//
// Asura_PS3_SPU_Skinning.cpp
//
// Author: Ben Weston 28/5/2008
//
// TEMPORARY! Really straightforward interface for running skinning code on SPUs
// Todo: use this and whatever Kevin does & Morgan's GeomCulling stuff as a basis for designing some kind of SPU task manager.
//


/* The model being followed here is:

spu DMA								|			spu processor
write processed vertices			|	process vertices
read next vertices					|

Since we know how many verts an SPU can take, we create a list of tasks and stick them in a queue.
But that's a pretty laborious way to do something we can do in the SPUs.
So ideally it should work like this:

data: static block of character data
task: process this ^ (pointer to that data)
spu: request next N vertices - update the read-from pointer shared between all processes.
	setup output address - update write-to pointer to where the end of the outputted data will be.

*/

//
// Includes
//

#include "Core/Asura.h"

// This file's header
#include "Asura_PS3_SPU_Skinning.h"

// Asura
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_CommandConsole.h"

// Asura Platform
#include "Animation/Asura_Platform_Animation_RenderCache.h"
#include "PPU_SPU_Utils/Asura_PS3_SPUGroupController.h"

// Platform
#include <cell/gcm.h>
#include <cell/sync/lfqueue.h>
#include <spu_printf.h>


//
// Macros
//

#define NUM_SPU_SKINNING_THREADS 5

#if (!defined(_DEBUG) && NUM_SPU_SKINNING_THREADS < 4)
#error NUM_SPU_SKINNING_THREADS is low in non-debug build, did you mean to do that?
#endif


#define uMAX_SPU_SKINNING_TASKS 64


//
// Variables
//

Asura_ProducerConsumerQueue		Asura_PS3_SPU_Skinning::s_xSPUKickerThread( 1, 1 );

u_int							Asura_PS3_SPU_Skinning::s_uWriteBuffer = 0;
int								Asura_PS3_SPU_Skinning::s_iRSXBufferReads = 0;
int								Asura_PS3_SPU_Skinning::s_iRSXReadyValue = 0;
bool							Asura_PS3_SPU_Skinning::s_bFirstCharacter = true;
bool							Asura_PS3_SPU_Skinning::s_bRanLastFrame = false;
bool							Asura_PS3_SPU_Skinning::s_bRanThisFrame = false;
volatile bool					Asura_PS3_SPU_Skinning::s_bIsRunning = false;

extern u_int _binary_Skinning_SPU_elf_start;

static Asura_PS3_SPUGroupController*	g_pxSPUThreadGroupInfo = NULL;

static const Asura_PS3_SPU_SkinningJob	g_xTerminatorJob(0);

static Asura_Mutex	g_xSpuLaunchMutex;

struct SPUKickData {
	u_int	*m_puRsxWaitLabel;
	u_int	m_uRSXReadyValue;
};
static SPUKickData g_xKickData;

static CellSyncBarrier g_xBarrier;


static bool g_bDoSpuSkinning = true;
static bool g_bHighlightSpuSkins = false;


//
// Functions
//


void Asura_PS3_SPU_Skinning::Shutdown( void )
{
	g_pxSPUThreadGroupInfo->End(); // just in case
	
	delete g_pxSPUThreadGroupInfo;
}

void Asura_PS3_SPU_Skinning::Initialise( void )
{
	g_pxSPUThreadGroupInfo = new Asura_PS3_SPUGroupController( "Skinning_SPU",
																&_binary_Skinning_SPU_elf_start,
																NUM_SPU_SKINNING_THREADS,
																Asura_PS3_SPUGroupController::PRIORITY_HIGH,
																sizeof(Asura_PS3_SPU_SkinningJob), uMAX_SPU_SKINNING_TASKS );

	g_pxSPUThreadGroupInfo->SetTerminatorJob( &g_xTerminatorJob );
 	
	s_uWriteBuffer = 0;
	s_iRSXBufferReads = 0;
	s_iRSXReadyValue = 0;

	s_bIsRunning = false;
	s_bRanLastFrame = false;
	s_bRanThisFrame = false;
	
	
 #if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
 	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.Enable", &g_bDoSpuSkinning, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.Highlight", &g_bHighlightSpuSkins, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
 #endif

	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_0) = 0;
	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_1) = 0;
	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX) = 0;
	
	cellSyncBarrierInitialize(&g_xBarrier, NUM_SPU_SKINNING_THREADS);
}


void Asura_PS3_SPU_Skinning::BeginFrame( void )
{
	s_uWriteBuffer = 1-s_uWriteBuffer;
	s_bFirstCharacter = true;
	
	//RSX read flag, set by RSX once it's finished reading each vertex buffer
	//Wait for the last read
	s_iRSXReadyValue = s_iRSXBufferReads;
	s_iRSXBufferReads = 0;

	s_bRanLastFrame = s_bRanThisFrame;
	s_bRanThisFrame = false;
}
	
	
// check that ProcessSkin will run correctly, but don't actually run it
//BW this function used to check lots of things because it locally handled caching of bones & output vertices, now it does very little.
bool Asura_PS3_SPU_Skinning::IsSpuable()
{
	if ( !g_bDoSpuSkinning )
	{
		return false;
	}

	return true;
}


inline size_t Asura_PS3_SPU_Skinning::GetVertexSize( SPU_SKINNING_MODE eMode )
{
	switch ( eMode )
	{
	default: // fall through
	case SPU_SKINNING_FULL_CULL:
		{
			return sizeof(Asura_PS3_SPU_Skinning_OutputVertex);
			break;
		}
	case SPU_SKINNING_POS_ONLY_NO_CULL:
		{
			return sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly);
			break;
		}
	}
}


//
// ProcessSkin
//
// Ben Weston June '08
//
// Add a character's skin to the SPU task list, and return a pointer to where the vertex data will be.
// This assumes you're about to add the draw strip commands, so this puts a wait label in the RSX command buffer, where it will wait for SPU skinning to complete.
//
// pOutputVertexBuffer - vertex buffer for the output vertices - for SPU to write to, and RSX to read from //BW todo: cull vertices so we can have less [+implement a cap so don't need to alloc the full space]
//						 at present this should be the same size as the input vertices.
//
bool Asura_PS3_SPU_Skinning::ProcessSkin(	const Asura_PS3_CharacterRenderer_Skin* pxCharacter,
											Asura_PS3_SPU_SkinningConstants* pxShaderConstants,
											Asura_Platform_Animation_RenderCache* pxRenderCache,
											u_int uBuffer, u_int uSkin,
											u_int uNumVerts,
											SPU_SKINNING_MODE eMode )
{
	Asura_PS3_SPU_SkinningJob xJob;
	
	if ( !IsSpuable() )
	{
		Asura_Assert(0,"ProcessSkin: IsSpuable() is false, this function should not be called!");
		return false;
	}

	if ( s_bFirstCharacter && s_bRanLastFrame )
	{
		WaitForSPUs();
		//BW do this as late as possible, but before any of the input data changes for this frame.
		//BW actually this might be too late, since the first character's bone cache has already been updated. But provided SPUs have started by now that won't cause a problem
	}


	//
	// Sync with RSX
	//

	// ensure RSX doesn't draw until SPUs have finished (this is unlikely to happen)
	//BW I was trying to make it sync per task but that's not practical, plus the SPUs should be finished well before the RSX starts, so just wait for the last task to finish.
	u_int uLabel = s_uWriteBuffer?ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_1:ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_0; //double buffer since our SPUs start running (and possibly finish) 1 frame before RSX reads it
	u_int* pLabel = cellGcmGetLabelAddress(uLabel);

	if ( s_bFirstCharacter )
	{
 		*pLabel = 0; // init to something other than uSPU_SKINNING_RSX_FLAG

		//make RSX wait for SPU to complete before it draws the character (don't do this in BeginFrame, in case we don't draw any chars!)
		cellGcmSetWaitLabelInline(Asura_PS3_Graphics::GetGcmCurrentContext(),uLabel,uSPU_SKINNING_RSX_FLAG);
	}

	//
	// Start the SPUS (asynchronously, once RSX finishes with last frame's data)
	//

	if ( s_bFirstCharacter )
	{
		KickSPUs( pLabel );
	}

	// Get the vertex output buffer
	Asura_PS3_GCM_VertexBuffer* pOutputVertexBuffer = pxRenderCache->ObtainVertexBuffer(uBuffer,uSkin,uNumVerts,GetVertexSize(eMode));


	//Sanity check: make sure vertex buffer is in RSX-mapped memory (otherwise we can't render from it)
	//Note we map all of main memory to the RSX, but on devkits it's possible to exceed the 256MB limit of mappable memory.
	u_int uTestOffset;
	char* pOutputBuffer = reinterpret_cast<char*>(pOutputVertexBuffer->LockForWriting());
	Asura_Assert( cellGcmAddressToOffset(pOutputBuffer, &uTestOffset) == CELL_OK, "Asura_PS3_SPU_Skinning::ProcessSkin - FATAL ERROR! Vertex buffer is not in memory mapped to RSX." );
	Asura_Assert( cellGcmAddressToOffset(pOutputBuffer + uNumVerts*GetVertexSize(eMode), &uTestOffset) == CELL_OK, "Asura_PS3_SPU_Skinning::ProcessSkin - FATAL ERROR! Vertex buffer is not in memory mapped to RSX." );
	
	
	// Create SPU Job
	xJob.m_uVertsOutAddr = reinterpret_cast<uint64_t>(pOutputVertexBuffer->LockForWriting()); //pointer to output vertices;

	if ( eMode == SPU_SKINNING_POS_ONLY_NO_CULL )
	{
		xJob.m_uInputAddr = reinterpret_cast<uint64_t>(pxCharacter->m_xVertexBuffer.GetReadPtr());
	}
	else
 	{
		int iNumStrips = pxCharacter->GetNumStrips();
		Asura_PS3_SPU_Skinning_Batch* pxBatches = pxRenderCache->GetBatches( uSkin, iNumStrips );
		
		const u_short* pusIndexBuffer = static_cast<const u_short*>(pxCharacter->m_xIndexBuffer.GetReadPtr());
		const Asura_PS3_SPU_Skinning_InputVertex* pxVertexBuffer = static_cast<const Asura_PS3_SPU_Skinning_InputVertex*>(pxCharacter->m_xVertexBuffer.GetReadPtr());
		
		for ( int iStrip = 0; iStrip < iNumStrips; iStrip++ )
		{
			Asura_PS3_SPU_Skinning_Batch& xBatch = pxBatches[iStrip];
			
			const Asura_PS3_CharacterRenderer_Strip& xStrip = pxCharacter->GetStrip( iStrip );
		
			xBatch.m_uStartIndexAddr = reinterpret_cast<uint64_t>(&pusIndexBuffer[xStrip.m_uStartIndex]);
			xBatch.m_uNumIndices = xStrip.m_uNumIndices;
			
			// assume the first vertex referenced by the strip is the 1st one we need (this is how the tools setup the data).
			u_int uFirstVert = pusIndexBuffer[xStrip.m_uStartIndex];
			xBatch.m_uVertAddr = reinterpret_cast<uint64_t>(&pxVertexBuffer[uFirstVert]);
			
			if ( iStrip < iNumStrips-1 )
			{
				u_int uNextFirstVert = pusIndexBuffer[ xStrip.m_uStartIndex + xStrip.m_uNumIndices ]; //the first vertex of the next strip
				xBatch.m_uNumVerts = uNextFirstVert - uFirstVert;
			}
			else
			{
				xBatch.m_uNumVerts = uNumVerts - uFirstVert;
			}
//			xBatch.m_uCommand = um;


			// sanity check the assumption about how strips are arranged (that the first vert indexed will be the first vert in the vert buffer)
			// n.b. if these fire someone's probably messed with the tools code (which sorts vertex data by first use, hence this assumption will always be true)
			for ( u_int i=xStrip.m_uStartIndex; i < xStrip.m_uStartIndex+xStrip.m_uNumIndices; i++ )
			{
				Asura_Assertf(pusIndexBuffer[i] >= uFirstVert,("Vertex needed by strip is not in batch sent to SPU! Culling will go mad\n%d >= %d\nbatch %d index %d localindex %d",
																pusIndexBuffer[i], uFirstVert, iStrip, i, i - xStrip.m_uStartIndex ));
				Asura_Assertf(pusIndexBuffer[i] - uFirstVert < xBatch.m_uNumVerts,("Vertex needed by strip is not in batch sent to SPU! Culling will go mad\n%d-%d < %d\nbatch %d index %d localindex %d",
																						pusIndexBuffer[i], uFirstVert, xBatch.m_uNumVerts, iStrip, i, i - xStrip.m_uStartIndex ));
			}
		}
		
		pxBatches[iNumStrips].m_uNumVerts = 0;
		pxBatches[iNumStrips].m_uNumIndices = 0;

		Asura_PS3_GCM_IndexBuffer* pOutputIndexBuffer = pxRenderCache->ObtainIndexBuffer(uSkin,pxCharacter->GetIndexBuffer()->GetIndexBufferSize()/2);
		
		xJob.m_uInputAddr = reinterpret_cast<uint64_t>(pxBatches);
		xJob.m_uIndicesOutAddr = reinterpret_cast<uint64_t>(pOutputIndexBuffer->LockForWriting());
	}
	
	xJob.m_eMode = eMode;


	// Shader constants
	Asura_Mem::Copy( &xJob.m_xShaderConstants, pxShaderConstants, sizeof(Asura_PS3_SPU_SkinningConstants) );


	if ( g_bHighlightSpuSkins )
	{
		// mess with the matrix to make SPU skins look different
		xJob.m_xShaderConstants.m_uBoneArrayAddr += 64;
	}

	//write this last to avoid SPUs ever getting nonsense data (though if they're reading at this point some of them might think they're finished when some others keep running)
	xJob.m_uNumVerts = uNumVerts;


	cellSyncLFQueuePush( g_pxSPUThreadGroupInfo->GetQueue(), &xJob );

	s_bFirstCharacter = false;
	
	return true;
}

void Asura_PS3_SPU_Skinning::EndFrame( void )
{
	if ( s_bFirstCharacter ) //if we didn't draw anything (slightly naff way to check this)
	{
		if ( s_bRanLastFrame )
		{
			WaitForSPUs(); //but we did draw on the previous frame, make sure SPUs finish that (previous) frame's work
		}

		return;
	}

	g_pxSPUThreadGroupInfo->PushTerminators(); // spus will take up time until this is done. - so do it as soon as we know there's no more characters
}

//
// Asura_PS3_SPU_Skinning::NotifyBufferUsed
//
// Call this whenever you've added render calls that read from a buffer filled by Asura_PS3_SPU_Skinning::ProcessSkin
// the SPU skinning uses this for synchronisation
//
void Asura_PS3_SPU_Skinning::NotifyBufferUsed( void )
{
	s_iRSXBufferReads++;

	cellGcmSetWriteTextureLabelInline(
		Asura_PS3_Graphics::GetGcmCurrentContext(),
		ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX, //note buffer ID's flipped, since we're making next-frame's SPUs wait for this
		s_iRSXBufferReads); //this label increments each time any of our vertex buffers is used for rendering (since we don't know which one's used last until the end of the frame), then we can just wait for the last usage
}

void Asura_PS3_SPU_Skinning::KickSPUs( u_int* pLabel )
{
	s_bRanThisFrame = true;

	//kick the SPUs, after the RSX has finished rendering the skinned data from the previous frame
	g_xKickData.m_puRsxWaitLabel = pLabel;
	g_xKickData.m_uRSXReadyValue = s_iRSXReadyValue; // need to pass this like this, since the val could change before the async kick
	s_xSPUKickerThread.AddToQueue(KickSPUsAsync, &g_xKickData, sizeof(g_xKickData));
}

// doing this in it's own thread is a bit of a waste, but we have to wait for RSX to finish reading from the vertex buffers
void Asura_PS3_SPU_Skinning::KickSPUsAsync( void* pxData, u_int uSize )
{
	Asura_Assert(uSize == sizeof(SPUKickData), "WTF?");
	SPUKickData* pxSPUKickData = reinterpret_cast<SPUKickData*>(pxData);

	// wait for the RSX to finish reading last frame's data
	u_int* pLabel = cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX);
	if ( pxSPUKickData->m_uRSXReadyValue > 0 ) // if we drew anything from our vertex buffers last time
	{
		Asura_TickCount xStartTicks = Asura_Timers::GetCurTicks();
	
		while ( *pLabel != pxSPUKickData->m_uRSXReadyValue
				)//&& Asura_Timers::GetElapsedTicks(xStartTicks) < Asura_Timers::SecondsToTicks(0.1f) )
		{
			Asura_Thread_System::YieldCurrentThread();
		}
		
		if ( *pLabel != pxSPUKickData->m_uRSXReadyValue )
		{
			Asura_OutputToDebugger_Retail(("KickSPUsAsync - wait for RSX timed out.\n"));
			return;
		}
	}

	// start the spus
	g_xSpuLaunchMutex.Lock();
	
	// clear it for next time
	*pLabel = 0xfeedbac2; // so we can spot it when debugging. N.B. this must be outside the range [0,s_iRSXBufferReads]

	Asura_Assert(!s_bIsRunning,"SPUs already running"); // sanity check (could be async thread issue)
	
	uint64_t uLabelAddr = reinterpret_cast<uint64_t>(pxSPUKickData->m_puRsxWaitLabel);
	uint64_t uBarrierAddr = reinterpret_cast<uint64_t>(&g_xBarrier);
	
	g_pxSPUThreadGroupInfo->Begin(
		NULL, // will be replaced with SPU num
		NULL, // will be replaced with pointer to queue
		&uLabelAddr,
		&uBarrierAddr
		);

	s_bIsRunning = true;

	g_xSpuLaunchMutex.Unlock();
}


void Asura_PS3_SPU_Skinning::WaitForSPUs()
{
	// make sure SPUs have actually started!
	Asura_TickCount xStartTicks = Asura_Timers::GetCurTicks();

	// wait til RSX has finished reading previous frame's SPU output
	bool bIsRunning;

	g_xSpuLaunchMutex.Lock();
	bIsRunning = s_bIsRunning;
	g_xSpuLaunchMutex.Unlock();

	while ( !bIsRunning
			&& Asura_Timers::GetElapsedTicks(xStartTicks) < Asura_Timers::SecondsToTicks(1.0f) )
	{
		Asura_Thread_System::YieldCurrentThread();

		g_xSpuLaunchMutex.Lock();
		bIsRunning = s_bIsRunning;
		g_xSpuLaunchMutex.Unlock();
	}

	if ( !bIsRunning )
	{
		Asura_OutputToDebugger_Retail(("WaitForSPUs - SPU kicker thread seems to be stuck.\n"));
		//ASURA_PS3_DEBUGBREAK();
		// this would happen if the kicker thread is stuck waiting for the RSX, which is waiting for the SPU
		// but the only time I've seen this has been when I've erroneously called this function when SPUs weren't meant to be running (i.e. wait for something that's not happening!)

		return;
	}


	// wait for SPUs to finish
	g_pxSPUThreadGroupInfo->End();

	g_xSpuLaunchMutex.Lock();
	s_bIsRunning = false;
	g_xSpuLaunchMutex.Unlock();
}

//BW
// The async kick thread is not very effective, it doesn't run unless the main thread yields, which means generally the SPUs aren't kicked until we WaitForSPUs()
// => after each character is added we could do a "try kick"
// then end frame does force kick (since doesn't matter if we wait for rsx at end of frame)
// and WaitForSPUs() doesn't wait for them to start (which will no-longer be asynchronous), just for them to end.
