//
// Asura_Platform_Animation_RenderCache.cpp
//
// Ben Weston - 13/06/2008
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Animation/Asura_Platform_Animation_RenderCache.h"

// Asura
#include "Animation/Asura_Animation.h"
#include "Core/Memory/Asura_Mem.h"
#include "Rendering/Asura_Render.h"

// Asura Platform
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_GCM_IndexBuffer.h"
#include "Rendering/Asura_PS3_SPU_SkinningTypes.h"
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"



Asura_Platform_Animation_RenderCache*		Asura_Platform_Animation_RenderCache::s_apxCleanUp[2] = { NULL, NULL };
Asura_Platform_Animation_RenderCache*		Asura_Platform_Animation_RenderCache::s_pxActive = NULL;
u_int										Asura_Platform_Animation_RenderCache::s_uCleanupBuffer = 0;



//
// Static functions, for globally managing clean-up of render caches
//

Asura_Platform_Animation_RenderCache* Asura_Platform_Animation_RenderCache::Create()
{
	Asura_Platform_Animation_RenderCache* pxNew = new Asura_Platform_Animation_RenderCache;
	
	Asura_Assert(pxNew,"Asura_Platform_Animation_RenderCache: failed to allocate new rendercache object");
	
	if ( pxNew )
	{
		// add to active list
		pxNew->m_pxNext = s_pxActive;
		s_pxActive = pxNew;
		
		// note: do not allocate caches here, wait until they're needed
	}
	
	return pxNew;
}


void Asura_Platform_Animation_RenderCache::Cleanup()
{
	//called per-frame, clean up all render caches added to the cleanup list 2-frames ago
	
	s_uCleanupBuffer = 1-s_uCleanupBuffer;
	
	while ( s_apxCleanUp[s_uCleanupBuffer] )
	{
		Asura_Platform_Animation_RenderCache* pxThis = s_apxCleanUp[s_uCleanupBuffer];
		s_apxCleanUp[s_uCleanupBuffer] = pxThis->m_pxNext;
		delete pxThis; //call the destructor & free the memory
	}
}

void Asura_Platform_Animation_RenderCache::Initialise()
{
	s_apxCleanUp[0] = NULL;
	s_apxCleanUp[1] = NULL;
	s_pxActive = NULL;
	s_uCleanupBuffer = 0;
}

void Asura_Platform_Animation_RenderCache::Shutdown()
{
	Cleanup();
	Cleanup();

	s_apxCleanUp[0] = NULL;
	s_apxCleanUp[1] = NULL;
	s_pxActive = NULL;
	s_uCleanupBuffer = 0;
}


//
// Methods
//

Asura_Platform_Animation_RenderCache::Asura_Platform_Animation_RenderCache()
{
	for ( int j=0; j < ASURA_ANIMATION_SKIN_COUNT; j++ )
	{
		for ( int i=0; i < NUM_ASURA_PLATFORM_ANIMATION_RENDERCACHE_VERTEXBUFFERS; i++ )
		{
				m_apxVertexBuffer[i][j] = NULL;
				m_auVertexFrameStamp[i][j] = 0;
		}

		m_apxIndexBuffer[j] = NULL;
		m_apxBatches[j] = NULL;
	}
	m_pxBoneCache = NULL;
	m_uNumBonesAlloced = 0;
	m_uBoneFrameStamp = 0;
}

Asura_Platform_Animation_RenderCache::~Asura_Platform_Animation_RenderCache()
{
	for ( int j=0; j < ASURA_ANIMATION_SKIN_COUNT; j++ )
	{
		for ( int i=0; i < NUM_ASURA_PLATFORM_ANIMATION_RENDERCACHE_VERTEXBUFFERS; i++ )
		{
			if ( m_apxVertexBuffer[i][j] != NULL )
			{
				m_apxVertexBuffer[i][j]->Destroy();
				delete m_apxVertexBuffer[i][j];
				m_apxVertexBuffer[i][j] = NULL;
			}

			m_auVertexFrameStamp[i][j] = 0;
		}
		
		if ( m_apxIndexBuffer[j] )
		{
			m_apxIndexBuffer[j]->Shutdown();
			delete m_apxIndexBuffer[j];
			m_apxIndexBuffer[j] = NULL;
		}

		if ( m_apxBatches[j] != NULL )
		{
			delete[] m_apxBatches[j];
		}
	}

	if ( m_pxBoneCache != NULL )
	{
		Asura_DeallocateAlign(m_pxBoneCache);
		m_pxBoneCache = NULL;
	}

	m_uNumBonesAlloced = 0;
	m_uBoneFrameStamp = 0;
}


// Queue this cache for destruction after the RSX has finished reading from it.
void Asura_Platform_Animation_RenderCache::Destroy()
{
	// remove from the active list
	if ( s_pxActive == this )
	{
		s_pxActive = m_pxNext;
	}
	else
	{
		Asura_Platform_Animation_RenderCache* pxSearch = s_pxActive;
		while ( pxSearch &&  pxSearch->m_pxNext != this )
		{
			pxSearch = pxSearch->m_pxNext;
		}
		
		Asura_Assert( pxSearch, "Asura_Platform_Animation_RenderCache::Destroy - Trying to destroy non-active RenderCache, this should be impossible.\nAre you sure this hasn't already been deleted?" );
		if ( !pxSearch ) return;
		
		pxSearch->m_pxNext = m_pxNext;
	}
	
	//BW - optimisation - could free here if I know it wasn't drawn from last frame.
	
	// add to the cleanup list
	m_pxNext = s_apxCleanUp[s_uCleanupBuffer];
	s_apxCleanUp[s_uCleanupBuffer] = this;
}

Asura_AMatrix4x4* Asura_Platform_Animation_RenderCache::UpdateCachedBones( const Asura_AMatrix4x4* pxBones, u_int uNumBones )
{
	Asura_Assert( !IsBoneCacheCurrent(), "UpdateCachedBones - updating bone cache twice in the same frame for the same animation, this is wasteful" );

	// first check our cache exists & is big enough
	if ( uNumBones != m_uNumBonesAlloced )
	{
		if ( m_pxBoneCache != NULL )
		{
			Asura_Assert( 0, "Asura_Platform_Animation_RenderCache::UpdateCachedBones - Number of bones in animated skin has changed, this will cause a realloc & is probably a bad thing." );
			
			Asura_DeallocateAlign( m_pxBoneCache );
		}
		
		// alloc the space needed
		m_pxBoneCache = reinterpret_cast<Asura_AMatrix4x4*>(Asura_AllocateAlign( uNumBones * sizeof(Asura_AMatrix4x4), ASURA_ALIGN_16_BYTE ));

		m_uNumBonesAlloced = uNumBones;
	}
	
	// cache the bones
	Asura_Mem::Copy( m_pxBoneCache, pxBones, uNumBones * sizeof(Asura_AMatrix4x4) );

	SetBoneCacheCurrent();
	
	return m_pxBoneCache;
}

Asura_PS3_GCM_VertexBuffer* Asura_Platform_Animation_RenderCache::ObtainVertexBuffer( u_int uBuffer, u_int uSkin, u_int uNumVerts, size_t uVertexSize )
{
	Asura_PS3_GCM_VertexBuffer* pVertexBuffer = m_apxVertexBuffer[uBuffer][uSkin];
	
	// alloc if necessary
	if ( pVertexBuffer == NULL || uNumVerts > pVertexBuffer->GetNumVerts() || uVertexSize != pVertexBuffer->GetSize() )
	{
		if ( pVertexBuffer == NULL )
		{
			pVertexBuffer = new Asura_PS3_GCM_VertexBuffer;
		}
		else
		{
			Asura_Assertf( uVertexSize == pVertexBuffer->GetSize(), ("Asura_Platform_Animation_RenderCache::ObtainVertexBuffer - Cached vertex size in animated skin has changed (buffer %d, skin %d), this will cause a realloc & is probably a bad thing.", uBuffer, uSkin));
			Asura_Assertf( uNumVerts <= pVertexBuffer->GetNumVerts(), ("Asura_Platform_Animation_RenderCache::ObtainVertexBuffer - Cached number of verts in animated skin has increased (buffer %d, skin %d), this will cause a realloc & is probably a bad thing.", uBuffer, uSkin));
			//BW if the above assert fires, try always calling this with the same num verts
			// i.e. if you're changing the LOD of the skin on this animation: always call this with the numverts of the highest LOD skin, should make life simpler.

			pVertexBuffer->Destroy();
		}

		//create a buffer sized for this number of verts
		pVertexBuffer->Initialise( uVertexSize, uNumVerts, CELL_GCM_LOCATION_MAIN, true );
		
		m_apxVertexBuffer[uBuffer][uSkin] = pVertexBuffer;
	}

	return pVertexBuffer;
}


Asura_PS3_GCM_IndexBuffer* Asura_Platform_Animation_RenderCache::ObtainIndexBuffer( u_int uSkin, u_int uNumIndices )
{
	Asura_PS3_GCM_IndexBuffer* pIndexBuffer = m_apxIndexBuffer[uSkin];

	// alloc if necessary
	if ( pIndexBuffer == NULL || uNumIndices > pIndexBuffer->GetIndexBufferSize()/2 )
	{
		if ( pIndexBuffer == NULL )
		{
			pIndexBuffer = new Asura_PS3_GCM_IndexBuffer;
		}
		else
		{
			Asura_Assertf( uNumIndices <= pIndexBuffer->GetIndexBufferSize()/2, ("Asura_Platform_Animation_RenderCache::ObtainIndexBuffer - Cached number of verts in animated skin has increased (skin %d), this will cause a realloc & is probably a bad thing.", uSkin));
			//BW if the above assert fires, try always calling this with the same num verts
			// i.e. if you're changing the LOD of the skin on this animation: always call this with the numverts of the highest LOD skin, should make life simpler.

			pIndexBuffer->Shutdown();
		}

		//create a buffer sized for this number of verts
		pIndexBuffer->Initialise( uNumIndices, CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16, CELL_GCM_LOCATION_MAIN, true );

		m_apxIndexBuffer[uSkin] = pIndexBuffer;
	}

	return pIndexBuffer;
}


Asura_PS3_SPU_Skinning_Batch* Asura_Platform_Animation_RenderCache::GetBatches( u_int uSkin, u_int uNumStrips )
{
	if ( !m_apxBatches[uSkin] )
	{
		m_apxBatches[uSkin] = new Asura_PS3_SPU_Skinning_Batch[uNumStrips+1];
	}
	return m_apxBatches[uSkin];
}
