//////////////////////////////////////////////////////////////////////////////
//
// Asura_AI_Navigation.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Created by Kevin Lea - 24/07/2001 12:23:00
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_DISABLE_AI_CODE
#else
// header
#include "AI/Asura_AI_Navigation.h"

// asura
#include "AI/Asura_Nav_Types.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_AABB_Tree.h"
#include "Core/Asura_OrientedVolume.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Module.h"

#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Debugging/Asura_DebugLineRender.h"
#include "Interface/Asura_CommandConsole.h"
#endif

#ifdef ASURA_TOOLS
#include "Lighting/Tools_Environment_LOS.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#endif

// Axon
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Core/Axon_System.h"
#endif
//////////////////////////////////////////////////////////////////////////////
//
// Asura_AI_NavVertex
//
//////////////////////////////////////////////////////////////////////////////

enum { NUM_NAV_VERTEX_NORMALS = 255 };
// Forward declaration. The actual table is at the bottom of this file so it doesn't clutter the top.
extern const Asura_Vector_3 s_axNormalTable[ NUM_NAV_VERTEX_NORMALS ];

const Asura_Vector_3& Asura_AI_NavVertex::GetNormal() const
{
	return s_axNormalTable[ m_ucNormalIndex ];
}

Asura_AI_NavVertex* Asura_AI_NavVertex::GetLinkedWaypoint( const u_int uLinkIndex ) const
{
	return Asura_AI_Navigation_System::GetWaypoint( m_pxConnections[ uLinkIndex ].GetIndex() );
}

void Asura_AI_NavVertex::SetNormal( const Asura_Vector_3& xNormal )
{
    // Work out which normal in the table most closely corresponds to the ideal normal.
    float fBestDotProduct = -1.0f;
    float fDotProduct     = -1.0f;
    for ( u_char uc = 0; uc < NUM_NAV_VERTEX_NORMALS; ++uc )
    {
        fDotProduct = xNormal * s_axNormalTable[ uc ];
        if ( fDotProduct > fBestDotProduct )
        {
            fBestDotProduct = fDotProduct;
            m_ucNormalIndex = uc;
        }
    }
}

bool Asura_AI_NavVertex::Contains( const Asura_Vector_3& xPosition, float* pfDistanceSq /*= NULL*/ ) const
{
    // These are hemi-spheres.
    Asura_Vector_3 xToPosition      = xPosition - m_xPosition;
    const float    fToPositionMagSq = xToPosition.MagnitudeSquared();
	if ( pfDistanceSq ) { *pfDistanceSq = fToPositionMagSq; }
    const bool     bInsideRadius    = ( fToPositionMagSq < ( GetRadius() * GetRadius() ) );
    if ( !bInsideRadius ) { return false; }

	if ( fToPositionMagSq == 0.0f )
	{
		return true;
	}
	else
	{
		// Check for the angle between the vector to this position and the waypoint normal being small enough.
		// We allow up to 95 to account for surface undulations.
		// Normalise this vector first.
		xToPosition *= ( 1.0f / Asura_Maths::Sqrt( fToPositionMagSq ) );
		const bool bCorrectSide = ( ( xToPosition * GetNormal() ) > -0.0871557f );
		return bCorrectSide;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_AI_Navigation_System
//
//////////////////////////////////////////////////////////////////////////////

Asura_AI_Navigation_System*		  Asura_AI_Navigation_System::s_pxThis = NULL;
bool                              Asura_AI_Navigation_System::s_bWaypointsValid = false;
unsigned short                    Asura_AI_Navigation_System::s_usNumWaypoints = 0;
Asura_AI_NavVertex*               Asura_AI_Navigation_System::s_pxWaypointStorage = NULL;
unsigned int                      Asura_AI_Navigation_System::s_uNumConnections = 0;
Asura_AI_NavConnection*           Asura_AI_Navigation_System::s_pxNavConnectionStorage = NULL;
Asura_AI_PerVertexCoverpointData* Asura_AI_Navigation_System::s_pxPerVertexCoverpointData = NULL;
unsigned int                      Asura_AI_Navigation_System::s_uNumCoverpointIndices = 0;
unsigned short*                   Asura_AI_Navigation_System::s_pusCoverpointIndexStorage = NULL;
int                               Asura_AI_Navigation_System::s_iNumCoverpoints = 0;
Asura_AI_Coverpoint*              Asura_AI_Navigation_System::s_pxCoverpointStorage = NULL;
Asura_AABB_Tree*                  Asura_AI_Navigation_System::s_pxWaypointTree = NULL;
unsigned char*                    Asura_AI_Navigation_System::s_pucWaypointWeights = NULL;
float                             Asura_AI_Navigation_System::s_fWaypointWeightFactor = 5.0f;

const Asura_Vector_3*           Asura_AI_Navigation_System::s_pxWaypointCallbackTestPosition         = NULL;
float                           Asura_AI_Navigation_System::s_fWaypointCallbackMinDistance           = 0.0f;
u_short							Asura_AI_Navigation_System::s_usWaypointCallbackWaypointZone         = 0;
bool                            Asura_AI_Navigation_System::s_bWaypointCallbackShouldContainPosition = false;
bool                            Asura_AI_Navigation_System::s_bWaypointCallbackShouldSkipConnectionlessWaypoints = true;
int                             Asura_AI_Navigation_System::s_iWaypointCallbackWaypointID            = 0;

Asura_AStar*				Asura_AI_Navigation_System::s_pxAStarList = NULL;
Asura_OpenClosed*           Asura_AI_Navigation_System::s_pxAStarOpenClosedVector = NULL;
Asura_AStar_OpenList_Elem*  Asura_AI_Navigation_System::s_pxAStarOpenList = NULL;
int                         Asura_AI_Navigation_System::s_iNumElemsInOpenList = 0;

bool                        Asura_AI_Navigation_System::s_bRouteTargetValid = false;
unsigned short              Asura_AI_Navigation_System::s_usRouteTarget = 0;
float                       Asura_AI_Navigation_System::s_fPathfindRange = 0.0f;
Asura_Vector_3              Asura_AI_Navigation_System::s_xPathfindPosition = Asura_Maths::ZeroVector3;
bool                        Asura_AI_Navigation_System::s_bMidAStarIncremental = false;

bool                        Asura_AI_Navigation_System::s_bConnectionChangedThisFrame = false;
bool                        Asura_AI_Navigation_System::s_bConnectionChangedLastFrame = false;

void (*Asura_AI_Navigation_System::s_pfnBlockCallback)(u_int uWaypointIndex, u_int uConnectionIndex, const bool bBlock) = NULL;

u_int						Asura_AI_Navigation_System::s_uConnectionGeneration = CONNECTION_GENERATION_FIRST;
float						Asura_AI_Navigation_System::s_fMaxWaypointRadius = 0.0f;

#ifdef ASURA_TOOLS
Asura_LinkList< Asura_AI_Navigation_System::UserDefinedWaypoint* > Asura_AI_Navigation_System::s_xUDWPList;
//Asura_Vector_3* Asura_AI_Navigation_System::s_pxUserDefinedWaypointPositions = NULL;
//Asura_Quat*     Asura_AI_Navigation_System::s_pxUserDefinedWaypointOrientations = NULL;
//unsigned int*	Asura_AI_Navigation_System::s_puUserDefinedWaypointFlags = NULL;
//unsigned int    Asura_AI_Navigation_System::s_uNumUserDefinedWaypoints = 0;
#endif

#ifndef ASURA_NO_DEBUG_OUTPUT
void Asura_AI_Navigation_System::RenderWaypoints( const u_int uColour /*=0xffffffff*/, const u_int uNumSegments /*=4*/, const u_int uNumSlices /*=4*/ )
{
	for ( u_short us = 0; us < s_usNumWaypoints; ++us )
	{
		if ( ( s_pxWaypointStorage[ us ].m_xPosition - Asura_Camera::GetCameraData( 0 ).m_xPosition ).MagnitudeSquared() > Axon_System::GetRenderNavRangeSQ() )
		{
			continue;
		}
		Asura_DebugLineRender::AddSphere( s_pxWaypointStorage[ us ].m_xPosition, s_pxWaypointStorage[ us ].GetRadius(), uColour, uNumSegments, uNumSlices );
	}
}

void Asura_AI_Navigation_System::RenderCoverpoints( const u_int uColour /*=0x6f00ffff*/, const u_int uNumSegments /*=4*/, const u_int uNumSlices /*=4*/ )
{
	for ( int i = 0; i < s_iNumCoverpoints; ++i )
	{
		Asura_DebugLineRender::AddSphere( s_pxCoverpointStorage[ i ].m_xCentre, 0.3f, uColour, uNumSegments, uNumSlices );
	}
}

void Asura_AI_Navigation_System::RenderLinks( const u_int uColour /*=0xffffffff*/ )
{
	for ( int i = 0; i < s_usNumWaypoints; ++i )
	{
		const Asura_AI_NavVertex& xBaseWP = s_pxWaypointStorage[ i ];

		if ( ( xBaseWP.m_xPosition - Asura_Camera::GetCameraData( 0 ).m_xPosition ).MagnitudeSquared() > Axon_System::GetRenderNavRangeSQ() )
		{
			continue;
		}

		for ( int c = 0; c < xBaseWP.m_ucNumConnections; ++c )
		{
			const Asura_AI_NavVertex& xTargetWP = s_pxWaypointStorage[ xBaseWP.m_pxConnections[ c ].GetIndex() ];
			Asura_DebugLineRender::AddLine( xBaseWP.m_xPosition, xTargetWP.m_xPosition, uColour );
		}
	}
}

void Asura_AI_Navigation_System::RenderBlockedLinks( const u_int uColour /*=0xffff0000*/ )
{
	for ( int i = 0; i < s_usNumWaypoints; ++i )
	{
		const Asura_AI_NavVertex& xBaseWP = s_pxWaypointStorage[ i ];

		for ( int c = 0; c < xBaseWP.m_ucNumConnections; ++c )
		{
			if ( !( xBaseWP.m_pxConnections[ c ].GetFlags() & ASURA_AI_CONNECTION_FLAGS_BLOCKED ) )
			{
				continue;
			}

			const Asura_AI_NavVertex& xTargetWP = s_pxWaypointStorage[ xBaseWP.m_pxConnections[ c ].GetIndex() ];
			Asura_DebugLineRender::AddLine( xBaseWP.m_xPosition, xTargetWP.m_xPosition, uColour );
		}
	}
}
#endif

bool Asura_AI_Navigation_System::WaypointTreeCreateCallback( u_int uNumWaypoints, Asura_Bounding_Box* pxBBs, Asura_Vector_3* )
{
    Asura_Bounding_Box xTempBB;
    Asura_Matrix_3x3   xOrientation;

	for ( unsigned short us = 0; us < s_usNumWaypoints; ++us )
	{
        const Asura_AI_NavVertex& xWP = s_pxWaypointStorage[ us ];

        const float fRadius = xWP.GetRadius();

        // This is consistent with the way this function worked before, but it doesn't take into account
        // undulations in the surface the waypoint is on.
        const Asura_Vector_3 xMin( -fRadius, -Asura_Maths::Min( 2.0f, fRadius ), -fRadius );
        const Asura_Vector_3 xMax(  fRadius, 0.1f, fRadius );

        xTempBB.Reset();
        xTempBB.Set( xMin );
        xTempBB.Set( xMax );

        Asura_Maths::OrientationFromNormal( xOrientation, xWP.GetNormal() );
        pxBBs[ us ].FromOBB( xTempBB, xOrientation );
        pxBBs[ us ].Displace( xWP.m_xPosition );
	}

	// Let tree creator class calculate centres for us.
	return false;
}

bool Asura_AI_Navigation_System::WaypointTreeCallback( unsigned short usIndex )
{
	if ( s_pxWaypointStorage[ usIndex ].m_ucNumConnections == 0 && s_bWaypointCallbackShouldSkipConnectionlessWaypoints )
    {
        return true;
    }

    // Skip waypoints not in the desired zone, if necessary.
    if ( s_usWaypointCallbackWaypointZone != ASURA_NAVVERTEX_MAX_ZONE_ID &&
        s_pxWaypointStorage[ usIndex ].m_usZone != s_usWaypointCallbackWaypointZone )
    {
        return true;
    }

	const Asura_Vector_3 xDisp = *s_pxWaypointCallbackTestPosition - s_pxWaypointStorage[ usIndex ].m_xPosition;
	const float          fDist = xDisp.Magnitude() - s_pxWaypointStorage[ usIndex ].GetRadius();

    // Skip waypoints that don't contain the position, if necessary.
    if ( s_bWaypointCallbackShouldContainPosition && fDist > 0.0f )
    {
        return true;
    }

	if ( fDist < s_fWaypointCallbackMinDistance )
	{
		s_fWaypointCallbackMinDistance = fDist;
		s_iWaypointCallbackWaypointID  = usIndex;
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_AI_Navigation_System
//
//////////////////////////////////////////////////////////////////////////////

Asura_AI_Navigation_System::Asura_AI_Navigation_System()
{
	s_bWaypointsValid					= false;
	s_usNumWaypoints					= 0;
	s_pxWaypointStorage					= NULL;
	s_uNumConnections					= 0;
	s_pxNavConnectionStorage			= NULL;
	s_pxPerVertexCoverpointData			= NULL;
	s_uNumCoverpointIndices				= 0;
	s_pusCoverpointIndexStorage			= NULL;
	s_iNumCoverpoints					= 0;
	s_pxCoverpointStorage				= NULL;
	s_pxWaypointTree					= NULL;
	s_pucWaypointWeights				= NULL;
	// this is how much each increment of a waypoint weight adds to the path cost
	s_fWaypointWeightFactor				= 5.0f;

	//-----------------------------------------------------------------------------------------------------------
	// Old route finding interface that is due for removal.
	s_pxAStarList						= NULL;
	s_pxAStarOpenClosedVector			= NULL;
	s_pxAStarOpenList					= NULL;
	s_iNumElemsInOpenList				= 0;

	s_bRouteTargetValid					= false;
	s_usRouteTarget						= 0;
	s_fPathfindRange					= 0.0f;
	s_xPathfindPosition					= Asura_Maths::ZeroVector3;
	s_bMidAStarIncremental				= false;
	// Old route finding interface that is due for removal.
	//-----------------------------------------------------------------------------------------------------------

	s_bConnectionChangedThisFrame		= false;
	s_bConnectionChangedLastFrame		= false;
	s_uConnectionGeneration				= CONNECTION_GENERATION_FIRST;

	s_fMaxWaypointRadius				= 0.0f;

	// Don't need to call initialise again, as this nulls everything we need to.
}

Asura_AI_Navigation_System::~Asura_AI_Navigation_System()
{
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Asura_AI_Navigation_System::Initialise()
{
	s_pxThis = Project_CreateSystem();

	Reset();

	#ifdef ASURA_USE_NEW_NAV_STUFF
	Asura_Navigation::Initialise();
	#endif
}

void Asura_AI_Navigation_System::Shutdown()
{
	#ifdef ASURA_USE_NEW_NAV_STUFF
	Asura_Navigation::Shutdown();
	#endif

	Reset();

	delete s_pxThis;
	s_pxThis = NULL;
}

void Asura_AI_Navigation_System::Reset()
{
#ifdef ASURA_USE_NEW_NAV_STUFF
	Asura_Navigation::AStarData::ShudownAStarStorage( s_usNumWaypoints );
#endif

	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
	{
		Asura_ResourceDeallocateArray(s_pxWaypointStorage);
		Asura_ResourceDeallocateArray(s_pxNavConnectionStorage);
		Asura_ResourceDeallocateArray(s_pxPerVertexCoverpointData);
		Asura_ResourceDeallocateArray(s_pusCoverpointIndexStorage);
		Asura_ResourceDeallocateArray(s_pxCoverpointStorage);
		delete s_pxWaypointTree;
		Asura_ResourceDeallocateArray(s_pucWaypointWeights);

		//-----------------------------------------------------------------------------------------------------------
		// Old route finding interface that is due for removal.
		Asura_ResourceDeallocateArray(s_pxAStarList);
		Asura_ResourceDeallocateArray(s_pxAStarOpenClosedVector);
		Asura_ResourceDeallocateArray(s_pxAStarOpenList);
		// Old route finding interface that is due for removal.
		//-----------------------------------------------------------------------------------------------------------

		s_bWaypointsValid             = false;
		s_usNumWaypoints              = 0;
		s_pxWaypointStorage           = NULL;
		s_uNumConnections             = 0;
		s_pxNavConnectionStorage      = NULL;
		s_pxPerVertexCoverpointData   = NULL;
		s_uNumCoverpointIndices       = 0;
		s_pusCoverpointIndexStorage   = NULL;
		s_iNumCoverpoints             = 0;
		s_pxCoverpointStorage         = NULL;
		s_pxWaypointTree              = NULL;
		s_pucWaypointWeights          = NULL;

		//-----------------------------------------------------------------------------------------------------------
		// Old route finding interface that is due for removal.
		s_pxAStarList                 = NULL;
		s_pxAStarOpenClosedVector     = NULL;
		s_pxAStarOpenList             = NULL;
		s_iNumElemsInOpenList         = 0;

		s_bRouteTargetValid           = false;
		s_usRouteTarget               = 0;
		s_fPathfindRange              = 0.0f;
		s_xPathfindPosition           = Asura_Maths::ZeroVector3;
		// Old route finding interface that is due for removal.
		//-----------------------------------------------------------------------------------------------------------

		s_bConnectionChangedThisFrame = false;
		s_bConnectionChangedLastFrame = false;
		s_uConnectionGeneration		  = CONNECTION_GENERATION_FIRST;

		#ifdef ASURA_TOOLS
		for ( Asura_LinkList_It< UserDefinedWaypoint* > xUDWIter( s_xUDWPList ); !xUDWIter.Done(); xUDWIter.Next() )
		{
			GetInstance()->DestroyUserDefinedWaypoint( *xUDWIter.GetCurrent() );
		}
		s_xUDWPList.Clear();
		#endif
	}
	Asura_ResourcePopType();
}

void Asura_AI_Navigation_System::Update()
{
	if (s_bConnectionChangedThisFrame)
	{
		++s_uConnectionGeneration;
	}

	s_bConnectionChangedLastFrame = s_bConnectionChangedThisFrame;
	s_bConnectionChangedThisFrame = false;
}

void Asura_AI_Navigation_System::ReadUDWDataFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;

	// Ordinarily we'd chunk in the parent now.

	if ( uVersion == uCURRENT_CHUNK_VERSION )
	{
		u_int uNumUDWs;
		xStream >> uNumUDWs;
		for ( u_int u = 0; u < uNumUDWs; ++u )
		{
			UserDefinedWaypoint* pxUDW = CreateUserDefinedWaypoint();
			pxUDW->ReadFromChunkStream( xStream );

			// In tools we're interested in these.
			#ifdef ASURA_TOOLS
			s_xUDWPList.Push_Back( pxUDW );
			#else
			DestroyUserDefinedWaypoint( pxUDW );
			#endif
		}
	} 
	else
	{
		Asura_Assertf( false, ( "Asura_AI_Navigation_System::ReadFromChunkStream - bad chunk version (%u)", uVersion ) );
	}
}

#if ASURA_TOOLS
void Asura_AI_Navigation_System::WriteUDWDataToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << uCURRENT_CHUNK_VERSION;

	// Ordinarily we'd chunk out the parent now.

	xStream << s_xUDWPList.Size();
	for ( Asura_LinkList_ConstIt< UserDefinedWaypoint* > xUDWIter( s_xUDWPList ); !xUDWIter.Done(); xUDWIter.Next() )
	{
		( *xUDWIter.GetCurrent() )->WriteToChunkStream( xStream );
	}
}
#endif

//-----------------------------------------------------------------------------------------------------------
// Old route finding interface that is due for removal.

// (AAM - 03.09.2004) Xbox compiler optimiser produces corrupted code here...
//TODO: follow up with XDS
#if defined _XBOX
	#pragma optimize("", off)
#endif

void Asura_AI_Navigation_System::FindRoute( unsigned short* pusRoute, int& iLengthOfRoute, const int iBufferLength,
											const unsigned short usStart, const unsigned short usTarget,
											const unsigned uPathfinderFlags, void* pUserData )
{
	s_bRouteTargetValid = true;
	s_usRouteTarget     = usTarget;
	AStarSearch( pusRoute, iLengthOfRoute, iBufferLength,
				usStart, uPathfinderFlags, pUserData,
				Asura_AI_Navigation_System::DefaultGoalCostEstimateFunction,
				Asura_AI_Navigation_System::DefaultTraversalCostFunction,
				Asura_AI_Navigation_System::DefaultGoalAchievedFunction,
				Asura_AI_Navigation_System::DefaultLinkFlagsTestFunction );
}

void Asura_AI_Navigation_System::FindRouteCustom( unsigned short* pusRoute, int& iLengthOfRoute, const int iBufferLength,
												   const unsigned short usStart,
												   const unsigned uPathfinderFlags,
												   void* pUserData,
                                                   AANS_CostToGoalEstFn* pfnCostToGoalEstFn,
                                                   AANS_TraversalCostFn* pfnTraversalCostFn,
                                                   AANS_GoalAchievedFn*  pfnGoalAchievedFn,
                                                   AANS_LinkFlagsTestFn* pfnLinkFlagsTestFn )
{
	s_bRouteTargetValid = false;
	AStarSearch( pusRoute, iLengthOfRoute, iBufferLength,
				usStart, uPathfinderFlags, pUserData,
				pfnCostToGoalEstFn,
				pfnTraversalCostFn,
				pfnGoalAchievedFn,
				pfnLinkFlagsTestFn );
}

void Asura_AI_Navigation_System::FindRouteCustom( unsigned short* pusRoute, int& iLengthOfRoute, const int iBufferLength,
												   const unsigned short usStart,
                                                   const unsigned short usTarget,
                                                   const unsigned uPathfinderFlags,
												   void* pUserData,
                                                   AANS_CostToGoalEstFn* pfnCostToGoalEstFn,
                                                   AANS_TraversalCostFn* pfnTraversalCostFn,
                                                   AANS_GoalAchievedFn*  pfnGoalAchievedFn,
                                                   AANS_LinkFlagsTestFn* pfnLinkFlagsTestFn )
{
	s_bRouteTargetValid = true;
	s_usRouteTarget     = usTarget;
	AStarSearch( pusRoute, iLengthOfRoute, iBufferLength,
				usStart, uPathfinderFlags, pUserData,
				pfnCostToGoalEstFn,
				pfnTraversalCostFn,
				pfnGoalAchievedFn,
				pfnLinkFlagsTestFn );
}

float Asura_AI_Navigation_System::DefaultGoalCostEstimateFunction( const int iCurrentIndex, void* pUserData )
{
	const Asura_Vector_3 xDisp = s_pxWaypointStorage[ iCurrentIndex ].m_xPosition - s_pxWaypointStorage[ s_usRouteTarget ].m_xPosition;
	return xDisp.Magnitude();
}

float Asura_AI_Navigation_System::DefaultTraversalCostFunction( const int iWaypointIndex, const int iConnectionIndex, const u_int uPathfinderFlags, void* pUserData )
{
	//     cost of the link traversal + cost of the waypoint (if any)
	return s_pxWaypointStorage[ iWaypointIndex ].m_pxConnections[ iConnectionIndex ].GetLength() +
		s_pucWaypointWeights[ iWaypointIndex ] * s_fWaypointWeightFactor;
}

bool Asura_AI_Navigation_System::DefaultLinkFlagsTestFunction( const Asura_AI_NavConnection* pxLink, const unsigned int uPathfinderFlags, void* pUserData )
{
	const unsigned int uLinkFlags = pxLink->GetFlags();
	if ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_BLOCKED )
	{
		// ignore blocked waypoints
		return true;
	}
	const bool bCanFallFar = ( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_CANFALLFAR ) ? true : false;
	if ( ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_LONGFALL ) && !bCanFallFar )
	{
		// it's a long fall and we can't long fall - ignore
		return true;
	}
	const bool bCanFall = ( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_CANFALL ) ? true : false;
	if ( ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_FALL ) && !bCanFall )
	{
		// it's a fall link and we can't fall - ignore
		return true;
	}

	const bool bCanJump = ( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_CANJUMP ) ? true : false;
	if ( ( uLinkFlags & ASURA_AI_CONNECTION_FLAGS_JUMP ) && !bCanJump )
	{
		// it's a jump link and we can't jump - ignore
		return true;
	}

	return false;
}

#if defined _XBOX
	#pragma optimize("", on)
#endif
// Old route finding interface that is due for removal.
//-----------------------------------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------------------------------
// Old nav data query interface that is due for removal.
unsigned short Asura_AI_Navigation_System::FindWaypointNearestTo( const Asura_Vector_3& xPosition,
                                                                   const u_short usSearchInWaypointZone /*= ASURA_NAVVERTEX_MAX_ZONE_ID*/,
                                                                   const bool bWaypointMustContainCoord /*= false*/,
																   const bool bAllowLinearSearch /*= true*/,
                                                                   const bool bSkipConnectionlessWaypoints /*= true*/ )
{
	Asura_Assert( s_bWaypointsValid, "Waypoints have not been calculated" );
	int iWaypointIndex = -1;

	iWaypointIndex = FindNearestWaypoint_TreeSearch( xPosition, NULL, usSearchInWaypointZone, bWaypointMustContainCoord, bSkipConnectionlessWaypoints );
	if ( iWaypointIndex == -1 && bAllowLinearSearch )
    {
        iWaypointIndex = FindNearestWaypoint_LinearSearch( xPosition, NULL, usSearchInWaypointZone, bWaypointMustContainCoord );
    }

	return static_cast< unsigned short >( iWaypointIndex );
}

int Asura_AI_Navigation_System::FindNearestWaypoint_TreeSearch( const Asura_Vector_3& xPosition,
                                                                 float* pfIOMinDistance /*= NULL*/,
                                                                 const u_short usSearchInWaypointZone /*= ASURA_NAVVERTEX_MAX_ZONE_ID*/,
                                                                 const bool bWaypointMustContainCoord /*= false*/,
                                                                 const bool bSkipConnectionlessWaypoints /*= true*/ )
{
	Asura_Assert( s_bWaypointsValid, "Waypoints have not been calculated" );
	if ( !s_pxWaypointTree )
    {
        return -1;
    }

	Asura_Bounding_Box xBB;
	xBB.Reset();
	xBB.Set(xPosition);
	Asura_Vector_3 xP = xPosition;
	xP.x -= 0.5f;
	xP.z -= 0.5f;
	xP.y -= 1.0f;
	xBB.Set(xP);
	xP.x += 1.0f;
	xP.z += 1.0f;
	xP.y += 2.0f;
	xBB.Set(xP);
	
    Asura_AI_Navigation_System::s_pxWaypointCallbackTestPosition                     = &xPosition;
    Asura_AI_Navigation_System::s_fWaypointCallbackMinDistance                       = pfIOMinDistance ? *pfIOMinDistance : Asura_Maths::LargeFloat;
    Asura_AI_Navigation_System::s_usWaypointCallbackWaypointZone                     = usSearchInWaypointZone;
    Asura_AI_Navigation_System::s_bWaypointCallbackShouldContainPosition             = bWaypointMustContainCoord;
    Asura_AI_Navigation_System::s_bWaypointCallbackShouldSkipConnectionlessWaypoints = bSkipConnectionlessWaypoints;
    Asura_AI_Navigation_System::s_iWaypointCallbackWaypointID                        = -1;

	s_pxWaypointTree->TestTree( xBB, Asura_AI_Navigation_System::WaypointTreeCallback );

	// Return min distance if pointer was provided for it.
	if ( pfIOMinDistance != NULL )
    {
        *pfIOMinDistance = Asura_AI_Navigation_System::s_fWaypointCallbackMinDistance;
    }

	return Asura_AI_Navigation_System::s_iWaypointCallbackWaypointID;
}

int Asura_AI_Navigation_System::FindNearestWaypoint_LinearSearch( const Asura_Vector_3& xPosition,
                                                                   float* pfIOMinDistance /*= NULL*/,
                                                                   const u_short usSearchInWaypointZone /*= ASURA_NAVVERTEX_MAX_ZONE_ID*/,
                                                                   const bool bWaypointMustContainCoord /*= false*/,
                                                                   const bool bSkipConnectionlessWaypoints /*= true*/ )
{
	Asura_Assert( s_bWaypointsValid, "Waypoints have not been calculated" );

    float fMin        = pfIOMinDistance ? *pfIOMinDistance : Asura_Maths::LargeFloat;
	int   iWaypointID = -1;

    for ( int i = 0; i < s_usNumWaypoints; ++i )
	{
        if ( ( s_pxWaypointStorage[ i ].m_ucNumConnections == 0 ) && bSkipConnectionlessWaypoints )
        {
            continue;
        }

        if ( usSearchInWaypointZone != ASURA_NAVVERTEX_MAX_ZONE_ID && s_pxWaypointStorage[ i ].m_usZone != usSearchInWaypointZone )
        {
            continue;
        }

		const Asura_Vector_3 xDisp = xPosition - s_pxWaypointStorage[ i ].m_xPosition;
		const float          fDist = xDisp.Magnitude() - s_pxWaypointStorage[ i ].GetRadius();
		if ( fDist > fMin )
		{
            continue;
		}

        if ( !bWaypointMustContainCoord || fDist < 0.0f )
        {
            fMin        = fDist;
            iWaypointID = i;
        }
    }

    // Return min distance if pointer was provided for it.
	if ( pfIOMinDistance != NULL )
    {
        *pfIOMinDistance = fMin;
    }

	return iWaypointID;
}

bool Asura_AI_Navigation_System::IsPositionInWaypoint( const Asura_Vector_3& xPosition, const int iWaypointID )
{
	if ( iWaypointID < 0 )
	{
		return false;
	}
	if ( iWaypointID >= s_usNumWaypoints )
	{
		return false;
	}

	const Asura_Vector_3 xOffset = s_pxWaypointStorage[ iWaypointID ].m_xPosition - xPosition;
	const float          fRadius = s_pxWaypointStorage[ iWaypointID ].GetRadius();
	return ( xOffset.MagnitudeSquared() < fRadius * fRadius );
}

bool Asura_AI_Navigation_System::IsPositionInWaypoint( const Asura_Vector_3& xPosition )
{
	const int            iWaypointID = FindWaypointNearestTo( xPosition );
	const Asura_Vector_3 xOffset     = s_pxWaypointStorage[ iWaypointID ].m_xPosition - xPosition;
	const float          fRadius     = s_pxWaypointStorage[ iWaypointID ].GetRadius();
	return ( xOffset.MagnitudeSquared() < fRadius * fRadius );
}
// Old nav data query interface that is due for removal.
//-----------------------------------------------------------------------------------------------------------

#ifdef ASURA_USE_NEW_NAV_STUFF
u_int Asura_AI_Navigation_System::GetNearestWaypointID( const Asura_Vector_3& xPosition, const Asura_Navigation::TestParams& xTestParams /*= s_xStandardTParams*/ )
{
	class NearestWaypointHelper : public Asura_AABB_Tree_Searcher
	{
	public:
		NearestWaypointHelper( const Asura_Vector_3& xHelperTestPosition, const Asura_Navigation::TestParams& xHelperTestParams )
		: m_xTestPosition( xHelperTestPosition )
		, m_xTestParams( xHelperTestParams )
		, m_usBestWPID( ASURA_INVALID_WAYPOINT_ID )
		, m_fBestDistance( Asura_Maths::LargeFloat )
		{
		}
		virtual bool VisitNode( const u_short usNodeIndex )
		{
			const Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( usNodeIndex );
			if ( !pxWaypoint )
			{
				return true;
			}

			if ( !m_xTestParams.IsValidWaypoint( *pxWaypoint ) )
			{
				return true;
			}

			const float fDistance = ( pxWaypoint->m_xPosition - m_xTestPosition ).Magnitude() - pxWaypoint->GetRadius();
			if ( fDistance < m_fBestDistance )
			{
				m_fBestDistance = fDistance;
				m_usBestWPID    = usNodeIndex;
			}

			return true;
		}

		const Asura_Vector_3&               m_xTestPosition;
		const Asura_Navigation::TestParams& m_xTestParams;
		u_short                             m_usBestWPID;
		float                               m_fBestDistance;
	};

	NearestWaypointHelper xHelper( xPosition, xTestParams );

	// Find all of the valid waypoints we're inside. Pick the nearest to us.
	Asura_Bounding_Box xBB;
	xBB.Reset();
	xBB.Set( xPosition );
	xBB.Grow( 5.0f );
	s_pxWaypointTree->TestTree( xBB, xHelper );
	if ( xHelper.m_usBestWPID != ASURA_INVALID_WAYPOINT_ID )
	{
		return xHelper.m_usBestWPID;
	}

	// If we've found nothing then expand the search.
	xBB.Grow( 15.0f );
	s_pxWaypointTree->TestTree( xBB, xHelper );
	if ( xHelper.m_usBestWPID != ASURA_INVALID_WAYPOINT_ID )
	{
		return xHelper.m_usBestWPID;
	}

	// Finally use a linear search if we are allowed.
	if ( xTestParams.m_bAllowLinearSearch )
	{
        u_short usBestWPID    = ASURA_INVALID_WAYPOINT_ID;
		float   fBestDistance = Asura_Maths::LargeFloat;

		for ( u_short usWP = 0; usWP < s_usNumWaypoints; ++usWP )
		{
			const Asura_AI_NavVertex& xWP = s_pxWaypointStorage[ usWP ];
			if ( !xTestParams.IsValidWaypoint( xWP ) ) { continue; }

			const float fDistance = ( xWP.m_xPosition - xPosition ).Magnitude() - xWP.GetRadius();
			if ( fDistance < fBestDistance )
			{
				fBestDistance = fDistance;
				usBestWPID    = usWP;
			}
		}

		return usBestWPID;
	}
	else
	{
		return ASURA_INVALID_WAYPOINT_ID;
	}
}

void Asura_AI_Navigation_System::UpdateNearestWaypointID( const Asura_Vector_3& xPosition, u_int& uCurrentWaypointID, const Asura_Navigation::TestParams& xParams /*= s_xStandardTParams*/ )
{
    if ( uCurrentWaypointID == ASURA_INVALID_WAYPOINT_ID )
    {
        // We'll have to do an exhaustive search.
        uCurrentWaypointID = GetNearestWaypointID( xPosition, xParams );
    }
    else
    {
        // We can use our last value to attempt a best guess.
        // See if we're still in our current waypoint.
        // If we're not then test that waypoint's connections to see if we're in a neighbouring waypoint.
        float fNearestDistanceSq = Asura_Maths::LargeFloat;
		u_int uNearestWaypoint   = ASURA_INVALID_WAYPOINT_ID;

        const Asura_AI_NavVertex* pxNavVertex = GetWaypoint( uCurrentWaypointID );
		if ( pxNavVertex && xParams.IsValidWaypoint( *pxNavVertex ) )
        {
			float fDistSq = Asura_Maths::LargeFloat;
            if ( pxNavVertex->Contains( xPosition, &fDistSq ) )
            {
                fNearestDistanceSq = fDistSq;
                uNearestWaypoint  = uCurrentWaypointID;
            }

            for ( u_char ucLinkIndex = 0; ucLinkIndex < pxNavVertex->m_ucNumConnections; ++ucLinkIndex )
            {
                if ( !xParams.IsValidLink( pxNavVertex->GetLink( ucLinkIndex ) ) )
                {
                    continue;
                }

                const Asura_AI_NavConnection* pxConnection    = &pxNavVertex->m_pxConnections[ ucLinkIndex ];
                const Asura_AI_NavVertex*     pxNextNavVertex = GetWaypoint( pxConnection->m_usIndex );
				if ( xParams.IsValidWaypoint( *pxNextNavVertex ) && pxNextNavVertex->Contains( xPosition, &fDistSq ) )
				{
					if ( fDistSq < fNearestDistanceSq )
					{
						fNearestDistanceSq = fDistSq;
						uNearestWaypoint  = pxConnection->m_usIndex;
					}
				}
            }
        }

        if ( uNearestWaypoint != ASURA_INVALID_WAYPOINT_ID )
        {
            uCurrentWaypointID = uNearestWaypoint;
        }
        else
        {
            // Still no good - we'll have to do an exhaustive search.
            uCurrentWaypointID = GetNearestWaypointID( xPosition, xParams );
        }
    }
}

bool Asura_AI_Navigation_System::ResolveToValidPosition( Asura_Vector_3& xPosition, const Asura_Navigation::TestParams& xParams /*= s_xStandardTParams*/ )
{
	const u_int uNearestWaypointID = GetNearestWaypointID( xPosition, xParams );
	const Asura_AI_NavVertex* pxNearestWaypoint = GetWaypoint( uNearestWaypointID );
	if ( !pxNearestWaypoint ) { return false; }

	Asura_Vector_3 xWaypointToPosition = xPosition - pxNearestWaypoint->m_xPosition;
	Asura_Vector_3 xDirection          = xWaypointToPosition;
	const float	   fWaypointToPosition = xDirection.Normalise();

	// Bring the position inside.
	if ( fWaypointToPosition > pxNearestWaypoint->GetRadius() - xParams.m_fActorRadius )
	{
		xPosition           = pxNearestWaypoint->m_xPosition + ( xDirection * ( pxNearestWaypoint->GetRadius() - xParams.m_fActorRadius ) );
		xWaypointToPosition = xPosition - pxNearestWaypoint->m_xPosition;
	}

	// Resolve the position to the surface.
	const Asura_Vector_3& xWaypointNormal = pxNearestWaypoint->GetNormal();
	const float fDotProduct = xWaypointToPosition * xWaypointNormal;
	xPosition = xPosition - ( xWaypointNormal * fDotProduct );
	return true;
}

bool Asura_AI_Navigation_System::PositionIsInWaypointSpace( const Asura_Vector_3& xPosition, const Asura_Navigation::TestParams& xParams /*= s_xStandardTParams*/ )
{
	const u_int uWPID = GetNearestWaypointID( xPosition, xParams );
	return PositionIsInsideWaypoint( xPosition, uWPID );
}

bool Asura_AI_Navigation_System::PositionIsInsideWaypoint( const Asura_Vector_3& xPosition, const u_int uWPID )
{
	const Asura_AI_NavVertex* pxWP = GetWaypoint( uWPID );
	return pxWP ? pxWP->Contains( xPosition ) : false;
}

bool Asura_AI_Navigation_System::AtPosition( const Asura_Vector_3& xCurrentPosition, const Asura_Vector_3& xDesiredPosition, const Asura_Navigation::APParams& xParams /*= s_xStandardAPParams*/ )
{
	// Spherical test.
	float fLargestRangeSq = Asura_Maths::Max( xParams.m_fMaxCoplanarOffset, xParams.m_fMaxPerpendicularOffset );
	fLargestRangeSq *= fLargestRangeSq;
	if ( ( xCurrentPosition - xDesiredPosition ).MagnitudeSquared() > fLargestRangeSq )
	{
		return false;
	}
	else if ( xParams.m_bSphericalTestOnly )
	{
		return true;
	}

	// Full orientated volume test.
	Asura_OrientedVolume xOrientedVolume;
	xOrientedVolume.m_xOrient.Identity();
	xOrientedVolume.m_xBox.Reset();
	xOrientedVolume.m_xBox.Set( Asura_Vector_3( -xParams.m_fMaxCoplanarOffset, -xParams.m_fMaxPerpendicularOffset, -xParams.m_fMaxCoplanarOffset ) );
	xOrientedVolume.m_xBox.Set( Asura_Vector_3(  xParams.m_fMaxCoplanarOffset,  xParams.m_fMaxPerpendicularOffset,  xParams.m_fMaxCoplanarOffset ) );
	xOrientedVolume.m_xBox.Displace( xDesiredPosition );

	if ( xParams.m_bUseNormal )
	{
		Asura_Matrix_3x3 xOrientationMatrix;
		Asura_Maths::OrientationFromNormal( xOrientationMatrix, xParams.m_xNormal );
		xOrientedVolume.m_xOrient.FromMatrix( xOrientationMatrix );
	}

	return ( xOrientedVolume.Contains( xCurrentPosition ) );
}

static bool RayIntersectsCylinder( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection,
								   const Asura_Vector_3& xCylinderStart, const Asura_Vector_3& xCylinderNormal,
								   const float fCylinderHeight, const float fCylinderRadius,
								   float* fMin, float* fMax )
{
	// Adapted from RayIntersectsCylinder in Asura_Maths.
	// That only checks for the first intersection. This has been modified as we need both places where the line intersects the cylinder.

	// First test against each cap.
	const Asura_Vector_3	xRayEnd								= xRayOrigin + xRayDirection;
	const float				fRadiusSq							= fCylinderRadius * fCylinderRadius;
	const Asura_Vector_3	xCylinderEnd						= xCylinderStart + (xCylinderNormal * fCylinderHeight);
	const float				fRayDotCylinderAxis					= xRayDirection * xCylinderNormal;
	const Asura_Vector_3	xRayOriginToCylinderStart			= xCylinderStart - xRayOrigin;
	const Asura_Vector_3	xRayEndToCylinderStart				= xCylinderStart - xRayEnd;
	const float				fDistanceROToCylinderStartAlongAxis	= xRayOriginToCylinderStart * xCylinderNormal;
	const float				fDistanceREToCylinderStartAlongAxis	= xRayEndToCylinderStart * xCylinderNormal;
	Asura_Vector_3			xIntersection[2]					= {Asura_Maths::ZeroVector3, Asura_Maths::ZeroVector3};
	u_int					uNumIntersections					= 0;

	//----------------------------
	// Start Cap
	//----------------------------
	if ((((fDistanceROToCylinderStartAlongAxis < 0.0f) && (fRayDotCylinderAxis >= 0.0f)) ||
		((fDistanceROToCylinderStartAlongAxis > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
		(fRayDotCylinderAxis != 0.0f))
	{
		// Ray points towards start cap. Test for intersection point.
		const float				fIntersectionDistance	= fDistanceROToCylinderStartAlongAxis / fRayDotCylinderAxis;
		const Asura_Vector_3	xIntersectionPoint		= xRayOrigin + xRayDirection * fIntersectionDistance;
		const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderStart).MagnitudeSquared();

		if (fDistanceOnCapSquared <= fRadiusSq)
		{
			Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
			xIntersection[uNumIntersections] = xIntersectionPoint;
			++uNumIntersections;
		}
	}
	
	if ((((fDistanceREToCylinderStartAlongAxis < 0.0f) && (fRayDotCylinderAxis >= 0.0f)) ||
		((fDistanceREToCylinderStartAlongAxis > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
		(fRayDotCylinderAxis != 0.0f))
	{
		// Ray points towards start cap. Test for intersection point.
		const float				fIntersectionDistance	= fDistanceREToCylinderStartAlongAxis / -fRayDotCylinderAxis;
		const Asura_Vector_3	xIntersectionPoint		= xRayEnd - xRayDirection * fIntersectionDistance;
		const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderStart).MagnitudeSquared();

		if (fDistanceOnCapSquared <= fRadiusSq)
		{
			Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
			xIntersection[uNumIntersections] = xIntersectionPoint;
			++uNumIntersections;
		}
	}

	//----------------------------
	// End Cap
	//----------------------------
	if (uNumIntersections < 2)
	{
		if (((((fDistanceROToCylinderStartAlongAxis + fCylinderHeight) > 0.0f) && (fRayDotCylinderAxis <= 0.0f)) ||
			(((fDistanceROToCylinderStartAlongAxis + fCylinderHeight) > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
			(fRayDotCylinderAxis != 0.0f))
		{
			// Ray points towards end cap. Test for intersection point.
			const float				fIntersectionDistance	= (fDistanceROToCylinderStartAlongAxis + fCylinderHeight) / fRayDotCylinderAxis;
			const Asura_Vector_3	xIntersectionPoint		= xRayOrigin + xRayDirection * fIntersectionDistance;
			const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderEnd).MagnitudeSquared();

			if (fDistanceOnCapSquared <= fRadiusSq)
			{
				Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
				xIntersection[uNumIntersections] = xIntersectionPoint;
				++uNumIntersections;
			}
		}
	}
	
	if (uNumIntersections < 2)
	{
		if (((((fDistanceREToCylinderStartAlongAxis + fCylinderHeight) < 0.0f) && (fRayDotCylinderAxis >= 0.0f)) ||
			(((fDistanceREToCylinderStartAlongAxis + fCylinderHeight) > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
			(fRayDotCylinderAxis != 0.0f))
		{
			// Ray points towards end cap. Test for intersection point.
			const float				fIntersectionDistance	= (fDistanceREToCylinderStartAlongAxis + fCylinderHeight) / -fRayDotCylinderAxis;
			const Asura_Vector_3	xIntersectionPoint		= xRayEnd - xRayDirection * fIntersectionDistance;
			const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderEnd).MagnitudeSquared();

			if (fDistanceOnCapSquared <= fRadiusSq)
			{
				Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
				xIntersection[uNumIntersections] = xIntersectionPoint;
				++uNumIntersections;
			}
		}
	}

	// Only do the following if we haven't hit both end caps
	if (uNumIntersections < 2)
	{
		// Test the sides of the cylinder
		Asura_Vector_3	xPerpendicularDirection			= xRayDirection ^ xCylinderNormal;
		const float		fPerpendicularLength			= xPerpendicularDirection.Normalise();
		const float		fShortestDistanceToAxis			= xRayOriginToCylinderStart * xPerpendicularDirection;
		const float		fShortestDistanceToAxisSquared	= fShortestDistanceToAxis * fShortestDistanceToAxis;

		if (fShortestDistanceToAxisSquared <= fRadiusSq)
		{
			// Calculate distance to closest point on ray to axis.
			const Asura_Vector_3	xD1							= xRayOriginToCylinderStart ^ xCylinderNormal;
			const float				fDistanceToClosestPoint		= (xD1 * xPerpendicularDirection) / fPerpendicularLength;
			// Calculate distance along ray from closest point to cylinder surface.
			const Asura_Vector_3	xD2							= xPerpendicularDirection ^ xCylinderNormal;
			const float				fS							= Asura_Maths::Abs(Asura_Maths::Sqrt(fRadiusSq - fShortestDistanceToAxisSquared) / (xRayDirection * xD2));
			const float				fFirstIntersectionDistance	= fDistanceToClosestPoint - fS;
			const Asura_Vector_3	xFirstIntersectionPoint		= xRayOrigin + xRayDirection * fFirstIntersectionDistance;

			if (fFirstIntersectionDistance >= 0.0f)
			{
				Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
				xIntersection[uNumIntersections] = xFirstIntersectionPoint;
				++uNumIntersections;
			}

			// Do an extra number intersections check here to avoid the three intersections crash. 
			if (uNumIntersections < 2)
			{
				const float	fSecondIntersectionDistance = fDistanceToClosestPoint + fS;

				if (fFirstIntersectionDistance < fSecondIntersectionDistance)
				{
					const Asura_Vector_3	xSecondIntersectionPoint	= xRayOrigin + xRayDirection * fSecondIntersectionDistance;

					if (fSecondIntersectionDistance >= 0.f)
					{
						Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
						xIntersection[uNumIntersections] = xSecondIntersectionPoint;
						++uNumIntersections;
					}
				}
				else
				{
					Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
					xIntersection[uNumIntersections] = xFirstIntersectionPoint;
					++uNumIntersections;
				}
			}
		}
	}

	// Calculate our parametric values
	if (uNumIntersections == 1)
	{
		float fDistanceRayOriginToIntersection	= (xIntersection[0] - xRayOrigin).Magnitude();
		float fLineDistance						= xRayDirection.Magnitude();

		*fMin = 0.0f;
		*fMax = fDistanceRayOriginToIntersection / fLineDistance;

		return true;
	}
	// We have two intersections, so we can find out what portion of the ray our cylinder covers
	else if (uNumIntersections == 2)
	{
		float fDistanceToFirstIntersection	= (xIntersection[0] - xRayOrigin).Magnitude();
		float fDistanceToSecondIntersection	= (xIntersection[1] - xRayOrigin).Magnitude();
		float fLineDistance					= xRayDirection.Magnitude();

		// Smallest value to be fMin, and largest value to be fMax
		if (fDistanceToFirstIntersection > fDistanceToSecondIntersection)
		{
			*fMin = fDistanceToFirstIntersection / fLineDistance;
			*fMax = fDistanceToSecondIntersection / fLineDistance;
		}
		else
		{
			*fMin = fDistanceToFirstIntersection / fLineDistance;
			*fMax = fDistanceToSecondIntersection / fLineDistance;
		}
		return true;
	}

	// We haven't intersected with anything so set fMin and fMax to zero and retrun
	*fMin = 0.0f;
	*fMax = 0.0f;
	return false;
}

static Asura_Vector_3 s_xLineStart     = Asura_Maths::ZeroVector3;
static Asura_Vector_3 s_xLineFinish    = Asura_Maths::ZeroVector3;
static Asura_Vector_3 s_xLineDirection = Asura_Maths::ZeroVector3;

enum { MAX_HITS_ALLOWED = 32 };
static struct ParametricHit
{
	float m_fMin;
	float m_fMax;
	bool  m_bUsed;
}     s_axParametricHits[ MAX_HITS_ALLOWED ];
static u_int s_uNumParametricHits = 0;

static const Asura_Navigation::TestParams* s_pxTestParams = NULL;

static bool WaypointSpaceTreeCallback( const u_int uWaypointID )
{
	const Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( uWaypointID );

	// Check this waypoint matches our test params.
	if ( !s_pxTestParams->IsValidWaypoint( *pxWaypoint ) )
	{
		return true;
	}

	// Work out the parametric portion of the line this box covers.
	if ( s_uNumParametricHits < MAX_HITS_ALLOWED )
	{
		const Asura_Vector_3&	xCylinderNormal		= pxWaypoint->GetNormal();
		const Asura_Vector_3	xCylinderPosition	= pxWaypoint->m_xPosition - xCylinderNormal;
		const float				fCylinderRadius		= pxWaypoint->GetRadius();
		const float				fCylinderHeight		= fCylinderRadius + 1.0f;
		float fMin, fMax;

		if ( RayIntersectsCylinder( s_xLineStart, s_xLineDirection, xCylinderPosition, xCylinderNormal, fCylinderHeight, fCylinderRadius, &fMin, &fMax ) )
		{
			// This is a hit. See if we can merge it with any existing hits. If we can't then add it on the end.
			bool bMerged = false;
			for ( u_int u = 0; u < s_uNumParametricHits && !bMerged; ++u )
			{
				if ( s_axParametricHits[ u ].m_fMin <= fMin && fMax <= s_axParametricHits[ u ].m_fMax )
				{
					// This hit is completely enclosed in an already existing hit, so it can be ignored.
					bMerged = true;
				}
				else if ( fMin <= s_axParametricHits[ u ].m_fMin && s_axParametricHits[ u ].m_fMin <= fMax )
				{
					// We intersect the lower bound. We can expand this existing hit.
					fMin = Asura_Maths::Min( fMin, s_axParametricHits[ u ].m_fMin );
					fMax = Asura_Maths::Max( fMax, s_axParametricHits[ u ].m_fMax );
					s_axParametricHits[ u ].m_fMin = fMin;
					s_axParametricHits[ u ].m_fMax = fMax;
					bMerged = true;
				}
				else if ( fMin <= s_axParametricHits[ u ].m_fMax && s_axParametricHits[ u ].m_fMax <= fMax )
				{
					// We intersect the upper bound. We can expand this existing hit.
					fMin = Asura_Maths::Min( fMin, s_axParametricHits[ u ].m_fMin );
					fMax = Asura_Maths::Max( fMax, s_axParametricHits[ u ].m_fMax );
					s_axParametricHits[ u ].m_fMin = fMin;
					s_axParametricHits[ u ].m_fMax = fMax;
					bMerged = true;
				}
			}

			if ( !bMerged )
			{
				s_axParametricHits[ s_uNumParametricHits ].m_fMin = fMin;
				s_axParametricHits[ s_uNumParametricHits ].m_fMax = fMax;
				++s_uNumParametricHits;
			}
		}

		// If we now cover the whole of the line then we can terminate the waypoint graph test.
		return ( 0.0f < fMin || fMax < 1.0f );
	}
	else
	{
		Asura_Assert( false, "WaypointSpaceTreeCallback - run out of hit space" );
		return false;
	}
}

bool Asura_AI_Navigation_System::LineSegmentIsInWaypointSpace( const Asura_Vector_3& xStart, const Asura_Vector_3& xFinish, const Asura_Navigation::TestParams& xParams /*= s_xStandardTParams*/ )
{
	// Test the waypoint BB tree with the line.
	// Each waypoint returned covers another section of the line.
	// If the line is entirely covered when the test is finished then this waypoint stays in waypoint space.
	Asura_Vector_3 xLineDirection = xFinish - xStart;

	static const float fEpsilon = 0.000001f;

	s_pxTestParams	          = &xParams;
	s_xLineStart              = xStart;
	s_xLineDirection		  = xLineDirection;
	s_uNumParametricHits      = 0;

	// We expand the distance to test the tree with to ensure we hit boxes that we be just past our finish position but still cover the line.
	float fLineLength = xLineDirection.Normalise() + 15.0f;
	SearchWaypoints( xStart, xLineDirection, &fLineLength, WaypointSpaceTreeCallback );

	if ( s_uNumParametricHits == 0 )
	{
		return false;
	}

	// Iterate over the hits, merging them until they are all used.
	u_int u;
	for ( u = 0; u < s_uNumParametricHits; ++u )
	{
		s_axParametricHits[ u ].m_bUsed = false;
	}
	float fMin = s_axParametricHits[ 0 ].m_fMin, fMax = s_axParametricHits[ 0 ].m_fMax;
	s_axParametricHits[ 0 ].m_bUsed = true;
	u_int uPrevUsedCount = 1;
	u_int uUsedCount     = 1;
	do 
	{
		uPrevUsedCount = uUsedCount;

		for ( u = 0; u < s_uNumParametricHits; ++u )
		{
			if ( !s_axParametricHits[ u ].m_bUsed )
			{
				if ( s_axParametricHits[ u ].m_fMin < fMin && fMin < s_axParametricHits[ u ].m_fMax )
				{
					fMin = s_axParametricHits[ u ].m_fMin;
					s_axParametricHits[ u ].m_bUsed = true;
				}
				if ( s_axParametricHits[ u ].m_fMin < fMax && fMax < s_axParametricHits[ u ].m_fMax )
				{
					fMax = s_axParametricHits[ u ].m_fMax;
					s_axParametricHits[ u ].m_bUsed = true;
				}
				if ( s_axParametricHits[ u ].m_bUsed )
				{
					++uUsedCount;
				}
			}
		}

		if ( fMin <= 0.0f && 1.0f <= fMax )
		{
			break;
		}

		// We stop if we've either used all of the hits up or we used none up this iteration (meaning we can't merge any further).
	} while ( ( uUsedCount < s_uNumParametricHits ) && ( uUsedCount != uPrevUsedCount ) );

	return ( fMin <= 0.0f && 1.0f <= fMax );
}
#endif // #ifdef ASURA_USE_NEW_NAV_STUFF

void Asura_AI_Navigation_System::SearchWaypoints( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection,
												   float* pfDistance, bool( *pfnCallback )( const u_int uIndex ) )
{
	if ( s_pxWaypointTree )
	{
		s_pxWaypointTree->TestTreeWithRay( xRayOrigin, xRayDirection, pfDistance, pfnCallback );
	}
}

void Asura_AI_Navigation_System::SearchWaypoints( const Asura_Bounding_Box& xBB, bool( *pfnCallback )( u_short usWaypointIdx ) )
{
	if ( s_pxWaypointTree )
	{
		s_pxWaypointTree->TestTree( xBB, pfnCallback );
	}
}

void Asura_AI_Navigation_System::AddWaypointWeight( const unsigned short usWaypointIndex, const unsigned char ucWaypointWeight )
{
	const int iTotal = s_pucWaypointWeights[ usWaypointIndex ] + ucWaypointWeight;
	Asura_Assert( iTotal <= 255, "Asura_AI_Navigation_System::AddWaypointWeight - value overflow" );
	s_pucWaypointWeights[ usWaypointIndex ] = static_cast< unsigned char >( Asura_Maths::Min( iTotal, 255 ) );
}

void Asura_AI_Navigation_System::RemoveWaypointWeight( const unsigned short usWaypointIndex, const unsigned char ucWaypointWeight )
{
	const int iTotal = s_pucWaypointWeights[ usWaypointIndex ] - ucWaypointWeight;
	Asura_Assert( iTotal >= 0, "Asura_AI_Navigation_System::RemoveWaypointWeight - value underflow" );
	s_pucWaypointWeights[ usWaypointIndex ] = static_cast< unsigned char >( Asura_Maths::Max( iTotal, 0 ) );
}

unsigned char Asura_AI_Navigation_System::GetWaypointWeight( const unsigned short usWaypointIndex )
{
	return s_pucWaypointWeights[ usWaypointIndex ];
}

int Asura_AI_Navigation_System::GetNumCoverpointsForWaypoint( const int iWaypointID )
{
	if ( s_iNumCoverpoints )
	{
		Asura_Assert( 0 <= iWaypointID && iWaypointID < s_usNumWaypoints, "WaypointID out of range" );
		return s_pxPerVertexCoverpointData[ iWaypointID ].m_usNumCoverpoints;
	}
	else
	{
		return 0;
	}
}

Asura_AI_PerVertexCoverpointData* Asura_AI_Navigation_System::GetCoverpointData( const int iWaypointID )
{
	if ( s_iNumCoverpoints )
	{
		Asura_Assert( 0 <= iWaypointID && iWaypointID < s_usNumWaypoints, "WaypointID out of range" );
		return &s_pxPerVertexCoverpointData[ iWaypointID ];
	}
	else
	{
		return NULL;
	}
}

/// Kevin Lea - 26/06/2002 14:18:35 - this probably wants to expanded to take multiple threats into account.
/// Asura_AI_Tactics.cpp?
bool Asura_AI_Navigation_System::GetNearestCover(const Asura_Vector_3& xThreatPosition,const Asura_Vector_3& xCurrentPosition, Asura_Vector_3& xCoverPosition)
{
	int iIndex = GetNearestCoverIndex(xThreatPosition, xCurrentPosition);
	if(iIndex != -1)
	{
		xCoverPosition = s_pxCoverpointStorage[iIndex].m_xCentre;
		return true;
	}
	return false;
}

int Asura_AI_Navigation_System::GetNearestCoverIndex(const Asura_Vector_3& xThreatPosition,const Asura_Vector_3& xCurrentPosition, bool bRequireExits /* = false */)
{
	Asura_Assert(s_pxWaypointStorage, "No waypoints");
	if(!s_pxWaypointStorage) return -1;
	const int iWayPoint = FindWaypointNearestTo(xThreatPosition);
	if (!s_pxPerVertexCoverpointData[iWayPoint].m_usNumCoverpoints || !s_pxPerVertexCoverpointData[iWayPoint].m_pusCoverpoints) return -1;

	float fMin = Asura_Maths::LargeFloat;
	int iNearestCoverPoint = -1;
	for (int c=0; c<s_pxPerVertexCoverpointData[iWayPoint].m_usNumCoverpoints; c++)
	{
		Asura_Assert(s_pxPerVertexCoverpointData[iWayPoint].m_pusCoverpoints, "No coverpoints");

		const int iCoverPoint = s_pxPerVertexCoverpointData[iWayPoint].m_pusCoverpoints[c];
		const Asura_AI_Coverpoint& xCoverpoint = s_pxCoverpointStorage[iCoverPoint];
		// Reject if in use or no exits (if specified
		if(xCoverpoint.IsLocked()) continue;
		if(!xCoverpoint.IsValid()) continue;
		if(bRequireExits && !xCoverpoint.m_ucLongExits && !xCoverpoint.m_ucShortExits)  continue;

		const Asura_Vector_3 xDiff = xCoverpoint.m_xCentre - xCurrentPosition;
		const float fDist = xDiff.Magnitude();

		if (fDist < fMin)
		{
			fMin = fDist;
			iNearestCoverPoint = iCoverPoint;
		}
	}
	return iNearestCoverPoint;
}

Asura_AI_Coverpoint* Asura_AI_Navigation_System::GetCoverpoint( const int iCoverpointID )
{
	Asura_Assert( 0 <= iCoverpointID && iCoverpointID < s_iNumCoverpoints, "Asura_AI_Navigation_System::GetCoverpoint - coverpoint ID out of range" );
	return &s_pxCoverpointStorage[ iCoverpointID ];
}

bool Asura_AI_Navigation_System::GetAllCoverpoints(const Asura_Vector_3& xThreatPosition, unsigned short** ppusCoverpoints, int* piNumberOfCoverpoints)
{
	if ( s_pxPerVertexCoverpointData )
	{
		const int iWaypoint = FindWaypointNearestTo( xThreatPosition );
	
		if( s_pxPerVertexCoverpointData[iWaypoint].m_usNumCoverpoints == 0 ) 
		{
			return false;
		}
		else
		{
			*ppusCoverpoints = s_pxPerVertexCoverpointData[iWaypoint].m_pusCoverpoints;
			*piNumberOfCoverpoints = s_pxPerVertexCoverpointData[iWaypoint].m_usNumCoverpoints;
			return true;
		}
	}
	return false;
}

int Asura_AI_Navigation_System::GetCoverpointNearestTo(const Asura_Vector_3& xPosition)
{

	// warning, this function searches through _all_ coverpoints, it is pretty slow.
	
	Asura_Vector_3 xOffset;
	float fDistance2 = 0.0f;
	float fBestDistance2 = Asura_Maths::LargeFloat;
	int iBestCoverpoint = -1;

	for(int i=0; i<s_iNumCoverpoints;++i)
	{
		if(s_pxCoverpointStorage[i].IsValid())
		{
			xOffset = xPosition - s_pxCoverpointStorage[i].m_xCentre;

			fDistance2 = xOffset.x*xOffset.x + xOffset.z*xOffset.z;

			if(fDistance2 < fBestDistance2)
			{
				fBestDistance2 = fDistance2;
				iBestCoverpoint = i;
			}
		}
	}

	return iBestCoverpoint;

}

void Asura_AI_Navigation_System::MarkConnectionAsBlocked(const Asura_AI_NavConnection_Index xConnectionIndex, const bool bBlock)
{
	Asura_Assert( xConnectionIndex.m_usWaypointIndex < s_usNumWaypoints, "Waypoint index out of range - this is probably because the entity file is out of sync with the nav file" );
	if (xConnectionIndex.m_usWaypointIndex >= s_usNumWaypoints) return;

	const Asura_AI_NavVertex* pxWaypoint = &s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex];

	Asura_Assert( xConnectionIndex.m_usConnectionIndex < pxWaypoint->m_ucNumConnections, "Connection index out of range - this is probably because the entity file is out of sync with the nav file" );
	if (xConnectionIndex.m_usConnectionIndex >= pxWaypoint->m_ucNumConnections) return;

	if (bBlock)
	{
		// (TI - 30.05.2007) - Increase block count
		++s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usBlockCount;

		if ( !(s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usFlags & ASURA_AI_CONNECTION_FLAGS_BLOCKED) )
		{
			s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usFlags |= ASURA_AI_CONNECTION_FLAGS_BLOCKED;
			s_bConnectionChangedThisFrame = true;
		}
	}
	else
	{
        // (TI - 30.05.2007) - Decrease block count
		Asura_Assert( s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usBlockCount > 0, 
					  "Block count has been reduced to less than 0!" );
		--s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usBlockCount;

		if ( s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usFlags & ASURA_AI_CONNECTION_FLAGS_BLOCKED &&
			 s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usBlockCount == 0 )
		{
			s_pxWaypointStorage[xConnectionIndex.m_usWaypointIndex].m_pxConnections[xConnectionIndex.m_usConnectionIndex].m_usFlags &= ~ASURA_AI_CONNECTION_FLAGS_BLOCKED;
			s_bConnectionChangedThisFrame = true;
		}
	}

	if (s_pfnBlockCallback)
	{
		s_pfnBlockCallback(xConnectionIndex.m_usWaypointIndex , xConnectionIndex.m_usConnectionIndex, bBlock );
	}
}

u_int Asura_AI_Navigation_System::GetConnectionOffset(const Asura_AI_NavConnection* pxConnection)
{
	Asura_Assert(pxConnection >= s_pxNavConnectionStorage && pxConnection < s_pxNavConnectionStorage + s_uNumConnections, "GetConnectionIndex connection not in storage");
	if (pxConnection >= s_pxNavConnectionStorage && pxConnection < s_pxNavConnectionStorage + s_uNumConnections)
	{
		return ((pxConnection - s_pxNavConnectionStorage) / sizeof(Asura_AI_NavConnection));
	}

	return ASURA_INVALID_CONNECTION_OFFSET;
}


void Asura_AI_Navigation_System::SetBlockCallback(void (*pfnCallback)(u_int uWaypointIndex, u_int uConnectionIndex, const bool bBlock))
{
	s_pfnBlockCallback = pfnCallback;
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Asura_AI_Navigation_System::InitialiseAStarStorage()
{
	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);

	#ifdef ASURA_USE_NEW_NAV_STUFF
	Asura_Navigation::AStarData::InitialiseAStarStorage( s_usNumWaypoints );
	#endif

	//-----------------------------------------------------------------------------------------------------------
	// Old route finding interface that is due for removal.
	// Always allocating this for now, as there are other functions to be moved over to the new stuff.
	{
		Asura_ResourceDeallocateArray(s_pxAStarList);
		Asura_ResourceAllocateArray(s_pxAStarList, s_usNumWaypoints);

		for ( unsigned short us = 0; us < s_usNumWaypoints; ++us )
		{
			s_pxAStarList[ us ].m_iIndex = us;
		}

		Asura_ResourceDeallocateArray(s_pxAStarOpenClosedVector);
		Asura_ResourceAllocateArray(s_pxAStarOpenClosedVector, s_usNumWaypoints);

		Asura_ResourceDeallocateArray(s_pxAStarOpenList);
		Asura_ResourceAllocateArray(s_pxAStarOpenList, s_usNumWaypoints);
	}
	// Old route finding interface that is due for removal.
	//-----------------------------------------------------------------------------------------------------------

	Asura_ResourcePopType();
}

void Asura_AI_Navigation_System::CreateWaypointTree()
{
	if ( s_usNumWaypoints == 0 )
	{
		Asura_Assert( false, "Asura_AI_Navigation_System::CreateWaypointTree - no waypoints" );
		return;
	}

	if ( s_pxWaypointTree != NULL )
	{
		delete s_pxWaypointTree;
	}

	s_pxWaypointTree = Asura_AABB_Tree_Creator::Create( s_usNumWaypoints, Asura_AI_Navigation_System::WaypointTreeCreateCallback );
}

//-----------------------------------------------------------------------------------------------------------
// Old route finding interface that is due for removal.
void Asura_AI_Navigation_System::AStarSearch( unsigned short* pusRoute, int& iLengthOfRoute, const int iBufferLength,
											   const unsigned short usStart,
											   const unsigned uPathfinderFlags,
											   void* pUserData,
											   AANS_CostToGoalEstFn* pfnCostToGoalEstFn,
											   AANS_TraversalCostFn* pfnTraversalCostFn,
											   AANS_GoalAchievedFn*  pfnGoalAchievedFn,
											   AANS_LinkFlagsTestFn* pfnLinkFlagsTestFn )
{
	if ( usStart >= s_usNumWaypoints )
	{
		Asura_Assert( false, "Asura_AI_Navigation_System::AStarSearch - invalid start waypoint" );
		iLengthOfRoute = 0;
		return;
	}

	if ( s_usNumWaypoints == 0 )
	{
		Asura_Assert( false, "Asura_AI_Navigation_System::AStarSearch - no waypoints" );
		iLengthOfRoute = 0;
		return;
	}

	// Waypoints must be in the same zone for there to be a valid route between them.
    if ( s_bRouteTargetValid &&
        ( !( uPathfinderFlags & ASURA_AI_PATHFINDER_FLAGS_IGNOREZONES ) ) &&
        s_pxWaypointStorage[ usStart ].m_usZone != s_pxWaypointStorage[ s_usRouteTarget ].m_usZone )
    {
        iLengthOfRoute = 0;
		return;
	}

	// initialise the open list element count to 0
	s_iNumElemsInOpenList = 0;
	// make all nodes not open and not closed
	Asura_Mem::Set( s_pxAStarOpenClosedVector, 0, s_usNumWaypoints * sizeof( Asura_OpenClosed ) );

	int iCurrentIndex = usStart;
	Asura_AStar*      pxCurrentNode       = &s_pxAStarList[ iCurrentIndex ];
	Asura_OpenClosed* pxCurrentOpenClosed = &s_pxAStarOpenClosedVector[ iCurrentIndex ];

	// add first node to open map and open list
	pxCurrentOpenClosed->m_bOpen = true;
	pxCurrentNode->m_fCostToHere = 0.0f;
	pxCurrentNode->m_fCostToGoal = pfnCostToGoalEstFn( iCurrentIndex, pUserData );
	pxCurrentNode->m_pxParent    = NULL;

	AddNodeToOpenList( pxCurrentNode );

	while ( ( pxCurrentNode = GetOpenListNode() ) != NULL )
	{
		iCurrentIndex       = pxCurrentNode->m_iIndex;
		pxCurrentOpenClosed = &s_pxAStarOpenClosedVector[ iCurrentIndex ];

		if ( pfnGoalAchievedFn( iCurrentIndex, pUserData ) )
		{
			// done
			break;
		}

		// visit all the neighbours of the current node, and decide whether to add them to the open list
		for ( u_char uc = 0; uc < s_pxWaypointStorage[ iCurrentIndex ].m_ucNumConnections; ++uc )
		{			
			if ( pfnLinkFlagsTestFn( &s_pxWaypointStorage[ iCurrentIndex ].m_pxConnections[ uc ], uPathfinderFlags, pUserData ) )
			{
				continue;
			}

			const float fTraversalCost = pfnTraversalCostFn( iCurrentIndex, uc, uPathfinderFlags, pUserData );
			// if we get a negative cost then disregard this link
			if ( fTraversalCost < 0.0f )
			{
				continue;
			}
			const float fNewCostToHere = pxCurrentNode->m_fCostToHere + fTraversalCost;

			// get a pointer to this neighbour
			const int iIndex = s_pxWaypointStorage[ iCurrentIndex ].m_pxConnections[ uc ].m_usIndex;
			Asura_AStar*      pxNewNode       = &s_pxAStarList[ iIndex ];
			Asura_OpenClosed* pxNewOpenClosed = &s_pxAStarOpenClosedVector[ iIndex ];

			// if we've already visited this node before && we had a better cost previously then move on
			if ( ( pxNewOpenClosed->m_bOpen || pxNewOpenClosed->m_bClosed ) && pxNewNode->m_fCostToHere <= fNewCostToHere )
			{
				continue;
			}

			pxNewNode->m_pxParent    = pxCurrentNode;
			pxNewNode->m_fCostToHere = fNewCostToHere;
			pxNewNode->m_fCostToGoal = pfnCostToGoalEstFn( iIndex, pUserData );

			pxNewOpenClosed->m_bClosed = false;
			pxNewOpenClosed->m_bOpen   = true;

			AddNodeToOpenList( pxNewNode );
		}

		pxCurrentOpenClosed->m_bClosed = true;
		pxCurrentOpenClosed->m_bOpen   = false;
	}

	// copy the route out
	Asura_AStar* pxSavedEndNode = pxCurrentNode;
	iLengthOfRoute = 0;

	// cycle back up the node list to count how many there are
	while ( pxCurrentNode != NULL )
	{
		++iLengthOfRoute;
		pxCurrentNode = pxCurrentNode->m_pxParent;
	}

	if ( iLengthOfRoute >= iBufferLength )
	{
		Asura_Assert( false, "Asura_AI_Navigation_System::AStarSearch - route is too long for the route buffer" );
		if ( s_bRouteTargetValid )
		{
			Asura_OutputToDebugger( ( "Asura_AI_Navigation_System::AStarSearch - route between %d and %d is too long for the route buffer\n",
									usStart, s_usRouteTarget ) );
		}
		else
		{
			Asura_OutputToDebugger( ( "Asura_AI_Navigation_System::AStarSearch - route from %d is too long for the route buffer\n",
									usStart ) );
		}
		iLengthOfRoute = 0;
	}
	else
	{
		pxCurrentNode = pxSavedEndNode;

		// then cycle back down again and write out the route
		int iIndex = iLengthOfRoute - 1;
		while ( pxCurrentNode != NULL )
		{
			pusRoute[ iIndex-- ] = static_cast< unsigned short >( pxCurrentNode->m_iIndex );
			pxCurrentNode = pxCurrentNode->m_pxParent;
		}
	}
}

void Asura_AI_Navigation_System::PreAStarSearch( const unsigned short usStart,
												  const unsigned short usFinish,
												  AANS_CostToGoalEstFn* pfnCostToGoalEstFn,
												  void* pUserData )
{
	// initialise the open list element count to 0
	s_iNumElemsInOpenList = 0;
	// make all nodes not open and not closed
	Asura_Mem::Set( s_pxAStarOpenClosedVector, 0, s_usNumWaypoints * sizeof( Asura_OpenClosed ) );

	int iCurrentIndex = usStart;
	Asura_AStar*      pxCurrentNode       = &s_pxAStarList[ iCurrentIndex ];
	Asura_OpenClosed* pxCurrentOpenClosed = &s_pxAStarOpenClosedVector[ iCurrentIndex ];

	// add first node to open map and open list
	pxCurrentOpenClosed->m_bOpen = true;
	pxCurrentNode->m_fCostToHere = 0.0f;
	pxCurrentNode->m_fCostToGoal = pfnCostToGoalEstFn( iCurrentIndex, pUserData );
	pxCurrentNode->m_pxParent    = NULL;

	AddNodeToOpenList( pxCurrentNode );

    s_usRouteTarget        = usFinish;
	s_bRouteTargetValid    = true;
	s_bMidAStarIncremental = true;
}

void Asura_AI_Navigation_System::PreAStarSearch( const unsigned short usStart, AANS_CostToGoalEstFn* pfnCostToGoalEstFn, void* pUserData )
{
	// initialise the open list element count to 0
	s_iNumElemsInOpenList = 0;
	// make all nodes not open and not closed
	Asura_Mem::Set( s_pxAStarOpenClosedVector, 0, s_usNumWaypoints * sizeof( Asura_OpenClosed ) );

	int iCurrentIndex = usStart;
	Asura_AStar*      pxCurrentNode       = &s_pxAStarList[ iCurrentIndex ];
	Asura_OpenClosed* pxCurrentOpenClosed = &s_pxAStarOpenClosedVector[ iCurrentIndex ];

	// add first node to open map and open list
	pxCurrentOpenClosed->m_bOpen = true;
	pxCurrentNode->m_fCostToHere = 0.0f;
	pxCurrentNode->m_fCostToGoal = pfnCostToGoalEstFn( iCurrentIndex, pUserData );
	pxCurrentNode->m_pxParent    = NULL;

	AddNodeToOpenList( pxCurrentNode );

	s_bRouteTargetValid    = false;
	s_bMidAStarIncremental = true;
}

bool Asura_AI_Navigation_System::AStarSearchIncremental( unsigned short* pusRoute, int& iLengthOfRoute, const int iBufferLength,
														  const unsigned uPathfinderFlags,
														  void* pUserData,
														  AANS_CostToGoalEstFn* pfnCostToGoalEstFn,
														  AANS_TraversalCostFn* pfnTraversalCostFn,
														  AANS_GoalAchievedFn*  pfnGoalAchievedFn,
														  AANS_LinkFlagsTestFn* pfnLinkFlagsTestFn )
{
	if ( !s_bMidAStarIncremental )
	{
		iLengthOfRoute = 0;
		return true;
	}

	int               iCurrentIndex = 0;
	Asura_AStar*      pxCurrentNode;
	Asura_OpenClosed* pxCurrentOpenClosed;

	if ( ( pxCurrentNode = GetOpenListNode() ) != NULL )
	{
		iCurrentIndex       = pxCurrentNode->m_iIndex;
		pxCurrentOpenClosed = &s_pxAStarOpenClosedVector[ iCurrentIndex ];

		if ( !pfnGoalAchievedFn( iCurrentIndex, pUserData ) )
		{
			// visit all the neighbours of the current node, and decide whether to add them to the open list
			for ( u_char uc = 0; uc < s_pxWaypointStorage[ iCurrentIndex ].m_ucNumConnections; ++uc )
			{
				if ( pfnLinkFlagsTestFn( &s_pxWaypointStorage[ iCurrentIndex ].m_pxConnections[ uc ], uPathfinderFlags, pUserData ) )
				{
					continue;
				}

				const float fTraversalCost = pfnTraversalCostFn( iCurrentIndex, uc, uPathfinderFlags, pUserData );
				// if we get a negative cost then disregard this link
				if ( fTraversalCost < 0.0f )
				{
					continue;
				}
				const float fNewCostToHere = pxCurrentNode->m_fCostToHere + fTraversalCost;

				// get a pointer to this neighbour
				const int iIndex = s_pxWaypointStorage[ iCurrentIndex ].m_pxConnections[ uc ].m_usIndex;
				Asura_AStar*      pxNewNode       = &s_pxAStarList[ iIndex ];
				Asura_OpenClosed* pxNewOpenClosed = &s_pxAStarOpenClosedVector[ iIndex ];

				// if we've already visited this node before && we had a better cost previously then move on
				if ( ( pxNewOpenClosed->m_bOpen || pxNewOpenClosed->m_bClosed ) && pxNewNode->m_fCostToHere <= fNewCostToHere )
				{
					continue;
				}

				pxNewNode->m_pxParent    = pxCurrentNode;
				pxNewNode->m_fCostToHere = fNewCostToHere;
				pxNewNode->m_fCostToGoal = pfnCostToGoalEstFn( iIndex, pUserData );

				pxNewOpenClosed->m_bClosed = false;
				pxNewOpenClosed->m_bOpen   = true;

				AddNodeToOpenList( pxNewNode );
			}

			pxCurrentOpenClosed->m_bClosed = true;
			pxCurrentOpenClosed->m_bOpen   = false;
		}
	}

	if ( pxCurrentNode == NULL )
	{
		iLengthOfRoute         = 0;
		s_bMidAStarIncremental = false;

		return true;
	}
	else if ( pfnGoalAchievedFn( iCurrentIndex, pUserData ) )
	{
		// copy the route out
		Asura_AStar* pxSavedEndNode = pxCurrentNode;
		iLengthOfRoute = 0;

		// cycle back up the node list to count how many there are
		while ( pxCurrentNode != NULL )
		{
			++iLengthOfRoute;
			pxCurrentNode = pxCurrentNode->m_pxParent;
		}

		if ( iLengthOfRoute >= iBufferLength )
		{
			Asura_Assert( false, "Asura_AI_Navigation_System::AStarSearch - route is too long for the route buffer" );
			iLengthOfRoute = 0;
			return true;
		}

		pxCurrentNode = pxSavedEndNode;

		// then cycle back down again and write out the route
		int iIndex = iLengthOfRoute - 1;
		while ( pxCurrentNode != NULL )
		{
			pusRoute[ iIndex-- ] = static_cast< unsigned short >( pxCurrentNode->m_iIndex );
			pxCurrentNode = pxCurrentNode->m_pxParent;
		}

		s_bMidAStarIncremental = false;

		return true;
	}
	else
	{
		return false;
	}
}

void Asura_AI_Navigation_System::AddNodeToOpenList( const Asura_AStar* pxNode )
{
	// add the node to the end of the list
	s_pxAStarOpenList[ s_iNumElemsInOpenList ].m_fTotalCost = pxNode->m_fCostToHere + pxNode->m_fCostToGoal;
	s_pxAStarOpenList[ s_iNumElemsInOpenList ].m_iIndex     = pxNode->m_iIndex;

	++s_iNumElemsInOpenList;
}

Asura_AStar* Asura_AI_Navigation_System::GetOpenListNode()
{
	// Do a linear search of the open list for the node with the lowest value.
	// Remove the node from the open list, and return it.
	if ( s_iNumElemsInOpenList == 0 )
	{
		return NULL;
	}

	float fMinCost   = s_pxAStarOpenList[ 0 ].m_fTotalCost;
	int   iBestIndex = 0;
	for ( int i = 1; i < s_iNumElemsInOpenList; ++i )
	{
		if ( s_pxAStarOpenList[ i ].m_fTotalCost < fMinCost )
		{
			fMinCost   = s_pxAStarOpenList[ i ].m_fTotalCost;
			iBestIndex = i;
		}
	}

	const int iBestWPID = s_pxAStarOpenList[ iBestIndex ].m_iIndex;

	// Copy the last element in the open list over the position of the element we found, and decrement the size of the open list.
	s_pxAStarOpenList[ iBestIndex ] = s_pxAStarOpenList[ --s_iNumElemsInOpenList ];

	return &s_pxAStarList[ iBestWPID ];
}

bool Asura_AI_Navigation_System::DefaultGoalAchievedFunction( const int iCurrentIndex, void* pUserData )
{
	return ( s_usRouteTarget == static_cast< unsigned short >( iCurrentIndex ) );
}
// Old route finding interface that is due for removal.
//-----------------------------------------------------------------------------------------------------------

//////////////////////////////////////////////////////////////////////////////
//
// Asura_AI_Navigation_System::UserDefinedWaypoint
//
//////////////////////////////////////////////////////////////////////////////

u_int Asura_AI_Navigation_System::UserDefinedWaypoint::s_uCurrentID = uINVALID_ID;

Asura_AI_Navigation_System::UserDefinedWaypoint::UserDefinedWaypoint()
: m_xPosition( Asura_Maths::ZeroVector3 )
, m_xOrientQ( Asura_Maths::IdentityQuat )
, m_uFlags( 0 )
, m_uID( uINVALID_ID )
{
}

void Asura_AI_Navigation_System::UserDefinedWaypoint::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;

	// Ordinarily we'd chunk in the parent now.

	if ( uVersion == uCURRENT_CHUNK_VERSION )
	{
		xStream >> m_uID;
		UpdateCurrentID( m_uID );
		xStream >> m_xPosition;
		xStream >> m_xOrientQ;
		xStream >> m_uFlags;
	} 
	else
	{
		Asura_Assertf( false, ( "Asura_AI_Navigation_System::UserDefinedWaypoint::ReadFromChunkStream - bad chunk version (%u)", uVersion ) );
	}
}

void Asura_AI_Navigation_System::UserDefinedWaypoint::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << uCURRENT_CHUNK_VERSION;

	// Ordinarily we'd chunk out the parent now.

	xStream << m_uID;
	xStream << m_xPosition;
	xStream << m_xOrientQ;
	xStream << m_uFlags;
}

Asura_AI_Navigation_System::UserDefinedWaypoint* Asura_AI_Navigation_System::CreateUserDefinedWaypoint()
{
	return new UserDefinedWaypoint;
}

void Asura_AI_Navigation_System::DestroyUserDefinedWaypoint( UserDefinedWaypoint* pxUDW )
{
	delete pxUDW;
}

#ifdef ASURA_TOOLS
void Asura_AI_Navigation_System::ResetForWaypointGeneration()
{
	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
	{
		Asura_ResourceDeallocateArray(s_pxWaypointStorage);
		Asura_ResourceDeallocateArray(s_pxNavConnectionStorage);
		Asura_ResourceDeallocateArray(s_pxPerVertexCoverpointData);
		Asura_ResourceDeallocateArray(s_pusCoverpointIndexStorage);
		Asura_ResourceDeallocateArray(s_pxCoverpointStorage);
		delete   s_pxWaypointTree;
		Asura_ResourceDeallocateArray(s_pucWaypointWeights);
		//-----------------------------------------------------------------------------------------------------------
		// Old route finding interface that is due for removal.
		Asura_ResourceDeallocateArray(s_pxAStarList);
		Asura_ResourceDeallocateArray(s_pxAStarOpenClosedVector);
		Asura_ResourceDeallocateArray(s_pxAStarOpenList);
		// Old route finding interface that is due for removal.
		//-----------------------------------------------------------------------------------------------------------

		s_bWaypointsValid             = false;
		s_usNumWaypoints              = 0;
		s_pxWaypointStorage           = NULL;
		s_uNumConnections             = 0;
		s_pxNavConnectionStorage      = NULL;
		s_pxPerVertexCoverpointData   = NULL;
		s_uNumCoverpointIndices       = 0;
		s_pusCoverpointIndexStorage   = NULL;
		s_iNumCoverpoints             = 0;
		s_pxCoverpointStorage         = NULL;
		s_pxWaypointTree              = NULL;
		s_pucWaypointWeights          = NULL;

		//-----------------------------------------------------------------------------------------------------------
		// Old route finding interface that is due for removal.
		s_pxAStarList                 = NULL;
		s_pxAStarOpenClosedVector     = NULL;
		s_pxAStarOpenList             = NULL;
		s_iNumElemsInOpenList         = 0;

		s_bRouteTargetValid           = false;
		s_usRouteTarget               = 0;
		s_fPathfindRange              = 0.0f;
		s_xPathfindPosition           = Asura_Maths::ZeroVector3;
		// Old route finding interface that is due for removal.
		//-----------------------------------------------------------------------------------------------------------

		s_bConnectionChangedThisFrame = false;
		s_bConnectionChangedLastFrame = false;
		s_uConnectionGeneration		  = CONNECTION_GENERATION_FIRST;
	}
	Asura_ResourcePopType();
}

void Asura_AI_Navigation_System::CreateWaypoint( const Asura_Vector_3& xPos, const float fRadius )
{
	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);

	Asura_AI_NavVertex* pxWaypoints;

	Asura_ResourceAllocateArray(pxWaypoints, s_usNumWaypoints + 1);

	if ( pxWaypoints == NULL )
	{
		Asura_Assert( false, "Out of memory while creating another waypoint." );
		Asura_ResourcePopType();
		return;
	}

	Asura_AI_NavVertex* pxWaypointSource = s_pxWaypointStorage;
	Asura_AI_NavVertex* pxWaypointSourceEnd = &s_pxWaypointStorage[ s_usNumWaypoints ];
	Asura_AI_NavVertex* pxWaypointDest = pxWaypoints;

	while ( pxWaypointSource < pxWaypointSourceEnd )
	{
		*pxWaypointDest++ = *pxWaypointSource++;
	}

	pxWaypointDest->m_xPosition = xPos;
	pxWaypointDest->SetRadius( fRadius );
	pxWaypointDest->m_ucNumConnections = 0;
	pxWaypointDest->m_pxConnections = NULL;
	pxWaypointDest->m_usZone = 0;
	pxWaypointDest->m_usFlags = 0;

	Asura_ResourceDeallocateArray( s_pxWaypointStorage );
	s_pxWaypointStorage = pxWaypoints;

	++s_usNumWaypoints;

	Asura_ResourcePopType();
}

Asura_AI_Navigation_System::UserDefinedWaypoint* Asura_AI_Navigation_System::AddUserDefinedWaypoint()
{
	Asura_ResourcePushType( ASURA_RESOURCE_WAYPOINT );

	UserDefinedWaypoint* pxUDW = GetInstance()->CreateUserDefinedWaypoint();
	pxUDW->AquireID();
	s_xUDWPList.Push_Back( pxUDW );

	Asura_ResourcePopType();

	return pxUDW;
}

void Asura_AI_Navigation_System::DeleteUserDefinedWaypoint( UserDefinedWaypoint* pxUDW )
{
	Asura_ResourcePushType( ASURA_RESOURCE_WAYPOINT );

	Asura_ListElt< UserDefinedWaypoint* >* pxUDWElt = s_xUDWPList.Find( pxUDW );
	if ( pxUDWElt )
	{
		s_xUDWPList.Remove( pxUDWElt );
		GetInstance()->DestroyUserDefinedWaypoint( pxUDW );
	}
	else
	{
		Asura_Assert( false, "Asura_AI_Navigation_System::DeleteUserDefinedWaypoint - could not find waypoint to delete" );
	}

	Asura_ResourcePopType();
}
#endif

#ifdef ASURA_TOOLS
void Asura_AI_Navigation_System::GenerateSubsetForExport()
{
	//DeleteSubsetForExport();

	//if(!m_usNumWaypoints || !m_bWaypointsValid) return;

	////no need to do anything if exporting all modules.
	//if(Tools_Level_Export::IsExportingAllModules())
	//{
	//	return;
	//}

	////allocate and initialise the storage for the waypoint subset
	//m_pxSubsetForExport = new Asura_AI_WaypointSubsetForExport;

	//m_pxSubsetForExport->m_usWaypointConversionSize = m_usNumWaypoints;
	//m_pxSubsetForExport->m_piWaypointConv = new int[m_usNumWaypoints];

	//m_pxSubsetForExport->m_pxHandler = new Asura_AI_Navigation_Handler;
	//Asura_AI_Navigation_Handler* pxDestHandler = m_pxSubsetForExport->m_pxHandler;

	////get a module for each waypoint , and determine whether it is being exported
	//const Asura_AI_NavVertex* pxWaypoint = m_pxWaypointStorage;
	//Tools_Environment_LOS::EnsureDataIsValid();
	//const Asura_Vector_3 xLOSDir(0.0f, 1.0f, 0.0f);

	////initialise common LOS test settings
	//Tools_Environment_LOS_Data xLOSData;
	//xLOSData.SetRayDirection(xLOSDir);
	//xLOSData.SetReturnOnFirstHit(true);
	//xLOSData.SetPolyFlagsToIgnore(0);
	//xLOSData.SetPolyFlagsToRequire(AsuraPolygonFlag_Floor);

	//for(u_short usWP=0; usWP<m_usNumWaypoints; usWP++, pxWaypoint++)
	//{
	//	//Do an los test down from slightly above the waypoint position
	//	//Since the waypoints can be generated at a vertex, it is possible that the LOS test will
	//	//fall between the cracks,and fail to hit a polygon.
	//	//If the first test fails ,then try a few more slightly to the side in the hope of getting a hit nearby.
	//	//RAR 09/04/2008
	//	bool bKeep = false;
	//	for(int iPass=0;iPass<5;iPass++)
	//	{
	//		Asura_Vector_3 xPos = pxWaypoint->m_xPosition;
	//		xPos.y -= 0.5f;

	//		switch(iPass)
	//		{
	//		case 1 : xPos.x -= 0.1f; break;
	//		case 2 : xPos.x += 0.1f; break;
	//		case 3 : xPos.z -= 0.1f; break;
	//		case 4 : xPos.z += 0.1f; break;
	//		default : break;
	//		}


	//		xLOSData.SetRayOrigin(xPos);
	//		if(Tools_Environment_LOS::TestWithRay(xLOSData))
	//		{
	//			if(xLOSData.GetZoneHash()==0 && Tools_Level_Export::ShouldIncludeModuleInExport(xLOSData.GetModuleIndex()))
	//			{
	//				bKeep = true;
	//			}
	//			break;
	//		}
	//	}

	//	if(bKeep)
	//	{
	//		m_pxSubsetForExport->m_piWaypointConv[usWP] = pxDestHandler->m_usNumWaypoints++;
	//	}
	//	else
	//	{
	//		m_pxSubsetForExport->m_piWaypointConv[usWP] = -1;
	//	}
	//}

	//if(pxDestHandler->m_usNumWaypoints == m_usNumWaypoints)
	//{
	//	//ended up keeping all the waypoints,so can skip doing the rest
	//	DeleteSubsetForExport();
	//	return;
	//}

	//Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
	//{

	//	//Copy the waypoints to the new handler, and count the number of connections used by them
	//	Asura_ResourceAllocateArray(pxDestHandler->m_pxWaypointStorage, pxDestHandler->m_usNumWaypoints);

	//	const Asura_AI_NavVertex* pxSourceWaypoint = m_pxWaypointStorage;
	//	Asura_AI_NavVertex* pxDestWaypoint = pxDestHandler->m_pxWaypointStorage;

	//	for(u_short usWP=0; usWP<m_usNumWaypoints; usWP++, pxSourceWaypoint++)
	//	{
	//		if(m_pxSubsetForExport->m_piWaypointConv[usWP] != -1)
	//		{
	//			*pxDestWaypoint = *pxSourceWaypoint;
	//			pxDestWaypoint->m_ucNumConnections = 0;
	//			pxDestWaypoint->m_pxConnections = 0;

	//			//check which connections still lead to valid waypoints
	//			const Asura_AI_NavConnection* pxConnection = pxSourceWaypoint->m_pxConnections;
	//			for(u_int i=0; i<pxSourceWaypoint->m_ucNumConnections; i++, pxConnection++)
	//			{
	//				int iConvertedDestIndex = m_pxSubsetForExport->m_piWaypointConv[pxConnection->m_usIndex];
	//				if( iConvertedDestIndex != -1)
	//				{
	//					pxDestHandler->m_uNumConnections++;
	//					pxDestWaypoint->m_ucNumConnections++;
	//				}
	//			}

	//			pxDestWaypoint++;
	//		}
	//	}
	//	Asura_Assert(pxDestWaypoint - pxDestHandler->m_pxWaypointStorage == pxDestHandler->m_usNumWaypoints,"");


	//	//copy and convert the connections
	//	Asura_ResourceAllocateArray(pxDestHandler->m_pxNavConnectionStorage, pxDestHandler->m_uNumConnections);

	//	pxSourceWaypoint = m_pxWaypointStorage;
	//	pxDestWaypoint = pxDestHandler->m_pxWaypointStorage;
	//	Asura_AI_NavConnection* pxDestConnection = pxDestHandler->m_pxNavConnectionStorage;
	//	for(u_short usWP=0; usWP<m_usNumWaypoints; usWP++, pxSourceWaypoint++)
	//	{
	//		if(m_pxSubsetForExport->m_piWaypointConv[usWP] != -1)
	//		{
	//			if(pxDestWaypoint->m_ucNumConnections)
	//			{
	//				pxDestWaypoint->m_pxConnections = pxDestConnection;

	//				const Asura_AI_NavConnection* pxConnection = pxSourceWaypoint->m_pxConnections;
	//				for(u_int i=0; i<pxSourceWaypoint->m_ucNumConnections; i++, pxConnection++)
	//				{
	//					int iConvertedDestIndex = m_pxSubsetForExport->m_piWaypointConv[pxConnection->m_usIndex];
	//					if( iConvertedDestIndex != -1)
	//					{
	//						*pxDestConnection = *pxConnection;
	//						pxDestConnection->m_usIndex = (u_short)iConvertedDestIndex;
	//						pxDestConnection++;
	//					}
	//				}
	//				Asura_Assert(pxDestConnection - pxDestWaypoint->m_pxConnections == pxDestWaypoint->m_ucNumConnections, "");
	//			}

	//			pxDestWaypoint++;
	//		}
	//	}
	//	Asura_Assert(pxDestConnection - pxDestHandler->m_pxNavConnectionStorage == (int)pxDestHandler->m_uNumConnections,"");


	//	if(m_pxPerVertexCoverpointData)
	//	{
	//		//determine the number of coverpoint and coverpoint indices required ,and build a conversion array
	//		int* piCoverpointConv = new int[m_iNumCoverpoints];
	//		for(int i=0;i<m_iNumCoverpoints;i++)
	//		{
	//			piCoverpointConv[i] = -1;
	//		}

	//		const Asura_AI_PerVertexCoverpointData* pxPerVertexCoverpointData = m_pxPerVertexCoverpointData;
	//		for(u_short usWP=0; usWP<m_usNumWaypoints; usWP++, pxPerVertexCoverpointData++)
	//		{
	//			if(m_pxSubsetForExport->m_piWaypointConv[usWP] != -1)
	//			{
	//				pxDestHandler->m_uNumCoverpointIndices += pxPerVertexCoverpointData->m_usNumCoverpoints;

	//				for(u_int i=0;i<pxPerVertexCoverpointData->m_usNumCoverpoints; i++)
	//				{
	//					u_short usSourceIndex = pxPerVertexCoverpointData->m_pusCoverpoints[i];
	//					if(piCoverpointConv[usSourceIndex] == -1)
	//					{
	//						piCoverpointConv[usSourceIndex] = pxDestHandler->m_iNumCoverpoints++;
	//					}
	//				}
	//			}
	//		}

	//		//copy and convert the coverpoint indices
	//		Asura_ResourceAllocateArray(pxDestHandler->m_pxPerVertexCoverpointData, pxDestHandler->m_usNumWaypoints);
	//		Asura_ResourceAllocateArray(pxDestHandler->m_pusCoverpointIndexStorage, pxDestHandler->m_uNumCoverpointIndices);

	//		pxPerVertexCoverpointData = m_pxPerVertexCoverpointData;
	//		Asura_AI_PerVertexCoverpointData* pxDestPerVertexCoverpointData = pxDestHandler->m_pxPerVertexCoverpointData;
	//		u_short* pusDestCoverpointIndex = pxDestHandler->m_pusCoverpointIndexStorage;
	//		for(u_short usWP=0; usWP<m_usNumWaypoints; usWP++, pxPerVertexCoverpointData++)
	//		{
	//			if(m_pxSubsetForExport->m_piWaypointConv[usWP] != -1)
	//			{
	//				pxDestPerVertexCoverpointData->m_usNumCoverpoints = pxPerVertexCoverpointData->m_usNumCoverpoints;
	//				pxDestPerVertexCoverpointData->m_pusCoverpoints = pusDestCoverpointIndex;

	//				for(u_int i=0;i<pxPerVertexCoverpointData->m_usNumCoverpoints; i++)
	//				{
	//					*pusDestCoverpointIndex++ = (u_short)piCoverpointConv[pxPerVertexCoverpointData->m_pusCoverpoints[i]];
	//				}

	//				pxDestPerVertexCoverpointData++;
	//			}
	//		}
	//		Asura_Assert(pxDestPerVertexCoverpointData - pxDestHandler->m_pxPerVertexCoverpointData == (int)pxDestHandler->m_usNumWaypoints, "");
	//		Asura_Assert(pusDestCoverpointIndex - pxDestHandler->m_pusCoverpointIndexStorage == (int) pxDestHandler->m_uNumCoverpointIndices, "");

	//		//copy the coverpoints used
	//		Asura_ResourceAllocateArray(pxDestHandler->m_pxCoverpointStorage, pxDestHandler->m_iNumCoverpoints);

	//		const Asura_AI_Coverpoint* pxSourceCoverpoint = m_pxCoverpointStorage;
	//		for(int i=0; i<m_iNumCoverpoints; i++, pxSourceCoverpoint++)
	//		{
	//			int iDestIndex = piCoverpointConv[i];
	//			if(iDestIndex != -1)
	//			{
	//				pxDestHandler->m_pxCoverpointStorage[iDestIndex] = *pxSourceCoverpoint;
	//			}
	//		}
	//		delete [] piCoverpointConv;
	//	}

	//	//build a new tree from the waypoints used
	//	if(pxDestHandler->m_usNumWaypoints)
	//	{
	//		pxDestHandler->CreateWaypointTree();
	//	}

	//}
	//Asura_ResourcePopType();

	//pxDestHandler->m_bWaypointsValid = true;
}

void Asura_AI_Navigation_System::DeleteSubsetForExport()
{
	//delete m_pxSubsetForExport;
	//m_pxSubsetForExport = 0;
}

bool Asura_AI_Navigation_System::ConvertNavConnectionIndexForExport(const Asura_AI_NavConnection_Index& xInput, Asura_AI_NavConnection_Index& xOutput)
{
	return false;

	//if(!m_pxSubsetForExport)
	//{
	//	//No conversion required
	//	xOutput = xInput;
	//	return true;
	//}

	//if(m_pxSubsetForExport->m_usWaypointConversionSize != m_usNumWaypoints)
	//{
	//	Asura_Assert(m_pxSubsetForExport->m_usWaypointConversionSize == m_usNumWaypoints, "Number of waypoints changed unexpectedly");
	//	//bit of a lost cause in this case
	//	return false;
	//}

	////check to see if the connection is valid in the original set of waypoints ,and dispose if not
	//if(xInput.m_usWaypointIndex >= m_usNumWaypoints) return false;
	//const Asura_AI_NavVertex* pxSourceWaypoint = &m_pxWaypointStorage[xInput.m_usWaypointIndex];

	//if(xInput.m_usConnectionIndex >= pxSourceWaypoint->m_ucNumConnections) return false;

	////Source maps to a valid waypoint in the subset?
	//int iConvertedSourceIndex = m_pxSubsetForExport->m_piWaypointConv[xInput.m_usWaypointIndex];
	//if(iConvertedSourceIndex == -1) return false;
	//xOutput.m_usWaypointIndex = (u_short)iConvertedSourceIndex;

	/////check to see if the connected waypoint is valid
	//u_short usDestIndex = pxSourceWaypoint->m_pxConnections[xInput.m_usConnectionIndex].m_usIndex;
	//if(usDestIndex >= m_usNumWaypoints) return false;

	////destination exists in the subset?
	//int iConvertedDestIndex = m_pxSubsetForExport->m_piWaypointConv[usDestIndex];
	//if(iConvertedDestIndex == -1) return false;

	////search the converted waypoint for the connection to this waypoint
	//const Asura_AI_NavVertex* pxConvertedSourceWaypoint = &m_pxSubsetForExport->m_pxHandler->m_pxWaypointStorage[iConvertedSourceIndex];
	//for(u_int i=0; i< pxConvertedSourceWaypoint->m_ucNumConnections; i++)
	//{
	//	if(pxConvertedSourceWaypoint->m_pxConnections[i].m_usIndex == iConvertedDestIndex)
	//	{
	//		//found the connection,and done
	//		xOutput.m_usConnectionIndex = (u_short)i;
	//		return true;
	//	}
	//}

	////Failed to find the connection.
	////Don't think this should be possible if both source and destination waypoints exist.
	//Asura_Assert(0,"Failed to convert waypoint connection index for an unexpected reason.");
	//return false;
}

Asura_AI_WaypointSubsetForExport::Asura_AI_WaypointSubsetForExport()
{
	//m_usWaypointConversionSize = 0;
	//m_piWaypointConv = 0;

	//m_pxHandler = 0;
}

Asura_AI_WaypointSubsetForExport::~Asura_AI_WaypointSubsetForExport()
{
	//delete [] m_piWaypointConv;
	//m_pxHandler->Shutdown();
	//delete m_pxHandler;
}
#endif // ASURA_TOOLS

// Stuck this at the bottom so it doesn't get in the way above.
const Asura_Vector_3 s_axNormalTable[ NUM_NAV_VERTEX_NORMALS ] =
{
	Asura_Vector_3(0.000000000000000F, 0.000000000000000F, 1.000000000000000F),
	Asura_Vector_3(0.219604492187500F, 0.000000000000000F, 0.975524902343750F),
	Asura_Vector_3(0.067810058593750F, 0.208862304687500F, 0.975524902343750F),
	Asura_Vector_3(0.428466796875000F, 0.000000000000000F, 0.903503417968750F),
	Asura_Vector_3(0.289794921875000F, 0.210510253906250F, 0.933593750000000F),
	Asura_Vector_3(0.132385253906250F, 0.407531738281250F, 0.903503417968750F),
	Asura_Vector_3(0.616455078125000F, 0.000000000000000F, 0.787353515625000F),
	Asura_Vector_3(0.502075195312500F, 0.212341308593750F, 0.838317871093750F),
	Asura_Vector_3(0.357116699218750F, 0.411865234375000F, 0.838317871093750F),
	Asura_Vector_3(0.190490722656250F, 0.586303710937500F, 0.787353515625000F),
	Asura_Vector_3(0.774353027343750F, 0.000000000000000F, 0.632690429687500F),
	Asura_Vector_3(0.690979003906250F, 0.212646484375000F, 0.690856933593750F),
	Asura_Vector_3(0.569152832031250F, 0.413513183593750F, 0.710571289062500F),
	Asura_Vector_3(0.415771484375000F, 0.591430664062500F, 0.690856933593750F),
	Asura_Vector_3(0.239257812500000F, 0.736450195312500F, 0.632690429687500F),
	Asura_Vector_3(0.894409179687500F, 0.000000000000000F, 0.447204589843750F),
	Asura_Vector_3(0.842224121093750F, 0.208862304687500F, 0.496948242187500F),
	Asura_Vector_3(0.748901367187500F, 0.407531738281250F, 0.522460937500000F),
	Asura_Vector_3(0.619018554687500F, 0.586303710937500F, 0.522460937500000F),
	Asura_Vector_3(0.458862304687500F, 0.736450195312500F, 0.496948242187500F),
	Asura_Vector_3(0.276367187500000F, 0.850646972656250F, 0.447204589843750F),
	Asura_Vector_3(-0.177673339843750F, 0.129089355468750F, 0.975524902343750F),
	Asura_Vector_3(-0.110656738281250F, 0.340698242187500F, 0.933593750000000F),
	Asura_Vector_3(-0.346679687500000F, 0.251831054687500F, 0.903503417968750F),
	Asura_Vector_3(-0.046813964843750F, 0.543151855468750F, 0.838317871093750F),
	Asura_Vector_3(-0.281311035156250F, 0.466918945312500F, 0.838317871093750F),
	Asura_Vector_3(-0.498718261718750F, 0.362365722656250F, 0.787353515625000F),
	Asura_Vector_3(0.011230468750000F, 0.722839355468750F, 0.690856933593750F),
	Asura_Vector_3(-0.217407226562500F, 0.669128417968750F, 0.710571289062500F),
	Asura_Vector_3(-0.433959960937500F, 0.578186035156250F, 0.690856933593750F),
	Asura_Vector_3(-0.626464843750000F, 0.455139160156250F, 0.632690429687500F),
	Asura_Vector_3(0.061584472656250F, 0.865539550781250F, 0.496948242187500F),
	Asura_Vector_3(-0.156127929687500F, 0.838195800781250F, 0.522460937500000F),
	Asura_Vector_3(-0.366333007812500F, 0.769897460937500F, 0.522460937500000F),
	Asura_Vector_3(-0.558593750000000F, 0.664001464843750F, 0.496948242187500F),
	Asura_Vector_3(-0.723571777343750F, 0.525695800781250F, 0.447204589843750F),
	Asura_Vector_3(-0.177673339843750F, -0.129089355468750F, 0.975524902343750F),
	Asura_Vector_3(-0.358215332031250F, 0.000000000000000F, 0.933593750000000F),
	Asura_Vector_3(-0.346679687500000F, -0.251831054687500F, 0.903503417968750F),
	Asura_Vector_3(-0.531005859375000F, 0.123229980468750F, 0.838317871093750F),
	Asura_Vector_3(-0.531005859375000F, -0.123229980468750F, 0.838317871093750F),
	Asura_Vector_3(-0.498718261718750F, -0.362365722656250F, 0.787353515625000F),
	Asura_Vector_3(-0.684020996093750F, 0.234069824218750F, 0.690856933593750F),
	Asura_Vector_3(-0.703552246093750F, 0.000000000000000F, 0.710571289062500F),
	Asura_Vector_3(-0.684020996093750F, -0.234069824218750F, 0.690856933593750F),
	Asura_Vector_3(-0.626464843750000F, -0.455139160156250F, 0.632690429687500F),
	Asura_Vector_3(-0.804138183593750F, 0.326049804687500F, 0.496948242187500F),
	Asura_Vector_3(-0.845397949218750F, 0.110473632812500F, 0.522460937500000F),
	Asura_Vector_3(-0.845397949218750F, -0.110473632812500F, 0.522460937500000F),
	Asura_Vector_3(-0.804138183593750F, -0.326049804687500F, 0.496948242187500F),
	Asura_Vector_3(-0.723571777343750F, -0.525695800781250F, 0.447204589843750F),
	Asura_Vector_3(0.067810058593750F, -0.208862304687500F, 0.975524902343750F),
	Asura_Vector_3(-0.110656738281250F, -0.340698242187500F, 0.933593750000000F),
	Asura_Vector_3(0.132385253906250F, -0.407531738281250F, 0.903503417968750F),
	Asura_Vector_3(-0.281311035156250F, -0.466918945312500F, 0.838317871093750F),
	Asura_Vector_3(-0.046813964843750F, -0.543151855468750F, 0.838317871093750F),
	Asura_Vector_3(0.190490722656250F, -0.586303710937500F, 0.787353515625000F),
	Asura_Vector_3(-0.433959960937500F, -0.578186035156250F, 0.690856933593750F),
	Asura_Vector_3(-0.217407226562500F, -0.669128417968750F, 0.710571289062500F),
	Asura_Vector_3(0.011230468750000F, -0.722839355468750F, 0.690856933593750F),
	Asura_Vector_3(0.239257812500000F, -0.736450195312500F, 0.632690429687500F),
	Asura_Vector_3(-0.558593750000000F, -0.664001464843750F, 0.496948242187500F),
	Asura_Vector_3(-0.366333007812500F, -0.769897460937500F, 0.522460937500000F),
	Asura_Vector_3(-0.156127929687500F, -0.838195800781250F, 0.522460937500000F),
	Asura_Vector_3(0.061584472656250F, -0.865539550781250F, 0.496948242187500F),
	Asura_Vector_3(0.276367187500000F, -0.850646972656250F, 0.447204589843750F),
	Asura_Vector_3(0.289794921875000F, -0.210510253906250F, 0.933593750000000F),
	Asura_Vector_3(0.357116699218750F, -0.411865234375000F, 0.838317871093750F),
	Asura_Vector_3(0.502075195312500F, -0.212341308593750F, 0.838317871093750F),
	Asura_Vector_3(0.415771484375000F, -0.591430664062500F, 0.690856933593750F),
	Asura_Vector_3(0.569152832031250F, -0.413513183593750F, 0.710571289062500F),
	Asura_Vector_3(0.690979003906250F, -0.212646484375000F, 0.690856933593750F),
	Asura_Vector_3(0.458862304687500F, -0.736450195312500F, 0.496948242187500F),
	Asura_Vector_3(0.619018554687500F, -0.586303710937500F, 0.522460937500000F),
	Asura_Vector_3(0.748901367187500F, -0.407531738281250F, 0.522460937500000F),
	Asura_Vector_3(0.842224121093750F, -0.208862304687500F, 0.496948242187500F),
	Asura_Vector_3(0.952026367187500F, -0.129089355468750F, 0.277343750000000F),
	Asura_Vector_3(0.952026367187500F, 0.129089355468750F, 0.277343750000000F),
	Asura_Vector_3(0.963134765625000F, -0.251831054687500F, 0.093933105468750F),
	Asura_Vector_3(0.995239257812500F, 0.000000000000000F, 0.097106933593750F),
	Asura_Vector_3(0.963134765625000F, 0.251831054687500F, 0.093933105468750F),
	Asura_Vector_3(0.927246093750000F, -0.362365722656250F, -0.093933105468750F),
	Asura_Vector_3(0.987304687500000F, -0.123229980468750F, -0.100036621093750F),
	Asura_Vector_3(0.987304687500000F, 0.123229980468750F, -0.100036621093750F),
	Asura_Vector_3(0.927246093750000F, 0.362365722656250F, -0.093933105468750F),
	Asura_Vector_3(0.846069335937500F, -0.455139160156250F, -0.277343750000000F),
	Asura_Vector_3(0.923828125000000F, -0.234069824218750F, -0.302856445312500F),
	Asura_Vector_3(0.950195312500000F, 0.000000000000000F, -0.311462402343750F),
	Asura_Vector_3(0.923828125000000F, 0.234069824218750F, -0.302856445312500F),
	Asura_Vector_3(0.846069335937500F, 0.455139160156250F, -0.277343750000000F),
	Asura_Vector_3(0.723571777343750F, -0.525695800781250F, -0.447204589843750F),
	Asura_Vector_3(0.804138183593750F, -0.326049804687500F, -0.496948242187500F),
	Asura_Vector_3(0.845397949218750F, -0.110473632812500F, -0.522460937500000F),
	Asura_Vector_3(0.845397949218750F, 0.110473632812500F, -0.522460937500000F),
	Asura_Vector_3(0.804138183593750F, 0.326049804687500F, -0.496948242187500F),
	Asura_Vector_3(0.723571777343750F, 0.525695800781250F, -0.447204589843750F),
	Asura_Vector_3(0.416931152343750F, 0.865539550781250F, 0.277343750000000F),
	Asura_Vector_3(0.171386718750000F, 0.945312500000000F, 0.277343750000000F),
	Asura_Vector_3(0.537170410156250F, 0.838195800781250F, 0.093933105468750F),
	Asura_Vector_3(0.307556152343750F, 0.946533203125000F, 0.097106933593750F),
	Asura_Vector_3(0.058044433593750F, 0.993835449218750F, 0.093933105468750F),
	Asura_Vector_3(0.631164550781250F, 0.769897460937500F, -0.093933105468750F),
	Asura_Vector_3(0.422302246093750F, 0.900878906250000F, -0.100036621093750F),
	Asura_Vector_3(0.187805175781250F, 0.977050781250000F, -0.100036621093750F),
	Asura_Vector_3(-0.058044433593750F, 0.993835449218750F, -0.093933105468750F),
	Asura_Vector_3(0.694335937500000F, 0.664001464843750F, -0.277343750000000F),
	Asura_Vector_3(0.508056640625000F, 0.806274414062500F, -0.302856445312500F),
	Asura_Vector_3(0.293640136718750F, 0.903686523437500F, -0.311462402343750F),
	Asura_Vector_3(0.062805175781250F, 0.950927734375000F, -0.302856445312500F),
	Asura_Vector_3(-0.171386718750000F, 0.945312500000000F, -0.277343750000000F),
	Asura_Vector_3(0.558593750000000F, 0.664001464843750F, -0.496948242187500F),
	Asura_Vector_3(0.366333007812500F, 0.769897460937500F, -0.522460937500000F),
	Asura_Vector_3(0.156127929687500F, 0.838195800781250F, -0.522460937500000F),
	Asura_Vector_3(-0.061584472656250F, 0.865539550781250F, -0.496948242187500F),
	Asura_Vector_3(-0.276367187500000F, 0.850646972656250F, -0.447204589843750F),
	Asura_Vector_3(-0.694335937500000F, 0.664001464843750F, 0.277343750000000F),
	Asura_Vector_3(-0.846069335937500F, 0.455139160156250F, 0.277343750000000F),
	Asura_Vector_3(-0.631164550781250F, 0.769897460937500F, 0.093933105468750F),
	Asura_Vector_3(-0.805175781250000F, 0.584960937500000F, 0.097106933593750F),
	Asura_Vector_3(-0.927246093750000F, 0.362365722656250F, 0.093933105468750F),
	Asura_Vector_3(-0.537170410156250F, 0.838195800781250F, -0.093933105468750F),
	Asura_Vector_3(-0.726257324218750F, 0.680053710937500F, -0.100036621093750F),
	Asura_Vector_3(-0.871215820312500F, 0.480529785156250F, -0.100036621093750F),
	Asura_Vector_3(-0.963134765625000F, 0.251831054687500F, -0.093933105468750F),
	Asura_Vector_3(-0.416931152343750F, 0.865539550781250F, -0.277343750000000F),
	Asura_Vector_3(-0.609802246093750F, 0.732360839843750F, -0.302856445312500F),
	Asura_Vector_3(-0.768737792968750F, 0.558532714843750F, -0.311462402343750F),
	Asura_Vector_3(-0.884948730468750F, 0.353637695312500F, -0.302856445312500F),
	Asura_Vector_3(-0.952026367187500F, 0.129089355468750F, -0.277343750000000F),
	Asura_Vector_3(-0.458862304687500F, 0.736450195312500F, -0.496948242187500F),
	Asura_Vector_3(-0.619018554687500F, 0.586303710937500F, -0.522460937500000F),
	Asura_Vector_3(-0.748901367187500F, 0.407531738281250F, -0.522460937500000F),
	Asura_Vector_3(-0.842224121093750F, 0.208862304687500F, -0.496948242187500F),
	Asura_Vector_3(-0.894409179687500F, 0.000000000000000F, -0.447204589843750F),
	Asura_Vector_3(-0.846069335937500F, -0.455139160156250F, 0.277343750000000F),
	Asura_Vector_3(-0.694335937500000F, -0.664001464843750F, 0.277343750000000F),
	Asura_Vector_3(-0.927246093750000F, -0.362365722656250F, 0.093933105468750F),
	Asura_Vector_3(-0.805175781250000F, -0.584960937500000F, 0.097106933593750F),
	Asura_Vector_3(-0.631164550781250F, -0.769897460937500F, 0.093933105468750F),
	Asura_Vector_3(-0.963134765625000F, -0.251831054687500F, -0.093933105468750F),
	Asura_Vector_3(-0.871215820312500F, -0.480529785156250F, -0.100036621093750F),
	Asura_Vector_3(-0.726257324218750F, -0.680053710937500F, -0.100036621093750F),
	Asura_Vector_3(-0.537170410156250F, -0.838195800781250F, -0.093933105468750F),
	Asura_Vector_3(-0.952026367187500F, -0.129089355468750F, -0.277343750000000F),
	Asura_Vector_3(-0.884948730468750F, -0.353637695312500F, -0.302856445312500F),
	Asura_Vector_3(-0.768737792968750F, -0.558532714843750F, -0.311462402343750F),
	Asura_Vector_3(-0.609802246093750F, -0.732360839843750F, -0.302856445312500F),
	Asura_Vector_3(-0.416931152343750F, -0.865539550781250F, -0.277343750000000F),
	Asura_Vector_3(-0.842224121093750F, -0.208862304687500F, -0.496948242187500F),
	Asura_Vector_3(-0.748901367187500F, -0.407531738281250F, -0.522460937500000F),
	Asura_Vector_3(-0.619018554687500F, -0.586303710937500F, -0.522460937500000F),
	Asura_Vector_3(-0.458862304687500F, -0.736450195312500F, -0.496948242187500F),
	Asura_Vector_3(-0.276367187500000F, -0.850646972656250F, -0.447204589843750F),
	Asura_Vector_3(0.171386718750000F, -0.945312500000000F, 0.277343750000000F),
	Asura_Vector_3(0.416931152343750F, -0.865539550781250F, 0.277343750000000F),
	Asura_Vector_3(0.058044433593750F, -0.993835449218750F, 0.093933105468750F),
	Asura_Vector_3(0.307556152343750F, -0.946533203125000F, 0.097106933593750F),
	Asura_Vector_3(0.537170410156250F, -0.838195800781250F, 0.093933105468750F),
	Asura_Vector_3(-0.058044433593750F, -0.993835449218750F, -0.093933105468750F),
	Asura_Vector_3(0.187805175781250F, -0.977050781250000F, -0.100036621093750F),
	Asura_Vector_3(0.422302246093750F, -0.900878906250000F, -0.100036621093750F),
	Asura_Vector_3(0.631164550781250F, -0.769897460937500F, -0.093933105468750F),
	Asura_Vector_3(-0.171386718750000F, -0.945312500000000F, -0.277343750000000F),
	Asura_Vector_3(0.062805175781250F, -0.950927734375000F, -0.302856445312500F),
	Asura_Vector_3(0.293640136718750F, -0.903686523437500F, -0.311462402343750F),
	Asura_Vector_3(0.508056640625000F, -0.806274414062500F, -0.302856445312500F),
	Asura_Vector_3(0.694335937500000F, -0.664001464843750F, -0.277343750000000F),
	Asura_Vector_3(-0.061584472656250F, -0.865539550781250F, -0.496948242187500F),
	Asura_Vector_3(0.156127929687500F, -0.838195800781250F, -0.522460937500000F),
	Asura_Vector_3(0.366333007812500F, -0.769897460937500F, -0.522460937500000F),
	Asura_Vector_3(0.558593750000000F, -0.664001464843750F, -0.496948242187500F),
	Asura_Vector_3(0.805175781250000F, 0.584960937500000F, -0.097106933593750F),
	Asura_Vector_3(0.726257324218750F, 0.680053710937500F, 0.100036621093750F),
	Asura_Vector_3(0.871215820312500F, 0.480529785156250F, 0.100036621093750F),
	Asura_Vector_3(0.609802246093750F, 0.732360839843750F, 0.302856445312500F),
	Asura_Vector_3(0.768737792968750F, 0.558532714843750F, 0.311462402343750F),
	Asura_Vector_3(0.884948730468750F, 0.353637695312500F, 0.302856445312500F),
	Asura_Vector_3(-0.307556152343750F, 0.946533203125000F, -0.097106933593750F),
	Asura_Vector_3(-0.422302246093750F, 0.900878906250000F, 0.100036621093750F),
	Asura_Vector_3(-0.187805175781250F, 0.977050781250000F, 0.100036621093750F),
	Asura_Vector_3(-0.508056640625000F, 0.806274414062500F, 0.302856445312500F),
	Asura_Vector_3(-0.293640136718750F, 0.903686523437500F, 0.311462402343750F),
	Asura_Vector_3(-0.062805175781250F, 0.950927734375000F, 0.302856445312500F),
	Asura_Vector_3(-0.995239257812500F, 0.000000000000000F, -0.097106933593750F),
	Asura_Vector_3(-0.987304687500000F, -0.123229980468750F, 0.100036621093750F),
	Asura_Vector_3(-0.987304687500000F, 0.123229980468750F, 0.100036621093750F),
	Asura_Vector_3(-0.923828125000000F, -0.234069824218750F, 0.302856445312500F),
	Asura_Vector_3(-0.950195312500000F, 0.000000000000000F, 0.311462402343750F),
	Asura_Vector_3(-0.923828125000000F, 0.234069824218750F, 0.302856445312500F),
	Asura_Vector_3(-0.307556152343750F, -0.946533203125000F, -0.097106933593750F),
	Asura_Vector_3(-0.187805175781250F, -0.977050781250000F, 0.100036621093750F),
	Asura_Vector_3(-0.422302246093750F, -0.900878906250000F, 0.100036621093750F),
	Asura_Vector_3(-0.062805175781250F, -0.950927734375000F, 0.302856445312500F),
	Asura_Vector_3(-0.293640136718750F, -0.903686523437500F, 0.311462402343750F),
	Asura_Vector_3(-0.508056640625000F, -0.806274414062500F, 0.302856445312500F),
	Asura_Vector_3(0.805175781250000F, -0.584960937500000F, -0.097106933593750F),
	Asura_Vector_3(0.871215820312500F, -0.480529785156250F, 0.100036621093750F),
	Asura_Vector_3(0.726257324218750F, -0.680053710937500F, 0.100036621093750F),
	Asura_Vector_3(0.884948730468750F, -0.353637695312500F, 0.302856445312500F),
	Asura_Vector_3(0.768737792968750F, -0.558532714843750F, 0.311462402343750F),
	Asura_Vector_3(0.609802246093750F, -0.732360839843750F, 0.302856445312500F),
	Asura_Vector_3(0.000000000000000F, 0.000000000000000F, -1.000000000000000F),
	Asura_Vector_3(-0.067810058593750F, 0.208862304687500F, -0.975524902343750F),
	Asura_Vector_3(0.177673339843750F, 0.129089355468750F, -0.975524902343750F),
	Asura_Vector_3(-0.132385253906250F, 0.407531738281250F, -0.903503417968750F),
	Asura_Vector_3(0.110656738281250F, 0.340698242187500F, -0.933593750000000F),
	Asura_Vector_3(0.346679687500000F, 0.251831054687500F, -0.903503417968750F),
	Asura_Vector_3(-0.190490722656250F, 0.586303710937500F, -0.787353515625000F),
	Asura_Vector_3(0.046813964843750F, 0.543151855468750F, -0.838317871093750F),
	Asura_Vector_3(0.281311035156250F, 0.466918945312500F, -0.838317871093750F),
	Asura_Vector_3(0.498718261718750F, 0.362365722656250F, -0.787353515625000F),
	Asura_Vector_3(-0.239257812500000F, 0.736450195312500F, -0.632690429687500F),
	Asura_Vector_3(-0.011230468750000F, 0.722839355468750F, -0.690856933593750F),
	Asura_Vector_3(0.217407226562500F, 0.669128417968750F, -0.710571289062500F),
	Asura_Vector_3(0.433959960937500F, 0.578186035156250F, -0.690856933593750F),
	Asura_Vector_3(0.626464843750000F, 0.455139160156250F, -0.632690429687500F),
	Asura_Vector_3(-0.219604492187500F, 0.000000000000000F, -0.975524902343750F),
	Asura_Vector_3(-0.428466796875000F, 0.000000000000000F, -0.903503417968750F),
	Asura_Vector_3(-0.289794921875000F, 0.210510253906250F, -0.933593750000000F),
	Asura_Vector_3(-0.616455078125000F, 0.000000000000000F, -0.787353515625000F),
	Asura_Vector_3(-0.502075195312500F, 0.212341308593750F, -0.838317871093750F),
	Asura_Vector_3(-0.357116699218750F, 0.411865234375000F, -0.838317871093750F),
	Asura_Vector_3(-0.774353027343750F, 0.000000000000000F, -0.632690429687500F),
	Asura_Vector_3(-0.690979003906250F, 0.212646484375000F, -0.690856933593750F),
	Asura_Vector_3(-0.569152832031250F, 0.413513183593750F, -0.710571289062500F),
	Asura_Vector_3(-0.415771484375000F, 0.591430664062500F, -0.690856933593750F),
	Asura_Vector_3(-0.067810058593750F, -0.208862304687500F, -0.975524902343750F),
	Asura_Vector_3(-0.132385253906250F, -0.407531738281250F, -0.903503417968750F),
	Asura_Vector_3(-0.289794921875000F, -0.210510253906250F, -0.933593750000000F),
	Asura_Vector_3(-0.190490722656250F, -0.586303710937500F, -0.787353515625000F),
	Asura_Vector_3(-0.357116699218750F, -0.411865234375000F, -0.838317871093750F),
	Asura_Vector_3(-0.502075195312500F, -0.212341308593750F, -0.838317871093750F),
	Asura_Vector_3(-0.239257812500000F, -0.736450195312500F, -0.632690429687500F),
	Asura_Vector_3(-0.415771484375000F, -0.591430664062500F, -0.690856933593750F),
	Asura_Vector_3(-0.569152832031250F, -0.413513183593750F, -0.710571289062500F),
	Asura_Vector_3(-0.690979003906250F, -0.212646484375000F, -0.690856933593750F),
	Asura_Vector_3(0.177673339843750F, -0.129089355468750F, -0.975524902343750F),
	Asura_Vector_3(0.346679687500000F, -0.251831054687500F, -0.903503417968750F),
	Asura_Vector_3(0.110656738281250F, -0.340698242187500F, -0.933593750000000F),
	Asura_Vector_3(0.498718261718750F, -0.362365722656250F, -0.787353515625000F),
	Asura_Vector_3(0.281311035156250F, -0.466918945312500F, -0.838317871093750F),
	Asura_Vector_3(0.046813964843750F, -0.543151855468750F, -0.838317871093750F),
	Asura_Vector_3(0.626464843750000F, -0.455139160156250F, -0.632690429687500F),
	Asura_Vector_3(0.433959960937500F, -0.578186035156250F, -0.690856933593750F),
	Asura_Vector_3(0.217407226562500F, -0.669128417968750F, -0.710571289062500F),
	Asura_Vector_3(-0.011230468750000F, -0.722839355468750F, -0.690856933593750F),
	Asura_Vector_3(0.358215332031250F, 0.000000000000000F, -0.933593750000000F),
	Asura_Vector_3(0.531005859375000F, 0.123229980468750F, -0.838317871093750F),
	Asura_Vector_3(0.531005859375000F, -0.123229980468750F, -0.838317871093750F),
	Asura_Vector_3(0.684020996093750F, 0.234069824218750F, -0.690856933593750F),
	Asura_Vector_3(0.703552246093750F, 0.000000000000000F, -0.710571289062500F),
	Asura_Vector_3(0.684020996093750F, -0.234069824218750F, -0.690856933593750F),
	Asura_Vector_3(0.0f,-1.0f,0.0f),
	Asura_Vector_3(0.0f,1.0f,0.0f),
	Asura_Vector_3(1.0f,0.0f,0.0f)
};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif  //#else not ASURA_DISABLE_AI_CODE
