//*******************************************************************
//
// File: Asura_Affector_System.h                    
//
// Created: 9 Nov 2006
//
// Description: 
//
//*******************************************************************

#ifndef _ASURA_AFFECTOR_SYSTEM_H_
#define _ASURA_AFFECTOR_SYSTEM_H_

#if defined( ASURA_USE_AFFECTORS) && 1 == ASURA_USE_AFFECTORS

// other libraries
#include "Core/Asura_Handle.h"
#include "Affectors/Asura_Affector.h"
#include "Affectors/Asura_Affector_Global.h"
#include "SpecialFX/Asura_PerlinNoise_System.h"

// internal dependencies

// forward declarations
class Asura_Affector;
class Asura_Affector_Global;
template<typename T> class Asura_TrackingHandlePool;

//*******************************************************************
// Constants
//*******************************************************************


//*******************************************************************
// Class Definitions
//*******************************************************************
class Asura_Affector_System
{
public:
	static void Initialise(void);
	static void Reset(void);
	static void Update(void);
	static void Shutdown(void);

	// Global affector functions...
	static Asura_Handle				CreateGlobalAffector(void);
	static void						DestroyGlobalAffector(const Asura_Handle iHandle);
	static Asura_Affector_Global*	GetGlobalAffectorPointer(const Asura_Handle iHandle);

	// Local affector functions...
	static Asura_Handle		CreateAffector(void);
	static void				DestroyAffector(const Asura_Handle iHandle);
	static Asura_Affector*	GetAffectorPointer(const Asura_Handle iHandle);

	// Function to create a list of affectors which intersect the specified bounding-box...
	static void MakeAffectorListForBoundingBox(const Asura_Bounding_Box& xBoundingBox, const u_int uAffectorArraySizeMax, Asura_Affector** ppxAffectors, u_int& uAffectorArraySize) ASURA_CODE_SECTION("affector_list");

	// ART - 08/12/2006 - Moved this code into a common place to prevent it from being 
	// duplicated in every place it's needed...
	static Asura_AVector3 CalculateForceAtPosition(const Asura_AVector3& xPosition, const float fRadius, const Asura_Affector* const* ppxAffectors, const u_int uNumAffectors);
	static Asura_AVector3 CalculateLocalAffectorForceAtPosition(const Asura_AVector3& xPosition, const float fRadius, const Asura_Affector* const* ppxAffectors, const u_int uNumAffectors);
	static Asura_AVector3 CalculateGlobalAffectorForceAtPosition(const Asura_AVector3& xPosition);

	// Accessor functions...
	static bool	HasActiveAffectors(void)		{ return(s_uNumActiveAffectors > 0); }
	static bool	HasActiveLocalAffectors(void)	{ return(s_uNumActiveLocalAffectors > 0); }
	static bool	HasActiveGlobalAffectors(void)	{ return(s_uNumActiveGlobalAffectors > 0); }


private:
	static Asura_TrackingHandlePool<Asura_Affector>*		s_pxDynamicAffectorList;
	static Asura_TrackingHandlePool<Asura_Affector_Global>*	s_pxDynamicGlobalAffectorList;

	static float											s_fGlobalAffectorStrength;
	static float											s_fGlobalAffectorTime;

	static u_int											s_uNumActiveAffectors;
	static u_int											s_uNumActiveGlobalAffectors;
	static u_int											s_uNumActiveLocalAffectors;
};



//*******************************************************************
// Inline Function Definitions
//*******************************************************************

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Affector_System::CalculateLocalAffectorForceAtPosition
// Author       : DLH
// Date         : 12 December 2006
////////////////////////////////////////////////////////////////////////////////
inline Asura_AVector3 Asura_Affector_System::CalculateLocalAffectorForceAtPosition(const Asura_AVector3& xPosition, const float fRadius, const Asura_Affector* const* ppxAffectors, const u_int uNumAffectors)
{
	Asura_AVector3 xForce;

	xForce.Zero();

	// Now sum up the affectors which apply to this position...
	for( u_int uAffectorIndex = 0; uAffectorIndex < uNumAffectors; uAffectorIndex++)
	{
		const Asura_Affector*	pxAffector			= ppxAffectors[uAffectorIndex];
		const Asura_AVector3&	xAffectorPosition	= pxAffector->GetPositionA();
		const Asura_AVector3	xAffectorToInstance(xPosition - xAffectorPosition);
		Asura_AVector3			xAffectorForce(xAffectorToInstance);

		const float				fAffectorToInstanceDistance		= xAffectorForce.NormaliseReturningMagnitude();
		const float				fAffectorRange					= pxAffector->GetRange();
		const float				fAffectorToInstanceMaximumRange	= fAffectorRange + fRadius;

		if(fAffectorToInstanceMaximumRange >= fAffectorToInstanceDistance)
		{
			const float	fAffectorDistanceStrength		= 1.0f - (fAffectorToInstanceDistance / fAffectorRange);
			const float	fAffectorDistanceStrengthScale	= Asura_Maths::Max(0.0f, fAffectorDistanceStrength);

			xAffectorForce	*= fAffectorDistanceStrengthScale * pxAffector->GetStrength();
			xForce			+= xAffectorForce;
		}
	}

	return( xForce);
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Affector_System::CalculateGlobalAffectorForceAtPosition
// Author       : DLH
// Date         : 12 December 2006
////////////////////////////////////////////////////////////////////////////////
inline Asura_AVector3 Asura_Affector_System::CalculateGlobalAffectorForceAtPosition(const Asura_AVector3& xPosition)
{
	Asura_AVector3 xForce;

	// Ensure that the base force includes any Perlin noise we may need...
	if(s_fGlobalAffectorStrength == 0.0f)
	{
		xForce.Zero();
	}
	else
	{
		Asura_PerlinNoise_System::CalculateNoiseVector(xPosition, s_fGlobalAffectorTime, xForce);
		xForce	*= s_fGlobalAffectorStrength;
	}

	return( xForce);
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Affector_System::CalculateForceAtPosition
// Author       : Andy Tate
// Date         : 8-Dec-2006
////////////////////////////////////////////////////////////////////////////////

// DLH - 12 December 2006 08:37:13 - 
inline Asura_AVector3 Asura_Affector_System::CalculateForceAtPosition (const Asura_AVector3& xPosition,
																const float fRadius,
																const Asura_Affector* const* ppxAffectors,
																const u_int uNumAffectors)
{
	Asura_AVector3 xForce;

	xForce = CalculateGlobalAffectorForceAtPosition( xPosition );
	xForce += CalculateLocalAffectorForceAtPosition( xPosition, fRadius, ppxAffectors, uNumAffectors);

	return( xForce);
}

/// End of function 'Asura_Affector_System::CalculateForceAtPosition'. ///


//*******************************************************************
//
// (c) 2006 Rebellion
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************

#endif //defined( ASURA_USE_AFFECTORS) && 1 == ASURA_USE_AFFECTORS

#endif // _ASURA_AFFECTOR_SYSTEM_H_


/// End of file. ///
