
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_AnimationBlend.cpp
// Author : Andy Tate
// Date   : 21-Sep-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Animation/Asura_AnimationBlend.h"

// Asura independent header-files...
#include "Animation/Asura_AnimationController_Group.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h" 
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
static const char* s_aszTweenStateString[] =
{
	"STATE_OUT",
	"STATE_TWEEN_IN",
	"STATE_IN",
	"STATE_TWEEN_OUT"
};
#endif

////////////////////////////////////////////////////////////////////////////////
// Defines...
////////////////////////////////////////////////////////////////////////////////

#if 1
#define LOG_OUTPUT(s)
#else
#define LOG_OUTPUT(s)	Asura_OutputToDebugger_Retail(s)
#endif


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::ChangeState
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationBlend_Tween::ChangeState (const State eNewState)
{
	m_bHasNewState	= true;
	m_eNewState		= eNewState;
	return(true);
}

/// End of function 'Asura_AnimationBlend_Tween::ChangeState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::BeginNewState
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationBlend_Tween::BeginNewState (const State eNewState,
												Asura_AnimationController& xPreviousController)
{
	// Decide what to do based on the destination state...
	switch(eNewState)
	{
		case STATE_OUT:
		{
			m_pxNextController->HandleMessage(Asura_AnimationController::DESTROY_MSG);
			return(true);
		}

		case STATE_TWEEN_IN:
		{
			m_pxNextController->HandleMessage(Asura_AnimationController::PAUSE_MSG);
			return(true);
		}

		case STATE_IN:
		{
			// Check if we should unpause the new anim (happens if this was a synced blend)
			if ( m_bPauseOnStateIn )
			{
				m_pxNextController->HandleMessage(Asura_AnimationController::PAUSE_MSG);
			}
			else if ( ( m_bSyncToNext || m_bPlayOnStateIn ) )
			{
				m_pxNextController->HandleMessage(Asura_AnimationController::PLAY_MSG);
			}

			if(!m_pxNextController->IsUsingSubsetOfBones())
			{
				xPreviousController.HandleMessage(Asura_AnimationController::DESTROY_FROM_TWEEN_IN_MSG);
			}
			return(true);
		}

		case STATE_TWEEN_OUT:
		{
			return(true);
		}

		default:
		{
			Asura_Assert(false, "Unexpected state for tween.");
			return(false);
		}
	}	// End of switch.
}

/// End of function 'Asura_AnimationBlend_Tween::BeginNewState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::Asura_AnimationBlend_AnimatedTween
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationBlend_AnimatedTween::BeginNewState (const State eNewState,
														Asura_AnimationController& xPreviousController)
{
	// Decide what to do based on the destination state...
	switch(eNewState)
	{
		case STATE_OUT:
		{
			m_pxNextController->HandleMessage(Asura_AnimationController::DESTROY_MSG);
			return(true);
		}

		case STATE_TWEEN_IN:
		{
			m_pxNextController->HandleMessage(Asura_AnimationController::PLAY_MSG);
			return(true);
		}

		case STATE_IN:
		{
			if ( m_bPauseOnStateIn )
			{
				m_pxNextController->HandleMessage(Asura_AnimationController::PAUSE_MSG);
			}
			else if ( ( m_bSyncToNext || m_bPlayOnStateIn ) )
			{
				m_pxNextController->HandleMessage(Asura_AnimationController::PLAY_MSG);
			}

			if(!m_pxNextController->IsUsingSubsetOfBones())
			{
				xPreviousController.HandleMessage(Asura_AnimationController::DESTROY_FROM_TWEEN_IN_MSG);
			}
			return(true);
		}

		case STATE_TWEEN_OUT:
		{
			return(true);
		}

		default:
		{
			Asura_Assert(false, "Unexpected state for tween.");
			return(false);
		}
	}	// End of switch.
}

/// End of function 'Asura_AnimationBlend_AnimatedTween::BeginNewState'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend::~Asura_AnimationBlend
// Author       : Andy Tate
// Date         : 22-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationBlend::~Asura_AnimationBlend ()
{
	// Destroy any connected controllers...
	delete(m_pxNextController);
}

/// End of function 'Asura_AnimationBlend::~Asura_AnimationBlend'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend::Update
// Author       : Anthony Gregson
// Date         : 21-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationBlend::Update (const float fTimeStep,
										 Asura_AnimationController& xPreviousController)
{
	// Do we need to sync our animations?
	if(m_bSyncToNext)
	{
		Asura_AnimationMessage_SyncTimerTo	xSyncTimerTo;

		// Yes - sync the next controller to the previous controller...
		xSyncTimerTo.m_fTimeStep		= fTimeStep;
		xSyncTimerTo.m_uToAnimationHash	= xPreviousController.GetControllerHash();
		m_pxNextController->HandleMessage(Asura_AnimationController::SYNC_TIMER_TO_MSG, reinterpret_cast<void*>(&xSyncTimerTo));
	}
}

/// End of function 'Asura_AnimationBlend::Update'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend::PrintDebugInfo
// Author       : KOR
// Date         : 21/02/2007
////////////////////////////////////////////////////////////////////////////////

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

void Asura_AnimationBlend::PrintDebugInfo(u_int& uRow,
										  u_int& uCol,
										  const Asura_AnimationController_Group* pxParentGroup /*= NULL*/,
										  const char* szFilename /*= NULL*/) const
{
	const bool	bBlend	= !m_pxNextController || !m_pxNextController->AreAllBonesUsingDeltas() || ( m_pxPreviousController && m_pxPreviousController->AreAllBonesUsingDeltas() && ( pxParentGroup->GetDeltasUseBlends() || GetDeltasUseBlends() ) );
	const char*	szType	= (bBlend) ? "Blend" : "Add";
	if ( szFilename == NULL )
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "> %s, Weight = %1.2f", szType, Asura_Maths::FtoD(m_fWeight) ));
	}
	else
	{
		for ( u_int uItr = 0; uItr < uCol; ++uItr )
		{
			Asura_SimpleLogfileOutput(( szFilename, " " ));
		}
		Asura_SimpleLogfileOutput(( szFilename, "> %s, Weight = %1.2f\n", szType, Asura_Maths::FtoD(m_fWeight) ));
	}
}

#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)


void Asura_AnimationBlend::SetWeight( const float fWeight )
{
    Asura_Assertf( ((fWeight >= 0.0f) && (fWeight <= 1.0001f)), ("Bad weight (%f) specified for blend.", Asura_Maths::FtoD(fWeight)) );
    m_fWeight = fWeight;
}

Asura_AnimationBlend* Asura_AnimationBlend::Duplicate() const
{
	Asura_AnimationBlend* pxNewBlend = new Asura_AnimationBlend(m_fWeight);
	pxNewBlend->m_uBlendFlags = m_uBlendFlags;
	return pxNewBlend;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::Asura_AnimationBlend_Tween
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationBlend_Tween::Asura_AnimationBlend_Tween (const float fTweenInDuration,
														const bool bSyncToNext)
: Asura_AnimationBlend(0.0f, bSyncToNext)
, m_uTweenFlags(0)
, m_fTimer(0.0f)
, m_fTweenInDuration(fTweenInDuration)
, m_fTweenOutDuration(0.0f)
{
	// Begin the tween...
	ChangeState(STATE_TWEEN_IN);
}

/// End of function 'Asura_AnimationBlend_Tween::Asura_AnimationBlend_Tween'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::Asura_AnimationBlend_Tween
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationBlend_Tween::Asura_AnimationBlend_Tween (const float fTweenInDuration,
														const float fTweenOutDuration,
														const bool bSyncToNext)
: Asura_AnimationBlend(0.0f, bSyncToNext)
, m_uTweenFlags(0)
, m_fTimer(0.0f)
, m_fTweenInDuration(fTweenInDuration)
, m_fTweenOutDuration(fTweenOutDuration)
{
	// Begin the tween...
	ChangeState(STATE_TWEEN_IN);
}

/// End of function 'Asura_AnimationBlend_Tween::Asura_AnimationBlend_Tween'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::Update
// Author       : Andy Tate
// Date         : 21-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationBlend_Tween::Update (const float fTimeStep,
										 Asura_AnimationController& xPreviousController)
{
	float fTimeLeft = fTimeStep;
	while ( fTimeLeft > 0.0f || m_bHasNewState )
	{
		float fTimeSpentThisLoop = 0.0f;
		// Do we need to change state?
		if(m_bHasNewState && BeginNewState(static_cast<State>(m_eNewState), xPreviousController))
		{
			m_eState		= m_eNewState;
			m_bHasNewState	= false;
			m_fTimer		= 0.0f;
		}

		// Parent now does the syncing, so call that now
		Asura_AnimationBlend::Update(fTimeStep, xPreviousController);

		// What state are we in?
		switch(m_eState)
		{
			case STATE_OUT:
			{
				fTimeSpentThisLoop = fTimeLeft;
				break;
			}

			case STATE_TWEEN_IN:
			{
				float fFinalTime = m_fTimer + fTimeLeft;
				// Update the blend-weight...
				Asura_Assert(m_fTimer >= 0.0f, "timer went negative");
				if( fFinalTime < m_fTweenInDuration)
				{
					Asura_Assert(m_fTweenInDuration > 0.0f, "invalid tween duration");
					if (m_fTweenInDuration > 0.0f)
					{
						UpdateWeightFromTween( fFinalTime / m_fTweenInDuration );
					}
					fTimeSpentThisLoop = fTimeLeft;
				}
				else
				{
					UpdateWeightFromTween(1.0f);
					fTimeSpentThisLoop = m_fTweenInDuration - m_fTimer;
					ChangeState(STATE_IN);
				}
				break;
			}

			case STATE_IN:
			{
				const float	fTimeUntilFinished	= m_pxNextController->GetTimeUntilFinished();
				fTimeSpentThisLoop = fTimeLeft;

				// Has our controller finished?
				if ( fTimeUntilFinished < fTimeLeft )
				{
					if (m_fTweenOutDuration > 0.0f) // Zero tween out time -> don't tween out (or even remove this animation).
					{
						fTimeSpentThisLoop -= ( fTimeLeft - fTimeUntilFinished ); // i.e. spend this much time in STATE_TWEEN_OUT this frame
						ChangeState(STATE_TWEEN_OUT);
					}
				}
				break;
			}

			case STATE_TWEEN_OUT:
			{
				if (m_fTweenOutDuration > 0.0f)
				{
					if ( m_fTimer + fTimeLeft < m_fTweenOutDuration )
					{
						UpdateWeightFromTween( 1.0f - ( ( m_fTimer + fTimeLeft ) / m_fTweenOutDuration ) );
						fTimeSpentThisLoop = fTimeLeft;
					}
					else
					{
						UpdateWeightFromTween( 0.0f );
						fTimeSpentThisLoop = m_fTweenOutDuration - m_fTimer;
						ChangeState(STATE_OUT);
					}
				}
				else
				{
					UpdateWeightFromTween(0.0f);
					fTimeSpentThisLoop = fTimeLeft;
					ChangeState(STATE_OUT);
				}
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected state for tween.");
				break;
			}
		}		// End of switch.

		m_fTimer += fTimeSpentThisLoop;
		fTimeLeft -= fTimeSpentThisLoop;
	}

	LOG_OUTPUT(("Tween: %08X - State = %u, fTimeStep = %f, m_fWeight = %f\n", this, m_eState, fTimeStep, m_fWeight));
}

/// End of function 'Asura_AnimationBlend_Tween::Update'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Tween::PrintDebugInfo
// Author       : KOR
// Date         : 21/02/2007
////////////////////////////////////////////////////////////////////////////////

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

void Asura_AnimationBlend_Tween::PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup /*= NULL*/, const char* szFilename /*= NULL*/) const
{
	const bool	bBlend	= !m_pxNextController || !m_pxNextController->AreAllBonesUsingDeltas() || ( m_pxPreviousController && m_pxPreviousController->AreAllBonesUsingDeltas() && ( pxParentGroup->GetDeltasUseBlends() || GetDeltasUseBlends() ) );
	const char*	szType	= (bBlend) ? "Blend" : "Add";
	if ( szFilename == NULL )
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "> Tween %s, Weight = %1.2f, State = %s", szType, Asura_Maths::FtoD(m_fWeight), s_aszTweenStateString[ m_eState ] ));
	}
	else
	{
		for ( u_int uItr = 0; uItr < uCol; ++uItr )
		{
			Asura_SimpleLogfileOutput(( szFilename, " " ));
		}
		Asura_SimpleLogfileOutput(( szFilename, "> Tween %s, Weight = %1.2f, State = %s\n", szType, Asura_Maths::FtoD(m_fWeight), s_aszTweenStateString[ m_eState ] ));
	}
}

#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

/// End of function 'Asura_AnimationBlend_Tween::PrintDebugInfo'. ///

void Asura_AnimationBlend_Tween::ChangeTweenDuration(const float fTweenInDuration, const float fTweenOutDuration/* = 0.0f*/)
{
	m_fTweenInDuration = fTweenInDuration;
	m_fTweenOutDuration = fTweenOutDuration;

	// We need to update the timer if we're mid-tween.
	switch (m_eState)
	{
		default:
		break;

		case STATE_TWEEN_IN:
		{
			m_fTimer = m_fWeight * m_fTweenInDuration;
		}
		break;

		case STATE_TWEEN_OUT:
		{
			m_fTimer = ( 1.0f - m_fWeight ) * m_fTweenOutDuration;
		}
		break;
	}
}

Asura_AnimationBlend* Asura_AnimationBlend_Tween::Duplicate() const
{
	Asura_AnimationBlend_Tween* pxNewBlend = new Asura_AnimationBlend_Tween(m_fTweenInDuration, m_fTweenOutDuration, m_bSyncToNext);
	pxNewBlend->m_fWeight = m_fWeight;
	pxNewBlend->m_fTimer = m_fTimer;
	pxNewBlend->m_uBlendFlags = m_uBlendFlags;
	pxNewBlend->m_uTweenFlags = m_uTweenFlags;
	return pxNewBlend;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween (const float fTweenInDuration,
																		const bool bSyncToNext)
: Asura_AnimationBlend_Tween(fTweenInDuration, bSyncToNext)
{
	// Nothing more to initialise...
}

/// End of function 'Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween
// Author       : Andy Tate
// Date         : 24-Oct-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween (const float fTweenInDuration,
																		const float fTweenOutDuration,
																		const bool bSyncToNext)
: Asura_AnimationBlend_Tween(fTweenInDuration, fTweenOutDuration, bSyncToNext)
{
	// Nothing more to initialise...
}

/// End of function 'Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween'. ///

void Asura_AnimationBlend_AnimatedTween::Update (const float fTimeStep,
												 Asura_AnimationController& xPreviousController)
{
	float fTimeLeft = fTimeStep;
	while ( fTimeLeft > 0.0f || m_bHasNewState )
	{
		float fTimeSpentThisLoop = 0.0f;
		// Do we need to change state?
		if(m_bHasNewState && BeginNewState(static_cast<State>(m_eNewState), xPreviousController))
		{
			m_eState		= m_eNewState;
			m_bHasNewState	= false;
			m_fTimer		= 0.0f;
		}

		// Parent now does the syncing, so call that now
		Asura_AnimationBlend::Update(fTimeStep, xPreviousController);

		// What state are we in?
		switch(m_eState)
		{
			case STATE_OUT:
			{
				fTimeSpentThisLoop = fTimeLeft;
				break;
			}

			case STATE_TWEEN_IN:
			{
				float fFinalTime = m_fTimer + fTimeLeft;
				// Update the blend-weight...
				Asura_Assert(m_fTimer >= 0.0f, "timer went negative");
				if( fFinalTime < m_fTweenInDuration)
				{
					Asura_Assert(m_fTweenInDuration > 0.0f, "invalid tween duration");
					if (m_fTweenInDuration > 0.0f)
					{
						UpdateWeightFromTween( fFinalTime / m_fTweenInDuration );
					}
					fTimeSpentThisLoop = fTimeLeft;
				}
				else
				{
					UpdateWeightFromTween(1.0f);
					fTimeSpentThisLoop = m_fTweenInDuration - m_fTimer;
					ChangeState(STATE_IN);
				}
				break;
			}

			case STATE_IN:
			{
				const float	fTimeUntilFinished	= m_pxNextController->GetTimeUntilFinished();

				// Has our controller finished?
				if(fTimeUntilFinished < m_fTweenOutDuration)
				{
					fTimeSpentThisLoop = Asura_Maths::Min( fTimeUntilFinished, fTimeLeft );
					ChangeState(STATE_TWEEN_OUT);
				}
				else
				{
					fTimeSpentThisLoop = fTimeLeft;
				}

				break;
			}

			case STATE_TWEEN_OUT:
			{
				// Update the blend-weight...
				Asura_Assert( m_fTimer <= m_fTweenOutDuration, "invalid timer value" );
				if( fTimeLeft < m_fTweenOutDuration - m_fTimer )
				{
					Asura_Assert(m_fTweenOutDuration > 0.0f, "invalid tween duration");
					if (m_fTweenOutDuration > 0.0f)
					{
						UpdateWeightFromTween( 1.0f - ( ( m_fTimer + fTimeLeft ) / m_fTweenOutDuration ) );
						fTimeSpentThisLoop = fTimeLeft;
					}
				}
				else
				{
					UpdateWeightFromTween(0.0f);
					fTimeSpentThisLoop = m_fTweenOutDuration - m_fTimer;
					ChangeState(STATE_OUT);
				}
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected state for tween.");
				break;
			}
		}		// End of switch.

		m_fTimer += fTimeSpentThisLoop;
		fTimeLeft -= fTimeSpentThisLoop;
	}

	LOG_OUTPUT(("Tween: %08X - State = %u, fTimeStep = %f, m_fWeight = %f\n", this, m_eState, fTimeStep, m_fWeight));
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_AnimatedTween::PrintDebugInfo
// Author       : KOR
// Date         : 21/02/2007
////////////////////////////////////////////////////////////////////////////////

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

void Asura_AnimationBlend_AnimatedTween::PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup /*= NULL*/, const char* szFilename /*= NULL*/) const
{
	const bool	bBlend	= !m_pxNextController || !m_pxNextController->AreAllBonesUsingDeltas() || ( m_pxPreviousController && m_pxPreviousController->AreAllBonesUsingDeltas() && ( pxParentGroup->GetDeltasUseBlends() || GetDeltasUseBlends() ) );
	const char*	szType	= (bBlend) ? "Blend" : "Add";
	if ( szFilename == NULL )
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "> Animated Tween %s, Weight = %1.2f, State = %s", szType, Asura_Maths::FtoD(m_fWeight), s_aszTweenStateString[ m_eState ] ));
	}
	else
	{
		for ( u_int uItr = 0; uItr < uCol; ++uItr )
		{
			Asura_SimpleLogfileOutput(( szFilename, " " ));
		}
		Asura_SimpleLogfileOutput(( szFilename, "> Animated Tween %s, Weight = %1.2f, State = %s\n", szType, Asura_Maths::FtoD(m_fWeight), s_aszTweenStateString[ m_eState ] ));
	}
}

#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

/// End of function 'Asura_AnimationBlend_AnimatedTween::PrintDebugInfo'. ///

Asura_AnimationBlend* Asura_AnimationBlend_AnimatedTween::Duplicate() const
{
	Asura_AnimationBlend_AnimatedTween* pxNewBlend = new Asura_AnimationBlend_AnimatedTween(m_fTweenInDuration, m_fTweenOutDuration, m_bSyncToNext);
	pxNewBlend->m_fWeight = m_fWeight;
	pxNewBlend->m_fTimer = m_fTimer;
	pxNewBlend->m_uBlendFlags = m_uBlendFlags;
	pxNewBlend->m_uTweenFlags = m_uTweenFlags;
	return pxNewBlend;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_TweenOut::Update
// Author       : Andy Tate
// Date         : 22-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationBlend_TweenOut::Update (const float fTimeStep,
											Asura_AnimationController& xPreviousController)
{
	// Has our controller finished?
	if(m_pxNextController->IsFinished())
	{
		// Yes - begin tweening out...
		if(m_fInitialWeight < 0.0f)
		{
			m_fInitialWeight	= m_fWeight;
		}

		// Update the time...
		m_fTimer	+= fTimeStep;

		// Update the blend-weight...
		if(m_fTimer < m_fDuration)
		{
			m_fWeight	= m_fInitialWeight *  (1 - m_fTimer / m_fDuration );
		}
		else
		{
			// We've finished the tween - destroy our animation...
			m_pxNextController->HandleMessage(Asura_AnimationController::DESTROY_MSG);
			m_fWeight	= 0.0f;
		}
	}

	LOG_OUTPUT(("TweenOut: fTimeStep = %f, m_fWeight = %f\n", fTimeStep, m_fWeight));
}

/// End of function 'Asura_AnimationBlend_TweenOut::Update'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_TweenOut::PrintDebugInfo
// Author       : KOR
// Date         : 21/02/2007
////////////////////////////////////////////////////////////////////////////////

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

void Asura_AnimationBlend_TweenOut::PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup /*= NULL*/, const char* szFilename /*= NULL*/) const
{
	const bool	bBlend	= !m_pxNextController || !m_pxNextController->AreAllBonesUsingDeltas() || ( m_pxPreviousController && m_pxPreviousController->AreAllBonesUsingDeltas() && ( pxParentGroup->GetDeltasUseBlends() || GetDeltasUseBlends() ) );
	const char*	szType	= (bBlend) ? "Blend" : "Add";
	if ( szFilename == NULL )
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "> Tween Out %s, Weight = %1.2f", szType, Asura_Maths::FtoD(m_fWeight) ));
	}
	else
	{
		for ( u_int uItr = 0; uItr < uCol; ++uItr )
		{
			Asura_SimpleLogfileOutput(( szFilename, " " ));
		}
		Asura_SimpleLogfileOutput(( szFilename, "> Tween Out %s, Weight = %1.2f\n", szType, Asura_Maths::FtoD(m_fWeight) ));
	}
}

#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

/// End of function 'Asura_AnimationBlend_TweenOut::PrintDebugInfo'. ///

Asura_AnimationBlend* Asura_AnimationBlend_TweenOut::Duplicate() const
{
	Asura_AnimationBlend_TweenOut* pxNewBlend = new Asura_AnimationBlend_TweenOut(m_fDuration);
	pxNewBlend->m_fWeight = m_fWeight;
	pxNewBlend->m_fTimer = m_fTimer;
	pxNewBlend->m_fInitialWeight = m_fInitialWeight;
	// Should always be false, but can't hurt to copy it.
	pxNewBlend->m_bSyncToNext = m_bSyncToNext;
	return pxNewBlend;
}



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Secondary::Asura_AnimationBlend_Secondary
// Author       : Richard Rice
// Date         : 25-Apr-2008
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationBlend_Secondary::Asura_AnimationBlend_Secondary ( const float fWeight,
																		const float fTweenInDuration,
																		const float fTweenOutDuration)
: Asura_AnimationBlend_AnimatedTween(fTweenInDuration, fTweenOutDuration, false)
,m_fSecondaryWeight(fWeight)
{
}

/// End of function 'Asura_AnimationBlend_AnimatedTween::Asura_AnimationBlend_AnimatedTween'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Secondary::PrintDebugInfo
// Author       : Richard Rice
// Date         : 25-Apr-2008
////////////////////////////////////////////////////////////////////////////////

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

void Asura_AnimationBlend_Secondary::PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup /*= NULL*/, const char* szFilename /*= NULL*/) const
{
	const bool	bBlend	= !m_pxNextController || !m_pxNextController->AreAllBonesUsingDeltas() || ( m_pxPreviousController && m_pxPreviousController->AreAllBonesUsingDeltas() && ( pxParentGroup->GetDeltasUseBlends() || GetDeltasUseBlends() ) );
	const char*	szType	= (bBlend) ? "Blend" : "Add";
	if ( szFilename == NULL )
	{
		Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "> Secondary Tween %s, Weight = %1.2f, State = %s", szType, Asura_Maths::FtoD(m_fWeight), s_aszTweenStateString[ m_eState ] ));
	}
	else
	{
		for ( u_int uItr = 0; uItr < uCol; ++uItr )
		{
			Asura_SimpleLogfileOutput(( szFilename, " " ));
		}
		Asura_SimpleLogfileOutput(( szFilename, "> Secondary Tween %s, Weight = %1.2f, State = %s\n", szType, Asura_Maths::FtoD(m_fWeight), s_aszTweenStateString[ m_eState ] ));
	}
}

#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)

/// End of function 'Asura_AnimationBlend_AnimatedTween::PrintDebugInfo'. ///

Asura_AnimationBlend* Asura_AnimationBlend_Secondary::Duplicate() const
{
	Asura_AnimationBlend_Secondary* pxNewBlend = new Asura_AnimationBlend_Secondary(m_fSecondaryWeight,m_fTweenInDuration, m_fTweenOutDuration);
	pxNewBlend->m_fWeight = m_fWeight;
	pxNewBlend->m_fTimer = m_fTimer;
	pxNewBlend->m_uBlendFlags = m_uBlendFlags;
	pxNewBlend->m_uTweenFlags = m_uTweenFlags;
	return pxNewBlend;
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationBlend_Secondary::Asura_AnimationBlend_BeginNewState
// Author       : Richard Rice
// Date         : 25-Apr-2008
////////////////////////////////////////////////////////////////////////////////

bool Asura_AnimationBlend_Secondary::BeginNewState (const State eNewState,
														Asura_AnimationController& xPreviousController)
{
	// Decide what to do based on the destination state...
	//Identical to animated tween, except the previous animation isn't removed when this finishes
	//tweening in
	switch(eNewState)
	{
		case STATE_OUT:
		{
			m_pxNextController->HandleMessage(Asura_AnimationController::DESTROY_MSG);
			return(true);
		}

		case STATE_TWEEN_IN:
		{
			m_pxNextController->HandleMessage(Asura_AnimationController::PLAY_MSG);
			return(true);
		}

		case STATE_IN:
		{
			return(true);
		}

		case STATE_TWEEN_OUT:
		{
			return(true);
		}

		default:
		{
			Asura_Assert(false, "Unexpected state for tween.");
			return(false);
		}
	}	// End of switch.
}


/// End of file. ///
