
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_AnimationController.cpp
// Author : Andy Tate
// Date   : 14-Sep-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Animation/Asura_AnimationController.h"

// Asura independent header-files...
#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
#include "Animation/Asura_Animation.h"
#endif
#include "Animation/Asura_AnimationBlend.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController::Asura_AnimationController
// Author       : Andy Tate
// Date         : 14-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController::Asura_AnimationController (const Asura_Hash_ID uControllerHash,
													  const u_int uInitialFlags,
													  const u_int uPriority)
: m_szName(NULL)
, m_pxAsuraAnimation(NULL)
, m_uControllerHash(uControllerHash)
, m_uParentControllerHash(ASURA_HASH_ID_UNSET)
, m_uControllerFlags(uInitialFlags)
, m_uPriority(uPriority)
, m_pxBlendToNextController(NULL)
, m_pxBlendToPreviousController(NULL)
{
	// Nothing more to initialise...
}

/// End of function 'Asura_AnimationController::Asura_AnimationController'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController::~Asura_AnimationController
// Author       : Andy Tate
// Date         : 14-Sep-2006
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController::~Asura_AnimationController ()
{
	// Destroy the next controller...
	delete(m_pxBlendToNextController);
}

/// End of function 'Asura_AnimationController::~Asura_AnimationController'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController::ReplaceBlendToNextController
// Author       : Anthony Gregson
// Date         : 11/12/06
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController::ReplaceBlendToNextController(Asura_AnimationBlend* pxNewBlend)
{
	Asura_Assert(pxNewBlend, "No replacement blend");
	if (!pxNewBlend) { return; }

	if (m_pxBlendToNextController)
	{
		// Next controller exists, replace it

		// Get the controller we're blending to
		Asura_AnimationController* pxNextController = m_pxBlendToNextController->GetNextController();
		// Unlink both controllers from the old blend, so that deleting it won't delete them too
		m_pxBlendToNextController->SetNextController(NULL);
		m_pxBlendToNextController->SetPreviousController(NULL);
		// And get rid of it
		delete m_pxBlendToNextController;

		// Link in controllers to the new blend
		pxNewBlend->SetNextController(pxNextController);
		pxNewBlend->SetPreviousController(this);
		// Link in new blend to the controllers
		SetBlendToNextController(pxNewBlend);
		pxNextController->SetBlendToPreviousController(pxNewBlend);
	}
	else
	{
		// No next controller, clean up the blend
		Asura_Assert(!pxNewBlend->GetNextController() && !pxNewBlend->GetPreviousController(), "Expected replacement blend to be unconnected to any controllers");
		delete pxNewBlend;
	}
}

/// End of function 'Asura_AnimationController::ReplaceBlendToNextController'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController::ReplaceBlendToPreviousController
// Author       : Anthony Gregson
// Date         : 11/12/06
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController::ReplaceBlendToPreviousController(Asura_AnimationBlend* pxNewBlend)
{
	if (m_pxBlendToPreviousController)
	{
		// Previous controller exists, replace it

		// Get the controller we're blending from
		Asura_AnimationController* pxPreviousController = m_pxBlendToPreviousController->GetPreviousController();
		// Unlink both controllers from the old blend, so that deleting it won't delete them too
		m_pxBlendToPreviousController->SetNextController(NULL);
		m_pxBlendToPreviousController->SetPreviousController(NULL);
		// And get rid of it
		delete m_pxBlendToPreviousController;

		// Link in controllers to the new blend
		pxNewBlend->SetNextController(this);
		pxNewBlend->SetPreviousController(pxPreviousController);
		// Link in new blend to the controllers
		pxPreviousController->SetBlendToNextController(pxNewBlend);
		SetBlendToPreviousController(pxNewBlend);
	}
	else
	{
		// No previous controller, clean up the blend
		Asura_Assert(!pxNewBlend->GetNextController() && !pxNewBlend->GetPreviousController(), "Expected replacement blend to be unconnected to any controllers");
		delete pxNewBlend;
	}
}

/// End of function 'Asura_AnimationController::ReplaceBlendToPreviousController'. ///

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_AnimationController::PrintDebugInfo
// Author       : Kim Randell
// Date         : 07/02/2007
////////////////////////////////////////////////////////////////////////////////

void Asura_AnimationController::PrintDebugInfo(u_int& uRow,
											   u_int& uCol,
											   const Asura_AnimationController_Group* pxParentGroup,
											   const char* szFilename /*= NULL*/) const
{
	if(m_szName == NULL)
	{
		if ( szFilename == NULL )
		{
			Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "Pri %d, Controller 0x%08x", m_uPriority, m_uControllerHash ));
		}
		else
		{
			for ( u_int uItr = 0; uItr < uCol; ++uItr )
			{
				Asura_SimpleLogfileOutput(( szFilename, " " ));
			}
			Asura_SimpleLogfileOutput(( szFilename, "Pri %d, Controller 0x%08x\n", m_uPriority, m_uControllerHash ));
		}
	}
	else
	{
		if ( szFilename == NULL )
		{
			Asura_DebugTextPrint(( uCol, uRow++, 0xffffffff, 0, "Pri %d, Controller 0x%08x \"%s\"", m_uPriority, m_uControllerHash, m_szName));
		}
		else
		{
			for ( u_int uItr = 0; uItr < uCol; ++uItr )
			{
				Asura_SimpleLogfileOutput(( szFilename, " " ));
			}
			Asura_SimpleLogfileOutput(( szFilename, "Pri %d, Controller 0x%08x \"%s\"\n", m_uPriority, m_uControllerHash, m_szName ));
		}
	}
}

/// End of function 'Asura_AnimationController::ReplaceBlendToPreviousController'. ///
#endif // #if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)


/// End of file. ///
