
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_AnimationController.h
// Author : Andy Tate
// Date   : 14-Sep-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_ANIMATIONCONTROLLER_H__
#define __ASURA_ANIMATIONCONTROLLER_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Hash.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation;
class Asura_Bounding_Box;
class Asura_AnimationBlend;
class Asura_AnimationController_Group;
class Asura_Hierarchy_Skin;
class Asura_Hierarchy_Tween;

////////////////////////////////////////////////////////////////////////////////
// Defines...
////////////////////////////////////////////////////////////////////////////////

#define AAC_FLAG_ACTIVE					0x00000001
#define AAC_FLAG_DESTROY_WHEN_FINISHED	0x00000002
// If set, the controller's hash may be changed from the passed-in value so that it definitely gets added to the tree.
// Obviously, this means you have to store the hash of your new controller if you want to access it later,
// as it may not be what you originally asked for.
#define AAC_FLAG_ENSURE_UNIQUE_HASH		0x00000004
//This flag allows an animation to be destroyed by the internal DESTROY_MSG message ,and is assumed to be
//enabled if using AAC_FLAG_DESTROY_WHEN_FINISHED.
//Its main purpose is to allow primary animations that have been tarted using the wrapper functions 
//to be removed when a new primary has finished tweening in.
//RAR 29/11/2006
#define AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE 0x00000008

////////////////////////////////////////////////////////////////////////////////
// Structure declarations...
////////////////////////////////////////////////////////////////////////////////

struct Asura_AnimationMessage_SyncTimerTo
{
	Asura_Hash_ID	m_uToAnimationHash;
	float			m_fTimeStep;
};


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_AnimationController
// Author       : Andy Tate
// Date         : 14-Sep-2006
//
// Description  : This is an abstract class which is used to allow instances of
//				  animations and groups of animations to be connected together
//				  in order to provide a more powerful method of tweening between
//				  multiple animations.
////////////////////////////////////////////////////////////////////////////////

class Asura_AnimationController
{
public:
	// Public enumerated types...
	enum MessageType
	{
		PLAY_MSG,
		PAUSE_MSG,
		STOP_MSG,
		DESTROY_MSG,
		DESTROY_FROM_TWEEN_IN_MSG,//used when a tween in message to clear out any earlier animations.
		SYNC_TIMER_TO_MSG,

		MAX_MESSAGE_TYPES
	};

	// Public functions...
	virtual ~Asura_AnimationController ();

	// Virtual activation functions...
	virtual	void	Activate	(void)	{ m_uControllerFlags |= AAC_FLAG_ACTIVE; }
	virtual	void	Deactivate	(void)	{ m_uControllerFlags &= ~AAC_FLAG_ACTIVE; }

	// Abstract function to get the interval flags which were appropriate during the last frame...
	virtual int		GetFlagsForLastFrame	(void) const = 0;

	// Abstract bone-update function...
	virtual	u_int	TweenBoneData	(const Asura_Hierarchy_Skin& xSkin, const u_int uFirstBone, const u_int uLastBone, const float fWeight, const bool bDeltasUseBlends, Asura_AnimationController_Group* pxParentGroup)	= 0;

	// Abstract bounding-box function...
	virtual	u_int	TweenBoundingBox	(Asura_Bounding_Box& xBB, const Asura_Hierarchy_Skin& xSkin, const float fWeight, const bool bDeltasUseBlends, const bool bSlerpVelocity, Asura_AnimationController_Group* pxParentGroup)	= 0;

	// Abstract message handling functions...
	virtual	bool	HandleMessage	(const u_int uType, void* pvUserData = NULL)	= 0;

	// Abstract timer update functions...
	virtual	void	UpdateTimers	(const float fTimeStep, bool bIgnoreTimerFrameStamp)	= 0;

	// Abstract accessor functions...
	virtual	bool	AreAllBonesUsingDeltas	(void) const	= 0;
	virtual	bool	IsUsingSubsetOfBones	(void) const	= 0;
	virtual	bool	IsFinished				(void) const	= 0;
	virtual bool	IsReadyForDestruction	(void) const	= 0;
	virtual	float	GetTimeUntilFinished	(void) const	= 0;

	// Accessor functions...
	void		SetName(const char* szName)	{ m_szName = szName; }
	const char*	GetName(void) const			{ return(m_szName); }

	void						SetAsuraAnimation	(Asura_Animation* pxAsuraAnimation)	{ m_pxAsuraAnimation = pxAsuraAnimation; }
	Asura_Animation*			GetAsuraAnimation	(void)								{ return(m_pxAsuraAnimation); }
	const Asura_Animation*		GetAsuraAnimation	(void) const						{ return(m_pxAsuraAnimation); }

	bool						IsActive				(void) const	{ return((m_uControllerFlags & AAC_FLAG_ACTIVE) != 0); }

	void						SetBlendToNextController	(Asura_AnimationBlend* pxController)	{ m_pxBlendToNextController = pxController; }
	Asura_AnimationBlend*		GetBlendToNextController	(void) const							{ return(m_pxBlendToNextController); }

	void						SetBlendToPreviousController	(Asura_AnimationBlend* pxController)	{ m_pxBlendToPreviousController = pxController; }
	Asura_AnimationBlend*		GetBlendToPreviousController	(void) const							{ return(m_pxBlendToPreviousController); }

	// Blend replacement functions. Similar to sets, but untie and clear up the original blend too.
	// These only ever work if a blend already exists, since only then is there a next/previous controller to blend to.
	void ReplaceBlendToNextController(Asura_AnimationBlend* pxNewBlend);
	void ReplaceBlendToPreviousController(Asura_AnimationBlend* pxNewBlend);

	Asura_Hash_ID				GetControllerHash		(void) const				{ return(m_uControllerHash); }
	void						SetControllerHash		(const Asura_Hash_ID uHash) { m_uControllerHash = uHash; }

	Asura_Hash_ID				GetParentControllerHash	(void) const				{ return(m_uParentControllerHash); }
	void						SetParentControllerHash	(const Asura_Hash_ID uHash)	{ m_uParentControllerHash = uHash; }

	void						SetPriority	(const u_int uPriority)	{ m_uPriority = uPriority; }
	u_int						GetPriority	(void) const			{ return(m_uPriority); }

	u_int						GetControllerFlags() const			{ return m_uControllerFlags; }
	void						SetControllerFlags(u_int uFlags)	{ m_uControllerFlags = uFlags; }

	virtual Asura_AnimationController*	Duplicate(Asura_Animation* pxDestAnimation) const = 0;

#if (defined ASURA_USE_DEBUG_TEXT || defined ASURA_LOGFILES_ENABLED)
	virtual void PrintDebugInfo(u_int& uRow, u_int& uCol, const Asura_AnimationController_Group* pxParentGroup = NULL, const char* szFilename = NULL) const;
#endif

protected:
	// Protected variables...
	const char*					m_szName;						// The name associated with this controller.

	Asura_Animation*			m_pxAsuraAnimation;				// A pointer to the animation which updates this controller.

	Asura_Hash_ID				m_uControllerHash;				// The unique hash-ID used for this controller.
	Asura_Hash_ID				m_uParentControllerHash;		// The unique hash-ID used for this controller's parent group-controller.
	u_int						m_uControllerFlags;						// Single accessor to allow flags to be easily cleared.
	u_int						m_uPriority;					// The priority of this controller - used for storage.

	Asura_AnimationBlend*		m_pxBlendToNextController;		// A pointer to the blend to apply to the next controller.
	Asura_AnimationBlend*		m_pxBlendToPreviousController;	// A pointer to the blend to apply to the previous controller.


	// Protected functions...
	Asura_AnimationController (const Asura_Hash_ID uControllerHash, const u_int uInitialFlags, const u_int uPriority);


private:
	// Private variables...


	// Private functions...
	Asura_AnimationController (void);
	Asura_AnimationController (const Asura_AnimationController& xOther);

	Asura_AnimationController&	operator=	(const Asura_AnimationController& xOther);

};   // End of class 'Asura_AnimationController'.


#endif   // ifndef __ASURA_ANIMATIONCONTROLLER_H__


/// End of file. ///
