//
// Asura_TextureAnimation.cpp
//
// Created by Richard Rice - 13/05/2005
// 
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Animation/Asura_TextureAnimation.h"

#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Stream.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_NetMsg.h"


/////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_TextureAnimationOutputComplete **
**--------------------------------------*/


struct Asura_TextureAnimationOutputComplete
{
	float m_fUOffset;
	float m_fVOffset;
	u_int m_uAddedMaterialFlags;
	u_int m_uRemovedMaterialFlags;
};

////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_TextureAnimation_Handler **
**--------------------------------*/

//cached results from most recent use of GetAnimationState
static Asura_Hash_ID g_uCachedAnimHash;
static const Asura_TextureAnimationController* g_pxCachedAnimController;
static Asura_TextureAnimationOutputComplete g_xCachedOutput;
static u_int g_uCachedFrameStamp;
static bool g_bCachedReturnValue;


Asura_TextureAnimation_Handler::Asura_TextureAnimation_Handler()
{
}

Asura_TextureAnimation_Handler::~Asura_TextureAnimation_Handler()
{
}

void Asura_TextureAnimation_Handler::Reset()
{
	m_xAnimTree.Reset();
}

void Asura_TextureAnimation_Handler::Update()
{
	m_xAnimTree.Traverse(UpdateAnimationControllersCallback);	
}

void Asura_TextureAnimation_Handler::UpdateAnimationControllersCallback(Asura_Hash_ID uHash,Asura_TextureAnimation* pxAnim)
{
	pxAnim->UpdateGlobalControllerTimer();
}


bool Asura_TextureAnimation_Handler::NewAnimation(const char* pszName,Asura_TextureAnimation*& pxOutputAnim)
{
	Asura_Hash_ID uHash = Asura_GetHashID(pszName);

	bool bDidExist = true;
	pxOutputAnim = m_xAnimTree.AddOrFindNode(uHash,&bDidExist,true);
	if(pxOutputAnim && !bDidExist)
	{
		//found index for new anim
		pxOutputAnim->Initialise(pszName);
		return true;
	}
	else
	{
		//already exists
		return false;
	}
}

Asura_TextureAnimation* Asura_TextureAnimation_Handler::GetAnimationByHash(Asura_Hash_ID uHash)
{
	return m_xAnimTree.GetDataByHash(uHash);
}

#ifdef ASURA_TOOLS
Asura_TextureAnimation* Asura_TextureAnimation_Handler::GetAnimationByIndex(int iIndex)
{
	Asura_Assert(iIndex>=0 && iIndex<m_xAnimTree.GetNumberOfNodes(),"");
	
	return m_xAnimTree.GetDataByIndex(iIndex);
}
#endif

int Asura_TextureAnimation_Handler::GetNumberOfAnimations()
{
	return m_xAnimTree.GetNumberOfNodes();
}


bool Asura_TextureAnimation_Handler::GetAnimationState(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,Asura_TextureAnimationOutput* pxOutput)
{
	if(uAnimHash == ASURA_HASH_ID_UNSET) return false;

	if(UpdateCachedAnimationState(uAnimHash,pxController))
	{
		pxOutput->m_fUOffset = g_xCachedOutput.m_fUOffset;
		pxOutput->m_fVOffset = g_xCachedOutput.m_fVOffset;
		return true;
	}
	return false;
}

u_int Asura_TextureAnimation_Handler::GetMaterialFlags(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,u_int uInitialFlags)
{

	if(uAnimHash)
	{
		if(UpdateCachedAnimationState(uAnimHash,pxController))
		{
			return (uInitialFlags | g_xCachedOutput.m_uAddedMaterialFlags) &~ g_xCachedOutput.m_uRemovedMaterialFlags;
		}
	}
	return uInitialFlags;
}

bool Asura_TextureAnimation_Handler::GetChangedMaterialFlags(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,u_int* puOutputAddedFlags, u_int* puOutputRemovedFlags)
{
	if(uAnimHash)
	{
		if(UpdateCachedAnimationState(uAnimHash,pxController))
		{
			if(puOutputAddedFlags) *puOutputAddedFlags = g_xCachedOutput.m_uAddedMaterialFlags;
			if(puOutputRemovedFlags) *puOutputRemovedFlags = g_xCachedOutput.m_uRemovedMaterialFlags;
			return true;
		}
	}
	return false;
}


// Update from message data.
void Asura_TextureAnimation_Handler::Update(const Asura_Message_EnvironmentTextureAnimation* pxData)
{
	if (!pxData) return;

	Asura_TextureAnimation* pxAnim = GetAnimationByHash(pxData->m_uAnimHash);
	if(pxAnim)
	{
		pxAnim->SetSequence(pxData->m_ucSequence);
		pxAnim->SetSpeed(pxData->m_sSpeed/100.0f);
	}
}

bool Asura_TextureAnimation_Handler::UpdateCachedAnimationState(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController)
{

	#ifdef ASURA_TOOLS
	//check for all anims being disabled
	if(Asura_TextureAnimation_System::IsDisabled()) return false;
	#endif

	//are the cached results okay?
	u_int uFrameNumber = Asura_Timers::GetFrameNumber();
	if( uAnimHash == g_uCachedAnimHash && g_pxCachedAnimController == pxController && uFrameNumber == g_uCachedFrameStamp)
	{
		return g_bCachedReturnValue;
	}

	//find the anim
	const Asura_TextureAnimation* pxAnim = GetAnimationByHash(uAnimHash);
	
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// MGH 12/6/2006: There were some calls to check material flags that had the
	// wrong handler set, causing the cached value to be stomped on.
	// I've fixed as many of those calls as I could, but this is a fallback
	// for the remaining ones.
	if (!pxAnim)
	{
		//Asura_Assert(0, "Couldn't find texture animation in the current handler");
		pxAnim = Asura_TextureAnimation_System::FindAnimationByHash(uAnimHash);
	}
#endif

	if(pxAnim)
	{
		//and get the UV offsets from it
		g_bCachedReturnValue = pxAnim->GetAnimationState(pxController,&g_xCachedOutput);
	}
	else
	{
		g_bCachedReturnValue = false;
	}
	g_uCachedAnimHash = uAnimHash;
	g_pxCachedAnimController = pxController;
	g_uCachedFrameStamp = uFrameNumber;


	return g_bCachedReturnValue;
}

#ifdef ASURA_TOOLS
void Asura_TextureAnimation_Handler::DeleteAnimation(Asura_Hash_ID uAnimHash)
{
	//delete from the hash tree
	m_xAnimTree.DeleteNodeByHash(uAnimHash);
	
}

bool Asura_TextureAnimation_Handler::RenameAnimation(const char* pszOldName,const char* pszNewName)
{
	Asura_Hash_ID uOldHash = Asura_GetHashID(pszOldName);
	Asura_Hash_ID uNewHash = Asura_GetHashID(pszNewName);

	//find the anim to be renamed
	Asura_TextureAnimation* pxAnim = GetAnimationByHash(uOldHash);
	if(!pxAnim) return false;

	//update hash tree
	if(!m_xAnimTree.ReplaceHash(uOldHash,uNewHash)) return false;

	//if that worked ,its okay to update the name
	pxAnim->SetName(pszNewName);
	return true;
}

#endif //ASURA_TOOLS

void Asura_TextureAnimation_Handler::ResetAnimationControllers()
{
	m_xAnimTree.Traverse(ResetAnimationControllersCallback);
}

void Asura_TextureAnimation_Handler::ResetAnimationControllersCallback(Asura_Hash_ID uHash, Asura_TextureAnimation* pxAnim)
{
	pxAnim->ResetAnimationController();
}


#if defined(ASURA_TOOLS)
u_int Asura_TextureAnimation_Handler::GetMaterialFlagsAddedByAnimation(Asura_Hash_ID uHash)
{
	Asura_TextureAnimation* pxAnim = GetAnimationByHash(uHash);
	if(!pxAnim) return 0;

	return pxAnim->GetMaterialFlagsAddedByAnimation();
}

u_int Asura_TextureAnimation_Handler::GetMaterialFlagsRemovedByAnimation(Asura_Hash_ID uHash)
{
	Asura_TextureAnimation* pxAnim = GetAnimationByHash(uHash);
	if(!pxAnim) return 0;

	return pxAnim->GetMaterialFlagsRemovedByAnimation();
}
#endif	// defined(ASURA_TOOLS)



////////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_TextureAnimation_System **
**-------------------------------*/

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
Asura_TemplateHashTree<Asura_TextureAnimation_Handler>	Asura_TextureAnimation_System::s_xHandlerTree;
Asura_TextureAnimation_Handler*						Asura_TextureAnimation_System::s_pxCurrentHandler	= NULL;
#else
Asura_TextureAnimation_Handler		Asura_TextureAnimation_System::s_xHandler;
#endif


#ifdef ASURA_TOOLS
bool Asura_TextureAnimation_System::s_bDisabled = false;
#endif

void Asura_TextureAnimation_System::Initialise()
{
	// Nothing to initialise...
}

void Asura_TextureAnimation_System::Shutdown()
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	Asura_TextureAnimation_Handler&	xHandler = GetCurrentHandler();

	// Reset the current instance...
	xHandler.Reset();
#endif
}

void Asura_TextureAnimation_System::Reset()
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();
#else
	// Reset the current instance...
	GetCurrentHandler().Reset();
#endif
}

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
void Asura_TextureAnimation_System::ReparentHandler(Asura_TextureAnimation_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash)
{
	Asura_TextureAnimation_Handler*	pxParentHandler	= s_xHandlerTree.GetDataByHash(uNewParentHash);

	if(pxParentHandler != NULL)
	{
		xSourceHandler.m_xAnimTree.SetParent(&(pxParentHandler->m_xAnimTree));
	}
}

void Asura_TextureAnimation_System::ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash)
{
	Asura_TextureAnimation_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uSourceHash);

	if(pxHandler != NULL)
	{
		ReparentHandler(*pxHandler, uNewParentHash);
	}
}

Asura_TextureAnimation_Handler* Asura_TextureAnimation_System::CreateHandler (const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)
{
	Asura_TextureAnimation_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHash);

		if(pxHandler != NULL)
		{
			ReparentHandler(uHash, uParentHandlerHash);
		}
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have a texture animation handler with this hash-ID.");
	}

	// Return the handler...
	return(pxHandler);
}

bool Asura_TextureAnimation_System::DestroyHandler (const Asura_Hash_ID uHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

void Asura_TextureAnimation_System::SetCurrentHandler(Asura_TextureAnimation_Handler* pxHandler)
{
	if(pxHandler != s_pxCurrentHandler)
	{
		//invalidate cache of last anim result fetched
		g_uCachedAnimHash = 0;

		//and change the handler
		s_pxCurrentHandler = pxHandler;

	}
}

void Asura_TextureAnimation_System::WriteHandlerReport(u_int uHandler, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 64;
	char szLogBuffer[uLOGBUFFERSIZE];

	Asura_TextureAnimation_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHandler);
	if (pxHandler)
	{
		if (!pxHandler->m_xAnimTree.GetNumberOfNodes())
		{
			xLog << "Tex Anim:\t\t\tNone\n";
		}
		else
		{
			Asura_TemplateHashTree_Iterator<Asura_TextureAnimation> xIt(pxHandler->m_xAnimTree);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_TextureAnimation* pxTexAnim = xIt.GetCurrent();
				Asura_PrintfToBfr( szLogBuffer, "Tex Anim:\t\t\t%s\n", uLOGBUFFERSIZE, true, pxTexAnim->GetName());
				xLog << szLogBuffer;
			}		
		}
	}
	else
	{
		xLog << "Tex Anim:\t\t\tNo Handler\n";
	}
}

/////////////////////////////////////
// Asura_TextureAnimation_UserData //
/////////////////////////////////////
class Asura_TextureAnimation_UserData
{
public:
	Asura_TextureAnimation_UserData(Asura_Hash_ID uAnimHash)
		: m_uAnimHash(uAnimHash), m_pxTexAnim(0), m_pxHandler(0)
	{}

	Asura_Hash_ID m_uAnimHash; // Input

	Asura_TextureAnimation* m_pxTexAnim;		 // Output
	Asura_TextureAnimation_Handler* m_pxHandler; // Output
};

/////////////////////////
// FindAnimationByHash //
/////////////////////////
// Find animation by looking through all the handlers if necessary.
Asura_TextureAnimation* Asura_TextureAnimation_System::FindAnimationByHash(Asura_Hash_ID uAnimHash)
{
	Asura_TextureAnimation_UserData xUserData(uAnimHash);
	s_xHandlerTree.Traverse(FindAnimation_Callback, &xUserData);
	return xUserData.m_pxTexAnim;
}

////////////////////////////
// FindAnimation_Callback //
////////////////////////////
bool Asura_TextureAnimation_System::FindAnimation_Callback(Asura_Hash_ID uHash, Asura_TextureAnimation_Handler* pxHandler, void* pxUserData)
{
	Asura_TextureAnimation_UserData* pxTexAnimUserData = static_cast<Asura_TextureAnimation_UserData*>(pxUserData);
	Asura_TextureAnimation* pxTexAnim = pxHandler->GetAnimationByHash(pxTexAnimUserData->m_uAnimHash);
	if (pxTexAnim)
	{
		pxTexAnimUserData->m_pxHandler = pxHandler;
		pxTexAnimUserData->m_pxTexAnim = pxTexAnim;
		return false; // Found it, so stop looking.
	}
	return true; // Carry on looking.
}


#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)




////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Asura_TextureAnimation **
**------------------------*/

Asura_TextureAnimation::Asura_TextureAnimation()
{
	m_uNumberOfSequences = 0;
	m_pxSequences = 0;
}

Asura_TextureAnimation::~Asura_TextureAnimation()
{
	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(m_pxSequences);
	}
	Asura_ResourcePopType();
}

void Asura_TextureAnimation::Initialise(const char* pszName)
{
	m_xName = pszName;
}

#ifdef ASURA_TOOLS
void Asura_TextureAnimation::Copy(const Asura_TextureAnimation* pxSourceAnim)
{
	if(this == pxSourceAnim) return;

	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		m_uNumberOfSequences = pxSourceAnim->m_uNumberOfSequences;

		Asura_ResourceDeallocateArray(m_pxSequences);
		Asura_ResourceAllocateArray(m_pxSequences, m_uNumberOfSequences);
		
		for(u_int i=0;i<m_uNumberOfSequences;i++)
		{
			m_pxSequences[i].Copy(&pxSourceAnim->m_pxSequences[i]);
		}	
	}
	Asura_ResourcePopType();
}
#endif ///ASURA_TOOLS

bool Asura_TextureAnimation::GetAnimationState(const Asura_TextureAnimationController* pxController,Asura_TextureAnimationOutputComplete* pxOutput) const
{
	//do we need to use our own controller?
	bool bInternalController = pxController ? false : true;
	if(!pxController)
	{
		pxController = &m_xAnimController;
	}

	//get the correct sequence
	u_int uSeq = pxController->GetSequence();

	if(uSeq >= m_uNumberOfSequences)
	{
		Asura_Assertf(0,("Invalid sequence index %d being used for texture animation '%s'",uSeq,m_xName.GetString()));
		return false;
	}

	const Asura_TextureAnimationSequence* pxSeq = &m_pxSequences[uSeq];
	float fDuration = pxSeq->GetDuration();

	#ifdef ASURA_TOOLS
	if(pxController->m_iFrameOverride>=0 && pxController->m_iFrameOverride < (int)pxSeq->m_uNumberOfFrames)
	{
		//Force the frame that is being edited in tools to be shown regardless of the timer
		const Asura_TextureAnimationFrame* pxFrame = &pxSeq->m_pxFrames[pxController->m_iFrameOverride];
	   	pxOutput->m_fUOffset = pxFrame->m_fU;
	   	pxOutput->m_fVOffset = pxFrame->m_fV;
		pxOutput->m_uAddedMaterialFlags = pxFrame->m_uAddedMaterialFlags;
		pxOutput->m_uRemovedMaterialFlags = pxFrame->m_uRemovedMaterialFlags;
		return true;
	}
	#endif //ASURA_TOOLS


	//get a copy of the timer
	float fTimer = pxController->m_fTimer;


	//double the duration for the purposes of looping if going back and forth
	float fModifiedDuration = pxSeq->m_bLoopBackAndForth ? fDuration*2.0f : fDuration;

	//make sure its in range 
	if(pxController->m_fTimer > fModifiedDuration)
	{
		//end of anim.. loop or clamp as needed
		if(pxSeq->m_bLoop || pxSeq->m_bLoopBackAndForth)
		{
			if(fModifiedDuration)
			{
				fTimer -= Asura_Maths::FloatToInt_Chop(fTimer/fModifiedDuration) * fModifiedDuration;
				fTimer = Asura_Maths::ClampToRange(fTimer,0.0f,fModifiedDuration);
			}
			else
			{
				fTimer = 0;
			}
		}
		else
		{
			fTimer = fModifiedDuration;
		}
	}
	else if(fTimer < 0) //in case speed is negative
	{
		if(pxSeq->m_bLoop || pxSeq->m_bLoopBackAndForth)
		{
			if(fModifiedDuration)
			{
				fTimer -= (Asura_Maths::FloatToInt_Chop(fTimer/fModifiedDuration)-1) * fModifiedDuration;
				fTimer = Asura_Maths::ClampToRange(fTimer,0.0f,fModifiedDuration);
			}
			else
			{
				fTimer = 0;
			}
		}
		else
		{
			fTimer = 0;
		}
	}

	if(bInternalController)
	{
		//If using the internal controller ,then update the clamped timer to reduce problems
		//from it getting to large.
		//(Can't do this for external controllers , since they may need to control multiple
		//animations with different durations).
		pxController->m_fTimer = fTimer;
	}

	if(pxSeq->m_bLoopBackAndForth)
	{

		//adjust the timer if its in the back part of its journey
		if(fTimer > fDuration)
		{
			fTimer = fModifiedDuration - fTimer;
			fTimer = Asura_Maths::ClampToRange(fTimer,0.0f,fDuration);
		}
	}

	Asura_Assert(fDuration,"");
	float fProportionalTime = fTimer/fDuration;

	//get the current frame
	u_int uFrame;
	for(uFrame = pxSeq->m_uNumberOfFrames-1;uFrame;uFrame--)
	{
		if(pxSeq->m_pxFrames[uFrame].m_fTime<=fProportionalTime) break;
	}

	const Asura_TextureAnimationFrame* pxFrame = &pxSeq->m_pxFrames[uFrame];

	if(pxSeq->m_bRandom)
	{
		//Generate random UV offsets.
		//Hopefully generating a new value each time this gets called should still allow connected
		//polygons to be animated consistently.
		pxOutput->m_fUOffset = Asura_Random::GetFloat(0.0f,pxFrame->m_fU);
		pxOutput->m_fVOffset = Asura_Random::GetFloat(0.0f,pxFrame->m_fV);

	}
	else if(pxSeq->m_bInterpolate)
	{
		//interpolate to next frame
		if(uFrame!=pxSeq->m_uNumberOfFrames-1)
		{
			const Asura_TextureAnimationFrame* pxNextFrame = pxFrame+1;
			float fFrameDuration = pxNextFrame->m_fTime-pxFrame->m_fTime;

			Asura_Assert(fFrameDuration>=0,"");

			//get the proportion through the frame
			float fLerp = (fProportionalTime-pxFrame->m_fTime)/fFrameDuration;
			fLerp = Asura_Maths::ClampToRange(fLerp,0.0f,1.0f);

			//and fill in the interpolated UVs
			pxOutput->m_fUOffset = pxFrame->m_fU + (pxNextFrame->m_fU-pxFrame->m_fU)*fLerp;
			pxOutput->m_fVOffset = pxFrame->m_fV + (pxNextFrame->m_fV-pxFrame->m_fV)*fLerp;

		}
		else
		{
			//On last frame. Not going to interpolate back to first frame by default , since
			//this may well end up with interpolating the wrong way for scrolly effects.
			//(And it can always be achieved if desired by making the last frame the same as the first)
			pxOutput->m_fUOffset = pxFrame->m_fU;
			pxOutput->m_fVOffset = pxFrame->m_fV;
		}
	}
	else
	{
		//just take the frame as is
		pxOutput->m_fUOffset = pxFrame->m_fU;
		pxOutput->m_fVOffset = pxFrame->m_fV;
 
	}
	pxOutput->m_uAddedMaterialFlags = pxFrame->m_uAddedMaterialFlags;
	pxOutput->m_uRemovedMaterialFlags = pxFrame->m_uRemovedMaterialFlags;


	return true;
}

Asura_TextureAnimationSequence& Asura_TextureAnimationSequence::operator = (const Asura_TextureAnimationSequence& xSeq)
{
	if(this != &xSeq)
	{
		Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
		{
			Asura_ResourceDeallocateArray(m_pxFrames);

			m_uNumberOfFrames = xSeq.m_uNumberOfFrames;
			Asura_ResourceAllocateArray(m_pxFrames, m_uNumberOfFrames);
			for(u_int i=0;i<m_uNumberOfFrames;i++)
			{
				m_pxFrames[i] = xSeq.m_pxFrames[i];
			}

			m_fDuration = xSeq.m_fDuration;
			m_uFlags = xSeq.m_uFlags;
		}
		Asura_ResourcePopType();
	}
	return *this;
}

#ifdef ASURA_TOOLS
void Asura_TextureAnimation::CreateDefaultSequence()
{
	Asura_Assert(!m_uNumberOfSequences,"This should be called on a newly created sequence");

	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		m_uNumberOfSequences = 1;
		Asura_ResourceAllocateArray(m_pxSequences, 1);
		m_pxSequences->CreateDefaultFrame();
	}
	Asura_ResourcePopType();
}

void Asura_TextureAnimation::AddSequence()
{
	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		Asura_TextureAnimationSequence* pxSequences;
		Asura_ResourceAllocateArray(pxSequences, m_uNumberOfSequences+1);

		//copy over the old sequences
		for(u_int i=0;i<m_uNumberOfSequences;i++)
		{
			pxSequences[i] = m_pxSequences[i];
		}
		//initialise the new one
		pxSequences[m_uNumberOfSequences].CreateDefaultFrame();

		//and replace
		Asura_ResourceDeallocateArray(m_pxSequences);
		m_pxSequences = pxSequences;
		m_uNumberOfSequences++;
	}
	Asura_ResourcePopType();
}

void Asura_TextureAnimation::DeleteSequence(u_int uSequence)
{
	Asura_Assert(m_uNumberOfSequences>1,"");
	Asura_Assert(uSequence<m_uNumberOfSequences,"");

	//shuffle the sequences down to replace the deleted one
	m_uNumberOfSequences--;
	for(u_int i=uSequence;i<m_uNumberOfSequences-1;i++)
	{
		m_pxSequences[i] = m_pxSequences[i+1];
	}

}

#endif //ASURA_TOOLS

#ifdef ASURA_TOOLS
u_int Asura_TextureAnimation::GetMaterialFlagsAddedByAnimation() const 
{
	u_int uFlags = 0;
	for(u_int i=0;i<m_uNumberOfSequences;i++)
	{
		uFlags |= m_pxSequences[i].GetMaterialFlagsAddedBySequence();
	}
	return uFlags;
}

u_int Asura_TextureAnimation::GetMaterialFlagsRemovedByAnimation() const
{
	u_int uFlags = 0;
	for(u_int i=0;i<m_uNumberOfSequences;i++)
	{
		uFlags |= m_pxSequences[i].GetMaterialFlagsRemovedBySequence();
	}
	return uFlags;
}

#endif //ASURA_TOOLS

void Asura_TextureAnimation::UpdateGlobalControllerTimer()
{
	m_xAnimController.UpdateTimer(Asura_Timers::GetGameTimeSlice());
}

void Asura_TextureAnimation_System::HandleMessage(const Asura_Message* pxMessage)
{
	switch(pxMessage->GetMsgID())
	{
		case ASURA_MSG_ENVTEXTUREANIMATION :
		{
			const Asura_Message_EnvironmentTextureAnimation* pxData = reinterpret_cast
				<const Asura_Message_EnvironmentTextureAnimation*>(pxMessage->GetData());

			#if defined(ASURA_USE_MULTIPLE_HANDLERS)
			// Update all the handlers.
			// MGH 12/6/2006: Annoyingly there is no traversal function with a constant user data pointer.
			s_xHandlerTree.Traverse(EnvTextureAnimation_Callback, const_cast<Asura_Message_EnvironmentTextureAnimation*>(pxData));
			#else
			s_xHandler.Update(pxData);
			#endif

			break;
		}
		default :
			Asura_Assert(0,"Unexpected message sent to Asura_TextureAnimation::HandleMessage()");
			break;

	}
}



////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_TextureAnimationSequence **
**--------------------------------*/

Asura_TextureAnimationSequence::Asura_TextureAnimationSequence()
{
	m_uNumberOfFrames = 0;
	m_pxFrames = 0;

	m_fDuration = 0.0f;
	m_uFlags = 0;
	m_bLoop = true;
}

Asura_TextureAnimationSequence::~Asura_TextureAnimationSequence()
{
	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		Asura_ResourceDeallocateArray(m_pxFrames);
	}
	Asura_ResourcePopType();
}


#ifdef ASURA_TOOLS
void Asura_TextureAnimationSequence::Copy(const Asura_TextureAnimationSequence* pxSourceSeq)
{
	if(this == pxSourceSeq) return;

	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		m_uNumberOfFrames = pxSourceSeq->m_uNumberOfFrames;

		Asura_ResourceDeallocateArray(m_pxFrames);
		Asura_ResourceAllocateArray(m_pxFrames, m_uNumberOfFrames);

		for(u_int i=0;i<m_uNumberOfFrames;i++)
		{
			m_pxFrames[i] = pxSourceSeq->m_pxFrames[i];
		}

		m_fDuration = pxSourceSeq->m_fDuration;
		m_uFlags = pxSourceSeq->m_uFlags;
	}
	Asura_ResourcePopType();
}
#endif ///ASURA_TOOLS

#ifdef ASURA_TOOLS
void Asura_TextureAnimationSequence::CreateDefaultFrame()
{
	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		Asura_Assert(!m_uNumberOfFrames,"This should be called on a newly created frame");

		m_fDuration = 1.0f;

		m_uNumberOfFrames = 1;
		Asura_ResourceAllocateArray(m_pxFrames, 1);
		m_pxFrames->m_fU = 0.0f;
		m_pxFrames->m_fV = 0.0f;
		m_pxFrames->m_fTime = 0.0f;
		m_pxFrames->m_uAddedMaterialFlags = 0;
		m_pxFrames->m_uRemovedMaterialFlags = 0;
	}
	Asura_ResourcePopType();
}

void Asura_TextureAnimationSequence::AddFrameAfter(int iPrevFrame)
{
	Asura_ResourcePushType(ASURA_RESOURCE_TEXTURE_ANIMATION);
	{
		Asura_Assert(iPrevFrame>=-1 && iPrevFrame<(int)m_uNumberOfFrames,"");
		Asura_TextureAnimationFrame* pxFrames;
		
		Asura_ResourceAllocateArray(pxFrames, m_uNumberOfFrames+1);

		//copy over the old frames
		u_int uDestFrame = 0;
		for(u_int i=0;i<m_uNumberOfFrames;i++,uDestFrame++)
		{
			if((int)i==iPrevFrame+1) uDestFrame++;

			pxFrames[uDestFrame] = m_pxFrames[i];
		}
		//initialise the new one
		int iNewFrame = iPrevFrame+1;
		if(iNewFrame)
		{
			pxFrames[iNewFrame] = pxFrames[iNewFrame-1];
		}
		else
		{
			pxFrames[iNewFrame].m_fU = 0;
			pxFrames[iNewFrame].m_fV = 0;
			pxFrames[iNewFrame].m_fTime = 0;
		}

		//and replace
		Asura_ResourceDeallocateArray(m_pxFrames);
		m_pxFrames = pxFrames;
		m_uNumberOfFrames++;
	}
	Asura_ResourcePopType();
}

void Asura_TextureAnimationSequence::DeleteFrame(u_int uFrame)
{
	Asura_Assert(m_uNumberOfFrames>1,"");
	Asura_Assert(uFrame<m_uNumberOfFrames,"");

	//shuffle the frames down to replace the deleted one
	m_uNumberOfFrames--;
	for(u_int i=uFrame;i<m_uNumberOfFrames-1;i++)
	{
		m_pxFrames[i] = m_pxFrames[i+1];
	}
	//make sure frame 0 still starts at time 0
	m_pxFrames[0].m_fTime = 0;

}

#endif //ASURA_TOOLS


#ifdef ASURA_TOOLS
u_int Asura_TextureAnimationSequence::GetMaterialFlagsAddedBySequence() const
{
	u_int uFlags = 0;
	for(u_int i=0;i<m_uNumberOfFrames;i++)
	{
		uFlags |= m_pxFrames[i].m_uAddedMaterialFlags;
	}
	return uFlags;
}

u_int Asura_TextureAnimationSequence::GetMaterialFlagsRemovedBySequence() const
{
	u_int uFlags = 0;
	for(u_int i=0;i<m_uNumberOfFrames;i++)
	{
		uFlags |= m_pxFrames[i].m_uRemovedMaterialFlags;
	}
	return uFlags;
}

#endif //ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_TextureAnimationController **
**----------------------------------*/

Asura_TextureAnimationController::Asura_TextureAnimationController()
: m_pxSelfIllumination(0)
{
	Reset();
}

Asura_TextureAnimationController::~Asura_TextureAnimationController()
{
	DeleteSelfIllumination();
}

void Asura_TextureAnimationController::Reset()
{
	m_fTimer = 0;
	m_uFrameStamp = 0xffffffff;
	m_uSequence = 0;
	m_fSpeed = 1.0f;
	#ifdef ASURA_TOOLS
	m_iFrameOverride = -1;
	#endif

	DeleteSelfIllumination();
}

void Asura_TextureAnimationController::SetSequence(u_int uSequence)
{
	if(uSequence != m_uSequence)
	{
		m_uSequence = uSequence;
		m_fTimer = 0;
	}
}

// Delete the self-illumination properties.
void Asura_TextureAnimationController::DeleteSelfIllumination()
{
	if (m_pxSelfIllumination)
	{
		delete m_pxSelfIllumination;
		m_pxSelfIllumination = 0;
	}
}

void Asura_TextureAnimationController::SetSelfIllumination(float fContribution, float fColourise, const Asura_RGBA& xColour)
{
	if (!m_pxSelfIllumination)
	{
		m_pxSelfIllumination = new Asura_TextureAnimationController_SelfIllumination;
	}
	m_pxSelfIllumination->Set(fContribution, fColourise, xColour);
}

void Asura_TextureAnimationController::SetSelfIlluminationEnable(bool bEnable) 
{
	if (m_pxSelfIllumination)
	{
		m_pxSelfIllumination->SetEnabled(bEnable);
	}
}

///////////////////////////////////////////////////////////////////////////////
// Asura_TextureAnimationController_SelfIllumination
///////////////////////////////////////////////////////////////////////////////
Asura_TextureAnimationController_SelfIllumination::Asura_TextureAnimationController_SelfIllumination()
{
	m_fContribution = 0.0f;
	m_fColourise	= 0.0f;

	m_xColour.m_fR  = 0.0f;
	m_xColour.m_fG  = 0.0f;
	m_xColour.m_fB  = 0.0f;
	m_xColour.m_fA  = 1.0f;

	m_bEnabled = false;
}


// network message

#ifdef ASURA_NETMSG_PACK_ALL_DATA
void Asura_Message_EnvironmentTextureAnimation::RegisterFields(Asura_NetMsg* pxNetMsg) const
{
	Asura_Assert(pxNetMsg->GetMsgID() == ASURA_MSG_ENVTEXTUREANIMATION, "unexpected message ID");

	Asura_NetMsg_RegisterFieldSimple(Asura_FieldType_uint, m_uAnimHash);
	Asura_NetMsg_RegisterFieldSimple(Asura_FieldType_short, m_sSpeed);
	Asura_NetMsg_RegisterFieldSimple(Asura_FieldType_uchar, m_ucSequence);

#ifdef ASURA_USE_STREAMING
#error Asura_Hash_ID m_uZoneHash; is not dealt with
#endif // ASURA_USE_STREAMING
}
#endif // ASURA_NETMSG_PACK_ALL_DATA
