//
// Asura_TextureAnimation.h
//
// Created by Richard Rice - 13/05/2005
// 
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_TEXTUREANIMATION_H__
#define __ASURA_TEXTUREANIMATION_H__

#include "Core/Asura_String.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_HashTree.h"
#include "Rendering/Asura_Colour.h"


//maximum storage space for texture anims
#ifndef ASURA_MAX_TEXTURE_ANIMATIONS
#define ASURA_MAX_TEXTURE_ANIMATIONS 100
#endif


//10 zillion and one forward declarations
class Asura_HashTree;
class Asura_Stream;
class Asura_TextureAnimation;
class Asura_TextureAnimationSequence;
struct Asura_TextureAnimationFrame;
struct Asura_TextureAnimationOutput;
struct Asura_TextureAnimationOutputComplete;
class Asura_TextureAnimationController;
class Asura_TextureAnimationController_SelfIllumination;
struct Asura_Message_EnvironmentTextureAnimation;
class Asura_NetMsg;

/////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_TextureAnimation_Handler **
**--------------------------------*/

//Instance of texture animation system per streaming zone
class Asura_TextureAnimation_Handler
{
	friend class Asura_TextureAnimation_System;

public :
	Asura_TextureAnimation_Handler();
	~Asura_TextureAnimation_Handler();

	void Reset();

	//per frame update (updates the timers of the global controllers)
	void Update();

	//Creates a new animation with the given name.
	//Returns false if it already exists , and points at the existing anim
	bool NewAnimation(const char* pszName,Asura_TextureAnimation*& pxOutputAnim);

	Asura_TextureAnimation* GetAnimationByHash(Asura_Hash_ID uHash);
	#ifdef ASURA_TOOLS
	Asura_TextureAnimation* GetAnimationByIndex(int iIndex);
	#endif


	// Update from message data.
	void Update(const Asura_Message_EnvironmentTextureAnimation* pxData);

	int GetNumberOfAnimations();

	//Get the current uv offsets for the texture animation.
	//Returns false if it fails , leaving the output untouched.
	//If pxController is NULL then animation uses its own internal controller
	bool GetAnimationState(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,Asura_TextureAnimationOutput* pxOutput);

	//Gets the material flags as modified by the animation
	u_int GetMaterialFlags(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,u_int uInitialFlags);

	//Gets the material flags that have been added or removed by the animation
	//Returns false if it fails , leaving the output untouched.
	bool GetChangedMaterialFlags(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,u_int* puOutputAddedFlags, u_int* puOutputRemovedFlags);

	#ifdef ASURA_TOOLS
	//deletes the animation specified
	void DeleteAnimation(Asura_Hash_ID uAnimHash);

	bool RenameAnimation(const char* pszOldName,const char* pszNewName);
	#endif

	//Resets all the global anim controllers back to the initial states
	void ResetAnimationControllers();


	#ifdef ASURA_TOOLS
	u_int GetMaterialFlagsAddedByAnimation(Asura_Hash_ID uHash);
	u_int GetMaterialFlagsRemovedByAnimation(Asura_Hash_ID uHash);
	#endif

private :
	bool UpdateCachedAnimationState(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController);

	static void ResetAnimationControllersCallback(Asura_Hash_ID uHash,Asura_TextureAnimation* pxAnim);
	static void UpdateAnimationControllersCallback(Asura_Hash_ID uHash,Asura_TextureAnimation* pxAnim);

	Asura_TemplateHashTree<Asura_TextureAnimation> m_xAnimTree;

};


/////////////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------------------**
** Asura_TextureAnimation_System **
**-------------------------------*/

//The main animation system thingy
class Asura_TextureAnimation_System
{
	friend class Asura_TextureAnimation_Handler;

public :
	static void Initialise();
	static void Shutdown();
	static void Reset();

	static void Update() {GetCurrentHandler().Update();}

	//Creates a new animation with the given name.
	//Returns false if it already exists , and points at the existing anim
	static bool NewAnimation(const char* pszName,Asura_TextureAnimation*& pxOutputAnim)	{return GetCurrentHandler().NewAnimation(pszName,pxOutputAnim);}

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Find animation, by looking through all of the handlers if necessary.
	static Asura_TextureAnimation* FindAnimationByHash(Asura_Hash_ID uAnimHash);
#endif

	static Asura_TextureAnimation* GetAnimationByHash(Asura_Hash_ID uHash)	{return GetCurrentHandler().GetAnimationByHash(uHash);}
#ifdef ASURA_TOOLS
	static Asura_TextureAnimation* GetAnimationByIndex(int iIndex)	{return GetCurrentHandler().GetAnimationByIndex(iIndex);}
#endif

	static int GetNumberOfAnimations()	{return GetCurrentHandler().GetNumberOfAnimations();}

	//Get the current uv offsets for the texture animation.
	//Returns false if it fails , leaving the output untouched.
	//If pxController is NULL then animation uses its own internal controller
	static bool GetAnimationState(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,Asura_TextureAnimationOutput* pxOutput)
	{ return GetCurrentHandler().GetAnimationState(uAnimHash,pxController,pxOutput); }

	//Gets the material flags as modified by the animation
	static u_int GetMaterialFlags(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,u_int uInitialFlags)
	{return GetCurrentHandler().GetMaterialFlags(uAnimHash,pxController,uInitialFlags); }

	//Gets the material flags that have been added or removed by the animation
	//Returns false if it fails , leaving the output untouched.
	static bool GetChangedMaterialFlags(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController,u_int* puOutputAddedFlags, u_int* puOutputRemovedFlags)
	{return GetCurrentHandler().GetChangedMaterialFlags(uAnimHash, pxController, puOutputAddedFlags, puOutputRemovedFlags);}

	#ifdef ASURA_TOOLS
	//deletes the animation specified
	static void DeleteAnimation(Asura_Hash_ID uAnimHash)	{ GetCurrentHandler().DeleteAnimation(uAnimHash); }

	static bool RenameAnimation(const char* pszOldName,const char* pszNewName)	{ return GetCurrentHandler().RenameAnimation(pszOldName,pszNewName); }
	#endif

	//Resets all the global anim controllers back to the initial states
	static void ResetAnimationControllers()	{ GetCurrentHandler().ResetAnimationControllers(); }

	#ifdef ASURA_TOOLS
	//Disables all animations. Used as option in texturing mode to make texturing animating
	//stuff easier.
	static void DisableAnimations(bool bDisable){s_bDisabled = bDisable;}
	static bool IsDisabled() {return s_bDisabled;}
	#endif

	#ifdef ASURA_TOOLS
	static u_int GetMaterialFlagsAddedByAnimation(Asura_Hash_ID uHash)		{ return GetCurrentHandler().GetMaterialFlagsAddedByAnimation(uHash); }
	static u_int GetMaterialFlagsRemovedByAnimation(Asura_Hash_ID uHash)	{ return GetCurrentHandler().GetMaterialFlagsRemovedByAnimation(uHash); }
	#endif

	//For handling the environmental texture animation update messages
	static void HandleMessage(const class Asura_Message* pxMessage);


#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Handler functions...
	static	void							ReparentHandler			(Asura_TextureAnimation_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash);
	static	void							ReparentHandler			(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash);

	static	Asura_TextureAnimation_Handler*	CreateHandler			(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash);
	static	bool							DestroyHandler			(const Asura_Hash_ID uHash);
	static	void							WriteHandlerReport		(u_int uHandler, Asura_Stream& xLog);

	static	void							CreateAndSetHandler		(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash) { s_pxCurrentHandler = CreateHandler(uHash, uParentHandlerHash); }
	static	void							SetCurrentHandler		(Asura_TextureAnimation_Handler* pxHandler);
	static	void							SetCurrentHandler		(const Asura_Hash_ID uHash)	{ s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static	Asura_TextureAnimation_Handler&	GetCurrentHandler		(void)						{ return(*s_pxCurrentHandler); }
	static	Asura_TextureAnimation_Handler*	GetHandler				(const Asura_Hash_ID uHash)	{ return(s_xHandlerTree.GetDataByHash(uHash)); }
#else
	// Private functions...
	static	Asura_TextureAnimation_Handler&	GetCurrentHandler		(void)						{ return(s_xHandler); }
#endif


private :
//	static bool UpdateCachedAnimationState(Asura_Hash_ID uAnimHash,const Asura_TextureAnimationController* pxController);

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	static bool EnvTextureAnimation_Callback(Asura_Hash_ID uHash, Asura_TextureAnimation_Handler* pxHandler, void* pxUserData)
	{	
		pxHandler->Update(static_cast<Asura_Message_EnvironmentTextureAnimation*>(pxUserData));
		return true; // Traverse all handlers.
	}

	static bool FindAnimation_Callback(Asura_Hash_ID uHash, Asura_TextureAnimation_Handler* pxHandler, void* pxUserData);

	// Private variables...
	static	Asura_TextureAnimation_Handler*							s_pxCurrentHandler;		// Currently active handler.
	static	Asura_TemplateHashTree<Asura_TextureAnimation_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.
#else
	// Private variables...
	static	Asura_TextureAnimation_Handler							s_xHandler;				// Single static-instance of a zone.
#endif

	#ifdef ASURA_TOOLS
	static bool s_bDisabled;
	#endif
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_TextureAnimationController **
**----------------------------------*/

//animation controller keeps track of the current sequence and the timer
class Asura_TextureAnimationController
{
public :
	Asura_TextureAnimationController();
	~Asura_TextureAnimationController();

	//Changes the current sequence. If the sequence is different it will also reset the timer
	void SetSequence(u_int uSequence);

	u_int GetSequence() const{return m_uSequence;}


	//Sets a timescale/speed multiplier on the animation.
	//A speed of zero will pause the anim.
	//Negative values will reverse it.
	void SetSpeed(float fSpeed){m_fSpeed = fSpeed;}
	float GetSpeed() const {return m_fSpeed;}

	#ifdef ASURA_TOOLS
	//for editing.. forces the specified frame to be drawn
	void SetFrameOverride(int iFrame){m_iFrameOverride = iFrame;}
	void ClearFrameOverride(){m_iFrameOverride = -1;}
	#endif

	void UpdateTimer(const float fFrameTime)
	{
		m_fTimer += m_fSpeed*fFrameTime;
	}


	//reset back to initial state
	void Reset();

	void SetSelfIllumination(float fContribution, float fColourise, const Asura_RGBA& xColour);
	void SetSelfIlluminationEnable(bool bEnable);

	const Asura_TextureAnimationController_SelfIllumination* GetSelfIllumination() const { return m_pxSelfIllumination; }

private :
	void DeleteSelfIllumination();

	mutable float m_fTimer;
	mutable u_int m_uFrameStamp;
	u_int m_uSequence;
	float m_fSpeed;

	#ifdef ASURA_TOOLS
	int m_iFrameOverride;
	#endif

	// Modify the material response self-illumination properties.
	Asura_TextureAnimationController_SelfIllumination* m_pxSelfIllumination;
			
	friend class Asura_TextureAnimation;

};

/////////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Asura_TextureAnimation **
**------------------------*/

//details of a single animation , which may contain multiple sequences.
//(each material gets assigned at most one texture animation , which remains fixed...
//any changes get done through starting/stopping , or changing sequence).

class Asura_TextureAnimation
{
public :
	Asura_TextureAnimation();
	~Asura_TextureAnimation();

	void Initialise(const char* pszName);

	#ifdef ASURA_TOOLS
	void Copy(const Asura_TextureAnimation* pxSourceAnim);
	#endif

	//Get the current uv offsets for the texture animation.
	//Returns false if it fails , leaving the output untouched.
	//If pxController is NULL then animation uses its own internal controller
	bool GetAnimationState(const Asura_TextureAnimationController* pxController,Asura_TextureAnimationOutputComplete* pxOutput) const;

	#ifdef ASURA_TOOLS
	//Creates one sequence with one frame
	void CreateDefaultSequence();

	void AddSequence();
	void DeleteSequence(u_int uSeq);
	#endif

	const char* GetName() const {return m_xName.GetString();}

	#ifdef ASURA_TOOLS
	void SetName(const char* pszName){m_xName = pszName;}
	#endif

	//Functions for updating the global anim controller for this anim
	//(for those things that don't have their own controller)
	void ResetAnimationController(){m_xAnimController.Reset();}
	void SetSequence(u_int uSequence){m_xAnimController.SetSequence(uSequence);}
	void SetSpeed(float fSpeed){m_xAnimController.SetSpeed(fSpeed);}
	void UpdateGlobalControllerTimer();

	#ifdef ASURA_TOOLS
	void SetFrameOverride(int iFrame){m_xAnimController.SetFrameOverride(iFrame);}
	void ClearFrameOverride(){m_xAnimController.ClearFrameOverride();}
	#endif

	#ifdef ASURA_TOOLS
	u_int GetMaterialFlagsAddedByAnimation() const;
	u_int GetMaterialFlagsRemovedByAnimation() const;
	#endif


private :

	Asura_String m_xName;

	u_int m_uNumberOfSequences;
	Asura_TextureAnimationSequence* m_pxSequences;

	//global anim controller that is used for any instances of this anim that
	//aren't controlled by entities. Essentially for keeping track of the
	//state of anims playing on the environment
	mutable Asura_TextureAnimationController m_xAnimController;

	friend class Asura_Chunk_TextureAnimation;
	friend class Tools_TextureAnimation;
	friend class MayaAsuraImport;
	friend class MaxAsuraImport;
	friend class AsuraExporter;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_TextureAnimationSequence **
**--------------------------------*/

//A sequence is a series of frames that will be animated between , 
//possibly looping once the end is reached

class Asura_TextureAnimationSequence
{
public :
	Asura_TextureAnimationSequence();
	~Asura_TextureAnimationSequence();

	#ifdef ASURA_TOOLS
	void Copy(const Asura_TextureAnimationSequence* pxSourceSeq);
	#endif


	float GetDuration() const {return m_fDuration;}

	#ifdef ASURA_TOOLS
	//Initialise a new sequence with a default frame
	void CreateDefaultFrame();

	void AddFrameAfter(int iPrevFrame);
	void DeleteFrame(u_int uFrame);
	#endif

	Asura_TextureAnimationSequence& operator = (const Asura_TextureAnimationSequence& xSeq);

	#ifdef ASURA_TOOLS
	u_int GetMaterialFlagsAddedBySequence() const;
	u_int GetMaterialFlagsRemovedBySequence() const;
	#endif


private :
	u_int m_uNumberOfFrames;
	Asura_TextureAnimationFrame* m_pxFrames;

	float m_fDuration;
	
	union
	{
		u_int m_uFlags;
		struct
		{
			u_int m_bLoop : 1;
			u_int m_bInterpolate : 1;//interpolates UV offsets between frames
			u_int m_bRandom : 1;//Picks a random offset from 0 up to the value in the frame (for static effects)
			u_int m_bLoopBackAndForth : 1; //animation reverses upon reaching the end , instead of returning to the start
		};
	};

	friend class Asura_TextureAnimation;
	friend class Asura_Chunk_TextureAnimation;
	friend class Tools_TextureAnimation;
	friend class MayaAsuraImport;
	friend class MaxAsuraImport;
	friend class AsuraExporter;

};

/////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------**
** Asura_TextureAnimationFrame **
**-----------------------------*/

//A single frame in the sequence

struct Asura_TextureAnimationFrame
{
	float m_fU;
	float m_fV;
	float m_fTime;//time for this frame between 0 and 1

	u_int m_uAddedMaterialFlags;
	u_int m_uRemovedMaterialFlags;
};




/////////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------**
** Asura_TextureAnimationOutput **
**------------------------------*/

//Structure for giving the results of the current state of the animation.
//Currently this is just a UV offset

struct Asura_TextureAnimationOutput
{
	Asura_TextureAnimationOutput()
	{
		m_fUOffset = 0;
		m_fVOffset = 0;
	}
	
	float m_fUOffset;
	float m_fVOffset;
};



/////////////////////////////////////////////////////////////////////////////////////////////////////
/*----------**
** Messages **
**----------*/

#ifdef ASURA_TOOLS
//defines for flags for use on the editing overlay
#define ASURA_MESSAGE_TEXTUREANIMATION_CHANGESEQUENCE 0x00000001
#define ASURA_MESSAGE_TEXTUREANIMATION_CHANGESPEED 0x00000002
#endif

//This message is stored in the static message system , and so care needs to be taken about changing it
struct Asura_Message_TextureAnimation
{
	union
	{
		u_int m_uFlags;
		struct
		{
			// NOTE: These bit's are written out, adding any new bits means
			// ~~~~~ that you will need to add those bits to the bitfield
			//       preparation code in "WriteToChunkStream".
			u_int m_bChangeSequence:1;
			u_int m_bChangeSpeed:1;
		};
	};
	u_int m_uSequence;//sequence to switch to (if flag set)
	float m_fSpeed;//speed to change to (if flag set)
};


//update message containing current state of an environmental texture animation
struct Asura_Message_EnvironmentTextureAnimation
{
	Asura_Hash_ID m_uAnimHash;
	short m_sSpeed;//speed scaled by 100
	unsigned char m_ucSequence;
	#ifdef ASURA_USE_STREAMING
	Asura_Hash_ID m_uZoneHash;
	#endif

#ifdef ASURA_NETMSG_PACK_ALL_DATA
	void RegisterFields(Asura_NetMsg* pxNetMsg) const;
	static u_char GetNumFields() { return 3; }
#endif
};

///////////////////////////////////////////////////////
// Asura_TextureAnimationController_SelfIllumination //
///////////////////////////////////////////////////////
// Self-illumination properties for animation's texture which override those in the material response.
class Asura_TextureAnimationController_SelfIllumination
{
public:
	Asura_TextureAnimationController_SelfIllumination();

	void SetEnabled(bool bEnabled) { m_bEnabled = bEnabled; }
	
	void Set(float fContribution, float fColourise, const Asura_RGBA& xColour)
	{
		m_fContribution = fContribution;
		m_fColourise	= fColourise;
		m_xColour		= xColour;
		m_bEnabled		= true;
	}

	float GetContribution() const { return m_fContribution; }
	float GetColourise() const { return m_fColourise; }
	const Asura_RGBA& GetColour() const { return m_xColour; }

	bool IsEnabled() const { return m_bEnabled; }

private:
	float m_fContribution;
	float m_fColourise;
	Asura_RGBA m_xColour;
	
	bool m_bEnabled;
};

#endif // __ASURA_TEXTUREANIMATION_H__
