#ifdef ASURA_USE_ANIMATION_IK_SOLVER

///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Animation_IK_Chain.h
// Author : Will Plano
// Date   : 08/08/2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_ANIMATION_IK_CHAIN_H__
#define __ASURA_ANIMATION_IK_CHAIN_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"
#include "Physics/Asura_Physics_EnvironmentCache.h"

class Asura_Animation;
class Asura_Hierarchy_Bone_RestFrame;
class Asura_Animation_IK_Link;

////////////////////////////////////////////////////////////////////////////////
// Class  : Asura_Animation_IK_Chain
// Author : Will Plano
// Date   : 08/08/2007
//
// Description  : Class to store and handle IK chains.
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation_IK_Chain
{
public:
	// Use this constructor if we don't yet know the bone IDs or don't plan to attach the IK Chain to an animation at all.
	Asura_Animation_IK_Chain( const u_int uNumBones );

	// The arrays of bone IDs should be terminated with a -1
	Asura_Animation_IK_Chain( const int* piBoneIDs, const int* piTwistBoneIDs );

	virtual ~Asura_Animation_IK_Chain();

	virtual void Shutdown();

	void SetBonesArrays( const int* piBonesArray, const int* piTwistBonesArray = NULL );

	float GetLength() const { return m_fLength; }
	float GetExtendLimit() const { return m_fExtendLimit; }
	void SetExtendLimit( float fExtendLimit ) { m_fExtendLimit = fExtendLimit; }
	void SetDistanceFadeRange( float fStartFade, float fEndFade ) { m_fDistanceFadeStart = fStartFade; m_fDistanceFadeEnd = fEndFade; }
	float GetDistanceFadeStart() const { return m_fDistanceFadeStart; }
	float GetDistanceFadeEnd() const { return m_fDistanceFadeEnd; }
	void SetDistanceFade( float fFade ) { m_fDistanceFade = fFade; }

	void SetMoveLimit( float fMoveLimit ) { m_fMoveLimit = fMoveLimit; }
	void SetPosSmoothing( bool bSet ) { m_bNoPosSmoothing = !bSet; }
	void SetOriSmoothing( bool bSet ) { m_bNoOriSmoothing = !bSet; }
	void SetPosSmoothingSpeed( float fPosSmoothingSpeed ) { m_fPosSmoothingSpeed = fPosSmoothingSpeed; }
	void SetOriSmoothingSpeed( float fOriSmoothingSpeed ) { m_fOriSmoothingSpeed = fOriSmoothingSpeed; }

	// This function sets the anim positions and orientations of the chain links from the data in the associated Asura_Animation,
	virtual bool GetBoneDataFromAnim( const Asura_Animation& xAnim );

	// This function sets the orientations of the bones of the animation's tween store to the values stored in the IK Chain.
	virtual bool PushBoneDataToAnim( Asura_Animation& xAnim, const float fSolverLerp = 1.0f );

	const Asura_AVector3& GetBasePosition() const;
	const Asura_AVector3& GetEndPosition() const;
	float GetGroundOffset() const { return m_fGroundOffset; }

	Asura_Animation_IK_Chain* GetNextChain() const { return m_pxNextChain; }
	void SetNextChain( Asura_Animation_IK_Chain* xNextChain ) { m_pxNextChain = xNextChain; }

	// This function sets a 'fade' value to move to, it should be in the range 0.0f - 1.0f,
	// the bNow parameter sets whether the fade should be instant
	void SetTargetLerp( float fLerp, bool bNow = false );

	void SetAnimVeto( bool bAnimVeto ) { m_bAnimVeto = bAnimVeto; }

	void Enable( bool bNow = false ) { SetTargetLerp( 1.0f, bNow ); }
	void Disable( bool bNow = false ) { SetTargetLerp( 0.0f, bNow ); }
	bool IsActive() { return ( !m_bAnimVeto && ( m_fTargetLerp > 0.0f || GetCurrentLerp() > 0.0f ) ); }
	float GetCurrentLerp() { UpdateLerpTimer(); return m_fLerp * m_fDistanceFade; }
	float GetTargetLerp() const { return m_fTargetLerp; }

	bool IsRootBoneMover() const { return m_bRootBoneMover; }

	// Move all positions and targets by the specified vector
	void MoveEverything( const Asura_AVector3& xPosDelta, const Asura_Quat& xRotQuat = Asura_Maths::IdentityQuat, const Asura_AVector3& xRotCentre = Asura_Maths::ZeroAVector3 );
	
	// Sets a vector to try to orient the middle bone towards,
	// only really relevant for 2-bone IK, may change if/when we move to more general case.
	void SetElbowHintDir( const Asura_AVector3& xElbowHintDir ) { m_xElbowHintDir = xElbowHintDir; }

	// If a chain is set as a root bone mover, it will be used in the calculation of the root bone offset.
	void SetIsRootBoneMover( bool bSet = true ) { m_bRootBoneMover = bSet; }

	void SetDoJointConstraints( bool bSet = true ) { m_bDoJointConstraints = bSet; }
	void SetDoCollisionConstraints( bool bSet = true ) { m_bDoCollisionConstraints = bSet; }

	// Use this to force an update of the target position/orientation, e.g. when a parent bone has been moved.
	void SetNeedsTargetUpdate() { m_uFrameStamp = 0; }

	void SetAxesFromRestPose( Asura_Hierarchy_Bone_RestFrame* pxRestPose );

	// This functions will call CalculateTarget if it has not been called this frame before setting the values,
	// it returns false if the target pos & ori could not be calculated.
	bool GetTarget( Asura_AVector3* pxPos = NULL, Asura_Quat* pxQuatDelta = NULL, const bool bForceUpdate = false );

	bool Solve();

	bool SolveChildChains();

	// Functions to add/remove chains to/from the solver
	void AddChildChain( Asura_Animation_IK_Chain& xNewChildChain );
	void RemoveChildChain( Asura_Animation_IK_Chain& xChildChainToRemove );

	//Used to mark chains created by templates, that will need to be deallocated by the IK solver
	void SetOwnedBySolver(bool bSet) {m_bOwnedBySolver = true;}
	bool IsOwnedBySolver() const {return m_bOwnedBySolver;}

	void SetBoneCollisionRadius( const float fSet ){ m_fBoneCollisionRadius = fSet; }

	int GetFirstLinkBoneID() const;

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	static void AddConsoleVars();
#endif

protected:
	virtual bool CalculateTarget() = 0;

	virtual bool SolveInitFunction();	// return true if successful
	virtual bool SolveLoopFunction() { return false; }	// return true if some kind of constraint was applied
	virtual bool SolveEndFunction();	// return true if successful

	virtual bool CollideNodes();
	virtual bool SolveChain();
	bool SatisfyLinearConstraints();
	bool SatisfyAngularConstraints();

	void UpdateLerpTimer();

	Asura_Physics_EnvironmentCache	m_xEnvironmentCache;	// The environment-cache to use for collisions.

	Asura_AVector3 m_xElbowHintDir;				// may change when we move to >2 bone IK
	Asura_AVector3 m_xTargetPosition;			// don't use this directly, use GetTargetPosition()
	Asura_AVector3 m_xAnimPosition;

	Asura_Quat m_xTargetOrientationDelta;		// don't use this directly, use GetTargetOrientation()
	Asura_Quat m_xChainOri;
	Asura_Quat m_xAnimOrientation;
	Asura_Quat m_xAnimInvOrientation;

	u_int m_uFrameStamp;
	u_int m_uLerpFrameStamp;
	u_int m_uNumLinks;

	union
	{
		u_int m_uIKChainFlags;
		struct
		{
			u_int	m_bRootBoneMover				: 1,	// If this is set then the chain will be used in the calculation of the root bone offset
					m_bDoJointConstraints			: 1,	// Set this to honour angular and linear constraints
					m_bDoCollisionConstraints		: 1,	// Set this to test the chain against collisions ( calls the virtual function CollideNodes() in the solve loop which by default collides nodes with the environment )
					m_bAnimVeto						: 1,	// Anim flag to turn off IK.
					m_bTargetOK						: 1,	// Flag set if target is OK to use.
					m_bNoPosSmoothing				: 1,	// Dont smooth out position change.
					m_bNoOriSmoothing				: 1,	// Dont smooth out orientation change.
					m_bOwnedBySolver				: 1,	// Chain's deallocation will be handled by the IK Solver
													: 0;	// 24 bits free
		};
	};

	float m_fLength;
	float m_fLerp;
	float m_fTargetLerp;
	float m_fDistanceFade;
	float m_fDistanceFadeStart;
	float m_fDistanceFadeEnd;
	float m_fGroundOffset;
	float m_fExtendLimit;
	float m_fMoveLimit;
	float m_fPosSmoothingSpeed;
	float m_fOriSmoothingSpeed;
	float m_fBoneCollisionRadius;

	Asura_Animation_IK_Link* m_pxFirstLink;			// Pointer to start of list of links
	Asura_Animation_IK_Link* m_pxLastLink;			// Pointer to end of list of links
	Asura_Animation_IK_Chain* m_pxNextChain;		// Next pointer to make a list of these guys
	Asura_Animation_IK_Chain* m_pxFirstChildChain;	// Pointer to first child chain

}; // end class Asura_Animation_IK_Chain

#endif   // ifndef __ASURA_ANIMATION_IK_CHAIN_H__

#endif // ASURA_USE_ANIMATION_IK_SOLVER

/// End of file. ///
