#ifdef ASURA_USE_ANIMATION_IK_SOLVER

///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Animation_IK_Solver.h
// Author : Will Plano
// Date   : 08/08/2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_ANIMATION_IK_SOLVER_H__
#define __ASURA_ANIMATION_IK_SOLVER_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"

#ifndef ASURA_IK_FADE_NEAR_RANGE
#define ASURA_IK_FADE_NEAR_RANGE 15.0f
#endif

#ifndef ASURA_IK_FADE_FAR_RANGE
#define ASURA_IK_FADE_FAR_RANGE 30.0f
#endif

////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation;
class Asura_Animation_IK_Chain;
class Asura_Animation_IK_Template;

////////////////////////////////////////////////////////////////////////////////
// Class  : Asura_Animation_IK_Solver
// Author : Will Plano
// Date   : 08/08/2007
//
// Description  : Class to handle solution of inverse kinematics problems.
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation_IK_Solver
{
public:
	enum RootBoneLerpState
	{
		ROOT_BONE_OUT,
		ROOT_BONE_LERP_IN,
		ROOT_BONE_IN,
		ROOT_BONE_LERP_OUT
	};

	Asura_Animation_IK_Solver();
	~Asura_Animation_IK_Solver();
	
	void Shutdown();

	//Allocate and setup an new IK solver from a template.
	//Returns 0 if the template wasn't found
	static Asura_Animation_IK_Solver* CreateFromTemplate(Asura_Hash_ID uTemplate, const Asura_Animation* pxAnim);

	// Functions to add/remove chains to/from the solver
	void AddChain( Asura_Animation_IK_Chain& xNewChain );
	void RemoveChain( Asura_Animation_IK_Chain& xChainToRemove );
	// The solver doesn't handle allocation / deallocation of chains.
	// It does now in the case of chains created from templates . RAR 12/05/2008
	void Clear(); 

	Asura_Animation_IK_Chain* GetFirstChain() const { return m_pxFirstChain; }

	// Sets a parameter to control fading in and out of IK-ness, set to 1.0f for full IK control, 0.0f to turn IK off and anywhere in between to fade.
	void SetLerp( const float fLerp ){ m_fLerp = fLerp; }
	void Enable() { SetLerp( 1.0f ); }
	void Disable() { SetLerp( 0.0f ); }
	float GetLerp() const { return m_fLerp; }
	bool IsActive() const { return m_fLerp > 0.0f; }

	void SetMinDist( float fMinDist ) { m_fMinDist = fMinDist;}
	void SetMaxDist( float fMaxDist ) { m_fMaxDist = fMaxDist;}

	void EnableRootBoneMovement( const float fLerpTime );
	void DisableRootBoneMovement( const float fLerpTime );
	void SetRootRootBoneMaxMovement( const float fMaxDist ) { m_fRootBoneMaxMovement = fMaxDist; }
	float GetRootRootBoneMaxMovement() const { return m_fRootBoneMaxMovement; }

	// The meat of the IKSolver, this function updates the attached IK chains from the animation,
	// solves the chains, calculates the desired root bone offset and sets the tween store to the new bone data.
	void UpdateBones( Asura_Animation& xAnim );

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	static void AddConsoleVars();
#endif

private:
	void UpdateRootBoneLerp();

	float						m_fLerp;
	float m_fMinDist;
	float m_fMaxDist;

	Asura_Animation_IK_Chain*	m_pxFirstChain;
	Asura_Vector_3				m_xSmoothedPositionDelta;

	u_int						m_uLastKnwonVisibilityNode;
	RootBoneLerpState			m_eRootBoneLerpState;
	float						m_fRootBoneLerp;
	float						m_fRootBoneLerpTime;
	float						m_fRootBoneMaxMovement;

}; // end class Asura_Animation_IK_Solver
#endif // ifndef __ASURA_ANIMATION_IK_SOLVER_H__

#endif // ASURA_USE_ANIMATION_IK_SOLVER

/// End of file. ///
