//
// Asura_Attractor_Compiled.cpp
//
// Created by JGP 10/05/06
//
// Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_ATTRACTORS

// This file's header
#include "Attractors/Asura_Attractor_Compiled.h"

#include "Attractors/Asura_Attractor_Node_Data.h"
#include "Core/Asura_Maths.h"

bool Asura_Attractor_Compiled::Contains( const Asura_Vector_3& xPos, float fEndAdjustment /* = 0.f */, float fRadiusOverride /* = 0.f */) const
{
	const Asura_Vector_3& xNode1 = m_axNodes[0].GetPosition();
	const Asura_Vector_3& xNode2 = m_axNodes[1].GetPosition();

	//get a normalised vector between the two nodes
	Asura_Vector_3 xAxis = xNode2 - xNode1;
	const float fAxisLength = xAxis.Normalise();
	Asura_Assertf(fAxisLength > 0.0001f, ("Ridiculously short attractor? (ID: %i)", GetID()));

	// Fail if the position falls outside the caps.
	const Asura_Vector_3 xNode1ToPos = xPos - xNode1;
	const float fAxisDistance = xNode1ToPos * xAxis;
	if(fAxisDistance < -fEndAdjustment || fAxisDistance > fAxisLength + fEndAdjustment)
	{
		return false;
	}

	// Fail if the perpendicular distance is larger than the radius.
	const float fRadius = (fRadiusOverride > 0.f) ? fRadiusOverride : GetRadius();
	const float fRadiusSquared = fRadius * fRadius;
	const Asura_Vector_3 xClosestPointOnAxis = xNode1 + (xAxis * fAxisDistance);
	const Asura_Vector_3 xPerpendicularDir = xPos - xClosestPointOnAxis;
	if(xPerpendicularDir.MagnitudeSquared() > fRadiusSquared)
	{
		return false;
	}

	// Must be inside the cylinder at this point.
	return true;
}

void Asura_Attractor_Compiled::GetRelativePosition(	const Asura_Vector_3& xPos,
													float fEndAdjustment /* = 0.f */,
													float* pfDistanceAlongAxis /* = NULL */,
													Asura_Vector_3* pxClosestPointOnAxis /* = NULL */,
													float* pfDistanceFromAxis /* = NULL */,
													Asura_Vector_3* pxVectorFromAxis /* = NULL */,
													bool bClampClosestPosToEnds /* = true  */) const
{
	const Asura_Vector_3& xNode1 = m_axNodes[0].GetPosition();
	const Asura_Vector_3& xNode2 = m_axNodes[1].GetPosition();

	//get a normalised vector between the two nodes
	Asura_Vector_3 xAxis = xNode2 - xNode1;
	const float fAxisLength = xAxis.Normalise();
	// Fail if the nodes are too close together
	Asura_Assertf(fAxisLength > 0.0001f, ("Ridiculously short attractor? (ID: %i)", GetID()));

	const Asura_Vector_3 xNode1ToPos = xPos - xNode1;
	const float fUnclampedAxisDistance = xNode1ToPos * xAxis;
	const float fOutputAxisDistance = bClampClosestPosToEnds ? Asura_Maths::ClampToRange(fUnclampedAxisDistance, -fEndAdjustment, fAxisLength + fEndAdjustment) : fUnclampedAxisDistance;	
	const Asura_Vector_3 xClosestPointOnAxis = xNode1 + (xAxis * fOutputAxisDistance);
	const Asura_Vector_3 xPerpendicularDir = xPos - xClosestPointOnAxis;

	// If this comes up high on the profiler, we could only calculate the specific values asked for.
	// However, it shouldn't really be called that often, and excessive branching will probably be
	// just as expensive as the calculations themselves, which are fairly simple vector arithmetic.
	if(pfDistanceAlongAxis)
	{
		*pfDistanceAlongAxis = fOutputAxisDistance;
	}
	if(pxClosestPointOnAxis)
	{
		*pxClosestPointOnAxis = xClosestPointOnAxis;
	}
	if(pfDistanceFromAxis)
	{
		*pfDistanceFromAxis = xPerpendicularDir.Magnitude();
	}
	if(pxVectorFromAxis)
	{
		*pxVectorFromAxis = xPerpendicularDir;
	}
}

float Asura_Attractor_Compiled::GetAxisLength() const
{
	const Asura_Vector_3& xNode1 = m_axNodes[0].GetPosition();
	const Asura_Vector_3& xNode2 = m_axNodes[1].GetPosition();
	const Asura_Vector_3 xAxis = xNode2 - xNode1;
	return xAxis.Magnitude();
}

void Asura_Attractor_Compiled::GetAxisDirection(Asura_Vector_3& xDirection) const
{
	const Asura_Vector_3& xNode1 = m_axNodes[0].GetPosition();
	const Asura_Vector_3& xNode2 = m_axNodes[1].GetPosition();
	const Asura_Vector_3 xAxis = xNode2 - xNode1;
	xDirection = xAxis;
	xDirection.Normalise();
}

bool Asura_Attractor_Compiled::LOSTest(const Asura_Vector_3& xPos, const Asura_Vector_3& xDir, float fDistance) const
{
	//make sure the cylinder radius is non-zero
	const float fRadius = GetRadius();
	if(fRadius <= 0.0f) return false;
	const float fRadiusSquared = fRadius*fRadius;


	const Asura_Vector_3& xNode1 = m_axNodes[0].GetPosition();
	const Asura_Vector_3& xNode2 = m_axNodes[1].GetPosition();

	//get a normalised vector between the two nodes
	Asura_Vector_3 xAxis = xNode2-xNode1;
	float fMag = xAxis.Magnitude();
	if(fMag<0.0001f) return false;
	xAxis /= fMag;

	//do a sphere test on the centre of the line before checking anything else
	Asura_Vector_3 xCentre = (xNode1+xNode2)*0.5f;
	float fIntersectionDist;
	if(!Asura_Maths::RayIntersectsSphere(xPos,xDir,xCentre,fRadius+fMag,&fIntersectionDist)) return false;
	if(fIntersectionDist > fDistance) return false;

	// Test against spheres at the nodes for the ends of our capsule.
	if (Asura_Maths::RayIntersectsSphere(xPos, xDir, xNode1, fRadius, &fIntersectionDist))
	{
		if (fIntersectionDist <= fDistance) { return true; }
	}
	if (Asura_Maths::RayIntersectsSphere(xPos, xDir, xNode2, fRadius, &fIntersectionDist))
	{
		if (fIntersectionDist <= fDistance) { return true; }
	}

	////Check for the closest intersection between the LOS line and the axis

	//Get a perpendicular to the axis that is in the plane of the two lines
	Asura_Vector_3 xPerp = xAxis ^ xDir;
	Asura_Vector_3 xAxisPerp = xAxis ^ xPerp;
	fMag = xAxisPerp.Magnitude();

	if(fMag > 0.0001f)
	{
		xAxisPerp /= fMag;
		
		//get the distance until the LOS line is closest to the axis
		fIntersectionDist = (xNode1-xPos)*xAxisPerp / (xDir*xAxisPerp);

		if(fIntersectionDist >= 0.0f && fIntersectionDist <= fDistance)
		{
			//Is the closest point within the radius?
			Asura_Vector_3 xIntersection = xPos + xDir*fIntersectionDist;
			Asura_Vector_3 xOffset = xIntersection - xNode1;
			xOffset -= xAxis*(xOffset*xAxis);
			if(xOffset.MagnitudeSquared() <= fRadiusSquared)
			{
				//Is it within the axis line segment
				float fDot1 = (xIntersection-xNode1)*xAxis;
				float fDot2 = (xNode2-xIntersection)*xAxis;
				if(fDot1 * fDot2 >= 0.0f) return true;
			}
		}
	}

	//// Is the start or end point within the cylinder?
	{
		const Asura_Vector_3& xIntersection = xPos;
		Asura_Vector_3 xPerpOffset = xIntersection - xNode1;
		xPerpOffset -= xAxis*(xPerpOffset*xAxis);
		if(xPerpOffset.MagnitudeSquared() <= fRadiusSquared)
		{
			float fDot1 = (xIntersection-xNode1)*xAxis;
			float fDot2 = (xNode2-xIntersection)*xAxis;
			if(fDot1 * fDot2 >= 0.0f) return true;
		}
	}

	{
		Asura_Vector_3 xIntersection = xPos + xDir*fDistance;
		Asura_Vector_3 xPerpOffset = xIntersection - xNode1;
		xPerpOffset -= xAxis*(xPerpOffset*xAxis);
		if(xPerpOffset.MagnitudeSquared() <= fRadiusSquared)
		{
			float fDot1 = (xIntersection-xNode1)*xAxis;
			float fDot2 = (xNode2-xIntersection)*xAxis;
			if(fDot1 * fDot2 >= 0.0f) return true;
		}
	}


	return false;
}

bool Asura_Attractor_Compiled::CouldIntersect(const Asura_Bounding_Box& xBB) const
{
	// If performance becomes a problem this could be cached.
	Asura_Bounding_Box xAttractorBB;
	xAttractorBB.Reset();
	xAttractorBB.Set(m_axNodes[0].GetPosition());
	xAttractorBB.Set(m_axNodes[1].GetPosition());
	xAttractorBB.Grow(GetRadius());
	return xAttractorBB.Intersects(xBB);
}


#endif // ASURA_USE_ATTRACTORS
