//////////////////////////////////////////////////////////////////////////////
//
// Asura_Blueprint.cpp
//
// Created by NJD - 06 September 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Blueprint/Asura_Blueprint.h"

// asura
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Profiling_System.h"

#ifdef ASURA_TOOLS
// tools
#include "Blueprint/Tools_Blueprint_Edit_Object.h"
#endif 

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS
const char* s_apszOverrideTypeStrings[ ABOT_END_OVERRIDE_TYPES ] =
{
    "None",
    "Scale",
    "Offset",
    "Fixed"
};

extern const char* s_apszBooleanStrings[ 2 ] =
{
    "false",
    "true"
};
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Value
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS
const char* Asura_Value::s_apszValueTypeStrings[ VT_END_VALUE_TYPES ] =
{
    "Int",
    "Float",
    "Boolean",
    "Hash String"
};

Asura_TemplateHashTree< Asura_String > Asura_Value::s_xHashIDStringTree;
#endif

void Asura_Value::SetValue( const int iValue )
{
    m_eValueType      = VT_INT;
	m_xValue.m_iValue = iValue;
}

void Asura_Value::SetValue( const float fValue )
{
    m_eValueType      = VT_FLOAT;
	m_xValue.m_fValue = fValue;
}

void Asura_Value::SetValue( const bool bValue )
{
    m_eValueType      = VT_BOOL;
	m_xValue.m_bValue = bValue;
}

void Asura_Value::SetValue( const Asura_Hash_ID uHashID, const Asura_Char* pszString /*= NULL*/ )
{
    m_eValueType      = VT_HASH;
    m_xValue.m_uValue = uHashID;

#ifdef ASURA_TOOLS
    m_uHashDisplayTextHash = pszString ? Asura_GetHashID( pszString ) : ASURA_HASH_ID_UNSET;

    if ( m_uHashDisplayTextHash != ASURA_HASH_ID_UNSET )
    {
        Asura_String* pxString = s_xHashIDStringTree.GetDataByHash( m_uHashDisplayTextHash );
        if ( !pxString )
        {
            // Not found, add a new node and update the string.
            pxString = s_xHashIDStringTree.AddNode( m_uHashDisplayTextHash );
            if ( pxString )
            {	
                *pxString = pszString;
            }
        }
    }
#endif
}

void Asura_Value::SetValue( const Asura_Value& xValue )
{
    m_xValue     = xValue.m_xValue;
    m_eValueType = xValue.m_eValueType;
#ifdef ASURA_TOOLS
    m_uHashDisplayTextHash = xValue.m_uHashDisplayTextHash;
#endif
}

#ifdef ASURA_TOOLS
const Asura_Char* Asura_Value::GetHashStringValue() const
{
    if ( m_uHashDisplayTextHash == ASURA_HASH_ID_UNSET )
    {
        return "<UNSET>";
    }

	Asura_String* pxString = Asura_Value::s_xHashIDStringTree.GetDataByHash( m_uHashDisplayTextHash );
	if ( pxString )
	{
		return pxString->GetString();
	}

	// Returning NULL is really an error case: we've totally failed to find a string to match this hash ID.
	Asura_Assertf( false, ( "Asura_Value::GetHashStringValue - failed to find hash string value for hash ID value '0x%X'", m_xValue.m_uValue ) );

	return NULL;
}
#endif

void Asura_Value::ApplyOverride( const Asura_Blueprint_Override_Type eOverrideType, const Asura_Value& xValue )
{
#ifdef ASURA_TOOLS
    Asura_Assert( IsValidOverride( GetValueType(), xValue.GetValueType(), eOverrideType ), "Asura_Value::ApplyOverride - bad override" );
#endif

    switch ( eOverrideType )
    {
        case ABOT_SCALE:
        {
            switch ( xValue.GetValueType() )
            {
                case VT_INT:
                {
                    const int iValue = xValue.GetIntValue();
                    switch ( GetValueType() )
                    {
                        case VT_INT:   SetValue( GetIntValue() * iValue );   break;
                        case VT_FLOAT: SetValue( GetFloatValue() * iValue ); break;
                        default: break;
                    }
                } break;

                case VT_FLOAT:
                {
                    const float fValue = xValue.GetFloatValue();
                    switch ( GetValueType() )
                    {
                        case VT_INT:   SetValue( Asura_Maths::FloatToInt_Near( GetIntValue() * fValue ) ); break;
                        case VT_FLOAT: SetValue( GetFloatValue() * fValue );                               break;
                        default: break;
                    }
                } break;

                //case VT_BOOL:
                //{
                    // Can't apply a scale boolean override.
                //} break;

                //case VT_HASH:
                //{
                    // Can't apply a scale hash ID override.
                //} break;

                default:
                {
                    Asura_Assert( false, "Asura_Value::ApplyOverride - unhandled value type." );
                } break;
            }
        } break;

        case ABOT_OFFSET:
        {
            switch ( xValue.GetValueType() )
            {
                case VT_INT:
                {
                    const int iValue = xValue.GetIntValue();
                    switch ( GetValueType() )
                    {
                        case VT_INT:   SetValue( GetIntValue() + iValue );   break;
                        case VT_FLOAT: SetValue( GetFloatValue() + iValue ); break;
                        default: break;
                    }
                } break;

                case VT_FLOAT:
                {
                    const float fValue = xValue.GetFloatValue();
                    switch ( GetValueType() )
                    {
                        case VT_INT:   SetValue( Asura_Maths::FloatToInt_Near( GetIntValue() + fValue ) ); break;
                        case VT_FLOAT: SetValue( GetFloatValue() + fValue );                               break;
                        default: break;
                    }
                } break;

                //case VT_BOOL:
                //{
                    // Can't apply an offset boolean override.
                //} break;

                //case VT_HASH:
                //{
                    // Can't apply an offset hash ID override.
                //} break;

                default:
                {
                    Asura_Assert( false, "Asura_Value::ApplyOverride - unhandled value type." );
                } break;
            }
        } break;

        case ABOT_FIXED:
        {
            switch ( xValue.GetValueType() )
            {
                case VT_INT:
                {
                    switch ( GetValueType() )
                    {
                        case VT_INT:   SetValue( xValue.GetIntValue() );                         break;
                        case VT_FLOAT: SetValue( static_cast< float >( xValue.GetIntValue() ) ); break;
                        default: break;
                    }
                } break;

                case VT_FLOAT:
                {
                    switch ( GetValueType() )
                    {
                        case VT_INT:   SetValue( Asura_Maths::FloatToInt_Near( xValue.GetFloatValue() ) ); break;
                        case VT_FLOAT: SetValue( xValue.GetFloatValue() );                                 break;
                        default: break;
                    }
                } break;

                case VT_BOOL:
                {
                    switch ( GetValueType() )
                    {
                        case VT_BOOL: SetValue( xValue.GetBoolValue() ); break;
                        default: break;
                    }
                } break;

				case VT_HASH:
				{
					switch ( GetValueType() )
					{
    					case VT_HASH:
#ifdef ASURA_TOOLS
                            SetValue( xValue.GetHashIDValue(), xValue.GetHashStringValue() );
#else
                            SetValue( xValue.GetHashIDValue() );
#endif
                            break;
                        default: break;
					}
				} break;

                default:
                {
                    Asura_Assert( false, "Asura_Value::ApplyOverride - unhandled value type." );
                } break;
            }
        } break;

        case ABOT_NONE:
        {
        } break;

        default:
        {
            Asura_Assert( false, "Axon_Value_Base::ApplyOverride - unexpected override type" );
        } break;
    }
}

#ifdef ASURA_TOOLS
bool Asura_Value::IsValidOverride( const VALUE_TYPE eBaseValueType, const VALUE_TYPE eOverrideValueType, const Asura_Blueprint_Override_Type eOverrideType )
{
    switch ( eOverrideType )
    {
        case ABOT_SCALE:
        {
            switch ( eOverrideValueType )
            {
                case VT_INT:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_INT:
                        case VT_FLOAT: return true;
                        default:       return false;
                    }
                } break;

                case VT_FLOAT:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_INT:
                        case VT_FLOAT: return true;
                        default:       return false;
                    }
                } break;

                //case VT_BOOL:
                //{
                    // Can't apply a scale boolean override to anything.
                //    return false;
                //} break;

                //case VT_HASH:
                //{
                    // Can't apply a scale hash ID override to anything.
                //    return false;
                //} break;

                default:
                {
                    return false;
                } break;
            }
        } break;

        case ABOT_OFFSET:
        {
            switch ( eOverrideValueType )
            {
                case VT_INT:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_INT:
                        case VT_FLOAT: return true;
                        default:       return false;
                    }
                } break;

                case VT_FLOAT:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_INT:
                        case VT_FLOAT: return true;
                        default:       return false;
                    }
                } break;

                //case VT_BOOL:
                //{
                    // Can't apply an offset boolean override to anything.
                //    return false;
                //} break;

                //case VT_HASH:
                //{
                    // Can't apply an offset hash ID override to anything.
                //    return false;
                //} break;

                default:
                {
                    return false;
                } break;
            }
        } break;

        case ABOT_FIXED:
        {
            switch ( eOverrideValueType )
            {
                case VT_INT:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_INT:
                        case VT_FLOAT: return true;
                        default:       return false;
                    }
                } break;

                case VT_FLOAT:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_INT:
                        case VT_FLOAT: return true;
                        default:       return false;
                    }
                } break;

                case VT_BOOL:
                {
                    switch ( eBaseValueType )
                    {
                        case VT_BOOL: return true;
                        default:      return false;
                    }
                } break;

				case VT_HASH:
				{
					switch ( eBaseValueType )
					{
						case VT_HASH: return true;
						default:      return false;
					}
				} break;

                default:
                {
                    return false;
                } break;
            }
        } break;

        case ABOT_NONE: return true;
        default:        return false;
    }
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Parameter
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS
Asura_TemplateHashTree< Asura_String > Asura_Parameter::s_xGroupNameTree;
const char**                           Asura_Parameter::s_ppszGroupNameList = NULL;

void Asura_Parameter::BuildGroupNameTreeAndList()
{
    s_xGroupNameTree.Reset();

    // Build the tree first for speed.
    // Always add the 'no group' empty string.
    s_xGroupNameTree.AddNode( ASURA_HASH_ID_UNSET, Asura_String() );

	for ( Asura_Blueprint_Parameter_ConstIterator xBPParamIter; !xBPParamIter.Done(); xBPParamIter.Next() )
	{
		const Asura_Parameter* pxParam = xBPParamIter.GetCurrent();
		if ( pxParam->GetOverrideType() == ABOT_NONE )
		{
			const Asura_String& xGroupName       = pxParam->GetGroupName();
			const Asura_Hash_ID uGroupNameHashID = Asura_GetHashID( xGroupName.GetString() );
			s_xGroupNameTree.AddNode( uGroupNameHashID, xGroupName );
		}
	}

	// Now build the list.
    BuildGroupListFromTree();
}

void Asura_Parameter::ReleaseGroupNameStorage()
{
    s_xGroupNameTree.Reset();
    delete [] s_ppszGroupNameList;
    s_ppszGroupNameList = NULL;
}

int Asura_Parameter::GetGroupListIndex( const Asura_String& xGroupName )
{
    // The index into the list is identical to the index into the tree.
    if ( xGroupName.Compare( "*None*" ) == 0 )
    {
        // We're mapping the blank 'no group' string to "*None*" in the UI.
        return 0;
    }
    else
    {
        return s_xGroupNameTree.GetIndexOfHash( Asura_GetHashID( xGroupName.GetString() ) );
    }
}

void Asura_Parameter::GetGroupName( const int iGroupListIndex, Asura_String& xGroupName )
{
    // If we have no blueprints loaded this array won't have been set up.
    if ( !s_ppszGroupNameList )
    {
        xGroupName.MakeNull();
    }
    else if ( strcmp( s_ppszGroupNameList[ iGroupListIndex ], "*None*" ) == 0 )
    {
        // We're mapping the blank 'no group' string to "*None*" in the UI.
        xGroupName = Asura_String();
    }
    else
    {
        xGroupName = s_ppszGroupNameList[ iGroupListIndex ];
    }
}

void Asura_Parameter::AddGroupName( const Asura_String& xGroupName )
{
    // We should have already checked that this isn't a duplicate.
    s_xGroupNameTree.AddNode( Asura_GetHashID( xGroupName.GetString() ), xGroupName );
    BuildGroupListFromTree();
}

void Asura_Parameter::BuildGroupListFromTree()
{
    delete [] s_ppszGroupNameList;
    s_ppszGroupNameList = NULL;

    const int iNumTreeNodes = s_xGroupNameTree.GetNumberOfNodes();
    if ( iNumTreeNodes )
    {
        s_ppszGroupNameList = new const char* [ iNumTreeNodes ];

        for ( int iNodeIndex = 0; iNodeIndex < iNumTreeNodes; ++iNodeIndex )
        {
            const Asura_String* pxGroupName = s_xGroupNameTree.GetDataByIndex( iNodeIndex );
            if ( !pxGroupName ) { continue; }

            // We're mapping the blank 'no group' string to "*None*" in the UI.
            if ( pxGroupName->IsBlank() )
            {
                s_ppszGroupNameList[ iNodeIndex ] = "*None*";
            } 
            else
            {
                s_ppszGroupNameList[ iNodeIndex ] = pxGroupName->GetString();
            }
        }
    }
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Blueprint
//
//////////////////////////////////////////////////////////////////////////////

Asura_Blueprint::Asura_Blueprint()
: m_uHashID( ASURA_HASH_ID_UNSET )
, m_uParentHashID( ASURA_HASH_ID_UNSET )
, m_pxParent( NULL )
, m_xParameters()
#ifdef ASURA_TOOLS
, m_bExported(false)
, m_bChildBlueprintsExported(false)
, m_bToBeDeleted( false )
#endif
{
}

Asura_Blueprint::Asura_Blueprint( const Asura_Blueprint& xRhs )
: m_uHashID( ASURA_HASH_ID_UNSET )
, m_uParentHashID( ASURA_HASH_ID_UNSET )
, m_pxParent( NULL )
, m_xParameters()
#ifdef ASURA_TOOLS
, m_bExported(false)
, m_bChildBlueprintsExported(false)
, m_bToBeDeleted( false )
#endif
{
	*this = xRhs;
}

bool Asura_Blueprint::GetParameterValueByHashID( const Asura_Hash_ID uParameterHashID, Asura_Value& xValue ) const
{
    Asura_Parameter* pxParameter = GetParameterByHashID( uParameterHashID, false );
    if ( pxParameter )
    {
        if ( pxParameter->GetOverrideType() == ABOT_NONE )
        {
            xValue = pxParameter->GetValue();
            return true;
        }
        else
        {
            if ( m_pxParent && m_pxParent->GetParameterValueByHashID( uParameterHashID, xValue ) )
            {
                xValue.ApplyOverride( pxParameter->GetOverrideType(), pxParameter->GetValue() );
                return true;
            }
            else
            {
                return false;
            }
        }
    }
    else if ( m_pxParent )
    {
        return m_pxParent->GetParameterValueByHashID( uParameterHashID, xValue );
    }
    else
    {
        return false;
    }
}

bool Asura_Blueprint::GetParameterValueByIndex( const int iParameterIndex, Asura_Value& xValue ) const
{
	const Asura_Parameter* pxParameter = GetParameterByIndex( iParameterIndex );
	if ( pxParameter )
	{
        if ( pxParameter->GetOverrideType() == ABOT_NONE )
        {
            xValue = pxParameter->GetValue();
            return true;
        }
        else
        {
            if ( m_pxParent && m_pxParent->GetParameterValueByHashID( pxParameter->GetHashID(), xValue ) )
            {
                xValue.ApplyOverride( pxParameter->GetOverrideType(), pxParameter->GetValue() );
                return true;
            }
            else
            {
                return false;
            }
        }
    }
    else
    {
        Asura_Assert( false, "Asura_Blueprint::GetParameterValueByIndex - bad index" );
        return false;
    }
}

void Asura_Blueprint::EstablishLinkage( const Asura_Hash_ID uTypeHashID )
{
    if ( m_uParentHashID == ASURA_HASH_ID_UNSET )
    {
        m_pxParent = NULL;
    }
    else
    {
        m_pxParent = Asura_Blueprint_System::GetBlueprintByHashID( uTypeHashID, m_uParentHashID );
        Asura_Assert( m_pxParent, "Asura_Blueprint::EstablishLinkage - could not find parent" );
    }
}

bool Asura_Blueprint::IsChildOf( const Asura_Hash_ID uBlueprintTypeHashID, const Asura_Hash_ID uParentBlueprintHashID, const bool bOnlyDirectChild ) const
{
    if ( m_uParentHashID == ASURA_HASH_ID_UNSET || !m_pxParent )
    {
        return false;
    }
    else if ( m_uParentHashID == uParentBlueprintHashID )
    {
        return true;
    }
    else if ( bOnlyDirectChild )
    {
        return false;
    }
    else
    {
		return m_pxParent->IsChildOf( uBlueprintTypeHashID, uParentBlueprintHashID, false );
    }
}

int Asura_Blueprint::GetNumParameters() const
{
	return static_cast< int >( m_xParameters.GetNumEntries() );
}

Asura_Parameter* Asura_Blueprint::GetParameterByIndex( const int iParameterIndex ) const
{
	return m_xParameters.GetByIndex( iParameterIndex );
}

Asura_Parameter* Asura_Blueprint::GetParameterByHashID( const Asura_Hash_ID uParameterHashID, const bool bLookUpHierarchy /*=false*/ ) const
{
	// See if we have it first.
	Asura_Parameter* pxParam = m_xParameters.GetByHash( uParameterHashID );

	// If not then look up the hierarchy if we should.
	if ( !pxParam && bLookUpHierarchy )
	{
		const Asura_Blueprint* pxParent = GetParent();
		if ( pxParent )
		{
			pxParam = pxParent->GetParameterByHashID( uParameterHashID, true );
		}
	}

	return pxParam;
}

float Asura_Blueprint::GetParameterAsFloat( Asura_Hash_ID uParameterID, float fDefaultValue ) const
{
	float fValue = fDefaultValue;

	Asura_Value xTemp;
	if( GetParameterValueByHashID( uParameterID, xTemp ))
	{
		if	(xTemp.GetValueType() == Asura_Value::VT_FLOAT)  
		{
			fValue = xTemp.GetFloatValue();		
		}
		else
		{
#ifdef ASURA_TOOLS
			Asura_Assertf(0, ("GetBlueprintValue as Float, value was found, but type was unexpected... wrong default value? (%s) , %s", GetNameIncludingAllParentNames().GetString(), GetParameterByHashID( uParameterID, true)->GetName().GetString()));
#else
			Asura_Assert(0, "GetBlueprintValue as Float, value was found, but type was unexpected... wrong default value? ");
#endif
		}

	}
#ifdef ASURA_ASSERT
	else
	{
#ifdef ASURA_TOOLS
		Asura_Assertf( false, ( "GetBlueprintValue as Float, value with hash ID %08X not found, using default value (%f) (blueprint path: '%s')", uParameterID, Asura_Maths::FtoD(fDefaultValue), GetNameIncludingAllParentNames().GetString() ) );
#else
		Asura_Assertf( false, ( "GetBlueprintValue as Float, value with hash ID %08X not found, using default value (%f)", uParameterID, Asura_Maths::FtoD(fDefaultValue) ) );
#endif
	}	
#endif

	return fValue;
}

int Asura_Blueprint::GetParameterAsInt( Asura_Hash_ID uParameterID, int iDefaultValue ) const
{
	int iValue = iDefaultValue;

	Asura_Value xTemp;
	if( GetParameterValueByHashID( uParameterID, xTemp ))
	{
		if (xTemp.GetValueType() == Asura_Value::VT_INT) 
		{
			iValue = xTemp.GetIntValue();		
		}	
		else
		{
#ifdef ASURA_TOOLS
			Asura_Assertf(0, ("GetBlueprintValue as Int, value was found, but type was unexpected... wrong default value? (%s) , %s", GetNameIncludingAllParentNames().GetString(), GetParameterByHashID( uParameterID, true)->GetName().GetString()));
#else
			Asura_Assert(0, "GetBlueprintValue as Int, value was found, but type was unexpected... wrong default value? ");
#endif
		}
	}
#ifdef ASURA_ASSERT
	else
	{
#ifdef ASURA_TOOLS
		Asura_Assertf( false, ( "GetBlueprintValue as Int, value with hash ID %08X not found, using default value (%d) (blueprint path: '%s')", uParameterID, iDefaultValue, GetNameIncludingAllParentNames().GetString() ) );
#else
		Asura_Assertf( false, ( "GetBlueprintValue as Int, value with hash ID %08X not found, using default value (%d)", uParameterID, iDefaultValue ) );
#endif
	}
#endif

	return iValue;
}

u_int Asura_Blueprint::GetParameterAsUInt( Asura_Hash_ID uParameterID, u_int uDefaultValue ) const
{
	Asura_Hash_ID uValue = uDefaultValue;

	Asura_Value xTemp;
	if( GetParameterValueByHashID( uParameterID, xTemp ))
	{
		if (xTemp.GetValueType() == Asura_Value::VT_INT)  
		{
			uValue = static_cast< u_int >( xTemp.GetIntValue() );
		}	
		else
		{
#ifdef ASURA_TOOLS
			Asura_Assertf(0, ("GetBlueprintValue as UInt, value was found, but type was unexpected... wrong default value? (%s) , %s", GetNameIncludingAllParentNames().GetString(), GetParameterByHashID( uParameterID, true)->GetName().GetString()));
#else
			Asura_Assert(0, "GetBlueprintValue as UInt, value was found, but type was unexpected... wrong default value? ");
#endif		
		}
	}
#ifdef ASURA_ASSERT
	else
	{
#ifdef ASURA_TOOLS
		Asura_Assertf( false, ( "GetBlueprintValue as UInt, value with hash ID %08X not found, using default value (%08X)  (blueprint path: '%s')", uParameterID, uDefaultValue, GetNameIncludingAllParentNames().GetString() ) );
#else
		Asura_Assertf( false, ( "GetBlueprintValue as UInt, value with hash ID %08X not found, using default value (%08X)", uParameterID, uDefaultValue ) );
#endif
	}
#endif

	return uValue;
}

bool Asura_Blueprint::GetParameterAsBool( Asura_Hash_ID uParameterID, bool bDefaultValue ) const
{
	bool bValue = bDefaultValue;

	Asura_Value xTemp;
	if( GetParameterValueByHashID( uParameterID, xTemp ))
	{
		if (xTemp.GetValueType() == Asura_Value::VT_BOOL) 
		{
			bValue = xTemp.GetBoolValue();		
		}
		else
		{
#ifdef ASURA_TOOLS
			Asura_Assertf(0, ("GetBlueprintValue as Bool, value was found, but type was unexpected... wrong default value? (%s) , %s", GetNameIncludingAllParentNames().GetString(), GetParameterByHashID( uParameterID, true)->GetName().GetString()));
#else
			Asura_Assert(0, "GetBlueprintValue as Bool, value was found, but type was unexpected... wrong default value? ");
#endif		
		}
	}
#ifdef ASURA_ASSERT
	else
	{
#ifdef ASURA_TOOLS
		Asura_Assertf( false, ( "GetBlueprintValue as Bool, value with hash ID %08X not found, using default value (%s) (blueprint path: '%s')", uParameterID, bDefaultValue ? "true" : "false", GetNameIncludingAllParentNames().GetString() ) );
#else
		Asura_Assertf( false, ( "GetBlueprintValue as Bool, value with hash ID %08X not found, using default value (%s)", uParameterID, bDefaultValue ? "true" : "false" ) );
#endif
	}
#endif

	return bValue;
}

Asura_Hash_ID Asura_Blueprint::GetParameterAsHash( Asura_Hash_ID uParameterID, Asura_Hash_ID uDefaultValue ) const
{
	Asura_Hash_ID uValue = uDefaultValue;

	Asura_Value xTemp;
	if( GetParameterValueByHashID( uParameterID, xTemp ))
	{
		if (xTemp.GetValueType() == Asura_Value::VT_HASH)  
		{
			uValue = xTemp.GetHashIDValue();
		}
		else
		{
#ifdef ASURA_TOOLS
			Asura_Assertf(0, ("GetBlueprintValue as Hash, value was found, but type was unexpected... wrong default value? (%s) , %s", GetNameIncludingAllParentNames().GetString(), GetParameterByHashID( uParameterID, true)->GetName().GetString()));
#else
			Asura_Assert(0, "GetBlueprintValue as Hash, value was found, but type was unexpected... wrong default value? ");
#endif		
		}
	}
#ifdef ASURA_ASSERT
	else
	{
#ifdef ASURA_TOOLS
		Asura_Assertf( false, ( "GetBlueprintValue as Hash, value with hash ID %08X not found, using default value (%u) (blueprint path: '%s')", uParameterID, uDefaultValue, GetNameIncludingAllParentNames().GetString() ) );
#else
		Asura_Assertf( false, ( "GetBlueprintValue as Hash, value with hash ID %08X not found, using default value (%u)", uParameterID, uDefaultValue ) );
#endif
	}
#endif

	return uValue;
}

#ifdef ASURA_TOOLS
Asura_String Asura_Blueprint::GetNameIncludingAllParentNames() const
{
	const Asura_String szSpecialString(".");

	Asura_String xName = GetName();

	Asura_Blueprint* pxParentBlueprint = GetParent();
	while ( pxParentBlueprint )
	{
		xName.Prepend( szSpecialString.GetString(), szSpecialString.GetLength() );
		xName.Prepend( pxParentBlueprint->GetName().GetString(), pxParentBlueprint->GetName().GetLength() );
		pxParentBlueprint = pxParentBlueprint->GetParent();
	}

	return xName;
}

void Asura_Blueprint::AddParameter( const Asura_String& xName, const Asura_String& xGroupName,
									const Asura_Blueprint_Override_Type eOverrideType, const Asura_Value& xValue,
									const Asura_Hash_ID uTemplate /*= ASURA_HASH_ID_UNSET*/ )
{
    // We have already checked this parameter doesn't exist.
    const Asura_Hash_ID uParameterHashID = Asura_GetHashID( xName.GetString() );

	Asura_Parameter xNewParameter;
	xNewParameter.SetName( xName );
	xNewParameter.SetGroupName( xGroupName );
	xNewParameter.SetValue( xValue );
	xNewParameter.SetHashID( uParameterHashID );
	xNewParameter.SetOverrideType( eOverrideType );
	xNewParameter.SetTemplate( uTemplate );

	m_xParameters.AddEntry( xNewParameter );
}

void Asura_Blueprint::DeleteParameter( const Asura_String& xName )
{
	const Asura_Hash_ID uParameterHashID = Asura_GetHashID( xName.GetString() );
	m_xParameters.RemoveEntry( uParameterHashID );
}

void Asura_Blueprint::RenameParameter( const Asura_String& xOldName, const Asura_String& xNewName )
{
	const Asura_Hash_ID uOldParameterHashID = Asura_GetHashID( xOldName.GetString() );
	const Asura_Hash_ID uNewParameterHashID = Asura_GetHashID( xNewName.GetString() );

	Asura_Parameter* pxParameter = GetParameterByHashID( uOldParameterHashID );
	if ( pxParameter )
	{
		// Must take a local copy, as RemoveEntry will invalidate the old parameter.
		Asura_Parameter xParameter = *pxParameter;
		xParameter.SetName( xNewName );
		xParameter.SetHashID( uNewParameterHashID );

		m_xParameters.RemoveEntry( uOldParameterHashID );
		m_xParameters.AddEntry( xParameter );
	}
}

void Asura_Blueprint::WriteToChunkStream( Asura_Chunk_Stream& xStream )
{
    xStream << CURRENT_CHUNK_VERSION;

    xStream << m_uHashID;
    xStream << m_uParentHashID;
    xStream << m_xName;
	const int iNumParameters = GetNumParameters();
    xStream << iNumParameters;

    for ( int i = 0; i < iNumParameters; ++i )
    {
        Asura_Parameter* pxParameter = GetParameterByIndex( i );
        Asura_Assert( pxParameter, "Asura_Blueprint::WriteToChunkStream - could not get parameter" );
        if ( !pxParameter ) { continue; }

        xStream << pxParameter->GetHashID();
        xStream << pxParameter->GetName();
        xStream << pxParameter->GetGroupName();
        xStream << pxParameter->GetOverrideType();
		xStream << pxParameter->GetTemplate();
		xStream << pxParameter->GetValue().GetValueType();
		switch ( pxParameter->GetValue().GetValueType() )
		{
			default: Asura_Assert( false, "Asura_Blueprint::WriteToChunkStream - bad value type" ); // fallthrough
			case Asura_Value::VT_INT:   xStream << pxParameter->GetValue().GetIntValue();     break;
			case Asura_Value::VT_FLOAT: xStream << pxParameter->GetValue().GetFloatValue();   break;
			case Asura_Value::VT_BOOL:  xStream << pxParameter->GetValue().GetBoolValue();    break;
			case Asura_Value::VT_HASH:
			{
				xStream << pxParameter->GetValue().GetHashIDValue();
                xStream << pxParameter->GetValue().GetHashDisplayTextHash();

				if ( pxParameter->GetValue().GetHashDisplayTextHash() != ASURA_HASH_ID_UNSET )
				{
                    const Asura_Char* pszString = pxParameter->GetValue().GetHashStringValue();
					if ( pszString )
					{
                        Asura_String xString( pszString );
						xStream << xString;
					}
					else
					{
                        // We must write something out here, otherwise we'll load the wrong amount of data in.
						xStream << Asura_String();
					}
				}
                break;
			}
		}
    }
}
#endif // ASURA_TOOLS

void Asura_Blueprint::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
    u_int uVersion;
    xStream >> uVersion;

	if ( uVersion <= 7 )
	{
		xStream >> m_uHashID;
		xStream >> m_uParentHashID;
#ifdef ASURA_TOOLS
		xStream >> m_xName;
#else
		{
			Asura_String xDummyString;
			xStream >> xDummyString;
		}
#endif
		int iNumParameters;
		xStream >> iNumParameters;

		if ( iNumParameters > 0 )
		{
			m_xParameters.Preallocate( iNumParameters );
			for ( int i = 0; i < iNumParameters; ++i )
			{
				Asura_Parameter xParameter;
				ReadParameterFromStream( uVersion, xStream, xParameter );

				m_xParameters.AddEntryWithPreallocation( xParameter );
			}
		}
	}
}

Asura_Blueprint& Asura_Blueprint::operator=( const Asura_Blueprint& xRhs )
{
    if ( this != &xRhs )
    {
        m_uHashID        = xRhs.m_uHashID;
        m_uParentHashID  = xRhs.m_uParentHashID;
        m_pxParent       = xRhs.m_pxParent;

		const int iNumParameters = xRhs.GetNumParameters();
		if ( iNumParameters > 0 )
		{
			m_xParameters.Preallocate( iNumParameters );
			for ( int i = 0; i < iNumParameters; ++i )
			{
				m_xParameters.AddEntryWithPreallocation( *(xRhs.GetParameterByIndex( i )) );
			}
		}
		else
		{
			m_xParameters.Clear();
		}

#ifdef ASURA_TOOLS
		m_bExported					= xRhs.m_bExported;
		m_bChildBlueprintsExported	= xRhs.m_bChildBlueprintsExported;
		m_bToBeDeleted				= xRhs.m_bToBeDeleted;
		m_xName						= xRhs.m_xName;
#endif
    }

    return *this;
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Asura_Blueprint::ReadParameterFromStream( const int iChunkVersion, class Asura_Chunk_Stream& xStream, Asura_Parameter& xParameter )
{
	if ( iChunkVersion == 7 || iChunkVersion == 6 )
	{
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

#ifdef ASURA_TOOLS
        Asura_String xString;
        xStream >> xString;
        xParameter.SetName( xString );
        xStream >> xString;
        xParameter.SetGroupName( xString );
#else
        {
        Asura_String xDummyString;
        xStream >> xDummyString;
        xStream >> xDummyString;
        }
#endif

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

	    Asura_Hash_ID uTemplate;
	    xStream >> uTemplate;
	    xParameter.SetTemplate( uTemplate );

	    Asura_Value xValue;
	    int iValueType;
	    xStream >> iValueType;
	    switch ( iValueType )
	    {
		    default: Asura_Assert( false, "Asura_Blueprint::ReadFromChunkStream - bad value type" ); // fallthrough
		    case Asura_Value::VT_INT:
		    {
			    int iValue;
			    xStream >> iValue;
			    xValue.SetValue( iValue );
		    } break;

		    case Asura_Value::VT_FLOAT:
		    {
			    float fValue;
			    xStream >> fValue;
			    xValue.SetValue( fValue );
		    } break;

		    case Asura_Value::VT_BOOL:
		    {
			    bool bValue;
			    xStream >> bValue;
			    xValue.SetValue( bValue );
		    } break;

		    case Asura_Value::VT_HASH:
		    {
			    Asura_Hash_ID uValue;
			    xStream >> uValue;
                Asura_Hash_ID uHashDisplayTextHash;
                xStream >> uHashDisplayTextHash;

			    if ( uHashDisplayTextHash != ASURA_HASH_ID_UNSET )
			    {
                    Asura_String xString;
				    xStream >> xString;
                    xValue.SetValue( uValue, xString.GetString() );
			    }
                else
                {
                    xValue.SetValue( uValue );
                }
		    } break;
	    }
    	xParameter.SetValue( xValue );

#ifdef ASURA_TOOLS
	    if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
	    {
		    Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
	    }
#endif
	}
    else if ( iChunkVersion == 5 )
    {
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

#ifdef ASURA_TOOLS
        Asura_String xString;
        xStream >> xString;
        xParameter.SetName( xString );
        xStream >> xString;
        xParameter.SetGroupName( xString );
#else
        {
        Asura_String xDummyString;
        xStream >> xDummyString;
        xStream >> xDummyString;
        }
#endif

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

	    Asura_Hash_ID uTemplate;
	    xStream >> uTemplate;
	    xParameter.SetTemplate( uTemplate );

	    u_int uArraySize;
	    xStream >> uArraySize;

	    for ( u_int j = 0; j < uArraySize; ++j )
	    {
		    Asura_Value xValue;
		    int iValueType;
		    xStream >> iValueType;
		    switch ( iValueType )
		    {
			    default: Asura_Assert( false, "Asura_Blueprint::ReadFromChunkStream - bad value type" ); // fallthrough
			    case Asura_Value::VT_INT:
			    {
				    int iValue;
				    xStream >> iValue;
				    xValue.SetValue( iValue );
			    } break;

			    case Asura_Value::VT_FLOAT:
			    {
				    float fValue;
				    xStream >> fValue;
				    xValue.SetValue( fValue );
			    } break;

			    case Asura_Value::VT_BOOL:
			    {
				    bool bValue;
				    xStream >> bValue;
				    xValue.SetValue( bValue );
			    } break;

			    case Asura_Value::VT_HASH:
			    {
				    Asura_Hash_ID uValue;
				    xStream >> uValue;
                    Asura_Hash_ID uHashDisplayTextHash;
                    xStream >> uHashDisplayTextHash;

				    if ( uHashDisplayTextHash != ASURA_HASH_ID_UNSET )
				    {
                        Asura_String xString;
					    xStream >> xString;
                        xValue.SetValue( uValue, xString.GetString() );
				    }
                    else
                    {
                        xValue.SetValue( uValue );
                    }
			    } break;
		    }
		    xParameter.SetValue( xValue );

#ifdef ASURA_TOOLS
		    if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
		    {
			    Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
		    }
#endif
	    }
    }
	// We're only supporting older versions in tools.
#ifdef ASURA_TOOLS
    else if ( iChunkVersion == 4 )
    {
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

#ifdef ASURA_TOOLS
        Asura_String xName;
        xStream >> xName;
        xParameter.SetName( xName );
#else
        {
        Asura_String xDummyString;
        xStream >> xDummyString;
        }
#endif

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

	    Asura_Hash_ID uTemplate;
	    xStream >> uTemplate;
#ifdef ASURA_TOOLS
	    xParameter.SetTemplate( uTemplate );
#endif

	    u_int uArraySize;
	    xStream >> uArraySize;

	    for ( u_int j = 0; j < uArraySize; ++j )
	    {
		    Asura_Value xValue;
		    int iValueType;
		    xStream >> iValueType;
		    switch ( iValueType )
		    {
			    default: Asura_Assert( false, "Asura_Blueprint::ReadFromChunkStream - bad value type" ); // fallthrough
			    case Asura_Value::VT_INT:
			    {
				    int iValue;
				    xStream >> iValue;
				    xValue.SetValue( iValue );
			    } break;

			    case Asura_Value::VT_FLOAT:
			    {
				    float fValue;
				    xStream >> fValue;
				    xValue.SetValue( fValue );
			    } break;

			    case Asura_Value::VT_BOOL:
			    {
				    bool bValue;
				    xStream >> bValue;
				    xValue.SetValue( bValue );
			    } break;

			    case Asura_Value::VT_HASH:
			    {
				    Asura_Hash_ID uValue;
				    xStream >> uValue;
                    Asura_Hash_ID uHashDisplayTextHash;
                    xStream >> uHashDisplayTextHash;

				    if ( uHashDisplayTextHash != ASURA_HASH_ID_UNSET )
				    {
                        Asura_String xString;
					    xStream >> xString;
                        xValue.SetValue( uValue, xString.GetString() );
				    }
                    else
                    {
                        xValue.SetValue( uValue );
                    }
			    } break;
		    }
    	    
		    xParameter.SetValue( xValue );

#ifdef ASURA_TOOLS
		    if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
		    {
			    Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
		    }
#endif
	    }
    }
    else if ( iChunkVersion == 3 )
    {
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

#ifdef ASURA_TOOLS
        Asura_String xName;
        xStream >> xName;
        xParameter.SetName( xName );
#else
        {
        Asura_String xDummyString;
        xStream >> xDummyString;
        }
#endif

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

	    Asura_Hash_ID uTemplate;
	    xStream >> uTemplate;
#ifdef ASURA_TOOLS
	    xParameter.SetTemplate( uTemplate );
#endif

	    u_int uArraySize;
	    xStream >> uArraySize;

	    for ( u_int j = 0; j < uArraySize; ++j )
	    {
		    Asura_Value xValue;
		    int iValueType;
		    xStream >> iValueType;
		    switch ( iValueType )
		    {
			    default: Asura_Assert( false, "Asura_Blueprint::ReadFromChunkStream - bad value type" ); // fallthrough
			    case Asura_Value::VT_INT:
			    {
				    int iValue;
				    xStream >> iValue;
				    xValue.SetValue( iValue );
			    } break;

			    case Asura_Value::VT_FLOAT:
			    {
				    float fValue;
				    xStream >> fValue;
				    xValue.SetValue( fValue );
			    } break;

			    case Asura_Value::VT_BOOL:
			    {
				    bool bValue;
				    xStream >> bValue;
				    xValue.SetValue( bValue );
			    } break;

                case 4: // Asura_Value::VT_STRING:
			    case Asura_Value::VT_HASH:
			    {
				    Asura_Hash_ID uValue;
				    xStream >> uValue;

				    Asura_String xString;
				    if ( uValue != ASURA_HASH_ID_UNSET )
				    {
					    xStream >> xString;
				    }
				    xValue.SetValue( uValue, xString.GetString() );
			    } break;
		    }
    	    
		    xParameter.SetValue( xValue );

#ifdef ASURA_TOOLS
		    if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
		    {
			    Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
		    }
#endif
	    }
    }
    else if ( iChunkVersion == 2 )
    {
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

#ifdef ASURA_TOOLS
        Asura_String xName;
        xStream >> xName;
        xParameter.SetName( xName );
#else
        {
        Asura_String xDummyString;
        xStream >> xDummyString;
        }
#endif

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

	    Asura_Hash_ID uTemplate;
	    xStream >> uTemplate;
#ifdef ASURA_TOOLS
	    xParameter.SetTemplate( uTemplate );
#endif

	    Asura_Value xValue;
	    int iValueType;
	    xStream >> iValueType;
	    switch ( iValueType )
	    {
		    default: Asura_Assert( false, "Asura_Blueprint::ReadFromChunkStream - bad value type" ); // fallthrough
		    case Asura_Value::VT_INT:
		    {
			    int iValue;
			    xStream >> iValue;
			    xValue.SetValue( iValue );
		    } break;

		    case Asura_Value::VT_FLOAT:
		    {
			    float fValue;
			    xStream >> fValue;
			    xValue.SetValue( fValue );
		    } break;

		    case Asura_Value::VT_BOOL:
		    {
			    bool bValue;
			    xStream >> bValue;
			    xValue.SetValue( bValue );
		    } break;

            case 4: // Asura_Value::VT_STRING:
		    case Asura_Value::VT_HASH:
		    {
			    Asura_Hash_ID uValue;
			    xStream >> uValue;

			    Asura_String xString;
			    if ( uValue != ASURA_HASH_ID_UNSET )
			    {
				    xStream >> xString;
			    }
			    xValue.SetValue( uValue, xString.GetString() );
		    } break;
	    }
	    
	    xParameter.SetValue( xValue );

#ifdef ASURA_TOOLS
	    if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
	    {
		    Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
	    }
#endif
    }
    else if ( iChunkVersion == 1 )
    {
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

        Asura_String xName;
        xStream >> xName;
        xParameter.SetName( xName );

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

	    Asura_Value xValue;
	    int iValueType;
	    xStream >> iValueType;
	    switch ( iValueType )
	    {
		    default: Asura_Assert( false, "Asura_Blueprint::ReadFromChunkStream - bad value type" ); // fallthrough
		    case Asura_Value::VT_INT:
		    {
			    int iValue;
			    xStream >> iValue;
			    xValue.SetValue( iValue );
		    } break;

		    case Asura_Value::VT_FLOAT:
		    {
			    float fValue;
			    xStream >> fValue;
			    xValue.SetValue( fValue );
		    } break;

		    case Asura_Value::VT_BOOL:
		    {
			    bool bValue;
			    xStream >> bValue;
			    xValue.SetValue( bValue );
		    } break;

            case 4: // Asura_Value::VT_STRING:
		    case Asura_Value::VT_HASH:
		    {
			    Asura_Hash_ID uValue;
			    xStream >> uValue;

			    Asura_String xString;
			    if ( uValue != ASURA_HASH_ID_UNSET )
			    {
				    xStream >> xString;
			    }
			    xValue.SetValue( uValue, xString.GetString() );
		    } break;
	    }
	    
	    xParameter.SetValue( xValue );

	    if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
	    {
		    Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
	    }
    }
    else if ( iChunkVersion == 0 )
    {
        Asura_Hash_ID uHashID;
        xStream >> uHashID; 
        xParameter.SetHashID( uHashID );

        Asura_String xName;
        xStream >> xName;
        xParameter.SetName( xName );

        int iOverrideType;
        xStream >> iOverrideType;
        xParameter.SetOverrideType( static_cast< Asura_Blueprint_Override_Type >( iOverrideType ) );

        Asura_Value xValue;
        int iValueType;
        xStream >> iValueType;
        switch ( iValueType )
        {
            default: Asura_Assert( false, "Asura_Blueprint::WriteToChunkStream - bad value type" ); // fallthrough
            case Asura_Value::VT_INT:
            {
                int iValue;
                xStream >> iValue;
                xValue.SetValue( iValue );
            } break;

            case Asura_Value::VT_FLOAT:
            {
                float fValue;
                xStream >> fValue;
                xValue.SetValue( fValue );
            } break;

            case Asura_Value::VT_BOOL:
            {
                bool bValue;
                xStream >> bValue;
                xValue.SetValue( bValue );
            } break;

            // For version 0 we convert what were U_INT values into INTs.
            case Asura_Value::VT_HASH:
            {
                u_int uValue;
                xStream >> uValue;
                Asura_Assert( uValue < INT_MAX, "Asura_Blueprint::ReadFromChunkStream - value clamped when converting U_INT to INT" );
                xValue.SetValue( static_cast< int >( uValue ) );
            } break;
        }
        xParameter.SetValue( xValue );

		if ( xParameter.GetTemplate() != ASURA_HASH_ID_UNSET )
		{
			Tools_Blueprint_Edit_Object::HandleTemplatedTypeLoad( &xParameter );
		}
    }
#endif // ASURA_TOOLS - only supporting older versions in tools
    else
    {
        Asura_Assertf( false, ( "Asura_Blueprint::ReadParameterFromStream - unhandled chunk version (%d)", iChunkVersion ) );
    }
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Blueprint::EnumerateParameters
// Author       : Andy Tate
// Date         : 22-Jan-2008
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Blueprint::EnumerateParameters(const char* szBaseText,
										   const u_int uNumDigits,
										   Asura_Blueprint_ParameterEnumerator& xEnumerator) const
{
	char	szParameterName[256];

	u_int	uCount	= 0;

	// Begin enumeration...
	xEnumerator.BeginEnumeration();

	// Loop until we cannot find the number parameter...
	for(;;)
	{
		Asura_Hash_ID	uParameterHash;
		Asura_Value		xValue;

		// Create a new parameter-name...
		Asura_PrintfToBfr(szParameterName, xEnumerator.GetParameterFormat(), 256, true, szBaseText, uNumDigits, (uCount + 1));
		uParameterHash	= Asura_GetHashID(szParameterName);

		// Does this parameter-hash exist?
		if(GetParameterValueByHashID(uParameterHash, xValue))
		{
			// Yes - time to enumerate this data...
			if(xEnumerator.Enumerate(uCount, szParameterName, xValue))
			{
				// Enumeration successful - try the next entry...
				uCount++;
			}
			else
			{
				// Enumeration failed - abort...
				Asura_Assert(false, "Failed to enumerate parameter.");
				xEnumerator.EndEnumeration(false);
				return(uCount);
			}
		}
		else
		{
			// No parameter - that's the end of our enumeration...
			xEnumerator.EndEnumeration(true);
			return(uCount);
		}
	}
}

/// End of function 'Asura_Blueprint::EnumerateParameters'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Blueprint::CountEnumeratedParameters
// Author       : Anthony Gregson
// Date         : 11/03/08
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Blueprint::CountEnumeratedParameters(const char* szBaseText, const u_int uNumDigits) const
{
	Asura_Blueprint_DummyParameterEnumerator xDummyEnumerator;

	return EnumerateParameters(szBaseText, uNumDigits, xDummyEnumerator);
}

/// End of function 'Asura_Blueprint::CountEnumeratedParameters'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Blueprint::EnumerateParameters
// Author       : Anthony Gregson
// Date         : 21/04/09
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Blueprint::EnumerateParameters(const Asura_Hash_ID uBaseHash, const u_int uNumDigits, Asura_Blueprint_ParameterEnumerator_ByHash& xEnumerator) const
{
	Asura_Assert(uNumDigits < 8, "How many parameters are you expecting to enumerate?");
	char	szExtension[8];

	u_int	uCount	= 0;

	// Begin enumeration...
	xEnumerator.BeginEnumeration();

	// Loop until we cannot find the number parameter...
	for(;;)
	{
		Asura_Hash_ID	uParameterHash;
		Asura_Value		xValue;

		// Create a new parameter-name...
		Asura_PrintfToBfr(szExtension, "%0*d", 8, true, uNumDigits, (uCount + 1));
		uParameterHash	= Asura_ConcatenateHashID(uBaseHash, szExtension);

		// Does this parameter-hash exist?
		if(GetParameterValueByHashID(uParameterHash, xValue))
		{
			// Yes - time to enumerate this data...
			if(xEnumerator.Enumerate(uCount, xValue))
			{
				// Enumeration successful - try the next entry...
				uCount++;
			}
			else
			{
				// Enumeration failed - abort...
				Asura_Assert(false, "Failed to enumerate parameter.");
				xEnumerator.EndEnumeration(false);
				return(uCount);
			}
		}
		else
		{
			// No parameter - that's the end of our enumeration...
			xEnumerator.EndEnumeration(true);
			return(uCount);
		}
	}
}

/// End of function 'Asura_Blueprint::EnumerateParameters'. ///


//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
