
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_BackgroundChunkProcessing_System.cpp
// Author : Andy Tate
// Date   : 25-Jul-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "ChunkSystem/Asura_BackgroundChunkProcessing_System.h"

// Asura independent header-files...
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_ProcessChunkTimingList.h"
#include "Core/Asura_Maths.h"

#if defined(ASURA_USE_BACKGROUND_CHUNK_PROCESSING)

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// Static variables...
bool						Asura_BackgroundChunkProcessing_System::s_bInitialised	= false;

Asura_SyncEvent				Asura_BackgroundChunkProcessing_System::s_xReadyEvent ("BCP Ready", false);
Asura_SyncEvent				Asura_BackgroundChunkProcessing_System::s_xCompleteEvent ("BCP Complete", false);

Asura_TickCount				Asura_BackgroundChunkProcessing_System::s_xStartTicks	= 0;
Asura_TickCount				Asura_BackgroundChunkProcessing_System::s_xTicksAllowed	= 0;

Asura_Chunk_File_Stream*	Asura_BackgroundChunkProcessing_System::s_pxFileStream	= NULL;


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BackgroundChunkProcessing_System::UpdateThread
// Author       : Andy Tate
// Date         : 25-Jul-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_BackgroundChunkProcessing_System::UpdateThread (const void* pUserData)
{
	// Wait forever...
	for (;;)
	{
		// Do we have data to process?
		s_xReadyEvent.WaitForSignal();
		s_xReadyEvent.ClearSignal();

		// Process the chunk...
		if(!Asura_ChunkLoading_System::ProcessChunk(*s_pxFileStream))
		{
			// All done...
			s_pxFileStream	= NULL;
		}

		// We're done with this chunk...
		if (!s_xCompleteEvent.IsSignalled())
		{
			s_xCompleteEvent.SetSignal();
		}
	}
}

/// End of function 'Asura_BackgroundChunkProcessing_System::UpdateThread'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BackgroundChunkProcessing_System::Initialise
// Author       : Andy Tate
// Date         : 25-Jul-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_BackgroundChunkProcessing_System::Initialise (void)
{
	// Create the thread to use...
	if(Asura_Thread_System::CreateThread("Chunk Processing", UpdateThread, NULL))
	{
		s_bInitialised	= true;
	}
	else
	{
		Asura_Assert(false, "Failed to create background chunk-processing thread");
	}
}

/// End of function 'Asura_BackgroundChunkProcessing_System::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BackgroundChunkProcessing_System::Shutdown
// Author       : Andy Tate
// Date         : 25-Jul-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_BackgroundChunkProcessing_System::Shutdown (void)
{
	// Should destroy the thread here - but we don't have that functionality...
	s_bInitialised	= false;

	// Remove the signals...
	if(s_xReadyEvent.IsSignalled())
	{
		s_xReadyEvent.ClearSignal();
	}
	if(s_xCompleteEvent.IsSignalled())
	{
		s_xCompleteEvent.ClearSignal();
	}
}

/// End of function 'Asura_BackgroundChunkProcessing_System::Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BackgroundChunkProcessing_System::BeginProcessing
// Author       : Andy Tate
// Date         : 25-Jul-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_BackgroundChunkProcessing_System::BeginProcessing (Asura_Chunk_File_Stream& xStream)
{
	// Yes - are we already processing a chunk?
	if(s_pxFileStream == NULL)
	{
		// No - time to begin...
		s_pxFileStream	= &xStream;
		return(true);
	}
	else
	{
		Asura_Assert(false, "Already processing a chunk in the background - cannot process a second one.");
		return(false);
	}
}

/// End of function 'Asura_BackgroundChunkProcessing_System::BeginProcessing'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BackgroundChunkProcessing_System::WaitForProcessing
// Author       : Andy Tate
// Date         : 25-Jul-2006
////////////////////////////////////////////////////////////////////////////////
DEBUGVAR float dbg_fUnlimitedProcessingTimeMin = Asura_Maths::LargeFloat * 0.5f;

bool Asura_BackgroundChunkProcessing_System::WaitForProcessing (const float fTimeAllowed)
{
	// Do we have a chunk to process?
	if(s_pxFileStream == NULL)
	{
		Asura_Assert(false, "No chunk to be processed in the background.");
		return(true);
	}
	else
	{
		// Are we using the background chunk-processing?
		if(s_bInitialised)
		{
			// Yes - set the amount of time we have...
			s_xStartTicks	= Asura_Timers::GetCurTicks();
			s_xTicksAllowed	= (fTimeAllowed < dbg_fUnlimitedProcessingTimeMin) ? Asura_Timers::SecondsToTicks(fTimeAllowed) : 0; // Use 0 to signify unlimited time.

#if defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)
			// Begin time how long the process is going to take...
			Asura_ProcessChunkTimingList::BeginTiming();
#endif	// defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)

			// Inform the thread that we've got some data...
			if (!s_xReadyEvent.IsSignalled())
			{
				s_xReadyEvent.SetSignal();
			}

			s_xCompleteEvent.WaitForSignal();
			s_xCompleteEvent.ClearSignal();

#if defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)
			// Stop timing how long processing this chunk took...
			Asura_ProcessChunkTimingList::EndTiming(Asura_ChunkLoading_System::GetCurrentChunkID());
#endif	// defined(ASURA_CHUNK_LOADERS_ALLOW_DEBUG_TIMING)

			// Return whether we've finished processing...
			return(s_pxFileStream != NULL);
		}
		else
		{
			// Process as normal...
			return(Asura_ChunkLoading_System::ProcessChunk(*s_pxFileStream));
		}
	}
}

/// End of function 'Asura_BackgroundChunkProcessing_System::WaitForProcessing'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BackgroundChunkProcessing_System::YieldProcessing
// Author       : Andy Tate
// Date         : 25-Jul-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_BackgroundChunkProcessing_System::YieldProcessing (void)
{
	// Do we have a chunk to process?
	if(s_bInitialised && (s_pxFileStream != NULL))
	{
		const Asura_TickCount	xTicksElapsed	= Asura_Timers::GetElapsedTicks(s_xStartTicks);

		// Have we taken long enough?
		// 0 ticks allowed signifies unlimited time to process.
		if((s_xTicksAllowed > 0) && (xTicksElapsed >= s_xTicksAllowed))
		{
			// Yes - suspend what we're doing until we're told to resume...
			if (!s_xCompleteEvent.IsSignalled())
			{
				s_xCompleteEvent.SetSignal();
			}

			s_xReadyEvent.WaitForSignal();
			s_xReadyEvent.ClearSignal();
		}
	}
}

/// End of function 'Asura_BackgroundChunkProcessing_System::YieldProcessing'. ///

#endif	// defined(ASURA_USE_BACKGROUND_CHUNK_PROCESSING)


/// End of file. ///
