/**
*
* \file Asura_Chunk_AnimationTable.cpp
*
* Created by Richard May - 18/06/2002 14:27:18
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/



/************/
/* Includes.*/
/************/

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_Chunk_AnimationTable.h"

#ifdef ASURA_DISABLE_ANIM_TABLE_CODE

#include "Debugging/Asura_Assert.h"
bool Asura_Chunk_AnimationTable::Process ( Asura_Handle fileHandle )
{
    Asura_Assert(false, "Anim table not supported");
    return false;
}

bool Asura_Chunk_ExportedAnimationTable::Process(Asura_Handle iFileHandle) const
{
    Asura_Assert(false, "Exported anim table not supported");
    return false;
}

#else // not def ASURA_DISABLE_ANIM_TABLE_CODE

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_String.h"
#include "Core/Memory/Asura_Mem.h"

#ifdef ASURA_MAIN
#include "Animation/Asura_AnimTable.h"
#include "Animation/Asura_AnimTable_System.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#endif

#if defined(ASURA_MAIN) && defined(ASURA_TOOLS)
#include "Hierarchy/Tools_Animation_Files.h"
#endif

/**************/
/* Functions. */
/**************/

#ifdef ASURA_MAIN

static inline bool ReadHashIDForThisVersion( int iVersion, Asura_Handle iFileHandle, Asura_Hash_ID& uIDOut )
{
#ifdef ASURA_TOOLS
	if( iVersion >= 4 )
	{
		return Asura_FileAccess::ReadData(iFileHandle, &uIDOut, 4);
	}
	else
	{
		const char* szString = Asura_FileAccess::ReadString(iFileHandle);
		if (szString)
		{
			uIDOut = Asura_GetHashID(szString);
			return true;
		}
		return false;
	}
#else
    // (MP - 25.01.2005) Don't bother supporting ancient chunk versions
    // outside of tools.
	return Asura_FileAccess::ReadData( iFileHandle, &uIDOut, 4 );
#endif
}

/*************************************************************
 *
 * Name          : Asura_Chunk_AnimationTable::Process
 *
 * Author        : Richard May
 *
 * Description   : 
 *
 * Parameters    : void
 *
 * Return value  : void
 *
 *************************************************************/
bool Asura_Chunk_AnimationTable::Process ( Asura_Handle fileHandle )
{
#ifdef ASURA_TOOLS
	Asura_Assert( 1 <= Version && Version <= CURRENT_VERSION, "Unsupported Chunk version!" );
#else
    // (MP - 25.01.2005) Don't bother supporting ancient chunk versions
    // outside of tools.
	Asura_Assert( Version >= 4, "Unsupported Chunk version!" );
#endif

    // (MP - 22.10.2002) Some anim table chunks are being exported without
    // any data. Deal with that case.
    if( sizeof(Asura_Chunk_AnimationTable) == this->Size )
    {
#if _DEBUG
        Asura_OutputToDebugger(( "Skipping empty anim table chunk in file \"%s\".\n", Asura_FileAccess::GetFilename(fileHandle) ));
#endif
        return true;
    }
	
	Asura_Hash_ID uID;
	Asura_Hash_ID uMasterID = ASURA_HASH_ID_UNSET;

	// read all the header information first
	bool bSuccess = ReadHashIDForThisVersion(Version, fileHandle, uID);
	if (!bSuccess)
	{
		Asura_Assert(false, "Error reading data");
		return false;
	}
	Asura_AnimTable* pTable = Asura_AnimTable_System::GetAnimTable(uID);
	if(!pTable)
	{
		// attempt to create a new table
		pTable = Asura_AnimTable_System::AddNewTable(uID,NumberOfPrimaryStates,NumberOfSecondaryStates);
		Asura_Assert( pTable, "Failed To Create AnimTable!" );
#ifdef ASURA_TOOLS
		pTable->SetName(Asura_FileAccess::GetFilename(fileHandle));
#endif
		// skip the rest of this chunk
		if( !pTable )
		{
			// RAR 26/02/03 - The amount being skipped on some of these failure cases is wrong.
			// just return false and let the loaders sort out the position
			
			//int size = Size - (sizeof(*this) + name.GetSizeInInts()*4 + tableType.GetSizeInInts()*4);
			//if( tableType=="Slave" ) size -= (masterTable.GetSizeInInts()*4);
			//Asura_FileAccess::SkipData(fileHandle,size);
			
			return false;
		}	
	}
	
	bool bSlave = false;
	bSuccess = ReadHashIDForThisVersion(Version, fileHandle, uID);
	if (!bSuccess)
	{
		Asura_Assert(false, "Error reading data");
		return false;
	}
	if( uID == uASURA_ANIMTABLE_HASHID_SLAVE )
	{
		bSlave = true;
	}
	
	Asura_AnimTable* pMasterTable = NULL;
	if(bSlave)
	{
		// if it's a slave table, the next entry is the name of its master table		
		bSuccess = ReadHashIDForThisVersion(Version, fileHandle, uMasterID);
		if (!bSuccess)
		{
			Asura_Assert(false, "Error reading data");
			return false;
		}
		#ifdef ASURA_TOOLS
		//Use a seperate function in tools which has the possibility of loading the master table if it is missing. RAR 02/05/2007
		pMasterTable = Tools_Animation_Files::GetAnimTable( uMasterID );
		#else
		pMasterTable = Asura_AnimTable_System::GetAnimTable( uMasterID );
		#endif
	}

	int state;
	bSuccess = ReadHashIDForThisVersion(Version, fileHandle, uID);
	if (!bSuccess)
	{
		Asura_Assert(false, "Error reading data");
		return false;
	}
	state = Asura_AnimTable_System::ConvertHashToAnimState(uID);

	// check to see whether this state already has data in it's table
	if( ( state < static_cast<int>(pTable->GetNoOfWeaponStates()) ) &&
		( pTable->GetWeaponStateExists(state) ) &&
		( pTable->GetAnimID(state,0,0) != ASURA_HASH_ID_UNSET ) )
	{
		// if it has, the table has already been loaded

#if defined(ASURA_TOOLS) && defined(ASURA_ENABLE_HASH_TO_FILENAME_LOOKUPS)
        const char* szFilename = Asura_HashedFilename_System::GetFilename( pTable->m_uNameHash, false );
        if( szFilename )
        {
			Asura_Assertf( false, ("ERROR: Anim table '%s' (hash: 0x%08X) already contains data for this weapon state.", szFilename, pTable->m_uNameHash) );
        }
        else
#endif
        {
		    Asura_Assertf( false, ("ERROR: Anim table (name hash: 0x%08X) already contains data for this weapon state.", pTable->m_uNameHash) );
        }

#if  !ASURA_ANIMTABLE_ALLOW_OVERWRITE
		// skip the rest of the file
		// RAR 26/02/03 - The amount being skipped on some of these failure cases is wrong.
		// just return false and let the loaders sort out the position
		/*
		int size = Size - (sizeof(*this) + name.GetSizeInInts()*4 + tableType.GetSizeInInts()*4);
		if( tableType=="Slave" ) size -= (masterTable.GetSizeInInts()*4);
		Asura_FileAccess::SkipData(fileHandle,size);
		*/

		return false;	
#endif
	}	

#ifdef ASURA_TOOLS
	pTable->SetParentTableHashID( uMasterID );
#endif

	const Asura_Hash_ID uZeroHash = 0x00000030; // Asura_GetHashID("0")
	// over-write the elements of the table with the slave specific ones
	for( unsigned int primary=0; primary<pTable->GetNoOfPrimaryStates(); ++primary )
	{	
		for( unsigned int secondary=0; secondary<pTable->GetNoOfSecondaryStates(); ++secondary )
        {
#ifdef ASURA_TOOLS
			if( (Version == 1) && (secondary == 1) )
			{
				/// Richard May - 04/07/2002 16:08:59 - version 1 files have one less column 
				/// (one less secondary state) so this compensates for that, ensuring that 
				/// the rest of the table is loaded as correctly as possible
				++secondary;
			}
#endif

			bSuccess = ReadHashIDForThisVersion(Version, fileHandle, uID);
			if( bSuccess )
			{
				pTable->SetAnimID( uID, state, primary, secondary );
			} // if tableEntry
		} // for secondary
	} // for primary

    // Process flag data, if present.
	Asura_AnimTable_AnimFlags* pxFlagsOut = NULL;
    if( Version >= 5 )
    {
        if( Flags & uASURA_CHUNK_ANIMTABLE_HAS_ENTRY_FLAGS )
        {
            pxFlagsOut = pTable->AllocateAnimFlagsForState( state );

            int iDataSize = NumberOfPrimaryStates * NumberOfSecondaryStates * sizeof(Asura_AnimTable_AnimFlags);
            Asura_FileAccess::ReadData( fileHandle, pxFlagsOut, iDataSize );

            // Deal with padding.
            int iOffset = iDataSize & 3;
            if( iOffset )
            {
                Asura_FileAccess::SkipData( fileHandle, 4 - iOffset );
            }
		}
    }
    
	// copy in the data from the master table if applicable
	if(bSlave)
	{
		if( pMasterTable )
		{
			if(	(pTable->GetNoOfPrimaryStates() < pMasterTable->GetNoOfPrimaryStates()) ||
				(pTable->GetNoOfSecondaryStates() < pMasterTable->GetNoOfSecondaryStates())  )
			{
#if defined(ASURA_TOOLS) && defined(ASURA_ENABLE_HASH_TO_FILENAME_LOOKUPS)
                const char* szFilename = Asura_HashedFilename_System::GetFilename( pTable->m_uNameHash, false );
                if( szFilename )
                {
			        Asura_Assertf( false, ("ERROR: Anim table '%s' (hash: 0x%08X) is smaller than master table.", szFilename, pTable->m_uNameHash) );
                }
                else
#endif
                {
			        Asura_Assertf( false, ("ERROR: Anim table (name hash: 0x%08X) is smaller than master table.", pTable->m_uNameHash) );
                }
				// RAR 26/02/03 - The amount being skipped on some of these failure cases is wrong.
				// just return false and let the loaders sort out the position
				
				//int size = Size - (sizeof(*this) + name.GetSizeInInts()*4 + tableType.GetSizeInInts()*4);
				//if( tableType=="Slave" ) size -= (masterTable.GetSizeInInts()*4);
				//Asura_FileAccess::SkipData(fileHandle,size);

				return false;
			}
			
			/// Richard May - 20/06/2002 10:17:49 - 
			/// the entirity of the master table is copied into the slave table initially
			/// these entries will then be over-written by the slave table's own specific
			/// animations where applicable

			/// Richard May - 27/09/2004 11:03:15 - Now checks whether the master table contains
			/// data for this state 
			if( (u_int)state < pMasterTable->GetNoOfWeaponStates() )
			{
				// allocate flags if master has them but slave does not PDC 3/8/2007
				if (pxFlagsOut == NULL && pMasterTable->HasAnyAnimFlagData())
				{
		            pxFlagsOut = pTable->AllocateAnimFlagsForState( state );
		            int iDataSize = NumberOfPrimaryStates * NumberOfSecondaryStates * sizeof(Asura_AnimTable_AnimFlags);
					Asura_Mem::Set(pxFlagsOut, 0, iDataSize);
				}

				for( unsigned int primary=0; primary<pMasterTable->GetNoOfPrimaryStates(); ++primary )
				{
					for( unsigned int secondary=0; secondary<pMasterTable->GetNoOfSecondaryStates(); ++secondary )
					{
						// copy anim data from master if not in slave PDC 3/8/2007
						if (pTable->GetAnimID(state,primary,secondary) == uZeroHash)
						{
							pTable->SetAnimID( pMasterTable->GetAnimID(state,primary,secondary), state, primary, secondary );
							if (pxFlagsOut)
							{
								pxFlagsOut[primary * NumberOfSecondaryStates + secondary] = pMasterTable->GetAnimFlags(state, primary, secondary);
							}
						}
					}
				}
			}
		}
		else
		{
			// need to load the master table here perhaps?
#if defined(ASURA_TOOLS) && defined(ASURA_ENABLE_HASH_TO_FILENAME_LOOKUPS)
            const char* szFilename = Asura_HashedFilename_System::GetFilename( pTable->m_uNameHash, false );
            if( szFilename )
            {
			    Asura_Assertf( false, ("ERROR: Anim table '%s' (hash: 0x%08X) has missing master table. They must be loaded before slaves.", szFilename, pTable->m_uNameHash) );
            }
            else
#endif
            {
		        Asura_Assertf( false, ("ERROR: Anim table (name hash: 0x%08X) has missing master table. They must be loaded before slaves.", pTable->m_uNameHash) );
            }
		}
	}

    return true;
}
/* end of function Process */

#endif // #ifdef ASURA_MAIN


//////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------------------**
** Asura_Chunk_ExportedAnimationTable **
**------------------------------------*/

#define ASURA_CHUNK_EXPORTEDANIMATIONTABLE_VERSION 1

#ifdef ASURA_MAIN

bool Asura_Chunk_ExportedAnimationTable::Process(Asura_Chunk_Stream& xStream) const
{
	Asura_Assertf(Version<=ASURA_CHUNK_EXPORTEDANIMATIONTABLE_VERSION,("Unknown chunk version %d for exported animation table",Version));

	Asura_Hash_ID uNameHash;
	u_int uNumWeapons;		
	u_int uNumPrimary;		
	u_int uNumSecondary;

	xStream >> uNameHash;
	xStream >> uNumWeapons;
	xStream >> uNumPrimary;
	xStream >> uNumSecondary;

	//get/create the anim table
	Asura_AnimTable* pxTable = Asura_AnimTable_System::GetAnimTable(uNameHash);
	if(pxTable)
	{
		//make sure the existing table is the same size
		if( (pxTable->GetNoOfPrimaryStates() != uNumPrimary) || (pxTable->GetNoOfSecondaryStates() != uNumSecondary) )
		{
			Asura_Assertf(false,("Animation table 0x%x already exists , but with a different number of primary/secondary states.",uNameHash));
			return false;
		}
		
		//Unsure on best behaviour here... for now ,will assert and skip the chunk.
		//RAR 19/01/2006
		Asura_Assertf(false,("Animation table 0x%x already exists",uNameHash));
		return false;
	}
	else
	{
		pxTable = Asura_AnimTable_System::AddNewTable(uNameHash,uNumPrimary,uNumSecondary);
		Asura_Assert(pxTable,"Error creating anim table");
	}

	if(!uNumWeapons || !uNumPrimary || !uNumSecondary)
	{
		Asura_Assertf(false,("Animation table 0x%x has no entries",uNameHash));
		return false;
	}

	//prod the last element now to avoid having to resize the table as we go
	pxTable->SetAnimID(0,uNumWeapons-1,uNumPrimary-1,uNumSecondary-1);

	//fill in the rest
	for(u_int i=0;i<uNumWeapons;i++)
	{
		for(u_int j=0;j<uNumPrimary;j++)
		{
			for(u_int k=0;k<uNumSecondary;k++)
			{
				Asura_Hash_ID uAnimHash;
				xStream >> uAnimHash;
				pxTable->SetAnimID(uAnimHash,i,j,k);	
			}
		}
	}

    // Process flag data, if present.
    if( Version >= 1 && (Flags & uASURA_CHUNK_ANIMTABLE_HAS_ENTRY_FLAGS) && uNumWeapons )
    {
        pxTable->PreallocateAnimFlags();
        for( u_int uWeapon = 0; uWeapon < uNumWeapons; ++uWeapon )
        {
            Asura_AnimTable_AnimFlags* pxFlagsOut = pxTable->AllocateAnimFlagsForState( uWeapon );
            xStream.Read( uNumPrimary * uNumSecondary * sizeof(Asura_AnimTable_AnimFlags), pxFlagsOut );
        }
        xStream.Skip4BytePadding();
    }

	return true;

}

bool Asura_Chunk_ExportedAnimationTable::Process(Asura_Handle iFileHandle) const
{
	Asura_Chunk_File_Stream	xStream (iFileHandle);

	// Call the other process function...

	return	(Process(xStream));
}

#endif // #ifdef ASURA_MAIN

#ifdef ASURA_TOOLS

void Asura_Chunk_ExportedAnimationTable::Create( const Asura_AnimTable* pxTable, Asura_Chunk_Stream& xStream )
{
    bool bAnimFlagData = pxTable->HasAnyAnimFlagData();

    int iChunkFlags = 0;
    if( bAnimFlagData )
    {
        iChunkFlags |= uASURA_CHUNK_ANIMTABLE_HAS_ENTRY_FLAGS;
    }

	xStream.WriteChunk_Start( ASURA_CHUNK_EXPORTEDANIMATIONTABLE, ASURA_CHUNK_EXPORTEDANIMATIONTABLE_VERSION, iChunkFlags );

	u_int uNumWeapons = pxTable->GetNoOfWeaponStates();
	u_int uNumPrimary = pxTable->GetNoOfPrimaryStates();
	u_int uNumSecondary = pxTable->GetNoOfSecondaryStates();

	xStream << pxTable->m_uNameHash;
	xStream << uNumWeapons;
	xStream << uNumPrimary;
	xStream << uNumSecondary;

	for(u_int i=0;i<uNumWeapons;i++)
	{
		for(u_int j=0;j<uNumPrimary;j++)
		{
			for(u_int k=0;k<uNumSecondary;k++)
			{
				xStream << pxTable->GetAnimID(i,j,k);	
			}
		}
	}

    if( bAnimFlagData )
    {
        for( u_int uWeapon = 0; uWeapon < uNumWeapons; ++uWeapon )
        {
            const Asura_AnimTable_AnimFlags* pxFlags = pxTable->GetAllAnimFlagsForState( uWeapon );

            for( u_int uPrimary = 0; uPrimary < uNumPrimary; ++uPrimary )
            {
                for( u_int uSecondary = 0; uSecondary < uNumSecondary; ++uSecondary )
                {
                    const Asura_AnimTable_AnimFlags& xFlags = ( pxFlags ? pxFlags[uPrimary * uNumSecondary + uSecondary] : Asura_AnimTable::s_xDefaultAnimFlags );

                    Asura_WriteToStream_Bitfield_Wrapper<u_char> xBitfieldWrapper( xFlags.m_ucAllFlags );
                    xBitfieldWrapper.Add( xFlags.m_bMirrored, 1 );
                    xBitfieldWrapper.Add( xFlags.m_bReversed, 1 );

                    xStream << xBitfieldWrapper;
                }
            }
        }

        xStream.Write4BytePadding();
    }

	xStream.WriteChunk_End();
}


void Asura_Chunk_ExportedAnimationTable::Create(const class Asura_AnimTable* pxTable,class Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream	xStream (pxSaveBlock);


	// Call the other version...

	Create	(pxTable, xStream);
}

#endif // ASURA_TOOLS

#endif // else not def ASURA_DISABLE_ANIM_TABLE_CODE
