
/**
*
* \file Asura_Chunk_Carpet_Styles.cpp
*
* Created by Dave Reed - 24/11/2005
*
* Copyright (C) 2002 Rebellion (Derby), All Rights Reserved.
*
**/

#include "Core/Asura.h"

#ifdef ASURA_USE_CARPET

#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"

#include "ChunkSystem/Asura_Chunk_Carpet_Styles.h"
#include "Carpet/Asura_Carpet_StyleManager.h"


#ifdef ASURA_TOOLS


const uint32 uASURA_CARPET_STYLES_CHUNK_VERSION = 2;

bool Asura_Chunk_Carpet_Styles::Create(char* szFileName)
{
	int iNumberOfStyles = Asura_Carpet_StyleManager::GetNumStyles();

	//Calculate size of chunk: Header + data
	int iChunkSize = sizeof(Asura_Chunk_Carpet_Styles);
	// +iNumberOfStyles
	iChunkSize += sizeof(int);
	// +iNextUID
	iChunkSize += sizeof(int);

	for(int i=0;i<iNumberOfStyles;i++)
	{
		Asura_Carpet_Style* pxStyle=Asura_Carpet_StyleManager::GetStyle(i);

		// Can't save null strings, but can save empty ones
		if (pxStyle->xName.IsNull()) pxStyle->xName="";
		if (pxStyle->xTextureName.IsNull()) pxStyle->xTextureName="";		

		iChunkSize+=(pxStyle->xName.GetSizeInInts()*4);
		iChunkSize+=(pxStyle->xTextureName.GetSizeInInts()*4);
		
		iChunkSize+=(6*4);
		iChunkSize+=(ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH*4);
		iChunkSize+=(ASURA_CARPET_BILLBOARDS_PER_STYLE*7*4);		
	}

	//Create saveblock and chunk
	Asura_SaveBlock* pxSaveBlock = new Asura_SaveBlock();
	Asura_Assert(pxSaveBlock, "Unable to create saveblock");

	if(pxSaveBlock)
	{
		Asura_Chunk_Carpet_Styles* pxChunk = (Asura_Chunk_Carpet_Styles*)pxSaveBlock->GetAccessPointer(iChunkSize);

		// Fill chunk header
		pxChunk->ID		 = static_cast<ASURA_CHUNKID>(ASURA_CHUNK_CARPET_STYLES);
		pxChunk->Size    = iChunkSize;
		pxChunk->Version = uASURA_CARPET_STYLES_CHUNK_VERSION;
		pxChunk->Flags   = 0;
		char *pcChunkBuffer = (char*)(pxChunk+1);

		// Store style count
		ASURA_CHUNK_FILL(pcChunkBuffer, iNumberOfStyles, int);
		int iNextUID=Asura_Carpet_StyleManager::GetNextUID();
		ASURA_CHUNK_FILL(pcChunkBuffer, iNextUID, int);

		for(int i=0;i<iNumberOfStyles;i++)
		{
			Asura_Carpet_Style* pxStyle=Asura_Carpet_StyleManager::GetStyle(i);			

			ASURA_CHUNK_FILL_STRING(pcChunkBuffer, pxStyle->xName);
			ASURA_CHUNK_FILL_STRING(pcChunkBuffer, pxStyle->xTextureName);

			ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->iTextureID,   int);
			ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->iUID,		   int);
			ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->fDensity,     float);
			ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->fLodFadeNear, float);
			ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->fLodFadeFar,  float);

			ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->fMovementScale,  float);

			for(int j=0;j<ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH;j++)
			{
				ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->aiSequence[j], int);
			}

			for(int j=0;j<ASURA_CARPET_BILLBOARDS_PER_STYLE;j++)
			{
				for(int k=0;k<2;k++)
				{				
					ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->axBillboards[j].fU[k]  , float);
					ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->axBillboards[j].fV[k]  , float);
				}

				ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->axBillboards[j].fWidth		, float);
				ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->axBillboards[j].fHeight	, float);
				ASURA_CHUNK_FILL(pcChunkBuffer, pxStyle->axBillboards[j].fMinScale	, float);
			}
		}

		//Terminate and save chunk
		pxSaveBlock->AddAsuraTerminator();
		bool bSuccess=pxSaveBlock->SaveFile( szFileName, false );
		delete pxSaveBlock;

		return bSuccess;		
	}

	return false;
}



bool Asura_Chunk_Carpet_Styles::Process(Asura_Handle iFileHandle) const
{
	if (Version<1)
	{
		return false;
	}

	Asura_Carpet_StyleManager::Reset();
	int iUID;
	int iNumberOfStyles=0;		

	ASURA_CHUNK_FILEEXTRACT(iFileHandle, iNumberOfStyles, int);	
	ASURA_CHUNK_FILEEXTRACT(iFileHandle, iUID, int);
	Asura_Carpet_StyleManager::SetNextUID(iUID);


	// Add the styles
	for (int i=0;i<iNumberOfStyles;i++)
	{
		// Would be better to preallocate this buffer, as CreateStyle will keep realloc-ing
		const int iStyle=Asura_Carpet_StyleManager::CreateStyle();
		Asura_Carpet_Style* const pxStyle=Asura_Carpet_StyleManager::GetStyle(iStyle);

		ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle, pxStyle->xName);
		ASURA_CHUNK_FILEEXTRACT_STRING(iFileHandle, pxStyle->xTextureName);

		ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->iTextureID,   int);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->iUID,		   int);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->fDensity,     float);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->fLodFadeNear, float);
		ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->fLodFadeFar,  float);

		if (Version>1)
		{
			ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->fMovementScale,  float);
		}
		else
		{
			pxStyle->fMovementScale = 1.0f;
		}

		for(int j=0;j<ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH;j++)
		{
			ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->aiSequence[j], int);
		}

		for(int j=0;j<ASURA_CARPET_BILLBOARDS_PER_STYLE;j++)
		{
			for(int k=0;k<2;k++)
			{				
				ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->axBillboards[j].fU[k]  , float);
				ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->axBillboards[j].fV[k]  , float);
			}

			ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->axBillboards[j].fWidth		, float);
			ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->axBillboards[j].fHeight	, float);
			ASURA_CHUNK_FILEEXTRACT(iFileHandle, pxStyle->axBillboards[j].fMinScale	, float);
		}
		// Need to get the correct texture ID
		bool bWasNew = false;
		int iTextureIndex = Asura_Texture_Management::AddTexture(pxStyle->xTextureName, &bWasNew);
		pxStyle->iTextureID = iTextureIndex;
		if (Asura_Material_System::GetMaterialIDForTexture(iTextureIndex, false) == ASURA_MATERIAL_NONE)
		{
			// Mark it as transparent too.
			int iMaterialID = Asura_Material_System::CreateMaterial();
			if(iMaterialID != ASURA_MATERIAL_NONE)
			{
				Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(iMaterialID);
				Asura_Assert(pxMaterial,"");
				pxMaterial->m_iTextureID[0] = iTextureIndex;
				pxMaterial->m_uMaterialFlags = AsuraMaterialFlag_Transparent;
			}
		}
	}	

	return true;
}
#endif	// ASURA_TOOLS

#endif // ASURA_USE_CARPET
