//////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_Fog.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_Chunk_Fog.h"

// Engine independent.
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_SaveBlock.h"
#endif
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Fog.h"

// T Y P E S /////////////////////////////////////////////////////////////////
struct ChunkData
{
    enum { VERSION = 1 };

    Asura_RGBA		xColour;
    float			fNearPlane;
    float			fFarPlane;
    float			fValueAtFarPlane;
    float			fSkyboxValue;
	Asura_Vector_2	xGraphFogExtraPoint;

};

struct ChunkData_V0
{
    Asura_RGBA xColour;
    float      fNearPlane;
    float      fFarPlane;
    float      fValueAtFarPlane;
    float      fSkyboxValue;
};

// C O N S T A N T S /////////////////////////////////////////////////////////

enum
{
    FLAG_ENABLED = 1 << 0,
	FLAG_GRAPHFOGENABLED = 1 << 1,
};

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#ifdef ASURA_TOOLS
void Asura_Chunk_Fog::Create( Asura_SaveBlock* pxSaveBlock )
{
    const u_int uSize = sizeof(Asura_Chunk_Fog) + sizeof(ChunkData);

    char* const pcOutStart = static_cast<char*>( pxSaveBlock->GetAccessPointer(uSize) );
    char* pcOut = pcOutStart;

    // Construct flags. These are used in place of any booleans since we
    // don't want to rely on the size of that type on different platforms.
    int iFlags = 0;
    if( Asura_Fog::IsFogEnabled() ) { iFlags |= FLAG_ENABLED; }
	if( Asura_Fog::IsGraphFogEnabled() ) {iFlags |= FLAG_GRAPHFOGENABLED; }

    // Header.
    {
        Asura_Chunk_Fog xChunkHdr;

	    xChunkHdr.ID = ASURA_CHUNK_FOG;
	    ASURA_CHUNK_VERIFY( xChunkHdr.ID, int );
	    xChunkHdr.Size = uSize;
	    ASURA_CHUNK_VERIFY( xChunkHdr.Size, int );
        xChunkHdr.Version = ChunkData::VERSION;
	    ASURA_CHUNK_VERIFY( xChunkHdr.Version, int );
	    xChunkHdr.Flags = iFlags;
	    ASURA_CHUNK_VERIFY( xChunkHdr.Flags, int );

        ASURA_CHUNK_FILL( pcOut, xChunkHdr, Asura_Chunk_Fog );
    }

    // Remaining data.
    {
        ChunkData xData;

        xData.xColour = Asura_Fog::GetInitialFogColour();
        ASURA_CHUNK_ARRAY_VERIFY( xData.xColour, float, 4 );
        xData.fNearPlane = Asura_Fog::GetInitialFogNearPlane();
        ASURA_CHUNK_VERIFY( xData.fNearPlane, float );
        xData.fFarPlane = Asura_Fog::GetInitialFogFarPlane();
        ASURA_CHUNK_VERIFY( xData.fFarPlane, float );
        xData.fValueAtFarPlane = Asura_Fog::GetInitialFogValueAtFarPlane();
        ASURA_CHUNK_VERIFY( xData.fValueAtFarPlane, float );
        xData.fSkyboxValue = Asura_Fog::GetInitialSkyboxFogBlendFactor();
        ASURA_CHUNK_VERIFY( xData.fSkyboxValue, float );
		xData.xGraphFogExtraPoint = Asura_Fog::GetInitialGraphFogPoint();
		ASURA_CHUNK_ARRAY_VERIFY( xData.xGraphFogExtraPoint, float, 2 );

        ASURA_CHUNK_FILL( pcOut, xData, ChunkData );
    }


    Asura_Assert( static_cast<u_int>(pcOut - pcOutStart) == uSize, "" );
}
#endif // #ifdef ASURA_TOOLS

bool Asura_Chunk_Fog::Process( Asura_Handle iFileHandle ) const
{
    // Process flags.
    Asura_Fog::EnableFog( 0 != (Flags & FLAG_ENABLED) );
	Asura_Fog::EnableGraphFog( 0 != (Flags & FLAG_GRAPHFOGENABLED) );

    switch( Version )
    {

// Don't bother supporting ancient chunk versions outside tools.
#ifdef ASURA_TOOLS

        case 0:
	    {
		    // Read data.
		    ChunkData_V0 xData;
		    Asura_FileAccess::ReadData( iFileHandle, &xData, sizeof(ChunkData_V0) );

		    // Process.
		    Asura_Fog::SetInitialFogColour( xData.xColour );
		    Asura_Fog::SetInitialFogNearPlane( xData.fNearPlane );
		    Asura_Fog::SetInitialFogFarPlane( xData.fFarPlane );
		    Asura_Fog::SetInitialFogValueAtFarPlane( xData.fValueAtFarPlane );
		    Asura_Fog::SetInitialSkyboxFogBlendFactor( xData.fSkyboxValue );

            break;
	    }

#endif // #ifdef ASURA_TOOLS

        case 1:
        {
		    // Read data.
		    ChunkData xData;
		    Asura_FileAccess::ReadData( iFileHandle, &xData, sizeof(ChunkData) );

			// EP.15.JAN.2009. After changes in fog system, fog always have two values. Curent fog and Initial fog. By calling the SetInitial... we always set Current fog also.
		    Asura_Fog::SetInitialFogColour( xData.xColour );
		    Asura_Fog::SetInitialFogNearPlane( xData.fNearPlane );
		    Asura_Fog::SetInitialFogFarPlane( xData.fFarPlane );
		    Asura_Fog::SetInitialFogValueAtFarPlane( xData.fValueAtFarPlane );
		    Asura_Fog::SetInitialSkyboxFogBlendFactor( xData.fSkyboxValue );
		    Asura_Fog::SetInitialGraphFogPoint( xData.xGraphFogExtraPoint );

            break;
	    }

        default:
        {
            Asura_Assert(false, "Unsupported version number");
            break;
        }
    }

    return true;
}
