//
// Asura_Chunk_GI_Solution.cpp
//
// Created by KJFL - 2004-07-20 10:24:59
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_Chunk_GI_Solution.h"

#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Macros.h"
#endif
#include "ChunkSystem/Asura_Chunk_Stream.h"

#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Light/Asura_GI_Solution.h"

#ifdef ASURA_TOOLS
#include "Lighting/Tools_LightSets.h"
#include "Load&Save/Tools_SaveAsura.h"
#endif


bool Asura_Chunk_GI_Solution::Process( Asura_Handle hFile ) const
{	
	// first check the validity of the chunk 
	//return false;
	// Too new?
    if( Version > CHUNK_VERSION )
    {
        Asura_Assert( false, "Chunk version is too new. You need a new executable." );
        return false;
    }
	
	// Too old?
    if( Version < 2 ) 
    {
        Asura_OutputToDebugger(( "INFO: Loading chunk v%i (current version is v%i.)\n", Version, CHUNK_VERSION ));
		// don't need to exit but we may need to touch up older data
    }

	// (AAM - 09.02.2009) No longer compressing samples in next-gen

	#ifdef ASURA_TARGET_SM3
		const bool bNextGen = true;
	#else
		const bool bNextGen = false;
	#endif

	const bool bUncompressed = ((Flags & ChunkFlag_SamplesUncompressed) != 0);

	if(bNextGen != bUncompressed)
	{
		Asura_Assert(false,"Mixing current-gen and next-gen GI solution chunks. Recalc and resave.");
		return false;
	}

	float afDecompressCoefficientMinimum[12];
	float afDecompressCoefficientScale[12];

	if(!bUncompressed)
	{
		// Decompression coefficients only exist in the chunk if "uncompressed" flag isn't set
		Asura_FileAccess::ReadData(hFile,afDecompressCoefficientMinimum,sizeof(afDecompressCoefficientMinimum));
		Asura_FileAccess::ReadData(hFile,afDecompressCoefficientScale,sizeof(afDecompressCoefficientScale));
	}

	// if there are no samples, then we are done
	if (m_uNumberOfSamples == 0) return false;

	u_int uLightSets = 0;
	if(Version > 2)
	{
		Asura_FileAccess::ReadData(hFile,&uLightSets,sizeof(u_int));	
	}	
	
	#ifdef ASURA_TOOLS
	if(!Tools_LightSets::IsUsedLightSetCombination(uLightSets))
	{
		//not using the light set that this data is for anymore , so ditch it.
		return false;
	}

	//Change to correct handler for this chunk
	u_int uInitialLightSets = Tools_LightSets::GetCurrentLightSets();
	Tools_LightSets::SetCurrentLightSets(uLightSets,false);
	#endif // ASURA_TOOLS

	
	// (Obviously calling Reset prior to the allocation of new memory would solve any memory leaks,
	// but I'd rather know about them, since it shouldn't be happening in the first place.)

	Asura_ResourcePushType(ASURA_RESOURCE_GLOBAL_ILLUMATION);
	{
		if (Asura_GI_Solution::GetCurrentHandler().m_pxSamples)
		{
			#ifdef ASURA_TOOLS
			if(Tools_SaveAsura::IsEditingMission())
			{
				//May end up with a second copy of the sample data when loading the common .gi file
				//Ignore it in preference to that from the mission .gi file would should have been loaded earlier.
				//RAR 15/01/2008
				return false;
			}
			#endif
			
			Asura_Assert( false, "Samples already loaded?" );

			Asura_GI_Solution::GetCurrentHandler().Reset();
		}

		Asura_GI_Solution::GetCurrentHandler().m_uNumberOfSamples = m_uNumberOfSamples;

		#ifndef ASURA_TARGET_SM3
		for (u_int u=0; u<12; u++)
		{
			Asura_GI_Solution::GetCurrentHandler().m_afDecompressCoefficientMinimum[u] = afDecompressCoefficientMinimum[u]; 
			Asura_GI_Solution::GetCurrentHandler().m_afDecompressCoefficientScale[u] = afDecompressCoefficientScale[u];
		}
		#endif

		Asura_ResourceAllocateArray(Asura_GI_Solution::GetCurrentHandler().m_pxSamples, m_uNumberOfSamples);

		if (!Asura_GI_Solution::GetCurrentHandler().m_pxSamples)
		{
			// we've got a memory problem
			Asura_ResourcePopType();
			return false;
		}

		// load 'em in as quick as possible
		Asura_FileAccess::ReadData
			(
			hFile,
			Asura_GI_Solution::GetCurrentHandler().m_pxSamples,
			m_uNumberOfSamples * sizeof ( Asura_GI_IrradianceSample )
			);

#ifdef ASURA_TOOLS
		if(Flags & ChunkFlag_SamplesValid)
		{
			Asura_GI_Solution::GetCurrentHandler().m_bSamplesValid = true;
		}
#endif
	}
	Asura_ResourcePopType();

	#ifdef ASURA_TOOLS
	//Restore the original light set
	Tools_LightSets::SetCurrentLightSets(uInitialLightSets,false);
	#endif
	
	// success
	return true;
}

bool Asura_Chunk_GI_Solution::Create( Asura_SaveBlock& xSaveBlock )
{
#ifdef ASURA_TOOLS
//	Asura_GI_Solution::Tools_InitialiseFromPhotonMappingSystem();
	
	Asura_Chunk_SaveBlock_Stream	xStream (&xSaveBlock);

	return Create(xStream);

#else	// not defined(ASURA_TOOLS)
	return true;
#endif // end else not defined(ASURA_TOOLS)

}



bool Asura_Chunk_GI_Solution::Create( Asura_Chunk_Stream& xStream )
{
#ifdef ASURA_TOOLS
//	Asura_GI_Solution::Tools_InitialiseFromPhotonMappingSystem();

	int iChunkFlags = 0;
	if(Asura_GI_Solution::GetCurrentHandler().m_bSamplesValid) iChunkFlags |= ChunkFlag_SamplesValid;

#ifdef ASURA_TARGET_SM3
	// Next-gen projects always write uncompressed samples
	iChunkFlags |= ChunkFlag_SamplesUncompressed;
#endif
	
	xStream.WriteChunk_Start( ASURA_CHUNK_GI_SOLUTION, CHUNK_VERSION, iChunkFlags );

	
	//const u_int uChunkSize = sizeof ( Asura_Chunk_GI_Solution );
	const u_int uNumberOfSamples = Asura_GI_Solution::GetCurrentHandler().m_uNumberOfSamples;

	xStream << uNumberOfSamples;

#ifndef ASURA_TARGET_SM3

	if(Tools_Level_Export::IsExport() && Tools_Level_Export::ExportFullBright())
	{
		//Doing export with fullbright flag , so force all the samples to an overhead
		//white light. RAR 16/09/04
		for (u_int u=0; u<12; u++)
		{
			float f = ((u%4) == 2) ?  -1.0f : 0.0f;

			xStream << f;

		}
	
		float f = 0.0f;

		for (u_int u=0; u<12; u++)
		{
			xStream << f;
		}
	
	}
	else
	{
		for (u_int u=0; u<12; u++)
		{
			xStream << Asura_GI_Solution::GetCurrentHandler().m_afDecompressCoefficientMinimum[u];
		}

		for (u_int u=0; u<12; u++)
		{
			xStream <<  Asura_GI_Solution::GetCurrentHandler().m_afDecompressCoefficientScale[u];
		}
	}

#endif // ndef ASURA_TARGET_SM3

	u_int uLightSets = Tools_LightSets::GetCurrentLightSets();
	xStream << uLightSets;

   	const Asura_Vector_3& xPositionOffset = Tools_Level_Export::GetExportCentre();

	for ( u_int u=0; u<uNumberOfSamples; ++u )
	{
		//RAR 08/03/2006 Need to be careful with recenreing these for level exports , since the y value
		//uses some bits for other stuff.
		float fOutputY =  (Asura_GI_Solution::GetCurrentHandler().m_pxSamples[u].m_afPosition[1] - xPositionOffset.y);
		int iOutputY = *(int*)&fOutputY;
		iOutputY = (iOutputY &~ 3) | (Asura_GI_Solution::GetCurrentHandler().m_pxSamples[u].GetPlaneAxis());
		fOutputY = *(float*)&iOutputY;

		
		xStream << Asura_GI_Solution::GetCurrentHandler().m_pxSamples[u].m_afPosition[0] - xPositionOffset.x;
		xStream << fOutputY;
		xStream << Asura_GI_Solution::GetCurrentHandler().m_pxSamples[u].m_afPosition[2] - xPositionOffset.z;

		for ( u_int v=0; v<12; ++v)
		{
			#ifdef ASURA_TARGET_SM3
				if(Tools_Level_Export::IsExport() && Tools_Level_Export::ExportFullBright())
				{
					// Current-gen did 'full bright' exports by fudging the decompression coeffs
					// above. These don't exist any more, so modify the samples themselves
					const float fCoeff = ((v%4) == 2) ?  -1.0f : 0.0f;
					xStream << fCoeff;	
				}
				else
				{
					xStream << Asura_GI_Solution::GetCurrentHandler().m_pxSamples[u].m_afCoefficients[v];
				}
			#else
				xStream << Asura_GI_Solution::GetCurrentHandler().m_pxSamples[u].m_aucCompressedCoefficients[v];
			#endif
		}
	}

	xStream.WriteChunk_End();
#endif

    return true;

}

