//
// Asura_Chunk_Lights.cpp
//
// Created by Kevin Lea - 12/06/2001 11:18:38
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//


#include "Core/Asura.h"

#include "Core/Asura_PlatformBasics.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Lights.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"

#include "Light/Asura_Light.h"

#include "FileAccess/Asura_FileAccess.h"
#include "ChunkSystem/Asura_SaveBlock.h"



#define ASURA_CHUNK_LIGHT_VERSION 5
// (AW - 26.02.2004) - Ambient entity lights (3 floats) added for version 5.
/// Kevin Lea - 05/02/2003 18:23:17 - new with version 4
/// BottomAmbience (3 floats) & a flag (u_int) at the beginning of the chunk

//old version of the light structure
struct Asura_Light_V2
{
	// name?

	Asura_Vector_3 Position;
	Asura_Vector_3 Direction;

	float R;
	float G;
	float B;

	float Brightness;
	float Range;
	float Angle;

	float ShadowStrength;

	float Padding[3]; // for future expansion

	/// Kevin Lea - 26/09/2002 11:28:36 - added a union so that all the
	/// light flags can be copied just by using the 'm_uFlags' member
	union
	{
		struct
		{
			u_int	HasCorona:1,
					InfiniteLight:1,
					AffectsEntities:1,
					AffectsEnvironment:1,
					SpareBits:28;
		};

		u_int		m_uFlags;
	};
	
	/// Kevin Lea - 26/09/2002 11:33:28 - left IsVolumetric as a bool rather than
	/// put it in the flags above since this would involve a more complicated change
	/// to the light chunk loaders. Next time someone does a big change to this class
	/// which results in a different size chunk being saved, could they please tidy
	/// things up. Ta.
	bool IsVolumetric;

	float BrightnessOverRange;
	Asura_Vector_3 OldPosition;
	float OldRange;
	bool HasChanged;
};


bool Asura_Chunk_Lights::Process(Asura_Handle fileHandle) const
{
	Asura_Assert(Version<=ASURA_CHUNK_LIGHT_VERSION,"Unknown chunk version");

#ifndef ASURA_TOOLS
    // Don't bother supporting ancient chunk versions outside of tools.
    Asura_Assert( Version >= 5, "" );
#endif

	Asura_Lighting_System::AmbientR = AmbientR;
	Asura_Lighting_System::AmbientG = AmbientG;
	Asura_Lighting_System::AmbientB = AmbientB;

#ifdef ASURA_TOOLS
    if( Version < 5 )
    {
		Asura_Lighting_System::s_fEntityAmbientR = 0.0f;
		Asura_Lighting_System::s_fEntityAmbientG = 0.0f;
		Asura_Lighting_System::s_fEntityAmbientB = 0.0f;
    }
    else
#endif
	{
		Asura_FileAccess::ReadData(fileHandle, &Asura_Lighting_System::s_fEntityAmbientR, sizeof(float));
		Asura_FileAccess::ReadData(fileHandle, &Asura_Lighting_System::s_fEntityAmbientG, sizeof(float));
		Asura_FileAccess::ReadData(fileHandle, &Asura_Lighting_System::s_fEntityAmbientB, sizeof(float));
	}

    Asura_Lighting_System::s_uEntityAmbientRGB = (Asura_Maths::FloatToInt(Asura_Lighting_System::s_fEntityAmbientR * 255.0f) << 16) | 
							(Asura_Maths::FloatToInt(Asura_Lighting_System::s_fEntityAmbientG * 255.0f) << 8) | 
							(Asura_Maths::FloatToInt(Asura_Lighting_System::s_fEntityAmbientB * 255.0f));

#ifdef ASURA_TOOLS
    if( Version < 4 )
	{
		/// Kevin Lea - 05/02/2003 18:07:19 -  default new features to off
		Asura_Lighting_System::s_bUseHemisphereAmbient = false;

		Asura_Lighting_System::s_fBottomAmbientR = AmbientR;
		Asura_Lighting_System::s_fBottomAmbientG = AmbientG;
		Asura_Lighting_System::s_fBottomAmbientB = AmbientB;
	}
    else
#endif
    {
		u_int uUseHemisphereAmbient;
		Asura_FileAccess::ReadData(fileHandle,&Asura_Lighting_System::s_fBottomAmbientR,sizeof(float));
		Asura_FileAccess::ReadData(fileHandle,&Asura_Lighting_System::s_fBottomAmbientG,sizeof(float));
		Asura_FileAccess::ReadData(fileHandle,&Asura_Lighting_System::s_fBottomAmbientB,sizeof(float));
		Asura_FileAccess::ReadData(fileHandle,&uUseHemisphereAmbient,sizeof(u_int));

		Asura_Lighting_System::s_bUseHemisphereAmbient = (uUseHemisphereAmbient!=0);
	}

#ifndef ASURA_TOOLS
	// allocate just enough memory for the number of lights
	Asura_Assert(!Asura_Lighting_System::LightStorage, "lights already allocated, more than one light chunk?");
	Asura_Lighting_System::LightStorage = new Asura_Light[NumberOfLights];
#endif

	Asura_Light source;
	for(int i=0; i<NumberOfLights; i++)
	{
		Asura_Light* light = Asura_Lighting_System::AllocateLight();
		Asura_Assert(light, "failed to allocate Asura_Light");
		if (!light) return false;
			
#ifdef ASURA_TOOLS
		if(Version<3)
		{
			//fun fun.... copy from old version of light
			Asura_Light_V2 xSourceV2;
			Asura_FileAccess::ReadData(fileHandle,&xSourceV2,sizeof(xSourceV2));
			source.SetToDefault();

			source.Position = xSourceV2.Position;
			source.Direction = xSourceV2.Direction;
			source.R = xSourceV2.R;
			source.G = xSourceV2.G;
			source.B = xSourceV2.B;
			source.Brightness = xSourceV2.Brightness;
			source.Range = xSourceV2.Range;
			source.Angle = xSourceV2.Angle;
			source.ShadowStrength = xSourceV2.ShadowStrength;
			source.m_uFlags = xSourceV2.m_uFlags;
			source.IsVolumetric = xSourceV2.IsVolumetric;


			/// Kevin Lea - 26/09/2002 16:05:48 - cope with older chunks by setting default settings
			if (Version < 2)
			{
				source.AffectsEntities = false;
				source.AffectsEnvironment = true;
			}

			if(source.Angle>0.0f && source.Angle<360.0f && !source.AffectsEntities && !source.InfiniteLight)
			{
				//adjust brightness to take into account changes to lighting calculation
				source.Brightness *= ((1.0f - Asura_Maths::Cosine((Asura_Maths::Pi/360.0f)*source.Angle))/2.0f);
			}

		}
		else
#endif
		{
			Asura_FileAccess::ReadData(fileHandle,&source,sizeof(Asura_Light));
		}

		// Kevin Lea - 04/10/2001 16:04:01 - setting the light twice makes sure the 'old' position is equal to the new position
		light->Set(source);
		light->Set(source);

	}

	Asura_Lighting_System::MakeStaticPhotonsList();

	return true;

}

#ifdef ASURA_TOOLS
void Asura_Chunk_Lights::Create(Asura_SaveBlock& saveBlock)
{
	int numberOfLights = Asura_Lighting_System::NumberOfActiveLights;

	int chunkSize = sizeof(Asura_Chunk_Lights) + sizeof(float)*6 + sizeof(u_int) + sizeof(Asura_Light)*numberOfLights;

	Asura_Chunk_Lights *chunk = (Asura_Chunk_Lights*)saveBlock.GetAccessPointer(chunkSize);

	chunk->ID = ASURA_CHUNK_LIGHTS;
	ASURA_CHUNK_VERIFY(chunk->ID, int);
	chunk->Size = chunkSize;
	ASURA_CHUNK_VERIFY(chunk->Size, int);
	chunk->Version = ASURA_CHUNK_LIGHT_VERSION;
	ASURA_CHUNK_VERIFY(chunk->Version, int);
	chunk->Flags = 0;
	ASURA_CHUNK_VERIFY(chunk->Flags, int);

	chunk->NumberOfLights = numberOfLights;
	ASURA_CHUNK_VERIFY(chunk->NumberOfLights, int);

	chunk->AmbientR = Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::AmbientR);
	ASURA_CHUNK_VERIFY(chunk->AmbientR, float);
	chunk->AmbientG = Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::AmbientG);
	ASURA_CHUNK_VERIFY(chunk->AmbientG, float);
	chunk->AmbientB = Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::AmbientB);
	ASURA_CHUNK_VERIFY(chunk->AmbientB, float);

	char* pcDest = (char*)(chunk+1);

	ASURA_CHUNK_FILL(pcDest, Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::s_fEntityAmbientR), float);
	ASURA_CHUNK_VERIFY(*(pcDest-sizeof(float)), float);
	ASURA_CHUNK_FILL(pcDest, Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::s_fEntityAmbientG), float);
	ASURA_CHUNK_VERIFY(*(pcDest-sizeof(float)), float);
	ASURA_CHUNK_FILL(pcDest, Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::s_fEntityAmbientB), float);
	ASURA_CHUNK_VERIFY(*(pcDest-sizeof(float)), float);

	u_int uUseHemisphereAmbient = Asura_Lighting_System::s_bUseHemisphereAmbient;
	ASURA_CHUNK_VERIFY(uUseHemisphereAmbient, u_int);

	ASURA_CHUNK_FILL(pcDest,Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::s_fBottomAmbientR),float);
	ASURA_CHUNK_VERIFY(*(pcDest-sizeof(float)), float);
	ASURA_CHUNK_FILL(pcDest,Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::s_fBottomAmbientG),float);
	ASURA_CHUNK_VERIFY(*(pcDest-sizeof(float)), float);
	ASURA_CHUNK_FILL(pcDest,Tools_Level_Export::ScaleLightBrightness(Asura_Lighting_System::s_fBottomAmbientB),float);
	ASURA_CHUNK_VERIFY(*(pcDest-sizeof(float)), float);
	ASURA_CHUNK_FILL(pcDest,uUseHemisphereAmbient,u_int);

	Asura_Light* dest = (Asura_Light*)pcDest;
	
	for(int i=0; i<numberOfLights; i++)
	{
		*dest = Asura_Lighting_System::LightStorage[i];
		//Light brightness may need to be scaled during level exports. RAR 09/02/2004
		dest->Brightness = Tools_Level_Export::ScaleLightBrightness(dest->Brightness);

		ASURA_CHUNK_ARRAY_VERIFY(dest->Position, float, 3);
		ASURA_CHUNK_ARRAY_VERIFY(dest->Direction, float, 3);
		ASURA_CHUNK_VERIFY(dest->R, float);
		ASURA_CHUNK_VERIFY(dest->G, float);
		ASURA_CHUNK_VERIFY(dest->B, float);
		ASURA_CHUNK_VERIFY(dest->Brightness, float);
		ASURA_CHUNK_VERIFY(dest->Range, float);
		ASURA_CHUNK_VERIFY(dest->m_fInnerRange, float);
		ASURA_CHUNK_VERIFY(dest->Angle, float);
		ASURA_CHUNK_VERIFY(dest->ShadowStrength, float);
		ASURA_CHUNK_ARRAY_VERIFY(dest->m_xBoundingBox, float, 6);
		// KOR - 12/05/2003 19:41:09 - Flags are referenced by both bitmasks and bitfields, which is...
		// problematic. Fortunately, it looks like the only time bitmasks are used is for tools,
		// so we want to use the bitfield version.
		ASURA_CHUNK_BITFIELD_VERIFY(dest->m_uFlags, u_int);
		ASURA_CHUNK_VERIFY(dest->BrightnessOverRange, float);
		ASURA_CHUNK_ARRAY_VERIFY(dest->OldPosition, float, 3);
		ASURA_CHUNK_VERIFY(dest->OldRange, float);
		ASURA_CHUNK_VERIFY(dest->HasChanged, bool);
		dest++;
	}

}
#endif

bool Asura_Chunk_LightVisibilities::Process(Asura_Handle fileHandle) const
{
#ifndef ASURA_TOOLS
    Asura_Assert(false, "shouldn't be loading light visibilities outside Tools");
    return false;
#else
	
	for(int i=0; i<NumberOfVolumetricLights; i++)
	{
		int index;
		int numIndices;

		Asura_FileAccess::ReadData(fileHandle,&index,sizeof(int));
		Asura_FileAccess::ReadData(fileHandle,&numIndices,sizeof(int));

		
		Asura_FileAccess::ReadData(fileHandle,&Asura_Lighting_System::Visibility[index].MaxRange,sizeof(float));

		if(Asura_Lighting_System::Visibility[index].MaxRange==0.0f)
		{
			Asura_Lighting_System::Visibility[index].MaxRange = Asura_Lighting_System::LightStorage[index].Range;
		}

		if(numIndices)
		{
			Asura_Assert(!Asura_Lighting_System::Visibility[index].Indices,"");

			//allocate space for visibilities
			//Asura_Lighting_System::Visibility[index].Indices = (int*)realloc(Asura_Lighting_System::Visibility[index].Indices,sizeof(int)*numIndices);
			Asura_Lighting_System::Visibility[index].Indices = (int*)Asura_Allocate(sizeof(int)*numIndices);
			Asura_Assert(Asura_Lighting_System::Visibility[index].Indices,"Allocation failed");
			Asura_Lighting_System::Visibility[index].NumberAllocated = numIndices;
		}

		
		for(int l=0; l<numIndices; l++)
		{
			int indexToAdd;
			Asura_FileAccess::ReadData(fileHandle,&indexToAdd,sizeof(int));
			Asura_Lighting_System::Visibility[index].AddIndex(indexToAdd,NULL);
		}
	
	}


	Asura_Lighting_System::Update();
//	Asura_Lighting_System::ForceLightingUpdate = true;


	return true;
#endif
}

#ifdef ASURA_TOOLS
void Asura_Chunk_LightVisibilities::Create(Asura_SaveBlock& saveBlock)
{
	int numberOfLights = 0;
	int numberOfIndices = 0;
	int i;

	for (i=0; i<Asura_Lighting_System::NumberOfActiveLights; i++)
	{
		if (Asura_Lighting_System::LightStorage[i].IsVolumetric)
		{
			numberOfLights++;
			numberOfIndices+=Asura_Lighting_System::Visibility[i].NumberOfVisibleVertices;
		}
	}

	int chunkSize = sizeof(Asura_Chunk_LightVisibilities)+sizeof(int)*(numberOfLights*3+numberOfIndices);

	Asura_Chunk_LightVisibilities *chunk = (Asura_Chunk_LightVisibilities*)saveBlock.GetAccessPointer(chunkSize);

	chunk->ID = ASURA_CHUNK_LIGHTVISIBILITIES;
	ASURA_CHUNK_VERIFY(chunk->ID, int);
	chunk->Size = chunkSize;
	ASURA_CHUNK_VERIFY(chunk->Size, int);
	chunk->Version = 1;
	ASURA_CHUNK_VERIFY(chunk->Version, int);
	chunk->Flags = 0;
	ASURA_CHUNK_VERIFY(chunk->Flags, int);

	chunk->NumberOfVolumetricLights = numberOfLights;
	ASURA_CHUNK_VERIFY(chunk->NumberOfVolumetricLights, int);

	int* dest = (int*)(chunk+1);

	for(i=0; i<Asura_Lighting_System::NumberOfActiveLights; i++)
	{
		if (!Asura_Lighting_System::LightStorage[i].IsVolumetric) continue;
		
		*dest = i;
		ASURA_CHUNK_VERIFY(*dest, int);
		dest++;
		*dest = Asura_Lighting_System::Visibility[i].NumberOfVisibleVertices;
		ASURA_CHUNK_VERIFY(*dest, int);
		dest++;
		*dest = *((int*)&Asura_Lighting_System::Visibility[i].MaxRange);
		ASURA_CHUNK_VERIFY(*dest, int);
		dest++;

		for (int l=0; l<Asura_Lighting_System::Visibility[i].NumberOfVisibleVertices; l++)
		{
			*dest = Asura_Lighting_System::Visibility[i].Indices[l];
			ASURA_CHUNK_VERIFY(*dest, int);
			dest++;
		}
		
	}

}
///////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------------**
** Asura_Chunk_LightScale **
**------------------------*/
#ifdef ASURA_TOOLS

#define ASURA_CHUNK_LIGHTSCALE_VERSION 0

bool Asura_Chunk_LightScale::Process(Asura_Handle iFileHandle) const
{
	if(Version>ASURA_CHUNK_LIGHTSCALE_VERSION)
	{
		Asura_Assert(Version<=ASURA_CHUNK_LIGHTSCALE_VERSION,"Unrecognised light scale chunk version");
		return false;
	}
	if(m_iNumberOfScaleValues > ASURA_NUM_LIGHT_SCALE_VALUES)
	{
		Asura_Assert(m_iNumberOfScaleValues <= ASURA_NUM_LIGHT_SCALE_VALUES,"Too many light scale values found in chunk");
		return false;
	}

	for(int i=0;i<m_iNumberOfScaleValues;i++)
	{
		ASURA_CHUNK_FILEEXTRACT(iFileHandle,Asura_Lighting_System::s_afPrelightScale[i],float)
		ASURA_CHUNK_FILEEXTRACT(iFileHandle,Asura_Lighting_System::s_afRuntimeLightScale[i],float)
	}
	return true;
}

void Asura_Chunk_LightScale::Create(Asura_SaveBlock* pxSaveBlock)
{
	//determine size of chunk
	int iChunkSize = sizeof(Asura_Chunk_LightScale);
	iChunkSize += 2*sizeof(float)*ASURA_NUM_LIGHT_SCALE_VALUES;

	//allocate chunk
	Asura_Chunk_LightScale* pxChunk = (Asura_Chunk_LightScale*) pxSaveBlock->GetAccessPointer(iChunkSize);

	//and fill it in
	pxChunk->ID = ASURA_CHUNK_LIGHTSCALE;
	ASURA_CHUNK_VERIFY(pxChunk->ID, int);
	pxChunk->Size = iChunkSize;
	ASURA_CHUNK_VERIFY(pxChunk->Size, int);
	pxChunk->Version = ASURA_CHUNK_LIGHTSCALE_VERSION;
	ASURA_CHUNK_VERIFY(pxChunk->Version, int);
	pxChunk->Flags = 0;
	ASURA_CHUNK_VERIFY(pxChunk->Flags, int);

	pxChunk->m_iNumberOfScaleValues = ASURA_NUM_LIGHT_SCALE_VALUES;
	ASURA_CHUNK_VERIFY(pxChunk->m_iNumberOfScaleValues, float);


	char* pcChunkBuffer = (char*)(pxChunk+1);

	for(int i=0;i<ASURA_NUM_LIGHT_SCALE_VALUES;i++)
	{
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,Asura_Lighting_System::s_afPrelightScale[i],float)
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer,Asura_Lighting_System::s_afRuntimeLightScale[i],float)
	}
}


#endif // ASURA_TOOLS

///////////////////////////////////////////////////////////////////////////////////////////////////
#include "ChunkSystem/Asura_Huffman.h"
#include "Core/Asura_String.h"
#include "ChunkSystem/Asura_Chunk_Texture.h"
#include "ChunkSystem/Asura_Chunk_Corona.h"
#include "Core/Asura_NameTree.h"
#include "SpecialFX/Asura_StaticCorona_System.h"

bool SaveLights(const char* filename, bool bMissionFile)
{
	Asura_SaveBlock saveBlock;

	Asura_Chunk_Lights::Create(saveBlock);
	Asura_Chunk_LightVisibilities::Create(saveBlock);
//	Asura_Chunk_SkyBox::Create(saveBlock);

	//save textures used by coronas
	Asura_NameTree xTextureTree;
	Asura_StaticCorona_System::BuildTextureTree(&xTextureTree);
	if(xTextureTree.GetNumberOfNodes())
	{
		Asura_Chunk_TextureNames::Create(&xTextureTree,&saveBlock);
	}

	//and save the coronas themselves
	Asura_Chunk_Corona::Create(&saveBlock,false,bMissionFile);

	//save the tools light scale chunk
	Asura_Chunk_LightScale::Create(&saveBlock);

	saveBlock.AddAsuraTerminator();

	return saveBlock.SaveFile(filename,true);
}
#endif//ASURA_TOOLS
