//
// Asura_Chunk_Stream.cpp
//
// Created by KOR - 23/06/2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_Chunk_Stream.h"

// Asura headers
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_HalfFloat.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Maths/Asura_CompressedVector3_101010N.h"
#include "Core/Memory/Asura_Mem.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Colour.h"

#ifdef TEST_ENTITY_CHUNK_STREAM
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Asura_Entity.h"
#endif

// Macros for swapping endian and reading/writing basic types
#if defined ASURA_TOOLS || !defined ASURA_MAIN

    #define ASURA_CHUNK_STREAM_WRITE(xValue, xType) \
		Asura_Assert(m_bHasSetSwapEndian, "Haven't set whether to endian swap for this stream yet."); \
		xType xTemp = xValue; \
		if(m_bSwapEndian) { Asura_Endian::SwapBytes(reinterpret_cast<char*>(&xTemp), sizeof(xType)); }  \
		Write(sizeof(xType), reinterpret_cast<const void*>(&xTemp)); \
		return *this;
	
    #define ASURA_CHUNK_STREAM_READ(xValue, xType) \
        xType xTemp; \
        Read(sizeof(xType), reinterpret_cast<void*>(&xTemp)); \
		if(m_bSwapEndian) { Asura_Endian::SwapBytes(reinterpret_cast<char*>(&xTemp), sizeof(xType)); }  \
        xValue = xTemp; \
        return *this;

#else // not def ASURA_TOOLS and def ASURA_MAIN

    #define ASURA_CHUNK_STREAM_WRITE( xValue, type ) \
        Write( sizeof(type), &xValue ); \
        return *this;

    #define ASURA_CHUNK_STREAM_READ( xValue, type ) \
        Read( sizeof(type), &xValue ); \
        return *this;

#endif // else not def ASURA_TOOLS and def ASURA_MAIN


// Static variables...

u_char	Asura_Chunk_Stream::s_ucPaddingValue	= 0x00;


bool Asura_Chunk_Stream::ReadChunk_GetNextID( char* const pcChunkBuffer )
{
	// Error checking
	if(m_eState == STREAM_ERROR) return false;
	Asura_Assert(m_eState == STREAM_READY, "Unexpected state");

	// JAT 15/04/07 assumes aligned to a 4-byte boundary
	Asura_Assert( ( ( (u_int)(pcChunkBuffer) % 4 ) == 0 ), "Start of chunk not 4-byte aligned" );

	Asura_Chunk_Header* const pxHeader = reinterpret_cast<Asura_Chunk_Header*>( pcChunkBuffer );

	m_iChunkSize = 0;

	// Read in the chunk ID to the given buffer
	Read( 4, &(pxHeader->ID) );

	// Get the ID 
	m_iChunkID = pxHeader->ID;

	// Set state
	m_eState = m_iChunkID ? STREAM_READ_ID : STREAM_READ_END;

	// Valid if non-zero
	return (m_iChunkID != 0);
}

bool Asura_Chunk_Stream::ReadChunk_GetNextHeader( char* const pcChunkBuffer, const int iHeaderSize )
{
	// Error checking
	if(m_eState == STREAM_ERROR) return false;
	Asura_Assert(m_eState == STREAM_READ_ID, "Must call GetNewChunkID before GetNewChunkHeader");

	// JAT 15/04/07 assumes aligned to a 4-byte boundary
	Asura_Assert( ( ( (u_int)(pcChunkBuffer) % 4 ) == 0 ), "Start of chunk not 4-byte aligned" );

	Asura_Chunk_Header* const pxHeader = reinterpret_cast<Asura_Chunk_Header*>( pcChunkBuffer );

	// Read in chunk header without chunk ID, as that will have been done previously
	const int iBytesRead = Read( iHeaderSize - 4, &( pxHeader->Size ) );

	// Get the size
	const int iChunkSize = pxHeader->Size;

	// Use chunk size for verification
	Asura_Assert(iChunkSize >= static_cast<int>(sizeof(Asura_Chunk_Header)), "Chunk smaller than header size");
#ifdef ASURA_CHUNK_STREAM_ALLOW_OTHER_IO
	m_iExpectedEndPosition = GetPosition() + iChunkSize - iHeaderSize;
#else
	m_iByteCounter = iChunkSize - iHeaderSize;
#endif
	// Set state
	m_eState = STREAM_READ_HEADER;
	// Valid if whole header was read.
	return (iBytesRead == (iHeaderSize - 4));
}

bool Asura_Chunk_Stream::ReadChunk_End(bool bFinishedReading)
{
	// Error checking
	if(m_eState == STREAM_ERROR) return false;
	Asura_Assert(m_eState == STREAM_READ_HEADER, "Must call GetNewChunkHeader before EndChunkRead");

	// Validate stream position
#ifdef ASURA_CHUNK_STREAM_ALLOW_OTHER_IO
	int iPositionDifference = m_iExpectedEndPosition - GetPosition();
#else
	int iPositionDifference = m_iByteCounter;
#endif
	if(iPositionDifference)
	{
		if(iPositionDifference < 0)
		{
			// Gone past. This should not happen once we are fully moved over to the new system, as it will be caught on the read
			Asura_Assertf(0, ("Unexpected end position for chunk %c%c%c%c: went %i bytes past end of chunk.\nSubsequent chunks in file unrecoverable",
                              m_acTextID[0], m_acTextID[1], m_acTextID[2], m_acTextID[3],
							  -iPositionDifference)
							 );
			m_eState = STREAM_ERROR;
			return false;
		}
		else
		{
			// Not reached - skip data to end
			if ( Skip(iPositionDifference) != static_cast<u_int>( iPositionDifference ) )
			{
				Asura_Assert( false, "Failed to skip data to end" );
				m_eState = STREAM_ERROR;
				return false;
			}

			if(bFinishedReading)
			{
				Asura_Assertf(0, ("Unexpected end position for chunk %c%c%c%c: didn't read %i bytes of chunk - skipping extra data",
	                              m_acTextID[0], m_acTextID[1], m_acTextID[2], m_acTextID[3],
								  iPositionDifference)
								 );
			}
			else
			{
				iPositionDifference = 0;
			}
		}
	}
	// Read for next chunk
	m_eState = STREAM_READY;
	return true;
}

void Asura_Chunk_Stream::WriteChunk_Start(int iChunkID, int iVersion, int iFlags)
{
#ifdef ASURA_TOOLS
	if(!m_bHasSetSwapEndian)
	{
		SetSwapEndian(Tools_Level_Export::ShouldExportBigEndian());
	}
#endif
	Asura_Assert(m_eState == STREAM_READY, "Unexpected stream state");
	m_iChunkSize = 0;
	m_uChunkStartPosition = GetPosition();
	m_eState = STREAM_WRITING_CHUNK;
	// Use placeholder for size
	*this << iChunkID << m_iChunkSize << iVersion << iFlags;
}

void Asura_Chunk_Stream::WriteChunk_End()
{
    int iChunkSizeToWrite = m_iChunkSize;
#if defined ASURA_TOOLS || !defined ASURA_MAIN
    if( m_bSwapEndian )
    {
        iChunkSizeToWrite = Asura_Endian::SwapBytes( iChunkSizeToWrite );
    }
#endif

    Asura_Assert(m_eState == STREAM_WRITING_CHUNK, "Unexpected stream state");
// ARC - 09 February 2006 14:56:57 - Removed as per Mike Healey's suggestion
    //Asura_Assert(((m_iChunkSize & 3) == 0), "Chunk not word aligned"); // pad here if this happens?

    // Insert size
    Write( sizeof(int), &iChunkSizeToWrite, m_uChunkStartPosition + 4 );

    m_eState = STREAM_READY;
}

void Asura_Chunk_Stream::WriteResourceChunk_Start( u_int uResourceType, u_int uResourceSubType, int iVersion, int iFlags, const Asura_String& xFilename )
{
#ifdef ASURA_TOOLS
	if(!m_bHasSetSwapEndian)
	{
		SetSwapEndian(Tools_Level_Export::ShouldExportBigEndian());
	}
#endif
	Asura_Assert(m_eState == STREAM_READY, "Unexpected stream state");
	
	m_iChunkSize = 0; 	// Use placeholder for size
	m_uChunkStartPosition = GetPosition();
	m_eState = STREAM_WRITING_CHUNK;

	// first the usual chunk info
	*this << ASURA_CHUNK_RESOURCEFILE << m_iChunkSize << iVersion << iFlags;

	// KJFL - 02 January 2007 - and then the resource-chunk specific data...
	// The ResourceFileSize is equal to the size of the whole chunk minus the header and the filename string,
	// so we'll store that value for later, and write a dummy value to the stream to touch up later.
	m_uResourceFileSizeOffset = ( sizeof(Asura_Chunk_ResourceFile) + xFilename.GetSizeInInts()*4 );
	*this << uResourceType << uResourceSubType << m_uResourceFileSizeOffset;

	*this << xFilename;
}

void Asura_Chunk_Stream::WriteResourceChunk_End( void )
{
    int iChunkSizeToWrite = m_iChunkSize;
	int iResourceSizeToWrite = m_iChunkSize - m_uResourceFileSizeOffset; 

#if defined ASURA_TOOLS || !defined ASURA_MAIN
	// Remember we may need to swap the endianess of the sizes 
    if( m_bSwapEndian )
    {
        iChunkSizeToWrite = Asura_Endian::SwapBytes( iChunkSizeToWrite );
		iResourceSizeToWrite = Asura_Endian::SwapBytes( iResourceSizeToWrite );
    }
#endif

    Asura_Assert(m_eState == STREAM_WRITING_CHUNK, "Unexpected stream state");

	// Find to were we're meant to be writing
   	Asura_Chunk_ResourceFile* pxDummy = 0;
	u_int uChunkSizeOffset = (u_int)(&(pxDummy->Size));
	u_int uFileSizeOffset = (u_int)(&(pxDummy->FileSize));

	// Insert size of chunk
    Write( sizeof(int), &iChunkSizeToWrite, m_uChunkStartPosition + uChunkSizeOffset );

    // Insert size of resource data
    Write( sizeof(int), &iResourceSizeToWrite, m_uChunkStartPosition + uFileSizeOffset );

    m_eState = STREAM_READY;
}


void Asura_Chunk_Stream::WritePadding (const u_int uNumBytes, const u_char ucPadValue)
{
	// Simply write the pad value the requisite number of times...
	for (u_int u = 0; u < uNumBytes; u++)
	{
		*this << ucPadValue;
	}
}

void Asura_Chunk_Stream::SkipPadding (const u_int uNumBytes)
{
	// Do we have any data to skip?
	if (0 < uNumBytes)
	{
		// Yes - skip the requisite number of bytes...
		Skip	(uNumBytes);
	}
}


// Stream operators for basic types - these do endian swapping if necessary
Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (int& iValue)
{
	ASURA_CHUNK_STREAM_READ(iValue, int);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const int iValue)
{
	ASURA_CHUNK_STREAM_WRITE(iValue, int);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (u_int& uValue)
{
	ASURA_CHUNK_STREAM_READ(uValue, u_int);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const u_int uValue)
{
	ASURA_CHUNK_STREAM_WRITE(uValue, u_int);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (float& fValue)
{
	ASURA_CHUNK_STREAM_READ(fValue, float);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const float fValue)
{
	ASURA_CHUNK_STREAM_WRITE(fValue, float);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (short& sValue)
{
	ASURA_CHUNK_STREAM_READ(sValue, short);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const short sValue)
{
	ASURA_CHUNK_STREAM_WRITE(sValue, short);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (u_short& usValue)
{
	ASURA_CHUNK_STREAM_READ(usValue, u_short);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const u_short usValue)
{
	ASURA_CHUNK_STREAM_WRITE(usValue, u_short);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (char& cValue)
{
	ASURA_CHUNK_STREAM_READ(cValue, char);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const char cValue)
{
	ASURA_CHUNK_STREAM_WRITE(cValue, char);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (u_char& ucValue)
{
	ASURA_CHUNK_STREAM_READ(ucValue, u_char);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const u_char ucValue)
{
	ASURA_CHUNK_STREAM_WRITE(ucValue, u_char);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (int8& cValue)
{
	ASURA_CHUNK_STREAM_READ(cValue, int8);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const int8 cValue)
{
	ASURA_CHUNK_STREAM_WRITE(cValue, int8);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (bool& bValue)
{
	u_char ucValue;
	*this >> ucValue;
	bValue = (ucValue) ? true : false;
	return *this;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const bool bValue)
{
	u_char ucValue = static_cast<u_char>(bValue);
	ASURA_CHUNK_STREAM_WRITE(ucValue, u_char);
}

#if defined(ASURA_UNICODE_CHAR_DISTINCT_TYPE)

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_UnicodeChar& wcValue)
{
	ASURA_CHUNK_STREAM_READ(wcValue, Asura_UnicodeChar);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_UnicodeChar wcValue)
{
	ASURA_CHUNK_STREAM_WRITE(wcValue, Asura_UnicodeChar);
}

#endif	// defined(ASURA_UNICODE_CHAR_DISTINCT_TYPE)


// Stream operators for Asura structures.
// NOTE: These should not use Read/Write directly, but instead use the stream operators on the basic types
Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_CompressedVector3_101010N& xValue)
{
	return *this >> xValue.m_uValue;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_CompressedVector3_101010N& xValue)
{
	return *this << xValue.m_uValue;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_HalfFloat& xValue)
{
	return *this >> xValue.m_usBits;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_HalfFloat& xValue)
{
	return *this << xValue.m_usBits;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Comp_Quat& xValue)
{
	u_int uBits;
	*this >> uBits;
	xValue.SetBits( uBits );
	return *this;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Comp_Quat& xValue)
{
	return *this << xValue.GetBits();
}
Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Comp_Quat_Fast& xValue)
{
	u_int uBits;
	*this >> uBits;
	xValue.SetBits( uBits );
	return *this;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Comp_Quat_Fast& xValue)
{
	return *this << xValue.GetBits();
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_RGBA& xValue)
{
	return *this >> xValue.m_fR >> xValue.m_fG >> xValue.m_fB >> xValue.m_fA;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Quat& xValue)
{
	return *this >> xValue.x >> xValue.y >> xValue.z >> xValue.w;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Quat& xValue)
{
	return *this << xValue.x << xValue.y << xValue.z << xValue.w;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Euler_3& xValue)
{
	return *this >> xValue.x >> xValue.y >> xValue.z;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Euler_3& xValue)
{
	return *this << xValue.x << xValue.y << xValue.z;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_RGBA& xValue)
{
	return *this << xValue.m_fR << xValue.m_fG << xValue.m_fB << xValue.m_fA;
}

void Asura_Chunk_Stream::ReadString (char* szReadStringBuffer,u_int uBufferSize)
{
	uBufferSize &=~ 3;//buffer size must be a multiple of 4

	char* pcDest = szReadStringBuffer;
	
    bool bEndFound = false;
    u_int uReadStringLength = 0;
	do
	{
		Read( 4, pcDest);
        
        for( u_int u = 4; u--; ++uReadStringLength )
        {
            if( !*pcDest++ )
            {
                bEndFound = true;
                break;
            }
        }
	}
	while( !bEndFound && uReadStringLength < uBufferSize );

    // Ensure it's null terminated if the string was too long.
    if( uReadStringLength >= uBufferSize )
    {
        szReadStringBuffer[uBufferSize - 1] = '\0';
		Asura_Assert(0,"String was too long to read into buffer");

		//skip the remainder
		char szTemp[4];
		do
		{
			pcDest = szTemp;
			Read( 4, pcDest);
	        
	        for( u_int u = 4; u--;)
	        {
	            if( !*pcDest++ )
	            {
					//found terminator
					return;
	            }
	        }
		}
		while(true);
    }
}

void Asura_Chunk_Stream::ReadString (Asura_String* pxString)
{
	static const u_int READ_STRING_BUFFER_SIZE = 512; // Must be a multiple of 4.
	char szReadStringBuffer[READ_STRING_BUFFER_SIZE] = "";

	ReadString(szReadStringBuffer, READ_STRING_BUFFER_SIZE);

	//If a string pointer was given , fill it in now
	if(pxString)
	{
		//check for the null string identifier
		if(*(int*)&szReadStringBuffer[0] == Asura_String::NULL_STRING_CHUNK)
		{
			pxString->MakeNull();
		}
		else
		{
			//just assign to the Asura_String
			*pxString = szReadStringBuffer;
		}
	}
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_String& xString)
{
    if( xString.GetString() )
    {
		int iZero = 0;
		Write(xString.GetSizeInBytes(), xString.GetString());
		const int iNumPadBytes = (xString.GetSizeInInts() * 4) - xString.GetSizeInBytes();
		if(iNumPadBytes) Write(iNumPadBytes, reinterpret_cast<void*>(&iZero));
    }
	else
	{
		return *this << Asura_String::NULL_STRING_CHUNK;
	}
	return (*this);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Vector_2& xValue)
{
	return *this >> xValue.x >> xValue.y;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Vector_2& xValue)
{
	return *this << xValue.x << xValue.y;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Vector_3& xValue)
{
	return *this >> xValue.x >> xValue.y >> xValue.z;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Vector_3& xValue)
{
	return *this << xValue.x << xValue.y << xValue.z;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_AVector3& xValue)
{
	// w is ignored.
	Asura_Vector_3 xTemp;
	Asura_Chunk_Stream& xRet = *this >> xTemp.x >> xTemp.y >> xTemp.z;
	xValue = xTemp;// copy temp to vector 4. (vector 4 doesn't give access to internals)
	return xRet;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_AVector3& xValue)
{
	return *this << xValue.GetX() << xValue.GetY() << xValue.GetZ();	// << xValue.m_fW;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Matrix_3x3& xValue)
{
    Asura_ReadFromStream_Array_Wrapper<float>	xTempArray (9, &xValue.Element[0][0]);

    return *this >> xTempArray;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Matrix_3x3& xValue)
{
    return *this << Asura_WriteToStream_Array_Wrapper<float>(9, &xValue.Element[0][0]);
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_AMatrix3x3& xValue)
{
    Asura_AVector3 xTemp;
    *this >> xTemp;
    xValue.SetColumn0( xTemp );
    *this >> xTemp;
    xValue.SetColumn1( xTemp );
    *this >> xTemp;
    xValue.SetColumn2( xTemp );
    return *this;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_AMatrix3x3& xValue)
{
    *this << xValue.GetColumn0();
    *this << xValue.GetColumn1();
    *this << xValue.GetColumn2();
    return *this;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator>> (Asura_Bounding_Box& xValue)
{
	Asura_ReadFromStream_Array_Wrapper<float>	xTempArray (6, xValue.m_fElement);
	
	return *this >> xTempArray;
}

Asura_Chunk_Stream& Asura_Chunk_Stream::operator<< (const Asura_Bounding_Box& xValue)
{
	return *this << Asura_WriteToStream_Array_Wrapper<float>(6, xValue.m_fElement);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_File_Stream
///////////////////////////////////////////////////////////////////////////////////////////////////

u_int Asura_Chunk_File_Stream::Read(u_int uNumberOfBytes, void* pData)
{
	if(Asura_FileAccess::ReadData(m_iFileHandle, pData, uNumberOfBytes))
	{
		m_iChunkSize += uNumberOfBytes;
		return uNumberOfBytes;
	}
	else
	{
		// (MH - 29.02.2008) fail safely on read errors by zeroing memory.
		Asura_Mem::Set(pData, 0, uNumberOfBytes);
	}
	return 0;
}

u_int Asura_Chunk_File_Stream::Skip(u_int uNumberOfBytes)
{
	if(Asura_FileAccess::SkipData(m_iFileHandle, uNumberOfBytes))
	{
		m_iChunkSize += uNumberOfBytes;
		return uNumberOfBytes;
	}
	return 0;
}

// Cannot write to file; file export is done by writing to buffer in Asura_ProtoSaveBlock and then saving the whole lot
u_int Asura_Chunk_File_Stream::Write(u_int uNumberOfBytes, const void* pData)
{
	Asura_Assert(0, "Should not try to write to file through this interface");
	return 0;
}

u_int Asura_Chunk_File_Stream::Write(u_int uNumberOfBytes, const void* pData, u_int uPosition)
{
	Asura_Assert(0, "Should not try to write to file through this interface");
	return 0;
}

u_int Asura_Chunk_File_Stream::GetPosition() const
{
	return Asura_FileAccess::GetFilePosition(m_iFileHandle);
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_SaveBlock_Stream
///////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Chunk_SaveBlock_Stream::Asura_Chunk_SaveBlock_Stream(Asura_SaveBlock_Base* pxSaveBlock)
: m_pxSaveBlock(pxSaveBlock)
{
}

#if defined ASURA_TOOLS || !defined ASURA_MAIN
Asura_Chunk_SaveBlock_Stream::Asura_Chunk_SaveBlock_Stream(Asura_SaveBlock_Base* pxSaveBlock, const bool bSwapEndian)
: m_pxSaveBlock(pxSaveBlock)
{
	SetSwapEndian(bSwapEndian);
}
#endif	//defined ASURA_TOOLS || !defined ASURA_MAIN

u_int Asura_Chunk_SaveBlock_Stream::Read(u_int uNumberOfBytes, void* pData)
{
	Asura_Assert(false, "reading from saveblock stream not supported - use Asura_Chunk_SaveBlockReader_Stream");
	return 0;
}

u_int Asura_Chunk_SaveBlock_Stream::Skip(u_int uNumberOfBytes)
{
	Asura_Assert(false, "reading from saveblock stream not supported - use Asura_Chunk_SaveBlockReader_Stream");
	return 0;
}

u_int Asura_Chunk_SaveBlock_Stream::Write(u_int uNumberOfBytes, const void* pData)
{
	Asura_Assert(m_pxSaveBlock, "No save block");
	if(m_pxSaveBlock)
	{
		m_pxSaveBlock->AddData(pData, uNumberOfBytes);
		m_iChunkSize += uNumberOfBytes;
		return uNumberOfBytes;
	}
	return 0;
}

u_int Asura_Chunk_SaveBlock_Stream::Write(u_int uNumberOfBytes, const void* pData, u_int uPosition)
{
	Asura_Assert(uNumberOfBytes == 4, "MH assumption - this method should only ever be used to modify u_int chunk size"	);
	Asura_Assert(m_pxSaveBlock, "No save block");
	if(m_pxSaveBlock && m_pxSaveBlock->AddPatchData(pData, uNumberOfBytes, uPosition))
	{
		return sizeof(u_int);
	}
	return 0;
}

u_int Asura_Chunk_SaveBlock_Stream::GetPosition() const
{
	return m_pxSaveBlock->GetBytesWritten();
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_SaveBlockReader_Stream
///////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Chunk_SaveBlockReader_Stream::Asura_Chunk_SaveBlockReader_Stream(const Asura_SaveBlock_Base* pxSaveBlock)
: m_pxSaveBlock(pxSaveBlock)
{
	m_pcPeekedDataPtr = NULL;
	m_uPeekedBytesRemaining = 0;

	pxSaveBlock->RestartRead();
}

u_int Asura_Chunk_SaveBlockReader_Stream::Read(u_int uNumberOfBytes, void* pData)
{
	u_int uBytesRead = 0;

	// use up any peeked data we may have
	if (m_uPeekedBytesRemaining)
	{
		u_int uPeekedBytesToUse = Asura_Maths::Min(uNumberOfBytes, m_uPeekedBytesRemaining);
		m_uPeekedBytesRemaining -= uPeekedBytesToUse;
		uNumberOfBytes -= uPeekedBytesToUse;
		uBytesRead = uPeekedBytesToUse;
		Asura_Mem::Copy(pData, m_pcPeekedDataPtr, uPeekedBytesToUse);
		pData = reinterpret_cast<char*>(pData) + uPeekedBytesToUse;
		m_pcPeekedDataPtr += uPeekedBytesToUse;
	}

	// read any remainder from stream as normal.
	uBytesRead += m_pxSaveBlock->ReadData(pData, uNumberOfBytes);

	return uBytesRead;
}

int Asura_Chunk_SaveBlockReader_Stream::PeekInt()
{
	// if we're peeking again at data we already peeked, just copy that
	int iValue;
	if (m_uPeekedBytesRemaining)
	{
		Asura_Assert(m_uPeekedBytesRemaining == 4, "should be reading 4 bytes");
		Asura_Mem::Copy(&iValue, m_pcPeekedDataPtr, 4);
	}
	else
	{
		// read the data as normal
		#if defined ASURA_ASSERT
			const u_int uBytesRead =
		#endif
			m_pxSaveBlock->ReadData(&iValue, 4);

		Asura_Assert(uBytesRead == 4, "failed to read int");

		// copy to peek cache
		Asura_Mem::Copy(m_acPeekedBytes, &iValue, 4);
		m_pcPeekedDataPtr = m_acPeekedBytes;
		m_uPeekedBytesRemaining = 4;
	}

	return iValue;
}

u_int Asura_Chunk_SaveBlockReader_Stream::Skip(u_int uNumberOfBytes)
{
	Asura_Assert(!m_uPeekedBytesRemaining, "CHUNK ERROR: skipping peeked bytes not implemented");
	return m_pxSaveBlock->SkipData(uNumberOfBytes);
}

u_int Asura_Chunk_SaveBlockReader_Stream::Write(u_int uNumberOfBytes, const void* pData)
{
	Asura_Assert(false, "writing to saveblock reader stream not supported - use Asura_Chunk_SaveBlock_Stream");
	return 0;
}

u_int Asura_Chunk_SaveBlockReader_Stream::Write(u_int uNumberOfBytes, const void* pData, u_int uPosition)
{
	Asura_Assert(false, "writing to saveblock reader stream not supported - use Asura_Chunk_SaveBlock_Stream");
	return 0;
}

u_int Asura_Chunk_SaveBlockReader_Stream::GetPosition() const
{
	Asura_Assert(m_pxSaveBlock->GetBytesRead() >= m_uPeekedBytesRemaining, "peeked more bytes than were read from save block");
	// don't include peeked bytes in where we have got to.
	return m_pxSaveBlock->GetBytesRead() - m_uPeekedBytesRemaining;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_Memory_Stream
///////////////////////////////////////////////////////////////////////////////////////////////////

u_int Asura_Chunk_Memory_Stream::Read (u_int uNumberOfBytes, void* pData)
{
	const u_char*	pucStart	= &m_pucStartOfBlock [m_uCurrentPosition];

	// Do we have enough room?
	if ((uNumberOfBytes + m_uCurrentPosition) > m_uSizeOfBlock)
	{
		// No - report an error and don't write anything...
		Asura_Assert	(false, "Not enough data to be read.");
		return			(0);
	}

	else
	{
		// Yes - copy the data and update the pointers...
		Asura_Mem::SafeCopy	(pData, pucStart, uNumberOfBytes);

		m_uCurrentPosition	+= uNumberOfBytes;
		m_iChunkSize		+= uNumberOfBytes;

		// Return the number of bytes written...
		return	(uNumberOfBytes);
	}
}

u_int Asura_Chunk_Memory_Stream::Skip (u_int uNumberOfBytes)
{
	// Do we have enough room?
	if ((uNumberOfBytes + m_uCurrentPosition) > m_uSizeOfBlock)
	{
		// No - report an error and don't write anything...
		Asura_Assert	(false, "Not enough data to be skipped.");
		return			(0);
	}

	else
	{
		// Yes - update the pointers...
		m_uCurrentPosition	+= uNumberOfBytes;
		m_iChunkSize		+= uNumberOfBytes;

		// Return the number of bytes written...
		return	(uNumberOfBytes);
	}
}

u_int Asura_Chunk_Memory_Stream::Write (u_int uNumberOfBytes, const void* pData)
{
	if(m_bReadOnly)
	{
		Asura_Assert(0,"Attempting to write to a read-only stream");
		return 0;
	}

	// Do we have enough room?
	if ((uNumberOfBytes + m_uCurrentPosition) > m_uSizeOfBlock)
	{
		// not enough space - can we grow the buffer to fit?
		if ( m_bOwnsBuffer )
		{
			// pass on an estimate of the minimum new space we need
			GrowBuffer( uNumberOfBytes + m_uCurrentPosition - m_uSizeOfBlock );
		}
		else
		{
			// No - report an error and don't write anything...

			Asura_Assert	(false, "Not enough room to write data.");
			return			(0);
		}
	}
	
	{
		u_char*	pucStart	= &m_pucStartOfBlock [m_uCurrentPosition];

		// Yes - copy the data and update the pointers...
		Asura_Mem::Copy	(pucStart, pData, uNumberOfBytes);

		m_uCurrentPosition	+= uNumberOfBytes;
		m_iChunkSize		+= uNumberOfBytes;

		// Return the number of bytes written...
		return	(uNumberOfBytes);
	}
}

u_int Asura_Chunk_Memory_Stream::Write (u_int uNumberOfBytes, const void* pData, u_int uPosition)
{
	if(m_bReadOnly)
	{
		Asura_Assert(0,"Attempting to write to a read-only stream");
		return 0;
	}


	// Do we have enough room?
	if ( (uNumberOfBytes + uPosition) > m_uSizeOfBlock )
	{
		// not enough space - can we grow the buffer to fit?
		if ( m_bOwnsBuffer )
		{
			// pass on an estimate of the minimum new space we need
			GrowBuffer( uNumberOfBytes + uPosition - m_uSizeOfBlock );
		}
		else
		{
			// No - report an error and don't write anything...

			Asura_Assert	(false, "Not enough room to write data.");
			return			(0);
		}
	}

	{
		// Yes - copy the data and return the number of bytes written...

		u_char*	pucStart = &m_pucStartOfBlock [uPosition];
		Asura_Mem::Copy	(pucStart, pData, uNumberOfBytes);
		return	(uNumberOfBytes);
	}
}

void Asura_Chunk_Memory_Stream::BeginPeek()
{
	Asura_Assert(!m_bPeekEnabled, "peek already enabled");
	m_bPeekEnabled = true;

	m_uPrePeekCurrentPosition = m_uCurrentPosition;
	m_iPrePeekChunkSize = m_iChunkSize;
}

void Asura_Chunk_Memory_Stream::EndPeek()
{
	Asura_Assert(m_bPeekEnabled, "peek was not already enabled");
	m_bPeekEnabled = false;

	m_uCurrentPosition = m_uPrePeekCurrentPosition;
	m_iChunkSize = m_iPrePeekChunkSize;
}

u_int Asura_Chunk_Memory_Stream::GetPosition (void) const
{
	return	(m_uCurrentPosition);
}

int Asura_Chunk_Memory_Stream::PeekInt()
{
	BeginPeek();
	int iValue;
	*this >> iValue;
	EndPeek();
	return iValue;
}

void Asura_Chunk_Memory_Stream::SetPosition(u_int uPosition)
{ 
	Asura_Assert((uPosition < m_uSizeOfBlock), "stream position out of range");
	if (uPosition < m_uSizeOfBlock)
	{
		m_uCurrentPosition = uPosition; 
	}
}

void Asura_Chunk_Memory_Stream::GrowBuffer( u_int uMinimumRequired )
{
	Asura_Assert( m_bOwnsBuffer == true, "Code bug: memory stream shouldn't be trying to change buffer size" );

	Asura_Assert( m_bReadOnly == false, "Code bug: memory stream shouldn't be trying to change buffer size" );

	Asura_Assert( m_uSizeOfBlock, "Code bug: block size is zero!" );

	// keep doubling the size until we fit
	u_int uSizeIncrease = m_uSizeOfBlock;

	while ( uSizeIncrease <= uMinimumRequired )
	{
		uSizeIncrease*=2;
	}

	const u_int uNewSizeOfBlock = uSizeIncrease + m_uSizeOfBlock;

	u_char* pucNewBuffer = new u_char [uNewSizeOfBlock];

	// copy the original data into it
	if(pucNewBuffer == NULL)
	{
		Asura_Assertf(false, ("Asura_Chunk_Memory_Stream::GrowBuffer(): Failed to allocate %u bytes of memory - CRASH IMMINENT!!!", uNewSizeOfBlock));
	}
	else
	{
		Asura_Mem::Copy( pucNewBuffer, m_pucStartOfBlock, m_uSizeOfBlock );
	}

	// delete the original buffer, and update the member vars to know about the new buffer instead	
	delete [] m_pucStartOfBlock;
	m_pucStartOfBlock = pucNewBuffer;
	m_uSizeOfBlock = uNewSizeOfBlock;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Asura_Chunk_ChunkBuffer_Stream
///////////////////////////////////////////////////////////////////////////////////////////////////

u_int Asura_Chunk_ChunkBuffer_Stream::Read(u_int uNumberOfBytes, void* pData)
{
    // copy the data and update the pointers...
    Asura_Mem::Copy	(pData, m_pcMovingStartOfBlock, uNumberOfBytes);

    m_pcMovingStartOfBlock += uNumberOfBytes;
    m_uCurrentPosition	   += uNumberOfBytes;
    m_iChunkSize		   += uNumberOfBytes;

    // Return the number of bytes written...
    return	(uNumberOfBytes);
}

u_int Asura_Chunk_ChunkBuffer_Stream::Skip(u_int uNumberOfBytes)
{
    // update the pointers...
    m_pcMovingStartOfBlock += uNumberOfBytes;
    m_uCurrentPosition	   += uNumberOfBytes;
    m_iChunkSize		   += uNumberOfBytes;

    // Return the number of bytes written...
    return	(uNumberOfBytes);
}

u_int Asura_Chunk_ChunkBuffer_Stream::Write(u_int uNumberOfBytes, const void* pData)
{
    if(m_bReadOnly)
    {
        Asura_Assert(false,"Asura_Chunk_ChunkBuffer_Stream::Write - Attempting to write to a read-only stream");
        return 0;
    }

    // copy the data and update the pointers...
    Asura_Mem::Copy	(m_pcMovingStartOfBlock, pData, uNumberOfBytes);

    m_pcMovingStartOfBlock += uNumberOfBytes;
    m_uCurrentPosition	   += uNumberOfBytes;
    m_iChunkSize		   += uNumberOfBytes;

    // Return the number of bytes written...
    return	(uNumberOfBytes);
}

u_int Asura_Chunk_ChunkBuffer_Stream::Write (u_int uNumberOfBytes, const void* pData, u_int uPosition)
{
    if(m_bReadOnly)
    {
        Asura_Assert(false,"Asura_Chunk_ChunkBuffer_Stream::Write - Attempting to write to a read-only stream");
        return 0;
    }

    if ( uPosition > m_uCurrentPosition )
    {
        Asura_Assert( false, "Asura_Chunk_ChunkBuffer_Stream::Write - attempting to write beyond our buffer" );
        return 0;
    }

    // copy the data...
    Asura_Mem::Copy	(m_pcStartOfBlock + uPosition, pData, uNumberOfBytes);

    // Return the number of bytes written...
    return	(uNumberOfBytes);
}

int Asura_Chunk_ChunkBuffer_Stream::PeekInt()
{
	BeginPeek();
	int iValue;
	*this >> iValue;
	EndPeek();
	return iValue;
}

void Asura_Chunk_ChunkBuffer_Stream::BeginPeek()
{
	Asura_Assert(!m_bPeekEnabled, "peek already enabled");
	m_bPeekEnabled = true;

	m_pcPrePeekMovingStartOfBlock = m_pcMovingStartOfBlock;
	m_uPrePeekCurrentPosition = m_uCurrentPosition;
	m_iPrePeekChunkSize = m_iChunkSize;
}

void Asura_Chunk_ChunkBuffer_Stream::EndPeek()
{
	Asura_Assert(m_bPeekEnabled, "peek was not already enabled");
	m_bPeekEnabled = false;

	m_pcMovingStartOfBlock = m_pcPrePeekMovingStartOfBlock;
	m_uCurrentPosition = m_uPrePeekCurrentPosition;
	m_iChunkSize = m_iPrePeekChunkSize;
}


//////////////////////////////////////////////////////////////////////////
// Asura_Chunk_Dummy_Stream
//////////////////////////////////////////////////////////////////////////
int Asura_Chunk_Dummy_Stream::PeekInt()
{
	BeginPeek();
	int iValue;
	*this >> iValue;
	EndPeek();
	return iValue;
}

// KOR - temporary testing framework for moving from chunkbuffers to chunk streams
#ifdef TEST_ENTITY_CHUNK_STREAM
#define TestStreamAssert(__Condition, __Error) \
	Asura_Assertf(__Condition, ("******************************\n*** CRITICAL ERROR ***\n******************************\n" \
	"Please report this assert to kim.randell@rebellion.co.uk\n" \
	"Entity classification: %s\n" __Error, Asura_Network::GetEntityClassificationDesc(pxEntity->GetClassification()) ) ); 

void TestStreamWrite(Asura_Entity_Server* pxEntity, const char* pcChunkBufferStart, const u_int uChunkSize)
{
	// Create a double size buffer just in case.
	Asura_Chunk_Memory_Stream xWriteStream(uChunkSize * 2);
	xWriteStream.SetSwapEndian(false);
	const char* pcStreamBuffer = reinterpret_cast<const char*>(xWriteStream.GetStartAddress());
	// Test write
	pxEntity->WriteToChunkStream(xWriteStream);
	TestStreamAssert(xWriteStream.GetPosition() == uChunkSize, "Stream write did not write number of bytes returned by GetChunkBufferSize");
	for(u_int u = 0; u < uChunkSize; ++u)
	{
		TestStreamAssert(pcChunkBufferStart[u] == pcStreamBuffer[u], "Stream write did not match original chunk buffer");
	}
}
#endif

// eof
