//
// Asura_Chunk_Texture.cpp
//
// Created by Kevin Lea - 12/06/2001 11:20:35
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "ChunkSystem/Asura_Chunk_Texture.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Macros.h"
#endif
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_NameTree.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Memory/Asura_Mem.h"
#include "FileAccess/Asura_FileAccess.h"

#define ASURA_CHUNK_TEXTURENAMES_VERSION 3

bool Asura_Chunk_TextureNames::Process(Asura_Handle fileHandle) const
{
	Asura_Assert(Version<=ASURA_CHUNK_TEXTURENAMES_VERSION,"Unknown chunk version");

	Asura_Texture_Management::NewConversionList(NumberOfTextures);

	
	if(Version<3)
	{
		//create conversion list for materials as well
		Asura_Material_System::NewConversionList(NumberOfTextures);
	}
	

	for(int i=0; i<NumberOfTextures; i++)
	{
		const char* szTextureName = Asura_FileAccess::ReadString( fileHandle );

#if defined ASURA_AUTO_REPLACE_FORWARD_SLASH_IN_TEXTURE_NAMES
		// JAT 080206 - need to change '/' to '\' in the texture names to get PS2 to work
		char* szPtr = const_cast<char*>(szTextureName);
		Asura_StringUtil::ReplaceAllChars( szPtr, '/', '\\' );
#endif // #if defined ASURA_AUTO_FIX_FORWARD_SLASH_IN_TEXTURE_NAMES

		int iTextureID = Asura_Texture_Management::AddTexture( szTextureName, i );
		if(Version<3)
		{
			//for earlier files , a default material needs to be created
			int iMaterialID = Asura_Material_System::CreateMaterial(i);
			if(iMaterialID != ASURA_MATERIAL_NONE)
			{
				Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(iMaterialID);
				Asura_Assert(pxMaterial,"");

				pxMaterial->m_iTextureID[0] = iTextureID;
			}
		}
	
		Asura_ChunkLoading_System::YieldBackgroundProcessing();
	}
	return true;
}

#ifdef ASURA_TOOLS
bool Asura_Chunk_TextureNames::Process(const char* pcChunkBuffer) const
{
	Asura_Assert(Version<=ASURA_CHUNK_TEXTURENAMES_VERSION,"Unknown chunk version");

	Asura_Texture_Management::NewConversionList(NumberOfTextures);
	
	

	for(int i=0; i<NumberOfTextures; i++)
	{
		Asura_Texture_Management::AddTexture( pcChunkBuffer, i );
		pcChunkBuffer += (strlen(pcChunkBuffer)+4) & ~3;

	}
	return true;
}
#endif

#ifdef ASURA_TOOLS
void Asura_Chunk_TextureNames::Create( const Asura_NameTree* pxNameTree, Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(pxNameTree,xStream);
}

void Asura_Chunk_TextureNames::Create( const Asura_NameTree* pxNameTree, Asura_Chunk_Stream& xStream)
{

	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_TEXTURENAMES, ASURA_CHUNK_TEXTURENAMES_VERSION, uChunkFlags);

	xStream << pxNameTree->GetNumberOfNodes();

	const Asura_String** ppxNameList = new const Asura_String*[pxNameTree->GetNumberOfNodes()];
	pxNameTree->MakeNameList(ppxNameList);

	for (int i=0; i<pxNameTree->GetNumberOfNodes(); i++)
	{
		const Asura_String* pxName = ppxNameList[i];
		xStream << *pxName;
	}
	
	delete [] ppxNameList;

	xStream.WriteChunk_End();
}

#endif

#define ASURA_CHUNK_TEXTUREFLAGS_VERSION 1

bool Asura_Chunk_TextureFlags::Process(Asura_Handle fileHandle) const
{
	Asura_Chunk_File_Stream xStream(fileHandle);
	return Process(xStream);
}

bool Asura_Chunk_TextureFlags::Process(Asura_Chunk_Stream& xStream) const
{
	Asura_Assert(Version<=ASURA_CHUNK_TEXTUREFLAGS_VERSION,"Unknown chunk version");

	for(int i=0; i<NumberOfTextures; i++)
	{
		//read in the flags
		int iFlags;
		xStream >> iFlags;
		if(iFlags)
		{
			if(Version==0)
			{
				//need to copy off some of the flags into the material flags instead
				int iMaterialID = Asura_Material_System::ConvertIndex(i);
				Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(iMaterialID);

				if(pxMaterial) pxMaterial->m_uMaterialFlags |= iFlags & AsuraMaterialFlagsToCopyFromTextureFlags;
				iFlags &=~ AsuraMaterialFlagsToCopyFromTextureFlags;
			}

			#ifdef ASURA_TOOLS
			//Clear the texture usage flags , since these will be generated as the materials are loaded. RAR 08/05/2006
			iFlags &=~ ASURA_TEXTURE_USAGE_FLAGS;
			#endif
			
			int iNewIndex = Asura_Texture_Management::ConvertIndex(i);
			//combine with any existing flags.
			iFlags |= Asura_Texture_Management::GetTextureFlags(iNewIndex);
			Asura_Texture_Management::SetTextureFlags(iNewIndex,iFlags);
		}
	}
	return true;
}

#ifdef ASURA_TOOLS
void Asura_Chunk_TextureFlags::Create(int iNumberOfTextures, const int* piTextureFlags,Asura_SaveBlock* pxSaveBlock)
{
	Asura_Chunk_SaveBlock_Stream xStream(pxSaveBlock);
	Create	(iNumberOfTextures, piTextureFlags, xStream);
}

void Asura_Chunk_TextureFlags::Create(int iNumberOfTextures, const int* piTextureFlags, Asura_Chunk_Stream& xStream)
{
		
	u_int uChunkFlags = 0;
	xStream.WriteChunk_Start(ASURA_CHUNK_TEXTUREFLAGS, ASURA_CHUNK_TEXTUREFLAGS_VERSION, uChunkFlags);


	xStream << iNumberOfTextures;
	
	for(int i=0; i<iNumberOfTextures; i++)
	{
		xStream << piTextureFlags[i];
	}
	xStream.WriteChunk_End();

}
#endif
