/* MD5C.C - RSA Data Security, Inc., MD5 message-digest algorithm
 */

/* Copyright (C) 1991-2, RSA Data Security, Inc. Created 1991. All
rights reserved.

License to copy and use this software is granted provided that it
is identified as the "RSA Data Security, Inc. MD5 Message-Digest
Algorithm" in all material mentioning or referencing this software
or this function.

License is also granted to make and use derivative works provided
that such works are identified as "derived from the RSA Data
Security, Inc. MD5 Message-Digest Algorithm" in all material
mentioning or referencing the derived work.

RSA Data Security, Inc. makes no representations concerning either
the merchantability of this software or the suitability of this
software for any particular purpose. It is provided "as is"
without express or implied warranty of any kind.

These notices must be retained in any copies of any part of this
documentation and/or software.
 */

#include "Core/Asura.h"
#include "Core/Asura_Checksum.h"
#include "Core/Memory/Asura_Mem.h"

#include <string.h>


/* Constants for MD5Transform routine.
 */

#define S11 7
#define S12 12
#define S13 17
#define S14 22
#define S21 5
#define S22 9
#define S23 14
#define S24 20
#define S31 4
#define S32 11
#define S33 16
#define S34 23
#define S41 6
#define S42 10
#define S43 15
#define S44 21

#define MD5_memcpy Asura_Mem::Copy
#define MD5_memset memset

static u_char PADDING[64] = 
{
  0x80, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

/* F, G, H and I are basic MD5 functions.
 */
#define F(x, y, z) (((x) & (y)) | ((~x) & (z)))
#define G(x, y, z) (((x) & (z)) | ((y) & (~z)))
#define H(x, y, z) ((x) ^ (y) ^ (z))
#define I(x, y, z) ((y) ^ ((x) | (~z)))

/* ROTATE_LEFT rotates x left n bits.
 */
#define ROTATE_LEFT(x, n) (((x) << (n)) | ((x) >> (32-(n))))

/* FF, GG, HH, and II transformations for rounds 1, 2, 3, and 4.
Rotation is separate from addition to prevent recomputation.
 */
#define FF(a, b, c, d, x, s, ac) \
{ \
    (a) += F ((b), (c), (d)) + (x) + (u_int)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}

#define GG(a, b, c, d, x, s, ac) \
{ \
    (a) += G ((b), (c), (d)) + (x) + (u_int)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}

#define HH(a, b, c, d, x, s, ac) \
{ \
    (a) += H ((b), (c), (d)) + (x) + (u_int)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}
#define II(a, b, c, d, x, s, ac) \
{ \
    (a) += I ((b), (c), (d)) + (x) + (u_int)(ac); \
    (a) = ROTATE_LEFT ((a), (s)); \
    (a) += (b); \
}

/* MD5 initialization. Begins an MD5 operation, writing a new context.
 */
void Asura_Checksum::Initialise(void)
{
    m_xContext.uCount[0] = 0;
    m_xContext.uCount[1] = 0;

    // Load magic initialization constants.
    m_xContext.uState[0] = 0x67452301;
    m_xContext.uState[1] = 0xefcdab89;
    m_xContext.uState[2] = 0x98badcfe;
    m_xContext.uState[3] = 0x10325476;
}

/* MD5 block update operation. Continues an MD5 message-digest
  operation, processing another message block, and updating the
  context.
 */
void Asura_Checksum::Update(const u_char* pucInputBuffer, u_int uInputLen)
{
    u_int i, uIndex, uPartLen;

    /* Compute number of bytes mod 64 */
    uIndex = (u_int)((m_xContext.uCount[0] >> 3) & 0x3F);

    /* Update number of bits */
    if ((m_xContext.uCount[0] += (uInputLen << 3)) < (uInputLen << 3))
    {
        m_xContext.uCount[1]++;
    }
    m_xContext.uCount[1] += (uInputLen >> 29);

    uPartLen = 64 - uIndex;

    /* Transform as many times as possible.
    */
    if (uInputLen >= uPartLen) 
    {
        Asura_Mem::Copy(&m_xContext.ucBuffer[uIndex], pucInputBuffer, uPartLen);
        Transform(m_xContext.uState, m_xContext.ucBuffer);

        for (i = uPartLen; i + 63 < uInputLen; i += 64)
        {
            Transform(m_xContext.uState, &pucInputBuffer[i]);
        }

        uIndex = 0;
    }
    else
    {
        i = 0;
    }

    /* Buffer remaining input */
    Asura_Mem::Copy(&m_xContext.ucBuffer[uIndex], &pucInputBuffer[i], uInputLen-i);
}

/* MD5 finalization. Ends an MD5 message-digest operation, writing the
  the message digest and zeroizing the context.
 */

void Asura_Checksum::Finalise(u_char aucOutput[CHECKSUM_LENGTH])
{
    u_char ucBits[8];
    u_int  uIndex, uPadLen;

    /* Save number of bits */
    Encode(ucBits, m_xContext.uCount, 8);

    /* Pad out to 56 mod 64.
    */
    uIndex  = (m_xContext.uCount[0] >> 3) & 0x3f;
    uPadLen = (uIndex < 56) ? (56 - uIndex) : (120 - uIndex);
    Update(PADDING, uPadLen);

    /* Append length (before padding) */
    Update(ucBits, 8);
    /* Store state in digest */
    Encode(aucOutput, m_xContext.uState, CHECKSUM_LENGTH);

    /* Zeroize sensitive information.
    */
    memset(reinterpret_cast<char*>(&m_xContext), 0, sizeof(m_xContext));
}

/* MD5 basic transformation. Transforms state based on block.
 */
void Asura_Checksum::Transform(u_int auState[4], const u_char* ucBlock)
{
    u_int a = auState[0];
    u_int b = auState[1];
    u_int c = auState[2]; 
    u_int d = auState[3];
    u_int x[16];

    Decode (x, ucBlock, 64);

    /* Round 1 */
    FF (a, b, c, d, x[ 0], S11, 0xd76aa478); /* 1 */
    FF (d, a, b, c, x[ 1], S12, 0xe8c7b756); /* 2 */
    FF (c, d, a, b, x[ 2], S13, 0x242070db); /* 3 */
    FF (b, c, d, a, x[ 3], S14, 0xc1bdceee); /* 4 */
    FF (a, b, c, d, x[ 4], S11, 0xf57c0faf); /* 5 */
    FF (d, a, b, c, x[ 5], S12, 0x4787c62a); /* 6 */
    FF (c, d, a, b, x[ 6], S13, 0xa8304613); /* 7 */
    FF (b, c, d, a, x[ 7], S14, 0xfd469501); /* 8 */
    FF (a, b, c, d, x[ 8], S11, 0x698098d8); /* 9 */
    FF (d, a, b, c, x[ 9], S12, 0x8b44f7af); /* 10 */
    FF (c, d, a, b, x[10], S13, 0xffff5bb1); /* 11 */
    FF (b, c, d, a, x[11], S14, 0x895cd7be); /* 12 */
    FF (a, b, c, d, x[12], S11, 0x6b901122); /* 13 */
    FF (d, a, b, c, x[13], S12, 0xfd987193); /* 14 */
    FF (c, d, a, b, x[14], S13, 0xa679438e); /* 15 */
    FF (b, c, d, a, x[15], S14, 0x49b40821); /* 16 */

    /* Round 2 */
    GG (a, b, c, d, x[ 1], S21, 0xf61e2562); /* 17 */
    GG (d, a, b, c, x[ 6], S22, 0xc040b340); /* 18 */
    GG (c, d, a, b, x[11], S23, 0x265e5a51); /* 19 */
    GG (b, c, d, a, x[ 0], S24, 0xe9b6c7aa); /* 20 */
    GG (a, b, c, d, x[ 5], S21, 0xd62f105d); /* 21 */
    GG (d, a, b, c, x[10], S22,  0x2441453); /* 22 */
    GG (c, d, a, b, x[15], S23, 0xd8a1e681); /* 23 */
    GG (b, c, d, a, x[ 4], S24, 0xe7d3fbc8); /* 24 */
    GG (a, b, c, d, x[ 9], S21, 0x21e1cde6); /* 25 */
    GG (d, a, b, c, x[14], S22, 0xc33707d6); /* 26 */
    GG (c, d, a, b, x[ 3], S23, 0xf4d50d87); /* 27 */
    GG (b, c, d, a, x[ 8], S24, 0x455a14ed); /* 28 */
    GG (a, b, c, d, x[13], S21, 0xa9e3e905); /* 29 */
    GG (d, a, b, c, x[ 2], S22, 0xfcefa3f8); /* 30 */
    GG (c, d, a, b, x[ 7], S23, 0x676f02d9); /* 31 */
    GG (b, c, d, a, x[12], S24, 0x8d2a4c8a); /* 32 */

    /* Round 3 */
    HH (a, b, c, d, x[ 5], S31, 0xfffa3942); /* 33 */
    HH (d, a, b, c, x[ 8], S32, 0x8771f681); /* 34 */
    HH (c, d, a, b, x[11], S33, 0x6d9d6122); /* 35 */
    HH (b, c, d, a, x[14], S34, 0xfde5380c); /* 36 */
    HH (a, b, c, d, x[ 1], S31, 0xa4beea44); /* 37 */
    HH (d, a, b, c, x[ 4], S32, 0x4bdecfa9); /* 38 */
    HH (c, d, a, b, x[ 7], S33, 0xf6bb4b60); /* 39 */
    HH (b, c, d, a, x[10], S34, 0xbebfbc70); /* 40 */
    HH (a, b, c, d, x[13], S31, 0x289b7ec6); /* 41 */
    HH (d, a, b, c, x[ 0], S32, 0xeaa127fa); /* 42 */
    HH (c, d, a, b, x[ 3], S33, 0xd4ef3085); /* 43 */
    HH (b, c, d, a, x[ 6], S34,  0x4881d05); /* 44 */
    HH (a, b, c, d, x[ 9], S31, 0xd9d4d039); /* 45 */
    HH (d, a, b, c, x[12], S32, 0xe6db99e5); /* 46 */
    HH (c, d, a, b, x[15], S33, 0x1fa27cf8); /* 47 */
    HH (b, c, d, a, x[ 2], S34, 0xc4ac5665); /* 48 */

    /* Round 4 */
    II (a, b, c, d, x[ 0], S41, 0xf4292244); /* 49 */
    II (d, a, b, c, x[ 7], S42, 0x432aff97); /* 50 */
    II (c, d, a, b, x[14], S43, 0xab9423a7); /* 51 */
    II (b, c, d, a, x[ 5], S44, 0xfc93a039); /* 52 */
    II (a, b, c, d, x[12], S41, 0x655b59c3); /* 53 */
    II (d, a, b, c, x[ 3], S42, 0x8f0ccc92); /* 54 */
    II (c, d, a, b, x[10], S43, 0xffeff47d); /* 55 */
    II (b, c, d, a, x[ 1], S44, 0x85845dd1); /* 56 */
    II (a, b, c, d, x[ 8], S41, 0x6fa87e4f); /* 57 */
    II (d, a, b, c, x[15], S42, 0xfe2ce6e0); /* 58 */
    II (c, d, a, b, x[ 6], S43, 0xa3014314); /* 59 */
    II (b, c, d, a, x[13], S44, 0x4e0811a1); /* 60 */
    II (a, b, c, d, x[ 4], S41, 0xf7537e82); /* 61 */
    II (d, a, b, c, x[11], S42, 0xbd3af235); /* 62 */
    II (c, d, a, b, x[ 2], S43, 0x2ad7d2bb); /* 63 */
    II (b, c, d, a, x[ 9], S44, 0xeb86d391); /* 64 */

    auState[0] += a;
    auState[1] += b;
    auState[2] += c;
    auState[3] += d;

    /* Zeroize sensitive information.
    */
    memset(reinterpret_cast<u_char*>(x), 0, sizeof(x));
}

/* Encodes input (UINT4) into output (unsigned char). Assumes len is
  a multiple of 4.
 */

void Asura_Checksum::Encode(u_char* pucOutput, const u_int* puInput, u_int uLength)
{
    u_int i, j;

    for (i = 0, j = 0; j < uLength; i++, j += 4) 
    {
        pucOutput[j]   = static_cast<u_char>( puInput[i] & 0xff);
        pucOutput[j+1] = static_cast<u_char>((puInput[i] >> 8) & 0xff);
        pucOutput[j+2] = static_cast<u_char>((puInput[i] >> 16) & 0xff);
        pucOutput[j+3] = static_cast<u_char>((puInput[i] >> 24) & 0xff);
    }
}

/* Decodes input (unsigned char) into output (UINT4). Assumes len is
  a multiple of 4.
 */
void Asura_Checksum::Decode(u_int* puOutput, const u_char* pucInput, u_int uLength)
{
    u_int i, j;

    for (i = 0, j = 0; j < uLength; i++, j += 4)
    {
        puOutput[i] = (pucInput[j]) | 
                       ((pucInput[j+1]) << 8) |
                        ((pucInput[j+2]) << 16) | 
                         ((pucInput[j+3]) << 24);
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_Checksum::EncodeFromBuffer().
// PARAMETERS   pucBuffer                  - 
//              uDataLength                - 
//              aucOutput[CHECKSUM_LENGTH] - 
// RETURNS      Nothing.
// DESCRIPTION  static function acting as a standalone wrapper for one-off checksums
// AUTHOR       Mike Healey - 22.09.2003.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_Checksum::EncodeFromBuffer(const u_char* pucBuffer, u_int uDataLength, u_char aucOutput[CHECKSUM_LENGTH])
{
	Asura_Checksum xTempChecksumContext;

	//initialise the context
	xTempChecksumContext.Initialise();

	//perform the encoding
	xTempChecksumContext.Update(pucBuffer, uDataLength);

	//and extract the result
	xTempChecksumContext.Finalise(aucOutput);
}
// End function Asura_Checksum::EncodeFromBuffer().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_Checksum::Compare().
// PARAMETERS   aucChecksum1[CHECKSUM_LENGTH] - 
//              aucChecksum2[CHECKSUM_LENGTH] - 
// RETURNS      
// DESCRIPTION  static function to compare checksums
// AUTHOR       Mike Healey - 22.09.2003.
//
//////////////////////////////////////////////////////////////////////////////
bool Asura_Checksum::Compare( const u_char aucChecksum1[CHECKSUM_LENGTH], const u_char aucChecksum2[CHECKSUM_LENGTH] )
{
    return( !memcmp(aucChecksum1, aucChecksum2, CHECKSUM_LENGTH) );
}
// End function Asura_Checksum::Compare().
