//
// Asura_DynamicBVH.h
//
// Created by KJFL - 2 October 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_DYNAMICBVH_H__
#define __ASURA_DYNAMICBVH_H__

#include "Core/Asura_Classified_List.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Maths.h"

// node for the BVH tree
template<class T> class Asura_DynamicBVH_Node
{
	public: 
		// 6 floats + u_ints + pointer = 32 bytes... *IF* a pointer is 32bits.

		Asura_Bounding_Box		m_xAABB;
		u_int					m_uLeftChild;
		T*						m_pxLeafObject;
};

// this leaf information is only needed during creation
template <class T> class Asura_DynamicBVH_ConstructionLeaf
{
	public:
		
		Asura_Bounding_Box		m_xAABB;
		Asura_Vector_3			m_xCentre;
		T*						m_pxLeafObject;
};

// element of the construction stack used when making the tree
template <class T> class Asura_DynamicBVH_ConstructionStack
{
	public:

		Asura_DynamicBVH_Node<T>*	m_pxNode;
		u_int						m_uStartIndex;
		u_int						m_uEndIndex;
};

/*
// Example functor for LOS tests with multiple results returned.
template <class T> class Asura_DynamicBVH_Functor
{
public:
	// Check for a LOS hit on the leaf object.
	// If there is a hit then update the output data and maximum distance and return true.
	// If there is no hit then just return false and leave the output data and maximum distance unchanged.
	bool CheckEntity(T* pxLeafObject, const Asura_LOS_InputData& xInputData);
	
	// Return the initial maximum distance.
	float GetInitialMaxDistance();

	// Get the current maximum distance.
	float GetMaxDistance();

	// Return true to stop the search (typically happens when testing
	// for multiple hits and the output array is too full to continue).
	bool ShouldStop();
};
*/

template< class T >
class Asura_DynamicBVH
{

public:

    typedef void(*VOID_CALLBACK)(T&);

	Asura_DynamicBVH( void )
	{
        m_pxTree = NULL;
        Reset();
	}

	~Asura_DynamicBVH( void )
	{
		Reset();
	}

    // The iterator must provide Restart(), Next(), Done() and
    // GetCurrent() methods, the latter returning a non-const pointer.
    // The BBRetrievalWrapper class must provide a static
    // GetAABB(T&, Asura_Bounding_Box&) method.
    template< class BBRetrievalWrapper, class Iterator >
	void Create( Iterator& xIt );

    // See above requirements for BBRetrievalFunctor.
    template< class BBRetrievalWrapper >
	void Refit( void );

    // If you wish to use these TestRay() methods then T must provide a
    // LOSTest() method.
	// The output data is an accumulated result and as such is not written to if there are no hits.
	// Users calling this function in isolation must first call Reset( float fInitialMaximumRange ) on xOutputData.
	void TestRay( const Asura_LOS_InputData& xInputData, Asura_LOS_OutputData& xOutputData ) const;
	template<class F> void TestRay_Functor( const Asura_LOS_InputData& xInputData, F& xFunctor ) const;

    // The CallbackFunctor class must provide a () operator that takes
    // a single T& argument.
    template< class CallbackFunctor >
    void TestPoint( const Asura_Vector_3& xPoint, CallbackFunctor& xCBFunctor ) const;
    template< class CallbackFunctor >
    void TestAABB( const Asura_Bounding_Box& xAABB, CallbackFunctor& xCBFunctor ) const;

    // Plain callback function of the above, when operating on global
    // data and/or thread safety isn't an issue.
    void TestPoint( const Asura_Vector_3& xPoint, const VOID_CALLBACK pfnCallback ) const;
    void TestAABB( const Asura_Vector_3& xPoint, const VOID_CALLBACK pfnCallback ) const;

	u_int GetNumberOfLeaves( void ) const { return m_uNumberOfLeaves; }
	u_int GetMaxDepth( void ) const { return m_uMaxDepth; }

	void Reset( void );

private:

    // Wrapper for plain function callback versions of functions.
    template< void(*pfnCallback)(T&) >
    class CallbackFunctionWrapper
    {
    public:
        inline void operator()( T& xPhoton ) const { pfnCallback( xPhoton ); }
    };

	// As a last resort, if for some reason the hierarchy is too unbalanced to be used,
	// do a brute force O(n) test through the list. Also handy for checking the benefit
	// the BVH gives.
	void TestRay_BruteForce( const Asura_LOS_InputData& xInputData, Asura_LOS_OutputData& xOutputData ) const;
	template<class F> void TestRay_BruteForce_Functor( const Asura_LOS_InputData& xInputData, F& xFunctor ) const;

	// KJFL - 16 October 2007 - even faster Ray-AABB intersection code.
	// Based on SAT & removed all divides. Could make it even faster
	// if we didn't want the distance to intersection.
	inline bool DoesRayIntersectAABB( const Asura_Vector_3& xAlpha, const Asura_Vector_3& xOneOverBeta, const Asura_Bounding_Box& xAABB, float& fDistance) const;
	
	Asura_DynamicBVH_Node<T>*	m_pxTree;
	u_int						m_uNumberOfLeaves;
	u_int						m_uNumberOfNodes;
	u_int						m_uMaxDepth;

};


template< class T >
inline bool Asura_DynamicBVH<T>::DoesRayIntersectAABB( const Asura_Vector_3& xAlpha, const Asura_Vector_3& xOneOverBeta, const Asura_Bounding_Box& xAABB, float& fDistance) const
{
	float fMax = fDistance;	// max distance ray can travel 
	fDistance = 0.0f;		// want to return 0 if ray origin inside AABB

	// KJFL - 2 November 2007 - TODO
	// change format for bounding box to be these vectors, methinks?
	const Asura_Vector_3 xMin( xAABB.MinX, xAABB.MinY, xAABB.MinZ );
	const Asura_Vector_3 xMax( xAABB.MaxX, xAABB.MaxY, xAABB.MaxZ );

	// For all three slabs
	for (int i = 0; i < 3; i++)
	{
		const float fOneOverBeta = xOneOverBeta[i];

		if ( fOneOverBeta )		
		{
			// Compute intersection values of ray with near and far plane of slab
			float fDist1 = (xMin[i] - xAlpha[i]) * fOneOverBeta;
			float fDist2 = (xMax[i] - xAlpha[i]) * fOneOverBeta;

			// Make fDist1 be intersection with near plane, fDist2 with far plane
			Asura_Maths::Sort( fDist1, fDist2 );

			fDistance = Asura_Maths::Max( fDist1, fDistance );
			fMax = Asura_Maths::Min( fDist2, fMax );

			// If the interval inverts, there is no intersection
			if (fDistance > fMax)
			{
				return false;
			}
		}
		// Ray is parallel to slab. No hit if origin not within slab
		else if (xAlpha[i] < xMin[i] || xAlpha[i] > xMax[i])
		{
			return false;
		}
	}

	return true;
}


template <class T> void Asura_DynamicBVH<T>::Reset( void )
{
	delete [] m_pxTree;
	m_pxTree = NULL;
	m_uNumberOfLeaves = 0;
	m_uNumberOfNodes = 0;
	m_uMaxDepth = 0;
}


template< class T >
template< class BBRetrievalWrapper, class Iterator >
void Asura_DynamicBVH<T>::Create( Iterator& xIt )
{
	// destroy old data
	Reset();

	for( xIt.Restart(); !xIt.Done(); xIt.Next() )
	{
		++m_uNumberOfLeaves;
	}

	if ( m_uNumberOfLeaves == 0 ) return;

	m_uNumberOfNodes = m_uNumberOfLeaves*2 - 1;

	m_pxTree = new Asura_DynamicBVH_Node<T>[ m_uNumberOfNodes ];

	// KJFL - 2 November 2007 - TODO
	// Would it be better to keep this memory around? Or throw away after construction as currently done?
	Asura_DynamicBVH_ConstructionLeaf<T>* pxLeafConstruction = new Asura_DynamicBVH_ConstructionLeaf<T>[ m_uNumberOfNodes ];

	Asura_Bounding_Box& xRootAABB = m_pxTree[0].m_xAABB;

	xRootAABB.Reset();

	u_int uOut = 0;
	for( xIt.Restart(); !xIt.Done(); xIt.Next() )
	{
		T* pxObject = xIt.GetCurrent();

		Asura_Bounding_Box xAABB;
        BBRetrievalWrapper::GetAABB( *pxObject, xAABB );
		
		pxLeafConstruction[uOut].m_xAABB = xAABB;
		pxLeafConstruction[uOut].m_xCentre = xAABB.CalculateCentre();
		pxLeafConstruction[uOut].m_pxLeafObject = pxObject;
		++uOut;

		xRootAABB.Set( xAABB );
	}
    Asura_Assert( uOut == m_uNumberOfLeaves, "" );

	u_int* puLeafIndices = new u_int[m_uNumberOfLeaves];

	for (u_int u=0; u<m_uNumberOfLeaves; ++u)
	{
		puLeafIndices[u] = u;
	}

	Asura_DynamicBVH_ConstructionStack<T> axStack[32];
	u_int uStack = 0;

	// push root onto stack
	axStack[ uStack ].m_pxNode = m_pxTree;
	axStack[ uStack ].m_uStartIndex = 0;
	axStack[ uStack ].m_uEndIndex = m_uNumberOfLeaves;
	++uStack;

	u_int uFreeNode = 1;


	while ( uStack )
	{
		// pop something off the stack to work on
		--uStack;

		m_uMaxDepth = Asura_Maths::Max( m_uMaxDepth, uStack );
	
		// If we run out of stack space reset everything and give up.
		// 1 less than stack size because we may add *two* children at the end of the loop,
		// so we need space for both.
		if ( uStack >= 31 )
		{
			Asura_Assert( false, "Code bug: stack overflow - tree must be very unbalanced" );

			delete [] puLeafIndices;
			delete [] pxLeafConstruction;

			Reset();

			return;
		}

		Asura_DynamicBVH_Node<T>* pxNode = axStack[ uStack ].m_pxNode;
		u_int uStart = axStack[ uStack ].m_uStartIndex;
		u_int uEnd = axStack[ uStack ].m_uEndIndex;

		if ( uEnd - uStart > 1) // more than one leaf
		{
			float fTestValue;
			u_int uAxis;
			Asura_Bounding_Box xLeftBB, xRightBB;

			u_int uRight = uStart;

			{

				float fScore[3];

				fScore[0] = pxNode->m_xAABB.MaxX - pxNode->m_xAABB.MinX;
				fScore[1] = pxNode->m_xAABB.MaxY - pxNode->m_xAABB.MinY;
				fScore[2] = pxNode->m_xAABB.MaxZ - pxNode->m_xAABB.MinZ;

				uAxis = 2;
				if (fScore[0]>fScore[1])
				{
					if (fScore[0]>fScore[2])
					{
						uAxis = 0;
					}
				}
				else
				{
					if (fScore[1]>fScore[2])
					{
						uAxis = 1;
					}
				}

				fTestValue = pxNode->m_xAABB.CalculateCentre().GetElement( uAxis );

				xLeftBB.Reset();
				xRightBB.Reset();

				for (u_int u = uStart; u<uEnd; ++u)
				{
					u_int uIndex = puLeafIndices[u];

					const Asura_Bounding_Box& xLeafAABB = pxLeafConstruction[uIndex].m_xAABB;

					if ( pxLeafConstruction[uIndex].m_xCentre.GetElement(uAxis) > fTestValue )
					{
						puLeafIndices[ u ] = puLeafIndices[ uRight ];
						puLeafIndices[ uRight ] = uIndex;

						++uRight;

						xRightBB.Set( xLeafAABB );
					}
					else 
					{

						xLeftBB.Set( xLeafAABB );
					}
				}

			}

			if ( uRight == uStart )
			{
				uRight = (uStart+uEnd+1)/2;
				uAxis = 0;
				xLeftBB = pxNode->m_xAABB;
				xRightBB = pxNode->m_xAABB;
			}

			pxNode->m_uLeftChild = uFreeNode++;
			uFreeNode++;
			pxNode->m_pxLeafObject = 0;

			Asura_DynamicBVH_Node<T>* pxLeftNode = &m_pxTree[pxNode->m_uLeftChild];
			Asura_DynamicBVH_Node<T>* pxRightNode = &m_pxTree[pxNode->m_uLeftChild + 1];

			pxLeftNode->m_xAABB = xLeftBB;
			pxRightNode->m_xAABB = xRightBB;

			// push children onto stack
			axStack[ uStack ].m_pxNode = pxLeftNode;
			axStack[ uStack ].m_uStartIndex = uRight;
			axStack[ uStack ].m_uEndIndex = uEnd;
			uStack++;

			axStack[ uStack ].m_pxNode = pxRightNode;
			axStack[ uStack ].m_uStartIndex = uStart;
			axStack[ uStack ].m_uEndIndex = uRight;
			uStack++;

		}
		else // We're a leaf node
		{
			pxNode->m_uLeftChild = 0;
			pxNode->m_pxLeafObject = pxLeafConstruction[ puLeafIndices[uStart] ].m_pxLeafObject;
		}				

	}

	delete [] puLeafIndices;
	delete [] pxLeafConstruction;
}

template< class T >
template< class BBRetrievalWrapper >
void Asura_DynamicBVH<T>::Refit( void )
{
	// O(n) linear pass through the tree (backwards atm)
	for ( u_int u=m_uNumberOfNodes; u-- > 0; )
	{
		Asura_DynamicBVH_Node<T>& xNode = m_pxTree[ u ];

		T* pxObject = xNode.m_pxLeafObject;

		// if you're a leaf node, update your BB
		if ( pxObject )
		{
            BBRetrievalWrapper::GetAABB( *pxObject, xNode.m_xAABB );
		}
		else // otherwise make your BB the union of your children's
		{
			xNode.m_xAABB = m_pxTree[ xNode.m_uLeftChild ].m_xAABB;
			xNode.m_xAABB.Set( m_pxTree[ xNode.m_uLeftChild + 1 ].m_xAABB );
		}

	}
}

template <class T> void Asura_DynamicBVH<T>::TestRay( const Asura_LOS_InputData& xInputData, Asura_LOS_OutputData& xOutputData ) const
{
	if ( !m_pxTree )
	{
		TestRay_BruteForce( xInputData, xOutputData );
		return;
	}

	const Asura_Vector_3& xRayDirection = xInputData.m_xDirection;
	const Asura_Vector_3& xRayOrigin = xInputData.m_xOrigin;

	Asura_Vector_3 xOneOverBeta;

	// precalc the divides we need for the ray-AABB intersection tests
	for (int i = 0; i < 3; i++)
	{
		const float fBETAEPSILON = 0.000001f;

		// avoid denormals when divisor is very small
		if ( Asura_Maths::Abs( xRayDirection[i]) < fBETAEPSILON )
		{
			xOneOverBeta[i] = 0.0f;
		}
		else
		{
			xOneOverBeta[i] = 1.0f / xRayDirection[i];
		}
	}

	float fMaxDistance = xInputData.m_fInitialMaximumRange;
	float fDistance = fMaxDistance;

	if ( !DoesRayIntersectAABB( xRayOrigin, xOneOverBeta, m_pxTree[0].m_xAABB, fDistance) ) return;

	const Asura_Guid uIgnore = xInputData.m_uEntityToIgnore;

	u_int auStack[32];
	u_int uStack = 0;

	// push root onto stack
	auStack[ uStack ] = 0;
	++uStack;

	while ( uStack )
	{
		// pop something off the stack to work on
		// no need to check stack size because Create() routine won't allow invalid tree depths
		--uStack;
		const Asura_DynamicBVH_Node<T>& xNode = m_pxTree[ auStack[ uStack ] ];

		if ( xNode.m_pxLeafObject )
		{
			if ( uIgnore == ASURA_GUID_UNREGISTERED || uIgnore != xNode.m_pxLeafObject->GetGuid() )
			{
                if ( xNode.m_pxLeafObject->LOSTest( &xInputData, &xOutputData ) )
				{
					fMaxDistance = xOutputData.m_fDistanceToIntersection;
				}
			}
		}
		else
		{
			// Left node is always followed by right node in memory - should improve caching.
			const Asura_DynamicBVH_Node<T>& xLeftNode = m_pxTree[ xNode.m_uLeftChild ];
			const Asura_DynamicBVH_Node<T>& xRightNode = m_pxTree[ xNode.m_uLeftChild + 1 ];

			// TODO - make a ray vs box that tests against two boxes simultaneously?
			float fLeftDistance = fMaxDistance;

			bool bHitLeft = DoesRayIntersectAABB( xRayOrigin, xOneOverBeta, xLeftNode.m_xAABB, fLeftDistance );

			float fRightDistance = fMaxDistance;

			bool bHitRight = DoesRayIntersectAABB( xRayOrigin, xOneOverBeta, xRightNode.m_xAABB, fRightDistance );

			// KJFL - 12 October 2007 - cunning bit: determine order on stack 
			// by distance to nodes - refines test quicker.
			if ( bHitLeft && bHitRight && (fLeftDistance < fRightDistance) )
			{
				auStack[ uStack++ ] = xNode.m_uLeftChild + 1;
				auStack[ uStack++ ] = xNode.m_uLeftChild;
			}
			else 
			{	
				if (bHitLeft)
				{
					auStack[ uStack++ ] = xNode.m_uLeftChild;
				}
				if (bHitRight)
				{
					auStack[ uStack++ ] = xNode.m_uLeftChild + 1;
				}
			}
		}
	}
}


template <class T> void Asura_DynamicBVH<T>::TestRay_BruteForce( const Asura_LOS_InputData& xInputData, Asura_LOS_OutputData& xOutputData ) const
{
	// As a last resort, if for some reason the hierarchy is too unbalanced to be used,
	// do a brute force O(n) test through the list. Also handy for checking the benefit
	// the BVH gives.
	// NOTE: this will also get called if the list is empty, since that results in an empty BVH.

	const Asura_Guid uIgnore = xInputData.m_uEntityToIgnore;

	for( Asura_Classified_List_It<T> xIt(T::GetLineOfSightList()); !xIt.Done(); xIt.Next() )
	{
		T* pxObject = xIt.GetCurrent();

		if( pxObject->IsDestroyed() ) { continue; }

		if ( uIgnore == ASURA_GUID_UNREGISTERED || uIgnore != pxObject->GetGuid() )
		{
			pxObject->LOSTest( &xInputData, &xOutputData );
		}	
	}
}

// MGH 3/2/2008: Variant of TestRay that uses a functor to allow multiple LOS hits to be accumulated.
// Used by Asura_LOS_Penetration (e.g. for bullet penetration and sound occlusion).

template <class T> template<class F> void Asura_DynamicBVH<T>::TestRay_Functor( const Asura_LOS_InputData& xInputData, F& xFunctor ) const
{
	if ( xFunctor.ShouldStop() )
	{
		// Stop the search. Typically this happens when the output array is full.
		return;
	}

	if ( !m_pxTree )
	{
		TestRay_BruteForce_Functor( xInputData, xFunctor );
		return;
	}

	const Asura_Vector_3& xRayDirection = xInputData.m_xDirection;
	const Asura_Vector_3& xRayOrigin = xInputData.m_xOrigin;

	Asura_Vector_3 xOneOverBeta;

	// precalc the divides we need for the ray-AABB intersection tests
	for (int i = 0; i < 3; i++)
	{
		const float fBETAEPSILON = 0.000001f;

		// avoid denormals when divisor is very small
		if ( Asura_Maths::Abs( xRayDirection[i]) < fBETAEPSILON )
		{
			xOneOverBeta[i] = 0.0f;
		}
		else
		{
			xOneOverBeta[i] = 1.0f / xRayDirection[i];
		}
	}

	float fMaxDistance = xFunctor.GetInitialMaxDistance();
	float fDistance = fMaxDistance;

	if ( !DoesRayIntersectAABB( xRayOrigin, xOneOverBeta, m_pxTree[0].m_xAABB, fDistance) ) return;

	const Asura_Guid uIgnore = xInputData.m_uEntityToIgnore;

	u_int auStack[32];
	u_int uStack = 0;

	// push root onto stack
	auStack[ uStack ] = 0;
	++uStack;

	while ( uStack )
	{
		// pop something off the stack to work on
		// no need to check stack size because Create() routine won't allow invalid tree depths
		--uStack;
		const Asura_DynamicBVH_Node<T>& xNode = m_pxTree[ auStack[ uStack ] ];

		if ( xNode.m_pxLeafObject )
		{
			if ( uIgnore == ASURA_GUID_UNREGISTERED || uIgnore != xNode.m_pxLeafObject->GetGuid() )
			{
				if ( xFunctor.CheckEntity(xNode.m_pxLeafObject, xInputData) )
				{
					fMaxDistance = xFunctor.GetMaxDistance();
				}
				if ( xFunctor.ShouldStop() )
				{
					// Stop the search. Typically this happens when the output array is full.
					return;
				}	
			}
		}
		else
		{
			// Left node is always followed by right node in memory - should improve caching.
			const Asura_DynamicBVH_Node<T>& xLeftNode = m_pxTree[ xNode.m_uLeftChild ];
			const Asura_DynamicBVH_Node<T>& xRightNode = m_pxTree[ xNode.m_uLeftChild + 1 ];

			// TODO - make a ray vs box that tests against two boxes simultaneously?
			float fLeftDistance = fMaxDistance;

			bool bHitLeft = DoesRayIntersectAABB( xRayOrigin, xOneOverBeta, xLeftNode.m_xAABB, fLeftDistance );

			float fRightDistance = fMaxDistance;

			bool bHitRight = DoesRayIntersectAABB( xRayOrigin, xOneOverBeta, xRightNode.m_xAABB, fRightDistance );

			// KJFL - 12 October 2007 - cunning bit: determine order on stack 
			// by distance to nodes - refines test quicker.
			if ( bHitLeft && bHitRight && (fLeftDistance < fRightDistance) )
			{
				auStack[ uStack++ ] = xNode.m_uLeftChild + 1;
				auStack[ uStack++ ] = xNode.m_uLeftChild;
			}
			else 
			{	
				if (bHitLeft)
				{
					auStack[ uStack++ ] = xNode.m_uLeftChild;
				}
				if (bHitRight)
				{
					auStack[ uStack++ ] = xNode.m_uLeftChild + 1;
				}
			}
		}
	}
}

template <class T> template <class F> void Asura_DynamicBVH<T>::TestRay_BruteForce_Functor( const Asura_LOS_InputData& xInputData, F& xFunctor ) const
{
	// As a last resort, if for some reason the hierarchy is too unbalanced to be used,
	// do a brute force O(n) test through the list. Also handy for checking the benefit
	// the BVH gives.
	// NOTE: this will also get called if the list is empty, since that results in an empty BVH.

	const Asura_Guid uIgnore = xInputData.m_uEntityToIgnore;

	for( Asura_Classified_List_It<T> xIt(T::GetLineOfSightList()); !xIt.Done(); xIt.Next() )
	{
		T* pxObject = xIt.GetCurrent();

		if( pxObject->IsDestroyed() ) { continue; }

		if ( uIgnore == ASURA_GUID_UNREGISTERED || uIgnore != pxObject->GetGuid() )
		{
			xFunctor.CheckEntity(pxObject, xInputData);

			if (xFunctor.ShouldStop())
			{
				// Stop the search. Typically this happens when the output array is full.
				break;
			}
		}	
	}
}


template< class T >
template< class CallbackFunctor >
void Asura_DynamicBVH<T>::TestPoint( const Asura_Vector_3& xPoint, CallbackFunctor& xCBFunctor ) const
{
	if( !m_pxTree || !m_pxTree[0].m_xAABB.Contains(xPoint) )
	{
		return;
	}

	u_int auNodeStack[32];
	u_int uNodeStackIdx = 0;

	// push root onto stack
	auNodeStack[uNodeStackIdx++] = 0;

	while( uNodeStackIdx )
	{
		// pop something off the stack to work on
		// no need to check stack size because Create() routine won't allow invalid tree depths
		const Asura_DynamicBVH_Node<T>& xNode = m_pxTree[ auNodeStack[--uNodeStackIdx] ];

		if( xNode.m_pxLeafObject )
		{
            xCBFunctor( *xNode.m_pxLeafObject );
		}
		else
		{
			const u_int uLeftChild = xNode.m_uLeftChild;
			if( m_pxTree[uLeftChild].m_xAABB.Contains(xPoint) )
			{
				auNodeStack[uNodeStackIdx++] = uLeftChild;
			}

			const u_int uRightChild = uLeftChild + 1;
			if( m_pxTree[uRightChild].m_xAABB.Contains(xPoint) )
			{
				auNodeStack[uNodeStackIdx++] = uRightChild;
			}
		}
	}
}


template< class T >
template< class CallbackFunctor >
void Asura_DynamicBVH<T>::TestAABB( const Asura_Bounding_Box& xAABB, CallbackFunctor& xCBFunctor ) const
{
	if( !m_pxTree || !m_pxTree[0].m_xAABB.Intersects(xAABB) )
	{
		return;
	}

	u_int auNodeStack[32];
	u_int uNodeStackIdx = 0;

	// push root onto stack
	auNodeStack[uNodeStackIdx++] = 0;

	while( uNodeStackIdx )
	{
		// pop something off the stack to work on
		// no need to check stack size because Create() routine won't allow invalid tree depths
		const Asura_DynamicBVH_Node<T>& xNode = m_pxTree[ auNodeStack[--uNodeStackIdx] ];

		if( xNode.m_pxLeafObject )
		{
            xCBFunctor( *xNode.m_pxLeafObject );
		}
		else
		{
			const u_int uLeftChild = xNode.m_uLeftChild;
			if( m_pxTree[uLeftChild].m_xAABB.Intersects(xAABB) )
			{
				auNodeStack[uNodeStackIdx++] = uLeftChild;
			}

			const u_int uRightChild = uLeftChild + 1;
			if( m_pxTree[uRightChild].m_xAABB.Intersects(xAABB) )
			{
				auNodeStack[uNodeStackIdx++] = uRightChild;
			}
		}
	}
}


template< class T >
void Asura_DynamicBVH<T>::TestPoint( const Asura_Vector_3& xPoint, const VOID_CALLBACK pfnCallback ) const
{
    CallbackFunctionWrapper<pfnCallback> xWrapper;
    TestPoint( xPoint, xWrapper );
}


template< class T >
void Asura_DynamicBVH<T>::TestAABB( const Asura_Vector_3& xPoint, const VOID_CALLBACK pfnCallback ) const
{
    CallbackFunctionWrapper<pfnCallback> xWrapper;
    TestAABB( xPoint, xWrapper );
}

#endif//__ASURA_DYNAMICBVH_H__


