//
// Asura_Graph.h
//
// Created by KOR - 06/02/2004 17:02:08
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_GRAPH_H__
#define __ASURA_GRAPH_H__

#ifndef ASURA_GRAPH_LEAN_AND_MEAN
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Handle.h"
#endif // #ifndef ASURA_GRAPH_LEAN_AND_MEAN
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"



// Defines
#define ASURA_GRAPH_CHUNK_FLAG_SAMPLES (0x00000001)

//Default tolerance level to use when determining how many samples to generate
//(0.05 being 5%) RAR 04/10/2005
#define ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE (0.05f)

// (AW - 09.11.2004) - Graph version numbers start at 10000 to get around fact that original
// incarnation didn't have a version number (I'm assuming that NumSplinePoints will always
// be less than 10000 to make this little hack work)
// Version 10002 - KOR - 22/12/2004 - changed to explicitly separate exported (samples only) and tools (source data only) versions
// Version 10003 - AAM - 13/03/2007 - Added domain min and max
const u_int uASURA_GRAPH_VERSION_NUMBER = 10003;

#ifndef ASURA_GRAPH_LEAN_AND_MEAN
template <class T> class Asura_Graph_Point
{
	public:
		// Constructor
		Asura_Graph_Point(const T& xValue, float fTime)
		: m_xValue(xValue)
		, m_fTime(fTime)
		, m_pxNext(NULL)
		, m_pxPrev(NULL)
		{
		}
		// Destructor
		~Asura_Graph_Point()
		{
		}
		void DeleteRecursive()
		{
			if(m_pxNext)
			{
				m_pxNext->DeleteRecursive();
				delete m_pxNext;
			}
		}
		// Members
		T						m_xValue;
		float					m_fTime;
		Asura_Graph_Point<T>*	m_pxNext;
		Asura_Graph_Point<T>*	m_pxPrev;
};

// (AW - 09.11.2004) - Little class that encapsulates data required to build a graph (generally only
// exists in tools)
template <class T> class Asura_Graph_Source_Data
{
public:
	Asura_Graph_Source_Data()
	: m_pxSpline_Points(NULL)
	, m_uFlags(0)
	{

	}
	~Asura_Graph_Source_Data()
	{
		if(m_pxSpline_Points)
		{
			// This will recursively delete the list
			m_pxSpline_Points->DeleteRecursive();
			delete m_pxSpline_Points;
		}
	}
	const Asura_Graph_Source_Data& operator = (const Asura_Graph_Source_Data& xOther)
	{
		if(this != &xOther)
		{
			if(m_pxSpline_Points)
			{
				m_pxSpline_Points->DeleteRecursive();
				delete m_pxSpline_Points;
				m_pxSpline_Points = 0;
			}
			m_fClampMin = xOther.m_fClampMin;
			m_fClampMax = xOther.m_fClampMax;
			m_uFlags = xOther.m_uFlags;
			
			const Asura_Graph_Point<T>* pxSourcePoint = xOther.m_pxSpline_Points;
			Asura_Graph_Point<T>* pxPrevPoint = 0;

			while(pxSourcePoint)
			{
				Asura_Graph_Point<T>* pxCopy = new Asura_Graph_Point<T>(pxSourcePoint->m_xValue, pxSourcePoint->m_fTime);
				pxCopy->m_pxPrev = pxPrevPoint;
				if(pxPrevPoint)
				{
					pxPrevPoint->m_pxNext = pxCopy;
				}
				else
				{
					m_pxSpline_Points = pxCopy;
				}

				pxPrevPoint = pxCopy;
				pxSourcePoint = pxSourcePoint->m_pxNext;
			}
		}
		return *this;
	}

	// Linked list of spline points
	Asura_Graph_Point<T>* m_pxSpline_Points;
	// Clamping data.
	float m_fClampMin;
	float m_fClampMax;
	union
	{
		u_int m_uFlags;
		struct
		{
			u_int		m_bUseClampMin : 1,
						m_bUseClampMax : 1,
						SpareBits : 30;
		};
	};
};
#endif // #ifndef ASURA_GRAPH_LEAN_AND_MEAN


template <class T> class Asura_Graph
{
	public:

		// main workhorse lookup function
		T GetValueAtTime_Discrete(float fTime) const 
		{
			// Discrete version
			Asura_Assert(m_uNumberOfDiscreteSamples && m_pxDiscreteSamples, "No discrete samples");
			int iNumDiscreteSamplesMinus1 = static_cast<int>(m_uNumberOfDiscreteSamples - 1);
			float fIndex = (fTime - m_fDomainMin) / (m_fDomainMax - m_fDomainMin) * static_cast< float >( iNumDiscreteSamplesMinus1 );
			int iIndex = Asura_Maths::FloatToInt_Chop(fIndex);
			// Range checking
			if(iIndex < 0)
			{
				return m_pxDiscreteSamples[0];
			}
			if(iIndex >= iNumDiscreteSamplesMinus1)
			{
				return m_pxDiscreteSamples[iNumDiscreteSamplesMinus1];
			}
			const float fScale = fIndex - iIndex;
			return m_pxDiscreteSamples[iIndex] * (1.f - fScale) + m_pxDiscreteSamples[iIndex + 1] * fScale;
		}

		float GetDomainMin() const { return m_fDomainMin; }
		float GetDomainMax() const { return m_fDomainMax; }

#ifdef ASURA_GRAPH_LEAN_AND_MEAN

		Asura_Graph()
		: m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
		, m_fDomainMin( 0.0f )
		, m_fDomainMax( 1.0f )
		{}

		~Asura_Graph()
		{
			delete[] m_pxDiscreteSamples;
		}
		
		T GetValueAtTime(float fTime) const
		{
			return GetValueAtTime_Discrete(fTime);
		}
		
		// SetFromDataBlock allocates and fills out samples from input block of data and 
		// returns ptr to memory immediately after the input block.
		const void* SetFromDataBlock( const void* pvDiscreteSampleData, u_int uNumDiscreteSamples, float fDomainMin, float fDomainMax )
		{
			const T* pxDiscreteSamples = static_cast<const T*>( pvDiscreteSampleData );

			m_pxDiscreteSamples = new T[uNumDiscreteSamples];
			Asura_Mem::Copy( m_pxDiscreteSamples, pxDiscreteSamples, sizeof(T) * uNumDiscreteSamples );
			m_uNumberOfDiscreteSamples = uNumDiscreteSamples;

            m_fDomainMin = fDomainMin;
            m_fDomainMax = fDomainMax;

			return pxDiscreteSamples + uNumDiscreteSamples;
		}

#else

		// Constructors
		// Default. This will not be valid - it will have to be filled out from a chunk buffer later
		Asura_Graph()
		: m_pxGraphSourceData(NULL)
		, m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
#ifdef ASURA_TOOLS
		, m_fSampleTolerance(ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE)
		, m_bProportionalSampleTolerance(true)
		, m_uMaxNumberOfDiscreteSamples(0)
#endif
		, m_fDomainMin(0.0f)
		, m_fDomainMax(1.0f)
		{}

		// Copy constructor
		Asura_Graph(const Asura_Graph& pxOriginal)
		: m_pxGraphSourceData(NULL)
		, m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
#ifdef ASURA_TOOLS
		, m_fSampleTolerance(ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE)
		, m_bProportionalSampleTolerance(true)
		, m_uMaxNumberOfDiscreteSamples(0)
#endif
		, m_fDomainMin(0.0f)
		, m_fDomainMax(1.0f)
		{
			CopyFrom(pxOriginal);
		}
		// Blank, with start and end points
		Asura_Graph(u_int uDiscreteSamples, T xStartValue, T xEndValue, float fDomainMin = 0.0f, float fDomainMax = 1.0f)
		: m_pxGraphSourceData(NULL)
		, m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
#ifdef ASURA_TOOLS
		, m_fSampleTolerance(ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE)
		, m_bProportionalSampleTolerance(true)
#endif
		{
			Initialise(uDiscreteSamples, xStartValue, xEndValue, fDomainMin, fDomainMax);
		}

		//assignment operator
		Asura_Graph& operator = (const Asura_Graph& xOriginal)
		{
			if(this != &xOriginal)
			{
				CopyFrom(xOriginal);
			}
			return *this;
		}

		void Initialise(u_int uDiscreteSamples, T xStartValue, T xEndValue, float fDomainMin = 0.0f, float fDomainMax = 1.0f)
		{
			// DMR - 18/05/07 - Make it safe to call Initialise more than once - it would previously leak memory if this happened			
			if(m_pxDiscreteSamples)
			{
				delete [] m_pxDiscreteSamples;
				m_pxDiscreteSamples = 0;
			}
			if(m_pxGraphSourceData)
			{
				delete m_pxGraphSourceData;
				m_pxGraphSourceData = 0;
			}
			

			m_pxGraphSourceData = new Asura_Graph_Source_Data<T>;
			m_uNumberOfDiscreteSamples = uDiscreteSamples;
#ifdef ASURA_TOOLS
			m_uMaxNumberOfDiscreteSamples = uDiscreteSamples;
#endif

			m_fDomainMin = fDomainMin;
			m_fDomainMax = fDomainMax;

			m_pxDiscreteSamples = new T[uDiscreteSamples];
			// Must start with 4 points - start and end, with extra control points
			AddPointInternal((xStartValue * 2.0f) - xEndValue, (fDomainMin * 2.0f) - fDomainMax);
			AddPointInternal(xStartValue, fDomainMin);
			AddPointInternal(xEndValue, fDomainMax);
			AddPointInternal((xEndValue * 2.0f) - xStartValue, (fDomainMax * 2.0f) - fDomainMin);
			// End points should be correct at this point, but just in case:
			ReCalculateEndPoints();
			ReCalculateDiscreteSamples();
		}

		// From chunk buffer. Behaviour depends on whether it's tools or not
		Asura_Graph(const char*& pcChunkBuffer)
		: m_pxGraphSourceData(NULL)
		, m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
#ifdef ASURA_TOOLS
		, m_fSampleTolerance(ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE)
		, m_bProportionalSampleTolerance(true)
#endif
		, m_fDomainMin(0.0f)
		, m_fDomainMax(1.0f)
		{
			Asura_Chunk_ChunkBuffer_Stream xStream( pcChunkBuffer );
			ReadFromChunkStream( xStream );
		}

		// From stream
		Asura_Graph(Asura_Chunk_Stream& xStream)
		: m_pxGraphSourceData(NULL)
		, m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
#ifdef ASURA_TOOLS
		, m_fSampleTolerance(ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE)
		, m_bProportionalSampleTolerance(true)
#endif
		, m_fDomainMin( 0.0f )
		, m_fDomainMax( 1.0f )
		{
			ReadFromChunkStream(xStream);
		}

		// From filehandle
		Asura_Graph(Asura_Handle iFileHandle)
		: m_pxGraphSourceData(NULL)
		, m_uNumberOfDiscreteSamples(0)
		, m_pxDiscreteSamples(NULL)
#ifdef ASURA_TOOLS
		, m_fSampleTolerance(ASURA_GRAPH_DEFAULT_SAMPLE_TOLERANCE)
		, m_bProportionalSampleTolerance(true)
#endif
		, m_fDomainMin( 0.0f )
		, m_fDomainMax( 1.0f )
		{
			ProcessFromFileHandle(iFileHandle);
		}

		void ReadFromChunkStream(Asura_Chunk_Stream& xStream)
		{
			// delete any old data we may have
			if(m_pxDiscreteSamples)
			{
				delete [] m_pxDiscreteSamples;
				m_pxDiscreteSamples = 0;
			}
			if(m_pxGraphSourceData)
			{
				delete m_pxGraphSourceData;
				m_pxGraphSourceData = 0;
			}
			

			u_int uVersionNumber;
            xStream >> uVersionNumber;
            if ( uVersionNumber >= 10002 )
            {
                // Check whether it's spline data or samples
                u_int uChunkFlags = 0;
                xStream >> uChunkFlags;
#ifdef ASURA_TOOLS
                if(!(uChunkFlags & ASURA_GRAPH_CHUNK_FLAG_SAMPLES))
                {
                    // Spline data
                    u_int uNumSplinePoints;
                    xStream >> uNumSplinePoints;
                    m_pxGraphSourceData = new Asura_Graph_Source_Data<T>;
                    for(u_int u = 0; u < uNumSplinePoints; ++u)
                    {
                        T xValue;
                        float fTime;
                        xStream >> xValue;
                        xStream >> fTime;
                        AddPointInternal(xValue, fTime);
                    }
                    xStream >> m_pxGraphSourceData->m_fClampMin;
                    xStream >> m_pxGraphSourceData->m_fClampMax;
                    xStream >> m_pxGraphSourceData->m_uFlags;
                    xStream >> m_uNumberOfDiscreteSamples;

					if( uVersionNumber >= 10003 )
					{
						xStream >> m_fDomainMin;
						xStream >> m_fDomainMax;
					}

                    m_uMaxNumberOfDiscreteSamples = m_uNumberOfDiscreteSamples;
                    m_pxDiscreteSamples = new T[m_uNumberOfDiscreteSamples];
                    ReCalculateDiscreteSamples();
                }
                else
#else//ASURA_TOOLS
                Asura_Assert(uChunkFlags & ASURA_GRAPH_CHUNK_FLAG_SAMPLES, "Should not be loading spline data outside tools");
#endif
                {
                    // Samples
                    xStream >> m_uNumberOfDiscreteSamples;
                    m_pxDiscreteSamples = new T[m_uNumberOfDiscreteSamples];
#ifdef ASURA_TOOLS
                    m_uMaxNumberOfDiscreteSamples = m_uNumberOfDiscreteSamples;
#endif
                    // Copy samples themselves
                    for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
                    {
                        xStream >> m_pxDiscreteSamples[u];
                    }
                }
            }
            else
            {
                // Old version has spline data and samples
                u_int uNumSplinePoints;
                if (uVersionNumber < 10000)
                {
                    // (AW - 09.11.2004) - Hack to get around fact that original version had no version number.
                    uNumSplinePoints = uVersionNumber;
                    uVersionNumber = 10000;
                }
                else
                {
                    xStream >> uNumSplinePoints;
                }
#ifdef ASURA_TOOLS
                m_pxGraphSourceData = new Asura_Graph_Source_Data<T>;
                for(u_int u = 0; u < uNumSplinePoints; ++u)
                {
                    T xValue;
                    float fTime;
                    xStream >> xValue;
                    xStream >> fTime;
                    AddPointInternal(xValue, fTime);
                }
                if (uVersionNumber > 10000)
                {
                    xStream >> m_pxGraphSourceData->m_fClampMin;
                    xStream >> m_pxGraphSourceData->m_fClampMax;
                    xStream >> m_pxGraphSourceData->m_uFlags;
                }
#else
                u_int uBytesToSkip = uNumSplinePoints * (sizeof(T) + sizeof(float));
                if (uVersionNumber > 10000)
                {
                    uBytesToSkip += (2 * sizeof(float)) + sizeof(u_int);
                }
				xStream.Skip( uBytesToSkip );
#endif
                // Read in discrete samples
                xStream >> m_uNumberOfDiscreteSamples;
#ifdef ASURA_TOOLS
                m_uMaxNumberOfDiscreteSamples = m_uNumberOfDiscreteSamples;
#endif
                m_pxDiscreteSamples = new T[m_uNumberOfDiscreteSamples];
                // Copy samples themselves
                for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
                {
                    xStream >> m_pxDiscreteSamples[u];
                }
            }
        }

		void ProcessFromFileHandle(Asura_Handle iFileHandle)
		{
			Asura_Chunk_File_Stream xStream( iFileHandle );
			ReadFromChunkStream( xStream );
		}

		void WriteToChunkStream(Asura_Chunk_Stream& xStream, bool bSamplesOnly, bool bIntegrateSamples = false) const
		{
			u_int uVersionNumber = uASURA_GRAPH_VERSION_NUMBER;
			xStream << uVersionNumber;
			// Can only save spline data in tools
			if(!bSamplesOnly)
			{
				u_int uChunkFlags = 0;
				xStream << uChunkFlags;
				// Fill in spline points
				u_int uNumSplinePoints = GetNumberOfSplinePoints();
				xStream << uNumSplinePoints;
				Asura_Graph_Point<T>* pxPoint = GetControlPoints();
				while(pxPoint)
				{
					xStream << pxPoint->m_xValue;
					xStream << pxPoint->m_fTime;
					pxPoint = pxPoint->m_pxNext;
				}
				Asura_Assert(m_pxGraphSourceData, "ERROR: No source data to save. Using dummy values");
				if (m_pxGraphSourceData)
				{
					xStream << m_pxGraphSourceData->m_fClampMin;
					xStream << m_pxGraphSourceData->m_fClampMax;
					xStream << m_pxGraphSourceData->m_uFlags;
				}
				else
				{
					// Fill out with some dummy data so that it doesn't break the chunk load
					float fZero = 0.f;
					xStream << fZero;
					xStream << fZero;
					xStream << uChunkFlags; // uChunkFlags is zero
				}
				// Max samples.
#ifdef ASURA_TOOLS
				xStream << m_uMaxNumberOfDiscreteSamples;
#else
				xStream << m_uNumberOfDiscreteSamples;
#endif

				xStream << m_fDomainMin;
				xStream << m_fDomainMax;
			}
			else
			{
				u_int uChunkFlags = ASURA_GRAPH_CHUNK_FLAG_SAMPLES;
				xStream << uChunkFlags;
				if(IsConstant())
				{
					if(bIntegrateSamples)
					{
						// Constant graph integrates to linear
						// Two samples
						u_int uTwo = 2;
						xStream << uTwo;
						// First
						T xValue = GetAreaUnderGraph( m_fDomainMin );
						xStream << xValue;
						// Last
						xValue = GetAreaUnderGraph( m_fDomainMax );
						xStream << xValue;
					}
					else
					{
						T xFirstValue = GetValueAtTime( m_fDomainMin );
						// Only one sample
						u_int uOne = 1;
						xStream << uOne;
						// Which is the constant
						xStream << xFirstValue;
					}
				}
				// Check for linearity (just two samples)
				// Linear integrates to curve, so keep number of samples if integrating
				else if(IsLinear() && !bIntegrateSamples)
				{
					// Two samples
					u_int uTwo = 2;
					xStream << uTwo;
					// First
					T xValue = GetValueAtTime( m_fDomainMin );
					xStream << xValue;
					// Last
					xValue = GetValueAtTime( m_fDomainMax );
					xStream << xValue;
				}
				else
				{
					// Fill in discrete samples
					xStream << m_uNumberOfDiscreteSamples;
					// Copy samples themselves
					for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
					{
						if(bIntegrateSamples)
						{
							const float fProportion = static_cast< float >(u) / (m_uNumberOfDiscreteSamples - 1);
							const float fX = Asura_Maths::Lerp( m_fDomainMin, m_fDomainMax, fProportion );
							T xValue = GetAreaUnderGraph( fX );
							xStream << xValue;
						}
						else
						{
							xStream << m_pxDiscreteSamples[u];
						}
					}
				}
			}
		}

		void FillInChunkBuffer(char *&pcChunkBuffer, bool bSamplesOnly, bool bIntegrateSamples = false) const
		{
			Asura_Chunk_ChunkBuffer_Stream xStream( pcChunkBuffer );
			WriteToChunkStream( xStream, bSamplesOnly, bIntegrateSamples );
		}
		
		u_int GetChunkBufferSize(bool bSamplesOnly, bool bIntegrateSamples = false) const
		{
			u_int uChunkSize = sizeof(u_int);	//Version number.
			uChunkSize += sizeof(u_int); // Chunk flags
			if(!bSamplesOnly)
			{
				uChunkSize += sizeof(u_int); // uNumSplinePoints 
				Asura_Graph_Point<T>* pxPoint = GetControlPoints();
				while(pxPoint)
				{
					uChunkSize += sizeof(T); // value at point
					uChunkSize += sizeof(float); // time at point
					pxPoint = pxPoint->m_pxNext;
				}
				uChunkSize += sizeof(float) * 2;  //Clamp min and max.
				uChunkSize += sizeof(u_int); // Source data flags.
				uChunkSize += sizeof(u_int); // number of discrete samples
				return uChunkSize;
			}
			else
			{
				uChunkSize += sizeof(u_int); // number of discrete samples
				int iNumberOfSamples;
				if(IsConstant())
				{
					if(bIntegrateSamples)
					{
						iNumberOfSamples = 2;
					}
					else
					{
						iNumberOfSamples = 1;
					}
				}
				else if(IsLinear() && !bIntegrateSamples)
				{
					iNumberOfSamples = 2;
				}
				else
				{
					iNumberOfSamples = m_uNumberOfDiscreteSamples;
				}
				uChunkSize += iNumberOfSamples * sizeof(T); // discrete samples
				return uChunkSize;
			}
		}

		// WriteDataBlock() fills out a block of memory with just the discrete samples and 
		// returns ptr to memory immediately after the block.
		void* WriteDataBlock(void* pDataBlock) const
		{
			u_int uNumBytes = GetDataBlockSize();
			Asura_Mem::Copy( pDataBlock, m_pxDiscreteSamples, uNumBytes );
			return( static_cast<char*>(pDataBlock) + uNumBytes );
		}
		// GetDataBlockSize() returns the size of the memory needed for WriteDataBlock()
		u_int GetDataBlockSize() const { return sizeof(T) * m_uNumberOfDiscreteSamples; }

		// Destructor
		~Asura_Graph()
		{
			if(m_pxDiscreteSamples) delete [] m_pxDiscreteSamples;
			if(m_pxGraphSourceData)	delete m_pxGraphSourceData;
		}

		void AddPoint(T xValue, float fTime)
		{
			AddPointInternal(xValue, fTime);
			ReCalculateEndPoints();
		}

		void RemovePoint(Asura_Graph_Point<T>* pxPoint)
		{
			RemovePointInternal(pxPoint);
			ReCalculateEndPoints();
		}

		void SetClampingValues(bool bUseClampMin, bool bUseClampMax, float fMinClamp, float fMaxClamp)
		{
			Asura_Assert(!bUseClampMin || !bUseClampMax || fMaxClamp > fMinClamp, "Minimum clamp is more or equal to maximum clamp");
			Asura_Assert(m_pxGraphSourceData, "Graph needs source data to set clamping values");
			if(!m_pxGraphSourceData) return;
			m_pxGraphSourceData->m_bUseClampMin = bUseClampMin;
			m_pxGraphSourceData->m_bUseClampMax = bUseClampMax;
			m_pxGraphSourceData->m_fClampMin = fMinClamp;
			m_pxGraphSourceData->m_fClampMax = fMaxClamp;
			Update();
		}

#ifdef ASURA_TOOLS
		//Sets tolerance level for determining how few samples can be got away.
		//If bProportional then the tolerance is multipled by the difference between the max and min values
		//to determine what the allowed error is.
		void SetSampleTolerance(float fTol, bool bProportional = true)
		{
			if(m_fSampleTolerance == fTol && bProportional == m_bProportionalSampleTolerance) return;
			m_fSampleTolerance = fTol;
			m_bProportionalSampleTolerance = bProportional;
			ReCalculateDiscreteSamples();

		}
#endif

		bool GetClampMin(float &fClampMin)
		{
			if (m_pxGraphSourceData->m_bUseClampMin)
			{
				fClampMin = m_pxGraphSourceData->m_fClampMin;
				return true;
			}
			return false;
		}

		bool GetClampMax(float &fClampMax)
		{
			if (m_pxGraphSourceData->m_bUseClampMax)
			{
				fClampMax = m_pxGraphSourceData->m_fClampMax;
				return true;
			}
			return false;
		}

		void Update()
		{
			ApplyClamp();
			ReCalculateEndPoints();	
			ReCalculateDiscreteSamples();
		}

		inline void GetDiscreteIndexAndScale(const float fTime, u_int &uIndex, float &fScale) const
		{
			Asura_Assert(m_uNumberOfDiscreteSamples, "No discrete samples");
			// Use of clamp-to-range means that this will clamp fScale to 1.0f if it goes past the end of the last bin
			float fIndex = Asura_Maths::ClampToRange<float>(fTime, 0.0f, 1.0f) * (m_uNumberOfDiscreteSamples - 1);
			uIndex = Asura_Maths::FloatToInt_Chop(fIndex);
			fScale = fIndex - uIndex;
		}

		inline T GetDiscreteValueByIndexAndScale(const u_int uIndex, const float fScale) const
		{
			return T( m_pxDiscreteSamples[uIndex] * (1.f - fScale) + m_pxDiscreteSamples[uIndex + 1] * fScale );
		}

		inline const T* GetDiscreteSamples() const
		{
			return m_pxDiscreteSamples;
		}
		// Non const version - used for nasty stuff like inserting a vector 2 graph into a vector 3 one. Should be avoided.
		inline T* GetDiscreteSamplesForEdit()
		{
			return m_pxDiscreteSamples;
		}

		T GetValueAtTime(float fTime) const
		{
			//Asura_Assert(m_pxSpline_Points, "No points in graph - something seriously broken here, as four points are created in the constructor.");
			if (GetControlPoints())
			{
				Asura_Graph_Point<T>* pxPoint = GetControlPoints();
				// Find last point before specified time
				while(pxPoint->m_pxNext && pxPoint->m_pxNext->m_fTime < fTime)
				{
					pxPoint = pxPoint->m_pxNext;
				}
				if(pxPoint->m_pxNext)
				{
					// Between two points

					float fProportion = (fTime - pxPoint->m_fTime) / (pxPoint->m_pxNext->m_fTime - pxPoint->m_fTime);

					if(!pxPoint->m_pxPrev || !pxPoint->m_pxNext->m_pxNext)
					{
						// Linear interpolation
						return (pxPoint->m_pxNext->m_xValue * fProportion) + pxPoint->m_xValue * (1.0f - fProportion);
					}

					// Catmull-Rom Spline
					float fProp2 = fProportion * fProportion;
					float fProp3 = fProportion * fProp2;

					T xValue0 = pxPoint->m_pxPrev->m_xValue;
					T xValue1 = pxPoint->m_xValue;
					T xValue2 = pxPoint->m_pxNext->m_xValue;
					T xValue3 = pxPoint->m_pxNext->m_pxNext->m_xValue;

					// No unary minus on Asura_Vector_n forces odd format
					T a = (                  xValue1 * 3.0f - xValue2 * 3.0f + xValue3 - xValue0);
					T b = (xValue0 *  2.0f - xValue1 * 5.0f + xValue2 * 4.0f - xValue3);
					T c = (xValue0 * -1.0f +                  xValue2);
					T xReturnValue = ((a * fProp3) + (b * fProp2) + (c * fProportion)) * 0.5f + xValue1;
					ClampValue(xReturnValue);
					return xReturnValue;
				}
				else
				{
					// Past last point. Just return final value
					return pxPoint->m_xValue;
				}
			}
			else
			{
				return GetValueAtTime_Discrete(fTime);
			}
		}
		
		// Basic accessor methods
		inline u_int GetNumDiscreteSamples() const { return m_uNumberOfDiscreteSamples; }
		void SetNumDiscreteSamples(u_int uNumDiscreteSamples)
		{
#ifdef ASURA_TOOLS
			if(uNumDiscreteSamples == m_uMaxNumberOfDiscreteSamples) return;
#else
			if(uNumDiscreteSamples == m_uNumberOfDiscreteSamples) return;
#endif
			if(m_pxDiscreteSamples)
			{
				delete [] m_pxDiscreteSamples;
			}
			m_pxDiscreteSamples = new T[uNumDiscreteSamples];
			m_uNumberOfDiscreteSamples = uNumDiscreteSamples;
#ifdef ASURA_TOOLS
			m_uMaxNumberOfDiscreteSamples = uNumDiscreteSamples;
#endif
			ReCalculateDiscreteSamples();
		}
		inline T GetDiscreteSample(u_int uIndex) const
		{
			Asura_Assert(uIndex < GetNumDiscreteSamples(), "Invalid index");
			return m_pxDiscreteSamples[uIndex];
		}
		inline Asura_Graph_Point<T>* GetControlPoints() const 
		{ 
			if (m_pxGraphSourceData)
			{
				return m_pxGraphSourceData->m_pxSpline_Points; 
			}
			return NULL;
		}

		T GetAverageValue() const
		{
			Asura_Assert(m_uNumberOfDiscreteSamples, "No samples");
			// (AW - 21/10/05) - Added code to handle case where there is only one sample.
			if (m_uNumberOfDiscreteSamples == 1)
			{
				return GetDiscreteSample(0);
			}
			// (AW - 21/10/05) - Average of the graph is not the same as the average of the points.
			// The first and last points only contribute 1/2 as much to the average as other points do.
			T xAverage = (GetDiscreteSample(0) + GetDiscreteSample(m_uNumberOfDiscreteSamples - 1)) * 0.5f;
			for (u_int i=1;i<m_uNumberOfDiscreteSamples-1;i++)
			{
				xAverage += GetDiscreteSample(i);
			}
			xAverage /= static_cast<float>(m_uNumberOfDiscreteSamples - 1);
			return xAverage;
		}

		T GetMinDiscreteValue() const
		{
			Asura_Assert( m_uNumberOfDiscreteSamples > 0, "No samples" );

			T tMin = GetDiscreteSample(0);

			for( u_int u = 1; u < m_uNumberOfDiscreteSamples; ++u )
			{
				tMin = Asura_Maths::Min( tMin, GetDiscreteSample(u) );
			}

			return tMin;
		}

		T GetMaxDiscreteValue() const
		{
			Asura_Assert( m_uNumberOfDiscreteSamples > 0, "No samples" );

			T tMax = GetDiscreteSample(0);

			for( u_int u = 1; u < m_uNumberOfDiscreteSamples; ++u )
			{
				tMax = Asura_Maths::Max( tMax, GetDiscreteSample(u) );
			}

			return tMax;
		}

		// Returns the area under the graph between m_fDomainMin and fEndX.
		T GetAreaUnderGraph( float fEndX ) const
		{
			// (AW - 12.07.2004) - Useful function if it is a speed/time graph or something. Doesn't
			// make any sense to use this on a colour or a vector. Uses the discrete values NOT the 
			// spline values.
			Asura_Assert(fEndX >= m_fDomainMin && fEndX <= m_fDomainMax, "End X out of range");
			// First check for the constant graph case.
			if(m_uNumberOfDiscreteSamples == 1)
			{
				return GetDiscreteSample(0) * (fEndX - m_fDomainMin);
			}
			const float fNumberOfSegments = static_cast<float>(m_uNumberOfDiscreteSamples - 1);
			const float fSegmentSize = (m_fDomainMax - m_fDomainMin) / fNumberOfSegments;
			T xArea = GetZeroElement();	
			// Iterate through the segments and add up the area of each.
			for (u_int i = 1; i < m_uNumberOfDiscreteSamples; ++i)
			{
				const float fX = m_fDomainMin + fSegmentSize * i;
				if (fX <= fEndX)
				{
					// Add entire area under segment
					const T xAverageValueDuringSegment = (GetDiscreteSample(i - 1) + GetDiscreteSample(i)) * 0.5f;
					xArea += xAverageValueDuringSegment * fSegmentSize;
				}
				else
				{
					// Add proportion of area under segment.
					const T xStartValue = GetDiscreteSample(i - 1);
					const T xEndValue = GetDiscreteSample(i);
                    // (NY - 11.04.2007) the gradient is (xEndValue - xStartValue) / fSegmentSize
					// since fSegmentSize = 1.0f / static_cast<float>(m_uNumberOfDiscreteSamples - 1),
					// we can ditch the divide
					const T xGradientOfSegment = (xEndValue - xStartValue) * fNumberOfSegments;
					const float fXInSegment = fEndX - fX + fSegmentSize;
					xArea += (xStartValue + xGradientOfSegment * fXInSegment * 0.5f) * fXInSegment;
					break;
				}
			}
			return xArea;
		}
		// Replaces discrete samples with the area under the graph
		// This will make it incompatible with the spline points, so it should only be done
		// when the graph will not be changed in future i.e., when loading in a non-Tools build.
		void IntegrateDiscreteSamples()
		{
#ifdef ASURA_TOOLS
			Asura_Assert(0, "Should not call this function in a tools build; discrete samples will no longer match spline points");
#endif
			T* pxNewSamples = new T[m_uNumberOfDiscreteSamples];
			// Get area at each sample
			for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
			{
				const float fProportion = static_cast< float >(u) / (m_uNumberOfDiscreteSamples - 1);
				const float fX = Asura_Maths::Lerp( m_fDomainMin, m_fDomainMax, fProportion );
				pxNewSamples[u] = GetAreaUnderGraph( fX );
			}
			// Replace samples
			for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
			{
				m_pxDiscreteSamples[u] = pxNewSamples[u];
			}
			delete [] pxNewSamples;
		}

		u_int GetNumberOfSplinePoints() const
		{
			u_int u = 0;
			Asura_Graph_Point<T>* pxPoint = GetControlPoints();
			while(pxPoint)
			{
				pxPoint = pxPoint->m_pxNext;
				++u;
			};
			return u;
		}

		// Adds a point without recalculating endpoints
		void AddPointInternal(T xValue, float fTime)
		{
			if(!GetControlPoints())
			{
				// Create the list
				if(!m_pxGraphSourceData)
				{
					m_pxGraphSourceData = new Asura_Graph_Source_Data<T>;
				}
				m_pxGraphSourceData->m_pxSpline_Points = new Asura_Graph_Point<T>(xValue, fTime);
			}
			else
			{
				Asura_Graph_Point<T>* pxPoint = GetControlPoints();
				// Find last point before specified X
				while(pxPoint->m_pxNext && pxPoint->m_pxNext->m_fTime <= fTime)
				{
					pxPoint = pxPoint->m_pxNext;
				}
				// Insert new point
				Asura_Graph_Point<T>* pxNewPoint = new Asura_Graph_Point<T>(xValue, fTime);
				// Link new point to next point
				pxNewPoint->m_pxNext = pxPoint->m_pxNext;
				if(pxPoint->m_pxNext)
				{
					pxPoint->m_pxNext->m_pxPrev = pxNewPoint;
				}
				// Link current point to new point
				pxPoint->m_pxNext = pxNewPoint;
				pxNewPoint->m_pxPrev = pxPoint;
			}
		}

		inline Asura_Graph& operator *= ( float fScale )
		{
			Asura_Graph_Point<T>* pxPoint = GetControlPoints()->m_pxNext;
			// Skip first and last
			while(pxPoint->m_pxNext)
			{
				pxPoint->m_xValue *= fScale;
				pxPoint = pxPoint->m_pxNext;
			}
#ifdef ASURA_TOOLS
			Update();
#else
			for(u_int uSample = 0; uSample < m_uNumberOfDiscreteSamples; ++uSample)
			{
				m_pxDiscreteSamples[uSample] *= fScale;
			}
#endif
			return *this;
		}
		// Returns true iff GetValueAtTime returns the same value for all times
		bool IsConstant() const
		{
			Asura_Graph_Point<T>*	pxPoint	= GetControlPoints();

			// Check for a constant graph
			if(pxPoint == NULL)
			{
				const T	xFirstValue	= GetValueAtTime_Discrete(m_fDomainMin);

				for(u_int u = 0; u < m_uNumberOfDiscreteSamples; u++)
				{
					if(m_pxDiscreteSamples[u] != xFirstValue)
					{
						return(false);
					}
				}
				return(true);
			}
			else
			{
				const T	xFirstValue	= GetValueAtTime(m_fDomainMin);

				while(pxPoint != NULL)
				{
					if(pxPoint->m_xValue == xFirstValue)
					{
						pxPoint = pxPoint->m_pxNext;
					}
					else
					{
						return(false);
					}
				}
				return(true);
			}
		}
		// Returns true if the graph is a straight line (i.e. has no spline points aside from the start and end)
		bool IsLinear() const
		{
			return (GetNumberOfSplinePoints() == 4);
		}

#ifdef ASURA_TOOLS
		Asura_Graph_Source_Data<T>* GetGraphSourceData() {return m_pxGraphSourceData;}
		const Asura_Graph_Source_Data<T>* GetGraphSourceData() const {return m_pxGraphSourceData;}
#endif

	private:
		void CopyFrom(const Asura_Graph& xOriginal)
		{
			//delete old data
			if(m_pxDiscreteSamples)
			{
				delete [] m_pxDiscreteSamples;
				m_pxDiscreteSamples = 0;
			}
			if(m_pxGraphSourceData)
			{
				delete m_pxGraphSourceData;
				m_pxGraphSourceData = 0;
			}
			//and copy new values
			m_uNumberOfDiscreteSamples = xOriginal.m_uNumberOfDiscreteSamples;

			m_fDomainMin = xOriginal.m_fDomainMin;
			m_fDomainMax = xOriginal.m_fDomainMax;

#ifdef ASURA_TOOLS
			m_uMaxNumberOfDiscreteSamples = xOriginal.m_uMaxNumberOfDiscreteSamples;
			m_fSampleTolerance = xOriginal.m_fSampleTolerance;
			m_bProportionalSampleTolerance = xOriginal.m_bProportionalSampleTolerance;
			m_pxDiscreteSamples = new T[m_uMaxNumberOfDiscreteSamples];
#else
			m_pxDiscreteSamples = new T[m_uNumberOfDiscreteSamples];
#endif

			if (xOriginal.m_pxGraphSourceData)
			{
				m_pxGraphSourceData = new Asura_Graph_Source_Data<T>;
				m_pxGraphSourceData->m_fClampMin = xOriginal.m_pxGraphSourceData->m_fClampMin;
				m_pxGraphSourceData->m_fClampMax = xOriginal.m_pxGraphSourceData->m_fClampMax;
				m_pxGraphSourceData->m_bUseClampMin = xOriginal.m_pxGraphSourceData->m_bUseClampMin;
				m_pxGraphSourceData->m_bUseClampMax = xOriginal.m_pxGraphSourceData->m_bUseClampMax;

				const Asura_Graph_Point<T>* pxPoint = xOriginal.GetControlPoints();
				while(pxPoint)
				{
					AddPointInternal(pxPoint->m_xValue, pxPoint->m_fTime);
					pxPoint = pxPoint->m_pxNext;
				}
			}
#ifdef ASURA_TOOLS
			Update();
#else
			for(u_int uSample = 0; uSample < m_uNumberOfDiscreteSamples; ++uSample)
			{
				m_pxDiscreteSamples[uSample] = xOriginal.m_pxDiscreteSamples[uSample];
			}
#endif
		}

		void ApplyClamp()
		{
			Asura_Assert(m_pxGraphSourceData, "Graph needs source data to set clamping values");
			if (GetNumberOfSplinePoints() >= 4)
			{
				Asura_Graph_Point<T>* pxPoint = GetControlPoints()->m_pxNext;
				while (pxPoint->m_pxNext)
				{
					ClampValue(pxPoint->m_xValue);
					pxPoint = pxPoint->m_pxNext;
				}
			}
		}

		T GetZeroElement() const;
		void ClampValue(T &xValue) const
		{
			Asura_Assert(m_pxGraphSourceData, "Graph needs source data to set clamping values");
			if (m_pxGraphSourceData->m_bUseClampMin)
			{
				ClampToMinimum(xValue, m_pxGraphSourceData->m_fClampMin);
			}
			if (m_pxGraphSourceData->m_bUseClampMax)
			{
				ClampToMaximum(xValue, m_pxGraphSourceData->m_fClampMax);
			}
		}
		void ClampToMinimum(T& xValue, float fClampMinimum) const;
		void ClampToMaximum(T& xValue, float fClampMaximum) const;

#ifdef ASURA_TOOLS
		//get min/max of 2 values , considering each component seperately
		void GetComponentWiseMin(const T& xVal1,const T& xVal2,T& xOutput) const;
		void GetComponentWiseMax(const T& xVal1,const T& xVal2,T& xOutput) const;

		//returns true if the difference between the 2 values is less than the tolerance
		bool WithinTolerance(const T& xVal1,const T& xVal2,const T& xTol) const;

		//sets all the components in the value to fSet
		void SetComponents(T& xVal,float fSet) const;
#endif

		void RemovePointInternal(Asura_Graph_Point<T>* pxPointToDelete)
		{
			Asura_Assert(pxPointToDelete->m_pxNext && pxPointToDelete->m_pxPrev, "Should not delete end control points.");
			if(!pxPointToDelete->m_pxPrev->m_pxPrev || !pxPointToDelete->m_pxNext->m_pxNext)
			{
				// Don't delete first or last points
				// Don't assert either though, as the graph gadget will allow this action
				return;
			}
			Asura_Graph_Point<T>* pxPoint = GetControlPoints();
			while(pxPointToDelete != pxPoint)
			{
				pxPoint = pxPoint->m_pxNext;
				if(!pxPoint)
				{
					Asura_Assert(0, "Point to delete not found.");
					return;
				}
			}
			pxPoint->m_pxPrev->m_pxNext = pxPoint->m_pxNext;
			pxPoint->m_pxNext->m_pxPrev = pxPoint->m_pxPrev;
			delete pxPoint;
		}

		void ReCalculateEndPoints()
		{
			Asura_Assert(GetControlPoints(), "No points in graph - something seriously broken here, as four points are created in the constructor.");
			if(!GetControlPoints()) return;
			// Set first value
			GetControlPoints()->m_xValue = GetControlPoints()->m_pxNext->m_xValue * 2.0f - GetControlPoints()->m_pxNext->m_pxNext->m_xValue;
			// Find last point
			Asura_Graph_Point<T>* pxPoint = GetControlPoints();
			while(pxPoint->m_pxNext)
			{
				pxPoint = pxPoint->m_pxNext;
			}
			// Set last value
			pxPoint->m_xValue = pxPoint->m_pxPrev->m_xValue * 2.0f - pxPoint->m_pxPrev->m_pxPrev->m_xValue;
		}

		void ReCalculateDiscreteSamples()
		{
// (MP - 27.04.2005) There's not really any reason why you shouldn't be able
// to have graphs that are intended solely for continuous sampling.
//			Asura_Assert(m_uNumberOfDiscreteSamples, "No discrete samples");

#ifdef ASURA_TOOLS
			m_uNumberOfDiscreteSamples = m_uMaxNumberOfDiscreteSamples;
			if(m_fSampleTolerance>0.0f && m_uMaxNumberOfDiscreteSamples>2)
			{
				T xTol;
				if(m_bProportionalSampleTolerance)
				{
					//get the minimum and maximum values , so we can determine a suitable tolerance level
					T xMin = GetValueAtTime(0.0f);
					T xMax = xMin;

					// Skip end control points
					const Asura_Graph_Point<T>* pxPoint = GetControlPoints()->m_pxNext;
					while(pxPoint->m_pxNext)
					{
						GetComponentWiseMin(xMin,pxPoint->m_xValue,xMin);
						GetComponentWiseMax(xMax,pxPoint->m_xValue,xMax);

						pxPoint = pxPoint->m_pxNext;
					}
					//if there are clamps, include those as well
					float fClamp;
					if(GetClampMin(fClamp))
					{
						SetComponents(xMin,fClamp);
					}
					if(GetClampMax(fClamp))
					{
						SetComponents(xMax,fClamp);
					}

					//use the tolerance multiplier on the range to determine the overall tolerance
					xTol = (xMax-xMin) * m_fSampleTolerance;
				}
				else
				{
					SetComponents(xTol, m_fSampleTolerance);
				}
				
				//Try each number of samples to determine how few we can get away with
				//A bit ugly ,but it doesn't seem to be so slow as to be a big problem.
				for(m_uNumberOfDiscreteSamples=2;m_uNumberOfDiscreteSamples<=m_uMaxNumberOfDiscreteSamples;m_uNumberOfDiscreteSamples++)
				{
					//calculate samples
					for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
					{
						const float fProportion = static_cast< float >(u) / (m_uNumberOfDiscreteSamples - 1);
						const float fX = Asura_Maths::Lerp( m_fDomainMin, m_fDomainMax, fProportion );
						m_pxDiscreteSamples[u] = GetValueAtTime( fX );
					}

					if(m_uNumberOfDiscreteSamples==m_uMaxNumberOfDiscreteSamples) break;

					//check to see if any error is too large
					bool bOkay = true;
					for(u_int u = 1; u < m_uMaxNumberOfDiscreteSamples-1; ++u)
					{
						const float fProportion = static_cast< float >(u) / (m_uMaxNumberOfDiscreteSamples - 1);
						const float fX = Asura_Maths::Lerp( m_fDomainMin, m_fDomainMax, fProportion );

						if(!WithinTolerance(GetValueAtTime(fX),GetValueAtTime_Discrete(fX),xTol))
						{
							bOkay = false;
							break;
						}
					}
					if(bOkay) break;
				}
			}
			else
#endif // ASURA_TOOLS
			{
				if(m_uNumberOfDiscreteSamples == 1)
				{
					m_pxDiscreteSamples[0] = GetValueAtTime(0.f);
					return;
				}
            
				for(u_int u = 0; u < m_uNumberOfDiscreteSamples; ++u)
				{
					const float fProportion = static_cast< float >(u) / (m_uNumberOfDiscreteSamples - 1);
					const float fX = Asura_Maths::Lerp( m_fDomainMin, m_fDomainMax, fProportion );
					m_pxDiscreteSamples[u] = GetValueAtTime( fX );
				}
			}
		}
	


		// Data for creating discrete points. (Spline points, min and max clamps)
		// Generally null if not in tools.
		Asura_Graph_Source_Data<T>*	m_pxGraphSourceData;
#endif // else not #ifndef ASURA_GRAPH_LEAN_AND_MEAN

		// Discrete samples
		// Dependent on spline points in tools
		unsigned int			m_uNumberOfDiscreteSamples;
		T*						m_pxDiscreteSamples;

#ifdef ASURA_TOOLS
		float					m_fSampleTolerance;//tolerance level for determining number of samples to generate
		bool					m_bProportionalSampleTolerance;//if true tolerance is a proportion of the range covered , otherwise it is an absolute value
		unsigned int			m_uMaxNumberOfDiscreteSamples;//the array size for m_pxDiscreteSamples
#endif //ASURA_TOOLS

		float					m_fDomainMin;
		float					m_fDomainMax;


};

#endif//__ASURA_GRAPH_H__
