//////////////////////////////////////////////////////////////////////////////
//
// Asura_HandleTreeTemplate.h
//
// Created by Mike Healey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// template definition for generic binary tree class for handle maintainance.
// based on Asura_NameTree which does an equivalent job for filenames.
// Template parameter could be enum (for sound IDs) string (for texture names,
// if using our string class >, <, = operators was OK) or anything else 
// supporting those operators.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_HANDLETREETEMPLATE_H___
#define ___ASURA_HANDLETREETEMPLATE_H___ 1

#include <stddef.h>         // for NULL
#include "Core/Asura_Handle.h"
#include "Debugging/Asura_Assert.h"

#include "Core/Asura_LinkList.h"

#if defined( _DEBUG ) || defined( ASURA_TOOLS )
#define ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH		(1)
#endif // #if defined( _DEBUG ) || defined( ASURA_TOOLS )

// JAT 18/09/06 - give an assert if recursing more than this limit
#if !defined( ASURA_HANDLETREETEMAPLTE_MAX_STACK_DEPTH )
#define ASURA_HANDLETREETEMAPLTE_MAX_STACK_DEPTH 4000
#endif // #if !defined( ASURA_HANDLETREETEMAPLTE_MAX_STACK_DEPTH )


template <class T> class Asura_HandleTree_Node
{
	public:
 
		T Value;
		int Index;
		Asura_HandleTree_Node *Lower;
		Asura_HandleTree_Node *Higher;

		Asura_HandleTree_Node(const T & _value, const int __index)
		{
			Lower = NULL;
			Higher = NULL;	

			Value = _value;
			Index = __index;
		}
		~Asura_HandleTree_Node()
		{
		}
		
	    inline bool operator == ( const Asura_HandleTree_Node& v) const
		{
		    return (Value == v.Value);
		}

	    inline bool operator > ( const Asura_HandleTree_Node& v) const
		{
		    return (Value > v.Value);
		}

	    inline bool operator < ( const Asura_HandleTree_Node& v) const
		{
            return (Value < v.Value);
        }
};

template <class T> class Asura_HandleTree
{
		
		Asura_HandleTree_Node<T>* Root;
		int NumberOfNodes;
        int MaxNodes;
        Asura_Handle_List* HandleList;

		//MKR 15/09/06: for doing ascending order non recursive traversal - used by Balance()
		struct Recurse_Level
		{
			Recurse_Level() {}
			Recurse_Level(Asura_HandleTree_Node<T>* pxNode) { m_pxNode = pxNode; m_eStage = LOWER; }
			enum Recurse_Level_Stage
			{
				LOWER,
				INNER,
				UPPER,
				MAX
			};
			Asura_HandleTree_Node<T>* m_pxNode;
			Recurse_Level_Stage m_eStage;
		};	
        
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
		// JAT 18/09/06 - track recursion depth and try to trap problems
		u_int m_uStackDepth;
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)

#if defined( ASURA_TOOLS )
		// DLH - 07 September 2006 12:40:37 - Records the maximum depth that the tree gets to. Used for Debugging only.
        u_int m_uDepthMax;
public:
	u_int GetMaxDepth() { return m_uDepthMax; }
#endif

	public:

		Asura_HandleTree(int maximumNodes)
		{
#if defined( ASURA_TOOLS)
			this->m_uDepthMax = 0;
#endif

			Root = NULL;
		    NumberOfNodes = 0;
            MaxNodes = maximumNodes;
            HandleList = new Asura_Handle_List(maximumNodes);
		}
		~Asura_HandleTree()
		{
			FreeAll();
            delete HandleList;
		}

        int GetNumberOfNodes() const { return NumberOfNodes; }
        int GetMaxNodes() const { return MaxNodes; }
        const Asura_Handle_List* GetHandleList() const { return HandleList; }
        
		// KJL 08/03/00 17:08:16 - returns true if the node existed already
		bool FindNodeIndex(const T _value, const bool __addToTreeIfRequired, int * __index)
		{		
            if (!Root)
            {
                if (__addToTreeIfRequired)
                {
                    Asura_Assert(__index, "index should not be NULL if adding");
                    Asura_Handle handle;
					#ifdef ASURA_ASSERT
					bool success = 
					#endif
					HandleList->GetNewHandle(handle);
                    Asura_Assert(success, "No more space!");
				    Root = new Asura_HandleTree_Node<T>(_value, handle);
				    NumberOfNodes++;
				    *__index = handle;
				    
#if defined( ASURA_TOOLS)
					this->m_uDepthMax = 1;
#endif
                }
                return false;
            }
            else
			{
#if defined( ASURA_TOOLS)
				u_int uDepth = 1;
#endif
			    Asura_HandleTree_Node<T> *nodePtr = Root;
			    
			    while (nodePtr->Value != _value)
			    {
				    while (nodePtr->Value < _value)
					{
#if defined( ASURA_TOOLS)
						uDepth += 1;//MKR 15/09/06: moved this out of the lower if statement
#endif
					    if (nodePtr->Higher)
					    {
						    nodePtr = nodePtr->Higher;

						    continue;
					    }
					    else if (__addToTreeIfRequired)
					    {
                            Asura_Assert(__index, "index should not be NULL if adding");
                            Asura_Handle handle;
							#ifdef ASURA_ASSERT
							bool success = 
							#endif
							HandleList->GetNewHandle(handle);
                            Asura_Assert(success, "No more space!");
						    nodePtr->Higher = new Asura_HandleTree_Node<T>(_value, handle);
					        NumberOfNodes++;
						    *__index = handle;
#if defined( ASURA_TOOLS)
							this->m_uDepthMax = Asura_Maths::Max( this->m_uDepthMax, uDepth);
#endif
					    }
					    return false;
				    }
				    while (nodePtr->Value > _value)
					{
#if defined( ASURA_TOOLS)
						uDepth += 1;//MKR 15/09/06: moved this out of the lower if statement
#endif
					    if (nodePtr->Lower)
					    {
						    nodePtr = nodePtr->Lower;

						    continue;
					    }
					    else if (__addToTreeIfRequired)
					    {
                            Asura_Assert(__index, "index should not be NULL if adding");
                            Asura_Handle handle;
							#ifdef ASURA_ASSERT
							bool success = 
							#endif
							HandleList->GetNewHandle(handle);
                            Asura_Assert(success, "No more space!");
						    nodePtr->Lower = new Asura_HandleTree_Node<T>(_value, handle);
					        NumberOfNodes++;
						    *__index = handle;
#if defined( ASURA_TOOLS)
							this->m_uDepthMax = Asura_Maths::Max( this->m_uDepthMax, uDepth);
#endif
					    }
					    return false;
				    }

			    }
			    if (__index)
                {
                    *__index = nodePtr->Index;
                }
			    return true;
            }
		}

		Asura_Handle AddNode(const T _value)
		{
			Asura_Handle	iHandle;

			if(FindNodeIndex(_value, true, &iHandle))
			{
				return(ASURA_INVALID_HANDLE);
			}
			else
			{
				return(iHandle);
			}
		}

        bool DeleteNode(const T _value)
        {
            if (!Root)
            {
                return false;
            }
            else
            {
			    Asura_HandleTree_Node<T> *nodePtr = Root;
			    Asura_HandleTree_Node<T> **nodePtrPtr = &Root;
			    
			    while (nodePtr->Value != _value)
			    {
				    while (nodePtr->Value < _value)
				    {
					    if (nodePtr->Higher)
					    {
						    nodePtrPtr = &nodePtr->Higher;
					        nodePtr = *nodePtrPtr;
						    continue;
					    }
					    return false;
				    }
				    while (nodePtr->Value > _value)
				    {
					    if (nodePtr->Lower)
					    {
					        nodePtrPtr = &nodePtr->Lower;
					        nodePtr = *nodePtrPtr;
						    continue;
					    }
					    return false;
				    }
			    }

			    Asura_HandleTree_Node<T> *nodeToDelete = nodePtr;
			    //found the node to delete , if it has children they will need to be transferred up
			    if(nodeToDelete->Lower && nodeToDelete->Higher)
			    {
				    //The deleted node has 2 children  , 1 will need to be found a place beneath the other
				    //Pick semi-randomly to keep some chance of things not being unbalanced
				    Asura_HandleTree_Node<T>* nodeToTransfer;
				    if(NumberOfNodes%2)
				    {
					    *nodePtrPtr = nodeToDelete->Lower;
					    nodeToTransfer = nodeToDelete->Higher;
				    }
				    else
				    {
					    *nodePtrPtr = nodeToDelete->Higher;
					    nodeToTransfer = nodeToDelete->Lower;
				    }
				    nodePtr = *nodePtrPtr;

				    //search down for somewhere to stick the transferred node
				    TryNextNode :
				    
				    if(nodePtr < nodeToTransfer)
				    {
					    if(nodePtr->Higher)
					    {
						    nodePtr = nodePtr->Higher;
						    goto TryNextNode;
					    }
					    else
					    {
						    nodePtr->Higher = nodeToTransfer;
					    }
				    }
				    else
				    {
					    if(nodePtr->Lower)
					    {
						    nodePtr = nodePtr->Lower;
						    goto TryNextNode;
					    }
					    else
					    {
						    nodePtr->Lower = nodeToTransfer;
					    }
				    }
			    }
			    else if(nodeToDelete->Lower)
			    {
				    *nodePtrPtr = nodeToDelete->Lower;
			    }
			    else
			    {
				    *nodePtrPtr = nodeToDelete->Higher;
			    }

			    // return handle
                HandleList->ReleaseHandle(nodeToDelete->Index);

			    //get rid of the node ,and update the counters
			    NumberOfNodes--;
			    delete nodeToDelete;

			    return true;
            }
        }

		void FreeAll(void)
		{
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			// JAT 18/09/06 - track recursion depth and try to trap problems
			m_uStackDepth = 0;
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			if (Root)
            {
                FreeAllRecursive(Root);
                Root = NULL;
            }
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			Asura_Assert( m_uStackDepth == 0, "Problem tracking stack depth" );
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
		}

		//MKR 15/09/06: the return value of the callback is not used anywhere - is this a relic?
        void TraverseFromRoot(bool (*function)(const T _value))
        {
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			// JAT 18/09/06 - track recursion depth and try to trap problems
			m_uStackDepth = 0;
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
            if (Root)
            {
                Traverse(Root, function);
            }
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			Asura_Assert( m_uStackDepth == 0, "Problem tracking stack depth" );
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
        }

private:
        
        void FreeAllRecursive(Asura_HandleTree_Node<T> *nodePtr)
		{
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			m_uStackDepth++;
			Asura_Assert( m_uStackDepth < ASURA_HANDLETREETEMAPLTE_MAX_STACK_DEPTH, "Too much recursion stack likely to be corrupted" );
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			if (nodePtr->Lower)
			{
				FreeAllRecursive(nodePtr->Lower);
			}
			if (nodePtr->Higher)
			{
				FreeAllRecursive(nodePtr->Higher);
			}

			// return handle
            HandleList->ReleaseHandle(nodePtr->Index);
			delete nodePtr;
            NumberOfNodes--;
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			m_uStackDepth--;
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
		}

		//MKR 15/09/06: this is not used anywhere
		void Traverse(Asura_HandleTree_Node<T> *nodePtr, void (*function)(T _value, int _index))
		{
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			m_uStackDepth++;
			Asura_Assert( m_uStackDepth < ASURA_HANDLETREETEMAPLTE_MAX_STACK_DEPTH, "Too much recursion stack likely to be corrupted" );
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			if (nodePtr->Lower)
			{
				Traverse(nodePtr->Lower, function);
			}
			if (nodePtr->Higher)
			{
				Traverse(nodePtr->Higher, function);
			}

			if (nodePtr!=Root)//MKR 15/09/06: this if statement shouldnt be here because this method is private so the user cant get hold of Root
			{
				function(nodePtr->Value, nodePtr->Index);
			}
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			m_uStackDepth--;
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
		}

        void Traverse(Asura_HandleTree_Node<T> *nodePtr, bool (*function)(const T _value))
		{
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			m_uStackDepth++;
			Asura_Assert( m_uStackDepth < ASURA_HANDLETREETEMAPLTE_MAX_STACK_DEPTH, "Too much recursion stack likely to be corrupted" );
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			if (nodePtr->Lower)
			{
				Traverse(nodePtr->Lower, function);
			}
			if (nodePtr->Higher)
			{
				Traverse(nodePtr->Higher, function);
			}

			function(nodePtr->Value);
#if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
			m_uStackDepth--;
#endif // #if defined( ASURA_HANDLETREETEMPLATE_TRACK_STACK_DEPTH	)
		}
public:
	// DLH - 07 September 2006 12:41:52 - This is a dynamic version of the above code. 
	// DLH - 07 September 2006 12:42:13 - If a large amount of sequential items are added to the tree then the recursive verion runs out of stack
	// DLH - 07 September 2006 12:43:01 - Original done to fix ShellShock2 6000+ sequential Asura_Sound_ID
	// MKR 15/09/06: this does NOT work in ascending order - if you want that then copy Balance()
		void TraverseFromRootNonRecursive(bool (*function)(const T _value))
		{
			if(!Root)
			{
				return;
			}

			// DLH - 07 September 2006 12:36:07 - List of nodes that are to have their node's data processed by function in the lists order
			Asura_LinkList< const Asura_HandleTree_Node<T>*> xList_pxNode_ToRunFunction;
			// DLH - 07 September 2006 12:36:47 - The nodes that have not been processed yet
			Asura_LinkList< const Asura_HandleTree_Node<T>*> xList_pxNode_ToProcess;

			// DLH - 07 September 2006 12:37:05 - Start with the present root node
			xList_pxNode_ToProcess.Push_Front( Root);

			while( NULL != xList_pxNode_ToProcess.Front())
			{
				// DLH - 07 September 2006 12:37:27 - Get the first item on the process list as it is the deepest lowest unprocessed node. Remove it from the list
				Asura_ListElt< const Asura_HandleTree_Node<T>*>* pxItNodeProcess = xList_pxNode_ToProcess.Front();

				const Asura_HandleTree_Node<T>* pxNode = pxItNodeProcess->Data();

				xList_pxNode_ToProcess.Remove( pxItNodeProcess);

				xList_pxNode_ToRunFunction.Push_Front( pxNode);

				// DLH - 07 September 2006 12:38:42 - Add children nodes to be processed. Highest is added first so it is processed after the lowest node
				{
					const Asura_HandleTree_Node<T>* pxNodeHigher = pxNode->Higher;
					if( NULL != pxNodeHigher)				
					{
						xList_pxNode_ToProcess.Push_Front( pxNodeHigher);
					}
				}

				{
					const Asura_HandleTree_Node<T>* pxNodeLower = pxNode->Lower;
					if( NULL != pxNodeLower)				
					{
						xList_pxNode_ToProcess.Push_Front( pxNodeLower);
					}
				}
			}

			// DLH - 07 September 2006 12:39:52 - run the function on each node in the list
			for( Asura_ListElt< const Asura_HandleTree_Node<T>*>* pxItNodeFuncation = xList_pxNode_ToRunFunction.Front(); NULL != pxItNodeFuncation; pxItNodeFuncation = pxItNodeFuncation->Next())
			{
				const Asura_HandleTree_Node<T>* pxNode = pxItNodeFuncation->Data();

				function( pxNode->Value);
			}
		}		

private:
		Asura_HandleTree_Node<T>* Balance_BuildTreeFromArray(Asura_HandleTree_Node<T>** axNodeArray, 
			int iLower, int iHigher, u_int uiDepth)
		{
#ifdef ASURA_TOOLS
			if(uiDepth > m_uDepthMax)
				m_uDepthMax = uiDepth;
#endif
			int iMiddle = (iLower + iHigher) / 2;
			Asura_HandleTree_Node<T>* pxMiddle = axNodeArray[iMiddle];

			if(iMiddle > iLower)
				pxMiddle->Lower = Balance_BuildTreeFromArray(axNodeArray, iLower, iMiddle - 1, uiDepth + 1);
			else
				pxMiddle->Lower = NULL;
			if(iHigher > iMiddle)
				pxMiddle->Higher = Balance_BuildTreeFromArray(axNodeArray, iMiddle + 1, iHigher, uiDepth + 1);
			else
				pxMiddle->Higher = NULL;

			return pxMiddle;
		}		
public:
		//MKR 12/09/06: Balance() balances the binary tree
		void Balance()
		{
#ifdef ASURA_TOOLS		
			m_uDepthMax = 0;
#endif
			if(!Root)
				return;

			Asura_HandleTree_Node<T>** axTreeArray = new Asura_HandleTree_Node<T>*[NumberOfNodes];
			int iNodeArrayIndex = 0;

			Asura_LinkList<Recurse_Level> xStack;
			xStack.Push_Front(Recurse_Level(Root));

			do
			{
				Asura_ListElt<Recurse_Level>* pxListFront = xStack.Front();			
				Recurse_Level& xStackTopLevel = pxListFront->Data();

				switch(xStackTopLevel.m_eStage)
				{
				case Recurse_Level::LOWER:
					if(xStackTopLevel.m_pxNode->Lower)
						xStack.Push_Front(Recurse_Level(xStackTopLevel.m_pxNode->Lower));
					xStackTopLevel.m_eStage = Recurse_Level::INNER;
					break;
				case Recurse_Level::INNER:
					axTreeArray[iNodeArrayIndex++] = xStackTopLevel.m_pxNode;
					xStackTopLevel.m_eStage = Recurse_Level::UPPER;
					break;
				case Recurse_Level::UPPER:
					if(xStackTopLevel.m_pxNode->Higher)
						xStack.Push_Front(Recurse_Level(xStackTopLevel.m_pxNode->Higher));
					xStackTopLevel.m_eStage = Recurse_Level::MAX;
					break;
				case Recurse_Level::MAX:		
					xStack.Remove(pxListFront);
					break;
				};			
			}while(xStack.Front());

			Root = Balance_BuildTreeFromArray(axTreeArray, 0, NumberOfNodes - 1, 1);

			delete[] axTreeArray;
		}


};
// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_HANDLETREETEMPLATE_H___
