///////////////////////////////////////////////////////////////////////////////
//
// Asura_HashTable.h
//
// Created by KJFL - 2004-07-12 17:30:13
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// MGH 2007-10-17: The old class in this file was renamed from Asura_HashTable
// to Tools_HashTable, in order to make space for a shiny new hash table class
// that has more functionality.
// MGH 2007-10-18: New class added.
//
//
// Description:
//
//	Hash table implementation, with an interface inteneded to be compatible
//  with Asura_List.
//
//  The class T needs to provide a hash key (of type Asura_Hash_ID or some
//  type that can be promoted to it) by implementing GetHashID(). This
//  doesn't need to be unique; duplicate keys are allowed. If duplicate
//  entries are not desirable then they can be suppressed by using
//  AddUniqueObject() or manually by using one of the find methods first.
//
//  If the AddUniqueObject(), FindObject() or RemoveObject() methods are
//  required (ie. any manipulation except via hash key) then it also needs to
//  provide an equality test by implementing operator==, which is used to
//  search for objects stored in the table. Entries that are 'equal' should
//  have the same hash key.
//
//  Optionally a memory pool or similar can be used to supply memory for nodes.
//
//  The table size passed in must be a power of 2.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_HASHTABLE_H__
#define __ASURA_HASHTABLE_H__

// Includes
#include "Debugging/Asura_Assert.h"
#include "Core/Asura_Maths.h"

// Forward declarations
template< typename T > class Asura_HashTable_It;
template< typename T > class Asura_HashTable_ConstIt;

// Typedefs
typedef void*( Asura_HashTable_MemSupplyFuncPtr() );
typedef void( Asura_HashTable_MemReturnFuncPtr( void* ) );


// Node
template< typename T >
class Asura_HashTable_Node
{
public:
    T* m_pxObject;
    Asura_HashTable_Node< T >* m_pxPrev;
    Asura_HashTable_Node< T >* m_pxNext;
};


// Hash Table
template< typename T >
class Asura_HashTable
{
	// Friends
    friend class Asura_HashTable_It< T >;
	friend class Asura_HashTable_ConstIt< T >;

	// Typedefs
	typedef Asura_HashTable_Node< T > Node;

public:
    Asura_HashTable< T >(u_int uTableSize)
    : m_ppxHead( 0 )
	, m_uTableSize( uTableSize )
	, m_uMask( 0 )
    , m_iNumEntries( 0 )
    {
		// Must be a power of two, so we can just mask.
		Asura_Assert( m_uTableSize == (1u << Asura_Maths::GetCeilLog2(m_uTableSize)), "Table size is not power of two!" );

		m_ppxHead = new Node*[m_uTableSize];

		for (u_int u=0; u<m_uTableSize; ++u)
		{
			m_ppxHead[u] = 0;
		}

		m_uMask = m_uTableSize - 1;
    }

    ~Asura_HashTable< T >()
	{
		Clear();
		delete [] m_ppxHead;
	}

    // Provides functionality for supplying your own memory for the nodes.
    // Get the memory element size and create your mempool or whatever.
    // Implement the supply and return functions to supply and return memory from your storage.
    // Obviously you must either provide both of these or neither, otherwise things go bad.
    static size_t GetMemoryElementSize() { return sizeof( Node ); }
    static void SetMemorySupplyFunction( Asura_HashTable_MemSupplyFuncPtr pfnMemSupplyFuncPtr )
	{
		s_pMemSupplyFuncPtr = pfnMemSupplyFuncPtr;
	}
    static void SetMemoryReturnFunction( Asura_HashTable_MemReturnFuncPtr pfnMemReturnFuncPtr )
	{
		s_pMemReturnFuncPtr = pfnMemReturnFuncPtr; 
	}

	// Add a new entry (no test is made to check whether it is a duplicate).
    void AddObject( T* pxObject )
    {
		Asura_Assert(pxObject, "Null pointer added");
		if (!pxObject) return;

        Node* pxNewNode = NULL;
        if ( s_pMemSupplyFuncPtr == NULL )
        {
            pxNewNode = new Node;
        }
        else
        {
            pxNewNode = static_cast< Node* >( s_pMemSupplyFuncPtr() );
        }

		const u_int uRow = GetRow(pxObject);

        pxNewNode->m_pxObject = pxObject;
        pxNewNode->m_pxNext   = m_ppxHead[uRow];
        pxNewNode->m_pxPrev   = 0;

        if ( pxNewNode->m_pxNext )
        {
            pxNewNode->m_pxNext->m_pxPrev = pxNewNode;
        }

        m_ppxHead[uRow] = pxNewNode;

        ++m_iNumEntries;
    }

	// Attempts to insert the specific data into the table. If identical data
	// already exists then the table is not modified and a pointer to that other
	// data is returned. Otherwise returns the original input pointer,
	// on successfully inserting new distinct data.
	T* AddUniqueObject( T* pxObject, bool bAssertAboutDuplicates )
	{
		Asura_Assert(pxObject, "Null pointer inserted");
		if (pxObject)
		{
			Node* pxNode = FindNode(*pxObject);
			if (pxNode)
			{
                Asura_Assert( !bAssertAboutDuplicates, "Duplicate entry!" );
				return pxNode->m_pxObject; 
			}

			AddObject( pxObject );
		}

		return pxObject;
	}

	// Using the equality test (operator==) for T, search for the relevant object.
	T* FindObject( const T& xTestObject )
	{
		const Node* pxNode = FindNode(xTestObject);
		return( pxNode ? pxNode->m_pxObject : NULL );
	}
	const T* FindObject( const T& xTestObject ) const
	{
		const Node* pxNode = FindNode(xTestObject);
		return( pxNode ? pxNode->m_pxObject : NULL );
	}

    // Find by hash ID alone, returning the first match.
	T* FindObjectByHashID( const Asura_Hash_ID uHashID )
	{
		const Node* pxNode = FindNodeByHashID( uHashID );
		return( pxNode ? pxNode->m_pxObject : NULL );
	}
	const T* FindObjectByHashID( const Asura_Hash_ID uHashID ) const
	{
		const Node* pxNode = FindNodeByHashID( uHashID );
		return( pxNode ? pxNode->m_pxObject : NULL );
	}

	// Using the equality test (operator==) for T, search for the relevant object.
	// If found then remove the corresponding node and return the object,
	// otherwise just return null.
	T* RemoveObject( const T& xTestObject )
	{
		Node* pxNode = FindNode(xTestObject);
		if (pxNode)
		{
			T* const pxObject = pxNode->m_pxObject;
			RemoveNodeFromRow(pxNode, GetRow(pxObject));
			return pxObject;
		}

		return NULL;
	}

    // Find by hash ID alone, removing (and returning) the first match.
	T* RemoveObjectByHashID( const Asura_Hash_ID uHashID )
	{
		Node* pxNode = FindNodeByHashID( uHashID );
		if (pxNode)
		{
			T* const pxObject = pxNode->m_pxObject;
			RemoveNodeFromRow(pxNode, GetRow(pxObject));
			return pxObject;
		}

		return NULL;
	}

	// Clear out all the entries, freeing up node memory (but not object memory).
    void Clear()
    {
		for (u_int u=0; u<m_uTableSize; ++u)
		{
			ClearRow(u);
		}

		Asura_Assert(m_iNumEntries == 0, "Unexpected number of entries");
        m_iNumEntries = 0;
    }

	bool IsEmpty() const { return (m_iNumEntries == 0); }

    int GetNumEntries() const { return m_iNumEntries; }

	u_int GetTableSize() const { return m_uTableSize; }

protected:

    inline u_int GetRowForHashID( const Asura_Hash_ID uHashID ) const
	{
		return( uHashID & m_uMask );
	}

	// Get row of hash table for given object.
	inline u_int GetRow(const T* pxObject) const
	{
		return GetRowForHashID( pxObject->GetHashID() );
	}

	// Clear a row of the hash table.
	void ClearRow(const u_int uRow)
	{
		Node* pxNode = m_ppxHead[uRow];

        while ( pxNode )
        {
            Node* pxToDelete = pxNode;

            pxNode = pxNode->m_pxNext;

			FreeNode(pxToDelete);
			--m_iNumEntries;
        }
	
		m_ppxHead[uRow] = 0;
	}

	// Free up the memory associated with a node.
	void FreeNode(Node* pxToDelete)
	{
		if ( s_pMemReturnFuncPtr == NULL )
        {
            delete pxToDelete;
        }
        else
        {
            s_pMemReturnFuncPtr( pxToDelete );
        }
	}

	// Remove the specified node.
	void RemoveNodeFromRow(Node* pxNode, u_int uRow)
	{
		if (pxNode)
		{
			if ( pxNode->m_pxNext )
            {
                pxNode->m_pxNext->m_pxPrev = pxNode->m_pxPrev;
            }
            
            if ( pxNode->m_pxPrev )
            {
                pxNode->m_pxPrev->m_pxNext = pxNode->m_pxNext;
            }
            else
            {
				Asura_Assert(m_ppxHead[uRow] == pxNode, "Error: Inconsistent head node pointer.");
				m_ppxHead[uRow] = pxNode->m_pxNext;
            }

			FreeNode(pxNode);
            --m_iNumEntries;
		}
	}

	// Using the equality test (operator==) for T, search for the relevant node.
	Node* FindNode( const T& xTestObject ) const
    {
		const u_int uRow = GetRow(&xTestObject);
        Node* pxNode = m_ppxHead[uRow];

        while ( pxNode )
        {
            if (*(pxNode->m_pxObject) == xTestObject )
            {
                return pxNode;
            }

            pxNode = pxNode->m_pxNext;
        }

        return NULL;
    }

	Node* FindNodeByHashID( const Asura_Hash_ID uHashID ) const
    {
		const u_int uRow = GetRowForHashID( uHashID );
        Node* pxNode = m_ppxHead[uRow];

        while( pxNode )
        {
            if( pxNode->m_pxObject->GetHashID() == uHashID )
            {
                return pxNode;
            }

            pxNode = pxNode->m_pxNext;
        }

        return NULL;
    }

private:
	// Disallow copying of this class.
	Asura_HashTable(const Asura_HashTable& xOther);
	Asura_HashTable& operator=(const Asura_HashTable& xOther);

protected:
    Node** m_ppxHead;

	u_int m_uTableSize;
	u_int m_uMask;

 	int m_iNumEntries;

	static Asura_HashTable_MemSupplyFuncPtr* s_pMemSupplyFuncPtr;
    static Asura_HashTable_MemReturnFuncPtr* s_pMemReturnFuncPtr;
};

template< typename T > Asura_HashTable_MemSupplyFuncPtr* Asura_HashTable< T >::s_pMemSupplyFuncPtr = NULL;
template< typename T > Asura_HashTable_MemReturnFuncPtr* Asura_HashTable< T >::s_pMemReturnFuncPtr = NULL;

// Iterator
template< typename T >
class Asura_HashTable_It
{
public:
    // Constructor for iterator, also starts the iteration.
    Asura_HashTable_It< T >( Asura_HashTable< T >* pxTable )
	: m_pxTable( pxTable )
	, m_pxCurrentNode( NULL )
	, m_uCurrentRow(0)
    {
        Asura_Assert( pxTable, "Null table pointer" );
        Next();
    }

    // Gets the current entry in the iteration.
    T* GetCurrent() const
    {
		return m_pxCurrentNode->m_pxObject;
    }

    Asura_Hash_ID GetCurrentHashID( void ) const
    {
        return m_pxCurrentNode->m_pxObject->GetHashID();
    }

    // Deletes the current entry and moves to the next in the table.
    void DeleteCurrent()
    {
		if ( m_pxCurrentNode )
        {
            Asura_HashTable_Node< T >* pxToDelete = m_pxCurrentNode;
			const u_int uRow = m_uCurrentRow;
            Next();
			// NOTE: we need to specify the row as the contents of m_pxObject may already have been deleted.
			m_pxTable->RemoveNodeFromRow(pxToDelete, uRow);
        }
    }

    // Advance to the next entry in the table.
    void Next()
    {
		// Return to our place in the loop, if this isn't the first time.
		if (m_pxCurrentNode) goto ReEnterLoop;

		for ( ; m_uCurrentRow < m_pxTable->GetTableSize(); ++m_uCurrentRow)
		{
			m_pxCurrentNode = m_pxTable->m_ppxHead[m_uCurrentRow];

			while (m_pxCurrentNode)
			{
				return;

				ReEnterLoop :

				m_pxCurrentNode = m_pxCurrentNode->m_pxNext;
			}
		}
    }

    // Indicates whether the iteration has finished.
    bool Done() const
    {
        return ( m_pxCurrentNode == NULL );
    }

    // Restarts the iteration.
    void Restart()
    {
        m_pxCurrentNode = NULL;
		m_uCurrentRow = 0;
        Next();
    }

private:
    Asura_HashTable< T >* m_pxTable;
	Asura_HashTable_Node< T >* m_pxCurrentNode;
	u_int m_uCurrentRow;
};


// Constant Iterator
template< typename T >
class Asura_HashTable_ConstIt
{
public:
    // Constructor for iterator, also starts the iteration.
    Asura_HashTable_ConstIt< T >( const Asura_HashTable< T >* pxTable )
	: m_pxTable( pxTable )
	, m_pxCurrentNode( NULL )
	, m_uCurrentRow(0)
    {
        Asura_Assert( pxTable, "Null table pointer" );
        Next();
    }

    // Gets the current entry in the iteration.
    const T* GetCurrent() const
    {
		return m_pxCurrentNode->m_pxObject;
    }

    Asura_Hash_ID GetCurrentHashID( void ) const
    {
        return m_pxCurrentNode->m_pxObject->GetHashID();
    }

    // Advance to the next entry in the table.
    void Next()
    {
		// Return to our place in the loop, if this isn't the first time.
		if (m_pxCurrentNode) goto ReEnterLoop;

		for ( ; m_uCurrentRow < m_pxTable->GetTableSize(); ++m_uCurrentRow)
		{
			m_pxCurrentNode = m_pxTable->m_ppxHead[m_uCurrentRow];

			while (m_pxCurrentNode)
			{
				return;

				ReEnterLoop :

				m_pxCurrentNode = m_pxCurrentNode->m_pxNext;
			}
		}
    }

    // Indicates whether the iteration has finished.
    bool Done() const
    {
        return ( m_pxCurrentNode == NULL );
    }

    // Restarts the iteration.
    void Restart()
    {
        m_pxCurrentNode = NULL;
		m_uCurrentRow = 0;
        Next();
    }

private:
    const Asura_HashTable< T >* m_pxTable;
	const Asura_HashTable_Node< T >* m_pxCurrentNode;
	u_int m_uCurrentRow;
};


#endif//__ASURA_HASHTABLE_H__
