//
// Asura_HashTree.h
//
// Created by Richard Rice - 08/10/2002 
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//


#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_HASHTREE_H__
#define __ASURA_HASHTREE_H__

#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Debugging/Asura_Assert.h"

class Asura_Chunk_Stream;
template <class T> class Asura_TemplateHashTree_Iterator;
template <class T> class Asura_TemplateHashTree_ConstIterator;

class Asura_HashTree_Node
{
	public:
 
		Asura_Hash_ID m_uHash;
		int m_iIndex;
		class Asura_HashTree_Node *m_pxLower;
		class Asura_HashTree_Node *m_pxHigher;

		Asura_HashTree_Node(const Asura_Hash_ID uHash, int iIndex)
		{
			m_pxLower = 0;
			m_pxHigher = 0;	

			m_uHash = uHash;
			m_iIndex = iIndex;
		}

		// Recursive function to retrieve depth of nodes in this (sub)tree
		u_int GetDepth() const
		{
			// Default to zero sub-tree depth
			u_int uLower = 0;
			u_int uHigher = 0;
			// Retrieve depth of lower and higher sub-trees if they exist
			if(m_pxLower)	uLower  = m_pxLower->GetDepth();
			if(m_pxHigher)	uHigher = m_pxHigher->GetDepth();
			// Count this node, plus the largest of the lower and higher sub-trees
			return 1 + ((uLower > uHigher) ? uLower : uHigher);
		}
};

class Asura_HashTree
{
	public:
		
		Asura_HashTree()
		{
			m_iNumberOfNodes = 0;
			m_pxFirstRoot = 0;
			m_iMaximumNodes = 0x7fffffff;

			m_puHandleList = 0;
		}

		Asura_HashTree(int iMaximumNodes)
		{
			m_iNumberOfNodes = 0;
			m_pxFirstRoot = 0;
			m_iMaximumNodes = iMaximumNodes;

			m_puHandleList = new Asura_Handle_List(iMaximumNodes);
		}

		~Asura_HashTree()
		{
			FreeAll();

			if (m_puHandleList) delete m_puHandleList;
		}

        // (MH - 15.12.2002) should be no need to delete these in order to reset them.
        void Reset()
        {
			FreeAll();
			m_iNumberOfNodes = 0;
			m_pxFirstRoot = 0;

			if (m_puHandleList)
            {
                m_puHandleList->Reset();
            }
        }

		Asura_Hash_ID FindHashIDByIndex( int __iIndex ) const
		{
			Asura_HashTree_Node* pxNode = m_pxFirstRoot;

            // Empty tree?
            if( !pxNode )
			{
				return ASURA_HASH_ID_UNSET;
			}

			// Search.
			Asura_HashTree_Node* pxFoundNode = FindNodeByIndexRecursive(pxNode, __iIndex);
			return pxFoundNode ? pxFoundNode-> m_uHash: ASURA_HASH_ID_UNSET;
		}

		// Returns true if the node existed already
		bool FindNodeIndexByName(const char* szSearchName, bool bAddToTreeIfRequired, int& iIndex)
		{
			Asura_Assert(szSearchName,"");
			return FindNodeIndexByHash(Asura_GetHashID(szSearchName),bAddToTreeIfRequired,iIndex);
		}


		// Returns true if the node existed already
		bool FindNodeIndexByHash(Asura_Hash_ID uSearchHash, bool bAddToTreeIfRequired, int& iIndex)
		{
			if (!m_pxFirstRoot)
			{
				if (bAddToTreeIfRequired)
				{
					m_pxFirstRoot = new Asura_HashTree_Node(uSearchHash,m_iNumberOfNodes++);

					if (m_puHandleList)
					{
						#ifdef ASURA_ASSERT
						bool bSuccess = 
						#endif
						m_puHandleList->GetNewHandle(m_pxFirstRoot->m_iIndex);

						Asura_Assert(bSuccess,"No more space!");
					}
					iIndex =  m_pxFirstRoot->m_iIndex;
				}
				return false;
			}

			Asura_HashTree_Node **ppxNode = &m_pxFirstRoot;
			while(*ppxNode)
			{
				if((*ppxNode)->m_uHash==uSearchHash)
				{
					iIndex = (*ppxNode)->m_iIndex;
					return true;
				}
				else if(uSearchHash<(*ppxNode)->m_uHash)
				{
					ppxNode = &((*ppxNode)->m_pxLower);
				}
				else
				{
					ppxNode = &((*ppxNode)->m_pxHigher);
				}
			}
			//failed to find node
			if(bAddToTreeIfRequired)
			{
				*ppxNode = new Asura_HashTree_Node(uSearchHash,m_iNumberOfNodes++);
			
				if (m_puHandleList)
				{
					#ifdef ASURA_ASSERT
					bool bSuccess = 
					#endif
					m_puHandleList->GetNewHandle((*ppxNode)->m_iIndex);
					Asura_Assert(bSuccess,"No more space!");
				}
				iIndex = (*ppxNode)->m_iIndex;
			}
			return false;
		}


		// Delete a node and get its index, returns true if the node was found.
		// Note!!
		// The node with index NumberOfNodes-1 will have its index changed to the index
		// of the node being deleted. (since the indeces must be consecutive)
		bool DeleteNodeByHash(Asura_Hash_ID uSearchHash,int& iOutputIndex)
		{
			if(!DeleteNodeByHashInternal(uSearchHash,iOutputIndex)) return false;

			//need to alter the index of the last node to take the place of the deleted one
			if(m_pxFirstRoot)
			{
				Asura_HashTree_Node* pxNode = FindNodeByIndexRecursive(m_pxFirstRoot,m_iNumberOfNodes);
				if(pxNode)
				{
					pxNode->m_iIndex = iOutputIndex;
				}
			}

			return true;//search and destroy mission accomplished...
		}

		bool DeleteNodeByHash(Asura_Hash_ID uSearchHash)
		{
			Asura_Assert(m_puHandleList!=0,"This HashTree was not constructed with a Handle List");

			int iOutputIndex = ASURA_INVALID_HANDLE;

			if(!DeleteNodeByHashInternal(uSearchHash,iOutputIndex)) return false;

			m_puHandleList->ReleaseHandle(iOutputIndex);

			return true;
		}


		//'Renames' an existing node in the hash tree , replacing the old hash value with the new one
		bool ReplaceHash( Asura_Hash_ID uOldHash , Asura_Hash_ID uNewHash )
		{
			//make sure an entry with the new hash doesn't already exist
			int iJunkIndex;
			if(FindNodeIndexByHash(uNewHash, false, iJunkIndex))
			{
				return false;
			}
			//delete the old node , and get the index for it
			int iIndex;
            if( !DeleteNodeByHashInternal(uOldHash, iIndex) ) { return false; }

	 		//search for the point to reinsert with the new hash
			Asura_HashTree_Node **ppxInsertionPoint = &m_pxFirstRoot;
			while(*ppxInsertionPoint)
			{
				if( uNewHash > (*ppxInsertionPoint)->m_uHash )
				{
					ppxInsertionPoint = &(*ppxInsertionPoint)->m_pxHigher;
				}
				else
				{
					ppxInsertionPoint = &(*ppxInsertionPoint)->m_pxLower;
				}
			}

			//create the new node , and update the counters
			*ppxInsertionPoint = new Asura_HashTree_Node( uNewHash, iIndex );
			m_iNumberOfNodes++;

			return true;
		}
		
		void MakeHashList(Asura_Hash_ID* pxHashList) const
		{
			if (m_pxFirstRoot) MakeHashListRecursive(m_pxFirstRoot,pxHashList);
		}

		

		inline int GetNumberOfNodes(void) const
		{
			return m_iNumberOfNodes;
		}

		// Function to retrieve depth of nodes in this tree.
		// It is not recommended that you call this per frame, as it recurses down every node in the tree
		// (I can't see why you'd want to anyway) - it's mainly intended as a debug check for tree balance.
		int GetDepth() const
		{
			if (m_pxFirstRoot)
			{
				// Call recursive get depth on first root
				return m_pxFirstRoot->GetDepth();
			}
			else
			{
				// No nodes
				return 0;
			}
		}
	
	private:

		void FreeAll(void)
		{
			if (m_pxFirstRoot) FreeAllRecursive(m_pxFirstRoot);
		}

		void FreeAllRecursive(Asura_HashTree_Node *pxNode)
		{
			if (pxNode->m_pxLower)
			{
				FreeAllRecursive(pxNode->m_pxLower);
			}
			if (pxNode->m_pxHigher)
			{
				FreeAllRecursive(pxNode->m_pxHigher);
			}

			delete pxNode;
		}


		void MakeHashListRecursive(Asura_HashTree_Node* pxNode, Asura_Hash_ID* pxHashList) const
		{
			if (pxNode->m_pxLower)
			{
				MakeHashListRecursive(pxNode->m_pxLower,pxHashList);
			}
			if (pxNode->m_pxHigher)
			{
				MakeHashListRecursive(pxNode->m_pxHigher,pxHashList);
			}
			
			pxHashList[pxNode->m_iIndex] = pxNode->m_uHash;
		}
        
        Asura_HashTree_Node* FindNodeByIndexRecursive(Asura_HashTree_Node* __pxNode, int __iIndex ) const
        {
            if( __pxNode->m_iIndex == __iIndex )
            {
                return __pxNode;
            }
            else
            {
                Asura_HashTree_Node* pxRetVal = NULL;

                if( __pxNode->m_pxLower )
                {
                    pxRetVal = FindNodeByIndexRecursive( __pxNode->m_pxLower, __iIndex );
                }

                if( !pxRetVal && __pxNode->m_pxHigher )
                {
                    pxRetVal = FindNodeByIndexRecursive( __pxNode->m_pxHigher, __iIndex );
                }

                return pxRetVal;
            }
        }

		bool DeleteNodeByHashInternal(Asura_Hash_ID uSearchHash,int& iOutputIndex)
		{
			if(!m_pxFirstRoot) return false;

			Asura_HashTree_Node **ppxNode = &m_pxFirstRoot;

			while(*ppxNode)
			{
				if((*ppxNode)->m_uHash==uSearchHash)
				{
					//found the node
					break;
				}
				else if(uSearchHash<(*ppxNode)->m_uHash)
				{
					ppxNode = &((*ppxNode)->m_pxLower);
				}
				else
				{
					ppxNode = &((*ppxNode)->m_pxHigher);
				}
				
			}
			if(!*ppxNode) return false;
			
			Asura_HashTree_Node *pxNodeToDelete = *ppxNode;
			//found the node to delete , if it has children they will need to be transferred up
			if(pxNodeToDelete->m_pxLower && pxNodeToDelete->m_pxHigher)
			{
				//The deleted node has 2 children  , 1 will need to be found a place beneath the other
				//Pick semi-randomly to keep some chance of things not being unbalanced
				Asura_HashTree_Node* pxNodeToTransfer;
				if(m_iNumberOfNodes%2)
				{
					*ppxNode = pxNodeToDelete->m_pxLower;
					pxNodeToTransfer = pxNodeToDelete->m_pxHigher;
				}
				else
				{
					*ppxNode = pxNodeToDelete->m_pxHigher;
					pxNodeToTransfer = pxNodeToDelete->m_pxLower;
				}
				Asura_HashTree_Node *pxNode = *ppxNode;

				//search down for somewhere to stick the transferred node
				TryNextNode :
				
				if(pxNode->m_uHash<pxNodeToTransfer->m_uHash)
				{
					if(pxNode->m_pxHigher)
					{
						pxNode = pxNode->m_pxHigher;
						goto TryNextNode;
					}
					else
					{
						pxNode->m_pxHigher = pxNodeToTransfer;
					}
				}
				else
				{
					if(pxNode->m_pxLower)
					{
						pxNode = pxNode->m_pxLower;
						goto TryNextNode;
					}
					else
					{
						pxNode->m_pxLower = pxNodeToTransfer;
					}
				}
			}
			else if(pxNodeToDelete->m_pxLower)
			{
				*ppxNode = pxNodeToDelete->m_pxLower;
			}
			else
			{
				*ppxNode = pxNodeToDelete->m_pxHigher;
			}

			iOutputIndex = pxNodeToDelete->m_iIndex;


			//get rid of the node ,and update the counters
			m_iNumberOfNodes--;
			delete pxNodeToDelete;

			return true;
		}

		Asura_HashTree_Node *m_pxFirstRoot;
		Asura_Handle_List *m_puHandleList;

		int m_iNumberOfNodes;
		int m_iMaximumNodes;
		
};

//////////////////////////////////////////////////////////////////////////////////////////////////

//Templatised version of hash tree for storing data sorted by hash value. RAR 09/12/2003

template <class T> class Asura_TemplateHashTree_Node
{
public:
	Asura_Hash_ID m_uHash;
	T m_xData;
	Asura_TemplateHashTree_Node<T>* m_pxLower;
	Asura_TemplateHashTree_Node<T>* m_pxHigher;

	Asura_TemplateHashTree_Node(const Asura_Hash_ID uHash,const T& xData)
	: m_uHash(uHash)
	, m_xData(xData)
	, m_pxLower(NULL)
	, m_pxHigher(NULL)
	{
	}

	Asura_TemplateHashTree_Node(const Asura_Hash_ID uHash,Asura_Chunk_Stream& xStream)
	: m_uHash(uHash)
	, m_xData(uHash, xStream)
	, m_pxLower(NULL)
	, m_pxHigher(NULL)
	{
	}

	Asura_TemplateHashTree_Node(const Asura_Hash_ID uHash)
	: m_uHash(uHash)
	, m_xData()
	, m_pxLower(NULL)
	, m_pxHigher(NULL)
	{
	}
};

template <class T> class Asura_TemplateHashTree
{
friend class Asura_TemplateHashTree_Iterator< T >;
friend class Asura_TemplateHashTree_ConstIterator< T >;

public:
	typedef void TraverseFunction(Asura_Hash_ID uHash, T* pxData);
	typedef bool TraverseWithUserDataFunction(Asura_Hash_ID uHash, T* pxData, void* pUserData);
	typedef void TraverseFunction_Const(Asura_Hash_ID uHash, T const* pxData);
	typedef bool TraverseWithUserDataFunction_Const(Asura_Hash_ID uHash, T const* pxData, void* pUserData);
	
	Asura_TemplateHashTree()
	{
		m_iNumberOfNodes = 0;
		m_pxFirstRoot = NULL;
		m_bTraversalInProgress = false;
		m_uIterationCount = 0;

		m_pxDeletionHashList = NULL;
		m_uNumberOfNodesToDelete = 0;

		m_pxParentTree = NULL;
#ifdef ASURA_TOOLS
		m_ppxNodes = NULL;
#endif
	}

	~Asura_TemplateHashTree()
	{
		FreeAll();
	}

	Asura_TemplateHashTree& operator =( const Asura_TemplateHashTree& xRHS)
	{
		this->Reset();
		
		TraverseWithUserDataFunction_Const* pfnCallback = this->OperatorEquTraverseCallback;
		
		xRHS.Traverse( pfnCallback, reinterpret_cast< void*>( this));
		
		return( *this);
	}
	
    // (MH - 15.12.2002) should be no need to delete these in order to reset them.
    void Reset()
    {
		FreeAll();
		m_iNumberOfNodes = 0;
		m_pxFirstRoot = 0;
	}

	T* GetDataByName(const char* szSearchName) const
	{
		Asura_Assert(szSearchName,"");
		return GetDataByHash(Asura_GetHashID(szSearchName));
	}

	T* GetDataByHash(Asura_Hash_ID uSearchHash, bool bSearchParentTree = true) const
	{
		Asura_TemplateHashTree_Node<T>* const*	ppxNode	= &m_pxFirstRoot;
		while(*ppxNode)
		{
			if((*ppxNode)->m_uHash==uSearchHash)
			{
				return &((*ppxNode)->m_xData);
			}
			else if(uSearchHash<(*ppxNode)->m_uHash)
			{
				ppxNode = &((*ppxNode)->m_pxLower);
			}
			else
			{
				ppxNode = &((*ppxNode)->m_pxHigher);
			}
		}

		// Failed to find a match...
		if(m_pxParentTree != NULL && bSearchParentTree)
		{
			return(m_pxParentTree->GetDataByHash(uSearchHash));
		}
		else
		{
			return(NULL);
		}
	}

#ifdef ASURA_TOOLS
	//GetDataByIndex added to help with conversion of old systems to using template hash tree.
	//Hopefully this functionality will only be needed by tools , so the allocation of the extra
	//array won't be an issue. RAR 13/03/2006
	//Note that the indexing is based on hash order , rather than the order in which nodes are added
	T* GetDataByIndex(int iIndex) const
	{
		if(iIndex<0 || iIndex>=m_iNumberOfNodes)
		{
			Asura_Assert(iIndex>=0 && iIndex<m_iNumberOfNodes,"Index out of range");
			return 0;
		}

		//Build array of nodes if required. RAR 13/03/2006
		if(!m_ppxNodes)
		{
			m_ppxNodes = new Asura_TemplateHashTree_Node<T>*[m_iNumberOfNodes];
#ifdef ASURA_ASSERT
			int iNextIndex = 
#endif
			FillNodeArrayRecursive(m_pxFirstRoot,0);
			Asura_Assert(iNextIndex == m_iNumberOfNodes,"Error while indexing nodes");

		}

		return &m_ppxNodes[iIndex]->m_xData;
	}

	//AJG - 05/10/06 - Added this to match GetDataByIndex, so the hash id of a node can be found from the same index.
	Asura_Hash_ID GetHashByIndex(int iIndex) const
	{
		if(iIndex<0 || iIndex>=m_iNumberOfNodes)
		{
			Asura_Assert(iIndex>=0 && iIndex<m_iNumberOfNodes,"Index out of range");
			return 0;
		}

		//Build array of nodes if required. RAR 13/03/2006
		if(!m_ppxNodes)
		{
			m_ppxNodes = new Asura_TemplateHashTree_Node<T>*[m_iNumberOfNodes];
#ifdef ASURA_ASSERT
			int iNextIndex = 
#endif
			FillNodeArrayRecursive(m_pxFirstRoot,0);
			Asura_Assert(iNextIndex == m_iNumberOfNodes,"Error while indexing nodes");

		}

		return m_ppxNodes[iIndex]->m_uHash;
	}

	int GetIndexOfHash(const Asura_Hash_ID uHash)
	{
		//Build array of nodes if required.
		if(!m_ppxNodes)
		{
			if(!m_pxFirstRoot) return -1;
			m_ppxNodes = new Asura_TemplateHashTree_Node<T>*[m_iNumberOfNodes];
#ifdef ASURA_ASSERT
			int iNextIndex = 
#endif
			FillNodeArrayRecursive(m_pxFirstRoot,0);
			Asura_Assert(iNextIndex == m_iNumberOfNodes,"Error while indexing nodes");
		}

		return FindIndexBinarySearch(uHash);
	}
#endif //ASURA_TOOLS

	//adds node to tree , returns pointer to data if successful
	T* AddNode(Asura_Hash_ID uHash, const T& xData)
	{
		if(m_bTraversalInProgress)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during a traversal.");
			return(NULL);
		}
		else if(m_uIterationCount > 0)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during iteration.");
			return(NULL);
		}
		else
		{
			Asura_TemplateHashTree_Node<T> **ppxNode = &m_pxFirstRoot;
			//find point to insert new node
			while(*ppxNode)
			{
				if((*ppxNode)->m_uHash==uHash)
				{
					//already exists
					return NULL;
				}
				else if(uHash<(*ppxNode)->m_uHash)
				{
					ppxNode = &((*ppxNode)->m_pxLower);
				}
				else
				{
					ppxNode = &((*ppxNode)->m_pxHigher);
				}
			}

			*ppxNode = new Asura_TemplateHashTree_Node<T>(uHash,xData);
			m_iNumberOfNodes++;

#ifdef ASURA_TOOLS
			InvalidateNodeArray();
#endif

			return &((*ppxNode)->m_xData);
		}
	}

	//adds node to tree , returns pointer to data if successful
	T* AddNode(Asura_Hash_ID uHash, Asura_Chunk_Stream& xStream)
	{
		if(m_bTraversalInProgress)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during a traversal.");
			return(NULL);
		}
		else if(m_uIterationCount > 0)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during iteration.");
			return(NULL);
		}
		else
		{
			Asura_TemplateHashTree_Node<T>** ppxNode = &m_pxFirstRoot;
			Asura_TemplateHashTree_Node<T>* pxNewNode = new Asura_TemplateHashTree_Node<T>(uHash,xStream);
			//find point to insert new node
			while(*ppxNode)
			{
				if((*ppxNode)->m_uHash==uHash)
				{
					//already exists
					delete(pxNewNode);
					return NULL;
				}
				else if(uHash<(*ppxNode)->m_uHash)
				{
					ppxNode = &((*ppxNode)->m_pxLower);
				}
				else
				{
					ppxNode = &((*ppxNode)->m_pxHigher);
				}
			}

			*ppxNode = pxNewNode;
			m_iNumberOfNodes++;

#ifdef ASURA_TOOLS
			InvalidateNodeArray();
#endif

			return &((*ppxNode)->m_xData);
		}
	}

	//adds node to tree , returns pointer to data if successful
	T* AddNode(Asura_Hash_ID uHash)
	{
		if(m_bTraversalInProgress)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during a traversal.");
			return(NULL);
		}
		else if(m_uIterationCount > 0)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during iteration.");
			return(NULL);
		}
		else
		{
			Asura_TemplateHashTree_Node<T> **ppxNode = &m_pxFirstRoot;
			//find point to insert new node
			while(*ppxNode)
			{
				if((*ppxNode)->m_uHash==uHash)
				{
					//already exists
					return 0;
				}
				else if(uHash<(*ppxNode)->m_uHash)
				{
					ppxNode = &((*ppxNode)->m_pxLower);
				}
				else
				{
					ppxNode = &((*ppxNode)->m_pxHigher);
				}
			}

			*ppxNode = new Asura_TemplateHashTree_Node<T>(uHash);
			m_iNumberOfNodes++;

#ifdef ASURA_TOOLS
			InvalidateNodeArray();
#endif

			return &((*ppxNode)->m_xData);
		}
	}

	// KJFL - 24 October 2007 - 
	// Add a new node with this hash or return a pointer to an existing one. Originally I tried 
	// adding a bool parameter to the above AddNode( Asura_Hash_ID uHash ) method, but that can
	// create compile errors about ambiguity with AddNode(Asura_Hash_ID uHash,const T& xData)
	// e.g. people using int or pointer types for T and passing 0 as a the xData parameter.
	T* AddOrFindNode( Asura_Hash_ID uHash, bool* pbDidExist = NULL, bool bSearchParentTree = true )
	{


		Asura_TemplateHashTree_Node<T> **ppxNode = &m_pxFirstRoot;
		//find point to insert new node
		while(*ppxNode)
		{
			if((*ppxNode)->m_uHash==uHash)
			{
				if (pbDidExist)
				{
					*pbDidExist = true;
				}
				//already exists
				return &((*ppxNode)->m_xData);
			}
			else if(uHash<(*ppxNode)->m_uHash)
			{
				ppxNode = &((*ppxNode)->m_pxLower);
			}
			else
			{
				ppxNode = &((*ppxNode)->m_pxHigher);
			}
		}

		if (m_pxParentTree && bSearchParentTree)
		{
			T* pxFoundData = m_pxParentTree->GetDataByHash(uHash, bSearchParentTree);
			if (pxFoundData) return pxFoundData;
		}

		//Node doesn't exist ,so complain if a traversal/iteration is happening that would prevent
		//us from adding a new one safely. RAR 14/11/2008
		if(m_bTraversalInProgress)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during a traversal.");
			return(NULL);
		}
		else if(m_uIterationCount > 0)
		{
			Asura_Assert(false, "Cannot add a hash-tree node during iteration.");
			return(NULL);
		}


		*ppxNode = new Asura_TemplateHashTree_Node<T>(uHash);
		m_iNumberOfNodes++;

#ifdef ASURA_TOOLS
		InvalidateNodeArray();
#endif
		if (pbDidExist)
		{
			*pbDidExist = false;
		}
		return &((*ppxNode)->m_xData);
	}



	// Delete a nodes from the tree
	bool DeleteNodeByHash(Asura_Hash_ID uSearchHash)
	{
		if(m_bTraversalInProgress || (m_uIterationCount > 0))
		{
			if(m_pxDeletionHashList == NULL)
			{
				m_pxDeletionHashList = new Asura_Hash_ID [m_iNumberOfNodes];
			}
			m_pxDeletionHashList[m_uNumberOfNodesToDelete++] = uSearchHash;
			return(true);
		}
		else
		{
			Asura_TemplateHashTree_Node<T>* pxNode = UnlinkNodeByHash(uSearchHash);
			if(pxNode)
			{
				delete pxNode;
				return true;
			}
			return false;
		}
	}

	//'Renames' an existing node in the hash tree , replacing the old hash value with the new one
	bool ReplaceHash( Asura_Hash_ID uOldHash , Asura_Hash_ID uNewHash )
	{
		if(m_bTraversalInProgress)
		{
			Asura_Assert(false, "Cannot replace a hash-tree node during a traversal.");
			return(false);
		}
		else if(m_uIterationCount > 0)
		{
			Asura_Assert(false, "Cannot replace a hash-tree node during iteration.");
			return(false);
		}
		else
		{
			//make sure an entry with the new hash doesn't already exist
			if(GetDataByHash(uNewHash))
			{
				return false;
			}
			//Unlink the node being renamed
			Asura_TemplateHashTree_Node<T>* pxRenamedNode = UnlinkNodeByHash(uOldHash);
			if(!pxRenamedNode) return false;

			pxRenamedNode->m_uHash = uNewHash;

			//search for the point to reinsert with the new hash
			Asura_TemplateHashTree_Node<T> **ppxInsertionPoint = &m_pxFirstRoot;
			while(*ppxInsertionPoint)
			{
				if( uNewHash > (*ppxInsertionPoint)->m_uHash )
				{
					ppxInsertionPoint = &(*ppxInsertionPoint)->m_pxHigher;
				}
				else
				{
					ppxInsertionPoint = &(*ppxInsertionPoint)->m_pxLower;
				}
			}

			//Reinsert the node
			*ppxInsertionPoint = pxRenamedNode;
			m_iNumberOfNodes++;

			return true;
		}
	}

	inline int GetNumberOfNodes() const { return m_iNumberOfNodes; }

	inline Asura_Hash_ID* MakeHashList() const
	{
		// Do we have a root node?
		if (NULL == m_pxFirstRoot)
		{
			return (NULL);
		}
		else
		{
			Asura_Hash_ID* auHashIDs = new Asura_Hash_ID [m_iNumberOfNodes];

			// Simply call the recursive function...
			MakeHashListRecursive(m_pxFirstRoot, auHashIDs);
			return (auHashIDs);
		}
	}

	inline T** MakeDataList() const
	{
		// Do we have a root node?
		if (NULL == m_pxFirstRoot)
		{
			// No - return NULL...
			return	(NULL);
		}
		else
		{
			T**	ppxData	= new T* [m_iNumberOfNodes];

			// Simply call the recursive function...
			MakeDataListRecursive(m_pxFirstRoot, ppxData);
			return(ppxData);
		}
	}

	//calls 'pfnFunc' on each entry in tree
	void Traverse(TraverseFunction *pfnFunc)
	{
		m_bTraversalInProgress = true;
		if(m_pxFirstRoot) TraverseRecursive(m_pxFirstRoot,pfnFunc);
		FreeDeletedNodesList();
		m_bTraversalInProgress = false;
	}

	//Alternative traversal function which passes a pointer to some user data to calback function.
	//Return true to continue traversal , and false to stop.
	//RAR 21/03/2006
	void Traverse(TraverseWithUserDataFunction *pfnFunc, void* pUserData)
	{
		m_bTraversalInProgress = true;
		if(m_pxFirstRoot) TraverseRecursive(m_pxFirstRoot,pfnFunc,pUserData);
		FreeDeletedNodesList();
		m_bTraversalInProgress = false;
	}

	// RDM - 19/04/06 - Const versions of above
	void Traverse(TraverseFunction_Const *pfnFunc) const
	{
		m_bTraversalInProgress = true;
		if(m_pxFirstRoot) TraverseRecursive(m_pxFirstRoot,pfnFunc);
		Asura_Assert( m_pxDeletionHashList == NULL, "Cannot delete during const traversal!" );
		m_bTraversalInProgress = false;
	}

	void Traverse(TraverseWithUserDataFunction_Const *pfnFunc, void* pUserData) const
	{
		m_bTraversalInProgress = true;
		if(m_pxFirstRoot) TraverseRecursive(m_pxFirstRoot,pfnFunc,pUserData);
		Asura_Assert( m_pxDeletionHashList == NULL, "Cannot delete during const traversal!" );
		m_bTraversalInProgress = false;
	}

	void SetParent(Asura_TemplateHashTree<T>* pxParent)
	{
		m_pxParentTree	= pxParent;
	}

private:
	Asura_TemplateHashTree( const Asura_TemplateHashTree&);

	void BeginIteration(void)
	{
		m_uIterationCount++;
	}

	void EndIteration(void)
	{
		if(m_uIterationCount == 0)
		{
			Asura_Assert(false, "Mismatch between call to \"BeginIteration()\" and \"EndIteration\".");
		}
		else
		{
			m_uIterationCount--;

			if(m_uIterationCount == 0)
			{
				FreeDeletedNodesList();
			}
		}
	}

	void FreeAll(void)
	{
		Asura_Assert((m_uIterationCount == 0), "This hash-tree is still being iterated over.");

		if (m_pxFirstRoot) FreeAllRecursive(m_pxFirstRoot);

		delete[](m_pxDeletionHashList);
		m_pxDeletionHashList = NULL;
		m_uNumberOfNodesToDelete = 0;

		#ifdef ASURA_TOOLS
		InvalidateNodeArray();
		#endif
	}


	void FreeAllRecursive(Asura_TemplateHashTree_Node<T> *pxNode)
	{
		if (pxNode->m_pxLower)
		{
			FreeAllRecursive(pxNode->m_pxLower);
		}
		if (pxNode->m_pxHigher)
		{
			FreeAllRecursive(pxNode->m_pxHigher);
		}

		delete pxNode;
	}

	static bool OperatorEquTraverseCallback( Asura_Hash_ID uHash, T const* pxData, void* pUserData)
	{
		Asura_TemplateHashTree* pxLHS = reinterpret_cast< Asura_TemplateHashTree*>( pUserData);
		
		pxLHS->AddNode( uHash, *pxData);
		
		return( true);
	}
	
	//Unlinks the specified node from the tree and returns it.
	//The count of the number of nodes is decreased in the process.
	//RAR 15/03/2006
	Asura_TemplateHashTree_Node<T>* UnlinkNodeByHash(Asura_Hash_ID uSearchHash)
	{
		if(!m_pxFirstRoot) return 0;

		Asura_TemplateHashTree_Node<T> **ppxNode = &m_pxFirstRoot;

		while(*ppxNode)
		{
			if((*ppxNode)->m_uHash==uSearchHash)
			{
				//found the node
				break;
			}
			else if(uSearchHash<(*ppxNode)->m_uHash)
			{
				ppxNode = &((*ppxNode)->m_pxLower);
			}
			else
			{
				ppxNode = &((*ppxNode)->m_pxHigher);
			}
			
		}
		if(!*ppxNode) return 0;
		
		Asura_TemplateHashTree_Node<T> *pxNodeToUnlink = *ppxNode;
		//found the node to delete , if it has children they will need to be transferred up
		if(pxNodeToUnlink->m_pxLower && pxNodeToUnlink->m_pxHigher)
		{
			//The deleted node has 2 children  , 1 will need to be found a place beneath the other
			//Pick semi-randomly to keep some chance of things not being unbalanced
			Asura_TemplateHashTree_Node<T>* pxNodeToTransfer;
			if(m_iNumberOfNodes%2)
			{
				*ppxNode = pxNodeToUnlink->m_pxLower;
				pxNodeToTransfer = pxNodeToUnlink->m_pxHigher;
			}
			else
			{
				*ppxNode = pxNodeToUnlink->m_pxHigher;
				pxNodeToTransfer = pxNodeToUnlink->m_pxLower;
			}
			Asura_TemplateHashTree_Node<T> *pxNode = *ppxNode;

			//search down for somewhere to stick the transferred node
			TryNextNode :
			
			if(pxNode->m_uHash<pxNodeToTransfer->m_uHash)
			{
				if(pxNode->m_pxHigher)
				{
					pxNode = pxNode->m_pxHigher;
					goto TryNextNode;
				}
				else
				{
					pxNode->m_pxHigher = pxNodeToTransfer;
				}
			}
			else
			{
				if(pxNode->m_pxLower)
				{
					pxNode = pxNode->m_pxLower;
					goto TryNextNode;
				}
				else
				{
					pxNode->m_pxLower = pxNodeToTransfer;
				}
			}
		}
		else if(pxNodeToUnlink->m_pxLower)
		{
			*ppxNode = pxNodeToUnlink->m_pxLower;
		}
		else
		{
			*ppxNode = pxNodeToUnlink->m_pxHigher;
		}

		//The unlinked node no longer has any children
		pxNodeToUnlink->m_pxLower = 0;
		pxNodeToUnlink->m_pxHigher = 0;

		m_iNumberOfNodes--;

		#ifdef ASURA_TOOLS
		InvalidateNodeArray();
		#endif

		return pxNodeToUnlink;
	}

	Asura_Hash_ID* MakeHashListRecursive(Asura_TemplateHashTree_Node<T>* pxNode, Asura_Hash_ID* pxHashList) const
	{
		if (pxNode->m_pxLower)
		{
			pxHashList	= MakeHashListRecursive (pxNode->m_pxLower, pxHashList);
		}
		if (pxNode->m_pxHigher)
		{
			pxHashList	= MakeHashListRecursive (pxNode->m_pxHigher, pxHashList);
		}
		
		*(pxHashList++)	= pxNode->m_uHash;

		return	(pxHashList);
	}

	T** MakeDataListRecursive(Asura_TemplateHashTree_Node<T>* pxNode, T** ppxData) const
	{
		if (pxNode->m_pxLower)
		{
			ppxData	= MakeDataListRecursive (pxNode->m_pxLower, ppxData);
		}
		if (pxNode->m_pxHigher)
		{
			ppxData	= MakeDataListRecursive (pxNode->m_pxHigher, ppxData);
		}

		*(ppxData++)	= &(pxNode->m_xData);

		return	(ppxData);
	}

	void TraverseRecursive(Asura_TemplateHashTree_Node<T>* pxNode, TraverseFunction* pfnFunc)
	{
		if (pxNode->m_pxLower)
		{
			TraverseRecursive(pxNode->m_pxLower,pfnFunc);
		}
		if (pxNode->m_pxHigher)
		{
			TraverseRecursive(pxNode->m_pxHigher,pfnFunc);
		}

		pfnFunc(pxNode->m_uHash,&pxNode->m_xData);
	}

	bool TraverseRecursive(Asura_TemplateHashTree_Node<T>* pxNode, TraverseWithUserDataFunction* pfnFunc, void* pUserData)
	{
		if (pxNode->m_pxLower)
		{
			if(!TraverseRecursive(pxNode->m_pxLower,pfnFunc,pUserData)) return false;
		}
		if (pxNode->m_pxHigher)
		{
			if(!TraverseRecursive(pxNode->m_pxHigher,pfnFunc,pUserData)) return false;
		}

		return pfnFunc(pxNode->m_uHash,&pxNode->m_xData,pUserData);
	}

	// const versions
	void TraverseRecursive(Asura_TemplateHashTree_Node<T>* pxNode, TraverseFunction_Const* pfnFunc)	const
	{
		if (pxNode->m_pxLower)
		{
			TraverseRecursive(pxNode->m_pxLower,pfnFunc);
		}
		if (pxNode->m_pxHigher)
		{
			TraverseRecursive(pxNode->m_pxHigher,pfnFunc);
		}

		pfnFunc(pxNode->m_uHash,&pxNode->m_xData);
	}

	bool TraverseRecursive(Asura_TemplateHashTree_Node<T>* pxNode, TraverseWithUserDataFunction_Const* pfnFunc, void* pUserData) const
	{
		if (pxNode->m_pxLower)
		{
			if(!TraverseRecursive(pxNode->m_pxLower,pfnFunc,pUserData)) return false;
		}
		if (pxNode->m_pxHigher)
		{
			if(!TraverseRecursive(pxNode->m_pxHigher,pfnFunc,pUserData)) return false;
		}

		return pfnFunc(pxNode->m_uHash,&pxNode->m_xData,pUserData);
	}

#ifdef ASURA_TOOLS
	int FillNodeArrayRecursive(Asura_TemplateHashTree_Node<T>* pxNode,int iIndex) const
	{
		if (pxNode->m_pxLower)
		{
			iIndex = FillNodeArrayRecursive (pxNode->m_pxLower,iIndex);
		}
		m_ppxNodes[iIndex++] = pxNode;
		if (pxNode->m_pxHigher)
		{
			iIndex = FillNodeArrayRecursive (pxNode->m_pxHigher,iIndex);
		}
		return iIndex;
	}

	bool FindIndexRecursive(const Asura_Hash_ID uHash, Asura_TemplateHashTree_Node<T>* pxNode, int& iIndex)
	{
		if (pxNode->m_pxLower)
		{
			if (FindIndexRecursive(uHash, pxNode->m_pxLower, iIndex)) { return true; }
		}
		if (pxNode->m_uHash == uHash) { return true; }
		++iIndex;
		if (pxNode->m_pxHigher)
		{
			if (FindIndexRecursive(uHash, pxNode->m_pxHigher, iIndex)) { return true; }
		}
		return false;
	}

	int FindIndexBinarySearch(const Asura_Hash_ID uHashToFind)
	{
		int iLeftIndex = 0;
		int iRightIndex = m_iNumberOfNodes - 1;
		while (iLeftIndex <= iRightIndex)
		{
			int iIndex = ( iLeftIndex + iRightIndex ) / 2;
			const Asura_Hash_ID uThisHash = m_ppxNodes[iIndex]->m_uHash;
			if (uThisHash == uHashToFind)
			{
				return iIndex;
			}
			else if (uThisHash > uHashToFind)
			{
				// look lower
				iRightIndex = iIndex - 1;
			}
			else
			{
				// look higher
				iLeftIndex = iIndex + 1;
			}
		}
		// failed
		return -1;
	}

	void InvalidateNodeArray()
	{
		if(m_ppxNodes)
		{
			delete [] m_ppxNodes;
			m_ppxNodes = 0;
		}
	}
#endif //ASURA_TOOLS

	void FreeDeletedNodesList(void)
	{
		if(m_pxDeletionHashList != NULL)
		{
			for(u_int u = 0; u < m_uNumberOfNodesToDelete; u++)
			{
				const Asura_Hash_ID				uHash	= m_pxDeletionHashList[u];
				Asura_TemplateHashTree_Node<T>*	pxNode	= UnlinkNodeByHash(uHash);

				delete(pxNode);
			}

			delete[](m_pxDeletionHashList);
			m_pxDeletionHashList = NULL;
			m_uNumberOfNodesToDelete = 0;
		}
	}


	Asura_TemplateHashTree_Node<T> *m_pxFirstRoot;
	int m_iNumberOfNodes;
	mutable bool m_bTraversalInProgress;
	u_int m_uIterationCount;

	Asura_Hash_ID* m_pxDeletionHashList;
	u_int m_uNumberOfNodesToDelete;

	Asura_TemplateHashTree<T>*	m_pxParentTree;		// Parent tree to perform checks in.

#ifdef ASURA_TOOLS
	//For lazy evaluation of accessing nodes by index while in tools. RAR 13/03/2006
	mutable Asura_TemplateHashTree_Node<T>** m_ppxNodes;
#endif
};


#ifdef ASURA_TOOLS
//When accessing the hash tree entries by index in a template hash tree, they are accessed in hash value order , so saving them consecutively
//will result in a perfectly unbalanced tree upon load.
//This iterator visits the index values in an order that should result in a balanced tree.
//Its wrapped in ASURA_TOOLS since GetDataByIndex() is tools only
//RAR 17/08/2006 


class Asura_TemplateHashTree_SaveIterator
{
public  :
	//Creates and starts iterator for tree with iSize nodes
	Asura_TemplateHashTree_SaveIterator(int iSize)
	{
		m_iSize = iSize;

		//Find highest power of 2 greater than or equal to the iteration size
		m_iPower = 1;
		m_iBits = 0;
			
		while(m_iPower < m_iSize)
		{
			m_iPower <<= 1;
			m_iBits++;
		}
		//Get an offset for recentreing the iteration ,so that the first value produced will be iSize/2 (give or take rounding)
		m_iOffset = (m_iPower - m_iSize)/2;

		m_iPos = 0;
		Next();
	}

	//get current value in iteration (index to use into hash tree)
	int GetCurrent() const {return m_iValue;}

	//retuns true if iteration has finished
	bool Done() const {return m_iPos > m_iPower;}

	//move onto next value
	void Next()
	{
		while(m_iPos <= m_iPower)
		{
			m_iPos++;

			//Get the current value by reversing the order of the bits in the current position
			m_iValue = 0;
			int iBit = 1;
			for(int i=0;i<m_iBits;i++,iBit<<=1)
			{
				m_iValue <<= 1;
				if(m_iPos & iBit) m_iValue++;
			}
			//Apply the recentreing offset
			m_iValue -= m_iOffset;

			if(m_iValue>=0 && m_iValue < m_iSize) return;
		}
	}

private :
	int m_iSize;
	int m_iPower;
	int m_iBits;
	int m_iOffset;

	int m_iPos;
	int m_iValue;
};
#endif //ASURA_TOOLS


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_TemplateHashTree_Iterator
// Author       : Andy Tate
// Date         : 3-Dec-2007
//
// Description  : Templified class used to iterate through members of the
//				  "Asura_TemplateHashTree" class.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_TemplateHashTree_Iterator
{
public:
	// Public functions...
	Asura_TemplateHashTree_Iterator(Asura_TemplateHashTree<T>& xHashTree)
	: m_xHashTree(xHashTree)
	, m_ppxNodes(NULL)
	, m_uNumNodes(0)
	, m_uCurrentNode(0)
	, m_bIterating( false )
	{
		Begin();
	}
	~Asura_TemplateHashTree_Iterator()
	{
		End();
	}

	void Restart()
	{
		End();
		Begin();
	}

	// Query function to see whether iteration is complete...
	bool Done() const
	{
		return(m_ppxNodes == NULL);
	}

	// Function to step onto the next item to be iterated...
	void Next()
	{
		m_uCurrentNode++;

		if(m_uCurrentNode == m_uNumNodes)
		{
			End();
		}
	}

	// Accessor function to get the current entry...
	T* GetCurrent() const
	{
		if(m_ppxNodes == NULL)
		{
			return(NULL);
		}
		else
		{
			return(&(m_ppxNodes[m_uCurrentNode]->m_xData));
		}
	}

	Asura_Hash_ID GetCurrentHash() const
	{
		if(m_ppxNodes == NULL)
		{
			return(ASURA_HASH_ID_UNSET);
		}
		else
		{
			return(m_ppxNodes[m_uCurrentNode]->m_uHash);
		}
	}

protected:
	void Begin()
	{
		if (!m_bIterating)
		{
			m_uCurrentNode	= 0;
			m_uNumNodes		= m_xHashTree.GetNumberOfNodes();

			if(m_uNumNodes == 0)
			{
				m_ppxNodes	= NULL;
			}
			else
			{
				m_xHashTree.BeginIteration();

				m_ppxNodes	=  new Asura_TemplateHashTree_Node<T>* [m_uNumNodes];

				if(m_xHashTree.m_pxFirstRoot != NULL)
				{
					FillItemArray(m_xHashTree.m_pxFirstRoot);
				}

				m_uCurrentNode	= 0;
			}

			m_bIterating = true;
		}
	}

	void End()
	{
		if( m_bIterating && (m_ppxNodes != NULL) )
		{
			delete[](m_ppxNodes);

			m_ppxNodes		= NULL;
			m_uCurrentNode	= 0;
			m_uNumNodes		= 0;

			m_xHashTree.EndIteration();

			m_bIterating = false;
		}
	}

	// Recursive function used to fill the item array with data...
	void FillItemArray(Asura_TemplateHashTree_Node<T>* pxNode)
	{
		Asura_TemplateHashTree_Node<T>*	pxLower		= pxNode->m_pxLower;
		Asura_TemplateHashTree_Node<T>*	pxHigher	= pxNode->m_pxHigher;

		if(pxLower != NULL)
		{
			FillItemArray(pxLower);
		}

		m_ppxNodes[m_uCurrentNode++]	= pxNode;

		if(pxHigher != NULL)
		{
			FillItemArray(pxHigher);
		}
	}

private:
	// Private variables...
	Asura_TemplateHashTree<T>&			m_xHashTree;	// A reference to the hash-tree to iterate over.
	Asura_TemplateHashTree_Node<T>**	m_ppxNodes;		// An array of pointers to the data to be iterated over.
	u_int								m_uNumNodes;	// The number of nodes to be iterated over.
	u_int								m_uCurrentNode;	// The index of the current node during iteration.
	bool								m_bIterating;	// Flag to enforce ordering of Begin() End()

	// Private functions...
	Asura_TemplateHashTree_Iterator(void);
	Asura_TemplateHashTree_Iterator(const Asura_TemplateHashTree_Iterator& xOther);

	Asura_TemplateHashTree_Iterator&	operator=(const Asura_TemplateHashTree_Iterator& xOther);

};   // End of class 'Asura_TemplateHashTree_Iterator'.


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_TemplateHashTree_ConstIterator
// Author       : Andy Tate
// Date         : 3-Dec-2007
//
// Description  : Templified class used to iterate through members of the
//				  "Asura_TemplateHashTree" class.
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_TemplateHashTree_ConstIterator
{
public:
	// Public functions...
	Asura_TemplateHashTree_ConstIterator(Asura_TemplateHashTree<T> const& xHashTree)
	: m_xHashTree(const_cast<Asura_TemplateHashTree<T>&>(xHashTree))
	, m_ppxNodes(NULL)
	, m_uNumNodes(0)
	, m_uCurrentNode(0)
	{
		Begin();
	}
	~Asura_TemplateHashTree_ConstIterator()
	{
		End();
	}

	void Restart()
	{
		End();
		Begin();
	}

	// Query function to see whether iteration is complete...
	bool Done() const
	{
		return(m_ppxNodes == NULL);
	}

	// Function to step onto the next item to be iterated...
	void Next()
	{
		m_uCurrentNode++;

		if(m_uCurrentNode == m_uNumNodes)
		{
			End();
		}
	}

	// Accessor function to get the current entry...
	T const* GetCurrent() const
	{
		if(m_ppxNodes == NULL)
		{
			return(NULL);
		}
		else
		{
			return(&(m_ppxNodes[m_uCurrentNode]->m_xData));
		}
	}

	Asura_Hash_ID GetCurrentHash() const
	{
		if(m_ppxNodes == NULL)
		{
			return(ASURA_HASH_ID_UNSET);
		}
		else
		{
			return(m_ppxNodes[m_uCurrentNode]->m_uHash);
		}
	}

protected:
	void Begin()
	{
		m_uCurrentNode	= 0;
		m_uNumNodes		= m_xHashTree.GetNumberOfNodes();

		if(m_uNumNodes == 0)
		{
			m_ppxNodes	= NULL;
		}
		else
		{
			m_xHashTree.BeginIteration();

			m_ppxNodes	= new Asura_TemplateHashTree_Node<T> const* [m_uNumNodes];

			if(m_xHashTree.m_pxFirstRoot != NULL)
			{
				FillItemArray(m_xHashTree.m_pxFirstRoot);
			}

			m_uCurrentNode	= 0;
		}
	}

	void End()
	{
		if(m_ppxNodes != NULL)
		{
			delete[](m_ppxNodes);

			m_ppxNodes		= NULL;
			m_uCurrentNode	= 0;
			m_uNumNodes		= 0;

			m_xHashTree.EndIteration();
		}
	}

	// Recursive function used to fill the item array with data...
	void FillItemArray(Asura_TemplateHashTree_Node<T> const* pxNode)
	{
		Asura_TemplateHashTree_Node<T> const*	pxLower		= pxNode->m_pxLower;
		Asura_TemplateHashTree_Node<T> const*	pxHigher	= pxNode->m_pxHigher;

		if(pxLower != NULL)
		{
			FillItemArray(pxLower);
		}

		m_ppxNodes[m_uCurrentNode++]	= pxNode;

		if(pxHigher != NULL)
		{
			FillItemArray(pxHigher);
		}
	}

private:
	// Private variables...
	Asura_TemplateHashTree<T>&				m_xHashTree;	// A reference to the hash-tree to iterate over.

	Asura_TemplateHashTree_Node<T> const**	m_ppxNodes;		// An array of pointers to the data to be iterated over.
	u_int									m_uNumNodes;	// The number of items to be iterated over.
	u_int									m_uCurrentNode;	// The index of the current item during iteration.

	// Private functions...
	Asura_TemplateHashTree_ConstIterator(void);
	Asura_TemplateHashTree_ConstIterator(const Asura_TemplateHashTree_ConstIterator& xOther);

	Asura_TemplateHashTree_ConstIterator&	operator=(const Asura_TemplateHashTree_ConstIterator& xOther);

};   // End of class 'Asura_TemplateHashTree_ConstIterator'.


#endif // __ASURA_HASHTREE_H__


/// End of file. ///
