//////////////////////////////////////////////////////////////////////////////
//
// Asura_HashedList.h
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
// (MH - 04.09.2002) template-ised (and simplified) entity list class.
// Used for derived entities which cannot use static_cast or C-style
// cast from Asura_Entity to their type, e.g. entities derived with 
// virtual base class (multiple inheritance) where such a cast is 
// explicitly disallowed. A static_cast from the type to Asura_Entity
// must however be possible or a compile error will occur.
// NOTE: it can be used for non-entities provided they have a GetGuid()
// function and an IsDestroyed() function. I have therefore moved it to 
// its own file and removed references to entity.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_HASHEDLIST_H___
#define ___ASURA_HASHEDLIST_H___ 1

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Debugging/Asura_Assert.h"
#include "Network/Asura_Guid.h"

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Hashed_Node / Asura_Hashed_List
//
//////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_Hashed_Node
{
	public:
		T* m_pxObject;
		Asura_Hashed_Node<T>* m_pxNext;
		bool m_bDeleted;
};

// Forward declaration
template <class T> class Asura_Hashed_List_It;
template <class T> class Asura_Hashed_List_ConstIt;
template <class T> class Asura_Hashed_List_Deleter_It;

// Templatised list.
template <class T> class Asura_Hashed_List
{
	friend class Asura_Hashed_List_It<T>;
    friend class Asura_Hashed_List_ConstIt<T>;
    friend class Asura_Hashed_List_Deleter_It<T>;

	public:
		//If bTreatAsuraGuidUnregisteredAsANormalID then objects with a Guid/ID of 999 can be searched for normally , without
		//it being treated as a special case
		Asura_Hashed_List<T>( const u_int uLog2HashSize = 2, bool bTreatAsuraGuidUnregisteredAsANormalID = false )
		: m_bTreatAsuraGuidUnregisteredAsANormalID(bTreatAsuraGuidUnregisteredAsANormalID) 
		{
			Asura_Assert( uLog2HashSize <= 15, "Unexpected hash size (actual size == 1<<hashSize)" );
			m_uHashSize = 1 << uLog2HashSize;

			m_ppxHeadOfList = new Asura_Hashed_Node<T>*[m_uHashSize];
			for (unsigned int i=0; i<m_uHashSize; i++)
			{
				m_ppxHeadOfList[i] = NULL;
			}

			m_iNumEntries = 0;
			m_uNumIterationsInProgress = 0;
		}

		~Asura_Hashed_List<T>()
        {
			Asura_Assert( (m_uNumIterationsInProgress == 0), "Destroying list with iterator still active. There's a fair chance this will cause a crash." );
#ifdef ASURA_ASSERT
			if (m_ppxHeadOfList)
			{
				for(u_int u = 0; u < m_uHashSize; ++u)
				{
					Asura_Assert( !m_ppxHeadOfList[u], "Destroying hashed list with items still in it, if this is intended you must clear it first." );
				}
			}
#endif
			Clear();
			delete [] m_ppxHeadOfList;
        }

		void Clear()
		{
			Asura_Assert( (m_uNumIterationsInProgress == 0), "Trying to clear the list while iterations are still in progress!" );
			if(!m_ppxHeadOfList) return;

			for(u_int u = 0; u < m_uHashSize; ++u)
			{
				Asura_Hashed_Node<T>* pxNode = m_ppxHeadOfList[u];
				Asura_Hashed_Node<T>* pxNext;
				while(pxNode)
				{
					pxNext = pxNode->m_pxNext;
					delete pxNode;
					pxNode = pxNext;
				}
				m_ppxHeadOfList[u] = NULL;
			}

			m_iNumEntries = 0;
		}

		void AddToList(T* pxObject)
        {
			T*	pxExistingObject	= FindObject(pxObject->GetGuid());

			if(pxExistingObject == NULL)
			{
				int iIndex = GetHashIndex(pxObject->GetGuid());

				Asura_Hashed_Node<T>* pxNewNode = new Asura_Hashed_Node<T>;

				pxNewNode->m_pxObject = pxObject;
				pxNewNode->m_pxNext   = m_ppxHeadOfList[iIndex];
				pxNewNode->m_bDeleted = false;

				m_ppxHeadOfList[iIndex] = pxNewNode;
				m_iNumEntries++;
			}
			else
			{
				Asura_Assertf( false, ("Object (GUID %u) already exists in hashed list", pxObject->GetGuid()) );
			}
        }

        bool RemoveFromListIfPresent( T* pxObject )
        {
			int iIndex = GetHashIndex(pxObject->GetGuid());
		        
			Asura_Hashed_Node<T>* pxNode = m_ppxHeadOfList[iIndex];
			Asura_Hashed_Node<T>* pxPrev = NULL;

	        while (pxNode)
	        {
		        if (pxNode->m_pxObject == pxObject)
		        {
					// RDM - 22/06/06 - Nodes are no longer deleted directly if there's an iteration 
					// currently in progress: they are flagged as deleted and removed by FlushDeletedNodes later
					if( m_uNumIterationsInProgress > 0 )
					{
						pxNode->m_bDeleted = true;
					}
					else
					{
						if (pxPrev)
						{
							pxPrev->m_pxNext = pxNode->m_pxNext;
						}
						else
						{
							m_ppxHeadOfList[iIndex] = pxNode->m_pxNext;
						}

						/// Kevin Lea - 29/11/2001 17:27:43 - add node to free cache?
						delete pxNode;
						m_iNumEntries--;
					}
			        return true;
		        }

				pxPrev = pxNode;
		        pxNode = pxNode->m_pxNext;
	        }

            return false;
        }

		void RemoveFromList( T* pxObject )
        {
#ifdef ASURA_ASSERT
			const bool bFound =
#endif
            RemoveFromListIfPresent( pxObject );
			Asura_Assert( bFound, "Asura_HashedList< T >::RemoveFromList - node not found" );
        }

		inline T* FindObject( const Asura_Guid uGuid, const bool bIgnoreDestroyedEntities = true ) const
        {
            // (MH - 29.10.2002) not sure about this - probably OK to return NULL. Maybe Assert?
			if (uGuid == ASURA_GUID_UNREGISTERED && !m_bTreatAsuraGuidUnregisteredAsANormalID)
            {
                return NULL;
            }

			const int iIndex = GetHashIndex(uGuid);
			Asura_Hashed_Node<T>* pxNode = m_ppxHeadOfList[iIndex];
			while ( pxNode )
	        {
			    if( !pxNode->m_bDeleted && pxNode->m_pxObject->GetGuid() == uGuid )
                {
		            if( bIgnoreDestroyedEntities && pxNode->m_pxObject->IsDestroyed() )
                    {
						// node not valid, skip this node
                    }
                    else
		            {
				        return pxNode->m_pxObject;
		            }
                }
				pxNode = pxNode->m_pxNext;
	        }

	        return NULL;
        }

		inline T* FindEntity( const Asura_Guid uGuid, const bool bIgnoreDestroyedEntities = true ) const
        {
			return FindObject( uGuid, bIgnoreDestroyedEntities );
        }

		int GetHashIndex( const Asura_Guid uGuid ) const
		{
			return (uGuid)&(m_uHashSize-1);
		}

		bool IsEmpty() const { return m_iNumEntries == 0; }
        int GetNumEntries() const { return m_iNumEntries; }

		void FlushDeletedNodes( void ) const
		{
            Asura_Hashed_Node<T>* pxCurrentNode = NULL;			
			
			for( u_int uHashIndex=0; uHashIndex<m_uHashSize; uHashIndex++ )
			{
				pxCurrentNode = m_ppxHeadOfList[uHashIndex];
				Asura_Hashed_Node<T>* pxPreviousNode = NULL;

				while( pxCurrentNode )
				{
					if( pxCurrentNode->m_bDeleted )
					{											
						if( pxPreviousNode )
						{
							pxPreviousNode->m_pxNext = pxCurrentNode->m_pxNext;
						}
						else
						{
							m_ppxHeadOfList[uHashIndex] = pxCurrentNode->m_pxNext;
						}

						Asura_Hashed_Node<T>* pxTempNode = pxCurrentNode;						
						pxCurrentNode = pxCurrentNode->m_pxNext;

						delete pxTempNode;
						m_iNumEntries--;
					}
					else
					{

						pxPreviousNode = pxCurrentNode;
						pxCurrentNode  = pxCurrentNode->m_pxNext;
					}
				}
			}
		}

		void StartIteration( void ) const
		{
			++m_uNumIterationsInProgress;
		}

		void FinishIteration( void ) const
		{
			Asura_Assert( (m_uNumIterationsInProgress > 0), "Asura_HashedList< T >::FinishIteration - there's no iteration in progress?" );

			--m_uNumIterationsInProgress;
			if( m_uNumIterationsInProgress == 0 )
			{
				// no more iterations in progress, so it's now safe to remove any nodes flagged as deleted
				FlushDeletedNodes();		
			}
		}

		const Asura_Hashed_Node<T>*	GetHeadOfList( u_int uIndex ) const	{ return m_ppxHeadOfList[ uIndex ]; }
		u_int						GetHashSize() const					{ return m_uHashSize; }

    protected:
		mutable Asura_Hashed_Node<T>** m_ppxHeadOfList;
		u_int                          m_uHashSize;
		mutable u_int m_uNumIterationsInProgress;
        mutable int m_iNumEntries;
		bool m_bTreatAsuraGuidUnregisteredAsANormalID;
};

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Hashed_List_It
//
//////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_Hashed_List_It
{

public:

	//constructor for iterator , also starts the iteration
	Asura_Hashed_List_It<T>( const Asura_Hashed_List<T>* pxList, const bool bIgnoreDestroyedEntities = true )
	{
		Asura_Assert(pxList,"");
		m_pxList = pxList;
		m_pxList->StartIteration();

		Restart( bIgnoreDestroyedEntities );
	}

	~Asura_Hashed_List_It<T>()
	{
		m_pxList->FinishIteration();
	}

	//gets the current entity in the iteration
	T* GetCurrent() const
	{
		Asura_Assert(m_pxCurrentNode,"");
		return m_pxCurrentNode->m_pxObject;
	}

    T* operator()( void ) const { return GetCurrent(); }

	//advance to the next entity in the list
	void Next( const bool bIgnoreDestroyedEntities = true )
	{
		//return to our place in the loop , if this isn't the first time
		if (m_pxCurrentNode) goto ReEnterLoop;
		
		for (; m_uHashIndex < m_pxList->m_uHashSize; m_uHashIndex++)
		{
			m_pxCurrentNode = m_pxList->m_ppxHeadOfList[m_uHashIndex];

			while(m_pxCurrentNode)
			{
                if ( !m_pxCurrentNode->m_bDeleted && ( !bIgnoreDestroyedEntities || !m_pxCurrentNode->m_pxObject->IsDestroyed() ) )
                {
				    return;
                }

				ReEnterLoop:

				m_pxCurrentNode = m_pxCurrentNode->m_pxNext;
			}
		}
	}

	//indicates whether the iteration has finished
	bool Done() const {return m_pxCurrentNode == 0;}

	//restarts the iteration
	void Restart( const bool bIgnoreDestroyedEntities = true )
	{
		//reset values
		m_uHashIndex = 0;
		m_pxCurrentNode = 0;
		//and find the first entry
		Next( bIgnoreDestroyedEntities );
	}

private:

    // Prevent copies being made, since currently it's unsafe.
    Asura_Hashed_List_It<T>( const Asura_Hashed_List_It<T>& xOtherIt );
    Asura_Hashed_List_It<T>& operator=( const Asura_Hashed_List_It<T>& xOtherIt );

	const Asura_Hashed_List<T>* m_pxList;
	unsigned int m_uHashIndex;
	const Asura_Hashed_Node<T>* m_pxCurrentNode;

};

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Hashed_List_ConstIt
//
//////////////////////////////////////////////////////////////////////////////

template< typename T >
class Asura_Hashed_List_ConstIt : public Asura_Hashed_List_It<T>
{

public:

    Asura_Hashed_List_ConstIt<T>( const Asura_Hashed_List<T>* pxList ) : Asura_Hashed_List_It<T>( pxList ) { }

	const T* GetCurrent( void ) const { return Asura_Hashed_List_It<T>::GetCurrent(); }

    const T* operator()( void ) const { return GetCurrent(); }

private:

    // Prevent copies being made, since currently it's unsafe.
    Asura_Hashed_List_ConstIt<T>( const Asura_Hashed_List_ConstIt<T>& xOtherIt );
    Asura_Hashed_List_ConstIt<T>& operator=( const Asura_Hashed_List_ConstIt<T>& xOtherIt );

};

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Hashed_List_Deleter_It
//
//////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_Hashed_List_Deleter_It
{

public:

	//constructor for iterator , also Asura_Hashed_List_Deleter_It the iteration
	Asura_Hashed_List_Deleter_It<T>(Asura_Hashed_List<T>* pxList)
	{
		Asura_Assert(pxList,"");
		m_pxList = pxList;
		Restart();
	}

	//gets the current entity in the iteration
	T* GetCurrent() const
	{
		Asura_Assert(m_pxCurrentNode,"");
		return m_pxCurrentNode->m_pxObject;
	}

    //deletes the current entity and moves to the next in the list
    void DeleteCurrent()
    {
		Asura_Assert(m_pxCurrentNode,"");
        Asura_Hashed_Node<T>* pxToDelete = m_pxCurrentNode;
        Next();
        m_pxList->RemoveFromList( pxToDelete->m_pxObject );
    }

	//advance to the next entity in the list
	void Next()
	{
		//return to our place in the loop , if this isn't the first time
		if (m_pxCurrentNode) goto ReEnterLoop;
		
		for (; m_uHashIndex < m_pxList->m_uHashSize; m_uHashIndex++)
		{
			m_pxCurrentNode = m_pxList->m_ppxHeadOfList[m_uHashIndex];

			while(m_pxCurrentNode)
			{
				return;

				ReEnterLoop :

				m_pxCurrentNode = m_pxCurrentNode->m_pxNext;
			}
		}
	}

	//indicates whether the iteration has finished
	bool Done() const {return m_pxCurrentNode == 0;}

	//restarts the iteration
	void Restart()
	{
		//reset values
		m_uHashIndex = 0;
		m_pxCurrentNode = 0;
		//and find the first entry
		Next();
	}

private:

    // Prevent copies being made, since currently it's unsafe.
    Asura_Hashed_List_Deleter_It<T>( const Asura_Hashed_List_Deleter_It<T>& xOtherIt );
    Asura_Hashed_List_Deleter_It<T>& operator=( const Asura_Hashed_List_Deleter_It<T>& xOtherIt );

	Asura_Hashed_List<T>* m_pxList;
	unsigned int m_uHashIndex;
	Asura_Hashed_Node<T>* m_pxCurrentNode;

};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // #ifndef ___ASURA_HASHEDLIST_H___
