//
// Asura_LBAABB_Tree.h
//
// Created by KJFL - Wednesday, January 19, 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_LBAABB_TREE_H__
#define __ASURA_LBAABB_TREE_H__

#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_ResourceAllocator.h"

class Asura_Chunk_Stream;

typedef void(*Asura_LBAABB_Callback)(const u_int uIndex);

// 17/08/05 - Andrew Weinkove
//
// This LBAABB tree structure is fairly simple. Each child bounding box is constructed
// by taking an offsetted version of the parent bounding box then clipping it to the 
// parent bounding box.
//
// The structure is 48 bits per node - in order to prevent padding of the structure to
// 64 bits, the nodes are stored in pairs.
class Asura_LBAABB_Tree_Double_Node
{
	public:
		union
		{
			struct
			{
				// Don't actually reference these anywhere, but they are useful to help 
				// understand the structure and handy in the debugger.
				char m_cNode1LeftOffsetX;
				char m_cNode1LeftOffsetY;
				char m_cNode1LeftOffsetZ;
				char m_cNode1RightOffsetX;
				char m_cNode1RightOffsetY;
				char m_cNode1RightOffsetZ;
				char m_cNode2LeftOffsetX;
				char m_cNode2LeftOffsetY;
				char m_cNode2LeftOffsetZ;
				char m_cNode2RightOffsetX;
				char m_cNode2RightOffsetY;
				char m_cNode2RightOffsetZ;
			};
			char m_acElements[12];
			char m_aacElements[2][6];
		};

		void ReadFromChunkStream (Asura_Chunk_Stream& xStream);
		void WriteToChunkStream (Asura_Chunk_Stream& xStream) const;
};


// KJFL - Friday, January 28, 2005
//
// A stack of these elements is required for non-recursive tree-search.
// Each stores a node index, and the parent bounding box of that node.
//
class Asura_LBAABB_Tree_StackElement
{
	public:

		u_int				m_uIndex;
		Asura_Bounding_Box	m_xBB;
};

// KJFL - Friday, January 28, 2005 - 
//
// Intentionally minimal tree class merely containing query methods.
// Creation of trees is done with a corresponding Tools class.
//
// Note no allocation or deallocation of node data is done here; this
// is to allow nodes to be allocated in one huge array whose memory
// is handled separately. This means many trees can be loaded into
// memory in one go in a single allocated chunk (after which the node
// pointers will have to be set up accordingly).
//
// Note all test methods are entirely const.
//		
class Asura_LBAABB_Tree
{
	public:
        
		Asura_LBAABB_Tree()
		{
			m_uNumberOfNodes = 0;
   			m_xRootBoundingBox.Reset();
			m_pxNodes = NULL;
		}

		void TestTreeWithPoint( const Asura_Vector_3& xPos, Asura_LBAABB_Callback pfnOutputCallback) const; 

		void TestTreeWithBB( const Asura_Bounding_Box& xTestBB, Asura_LBAABB_Callback pfnOutputCallback, const u_int uCache = 0  ) const; 
		
		void TestTreeWithRay( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, float* pfDistance, Asura_LBAABB_Callback pfnOutputCallback, const u_int uCache = 0 ) const;

		const Asura_Bounding_Box& GetBoundingBox( void ) const
		{
			return m_xRootBoundingBox;
		}

		// Number of nodes in the structure.
		u_int GetNumberOfNodes( void ) const
		{
			return m_uNumberOfNodes;
		}

		// Number of double nodes required to represent the nodes.
		u_int GetNumberOfDoubleNodes(void) const
		{
			return (m_uNumberOfNodes + 1) >> 1;
		}

		// Warning! Asura_Physics_Environment allocates nodes in one big heap in exported levels.
		// If this is the case then do not delete them per-tree.
		void DeleteNodes( void )
		{
			Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
			{
				Asura_ResourceDeallocateArray(m_pxNodes);
				m_pxNodes = 0;
			}
			Asura_ResourcePopType();
		}

		// JGP 11/05/06 - New functions for reading header and nodes
		void ReadTreeAndNodesFromChunkStream( Asura_Chunk_Stream& xStream );
		void WriteTreeAndNodesToChunkStream( Asura_Chunk_Stream& xStream, const Asura_Vector_3& xDisplacement = Asura_Maths::ZeroVector3 );

		void ReadHeaderFromChunkStream (Asura_Chunk_Stream& xStream);
		//Displacement is position to add to tree upon save.
		void WriteHeaderToChunkStream (Asura_Chunk_Stream& xStream,const Asura_Vector_3& xDisplacement = Asura_Maths::ZeroVector3) const;

	private:
		// Optimised version of TestTreeWithRay for the surprisingly common case where direction is (0, 1, 0).
		void TestTreeWithRay_VertDown( const Asura_Vector_3& xRayOrigin, float* pfDistance, void(*pfnRayTestCallback)( const u_int uIndex ), const u_int uCache = 0 ) const;

		friend class Asura_Chunk_EnvironmentLBAABBTrees; // TODO: Move Asura_Chunk_EnvironmentLBAABBTrees over to using the fuller read and write functions so this friendship is no longer needed.
		friend class Tools_LBAABB_Tree;
		friend class Asura_Physics_Zone;

		u_int						m_uNumberOfNodes;
   		Asura_Bounding_Box			m_xRootBoundingBox;
	    Asura_LBAABB_Tree_Double_Node* m_pxNodes;
};


#ifdef ASURA_TOOLS


class Tools_LBAABB_Tree_Node
{
	public:
		u_int 	m_uLeftIndex;
		u_int 	m_uRightIndex;

		u_int 	m_uNumberOfLeavesToLeft;
		u_int 	m_uNumberOfLeavesToRight;

		bool	m_bLeftIsLeaf;
		bool	m_bRightIsLeaf;
};

class Tools_LBAABB_Tree
{
	public:
		// Caller is responsible for deleting the returned array.
		static u_int* CreateRunTimeTree( Asura_LBAABB_Tree& xTree, const u_int uNumberOfLeaves, const Asura_Bounding_Box* pxLeafBBs, const Asura_Vector_3* pxLeafCentres  )
		{
			Tools_LBAABB_Tree* pxTree = new Tools_LBAABB_Tree( uNumberOfLeaves, pxLeafBBs, pxLeafCentres );

			xTree = *(pxTree->m_pxRunTimeTree);

			u_int* puTranslationArray = pxTree->m_puLeafTranslationArray;
			delete pxTree;

			return puTranslationArray;
		}
    	
	private:

		Tools_LBAABB_Tree( const u_int uNumberOfLeaves, const Asura_Bounding_Box* pxLeafBBs, const Asura_Vector_3* pxLeafCentres  );

		~Tools_LBAABB_Tree();
	
		void SortNodes( const u_int uCurrentNode, const u_int uStartLeaf, const u_int uEndLeaf, const Asura_Bounding_Box& xParentBB );

		Tools_LBAABB_Tree_Node* 			m_pxNodes;
		u_int								m_uNumberOfNodes;
		Asura_Bounding_Box					m_xRootBoundingBox;
		Asura_LBAABB_Tree*					m_pxRunTimeTree;
		u_int*								m_puLeafTranslationArray;


		static int Tools_LBAABB_Tree::CompareForLeavesInX( const void *pC1, const void *pC2 );
		static int Tools_LBAABB_Tree::CompareForLeavesInY( const void *pC1, const void *pC2 );
		static int Tools_LBAABB_Tree::CompareForLeavesInZ( const void *pC1, const void *pC2 );

		static u_int*						s_puLeafArray;
		static u_int*						s_puLeafArraySortedInX;
		static u_int*						s_puLeafArraySortedInY;
		static u_int*						s_puLeafArraySortedInZ;
		static const Asura_Bounding_Box*	s_pxLeafBBs;
		static const Asura_Vector_3*		s_pxLeafCentres;


};

#endif //ASURA_TOOLS

#endif//__ASURA_LBAABB_TREE_H__
