//////////////////////////////////////////////////////////////////////////////
//
// File: Asura_LOS_Penetration.cpp
//
// Created by MGH 13/12/2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//
// Description: Class does LOS tests for penetrating projectiles (eg bullets).
//              Given stopping powers of various materials, it returns
//              multiple hits.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "Core/Asura_LOS_Penetration.h"

#include "Animation/Asura_AnimationAttachment.h"
#include "Core/Asura_DynamicBVH.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_MaterialFlags.h"
#include "Entity/Asura_Entity_Renderable.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "Network/Asura_Network.h"
#include "Physics/Asura_Physics_Environment.h"
#include "Rendering/Asura_Module.h"

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "Entity/Tools_Entity_Object.h"
#endif //ASURA_TOOLS

/////////////
// Defines //
/////////////
// Provide a default value for the maximum number of hits that can be returned.
#ifndef ASURA_MAX_NUM_LOS_PENETRATION_HITS
	#define ASURA_MAX_NUM_LOS_PENETRATION_HITS 12
#endif //ASURA_MAX_NUM_LOS_PENETRATION_HITS

///////////////
// Constants //
///////////////
// When we restart a LOS test, move the new origin on by this extra amount so that
// we don't hit the same polygon again.
const float fASURA_ADJUST_DISTANCE = 0.01f;

/////////////////
// Static data //
/////////////////
static Asura_FixedArray<Asura_LOS_OutputData, ASURA_MAX_NUM_LOS_PENETRATION_HITS> g_axOutputArray;
static Asura_FixedArray<float, ASURA_MAX_NUM_LOS_PENETRATION_HITS> g_afStoppingPower;

// Indices into output/stopping arrays, in order of increasing distance.
static Asura_FixedArray<int, ASURA_MAX_NUM_LOS_PENETRATION_HITS> g_aiSortIndex;

// Temporary copy of input data.
static Asura_LOS_InputData g_xInputData(Asura_Maths::ZeroVector3, Asura_Maths::ZeroVector3, ASURA_GUID_UNREGISTERED);

int Asura_LOS_Penetration::s_iNumHits = 0;
float Asura_LOS_Penetration::s_fMaxDistance = 0.0f;

/////////////////
// ResetOutput //
/////////////////
// Helper function to reset LOS output data to initial value
// MGH 5/2/2008: Replaced with call to Asura_LOS_OutputData::Reset().
void ResetOutput(Asura_LOS_OutputData& xOutput, float fMaxDistance)
{
	xOutput.Reset(fMaxDistance);
}


////////////
// GetHit //
////////////
// Return results of LOS test. Sorted in order of increasing distance.
const Asura_LOS_OutputData* Asura_LOS_Penetration::GetHit(int iHitNumber)
{
	if (iHitNumber>=0 && iHitNumber<s_iNumHits)
	{
		const int iIndex = g_aiSortIndex[iHitNumber];
		if (iIndex>=0 && iIndex<ASURA_MAX_NUM_LOS_PENETRATION_HITS)
		{
			return &g_axOutputArray[iIndex];
		}
	}

	Asura_Assert(0, "Invalid LOS hit index");
	return 0;
}

///////////////////////////
// GetTotalStoppingPower //
///////////////////////////
// Gives the cumulative stopping power up to and including the specified hit,
// hits are sorted by distance as for GetHit().
// Returns 0.0f (not slowed) to 1.0f (stopped).
// Note: passing in iHitNumber equal to -1 will return 0.0f.
float Asura_LOS_Penetration::GetTotalStoppingPower(int iHitNumber)
{
	if (iHitNumber<0)
	{
		return 0.0f;
	}
	if (iHitNumber>=s_iNumHits)
	{
		return fASURA_MAX_STOPPING_POWER;
	}

	float fTotal = 0.0f;

	for (int i=0; i<=iHitNumber; ++i)
	{
		const int iIndex = g_aiSortIndex[i];
		if (iIndex>=0 && iIndex<ASURA_MAX_NUM_LOS_PENETRATION_HITS)
		{
			fTotal += g_afStoppingPower[iIndex];
		}
		else
		{
			Asura_Assert(0, "Invalid LOS sort index");
		}
	}

	// In case of numerical error during the summation...
	if (fTotal>fASURA_MAX_STOPPING_POWER*0.9999f)
	{
		fTotal = fASURA_MAX_STOPPING_POWER;
	}
	return fTotal;
}

//////////////////
// HasHitEntity //
//////////////////
// Search through the impacts to see whether a specific entity was hit.
// Useful when we want to check whether a bullet hit its intended target.
bool Asura_LOS_Penetration::HasHitEntity(Asura_Guid uGuid)
{
	for (int i=0; i<s_iNumHits; ++i)
	{
		const Asura_LOS_OutputData* const pxHit = GetHit(i);
		if (pxHit && pxHit->m_uHitEntity == uGuid)
		{
			return true;
		}
	}

	return false;
}


//
// Functor for Asura_DynamicBVH::TestRay_Functor().
//
class Asura_DynamicBVH_Penetration_Functor
{
public:
	Asura_DynamicBVH_Penetration_Functor(int iLOSCounter, float fInitialMaxDistance, const Asura_Penetration& xFunctor)
		: m_xFunctor(xFunctor)
		, m_iLOSCounter(iLOSCounter)
		, m_fInitialMaxDistance(fInitialMaxDistance)
		, m_bStop(iLOSCounter == 0)
	{
	}

	// Check for a LOS hit on the leaf object.
	// If there is a hit then update the output data and maximum distance and return true.
	// If there is no hit then just return false and leave the output data
	// and maximum distance unchanged.
	bool CheckEntity(Asura_Entity_Renderable* pxEntity, const Asura_LOS_InputData& xInputData)
	{
		Asura_Assert(pxEntity, "Null leaf object pointer");

		const int iNumHits = Asura_LOS_Penetration::GetNumberOfHits();
		const bool bHit = pxEntity->LOSTest(&xInputData, &g_axOutputArray[iNumHits]);
		if (bHit)
		{
			if (!Asura_LOS_Penetration::RegisterHit(m_xFunctor, pxEntity->GetClassification()))
			{
				// Output array is full.
				m_bStop = true;
			}

			if (m_iLOSCounter > 0)
			{
				--m_iLOSCounter;
				if (m_iLOSCounter == 0)
				{
					// Hit limit of counter.
					m_bStop = true;
				}
			}
		}
		return bHit;
	}
	
	// Return the initial maximum distance.
	float GetInitialMaxDistance() const { return m_fInitialMaxDistance; }

	// Get the current maximum distance.
	float GetMaxDistance() { return Asura_LOS_Penetration::GetMaxDistance(); }

	// Return true to stop the search (typically happens when testing
	// for multiple hits and the output array is too full to continue).
	bool ShouldStop() const { return m_bStop; }

private:
	const Asura_Penetration& m_xFunctor;
	int m_iLOSCounter;
	float m_fInitialMaxDistance;

	bool m_bStop;
};



///////////////////////////
// TestAgainstEverything //
///////////////////////////
// Do a LOS test against everything, using the given input data.
// A functor provides callbacks for the stopping power of different materials and entities.
void Asura_LOS_Penetration::TestAgainstEverything(const Asura_LOS_InputData* const pxData,
												  const Asura_Penetration& xFunctor)
{
	// First check the environment
	TestAgainstEnvironment(pxData, xFunctor);

	if (!pxData) return;

	if (s_iNumHits >= ASURA_MAX_NUM_LOS_PENETRATION_HITS)
	{
		// We have run out of space. Skip the entity test.
		// Note: This should never happen, we reserved one slot for the entities.
		Asura_Assert(0, "Error ran out of space in the LOS array");
		return;
	}

	// Initialise output data, copy across max range from the environment test.
	ResetOutput(g_axOutputArray[s_iNumHits], GetMaxDistance());

    const Asura_Guid uIgnore = pxData->m_uEntityToIgnore;
	
	int iLoSHitCounter = xFunctor.GetMaxEntityLoSHits();

	// MGH 14/12/2004: Note: this linear search over entities is due to be replaced
	// (at some time in the future) by a more efficient method.
	// Ideally the new algorithm will test the entities in order from closest to furthest.
	
	//TY 03.12.07 - This function is useful for Tools sound occlusion, but since we have no client entities
	// we need to fudge the test a little, i.e. use server entities in tools mode
	#ifdef ASURA_TOOLS
	if ( !Tools_System::IsActive() && Asura_Network::IsClient() )
	#else
    if ( Asura_Network::IsClient() )
	#endif
    {
		// MGH 4/2/2008: Use the shiny new Dynamic BVH code, to speed up the LOS testing.
		Asura_DynamicBVH_Penetration_Functor xCallback(iLoSHitCounter, GetMaxDistance(), xFunctor);
		Asura_DynamicBVH<Asura_Entity_Renderable>& xBVH = Asura_Entity_System::GetLineOfSightBVH();
		xBVH.TestRay_Functor(*pxData, xCallback);
    }
    else
    {
		// Dedicated servers don't have client entities.
		Asura_Hashed_List_It<Asura_ServerEntity_PhysicalObject> xIt(Asura_ServerEntity_PhysicalObject::GetList(), true);
        for ( ; (!xIt.Done() && iLoSHitCounter!=0); xIt.Next() )
        {
            const Asura_ServerEntity_PhysicalObject* const pxEntity = xIt.GetCurrent();
			if (pxEntity->IsDestroyed()) continue;
			
			#ifdef ASURA_TOOLS
			//tools object's have some additional early out tests for LoS
			if(Tools_System::IsActive())
			{
				Tools_Entity_Object* pxToolsEntity = Tools_Entity_Object::FindEntity(pxEntity->GetGuid());
				if(pxToolsEntity && !pxToolsEntity->ShouldTestForLOS()) 
				{
					continue;
				}
			}
			#endif //ASURA_TOOLS

            if ( uIgnore == ASURA_GUID_UNREGISTERED || uIgnore != pxEntity->GetGuid() )
		    {
			    if (pxEntity->LOSTest(pxData, &g_axOutputArray[s_iNumHits]))
				{
					if (!RegisterHit(xFunctor, pxEntity->GetClassification()))
					{
						break;
					}
					if(iLoSHitCounter > 0)
					{
						--iLoSHitCounter;
					}
				}
		    }
        }
    }

	// In theory the hits should be sorted by distance and the max distance
	// should have been set by this point.
	float fStoppingPower = 0.0f;
	for (int i=0; i<s_iNumHits; ++i)
	{
		const int iIndex = g_aiSortIndex[i];
		const float fThisDistance = g_axOutputArray[iIndex].m_fDistanceToIntersection;
		
		// Sanity check. Make sure that the array is properly sorted.
#ifdef ASURA_ASSERT
		Asura_Assert(s_fMaxDistance >= fThisDistance, "Error: bad max distance in LOS");
		if (i>0)
		{
			const int iPrev = g_aiSortIndex[i-1];
			const float fPrevDistance = g_axOutputArray[iPrev].m_fDistanceToIntersection;
			Asura_Assert(fPrevDistance <= fThisDistance, "Error: badly sorted LOS array");
		}
#endif //ASURA_ASSERT
		
		fStoppingPower += g_afStoppingPower[iIndex];

		if (fStoppingPower >= fASURA_MAX_STOPPING_POWER)
		{
			// Adjust number of hits to exclude obsolete ones beyond max distance.
			s_iNumHits = i+1;

			// Max distance should already have been set, but no harm in setting it again.
			Asura_Assert(s_fMaxDistance == fThisDistance, "Error: wrong max distance in LOS");
			s_fMaxDistance = fThisDistance;
			break;
		}	
	}
}

/////////////////
// RegisterHit //
/////////////////
// Register a possible hit on an entity.
// Return false if the array is full and no further hits can be stored.
bool Asura_LOS_Penetration::RegisterHit(const Asura_Penetration& xFunctor,
										unsigned short usClassification)
{
	if (s_iNumHits >= ASURA_MAX_NUM_LOS_PENETRATION_HITS)
	{
		// MGH 30/1/2008: It should never be possible to reach this code.
		Asura_Assert(0, "Error: trying to write past end of LOS penetration arrays");
		return false;
	}

	const Asura_LOS_OutputData& xOutput = g_axOutputArray[s_iNumHits];
	const float fThisStoppingPower = xFunctor.GetStoppingPower(xOutput, usClassification);
	g_afStoppingPower[s_iNumHits] = fThisStoppingPower;
	g_aiSortIndex[s_iNumHits] = s_iNumHits;

	const float fThisDistance = xOutput.m_fDistanceToIntersection;

	float fTotalStoppingPower = 0.0f;
	bool bInserted = false;

	// Sort by distance and keep track of the stopping power.
	for (int i=0; i<=s_iNumHits; ++i)
	{
		const int iIndex = g_aiSortIndex[i];
		const Asura_LOS_OutputData& xSortedOutput = g_axOutputArray[iIndex];

		float fSortedDistance = xSortedOutput.m_fDistanceToIntersection;

		if (bInserted || fSortedDistance <= fThisDistance)
		{
			// Element is in correct place in array.
			fTotalStoppingPower += g_afStoppingPower[iIndex];
		}
		else
		{
			// New element needs to be inserted here.
			if (s_iNumHits>0)
			{
				// Move other elements upwards to make space.
				for (int j=s_iNumHits-1; j>=i; --j)
				{
					g_aiSortIndex[j+1] = g_aiSortIndex[j];
				}
				// Insert new element.
				g_aiSortIndex[i] = s_iNumHits;
				bInserted = true;
			}

			fTotalStoppingPower += fThisStoppingPower;
			fSortedDistance = fThisDistance;
		}

		if (fTotalStoppingPower >= fASURA_MAX_STOPPING_POWER)
		{
			// Projectile has been stopped. Update maximum distance.
			s_fMaxDistance = fSortedDistance;
			break;
		}
	}

	s_iNumHits++;

	if (s_iNumHits>=ASURA_MAX_NUM_LOS_PENETRATION_HITS)
	{
#ifdef ASURA_ASSERT
		// MGH 31/1/2008: If the maximum number of hits is restricted so as not to exceed
		// the size of the array, then there is no need to assert that the array is full.
		const int iMaxEnvHits = xFunctor.GetMaxEnvironmentLoSHits();
		const int iMaxEntHits = xFunctor.GetMaxEntityLoSHits() ;
		const bool bCantOverflow = ( iMaxEnvHits >= 0 &&
									 iMaxEntHits >= 0 && 
									(iMaxEnvHits + iMaxEntHits) <= ASURA_MAX_NUM_LOS_PENETRATION_HITS );

		// Could happen if we shoot through a large number of entities.
		Asura_Assert(bCantOverflow, "Warning: ran out of space in the LOS array");
#endif //ASURA_ASSERT

		// Break out of the loop over entities.
		// If the entities are being tested in order from closest to furthest then
		// this means that hits on any further away entities will be ignored.
		return false;
	}

	// Set up the next output.
	ResetOutput(g_axOutputArray[s_iNumHits], GetMaxDistance());

	return true;
}


////////////////////////////
// TestAgainstEnvironment //
////////////////////////////
void Asura_LOS_Penetration::TestAgainstEnvironment(const Asura_LOS_InputData* const pxData,
												   const Asura_Penetration& xFunctor)
{
	s_iNumHits = 0;
	s_fMaxDistance = 0.0f; // Furthest hit so far.

	Asura_Assert(pxData, "Null input data pointer");
	if (!pxData)
	{
		return;
	}

    ++Asura_LOS::s_uLOSTestCount;

	Asura_LOS::s_uPolyFlagsToIgnore  = pxData->m_uPolyFlagsToIgnore;
	Asura_LOS::s_uPolyFlagsToTestFor = pxData->m_uPolyFlagsToTestFor;

	float fTotalStoppingPower = 0.0f;

	// Point to original input data initially.
	const Asura_LOS_InputData* pxInputData = pxData;

	int iLoSHitCounter = xFunctor.GetMaxEnvironmentLoSHits();

	while ( iLoSHitCounter != 0 )
	{
		Asura_LOS_OutputData& xOutput = g_axOutputArray[s_iNumHits];

		// Setup output data and call LOS routine.
		// Initialise output data, copy across max range
		ResetOutput(xOutput, pxInputData->m_fInitialMaximumRange);

		xOutput.m_bHitEnvironment = Asura_Physics_Environment::FindPolyInLOS( pxInputData, &xOutput );

		if (!xOutput.m_bHitEnvironment)
		{
			// Didn't hit anything. Stop searching.
			break;
		}
		
		// Hit environment.
		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			g_afStoppingPower[s_iNumHits] = xFunctor.GetStoppingPower(xOutput.m_uMaterialResponseHashID);
		#else
			g_afStoppingPower[s_iNumHits] = xFunctor.GetStoppingPower(xOutput.m_sMaterialID);
		#endif

		fTotalStoppingPower += g_afStoppingPower[s_iNumHits];

		s_iNumHits++;
		if (s_iNumHits >= ASURA_MAX_NUM_LOS_PENETRATION_HITS-1)
		{
			// Run out of space in the array, make this hit stop the projectile.
			// Note: we have reserved one slot in the array for any entity test.
			g_afStoppingPower[s_iNumHits-1] = fASURA_MAX_STOPPING_POWER;
			fTotalStoppingPower = fASURA_MAX_STOPPING_POWER;
		}

		if (fTotalStoppingPower>=fASURA_MAX_STOPPING_POWER)
		{
			// Projectile ran out of power.
			break;
		}

		// Set up next test
		if (s_iNumHits == 1)
		{
			g_xInputData = *pxData;		 // Copy input data to temporary data.
			pxInputData = &g_xInputData; // Set pointer to temporary data.
		}
		
		// Adjust the origin and distance.
		const float fMoveDistance = xOutput.m_fDistanceToIntersection + fASURA_ADJUST_DISTANCE;
		g_xInputData.m_xOrigin += g_xInputData.m_xDirection * fMoveDistance;
		g_xInputData.m_fInitialMaximumRange -= fMoveDistance;

		// (AW - 30.03.2005) - Adjust the remaining range if 
		// we've reached the stopping power threshold.
		if (xFunctor.GetThresholdBeforeMaxDistanceKicksIn() < fTotalStoppingPower)
		{
			g_xInputData.m_fInitialMaximumRange = Asura_Maths::Min(g_xInputData.m_fInitialMaximumRange, xFunctor.GetMaxDistanceAfterEnvironmentImpact());
		}

		if (g_xInputData.m_fInitialMaximumRange < 0.0f)
		{
			// Projectile max range is just beyond the last hit.
			break;
		}

		if( xFunctor.GetMaxEnvironmentLoSHits() > 0)
		{
			//decrement our LoS hit limiter (assuming it is a valid number)
			-- iLoSHitCounter;
		}
	}

	// Fix up distances in the output which are currently relative to the previous
	// intersection. We want them to be relative to the initial input data.
	s_fMaxDistance = 0.0f;
	for (int i=0; i<s_iNumHits; ++i)
	{
		Asura_LOS_OutputData& xOutput = g_axOutputArray[i];
		s_fMaxDistance += xOutput.m_fDistanceToIntersection;

		if (i>0)
		{
			s_fMaxDistance += fASURA_ADJUST_DISTANCE;
			xOutput.m_fDistanceToIntersection = s_fMaxDistance;
		}

		// Results are already in order of increasing distance.
		g_aiSortIndex[i] = i;
	}
	
	if (fTotalStoppingPower<fASURA_MAX_STOPPING_POWER)
	{
		if (s_iNumHits)
		{
			// MGH 31/1/2008: Fixed this, as it wasn't adding back the adjustment we subtracted off earlier.
			// (AW - 30.03.2005) - Add on the remaining distance from the LOS.
			s_fMaxDistance += g_xInputData.m_fInitialMaximumRange + fASURA_ADJUST_DISTANCE;
		}
		else
		{
			// No impacts for the projectile.
			s_fMaxDistance = pxData->m_fInitialMaximumRange;
		}
	}
}

//////////
// Dump //
//////////
// Print out contents of arrays, for debugging purposes.
#ifdef _DEBUG
#include "Debugging/Asura_Logfile.h"
void Asura_LOS_Penetration::Dump()
{
	Asura_OutputToDebugger(("LOS: NumHits = %d  MaxDistance = %f\n", s_iNumHits, Asura_Maths::FtoD(s_fMaxDistance)));
	
	float fStop = 0.0f;
	for (int i=0; i<s_iNumHits; ++i)
	{
		const int iIndex = g_aiSortIndex[i];
		const float fDist = g_axOutputArray[iIndex].m_fDistanceToIntersection;
		const Asura_Guid uGuid = g_axOutputArray[iIndex].m_uHitEntity;

		fStop += g_afStoppingPower[iIndex];

		// Index in sort array, index in output array, distance, stopping power, cumulative stopping power, (guid).
		if (g_axOutputArray[iIndex].m_bHitEnvironment)
		{
			Asura_OutputToDebugger(("%3d: (%d)  dist=%11.6f   stop=%.3f  total=%.3f\n", i, iIndex, Asura_Maths::FtoD(fDist), Asura_Maths::FtoD(g_afStoppingPower[iIndex]), Asura_Maths::FtoD(fStop)));
		}
		else 
		{
			Asura_OutputToDebugger(("%3d: (%d)  dist=%11.6f   stop=%.3f  total=%.3f  guid=%u\n", i, iIndex, Asura_Maths::FtoD(fDist), Asura_Maths::FtoD(g_afStoppingPower[iIndex]), Asura_Maths::FtoD(fStop), uGuid));
		}
	}
}
#endif //_DEBUG
