//////////////////////////////////////////////////////////////////////////////
//
// Asura_MaterialResponse_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#include "Core/Asura_MaterialResponse_System.h"

// Engine independent.
#include "Animation/Asura_TextureAnimation.h"
#include "Core/Asura_HashTable.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_MaterialFlags.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Debug_Timers.h"

// Engine tools.
#if defined ASURA_TOOLS
#include "Materials/Tools_MaterialResponse_System.h"
#include "Rendering/Asura_Direct3D_MaterialResponse.h"
#endif

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Asura_HashTable<Asura_MaterialResponse> g_xMaterialHashTable( ASURA_MATERIAL_RESPONSE_HASH_TABLE_SIZE );

#if defined ASURA_TOOLS
// KJFL - 22 December 2006 - slightly hacky fake material for when in Tools,
// most likely want to refactor this.
// RAR 15/06/2007 - Need to be able to have at least the most recent two materials
// stay valid to deal with having two materials per polygon.
// Not so nice ,but I guess it will work.
static Asura_MaterialResponse g_axToolsMaterial[2];
static int g_iToolsMaterialIndex = 0;
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_MaterialResponse_System::MaterialProjectContextCallback
    Asura_MaterialResponse_System::s_pfnMaterialProjectContextCallback;

Asura_MaterialResponse_System::AlbedoModifierCallback
    Asura_MaterialResponse_System::s_pfnAlbedoModifierCallback;

Asura_MaterialResponse_System::OverrideMapTree Asura_MaterialResponse_System::s_xOverrideMapTree;

Asura_Hash_ID   Asura_MaterialResponse_System::s_uCurrentOverrideMapHash    = ASURA_HASH_ID_UNSET;
Asura_MaterialResponse_OverrideMap* Asura_MaterialResponse_System::s_pxCurrentOverrideMap = NULL;

int Asura_MaterialResponse_System::s_iNumGlobalMaterials = 0;
bool Asura_MaterialResponse_System::s_bGlobalResetPointHit = false;

u_int Asura_MaterialResponse_System::s_uNumMaterialsTextureLines = 0;

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
Asura_Hash_ID   Asura_MaterialResponse_System::s_uDefaultSurfaceTypeHash    = ASURA_HASH_ID_UNSET;
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

#if defined(ASURA_TOOLS)
Asura_Hash_ID*  Asura_MaterialResponse_System::s_puTraversalHashes  = NULL;
const char**    Asura_MaterialResponse_System::s_pszTraversalNames  = NULL;
u_int           Asura_MaterialResponse_System::s_uTraversalIndex    = 0;

bool            Asura_MaterialResponse_System::s_bUseUVOrientMaterial = false;
#endif  // defined(ASURA_TOOLS)


/******************************************************************************/
/* Asura_MaterialResponse_System Static Class Functions                       */
/******************************************************************************/

void Asura_MaterialResponse_System::Initialise()
{
    ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_MaterialResponse_System::Initialise");

    s_uNumMaterialsTextureLines = 0;

    Platform_Initialise();
    
#if defined ASURA_TOOLS
    Tools_MaterialResponse_System::Initialise();
#endif
}


void Asura_MaterialResponse_System::Reset(const bool bFullReset)
{
    Platform_Reset( bFullReset );

#if defined ASURA_TOOLS
    Tools_MaterialResponse_System::Reset();
#endif

    for( Asura_HashTable_It<Asura_MaterialResponse> xIt(&g_xMaterialHashTable); !xIt.Done(); )
    {
        Asura_MaterialResponse* pxMR = xIt.GetCurrent();
        if( bFullReset || pxMR->GetRemoveOnReset() )
        {
            delete pxMR;
            xIt.DeleteCurrent();
        }
        else
        {
            xIt.Next();
        }
    }

    if( bFullReset )
    {
        Asura_Assert( !g_xMaterialHashTable.GetNumEntries(), "" );
        s_xOverrideMapTree.Reset();
    }

    // (AAM - 19.10.2007) TODO: this assumes there are no slave materials in the global set, since
    // they won't have their own response lines. Don't currently have time to fix that case!
    s_uNumMaterialsTextureLines = GetNumGlobalMaterials();
}

void Asura_MaterialResponse_System::Shutdown()
{
    Reset( true );
    Platform_Shutdown();

#if defined ASURA_TOOLS
    Tools_MaterialResponse_System::Shutdown();
#endif
}


Asura_MaterialResponse* Asura_MaterialResponse_System::CreateNewMaterial( const Asura_Hash_ID uHashID )
{
    // Duplicate?
    if( g_xMaterialHashTable.FindObjectByHashID(uHashID) )
    {
        return NULL;
    }

    Asura_MaterialResponse* pxNewMR = new Asura_MaterialResponse();
    if( !pxNewMR )
    {
        return NULL;
    }

    pxNewMR->m_uHashID = uHashID;

    g_xMaterialHashTable.AddObject( pxNewMR );

    return pxNewMR;
}


const Asura_MaterialResponse* Asura_MaterialResponse_System::GetMaterialByHashID( const Asura_Hash_ID uNameHashID, const bool bCheckForOverrides )
{
    Asura_Hash_ID uSearchHash;
    if( bCheckForOverrides )
    {
        uSearchHash = GetOverriddenHash( uNameHashID );
    }
    else
    {
        uSearchHash = uNameHashID;
    }

    #ifdef ASURA_TOOLS
    if(s_bUseUVOrientMaterial && bCheckForOverrides)
    {
        //Force everything to use the UV Orientation material
        uSearchHash = 0x9CE61B5B;//Asura_GetHashID("Tools/UVOrientation")
    }
    #endif

    // Ensure that we always have a valid material-response...
    if(uSearchHash == ASURA_HASH_ID_UNSET)
    {
        uSearchHash = ASURA_DEFAULT_MATERIAL_RESPONSE;
    }

#if !defined ASURA_TOOLS

    Asura_MaterialResponse* pxMaterial = g_xMaterialHashTable.FindObjectByHashID( uSearchHash );

    if (!pxMaterial)
    {
        pxMaterial = g_xMaterialHashTable.FindObjectByHashID( ASURA_DEFAULT_MATERIAL_RESPONSE );
        Asura_Assert( pxMaterial, "Default material not available" );
    }

    // Validate the texture IDs... may be able to do this post load once we can guarantee all the
    // textures are loaded etc rather than each time material is referenced!
    if ( pxMaterial ) pxMaterial->ValidateTextureIDs();

    return pxMaterial;

#else

    int iMaterialIndex = 0;

    Tools_MaterialResponse* pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( uSearchHash, &iMaterialIndex );

    if( !pxMaterial )
    {
        // Look for the default material if we couldn't find the one requested
        pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( ASURA_DEFAULT_MATERIAL_RESPONSE, &iMaterialIndex );
    }

    if ( pxMaterial )
    {
        //Get the static material to fill in. Need to cycle through 2 slots to allow for polygons
        //that have two materials.
        Asura_MaterialResponse* pxReturnMaterial = &g_axToolsMaterial[g_iToolsMaterialIndex];
        g_iToolsMaterialIndex = (g_iToolsMaterialIndex+1)%2;
        
        // make a fake material and return that
        (*pxReturnMaterial) = *(pxMaterial);

        int iParentIndex;
        Tools_MaterialResponse* pxParent = Tools_MaterialResponse_System::GetMaterialByHashID( pxMaterial->GetParentHash(), &iParentIndex );    

        Asura_Assertf( pxMaterial->GetParentHash() == ASURA_HASH_ID_UNSET || pxParent, ("Parent material of '%s' doesn't exist.", pxMaterial->GetName().GetString() ) );

        if( pxParent )
        {
            // Do we need to allocate a row in the material response texture?
            if (pxParent->GetResponseRowIndex() >= ASURA_PC_MATERIALS_TEXTURE_HEIGHT)
            {
                // Allocate it
                // If we've successfully allocated a row, mark the texture as dirty
                if (pxParent->AllocateResponseRowIndex())
                {
                    Asura_Direct3D_MaterialResponse::SetMaterialsTextureValid(false);
                }
            }

            pxReturnMaterial->m_iMaterialIndex = pxParent->GetResponseRowIndex(); //iParentIndex;
            pxReturnMaterial->m_fAmbientContribution = pxParent->GetAmbientContribution() / 100.0f;
            pxReturnMaterial->m_fPrelightingColourise = pxParent->GetColouriseAmbient() / 100.0f;
            pxReturnMaterial->m_xAmbientColour = pxParent->GetAmbientColour();
            pxReturnMaterial->m_eBlendMode = pxParent->GetBlendMode();
            pxReturnMaterial->m_bZPrePass = pxParent->GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_OPAQUE && pxParent->GetIsZPrePass();
            pxReturnMaterial->m_eRenderType = pxParent->GetRenderType();

            pxReturnMaterial->m_fSelfIlluminationContribution = pxParent->m_fSelfIlluminationContribution / 100.0f;
            pxReturnMaterial->m_fSelfIlluminationColourise = pxParent->m_fColouriseSelfIllumination / 100.0f;
            pxReturnMaterial->m_xSelfIlluminationColour = pxParent->m_xSelfIlluminationColour;
            //pxReturnMaterial->m_fDiffuseGraphScale = pxParent->m_xDiffuseGraph.GetMaxDiscreteValue()/ 100.0f;
            //pxReturnMaterial->m_fSpecularGraphScale = pxParent->m_xSpecularGraph.GetMaxDiscreteValue() / 100.0f;
            pxReturnMaterial->m_fReflectivity = pxParent->m_fReflectivity / 100.0f;
            pxReturnMaterial->m_bDecal = pxParent->GetIsDecal();
        }
        else
        {
            // Do we need to allocate a row in the material response texture?
            if (pxMaterial->GetResponseRowIndex() >= ASURA_PC_MATERIALS_TEXTURE_HEIGHT)
            {
                // Allocate it
                // If we've successfully allocated a row, mark the texture as dirty
                if (pxMaterial->AllocateResponseRowIndex())
                {
                    Asura_Direct3D_MaterialResponse::SetMaterialsTextureValid(false);
                }
            }

            pxReturnMaterial->m_iMaterialIndex = pxMaterial->GetResponseRowIndex();
        }

        return pxReturnMaterial;    
    }
    else
    {
        return NULL;
    }

#endif
}


Asura_MaterialResponse_OverrideMap* Asura_MaterialResponse_System::CreateOverrideMap( const Asura_Hash_ID uHashID, u_int uExpectedNumOverrides )
{
    Asura_MaterialResponse_OverrideMap* pxNewOverrideMap = s_xOverrideMapTree.AddNode( uHashID );

    if( pxNewOverrideMap )
    {
        pxNewOverrideMap->AllocateHashTable( uExpectedNumOverrides );
    }
    else
    {
        Asura_Assertf( false, ("Failed to create material-response override-map for hash ID 0x%08X. Duplicate?", uHashID) );
    }

    return pxNewOverrideMap;
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::DestroyOverrideMap
// Author       : Andy Tate
// Date         : 26-Apr-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MaterialResponse_System::DestroyOverrideMap(const Asura_Hash_ID uHash)
{
    if(s_xOverrideMapTree.DeleteNodeByHash(uHash))
    {
        if( uHash == s_uCurrentOverrideMapHash )
        {
            ResetCurrentOverrideMap();
        }
        return(true);
    }
    else
{
        Asura_Assertf(false, ("Failed to delete material-response override-map for hash-ID %08X", uHash));
        return(false);
    }
}

/// End of function 'Asura_MaterialResponse_System::DestroyOverrideMap'. ///

/*
void Asura_MaterialResponse_System::SetNumGlobalMaterials(const int iNumGlobalMaterials)
{
    s_iNumGlobalMaterials = iNumGlobalMaterials;
}*/

int Asura_MaterialResponse_System::GetNumGlobalMaterials()
{
    return s_iNumGlobalMaterials;
}

bool Asura_MaterialResponse_System::IsGlobalResetPointHit(void)
{
    return s_bGlobalResetPointHit;
}

void Asura_MaterialResponse_System::SetGlobalResetPointHit(void)
{
    s_bGlobalResetPointHit = true;
    s_iNumGlobalMaterials = g_xMaterialHashTable.GetNumEntries();
}

Asura_Hash_ID Asura_MaterialResponse_System::CombineShaderHashes(
	const Asura_MaterialResponse* const pxMaterial1,
	const Asura_MaterialResponse* const pxMaterial2, const bool bSkin )
{
	const Asura_Hash_ID uPartialHash1 =
		(pxMaterial1 ? pxMaterial1->GetPartialShaderHash() : ASURA_HASH_ID_UNSET);

	const Asura_Hash_ID uPartialHash2 =
		(pxMaterial2 ? pxMaterial2->GetPartialShaderHash() : ASURA_HASH_ID_UNSET);

	// Extract the secondary texture slots used 
	const u_int uSecondaryTextures = uPartialHash2 & uASURA_MRPS_HASH_BITMASK_TEXTURES;

	// Push them up one into the secondary bits
	Asura_Hash_ID uCombinedHash = uPartialHash1 | (uSecondaryTextures << 1);

	if( bSkin )
	{
		// This can go if we get round to unifying skin & environment pixel shaders
		uCombinedHash |= uASURA_MRPS_HASH_BITMASK_SKIN;
	}

	return uCombinedHash;
}

#if defined(ASURA_TOOLS)

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::Traverse_SetOverrideMapHashes
// Author       : Andy Tate
// Date         : 3-May-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MaterialResponse_System::Traverse_SetOverrideMapHashes(Asura_Hash_ID uHash,
                                                                  Asura_MaterialResponse_OverrideMap* pxOverrideMap)
{
    s_puTraversalHashes[s_uTraversalIndex++]    = uHash;
}

/// End of function 'Asura_MaterialResponse_System::Traverse_SetOverrideMapHashes'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::CreateOverrideMapHashArray
// Author       : Andy Tate
// Date         : 3-May-2007
////////////////////////////////////////////////////////////////////////////////

Asura_Hash_ID* Asura_MaterialResponse_System::CreateOverrideMapHashArray(void)
{
    const u_int uNumOverrideMaps    = GetNumOverrideMaps();

    if(uNumOverrideMaps == 0)
    {
        return(NULL);
    }
    else
    {
        s_puTraversalHashes = new Asura_Hash_ID [uNumOverrideMaps];
        s_uTraversalIndex   = 0;

        s_xOverrideMapTree.Traverse(Traverse_SetOverrideMapHashes);
        return(s_puTraversalHashes);
    }
}

/// End of function 'Asura_MaterialResponse_System::CreateOverrideMapHashArray'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::DestroyOverrideMapHashes
// Author       : Andy Tate
// Date         : 3-May-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MaterialResponse_System::DestroyOverrideMapHashArray(Asura_Hash_ID* puOverrideMapHashes)
{
    delete[](puOverrideMapHashes);
}

/// End of function 'Asura_MaterialResponse_System::DestroyOverrideMapHashes'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::Traverse_SetOverrideMapNames
// Author       : Andy Tate
// Date         : 3-May-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MaterialResponse_System::Traverse_SetOverrideMapNames(Asura_Hash_ID uHash,
                                                                 Asura_MaterialResponse_OverrideMap* pxOverrideMap)
{
    s_pszTraversalNames[s_uTraversalIndex++]    = pxOverrideMap->GetName();
}

/// End of function 'Asura_MaterialResponse_System::Traverse_SetOverrideMapNames'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::CreateOverrideMapNameArray
// Author       : Andy Tate
// Date         : 3-May-2007
////////////////////////////////////////////////////////////////////////////////

const char** Asura_MaterialResponse_System::CreateOverrideMapNameArray(void)
{
    const u_int uNumOverrideMaps    = GetNumOverrideMaps();

    if(uNumOverrideMaps == 0)
    {
        return(NULL);
    }
    else
    {
        s_pszTraversalNames = new const char*[uNumOverrideMaps];
        s_uTraversalIndex   = 0;

        s_xOverrideMapTree.Traverse(Traverse_SetOverrideMapNames);
        return(s_pszTraversalNames);
    }
}

/// End of function 'Asura_MaterialResponse_System::CreateOverrideMapNameArray'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_System::DestroyOverrideMapNames
// Author       : Andy Tate
// Date         : 3-May-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MaterialResponse_System::DestroyOverrideMapNameArray(const char** pszOverrideMapNames)
{
    delete[](pszOverrideMapNames);
}

/// End of function 'Asura_MaterialResponse_System::DestroyOverrideMapNames'. ///

#endif  // defined(ASURA_TOOLS)


void Asura_MaterialResponse_System::ResetCurrentOverrideMap( void )
{
    s_uCurrentOverrideMapHash = ASURA_HASH_ID_UNSET;
    s_pxCurrentOverrideMap = NULL;
}


void Asura_MaterialResponse_System::SetCurrentOverrideMap( const Asura_Hash_ID uHash )
{
    s_uCurrentOverrideMapHash = uHash;
    s_pxCurrentOverrideMap = GetOverrideMap( uHash );
}


/******************************************************************************/
/* Asura_MaterialResponse Class Functions                                     */
/******************************************************************************/

Asura_MaterialResponse::Asura_MaterialResponse()
    : m_fAmbientContribution( 1.0f )
    , m_xAmbientColour( 1.0f, 1.0f, 1.0f, 0.0f )
    , m_fPrelightingColourise( 0.0f )
    , m_fReflectivity( 0.0f )
    , m_fSelfIlluminationContribution( 0.0f )
    , m_xSelfIlluminationColour( 0.0f, 0.0f, 0.0f, 0.0f )
    , m_fSelfIlluminationColourise( 0.0f )
    , m_fParallaxScale( 0.04f )
    , m_fDiffuseGraphScale( 1.0f )
    , m_fSpecularGraphScale( 1.0f )
    , m_eBlendMode( BLEND_MODE_OPAQUE )
    , m_eRenderType( RENDER_TYPE_DEFAULT )
    , m_uNonChunkedFlags(0)
    , m_uChunkedFlags(0)
    , m_uProjectContext(0)
    , m_eSpecialFXMode( SPECIALFX_MODE_NONE )
    , m_xWaterReflectionColour( 0.5f, 0.5f, 0.5f, 0.5f )
    , m_xWaterRefractionColour( 0.5f, 0.5f, 0.5f, 0.5f )
    , m_xWaterFogColour( 0.25f, 0.25f, 0.25f, 0.25f )
    , m_fWaterFogStrength( 1.0f )   
    , m_fWaterRippleScale( 1.0f )
    , m_fWaterDistortionScale( 1.0f )
    , m_fWaterEdgeSoftness( 1.0f )
#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    , m_uSurfaceTypeHash(Asura_MaterialResponse_System::GetDefaultSurfaceTypeHash())
#else
    , m_uSurfaceType(0)
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    , m_uPartialShaderHash( ASURA_HASH_ID_UNSET )
{
    for( u_int u = 0; u < NUMBER_OF_TEXTURE_SLOTS; ++u )
    {
        m_auTextureHashes[u] = ASURA_HASH_ID_UNSET;
        m_aiTextures[u] = ASURA_TEXTURE_NONE;
    }

    m_fAlbedoDetailMapUVScale = 4.0f;
    m_fAlbedoDetailMapValueScale = 0.25f;
    m_fNormalDetailMapUVScale = 4.0f;
    m_fNormalDetailMapValueScale = 0.25f;

    m_uTextureAnimHashID = ASURA_HASH_ID_UNSET;

    for ( u_int u = 0; u <2; ++u )
    {
        m_afWaterUVControl[ u*3 + 0 ]   = 0.0f;   // u scroll speed
        m_afWaterUVControl[ u*3 + 1 ]   = 0.0f;   // v scroll speed
        m_afWaterUVControl[ u*3 + 2 ]   = 1.0f;   // uv scale

        m_afWaterWaveControl[ u*3 + 0 ] = 0.0f;   // amplitude
        m_afWaterWaveControl[ u*3 + 1 ] = 1.0f;   // frequency
        m_afWaterWaveControl[ u*3 + 2 ] = 1.0f;   // speed
    }

    m_bRemoveOnReset = Asura_MaterialResponse_System::IsGlobalResetPointHit();
}

Asura_MaterialResponse::~Asura_MaterialResponse()
{
}

#if defined( ASURA_TOOLS )

const Asura_MaterialResponse& Asura_MaterialResponse::operator =( const Tools_MaterialResponse& xOther )
{
    m_fAmbientContribution          = xOther.m_fAmbientContribution / 100.0f;
    m_fPrelightingColourise         = xOther.m_fColouriseAmbient / 100.0f;
    m_xAmbientColour                = xOther.m_xAmbientColour;

    m_fReflectivity                 = xOther.m_fReflectivity / 100.0f;

    m_fSelfIlluminationContribution = xOther.m_fSelfIlluminationContribution / 100.0f;
    m_fSelfIlluminationColourise    = xOther.m_fColouriseSelfIllumination / 100.0f;
    m_xSelfIlluminationColour       = xOther.m_xSelfIlluminationColour;

    m_fParallaxScale                = xOther.m_fParallaxScale / 100.0f;
    //m_fDiffuseGraphScale          = xOther.m_xDiffuseGraph.GetMaxDiscreteValue() / 100.0f;
    //m_fSpecularGraphScale         = xOther.m_xSpecularGraph.GetMaxDiscreteValue() / 100.0f;

    for( u_int u = 0; u < NUMBER_OF_TEXTURE_SLOTS; ++u )
    {
        m_aiTextures[u]             = xOther.GetTexture(u);
    }

    m_uTextureAnimHashID            = xOther.GetTextureAnim();

    m_eBlendMode                    = xOther.GetBlendMode();
    m_eRenderType                   = xOther.GetRenderType();

    m_uChunkedFlags                 = xOther.m_uChunkedFlags;
    if (m_eBlendMode == BLEND_MODE_OPAQUE)
    {
        m_bZPrePass = false;
    }
    m_uProjectContext               = xOther.m_uProjectContext;

    m_eSpecialFXMode                = xOther.GetSpecialFXMode();

    // SpecialFX mode: Water
    m_xWaterReflectionColour        = xOther.m_xWaterReflectionColour;
    m_xWaterRefractionColour        = xOther.m_xWaterRefractionColour;
    m_xWaterFogColour               = xOther.m_xWaterFogColour;
    m_fWaterFogStrength             = xOther.m_fWaterFogStrength;
    m_fWaterRippleScale             = xOther.m_fWaterRippleScale;
    m_fWaterDistortionScale         = xOther.m_fWaterDistortionScale;
    m_fWaterEdgeSoftness            = xOther.m_fWaterEdgeSoftness;

    for ( u_int u = 0; u < 6; ++u )
    {
        m_afWaterUVControl[ u ]     = xOther.m_afWaterUVControl[ u ];
        m_afWaterWaveControl[ u ]   = xOther.m_afWaterWaveControl[ u ];
    }

    m_fAlbedoDetailMapUVScale = xOther.m_fAlbedoDetailMapUVScale;
    m_fAlbedoDetailMapValueScale = xOther.m_fAlbedoDetailMapValueScale;
    m_fNormalDetailMapUVScale = xOther.m_fNormalDetailMapUVScale;
    m_fNormalDetailMapValueScale = xOther.m_fNormalDetailMapValueScale;

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    m_uSurfaceTypeHash              = xOther.m_uSurfaceTypeHash;
#else
    m_uSurfaceType                  = xOther.m_uSurfaceType;
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

    m_uPartialShaderHash			= xOther.CalculatePartialShaderHash();

    return *this;
}

#endif // defined ASURA_TOOLS




void Asura_MaterialResponse::ValidateTextureIDs()
{
    if (m_bTextureIDsValid) return;

    m_bTextureIDsValid = true;

    for( u_int u = 0; u < NUMBER_OF_TEXTURE_SLOTS; ++u )
    {
        m_aiTextures[u] = Asura_Texture_Management::GetIndexFromHashID( m_auTextureHashes[u] );
    }
}

bool Asura_MaterialResponse::GetAnimatedFlagState( const u_int uMaterialFlags ) const
{

    const Asura_Hash_ID uTextureAnimHash = GetTextureAnim();
    if( uTextureAnimHash != ASURA_HASH_ID_UNSET )
    {
        const u_int uAnimatedFlags = Asura_TextureAnimation_System::GetMaterialFlags( uTextureAnimHash, NULL, 0 );
        return (uAnimatedFlags & uMaterialFlags) != 0;
    }

    return false;
}

bool Asura_MaterialResponse::GetDontRender() const
{
    return (m_eRenderType == RENDER_TYPE_DONT_RENDER_BUT_CAST_SHADOW) || GetAnimatedFlagState( AsuraMaterialFlag_DontRender );
}

bool Asura_MaterialResponse::GetDontCastShadow() const
{
    return (m_eRenderType == RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW) || GetAnimatedFlagState( AsuraMaterialFlag_DontRender );
}

bool Asura_MaterialResponse::GetIgnoreLOS() const
{
    return GetAnimatedFlagState( AsuraMaterialFlag_IgnoreLOS );
}

bool Asura_MaterialResponse::GetNoCollisions() const
{
    return GetAnimatedFlagState( AsuraMaterialFlag_NoCollisions );
}

// Get all of the self-illumination properties in one go.
void Asura_MaterialResponse::GetSelfIllumination(const Asura_TextureAnimationController* const pxTextureAnimationController,
                                                 float& fContribution, float& fColourise, Asura_RGBA& xColour) const
{
    // Check for a self-illumination override in the texture controller.
    const Asura_TextureAnimationController_SelfIllumination* pxSelf = pxTextureAnimationController ? pxTextureAnimationController->GetSelfIllumination() : 0;
    if (pxSelf && pxSelf->IsEnabled())
    {
        fContribution = pxSelf->GetContribution();
        fColourise    = pxSelf->GetColourise();
        xColour       = pxSelf->GetColour();
        return;
    }

    fContribution = m_fSelfIlluminationContribution;
    fColourise    = m_fSelfIlluminationColourise;
    xColour       = m_xSelfIlluminationColour;

    if (m_uTextureAnimHashID)
    {
        //see if the self illumination material flag has been added or removed by the anim
        u_int uAddedFlags,uRemovedFlags;
        if (Asura_TextureAnimation_System::GetChangedMaterialFlags(m_uTextureAnimHashID, pxTextureAnimationController, &uAddedFlags, &uRemovedFlags))
        {
            if (uRemovedFlags & AsuraMaterialFlag_SelfIlluminate) 
            {
                fContribution = 0.0f;
            }
        }
    }
}


//////////////////////////////////////////////////////////////////////////
/// Asura_MaterialResponse_OverrideMap
//////////////////////////////////////////////////////////////////////////


Asura_MaterialResponse_OverrideMap::~Asura_MaterialResponse_OverrideMap( void )
{
    if( m_pxOverrideHashTable )
    {
        for( Asura_HashTable_It<Asura_MaterialResponse_Override_HashTableWrapper> xIt(m_pxOverrideHashTable); !xIt.Done(); xIt.Next() )
        {
            delete xIt.GetCurrent();
        }
        delete m_pxOverrideHashTable;
    }

    delete m_pxDefaultOverride;
}


Asura_MaterialResponse_Override* Asura_MaterialResponse_OverrideMap::CreateDefaultOverride(void)
{
    if(m_pxDefaultOverride == NULL)
    {
        m_pxDefaultOverride = new Asura_MaterialResponse_Override;
        return(m_pxDefaultOverride);
    }
    else
    {
        Asura_Assert(false, "Already have a default material-response override for this override-map, cannot create a second one.");
        return(NULL);
    }
}

/// End of function 'Asura_MaterialResponse_OverrideMap::CreateDefaultOverride'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MaterialResponse_OverrideMap::DestroyDefaultOverride
// Author       : Andy Tate
// Date         : 26-Apr-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MaterialResponse_OverrideMap::DestroyDefaultOverride(void)
{
    if(m_pxDefaultOverride == NULL)
    {
        return(false);
    }
    else
    {
        delete(m_pxDefaultOverride);
        m_pxDefaultOverride = NULL;
        return(true);
    }
}


Asura_MaterialResponse_Override* Asura_MaterialResponse_OverrideMap::CreateOverride( const Asura_Hash_ID uMaterialHashIDToOverride )
{
    Asura_Assert( m_pxOverrideHashTable, "" );

    // Duplicate?
    if( m_pxOverrideHashTable->FindObjectByHashID(uMaterialHashIDToOverride) )
    {
		Asura_Assertf( false, ("Attemping to create duplicate material response override (hash ID 0x%08X)", uMaterialHashIDToOverride) );
        return NULL;
    }

    Asura_MaterialResponse_Override* pxNewOverride = new Asura_MaterialResponse_Override();
    if( !pxNewOverride )
    {
        return NULL;
    }

    pxNewOverride->SetMaterialHashIDToOverride( uMaterialHashIDToOverride );

    m_pxOverrideHashTable->AddObject( static_cast<Asura_MaterialResponse_Override_HashTableWrapper*>(pxNewOverride) );

    return pxNewOverride;
}


bool Asura_MaterialResponse_OverrideMap::DestroyOverride( const Asura_Hash_ID uMaterialHashIDToOverride )
{
    Asura_MaterialResponse_Override* pxOverride = m_pxOverrideHashTable->RemoveObjectByHashID( uMaterialHashIDToOverride );
    if( pxOverride )
    {
        delete pxOverride;
        return true;
    }
    else
    {
        Asura_Assertf( false, ("No material response override with hash ID 0x%08X", uMaterialHashIDToOverride) );
        return false;
    }
}


#if defined(ASURA_TOOLS)

Asura_Hash_ID* Asura_MaterialResponse_OverrideMap::CreateMaterialHashIDsToOverrideArray( void ) const
{
    const u_int uNumOverrides = GetNumOverrides();
    if( !uNumOverrides )
    {
        return NULL;
    }

    Asura_Hash_ID* puMaterialHashIDsToOverride = new Asura_Hash_ID[uNumOverrides];

    u_int uOut = 0;
    for( Asura_HashTable_ConstIt<Asura_MaterialResponse_Override_HashTableWrapper> xIt(m_pxOverrideHashTable); !xIt.Done(); xIt.Next() )
    {
        puMaterialHashIDsToOverride[uOut++] = xIt.GetCurrentHashID();
    }
    Asura_Assert( uOut == uNumOverrides, "" );

    return puMaterialHashIDsToOverride;
}


void Asura_MaterialResponse_OverrideMap::DestroyMaterialHashIDsToOverrideArray( Asura_Hash_ID* puMaterialHashIDsToOverride ) const
{
    delete[] puMaterialHashIDsToOverride;
}

#endif  // defined(ASURA_TOOLS)

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM


/// End of file. ///
