//////////////////////////////////////////////////////////////////////////////
//
// Asura_MaterialResponse_System.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#if defined ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#if !defined __ASURA_MATERIALRESPONSE_SYSTEM_H__
#define __ASURA_MATERIALRESPONSE_SYSTEM_H__

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Engine independent.
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_HashTable.h"
#include "Core/Asura_HashTree.h"
#include "Rendering/Asura_Colour.h"

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#if !defined ASURA_DEFAULT_MATERIAL_RESPONSE
    #define ASURA_DEFAULT_MATERIAL_RESPONSE 0x5C13D641 // Hash of "Default"
#endif

#if !defined ASURA_MATERIAL_RESPONSE_TEXTURE_WIDTH
    #define ASURA_MATERIAL_RESPONSE_TEXTURE_WIDTH (256)
#endif

#if !defined ASURA_MATERIAL_RESPONSE_HASH_TABLE_SIZE
    #define ASURA_MATERIAL_RESPONSE_HASH_TABLE_SIZE (1024)
#endif

// Currently have the last graph texel repeated and 3 data texels
#define ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH (ASURA_MATERIAL_RESPONSE_TEXTURE_WIDTH - 4)

/******************************************************************************/
/* Classes                                                                    */
/******************************************************************************/

class Asura_Animation;
class Asura_TextureAnimationController;
#if defined( ASURA_TOOLS )
    class Tools_MaterialResponse;
#endif

////////////////////////////////////////////////////////////////////////////////
//
// Asura_MaterialResponse
//
// Cross-platform material response definition. Tools_MaterialResponse contains
// more info for actual editing
//
////////////////////////////////////////////////////////////////////////////////

static const u_int ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID = 0xFFFFFFFF;

class Asura_MaterialResponse
{
public:

    // Constructors / destructor
    Asura_MaterialResponse();
    ~Asura_MaterialResponse();

    // enums

    // KJFL - 04 January 2007 - I've explicitly set the values for the members of this
    // enum to stop people adding new entries in the middle - add new additions to the 
    // end of the list.
    enum TEXTURE_SLOT_ID
    {
        // Ensure this matches order of Tools_MaterialResponse_System::TextureUsage
        // and the description array in Tools_MaterialResponse_System::s_aszTextureSlotDescription
        ALBEDO_TEXTURE = 0,
        NORMAL_MAP = 1,
        SPECULAR_MAP = 2,
        TRANSPARENCY_MAP = 3,
        SELFILLUMINATION_MAP = 4,
        ENVIRONMENT_MAP = 5,
        REFLECTIVITY_MASK = 6,
        PARALLAX_MAP = 7,
        DETAIL_MAP = 8,
        DETAIL_NORMAL_MAP = 9,

        NUMBER_OF_TEXTURE_SLOTS
    };

    // KJFL - 17 January 2007 - I've explicitly set the values for the members of this
    // enum to stop people adding new entries in the middle - add new additions to the 
    // end of the list.
    enum UV_SET_ID
    {
        UV_SET_1 = 0,
        UV_SET_2 = 1,
        NUMBER_OF_UV_SETS
    };

    // KJFL - 17 January 2007 - I've explicitly set the values for the members of this
    // enum to stop people adding new entries in the middle - add new additions to the 
    // end of the list. There's also a static string array for Tools you need to change
    // in Tools_MaterialResponse_System.cpp
    enum BLEND_MODE
    {
        BLEND_MODE_OPAQUE = 0,

        BLEND_MODE_ADDITIVE = 1,
        BLEND_MODE_BACKGROUNDBLEND = 2, // background blend is "normal" alpha transparency

        NUMBER_OF_BLEND_MODES
    };

    // ART - 07/04/2008 - Replaced the old set of flags which were mutually exclusive but
    // were able to be set in an invalid manner due to how they were accessed in Tools.
    enum RENDER_TYPE
    {
        RENDER_TYPE_DEFAULT = 0,
        RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW,
        RENDER_TYPE_DONT_RENDER_BUT_CAST_SHADOW,

        NUMBER_OF_RENDER_TYPES
    };


    // DMR - 22 May 2007 - Added special effects modes - add new additions to the 
    // end of the list. There's also a static string array for Tools you need to change
    // in Tools_MaterialResponse_System.cpp
    enum SPECIALFX_MODE
    {
        SPECIALFX_MODE_NONE  = 0,
        SPECIALFX_MODE_WATER = 1,
        NUMBER_OF_SPECIALFX_MODES
    };

    // Functions

    Asura_Hash_ID GetHashID( void ) const { return m_uHashID; }

    float GetAmbientContribution() const { return m_fAmbientContribution; }
    const Asura_RGBA& GetAmbientColour() const { return m_xAmbientColour; }
    float GetPrelightingColourise() const { return m_fPrelightingColourise; }

    float GetReflectivity() const { return m_fReflectivity; }

    void GetSelfIllumination(const Asura_TextureAnimationController* const pxTextureAnimationController,
                             float& fContribution, float& fColourise, Asura_RGBA& xColour) const;

    float GetParallaxScale() const { return m_fParallaxScale; }

    int GetTexture( const TEXTURE_SLOT_ID eSlot ) const { return m_aiTextures[ eSlot ]; }

    int GetMaterialIndex( void ) const { return m_iMaterialIndex; }

    Asura_Hash_ID GetTextureAnim() const {return m_uTextureAnimHashID;}

    BLEND_MODE     GetBlendMode( void ) const { return m_eBlendMode; }
    RENDER_TYPE    GetRenderType( void ) const { return m_eRenderType; }
    SPECIALFX_MODE GetSpecialFXMode( void ) const { return m_eSpecialFXMode; }

    Asura_RGBA   GetWaterReflectionColour()    const {return  m_xWaterReflectionColour;}
    Asura_RGBA   GetWaterRefractionColour()    const {return  m_xWaterRefractionColour;}
    Asura_RGBA   GetWaterFogColour()           const {return  m_xWaterFogColour;}
    float        GetWaterFogStrength()         const {return  m_fWaterFogStrength;}
    float        GetWaterDistortionScale()     const {return  m_fWaterDistortionScale;}
    float        GetWaterRippleScale()         const {return  m_fWaterRippleScale;}
    float        GetWaterEdgeSoftness()        const {return  m_fWaterEdgeSoftness;}
    float        GetWaterUVControl(int iIdx)   const {return  m_afWaterUVControl[iIdx];}
    float        GetWaterWaveControl(int iIdx) const {return  m_afWaterWaveControl[iIdx];}

    bool GetRemoveOnReset(void) const { return m_bRemoveOnReset; }

    // These flags are controlled via texture animation
    bool GetDontRender() const;
    bool GetDontCastShadow() const;
    bool GetIgnoreLOS() const;
    bool GetNoCollisions() const;
    bool GetIsDecal() const { return m_bDecal; }
    bool GetZPrePass() const { return m_bZPrePass; }
    bool GetIgnoreSpecularPrelighting() const { return m_bIgnoreSpecularPrelighting; }

    float GetAlbedoDetailMapUVScale() const { return m_fAlbedoDetailMapUVScale;}
    float GetAlbedoDetailMapValueScale() const { return m_fAlbedoDetailMapValueScale;}
    float GetNormalDetailMapUVScale() const { return m_fNormalDetailMapUVScale;}
    float GetNormalDetailMapValueScale() const { return m_fNormalDetailMapValueScale;}

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    inline Asura_Hash_ID GetSurfaceTypeHash(void) const;
#else
    u_int GetSurfaceType() const {return m_uSurfaceType;}
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

    // Blended materials (excluding decals) must be rendered after deferred lights
    bool IsTransparent() const { return GetBlendMode() != BLEND_MODE_OPAQUE && !GetIsDecal(); }

    float GetDiffuseGraphScale() const { return m_fDiffuseGraphScale; }
    float GetSpecularGraphScale() const { return m_fSpecularGraphScale; }

    u_int GetProjectContext() const {return m_uProjectContext;}

	Asura_Hash_ID GetPartialShaderHash() const { return m_uPartialShaderHash; }

protected:
    
    friend class Asura_MaterialResponse_System;
    friend class Asura_Chunk_MaterialResponse;

    // Operators
#if defined( ASURA_TOOLS )
    const Asura_MaterialResponse& operator =( const Tools_MaterialResponse& xOther );
#endif

    void ValidateTextureIDs();
    bool GetAnimatedFlagState( const u_int uMaterialFlags ) const;

    // Data

    Asura_Hash_ID   m_uHashID;
    int             m_iMaterialIndex;

    Asura_FixedArray< Asura_Hash_ID,    NUMBER_OF_TEXTURE_SLOTS > m_auTextureHashes;
    Asura_FixedArray< int,              NUMBER_OF_TEXTURE_SLOTS > m_aiTextures;

    float           m_fAmbientContribution;
    Asura_RGBA      m_xAmbientColour;
    float           m_fPrelightingColourise;

    float           m_fReflectivity;

    float           m_fSelfIlluminationContribution;
    Asura_RGBA      m_xSelfIlluminationColour;
    float           m_fSelfIlluminationColourise;

    float           m_fParallaxScale;
    float           m_fDiffuseGraphScale;
    float           m_fSpecularGraphScale;

    float           m_fAlbedoDetailMapUVScale;
    float           m_fAlbedoDetailMapValueScale;
    float           m_fNormalDetailMapUVScale;
    float           m_fNormalDetailMapValueScale;


    Asura_Hash_ID   m_uTextureAnimHashID;

    BLEND_MODE      m_eBlendMode;
    RENDER_TYPE     m_eRenderType;

    union
    {
        // These flags are used during runtime for the state of the material, and aren't chunked
        u_int       m_uNonChunkedFlags;

        struct
        {
            u_int   m_bRemoveOnReset        : 1;
            u_int   m_bTextureIDsValid      : 1;
        };
    };

    union
    {
        // These flags are material properties, serialised in Asura_Chunk_MaterialResponse. They
        // should match the flags in Tools_MaterialResponse
        u_int       m_uChunkedFlags;

        struct
        {
            u_int   m_bDecal                     : 1;
            u_int   m_bZPrePass                  : 1;
            u_int   m_bIgnoreSpecularPrelighting : 1;
        };
    };

    // Can be used for whatever the project wants, no assumption is made by asura over what this means
    // other than ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID will be passed for the second ID when
    // there is no second material
    u_int           m_uProjectContext;

    SPECIALFX_MODE  m_eSpecialFXMode;

    // SpecialFX mode: Water
    Asura_RGBA      m_xWaterReflectionColour;
    Asura_RGBA      m_xWaterRefractionColour;
    Asura_RGBA      m_xWaterFogColour;
    float           m_fWaterFogStrength;
    float           m_afWaterUVControl[6];          // 2 layers, U/V scroll speed and scale for each
    float           m_fWaterRippleScale;
    float           m_fWaterDistortionScale;
    float           m_fWaterEdgeSoftness;
    float           m_afWaterWaveControl[6];        // Amplitude, Frequency, Speed along the X and Z axes

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    Asura_Hash_ID   m_uSurfaceTypeHash;             // Allows surfaces to be specified by designer.
#else
    u_int           m_uSurfaceType;                 // Takes values from a project defined enumeration. Used for specifying wood/stone/metal etc for playing footstep sounds , or similar
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

	Asura_Hash_ID	m_uPartialShaderHash;
};


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_MaterialResponse_Override
// Author       : Andy Tate
// Date         : 26-Apr-2007
//
// Description  : Class used to hold the data needed when mapping material-response
//                data.
////////////////////////////////////////////////////////////////////////////////

class Asura_MaterialResponse_Override
{

public:

    // Public functions...
    Asura_MaterialResponse_Override  (void)                                             {}
    Asura_MaterialResponse_Override  (const Asura_MaterialResponse_Override& xOther)    { *this = xOther; }
//  ~Asura_MaterialResponse_Override ();

    Asura_MaterialResponse_Override& operator= (const Asura_MaterialResponse_Override& xOther)
    {
#if defined(ASURA_TOOLS)
        m_xOverrideName = xOther.m_xOverrideName;
#endif
        m_uMaterialHashIDToOverride = xOther.m_uMaterialHashIDToOverride;
        m_uOverrideHashID = xOther.m_uOverrideHashID;
        return(*this);
    }

    // Accessor functions...
#if defined(ASURA_TOOLS)
    void                SetOverrideName (const Asura_String& xName) { m_xOverrideName = xName; }
    const Asura_String& GetOverrideName (void) const                { return m_xOverrideName; }
#endif

    void SetMaterialHashIDToOverride( const Asura_Hash_ID uHashID ) { m_uMaterialHashIDToOverride = uHashID; }
    Asura_Hash_ID GetMaterialHashIDToOverride( void ) const { return m_uMaterialHashIDToOverride; }

    void            SetOverrideHashID( const Asura_Hash_ID uHashID ) { m_uOverrideHashID = uHashID; }
    Asura_Hash_ID   GetOverrideHashID( void ) const { return m_uOverrideHashID; }


protected:

#if defined(ASURA_TOOLS)
    Asura_String  m_xOverrideName;              // The name of the replacement material response.
#endif
    Asura_Hash_ID m_uMaterialHashIDToOverride;  // The hash ID of the material response that should be overriden.
    Asura_Hash_ID m_uOverrideHashID;            // The hash ID of the material response to replace it with.

};   // End of class 'Asura_MaterialResponse_Override'.


// Subclass dealing with some extra details for storage in hash tables.
class Asura_MaterialResponse_Override_HashTableWrapper : public Asura_MaterialResponse_Override
{
public:
    Asura_Hash_ID GetHashID( void ) const { return m_uMaterialHashIDToOverride; }
};


////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_MaterialResponse_OverrideMap
// Author       : Andy Tate
// Date         : 26-Apr-2007
//
// Description  : Class used to map material-response hash-IDs to other materials,
//                can be used for changing appearance in e.g. vision-modes.
////////////////////////////////////////////////////////////////////////////////
class Asura_MaterialResponse_OverrideMap
{

public:

    Asura_MaterialResponse_OverrideMap( void )
        : m_pxOverrideHashTable( NULL )
        , m_pxDefaultOverride( NULL )
    {
    }

    ~Asura_MaterialResponse_OverrideMap( void );

    void AllocateHashTable( u_int uExpectedNumOverrides )
    {
        Asura_Assert( !m_pxOverrideHashTable, "" );
        m_pxOverrideHashTable = new OverrideHashTable( 1 << Asura_Maths::GetFloorLog2(uExpectedNumOverrides) );
    }

#if defined(ASURA_TOOLS)
    // Methods for retrieving an array of all MR hashes that this map
    // overrides.
    Asura_Hash_ID* CreateMaterialHashIDsToOverrideArray( void ) const;
    void DestroyMaterialHashIDsToOverrideArray( Asura_Hash_ID* puOverrideHashes ) const;
#endif

    // Accessor function...
    Asura_MaterialResponse_Override*        CreateDefaultOverride   (void);
    bool                                    DestroyDefaultOverride  (void);
    Asura_MaterialResponse_Override*        GetDefaultOverride      (void)          { return(m_pxDefaultOverride); }
    const Asura_MaterialResponse_Override*  GetDefaultOverride      (void) const    { return(m_pxDefaultOverride); }

    Asura_MaterialResponse_Override*        CreateOverride  ( const Asura_Hash_ID uMaterialHashIDToOverride );
    bool                                    DestroyOverride ( const Asura_Hash_ID uMaterialHashIDToOverride );

    Asura_MaterialResponse_Override*        GetOverride     (const Asura_Hash_ID uHash)         { return m_pxOverrideHashTable->FindObjectByHashID( uHash ); }
    const Asura_MaterialResponse_Override*  GetOverride     (const Asura_Hash_ID uHash) const   { return m_pxOverrideHashTable->FindObjectByHashID( uHash ); }
    u_int                                   GetNumOverrides (void) const                        { return m_pxOverrideHashTable->GetNumEntries(); }

#if defined(ASURA_TOOLS)
    void        SetName (const char* szName)    { m_xName = szName; }
    const char* GetName (void) const            { return(m_xName.GetString()); }
#endif

private:

    typedef Asura_HashTable<Asura_MaterialResponse_Override_HashTableWrapper> OverrideHashTable;

    OverrideHashTable*               m_pxOverrideHashTable;  // Hash table of overrides for current materials.
    Asura_MaterialResponse_Override* m_pxDefaultOverride;    // A pointer to the override used for any hash IDs not in our tree. This may be NULL if no default is required.

#if defined(ASURA_TOOLS)
    Asura_String                        m_xName;                // The name of this override-map (used by Tools).
#endif

    // Prevent copying.
    Asura_MaterialResponse_OverrideMap  (const Asura_MaterialResponse_OverrideMap& xOther);
    Asura_MaterialResponse_OverrideMap& operator=   (const Asura_MaterialResponse_OverrideMap& xOther);

};   // End of class 'Asura_MaterialResponse_OverrideMap'.


////////////////////////////////////////////////////////////////////////////////
//
// Asura_MaterialResponse_System
//
// Manager class for the material response system
//
////////////////////////////////////////////////////////////////////////////////

class Asura_MaterialResponse_System
{
public:
    typedef void (*MaterialProjectContextCallback)(u_int uPrimaryContext, u_int uSecondaryContext);
    typedef void (*AlbedoModifierCallback)(const Asura_Animation* const xAnim, const u_int uContext, Asura_RGBA& xColour);

    static void Initialise();
    static void Reset( const bool bFullReset );
    static void Shutdown();

    static void SetMaterialProjectContextCallback(MaterialProjectContextCallback pfnMaterialProjectContextCallback);
    static MaterialProjectContextCallback GetMaterialProjectContextCallback();

    static void SetAlbedoModifierCallback(AlbedoModifierCallback pfnCallback);
    static AlbedoModifierCallback GetAlbedoModifierCallback();

    // Material-response accessor functions...
    static Asura_MaterialResponse* CreateNewMaterial( const Asura_Hash_ID uHashID );
    static const Asura_MaterialResponse* GetMaterialByHashID( const Asura_Hash_ID uNameHashID, const bool bCheckForOverrides = true );

    static void InitialiseMaterial( const Asura_Hash_ID uNameHash,
                                    const Asura_MaterialResponse& xMaterial )
    {
        Platform_InitialiseMaterial( uNameHash, xMaterial );
    }

    // Material-response override-map accessor functions...
    static Asura_MaterialResponse_OverrideMap*  CreateOverrideMap( const Asura_Hash_ID uHashID, u_int uExpectedNumOverrides );
    static bool                                 DestroyOverrideMap( const Asura_Hash_ID uHashID );

    static Asura_MaterialResponse_OverrideMap*  GetOverrideMap              (const Asura_Hash_ID uHash) { return(s_xOverrideMapTree.GetDataByHash(uHash)); }
    static u_int                                GetNumOverrideMaps          (void)  { return(s_xOverrideMapTree.GetNumberOfNodes()); }

    static void                                 ResetCurrentOverrideMap     (void);
    static void                                 SetCurrentOverrideMap       (const Asura_Hash_ID uHash);
    static Asura_MaterialResponse_OverrideMap*  GetCurrentOverrideMap       (void)  { return s_pxCurrentOverrideMap; }
    static Asura_Hash_ID                        GetCurrentOverrideMapHash   (void)  { return(s_uCurrentOverrideMapHash); }

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    static void             SetDefaultSurfaceTypeHash(const Asura_Hash_ID uHash)    { s_uDefaultSurfaceTypeHash = uHash; }
    static Asura_Hash_ID    GetDefaultSurfaceTypeHash(void)                         { return(s_uDefaultSurfaceTypeHash); }
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

#if defined(ASURA_TOOLS)
    // Functions used to access the names for the overrides...
    static const char** CreateOverrideMapNameArray(void);
    static void         DestroyOverrideMapNameArray(const char** pszOverrideMapNames);

    static Asura_Hash_ID*   CreateOverrideMapHashArray(void);
    static void             DestroyOverrideMapHashArray(Asura_Hash_ID* pszOverrideMapHashes);
#endif  // defined(ASURA_TOOLS)

    static int GetNumGlobalMaterials(void);

    static bool IsGlobalResetPointHit(void);
    static void SetGlobalResetPointHit();

    static u_int GetNumMaterialsTextureLines() { return( s_uNumMaterialsTextureLines);}
    static void SetNumMaterialsTextureLines( const u_int uNum ) { s_uNumMaterialsTextureLines = uNum; }

	static Asura_Hash_ID CombineShaderHashes( const Asura_MaterialResponse* const pxMaterial1,
		const Asura_MaterialResponse* const pxMaterial2, const bool bSkin );

#ifdef ASURA_TOOLS
    //This forces everything to be rendered using a special material for showing off the orientation of the texturing
    static void SetUseUVOrientMaterial(bool bSet){s_bUseUVOrientMaterial = bSet;}
    static bool GetUseUVOrientMaterial(){return s_bUseUVOrientMaterial;}
#endif

    static Asura_Hash_ID GetOverriddenHash( const Asura_Hash_ID uNameHashID );

    // JAT 16/05/07 added an enum for the exported texture data
    enum { EXPORTED_TEXTURE_VERSION = 2 };

protected:
    static bool s_bGlobalResetPointHit;
    static int s_iNumGlobalMaterials;

    static void Platform_Initialise();
    static void Platform_Reset( const bool bFullReset );
    static void Platform_Shutdown();

    static void Platform_InitialiseMaterial( const Asura_Hash_ID uNameHash,
                                             const Asura_MaterialResponse& xMaterial );

    // Override accessor functions...
    static Asura_MaterialResponse_Override* GetMaterialResponseOverride( const Asura_Hash_ID uNameHashID );

#if defined(ASURA_TOOLS)
    static void Traverse_SetOverrideMapHashes(Asura_Hash_ID uHash, Asura_MaterialResponse_OverrideMap* pxOverrideMap);
    static void Traverse_SetOverrideMapNames(Asura_Hash_ID uHash, Asura_MaterialResponse_OverrideMap* pxOverrideMap);
#endif  // defined(ASURA_TOOLS)

private:

    typedef Asura_TemplateHashTree<Asura_MaterialResponse_OverrideMap> OverrideMapTree;

    // Private variables...
    static OverrideMapTree                     s_xOverrideMapTree;          // Tree of material-response override-maps, accessed via hash-ID.
    static Asura_Hash_ID                       s_uCurrentOverrideMapHash;   // Hash-ID of the currently active override-map.
    static Asura_MaterialResponse_OverrideMap* s_pxCurrentOverrideMap;
    static u_int                               s_uNumMaterialsTextureLines;

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
    static Asura_Hash_ID    s_uDefaultSurfaceTypeHash;  // Hash-ID of the default surface-type to use when creating a meterial.
#endif  // defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

#if defined(ASURA_TOOLS)
    static Asura_Hash_ID*   s_puTraversalHashes;        // A pointer to the static array of hash-IDs (to be used during traversal).
    static const char**     s_pszTraversalNames;        // A pointer to the static array of names (to be used during traversal).
    static u_int            s_uTraversalIndex;          // Index used when traversing one of the trees.
#endif  // defined(ASURA_TOOLS)

    static MaterialProjectContextCallback s_pfnMaterialProjectContextCallback;
    static AlbedoModifierCallback s_pfnAlbedoModifierCallback;

#ifdef ASURA_TOOLS
    static bool             s_bUseUVOrientMaterial;
#endif
};


//////////////////////////////////////////////////////////////////////////
// Inlined functions...
//////////////////////////////////////////////////////////////////////////

inline void Asura_MaterialResponse_System::SetMaterialProjectContextCallback(
    MaterialProjectContextCallback pfnMaterialProjectContextCallback )
{
    s_pfnMaterialProjectContextCallback = pfnMaterialProjectContextCallback;
}

inline Asura_MaterialResponse_System::MaterialProjectContextCallback
    Asura_MaterialResponse_System::GetMaterialProjectContextCallback()
{
    return s_pfnMaterialProjectContextCallback;
}

inline void Asura_MaterialResponse_System::SetAlbedoModifierCallback(
    AlbedoModifierCallback pfnCallback )
{
    s_pfnAlbedoModifierCallback = pfnCallback;
}

inline Asura_MaterialResponse_System::AlbedoModifierCallback
    Asura_MaterialResponse_System::GetAlbedoModifierCallback()
{
    return s_pfnAlbedoModifierCallback;
}


inline Asura_MaterialResponse_Override* Asura_MaterialResponse_System::GetMaterialResponseOverride( const Asura_Hash_ID uNameHashID )
{
    if( s_pxCurrentOverrideMap )
    {
        Asura_MaterialResponse_Override* pxOverride = s_pxCurrentOverrideMap->GetOverride( uNameHashID );
        if( pxOverride )
        {
            return pxOverride;
        }
        else
        {
            return s_pxCurrentOverrideMap->GetDefaultOverride();
        }
    }
    else
    {
        return NULL;
    }
}


inline Asura_Hash_ID Asura_MaterialResponse_System::GetOverriddenHash( const Asura_Hash_ID uNameHashID )
{
    const Asura_MaterialResponse_Override* pxOverride = GetMaterialResponseOverride( uNameHashID );
    if( pxOverride )
    {
        return pxOverride->GetOverrideHashID();
    }
    else
    {
        return uNameHashID;
    }
}

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

inline Asura_Hash_ID Asura_MaterialResponse::GetSurfaceTypeHash(void) const
{
    return((m_uSurfaceTypeHash == ASURA_HASH_ID_UNSET) ? Asura_MaterialResponse_System::GetDefaultSurfaceTypeHash() : m_uSurfaceTypeHash);
}

#endif  //defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Bitmasks for the pixel shader hash IDs
// First come 20 bits for textures slots, then 6 bits for colourise options
const u_int uASURA_MRPS_HASH_BITMASK_BLEND_INTO_ALBEDO = 0x04000000;
const u_int uASURA_MRPS_HASH_BITMASK_IGNORE_SPECULAR   = 0x08000000;
const u_int uASURA_MRPS_HASH_BITMASK_SKIN              = 0x10000000;
const u_int uASURA_MRPS_HASH_BITMASK_TRANSPARENT       = 0x20000000;
// Only two bits left!

const u_int uASURA_MRPS_HASH_BITMASK_TEXTURES =
	(1 << (Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS * 2 )) - 1;

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#endif // !defined __ASURA_MATERIALRESPONSE_SYSTEM_H__


/// End of file. ///
