//
// Asura_Material_System.cpp
//
// Created by Richard Rice - 10/02/03
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Core/Asura_Material_System.h"
#include "Debugging/Asura_Assert.h"

#include "Animation/Asura_TextureAnimation.h"
#include "Core/Asura_Stream.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_TextureManagement.h"


Asura_Material* Asura_Material_System::s_pxMaterials = 0;
Asura_Handle_List* Asura_Material_System::s_pxHandleList = 0;
int* Asura_Material_System::s_piLoadGroups = 0;

int* Asura_Material_System::s_piConversionArray = 0;
int Asura_Material_System::s_iConversionArraySize = 0;

#if defined( ASURA_TOOLS)
	u_int Asura_Material_System::s_uAllowedRenderFlags = static_cast< u_int>( ~0);
#endif

void Asura_Material::Initialise()
{
	for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
	{
		m_iTextureID[i] = ASURA_TEXTURE_NONE;
		m_iNormalMapID[i] = ASURA_TEXTURE_NONE;
		m_iSpecularMapID[i] = ASURA_TEXTURE_NONE;	// VGP 9Feb2006
	}
	m_uMaterialFlags = 0;
	m_uProjectFlags = 0;
	m_uAnimHash = ASURA_HASH_ID_UNSET;
	m_uSpecularColour = 0xffffffff;

	#ifdef ASURA_TOOLS
	m_xMaterialName.MakeNull();
	#endif
}

#ifdef ASURA_TOOLS
bool Asura_Material::MaterialsAreEquivalent(const Asura_Material* pxMat1,const Asura_Material* pxMat2,bool bCompareNames)
{
	for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
	{
		if(pxMat1->m_iTextureID[i] != pxMat2->m_iTextureID[i]) return false;
		if(pxMat1->m_iNormalMapID[i] != pxMat2->m_iNormalMapID[i]) return false;
		if(pxMat1->m_iSpecularMapID[i] != pxMat2->m_iSpecularMapID[i]) return false;	// VGP 9Feb2006
	}
	if(pxMat1->m_uMaterialFlags != pxMat2->m_uMaterialFlags) return false;
	if(pxMat1->m_uProjectFlags != pxMat2->m_uProjectFlags) return false;
	if(pxMat1->m_uAnimHash != pxMat2->m_uAnimHash) return false;
	if(pxMat1->m_uSpecularColour != pxMat2->m_uSpecularColour) { return false; }
	
	if(bCompareNames)
	{
		if(pxMat1->m_xMaterialName.IsBlank() != pxMat2->m_xMaterialName.IsBlank()) return false;
		if(!pxMat1->m_xMaterialName.IsBlank())
		{
			if(pxMat1->m_xMaterialName.Compare(pxMat2->m_xMaterialName)) return false;
		}
	}
	return true;
}
#endif //ASURA_TOOLS

void Asura_Material_System::Initialise()
{
	Asura_Assert(!s_pxMaterials,"Already initialised");

	s_pxMaterials = new Asura_Material[ASURA_MAX_NO_OF_MATERIALS];
	Asura_Assert(s_pxMaterials,"Failed to allocated materials");

	s_pxHandleList = new Asura_Handle_List(ASURA_MAX_NO_OF_MATERIALS);
	Asura_Assert(s_pxHandleList,"");

	s_piLoadGroups = new int[ASURA_MAX_NO_OF_MATERIALS];
	Asura_Assert(s_piLoadGroups,"");

	//mark all the materials as invalid
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		s_piLoadGroups[i] = ASURA_TEXTURE_NOT_LOADED;
	}
}

void Asura_Material_System::Shutdown()
{
	delete [] s_pxMaterials;
	s_pxMaterials = 0;

	delete s_pxHandleList;
	s_pxHandleList = 0;

	delete [] s_piLoadGroups;
	s_piLoadGroups = 0;

	delete [] s_piConversionArray;
	s_piConversionArray = 0;
	s_iConversionArraySize = 0;
}

void Asura_Material_System::Reset()
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	Reset( Asura_Texture_Management::GetTextureResetGroup() );
#else
	Asura_Assert(s_pxHandleList,"");
	
	s_pxHandleList->Reset();

	//mark all the materials as invalid
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		s_piLoadGroups[i] = ASURA_TEXTURE_NOT_LOADED;

		#ifdef ASURA_TOOLS
		s_pxMaterials[i].m_xMaterialName.MakeNull();
		#endif //ASURA_TOOLS
	}

	delete [] s_piConversionArray;
	s_piConversionArray = 0;
	s_iConversionArraySize = 0;
#endif
}

void Asura_Material_System::Reset(int iGroup)
{
	if(iGroup == ASURA_TEXTURE_NOT_LOADED)
	{
		//not a group that can be reset
		return;
	}

	Asura_Assert(s_pxHandleList && s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] == iGroup)
		{
			DestroyMaterial(i+ASURA_MATERIAL_FIRSTID);
		}
	}
}

Asura_Material* Asura_Material_System::GetMaterial(int iMaterialID)
{
	if(iMaterialID == ASURA_MATERIAL_NONE) return 0;

	Asura_Assertf( ASURA_MATERIAL_FIRSTID <= iMaterialID && iMaterialID < (ASURA_MAX_NO_OF_MATERIALS + ASURA_MATERIAL_FIRSTID),("Invalid material index: 0x%08X", iMaterialID));
	int iIndex = iMaterialID - ASURA_MATERIAL_FIRSTID;
	Asura_Assertf(iIndex>=0 && iIndex<ASURA_MAX_NO_OF_MATERIALS,("Invalid material index: %i", iIndex));
	if(iIndex>=0 && iIndex<ASURA_MAX_NO_OF_MATERIALS)
	{
		if(s_piLoadGroups[iIndex] != ASURA_TEXTURE_NOT_LOADED)
		{
			return &s_pxMaterials[iIndex];
		}
	}

	return 0;
}

void Asura_Material_System::SetMaterial(int iMaterialID, const Asura_Material& xNewMat)
{
	Asura_Material* pxDestMat = GetMaterial(iMaterialID);
	if(!pxDestMat) return;

	#ifdef ASURA_TOOLS
	//take a backup up the material to determine changes
	Asura_Material xOriginalMat = *pxDestMat;
	#endif

	*pxDestMat = xNewMat;

	#ifdef ASURA_TOOLS
	//Check for any textures that have been added or removed , and update the usage flags for them. RAR 08/05/2006

	for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
	{
		if(xNewMat.m_iTextureID[i] != xOriginalMat.m_iTextureID[i])
		{
			UpdateTextureUsageFlags(xNewMat.m_iTextureID[i]);
			UpdateTextureUsageFlags(xOriginalMat.m_iTextureID[i]);
		}

		if(xNewMat.m_iNormalMapID[i] != xOriginalMat.m_iNormalMapID[i])
		{
			UpdateTextureUsageFlags(xNewMat.m_iNormalMapID[i]);
			UpdateTextureUsageFlags(xOriginalMat.m_iNormalMapID[i]);
		}

		if(xNewMat.m_iSpecularMapID[i] != xOriginalMat.m_iSpecularMapID[i])
		{
			UpdateTextureUsageFlags(xNewMat.m_iSpecularMapID[i]);
			UpdateTextureUsageFlags(xOriginalMat.m_iSpecularMapID[i]);
		}
	}
	#endif //ASURA_TOOLS
}

int Asura_Material_System::GetTextureID(int iMaterialID,int iTextureIndex /* = 0 */)
{
	Asura_Assert(iTextureIndex>=0 && iTextureIndex<ASURA_MATERIAL_MAXTEXTURES,"");	


	if(iMaterialID == ASURA_MATERIAL_NONE) return ASURA_TEXTURE_NONE;
	Asura_Material* pxMaterial = GetMaterial(iMaterialID);
	return pxMaterial ? pxMaterial->m_iTextureID[iTextureIndex] : ASURA_TEXTURE_NONE;
}

int Asura_Material_System::GetNormalMapID(int iMaterialID, int iNormalMapIndex /* = 0 */)
{
	Asura_Assert(iNormalMapIndex>=0 && iNormalMapIndex<ASURA_MATERIAL_MAXTEXTURES,"");	
	if(iMaterialID == ASURA_MATERIAL_NONE) return ASURA_TEXTURE_NONE;
	Asura_Material* pxMaterial = GetMaterial(iMaterialID);
	return pxMaterial ? pxMaterial->m_iNormalMapID[iNormalMapIndex] : ASURA_TEXTURE_NONE;
}

// VGP 9Feb2006
int Asura_Material_System::GetSpecularMapID(int iMaterialID, int iSpecularMapIndex /* = 0 */)
{
	Asura_Assert(iSpecularMapIndex>=0 && iSpecularMapIndex<ASURA_MATERIAL_MAXTEXTURES, "Invalid specular map index");

	if (iMaterialID == ASURA_MATERIAL_NONE) 
		return ASURA_TEXTURE_NONE;

	Asura_Material* pxMaterial = GetMaterial(iMaterialID);

	return pxMaterial ? pxMaterial->m_iSpecularMapID[iSpecularMapIndex] : ASURA_TEXTURE_NONE;
}

int Asura_Material_System::CreateMaterial()
{
	Asura_Assert(s_pxHandleList && s_pxMaterials,"");

	Asura_Handle iIndex;
	if(!s_pxHandleList->GetNewHandle(iIndex))
	{
		Asura_Assertf(0,("Unable to create another material. %d materials used already", s_pxHandleList->GetNumberOfActiveHandles()));
		return ASURA_MATERIAL_NONE;
	}

	//initialise the new material
	Asura_Material* pxMaterial = &s_pxMaterials[iIndex];
	Asura_Assert(s_piLoadGroups[iIndex] == ASURA_TEXTURE_NOT_LOADED,"");
	s_piLoadGroups[iIndex] = Asura_Texture_Management::GetTextureResetGroup();

	pxMaterial->Initialise();

	//and return the index
	return iIndex+ASURA_MATERIAL_FIRSTID;
}

int Asura_Material_System::CreateMaterial(int iOriginalIndex)
{
	Asura_Assert(s_piConversionArray,"");
	Asura_Assert(iOriginalIndex>=0 && iOriginalIndex<s_iConversionArraySize,"");

	int iMaterialID = CreateMaterial();
	s_piConversionArray[iOriginalIndex] = iMaterialID;

	return iMaterialID;
}

void Asura_Material_System::DestroyMaterial(int iMaterialID)
{
	Asura_Assert(s_pxHandleList && s_pxMaterials,"");
	int iIndex = iMaterialID - ASURA_MATERIAL_FIRSTID;
	Asura_Assert(iIndex>=0 && iIndex<ASURA_MAX_NO_OF_MATERIALS,"Invalid material index");
	if(iIndex>=0 && iIndex<ASURA_MAX_NO_OF_MATERIALS)
	{
		Asura_Assert(s_piLoadGroups[iIndex] != ASURA_TEXTURE_NOT_LOADED,"Material destroyed twice?");

		//release the handle , and mark the material to show it isn't in use
		s_pxHandleList->ReleaseHandle(iIndex);

		s_piLoadGroups[iIndex] = ASURA_TEXTURE_NOT_LOADED;

		#ifdef ASURA_TOOLS
		s_pxMaterials[iIndex].m_xMaterialName.MakeNull();
		#endif //ASURA_TOOLS

	}
}

void Asura_Material_System::NewConversionList(int iLength)
{
	delete [] s_piConversionArray;

	s_iConversionArraySize = iLength;
	s_piConversionArray = new int [s_iConversionArraySize];
}

int Asura_Material_System::ConvertIndex(int iOriginalIndex)
{
	//-1 is saved out for any untextured polygons , in most cases this gets checked before we get here
	//but may as well deal with the case anyway. RAR 11/06/03
	if(iOriginalIndex == -1) return ASURA_MATERIAL_NONE;

	Asura_Assert(s_piConversionArray,"");
	if(iOriginalIndex>=0 && iOriginalIndex<s_iConversionArraySize)
	{
		return s_piConversionArray[iOriginalIndex];
	}
	Asura_Assert(iOriginalIndex>=0 && iOriginalIndex<s_iConversionArraySize,"Invalid original index");
	return ASURA_MATERIAL_NONE;
}

#ifdef ASURA_TOOLS
int Asura_Material_System::ConvertIDToIndex(int iMaterialID,const int* piMaterialIDToIndexConversion)
{
	//ASURA_MATERIAL_NONE doesn't need to be converted
	if(iMaterialID == ASURA_MATERIAL_NONE) return ASURA_MATERIAL_NONE;

	if(!piMaterialIDToIndexConversion)
	{
		//Allowing for the possiblity of no conversion array to make the saving code simpler
		Asura_Assert(iMaterialID < ASURA_MAX_NO_OF_MATERIALS,"Material doesn't have a valid index , and no conversion array has been provided");
		return iMaterialID;
	}
	else if(iMaterialID >= ASURA_MATERIAL_FIRSTID)
	{
		int iIndex = iMaterialID - ASURA_MATERIAL_FIRSTID;
		if(iIndex < ASURA_MAX_NO_OF_MATERIALS)
		{
			//material id is in a suitable range ,so return the converted value
			return piMaterialIDToIndexConversion[iIndex];
		}
	}
	//material ID is out of range , so best we can do is return no material
	Asura_Assertf(0,("Can't convert invalid material ID %d\nIt should be between %d and %d",iMaterialID,ASURA_MATERIAL_FIRSTID,ASURA_MATERIAL_FIRSTID+ASURA_MAX_NO_OF_MATERIALS));
	return ASURA_MATERIAL_NONE;
}
#endif //ASURA_TOOLS

int Asura_Material_System::GetMaterialIDForTexture(int iTextureID,bool bCreate)
{
	if(iTextureID == ASURA_TEXTURE_NONE) return ASURA_MATERIAL_NONE;
	
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			if(s_pxMaterials[i].m_iTextureID[0] == iTextureID) return i+ASURA_MATERIAL_FIRSTID;
		}
	}	

	//failed to find material for texture. create one if needed
	if(!bCreate) return ASURA_MATERIAL_NONE;

	int iMaterialID = CreateMaterial();
	if(iMaterialID != ASURA_MATERIAL_NONE)
	{
		Asura_Material* pxMaterial = GetMaterial(iMaterialID);
		pxMaterial->m_iTextureID[0] = iTextureID;

		#ifdef ASURA_TOOLS
		UpdateTextureUsageFlags(iTextureID);
		#endif
	}

	return iMaterialID;
}

#ifdef ASURA_TOOLS
unsigned int Asura_Material_System::GetMaterialFlagsUsedWithTexture(int iTextureID)
{
	if(iTextureID == ASURA_TEXTURE_NONE) return 0;

	unsigned int uMaterialFlags = 0;

	//scan through all the materials for those using the texture
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			for(int j=0;j<ASURA_MATERIAL_MAXTEXTURES;j++)
			{
				if(s_pxMaterials[i].m_iTextureID[j] == iTextureID)
				{
					uMaterialFlags |= s_pxMaterials[i].m_uMaterialFlags;
					if(s_pxMaterials[i].m_uAnimHash)
					{
						uMaterialFlags |= Asura_TextureAnimation_System::GetMaterialFlagsAddedByAnimation(s_pxMaterials[i].m_uAnimHash);
					}
				}
			}
		}
	}	
	return uMaterialFlags;
}

// (SB - 22/05/07) Equivalent to the above for project flags
unsigned int Asura_Material_System::GetProjectFlagsUsedWithTexture(int iTextureID)
{
	if(iTextureID == ASURA_TEXTURE_NONE) return 0;

	unsigned int uMaterialFlags = 0;

	//scan through all the materials for those using the texture
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			for(int j=0;j<ASURA_MATERIAL_MAXTEXTURES;j++)
			{
				if(s_pxMaterials[i].m_iTextureID[j] == iTextureID)
				{
					uMaterialFlags |= s_pxMaterials[i].m_uProjectFlags;
				}
			}
		}
	}	
	return uMaterialFlags;
}

unsigned int Asura_Material_System::GetAndedMaterialFlagsUsedWithTexture(int iTextureID)
{
	if(iTextureID == ASURA_TEXTURE_NONE) return 0;

	unsigned int uMaterialFlags = 0;
	bool bFirstFound = true;

	//scan through all the materials for those using the texture
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			for(int j=0;j<ASURA_MATERIAL_MAXTEXTURES;j++)
			{
				if(s_pxMaterials[i].m_iTextureID[j] == iTextureID)
				{
					if(bFirstFound)
					{
						uMaterialFlags = s_pxMaterials[i].m_uMaterialFlags;
						bFirstFound = false;
					}
					else
					{
						uMaterialFlags &= s_pxMaterials[i].m_uMaterialFlags;
					}
					if(s_pxMaterials[i].m_uAnimHash)
					{
						uMaterialFlags &=~ Asura_TextureAnimation_System::GetMaterialFlagsRemovedByAnimation(s_pxMaterials[i].m_uAnimHash);
					}

				}
			}
		}
	}	
	return uMaterialFlags;
}

// (SB - 22/05/07) Equivalent to the above for project flags
unsigned int Asura_Material_System::GetAndedProjectFlagsUsedWithTexture(int iTextureID)
{
	if(iTextureID == ASURA_TEXTURE_NONE) return 0;

	unsigned int uMaterialFlags = 0;
	bool bFirstFound = true;

	//scan through all the materials for those using the texture
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			for(int j=0;j<ASURA_MATERIAL_MAXTEXTURES;j++)
			{
				if(s_pxMaterials[i].m_iTextureID[j] == iTextureID)
				{
					if(bFirstFound)
					{
						uMaterialFlags = s_pxMaterials[i].m_uProjectFlags;
						bFirstFound = false;
					}
					else
					{
						uMaterialFlags &= s_pxMaterials[i].m_uProjectFlags;
					}
				}
			}
		}
	}	
	return uMaterialFlags;
}

void Asura_Material_System::UpdateTextureUsageFlags(int iTextureID)
{
	if(iTextureID == ASURA_TEXTURE_NONE) return;

	//Get the current texture flags ,and clear the usage flags
	int iTextureFlags = Asura_Texture_Management::GetTextureFlags(iTextureID);
	iTextureFlags &=~ ASURA_TEXTURE_USAGE_FLAGS;

	//scan through all the materials for those using the texture
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			for(int j=0;j<ASURA_MATERIAL_MAXTEXTURES;j++)
			{
				if(s_pxMaterials[i].m_iTextureID[j] == iTextureID)
				{
					iTextureFlags |= AsuraTextureFlag_DiffuseMap;
				}
				if(s_pxMaterials[i].m_iNormalMapID[j] == iTextureID)
				{
					iTextureFlags |= AsuraTextureFlag_NormalMap;
				}
				if(s_pxMaterials[i].m_iSpecularMapID[j] == iTextureID)
				{
					iTextureFlags |= AsuraTextureFlag_SpecularMap;
				}
			}
		}
	}

	Asura_Texture_Management::SetTextureFlags(iTextureID,iTextureFlags);
}


void Asura_Material_System::ReplaceTextureAnimation(Asura_Hash_ID uOldHash,Asura_Hash_ID uNewHash)
{
	Asura_Assert(s_pxMaterials,"");
	for(int i=0;i<ASURA_MAX_NO_OF_MATERIALS;i++)
	{
		if(s_piLoadGroups[i] != ASURA_TEXTURE_NOT_LOADED)
		{
			if(s_pxMaterials[i].m_uAnimHash == uOldHash)
			{
				s_pxMaterials[i].SetTextureAnim(uNewHash);
			}
		}
	}	
}

#endif // ASURA_TOOLS

bool Asura_Material_System::IsMaterialNormalMapped( int iMaterialID )
{
    return( ASURA_TEXTURE_NONE != GetNormalMapID(iMaterialID, 0) );
}

bool Asura_Material_System::IsMaterialDualNormalMap( int iMaterialID )
{
    return( ASURA_TEXTURE_NONE != GetNormalMapID(iMaterialID, 1) );
}

bool Asura_Material_System::IsMaterialDualTexture( int iMaterialID )
{
    return( ASURA_TEXTURE_NONE != GetTextureID(iMaterialID, 1) );
}

// VGP 9Feb2006
bool Asura_Material_System::IsMaterialSpecularMapped( int iMaterialID )
{
	return	(IsMaterialSpecularSoft(iMaterialID) || IsMaterialSpecularSharp(iMaterialID)) &&
			(ASURA_TEXTURE_NONE != GetSpecularMapID(iMaterialID, 0));
}

// VGP 9Feb2006
bool Asura_Material_System::IsMaterialDualSpecularMap( int iMaterialID )
{
	return	(IsMaterialSpecularMapped(iMaterialID)) &&
			(ASURA_TEXTURE_NONE != GetSpecularMapID(iMaterialID, 1));
}

bool Asura_Material_System::GetTextureAnimationState(int iMaterialID,const class Asura_TextureAnimationController* pxController,struct Asura_TextureAnimationOutput* pxOutput)
{

	Asura_Material* pxMaterial = GetMaterial(iMaterialID);
	if(!pxMaterial) return false;
	if(!pxMaterial->m_uAnimHash) return false;

	return Asura_TextureAnimation_System::GetAnimationState(pxMaterial->m_uAnimHash,pxController,pxOutput);

}

u_int Asura_Material_System::GetAnimatedMaterialFlags(const Asura_Material* pxMaterial)
{
	return Asura_TextureAnimation_System::GetMaterialFlags(pxMaterial->m_uAnimHash,0,pxMaterial->m_uMaterialFlags);
}

//
//	int Asura_Material_System::CompareMaterials( const int iMat1, const int iMat2 )
//
// Helper function to compare two materials' properties; designed to be used with qsort.
//
// Returns	-1 if iMat1 should go before iMat2
//
//			 0 if materials are identical
//
//			+1 if iMat1 should go after iMat2
//
// Order of importance is determined by:
//
// 1. Required render order e.g. transparent polys need to go last
// 2. Vertex size. Some materials (e.g. normal mapped ones) have more data
// 3. Texture (to group materials of the same texture together)
// 4. Material ID
// 
int Asura_Material_System::CompareMaterials( const int iMat1, const int iMat2 )
{
	// early out - are these the same material?
	if ( iMat1 == iMat2 ) return 0;


	// 1. Render order

	// A. Alpha decal vs non-Alpha decal (the former go last)

	if ( Asura_Material_System::IsMaterialAlphaDecal(iMat1) && !Asura_Material_System::IsMaterialAlphaDecal(iMat2) )
	{
		return 1;
	}
	if ( !Asura_Material_System::IsMaterialAlphaDecal(iMat1) && Asura_Material_System::IsMaterialAlphaDecal(iMat2) )
	{
		return -1;
	}

	// B. Additive-alpha vs not (the former go last)

	if ( Asura_Material_System::IsMaterialPS2AdditiveAlpha(iMat1) && !Asura_Material_System::IsMaterialPS2AdditiveAlpha(iMat2) )
	{
		return 1;
	}
	if ( !Asura_Material_System::IsMaterialPS2AdditiveAlpha(iMat1) && Asura_Material_System::IsMaterialPS2AdditiveAlpha(iMat2) )
	{
		return -1;
	}

    // C. Transparent vs non-transparent (the former go last)

	if ( Asura_Material_System::IsMaterialTransparent(iMat1) && !Asura_Material_System::IsMaterialTransparent(iMat2) )
	{
		return 1;
	}
	if ( !Asura_Material_System::IsMaterialTransparent(iMat1) && Asura_Material_System::IsMaterialTransparent(iMat2) )
	{
		return -1;
	}


	// Couldn't make a decision by render order, let's try 
	// 2. Vertex Size

	// TODO! Need some sort of get vertex size / vertex component flags thingy perhaps.


	// Couldn't make a decision by vertex size; let's try 
	// 3. Texture
	
	// A. Primary texture first
	{
		const int iTexture1 = Asura_Material_System::GetTextureID( iMat1 );
		const int iTexture2 = Asura_Material_System::GetTextureID( iMat2 );

		if (iTexture1 < iTexture2) return 1;
		if (iTexture1 > iTexture2) return -1;
	}
	// B. Then secondary texture
	{
		const int iTexture1 = Asura_Material_System::GetTextureID( iMat1, 1 );
		const int iTexture2 = Asura_Material_System::GetTextureID( iMat2, 1 );

		if (iTexture1 < iTexture2) return 1;
		if (iTexture1 > iTexture2) return -1;
	}

	// TODO - normal maps? reflectance maps? etc etc


	// Couldn't make a decision so far; let's just separate them based on material ID
	// 4. Material IDs

	if (iMat1 < iMat2) return 1;
	if (iMat1 > iMat2) return -1;
	
    return 0;
}

void Asura_Material_System::WriteResetGroupReport(u_int uGroup, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 96;
	char szLogBuffer[uLOGBUFFERSIZE];

	u_int uCount = 0;
	for (u_int u = 0; u < ASURA_MAX_NO_OF_MATERIALS; u++)
	{
		if (s_piLoadGroups[u] == static_cast<int>(uGroup))
		{
			uCount++;
		}
	}

	if (uCount == 0)
	{
		xLog << "Material:\t\t\tNone\n";
	}
	else
	{
		Asura_PrintfToBfr(szLogBuffer,"Material:\t\t\t%d Materials\n",uLOGBUFFERSIZE,true,uCount);
		xLog << szLogBuffer;
	}
}
