//
// Asura_Material_System.h
//
// Created by Richard Rice - 10/02/03
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_MATERIAL_SYSTEM_H__
#define __ASURA_MATERIAL_SYSTEM_H__

#include "Core/Asura_Handle.h"
#include "Core/Asura_MaterialFlags.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Hash.h"

#ifndef ASURA_MAX_NO_OF_MATERIALS
#ifdef ASURA_TOOLS
#define ASURA_MAX_NO_OF_MATERIALS 2000
#else
#define ASURA_MAX_NO_OF_MATERIALS 500
#endif
#endif

//Offset to apply to material indeces , so as to make it obvious if a material is being used instead
//of a texture id or vice versa
//This may or may not be a permanent feature , but at least for now it should help in tracking down
//parts of the code that haven't been updated
#define ASURA_MATERIAL_FIRSTID 1000

//Number of Texture IDs per material.
//This will most likely be assumed to be 2 in many places , but I'll to get as much as possible
//to use this in case we ever want to increase it again...
#define ASURA_MATERIAL_MAXTEXTURES 2

class Asura_Stream;

class Asura_Material
{
public :
	int m_iTextureID[ASURA_MATERIAL_MAXTEXTURES];
	int m_iNormalMapID[ASURA_MATERIAL_MAXTEXTURES];
	int m_iSpecularMapID[ASURA_MATERIAL_MAXTEXTURES];	// VGP 9Feb2006
	unsigned int m_uMaterialFlags;
	unsigned int m_uProjectFlags;
	Asura_Hash_ID m_uAnimHash;//ID of texture anim
	u_int m_uSpecularColour;

	void Initialise();


	#ifdef ASURA_TOOLS
	Asura_String m_xMaterialName;
	#endif

	#ifdef ASURA_TOOLS
	//checks to see if two materials are the same.
	//bCompareNames determines whether it will care if the material names differ
	static bool MaterialsAreEquivalent(const Asura_Material* pxMat1,const Asura_Material* pxMat2,bool bCompareNames);
	#endif

	void SetTextureAnim(Asura_Hash_ID uAnimHash)
	{
		m_uAnimHash = uAnimHash;
		if(uAnimHash)
			m_uMaterialFlags |= AsuraMaterialFlag_TextureAnim;
		else
			m_uMaterialFlags &=~ AsuraMaterialFlag_TextureAnim;
	}


};



class Asura_Material_System
{
public :
#if defined( ASURA_TOOLS)
	enum AllowedRenderFlags
	{
		ALLOWED_RENDER_FLAG_DIFFUSE1	= 0x0001,
		ALLOWED_RENDER_FLAG_DIFFUSE2	= 0x0002,
		ALLOWED_RENDER_FLAG_SPECULAR1	= 0x0004,
		ALLOWED_RENDER_FLAG_SPECULAR2	= 0x0008,
		ALLOWED_RENDER_FLAG_NORMAL1		= 0x0010,
		ALLOWED_RENDER_FLAG_NORMAL2		= 0x0020,
	};
	
	static void SetAllowedRenderFlag( const u_int uAllowedRenderFlag, const u_int uAllowedRenderMask) { s_uAllowedRenderFlags = (s_uAllowedRenderFlags & ~uAllowedRenderMask) | uAllowedRenderFlag; }
	static bool GetAllowedRenderFlag( const u_int uAllowedRenderFlag) { return( uAllowedRenderFlag == (s_uAllowedRenderFlags & uAllowedRenderFlag)); }
#endif

	static void Initialise();
	static void Shutdown();
	static void Reset();
	static void Reset(int iGroup);

	static Asura_Material* GetMaterial(int iMaterialID);
	static void SetMaterial(int iMaterialID,const Asura_Material& xMat);

	static int GetTextureID(int iMaterialID,int iTextureIndex = 0);
	static int GetNormalMapID(int iMaterialID, int iNormalMapIndex = 0);
	static int GetSpecularMapID(int iMaterialID, int iNormalMapIndex = 0);	// VGP 9Feb2006


	static inline unsigned GetMaterialFlags(int iMaterialID)
	{
		Asura_Material* pxMaterial = GetMaterial(iMaterialID);
		if(!pxMaterial) return 0;
		if(!pxMaterial->m_uAnimHash) return pxMaterial->m_uMaterialFlags;

		//has an animation , so need to check if it is altering the flags
		return GetAnimatedMaterialFlags(pxMaterial);
	}

	//Gets the flags without the overhead of checking for changes caused by animations
	static inline unsigned GetMaterialFlagsIgnoringAnim(int iMaterialID)
	{
		Asura_Material* pxMaterial = GetMaterial(iMaterialID);
		return pxMaterial ? pxMaterial->m_uMaterialFlags : 0;
	}

	static inline unsigned GetProjectFlags(int iMaterialID)
	{
		Asura_Material* pxMaterial = GetMaterial(iMaterialID);
		return pxMaterial ? pxMaterial->m_uProjectFlags : 0;
	}

	//allocates a new material , and returns the index. ASURA_MATERIAL_NONE if it fails
	static int CreateMaterial();
	static int CreateMaterial(int iOriginalIndex);

	//find/create a material using the given texture index (as the primary texture)
	static int GetMaterialIDForTexture(int iTextureID,bool bCreate);

	static void DestroyMaterial(int iMaterialID);

	static void NewConversionList(int iLength);
	static int ConvertIndex(int iOriginalIndex);

	#ifdef ASURA_TOOLS
	//Converts a material ID to an index suitable for saving. Uses a conversion array created by Tools_SaveAsura::PrepareMeshMaterialConversionLists
	//RAR 09/06/2006
	static int ConvertIDToIndex(int iMaterialID,const int* piMaterialIDToIndexConversion);
	#endif


	static inline bool IsMaterialSelfIlluminating(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_SelfIlluminate)!=0;
	}
	static inline bool IsMaterialTransparent(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_Transparent)!=0;
	}
	static inline bool IsMaterialReflective(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_Reflective)!=0;
	}
	static inline bool IsMaterialFMV(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_FMV)!=0;
	}
	static inline bool IsMaterialUsingDetailMap(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_UseDetailMap)!=0;
	}
	static inline bool IsMaterialFlowingWater(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_FlowingWater)!=0;
	}
	static inline bool IsMaterialWaterfall(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_Waterfall)!=0;
	}
	static inline bool IsMaterialSubdividePolygon(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_SubdividePolygon)!=0;
	}
	static inline bool IsMaterial3dDecal(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_3dDecal)!=0;
	}

	static inline bool IsMaterialPS2AdditiveAlpha(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_PS2AdditiveAlpha)!=0;
	}
	static inline bool IsMaterialRainSplashes(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_RainSplashes)!=0;
	}
	static inline bool IsMaterialSpecularSoft(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_Specular_Soft)!=0;
	}
	static inline bool IsMaterialSpecularSharp(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_Specular_Sharp)!=0;
	}
	static inline bool IsMaterialAlphaDecal(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_AlphaDecal)!=0;
	}
	static inline bool IsMaterialParallaxMapping(int iMaterialID)
	{
		// KFJL - This flag isn't altered by texture anims, but it could be in theory
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_ParallaxMapping)!=0;
	}
	static inline bool IsMaterialTextureAnim(int iMaterialID)
	{
		//RAR This flag isn't altered by texture anims. 16/02/2006
		return (GetMaterialFlagsIgnoringAnim(iMaterialID)&AsuraMaterialFlag_TextureAnim)!=0;
	}
	static inline bool IsMaterialDontRender(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_DontRender)!=0;
	}
	static inline bool IsMaterialNoCollisions(int iMaterialID)
	{
		return (GetMaterialFlags(iMaterialID)&AsuraMaterialFlag_NoCollisions)!=0;
	}
    static bool IsMaterialNormalMapped( int iMaterialID );
    static bool IsMaterialDualNormalMap( int iMaterialID );
	static bool IsMaterialDualTexture( int iMaterialID );
    static bool IsMaterialSpecularMapped( int iMaterialID );	// VGP 9Feb2006
    static bool IsMaterialDualSpecularMap( int iMaterialID );	// VGP 9Feb2006

	#ifdef ASURA_TOOLS
	//OR's together flags for all materials using the texture
	static unsigned int GetMaterialFlagsUsedWithTexture(int iTextureID);
	// (SB - 22/05/07) Equivalent to the above for project flags
	static unsigned int GetProjectFlagsUsedWithTexture(int iTextureID);
	//AND's together flags for all materials using the texture (yeah , snappy name)
	static unsigned int GetAndedMaterialFlagsUsedWithTexture(int iTextureID);
	// (SB - 22/05/07) Equivalent to the above for project flags
	static unsigned int GetAndedProjectFlagsUsedWithTexture(int iTextureID);

	//Updates texture usage flags by checking for all materials that use the texture. RAR 08/05/2006
	static void UpdateTextureUsageFlags(int iTextureID);


	//Updates all materials using the old animation hash to use the new value instead
	static void ReplaceTextureAnimation(Asura_Hash_ID uOldHash,Asura_Hash_ID uNewHash);
	#endif
	
    static int TranslateMaterialIDToArrayIdx( int iMaterialID )
    {
        return( ASURA_MATERIAL_NONE == iMaterialID ? ASURA_MATERIAL_NONE : iMaterialID - ASURA_MATERIAL_FIRSTID );
    }
    static int TranslateArrayIdxToMaterialID( int iArrayIdx )
    {
        return( ASURA_MATERIAL_NONE == iArrayIdx ? ASURA_MATERIAL_NONE : iArrayIdx + ASURA_MATERIAL_FIRSTID );
    }

	//Get the current uv offsets for the texture animation on the material.
	//Returns false if it fails , or there is no animation , leaving the output untouched.
	//If pxController is NULL then animation uses its own internal controller
	static bool GetTextureAnimationState(int iMaterialID,const class Asura_TextureAnimationController* pxController,struct Asura_TextureAnimationOutput* pxOutput);

	//get the texture animation used by a material
	static Asura_Hash_ID GetTextureAnimation(int iMaterialID)
	{
		Asura_Material* pxMaterial = GetMaterial(iMaterialID);
		return pxMaterial ? pxMaterial->m_uAnimHash : ASURA_HASH_ID_UNSET;
	}


	// Helper function to compare two materials' properties; designed to be used with qsort.
    //
	// Returns	-1 if iMat1 should go before iMat2
	//
	//			 0 if materials are identical
	//
	//			+1 if iMat1 should go after iMat2
	//
	// Order of importance is determined by:
	//
	// 1. Required render order e.g. transparent polys need to go last
	// 2. Vertex size. Some materials (e.g. normal mapped ones) have more data
	// 3. Texture (to group materials of the same texture together)
	// 4. Material ID
	// 
	static int CompareMaterials( const int iMat1, const int iMat2 );
	
	static u_int GetActiveMaterialCount() 
	{ 
		Asura_Assert( NULL != s_pxHandleList, "Missing Material Handle List");
		const u_int uCount = s_pxHandleList->GetNumberOfActiveHandles();
		return( uCount);
	}

	static u_int GetFreeMaterialCount() 
	{ 
		Asura_Assert( NULL != s_pxHandleList, "Missing Material Handle List");
		const u_int uCountTotal = s_pxHandleList->GetTotalNumberOfHandles();
		const u_int uCountActive = s_pxHandleList->GetNumberOfActiveHandles();
		Asura_Assert( uCountActive <= uCountTotal, "Material Handle list size problem");
		const u_int uCount = uCountTotal - uCountActive;
		return( uCount);
	}

	static void WriteResetGroupReport(u_int uGroup, Asura_Stream& xLog);

private :
	static u_int GetAnimatedMaterialFlags(const Asura_Material* pxMaterial);

	static Asura_Material* s_pxMaterials;
	static Asura_Handle_List* s_pxHandleList;
	static int* s_piLoadGroups;

	static int* s_piConversionArray;
	static int s_iConversionArraySize;

#if defined( ASURA_TOOLS)
	static u_int s_uAllowedRenderFlags;
#endif
};



#endif //__ASURA_MATERIAL_SYSTEM_H__
