////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Core/Asura_MemTracker.h"

#ifdef ASURA_ENABLE_MEMTRACKER

#include "Core/Asura_UserInput.h"
#include "Rendering/Asura_Render.h"
#include "Core/Memory/Asura_Mem.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//#define DRAW_REAL_HEAP
//#define DEBUG_COMPARE_AGAINST_HEAP_WALK

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Asura_MemTracker* Asura_MemTracker::s_pxObject = NULL;
bool Asura_MemTracker::s_bShowMemTracker = true;

#ifdef DRAW_REAL_HEAP
static unsigned int g_auBuckets2 [ASURA_MEMTRACKER_NUM_BUCKETS];
#endif

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Asura_MemTracker* Asura_MemTracker::GetInstance ()
{
	if (s_pxObject == NULL)
	{
		s_pxObject = new Asura_MemTracker ();
	}

	return s_pxObject;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::DestroyInstance ()
{
	delete s_pxObject;
	s_pxObject = NULL;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::Initialise_ConsoleCommands ()
{
	Asura_CommandConsole::AddVar (L"Debug.EnableMemTracker", &s_bShowMemTracker, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Asura_MemTracker::Asura_MemTracker ()
{
	Asura_Mem::Set (m_auBuckets, 0x00, sizeof (unsigned int) * ASURA_MEMTRACKER_NUM_BUCKETS);

	m_uMaxAllocations = 0;

	m_uMemoryMin = 0;
	m_uMemoryMax = 0x80000000; // 2gb (virtual memory available to pc process)
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::SetMemoryMin (const unsigned int uMemoryMin)
{
	m_uMemoryMin = uMemoryMin;

	Asura_Assert (m_uMemoryMin < m_uMemoryMax, "Asura_MemTracker: Memory ranges are invalid");

	m_fCoefficient = static_cast <float> (ASURA_MEMTRACKER_NUM_BUCKETS) / static_cast <float> (m_uMemoryMax - m_uMemoryMin);
	m_fBucketSize = static_cast <float> (m_uMemoryMax - m_uMemoryMin) / static_cast <float> (ASURA_MEMTRACKER_NUM_BUCKETS);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::SetMemoryMax (const unsigned int uMemoryMax)
{
	m_uMemoryMax = uMemoryMax;

	Asura_Assert (m_uMemoryMin < m_uMemoryMax, "Asura_MemTracker: Memory ranges are invalid");

	m_fCoefficient = static_cast <float> (ASURA_MEMTRACKER_NUM_BUCKETS) / static_cast <float> (m_uMemoryMax - m_uMemoryMin);
	m_fBucketSize = static_cast <float> (m_uMemoryMax - m_uMemoryMin) / static_cast <float> (ASURA_MEMTRACKER_NUM_BUCKETS);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::AddAllocation (void* pAddress, unsigned int uSize, const bool bUse2ndTracker /* = FALSE */)
{
	unsigned int uAddress = reinterpret_cast <unsigned int> (pAddress);

	Asura_Assert ((uAddress >= m_uMemoryMin) && ((uAddress + uSize) <= m_uMemoryMax), "Asura_MemTracker: Out of bounds of memory track's min and max values");

	if ((uAddress < m_uMemoryMin) || ((uAddress + uSize) > m_uMemoryMax))
	{
		return;
	}

	float fStartBlockUsedCoeff = m_fCoefficient * static_cast <float> (uAddress - m_uMemoryMin);
	const unsigned int uStartIndex = Asura_Maths::FloatToInt_Chop (fStartBlockUsedCoeff);
	fStartBlockUsedCoeff -= static_cast <float> (uStartIndex);

	float fEndBlockUsedCoeff = m_fCoefficient * static_cast <float> ((uAddress + uSize) - m_uMemoryMin);
	const unsigned int uEndIndex = Asura_Maths::FloatToInt_Chop (fEndBlockUsedCoeff);
	fEndBlockUsedCoeff -= static_cast <float> (uEndIndex);

	Asura_Assert (uStartIndex < ASURA_MEMTRACKER_NUM_BUCKETS, "Asura_MemTracker: Invalid bucket index");
	Asura_Assert (uEndIndex < ASURA_MEMTRACKER_NUM_BUCKETS, "Asura_MemTracker: Invalid bucket index");

	if ((uStartIndex >= ASURA_MEMTRACKER_NUM_BUCKETS) || (uEndIndex >= ASURA_MEMTRACKER_NUM_BUCKETS))
	{
		return;
	}

	for (unsigned int n = uStartIndex; n <= uEndIndex; n ++)
	{
		unsigned int uNumBytes;

		// Calculate the number of bytes to put into this bucket
		if (uStartIndex == uEndIndex)
		{
			uNumBytes = static_cast <unsigned int> ((fEndBlockUsedCoeff - fStartBlockUsedCoeff) * m_fBucketSize);
		}
		else if (n == uStartIndex)
		{
			uNumBytes = static_cast <unsigned int> ((1.0f - fStartBlockUsedCoeff) * m_fBucketSize);
		}
		else if (n == uEndIndex)
		{
			uNumBytes = static_cast <unsigned int> (fEndBlockUsedCoeff * m_fBucketSize);
		}
		else
		{
			uNumBytes = static_cast <unsigned int> (m_fBucketSize);
		}

		Asura_Assert ((m_auBuckets [n] + uNumBytes) <= static_cast <unsigned int> (m_fBucketSize), "Asura_MemTracker: Bucket overflow");

		#ifdef DRAW_REAL_HEAP
		if (bUse2ndTracker)
		{
			g_auBuckets2 [n] ++;
		}
		else
		#endif
		{
			m_auBuckets [n] += uNumBytes;
		}

		m_uMaxAllocations = Asura_Maths::Max(m_uMaxAllocations, m_auBuckets [n]);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::RemoveAllocation (void* pAddress, unsigned int uSize)
{
	unsigned int uAddress = reinterpret_cast <unsigned int> (pAddress);

	Asura_Assert ((uAddress >= m_uMemoryMin) && ((uAddress + uSize) <= m_uMemoryMax), "Asura_MemTracker: Out of bounds of memory track's min and max values");

	if ((uAddress < m_uMemoryMin) || ((uAddress + uSize) > m_uMemoryMax))
	{
		return;
	}

	float fStartBlockUsedCoeff = m_fCoefficient * static_cast <float> (uAddress - m_uMemoryMin);
	const unsigned int uStartIndex = Asura_Maths::FloatToInt_Chop (fStartBlockUsedCoeff);
	fStartBlockUsedCoeff -= static_cast <float> (uStartIndex);

	float fEndBlockUsedCoeff = m_fCoefficient * static_cast <float> ((uAddress + uSize) - m_uMemoryMin);
	const unsigned int uEndIndex = Asura_Maths::FloatToInt_Chop (fEndBlockUsedCoeff);
	fEndBlockUsedCoeff -= static_cast <float> (uEndIndex);

	Asura_Assert (uStartIndex < ASURA_MEMTRACKER_NUM_BUCKETS, "Asura_MemTracker: Invalid bucket index");
	Asura_Assert (uEndIndex < ASURA_MEMTRACKER_NUM_BUCKETS, "Asura_MemTracker: Invalid bucket index");

	if ((uStartIndex >= ASURA_MEMTRACKER_NUM_BUCKETS) || (uEndIndex >= ASURA_MEMTRACKER_NUM_BUCKETS))
	{
		return;
	}

	for (unsigned int n = uStartIndex; n <= uEndIndex; n ++)
	{
		unsigned int uNumBytes;

		// Calculate the number of bytes to put into this bucket
		if (uStartIndex == uEndIndex)
		{
			uNumBytes = static_cast <unsigned int> ((fEndBlockUsedCoeff - fStartBlockUsedCoeff) * m_fBucketSize);
		}
		else if (n == uStartIndex)
		{
			uNumBytes = static_cast <unsigned int> ((1.0f - fStartBlockUsedCoeff) * m_fBucketSize);
		}
		else if (n == uEndIndex)
		{
			uNumBytes = static_cast <unsigned int> (fEndBlockUsedCoeff * m_fBucketSize);
		}
		else
		{
			uNumBytes = static_cast <unsigned int> (m_fBucketSize);
		}

		Asura_Assert (m_auBuckets [n] >= uNumBytes, "Asura_MemTracker: Will cause negative value in bucket.  Maybe a delete on invalid memory?");

		m_auBuckets [n] -= uNumBytes;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::Render ()
{
	if ((s_pxObject == NULL) || (!s_bShowMemTracker))
	{
		return;
	}

	#ifdef DEBUG_COMPARE_AGAINST_HEAP_WALK
	if (Asura_UserInput::DebouncedKeyPressed (ASURA_KEYID_SPACE))
	{
		static u_int uStage = 0;

		switch (uStage)
		{
			case 0:
			{
				s_pxObject->WalkProcessHeap ();
				uStage ++;
				break;
			}

			case 1:
			{
				static BYTE* pMemory = NULL;

				if (pMemory)
				{
					delete [] pMemory;
					pMemory = NULL;
				}
				else
				{
					pMemory = new BYTE [1024 * 1024 * 5];
				}

				//uStage ++;

				break;
			}
		}
	}
	#endif

	Asura_Render::FlushDisplay ();

	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;

	const float fBarX = 100.0f;
	const float fBarY = 100.0f;

	float fX = fBarX;
	float fY = fBarY;
	float uWidth = 0;
	float fHeight = 25.0f;

	#if DRAW_REAL_HEAP
	for (unsigned int i = 0; i < 2; i ++)
	#endif
	{
		unsigned int uLastColour = 0x00000000;

		#if DRAW_REAL_HEAP
		unsigned int* puBucketArray = (i == 0) ? s_pxObject->m_auBuckets : g_auBuckets2;

		if (i == 1)
		{
			fX = fBarX;
			fY += fHeight;
			uWidth = 0;
		}
		#else
		unsigned int* puBucketArray = s_pxObject->m_auBuckets;
		#endif

		for (unsigned int n = 0; n < ASURA_MEMTRACKER_NUM_BUCKETS; n ++)
		{
			unsigned int uColour;

			if (puBucketArray [n] == 0)
			{
				uColour = 0xffffffff;
			}
			else
			{
				unsigned int uColourChannel = static_cast <unsigned int> (255.0f * (puBucketArray [n] / s_pxObject->m_fBucketSize));
				uColourChannel = 255 - uColourChannel;
				Asura_Assert (uColourChannel < 256, "Asura_MemTracker: Error in maths - colour channel shouldn't be greater than 255");
				uColour = 0xff000000 | (uColourChannel << 16) | (uColourChannel << 8) | uColourChannel;
			}

			if (uColour == uLastColour)
			{
				uWidth ++;
			}
			else
			{
				if (uWidth > 0)
				{
					// Flush the previous colour block

					Asura_Render::Overlay.SetRectXY (fX, fY, static_cast <float> (uWidth), fHeight);
					Asura_Render::Overlay.Colour = uLastColour;
					Asura_Render::Overlay.Render ();
				}

				// Reset for new colour block

				fX += static_cast <float> (uWidth);
				uWidth = 1;

				uLastColour = uColour;
			}
		}

		if (uWidth > 0)
		{
			// Flush the last colour block

			Asura_Render::Overlay.SetRectXY (fX, fY, static_cast <float> (uWidth), fHeight);
			Asura_Render::Overlay.Colour = uLastColour;
			Asura_Render::Overlay.Render ();
		}

		// Draw a border around the whole bar

		const float fLeft = fBarX;
		const float fTop = fBarY;
		const float fRight = fBarX + ASURA_MEMTRACKER_NUM_BUCKETS;
		const float fBottom = fBarY + fHeight;

		Asura_Render::Render2DLine (fLeft, fTop, fRight, fTop, 0xff000000);
		Asura_Render::Render2DLine (fRight, fTop, fRight, fBottom, 0xff000000);
		Asura_Render::Render2DLine (fRight, fBottom, fLeft, fBottom, 0xff000000);
		Asura_Render::Render2DLine (fLeft, fBottom, fLeft, fTop, 0xff000000);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_MemTracker::WalkProcessHeap ()
{
	#if DRAW_REAL_HEAP

	const HANDLE pHeap = GetProcessHeap ();

	PROCESS_HEAP_ENTRY xHeapEntry;

	memset (&xHeapEntry, 0x00, sizeof (PROCESS_HEAP_ENTRY));

	memset (g_auBuckets2, 0x00, sizeof (unsigned int) * ASURA_MEMTRACKER_NUM_BUCKETS);

	while (HeapWalk (pHeap, &xHeapEntry))
	{
		AddAllocation (xHeapEntry.lpData, xHeapEntry.cbData, true);
	}

	Asura_Assert (GetLastError () == ERROR_NO_MORE_ITEMS, "Asura_MemTracker: Error when walking heap");

	#endif
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#endif // ASURA_ENABLE_MEMTRACKER

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
