// Asura_MemHeap.cpp
// Copyright (c) 2007 Rebellion. All rights reserved.
// TY - 30.10.07

//This is essentially a wrapped version of doug lee malloc (dlm)

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"						//Main asura header

#include "Core/Asura_MemHeap.h"				//My header

#include "Core/Asura_Timers.h"				//Other asura headers
#include "Debugging/Asura_Logfile.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
#define MEMHEAP_CHECK(X,Y)  if(!(X)) { Asura_Assert(false,Y);	++m_iIntegrityFailures; }
#endif

// M E M B E R   F U N C T I O N   D E F I N I T I O N S //////////////////////

Asura_MemHeap::Asura_MemHeap( void )
{
	InitData();

    // Defaults.
	Asura_Memheap_Core_Details xCoreDetails;
	SetCoreDetails(&xCoreDetails);
}


Asura_MemHeap::Asura_MemHeap( void* pvMemory, const size_t uSize )
{
	InitData();

	//Treat this as a single page of memory
	Asura_Memheap_Core_Details xDetails;
	xDetails.m_iMaxPages = 1;
	xDetails.m_uPageSize = 0;
	
	SetCoreDetails(&xDetails);

	m_pcBaseMem = static_cast<char*>(pvMemory);
	m_uBaseMemSize = uSize;
}


Asura_MemHeap::Asura_MemHeap( const Asura_Memheap_Core_Details& xCoreDetails )
{
	InitData();
	SetCoreDetails( &xCoreDetails );
}


#if ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
Asura_MemHeap::Asura_MemHeap(Asura_MemHeap* pxParent, const Asura_Subheap_Core_Details* pCoreDetails)
{
  	Asura_Assert( pxParent, "Attempting to construct a mem heap without a valid parent" );

  	InitData();
  		
  	m_pxParent = pxParent;

  	if(pCoreDetails)
  	{
  		SetCoreDetails(pCoreDetails);
  	}
  	else
  	{
  		//Use defaults
  		Asura_Subheap_Core_Details xCoreDetails;
  		SetCoreDetails(&xCoreDetails);
  	}
}
#endif


void Asura_MemHeap::MutexAcquire () const
{
	if(m_bThreadSafetyEnabled)
	{		
		m_xMutex.Lock();
	}
	else if(Asura_Thread_System::GetCurrentThreadID() != m_uCreationThread)
	{
		Asura_Assert(m_bThreadSafetyWarningDisabled,"Asura_Memheap is being used by multiple threads, but thread safety is disabled");
	}
}


void Asura_MemHeap::MutexRelease () const
{	
	if(m_bThreadSafetyEnabled)
	{
		m_xMutex.Unlock();
	}
}


void	Asura_MemHeap::SetCoreDetails	(const Asura_Memheap_Core_Details* pxDetails)
{
	Asura_Assert(pxDetails,		"Cannot set Asura_Memheap core details, pointer is null");
  	Asura_Assert(!m_pxParent,	"Cannot set Asura_Memheap core details, we are a subheap");

	if(pxDetails)
	{
		m_pfnAllocFunction   = pxDetails->m_pfnAllocator;
		m_pfnTrimFunction    = pxDetails->m_pfnTrimmer;

		SetPageSize		(pxDetails->m_uPageSize);
		SetMaxPages		(pxDetails->m_iMaxPages);	
		SetTopPad		(pxDetails->m_uPad);
		SetTrimThreshold(pxDetails->m_uTrim);

		if(pxDetails->m_bIsContiguous)
		{
			SetContiguous();
		}
		else
		{
			SetNoncontiguous();
		}
	}	
}


#if ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
void	Asura_MemHeap::SetCoreDetails (const Asura_Subheap_Core_Details* pxDetails) 
{
  	Asura_Assert(pxDetails,		"Cannot set Asura_Memheap core details, pointer is null");
  	if(pxDetails)
  	{
  		SetPageSize	(pxDetails->m_uPageSize);
  		SetMaxPages	(pxDetails->m_iMaxPages);	
  		SetTopPad	(pxDetails->m_uPad);
  	}	
  	SetTrimThreshold(0);
  	m_pfnAllocFunction   = NULL;
  	m_pfnTrimFunction    = NULL;
  	SetNoncontiguous();
}
#endif


void Asura_MemHeap::SetPageSize (unsigned int	uPageSize)
{
	m_xMallocState.m_uPagesize = uPageSize;	
}


void Asura_MemHeap::SetMaxPages	(int iMaxPages)
{
	m_xMallocState.m_iMaxPages = iMaxPages;
}


void Asura_MemHeap::SetTrimThreshold (size_t uTrim)
{
	m_xMallocState.m_uTrimThresh = uTrim;		
}


void Asura_MemHeap::SetTopPad (size_t uPad)
{
	m_xMallocState.m_uTopPad = uPad;
}


size_t Asura_MemHeap::GetFreeAllocatedSpace() const
{
	MutexAcquire();

	UpdateMallinfo();

	size_t uRet = 0;

	if(m_pcBaseMem && (m_xMallinfo.m_uSystemMemCalls == 0))
	{
		//If we havent called for system memory but are using fixed memory, return the total memory count
		uRet = m_uBaseMemSize;	
	}	
	else
	{
		uRet = static_cast<size_t>(m_xMallinfo.m_iTotalFree);
	}
	
	MutexRelease();

	return uRet;
}


Asura_MemHeap::~Asura_MemHeap()
{
	void* pvMem;
	//If we are using the default allocator, free our allocations
	pvMem = m_xDefaultAllocs.GetHeadObject();
	while(pvMem)
	{
		Asura_Deallocate(pvMem);
		m_xDefaultAllocs.RemoveFromList(static_cast<char*>(pvMem));
		pvMem = m_xDefaultAllocs.GetHeadObject();
	}
}

void Asura_MemHeap::InitData ()
{
	m_pcBaseMem			= 0;
	m_uBaseMemSize		= 0;
	m_pfnAllocFunction	= 0;
	m_pfnTrimFunction	= 0;
  	m_pxParent			= 0;

	m_xDefaultAllocs.ClearList();
	
	m_bThreadSafetyEnabled = false;
	m_bThreadSafetyWarningDisabled = false;

	m_bUpdateMallinfo = true;

	m_uCreationThread = Asura_Thread_System::GetCurrentThreadID();

	MallocInitState();
	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	m_iIntegrityFailures = 0;
	Checks_MallocState();
	#endif
}

void 	Asura_MemHeap::SetFastchunks ()
{
	m_xMallocState.m_uMaxFastSize |=  (uASURA_MEMHEAP_FASTCHUNKS_BIT | uASURA_MEMHEAP_ANYCHUNKS_BIT);
}


bool 	Asura_MemHeap::HaveFastchunks () const
{
	if(m_xMallocState.m_uMaxFastSize & uASURA_MEMHEAP_FASTCHUNKS_BIT)
	{
		return true;
	}
	return false;
}


bool 	Asura_MemHeap::HaveAnyChunks () const
{
	if(m_xMallocState.m_uMaxFastSize &  uASURA_MEMHEAP_ANYCHUNKS_BIT)
	{
		return true;
	}
	return false;
}


void 	Asura_MemHeap::SetAnyChunks ()
{
	m_xMallocState.m_uMaxFastSize |=  uASURA_MEMHEAP_ANYCHUNKS_BIT;
}


void 	Asura_MemHeap::ClearFastChunks ()
{
	m_xMallocState.m_uMaxFastSize &= ~(uASURA_MEMHEAP_FASTCHUNKS_BIT);
}


void Asura_MemHeap::SetInuseBitAtOffset (Asura_Memheap_Malloc_Chunk* pChunk, size_t uOffset)
{
	Asura_Memheap_Malloc_Chunk* pOffsetChunk = reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(reinterpret_cast<char*>(pChunk) + uOffset);
	pOffsetChunk->m_uSize |= uASURA_MEMHEAP_PREV_INUSE;
}


int	Asura_MemHeap::GetInuseBitAtOffset (Asura_Memheap_Malloc_Chunk* pChunk, size_t uOffset) const
{
	Asura_Memheap_Malloc_Chunk* pOffsetChunk = reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(reinterpret_cast<char*>(pChunk) + uOffset);
	return static_cast<int>(pOffsetChunk->m_uSize & uASURA_MEMHEAP_PREV_INUSE);
}


u_int Asura_MemHeap::GetSizeFromRequest (size_t uRequest) const
{
	if(uRequest + sizeof(size_t) + uASURA_MEMHEAP_MALLOC_ALIGN_MASK < uASURA_MEMHEAP_MINSIZE)
	{
		return uASURA_MEMHEAP_MINSIZE;
	}
	else
	{
		return static_cast<u_int>((uRequest) + sizeof(size_t) + uASURA_MEMHEAP_MALLOC_ALIGN_MASK) & ~uASURA_MEMHEAP_MALLOC_ALIGN_MASK;
	}													
}


void Asura_MemHeap::SetMaxFast (size_t uSize)
{
	if(uSize == 0)
	{
		//This should not be hit
		m_xMallocState.m_uMaxFastSize = uASURA_MEMHEAP_SMALLBIN_WIDTH;
	}
	else
	{
		m_xMallocState.m_uMaxFastSize = GetSizeFromRequest(uSize) | (m_xMallocState.m_uMaxFastSize & (uASURA_MEMHEAP_FASTCHUNKS_BIT | uASURA_MEMHEAP_ANYCHUNKS_BIT));
	}	
}


bool Asura_MemHeap::InSmallbinRange(size_t uSize) const
{
	return (uSize < uASURA_MEMHEAP_MIN_LARGE_SIZE);
}


bool Asura_MemHeap::ChunkInUse (Asura_Memheap_Malloc_Chunk* pChunk) const
{
	Asura_Memheap_Malloc_Chunk* pNextChunk = reinterpret_cast<Asura_Memheap_Malloc_Chunk*>((reinterpret_cast<char*>(pChunk) + (pChunk->m_uSize & ~uASURA_MEMHEAP_PREV_INUSE)));
	return (pNextChunk->m_uSize & uASURA_MEMHEAP_PREV_INUSE);
}


void Asura_MemHeap::UnlinkChunk	(const Asura_Memheap_Malloc_Chunk* pxChunk)
{
	//Update the forward and backward pointers of the surrounding chunks
	Asura_Memheap_Malloc_Chunk* pxNext = pxChunk->m_pxNextChunk;
	Asura_Memheap_Malloc_Chunk* pxPrev = pxChunk->m_pxPrevChunk; 

	pxNext->m_pxPrevChunk = pxPrev;
	pxPrev->m_pxNextChunk = pxNext;
}


bool	Asura_MemHeap::IsContiguous () const
{
	return m_xMallocState.m_bContiguous;
}


void	Asura_MemHeap::SetNoncontiguous	()
{	
	m_xMallocState.m_bContiguous = false;

#if !ASURA_MEMHEAP_SUPPORT_NONCONTIGUITIES
    // (MP - 06.05.2008) Non-contiguous memheaps (in the sense of the memory
    // returned by the m_pfnAllocFunction callback) have a problem, inherited
    // from the original DLM code:
    //
    //     * Memory isn't ever (significantly) trimmed. Because the code
    //       always seeks to keep a small amount (uASURA_MEMHEAP_MINSIZE) of
    //       memory back in the top chunk, it will only continue trimming if
    //       over time it can consolidate free'd memory into this top chunk.
    //       However, when blocks of memory given to the memheap are
    //       genuinely non-contiguous this will never happen and it will
    //       never trim any blocks in their entirety because of this small
    //       amount of memory retained at the start. (Though if the callback
    //       supports partial trimming of a block then that may happen at
    //       least, so that it retains only a single page of the last block.)
    //
    //           * This could be fixed by recognising in the trim code the
    //             case that the entire top chunk, as originally fetched from
    //             the 'system', is free and that another chunk has at least
    //             uASURA_MEMHEAP_MINSIZE bytes available but there doesn't
    //             seem to be sufficient information for this and isn't
    //             something that the original code was really intended to
    //             cope with.
    //
    //           * Another option would be to support non-contiguities
    //             without trimming. However, this is unlikely to be very
    //             viable in most game situations; without trimming there's
    //             no way of guaranteeing that a sufficiently large
    //             contiguous block is available to load the next level.
    //             (Suppose, for example, that some particular type of data -
    //             say, environment geometry - allocates a single large block.
    //             If one level allocates 5MB and no trimming occurs then
    //             it may not be possible to support the next level requiring
    //             a contiguous 10MB block.)
    //
    //           * It's worth also mentioning that the memheap unit tests
    //             we've had previously haven't detected this problem since
    //             they don't guarantee that the memory they're handing to
    //             the test memheap is non-contiguous. Even if they did they
    //             would need to also explicitly verify that when they've
    //             deallocated everything again that the memheap has trimmed
    //             all memory given to it (except for a single page that it
    //             will never give up).
    //
    // There was also another problem but this was straightforward to fix
    // (and has been):
    //
    //     * The memory isn't aligned properly if the min malloc alignment is
    //       greater than 8.
    //
    // Note that disabling supported for non-contiguous memory also disables
    // support for sub-heaps since they're of no real use if other stuff
    // isn't allowed to also allocate from the parent heap as well as the
    // sub-heap. By allowing this though the memory given to the sub-heap is
    // typically going to be non-contiguous.)
    Asura_Assert( false, "Non-contiguous memheaps not supported!" );
#endif
}


void	Asura_MemHeap::SetContiguous		()
{
	m_xMallocState.m_bContiguous = true;
}


Asura_Memheap_Malloc_Chunk* Asura_MemHeap::NextChunk(Asura_Memheap_Malloc_Chunk* pChunk) const
{	
	//Get the next chunk
	return reinterpret_cast<Asura_Memheap_Malloc_Chunk*>((reinterpret_cast<char*>(pChunk) + (pChunk->m_uSize & ~uASURA_MEMHEAP_PREV_INUSE) ));
}


Asura_Memheap_Malloc_Chunk* Asura_MemHeap::PrevChunk(Asura_Memheap_Malloc_Chunk* pChunk) const
{
	//Get the previous chunk
	return reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(reinterpret_cast<char*>(pChunk) - pChunk->m_uPrevSize );
}


Asura_Memheap_Malloc_Chunk*	Asura_MemHeap::GetChunkFromMemory	(void* pMem) const
{
	//Get a chunk from a memory address
	return reinterpret_cast<Asura_Memheap_Malloc_Chunk*>((static_cast<char*>(pMem) - 2*sizeof(size_t)));
}


void*	Asura_MemHeap::GetMemoryFromChunk	(void* pBase) const
{
	//Get a memory address for a chunk (i.e. seek past the header)
	return (static_cast<char*>(pBase) + 2*sizeof(size_t));
}


size_t Asura_MemHeap::ChunkSize (Asura_Memheap_Malloc_Chunk* pChunk) const
{
	return (pChunk->m_uSize & ~(uASURA_MEMHEAP_SIZE_BITS));
}


bool Asura_MemHeap::PrevChunkInUse (Asura_Memheap_Malloc_Chunk* pChunk) const
{
	return (pChunk->m_uSize & uASURA_MEMHEAP_PREV_INUSE);
}


u_int Asura_MemHeap::BlockFromIndex(u_int u) const
{
	return (u >> uASURA_MEMHEAP_BINMAPSHIFT);
}


u_int Asura_MemHeap::BitFromIndex (u_int u) const
{
	return ((1U << (u & ((1U << uASURA_MEMHEAP_BINMAPSHIFT)-1))));
}


Asura_Memheap_Malloc_Chunk* Asura_MemHeap::GetNextBin(Asura_Memheap_Malloc_Chunk* pChunk) const
{
	//get the next bin
	return reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(reinterpret_cast<char*>(pChunk) + (sizeof(Asura_Memheap_Malloc_Chunk*) << 1));
}


Asura_Memheap_Malloc_Chunk* Asura_MemHeap::GetBinAt(u_int uIndex) const
{
	//get the bin at index u
	Asura_Memheap_Malloc_State* pxState = const_cast<Asura_Memheap_Malloc_State*>(&m_xMallocState);	//yuck, sorry
	char* pcChunk = reinterpret_cast<char*>( &(pxState->m_xMemoryBins[uIndex << 1])) - (sizeof(size_t) << 1 ) ;
	return reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(pcChunk);
}


u_int Asura_MemHeap::GetBinmap (u_int uIndex) const
{
	return (m_xMallocState.m_uBinmap[BlockFromIndex(uIndex)] & BitFromIndex(uIndex));
}


Asura_Memheap_Malloc_Chunk* Asura_MemHeap::GetChunkAtOffset(Asura_Memheap_Malloc_Chunk* pChunk, size_t uOffset) const
{
	return reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(reinterpret_cast<char*>(pChunk) + uOffset);
}


Asura_Memheap_Malloc_Chunk* Asura_MemHeap::GetUnsortedChunks() const
{
	return GetBinAt(1);
}


void Asura_MemHeap::SetHeadSize(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize) 
{
	pChunk->m_uSize = ((pChunk->m_uSize & uASURA_MEMHEAP_PREV_INUSE) | uSize);
}


void Asura_MemHeap::SetHead(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize)
{
	pChunk->m_uSize = uSize;
}


void Asura_MemHeap::SetFoot(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize)
{
	(((Asura_Memheap_Malloc_Chunk*)((char*)(pChunk) + (uSize)))->m_uPrevSize = (uSize));
}


void Asura_MemHeap::MarkBin (size_t u)	
{
	m_xMallocState.m_uBinmap[BlockFromIndex(u)] |= BitFromIndex(u);
}

bool Asura_MemHeap::IsAligned (void* pMem) const
{
	//Is this memory aligned to our prespecified malloc alignment
	return ((reinterpret_cast<uintptr_t>(pMem) & uASURA_MEMHEAP_MALLOC_ALIGN_MASK) == 0);
}


u_int Asura_MemHeap::GetBinIndex(size_t uSize) const
{
	if(InSmallbinRange(uSize))
	{
		return GetSmallbinIndex(uSize);
	}
	return GetLargebinIndex(uSize);
}


size_t Asura_MemHeap::GetMaxFastSize() const
{
	return (m_xMallocState.m_uMaxFastSize & ~(uASURA_MEMHEAP_FASTCHUNKS_BIT | uASURA_MEMHEAP_ANYCHUNKS_BIT));
}


u_int Asura_MemHeap::GetLargebinIndex(size_t uSize) const
{
	unsigned int x = uSize >> uASURA_MEMHEAP_SMALLBIN_WIDTH; 
	unsigned int m;            // bit position of highest set bit of m 

	if (x >= 0x10000) 
	{
		return uASURA_MEMHEAP_NBINS-1;
	}

	//Based on branch-free nlz algorithm in chapter 5 of Henry
	//S. Warren Jr's book "Hacker's Delight".
	unsigned int n = ((x - 0x100) >> 16) & 8;
	x <<= n; 
	m = ((x - 0x1000) >> 16) & 4;
	n += m; 
	x <<= m; 
	m = ((x - 0x4000) >> 16) & 2;
	n += m; 
	x = (x << m) >> 14;
	m = 13 - n + (x & ~(x>>1));

	/* Use next 2 bits to create finer-granularity bins */
	return uASURA_MEMHEAP_NSMALLBINS + (m << 2) + ((uSize >> (m + 6)) & 3);
}


u_int Asura_MemHeap::GetFastbinIndex(size_t uSize) const
{
	return static_cast<u_int>((uSize >> 3) - 2);
}


u_int Asura_MemHeap::GetSmallbinIndex(size_t uSize) const
{
	return static_cast<u_int>(uSize >> 3);
}


void* Asura_MemHeap::Alloc(size_t uBytes, size_t uAlignment)
{
	size_t							uRequestSize;			// padded  request size 
	char*							pcMem;					// memory returned by malloc call 
	Asura_Memheap_Malloc_Chunk*		pxChunk;				// corresponding chunk 
	char*							pcAlignMem;				// alignment point within p 
	Asura_Memheap_Malloc_Chunk*		pxNewChunk;				// chunk to return 
	size_t							uNewSize;				// its size 
	size_t							uLeadSize;				// leading space before alignment point 
	Asura_Memheap_Malloc_Chunk*		pxRemainderChunk;		// spare room at end to split off 
	size_t							m_uRemainderSize;		// its size 
	size_t							uSize;

	MutexAcquire();

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_MallocState();
	#endif

	++ m_xMallinfo.m_uRequestedAllocations;
	m_bUpdateMallinfo = true;

	// If need less alignment than we give anyway, just relay to malloc 
	if (uAlignment <=  uASURA_MEMHEAP_MALLOC_ALIGNMENT) 
	{
		void* pRet = Alloc(uBytes);
		MutexRelease();
		return pRet;
	}

	// Otherwise, ensure that it is at least a minimum chunk size 
	if (uAlignment <  uASURA_MEMHEAP_MINSIZE) 
	{
		uAlignment = uASURA_MEMHEAP_MINSIZE;
	}

	// Make sure alignment is power of 2 (in case MINSIZE is not).  
	if ((uAlignment & (uAlignment - 1)) != 0) 
	{
		size_t uAdjust = uASURA_MEMHEAP_MALLOC_ALIGNMENT * 2;
		while (uAdjust < uAlignment) 
		{
			uAdjust <<= 1;
		}
		uAlignment = uAdjust;
	}

	uRequestSize = GetSizeFromRequest(uBytes);

	//Strategy: find a spot within that chunk that meets the alignment
	//request, and then possibly free the leading and trailing space.

	// Call alloc with worst case padding to hit alignment.

	-- m_xMallinfo.m_uRequestedAllocations; //this will be incremented again in Alloc
	m_bUpdateMallinfo = true;

	pcMem  = static_cast<char*>(Alloc(uRequestSize + uAlignment + uASURA_MEMHEAP_MINSIZE));

	if (pcMem == 0) 
	{
		MutexRelease();
		return 0;
	}

	pxChunk = GetChunkFromMemory(pcMem);

	if ((reinterpret_cast<uintptr_t>(pcMem) % uAlignment) != 0) 
	{ 
		/* misaligned 
		Find an aligned spot inside chunk.  Since we need to give back
		leading space in a chunk of at least MINSIZE, if the first
		calculation places us at a spot with less than MINSIZE leader,
		we can move to the next aligned spot -- we've allocated enough
		total room so that this is always possible.*/

		void* pArg = reinterpret_cast<void*>((reinterpret_cast<uintptr_t>(pcMem + uAlignment - 1) & -(static_cast<int>(uAlignment))));
		pcAlignMem = (char*)GetChunkFromMemory(pArg);
		if (static_cast<size_t>(pcAlignMem - reinterpret_cast<char*>(pxChunk)) < uASURA_MEMHEAP_MINSIZE)
		{
			pcAlignMem += uAlignment;
		}

		pxNewChunk = (Asura_Memheap_Malloc_Chunk*)pcAlignMem;
		uLeadSize = pcAlignMem - (char*)(pxChunk);
		uNewSize = ChunkSize(pxChunk) - uLeadSize;

		/* Otherwise, give back leader, use the rest */
		SetHead(pxNewChunk, uNewSize |  uASURA_MEMHEAP_PREV_INUSE);
		SetInuseBitAtOffset(pxNewChunk, uNewSize);
		SetHeadSize(pxChunk, uLeadSize);
		
		Free(GetMemoryFromChunk(pxChunk));
		
		pxChunk = pxNewChunk;

		Asura_Assert (uNewSize >= uRequestSize && ((reinterpret_cast<uintptr_t>(GetMemoryFromChunk(pxChunk))) % uAlignment) == 0, "[Asura_Memheap Internal Checks] Alignment incorrect");
	}

	/* Also give back spare room at the end */
	uSize = ChunkSize(pxChunk);
	if (static_cast<size_t>(uSize) > static_cast<size_t>(uRequestSize +  uASURA_MEMHEAP_MINSIZE)) 
	{
		m_uRemainderSize = uSize - uRequestSize;
		pxRemainderChunk = GetChunkAtOffset(pxChunk, uRequestSize);
		SetHead(pxRemainderChunk, m_uRemainderSize |  uASURA_MEMHEAP_PREV_INUSE);
		SetHeadSize(pxChunk, uRequestSize);
		Free(GetMemoryFromChunk(pxRemainderChunk));
	}

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_InUseChunk(pxChunk);
	#endif

	void* pRet = GetMemoryFromChunk(pxChunk);
	
	MutexRelease();
	return pRet;
}


void *Asura_MemHeap::Alloc ( size_t uBytes )
{
	// Alloc function lifted from DLM
	
	size_t							uRequestSize;		// normalized request size 
	unsigned int					uBinIndex;			// associated bin index 
	Asura_Memheap_Malloc_Chunk*		pxBinChunk;			// associated bin 
	Asura_Memheap_Malloc_Chunk**	ppxFastbinChunk;	// associated fastbin 

	Asura_Memheap_Malloc_Chunk*		pxSelectedChunk;	// inspected/selected chunk 
	size_t							uSelectedSize;		// its size 
	int								uSelectedIndex;		// its bin index 

	Asura_Memheap_Malloc_Chunk*		pxRemainderChunk;	// remainder from a split 
	size_t							uRemainderSize;		// its size 

	unsigned int					uBlock;				// bit map traverser 
	unsigned int					uBit;				// bit map traverser 
	unsigned int					uMapWord;			// current word of binmap

	Asura_Memheap_Malloc_Chunk*		pxNext;				// misc temp for linking 
	Asura_Memheap_Malloc_Chunk*		pxPrev;				// misc temp for linking 

	/*Convert request size to internal form by adding SIZE_SZ bytes
	overhead plus possibly more to obtain necessary alignment and/or
	to obtain a size of at least uASURA_MEMHEAP_MINSIZE, the smallest allocatable
	size. Also, checked_request2size traps (returning 0) request sizes
	that are so large that they wrap around zero when padded and aligned.*/

	MutexAcquire();

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_MallocState();
	#endif

	++ m_xMallinfo.m_uRequestedAllocations;
	m_bUpdateMallinfo = true;

	uRequestSize = GetSizeFromRequest(uBytes);

	//Bypass search if no frees yet
	if (HaveAnyChunks()) 
	{
		//If the size qualifies as a fastbin, first check corresponding bin.
		if (static_cast<size_t>(uRequestSize) <= static_cast<size_t>(m_xMallocState.m_uMaxFastSize)) 
		{ 
			ppxFastbinChunk = &(m_xMallocState.m_xFastbins[(GetFastbinIndex(uRequestSize))]);
			if ( (pxSelectedChunk = *ppxFastbinChunk) != 0) 
			{
				*ppxFastbinChunk = pxSelectedChunk->m_pxNextChunk;
				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_RemallocedChunk(pxSelectedChunk, uRequestSize);
				#endif
				MutexRelease();
				return GetMemoryFromChunk(pxSelectedChunk);
			}
		}

		/*If a small request, check regular bin.  Since these "smallbins"
		hold one size each, no searching within bins is necessary.
		(For a large request, we need to wait until unsorted chunks are
		processed to find best fit. But for small ones, fits are exact
		anyway, so we can check now, which is faster.)*/

		if (InSmallbinRange(uRequestSize)) 
		{
			uBinIndex = GetSmallbinIndex(uRequestSize);
			pxBinChunk = GetBinAt(uBinIndex);

			if ( (pxSelectedChunk = pxBinChunk->m_pxPrevChunk) != pxBinChunk) 
			{
				pxPrev = pxSelectedChunk->m_pxPrevChunk;
				SetInuseBitAtOffset(pxSelectedChunk, uRequestSize);
				pxBinChunk->m_pxPrevChunk = pxPrev;
				pxPrev->m_pxNextChunk = pxBinChunk;

				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
				#endif
				MutexRelease();
				return GetMemoryFromChunk(pxSelectedChunk);
			}
		}

		/* 
		If this is a large request, consolidate fastbins before continuing.
		While it might look excessive to kill all fastbins before
		even seeing if there is space available, this avoids
		fragmentation problems normally associated with fastbins.
		Also, in practice, programs tend to have runs of either small or
		large requests, but less often mixtures, so consolidation is not 
		invoked all that often in most programs. And the programs that
		it is called frequently in otherwise tend to fragment.
		*/

		else 
		{
			uBinIndex = GetLargebinIndex(uRequestSize);
			if (HaveFastchunks()) 
			{
				MallocConsolidate();
			}
		}

		/*Process recently freed or remaindered chunks, taking one only if
		it is exact fit, or, if this a small request, the chunk is remainder from
		the most recent non-exact fit.  Place other traversed chunks in
		bins.  Note that this step is the only place in any routine where
		chunks are placed in bins.*/

		while ( (pxSelectedChunk = GetUnsortedChunks()->m_pxPrevChunk) != GetUnsortedChunks()) 
		{
			pxPrev = pxSelectedChunk->m_pxPrevChunk;
			uSelectedSize = ChunkSize(pxSelectedChunk);

			/*If a small request, try to use last remainder if it is the
			only chunk in unsorted bin.  This helps promote locality for
			runs of consecutive small requests. This is the only
			exception to best-fit, and applies only when there is
			no exact fit for a small chunk.*/

			if ((InSmallbinRange(uRequestSize))							&& 
				(pxPrev == GetUnsortedChunks())							&& 
				(pxSelectedChunk == m_xMallocState.m_xRemainderChunk)	&&
				(uSelectedSize > (uRequestSize + uASURA_MEMHEAP_MINSIZE)) )
			{

				/* split and reattach remainder */
				uRemainderSize = uSelectedSize - uRequestSize;
				pxRemainderChunk = GetChunkAtOffset(pxSelectedChunk, uRequestSize);
				GetUnsortedChunks()->m_pxPrevChunk = GetUnsortedChunks()->m_pxNextChunk = pxRemainderChunk;
				m_xMallocState.m_xRemainderChunk = pxRemainderChunk; 
				pxRemainderChunk->m_pxPrevChunk = pxRemainderChunk->m_pxNextChunk = GetUnsortedChunks();

				SetHead(pxSelectedChunk		,uRequestSize   | uASURA_MEMHEAP_PREV_INUSE);
				SetHead(pxRemainderChunk	,uRemainderSize | uASURA_MEMHEAP_PREV_INUSE);
				SetFoot(pxRemainderChunk	,uRemainderSize);

				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
				#endif
				MutexRelease();
				return GetMemoryFromChunk(pxSelectedChunk);
			}

			/* remove from unsorted list */
			GetUnsortedChunks()->m_pxPrevChunk = pxPrev;
			pxPrev->m_pxNextChunk = GetUnsortedChunks();

			// Take now instead of binning if exact fit 
			if (uSelectedSize == uRequestSize) 
			{
				SetInuseBitAtOffset(pxSelectedChunk, uSelectedSize);
				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
				#endif
				MutexRelease();
				return GetMemoryFromChunk(pxSelectedChunk);
			}

			// place chunk in bin 
			if (InSmallbinRange(uSelectedSize)) 
			{
				uSelectedIndex = GetSmallbinIndex(uSelectedSize);
				pxPrev = GetBinAt(uSelectedIndex);
				pxNext = pxPrev->m_pxNextChunk;
			}
			else 
			{
				uSelectedIndex = GetLargebinIndex(uSelectedSize);
				pxPrev = GetBinAt(uSelectedIndex);
				pxNext = pxPrev->m_pxNextChunk;

				if (pxNext != pxPrev) 
				{
					// if smaller than smallest, place first
					if (static_cast<size_t>(uSelectedSize) < static_cast<size_t>(pxPrev->m_pxPrevChunk->m_uSize)) 
					{
						pxNext = pxPrev;
						pxPrev = pxPrev->m_pxPrevChunk;
					}
					else if (static_cast<size_t>(uSelectedSize) >= static_cast<size_t>(uASURA_MEMHEAP_MIN_LARGE_SIZE)) 
					{

						// maintain large bins in sorted order
						uSelectedSize |= uASURA_MEMHEAP_PREV_INUSE; /* Or with inuse bit to speed comparisons */
						while (static_cast<size_t>(uSelectedSize) < static_cast<size_t>(pxNext->m_uSize))
						{
							pxNext = pxNext->m_pxNextChunk;
						}
						pxPrev = pxNext->m_pxPrevChunk;
					}
				}
			}

			MarkBin(uSelectedIndex);
			pxSelectedChunk->m_pxPrevChunk = pxPrev;
			pxSelectedChunk->m_pxNextChunk = pxNext;
			pxNext->m_pxPrevChunk = pxSelectedChunk;
			pxPrev->m_pxNextChunk = pxSelectedChunk;
		}

		/* If a large request, scan through the chunks of current bin to find one that fits.  (This will be the smallest 
		that fits unless FIRST_SORTED_BIN_SIZE has been changed from default [TY: it hasn't, replaced with 
		uASURA_MEMHEAP_MIN_LARGE_SIZE]) This is the only step where an unbounded number of chunks might be scanned
		without doing anything useful with them. However the lists tend to be short.*/

		if (!InSmallbinRange(uRequestSize)) 
		{
			pxBinChunk = GetBinAt(uBinIndex);

			for (pxSelectedChunk = pxBinChunk->m_pxPrevChunk; pxSelectedChunk != pxBinChunk; pxSelectedChunk = pxSelectedChunk->m_pxPrevChunk) 
			{
				uSelectedSize = ChunkSize(pxSelectedChunk);

				if (static_cast<size_t>(uSelectedSize) >= static_cast<size_t>(uRequestSize)) 
				{
					uRemainderSize = uSelectedSize - uRequestSize;
					UnlinkChunk(pxSelectedChunk);

					// Exhaust 
					if (uRemainderSize < uASURA_MEMHEAP_MINSIZE)  
					{
						SetInuseBitAtOffset(pxSelectedChunk, uSelectedSize);
						#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
						Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
						#endif
						MutexRelease();
						return GetMemoryFromChunk(pxSelectedChunk);
					}
					// Split
					else 
					{
						pxRemainderChunk = GetChunkAtOffset(pxSelectedChunk, uRequestSize);
						GetUnsortedChunks()->m_pxPrevChunk = GetUnsortedChunks()->m_pxNextChunk = pxRemainderChunk;
						pxRemainderChunk->m_pxPrevChunk = pxRemainderChunk->m_pxNextChunk = GetUnsortedChunks();
						SetHead(pxSelectedChunk		, uRequestSize   | uASURA_MEMHEAP_PREV_INUSE);
						SetHead(pxRemainderChunk	, uRemainderSize | uASURA_MEMHEAP_PREV_INUSE);
						SetFoot(pxRemainderChunk	, uRemainderSize);
						#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
						Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
						#endif	
						MutexRelease();
						return GetMemoryFromChunk(pxSelectedChunk);
					} 
				}
			}    
		}

		/*	Search for a chunk by scanning bins, starting with next largest
		bin. This search is strictly by best-fit; i.e., the smallest
		(with ties going to approximately the least recently used) chunk
		that fits is selected.
		The bitmap avoids needing to check that most blocks are nonempty.	*/

		++uBinIndex;
		pxBinChunk = GetBinAt(uBinIndex);
		uBlock = BlockFromIndex(uBinIndex);
		uMapWord = m_xMallocState.m_uBinmap[uBlock];
		uBit = BitFromIndex(uBinIndex);

		while(true)
		{
			// Skip rest of block if there are no more set bits in this block.
			if (uBit > uMapWord || uBit == 0) 
			{
				bool bFullBreak = false;
				do 
				{
					if (++uBlock >= uASURA_MEMHEAP_BINMAPSIZE)  // out of bins
					{
						bFullBreak = true;
						break;
					}
				} 
				while ( (uMapWord = m_xMallocState.m_uBinmap[uBlock]) == 0);

				if(bFullBreak)
				{
					break;
				}

				pxBinChunk = GetBinAt(uBlock << uASURA_MEMHEAP_BINMAPSHIFT);
				uBit = 1;
			}

			// Advance to bin with set bit. There must be one.
			while ((uBit & uMapWord) == 0) 
			{
				pxBinChunk = GetNextBin(pxBinChunk);
				uBit <<= 1;
				Asura_Assert(uBit != 0,"[Asura_Memheap Internal Checks] No bins left");
			}

			// Inspect the bin. It is likely to be non-empty
			pxSelectedChunk = pxBinChunk->m_pxPrevChunk;

			//  If a false alarm (empty bin), clear the bit.
			if (pxSelectedChunk == pxBinChunk) 
			{
				m_xMallocState.m_uBinmap[uBlock] = uMapWord &= ~uBit; // Write through
				pxBinChunk = GetNextBin(pxBinChunk);
				uBit <<= 1;
			}

			else 
			{
				uSelectedSize = ChunkSize(pxSelectedChunk);

				//  We know the first chunk in this bin is big enough to use.
				Asura_Assert(static_cast<size_t>(uSelectedSize) >= static_cast<size_t>(uRequestSize),"[Asura_Memheap Internal Checks] Chunk is too small");

				uRemainderSize = uSelectedSize - uRequestSize;

				// unlink
				pxPrev = pxSelectedChunk->m_pxPrevChunk;
				pxBinChunk->m_pxPrevChunk = pxPrev;
				pxPrev->m_pxNextChunk = pxBinChunk;

				// Exhaust
				if (uRemainderSize < uASURA_MEMHEAP_MINSIZE) 
				{
					SetInuseBitAtOffset(pxSelectedChunk, uSelectedSize);
					#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
					Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
					#endif
					MutexRelease();
					return GetMemoryFromChunk(pxSelectedChunk);
				}

				// Split
				else 
				{
					pxRemainderChunk = GetChunkAtOffset(pxSelectedChunk, uRequestSize);

					GetUnsortedChunks()->m_pxPrevChunk = GetUnsortedChunks()->m_pxNextChunk = pxRemainderChunk;
					pxRemainderChunk->m_pxPrevChunk = pxRemainderChunk->m_pxNextChunk = GetUnsortedChunks();
					// advertise as last remainder
					if (InSmallbinRange(uRequestSize)) 
					{
						m_xMallocState.m_xRemainderChunk = pxRemainderChunk; 
					}
					SetHead(pxSelectedChunk, uRequestSize | uASURA_MEMHEAP_PREV_INUSE);
					SetHead(pxRemainderChunk, uRemainderSize | uASURA_MEMHEAP_PREV_INUSE);
					SetFoot(pxRemainderChunk, uRemainderSize);
					#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
					Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
					#endif
					MutexRelease();
					return GetMemoryFromChunk(pxSelectedChunk);
				}
			}
		}
	}
	else
	{
		// initialization check
		if (m_xMallocState.m_uMaxFastSize == 0) 
		{
			MallocConsolidate();
		}		
	}

	/*use_top:    
	If large enough, split off the chunk bordering the end of memory
	(held in m_pxMallocState.top). Note that this is in accord with the best-fit
	search rule.  In effect, m_pxMallocState.top is treated as larger (and thus
	less well fitting) than any other available chunk since it can
	be extended to be as large as necessary (up to system
	limitations).

	We require that m_pxMallocState.top always exists (i.e., has size >=
	uASURA_MEMHEAP_MINSIZE) after initialization, so if it would otherwise be
	exhuasted by current request, it is replenished. (The main
	reason for ensuring it exists is that we may need uASURA_MEMHEAP_MINSIZE space
	to put in fenceposts in sysmalloc.)*/

	pxSelectedChunk = m_xMallocState.m_xTopChunk;
	uSelectedSize = ChunkSize(pxSelectedChunk);

	if (static_cast<size_t>(uSelectedSize) >= static_cast<size_t>(uRequestSize + uASURA_MEMHEAP_MINSIZE)) 
	{
		uRemainderSize = uSelectedSize - uRequestSize;
		pxRemainderChunk = GetChunkAtOffset(pxSelectedChunk, uRequestSize);
		m_xMallocState.m_xTopChunk = pxRemainderChunk;
		SetHead(pxSelectedChunk, uRequestSize | uASURA_MEMHEAP_PREV_INUSE);
		SetHead(pxRemainderChunk, uRemainderSize | uASURA_MEMHEAP_PREV_INUSE);

		#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
		Checks_MallocedChunk(pxSelectedChunk, uRequestSize);
		#endif
		MutexRelease();
		return GetMemoryFromChunk(pxSelectedChunk);
	}

	//If no space in top, relay to handle system-dependent cases 
    void* pvMem = SystemAlloc(uRequestSize);
	MutexRelease();
	return pvMem;
}

void Asura_MemHeap::MallocConsolidate()
{
	Asura_Memheap_Malloc_Chunk**	ppxFastbinChunk;		// current fastbin being consolidated 
	Asura_Memheap_Malloc_Chunk**	ppxLastFastbinChunk;	// last fastbin (for loop control) 
	Asura_Memheap_Malloc_Chunk* 	pxCurrentChunk;			// current chunk being consolidated 
	Asura_Memheap_Malloc_Chunk* 	pxPendingChunk;			// next chunk to consolidate 
	Asura_Memheap_Malloc_Chunk* 	pxUnsortedChunk;		// bin header 
	Asura_Memheap_Malloc_Chunk* 	pxFirstUnsortedChunk;	// chunk to link to 
	
	// These have same use as in Free()
	Asura_Memheap_Malloc_Chunk*		pxNextChunk;
	size_t							uSize;
	size_t							uNextSize;
	size_t							uPrevSize;
	int								iNextInUse;

	ClearFastChunks();

	pxUnsortedChunk = GetUnsortedChunks();

	/*Remove each chunk from fast bin and consolidate it, placing it
	then in unsorted bin. Among other reasons for doing this,
	placing in unsorted bin avoids needing to calculate actual bins
	until malloc is sure that chunks aren't immediately going to be
	reused anyway.*/

	ppxLastFastbinChunk = &(m_xMallocState.m_xFastbins[GetFastbinIndex(m_xMallocState.m_uMaxFastSize)]);
	ppxFastbinChunk = &(m_xMallocState.m_xFastbins[0]);
	do 
	{
		if ( (pxCurrentChunk = *ppxFastbinChunk) != 0) 
		{
			*ppxFastbinChunk = 0;

			do 
			{
				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_InUseChunk(pxCurrentChunk);
				#endif
				pxPendingChunk = pxCurrentChunk->m_pxNextChunk;

				// Slightly streamlined version of consolidation code in free()
				uSize = pxCurrentChunk->m_uSize & ~uASURA_MEMHEAP_PREV_INUSE;
				pxNextChunk = GetChunkAtOffset(pxCurrentChunk, uSize);
				uNextSize = ChunkSize(pxNextChunk);

				if (!PrevChunkInUse(pxCurrentChunk)) 
				{
					uPrevSize = pxCurrentChunk->m_uPrevSize;
					uSize += uPrevSize;
					pxCurrentChunk = GetChunkAtOffset(pxCurrentChunk, -(static_cast<int>(uPrevSize)));
					UnlinkChunk(pxCurrentChunk);
				}

				if (pxNextChunk != m_xMallocState.m_xTopChunk) 
				{
					iNextInUse = GetInuseBitAtOffset(pxNextChunk, uNextSize);
					SetHead(pxNextChunk, uNextSize);

					if (!iNextInUse) 
					{
						uSize += uNextSize;
						UnlinkChunk(pxNextChunk);
					}

					pxFirstUnsortedChunk = pxUnsortedChunk->m_pxNextChunk;
					pxUnsortedChunk->m_pxNextChunk = pxCurrentChunk;
					pxFirstUnsortedChunk->m_pxPrevChunk = pxCurrentChunk;

					SetHead(pxCurrentChunk, uSize | uASURA_MEMHEAP_PREV_INUSE);
					pxCurrentChunk->m_pxPrevChunk = pxUnsortedChunk;
					pxCurrentChunk->m_pxNextChunk = pxFirstUnsortedChunk;
					SetFoot(pxCurrentChunk, uSize);
				}
				else 
				{
					uSize += uNextSize;
					SetHead(pxCurrentChunk, uSize | uASURA_MEMHEAP_PREV_INUSE);
					m_xMallocState.m_xTopChunk = pxCurrentChunk;
				}
			} 
			while ( (pxCurrentChunk = pxPendingChunk) != 0);
		}
	} 
	while (ppxFastbinChunk++ != ppxLastFastbinChunk);
}

void Asura_MemHeap::MallocInitState()
{
	Asura_Memheap_Malloc_Chunk*	pBinChunk;
	
	// Establish circular links for normal bins
	for (u_int u = 1; u < uASURA_MEMHEAP_NBINS; ++u) 
	{ 
		pBinChunk = GetBinAt(u);
		pBinChunk->m_pxNextChunk = pBinChunk;
		pBinChunk->m_pxPrevChunk = pBinChunk;
	}

	m_xMallocState.m_uTopPad		= uASURA_MEMHEAP_DEFAULT_TOP_PAD;
	m_xMallocState.m_uTrimThresh	= uASURA_MEMHEAP_DEFAULT_TRIM_THRESHOLD;

	SetMaxFast(uASURA_MEMHEAP_DEFAULT_MXFAST);

	m_xMallocState.m_xTopChunk	= GetUnsortedChunks();
	m_xMallocState.m_uPagesize	= uASURA_MEMHEAP_PAGESIZE;
	m_xMallocState.m_iMaxPages	= -1;
}

void Asura_MemHeap::Free(void* pMem)
{
	Asura_Memheap_Malloc_Chunk*		pxMemChunk;			// chunk corresponding to mem 
	size_t							uMemChunkSize;		// its size 
	Asura_Memheap_Malloc_Chunk**	pxFastbinChunk;		// associated fastbin 
	Asura_Memheap_Malloc_Chunk*		pxPendingChunk;		// next contiguous chunk 
	size_t							uPendingChunkSize;	// its size 
	int								iPendingChunkUsed;	// true if nextchunk is used 
	size_t							uPrevChunkSize;		// size of previous contiguous chunk 
	Asura_Memheap_Malloc_Chunk*		pxPrevChunk;		// misc temp for linking 
	Asura_Memheap_Malloc_Chunk*		pxNextChunk;		// misc temp for linking 
	
	MutexAcquire();

	if (pMem != 0) 
	{
		m_bUpdateMallinfo = true;

		pxMemChunk = GetChunkFromMemory(pMem);
		uMemChunkSize = ChunkSize(pxMemChunk);

		#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
		Checks_InUseChunk(pxMemChunk);
		#endif	

		//If eligible, place chunk on a fastbin so it can be found and used quickly in malloc.

		if (static_cast<size_t>(uMemChunkSize) <= static_cast<size_t>(m_xMallocState.m_uMaxFastSize)) 
		{
			SetFastchunks();
			pxFastbinChunk = &(m_xMallocState.m_xFastbins[GetFastbinIndex(uMemChunkSize)]);
			pxMemChunk->m_pxNextChunk = *pxFastbinChunk;
			*pxFastbinChunk = pxMemChunk;
		}
		else // Consolidate other chunks as they arrive.
		{
			SetAnyChunks();

			pxPendingChunk = GetChunkAtOffset(pxMemChunk, uMemChunkSize);
			uPendingChunkSize = ChunkSize(pxPendingChunk);

			/* consolidate backward */
			if (!PrevChunkInUse(pxMemChunk)) 
			{
				uPrevChunkSize = pxMemChunk->m_uPrevSize;
				uMemChunkSize += uPrevChunkSize;
				pxMemChunk = GetChunkAtOffset(pxMemChunk, -static_cast<int>(uPrevChunkSize));
				UnlinkChunk(pxMemChunk);
			}

			if (pxPendingChunk != m_xMallocState.m_xTopChunk) 
			{
				/* get and clear inuse bit */
				iPendingChunkUsed = GetInuseBitAtOffset(pxPendingChunk, uPendingChunkSize);
				SetHead(pxPendingChunk, uPendingChunkSize);

				/* consolidate forward */
				if (!iPendingChunkUsed) 
				{
					UnlinkChunk(pxPendingChunk);
					uMemChunkSize += uPendingChunkSize;
				}

				/*
				Place the chunk in unsorted chunk list. Chunks are
				not placed into regular bins until after they have
				been given one chance to be used in malloc.
				*/

				pxPrevChunk = GetUnsortedChunks();
				pxNextChunk = pxPrevChunk->m_pxNextChunk;
				pxMemChunk->m_pxPrevChunk = pxPrevChunk;
				pxMemChunk->m_pxNextChunk = pxNextChunk;
				pxPrevChunk->m_pxNextChunk = pxMemChunk;
				pxNextChunk->m_pxPrevChunk = pxMemChunk;

				SetHead(pxMemChunk, uMemChunkSize | uASURA_MEMHEAP_PREV_INUSE);
				SetFoot(pxMemChunk, uMemChunkSize);
					
				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Check_FreeChunk(pxMemChunk);
				#endif
			}
			else  //If the chunk borders the current high end of memory, consolidate into top				
			{
				uMemChunkSize += uPendingChunkSize;
				SetHead(pxMemChunk, uMemChunkSize | uASURA_MEMHEAP_PREV_INUSE);
				m_xMallocState.m_xTopChunk = pxMemChunk;
				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_Chunk(pxMemChunk);
				#endif
			}

			/* If freeing a large space, consolidate possibly-surrounding
			chunks. Then, if the total unused topmost memory exceeds trim
			threshold, ask malloc_trim to reduce top.

			Unless max_fast is 0, we don't know if there are fastbins
			bordering top, so we cannot tell for sure whether threshold
			has been reached unless fastbins are consolidated.  But we
			don't want to consolidate on each free.  As a compromise,
			consolidation is performed if uASURA_MEMHEAP_FASTBIN_CONSOLIDATION_THRESHOLD
			is reached.	*/

			if (static_cast<size_t>(uMemChunkSize) >= uASURA_MEMHEAP_FASTBIN_CONSOLIDATION_THRESHOLD) 
			{ 
				if (HaveFastchunks()) 
				{
					MallocConsolidate();
				}

				//Trim if possible
				if(m_pfnTrimFunction)
				{
					if (static_cast<size_t>(ChunkSize(m_xMallocState.m_xTopChunk)) >= static_cast<size_t>(m_xMallocState.m_uTrimThresh))
					{
						SystemTrim(m_xMallocState.m_uTopPad);
					}
				}				
			}

		}		
	}
	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	//Checks_MallocState(); //TY 25.10.07: Stats may not have been updated if we came here from Alloc (bug I think)
	#endif
	MutexRelease();
}


void *Asura_MemHeap::Realloc ( void *pvMemory, size_t uBytes )
{
	size_t								uNumBytes;				// padded request size

	Asura_Memheap_Malloc_Chunk*			pxOldChunk;			// chunk corresponding to pvMemory
	size_t								uOldSize;		// its size

	Asura_Memheap_Malloc_Chunk*			pxNewChunk;			// chunk to return
	size_t								uNewSize;		// its size
	void*								pNewMem;			// corresponding user mem

	Asura_Memheap_Malloc_Chunk*			pxNextChunk;			// next contiguous chunk after oldp

	Asura_Memheap_Malloc_Chunk*			pxRemainderChunk;		// extra space at end of newp
	size_t								uRemainderSize;	// its size
	
	//We can optionally support freeing when zero bytes is passed here
	/*if (!uBytes) 
	{
		Free(pvMemory);
		return 0;
	}*/

	// realloc of null is supposed to be same as malloc
	if (pvMemory == 0)
	{
		return Alloc(uBytes);
	}

	uNumBytes = GetSizeFromRequest(uBytes);

	pxOldChunk    = GetChunkFromMemory(pvMemory);
	uOldSize = ChunkSize(pxOldChunk);

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_InUseChunk(pxOldChunk);
	#endif

	if (static_cast<size_t>(uOldSize) >= static_cast<size_t>(uNumBytes)) 
	{
		// already big enough; split below
		pxNewChunk = pxOldChunk;
		uNewSize = uOldSize;
	}
	else 
	{
		pxNextChunk = GetChunkAtOffset(pxOldChunk, uOldSize);

		// Try to expand forward into top 
		if (pxNextChunk == m_xMallocState.m_xTopChunk && static_cast<size_t>(uNewSize = uOldSize + ChunkSize(pxNextChunk)) >= static_cast<size_t>(uNumBytes + uASURA_MEMHEAP_MINSIZE)) // yuck
		{
			SetHeadSize(pxOldChunk, uNumBytes);
			m_xMallocState.m_xTopChunk = GetChunkAtOffset(pxOldChunk, uNumBytes);
			SetHead(m_xMallocState.m_xTopChunk, (uNewSize - uNumBytes) | uASURA_MEMHEAP_PREV_INUSE);
			return GetMemoryFromChunk(pxOldChunk);
		}
		else if (	pxNextChunk != m_xMallocState.m_xTopChunk 
				&&	!ChunkInUse(pxNextChunk) 
				&&	static_cast<size_t>(uNewSize = uOldSize + ChunkSize(pxNextChunk)) >= static_cast<size_t>(uNumBytes))
		{
			// Try to expand forward into next chunk, split off remainder below
			pxNewChunk = pxOldChunk;
			UnlinkChunk(pxNextChunk);
		}
		else 
		{
			// allocate, copy, free
			pNewMem = Alloc(uNumBytes - uASURA_MEMHEAP_MALLOC_ALIGN_MASK);
			if (pNewMem == 0)
			{
				return 0;
			}

			pxNewChunk = GetChunkFromMemory(pNewMem);
			uNewSize = ChunkSize(pxNewChunk);

			//Avoid copy if newp is next chunk after oldp.
			if (pxNewChunk == pxNextChunk) 
			{
				uNewSize += uOldSize;
				pxNewChunk = pxOldChunk;
			}
			else 
			{
				/*Unroll copy of <= 36 bytes (72 if 8byte sizes)
				We know that contents have an odd number of
				size_t-sized words; minimally 3.*/

				size_t uCopySize = uOldSize - sizeof(size_t);
				size_t* pxSource = static_cast<size_t*>(pvMemory);
				size_t* pxDest = static_cast<size_t*>(pNewMem);

				unsigned int uCopies = uCopySize / sizeof(size_t);
				Asura_Assert(uCopies >= 3, "Asura_Memheap: expected to copy more data than this in realloc");

				if (uCopies > 9)
				{
					Asura_Mem::Copy(pxDest, pxSource, uCopySize);
				}
				else 
				{
					*(pxDest+0) = *(pxSource+0);
					*(pxDest+1) = *(pxSource+1);
					*(pxDest+2) = *(pxSource+2);
					if (uCopies > 4)
					{
						*(pxDest+3) = *(pxSource+3);
						*(pxDest+4) = *(pxSource+4);
						if (uCopies > 6)
						{
							*(pxDest+5) = *(pxSource+5);
							*(pxDest+6) = *(pxSource+6);
							if (uCopies > 8)
							{
								*(pxDest+7) = *(pxSource+7);
								*(pxDest+8) = *(pxSource+8);
							}
						}
					}
				}

				Free(pvMemory);
				#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
				Checks_InUseChunk(pxNewChunk);
				#endif
				return GetMemoryFromChunk(pxNewChunk);
			}
		}
	}

	// If possible, free extra space in old or extended chunk
	Asura_Assert(static_cast<size_t>(uNewSize) >= static_cast<size_t>(uNumBytes),"Padded request size should be at least as big as requested memory size");

	uRemainderSize = uNewSize - uNumBytes;

	if (uRemainderSize < uASURA_MEMHEAP_MINSIZE) 
	{ 
		// not enough extra to split off
		SetHeadSize(pxNewChunk, uNewSize);
		SetInuseBitAtOffset(pxNewChunk, uNewSize);
	}
	else 
	{ 
		// split remainder
		pxRemainderChunk = GetChunkAtOffset(pxNewChunk, uNumBytes);
		SetHeadSize(pxNewChunk, uNumBytes);
		SetHead(pxRemainderChunk, uRemainderSize | uASURA_MEMHEAP_PREV_INUSE);
		// Mark remainder as inuse so free() won't complain
		SetInuseBitAtOffset(pxRemainderChunk, uRemainderSize);
		Free(GetMemoryFromChunk(pxRemainderChunk)); 
	}

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_InUseChunk(pxNewChunk);
	#endif
	return GetMemoryFromChunk(pxNewChunk);
}


void Asura_MemHeap::UpdateMallinfo() const
{
	if(!m_bUpdateMallinfo)
	{
		//If we dont need to update, dont
		return;
	}

	unsigned int u;
	Asura_Memheap_Malloc_Chunk* pxBaseChunk;
	Asura_Memheap_Malloc_Chunk* pxChunk;
	size_t uAvailable;
	size_t uFastAvailable;
	int iNumBlocks;
	int iNumFastBlocks;

	/* Ensure initialization */
	if( !m_xMallocState.m_xTopChunk ) 
	{
		return;
	}

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_MallocState();
	#endif

	// Account for top
	uAvailable = ChunkSize(m_xMallocState.m_xTopChunk);
	iNumBlocks = 1;  // top always exists

	// traverse fastbins
	iNumFastBlocks = 0;
	uFastAvailable = 0;

	for (u = 0; u < uASURA_MEMHEAP_NFASTBINS; ++u) 
	{
		for (pxChunk = m_xMallocState.m_xFastbins[u]; pxChunk != 0; pxChunk = pxChunk->m_pxNextChunk) 
		{
			++iNumFastBlocks;
			uFastAvailable += ChunkSize(pxChunk);
		}
	}

	uAvailable += uFastAvailable;

	// traverse regular bins 
	for (u = 1; u < uASURA_MEMHEAP_NBINS; ++u) 
	{
		pxBaseChunk = GetBinAt(u);
		for (pxChunk = pxBaseChunk->m_pxPrevChunk; pxChunk != pxBaseChunk; pxChunk = pxChunk->m_pxPrevChunk) 
		{
			++iNumBlocks;
			uAvailable += ChunkSize(pxChunk);
		}
	}

	m_xMallinfo.m_iFastbinBlocks		= iNumFastBlocks;
	m_xMallinfo.m_iFreeChunks			= iNumBlocks;
	m_xMallinfo.m_iTotalFree			= static_cast<int>(uAvailable);
	m_xMallinfo.m_iTotalAllocated		= static_cast<int>(m_xMallocState.m_uSystemAllocated - uAvailable);
	m_xMallinfo.m_iSystemAllocated		= static_cast<int>(m_xMallocState.m_uSystemAllocated);
	m_xMallinfo.m_iSpaceInFreeFast		= static_cast<int>(uFastAvailable);
	m_xMallinfo.m_iKeepCost			= static_cast<int>(ChunkSize(m_xMallocState.m_xTopChunk));
	m_xMallinfo.m_iMaxSystemAllocated	= static_cast<int>(m_xMallocState.m_uMaxSystemAllocated);

	m_bUpdateMallinfo = false;
}


Asura_MallocInfo Asura_MemHeap::GetMallocInfo( void ) const
{
    // Need to create a copy and return by value to ensure thread-safety.
    // (Returning a const reference to the internal m_xMallinfo means the
    // members could change while the caller is inspecting them.)
    Asura_MallocInfo xMallInfoCopy;

	MutexAcquire();

	UpdateMallinfo();
    xMallInfoCopy = m_xMallinfo;

	MutexRelease();

    return xMallInfoCopy;
}


void Asura_MemHeap::PrintMallocInfo () const
{	
	MutexAcquire();

	UpdateMallinfo();

	int iUsedPercent = 0;
	if(m_xMallinfo.m_iSystemAllocated && m_xMallinfo.m_iTotalAllocated)
	{
		iUsedPercent = static_cast<int>(100.0f / static_cast<float>(m_xMallinfo.m_iSystemAllocated) * static_cast<float>(m_xMallinfo.m_iTotalAllocated));
	}
		
	Asura_OutputToDebugger_Retail(("Asura_Memheap snapshot\n"));
	Asura_OutputToDebugger_Retail(("Free chunks: %d\n"										,m_xMallinfo.m_iFreeChunks));
	Asura_OutputToDebugger_Retail(("Fastbin blocks: %d\n"									,m_xMallinfo.m_iFastbinBlocks));
	Asura_OutputToDebugger_Retail(("Space availible in freed fastbin blocks:%d\n"			,m_xMallinfo.m_iSpaceInFreeFast));
	Asura_OutputToDebugger_Retail(("Total used space: %d of %d (%d percent)\n"				,m_xMallinfo.m_iTotalAllocated,m_xMallinfo.m_iSystemAllocated,iUsedPercent));
	Asura_OutputToDebugger_Retail(("available free space: %d\n"								,m_xMallinfo.m_iTotalFree));
	Asura_OutputToDebugger_Retail(("Topmost, releasable space: %d\n"						,m_xMallinfo.m_iKeepCost));	

	MutexRelease();
}

#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS

int	Asura_MemHeap::GetNumIntegrityFailures ()
{
	//Read and reset the number of integrity failures recorded (for unit tests)
	int iRet = m_iIntegrityFailures;
	m_iIntegrityFailures = 0;
	return iRet;
}

void Asura_MemHeap::Checks_InUseChunk(Asura_Memheap_Malloc_Chunk* pxChunk) const
{
	Asura_Memheap_Malloc_Chunk* pxNextChunk;
	Checks_Chunk(pxChunk);

	//Check whether it claims to be in use ... 
	Asura_Assert(ChunkInUse(pxChunk),"[Asura_Memheap Internal Checks] Chunk is in use");

	pxNextChunk = NextChunk(pxChunk);

	// ... and is surrounded by OK chunks.
	//Since more things can be checked with free chunks than inuse ones,
	//if an in use chunk borders them and debug is on, it's worth doing them.
	
	if (!PrevChunkInUse(pxChunk))  
	{
		// Note that we cannot even look at prev unless it is not inuse
		Asura_Memheap_Malloc_Chunk* prv = PrevChunk(pxChunk);
		Asura_Assert(NextChunk(prv) == pxChunk,"[Asura_Memheap Internal Checks] chunk pointers inconsistant");
		Check_FreeChunk(prv);
	}

	if (pxNextChunk == m_xMallocState.m_xTopChunk) 
	{
		Asura_Assert(PrevChunkInUse(pxNextChunk),"[Asura_Memheap Internal Checks] chunk pointers inconsistant");
		Asura_Assert(ChunkSize(pxNextChunk) >= uASURA_MEMHEAP_MINSIZE,"[Asura_Memheap Internal Checks] chunk too small");
	}
	else if (!ChunkInUse(pxNextChunk))
	{
		Check_FreeChunk(pxNextChunk);
	}
}

void Asura_MemHeap::Checks_MallocState () const
{
	unsigned int				u;
	Asura_Memheap_Malloc_Chunk*	pxChunk;
	Asura_Memheap_Malloc_Chunk*	pCheckChunk;
	Asura_Memheap_Malloc_Chunk*	pxNormalBinChunk;
	unsigned int				uBitBin;
	bool						bEmpty;
	unsigned int				uIndex;
	size_t						uChunkSize;
	size_t						uTotalUsage = 0;
	unsigned int				uMaxFastBin;

	// internal size_t must be no wider than pointer type 
	MEMHEAP_CHECK(sizeof(size_t) <= sizeof(char*),"[Asura_Memheap Internal Checks] internal size_t must be no wider than pointer type");
		
	// alignment is a power of 2
	MEMHEAP_CHECK((uASURA_MEMHEAP_MALLOC_ALIGNMENT & (uASURA_MEMHEAP_MALLOC_ALIGNMENT-1)) == 0,"[Asura_Memheap Internal Checks] alignment not power of two");

	/* cannot run remaining checks until fully initialized */
	if (m_xMallocState.m_xTopChunk == 0 || m_xMallocState.m_xTopChunk == GetUnsortedChunks())
	{
		return;
	}

	// pagesize is a power of 2
	MEMHEAP_CHECK((m_xMallocState.m_uPagesize & (m_xMallocState.m_uPagesize-1)) == 0,"[Asura_Memheap Internal Checks] Pagesize is not power of 2");

	// properties of fastbins

	// max_fast is in allowed range
	MEMHEAP_CHECK(GetMaxFastSize() <= GetSizeFromRequest(uASURA_MEMHEAP_MAX_FAST_SIZE),"[Asura_Memheap Internal Checks] maxfast is not in allowed range");

	uMaxFastBin = GetFastbinIndex(m_xMallocState.m_uMaxFastSize);

	for (u = 0; u < uASURA_MEMHEAP_NFASTBINS; ++u) 
	{
		pxChunk = m_xMallocState.m_xFastbins[u];

		// all bins past max_fast are empty
		if (u > uMaxFastBin)
		{
			MEMHEAP_CHECK(pxChunk == 0,"[Asura_Memheap Internal Checks] all bins past max_fast are empty");
		}

		while (pxChunk != 0) 
		{
			// each chunk claims to be inuse
			Checks_InUseChunk(pxChunk);
			uTotalUsage += ChunkSize(pxChunk);
			// chunk belongs in this bin
			MEMHEAP_CHECK(GetFastbinIndex(ChunkSize(pxChunk)) == u,"[Asura_Memheap Internal Checks] fastbin index mismatch");
			pxChunk = pxChunk->m_pxNextChunk;
		}
	}

	if (uTotalUsage != 0)
	{
		MEMHEAP_CHECK(HaveFastchunks(),"[Asura_Memheap Internal Checks] fastchunks dont match state");
	}
	else if (!HaveFastchunks())
	{
		MEMHEAP_CHECK(uTotalUsage == 0,"[Asura_Memheap Internal Checks] fastchunks dont match state");
	}

	// check normal bins 
	for (u = 1; u < uASURA_MEMHEAP_NBINS; ++u) 
	{
		pxNormalBinChunk = GetBinAt(u);

		// binmap is accurate (except for bin 1 == unsorted_chunks)
		if (u >= 2) 
		{
			uBitBin = GetBinmap(u);
			bEmpty = (pxNormalBinChunk->m_pxPrevChunk == pxNormalBinChunk);
			if (!uBitBin)
			{
				MEMHEAP_CHECK(bEmpty,"[Asura_Memheap Internal Checks] binmap records inconsistant");
			}
			else if (!bEmpty)
			{
				MEMHEAP_CHECK(uBitBin,"[Asura_Memheap Internal Checks] binmap records inconsistant");
			}
		}

		for (pxChunk = pxNormalBinChunk->m_pxPrevChunk; pxChunk != pxNormalBinChunk; pxChunk = pxChunk->m_pxPrevChunk) 
		{
			// each chunk claims to be free
			Check_FreeChunk(pxChunk);
			uChunkSize = ChunkSize(pxChunk);
			uTotalUsage += uChunkSize;
			if (u >= 2) 
			{
				// chunk belongs in bin
				uIndex = GetBinIndex(uChunkSize);
				MEMHEAP_CHECK(uIndex == u,"[Asura_Memheap Internal Checks] bin index incorrect");
				// lists are sorted
				if (static_cast<size_t>(uChunkSize) >= static_cast<size_t>(uASURA_MEMHEAP_MIN_LARGE_SIZE)) 
				{
					MEMHEAP_CHECK((pxChunk->m_pxPrevChunk == pxNormalBinChunk) || (static_cast<size_t>(ChunkSize(pxChunk->m_pxPrevChunk)) >= static_cast<size_t>(ChunkSize(pxChunk))),"[Asura_Memheap Internal Checks] bins not sorted correctly");
				}
			}
			// chunk is followed by a legal chain of inuse chunks
			for (pCheckChunk = NextChunk(pxChunk); (pCheckChunk != m_xMallocState.m_xTopChunk && ChunkInUse(pCheckChunk) && static_cast<size_t>(ChunkSize(pCheckChunk)) >= uASURA_MEMHEAP_MINSIZE);pCheckChunk = NextChunk(pCheckChunk))
			{
				Checks_InUseChunk(pCheckChunk);
			}
		}
	}

	// top chunk is OK
	Checks_Chunk(m_xMallocState.m_xTopChunk);

	// sanity checks for statistics
	MEMHEAP_CHECK(uTotalUsage <= static_cast<size_t>(m_xMallocState.m_uMaxSystemAllocated),								"[Asura_Memheap Internal Checks] total memory does not match expected");
	MEMHEAP_CHECK(static_cast<size_t>(m_xMallocState.m_uSystemAllocated) <= static_cast<size_t>(m_xMallocState.m_uMaxSystemAllocated),"[Asura_Memheap Internal Checks] assigned memory greater than maximum allowed");		
}


void Asura_MemHeap::Checks_RemallocedChunk(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize) const
{
	size_t uInternalSize = pChunk->m_uSize & ~uASURA_MEMHEAP_PREV_INUSE;

	Checks_InUseChunk(pChunk);

	// Legal size ...
	MEMHEAP_CHECK((uInternalSize & uASURA_MEMHEAP_MALLOC_ALIGN_MASK) == 0,"[Asura_Memheap Internal Checks] invalid chunk size");
	MEMHEAP_CHECK(static_cast<size_t>(uInternalSize) >= uASURA_MEMHEAP_MINSIZE,"[Asura_Memheap Internal Checks] chunk too small");
	// ... and alignment
	MEMHEAP_CHECK(IsAligned(GetMemoryFromChunk(pChunk)),"[Asura_Memheap Internal Checks] chunk not aligned correctly");	
	MEMHEAP_CHECK(static_cast<int>(uInternalSize) - static_cast<int>(uSize) >= 0, "[Asura_Memheap Internal Checks] chunk size mismatch");
	MEMHEAP_CHECK(static_cast<int>(uInternalSize) - static_cast<int>(uSize + uASURA_MEMHEAP_MINSIZE) < 0, "[Asura_Memheap Internal Checks] chunk size mismatch");
}


void Asura_MemHeap::Checks_MallocedChunk(Asura_Memheap_Malloc_Chunk* pChunk, size_t uSize) const
{
	// same as recycled case ... 
	Checks_RemallocedChunk(pChunk, uSize);

	/*... plus,  must obey implementation invariant that prev_inuse is
	always true of any allocated chunk; i.e., that each allocated
	chunk borders either a previously allocated and still in-use
	chunk, or the base of its memory arena. This is ensured
	by making all allocations from the the `lowest' part of any found
	chunk.  This does not necessarily hold however for chunks
	recycled via fastbins.	*/

	MEMHEAP_CHECK(PrevChunkInUse(pChunk),"[Asura_Memheap Internal Checks] chunk does not follow a used chunk");
}


void Asura_MemHeap::Check_FreeChunk(Asura_Memheap_Malloc_Chunk* pChunk) const
{
	size_t uSize = pChunk->m_uSize & ~uASURA_MEMHEAP_PREV_INUSE;
	Asura_Memheap_Malloc_Chunk* pxNextChunk = GetChunkAtOffset(pChunk, uSize);

	Checks_Chunk(pChunk);

	// Chunk must claim to be free ...
	MEMHEAP_CHECK(!ChunkInUse(pChunk),"[Asura_Memheap Internal Checks] free chunk thinks it is in use");	

	// Unless a special marker, must have OK fields 
	if ((size_t)(uSize) >= uASURA_MEMHEAP_MINSIZE)
	{
		MEMHEAP_CHECK((uSize & uASURA_MEMHEAP_MALLOC_ALIGN_MASK) == 0,"[Asura_Memheap Internal Checks] free chunk invalid size");
		MEMHEAP_CHECK(IsAligned(GetMemoryFromChunk(pChunk)),"[Asura_Memheap Internal Checks] free chunk not aligned");
		// ... matching footer field 
		MEMHEAP_CHECK(pxNextChunk->m_uPrevSize == uSize,"[Asura_Memheap Internal Checks] footer chunk data does not match");
		// ... and is fully consolidated 
		MEMHEAP_CHECK(PrevChunkInUse(pChunk),"[Asura_Memheap Internal Checks] free chunk does not follow a used chunk");
		MEMHEAP_CHECK (pxNextChunk == m_xMallocState.m_xTopChunk || ChunkInUse(pxNextChunk),"[Asura_Memheap Internal Checks] free chunk is not followed by anything");

		// ... and has minimally sane links 
		MEMHEAP_CHECK(pChunk->m_pxNextChunk->m_pxPrevChunk == pChunk, "[Asura_Memheap Internal Checks] chunk pointers invalid");
		MEMHEAP_CHECK(pChunk->m_pxPrevChunk->m_pxNextChunk == pChunk, "[Asura_Memheap Internal Checks] chunk pointers invalid");
	}
	else // markers are always of size size_t 
	{
		MEMHEAP_CHECK(uSize == sizeof(size_t),"[Asura_Memheap Internal Checks] marker is incorrect size");
	}
}

void Asura_MemHeap::Checks_Chunk(Asura_Memheap_Malloc_Chunk* pChunk) const
{
	size_t  uSize = ChunkSize(pChunk);
	// min and max possible addresses assuming contiguous allocation
	char* max_address = (char*)(m_xMallocState.m_xTopChunk) + ChunkSize(m_xMallocState.m_xTopChunk);
	char* min_address = max_address - m_xMallocState.m_uSystemAllocated;

	// Has legal address ... 
	if (pChunk != m_xMallocState.m_xTopChunk) 
	{
		if (IsContiguous()) 
		{
			MEMHEAP_CHECK(((char*)pChunk) >= min_address, "[Asura_Memheap Internal Checks] chunk address not valid (too low)");
			MEMHEAP_CHECK(((char*)pChunk + uSize) <= ((char*)(m_xMallocState.m_xTopChunk)), "[Asura_Memheap Internal Checks] chunk address not valid (too high)");
		}
	}
	else
	{
		// top size is always at least uASURA_MEMHEAP_MINSIZE
		MEMHEAP_CHECK(static_cast<size_t>(uSize) >= uASURA_MEMHEAP_MINSIZE, "[Asura_Memheap Internal Checks] top chunk too small");
		// top predecessor always marked inuse 
		MEMHEAP_CHECK(PrevChunkInUse(pChunk), "[Asura_Memheap Internal Checks] top chunk previous chunk should be marked as in use");
	}
}
#endif //#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS


void* Asura_MemHeap::SystemAlloc(size_t uBytes)
{
	//SystemAlloc is called when we need more internal memory for our heap, the name is slightly 
	//misleading as the allocation can potentially come from our parent heap if we are a subheap

	Asura_Memheap_Malloc_Chunk*		pOldTopChunk;		// incoming value of m_pxMallocState.top 
	size_t							uOldTopSize;		// its size 
	char*							pcOldEndAddress;	// its end address 

	size_t							uSize;				// arg to first Memory call 
	char*							pcMem;				// return value from memory allocator 

	size_t							uCorrection;		// arg to 2nd memory allocation call 
	char*							pcSecondMem;		// 2nd return val 

	size_t							uFrontMisalign;		// unusable bytes at front of new space 
	size_t							uEndMisalign;		// partial page left at end of new space 
	char*							pcAlignedMem;		// aligned offset into brk 

	Asura_Memheap_Malloc_Chunk*		pxChunk;			// the allocated/returned chunk 
	Asura_Memheap_Malloc_Chunk*		pxRemainderChunk;	// remainder from allocation 
	size_t							uRemainderSize;		// its size 

	size_t							uSum;				// for updating stats 

	size_t							uPageMask = static_cast<size_t>(m_xMallocState.m_uPagesize) - 1;

	/*If there is space available in fastbins, consolidate and retry
	malloc from scratch rather than getting memory from system.  This
	can occur only if nb is in smallbin range so we didn't consolidate
	upon entry to malloc. It is much easier to handle this case here
	than in malloc proper.*/

	if (HaveFastchunks()) 
	{
		Asura_Assert(InSmallbinRange(uBytes), "[Asura_Memheap Internal Checks] have fastchunks but not in smallbin range");
		MallocConsolidate();
		return Alloc(uBytes - uASURA_MEMHEAP_MALLOC_ALIGN_MASK);
	}

	// Record incoming configuration of top

	pOldTopChunk  = m_xMallocState.m_xTopChunk;
	uOldTopSize = ChunkSize(pOldTopChunk);
	pcOldEndAddress  = reinterpret_cast<char*>(GetChunkAtOffset(pOldTopChunk, uOldTopSize));

	pcMem		= NULL; 
	pcSecondMem = NULL;

	//If not the first time through, we require old_size to be
	//at least uASURA_MEMHEAP_MINSIZE and to have prev_inuse set.

	Asura_Assert((pOldTopChunk == GetUnsortedChunks() && uOldTopSize == 0) || (static_cast<size_t>(uOldTopSize) >= uASURA_MEMHEAP_MINSIZE && PrevChunkInUse(pOldTopChunk)),"[Asura_Memheap Internal Checks] Unexpected size");

	// Precondition: not enough current space to satisfy nb request
	Asura_Assert(static_cast<size_t>(uOldTopSize) < static_cast<size_t>(uBytes + uASURA_MEMHEAP_MINSIZE),"[Asura_Memheap Internal Checks] not enough current space to satisfy request");

	// Precondition: all fastbins are consolidated
	Asura_Assert(!HaveFastchunks(),"[Asura_Memheap Internal Checks] fastbins not consolidated");

	// Request enough space for nb + pad + overhead

	uSize = uBytes + m_xMallocState.m_uTopPad + uASURA_MEMHEAP_MINSIZE;

	/*If contiguous, we can subtract out existing space that we hope to
	combine with new space. We add it back later only if
	we don't actually get contiguous space.*/

	if (IsContiguous())
	{
		uSize -= uOldTopSize;
	}

	/*Round to a multiple of page size.	If our memory provider is not contiguous, this ensures that we only call it
	with whole-page arguments.  And if it is contiguous and	this is not first time through, this preserves page-alignment of
	previous calls. Otherwise, we correct to page-align below. On the other hand if we are using preassigned memory, just
	always 'allocate' that much all at once*/

	if(!m_pcBaseMem)
	{
		uSize = (uSize + uPageMask) & ~uPageMask;	
	}	
	else
	{
		uSize = m_uBaseMemSize;
	}

	pcMem = static_cast<char*>(MoreMemory(uSize));

	if (pcMem) 
	{
		m_xMallocState.m_uSystemAllocated += uSize;

		//If MoreMemory extends previous space, we can likewise extend top size.		

		if (pcMem == pcOldEndAddress && pcSecondMem == NULL) 
		{
			SetHead(pOldTopChunk, (uSize + uOldTopSize) | uASURA_MEMHEAP_PREV_INUSE);
		}
		else 
		{
			/*Otherwise, make adjustments:

			* If the first time through or noncontiguous, we need to call sbrk
			just to find out where the end of memory lies.

			* We need to ensure that all returned chunks from malloc will meet
			MALLOC_ALIGNMENT

			* If there was an intervening foreign sbrk, we need to adjust sbrk
			request size to account for fact that we will not be able to
			combine new space with existing space in old_top.

			* Almost all systems internally allocate whole pages at a time, in
			which case we might as well use the whole last page of request.
			So we allocate enough more memory to hit a page boundary now,
			which in turn causes future contiguous calls to page-align.	*/

			uFrontMisalign = 0;
			uEndMisalign = 0;
			uCorrection = 0;
			pcAlignedMem = pcMem;

			/*If MoreMemory returns an address lower than we have seen before,
			we know it isn't really contiguous.  This and some subsequent
			checks help cope with non-conforming MORECORE functions and
			the presence of "foreign" calls to MORECORE from outside of
			malloc or by other threads.  We cannot guarantee to detect
			these in all cases, but cope with the ones we do detect.*/
			if (IsContiguous() && uOldTopSize != 0 && pcMem < pcOldEndAddress) 
			{
				SetNoncontiguous();
			}

			/* handle contiguous cases */
			if (IsContiguous()) 
			{ 
				/* We can tolerate forward non-contiguities here (usually due
				to foreign calls) but treat them as part of our space for
				stats reporting.*/
				if (uOldTopSize != 0) 
				{
					m_xMallocState.m_uSystemAllocated += reinterpret_cast<uintptr_t>(pcMem) - reinterpret_cast<uintptr_t>(pcOldEndAddress);
				}

				/* Guarantee alignment of first new chunk made from this space */

				uFrontMisalign = reinterpret_cast<uintptr_t>(GetMemoryFromChunk(pcMem)) & uASURA_MEMHEAP_MALLOC_ALIGN_MASK;
				if (uFrontMisalign > 0) 
				{
					/*Skip over some bytes to arrive at an aligned position.
					We don't need to specially mark these wasted front bytes.
					They will never be accessed anyway because
					prev_inuse of m_pxMallocState.top (and any chunk created from its start)
					is always true after initialization.*/

					uCorrection = uASURA_MEMHEAP_MALLOC_ALIGNMENT - uFrontMisalign;
					pcAlignedMem += uCorrection;
				}

				/*
				If this isn't adjacent to existing space, then we will not
				be able to merge with old_top space, so must add to 2nd request.
				*/

				uCorrection += uOldTopSize;

				/* Extend the end address to hit a page boundary */
				uEndMisalign = reinterpret_cast<uintptr_t>(pcMem) + uSize + uCorrection;
				uCorrection += ((uEndMisalign + uPageMask) & ~uPageMask) - uEndMisalign;

				pcSecondMem = (char*)MoreMemory(uCorrection);

				if (pcSecondMem == NULL) 
				{
					//If can't allocate correction, try to at least find out current
					//end of mem.  It might be enough to proceed without failing.
					uCorrection = 0;
					pcSecondMem = static_cast<char*>(MoreMemory(0));
					// (MH - 02.01.2003) attempt to get the [stats] correct
					if (pcSecondMem != NULL)
					{
						m_xMallocState.m_uSystemAllocated += reinterpret_cast<uintptr_t>(pcSecondMem) - reinterpret_cast<uintptr_t>(pcMem) - uSize;
					}
				}
				else if (pcSecondMem < pcMem) 
				{
					/*If the second call gives noncontiguous space even though
					it says it won't, the only course of action is to ignore
					results of second call, and conservatively estimate where
					the first call left us. Also set noncontiguous, so this
					won't happen again, leaving at most one hole.

					Note that this check is intrinsically incomplete.  Because
					MORECORE is allowed to give more space than we ask for,
					there is no reliable way to detect a noncontiguity
					producing a forward gap for the second call.*/
					pcSecondMem = pcMem + uSize;
					uCorrection = 0;
					SetNoncontiguous();
				}

			}			
			else 
			{
                // Deal with alignment in same way as contiguous case.
				uFrontMisalign = reinterpret_cast<uintptr_t>(GetMemoryFromChunk(pcMem)) & uASURA_MEMHEAP_MALLOC_ALIGN_MASK;
				if( uFrontMisalign ) 
				{
					pcAlignedMem += (uASURA_MEMHEAP_MALLOC_ALIGNMENT - uFrontMisalign);
				}

				// Find out current end of memory
				if (pcSecondMem == NULL) 
				{
					pcSecondMem = static_cast<char*>(MoreMemory(0));
					m_xMallocState.m_uSystemAllocated += reinterpret_cast<uintptr_t>(pcSecondMem) - reinterpret_cast<uintptr_t>(pcMem) - uSize;
				}
			}

			// Adjust top based on results of second sbrk
			if (pcSecondMem != NULL) 
			{
				m_xMallocState.m_xTopChunk = reinterpret_cast<Asura_Memheap_Malloc_Chunk*>(pcAlignedMem);
				SetHead(m_xMallocState.m_xTopChunk, (reinterpret_cast<uintptr_t>(pcSecondMem) - reinterpret_cast<uintptr_t>(pcAlignedMem) + uCorrection) | uASURA_MEMHEAP_PREV_INUSE);
				m_xMallocState.m_uSystemAllocated += uCorrection;

				/*If not the first time through, we either have a
				gap due to foreign sbrk or a non-contiguous region.  Insert a
				double fencepost at old_top to prevent consolidation with space
				we don't own. These fenceposts are artificial chunks that are
				marked as inuse and are in any case too small to use.  We need
				two to make sizes and alignments work out.*/

				if (uOldTopSize != 0) 
				{
					/* Shrink old_top to insert fenceposts, keeping size a
					multiple of uASURA_MEMHEAP_MALLOC_ALIGNMENT. We know there is at least
					enough space in old_top to do this.	*/
					uOldTopSize = (uOldTopSize - 3*sizeof(size_t)) & ~uASURA_MEMHEAP_MALLOC_ALIGN_MASK;
					SetHead(pOldTopChunk, uOldTopSize | uASURA_MEMHEAP_PREV_INUSE);

					/*Note that the following assignments completely overwrite
					old_top when old_size was previously uASURA_MEMHEAP_MINSIZE.  This is
					intentional. We need the fencepost, even if old_top otherwise gets
					lost.*/
					GetChunkAtOffset(pOldTopChunk, uOldTopSize)->m_uSize =	sizeof(size_t)|uASURA_MEMHEAP_PREV_INUSE;

					GetChunkAtOffset(pOldTopChunk, uOldTopSize + sizeof(size_t))->m_uSize = sizeof(size_t)|uASURA_MEMHEAP_PREV_INUSE;

					//If possible, release the rest, suppressing trimming.
					if (uOldTopSize >= uASURA_MEMHEAP_MINSIZE) 
					{
						size_t uTemp = m_xMallocState.m_uTrimThresh;
						m_xMallocState.m_uTrimThresh = (size_t)(-1);
						Free(GetMemoryFromChunk(pOldTopChunk));
						m_xMallocState.m_uTrimThresh = uTemp;
					}
				}
			}
		}

		// Update statistics 
		uSum = m_xMallocState.m_uSystemAllocated;
		if (uSum > static_cast<size_t>(m_xMallocState.m_uMaxSystemAllocated))
		{
			m_xMallocState.m_uMaxSystemAllocated = uSum;
		}

		#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
		Checks_MallocState();
		#endif

		// finally, do the allocation

		pxChunk = m_xMallocState.m_xTopChunk;
		uSize = ChunkSize(pxChunk);

		// check that one of the above allocation paths succeeded
		if (static_cast<size_t>(uSize) >= static_cast<size_t>(uBytes + uASURA_MEMHEAP_MINSIZE)) 
		{
			uRemainderSize = uSize - uBytes;
			pxRemainderChunk = GetChunkAtOffset(pxChunk, uBytes);
			m_xMallocState.m_xTopChunk = pxRemainderChunk;
			SetHead(pxChunk, uBytes | uASURA_MEMHEAP_PREV_INUSE);
			SetHead(pxRemainderChunk, uRemainderSize | uASURA_MEMHEAP_PREV_INUSE);
			#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
			Checks_MallocedChunk(pxChunk, uBytes);
			#endif
			return GetMemoryFromChunk(pxChunk);
		}

	}

	// catch all failure paths
	Asura_Assert(false,"Cannot allocate more memory for memheap");
	return 0;
}


int Asura_MemHeap::SystemTrim(size_t pad)
{
	//Give memory back to the system
	int  iTopSize		= 0;	// Amount of top-most memory
	int  iExtra			= 0;	// Amount to release
	int  iReleased		= 0;	// Amount actually released
	char* pcCurrentMem	= 0;	// address returned by pre-check mem call
	char* pcNewMem		= 0;	// address returned by post-check mem call

	m_xMallocState.m_uPagesize = m_xMallocState.m_uPagesize;
	iTopSize = static_cast<int>(ChunkSize(m_xMallocState.m_xTopChunk));

	/* Release in pagesize units, keeping at least one page */
	iExtra = static_cast<int>(((iTopSize - pad - uASURA_MEMHEAP_MINSIZE + (m_xMallocState.m_uPagesize-1)) / m_xMallocState.m_uPagesize - 1) * m_xMallocState.m_uPagesize);

	if (iExtra > 0) 
	{
		//Only proceed if end of memory is where we last set it.This avoids problems if there were foreign sbrk calls.
		//TY - This check is somewhat redundant now that the end of memory is store locally, it provides some sanity checking though
		pcCurrentMem = static_cast<char*>(MoreMemory(0));
		if (pcCurrentMem == reinterpret_cast<char*>(m_xMallocState.m_xTopChunk) + iTopSize) 
		{
			//TY - It used to be the case here that dlm would trim, ignore the return value and call
			//the allocator function with zero to get the end of memory. This existed mainly to cater
			//for the situation where less memory was trimmed than requested, however I do not think
			//it is too much to ask that the trim function trims as much as requested.
			//Trim
			if(m_pfnTrimFunction)
			{
				//trim (returns the new end of memory)
				pcNewMem = static_cast<char*>(m_pfnTrimFunction(iExtra));
			}
			if (pcNewMem != 0) 
			{
				//update our end of memory
				m_xMallocState.m_pcEndOfMem = pcNewMem;
				iReleased = static_cast<int>(pcCurrentMem - pcNewMem);

				if (iReleased != 0) 
				{
					/* Success. Adjust top. */
					m_xMallocState.m_uSystemAllocated -= iReleased;
					SetHead(m_xMallocState.m_xTopChunk, (iTopSize - iReleased) | uASURA_MEMHEAP_PREV_INUSE);
					#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
					Checks_MallocState();
					#endif
					return 1;
				}
			}
		}
	}
	return 0;
}


void* Asura_MemHeap::MoreMemory (size_t uSize)
{
	char *pcAlloc = 0;
	size_t uAllocSize = uSize;
	
	if(!uSize)
	{
		return m_xMallocState.m_pcEndOfMem;
	}
	//check if we have reached, or are going to reach, the maximum allowed memory for this heap
	if(!m_pcBaseMem && uSize && (m_xMallocState.m_iMaxPages != -1) &&
	  (m_xMallocState.m_uSystemAllocated + uSize > m_xMallocState.m_iMaxPages*m_xMallocState.m_uPagesize) )
	{
		//over the limit
		Asura_Assert(false,"Memheap allocation limit reached");
		return 0;
	}
	if(m_pcBaseMem)
	{
		//In this case we were constructed from a blob of memory so just return it
		if(m_xMallinfo.m_uSystemMemCalls)
		{
			//If we have already allocated memory we cannot allow any more
			pcAlloc = 0;
		}
		else
		{
			pcAlloc = m_pcBaseMem;
		}		
	}
  	else if(m_pxParent)
  	{
  		//In this case we are a subheap, so request memory from our parent rather than using new
  		pcAlloc = static_cast<char*>(m_pxParent->Alloc(uAllocSize));
  	}
	else if(m_pfnAllocFunction)
	{
		//if we have been supplied with a memory allocation function, use it
		pcAlloc = static_cast<char*>(m_pfnAllocFunction(uAllocSize));
	}
	else
	{
		pcAlloc = static_cast<char*>(DefaultMemoryAllocator(uAllocSize));
	}
	if(pcAlloc)
	{
		++ m_xMallinfo.m_uSystemMemCalls;
		m_xMallocState.m_pcEndOfMem = pcAlloc + uSize;
		m_bUpdateMallinfo = true;
	}
	return pcAlloc;
}


void* Asura_MemHeap::DefaultMemoryAllocator	(size_t uSize)
{
	void* pMem = Asura_Allocate(uSize);
	if(pMem)
	{
		m_xDefaultAllocs.AddToList( static_cast<char*>(pMem) );
	}
	return pMem;
}


void Asura_MemHeap::Wipe()
{
	Asura_Assert(m_pcBaseMem,"Cannot wipe a dynamic memory heap");
	//Note: This is for fixed memory heaps only, otherwise the memory that has been allocated
	//      could be all over the place and it is not sensible to try to consolidate it.

	//Save the old settings

	MutexAcquire();

	size_t	uTrimThresh = m_xMallocState.m_uTrimThresh;
	size_t	uTopPad = m_xMallocState.m_uTopPad;
	unsigned int uPagesize = m_xMallocState.m_uPagesize;
	int	iMaxPages = m_xMallocState.m_iMaxPages;

	Asura_Mem::Set(&m_xMallocState,0,sizeof(Asura_Memheap_Malloc_State));
	Asura_Mem::Set(&m_xMallinfo,0,sizeof(Asura_MallocInfo));

	m_bUpdateMallinfo = true;

	m_xMallocState.m_uTrimThresh = uTrimThresh;
	m_xMallocState.m_uTopPad = uTopPad;
	m_xMallocState.m_uPagesize = uPagesize;
	m_xMallocState.m_iMaxPages = iMaxPages;

	Asura_Memheap_Malloc_Chunk*	pBinChunk;

	// Establish circular links for normal bins
	for (u_int u = 1; u < uASURA_MEMHEAP_NBINS; ++u) 
	{ 
		pBinChunk = GetBinAt(u);
		pBinChunk->m_pxNextChunk = pBinChunk;
		pBinChunk->m_pxPrevChunk = pBinChunk;
	}
	SetMaxFast(uASURA_MEMHEAP_DEFAULT_MXFAST);

	m_xMallocState.m_xTopChunk	= GetUnsortedChunks();

	#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
	Checks_MallocState();
	#endif
	MutexRelease();
}


void Asura_MemHeap::EnableThreadSafety ()
{
	m_bThreadSafetyEnabled = true;
}

void Asura_MemHeap::DisableThreadSafety ()
{
	m_bThreadSafetyEnabled = false;
}

void Asura_MemHeap::DisableThreadSafetyWarning()
{
	m_bThreadSafetyWarningDisabled = true;
}

// E N D   O F   F I L E /////////////////////////////////////////////////////
