/**
*
* \file Asura_MemoryManagement.cpp
*
* Created by Kevin Lea - 23/10/2001 10:48:13
*
* Copyright (C) 2001 Rebellion, All Rights Reserved.
*
* \brief
*
**/

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED

// This file's header.
#include "Core/Asura_MemoryManagement.h"

// Stdlib independent.
#include <stdlib.h>
#if defined ASURA_PC_SSE
#include <xmmintrin.h>
#endif

// Engine independent.
#if defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ALL) || defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ANONYMOUS)
#include "Core/Asura_CallStack.h"
#endif
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_MemoryManagement_Enums.h"
#include "Core/Asura_NameTree.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_MemoryBank.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Core/Asura_MemTracker.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// D E F I N E S   E T C . ///////////////////////////////////////////////////

static const Asura_Char* MEMORYLOGFILE = "Logfiles\\Memory.txt";
static const Asura_Char* szMEMORY_LEAK_FILE = "Memleaks.txt";
static const Asura_Char* MALLOC_DUMP_DEFAULT_PATH = "Logfiles\\";
static const Asura_Char* MALLOC_INFO_DUMP_DEFAULT_FILE = "MallocInfo.txt";
static const Asura_Char* MALLOC_SUMMARY_DUMP_DEFAULT_FILE = "MallocSummary.txt";

static const char* szUNKNOWN_FILE = "<unknown>";

#define ASURA_USE_MEMPOOL_FOR_MEMORY_NODES 1

#ifdef ASURA_MEMORYMANAGEMENT_FILL
//#define ASURA_MEMORYMANAGEMENT_GUARDS 1 // turn on checking of memory before and/or after allowed block (e.g. array overwrite)

// ASURA_MEMORYMANAGEMENT_GUARDS_START is the number of bytes added to the start of an allocated 
// block which should remain untouched. It should match the default malloc alignment (16 is using 
// vector stuff like SSE (PC) or VFPU (PSP)) or can be 4 on non-vectorised platforms. 
// 1 byte is always used at the end of a block. This catches off-by-one errors and has minimal impact on memory usage.
#ifdef ASURA_MEMORYMANAGEMENT_GUARDS
//#define ASURA_MEMORYMANAGEMENT_GUARDS_START 16
#ifndef ASURA_MEMORYMANAGEMENT_GUARDS_START
#define ASURA_MEMORYMANAGEMENT_GUARDS_START 0
#endif
#endif // #ifdef ASURA_MEMORYMANAGEMENT_GUARDS
#endif // #ifdef ASURA_MEMORYMANAGEMENT_FILL

#define ASURA_MM_HASH_TABLE_SIZE		1024
#define ASURA_NUM_MEMORY_NODES_PER_POOL	8192

#if defined ASURA_PC_SSE
//	#define ASURA_MM_GET_MEM(N)			(_mm_malloc((N),sizeof (__m128)))
void* ASURA_MM_GET_MEM( size_t N) 
{
		void* pMemory = _mm_malloc((N),sizeof (__m128));
		Asura_Assertf( NULL != pMemory, ("Cannot allocate %d bytes of memory", N));
		return( pMemory);
}
	#define ASURA_MM_FREE_MEM(P)		(_mm_free(P))
#else
	#define ASURA_MM_GET_MEM(N)			(malloc(N))
	#define ASURA_MM_FREE_MEM(P)		(free(P))
#endif

// C L A S S E S /////////////////////////////////////////////////////////////
class Asura_MemoryNode
{
	public:

#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
		void CheckGuardBytes() const
		{
			// JAT 15/06/07 added #if line to stop a compiler warning
#if ASURA_MEMORYMANAGEMENT_GUARDS_START && !(ASURA_MEMORYMANAGEMENT_GUARDS_START & 3)
			const u_int* puActualStart = static_cast<u_int*>(this->ActualAddress);
#endif // #if ASURA_MEMORYMANAGEMENT_GUARDS_START == 4 || ASURA_MEMORYMANAGEMENT_GUARDS_START == 16 
			const u_char* pucActualStart = static_cast<u_char*>(this->ActualAddress);
			const u_char* pucReportedStart = static_cast<u_char*>(this->GetReportedAddress());
			const u_char* pucStartOfEndGuard = pucReportedStart + this->GetReportedSize();
			u_int uNumEndGuardBytes = (u_int)( (pucActualStart + this->ActualSize) - pucStartOfEndGuard );		// PAZ - fix for PC.x64 build ** POINTER ARITHMETIC ** (safe)

#if (ASURA_MEMORYMANAGEMENT_GUARDS_START == 4)
			{
				Asura_Assert(*puActualStart == Asura_MemoryManagement_System::uiFillValueAllocate, "guard bytes overwritten at start");
			}
#elif ASURA_MEMORYMANAGEMENT_GUARDS_START && !(ASURA_MEMORYMANAGEMENT_GUARDS_START & 3)
			{
				for( int i = 0; i < (ASURA_MEMORYMANAGEMENT_GUARDS_START >> 2); i++ )
				{
					if (*puActualStart != Asura_MemoryManagement_System::uiFillValueAllocate)
					{
						Asura_Assert(false, "guard bytes overwritten at start");
						break;
					}
					puActualStart++;
				}
			}
#elif (ASURA_MEMORYMANAGEMENT_GUARDS_START != 0)
#error "Unhandled number of guard bytes!";
#endif

			const u_char* pucEndGuard = pucStartOfEndGuard;
			for (u_int u = 0; u < uNumEndGuardBytes; ++u)
			{
				if (*pucEndGuard != Asura_MemoryManagement_System::ucFillValueAllocate)
				{
					Asura_Assert(false, "guard bytes overwritten at end");
					break;
				}
			}
		}
#endif

		// accessors for the bitfield members and computed values
		inline void* GetReportedAddress() const
		{
			size_t uReportedAddress = reinterpret_cast<size_t>(ActualAddress);
#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL) && ASURA_MEMORYMANAGEMENT_GUARDS_START
			uReportedAddress += ASURA_MEMORYMANAGEMENT_GUARDS_START;
#endif			
			if (uAlignShift)
			{
				u_int uAlignment = (1 << uAlignShift) & ~1; // 0 shift means 0 extra alignment bytes, so AND with ~1 to convert 1->0.
				u_int uOffset = (uAlignment - static_cast<u_int>(uReportedAddress) & (uAlignment-1));
				uReportedAddress += uOffset;
			}
			return reinterpret_cast<void*>(uReportedAddress);
		}
		
		inline size_t GetReportedSize() const
		{
			u_int uExtraBytes = (1 << uAlignShift) & ~1; // 0 shift means 0 extra alignment bytes, so AND with ~1 to convert 1->0.
#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
			uExtraBytes += ASURA_MEMORYMANAGEMENT_GUARDS_START + 1;// need (default alignment) bytes at start and one byte at the end
#endif
			return ActualSize - uExtraBytes;
		}

		inline void SetLine(int iLine)
		{
			// clamp to range 0-32767
			uLine = Asura_Maths::Min(iLine, (1<<15)-1);
		}

		inline void SetFrameNum(u_int uFrameNumber)
		{
			// clamp to range 0-32767
			uFrameNum = Asura_Maths::Min(uFrameNumber, static_cast<u_int>((1<<12)-1));
		}

		void*					ActualAddress;
		size_t					ActualSize;
		const char*             pFilename;
								
		u_int					uAlignShift:3;			// if non-zero, alignment = 1<<uAlignShift (0-128)
		u_int					uAllocationType:2;		// 0-3 (2 bits) - never store delete events
		u_int                   uLine:15;				// 0-32767 (15 bits)
		u_int                   uFrameNum:12;			// 12 bits available (0-4095)

#if defined ASURA_MEMORYMANAGEMENT_USE_RETURN_ADDRESS
        u_int                   uReturnAddr;
#endif
#if defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ALL) || defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ANONYMOUS)
        void*                   pvStackDump;
#endif

		Asura_MemoryNode*		NextNode;
};

// (MH - 17.03.2008) 
// Asura_MemoryNodeMempool is copy & pasted from Asura_ResizingMempool. 
// While I dislike doing this, KJFL wanted the memory management
// version of the class to be insulated from future changes.
// It's also easier to manage making the changes to allow custom allocators
// to be used, i.e. replacing new/delete with Asura_MemoryUtility calls.
class Asura_MemoryNodeMempool_Node 
{
	friend class Asura_MemoryNodeMempool;

private:

	static Asura_MemoryNodeMempool_Node* Create(u_int uNumberOfElementsPerPool, u_int uSizeOfElementInBytes);
	// removed ctor/dtor. Put ctor code into Create() function and replaced dtor with Destroy() function
	static void Destroy(Asura_MemoryNodeMempool_Node* pxNode);

	void* Allocate(void)
	{
		if (m_pucNextFreeElement==NULL) return NULL;

		void* p = (void*)m_pucNextFreeElement;

		m_pucNextFreeElement = *((unsigned char**)p);

		m_uNumberOfAllocatedBlocks++;

		return p;
	}


	void Deallocate(void* pElement) 
	{
		*((unsigned char**)(pElement)) = m_pucNextFreeElement;
		m_pucNextFreeElement = (unsigned char*)pElement;

		m_uNumberOfAllocatedBlocks--;
	}

	void Reset(u_int uNumberOfElementsPerPool, u_int uSizeOfElementInBytes)
	{
		m_pucNextFreeElement = m_pucMemoryPool;

		for (u_int i=0; i<uNumberOfElementsPerPool-1; i++)
		{
			unsigned char* node = &m_pucMemoryPool[i*uSizeOfElementInBytes];
			unsigned char* nextNode = &m_pucMemoryPool[(i+1)*uSizeOfElementInBytes];

			*((unsigned char**)node) = nextNode;
		}

		*((unsigned char**)(&m_pucMemoryPool[(uNumberOfElementsPerPool-1)*uSizeOfElementInBytes])) = 0;

	}


	unsigned char*					m_pucMemoryPool;
	unsigned char*					m_pucNextFreeElement;
	u_int							m_uNumberOfAllocatedBlocks;
	Asura_MemoryNodeMempool_Node*	m_pxNextNode;
};

Asura_MemoryNodeMempool_Node* Asura_MemoryNodeMempool_Node::Create(	u_int uNumberOfElementsPerPool, 
																	  u_int uSizeOfElementInBytes)
{
	Asura_MemoryNodeMempool_Node* pxNode = static_cast<Asura_MemoryNodeMempool_Node*>
		(Asura_MemoryUtility::Allocate(sizeof(Asura_MemoryNodeMempool_Node)));

	if (pxNode == NULL)
	{
		return NULL;
	}

	// ctor used to do these 4 lines:
	pxNode->m_uNumberOfAllocatedBlocks = 0;
	pxNode->m_pucMemoryPool = 0;
	pxNode->m_pxNextNode = 0;
	pxNode->m_pucNextFreeElement = 0;

	pxNode->m_pucMemoryPool = static_cast<unsigned char*>
		(Asura_MemoryUtility::Allocate(uNumberOfElementsPerPool*uSizeOfElementInBytes));

	if (pxNode->m_pucMemoryPool == NULL)
	{
		Asura_MemoryUtility::Deallocate(pxNode);

		return NULL;
	}

	pxNode->Reset(uNumberOfElementsPerPool,uSizeOfElementInBytes);

	return pxNode;
}

// static destroy function
void Asura_MemoryNodeMempool_Node::Destroy(Asura_MemoryNodeMempool_Node* pxNode)
{
	// destructor used to do this.

	if (!pxNode)
	{
		return;
	}

	// free the memory pool
	Asura_MemoryUtility::Deallocate(pxNode->m_pucMemoryPool);

	// recursively delete next node & its mempool
	Destroy(pxNode->m_pxNextNode);

	// free the node memory
	Asura_MemoryUtility::Deallocate(pxNode);
}

class Asura_MemoryNodeMempool
{
public:
	// replaced ctor/dtor with Init/Shutdown
	void Initialise( u_int uNumberOfElementsPerPool, u_int uSizeOfElementInBytes )
	{
		m_pxHeadOfPoolList = NULL;

		m_uNumberOfElementsPerPool = uNumberOfElementsPerPool;
		m_uSizeOfElementInBytes = uSizeOfElementInBytes;

		Asura_Assert( (uNumberOfElementsPerPool * uSizeOfElementInBytes) < static_cast<u_int>(~0), "This is a very large memory pool.");
	}

	void Shutdown()
	{
		Asura_MemoryNodeMempool_Node::Destroy(m_pxHeadOfPoolList);
	}

	void* Allocate(void)
	{
		/// Kevin Lea - 11/09/2002 12:28:52 - we have four possible situations
		//
		// 1) Nothing has been allocated before (m_pxHeadOfPoolList == NULL) and
		//    so we allocated a new pool and an element from that.
		//
		// 2) The pool pointed to by m_pxHeadOfPoolList has space and we allocate
		//    an element from that.
		//
		// 3) The pool pointed to by m_pxHeadOfPoolList has no space; we carry on
		//    down to the next pool looking for a free element to use.
		//
		// 4) We find we have no space at all, and so we allocate a new pool and 
		//    place it at the head of the list (which means next time we call this
		//    function we at least point to a pool with lots of free space).

		Asura_MemoryNodeMempool_Node* pxNode = m_pxHeadOfPoolList;

		/// Kevin Lea - 11/09/2002 15:14:55 - check for cases 2) and 3)
		while(pxNode)
		{
			void* pMemory = pxNode->Allocate();

			/// Kevin Lea - 11/09/2002 15:12:39 - did we find a free element?
			if (pMemory)
			{
				return pMemory;
			}

			/// Kevin Lea - 11/09/2002 15:14:42 - try next pool then
			pxNode = pxNode->m_pxNextNode;
		}

		/// Kevin Lea - 11/09/2002 15:15:10 - okay then, new block it is
		if (AddANewPool())
		{
			return m_pxHeadOfPoolList->Allocate();
		}
		else /// Kevin Lea - 11/09/2002 12:44:53 - operation failed
		{
			return 0;
		}
	}

	void Deallocate(void *pElement)
	{
		Asura_MemoryNodeMempool_Node* pxNode = m_pxHeadOfPoolList;
		Asura_MemoryNodeMempool_Node** ppxPrevious = &m_pxHeadOfPoolList;

		u_int uPoolSize = m_uSizeOfElementInBytes*m_uNumberOfElementsPerPool;

		while(pxNode)
		{
			if ( (pElement >= pxNode->m_pucMemoryPool) && (pElement < (pxNode->m_pucMemoryPool + uPoolSize)) )
			{
#if defined(_DEBUG) && defined(ASURA_ASSERT)

				int iElementIndex = static_cast<int>(((unsigned char*)pElement-pxNode->m_pucMemoryPool)/m_uSizeOfElementInBytes);
				Asura_Assert( (iElementIndex*m_uSizeOfElementInBytes + pxNode->m_pucMemoryPool) == ((unsigned char*)pElement),"Pointer not on element boundary!");

#endif

				pxNode->Deallocate(pElement);

				if (pxNode->m_uNumberOfAllocatedBlocks == 0)
				{
					*ppxPrevious = pxNode->m_pxNextNode;

					pxNode->m_pxNextNode = NULL;

					Asura_MemoryNodeMempool_Node::Destroy(pxNode);
				}
				return;
			}

			ppxPrevious = &(pxNode->m_pxNextNode);
			pxNode = pxNode->m_pxNextNode;
		}

		Asura_Assert(false,"Element not found in memory pool!");
	}

	void Reset(void)
	{
		if(!m_pxHeadOfPoolList) return;

		/// Richard May - 30/04/2004 12:59:41 - Resetting a non-empty pool is now illegal. It's the users
		/// responsibilty to clear up their allocations rather than relying on the pool to do it for them
		Asura_Assert( !m_pxHeadOfPoolList->m_uNumberOfAllocatedBlocks, "Resetting a non-empty pool! Must clear up allocations before reset!" );
		if( m_pxHeadOfPoolList->m_uNumberOfAllocatedBlocks ) return;

		Asura_MemoryNodeMempool_Node::Destroy(m_pxHeadOfPoolList);
		m_pxHeadOfPoolList = 0;
	}

	// KJFL - 24 January 2007 - a "force deallocate" method like the original Reset() used to be. 
	// This will delete all memory allocated by the pool in a rapid manner, but remember any pointers 
	// returned by the pool will now be invalid. So, only use if you really know what you're doing.
	// (If you know you want to wipe the slate clean, zeroing any pointers/lists you may have and
	// calling this method is obviously a lot faster than going through all your pointers/lists and 
	// calling deallocate per element.)
	void ForceReset( void )
	{
		if (m_pxHeadOfPoolList)
		{
			// each node automatically deletes the next one, no need to recurse ourselves
			Asura_MemoryNodeMempool_Node::Destroy(m_pxHeadOfPoolList);
			m_pxHeadOfPoolList = 0;
		}
	}

	bool IsEmpty( void ) const
	{
		return( NULL == m_pxHeadOfPoolList );
	}

	// Returns whether this pool allocated this memory.
	bool AllocatedThisMemory( const void* const pMemory ) const
	{
		const u_int uPoolSize = m_uSizeOfElementInBytes * m_uNumberOfElementsPerPool;

		Asura_MemoryNodeMempool_Node* pxNode = m_pxHeadOfPoolList;
		while ( pxNode )
		{
			if ( ( pxNode->m_pucMemoryPool <= pMemory ) && ( pMemory < ( pxNode->m_pucMemoryPool + uPoolSize ) ) )
			{
				const u_int uElementIndex = static_cast< u_int>(( static_cast< const unsigned char* const >( pMemory ) - pxNode->m_pucMemoryPool ) / m_uSizeOfElementInBytes);
				return ( uElementIndex < m_uNumberOfElementsPerPool );
			}

			pxNode = pxNode->m_pxNextNode;
		}

		return false;
	}

private:

	bool AddANewPool()
	{
		Asura_MemoryNodeMempool_Node* pxNode = Asura_MemoryNodeMempool_Node::Create(m_uNumberOfElementsPerPool, m_uSizeOfElementInBytes);

		if (pxNode == NULL)
		{
			/// Kevin Lea - 11/09/2002 12:41:55 - uh-oh, we're out of memory
			Asura_Assert(false, "Out of memory!");

			return false;
		}

		/// Kevin Lea - 11/09/2002 12:43:50 - pop the new pool onto the head of the list
		pxNode->m_pxNextNode = m_pxHeadOfPoolList;
		m_pxHeadOfPoolList = pxNode;

		return true;
	}

	Asura_MemoryNodeMempool_Node*	m_pxHeadOfPoolList;

	u_int						m_uNumberOfElementsPerPool;
	u_int						m_uSizeOfElementInBytes;
};


const int iASURA_MM_STATE_UNINITIALISED = -1;
const u_int uASURA_MM_STATE_UNINITIALISED = 0xffffffff;

// KOR - 27/01/2006 - store certain properties of the core,
// so that we can compare against previous occasions when we expect
// the memory environment to be the same (e.g. after core reset, before 
// level load, after loading a particular environment, etc.)
// See declaration of Asura_MemoryManagement_System::Checkpoint in Asura_MemoryManagement.h
struct Asura_MemoryManagement_Core_State
{
	u_int	m_uTotalMemoryCurrentlyAlloced;
	int		m_iNumberOfAllocations;
};

class Asura_MemoryManagement_Core
{
	public:
		void Initialise(void);
		void Shutdown(void);

        void DumpMallocInfo(const char* szFilenameNoPath = NULL, const char* szPath = NULL, u_int uFlags = ASURA_MALLOCDUMPFLAG_DEFAULTS);
		void DumpMallocSummary(const char* szFilenameNoPath = NULL, const char* szPath = NULL);
		void* Allocate( size_t uNumBytes, ASURA_MM_EVENT_ID eEventID, u_int uAlignment, const char* szFile, int iLine, u_int uReturnAddr );
		void Deallocate( void* pAddr, ASURA_MM_EVENT_ID eEventID, const char* szFile, int iLine);

#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
        void CheckMallocGuardBytes();
#endif
		// KOR - 30/01/2006 - see declaration of Asura_MemoryManagement_System::Checkpoint in Asura_MemoryManagement.h
		void Checkpoint(int iState);
		u_int GetTotalCurrentlyAlloced() const { return m_uTotalMemoryCurrentlyAlloced; }

    private:
		const char* GetExtraNodeInfo(const Asura_MemoryNode* pxNode);

		Asura_MemoryNode* AcquireMemoryNode(void);
		void ReleaseMemoryNode(Asura_MemoryNode* node, Asura_MemoryNode* pxPrevNode);	
		void InsertMemoryNode(Asura_MemoryNode* node);
		Asura_MemoryNode* FindMemoryNode(void* address, Asura_MemoryNode** ppxPrevNode);
		void FreeMemoryNodeMem(Asura_MemoryNode* node); // unlike ReleaseMemoryNode, this doesn't fix up the hash table/linked lists.

		u_int GetHashIndex( void* pAddr ) const
        {
			// This used to be "reinterpret_cast<u_int>" but because PS3 pointers are 64-bit
			// and u_int is 32-bit you get a compiler error. So I've changed it to cast to 
			// "size_t" (64-bit on the PS3) and then cast this down to a "u_int" - ART.
        	return( (static_cast<u_int>(reinterpret_cast<size_t>(pAddr)) >> 4) & (ASURA_MM_HASH_TABLE_SIZE - 1) );
        }

		Asura_MemoryNode* HashTable[ASURA_MM_HASH_TABLE_SIZE];

#ifdef ASURA_USE_MEMPOOL_FOR_MEMORY_NODES
		Asura_MemoryNodeMempool* m_pxMemoryNodePool;
#endif
		int NumberOfActiveMemoryNodes;

		int EventCount_New;
		int EventCount_Delete;
		int EventCount_NewArray;
		int EventCount_DeleteArray;
		int EventCount_NewAlign;
		int EventCount_DeleteAlign;
		int EventCount_Alloc;
		int EventCount_Dealloc;

		u_int m_uTotalMemoryAllocedOverall;
		u_int m_uTotalMemoryCurrentlyAlloced;
		u_int m_uPeakMemoryUsage;

		int m_iTotalMemoryAllocations;
		int m_iCurrentMemoryAllocations;
		int m_iPeakMemoryAllocations;

		Asura_MemoryManagement_Core_State m_axPreviousStates[ASURA_MM_STATE_NUM_TYPES];

		#ifdef ASURA_ENABLE_MEMTRACKER
		Asura_MemTracker* m_pxMemTracker;
		#endif

    public:
        
        // if bEnableMemoryManager is false malloc is used instead of allocating 
        // nodes (this gets used during shutdown, and during malloc dump)
        bool bEnableMemoryManager;      
};

// G L O B A L S /////////////////////////////////////////////////////////////

static Asura_MemoryManagement_Core* MMCore=NULL;

DEBUGVAR bool  dbg_bLogAllMemAllocs        = false;
DEBUGVAR bool  dbg_bLogAllMemDeallocs      = false;
DEBUGVAR u_int dbg_uMinAllocSizeToReport   = 0;
DEBUGVAR u_int dbg_uMinDeallocSizeToReport = 0;

#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
static bool g_bCheckMallocGuardBytesOnce = false;
static bool g_bCheckMallocGuardBytesAlways = false;
#endif

static bool g_bClearedMemoryLogfile = false;

// (MP - 17.09.2002) This function solves the problem of trying to clear the
// memory logfile too early; the memory manager obviously gets initialised
// before almost anything else and, in particular, before the PS2 version
// sets up the directory where it should be outputting the logfile.
static void ClearMemoryLogfileIfNecessary( void )
{
    if( !g_bClearedMemoryLogfile && Asura_Logfile_CanLogYet() )
    {
        Asura_SimpleLogfileClear( MEMORYLOGFILE );
        g_bClearedMemoryLogfile = true;
    }
}

// G L O B A L   F U N C T I O N S ///////////////////////////////////////////

// static void ReleaseMemoryManager(void);
void ReleaseMemoryManager(void)
{	
	if (MMCore)
	{
		MMCore->Shutdown();
		ASURA_MM_FREE_MEM( MMCore );
		MMCore = NULL;
		Asura_MemoryManagement_System::Platform_Shutdown();
	}
	
}

// (MP - 09.08.2006) Non-inline versions of the new and delete operators so
// that external libraries can link with them if required. Without this the
// compiler may never generate versions of all of these functions (if we
// don't call it within our own source). Previously we used to declare these
// as inline in the header but this will often stop the compiler emitting
// an externally linkable function, so that external library code, etc., will
// end up linking with the standard implementations from libc or wherever.
// The performance difference should be negligible - typically the internal
// malloc work is expensive anyway, and of course memory management is turned
// off in most non-debug / non-tools builds

#include "Core/Asura_MemoryManagement_DebugOff.h"

void* operator new( size_t size ) 
{
    return Asura_MemoryManagement_System::Allocate( size, ASURA_NEW );
}
void* operator new( size_t size, const std::nothrow_t& ) throw()
{
    return Asura_MemoryManagement_System::Allocate( size, ASURA_NEW );
}
void* operator new( size_t size, const char* szFile, int iLine, u_int uReturnAddr )
{
    return Asura_MemoryManagement_System::Allocate( size, ASURA_NEW, 0, szFile, iLine, uReturnAddr );
}

void* operator new[]( size_t size )
{
    return Asura_MemoryManagement_System::Allocate( size, ASURA_NEWARRAY );
} 
void* operator new[]( size_t size, const std::nothrow_t& ) throw()
{
    return Asura_MemoryManagement_System::Allocate( size, ASURA_NEWARRAY );
} 
void* operator new[]( size_t size, const char* szFile, int iLine, u_int uReturnAddr )
{
    return Asura_MemoryManagement_System::Allocate( size, ASURA_NEWARRAY, 0, szFile, iLine, uReturnAddr );
}

void operator delete( void* pvMemory )
{
    Asura_MemoryManagement_System::Deallocate( pvMemory, ASURA_DELETE );
}
void operator delete( void* pvMemory, const std::nothrow_t& ) throw()
{
    Asura_MemoryManagement_System::Deallocate( pvMemory, ASURA_DELETE );
}
void operator delete( void* pvMemory, const char* szFile, int iLine, u_int uReturnAddr )
{
    Asura_MemoryManagement_System::Deallocate( pvMemory, ASURA_DELETE, szFile, iLine );
}

void operator delete[]( void* pvMemory )
{   
    Asura_MemoryManagement_System::Deallocate( pvMemory, ASURA_DELETEARRAY );
}
void operator delete[]( void* pvMemory, const std::nothrow_t& ) throw()
{
    Asura_MemoryManagement_System::Deallocate( pvMemory, ASURA_DELETEARRAY );
}
void operator delete[]( void* pvMemory, const char* szFile, int iLine, u_int uReturnAddr )
{   
    Asura_MemoryManagement_System::Deallocate( pvMemory, ASURA_DELETEARRAY, szFile, iLine );
}

#include "Core/Asura_MemoryManagement_DebugOn.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#ifdef ASURA_MEMORYMANAGEMENT_FILL  
// these are the same as Microsoft's
const u_char Asura_MemoryManagement_System::ucFillValueAllocate = 0xCD;
const u_int  Asura_MemoryManagement_System::uiFillValueAllocate = 0xCDCDCDCD;
const u_char Asura_MemoryManagement_System::ucFillValueFree     = 0xDD;
const u_int  Asura_MemoryManagement_System::uiFillValueFree     = 0xDDDDDDDD;
#endif

#ifdef ASURA_MEMORYBANK_ALLOW_STATS
static Asura_MemoryBank_Stats g_axStats[ASURA_MAX_RESOURCE_TYPES];
bool Asura_MemoryManagement_System::s_bEnableStatLogging = false;
bool Asura_MemoryManagement_System::s_bLogUnspecifiedAllocations = false;
bool Asura_MemoryManagement_System::s_bRenderTypesWithZeroAllocations = false;
#endif

bool Asura_MemoryManagement_System::s_bMultithreadingEnabled = true;
bool Asura_MemoryManagement_System::s_bMemoryLeak = false;
CheckpointCallback* Asura_MemoryManagement_System::s_pfnCheckpointCallback = NULL;

static bool g_bReportMemoryLeaksOnShutdown = true;
static bool g_bIgnoreNewAndDeleteTypeMismatch = false;
static u_int g_uMaxMemToAllocate = 0;

// constructor for initialiser objects: see comment in header file at Asura_MMS_InitialiserObject
Asura_MemoryManagement_System::Asura_MemoryManagement_System()
{
	if (MMCore != NULL)
	{
		/// Kevin Lea - 23/10/2001 16:35:28 - another file has already initialised
		/// the system; this is expected behaviour
		return;
	}
	Platform_Initialise();

	MMCore = (Asura_MemoryManagement_Core*)ASURA_MM_GET_MEM( sizeof(Asura_MemoryManagement_Core) );

	MMCore->Initialise();

	atexit( ReleaseMemoryManager );  
}

void* Asura_MemoryManagement_System::Allocate( size_t size, ASURA_MM_EVENT_ID eEventID, u_int uAlignment, const char* szFile, int iLine, u_int uReturnAddr )
{
#ifdef ASURA_MEMORYBANK_ALLOW_STATS
	if( s_bEnableStatLogging )
	{
		const u_int uResourceType = Asura_ResourceAllocator::GetCurrentResourceType();
		if( (uResourceType != ASURA_RESOURCE_UNSPECIFIED) || s_bLogUnspecifiedAllocations )
		{
			g_axStats[uResourceType].AddAllocation( (u_int)size );		// PAZ - fix for PC.x64 build
		}		
	}
#endif

	if (MMCore && MMCore->bEnableMemoryManager)
	{
		Platform_Lock();
        void* pMem = MMCore->Allocate( size, eEventID, uAlignment, szFile, iLine, uReturnAddr );
		Platform_Unlock();
		return pMem;
	}
	else
	{
		return ASURA_MM_GET_MEM( size );
	}
}

void Asura_MemoryManagement_System::Deallocate( void* pAddr, ASURA_MM_EVENT_ID eEventID, const char* szFile, int iLine )
{
	if (MMCore && MMCore->bEnableMemoryManager)
	{
		Platform_Lock();
		MMCore->Deallocate( pAddr, eEventID, szFile, iLine );
        Platform_Unlock();
	}
	else
	{
		ASURA_MM_FREE_MEM( pAddr );
	}
}

#if defined(ASURA_MEMORYMANAGEMENT_FILL)
void Asura_MemoryManagement_System::CheckMallocGuardBytes(void)
{
#if defined(ASURA_MEMORYMANAGEMENT_GUARDS)
	MMCore->CheckMallocGuardBytes();
#endif	//defined(ASURA_MEMORYMANAGEMENT_GUARDS)
}
#endif	//defined(ASURA_MEMORYMANAGEMENT_FILL)


#if defined(ASURA_MEMORYBANK_ALLOW_STATS)

void Asura_MemoryManagement_System::ResetStatLogging( void )
{
	for( u_int u=0; u<ASURA_MAX_RESOURCE_TYPES; ++u )
	{
		g_axStats[u].Reset();
	}
}

void Asura_MemoryManagement_System::DebugRenderStats( const Asura_MemoryBank_StatsRender& xRender )
{
	const u_int uSizeOfTempText = 256;
	char szTempTextBuffer[uSizeOfTempText];

	// RDM - 15/12/06 - This code has been borrowed from the memory banks that had a concept of maximum size and 
	// an easily accessible total allocated memory. Unfortunately the main memory management doesn't (and can't
	// as far as the PC is concerned) have a concept of size. Ultimately it might be possible to add these as 
	// platform functions but for now I'm leaving this stuff out.
	//const float	fUsedMb			= 0.0f;//static_cast<float>(static_cast<int>(m_uCurrentSize)) / (1024.0f * 1024.0f);
	//const float	fTotalMb		= 1.0f;//static_cast<float>(static_cast<int>(m_uSizeOfData)) / (1024.0f * 1024.0f);
	//const float	fPercentageUsed	= (fUsedMb / fTotalMb) * 100.0f;

	// Render the title...
	xRender.RenderTitle( Asura_MemoryBank_StatsRender::NAME_COLUMN, "Main Memory" );
	xRender.RenderTitle( Asura_MemoryBank_StatsRender::NUM_ALLOCATIONS_COLUMN, "Allocs");
	xRender.RenderTitle( Asura_MemoryBank_StatsRender::SIZE_COLUMN, "Mb" );

// RDM - 15/12/06 - See comment above
	u_int uTotalBytesAllocated = 0;
	for( u_int u=0; u<ASURA_MAX_RESOURCE_TYPES; u++ )
	{
		if( (u==ASURA_RESOURCE_UNSPECIFIED) && !s_bLogUnspecifiedAllocations ) continue;

		uTotalBytesAllocated += g_axStats[u].GetNumBytesAllocated();
	}		

	// Do we have any stats?
	u_int uRow = 0;
	const char** pszStatsNames = Asura_ResourceAllocator::GetResourceNameArray();

	// Yes - output the specific stats...
	for( u_int u=0; u<ASURA_MAX_RESOURCE_TYPES; u++ )
	{	
		if( !s_bLogUnspecifiedAllocations && (u == ASURA_RESOURCE_UNSPECIFIED) ) continue;
		if( (g_axStats[u].GetNumBytesAllocated() == 0) && !s_bRenderTypesWithZeroAllocations ) continue;

		g_axStats[u].SetName(pszStatsNames[u]);
		g_axStats[u].DebugRender( xRender, uRow );
		uRow++;
	}

	// Render the summary...
	const float fTotalMb = static_cast<float>( static_cast<int>(uTotalBytesAllocated) ) / (1024.0f * 1024.0f);
	Asura_PrintfToBfr( szTempTextBuffer, "Total Tracked Allocations: %.3fMb. Note this is NOT TOTAL USED MEMORY!", uSizeOfTempText, true, Asura_Maths::FtoD(fTotalMb) );
	//Asura_PrintfToBfr( szTempTextBuffer, "Used %.3fMb of %.3fMb (%.3f%%)", uSizeOfTempText, true, Asura_Maths::FtoD(fUsedMb), Asura_Maths::FtoD(fTotalMb), Asura_Maths::FtoD(fPercentageUsed) );
	xRender.RenderRow( (uRow + 1), szTempTextBuffer );
}

#endif	// defined(ASURA_MEMORYBANK_ALLOW_STATS)

void Asura_MemoryManagement_System::SetIgnoreNewAndDeleteTypeMismatch(bool bIgnore)
{
	g_bIgnoreNewAndDeleteTypeMismatch = bIgnore;
}

// SetMaxMemToAllocate()/GetMaxMemToAllocate() - set/get max memory that will be allocated
// via Asura_MemoryManagement_System. This allows consoles to restrict memory usage when 
// running in devkit modes which have extra memory.
void Asura_MemoryManagement_System::SetMaxMemToAllocate( u_int uMaxBytes )
{
	g_uMaxMemToAllocate = uMaxBytes;
}
u_int Asura_MemoryManagement_System::GetMaxMemToAllocate()
{
	return g_uMaxMemToAllocate;
}
// GetTotalMemAllocated() = how much memory is currently outstanding.
u_int Asura_MemoryManagement_System::GetTotalMemAllocated()
{
	if (MMCore && MMCore->bEnableMemoryManager)
	{
		return MMCore->GetTotalCurrentlyAlloced();
	}
	return 0;
}

// (MH - 24.09.2004) added functions which may be externed and called from .c files (e.g. in middleware).
extern "C" {
	void* Asura_C_Allocate( size_t uNumBytes, u_int uAlignment, const char* szFile, int iLine, u_int uReturnAddr )
	{
		return Asura_MemoryManagement_System::Allocate(uNumBytes, ASURA_ALLOC, uAlignment, szFile, iLine, uReturnAddr);
	}
	void Asura_C_Deallocate( void* pAddr, const char* szFile, int iLine )
	{
		Asura_MemoryManagement_System::Deallocate(pAddr, ASURA_DEALLOC, szFile, iLine);
	}
};

// KOR - 30/01/2006 - see declaration in Asura_MemoryManagement.h
void Asura_MemoryManagement_System::Checkpoint(int iState)
{
	if (MMCore && MMCore->bEnableMemoryManager)
	{
		Platform_Lock();

		MMCore->Checkpoint(iState);

		if(s_pfnCheckpointCallback)
		{
			s_pfnCheckpointCallback(iState);
		}

		Platform_Unlock();
	}	
}

void Asura_MemoryManagement_System::DumpMallocInfo(const char* szFilenameNoPath /* = NULL */, const char* szPath /* = NULL */, u_int uFlags /* = ASURA_MALLOCDUMPFLAG_DEFAULTS */)
{
	if (MMCore && MMCore->bEnableMemoryManager)
	{
		Platform_Lock();

        MMCore->DumpMallocInfo(szFilenameNoPath, szPath, uFlags);

        Platform_Unlock();
	}
}

void Asura_MemoryManagement_System::DumpMallocSummary(const char* szFilenameNoPath, const char* szPath)
{
	if (MMCore && MMCore->bEnableMemoryManager)
	{
		Platform_Lock();

		MMCore->DumpMallocSummary(szFilenameNoPath, szPath);

		Platform_Unlock();
	}
}

#ifndef ASURA_MALLOCDUMP_BUFFER_SIZE
#define ASURA_MALLOCDUMP_BUFFER_SIZE (1024 * 1024)
#endif

static const char*	g_szMallocDumpFilename = NULL;
static bool			g_bMallocDumpFileNeedsClearing = false;
static char			g_szMallocDumpOutputBfr[ASURA_MALLOCDUMP_BUFFER_SIZE];
static u_int		g_uMallocDumpOutputBfrLenUsed = 0;

void Asura_MemoryManagement_System::BeginMallocDump( const char* szFilename )
{
	g_szMallocDumpFilename         = szFilename;
	g_bMallocDumpFileNeedsClearing = true;
	g_uMallocDumpOutputBfrLenUsed  = 0;
}

void Asura_MemoryManagement_System::AddMallocDumpLine( const char* szLine, u_int uLen )
{
	// Flush?
	if( g_uMallocDumpOutputBfrLenUsed + uLen >= ASURA_MALLOCDUMP_BUFFER_SIZE )
	{
		FlushMallocDump();
	}

	const u_int uCharsToCopy = Asura_Maths::Min( uLen + 1, ASURA_MALLOCDUMP_BUFFER_SIZE - g_uMallocDumpOutputBfrLenUsed );
	Asura_Mem::Copy( &g_szMallocDumpOutputBfr[g_uMallocDumpOutputBfrLenUsed], szLine, uCharsToCopy );
	g_uMallocDumpOutputBfrLenUsed += (uCharsToCopy - 1);
}

void Asura_MemoryManagement_System::EndMallocDump( void )
{
	FlushMallocDump();
}

void Asura_MemoryManagement_System::FlushMallocDump()
{
	if( g_bMallocDumpFileNeedsClearing )
	{
		Asura_SimpleLogfileClear( g_szMallocDumpFilename );
		g_bMallocDumpFileNeedsClearing = false;
	}

	if( g_uMallocDumpOutputBfrLenUsed )
	{
		Asura_SimpleLogfileOutput_PreFormatted( g_szMallocDumpFilename, g_szMallocDumpOutputBfr, g_uMallocDumpOutputBfrLenUsed );
		g_uMallocDumpOutputBfrLenUsed = 0;
	}
}


void Asura_MemoryManagement_System::SetReportMemoryLeaksOnShutdown(bool bEnable)
{
	g_bReportMemoryLeaksOnShutdown = bEnable;
}

void Asura_MemoryManagement_Core::Initialise(void)
{
	EventCount_New = 0;
	EventCount_Delete = 0;
	EventCount_NewArray = 0;
	EventCount_DeleteArray = 0;
	EventCount_NewAlign = 0;
	EventCount_DeleteAlign = 0;
	EventCount_Alloc = 0;
	EventCount_Dealloc = 0;

	m_uTotalMemoryAllocedOverall = 0;
	m_uTotalMemoryCurrentlyAlloced = 0;
	m_uPeakMemoryUsage = 0;

	m_iTotalMemoryAllocations	= 0;
	m_iCurrentMemoryAllocations = 0;
	m_iPeakMemoryAllocations	= 0;

	NumberOfActiveMemoryNodes = 0;

	for (int i=0; i<ASURA_MM_HASH_TABLE_SIZE; ++i)
	{
		HashTable[i] = NULL;
	}

	for (int iState = 0; iState < ASURA_MM_STATE_NUM_TYPES; ++iState)
	{
		m_axPreviousStates[iState].m_iNumberOfAllocations = iASURA_MM_STATE_UNINITIALISED;
		m_axPreviousStates[iState].m_uTotalMemoryCurrentlyAlloced = uASURA_MM_STATE_UNINITIALISED;
	}

	Asura_MemoryUtility::Initialise();

	// init the node mempool
#ifdef ASURA_USE_MEMPOOL_FOR_MEMORY_NODES
	m_pxMemoryNodePool = static_cast<Asura_MemoryNodeMempool*>(Asura_MemoryUtility::Allocate(sizeof(Asura_MemoryNodeMempool)));
	m_pxMemoryNodePool->Initialise(ASURA_NUM_MEMORY_NODES_PER_POOL, sizeof(Asura_MemoryNode));
#endif

	bEnableMemoryManager = true;

    g_bClearedMemoryLogfile = false;

#ifdef ASURA_ENABLE_MEMTRACKER
	m_pxMemTracker = NULL; // This needs setting to NULL because Asura_MemTracker::GetInstance called 'new' internally, which calls our memtracking stuff
	m_pxMemTracker = Asura_MemTracker::GetInstance ();

	m_pxMemTracker->SetMemoryMin (Asura_MemoryManagement_System::Platform_GetAddressSpaceStart());
	m_pxMemTracker->SetMemoryMax (Asura_MemoryManagement_System::Platform_GetAddressSpaceEnd());
#endif
}

void Asura_MemoryManagement_Core::Shutdown(void)
{
    // MH note: you won't be able to use logfiles in this function since 
    // they will have been closed down. This also precludes Asserts since
    // they write to a logfile.
	bEnableMemoryManager = false;

    // Check for mem leaks.
    bool bAnyMemLeaks = false;
    for( int i = 0; i < ASURA_MM_HASH_TABLE_SIZE; ++i )
    {
        if( HashTable[i] )
        {
            bAnyMemLeaks = true;
            break;
        }
    }

    // Report them if required.
    if( bAnyMemLeaks && g_bReportMemoryLeaksOnShutdown )
    {
        DumpMallocInfo( szMEMORY_LEAK_FILE, NULL, (ASURA_MALLOCDUMPFLAG_DEFAULTS | ASURA_MALLOCDUMPFLAG_CALLSTACK | ASURA_MALLOCDUMPFLAG_REPORTEDADDR | ASURA_MALLOCDUMPFLAG_ISLEAKREPORT) );
    }

    // And then deal with them.
    for( int i = 0; i < ASURA_MM_HASH_TABLE_SIZE; ++i )
    {
        while( HashTable[i] )
        {
            Asura_MemoryNode* pxNode = HashTable[i];
            HashTable[i] = pxNode->NextNode;
            FreeMemoryNodeMem( pxNode );
        }
    }

#ifdef ASURA_USE_MEMPOOL_FOR_MEMORY_NODES
	// shutdown the node mempool
	m_pxMemoryNodePool->Shutdown();
	Asura_MemoryUtility::Deallocate(m_pxMemoryNodePool);
	m_pxMemoryNodePool = NULL;
#endif

    Asura_MemoryUtility::Shutdown();

    if( bAnyMemLeaks && g_bReportMemoryLeaksOnShutdown )
    {
		Asura_MemoryManagement_System::s_bMemoryLeak = true;
        
        Asura_Assert(0, "ASURA MEMORY MANAGER DETECTED MEMORY LEAKS - PLEASE SEE LOGFILE");
#ifdef ASURA_LOGFILES_ENABLED
        Asura_OutputToDebugger_Retail((
            "\n"
            "***************************************************************\n"
            "ASURA MEMORY MANAGER DETECTED MEMORY LEAKS - PLEASE SEE LOGFILE\n"
            "***************************************************************\n"
            "\n"
        ));
#else
        Asura_OutputToDebugger_Retail((
            "\n"
            "************************************************************************************************\n"
            "ASURA MEMORY MANAGER DETECTED MEMORY LEAKS - TURN ON ASURA_LOGFILES_ENABLED TO GET LIST OF LEAKS\n"
            "************************************************************************************************\n"
            "\n"
        ));
#endif
    }
	else
	{
		Asura_OutputToDebugger_Retail(("Asura MemoryManagement Shutdown - no leaks detected\n"));
	}

    //Asura_Logfile_Retail((MEMORYLOGFILE,"MemoryManagement Shutdown Successfully"));

	#ifdef ASURA_ENABLE_MEMTRACKER
	Asura_MemTracker::DestroyInstance ();
	#endif
}

Asura_MemoryNode* Asura_MemoryManagement_Core::AcquireMemoryNode(void)
{
	++NumberOfActiveMemoryNodes;

#ifdef ASURA_USE_MEMPOOL_FOR_MEMORY_NODES
	Asura_Assert(m_pxMemoryNodePool, "node mempool was not created?");
	Asura_MemoryNode* node = static_cast<Asura_MemoryNode*>(m_pxMemoryNodePool->Allocate());
#else
	Asura_MemoryNode* node = static_cast<Asura_MemoryNode*>(Asura_MemoryUtility::Allocate(sizeof(Asura_MemoryNode)));
#endif
        Asura_Assert(node, "failed to acquire memory node");
	return node;
}

void Asura_MemoryManagement_Core::InsertMemoryNode(Asura_MemoryNode* node)
{
	u_int hashIndex = GetHashIndex( node->GetReportedAddress() );

	node->NextNode = HashTable[hashIndex];
	HashTable[hashIndex] = node;
}

Asura_MemoryNode* Asura_MemoryManagement_Core::FindMemoryNode(void* address, Asura_MemoryNode** ppxPrevNode)
{
	Asura_MemoryNode *node = HashTable[GetHashIndex( address )];

	Asura_MemoryNode* pxPrevNode = NULL;
	while (node && node->GetReportedAddress() != address)
	{
		pxPrevNode = node;
		node = node->NextNode;
	}

	*ppxPrevNode = pxPrevNode;
	return node;
}

void Asura_MemoryManagement_Core::ReleaseMemoryNode(Asura_MemoryNode* node, Asura_MemoryNode* pxPrevNode)
{
	if (node)
	{
#if defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ALL) || defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ANONYMOUS)
        if( node->pvStackDump )
        {
            Asura_CallStack::Platform_DeallocStackDump( node->pvStackDump );
            node->pvStackDump = NULL;
        }
#endif

		if (pxPrevNode)
		{
			pxPrevNode->NextNode = node->NextNode;
		}
		else
		{
			u_int hashIndex = GetHashIndex( node->GetReportedAddress() );
			HashTable[hashIndex] = node->NextNode;
		}
	
		FreeMemoryNodeMem(node);

		--NumberOfActiveMemoryNodes;
	}
}

void Asura_MemoryManagement_Core::FreeMemoryNodeMem(Asura_MemoryNode* pxNode)
{
#ifdef ASURA_USE_MEMPOOL_FOR_MEMORY_NODES
	Asura_Assert(m_pxMemoryNodePool, "node mempool was not created?");
	m_pxMemoryNodePool->Deallocate(pxNode);
#else
	Asura_MemoryUtility::Deallocate(pxNode);
#endif
}


#if !defined( DBG_BDUMPMALLOCINFOWHENOUTOFMEMORY)
	#define DBG_BDUMPMALLOCINFOWHENOUTOFMEMORY false
#endif
DEBUGVAR bool dbg_bDumpMallocInfoWhenOutOfMemory = DBG_BDUMPMALLOCINFOWHENOUTOFMEMORY;

static bool g_bInsideAllocate = false;

void* Asura_MemoryManagement_Core::Allocate( size_t uNumBytes, ASURA_MM_EVENT_ID eEventID, u_int uAlignment, const char* szFile, int iLine, u_int uReturnAddr )
{
    if( !szFile )
    {
        szFile = szUNKNOWN_FILE;
    }

#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
    if (g_bCheckMallocGuardBytesOnce)
    {
        CheckMallocGuardBytes();
        g_bCheckMallocGuardBytesOnce = false;
    }
    else if (g_bCheckMallocGuardBytesAlways)
    {
        CheckMallocGuardBytes();
    }
#endif

	/// Kevin Lea - 24/10/2001 15:22:17 - ANSI allows allocations of zero bytes (!?)
    // (MP - 07.10.2002) Worth noting that certain parts of DirectX depend
    // upon the fact that you can successfully allocate and deallocate 0 byte
    // blocks of memory.
    if( !uNumBytes ) { uNumBytes = 1; }

    ClearMemoryLogfileIfNecessary();
	
	if (g_bInsideAllocate)
	{
		Asura_Assert(false, "recursive call to allocate");
	}

	g_bInsideAllocate = true;

	switch( eEventID )
	{
		case ASURA_NEW:
		{
			++EventCount_New;
			Asura_Assert( !uAlignment, "cannot be aligned" );
			break;
		}
		case ASURA_NEWARRAY:
		{
			++EventCount_NewArray;
			Asura_Assert( !uAlignment, "cannot be aligned" );
			break;
		}
		case ASURA_NEWALIGN:
		{
			++EventCount_NewAlign;
			break;
		}
		case ASURA_ALLOC:
		{
			++EventCount_Alloc;
			Asura_Assert( !uAlignment, "cannot be aligned" );
			break;
		}
		default:
		{
			Asura_Assert( 0, "Unexpected case" );
			g_bInsideAllocate = false;
            return NULL;
		}
	}

    // (MP - 07.10.2002) Allocate the memory first; this solves the problem
    // of trying to release uninitialised memory nodes (and the subsequent
    // crash.)
    u_int uBytesToAllocate = (u_int)uNumBytes + uAlignment;		// PAZ - fix for PC.x64 build
#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
    uBytesToAllocate += ASURA_MEMORYMANAGEMENT_GUARDS_START + 1;// need (default alignment) bytes at start and one byte at the end
#endif

	void* pMemory;
	if (g_uMaxMemToAllocate && (m_uTotalMemoryCurrentlyAlloced + uBytesToAllocate > g_uMaxMemToAllocate))
	{
		Asura_Assert( false, "Out of memory (based on limit set by SetMaxMemToAllocate)" );
		g_bInsideAllocate = false;
		return NULL;
	}
	else
	{
		pMemory = ASURA_MM_GET_MEM(uBytesToAllocate);
	}

    if( !pMemory )
    {
	    Asura_Assert( false, "Out of memory!" );
		Asura_OutputToDebugger_Retail(( "out of memory (tried to alloc %u bytes, %u bytes allocated\n", uBytesToAllocate, m_uTotalMemoryCurrentlyAlloced));

        // (MP - 07.10.2002) Seeing as we're regarding running out of memory
        // as an error, it'll be useful to be able to see what's currently
        // allocated.
        if( dbg_bDumpMallocInfoWhenOutOfMemory )
        {
			bool bDumpSummary = true;
			if (bDumpSummary)
			{
				DumpMallocSummary();
			}
			else
			{
				DumpMallocInfo();
			}
        }

        g_bInsideAllocate = false;
        return NULL;
    }

	Asura_MemoryNode* node = AcquireMemoryNode();
    if( !node )
    {
	    Asura_Assert( false, "Couldn't acquire memory node." );
        ASURA_MM_FREE_MEM( pMemory );
        g_bInsideAllocate = false;
        return NULL;
    }

	// store actual address and size in the node
	node->ActualAddress = pMemory;
	node->ActualSize = uBytesToAllocate;

	#ifdef ASURA_ENABLE_MEMTRACKER
	if (m_pxMemTracker)
	{
		m_pxMemTracker->AddAllocation (pMemory, uBytesToAllocate);
	}
	#endif

	// offset pMemory for guard bytes and for alignment
#if defined(ASURA_MEMORYMANAGEMENT_GUARDS_START) && defined(ASURA_MEMORYMANAGEMENT_FILL)
    pMemory = (void*)((unsigned char*)pMemory + ASURA_MEMORYMANAGEMENT_GUARDS_START);
#endif

    if( uAlignment )
	{
		u_int offset = (uAlignment - static_cast<u_int>(reinterpret_cast<size_t>(pMemory)) & (uAlignment-1));
		pMemory = (void*)((unsigned char*)pMemory + offset);
	}

	// store alignment in the node - using switch so we can check its a power of 2 which is <=128
	u_int uAlignmentShift = 0;
	switch (uAlignment)
	{
		case 0: break;
		case 2: uAlignmentShift = 1; break;
		case 4: uAlignmentShift = 2; break;
		case 8: uAlignmentShift = 3; break;
		case 16: uAlignmentShift = 4; break;
		case 32: uAlignmentShift = 5; break;
		case 64: uAlignmentShift = 6; break;
		case 128: uAlignmentShift = 7; break;
		default:
	{
			Asura_Assert(false, "unsupported alignment");
			break;
	}
	}
	node->uAlignShift = uAlignmentShift;

	// finish setting up the node
	node->uAllocationType = static_cast<u_int>(eEventID);
    node->pFilename = szFile;
    node->SetLine(iLine);
    node->SetFrameNum(Asura_Timers::GetFrameNumber());
#if defined ASURA_MEMORYMANAGEMENT_USE_RETURN_ADDRESS
    node->uReturnAddr = uReturnAddr;
#endif
#if defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ALL) || defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ANONYMOUS)
    node->pvStackDump = Asura_CallStack::Platform_CreateStackDump( 3 );
#endif

    if( dbg_bLogAllMemAllocs && uNumBytes >= dbg_uMinAllocSizeToReport )
    {
        Asura_SimpleLogfileOutput(( MEMORYLOGFILE,
                                    "Alloc   in frame %4i: [0x%08x] %10i bytes at line %4i of <%s>.",
                                    Asura_Timers::GetFrameNumber(),
                                    pMemory,
                                    uNumBytes,
                                    iLine,
                                    szFile ));
    }

#ifdef ASURA_MEMORYMANAGEMENT_PROCESS_ALLOCATED_MEM
    Asura_MemoryManagement_System::Platform_ProcessAllocatedMem( node->ActualAddress,
                                                                 node->ActualSize,
                                                                 eEventID,
                                                                 uAlignment,
                                                                 pMemory,
                                                                 uNumBytes );
#endif

#ifdef ASURA_MEMORYMANAGEMENT_FILL
    Asura_MemoryManagement_System::Platform_FillMem( node->ActualAddress,
                                                     (unsigned)node->ActualSize,		// PAZ - fix for PC.x64 build
                                                     eEventID,
                                                     uAlignment,
                                                     pMemory,
                                                     uNumBytes );
#endif

	InsertMemoryNode(node);
	
	m_uTotalMemoryAllocedOverall += (u_int)node->ActualSize;		// PAZ - fix for PC.x64 build
	m_uTotalMemoryCurrentlyAlloced += (u_int)node->ActualSize;		// PAZ - fix for PC.x64 build

	if (m_uTotalMemoryCurrentlyAlloced > m_uPeakMemoryUsage) 
	{
		m_uPeakMemoryUsage = m_uTotalMemoryCurrentlyAlloced;
	}

	m_iTotalMemoryAllocations++;
	m_iCurrentMemoryAllocations++;
	if( m_iCurrentMemoryAllocations > m_iPeakMemoryAllocations )
	{
		m_iPeakMemoryAllocations = m_iCurrentMemoryAllocations;
	}

	g_bInsideAllocate = false;
	return pMemory;
}

void Asura_MemoryManagement_Core::Deallocate( void* pAddr, enum ASURA_MM_EVENT_ID eEventID, const char* szFile, int iLine )
{
#if defined(ASURA_LOGFILES_ENABLED)
	const char*	szFileToOutput	= (NULL == szFile) ? szUNKNOWN_FILE : szFile;
#endif	//defined(ASURA_LOGFILES_ENABLED)

	/// Kevin Lea - 24/10/2001 09:42:39 - ANSI allows delete/free to be called
	/// with null pointers. I originally thought of asserting on this anyway,
	/// to see if there were any logic errors in Asura, but other systems such
	/// as Windows and DirectX quite happily delete null pointers all the time.
    if( !pAddr ) { return; }

    ClearMemoryLogfileIfNecessary();

	switch( eEventID )
	{
		case ASURA_DELETE:
		{
			++EventCount_Delete;
			break;
		}
		case ASURA_DELETEARRAY:
		{
			++EventCount_DeleteArray;
			break;
		}
		case ASURA_DELETEALIGN:
		{
			++EventCount_DeleteAlign;
			break;
		}
		case ASURA_DEALLOC:
		{
			++EventCount_Dealloc;
			break;
		}
		default:
		{
			Asura_Assert( 0, "Unexpected case" );
			break;
		}
	}
	
	Asura_MemoryNode* pxPrevNode;
	Asura_MemoryNode* node = FindMemoryNode( pAddr, &pxPrevNode );

	Asura_Assert( node!=0, "No matching address found for deallocate" );

	if(node==0) 
    {
        Asura_SimpleLogfileOutput(( MEMORYLOGFILE, "No matching address found for deallocate" ));
        Asura_SimpleLogfileOutput((MEMORYLOGFILE, "called from %s, line %d", szFileToOutput, iLine ));
        return;
    }

	#ifdef ASURA_ENABLE_MEMTRACKER
	if (m_pxMemTracker)
	{
		m_pxMemTracker->RemoveAllocation (node->ActualAddress, node->ActualSize);
	}
	#endif

    if( dbg_bLogAllMemDeallocs && node->GetReportedSize() >= dbg_uMinDeallocSizeToReport )
    {
        Asura_SimpleLogfileOutput(( MEMORYLOGFILE,
                                    "Dealloc in frame %4i: [0x%08x] %10i bytes at line %4i of <%s>.",
                                    Asura_Timers::GetFrameNumber(),
                                    node->GetReportedAddress(),
                                    node->GetReportedSize(),
                                    iLine,
                                    szFileToOutput ));
    }

   	if(!g_bIgnoreNewAndDeleteTypeMismatch)
	{
		if( eEventID == ASURA_DELETE && node->uAllocationType!= ASURA_NEW )
		{
			/// Kevin Lea - 30/10/2001 17:17:42 - TODO, insert logging info here
    	    Asura_Assert(0, "delete called for memory not alloced with new");
			Asura_SimpleLogfileOutput((MEMORYLOGFILE, "delete called for memory not alloced with new\n" ));
    	    Asura_SimpleLogfileOutput((MEMORYLOGFILE, "no information on location of delete\n" ));
		}
		else if( eEventID == ASURA_DELETEARRAY && node->uAllocationType!= ASURA_NEWARRAY )
		{
			/// Kevin Lea - 30/10/2001 17:17:42 - TODO, insert logging info here
    	    Asura_Assert(0, "delete [] called for memory not alloced with new []");
			Asura_SimpleLogfileOutput(( MEMORYLOGFILE, "delete [] called for memory not alloced with new []\n" ));
    	    Asura_SimpleLogfileOutput((MEMORYLOGFILE, "no information on location of delete\n" ));
		}
		else if( eEventID == ASURA_DELETEALIGN && node->uAllocationType!= ASURA_NEWALIGN )
		{
			/// Kevin Lea - 30/10/2001 17:17:42 - TODO, insert logging info here
			Asura_Assert(0, "delete align called for memory not alloced with new align");
    	    Asura_SimpleLogfileOutput(( MEMORYLOGFILE, "delete align called for memory not alloced with new align\n" ));
    	    Asura_SimpleLogfileOutput((MEMORYLOGFILE, "called from %s, line %d\n", szFileToOutput, iLine ));
		}
		else if( eEventID == ASURA_DEALLOC && node->uAllocationType != ASURA_ALLOC )
		{
			Asura_Assert(0, "Deallocate called for memory not alloced by Allocate");
    	    Asura_SimpleLogfileOutput(( MEMORYLOGFILE, "Deallocate called for memory not alloced my Allocate\n" ));
    	    Asura_SimpleLogfileOutput((MEMORYLOGFILE, "called from %s, line %d\n", szFileToOutput, iLine ));
		}
	}

	m_uTotalMemoryCurrentlyAlloced -= (u_int)node->ActualSize;		// PAZ - fix for PC.x64 build
	m_iCurrentMemoryAllocations--;

#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
	node->CheckGuardBytes();
#endif
	
	void* actualAddress = node->ActualAddress;
#ifdef ASURA_MEMORYMANAGEMENT_FILL
	const u_int actualSize = (u_int)node->ActualSize;		// PAZ - fix for PC.x64 build
#endif

	ReleaseMemoryNode( node, pxPrevNode );

#ifdef ASURA_MEMORYMANAGEMENT_FILL
    // MH note: we must do this before calling free since we can't write to 
    // memory that has been freed. On PS2, the memory manager writes 8 bytes of
    // data at the start of a freed block, so unless we write our own malloc/
    // free we can't fill the entire free'd block.
    // (MP - 30.07.2002) Not bothering to make alignment, etc., available
    // to deallocation fills at the moment; they weren't necessary for the
    // PS2 code and no other systems fill manually at the moment. You'll
    // probably need to add an alignment field to the nodes if this is ever
    // required.
    Asura_MemoryManagement_System::Platform_FillMem( actualAddress, actualSize, eEventID );
#endif

	ASURA_MM_FREE_MEM( actualAddress);

    Asura_MemoryManagement_System::Platform_PostDeallocate();
}

static char* g_szMallocDumpLineBfr = NULL;
static const u_int MALLOC_DUMP_LINE_BFR_SIZE = 1024;

static u_int PrepareMallocDumpLine( const char* szFmt, ... )
{
	va_list xArgs;
	va_start( xArgs, szFmt );
    u_int uLen = Asura_PrintfToBfr( g_szMallocDumpLineBfr, szFmt, xArgs, MALLOC_DUMP_LINE_BFR_SIZE, true );
	va_end( xArgs );
    return uLen;
}

DEBUGVAR bool dbg_bDumpFullInfoAtCheckpoint = false;

// KOR - 30/01/2006 - see declaration of Asura_MemoryManagement_System::Checkpoint in Asura_MemoryManagement.h
void Asura_MemoryManagement_Core::Checkpoint(int iState)
{
	Asura_Assert(0 <= iState && iState < ASURA_MM_STATE_NUM_TYPES, "State out of range");
	if(iState < 0 || iState >= ASURA_MM_STATE_NUM_TYPES) { return; }

	static int ls_aiCounter[ASURA_MM_STATE_NUM_TYPES];

	const int iAllocs = NumberOfActiveMemoryNodes;
	char szLogfileName[sizeof("Memory_Checkpoint_LXXX_State_XX-X.txt")];
	Asura_MemoryManagement_Core_State& xState = m_axPreviousStates[iState];
	// Check that this isn't the first time we have encountered this state this run
	if(xState.m_iNumberOfAllocations != iASURA_MM_STATE_UNINITIALISED)
	{
		if(xState.m_iNumberOfAllocations != iAllocs || xState.m_uTotalMemoryCurrentlyAlloced != m_uTotalMemoryCurrentlyAlloced)
		{
			// Dump to logfile for diffing.
			Asura_PrintfToBfr(szLogfileName, "Memory_Checkpoint_Leak_State_%02i-%1i.txt", sizeof(szLogfileName), true, iState, ls_aiCounter[iState] ^ 1);
			if(dbg_bDumpFullInfoAtCheckpoint)
			{
				DumpMallocInfo( szLogfileName, NULL, (ASURA_MALLOCDUMPFLAG_DEFAULTS | ASURA_MALLOCDUMPFLAG_REPORTEDADDR) );
			}
			else
			{
				DumpMallocSummary(szLogfileName);
			}

			// KOR - 13/02/2006 - If you've hit this assert, you should find that you have two logfiles named
			// Memory_Checkpoint_Leak_State_<i>.txt and Memory_Checkpoint_Last_State_<i>.txt, where <i> is the
			// current value for iState. By doing a plain text diff on these files, you can find out what the
			// change in the memory environment is. If the logfiles are missing, ensure that ASURA_LOGFILES_ENABLED
			// is defined for your current build and try to recreate the leak.
			//
			// N.B. If you continue running beyond this point, Memory_Checkpoint_Last_State_<i>.txt
			// will be over-written (see the bottom of this function), so make sure you copy off the versions
			// you want to compare. This is to avoid having a ridiculous number of logfiles clogging the directory.
            // (MH - 11.06.2007) - to give you time to grab copies if the assert is ignored automatically, I've 
			// changed this to store copies with -0 and -1 at the end of the filename.
			Asura_Assertf(0, ("Memory environment has changed since the last time in this state. Previous: %i allocs, %u bytes. Current: %i allocs, %u bytes. See comment for explanation (compare files ending in %02i-%1i).",
				xState.m_iNumberOfAllocations, xState.m_uTotalMemoryCurrentlyAlloced, iAllocs, m_uTotalMemoryCurrentlyAlloced, iState, ls_aiCounter[iState] ^ 1));
		}
	}
	// Now update the stored state.
	xState.m_iNumberOfAllocations = iAllocs;
	xState.m_uTotalMemoryCurrentlyAlloced = m_uTotalMemoryCurrentlyAlloced;

	// change the counter index
	ls_aiCounter[iState] ^= 1;

	// Dump to logfile
	Asura_PrintfToBfr(szLogfileName, "Memory_Checkpoint_Last_State_%02i-%1i.txt", sizeof(szLogfileName), true, iState, ls_aiCounter[iState] ^ 1);
	if(dbg_bDumpFullInfoAtCheckpoint)
	{
		DumpMallocInfo( szLogfileName, NULL, (ASURA_MALLOCDUMPFLAG_DEFAULTS | ASURA_MALLOCDUMPFLAG_REPORTEDADDR) );
	}
	else
	{
		DumpMallocSummary(szLogfileName);
	}
}

void Asura_MemoryManagement_Core::DumpMallocInfo(const char* szFilenameNoPath /* = NULL */, const char* szPath /* = NULL */, u_int uFlags /* = ASURA_MALLOCDUMPFLAG_DEFAULTS */)
{
    bool bWasEnabled = bEnableMemoryManager;
    bEnableMemoryManager = false;// turn off while we do the dump

    char szFileName[256];
    Asura_PrintfToBfr(szFileName, "%s%s", 256, true, 
        szPath ? szPath : MALLOC_DUMP_DEFAULT_PATH, 
        szFilenameNoPath ? szFilenameNoPath : MALLOC_INFO_DUMP_DEFAULT_FILE);

#if defined( DBG_BDUMPMALLOCINFOWHENOUTOFMEMORY_XML) && defined( ASURA_TOOLS)
	const u_int uXMLTempStringSizeMax = 256;
	char szXMLTempString[ uXMLTempStringSizeMax] = {0};
    char szFileNameXML[256];
    Asura_PrintfToBfr(szFileNameXML, "%s.xml", 256, true, szFileName);
    
    Asura_SimpleLogfileOpenXML( szFileNameXML);
    Asura_SimpleLogfileOutput((szFileNameXML, "<MemoryDump>\n"));
#endif

    Asura_OutputToDebugger(( "Dumping malloc info to \"%s\".\n", szFileName ));

    Asura_MemoryManagement_System::BeginMallocDump( szFileName );

    char szLineBfr[MALLOC_DUMP_LINE_BFR_SIZE];
    g_szMallocDumpLineBfr = szLineBfr;
    #define ADD_LINE( args ) \
    { \
        u_int uLen = PrepareMallocDumpLine args; \
        Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, uLen ); \
    }
    #define ADD_LINE_DIRECT( formatted_string ) \
    { \
        Asura_MemoryManagement_System::FlushMallocDump(); \
		Asura_SimpleLogfileOutput_PreFormatted( g_szMallocDumpFilename, formatted_string, strlen(formatted_string) ); \
    }

    ADD_LINE(( "Memory dump: %u bytes outstanding in %i allocations.\n", m_uTotalMemoryCurrentlyAlloced, NumberOfActiveMemoryNodes));

    if( uFlags & ASURA_MALLOCDUMPFLAG_FRAMENUM )
    {
        // Record current frame number for reference.
        ADD_LINE(( "Current frame number: %u\n", Asura_Timers::GetFrameNumber() ));
    }

    ADD_LINE(( "\n" ));

	if(uFlags & ASURA_MALLOCDUMPFLAG_FILEANDLINE)	ADD_LINE(( "File and line\t" ));
	if(uFlags & ASURA_MALLOCDUMPFLAG_ACTUALSIZE)	ADD_LINE(( "Bytes\t" ));
    if(uFlags & ASURA_MALLOCDUMPFLAG_REPORTEDSIZE)  ADD_LINE(( "Reported size\t" ));
	if(uFlags & ASURA_MALLOCDUMPFLAG_RETURNADDR)	ADD_LINE(( "Return addr\t" ));
	if(uFlags & ASURA_MALLOCDUMPFLAG_EXTRAINFO)		ADD_LINE(( "Extra info\t" ));
	if(uFlags & ASURA_MALLOCDUMPFLAG_ACTUALADDRESS)	ADD_LINE(( "Actual address\t" ));
    if(uFlags & ASURA_MALLOCDUMPFLAG_REPORTEDADDR)  ADD_LINE(( "Reported addr\t" ));
	if(uFlags & ASURA_MALLOCDUMPFLAG_FRAMENUM)		ADD_LINE(( "Frame\t" ));
	ADD_LINE(( "\n\n" ));

#if _DEBUG
	int iNumOutputToDebugger = 0;
#endif

	for (int i = 0; i < ASURA_MM_HASH_TABLE_SIZE; i++)
    {
        for( const Asura_MemoryNode* pxNode = HashTable[i]; pxNode; pxNode = pxNode->NextNode )
        {
#if _DEBUG
            if( uFlags & ASURA_MALLOCDUMPFLAG_ISLEAKREPORT )
            {
			    // in debug builds output the first few to the debugger for easy double-clicking to get to the file and line.
			    if (iNumOutputToDebugger++ < 16)
			    {
				    Asura_OutputToDebugger(( "%s(%u)\n\t(%i bytes)\tframe %i\t%s\n", pxNode->pFilename, 
					    pxNode->uLine, pxNode->GetReportedSize(), pxNode->uFrameNum, GetExtraNodeInfo(pxNode)  ));
			    }
            }
#endif // DEBUG

			if(uFlags & ASURA_MALLOCDUMPFLAG_FILEANDLINE)	ADD_LINE(( "%s @ %u\t", pxNode->pFilename, pxNode->uLine ));
			if(uFlags & ASURA_MALLOCDUMPFLAG_ACTUALSIZE)	ADD_LINE(( "%u\t", pxNode->ActualSize ));
			if(uFlags & ASURA_MALLOCDUMPFLAG_REPORTEDSIZE)	ADD_LINE(( "%u\t", pxNode->GetReportedSize() ));
#if defined ASURA_MEMORYMANAGEMENT_USE_RETURN_ADDRESS
			if(uFlags & ASURA_MALLOCDUMPFLAG_RETURNADDR)	ADD_LINE(( "0x%08X\t", pxNode->uReturnAddr ));
#endif
			if(uFlags & ASURA_MALLOCDUMPFLAG_EXTRAINFO)		ADD_LINE(( "%s\t", GetExtraNodeInfo(pxNode) ));
			if(uFlags & ASURA_MALLOCDUMPFLAG_ACTUALADDRESS)	ADD_LINE(( "0x%08X\t", pxNode->ActualAddress ));
            if(uFlags & ASURA_MALLOCDUMPFLAG_REPORTEDADDR)  ADD_LINE(( "0x%08X\t", pxNode->GetReportedAddress() ));
			if(uFlags & ASURA_MALLOCDUMPFLAG_FRAMENUM)		ADD_LINE(( "%u\t", pxNode->uFrameNum ));
			ADD_LINE(( "\n" ));

#if defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ALL) || defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ANONYMOUS)
            if( uFlags & ASURA_MALLOCDUMPFLAG_CALLSTACK )
            {
#if defined(ASURA_MEMORYMANAGEMENT_STACK_WALK_ANONYMOUS)
                if( !strcmp(pxNode->pFilename, szUNKNOWN_FILE) )
#endif
                {
                    ADD_LINE(( "Callstack:\n" ));
                    if( pxNode->pvStackDump )
                    {
                        char szStackDump[4096];
                        Asura_CallStack::Platform_GetHumanReadableStackDump( pxNode->pvStackDump, szStackDump, sizeof(szStackDump) );
                        ADD_LINE_DIRECT(( szStackDump )); // Bypass buffering since this may be very large.
                        ADD_LINE(( "\n" ));
                    }
                    else
                    {
                        ADD_LINE(( "UNAVAILABLE\n" ));
                    }
                }
            }
#endif
#if defined( DBG_BDUMPMALLOCINFOWHENOUTOFMEMORY_XML) && defined( ASURA_TOOLS)
		    Asura_SimpleLogfileOutput((szFileNameXML, "<Node"));
			if(uFlags & ASURA_MALLOCDUMPFLAG_FILEANDLINE)	
			{ 
				Asura_SimpleLogfileConvertToXML( pxNode->pFilename, szXMLTempString, uXMLTempStringSizeMax);
				Asura_SimpleLogfileOutput((szFileNameXML, " file='%s' line='%d'", szXMLTempString, pxNode->uLine ));
			}
			if(uFlags & ASURA_MALLOCDUMPFLAG_ACTUALSIZE) { Asura_SimpleLogfileOutput(( szFileNameXML, " ActualSize='%i'", pxNode->ActualSize ));}
			if(uFlags & ASURA_MALLOCDUMPFLAG_REPORTEDSIZE) { Asura_SimpleLogfileOutput(( szFileNameXML, " ReportedSize='%i'", pxNode->GetReportedSize() ));}
#if defined ASURA_MEMORYMANAGEMENT_USE_RETURN_ADDRESS
			if(uFlags & ASURA_MALLOCDUMPFLAG_RETURNADDR) { Asura_SimpleLogfileOutput(( szFileNameXML, " ReturnAddr='0%08'", pxNode->uReturnAddr ));}
#endif
			if(uFlags & ASURA_MALLOCDUMPFLAG_ACTUALADDRESS) { Asura_SimpleLogfileOutput(( szFileNameXML, " ActualAddr='0%08'", pxNode->ActualAddress ));}
            if(uFlags & ASURA_MALLOCDUMPFLAG_REPORTEDADDR) { Asura_SimpleLogfileOutput(( szFileNameXML, " ReportedAddr='0%08'", pxNode->GetReportedAddress() ));}
			if(uFlags & ASURA_MALLOCDUMPFLAG_FRAMENUM) { Asura_SimpleLogfileOutput(( szFileNameXML, " Frame='%i'", pxNode->uFrameNum ));}
			
		    Asura_SimpleLogfileOutput((szFileNameXML, " />\n"));
#endif
        }
    }

    #undef ADD_LINE_DIRECT
    #undef ADD_LINE
    g_szMallocDumpLineBfr = NULL;

    Asura_MemoryManagement_System::EndMallocDump();

#if defined( DBG_BDUMPMALLOCINFOWHENOUTOFMEMORY_XML) && defined( ASURA_TOOLS)
    Asura_SimpleLogfileOutput((szFileNameXML, "</MemoryDump>"));
    Asura_SimpleLogfileCloseXML( szFileNameXML);
#endif

    Asura_OutputToDebugger(( "Finished dumping malloc info to \"%s\".\n", szFileName ));

    bEnableMemoryManager = bWasEnabled;
}

struct DumpMallocSummaryNode
{
	const char*	m_szFilename;
	u_int		m_uLine;
	u_int		m_uNumAllocations;
	u_int		m_uTotalActualSize;
	u_int		m_uTotalReportedSize;
};

static void DumpMallocSummaryOutputData (Asura_Hash_ID uHash, const DumpMallocSummaryNode* pxNode)
{
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "%s\t", pxNode->m_szFilename ));
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "%i\t", pxNode->m_uLine ));
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "%i\t", pxNode->m_uTotalActualSize ));
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "%i\t", pxNode->m_uTotalReportedSize ));
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "%i\t", pxNode->m_uNumAllocations ));
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "\n" ));
}


// MGH 21/2/2008: Same as Asura_GetHashID, but without the Asura_Universal_HashLookup code,
// which causes asserts on shutdown if the memory manager was disabled when adding nodes.
Asura_Hash_ID Asura_GetHashID_Simple( const char* szString )
{
	if(!szString) return ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uHash = 0;

	char cThisChar;
	for( ; *szString ; ++szString )
	{
		cThisChar = *szString;
		
		if( cThisChar >= 'A' && cThisChar <= 'Z' )
		{
			cThisChar = static_cast<char>((cThisChar ) - 'A' + 'a');
		} 
		else if( cThisChar == '\\' )
		{
			// Convert back slash '\\' to forward slash '/'.
			cThisChar = '/';
		}

		uHash = ( uHash << 5 ) - uHash + cThisChar;
	}

	return uHash;
}


void Asura_MemoryManagement_Core::DumpMallocSummary(const char* szFilenameNoPath, const char* szPath)
{
	bool bWasEnabled = bEnableMemoryManager;
	bEnableMemoryManager = false;// turn off while we do the dump

	char szFileName[256];
	Asura_PrintfToBfr(szFileName, "%s%s", 256, true, 
		szPath ? szPath : MALLOC_DUMP_DEFAULT_PATH, 
		szFilenameNoPath ? szFilenameNoPath : MALLOC_SUMMARY_DUMP_DEFAULT_FILE);

	Asura_OutputToDebugger(( "Dumping malloc summary to \"%s\".\n", szFileName ));

	Asura_MemoryManagement_System::BeginMallocDump( szFileName );

	char szLineBfr[MALLOC_DUMP_LINE_BFR_SIZE];
	g_szMallocDumpLineBfr = szLineBfr;

	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "Memory dump: %u bytes outstanding in %i allocations.\n\n", m_uTotalMemoryCurrentlyAlloced, NumberOfActiveMemoryNodes));
	Asura_MemoryManagement_System::AddMallocDumpLine( g_szMallocDumpLineBfr, PrepareMallocDumpLine( "File\tLine\tActual Bytes\tReported Bytes\tNum Allocations\n\n" ));

	{
		Asura_TemplateHashTree<DumpMallocSummaryNode>	xSummaryHashTree;

		for (int i = 0; i < ASURA_MM_HASH_TABLE_SIZE; i++)
		{
			for( const Asura_MemoryNode* pxNode = HashTable[i]; pxNode; pxNode = pxNode->NextNode )
			{
				const char*			szFilename		= pxNode->pFilename;
				const u_int			uLine			= pxNode->uLine;
				const Asura_Hash_ID	uHash			= Asura_GetHashID_Simple(szFilename) + uLine;
				DumpMallocSummaryNode*		pxSummaryNode	= xSummaryHashTree.GetDataByHash(uHash);

				if(pxSummaryNode == NULL)
				{
					pxSummaryNode	= xSummaryHashTree.AddNode(uHash);

					if(pxSummaryNode == NULL)
					{
						Asura_Assert(false, "Failed to add node when creating summary hash-tree.");
					}
					else
					{
						pxSummaryNode->m_szFilename			= szFilename;
						pxSummaryNode->m_uLine				= uLine;
						pxSummaryNode->m_uNumAllocations	= 1;
						pxSummaryNode->m_uTotalActualSize	= (u_int)pxNode->ActualSize;			// PAZ - fix for PC.x64 build
						pxSummaryNode->m_uTotalReportedSize	= (u_int)pxNode->GetReportedSize();		// PAZ - fix for PC.x64 build
					}
				}
				else
				{
					Asura_Assertf(uLine == pxSummaryNode->m_uLine && Asura_GetHashID_Simple(szFilename) == Asura_GetHashID_Simple(pxSummaryNode->m_szFilename),
						( "Unexpected file and line, these should be the same - possible hash collision in DumpMallocSummary\n"
						  "HashTable Node line %d file %s\n"
						  "Summary Node line %d file %s\n", uLine, szFilename, pxSummaryNode->m_uLine, pxSummaryNode->m_szFilename ) );
					pxSummaryNode->m_uTotalActualSize	+= (u_int)pxNode->ActualSize;			// PAZ - fix for PC.x64 build
					pxSummaryNode->m_uTotalReportedSize	+= (u_int)pxNode->GetReportedSize();	// PAZ - fix for PC.x64 build
					pxSummaryNode->m_uNumAllocations++;
				}
			}
		}

		xSummaryHashTree.Traverse(DumpMallocSummaryOutputData);
	}

	g_szMallocDumpLineBfr = NULL;

	Asura_MemoryManagement_System::EndMallocDump();

	Asura_OutputToDebugger(( "Finished dumping malloc summary to \"%s\".\n", szFileName ));

	bEnableMemoryManager = bWasEnabled;
}

const char* Asura_MemoryManagement_Core::GetExtraNodeInfo(const Asura_MemoryNode* pxNode)
{
	const char* szExtraInfo = "";
	if (pxNode->pFilename)
	{
		if( strstr(pxNode->pFilename, "Asura_String.h") || strstr(pxNode->pFilename, "asura_string.h") )
		{
			szExtraInfo = reinterpret_cast<const char*>( pxNode->GetReportedAddress() );
		}
		else if( strstr(pxNode->pFilename, "Asura_String.cpp") || strstr(pxNode->pFilename, "asura_string.cpp") )
		{
			// MH 21.9.07 - only report these if they're longer than 4 bytes - this allows a check
			// on whether it's unicode or not. Previously you had to manually #if char or unicode char.
			if (pxNode->GetReportedSize() > 4)
			{
				const char* pcBuffer = reinterpret_cast<const char*>( pxNode->GetReportedAddress() );
				// NB: the check isn't very good. It assumes that the Unicode string is storing text which 
				// could have been stored in a char array, but that's generally the case with English.
				const char* szFormatString;
				if ( (*pcBuffer) && (*(pcBuffer + 1)) )
				{
					// probably not unicode.
					szFormatString = "%s";
				}
				else
				{
					// probably unicode
					szFormatString = "%S";
				}

				// Print it into a buffer.
				const u_int uBufferSize = 128;
				static char szBuffer[uBufferSize];
				Asura_PrintfToBfr(szBuffer, szFormatString, uBufferSize, true, pxNode->GetReportedAddress());
				szExtraInfo = szBuffer;
			}
		}
		else if( 64 != pxNode->uLine && (strstr(pxNode->pFilename, "Asura_NameTree.h") || strstr(pxNode->pFilename, "asura_nametree.h")) )
		{
			szExtraInfo = (reinterpret_cast<const Asura_NameTree_Node*>(pxNode->GetReportedAddress()))->Name.GetString();
		}
	}
	return szExtraInfo;
}

#if defined(ASURA_MEMORYMANAGEMENT_GUARDS) && defined(ASURA_MEMORYMANAGEMENT_FILL)
void Asura_MemoryManagement_Core::CheckMallocGuardBytes()
{
    for (int i = 0; i < ASURA_MM_HASH_TABLE_SIZE; i++)
    {
        Asura_MemoryNode *node = HashTable[i];
        while (node)
        {
			node->CheckGuardBytes();
            node = node->NextNode;
        }
    }
}
#endif

const char* Asura_MemoryManagement_System::GetDefaultMallocDumpPath()
{
	return MALLOC_DUMP_DEFAULT_PATH;
}

const char* Asura_MemoryManagement_System::GetMemoryLeakFilename()
{
	return szMEMORY_LEAK_FILE;
}
   
#else//#if ASURA_MEMORYMANAGEMENT_ENABLED

bool Asura_MemoryManagement_System::s_bMultithreadingEnabled = true;

#if ASURA_MEMORYMANAGEMENT_PLATFORM_INIT_WHEN_DISABLED
// constructor for initialiser objects: see comment in header file at Asura_MMS_InitialiserObject
Asura_MemoryManagement_System::Asura_MemoryManagement_System()
{
    static bool bOnce = true;
    if( !bOnce ) { return; }
    bOnce = false;

	Platform_Initialise();
	atexit( Asura_MemoryManagement_System::Platform_Shutdown );  
}
#endif

#if !defined (ASURA_PC_SSE)
#include <stdio.h>	// for printf when out of memory
void* Asura_Allocate(unsigned int uBytes)
{
	void* pRet = malloc(uBytes);
	if (!pRet)
	{
		printf( "out of memory (tried to alloc %u bytes)\n", uBytes );
	}
	return pRet;
}
#endif

#endif//#else #if ASURA_MEMORYMANAGEMENT_ENABLED

/*
    Utility memory allocation. Usually for debug purposes.
	The 'Platform_' style subfunctions are sort of static
	''advising'' of the real entry points. Some overhead, but
	then utility memory is not supposed to be performance
	critical.
*/

#if defined ASURA_MEMORYMANAGEMENT_ENABLED
void Asura_MemoryUtility::Initialise(void)
{
	Platform_Initialise();
}

void Asura_MemoryUtility::Shutdown(void)
{
	Platform_Shutdown();
}

/*
	'PreAllocate' can either do nothing or some preparatory work and
	return 0, or do the actual allocation and return the block address.
*/
void* Asura_MemoryUtility::Allocate(/*const*/ size_t size)
{
	{
		void *const pAddr = Platform_PreAllocate(size);

		if (pAddr != 0)
		{
			return pAddr;
		}
	}

	{
		void *const pAddr = malloc(size);

		return Platform_PostAllocate(pAddr);
	}
}

/*
	'PreDeallocate' can either do nothing or some preparatory work and
	return 'false', or do the actual deallocation and return 'true'.
*/
void Asura_MemoryUtility::Deallocate(void* /*const*/ pAddr)
{
	if (Platform_PreDeallocate(pAddr))
	{
		return;
	}

	free(pAddr);

	Platform_PostDeallocate(pAddr);
}

/*
	C wrappers.
*/

extern "C" void* Asura_C_UtilityAllocate(/*const*/ size_t uNumBytes)
{
	return Asura_MemoryUtility::Allocate(uNumBytes);
}

extern "C" void Asura_C_UtilityDeallocate(void* /*const*/ pAddr)
{
	Asura_MemoryUtility::Deallocate(pAddr);
}

#endif

