//
// Asura_NameTree.h
//
// Created by Kevin Lea - 12/06/2001 10:59:56
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_NAMETREE_H__
#define __ASURA_NAMETREE_H__

// Engine independent.
#include "Core/Asura_Handle.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

class Asura_NameTree_Node
{
	public:
 
		Asura_String Name;
		int Index;
		class Asura_NameTree_Node *Lower;
		class Asura_NameTree_Node *Higher;

		Asura_NameTree_Node( const char* szName, int iIndex )
		{
			Lower = 0;
			Higher = 0;	

			Name = szName;
			Index = iIndex;
		}
		~Asura_NameTree_Node()
		{
		}
};

class Asura_NameTree
{
	public:
		
		Asura_NameTree()
		{
			NumberOfNodes = 0;
			FirstRoot = 0;
			MaximumNodes = 0x7fffffff;
			m_bIgnoreExtensions = true;

			HandleList = 0;
		}

		Asura_NameTree(int maximumNodes)
		{
			NumberOfNodes = 0;
			FirstRoot = 0;
			MaximumNodes = maximumNodes;
			m_bIgnoreExtensions = true;

			HandleList = new Asura_Handle_List(maximumNodes);
		}

		virtual ~Asura_NameTree()
		{
			FreeAll();

			if (HandleList) delete HandleList;
		}


        // (MH - 15.12.2002) Reset method to help avoid memory fragmentation
        void Reset()
        {
            FreeAll();
			NumberOfNodes = 0;
			FirstRoot = 0;

            if (HandleList)
            {
                HandleList->Reset();
            }
        }
        
		/// Kevin Lea - 17/10/2001 12:34:43 - - returns pointer to string if
		/// a match is found, returns null otherwise
		const Asura_String* FindNodeNameByIndex( int __iIndex ) const
		{
			const Asura_NameTree_Node* pxNode = FirstRoot;

            // Empty tree?
            if( !pxNode )
			{
				return NULL;
			}

            // Search.
            return FindNodeNameByIndexRecursive( pxNode, __iIndex );
		}

		// KJL 08/03/00 17:08:16 - returns true if the node existed already
		bool FindNodeIndexByName( const Asura_String& xSearchName, bool bAddToTreeIfRequired, int& iIndex )
		{
			return FindNodeIndexByName( xSearchName.GetString(), bAddToTreeIfRequired, iIndex );
		}
		bool FindNodeIndexByName( const char* szSearchName, bool bAddToTreeIfRequired, int& iIndex )
		{
			if (!FirstRoot)
			{
				if( bAddToTreeIfRequired )
				{
					FirstRoot = new Asura_NameTree_Node( szSearchName, NumberOfNodes++ );

					if (HandleList)
					{
						#ifdef ASURA_ASSERT
						bool success = 
						#endif
						HandleList->GetNewHandle(FirstRoot->Index);

						Asura_Assert(success,"No more space!");
					}

					iIndex = FirstRoot->Index;
				}
				return false;
			}

			Asura_NameTree_Node* nodePtr = FirstRoot;
			while( !Equals(nodePtr->Name, szSearchName) )
			{
				while( LessThan(nodePtr->Name, szSearchName) )
				{
					if (nodePtr->Higher)
					{
						nodePtr = nodePtr->Higher;
						continue;
					}
					else if( bAddToTreeIfRequired )
					{
						nodePtr->Higher = new Asura_NameTree_Node( szSearchName, NumberOfNodes++ );
						
						if (HandleList)
						{
							#ifdef ASURA_ASSERT
							bool success = 
							#endif
							HandleList->GetNewHandle(nodePtr->Higher->Index);
				
							Asura_Assert(success,"No more space!");
						}
						iIndex = nodePtr->Higher->Index;
					}
					return false;
				}

				while( GreaterThan(nodePtr->Name, szSearchName) )
				{
					if (nodePtr->Lower)
					{
						nodePtr = nodePtr->Lower;
						continue;
					}
					else if( bAddToTreeIfRequired )
					{
						nodePtr->Lower = new Asura_NameTree_Node( szSearchName, NumberOfNodes++ );
					
						if (HandleList)
						{
							#ifdef ASURA_ASSERT
							bool success = 
							#endif
							HandleList->GetNewHandle(nodePtr->Lower->Index);

							Asura_Assert(success,"No more space!");
						}
						iIndex = nodePtr->Lower->Index;
					}
					return false;
				}
			}

			iIndex = nodePtr->Index;
			return true;
		}

		// This adds a node at a specified index if a node with that name doesn't already exist,
		// returning true if the node was added sucessfully.
		// No attempt is made to avoid clashes in index numbers , which will most likely occur
		// if this is used in combination with FindNodeIndexByName
		// RAR 30/05/2002
		bool InsertNode( const Asura_String& xSearchName, int iIndex )
		{
			return InsertNode( xSearchName.GetString(), iIndex );
		}
		bool InsertNode( const char* szSearchName, int iIndex )
		{
			Asura_NameTree_Node **nodePtrPtr = &FirstRoot;
			Asura_NameTree_Node *nodePtr = *nodePtrPtr;

			//search for the place to insert the new node
			while(nodePtr)
			{
				if( LessThan(nodePtr->Name, szSearchName) )
				{
					nodePtrPtr = &nodePtr->Higher;
					nodePtr = *nodePtrPtr;
				}
				else if( GreaterThan(nodePtr->Name, szSearchName) )
				{
					nodePtrPtr = &nodePtr->Lower;
					nodePtr = *nodePtrPtr;
				}
				else
				{
					//node already exists
					return false;
				}
			}

			//add the node here
			*nodePtrPtr = new Asura_NameTree_Node( szSearchName, iIndex );
			NumberOfNodes++;

			return true;

		}

		// Delete a node and get its index, returns true if the node was found.
		// Note!!
		// The node with index NumberOfNodes-1 will have its index changed to the index
		// of the node being deleted. (since the indeces must be consecutive)
		bool DeleteNodeByName( const Asura_String& xSearchName, int& iOutputIndex )
		{
            return DeleteNodeByName( xSearchName.GetString(), iOutputIndex );
        }
		bool DeleteNodeByName( const char* szSearchName, int& iOutputIndex )
		{
            if( !DeleteNodeByNameInternal(szSearchName, iOutputIndex) ) { return false; }

			//need to alter the index of the last node to take the place of the deleted one
			if(FirstRoot)
			{
				Asura_NameTree_Node* nodePtr = FindNodeByIndexRecursive(FirstRoot,NumberOfNodes);
				if(nodePtr)
				{
					nodePtr->Index = iOutputIndex;
				}
			}

			return true;//search and destroy mission accomplished...
		}

		bool DeleteNodeByName( const Asura_String& xSearchName )
		{
            return DeleteNodeByName( xSearchName.GetString() );
        }
		bool DeleteNodeByName( const char* szSearchName )
		{
			Asura_Assert(HandleList!=0,"This NameTree was not constructed with a Handle List");

			int outputIndex;
            if( !DeleteNodeByNameInternal(szSearchName, outputIndex) ) { return false; }

			HandleList->ReleaseHandle(outputIndex);

			return true;
		}

		//returns true if the node was sucessfully renamed
		bool RenameNode( const Asura_String& xOldName, const Asura_String& xNewName )
		{
            return RenameNode( xOldName.GetString(), xNewName.GetString() );
        }
		bool RenameNode( const char* szOldName, const char* szNewName )
		{
			//delete the old node , and get the index for it
			int index;
            if( !DeleteNodeByNameInternal(szOldName, index) ) { return false; }

	 		//search for the point to reinsert with the new name
			Asura_NameTree_Node **insertionPtr = &FirstRoot;
			while(*insertionPtr)
			{
				if( LessThan((*insertionPtr)->Name, szNewName) )
				{
					insertionPtr = &(*insertionPtr)->Higher;
				}
				else
				{
					insertionPtr = &(*insertionPtr)->Lower;
				}
			}

			//create the new node , and update the counters
			*insertionPtr = new Asura_NameTree_Node( szNewName, index );
			NumberOfNodes++;

			return true;
		}
		
		// KJL - 14-Mar-01 11:21:24 - call 'function' with every entry in the tree
		void Traverse( void (*function)(const Asura_String& xName, int index) ) const
		{
			if (FirstRoot) TraverseRecursive(FirstRoot,function);
		}

		void MakeNameList( const Asura_String** nameList ) const
		{
			if (FirstRoot) MakeNameListRecursive(FirstRoot,nameList);
		}

		void MakeNameList( const char** ppszNameList ) const
		{
			if (FirstRoot) MakeNameListRecursive(FirstRoot,ppszNameList);
		}

 		//get alphabetically sorted list of names
 		void MakeSortedNameList(const char** ppszNameList) const
		{
			int iIndex = 0;
			if (FirstRoot) MakeSortedNameListRecursive(FirstRoot,ppszNameList,iIndex);
		}

		inline int GetNumberOfNodes(void) const
		{
			return NumberOfNodes;
		}

		// KJL - 14-Mar-01 11:45:52 - N.B. returns the number of bytes
		int GetSizeOfNamesInIntChunks( void ) const { return( FirstRoot ? GetSizeOfNamesInIntChunksRecursive(FirstRoot) : 0 ); }
	
	private:

		void FreeAll(void)
		{
			if (FirstRoot) FreeAllRecursive(FirstRoot);
		}

		void FreeAllRecursive(Asura_NameTree_Node *nodePtr)
		{
			if (nodePtr->Lower)
			{
				FreeAllRecursive(nodePtr->Lower);
			}
			if (nodePtr->Higher)
			{
				FreeAllRecursive(nodePtr->Higher);
			}

			delete nodePtr;
		}

		void TraverseRecursive( const Asura_NameTree_Node* nodePtr, void (*function)(const Asura_String& name, int index) ) const
		{
			if (nodePtr->Lower)
			{
				TraverseRecursive(nodePtr->Lower,function);
			}
			if (nodePtr->Higher)
			{
				TraverseRecursive(nodePtr->Higher,function);
			}

			function(nodePtr->Name,nodePtr->Index);
		}

		void MakeNameListRecursive( Asura_NameTree_Node* nodePtr, const Asura_String** nameList ) const
		{
			if (nodePtr->Lower)
			{
				MakeNameListRecursive(nodePtr->Lower,nameList);
			}
			if (nodePtr->Higher)
			{
				MakeNameListRecursive(nodePtr->Higher,nameList);
			}
			
			nameList[nodePtr->Index] = &(nodePtr->Name);
		}

		void MakeNameListRecursive( Asura_NameTree_Node* nodePtr, const char** ppszNameList ) const
		{
			if (nodePtr->Lower)
			{
				MakeNameListRecursive(nodePtr->Lower,ppszNameList);
			}
			if (nodePtr->Higher)
			{
				MakeNameListRecursive(nodePtr->Higher,ppszNameList);
			}
			
			ppszNameList[nodePtr->Index] = nodePtr->Name.GetString();
		}

		void MakeSortedNameListRecursive(Asura_NameTree_Node* pxNode,const char** ppszNameList,int& iNextIndex) const
		{
			if(pxNode->Lower) MakeSortedNameListRecursive(pxNode->Lower,ppszNameList,iNextIndex);
			ppszNameList[iNextIndex++] = pxNode->Name.GetString();
			if(pxNode->Higher) MakeSortedNameListRecursive(pxNode->Higher,ppszNameList,iNextIndex);
		}

        Asura_NameTree_Node* FindNodeByIndexRecursive(Asura_NameTree_Node* __pxNode, int __iIndex ) const
        {
            if( __pxNode->Index == __iIndex )
            {
                return __pxNode;
            }
            else
            {
                Asura_NameTree_Node* pxRetVal = NULL;

                if( __pxNode->Lower )
                {
                    pxRetVal = FindNodeByIndexRecursive( __pxNode->Lower, __iIndex );
                }

                if( !pxRetVal && __pxNode->Higher )
                {
                    pxRetVal = FindNodeByIndexRecursive( __pxNode->Higher, __iIndex );
                }

                return pxRetVal;
            }
        }

        const Asura_String* FindNodeNameByIndexRecursive( const Asura_NameTree_Node* __pxNode, int __iIndex ) const
        {
            if( __pxNode->Index == __iIndex )
            {
                return &__pxNode->Name;
            }
            else
            {
                const Asura_String* pxRetVal = NULL;

                if( __pxNode->Lower )
                {
                    pxRetVal = FindNodeNameByIndexRecursive( __pxNode->Lower, __iIndex );
                }

                if( !pxRetVal && __pxNode->Higher )
                {
                    pxRetVal = FindNodeNameByIndexRecursive( __pxNode->Higher, __iIndex );
                }

                return pxRetVal;
            }
        }

		bool DeleteNodeByNameInternal( const char* szSearchName, int& iOutputIndex )
		{
			if(!FirstRoot) return false;

			Asura_NameTree_Node **nodePtrPtr = &FirstRoot;
			Asura_NameTree_Node *nodePtr = *nodePtrPtr;

			while( !Equals(nodePtr->Name, szSearchName) )
			{
				if( nodePtr->Higher && LessThan(nodePtr->Name, szSearchName) )
				{
					nodePtrPtr = &nodePtr->Higher;
					nodePtr = *nodePtrPtr;
				}
				else if( nodePtr->Lower && GreaterThan(nodePtr->Name, szSearchName) )
				{
					nodePtrPtr = &nodePtr->Lower;
					nodePtr = *nodePtrPtr;
				}
				else
				{
					//failed to find node
					return false;
				}
			}
			
			Asura_NameTree_Node *nodeToDelete = nodePtr;
			//found the node to delete , if it has children they will need to be transferred up
			if(nodeToDelete->Lower && nodeToDelete->Higher)
			{
				//The deleted node has 2 children  , 1 will need to be found a place beneath the other
				//Pick semi-randomly to keep some chance of things not being unbalanced
				Asura_NameTree_Node* nodeToTransfer;
				if(NumberOfNodes%2)
				{
					*nodePtrPtr = nodeToDelete->Lower;
					nodeToTransfer = nodeToDelete->Higher;
				}
				else
				{
					*nodePtrPtr = nodeToDelete->Higher;
					nodeToTransfer = nodeToDelete->Lower;
				}
				nodePtr = *nodePtrPtr;

				//search down for somewhere to stick the transferred node
				TryNextNode :
				
				if( LessThan(nodePtr->Name, nodeToTransfer->Name.GetString()) )
				{
					if(nodePtr->Higher)
					{
						nodePtr = nodePtr->Higher;
						goto TryNextNode;
					}
					else
					{
						nodePtr->Higher = nodeToTransfer;
					}
				}
				else
				{
					if(nodePtr->Lower)
					{
						nodePtr = nodePtr->Lower;
						goto TryNextNode;
					}
					else
					{
						nodePtr->Lower = nodeToTransfer;
					}
				}
			}
			else if(nodeToDelete->Lower)
			{
				*nodePtrPtr = nodeToDelete->Lower;
			}
			else
			{
				*nodePtrPtr = nodeToDelete->Higher;
			}

			iOutputIndex = nodeToDelete->Index;

			//get rid of the node ,and update the counters
			NumberOfNodes--;
			delete nodeToDelete;

			return true;
		}

		bool GreaterThan( const Asura_String& xString1, const char* szString2 )
		{
			if(m_bIgnoreExtensions)
			{
				return xString1.GreaterThanIgnoringExtension( szString2 );
			}
			else
			{
				return xString1.Compare(szString2) > 0;
			}
		}

		bool LessThan( const Asura_String& xString1, const char* szString2 )
		{
			if(m_bIgnoreExtensions)
			{
				return xString1.LessThanIgnoringExtension( szString2 );
			}
			else
			{
				return xString1.Compare(szString2) < 0;
			}
		}

		bool Equals( const Asura_String& xString1, const char* szString2 )
		{
			if(m_bIgnoreExtensions)
			{
				return xString1.EqualsIgnoringExtension( szString2 );
			}
			else
			{
				return xString1.Compare(szString2) == 0;
			}
		}

		int GetSizeOfNamesInIntChunksRecursive( const Asura_NameTree_Node* pxNode ) const
        {
            if( !pxNode ) { return 0; }

            return( (pxNode->Name.GetSizeInInts() << 2)
                    +
                    GetSizeOfNamesInIntChunksRecursive(pxNode->Lower)
                    +
                    GetSizeOfNamesInIntChunksRecursive(pxNode->Higher) );
        }

		Asura_NameTree_Node *FirstRoot;
		Asura_Handle_List *HandleList;

		int NumberOfNodes;
		int MaximumNodes;
		
protected :
		bool m_bIgnoreExtensions;
		
};

class Asura_StringTree : public Asura_NameTree
{
public :
	Asura_StringTree()
	:Asura_NameTree()
	{
		m_bIgnoreExtensions = false;
	}	
};

#endif//__ASURA_NAMETREE_H__
