//
// Asura_Spline.h
//
// Created by KOR - 18/01/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// An Asura_Spline consists of a series of templated points along with methods
// to access the length of the spline and the value at any given point along it.

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_SPLINE_H__
#define __ASURA_SPLINE_H__

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

class Asura_Quat;
class Asura_RGBA;
class Asura_Vector_3;

// Flags are stored in a u_char for each point. I'm reserving the bottom four bits for Asura_Spline,
// then anything else can hijack the other four bits for their own purposes.
enum Asura_Spline_Point_Flags
{
	Asura_Spline_Point_Flag_Straight = 0x01,
	Asura_Spline_Point_Flag_First_Unreserved_Flag = 0x10,
};

static const u_int uASURA_SPLINE_NUM_INTERMEDIATE_POINTS = 20;

template <typename T> inline void AddArrayElement(T*& pxArray, u_int uNumElements, const T& xValue, u_int uInsertAfter)
{
	T* pxNewArray = NULL;
	Asura_ResourceAllocateArray(pxNewArray, uNumElements + 1);
	for(u_int u = 0; u < uNumElements + 1; ++u)
	{
		if(u <= uInsertAfter)
		{
			pxNewArray[u] = pxArray[u];
		}
		else if(u > uInsertAfter + 1)
		{
			pxNewArray[u] = pxArray[u - 1];
		}
		else
		{
			pxNewArray[u] = xValue;
		}
	}
	Asura_ResourceDeallocateArray(pxArray);
	pxArray = pxNewArray;
}

template <typename T> inline void RemoveArrayElement(T*& pxArray, u_int uNumElements, u_int uIndex)
{
	T* pxNewArray = NULL;
	Asura_ResourceAllocateArray(pxNewArray, uNumElements - 1);
	for(u_int u = 0; u < uNumElements - 1; ++u)
	{
		if(u < uIndex)
		{
			pxNewArray[u] = pxArray[u];
		}
		else
		{
			pxNewArray[u] = pxArray[u + 1];
		}
	}
	Asura_ResourceDeallocateArray(pxArray);
	pxArray = pxNewArray;
}

template <class T> class Asura_Spline
{
public:
	Asura_Spline()
	: m_uNumberOfPoints(0)
	, m_pxPositions(NULL)
	, m_pucSectionFlags(NULL)
	, m_pfCumulativeLengths(NULL)
	, m_fInverseTotalLength(0.f)
	{
	}

	// Note that unlike SetPointPositions, this takes the position of the control points as well.
	// The points are deep copied.
	Asura_Spline(u_int uNumPoints, T* pxPoints)
	: m_uNumberOfPoints(uNumPoints)
	, m_pxPositions(NULL)
	, m_pucSectionFlags(NULL)
	, m_pfCumulativeLengths(NULL)
	, m_fInverseTotalLength(0.f)
	{
		Asura_ResourceAllocateArray(m_pxPositions, m_uNumberOfPoints);
		Asura_ResourceAllocateArray(m_pfCumulativeLengths, uNumPoints - 2);
		Asura_ResourceAllocateArray(m_pucSectionFlags, uNumPoints - 2);

		// Copy the points into the array
		for(u_int u = 0; u < uNumPoints; ++u)
		{
			m_pxPositions[u] = pxPoints[u];
		}
		// Zero other properties
		for(u_int u = 0; u < uNumPoints - 2; ++u)
		{
			m_pfCumulativeLengths[u] = 0.f;
			m_pucSectionFlags[u] = 0;
		}
		CalculateLengths();
	}

	virtual ~Asura_Spline()
	{
		Asura_ResourceDeallocateArray(m_pxPositions);
		Asura_ResourceDeallocateArray(m_pfCumulativeLengths);
		Asura_ResourceDeallocateArray(m_pucSectionFlags);
	}

	//assignment operator
	Asura_Spline& operator = (const Asura_Spline& xOriginal)
	{
		if(this != &xOriginal)
		{
			CopyFrom(&xOriginal);
		}
		return *this;
	}

	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream)
	{
		u_int uVersion;
		xStream >> uVersion;
		Asura_Assert(uVersion <= CHUNK_VERSION, "Unrecognised chunk version");

#ifdef ASURA_ASSERT
		const u_int uOldNumberOfPoints = m_uNumberOfPoints;
#endif
		xStream >> m_uNumberOfPoints;
		Asura_Assert(( m_uNumberOfPoints == uOldNumberOfPoints ) || ( uOldNumberOfPoints == 0 ), "Number of track points has changed since last load." );
		Asura_Assert(m_uNumberOfPoints >= 3, "Should be at least 3 control points for spline");
		const u_int uNumberOfPositionsIgnoringEnds = m_uNumberOfPoints - 2;
		if(!m_pxPositions)
		{
			Asura_ResourceAllocateArray(m_pxPositions, m_uNumberOfPoints);
		}
		// Length and flags don't need the end points.
		if(!m_pfCumulativeLengths)
		{
			Asura_ResourceAllocateArray(m_pfCumulativeLengths, uNumberOfPositionsIgnoringEnds);
		}
		if(!m_pucSectionFlags)
		{
			Asura_ResourceAllocateArray(m_pucSectionFlags, uNumberOfPositionsIgnoringEnds);
		}

		Asura_ReadFromStream_Array_Wrapper<T> xPositionArray(m_uNumberOfPoints, m_pxPositions);
		xStream >> xPositionArray;
		Asura_ReadFromStream_Array_Wrapper<float> xLengthsArray(uNumberOfPositionsIgnoringEnds, m_pfCumulativeLengths);
		xStream >> xLengthsArray;
		Asura_ReadFromStream_Array_Wrapper<u_char> xFlagsArray(uNumberOfPositionsIgnoringEnds, m_pucSectionFlags);
		xStream >> xFlagsArray;

		xStream >> m_fInverseTotalLength;
	}
	
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const
	{
		xStream << CHUNK_VERSION;
		xStream << m_uNumberOfPoints;

		const u_int uNumberOfPositionsIgnoringEnds = m_uNumberOfPoints - 2;
		Asura_WriteToStream_Array_Wrapper<Asura_Vector_3> xPositionArray(m_uNumberOfPoints, m_pxPositions);
		xStream << xPositionArray;
		Asura_WriteToStream_Array_Wrapper<float> xLengthsArray(uNumberOfPositionsIgnoringEnds, m_pfCumulativeLengths);
		xStream << xLengthsArray;
		Asura_WriteToStream_Array_Wrapper<u_char> xFlagsArray(uNumberOfPositionsIgnoringEnds, m_pucSectionFlags);
		xStream << xFlagsArray;

		xStream << m_fInverseTotalLength;
	}

	float GetCumulativeLength(u_int uIndex) const
	{
		Asura_Assert( uIndex < m_uNumberOfPoints - 2, "Index out of range" );
		return m_pfCumulativeLengths[uIndex];
	}

	float GetTotalLength() const
	{
		return GetCumulativeLength(m_uNumberOfPoints - 3);
	}

	// This returns the actual distance used in the calculation (clamped to the valid range along the spline).
	float GetPositionFromDistance(float fDistance, T& xOutput) const
	{
		u_int uSection = 0;
		float fLerp = 0.f;
		GetLerpAndSectionFromDistance(fDistance, fLerp, uSection);
		GetPositionInSection(uSection, fLerp, xOutput);

		return fDistance;
	}

	// This returns the actual distance used in the calculation (clamped to the valid range along the spline).
	float GetTangentFromDistance(float fDistance, T& xOutput) const
	{
		u_int uSection = 0;
		float fLerp = 0.f;
		GetLerpAndSectionFromDistance(fDistance, fLerp, uSection);
		GetTangentInSection(uSection, fLerp, xOutput);

		return fDistance;
	}

	float GetProportionAlongSpline(float fDistance) const
	{
		return fDistance * m_fInverseTotalLength;
	}

	u_int GetNumberOfPoints() const
	{
		return m_uNumberOfPoints;
	}

	const T& GetPointPosition(u_int uIndex) const
	{
		Asura_Assert(uIndex < m_uNumberOfPoints, "Index out of range");
		return m_pxPositions[uIndex];
	}

	u_char GetSectionFlags(u_int uIndex) const
	{
		Asura_Assert(uIndex < m_uNumberOfPoints - 2, "Index out of range");
		return m_pucSectionFlags[uIndex];
	}

	void RenderLines(const Asura_RGBA& xBaseColour, const Asura_RGBA& xCustomColour, const bool bHighLOD = true);

	void CopyFrom(const Asura_Spline* pxFrom)
	{
		// Clean up anything we have around.
		Asura_ResourceDeallocateArray(m_pxPositions);
		Asura_ResourceDeallocateArray(m_pfCumulativeLengths);
		Asura_ResourceDeallocateArray(m_pucSectionFlags);

		m_uNumberOfPoints = pxFrom->m_uNumberOfPoints;
		const u_int uNumberOfPositionsIgnoringEnds = m_uNumberOfPoints - 2;
		Asura_ResourceAllocateArray(m_pxPositions, m_uNumberOfPoints);
		Asura_ResourceAllocateArray(m_pfCumulativeLengths, uNumberOfPositionsIgnoringEnds);
		Asura_ResourceAllocateArray(m_pucSectionFlags, uNumberOfPositionsIgnoringEnds);
		for(u_int u = 0; u < uNumberOfPositionsIgnoringEnds; ++u)
		{
			m_pxPositions[u + 1] = pxFrom->m_pxPositions[u + 1];
			m_pfCumulativeLengths[u] = pxFrom->m_pfCumulativeLengths[u];
			m_pucSectionFlags[u] = pxFrom->m_pucSectionFlags[u];
		}
		// End points.
		m_pxPositions[0] = pxFrom->m_pxPositions[0];
		m_pxPositions[m_uNumberOfPoints - 1] = pxFrom->m_pxPositions[m_uNumberOfPoints - 1];
	}

	void CalculateLengths()
	{
		if(m_uNumberOfPoints <= 3) { return; }
		//determine the section lengths
		const u_int uNumSections = m_uNumberOfPoints - 2;
		if(!m_pfCumulativeLengths)
		{
			Asura_ResourceAllocateArray(m_pfCumulativeLengths, uNumSections);
		}

		float fTrackLength = 0.f;
		m_pfCumulativeLengths[0] = 0.f;
		for(u_int uSection = 0; uSection < uNumSections - 1; ++uSection)
		{
			float fLength = 0.0f;

			if (m_pucSectionFlags[uSection] & Asura_Spline_Point_Flag_Straight)
			{
				fLength = (m_pxPositions[uSection + 2] - m_pxPositions[uSection + 1]).Magnitude();
			}
			else
			{
				// to approximate distance along curve, divide into 20 straight 
				// segments and add up the lengths of the segments.
				T axSegmentPoints[uASURA_SPLINE_NUM_INTERMEDIATE_POINTS + 1];

				for (u_int uSubSection = 0; uSubSection <= uASURA_SPLINE_NUM_INTERMEDIATE_POINTS; ++uSubSection)
				{
					float f = static_cast<float>(uSubSection) / (float)uASURA_SPLINE_NUM_INTERMEDIATE_POINTS;
					GetPositionInSection(uSection, f, axSegmentPoints[uSubSection]);
				}

				for (u_int uSubSection = 0; uSubSection < uASURA_SPLINE_NUM_INTERMEDIATE_POINTS; ++uSubSection)
				{
					fLength += (axSegmentPoints[uSubSection] - axSegmentPoints[uSubSection + 1]).Magnitude();
				}
			}

			fTrackLength += fLength;

			m_pfCumulativeLengths[uSection+1] = fTrackLength;
		}
		m_fInverseTotalLength = (fTrackLength > 0.f) ? 1.f / fTrackLength : 0.f;
	}

	// It is assumed that these will not be modified outside tools.
	// If it turns out that it would be useful in some situation, feel free
	// to remove these guards, but bear in mind that it will result in dynamic allocations
#ifdef ASURA_TOOLS
	virtual void SetPointPositions(u_int uNumberOfPoints, const T* pxPoints)
	{
		Asura_Assert(uNumberOfPoints >= 1 && pxPoints, "No points");
		if(!uNumberOfPoints || !pxPoints) { return; }

		// Clean up old versions
		Asura_ResourceDeallocateArray(m_pxPositions);
		Asura_ResourceDeallocateArray(m_pfCumulativeLengths);
		Asura_ResourceDeallocateArray(m_pucSectionFlags);

		m_uNumberOfPoints = uNumberOfPoints + 2;
		Asura_ResourceAllocateArray(m_pxPositions, m_uNumberOfPoints);
		Asura_ResourceAllocateArray(m_pfCumulativeLengths, uNumberOfPoints);
		Asura_ResourceAllocateArray(m_pucSectionFlags, uNumberOfPoints);

		// Copy the points into the array
		for(u_int u = 0; u < uNumberOfPoints; ++u)
		{
			m_pxPositions[u + 1] = pxPoints[u];
			m_pfCumulativeLengths[u] = 0.f;
			m_pucSectionFlags[u] = 0;
		}
		OnChanged();
	}

	T& GetPointPosition(u_int uIndex)
	{
		Asura_Assert(uIndex < m_uNumberOfPoints, "Index out of range");
		return m_pxPositions[uIndex];
	}

	void SetPointPosition(u_int uIndex, const T& xPos)
	{
		Asura_Assert(uIndex < m_uNumberOfPoints, "Index out of range");
		m_pxPositions[uIndex] = xPos;
		OnChanged();
	}

	void SetSectionFlags(u_int uIndex, const u_char ucFlags)
	{
		Asura_Assert(uIndex < m_uNumberOfPoints - 2, "Index out of range");
		m_pucSectionFlags[uIndex] = ucFlags;
		OnChanged();
	}

	void AddSectionFlag(u_int uIndex, const u_char uFlag)
	{
		Asura_Assert(uIndex < m_uNumberOfPoints - 2, "Index out of range");
		m_pucSectionFlags[uIndex] |= uFlag;
		OnChanged();
	}

	void RemoveSectionFlag(u_int uIndex, const u_char uFlag)
	{
		Asura_Assert(uIndex < m_uNumberOfPoints - 2, "Index out of range");
		m_pucSectionFlags[uIndex] &= ~uFlag;
		OnChanged();
	}

	virtual void CalculateEndControlPoints()
	{
		Asura_Assert(m_pxPositions, "No positions");
		if(m_uNumberOfPoints < 4) { return; }

		m_pxPositions[0] = m_pxPositions[1] * 2.0f - m_pxPositions[2];
		m_pxPositions[m_uNumberOfPoints - 1] = m_pxPositions[m_uNumberOfPoints - 2] * 2.0f - m_pxPositions[m_uNumberOfPoints-3];
	}
	
	virtual void AddPoint(const u_int uIndexAfter, const T& xPosition)
	{
		Asura_Assert( 0 < uIndexAfter && uIndexAfter < m_uNumberOfPoints - 2, "Index out of range" );
		AddArrayElement<Asura_Vector_3>(m_pxPositions, m_uNumberOfPoints, xPosition, uIndexAfter);
		AddArrayElement<u_char>(m_pucSectionFlags, m_uNumberOfPoints - 2, 0, uIndexAfter - 1);
		++m_uNumberOfPoints;

		// Just delete the lengths, they will be reallocated by CalculateLengths
		Asura_ResourceDeallocateArray(m_pfCumulativeLengths);

		OnChanged();
	}
	
	virtual void RemovePoint(const u_int uIndex)
	{
		Asura_Assert( m_uNumberOfPoints >= 4, "Must leave at least 3 control points in a spline" );
		if ( m_uNumberOfPoints <= 3)
		{
			return;
		}
		RemoveArrayElement<Asura_Vector_3>(m_pxPositions, m_uNumberOfPoints, uIndex);
		RemoveArrayElement<u_char>(m_pucSectionFlags, m_uNumberOfPoints - 2, uIndex);
		--m_uNumberOfPoints;

		// Just delete the lengths, they will be reallocated by CalculateLengths
		Asura_ResourceDeallocateArray(m_pfCumulativeLengths);

		OnChanged();
	}

	virtual void ReverseSpline()
	{
		const u_int uNumPoints = GetNumberOfPoints() - 2;
		Asura_Vector_3* pxPoints = NULL;
		Asura_ResourceAllocateArray( pxPoints, uNumPoints );
		u_int uArrayIndex = uNumPoints;
		for ( u_int u = 0; u < uNumPoints; u++ )
		{
			pxPoints[u] = GetPointPosition( uArrayIndex );
			--uArrayIndex;
		}
		SetPointPositions( uNumPoints, pxPoints );
		Asura_ResourceDeallocateArray( pxPoints );
	}

	virtual void OnChanged()
	{
		CalculateEndControlPoints();
		CalculateLengths();
	}

	virtual void ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
	{
		const u_int uNumPoints = GetNumberOfPoints();
		for ( u_int u = 0; u < uNumPoints; u++ )
		{
			SetPointPosition( u, pxInstance->TransformPosition( GetPointPosition( u ) ) );
		}
	}
#endif // ASURA_TOOLS

protected:
	void GetPositionInSection(u_int uSection, float fLerpValue, T& xOutput) const
	{
		const bool bStraightBefore = ( uSection > 0 ) ? ( m_pucSectionFlags[ uSection - 1 ] & Asura_Spline_Point_Flag_Straight ) != 0 : false;
		const bool bStraight	   = ( m_pucSectionFlags[ uSection ] & Asura_Spline_Point_Flag_Straight ) != 0;
		const bool bStraightAfter  = ( uSection < ( m_uNumberOfPoints - 3 ) ) ? ( m_pucSectionFlags[ uSection + 1 ] & Asura_Spline_Point_Flag_Straight ) != 0 : false;
		// Section 0 is the spline between positions 1 and 2, etc.
		Asura_CalcCRInterpolatedPosition(
			m_pxPositions[ uSection ], m_pxPositions[ uSection + 1 ], m_pxPositions[ uSection + 2 ], m_pxPositions[ uSection + 3 ],
			fLerpValue,
			bStraightBefore, bStraight, bStraightAfter, xOutput );
	}
	void GetTangentInSection(u_int uSection, float fLerpValue, T& xOutput) const
	{
		const bool bStraightBefore = ( uSection > 0 ) ? ( m_pucSectionFlags[ uSection - 1 ] & Asura_Spline_Point_Flag_Straight ) != 0 : false;
		const bool bStraight	   = ( m_pucSectionFlags[ uSection ] & Asura_Spline_Point_Flag_Straight ) != 0;
		const bool bStraightAfter  = ( uSection < ( m_uNumberOfPoints - 3 ) ) ? ( m_pucSectionFlags[ uSection + 1 ] & Asura_Spline_Point_Flag_Straight ) != 0 : false;
		// Section 0 is the spline between positions 1 and 2, etc.
		Asura_CalcCRTangent(
			m_pxPositions[ uSection ], m_pxPositions[ uSection + 1 ], m_pxPositions[ uSection + 2 ], m_pxPositions[ uSection + 3 ],
			fLerpValue,
			bStraightBefore, bStraight, bStraightAfter, xOutput );
	}

	// Gets the section and proportion along it from distance
	float GetLerpAndSectionFromDistance(float fDistance, float& fLerp, u_int& uSection) const
	{
		Asura_Assert(m_uNumberOfPoints >= 4,"Shouldn't be using this function on single point tracks");

		//make sure the distance is in range
		const float fActualDistance = Asura_Maths::ClampToRange(fDistance, 0.f, GetTotalLength());

		//find the section that we are on
		for(uSection = 0; uSection < m_uNumberOfPoints - 2; ++uSection)
		{
			if(m_pfCumulativeLengths[uSection + 1] >= fActualDistance)
			{
				break;
			}
		}

		//get the proportion along the section
		const float fSectionStart = m_pfCumulativeLengths[uSection];
		const float fSectionLength = m_pfCumulativeLengths[uSection+1] - fSectionStart;
		if(fSectionLength > 0.f && fActualDistance > fSectionStart)
		{
			fLerp = (fActualDistance - fSectionStart) / fSectionLength;
		}
		else
		{
			fLerp = 0.f;
		}
		return fActualDistance;
	}

	bool RenderSection( const bool bHighLOD,
						const Asura_Vector_3& xP0,
						const Asura_Vector_3& xD0Tangent,
						const float fD0,
						const float fD1,
						Asura_Vector_3& xP1,
						Asura_Vector_3& xD1Tangent );
	void CalcColours( const Asura_RGBA& xBaseColour, const Asura_RGBA& xCustomColour );

	enum { NUM_COLOURS = 16 };
	static Asura_RGBA s_xBaseColour;
	static Asura_RGBA s_xCustomColour;
	static u_int s_auSplineColours[ NUM_COLOURS ];


	enum { CHUNK_VERSION = 0 };	

	// Number of points including two extra control points for the ends.
	u_int	m_uNumberOfPoints;
	// Position and flags are control data.
	T*		m_pxPositions;
	u_char* m_pucSectionFlags;
	// Lengths are derived from control data.
	float*	m_pfCumulativeLengths;
	float	m_fInverseTotalLength;
};

class Asura_Oriented_Spline : public Asura_Spline<Asura_Vector_3>
{
public:
	Asura_Oriented_Spline();
	virtual ~Asura_Oriented_Spline();

	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	float GetPosAndOriFromDistance(float fDistance, Asura_Vector_3& xPosition, Asura_Quat& xOrientation) const;
	const Asura_Quat& GetPointOrientation(u_int uIndex) const;

	void CopyFrom(const Asura_Oriented_Spline* pxFrom);

#ifdef ASURA_TOOLS
	virtual void SetPointPositions(u_int uNumberOfPoints, const Asura_Vector_3* pxPoints);
	virtual void AddPoint(const u_int uIndexAfter, const Asura_Vector_3& xPosition);
	virtual void AddPointWithOri(const u_int uIndexAfter, const Asura_Vector_3& xPos, const Asura_Quat& xOri);
	virtual void RemovePoint(const u_int uIndex);
	virtual void ReverseSpline();

	Asura_Quat& GetPointOrientation(u_int uIndex);
	void SetPointOrientation(u_int uIndex, const Asura_Quat& xQuat);

	virtual void ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance);
#endif

protected:
	enum { CHUNK_VERSION = 0 };	

	Asura_Quat* m_pxOrientations;

	typedef Asura_Spline<Asura_Vector_3> PARENT;
};

template< typename T >
void Asura_CalcCRInterpolatedPosition( const T& xP0, const T& xP1, const T& xP2, const T& xP3, const float fLerpValue,
									   const bool bStraightBefore, const bool bStraight, const bool bStraightAfter,
									   T& xOutput )
{
	Asura_Assert( 0.0f <= fLerpValue && fLerpValue <= 1.0f, "Asura_CalcCRInterpolatedPosition - interpolation parameter out of range" );

	if ( bStraight )
	{
		xOutput = xP1 * ( 1.0f - fLerpValue ) + xP2 * fLerpValue;
		return;
	}

	const float fLerpSquared = fLerpValue * fLerpValue;
	const float fLerpCubed = fLerpSquared * fLerpValue;

	if ( bStraightBefore )
	{
		if ( bStraightAfter )
		{
			// Straight before and after.
			xOutput = xP0 * ( -0.5f * fLerpCubed + fLerpSquared - 0.5f * fLerpValue ) +
				xP1 * ( 2.5f * fLerpCubed - 4.0f * fLerpSquared + 0.5f * fLerpValue + 1.0f ) +
				xP2 * ( -2.5f * fLerpCubed + 3.5f * fLerpSquared ) +
				xP3 * ( 0.5f * fLerpCubed - 0.5f * fLerpSquared );
		}
		else
		{
			// Just straight before.
			xOutput = xP0 * ( -0.5f * fLerpCubed + fLerpSquared - 0.5f * fLerpValue ) +
				xP1 * ( 2.0f * fLerpCubed - 3.5f * fLerpSquared + 0.5f * fLerpValue + 1.0f ) +
				xP2 * ( -2.0f * fLerpCubed + 3.0f * fLerpSquared ) +
				xP3 * ( 0.5f * fLerpCubed - 0.5f * fLerpSquared );
		}
	}
	else if ( bStraightAfter )
	{
		// Just straight after.
		xOutput = xP0 * ( -0.5f * fLerpCubed + fLerpSquared - 0.5f * fLerpValue ) +
			xP1 * ( 2.0f * fLerpCubed - 3.0f * fLerpSquared + 1.0f ) +
			xP2 * ( -2.0f * fLerpCubed + 2.5f * fLerpSquared + 0.5f * fLerpValue ) +
			xP3 * ( 0.5f * fLerpCubed - 0.5f * fLerpSquared );
	}
	else
	{
		// Normal.
		xOutput = xP0 * ( -0.5f * fLerpCubed + fLerpSquared - 0.5f * fLerpValue ) +
			xP1 * ( 1.5f * fLerpCubed - 2.5f * fLerpSquared + 1.0f ) +
			xP2 * ( -1.5f * fLerpCubed + 2.0f * fLerpSquared + 0.5f * fLerpValue ) +
			xP3 * ( 0.5f * fLerpCubed - 0.5f * fLerpSquared );
	}
}

template< typename T >
void Asura_CalcCRTangent( const T& xP0, const T& xP1, const T& xP2, const T& xP3, const float fLerpValue,
						  const bool bStraightBefore, const bool bStraight, const bool bStraightAfter,
						  T& xOutput )
{
	Asura_Assert( 0.0f <= fLerpValue && fLerpValue <= 1.0f, "Asura_CalcCRTangent - interpolation parameter out of range" );

	if ( bStraight )
	{
		xOutput = xP2 - xP1;
		return;
	}

	const float fLerpSquared = fLerpValue * fLerpValue;

	if ( bStraightBefore )
	{
		if ( bStraightAfter )
		{
			// Straight before and after.
			xOutput = xP0 * ( -1.5f * fLerpSquared + 2.0f * fLerpValue - 0.5f ) +
				xP1 * ( 7.5f * fLerpSquared - 8.0f * fLerpValue + 0.5f ) +
				xP2 * ( -7.5f * fLerpSquared + 7.0f * fLerpValue ) +
				xP3 * ( 1.5f * fLerpSquared - fLerpValue );
		}
		else
		{
			// Just straight before.
			xOutput = xP0 * ( -1.5f * fLerpSquared + 2.0f * fLerpValue - 0.5f ) +
				xP1 * ( 6.0f * fLerpSquared - 7.0f * fLerpValue + 0.5f ) +
				xP2 * ( -6.0f * fLerpSquared + 6.0f * fLerpValue ) +
				xP3 * ( 1.5f * fLerpSquared - fLerpValue );
		}
	}
	else if ( bStraightAfter )
	{
		// Just straight after.
		xOutput = xP0 * ( -1.5f * fLerpSquared + 2.0f * fLerpValue - 0.5f ) +
			xP1 * ( 6.0f * fLerpSquared - 6.0f * fLerpValue ) +
			xP2 * ( -6.0f * fLerpSquared + 5.0f * fLerpValue + 0.5f ) +
			xP3 * ( 1.5f * fLerpSquared - fLerpValue );
	}
	else
	{
		// Normal.
		xOutput = xP0 * ( -1.5f * fLerpSquared + 2.0f * fLerpValue - 0.5f ) + 
			xP1 * ( 4.5f * fLerpSquared - 5.0f * fLerpValue ) +
			xP2 * ( -4.5f * fLerpSquared + 4.0f * fLerpValue + 0.5f ) +
			xP3 * ( 1.5f * fLerpSquared - fLerpValue );
	}
}

#endif//__ASURA_SPLINE_H__
