//
// Asura_String.h
//
// Created by Kevin Lea - 12/06/2001 11:00:49
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_STRING_H__
#define __ASURA_STRING_H__

// KJL - 14-Feb-01 14:56:35
// 
// Asura_String.h
//
//    There are two distinct categories of text in Asura - 
//
// 1. things which need to be displayed to the player, and thus need to be localised
// 2. internal strings, such as filenames, Assert text descriptions, etc..
//
//    I had originally planned to write a Unicode string class and use it everywhere,
// but on reflection it seemed that only the first category of text needed to be
// in Unicode, since it would just be waste of memory to store things like filenames
// in 16bit. Furthermore, only the 'internal' strings need to be manipulated (e.g. 
// comparing filenames, extracting substrings) whilst the Unicode strings would just
// be used as is. (The only exception I can think of are strings entered by the
// player, such as the player's name.)
//
//    So, I've made two classes Asura_String (ANSI) and Asura_UString (Unicode),
// with Asura_String containing lots of useful overloading. Any functions which output
// or display text to the player *must* operate with Asura_UString.

// Stdlib independent.
#include <ctype.h>
#include <limits.h>
#include <stdarg.h>
#include <string.h>

// Engine independent.
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_Hash.h"
#include "Core/Memory/Asura_Mem.h"

class Asura_String;

class Asura_StringUtil
{

    public:

        enum
        {
            STRIP_LEADING   = 0x1,
            STRIP_TRAILING  = 0x2,
            STRIP_BOTH_ENDS = STRIP_LEADING | STRIP_TRAILING,
        };

        // C L A S S   M E T H O D S /////////////////////////////////////////

        // Dest value will not be altered if format is invalid (ie. if the
        // function returns false.)
        static bool GetBoolValue( const Asura_Char* szToConvert, bool* pbValue );
        static bool GetFloatValue( const Asura_Char* szToConvert, float* pfValue );
        static bool GetIntValue( const Asura_Char* szToConvert, int* piValue, bool bForceHex = false );
        // And wide-character versions.
        static bool GetBoolValue( const Asura_UnicodeChar* wszToConvert, bool* pbValue );
        static bool GetFloatValue( const Asura_UnicodeChar* wszToConvert, float* pfValue );
        static bool GetIntValue( const Asura_UnicodeChar* wszToConvert, int* piValue, bool bForceHex = false );

        static void ToString( int iValue, Asura_String& xString );
        static void ToString( u_int uValue, Asura_String& xString );
        static void ToString( float fValue, Asura_String& xString );

        static void Reverse( Asura_Char* szString );

        // Removes leading and/or trailing whitespace from the specified
        // string. Modifies the buffer passed. Control exactly what's
        // stripped by passing a bitwise-ORing of STRIP_... values from the
        // enum above. Non-const versions modifies string in place. Const
        // versions simply return information about changes required to
        // remove quotes.
        static Asura_Char* StripWhitespace( Asura_Char* szBfr, u_int uStripFlags );
        static const Asura_Char* StripWhitespace( const Asura_Char* szBfr, u_int uStripFlags, u_int* puNewLen );
        static Asura_UnicodeChar* StripWhitespace( Asura_UnicodeChar* wszBfr, u_int uStripFlags );
        static const Asura_UnicodeChar* StripWhitespace( const Asura_UnicodeChar* wszBfr, u_int uStripFlags, u_int* puNewLen );

        // Creates a copy of the specified string. Allocates memory via new[].
        static Asura_Char* Duplicate( const Asura_Char* szBfr );
        static Asura_UnicodeChar* Duplicate( const Asura_UnicodeChar* wszBfr );

        // Non-const versions modifies string in place. Const versions simply
        // return information about changes required to remove quotes. These
        // methods assume quotes to be stripped are at the extremes of the
        // string, and in particular that there is no leading or trailing
        // whitespace.
        static Asura_Char* StripQuotes( Asura_Char* wszBfr );
        static const Asura_Char* StripQuotes( const Asura_Char* wszBfr, u_int* puNewLen );
        static Asura_UnicodeChar* StripQuotes( Asura_UnicodeChar* wszBfr );
        static const Asura_UnicodeChar* StripQuotes( const Asura_UnicodeChar* wszBfr, u_int* puNewLen );

		// Removes the file extension from the specified string - returning whether we did anything or not...
		static bool StripExtension(Asura_Char* szString);

        // Creates a new string, via new[], with the specified substring
        // replaced with the new text given.
        static Asura_Char* Replace( const Asura_Char* szOriginal, const Asura_Char* szToReplace, const Asura_Char* szReplacementText );
        static Asura_Char* Replace( const Asura_Char* szOriginal, u_int uToReplaceStart, u_int uToReplaceLen, const Asura_Char* szReplacementText );
        
		// JAT 080206 - Replace all instances of the specified character with a different character e.g. change all '/' -> '\'
		static void ReplaceAllChars( char* const szString, const Asura_Char cSearch, const Asura_Char cReplace );

        static bool StartsWith( const Asura_Char* szString, const Asura_Char* szPrefix, bool bIgnoreCase );
        static bool EndsWith( const Asura_Char* szString, const Asura_Char* szSuffix, bool bIgnoreCase );

        // Returns true iff the string matches the specified wildcard string.
        // The wildcard characters supported are as in Windows: * and ?.
        static bool CompareAgainstWildcardString( const Asura_Char* szString, const Asura_Char* szWildcard, bool bIgnoreCase );

		// Return true iff the string contains the other string.
		static bool Contains( const Asura_Char* szString, const Asura_Char* szFindString, int* piPositionInString = NULL );
		static bool Contains( const Asura_UnicodeChar* wszString, const Asura_UnicodeChar* wszFindString, int* piPositionInString = NULL );

        // In-place conversion.
        static void MakeLowercase( Asura_Char* szString );
        static void MakeUppercase( Asura_Char* szString );

		// ARC - 23 February 2007 10:52:29 - Find the indexed string part within the passed string (if it exists)
		// e.g. given the string "homer.homerball.speed", and a separator of ".", index 0 will find "homer", index 1 will find "homerball", etc
		static bool FindStringPartByIndex(const Asura_UnicodeChar* wszString, const Asura_UnicodeChar wcSeparator, u_int uStringPartIndex, u_int& uStringPartStart, u_int& uStringPartLength);
		static bool FindStringPartByIndex(const Asura_Char* szString, const Asura_Char cSeparator, u_int uStringPartIndex, u_int& uStringPartStart, u_int& uStringPartLength);

		static u_int CountInstancesOfCharacter(const Asura_UnicodeChar* wszString, const Asura_UnicodeChar wcCharacter);
		static u_int CountInstancesOfCharacter(const Asura_Char* szString, const Asura_Char cCharacter);
};

class Asura_String
{
	public:
        			
		// constructors and destructors
		Asura_String()
		{
			Buffer = 0;
			LengthInCharacters = 0;
			MemoryAllocatedInBytes = 0;
		}

		Asura_String(const Asura_String& string)
		{	
			AssignFrom(string);
		}
		Asura_String(const Asura_Char* string)
		{
			AssignFrom(string);
		}
		Asura_String(const Asura_Char* string, unsigned int uLen)
		{
			Buffer = 0;
			LengthInCharacters = 0;
			MemoryAllocatedInBytes = 0;

			if (string && uLen)
			{
				Append(string, uLen);
			}
		}

		~Asura_String()
		{
			MakeNull();
		}

		// accessors
		Asura_Char* GetString( void )
		{
			return Buffer;
		}
        const Asura_Char* GetString( void ) const
        {
            return Buffer;
        }
		int* IntCopy(int* dest) const
		{
            int* piRetVal = dest;

            if( Buffer )
            {
			    int namelength = GetSizeInInts();
			    for(int i=0; i<namelength;i++)
			    {
				    dest[i] = 0;
			    }
			    strcpy((char*)dest,Buffer);

			    piRetVal += namelength;
            }
			else
			{
				dest[0] = NULL_STRING_CHUNK;
				piRetVal ++;
			}

            return piRetVal;
		}

		int GetLength(void) const
		{
			return LengthInCharacters;
		}
			
		int GetSizeInBytes(void) const
		{
			return MemoryAllocatedInBytes;
		}

		int GetSizeInInts(void) const
		{
			if(IsNull())
			{
				return 1;
			}
			else
			{
				return ((GetSizeInBytes()+3)/4);
			}
		}

		Asura_Hash_ID GetHash(void) const
		{
			return(Asura_GetHashID(Buffer));
		}

		Asura_Hash_ID GetHash_IgnoreExtension(void) const
		{
			return(Asura_GetHashID_IgnoreExtension(Buffer));
		}

		bool IsNull(void) const
		{
			return (MemoryAllocatedInBytes == 0);
		}

		void MakeNull(void)
		{
			if (Buffer)
			{
				delete[] Buffer;
    			Buffer = NULL;
			}
			MemoryAllocatedInBytes = 0;
			LengthInCharacters = 0;
		}

		/// Kevin Lea - 25/04/2002 17:36:49 - returns true if Buffer is null OR contains the string ""
		bool IsBlank(void) const
		{
			if (Buffer)
			{
				return (*Buffer==0);
			}

			return true;
		}

		void MakeBlank(void)
		{
			if (Buffer)
			{
				delete [] Buffer;
			}
			AssignFrom("");
		}

		bool IsValidIndex(const int index) const
		{
			return (index>=0 && index<LengthInCharacters);
		}

		Asura_Char& operator [](int index)
		{
			return Buffer[index];
		}
        const Asura_Char& operator[]( int i ) const
        {
            return Buffer[i];
        }
		
		// assignment
		Asura_String& operator =( const Asura_Char* szSource )
		{
            if( !Buffer || szSource != Buffer )
            {
			    MakeNull();
			    AssignFrom( szSource );
            }
            return *this;
		}

        //lint -e1529 // not first checking for assignment to this; we do in the Asura_Char* operator instead
		Asura_String& operator =( const Asura_String& xSource )
		{
            *this = xSource.Buffer;
            return *this;
		}
        //lint +e1529
	
// (MP - 02.08.2002) Removed corresponding AssignFrom() (see below) so this
// has to go too.
//		void operator =(const Asura_UnicodeChar* string)
//		{
//			MakeNull();
//			AssignFrom(string);
//		}

		void operator +=( const Asura_Char cSource )
        {
			Append( cSource );
		}
		void operator +=( const Asura_Char* szSource )
        {
            if (szSource)
			{
				Append( szSource, u_int( strlen(szSource)) );
			}
		}
		void operator +=( const Asura_String& xSource )
		{
            Append( xSource.Buffer, xSource.LengthInCharacters );
		}

		// JAT 15/05/07 - this assumes both strings have an extension so "jake.asr" won't match "jake"
		bool EqualsIgnoringExtension( const char* szOther ) const
		{
			for (int i=LengthInCharacters-1; i>=0; --i)
			{
				if (Buffer[i] == '.')
				{
					return !strnicmp( Buffer, szOther, i + 1 );
				}
			}
		
			return !stricmp( Buffer, szOther );
		}

		// JAT 15/05/07 - this assumes both strings have an extension so "jake.asr" won't match "jake"
		bool LessThanIgnoringExtension( const char* szOther ) const
		{
			for (int i=LengthInCharacters-1; i>=0; --i)
			{
				if (Buffer[i] == '.')
				{
					return (strnicmp(Buffer, szOther, i + 1 )<=-1); 
				}
			}
					
			return (stricmp(Buffer, szOther)<=-1);
		}
		// JAT 15/05/07 - this assumes both strings have an extension so "jake.asr" won't match "jake"
		bool GreaterThanIgnoringExtension( const char* szOther ) const
		{
			for (int i=LengthInCharacters-1; i>=0; --i)
			{
				if (Buffer[i] == '.')
				{
					return (strnicmp(Buffer, szOther, i + 1 )>=1);
				}
			}
					
			return (stricmp(Buffer, szOther)>=1);
		}

        bool EqualsIgnoringExtension( const Asura_String& xOther ) const { return EqualsIgnoringExtension( xOther.GetString() ); }
        bool LessThanIgnoringExtension( const Asura_String& xOther ) const { return LessThanIgnoringExtension( xOther.GetString() ); }
        bool GreaterThanIgnoringExtension( const Asura_String& xOther ) const { return GreaterThanIgnoringExtension( xOther.GetString() ); }

		// compare two strings - same return values as strcmp
        // ie. -1 if this <  string
        //      0 if this == string
        //     +1 if this >  string
		int Compare(const Asura_String& string, bool ignoreCase=true) const
		{
            return Compare( string.Buffer, ignoreCase );
		}
		
		int Compare(const Asura_Char* szStringBuffer, bool ignoreCase=true) const
		{
			// check for empty string cases first
			if( !Buffer || !(*Buffer) )
			{
				if( !szStringBuffer || !*szStringBuffer )
				{
					return 0;
				}
				else
				{
					return -1;
				}
			}
			else
			{
				if( !szStringBuffer || !*szStringBuffer )
				{
					return 1;
				}
			}

			if (ignoreCase)
			{
				return stricmp(Buffer,szStringBuffer);				
			}
			else
			{
				return strcmp(Buffer,szStringBuffer);
			}
		}
        
        bool operator < ( const Asura_Char* szBfr ) const { return( Compare(szBfr) <= -1 ); }
        bool operator > ( const Asura_Char* szBfr ) const { return( Compare(szBfr) >=  1 ); }
        bool operator ==( const Asura_Char* szBfr ) const { return( Compare(szBfr) ==  0 ); }
        bool operator !=( const Asura_Char* szBfr ) const { return( Compare(szBfr) !=  0 ); }

        bool operator < ( const Asura_String& xOther ) const { return( Compare(xOther.Buffer) <= -1 ); }
        bool operator > ( const Asura_String& xOther ) const { return( Compare(xOther.Buffer) >=  1 ); }
        bool operator ==( const Asura_String& xOther ) const { return( Compare(xOther.Buffer) ==  0 ); }
        bool operator !=( const Asura_String& xOther ) const { return( Compare(xOther.Buffer) !=  0 ); }

		bool StartsWith( const Asura_Char* szPrefix, bool bIgnoreCase = true ) const
		{
            return Asura_StringUtil::StartsWith( Buffer, szPrefix, bIgnoreCase );
        }

        bool EndsWith( const Asura_Char* szSuffix, bool bIgnoreCase = true ) const
        {
            return Asura_StringUtil::EndsWith( Buffer, szSuffix, bIgnoreCase );
        }
		
		bool Contains( const Asura_Char* szString, int* piPositionInString = NULL ) const
		{			
            return Asura_StringUtil::Contains( static_cast<char*>(Buffer), szString, piPositionInString );
		}

		void TidyFilename();

		bool SplitFilename(Asura_String* path, Asura_String* filename, Asura_String* extension) const
		{			
			if( IsNull() ) return false; // don't split if there's nothing to split

			Asura_String workspace = *this;
			int p;
			// grab extension
			{
                for(p = workspace.GetLength(); workspace[p]!='.' && p!=0; p--) { }

				if (workspace[p]=='.')
				{
					workspace[p] = 0;

					if (extension)
					{
						*extension = (&workspace[p+1]);
					}
				}
				else if (extension)
				{
					extension->MakeNull();
				}
			}

			// grab filename
			{
                for(p = workspace.GetLength(); (workspace[p]!='\\' && workspace[p]!='/') && p!=0; p--) { }

				if (filename)
				{
					// (AAM - 15.11.2006) Used not to check for the slashes here, but that resulted in
					// inputs of the form "\\foo.bar" producing a filename of "\\foo" rather than "foo"

					if (p==0 && workspace[p] != '\\' && workspace[p] != '/')
					{
						*filename = (&workspace[0]);
					}
					else
					{
						*filename = (&workspace[p+1]);
					}

				}

				if (workspace[p]=='\\' || workspace[p]=='/')
				{
					// there should be a path too
					if (path)
					{		
						workspace[p] = 0;
						*path=(&workspace[0]);
					}
				}
			
			}
			return true;

		}

        void MakeLowercase( void )
        {
            Asura_StringUtil::MakeLowercase( Buffer );
        }

        void MakeUppercase( void )
        {
            Asura_StringUtil::MakeUppercase( Buffer );
        }

        void Reverse()
        {
            Asura_StringUtil::Reverse( Buffer );
        }

        void Replace( const Asura_Char* szToReplace, const Asura_Char* szReplacementText )
        {
            Asura_Char* szNew = Asura_StringUtil::Replace( Buffer, szToReplace, szReplacementText );
            
            delete[] Buffer;

            Buffer = szNew;
            LengthInCharacters = int( strlen( Buffer ));
            MemoryAllocatedInBytes = LengthInCharacters + 1;
        }

        void Replace( u_int uToReplaceStart, u_int uToReplaceLen, const Asura_Char* szReplacementText )
        {
            Asura_Char* szNew = Asura_StringUtil::Replace( Buffer, uToReplaceStart, uToReplaceLen, szReplacementText );
            
            delete[] Buffer;

            Buffer = szNew;
            LengthInCharacters = int( strlen( Buffer ));
            MemoryAllocatedInBytes = LengthInCharacters + 1;
        }

		//reduce the length of string if above iMaxLength
		void Truncate(int iMaxLength)
		{
			if(iMaxLength<LengthInCharacters)
			{
				Buffer[iMaxLength] = 0;
				LengthInCharacters = iMaxLength;
	            MemoryAllocatedInBytes = LengthInCharacters + 1;//not strictly speaking true... but leave like this so that other stuff still works
			}
		}

        // Uses strip flag values from Asura_StringUtil.
        void StripWhitespace( u_int uStripFlags )
        {
            u_int uNewLen;
            const char* szStripped = Asura_StringUtil::StripWhitespace( static_cast<const char*>(Buffer), uStripFlags, &uNewLen );
            memmove( Buffer, szStripped, uNewLen + 1 );
			Buffer[uNewLen] = 0;
            LengthInCharacters = uNewLen;
        }

		bool StripExtension(void)
		{
			for(int i = (LengthInCharacters - 1); i >= 0; i--)
			{
				if(Buffer[i] == '.')
				{
					// We've found our extension - time to truncate the string here...
					Buffer[i]				= '\0';
					LengthInCharacters		= i;
					MemoryAllocatedInBytes	= i + 1;//not strictly speaking true... but leave like this so that other stuff still works
					return(true);
				}
			}

			// No extension found...
			return(false);
		}

		// JAT 080206 - replace '\' with '/'
		void ReplaceAllChars( const Asura_Char cSearch, const Asura_Char cReplace )
		{
            Asura_StringUtil::ReplaceAllChars( static_cast<char*>(Buffer), cSearch, cReplace );
		}

		void Append( const Asura_Char* szSource, u_int uSourceLen )
		{
			if( uSourceLen )
			{
				if( ExpandBy(uSourceLen + (IsNull() ? 1 : 0)) )
				{
					strncpy( &Buffer[LengthInCharacters], szSource, uSourceLen);
					LengthInCharacters += uSourceLen;
					Buffer[LengthInCharacters] = '\0';
				}
			}
		}

		void Prepend( const Asura_Char* szSource, u_int uSourceLen )
		{
			if ( uSourceLen )
			{
				if ( ExpandBy( uSourceLen + ( IsNull() ? 1 : 0 ) ) )
				{
					// strncpy undefined if source and destination strings overlap.
					Asura_Mem::Move( &Buffer[ uSourceLen ], &Buffer[ 0 ], LengthInCharacters * sizeof( Asura_Char ) );
					strncpy( &Buffer[ 0 ], szSource, uSourceLen );
					LengthInCharacters += uSourceLen;
					Buffer[LengthInCharacters] = '\0';
				}
			}
		}

		void Prepend( const Asura_String& xString )
		{
			Prepend( xString.GetString(), xString.GetLength() );
		}

		Asura_String SubString(u_int uStart, u_int uLen) const
		{
			if (uStart > (u_int)LengthInCharacters)
			{
				uStart = LengthInCharacters;
			}
			if (uLen > LengthInCharacters - uStart)
			{
				uLen = LengthInCharacters - uStart;
			}
			return Asura_String(Buffer + uStart, uLen);
		}

	private:

		bool ExpandBy(int bytes)
		{
			int newSize = MemoryAllocatedInBytes + bytes;
	
			Asura_Char* tempBuffer = new Asura_Char[newSize];
            if( tempBuffer )
            {
			    if (Buffer)
			    {
				    strcpy(tempBuffer,Buffer);
				    delete[] Buffer;
                }

				Buffer = tempBuffer;
			    MemoryAllocatedInBytes = newSize;

                return true;
            }
            else
            {
                return false;
            }		
		}
		
    	void AssignFrom( const Asura_Char* szSource )
		{
			if( !szSource )
			{
				Buffer = NULL;
				LengthInCharacters = MemoryAllocatedInBytes = 0;
				return;
			}

			LengthInCharacters = int( strlen( szSource ));
			MemoryAllocatedInBytes = (LengthInCharacters + 1) * sizeof(Asura_Char);
			Buffer = new Asura_Char[MemoryAllocatedInBytes];
            if( Buffer )
            {
			    strcpy( Buffer, szSource );
            }
            else
            {
                MakeNull();
            }
		}
		void AssignFrom( const Asura_String& xSource )
		{
			AssignFrom( xSource.Buffer );
		}

		void Append( const Asura_Char cSource )
		{
			if( ExpandBy(1 + (IsNull() ? 1 : 0)) )
			{
				Buffer[LengthInCharacters] = cSource;
				Buffer[LengthInCharacters + 1] = '\0';
				LengthInCharacters += 1;
			}
		}

		Asura_Char* Buffer;          // (Buffer == NULL) iff (MemoryAllocatedInBytes == NULL)
		int LengthInCharacters;
		int MemoryAllocatedInBytes;

	public:
		//Identifier to write out for null strings.Doing it as a palindrome to save having to worry about
		//endianess issues. It has to be a non-zero value to preserve the old behaviour for
		//saving/loading blank strings. RAR 13/09/2005
		//(Shoving it at the end , so that the variable display in the debugger is clearer)
        enum { NULL_STRING_CHUNK = 0x00ffff00 };
};

Asura_String operator+( const Asura_String& xLhs, const Asura_String& xRhs );

u_int Asura_PrintfToBfr( Asura_Char* szBfr, const Asura_Char* szFormat, u_int uMaxChars = UINT_MAX, bool bForceNul = true, ... );
u_int Asura_PrintfToBfr( Asura_UnicodeChar* wszBfr, const Asura_UnicodeChar* wszFormat, u_int uMaxChars = UINT_MAX, bool bForceNul = true, ... );

u_int Asura_PrintfToBfr( Asura_Char* szBfr, const Asura_Char* szFormat, va_list xArgs, u_int uMaxChars = UINT_MAX, bool bForceNul = true );
u_int Asura_PrintfToBfr( Asura_UnicodeChar* wszBfr, const Asura_UnicodeChar* wszFormat, va_list xArgs, u_int uMaxChars = UINT_MAX, bool bForceNul = true );

u_int Asura_UnicodeToUTF8( Asura_UnicodeChar const* wszSource, Asura_Char* szDestination, int iDestLength = -1 );
u_int Asura_UTF8ToUnicode( Asura_Char const* szSource, Asura_UnicodeChar* wszDestination, int iDestLength = -1 );

#endif//__ASURA_STRING_H__
