//
// Asura_TextureManagement.cpp
//
// Created by Kevin Lea - 12/06/2001 11:03:36
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Core/Asura_TextureManagement.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Stream.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_NameTree.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_Debug_Timers.h"

#ifdef ASURA_TOOLS
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#endif

#if defined (ASURA_USE_TEXTURE_STREAMING)
#include "Core/Asura_Texture_Streaming.h"
#endif

DEBUGVAR bool dbg_bReportDestroyedTextures = false;

Asura_Mutex Asura_Texture_Management::s_xMutex;
Asura_TemplateHashTree<Asura_TextureTreeData>* Asura_Texture_Management::s_pxTextureTree;
Asura_Handle_List* Asura_Texture_Management::s_pxTextureHandleList;
int Asura_Texture_Management::NumberOfTextures;

int Asura_Texture_Management::Loaded[ASURA_MAX_NO_OF_TEXTURES];
int Asura_Texture_Management::TextureFlags[ASURA_MAX_NO_OF_TEXTURES];
bool Asura_Texture_Management::s_bTexturesAddedSinceLastLoad = false;
int Asura_Texture_Management::s_iLastAddedTextureID = ASURA_TEXTURE_NONE;

int* Asura_Texture_Management::ConversionList;
int Asura_Texture_Management::ConversionListSize;
u_int Asura_Texture_Management::s_uCurResetGroup;

#if defined(ASURA_LOGFILES_ENABLED)
static const Asura_Char* TEXTURE_REPORT_DEFAULT_PATH = "Logfiles\\";
static const Asura_Char* TEXTURE_REPORT_DEFAULT_FILE = "TextureReport.txt";
#endif	// defined(ASURA_LOGFILES_ENABLED)


void Asura_Texture_Management::Initialise(void)
{
	#if defined (ASURA_USE_TEXTURE_STREAMING)
	Asura_Texture_Streaming_System::Initialise();
	#endif

	NumberOfTextures = 0;

	s_pxTextureTree = new Asura_TemplateHashTree<Asura_TextureTreeData>;
	s_pxTextureHandleList = new Asura_Handle_List(ASURA_MAX_NO_OF_TEXTURES);

	ConversionList = 0;
    ConversionListSize = 0;
    s_uCurResetGroup = ASURA_TEXTURE_RESETTABLE_GROUP_DEFAULT;
    // (MH - 17.12.2002) added these for safety, global data might not be 0'd for some platforms?
    for (int i = 0; i < ASURA_MAX_NO_OF_TEXTURES; ++i)
    {
        Loaded[i] = ASURA_TEXTURE_UNUSED;
        TextureFlags[i] = 0;
    }
    
    Platform_Initialise();
    PlatformProject_Initialise();
}

void Asura_Texture_Management::DestroyTexture( int iTexID )
{
	s_xMutex.Lock();
	{
		// if texture management is shutdown, texture is already destroyed. This
		// happens on PS2 builds when Shutdown() gets called.
		if (s_pxTextureTree)
		{
			Asura_TextureTreeData* pxData = GetTextureTreeDataFromID( iTexID);
			if(pxData)
			{
				// Allow platform-specific code to take care of deallocating memory and
				// other resources.
				//Not clear whether it is desirable to do this for textures that were never loaded
				//in the first place... however it looks like all platforms cope either way
				//at the moment.  RAR 20/04/2005
				Platform_DestroyTexture( iTexID );

				// Before the texture ID becomes invalid, we should probably flag that
				// this texture is no longer loaded, for when the index gets reused.
				Loaded[iTexID] = ASURA_TEXTURE_UNUSED;


				// Remove from hash tree.
				if( dbg_bReportDestroyedTextures )
				{
					Asura_OutputToDebugger(( "DestroyTexture(): %s (%i).\n", pxData->m_xFilename.GetString(), iTexID ));
				}
				s_pxTextureTree->DeleteNodeByHash( pxData->m_uInternalHash );

				//And the handle list
				s_pxTextureHandleList->ReleaseHandle(iTexID);

				if(s_iLastAddedTextureID == iTexID)
				{
					s_iLastAddedTextureID = ASURA_TEXTURE_NONE;
				}


				// Update texture count.
				--NumberOfTextures;
			}
		}
	}
	s_xMutex.Unlock();
}


void Asura_Texture_Management::Reset(void)
{
	#if defined (ASURA_USE_TEXTURE_STREAMING)
	Asura_Texture_Streaming_System::Reset();
	#endif

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	Reset( s_uCurResetGroup );
#else
	s_xMutex.Lock();
	{
		for(int i=0; i<ASURA_MAX_NO_OF_TEXTURES; i++)
		{
			if (Loaded[i] >= ASURA_TEXTURE_START_OF_RESETTABLE_GROUPS || Loaded[i] == ASURA_TEXTURE_NOT_LOADED)
			{
				DestroyTexture(i);
			}
		}

		// (MH - 15.12.2002) 
		if (ConversionList) delete [] ConversionList;
		ConversionList = NULL;
		ConversionListSize = 0;

		Platform_Reset();
	}
	s_xMutex.Unlock();
#endif
}

void Asura_Texture_Management::Reset(int iGroup)
{
	s_xMutex.Lock();
	{
		for(int i=0; i<ASURA_MAX_NO_OF_TEXTURES; i++)
		{
			if (Loaded[i] == iGroup)
			{
				DestroyTexture(i);
			}
		}

		if (ConversionList) delete [] ConversionList;
		ConversionList = NULL;
		ConversionListSize = 0;

		Platform_Reset();
	}
	s_xMutex.Unlock();
}

void Asura_Texture_Management::Shutdown(void)
{
	#if defined (ASURA_USE_TEXTURE_STREAMING)
	Asura_Texture_Streaming_System::Shutdown();
	#endif

	UnloadTextures();
	NumberOfTextures = 0;

	if (s_pxTextureTree)
	{
		delete s_pxTextureTree;
		s_pxTextureTree = 0;
	}
	if(s_pxTextureHandleList)
	{
		delete s_pxTextureHandleList;
		s_pxTextureHandleList = 0;
	}

	if (ConversionList)
	{
		delete [] ConversionList;
		ConversionList = 0;
        ConversionListSize = 0;
	}

    Platform_Shutdown();
}


void Asura_Texture_Management::NewConversionList(int length)
{
	s_xMutex.Lock();
	{
		if (ConversionList) delete [] ConversionList;

		ConversionList = new int[length];

		// (MP - 04.04.2002) Record the size of the conversion list so we can
		// verify indices passed elsewhere.
		ConversionListSize = length;
	}
	s_xMutex.Unlock();
}

int Asura_Texture_Management::ConvertIndex(int index)
{
	Asura_Assert(ConversionList,"");
    Asura_Assert( 0 <= index && index < ConversionListSize, "Invalid texture index passed for conversion." );

	if ( (index<0) || (index>=ConversionListSize) )
	{
		return 0;
	}

	return ConversionList[index];
}

int Asura_Texture_Management::GetLastAddedTexID( void )
{
    return( s_iLastAddedTextureID);
}

int Asura_Texture_Management::AddTexture( const char* szName, int originalIndex )
{
	int	iNewIndex;
	s_xMutex.Lock();
	{
		iNewIndex = AddTextureWithResetGroup( szName );
		
		Asura_Assert( 0 <= originalIndex && originalIndex < ConversionListSize, "Invalid original index." );
		ConversionList[originalIndex] = iNewIndex;
	}
	s_xMutex.Unlock();
	return iNewIndex;
}

int Asura_Texture_Management::AddTextureWithResetGroup( const char* szName, u_int uGroup, bool* pbWasNew )
{
	int	iID;

    if( !szName || !*szName ) { return ASURA_TEXTURE_NONE; }

	s_xMutex.Lock();
	{
		// JAT 13/02/06 - make sure there are no '/' in the name
		Asura_Assertf( ( Asura_StringUtil::Contains( szName, "/" ) == false ), ( "Textures must not contain '/' in them %s", szName ) );
#if defined ASURA_AUTO_REPLACE_FORWARD_SLASH_IN_TEXTURE_NAMES
		// JAT - very nasty change the '/' to '\'
		char* szPtr = const_cast<char*>(szName);
		Asura_StringUtil::ReplaceAllChars( szPtr, '/', '\\' );
#endif // #if defined ASURA_AUTO_REPLACE_FORWARD_SLASH_IN_TEXTURE_NAMES

		const u_int uBackup = GetTextureResetGroup();
		SetTextureResetGroup( uGroup );

#ifdef ASURA_USE_TEXTURE_LOG
		AddToTextureLog("%s\n",szName);
#endif

		//Is the texture already loaded?
		iID = GetTextureIndex(szName);

		if( pbWasNew )  *pbWasNew = (iID == ASURA_TEXTURE_NONE);


		if( iID == ASURA_TEXTURE_NONE )
		{
			if(NumberOfTextures < ASURA_MAX_NO_OF_TEXTURES)
			{
				//Get new entry in tree
				Asura_Hash_ID uHash = GetInternalHash(szName,uGroup);
				Asura_TextureTreeData* pxData = s_pxTextureTree->AddNode(uHash);

				if(pxData)
				{
					//get texture ID from handle tree
					s_pxTextureHandleList->GetNewHandle(iID);
					Asura_Assert(iID != ASURA_TEXTURE_NONE,"Failed to get new texture DI");


					//and fill in the data
					pxData->m_iTextureID = iID;
					pxData->m_uInternalHash = uHash;
					pxData->m_xFilename = szName;

					Asura_Assert(Loaded[iID] == ASURA_TEXTURE_UNUSED,"Index already in use in Loaded array");
					NumberOfTextures++;
					Loaded[iID] = ASURA_TEXTURE_NOT_LOADED;
					TextureFlags[iID] = 0;

					//Set flag so LoadTextures() will know there is something to do
					s_bTexturesAddedSinceLastLoad = true;
					s_iLastAddedTextureID = iID;

				}
				else
				{
					Asura_Assert(0,"Texture has some hash ID as an earlier texture");
				}
			}
			else
			{
				Asura_Assertf( false, ("Too many textures loaded. %d maximum.", ASURA_MAX_NO_OF_TEXTURES));
			}



		}
#ifdef ASURA_USE_TEXTURE_LOG
		else
		{
			//check for existing entry having a different extension
			const Asura_String* pxExistingName = GetTextureName(iID);

			if(pxExistingName && pxExistingName->Compare(szName))
			{
				const char* szExtension = strrchr(pxExistingName->GetString(),'.');
				Asura_Assertf(0,("Texture %s previously added with a different extension (%s)",szName,szExtension));
				AddToTextureLog("Error - %s previously added with a different extension (%s)\n",szName,szExtension);
			}
		}
#endif


		SetTextureResetGroup( uBackup );
	}
	s_xMutex.Unlock();
    return iID;
}

const Asura_String* Asura_Texture_Management::GetTextureName(int iIndex)
{
	Asura_TextureTreeData* pxData = GetTextureTreeDataFromID(iIndex);
	return pxData ? &pxData->m_xFilename : 0;
}

void Asura_Texture_Management::MakeNameList( const Asura_String** ppxNameList )
{
	s_xMutex.Lock();
	{
		s_pxTextureTree->Traverse(MakeNameListCallback,ppxNameList);
	}
	s_xMutex.Unlock();
}

bool Asura_Texture_Management::MakeNameListCallback(Asura_Hash_ID uHash, Asura_TextureTreeData* pxData,void* pUserData)
{
	const Asura_String** ppxNameList = (const Asura_String**) pUserData;
	ppxNameList[pxData->m_iTextureID] = &pxData->m_xFilename;

	return true;
}

void Asura_Texture_Management::LoadTextures(void)
{
	s_xMutex.Lock();
	{
		if(s_bTexturesAddedSinceLastLoad)
		{
			s_bTexturesAddedSinceLastLoad = false;

			if( s_pxTextureTree )
			{
//				ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Texture_Management::LoadTextures");
	
				s_pxTextureTree->Traverse( LoadTexturesCallback );
			}

#if defined(ASURA_TOOLS) && defined(ASURA_MAIN)
			//After loading textures on PC tools may need to get rid of any DDS conversion messages
			//RAR 01/07/04
			Tools_Level_Export::ClearProgressMessage();
#endif
		}
	}
	s_xMutex.Unlock();
}

void Asura_Texture_Management::LoadTexturesCallback(Asura_Hash_ID uHash, Asura_TextureTreeData* pxData)
{
	
	#if 0//defined (ASURA_USE_TEXTURE_STREAMING)
	#if defined ASURA_TOOLS
	// Tools-side texture streaming

	{
		// TODO - need some sort of flag for streaming textures? Currently just looking for "environment" textures.
		if ( pxData->m_xFilename.Contains("environment") || pxData->m_xFilename.Contains("avp_demo") || pxData->m_xFilename.Contains("test" ) )
		{

			if( IsTextureLoaded( pxData->m_iTextureID ) || Loaded[ pxData->m_iTextureID ] == ASURA_TEXTURE_ERROR_LOADING ) return;
			
			// Set tools data for width/height to 0 to indicate that it is unknown yet
		//	Tools_Texture_Management::StoreTextureSize( pxData->m_iTextureID, 0, 0 );
	
			Loaded[ pxData->m_iTextureID ] = s_uCurResetGroup;

			Asura_Texture_Streaming_System::RegisterTexture( pxData->m_iTextureID );		
			return;
		}
	}

	#endif
	#endif

	Platform_LoadTexture(pxData->m_xFilename,pxData->m_iTextureID);
}

void Asura_Texture_Management::UnloadTextures(void)
{
	s_xMutex.Lock();
	{
		// (MH - 03.05.2002) texture indices are not necessarily continuous 
		// so destroy all possible textures. DestroyTexture should handle any empty slots.
		//for(int i=0; i<NumberOfTextures; i++)
		for(int i=0; i<ASURA_MAX_NO_OF_TEXTURES; i++)
		{
			DestroyTexture(i);
		}
	}
	s_xMutex.Unlock();
}

void Asura_Texture_Management::SetTextureResetGroup( u_int uGroup )
{
	if( uGroup != s_uCurResetGroup )
	{
		// load any pending textures
		LoadTextures();

		// and from now on all loaded textures are resetable
		s_uCurResetGroup = uGroup;
	}
}

bool Asura_Texture_Management::LoadAsuraFileUsingResetGroup( const char* szFilename, u_int uGroup )
{
    const u_int uBackup = GetTextureResetGroup();
    SetTextureResetGroup( uGroup );
    bool bOK = Asura_ChunkLoading_System::LoadAsuraFile( szFilename );
    SetTextureResetGroup( uBackup );
    return bOK;
}

Asura_Hash_ID Asura_Texture_Management::GetHashIDFromIndex(int iIndex)
{
	Asura_Assert(s_pxTextureTree,"");
	if(iIndex == ASURA_TEXTURE_NONE) return ASURA_HASH_ID_UNSET;

	Asura_TextureTreeData* pxData = GetTextureTreeDataFromID(iIndex);

	if(pxData)
	{
		//Note that for external purposes we use the hash ID from the filename , not the internal hash ID. RAR 21/03/2006
		return Asura_GetHashID_IgnoreExtension(pxData->m_xFilename.GetString());
	}
	else
	{
		return ASURA_HASH_ID_UNSET;
	}
}

int Asura_Texture_Management::GetIndexFromHashID(Asura_Hash_ID uBaseHash)
{
	Asura_Assert(s_pxTextureTree,"");
	if(uBaseHash == ASURA_HASH_ID_UNSET) return ASURA_TEXTURE_NONE;

	//Try current load group.
	//(For non-streaming this actually checks all load groups)
	Asura_Hash_ID uHash = GetInternalHash(uBaseHash, s_uCurResetGroup);
	Asura_TextureTreeData* pxData = s_pxTextureTree->GetDataByHash(uHash);
	if(pxData)	return pxData->m_iTextureID;

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	//For streaming , we may want to check the common and permanent textures as well. RAR 21/03/2006
	if(s_uCurResetGroup > ASURA_TEXTURE_RESETTABLE_GROUP_COMMON)
	{
		uHash = GetInternalHash(uBaseHash, ASURA_TEXTURE_RESETTABLE_GROUP_COMMON);
		pxData = s_pxTextureTree->GetDataByHash(uHash);
		if(pxData)	return pxData->m_iTextureID;
	}

	if(s_uCurResetGroup != ASURA_TEXTURE_LOADED_PERMANENT)
	{
		uHash = GetInternalHash(uBaseHash, ASURA_TEXTURE_LOADED_PERMANENT);
		pxData = s_pxTextureTree->GetDataByHash(uHash);
		if(pxData)	return pxData->m_iTextureID;
	}
#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)

	return ASURA_TEXTURE_NONE;
}


static Asura_TextureTreeData* g_pxTextureTreeData = 0;

Asura_TextureTreeData* Asura_Texture_Management::GetTextureTreeDataFromID(int iTexID)
{
	s_xMutex.Lock();
	{
		g_pxTextureTreeData = 0;
		s_pxTextureTree->Traverse(GetTextureTreeDataCallback,&iTexID);
	}
	s_xMutex.Unlock();
	return g_pxTextureTreeData;
}

bool Asura_Texture_Management::GetTextureTreeDataCallback(Asura_Hash_ID uHash,Asura_TextureTreeData* pxData,void* pIndex)
{
	if(pxData->m_iTextureID == *(int*)pIndex)
	{
		g_pxTextureTreeData = pxData;
		return false;
	}
	return true;
}



#ifdef ASURA_LOGFILES_ENABLED

static char* g_szDumpLogfileName = 0;

void Asura_Texture_Management::DumpSingleTexture(  Asura_Hash_ID uHash, Asura_TextureTreeData* pxData )
{
	char* szTextureStatus = 0;
	switch(Loaded[pxData->m_iTextureID])
	{
	case ASURA_TEXTURE_UNUSED: szTextureStatus = "ASURA_TEXTURE_UNUSED"; break;
	case ASURA_TEXTURE_NOT_LOADED: szTextureStatus = "ASURA_TEXTURE_NOT_LOADED"; break;
	case ASURA_TEXTURE_LOADED_PERMANENT: szTextureStatus = "ASURA_TEXTURE_LOADED_PERMANENT"; break;
	case ASURA_TEXTURE_RESETTABLE_GROUP_COMMON: szTextureStatus = "ASURA_TEXTURE_RESETTABLE_GROUP_COMMON"; break;
	case ASURA_TEXTURE_RESETTABLE_GROUP_DEFAULT: szTextureStatus = "ASURA_TEXTURE_RESETTABLE_GROUP_DEFAULT"; break;
	case ASURA_TEXTURE_RESETTABLE_GROUP_CUTSCENE_DATA: szTextureStatus = "ASURA_TEXTURE_RESETTABLE_GROUP_CUTSCENE_DATA"; break;
	case ASURA_TEXTURE_RESETTABLE_GROUP01: szTextureStatus = "ASURA_TEXTURE_RESETTABLE_GROUP01"; break;
	case ASURA_TEXTURE_RESETTABLE_GROUP02: szTextureStatus = "ASURA_TEXTURE_RESETTABLE_GROUP02"; break;
	case ASURA_TEXTURE_RESETTABLE_GROUP03: szTextureStatus = "ASURA_TEXTURE_RESETTABLE_GROUP03"; break;
	default: szTextureStatus = "<<Unknown>>"; break;
	}
	Asura_SimpleLogfileOutput((g_szDumpLogfileName, "%i: [%s] %s\n", pxData->m_iTextureID, szTextureStatus, pxData->m_xFilename.GetString()));
}

void Asura_Texture_Management::DumpRegisteredTextures(const char* szFilenameNoPath /* = NULL */, const char* szPath /* = NULL */)
{
	char szFileName[256];
    Asura_PrintfToBfr(szFileName, "%s%s", 256, true, 
        szPath ? szPath : TEXTURE_REPORT_DEFAULT_PATH, 
        szFilenameNoPath ? szFilenameNoPath : TEXTURE_REPORT_DEFAULT_FILE);

    Asura_OutputToDebugger(( "Dumping texture report to \"%s\".\n", szFileName ));

	Asura_SimpleLogfileClear(szFileName);
	g_szDumpLogfileName = szFileName;
	Asura_SimpleLogfileOutput_PreFormatted(szFileName, "\n\n", 2);

	s_pxTextureTree->Traverse(DumpSingleTexture);
}

#endif



#ifdef ASURA_USE_TEXTURE_LOG

static bool g_bClearLogfile = true;
static bool g_bNeedFileHeading = true;
static char g_szLastAsuraFile[300] = "";

void Asura_Texture_Management::SetCurrentAsuraFile(const char* szFilename)
{
	if(szFilename)
	{
		strncpy(g_szLastAsuraFile,szFilename,300);
		g_szLastAsuraFile[299] = 0;
		g_bNeedFileHeading = true;
	}
	else if(g_szLastAsuraFile[0])
	{
		g_szLastAsuraFile[0] = 0;
		g_bNeedFileHeading = true;
	}
}

void Asura_Texture_Management::AddToTextureLog(const char* szFormat,...)
{
	char szBuffer[500];

	va_list ap;
	va_start(ap, szFormat);
    Asura_PrintfToBfr( szBuffer, szFormat, ap, 500, true );
	va_end(ap);

	const char* szLogFile = "Logfiles\\TextureLog.txt";
	
	if(g_bClearLogfile)
	{
		//init logfile
		Asura_SimpleLogfileClear(szLogFile);
		g_bClearLogfile = false;
	}

	if(g_bNeedFileHeading)
	{
		//add heading containing asura filename
		if(g_szLastAsuraFile[0])
		{
			Asura_SimpleLogfileOutput((szLogFile,"\nFile : %s \n",g_szLastAsuraFile));
		}
		else
		{
			Asura_SimpleLogfileOutput((szLogFile,"\nFile : None\n"));
		}
		g_bNeedFileHeading = false;
	}

	Asura_SimpleLogfileOutput((szLogFile,"%s",szBuffer));
}

#endif //ASURA_USE_TEXTURE_LOG

void Asura_Texture_Management::WriteResetGroupReport(u_int uGroup, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 96;
	char szLogBuffer[uLOGBUFFERSIZE];

	bool bFirst = true;

	Asura_TemplateHashTree_Iterator<Asura_TextureTreeData> xIt(*s_pxTextureTree);
	for ( ; !xIt.Done(); xIt.Next() )
	{
		Asura_TextureTreeData* pxData = xIt.GetCurrent();
		if (pxData && (Loaded[pxData->m_iTextureID] == static_cast<int>(uGroup)))
		{
			Asura_PrintfToBfr(szLogBuffer,"Texture:\t\t\t%s\n",uLOGBUFFERSIZE,true,pxData->m_xFilename.GetString());
			xLog << szLogBuffer;
			if (bFirst) bFirst = false;
		}
	}

	if (bFirst)
	{
		xLog << "Texture:\t\t\tNone\n";
	}
}

#ifdef ASURA_TOOLS
bool Asura_Texture_Management::RenameTexture(int iTextureID,const char* szNewFilename)
{

	Asura_TextureTreeData* pxData = GetTextureTreeDataFromID(iTextureID);
	if(!pxData) return false;

	Asura_Assert(Loaded[iTextureID] == (int)s_uCurResetGroup || !IsTextureLoaded(iTextureID),"This is probably going to result in the texture's load group being changed , which is probably not so good.");

	//Get the new hash value to use
	Asura_Hash_ID uNewHash = GetInternalHash(szNewFilename,s_uCurResetGroup);
	if(uNewHash == pxData->m_uInternalHash)
	{
		//Hash value unchanged (most likely changing the extension).
		//If the filename is unchanged , then do nothing and return false , since no rename took place
		if(!pxData->m_xFilename.Compare(szNewFilename)) return false;

	}
	else
	{
		if(!s_pxTextureTree->ReplaceHash(pxData->m_uInternalHash,uNewHash)) return false;
	}

	//Move has worked , so update the texture tree data
	pxData->m_uInternalHash = uNewHash;
	pxData->m_xFilename = szNewFilename;

	return true;
}
#endif // ASURA_TOOLS
