//
// Asura_TextureManagement.h
//
// Created by Kevin Lea - 12/06/2001 11:01:49
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_TEXTUREMANAGEMENT_H__
#define __ASURA_TEXTUREMANAGEMENT_H__

#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureFlags.h"

class Asura_HashTree;
class Asura_Stream;
class Asura_Mutex;
class Asura_NameTree;
template <class T> class Asura_TemplateHashTree;

#ifndef ASURA_MAX_NO_OF_TEXTURES
// RICHARDS	#define ASURA_MAX_NO_OF_TEXTURES 500
#ifdef ASURA_TOOLS
//Allow more textures for the tools , since it has the habit of loading up more or less everything. RAR 10/06/03
#define ASURA_MAX_NO_OF_TEXTURES 2000
#else
#define ASURA_MAX_NO_OF_TEXTURES 250
#endif
#endif //ASURA_MAX_NO_OF_TEXTURES

// (MP - 08.07.2003) NB: Tiled texture stuff now shares these groups (and
// the current one as dictated by the Asura_Texture_Management class.)
enum Asura_LoadedTextureType
{
	ASURA_TEXTURE_ERROR_LOADING = -2,//A failed attempt has been made at loading this texture. RAR 07/03/2006
	ASURA_TEXTURE_UNUSED = -1,//for marking slots in Loaded array that haven't been used yet. RAR 20/10/2005
    ASURA_TEXTURE_NOT_LOADED = 0,
    ASURA_TEXTURE_LOADED_PERMANENT,
    ASURA_TEXTURE_START_OF_RESETTABLE_GROUPS,
	ASURA_TEXTURE_RESETTABLE_GROUP_COMMON = ASURA_TEXTURE_START_OF_RESETTABLE_GROUPS,
    ASURA_TEXTURE_RESETTABLE_GROUP_DEFAULT,
    ASURA_TEXTURE_RESETTABLE_GROUP_CUTSCENE_DATA,
    ASURA_TEXTURE_RESETTABLE_GROUP01,
    ASURA_TEXTURE_RESETTABLE_GROUP02,
    ASURA_TEXTURE_RESETTABLE_GROUP03,   // probably want to replace these with project IDs.
#if defined(ASURA_USE_STREAMING)
	ASURA_TEXTURE_RESETTABLE_STREAMING_SKYBOX,
	ASURA_TEXTURE_RESETTABLE_STREAMING_GROUP01,
#endif	// defined(ASURA_USE_STREAMING)
};

/*
Notes on changes for streaming system :

To allow for the same texture to appear in multiple load groups , the system is being changed to use a template hash
tree instead of a name tree. The hash values used will be a combination of the hash value taken from the filename ,and
a value taken from the load group.

These modified hash values are only used internally. For external purposes any hash values are solely based on the filename.

When searching for a texture index , the system will first check in the current load group , then  check in the common 
load group (ASURA_TEXTURE_RESETTABLE_GROUP_COMMON) and finally in the permanent group.

It is expected that permanent graphics will go in ASURA_TEXTURE_LOADED_PERMANENT. Graphics that are common to all zones will go
in ASURA_TEXTURE_RESETTABLE_GROUP_COMMON (in practice these may be as permanent as the permanent textures). And textures
for individual zones will use subsequent groups.

Finally ,if the streaming system isn't being used , then the internal hash values aren't modified by the load group , so
there will be only one copy of each texture loaded as before.

RAR 21/03/2006
*/

//define that determines whether logging of loaded textures should be done. RAR 07/03/2006
#if defined(ASURA_MAIN) && defined(ASURA_TOOLS) && defined(ASURA_LOGFILES_ENABLED)
#define ASURA_USE_TEXTURE_LOG 1
#endif

//data stored in hash tree. RAR 21/03/2006
struct Asura_TextureTreeData
{
	int m_iTextureID;	
	Asura_Hash_ID m_uInternalHash;
	Asura_String m_xFilename;	
};


class Asura_Texture_Management
{
	public:

		static void Initialise(void);
		static void Shutdown(void);
		static void Reset(void);
        static void Reset(int iGroup);
		
		static void NewConversionList(int length);

		// Kevin Lea - 14/09/2001 15:24:22 - called from Asura_Chunk_ResourceFile
		static void LoadTextureFromResourceFile(Asura_Handle fileHandle,int fileSize, const Asura_Char* fileName, int iTextureFlags, u_int uChunkFlags);


		// KJL - 15-Mar-01 15:37:03 - used by mesh loaders (e.g. to load environment textures)
		static int AddTexture( const char*         szName, int iOriginalIndex );
        static int AddTexture( const Asura_String& xName,  int iOriginalIndex ) { return AddTexture( xName.GetString(), iOriginalIndex ); }

        static int AddTextureWithResetGroup( const char*         szName, u_int uGroup = s_uCurResetGroup, bool* pbWasNew = NULL );
        static int AddTextureWithResetGroup( const Asura_String&  xName, u_int uGroup = s_uCurResetGroup, bool* pbWasNew = NULL ) { return AddTextureWithResetGroup( xName.GetString(), uGroup, pbWasNew ); }

		// KJL - 15-Mar-01 15:37:08 - used to explicitly load a texture (e.g. request to load a font)
        static int AddTexture( const char*         szName, bool* pbWasNew = NULL ) { return AddTextureWithResetGroup( szName, s_uCurResetGroup, pbWasNew ); }
        static int AddTexture( const Asura_String& xName,  bool* pbWasNew = NULL ) { return AddTextureWithResetGroup( xName,  s_uCurResetGroup, pbWasNew ); }

		static int ConvertIndex(int index);
        static int GetLastAddedTexID( void );

		// KJL - 14-Mar-01 11:23:22 - returns -1 if no texture of that name exists
		static int GetTextureIndex( const Asura_String& name ){return GetTextureIndex(name.GetString());}
		static int GetTextureIndex( const char* pszName){return GetIndexFromHashID(Asura_GetHashID_IgnoreExtension(pszName));}

		/// Kevin Lea - 20/08/2002 16:52:55 - returns null if texture not found
		static const Asura_String* GetTextureName(int iIndex);

		static inline int GetNumberOfTextures(void)
		{
			return NumberOfTextures;
		}
		static inline int GetTextureFlags(int i)
		{
			if (i == ASURA_TEXTURE_NONE)
			{
				return 0;
			}

			return TextureFlags[i];
		}
		static inline void SetTextureFlags(int iTexID,int iFlags)
		{
			if( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES )
			{
				TextureFlags[iTexID] = iFlags;
			}
			else
			{
				Asura_Assert( false, "Invalid parameter passed to Asura_Texture_Management::SetTextureFlags()." );
			}
		}

		static inline void AddTextureFlags(int iTexID,int iFlags)
		{		
			if( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES )
			{
				TextureFlags[iTexID] |= iFlags;
			}
			else
			{
				Asura_Assert( false, "Invalid parameter passed to Asura_Texture_Management::AddTextureFlags()." );
			}
		}

        static inline bool IsTextureIDValid( int iTexID ) { return( NULL != GetTextureName(iTexID) ); }
        
		static inline bool IsTextureLoaded(int textureIndex) 
		{
			if (textureIndex == ASURA_TEXTURE_NONE)
			{
				return false;
			}

			return Loaded[textureIndex] > ASURA_TEXTURE_NOT_LOADED;
		}
        static inline int GetTextureLoadState( int iTexID )
        {
            return( ASURA_TEXTURE_NONE == iTexID ? ASURA_TEXTURE_UNUSED : Loaded[iTexID] );
        }

		// (AAM - 23.03.2006) Should rarely be needed, usually set by member functions. 360 has other texture bundles though
		static inline void SetTextureLoadState( const int iTexID, const u_int uLoadedType )
		{
			if( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES )
			{
				Loaded[ iTexID ] = uLoadedType;
			}
			else
			{
				Asura_Assert( false, "Invalid parameter passed to Asura_Texture_Management::SetTextureLoadState()." );
			}
		}

		static inline bool IsTextureFMV(int textureIndex)
		{
			if (textureIndex == ASURA_TEXTURE_NONE)
			{
				return false;
			}

			return (TextureFlags[textureIndex]&AsuraTextureFlag_FMV)!=0;
		} 

		// DWW TEST 28/06/06 returns if the texture is a ad receiver
		static inline bool IsTextureAd(int iTextureIndex)
		{
			if (iTextureIndex == ASURA_TEXTURE_NONE)
			{
				return false;
			}

			return (TextureFlags[iTextureIndex]&AsuraTextureFlag_AdReceiver)!=0;
		}

		static void MakeNameList( const Asura_String** nameList );
		
		static void LoadTextures(void);
		static void UnloadTextures(void);

		static void DestroyTexture(int index);

		#ifdef ASURA_TOOLS
		//Changes texture name for texture 
		//Returns true if rename was succesful. RAR 21/03/2006
		static bool RenameTexture(int iTextureID,const char* szNewFilename);
		#endif

		//Attempts to reload the texture using the filename stored for the texture. Only supported on the PC
		//at the moment (since consoles will be loading the textures from concatenated files) , and basically
		//is here for tools use. (RAR 8/8/2002)
		static void ReloadTexture(int index);

        // (MH - 17.12.2002) all textures with group >= ASURA_TEXTURE_START_OF_RESETTABLE_GROUPS
        // will be unloaded by Reset(). Only textures with this group get unloaded by Reset(iGroup)
        // Textures get allocated the group set here when they are loaded.
        static u_int GetTextureResetGroup( void ) { return s_uCurResetGroup; }
        static void SetTextureResetGroup( u_int uGroup );
        static bool LoadAsuraFileUsingResetGroup( const char* szFilename, u_int uGroup = s_uCurResetGroup );

		//RAR 19/05/03 - Gets hash value of the name of a texture from its index
		static Asura_Hash_ID GetHashIDFromIndex(int iIndex);

		//RAR 19/05/03 - Gets the index of a texture from the hash value of its name
		static int GetIndexFromHashID(Asura_Hash_ID uHash);

#ifdef ASURA_LOGFILES_ENABLED
		// RJF - 06 June 2005 - Report the current textures to a file
		static void DumpRegisteredTextures(const char* szFilenameNoPath = NULL, const char* szPath = NULL);
#endif

#ifdef ASURA_USE_TEXTURE_LOG
		//sets the asura file currently being loaded , to give some context to the other logfile messages
		static void SetCurrentAsuraFile(const char* szFilename);

		//Output to logfile
		static void AddToTextureLog(const char* szFormat,...);
#endif
		static void WriteResetGroupReport(u_int uGroup, Asura_Stream& xLog);

		static bool IsTextureCubemap( const int iIndex ) { return Platform_IsTextureCubemap( iIndex ); }

	private:

		static Asura_TextureTreeData* GetTextureTreeDataFromID(int iTexID);
		static Asura_Hash_ID GetInternalHash(const char* szName,u_int uLoadGroup) {return GetInternalHash(Asura_GetHashID_IgnoreExtension(szName),uLoadGroup);}
		static Asura_Hash_ID GetInternalHash(Asura_Hash_ID uBaseHash,u_int uLoadGroup)
		{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
			return uBaseHash + (uLoadGroup<<16);
#else
			return uBaseHash;
#endif	//defined(ASURA_USE_MULTIPLE_HANDLERS)
		}

		static bool MakeNameListCallback(Asura_Hash_ID uHash, Asura_TextureTreeData* pxData,void* pUserData);
		static void LoadTexturesCallback(Asura_Hash_ID uHash, Asura_TextureTreeData* pxData);
		static bool GetTextureTreeDataCallback(Asura_Hash_ID uHash, Asura_TextureTreeData* pxData,void* pUserData);

#ifdef ASURA_LOGFILES_ENABLED
		static void DumpSingleTexture( Asura_Hash_ID uHash, Asura_TextureTreeData* pxData );
#endif

		static Asura_Mutex s_xMutex;
		static Asura_TemplateHashTree<Asura_TextureTreeData>* s_pxTextureTree;
		static Asura_Handle_List* s_pxTextureHandleList;
		static int NumberOfTextures;
	
		static int Loaded[ASURA_MAX_NO_OF_TEXTURES];// (MH - 17.12.2002) was bool, now code from enum Asura_LoadedTextureType
		static int TextureFlags[ASURA_MAX_NO_OF_TEXTURES];

		static bool s_bTexturesAddedSinceLastLoad;
		static int s_iLastAddedTextureID;
		
		static int* ConversionList;
        static int ConversionListSize;

        static u_int s_uCurResetGroup;

		// platform specific
		// KJL - 15-Mar-01 15:57:46 - implemented in Asura_PlatformTextureManagement.cpp
        static void Platform_Initialise( void );
        static void Platform_Shutdown( void );
        static void Platform_Reset( void );
        static void Platform_LoadTexture( const Asura_String& xName, int iTexID );
        static void Platform_LoadTextureFromResource( Asura_Handle iFileHandle, int iFileSize, int iTexID, u_int uChunkFlags );
        static void Platform_DestroyTexture( int iTexID );
		static bool Platform_IsTextureCubemap( const int iTexID );
        
        // (MH - 30.12.2002) load common graphics for your platform / project
        static void PlatformProject_Initialise( void );
	
};

#endif//__ASURA_TEXTUREMANAGEMENT_H__
