//////////////////////////////////////////////////////////////////////////////
//
// Asura_TiledTextureManagement.h
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// System for managing tiled textures (for HUD, decals etc.)
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_TILEDTEXTUREMANAGEMENT_H___
#define ___ASURA_TILEDTEXTUREMANAGEMENT_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"

#ifdef ASURA_TOOLS
#include "Misc/Tools_DynamicArray.h"
#endif

// F O R W A R D  D E C L A R A T I O N S ////////////////////////////////////

class Asura_Stream;

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_TiledTextureInfo
{
	public:

		float GetHeightInPixels() const { return m_TiledToHeight * (m_fv2 - m_fv1); }
		float GetWidthInPixels() const { return m_TiledToWidth * (m_fu2 - m_fu1); }

		Asura_String m_xNameOfTextureUsed;	// name of texture used - will be part of texture management - only req for v1 chunks and TOOLS (would be nice to obsolete the v1 chunks) NDB
		
		int	  m_iIndexOfTextureUsed;	// to try and speed things up
		float m_fu1, m_fv1;			// top left of tile cutfrom
		float m_fu2, m_fv2;			// bottom right of cutfrom
		float m_TiledToWidth;
		float m_TiledToHeight;
		u_int m_uTileHashID;
		u_int m_uTextureHashID;      // hash id generated from parent texture name and compatable with Asura_Texture_Management
		u_int m_uResetGroup;

#ifdef ASURA_TOOLS
		int  m_iAsuraFileNameIndex;  // position of this tile's parent files name in the asura file name container in tile management class
#endif
};

// Tiled Texture Management
//
// Provides a repository for all loaded tiled textures. Each tiled texture is essentially meta-data defined by Asura_TiledTextureInfo 
// and reffering to a sub-region of a parent texture. This class stores the tiled texture meta-data and provides methods for tiled 
// texture retrival. The design of the interface broadly follows that of the Asura_Texture_Management class.
//
// in Tools builds the class also maintains a collection of all the asura files from which tiled textures have been loaded and 
// provides a mechanism for retriving the path to the file from which a tile was originally loaded. Asura file names are added 
// to the collection by the Asura_Chunk_TiledTexture::Process() method.
class Asura_TiledTextureManagement
{
	public:

		static const int ASURA_TILEDTEXTURE_UNKNOWN_FILE;
		static const int ASURA_TILEDTEXTURE_NONE;

		static void Initialise( void );
		static void Reset( void );
		static void Reset( u_int uResetGroup );
		static void Shutdown( void );

		static Asura_Handle AddTile( const Asura_Hash_ID uHashIDOfTile, const Asura_TiledTextureInfo& xNewTile ); // deprecated as hash ID is stored in TiledTextureInfo - NDB 181007
		static Asura_Handle AddTile( const Asura_TiledTextureInfo& xNewTile );
		static bool DeleteTile( Asura_Hash_ID uHashIDOfTile );

		// destroys all tiles whose parent texture hash matches the argument (causes the texture to be destoryed in Asura_Texture_Management as well) 
		static bool DestroyTexture( const Asura_Hash_ID uHashIDOfTiledTexture );
		static bool DestroyTexture( const char* szTextureName );
		
		// It should no longer be nessary to use this method as texture hash id's are stored for each tile in the Asura_TiledTextureInfo
		// structure. The SetTextureHandlesFromHash() method bellow should be faster and have the same effect.
		//
		// Version 2 Texture tiler chunks are loaded after the parent texture data and a texture index is retrived from the Asura_Texture_Management 
		// class when tiles are read in from a Version 2 chunk file. This makes both the SetTextureHandlesFrom*() methods obsolete when working solely 
		// with Version 2 Tiled Texture Chunk Files - NDB 141107
		static void SetTextureHandlesFromNames( void );

		static void SetTextureHandlesFromHash( void );

		// returns the index of the tile with the coresponding hash ID 
		static Asura_Handle	GetHandleOfTile( Asura_Hash_ID uHashIDOfTile );
		static const Asura_TiledTextureInfo* GetTileInfoByHash( Asura_Hash_ID uHashIDOfTile );
		
		// returns a pointer to the tile with index equal to the argument if that tile is "in use" and the index is within range (asserts if out of range), returns NULL otherwise 
		static const Asura_TiledTextureInfo* GetTileInfo( Asura_Handle iHandle );

		// similar to GetTileInfo() but does not check the state of "in use" and clamps the arguement to the range of the tile collection
		static const Asura_TiledTextureInfo* GetTileInfoByID( u_int uID );

		static void SetResetGroup( Asura_Handle iHandle, u_int uResetGroup );

		// returns the number of tiles managed
		static int Size(){ return s_iTileCount; }
		// returns the number that's higher then the highest index used so far
		static int MaxIndex() { return s_iTileMaxIndex; }

		// fills the container provided with the ID numbers of all (upto a maximum of iSize) the managed tiles.
		static const u_int* MakeTileIDList( u_int* puContainer,const int iSize );

		static void WriteResetGroupReport(u_int uGroup, Asura_Stream& xLog);

#ifdef ASURA_TOOLS	
		static const int AddAsuraFileName( const Asura_String& xFileName );
		static void	 LogAsuraFileName( const Asura_Handle iTileHandle,const int iAsuraFileNameIndex );
		static const Asura_String* const GetAsuraFileName( const int iTileIndex );
		static const Asura_String* const GetAsuraFileNameByHash( const u_int uTileHashID );
#endif	

	private:

		// <-- private static methods -->
		
		// sets the "in use" state of a tile to false and optionally removes the tile's parent texture 
		// from Asura_Texture_Management if it is not refferenced by any other tiles
		static void DeleteTileByIndex( const int iIdx, const bool bDestroyUnusedTextures = true );
		
		// returns true if any tile's parent texture index is equal to the arguement
		static bool IsTextureUsedByAnyTile( int iTextureID );
        static void ResetInternal( u_int uMinResetGroup, u_int uMaxResetGroup );
        static void PlatformProject_Initialise( void );

		// <-- private static data members -->
		static int s_iTileMaxIndex; // larger than the highest tile index ever used. iterations over the tile collection can be limited to this value
		static int s_iTileCount;    // the number of tiles that have been added and not deleted by this class

	
#ifdef ASURA_TOOLS
		
		static Tools_DynamicArray<Asura_String> s_xAsuraFileNameCollection; // collection of full paths to all asr files from which tiles have been loaded
#endif

};


// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_TILEDTEXTUREMANAGEMENT_H___
