//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE        :  Asura_Time_System.h
//
//  CREATED     :  Adam A. Brown  (9.1.2004)
//
//  Copyright (C) 2004 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_TIME_SYSTEM_H__
#define __ASURA_TIME_SYSTEM_H__

// The classes here deal with system date and time, its formatting and display.
// For game time intervals, see Asura_Timers

#include "Core/Asura_PlatformBasics.h"

////////////////////////////////////////
// enums
////////////////////////////////////////
enum ASURA_DATE_FORMAT
{
	ASURA_DATE_DDMMYY,
	ASURA_DATE_DDMMYYYY,
	ASURA_DATE_MMDDYY,
	ASURA_DATE_MMDDYYYY,
	ASURA_DATE_YYMMDD,
	ASURA_DATE_YYYYMMDD,
};


//  AAB:  (12/01/2004 11:47:34):  I have assumed that am and pm don't need localising.
// In general, there doesn't appear to be any accepted convention aboud dealing with
// twelve-hour postfixes in other languages, mainly because (in Europe at least), the
// 24-hour clock is universally used.
enum ASURA_TIME_FORMAT
{
	ASURA_24_HOUR,
	ASURA_12_HOUR,
};


//  AAB:  (12/01/2004 16:21:46):  Currently, times will all print before dates.
enum ASURA_DATE_STRING_FORMAT
{
	ASURA_DATE_MINIMAL,
	ASURA_DATE_SHORT,
	ASURA_DATE_LONG,		//  AAB:  (12/01/2004 16:20:57):  Long format doesn't 
							// currently support multiple orderings of weekday / month 
							// / year.
};


//------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////
//
// Asura_Time - class for storing date and time information. Everything is
// measured from zero at midnight, 1st January 00 (1 BC if you're being pedantic)
//
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
class Asura_Time
{
public:

	Asura_Time(){}

	Asura_Time(unsigned short usYear, unsigned char ucMonth, unsigned char ucDay, unsigned char ucHour, unsigned char ucMinute, unsigned char ucSecond)
	: m_usYear(usYear)
	, m_ucMonth(ucMonth)
	, m_ucDayOfMonth(ucDay)
	, m_ucHour(ucHour)
	, m_ucMinute(ucMinute)
	, m_ucSecond(ucSecond)
	{
	}

	// uLibCTime is number of seconds since 00:00 UCT (GMT) on 1/1/1970 
	void FillFromLibCTime(u_int uLibCTime)
	{
		Platform_FillFromLibCTime(uLibCTime);
	}

	void ConvertUTCToLocalTime();

	bool operator ==(const Asura_Time& xTime) const
	{
		return
		(
			m_usYear == xTime.m_usYear &&
			m_ucMonth == xTime.m_ucMonth &&
			m_ucDayOfMonth == xTime.m_ucDayOfMonth &&
			m_ucHour == xTime.m_ucHour &&
			m_ucMinute == xTime.m_ucMinute &&
			m_ucSecond == xTime.m_ucSecond
		);
	}

	bool operator >(const Asura_Time& xTime) const
	{
		if (m_usYear < xTime.m_usYear)					{	return false;	}
		else if (m_usYear > xTime.m_usYear)				{	return true;	}
		else if (m_ucMonth < xTime.m_ucMonth)			{	return false;	}
		else if (m_ucMonth > xTime.m_ucMonth)			{	return true;	}
		else if (m_ucDayOfMonth < xTime.m_ucDayOfMonth)	{	return false;	}
		else if (m_ucDayOfMonth > xTime.m_ucDayOfMonth)	{	return true;	}
		else if (m_ucHour < xTime.m_ucHour)				{	return false;	}
		else if (m_ucHour > xTime.m_ucHour)				{	return true;	}
		else if (m_ucMinute < xTime.m_ucMinute)			{	return false;	}
		else if (m_ucMinute > xTime.m_ucMinute)			{	return true;	}
		else if (m_ucSecond <= xTime.m_ucSecond)		{	return false;	}
		{	return true;	}
	}

	u_short m_usYear;		// 0 AD = 0
	u_char m_ucMonth;		// Jan = 0
	u_char m_ucDayOfMonth;	
	u_char m_ucDayOfWeek;	// Sun = 0
	u_char m_ucHour;		// Midnight = 0
	u_char m_ucMinute;		
	u_char m_ucSecond;

private:

	void Platform_FillFromLibCTime(u_int uLibCTime);	// uLibCTime is number of seconds since 00:00 UCT (GMT) on 1/1/1970 
};



//------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////
//
// Asura_Time_System
//
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------
class Asura_Time_System
{
	friend class Asura_Time;	// to access GetDayNumber

public:

	//Main public methods.
	static const Asura_Time& GetCurrentDateAndTime();	
	static const Asura_UnicodeChar* GetDateAndTimeString(const Asura_Time& xTime, ASURA_DATE_STRING_FORMAT eStringFormat);
	static const Asura_UnicodeChar* GetTimeString(const Asura_Time& xTime);
	static const Asura_UnicodeChar* GetDateString(const Asura_Time& xTime, ASURA_DATE_STRING_FORMAT eStringFormat);

	//Utility functions
	static const Asura_UnicodeChar* GetCurrentDateAndTimeString(ASURA_DATE_STRING_FORMAT eStringFormat);
	static const Asura_UnicodeChar* GetCurrentDateString(ASURA_DATE_STRING_FORMAT eStringFormat);
	static const Asura_UnicodeChar* GetCurrentTimeString();	
	static u_int GetDaysSince( const Asura_Time& xEndTime, const Asura_Time& xStartTime );
	static bool  GetTimeFromAsctimeDate( const char* szTimeString, Asura_Time& xOutput );
	static int GetTimeZoneBias() { return Platform_GetTimeZoneBias(); } // in minutes, where UTC = local time + bias 
	
private:

	static ASURA_DATE_FORMAT Project_GetDateFormat();
	static ASURA_TIME_FORMAT Project_GetTimeFormat();
	static const Asura_UnicodeChar* Project_GetDayOfWeekName(int iWeekDay);
	static const Asura_UnicodeChar* Project_GetMonthName(int iMonth);
	static const Asura_UnicodeChar* Project_GetDayOfWeekShortName(int iWeekDay);

	static int GetDayOfWeek(const Asura_Time& xTime);
	static int GetDayNumber(const Asura_Time& xTime);	// returns number of days since 1/1/70
	static int GetNextLeapYear(const Asura_Time& xTime);
	static int GetPrevLeapYear(const Asura_Time& xTime);
	static int GetLeapYearNumber(unsigned short usYear);
	static int GetDaysSoFarInYear(const Asura_Time& xTime,bool bIncludeLeap=true);
	static bool IsLeapYear(unsigned short usYear);

	static void Platform_GetCurrentDateAndTime();
	static int Platform_GetTimeZoneBias(); // in minutes, where UTC = local time + bias 

	static int s_aiDaysInMonthsNormal[12];
	static int s_aiDaysInMonthsLeap[12];
	static Asura_Time s_xTime;
	static Asura_UnicodeChar s_wszDateBuffer[64];
	static Asura_UnicodeChar s_wszDateAndTimeBuffer[64];
	static Asura_UnicodeChar s_wszTimeBuffer[16];
};



#endif//__ASURA_TIME_SYSTEM_H__
