//
// Asura_Timers.cpp
//
// Created by Kevin Lea - 05/10/2001 16:38:13
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_Timers.h"

// Asura includes
#include "Core/Asura_Maths.h"
#include "Core/Asura_Multithread.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

// (MP - 21.02.2007) Project stuff can break with higher timescales and I
// think it's most useful to be able to reduce this. For the time being at 
// least, clamp the maximum relatively low.
static const float fMAX_TIMESCALE_VIA_DVS = 2.0f;

// D E B U G V A R S /////////////////////////////////////////////////////////

DEBUGVAR float dbg_fMinFrameRate = 5.0f;

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// Proportion of each frame time which can be changed to sync the global time.
// so if a frame takes 0.05s, and the ratio is 0.2, we can add on 0.04s to 
// 0.06s to correct clock drift. Any further change will wait till next frame.
// I'm allowing the clock to be advanced significantly, but only reduced by 
// adding as little as 1% of the actual frame time. Never subtract from the 
// global time - we don't want time flowing backwards!
DEBUGVAR float dbg_fMaxGlobalTimeChangeRatioBackwards = 0.99f;
DEBUGVAR float dbg_fMaxGlobalTimeChangeRatioForwards = 1.01f;
DEBUGVAR float dbg_fMinGlobalTimeAdjust = 0.01f;	// only sync up if we're badly out

DEBUGVAR bool dbg_bPrintTimeSyncs = false;

#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
float Asura_Timers::s_fLocalUnscaledTime;
float Asura_Timers::s_fLocalScaledTime;
float Asura_Timers::s_fPreviousFrameTime;
float Asura_Timers::s_fFrameRate;
float Asura_Timers::s_fSmoothedFrameRate;
float Asura_Timers::s_fFrameTime;
float Asura_Timers::s_fGameTime;
float Asura_Timers::s_fGameTimeScale;
float Asura_Timers::s_fTickFrequency;
unsigned int Asura_Timers::s_uFrameNumber;
unsigned int Asura_Timers::s_uMilliSecondsSinceStart;
unsigned int Asura_Timers::s_uMicroSecondsRemainder;
float Asura_Timers::s_fServerAbsoluteTime;
#ifdef ASURA_TOOLS
float Asura_Timers::s_fUncappedPhase;
#endif // ASURA_TOOLS
float Asura_Timers::s_fTicksPerMilliSecond;
float Asura_Timers::s_fTicksPerMicroSecond;

float Asura_Timers::s_fFixedFrameRate;
float Asura_Timers::s_fFixedFrameTime;
float Asura_Timers::s_fActualFrameRate;
float Asura_Timers::s_fActualFrameTime;
float Asura_Timers::s_fMaxFrameRate;
bool Asura_Timers::s_bSystemPause = false;

// G L O B A L S /////////////////////////////////////////////////////////////

static float g_fFixedGameTime = 0.0f;

static const unsigned NUM_FRAME_RATE_ENTRIES = 10;
static const float ONE_OVER_NUM_FRAME_RATE_ENTRIES = 1.0f / NUM_FRAME_RATE_ENTRIES;
static float g_afFPSTable[NUM_FRAME_RATE_ENTRIES];
static unsigned g_uCurFPSTableEntry = 0;

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// Table to average out global time difference over about 3 seconds.
// In theory we receive messages about every 0.1s from server. This is enforced
// by the function which processes them.
static const unsigned ASURA_NUM_GLOBAL_TIME_ENTRIES = 30;
static const float ASURA_ONE_OVER_NUM_GLOBAL_TIME_ENTRIES = 1.0f / ASURA_NUM_GLOBAL_TIME_ENTRIES;
static float g_afGlobalTimeTable[ASURA_NUM_GLOBAL_TIME_ENTRIES];
static unsigned	g_uCurGlobalTimeEntry = 0;

// number of seconds our global time is off from last update.
static float g_fGlobalTimeSyncOffset = 0.0f;
// first update received after Reset() forces instant sync
static bool g_bGlobalTimeHasBeenSynchronised = false;

#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_Timers::Initialise( void )
{
	Platform_Initialise();

    s_uFrameNumber = 0;
	s_uMilliSecondsSinceStart = 0;

	Reset();

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#ifndef ASURA_COMMAND_CONSOLE_READONLY
    Asura_CommandConsole::AddVar( L"Game.TimeScale", &s_fGameTimeScale, 0.0f, fMAX_TIMESCALE_VIA_DVS, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif //ifndef ASURA_COMMAND_CONSOLE_READONLY
#endif //ASURA_SUPPORT_COMMAND_CONSOLE
}

void Asura_Timers::Reset( void )
{
	Platform_Reset();

	s_fTickFrequency = Platform_GetTickFrequency();
	s_fTicksPerMilliSecond = s_fTickFrequency * 0.001f;
	s_fTicksPerMicroSecond = s_fTickFrequency * 0.000001f;
	
	s_fFrameRate = Platform_GetDefaultFrameRate();
    s_fSmoothedFrameRate = s_fFrameRate;
	s_fFrameTime = 1.0f/s_fFrameRate;
	s_fPreviousFrameTime = s_fFrameTime;
	s_fLocalUnscaledTime = 0.0f;
	s_fLocalScaledTime = 0.0f;
	g_fFixedGameTime = 0.0f;
	s_fFixedFrameRate = 0.0f;
	s_fFixedFrameTime = 0.0f;
	s_fActualFrameRate = s_fFrameRate;
	s_fActualFrameTime = s_fFrameTime;

	#ifdef ASURA_TOOLS
	s_fUncappedPhase = 0.0f;
	#endif //ASURA_TOOLS

	for (u_int u = 0; u < NUM_FRAME_RATE_ENTRIES; ++u)
	{
		g_afFPSTable[u] = s_fFrameRate;
	}

	SetGameTimeScale(1.0f);

	s_fServerAbsoluteTime = 0.0f;

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	g_fGlobalTimeSyncOffset = 0.0f;
	g_bGlobalTimeHasBeenSynchronised = false;
#endif
}

void Asura_Timers::Shutdown( void )
{
	Platform_Shutdown();
}

void Asura_Timers::HandlePrimaryTimer(void)
{
	Platform_Update();

    // Cap frame-rate. Platform code can deal with this in Platform_Update()
    // above if it wishes, in which case the code below should end up doing
    // nothing except a single check that we've accumulated sufficient frame
    // time.
	{
        const float fEffectiveMinFrameTime = GetEffectiveMinFrameTime();
        if( fEffectiveMinFrameTime > 0.0f )
        {
		    while( GetTimeInFrame_Uncapped() < fEffectiveMinFrameTime )
		    {
			    Asura_Thread_System::YieldCurrentThread();
		    }
        }
	}

	const u_int uTicksThisFrame = Platform_GetTicksThisFrame();
	const float fTicksThisFrame = static_cast<float>(uTicksThisFrame);
	{
		int iElapsedMillisecs = Asura_Maths::FloatToInt_Chop(fTicksThisFrame / s_fTicksPerMilliSecond);
		const int iElapsedMicrosecs = Asura_Maths::FloatToInt_Chop(fTicksThisFrame / s_fTicksPerMicroSecond);
		const int iRemainderedMicrosecs = Asura_Maths::ClampToRange(iElapsedMicrosecs - (iElapsedMillisecs * 1000), 0, 999);
		s_uMicroSecondsRemainder += iRemainderedMicrosecs;
		if (s_uMicroSecondsRemainder >= 1000)
		{
			s_uMicroSecondsRemainder -= 1000;
			iElapsedMillisecs++;
		}
		s_uMilliSecondsSinceStart += iElapsedMillisecs;
	}
	s_fPreviousFrameTime = s_fFrameTime;

	const float fUncappedFrameTime = fTicksThisFrame / s_fTickFrequency;

    // set unscaled frametime and framerate
    {
	    if( !uTicksThisFrame )
	    {
		    s_fFrameRate = Platform_GetDefaultFrameRate();
	    }
	    else
	    {
		    s_fFrameRate = s_fTickFrequency / fTicksThisFrame;
	    }

        if (s_fFrameRate < dbg_fMinFrameRate)
        {
		    s_fFrameRate = dbg_fMinFrameRate;
        }

        s_fFrameTime = 1.0f / s_fFrameRate;
    }

	// check Platform_Update() clamped framerate if requested
	if (s_fMaxFrameRate > 0.0f)
	{
		Asura_Assert(int(s_fFrameRate) <= int(s_fMaxFrameRate), "Platform_Update() didn't clamp framerate");
	}

	s_fActualFrameRate = s_fFrameRate;
	s_fActualFrameTime = s_fFrameTime;

	// increment time stamp counter
	s_fLocalUnscaledTime += s_fFrameTime;

	#ifdef ASURA_TOOLS
	s_fUncappedPhase += fUncappedFrameTime;
	#endif //ASURA_TOOLS

	// set scaled gametime
	if(g_fFixedGameTime>0)
	{
		s_fGameTime = g_fFixedGameTime * GetGameTimeScale();
	}
	else
	{
		s_fGameTime = s_fFrameTime * GetGameTimeScale();
	}

	// increment game time stamp
	s_fLocalScaledTime += s_fGameTime;

	s_uFrameNumber++;

// MH 12.4.02 -- code to average out frame rate
// (MH - 26.09.2002) sadly float inaccuracies mean this results in cumulative errors.
// We will have to do the sum every frame instead. The method below would work if we
// kept the sum of integer ticks, but that would require a 64-bit long int type on 
// PS2 and probably other platforms too, so probably slower than doing the whole sum.
//	g_fFPSSum -= g_afFPSTable[g_uCurFPSTableEntry];
//	g_fFPSSum += s_fFrameRate;
//	g_afFPSTable[g_uCurFPSTableEntry++] = s_fFrameRate;
//	if( g_uCurFPSTableEntry == NUM_FRAME_RATE_ENTRIES )
//	{
//		g_uCurFPSTableEntry = 0;
//	}
//	s_fSmoothedFrameRate = g_fFPSSum * ONE_OVER_NUM_FRAME_RATE_ENTRIES;
    g_afFPSTable[g_uCurFPSTableEntry++] = s_fFrameRate;
	if( g_uCurFPSTableEntry == NUM_FRAME_RATE_ENTRIES )
	{
		g_uCurFPSTableEntry = 0;
	}
    float fFPSSum = 0;
    for (u_int i = 0; i < NUM_FRAME_RATE_ENTRIES; i++)
    {
        fFPSSum += g_afFPSTable[i];
    }
    s_fSmoothedFrameRate = fFPSSum * ONE_OVER_NUM_FRAME_RATE_ENTRIES;

	// update global game time
	float fGlobalTimeIncrement = fUncappedFrameTime;

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	if (g_bGlobalTimeHasBeenSynchronised)
	{
		// this should only happen on the client, as the server will never 
		// send a message to itself to synchronise.

		// update each table entry's estimate of time on the server
		for (u_int i = 0; i < ASURA_NUM_GLOBAL_TIME_ENTRIES; i++)
		{
			g_afGlobalTimeTable[i] += fGlobalTimeIncrement;
		}
		
		// apply the average offset (calculated the last time we processed
		// a time offset) if there is a big enough difference.
		if (g_fGlobalTimeSyncOffset > dbg_fMinGlobalTimeAdjust)
		{
			// we're ahead of the server, reduce the time added this frame
			Asura_Assert(dbg_fMaxGlobalTimeChangeRatioBackwards <= 1.0f, "time will go backwards");
			float fMaxAbsChange = s_fGameTime * dbg_fMaxGlobalTimeChangeRatioBackwards;
			if (g_fGlobalTimeSyncOffset < fMaxAbsChange)
			{
				fMaxAbsChange = g_fGlobalTimeSyncOffset;
			}
			fGlobalTimeIncrement -= fMaxAbsChange;
			g_fGlobalTimeSyncOffset -= fMaxAbsChange;
		}
		else if (g_fGlobalTimeSyncOffset < -dbg_fMinGlobalTimeAdjust)
		{
			// we're behind the server, increase the time added this frame
			float fMaxAbsChange = s_fGameTime * dbg_fMaxGlobalTimeChangeRatioForwards;
			if ((-g_fGlobalTimeSyncOffset) < fMaxAbsChange)
			{
				fMaxAbsChange = -g_fGlobalTimeSyncOffset;
			}
			fGlobalTimeIncrement += fMaxAbsChange;
			g_fGlobalTimeSyncOffset += fMaxAbsChange;
		}
	}
#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	
	s_fServerAbsoluteTime += fGlobalTimeIncrement;
}
  
float Asura_Timers::GetTimeInFrame(void)
{
	const float fTicksThisFrame = static_cast< float >( Platform_GetTicksInFrame() );

	float fTimeInframe = ( fTicksThisFrame / s_fTickFrequency );
	float fMaxFrameTime = 1.0f / dbg_fMinFrameRate;

	if ( fTimeInframe > fMaxFrameTime )
	{
		return fMaxFrameTime;
	}

	return fTimeInframe;
}

float Asura_Timers::GetTimeInFrame_Uncapped(void)
{
	const float fTicksThisFrame = static_cast< float >( Platform_GetTicksInFrame() );

	return ( fTicksThisFrame / s_fTickFrequency );
}

void Asura_Timers::SetGameTimeScale(float fScale)
{
	// If the value is too small we want to set it to 0. Assert so that the user knows we have changed it
	if (fScale > 0.0f && fScale < 1e-6f)
	{
		Asura_Assert(false,"SetGameTimeScale(float scale): fScale is too low. Setting to 0.");
		fScale = 0;
	}
	s_fGameTimeScale = fScale;
}

#ifdef ASURA_TOOLS
void Asura_Timers::SetFixedGameTime(float fTime)
{
	g_fFixedGameTime = fTime;
}

float Asura_Timers::GetFixedGameTime()
{
	return g_fFixedGameTime;
}
#endif


void Asura_Timers::UpdateServerAbsoluteTime(float fNewServerTime, float fCurrentLatency, bool bInstant /* = false */)
{
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	// (MH - 11.09.2003) - if we want to guess the time on the server, we should 
	// add on the time it likely took for this message to arrive. If we don't, our 
	// global time lags behind the server which is what we want for displaying our 
	// lagged view of the world to the player. Small differences in latency should 
	// be sorted out by the global time smoothing which only allows the clock to 
	// be advanced by frametime +/- 1%. 
	float fCurrentTimeOnServer = fNewServerTime;// + (fCurrentLatency / 2.0f);
	if (bInstant || !g_bGlobalTimeHasBeenSynchronised)
	{
		g_bGlobalTimeHasBeenSynchronised = true;
		s_fServerAbsoluteTime = fCurrentTimeOnServer;
		g_fGlobalTimeSyncOffset = 0.0f;

		// set the whole table to match this value
		for (u_int i = 0; i < ASURA_NUM_GLOBAL_TIME_ENTRIES; i++)
		{
			g_afGlobalTimeTable[i] = fCurrentTimeOnServer;
		}
	}
	else
	{
		// add this estimate to our table
		g_afGlobalTimeTable[g_uCurGlobalTimeEntry++] = fCurrentTimeOnServer;
		if( g_uCurGlobalTimeEntry == ASURA_NUM_GLOBAL_TIME_ENTRIES )
		{
			g_uCurGlobalTimeEntry = 0;
		}

		// find the average estimate
		float fTimeSum = 0;
		for (u_int i = 0; i < ASURA_NUM_GLOBAL_TIME_ENTRIES; i++)
		{
			fTimeSum += g_afGlobalTimeTable[i];
		}

		float fEstimatedTimeOnServer = (fTimeSum * ASURA_ONE_OVER_NUM_GLOBAL_TIME_ENTRIES);
		
	    g_fGlobalTimeSyncOffset = s_fServerAbsoluteTime - fEstimatedTimeOnServer;
	}

	if (Asura_Maths::Abs(g_fGlobalTimeSyncOffset) > 1.0f)
	{
		Asura_OutputToDebugger_Retail(("WARNING: time diff from server = %.3f. latency %.3f.\n", Asura_Maths::FtoD(g_fGlobalTimeSyncOffset), Asura_Maths::FtoD(fCurrentLatency)));
	}
	else if (dbg_bPrintTimeSyncs)
	{
		Asura_OutputToDebugger(("time diff %.3f. latency %.3f.\n", Asura_Maths::FtoD(g_fGlobalTimeSyncOffset), Asura_Maths::FtoD(fCurrentLatency)));
	}
#else
	Asura_Assert(false, "shouldn't receive time sync messages if not supporting external network");
#endif
}


void Asura_Timers::NeedToResyncServerAbsoluteTime()
{
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	g_bGlobalTimeHasBeenSynchronised = false;
#endif
}

unsigned int Asura_Timers::GetMSecsSinceStart_Continuous()
{
	u_int uMsecsInFrame = Asura_Maths::FloatToInt(GetTimeInFrame_Uncapped() * 1000.0f);
	return GetMSecsSinceStart() + uMsecsInFrame;
}

bool Asura_Timers::HasServerAbsoluteTimeBeenSynchronised()
{
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	return g_bGlobalTimeHasBeenSynchronised;
#else
	return false;//Hopefully doesn't matter what is returned here in this case. RAR 28/02/2005
#endif
}


float Asura_Timers::GetEffectiveMinFrameTime( void )
{
    // Presumably fixed framerate should take precedence since that's used
    // in specific situations when we want to run at an exact framerate,
    // including faking frame times, etc.
    if( s_fFixedFrameTime > 0.0f )
    {
        return s_fFixedFrameTime;
    }
    else if( s_fMaxFrameRate > 0.0f )
    {
        return( 1.0f / s_fMaxFrameRate );
    }
    else
    {
        return 0.0f;
    }
}