//////////////////////////////////////////////////////////////////////////////
//
// Asura_MemAllocator.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
// Description:
//
//     class to manage allocation of a block of memory e.g. VRAM where you
//     don't want the allocation records to be kept in the memory arena.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Memory/Asura_MemAllocator.h"
#include "Core/Asura_Maths.h"

#include "Debugging/Asura_Assert.h"

// T Y P E D E F S ///////////////////////////////////////////////////////////

//For 32/64 bit portability, note that this is something of a hack, we should really have the
//'pointer as a u_int' type defined in a global header somewhere. Since currently Windows is the only
//64 bit platform we can get away with it, future 64 bit platforms will need to define it appropriately
//uintptr_t might be defined on all platforms, but it's not guaranteed so sticking with this...

#if defined(_WIN64) || defined(_WIN32)	//Windows
typedef UINT_PTR Asura_PointerAsUnsigned;
#else									//Other platforms (which are currently all 32bit)
typedef u_int Asura_PointerAsUnsigned;
#endif

// D E B U G V A R S /////////////////////////////////////////////////////////

#if _DEBUG
DEBUGVAR bool dbg_bMemAllocatorIntegrityChecks = false;
#endif

// C L A S S E S /////////////////////////////////////////////////////////////
class Asura_MemAllocator_AllocRecord
{
public:

	void* m_pAddress;
	u_int m_uSize;
	Asura_MemAllocator_AllocRecord* m_pxNext;
};


// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
Asura_MemAllocator::Asura_MemAllocator( void* pArena, u_int uSize, u_int uMinAlignment, u_int uMaximumAllocRecords, u_int uMinAllocSize /*= 0*/ )
: m_pArena(pArena)
, m_uArenaSize(uSize)
, m_uMinAlignment(uMinAlignment)
, m_xRecordPool( sizeof( Asura_MemAllocator_AllocRecord ), uMaximumAllocRecords )
, m_uMinAllocSize(uMinAllocSize)
, m_pxFreeList(NULL)
, m_pxUsedList(NULL)
, m_uFreeSpace(uSize)
{
	Asura_Assert(uMinAlignment > 0 && (uMinAlignment & (uMinAlignment - 1)) == 0, "must specify power of two alignment");
	// init free list to entire arena
	m_pxFreeList = CreateAllocRecord( pArena, uSize, NULL );
}

Asura_MemAllocator::~Asura_MemAllocator()
{
	Asura_Assert(GetLargestFreeBlock() == m_uArenaSize, "all memory should be free at destruction");
	Asura_Assert(GetNumUsedBlocks() == 0, "outstanding allocations in Asura_MemAllocator on destruction");
	Asura_Assert(GetNumFreeBlocks() == 1, "fragmented free list in Asura_MemAllocator on destruction");
	Asura_Assert(m_xRecordPool.AllocatedThisMemory(m_pxFreeList), "free list pointer incorrect");
	Asura_Assert(!m_pxUsedList, "outstanding allocations in Asura_MemAllocator on destruction");
	Asura_Assert(m_pxFreeList && m_pxFreeList->m_uSize == m_uArenaSize && m_pxFreeList->m_pAddress == m_pArena, 
		"outstanding allocations in Asura_MemAllocator on destruction");
	ReturnAllocRecord(m_pxFreeList);
}

void Asura_MemAllocator::Wipe()
{
	MutexAcquire();
	{
		m_xRecordPool.Reset();
		m_uFreeSpace = m_uArenaSize;
		m_pxFreeList = NULL;
		m_pxUsedList = NULL;

		// init free list to entire arena
		m_pxFreeList = CreateAllocRecord( m_pArena, m_uArenaSize, NULL );
	}
	MutexRelease();
}

Asura_MemAllocator_AllocRecord* Asura_MemAllocator::CreateAllocRecord( void* pAddress, 
																	   u_int uSize, 
																	   Asura_MemAllocator_AllocRecord* pxNext )
{
	Asura_MemAllocator_AllocRecord* pxRecord = static_cast<Asura_MemAllocator_AllocRecord*>(m_xRecordPool.Allocate());

	if (pxRecord)
	{
		pxRecord->m_pAddress = pAddress;
		pxRecord->m_uSize = uSize;
		pxRecord->m_pxNext = pxNext;
	}
	else
	{
		Asura_Assert(pxRecord, "Failed to create alloc record. Possible cause: Maximum number of allocs is to small.");
	}

	return pxRecord;
}
void Asura_MemAllocator::ReturnAllocRecord( Asura_MemAllocator_AllocRecord* pxRecord )
{
	m_xRecordPool.Deallocate( pxRecord );
}

void* Asura_MemAllocator::Allocate( u_int uBytes, u_int uAlignment )
{
	void* pMem = NULL;

	MutexAcquire();
	{
		Asura_Assert(uAlignment > 0 && (uAlignment & (uAlignment - 1)) == 0, "must specify power of two alignment");

#if _DEBUG
		if (dbg_bMemAllocatorIntegrityChecks)
		{
			IntegrityCheck();
		}
#endif

		//increase the size of the allocation to be at least the min allocation size
		uBytes = Asura_Maths::Max(m_uMinAllocSize, uBytes);
		// align the allocation size to the min alignment allowed by the heap
		uBytes = (uBytes + m_uMinAlignment - 1) & ~(m_uMinAlignment - 1);
		// alignment is whichever is larger of min and requested alignment
		uAlignment = Asura_Maths::Max(m_uMinAlignment, uAlignment);

		Asura_MemAllocator_AllocRecord* pxBlock = m_pxFreeList;
		Asura_MemAllocator_AllocRecord* pxBestBlock = NULL;
		Asura_MemAllocator_AllocRecord* pxPrevBlock = NULL;
		Asura_MemAllocator_AllocRecord* pxBestPrevBlock = NULL; // store prev block to fix things up if free block is got rid of
		u_int uSmallestBlockSize = 0xffffffff;
		u_int uBytesToLeaveFreeAtStart = 0; // for alignment
		while (pxBlock)
		{
			// search for best-fit block
			const u_int uSize = pxBlock->m_uSize;
			// big enough?
			if (uSize >= uBytes)
			{
				char* pcAlignedAddress = reinterpret_cast<char*>
					( (reinterpret_cast<Asura_PointerAsUnsigned>(pxBlock->m_pAddress) + uAlignment - 1) & ~(uAlignment - 1));
				u_int uExtraBytes = pcAlignedAddress - reinterpret_cast<char*>(pxBlock->m_pAddress);
				// block big enough for the extra bytes for alignment?
				if (uSize >= uBytes + uExtraBytes)
				{
					// is it smaller that the best one so far?
					if (uSize < uSmallestBlockSize)
					{
						uSmallestBlockSize = uSize;
						pxBestBlock = pxBlock;
						pxBestPrevBlock = pxPrevBlock;
						uBytesToLeaveFreeAtStart = uExtraBytes;
						if (uSize == uBytes)
						{
							// can't get better than this.
							break;
						}
					}
				}
			}
			pxPrevBlock = pxBlock;
			pxBlock = pxBlock->m_pxNext;
		}

		// did we find a suitable block?
		if (pxBestBlock)
		{
			// use memory from the beginnning of the free block, but offset for alignment if necessary
			pMem = reinterpret_cast<char*>(pxBestBlock->m_pAddress) + uBytesToLeaveFreeAtStart;
			// sort out the free block
			pxBestBlock->m_uSize -= uBytes;

			// any space left in block?
			if (pxBestBlock->m_uSize)
			{
				// check if we have to create a free block at the start
				if (!uBytesToLeaveFreeAtStart)
				{
					// set new block pointer
					pxBestBlock->m_pAddress = reinterpret_cast<char*>(pxBestBlock->m_pAddress) + uBytes;
				}
				else
				{
					// if the block was just big enough for alignment and the allocation we don't need 
					// to do anything as it's already pointing to the start and is the right size.
					if (pxBestBlock->m_uSize != uBytesToLeaveFreeAtStart)
					{
						// leave the extra bit at the start where it is and create a new record
						// for the remaining free mem and insert it before the alignment part 
						// in the free list.
						Asura_Assert(pxBestBlock->m_uSize > uBytesToLeaveFreeAtStart, "block size should include bytes to leave at start");
						u_int uFreeMemAfterSize = pxBestBlock->m_uSize - uBytesToLeaveFreeAtStart;
						void* pFreeMemAfter = reinterpret_cast<char*>(pMem) + uBytes;

						// adjust the block size of the initial part
						pxBestBlock->m_uSize = uBytesToLeaveFreeAtStart;

						// create and insert new free record
						Asura_MemAllocator_AllocRecord* pxNewFreeBlock = CreateAllocRecord( pFreeMemAfter, uFreeMemAfterSize, pxBestBlock );
						if (pxNewFreeBlock)
						{
							if (pxBestPrevBlock)
							{
								pxBestPrevBlock->m_pxNext = pxNewFreeBlock;
							}
							else
							{
								Asura_Assert(pxBestBlock == m_pxFreeList, "internal bug");
								m_pxFreeList = pxNewFreeBlock;
							}
						}
					}
				}
			}
			else
			{
				Asura_Assert(!uBytesToLeaveFreeAtStart, "internal bug");
				// no longer free - fix up linked list and return block
				if (pxBestPrevBlock)
				{
					pxBestPrevBlock->m_pxNext = pxBestBlock->m_pxNext;
				}
				else
				{
					Asura_Assert(m_pxFreeList == pxBestBlock, "internal bug");
					m_pxFreeList = pxBestBlock->m_pxNext;
				}
				ReturnAllocRecord( pxBestBlock );
			}

			// link new record to the head of the used list
			Asura_MemAllocator_AllocRecord* pxNewRecord = CreateAllocRecord( pMem, uBytes, m_pxUsedList );
			if (pxNewRecord)
			{
				m_pxUsedList = pxNewRecord;
			}

			// update free space counter
			Asura_Assert(m_uFreeSpace >= uBytes, "internal bug - free space counter incorrect");
			m_uFreeSpace -= uBytes;
		}

#if _DEBUG
		if (dbg_bMemAllocatorIntegrityChecks)
		{
			IntegrityCheck();
		}
#endif
	}
	MutexRelease();
	return pMem;
}

void Asura_MemAllocator::Free( void* pAddress )
{
	MutexAcquire();
	{
		Asura_Assert(m_pxUsedList, "free called when no outstanding records exist");

#if _DEBUG
		if (dbg_bMemAllocatorIntegrityChecks)
		{
			IntegrityCheck();
		}
#endif

		Asura_MemAllocator_AllocRecord* pxBlock = m_pxUsedList;
		Asura_MemAllocator_AllocRecord* pxPrevBlock = NULL;
		while (pxBlock)
		{
			if (pxBlock->m_pAddress == pAddress)
			{
				const u_int uSize = pxBlock->m_uSize;

				// found it. fix up the free list by seeing if any free block starts 
				// at the end address or ends at the start address.
				void* const pEndAddress = reinterpret_cast<char*>(pAddress) + uSize;

				Asura_MemAllocator_AllocRecord* pxFreeBlock = m_pxFreeList;
				Asura_MemAllocator_AllocRecord* pxPrevFreeBlock = NULL;
				Asura_MemAllocator_AllocRecord* pxMergedFreeBlock = NULL;
				// keep going until we've looked at everything or merged top and bottom
				while (pxFreeBlock)
				{
					if (pxFreeBlock->m_pAddress == pEndAddress)
					{
						// there is a free block immediately after the block we're freeing
						if (!pxMergedFreeBlock)
						{
							// grow the free block downwards
							pxFreeBlock->m_pAddress = pAddress;
							pxFreeBlock->m_uSize += uSize;
							pxMergedFreeBlock = pxFreeBlock;
						}
						else
						{
							// already merged another block so this is the other side. merge this free block into the other one.
							pxMergedFreeBlock->m_uSize += pxFreeBlock->m_uSize;
							// and return the free block.
							// (first fix up free list)
							if (pxPrevFreeBlock)
							{
								pxPrevFreeBlock->m_pxNext = pxFreeBlock->m_pxNext;
							}
							else
							{
								Asura_Assert(m_pxFreeList == pxFreeBlock, "internal bug");
								m_pxFreeList = pxFreeBlock->m_pxNext;
							}
							ReturnAllocRecord(pxFreeBlock);
							break;
						}
					}
					else if (reinterpret_cast<char*>(pxFreeBlock->m_pAddress) + pxFreeBlock->m_uSize == pAddress)
					{
						// there is a free block immediately before the block we're freeing
						if (!pxMergedFreeBlock)
						{
							// grow free block up
							pxFreeBlock->m_uSize += uSize;
							pxMergedFreeBlock = pxFreeBlock;
						}
						else
						{
							// already merged another block so this is the other side. merge this free block into the other one.
							pxMergedFreeBlock->m_pAddress = pxFreeBlock->m_pAddress;
							pxMergedFreeBlock->m_uSize += pxFreeBlock->m_uSize;
							// and return the free block.
							// (first fix up free list)
							if (pxPrevFreeBlock)
							{
								pxPrevFreeBlock->m_pxNext = pxFreeBlock->m_pxNext;
							}
							else
							{
								Asura_Assert(m_pxFreeList == pxFreeBlock, "internal bug");
								m_pxFreeList = pxFreeBlock->m_pxNext;
							}
							ReturnAllocRecord(pxFreeBlock);
							break;
						}
					}

					pxPrevFreeBlock = pxFreeBlock;
					pxFreeBlock = pxFreeBlock->m_pxNext;
				}


				// fix up the used block list
				if (pxPrevBlock)
				{
					pxPrevBlock->m_pxNext = pxBlock->m_pxNext;
				}
				else
				{
					Asura_Assert(m_pxUsedList == pxBlock, "internal bug");
					m_pxUsedList = pxBlock->m_pxNext;
				}

				// if we didn't merge anything we must create a new record for the free block.
				// might as well just put this block in the free list (at the head).
				if (!pxMergedFreeBlock)
				{
					pxBlock->m_pxNext = m_pxFreeList;
					m_pxFreeList = pxBlock;
				}
				else
				{
					// block no longer needed - return it.
					ReturnAllocRecord(pxBlock);
				}

				// update free space counter
				m_uFreeSpace += uSize;

#if _DEBUG
				if (dbg_bMemAllocatorIntegrityChecks)
				{
					IntegrityCheck();
				}
#endif
				MutexRelease();
				return;
			}
			pxPrevBlock = pxBlock;
			pxBlock = pxBlock->m_pxNext;
		}

		Asura_Assert(false, "no record for allocated block");
	}
	MutexRelease();
}

u_int Asura_MemAllocator::GetLargestFreeBlock() const
{
	u_int uLargest = 0;

	MutexAcquire();
	{
		Asura_MemAllocator_AllocRecord* pxFreeBlock = m_pxFreeList;
		while (pxFreeBlock)
		{
			if (pxFreeBlock->m_uSize > uLargest)
			{
				uLargest = pxFreeBlock->m_uSize;
			}
			pxFreeBlock = pxFreeBlock->m_pxNext;
		}
	}
	MutexRelease();
	return uLargest;
}

u_int Asura_MemAllocator::GetNumUsedBlocks() const
{
	u_int uNum = 0;

	MutexAcquire();
	{
		Asura_MemAllocator_AllocRecord* pxUsedBlock = m_pxUsedList;
		while (pxUsedBlock)
		{
			++uNum;
			pxUsedBlock = pxUsedBlock->m_pxNext;
		}
	}
	MutexRelease();
	return uNum;
}

u_int Asura_MemAllocator::GetNumFreeBlocks() const
{
	u_int uNum = 0;

	MutexAcquire();
	{
		Asura_MemAllocator_AllocRecord* pxFreeBlock = m_pxFreeList;
		while (pxFreeBlock)
		{
			++uNum;
			pxFreeBlock = pxFreeBlock->m_pxNext;
		}
	}
	MutexRelease();
	return uNum;
}

bool Asura_MemAllocator::IntegrityCheck() const
{
	bool bOK = true;

	MutexAcquire();
	{
		// check we can compute the used space correctly and that it matches our internal record of it.
		{
			u_int uUsedSize = 0;
			Asura_MemAllocator_AllocRecord* pxUsedBlock = m_pxUsedList;
			while (pxUsedBlock)
			{
				uUsedSize += pxUsedBlock->m_uSize;
				pxUsedBlock = pxUsedBlock->m_pxNext;
			}

			Asura_Assert(uUsedSize == GetUsedSpace(), "computed used space didn't match stored used space");	
			bOK = uUsedSize == GetUsedSpace();
		}
		// check we can compute the free space correctly and that it matches our internal record of it.
		{
			u_int uFreeSize = 0;
			Asura_MemAllocator_AllocRecord* pxFreeBlock = m_pxFreeList;
			while (pxFreeBlock)
			{
				uFreeSize += pxFreeBlock->m_uSize;
				pxFreeBlock = pxFreeBlock->m_pxNext;
			}

			Asura_Assert(uFreeSize == GetFreeSpace(), "computed free space didn't match stored free space");	
			if (bOK)
			{
				bOK = uFreeSize == GetFreeSpace();
			}
		}
	}
	MutexRelease();
	return bOK;
}

#include "UnitTest/Asura_UnitTest.h"

#ifdef ASURA_SUPPORT_UNITTEST

ASURA_UNITTEST(MemAllocator, GeneralTest)
{
	const u_int uMinAlign = 4;
	const u_int uInitialAllocRecords = 16;
	const u_int uHeapSize = 64;
	char acBuffer[uHeapSize];
	Asura_MemAllocator xHeap( acBuffer, uHeapSize, uMinAlign, uInitialAllocRecords );

	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());

	// alloc all
	void* pMem = xHeap.Allocate(uHeapSize);
	ASURA_UNITTEST_CHECK( pMem != NULL );
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == uHeapSize);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 1);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 0);

	// free all
	xHeap.Free(pMem);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 0);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 0);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// alloc 1 (will do 4)
	pMem = xHeap.Allocate(1);
	ASURA_UNITTEST_CHECK( pMem != NULL );
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 4);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 1);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// alloc another 1 (will do 4)
	void* pMem2 = xHeap.Allocate(1);
	ASURA_UNITTEST_CHECK( pMem2 != NULL );
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 8);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 2);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// alloc another 1 (will do 4)
	void* pMem3 = xHeap.Allocate(1);
	ASURA_UNITTEST_CHECK( pMem3 != NULL );
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 12);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 3);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// free the middle one
	xHeap.Free(pMem2);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 8);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 2);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 2);

	// alloc 4 (should reuse that free space in the middle)
	pMem2 = xHeap.Allocate(4);
	ASURA_UNITTEST_CHECK( pMem2 != NULL );
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 12);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 3);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// alloc what should be rest of heap (64 - (4 + 4 + 4))
	void* pMem4 = xHeap.Allocate(52);
	ASURA_UNITTEST_CHECK( pMem4 != NULL );
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 64);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 4);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 0);

	// free 1st one
	xHeap.Free(pMem);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 60);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 3);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// free last one
	xHeap.Free(pMem4);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 8);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 2);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 2);

	// free 1st one of 2 remaining (should have space on one side)
	xHeap.Free(pMem2);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 4);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 1);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 2);

	// free last one (should have space on both sides)
	xHeap.Free(pMem3);
	// check all is as it should be
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	ASURA_UNITTEST_CHECK(xHeap.GetUsedSpace() == 0);
	ASURA_UNITTEST_CHECK(xHeap.GetFreeSpace() == uHeapSize);
	ASURA_UNITTEST_CHECK(xHeap.GetLargestFreeBlock() == uHeapSize);
	ASURA_UNITTEST_CHECK(xHeap.GetNumUsedBlocks() == 0);
	ASURA_UNITTEST_CHECK(xHeap.GetNumFreeBlocks() == 1);

	// check alignment is working
	pMem = xHeap.Allocate(4, 16);
	Asura_Assert(pMem && (reinterpret_cast<Asura_PointerAsUnsigned>(pMem) & 15) == 0, "memory was not aligned");
	ASURA_UNITTEST_CHECK(pMem && (reinterpret_cast<Asura_PointerAsUnsigned>(pMem) & 15) == 0);
	pMem2 = xHeap.Allocate(4, 16);
	Asura_Assert(pMem2 && (reinterpret_cast<Asura_PointerAsUnsigned>(pMem2) & 15) == 0, "memory was not aligned");
	ASURA_UNITTEST_CHECK(pMem2 && (reinterpret_cast<Asura_PointerAsUnsigned>(pMem) & 15) == 0);
	pMem3 = xHeap.Allocate(4, 32);
	Asura_Assert(pMem3 && (reinterpret_cast<Asura_PointerAsUnsigned>(pMem3) & 31) == 0, "memory was not aligned");
	ASURA_UNITTEST_CHECK(pMem3 && (reinterpret_cast<Asura_PointerAsUnsigned>(pMem) & 15) == 0);
	pMem4 = xHeap.Allocate(16);
	Asura_Assert(pMem4, "aligned memory allocs waste memory");
	ASURA_UNITTEST_CHECK(pMem4);

	// now free them all and check the heap has survived
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	xHeap.Free(pMem3);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	xHeap.Free(pMem4);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	xHeap.Free(pMem);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
	xHeap.Free(pMem2);
	ASURA_UNITTEST_CHECK(xHeap.IntegrityCheck());
}

#endif // #ifdef ASURA_SUPPORT_UNITTEST


void Asura_MemAllocator::MutexAcquire(void) const
{
	m_xMutex.Lock();
}

void Asura_MemAllocator::MutexRelease(void) const
{
	m_xMutex.Unlock();
}

// eof
