///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ResourceSetWithOverride.h
// Author : Andy Southgate
// Date   : 15-Jan-2009
//
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_RESOURCESETWITHOVERRIDE_H__
#define __ASURA_RESOURCESETWITHOVERRIDE_H__


#include "Core/Asura_List.h"

//////////////////////////////////////////////////////////////////////////
//
// class Asura_ResourceSetOverride
//
// Stores the hash and name (only in tools) of the extry to pick instead
//
// Note templated just so that each type T gives a different typed
// override which cannot be assigned to each other
//
//////////////////////////////////////////////////////////////////////////
template <typename T> class Asura_ResourceSetOverride
{
private:
	// Private variables...
#ifdef ASURA_TOOLS
	Asura_String	m_xOldName;		// The name of the entry to be replaced (used by Tools).
	Asura_String	m_xNewName;		// The name of the entry to be used instead(used by Tools).

#endif	// defined(ASURA_TOOLS)
	Asura_Hash_ID	m_uHash;		// The hash-ID of the entry to map to.

public:
	Asura_ResourceSetOverride() {}
	Asura_ResourceSetOverride(const Asura_ResourceSetOverride &xOther) {*this = xOther;}

	Asura_ResourceSetOverride & operator= (const Asura_ResourceSetOverride &xOther);

#ifdef ASURA_TOOLS
	void				SetOldName	(const Asura_String& xName)	{ m_xOldName = xName; }
	void				SetNewName	(const Asura_String& xName)	{ m_xNewName = xName; m_uHash = m_xNewName.GetHash(); }
	const Asura_String&	GetOldName	(void) const				{ return(m_xOldName); }
	const Asura_String&	GetNewName	(void) const				{ return(m_xNewName); }
#endif	// defined(ASURA_TOOLS)

	void			SetHash	(const Asura_Hash_ID uHash)	;
	Asura_Hash_ID	GetHash	(void) const				{ return(m_uHash); }
};
//////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSetOverrideMap_Iterator;
template <class T> class Asura_ResourceSetOverrideMap_ConstIterator;

//////////////////////////////////////////////////////////////////////////
//
// class Asura_ResourceSetOverrideMap
//
// Keeps a map of Asura_ResourceSetOverride<T>'s
//
//////////////////////////////////////////////////////////////////////////
template <typename T> class Asura_ResourceSetOverrideMap
{
	friend class Asura_ResourceSetOverrideMap_Iterator<T>;
	friend class Asura_ResourceSetOverrideMap_ConstIterator<T>;
private:
	// Private typedefs...
	typedef Asura_TemplateHashTree< Asura_ResourceSetOverride<T> > HashTree;

#ifdef ASURA_TOOLS
	Asura_String	m_xName;				// The name of this override-map (used by Tools).
#endif
	HashTree		m_xHashTree;			// Hash-tree of overrides.

public:
	Asura_ResourceSetOverrideMap  (void)		{}
	~Asura_ResourceSetOverrideMap ()			{}

	Asura_Hash_ID GetOverride (Asura_Hash_ID uHash) const;

	Asura_ResourceSetOverride<T> *		GetOverrideEntry(Asura_Hash_ID uHash)		{ return m_xHashTree.GetDataByHash(uHash);}
	const Asura_ResourceSetOverride<T> *GetOverrideEntry(Asura_Hash_ID uHash) const { return m_xHashTree.GetDataByHash(uHash);}

	void CreateOverride (Asura_Hash_ID uHashFrom, Asura_Hash_ID uHashTo);
#ifdef ASURA_TOOLS
	void CreateOverride (Asura_String xFrom, Asura_String xTo);
#endif

	void RemoveOverride (Asura_Hash_ID uHashFrom) { m_xHashTree.DeleteNodeByHash(uHashFrom);}

#ifdef ASURA_TOOLS

	void		SetName	(const char* szName)	{ m_xName = szName; }
	const char*	GetName	(void) const			{ return(m_xName.GetString()); }

	u_int GetNumEntries (void) const			{ return m_xHashTree.GetNumberOfNodes();}

// 	Asura_ResourceSetOverride<T> *		GetOverrideEntryByIndex(u_int uIndex)		{ return m_xHashTree.GetDataByIndex(uIndex);}
// 	const Asura_ResourceSetOverride<T> *GetOverrideEntryByIndex(u_int uIndex) const { return m_xHashTree.GetDataByIndex(uIndex);}
#endif	// defined(ASURA_TOOLS)
};

//////////////////////////////////////////////////////////////////////////

template <typename T> class Asura_ResourceSetWithOverride_ResourceIterator;
template <typename T> class Asura_ResourceSetWithOverride_ConstResourceIterator;
template <typename T> class Asura_ResourceSetWithOverride_OverrideMapIterator;
template <typename T> class Asura_ResourceSetWithOverride_ConstOverrideMapIterator;

//////////////////////////////////////////////////////////////////////////
//
// class Asura_ResourceSetWithOverride
//
// Works as an Asura_ResourceSet but with the concept of override maps
//
//////////////////////////////////////////////////////////////////////////
template <typename T> class Asura_ResourceSetWithOverride
{
	friend class Asura_ResourceSetWithOverride_ResourceIterator<T>;
	friend class Asura_ResourceSetWithOverride_ConstResourceIterator<T>;
	friend class Asura_ResourceSetWithOverride_OverrideMapIterator<T>;
	friend class Asura_ResourceSetWithOverride_ConstOverrideMapIterator<T>;

private:
	typedef Asura_TemplateHashTree< Asura_ResourceSetOverrideMap<T> > HashTree;

	Asura_ResourceSet<T>	m_xResourceSet;			// Resource set of the resources
	
	Asura_Hash_ID			m_uOverrideMapEnabled;	// The ID of the currently enabled override map (or ASURA_HASH_ID_UNSET)
	HashTree				m_xOverrideMaps;		// Hashtree of all loaded override maps

public:
	// Public functions...
	Asura_ResourceSetWithOverride(const char* szName)	: m_xResourceSet(szName)	{}
	virtual ~Asura_ResourceSetWithOverride() {FreeAllOverrideMaps();}

	// Asura_ResourceSet_Base Functions

	// Reset functions...
	void	Reset(void)									{ m_xResourceSet.Reset(); FreeAllOverrideMaps();}
	void	ResetBank(const Asura_Hash_ID uBankHash)	{ m_xResourceSet.ResetBank(); }

	// Accessor functions...
	const Asura_String&	GetName(void) const	{ return(m_xResourceSet.GetName()); }
	Asura_Hash_ID		GetHash(void) const	{ return(m_xResourceSet.GetHash()); }

	// Asura_ResourceSetFunctions

	// Resource manipulation functions...
	bool		AddResource(const Asura_Hash_ID uResourceHash, T& xData, bool bAssertIfAlreadyInBank = true)			{ return m_xResourceSet.AddResource(uResourceHash, xData, bAssertIfAlreadyInBank); }

	// Accessor functions...
	T*			GetResource(const Asura_Hash_ID uResourceHash)						{ return(m_xResourceSet.GetResource(GetOverride(uResourceHash))); }
	const T*	GetResource(const Asura_Hash_ID uResourceHash) const				{ return(m_xResourceSet.GetResource(GetOverride(uResourceHash))); }
	
	T*			GetResourceWithoutOverride(const Asura_Hash_ID uResourceHash)		{ return(m_xResourceSet.GetResource(uResourceHash)); }
	const T*	GetResourceWithoutOverride(const Asura_Hash_ID uResourceHash) const	{ return(m_xResourceSet.GetResource(uResourceHash)); }

	u_int		GetNumResources(void) const											{ return(m_xResourceSet.GetNumResources()); }

	// and the overridemap functions

	void		EnableOverrideMap(Asura_Hash_ID uHash, bool bDisableAllPreviouslyEnabledMaps = true);
	void		DisableOverrideMap(Asura_Hash_ID uHash);
	bool		IsOverrideMapEnabled(Asura_Hash_ID uHash)	{ return m_uOverrideMapEnabled == uHash; }
	void		DisableAllOverrideMaps()					{ m_uOverrideMapEnabled = ASURA_HASH_ID_UNSET;	}

	Asura_ResourceSetOverrideMap<T> *	AllocateOverrideMap(Asura_Hash_ID uHash)	{ return m_xOverrideMaps.AddOrFindNode(uHash); }
	Asura_ResourceSetOverrideMap<T> *	GetOverrideMapByHash(Asura_Hash_ID uHash)	{ return m_xOverrideMaps.GetDataByHash(uHash); }
	void FreeOverrideMap(Asura_Hash_ID uHash)										{ DisableOverrideMap(uHash); m_xOverrideMaps.DeleteNodeByHash(uHash); }
	void FreeAllOverrideMaps()														{ m_uOverrideMapEnabled = ASURA_HASH_ID_UNSET; m_xOverrideMaps.Reset(); }

	u_int GetNumOverrideMaps() const		{ return m_xOverrideMaps.GetNumberOfNodes();}
	u_int GetNumEnabledOverrideMaps() const { return m_uOverrideMapEnabled != ASURA_HASH_ID_UNSET;}

	Asura_Hash_ID GetOverride(Asura_Hash_ID uHash) const;
#ifdef ASURA_TOOLS
	bool GetOverrideNewName(Asura_Hash_ID uHash, Asura_String& xReturnString) const;
	bool GetOverrideOldName(Asura_Hash_ID uHash, Asura_String& xReturnString) const;
#endif

#if defined(ASURA_TOOLS)
	bool	DeleteResource(const Asura_Hash_ID uResourceHash)												{ return m_xResourceSet.DeleteResource(uResourceHash); }
	bool	RelocateResource(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash)	{ return m_xResourceSet.RelocateResource(uOldResourceHash, uNewResourceHash); }

	T*			GetResourceByIndex(const u_int uIndex)														{ return(m_xResourceSet.GetResourceByIndex(uIndex)); }
	const T*	GetResourceByIndex(const u_int uIndex) const												{ return(m_xResourceSet.GetResourceByIndex(uIndex)); }
#endif	//defined(ASURA_TOOLS)
};


//////////////////////////////////////////////////////////////////////////
//
// class Asura_ResourceSetWithOverride_ResourceIterator
//
// Iterate over an Asura_ResourceSetWithOverride enumerating the resources
//
//////////////////////////////////////////////////////////////////////////
template <class T> class Asura_ResourceSetWithOverride_ResourceIterator
{
private:
	// Private variables...
	Asura_ResourceSet_Iterator<T>	m_xIterator;	// ConstIterator for the unified data.

	// Private functions...
	Asura_ResourceSetWithOverride_ResourceIterator(void);
	Asura_ResourceSetWithOverride_ResourceIterator(const Asura_ResourceSetWithOverride_ResourceIterator<T>& xOther);
	Asura_ResourceSetWithOverride_ResourceIterator<T>&	operator=(const Asura_ResourceSetWithOverride_ResourceIterator<T>& xOther);

public:
	// Public functions...
	Asura_ResourceSetWithOverride_ResourceIterator(Asura_ResourceSetWithOverride<T>& xResourceSet) : m_xIterator(xResourceSet.m_xResourceSet) {}

	~Asura_ResourceSetWithOverride_ResourceIterator()	{}

	// Manipulation functions...
	void			Next(void)					{ m_xIterator.Next(); }
	void			End(void)					{ m_xIterator.End(); }

	bool			Done(void) const			{ return(m_xIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xIterator.GetCurrentHash()); }
	T*				GetCurrent(void)			{ return(m_xIterator.GetCurrent()); }
};
//////////////////////////////////////////////////////////////////////////
template <class T> class Asura_ResourceSetWithOverride_ConstResourceIterator
{
private:
	// Private variables...
	Asura_ResourceSet_ConstIterator<T>	m_xIterator;	// ConstIterator for the unified data.

	// Private functions...
	Asura_ResourceSetWithOverride_ConstResourceIterator(void);
	Asura_ResourceSetWithOverride_ConstResourceIterator(const Asura_ResourceSetWithOverride_ConstResourceIterator<T>& xOther);
	Asura_ResourceSetWithOverride_ConstResourceIterator<T>&	operator=(const Asura_ResourceSetWithOverride_ConstResourceIterator<T>& xOther);

public:
	// Public functions...
	Asura_ResourceSetWithOverride_ConstResourceIterator(const Asura_ResourceSetWithOverride<T>& xResourceSet) : m_xIterator(xResourceSet.m_xResourceSet) {}

	~Asura_ResourceSetWithOverride_ConstResourceIterator()	{}

	// Manipulation functions...
	void			Next(void)					{ m_xIterator.Next(); }
	void			End(void)					{ m_xIterator.End(); }

	bool			Done(void) const			{ return(m_xIterator.Done()); }
	Asura_Hash_ID	GetCurrentHash(void) const	{ return(m_xIterator.GetCurrentHash()); }
	const T*		GetCurrent(void) const		{ return(m_xIterator.GetCurrent()); }
};
//////////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////////
//
// class Asura_ResourceSetWithOverride_ResourceIterator
//
// Iterate over an Asura_ResourceSetWithOverride enumerating the override maps
//
//////////////////////////////////////////////////////////////////////////
template <class T> class Asura_ResourceSetWithOverride_OverrideMapIterator
{
private:

	// Private variables...
	Asura_TemplateHashTree_Iterator<Asura_ResourceSetOverrideMap<T> >	m_xIterator;	// Iterator for the unified data.
	// Private functions...
	Asura_ResourceSetWithOverride_OverrideMapIterator(void);
	Asura_ResourceSetWithOverride_OverrideMapIterator(const Asura_ResourceSetWithOverride_OverrideMapIterator& xOther);
	Asura_ResourceSetWithOverride_OverrideMapIterator&	operator=(const Asura_ResourceSetWithOverride_OverrideMapIterator& xOther);

public:
	// Public functions...
	Asura_ResourceSetWithOverride_OverrideMapIterator(Asura_ResourceSetWithOverride<T>& xResourceSet) : m_xIterator(xResourceSet.m_xOverrideMaps) {}
	~Asura_ResourceSetWithOverride_OverrideMapIterator()	{}

	// Manipulation functions...
	void								Next(void)					{ m_xIterator.Next();}
	void								End(void)					{ m_xIterator.End();}

	bool								Done(void) const			{ return m_xIterator.Done();}
	Asura_ResourceSetOverrideMap<T>*	GetCurrent(void) const		{ return m_xIterator.GetCurrent();}
	Asura_Hash_ID						GetCurrentHash(void) const	{ return m_xIterator.GetCurrentHash();}
};
//////////////////////////////////////////////////////////////////////////
template <class T> class Asura_ResourceSetWithOverride_ConstOverrideMapIterator
{
private:
	// Private variables...
	Asura_TemplateHashTree_ConstIterator<Asura_ResourceSetOverrideMap<T> >	m_xIterator;	// Iterator for the unified data.
	// Private functions...
	Asura_ResourceSetWithOverride_ConstOverrideMapIterator(void);
	Asura_ResourceSetWithOverride_ConstOverrideMapIterator(const Asura_ResourceSetWithOverride_ConstOverrideMapIterator& xOther);
	Asura_ResourceSetWithOverride_ConstOverrideMapIterator&	operator=(const Asura_ResourceSetWithOverride_ConstOverrideMapIterator& xOther);
public:
	// Public functions...
	Asura_ResourceSetWithOverride_ConstOverrideMapIterator(const Asura_ResourceSetWithOverride<T>& xResourceSet) : m_xIterator(xResourceSet.m_xOverrideMaps) {}
	~Asura_ResourceSetWithOverride_ConstOverrideMapIterator()	{}

	// Manipulation functions...
	void									Next(void)					{ m_xIterator.Next();}
	void									End(void)					{ m_xIterator.End();}

	bool									Done(void) const			{ return m_xIterator.Done();}
	const Asura_ResourceSetOverrideMap<T>*	GetCurrent(void) const		{ return m_xIterator.GetCurrent();}
	Asura_Hash_ID							GetCurrentHash(void) const	{ return m_xIterator.GetCurrentHash();}
};
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
//
// class Asura_ResourceSetWithOverride_ResourceIterator
//
// Iterate over an Asura_ResourceSetOverrideMap enumerating the overrides
//
//////////////////////////////////////////////////////////////////////////
template <class T> class Asura_ResourceSetOverrideMap_Iterator
{
private:
	// Private variables...
	Asura_TemplateHashTree_Iterator<Asura_ResourceSetOverride<T> > m_xIterator;	// Iterator for the unified data.

	// Private functions...
	Asura_ResourceSetOverrideMap_Iterator(void);
	Asura_ResourceSetOverrideMap_Iterator(const Asura_ResourceSetOverrideMap_Iterator& xOther);
	Asura_ResourceSetOverrideMap_Iterator&	operator=(const Asura_ResourceSetOverrideMap_Iterator& xOther);
public:
	// Public functions...
	Asura_ResourceSetOverrideMap_Iterator(Asura_ResourceSetOverrideMap<T>& xOverrideMap) : m_xIterator(xOverrideMap.m_xHashTree) {}
	~Asura_ResourceSetOverrideMap_Iterator()	{}

	// Manipulation functions...
	void			Next(void) { m_xIterator.Next();}
	void			End(void) { m_xIterator.End();}

	bool			Done(void) const {  { return m_xIterator.Done();}}
	Asura_ResourceSetOverride<T>*	GetCurrent(void) const { return m_xIterator.GetCurrent();}
	Asura_Hash_ID	GetCurrentHash(void) const { return m_xIterator.GetCurrentHash();}
};
//////////////////////////////////////////////////////////////////////////
template <class T> class Asura_ResourceSetOverrideMap_ConstIterator
{
private:
	// Private variables...
	Asura_TemplateHashTree_ConstIterator<Asura_ResourceSetOverride<T> > m_xIterator;	// Iterator for the unified data.

	// Private functions...
	Asura_ResourceSetOverrideMap_ConstIterator(void);
	Asura_ResourceSetOverrideMap_ConstIterator(const Asura_ResourceSetOverrideMap_ConstIterator& xOther);
	Asura_ResourceSetOverrideMap_ConstIterator&	operator=(const Asura_ResourceSetOverrideMap_ConstIterator& xOther);
public:
	// Public functions...
	Asura_ResourceSetOverrideMap_ConstIterator(const Asura_ResourceSetOverrideMap<T>& xOverrideMap) : m_xIterator(xOverrideMap.m_xHashTree) {}
	~Asura_ResourceSetOverrideMap_ConstIterator()	{}

	// Manipulation functions...
	void								Next(void)					{ m_xIterator.Next();}
	void								End(void)					{ m_xIterator.End();}
	bool								Done(void) const			{ return m_xIterator.Done();}
	const Asura_ResourceSetOverride<T>*	GetCurrent(void) const		{ return m_xIterator.GetCurrent();}
	Asura_Hash_ID						GetCurrentHash(void) const	{ return m_xIterator.GetCurrentHash();}
};
//////////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetOverride::operator=
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline Asura_ResourceSetOverride<T> & Asura_ResourceSetOverride<T>::operator= (const Asura_ResourceSetOverride<T> &xOther)
{
#ifdef ASURA_TOOLS
	m_xOldName = xOther.m_xOldName;
	m_xNewName = xOther.m_xOldName;
#endif	// defined(ASURA_TOOLS)
	m_uHash	= xOther.m_uHash;
	return(*this);
}
/// End of function 'Asura_ResourceSetOverride::operator='. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetOverride::SetHash
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline void Asura_ResourceSetOverride<T>::SetHash (const Asura_Hash_ID uHash)	
{ 
	m_uHash = uHash;
#ifdef ASURA_TOOLS
	m_xNewName = "";
#endif
}
/// End of function 'Asura_ResourceSetOverride::SetHash'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetOverrideMap<T>::GetOverride
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline Asura_Hash_ID Asura_ResourceSetOverrideMap<T>::GetOverride (Asura_Hash_ID uHash) const 
{
	const Asura_ResourceSetOverride<T> *pxEntry = GetOverrideEntry(uHash);
	return pxEntry ? pxEntry->GetHash() : uHash;
}
/// End of function 'Asura_ResourceSetOverride::SetHash'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetOverrideMap<T>::CreateOverride
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline void Asura_ResourceSetOverrideMap<T>::CreateOverride (Asura_Hash_ID uHashFrom, Asura_Hash_ID uHashTo)
{
	Asura_ResourceSetOverride<T> * pOverride = m_xHashTree.AddNode(uHashFrom);
	if (pOverride)
	{
		pOverride->SetHash(uHashTo);
	}
}
/// End of function 'Asura_ResourceSetOverride::SetHash'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetOverrideMap<T>::CreateOverride
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_TOOLS
template <typename T> inline void Asura_ResourceSetOverrideMap<T>::CreateOverride (Asura_String xFrom, Asura_String xTo)
{
	Asura_ResourceSetOverride<T> * pOverride = m_xHashTree.AddNode(xFrom.GetHash());
	if (pOverride)
	{
		pOverride->SetOldName(xFrom);
		pOverride->SetNewName(xTo);
	}
}
#endif // defined(ASURA_TOOLS)
/// End of function 'Asura_ResourceSetOverride::SetHash'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetWithOverride<T>::GetOverride
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline Asura_Hash_ID Asura_ResourceSetWithOverride<T>::GetOverride(Asura_Hash_ID uHash) const
{
	Asura_ResourceSetOverrideMap<T>* pxMap = m_uOverrideMapEnabled != ASURA_HASH_ID_UNSET ? m_xOverrideMaps.GetDataByHash(m_uOverrideMapEnabled) : NULL;

	if (pxMap)
	{
		uHash = pxMap->GetOverride(uHash);
	}
	return uHash;
}
/// End of function 'Asura_ResourceSetWithOverride<T>::GetOverride'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetWithOverride<T>::GetOverrideNewName
// Author       : Robert Myall
// Date         : 8-May-2009
////////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_TOOLS
template <typename T> inline bool Asura_ResourceSetWithOverride<T>::GetOverrideNewName(Asura_Hash_ID uHash, Asura_String& xReturnString) const
{
	Asura_ResourceSetOverrideMap<T>* pxMap = m_uOverrideMapEnabled != ASURA_HASH_ID_UNSET ? m_xOverrideMaps.GetDataByHash(m_uOverrideMapEnabled) : NULL;

	if (pxMap)
	{
		const Asura_ResourceSetOverride<T>* pxEntry = pxMap->GetOverrideEntry( uHash );
		if (pxEntry)
		{
			xReturnString = pxEntry->GetNewName();
			return true;
		}
	}
	return false;
}
#endif
/// End of function 'Asura_ResourceSetWithOverride<T>::GetOverrideNewName'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetWithOverride<T>::GetOverrideOldName
// Author       : Robert Myall
// Date         : 8-May-2009
////////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_TOOLS
template <typename T> inline bool Asura_ResourceSetWithOverride<T>::GetOverrideOldName(Asura_Hash_ID uHash, Asura_String& xReturnString) const
{
	Asura_ResourceSetOverrideMap<T>* pxMap = m_uOverrideMapEnabled != ASURA_HASH_ID_UNSET ? m_xOverrideMaps.GetDataByHash(m_uOverrideMapEnabled) : NULL;

	if (pxMap)
	{
		const Asura_ResourceSetOverride<T>* pxEntry = pxMap->GetOverrideEntry( uHash );
		if (pxEntry)
		{
			xReturnString = pxEntry->GetOldName();
			return true;
		}
	}
	return false;
}
#endif
/// End of function 'Asura_ResourceSetWithOverride<T>::GetOverrideOldName'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetWithOverride<T>::EnableOverrideMap
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline void Asura_ResourceSetWithOverride<T>::EnableOverrideMap(Asura_Hash_ID uHash, bool bDisableAllPreviouslyEnabledMaps/* = true*/)
{
	if (bDisableAllPreviouslyEnabledMaps)
	{
		DisableAllOverrideMaps();
	} else {
		Asura_Assert(0, "Multiple enabled override maps not currently supported");
	}
	m_uOverrideMapEnabled = uHash;

}
/// End of function 'Asura_ResourceSetWithOverride<T>::GetOverride'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ResourceSetWithOverride<T>::DisableOverrideMap
// Author       : Andy Southgate
// Date         : 15-Jan-2009
////////////////////////////////////////////////////////////////////////////////
template <typename T> inline void Asura_ResourceSetWithOverride<T>::DisableOverrideMap(Asura_Hash_ID uHash)
{
	if (m_uOverrideMapEnabled == uHash)
	{
		m_uOverrideMapEnabled = ASURA_HASH_ID_UNSET;
	}
}
/// End of function 'Asura_ResourceSetWithOverride<T>::DisableOverrideMap'. ///


#endif //__ASURA_RESOURCESETWITHOVERRIDE_H__
