
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ResourceSet_UnifiedResource.h
// Author : Andy Tate
// Date   : 17-Jun-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_RESOURCESET_UNIFIEDRESOURCE_H__
#define __ASURA_RESOURCESET_UNIFIEDRESOURCE_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_ResourceSet_UnifiedResource
// Author       : Andy Tate
// Date         : 17-Jun-2008
//
// Description  : Template class used to store multiple instances of a single
//				  Asura-resource (one per bank).
////////////////////////////////////////////////////////////////////////////////

template <class T> class Asura_ResourceSet_UnifiedResource
{
public:
	// Public functions...
	Asura_ResourceSet_UnifiedResource(void);
	~Asura_ResourceSet_UnifiedResource();

	// Resource manipulation functions...
	void	Initialise(const Asura_Hash_ID uBankHash, T& xData);

	Asura_ResourceSet_UnifiedResource<T>*	CreateNextResource(const Asura_Hash_ID uBankHash, T& xData);
	bool									DestroyNextResource(void);

	void	CopyData(const Asura_ResourceSet_UnifiedResource<T>& xOther);

	// Accessor functions...
	Asura_Hash_ID	GetBankHash(void) const	{ return(m_uBankHash); }

#if defined(ASURA_TOOLS)
	T*				GetData(const Asura_Hash_ID uBankHash);
#endif	//defined(ASURA_TOOLS)

	T*				GetData(void)			{ return(m_pxData); }
	const T*		GetData(void) const		{ return(m_pxData); }

	Asura_ResourceSet_UnifiedResource*	GetNext(void)		{ return(m_pxNext); }
	Asura_ResourceSet_UnifiedResource*	GetPrevious(void)	{ return(m_pxPrevious); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private typedefs...
	typedef Asura_ResourceSet_UnifiedResource<T> UnifiedResource;

	// Private variables...
	Asura_Hash_ID						m_uBankHash;	// The hash-ID of resource-set "bank" in which this resource can be found.
	T*									m_pxData;		// A pointer to the data stored for this resource in this bank.
	Asura_ResourceSet_UnifiedResource*	m_pxPrevious;	// A pointer to this resource stored for the previous bank.
	Asura_ResourceSet_UnifiedResource*	m_pxNext;		// A pointer to this resource stored for the next bank.

	// Private functions...
	Asura_ResourceSet_UnifiedResource(const Asura_Hash_ID uBankHash, T& xData, UnifiedResource* pxPrevious, UnifiedResource* pxNext);
	Asura_ResourceSet_UnifiedResource(const Asura_ResourceSet_UnifiedResource& xOther);

	Asura_ResourceSet_UnifiedResource&	operator=(const Asura_ResourceSet_UnifiedResource& xOther);

};   // End of class 'Asura_ResourceSet_UnifiedResource'.



////////////////////////////////////////////////////////////////////////////////
// Inlined functions...
////////////////////////////////////////////////////////////////////////////////

template <class T> Asura_ResourceSet_UnifiedResource<T>::Asura_ResourceSet_UnifiedResource(void)
: m_uBankHash(ASURA_HASH_ID_UNSET)
, m_pxData(NULL)
, m_pxPrevious(NULL)
, m_pxNext(NULL)
{
}

template <class T> Asura_ResourceSet_UnifiedResource<T>::Asura_ResourceSet_UnifiedResource(const Asura_Hash_ID uBankHash, T& xData, UnifiedResource* pxPrevious, UnifiedResource* pxNext)
: m_uBankHash(uBankHash)
, m_pxData(&xData)
, m_pxPrevious(pxPrevious)
, m_pxNext(pxNext)
{
}

template <class T> Asura_ResourceSet_UnifiedResource<T>::~Asura_ResourceSet_UnifiedResource()
{
	delete(m_pxNext);
}

template <class T> void Asura_ResourceSet_UnifiedResource<T>::Initialise(const Asura_Hash_ID uBankHash, T& xData)
{
	m_uBankHash	= uBankHash;
	m_pxData	= &xData;
}

template <class T> Asura_ResourceSet_UnifiedResource<T>* Asura_ResourceSet_UnifiedResource<T>::CreateNextResource(const Asura_Hash_ID uBankHash, T& xData)
{
	UnifiedResource*	pxNewResource	= new UnifiedResource (uBankHash, xData, this, m_pxNext);

	// Insert the new resource between the current one and the next one...
	if(m_pxNext != NULL)
	{
		m_pxNext->m_pxPrevious	= pxNewResource;
	}
	m_pxNext	= pxNewResource;
	return(pxNewResource);
}

template <class T> bool Asura_ResourceSet_UnifiedResource<T>::DestroyNextResource(void)
{
	if(m_pxNext == NULL)
	{
		return(false);
	}
	else
	{
		UnifiedResource*	pxBefore	= this;
		UnifiedResource*	pxAfter		= m_pxNext->m_pxNext;

		// Unlink the resource we're about to destroy - prevents further resource from being destroyed...
		m_pxNext->m_pxPrevious	= NULL;
		m_pxNext->m_pxNext		= NULL;
		delete(m_pxNext);

		// Link the previous/next resources together...
		if(pxAfter != NULL)
		{
			pxAfter->m_pxPrevious	= pxBefore;
		}

		m_pxNext	= pxAfter;
		return(true);
	}
}

template <class T> void Asura_ResourceSet_UnifiedResource<T>::CopyData(const Asura_ResourceSet_UnifiedResource<T>& xOther)
{
	// Only copy the data, not the links to the next/previous resources...
	m_uBankHash	= xOther.m_uBankHash;
	m_pxData	= xOther.m_pxData;
}

#if defined(ASURA_TOOLS)

template <class T> T* Asura_ResourceSet_UnifiedResource<T>::GetData(const Asura_Hash_ID uBankHash)
{
	for(UnifiedResource* pxUnifiedResource = this; (pxUnifiedResource != NULL); pxUnifiedResource = pxUnifiedResource->GetNext())
	{
		if(pxUnifiedResource->GetBankHash() == uBankHash)
		{
			return(pxUnifiedResource->GetData());
		}
	}

	return(NULL);
}

#endif	//defined(ASURA_TOOLS)

#endif   // ifndef __ASURA_RESOURCESET_UNIFIEDRESOURCE_H__


/// End of file. ///
