//
// Asura_Task_Stack.cpp
//
// Created by KJFL - 17 January 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_TASK_SYSTEM

// This file's header.
#include "Core/Threads/Asura_Task_Stack.h"

#include "Core/Threads/Asura_Task_System.h"

#include "Core/Asura_Atomic.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_StringBuffer.h"

Asura_Task_Stack::Asura_Task_Stack( const char* szStackName, u_int uMaxStackSize, Asura_TaskFunction pfnTask )
: m_szStackName( szStackName )
, m_uMaxStackSize( uMaxStackSize )
, m_uCurrentStackSize( 0 )
, m_ppTaskStack( 0 )
, m_pxStackSpace( 0 )
, m_bTerminateTasks( false )
, Asura_Task( pfnTask )
{
	m_pxStackSpace = new Asura_Semaphore( "ATS_StackSpace", uMaxStackSize, uMaxStackSize );

	m_ppTaskStack = new void* [ m_uMaxStackSize ];

	// Tell the system about this task
	Asura_Task_System::AddTask( this );
}

Asura_Task_Stack::~Asura_Task_Stack()
{
	m_bTerminateTasks = true;

	// need a lock to make sure no one else is about to signal work is available
	m_xStackAccessMutex.Lock();
	SignalWorkIsNotAvailable();
	m_xStackAccessMutex.Unlock();

	// Ask to be removed from the system - this will not return until any in-progress
	// tasks have completed.
	Asura_Task_System::RemoveTask( this );

	// now it's safe to delete things
	delete m_pxStackSpace;
	delete [] m_ppTaskStack;
}

bool Asura_Task_Stack::AddToStack_NonBlocking( void* pStackData )
{
	// if we've terminated already, just ignore it
	if ( m_bTerminateTasks ) return false;

	// Try to add to the Stack. If it's full, return false.
	if ( m_uCurrentStackSize == m_uMaxStackSize ) return false;

	m_xStackAccessMutex.Lock();

	// Check to see if anyone managed to fill the Stack again before we got to it!
	if ( m_uCurrentStackSize == m_uMaxStackSize || m_bTerminateTasks )
	{
		// unlock the Stack, and admit defeat
		m_xStackAccessMutex.Unlock();
		return false;
	}

	AddToStack_PostLock( pStackData );

	return true;
}

void Asura_Task_Stack::AddToStack( void* pStackData )
{
	// if we've terminated already, just ignore it
	if ( m_bTerminateTasks ) return;

	// Try to add to the Stack. If it's full, wait for space to appear.
	bool bAquiredLock = false;

	do
	{
		// wait until we have some space
		m_pxStackSpace->Wait();

		// try and lock the Stack so we can add to it
		m_xStackAccessMutex.Lock();

		if ( m_bTerminateTasks )
		{
			m_xStackAccessMutex.Unlock();
			return;
		}

		// check to see if anyone managed to fill the Stack again before we got to it!
		if ( m_uCurrentStackSize == m_uMaxStackSize  )
		{
			// unlock the Stack, and we'll just have to wait again
			m_xStackAccessMutex.Unlock();
		}
		else
		{
			// ok, we've got space and the Stack is locked
			bAquiredLock = true;
		}
	}
	while (!bAquiredLock);


	AddToStack_PostLock( pStackData );
} 

void Asura_Task_Stack::AddToStack_PostLock( void* pData )
{
	// signal early as possible, though routines will hit our lock
	SignalWorkIsAvailable();

	m_ppTaskStack[ m_uCurrentStackSize++ ] = pData;

	m_xStackAccessMutex.Unlock();
}

void* Asura_Task_Stack::GetTaskData( void )
{
	m_xStackAccessMutex.Lock();

    // Try to get something from the Stack. Return 0 if nothing is available.
	if (m_uCurrentStackSize==0 || m_bTerminateTasks )
	{
		//SignalWorkIsNotAvailable();
		m_xStackAccessMutex.Unlock();
		return 0;
	}

	// GET FROM Stack HERE
	void* pData = m_ppTaskStack[ --m_uCurrentStackSize ];

	// the Stack has some space now
	m_pxStackSpace->Signal();

	m_xStackAccessMutex.Unlock();

	return pData;
}

u_int Asura_Task_Stack::GetMaxStackSize( void ) const
{
	// Max Stack size is a simple constant so nothing clever has to be done here.
	return m_uMaxStackSize;
}

u_int Asura_Task_Stack::GetCurrentStackSize( void ) const
{
	// Internally we'd need to lock if we wanted to be sure of this, so this method
	// is really only for external things that want to query how things are going.
	return m_uCurrentStackSize;
}

#endif // #ifdef ASURA_USE_TASK_SYSTEM

