//
// Asura_Task_System.h
//
// Created by KJFL - 17 January 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_TASK_SYSTEM_H__
#define __ASURA_TASK_SYSTEM_H__

#ifdef ASURA_USE_TASK_SYSTEM

class Asura_Task;
class Asura_Semaphore;

typedef void (*Asura_TaskFunction)(const void*);

// KJFL - 16 January 2008 - This class manages the worker threads used by the tasks. Since 
// creating threads has a performance and memory overhead, instead this system creates worker
// threads when Initialised that persist for the duration of the main thread. These worker threads
// will go to sleep when there is no work available, thus taking negligible CPU time. As soon as
// there is work to do, they will be signalled and start grabbing tasks from the current task queue.
class Asura_Task_System
{
public:

	// Standard system methods - to be called by Asura_Core only
	static void Initialise( void );
	static void Reset( void );
	static void Shutdown( void );

	// Get the number of worker threads the system created
	static u_int GetNumberOfWorkerThreads( void );

	// Get the number of worker threads which are actually active and processing tasks
	static u_int GetNumberOfActiveThreads( void );

	// Add this Task to the system. Currently there is only provision for one task, but
	// we could make this a queue of tasks. 
	static void AddTask( Asura_Task* pxTask );

	// Remove this Task from the system - this will cause the calling thread to yield
	// until the worker threads have finished their current tasks.
	static void RemoveTask( Asura_Task* pxTask );

	// Called by task queues when they have some new work to be done.
	static void SignalWorkIsAvailable( void );

	// Called by task queues when then temporarily run out of work or wish to terminate.
	static void SignalWorkIsNotAvailable( void );

	// Yields (i.e. tries not to use any processor time) until current task is finished.
	static void YieldUntilTaskFinished( void );

private:

	// The main function of each worker thread.
	static void WorkerThreadFunction( const void* pData );

	// Encapsulation of a task. KJFL - This may go in refactoring.
	static void DoTask( void );

	static Asura_Task* volatile	s_pxCurrentTask;

	static Asura_Semaphore*		s_pxTaskAvailable;
	static Asura_Semaphore*		s_pxTaskFinished;

	static Asura_Semaphore*		s_pxSubTaskAvailable;
	static Asura_Semaphore*		s_pxThreadsTerminated;
	
	static u_int volatile		s_uNumberOfWorkerThreads;
	static u_int volatile		s_uNumberOfActiveThreads;

	static bool	volatile		s_bTerminateThreads;
	static bool	volatile		s_bTerminateTask;
	static bool	volatile		s_bTaskFinished;

};

class Asura_Task
{
	public:	

		// KJFL - ideally I'd handing Adding and Removing the task to the Asura_Task_System here,
		// but the ordering of constructor/destructor calls is counter the order we'd need.
		// i.e. we want to remove the task *before* we destroy any member data of the derived
		// classes.
		Asura_Task( Asura_TaskFunction pfnTaskFn )
		{
			m_pfnTaskFn = pfnTaskFn;
		}
		virtual ~Asura_Task()
		{
		}

		// Returns a pointer to the function the worker threads should use
		// - as passed in to the constructor.
		virtual Asura_TaskFunction GetTaskFunction( void ) const { return m_pfnTaskFn; }

		// Get something to pass as a parameter to the Task Function.
		// Should returns zero if there is no work.
		virtual void* GetTaskData( void ) = 0;

	protected:

		// Called when the Task has some new work to be done.
		void SignalWorkIsAvailable( void ) const { Asura_Task_System::SignalWorkIsAvailable(); }

		// Called when the Task wishes to terminate.
		void SignalWorkIsNotAvailable( void ) const { Asura_Task_System::SignalWorkIsNotAvailable(); }

		Asura_TaskFunction m_pfnTaskFn;

};

#endif //#ifdef ASURA_USE_TASK_SYSTEM

#endif//__ASURA_TASK_SYSTEM_H__

