#include "Core/Asura.h"

#include "Cutscene/Asura_Cutscene.h"

#include "Animation/Asura_Animation.h"        // for checking if skins/anims need loading
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_HashedLocalisedText.h"
#include "Core/Asura_NameTree.h"
#include "Core/Asura_Timers.h"
#include "Cutscene/Asura_Cutscene_Actor.h"
#include "Cutscene/Asura_Cutscene_Attachment.h"
#include "Cutscene/Asura_Cutscene_Event.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Entity/Asura_Entity.h"
#include "Entity/Asura_StaticMessage_System.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif
#include "Network/Asura_Network.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"   // for Asura_Render::CutsceneHasCameraControl
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Audio_Voice.h"
#include "Sound/Asura_Sound.h"
#include "Sound/Asura_StreamingBackgroundSound_System.h"
#include "Sound/Asura_Music.h"
#include "Sound/Asura_MetaMusic_System.h"
#include "Sound/Asura_Sound_Mixer.h"

#ifdef ASURA_TOOLS
#include "Hierarchy/Tools_Animation_Files.h"
#include "Lighting/Tools_LightSets.h"
#endif

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_Cutscene::ToolsPlayback;
#ifdef ASURA_TOOLS
bool Asura_Cutscene::s_bDisableSubtitles = false;
bool Asura_Cutscene::s_bDisableSoundTrack = false;
#endif

// C L A S S   M E T H O D S /////////////////////////////////////////////////
Asura_Cutscene::Asura_Cutscene(const Asura_String& name)
{
	CutsceneName = name;
	CutsceneDirectory = "";

	NumberOfActors = 0;
	Actors = 0;
	NumberOfEvents = 0;
	Events = 0;
	UseSoundTrack = false;
	Playing = false;
    m_bHasEverPlayed = false;
	Timer = 0.0f;
	HadAnUpdate = false;
	m_fCutsceneStartTime = 0;
	CutsceneEndTime = 0;
	m_usEndCutsceneMessageBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;
	m_iCutsceneFlags = AsuraCutsceneFlag_Widescreen|AsuraCutsceneFlag_Skippable;    // (MH - 25.11.2002) this is a good default I think.
	#ifdef ASURA_TOOLS
	m_fInitialFOV_UsedForUpdatingOldCutscenes = Project_GetDefaultInitialFOV();
	#endif
	m_uSubtitleHash = ASURA_HASH_ID_UNSET;
	m_fSubtitleStartTime = 0.0f;
	m_fSubtitleEndTime = 0.0f;
    m_bWaitingForSoundTrack = false;
	m_bAdvancingToTime = false;
	m_bSkipMessagesAfterLoad = false;
	m_uEnabledEvents = 0xffffffff;//default to all enabled
	m_fSoundTrackVolume = 1.0f;
	m_pxActiveCamera = 0;
	m_iNumberOfCameras = -1;
	m_bIsControllingCamera = false;
	m_fInitialNearClip = -1.0f;
	m_fInitialFarClip = -1.0f;
	m_fAudioStartTime = 0.0f;

	SoundTrackHandle = ASURA_INVALID_HANDLE;
	m_fCurrentVolume = 1.0f;

    PlaysOnServer = false;

	#ifdef ASURA_TOOLS
	m_bRecordingWavFile = false;
	#endif //ASURA_TOOLS

	#if defined(ASURA_TOOLS) || defined(ASURA_USE_STREAMING)
	m_uGuid = ASURA_GUID_UNREGISTERED;
	#endif

	m_uLightSets = 0xffffffff;
	m_uMixerPreset = 0;
	m_uMixerPresetUsed = 0;

	Position.Zero();

    m_pxProfileData = NULL;

}

#ifdef ASURA_TOOLS
void Asura_Cutscene::Copy(const Asura_Cutscene* pxCutscene)
{
	Asura_Assert(pxCutscene,"");
	Asura_Assert(!NumberOfActors && !NumberOfEvents,"Should only be used on a newly allocated cutscene");

	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		CutsceneDirectory = pxCutscene->CutsceneDirectory;

		SoundTrackName = pxCutscene->SoundTrackName;

		m_fCutsceneStartTime = pxCutscene->m_fCutsceneStartTime;
		CutsceneEndTime = pxCutscene->CutsceneEndTime;
		m_iCutsceneFlags = pxCutscene->m_iCutsceneFlags;
		PlaysOnServer = pxCutscene->PlaysOnServer;


		UseSoundTrack = pxCutscene->UseSoundTrack;
		Position = pxCutscene->Position;

		m_uEnabledEvents = pxCutscene->m_uEnabledEvents;//probably don't need to copy this... but what the hell.
		m_uLightSets = pxCutscene->m_uLightSets;
		m_uMixerPreset = pxCutscene->m_uMixerPreset;

		//allocate actors
		NumberOfActors = pxCutscene->NumberOfActors;
		Asura_ResourceAllocateArray(Actors, NumberOfActors);
		for(int i=0;i<NumberOfActors;i++)
		{
			Actors[i] = Asura_Cutscene_Actor::AllocateActor(pxCutscene->Actors[i]->Type,PlaysOnServer,this);
			Asura_Assert(Actors[i],"");
		}

		m_usEndCutsceneMessageBlock = CopyMessageBlock(pxCutscene,pxCutscene->m_usEndCutsceneMessageBlock);	

		//copy actors
		for(int i=0;i<NumberOfActors;i++)
		{
			Actors[i]->Copy(pxCutscene->Actors[i]);
		}

		//copy events
		NumberOfEvents = pxCutscene->NumberOfEvents;
		Asura_ResourceAllocateArray(Events, NumberOfEvents);
		for(int i=0;i<NumberOfEvents;i++)
		{
			Events[i] = Asura_Cutscene_Event::AllocateEvent(pxCutscene->Events[i]->m_iType,this);
			Asura_Assert(Events[i],"");

			Events[i]->Copy(pxCutscene->Events[i]);
		}
	}
	Asura_ResourcePopType();
}

void Asura_Cutscene::ListShapesUsed(List<Asura_Hash_ID>& xShapeList)
{
	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->ListShapesUsed(xShapeList);
	}
}

void Asura_Cutscene::BuildUsedTextureTree(Asura_NameTree* pxNameTree)
{
	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->BuildUsedTextureTree(pxNameTree);
	}
}
void Asura_Cutscene::AddExtraFilesToList(List<Asura_String>& xNameList) const
{
	//add extra files for actors (not needed yet , but no harm in having it...)
	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->AddExtraFilesToList(xNameList);
	}

	//and for events
	for(int i=0;i<NumberOfEvents;i++)
	{
		Events[i]->AddExtraFilesToList(xNameList);
	}

}

void Asura_Cutscene::CalculateRelativePosition(const Asura_Matrix_3x3& xCutsceneOrientInverse,const Asura_Vector_3 xCutscenePosition)
{

	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->CalculateRelativePosition(xCutsceneOrientInverse,xCutscenePosition);
	}
	for(int i=0;i<NumberOfEvents;i++)
	{
		Events[i]->CalculateRelativePosition(xCutsceneOrientInverse,xCutscenePosition);
	}
}

void Asura_Cutscene::UpdateRelativeToPosition(const Asura_Matrix_3x3& xCutsceneOrient,const Asura_Vector_3 xCutscenePosition)
{

	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->UpdateRelativeToPosition(xCutsceneOrient,xCutscenePosition);
	}
	for(int i=0;i<NumberOfEvents;i++)
	{
		Events[i]->UpdateRelativeToPosition(xCutsceneOrient,xCutscenePosition);
	}
}

#endif //ASURA_TOOLS

Asura_Cutscene::~Asura_Cutscene()
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		if(Playing) Asura_Cutscene::Finish();

		if( m_pxProfileData ) { delete m_pxProfileData; }

		int i;
		for(i=0;i<NumberOfActors;i++)
		{
			delete Actors[i];
			Actors[i] = NULL;
		}
		Asura_ResourceDeallocateArray(Actors);
		Actors = NULL;
		NumberOfActors = 0;

		for(i=0;i<NumberOfEvents;i++)
		{
			delete Events[i];
			Events[i] = NULL;
		}
		Asura_ResourceDeallocateArray(Events);
		Events = NULL;
		NumberOfEvents = 0;
	}
	Asura_ResourcePopType();
}

Asura_Cutscene* Asura_Cutscene::AllocateCutscene(int iCutsceneType,const Asura_String& xName)
{
	Asura_Cutscene* pxCutscene = Project_AllocateCutscene(iCutsceneType,xName);
	if(!pxCutscene)
	{
		switch(iCutsceneType)
		{
			case ACT_StandardCutscene :
				pxCutscene = new Asura_Cutscene(xName);
				break;
			default :
				Asura_Assert(0,"Unknown cutscene type");
				break;
		}
	}
	return pxCutscene;
}

void Asura_Cutscene::PreDestructionCleanup()
{
	Asura_StaticMessage_System::DeleteMessageBlock(m_usEndCutsceneMessageBlock);
	m_usEndCutsceneMessageBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;

	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->PreDestructionCleanup();
	}
	for(int i=0;i<NumberOfEvents;i++)
	{
		Events[i]->PreDestructionCleanup(this);
	}
	
}



void Asura_Cutscene::Initialise()
{
	if(Playing) Finish();
	
	Timer = 0;
	Playing = true;
    m_bHasEverPlayed = true;
	HadAnUpdate = false;
	m_uSubtitleHash = ASURA_HASH_ID_UNSET;
	m_fSubtitleEndTime = 0.0f;
	m_pxActiveCamera = 0;
	m_bSkipMessagesAfterLoad = false;
	m_fInitialNearClip = -1.0f;
	m_fInitialFarClip = -1.0f;

	//initialise actors
	int i;
	for(i=0;i<NumberOfActors;i++)
	{
		Actors[i]->Initialise();
	}

	for(i=0;i<NumberOfActors;i++)
	{
		//Attachments dealt with in a second pass , since its important that all the actor entities
		//are created first
		Actors[i]->InitialiseAttachments();
	}

	if( ControlsCamera() )
    {
        Asura_Render::SetCutsceneHasCameraControl(true);
		m_bIsControllingCamera =true;
    }

	if(m_iCutsceneFlags & AsuraCutsceneFlag_EndAtLastTrackPoint)
	{
		//update the cutscene end time , by checking the last track time for each actor
		CutsceneEndTime = 0;
		for(i=0;i<NumberOfActors;i++)
		{
			float fTime = Actors[i]->GetLastPointTime();
			if(fTime>CutsceneEndTime) CutsceneEndTime = fTime;
		}
	}

	//and events
	for(i=0;i<NumberOfEvents;i++)
	{
		Events[i]->Initialise();
	}

	SetMixerPreset(m_uMixerPreset);

	#ifdef ASURA_TOOLS
	if(!s_bDisableSoundTrack && !m_bRecordingWavFile)//(in tools the individual wav files will be used instead)
	#endif
	{
		if( ShouldUseSoundTrack() )
		{
			if(!(m_iCutsceneFlags & AsuraCutsceneFlag_DontPauseStreams))
			{
				if (CutsceneSoundTrackIsExclusive())
				{
					//(RAR 29/06/2005) Pause meta music system... by default allow it to have a short fade out
					// AUDIOTODO: Dynamic music system should pause here
					#ifndef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
					Asura_MetaMusic_System::Pause(false);
					#endif
					// (MH - 02.03.2005) next time Asura_StreamingBackgroundSound_System::Update is called
					// it will stop all, so we might as well stop it now so we can use its handle.
					Asura_StreamingBackgroundSound_System::StopAll();

				}
				else
				{
					//(RAR  29/06/2005) Meta music has to be paused even if track isn't exlusive ,since
					//otherwise it will use too many streams at once.
					#ifndef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
					Asura_MetaMusic_System::Pause(false);
					#endif
				}

				//Music has just been paused rather than released , so there may be a shortage
				//of sound handles... if so , we need to be a bit more brutal with it
				//RAR 19/08/2005

				#ifdef ASURA_USE_AUDIO_SYSTEM
					// AUDIOTODO
				#else
					if(Asura_Sound_System::GetNumberActiveStreamingSounds() == Asura_Sound_System::GetMaxStreamingSounds())
					{
						//First option is to get rid of the secondary music track (which would be stopping anyway)
						if(!Asura_Music_System::StopSecondary())
						{
							//Failing that we need to get the meta music system to pause in such
							//a way as to release all sound handles
							//(This should work even if the meta music system isn't being used as such)
							Asura_MetaMusic_System::PauseAndReleaseSound(true);
							//Now paused twice , so unpause to get the count down to one
							Asura_MetaMusic_System::Unpause();

							//if for some reason there still aren't enough sound handles.. just
							//stop everthing
							if(Asura_Sound_System::GetNumberActiveStreamingSounds() == Asura_Sound_System::GetMaxStreamingSounds())
							{
								Asura_Sound_System::EnsureAllStreamingSoundsAreStopped();
							}
						}
					}
				#endif // def ASURA_USE_AUDIO_SYSTEM
			}

			#ifndef ASURA_USE_AUDIO_SYSTEM
				Asura_Assert(Asura_Sound_System::GetNumberActiveStreamingSounds()!=Asura_Sound_System::GetMaxStreamingSounds(),"Cutscene needs to play a sound track, but there aren't enough streams available.");
			#endif

			//tell the cutscene system that we are playing a soundtrack to allow other systems to query it
			Asura_Cutscene_System::SetCutscenePlayingSoundTrack(this,true);

			char szSoundName[256];
			LocaliseSoundName(SoundTrackName.GetString(),szSoundName,256);

			#ifdef ASURA_USE_AUDIO_SYSTEM
				SoundTrackHandle = Asura_Audio_System::CreateStreamingVoice( Asura_GetHashID( szSoundName ), false, false, m_fAudioStartTime);
			#else
				SoundTrackHandle = Asura_Sound_System::CreateStreamingSound( szSoundName);
			#endif

    	    if( ASURA_INVALID_HANDLE != SoundTrackHandle )
    	    {
				#ifdef ASURA_USE_AUDIO_SYSTEM
					Asura_Audio_Voice* const pxVoice = Asura_Audio_System::GetVoice(SoundTrackHandle);
					if(pxVoice)
					{
						pxVoice->Start();
					}
				#else
				    Asura_Sound_System::PlayStreamingSound(SoundTrackHandle,false,true);// no loop, preload
				#endif

				if (ShouldPauseSoundTrackUntilFirstFrame())
				{
					m_bWaitingForSoundTrack = true;

    				// (MP - 19.06.2003) Pause it immediately. This is necessary to get
    				// around the lengthy first frame updates for cutscenes that play at
    				// the start of a level.

					#ifdef ASURA_USE_AUDIO_SYSTEM
						// AUDIOTODO: support pausing
					#else
	    				Asura_Sound_System::PauseStreamingSound( SoundTrackHandle );
					#endif
				}

				if(m_iCutsceneFlags & AsuraCutsceneFlag_TakeLengthFromSoundTrack)
				{
					//update the cutscene length to match the sound track. RAR 11/04/2007

					#ifdef ASURA_USE_AUDIO_SYSTEM
						// AUDIOTODO
						float fSoundTrackLength = 0.0f;
					#else
						float fSoundTrackLength = Asura_Sound_System::GetStreamingSoundLength(SoundTrackHandle);
					#endif

					if(fSoundTrackLength > 0.0f)
					{
						SetCutsceneEndTimeFromRealTimeDuration(fSoundTrackLength);
					}
				}
    	    }
			else
			{
				Asura_Assertf( false, ("Failed to find soundtrack '%s' for cutscene '%s'.",
					szSoundName, GetName().GetString()) );
			}
		}
	}

	//if there is a start time , immediately advance to it
	if(m_fCutsceneStartTime)
	{
		AdvanceToTime(m_fCutsceneStartTime);
	}

    Asura_Cutscene_System::CutsceneStarted(this);
}

void Asura_Cutscene::Finish()
{
	if(!Playing) return;

	Playing = false;

	int i;
	for(i=0;i<NumberOfEvents;i++)
	{
		Events[i]->Finish();
	}
	for(i=0;i<NumberOfActors;i++)
	{
		Actors[i]->Finish();
	}

	
 	Asura_Cutscene_System::SetCutscenePlayingSoundTrack(this, false);
	
	if( ShouldUseSoundTrack() )
	{
 		if(SoundTrackHandle!=ASURA_INVALID_HANDLE)
		{
			#ifdef ASURA_USE_AUDIO_SYSTEM
				Asura_Audio_System::DestroyVoice(SoundTrackHandle);
			#else
				Asura_Sound_System::StopStreamingSound(SoundTrackHandle);
				Asura_Sound_System::ReleaseStreamingSound(SoundTrackHandle);
			#endif
			SoundTrackHandle = ASURA_INVALID_HANDLE;
		}

		#ifdef ASURA_TOOLS
		if(!s_bDisableSoundTrack && !m_bRecordingWavFile)
		#endif
		{
			if (!(m_iCutsceneFlags & AsuraCutsceneFlag_DontPauseStreams))
			{
				// AUDIOTODO
				#ifndef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
				Asura_MetaMusic_System::Unpause();
				#endif
			}
		}
	}

	SetMixerPreset(0);

	//reset clip planes if needed
	UpdateClipPlanes(-1.0f,-1.0f);

    if( ControlsCamera() )
    {
        Asura_Render::SetCutsceneHasCameraControl(false);
    }
	m_bIsControllingCamera = false;

    // (MH - 09.08.2002) problem with return above?
    Asura_Cutscene_System::CutsceneFinished(this);

    if( m_pxProfileData && m_pxProfileData->bProfiling )
    {
#if defined ASURA_SUPPORT_COMMAND_CONSOLE || (!defined ASURA_NO_DEBUG_OUTPUT && !defined NDEBUG)
        float fAvgFPS = ( m_pxProfileData->uNumSamples ? m_pxProfileData->fFPSSum / static_cast<float>(m_pxProfileData->uNumSamples) : 0.0f );
#endif

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
        Asura_CommandConsole::Printf( "Cutscene '%s' finished. Average FPS: %.1f. Min FPS: %.1f. Max tris: %u. Avg tris: %u.",
                                      this->CutsceneName.GetString(),
                                      Asura_Maths::FtoD(fAvgFPS),
                                      Asura_Maths::FtoD(m_pxProfileData->fMinFPS),
                                      m_pxProfileData->uMaxTris,
                                      m_pxProfileData->uTotalTris / Asura_Maths::Max(1u, m_pxProfileData->uNumSamples) );
#else
		Asura_OutputToDebugger(( "Cutscene '%s' finished. Average FPS: %.1f. Min FPS: %.1f. Max tris: %u. Avg tris: %u.\n",
                                      this->CutsceneName.GetString(),
                                      Asura_Maths::FtoD(fAvgFPS),
                                      Asura_Maths::FtoD(m_pxProfileData->fMinFPS),
                                      m_pxProfileData->uMaxTris,
                                      m_pxProfileData->uTotalTris / Asura_Maths::Max(1u, m_pxProfileData->uNumSamples) ));
#endif

        m_pxProfileData->bProfiling = false;
    }
}

// virtual function to do things that should be done when the cutscene has finished playing normally
void Asura_Cutscene::PerformEndOfCutsceneTasks()
{
	if(m_usEndCutsceneMessageBlock!=ASURA_STATICMESSAGE_EMPTYBLOCK)
	{
		// MGH 4/1/2006: Only send messages if this is the server machine.
		// In multiplayer games the server was receiving multiple copies of the messages
		// (one from each machine). This test should eliminate the problem.
		// MGH 15/1/2007: Don't send messages if the m_bSkipMessagesAfterLoad is set; this is
		// to help with the reloading of playing cutscenes in streaming zones.
		if (Asura_Network::IsServer() && !m_bSkipMessagesAfterLoad)
		{
		#if defined(ASURA_USE_STREAMING)
			Asura_StaticMessage_System::SendMessages(m_usEndCutsceneMessageBlock, m_uGuid);
		#else
			Asura_StaticMessage_System::SendMessages(m_usEndCutsceneMessageBlock, ASURA_GUID_UNREGISTERED);
		#endif
		}
	}
}

DEBUGVAR float dbg_fCutsceneForceFrameTime = -1.0f;
//DEBUGVAR float dbg_fCutsceneForceFOV       = 78.0f;

void Asura_Cutscene::Update(float fFrameTime /*=-1*/)
{	
//    {
//        Asura_Camera::SetCameraIndex( 0 );
//
//        if( Asura_Camera::GetHorizontalFieldOfView() == Project_GetDefaultInitialFOV() )
//        {
//		    float fWidescreenRatioModifier = 1.0f;
//		    if (m_iCutsceneFlags & AsuraCutsceneFlag_Widescreen)
//		    {
//			    fWidescreenRatioModifier = Asura_Cutscene_System::GetWidescreenDisplayedXOverYRatio();
//		    }
//        
//            Asura_Camera::SetFieldOfView( Asura_Maths::DegToRad(dbg_fCutsceneForceFOV), fWidescreenRatioModifier );
//        }
//    }

	// RWM - 2008-02-26 - m_bWaitingForSoundTrack only true if ShouldPauseSoundTrackUntilFirstFrame returns true, so no need to check that again here
	if ((SoundTrackHandle != ASURA_INVALID_HANDLE) && m_bWaitingForSoundTrack && !m_bAdvancingToTime)
    {
		// (MP - 19.06.2003) To get around the lengthy first frame updates,
		// we pause the streaming sound initially and only unpause it once
		// we're past the first frame and it's finished preloading.
		// (RAR - 12.06.2005) Changed to ignore frames when advancing time , since those
		//aren't proper frames as such anyway.

		#ifdef ASURA_USE_AUDIO_SYSTEM
			// AUDIOTODO
			if( HadAnUpdate )
			{
				m_bWaitingForSoundTrack = false;
			}
		#else
			if( Asura_Sound_System::StreamingSoundIsReady(SoundTrackHandle) && HadAnUpdate )
			{
			    Asura_Sound_System::ResumeStreamingSound( SoundTrackHandle );
				m_bWaitingForSoundTrack = false;
			}
		#endif

        // pause on first frame. Maybe we should clear screen or soemthing.
		//RAR 04/08/2005 - Don't pause if being previewed in cutscene mode , since it plays havoc with the timeline slider
		if(!ToolsPlayback)
		{
        	fFrameTime = 0;
		}
    }

    if( dbg_fCutsceneForceFrameTime >= 0.0f && !m_bWaitingForSoundTrack )
    {
        fFrameTime = dbg_fCutsceneForceFrameTime;
    }
    
	Asura_Assert(Playing,"");
	//If the cutscene has reached the end time , stop.
	//Except while in tools , where we want to hold the last frame
	if(!(m_iCutsceneFlags & AsuraCutsceneFlag_LoopCutscene))
	{
		if(CutsceneEndTime && (Timer==CutsceneEndTime) && !ToolsPlayback)
		{
			// (MH - 31.10.2002) this used to be after sending the end of cutscene messages.
    	    // I moved it so this cutscene shuts down before starting the next one, to avoid overlaps.
    	    // I suppose you might want it the other way if the next cutscene is supposed to take over
    	    // any entities created by this this cutscene, but that doesn't happen at the moment.
    	    Finish();

    	    PerformEndOfCutsceneTasks();

			return;
		}
	}

	
	if(fFrameTime<0)
	{
		//default to the game frame time
		fFrameTime = Asura_Timers::GetGameTimeSlice();
	}


	Timer += fFrameTime;

	if(CutsceneEndTime && (Timer>=CutsceneEndTime))
	{
		if(m_iCutsceneFlags & AsuraCutsceneFlag_LoopCutscene)
		{
			//Determine where the cutscene loops back to , and the length of the looped section of cutscene
			float fLoopPoint = (m_iCutsceneFlags & AsuraCutsceneFlag_LoopToStartTime) ? m_fCutsceneStartTime : 0.0f;
			float fCutsceneLength = CutsceneEndTime - fLoopPoint;
			//Loop the timer back to the appropriate spot
			if(fCutsceneLength > 0.0f)
			{
				Timer = Asura_Maths::FloatMod(Timer-fLoopPoint, fCutsceneLength);
				Timer += fLoopPoint;
				Timer = Asura_Maths::ClampToRange(Timer, fLoopPoint, CutsceneEndTime);
			}
			else
			{
				Timer = fLoopPoint;
			}


			fFrameTime = Timer;
			//restart the cutscene
			//Temporarily zero the start time to prevent a second update being called in Initialise.
			//The start time has already been included in fFrameTime if needed.
			float fStartTimeCopy = m_fCutsceneStartTime;
			m_fCutsceneStartTime = 0;
			Finish();

#ifdef ASURA_SUPPORT_SERVERCUTSCENES				
			// MGH 5/10/2005: Fix problem with looping server cutscenes.
			if (PlaysOnServer)
			{
				Asura_Entity_System::PruneEntityLists();
			}
#endif //ASURA_SUPPORT_SERVERCUTSCENES

			Initialise();
			m_fCutsceneStartTime = fStartTimeCopy;
			//Timer will have been reset by restarting the cutscene , so set it again
			Timer = fFrameTime;
		}
		else
		{
			//make sure the timer doesn't overshoot the end of the cutscene
			fFrameTime -= Timer-CutsceneEndTime;
			if(fFrameTime<0) fFrameTime = 0;
			Timer = CutsceneEndTime;
		}
	}

		
	//check for events that need to be triggered
	for(int i=0;i<NumberOfEvents;i++)
	{
		Events[i]->Update(this);
	}

	//update the actors
	
	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->Update(fFrameTime);
	}

    if( m_pxProfileData && m_pxProfileData->bProfiling )
    {
        u_int uTris = Asura_Render::GetNumberOfTrianglesRendered();

        m_pxProfileData->uTotalTris += uTris;

        if( uTris > m_pxProfileData->uMaxTris )
        {
            m_pxProfileData->uMaxTris = uTris;
        }

        const float fFPS = Asura_Timers::GetFrameRate();

        if( fFPS < m_pxProfileData->fMinFPS )
        {
            m_pxProfileData->fMinFPS = fFPS;
        }

        m_pxProfileData->fFPSSum += fFPS;

        ++m_pxProfileData->uNumSamples;
    }

	HadAnUpdate = true;
}

void Asura_Cutscene::AdvanceToTime(float time)
{
	Asura_Assert(Playing, "cannot advance to time in cutscene that's not playing.");
	if (!Playing) return;

	Asura_Assert(time>=0,"");
	if(time<Timer)
	{
		Finish();
		Initialise();
	}
	m_bAdvancingToTime = true;
	Update(time-Timer);

	if (m_bSkipMessagesAfterLoad)
	{
		// Flush out any finishing messages, before we clear the skip messages flag.
		Update(0.0f);
	}

	m_bAdvancingToTime = false;
}

void Asura_Cutscene::SkipToEndOfCutscene()
{
	if(Playing)
	{
		//If there is an end time , and this isn't a looping cutscene , skip to the end
		if(CutsceneEndTime && !(m_iCutsceneFlags & AsuraCutsceneFlag_LoopCutscene))
		{
			//Stick timer just beyond the end of the cutscene , to make sure rounding errors don't
			//cause any problems.
			AdvanceToTime(CutsceneEndTime+1.0f);
			
			if(Playing)
			{
				//At this point the cutscene will be at the end time , but not actually finished.
				//One last Update should do the trick
				Update();
			}
		}
		else
		{
			//just stop
			Finish();

			if(!ToolsPlayback)
			{
		        PerformEndOfCutsceneTasks();
			}
		}
	}
}

void Asura_Cutscene::Render()
{
	if(!Playing) return;

	//make sure update has been called at least once before rendering
	if(!HadAnUpdate) Update(0.0f);

	int i;
	for(i=0;i<NumberOfActors;i++)
	{
		Actors[i]->Render();
	}

	// Ali Jakamy
	// Render events
	for(i=0;i<NumberOfEvents;i++)
	{
		Events[i]->Render(this);
	}

}

void Asura_Cutscene::RenderShadows()
{
	if( !Playing ) return;

	if( !HadAnUpdate )
	{
		Update( 0.0f );
	}

	for( int i = 0; i < NumberOfActors; ++i )
	{
		Actors[i]->RenderShadow();
	}
}

void Asura_Cutscene::SetSubtitle(Asura_Hash_ID uTextHash,float fEndTime)
{
	m_uSubtitleHash      = uTextHash;
    m_fSubtitleStartTime = Timer;
	m_fSubtitleEndTime   = fEndTime;
}

void Asura_Cutscene::RenderSubtitles()
{
	//current subtitle?
	if(!m_uSubtitleHash) return;
	//has it expired yet?
	if(Timer > m_fSubtitleEndTime) return;
	
	#ifdef ASURA_TOOLS
	if(s_bDisableSubtitles) return;
	#endif

	const Asura_UnicodeChar* wszText = Asura_HashedLocalisedText::GetString(uASURA_HASHEDTEXTPAGE_CUTSCENE, m_uSubtitleHash);
	if(!wszText) return;

    // Fade based on proximity to either end of its duration.
    float fTimeFromEnd = Asura_Maths::Min( Timer - m_fSubtitleStartTime, m_fSubtitleEndTime - Timer );
    float fAlpha = Asura_Maths::Min( 1.0f, fTimeFromEnd / fASURA_DEFAULT_SUBTITLE_FADE_TIME );

	RenderSubtitle( wszText, fAlpha );
}


DEBUGVAR float dbg_fMinSubtitleRectHeight = 0.125f;
void Asura_Cutscene::GetSubtitleRect(Asura_Bounding_Rectangle& xRect)
{
	xRect.MinX = 0.06f;
	xRect.MaxX = 0.94f;
	// KOR - 07/03/2006 - allow subtitles to render when there are little to no widescreen bars,
	// otherwise align them with the top of the bottom bar.
	xRect.MinY = Asura_Maths::Min(1.f - dbg_fMinSubtitleRectHeight, 1.0f - Asura_Cutscene_System::GetWidescreenBarHeight());
	xRect.MaxY = 1.0f;
}


void Asura_Cutscene::RenderSubtitle( const Asura_UnicodeChar* wszText, float fAlpha ) const
{
    Asura_Bounding_Rectangle xRect;
	GetSubtitleRect(xRect);

	Asura_Font_System::SetFont( Asura_Cutscene_System::GetSubtitleFontIndex() );

    u_int uColour = 0x00FFFFFF;
    uColour |= static_cast<u_int>(Asura_Maths::FloatToInt(fAlpha * 255.0f)) << 24u;

	Asura_Font_System::Print( xRect, 0.0f, 0.0f, 0.75f, 0.9f, uColour,
							 ASURA_TEXTFORMAT_RELATIVEPOSITION|ASURA_TEXTFORMAT_CENTREJUSTIFIED|ASURA_TEXTFORMAT_VERTICALCENTREJUSTIFIED|ASURA_TEXTFORMAT_WORDWRAP,
							 wszText);
}

#ifdef ASURA_TOOLS

static const char* CUTSCENES_LOCALFILES_LOG = "logfiles\\local_files.txt";
void Asura_Cutscene::LoadFiles()
{
	int i;

	for(i=0;i<NumberOfActors;i++)
	{
		Actors[i]->LoadFiles();
#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
        extern bool evilg_bUsingLocalFileNotOnNetwork;
        if (evilg_bUsingLocalFileNotOnNetwork)
        {
            Asura_Logfile_Retail((CUTSCENES_LOCALFILES_LOG, "Cutscene %s used local file which was not found on the network.", 
                CutsceneName.GetString()));
            evilg_bUsingLocalFileNotOnNetwork = false;
        }
#endif
    }

	for(i=0;i<NumberOfEvents;i++)
	{
		Events[i]->LoadFiles();
#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
        extern bool evilg_bUsingLocalFileNotOnNetwork;
        if (evilg_bUsingLocalFileNotOnNetwork)
        {
            Asura_Logfile_Retail((CUTSCENES_LOCALFILES_LOG, "Cutscene %s used local file which was not found on the network.", 
                CutsceneName.GetString()));
            evilg_bUsingLocalFileNotOnNetwork = false;
        }
#endif
    }

    // (MH - 24.07.2002) added this bit to load files from the name lists.
    // This is used for extra anims added to cutscenes e.g. for recoil anims in GF2.
    // Also, any other anims loaded here might actually not be required. Use PruneNameLists()
    // to remove those.
    Asura_Assert(SkinFileNames.GetNumEntries() == SkinNames.GetNumEntries(), "");
    for (i = 0; i < SkinFileNames.GetNumEntries(); i++)
    {
        Asura_String & skinname = SkinNames.GetName(i);
        Asura_String & filename = SkinFileNames.GetName(i);
        if (filename.GetLength())
        {
            
            const Asura_Hierarchy_Skin* pxSkin = Asura_Animation_System::GetSkin(skinname.GetString());
            if(!pxSkin)
            {
                bool bOK = LoadAsuraFile(filename.GetString());
                Asura_Assertf(bOK, ("Cutscene %s referenced file %s which was not found.", 
                        CutsceneName.GetString(), filename.GetString()));

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
                extern bool evilg_bUsingLocalFileNotOnNetwork;
                if (evilg_bUsingLocalFileNotOnNetwork)
                {
                    Asura_Logfile_Retail((CUTSCENES_LOCALFILES_LOG, "Cutscene %s used local file %s which was not found on the network.", 
                        CutsceneName.GetString(), filename.GetString()));
                }
                else if (!bOK)
                {
                    Asura_Logfile_Retail((CUTSCENES_LOCALFILES_LOG, "Cutscene %s referenced file %s which was not found.", 
                        CutsceneName.GetString(), filename.GetString()));
                }
#endif
            	pxSkin = Asura_Animation_System::GetSkin(skinname.GetString());
            }
			if(pxSkin) Tools_Animation_Files::LoadSoundsForSkin(pxSkin);
        }
    }
    for (i = 0; i < AnimationFileNames.GetNumEntries(); i++)
    {
        const char * animname = AnimationNames.GetName(i).GetString();
        Asura_String & filename = AnimationFileNames.GetName(i);
        if (filename.GetLength())
        {
            const Asura_Hierarchy_Anim* pxAnim = Asura_Animation_System::GetAnim(animname);
            if (!pxAnim)
            {
                bool bOK = LoadAsuraFile(filename.GetString());
                Asura_Assertf(bOK, ("Cutscene %s referenced file %s which was not found.", 
                        CutsceneName.GetString(), filename.GetString()));

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
                extern bool evilg_bUsingLocalFileNotOnNetwork;
                if (evilg_bUsingLocalFileNotOnNetwork)
                {
                    Asura_Logfile_Retail((CUTSCENES_LOCALFILES_LOG, "Cutscene %s used local file %s which was not found on the network.", 
                        CutsceneName.GetString(), filename.GetString()));
                }
                else if (!bOK)
                {
                    Asura_Logfile_Retail((CUTSCENES_LOCALFILES_LOG, "Cutscene %s referenced file %s which was not found.", 
                        CutsceneName.GetString(), filename.GetString()));
                }
#endif
            	pxAnim = Asura_Animation_System::GetAnim(animname);
            }
			if(pxAnim) Tools_Animation_Files::LoadSoundsForAnim(pxAnim);
        }
    }
}

#ifdef ASURA_TOOLS
void Asura_Cutscene::ReloadAnimations()
{
    for (int i = 0; i < AnimationFileNames.GetNumEntries(); i++)
    {
        const char * pszAnimName = AnimationNames.GetName(i).GetString();
        Asura_String & xFilename = AnimationFileNames.GetName(i);
        if (xFilename.GetLength())
        {
			//if animation is already loaded , flag it to allow overwrite
            Asura_Hierarchy_Anim* pxAnim = Asura_Animation_System::GetAnim(pszAnimName);
			if(pxAnim) pxAnim->m_bAllowOverwriteOnLoad = true;

			//(re)load the animation file
			#if defined ASURA_ASSERT
				bool bOK =
			#endif
				LoadAsuraFile(xFilename.GetString());

			Asura_Assertf(bOK, ("Cutscene %s referenced file %s which was not found.", 
				CutsceneName.GetString(), xFilename.GetString()));

			//clear the overwrite flag
			if(pxAnim) pxAnim->m_bAllowOverwriteOnLoad = false;
            
        }
    }
}
#endif //ASURA_TOOLS

void Asura_Cutscene::PruneNameLists()
{
    // (MH - 01.11.2002) prune the anim and skin tables
    for (int i = 0; i < AnimationFileNames.GetNumEntries(); i++)
    {
        if (AnimationFileNames.GetName(i).IsBlank() || AnimationFileNames.GetName(i).IsNull())
        {
            continue;
        }

        bool bUsed = false;
        for(int iActor = 0; iActor < NumberOfActors; iActor++)
	    {
		    if (Actors[iActor]->UsesAnim(i))
            {
                bUsed = true;
                break;
            }
	    }
        for(int iEvent = 0; iEvent < NumberOfEvents; iEvent++)
	    {
		    if (Events[iEvent]->UsesAnim(i))
            {
                bUsed = true;
                break;
            }
	    }
        if (!bUsed)
        {
            AnimationFileNames.RemoveName(i);
            AnimationNames.RemoveName(i);
        }
    }
    for (int i = 0; i < SkinFileNames.GetNumEntries(); i++)
    {
        if (SkinFileNames.GetName(i).IsBlank() || SkinFileNames.GetName(i).IsNull())
        {
            continue;
        }
        bool bUsed = false;
        for(int iActor = 0; iActor < NumberOfActors; iActor++)
	    {
		    if (Actors[iActor]->UsesSkin(i))
            {
                bUsed = true;
                break;
            }
	    }
        if (!bUsed)
        {
            SkinFileNames.RemoveName(i);
            SkinNames.RemoveName(i);
        }
    }
}

#include "Entity/Asura_Entity.h"

unsigned short Asura_Cutscene::CopyMessageBlock(const Asura_Cutscene* pxOriginalCutscene,unsigned short usBlockNumber)
{
	Asura_Assert(pxOriginalCutscene,"");

	//first do a stright copy of the messages
	usBlockNumber = Asura_StaticMessage_System::CopyMessageBlock(usBlockNumber);
	if(usBlockNumber==ASURA_STATICMESSAGE_EMPTYBLOCK) return usBlockNumber;

	//Now need to check any messages that were targeted at actors in the original cutscene , and redirect them
	//to the equivalent actors in the copy of the cutscene.
	int iNumberOfMessages = Asura_StaticMessage_System::GetNumberOfMessagesInBlock(usBlockNumber);
	for(int i=0;i<iNumberOfMessages;i++)
	{
		Asura_StaticMessage* pxMessage = Asura_StaticMessage_System::GetMessagePointer(usBlockNumber,i);
		Asura_Assert(pxMessage,"");
		if(pxMessage->m_uSendTo == ASURA_GUID_UNREGISTERED) continue;

		for(int i=0;i<pxOriginalCutscene->NumberOfActors;i++)
		{
			if(pxOriginalCutscene->Actors[i]->m_uGuid != pxMessage->m_uSendTo) continue;

			for(int j=0;j<NumberOfActors;j++)
			{
				if(Actors[i]->ActorName != pxOriginalCutscene->Actors[i]->ActorName) continue;

				//Message is targeted at an actor , make sure the equivalent actor has been given a guid
				if(Actors[j]->m_uGuid == ASURA_GUID_UNREGISTERED)
				{
					Actors[j]->m_uGuid = AssignActorGuid();
				}

				pxMessage->m_uSendTo = Actors[j]->m_uGuid;
				break;
			}

		}
	}

	return usBlockNumber;
}

#endif // #ifdef ASURA_TOOLS

void Asura_Cutscene::SetProfiling( bool bProfiling )
{
    // Can only profile on client.
    if( bProfiling && !Asura_Network::IsClient() )
    {
        Asura_Assert( false, "Shouldn't be activating cutscene profiling on server-only machines." );
        bProfiling = false;
    }

    if( bProfiling )
    {
        if( !m_pxProfileData )
        {
            m_pxProfileData = new ProfileData;
        }

        if( m_pxProfileData )
        {
            m_pxProfileData->bProfiling  = true;
            m_pxProfileData->uTotalTris  = 0;
            m_pxProfileData->uMaxTris    = 0;
            m_pxProfileData->fMinFPS     = Asura_Maths::LargeFloat;
            m_pxProfileData->fFPSSum     = 0.0f;
            m_pxProfileData->uNumSamples = 0;
        }
    }
    else if( m_pxProfileData )
    {
        m_pxProfileData->bProfiling = false;
    }
}

void Asura_Cutscene::ResetEventsForActor(Asura_Cutscene_Actor* pxActor)
{
	for(int i=0;i<NumberOfEvents;i++)
	{
		if(Events[i]->m_pxTargetActor == pxActor)
		{
			Events[i]->Finish();
			Events[i]->Initialise();
			Events[i]->m_fNextTriggerTime+=Timer;
		}
	}
}

#ifdef ASURA_TOOLS
void Asura_Cutscene::ResetGuidsAndMessages()
{
	Asura_StaticMessage_System::DeleteMessageBlock(m_usEndCutsceneMessageBlock);
	m_usEndCutsceneMessageBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;

	for(int i=0;i<NumberOfActors;i++)
	{
		Actors[i]->ResetGuidsAndMessages();
	}

	for(int i=0;i<NumberOfEvents;i++)
	{
		Events[i]->ResetGuidsAndMessages();
	}
}
#endif

bool Asura_Cutscene::ControlsCamera( void ) const
{
    if(m_iCutsceneFlags & AsuraCutsceneFlag_DisableCamera) return false;
    return( !Asura_ClientCutscene_Actor_Camera::Disable && GetNumberOfCameras() > 0 );
}

bool Asura_Cutscene::WouldCutsceneBePlayingAtThisTime(float fTimer) const
{
	if(CutsceneEndTime==0) return true;
	if(m_iCutsceneFlags & AsuraCutsceneFlag_LoopCutscene) return true;
	return(fTimer<CutsceneEndTime);
}


void Asura_Cutscene::SendAllMessages()
{
	//check for any trigger events
	for(int i=0;i<NumberOfEvents;i++)
	{
		if(Events[i]->m_iType == ACET_Trigger)
		{
			if(!Events[i]->ShouldSkipWhenAdvancingTime())
			{
				Events[i]->Trigger(this);
			}
		}
	}


	//deal with end of cutscene messages
	PerformEndOfCutsceneTasks();
}

void Asura_Cutscene::LocaliseSoundName(const char* pszInput,char* pszOutputBuffer,u_int uOutputBufferSize) const
{
	//default implementation , just use the name without changing it
	Asura_Assert(uOutputBufferSize > strlen(pszInput),"");
	strcpy(pszOutputBuffer,pszInput);
}

#ifdef ASURA_TOOLS
void Asura_Cutscene::DelocaliseSoundName(const char* pszInput,char* pszOutputBuffer,u_int uOutputBufferSize) const
{
	//default implementation , just use the name without changing it
	Asura_Assert(uOutputBufferSize > strlen(pszInput),"");
	strcpy(pszOutputBuffer,pszInput);
}
#endif

//
// Asura_Cutscene::FindCutsceneActor()
//
// Parameters:		uiGuid - The guid of the actor you want
// Return value:	Pointer to the actor, or NULL if not found
//
// This function finds the cutscene actor with the given guid and returns a pointer to it
//

Asura_Cutscene_Actor* Asura_Cutscene::FindCutsceneActor(Asura_Guid uiGuid)
{
	for (int i = 0; i < NumberOfActors; ++i)
	{
		if (Actors[i]->m_uGuid == uiGuid)
		{
			return Actors[i];
		}
	}

	return NULL;
}

Asura_Cutscene_Actor* Asura_Cutscene::GetActorByName(const char* pszName)
{
	if(!pszName) return 0;

	for(int i=0;i<NumberOfActors;i++)
	{
		if(!strcmp(pszName,Actors[i]->ActorName.GetString()))
		{
			return Actors[i];
		}
	}
	return 0;
}

Asura_Cutscene_Actor* Asura_Cutscene::GetActorByHash(Asura_Hash_ID uHash)
{
	Asura_Cutscene_Actor* pxResult = NULL;

	if(uHash != ASURA_HASH_ID_UNSET)
	{
		for(int i=0;i<NumberOfActors;i++)
		{
			if(uHash == Actors[i]->ActorName.GetHash())
			{
				pxResult = Actors[i];
				break;
			}
		}
	}
	
	return pxResult;
}

bool Asura_Cutscene::SetActiveCamera(Asura_Cutscene_Actor_Camera* pxActor)
{
	if(!pxActor || !pxActor->IsCamera() || !IsPlaying())
	{
		Asura_Assert(pxActor,"Not expecting to remove camera control using this");
		Asura_Assert(pxActor->IsCamera(),"Not a camera actor");
		Asura_Assert(IsPlaying(),"Can't set camera while not playing");
		return false;
	}

	m_pxActiveCamera = pxActor;
	return true;
}

bool Asura_Cutscene::SetActiveCameraByName(const char* pszName)
{
	//search for a matching camera
	for(int i=0;i<NumberOfActors;i++)
	{
		if(Actors[i]->IsCamera())
		{
			if(!stricmp(pszName,Actors[i]->ActorName.GetString()))
			{
				return SetActiveCamera((Asura_Cutscene_Actor_Camera*)Actors[i]);
			}
		}
	}
	//failed to find camera
	return false;
}

bool Asura_Cutscene::SetActiveCameraByHash(Asura_Hash_ID uHash)
{
	//search for a matching camera
	for(int i=0;i<NumberOfActors;i++)
	{
		if(Actors[i]->IsCamera())
		{
			if(Asura_GetHashID(Actors[i]->ActorName.GetString()) == uHash)
			{
				return SetActiveCamera((Asura_Cutscene_Actor_Camera*)Actors[i]);
			}
		}
	}
	//failed to find camera
	return false;
}


bool Asura_Cutscene::SetActiveCameraByIndex(u_int uIndex)
{
	Asura_Cutscene_Actor_Camera* pxCamera = GetCameraByIndex(uIndex);
	if(!pxCamera) return false;

	return SetActiveCamera(pxCamera);

}

u_int Asura_Cutscene::GetNumberOfCameras() const
{
	if(m_iNumberOfCameras<0)
	{
		//need to calculate it
		m_iNumberOfCameras = 0;
		for(int i=0;i<NumberOfActors;i++)
		{
			if(Actors[i]->IsCamera())
			{
				m_iNumberOfCameras++;
			}
		}
	}
	return (u_int)m_iNumberOfCameras;
}

Asura_Cutscene_Actor_Camera* Asura_Cutscene::GetCameraByIndex(u_int uIndex)
{
	u_int uCameraIndex = 0;
	for(int i=0;i<NumberOfActors;i++)
	{
		if(Actors[i]->IsCamera())
		{
			if(uCameraIndex==uIndex)
			{
				//found our camera
				return static_cast<Asura_Cutscene_Actor_Camera*>(Actors[i]);
			}
			uCameraIndex++;
		}
	}
	return 0;
}


const Asura_Cutscene_Actor_Camera* Asura_Cutscene::GetCameraByIndex(u_int uIndex) const
{
	return (const_cast<Asura_Cutscene*>(this))->GetCameraByIndex(uIndex);
}

bool Asura_Cutscene::ShouldRenderWidescreenBars( void ) const
{
	return( ControlsCamera() && (m_iCutsceneFlags & AsuraCutsceneFlag_Widescreen) );
}

#if defined(ASURA_TOOLS)

Asura_Guid Asura_Cutscene::AssignActorGuid() const
{
	return Asura_Entity_System::GetCreatedByToolsGuid_CutsceneFile();
}

void Asura_Cutscene::RegisterActorGuid(Asura_Guid uGuid) const
{
	Asura_Entity_System::RegisterToolsGuid(uGuid);
}

bool Asura_Cutscene::IsValidActorGuid(Asura_Guid uGuid) const
{
	return uGuid>=ASURA_GUID_TOOLS_CUTSCENEFILE_START && uGuid<=ASURA_GUID_TOOLS_CUTSCENEFILE_END;
}

#endif	// defined(ASURA_TOOLS)






void Asura_Cutscene::SetCutsceneEndTimeFromRealTimeDuration(float fRealTimeDuration)
{
	float fTimeScale = 1.0f;
	float fRealTime = 0.0f;
	float fCutsceneTime = m_fCutsceneStartTime;

	//Work out the required cutscene time to get the desired real time duration
	for(int i=0;i<NumberOfEvents;i++)
	{
		Asura_Cutscene_Event* pxEvent = Events[i];

		if(pxEvent->m_iType == ACET_TimeScale)
		{
			//add up the real time to this event
			if(pxEvent->m_fTriggerTime > fCutsceneTime)
			{
				float fTimeDiff = pxEvent->m_fTriggerTime-fCutsceneTime;
				fRealTime += fTimeDiff/fTimeScale;
				fCutsceneTime = pxEvent->m_fTriggerTime;
			}

			if(fRealTime >= fRealTimeDuration)
			{
				//timescale happens after end of desired cutscene length
				break;
			}
			
			//take note of the new time scale
			Asura_Cutscene_Event_TimeScale* pxTimeScaleEvent = (Asura_Cutscene_Event_TimeScale*) pxEvent;
			fTimeScale = pxTimeScaleEvent->m_fTimeScale;
			Asura_Assert(fTimeScale > 0.0f,"");
		}
	}

	//Add/subtract from the times to get to the desired result. There should be no further timescale events to
	//get in the way , so just use the timescale that we have now.
	fCutsceneTime += (fRealTimeDuration-fRealTime) * fTimeScale;

	CutsceneEndTime = fCutsceneTime;
}

void Asura_Cutscene::UpdateVolume(bool bForceUpdate)
{
	if(ASURA_INVALID_HANDLE == SoundTrackHandle) return;

	//Get the volume by combining the project multiplier (probably obsolete) ,the individual cutscene volume , 
	//and the mixer volume. 
	//RAR 30/04/2007
	float fVolume = GetVolume() * m_fSoundTrackVolume;
	const Asura_Sound_Mixer_Channel* pxMixer = Asura_Sound_Mixer_System::FindChannel(uMIXER_HASHID_CUTSCENES);
	if (!pxMixer)
	{
		// if project doesn't have a mixer called "cutscenes", then try "cutscene"
		pxMixer = Asura_Sound_Mixer_System::FindChannel(0x25C2AF8A); // Asura_GetHashID("Cutscene");
	}
	fVolume *= Asura_Sound_Mixer_System::GetMixerVolume(pxMixer);

	if((fVolume != m_fCurrentVolume) || bForceUpdate)
	{
		m_fCurrentVolume = fVolume;

		#ifdef ASURA_USE_AUDIO_SYSTEM
			Asura_Audio_Voice* const pxVoice = Asura_Audio_System::GetVoice(SoundTrackHandle);
			if(pxVoice)
			{
				pxVoice->SetVolume(m_fCurrentVolume);
			}
		#else
		    Asura_Sound_System::StreamingSoundSetVolume(SoundTrackHandle, m_fCurrentVolume);
		#endif
	}
}

#ifdef ASURA_TOOLS
bool Asura_Cutscene::ShouldSave() const
{
	if(Tools_Level_Export::IsExport())
	{
		if(m_iCutsceneFlags & AsuraCutsceneFlag_NeverIncludeInExport) return false;
		return IsMemberOfCurrentLightSet();
	}
	else
	{
		return true;
	}
}

// If one or more actors have a referance to the skin without the dont "export flag", return true
bool Asura_Cutscene::ShouldAddSkinToExport( const Asura_String& xSkinName, int iSkinNameListIndex ) const
{
	for (int i = 0; i < NumberOfActors; i++)
	{
		if (Actors[i]->IsHierarchyActor())
		{
			Asura_Cutscene_Actor_Hierarchy* pxHierActor = static_cast<Asura_Cutscene_Actor_Hierarchy*>(Actors[i]);
			if (pxHierActor)
			{
				if (((pxHierActor->m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_DONT_EXPORT_RESOURCES) == 0) 
					&& pxHierActor->UsesSkin(iSkinNameListIndex))
				{
					return true;
				}
				// If the DONT_EXPORT flag is set, we need to check the attachments separately
				else if ((pxHierActor->m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_DONT_EXPORT_RESOURCES) != 0)
				{
					for (int iAttachment = 0; iAttachment < pxHierActor->m_iNumberOfAttachments; iAttachment++)
					{
						if (pxHierActor->m_ppxAttachments[iAttachment] && pxHierActor->m_ppxAttachments[iAttachment]->UsesSkin( iSkinNameListIndex ))
						{
							return true;
						}
					}
				}
			}
		}
		else
		{
			if (Actors[i]->UsesSkin(iSkinNameListIndex))
			{
				return true;
			}
		}
	}

	//Asura_OutputToDebugger_Retail(("Cutscene %s: Not exporting skin %s\n", CutsceneName.GetString(), xSkinName.GetString()));
	return false;
}

bool Asura_Cutscene::ShouldAddAnimToExport( const Asura_String& xAnimName, int iAnimNameListIndex ) const
{
	for (int i = 0; i < NumberOfActors; i++)
	{
		if (Actors[i]->IsHierarchyActor())
		{
			Asura_Cutscene_Actor_Hierarchy* pxHierActor = static_cast<Asura_Cutscene_Actor_Hierarchy*>(Actors[i]);
			if (pxHierActor)
			{
				if (((pxHierActor->m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_DONT_EXPORT_RESOURCES) == 0) 
					&& pxHierActor->UsesAnim(iAnimNameListIndex))
				{
					return true;
				}
				// If the DONT_EXPORT flag is set, we need to check the attachments separately
				else if ((pxHierActor->m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_DONT_EXPORT_RESOURCES) != 0)
				{
					for (int iAttachment = 0; iAttachment < pxHierActor->m_iNumberOfAttachments; iAttachment++)
					{
						if (pxHierActor->m_ppxAttachments[iAttachment] && pxHierActor->m_ppxAttachments[iAttachment]->UsesAnim( iAnimNameListIndex))
						{
							return true;
						}
					}
				}
			}
		}
		else if (Actors[i]->UsesAnim(iAnimNameListIndex))
		{
			return true;
		}
	}

	for (int i = 0; i < NumberOfEvents; i++)
	{
		if (Events[i] && Events[i]->UsesAnim(iAnimNameListIndex))
		{
			return true;
		}
	}

	//Asura_OutputToDebugger_Retail(("Cutscene %s: Not exporting anim %s\n", CutsceneName.GetString(), xSkinName.GetString()));
	return false;
}

bool Asura_Cutscene::IsMemberOfCurrentLightSet() const
{
	return Tools_LightSets::IsMemberOfCurrentLightSets(m_uLightSets);
}

void Asura_Cutscene::AddActor(Asura_Cutscene_Actor* pxActor)
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		//enlarge the array of actors
		NumberOfActors++;
		Asura_Cutscene_Actor** ppxNewActors;
		Asura_ResourceAllocateArray(ppxNewActors, NumberOfActors);
		for(int i=0;i<NumberOfActors-1;i++)
		{
			ppxNewActors[i] = Actors[i];
		}
		//add on the new pxActor
		ppxNewActors[NumberOfActors-1] = pxActor;

		//replace the old array
		Asura_ResourceDeallocateArray(Actors);
		Actors = ppxNewActors;
	}
	Asura_ResourcePopType();
}

#endif //ASURA_TOOLS


void Asura_Cutscene::UpdateClipPlanes(float fNearClip, float fFarClip)
{
	Asura_Camera::SetCameraIndex(0);

	if(fNearClip >= 0.0f || m_fInitialNearClip >= 0.0f)
	{
		//store initial value if needed
		if(m_fInitialNearClip < 0.0f)
		{
			m_fInitialNearClip = Asura_Camera::GetNearClip();
		}
		if(fNearClip < 0.0f)
		{
			//want to reset the original value
			fNearClip = m_fInitialNearClip;
		}

		if(fNearClip != Asura_Camera::GetNearClip())
		{
			Asura_Camera::SetNearClip(fNearClip);
		}
	}

	if(fFarClip >= 0.0f || m_fInitialFarClip >= 0.0f)
	{
		//store initial value if needed
		if(m_fInitialFarClip < 0.0f)
		{
			m_fInitialFarClip = Asura_Camera::GetFarClip();
		}
		if(fFarClip < 0.0f)
		{
			//want to reset the original value
			fFarClip = m_fInitialFarClip;
		}

		if(fFarClip != Asura_Camera::GetFarClip())
		{
			Asura_Camera::SetFarClip(fFarClip);
		}
	}
}

void Asura_Cutscene::SetMixerPreset(Asura_Hash_ID uMixerPreset)
{
	//don't do anything if this is the same as the previous preset
	if(uMixerPreset == m_uMixerPresetUsed) return;	

	//unqueue the current preset
	if(m_uMixerPresetUsed)
	{
		Asura_Sound_Mixer_System::UnqueuePreset(m_uMixerPresetUsed);
		m_uMixerPresetUsed = 0;
	}

	//and use the new one
	if(uMixerPreset && uMixerPreset != uMIXER_HASHID_DEFAULT)
	{
		Asura_Sound_Mixer_System::QueuePreset(uMixerPreset);
		m_uMixerPresetUsed = uMixerPreset;
	}
}