#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _ASURA_CUTSCENE_H_
#define _ASURA_CUTSCENE_H_

#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_NameList.h"
#include "Core/Asura_String.h"
#include "Network/Asura_Guid.h"

#ifdef ASURA_TOOLS
#include "Misc/list_tem.hpp"
#endif

#define ASURA_CUTSCENE_USE_GAME_TIME (-1)

const float fASURA_DEFAULT_SUBTITLE_FADE_TIME = 0.25f;

class Asura_Chunk_Stream;
class Asura_Cutscene_Actor;
class Asura_Cutscene_Actor_Camera;
class Asura_Cutscene_Event;
class Asura_ServerCutscene_Actor;
class Asura_ServerCutscene_Event;

//enumeration of cutscene types
enum Asura_Cutscene_Type
{
	ACT_StandardCutscene,

	ACT_FirstProjectSpecific = 1000,
};

#define AsuraCutsceneFlag_EndAtLastTrackPoint   	0x00000001  //CutsceneEndTime is automatically set to the time of the last track point
#define AsuraCutsceneFlag_LoopCutscene          	0x00000002
#define AsuraCutsceneFlag_Widescreen            	0x00000004  // draw black bars top and bottom.
#define AsuraCutsceneFlag_StartsActive				0x00000008  //Cutscene should start playing immediately
#define AsuraCutsceneFlag_IgnoreModuleLinking		0x00000010	//render cutscene characters even if they are outside of modules
#define AsuraCutsceneFlag_ForceIntoMainFile			0x00000020  //When exporting , the skins/animations for this should go in the main file.(Assumed to be the case anyway for cutscenes that don't control the camera)
#define AsuraCutsceneFlag_AlwaysIncludeInExport		0x00000040	//Include cutscene in level exports , even when don't export cutscenes is checkedd
#define AsuraCutsceneFlag_DontExportToPS2			0x00000080  
#define AsuraCutsceneFlag_DontExportToGC			0x00000100  
#define AsuraCutsceneFlag_Skippable					0x00000200  //This cutscene can be skipped. Implementation is down to project specific code
#define AsuraCutsceneFlag_NeverIncludeInExport		0x00000400	//Never include cutscene in level exports, used for template cutscenes (e.g. in lightgun games).
#define AsuraCutsceneFlag_DontPauseStreams			0x00000800	//Do not pause other streaming sounds when this cutscene plays
#define AsuraCutsceneFlag_TakeLengthFromSoundTrack	0x00001000	//Causes the cutscene length to be set to be the same length as its sound track
#define AsuraCutsceneFlag_DisableCamera				0x00002000  //Disables cutscene camera control.
#define AsuraCutsceneFlag_Skybox					0x00004000	//Render the actors in the skybox
#define AsuraCutsceneFlag_LoopToStartTime			0x00008000  //Causes a looping cutscene to loop back to the start time instead of back to 0		
#define AsuraCutsceneFlag_ReservedForDerivedClasses 0x7ff00000

class Asura_Cutscene
{
public:
	Asura_Cutscene(const Asura_String& name);
	virtual ~Asura_Cutscene();

#ifdef ASURA_TOOLS
	virtual void Copy(const Asura_Cutscene* pxCutscene);
	unsigned short CopyMessageBlock(const Asura_Cutscene* pxOriginalCutscene,unsigned short usBlockNumber);
	void ListShapesUsed(List<Asura_Hash_ID>& xShapeList);
	virtual void BuildUsedTextureTree(class Asura_NameTree* pxNameTree);
	virtual void AddExtraFilesToList( List<Asura_String>& xNameList ) const;//called during level export to add extra asura files to export list

	//These two are for moving the members of a cutscene as the cutscene object is moved
	virtual void CalculateRelativePosition(const Asura_Matrix_3x3& xCutsceneOrientInverse,const Asura_Vector_3 xCutscenePosition);
	virtual void UpdateRelativeToPosition(const Asura_Matrix_3x3& xCutsceneOrient,const Asura_Vector_3 xCutscenePosition);
#endif

	static Asura_Cutscene* AllocateCutscene(int iCutsceneType,const Asura_String& xName);
	static Asura_Cutscene* Project_AllocateCutscene(int iCutsceneType,const Asura_String& xName);
    static float Project_GetDefaultInitialFOV();


	//Tools use.. remove any links this cutscene may have elsewhere before it gets banished from the level
	virtual void PreDestructionCleanup();

	//starts the cutscene playing , intialising the timer and actors
	virtual void Initialise();
	//performs cleanup at the end of the cutscene
	virtual void Finish();

	//Advances the cutscene to the end , triggering remaining events
	//If the cutscene has no end time , then this will just stop it.
	virtual void SkipToEndOfCutscene();

	//Gets volume multilpier to use for soundtrack (for project volume override)
	//Just to be confusing there is now a per-cutscene volume setting as well (m_fSoundTrackVolume)
	//This will have to be multiplied in as well to get the final result. RAR 10/03/2005
	virtual float GetVolume() const {return 1.0f;}

	//Goes through and sends all the static messages that would normally be sent as a result
	//of playing the cutscene , without actuall bothering with the tedious business of actually
	//playing it.
	void SendAllMessages();


    // called when cutscene finishes normally, so end of cutscene messages can be sent
    virtual void PerformEndOfCutsceneTasks();
    
	//Advances the cutscene. If no frame time is passed ,the time used is the
	//time of the current game frame.
	virtual void Update(float frameTime = ASURA_CUTSCENE_USE_GAME_TIME);

	void AdvanceToTime(float time);
	bool IsAdvancingToTime(){return m_bAdvancingToTime;}

	virtual void Render();
	virtual void RenderShadows();
	void RenderSubtitles();

	//gets the default rectangle to use for subtitles
	static void GetSubtitleRect(Asura_Bounding_Rectangle& xRect);
	//called by RenderSubtitles to display the current string
	virtual void RenderSubtitle( const Asura_UnicodeChar* wszText, float fAlpha ) const;

    virtual bool ShouldRenderWidescreenBars( void ) const;

	bool IsPlaying() const {return Playing;}
    virtual bool IsAllowedToPlay( void ) const { return true; }
    bool HasEverPlayed( void ) const { return m_bHasEverPlayed; }

	bool HasSoundTrack() const {return UseSoundTrack;}
	virtual bool ShouldUseSoundTrack( void ) 							const { return HasSoundTrack(); }
	virtual bool ShouldPauseSoundTrackUntilFirstFrame( void ) 	const { return true; }
	virtual bool SoundTrackContainsMusic()								const { return( m_bSoundTrackContainsMusic ); }

	//returns true if the cutscene plays forever , or if the time passed is less than the cutscene end time
	bool WouldCutsceneBePlayingAtThisTime(float fTimer) const;

// (MP - 10.03.2003) We don't want to do this outside of tools. The level
// exports should contain the necessary data.
#ifdef ASURA_TOOLS
	//Attempts to load any skins and animations needed by this cutscene, that aren't already loaded
	void LoadFiles();

	//Forces all animations to be reloaded from their original files.
	void ReloadAnimations();
#endif

	const Asura_String& GetName() const {return CutsceneName;}

	virtual int GetCutsceneType() const {return ACT_StandardCutscene;}
    // (MH - 22.07.2002) derived classes should override this function.

	//called by subtitle event to change the subtitle
	void SetSubtitle(Asura_Hash_ID uTextHash,float fEndTime);

	//sets all events for the actor as untriggered  , and updates the next trigger time
	void ResetEventsForActor(Asura_Cutscene_Actor* pxActor);

	//resets bitfield of enabled event IDs
	void EnableAllEvents(){m_uEnabledEvents = 0xffffffff;}

    virtual void SetProfiling( bool bProfiling );

    struct ProfileData
    {
        bool  bProfiling;
        u_int uTotalTris;
        u_int uMaxTris;
        float fMinFPS;
        float fFPSSum;
        u_int uNumSamples;
    };   

    // Only valid when playing.
    bool IsProfiling( void ) const { return( m_pxProfileData && m_pxProfileData->bProfiling ); }

    // Only valid after finished and if it was profiling when playing.
    const ProfileData* GetProfileData( void ) const { return m_pxProfileData; }

	//load/save functions. Defined in Asura_Chunk_System
	virtual void LoadFromChunk(class Asura_Chunk_Cutscene* chunk,Asura_Handle fileHandle);
	virtual void SaveToChunk(char*& chunkBuffer);
	virtual int GetChunkSize();
//	virtual void ReadFromChunkStream (Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream (Asura_Chunk_Stream& xStream) const;

    bool ControlsCamera( void ) const;
	//Similar to ControlsCamera() except returns true if the cutscene is currently actively
	//controlling the camera.
	bool IsControllingCamera() const {return m_bIsControllingCamera;}

	bool IsSkippable() const {return (m_iCutsceneFlags & AsuraCutsceneFlag_Skippable)!=0;}

#ifdef ASURA_TOOLS
	//Resets any guids and static messages so as to make cutscene
	//safe for exporting to other levels.
	virtual void ResetGuidsAndMessages();

    void PruneNameLists();

    static void DisableSubtitles(bool bDisable){s_bDisableSubtitles = bDisable;}

	static void DisableSoundTrack(bool bDisable){s_bDisableSoundTrack = bDisable;}

    virtual bool ShouldSave( void ) const;

    virtual bool NeedsController( void ) const { return true; }

	virtual bool ShouldAddSkinToExport( const Asura_String& xSkinName, int iSkinNameListIndex ) const;
	virtual bool ShouldAddAnimToExport( const Asura_String& xAnimName, int iAnimNameListIndex ) const;

	//Determines whether skins/anims should be added to an export at all for this cutscene
	virtual bool ShouldAddResourcesToExport() const {return ShouldSave();}

	bool IsMemberOfCurrentLightSet() const;
	u_int GetLightSets() const {return m_uLightSets;}
#endif

	bool IsUsingSoundTrack(){return SoundTrackHandle != ASURA_INVALID_HANDLE;}

	//Converts the input wav filename to a filename that includes the language directory.
	//This is used both on SoundTrackName and on the individual wav files used in tools
	virtual void LocaliseSoundName(const char* pszInput,char* pszOutputBuffer,u_int uOutputBufferSize) const;

	#ifdef ASURA_TOOLS
	//opposite version of the above , strips out the localised part of the filename
	virtual void DelocaliseSoundName(const char* pszInput,char* pszOutputBuffer,u_int uOutputBufferSize) const;
	#endif

	// (AAM - 29.04.2004) Get the cutscene actor with the given guid
	Asura_Cutscene_Actor* FindCutsceneActor(Asura_Guid uiGuid);
	Asura_Cutscene_Actor* GetActorByName(const char* pszName);
	Asura_Cutscene_Actor* GetActorByHash(Asura_Hash_ID uHash);

	//get the camera actor that is currently being used
	const Asura_Cutscene_Actor_Camera* GetActiveCamera() const{ return m_pxActiveCamera;}

	//for accessing camera by index number (not same as index in actor list)
	u_int GetNumberOfCameras() const;
	Asura_Cutscene_Actor_Camera* GetCameraByIndex(u_int uIndex);
	const Asura_Cutscene_Actor_Camera* GetCameraByIndex(u_int uIndex) const;

	//various means of setting the active camera actor
	bool SetActiveCamera(Asura_Cutscene_Actor_Camera* pxActor);
	bool SetActiveCameraByName(const char* pszName);
	bool SetActiveCameraByHash(Asura_Hash_ID uHash);
	bool SetActiveCameraByIndex(u_int uIndex);//index from 0 to number of cameras-1 (not index into actor array)
    void DisableCameraControl( void );
	
	#ifdef ASURA_TOOLS
	void InvalidateNumberOfCameras() {m_iNumberOfCameras=-1;}
	#endif

	//Sets the end time to match the overall real time duration , taking time scale events into account.
	void SetCutsceneEndTimeFromRealTimeDuration(float fRealTimeDuration);
	
	//Various functions for overriding assignment of guids to actors. Needed so that killmoves
	//can use a different guid range.

#if defined(ASURA_TOOLS)
	//Assigns next available actor guid
	virtual Asura_Guid AssignActorGuid() const;
	//Registers guid on load
	virtual void RegisterActorGuid(Asura_Guid uGuid) const;
	//Checks to see if guid is in correct range
	virtual bool IsValidActorGuid(Asura_Guid uGuid) const;
#endif	// defined(ASURA_TOOLS)

	//updates the current soundtrack volume if needed
	void UpdateVolume(bool bForceUpdate);

	bool IsSkyboxCutscene() const {return (m_iCutsceneFlags & AsuraCutsceneFlag_Skybox) ? true : false;}
	#ifdef ASURA_TOOLS
	//Enlarges actor array , and adds the actor on the end
	void AddActor(Asura_Cutscene_Actor* pxActor);
	#endif

	//Get the time that the cutscene is currently at
	float GetTimer() const {return Timer;}

	//Updates the near/far clip planes. If the values are  negative then the original settings are used
	void UpdateClipPlanes(float fNearClip, float fFarClip);

	//Used by mixer preset events to change the current preset being used 
	void SetMixerPreset(Asura_Hash_ID uMixerPreset);

//protected :
	Asura_String CutsceneName;
	Asura_String CutsceneDirectory;//additional directory to look in for animations/skins (subdiretory of 'Cutscenes')

	int NumberOfActors;
	Asura_Cutscene_Actor** Actors;

	int NumberOfEvents;
	Asura_Cutscene_Event** Events;

	Asura_String SoundTrackName;
	Asura_Handle SoundTrackHandle;
	float m_fSoundTrackVolume;//multiplier for cutscene volume

	float m_fCutsceneStartTime;//If set to non-zero then the cutscene will skip to this time when started
	float CutsceneEndTime;//If 0 , then the cutscene will continue forever

	int m_iCutsceneFlags;//from the AsuraCutsceneFlag defines above

    bool PlaysOnServer;

	unsigned short m_usEndCutsceneMessageBlock;

	#ifdef ASURA_TOOLS
	//Fov transferred to camera actors now.. this is just here for helping
	//with transferring the data in old cutscenes
	Asura_Angle m_fInitialFOV_UsedForUpdatingOldCutscenes;
	#endif

    // (MH - 23.07.2002) to reduce the number of strings being passed around, and to avoid
    // sending them across the network, maintain a list of skin and animation names and 
    // filenames here and just pass an index.
    // After some discussion with RAR we decided that because of problems of reference 
    // counting and variable size lists, Asura_NameTree isn't quite suited to this purpose, 
    // so I made Asura_NameList
    Asura_NameList SkinNames;
    Asura_NameList SkinFileNames;
    Asura_NameList AnimationNames;
    Asura_NameList AnimationFileNames;

	u_int m_uLightSets; //light sets that this cutscene should be exported for

	//hash ID of mixer preset to activate while this cutscene is playing
	Asura_Hash_ID m_uMixerPreset;
	//mixer preset that was activated (if any)
	Asura_Hash_ID m_uMixerPresetUsed;


public :
	//cutscene playback data
	bool Playing;
    bool m_bHasEverPlayed;
	float Timer;
	bool UseSoundTrack;
	bool m_bSoundTrackContainsMusic;
	bool HadAnUpdate;
	float m_fAudioStartTime;

	Asura_Hash_ID m_uSubtitleHash;
    float m_fSubtitleStartTime;
	float m_fSubtitleEndTime;//Time that subtitle stops being shown (0 means until replaced)

    bool m_bWaitingForSoundTrack;   // flag that we're waiting for preloading stream data.
	bool m_bAdvancingToTime;
	bool m_bSkipMessagesAfterLoad;	// Skip static messages when advancing to time.

	//bitfield of events IDs that are enabled
	u_int m_uEnabledEvents;

	//current camera actor
	Asura_Cutscene_Actor_Camera* m_pxActiveCamera;

	#ifdef ASURA_TOOLS
	bool m_bRecordingWavFile;//special flag used when..... recording a wav file.
	#endif

	#if defined(ASURA_TOOLS) || defined(ASURA_USE_STREAMING)
	Asura_Guid m_uGuid;//guid of cutscene controller
	#endif

//	friend class Asura_Cutscene_Actor;
//	friend class Asura_Cutscene_Actor_Camera;
//	friend class Asura_Cutscene_Actor_Hierarchy;
//	friend class Asura_Chunk_Cutscene;
//	friend class Tools_Cutscene_Object;
//	friend class Tools_Mode_Cutscene;
//	friend class Asura_Cutscene_System;

	Asura_Vector_3 Position;//mainly needed for editing purposes

	static bool ToolsPlayback;
	
protected:

	virtual bool CutsceneSoundTrackIsExclusive() const {return true;}

    ProfileData* m_pxProfileData;

	#ifdef ASURA_TOOLS
	//for disabling subtitles when rendering AVIs
	static bool s_bDisableSubtitles;

	static bool s_bDisableSoundTrack;
	#endif

private :
	mutable int m_iNumberOfCameras;
	bool m_bIsControllingCamera;//true if cutscene is currently controlling the camera

	float m_fCurrentVolume;//current overall soundtrak volume

	//Near/Far clip before being altered by the cutscene.
	//Negative if the cutscene hasn't had to alter them
	float m_fInitialNearClip;
	float m_fInitialFarClip;



};


#endif // _ASURA_CUTSCENE_H_
