#include "Core/Asura.h"
#include "Rendering/Asura_Render.h"

#include "Cutscene/Asura_Cutscene_Actor.h"

#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#ifdef ASURA_USE_ANIMATION_IK_SOLVER
#include "Animation/IK/Asura_Animation_IK_Solver.h"
#endif
#include "Core/Asura_TextureManagement.h"
#include "Cutscene/Asura_Cutscene_Track.h"
#include "Cutscene/Asura_Cutscene_Event.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Cutscene/Asura_Cutscene_Attachment.h"
#include "Cutscene/Asura_Cutscene_System.h"
#ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities
#include "Entity/Cutscene/Asura_ServerEntity_ServerCutsceneActor.h"
#endif
#include "Entity/PhysicalObject/Asura_ClientEntity_PhysicalObject.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "FacialAnimation/Asura_FacialAnimation_AnimationHandler_Cutscene.h"
#include "FacialAnimation/Asura_FacialAnimation_System.h"
#endif
#ifdef ASURA_TOOLS
#include "Network/Asura_Message.h"
#endif

#include "Debugging/Asura_Assert.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"

#ifdef ASURA_TOOLS
#include "Cutscene/Tools_CutsceneModeAVI.h"
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "FacialAnimation/Tools_FacialAnimation_System.h"
#endif
#include "Objects/Tools_Object.h"
#include "Objects/Tools_Object_System.h"
#endif

#include "Physics/Asura_Physics_Object.h"


////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Asura_Cutscene_Actor **
**----------------------*/


Asura_Cutscene_Actor::Asura_Cutscene_Actor(Asura_Cutscene* pxCutscene, int iType)
{
	Cutscene = pxCutscene;
	Type = iType;
	NumberOfTracks = 0;
	Tracks = 0;
	m_uGuid = ASURA_GUID_UNREGISTERED;
	m_uGuidToReplace = ASURA_GUID_UNREGISTERED;
	m_bAttached = false;
	m_uActorFlags = 0;
	CurrentTrack = 0;
	m_fTrackTimer  = 0;
	m_bTryToHideEntityAgain = false;
	m_pxDynamicObject = 0;
	#ifdef ASURA_TOOLS
	m_uToolsSortOrder = 0xffffffff;
	#endif
}

Asura_Cutscene_Actor::~Asura_Cutscene_Actor()
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		int i;
		for(i=0;i<NumberOfTracks;i++)
		{
			delete Tracks[i];
		}
		Asura_ResourceDeallocateArray(Tracks);
	}
	Asura_ResourcePopType();

	if(m_pxDynamicObject)
	{
		Asura_Physics_Object::Destroy(m_pxDynamicObject);
		m_pxDynamicObject = NULL;
	}
}


#ifdef ASURA_TOOLS
void Asura_Cutscene_Actor::Copy(Asura_Cutscene_Actor* pxActor)
{
	Asura_Assert(Type == pxActor->Type,"");

	ActorName = pxActor->ActorName;
	Position = pxActor->Position;
	Orientation = pxActor->Orientation;
	m_uActorFlags = pxActor->m_uActorFlags;
	m_uGuidToReplace = pxActor->m_uGuidToReplace;

	//copy the tracks
	CopyTracks(pxActor);
}

void Asura_Cutscene_Actor::CopyTracks(Asura_Cutscene_Actor* pxActor)
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		NumberOfTracks = pxActor->NumberOfTracks;
		Asura_ResourceAllocateArray(Tracks, NumberOfTracks);
		for(int i=0;i<NumberOfTracks;i++)
		{
			Tracks[i] = new Asura_Cutscene_Track;
			Tracks[i]->Copy(pxActor->Tracks[i]);
		}
	}
	Asura_ResourcePopType();
}

void Asura_Cutscene_Actor::CalculateRelativePosition(const Asura_Matrix_3x3& xCutsceneOrientInverse,const Asura_Vector_3 xCutscenePosition)
{
	//No real need to deal with the Position/Orientation in the actor , since they
	//are just calculated from the track positions. RAR 18/01/2006
	for(int i=0;i<NumberOfTracks;i++)
	{
		Tracks[i]->CalculateRelativePosition(xCutsceneOrientInverse,xCutscenePosition);
	}
}

void Asura_Cutscene_Actor::UpdateRelativeToPosition(const Asura_Matrix_3x3& xCutsceneOrient,const Asura_Vector_3 xCutscenePosition)
{
	//No real need to deal with the Position/Orientation in the actor , since they
	//are just calculated from the track positions. RAR 18/01/2006
	for(int i=0;i<NumberOfTracks;i++)
	{
		Tracks[i]->UpdateRelativeToPosition(xCutsceneOrient,xCutscenePosition);
	}
}

#endif // ASURA_TOOLS

Asura_Cutscene_Actor* Asura_Cutscene_Actor::AllocateActor(int iActorType,bool bServer,Asura_Cutscene* pxCutscene)
{
	Asura_Cutscene_Actor* pxActor = Project_AllocateActor(iActorType,bServer,pxCutscene);
	if(!pxActor)
	{
		switch(iActorType)
		{
			case ACAT_Camera :
				if (bServer)
                {
#ifdef ASURA_SUPPORT_SERVERCUTSCENES
                    pxActor = new Asura_ServerCutscene_Actor_Camera(pxCutscene);
#else // not #if ASURA_SUPPORT_SERVERCUTSCENES
					Asura_Assert(false, "server cutscenes not supported");
#endif // #else not #if ASURA_SUPPORT_SERVERCUTSCENES
                }
                else
                {
                    pxActor =  new Asura_ClientCutscene_Actor_Camera(pxCutscene) ;
                }
				break;

			case ACAT_CameraTarget :
				pxActor =  new Asura_Cutscene_Actor_CameraTarget(pxCutscene) ;
				break;

			case ACAT_Hierarchy :
				if(bServer)
				{
#ifdef ASURA_SUPPORT_SERVERCUTSCENES
					pxActor = new Asura_ServerCutscene_Actor_Hierarchy(pxCutscene);
#else // not #if ASURA_SUPPORT_SERVERCUTSCENES
					Asura_Assert(false, "server cutscenes not supported");
#endif // #else not #if ASURA_SUPPORT_SERVERCUTSCENES
				}
				else
				{
					pxActor = new Asura_ClientCutscene_Actor_Hierarchy(pxCutscene);
				}
				break;

			case ACAT_Shape :
				if(bServer)
				{
#ifdef ASURA_SUPPORT_SERVERCUTSCENES
					pxActor = new Asura_ServerCutscene_Actor_Shape(pxCutscene);
#else // not #if ASURA_SUPPORT_SERVERCUTSCENES
					Asura_Assert(false, "server cutscenes not supported");
#endif // #else not #if ASURA_SUPPORT_SERVERCUTSCENES
				}
				else
				{
					pxActor = new Asura_ClientCutscene_Actor_Shape(pxCutscene);
				}
				break;

			default :
				Asura_Assert(0,"Unknown actor type");
				break;

		}
	}
	return pxActor;
}


void Asura_Cutscene_Actor::Initialise()
{
	CurrentTrack = 0;
	Orientation.Identity();
	Position.x = Position.y = Position.z = -Asura_Maths::LargeFloat;
	m_bAttached = false;

	//RAR 14/10/2005 May be an issue here of not wanting to reset this with external track
	//control when looping cutscenes.. but will leave alone for the moment
	m_fTrackTimer = 0;

	HideEntityToReplace(true);
}

void Asura_Cutscene_Actor::Finish()
{
	HideEntityToReplace(false);


	if(m_pxDynamicObject)
	{
		Asura_Physics_Object::Destroy(m_pxDynamicObject);
		m_pxDynamicObject = NULL;
	}

}

void Asura_Cutscene_Actor::Update(float fFrameTime)
{
	if(m_bTryToHideEntityAgain)
	{
		HideEntityToReplace(true);
	}

	if(!NumberOfTracks) return;

	if(!(m_uActorFlags & ASURA_ACTORFLAG_EXTERNALTRACKCONTROL))
	{
		SetTrackTime(m_fTrackTimer+fFrameTime);
	}

	Asura_Cutscene_Track* pxTrack = Tracks[CurrentTrack];
	Asura_Cutscene_Track* pxNextTrack = CurrentTrack+1<NumberOfTracks ? Tracks[CurrentTrack+1] : 0;
	

	//get the position
	if(!m_bAttached)
	{
		pxTrack->GetPositionFromTime(m_fTrackTimer,&Position,pxNextTrack);
	}


	if(IsCollisionsEnabled())
	{
		if(!m_pxDynamicObject)
		{
			m_pxDynamicObject = Asura_Physics_Object::Create(ASURA_PHYSICS_OBJECT_EXTERNALCONTROL);
			if(!SetCollisionObject())
			{
				// Can't set the collision object at the moment, so give up and try next frame
				Asura_Assert(false, "Could not set collision object for cutscene actor");
				Asura_Physics_Object::Destroy(m_pxDynamicObject);
				m_pxDynamicObject = 0;
			}
		}
		
		if(m_pxDynamicObject)
		{
			m_pxDynamicObject->TeleportTo(Position);
			m_pxDynamicObject->SetOrientation(Orientation);
		}
	}

}

void Asura_Cutscene_Actor::SetTrackTime(float fTime)
{
	if(fTime<m_fTrackTimer) CurrentTrack = 0;

	m_fTrackTimer = fTime;

	//determine the current track;
	while(CurrentTrack+1<NumberOfTracks && m_fTrackTimer>=Tracks[CurrentTrack+1]->PointArrivalTimes[0])
	{
		CurrentTrack++;
	}
}

void Asura_Cutscene_Actor::UpdateOrientation()
{
	Asura_Assert(NumberOfTracks,"");
	if(!NumberOfTracks) return;

	Asura_Cutscene_Track* pxTrack = Tracks[CurrentTrack];
	Asura_Cutscene_Track* pxNextTrack = CurrentTrack+1<NumberOfTracks ? Tracks[CurrentTrack+1] : 0;


	//Determine the orientation.
	#if 1
	pxTrack->GetOrientationFromTime(m_fTrackTimer,&Orientation,pxNextTrack);
	#else
	//Get the current direction of the line
	Asura_Vector_3 direction;
	pxTrack->GetDirectionFromTime(m_fTrackTimer,&direction,pxNextTrack);

	//want a horizontal facing
	direction.y = 0;

	if(direction.x || direction.z)
	{
		direction.Normalise();
		Asura_Maths::OrientationFromDirection(Orientation,direction);
	}
	//else
	//{
	//	Orientation = pxTrack->Orientation;
	//}
	#endif
	
}

float Asura_Cutscene_Actor::GetLastPointTime() const
{
	if(!NumberOfTracks) return 0;
	
	return Tracks[NumberOfTracks-1]->GetEndTime();
}

#ifdef ASURA_TOOLS
void Asura_Cutscene_Actor::ResetGuidsAndMessages()
{
	m_uGuid = ASURA_GUID_UNREGISTERED;
	m_uGuidToReplace = ASURA_GUID_UNREGISTERED;
}
#endif

bool Asura_Cutscene_Actor::LoadAnimFile(const Asura_String& xAnimName,Asura_String& xAnimFileName)
{
	if(!xAnimName.GetLength()) return false;

	if(!Asura_Animation_System::GetAnim(xAnimName.GetString()))
	{
		//attempt to load the animation
		if(!xAnimFileName.IsBlank())
		{
			return LoadAsuraFile(xAnimFileName.GetString());
		}
		return false;
	}

	return true;
}

#ifdef ASURA_TOOLS
bool Asura_Cutscene_Actor::HandlesThisMessage( int iMsgID ) const
{
    switch( iMsgID )
    {
        case ASURA_MSG_ACTIVATE:
        case ASURA_MSG_DEACTIVATE:
        case ASURA_MSG_DESTRUCT:
            return true;

        default:
            return false;
    }
}
#endif

void Asura_Cutscene_Actor::SetExternalTrackControl(bool bSet)
{
	if(bSet)
		m_uActorFlags |= ASURA_ACTORFLAG_EXTERNALTRACKCONTROL;
	else
		m_uActorFlags &=~ ASURA_ACTORFLAG_EXTERNALTRACKCONTROL;
}


void Asura_Cutscene_Actor::SetCollisions(bool bSet)
{
	Asura_Assert(!bSet || IsCollisionsSupported(), "Trying to enable collisions on an actor type that does not support them!");
	if(bSet)
		m_uActorFlags |= ASURA_ACTORFLAG_COLLISIONS;
	else
		m_uActorFlags &=~ ASURA_ACTORFLAG_COLLISIONS;
}


void Asura_Cutscene_Actor::HideEntityToReplace(bool bHide)
{
	m_bTryToHideEntityAgain = false;
	if(m_uGuidToReplace == ASURA_GUID_UNREGISTERED) return;

	#ifdef ASURA_TOOLS
	if(Cutscene->ToolsPlayback)
	{
		//hide/unhide the tools object associated with the entity
		Tools_Object* pxObject = Tools_Object_System::GetObjectFromGuid(m_uGuidToReplace);
		if(pxObject) pxObject->SetHiddenFlag(ObjectHideFlag_ReplacedByCutscene,bHide);
		return;
	}
	#endif

	// tell the server or client entity to hide/unhide as required
	if(Cutscene->PlaysOnServer)
	{
		Asura_ServerEntity_PhysicalObject* pxEntity = Asura_ServerEntity_PhysicalObject::FindEntity(m_uGuidToReplace);		
		if(pxEntity)
		{
			pxEntity->HideForCutscene(bHide);
		}
	}
	else
	{
		Asura_ClientEntity_PhysicalObject* pxEntity = Asura_ClientEntity_PhysicalObject::FindEntity(m_uGuidToReplace);
		if(pxEntity)
		{
			pxEntity->HideForCutscene(bHide);
		}
		else if(bHide)
		{
			//Failed to find the client entity , but it may well be created later , so wait and try again next update
			//RAR 02/06/2006
			m_bTryToHideEntityAgain = true;
		}
	}
	
}


////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_Cutscene_Actor_Hierarchy **
**--------------------------------*/

Asura_Cutscene_Actor_Hierarchy::Asura_Cutscene_Actor_Hierarchy(Asura_Cutscene* _Cutscene, Asura_Cutscene_Actor_Type eActorType)
:Asura_Cutscene_Actor(_Cutscene,eActorType)
{
	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		m_aiSkinID[i] = -1;
	}
    AnimID = -1;
	m_iAnimationFlags = 0;
	m_fAnimationDelay = 0;
	m_uHierarchyFlags = ASURA_HIERARCHYACTORFLAG_NOSHADOW;
	m_uTextureHash = ASURA_HASH_ID_UNSET;

	m_iNumberOfAttachments = 0;
	m_ppxAttachments = 0;

	m_uIKTemplateID = 0;
	
	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	m_uFaceID = 0;
	#endif


	m_bIgnoreAnimMovement = false;

	m_pxCollisionObject = 0;

}

Asura_Cutscene_Actor_Hierarchy::~Asura_Cutscene_Actor_Hierarchy()
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		for(int i=0;i<m_iNumberOfAttachments;i++)
		{
			delete m_ppxAttachments[i];
		}
		Asura_ResourceDeallocateArray(m_ppxAttachments);
	}
	Asura_ResourcePopType();

	
	if(m_pxCollisionObject) delete static_cast<Asura_Collision_Box*>(m_pxCollisionObject);

}

#ifdef ASURA_TOOLS
void Asura_Cutscene_Actor_Hierarchy::Copy(Asura_Cutscene_Actor* _pxActor)
{
	Asura_ResourcePushType(ASURA_RESOURCE_CUTSCENE);
	{
		Asura_Cutscene_Actor_Hierarchy* pxActor = (Asura_Cutscene_Actor_Hierarchy*) _pxActor;
		Asura_Cutscene_Actor::Copy(pxActor);	

		m_iAnimationFlags = pxActor->m_iAnimationFlags;
		m_fAnimationDelay = pxActor->m_fAnimationDelay;
		m_uHierarchyFlags = pxActor->m_uHierarchyFlags;

		for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
		{
			if(pxActor->m_aiSkinID[i]==-1)
			{
				m_aiSkinID[i] = -1;
			}
			else
			{
				Asura_String& xName = pxActor->Cutscene->SkinNames.GetName(pxActor->m_aiSkinID[i]);
				Asura_String& xFileName = pxActor->Cutscene->SkinFileNames.GetName(pxActor->m_aiSkinID[i]);

				m_aiSkinID[i] = Cutscene->SkinNames.AddName(xName);

				#if defined ASURA_ASSERT
					int iIndex =
				#endif
					Cutscene->SkinFileNames.AddName(xFileName);

				Asura_Assert(iIndex==m_aiSkinID[i],"");
			}
		}

		if(pxActor->AnimID==-1)
		{
			AnimID = -1;
		}
		else
		{
			Asura_String& xName = pxActor->Cutscene->AnimationNames.GetName(pxActor->AnimID);
			Asura_String& xFileName = pxActor->Cutscene->AnimationFileNames.GetName(pxActor->AnimID);

			AnimID = Cutscene->AnimationNames.AddName(xName);

			#if defined ASURA_ASSERT
				int iIndex =
			#endif
				Cutscene->AnimationFileNames.AddName(xFileName);

			Asura_Assert(iIndex==AnimID,"");
		}

		m_uTextureHash = pxActor->m_uTextureHash;

		m_iNumberOfAttachments = pxActor->m_iNumberOfAttachments;
		if(m_iNumberOfAttachments)
		{
			Asura_ResourceAllocateArray(m_ppxAttachments, m_iNumberOfAttachments);
			for(int i=0;i<m_iNumberOfAttachments;i++)
			{
				m_ppxAttachments[i] = Asura_Cutscene_Attachment::AllocateAttachment(this,pxActor->m_ppxAttachments[i]->GetType());
				m_ppxAttachments[i]->Copy(pxActor->m_ppxAttachments[i]);
			}
		}

		m_uIKTemplateID = pxActor->m_uIKTemplateID;

		#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
		m_uFaceID = pxActor->m_uFaceID;
		#endif

	}
	Asura_ResourcePopType();
}

void Asura_Cutscene_Actor_Hierarchy::ListShapesUsed(List<Asura_Hash_ID>& xShapeList)
{
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->ListShapesUsed(xShapeList);
	}
}

void Asura_Cutscene_Actor_Hierarchy::BuildUsedTextureTree(Asura_NameTree* pxNameTree)
 {
	if (!Tools_Level_Export::IsExport() || ((m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_DONT_EXPORT_RESOURCES) == 0))
	{
		//add the replacement texture (if any)
		if(m_uTextureHash != ASURA_HASH_ID_UNSET)
		{
			int iTextureID = Asura_Texture_Management::GetIndexFromHashID(m_uTextureHash);
			if(iTextureID)
			{
				const Asura_String* pxName = Asura_Texture_Management::GetTextureName(iTextureID);
				Asura_Assert(pxName,"");

				int iJunkIndex;
				pxNameTree->FindNodeIndexByName(pxName->GetString(),true,iJunkIndex);

				//Skin textures need to have the default skin texture flags applied to them at some point
				//so they will export correctly. Will sort this out here for the moment...
				if(Tools_Level_Export::IsExport())
				{
					int iFlags = Asura_Texture_Management::GetTextureFlags(iTextureID);
					Asura_Texture_Management::SetTextureFlags(iTextureID,iFlags|Tools_Level_Export::This->GetDefaultCharacterTextureFlags());
				}
			}
		}
	}
	
	//Always include the attachments textures
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->BuildUsedTextureTree(pxNameTree);
	}
}

void Asura_Cutscene_Actor_Hierarchy::AddExtraFilesToList( List<Asura_String>& xNameList ) const
{
	Asura_Cutscene_Actor::AddExtraFilesToList(xNameList);

	//add files for attachments
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->AddExtraFilesToList(xNameList);
	}

	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	if(m_uFaceID)
	{
		Tools_Level_Export::AddFaceToExportList(NULL, m_uFaceID);
	}
	#endif

}

#endif // ASURA_TOOLS


void Asura_Cutscene_Actor_Hierarchy::Initialise()
{
	Asura_Cutscene_Actor::Initialise();

	if(NumberOfTracks)
	{
		//initialise position and orientation to first track settings
		Asura_Cutscene_Track* pxTrack = Tracks[CurrentTrack];
		Asura_Cutscene_Track* pxNextTrack = CurrentTrack+1<NumberOfTracks ? Tracks[CurrentTrack+1] : 0;

		//get the position
		pxTrack->GetPositionFromTime(0,&Position,pxNextTrack);
		pxTrack->GetOrientationFromTime(0,&Orientation,pxNextTrack);
	}

	m_bIgnoreAnimMovement = false;

	if(m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_INITIALLY_HIDDEN)
	{
		m_uHierarchyFlags |= ASURA_HIERARCHYACTORFLAG_CURRENTLY_HIDDEN;
	}
	else
	{
		m_uHierarchyFlags &= ~ASURA_HIERARCHYACTORFLAG_CURRENTLY_HIDDEN;
	}

// (MH - 18.07.2002) server version will need to send message to do this
//	//May not have a skin name yet if the cutscene is being played in the middle of editing
//	if(!SkinName.GetLength()) return;
//
//	//initialise the animation
//	Animation = new Asura_Animation(SkinName);
//	LastAnimTimer = 0;
//	Animation->TimersUpdatedBeforeRender = true;
//
//	if(!InitialAnimationName.GetLength()) return;
//	Animation->StartAnim(InitialAnimationName.GetString(),true,false);

}

void Asura_Cutscene_Actor_Hierarchy::Finish()
{
	if(m_pxCollisionObject)
	{
		delete static_cast<Asura_Collision_Box*>(m_pxCollisionObject);
		m_pxCollisionObject = NULL;
	}


	Asura_Cutscene_Actor::Finish();

// (MH - 18.07.2002) 
//	delete Animation;
//	Animation = 0;
}

void Asura_Cutscene_Actor_Hierarchy::Update(float frameTime)
{
	if(!NumberOfTracks) return;
	Asura_Cutscene_Actor::Update(frameTime);

	if(!m_bAttached)
	{
		UpdateOrientation();
	}

	//Check for animations that have movement data associated with them.
	//At the moment only use this in cases where no movement is being gained from the track
	//(so all the movement is coming from the animation).
	//This should prevent problems with using run animations that have this data , which
	//would result movement being applied twice.
	//RAR 30/04/2004
	Asura_Animation* pxAnim = GetAnimation();
	if(pxAnim && !m_bIgnoreAnimMovement)
	{
		Asura_Vector_3 xOffset;
		UpdateAnimTimers();
		if(pxAnim->GetMovementOffset(&xOffset))
		{
			if(Tracks[CurrentTrack]->GetSpeedFromTime(m_fTrackTimer) == 0.0f)
			{
				//The animation has movement data ,and the track itself is stationary.
				//So use the offset.
				Position += xOffset*Orientation;
			}
		}
	}

	if(pxAnim)
	{
		//Invalidate frame stamps ,in case a previous update has been done on the same frame
		pxAnim->InvalidateBonesUpdateFrameStamp();
		for(int i=0;i<m_iNumberOfAttachments;i++)
		{
			m_ppxAttachments[i]->InvalidateBonesUpdateFrameStamp();
		}
	}

	if(pxAnim && m_iNumberOfAttachments)
	{
		//Force the bones to be calculated now , since they may be needed
		//by the attachments. RAR 23/09/2005
		UpdateAnimTimers();
		Asura_Vector_3 xRenderPos = Position;
		if(Cutscene->IsSkyboxCutscene())
		{
			xRenderPos += *Asura_Camera::GetPosition();
		}
		pxAnim->Render(xRenderPos,Orientation,ASURA_HIERARCHY_RENDER_UPDATEBONESALWAYS|ASURA_HIERARCHY_RENDER_DONTRENDER);
		
	}

	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->Update(frameTime);
	}
}

void Asura_Cutscene_Actor_Hierarchy::PreDestructionCleanup()
{
	Asura_Cutscene_Actor::PreDestructionCleanup();

	Asura_Assert(Cutscene,"");

	if(AnimID!=-1)
	{
		Cutscene->AnimationNames.RemoveName(AnimID);
		Cutscene->AnimationFileNames.RemoveName(AnimID);
		AnimID = -1;
	}

	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		if(m_aiSkinID[i]!=-1)
		{
			Cutscene->SkinNames.RemoveName(m_aiSkinID[i]);
			Cutscene->SkinFileNames.RemoveName(m_aiSkinID[i]);
			m_aiSkinID[i] = -1;
		}
	}

	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->PreDestructionCleanup();
	}

}

void Asura_Cutscene_Actor_Hierarchy::LoadFiles()
{
	#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM) && defined(ASURA_TOOLS)
	if(m_uFaceID)
	{
		Tools_FacialAnimation_System::FindOrLoadFace(m_uFaceID);
	}
	#endif

	for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
		if(m_aiSkinID[i]==-1) continue;
    	const Asura_String & SkinName = Cutscene->SkinNames.GetName(m_aiSkinID[i]);
		if(!SkinName.GetLength()) continue;

		if(!Asura_Animation_System::GetSkin(SkinName.GetString()))
		{
			//skin not loaded , so attempt to do so
    	    Asura_String & SkinFileName = Cutscene->SkinFileNames.GetName(m_aiSkinID[i]);
			if(!SkinFileName.IsBlank())
			{
				LoadAsuraFile(SkinFileName.GetString());
			}
			else
			{
				//don't know the filename , try checking in a couple of places
				char filename[200];

				//first try in the Skins directory
				Asura_PrintfToBfr( filename, "Characters\\%s\\%s.asr", 200, true, SkinName.GetString(), SkinName.GetString() );
				if(!LoadAsuraFile(filename) && Cutscene->CutsceneDirectory.GetLength())
				{
					//try in the directory for this cutscene
					Asura_PrintfToBfr( filename, "Cutscenes\\%s\\Props\\%s.asr", 200, true, Cutscene->CutsceneDirectory.GetString(), SkinName.GetString() );
					LoadAsuraFile(filename);
				}

				if(Asura_Animation_System::GetSkin(SkinName.GetString()))
				{
					//store the filename
					SkinFileName = filename;
				}
			}
		}
	}

	if(AnimID==-1) return;
    Asura_String & InitialAnimationName = Cutscene->AnimationNames.GetName(AnimID);
    Asura_String & AnimationFileName = Cutscene->AnimationFileNames.GetName(AnimID);
    LoadAnimFile(InitialAnimationName,AnimationFileName);

}

bool Asura_Cutscene_Actor_Hierarchy::LoadAnimFile(const Asura_String& animName,Asura_String& animFileName)
{
	if(!animName.GetLength()) return false;

	if(!Asura_Cutscene_Actor::LoadAnimFile(animName,animFileName))
	{
		const Asura_String & SkinName = Cutscene->SkinNames.GetName(m_aiSkinID[0]);
    	if(!SkinName.GetLength()) return false;

		char filename[200];

		//first try in the Anims directory
		Asura_PrintfToBfr( filename, "Characters\\%s\\Anims\\%s.asr", 200, true, SkinName.GetString(), animName.GetString() );
		if(!LoadAsuraFile(filename) && Cutscene->CutsceneDirectory.GetLength())
		{
			//try in the directory for this cutscene
			Asura_PrintfToBfr( filename, "Cutscenes\\%s\\%s.asr", 200, true, Cutscene->CutsceneDirectory.GetString(), animName.GetString() );
			LoadAsuraFile(filename);
		}

		if(Asura_Animation_System::GetAnim(animName.GetString()))
		{
			//store the filename
			animFileName = filename;
			return true;
		}
		return false;
	}
	return true;
}

const char* Asura_Cutscene_Actor_Hierarchy::GetSkinName( void ) const
{
	if(m_aiSkinID[0]==-1) return 0;
	const Asura_String& xSkinName = Cutscene->SkinNames.GetName(m_aiSkinID[0]);
	return xSkinName.IsBlank() ? 0 : xSkinName.GetString();
}

void Asura_Cutscene_Actor_Hierarchy::ActivateAttachment(u_int uID,bool bActivate)
{
	//search for the attachment in the array
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		if(m_ppxAttachments[i]->GetID() == uID)
		{
			//found it
			m_ppxAttachments[i]->Activate(bActivate);
			break;
		}
	}
}

Asura_Cutscene_Attachment* Asura_Cutscene_Actor_Hierarchy::GetAttachmentByID(u_int uID)
{
	//search for the attachment in the array
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		if(m_ppxAttachments[i]->GetID() == uID)
		{
			//found it
			return m_ppxAttachments[i];
		}
	}

	return NULL;
}

Asura_Cutscene_Attachment* Asura_Cutscene_Actor_Hierarchy::GetAttachmentByIndex(int iIndex)
{
	if( (iIndex >= 0) && (iIndex < m_iNumberOfAttachments) )
	{
		return m_ppxAttachments[iIndex];
	}

	return NULL;
}



#ifdef ASURA_TOOLS
// (MH - 01.11.2002) code to track down unused skins and animations
bool Asura_Cutscene_Actor_Hierarchy::UsesSkin(int iSkinIndex)// note this is an index into the NameList NOT a hashID
{
    for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
	{
    	if (m_aiSkinID[i] == iSkinIndex)
    	{
    	    return true;
    	}
	}
    for (int i = 0; i < m_iNumberOfAttachments; i++)
    {
        if (m_ppxAttachments[i]->UsesSkin(iSkinIndex))
        {
            return true;
        }
    }
    return false;
}

bool Asura_Cutscene_Actor_Hierarchy::UsesAnim(int iAnimIndex)// note this is an index into the NameList NOT a hashID
{
    if (AnimID == iAnimIndex)
    {
        return true;
    }
    for (int i = 0; i < m_iNumberOfAttachments; i++)
    {
        if (m_ppxAttachments[i]->UsesAnim(iAnimIndex))
        {
            return true;
        }
    }
    return false;
}

void Asura_Cutscene_Actor_Hierarchy::ResetGuidsAndMessages()
{
	Asura_Cutscene_Actor::ResetGuidsAndMessages();

	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->ResetGuidsAndMessages();
	}
}

bool Asura_Cutscene_Actor_Hierarchy::NeedsUpdateBefore(const Asura_Cutscene_Actor* pxActor) const
{
	if(Asura_Cutscene_Actor::NeedsUpdateBefore(pxActor)) return true;

	//check to see if any attachments control the actor
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		if(m_ppxAttachments[i]->NeedsUpdateBefore(pxActor)) return true;
	}
	return false;
}
#endif


bool Asura_Cutscene_Actor_Hierarchy::SetCollisionObject()
{
	Asura_Assert(m_pxDynamicObject, "Was expecting the dynamic object to already exist by this point");
	if(!m_pxDynamicObject) return false;

	if(GetAnimation() && GetAnimation()->GetSkin())
	{
		if(GetAnimation()->GetSkin()->GetCollisionObject())
		{
			m_pxDynamicObject->SetCollisionObject(GetAnimation()->GetSkin()->GetCollisionObject());
		}
		else
		{
#ifdef ASURA_TOOLS
			if(!GetAnimation()->GetSkin()->m_pxSectionBoundingBoxData)
			{
				GetAnimation()->GetSkin()->InitSectionBoundingBoxes();
				GetAnimation()->GetSkin()->SetupBB();
			}
#endif
			m_pxCollisionObject = new Asura_Collision_Box(GetAnimation()->GetSkin()->GetCollisionBB());
			m_pxDynamicObject->SetCollisionObject(m_pxCollisionObject);
		}
		return true;
	}
	return false;
}

void Asura_Cutscene_Actor_Hierarchy::SetHidden(bool bHidden)
{
	if(bHidden)
	{
		m_uHierarchyFlags |= ASURA_HIERARCHYACTORFLAG_CURRENTLY_HIDDEN;
	}
	else
	{
		m_uHierarchyFlags &=~ASURA_HIERARCHYACTORFLAG_CURRENTLY_HIDDEN;
	}
}


////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_ClientCutscene_Actor_Hierarchy **
**--------------------------------------*/
bool Asura_ClientCutscene_Actor_Hierarchy::DontRender = false;

Asura_ClientCutscene_Actor_Hierarchy::Asura_ClientCutscene_Actor_Hierarchy(Asura_Cutscene* _Cutscene, Asura_Cutscene_Actor_Type eActorType)
:Asura_Cutscene_Actor_Hierarchy(_Cutscene, eActorType)
{
	Animation = 0;
	#ifdef ASURA_USE_ANIMATION_IK_SOLVER
	m_pxIKSolver = 0;
	#endif
	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	m_pxFaceAnimHandler = 0;
	#endif

}

void Asura_ClientCutscene_Actor_Hierarchy::Initialise()
{
    Asura_Cutscene_Actor_Hierarchy::Initialise();

    //May not have a skin name yet if the cutscene is being played in the middle of editing
    if(m_aiSkinID[0] == -1) return;
	const Asura_String& SkinName = Cutscene->SkinNames.GetName( m_aiSkinID[0] );
    if(!SkinName.GetLength()) return;

    //initialise the animation
    if( !Animation )
    {
	    Animation = new Asura_Animation(SkinName.GetString());
        Asura_Assert(Animation, "");

		//set up any additional skins
		for(int i=1; i<ASURA_ANIMATION_SKIN_COUNT; i++)
		{
			if(m_aiSkinID[i] == -1) continue;
			const Asura_String& xOtherSkinName = Cutscene->SkinNames.GetName( m_aiSkinID[i] );
			if(xOtherSkinName.IsBlank()) continue;

			Animation->SetSkin(xOtherSkinName.GetString(), i);
		}
    }
	//set replacement texture
	if(m_uTextureHash)
	{
		int iTextureID = Asura_Texture_Management::GetIndexFromHashID(m_uTextureHash);
		Animation->SetAnimTexID(iTextureID);
	}

	#ifdef ASURA_USE_ANIMATION_IK_SOLVER
	if(m_uIKTemplateID)
	{
		m_pxIKSolver = Asura_Animation_IK_Solver::CreateFromTemplate(m_uIKTemplateID, Animation);
		if(m_pxIKSolver) Animation->SetIKSolver(m_pxIKSolver);
	}
	#endif

	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	if(m_uFaceID)
	{
		m_pxFaceAnimHandler = new Asura_FacialAnimation_AnimationHandler_Cutscene(this, m_uFaceID);
		Asura_FacialAnimation_System::AddAnimationHandler(m_uGuid, m_pxFaceAnimHandler);
	}
	#endif
	


	LastAnimTimer = 0;
	Animation->TimersUpdatedBeforeRender = true;

	const char* szAnimName = 0;

	if(AnimID!=-1)
	{
		szAnimName = Cutscene->AnimationNames.GetName( AnimID ).GetString();
	}

    Animation->m_xPrimaryWrapper.StartAnim2(Asura_GetHashID(szAnimName),m_iAnimationFlags&~AnimationChangeFlag_EventSpecificFlags,0,m_fAnimationDelay);

	SetIgnoreAnimMovement((m_iAnimationFlags & AnimationChangeFlag_IgnoreMovement) ? true : false);

}

void Asura_ClientCutscene_Actor_Hierarchy::InitialiseAttachments()
{
	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->Initialise();
	}
}

void Asura_ClientCutscene_Actor_Hierarchy::Finish()
{
    Asura_Cutscene_Actor_Hierarchy::Finish();

	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->Finish();
	}
	
	#ifdef ASURA_USE_FACIAL_ANIMATION_SYSTEM
	if(m_pxFaceAnimHandler)
	{
		Asura_FacialAnimation_System::RemoveAnimationHandler(m_uGuid);
		delete m_pxFaceAnimHandler;
		m_pxFaceAnimHandler = 0;
	}
	#endif

    delete Animation;
    Animation = 0;

	#ifdef ASURA_USE_ANIMATION_IK_SOLVER
	delete m_pxIKSolver;
	m_pxIKSolver = 0;
	#endif

}

void Asura_ClientCutscene_Actor_Hierarchy::UpdateAnimTimers()
{
	if(!NumberOfTracks) return;
	if(DontRender) return;
	if(!Animation) return;

	Asura_Assert(Cutscene->Timer>=LastAnimTimer,"");
	if(Cutscene->Timer == LastAnimTimer) return;
	
	//update the anim duration based on the speed at this point on the track
	float fSpeed = Tracks[CurrentTrack]->GetSpeedFromTime(m_fTrackTimer);

	if(fSpeed > 0.0f)
	{
		Asura_Hash_ID uAnimID = Animation->m_xPrimaryWrapper.GetAnimHashID(ASURA_PRIMARY_ANIMATION_HANDLE);
		if(uAnimID != ASURA_HASH_ID_UNSET)
		{
			Animation->m_xPrimaryWrapper.SetAnimDurationBySpeed(fSpeed);
		}
	}
	
	Animation->UpdateTimers(Cutscene->Timer-LastAnimTimer, true);

	LastAnimTimer = Cutscene->Timer;
}

void Asura_ClientCutscene_Actor_Hierarchy::SetAlpha( float fAlpha )
{ 
	if(Animation) 
	{ 
		Animation->SetAlpha(fAlpha,true); 
	}
}

float Asura_ClientCutscene_Actor_Hierarchy::GetAlpha( void ) const
{
	return Animation ? Animation->GetAlpha() : 1.0f;
}


void Asura_ClientCutscene_Actor_Hierarchy::Render()
{
	//if this actor is attached to another , then it will be rendered when the attachment is rendered
	if(m_bAttached) return;

	if(!NumberOfTracks) return;
	if(DontRender) return;
	if(!Animation) return;
	Asura_Assert(Cutscene->Timer>=LastAnimTimer,"");

	//update the anim duration based on the speed at this point on the track
	if(Cutscene->Timer != LastAnimTimer)
	{
		float fSpeed = Tracks[CurrentTrack]->GetSpeedFromTime(m_fTrackTimer);

		if(fSpeed > 0.0f)
		{
			Animation->m_xPrimaryWrapper.SetAnimDurationBySpeed(fSpeed);
		}
	}

	unsigned uRenderFlags = ASURA_HIERARCHY_RENDER_UPDATEBONESALWAYS;
	#ifdef ASURA_TOOLS
	if(Cutscene->m_bRecordingWavFile)
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_DONTRENDER;
	}
	#endif

	#ifdef ASURA_TOOLS
	//RAR 17/05/2006 Some actors may be hidden while recording AVIs
	if(Tools_CutsceneModeAVI::ShouldHideActor(this))
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_DONTRENDER;
	}
	#endif


	if(Cutscene->m_iCutsceneFlags & AsuraCutsceneFlag_IgnoreModuleLinking || Cutscene->IsSkyboxCutscene())
	{
		//force cutscene characters to be drawn even when outside of modules
		uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_MODULELINK_TEST;
	}
	if(m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_NOSHADOW)
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_SHADOW;
	}
	if(m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_CURRENTLY_HIDDEN)
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_DONTRENDER;
	}
	if(m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_AMBIENTONLY)
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_FORCE_AMBIENT_COLOUR_ONLY;
	}
	Asura_Vector_3 xRenderPos = Position;
	if(Cutscene->IsSkyboxCutscene())
	{
		xRenderPos += *Asura_Camera::GetPosition();
	}

	//Calling UpdateTimers etc directly to get around the fact that the framestamps will
	//be set inconveniently from possbile earlier attempts at changing the timer. RAR 14/04/2005
	Animation->UpdateTimers(Cutscene->Timer-LastAnimTimer);
	Animation->Render(xRenderPos,Orientation,uRenderFlags);
	Animation->PlaySoundsForLastFrame();

	LastAnimTimer = Cutscene->Timer;

	for(int i=0;i<m_iNumberOfAttachments;i++)
	{
		m_ppxAttachments[i]->Render();
	}

}

void Asura_ClientCutscene_Actor_Hierarchy::RenderShadow()
{
	if( m_bAttached ) return;

	if( DontRender
	 || !Animation
	 || (m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_NOSHADOW)
	 || (m_uHierarchyFlags & ASURA_HIERARCHYACTORFLAG_CURRENTLY_HIDDEN) )
	{
		return;
	}

	#ifdef ASURA_TOOLS
		if( Cutscene->m_bRecordingWavFile || Tools_CutsceneModeAVI::ShouldHideActor( this ) ) return;
	#endif

	Animation->RenderShadow( Position, Orientation, 0 );

	for( int i = 0; i < m_iNumberOfAttachments; ++i )
	{
		m_ppxAttachments[i]->RenderShadow();
	}
}

//void Asura_ClientCutscene_Actor_Hierarchy::ChangeAnimation(Asura_String& animationName,float time,int flags,float tweenDuration)
void Asura_ClientCutscene_Actor_Hierarchy::ChangeAnimation(int animationID,float time,int flags,float tweenDuration, float fAnimDelay)
{
	Asura_Assert(Cutscene->Timer>=time,"");
	if(!Animation) return;
	
    Asura_String animationName = Cutscene->AnimationNames.GetName(animationID);
    if(!animationName.GetLength()) return;

	Asura_Hash_ID uAnimHash = animationName.GetHash();
	Animation->m_xPrimaryWrapper.StartAnim2(uAnimHash,flags,tweenDuration,time+fAnimDelay-LastAnimTimer);

	#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	if(m_pxFaceAnimHandler)
	{
		//see if there is a face animation to go with this
		const Asura_Hierarchy_Anim* pxAnim = Asura_Animation_System::GetAnimByHash(uAnimHash);
		if(pxAnim && pxAnim->HasFacialAnimation())
		{
			PlayFacialAnimation(uAnimHash, time+fAnimDelay, false);
		}
	}
	#endif

}

void Asura_ClientCutscene_Actor_Hierarchy::ChangeAnimation(Asura_Hash_ID xAnimHashID, float fTime, int iFlags, float fTweenDuration)
{
	Asura_Assert(Cutscene->Timer>=fTime,"");
	Animation->m_xPrimaryWrapper.StartAnim2(xAnimHashID,iFlags,fTweenDuration,fTime-LastAnimTimer);

	#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	if(m_pxFaceAnimHandler)
	{
		//see if there is a face animation to go with this
		const Asura_Hierarchy_Anim* pxAnim = Asura_Animation_System::GetAnimByHash(xAnimHashID);
		if(pxAnim && pxAnim->HasFacialAnimation())
		{
			PlayFacialAnimation(xAnimHashID, fTime, false);
		}
	}
	#endif

}

void Asura_ClientCutscene_Actor_Hierarchy::PlayAnimationToPos(int iAnimationID,float fTime,float fStartTimer,float fEndTimer,int iFlags,float fTweenDuration, float fAnimDelay)
{
	Asura_Assert(Cutscene->Timer>=fTime,"");
	if(!Animation) return;
	
    Asura_String xAnimationName = Cutscene->AnimationNames.GetName(iAnimationID);
    if(!xAnimationName.GetLength()) return;

	Animation->m_xPrimaryWrapper.PlayAnimToPos(xAnimationName.GetHash(),fStartTimer,fEndTimer,iFlags,fTweenDuration);
	Animation->m_xPrimaryWrapper.SetTimeDelay(fAnimDelay+fTime-LastAnimTimer);
}

int Asura_ClientCutscene_Actor_Hierarchy::GetAnimationFlagsForLastFrame() const
{
	if(!Animation) return 0;
	return Animation->m_xPrimaryWrapper.GetFlagsForAllAnimsLastFrame();
}

//Asura_Handle Asura_ClientCutscene_Actor_Hierarchy::PlaySecondaryAnimation(Asura_String& animationName,float time,int flags,float weight,float duration)
Asura_Handle Asura_ClientCutscene_Actor_Hierarchy::PlaySecondaryAnimation(int animationID,float time,int flags,float weight,float fTweenDuration,bool bAutoRemove, float fAnimDelay)
{
	if(!Animation) return ASURA_INVALID_HANDLE;
    Asura_String animationName = Cutscene->AnimationNames.GetName(animationID);
	if(!animationName.GetLength()) return ASURA_INVALID_HANDLE;
	if(bAutoRemove)
	{
		flags |=AsuraAnimationFlag_AutoRemove;
	}

	Asura_Handle iHandle = Animation->m_xPrimaryWrapper.AddOneShotAnim2(animationName.GetHash(),weight,flags,fTweenDuration,time+fAnimDelay-LastAnimTimer);
    return iHandle;
}

Asura_Handle Asura_ClientCutscene_Actor_Hierarchy::PlaySecondaryAnimationToPos(int iAnimationID,float fTime,float fStartTimer,float fEndTimer,int iFlags,float fWeight,float fTweenDuration,bool bAutoRemove, float fAnimDelay)
{
	Asura_Assert(Cutscene->Timer>=fTime,"");
	if(!Animation) return ASURA_INVALID_HANDLE;
	
    Asura_String xAnimationName = Cutscene->AnimationNames.GetName(iAnimationID);
    if(!xAnimationName.GetLength()) return ASURA_INVALID_HANDLE;

	if(bAutoRemove)
	{
		iFlags |=AsuraAnimationFlag_AutoRemove;
	}

	Asura_Handle iHandle = Animation->m_xPrimaryWrapper.PlayOneShotAnimToPos(xAnimationName.GetHash(),fWeight,fStartTimer,fEndTimer,iFlags);\
	if(iHandle != ASURA_INVALID_HANDLE)
	{
		Animation->m_xPrimaryWrapper.SetTimeDelay(fTime+fAnimDelay-LastAnimTimer,iHandle);
		Animation->m_xPrimaryWrapper.SetTweenDuration(fTweenDuration,iHandle);
	}

	return iHandle;
}


void Asura_ClientCutscene_Actor_Hierarchy::RemoveSecondaryAnimation(Asura_Handle handle)
{
	Asura_Assert(Animation,"");
	Animation->m_xPrimaryWrapper.RemoveOneShotAnim(handle);
}

void Asura_ClientCutscene_Actor_Hierarchy::ChangeTextureAnimation(int iSequence,float fSpeed,u_int uFlags)
{
	if(!Animation) return;

	//set speed/sequence according to flags
	if(uFlags & ASURA_CUTSCENEEVENT_TEXTUREANIMATION_CHANGESEQUENCE)
	{
		Animation->SetTextureAnimSequence(iSequence);
	}
	if(uFlags & ASURA_CUTSCENEEVENT_TEXTUREANIMATION_CHANGESPEED)
	{
		Animation->SetTextureAnimSpeed(fSpeed);
	}
}

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
void Asura_ClientCutscene_Actor_Hierarchy::PlayFacialAnimation(Asura_Hash_ID uAnimID, float fStartTime, bool bLooped)
{
	if(!m_pxFaceAnimHandler)
	{
		Asura_Assertf(0,("Attempting to play facial animation on cutscene actor '%s'.\nThis actor doesn't have a face setup",ActorName.GetString()));
		return;
	}

	m_pxFaceAnimHandler->SetActive(true);
	m_pxFaceAnimHandler->Cutscene_PlayFacialAnimation(uAnimID, fStartTime, bLooped);
	
}
#endif //defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)



void Asura_ClientCutscene_Actor_Hierarchy::GetSoundPosition(Asura_Vector_3& xPosition) const
{
	xPosition = Position;

	//get the zeroth bone position
	Asura_Vector_3 xOffset;
	if(Animation && Animation->GetOffsetFromCompassPoint(xOffset))
	{
		xPosition += xOffset * Orientation;
	}
}


#ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities
////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Asura_ServerCutscene_Actor_Hierarchy **
**--------------------------------------*/

Asura_ServerCutscene_Actor_Hierarchy::Asura_ServerCutscene_Actor_Hierarchy(Asura_Cutscene* _Cutscene)
:Asura_Cutscene_Actor_Hierarchy(_Cutscene)
{
    Entity = NULL;
#ifdef ASURA_TOOLS
    ToolsCutsceneActor = NULL;
#endif
}

void Asura_ServerCutscene_Actor_Hierarchy::Initialise()
{
    Asura_Cutscene_Actor_Hierarchy::Initialise();

    //May not have a skin name yet if the cutscene is being played in the middle of editing
	//if(!SkinName.GetLength()) return;
	//if(!InitialAnimationName.GetLength()) return;
    //Asura_Assert(SkinID != -1, "");
    //Asura_Assert(AnimID != -1, "");

    if (!Asura_Cutscene::ToolsPlayback)
    {
        //initialise the server entity, if derived class hasn't done it yet.
	    if (!Entity)
        {
			//Just grab the first skin for now. RAR 30/09/2008
			int iSkinID = m_aiSkinID[0];
            Entity = new Asura_ServerEntity_ServerCutsceneActorHierarchy(this, iSkinID, AnimID, m_iAnimationFlags&~AnimationChangeFlag_EventSpecificFlags, m_fAnimationDelay ,m_uGuid);
        }
        // make sure the client version is created now in case messages occur out of order.
        // standard update does create the entity, animation updates currently do not.
        Entity->MakeClientUpdateMessage();

    }
#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback)
    {
        ToolsCutsceneActor = new Asura_ClientCutscene_Actor_Hierarchy(Cutscene);
        // copy Asura_Cutscene_Actor
        ToolsCutsceneActor->NumberOfTracks = NumberOfTracks;
	    ToolsCutsceneActor->Tracks = Tracks;
        // copy Asura_Cutscene_ActorHierarchy
        ToolsCutsceneActor->AnimID = AnimID;
		for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
		{
       		ToolsCutsceneActor->m_aiSkinID[i] = m_aiSkinID[i];
		}
		ToolsCutsceneActor->m_iAnimationFlags = m_iAnimationFlags;
		ToolsCutsceneActor->m_fAnimationDelay = m_fAnimationDelay;
		ToolsCutsceneActor->m_iNumberOfAttachments = m_iNumberOfAttachments;
		ToolsCutsceneActor->m_ppxAttachments = m_ppxAttachments;
		ToolsCutsceneActor->m_uTextureHash = m_uTextureHash;
		ToolsCutsceneActor->m_uHierarchyFlags = m_uHierarchyFlags;
		ToolsCutsceneActor->m_uGuidToReplace = m_uGuidToReplace;
        ToolsCutsceneActor->Initialise();
    }
#endif
}

void Asura_ServerCutscene_Actor_Hierarchy::InitialiseAttachments()
{
    if (!Asura_Cutscene::ToolsPlayback)
    {
		for(int i=0;i<m_iNumberOfAttachments;i++)
		{
			m_ppxAttachments[i]->Initialise();
		}
    }
#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->InitialiseAttachments();
    }
#endif
}


#ifdef ASURA_TOOLS
void Asura_ServerCutscene_Actor_Hierarchy::Render()
{
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->Position = Position;
        ToolsCutsceneActor->Orientation = Orientation;
        // need to set last timer?
        ToolsCutsceneActor->Render();
    }
}
#endif

void Asura_ServerCutscene_Actor_Hierarchy::Finish()
{
    Asura_Cutscene_Actor_Hierarchy::Finish();

    if (Entity)
    {
        Entity->SetThisAndClientsAsDestroyed();
        Entity = 0;
    }

#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->Finish();
        ToolsCutsceneActor->AnimID = -1;
		for(int i=0; i<ASURA_ANIMATION_SKIN_COUNT; i++)
		{
			ToolsCutsceneActor->m_aiSkinID[i] = -1;
		}
        ToolsCutsceneActor->NumberOfTracks = 0;
        ToolsCutsceneActor->Tracks = 0;
		ToolsCutsceneActor->m_iNumberOfAttachments = 0;
		ToolsCutsceneActor->m_ppxAttachments = 0;
        delete ToolsCutsceneActor;
		ToolsCutsceneActor = 0;
    }
	else
#endif
	{
		for(int i=0;i<m_iNumberOfAttachments;i++)
		{
			m_ppxAttachments[i]->Finish();
		}	
	}
}

//void Asura_ServerCutscene_Actor_Hierarchy::ChangeAnimation(Asura_String& animationName,float time,int flags,float tweenDuration)
void Asura_ServerCutscene_Actor_Hierarchy::ChangeAnimation(int animationID,float time,int flags,float tweenDuration,float fAnimDelay)
{
	Asura_Assert(Cutscene->Timer>=time,"");
	if(animationID==-1) return;
    Asura_String animationName = Cutscene->AnimationNames.GetName(animationID);
	if(!animationName.GetLength()) return;

#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->ChangeAnimation(animationID, time, flags, tweenDuration, fAnimDelay);
    }
#endif

	if(!Entity) return;
	Entity->ChangeAnimation(animationID, time, flags, tweenDuration);
}

void Asura_ServerCutscene_Actor_Hierarchy::ChangeAnimation(Asura_Hash_ID xAnimHashID, float fTime, int iFlags, float fTweenDuration)
{
	Asura_Assert(Cutscene->Timer>=fTime,"");
	if(xAnimHashID == ASURA_HASH_ID_UNSET) 
	{
		return;
	}

#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->ChangeAnimation(xAnimHashID, fTime, iFlags, fTweenDuration);
    }
#endif

	if(!Entity) return;
	Entity->ChangeAnimation(xAnimHashID, fTime, iFlags, fTweenDuration);
}

void Asura_ServerCutscene_Actor_Hierarchy::PlayAnimationToPos(int iAnimationID,float fTime,float fStartTimer,float fEndTimer,int iFlags,float fTweenDuration, float fAnimDelay)
{
	Asura_Assert(Cutscene->Timer>=fTime,"");
	if(iAnimationID==-1) return;
    Asura_String xAnimationName = Cutscene->AnimationNames.GetName(iAnimationID);
	if(!xAnimationName.GetLength()) return;

#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->PlayAnimationToPos(iAnimationID, fTime, fStartTimer, fEndTimer, iFlags, fTweenDuration, fAnimDelay);
    }
#endif

	if(!Entity) return;
	Entity->PlayAnimationToPos(iAnimationID, fTime, fStartTimer, fEndTimer, iFlags, fTweenDuration);
}


//Asura_Handle Asura_ServerCutscene_Actor_Hierarchy::PlaySecondaryAnimation(Asura_String& animationName,float time,int flags,float weight,float duration)
Asura_Handle Asura_ServerCutscene_Actor_Hierarchy::PlaySecondaryAnimation(int animationID,float time,int flags,float weight,float fTweenDuration,bool bAutoRemove, float fAnimDelay)
{
	if(animationID==-1) return ASURA_INVALID_HANDLE;
    Asura_String animationName = Cutscene->AnimationNames.GetName(animationID);
	if(!animationName.GetLength()) return ASURA_INVALID_HANDLE;

#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        return ToolsCutsceneActor->PlaySecondaryAnimation(animationID, time, flags, weight, fTweenDuration,bAutoRemove, fAnimDelay);
    }
#endif
	//RAR 29/03/04 - bAutoRemove currently ignored for server cutscenes (will always be autoremoved)

	if(!Entity) return ASURA_INVALID_HANDLE;
    Entity->PlaySecondaryAnimation(animationID, time, flags, weight, fTweenDuration);
    return ASURA_INVALID_HANDLE;
}

Asura_Handle Asura_ServerCutscene_Actor_Hierarchy::PlaySecondaryAnimationToPos(int iAnimationID,float fTime,float fStartTimer,float fEndTimer,int iFlags,float fWeight,float fTweenDuration,bool bAutoRemove, float fAnimDelay)
{
	Asura_Assert(Cutscene->Timer>=fTime,"");
	
    Asura_String xAnimationName = Cutscene->AnimationNames.GetName(iAnimationID);
    if(!xAnimationName.GetLength()) return ASURA_INVALID_HANDLE;

	#ifdef ASURA_TOOLS
	    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
	    {
	        return ToolsCutsceneActor->PlaySecondaryAnimationToPos(iAnimationID, fTime, fStartTimer, fEndTimer, iFlags, fWeight, fTweenDuration,bAutoRemove, fAnimDelay);
	    }
	#endif
	//RAR 29/03/04 - bAutoRemove currently ignored for server cutscenes (will always be autoremoved)

	if(!Entity) return ASURA_INVALID_HANDLE;
    Entity->PlaySecondaryAnimationToPos(iAnimationID, fTime, fStartTimer, fEndTimer, iFlags, fWeight, fTweenDuration);
    return ASURA_INVALID_HANDLE;
}


void Asura_ServerCutscene_Actor_Hierarchy::RemoveSecondaryAnimation(Asura_Handle handle)
{
#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->RemoveSecondaryAnimation(handle);
        return;
    }
#endif
	Asura_Assert(0,"server cutscene actor doesn't support RemoveSecondaryAnimation");
}

void Asura_ServerCutscene_Actor_Hierarchy::ChangeTextureAnimation(int iSequence,float fSpeed,u_int uFlags)
{
	Asura_Assert(0,"Texture animation events not implemented for server cutscenes yet.");
	#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
	{
		ToolsCutsceneActor->ChangeTextureAnimation(iSequence,fSpeed,uFlags);
	}
	#endif
	
	if(Entity)
	{
		Entity->ChangeTextureAnimation(iSequence,fSpeed,uFlags);
	}
}


void Asura_ServerCutscene_Actor_Hierarchy::AddAttachment_Shape(u_int uID,Asura_Hash_ID uMarkerHash,Asura_Hash_ID uShapeHash)
{
	if(!Entity) return;
	Entity->AddAttachment_Shape(uID,uMarkerHash,uShapeHash);
}

void Asura_ServerCutscene_Actor_Hierarchy::AddAttachment_Hierarchy(u_int uID,Asura_Hash_ID uMarkerHash,int iSkinID,int iAnimID)
{
	if(!Entity) return;
	Entity->AddAttachment_Hierarchy(uID,uMarkerHash,iSkinID,iAnimID);
}

void Asura_ServerCutscene_Actor_Hierarchy::AddAttachment_Actor(u_int uID,Asura_Hash_ID uMarkerHash,Asura_Guid uActorID)
{
	if(!Entity) return;
	Entity->AddAttachment_Actor(uID,uMarkerHash, uActorID);
}

void Asura_ServerCutscene_Actor_Hierarchy::AddAttachment_Entity(u_int uID,Asura_Hash_ID uMarkerHash,Asura_Guid uEntityID)
{
	if(!Entity) return;
	Entity->AddAttachment_Entity(uID,uMarkerHash, uEntityID);
}


void Asura_ServerCutscene_Actor_Hierarchy::RemoveAttachment(u_int uID)
{
	if(!Entity) return;
	Entity->RemoveAttachment(uID);
}

void Asura_ServerCutscene_Actor_Hierarchy::NotifyDestruction(const Asura_Entity_Server* pxEntity)
{
    if (Entity)
    {
        Asura_Assert(Entity == pxEntity, "NotifyDestruction called by unknown entity");
    }
    Entity = NULL;
}

void Asura_ServerCutscene_Actor_Hierarchy::GetSoundPosition(Asura_Vector_3& xPosition) const
{
	#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && ToolsCutsceneActor)
    {
        ToolsCutsceneActor->GetSoundPosition(xPosition);
		return;
    }
	#endif
	xPosition = Position;
}
#endif // #ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities

/////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Asura_ClientCutscene_Actor_Camera **
**-----------------------------------*/
#include "Camera/Asura_Camera.h"

bool Asura_ClientCutscene_Actor_Camera::Disable = false;

Asura_Euler_3 Asura_Cutscene_Actor_Camera::s_xPlayerControlEuler(Asura_Maths::ZeroEuler3);

Asura_Cutscene_Actor_Camera::Asura_Cutscene_Actor_Camera(Asura_Cutscene* pxCutscene, int iType /* = ACAT_Camera*/)
:Asura_Cutscene_Actor(pxCutscene, iType)
, m_xTargetName("")
, m_fInitialFOV(60.f)
, m_fNearClip(ASURA_CAMERA_DEFAULT_NEAR_CLIP)
, m_fFarClip(ASURA_CAMERA_DEFAULT_FAR_CLIP)
, Target(NULL)
, m_pxAnim(NULL)
, m_fAnimTimer(0.f)
, m_fLastAnimUpdateTime(0.f)
, m_iAnimationFlags(0)
, m_fCurrentFOV(60.f)
, m_fShakeChangeStartTime(0.f)
, m_fShakeChangeEndTime(0.f)
, m_uCameraFlags(0)
, m_fPlayerControlYawLimit(0.f)
, m_fPlayerControlPitchLimit(0.f)
, m_fInitialPlayerControlYawLimit(0.f)
, m_fInitialPlayerControlPitchLimit(0.f)
{
	m_fInitialFOV = Asura_Cutscene::Project_GetDefaultInitialFOV();
}

#ifdef ASURA_TOOLS
void Asura_Cutscene_Actor_Camera::Copy(Asura_Cutscene_Actor* _pxActor)
{
	Asura_Cutscene_Actor_Camera* pxActor = (Asura_Cutscene_Actor_Camera*) _pxActor;
	m_xTargetName = pxActor->m_xTargetName;
	m_fInitialFOV = pxActor->m_fInitialFOV;
	m_fNearClip = pxActor->m_fNearClip;
	m_fFarClip = pxActor->m_fFarClip;
	m_xInitialShake = pxActor->m_xInitialShake;
	Asura_Cutscene_Actor::Copy(pxActor);
}
#endif

void Asura_Cutscene_Actor_Camera::Initialise()
{
	Asura_Cutscene_Actor::Initialise();

	m_fCurrentFOV = m_fInitialFOV;

	//need to find the camera target
	if(m_xTargetName.IsBlank())
	{
		Target = NULL;
	}
	else
	{
		Target = Cutscene->GetActorByName(m_xTargetName.GetString());
	}

	//The first camera in the list should be the initially active camera
	if(!Cutscene->GetActiveCamera())
	{
		Cutscene->SetActiveCamera(this);		
	}
	m_fShakeChangeStartTime = m_fShakeChangeEndTime = 0.f;
	m_xLastShake = m_xChangedShake = m_xInitialShake;

	m_pxAnim = NULL;

	// Not sure if we always want to reset here.
	s_xPlayerControlEuler.Zero();

	m_fPlayerControlPitchLimit = m_fInitialPlayerControlPitchLimit;
	m_fPlayerControlYawLimit = m_fInitialPlayerControlYawLimit;
}

void Asura_Cutscene_Actor_Camera::Update(float frameTime)
{
	if(!NumberOfTracks) return;

	Asura_Cutscene_Actor::Update(frameTime);

	Asura_Cutscene_Track* pxTrack = Tracks[CurrentTrack];
	Asura_Cutscene_Track* pxNextTrack = CurrentTrack+1<NumberOfTracks ? Tracks[CurrentTrack+1] : 0;


	//need to update the orientation
	bool bUseCameraTarget = Target && Target->NumberOfTracks;
	if(m_pxAnim && !(m_iAnimationFlags & AnimationChangeFlag_UseCameraTarget))
	{
		//If using a camera animation then we ignore the camera target for the intial orientation ,unless
		//flagged otherwise. RAR 02/09/2005
		bUseCameraTarget = false;
	}
	if(bUseCameraTarget)
	{
		//get the direction to the target
		Asura_Vector_3 direction = Target->Position-Position;
		if(direction.x || direction.y || direction.z)
		{
			direction.Normalise();
			Asura_Maths::OrientationFromDirection(Orientation,direction);
		}

		if(pxTrack->m_uTrackFlags & TrackFlag_RollFromPoints)
		{
			//multiply in the roll from the camera track
			Asura_Matrix_3x3 xMat;
			Asura_Euler_3 xEuler;
			//get the orientation
			pxTrack->GetOrientationFromTime(Cutscene->Timer,&xMat,pxNextTrack);
			//convert to euler and kill off yaw/pitch
			xEuler.FromMatrix(xMat);
			xEuler.x = 0;
			xEuler.y = 0;
			xEuler.ToMatrix(xMat);

			Orientation = Orientation * xMat;



		}
	}
	else if(!m_bAttached)
	{

		//Get the current direction of the line
		pxTrack->GetOrientationFromTime(m_fTrackTimer,&Orientation,pxNextTrack);
	}

	if(m_pxAnim)
	{
		//We have an animation , so apply the orientation/offset of the first bone to the camera.

		//update the anim timer
		m_fAnimTimer += Cutscene->Timer - m_fLastAnimUpdateTime;
		m_fLastAnimUpdateTime = Cutscene->Timer;

		if(m_pxAnim->Duration && (m_iAnimationFlags & AnimationChangeFlag_Loop))
		{
			while(m_fAnimTimer>m_pxAnim->Duration)
			{
				m_fAnimTimer -= m_pxAnim->Duration;
			}
		}

		//determine how far into animation we are
		float fTimer = m_fAnimTimer / (m_pxAnim->Duration ? m_pxAnim->Duration : 1.0f);
		fTimer = Asura_Maths::ClampToRange(fTimer,0.0f,1.0f);

		//get the position/orient from the animation
		Asura_Quat xAnimQuat;
		Asura_Vector_3 xAnimOffset;
		
		m_pxAnim->GetRelativeBonePosition(0,fTimer,&xAnimOffset,&xAnimQuat);
		
		//and apply to the camera's position/orient
		Asura_Matrix_3x3 xAnimMat;
		xAnimQuat.ToMatrix(xAnimMat);

		Position += xAnimOffset*Orientation;
		Orientation = Orientation*xAnimMat;
		
	}

	// JDS 21/03/07 - Added a handheld camera mode. Jitters the final orientation of the camera with Perlin
	// noise to give the effect of a handheld camera.
	// For this to work well, the track really needs smooth speed data
	const float fSpeed = pxTrack->GetSpeedFromTime(m_fTrackTimer);

	if(m_fTrackTimer > m_fShakeChangeEndTime)
	{
		// Just use the final shake
		m_xChangedShake.ApplyShake(m_fTrackTimer, fSpeed, Orientation);
	}
	else if(m_fTrackTimer <= m_fShakeChangeStartTime)
	{
		// Just use the initial shake
		m_xLastShake.ApplyShake(m_fTrackTimer, fSpeed, Orientation);
	}
	else
	{
		// Otherwise construct a lerped version
		Asura_Cutscene_Camera_Shake xShake;
		const float fDuration = m_fShakeChangeEndTime - m_fShakeChangeStartTime;
		Asura_Assert(fDuration > 0.f, "This should be impossible - should go into one of the clauses above");
		const float fLambda = (m_fTrackTimer - m_fShakeChangeStartTime) / fDuration;
		xShake.Lerp(m_xLastShake, m_xChangedShake, fLambda);
		xShake.ApplyShake(m_fTrackTimer, fSpeed, Orientation);
	}

	if(m_uCameraFlags & ASURA_CAMERAACTORFLAG_ALLOWPLAYERCONTROL)
	{
		Asura_Matrix_3x3 xPlayerControlOri;
		// First make sure euler is within limits.
		const Asura_Angle fEulerXLimit = Asura_Maths::DegToRad(m_fPlayerControlPitchLimit);
		const Asura_Angle fEulerYLimit = Asura_Maths::DegToRad(m_fPlayerControlYawLimit);
		s_xPlayerControlEuler.x = Asura_Maths::ClampToRange(s_xPlayerControlEuler.x, -fEulerXLimit, fEulerXLimit);
		s_xPlayerControlEuler.y = Asura_Maths::ClampToRange(s_xPlayerControlEuler.y, -fEulerYLimit, fEulerYLimit);
		// Don't allow roll control for now.
		s_xPlayerControlEuler.z = 0.f;
		s_xPlayerControlEuler.ToMatrix(xPlayerControlOri);
		Orientation = Orientation * xPlayerControlOri;
	}
}

void Asura_Cutscene_Actor_Camera::ChangeShake(const Asura_Cutscene_Camera_Shake& xShake, float fChangeDuration)
{
	m_xLastShake = m_xChangedShake;
	m_xChangedShake = xShake;
	m_fShakeChangeStartTime = m_fTrackTimer;
	m_fShakeChangeEndTime = m_fTrackTimer + fChangeDuration;
}

unsigned Asura_Cutscene_Actor_Camera::GetDefaultTrackFlags() const
{
	return TrackFlag_PitchAlongTrack;
}

const Asura_Vector_3& Asura_Cutscene_Actor_Camera::GetStartPosition( void ) const
{
    Asura_Assert( NumberOfTracks > 0, "" );

    const Asura_Cutscene_Track* pxStartTrack = Tracks[0];
    Asura_Assert( pxStartTrack->NumberOfPivots > 0, "" );

    return pxStartTrack->Pivots[0];
}

void Asura_Cutscene_Actor_Camera::ChangeAnimation(int iIndex,float fStartTime,int iEventAnimationFlags)
{
	m_pxAnim = 0;
	m_fAnimTimer = Cutscene->Timer - fStartTime;
	m_fLastAnimUpdateTime = Cutscene->Timer;
	m_iAnimationFlags = iEventAnimationFlags;
	if(iIndex != -1)
	{
		const char* pszAnimName = Cutscene->AnimationNames.GetName(iIndex).GetString();
		if(pszAnimName)
		{
			m_pxAnim = Asura_Animation_System::GetAnim(pszAnimName);
		}
	}
}

void Asura_Cutscene_Actor_Camera::SetCameraTarget(const char* pszName)
{
	if(!pszName)
	{
		Target = 0;
	}
	else
	{
		Target = Cutscene->GetActorByName(pszName);
	}
}

void Asura_Cutscene_Actor_Camera::UpdateAsuraCamera() const
{
	// [7/15/2008 AS] Added test of Asura_Render::GetCutsceneHasCameraControl()
	// allows us to set this to false in the debugger and use flycam mid cut-scene
	if(Cutscene->IsControllingCamera() && Asura_Render::GetCutsceneHasCameraControl())
	{
		Asura_Camera::SetPosition(Position);
		Asura_Camera::SetOrientation(Orientation);

		// (MH - 02.06.2003) added force to camera 0 here.
		Asura_Camera::SetCameraIndex(0);
		Asura_Camera::SetFieldOfView(m_fCurrentFOV);

		//Get the cutscene to update the clip planes , since it will be storing the pre-cutscene states if needed
		Cutscene->UpdateClipPlanes((m_uCameraFlags & ASURA_CAMERAACTORFLAG_USENEARCLIP) ? m_fNearClip : -1.0f,
									(m_uCameraFlags & ASURA_CAMERAACTORFLAG_USEFARCLIP) ? m_fFarClip : -1.0f);	
	}
}




void Asura_ClientCutscene_Actor_Camera::Update(float frameTime)
{
    Asura_Cutscene_Actor_Camera::Update(frameTime);

    //update the camera
	if(this == Cutscene->GetActiveCamera())
	{
		UpdateAsuraCamera();
	}
}



#ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities

Asura_ServerCutscene_Actor_Camera::Asura_ServerCutscene_Actor_Camera(Asura_Cutscene* pxCutscene)
:Asura_Cutscene_Actor_Camera(pxCutscene)
{
    m_pxEntity = NULL;
#ifdef ASURA_TOOLS
    m_pxToolsCutsceneActor = NULL;
#endif
}

void Asura_ServerCutscene_Actor_Camera::Initialise()
{
    Asura_Cutscene_Actor_Camera::Initialise();

    if (!Asura_Cutscene::ToolsPlayback)
    {
        //initialise the server entity, if derived class hasn't done it yet.
		//Only want one camera entity regardless of the number of cameras.. so
		//let the initial active camera create it.
	    if (!m_pxEntity && this==Cutscene->GetActiveCamera())
        {
            m_pxEntity = new Asura_ServerEntity_ServerCutsceneActorCamera(this);
        }
    }
#ifdef ASURA_TOOLS
    else
    {
        m_pxToolsCutsceneActor = new Asura_ClientCutscene_Actor_Camera(Cutscene);
        // copy Asura_Cutscene_Actor
        m_pxToolsCutsceneActor->NumberOfTracks = NumberOfTracks;
	    m_pxToolsCutsceneActor->Tracks = Tracks;
	    m_pxToolsCutsceneActor->m_uActorFlags = m_uActorFlags;
		m_pxToolsCutsceneActor->m_uGuidToReplace = m_uGuidToReplace;
        // copy Asura_Cutscene_Actor_Camera
		m_pxToolsCutsceneActor->SetInitialFOV(m_fInitialFOV);
        m_pxToolsCutsceneActor->Initialise();
    }
#endif
	// (AW - 01.08.2003) - Set the initial position and ori of target and camera
	// to fix first-frame cutscene glitches.
	if (Target)
	{
		Target->Update(0.0f);
	}
	Update(0.0f);
}

#ifdef ASURA_TOOLS
void Asura_ServerCutscene_Actor_Camera::Update(float fFrameTime)
{
    if (Asura_Cutscene::ToolsPlayback && m_pxToolsCutsceneActor)
    {
		m_pxToolsCutsceneActor->SetFOV(m_fCurrentFOV);
        m_pxToolsCutsceneActor->Update(fFrameTime);

		//update position/orient from the client actor , since tools uses it to place
		//the camera tools object. RAR 17/01/2006
		Position = m_pxToolsCutsceneActor->Position;
		Orientation = m_pxToolsCutsceneActor->Orientation;


		//the fake cutscene actor won't update the camera, since it will
		//be the server actor that is the active camera.. so check here
		if(this == Cutscene->GetActiveCamera())
		{
			m_pxToolsCutsceneActor->UpdateAsuraCamera();
		}

    }
    else
    {
        Asura_Cutscene_Actor_Camera::Update(fFrameTime);
    }

}
#endif

void Asura_ServerCutscene_Actor_Camera::Finish()
{
    Asura_Cutscene_Actor_Camera::Finish();

    if (m_pxEntity)
    {
        m_pxEntity->SetThisAndClientsAsDestroyed();
        m_pxEntity = 0;
    }

#ifdef ASURA_TOOLS
    if (Asura_Cutscene::ToolsPlayback && m_pxToolsCutsceneActor)
    {
        m_pxToolsCutsceneActor->Finish();
        m_pxToolsCutsceneActor->NumberOfTracks = 0;
        m_pxToolsCutsceneActor->Tracks = 0;
        delete m_pxToolsCutsceneActor;
    }
#endif
}

#ifdef ASURA_TOOLS
void Asura_ServerCutscene_Actor_Camera::ChangeAnimation(int iIndex,float fStartTime,int iAnimationFlags)
{
	//apply the anim change to this
	Asura_Cutscene_Actor_Camera::ChangeAnimation(iIndex,fStartTime,iAnimationFlags);
    
	//In tools , apply to the client camera as well
    if (Asura_Cutscene::ToolsPlayback && m_pxToolsCutsceneActor)
	{
		m_pxToolsCutsceneActor->ChangeAnimation(iIndex,fStartTime,iAnimationFlags);
	}

}

#endif //ASURA_TOOLS

void Asura_ServerCutscene_Actor_Camera::NotifyDestruction(const Asura_Entity_Server* pxEntity)
{
	if (m_pxEntity)
	{
		Asura_Assert(m_pxEntity == pxEntity, "NotifyDestruction called by unknown entity");
	}
	m_pxEntity = NULL;
}
#endif // #ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities


/////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------**
** Asura_Cutscene_Actor_Shape  **
**-----------------------------*/

Asura_Cutscene_Actor_Shape::Asura_Cutscene_Actor_Shape(Asura_Cutscene* pxCutscene)
:Asura_Cutscene_Actor(pxCutscene,ACAT_Shape)
{
	m_uShapeID = ASURA_HASH_ID_UNSET;
	m_uShapeFlags = ASURA_SHAPEACTORFLAG_NOSHADOW;
}

#ifdef ASURA_TOOLS
void Asura_Cutscene_Actor_Shape::Copy(Asura_Cutscene_Actor* _pxActor)
{
	Asura_Cutscene_Actor_Shape* pxActor = (Asura_Cutscene_Actor_Shape*) _pxActor;
	m_uShapeID = pxActor->m_uShapeID;
	m_uShapeFlags = pxActor->m_uShapeFlags;
	Asura_Cutscene_Actor::Copy(pxActor);
}

void Asura_Cutscene_Actor_Shape::ListShapesUsed(List<Asura_Hash_ID>& xShapeList)
{
	if(m_uShapeID != ASURA_HASH_ID_UNSET)
	{
		xShapeList.add_entry(m_uShapeID);		
	}
}
#endif //ASURA_TOOLS

void Asura_Cutscene_Actor_Shape::Update(float fFrameTime)
{
	if(!NumberOfTracks) return;
	Asura_Cutscene_Actor::Update(fFrameTime);
	if(!m_bAttached)
	{
		UpdateOrientation();
	}
}


/////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_ClientCutscene_Actor_Shape **
**----------------------------------*/

Asura_ClientCutscene_Actor_Shape::Asura_ClientCutscene_Actor_Shape(Asura_Cutscene* pxCutscene)
:Asura_Cutscene_Actor_Shape(pxCutscene)
{
}

void Asura_ClientCutscene_Actor_Shape::Initialise()
{
	Asura_Cutscene_Actor_Shape::Initialise();

	m_xShapeInstance.SetShape(m_uShapeID);

	if(m_uShapeFlags & ASURA_SHAPEACTORFLAG_NOSHADOW)
	{
		m_xShapeInstance.AddFlags(ASURA_SHAPE_INSTANCE_NO_SHADOW);
	}
	else
	{
		m_xShapeInstance.RemoveFlags(ASURA_SHAPE_INSTANCE_NO_SHADOW);
	}
}

void Asura_ClientCutscene_Actor_Shape::Render()
{
	//if this actor is attached to another , then it will be rendered when the attachment is rendered
	if(m_bAttached) return;

	#ifdef ASURA_TOOLS
	//RAR 17/05/2006 Some actors may be hidden while recording AVIs
	if(Tools_CutsceneModeAVI::ShouldHideActor(this))
	{
		return;
	}
	#endif


	if (Cutscene &&
		Cutscene->m_iCutsceneFlags & AsuraCutsceneFlag_IgnoreModuleLinking)
	{
		m_xShapeInstance.AddFlags(ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING);
	}
	else
	{
		m_xShapeInstance.RemoveFlags(ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING);
	}

	m_xShapeInstance.SetNewPosAndOri(Position,Orientation);
	m_xShapeInstance.Render();
}

void Asura_ClientCutscene_Actor_Shape::RenderShadow()
{
	if( m_bAttached ) return;

	if( m_uShapeFlags & ASURA_SHAPEACTORFLAG_NOSHADOW ) return;

	#if defined ASURA_TOOLS
		if( Tools_CutsceneModeAVI::ShouldHideActor( this ) ) return;
	#endif

	m_xShapeInstance.RenderShadow();
}


#ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities
/////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Asura_ServerCutscene_Actor_Shape **
**----------------------------------*/

Asura_ServerCutscene_Actor_Shape::Asura_ServerCutscene_Actor_Shape(Asura_Cutscene* pxCutscene)
:Asura_Cutscene_Actor_Shape(pxCutscene)
{
	m_pxEntity = NULL;
	#ifdef ASURA_TOOLS
	m_pxToolsCutsceneActor = NULL;
	#endif
}

void Asura_ServerCutscene_Actor_Shape::Initialise()
{
	Asura_Cutscene_Actor_Shape::Initialise();

	if(!Asura_Cutscene::ToolsPlayback)
	{
        //initialise the server entity, if derived class hasn't done it yet.
	    if (!m_pxEntity)
        {
            m_pxEntity = new Asura_ServerEntity_ServerCutsceneActorShape(this,m_uShapeID,m_uGuid);
        }
        // make sure the client version is created now in case messages occur out of order.
        m_pxEntity->MakeClientUpdateMessage();
	}
	#ifdef ASURA_TOOLS
	else
	{
		m_pxToolsCutsceneActor = new Asura_ClientCutscene_Actor_Shape(Cutscene);
        // copy Asura_Cutscene_Actor
        m_pxToolsCutsceneActor->NumberOfTracks = NumberOfTracks;
	    m_pxToolsCutsceneActor->Tracks = Tracks;
		m_pxToolsCutsceneActor->m_uGuidToReplace = m_uGuidToReplace;
        // copy Asura_Cutscene_ActorShape
		m_pxToolsCutsceneActor->m_uShapeID = m_uShapeID;
        m_pxToolsCutsceneActor->Initialise();
	}
	#endif
}

#ifdef ASURA_TOOLS
void Asura_ServerCutscene_Actor_Shape::Render()
{
    if (Cutscene->ToolsPlayback && m_pxToolsCutsceneActor)
    {
        m_pxToolsCutsceneActor->Position = Position;
        m_pxToolsCutsceneActor->Orientation = Orientation;

        m_pxToolsCutsceneActor->Render();
    }
}
#endif // ASURA_TOOLS

void Asura_ServerCutscene_Actor_Shape::Finish()
{
    Asura_Cutscene_Actor_Shape::Finish();

	
    if (m_pxEntity)
    {
        m_pxEntity->SetThisAndClientsAsDestroyed();
        m_pxEntity = 0;
    }

#ifdef ASURA_TOOLS
    if (Cutscene->ToolsPlayback && m_pxToolsCutsceneActor)
    {
        m_pxToolsCutsceneActor->Finish();
        m_pxToolsCutsceneActor->NumberOfTracks = 0;
        m_pxToolsCutsceneActor->Tracks = 0;
        delete m_pxToolsCutsceneActor;
    }
#endif
}


void Asura_ServerCutscene_Actor_Shape::NotifyDestruction(const Asura_Entity_Server* pxEntity)
{
    if (m_pxEntity)
    {
        Asura_Assert(m_pxEntity == pxEntity, "NotifyDestruction called by unknown entity");
    }
    m_pxEntity = NULL;
}
#endif // #ifdef ASURA_SUPPORT_SERVERCUTSCENES	// the Gunfighter-style cutscenes with entities

#ifdef ASURA_TOOLS
bool Asura_Cutscene_Actor_CameraTarget::NeedsUpdateBefore(const Asura_Cutscene_Actor* pxActor) const
{
	if(pxActor->IsCamera()) return true;//camera target needs to be before the camera
	return Asura_Cutscene_Actor::NeedsUpdateBefore(pxActor);
}
#endif
