//////////////////////////////////////////////////////////////////////////////
//
// Asura_CPUProfile.h
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Created 1st September 2004 by Mike Healey
//
// Description:
//
//		Interface for graphical CPU profile. Calls through to platform code.
//		I decided not to implement platform-independent code to collect and
//		display timing results because (1) we already have a good system in 
//		place on the PS2, and (2) profiling code should be as unintrusive as
//		possible and platform-specific code is generally a bit faster than 
//		independent code.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_CPUPROFILE_H___
#define ___ASURA_CPUPROFILE_H___ 1

// Asura_Platform_CPUProfile.h should provide the following defines
//#define ASURA_CPU_PROFILE__START(x)	PLATFORM_START_FUNCTION(x)
//#define ASURA_CPU_PROFILE__END(x)		PLATFORM_END_FUNCTION(x)
// It is up to each platform to put in place a way of disabling the profiling code
// for final builds. This is typically done using defines in Asura_ProjectSettings.h
#include "Debugging/Asura_Platform_CPUProfile.h"

// you can either begin/end or start/stop or mix 'n' match
#define ASURA_CPU_PROFILE__BEGIN ASURA_CPU_PROFILE__START
#define ASURA_CPU_PROFILE__STOP ASURA_CPU_PROFILE__END

// profile timings for each profile are accumulated and drawn in the order shown here.
// At the moment you can't display a graph of timings in the order they were collected.
// Possible additional functionality for this system would be to add that, and to add a
// table display which maps colours on to text.
// This could use the form.
// ASURA_CPU_PROFILE__START_LABEL(x,y) { ASURA_CPU_PROFILE__START(x); ASURA_CPU_PROFILE__REGISTER_LABEL(x,y); }
// (MH - 16.08.2005) implementing this now!

// platform can now define ASURA_CPU_PROFILE__START_LABEL directly (useful for SN tuner)
#ifndef ASURA_CPU_PROFILE__START_LABEL
// if the platform defined the register label macro, then use it otherwise ignore it
#ifdef ASURA_CPU_PROFILE__REGISTER_LABEL
#define ASURA_CPU_PROFILE__START_LABEL(x,y) \
	{ ASURA_CPU_PROFILE__REGISTER_LABEL(x,y); ASURA_CPU_PROFILE__START(x); }
#else
#define ASURA_CPU_PROFILE__START_LABEL(x,y) \
	{ ASURA_CPU_PROFILE__START(x); }
#endif
#endif // #ifndef ASURA_CPU_PROFILE__START_LABEL

#if !defined( ASURA_CPU_PROFILE__START_THREAD)
	#define ASURA_CPU_PROFILE__START_THREAD( x)
#endif

#if !defined( ASURA_CPU_PROFILE__STOP_THREAD)
	#define ASURA_CPU_PROFILE__STOP_THREAD()
#endif

#if defined( ASURA_USE_PROFILING_SYSTEM_MULTITHREADED)
	#define ASURA_CPU_PROFILE__START_MULTITHREADED( x) ASURA_CPU_PROFILE__START( x)
	#define ASURA_CPU_PROFILE__START_MULTITHREADED_LABEL( x, y) ASURA_CPU_PROFILE__START_LABEL( x, y)
	#define ASURA_CPU_PROFILE__STOP_MULTITHREADED( x) ASURA_CPU_PROFILE__STOP( x)
#else
	// DLH - 24 May 2007 11:53:55 - Do not profile Threaded tasks
	#define ASURA_CPU_PROFILE__START_MULTITHREADED( x) ASURA_CPU_PROFILE__START( x)
	#define ASURA_CPU_PROFILE__START_MULTITHREADED_LABEL( x, y)
	#define ASURA_CPU_PROFILE__STOP_MULTITHREADED( x)
#endif

// So, the engine labels are drawn first (leftmost), and the project ones are drawn afterwards.
// These colours happen to match the PS2 profiler's colours so don't change them.
// Entry 0 is reserved for the total frame marker. (Probably drawn in grey)

// DLH - 27 June 2006 11:00:17 - This is the original colour palette index list. This was reeduced down in size to 32 values from 16
/*
#define ASURA_CPU_PROFILE__ENGINE_BLUE		1 
#define ASURA_CPU_PROFILE__ENGINE_GREEN		2 
#define ASURA_CPU_PROFILE__ENGINE_RED		3 
#define ASURA_CPU_PROFILE__ENGINE_CYAN		4 
#define ASURA_CPU_PROFILE__ENGINE_YELLOW	5 
#define ASURA_CPU_PROFILE__ENGINE_MAGENTA	6 
#define ASURA_CPU_PROFILE__ENGINE_BLACK		7 
#define ASURA_CPU_PROFILE__ENGINE_WHITE		8 
#define ASURA_CPU_PROFILE__PROJECT_BLUE		9 
#define ASURA_CPU_PROFILE__PROJECT_GREEN	10
#define ASURA_CPU_PROFILE__PROJECT_RED		11
#define ASURA_CPU_PROFILE__PROJECT_CYAN		12
#define ASURA_CPU_PROFILE__PROJECT_YELLOW	13
#define ASURA_CPU_PROFILE__PROJECT_MAGENTA	14
#define ASURA_CPU_PROFILE__PROJECT_BLACK	15

#define ASURA_CPU_PROFILE__ENGINE_DARK_BLUE		32 
#define ASURA_CPU_PROFILE__ENGINE_DARK_GREEN	33
#define ASURA_CPU_PROFILE__ENGINE_DARK_RED		34
#define ASURA_CPU_PROFILE__ENGINE_ORANGE		35
#define ASURA_CPU_PROFILE__ENGINE_PINK			36
#define ASURA_CPU_PROFILE__ENGINE_BROWN			37
*/

enum Asura_CPUProfileIndex
{
	ASURA_CPU_PROFILE__TOTALTIME		= 0,
	ASURA_CPU_PROFILE__ENGINE_BLUE		= 1,
	ASURA_CPU_PROFILE__ENGINE_GREEN		= 2,
	ASURA_CPU_PROFILE__ENGINE_RED		= 3,
	ASURA_CPU_PROFILE__ENGINE_CYAN		= 4,
	ASURA_CPU_PROFILE__ENGINE_YELLOW	= 5,
	ASURA_CPU_PROFILE__ENGINE_MAGENTA	= 6,
	ASURA_CPU_PROFILE__ENGINE_BLACK		= 7,
	
	ASURA_CPU_PROFILE__ENGINE_WHITE		= 8,
	ASURA_CPU_PROFILE__PROJECT_BLUE		= 9,
	ASURA_CPU_PROFILE__PROJECT_GREEN	= 10,
	ASURA_CPU_PROFILE__PROJECT_RED		= 11,
	ASURA_CPU_PROFILE__PROJECT_CYAN		= 12,
	ASURA_CPU_PROFILE__PROJECT_YELLOW	= 13,
	ASURA_CPU_PROFILE__PROJECT_MAGENTA	= 14,
	ASURA_CPU_PROFILE__PROJECT_BLACK	= 15,
	
	ASURA_CPU_PROFILE__PROJECT_WHITE	= 16,
	ASURA_CPU_PROFILE__PROJECT_PALE_BLUE	= 17,
	ASURA_CPU_PROFILE__PROJECT_PALE_GREEN	= 18,
	ASURA_CPU_PROFILE__PROJECT_PALE_RED		= 19,
	ASURA_CPU_PROFILE__PROJECT_PALE_CYAN	= 20,
	ASURA_CPU_PROFILE__PROJECT_PALE_YELLOW	= 21,
	ASURA_CPU_PROFILE__PROJECT_PALE_MAGENTA	= 22,
	ASURA_CPU_PROFILE__PROJECT_PALE_BLACK	= 23,

	ASURA_CPU_PROFILE__ENGINE_GREY			= 24,
	ASURA_CPU_PROFILE__ENGINE_DARK_BLUE		= 25,
	ASURA_CPU_PROFILE__ENGINE_DARK_GREEN	= 26,
	ASURA_CPU_PROFILE__ENGINE_DARK_RED		= 27,
	ASURA_CPU_PROFILE__ENGINE_PALE_CYAN		= 28,
	ASURA_CPU_PROFILE__ENGINE_PALE_YELLOW	= 29,
	ASURA_CPU_PROFILE__ENGINE_PALE_MAGENTA	= 30,
	ASURA_CPU_PROFILE__ENGINE_PALE_BLACK	= 31,

	ASURA_CPU_PROFILE__ENGINE_DARKER_GREY	= 32,
	ASURA_CPU_PROFILE__ENGINE_DARKER_BLUE	= 33,
	ASURA_CPU_PROFILE__ENGINE_DARKER_RED	= 34,
	ASURA_CPU_PROFILE__ENGINE_DARKER_GREEN	= 35,

	ASURA_CPU_PROFILE__ENGINE_PURPLE		= 36,
	ASURA_CPU_PROFILE__ENGINE_DARK_PURPLE   = 37,

	ASURA_CPU_PROFILE__MAX_TIMERS			= 38
};


// KOR - 17/09/2007 - please keep the various profile bar Wiki pages up to date whenever
// changing these. http://asurawiki.rebellion.co.uk/twiki/bin/view/Programming/HowTo
#define ASURA_CPU_PROFILE__TOTALFRAME			0
#define ASURA_CPU_PROFILE__ENTITY_UPDATE				ASURA_CPU_PROFILE__ENGINE_BLUE
#define ASURA_CPU_PROFILE__SPECIALFX_UPDATE				ASURA_CPU_PROFILE__ENGINE_GREEN
#define ASURA_CPU_PROFILE__SPECIALFX_FINISHALLJOBS		ASURA_CPU_PROFILE__ENGINE_DARK_GREEN
#define ASURA_CPU_PROFILE__DYNAMICS						ASURA_CPU_PROFILE__ENGINE_RED
#define ASURA_CPU_PROFILE__RAGDOLLS						ASURA_CPU_PROFILE__ENGINE_CYAN
#define ASURA_CPU_PROFILE__WORLD_RENDER					ASURA_CPU_PROFILE__ENGINE_YELLOW
#define ASURA_CPU_PROFILE__ENTITY_RENDER				ASURA_CPU_PROFILE__ENGINE_MAGENTA
#define ASURA_CPU_PROFILE__SPECIALFX_RENDER				ASURA_CPU_PROFILE__ENGINE_BLACK
#define ASURA_CPU_PROFILE__NETWORKMESSAGES				ASURA_CPU_PROFILE__ENGINE_WHITE
#define ASURA_CPU_PROFILE__AXON_UPDATE					ASURA_CPU_PROFILE__ENGINE_PALE_YELLOW
#define ASURA_CPU_PROFILE__VISIBILITY_UPDATE			ASURA_CPU_PROFILE__ENGINE_DARKER_GREEN
#define ASURA_CPU_PROFILE__TEXTURE_STREAMING_UPDATE		ASURA_CPU_PROFILE__ENGINE_DARK_BLUE
#define ASURA_CPU_PROFILE__SOUND_UPDATE					ASURA_CPU_PROFILE__ENGINE_GREY

// ART - 17/05/06 - Streaming profile bars...
#define ASURA_CPU_PROFILE__STREAMING_UPDATE			ASURA_CPU_PROFILE__ENGINE_DARK_BLUE
#define ASURA_CPU_PROFILE__STREAMING_ENDOFFRAME		ASURA_CPU_PROFILE__ENGINE_DARK_RED

// JAT 24/03/06 match the PS2 colour index but not the colour on the PS2
#define ASURA_CPU_PROFILE__CARPET_RENDER		ASURA_CPU_PROFILE__ENGINE_PALE_MAGENTA
#define ASURA_CPU_PROFILE__BOTANICALS_RENDER	ASURA_CPU_PROFILE__ENGINE_PALE_BLACK
#define ASURA_CPU_PROFILE__BOTANICALS_UPDATE	ASURA_CPU_PROFILE__ENGINE_PALE_CYAN

#define ASURA_CPU_PROFILE__PRESENT_FRAME		ASURA_CPU_PROFILE__ENGINE_DARKER_RED

#define ASURA_CPU_PROFILE__CREATE_SHADOWS		ASURA_CPU_PROFILE__ENGINE_DARKER_GREY
#define ASURA_CPU_PROFILE__DEFERRED_LIGHTS		ASURA_CPU_PROFILE__ENGINE_DARKER_BLUE

#define ASURA_CPU_PROFILE__MISC_RENDER			ASURA_CPU_PROFILE__ENGINE_PURPLE
#define ASURA_CPU_PROFILE__MISC_UPDATE			ASURA_CPU_PROFILE__ENGINE_DARK_PURPLE

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_CPUPROFILE_H___
