///////////////////////////////////////////////////////////////////////////
//
// Asura_DebugOverlay.cpp
//
// Created by JAT - 3rd May 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// The Asura_DebugOverlay system - a static singleton class
// A very simple deferred debug rendering system see
//
// http://asurawiki.rebellion.co.uk/twiki/bin/view/Main/AsuraDebug
//
// * Unified graphics based debug overlay system - 
//    so different systems can have their own graphical debug overlays (called debug modes) 
// * Have base Asura modes e.g. Render, Physics, Input, Global, Entity, AI, Sound, Network, Messaging 
// * Supports project specific modes from the outset 
// * Self-contained system very little reliance on external systems (all done via registering callbacks) 
// * Deterministic rendering order - support screenspace and 3D, plus z-buffer test on/off 
// * Basic list of rendering primitives using existing render functionality 
// * Supports input handling and a callback per mode
//
/// !defined( ASURA_USE_DEBUGOVERLAY ) then the class and its interface doesn't exist
/// ASURA_USE_DEBUGOVERLAY == 0 then all the functions exist but they are inlined empty functions (which will be removed by the optimizer)
/// ASURA_USE_DEBUGOVERLAY == 1 then all the functions exist and they work
//
///////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if ASURA_USE_DEBUGOVERLAY == 1

// Default is a 32k buffer - this used to be 1MB!
#ifndef ASURA_DEBUGOVERLAY_MEMORY_SIZE
#define ASURA_DEBUGOVERLAY_MEMORY_SIZE					(32*1024)
#endif // #ifndef ASURA_DEBUGOVERLAY_MEMORY_SIZE

// Default maximum number of modes
#ifndef ASURA_DEBUGOVERLAY_MAX_NUM_MODES
#define ASURA_DEBUGOVERLAY_MAX_NUM_MODES				(16)
#endif // #ifndef ASURA_DEBUGOVERLAY_MAX_NUM_MODES

// This classes header file
#include "Debugging/Asura_DebugOverlay.h"

// Asura files
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_StringBuffer.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_DebugOverlay_Alphabet.h"
#include "Debugging/Asura_DebugOverlay_Modes.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Frustum.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

////////////////////////////////////////////////////////////////////////////////////////////////
//
// Put internal data and functions into a namespace to keep the interface clean in the header file (and not pollute the global namespace)
//
////////////////////////////////////////////////////////////////////////////////////////////////

namespace Asura_DebugOverlayData
{
	struct Asura_DebugOverlayMode
	{
		void Initialise( void );
		Asura_String64 m_xName;
		AsuraDebugOverlayCallbackFn m_pDoDebugFunc;
		int m_iMode;
		void* m_pxUserData;
	};
	static int s_iNumModes;

	enum ASURA_DO_COMMAND
	{
		ASURA_DO_SPHERE = 0,
		ASURA_DO_LINE ,
		ASURA_DO_ARROW,
		ASURA_DO_CROSS,
		ASURA_DO_OBB,
		ASURA_DO_AXIS,
		ASURA_DO_BOUNDING_BOX,
		ASURA_DO_FRUSTUM,
		ASURA_DO_CONSOLE_STRING,
		ASURA_DO_DEBUG_STRING,
		ASURA_DO_NUM_PRIMS
	};
	const int ASURA_DO_HEADER_SIZE = sizeof( u_int ) + sizeof( u_int );
	const int ASURA_DO_FLOAT_SIZE = sizeof( float );
	const int ASURA_DO_VECTOR_SIZE = 3 * ASURA_DO_FLOAT_SIZE;
	const int ASURA_DO_QUAT_SIZE = 4 * ASURA_DO_FLOAT_SIZE;
	const int ASURA_DO_UINT_SIZE = sizeof( u_int );
	const int ASURA_DO_ARGB_SIZE = sizeof( u_int );

	const int ASURA_DO_COMMAND_BIT_MASK = 0xFFFF;
	const int ASURA_DO_WORLD_SPACE_BIT_MASK = ( 1 << 16 );
	const int ASURA_DO_IGNORE_Z_BUFFER_BIT_MASK = ( 1 << 17 );

	static u_char* s_pucListStart = NULL;
	static u_char* s_pucListEnd = NULL;
	static u_char* s_pucListNext = NULL;
	static u_int s_uBufferUsedLastFrame = 0;
	static u_int s_uPeakBufferUsed = 0;
	static u_int s_auPrimMemSize[ASURA_DO_NUM_PRIMS] = 
	{
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_ARGB_SIZE + ASURA_DO_UINT_SIZE + ASURA_DO_UINT_SIZE,		// sphere
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,												// line
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_FLOAT_SIZE,	// arrow
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE + ASURA_DO_FLOAT_SIZE,													// cross
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,							// OBB
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_FLOAT_SIZE,							// axis
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_ARGB_SIZE,	// bounding box
		ASURA_DO_HEADER_SIZE + ASURA_DO_FLOAT_SIZE * 4 + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_ARGB_SIZE,						// frustum
		ASURA_DO_HEADER_SIZE + ASURA_DO_ARGB_SIZE,																								// console string - not a fixed size
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,																		// debug string - not a fixed size
	};
	static Asura_FixedArray<Asura_DebugOverlayMode, ASURA_DEBUGOVERLAY_MAX_NUM_MODES> s_xModes;

	static Asura_Vector_2 s_xConsoleStartPos( 0.05f, 0.21f );
	static Asura_Vector_2 s_xConsoleLineSize( 0.0f, 0.025f );
	static Asura_Vector_2 s_xConsoleCurrentPos( 0.0f, 0.0f );

	static u_int s_uDefaultConsoleColour;
	static u_int s_uModeTextColour;

	bool Initialise( const int iMemorySize ); 
	bool Shutdown( void ); 
	void Reset( void ); 
	void Render( void );
	void Update( void );

	bool RegisterMode( const char* const szName, const int iMode, AsuraDebugOverlayCallbackFn pCallbackFn, void* const pxUserData );
	int FindModeIndex( const int iMode );

	bool CanAdd( const enum ASURA_DO_COMMAND eCommand );
	bool AddData( const void* const pData, const size_t uNumBytes );
	bool AddFloat( const float fValue );
	bool AddUint( const u_int uValue );
	bool AddVector( const Asura_Vector_3& xVector );
	bool AddQuat( const Asura_Quat& xQuat );
	bool AddRGBA( const Asura_RGBA& xColour );
	bool AddARGB( const u_int uARGB );
	bool AddString( const char* const szString, const u_int uNumBytes );

	// For simple fixed sized primitives
	bool AddHeader( const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer );
	// For primitives with variable size e.g. text based ones
	bool AddHeader( const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes );

	bool ReadData( void* const pData, const size_t uNumBytes );
	bool ReadFloat( float& fValue );
	bool ReadUint( u_int& uValue );
	bool ReadVector( Asura_Vector_3& xVector );
	bool ReadQuat( Asura_Quat& xQuat );
	bool ReadRGBA( Asura_RGBA& xColour );
	bool ReadARGB( u_int& uARGB );
	bool ReadString( const char*& szString, const u_int uNumBytes );

	bool ReadHeader( u_int& uNumBytes, enum ASURA_DO_COMMAND& eCommand, bool& bWorldSpace, bool& bIgnoreZBuffer );

	void SortModeList( void );
	void ResetCommandList( void );
	void RenderCommandList( const u_char* const pucListEnd, const bool bIgnoreZBuffer );

	// Render functions for each command
	bool RenderSphere( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderFlatLine( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderArrow( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderCross( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderOBB( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderAxis( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderBoundingBox( const bool bWorldSpace, const bool bIgnoreZBuffer );
	bool RenderFrustum( const bool bWorldSpace, const bool bIgnoreZBuffer );

	bool RenderDebugString( const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes );
	bool RenderConsoleString( const u_int uNumBytes );

	void RenderStringScreenInternal( const char* const szString, const u_int uARGB, const Asura_Vector_2& xPos );
	void RenderStringWorldInternal( const char* const szString, const u_int uARGB, const Asura_Vector_3& xPos );

	void RenderStringInternal( const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const char* const szString, const u_int uARGB );
	void RenderLineStripInternal( const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const float* const pfCoords, const u_int uNumPoints, const u_int uARGB );
};

////////////////////////////////////////////////////////////////////////////////////////////////
//
// The functions themselves
//
////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_DebugOverlayData::Asura_DebugOverlayMode::Initialise( void )
{
	m_iMode = ASURA_DO_RESERVED;
	m_pDoDebugFunc = NULL;
	m_pxUserData = NULL;
	m_xName.MakeBlank();
}


// Initialise internals - allocate memory buffer
bool Asura_DebugOverlayData::Initialise( const int iMemorySize )
{
	s_pucListEnd = NULL;
	s_pucListNext = NULL;
	s_uBufferUsedLastFrame = 0;
	s_uPeakBufferUsed = 0;
	s_pucListStart = new u_char[iMemorySize];
	if ( s_pucListStart == NULL )
	{
		return false;
	}
	s_pucListEnd = s_pucListStart + iMemorySize;
	s_pucListNext = s_pucListStart;

	for ( int i = 0; i < ASURA_DEBUGOVERLAY_MAX_NUM_MODES; i++ )
	{
		s_xModes[i].Initialise();
	}
	s_iNumModes = 0;

	const Asura_RGBA xDefaultConsoleColour( 150.0f/255.0f, 1.0f, 180.0f/255.0f, 1.0f );
	s_uDefaultConsoleColour = xDefaultConsoleColour.ToUIntARGB();

	const Asura_RGBA xModeTextColour( 200.0f/255.0f, 50.0f/255.0f, 100.0f/255.0f, 1.0f );
	s_uModeTextColour = xModeTextColour.ToUIntARGB();

	return true;
}

// Free memory
bool Asura_DebugOverlayData::Shutdown( void )
{
	delete[] s_pucListStart;
	s_pucListStart = NULL;
	s_pucListEnd = NULL;
	s_pucListNext = NULL;
	return true;
}
	
void Asura_DebugOverlayData::ResetCommandList( void )
{
	// store current usage, reset the list back to its beginning and the console render pos to the start
	s_uBufferUsedLastFrame = s_pucListNext - s_pucListStart;
	s_uPeakBufferUsed = Asura_Maths::Max(s_uBufferUsedLastFrame, s_uPeakBufferUsed);

	s_pucListNext = s_pucListStart;
	s_xConsoleCurrentPos = s_xConsoleStartPos;
}

// Reset internal state
void Asura_DebugOverlayData::Reset( void )
{
	ResetCommandList();
}

void Asura_DebugOverlayData::RenderCommandList( const u_char* const pucListEnd, const bool bDrawIgnoreZBuffer )
{
	while ( s_pucListNext < pucListEnd )
	{
		// Read the next header from the list
		u_int uNumBytes;
		ASURA_DO_COMMAND eCommand;
		bool bWorldSpace;
		bool bIgnoreZBuffer;
		if ( ReadHeader( uNumBytes, eCommand, bWorldSpace, bIgnoreZBuffer ) == false )
		{
			Asura_Assert( false, "RenderCommandList ReadHeader failed" );
			return;
		}

		if ( bDrawIgnoreZBuffer != bIgnoreZBuffer )
		{
			// We have already read the header
			s_pucListNext += ( uNumBytes - ASURA_DO_HEADER_SIZE );
			continue;
		}
		bool bOK = false;

		switch ( eCommand )
		{
			case ASURA_DO_SPHERE:
			{
				bOK = RenderSphere( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_LINE:
			{
				bOK = RenderFlatLine( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_ARROW:
			{
				bOK = RenderArrow( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_CROSS:
			{
				bOK = RenderCross( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_OBB:
			{
				bOK = RenderOBB( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_AXIS:
			{
				bOK = RenderAxis( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_BOUNDING_BOX:
			{
				bOK = RenderBoundingBox( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_FRUSTUM:
			{
				bOK = RenderFrustum( bWorldSpace, bIgnoreZBuffer );
				break;
			}
			case ASURA_DO_CONSOLE_STRING:
			{
				bOK = RenderConsoleString( uNumBytes );
				break;
			}
			case ASURA_DO_DEBUG_STRING:
			{	
				bOK = RenderDebugString( bWorldSpace, bIgnoreZBuffer, uNumBytes );
				break;
			}
#if 0
			case GOURAUD_LINE:
			{
				RenderGouraudLine( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case TRIANGLE:
			{
				RenderTriangle( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case QUAD:
			{
				RenderQuad( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case CIRCLE:
			{
				RenderCircle( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case SPHERE_MATRIX:
			{
				RenderSphereMatrix( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case CYLINDER:
			{
				RenderCylinder( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case SAUSAGE:
			{
				RenderSausage( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case CONE:
			{
				RenderCone( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case CONE_MATRIX:
			{
				RenderConeMatrix( oColour, eDrawType, io_graphicsDevice );
				break;
			}
			case CUBE:
			{
				RenderCube( oColour, eDrawType, io_graphicsDevice );
				break;
			}
#endif // #if 0
			default:
			{
				Asura_Assertf( false, ( "RenderCommandList unknown eCommand %d\n", eCommand ) );
				break;
			}
		}
		if ( bOK == false )
		{
			// If the render fails stop reading the command list
			// Yuck having to use a const_cast!
			s_pucListNext = const_cast<u_char*>( pucListEnd );
		}
	}
}

int Asura_DebugOverlayData::FindModeIndex( const int iMode )
{
	for ( int i = 0; i < s_iNumModes; i++ )
	{
		if ( s_xModes[i].m_iMode == iMode )
		{
			return i;
		}
	}
	return -1;
}

void Asura_DebugOverlayData::Update( void )
{
	// Find the current mode and call its callback function
	const int iModeIndex = FindModeIndex( Asura_DebugOverlay::GetCurrentMode() );
	if ( iModeIndex < 0 )
	{
		return;
	}

	// Call the user-supplied callback function
	Asura_DebugOverlayData::Asura_DebugOverlayMode& xMode = s_xModes[iModeIndex];
	if ( xMode.m_pDoDebugFunc != NULL )
	{
		xMode.m_pDoDebugFunc( xMode.m_pxUserData );
	}
}

// JAT 21/10/07
// What should we do about split-screens?
// In CoreEngine we rendered the 3D debug data once per screen and the 2D data in full-screen part of the render loop

// Render the debug command list and info
void Asura_DebugOverlayData::Render( void )
{
	// Set the debug level string
	const char* szDebugLevel = "None";
	const int iModeIndex = FindModeIndex( Asura_DebugOverlay::GetCurrentMode() );
	if ( iModeIndex >= 0 )
	{
		szDebugLevel = s_xModes[iModeIndex].m_xName.GetString();
	}

	s_xConsoleCurrentPos = s_xConsoleStartPos;
	const float fFontHeight = Asura_DebugOverlay::GetUseVectorFont() ? 
		Asura_DebugOverlay::GetVectorFontWidth() / Asura_DebugOverlay::GetVectorFontWHRatio() : 
		ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
	s_xConsoleLineSize.y = fFontHeight / static_cast<float>(Asura_Viewport::GetHeight());

	// Display the debug mode string and memory usage
	Asura_String256 xDebugTitle( "Debug Mode: %s (%d) Buffer usage: %.02f%%", szDebugLevel, Asura_DebugOverlay::GetCurrentMode(), 
		Asura_Maths::FtoD( 100.0f * ( static_cast<float>(s_uBufferUsedLastFrame) ) / (static_cast<float>( s_pucListEnd - s_pucListStart ) ) ) );

	RenderStringScreenInternal( xDebugTitle.GetString(), s_uModeTextColour, s_xConsoleCurrentPos );
	s_xConsoleCurrentPos += s_xConsoleLineSize;

	const Asura_Vector_2 xConsoleStartPos = s_xConsoleCurrentPos;

	u_char* pucListEnd = s_pucListNext;

	// Need to decode the loop multiple times so as to draw things in the correct order

	// Render the stuff which needs z-buffering
	s_xConsoleCurrentPos = xConsoleStartPos;
	s_pucListNext = s_pucListStart;

	Asura_Render::CurrentStates.RequestZBufferEnabled( true );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
	RenderCommandList( pucListEnd, false );

	// Render the stuff which doesn't need z-buffering
	s_xConsoleCurrentPos = xConsoleStartPos;
	s_pucListNext = s_pucListStart;

	Asura_Render::CurrentStates.RequestZBufferEnabled( false );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
	RenderCommandList( pucListEnd, true );

	// Put back z-test and z-write
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool Asura_DebugOverlayData::RenderSphere( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xCentre;
	if ( ReadVector( xCentre ) == false )
	{
		Asura_Assert( false, "RenderSphere failed to read centre vector" );
		return false;
	}
	float fRadius;
	if ( ReadFloat( fRadius ) == false )
	{
		Asura_Assert( false, "RenderSphere failed to read radius value" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderSphere failed to read ARGB value" );
		return false;
	}
	u_int uNumSegments;
	if ( ReadUint( uNumSegments ) == false )
	{
		Asura_Assert( false, "RenderSphere failed to read num segments value" );
		return false;
	}
	u_int uNumSlices;
	if ( ReadUint( uNumSlices ) == false )
	{
		Asura_Assert( false, "RenderSphere failed to read num slices value" );
		return false;
	}
	if ( ( uNumSegments == 0 ) || ( uNumSlices == 0 ) )
	{
		return true;
	}

	// Add the slices to be rendered...
	// Don't do slice 0 or slice = uNumSlices because their radius is 0
	for ( u_int uSlice = 1; uSlice < uNumSlices; uSlice++ )
	{
		const Asura_Angle xSliceAngle = ( Asura_Maths::Deg180 * static_cast<float>( uSlice ) ) / uNumSlices;
		const float	fSliceRadius = fRadius * Asura_Maths::Sine( xSliceAngle );
		const float	fY = xCentre.y + ( fRadius * Asura_Maths::Cosine( xSliceAngle ) );

		Asura_Vector_3 xLastVertex;
		for ( u_int uSegment = 0; uSegment <= uNumSegments; uSegment++ )
		{
			const Asura_Angle xAngle = ( Asura_Maths::Deg360 * static_cast<float>( uSegment ) ) / uNumSegments;
			const float	fX = xCentre.x + ( fSliceRadius * Asura_Maths::Sine( xAngle ) );
			const float	fZ = xCentre.z + ( fSliceRadius * Asura_Maths::Cosine( xAngle ) );
			Asura_Vector_3 xThisVertex( fX, fY, fZ );

			if ( uSegment > 0 )
			{
				Asura_Render::RenderLine( xLastVertex, xThisVertex, uARGB );
			}
			xLastVertex = xThisVertex;
		}
	}

	// Add the segments to be rendered...
	for ( u_int uSegment = 0; uSegment < uNumSegments; uSegment++ )
	{
		Asura_Vector_3 xLastVertex;
		for ( u_int uSlice = 0; uSlice <= uNumSlices; uSlice++ )
		{
			const Asura_Angle xSliceAngle = ( Asura_Maths::Deg180 * static_cast<float>( uSlice ) ) / uNumSlices;
			const float	fSliceRadius = fRadius * Asura_Maths::Sine( xSliceAngle );
			const Asura_Angle xAngle = ( Asura_Maths::Deg360 * static_cast<float>( uSegment ) ) / uNumSegments;

			const float	fX = xCentre.x + ( fSliceRadius * Asura_Maths::Sine( xAngle ) );
			const float	fY = xCentre.y + ( fRadius * Asura_Maths::Cosine( xSliceAngle ) );
			const float	fZ = xCentre.z + ( fSliceRadius * Asura_Maths::Cosine( xAngle ) );
			Asura_Vector_3 xThisVertex( fX, fY, fZ );

			if ( uSlice > 0 )
			{
				Asura_Render::RenderLine( xLastVertex, xThisVertex, uARGB );
			}
			xLastVertex = xThisVertex;
		}
	}

	return true;
}

bool Asura_DebugOverlayData::RenderFlatLine( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xStart;
	if ( ReadVector( xStart ) == false )
	{
		Asura_Assert( false, "RenderLine failed to read start vector" );
		return false;
	}
	Asura_Vector_3 xEnd;
	if ( ReadVector( xEnd ) == false )
	{
		Asura_Assert( false, "RenderLine failed to read end vector" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderLine failed to read ARGB value" );
		return false;
	}
	Asura_Render::RenderLine( xStart, xEnd, uARGB );

	return true;
}

bool Asura_DebugOverlayData::RenderArrow( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xStart;
	if ( ReadVector( xStart ) == false )
	{
		Asura_Assert( false, "RenderArrow failed to read start vector" );
		return false;
	}
	Asura_Vector_3 xEnd;
	if ( ReadVector( xEnd ) == false )
	{
		Asura_Assert( false, "RenderArrow failed to read end vector" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderArrow failed to read ARGB value" );
		return false;
	}
	float fHeadProportion;
	if ( ReadFloat( fHeadProportion ) == false )
	{
		Asura_Assert( false, "RenderArrow failed to read head proportion value" );
		return false;
	}
	float fCrossSize;
	if ( ReadFloat( fCrossSize ) == false )
	{
		Asura_Assert( false, "RenderArrow failed to read cross size value" );
		return false;
	}

	Asura_Vector_3 xDirection( xEnd - xStart );
	const float fLength	= xDirection.Normalise();

	Asura_Vector_3 xHeadStart;
	Asura_Vector_3 xXAxis;
	Asura_Vector_3 xYAxis;

	// Calculate which axis our arrow-head is on...
	xHeadStart.Lerp( xEnd, xStart, fHeadProportion );

	if ( Asura_Maths::Abs( xDirection.GetY() ) > 0.707f )
	{
		xXAxis	= xDirection ^ Asura_Vector_3( 0.0f, 0.0f, 1.0f );
		xYAxis	= xXAxis ^ xDirection;
	}
	else
	{
		xXAxis	= xDirection ^ Asura_Vector_3( 0.0f, 1.0f, 0.0f );
		xYAxis	= xXAxis ^ xDirection;
	}

	// Now add the lines to be rendered...
	if ( fCrossSize > 0.0f )
	{
		Asura_Render::RenderLine( xStart - xDirection * fCrossSize, xEnd, uARGB );

		const Asura_Vector_3 xXHalfCross = xXAxis * fCrossSize;
		Asura_Render::RenderLine( xStart - xXHalfCross, xStart + xXHalfCross, uARGB );
		const Asura_Vector_3 xYHalfCross = xYAxis * fCrossSize;
		Asura_Render::RenderLine( xStart - xYHalfCross, xStart + xYHalfCross, uARGB );
	}
	else
	{
		Asura_Render::RenderLine(xStart, xEnd, uARGB);
	}

	const float fHeadSize = fLength * fHeadProportion;
	const Asura_Vector_3 xXHalfHead = xXAxis * fHeadSize;
	Asura_Render::RenderLine( xHeadStart + xXHalfHead, xEnd, uARGB );
	Asura_Render::RenderLine( xHeadStart - xXHalfHead, xEnd, uARGB );

	const Asura_Vector_3 xYHalfHead = xXAxis * fHeadSize;
	Asura_Render::RenderLine( xHeadStart + xYHalfHead, xEnd, uARGB );
	Asura_Render::RenderLine( xHeadStart - xYHalfHead, xEnd, uARGB );

	return true;
}

bool Asura_DebugOverlayData::RenderCross( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xPosition;
	if ( ReadVector( xPosition ) == false )
	{
		Asura_Assert( false, "RenderCross failed to read position vector" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderCross failed to read ARGB value" );
		return false;
	}
	float fSize;
	if ( ReadFloat( fSize ) == false )
	{
		Asura_Assert( false, "RenderCross failed to read size value" );
		return false;
	}
	const Asura_Vector_3 xXAxis( fSize, 0.0f,  0.0f );
	const Asura_Vector_3 xYAxis( 0.0f,  fSize, 0.0f );
	const Asura_Vector_3 xZAxis( 0.0f,  0.0f,  fSize );

	// Now add the lines to be rendered...
	Asura_Render::RenderLine( Asura_Vector_3( xPosition + xXAxis ), Asura_Vector_3( xPosition - xXAxis ), uARGB );
	Asura_Render::RenderLine( Asura_Vector_3( xPosition + xYAxis ), Asura_Vector_3( xPosition - xYAxis ), uARGB );
	Asura_Render::RenderLine( Asura_Vector_3( xPosition + xZAxis ), Asura_Vector_3( xPosition - xZAxis ), uARGB );

	return true;
}

bool Asura_DebugOverlayData::RenderOBB( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xCentre;
	if ( ReadVector( xCentre ) == false )
	{
		Asura_Assert( false, "RenderOBB failed to read centre vector" );
		return false;
	}
	Asura_Quat xQuat;
	if ( ReadQuat( xQuat ) == false )
	{
		Asura_Assert( false, "RenderOBB failed to read orient quaternion" );
		return false;
	}
	Asura_Vector_3 xHalfExtents;
	if ( ReadVector( xHalfExtents ) == false )
	{
		Asura_Assert( false, "RenderOBB failed to read half extents vector" );
		return false;
	}

	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderOBB failed to read ARGB value" );
		return false;
	}

	const float	fX = xHalfExtents.x;
	const float	fY = xHalfExtents.y;
	const float	fZ = xHalfExtents.z;
	Asura_AMatrix3x3 xOrientation;
	xQuat.ToMatrix( xOrientation );
	const Asura_AVector3 xCentreA( xCentre );

	Asura_AVector3 axVertices[8];

	// Generate the vertices...
	for( u_int u = 0; u < 8; u++ )
	{
		const Asura_AVector3 xVertex( ( ( ( u & 1 ) == 0 ) ? fX : -fX ), ( ( ( u & 2 ) == 0 ) ? fY : -fY ), (  ( ( u & 4 ) == 0 ) ? fZ : -fZ ) );
		axVertices[u] = ( xVertex * xOrientation ) + xCentreA;
	}

	// Now add the lines to be rendered...
	Asura_Render::RenderLine( axVertices [0], axVertices [1], uARGB );
	Asura_Render::RenderLine( axVertices [1], axVertices [3], uARGB );
	Asura_Render::RenderLine( axVertices [3], axVertices [2], uARGB );
	Asura_Render::RenderLine( axVertices [2], axVertices [0], uARGB );

	Asura_Render::RenderLine( axVertices [4], axVertices [5], uARGB );
	Asura_Render::RenderLine( axVertices [5], axVertices [7], uARGB );
	Asura_Render::RenderLine( axVertices [7], axVertices [6], uARGB );
	Asura_Render::RenderLine( axVertices [6], axVertices [4], uARGB );

	Asura_Render::RenderLine( axVertices [0], axVertices [4], uARGB );
	Asura_Render::RenderLine( axVertices [1], axVertices [5], uARGB );
	Asura_Render::RenderLine( axVertices [2], axVertices [6], uARGB );
	Asura_Render::RenderLine( axVertices [3], axVertices [7], uARGB );

	return true;
}

bool Asura_DebugOverlayData::RenderAxis( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xPosition;
	if ( ReadVector( xPosition ) == false )
	{
		Asura_Assert( false, "RenderAxis failed to read position vector" );
		return false;
	}
	Asura_Quat xQuat;
	if ( ReadQuat( xQuat ) == false )
	{
		Asura_Assert( false, "RenderAxis failed to read orient quaternion" );
		return false;
	}
	float fSize;
	if ( ReadFloat( fSize ) == false )
	{
		Asura_Assert( false, "RenderAxis failed to read size value" );
		return false;
	}
	float fAlpha;
	if ( ReadFloat( fAlpha ) == false )
	{
		Asura_Assert( false, "RenderAxis failed to read alpha value" );
		return false;
	}
	Asura_Matrix_3x3 xOrient;
	xQuat.ToMatrix( xOrient );
	Asura_Render::RenderAxes( xPosition, xOrient, fSize, fAlpha );

	return true;
}

bool Asura_DebugOverlayData::RenderBoundingBox( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	Asura_Vector_3 xMin;
	if ( ReadVector( xMin ) == false )
	{
		Asura_Assert( false, "RenderBoundingBox failed to read min vector" );
		return false;
	}
	Asura_Vector_3 xMax;
	if ( ReadVector( xMax ) == false )
	{
		Asura_Assert( false, "RenderBoundingBox failed to read max vector" );
		return false;
	}
	Asura_Vector_3 xPosition;
	if ( ReadVector( xPosition ) == false )
	{
		Asura_Assert( false, "RenderBoundingBox failed to read position vector" );
		return false;
	}
	Asura_Quat xQuat;
	if ( ReadQuat( xQuat ) == false )
	{
		Asura_Assert( false, "RenderBoundingBox failed to read orient quaternion" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderBoundingBox failed to read ARGB value" );
		return false;
	}

	const u_short ausIndices[24] =
	{
		0, 2, 2, 3, 3, 1, 1, 0,
		0, 4, 2, 6, 3, 7, 1, 5,
		4, 6, 6, 7, 7, 5, 5, 4,
	};

	Asura_Vector_3 axVertices[8];

	Asura_Bounding_Box xBB;
	xBB.MinX = xMin.x;
	xBB.MinY = xMin.y;
	xBB.MinZ = xMin.z;
	xBB.MaxX = xMax.x;
	xBB.MaxY = xMax.y;
	xBB.MaxZ = xMax.z;

	Asura_Matrix_3x3 xOrientation;
	xQuat.ToMatrix( xOrientation );

	// Calculate the vertices...
	for ( u_int u = 0; u < 8; u++ )
	{
		axVertices[u][0] = ( ( u & 1 ) == 0 ) ? xBB.MinX : xBB.MaxX;
		axVertices[u][1] = ( ( u & 2 ) == 0 ) ? xBB.MinY : xBB.MaxY;
		axVertices[u][2] = ( ( u & 4 ) == 0 ) ? xBB.MinZ : xBB.MaxZ;

		axVertices[u] = ( axVertices[u] * xOrientation ) + xPosition;
	}

	// Add them to be rendered...
	for ( u_int u = 0; u < 24; u += 2 )
	{
		const Asura_Vector_3& xStart = axVertices[ausIndices[u]];
		const Asura_Vector_3& xEnd = axVertices[ausIndices[u + 1]];

		Asura_Render::RenderLine( xStart, xEnd, uARGB );
	}

	return true;
}

bool Asura_DebugOverlayData::RenderFrustum( const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	float fNearPlane;
	if ( ReadFloat( fNearPlane ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read near plane value" );
		return false;
	}
	float fFarPlane;
	if ( ReadFloat( fFarPlane ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read far plane value" );
		return false;
	}
	float fTanHorizontalFOV;
	if ( ReadFloat( fTanHorizontalFOV ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read tan horizontal FOV" );
		return false;
	}
	float fTanVerticalFOV;
	if ( ReadFloat( fTanVerticalFOV ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read tan vertical FOV" );
		return false;
	}
	Asura_Frustum xFrustum;
	xFrustum.SetForFOV( fNearPlane, fFarPlane, fTanHorizontalFOV, fTanVerticalFOV );

	Asura_Vector_3 xPosition;
	if ( ReadVector( xPosition ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read position vector" );
		return false;
	}
	Asura_Quat xQuat;
	if ( ReadQuat( xQuat ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read orient quaternion" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderFrustum failed to read ARGB value" );
		return false;
	}

	Asura_Vector_3 axVertices[Asura_Frustum::MAX_VERTICES];

	// Generate the vertices...
	xFrustum.GenerateVertices( axVertices, xPosition, xQuat );

	// Now add the lines to be rendered...
	Asura_Render::RenderLine( axVertices[0], axVertices[1], uARGB );
	Asura_Render::RenderLine( axVertices[1], axVertices[3], uARGB );
	Asura_Render::RenderLine( axVertices[3], axVertices[2], uARGB );
	Asura_Render::RenderLine( axVertices[2], axVertices[0], uARGB );

	Asura_Render::RenderLine( axVertices[4], axVertices[5], uARGB );
	Asura_Render::RenderLine( axVertices[5], axVertices[7], uARGB );
	Asura_Render::RenderLine( axVertices[7], axVertices[6], uARGB );
	Asura_Render::RenderLine( axVertices[6], axVertices[4], uARGB );

	Asura_Render::RenderLine( axVertices[0], axVertices[4], uARGB );
	Asura_Render::RenderLine( axVertices[1], axVertices[5], uARGB );
	Asura_Render::RenderLine( axVertices[2], axVertices[6], uARGB );
	Asura_Render::RenderLine( axVertices[3], axVertices[7], uARGB );

	return true;
}

bool Asura_DebugOverlayData::RenderConsoleString( const u_int uNumBytes )
{
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderConsoleString failed to read ARGB value" );
		return false;
	}
	const u_int uStringLen = uNumBytes - s_auPrimMemSize[ASURA_DO_CONSOLE_STRING];
	const char* szString = NULL;
	if ( ReadString( szString, uStringLen ) == false )
	{
		Asura_Assert( false, "RenderConsoleString failed to read string data" );
		return false;
	}

	RenderStringScreenInternal( szString, uARGB, s_xConsoleCurrentPos );
	s_xConsoleCurrentPos += s_xConsoleLineSize;

	return true;
}

bool Asura_DebugOverlayData::RenderDebugString( const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes )
{
	Asura_Vector_3 xPos3D;
	if ( ReadVector( xPos3D ) == false )
	{
		Asura_Assert( false, "RenderDebugString failed to read pos value" );
		return false;
	}
	u_int uARGB;
	if ( ReadARGB( uARGB ) == false )
	{
		Asura_Assert( false, "RenderDebugString failed to read ARGB value" );
		return false;
	}
	const u_int uStringLen = uNumBytes - s_auPrimMemSize[ASURA_DO_DEBUG_STRING];
	const char* szString = NULL;
	if ( ReadString( szString, uStringLen ) == false )
	{
		Asura_Assert( false, "RenderDebugString failed to read string data" );
		return false;
	}

	if ( bWorldSpace == false )
	{
		const Asura_Vector_2 xPos2D( xPos3D.x, xPos3D.y );
		RenderStringScreenInternal( szString, uARGB, xPos2D );
	}
	else
	{
		RenderStringWorldInternal( szString, uARGB, xPos3D );
	}

	return true;
}

// Render text using the vector font data stored in Asura_DebugOverlay_Alphabet.h
void Asura_DebugOverlayData::RenderStringScreenInternal( const char* const szString, const u_int uARGB, const Asura_Vector_2& xPos )
{
	if ( szString == NULL )
	{
		return;
	}

	const float fFontWidth = Asura_DebugOverlay::GetUseVectorFont() ? 
		(Asura_DebugOverlay::GetVectorFontWidth() / Asura_Viewport::GetWidth()) : ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH;
	const float fFontHeight = Asura_DebugOverlay::GetUseVectorFont() ? 
		(fFontWidth / Asura_DebugOverlay::GetVectorFontWHRatio()) : ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
	const Asura_Vector_2 xScale( fFontWidth, fFontHeight );

	RenderStringInternal( xPos, xScale, szString, uARGB );
}


// Render text using the vector font data stored in Asura_DebugOverlay_Alphabet.h
void Asura_DebugOverlayData::RenderStringWorldInternal( const char* const szString, const u_int uARGB, const Asura_Vector_3& xPos )
{
	if ( szString == NULL )
	{
		return;
	}

	const float fFontWidth = Asura_DebugOverlay::GetUseVectorFont() ? 
		(Asura_DebugOverlay::GetVectorFontWidth() / Asura_Viewport::GetWidth()) : ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH;
	const float fFontHeight = Asura_DebugOverlay::GetUseVectorFont() ? 
		(fFontWidth / Asura_DebugOverlay::GetVectorFontWHRatio()) : ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;

	// Made a world vector to represent the screen x-vector
	Asura_Vector_3 xScreenPos = xPos;
	Asura_Camera::ProjectWorldCoordsToScreenSpace( xScreenPos );
	if ( xScreenPos.z <= 0.0f ) // behind the camera
	{
		return;
	}

	const float fDelta = 1.0e-3f;

	const Asura_Vector_3 xWorldX = xPos + Asura_Vector_3( fDelta, 0.0f, 0.0f );
	Asura_Vector_3 xScreenX = xWorldX;
	Asura_Camera::ProjectWorldCoordsToScreenSpace( xScreenX );
	xScreenX -= xScreenPos;

	const Asura_Vector_3 xWorldY = xPos + Asura_Vector_3( 0.0f, fDelta, 0.0f );
	Asura_Vector_3 xScreenY = xWorldY;
	Asura_Camera::ProjectWorldCoordsToScreenSpace( xScreenY );
	xScreenY -= xScreenPos;

	const Asura_Vector_3 xWorldZ = xPos + Asura_Vector_3( 0.0f, 0.0f, fDelta );
	Asura_Vector_3 xScreenZ = xWorldZ;
	Asura_Camera::ProjectWorldCoordsToScreenSpace( xScreenZ );
	xScreenZ -= xScreenPos;

	const Asura_Vector_3 xScreenXdir( xScreenX.GetX(), xScreenY.GetX(), xScreenZ.GetX() );
	const Asura_Vector_3 xScreenYdir( xScreenX.GetY(), xScreenY.GetY(), xScreenZ.GetY() );

	// This scales the text
	const float fScale = 6.0f;
	const float fXscale = fScale * ( xScreenXdir.Magnitude() / fDelta );
	const float fYscale = fScale * ( xScreenYdir.Magnitude() / fDelta );

	const Asura_Vector_2 xScale( fFontWidth * fXscale, fFontHeight * fYscale );

	const Asura_Vector_2 xCurrentPos( xScreenPos.x, xScreenPos.y );
	RenderStringInternal( xCurrentPos, xScale, szString, uARGB );
}

void Asura_DebugOverlayData::RenderStringInternal( const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const char* const szString, const u_int uARGB )
{
	if (Asura_DebugOverlay::GetUseVectorFont())
	{
		const u_int uLen = strlen( szString );
		Asura_Vector_2 xCurrentPos( xPos.x, xPos.y );

		for ( u_int u = 0; u < uLen; u++ )
		{
			unsigned char ucChar = szString[u];
			if ( ucChar > 127 ) 
			{
				ucChar = 0;
			}

			const Asura_DebugOverlay_Alphabet::Asura_AlphaCharacter& xChar = Asura_DebugOverlay_Alphabet::pxChars[ucChar];
			if ( xChar.m_uNumCoords1 > 0 )
			{
				RenderLineStripInternal( xCurrentPos, xScale, xChar.m_pfCoords1, xChar.m_uNumCoords1, uARGB );
			}
			if ( xChar.m_uNumCoords2 > 0 )
			{
				RenderLineStripInternal( xCurrentPos, xScale, xChar.m_pfCoords2, xChar.m_uNumCoords2, uARGB );
			}

			xCurrentPos.x += xScale.x;
		}
	}
	else
	{
		const u_int uScreenWidth	= Asura_Viewport::GetWidth();
		const u_int uScreenHeight	= Asura_Viewport::GetHeight();
		const float	fScreenWidth	= static_cast<float>(uScreenWidth);
		const float	fScreenHeight	= static_cast<float>(uScreenHeight);
		const int iX = Asura_Maths::FloatToInt(fScreenWidth * xPos.x / ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
		const int iY = Asura_Maths::FloatToInt(fScreenHeight * xPos.y / ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
		Asura_DebugText::AddTextBlock(iX, iY, uARGB, 0, szString);
	}
}

void Asura_DebugOverlayData::RenderLineStripInternal( const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const float* const pfCoords, const u_int uNumPoints, const u_int uARGB )
{
	const u_int uScreenWidth	= Asura_Viewport::GetWidth();
	const u_int uScreenHeight	= Asura_Viewport::GetHeight();
	const float	fScreenWidth	= static_cast<float>(uScreenWidth);
	const float	fScreenHeight	= static_cast<float>(uScreenHeight);

	// The input position is in 0->1 screen dimensions
	// The input scale is in 0->1 screen dimensions
	const float* pfXValues = pfCoords;
	const float* pfYValues = pfCoords + 1;
	const u_int uNumLines = uNumPoints - 1;

	const Asura_Vector_2 xScreenScale( fScreenWidth, fScreenHeight );
	const Asura_Vector_2 xPointScale( xScreenScale.x * xScale.x, xScreenScale.y * xScale.y );

	const Asura_Vector_2 xOrigin( xPos.x * xScreenScale.x, xPos.y * xScreenScale.y );
	const Asura_Vector_2 xFirstPoint( pfXValues[0], pfYValues[0] );
	pfXValues += 2;
	pfYValues += 2;

	Asura_Vector_2 xFrom = xOrigin + Asura_Vector_2( xFirstPoint.x * xPointScale.x, xFirstPoint.y * xPointScale.y );

	for ( u_int u = 0; u < uNumLines; u++ )
	{
		const Asura_Vector_2 xPoint( pfXValues[0], pfYValues[0] );
		pfXValues += 2;
		pfYValues += 2;

		const Asura_Vector_2 xTo = xOrigin + Asura_Vector_2( xPoint.x * xPointScale.x, xPoint.y * xPointScale.y );

		Asura_Render::Render2DLine( xFrom, xTo, uARGB );
		xFrom = xTo;
	}
}

// is there enough memory free in the list to add this type of primitive
bool Asura_DebugOverlayData::CanAdd( const enum ASURA_DO_COMMAND eCommand )
{
	const int iMemSize = s_auPrimMemSize[ eCommand ];
	if ( s_pucListNext + iMemSize > s_pucListEnd )
	{
		return false;
	}
	return true;
}

// Use memory copy to avoid any alignment issues
bool Asura_DebugOverlayData::AddData( const void* const pData, const size_t uNumBytes )
{
	const u_char* const pucNewListNext = s_pucListNext + uNumBytes;
	if ( pucNewListNext > s_pucListEnd )
	{
		return false;
	}	
	Asura_Mem::Copy( s_pucListNext, pData, uNumBytes );
	s_pucListNext += uNumBytes;
	return true;
}
bool Asura_DebugOverlayData::AddFloat( const float fValue )
{
	return AddData( &fValue, sizeof( float ) );
}

bool Asura_DebugOverlayData::AddUint( const u_int uValue )
{
	return AddData( &uValue, sizeof( u_int ) );
}

bool Asura_DebugOverlayData::AddVector( const Asura_Vector_3& xVector )
{
	if ( AddFloat( xVector.x ) == false ) 
	{
		return false;
	}
	if ( AddFloat( xVector.y ) == false )
	{
		return false;
	}
	if ( AddFloat( xVector.z ) == false )
	{
		return false;
	}
	return true;
}

bool Asura_DebugOverlayData::AddQuat( const Asura_Quat& xQuat )
{
	if ( AddFloat( xQuat.x ) == false ) 
	{
		return false;
	}
	if ( AddFloat( xQuat.y ) == false )
	{
		return false;
	}
	if ( AddFloat( xQuat.z ) == false )
	{
		return false;
	}
	if ( AddFloat( xQuat.w ) == false )
	{
		return false;
	}
	return true;
}

bool Asura_DebugOverlayData::AddRGBA( const Asura_RGBA& xColour )
{
	const u_int uColour = xColour.ToUIntARGB( true );
	return AddARGB( uColour );
}

bool Asura_DebugOverlayData::AddARGB( const u_int uARGB )
{
	return AddData( &uARGB, sizeof( u_int ) );
}

bool Asura_DebugOverlayData::AddString( const char* const szString, const u_int uNumBytes )
{
	return AddData( szString, uNumBytes );
}

// For simple fixed sized primitives
bool Asura_DebugOverlayData::AddHeader( const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer )
{
	// The ASURA_DO_CONSOLE_STRING & ASURA_DO_DEBUG_STRING commands are variable length so have to use the other AddHeader function
	Asura_Assertf( ( ( eCommand != ASURA_DO_CONSOLE_STRING ) && ( eCommand != ASURA_DO_DEBUG_STRING ) ), ( "AddHeader wrong eCommand %d", eCommand ) );
	const u_int uNumBytes = s_auPrimMemSize[eCommand];
	return AddHeader( eCommand, bWorldSpace, bIgnoreZBuffer, uNumBytes );
}

// For primitives with variable size e.g. text based ones
bool Asura_DebugOverlayData::AddHeader( const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes )
{
	const u_char* const pucNewListNext = s_pucListNext + uNumBytes;
	if ( pucNewListNext > s_pucListEnd )
	{
		Asura_Assertf( false, ( "AddHeader not enough memory adding eCommand %d uNumBytes %d", eCommand, uNumBytes ) );
		return false;
	}	
	// includes header size
	if ( AddUint( uNumBytes ) == false )
	{
		Asura_Assertf( false, ( "AddHeader failed to add command size eCommand %d uNumBytes %d", eCommand, uNumBytes ) );
		return false;
	}
	u_int uHeader = ( static_cast<u_int>( eCommand ) & ASURA_DO_COMMAND_BIT_MASK );
	uHeader |= ( bWorldSpace ? ASURA_DO_WORLD_SPACE_BIT_MASK : 0 );
	uHeader |= ( bIgnoreZBuffer ? ASURA_DO_IGNORE_Z_BUFFER_BIT_MASK : 0 );
	if ( AddUint( uHeader ) == false )
	{
		Asura_Assertf( false, ( "AddHeader failed to add header eCommand %d uNumBytes %d", eCommand, uNumBytes ) );
		return false;
	}
	return true;
}

// Use memory copy to avoid any alignment issues
bool Asura_DebugOverlayData::ReadData( void* const pData, const size_t uNumBytes )
{
	const u_char* const pucNewListNext = s_pucListNext + uNumBytes;
	if ( pucNewListNext > s_pucListEnd )
	{
		return false;
	}	
	Asura_Mem::Copy( pData, s_pucListNext, uNumBytes );
	s_pucListNext += uNumBytes;
	return true;
}
bool Asura_DebugOverlayData::ReadFloat( float& fValue )
{
	return ReadData( &fValue, sizeof( float ) );
}

bool Asura_DebugOverlayData::ReadUint( u_int& uValue )
{
	return ReadData( &uValue, sizeof( u_int ) );
}

bool Asura_DebugOverlayData::ReadVector( Asura_Vector_3& xVector )
{
	if ( ReadFloat( xVector.x ) == false ) 
	{
		return false;
	}
	if ( ReadFloat( xVector.y ) == false )
	{
		return false;
	}
	if ( ReadFloat( xVector.z ) == false )
	{
		return false;
	}
	return true;
}

bool Asura_DebugOverlayData::ReadQuat( Asura_Quat& xQuat )
{
	if ( ReadFloat( xQuat.x ) == false ) 
	{
		return false;
	}
	if ( ReadFloat( xQuat.y ) == false )
	{
		return false;
	}
	if ( ReadFloat( xQuat.z ) == false )
	{
		return false;
	}
	if ( ReadFloat( xQuat.w ) == false )
	{
		return false;
	}
	return true;
}

bool Asura_DebugOverlayData::ReadRGBA( Asura_RGBA& xColour )
{
	u_int uColour = 0;
	if ( ReadARGB( uColour ) == false )
	{
		return false;
	}

	xColour.FromUIntARGB( uColour );
	return true;
}

bool Asura_DebugOverlayData::ReadARGB( u_int& uARGB )
{
	return ReadData( &uARGB, sizeof( u_int ) );
}

bool Asura_DebugOverlayData::ReadString( const char*& szString, const u_int uNumBytes )
{
	const u_char* const pucNewListNext = s_pucListNext + uNumBytes;
	if ( pucNewListNext > s_pucListEnd )
	{
		return false;
	}	

	szString = reinterpret_cast<char*>( s_pucListNext );
	s_pucListNext += uNumBytes;
	return true;
}

// Extract data from the command list
bool Asura_DebugOverlayData::ReadHeader( u_int& uNumBytes, enum ASURA_DO_COMMAND& eCommand, bool& bWorldSpace, bool& bIgnoreZBuffer )
{
	if ( ReadUint( uNumBytes ) == false )
	{
		Asura_Assert( false, "ReadHeader failed reading command size" );
		return false;
	}
	u_int uHeader = 0;
	if ( ReadUint( uHeader ) == false )
	{
		Asura_Assert( false, "ReadHeader failed reading header" );
		return false;
	}
	eCommand = static_cast<ASURA_DO_COMMAND>( uHeader & ASURA_DO_COMMAND_BIT_MASK );
	bWorldSpace = ( uHeader & ASURA_DO_WORLD_SPACE_BIT_MASK ) ? true : false;
	bIgnoreZBuffer = ( uHeader & ASURA_DO_IGNORE_Z_BUFFER_BIT_MASK ) ? true : false;

	return true;
}

// Register a debug mode
bool Asura_DebugOverlayData::RegisterMode( const char* const szName, const int iMode, AsuraDebugOverlayCallbackFn pCallbackFn, void* const pxUserData )
{
	for ( int i = 0; i < ASURA_DEBUGOVERLAY_MAX_NUM_MODES; i++ )
	{
		const int iFoundMode = s_xModes[i].m_iMode;
		if ( iFoundMode == ASURA_DO_RESERVED )
		{
			s_xModes[i].m_iMode = iMode;
			s_xModes[i].m_pDoDebugFunc = pCallbackFn;
			s_xModes[i].m_pxUserData = pxUserData;
			s_xModes[i].m_xName = szName;
			SortModeList();
			return true;
		}
	}
	return false;
}

void Asura_DebugOverlayData::SortModeList( void )
{
	// Could do an insertion sort because modes are just added one after another but using bubble sort copes with deleting modes as well
	// Very simple bubble sort
	for ( int i = 0; i < ( ASURA_DEBUGOVERLAY_MAX_NUM_MODES - 1 ); i++ )
	{
		for ( int j = i + 1; j < ASURA_DEBUGOVERLAY_MAX_NUM_MODES; j++ )
		{
			const int iMode = s_xModes[i].m_iMode;
			const int jMode = s_xModes[j].m_iMode;

			if ( ( iMode == ASURA_DO_RESERVED ) || ( jMode < iMode ) )
			{
				if ( jMode != ASURA_DO_RESERVED )
				{
					// Swap j & i around
					Asura_DebugOverlayMode xTemp = s_xModes[i];
					s_xModes[i] = s_xModes[j];
					s_xModes[j] = xTemp;
				}
			}
		}
	}

	// Count how many modes there are
	s_iNumModes = 0;
	for ( int i = 0; i < ASURA_DEBUGOVERLAY_MAX_NUM_MODES; i++ )
	{
		const int iMode = s_xModes[i].m_iMode;
		if ( iMode != ASURA_DO_RESERVED )
		{
			// Because this is a sorted array the number of clients must equal the current index
			Asura_Assert( i == s_iNumModes, "" );
			s_iNumModes++;
		}
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Command console functions
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#define CONSOLE_CMDNAME_ACTIVATE			L"Debug.DebugOverlay.Activate"
#define CONSOLE_CMDNAME_SELECTMODE			L"Debug.DebugOverlay.SelectMode"
#define CONSOLE_CMDNAME_PREVMODE			L"Debug.DebugOverlay.PrevMode"
#define CONSOLE_CMDNAME_NEXTMODE			L"Debug.DebugOverlay.NextMode"
#define CONSOLE_VARNAME_USE_VECTOR_FONT 	L"Debug.DebugOverlay.UseVectorFont"
#define CONSOLE_VARNAME_VECTOR_FONT_WIDTH	L"Debug.DebugOverlay.VectorFontWidth"
#define CONSOLE_VARNAME_VECTOR_FONT_WHRATIO	L"Debug.DebugOverlay.VectorFontWHRatio"

static bool ConsoleCmd_Activate( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_ACTIVATE L": activate/de-activate the debug overlay system" );
	}
	else
	{
		Asura_DebugOverlay::SetActive( Asura_DebugOverlay::IsActive() ^ true );
	}
	return true;
}

static bool ConsoleCmd_SelectMode( int iMode, bool bInterrogate, bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_SELECTMODE L": select a specific debug overlay mode (integer)" );
	}
	else if (bInterrogate)
	{
		const int iCurrentMode = Asura_DebugOverlay::GetCurrentMode();
		int iModeIndex = Asura_DebugOverlayData::FindModeIndex( iCurrentMode );
		if ( iModeIndex >= 0 )
		{
			Asura_CommandConsole::Printf( L"    current mode is %d '%s'", iCurrentMode, Asura_DebugOverlayData::s_xModes[iModeIndex].m_xName.GetString() );
		}
	}
	else
	{
		Asura_DebugOverlay::SelectMode( iMode );
	}
	return true;
}

static bool ConsoleCmd_PrevMode( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_PREVMODE L": select the previous debug overlay mode" );
	}
	else
	{
		Asura_DebugOverlay::CycleDebugMode( -1 );
	}
	return true;
}

static bool ConsoleCmd_NextMode( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_NEXTMODE L": select the next debug overlay mode" );
	}
	else
	{
		Asura_DebugOverlay::CycleDebugMode( +1 );
	}
	return true;
}

#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE

////////////////////////////////////////////////////////////////////////////////////////////////
//
// Internal data & functions for Asura_DebugOverlay
//
////////////////////////////////////////////////////////////////////////////////////////////////

int Asura_DebugOverlay::s_iCurrentMode = ASURA_DO_NONE;
bool Asura_DebugOverlay::s_bActive = false;
bool Asura_DebugOverlay::s_bUseVectorFont = true;
float Asura_DebugOverlay::s_fVectorFontWidth = 12.0f;
float Asura_DebugOverlay::s_fVectorFontWHRatio = 0.5f;

// Initialise internals - allocate memory buffer
void Asura_DebugOverlay::Initialise( void )
{
	const int iMemorySize = ASURA_DEBUGOVERLAY_MEMORY_SIZE;
	if ( Asura_DebugOverlayData::Initialise( iMemorySize ) == false )
	{
		Asura_Assertf( false, ( "Asura_DebugOverlay failed to initialise with memory size of %dKB", iMemorySize/1024 ) );
		return;
	}
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_ACTIVATE, ConsoleCmd_Activate, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_SELECTMODE, ConsoleCmd_SelectMode, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_PREVMODE, ConsoleCmd_PrevMode, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_NEXTMODE, ConsoleCmd_NextMode, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_USE_VECTOR_FONT, &s_bUseVectorFont, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_VECTOR_FONT_WIDTH, &s_fVectorFontWidth, 1.0f, 32.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_VECTOR_FONT_WHRATIO, &s_fVectorFontWHRatio, 0.1f, 2.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
	Reset();
}

// Free memory
void Asura_DebugOverlay::Shutdown( void )
{
	Reset();
	Asura_DebugOverlayData::Shutdown();
	return;
}

// Reset internal state
void Asura_DebugOverlay::Reset( void )
{
	Asura_DebugOverlayData::Reset();
	s_iCurrentMode = ASURA_DO_NONE;
	return;
}
 
// Register a debug mode
bool Asura_DebugOverlay::RegisterMode( const char* const szName, const int iMode, AsuraDebugOverlayCallbackFn pCallbackFn, void* const pxUserData )
{
	if ( iMode == ASURA_DO_RESERVED )
	{
		Asura_Assertf( false, ( "RegisterMode '%s' ERROR iMode value %d is reserved for the debug overlay system\n", szName, iMode ) );
		return false;
	}

	const bool bReturn = Asura_DebugOverlayData::RegisterMode( szName, iMode, pCallbackFn, pxUserData );
	Asura_Assertf( bReturn, ( "RegisterMode ERROR couldn't add mode '%s' to list", szName ) );
	return bReturn;
}

// Choose a specific debug mode
bool Asura_DebugOverlay::SelectMode( const int iMode )
{
	const int iIndex = Asura_DebugOverlayData::FindModeIndex( iMode );
	if ( iIndex < 0 )
	{
		Asura_OutputToDebugger_Retail( ( "SelectMode can't find mode %d", iMode ) );
								       
		return false;
	}
	s_iCurrentMode = iMode;
	Asura_OutputToDebugger_Retail( ( "SelectMode '%s' Index %d Mode %d\n", 	
								     Asura_DebugOverlayData::s_xModes[iIndex].m_xName.GetString(), iIndex, s_iCurrentMode ) );
	return true;
}

// Change the debug mode: i.e. +1, -1
void Asura_DebugOverlay::CycleDebugMode( const int iIncrement )
{
	int iIndex = Asura_DebugOverlayData::FindModeIndex( s_iCurrentMode );
	iIndex += iIncrement;

	if ( iIndex >= Asura_DebugOverlayData::s_iNumModes )
	{
		iIndex = 0;
	}
	if ( iIndex < 0 )
	{
		iIndex = Asura_DebugOverlayData::s_iNumModes - 1;
	}
	s_iCurrentMode = Asura_DebugOverlayData::s_xModes[iIndex].m_iMode;
	Asura_OutputToDebugger_Retail( ( "CycleDebugMode %d '%s' Index %d Mode %d\n", 	
								     iIncrement, 
						   	         Asura_DebugOverlayData::s_xModes[iIndex].m_xName.GetString(), iIndex, s_iCurrentMode ) );
}

// Render the debug data - world space data, before the z-buffer clear
void Asura_DebugOverlay::Render( void )
{
	if ( s_bActive == true )
	{
		Asura_DebugOverlayData::Render();
	}

	// All done reset the list ready for the next cycle
	Asura_DebugOverlayData::ResetCommandList();

	return;
}

// Update the Asura_Debug system - calls the correct debug callback
void Asura_DebugOverlay::Update( void )
{
	// Do nothing if we aren't active
	if ( s_bActive == false )
	{
		return;
	}

	Asura_DebugOverlayData::Update();
	return;
}

bool Asura_DebugOverlay::ModeIsRegistered ( const int iMode )
{
	return ( Asura_DebugOverlayData::FindModeIndex( iMode ) != -1 );
}

// Drraw text in screenspace (not z-buffered)
void Asura_DebugOverlay::DrawTextScreen( const int iMode, const Asura_Vector_2& xPos, const u_int uARGB, const char* const szFmtText, ... )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	Asura_String2048 xBuffer;
	va_list xArgs;
	va_start( xArgs, szFmtText );
	xBuffer.VPrint( szFmtText, xArgs );

	const bool bWorldSpace = false;
	const bool bIgnoreZBuffer = true;
	const Asura_Vector_3 xPos3D( xPos.x, xPos.y, 0.0f );
	AddDebugString( bWorldSpace, bIgnoreZBuffer, xBuffer.GetString(), xPos3D, uARGB );
}

// Draw text in world space (z-buffered)
void Asura_DebugOverlay::DrawTextWorld( const int iMode, const Asura_Vector_3& xPos, const u_int uARGB, const char* const szFmtText, ... )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	Asura_String2048 xBuffer;
	va_list xArgs;
	va_start( xArgs, szFmtText );
	xBuffer.VPrint( szFmtText, xArgs );

	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = true;
	AddDebugString( bWorldSpace, bIgnoreZBuffer, xBuffer.GetString(), xPos, uARGB );
}

void Asura_DebugOverlay::DrawConsoleText( const int iMode, const char* const szFmtText, ... )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	Asura_String2048 xBuffer;
	va_list xArgs;
	va_start( xArgs, szFmtText );
	xBuffer.VPrint( szFmtText, xArgs );

	AddConsoleString( xBuffer.GetString(), Asura_DebugOverlayData::s_uDefaultConsoleColour );
}

void Asura_DebugOverlay::DrawConsoleText( const int iMode, const u_int uARGB, const char* const szFmtText, ... )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	Asura_String2048 xBuffer;

	va_list xArgs;
	va_start( xArgs, szFmtText );
	xBuffer.VPrint( szFmtText, xArgs );

	AddConsoleString( xBuffer.GetString(), uARGB );
}

void Asura_DebugOverlay::Printf( const int iMode, const char* const szFmtText, ... )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}

	Asura_String2048 xBuffer;

	va_list xArgs;
	va_start( xArgs, szFmtText );
	xBuffer.VPrint( szFmtText, xArgs );

	Asura_OutputToDebugger_Retail( ( "%s\n", xBuffer.GetString() ) );
}

// Add a sphere to the list
void Asura_DebugOverlay::AddSphere( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices )
{
	// Is there enough memory to add a sphere
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_SPHERE ) == false )
	{
		return;
	}
	
	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_SPHERE, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddSphere failed to add the header" );
		return;
	}
	if (  Asura_DebugOverlayData::AddVector( xCentre ) == false )
	{
		Asura_Assert( false, "AddSphere failed to add centre" );
		return;
	}
	if ( Asura_DebugOverlayData::AddFloat( fRadius ) == false )
	{
		Asura_Assert( false, "AddSphere failed to add radius" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddSphere failed to add ARGB" );
		return;
	}
	if ( Asura_DebugOverlayData::AddUint( uNumSegments ) == false )
	{
		Asura_Assert( false, "AddSphere failed to add num segments" );
		return;
	}
	if ( Asura_DebugOverlayData::AddUint( uNumSlices ) == false )
	{
		Asura_Assert( false, "AddSphere failed to add num slices" );
		return;
	}

	return;
}

// Add line to the list
void Asura_DebugOverlay::AddLine( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB )
{
	// Is there enough memory to add a line
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_LINE ) == false )
	{
		return;
	}
	
	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_LINE, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddLine failed to add header" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xStart ) == false )
	{
		Asura_Assert( false, "AddLine failed to add start" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xEnd ) == false )
	{
		Asura_Assert( false, "AddLine failed to add end" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddLine failed to add ARGB" );
		return;
	}

	return;
}

// Add arrow to the list
void Asura_DebugOverlay::AddArrow( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize )
{
	// Is there enough memory to add an arrow
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_ARROW ) == false )
	{
		return;
	}
	
	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_ARROW, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddArrow failed to add header" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xStart ) == false )
	{
		Asura_Assert( false, "AddArrow failed to add start" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xEnd ) == false )
	{
		Asura_Assert( false, "AddArrow failed to add end" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddArrow failed to add ARGB" );
		return;
	}
	if ( Asura_DebugOverlayData::AddFloat( fHeadProportion ) == false )
	{
		Asura_Assert( false, "AddArrow failed to add head proportion" );
		return;
	}
	if ( Asura_DebugOverlayData::AddFloat( fCrossSize ) == false )
	{
		Asura_Assert( false, "AddArrow failed to add head cross size" );
		return;
	}
}

void Asura_DebugOverlay::AddCross( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize )
{
	// Is there enough memory to add a cross
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_CROSS ) == false )
	{
		return;
	}

	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_CROSS, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddCross failed to add header" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xPosition ) == false )
	{
		Asura_Assert( false, "AddCross failed to add position" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddLine failed to add ARGB" );
		return;
	}
	if ( Asura_DebugOverlayData::AddFloat( fSize ) == false )
	{
		Asura_Assert( false, "AddLine failed to add size" );
		return;
	}

	return;
}

void Asura_DebugOverlay::AddOBB( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_OBB& xOBB, const u_int uARGB )
{
	// Is there enough memory to add an OBB
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_OBB ) == false )
	{
		return;
	}

	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_OBB, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddOBB failed to add header" );
		return;
	}
	const Asura_AVector3 xCentreA = xOBB.GetCentre();
	const Asura_AMatrix3x3 xOrientA = xOBB.GetOrientation();
	const Asura_AVector3 xHalfExtentsA = xOBB.GetHalfExtents();

	const Asura_Vector_3 xCentre( xCentreA );
	if ( Asura_DebugOverlayData::AddVector( xCentre ) == false )
	{
		Asura_Assert( false, "AddOBB failed to add centre" );
		return;
	}
	Asura_Quat xQuat;
	xQuat.FromMatrix( xOrientA );
	if ( Asura_DebugOverlayData::AddQuat( xQuat ) == false )
	{
		Asura_Assert( false, "AddOBB failed to add orientation" );
		return;
	}
	const Asura_Vector_3 xHalfExtents( xHalfExtentsA );
	if ( Asura_DebugOverlayData::AddVector( xHalfExtents ) == false )
	{
		Asura_Assert( false, "AddOBB failed to add half extents" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddOBB failed to add ARGB" );
		return;
	}

	return;
}

void Asura_DebugOverlay::AddAxis( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha )
{
	// Is there enough memory to add an axis
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_AXIS ) == false )
	{
		return;
	}

	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_AXIS, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddAxis failed to add header" );
	}
	if ( Asura_DebugOverlayData::AddVector( xPosition ) == false )
	{
		Asura_Assert( false, "AddAxis failed to add position" );
		return;
	}
	if ( Asura_DebugOverlayData::AddQuat( xOrientation ) == false )
	{
		Asura_Assert( false, "AddAxis failed to add orientation" );
		return;
	}
	if ( Asura_DebugOverlayData::AddFloat( fSize ) == false )
	{
		Asura_Assert( false, "AddAxis failed to add size" );
		return;
	}
	if ( Asura_DebugOverlayData::AddFloat( fAlpha ) == false )
	{
		Asura_Assert( false, "AddAxis failed to add alpha" );
		return;
	}

	return;
}

void Asura_DebugOverlay::AddBoundingBox( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB )
{
	// Is there enough memory to add a bounding box
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_BOUNDING_BOX ) == false )
	{
		return;
	}

	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_BOUNDING_BOX, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddBoundingBox failed to add header" );
	}
	const Asura_Vector_3 xMin = xBB.GetMin();
	if ( Asura_DebugOverlayData::AddVector( xMin ) == false )
	{
		Asura_Assert( false, "AddBoundingBox failed to add min" );
		return;
	}
	const Asura_Vector_3 xMax = xBB.GetMax();
	if ( Asura_DebugOverlayData::AddVector( xMax ) == false )
	{
		Asura_Assert( false, "AddBoundingBox failed to add max" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xPosition ) == false )
	{
		Asura_Assert( false, "AddBoundingBox failed to add position" );
		return;
	}
	if ( Asura_DebugOverlayData::AddQuat( xOrientation ) == false )
	{
		Asura_Assert( false, "AddBoundingBox failed to add orientation" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddBoundingBox failed to add ARGB" );
		return;
	}

	return;
}

void Asura_DebugOverlay::AddFrustum( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB )
{
	// Is there enough memory to add a frustum
	if ( Asura_DebugOverlayData::CanAdd( Asura_DebugOverlayData::ASURA_DO_FRUSTUM ) == false )
	{
		return;
	}

	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_FRUSTUM, bWorldSpace, bIgnoreZBuffer ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add header" );
	}
	const float fNearPlane = xFrustum.GetNearPlane();
	if ( Asura_DebugOverlayData::AddFloat( fNearPlane ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add near plane" );
		return;
	}
	const float fFarPlane = xFrustum.GetFarPlane();
	if ( Asura_DebugOverlayData::AddFloat( fFarPlane ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add far plane" );
		return;
	}
	const float fTanHorizontalFOV = xFrustum.GetTanHorizontalFOV();
	if ( Asura_DebugOverlayData::AddFloat( fTanHorizontalFOV ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add tan horizontal FOV" );
		return;
	}
	const float fTanVerticalFOV = xFrustum.GetTanVerticalFOV();
	if ( Asura_DebugOverlayData::AddFloat( fTanVerticalFOV ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add tan vertical FOV" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xPosition ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add position" );
		return;
	}
	if ( Asura_DebugOverlayData::AddQuat( xOrientation ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add orientation" );
		return;
	}

	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddFrustum failed to add ARGB" );
		return;
	}
	return;
}

void Asura_DebugOverlay::AddConsoleString( const char* const szText, const u_int uARGB )
{
	// Include the null terminator
	const int iStringLen = strlen( szText ) + 1;
	const u_int uNumBytes = Asura_DebugOverlayData::s_auPrimMemSize[Asura_DebugOverlayData::ASURA_DO_CONSOLE_STRING] + iStringLen;

	// Is there enough memory to add the text
	if ( Asura_DebugOverlayData::s_pucListNext + uNumBytes > Asura_DebugOverlayData::s_pucListEnd )
	{
		return;
	}
	const bool bWorldSpace = false;
	const bool bIgnoreZBuffer = true;

	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_CONSOLE_STRING, bWorldSpace, bIgnoreZBuffer, uNumBytes ) == false )
	{
		Asura_Assert( false, "AddConsoleString failed to add header" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddConsoleString failed to add colour" );
		return;
	}
	if ( Asura_DebugOverlayData::AddString( szText, iStringLen ) == false )
	{
		Asura_Assert( false, "AddConsoleString failed to add string" );
		return;
	}
}

void Asura_DebugOverlay::AddDebugString( const bool bWorldSpace, const bool bIgnoreZBuffer, const char* const szText, const Asura_Vector_3& xPos, const u_int uARGB )
{
	// Include the null terminator
	const int iStringLen = strlen( szText ) + 1;
	const u_int uNumBytes = Asura_DebugOverlayData::s_auPrimMemSize[Asura_DebugOverlayData::ASURA_DO_DEBUG_STRING] + iStringLen;

	// Is there enough memory to add the text
	if ( Asura_DebugOverlayData::s_pucListNext + uNumBytes > Asura_DebugOverlayData::s_pucListEnd )
	{
		return;
	}
	if ( Asura_DebugOverlayData::AddHeader( Asura_DebugOverlayData::ASURA_DO_DEBUG_STRING, bWorldSpace, bIgnoreZBuffer, uNumBytes ) == false )
	{
		Asura_Assert( false, "AddDebugString failed to add header" );
		return;
	}
	if ( Asura_DebugOverlayData::AddVector( xPos ) == false )
	{
		Asura_Assert( false, "AddDebugString failed to add pos" );
		return;
	}
	if ( Asura_DebugOverlayData::AddARGB( uARGB ) == false )
	{
		Asura_Assert( false, "AddDebugString failed to add colour" );
		return;
	}
	if ( Asura_DebugOverlayData::AddString( szText, iStringLen ) == false )
	{
		Asura_Assert( false, "AddDebugString failed to add string" );
		return;
	}
}

#endif // #if ASURA_USE_DEBUGOVERLAY == 1
