///////////////////////////////////////////////////////////////////////////
//
// Asura_DebugOverlay.h
//
// Created by JAT - 31 January 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// The Asura_DebugOverlay system - a static singleton class
// A very simple deferred debug rendering system see
//
// http://asurawiki.rebellion.co.uk/twiki/bin/view/Main/AsuraDebug
//
// * Unified graphics based debug overlay system - 
//    so different systems can have their own graphical debug overlays (called debug modes) 
// * Have base Asura modes e.g. Render, Physics, Input, Global, Entity, AI, Sound, Network, Messaging 
// * Supports project specific modes from the outset 
// * Self-contained system very little reliance on external systems (all done via registering callbacks) 
// * Deterministic rendering order - support screenspace and 3D, plus z-buffer test on/off 
// * Basic list of rendering primitives using existing render functionality 
// * Supports input handling and a callback per mode
//
// !defined( ASURA_USE_DEBUGOVERLAY ) then the class and its interface doesn't exist
// ASURA_USE_DEBUGOVERLAY == 0 then all the functions exist but they are inlined empty functions (which will be removed by the optimizer)
// ASURA_USE_DEBUGOVERLAY == 1 then all the functions exist and they work
//
///////////////////////////////////////////////////////////////////////////

#if defined( ASURA_USE_DEBUGOVERLAY )

#ifndef ASURA_DEBUGOVERLAY_HH
#define ASURA_DEBUGOVERLAY_HH

#include "Core/Asura_Maths.h"
#include "Debugging/Asura_DebugOverlay_Modes.h"

typedef void (*AsuraDebugOverlayCallbackFn)( void* const pxUserData );

// Forward declarations
class Asura_RGBA;
class Asura_Frustum;

// Static singleton class
class Asura_DebugOverlay
{
public:
	// Initialise internals - allocate memory buffer
	static void Initialise();
	// Free memory
	static void Shutdown( void );
 
	// Reset internal state
	static void Reset( void );
	
	// Register a debug mode
	static bool RegisterMode( const char* const szName, const int iMode, AsuraDebugOverlayCallbackFn pCallbackFn, void* const pxUserData );

	// Choose a specific debug mode
	static bool SelectMode( const int iMode );

	// Change the debug mode: i.e. +1, -1
	static void CycleDebugMode( const int iIncrement );

	// Render the debug data -  does world space data, then screen space data
	static void Render( void );

	// Update the Asura_Debug system - calls the correct debug callback
	static void Update( void );

	// Return true if a mode is already registered with the system
	static bool ModeIsRegistered ( const int iMode );

	// Primarily for the Asura_DebugOverlayData namespace :(
	static void SetActive( const bool bActive );
	static bool IsActive( void );
	static int GetCurrentMode( void );
	#if ASURA_USE_DEBUGOVERLAY == 1
	static bool GetUseVectorFont() { return s_bUseVectorFont; }
	static float GetVectorFontWidth() { return s_fVectorFontWidth; }
	static float GetVectorFontWHRatio() { return s_fVectorFontWHRatio; }
	#endif // #if ASURA_USE_DEBUGOVERLAY == 1

	// Draw a sphere in world space
	static void DrawSphereWorld( const int iMode, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments = 16, const u_int uNumSlices = 8 );

	// Draw a line in world space
	static void DrawLineWorld( const int iMode, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB );
	static void DrawLineWorld( const int iMode, const Asura_AVector3& xStart, const Asura_AVector3& xEnd, const u_int uARGB );

	// Draw an arrow in world space
	static void DrawArrowWorld( const int iMode, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion = 0.125f, const float fCrossSize = 0.0f );

	// Draw a cross (XYZ world axis set) in world space
	static void DrawCrossWorld( const int iMode, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize = 1.0f );

	// Draw an OBB in world space
	static void DrawOBBWorld( const int iMode, const Asura_OBB& xOBB, const u_int uARGB );

	// Draw text in world space (z-buffered)
	static void DrawTextWorld( const int iMode, const Asura_Vector_3& xPosition, const u_int uARGB, const char* const szFmtText, ... );

	// Quaternion functions...
	static void DrawAxisWorld( const int iMode, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize = 1.0f, const float fAlpha = 1.0f );
	static void DrawBoundingBoxWorld( const int iMode, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB );
	static void DrawFrustumWorld( const int iMode, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB );

	// 3x3 matrix functions...
	static void DrawAxisWorld( const int iMode, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const float fSize = 1.0f, const float fAlpha = 1.0f );
	static void DrawBoundingBoxWorld( const int iMode, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB );
	static void DrawFrustumWorld( const int iMode, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB );

	// Draw text in normalised screenspace [top-left is 0,0, bottom-right 1,1] (not z-buffered)
	static void DrawTextScreen( const int iMode, const Asura_Vector_2& xPosition, const u_int uARGB, const char* const szFmtText, ... );

	// Output a console message using the default console colour
	static void DrawConsoleText( const int iMode, const char* const szFmtText, ... );
	// Console message with a specific colour
	static void DrawConsoleText( const int iMode, const u_int uARGB, const char* const szFmtText, ... );

	// Printf debugging but filtered by the mode and if the debug overlay system is active
	static void Printf( const int iMode, const char* const szFmtText, ... );

private:

#if ASURA_USE_DEBUGOVERLAY == 1
	static int s_iCurrentMode;
	static bool s_bActive;
	static bool s_bUseVectorFont;
	static float s_fVectorFontWidth;
	static float s_fVectorFontWHRatio;
#endif // #if ASURA_USE_DEBUGOVERLAY == 1

	// Internal versions of the draw functions - so the function is ignored if the requested mode differs from the currently active mode (performance speed up)

	// Internal functions to add the commands to the command list
	static void AddSphere( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices );
	static void AddLine( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB );
	static void AddArrow( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize );
	static void AddCross( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize ); 
	static void AddOBB( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_OBB& xOBB, const u_int uARGB ); 
	static void AddAxis( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha );
	static void AddBoundingBox( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB );
	static void AddFrustum( const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB );

	static void AddConsoleString( const char* const szText, const u_int uARGB );
	static void AddDebugString( const bool bWorldSpace, const bool bIgnoreZBuffer, const char* const szText, const Asura_Vector_3& xPos, const u_int uARGB );
};

#if ASURA_USE_DEBUGOVERLAY == 1

inline void Asura_DebugOverlay::SetActive( const bool bActive ) { s_bActive = bActive; }

// Primarily for the Asura_DebugOverlayData namespace :(
inline bool Asura_DebugOverlay::IsActive( void ) { return s_bActive; }
inline int Asura_DebugOverlay::GetCurrentMode( void ) { return s_iCurrentMode; }

// If worried about code size and potential i-cache hits then don't inline these functions

// Draw a sphere in world space
inline void Asura_DebugOverlay::DrawSphereWorld( const int iMode, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddSphere( bWorldSpace, bIgnoreZBuffer, xCentre, fRadius, uARGB, uNumSegments, uNumSlices );
}

// Draw a line in world space
inline void Asura_DebugOverlay::DrawLineWorld( const int iMode, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddLine( bWorldSpace, bIgnoreZBuffer, xStart, xEnd, uARGB );
}

inline void Asura_DebugOverlay::DrawLineWorld( const int iMode, const Asura_AVector3& xStart, const Asura_AVector3& xEnd, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const Asura_Vector_3 xStart3( xStart );
	const Asura_Vector_3 xEnd3( xEnd );
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddLine( bWorldSpace, bIgnoreZBuffer, xStart3, xEnd3, uARGB );
}

// Draw a arrow in world space
inline void Asura_DebugOverlay::DrawArrowWorld( const int iMode, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddArrow( bWorldSpace, bIgnoreZBuffer, xStart, xEnd, uARGB, fHeadProportion, fCrossSize );
}

// Draw a cross in world space
inline void Asura_DebugOverlay::DrawCrossWorld( const int iMode, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddCross( bWorldSpace, bIgnoreZBuffer, xPosition, uARGB, fSize );
}

// Draw an OBB in world space
inline void Asura_DebugOverlay::DrawOBBWorld( const int iMode, const Asura_OBB& xOBB, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddOBB( bWorldSpace, bIgnoreZBuffer, xOBB, uARGB );
}

// Quaternion functions...
inline void Asura_DebugOverlay::DrawAxisWorld( const int iMode, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddAxis( bWorldSpace, bIgnoreZBuffer, xPosition, xOrientation, fSize, fAlpha );
}

inline void Asura_DebugOverlay::DrawBoundingBoxWorld( const int iMode, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddBoundingBox( bWorldSpace, bIgnoreZBuffer, xBB, xPosition, xOrientation, uARGB );
}

inline void Asura_DebugOverlay::DrawFrustumWorld( const int iMode, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	AddFrustum( bWorldSpace, bIgnoreZBuffer, xFrustum, xPosition, xOrientation, uARGB );
}

// 3x3 matrix functions...
inline void Asura_DebugOverlay::DrawAxisWorld( const int iMode, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const float fSize, const float fAlpha )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	Asura_Quat xQuat;
	xQuat.FromMatrix( xOrientation );
	AddAxis( bWorldSpace, bIgnoreZBuffer, xPosition, xQuat, fSize, fAlpha );
}

inline void Asura_DebugOverlay::DrawBoundingBoxWorld( const int iMode, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	Asura_Quat xQuat;
	xQuat.FromMatrix( xOrientation );
	AddBoundingBox( bWorldSpace, bIgnoreZBuffer, xBB, xPosition, xQuat, uARGB );
}

inline void Asura_DebugOverlay::DrawFrustumWorld( const int iMode, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB )
{
	if ( ( s_bActive == false ) || ( ( iMode != ASURA_DO_GLOBAL ) && ( iMode != s_iCurrentMode ) ) )
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = false;
	Asura_Quat xQuat;
	xQuat.FromMatrix( xOrientation );
	AddFrustum( bWorldSpace, bIgnoreZBuffer, xFrustum, xPosition, xQuat, uARGB );
}

#else // #if ASURA_USE_DEBUGOVERLAY == 1

// Just empty inline functions to effectively remove the code when the #define isn't defined
// Initialise internals - allocate memory buffer
inline void Asura_DebugOverlay::Initialise( void ) {}

// Free memory
inline void Asura_DebugOverlay::Shutdown( void ) {}
 
// Reset internal state
inline void Asura_DebugOverlay::Reset( void )  {}
	
// Register a debug mode
inline bool Asura_DebugOverlay::RegisterMode( const char* const szName, const int iMode, AsuraDebugOverlayCallbackFn pCallbackFn, void* const pxUserData ) { (void)szName; (void)iMode; (void)pCallbackFn; (void)pxUserData; return true; }

// Choose a specific debug mode
inline bool Asura_DebugOverlay::SelectMode( const int iMode ) { (void)iMode; return true; }

// Change the debug mode: i.e. +1, -1
inline void Asura_DebugOverlay::CycleDebugMode( const int iIncrement ) { (void)iIncrement; }

// Render the debug data -  does world space data, then screen space data
inline void Asura_DebugOverlay::Render( void ) {}

// Update the Asura_Debug system - calls the correct debug callback
inline void Asura_DebugOverlay::Update( void ) {}

inline void Asura_DebugOverlay::SetActive( const bool bActive ) { (void)bActive; }

// Primarily for the Asura_DebugOverlayData namespace :(
inline bool Asura_DebugOverlay::IsActive( void ) { return false; }
inline int GetCurrentMode( void ) { return -1; }

inline void Asura_DebugOverlay::DrawSphereWorld( const int iMode, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices ) { (void)iMode; (void)xCentre; (void)fRadius; (void)uARGB; (void)uNumSegments; (void)uNumSlices; }
inline void Asura_DebugOverlay::DrawLineWorld( const int iMode, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB ) { (void)iMode; (void)xStart; (void)xEnd; (void)uARGB; }
inline void Asura_DebugOverlay::DrawLineWorld( const int iMode, const Asura_AVector3& xStart, const Asura_AVector3& xEnd, const u_int uARGB ) { (void)iMode; (void)xStart; (void)xEnd; (void)uARGB; }
inline void Asura_DebugOverlay::DrawArrowWorld( const int iMode, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize ) { (void)xStart; (void)xEnd; (void)uARGB; (void)fHeadProportion; (void)fCrossSize; }
inline void Asura_DebugOverlay::DrawTextWorld( const int iMode, const Asura_Vector_3& xPosition, const u_int uARGB, const char* const szFmtText, ... ) { (void)iMode; (void)xPosition; (void)uARGB; (void)szFmtText; }
inline void Asura_DebugOverlay::DrawCrossWorld( const int iMode, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize ) { (void)iMode; (void)xPosition; (void)uARGB; (void)fSize; }
inline void Asura_DebugOverlay::DrawOBBWorld( const int iMode, const Asura_OBB& xOBB, const u_int uARGB ) { (void)iMode; (void)xOBB; (void)uARGB; }

inline void Asura_DebugOverlay::DrawAxisWorld( const int iMode, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const float fSize, const float fAlpha ) { (void)iMode; (void)xPosition; (void)xOrientation; (void)fSize; (void)fAlpha; }
inline void Asura_DebugOverlay::DrawBoundingBoxWorld( const int iMode, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB ) { (void)iMode; (void)xBB; (void)xPosition; (void)xOrientation; (void)uARGB; }
inline void Asura_DebugOverlay::DrawFrustumWorld( const int iMode, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB ) { (void)iMode; (void)xPosition; (void)xOrientation; (void)uARGB; }

inline void Asura_DebugOverlay::DrawAxisWorld( const int iMode, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha ) { (void)iMode; (void)xPosition; (void)xOrientation; (void)fSize; (void)fAlpha; }
inline void Asura_DebugOverlay::DrawBoundingBoxWorld( const int iMode, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB ) { (void)iMode; (void)xBB; (void)xPosition; (void)xOrientation; (void)uARGB; }
inline void Asura_DebugOverlay::DrawFrustumWorld( const int iMode, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB ) { (void)iMode; (void)xPosition; (void)xOrientation; (void)uARGB; }

inline void Asura_DebugOverlay::DrawTextScreen( const int iMode, const Asura_Vector_2& xPosition, const u_int uARGB, const char* const szFmtText, ... ) { (void)iMode; (void)xPosition; (void)uARGB; (void)szFmtText; }

inline void Asura_DebugOverlay::DrawConsoleText( const int iMode, const char* const szFmtText, ... ) { (void)iMode; (void)szFmtText; }
inline void Asura_DebugOverlay::DrawConsoleText( const int iMode, const u_int uARGB, const char* const szFmtText, ... ) { (void)iMode; (void)szFmtText; }

inline void Asura_DebugOverlay::Printf( const int iMode, const char* const szFmtText, ... ) { (void)iMode; (void)szFmtText; }

#endif // #if ASURA_USE_DEBUGOVERLAY == 1

#endif // #ifndef ASURA_DEBUGOVERLAY_HH

#endif // #if defined( ASURA_USE_DEBUGOVERLAY )
