//////////////////////////////////////////////////////////////////////////////
//
// Asura_Profiling_BarRenderer.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Debugging/Asura_Profiling_BarRenderer.h"

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int auPROFILING_DEFAULT_BACKGROUND_COLOURS[] = { 0xFF666666u, 0xFF333333u };

static const float fPROFILING_BAR_X    = 0.1f;
static const float fPROFILING_BAR_W    = 0.8f;
static const float fPROFILING_BAR_H    = 0.01f;
static const float fPROFILING_BORDER_H = 0.004f;

// G L O B A L S /////////////////////////////////////////////////////////////

// Settings for all bars being rendered.
static float g_fDivisionCountIncreaseDelay   = 0.5f;
static float g_fDivisionCountReductionDelay  = 1.5f;
static int   g_iMaxDivisions                 = 12;
static float g_fTargetFrameRate              = 60.0f;

// State for all bars being rendered.
static u_int g_uFrameNumberThisFrame             = 0xFFFFFFFF;
static float g_fFrameTimeThisFrame               = -1.0f;
static int   g_iNumDivisionsThisFrame            = -1;
static float g_fRemainingDivisionIncreaseDelay   = 0.0f;
static float g_fRemainingDivisionReductionDelay  = 0.0f;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void RenderSection( float fBarOffset, float fWidth, float fYOffset, u_int uARGB );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Profiling_BarRenderer_Data
//

Asura_Profiling_BarRenderer_Data::Asura_Profiling_BarRenderer_Data( void )
    : m_puBackgroundARGBs( NULL )
    , m_uNumBackgroundARGBs( 0 )
    , m_puBarARGBs( NULL )
    , m_uNumBarARGBs( 0 )
    , m_uFrameRemainderARGB( 0 )
    , m_fYOffset( -Asura_Maths::LargeFloat )
    , m_fTotalFrameTime( -1.0f )
    , m_pfTimers( NULL )
    , m_uNumTimers( 0 )
{
}


void Asura_Profiling_BarRenderer_Data::SetBackgroundColours( const u_int* puBackgroundARGBs, u_int uNumBackgroundARGBs )
{
    Asura_Assert( puBackgroundARGBs && uNumBackgroundARGBs, "" );

    m_puBackgroundARGBs   = puBackgroundARGBs;
    m_uNumBackgroundARGBs = uNumBackgroundARGBs;
}


void Asura_Profiling_BarRenderer_Data::SetBarColours( const u_int* puBarARGBs, u_int uNumBarARGBs, u_int uFrameRemainderARGB )
{
    Asura_Assert( puBarARGBs && uNumBarARGBs, "" );

    m_puBarARGBs          = puBarARGBs;
    m_uNumBarARGBs        = uNumBarARGBs;
    m_uFrameRemainderARGB = uFrameRemainderARGB;
}


void Asura_Profiling_BarRenderer_Data::SetUseDefaultBackgroundColours( void )
{
    SetBackgroundColours( auPROFILING_DEFAULT_BACKGROUND_COLOURS, sizeof(auPROFILING_DEFAULT_BACKGROUND_COLOURS) / sizeof(u_int) );
}


void Asura_Profiling_BarRenderer_Data::SetYOffset( float fYOffset )
{
    m_fYOffset = fYOffset;
}


void Asura_Profiling_BarRenderer_Data::SetNewFrameData( float fTotalFrameTime,
                                                        const float* pfTimers, u_int uNumTimers )
{
    m_fTotalFrameTime = fTotalFrameTime;
    m_pfTimers        = pfTimers;
    m_uNumTimers      = uNumTimers;
}

//
// Asura_Profiling_BarRenderer
//

void Asura_Profiling_BarRenderer::Render( Asura_Profiling_BarRenderer_Data& xData )
{
    Asura_Assert( xData.m_puBackgroundARGBs, "Background colours not set" );
    Asura_Assert( xData.m_puBarARGBs, "Bar colours not set" );
    Asura_Assert( xData.m_fYOffset > -Asura_Maths::LargeFloat, "Y offset not set" );
    Asura_Assert( xData.m_fTotalFrameTime >= 0.0f, "You must provide new data every frame" );

    Asura_Render::FlushDisplay();

    Asura_Render::Overlay.TextureIndex     = ASURA_TEXTURE_NONE;
    Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;

    const float fTargetFrameDuration = 1.0f / g_fTargetFrameRate;

    // If this is the first set of bars rendered this frame then calculate
    // new parameters. We calculate the number of divisions based on the
    // overall frame time, which should hopefully be limited by the worst
    // case of anything we happen to be profiling. Therefore if we render
    // multiple bars - CPU and GPU, say - then the number of divisions should
    // be consistent.
    u_int uFrameNumber = Asura_Timers::GetFrameNumber();
    if( g_uFrameNumberThisFrame != uFrameNumber || g_fFrameTimeThisFrame < 0.0f )
    {
        float fFrameTime = Asura_Timers::GetFrameTime();

        int iNumDivisionsForNewFrameTime = Asura_Maths::Min( Asura_Maths::FloatToInt_Chop(fFrameTime / fTargetFrameDuration) + 1, g_iMaxDivisions );

        bool bUseNewDivisionCount;
        if( g_fFrameTimeThisFrame < 0.0f || iNumDivisionsForNewFrameTime == g_iNumDivisionsThisFrame )
        {
            // If the number of divisions hasn't actually changed or if this
            // is the first frame then just go with it - we'll worry about
            // hysteresis when it's actually different.
            bUseNewDivisionCount = true;
        }
        else
        {
            // Otherwise check the appropriate timer.
            if( iNumDivisionsForNewFrameTime > g_iNumDivisionsThisFrame )
            {
                g_fRemainingDivisionIncreaseDelay -= fFrameTime;
                bUseNewDivisionCount = ( g_fRemainingDivisionIncreaseDelay <= 0.0f );

                // Reset other timer regardless.
                g_fRemainingDivisionReductionDelay = g_fDivisionCountReductionDelay;
            }
            else
            {
                g_fRemainingDivisionReductionDelay -= fFrameTime;
                bUseNewDivisionCount = ( g_fRemainingDivisionReductionDelay <= 0.0f );

                // Reset other timer regardless.
                g_fRemainingDivisionIncreaseDelay = g_fDivisionCountIncreaseDelay;
            }
        }

        if( bUseNewDivisionCount )
        {
            g_iNumDivisionsThisFrame           = iNumDivisionsForNewFrameTime;
            g_fFrameTimeThisFrame              = fFrameTime;
            g_fRemainingDivisionIncreaseDelay  = g_fDivisionCountIncreaseDelay;
            g_fRemainingDivisionReductionDelay = g_fDivisionCountReductionDelay;
        }

        g_uFrameNumberThisFrame = uFrameNumber;
    }

    const float fTotalDisplayedDuration = fTargetFrameDuration * g_iNumDivisionsThisFrame;

    // Render background divisions.
    {
        const float fDivisionW = (fPROFILING_BAR_W * Asura_Viewport::GetWidth()) / g_iNumDivisionsThisFrame;
        const float fDivisionY = xData.m_fYOffset * Asura_Viewport::GetHeight();
        const float fDivisionH = (fPROFILING_BAR_H + fPROFILING_BORDER_H * 2.0f) * Asura_Viewport::GetHeight();

        for( int i = 0; i < g_iNumDivisionsThisFrame; ++i )
        {
            float fX = fPROFILING_BAR_X * Asura_Viewport::GetWidth() + fDivisionW * i;
            Asura_Render::Overlay.Colour = xData.m_puBackgroundARGBs[i % xData.m_uNumBackgroundARGBs];
            Asura_Render::Overlay.SetRectXY( fX, fDivisionY, fDivisionW, fDivisionH );
            Asura_Render::Overlay.Render();
        }
    }

    float fProportionOffset = 0.0f;

    // Render individual timer bars.
    for( u_int u = 0; u < xData.m_uNumTimers; ++u )
    {
        float fThisTimer = xData.m_pfTimers[u];
        if( fThisTimer > 0.0f )
        {
            float fProportion = fThisTimer / fTotalDisplayedDuration;
            RenderSection( fProportionOffset, fProportion, xData.m_fYOffset, xData.m_puBarARGBs[u % xData.m_uNumBarARGBs] );
            fProportionOffset += fProportion;
        }
    }

    // And the remainder of the frame.
    {
        float fTotalOfTimers = 0.0f;
        for( u_int u = 0; u < xData.m_uNumTimers; ++u )
        {
            fTotalOfTimers += xData.m_pfTimers[u];
        }

        float fRemainingTime = Asura_Maths::Max( xData.m_fTotalFrameTime - fTotalOfTimers, 0.0f );
        if( fRemainingTime > 0.0f )
        {
            float fProportion = fRemainingTime / fTotalDisplayedDuration;
            RenderSection( fProportionOffset, fProportion, xData.m_fYOffset, xData.m_uFrameRemainderARGB );
        }
    }

    Asura_Render::FlushDisplay();

    // Invalidate the frame data passed in so that we can ensure the caller
    // is setting it every frame.
    xData.m_fTotalFrameTime = -1.0f;
}


void Asura_Profiling_BarRenderer::SetDivisionCountIncreaseDelay( float fDelay )
{
    Asura_Assert( 0.0f <= fDelay && fDelay < 10.0f, "Silly value specified" );
    g_fDivisionCountIncreaseDelay = fDelay;
}


void Asura_Profiling_BarRenderer::SetDivisionCountReductionDelay( float fDelay )
{
    Asura_Assert( 0.0f <= fDelay && fDelay < 10.0f, "Silly value specified" );
    g_fDivisionCountReductionDelay = fDelay;
}


void Asura_Profiling_BarRenderer::SetMaxDivisions( int iMax )
{
    Asura_Assert( 0 < iMax && iMax < 50, "Silly value specified" );
    g_iMaxDivisions = iMax;
}


void Asura_Profiling_BarRenderer::SetTargetFrameRate( float fTarget )
{
    Asura_Assert( 5.0f <= fTarget && fTarget <= 100.0f, "Silly value specified" );
    g_fTargetFrameRate = fTarget;
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void RenderSection( float fBarOffset, float fWidth, float fYOffset, u_int uARGB )
{
    float fX = (fPROFILING_BAR_X + fPROFILING_BAR_W * fBarOffset) * Asura_Viewport::GetWidth();
    float fY = (fYOffset + fPROFILING_BORDER_H) * Asura_Viewport::GetHeight();
    float fW = fPROFILING_BAR_W * fWidth * Asura_Viewport::GetWidth();
    float fH = fPROFILING_BAR_H * Asura_Viewport::GetHeight();
    Asura_Render::Overlay.SetRectXY( fX, fY, fW, fH );
    Asura_Render::Overlay.Colour = uARGB;
    Asura_Render::Overlay.Render();
}
