//////////////////////////////////////////////////////////////////////////////
//
// Asura_Profiling_System.cpp
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_PROFILING_SYSTEM

// header
#include "Debugging/Asura_Profiling_System.h"

// asura
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"

// G L O B A L S /////////////////////////////////////////////////////////////
static Asura_TickCount g_axStartTimes[ASURA_PROFILING_SYSTEM_MAX_TIMERS];

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Profiling_System::Timer
//
//////////////////////////////////////////////////////////////////////////////

void Asura_Profiling_System::Timer::Reset()
{
    m_xTotalDuration = static_cast< Asura_TickCount >( 0 );
	// Arbitrary large value which the min duration should fall below.
	m_xMinDuration   = static_cast< Asura_TickCount >( 1000000 );
	m_xMaxDuration   = static_cast< Asura_TickCount >( 0 );
    m_uTriggerCount = 0;
    m_ucInUse = m_ucTiming = 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Profiling_System
//
//////////////////////////////////////////////////////////////////////////////

Asura_Profiling_System::Timer Asura_Profiling_System::s_axTimers[ ASURA_PROFILING_SYSTEM_MAX_TIMERS ];

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Asura_Profiling_System::Initialise()
{
    for ( int i = 0; i < ASURA_PROFILING_SYSTEM_MAX_TIMERS; ++i )
    {
        s_axTimers[ i ].m_ucInUse = 0;
    }
}

Asura_Handle Asura_Profiling_System::CreateTimer()
{
    // return first free timer
    for ( int i = 0; i < ASURA_PROFILING_SYSTEM_MAX_TIMERS; ++i )
    {
        if ( s_axTimers[ i ].m_ucInUse == 0 )
        {
            s_axTimers[ i ].Reset();
            s_axTimers[ i ].m_ucInUse = 1;
            return i;
        }
    }

    Asura_Assert( false, "Asura_Profiling_System - run out of timers" );
    return ASURA_INVALID_HANDLE;
}

void Asura_Profiling_System::DestroyTimer( const Asura_Handle iTimerHandle )
{
    Asura_Assert( ( s_axTimers[ iTimerHandle ].m_ucInUse == 1 ), "Asura_Profiling_System - accessing unused timer" );

    s_axTimers[ iTimerHandle ].m_ucInUse = 0;
}

void Asura_Profiling_System::ResetTimer(const Asura_Handle iTimerHandle)
{
    Asura_Assert( ( s_axTimers[ iTimerHandle ].m_ucInUse == 1 ), "Asura_Profiling_System - accessing unused timer" );

    s_axTimers[iTimerHandle].Reset();
	s_axTimers[iTimerHandle].m_ucInUse = 1;
}

void Asura_Profiling_System::StartTiming( const Asura_Handle iTimerHandle )
{
    Asura_Assert( ( s_axTimers[ iTimerHandle ].m_ucInUse == 1 ), "Asura_Profiling_System - accessing unused timer" );

    if ( s_axTimers[ iTimerHandle ].m_ucInUse == 1 && s_axTimers[ iTimerHandle ].m_ucTiming == 0 )
    {
        s_axTimers[ iTimerHandle ].m_ucTiming = 1;

		g_axStartTimes[ iTimerHandle ] = Asura_Timers::GetCurTicks();
    }
}

Asura_TickCount Asura_Profiling_System::StopTiming( const Asura_Handle iTimerHandle )
{
    if ( s_axTimers[ iTimerHandle ].m_ucInUse == 1 && s_axTimers[ iTimerHandle ].m_ucTiming == 1 )
    {
		const Asura_TickCount xTicks = Asura_Timers::GetElapsedTicks(g_axStartTimes[ iTimerHandle ]);

        if ( xTicks < s_axTimers[ iTimerHandle ].m_xMinDuration )
        {
            s_axTimers[ iTimerHandle ].m_xMinDuration = xTicks;
        }
        // not else if!
        if ( xTicks > s_axTimers[ iTimerHandle ].m_xMaxDuration )
        {
            s_axTimers[ iTimerHandle ].m_xMaxDuration = xTicks;
        }

        s_axTimers[ iTimerHandle ].m_xTotalDuration += xTicks;
        ++s_axTimers[ iTimerHandle ].m_uTriggerCount;
        s_axTimers[ iTimerHandle ].m_ucTiming = 0;

		return(xTicks);
    }
	else
	{
		return(static_cast<Asura_TickCount>(0.0f));
	}
}

float Asura_Profiling_System::GetMinDuration( const Asura_Handle iTimerHandle )
{
	if(s_axTimers[iTimerHandle].m_uTriggerCount > 0)
	{
		return ( Asura_Timers::TicksToSeconds(s_axTimers[ iTimerHandle ].m_xMinDuration) );
	}
	else
	{
		return(Asura_Maths::LargeFloat);
	}
}

float Asura_Profiling_System::GetMaxDuration( const Asura_Handle iTimerHandle )
{
	if(s_axTimers[iTimerHandle].m_uTriggerCount > 0)
	{
	    return ( Asura_Timers::TicksToSeconds(s_axTimers[ iTimerHandle ].m_xMaxDuration) );
	}
	else
	{
		return(0.0f);
	}
}

float Asura_Profiling_System::GetAverageDuration( const Asura_Handle iTimerHandle )
{
	if(s_axTimers[iTimerHandle].m_uTriggerCount > 0)
	{
	    return ( Asura_Timers::TicksToSeconds(s_axTimers[ iTimerHandle ].m_xTotalDuration) / s_axTimers[ iTimerHandle ].m_uTriggerCount );
	}
	else
	{
		return(0.0f);
	}
}

unsigned int Asura_Profiling_System::GetTriggerCount( const Asura_Handle iTimerHandle )
{
    return ( s_axTimers[ iTimerHandle ].m_uTriggerCount );
}

float Asura_Profiling_System::GetTotalDuration( const Asura_Handle iTimerHandle )
{
	return (Asura_Timers::TicksToSeconds(s_axTimers[iTimerHandle].m_xTotalDuration));
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // #ifdef ASURA_USE_PROFILING_SYSTEM
