
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ScreenShot_System.h
// Author : Andy Tate
// Date   : 9-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_SCREENSHOT_SYSTEM_H__
#define __ASURA_SCREENSHOT_SYSTEM_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Stream;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_SCREENSHOT_SYSTEM
class Asura_ScreenShot_System
{
public:
	static const u_int	uMAX_MAGNIFICATION	= 16;	// 10240x7168 from 640x448 buffer.

	// Public typedefs...

	typedef void RenderFn (void);


	// Public functions...

	// RenderHiResScreenshot must be given a function which do a complete render cycle
	// including Asura_Render::BeginFrame/EndFrame.
	// If bPrompt is true and the platform code has set the prompt keys and can display
	// a message, confirmation will be asked for, otherwise the shot is taken immediately.
	static	void	RenderHiResScreenshot	(RenderFn* pfnRender, bool bPrompt = true);

	// TakeNormalScreenShot can be called at any point - the screenshot will be taken
	// during EndFrame().
	static	void	TakeNormalScreenShot	(void)	{ s_bTakeScreenShot = true; }

	// IsScreenShotPending returns true iff a screenshot will be taken in the next EndFrame().
	static	bool	IsScreenShotPending		(void)	{ return (s_bTakeScreenShot); }
	static	bool	IsTakingHiResScreenshot	(void)	{ return (s_bTakingHiResScreenshot); }
	static	bool	ShouldFreezeFrame		(void)	{ return s_bTakeScreenShot || s_bTakingHiResScreenshot || s_bWaitingForHiResShotInput; }

	// DrawBufferIsReady is called when draw buffer is ready to be saved out.
	static	void	DrawBufferIsReady		(void);

	// function to add command-console interface
	static	void	AddConsoleCommands		(void);

	// interface for selecting resolution, quality etc.
	// SetMagnification() - multiplier for each dimension. min 1, max 16 (returns false if out of range).
	static	bool	SetMagnification		(u_int uMag);
	static	u_int	GetMagnification		(void);

	// SetAntiAliasing() - sets new values for the current magnification.
	// typically sample 1.0f, with a minimum distance between samples of between 0.1f and 0.35f.
	// returns the number of samples generated. Max samples is currently 128.
	// returns 0 if parameters are invalid.
	static	u_int	SetAntiAliasing			(float fSampledPixelWidth, float fMinDistBetweenSamples);

	// GetAntiAliasing - returns number of AA samples used for the current magnification.
	// Optional parameters allow retrieval of sampling values.
	static	u_int	GetAntiAliasing			(float* pfSampledPixelWidth = NULL, float* pfMinDistBetweenSamples = NULL);

    // In situations where some funky custom rendering is going on, possibly
    // on a project-specific basis, this can be used to override the standard
    // process of filling out the FP buffer.
    static void SetAddScreenToFPBufferCallback( void(*pfnCB)(const Asura_IntVector_2&, const Asura_IntVector_2&, u_int, float*) );

	static void SetOverrideFileName (const char* pszFileName) { Asura_PrintfToBfr (s_szOverriddenFileName, "%s", 256, true, pszFileName); }

protected:
	// Protected variables...


	// Protected functions...


private:
	// Private variables...

	static	bool			s_bTakeScreenShot;
	static	bool			s_bTakingHiResScreenshot;
	static	bool			s_bWaitingForHiResShotInput;
	// keys to take or not take the hi-res screenshot. If the platform is able to display 
	// a prompt saying e.g. "Enter - take, Escape - cancel" before committing to taking the 
	// screenshot then it should set up these keys in Platform_BeginHiResRender()
	static  int				s_iAcceptKey;
	static	int				s_aiCancelKey[2];
	static	char			s_szOverriddenFileName [256];

	// use our own little predictable random generator so the results are always the same
	static	Asura_Ranrot	s_xRanrot;


	// Private functions...

	Asura_ScreenShot_System  (void);
	Asura_ScreenShot_System  (const Asura_ScreenShot_System& xOther);
	~Asura_ScreenShot_System ();

			Asura_ScreenShot_System&	operator=						(const Asura_ScreenShot_System& xOther);

	static	void				UpdateProgressBar				(u_int uNum, u_int uTotal);
	static	int					BMPTakeScreenShot				(const char* szFilename, const Asura_IntVector_2& xPosition, const Asura_IntVector_2& xSize, const u_int uFormat);

	static	void				WriteBatchFile					(int iFirstIndex, int iMag);
	static	int					BMPTakeScreenShot				(const char* szFilename, const Asura_IntVector_2& xPosition, const Asura_IntVector_2& xSize);
	static	u_int				WriteBMPHeader					(Asura_Stream& xStream, const Asura_IntVector_2& xSize);

	static	bool				SaveFloatBufferAsBMP			(int iIndex);

	static	void				InitSeededFloats				(void);
	static	float				GetSeededFloat					(float fMin, float fMax);

	static	void				GeneratePoissonDiscSamples		(void);

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	static	bool				ConsoleCmd_SetMagnification		(u_int uMagnification, bool bInterrogate, bool bHelp);
	static	bool				ConsoleCmd_SetAAMinDist			(float fDist, bool bInterrogate, bool bHelp);
	static	bool				ConsoleCmd_SetAASampleWidth		(float fWidth, bool bInterrogate, bool bHelp);
#endif	// defined(ASURA_SUPPORT_COMMAND_CONSOLE)

	// Platform specific functions...
	static	void				Platform_RenderProgressBar		(const Asura_Vector_2& xPosition, const Asura_Vector_2& xSize, const u_int uARGB);
	static	void				Platform_GetScreenDimensions	(Asura_IntVector_2* pxPosition, Asura_IntVector_2* pxSize, u_int* puFormat);
	static	void				Platform_WriteScreenAs24Bit		(Asura_Stream& xStream, const Asura_IntVector_2& xPosition, const Asura_IntVector_2& xSize, const u_int uFormat);
	static	void				Platform_AddScreenToFPBuffer	(const Asura_IntVector_2& xPosition, const Asura_IntVector_2& xSize, const u_int uFormat, float* pfFPBuffer);
	static	void				Platform_BeginHiResRender		(void);
	static	void				Platform_EndHiResRender			(void);
	static	int					Platform_TakeScreenShot			(const char* szFilename);
	static	void				Platform_NotTakingScreenShot	(void);
	static	float*				Platform_AllocateFloatBuffer	(u_int uNumFloats);
	static	void				Platform_FreeFloatBuffer		(float* pfBuffer);
	static	void				Platform_RenderPrompt			(void);

};   // End of class 'Asura_ScreenShot_System'.

#else // not #ifdef ASURA_USE_SCREENSHOT_SYSTEM

class Asura_ScreenShot_System
{
public:
	// Public typedefs...

	typedef void RenderFn (void);

	// Public functions...
	static	void	RenderHiResScreenshot	(RenderFn* pfnRender, bool bPrompt = true)	{}
	static	void	TakeNormalScreenShot	(void)					{}
	static	bool	IsScreenShotPending		(void)					{ return false; }
	static	bool	IsTakingHiResScreenshot	(void)					{ return false; }
	static	bool	ShouldFreezeFrame		(void)					{ return false; }
	static	void	DrawBufferIsReady		(void)					{}
	static	void	AddConsoleCommands		(void)					{}
	static	bool	SetMagnification		(u_int uMag)			{ return false; }
	static	u_int	GetMagnification		(void)					{ return 0; }
	static	u_int	SetAntiAliasing			(float, float)			{ return 0; }
	static	u_int	GetAntiAliasing			(float* pf1 = NULL, float* pf2 = NULL) { return 0; }

    static void SetAddScreenToFPBufferCallback( void(*pfnCB)(const Asura_IntVector_2&, const Asura_IntVector_2&, u_int, float*) ) { }

};

#endif // else not #ifdef ASURA_USE_SCREENSHOT_SYSTEM


#endif   // ifndef __ASURA_SCREENSHOT_SYSTEM_H__


/// End of file. ///
