///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_ClientContainerInstance.cpp
//
// Created by Martin G.Harris (6/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// Instance of a container owned by a client entity.
// Contains the data that differs between instances, such as sound event handles,
// PFX handles, photon handles etc.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Asura_ClientContainerInstance.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "Animation/Asura_Animation.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Logfile.h"
#include "Entity/Container/Asura_Container.h"
#include "Entity/Container/Asura_Container_System.h"
#include "Entity/Container/Asura_ClientEntity_Container.h"
#include "Light/Asura_Light.h"
#include "Network/Asura_Network.h"
#include "Physics/Asura_Physics_Object.h"
#include "Shadows/Asura_Shadow_Instance.h"
#include "SpecialFX/Asura_Corona_System.h"

#ifdef ASURA_USE_SCREENSHOT_SYSTEM
#include "Debugging/Asura_ScreenShot_System.h"
#endif //#ifdef ASURA_USE_SCREENSHOT_SYSTEM

#include "Sound/Asura_Sound_Event_Storage.h"
#include "Sound/Asura_Sound_Event_System.h"

#include "SpecialFX/PFX/Asura_PFX_Entity_Effect.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

#ifdef ASURA_TARGET_SM3
#include "Light/Asura_Photon.h"
#include "Projectors/Asura_Projector_System.h"
#include "Projectors/Asura_Projector.h"
#include "SpecialFX/Asura_LightShaft_System.h"
#endif //#ifdef ASURA_TARGET_SM3

// Tools
#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "Entity/Container/Tools_Container_System.h"
#include "Entity/Container/Tools_Entity_Container.h"
#endif //ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////
// Asura_ContainerInstance_Properties
///////////////////////////////////////////////////////////////////////////////
// Add to the dynamic physics count and store the physics pointer in the entity.
// Return true, if we initialised the pos/ori data in the physics object.
bool Asura_ContainerInstance_Properties::AddDynamicPhysics(Asura_Physics_Object* pxPhysics, u_int uUniqueID)
{
	bool bInitialisedPosOri = false;
	if (m_pxOwner)
	{
		bInitialisedPosOri = m_pxOwner->AddDynamicPhysics(pxPhysics, uUniqueID);
	}
	++m_iNumDynamicPhysics;
	return bInitialisedPosOri;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance::Asura_ClientContainerInstance(const Asura_Container* pxContainer)
: m_pxContainer(pxContainer)
, m_pxParentCollection(0)
, m_uInstanceFlags(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON)
, m_uUniqueID(uASURA_CONTAINER_UNIQUE_ID_INVALID)
{
	Asura_Assert(pxContainer, "Null container pointer");
}

Asura_Hash_ID Asura_ClientContainerInstance::GetContainerHash() const
{
	return (m_pxContainer ? m_pxContainer->GetHash() : ASURA_HASH_ID_UNSET);
}

u_int Asura_ClientContainerInstance::GetClassification() const
{
	return (m_pxContainer ? m_pxContainer->GetClassification() : static_cast<u_int>(AsuraContainerClass_NotSet));
}

// Returns classification of container, skipping over attachments.
u_int Asura_ClientContainerInstance::GetClassificationSkipAttachment() const
{
	u_int uClass = GetClassification();

	if (uClass == AsuraContainerClass_Attachment)
	{
		const Asura_Container_Attachment* const pxAttachment = static_cast<const Asura_Container_Attachment*>(m_pxContainer);
		uClass = pxAttachment->GetChildClassification();
	}

	return uClass;
}

// Get owner entity guid.
Asura_Guid Asura_ClientContainerInstance::GetOwnerGuid() const
{
	return (m_pxParentCollection ? m_pxParentCollection->GetOwnerEntityGuid() : ASURA_GUID_UNREGISTERED); 
}

void Asura_ClientContainerInstance::SetStrobe(bool bStrobeOn)
{
	if (bStrobeOn)
	{
		SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON);
	}
	else
	{
		ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON);
	}
}

//////////////////////////////////
// GetGraphColourFromCollection //
//////////////////////////////////
// Helper. Returns true if we need to update from the output.
bool Asura_ClientContainerInstance::GetGraphColourFromCollection(Asura_Container_GraphColour_Output& xOutput)
{
	if (m_pxParentCollection)
	{
		if (m_pxParentCollection->GetGraphColour(xOutput))
		{
			// Some GraphColour containers are currently active.
			SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_USING_GRAPH);
			return true;
		}
		else if (GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_USING_GRAPH))
		{
			// The GraphColour has just been disabled, we need to reset all of the parameters.
			ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_USING_GRAPH);
			return true;
		}
	}
	
	// Nothing to do. No active graphs.
	return false;
}


/////////////////
// ToolsRender //
/////////////////
// Render in Tools if needed. Helper function.
inline void Asura_ClientContainerInstance::ToolsRender(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
#ifdef ASURA_TOOLS
	Tools_Container_System::ToolsRender(GetContainerHash(), xPosition, xOrientation);
#endif //ASURA_TOOLS
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Mesh
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_Mesh::Asura_ClientContainerInstance_Mesh(const Asura_Container_Mesh* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_pxAnimation(0)
, m_pxPhysicsObject(0)
{	
}

Asura_ClientContainerInstance_Mesh::~Asura_ClientContainerInstance_Mesh()
{
	if (Asura_Container_System::IsDoingStateChange())
	{
		// Preserve animation timer.
		if (m_pxParentCollection && m_pxAnimation)
		{
			const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);
			if (pxContainerMesh && pxContainerMesh->ShouldPreserveAnimTimer())
			{
				// Store the timer in the collection.
				const Asura_Hash_ID uAnimID = pxContainerMesh->GetAnimHash();
				const float fTimer = m_pxAnimation->m_xPrimaryWrapper.GetTimer();
				m_pxParentCollection->StoreAnimTimer(uAnimID, fTimer);
			}
		}
	}

	// Release the skin and animation.
	delete m_pxAnimation;
	m_pxAnimation = 0;

	// Release any physics object.
	DestroyPhysicsObject();
}

///////////////////////
// IsIgnoreForDamage //
///////////////////////
bool Asura_ClientContainerInstance_Mesh::IsIgnoreForDamage() const
{
	const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);
	if (pxContainerMesh)
	{
		return pxContainerMesh->IsIgnoreForDamage();
	}
	return false;
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_Mesh::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	// Set the physics position, but only if it hasn't already been initialised from stored data.
	if (m_pxPhysicsObject && !GetFlag(ASURA_CONTAINER_INSTANCE_MESH_FLAG_INITIALISEDPOSORI))
	{
		m_pxPhysicsObject->TeleportTo(xPosition);
		m_pxPhysicsObject->SetOrientation(xOrientation);
	}

	ResetFlag(ASURA_CONTAINER_INSTANCE_MESH_FLAG_INITIALISEDPOSORI);

	if (m_pxAnimation)
	{
		// Update the bones and stored position/orientation. This is needed to initialise the LOS AABB and LOS test.
		m_pxAnimation->Update(xPosition, xOrientation, ANIMTIME_USEGAMETIME, ASURA_HIERARCHY_RENDER_UPDATEBONESALWAYS | ASURA_HIERARCHY_RENDER_DONTRENDER);
	}
}

/////////////////////
// CreateAnimation //
/////////////////////
// Set up the skin and animation.
void Asura_ClientContainerInstance_Mesh::CreateAnimation(const Asura_Container_Mesh* pxContainer)
{
	if (pxContainer && !m_pxAnimation)
	{
		const Asura_Hash_ID uSkinID = pxContainer->GetSkinHash();
		if (uSkinID != ASURA_HASH_ID_UNSET)
		{
			m_pxAnimation = new Asura_Animation(uSkinID);
			if (m_pxAnimation)
			{
				const Asura_Hash_ID uAnimID = pxContainer->GetAnimHash();
				const int iAnimFlags = pxContainer->GetAnimFlags();

				float fInitialParametricTime = 0.0f;

				if (m_pxParentCollection && uAnimID != ASURA_HASH_ID_UNSET)
				{
					// If this anim hasn't been played before, should we randomise it?
					const bool bRandomise = pxContainer->ShouldRandomiseAnimTimer();
					
					// If the anim has been played before, should we restore the timer?
					const bool bRestore = pxContainer->ShouldPreserveAnimTimer();

					// Reset, restore or randomise the initial timer.
					fInitialParametricTime = m_pxParentCollection->RestoreAnimTimer(uAnimID, bRestore, bRandomise);
				}

				m_pxAnimation->m_xPrimaryWrapper.StartAnim2(uAnimID, iAnimFlags, 0.0f, 0.0f);

				if (fInitialParametricTime != 0.0f)
				{
					m_pxAnimation->m_xPrimaryWrapper.SetTimer(fInitialParametricTime);
				}

				SetParamFromContainer(pxContainer);
			}
		}
	}
}


/////////////////////////
// CreatePhysicsObject //
/////////////////////////
void Asura_ClientContainerInstance_Mesh::CreatePhysicsObject(const Asura_Container_Mesh* pxContainer)
{
	if (!pxContainer) return;
	if (pxContainer->HasNoCollisions()) return;
	DestroyPhysicsObject();

	// By default physics object never moves.
	ASURA_PHYSICS_OBJECT_TYPE eType = ASURA_PHYSICS_OBJECT_STATIC;

	if (IsAttachedToMarker())
	{
		// Physics object follows the marker point.
		eType = ASURA_PHYSICS_OBJECT_EXTERNALCONTROL;
	}
	else if (pxContainer->IsPhysicsObject())
	{
		// Physics object which can be pushed about.
		SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS);

		if (Asura_Network::IsServer())
		{
			// Pos/ori updated using physics on the server.
			eType = ASURA_PHYSICS_OBJECT_DYNAMIC;
		}
		else
		{
			// Client-only machine just updates the pos/ori from server messages.
			eType = ASURA_PHYSICS_OBJECT_EXTERNALCONTROL;
		}
	}
	else if (pxContainer->UseMovementOffset())
	{
		// Physics object follows the animation compass point.
		eType = ASURA_PHYSICS_OBJECT_EXTERNALCONTROL;
	}

	m_pxPhysicsObject = Asura_Physics_Object::Create(eType);

	Asura_Assert(m_pxPhysicsObject, "Failed to create physics object.");
	if (!m_pxPhysicsObject) return;

	m_pxPhysicsObject->SetCollisionProxy(static_cast<Asura_Physics_Object::COLLISION_PROXY>(pxContainer->GetCollisionProxy()));

	if (m_pxAnimation)
	{
		const Asura_Hierarchy_Skin* pxSkin = m_pxAnimation->GetSkin();
		if (pxSkin)
		{
			const Asura_Collision_Object* pxCollObject = pxSkin->GetCollisionObject();
			if (pxCollObject)
			{
				m_pxPhysicsObject->SetCollisionObject(pxCollObject);
			}
		}
	}

	CopyPhysicsProperties();
}

///////////////////////////
// CopyPhysicsProperties //
///////////////////////////
void Asura_ClientContainerInstance_Mesh::CopyPhysicsProperties()
{
	// Set physics properties from container.
	if (m_pxParentCollection && m_pxPhysicsObject && GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS))
	{
		const Asura_ClientContainerInstance_Physics* const pxProperties = m_pxParentCollection->FindPhysics();
		if (pxProperties)
		{
			const Asura_Container_Physics_Param& xParam = pxProperties->GetParam();
			
			m_pxPhysicsObject->SetGravityScale(xParam.GetGravityScale());
			m_pxPhysicsObject->SetElasticity(xParam.GetElasticity());
			m_pxPhysicsObject->SetFriction(xParam.GetFriction());
			m_pxPhysicsObject->SetDrag(xParam.GetDrag());
			m_pxPhysicsObject->SetAngularDrag(xParam.GetAngularDrag());

			if (m_pxPhysicsObject->GetInvMass(false) != xParam.GetInvMass())
			{
				m_pxPhysicsObject->SetInvMass(xParam.GetInvMass());
			}
		}
	}
}

//////////////////////////
// DestroyPhysicsObject //
//////////////////////////
void Asura_ClientContainerInstance_Mesh::DestroyPhysicsObject()
{
	if (m_pxPhysicsObject)
	{
		Asura_Physics_Object::Destroy( m_pxPhysicsObject );
		m_pxPhysicsObject = 0;
		ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS);
	}
}


////////////////
// Initialise //
////////////////
// Initialise guid and attachments.
void Asura_ClientContainerInstance_Mesh::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
													Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);
	
	const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);		

	ResetFlag(ASURA_CONTAINER_INSTANCE_MESH_FLAG_INITIALISEDPOSORI);

	// Set up the skin and animation.
	CreateAnimation(pxContainerMesh);

	if (m_pxAnimation)
	{
		if (!m_pxPhysicsObject)
		{
			CreatePhysicsObject(pxContainerMesh);
		}
		else
		{
			CopyPhysicsProperties();
		}
	}

	const bool bOwnPhysics = (pxContainerMesh && pxContainerMesh->HasOwnPhysicsObject());
	if (bOwnPhysics || !GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS))
	{
		// Not using the shared physics object.
		ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);
	}
	else
	{
		SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);
	}

	// Collect some data.
	if (m_pxPhysicsObject)
	{
		m_pxPhysicsObject->SetAssociatedEntityGuid(GetOwnerGuid());
		
		++xData.m_iNumPhysics;

		if (GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS))
		{
			// If we have our own physics object then set the unique id to be that of the mesh,
			// if we are sharing the physics object then set it to be the collection.
			u_int uID = GetUniqueID();
			if (!bOwnPhysics)
			{
				// Strip out the last component of the unique ID.
				uID >>= 8;
			}

			const bool bInitialisedPosOri = xData.AddDynamicPhysics(m_pxPhysicsObject, uID);
			if (bInitialisedPosOri)
			{
				// Pos/ori has been initialised from stored data. Set the flag so that the teleport
				// won't reset the pos/ori back to their default values.
				SetFlag(ASURA_CONTAINER_INSTANCE_MESH_FLAG_INITIALISEDPOSORI);
			}
		}
	}
	
	if (m_pxAnimation && m_pxContainer)
	{
		const u_int uRenderFlags = pxContainerMesh->GetRenderFlags();

		++xData.m_iNumRender;

		if (pxContainerMesh->HasShadow() && (uRenderFlags & ASURA_HIERARCHY_RENDER_NO_SHADOW) == 0)
		{
			++xData.m_iNumRenderShadow;
		}

		if (!pxContainerMesh->IsIgnoreForLOS())
		{
			++xData.m_iNumLOS;

			if (m_pxPhysicsObject && m_pxPhysicsObject->GetType() != ASURA_PHYSICS_OBJECT_STATIC)
			{
				++xData.m_iNumMovingLOS;
			}
		}

		if (pxContainerMesh->HasSelfIllumination())
		{ 
			++xData.m_iNumUpdate;
		}
	}
}

///////////////////////////
// SetParamFromContainer //
///////////////////////////
// Set up initial parameters based on the container.
void Asura_ClientContainerInstance_Mesh::SetParamFromContainer(const Asura_Container_Mesh* pxContainer)
{
	if (!m_pxAnimation) return;
	if (!pxContainer) return;

	const float fAlpha = pxContainer->GetAlpha();
	if (fAlpha != 1.0f)
	{
		m_pxAnimation->SetAlpha(fAlpha, true);
	}

	const u_int uRenderFlags = pxContainer->GetRenderFlags();

	if (pxContainer->IsFullBright())
	{
		m_pxAnimation->SetRGB( 0.5f, 0.5f, 0.5f, false );
	}
	else if (uRenderFlags & ASURA_HIERARCHY_RENDER_FORCE_AMBIENT_COLOUR_ONLY)
	{
		const u_int uRGB = Asura_Lighting_System::s_uEntityAmbientRGB;
        m_pxAnimation->SetRGB( ((uRGB >> 16) & 0xFF) / 255.0f, ((uRGB >> 8) & 0xFF) / 255.0f, (uRGB & 0xFF) / 255.0f );
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_Mesh::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (!m_pxAnimation) return;
	if (!m_pxContainer) return;

	if (GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED))
	{
		return; // Disabled.
	}

	const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);

	u_int uRenderFlags = pxContainerMesh->GetRenderFlags();

	if (pxContainerMesh->IsFullBright())
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_FORCE_AMBIENT_COLOUR_ONLY;
	}
	if (pxContainerMesh->IsDontRender())
	{
		uRenderFlags |= ASURA_HIERARCHY_RENDER_DONTRENDER;
	}

	float fAnimTimestep = ANIMTIME_USEGAMETIME;	

#ifdef ASURA_USE_SCREENSHOT_SYSTEM
    // (MH - 30.03.2007) if rendering hi-res screenshot must use 0 for anim timestep.
	if (Asura_ScreenShot_System::ShouldFreezeFrame())
	{
		fAnimTimestep = 0.0f;
	}
#endif // #ifdef ASURA_USE_SCREENSHOT_SYSTEM


#ifdef ASURA_TOOLS
	// Sets colour for wireframe outline for showing selection in tools.
	m_pxAnimation->SetOutlineColour(Tools_Entity_Container::GetOutlineColour());
#endif //ASURA_TOOLS

	if (m_pxPhysicsObject && GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS))
	{
		// Use physics object's position.
		m_pxAnimation->Update(m_pxPhysicsObject->GetPosition(), m_pxPhysicsObject->GetOrientation(), fAnimTimestep, uRenderFlags);
	}
	else if (pxContainerMesh->UseMovementOffset())
	{
		// Apply offset from the animation.
		Asura_Vector_3 xPos = xPosition;
		Asura_Matrix_3x3 xMat = xOrientation;

		Asura_Vector_3 xPosOffset;
		Asura_Quat xQuatOffset;
		if (m_pxAnimation->GetMovementOffset(&xPosOffset, &xQuatOffset))
		{
			Asura_Matrix_3x3 xMatOffset;
			xQuatOffset.ToMatrix(xMatOffset);

			xPos = xPos + xPosOffset*xMat;
			xMat = xMat * xMatOffset;
		}
		
		m_pxAnimation->Update(xPos, xMat, fAnimTimestep, uRenderFlags);

		// Move the physics object to match the animation compass position.
		MovePhysics(xPos, xMat);
	}
	else
	{
    	m_pxAnimation->Update(xPosition, xOrientation, fAnimTimestep, uRenderFlags);

		if (IsAttachedToMarker())
		{
			// The compass point may not be animating, but our marker point may have moved.
			MovePhysics(xPosition, xOrientation);
		}
	}
}

/////////////////
// MovePhysics //
/////////////////
// Move the physics block.
void Asura_ClientContainerInstance_Mesh::MovePhysics(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_pxPhysicsObject)
	{
		if (!xPosition.ApproximatelyEquals(m_pxPhysicsObject->GetPosition()))
		{
			m_pxPhysicsObject->SetPosition(xPosition);
		}

		if (!xOrientation.ApproximatelyEquals(m_pxPhysicsObject->GetOrientation()))
		{
			m_pxPhysicsObject->SetOrientation(xOrientation);
		}
	}
}


//////////////////
// RenderShadow //
//////////////////
void Asura_ClientContainerInstance_Mesh::RenderShadow(Asura_Shadow* pxShadow)
{
#ifdef ASURA_USE_INDEPENDENT_SHADOW_SYSTEM
	if (!m_pxAnimation) return;
	if (!m_pxContainer) return;

	if (GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED))
	{
		return; // Disabled.
	}

	const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);
	if (!pxContainerMesh->HasShadow()) return;

	const u_int uRenderFlags = pxContainerMesh->GetRenderFlags();
	if (uRenderFlags & ASURA_HIERARCHY_RENDER_NO_SHADOW) return;
	

	if (pxShadow)
	{
		// Called from a marker point attachment.
		Asura_Shadow_Instance_Animation	xShadowInstance(*pxShadow, *m_pxAnimation, uRenderFlags, false);
		
		Asura_Matrix_3x3 xOriMat;
		const Asura_Quat& xQuat = m_pxAnimation->GetOrientQuat();
		xQuat.ToMatrix(xOriMat);
		xShadowInstance.Render(*pxShadow, m_pxAnimation->GetPosition(), xOriMat);
		return;
	}
	
	// Not attached to a marker point.
	Asura_Bounding_Box	xAABB;
	xAABB.SetToPoint(m_pxAnimation->GetBoundingSphereCentre());
	xAABB.Grow(m_pxAnimation->GetCoarseBoundingSphereIncludingEffectsRadius());

#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM
	// TODO - give containers an extra Asura_Visibility_Query member for its shadow?
	if ( m_xShadowVisibilityQuery.IsAABBVisibleFromCurrentCamera( xAABB ) )
#else
	if ( Asura_Module_System::DoesBBIntersectVisibleModule(xAABB) )
#endif
	{
		Asura_Matrix_3x3 xOriMat;
		const Asura_Quat& xQuat = m_pxAnimation->GetOrientQuat();
		xQuat.ToMatrix(xOriMat);

		m_pxAnimation->RenderShadow(m_pxAnimation->GetPosition(), xOriMat, uRenderFlags);
	}

#endif //ASURA_USE_INDEPENDENT_SHADOW_SYSTEM
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_Mesh::Update()
{
	UpdateFromGraphColour();
}

///////////////////////////
// UpdateFromGraphColour //
///////////////////////////
void Asura_ClientContainerInstance_Mesh::UpdateFromGraphColour()
{
	if (!m_pxAnimation) return;
	if (!m_pxContainer) return;

	const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);

	if (!pxContainerMesh->HasSelfIllumination())
	{
		// Not using any information from GraphColour/Strobe containers for self-illumination.
		return;
	}

	if (!GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON) &&
		pxContainerMesh->IsUsingStrobeForSelfIllumination())
	{
		// Strobe is off, so we want to turn off the self illumination temporarily.
		m_pxAnimation->SetSelfIlluminationEnable(false);
		return;
	}

	Asura_RGBA xColour(1.0f, 1.0f, 1.0f, 0.0f);
	float fContribution = pxContainerMesh->GetSelfIlluminationContribution();
	float fColourise = 0.0f;

	const bool bUseFlicker = pxContainerMesh->IsUsingGraphFlickerForSelfIllumination();
	const bool bUseColour  = pxContainerMesh->IsUsingGraphColourForSelfIllumination();

	if (bUseFlicker || bUseColour)
	{
		// Get the current graph colour data.
		Asura_Container_GraphColour_Output xOutput;
		if (!GetGraphColourFromCollection(xOutput))
		{
			m_pxAnimation->SetSelfIlluminationEnable(false);
			return;
		}

		if (bUseFlicker)
		{
			fContribution *= xOutput.m_fIntensity;
		}

		if (bUseColour)
		{
			xColour.m_fR *= xOutput.m_xColour.m_fR;
			xColour.m_fG *= xOutput.m_xColour.m_fG;
			xColour.m_fB *= xOutput.m_xColour.m_fB;
			fColourise = pxContainerMesh->GetSelfIlluminationColourise();
		}
	}

	m_pxAnimation->SetSelfIllumination(fContribution, fColourise, xColour);
}


////////////////////////
// GetLineOfSightAABB //
////////////////////////
bool Asura_ClientContainerInstance_Mesh::GetLineOfSightAABB( Asura_Bounding_Box& xAABB )
{
	if ( m_pxAnimation )
	{
		Asura_Matrix_3x3 xOriMat;
		const Asura_Quat& xQuat = m_pxAnimation->GetOrientQuat();
		xQuat.ToMatrix(xOriMat);

		m_pxAnimation->GetCoarseBoundingBox( xAABB, m_pxAnimation->GetPosition(), xOriMat );
		return true;
	}

	return false;
}

/////////////
// LOSTest //
/////////////
bool Asura_ClientContainerInstance_Mesh::LOSTest(const Asura_LOS_InputData* pxInput, Asura_LOS_OutputData* pxOutput) const
{
	if (!m_pxAnimation || !m_pxContainer)
	{
		return false;
	}

	const Asura_Container_Mesh* const pxContainerMesh = static_cast<const Asura_Container_Mesh*>(m_pxContainer);
	if (pxContainerMesh->IsIgnoreForLOS())
	{
		return false;
	}
		
	Asura_Matrix_3x3 xOriMat;
	const Asura_Quat& xQuat = m_pxAnimation->GetOrientQuat();
	xQuat.ToMatrix(xOriMat);

	if (m_pxAnimation->LOSTest(m_pxAnimation->GetPosition(), xOriMat, pxInput, pxOutput))
	{
		// Set attach ID, to identify the container instance.
		pxOutput->m_iHitAttachmentID = static_cast<int>(GetUniqueID());
		return true;
	}
	
	return false;
}

/////////////////////
// FindMarkerPoint //
/////////////////////
// Search for the specified marker point. If found then set up the attachment and return true.
bool Asura_ClientContainerInstance_Mesh::FindMarkerPoint(Asura_ClientContainerInstance_Attachment* pxAttachment) const
{
	if (pxAttachment && m_pxAnimation)
	{	
		const Asura_Hierarchy_Skin* const pxSkin = m_pxAnimation->GetSkin();
		if (pxSkin)
		{
			const Asura_Hash_ID uMarkerHash = pxAttachment->GetMarkerPointHash();

			int iMarkerIndex = -1;
			const bool bFound = pxSkin->GetMarkerPointIndex(uMarkerHash, iMarkerIndex);
			if (bFound)
			{
				Asura_AnimationAttachment_ContainerInstance_Internal* pxAnimAttach = new Asura_AnimationAttachment_ContainerInstance_Internal(pxAttachment);
				pxAttachment->SetMarkerData(this, iMarkerIndex, pxAnimAttach);
				pxAnimAttach->LinkAttachment(m_pxAnimation, uMarkerHash, true);
			}

			return bFound;
		}
	}

	return false;
}

//////////////////////////////
// SetAttachedToMarkerPoint //
//////////////////////////////
// Mesh is attached to a marker point, hence it can't have physics enabled.
void Asura_ClientContainerInstance_Mesh::SetAttachedToMarkerPoint()
{
	SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_ATTACHED_TO_MARKER);

	if (m_pxPhysicsObject)
	{
		m_pxPhysicsObject->ChangeType(ASURA_PHYSICS_OBJECT_EXTERNALCONTROL);
		ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DYNAMIC_PHYSICS);
		ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);
	}
}


///////////////////////
// GetBoundingSphere //
///////////////////////
bool Asura_ClientContainerInstance_Mesh::GetTightBoundingSphere( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxAnimation)
	{
		const float fTempRadius = m_pxAnimation->GetTightBoundingSphereRadius();
		if (fTempRadius > 0.0f)
		{
			fRadius	= fTempRadius;
			xCentre	= m_pxAnimation->GetBoundingSphereCentre();
			return true;
		}
	}
	return false;
}

bool Asura_ClientContainerInstance_Mesh::GetCoarseBoundingSphere( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxAnimation)
	{
		const float fTempRadius = m_pxAnimation->GetCoarseBoundingSphereRadius();
		if (fTempRadius > 0.0f)
		{
			fRadius	= fTempRadius;
			xCentre	= m_pxAnimation->GetBoundingSphereCentre();
			return true;
		}
	}
	return false;
}

//////////////////
// ApplyImpulse //
//////////////////
// Apply a linear impulse to a point on the object (position in world-space).
void Asura_ClientContainerInstance_Mesh::ApplyImpulse(const Asura_Vector_3& xImpulse,
													  const Asura_Vector_3& xPosition,
													  bool bIgnoreMass /*= true*/ )
{
	if (!Asura_Network::IsServer()) return;

	if (m_pxPhysicsObject && m_pxPhysicsObject->GetType() == ASURA_PHYSICS_OBJECT_DYNAMIC)
	{
		m_pxPhysicsObject->ApplyImpulse(xImpulse, xPosition, bIgnoreMass);
	}
}

/////////////////////////////
// FindFromRigidBodyHandle //
/////////////////////////////
// Search for the container that has the given rigid body handle.
Asura_ClientContainerInstance* Asura_ClientContainerInstance_Mesh::FindFromRigidBodyHandle(Asura_Handle iHandle)
{
	if (m_pxPhysicsObject && m_pxPhysicsObject->GetHandle() == iHandle)
	{
		return this;
	}
	return 0;
}

////////////////////////////////
// GetBasePosOriForAttachment //
////////////////////////////////
// Returns true, if the position and orientation were set.
bool Asura_ClientContainerInstance_Mesh::GetBasePosOriForAttachment(Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientation) const
{
	if (m_pxPhysicsObject && GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS))
	{
		xPosition = m_pxPhysicsObject->GetPosition();
		xOrientation = m_pxPhysicsObject->GetOrientation();
		return true;
	}

	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_SoundEvent
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_SoundEvent::Asura_ClientContainerInstance_SoundEvent(const Asura_Container_SoundEvent* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_iSoundEventHandle(ASURA_INVALID_HANDLE)
{
	if (Asura_Container_System::IsDoingCreateOrActivate() && pxContainer->IsTransitionOnly())
	{
		// Container entity is being created/loaded or activated. Not a transition, so disable.
		// This is so damage sounds (e.g. smashing glass) don't get played on loading checkpoints.
		SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED);
	}
}

Asura_ClientContainerInstance_SoundEvent::~Asura_ClientContainerInstance_SoundEvent()
{
	const float fFadeOutTime = 0.25f;
	Asura_Sound_Event_System::Stop(m_iSoundEventHandle, fFadeOutTime);
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_SoundEvent::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	if (!m_pxContainer) return;

	if (GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED))
	{
		return; // Disabled.
	}

	// Start the sound event if necessary, otherwise just update the position.
	if (m_iSoundEventHandle == ASURA_INVALID_HANDLE)
	{
		const Asura_Container_SoundEvent* pxContainerSE = static_cast<const Asura_Container_SoundEvent*>(m_pxContainer);
		const Asura_Hash_ID uSEHash = pxContainerSE->GetSoundEventHash();
		if (uSEHash != ASURA_HASH_ID_UNSET)
		{
			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::FindEvent(uSEHash);
			Asura_Assertf(pxEvent, ("Failed to find sound event (0x%08x) for container entity %u", uSEHash, GetOwnerGuid()));
			if (pxEvent)
			{
				Asura_Sound_Event_System::PlayEvent(pxEvent, &m_iSoundEventHandle, &xPosition, 1.0f, GetOwnerGuid() );
			}
		}
	}
	else
	{
		Asura_Sound_Event_System::UpdatePosition(m_iSoundEventHandle, xPosition);
	}
}

void Asura_ClientContainerInstance_SoundEvent::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_iSoundEventHandle != ASURA_INVALID_HANDLE)
	{
		if (IsAttachedToMarker())
		{
			// Sound events can only move if they are attached to the marker point of a mesh that is moving.
			Asura_Sound_Event_System::UpdatePosition(m_iSoundEventHandle, xPosition);
		}

		// Render if needed in Tools.
		ToolsRender(xPosition, xOrientation);
	}
}

// Initialise guid and attachments.
void Asura_ClientContainerInstance_SoundEvent::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
														  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	// Only need to call Render for sound events if they might be moving.
	if (IsAttachedToMarker())
	{
		++xData.m_iNumRender;
	}
}

void Asura_ClientContainerInstance_SoundEvent::Update()
{
}


///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Collection
///////////////////////////////////////////////////////////////////////////////

Asura_ClientContainerInstance_Collection::Asura_ClientContainerInstance_Collection(const Asura_Container_Collection* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_uNumChildren(0)
, m_ppxChildren(0)
, m_uOwnerGuid(ASURA_GUID_UNREGISTERED)
{
	m_xInputState.Reset();

	if (pxContainer)
	{
		m_uNumChildren = pxContainer->GetNumChildren();
		if (m_uNumChildren)
		{
			// Create a blank child array, and then fill out all of the health and collection containers.
			// We need these so that we can set up the health values from the server, before we
			// call ChangeState(). These health values are needed for the damage overrides, which
			// modify the collection states during the ChangeState() call.
			m_ppxChildren = new Asura_ClientContainerInstance*[m_uNumChildren];

			for (u_int u=0; u<m_uNumChildren; ++u)
			{
				const Asura_Hash_ID uChildContainerHash = pxContainer->GetChildContainerHash(u);
	
				// Create health containers and collections recursively.
				m_ppxChildren[u] = Asura_Container_System::CreateClientInstanceBase(uChildContainerHash);
			}
		}
	}
}

Asura_ClientContainerInstance_Collection::~Asura_ClientContainerInstance_Collection()
{
	DeleteAllChildren();
}

/////////////////
// ChangeState //
/////////////////
// Change to a new state. Create or delete children as necessary.
void Asura_ClientContainerInstance_Collection::ChangeState(const Asura_Container_InputState& xInputState)
{
	// Save the state.
	m_xInputState = xInputState;

	// Child array should already be set up.
	if (m_uNumChildren>0 && m_ppxChildren && m_pxContainer)
	{
		// Find the relevant state applying any damage overrides.
		const Asura_ClientContainerInstance_Health* const pxHealth = FindHealth();
		const float fHealthPercent = (pxHealth ? pxHealth->GetHealthPercent() : 100.0f);
		
		const Asura_Container_Collection* const pxCollection = static_cast<const Asura_Container_Collection*>(m_pxContainer);
		
		Asura_Assert(pxCollection->GetNumInputs() == xInputState.GetNumInputs(), "Mismatch in number of state inputs");

		// Apply any damage overrides.
		Asura_Container_OutputState xOutputState;
		pxCollection->ConvertState(xInputState, fHealthPercent, xOutputState);

		// Marker attachments are at the end of the array and need to be changed first in
		// order to allow them to unlink themselves from the relevant mesh.
		// So iterate over the array in reverse order.
		for (int i = static_cast<int>(m_uNumChildren) - 1; i >= 0; --i) 
		{
			const u_int u = static_cast<u_int>(i);
			const Asura_Hash_ID uChildContainerHash = pxCollection->GetChildContainerHash(u);
			
			Asura_Container_InputState xChildInputState;
			xOutputState.GetChildInputState(uChildContainerHash, xChildInputState);

			const bool bTurnOff = xChildInputState.IsOff();

			if (m_ppxChildren[u])
			{
				if ( bTurnOff )
				{
					// Don't destroy collections or health containers.
					const u_int uClass = m_ppxChildren[u]->GetClassificationSkipAttachment();
					if (!Asura_Container_System::IsBaseClassification(uClass) )
					{
						// Child has been switched off, delete it.
						// Note: We better zero out the child before destroying it, as the destruction
						// might initiate an iteration over this collection's children.
						Asura_ClientContainerInstance* const pxToDestroy = m_ppxChildren[u];
						m_ppxChildren[u] = 0;
						Asura_Container_System::DestroyInstance(pxToDestroy);
					}
				}
			}
			else if (!bTurnOff)
			{
				// Child doesn't exist, create if necessary.
				m_ppxChildren[u] = Asura_Container_System::CreateClientInstance(uChildContainerHash);
			}

			// Also call change state on children.
			if (m_ppxChildren[u])
			{
				m_ppxChildren[u]->ChangeState(xChildInputState);
			}
		}

		RefreshCachedChildData();
	}
}


////////////////
// FindHealth //
////////////////
// Find the relevant health container.
const Asura_ClientContainerInstance_Health* Asura_ClientContainerInstance_Collection::FindHealth() const
{
	if (!m_pxContainer) return 0;
	const Asura_Container_Collection* const pxCollection = static_cast<const Asura_Container_Collection* const>(m_pxContainer);
	const int iIndex = pxCollection->FindHealthIndex();

	if (iIndex >= 0 &&  static_cast<u_int>(iIndex) < m_uNumChildren)
	{
		const Asura_ClientContainerInstance* pxHealth = m_ppxChildren[iIndex];
		if (pxHealth && pxHealth->GetClassification() == AsuraContainerClass_Health)
		{
			return static_cast<const Asura_ClientContainerInstance_Health*>(pxHealth);
		}

		Asura_Assert(pxHealth, "Couldn't find the health container.");
	}

	return 0;
}

//////////////////////////////
// HasDamageOverrideChanged //
//////////////////////////////
// Return true if the damage override state will change as a result of the change in health.
bool Asura_ClientContainerInstance_Collection::HasDamageOverrideChanged(float fOldHealthPercent, float fNewHealthPercent) const
{
	if (!m_pxContainer) return false;

	const Asura_Container_Collection* const pxCollection = static_cast<const Asura_Container_Collection*>(m_pxContainer);
	return pxCollection->HasDamageOverrideChanged(m_xInputState, fOldHealthPercent, fNewHealthPercent);
}

///////////////
// SetStrobe //
///////////////
// Strobe only affects the children within its parent collection, it doesn't affect sub-collections.
void Asura_ClientContainerInstance_Collection::SetStrobe(bool bStrobeOn)
{
	PARENT::SetStrobe(bStrobeOn);

	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return;
	}
	
	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		Asura_ClientContainerInstance* pxChild = ConvertChild(m_ppxChildren[u]);
		if (pxChild)
		{
			pxChild->SetStrobe(bStrobeOn);
		}
	}
}

///////////////////////
// DeleteAllChildren //
///////////////////////
void Asura_ClientContainerInstance_Collection::DeleteAllChildren()
{
	if (m_ppxChildren)
	{
		// Marker attachments are at the end of the array and need to be deleted first in
		// order to allow them to unlink themselves from the relevant mesh.
		// So iterate over the array in reverse order.
		for (int i = static_cast<int>(m_uNumChildren) - 1; i >= 0; --i) 
		{
			// Note: We better zero out the child before destroying it, as the destruction
			// might initiate an iteration over this collection's children.
			Asura_ClientContainerInstance* const pxToDestroy = m_ppxChildren[i];
			m_ppxChildren[i] = 0;
			Asura_Container_System::DestroyInstance(pxToDestroy);
		}

		delete [] m_ppxChildren;
		m_ppxChildren = 0;
	}

	m_uNumChildren = 0;
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_Collection::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			m_ppxChildren[u]->TeleportTo(xPosition, xOrientation);
		}
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_Collection::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			m_ppxChildren[u]->Render(xPosition, xOrientation);
		}
	}
}

//////////////////
// RenderShadow //
//////////////////
void Asura_ClientContainerInstance_Collection::RenderShadow(Asura_Shadow* pxShadow)
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			m_ppxChildren[u]->RenderShadow(pxShadow);
		}
	}
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_Collection::Update()
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			m_ppxChildren[u]->Update();
		}
	}
}

////////////////////////
// GetLineOfSightAABB //
////////////////////////
// Returns the conservative axially-aligned WORLD SPACE box covering all of the container's children.
// Return true if the AABB was set.
bool Asura_ClientContainerInstance_Collection::GetLineOfSightAABB( Asura_Bounding_Box& xAABB )
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return false;
	}

	bool bSetAABB = false;
	
	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			Asura_Bounding_Box xTemp;
			if ( m_ppxChildren[u]->GetLineOfSightAABB(bSetAABB ? xTemp : xAABB) )
			{
				if (bSetAABB)
				{
					// Add the new bounding box.
					xAABB.Set(xTemp);
				}
				bSetAABB = true;
			}
		}
	}

	return bSetAABB;
}

/////////////
// LOSTest //
/////////////
// Do LOS test on all the children. Return true if the output data is changed.
bool Asura_ClientContainerInstance_Collection::LOSTest(const Asura_LOS_InputData* pxInput, Asura_LOS_OutputData* pxOutput) const
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return false;
	}

	bool bHit = false;
	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			if (m_ppxChildren[u]->LOSTest(pxInput, pxOutput))
			{
				bHit = true;
			}
		}
	}

	return bHit;
}

////////////////
// Initialise //
////////////////
void Asura_ClientContainerInstance_Collection::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
														  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);
	m_uOwnerGuid = xData.GetOwnerGuid();
	ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);

	// Root collection has ID of zero.
	if (!pxParentCollection)
	{
		m_uUniqueID = 0;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			// Note: This limits the instance to a depth of 4 not including the root collection.
			// It also limits the number of children in each to be 255.
			// If this isn't enough we'll need to rethink the method of generating the unique id. 
			u_int uNextID = ((m_uUniqueID<<8) + u + 1);
			if ((m_uUniqueID & 0xff000000) != 0 || u >= 0xff)
			{
				uNextID = uASURA_CONTAINER_UNIQUE_ID_INVALID;
				Asura_Assertf(0, ("Invalid Unique ID generated for container instance (0x%08x), entity %u.", GetContainerHash(), GetOwnerGuid()));
			}

			m_ppxChildren[u]->SetUniqueID(uNextID);
			m_ppxChildren[u]->Initialise(this, xData);

			// Cache the index for the mesh that has shared dynamic physics (there should only be one).
			const Asura_ClientContainerInstance* const pxConverted = ConvertChild(m_ppxChildren[u]);
			if (pxConverted && pxConverted->GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS))
			{
				SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);
				if (u < ucASURA_CONTAINER_MAX_CACHED_INDEX)
				{
					m_xCachedChildData.m_ucSharedPhysics = static_cast<u_char>(u);
				}
			}
		}
	}
}

//////////////////////
// FindFromUniqueID //
//////////////////////
// Find instance from unique ID.
Asura_ClientContainerInstance* Asura_ClientContainerInstance_Collection::FindFromUniqueID(u_int uUniqueID)
{
	if (uUniqueID == uASURA_CONTAINER_UNIQUE_ID_INVALID) return 0;
	if (uUniqueID == 0) return this;

	while ((uUniqueID & 0xff000000) == 0)
	{
		uUniqueID <<= 8;
	}

	const u_int uChild = ((uUniqueID & 0xff000000) >> 24) - 1;
	if (uChild >= 0xff || uChild >= m_uNumChildren)
	{
		Asura_Assert(0, "Invalid child ID");
		return 0;
	}

	if (m_ppxChildren[uChild])
	{
		return m_ppxChildren[uChild]->FindFromUniqueID(uUniqueID<<8);
	}

	return 0;
}

/////////////////////
// FindMarkerPoint //
/////////////////////
// Search for the specified marker point. If found then set up the attachment and return true.
bool Asura_ClientContainerInstance_Collection::FindMarkerPoint(Asura_ClientContainerInstance_Attachment* pxAttachment) const
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return false;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			const bool bFound = m_ppxChildren[u]->FindMarkerPoint(pxAttachment);
			if (bFound) return true;
		}
	}

	return false;
}

//////////////////////////////
// SetAttachedToMarkerPoint //
//////////////////////////////
void Asura_ClientContainerInstance_Collection::SetAttachedToMarkerPoint()
{
	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			m_ppxChildren[u]->SetAttachedToMarkerPoint();
		}
	}
}

///////////////////////
// GetBoundingSphere //
///////////////////////

bool Asura_ClientContainerInstance_Collection::GetTightBoundingSphere( float& fRadius, Asura_Vector_3& xCentre ) const
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return false;
	}

	bool bSet = false;

	float fNewRadius;
	Asura_Vector_3 xNewCentre;

	for (u_int u=0; u<m_uNumChildren; ++u)
	{		
		if (m_ppxChildren[u] && m_ppxChildren[u]->GetTightBoundingSphere(fNewRadius, xNewCentre))
		{
			if (!bSet)
			{
				// For first sphere, just copy the data.
				fRadius = fNewRadius;
				xCentre = xNewCentre;
				bSet = true;
			}
			else
			{
				// Combine with previous spheres.
				Asura_Vector_3 xDiff(xNewCentre - xCentre);
				const float	fCentreDiff = xDiff.ApproximatelyEquals( Asura_Maths::ZeroVector3 ) ? 0.0f : xDiff.Normalise();
				const float	fMaxRadius = fCentreDiff + fNewRadius;

				if (fMaxRadius > fRadius)
				{
					const float fCentreDisp = ( fMaxRadius - fRadius ) * 0.5f;
					xCentre += xDiff * fCentreDisp;
					fRadius += fCentreDisp;
				}
			}
		}
	}

	return bSet;
}

bool Asura_ClientContainerInstance_Collection::GetCoarseBoundingSphere( float& fRadius, Asura_Vector_3& xCentre ) const
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return false;
	}

	bool bSet = false;

	float fNewRadius;
	Asura_Vector_3 xNewCentre;

	for (u_int u=0; u<m_uNumChildren; ++u)
	{		
		if (m_ppxChildren[u] && m_ppxChildren[u]->GetCoarseBoundingSphere(fNewRadius, xNewCentre))
		{
			if (!bSet)
			{
				// For first sphere, just copy the data.
				fRadius = fNewRadius;
				xCentre = xNewCentre;
				bSet = true;
			}
			else
			{
				// Combine with previous spheres.
				Asura_Vector_3 xDiff(xNewCentre - xCentre);
				const float	fCentreDiff = xDiff.ApproximatelyEquals( Asura_Maths::ZeroVector3 ) ? 0.0f : xDiff.Normalise();
				const float	fMaxRadius = fCentreDiff + fNewRadius;

				if (fMaxRadius > fRadius)
				{
					const float fCentreDisp = ( fMaxRadius - fRadius ) * 0.5f;
					xCentre += xDiff * fCentreDisp;
					fRadius += fCentreDisp;
				}
			}
		}
	}

	return bSet;
}

bool Asura_ClientContainerInstance_Collection::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return false;
	}

	bool bSet = false;

	float fNewRadius;
	Asura_Vector_3 xNewCentre;

	for (u_int u=0; u<m_uNumChildren; ++u)
	{		
		if (m_ppxChildren[u] && m_ppxChildren[u]->GetCoarseBoundingSphereIncludingEffects(fNewRadius, xNewCentre))
		{
			if (!bSet)
			{
				// For first sphere, just copy the data.
				fRadius = fNewRadius;
				xCentre = xNewCentre;
				bSet = true;
			}
			else
			{
				// Combine with previous spheres.
				Asura_Vector_3 xDiff(xNewCentre - xCentre);
				const float	fCentreDiff = xDiff.ApproximatelyEquals( Asura_Maths::ZeroVector3 ) ? 0.0f : xDiff.Normalise();
				const float	fMaxRadius = fCentreDiff + fNewRadius;

				if (fMaxRadius > fRadius)
				{
					const float fCentreDisp = ( fMaxRadius - fRadius ) * 0.5f;
					xCentre += xDiff * fCentreDisp;
					fRadius += fCentreDisp;
				}
			}
		}
	}

	return bSet;
}

// Constants
const u_int uMAX_NUM_CACHED_GRAPHCOLOURS = 4;

////////////////////////////
// RefreshCachedChildData //
////////////////////////////
// Called after creation and after state changes (which can create/destroy children).
// Cache some information about the children to speed up some frequent operations.
void Asura_ClientContainerInstance_Collection::RefreshCachedChildData()
{
	// Clear out old data.
	m_xCachedChildData.Reset();

	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		// No graph colours.
		m_xCachedChildData.m_ucNumGraphColours = 0;
		return;
	}

	bool bFailed = false;

	// Count the number of enabled GraphColours and store the indices of up to 3 of them.
	u_int uNumGraphs = 0;

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		// Examine the children in the collection, skipping over attachments and ignoring sub-collections.
		const Asura_ClientContainerInstance* pxChild = ConvertChild(m_ppxChildren[u]);
		if (!pxChild) continue;
		const u_int uClassification = pxChild->GetClassification();

		if (uClassification == AsuraContainerClass_GraphColour)
		{
			if (u > ucASURA_CONTAINER_MAX_CACHED_INDEX)
			{
				bFailed = true;
				break;
			}

			if (uNumGraphs >= uMAX_NUM_CACHED_GRAPHCOLOURS)
			{
				// Too many colour graphs. Carry on looking for physics containers.
				bFailed = true;
				continue;
			}
			
			m_xCachedChildData.m_aucGraph[uNumGraphs] = static_cast<u_char>(u);
			++uNumGraphs;
		}
		else if (uClassification == AsuraContainerClass_Physics)
		{
			if (u < ucASURA_CONTAINER_MAX_CACHED_INDEX)
			{
				m_xCachedChildData.m_ucPhysics = static_cast<u_char>(u);
			}
		}
	}

	if (bFailed)
	{
		// Too many GraphColours active (more than 4), or too many children (more than 255)!
		m_xCachedChildData.ResetColourGraphs();
	}
	else
	{
 		m_xCachedChildData.m_ucNumGraphColours = static_cast<u_char>(uNumGraphs);
	}
}

////////////////////
// GetGraphColour //
////////////////////
// Get the current graph colour. Caller should initialise the output data before calling.
// Returns true if results have been returned (i.e there are enabled GraphColour containers).
bool Asura_ClientContainerInstance_Collection::GetGraphColour(Asura_Container_GraphColour_Output& xOutput)
{
	const u_int uNumGraphColours = m_xCachedChildData.m_ucNumGraphColours;
	if (uNumGraphColours == 0)
	{
		// No enabled GraphColours.
		return false;
	}

	// Note: we don't break out of the loop when we find a match, as there may be several
	// GraphColour containers active at once. We multiply their output data together.
	bool bFound = false;
	
	if (uNumGraphColours > uMAX_NUM_CACHED_GRAPHCOLOURS)
	{
		// No valid cached data, just do the slow linear search.
		for (u_int u=0; u<m_uNumChildren; ++u)
		{
			Asura_ClientContainerInstance* pxChild = ConvertChild(m_ppxChildren[u]);
			if (pxChild && pxChild->GetGraphColour(xOutput))
			{
				bFound = true;
			}
		}
	}
	else
	{
		// Faster version that goes straight to the GraphColours.
		for (u_int u=0; u<uNumGraphColours; ++u)
		{
			const u_int uIndex = m_xCachedChildData.m_aucGraph[u];
			Asura_Assertf(uIndex<m_uNumChildren, ("Out of range index (%u)", uIndex));

			Asura_ClientContainerInstance* pxChild = ConvertChild(m_ppxChildren[uIndex]);
			if (pxChild && pxChild->GetGraphColour(xOutput))
			{
				bFound = true;
			}
		}
	}

	return bFound;
}

/////////////////
// FindPhysics //
/////////////////
// Find the relevant physics properties.
const Asura_ClientContainerInstance_Physics* Asura_ClientContainerInstance_Collection::FindPhysics() const
{
	const u_int uIndex = m_xCachedChildData.m_ucPhysics;
	
	if (uIndex != ucASURA_CONTAINER_MAX_CACHED_INDEX && uIndex < m_uNumChildren)
	{
		Asura_ClientContainerInstance* pxChild = ConvertChild(m_ppxChildren[uIndex]);
		if (pxChild && pxChild->GetClassification() == AsuraContainerClass_Physics)
		{
			return static_cast<Asura_ClientContainerInstance_Physics*>(pxChild);
		}
	}
	return 0;
}

///////////////////
// FindBlueprint //
///////////////////
// Find the blueprint in the collection.
// Set bIncludeDisabled to include both enabled and disabled blueprint containers.
// If uBlueprintTypeHash != ASURA_HASH_ID_UNSET, then only return blueprints of the specified type.
const Asura_Container_Blueprint* Asura_ClientContainerInstance_Collection::FindBlueprint(Asura_Hash_ID uBlueprintTypeHash,
																						 bool bIncludeDisabled) const
{	
	if (!m_pxContainer) return 0;
	const Asura_Container_Collection* const pxCollection = static_cast<const Asura_Container_Collection*>(m_pxContainer);

	if (bIncludeDisabled)
	{
		// Just look for any blueprint, we don't care whether it is enabled or not.
		return pxCollection->FindBlueprintContainer(uBlueprintTypeHash);
	}

	u_int uFirst = 0;
	u_int uLast  = 0;

	if (!pxCollection->FindBlueprintIndices(uFirst, uLast) || m_ppxChildren==0)
	{
		return 0; // No blueprints.
	}

	// Search for any enabled blueprints (a non-null pointer means it is enabled).
	for (u_int u = uFirst; u <= uLast; ++u)
	{
		if (m_ppxChildren[u] && m_ppxChildren[u]->GetClassification() == AsuraContainerClass_Blueprint)
		{
			const Asura_ClientContainerInstance_Blueprint* const pxBPInstance = static_cast<const Asura_ClientContainerInstance_Blueprint*>(m_ppxChildren[u]);
			if (uBlueprintTypeHash == ASURA_HASH_ID_UNSET ||
				uBlueprintTypeHash == pxBPInstance->GetBlueprintTypeHash())
			{
				return pxBPInstance->GetBlueprintContainer();
			}
		}
	}

	return 0;
}

/////////////////////////////
// FindFromRigidBodyHandle //
/////////////////////////////
// Search for the container that has the given rigid body handle.
Asura_ClientContainerInstance* Asura_ClientContainerInstance_Collection::FindFromRigidBodyHandle(Asura_Handle iHandle)
{
	if (iHandle == ASURA_INVALID_HANDLE) return 0;

	// If the container is off, then the only possible children are collections, attachments and health containers.
	if (IsOff())
	{
		return 0;
	}

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			Asura_ClientContainerInstance* pxResult = m_ppxChildren[u]->FindFromRigidBodyHandle(iHandle);
			if (pxResult)
			{
				return pxResult;
			}
		}
	}
	return 0;
}

////////////////////////////////
// GetBasePosOriForAttachment //
////////////////////////////////
// Returns true, if the position and orientation were set.
bool Asura_ClientContainerInstance_Collection::GetBasePosOriForAttachment(Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientation) const
{
	const u_int uIndex = m_xCachedChildData.m_ucSharedPhysics;
	
	if (uIndex != ucASURA_CONTAINER_MAX_CACHED_INDEX && uIndex < m_uNumChildren)
	{
		Asura_ClientContainerInstance* pxChild = ConvertChild(m_ppxChildren[uIndex]);
		if (pxChild && pxChild->GetClassification() == AsuraContainerClass_Mesh)
		{
			const Asura_ClientContainerInstance_Mesh* const pxMesh = static_cast<Asura_ClientContainerInstance_Mesh*>(pxChild);
			return pxMesh->GetBasePosOriForAttachment(xPosition, xOrientation);
		}
	}

	return false;
}

//////////////
// Traverse //
//////////////
void Asura_ClientContainerInstance_Collection::Traverse(Asura_Container_Functor& xFunctor) const
{
	xFunctor.Process(this);

	for (u_int u=0; u<m_uNumChildren; ++u)
	{
		if (m_ppxChildren[u])
		{
			m_ppxChildren[u]->Traverse(xFunctor);
		}
	}
}

////////////////////
// StoreAnimTimer //
////////////////////
// Store animation timer. Mainly needed for state changes to damaged meshes, where the mesh is animated.
void Asura_ClientContainerInstance_Collection::StoreAnimTimer(Asura_Hash_ID uAnimHash, float fTimer)
{
	if (uAnimHash != ASURA_HASH_ID_UNSET)
	{
		#ifdef ASURA_ASSERT
		const bool bStored =
		#endif
		m_xCachedAnimTimers.StoreAnimTimer(uAnimHash, fTimer);

		Asura_Assert(bStored, "Couldn't store the animation timer in the container collection.");
	}
}

//////////////////////
// RestoreAnimTimer //
//////////////////////
// Retore animation timer. Mainly needed for state changes to damaged meshes, where the mesh is animated.
float Asura_ClientContainerInstance_Collection::RestoreAnimTimer(Asura_Hash_ID uAnimHash, bool bRestore, bool bRandomise)
{
	if (uAnimHash != ASURA_HASH_ID_UNSET)
	{
		// Search for the anim, if found then restore the timer.
		if (bRestore)
		{
			float fReturnTimer = 0.0f;
			if (m_xCachedAnimTimers.GetAnimTimer(uAnimHash, fReturnTimer))
			{
				return fReturnTimer;
			}
		}
		
		if (bRandomise)
		{
			// Return a random value in range [0,1].
			return Asura_Random::GetFloat();
		}
	}

	// Start anim at the beginning.
	return 0.0f;
}


///////////////////////////////////////////////////////////////////////////////
// CachedAnimTimers
///////////////////////////////////////////////////////////////////////////////

// Reset
void Asura_ClientContainerInstance_Collection::CachedAnimTimers::Reset()
{
	for (u_int u=0; u<uASURA_CONTAINER_MAX_STORED_ANIM_TIMERS; ++u)
	{
		m_auAnimHash[u] = ASURA_HASH_ID_UNSET;
		m_afParametricTimer[u] = 0.0f;
	}
}

// GetAnimTimer
// Search for animation/timer. Return true if found.
bool Asura_ClientContainerInstance_Collection::CachedAnimTimers::GetAnimTimer(Asura_Hash_ID uAnimHash, float& fTimer)
{
	for (u_int u=0; u<uASURA_CONTAINER_MAX_STORED_ANIM_TIMERS; ++u)
	{
		if (m_auAnimHash[u] == uAnimHash)
		{
			fTimer = m_afParametricTimer[u];
			return true;
		}
	}
	return false;
}

// StoreAnimTimer
// Store animation/timer. Return true if stored.
bool Asura_ClientContainerInstance_Collection::CachedAnimTimers::StoreAnimTimer(Asura_Hash_ID uAnimHash, float fTimer)
{
	for (u_int u=0; u<uASURA_CONTAINER_MAX_STORED_ANIM_TIMERS; ++u)
	{
		if (m_auAnimHash[u] == uAnimHash ||
			m_auAnimHash[u] == ASURA_HASH_ID_UNSET)
		{
			m_auAnimHash[u] = uAnimHash;
			m_afParametricTimer[u] = fTimer;
			return true;
		}
	}
	return false;
}


///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Attachment
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_Attachment::Asura_ClientContainerInstance_Attachment(const Asura_Container_Attachment* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_pxChild(0)
, m_pxParentMeshForMarker(0)
, m_iMarkerIndex(0)
, m_pxAnimAttachment(0)
{
	// Attachments just pass the state on to their child.
	CreateChild();
}

// Note: that deletion of marker attachments, must be done before their parent mesh is deleted.
// To achieve this marker attachments are put at the end of collections, which are deleted in reverse order.
Asura_ClientContainerInstance_Attachment::~Asura_ClientContainerInstance_Attachment()
{
	UnlinkFromParent();
	DeleteChild();
}

//////////////////////
// UnlinkFromParent //
//////////////////////
void Asura_ClientContainerInstance_Attachment::UnlinkFromParent()
{
	// Unlink from the parent mesh.
	if (m_pxAnimAttachment)
	{
		m_pxAnimAttachment->UnlinkAttachment();
		m_pxAnimAttachment = 0;
	}

	if (m_pxParentMeshForMarker)
	{	
		m_pxParentMeshForMarker = 0;
		m_iMarkerIndex = 0;
	}
}

/////////////////
// ChangeState //
/////////////////
// Change to a new state.
void Asura_ClientContainerInstance_Attachment::ChangeState(const Asura_Container_InputState& xInputState)
{
	// Unlink from parent as it might be about to be deleted, if it doesn't exist in this new state.
	// State changes are done in reverse order in collections, so that marker attachments will
	// have ChangeState() calls before their parent meshes are deleted.
	UnlinkFromParent();

	if (!m_pxChild && m_pxContainer && GetMarkerPointHash() != ASURA_HASH_ID_UNSET)
	{
		// There is no child. It may have been deleted as the result of a failed marker attachment.
		// Recreate it if needed.
		CreateChild();
	}

	if (m_pxChild)
	{
		m_pxChild->ChangeState(xInputState);
	}
}

/////////////////
// CreateChild //
/////////////////
// Create child if it doesn't already exist.
void Asura_ClientContainerInstance_Attachment::CreateChild()
{
	if (!m_pxChild && m_pxContainer)
	{
		const Asura_Container_Attachment* const pxContainer = static_cast<const Asura_Container_Attachment*>(m_pxContainer);
		const Asura_Hash_ID uChildContainerHash = pxContainer->GetChildContainerHash();
		m_pxChild = Asura_Container_System::CreateClientInstance(uChildContainerHash);
	}
}

/////////////////
// DeleteChild //
/////////////////
void Asura_ClientContainerInstance_Attachment::DeleteChild()
{
	// Note: We better zero out the child before destroying it, as the destruction
	// might initiate a call to this attachment's child.
	Asura_ClientContainerInstance* const pxToDestroy = m_pxChild;
	m_pxChild = 0;
	Asura_Container_System::DestroyInstance(pxToDestroy);
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_Attachment::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_pxChild && m_pxContainer)
	{
		Asura_Vector_3 xNewPos;
		Asura_Matrix_3x3 xNewOri;
		ApplyOffset(xPosition, xOrientation, xNewPos, xNewOri);
		m_pxChild->TeleportTo(xNewPos, xNewOri);
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_Attachment::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_pxParentMeshForMarker)
	{
		// We are relying on the animation container attachment to call Render on the child.
		return;
	}

	if (m_pxChild && m_pxContainer)
	{
		Asura_Vector_3 xNewPos;
		Asura_Matrix_3x3 xNewOri;
		ApplyOffset(xPosition, xOrientation, xNewPos, xNewOri);
		m_pxChild->Render(xNewPos, xNewOri);
	}
}

//////////////////
// RenderShadow //
//////////////////
void Asura_ClientContainerInstance_Attachment::RenderShadow(Asura_Shadow* pxShadow)
{
	if (m_pxParentMeshForMarker)
	{
		// We are relying on the animation container attachment to call RenderShadow on the child.
		return;
	}

	if (m_pxChild)
	{
		m_pxChild->RenderShadow(pxShadow);
	}
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_Attachment::Update()
{
	if (m_pxChild)
	{
		m_pxChild->Update();
	}
}

////////////////////////
// GetLineOfSightAABB //
////////////////////////
// Returns the conservative axially-aligned WORLD SPACE box covering the container.
// Return true if the AABB was set.
bool Asura_ClientContainerInstance_Attachment::GetLineOfSightAABB( Asura_Bounding_Box& xAABB )
{
	if (m_pxChild)
	{
		return m_pxChild->GetLineOfSightAABB(xAABB);
	}
	return false;
}

/////////////
// LOSTest //
/////////////
// Do LOS test. Return true if the output data is changed.
bool Asura_ClientContainerInstance_Attachment::LOSTest(const Asura_LOS_InputData* pxInput, Asura_LOS_OutputData* pxOutput) const
{
	if (m_pxChild)
	{
		return m_pxChild->LOSTest(pxInput, pxOutput);
	}
	return false;
}

////////////////
// Initialise //
////////////////
void Asura_ClientContainerInstance_Attachment::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
														  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	ResetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);

	Asura_Assert(m_pxParentMeshForMarker==0, "Parent mesh should have been cleared before calling Initialise.");

	if (m_pxChild)
	{
		const Asura_Hash_ID uMarkerHash = GetMarkerPointHash();
		if (uMarkerHash != ASURA_HASH_ID_UNSET)
		{
			// Search for marker point and attach to it.
			if (pxParentCollection)
			{
				const bool bFound = pxParentCollection->FindMarkerPoint(this);
				
				// For tools builds, don't bother asserting when Tools is active, as it's annoying.
#ifdef ASURA_ASSERT
#ifdef ASURA_TOOLS
				const bool bOk = bFound || Tools_System::IsActive();
#else
				const bool bOk = bFound;
#endif
				Asura_Assertf(bOk, ("Couldn't find container attachment marker point (0x%08x) for entity %u.", uMarkerHash, GetOwnerGuid()));
#endif //ASURA_ASSERT				

				if (bFound)
				{
					m_pxChild->SetAttachedToMarkerPoint();
				}
				else
				{
					// Failed to attach the child.
					// It is better to destroy it than have it lying around in the wrong place.
					DeleteChild();
				}
			}
		}
		else if (ShouldAttachToPhysicsObject())
		{
			m_pxChild->SetAttachedToMarkerPoint();
		}
	}

	if (m_pxChild)
	{
		// Note: Attachment just passes its unique ID along to its child.
		m_pxChild->SetUniqueID(GetUniqueID());
		m_pxChild->Initialise(pxParentCollection, xData);

		if (m_pxChild->GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS))
		{
			SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_SHARED_PHYSICS);
		}
	}
}

////////////////////////
// GetMarkerPointHash //
////////////////////////
Asura_Hash_ID Asura_ClientContainerInstance_Attachment::GetMarkerPointHash() const
{
	if (m_pxContainer)
	{
		const Asura_Container_Attachment* pxContainer = static_cast<const Asura_Container_Attachment*>(m_pxContainer);
		return pxContainer->GetMarkerPointHash();
	}
	return ASURA_HASH_ID_UNSET;
}

/////////////////////
// FindMarkerPoint //
/////////////////////
// Search for the specified marker point. If found then set up the attachment and return true.
bool Asura_ClientContainerInstance_Attachment::FindMarkerPoint(Asura_ClientContainerInstance_Attachment* pxAttachment) const
{
	if (pxAttachment == this)
	{
		// Don't attach to self.
		return false;
	}

	if (m_pxChild)
	{
		// Don't check child if this is a marker point attachment.
		if (GetMarkerPointHash() == ASURA_HASH_ID_UNSET)
		{
			return m_pxChild->FindMarkerPoint(pxAttachment);
		}
	}

	return false;
}

///////////////////////
// GetBoundingSphere //
///////////////////////
bool Asura_ClientContainerInstance_Attachment::GetTightBoundingSphere( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxChild)
	{
		return m_pxChild->GetTightBoundingSphere(fRadius, xCentre);
	}
	return false;
}

bool Asura_ClientContainerInstance_Attachment::GetCoarseBoundingSphere( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxChild)
	{
		return m_pxChild->GetCoarseBoundingSphere(fRadius, xCentre);
	}
	return false;
}

bool Asura_ClientContainerInstance_Attachment::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxChild)
	{
		return m_pxChild->GetCoarseBoundingSphereIncludingEffects(fRadius, xCentre);
	}
	return false;
}

//////////////////////////////
// SetAttachedToMarkerPoint //
//////////////////////////////
void Asura_ClientContainerInstance_Attachment::SetAttachedToMarkerPoint()
{
	if (m_pxChild)
	{
		m_pxChild->SetAttachedToMarkerPoint();
	}
}

// Helper
const Asura_Container_Attachment* Asura_ClientContainerInstance_Attachment::GetContainer() const
{
	return static_cast<const Asura_Container_Attachment*>(m_pxContainer);
}

/////////////////
// ApplyOffset //
/////////////////
void Asura_ClientContainerInstance_Attachment::ApplyOffset(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri,
															Asura_Vector_3& xNewPosition, Asura_Matrix_3x3& xNewOri) const
{
	if (m_pxContainer)
	{
		const Asura_Container_Attachment* pxAttachment = GetContainer();
		pxAttachment->ApplyOffset(xPosition, xOri, xNewPosition, xNewOri, this);
	}
	else
	{
		xNewPosition = xPosition;
		xNewOri = xOri;
	}
}

/////////////////////////////////
// ShouldAttachToPhysicsObject //
/////////////////////////////////
bool Asura_ClientContainerInstance_Attachment::ShouldAttachToPhysicsObject() const
{
	if (m_pxContainer)
	{
		const Asura_Container_Attachment* pxAttachment = GetContainer();
		return pxAttachment->ShouldAttachToPhysicsObject();
	}
	return false;
}

////////////////////////////////
// GetBasePosOriForAttachment //
////////////////////////////////
// Returns true, if the position and orientation were set.
bool Asura_ClientContainerInstance_Attachment::GetBasePosOriForAttachment(Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientation) const
{
	if (m_pxContainer)
	{
		const Asura_Container_Attachment* pxAttachment = GetContainer();
		if (pxAttachment->ShouldAttachToPhysicsObject())
		{
			// Use collection's shared dynamic physics object.
			Asura_Assert(m_pxParentCollection, "Null parent pointer");
			if (m_pxParentCollection)
			{
				return m_pxParentCollection->GetBasePosOriForAttachment(xPosition, xOrientation);
			}
		}
		else if (m_pxParentMeshForMarker)
		{
			// Use marker point.
			const Asura_Animation* const pxAnim = m_pxParentMeshForMarker->GetAnimation();
			Asura_Assert(pxAnim, "Null animation pointer");
			if (pxAnim)
			{
				return pxAnim->GetMarkerLocation( GetMarkerIndex(), 0, &xPosition, &xOrientation, true );
			}
		}
	}

	return false;
}


///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_PFX
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_PFX::Asura_ClientContainerInstance_PFX(const Asura_Container_PFX* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_uEffectID(ASURA_PFX_INVALID_EFFECT_ID)
, m_fTriggerTimer(0.0f)
{
	if (Asura_Container_System::IsDoingCreateOrActivate() && pxContainer->IsTransitionOnly())
	{
		// Container entity is being created/loaded or activated. Not a transition, so disable.
		// This is so damage effects (e.g. fragments) don't get played on loading checkpoints.
		SetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED);
	}
}

Asura_ClientContainerInstance_PFX::~Asura_ClientContainerInstance_PFX()
{
	// Stop effect immediately if we are deleting or deactivating the corresponding entity.
	// Stop the effect gradually if it is just a state change.
	const bool bImmediate = !Asura_Container_System::IsDoingStateChange();
	StopEffect(bImmediate);
}

/////////////////
// StartEffect //
/////////////////
void Asura_ClientContainerInstance_PFX::StartEffect(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	m_fTriggerTimer = 0.0f;

	if (m_uEffectID != ASURA_PFX_INVALID_EFFECT_ID)
	{
		StopEffect(true);
	}

	if (GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_DISABLED))
	{
		return; // Disabled.
	}

	const Asura_Container_PFX* pxContainer = static_cast<const Asura_Container_PFX*>(m_pxContainer);
	if (!pxContainer) return;

	const Asura_Entity_PFX_Effect_Data& xData = pxContainer->GetPFXData();
	const Asura_Hash_ID uEffectTemplateHash = xData.m_uEffectTemplateHash;
	if (uEffectTemplateHash == ASURA_HASH_ID_UNSET)
	{
		return;
	}

	// Create the particle effect
	const bool bPointSource = (xData.m_xSourceVolume.m_eType == ASURA_PFX_VOLUME_POINT);
	const Asura_PFX_Volume* pxVolOverride = (bPointSource ? 0 : &xData.m_xSourceVolume);

	m_uEffectID = Asura_PFX_Entity_Effect::Create(xPosition, xOrientation, uEffectTemplateHash, pxVolOverride, ASURA_PFX_EFFECT_CAMERA_FLAGS, GetOwnerGuid());

	Asura_Assertf(m_uEffectID != ASURA_PFX_INVALID_EFFECT_ID, ("Failed to create PFX effect (0x%08x) for container entity %d", uEffectTemplateHash, GetOwnerGuid()));

	Asura_PFX_Entity_Effect* pxEffect = Asura_PFX_System::GetEntityEffect(m_uEffectID);
	if (pxEffect)
	{
		// TODO we need to adjust the position and orientation.
		//pxEffect->SetCollisionPlane(xData.m_xCollisionPosition, xData.m_xCollisionDirection);
		pxEffect->SetIgnoreModuleLinking(xData.m_bIgnoreModuleLinking);
		pxEffect->SetSkyboxEffect(xData.m_bSkyboxEffect);
		pxEffect->SetWaterDistorterEffect(xData.m_bWaterDistorterEffect);
		if(xData.m_bUsesSpline)
		{
			pxEffect->SetSpline(xData.m_pxSpline);
		}
	}
}

////////////////
// StopEffect //
////////////////
void Asura_ClientContainerInstance_PFX::StopEffect(bool bImmediate)
{
	m_fTriggerTimer = 0.0f;

	if (m_uEffectID != ASURA_PFX_INVALID_EFFECT_ID)
	{
		// Check if effect is still active.
		Asura_PFX_Entity_Effect* pxEffect = GetEffect();
		if (pxEffect)
		{
			pxEffect->Stop(bImmediate);
			if (bImmediate)
			{
				m_uEffectID = ASURA_PFX_INVALID_EFFECT_ID;
			}
		}
		else
		{
			// Effect no longer exists , so we can chuck the ID away.
			m_uEffectID = ASURA_PFX_INVALID_EFFECT_ID;
		}
	}
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_PFX::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	if (!m_pxContainer) return;

	// Don't render in Tools object mode if the relevant option is disabled.
#ifdef ASURA_TOOLS
	if (!Tools_Container_System::ShouldRenderPFX())
	{
		StopEffect(true);
		return;
	}
#endif //ASURA_TOOLS

	// Start or teleport the effect.
	if (m_uEffectID == ASURA_PFX_INVALID_EFFECT_ID)
	{
		StartEffect(xPosition, xOrientation);
	}
	else
	{
		Asura_PFX_Entity_Effect* pxEffect = GetEffect();
		if (pxEffect)
		{
			pxEffect->SetPosition(xPosition, xOrientation, true);
		}
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_PFX::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_uEffectID == ASURA_PFX_INVALID_EFFECT_ID)
	{
		return; // No effect set up.
	}

	// Render if needed in Tools.
	ToolsRender(xPosition, xOrientation);

	Asura_PFX_Entity_Effect* pxEffect = GetEffect();

	// Update position/orientation if attached to a marker point.
	if (IsAttachedToMarker() && pxEffect)
	{
		pxEffect->SetPosition(xPosition, xOrientation, false);
	}

	if (pxEffect) return; // Effect still running.

	// Effect has finished, perhaps we should restart it.
	
	// Find the PFX data.
	const Asura_Container_PFX* pxContainer = static_cast<const Asura_Container_PFX*>(m_pxContainer);
	if (!pxContainer) return;
	const Asura_Entity_PFX_Effect_Data& xData = pxContainer->GetPFXData();

	if (xData.m_fMaxRetriggerTime > 0.0f)
	{
		if (m_fTriggerTimer > 0.0f)
		{
			m_fTriggerTimer -= Asura_Timers::GetGameTimeSlice();
			if (m_fTriggerTimer <= 0.0f)
			{
				m_fTriggerTimer = 0.0f;
				StartEffect(xPosition, xOrientation);
			}
		}
		else
		{
			// Start count down.
			m_fTriggerTimer = Asura_Random::GetFloat(xData.m_fMinRetriggerTime, xData.m_fMaxRetriggerTime);
		}
	}
	else
	{
		if (pxContainer->IsRetriggerAllowed())
		{
			StopEffect(true);
		}
	}
}

// Initialise guid and attachments.
void Asura_ClientContainerInstance_PFX::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
												   Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	if (IsAttachedToMarker())
	{
		++xData.m_iNumRender;
	}
	else
	{
		const Asura_Container_PFX* pxContainer = static_cast<const Asura_Container_PFX*>(m_pxContainer);
		if (pxContainer && pxContainer->GetPFXData().m_fMaxRetriggerTime > 0.0f)
		{
			++xData.m_iNumRender;
		}
	}
}

///////////////
// GetEffect //
///////////////
Asura_PFX_Entity_Effect* Asura_ClientContainerInstance_PFX::GetEffect()
{
	return Asura_PFX_System::GetEntityEffect(m_uEffectID); 
}

const Asura_PFX_Entity_Effect* Asura_ClientContainerInstance_PFX::GetEffect() const
{
	return Asura_PFX_System::GetEntityEffect(m_uEffectID);
}

bool Asura_ClientContainerInstance_PFX::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	Asura_PFX_Entity_Effect* pxEffect = Asura_PFX_System::GetEntityEffect(m_uEffectID);
	if ( pxEffect )
	{
		fRadius = pxEffect->GetBoundingBox().CalculateRadius();
		xCentre = pxEffect->GetBoundingBox().CalculateCentre();
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Corona
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_Corona::Asura_ClientContainerInstance_Corona(const Asura_Container_Corona* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_iCoronaHandle(ASURA_INVALID_HANDLE)
{
}

Asura_ClientContainerInstance_Corona::~Asura_ClientContainerInstance_Corona()
{
	DestroyCorona();
}

//////////////
// GetParam //
//////////////
const Asura_Container_Corona_Param& Asura_ClientContainerInstance_Corona::GetParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_Corona* pxContainer = static_cast<const Asura_Container_Corona*>(m_pxContainer);
	return pxContainer->GetParam();
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_Corona::Update()
{
	UpdateFromGraphColour();
}

///////////////
// SetStrobe //
///////////////
void Asura_ClientContainerInstance_Corona::SetStrobe(bool bStrobeOn)
{
	PARENT::SetStrobe(bStrobeOn);
	
	if (m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		const Asura_Container_Corona* pxContainer = static_cast<const Asura_Container_Corona*>(m_pxContainer);
		if (pxContainer && pxContainer->IsUsingStrobe())
		{
			Asura_Corona_System::SetFlag_Hidden(m_iCoronaHandle, !bStrobeOn);
		}
	}
}

//////////////////
// CreateCorona //
//////////////////
void Asura_ClientContainerInstance_Corona::UpdateCorona(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (!m_pxContainer) return;
	const Asura_Container_Corona* pxContainer = static_cast<const Asura_Container_Corona*>(m_pxContainer);

	if (m_iCoronaHandle == ASURA_INVALID_HANDLE)
	{
		m_iCoronaHandle = Asura_Corona_System::CreateCorona();
	
		if (m_iCoronaHandle == ASURA_INVALID_HANDLE)
		{
			Asura_Assert(0, "Failed to create corona");
			return;
		}
	}

	// Update the corona values
	UpdatePosition(xPosition, xOrientation);
	
	const int iTextureIndex = Asura_Texture_Management::GetIndexFromHashID(pxContainer->GetTextureHash());
	Asura_Corona_System::SetTextureIndex(m_iCoronaHandle, iTextureIndex);
	Asura_Corona_System::SetRadius(m_iCoronaHandle, pxContainer->GetRadius());
	Asura_Corona_System::SetVolumetricDepthScale(m_iCoronaHandle, pxContainer->GetDepth());
	
	// Flags
	Asura_Corona_System::SetFlag_DontFadeWithDistance(m_iCoronaHandle, pxContainer->HasNoDistanceFadeOut() );
	Asura_Corona_System::SetFlag_UseModuleLinking(m_iCoronaHandle, pxContainer->UsesVisibilityTest() );
	

	// Set the direction fadeout effect.
	const Asura_Vector_3 xDir(xOrientation.e13, xOrientation.e23, xOrientation.e33);
	Asura_Corona_System::SetDirection(m_iCoronaHandle, xDir);
	Asura_Corona_System::SetFlag_UseDirectionEffect(m_iCoronaHandle, pxContainer->IsUsingDirectionFadeout());

	const Asura_RGBA& xCoronaColour = pxContainer->GetColour();

	const float fColourScale = 1.0f;
	const float fColourMax   = 100.0f; // Allow values greater than 1 (assume HDR is enabled).

	const float fRed   = Asura_Maths::ClampToRange(xCoronaColour.m_fR * fColourScale, 0.0f, fColourMax);
	const float fGreen = Asura_Maths::ClampToRange(xCoronaColour.m_fG * fColourScale, 0.0f, fColourMax);
	const float fBlue  = Asura_Maths::ClampToRange(xCoronaColour.m_fB * fColourScale, 0.0f, fColourMax);	
	Asura_Corona_System::SetColour(m_iCoronaHandle, fRed, fGreen, fBlue);

	// Make sure that the strobe state is initialised correctly.
	if (!GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON))
	{
		SetStrobe(false);
	}
}

///////////////////////////
// UpdateFromGraphColour //
///////////////////////////
void Asura_ClientContainerInstance_Corona::UpdateFromGraphColour()
{
	if (m_iCoronaHandle == ASURA_INVALID_HANDLE) return;
	
	const Asura_Container_Corona_Param& xParam = GetParam();
	if (!xParam.IsUsingGraph())
	{
		// Not using an information from GraphColour container.
		return;
	}

	// Get the current graph colour data.
	Asura_Container_GraphColour_Output xOutput;
	if (!GetGraphColourFromCollection(xOutput))
	{
		return;
	}

	// Colour/intensity
	// Base colour.
	Asura_RGBA xColour(xParam.m_xCoronaColour);

	if (xParam.IsUsingGraphFlicker())
	{
		xColour *= xOutput.m_fIntensity;
	}

	if (xParam.IsUsingGraphColour())
	{
		xColour.m_fR *= xOutput.m_xColour.m_fR;
		xColour.m_fG *= xOutput.m_xColour.m_fG;
		xColour.m_fB *= xOutput.m_xColour.m_fB;
	}
	
	Asura_Corona_System::SetColour(m_iCoronaHandle, xColour.m_fR, xColour.m_fG, xColour.m_fB);
}

///////////////////
// DestroyCorona //
///////////////////
void Asura_ClientContainerInstance_Corona::DestroyCorona()
{
	if (m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Corona_System::DestroyCorona(m_iCoronaHandle);
		m_iCoronaHandle = ASURA_INVALID_HANDLE;
	}
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_Corona::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	if (m_iCoronaHandle == ASURA_INVALID_HANDLE)
	{
		// Create if needed and update.
		UpdateCorona(xPosition, xOrientation);
	}
	else
	{
		// Just update the position and direction.
		UpdatePosition(xPosition, xOrientation);
	}
}

////////////////////
// UpdatePosition //
////////////////////
// Just update the position and orientation
void Asura_ClientContainerInstance_Corona::UpdatePosition(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_iCoronaHandle == ASURA_INVALID_HANDLE) return;

	const Asura_Vector_3 xForwards(xOrientation.e13, xOrientation.e23, xOrientation.e33);
	Asura_Corona_System::SetPosition(m_iCoronaHandle, xPosition);
	Asura_Corona_System::SetDirection(m_iCoronaHandle, xForwards);
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_Corona::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (IsAttachedToMarker())
	{
		// Just update the position and direction.
		UpdatePosition(xPosition, xOrientation);
	}
}

////////////////
// Initialise //
////////////////
// Initialise guid and attachments.
void Asura_ClientContainerInstance_Corona::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
													  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	// Only need to call Render for coronas if they might be moving.
	if (IsAttachedToMarker())
	{
		++xData.m_iNumRender;
	}

	if (GetParam().IsUsingGraph())
	{
		++xData.m_iNumUpdate;
	}
}

bool Asura_ClientContainerInstance_Corona::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	fRadius = 0.0f;
	if ( m_iCoronaHandle != ASURA_INVALID_HANDLE )
	{
		bool bOK = Asura_Corona_System::GetRadius( m_iCoronaHandle, fRadius );
		return Asura_Corona_System::GetPosition( m_iCoronaHandle, xCentre ) && bOK;
	}
	else
	{
		return false;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_LightShaft
///////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_TARGET_SM3

Asura_ClientContainerInstance_LightShaft::Asura_ClientContainerInstance_LightShaft(const Asura_Container_LightShaft* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_pxLightShaft(0)
{
}

Asura_ClientContainerInstance_LightShaft::~Asura_ClientContainerInstance_LightShaft()
{
	DestroyLightShaft();
}

//////////////
// GetParam //
//////////////
const Asura_Container_LightShaft_Param& Asura_ClientContainerInstance_LightShaft::GetParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_LightShaft* pxContainer = static_cast<const Asura_Container_LightShaft*>(m_pxContainer);
	return pxContainer->GetParam();
}

////////////////////
// GetCoronaParam //
////////////////////
const Asura_Container_Corona_Param& Asura_ClientContainerInstance_LightShaft::GetCoronaParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_LightShaft* pxContainer = static_cast<const Asura_Container_LightShaft*>(m_pxContainer);
	return pxContainer->GetCoronaParam();
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_LightShaft::Update()
{
	UpdateFromGraphColour();
}

///////////////
// SetStrobe //
///////////////
void Asura_ClientContainerInstance_LightShaft::SetStrobe(bool bStrobeOn)
{
	PARENT::SetStrobe(bStrobeOn);
	
	if (m_pxLightShaft)
	{
		const Asura_Container_LightShaft* pxContainer = static_cast<const Asura_Container_LightShaft*>(m_pxContainer);
		if (pxContainer && pxContainer->IsUsingStrobe())
		{
			m_pxLightShaft->SetEnable(bStrobeOn);
		}
	}
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_LightShaft::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	if (m_pxLightShaft == 0)
	{
		// Create if needed and update.
		UpdateLightShaft(xPosition, xOrientation);
	}
	else
	{
		// Just update the position and direction.
		UpdatePosition(xPosition, xOrientation);
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_LightShaft::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (IsAttachedToMarker())
	{
		// Just update the position and direction.
		UpdatePosition(xPosition, xOrientation);
	}
}

////////////////
// Initialise //
////////////////
// Initialise guid and attachments.
void Asura_ClientContainerInstance_LightShaft::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
														  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	// Only need to call Render for lightshafts if they might be moving.
	if (IsAttachedToMarker())
	{
		++xData.m_iNumRender;
	}

	if (GetCoronaParam().IsUsingGraph())
	{
		++xData.m_iNumUpdate;
	}
}
	
//////////////////////
// UpdateLightShaft //
//////////////////////
void Asura_ClientContainerInstance_LightShaft::UpdateLightShaft(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (!m_pxContainer) return;

	if (m_pxLightShaft == 0)
	{
		m_pxLightShaft = Asura_LightShaft_System::CreateLightShaft();	

		if (!m_pxLightShaft)
		{
			Asura_Assert(0, "Failed to create light shaft");
			return;
		}

		// Set this to the origin, just in case UpdatePosition doesn't set it.
		m_pxLightShaft->m_xBB.SetToPoint(Asura_Maths::ZeroVector3);
		m_pxLightShaft->SetEnable(true);
	}

	const Asura_Container_LightShaft* pxLightShaft = static_cast<const Asura_Container_LightShaft*>(m_pxContainer);
	const Asura_Container_LightShaft_Param& xParam = pxLightShaft->GetParam();

	// Update cone geometry params.
	for (u_int u = 0; u < 2; ++u)
	{
		m_pxLightShaft->m_afDistances[u]  = xParam.m_afDistance[u];
		m_pxLightShaft->m_afRadii[u]      = xParam.m_afRadius[u];
		m_pxLightShaft->m_axColours[u]	  = xParam.m_axColour[u];						
	}

	// Update texturing params.
	for (u_int u = 0; u < 2; ++u)
	{
		const int iTex = Asura_Texture_Management::GetIndexFromHashID(xParam.m_auTextureHash[u]);
		m_pxLightShaft->m_aiTextureIndices[u]	  = static_cast<ASURA_TEXTURE_ID>(iTex);
		m_pxLightShaft->m_afTextureScrollRates[u] = xParam.m_afTextureScrollRate[u];
		m_pxLightShaft->m_afTextureUScales[u]	  = xParam.m_afTextureUScale[u];
	}

	// Distance fade
	m_pxLightShaft->m_fDistFadeRange = xParam.m_fDistFadeRange;
	m_pxLightShaft->m_fDistFadeStart = xParam.m_fDistFadeStart;
	m_pxLightShaft->m_fCloseFadeRange = xParam.m_fCloseFadeRange;
	m_pxLightShaft->m_fCloseFadeStart = xParam.m_fCloseFadeStart;
	
	// Corona
	const Asura_Container_Corona_Param& xCoronaParam = xParam.m_xCoronaParam;
	const int iTex = Asura_Texture_Management::GetIndexFromHashID(xCoronaParam.m_uCoronaTextureHash);
	m_pxLightShaft->m_iCoronaTextureIndex	= static_cast<ASURA_TEXTURE_ID>(iTex);
	m_pxLightShaft->m_fCoronaAngleRange		= xParam.m_fCoronaAngleRange;
	m_pxLightShaft->m_fCoronaSize			= xCoronaParam.m_fCoronaSize;
	m_pxLightShaft->m_fCoronaDepth			= xCoronaParam.m_fCoronaDepth;
	m_pxLightShaft->m_xCoronaColour			= xCoronaParam.m_xCoronaColour;

	// Update the pos/orient and BB.
	UpdatePosition(xPosition, xOrientation);

	// Make sure that the strobe state is initialised correctly.
	if (!GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON))
	{
		SetStrobe(false);
	}
}

///////////////////////////
// UpdateFromGraphColour //
///////////////////////////
void Asura_ClientContainerInstance_LightShaft::UpdateFromGraphColour()
{
	if (!m_pxLightShaft) return;

	const Asura_Container_Corona_Param& xCoronaParam = GetCoronaParam();
	if (!xCoronaParam.IsUsingGraph())
	{
		// Not using an information from GraphColour container.
		return;
	}

	// Get the current graph colour data.
	Asura_Container_GraphColour_Output xOutput;
	if (!GetGraphColourFromCollection(xOutput))
	{
		return;
	}

	// LightShaft
	const float fIntensity = (xCoronaParam.IsUsingGraphFlicker() ? xOutput.m_fIntensity : 1.0f);

	Asura_RGBA xColour(fIntensity, fIntensity, fIntensity, 1.0f);

	if (xCoronaParam.IsUsingGraphColour())
	{
		xColour.m_fR *= xOutput.m_xColour.m_fR;
		xColour.m_fG *= xOutput.m_xColour.m_fG;
		xColour.m_fB *= xOutput.m_xColour.m_fB;
	}
		
	m_pxLightShaft->SetInheritedColour(xColour);

	// Corona
	if (m_pxLightShaft->m_iCoronaTextureIndex > 0)
	{
		xColour.m_fR *= xCoronaParam.m_xCoronaColour.m_fR;
		xColour.m_fG *= xCoronaParam.m_xCoronaColour.m_fG;
		xColour.m_fB *= xCoronaParam.m_xCoronaColour.m_fB;

		m_pxLightShaft->m_xCoronaColour = xColour;
	}
}

////////////////////
// UpdatePosition //
////////////////////
// Just update the position and orientation
void Asura_ClientContainerInstance_LightShaft::UpdatePosition(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_pxLightShaft)
	{
		// Only recalculate the bounding box if the thing has actually moved.
		bool bMoved = false;
		if (!xPosition.ApproximatelyEquals(m_pxLightShaft->m_xPos))
		{
			m_pxLightShaft->m_xPos = xPosition;
			bMoved = true;
		}

		if (bMoved || !xOrientation.ApproximatelyEquals(m_pxLightShaft->m_xOrient))
		{
			m_pxLightShaft->m_xOrient = xOrientation;
			bMoved = true;
		}

		if (bMoved)
		{
			m_pxLightShaft->UpdateBoundingBox();
		}
	}
}

///////////////////////
// DestroyLightShaft //
///////////////////////
void Asura_ClientContainerInstance_LightShaft::DestroyLightShaft()
{
	if (m_pxLightShaft)
	{
		Asura_LightShaft_System::DestroyLightShaft(m_pxLightShaft);		
		m_pxLightShaft = 0;
	}
}

bool Asura_ClientContainerInstance_LightShaft::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxLightShaft)
	{
		fRadius = m_pxLightShaft->GetBoundingBox().CalculateRadius();
		xCentre = m_pxLightShaft->GetBoundingBox().CalculateCentre();
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Projector
///////////////////////////////////////////////////////////////////////////////

Asura_ClientContainerInstance_Projector::Asura_ClientContainerInstance_Projector(const Asura_Container_Projector* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_pxProjector(0)
{
}

Asura_ClientContainerInstance_Projector::~Asura_ClientContainerInstance_Projector()
{
	DestroyProjector();
}

void Asura_ClientContainerInstance_Projector::SetStrobe(bool bStrobeOn)
{
	PARENT::SetStrobe(bStrobeOn);
	
	if (!m_pxProjector) return;

	const Asura_Container_Projector_Param& xParam = GetParam();
	if (xParam.IsUsingStrobe())
	{
		if (bStrobeOn)
		{
			// Restore fade parameters.
			m_pxProjector->SetFade(xParam.m_fDistFadeStart, xParam.m_fDistFadeRange); ;
			// Restore shadow parameters.
			m_pxProjector->SetShadowParameters(GetShadowParam());
		}
		else
		{
			// Turn off the light by making sure it fails the distance test.
			m_pxProjector->SetFade(0.0f, 0.0f);
			// Turn off the shadow too.
			m_pxProjector->SetShadowParameters(Asura_ShadowParameters(0));
		}
	}
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_Projector::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	if (m_pxProjector == 0)
	{
		// Create if needed and update.
		UpdateProjector(xPosition, xOrientation);
	}
	else
	{
		// Just update the position and orientation.
		UpdatePosition(xPosition, xOrientation);
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_Projector::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (IsAttachedToMarker())
	{
		// Just update the position and orientation.
		UpdatePosition(xPosition, xOrientation);
	}

#ifdef ASURA_TOOLS
	// Render the frustum in Tools
	if ( m_pxProjector && Tools_System::IsActive() &&
		(Tools_System::IsInContainerMode() || Tools_System::IsInObjectMode()) )
	{
		if ( GetParam().m_uFlags & ASURA_CONTAINER_PROJECTOR_FLAG_TOOLS_RENDER_FRUSTRUM )
		{
			m_pxProjector->RenderFrustum();
		}
		// Update the limiting plane.
		ToolsRender(xPosition, xOrientation);
	}
#endif//ASURA_TOOLS
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_Projector::Update()
{
	UpdateFromGraphColour();
}

////////////////
// Initialise //
////////////////
// Initialise guid and attachments.
void Asura_ClientContainerInstance_Projector::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
														 Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	// Only need to call Render for projectors if they might be moving.
	if (IsAttachedToMarker())
	{
		++xData.m_iNumRender;
	}

	if (GetParam().IsUsingGraph())
	{
		++xData.m_iNumUpdate;
	}
}

//////////////
// GetParam //
//////////////
const Asura_Container_Projector_Param& Asura_ClientContainerInstance_Projector::GetParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_Projector* const pxProjector = static_cast<const Asura_Container_Projector*>(m_pxContainer);
	return pxProjector->GetParam();
}

////////////////////
// GetShadowParam //
////////////////////
const Asura_ShadowParameters& Asura_ClientContainerInstance_Projector::GetShadowParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_Projector* const pxProjector = static_cast<const Asura_Container_Projector*>(m_pxContainer);
	return pxProjector->GetShadowParam();
}
	
/////////////////////
// UpdateProjector //
/////////////////////
void Asura_ClientContainerInstance_Projector::UpdateProjector(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (!m_pxContainer) return;

	if (m_pxProjector == 0)
	{
		m_pxProjector = Asura_Projector_System::CreateProjector();	

		if (!m_pxProjector)
		{
			Asura_Assert(0, "Failed to create projector");
			return;
		}		
	}

	const Asura_Container_Projector_Param& xParam = GetParam();

	// Textures
	m_pxProjector->m_iTextureIndex = Asura_Texture_Management::GetIndexFromHashID(xParam.m_uTextureHash);
	m_pxProjector->m_iDetailTextureIndex = Asura_Texture_Management::GetIndexFromHashID(xParam.m_uDetailTextureHash);
				
	// Projection
	m_pxProjector->m_fFoV		= xParam.m_fFoV;
	m_pxProjector->m_fMinRange	= xParam.m_fMinRange;
	m_pxProjector->m_fMaxRange	= xParam.m_fMaxRange;
	m_pxProjector->m_fScale		= xParam.m_fParallelScale;

	// Colour
	m_pxProjector->m_fIntensity = 1.0f;
	m_pxProjector->m_afRGB[0] = xParam.m_afRGB[0];
	m_pxProjector->m_afRGB[1] = xParam.m_afRGB[1];
	m_pxProjector->m_afRGB[2] = xParam.m_afRGB[2];

	// Fall-off
	if (xParam.m_fFalloffInnerRatio >= 0.99f)
	{
		m_pxProjector->m_fFalloffScale = 100.0f;
	}
	else
	{
		m_pxProjector->m_fFalloffScale = 1.0f / (1.0f - xParam.m_fFalloffInnerRatio);
	}

	m_pxProjector->m_fFalloffPower  = xParam.m_fFalloffPower;

	// Detail texture scrolling
	m_pxProjector->m_fUScrollSpeed = xParam.m_fUScrollSpeed;
	m_pxProjector->m_fVScrollSpeed = xParam.m_fVScrollSpeed;
	m_pxProjector->m_fUScrollRange = xParam.m_fUScrollRange;
	m_pxProjector->m_fVScrollRange = xParam.m_fVScrollRange;

	// Texture params
	m_pxProjector->m_fDetailTextureScale = xParam.m_fDetailTextureScale;
	m_pxProjector->m_fUVAspect			 = xParam.m_fUVAspect;

	// Limiting plane
	m_pxProjector->m_xLimitingPlanePos    = xParam.m_xLimitingPlanePos * xOrientation + xPosition;
	m_pxProjector->m_xLimitingPlaneNormal = xParam.m_xLimitingPlaneNormal * xOrientation;

	// Types
	const u_int uFlags = xParam.m_uFlags;
	const u_int uType  = (uFlags & ASURA_CONTAINER_PROJECTOR_FLAG_TYPE_MASK);

	m_pxProjector->m_bFakeLight			= false;
	m_pxProjector->m_bDynamicLight		= false;
	m_pxProjector->m_bDynamicSpotlight	= false;
	m_pxProjector->m_eBlendMode			= ASURA_PROJECTOR_BLEND_ADD;

	switch (uType)
	{
	default:
		Asura_Assertf(0, ("Unknown projector type or blend mode (%u)", uType));
		// fallthrough

	case ASURA_CONTAINER_PROJECTOR_FLAG_FAKE_LIGHT:
		m_pxProjector->m_bFakeLight = true;
		break;

	case ASURA_CONTAINER_PROJECTOR_FLAG_DYNAMIC_LIGHT:
		m_pxProjector->m_bDynamicLight = true;
		break;

	case ASURA_CONTAINER_PROJECTOR_FLAG_DYNAMIC_SPOTLIGHT:
		m_pxProjector->m_bDynamicSpotlight = true;
		break;

	case ASURA_CONTAINER_PROJECTOR_FLAG_BLEND_ADD:
		m_pxProjector->m_eBlendMode = ASURA_PROJECTOR_BLEND_ADD;
		break;

	case ASURA_CONTAINER_PROJECTOR_FLAG_BLEND_MULTIPLY:
		m_pxProjector->m_eBlendMode = ASURA_PROJECTOR_BLEND_MULTIPLY;
		break;

	case ASURA_CONTAINER_PROJECTOR_FLAG_BLEND_MULTIPLYx2:
		m_pxProjector->m_eBlendMode = ASURA_PROJECTOR_BLEND_MULTIPLYX2;
		break;

	case ASURA_CONTAINER_PROJECTOR_FLAG_BLEND_STANDARD:
		m_pxProjector->m_eBlendMode = ASURA_PROJECTOR_BLEND_STANDARD;
		break;
	}

	// Flags
	const bool bParallel = (0 != (uFlags & ASURA_CONTAINER_PROJECTOR_FLAG_PARALLEL));
	m_pxProjector->m_eType = (bParallel ? ASURA_PROJECTOR_TYPE_PARALLEL : ASURA_PROJECTOR_TYPE_PERSPECTIVE);

	m_pxProjector->m_bUseLimitingPlane = (0 != (uFlags & ASURA_CONTAINER_PROJECTOR_FLAG_USE_LIMITING_PLANE));

	// Spotlight
	if (m_pxProjector->m_bDynamicSpotlight)
	{
		m_pxProjector->SetSpotlightAngularFalloff(xParam.m_fInnerFoV, xParam.m_fFoV);
	}
	else
	{
		m_pxProjector->m_fAngularFalloffOfs   = 0.0f;
		m_pxProjector->m_fAngularFalloffScale = 1.0f;
	}

	// Distance fade out
	m_pxProjector->m_fDistFadeRange = xParam.m_fDistFadeRange;
	m_pxProjector->m_fDistFadeStart = xParam.m_fDistFadeStart;

	// Shadows
	m_pxProjector->SetShadowParameters(GetShadowParam());

	// Update the pos/orient and vertices.
	m_pxProjector->m_xPos    = xPosition;
	m_pxProjector->m_xOrient = xOrientation;
	m_pxProjector->UpdateVertices();

	// Make sure that the strobe state is initialised correctly.
	if (!GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON))
	{
		SetStrobe(false);
	}
}

///////////////////////////
// UpdateFromGraphColour //
///////////////////////////
void Asura_ClientContainerInstance_Projector::UpdateFromGraphColour()
{
	if (!m_pxProjector) return;

	const Asura_Container_Projector_Param& xParam = GetParam();
	if (!xParam.IsUsingGraph())
	{
		// Not using an information from GraphColour container.
		return;
	}

	// Get the current graph colour data.
	Asura_Container_GraphColour_Output xOutput;
	if (!GetGraphColourFromCollection(xOutput))
	{
		return;
	}

	// Colour
	if (xParam.IsUsingGraphColour())
	{
		Asura_RGBA xColour(xParam.m_afRGB[0], xParam.m_afRGB[1], xParam.m_afRGB[2], 1.0f);
		xColour.m_fR *= xOutput.m_xColour.m_fR;
		xColour.m_fG *= xOutput.m_xColour.m_fG;
		xColour.m_fB *= xOutput.m_xColour.m_fB;

		m_pxProjector->SetRGB(xColour.m_fR, xColour.m_fG, xColour.m_fB);	
	}

	// Flicker
	if (xParam.IsUsingGraphFlicker())
	{
		m_pxProjector->SetIntensity(xOutput.m_fIntensity);
	}

	// Range
	if (xParam.IsUsingGraphRange())
	{
		const float fRangeScale = xOutput.m_xColour.m_fA;
		const float fNewFoV = Asura_Maths::ClampToRange(xParam.m_fFoV * fRangeScale, 0.0f, Asura_Maths::DegToRad(150.0f));

		const float fEpsilon = 1.0e-3f;
		if (Asura_Maths::Abs(m_pxProjector->m_fFoV - fNewFoV) > fEpsilon)
		{
			m_pxProjector->m_fFoV = fNewFoV;

			// If it's a spotlight, we need to scale it's inner FoV, too.
			if (m_pxProjector->m_bDynamicSpotlight)
			{		
				m_pxProjector->SetSpotlightAngularFalloff(xParam.m_fInnerFoV * fRangeScale, fNewFoV);
			}

			m_pxProjector->UpdateVertices();
		}
	}
}

////////////////////
// UpdatePosition //
////////////////////
// Just update the position and orientation
void Asura_ClientContainerInstance_Projector::UpdatePosition(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_pxProjector)
	{
		// Only recalculate the bounding box if the thing has actually moved.
		bool bMoved = false;
		if (!xPosition.ApproximatelyEquals(m_pxProjector->m_xPos))
		{
			m_pxProjector->m_xPos = xPosition;
			bMoved = true;
		}

		if (bMoved || !xOrientation.ApproximatelyEquals(m_pxProjector->m_xOrient))
		{
			m_pxProjector->m_xOrient = xOrientation;
			bMoved = true;
		}

		if (bMoved)
		{
			// Update the limiting plane and the vertices.
			if (m_pxContainer && m_pxProjector->m_bUseLimitingPlane)
			{
				const Asura_Container_Projector_Param& xParam = GetParam();
				m_pxProjector->m_xLimitingPlanePos    = xParam.m_xLimitingPlanePos * xOrientation + xPosition;
				m_pxProjector->m_xLimitingPlaneNormal = xParam.m_xLimitingPlaneNormal * xOrientation;
			}

			m_pxProjector->UpdateVertices();
		}
	}
}

///////////////////////
// DestroyProjector //
///////////////////////
void Asura_ClientContainerInstance_Projector::DestroyProjector()
{
	if (m_pxProjector)
	{
		Asura_Projector_System::DestroyProjector(m_pxProjector);		
		m_pxProjector = 0;
	}
}

bool Asura_ClientContainerInstance_Projector::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if( m_pxProjector )
	{
		fRadius = m_pxProjector->GetBoundingBox().CalculateRadius();
		xCentre = m_pxProjector->GetBoundingBox().CalculateCentre();
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Photon
///////////////////////////////////////////////////////////////////////////////

Asura_ClientContainerInstance_Photon::Asura_ClientContainerInstance_Photon(const Asura_Container_Photon* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_pxPhoton(0)
{
}

Asura_ClientContainerInstance_Photon::~Asura_ClientContainerInstance_Photon()
{
	DestroyPhoton();
}

void Asura_ClientContainerInstance_Photon::SetStrobe(bool bStrobeOn)
{
	PARENT::SetStrobe(bStrobeOn);

	if (m_pxPhoton && GetParam().IsUsingStrobe())
	{
		m_pxPhoton->SetAsActive(bStrobeOn);
	}
}

////////////////
// TeleportTo //
////////////////
void Asura_ClientContainerInstance_Photon::TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	PARENT::TeleportTo(xPosition, xOrientation);

	if (m_pxPhoton == 0)
	{
		// Create if needed and update.
		UpdatePhoton(xPosition, xOrientation);
	}
	else
	{
		// Just update the position and orientation.
		UpdatePosition(xPosition, xOrientation);
	}
}

////////////
// Render //
////////////
void Asura_ClientContainerInstance_Photon::Render(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (IsAttachedToMarker())
	{
		// Just update the position and orientation.
		UpdatePosition(xPosition, xOrientation);
	}

#ifdef ASURA_TOOLS
	// Render the volume in Tools
	if (m_pxPhoton && Tools_System::IsActive())
	{
		if ( GetParam().m_uFlags & ASURA_CONTAINER_PHOTON_FLAG_TOOLS_RENDER_LIGHT_VOLUME )
		{
			m_pxPhoton->RenderLightVolume();
		}
		// Update the limiting volume.
		ToolsRender(xPosition, xOrientation);
	}
#endif//ASURA_TOOLS
}

////////////
// Update //
////////////
void Asura_ClientContainerInstance_Photon::Update()
{
	UpdateFromGraphColour();
}

////////////////
// Initialise //
////////////////
// Initialise guid and attachments.
void Asura_ClientContainerInstance_Photon::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
													  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);

	// Only need to call Render for photons if they might be moving.
	if (IsAttachedToMarker())
	{
		++xData.m_iNumRender;
	}

	if (GetParam().IsUsingGraph())
	{
		++xData.m_iNumUpdate;
	}
}

//////////////
// GetParam //
//////////////
const Asura_Container_Photon_Param& Asura_ClientContainerInstance_Photon::GetParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_Photon* const pxPhoton = static_cast<const Asura_Container_Photon*>(m_pxContainer);
	return pxPhoton->GetParam();
}

////////////////////
// GetShadowParam //
////////////////////
const Asura_ShadowParameters& Asura_ClientContainerInstance_Photon::GetShadowParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_Photon* const pxPhoton = static_cast<const Asura_Container_Photon*>(m_pxContainer);
	return pxPhoton->GetShadowParam();
}

//////////////////
// UpdatePhoton //
//////////////////
void Asura_ClientContainerInstance_Photon::UpdatePhoton(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (!m_pxContainer) return;

	if (m_pxPhoton == 0)
	{
		m_pxPhoton = Asura_Photon_System::CreatePhoton();	

		if (!m_pxPhoton)
		{
			Asura_Assert(0, "Failed to create photon");
			return;
		}		
	}

	const Asura_Container_Photon_Param& xParam = GetParam();

	m_pxPhoton->SetAsActive(true);
	m_pxPhoton->SetIsDirectional(false);

	// Colour
	m_pxPhoton->SetColours(xParam.m_afRGB[0], xParam.m_afRGB[1], xParam.m_afRGB[2]);

	// Range
	m_pxPhoton->SetRange(xParam.m_fPhotonRange);

	// Light volume
	const bool bUseLightVolume = xParam.IsUsingLightVolume();
	m_pxPhoton->SetCustomLightVolumeEnable(bUseLightVolume);

	if (bUseLightVolume)
	{
		Asura_Bounding_Box xBB = xParam.m_xLightVolume;
		xBB.Displace(xPosition);
		m_pxPhoton->SetCustomLightVolume(xBB);

		Asura_Quat xTempQuat(xOrientation);
		m_pxPhoton->SetCustomLightVolumeOrient(xTempQuat * xParam.m_xLightVolumeOrient);
	}

	// Limiting angles
	if (xParam.IsUsingLimitingAngles())
	{
		m_pxPhoton->SetLimitAnglesEnabled(true);

		const float fInner = Asura_Maths::Cosine(0.5f * xParam.m_fLimitingAngleInner);
		const float fOuter = Asura_Maths::Cosine(0.5f * xParam.m_fLimitingAngleOuter);

		m_pxPhoton->SetLimitAngleInner(fInner);
		m_pxPhoton->SetLimitAngleOuter(fOuter);
	}
	else
	{		
		m_pxPhoton->SetLimitAnglesEnabled(false);
	}

	// Textures
	const int iProjCubeTexture = Asura_Texture_Management::GetIndexFromHashID(xParam.m_uProjCubeTextureHash);				
	m_pxPhoton->SetProjCubeTextureIndex(iProjCubeTexture);
	
	const int iProjCubeBlurredTexture = Asura_Texture_Management::GetIndexFromHashID(xParam.m_uProjCubeBlurredTextureHash);				
	m_pxPhoton->SetProjCubeBlurredTextureIndex(iProjCubeBlurredTexture);

	m_pxPhoton->SetProjCubeLerpDist(xParam.m_fProjCubeLerpDist);

	// Fall off
	m_pxPhoton->SetFalloffPower(xParam.m_fFalloffPower);

	if (xParam.m_fFalloffInnerRatio >= 0.99f)
	{
		m_pxPhoton->SetFalloffScale(100.0f);
	}
	else
	{
		m_pxPhoton->SetFalloffScale(1.0f / (1.0f - xParam.m_fFalloffInnerRatio));
	}

	// Distance fade out
	m_pxPhoton->SetDistanceFade(true, xParam.m_fDistFadeStart, xParam.m_fDistFadeRange);

	// Shadows
#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
	m_pxPhoton->SetShadowParameters(GetShadowParam());
#endif //defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)

	// Position/orientation
	m_pxPhoton->SetPosition(xPosition);
	m_pxPhoton->SetLightOrientation(xOrientation);

	// Make sure that the strobe state is initialised correctly.
	if (!GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON))
	{
		SetStrobe(false);
	}
}

///////////////////////////
// UpdateFromGraphColour //
///////////////////////////
void Asura_ClientContainerInstance_Photon::UpdateFromGraphColour()
{
	if (!m_pxPhoton) return;

	const Asura_Container_Photon_Param& xParam = GetParam();
	if (!xParam.IsUsingGraph())
	{
		// Not using an information from GraphColour container.
		return;
	}

	// Get the current graph colour data.
	Asura_Container_GraphColour_Output xOutput;
	if (!GetGraphColourFromCollection(xOutput))
	{
		return;
	}

	// Range
	if (xParam.IsUsingGraphRange())
	{
		m_pxPhoton->SetRange(xParam.m_fPhotonRange * xOutput.m_xColour.m_fA);
	}

	// Colour/intensity
	if (xParam.IsUsingGraphColour() || xParam.IsUsingGraphFlicker())
	{
		// Base colour.
		Asura_RGBA xColour(xParam.m_afRGB[0], xParam.m_afRGB[1], xParam.m_afRGB[2], 1.0f);

		if (xParam.IsUsingGraphFlicker())
		{
			xColour *= xOutput.m_fIntensity;
		}

		if (xParam.IsUsingGraphColour())
		{
			xColour.m_fR *= xOutput.m_xColour.m_fR;
			xColour.m_fG *= xOutput.m_xColour.m_fG;
			xColour.m_fB *= xOutput.m_xColour.m_fB;
		}
		
		m_pxPhoton->SetColours(xColour.m_fR, xColour.m_fG, xColour.m_fB);
	}
}

////////////////////
// UpdatePosition //
////////////////////
// Just update the position and orientation
void Asura_ClientContainerInstance_Photon::UpdatePosition(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation)
{
	if (m_pxPhoton)
	{
		const bool bMoved = (!xPosition.ApproximatelyEquals(m_pxPhoton->GetPosition()));
		if (bMoved)
		{
			m_pxPhoton->SetPosition(xPosition);
		}
		const bool bRotated = (!xOrientation.ApproximatelyEquals(m_pxPhoton->GetLightOrientation()));
		if (bRotated)
		{
			m_pxPhoton->SetLightOrientation(xOrientation);
		}

		if (m_pxContainer)
		{
			const Asura_Container_Photon_Param& xParam = GetParam();
			if (xParam.IsUsingLightVolume())
			{
				if (bMoved)
				{
					Asura_Bounding_Box xBB = xParam.m_xLightVolume;
					xBB.Displace(xPosition);
					m_pxPhoton->SetCustomLightVolume(xBB);
				}

				if (bRotated)
				{
					const Asura_Quat xTempQuat(xOrientation);
					m_pxPhoton->SetCustomLightVolumeOrient(xTempQuat * xParam.m_xLightVolumeOrient);
				}
			}
		}
	}
}

///////////////////
// DestroyPhoton //
///////////////////
void Asura_ClientContainerInstance_Photon::DestroyPhoton()
{
	if (m_pxPhoton)
	{
		Asura_Photon_System::DestroyPhoton(m_pxPhoton);		
		m_pxPhoton = 0;
	}
}

bool Asura_ClientContainerInstance_Photon::GetCoarseBoundingSphereIncludingEffects( float& fRadius, Asura_Vector_3& xCentre ) const
{
	if (m_pxPhoton)
	{
		fRadius = m_pxPhoton->GetRange();
		xCentre = m_pxPhoton->GetPosition();
		return true;
	}
	return false;
}

#endif //ifdef ASURA_TARGET_SM3


///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Strobe
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_Strobe::Asura_ClientContainerInstance_Strobe(const Asura_Container_Strobe* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_fStrobeTimer(0.0f)
{
	// Starts in the "on" strobe-state.
	m_fStrobeTimer = Asura_Timers::GetLocalScaledTime() + Asura_Random::GetFloat(pxContainer->GetMinTimeOn(), pxContainer->GetMaxTimeOn());
}

Asura_ClientContainerInstance_Strobe::~Asura_ClientContainerInstance_Strobe()
{
	// If the strobe has been deleted due to a state change. We should switch any
	// children of the parent collection to the "on" strobe-state, otherwise they
	// might be stuck in the "off" strobe-state forever.
	const bool bStateChange = Asura_Container_System::IsDoingStateChange();
	if (bStateChange)
	{
		const bool bStrobeOn = GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON);
		if (!bStrobeOn && m_pxParentCollection)
		{
			// Turn everything back on.
			m_pxParentCollection->SetStrobe(true);
		}
	}
}

////////////
// Update //
////////////
// Check whether we need to change the strobe state.
void Asura_ClientContainerInstance_Strobe::Update()
{
	Asura_Assert(m_pxContainer, "Null container!");
	if (!m_pxContainer) return;

	const float fCurrentTime = Asura_Timers::GetLocalScaledTime();
	if (fCurrentTime > m_fStrobeTimer)
	{
		const Asura_Container_Strobe* const pxStrobe = static_cast<const Asura_Container_Strobe*>(m_pxContainer);

		ToggleFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON);

		const bool bStrobeOn = GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON);
		if (bStrobeOn)
		{
			m_fStrobeTimer = fCurrentTime + Asura_Random::GetFloat(pxStrobe->GetMinTimeOn(), pxStrobe->GetMaxTimeOn());
		}
		else
		{
			m_fStrobeTimer = fCurrentTime + Asura_Random::GetFloat(pxStrobe->GetMinTimeOff(), pxStrobe->GetMaxTimeOff());
		}

		// Inform the parent collection (and its children) that the strobe has changed state.
		if (m_pxParentCollection)
		{
			m_pxParentCollection->SetStrobe(bStrobeOn);
		}
	}
}

// Initialise guid.
void Asura_ClientContainerInstance_Strobe::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
													  Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);
	++xData.m_iNumUpdate;

	// Initialise is called after a state change. Make sure that any newly created containers
	// have the correct strobe state (everything is created enabled by default).
	if (m_pxParentCollection)
	{
		const bool bStrobeOn = GetFlag(ASURA_CONTAINER_INSTANCE_FLAG_STROBE_ON);
		if (!bStrobeOn)
		{
			// If the collection is in the "off" strobe-state and there is a state change which
			// leaves the strobe alone, but creates new children, then this should ensure
			// that the children start with the correct "off" strobe-state (i.e. disabled/hidden).
			m_pxParentCollection->SetStrobe(bStrobeOn);
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_GraphColour
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_GraphColour::Asura_ClientContainerInstance_GraphColour(const Asura_Container_GraphColour* pxContainer)
 : Asura_ClientContainerInstance(pxContainer)
 , m_fOffset(Asura_Random::GetFloat())
 , m_fGraphStartTime(Asura_Timers::GetLocalScaledTime())
 , m_uCacheFrameNumber(0xffffffff)
{
}

Asura_ClientContainerInstance_GraphColour::~Asura_ClientContainerInstance_GraphColour()
{
}

// Initialise guid and attachments.
void Asura_ClientContainerInstance_GraphColour::Initialise(Asura_ClientContainerInstance_Collection* pxParentCollection,
														   Asura_ContainerInstance_Properties& xData)
{
	PARENT::Initialise(pxParentCollection, xData);
}

//////////////
// GetParam //
//////////////
const Asura_Container_GraphColour_Param& Asura_ClientContainerInstance_GraphColour::GetParam() const
{
	Asura_Assert(m_pxContainer, "Null container!");
	const Asura_Container_GraphColour* const pxContainer = static_cast<const Asura_Container_GraphColour*>(m_pxContainer);
	return pxContainer->GetParam();
}

////////////////////
// GetGraphColour //
////////////////////
// Get the current graph colour. Caller should initialise the output data before calling.
// Returns true if results have been returned (i.e there are enabled GraphColour containers).
bool Asura_ClientContainerInstance_GraphColour::GetGraphColour(Asura_Container_GraphColour_Output& xOutput)
{
	const Asura_Container_GraphColour_Param& xParam = GetParam();
	const bool bColourRangeEnabled = (xParam.IsColourEnabled() || xParam.IsRangeEnabled());

	if (!bColourRangeEnabled && !xParam.IsFlickerEnabled())
	{
		// Nothing is enabled.
		m_uCacheFrameNumber = 0xffffffff;
		return false;
	}

	if (m_uCacheFrameNumber == Asura_Timers::GetFrameNumber())
	{
		xOutput *= m_xCacheOutput;
		return true;
	}

	// Reset the cache.
	m_xCacheOutput.Reset();

	// If the duration is zero, then just stay with the initial random value.
	// Otherwise advance the timer looping if needed.
	const float fDuration = xParam.m_fGraphDuration;
	float fTimer = m_fOffset;

	if (fDuration > 0.0f)
	{
		const float fBaseTime = Asura_Timers::GetLocalScaledTime() - m_fGraphStartTime;
		if (xParam.IsLooped())
		{
			// Looping case, with random initial offset.
			fTimer = Asura_Maths::FloatMod ((fBaseTime/fDuration + m_fOffset), 1.0f);
		}
		else
		{
			// Non-looping case.
			fTimer = Asura_Maths::ClampToRange(fBaseTime/fDuration, 0.0f, 1.0f);
		}
	}

	// Colour (RGB) and range (A).
	if (bColourRangeEnabled)
	{
		const Asura_RGBA xRGBA = xParam.m_xColourRangeGraph.GetValueAtTime(fTimer);

		if (xParam.IsColourEnabled())
		{
			m_xCacheOutput.m_xColour.m_fR =  xRGBA.m_fR;
			m_xCacheOutput.m_xColour.m_fG =  xRGBA.m_fG;
			m_xCacheOutput.m_xColour.m_fB =  xRGBA.m_fB;
		}

		if (xParam.IsRangeEnabled())
		{
			m_xCacheOutput.m_xColour.m_fA = xRGBA.m_fA;
		}
	}

	// Flicker (intensity)
	if (xParam.IsFlickerEnabled())
	{
		// (SB - 13/03/08) Scale flicker amount by time scale
		// This is so flickering stops when paused and perceptively the
		// flicker slows when timescale is decreased
		const float fFlickerAmount = xParam.m_xFlickerGraph.GetValueAtTime(fTimer) * Asura_Timers::GetGameTimeScale();
		const float fMin = Asura_Maths::ClampToRange(1.0f - fFlickerAmount, 0.0f, 1.0f);

		m_xCacheOutput.m_fIntensity = Asura_Random::GetFloat(fMin, 1.0f);
	}

	// Cache the result
	m_uCacheFrameNumber = Asura_Timers::GetFrameNumber();
	xOutput *= m_xCacheOutput;
	return true;
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Health
///////////////////////////////////////////////////////////////////////////////

Asura_ClientContainerInstance_Health::Asura_ClientContainerInstance_Health(const Asura_Container_Health* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
, m_fHealth(pxContainer ? pxContainer->GetInitialHealth() : 0.0f)
{
}

void Asura_ClientContainerInstance_Health::SetHealth(float fHealth)
{
	m_fHealth = fHealth;
	
	const Asura_Container_Health* pxContainer = static_cast<const Asura_Container_Health*>(m_pxContainer);
	if (pxContainer)
	{
		m_fHealth = Asura_Maths::ClampToRange(m_fHealth, 0.0f, pxContainer->GetMaxHealth());
	}
}

//////////////////////
// GetHealthPercent //
//////////////////////
// Get the health as a percentage of the maximum value.
float Asura_ClientContainerInstance_Health::GetHealthPercent() const
{
	const Asura_Container_Health* const pxContainer = static_cast<const Asura_Container_Health*>(m_pxContainer);
	if (pxContainer)
	{
		const float fMax = pxContainer->GetMaxHealth();
		if (fMax > 0.0f && m_fHealth != fMax)
		{
			const float fPercent = 100.0f * m_fHealth / fMax;
			return Asura_Maths::ClampToRange(fPercent, 0.0f, 100.0f);
		}
	}

	return 100.0f;
}

//////////////////////////////
// HasDamageOverrideChanged //
//////////////////////////////
// Return true if the parent collection's damage override state will change as a
// result of the change in health.
bool Asura_ClientContainerInstance_Health::HasDamageOverrideChanged(float fOldHealthPercent, float fNewHealthPercent) const
{
	if (!m_pxParentCollection) return false;

	return m_pxParentCollection->HasDamageOverrideChanged(fOldHealthPercent, fNewHealthPercent);
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Physics
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_Physics::Asura_ClientContainerInstance_Physics(const Asura_Container_Physics* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
{
}

const Asura_Container_Physics_Param& Asura_ClientContainerInstance_Physics::GetParam() const
{
	Asura_Assert(m_pxContainer, "Null container pointer");
	const Asura_Container_Physics* pxProperties = static_cast<const Asura_Container_Physics*>(m_pxContainer);
	return pxProperties->GetParam();
}

///////////////////////////////////////////////////////////////////////////////
// Asura_ClientContainerInstance_Blueprint
///////////////////////////////////////////////////////////////////////////////
Asura_ClientContainerInstance_Blueprint::Asura_ClientContainerInstance_Blueprint(const Asura_Container_Blueprint* pxContainer)
: Asura_ClientContainerInstance(pxContainer)
{
}

Asura_Hash_ID Asura_ClientContainerInstance_Blueprint::GetBlueprintTypeHash() const
{
	Asura_Assert(m_pxContainer, "Null container pointer");
	const Asura_Container_Blueprint* pxBPContainer = static_cast<const Asura_Container_Blueprint*>(m_pxContainer);
	return pxBPContainer->GetBlueprintTypeHash();
}

Asura_Hash_ID Asura_ClientContainerInstance_Blueprint::GetBlueprintHash() const
{
	Asura_Assert(m_pxContainer, "Null container pointer");
	const Asura_Container_Blueprint* pxBPContainer = static_cast<const Asura_Container_Blueprint*>(m_pxContainer);
	return pxBPContainer->GetBlueprintHash();
}

#endif //ifdef ASURA_USE_CONTAINER_SYSTEM

