///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_Container_Messages.cpp
//
// Created by Martin G.Harris (22/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Asura_Container_Messages.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Debugging/Asura_Assert.h"
#include "Entity/Container/Asura_ClientEntity_Container.h"
#include "Entity/Container/Asura_Container_System.h"
#include "Entity/Container/Asura_ServerEntity_Container.h"
#include "Network/Asura_Message.h"
#include "Physics/Asura_Physics_Object.h"

#ifdef ASURA_SUPPORT_NETMSG_SYSTEM
#include "Network/Asura_NetMsg.h"
#endif //#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

///////////////////////////////////////////////////////////////////////////////
// Asura_Container_PosOrientVel
///////////////////////////////////////////////////////////////////////////////

////////////////////////
// SetDataFromPhysics //
////////////////////////
// Set the data from the physics object.
void Asura_Container_PosOrientVel::SetDataFromPhysics(const Asura_Physics_Object* pxPhysics)
{
	Asura_Assert(pxPhysics, "Null physics pointer");
	if (!pxPhysics) return;
	
	m_xPosition = pxPhysics->GetPosition();
	m_xOrient.FromMatrix(pxPhysics->GetOrientation());
	
	if (pxPhysics->IsAsleep() || pxPhysics->GetType() != ASURA_PHYSICS_OBJECT_DYNAMIC)
	{
		m_xLinVelocity.Zero();
		m_xAngVelocity.Zero();
	}
	else
	{
		pxPhysics->GetVelocities(m_xLinVelocity, m_xAngVelocity);
	}
}


///////////////////////////////////////////////////////////////////////////////
// Asura_Message_Container_Create
///////////////////////////////////////////////////////////////////////////////

////////////
// FillIn //
////////////
void Asura_Message_Container_Create::FillIn(const Asura_ServerEntity_Container* pxServer)
{
	Asura_Assert(pxServer, "Null server pointer");
	if (!pxServer) return;
	
	m_uContainerHash = pxServer->GetContainerHash();
		
	const u_int uNumInputs		 = pxServer->GetNumInputs();
	const u_int uNumPhysicsParam = pxServer->GetNumPhysicsParam();
	const u_int uNumFloatParam	 = pxServer->GetNumFloatParam();
	const u_int uNumProjectBytes = pxServer->GetNumBytesForProjectCreateMessageData();

	Asura_Assert(uNumInputs		  <= 0xff, "Too many inputs");
	Asura_Assert(uNumPhysicsParam <= 0xff, "Too many physics parameters");
	Asura_Assert(uNumFloatParam	  <= 0xff, "Too many float parameters");
	Asura_Assert(uNumProjectBytes <= 0xff, "Too much project-specific data in container create message");

	m_ucNumInputStates  = static_cast<u_char>(uNumInputs);
	m_ucNumPhysicsParam = static_cast<u_char>(uNumPhysicsParam);
	m_ucNumFloatParam   = static_cast<u_char>(uNumFloatParam);
	m_ucNumProjectBytes = static_cast<u_char>(uNumProjectBytes);

	m_xOriginPosOri.m_xPosition = pxServer->GetPosition();
	m_xOriginPosOri.m_xOrient   = pxServer->GetOrientationQuat();

	void* pData = (this + 1);

	// Append any input state hashes.
	if (m_ucNumInputStates > 0)
	{
		Asura_Hash_ID* puStateHash = reinterpret_cast<Asura_Hash_ID*>(pData);
		for (u_int u=0; u<uNumInputs; ++u)
		{
			*puStateHash = pxServer->GetInputState().GetStateHash(u);
			++puStateHash;
		}
		
		// Sanity check. Make sure that the data written matches the allocated message size.
		Asura_Assert( reinterpret_cast<char*>(puStateHash) == (reinterpret_cast<char*>(this) + GetSize(m_ucNumInputStates, 0, 0, 0)), "Error: Container create message size mismatch!" );

		pData = puStateHash;
	}

	// Append any physics parameters (i.e. position, orientation data).
	if (m_ucNumPhysicsParam > 0)
	{
		Asura_Container_PhysicsParam* pxPhysicsData = reinterpret_cast<Asura_Container_PhysicsParam*>(pData);
		pxServer->FillPhysicsParam(pxPhysicsData, m_ucNumPhysicsParam);

		// Sanity check. Make sure that the data written matches the allocated message size.
		Asura_Assert( reinterpret_cast<char*>(pxPhysicsData) == (reinterpret_cast<char*>(this) + GetSize(m_ucNumInputStates, m_ucNumPhysicsParam, 0, 0)), "Error: Container create message size mismatch!" );

		pData = pxPhysicsData;
	}
	
	// Append any float parameters (e.g. health values of collections).
	if (m_ucNumFloatParam > 0)
	{
		Asura_Container_FloatParam* pxFloatData = reinterpret_cast<Asura_Container_FloatParam*>(pData);
		pxServer->FillFloatParam(pxFloatData);

		// Sanity check. Make sure that the data written matches the allocated message size.
		Asura_Assert( reinterpret_cast<char*>(pxFloatData) == (reinterpret_cast<char*>(this) + GetSize(m_ucNumInputStates, m_ucNumPhysicsParam, m_ucNumFloatParam, 0)), "Error: Container create message size mismatch!" );

		pData = pxFloatData;
	}

	// Append any project-specific data.
	if (m_ucNumProjectBytes > 0)
	{
		// Zero the project data, and then fill it in.
		Asura_Mem::Set(pData, 0, m_ucNumProjectBytes);
		pxServer->FillProjectCreateMessageData(pData);
	}
}

/////////////
// FillOut //
/////////////
void Asura_Message_Container_Create::FillOut(Asura_ClientEntity_Container* pxClient, float fServerTime) const
{
	Asura_Assert(pxClient, "Null client pointer");
	if (pxClient)
	{
		pxClient->m_uContainerHash = m_uContainerHash;
		pxClient->ProcessNewPositionAndOrientationData(m_xOriginPosOri.m_xPosition, m_xOriginPosOri.m_xOrient, fServerTime);
		
		const void* const pStartData = (this + 1);
		const Asura_Hash_ID* const puStateHash = reinterpret_cast<const Asura_Hash_ID*>(pStartData);
		const Asura_Container_PhysicsParam* const pxPhysicsData = reinterpret_cast<const Asura_Container_PhysicsParam*>(puStateHash + m_ucNumInputStates);
		const Asura_Container_FloatParam* const pxFloatData = reinterpret_cast<const Asura_Container_FloatParam*>(pxPhysicsData + m_ucNumPhysicsParam);

		const void* const pProjectData = reinterpret_cast<const void*>(pxFloatData + m_ucNumFloatParam);

		// Make a copy of any float parameters (e.g. health). Must do this before setting the state.
		pxClient->ProcessFloatParam(m_ucNumFloatParam, pxFloatData, fServerTime);

		// Copy any physics pos/ori before setting up the state.
		pxClient->ProcessPhysicsParam(m_ucNumPhysicsParam, pxPhysicsData, fServerTime);

		// Copy any project-specific data.
		if (m_ucNumProjectBytes > 0)
		{
			pxClient->ProcessProjectCreateMessageData(m_ucNumProjectBytes, pProjectData);
		}

		// Copy input state hashes.
		Asura_Container_InputState xInputState;
		xInputState.SetNumInputs(m_ucNumInputStates);
		for (u_int u=0; u<xInputState.GetNumInputs(); ++u)
		{
			xInputState.SetStateHash(u, puStateHash[u]);
		}

		// Update state creating container instances as necessary.
		Asura_Container_System::SetDoingCreate(true);
		const bool bChanged = pxClient->SetInputState(xInputState, fServerTime);

		if (bChanged)
		{
			// The state changed. Call this after updating any physics data.
			pxClient->ProcessStateChange();
		}

		Asura_Container_System::SetDoingCreate(false);
	}
}


///////////////////////////////////////////////////////////////////////////////
// Asura_Message_Container_ClientUpdate
///////////////////////////////////////////////////////////////////////////////

////////////
// FillIn //
////////////
void Asura_Message_Container_ClientUpdate::FillIn(const Asura_ServerEntity_Container* pxServer)
{
	Asura_Assert(pxServer, "Null server pointer");
	if (!pxServer) return;
	
	Asura_Assertf(pxServer->GetNumPhysicsParam() == 0, ("Container entity (%u) using wrong update message!", pxServer->GetGuid()));

	const Asura_Container_InputState& xInputState = pxServer->GetInputState();
	for (u_int u = 0; u <uASURA_MAX_NUM_CONTAINER_INPUTS; ++u)
	{
		m_auStateHash[u] = xInputState.GetStateHash(u);
	}
}

/////////////
// FillOut //
/////////////
void Asura_Message_Container_ClientUpdate::FillOut(Asura_ClientEntity_Container* pxClient, float fServerTime) const
{
	Asura_Assert(pxClient, "Null client pointer");
	if (pxClient)
	{
		// Number of input states was sent in the create message and never changes.
		Asura_Container_InputState xInputState;
		xInputState.Set(pxClient->GetNumInputs(), m_auStateHash);

		const bool bChanged = pxClient->SetInputState(xInputState, fServerTime);
		if (bChanged)
		{
			// The state changed.
			pxClient->ProcessStateChange();
		}
	}
}

////////////
// NetMsg //
////////////
#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

void Asura_Message_Container_ClientUpdate::RegisterFields(u_short usMsgID, u_short usEntityClassification)
{
	Asura_Assert(usMsgID == ASURA_MSG_CLIENTUPDATE, "invalid msg ID");
	Asura_Assert(!Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification), "already exists");

	u_char ucNumFields = GetNumFields();

	Asura_NetMsg* pxNetMsg = new Asura_NetMsg(usMsgID, usEntityClassification, ucNumFields, sizeof(Asura_Message_Container_ClientUpdate));
	RegisterFields(pxNetMsg);
	pxNetMsg->EndRegistration();
}

void Asura_Message_Container_ClientUpdate::RegisterFields(class Asura_NetMsg* pxNetMsg)
{
	Asura_Message_Container_ClientUpdate xDummy;

	Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_uint, m_auStateHash, uASURA_MAX_NUM_CONTAINER_INPUTS, Asura_Threshold_AnyChange, 0.0f);
}
#endif//#ifdef ASURA_SUPPORT_NETMSG_SYSTEM


///////////////////////////////////////////////////////////////////////////////
// Asura_Message_Container_ClientUpdate_Extended1
///////////////////////////////////////////////////////////////////////////////

////////////
// FillIn //
////////////
void Asura_Message_Container_ClientUpdate_Extended1::FillIn(const Asura_ServerEntity_Container* pxServer)
{
	Asura_Assert(pxServer, "Null server pointer");
	if (!pxServer) return;
	
	const Asura_Container_InputState& xInputState = pxServer->GetInputState();
	for (u_int u = 0; u <uASURA_MAX_NUM_CONTAINER_INPUTS; ++u)
	{
		m_auStateHash[u] = xInputState.GetStateHash(u);
	}

	const u_int uNumServerPhysicsParam = pxServer->GetNumPhysicsParam();
	Asura_Assertf(uNumServerPhysicsParam == 1, ("Container entity (%u) has wrong number of dynamic physics objects!", pxServer->GetGuid()));

	m_xPhysicsParam.Reset();
	
	if (uNumServerPhysicsParam > 0)
	{
		Asura_Container_PhysicsParam* pxPhysicsData = &m_xPhysicsParam;
		pxServer->FillPhysicsParam(pxPhysicsData, 1);
	}
}

/////////////
// FillOut //
/////////////
void Asura_Message_Container_ClientUpdate_Extended1::FillOut(Asura_ClientEntity_Container* pxClient, float fServerTime) const
{
	Asura_Assert(pxClient, "Null client pointer");
	if (pxClient)
	{
		// Number of input states was sent in the create message and never changes.
		Asura_Container_InputState xInputState;
		xInputState.Set(pxClient->GetNumInputs(), m_auStateHash);

		const bool bChanged = pxClient->SetInputState(xInputState, fServerTime);

		// Pos/ori/ID for the dynamic physics block.
		// Must process this after setting the state.
		pxClient->ProcessPhysicsParam(1, &m_xPhysicsParam, fServerTime);

		if (bChanged)
		{
			// The state changed. Call this after updating any physics data.
			pxClient->ProcessStateChange();
		}
	}
}

////////////
// NetMsg //
////////////
#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

DEBUGVAR float dbg_fAsura_Container_PositionThreshold	 = 0.01f;
DEBUGVAR float dbg_fAsura_Container_OrientationThreshold = 0.01f;

void Asura_Message_Container_ClientUpdate_Extended1::RegisterFields(u_short usMsgID, u_short usEntityClassification)
{
	Asura_Assert(usMsgID == ASURA_MSG_CONTAINER_CLIENTUPDATE_EXTENDED1, "invalid msg ID");
	Asura_Assert(!Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification), "already exists");

	u_char ucNumFields = GetNumFields();

	Asura_NetMsg* pxNetMsg = new Asura_NetMsg(usMsgID, usEntityClassification, ucNumFields, sizeof(Asura_Message_Container_ClientUpdate_Extended1));
	RegisterFields(pxNetMsg);
	pxNetMsg->EndRegistration();
}

void Asura_Message_Container_ClientUpdate_Extended1::RegisterFields(class Asura_NetMsg* pxNetMsg)
{
	Asura_Message_Container_ClientUpdate_Extended1 xDummy;

	Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_uint, m_auStateHash, uASURA_MAX_NUM_CONTAINER_INPUTS, Asura_Threshold_AnyChange, 0.0f);

	Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_CompressedPosition, m_xPhysicsParam.m_xPosition, 1, Asura_Threshold_Magnitude, dbg_fAsura_Container_PositionThreshold);

	Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_CompressedQuat, m_xPhysicsParam.m_xOrient, 1, Asura_Threshold_Magnitude, dbg_fAsura_Container_OrientationThreshold);

	Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_uint, m_xPhysicsParam.m_uUniqueID, 1, Asura_Threshold_AnyChange, 0.0f);
	
}
#endif//#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

///////////////////////////////////////////////////////////////////////////////
// Asura_Message_Container_ClientUpdate_Extended
///////////////////////////////////////////////////////////////////////////////

////////////
// FillIn //
////////////
void Asura_Message_Container_ClientUpdate_Extended::FillIn(const Asura_ServerEntity_Container* pxServer)
{
	Asura_Assert(pxServer, "Null server pointer");
	if (pxServer)
	{
		const Asura_Container_InputState& xInputState = pxServer->GetInputState();
		for (u_int u = 0; u <uASURA_MAX_NUM_CONTAINER_INPUTS; ++u)
		{
			m_auStateHash[u] = xInputState.GetStateHash(u);
		}

		for (u_int u=0; u<uMAX_DYNAMIC_PHYSICS_PER_CONTAINER_ENTITY; ++u)
		{
			m_axPhysicsParam[u].Reset();
		}

		const u_int uNumServerPhysicsParam = pxServer->GetNumPhysicsParam();
		m_uNumPhysicsParam = Asura_Maths::ClampToRange(uNumServerPhysicsParam, 0u, uMAX_DYNAMIC_PHYSICS_PER_CONTAINER_ENTITY);
		Asura_Assertf(m_uNumPhysicsParam == uNumServerPhysicsParam, ("Container entity (%u) has too many dynamic physics objects!", pxServer->GetGuid()));

		if (m_uNumPhysicsParam > 0)
		{
			Asura_Container_PhysicsParam* pxPhysicsData = &m_axPhysicsParam[0];
			pxServer->FillPhysicsParam(pxPhysicsData, m_uNumPhysicsParam);
		}
	}
}

/////////////
// FillOut //
/////////////
void Asura_Message_Container_ClientUpdate_Extended::FillOut(Asura_ClientEntity_Container* pxClient, float fServerTime) const
{
	Asura_Assert(pxClient, "Null client pointer");
	if (pxClient)
	{
		Asura_Container_InputState xInputState;
		xInputState.Set(pxClient->GetNumInputs(), m_auStateHash);

		const bool bChanged = pxClient->SetInputState(xInputState, fServerTime);

		// Pos/ori/ID for any dynamic physics blocks.
		// Must process this after setting the state.
		const u_int uNum = Asura_Maths::ClampToRange(m_uNumPhysicsParam, 0u, uMAX_DYNAMIC_PHYSICS_PER_CONTAINER_ENTITY);

		if (uNum > 0)
		{
			pxClient->ProcessPhysicsParam(uNum, &m_axPhysicsParam[0], fServerTime);
		}

		if (bChanged)
		{
			// The state changed. Call this after updating any physics data.
			pxClient->ProcessStateChange();
		}
	}
}

////////////
// NetMsg //
////////////
#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

void Asura_Message_Container_ClientUpdate_Extended::RegisterFields(u_short usMsgID, u_short usEntityClassification)
{
	Asura_Assert(usMsgID == ASURA_MSG_CONTAINER_CLIENTUPDATE_EXTENDED, "invalid msg ID");
	Asura_Assert(!Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification), "already exists");

	u_char ucNumFields = GetNumFields();

	Asura_NetMsg* pxNetMsg = new Asura_NetMsg(usMsgID, usEntityClassification, ucNumFields, sizeof(Asura_Message_Container_ClientUpdate_Extended));
	RegisterFields(pxNetMsg);
	pxNetMsg->EndRegistration();
}

void Asura_Message_Container_ClientUpdate_Extended::RegisterFields(class Asura_NetMsg* pxNetMsg)
{
	Asura_Message_Container_ClientUpdate_Extended xDummy;

	Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_uint, m_auStateHash, uASURA_MAX_NUM_CONTAINER_INPUTS, Asura_Threshold_AnyChange, 0.0f);

	Asura_NetMsg_RegisterIntFieldBits(pxNetMsg, xDummy, Asura_FieldType_uint_bits, uBITS_FOR_MAX_DYNAMIC_PHYSICS, m_uNumPhysicsParam, 1, Asura_Threshold_AnyChange);

	for (u_int u=0; u<uMAX_DYNAMIC_PHYSICS_PER_CONTAINER_ENTITY; ++u)
	{
		Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_CompressedPosition, m_axPhysicsParam[u].m_xPosition, 1, Asura_Threshold_Magnitude, dbg_fAsura_Container_PositionThreshold);

		Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_CompressedQuat, m_axPhysicsParam[u].m_xOrient, 1, Asura_Threshold_Magnitude, dbg_fAsura_Container_OrientationThreshold);

		Asura_NetMsg_RegisterField(pxNetMsg, xDummy, Asura_FieldType_uint, m_axPhysicsParam[u].m_uUniqueID, 1, Asura_Threshold_AnyChange, 0.0f);
	}
}
#endif//#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

#endif // ASURA_USE_CONTAINER_SYSTEM

