///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_Container_System.h
//
// Created by Martin G.Harris (6/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_CONTAINER_SYSTEM_H__
#define __ASURA_CONTAINER_SYSTEM_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_HashTree.h"

// Forward declarations
class Asura_ClientContainerInstance;
class Asura_ServerContainerInstance;
class Asura_Container;


////////////////////////////
// Asura_Container_System //
////////////////////////////
class Asura_Container_System
{
public:
	static void Initialise();
	static void Reset();
	static void Shutdown();

	// Find the specified container.
	static Asura_Container* FindContainer(Asura_Hash_ID uContainerHash);

	// Create an instance of the specified container. Returns NULL if container is not found.	
	static Asura_ClientContainerInstance* CreateClientInstance(Asura_Hash_ID uContainerHash);
	static Asura_ServerContainerInstance* CreateServerInstance(Asura_Hash_ID uContainerHash);
	
	// As CreateClientInstance, but only creates certain basic types of container (i.e. health and collections).
	static Asura_ClientContainerInstance* CreateClientInstanceBase(Asura_Hash_ID uContainerHash);
	// Returns true if it is a basic container type (i.e. health or collection).
	static bool IsBaseClassification(u_int uClassification);

	// Destroy the instance.
	static void DestroyInstance(Asura_ClientContainerInstance* pxInstance);
	static void DestroyServerInstance(Asura_ServerContainerInstance* pxInstance);

	// Return the container tree.
	static Asura_TemplateHashTree<Asura_Container*>& GetTree() { return s_xContainerTree; }

	// Note: this flag behaves differently from those below; it isn't set for very long
	// and is only for internal use by the Asura_ClientContainerInstance classes.
	// Are we in the process of doing a state change (i.e. not destruction/deactivation)?
	static bool IsDoingStateChange() { return s_bDoingStateChange; }
	static void SetDoingStateChange(bool bChanging) { s_bDoingStateChange = bChanging; }
	
	// Note: the flags below can be used by the Asura_ClientContainerInstance classes
	// in their constructors/destructors, and also in PostChangeInitialisation.
	// If none of them are set then it must be an ordinary state change.

	// Are we in the process of activating a container entity?
	static bool IsDoingActivate() { return s_bDoingActivate; }
	static void SetDoingActivate(bool bActivating) { s_bDoingActivate = bActivating; }

	// Are we in the process of deactivating a container entity?
	static bool IsDoingDeactivate() { return s_bDoingDeactivate; }
	static void SetDoingDeactivate(bool bDeactivating) { s_bDoingDeactivate = bDeactivating; }

	// Are we in the process of creating a container entity?
	static bool IsDoingCreate() { return s_bDoingCreate; }
	static void SetDoingCreate(bool bCreating) { s_bDoingCreate = bCreating; }
	
	// Are we in the process of creating or activating a container entity?
	// Note: used by some of the Asura_ClientContainerInstance classes to differentiate
	// between create/load/activate and state transitions.
	static bool IsDoingCreateOrActivate() { return s_bDoingCreate || s_bDoingActivate; }

	// Are we in the process of doing a damage override change?
	static bool IsDoingDamageOverrideChange() { return s_bDoingDamageOverrideChange; }
	static void SetDoingDamageOverrideChange(bool bChanging) { s_bDoingDamageOverrideChange = bChanging; }

private:
	// Add container to the tree. Return true if added successfully.
	static bool AddContainer(Asura_Container* pxContainer);

	// Delete all of the containers and reset the tree.
	static void DeleteAllContainers();

	// Delete specified container and remove from tree. Return true if deleted.
	static bool DeleteContainer(Asura_Hash_ID uContainerHash);

private:
	static Asura_TemplateHashTree<Asura_Container*> s_xContainerTree;
	
	static bool s_bDoingStateChange;
	static bool s_bDoingActivate;
	static bool s_bDoingDeactivate;
	static bool s_bDoingCreate;
	static bool s_bDoingDamageOverrideChange;

	// Friends
	friend class Tools_Container_System;
	friend class Asura_Chunk_Container;
};


#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__ASURA_CONTAINER_SYSTEM_H__

