/**
*
* \file Asura_ClientEntity_CutsceneController.cpp
*
* Created by Kevin Lea - 15/04/2002 12:13:56
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "Entity/Cutscene/Asura_ClientEntity_CutsceneController.h"


#include "Camera/Asura_Camera.h"
#include "Cutscene/Asura_Cutscene_Actor.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Entity/Cutscene/Asura_ServerEntity_CutsceneController.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_NetMsg.h"
#include "Rendering/Asura_Render.h"


Asura_ClientEntity_CutsceneController::Asura_ClientEntity_CutsceneController()
{
	Classification = AsuraEntityClass_CutsceneController;
	m_bInitialised = false;
	m_bPlaying = false;
	m_uCutsceneHash = ASURA_HASH_ID_UNSET;
    m_pxCutscene = NULL;
}

Asura_ClientEntity_CutsceneController::~Asura_ClientEntity_CutsceneController()
{
	//Stop the cutscene if it still exists and is playing.
	//(During shutdown the cutscene will already have been deleted ,so m_pxCutscene won't be valid)
	if(m_uCutsceneHash != ASURA_HASH_ID_UNSET)
	{
		Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByHash(m_uCutsceneHash);
		if (pxCutscene && pxCutscene->IsPlaying())
		{
			Asura_Assert(pxCutscene == m_pxCutscene,"Found a different cutscene from expected one");
			pxCutscene->Finish();
		}
	}

	RemoveFromUpdateableClientList(this, false);
}

void Asura_ClientEntity_CutsceneController::Update(void)
{
    Asura_Assert(m_pxCutscene, "NULL cutscene pointer!");
	//Only update cutscene if it was started via this entity
	if(m_bPlaying)
	{
		if (m_pxCutscene && m_pxCutscene->IsPlaying())
		{
			m_pxCutscene->Update();
    	}
		else
		{
			m_bPlaying = false;
			// NY - 19 June 2008 - message back to the server so it doesn't have to directly poll the client
			Asura_Message* pxMsg = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
			Send(pxMsg, GetGuid());
			//stop updating
			RemoveFromUpdateableClientList(this, false);
		}
	}
}

void Asura_ClientEntity_CutsceneController::Render(void)
{
    Asura_Assert(m_pxCutscene, "NULL cutscene pointer!");
	//Only update cutscene if it was started via this entity
	if(m_bPlaying)
	{
		if (m_pxCutscene && m_pxCutscene->IsPlaying())
		{
			m_pxCutscene->Render();
		}
		else
		{
			m_bPlaying = false;
			// NY - 19 June 2008 - message back to the server so it doesn't have to directly poll the client
			Asura_Message* pxMsg = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
			Send(pxMsg, GetGuid());
			//stop updating
			RemoveFromUpdateableClientList(this, false);
		}
	}
}

bool Asura_ClientEntity_CutsceneController::HandleMessage(const Asura_Message* pxMessage)
{
	Asura_Assert(pxMessage,"");
	if (!pxMessage) return false;
	
	switch(pxMessage->GetMsgID())
	{
		default:
		{
			break;
		}

        case ASURA_MSG_CUTSCENECONTROLLER_PROFILE_CUTSCENE: // Intentional fallthrough.
        case ASURA_MSG_CUTSCENECONTROLLER_ACTIVATE:
		{
			const Asura_Msg_ActivateCutscene* pxData = (const Asura_Msg_ActivateCutscene*) pxMessage->GetData();

			if(!m_bInitialised)
			{
				//Need to find out which cutscene this controller controls from the message
				//Get the cutscene
				m_pxCutscene = Asura_Cutscene_System::GetCutsceneByHash( pxData->m_uHash );
				m_uCutsceneHash = pxData->m_uHash;

				if(!m_pxCutscene)
				{
					return false;
				}
				Asura_Assert(m_pxCutscene!=NULL,"Unable to find cutscene");

// (MP - 10.03.2003) We don't want to do this outside of tools. The level
// exports should contain the necessary data.
#ifdef ASURA_TOOLS
				m_pxCutscene->LoadFiles();
				m_pxCutscene->m_uGuid = GetGuid();
				Asura_Texture_Management::LoadTextures();
#elif defined(ASURA_USE_STREAMING)
				m_pxCutscene->m_uGuid = GetGuid();
#endif //defined(ASURA_USE_STREAMING)

				m_bInitialised = true;
			}
			
			Asura_Assert(!m_pxCutscene->IsPlaying(), "cutscene already playing when activated");
            if (!m_pxCutscene->IsPlaying())
			{
                bool bProfile = ( ASURA_MSG_CUTSCENECONTROLLER_PROFILE_CUTSCENE == pxMessage->GetMsgID() );
				float fTimer = 0;
				if(!bProfile)
				{
					fTimer = pxData->m_fTimer;
				}

                
				const bool bSkipMessagesAfterLoad = ((pxData->m_ucFlags & Asura_Msg_ActivateCutscene::FLAG_SKIP_MESSSAGES_AFTER_LOAD) != 0);
				const bool bOldUseSoundTrack = m_pxCutscene->UseSoundTrack;
				const bool bOldCameraDisable = Asura_ClientCutscene_Actor_Camera::Disable;

				if (bSkipMessagesAfterLoad &&
					fTimer >= m_pxCutscene->CutsceneEndTime &&
					m_pxCutscene->CutsceneEndTime != 0.0f &&
					(m_pxCutscene->m_iCutsceneFlags & AsuraCutsceneFlag_LoopCutscene) == 0)
				{
					// MGH 16/1/2007: The cutscene is being initialised, but will finish immediately,
					// don't start playing the soundtrack. Don't mess with the camera either.
					m_pxCutscene->UseSoundTrack = false;
					Asura_ClientCutscene_Actor_Camera::Disable = true;
				}

                m_pxCutscene->Initialise();
				m_pxCutscene->m_uEnabledEvents = pxData->m_uEnabledEvents;
                m_pxCutscene->SetProfiling( bProfile );
				m_bPlaying = true;
				//start updating
				AddToUpdateableClientList(this);

				if(fTimer!=0)
				{
					//skip to the correct part of the cutscene
					m_pxCutscene->m_bSkipMessagesAfterLoad = bSkipMessagesAfterLoad;
					m_pxCutscene->AdvanceToTime(fTimer);
					m_pxCutscene->m_bSkipMessagesAfterLoad = false;
				}

				// Restore the settings.
				if (bSkipMessagesAfterLoad)
				{
					m_pxCutscene->UseSoundTrack = bOldUseSoundTrack;
					Asura_ClientCutscene_Actor_Camera::Disable = bOldCameraDisable;
				}
                
			}
			
			return true;
		}
		case ASURA_MSG_DEACTIVATE:
		{
			if(m_pxCutscene)
			{
				if (m_pxCutscene->IsPlaying())
				{
					m_pxCutscene->Finish();
					m_bPlaying = false;
				}
			}
			//stop updating
			RemoveFromUpdateableClientList(this, false);
			return true;
		}


		case ASURA_MSG_CUTSCENECONTROLLER_CLIENTUPDATE:
		{
			Asura_Assert(0,"Unable to find cutscene");
			return true;
		}

		case ASURA_MSG_CUTSCENECONTROLLER_UPDATEEVENTSTATE :
		{
			if(m_pxCutscene)
			{
				const Asura_Message_CutsceneEventState* pxCutsceneMessage = static_cast<const Asura_Message_CutsceneEventState*>(pxMessage->GetData());
				m_pxCutscene->m_uEnabledEvents = pxCutsceneMessage->m_uEnabledEvents;
				
			}
			return true;
		}

		case ASURA_MSG_CUTSCENE_SETCAMERA :
		{
			const Asura_Message_Cutscene_SetCamera* pxCutsceneMessage = static_cast<const Asura_Message_Cutscene_SetCamera*>(pxMessage->GetData());

			if(m_pxCutscene && m_pxCutscene->IsPlaying() && pxCutsceneMessage->m_uCameraHash)
			{
				m_pxCutscene->SetActiveCameraByHash(pxCutsceneMessage->m_uCameraHash);
			}
			return true;
		}
	}

	Asura_Assert(0,"Unhandled message");
	return false; 
}


// NetMsg Registration
#ifdef ASURA_NETMSG_PACK_ALL_DATA
void Asura_Msg_ActivateCutscene::RegisterFields(Asura_NetMsg* pxNetMsg) const
{
	Asura_Assert(pxNetMsg->GetMsgID() == ASURA_MSG_CUTSCENECONTROLLER_ACTIVATE, "unexpected message ID");
	Asura_Assertf(pxNetMsg->GetEntityClassification() == AsuraEntityClass_NotSet, 
		("unexpected entity classification (%s)", Asura_Network::GetEntityClassificationDesc(pxNetMsg->GetEntityClassification())));

	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_uint, m_uHash );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_float, m_fTimer );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_uint, m_uEnabledEvents );
	Asura_NetMsg_RegisterIntFieldSimpleBits( m_ucFlags, 2 );
}
#endif // 	ASURA_NETMSG_PACK_ALL_DATA
