
////////////////////////////////////////////////////////////////////////////////
//
// Asura_ClientEntity_FMVController.cpp
// Author    : Jason Turner
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/FMVController/Asura_ClientEntity_FMVController.h"
#include "Entity/FMVController/Asura_ClientEntity_FMVSpeaker.h"
#include "Entity/FMVController/Asura_FMVController_Messages.h"
#include "FMV/Asura_VideoToTexture.h"
#include "Network/Asura_Message.h"

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#endif

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

Asura_ClientEntity_FMVController::Asura_ClientEntity_FMVController(Asura_Guid uGuid)
: Asura_ClientEntity_Activatable(uGuid)
, m_uFlags(0)
, m_bTriggered(false)
{
	Classification = AsuraEntityClass_FMVController;

	m_pxVideoRenderer = new Asura_VideoToTexture;
	Asura_Assert(m_pxVideoRenderer, "Asura_ClientEntity_FMVController::m_pxVideoRenderer is NULL!");

	m_xLinkedSpeakerGuids.Clear();

	AddToUpdateableClientList(this);
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

Asura_ClientEntity_FMVController::~Asura_ClientEntity_FMVController()
{
	Destroy();
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::Destroy()
{
	m_xLinkedSpeakerGuids.Clear();
	if (m_pxVideoRenderer)
	{
		delete m_pxVideoRenderer;
		m_pxVideoRenderer = NULL;
	}

	m_bTriggered = false;
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::CreateFromChunkStream( Asura_Guid uGuid, Asura_Chunk_Stream &xStream )
{
	Asura_ClientEntity_FMVController* pxNew = new Asura_ClientEntity_FMVController(uGuid);
	if (pxNew)
	{
		pxNew->ReadFromChunkStream(xStream);
	}
}

////////////////////////////////////////////////////////////////////////////////
// ReadFromChunkStream
// Must cope with all versions up to CURRENT_CHUNK_VERSION (be backwards compatible)
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	// Version.
	int iVersion;
	xStream >> iVersion;

	// Parent data.
	PARENT::ReadFromChunkStream( xStream );    

	//
	// Read our data...
	//
	Asura_Assert((iVersion>=0) && (iVersion <= CURRENT_CHUNK_VERSION), "Asura_ClientEntity_FMVController: Invalid chunk version");

	if ( iVersion >= 0 )
	{
		// This purposely doesn't use Asura_ReadFromStream_Bitfield_Wrapper for m_uFlags
		xStream >> m_uFlags;
		xStream >> m_xFMVName;

		if ( iVersion >= 1 )
		{
			//
			// Read in linked speaker Guid list
			//
			unsigned int uTotalGuids;
			xStream >> uTotalGuids;

			for(unsigned int uGuidCount = 0; uGuidCount < uTotalGuids; uGuidCount++)
			{
				unsigned int uThisGuid;
				xStream >> uThisGuid;
				m_xLinkedSpeakerGuids.Push_Back(uThisGuid);
			}
		}
		m_pxVideoRenderer->SetPlayFlags((Asura_VideoToTexture::ASURA_FMV_FLAGS) m_uFlags);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Called after ALL entities are created
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::PostCreateInitialise()
{
	for ( Asura_LinkList_ConstIt< Asura_Guid > xIt( m_xLinkedSpeakerGuids ); !xIt.Done(); xIt.Next() )
	{
		unsigned int uGuid = *xIt.GetCurrent();

		// Find the Speaker
		Asura_ClientEntity_FMVSpeaker* pxSpeaker = Asura_ClientEntity_FMVSpeaker::FindEntity(uGuid);
		if(pxSpeaker)
		{
			Asura_VideoToTexture::SpeakerProperty xSpeakerProperty;

			xSpeakerProperty.m_fInnerRange = pxSpeaker->m_fInnerRange;
			xSpeakerProperty.m_fOuterRange = pxSpeaker->m_fOuterRange;
			xSpeakerProperty.m_xPosition = pxSpeaker->m_xPosition;
			m_pxVideoRenderer->CreateSpeaker(xSpeakerProperty);
		}
		else
		{
			Asura_Assertf(false, ("FMV Speaker doesn't exist (yet) [%d]", uGuid));
		}
	}

	PARENT::PostCreateInitialise();
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_ClientEntity_FMVController::HandleMessage(const Asura_Message* pxMessage)
{
	if( IsDestroyed() ) { return true; } // ignore all further messages?

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			case ASURA_MSG_CLIENTCREATE:
			{
				ProcessCreateMessage( static_cast<const Asura_Message_FMVController_ClientCreate*>(pxMessage->GetData()), pxMessage->GetSentTime() );
				return true;
			}

			case ASURA_MSG_FMV_CONTROLLER_PLAY:
			case ASURA_MSG_FMV_CONTROLLER_PAUSE:
			{
				m_bTriggered = true;
				return true;
			}

			default:
			{
				return PARENT::HandleMessage(pxMessage);
			}
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::ProcessCreateMessage (const Asura_Message_FMVController_ClientCreate* pxData, float fTime)
{
	// Init this client
	m_xFMVName = pxData->m_szFMVName;
	m_uFlags = pxData->m_uFlags;
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::OnDestruction()
{
	PARENT::OnDestruction();
}

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_FMVController::Update()
{
	if( IsDestroyed() ) { return; }

	if (m_pxVideoRenderer && m_bTriggered)
	{
		switch (m_pxVideoRenderer->GetPlayerState())
		{
			case Asura_VideoToTexture::PLAYER_STATE_INITIALISED:
				{
					// this is our placeholder texture name to render the FMV into
					int iTex = Asura_Texture_Management::GetTextureIndex("/FMV_Output.bmp");
					Asura_Assert(iTex != ASURA_TEXTURE_NONE, "/FMV_Output.bmp doesn't exist in texture library");

					m_pxVideoRenderer->Open(m_xFMVName.GetString(), iTex);
					m_pxVideoRenderer->Play();
				}
				break;

			case Asura_VideoToTexture::PLAYER_STATE_COMPLETE:
				{
					Destroy();
				}
				break;

			default:
				break;
		}
	}

	PARENT::Update();
}

#endif // ASURA_USE_VIDEO_TO_TEXTURE_SYSTEM
