///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_ServerEntity_ConsoleVar.cpp
//
// Created by Martin G.Harris (20/6/2007)
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Entity allowing designers to add a run-time console variable in tools and modify
// the value using static messages. This allows easy hooking up of game logic to
// (for example) the HUD system.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Need command console for the console variables.
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#include "Entity/Intangible/Asura_ServerEntity_ConsoleVar.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Triggers/Asura_ServerEntity_CountedTrigger.h"
#include "Entity/Triggers/Asura_ServerEntity_TimeTrigger.h"
#include "Network/Asura_Message.h"


// Static data
Asura_Hashed_List<Asura_ServerEntity_ConsoleVar> Asura_ServerEntity_ConsoleVar::s_xList(4);


///////////////////////////////////
// Asura_ServerEntity_ConsoleVar //
///////////////////////////////////
Asura_ServerEntity_ConsoleVar::Asura_ServerEntity_ConsoleVar(Asura_Guid uGuid)
: Asura_Entity_Server(uGuid)
, m_eType(Asura_ClientEntity_ConsoleVar::TYPE_INVALID)
, m_uEntityToTrackGuid(ASURA_GUID_UNREGISTERED)
, m_uFlags(0)
, m_fTimeOfLastChangeMessage(0.0f)
{
	Classification = AsuraEntityClass_ConsoleVar;
	
	s_xList.AddToList(this);

	SetNeverUpdateClient();
}

Asura_ServerEntity_ConsoleVar::~Asura_ServerEntity_ConsoleVar()
{
	s_xList.RemoveFromList(this);
}

////////////////////////
// WriteToChunkStream //
////////////////////////
void Asura_ServerEntity_ConsoleVar::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	// Version
    xStream << static_cast<int>( CURRENT_CHUNK_VERSION );

	PARENT::WriteToChunkStream(xStream);

	// Our data.
	int iType = m_eType;
	xStream << iType;

	// Peek the client for the string.
	const Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());
	Asura_Assert(pxClient, "Couldn't find client entity");
	if (pxClient && pxClient->GetVarName())
	{
		const Asura_Char* szVarName = pxClient->GetVarName(); 
		u_int uLength = strlen(szVarName) + 1;
		xStream << uLength;

		for (u_int u=0; u<uLength; ++u)
		{
			xStream << szVarName[u];
		}

		// May as well align.
		xStream.Write4BytePadding();
	}
	else
	{
		// No string!
		u_int uLength = 0;
		xStream << uLength;
	}

	xStream << m_uEntityToTrackGuid;
	xStream << m_uFlags;

}

/////////////////////////
// ReadFromChunkStream //
/////////////////////////
// NOTE: If you change this you *must* update the corresponding function in Asura_ClientEntity_ConsoleVar.
void Asura_ServerEntity_ConsoleVar::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	// Version.
    int iVersion;
    xStream >> iVersion;

    // Parent data.
	PARENT::ReadFromChunkStream( xStream );    

    // Our data.
	Asura_Assert(iVersion>=0 && iVersion<=CURRENT_CHUNK_VERSION, "Asura_ServerEntity_ConsoleVar: Invalid chunk version");
    
	if ( iVersion >= 0 )
    {
		// Type
		int iType = -1;
		xStream >> iType;
		m_eType = static_cast<Asura_ClientEntity_ConsoleVar::TYPE>(iType);
		
		// Console Variable name
		u_int uLength = 0; // Length excluding the null terminator (number of charcters written).
		xStream >> uLength;

		// Any string longer than this is probably an error in reading the chunks.
		Asura_Assert(uLength<256, "Asura_ClientEntity_ConsoleVar: Error: String too long!");

		if (uLength>0)
		{
			// Skip the name.
			xStream.Skip(uLength*sizeof(Asura_Char));

			// May as well align.
			xStream.Skip4BytePadding();
		}
	}

	if (iVersion >= 1)
	{
		xStream >> m_uEntityToTrackGuid;
		xStream >> m_uFlags;
	}

}

///////////////////
// HandleMessage //
///////////////////
bool Asura_ServerEntity_ConsoleVar::HandleMessage( const Asura_Message* pxMessage )
{
	if (pxMessage && pxMessage->GetMsgID() == ASURA_MSG_CHANGE_CONSOLEVAR)
	{
		// Forward to clients.
		Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, GetGuid(), true);
		m_fTimeOfLastChangeMessage = Asura_Timers::GetServerAbsoluteTime();
		return true;
	}
	
	return PARENT::HandleMessage(pxMessage);
}

////////////
// Update //
////////////
void Asura_ServerEntity_ConsoleVar::Update()
{
	PARENT::Update();

	if (m_uEntityToTrackGuid != ASURA_GUID_UNREGISTERED)
	{
		UpdateFromTrackedEntity();
	}
}

/////////////////////////////
// UpdateFromTrackedEntity //
/////////////////////////////
// If we have been given an entity to track then continuously monitor its value
// and copy the data to the console variable whenever the value changes.
void Asura_ServerEntity_ConsoleVar::UpdateFromTrackedEntity()
{
	const Asura_Entity_Server* pxServer = Asura_Entity_Server::FindEntity(m_uEntityToTrackGuid);
	Asura_Assert(pxServer, "Couldn't find entity to track");
	if (!pxServer)
	{
		return;
	}

	const u_int uClass = pxServer->GetClassification();
	switch (uClass)
	{
	default:
		Asura_Assert(0, "Asura_ServerEntity_ConsoleVar: Unexpected classification for entity to track!");
		break;

	case AsuraEntityClass_TimeTrigger:
		{
			const Asura_ServerEntity_TimeTrigger* pxTime = static_cast<const Asura_ServerEntity_TimeTrigger*>(pxServer);

			if (IsTimeCountUp())
			{
				// Count up (time elapsed)
				const float fElapsedTime = pxTime->GetElapsedTime();
				SetData(fElapsedTime, true);
			}
			else
			{
				// Count down (time remaining).
				const float fTimeRemaining = pxTime->GetRemainingTime();
				SetData(fTimeRemaining, true);
			}
			break;
		}

	case AsuraEntityClass_CountedTrigger:
		{
			const Asura_ServerEntity_CountedTrigger* pxCounted = static_cast<const Asura_ServerEntity_CountedTrigger*>(pxServer);
			
			const int iCount = pxCounted->GetCount();
			SetData(iCount);
			break;
		}
	}
}

////////////
// GetVar //
////////////
// Helper function to look up variable.
const Asura_CommandConsole_VarRepository::Var* Asura_ServerEntity_ConsoleVar::GetVar() const
{
	const Asura_ClientEntity_ConsoleVar* pxClient = Asura_ClientEntity_ConsoleVar::FindEntity(GetGuid());
	Asura_Assert(pxClient, "Asura_ServerEntity_ConsoleVar: Couldn't find client entity");
	
	return pxClient ? pxClient->GetVar() : 0;
}

/////////////
// SetData //
/////////////
void Asura_ServerEntity_ConsoleVar::SetData(int iData)
{
	const Asura_CommandConsole_VarRepository::Var* pxVar = GetVar();
	if (!pxVar) return;

	const Asura_CommandConsole_VarRepository::Var::Data& xData = pxVar->m_xData;
	
	bool bChanged = false;

	switch (pxVar->m_eType)
	{
	case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL:
		if (xData.pbData && *xData.pbData != (iData!=0))
		{
			bChanged = true;
		}
		break;

	case Asura_CommandConsole_VarRepository::Var::TYPE_INT:
		if (xData.piData && *xData.piData != iData)
		{
			bChanged = true;
		}
		break;

	case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT:
		if (xData.pfData && *xData.pfData != static_cast<float>(iData))
		{
			bChanged = true;
		}
		break;

	default:
		Asura_Assert(0, "Unexpected console variable type.");
		break;
	}

	if (bChanged)
	{
		// Value has changed update all the clients.
		// MGH 26/6/2007: I'm assuming that integer values don't change very often,
		// so we should probably guarantee the messages.
		SendChangeMessage(static_cast<float>(iData), true);
	}
}


/////////////
// SetData //
/////////////
void Asura_ServerEntity_ConsoleVar::SetData(float fData, bool bChangesEveryFrame)
{
	const Asura_CommandConsole_VarRepository::Var* pxVar = GetVar();
	if (!pxVar) return;

	const Asura_CommandConsole_VarRepository::Var::Data& xData = pxVar->m_xData;
	
	bool bChanged = false;

	switch (pxVar->m_eType)
	{
	case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL:
		if (xData.pbData && *xData.pbData != (fData!=0.0f))
		{
			bChanged = true;
		}
		break;

	case Asura_CommandConsole_VarRepository::Var::TYPE_INT:
		if (xData.piData && *xData.piData != Asura_Maths::FloatToInt_Near(fData))
		{
			bChanged = true;
		}
		break;

	case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT:
		if (xData.pfData && *xData.pfData != fData)
		{
			bChanged = true;
		}
		break;

	default:
		Asura_Assert(0, "Unexpected console variable type.");
		break;
	}

	if (bChanged)
	{
		// Value has changed update all the clients.
		// MGH 26/6/2007: Don't guarantee these messages, if the data changes frequently.
		const bool bGuarantee = !bChangesEveryFrame;
		SendChangeMessage(fData, bGuarantee);
	}
}

///////////////////////
// SendChangeMessage //
///////////////////////
DEBUGVAR float dbg_fCONSOLE_VAR_UPDATE_TIME = 0.1f;

void Asura_ServerEntity_ConsoleVar::SendChangeMessage(float fData, bool bGuarantee)
{
	if (!bGuarantee && Asura_Network::IsExternalNetworkActive())
	{
		const float fDelta = (Asura_Timers::GetServerAbsoluteTime() - m_fTimeOfLastChangeMessage);
		if (fDelta >= 0.0f && fDelta < dbg_fCONSOLE_VAR_UPDATE_TIME)
		{
			// Wait before sending another update. This is to reduce the frequency of messages sent
			// from time triggers in MP games.
			return;
		}
	}

	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_CHANGE_CONSOLEVAR, sizeof(Asura_Message_Change_ConsoleVar));
	if (pxMessage)
	{
		Asura_Message_Change_ConsoleVar* pxData = static_cast<Asura_Message_Change_ConsoleVar*>(pxMessage->GetData());
		if (pxData)
		{
			pxData->m_iFlags = Asura_Message_Change_ConsoleVar::FLAG_OPERATION_SET;
			pxData->m_fData  = fData;
		}

		// Update all the clients.
		if (bGuarantee)
		{
			pxMessage->UseGuaranteedDelivery();
		}

		Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
		m_fTimeOfLastChangeMessage = Asura_Timers::GetServerAbsoluteTime();
	}
}

/////////////////////////////////////
// ApplyChangesForTemplateInstance //
/////////////////////////////////////
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"

void Asura_ServerEntity_ConsoleVar::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	m_uEntityToTrackGuid = pxInstance->ConvertGuid(m_uEntityToTrackGuid);

	PARENT::ApplyChangesForTemplateInstance(pxInstance);
}

bool Asura_ServerEntity_ConsoleVar::CanBeIncludedInTemplate() const
{
	return ( Classification == AsuraEntityClass_ConsoleVar );
}
#endif // #ifdef ASURA_TOOLS

#endif //ASURA_SUPPORT_COMMAND_CONSOLE
