//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_Node.cpp
//
// Created by NJD - 11 October 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Entity/Locations/Asura_ServerEntity_Node.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Camera/Asura_Camera.h"
#endif
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Locations/Spline/Asura_ServerNode_Spline.h"
#include "Network/Asura_Message.h"

#if defined (ASURA_USE_STREAMING) && !defined (ASURA_TOOLS)
#include "Streaming/Asura_Streaming_System.h"
#endif //#if defined (ASURA_USE_STREAMING) && !defined (ASURA_TOOLS)

// Tools
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_Node
//
//////////////////////////////////////////////////////////////////////////////

Asura_Hashed_List< Asura_ServerEntity_Node > Asura_ServerEntity_Node::s_xNodeList;

Asura_ServerEntity_Node::Asura_ServerEntity_Node(Asura_Guid uGuid)
: PARENT( uGuid )
, m_uNodeFlags( 0 )
, m_xPosition( Asura_Maths::ZeroVector3 )
, m_puSplines( NULL )
#ifdef ASURA_TOOLS
, m_uToolsObjectColour( 0xffffffff )
#endif
#ifndef ASURA_NO_DEBUG_OUTPUT
, m_xDebugTextPos( Asura_Maths::ZeroVector3 )
, m_bDebugTextPosOnScreen( false )
#endif
{
    Classification = AsuraEntityClass_Node;
    m_bObsoleteIsActive = true;
    m_bIsValid = true;

    SetNeverUpdateClient();
    SetNeverUpdates( true );

    s_xNodeList.AddToList( this );
}

Asura_ServerEntity_Node::~Asura_ServerEntity_Node()
{
    if ( m_puSplines )
    {
        Asura_ResourcePushType( ASURA_RESOURCE_GAMESCENE );
        Asura_ResourceDeallocateArray( m_puSplines );
        Asura_ResourcePopType();
        m_puSplines = NULL;
    }

    s_xNodeList.RemoveFromList( this );
}

u_int Asura_ServerEntity_Node::GetNumIncomingLinks() const
{
	// There's no list of splines unfortunately, so we have to take the longer route.
	u_int uNumIncomingLinks = 0;
	for ( Asura_Hashed_List_ConstIt< Asura_ServerEntity_Node > xNodeIter( &s_xNodeList ); !xNodeIter.Done(); xNodeIter.Next() )
	{
		const Asura_ServerEntity_Node* pxNode = xNodeIter.GetCurrent();
		if ( pxNode->IsBaseOrDerivativeOf( AsuraEntityClass_Spline ) )
		{
			const Asura_ServerNode_Spline* pxSpline = static_cast< const Asura_ServerNode_Spline* >( pxNode );
			if ( pxSpline->GetTargetGuid() == GetGuid() )
			{
				++uNumIncomingLinks;
			}
		}
	}
	return uNumIncomingLinks;
}

bool Asura_ServerEntity_Node::IsBaseOrDerivativeOf( u_short usClassification ) const
{
    // Buck stops here with this method.
    return ( usClassification == AsuraEntityClass_Node );
}

#ifndef ASURA_NO_DEBUG_OUTPUT
void Asura_ServerEntity_Node::RenderDebugInfo()
{
	m_xDebugTextPos = GetPosition();
	m_bDebugTextPosOnScreen = Asura_Camera::ProjectWorldCoordsToScreenSpace( m_xDebugTextPos, true );
}
#endif

#ifdef ASURA_TOOLS
void Asura_ServerEntity_Node::ApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance )
{
	PARENT::ApplyChangesForTemplateInstance(pxInstance );
	
	m_xPosition = pxInstance->TransformPosition(m_xPosition);
    if (m_puSplines)
    {
        for (u_int i=0;i<m_uNumberOfSplines;i++)
        {
            m_puSplines[i] = pxInstance->ConvertGuid(m_puSplines[i]);
        }
    }
}

bool Asura_ServerEntity_Node::CanBeIncludedInTemplate() const 
{
    return ( Classification == AsuraEntityClass_Node );
}

void Asura_ServerEntity_Node::AddNewSpline( const Asura_Guid uGuid )
{
    if ( m_uNumberOfSplines + 1 >= 255 )
    {
        Asura_Assert( false, "Asura_ServerEntity_Node::AddNewSpline - run out of spline space" );
        return;
    }

    // Check we don't already have this spline.
    for (u_int u=0;u<m_uNumberOfSplines;u++)
    {
        Asura_Assert(m_puSplines[u] != uGuid, "Asura_ServerEntity_Node::AddNewSpline - already have this spline");
        if (m_puSplines[u] == uGuid)
        {
            return;
        }
    }

    Asura_ResourcePushType(ASURA_RESOURCE_GAMESCENE);
    {
        Asura_Guid *puBuffer;
        Asura_ResourceAllocateArray(puBuffer, m_uNumberOfSplines + 1);
        for (u_int u=0;u<m_uNumberOfSplines;u++)
        {
            puBuffer[u] = m_puSplines[u];
        }
        puBuffer[m_uNumberOfSplines] = uGuid;
        m_uNumberOfSplines++;
        Asura_ResourceDeallocateArray(m_puSplines);
        m_puSplines = puBuffer;
    }
    Asura_ResourcePopType();
}

void Asura_ServerEntity_Node::RemoveSpline( const Asura_Guid uGuid )
{
    Asura_Assert(m_uNumberOfSplines, "Asura_ServerEntity_Node::RemoveSpline - could not find spline to remove");
    if (!m_uNumberOfSplines)
    {
        return;
    }

    Asura_ResourcePushType(ASURA_RESOURCE_GAMESCENE);
    {
        u_int uPosition = UINT_MAX;
        for (u_int u=0;u<m_uNumberOfSplines;u++)
        {
            if (m_puSplines[u] == uGuid)
            {
                uPosition = u;
            }
        }
		if (uPosition == UINT_MAX)
		{
			Asura_Assert(false, "Asura_ServerEntity_Node::RemoveSpline - could not find spline to remove");
			return;
		}

		// Copy the old splines across, or just remove the array entirely.
        if (m_uNumberOfSplines > 1)
        {
            Asura_Guid *puBuffer;
            Asura_ResourceAllocateArray(puBuffer, m_uNumberOfSplines - 1);
            for (u_int u=0;u<(m_uNumberOfSplines - 1);u++)
            {
                if (u<uPosition)
                {
                    puBuffer[u] = m_puSplines[u];
                }
                else
                {
                    puBuffer[u] = m_puSplines[u+1];
                }
            }
            Asura_ResourceDeallocateArray(m_puSplines);
            m_puSplines = puBuffer;
            m_uNumberOfSplines--;
        }
        else
        {
            Asura_ResourceDeallocateArray(m_puSplines);
            m_puSplines = NULL;
            m_uNumberOfSplines = 0;
        }
    }
    Asura_ResourcePopType();
}
#endif // #ifdef ASURA_TOOLS


void Asura_ServerEntity_Node::CreateFromChunkStream( Asura_Guid uGuid, Asura_Chunk_Stream& xStream )
{
    Asura_ServerEntity_Node* pxNode = new Asura_ServerEntity_Node( uGuid );
    if ( pxNode )
    {
        pxNode->ReadFromChunkStream( xStream );
    }
}

void Asura_ServerEntity_Node::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
    // (TI - 5.9.2004) - DON'T FORGET TO UPDATE ApplyChangesForTemplateInstance IF YOU CHANGE STUFF.

    // Version
    xStream << static_cast< int >( CURRENT_CHUNK_VERSION );

    // Parent.
    PARENT::WriteToChunkStream( xStream );

    // Our data.
    Asura_WriteToStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uNodeFlags );
    xFlagsBitfield.Add( m_bObsoleteIsActive, 1 );
    xFlagsBitfield.Add( m_bIsOccupied,       1 );
    xFlagsBitfield.Add( m_usSplittingPlane,  2 );
    xFlagsBitfield.Add( m_uNumberOfSplines,  8 );
    xFlagsBitfield.Add( m_bIsSmoothOverNode, 1 );
    xFlagsBitfield.Add( m_bIsValid,          1 );
	xFlagsBitfield.Add( m_uTraverseHelper,   2 );
    xStream << xFlagsBitfield;

    xStream << m_xPosition;

	Asura_WriteToStream_Array_Wrapper< Asura_Guid > xGuidArrayWrapper( m_uNumberOfSplines, m_puSplines );
	xStream << xGuidArrayWrapper;

#ifdef ASURA_TOOLS
	xStream << m_uToolsObjectColour;
#else
	xStream << 0;
#endif

    // (TI - 5.9.2004) - DON'T FORGET TO UPDATE ApplyChangesForTemplateInstance IF YOU CHANGE STUFF.
}

void Asura_ServerEntity_Node::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
    Asura_ResourcePushType( ASURA_RESOURCE_GAMESCENE );

    // Version.
    int iVersion;
    xStream >> iVersion;

    // Parent data.
    if ( iVersion > 0 )
    {
        PARENT::ReadFromChunkStream( xStream );
    }

    // Our data.
    if ( iVersion == CURRENT_CHUNK_VERSION )
	{
#ifdef ASURA_ASSERT
		u_int uNumSplines = m_uNumberOfSplines;
#endif
		Asura_Assert( ( m_uNumberOfSplines == uNumSplines ) || ( uNumSplines == 0 ), "Number of splines has changed since last load." );
		Asura_ReadFromStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uNodeFlags );
		xStream >> xFlagsBitfield;
		xStream >> m_xPosition;
		if ( m_uNumberOfSplines )
		{
			Asura_ResourceAllocateArray( m_puSplines, m_uNumberOfSplines );
			Asura_ReadFromStream_Array_Wrapper< Asura_Guid > xGuidArrayWrapper( m_uNumberOfSplines, m_puSplines );
			xStream >> xGuidArrayWrapper;
		}

#ifdef ASURA_TOOLS
		xStream >> m_uToolsObjectColour;
#else
		u_int uDummy;
		xStream >> uDummy;
#endif
	}
	else if ( iVersion == 2 )
    {
#ifdef ASURA_ASSERT
        u_int uNumSplines = m_uNumberOfSplines;
#endif
        Asura_Assert( ( m_uNumberOfSplines == uNumSplines ) || ( uNumSplines == 0 ), "Number of splines has changed since last load." );
        Asura_ReadFromStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uNodeFlags );
        xStream >> xFlagsBitfield;
        xStream >> m_xPosition;
        if ( m_uNumberOfSplines )
        {
            Asura_ResourceAllocateArray( m_puSplines, m_uNumberOfSplines );
			Asura_ReadFromStream_Array_Wrapper< Asura_Guid > xGuidArrayWrapper( m_uNumberOfSplines, m_puSplines );
			xStream >> xGuidArrayWrapper;
        }
    }
	else if ( iVersion == 1 )
    {
#ifdef ASURA_ASSERT
        u_int uNumSplines = m_uNumberOfSplines;
#endif
        Asura_Assert( ( m_uNumberOfSplines == uNumSplines ) || ( uNumSplines == 0 ), "Number of splines has changed since last load." );
        Asura_ReadFromStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uNodeFlags );
        xStream >> xFlagsBitfield;
        xStream >> m_xPosition;
        if ( m_uNumberOfSplines )
        {
            Asura_ResourceAllocateArray( m_puSplines, m_uNumberOfSplines );
			Asura_ReadFromStream_Array_Wrapper< Asura_Guid > xGuidArrayWrapper( m_uNumberOfSplines, m_puSplines );
			xStream >> xGuidArrayWrapper;
        }

		SetEntityActiveFromLoad( m_bObsoleteIsActive );
    }
    else if ( iVersion == 0 )
    {
#ifdef ASURA_ASSERT
        u_int uNumSplines = m_uNumberOfSplines;
#endif
        Asura_Assert( ( m_uNumberOfSplines == uNumSplines ) || ( uNumSplines == 0 ), "Number of splines has changed since last load." );
        Asura_ReadFromStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uNodeFlags );
        xStream >> xFlagsBitfield;
        xStream >> m_xPosition;
        if ( m_uNumberOfSplines )
        {
            Asura_ResourceAllocateArray( m_puSplines, m_uNumberOfSplines );
			Asura_ReadFromStream_Array_Wrapper< Asura_Guid > xGuidArrayWrapper( m_uNumberOfSplines, m_puSplines );
			xStream >> xGuidArrayWrapper;
        }

#ifdef ASURA_TOOLS
        SetIsValid( true );
#endif

		SetEntityActiveFromLoad( m_bObsoleteIsActive );
    }
    else
    {
        Asura_Assertf( false, ( "Asura_ServerEntity_Node::InitialiseFromChunkStream - unhandled chunk version (%d)", iVersion ) );
    }

#if defined (ASURA_USE_STREAMING) && !defined (ASURA_TOOLS)
	for (u_int u=0; u<m_uNumberOfSplines; ++u)
	{
		m_puSplines[u] = Asura_Streaming_System::MakeCurrentMasterZoneGuid(m_puSplines[u]);
	}
#endif //#if defined (ASURA_USE_STREAMING) && !defined (ASURA_TOOLS)

    Asura_ResourcePopType();
}

void Asura_ServerEntity_Node::TraverseStructure( const Asura_Guid uStartGuid, void(*pfnFunction)(Asura_ServerEntity_Node*,void*), void* pUserData )
{
	static const u_int s_uUntouchedID = 0;
	static const u_int s_uTouchedID   = 1;
	static const u_int s_uClosedID    = 2;

	// Mark all nodes as untouched.
	for ( Asura_Hashed_List_It< Asura_ServerEntity_Node > xNodeIter( GetList() ); !xNodeIter.Done(); xNodeIter.Next() )
	{
		xNodeIter.GetCurrent()->m_uTraverseHelper = s_uUntouchedID;
	}

	// Traverse the structure, calling the callback for all appropriate objects.
	// This node could actually be a spline. If it is we start from the spline's source node.
	Asura_LinkList< Asura_Guid > xPendingQueue;
	Asura_ServerEntity_Node* pxStartNode = FindEntity( uStartGuid );
	if ( !pxStartNode )
	{
		return;
	}
	else if ( pxStartNode->IsBaseOrDerivativeOf( AsuraEntityClass_Spline ) )
	{
		Asura_ServerNode_Spline* pxStartSpline = static_cast< Asura_ServerNode_Spline* >( pxStartNode );
		xPendingQueue.Push_Back( pxStartSpline->GetSourceGuid() );
	}
	else
	{
		xPendingQueue.Push_Back( pxStartNode->GetGuid() );
	}
	while ( xPendingQueue.Front() != NULL )
	{
		Asura_ListElt< Asura_Guid >* pxElt = xPendingQueue.Front();
		const Asura_Guid uBaseNodeGuid = pxElt->Data();
		xPendingQueue.Remove( pxElt );

		// Handle this node.
		Asura_ServerEntity_Node* pxBaseNode = FindEntity( uBaseNodeGuid );
		if ( !pxBaseNode ) { continue; }
		pfnFunction( pxBaseNode, pUserData );

		// Traverse all of the links out of this node.
		for ( u_int uLink = 0; uLink < pxBaseNode->GetNumberLinks(); ++uLink )
		{
			Asura_ServerEntity_Node* pxSplineNode = FindEntity( pxBaseNode->GetLink( uLink ) );
			if ( !pxSplineNode || !pxSplineNode->IsBaseOrDerivativeOf( AsuraEntityClass_Spline ) ) { continue; }

			// If the link has already been touched/closed then we can be sure the relevant node has already been touched.
			if ( pxSplineNode->m_uTraverseHelper != s_uUntouchedID ) { continue; }

			// Handle this link.
			pxSplineNode->m_uTraverseHelper = s_uClosedID;
			pfnFunction( pxSplineNode, pUserData );

			// If the target node has not already been touched then it needs expanding.
			Asura_ServerNode_Spline* pxSpline = static_cast< Asura_ServerNode_Spline* >( pxSplineNode );
			Asura_ServerEntity_Node* pxTargetNode = FindEntity( pxSpline->GetTargetGuid() );
			if ( !pxTargetNode || pxTargetNode->m_uTraverseHelper != s_uUntouchedID ) { continue; }
			pxTargetNode->m_uTraverseHelper = s_uTouchedID;
			xPendingQueue.Push_Back( pxTargetNode->GetGuid() );
		}

		// Also traverse all of the links into this node.
		for ( Asura_Hashed_List_It< Asura_ServerEntity_Node > xNodeIter( GetList() ); !xNodeIter.Done(); xNodeIter.Next() )
		{
			Asura_ServerEntity_Node* pxSplineNode = xNodeIter.GetCurrent();
			if ( !pxSplineNode || !pxSplineNode->IsBaseOrDerivativeOf( AsuraEntityClass_Spline ) ) { continue; }
			Asura_ServerNode_Spline* pxSpline = static_cast< Asura_ServerNode_Spline* >( pxSplineNode );
			if ( pxSpline->GetTargetGuid() != pxBaseNode->GetGuid() ) { continue; }

			// If the link has already been touched/closed then we can be sure the relevant node has already been touched.
			if ( pxSplineNode->m_uTraverseHelper != s_uUntouchedID ) { continue; }

			// Handle this link.
			pxSplineNode->m_uTraverseHelper = s_uClosedID;
			pfnFunction( pxSplineNode, pUserData );

			// If the source node has not already been touched then it needs expanding.
			Asura_ServerEntity_Node* pxSourceNode = FindEntity( pxSpline->GetSourceGuid() );
			if ( !pxSourceNode || pxSourceNode->m_uTraverseHelper != s_uUntouchedID ) { continue; }
			pxSourceNode->m_uTraverseHelper = s_uTouchedID;
			xPendingQueue.Push_Back( pxSourceNode->GetGuid() );
		}

		// We've finished expanding our source node.
		pxBaseNode->m_uTraverseHelper = s_uClosedID;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
