//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerNode_LiftSpline.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     spline class for lift entities to follow
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Locations/Spline/LiftSpline/Asura_ServerNode_LiftSpline.h"

// asura
#include "Core/Asura_Graph.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Cutscene/Asura_Cutscene_Track.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Locations/OrientedNode/LiftNode/Asura_ServerNode_LiftNode.h"
#include "Network/Asura_Message.h"
#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING)

//Tools
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerNode_LiftSpline
//
//////////////////////////////////////////////////////////////////////////////

Asura_Hashed_List< Asura_ServerNode_LiftSpline > Asura_ServerNode_LiftSpline::s_xLiftSplineList;

Asura_ServerNode_LiftSpline::Asura_ServerNode_LiftSpline(Asura_Guid uGuid)
: PARENT( uGuid )
, m_uLiftSplineFlags( 0 )
, m_pxSpeedOverTimeGraph(NULL)
, m_fTotalTimeRequired( 0.f )
, m_fDurationOverride(0.0f)
{
    Classification = AsuraEntityClass_LiftSpline;
	m_pxSpeedOverTimeGraph = new Asura_Graph<float>(16, 1.75f, 1.75f);
	m_pxSpeedOverTimeGraph->SetClampingValues(true, false, 0.0f, Asura_Maths::LargeFloat);

	s_xLiftSplineList.AddToList( this );
}

Asura_ServerNode_LiftSpline::~Asura_ServerNode_LiftSpline()
{
	delete m_pxSpeedOverTimeGraph;

	s_xLiftSplineList.RemoveFromList( this );
}

void Asura_ServerNode_LiftSpline::CreateFromChunkStream( Asura_Guid uGuid, Asura_Chunk_Stream& xStream )
{
    Asura_ServerNode_LiftSpline* pxNode = new Asura_ServerNode_LiftSpline( uGuid );
    if ( pxNode )
    {
        pxNode->ReadFromChunkStream( xStream );
    }
}

void Asura_ServerNode_LiftSpline::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	int iVersion;
	xStream >> iVersion;

	// Parent.
	PARENT::ReadFromChunkStream( xStream );

	if ( iVersion <= CURRENT_CHUNK_VERSION )
	{
		// Us.
		Asura_ReadFromStream_Bitfield_Wrapper< u_int > xFlagsBitfield( m_uLiftSplineFlags );
		xStream >> xFlagsBitfield;

		if (iVersion < 2)
		{
			m_bDistanceVsTimeGraph = 0;
		}


		delete m_pxSpeedOverTimeGraph;
		m_pxSpeedOverTimeGraph = new Asura_Graph< float >(xStream);
#ifdef ASURA_TOOLS
		if (m_bDistanceVsTimeGraph)
		{
			m_pxSpeedOverTimeGraph->SetClampingValues( true, true, 0.0f, 1.0f );
		}
		else
		{
			m_pxSpeedOverTimeGraph->SetClampingValues( true, false, 0.0f, Asura_Maths::LargeFloat );
		}
#endif //ASURA_TOOLS
		xStream >> m_fTotalTimeRequired;
		if (iVersion >= 1)
		{
			xStream >> m_fDurationOverride;
		}
		if (iVersion < 1)
		{
			u_int uNumDistanceTimeGraphEntries;
			float fDistance;
			float fCumulativeTime;
			xStream >> uNumDistanceTimeGraphEntries;
			for ( u_int i = 0; i < uNumDistanceTimeGraphEntries; i++ )
			{
				xStream >> fDistance;
				xStream >> fCumulativeTime;
			}
		}
	}
	else
	{
		Asura_Assertf( false, ( "Asura_ServerNode_LiftSpline::InitialiseFromChunkStream - unhandled chunk version (%d)", iVersion ) );
	}
}

void Asura_ServerNode_LiftSpline::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	// (TI - 5.9.2004) - DON'T FORGET TO UPDATE ApplyChangesForTemplateInstance IF YOU CHANGE STUFF.

	// Version
	xStream << static_cast< int >( CURRENT_CHUNK_VERSION );

	// Parent
	PARENT::WriteToChunkStream(xStream);

    bool bSamplesOnly = true;
	bool bIntegrate = false;
#ifdef ASURA_TOOLS
    bSamplesOnly = Tools_Level_Export::IsExport();
	bIntegrate = Tools_Level_Export::IsExport() && !m_bDistanceVsTimeGraph;
	if (bIntegrate)
	{
		m_bIntegratedGraph = true;
	}
#endif //ASURA_TOOLS

	// Our data
	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xFlagsBitfield (m_uLiftSplineFlags);
	xFlagsBitfield.Add(m_bTakeOrientationFromPoints, 1);
	xFlagsBitfield.Add(m_bDeprecated1, 1);
	xFlagsBitfield.Add(m_bOverrideDuration, 1);
	xFlagsBitfield.Add(m_bIntegratedGraph, 1);
	xFlagsBitfield.Add(m_bDistanceVsTimeGraph, 1);
	xStream << xFlagsBitfield;

#ifdef ASURA_TOOLS
	m_bIntegratedGraph = false;
#endif //ASURA_TOOLS

    m_pxSpeedOverTimeGraph->WriteToChunkStream( xStream, bSamplesOnly, bIntegrate );

	xStream << m_fTotalTimeRequired;
	xStream << m_fDurationOverride;
	// (TI - 5.9.2004) - DON'T FORGET TO UPDATE ApplyChangesForTemplateInstance IF YOU CHANGE STUFF.
}

float Asura_ServerNode_LiftSpline::GetAverageLinkSpeed() const
{
	return GetTotalLinkTime() > 0.001f ? GetTotalLength() / GetTotalLinkTime() : Asura_Maths::LargeFloat;
}

float Asura_ServerNode_LiftSpline::GetPositionFromTime( const float fTimeElapsed, Asura_Vector_3& xOutput, Asura_Quat *pxOutputOrientation ) const
{
	const float fProportion = GetProportionFromTime( fTimeElapsed );
	GetPositionFromProportion( fProportion, xOutput, pxOutputOrientation );
	return ( fTimeElapsed > GetTotalLinkTime() ) ? ( fTimeElapsed - GetTotalLinkTime() ) : 0.0f;
}

float Asura_ServerNode_LiftSpline::GetSpeedFromTime(const float fTimeElapsed) const
{
	float fProportion = GetProportionFromTime(fTimeElapsed);
	if (m_bDistanceVsTimeGraph)
	{
		Asura_Assert(0, "Unimplemented!");
		return 1.0f;
	}
	else if (m_bOverrideDuration)
	{
		return m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete(fProportion) * m_fTotalTimeRequired / m_fDurationOverride;
	}
	else
	{
		return m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete(fProportion);
	}
}

const Asura_Quat& Asura_ServerNode_LiftSpline::GetLinkOrientation(const int iIndex) const
{
	Asura_Assert( iIndex != static_cast<int>(m_uNumberOfPoints) - 1 && iIndex != 0, "Asura_ServerNode_LiftSpline::GetLinkOrientation - index out of range" );
	return GetPointOrientation(iIndex - 1);
}

float Asura_ServerNode_LiftSpline::GetProportionFromTime(float fTimeElapsed) const
{
	//sanity check
	if (m_fTotalTimeRequired <= 0.0f)
	{
		return 0.0f;
	}
	else if (m_bDistanceVsTimeGraph)
	{
		return m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete(fTimeElapsed / m_fDurationOverride);
	}
	else
	{
		const float fLength	= GetTotalLength();
		const float	fTime	= (m_bOverrideDuration)
							? (fTimeElapsed / m_fDurationOverride)
							: (fTimeElapsed / m_fTotalTimeRequired);
		const float	fValue	= (m_bIntegratedGraph)
							? m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete(fTime)
							: m_pxSpeedOverTimeGraph->GetAreaUnderGraph(fTime);

		return(fValue * m_fTotalTimeRequired / fLength);
	}
}

bool Asura_ServerNode_LiftSpline::IsBaseOrDerivativeOf(u_short usClassification) const
{
	if (usClassification == AsuraEntityClass_LiftSpline)
	{
		return true;
	}
	return PARENT::IsBaseOrDerivativeOf(usClassification);
}

#ifdef ASURA_TOOLS
void Asura_ServerNode_LiftSpline::OnChanged()
{
	PARENT::OnChanged();
	ChangedDistanceTimeGraph();
}

Asura_Graph<float>* Asura_ServerNode_LiftSpline::GetSpeedOverTimeGraph()
{
	return m_pxSpeedOverTimeGraph;
}

Asura_Quat& Asura_ServerNode_LiftSpline::GetLinkOrientation(const int iIndex)
{
	Asura_Assert( iIndex != static_cast<int>(m_uNumberOfPoints) - 1 && iIndex != 0, "Error : cannot move end control points this way!" );
	return GetPointOrientation(iIndex - 1);
}

void Asura_ServerNode_LiftSpline::SetDistanceVsTimeGraph( const bool bDistanceTimeGraph ) 
{
	if (bDistanceTimeGraph != (bool)m_bDistanceVsTimeGraph)
	{
		if (bDistanceTimeGraph)
		{
			if (!m_bOverrideDuration || m_fDurationOverride < 0.001f)
			{
				m_fDurationOverride = GetTotalLength() / GetAverageLinkSpeed();
			}
			m_pxSpeedOverTimeGraph->Initialise(16, 0.0f, 1.0f, 0.0f, 1.f);
			m_pxSpeedOverTimeGraph->SetClampingValues(true, true, 0.0f, 1.0f);
		}
		else
		{
			m_pxSpeedOverTimeGraph->Initialise(16, GetAverageLinkSpeed(), GetAverageLinkSpeed(), 0.0f, 1.f);
			m_pxSpeedOverTimeGraph->SetClampingValues(true, false, 0.0f, Asura_Maths::LargeFloat);
		}
		m_bDistanceVsTimeGraph = bDistanceTimeGraph;
	}
}

float Asura_ServerNode_LiftSpline::GetTimeFromProportion( float fProportion ) const
{
	const u_int uNumSamples = m_pxSpeedOverTimeGraph->GetNumDiscreteSamples();
	const float fTimeSlice  = ( m_pxSpeedOverTimeGraph->GetDomainMax() - m_pxSpeedOverTimeGraph->GetDomainMin() ) / ( uNumSamples - 1 );

	float fTimeBefore = m_pxSpeedOverTimeGraph->GetDomainMin();
	float fTimeAfter  = fTimeBefore + fTimeSlice;

	if ( m_bDistanceVsTimeGraph )
	{
		for ( u_int u = 0; u < uNumSamples - 1; ++u )
		{
			const float fValueAfter = m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete( fTimeAfter );
			if ( fValueAfter > fProportion )
			{
				const float fValueBefore = m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete( fTimeBefore );
				const float fSubProp	 = ( fProportion - fValueBefore ) / ( fValueAfter - fValueBefore );

				return Asura_Maths::Lerp( fTimeBefore, fTimeAfter, fSubProp ) * GetTotalLinkTime();
			}

			fTimeBefore = fTimeAfter;
			fTimeAfter += fTimeSlice;
		}
	} 
	else if ( m_bIntegratedGraph )
	{
		for ( u_int u = 0; u < uNumSamples - 1; ++u )
		{
			const float fValueAfter = m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete( fTimeAfter ) * GetTotalLinkTime() / GetTotalLength();
			if ( fValueAfter > fProportion )
			{
				const float fValueBefore = m_pxSpeedOverTimeGraph->GetValueAtTime_Discrete( fTimeBefore ) * GetTotalLinkTime() / GetTotalLength();
				const float fSubProp     = ( fProportion - fValueBefore ) / ( fValueAfter - fValueBefore );

				return Asura_Maths::Lerp( fTimeBefore, fTimeAfter, fSubProp ) * GetTotalLinkTime();
			}

			fTimeBefore = fTimeAfter;
			fTimeAfter += fTimeSlice;
		}
	}
	else
	{
		for ( u_int u = 0; u < uNumSamples - 1; ++u )
		{
			const float fValueAfter = m_pxSpeedOverTimeGraph->GetAreaUnderGraph( fTimeAfter ) * GetTotalLinkTime() / GetTotalLength();
			if ( fValueAfter > fProportion )
			{
				float fValueBefore = m_pxSpeedOverTimeGraph->GetAreaUnderGraph( fTimeBefore ) * GetTotalLinkTime() / GetTotalLength();
				float fSubProp = ( fProportion - fValueBefore ) / ( fValueAfter - fValueBefore );

				return Asura_Maths::Lerp( fTimeBefore, fTimeAfter, fSubProp ) * GetTotalLinkTime();
			}

			fTimeBefore = fTimeAfter;
			fTimeAfter += fTimeSlice;
		}
	}

	return GetTotalLinkTime();
}

void Asura_ServerNode_LiftSpline::ChangedDistanceTimeGraph()
{
	if (!m_bDistanceVsTimeGraph)
	{
		// (AW - 09.07.2004) - Some value has changed which means we need to update our distance-time graph.

		// (AW - 09.07.2004) - Figure out our average speed
		float fAverageSpeed = m_pxSpeedOverTimeGraph->GetAverageValue();
		Asura_Assert( fAverageSpeed > 0.01f && fAverageSpeed < 1000.0f, "Unexpected average speed" );
		fAverageSpeed = Asura_Maths::ClampToRange( fAverageSpeed, 0.01f, 1000.0f );

		// (AW - 09.07.2004) - Calculate time required and number of distance-time graph entries to start with.
		m_fTotalTimeRequired = GetTotalLength() / fAverageSpeed;
	}
}

bool Asura_ServerNode_LiftSpline::CanBeIncludedInTemplate() const 
{
    return ( GetClassification() == AsuraEntityClass_LiftSpline );
}

#endif //ASURA_TOOLS
