//////////////////////////////////////////////////////////////////////////////////////////
//
// FILE : Asura_ClientEntity_PhysicalObject.cpp
//
// Inherits from Asura_Entity_Renderable. Abstract base class for physical
// objects in the game that have shape, dynamics block and health.
//
// AUTHORS : Martin G. Harris, Richard M. Smith
//
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////

//////////////
// INCLUDES //
//////////////

#include "Core/Asura.h"

#include "Entity/PhysicalObject/Asura_ClientEntity_PhysicalObject.h"

#include "Animation/Asura_AnimationAttachment.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "Core/Asura_LOS.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "Network/Asura_Message.h"
#include "SpecialFX/PFX/Asura_PFX_Entity_Effect.h"
#include "SpecialFX/Asura_Skybox.h"
#include "Teleport/Asura_Teleport_Messages.h"

#ifdef ASURA_USE_MOVEMENT_SYSTEM
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Entity/PhysicalObject/Moves/Asura_Moves.h"
#endif

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#endif

//////////
// DATA //
//////////

Asura_Hashed_List<Asura_ClientEntity_PhysicalObject> Asura_ClientEntity_PhysicalObject::s_xList( 5 );

//////////////////////////////////////////////////////////////////////////////////////////
// Asura_ClientEntity_PhysicalObject                                                    //
//////////////////////////////////////////////////////////////////////////////////////////
Asura_ClientEntity_PhysicalObject::Asura_ClientEntity_PhysicalObject()
: PARENT()
#ifdef ASURA_USE_MOVEMENT_SYSTEM
, m_pxMovementSystem( NULL )
#endif
, m_fHealth(0.0f)
, m_fLastHealthServerTimestamp(0.0f)
, m_xShapeInstance()
, m_fLastShapeIDServerTimestamp(0.0f)
, m_uParentGuid(ASURA_GUID_UNREGISTERED)
, m_pxParentAttachment(NULL)
, m_fParentLinkServerTime(0.0f)
{
    m_bStaticDoneUpdate = false;
    m_bAnimChanged      = false;
	m_bClientControlsShapeSkinAndAnim = false;
	m_bIgnoreForLOS = false;
	m_bAttachedWithPhysics = false;
	m_bSkyboxObject = false;

	s_xList.AddToList(this);

	AddToLineOfSightList();
}

Asura_ClientEntity_PhysicalObject::~Asura_ClientEntity_PhysicalObject()
{
#ifdef ASURA_USE_MOVEMENT_SYSTEM
	delete(m_pxMovementSystem);
#endif

	if (m_uParentGuid != ASURA_GUID_UNREGISTERED)
	{
		// (MH - 22.01.2007) only unlink if parent hasn't already been destroyed (avoids assert)
		if (Asura_ClientEntity_PhysicalObject::FindEntityIncludingDestroyed(m_uParentGuid))
		{
			UnlinkFromParent(Asura_Maths::LargeFloat);
		}

	}
	if(m_bSkyboxObject)
	{
		Asura_Skybox::RemoveSkyboxObject(this);
	}

	s_xList.RemoveFromList(this);
}

void Asura_ClientEntity_PhysicalObject::InitialiseFromChunkBuffer(const char*& pcChunkBuffer)
{
	Asura_Assert(false, "Reading of ChunkBuffers in the client physical object is no longer supported.");
}

///////////////////
// HandleMessage //
///////////////////
bool Asura_ClientEntity_PhysicalObject::HandleMessage(const Asura_Message* pxMessage)
{	
#ifdef ASURA_USE_MOVEMENT_SYSTEM
	if ( m_pxMovementSystem && m_pxMovementSystem->HandleMessage( pxMessage ) )
	{
		return true;		
	}
#endif

	if (pxMessage)
    {
        switch (pxMessage->GetMsgID())
        {
            case ASURA_MSG_DESTRUCT:
	        {
		        SetAsDestroyed();
		        return true;
	        }

            case ASURA_MSG_CHANGE_ANIMATION:
            {
                const Asura_Message_PhysicalObject_ChangeAnimation* pxData = reinterpret_cast
                    <const Asura_Message_PhysicalObject_ChangeAnimation*>(pxMessage->GetData());
                Asura_Assert(pxMessage->GetDataSize() == sizeof(Asura_Message_PhysicalObject_ChangeAnimation), "MH: serious error. data size has changed");
				
				if(pxData->m_iFlags & Asura_PhysicalObject_AnimationFlag_UseStartTime)
				{
					//playing from a specified position.
					//Determine the end time to use , and whether the loop flag should be left
					//RAR 25/08/2006
					int iFlags = pxData->m_iFlags;
					float fEndTime;
					if(pxData->m_iFlags & Asura_PhysicalObject_AnimationFlag_UseEndTime)
					{
						fEndTime = pxData->m_fEndTime;
						iFlags &=~ AsuraAnimationFlag_Loop;
					}
					else
					{
						fEndTime = (pxData->m_iFlags & AsuraAnimationFlag_Reversed) ? 0.0f : 1.0f;
					}
					
					m_xShapeInstance.ChangeAnimation(pxData->m_uAnimID, iFlags, pxData->m_fTweenDuration, pxData->m_fStartTime, fEndTime);
				}
				else if(pxData->m_iFlags & Asura_PhysicalObject_AnimationFlag_UseEndTime)
				{
					//Playing animation to a specifiedposition 
					//If already playing it , then start from the current timer
					//RAR 25/08/2006
					Asura_Animation* pxAnim = m_xShapeInstance.GetAnimation();
					if(pxAnim && pxAnim->m_xPrimaryWrapper.GetPrimaryAnimHashID() == pxData->m_uAnimID)
					{
						pxAnim->m_xPrimaryWrapper.SetEndTime(pxData->m_fEndTime);
					}
					else
					{
						//determine start timer according to whether it is reversed
						float fStart = (pxData->m_iFlags & AsuraAnimationFlag_Reversed) ? 1.0f : 0.0f;
						m_xShapeInstance.ChangeAnimation(pxData->m_uAnimID, pxData->m_iFlags &~ AsuraAnimationFlag_Loop, pxData->m_fTweenDuration, fStart, pxData->m_fEndTime);

					}
				}
				else
				{
					m_xShapeInstance.ChangeAnimation(pxData->m_uAnimID, pxData->m_iFlags, pxData->m_fTweenDuration);
				}
				
                m_bAnimChanged = true;
                return true;
            }

			case ASURA_MSG_TEXTUREANIMATION :
			{
	            const Asura_Message_TextureAnimation* pxData = reinterpret_cast
	                <const Asura_Message_TextureAnimation*>(pxMessage->GetData());
	            Asura_Assert(pxMessage->GetDataSize() == sizeof(Asura_Message_TextureAnimation), "serious error. data size has changed");

				if(m_xShapeInstance.GetAnimation())
				{
					if(pxData->m_bChangeSequence)
					{
						m_xShapeInstance.SetTextureAnimSequence(pxData->m_uSequence);
					}
					if(pxData->m_bChangeSpeed)
					{
						m_xShapeInstance.SetTextureAnimSpeed(pxData->m_fSpeed);
					}
				}

				return true;

			}

			case ASURA_MSG_HIDE :
				SetInvisibleByPhysicalFlag(true);
				return true;

			case ASURA_MSG_UNHIDE :
				SetInvisibleByPhysicalFlag(false);
				return true;

			case ASURA_MSG_HIDEFORCUTSCENE :
			{
	            const unsigned char* pucHide = reinterpret_cast
	                <const unsigned char*>(pxMessage->GetData());

				SetInvisibleByCutscene((*pucHide) ? true : false);
				return true;
			}

			case ASURA_MSG_LINK_TO_PARENT:
			{
				const Asura_Message_PhysicalObject_LinkToParent*	pxData	= static_cast<const Asura_Message_PhysicalObject_LinkToParent*>(pxMessage->GetData());

				LinkToParent(pxData->GetParentGuid(), pxData->GetMarkerHash(), pxMessage->GetSentTime(), false);
				return(true);
			}

			case ASURA_MSG_LINK_TO_PARENT_WITH_PHYSICS:
			{
				const Asura_Message_PhysicalObject_LinkToParent*	pxData	= static_cast<const Asura_Message_PhysicalObject_LinkToParent*>(pxMessage->GetData());

				LinkToParent(pxData->GetParentGuid(), pxData->GetMarkerHash(), pxMessage->GetSentTime(), true);
				return(true);
			}

			case ASURA_MSG_LINK_TO_PARENT_BONE_RELATIVE:
			{
				const Asura_Message_PhysicalObject_LinkToParentBoneRelative*	pxData	= static_cast<const Asura_Message_PhysicalObject_LinkToParentBoneRelative*>(pxMessage->GetData());

				LinkToParent(pxData->GetParentGuid(), pxData->GetPosition(), pxData->GetOrientation(), pxData->GetBoneID(), pxMessage->GetSentTime());
				return(true);
			}

			case ASURA_MSG_LINK_TO_PARENT_STICKY:
			{
				const Asura_Message_PhysicalObject_LinkToParentSticky*	pxData	= static_cast<const Asura_Message_PhysicalObject_LinkToParentSticky*>(pxMessage->GetData());

				LinkToParent(pxData->GetParentGuid(), pxData->GetMarkerHash(), pxData->GetBoneID(), pxData->GetPosition(), pxMessage->GetSentTime());
				return(true);
			}

			case ASURA_MSG_UNLINK_FROM_PARENT:
			{
				UnlinkFromParent(pxMessage->GetSentTime());
				return(true);
			}

			case ASURA_MSGID_TELEPORT:
			{
				const Asura_Message_Teleport*	pxData	= static_cast<const Asura_Message_Teleport*>(pxMessage->GetData());

				if(pxData != NULL)
				{
					Asura_Vector_3		xPosition;
					Asura_Matrix_3x3	xOrientation;

					pxData->GetPosition(xPosition);
					pxData->GetOrientation(xOrientation);
					TeleportTo(xPosition, xOrientation);
				}
				return(true);
			}

	
			case ASURA_MSG_PLAY_PFX_ON_ENTITY:
			{
				const Asura_Msg_PlayPFXOnEntity* const pxData = static_cast<const Asura_Msg_PlayPFXOnEntity*>(pxMessage->GetData());
				ProcessPlayPFXOnEntity(pxData);
				return true;
			}

            default:
                break;
        }
    }

	return PARENT::HandleMessage(pxMessage);
}

////////////////////////////
// ProcessPlayPFXOnEntity //
////////////////////////////
// MGH 10/12/2007: Code copied and adapted from the Simpsons.
void Asura_ClientEntity_PhysicalObject::ProcessPlayPFXOnEntity(const Asura_Msg_PlayPFXOnEntity* pxData)
{
	if (!pxData) return;

	if (!IsEntityActive())
	{
		return;
	}

	Asura_Animation* const pxAnimation = GetAnimation();
	if (!pxAnimation)
	{
		Asura_Assertf(pxAnimation, ("Null animation, couldn't play PFX on entity %u.", GetGuid()));
		return;
	}

	const Asura_Hash_ID uPFXHash = pxData->m_uPFXEffectHash;
	if (uPFXHash == ASURA_HASH_ID_UNSET)
	{
		Asura_Assertf(0, ("PFX is unset, couldn't play PFX on entity %u.", GetGuid()));
		return;
	}

	const Asura_Hash_ID uMarkerHash = pxData->m_uMarkerHash;
	const u_int uFlags = pxData->m_uPlayPFXFlags;
	const bool bAttach = (0 != (uFlags & uASURA_PLAYPFXONENTITY_FLAG_ATTACH));

	const unsigned short usPFXFlags = ASURA_PFX_EFFECT_CAMERA_FLAGS;

	if (bAttach)
	{
		if (uMarkerHash != ASURA_HASH_ID_UNSET)
		{
			// Attach PFX to marker.
			Asura_AnimationAttachment_PFX* const pxPFX = new Asura_AnimationAttachment_PFX(uPFXHash, usPFXFlags, GetGuid());
			if (pxPFX)
			{
#ifdef ASURA_ASSERT
				const bool bOk = 
#endif // #ifdef ASURA_ASSERT
					pxPFX->LinkAttachment(pxAnimation, uMarkerHash, true);
				Asura_Assertf(bOk, ("Failed to attach PFX (0x%08x) to marker (0x%08x) for entity %u.", uPFXHash, uMarkerHash, GetGuid()));
			}
		}
		else
		{
			// Attach PFX to centre of bounding box.
			Asura_AnimationAttachment_BoneRelativePFX* const pxPFX = new Asura_AnimationAttachment_BoneRelativePFX(uPFXHash, usPFXFlags, GetGuid());
			if (pxPFX)
			{
				// If he's never had his bone data calculated then do it now.
				Asura_Hierarchy_Tween& xTweenStore = *(pxAnimation->GetTweenStore());
				if(!xTweenStore.IsPositionDataValid() && pxAnimation->GetSkin() )
				{
					// Ensure that the bounding-box is up-to-date...
					if(pxAnimation->GetNumTimerUpdates() == 0)
					{
						pxAnimation->UpdateTimers(0.0f, true);
					}
					
					// Ensure that the bones have been updated anyway...
					pxAnimation->UpdateBoneData(0, pxAnimation->GetSkin()->NumberOfBones - 1 );
			
					xTweenStore.SetPositionDataValid(true);
					xTweenStore.SetPositionDataInWorldspace(false);
				}

				const Asura_Vector_3& xWorldSpacePos = pxAnimation->GetBoundingSphereCentre();
				const Asura_Vector_3 xObjectSpacePos = GetCurrentOrientation().ApplyTranspose(xWorldSpacePos - GetCurrentPosition());
			
				const int iBone = 0;

				Asura_Vector_3 xBoneSpacePos;
				Asura_Quat xBoneSpaceOri;
				if (pxAnimation->GetBoneRelativePositionAndOrientation(xObjectSpacePos, Asura_Maths::IdentityQuat, iBone, xBoneSpacePos, xBoneSpaceOri))
				{
#ifdef ASURA_ASSERT
					const bool bOk = 
#endif // #ifdef ASURA_ASSERT
						pxPFX->LinkAttachment_BoneRelative(pxAnimation, xBoneSpacePos, Asura_Maths::IdentityQuat, iBone, true);
					Asura_Assertf(bOk, ("Failed to attach PFX (0x%08x) to bone for entity %u.", uPFXHash, GetGuid()));
				}
				else
				{
					Asura_Assertf(0, ("Failed to attach PFX (0x%08x) to bone for entity %u.", uPFXHash, GetGuid()));
				}
			}
		}
	}
	else if (uMarkerHash != ASURA_HASH_ID_UNSET)
	{
		// Start PFX at marker position, but don't attach.
		Asura_Vector_3 xPos;
		Asura_Matrix_3x3 xOri;
		if (pxAnimation->GetMarkerLocationFromHash(uMarkerHash, &xPos, &xOri, true))
		{
			Asura_PFX_Entity_Effect::Create(xPos, xOri, uPFXHash, 0, usPFXFlags, GetGuid());
		}
		else
		{
			Asura_Assertf(0, ("Failed to start PFX (0x%08x) at marker (0x%08x) for entity %u.", uPFXHash, uMarkerHash, GetGuid()));
		}
	}
	else
	{
		// Start PFX at centre of BB, but don't attach.	
		// If he's never had his bone data calculated then do it now.
		Asura_Hierarchy_Tween& xTweenStore = *(pxAnimation->GetTweenStore());
		if(!xTweenStore.IsPositionDataValid() && pxAnimation->GetSkin() )
		{
			// Ensure that the bounding-box is up-to-date...
			if(pxAnimation->GetNumTimerUpdates() == 0)
			{
				pxAnimation->UpdateTimers(0.0f, true);
			}

			// Ensure that the bones have been updated anyway...
			pxAnimation->UpdateBoneData(0, pxAnimation->GetSkin()->NumberOfBones - 1 );

			xTweenStore.SetPositionDataValid(true);
			xTweenStore.SetPositionDataInWorldspace(false);
		}

		Asura_PFX_Entity_Effect::Create( pxAnimation->GetBoundingSphereCentre(), GetCurrentOrientation(), uPFXHash, 0, usPFXFlags, GetGuid() );
	}
}


// --------------------------------------------------------------------------------------------------------------------------------------------------------------
// void Asura_ClientEntity_PhysicalObject::ProcessClientUpdate(const Asura_Message_PhysicalObject_ClientUpdate* pxData, float fServerTime, bool bProcessPosition)
// --------------------------------------------------------------------------------------------------------------------------------------------------------------
//
// Process a message from the server, setup anims, skins, physical; properties, etc...
// Also has an option (default is ON) to add the data passed by the server into the interpolator.
// Now, for internet players, the position is predicted by the client, and corrected by the server. 
// the data passed by the server in the default update message is therefore irrelevant and should 
// not be used to update the current position (the local player dynamics will takes care of it).
// --------------------------------------------------------------------------------------------------------------------------------------------------------------
void Asura_ClientEntity_PhysicalObject::ProcessClientUpdate(const Asura_Message_PhysicalObject_ClientUpdate* pxData, float fServerTime, bool bProcessPosAndOri /* = true */)
{ 
    Asura_Assert(pxData, "no data");

	UpdateHealth(pxData->m_fHealth, fServerTime);
    
	if(!m_bClientControlsShapeSkinAndAnim)
	{
		if(pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV)
		{
			//mesh for this object is part of the environment , so don't bother setting the shape
			SetShapeID(ASURA_HASH_ID_UNSET, fServerTime);
		}
		else if (pxData->m_uAnimID || (pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISSKIN))
    	{
			//Assume this is a skin and anim if the skin flag is set , or if an animation has been provided.
			//(The latter test is probably redundant , but I'm leaving in just in case I've missed something).
			//Assert added below to check if there any ways in which the skin flag may be being missed.
			//RAR 27/06/2006
			Asura_Assert(pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISSKIN,"Client update message has an animation , but the skin flag hasn't been set");
			
			SetSkinAndAnimID(pxData->m_uShapeFileID, pxData->m_uAnimID, pxData->m_iAnimFlags, pxData->m_fAnimTimer,fServerTime);

			if(pxData->m_iAnimFlags & Asura_PhysicalObject_AnimationFlag_UseEndTime)
			{
				//Playing animation to position , set the end timer if needed
				Asura_Animation* pxAnim = m_xShapeInstance.GetAnimation();
				if(pxAnim)
				{
					float fEndTime = pxAnim->m_xPrimaryWrapper.GetEndTime();
					float fDiff = pxData->m_fAnimEndTime - fEndTime;

					//Need to update the end time if currently looped ,or if the difference in end times is approximately an integer
					if(pxAnim->m_xPrimaryWrapper.GetLooped() || Asura_Maths::Abs(fDiff - (float)Asura_Maths::FloatToInt_Near(fDiff)) > 0.001f)
					{
						pxAnim->m_xPrimaryWrapper.SetEndTime(pxData->m_fAnimEndTime);
					}
				}
			}

            Asura_Animation* pxAnim = m_xShapeInstance.GetAnimation();
            if( pxAnim )
            {
                pxAnim->SetParentEntity( GetGuid() );
            }
    	}
    	else
    	{
    	    SetShapeID(pxData->m_uShapeFileID, fServerTime);
    	}

		// These flags are render related - take notice of them since the client is NOT in charge...
		if( pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_NOSHADOW )
		{
			m_xShapeInstance.AddFlags( ASURA_SHAPE_INSTANCE_NO_SHADOW );
		}
		else
		{
			m_xShapeInstance.RemoveFlags( ASURA_SHAPE_INSTANCE_NO_SHADOW );
		}

		if( pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_IGNOREMODULELINKING )
		{
			m_xShapeInstance.AddFlags( ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING );
		}
		else
		{
			m_xShapeInstance.RemoveFlags( ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING );
		}

		if(pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_FULLBRIGHT)
		{
			m_xShapeInstance.AddFlags( ASURA_SHAPE_INSTANCE_FULLBRIGHT );
		}
		else
		{
			m_xShapeInstance.RemoveFlags( ASURA_SHAPE_INSTANCE_FULLBRIGHT );
		}

		if(pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_USEANIMMOVEMENT)
		{
			m_xShapeInstance.AddFlags( ASURA_SHAPE_INSTANCE_USE_ANIM_MOVEMENT );
		}
		else
		{
			m_xShapeInstance.RemoveFlags( ASURA_SHAPE_INSTANCE_USE_ANIM_MOVEMENT );
		}
	}

	bool bSkyboxObject = (pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY) ? true : false;
	SetSkyboxObject(bSkyboxObject);

	// Nothing to do with rendering - take notice of the server's data...
	// Skybox objects should I think be ignored for LOS. RAR 15/02/2008
	bool bIgnoreForLOS = m_bSkyboxObject || (pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_IGNOREFORLOS);
	SetIgnoreForLOS(bIgnoreForLOS);

#ifdef ASURA_NO_NETWORK_LAG_COMPENSATION
    // (MH - 13.12.2002) SI pos and ori never seemed to get initialised. Try doing this here instead of in render().
    if (m_bIsEntityCurrentlyMoving)
    {
        Asura_Matrix_3x3 xMat;
        pxData->m_xOrientation.ToMatrix(xMat);
        m_xShapeInstance.SetNewPosAndOri(pxData->m_xPosition, xMat);
    }
#endif

	// ART - 30/10/2006 - Don't allow positional updates if we're hooked
	// to a parent entity, our data comes from here.
	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
		// (NRM - 11.07.2003) update the static objects once and once only
		if(pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_STATIC)
		{
			#ifdef ASURA_TOOLS
			//If currently being edited in tools , then process the update even if it has previously been done
			//for this entity. RAR 19/02/2008
			if(Tools_System::IsActive()) ForceUpdate();
			#endif
			
			if(!m_bStaticDoneUpdate || m_bAnimChanged)
			{
				ProcessNewPositionAndOrientationData(pxData->m_xPosition, pxData->m_xOrientation, fServerTime);
				Asura_Matrix_3x3 xOri;
				pxData->m_xOrientation.ToMatrix(xOri);
				if( m_xShapeInstance.GetShape() || m_xShapeInstance.GetAnimation() )
				{
					m_xShapeInstance.SetNewPosAndOri(pxData->m_xPosition, xOri);
					#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
					if( m_xShapeInstance.GetShape() )
					{
						m_xShapeInstance.CalculateModuleList();
					}				
					#endif //!defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
				}
				m_bStaticDoneUpdate = true;
				m_bAnimChanged = false;
			}
			else
			{
				m_bIsEntityCurrentlyMoving = false;
			}
		}
		else
		{
			if (bProcessPosAndOri)
			{
				ProcessNewPositionAndOrientationData(pxData->m_xPosition, pxData->m_xOrientation, fServerTime);
			}
		}
	}

	//Set invisibility flag
	SetInvisibleByPhysicalFlag((pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_INVISIBLE) ? true : false);

	if(m_xShapeInstance.GetAnimation())
	{
		//update texture animation sequence/speed in animation
		m_xShapeInstance.SetTextureAnimSpeed(pxData->m_sTextureAnimSpeed/100.0f);
		m_xShapeInstance.SetTextureAnimSequence(pxData->m_ucTextureAnimSequence);
	}

    // (MH - 22.02.2007) check parent link status
	if (fServerTime > m_fParentLinkServerTime)
	{
		if (m_uParentGuid != pxData->m_uParentGuid)
		{
			if (pxData->m_uParentGuid == ASURA_GUID_UNREGISTERED)
			{
				UnlinkFromParent(fServerTime);
			}
			else
			{
				LinkToParent( pxData->m_uParentGuid, pxData->m_uParentMarkerHashID, fServerTime, (pxData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ATTACHEDWITHPHYSICS) ? true : false );
			}
		}
	}

}

///////////////////////////////////////////////////
// Update health in response to a server message //
///////////////////////////////////////////////////
void Asura_ClientEntity_PhysicalObject::UpdateHealth(float fHealth, float fServerTime)
{
	if (fServerTime < m_fLastHealthServerTimestamp) return; // old message, ignore
	m_fLastHealthServerTimestamp = fServerTime; // update timestamp

	if (m_fHealth == fHealth) return; // no change

	// Health has changed!
	if (m_fHealth>0 && fHealth<=0)
	{
		// We just died!
		MakeDead();
		fHealth = 0.0f;
	}
    else
    {
        // (MH - 30.11.2002) added this so derived classes can do client-only effects e.g. flashing.
        HealthChanged(fHealth);
    }

	m_fHealth = fHealth;
}

// (MH - 09.10.2002) call InitialiseHealth at object creation time only
void Asura_ClientEntity_PhysicalObject::InitialiseHealth(float fHealth)
{
    m_fHealth = fHealth;
}

void Asura_ClientEntity_PhysicalObject::MakeDead()
{
}

void Asura_ClientEntity_PhysicalObject::SetEntityCurrentlyMoving(bool bMoving)
{
	if (m_xShapeInstance.IsShape())
	{
		Asura_Assert(bMoving != m_bIsEntityCurrentlyMoving, "Unnecessary call to Asura_Entity_Renderable::SetEntityCurrentlyMoving");
		if (m_bIsEntityCurrentlyMoving)
		{
			// We have just stopped moving.
			// Calculate the modules that we are in now for quick culling.
#ifndef ASURA_NO_NETWORK_LAG_COMPENSATION
			Asura_Matrix_3x3 xOri;
			GetLastReportedOrientationMatrix(xOri);
			m_xShapeInstance.SetNewPosAndOri(GetLastReportedPosition(), xOri);
#endif //ASURA_NO_NETWORK_LAG_COMPENSATION
			
			#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
			m_xShapeInstance.CalculateModuleList();
			#endif //defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
		}
		#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
		else
		{
			// We have started moving.
			// Make the quicker culling data invalid because re-calculating
			// each frame is a waste of time.
			m_xShapeInstance.InvalidateModuleList();
		}
		#endif //#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	}
	m_bIsEntityCurrentlyMoving = bMoving;
}


////////////
// Render //
////////////
void Asura_ClientEntity_PhysicalObject::Render()
{
	if(m_bSkyboxObject) return;

	// ART - 30/10/2006 - Don't allow rendering if we're hooked
	// to a parent entity, we will be rendered by the parent.
	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
#ifndef ASURA_NO_NETWORK_LAG_COMPENSATION
		const Asura_Vector_3&   xPos = GetCurrentPosition();
		const Asura_Matrix_3x3& xOri = GetCurrentOrientation();
		// (MH - 13.12.2002) do things like this in Update() / HandleMessage()
		if (m_bIsEntityCurrentlyMoving)
		{
			// If the entity is moving, then update its position.
			m_xShapeInstance.SetNewPosAndOri(xPos, xOri);
		}
#endif

		// KOR - 27/03/2006 - commented out assert. Unfortunately some project code
		// changes m_bInvisible directly rather than calling SetInvisible. We could
		// make m_bInvisible private, but would need to fix up all project code using it.
		//Asura_Assert(!m_bInvisible, "Shouldn't be calling render if we're invisible - SetInvisible will remove entity from renderable list");
		if ( !IsInvisible() && !m_xShapeInstance.ShouldNeverBeRendered())
		{
			m_xShapeInstance.Render();
		}
	}
	else
	{
		const Asura_Animation*	pxAnimation		= m_xShapeInstance.GetAnimation();
		const Asura_Vector_3&	xPosition		= pxAnimation->GetPosition();
		Asura_Matrix_3x3		xOrientation;

		pxAnimation->GetOrientQuat().ToMatrix(xOrientation);
		TeleportTo(xPosition, xOrientation);

	}

#if defined(ASURA_USE_MOVEMENT_SYSTEM)
	if(m_pxMovementSystem)
	{
		m_pxMovementSystem->Render();
	}
#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)
}

void Asura_ClientEntity_PhysicalObject::RenderInSkybox(const Asura_Vector_3& xOffset, float fRed, float fGreen, float fBlue)
{
	if ( !IsInvisible() && !m_xShapeInstance.ShouldNeverBeRendered())
	{
				
		m_xShapeInstance.AddFlags(ASURA_SHAPE_INSTANCE_NO_SHADOW | ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING | ASURA_SHAPE_INSTANCE_AMBIENT_ONLY);
		m_xShapeInstance.Platform_SetAmbientRGB((u_char)Asura_Maths::FloatToInt(fRed/2.0f), (u_char)Asura_Maths::FloatToInt(fGreen/2.0f), (u_char)Asura_Maths::FloatToInt(fBlue/2.0f));

		m_xShapeInstance.SetNewPosAndOri(GetCurrentPosition()+xOffset, GetCurrentOrientation());
		m_xShapeInstance.Render();
	}
}

void Asura_ClientEntity_PhysicalObject::RenderShadow()
{
	// ART - 30/10/2006 - Don't allow rendering if we're hooked
	// to a parent entity, we will be rendered by the parent.
	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
		if( !IsInvisible() && !m_xShapeInstance.ShouldNeverBeRendered() )
		{
			// JAT 28/02/07 only call render shadow if the shape is marked to cast shadows
			if ( ( m_xShapeInstance.GetFlags() & ASURA_SHAPE_INSTANCE_NO_SHADOW ) == 0 )
			{
				m_xShapeInstance.RenderShadow();
			}
		}
	}
}

/////////////
// LosTest //
/////////////

bool Asura_ClientEntity_PhysicalObject::LOSTest(const Asura_LOS_InputData* pxInputData, Asura_LOS_OutputData* pxOutputData)
{
    if ( IsInvisible() ) { return false; }
	if(m_bIgnoreForLOS) return false;

	if(!pxInputData || !pxOutputData)
	{
		Asura_Assert(0,"Null pointer");
		return false;
	}

	if (pxInputData->m_uEntityToIgnore == Guid) return false;

	// Do LOS test against the mesh or skin.
	const Asura_Vector_3&   xPos = GetCurrentPosition();
	const Asura_Matrix_3x3& xOri = GetCurrentOrientation();

    if (m_xShapeInstance.LOSTest( xPos, xOri, pxInputData, pxOutputData))
    {           
		// A closer hit. Set the guid.
        pxOutputData->m_uHitEntity = Guid;
        return true;
    }

    return false;
}

//////////////
// SetShape //
//////////////
void Asura_ClientEntity_PhysicalObject::SetShapeID(Asura_Hash_ID uShapeID, float fServerTime)
{
	if (fServerTime < m_fLastShapeIDServerTimestamp) return; // old message, ignore
	m_fLastShapeIDServerTimestamp = fServerTime;

	/// Richard May - 22/01/2003 09:49:27 - removed assertion as certain objects, like lights, may
	/// not need to have a shape
	//Asura_Assert(uShapeID != ASURA_HASH_ID_UNSET, "");

	m_xShapeInstance.SetShape( uShapeID );
}

//void Asura_ClientEntity_PhysicalObject::SetShapeScale(const float fScale)
//{
//	m_xShapeInstance.SetScale(fScale, fScale, fScale);
//}


void Asura_ClientEntity_PhysicalObject::SetSkinAndAnimID(Asura_Hash_ID uShapeID, Asura_Hash_ID uAnimID, int iAnimFlags, float fAnimTimer,float fServerTime)
{
    if (fServerTime < m_fLastShapeIDServerTimestamp) return; // old message, ignore
	m_fLastShapeIDServerTimestamp = fServerTime;

    Asura_Assert(uShapeID != ASURA_HASH_ID_UNSET, "");
	m_xShapeInstance.SetSkinAndAnimID(uShapeID, uAnimID, iAnimFlags, fAnimTimer);
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_ClientEntity_PhysicalObject::SetAmbientColour().
// PARAMETERS   uARGB         - ARGB to be set in the shape instance
//              bForceAmbient - no directional lighting if true.
// RETURNS      None.
// DESCRIPTION  Sets the relevant bits of the Asura_Shape_Instance
// AUTHOR       Mike Healey - 28.11.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_ClientEntity_PhysicalObject::SetAmbientColour(u_int uARGB, bool bForceAmbient)
{
    m_xShapeInstance.Platform_SetAmbientARGB(uARGB);
    if (bForceAmbient)
    {
        m_xShapeInstance.AddFlags(ASURA_SHAPE_INSTANCE_AMBIENT_ONLY);
    }
    else
    {
        m_xShapeInstance.RemoveFlags(ASURA_SHAPE_INSTANCE_AMBIENT_ONLY);
    }
}
// End function Asura_ClientEntity_PhysicalObject::SetAmbientColour().

void Asura_ClientEntity_PhysicalObject::InitialiseShapePositionAndOrientation(void)
{
	m_xShapeInstance.SetNewPosAndOri(GetCurrentPosition(), GetCurrentOrientation());

	#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	if( m_xShapeInstance.GetShape() )
	{
		m_xShapeInstance.CalculateModuleList();
	}				
	#endif// !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
}



void Asura_ClientEntity_PhysicalObject::HideForCutscene(bool bHide)
{
	//tell server entity
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_HIDEFORCUTSCENE,sizeof(unsigned char));
	if (pxMessage)
	{
		unsigned char* pucHide = (unsigned char*) pxMessage->GetData();
		*pucHide = bHide ? 1 : 0;
		pxMessage->To =	Guid;
		pxMessage->From = Guid;
		Asura_Network::Send(pxMessage);
	}

	//and deal with the hiding/unhiding
	SetInvisibleByCutscene(bHide);
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_PhysicalObject::LinkToParent
// Author       : Andy Tate
// Date         : 30-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_ClientEntity_PhysicalObject::LinkToParent (const Asura_Guid uParentGuid,
													  const Asura_Hash_ID uMarkerHash,
													  float fServerTime,
													  const bool bWithPhysics)
{
	if (m_fParentLinkServerTime > fServerTime)
	{
		// old message - ignore
		return true;
	}

	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
		Asura_ClientEntity_PhysicalObject*	pxParentClient	= Asura_ClientEntity_PhysicalObject::FindEntity(uParentGuid);

		if(pxParentClient == NULL)
		{
			// (MH - 22.02.2007) we should be able to deal with this in the update message now.
			//Asura_Assert(false, "Failed to find entity to link to.");
			return(false);
		}
		else
		{
			Asura_Shape_Instance&	xParentShapeInstance	= pxParentClient->GetShapeInstance();
			Asura_Animation*		pxParentAnimation		= xParentShapeInstance.GetAnimation();

			if(pxParentAnimation == NULL)
			{
				return(false);
			}
			else
			{
				Asura_Animation*	pxAnimation	= m_xShapeInstance.GetAnimation();

				if(pxAnimation == NULL)
				{
					Asura_Assert(false, "Child entity doesn't have an animation to link to.");
					return(false);
				}
				else
				{
					m_pxParentAttachment	= new Asura_AnimationAttachment_Animation (pxAnimation, false);

					if(m_pxParentAttachment->LinkAttachment(pxParentAnimation, uMarkerHash, true))
					{
						m_uParentGuid	= uParentGuid;
						m_fParentLinkServerTime = fServerTime;
                        // (NY - 07.03.2007) store whether we're using physics - if projects update the attached physics using messages, the client will need to know to message
						m_bAttachedWithPhysics = bWithPhysics;
						UpdateVisibilityWhileAttached();
						return(true);
					}
					else
					{
						Asura_Assert(false, "Failed to link to our parent entity.");
						m_pxParentAttachment = NULL;
						return(false);
					}
				}
			}
		}
	}
	else
	{
		Asura_Assert(false, "We already have a parent - cannot link to a second one.");
		return(false);
	}
}

/// End of function 'Asura_ClientEntity_PhysicalObject::LinkToParent'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_PhysicalObject::LinkToParent
// Author       : Andy Tate
// Date         : 6-Nov-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_ClientEntity_PhysicalObject::LinkToParent (const Asura_Guid uParentGuid,
													  const Asura_Vector_3& xBoneRelativePos,
													  const Asura_Quat& xBoneRelativeOri,
													  const u_int uBoneID,
													  float fServerTime)
{
	if (m_fParentLinkServerTime > fServerTime)
	{
		// old message - ignore
		return true;
	}

	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
		Asura_ClientEntity_PhysicalObject*	pxParentClient	= Asura_ClientEntity_PhysicalObject::FindEntity(uParentGuid);

		if(pxParentClient == NULL)
		{
			Asura_Assert(false, "Failed to find entity to link to.");
			return(false);
		}
		else
		{
			Asura_Shape_Instance&	xParentShapeInstance	= pxParentClient->GetShapeInstance();
			Asura_Animation*		pxParentAnimation		= xParentShapeInstance.GetAnimation();

			if(pxParentAnimation == NULL)
			{
				Asura_Assert(false, "Parent entity doesn't have an animation to link to.");
				return(false);
			}
			else
			{
				Asura_Animation*	pxAnimation	= m_xShapeInstance.GetAnimation();

				if(pxAnimation == NULL)
				{
					Asura_Assert(false, "Child entity doesn't have an animation to link to.");
					return(false);
				}
				else if(pxParentAnimation->LinkAnimation_BoneRelativeObjectSpace(pxAnimation, xBoneRelativePos, xBoneRelativeOri, uBoneID, false))
				{
					m_uParentGuid	= uParentGuid;
					m_fParentLinkServerTime = fServerTime;
					return(true);
				}
				else
				{
					Asura_Assert(false, "Failed to link to our parent entity.");
					return(false);
				}
			}
		}
	}
	else
	{
		Asura_Assert(false, "We already have a parent - cannot link to a second one.");
		return(false);
	}
}

/// End of function 'Asura_ClientEntity_PhysicalObject::LinkToParent'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_PhysicalObject::LinkToParent
// Author       : Andy Tate
// Date         : 6-Nov-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_ClientEntity_PhysicalObject::LinkToParent (const Asura_Guid uParentGuid,
													  const Asura_Hash_ID uMarkerHash,
													  const u_int uBoneID,
													  const Asura_Vector_3& xPosition,
													  float fServerTime)

{
	if (m_fParentLinkServerTime > fServerTime)
	{
		// old message - ignore
		return true;
	}

	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
		Asura_ClientEntity_PhysicalObject*	pxParentClient	= Asura_ClientEntity_PhysicalObject::FindEntity(uParentGuid);

		if(pxParentClient == NULL)
		{
			Asura_Assert(false, "Failed to find entity to link to.");
			return(false);
		}
		else
		{
			Asura_Shape_Instance&	xParentShapeInstance	= pxParentClient->GetShapeInstance();
			Asura_Animation*		pxParentAnimation		= xParentShapeInstance.GetAnimation();

			if(pxParentAnimation == NULL)
			{
				Asura_Assert(false, "Parent entity doesn't have an animation to link to.");
				return(false);
			}
			else
			{
				Asura_Animation*	pxAnimation	= m_xShapeInstance.GetAnimation();

				if(pxAnimation == NULL)
				{
					Asura_Assert(false, "Child entity doesn't have an animation to link to.");
					return(false);
				}
				else
				{
					m_pxParentAttachment	= new Asura_AnimationAttachment_Sticky (pxAnimation, false, uBoneID, xPosition);

					if(m_pxParentAttachment->LinkAttachment(pxParentAnimation, uMarkerHash, true))
					{
						m_uParentGuid	= uParentGuid;
						m_fParentLinkServerTime = fServerTime;
						UpdateVisibilityWhileAttached();
						return(true);
					}
					else
					{
						Asura_Assert(false, "Failed to link to our parent entity.");
						m_pxParentAttachment = NULL;
						return(false);
					}
				}
			}
		}
	}
	else
	{
		Asura_Assert(false, "We already have a parent - cannot link to a second one.");
		return(false);
	}
}

/// End of function 'Asura_ClientEntity_PhysicalObject::LinkToParent'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_PhysicalObject::UnlinkFromParent
// Author       : Andy Tate
// Date         : 30-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_ClientEntity_PhysicalObject::UnlinkFromParent (float fServerTime)
{
	if (m_fParentLinkServerTime > fServerTime)
	{
		// old message - ignore
		return true;
	}

	if(m_uParentGuid == ASURA_GUID_UNREGISTERED)
	{
		Asura_Assert(false, "We don't have a parent to unlink from.");
		return(false);
	}
	else
	{
		Asura_ClientEntity_PhysicalObject*	pxParentClient	= 
			Asura_ClientEntity_PhysicalObject::FindEntityIncludingDestroyed(m_uParentGuid);

		if(pxParentClient == NULL)
		{
			m_uParentGuid = ASURA_GUID_UNREGISTERED;
			m_fParentLinkServerTime = fServerTime;
            // (MH - 01.02.2007) set parent guid to unregistered anyway, and don't assert.
			// The client entity (e.g. player) might have been deleted and it doesn't know
			// about any linked entities so we always get here.
			return(false);
		}
		else
		{
			Asura_Shape_Instance&	xParentShapeInstance	= pxParentClient->GetShapeInstance();
			Asura_Animation*		pxParentAnimation		= xParentShapeInstance.GetAnimation();

			if(pxParentAnimation == NULL)
			{
				Asura_Assert(false, "Parent entity doesn't have an animation to unlink from.");
				return(false);
			}
			else
			{
				Asura_Animation*	pxAnimation	= m_xShapeInstance.GetAnimation();

				if(pxAnimation == NULL)
				{
					Asura_Assert(false, "Child entity doesn't have an animation to be unlinked.");
					return(false);
				}
				else if(pxParentAnimation->UnlinkAnimation(pxAnimation))
				{
					const Asura_Vector_3&	xPosition	= pxAnimation->GetPosition();
					Asura_Matrix_3x3		xOrientation;

					// Teleport the entity to the new position...
					pxAnimation->GetOrientQuat().ToMatrix(xOrientation);
					TeleportTo(xPosition, xOrientation);

					m_xShapeInstance.SetNewPosAndOri(xPosition, xOrientation);

					// ...before completing the unlinking...
					m_uParentGuid			= ASURA_GUID_UNREGISTERED;
					m_pxParentAttachment	= NULL;
					m_fParentLinkServerTime = fServerTime;
					m_bAttachedWithPhysics = false;
					return(true);
				}
				else
				{
					Asura_Assert(false, "Failed to unlink to our parent entity.");
					return(false);
				}
			}
		}
	}
}

/// End of function 'Asura_ClientEntity_PhysicalObject::UnlinkFromParent'. ///

void Asura_ClientEntity_PhysicalObject::UpdateVisibilityWhileAttached()
{
	if(!m_pxParentAttachment) return;

	//The attachment code won't know about the IsInvisible() status of the entity while rendering , so
	//get that to work by adding/removing the don't render flag from the flags to add
	//RAR 10/05/2007
	u_int uFlags = m_pxParentAttachment->GetRenderFlagsToAdd();
	if(IsInvisible()) uFlags |= ASURA_HIERARCHY_RENDER_DONTRENDER;
	else uFlags &=~ ASURA_HIERARCHY_RENDER_DONTRENDER;

	m_pxParentAttachment->SetRenderFlagsToAdd(uFlags);
}

void Asura_ClientEntity_PhysicalObject::OnVisibilityChanged()
{
	if(m_pxParentAttachment)
	{
		UpdateVisibilityWhileAttached();
	}

	if (GetShapeInstance().GetAnimation())
	{
		Asura_AnimationAttachment* pxAttachment = GetShapeInstance().GetAnimation()->GetFirstAttachment();
		while (pxAttachment)
		{
			pxAttachment->OnParentVisibilityChanged( IsInvisible() ); 
			pxAttachment = pxAttachment->GetNextAttachment();
		}
	}
}

// Returns the conservative axially-aligned WORLD SPACE box covering the entity. 
void Asura_ClientEntity_PhysicalObject::GetLineOfSightAABB( Asura_Bounding_Box& xAABB ) 
{
	Asura_Animation* pxAnim = m_xShapeInstance.GetAnimation();

	if ( pxAnim )
	{
		pxAnim->GetCoarseBoundingBox( xAABB, GetCurrentPosition(), GetCurrentOrientation() );
	}
	else
	{
		xAABB = m_xShapeInstance.GetBB();
	}
}

void Asura_ClientEntity_PhysicalObject::PreDestructionShutdown()
{
#ifdef ASURA_USE_MOVEMENT_SYSTEM
	// The movement system holds a reference to this physical object, so we need to make sure it cleans everything up before this begins to be deleted.
	if (m_pxMovementSystem)
	{
		m_pxMovementSystem->Reset();
	}
#endif //#ifdef ASURA_USE_MOVEMENT_SYSTEM

	PARENT::PreDestructionShutdown();
}

#ifdef ASURA_USE_MOVEMENT_SYSTEM
const Asura_Vector_3& Asura_ClientEntity_PhysicalObject::GetCurrentPosition( void ) const
{
	if(CanCurrentlyUseMoves() && m_pxMovementSystem->IsMoveInProgress())
	{
		const Asura_Move* pxCurrentMove = m_pxMovementSystem->GetCurrentMove();
		
		if(pxCurrentMove->GetPositionUsage() == Asura_Move::POSITION_SET_DIRECTLY)
		{
			m_pxMovementSystem->GetCurrentMovePosition(m_xCurrentPosition);				
			return m_xCurrentPosition;
		}
	}
	return PARENT::GetCurrentPosition();
}

const Asura_Matrix_3x3& Asura_ClientEntity_PhysicalObject::GetCurrentOrientation( void ) const
{
	if(CanCurrentlyUseMoves() && m_pxMovementSystem->IsMoveInProgress())
	{
		const Asura_Move* pxCurrentMove = m_pxMovementSystem->GetCurrentMove();

		if(pxCurrentMove->GetOrientationUsage() == Asura_Move::ORIENTATION_SET_DIRECTLY)
		{
			m_pxMovementSystem->GetCurrentMoveOrientation(m_xCurrentOrientation);
			return m_xCurrentOrientation;
		}
	}
	return PARENT::GetCurrentOrientation();
}

void Asura_ClientEntity_PhysicalObject::GetCurrentVelocity( Asura_Vector_3& xVelocity ) const
{
	if(CanCurrentlyUseMoves() && m_pxMovementSystem->IsMoveInProgress())
	{
		const Asura_Move* pxCurrentMove = m_pxMovementSystem->GetCurrentMove();

		if(pxCurrentMove->GetPositionUsage() == Asura_Move::POSITION_SET_DIRECTLY)
		{
			m_pxMovementSystem->GetCurrentMoveVelocity(xVelocity);
			return;
		}	
	}

	PARENT::GetCurrentVelocity( xVelocity );
}

void Asura_ClientEntity_PhysicalObject::Update()
{
	if( m_pxMovementSystem )
	{
		m_pxMovementSystem->Update();
	}
	PARENT::Update();
}

#endif //ASURA_USE_MOVEMENT_SYSTEM

// overrides of OnEntityActivate()/OnEntityDeactivate() - called when active status changes or is initialised. 
// Derived classes should always call parent versions if they override them. 
// We remove the entity from the renderable lists as appropriate on deactivation and put it back in on activation.
void Asura_ClientEntity_PhysicalObject::OnEntityActivate(void) 
{
	PARENT::OnEntityActivate();

	AddToLineOfSightList();
}

void Asura_ClientEntity_PhysicalObject::SetSkyboxObject(bool bSet)
{
	if(bSet != m_bSkyboxObject)
	{
		m_bSkyboxObject = bSet;
		//add/remove this object from the list in Asura_Skybox
		if(m_bSkyboxObject)
		{
			Asura_Skybox::AddSkyboxObject(this);
		}
		else
		{
			Asura_Skybox::RemoveSkyboxObject(this);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_PhysicalObject::SetIgnoreForLOS
// Author       : Andy Tate
// Date         : 17-Mar-2008
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_PhysicalObject::SetIgnoreForLOS(bool bSet)
{
	if(bSet && !m_bIgnoreForLOS)
	{
		RemoveFromLineOfSightList();
		m_bIgnoreForLOS	= true;
	}
	else if(!bSet && m_bIgnoreForLOS)
	{
		AddToLineOfSightList();
		m_bIgnoreForLOS	= false;
	}
}

/// End of function 'Asura_ClientEntity_PhysicalObject::SetIgnoreForLOS'. ///


/// End of file. ///
