//////////////////////////////////////////////////////////////////////////////////////////
//
// FILE : Asura_ClientEntity_PhysicalObject.h
//
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_CLIENT_ENTITY_PHYSICAL_OBJECT_H__
#define __ASURA_CLIENT_ENTITY_PHYSICAL_OBJECT_H__


//////////////
// INCLUDES //
//////////////

#include "Animation/Asura_AnimationAttachment.h"
#include "Entity/Asura_Entity.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_HashedList.h"
#include "Shapes/Asura_Shapes.h"

//////////////////////////
// FORWARD DECLARATIONS //
//////////////////////////

class Asura_Movement_System;
class Asura_Msg_PlayPFXOnEntity;

///////////////////////////////////////
// Asura_ClientEntity_PhysicalObject //
///////////////////////////////////////

class Asura_ClientEntity_PhysicalObject : public Asura_Entity_Renderable
{
protected:
	Asura_ClientEntity_PhysicalObject();
	virtual ~Asura_ClientEntity_PhysicalObject();

	virtual void InitialiseFromChunkBuffer(const char*& pcChunkBuffer);

public:
	inline bool IsDead() const { return m_fHealth <= 0.0f; }
	float GetHealth() const { return m_fHealth; }

	void SetShapeID(Asura_Hash_ID uShapeID, float fServerTime);
    void SetSkinAndAnimID(Asura_Hash_ID uShapeID, Asura_Hash_ID uAnimID, int iAnimFlags, float fAnimTimer,float fServerTime);
	void SetShapeScale(const float fScale);
    void SetAmbientColour(u_int uARGB, bool bAmbientOnly);// (MH - 28.11.2002) use this to flash objects.
        
	// DWE If you don't render the shape for a static object straight away,
	// it does have it's position/orientation initialised becuase the m_bNotMoving bool is set.
	// Then when you do want to render it asserts. Use this function to fix that.
	void InitialiseShapePositionAndOrientation(void);

	virtual bool HandleMessage(const Asura_Message* pxMessage);

	// This function can be overloaded to do any special FX when killed 
    // (eg play death anim, explosion etc). 
    // (MH - 15.10.2002) used to SetAsDestroyed(), this was agreed to be wrong for the default case.
    virtual void MakeDead();

    // Sub-classes should override these if they're not using the animation
    // pointer in the shape instance.
    virtual const Asura_Animation* GetAnimation( void ) const { return m_xShapeInstance.GetAnimation(); }
    virtual Asura_Animation* GetAnimation( void ) { return m_xShapeInstance.GetAnimation(); }

    inline const Asura_Shape_Instance& GetShapeInstance( void ) const { return m_xShapeInstance; }
    inline Asura_Shape_Instance& GetShapeInstance( void ) { return m_xShapeInstance; }


	//Gets the world position/orientation for a marker point , providing this entity
	//uses a skin/hierarchy that has a matching marker point.
	bool GetMarkerLocationFromHash( Asura_Hash_ID uMarkerHash, Asura_Vector_3* pxOutputPosition, Asura_Matrix_3x3* pxOutputOrient, bool bWorldSpace=true, int iAttachID=Asura_AnimationAttachment::INVALID_ID ) const
	{ return m_xShapeInstance.GetMarkerLocationFromHash(uMarkerHash,pxOutputPosition,pxOutputOrient,bWorldSpace,iAttachID); }

	//Called by cutscene code to hide/unhide the entity
	void HideForCutscene(bool bHide);

	virtual void RenderShadow();
		
	//called by Asura_Skybox to render skybox objects
	void RenderInSkybox(const Asura_Vector_3& xOffset, float fRed, float fGreen, float fBlue);

	// Parent linkage functions...
	virtual bool LinkToParent		(const Asura_Guid uParentGuid, const Asura_Hash_ID uMarkerHash, float fServerTime, const bool bWithPhysics);
	virtual bool LinkToParent		(const Asura_Guid uParentGuid, const Asura_Vector_3& xBoneRelativePos, const Asura_Quat& xBoneRelativeOri, const u_int uBoneID, float fServerTime);
	virtual bool LinkToParent		(const Asura_Guid uParentGuid, const Asura_Hash_ID uMarkerHash, const u_int uBoneID, const Asura_Vector_3& xPosition, float fServerTime);
	virtual bool UnlinkFromParent	(float fServerTime);

	Asura_Guid							GetParentGuid		(void) const	{ return(m_uParentGuid); }
	Asura_AnimationAttachment*			GetParentAttachment	(void)			{ return(m_pxParentAttachment); }
	const Asura_AnimationAttachment*	GetParentAttachment	(void) const	{ return(m_pxParentAttachment); }
	float								GetParentLinkServerTime(void) const	{ return(m_fParentLinkServerTime); }
	bool								IsAttachedWithPhysics() const		{ return(m_bAttachedWithPhysics); }


#ifdef ASURA_USE_MOVEMENT_SYSTEM
	virtual void Update(void);

	virtual bool CanCurrentlyUseMoves() const	{ return( false && ( m_pxMovementSystem!=NULL) ); }

	Asura_Movement_System*			GetMovementSystem()				{ return m_pxMovementSystem; }
	const Asura_Movement_System*	GetMovementSystem() const		{ return m_pxMovementSystem; }

	virtual const Asura_Vector_3&   GetCurrentPosition( void ) const;
	virtual const Asura_Matrix_3x3& GetCurrentOrientation( void ) const;
	virtual void					GetCurrentVelocity( Asura_Vector_3& xVelocity ) const;
#endif

	virtual void OnVisibilityChanged();

	// Returns the conservative axially-aligned WORLD SPACE box covering the entity. 
	virtual void GetLineOfSightAABB( Asura_Bounding_Box& xAABB );

	// The behaviour of physical objects before the creation of this function was to stop dialogue if dead, so we'll match that here.
	virtual bool GetCanPlayDialogue() const { return !IsDead(); }

	virtual void PreDestructionShutdown();

protected:
	// override of OnEntityActivate() - to handle adding entity to line of sight list
	// Derived classes should always call parent version if they override this. 
	virtual	void OnEntityActivate(void);

	virtual void Render();
	virtual bool LOSTest(const Asura_LOS_InputData* pxInputData, Asura_LOS_OutputData* pxOutputData);
    
    // note that in the case of internet player updates, the position and orientation
	// are set by the local player code directly, so a parameter has been added to 
	// this function so it can ignore them. 
	void ProcessClientUpdate(const class Asura_Message_PhysicalObject_ClientUpdate* pxData, float fServerTime, bool bProcessPosAndOri = true);

	void UpdateHealth(float fHealth, float fServerTime);
    virtual void HealthChanged(float fHealth) {}    // called by update health if not dead
    // (MH - 09.10.2002) call InitialiseHealth at object creation time only
    void InitialiseHealth(float fHealth);
	void SetEntityCurrentlyMoving(bool bMoving);
	void ForceUpdate(void) { m_bStaticDoneUpdate = false; }

	float GetHealthTimestamp() const { return m_fLastHealthServerTimestamp; }

	//Some derived client classes may have a better idea of what animation etc they should be using than the
	//server does. Setting this flag will cause the client to ignore the shape,skin and anim hash IDs
	//that get passed to it in the client update messages. RAR 22/05/2006
	void SetClientControlsShapeSkinAndAnim(bool bSet) {m_bClientControlsShapeSkinAndAnim = bSet;}
	bool GetClientControlsShapeSkinAndAnim() const {return m_bClientControlsShapeSkinAndAnim;}

	bool GetIgnoreForLOS() const { return m_bIgnoreForLOS; }
	
	//Used by entities that are linked to parent entities via attachments , since otherwise the IsInvisible() status would be ignored
	void UpdateVisibilityWhileAttached();

	// Process "Play PFX on Entity" message.
	void ProcessPlayPFXOnEntity(const Asura_Msg_PlayPFXOnEntity* pxData);

#ifdef ASURA_USE_MOVEMENT_SYSTEM
	Asura_Movement_System* m_pxMovementSystem;
#endif

	void SetSkyboxObject(bool bSet);
	void SetIgnoreForLOS(bool bSet);

private:
	float m_fHealth; // Note: Zero = dead, Positive = alive. (Negative not allowed).
	float m_fLastHealthServerTimestamp;

	Asura_Shape_Instance m_xShapeInstance;
	float m_fLastShapeIDServerTimestamp;
    bool  m_bStaticDoneUpdate;
    bool  m_bAnimChanged;
	bool  m_bClientControlsShapeSkinAndAnim;
	bool  m_bIgnoreForLOS;//Set according to the server entity flag
	bool  m_bAttachedWithPhysics;
	bool  m_bSkyboxObject;

	Asura_Guid					m_uParentGuid;			// The GUID used for the parent entity.
	Asura_AnimationAttachment*	m_pxParentAttachment;	// A pointer to the parent attachment.
	float						m_fParentLinkServerTime;// time of last update on the parent entity.

public:
	/* static members */
	static Asura_ClientEntity_PhysicalObject* FindEntity(Asura_Guid uGuid) { return s_xList.FindEntity(uGuid); }
	static Asura_ClientEntity_PhysicalObject* FindEntityIncludingDestroyed(Asura_Guid uGuid) { return s_xList.FindObject(uGuid,false); }
	static const Asura_Hashed_List<Asura_ClientEntity_PhysicalObject>* GetList() { return &s_xList; }

private:
	/* static data */
	static Asura_Hashed_List<Asura_ClientEntity_PhysicalObject> s_xList;

	typedef Asura_Entity_Renderable PARENT;

};


#endif //__ASURA_CLIENT_ENTITY_PHYSICAL_OBJECT_H__
