//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_PhysicsObject.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicsObject.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int uPHYSICS_OBJECT_CHUNK_VERSION = 5;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_ServerEntity_PhysicsObject
//

Asura_Hashed_List<Asura_ServerEntity_PhysicsObject> Asura_ServerEntity_PhysicsObject::s_xList( 5 );


Asura_ServerEntity_PhysicsObject::Asura_ServerEntity_PhysicsObject( bool bAllocateDynamicsBlock )
    : Asura_Entity_Server(ASURA_GUID_UNREGISTERED) // ignored
{
    Classification = AsuraEntityClass_PhysicsObject;

    s_xList.AddToList( this );

    m_bClientUpdated = false;
    m_uPhysicalObjectFlags |= ASURA_PHYSICALOBJECT_FLAG_NOSHADOW;



    // We'll allocate the physics object when initialising from our chunk
    // data or when creating these things in tools.
    m_pxDynamicObject = NULL;

}


Asura_ServerEntity_PhysicsObject::~Asura_ServerEntity_PhysicsObject( void )
{
    s_xList.RemoveFromList( this );
}


void Asura_ServerEntity_PhysicsObject::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
    u_int uPhysicsObjectType;
    if( m_pxDynamicObject )
    {
        uPhysicsObjectType = m_pxDynamicObject->GetType();
    }
    else
    {
        uPhysicsObjectType = ( IsStatic() ? ASURA_PHYSICS_OBJECT_STATIC : ASURA_PHYSICS_OBJECT_DYNAMIC );
    }

	const int	iCollisionProxy	= GetCollisionProxy();
	const float	fGravityScale	= GetGravityScale();
	const float	fElasticity		= GetElasticity();
	const float	fFriction		= GetFriction();
	const float	fDrag			= GetDrag();
	const float	fInvMass		= GetInvMass( false );
	const float	fAngularDrag 	= GetAngularDrag();

	// Write the data...
	xStream << uPHYSICS_OBJECT_CHUNK_VERSION;
	xStream << uPhysicsObjectType;
	xStream << iCollisionProxy;
	xStream << fGravityScale;
	xStream << fElasticity;
	xStream << fFriction;
	xStream << fDrag;
	xStream << fInvMass;
	xStream << fAngularDrag;
}


void Asura_ServerEntity_PhysicsObject::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert( uVersion <= uPHYSICS_OBJECT_CHUNK_VERSION, "Unsupported chunk version" );

	const bool bUpdateWaypointBlocking = ( 0 == uVersion );

	if(uVersion > 1)
	{
        ASURA_PHYSICS_OBJECT_TYPE ePhysicsObjectType;

        if( uVersion > 4 )
        {
            u_int uPhysicsObjectType;
            xStream >> uPhysicsObjectType;
            ePhysicsObjectType = static_cast<ASURA_PHYSICS_OBJECT_TYPE>( uPhysicsObjectType );
        }
        else
        {
            u_int uIsStatic;
		    xStream >> uIsStatic;
            ePhysicsObjectType = ( uIsStatic ? ASURA_PHYSICS_OBJECT_STATIC : ASURA_PHYSICS_OBJECT_DYNAMIC );
        }

        SetupPhysicsObject( ePhysicsObjectType, bUpdateWaypointBlocking );

		if(uVersion > 2)
		{
            int iCollisionProxy;
			xStream >> iCollisionProxy;
			SetCollisionProxy( static_cast<Asura_Physics_Object::COLLISION_PROXY>(iCollisionProxy) );

            float fGravityScale, fElasticity, fFriction, fDrag, fInvMass;
			xStream >> fGravityScale;
			xStream >> fElasticity;
			xStream >> fFriction;
			xStream >> fDrag;
			xStream >> fInvMass;
			// Fix up silently for static and external control objects, since pretty much all of them will have inv mass of 0 because of the old initialisation process.
			Asura_Assertf((fInvMass > 0.0f) || (ePhysicsObjectType == ASURA_PHYSICS_OBJECT_STATIC) || (ePhysicsObjectType == ASURA_PHYSICS_OBJECT_EXTERNALCONTROL), ("Entity %d has infinite mass. This should no longer be necessary for any type of physics object. Fixing up the value to 1, but this may need changing in tools.", Guid));
			if (fInvMass <= 0.0f) { fInvMass = 1.0f; }

			SetGravityScale( fGravityScale );
			SetElasticity( fElasticity );
			SetFriction( fFriction );
			SetDrag( fDrag );
			SetInvMass( fInvMass );

			if( uVersion > 3 )
			{
                float fAngularDrag;
				xStream >> fAngularDrag;

				SetAngularDrag( fAngularDrag );
            }
		}
	}
	else
	{
		SetupPhysicsObject( ASURA_PHYSICS_OBJECT_STATIC, bUpdateWaypointBlocking );
	}
}


void Asura_ServerEntity_PhysicsObject::FillInChunkBuffer( char*& pcChunkBfr ) 
{
	Asura_Chunk_ChunkBuffer_Stream	xStream(pcChunkBfr);

	// Always use the chunk-stream...
	Asura_ServerEntity_PhysicsObject::WriteToChunkStream(xStream);
}


int Asura_ServerEntity_PhysicsObject::GetChunkBufferSize( void )
{
	Asura_Chunk_Dummy_Stream xStream;
	Asura_ServerEntity_PhysicsObject::WriteToChunkStream(xStream);
    return xStream.GetPosition();
}


void Asura_ServerEntity_PhysicsObject::InitialiseFromChunkBuffer( const char*& pcChunkBfr )
{
	Asura_Chunk_ChunkBuffer_Stream	xStream(pcChunkBfr);

	// Always use the chunk-stream...
	Asura_ServerEntity_PhysicsObject::ReadFromChunkStream(xStream);
}


void Asura_ServerEntity_PhysicsObject::MakeClientUpdateMessage( void )
{
    // TODO: Shouldn't static objects just be using the 'do not always
    // update my client' and burst update functionality rather than manually
    // suppressing updates and then constantly sending and guaranteeing
    // messages when dead? (See also ForceUpdate() in the header which would
    // also not be necessary when RequestClientUpdate() is available.)
    // (MH - 03.01.2007) - removed the bit that sent updates guaranteed when dead.

    // Most static objects wouldn't change state very often
    // so there isn't any need to send update messages every 0.1s!
    // Should only send messages when a player connects or if the object changes state.
    if( IsStatic() )
    {
        bool bSendUpdateMessage = GetHealth() == 0.0f
                                  || Asura_Network::IsExternalNetworkActive()
                                  || !m_bClientUpdated;

        if( !bSendUpdateMessage ) { return; }
    }

    Asura_Message* pxMsg = Asura_Network::CreateMessage( ASURA_MSG_CLIENTUPDATE, sizeof(Asura_MsgData_PhysicsObject_ClientUpdate), GetClassification() );
    if( !pxMsg ) { return; }

    m_bClientUpdated = true;

    Asura_MsgData_PhysicsObject_ClientUpdate* pxMsgData = static_cast<Asura_MsgData_PhysicsObject_ClientUpdate*>( pxMsg->GetData() );
    pxMsgData->FillPhysics( this );
            
    Send( pxMsg, ASURA_GUID_MY_CLIENT_ENTITIES );
}

void Asura_ServerEntity_PhysicsObject::SetupPhysicsObject( ASURA_PHYSICS_OBJECT_TYPE ePhysicsObjectType, bool bUpdateWaypointBlocking )
{
	bool bChanging;
	if (m_pxDynamicObject)
	{
		bChanging = ( ePhysicsObjectType != m_pxDynamicObject->GetType() );
	}
	else
	{
		bChanging = true;
	}

    const bool bIsStatic = ( ASURA_PHYSICS_OBJECT_STATIC == ePhysicsObjectType );

	if (bChanging)
	{
        if( m_pxDynamicObject )
        {
            m_pxDynamicObject->ChangeType( ePhysicsObjectType );
        }
        else
        {
			m_pxDynamicObject = Asura_Physics_Object::Create( ePhysicsObjectType );
        }

		m_pxDynamicObject->SetAssociatedEntityGuid( GetGuid() );

        if( bIsStatic )
        {
			m_uPhysicalObjectFlags |= ASURA_PHYSICALOBJECT_FLAG_STATIC;
			m_bClientUpdated = false;
			SetBlockNavDataGenerationFlag( true );
        }
        else
        {
			m_uPhysicalObjectFlags &= ~ASURA_PHYSICALOBJECT_FLAG_STATIC;
			SetBlockNavDataGenerationFlag( false );
		}
	}

    if( bUpdateWaypointBlocking )
    {
        SetBlockWaypointLinksFlag( bIsStatic );
    }
}

#ifdef ASURA_TOOLS

bool Asura_ServerEntity_PhysicsObject::ShouldIncludeInLevelExport( void ) const
{
    if( IsMergedWithEnvironment()
        && GetClassification() == AsuraEntityClass_PhysicsObject
        && IsStatic() )
    {
        //Pure static objects that have been merged into the environment mesh are redundant
        //as far as the game is concerned. (Well ,hopefully). RAR 10/05/04
        return false;
    }

    return PARENT::ShouldIncludeInLevelExport();
}

#endif // #ifdef ASURA_TOOLS
