
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_MoveData_Animation.cpp
// Author : Andy Tate
// Date   : 7-Sep-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(ASURA_USE_MOVEMENT_SYSTEM)

// This file's header...
#include "Entity/PhysicalObject/Moves/Asura_MoveData_Animation.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/PhysicalObject/Moves/Asura_Moves.h"
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "FacialAnimation/Asura_FacialAnimation_System.h"
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::Asura_MoveData_Animation
// Author       : Andy Tate
// Date         : 6-Sep-2007
////////////////////////////////////////////////////////////////////////////////

Asura_MoveData_Animation::Asura_MoveData_Animation(const Asura_Hash_ID uAnimHash,
												   const u_int uFlags,
												   const u_int uPriority,
												   const Asura_Hash_ID uControllerGroupHash)
: m_uAnimHash(uAnimHash)
, m_uAnimFlags(uFlags)
, m_uAnimPriority(uPriority)
, m_uControllerGroupHash(uControllerGroupHash)
, m_pxAnimation(NULL)
, m_fTweenInDuration(DEFAULT_TWEEN_TIME)
, m_fTweenOutDuration(0.0f)
, m_fParametricStartTime(0.0f)
, m_fParametricEndTime(1.0f)
, m_uFlags(0)
{
	// Initialise the variables...
	m_bWaitForCompletion	= true;
}

/// End of function 'Asura_MoveData_Animation::Asura_MoveData_Animation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::Reset
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_Animation::Reset(void)
{
	PARENT::Reset();

	if(m_pxAnimation != NULL && !m_bLeaveAnimAtEnd)
	{
		m_pxAnimation->DestroyController(m_uControllerAnimationHash);
	}

	m_pxAnimation				= NULL;
	m_uControllerAnimationHash	= ASURA_HASH_ID_UNSET;

	ChangeState(WAITING);
}

/// End of function 'Asura_MoveData_Animation::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::Start
// Author       : Andy Tate
// Date         : 6-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_Animation::Start(const bool bLoopAtEnd, Asura_Animation& xAnimation)
{
	Asura_AnimationController_Group*		pxParentGroup			= xAnimation.GetControllerGroupFromHash(m_uControllerGroupHash);
	Asura_AnimationController_Animation*	pxAnimationController	= xAnimation.GetControllerAnimationFromHash(m_uAnimHash);
	Asura_AnimationBlend*					pxBlend					= (m_bAnimatedTween)
																	? new Asura_AnimationBlend_AnimatedTween (m_fTweenInDuration, m_fTweenOutDuration, false)
																	: new Asura_AnimationBlend_Tween (m_fTweenInDuration, m_fTweenOutDuration, false);
	u_int									uModifiedFlags			= m_uAnimFlags | AAC_FLAG_ACTIVE;

	if (!m_bDontCalculateVelocity)
	{
		uModifiedFlags |= AACA_FLAG_CALCULATE_VELOCITY | AACA_FLAG_CALCULATE_MOVEMENT_OFFSET;
	}
	if (!m_bDontCalculateAngularVelocity)
	{
		uModifiedFlags |= AACA_FLAG_CALCULATE_ANGULAR_VELOCITY | AACA_FLAG_CALCULATE_MOVEMENT_OFFSET;
	}
	
	if(bLoopAtEnd)
	{
		uModifiedFlags |= AACA_FLAG_LOOPED;
	}
	else
	{
		Asura_Assert((uModifiedFlags & AACA_FLAG_LOOPED)==0,"Loop anim flag set, m_bLoopAtEnd not set. Will cause move to end leaving animation playing!");
		uModifiedFlags &= ~AACA_FLAG_LOOPED;
	}

	// Update our animation flags based upon our settings...
	if((m_fTweenOutDuration <= 0.0f) && !bLoopAtEnd && !m_bLeaveAnimAtEnd)
	{
		uModifiedFlags	|= AAC_FLAG_DESTROY_WHEN_FINISHED;
	}
	else
	{
		uModifiedFlags	|= AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE;
	}

	// Ensure that we are able to create a new animation-controller...
	if(pxAnimationController != NULL)
	{
		xAnimation.DestroyController(m_uAnimHash);
	}

	// Create and play the appropriate animation...
	pxAnimationController	= xAnimation.CreateControllerAnimation(m_uAnimHash, uModifiedFlags, m_uAnimPriority, m_uAnimHash, pxBlend, pxParentGroup);

	if(pxAnimationController == NULL || pxAnimationController->GetAnimation() == NULL )
	{
		// Failed to create the animation...
		m_uControllerAnimationHash	= ASURA_HASH_ID_UNSET;
		Asura_Assertf(false, ("Failed to create animation for hash-ID %08X.", m_uAnimHash));
		return(false);
	}
	else
	{
		// Play the animation...
		pxAnimationController->SetParametricStartTime(m_fParametricStartTime);
		pxAnimationController->SetParametricEndTime(m_fParametricEndTime);
		pxAnimationController->Play();

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
		if(m_pxParentMove == NULL)
		{
			Asura_Assert(false, "\"m_pxParentMove\" has not been set when using animation move-data, something's gone wrong here.");
		}
		else
		{
			const Asura_Movement_System&			xMovementSystem				= m_pxParentMove->GetParentSystem();
			const Asura_Guid						uOwnerGuid					= xMovementSystem.GetOwnerGuid();
			Asura_FacialAnimation_AnimationHandler*	pxFacialAnimationHandler	= Asura_FacialAnimation_System::FindAnimationHandler(uOwnerGuid);

			// Play the facial-animation (if there is one)...
			if(pxFacialAnimationHandler != NULL)
			{
				Asura_FacialAnimation_FaceAnimation*	pxFaceAnimation	= Asura_FacialAnimation_System::FindFaceAnimation(m_uAnimHash);

				if(pxFaceAnimation != NULL)
				{
					pxFacialAnimationHandler->PlayFaceAnimation(m_uAnimHash);
				}
			}
		}
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)

		// Ensure that we actually start the timer...
		if(!xAnimation.UpdateTimers(ANIMTIME_USEGAMETIME, false))
		{
			//Failed to update the timers because they have already been done this fame.
			//Force an update with a zero frame time to make sure that the tween data store
			//is up to date with the new animation , but without causing the timers to
			//be advanced a second time.
			//This should hopefully prevent the animation from glitching while also preventing the
			//sounds/flags from being lost for the first frame.
			//RAR 13/11/2007
			xAnimation.UpdateTimers(0.0f, true);
		}


		// Return the hash-ID of the animation-controller...
		m_pxAnimation				= &xAnimation;
		m_uControllerAnimationHash	= pxAnimationController->GetControllerHash();

		ChangeState(STARTING);
		return(true);
	}
}

/// End of function 'Asura_MoveData_Animation::Start'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::Update
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_Animation::Update(const float fTimeStep)
{
	if(m_pxAnimation == NULL)
	{
		Asura_Assert(false, "No animation to use when updating move-data.");
		return(false);
	}
	else
	{
		// Firstly, make sure our animations are up-to-date...
		m_pxAnimation->UpdateTimers(fTimeStep);

		if(m_bWaitForCompletion)
		{
			const Asura_AnimationController_Animation*	pxAnimationController	= m_pxAnimation->GetControllerAnimationFromHash(m_uControllerAnimationHash);

			if(pxAnimationController == NULL)
			{
				Asura_Assert(false, "Cannot find the required animation for this move.");
				return(false);
			}
			else if(pxAnimationController->IsFinished())
			{
				return(false);
			}
			else
			{
				return(true);
			}
		}
		else
		{
			// We don't need to wait for the animation...
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_Animation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::Finish
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_Animation::Finish(void)
{
	ChangeState(FINISHED);
}

/// End of function 'Asura_MoveData_Animation::Finish'. ///

void Asura_MoveData_Animation::StopLoopedAnims()
{
	Asura_AnimationController_Animation*	pxAnimationController	= GetAnimationController();

	// Stop any looped animations for this move...
	if((pxAnimationController != NULL) && ((pxAnimationController->GetControllerFlags() & AACA_FLAG_LOOPED) != 0))
	{
		pxAnimationController->Stop();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::CalculateMovementOffset
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_Animation::CalculateMovementOffset(const float fParametricTimer,
													   Asura_AVector3* pxPosition,
													   Asura_Quat* pxOrientation) const
{
	const Asura_AnimationController_Animation*	pxAnimationController	= GetAnimationController();

	if(pxAnimationController == NULL)
	{
		Asura_Assert(false, "Cannot find specified animation controller for this move.");
		return(false);
	}
	else
	{
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= GetHierarchyAnim();

		if(pxHierarchyAnim == NULL)
		{
			Asura_Assert(false, "Cannot find the required animation for this move.");
			return(false);
		}
		else
		{
			const Asura_Hierarchy_Skin*	pxHierachySkin	= GetHierarchySkin();

			if(pxHierachySkin == NULL)
			{
				Asura_Assert(false, "This physical-object doesn't appear to have a skin.");
				return(false);
			}
			else
			{
				Asura_Vector_3	xPosition;

				if(pxHierarchyAnim->GetMovementOffset(pxHierachySkin, fParametricTimer, &xPosition, pxOrientation, false))
				{
					if(pxPosition != NULL)
					{
						*pxPosition	= xPosition;
					}

					return(true);
				}
				else
				{
					return(false);
				}
			}
		}
	}
}

/// End of function 'Asura_MoveData_Animation::CalculateMovementOffset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::CalculateVelocity
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_Animation::CalculateVelocity(const float fParametricTimer,
												 Asura_AVector3& xVelocity) const
{
	const Asura_AnimationController_Animation*	pxAnimationController	= GetAnimationController();

	if(pxAnimationController == NULL)
	{
		Asura_Assert(false, "Cannot find specified animation controller for this move.");
		return(false);
	}
	else
	{
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= GetHierarchyAnim();

		if(pxHierarchyAnim == NULL)
		{
			Asura_Assert(false, "Cannot find the required animation for this move.");
			return(false);
		}
		else
		{
			const Asura_Hierarchy_Skin*	pxHierachySkin	= GetHierarchySkin();

			if(pxHierachySkin == NULL)
			{
				Asura_Assert(false, "This physical-object doesn't appear to have a skin.");
				return(false);
			}
			else
			{
				Asura_Vector_3	xTempVelocity;

				if(pxHierarchyAnim->GetVelocity(pxHierachySkin, fParametricTimer, GetAnimationDuration(), &xTempVelocity))
				{
					xVelocity	= xTempVelocity;
					return(true);
				}
				else
				{
					return(false);
				}
			}
		}
	}
}

/// End of function 'Asura_MoveData_Animation::CalculateVelocity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::CalculateAngularVelocity
// Author       : Andy Tate
// Date         : 20-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_Animation::CalculateAngularVelocity(const float fParametricTimer,
														Asura_AVector3& xAxis,
														Asura_Angle& fAngle) const
{
	const Asura_AnimationController_Animation*	pxAnimationController	= GetAnimationController();

	if(pxAnimationController == NULL)
	{
		Asura_Assert(false, "Cannot find specified animation controller for this move.");
		return(false);
	}
	else
	{
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= GetHierarchyAnim();

		if(pxHierarchyAnim == NULL)
		{
			Asura_Assert(false, "Cannot find the required animation for this move.");
			return(false);
		}
		else
		{
			const Asura_Hierarchy_Skin*	pxHierachySkin	= GetHierarchySkin();

			if(pxHierachySkin == NULL)
			{
				Asura_Assert(false, "This physical-object doesn't appear to have a skin.");
				return(false);
			}
			else
			{
				Asura_Vector_3	xTempAxis;

				if(pxHierarchyAnim->GetAngularVelocity(fParametricTimer, GetAnimationDuration(), &xTempAxis, &fAngle))
				{
					xAxis	= xTempAxis;
					return(true);
				}
				else
				{
					return(false);
				}
			}
		}
	}
}

/// End of function 'Asura_MoveData_Animation::CalculateAngularVelocity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::GetAnimationController
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

const Asura_AnimationController_Animation* Asura_MoveData_Animation::GetAnimationController(void) const
{
	if(m_pxAnimation == NULL)
	{
		return(NULL);
	}
	else
	{
		return(m_pxAnimation->GetControllerAnimationFromHash(m_uControllerAnimationHash));
	}
}

/// End of function 'Asura_MoveData_Animation::GetAnimationController'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::GetAnimationController
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

Asura_AnimationController_Animation* Asura_MoveData_Animation::GetAnimationController(void)
{
	if(m_pxAnimation == NULL)
	{
		return(NULL);
	}
	else
	{
		return(m_pxAnimation->GetControllerAnimationFromHash(m_uControllerAnimationHash));
	}
}

/// End of function 'Asura_MoveData_Animation::GetAnimationController'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::GetAnimationDuration
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

float Asura_MoveData_Animation::GetAnimationDuration(void) const
{
	const Asura_Hierarchy_Anim*	pxHierarchyAnim	= GetHierarchyAnim();

	if(pxHierarchyAnim == NULL)
	{
		return(0.0f);
	}
	else
	{
		return(pxHierarchyAnim->GetDuration());
	}
}

/// End of function 'Asura_MoveData_Animation::GetAnimationDuration'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::GetHierarchyAnim
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

const Asura_Hierarchy_Anim* Asura_MoveData_Animation::GetHierarchyAnim(void) const
{
	return(Asura_Animation_System::GetAnimByHash(m_uAnimHash));
}

/// End of function 'Asura_MoveData_Animation::GetHierarchyAnim'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::GetHierarchySkin
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

const Asura_Hierarchy_Skin* Asura_MoveData_Animation::GetHierarchySkin(void) const
{
	if(m_pxAnimation == NULL)
	{
		return(NULL);
	}
	else
	{
		return(m_pxAnimation->GetSkin());
	}
}

/// End of function 'Asura_MoveData_Animation::GetHierarchySkin'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::IsLooped
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_Animation::IsLooped(void) const
{
	return((m_uAnimFlags & AACA_FLAG_LOOPED) != 0);
}

/// End of function 'Asura_MoveData_Animation::IsLooped'. ///


void Asura_MoveData_Animation::UpdateTweenData()
{
	Asura_AnimationController_Animation* pxController = GetAnimationController();
	if (!pxController) { return; }

	Asura_AnimationBlend* pxBlend = pxController->GetBlendToPreviousController();
	if (!pxBlend) { return; }
	if (!pxBlend->IsTween()) { return; }

	Asura_AnimationBlend_Tween* pxTween = static_cast<Asura_AnimationBlend_Tween*>(pxBlend);

	pxTween->ChangeTweenDuration(m_fTweenInDuration, m_fTweenOutDuration);
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_Animation::GetParametricTimer
// Author       : Andy Tate
// Date         : 19-Sep-2007
////////////////////////////////////////////////////////////////////////////////

float Asura_MoveData_Animation::GetParametricTimer(void) const
{
	const Asura_AnimationController_Animation*	pxAnimationController	= GetAnimationController();
	if(pxAnimationController)
	{
		const float	fParametricTimer = pxAnimationController->GetParametricTimer();
		return(fParametricTimer);
	}
	else
	{
		return(0.0f);
	}
}

/// End of function 'Asura_MoveData_Animation::GetParametricTimer'. ///


#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)


/// End of file. ///
