
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_MoveData_PositionInterpolation.cpp
// Author : Andy Tate
// Date   : 10-Sep-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(ASURA_USE_MOVEMENT_SYSTEM)

// This file's header...
#include "Entity/PhysicalObject/Moves/Asura_MoveData_PositionInterpolation.h"

// Asura independent header-files...
#include "Debugging/Asura_Assert.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::UpdateDurationBased
// Author       : Andy Tate
// Date         : 19-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::UpdateDurationBased(void)
{
	if(m_fTimer < m_fDuration)
	{
		const float	fParametricTimer	= m_fTimer / m_fDuration;

		// Duration based simply implies interpolation from initial -> final...
		m_xCurrent.Lerp(m_xInitial, m_xFinal, fParametricTimer);
		m_xActual	= m_xCurrent;
		return(true);
	}
	else
	{
		if(m_bSnapToFinal)
		{
			m_xCurrent	= m_xFinal;
			m_xActual	= m_xCurrent;
		}

		return(false);
	}
}

/// End of function 'Asura_MoveData_PositionInterpolation::UpdateDurationBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::UpdateDurationBased
// Author       : Andy Tate
// Date         : 19-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::UpdateDurationBased(const float fParametricTimer,
															   Asura_AVector3_InParam xLSMOCurrentPosition,
															   Asura_AVector3_InParam xLSMOFinalPosition,
															   const Asura_Quat& xWSCurrentOrientation,
															   const Asura_Quat& xWSFinalOrientation)
{
	// We're still updating with movement-offsets - calculate the current values...
	switch(m_eMovementOffsetOperation)
	{
		case IGNORE_MOVEMENT_OFFSET:
		{
			Asura_Assert(false, "Shouldn't be here if we don't have a movement-offset operation.");
			return(false);
		}

		case ADD_MOVEMENT_OFFSET:
		{
			// Simply add the movement-offset onto the interpolated position...
			m_xMovementOffsetFinal = xWSFinalOrientation.RotateVector(xLSMOFinalPosition);
			m_xMovementOffsetFinal	= m_xFinal + m_xMovementOffsetFinal;

			m_xMovementOffsetCurrent.Lerp(m_xInitial, m_xMovementOffsetFinal, fParametricTimer);
			m_xCurrent.Lerp(m_xInitial, m_xFinal, fParametricTimer);

			m_xActual = xWSCurrentOrientation.RotateVector(xLSMOCurrentPosition);
			m_xActual	= m_xCurrent + m_xActual;
			return(true);
		}

		case SUBTRACT_MOVEMENT_OFFSET:
		{
			// Simply subtract the movement-offset onto the interpolated position...
			m_xMovementOffsetFinal = xWSFinalOrientation.RotateVector(xLSMOFinalPosition);
			m_xMovementOffsetFinal	= m_xFinal - m_xMovementOffsetFinal;

			m_xMovementOffsetCurrent.Lerp(m_xInitial, m_xMovementOffsetFinal, fParametricTimer);
			m_xCurrent.Lerp(m_xInitial, m_xFinal, fParametricTimer);

			m_xActual = xWSCurrentOrientation.RotateVector(xLSMOCurrentPosition);
			m_xActual	= m_xMovementOffsetCurrent + m_xActual;
			return(true);
		}

		case STRETCH_MOVEMENT_OFFSET:
		{
			const float	fFinalMagnitude	= xLSMOFinalPosition.Magnitude();

			// Are we able to stretch?
			if(fFinalMagnitude == 0.0f)
			{
				// Simply subtract the movement-offset onto the interpolated position...
				m_xMovementOffsetFinal = xWSFinalOrientation.RotateVector(xLSMOFinalPosition);
				m_xMovementOffsetFinal	= m_xFinal - m_xMovementOffsetFinal;

				m_xMovementOffsetCurrent.Lerp(m_xInitial, m_xMovementOffsetFinal, fParametricTimer);
				m_xCurrent.Lerp(m_xInitial, m_xFinal, fParametricTimer);

				m_xActual = xWSCurrentOrientation.RotateVector(xLSMOCurrentPosition);
				m_xActual	= m_xMovementOffsetCurrent + m_xActual;
				return(true);
			}
			else
			{
				const float	fCurrentMagnitude	= xLSMOCurrentPosition.Magnitude();
				const float	fParametric			= fCurrentMagnitude / fFinalMagnitude;

				// Stretch the movement from the movement-offset over the range of the interpolated position...
				m_xMovementOffsetFinal = xWSFinalOrientation.RotateVector(xLSMOFinalPosition);
				m_xMovementOffsetFinal		= m_xInitial + m_xMovementOffsetFinal;

				m_xMovementOffsetCurrent = xWSCurrentOrientation.RotateVector(xLSMOCurrentPosition);
				m_xMovementOffsetCurrent	= m_xInitial + m_xMovementOffsetCurrent;

				m_xCurrent.Lerp(m_xInitial, m_xFinal, fParametricTimer);
				m_xActual.Lerp(m_xInitial, m_xFinal, fParametric);
				return(true);
			}
		}
		
		default:
		{
			Asura_Assert(false, "Unexpected movement-offset operation when updating position-interpolation.");
			return(false);
		}
	}		// End of switch.
}

/// End of function 'Asura_MoveData_PositionInterpolation::UpdateDurationBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::UpdateDurationBased
// Author       : Andy Tate
// Date         : 19-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::UpdateDurationBased(Asura_AVector3_InParam xLSMOCurrentPosition,
															   Asura_AVector3_InParam xLSMOFinalPosition,
															   const Asura_Quat& xWSCurrentOrientation,
															   const Asura_Quat& xWSFinalOrientation)
{
	if(m_fTimer < m_fDuration)
	{
		const float		fParametricTimer	= m_fTimer / m_fDuration;

		return(UpdateDurationBased(fParametricTimer, xLSMOCurrentPosition, xLSMOFinalPosition, xWSCurrentOrientation, xWSFinalOrientation));
	}
	else
	{
		// We've finished updating...
		if(m_bSnapToFinal)
		{
			UpdateDurationBased(1.0f, xLSMOCurrentPosition, xLSMOFinalPosition, xWSCurrentOrientation, xWSFinalOrientation);
			return(false);
		}
		else
		{
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_PositionInterpolation::UpdateDurationBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::UpdateSpeedBased
// Author       : Andy Tate
// Date         : 19-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::UpdateSpeedBased(const float fTimeStep)
{
	Asura_AVector3	xDirection(m_xFinal - m_xInitial);

	const float		fMagnitude	= xDirection.NormaliseReturningMagnitude();
	const float		fDistance	= m_fSpeed * fTimeStep;

	Asura_AVector3	xDistanceMoved;

	if(fDistance < fMagnitude)
	{
		xDistanceMoved	 = xDirection * fDistance;
		m_xCurrent		+= xDistanceMoved;
		m_xActual		 = m_xCurrent;

		return(true);
	}
	else
	{
		if(m_bSnapToFinal)
		{
			m_xCurrent	= m_xFinal;
		}
		else
		{
			xDistanceMoved	 = xDirection * fMagnitude;
			m_xCurrent		+= xDistanceMoved;
		}

		m_xActual	= m_xCurrent;
		return(false);
	}
}

/// End of function 'Asura_MoveData_PositionInterpolation::UpdateSpeedBased'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::Asura_MoveData_PositionInterpolation
// Author       : Andy Tate
// Date         : 11-Sep-2007
////////////////////////////////////////////////////////////////////////////////

Asura_MoveData_PositionInterpolation::Asura_MoveData_PositionInterpolation(void)
: m_fReachedFinalEpsilon(0.010f)
, m_fTimer(0.0f)
, m_fDuration(0.0f)
, m_fSpeed(0.0f)
, m_uFlags(0)
{
	m_bSnapToFinal	= true;
	m_eVelocityOperation = ADD_VELOCITY;
}

/// End of function 'Asura_MoveData_PositionInterpolation::Asura_MoveData_PositionInterpolation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::Reset
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_PositionInterpolation::Reset(void)
{
	PARENT::Reset();
	ChangeState(WAITING);

	m_xInitial.Zero();
	m_xCurrent.Zero();
	m_xFinal.Zero();

	m_xMovementOffsetCurrent.Zero();
	m_xMovementOffsetFinal.Zero();

	m_xActual.Zero();

	m_fTimer		= 0.0f;
	m_bInitialSet	= false;
	m_bFinalSet		= false;
}

/// End of function 'Asura_MoveData_PositionInterpolation::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::Start
// Author       : Andy Tate
// Date         : 12-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::Start(Asura_AVector3_InParam xInitial)
{
	SetInitial(xInitial);
	ChangeState(STARTING);
	return(true);
}

/// End of function 'Asura_MoveData_PositionInterpolation::Start'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::Update
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::Update(const float fTimeStep)
{
	if(m_bInitialSet)
	{
		if(m_bFinalSet)
		{
			// Increment the timer...
			m_fTimer	+= fTimeStep;

			// Is our move duration-based?
			if(m_bDurationBased)
			{
				return(UpdateDurationBased());
			}
			else if(m_bSpeedBased)
			{
				return(UpdateSpeedBased(fTimeStep));
			}
			else
			{
				Asura_Assert(false, "Position-interpolation not fully initialised.");
				return(false);
			}
		}
		else
		{
			Asura_Assert(false, "Initial value is set but not the final value for move position-interpolation - cannot update.");
			return(false);
		}
	}
	else
	{
		if(m_bFinalSet)
		{
			Asura_Assert(false, "Final value is set but not the initial value for move position-interpolation - cannot update.");
			return(false);
		}
		else
		{
			Asura_Assert(false, "Neither the initial nor final value is set for move position-interpolation - cannot update.");
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_PositionInterpolation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::Update
// Author       : Andy Tate
// Date         : 10-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::Update(const float fTimeStep,
												  Asura_AVector3_InParam xLSMOCurrentPosition,
												  Asura_AVector3_InParam xLSMOFinalPosition,
												  const Asura_Quat& xWSCurrentOrientation,
												  const Asura_Quat& xWSFinalOrientation)
{
	if(m_bInitialSet)
	{
		if(m_bFinalSet)
		{
			// Increment the timer...
			m_fTimer	+= fTimeStep;

			// Is our move duration-based?
			if(m_bDurationBased)
			{
				return(UpdateDurationBased(xLSMOCurrentPosition, xLSMOFinalPosition, xWSCurrentOrientation, xWSFinalOrientation));
			}
			else if(m_bSpeedBased)
			{
				return(UpdateSpeedBased(fTimeStep));
			}
			else
			{
				Asura_Assert(false, "Position-interpolation not fully initialised.");
				return(false);
			}
		}
		else
		{
			Asura_Assert(false, "Initial value is set but not the final value for move position-interpolation - cannot update.");
			return(false);
		}
	}
	else
	{
		if(m_bFinalSet)
		{
			Asura_Assert(false, "Final value is set but not the initial value for move position-interpolation - cannot update.");
			return(false);
		}
		else
		{
			Asura_Assert(false, "Neither the initial nor final value is set for move position-interpolation - cannot update.");
			return(false);
		}
	}
}

/// End of function 'Asura_MoveData_PositionInterpolation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::UpdateUsingVelocity
// Author       : Andy Tate
// Date         : 21-Sep-2007
////////////////////////////////////////////////////////////////////////////////

bool Asura_MoveData_PositionInterpolation::UpdateUsingVelocity(const float fTimeStep,
															   Asura_AVector3_InParam xVelocity)
{
	if(m_bInitialSet)
	{
		// Increment the timer...
		m_fTimer	+= fTimeStep;

		// Is our move duration-based?
		if(m_bDurationBased)
		{
			// Update the position...
			m_xCurrent	+= (xVelocity * fTimeStep);
			m_xActual	 = m_xCurrent;

			return(m_fTimer < m_fDuration);
		}
		else if(m_bSpeedBased)
		{
			Asura_Assert(false, "Speed based interpolation using velocity not yet implemented!!!");
			return(false);
		}
		else
		{
			Asura_Assert(false, "Position-interpolation not fully initialised.");
			return(false);
		}
	}
	else
	{
		Asura_Assert(false, "Neither the initial nor final value is set for move position-interpolation - cannot update.");
		return(false);
	}
}

/// End of function 'Asura_MoveData_PositionInterpolation::UpdateUsingVelocity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_MoveData_PositionInterpolation::Finish
// Author       : Andy Tate
// Date         : 12-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void Asura_MoveData_PositionInterpolation::Finish(void)
{
	ChangeState(FINISHED);
}

/// End of function 'Asura_MoveData_PositionInterpolation::Finish'. ///

#endif	// defined(ASURA_USE_MOVEMENT_SYSTEM)


/// End of file. ///
