//////////////////////////////////////////////////////////////////////////////
//
// Asura_Entity_NetworkFPSPlayer.h
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:	code to manage the movement of player-controlled entities in 
//				networked games. Client->Server updates should be as small as
//				possible, so any information which isn't likely to change in
//				every frame and has no bearing on player movement (e.g. 
//				respawn requests, weapon data) should go in separate messages
//				which preferably are not sent every frame.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_ENTITY_NETWORKFPSPLAYER_H___
#define ___ASURA_ENTITY_NETWORKFPSPLAYER_H___ 1

#include "Core/Asura_BitPacker.h"
#include "Core/Asura_Maths.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "Network/Asura_Player.h"

// C O N S T A N T S /////////////////////////////////////////////////////////
static const float fASURA_PLAYER_MAX_ORIENT_X = 1.4f;	// +/- clamp on looking up/down
static const u_int uASURA_PLAYER_MAX_COMMANDS_IN_UPDATE_PACKAGE = 4;

// T Y P E S   &   E N U M S /////////////////////////////////////////////////
enum Asura_NetworkFPSPlayer_State
{
	ASURA_FPS_PLAYERSTATE_NONE,
	ASURA_FPS_PLAYERSTATE_PLAYING,
	ASURA_FPS_PLAYERSTATE_DEAD,
	ASURA_FPS_PLAYERSTATE_SPECTATING,
	ASURA_FPS_PLAYERSTATE_WAITINGTORESPAWN,

	NUM_ASURA_FPS_PLAYERSTATES
};

enum Asura_NetworkFPSPlayer_Posture
{
	ASURA_FPS_PLAYERPOSTURE_STANDING,
	ASURA_FPS_PLAYERPOSTURE_CROUCHING,
	ASURA_FPS_PLAYERPOSTURE_PRONE,

	NUM_ASURA_FPS_PLAYERPOSTURES
};


// Movement flags. These are things that affect movement e.g. jumping,
// crouching, prone, aiming, sprinting.
enum
{
	ASURA_MOVEMENTFLAG_JUMP			= 0x0001,
	ASURA_MOVEMENTFLAG_CROUCHING	= 0x0002,
	ASURA_MOVEMENTFLAG_PRONE		= 0x0004,
	ASURA_MOVEMENTFLAG_AIMING		= 0x0008,
	ASURA_MOVEMENTFLAG_SPRINTING	= 0x0010,
};


// C L A S S E S /////////////////////////////////////////////////////////////

// message containing movement data, also includes methods for packing using Asura_BitPacker
class Asura_Msg_PlayerMovementCommands
{
	u_char m_ucFrameTime;		// time to run dynamics over (make 7 bits)
	signed char m_cWalkAxis;	// (8 bits)
	signed char m_cStrafeAxis;	// (8 bits)
	signed char m_cOrientX;		// (8 bits)
	signed short m_sOrientY;	// (16 bits)
	u_short m_usCommands;		// (5 bits)

private:

	enum
	{
		NumBits_FrameTime = 8,	// 0-255 ms
		NumBits_WalkAxis = 8,
		NumBits_StrafeAxis = 8,
		NumBits_OrientX = 8,
		NumBits_OrientY = 16,
		NumBits_Commands = 5,
	};

public:

	Asura_Msg_PlayerMovementCommands()
	{
		m_ucFrameTime = 0;
		m_cWalkAxis = 0;
		m_cStrafeAxis = 0;
		m_cOrientX = 0;	
		m_sOrientY = 0;	
		m_usCommands = 0;
	}

	float GetFrameTime() const 
	{
		float fValue = static_cast<float>(m_ucFrameTime);
		fValue /= 1000.0f;
		return fValue;  
	}
	float GetWalkAxis() const 
	{ 
		float fWalkAxis = static_cast<float>(m_cWalkAxis);
		fWalkAxis /= 127.0f;
		return fWalkAxis;
	}
	float GetStrafeAxis() const 
	{ 
		float fStrafeAxis = static_cast<float>(m_cStrafeAxis);
		fStrafeAxis /= 127.0f;
		return fStrafeAxis;
	}
	float GetOrientX() const 
	{ 
		float fOrientX = static_cast<float>(m_cOrientX);
		fOrientX /= 127.0f;
		fOrientX *= fASURA_PLAYER_MAX_ORIENT_X;
		return fOrientX;
	}
	float GetOrientY() const
	{ 
		float fOrientY = static_cast<float>(m_sOrientY);
		fOrientY /= 32767.0f;
		fOrientY *= Asura_Maths::Deg180;
		return fOrientY;
	}
	u_int GetFlags() const { return m_usCommands;	 }

	void SetFrameTime(float fFrameTime) 
	{
		float fFrameTimeMilliSecs = fFrameTime * 1000.0f;
		Asura_Assert(fFrameTimeMilliSecs >= 0.0f, "negative frame time");
		Asura_Assertf(fFrameTimeMilliSecs < static_cast<float>(1 << NumBits_FrameTime), 
			("frame time too long for packing, will clamp to %ums", (1 << NumBits_FrameTime) - 1));
		if (fFrameTimeMilliSecs >= static_cast<float>(1 << NumBits_FrameTime))
		{
			// clamp to top of range
			fFrameTimeMilliSecs = static_cast<float>(1 << NumBits_FrameTime) - 1.0f;
		}
		m_ucFrameTime = static_cast<u_char>(Asura_Maths::FloatToInt_Chop(fFrameTimeMilliSecs));
	}
	void SetWalkAxis(float fWalkAxis)
	{
		Asura_Assert(fWalkAxis >= -1.0f && fWalkAxis <= 1.0f, "walk axis too big");
		m_cWalkAxis	= static_cast<char>(Asura_Maths::FloatToInt_Chop(fWalkAxis * 127.0f));
	}
	void SetStrafeAxis(float fStrafeAxis)
	{
		Asura_Assert(fStrafeAxis >= -1.0f && fStrafeAxis <= 1.0f, "strafe axis too big");
		m_cStrafeAxis = static_cast<char>(Asura_Maths::FloatToInt_Chop(fStrafeAxis * 127.0f));
	}
	void SetOrientX(float fOrientX)
	{
		Asura_Assert(fOrientX >= -fASURA_PLAYER_MAX_ORIENT_X && fOrientX <= fASURA_PLAYER_MAX_ORIENT_X, "orient x out of range");
		m_cOrientX = static_cast<char>(Asura_Maths::FloatToInt_Chop((fOrientX / fASURA_PLAYER_MAX_ORIENT_X) * 127.0f));
	}
	void SetOrientY(float fOrientY) 
	{
		Asura_Assert(fOrientY >= -Asura_Maths::Deg180 && fOrientY <= Asura_Maths::Deg180, "orient Y not in range -180,180");
		m_sOrientY = static_cast<short>(Asura_Maths::FloatToInt_Chop((fOrientY / Asura_Maths::Deg180) * 32767.0f));
	}
	void SetFlags(u_int uFlags) { m_usCommands = static_cast<u_short>(uFlags); }


	static u_int GetNumBytesRequired(u_int uNumCommandSets) 
	{ 
		return (((NumBits_FrameTime + NumBits_WalkAxis + NumBits_StrafeAxis + NumBits_OrientX + NumBits_OrientY + NumBits_Commands) * uNumCommandSets) + 7) / 8; 
	}
	
	void Pack(Asura_BitPacker& xBitPacker) const
	{
		// frame time
		xBitPacker.Pack(m_ucFrameTime, NumBits_FrameTime);
		xBitPacker.Pack(m_cWalkAxis + 128, NumBits_WalkAxis);
		xBitPacker.Pack(m_cStrafeAxis + 128, NumBits_StrafeAxis);
		xBitPacker.Pack(m_cOrientX + 128, NumBits_OrientX);
		xBitPacker.Pack(m_sOrientY + 32768, NumBits_OrientY);
		xBitPacker.Pack(m_usCommands, NumBits_Commands);
	}

	void Unpack(Asura_BitUnpacker& xUnpacker)
	{
		// frame time
		m_ucFrameTime = static_cast<u_char>(xUnpacker.Unpack(NumBits_FrameTime));
		m_cWalkAxis = static_cast<char>(xUnpacker.Unpack(NumBits_WalkAxis) - 128);
		m_cStrafeAxis = static_cast<char>(xUnpacker.Unpack(NumBits_StrafeAxis) - 128);
		m_cOrientX = static_cast<char>(xUnpacker.Unpack(NumBits_OrientX) - 128);
		m_sOrientY = static_cast<short>(xUnpacker.Unpack(NumBits_OrientY) - 32768);
		m_usCommands = static_cast<u_short>(xUnpacker.Unpack(NumBits_Commands));
	}
};

class Asura_PlayerUpdatePackage
{
	Asura_PlayerUpdatePackage(const Asura_PlayerUpdatePackage&);	// declare copy constructor
	Asura_PlayerUpdatePackage& operator=(const Asura_PlayerUpdatePackage&);	// declare assignment operator
public:

	Asura_PlayerUpdatePackage();
	~Asura_PlayerUpdatePackage();

	float								m_fCreationTime;
	Asura_Vector_3						m_xPositionWhenSent;
	u_short								m_usNumCommands;
	u_short								m_usUpdateID;
	bool								m_bHasBeenSent;
	Asura_Msg_PlayerMovementCommands*	m_pxCommands;		// allocated array of uASURA_PLAYER_MAX_COMMANDS_IN_UPDATE_PACKAGE
};


class Asura_Msg_MultiplePlayerCommands_Header
{
public:

	// 0 returns most recent, 1 returns previous one, 2 returns two back
	void GetCommands(Asura_Msg_PlayerMovementCommands* pxCommands, u_int uFromMostRecent) const
	{
		const u_char* pucPtr = reinterpret_cast<const u_char*>(this+1);
		u_int uNumCommands = m_uNumCommands0;
		if (uFromMostRecent > 0)
		{
			pucPtr += Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(uNumCommands);
			uNumCommands = m_uNumCommandsPrev1;
			if (uFromMostRecent > 1)
			{
				pucPtr += Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(uNumCommands);
				uNumCommands = m_uNumCommandsPrev2;
			}
		}

		Asura_BitUnpacker xUnpacker(pucPtr, Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(uNumCommands));
		for (u_int u = 0; u < uNumCommands; ++u)
		{
			pxCommands[u].Unpack(xUnpacker);
		}
	}

	static u_char* PackCommands(const Asura_PlayerUpdatePackage* pxPackage, u_char* pucDest)
	{
		u_int uNumBytes = Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(pxPackage->m_usNumCommands);
		Asura_BitPacker xBitPacker(uNumBytes);

		const Asura_Msg_PlayerMovementCommands* pxCommands = pxPackage->m_pxCommands; 
		for (u_int u = 0; u < pxPackage->m_usNumCommands; ++u)
		{
			pxCommands->Pack(xBitPacker);
			pxCommands++;
		}
		const u_char* pucResult;
		u_int uLen;
		xBitPacker.GetResult(&pucResult, &uLen);
		Asura_Assert(uLen == uNumBytes, "bit packer packed wrong amount");
		memcpy(pucDest, pucResult, uNumBytes);
		return pucDest + uNumBytes;
	}

	static u_int GetSizeRequired(const Asura_PlayerUpdatePackage* pxPackage0, 
								const Asura_PlayerUpdatePackage* pxPackagePrev1, 
								const Asura_PlayerUpdatePackage* pxPackagePrev2)
	{
		return sizeof(Asura_Msg_MultiplePlayerCommands_Header) + 
			Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(pxPackage0->m_usNumCommands) + 
			Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(pxPackagePrev1->m_usNumCommands) + 
			Asura_Msg_PlayerMovementCommands::GetNumBytesRequired(pxPackagePrev2->m_usNumCommands);
	}

	static u_int CreateMessageHeaderAndData(void* const pData, const Asura_PlayerUpdatePackage* pxPackage0, 
							const Asura_PlayerUpdatePackage* pxPackagePrev1, 
							const Asura_PlayerUpdatePackage* pxPackagePrev2)
	{
		Asura_Msg_MultiplePlayerCommands_Header* pxHeader = reinterpret_cast
			<Asura_Msg_MultiplePlayerCommands_Header*>(pData);
		pxHeader->m_xPredictedPosition = pxPackage0->m_xPositionWhenSent;
		pxHeader->m_usUpdateID = pxPackage0->m_usUpdateID;
		pxHeader->m_uNumCommands0 = pxPackage0->m_usNumCommands;
		pxHeader->m_uNumCommandsPrev1 = pxPackagePrev1->m_usNumCommands;
		pxHeader->m_uNumCommandsPrev2 = pxPackagePrev2->m_usNumCommands;
		
		u_char* pucData = reinterpret_cast<u_char*>(pxHeader+1);
		pucData = PackCommands(pxPackage0, pucData);
		pucData = PackCommands(pxPackagePrev1, pucData);
		pucData = PackCommands(pxPackagePrev2, pucData);

		// return num bytes written
		return pucData - reinterpret_cast<u_char*>(pData);
	}
	
	Asura_Vector_3 m_xPredictedPosition;
	u_short m_usUpdateID;
	u_short m_uNumCommands0 : 5,
			m_uNumCommandsPrev1 : 5,
			m_uNumCommandsPrev2 : 5;
};


class Asura_Entity_NetworkFPSPlayer_Client : public Asura_Entity_Player
{

	public:
	
		Asura_Entity_NetworkFPSPlayer_Client();
		virtual ~Asura_Entity_NetworkFPSPlayer_Client();

		virtual void Initialise	( void );
		virtual void Shutdown	( void );
		virtual void Reset		( void );

		virtual void Teleport( const class Asura_Vector_3* pxPosition = NULL, const class Asura_Matrix_3x3* pxOrientation = NULL);

		void HandleFirstUpdate(const float fOrientY, const float fOrientX)
		{
			m_bReceivedFirstUpdate = true;
			m_fOrientY = fOrientY;
			m_fOrientX = fOrientX;
		}

		bool HasReceivedFirstUpdate( void ) const { return m_bReceivedFirstUpdate; }

		
		// IsDead() and controller input value functions must be supplied by derived class.
		virtual bool IsDead( void ) const = 0;
		virtual void GetWalkAxes( float& fWalk, float& fStrafe ) const = 0;
		virtual void GetLookAxes( float& fHoriz, float& fVert ) const = 0;
		virtual bool IsYAxisInverted( void ) const = 0;
		virtual bool IsJumpRequested( void ) const = 0;

	protected:

		// pointless pure virtual function
		void MakeCommandMessageForServer(void) {}
		// the function that does the work
		void MakeCommandsMessageForServer( bool bUsingMouse );

		void HandleMessage_PositionValidated(float fSentTime);
		void HandleMessage_PositionCorrected(const class Asura_MsgData_PlayerPositionCorrection* pxData, float fSentTime);

		float GetOrientX() const { return m_fOrientX; }
		float GetOrientY() const { return m_fOrientY; }
		virtual void CalculateViewDirection(float&	fDeltaVert, float&	fDeltaHoriz, 
											float	fAxisVert,	float	fAxisHoriz, 
											bool bUsingMouse) = 0;

		// protected data

		// posture data - set by client update function
		Asura_NetworkFPSPlayer_Posture m_ePosture;
		float m_fClampY;


	private:
	
		// private methods
		void BuildCommandPacket( bool bUsingMouse );
		bool ShouldUseMultipleCommands( void ) const;
		void UpdateTimers( void );
		bool IsTimeToActOnCommandPacket( void ) const;
		bool IsTimeToSendCommandPacket( void ) const;
		
		void ActOnCommandPacket( void );
		void SendSingleCommandPacket( void );
		void SendMultipleCommandPacket( void );

		void SetPostureFromCommands(const Asura_Msg_PlayerMovementCommands* pxCommandData);
	
		// private data
		bool m_bReceivedFirstUpdate;
		float m_fOrientX;
		float m_fOrientY;
		
		// data to manage command packets sent to server
		Asura_Msg_PlayerMovementCommands m_xCurrentCommands;
		float m_fTimeToNextCommandUpdate;
		float m_fGameTimeAtLastCommandUpdate;
		float m_fTimeToNextCommandSend;
		u_int m_uCurrentUpdatePackage;
		class Asura_PlayerUpdatePackage* m_pxUpdatePackageStorage;
		float m_fLastServerPositionConfirmationTimestamp;
		float m_fTimeOfLastPositionConfirmation;

		// data to do dynamics on client
		class Asura_Dynamic_Object* m_pxDynamicObject;
//		float m_fLinVelocity;
//		float m_fLinStrafeVelocity;
//		Asura_Vector_3 m_xLinearTargetVelocity;
//		Asura_Vector_3 m_xLastVelocity;
};



// Asura_Entity_NetworkFPSPlayer_Server is a base class which should be used in a multiple
// inheritance setup. It provides the server-side code for moving players around based on the 
// commands received from the player client entity.
// It requires a couple of project functions to deal with posture changes:-
// CanChangePosture() and ChangePosture() - these are pure virtual functions.
class Asura_Entity_NetworkFPSPlayer_Server : public virtual Asura_ServerEntity_PhysicalObject
{
protected:
	Asura_Entity_NetworkFPSPlayer_Server();
	virtual ~Asura_Entity_NetworkFPSPlayer_Server();

	void HandleMessage_MultipleCommands(const void* pxData, float fSentTime);
	void ProcessCommands(const Asura_Msg_PlayerMovementCommands* pxCommands);

	void ResetInternetPlay();
	
	bool IsInternetPlay() const { return m_pxPosPredictor != NULL; }
	void InternetPlay_ResetPosition(const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri);

	void GetInternetPlayPosition( Asura_Vector_3& xPos ) const;
	Asura_NetworkFPSPlayer_Posture GetInternetPlayPosture() const { return m_eFPSPlayerPosture; }

	// send update to clients every update cycle until commands have been received, and then 
	// only when the position has actually changed. Also send at least every 5 seconds.
	bool ShouldSendActorUpdate( void ) const;

	// call ActorUpdated when the client update package is sent out.
	void ActorUpdated( void );

	virtual bool CanChangePosture(Asura_NetworkFPSPlayer_Posture eDesiredPosture) const = 0;
	virtual void ChangePosture(Asura_NetworkFPSPlayer_Posture eNewPosture) = 0;

private:

	// posture update
	void SetPostureFromCommands(const Asura_Msg_PlayerMovementCommands* pxCommandData);

	// posture data
	Asura_NetworkFPSPlayer_Posture m_eFPSPlayerPosture;

	// data to process client command packets
	float	m_fLastUpdatePacketTime;
	u_short m_usLastUpdatePacketID;
	// Client will not send any commands until actor has been created. So keep 
	// sending those actor updates until a command is received, then can start
	// sending them only when commands are received.
	bool			m_bReceivedAnyCommands;
	mutable bool	m_bNeedActorUpdate;
	u_int			m_uTimeOfLastActorUpdate;
	class Asura_PosAndOriInterpolator* m_pxPosPredictor;
};


// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_ENTITY_NETWORKFPSPLAYER_H___
