//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_Projector.cpp
//
// Copyright (c) 2007 Rebellion (Derby). All rights reserved.
//
// Description:
//
// An entity to project a texture onto the environment
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_TARGET_SM3

// This file's header.
#include "Entity/Projector/Asura_ServerEntity_Projector.h"
#include "Entity/Projector/Asura_ClientEntity_Projector.h"

// Engine independent.
#include "Network/Asura_Message.h"
#include "Entity/Asura_Entity_Classification.h"

#include "Camera/Asura_Camera.h"

// asura tools
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif


// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int PROJECTOR_CHUNK_VERSION = 8;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#ifdef ASURA_TOOLS
Asura_ServerEntity_Projector* Asura_ServerEntity_Projector::CreateFromTools( Asura_Guid uGuid , const Asura_Vector_3& xCentre)
{
	Asura_ServerEntity_Projector* pxNew = new Asura_ServerEntity_Projector(uGuid);
	pxNew->m_xParams.m_xOrientation = *Asura_Camera::GetOrientation();
	pxNew->m_xParams.m_xPosition = xCentre;
	return pxNew;
}
#endif

Asura_ServerEntity_Projector::Asura_ServerEntity_Projector( Asura_Guid uGuid)
: PARENT( uGuid )
, m_xParams()
{
	Classification = AsuraEntityClass_Projector;
	SetNeverUpdateClient();
}

void Asura_ServerEntity_Projector::CreateFromChunkStream( Asura_Guid uGuid, Asura_Chunk_Stream &xStream )
{
	Asura_ServerEntity_Projector* pxNew = new Asura_ServerEntity_Projector( uGuid);
	pxNew->ReadFromChunkStream(xStream);
}

void Asura_ServerEntity_Projector::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;

	if (uVersion >= 7)
	{
		PARENT::ReadFromChunkStream(xStream);
	}

	ExtractParamsFromChunkStream(m_xParams, xStream, uVersion );
}

void Asura_ServerEntity_Projector::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << PROJECTOR_CHUNK_VERSION;

	PARENT::WriteToChunkStream(xStream);

	WriteParamsToChunkStream(m_xParams, xStream);
}

Asura_ServerEntity_Projector::ProjectorParams::ProjectorParams()
: m_xPosition(Asura_Maths::ZeroVector3)
, m_xOrientation(Asura_Maths::IdentityMatrix3x3)
, m_uTextureHash(ASURA_HASH_ID_UNSET)
, m_uFalloffTextureHash(ASURA_HASH_ID_UNSET)
, m_uDetailTextureHash(ASURA_HASH_ID_UNSET)
, m_fDetailTextureScale(4.0f)
, m_fMinRange(0.0f)
, m_fMaxRange(25.0f)
, m_iBlendMode(0)
, m_fFoV(Asura_Maths::Deg90 / 2.0f)
, m_fScale(1.0f)
, m_fAspect(1.0f)
, m_fFadeoutStart(100.0f)
, m_fFadeoutDist(10.0f)
, m_xRotSpeed(Asura_Maths::ZeroEuler3)
, m_xSwingDist(Asura_Maths::ZeroEuler3)
, m_fRotationOffset(0.0f)
, m_fIntensity(1.0f)
, m_uMasterEntity(ASURA_GUID_UNREGISTERED)
, m_xOffsetFromMaster(Asura_Maths::ZeroVector3)
, m_uMasterMarker(ASURA_HASH_ID_UNSET)
, m_uProjectorParams_Flags(0)
{
	m_bShowFrustum = true;
	m_bUseMasterOrient = true;

	for (u_int u = 0; u < 3; ++u)
	{
		m_afRGB[u] = 1.0f;
	}
}

void Asura_ServerEntity_Projector::ExtractParamsFromChunkStream( ProjectorParams &xParams , Asura_Chunk_Stream &xStream, u_int uVersion )
{
	Asura_Assert( uVersion <= PROJECTOR_CHUNK_VERSION, "Unhandled Asura_ServerEntity_Projector chunk version" );
	if (uVersion >= 7)
	{
		// Pos/Orient
		xStream >> xParams.m_xPosition;
		xStream >> xParams.m_xOrientation;

		// Texture-related
		xStream >> xParams.m_uTextureHash;
		xStream >> xParams.m_uFalloffTextureHash;
		xStream >> xParams.m_uDetailTextureHash;
		xStream >> xParams.m_fDetailTextureScale;
		xStream >> xParams.m_fMinRange;
		xStream >> xParams.m_fMaxRange;
		xStream >> xParams.m_iBlendMode;

		xStream >> xParams.m_fIntensity;

		// FoV + Type
		xStream >> xParams.m_fFoV;

		xStream >> xParams.m_fScale;
		xStream >> xParams.m_fAspect;

		// Fadeout Distance
		xStream >> xParams.m_fFadeoutStart;
		xStream >> xParams.m_fFadeoutDist;

		xStream >> xParams.m_xRotSpeed;
		xStream >> xParams.m_xSwingDist;

		xStream >> xParams.m_fRotationOffset;

		for (u_int u = 0; u < 3; ++u)
		{
			xStream >> xParams.m_afRGB[u];
		}
		xStream >> xParams.m_uMasterEntity;

		xStream >> xParams.m_uMasterMarker;
		xStream >> xParams.m_xOffsetFromMaster;

		// WGP : Nasty hack to peek the chunk data for version 7 and see if it looks like a parent guid 
		// This is because when version 7 got bumped to 8 I didn't spot the chunk version number in the header file with the same name as the one at the top of this file, so I only bumped the one in this file, but the other one is what got used.
		// there were only 10 chunked flags at version 7 so I'm going to say any number between 0x000003FF and 0xFF030000 (for endianness) must be a GUID and it's very unlikely for your combination of flags to happen to equal 999 )
		if ( uVersion == 7 && ( xStream.PeekUInt() == ASURA_GUID_UNREGISTERED || ( xStream.PeekUInt() > 0x3FF && xStream.PeekUInt() < 0xFF030000 ) ) )
		{
			Asura_Guid uDummyParentLightGuid;
			xStream >> uDummyParentLightGuid;
		}

		Asura_ReadFromStream_Bitfield_Wrapper<u_int>	xBitfield(xParams.m_uProjectorParams_Flags);

		// Write the data...
		xStream >> xBitfield;

#ifdef ASURA_TOOLS
		xStream >> xParams.m_xTextureName;
		xStream >> xParams.m_xFalloffTextureName;
		xStream >> xParams.m_xDetailTextureName;
#else // #ifdef ASURA_TOOLS
		Asura_String xDummy;
		xStream >> xDummy;
		xStream >> xDummy;
		xStream >> xDummy;
#endif // #ifdef ASURA_TOOLS

		return;
	}
	//NY - older version chunked flags out seperately
	xParams.m_uProjectorParams_Flags = 0;

	int iTemp;
	float fTemp;

	// Pos/Orient
	xStream >> xParams.m_xPosition;
	xStream >> xParams.m_xOrientation;

	// Texture-related
	xStream >> xParams.m_uTextureHash;
	xStream >> xParams.m_uFalloffTextureHash;

	if (uVersion>1)
	{
		xStream >> xParams.m_uDetailTextureHash;
		xStream >> xParams.m_fDetailTextureScale;
	}

	xStream >> xParams.m_fMinRange;
	xStream >> xParams.m_fMaxRange;
	xStream >> xParams.m_iBlendMode;

	xStream >> iTemp;
	xParams.m_bFakeLight = (iTemp!=0);

	if (uVersion>3)
	{
		xStream >> iTemp;
		xParams.m_bDynamicLight = (iTemp!=0);
		xStream >> xParams.m_fIntensity;
	}
	else
	{
		xParams.m_bDynamicLight = false;
		xParams.m_fIntensity = 1.0f;
	}

	// FoV + Type
	xStream >> fTemp;
	xParams.m_fFoV = Asura_Maths::DegToRad(fTemp);

	xStream >> iTemp;
	xParams.m_bParallel = (iTemp!=0);

	xStream >> xParams.m_fScale;
	xStream >> xParams.m_fAspect;

	// Fadeout Distance
	xStream >> xParams.m_fFadeoutStart;
	xStream >> xParams.m_fFadeoutDist;

	// Movement Params
	xStream >> iTemp;
	xParams.m_bMoving = (iTemp!=0);

	xStream >> iTemp;
	xParams.m_bSwingX = (iTemp!=0);
	xStream >> fTemp;
	xParams.m_xRotSpeed.x = Asura_Maths::DegToRad(fTemp);
	xStream >> fTemp;
	xParams.m_xSwingDist.x = Asura_Maths::DegToRad(fTemp);
	xStream >> iTemp;
	xParams.m_bSwingY = (iTemp!=0);
	xStream >> fTemp;
	xParams.m_xRotSpeed.y = Asura_Maths::DegToRad(fTemp);
	xStream >> fTemp;
	xParams.m_xSwingDist.y = Asura_Maths::DegToRad(fTemp);
	xStream >> iTemp;
	xParams.m_bSwingZ = (iTemp!=0);
	xStream >> fTemp;
	xParams.m_xRotSpeed.z = Asura_Maths::DegToRad(fTemp);
	xStream >> fTemp;
	xParams.m_xSwingDist.z = Asura_Maths::DegToRad(fTemp);

	xStream >> iTemp;
	xParams.m_bShowFrustum = (iTemp!=0);

	if (uVersion>2)
	{
		xStream >> xParams.m_fRotationOffset;
		xStream >> iTemp;
		xParams.m_bQuadIntensity = (iTemp!=0);
	}
	else
	{
		xParams.m_fRotationOffset = 0.0f;
		xParams.m_bQuadIntensity  = false;
	}

	// New since attachements were added
	if (uVersion>4)
	{
		for (u_int u = 0; u < 3; ++u)
		{
			xStream >> xParams.m_afRGB[u];
		}
		xStream >> xParams.m_uMasterEntity;
		bool bTemp;

		xStream >> bTemp;
		xParams.m_bUseMasterOrient = bTemp;

		if (uVersion>5)
		{
			u_char ucPad;
			for (u_int u = 0; u < 3; ++u)
			{
				xStream >> ucPad;
			}
		}

		xStream >> xParams.m_uMasterMarker;
		xStream >> xParams.m_xOffsetFromMaster;
		{
			Asura_Guid uDummyParentLightGuid;
			xStream >> uDummyParentLightGuid;
		}
	}
	else
	{
		xParams.m_afRGB[0] = 1.0f;
		xParams.m_afRGB[1] = 1.0f;
		xParams.m_afRGB[2] = 1.0f;
		xParams.m_uMasterEntity		= ASURA_GUID_UNREGISTERED;
		xParams.m_bUseMasterOrient  = true;
		xParams.m_uMasterMarker		= ASURA_HASH_ID_UNSET;
		xParams.m_xOffsetFromMaster = Asura_Maths::ZeroVector3;
	}

	// Now the texture names
#ifdef ASURA_TOOLS
	xStream >> xParams.m_xTextureName;
	xStream >> xParams.m_xFalloffTextureName;

	if (uVersion>1)
	{
		xStream >> xParams.m_xDetailTextureName;
	}
#else // #ifdef ASURA_TOOLS
	Asura_String xDummy;
	xStream >> xDummy;
	xStream >> xDummy;
	xStream >> xDummy;
#endif // #ifdef ASURA_TOOLS
}



Asura_ServerEntity_Projector::~Asura_ServerEntity_Projector( void )
{
}

void Asura_ServerEntity_Projector::WriteParamsToChunkStream( const ProjectorParams &xParams , Asura_Chunk_Stream& xStream)
{
	// Pos/Orient
	xStream << xParams.m_xPosition;
	xStream << xParams.m_xOrientation;

	// Texture-related
	xStream << xParams.m_uTextureHash;
	xStream << xParams.m_uFalloffTextureHash;
	xStream << xParams.m_uDetailTextureHash;
	xStream << xParams.m_fDetailTextureScale;
	xStream << xParams.m_fMinRange;
	xStream << xParams.m_fMaxRange;
	xStream << xParams.m_iBlendMode;

	xStream << xParams.m_fIntensity;

	// FoV + Type
	xStream << xParams.m_fFoV;

	xStream << xParams.m_fScale;
	xStream << xParams.m_fAspect;

	// Fadeout Distance
	xStream << xParams.m_fFadeoutStart;
	xStream << xParams.m_fFadeoutDist;

	xStream << xParams.m_xRotSpeed;
	xStream << xParams.m_xSwingDist;

	xStream << xParams.m_fRotationOffset;

	for (u_int u = 0; u < 3; ++u)
	{
		xStream << xParams.m_afRGB[u];
	}
	xStream << xParams.m_uMasterEntity;

	xStream << xParams.m_uMasterMarker;
	xStream << xParams.m_xOffsetFromMaster;

	Asura_WriteToStream_Bitfield_Wrapper<u_int>	xBitfield(xParams.m_uProjectorParams_Flags);

	// Generate the bitfield...
	xBitfield.Add(xParams.m_bFakeLight, 1);
	xBitfield.Add(xParams.m_bDynamicLight, 1);
	xBitfield.Add(xParams.m_bParallel, 1);
	xBitfield.Add(xParams.m_bMoving, 1);
	xBitfield.Add(xParams.m_bSwingX, 1);
	xBitfield.Add(xParams.m_bSwingY, 1);
	xBitfield.Add(xParams.m_bSwingZ, 1);
	xBitfield.Add(xParams.m_bShowFrustum, 1);
	xBitfield.Add(xParams.m_bQuadIntensity, 1);
	xBitfield.Add(xParams.m_bUseMasterOrient, 1);

	// Write the data...
	xStream << xBitfield;

#ifdef ASURA_TOOLS
	xStream << xParams.m_xTextureName;
	xStream << xParams.m_xFalloffTextureName;
	xStream << xParams.m_xDetailTextureName;
#else // #ifdef ASURA_TOOLS
	Asura_String xDummy;
	xStream << xDummy;
	xStream << xDummy;
	xStream << xDummy;
#endif // #ifdef ASURA_TOOLS
}


#ifdef ASURA_TOOLS
void Asura_ServerEntity_Projector::SetParams( const ProjectorParams& xNewParams )
{
    m_xParams = xNewParams;

	Asura_ClientEntity_Projector *pxClient = static_cast<Asura_ClientEntity_Projector*>(Asura_ClientEntity_Projector::FindEntity(GetGuid()));
	if (pxClient)
	{
		pxClient->SetParams( m_xParams );


		//NY - inelegant but effective way of keeping the omnipresent client entity up to date...
		if (IsEntityActive())
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_ACTIVATE );
			if( pxMessage )
			{
				pxMessage->UseGuaranteedDelivery();
				Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
			}
		}
		else
		{
			Asura_Message* pxMessage = Asura_Network::CreateMessage( ASURA_MSG_DEACTIVATE );
			if( pxMessage )
			{
				pxMessage->UseGuaranteedDelivery();
				Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
			}
		}
	}
}
#endif

#ifdef ASURA_TOOLS
void Asura_ServerEntity_Projector::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	PARENT::ApplyChangesForTemplateInstance(pxInstance);

	m_xParams.m_xPosition    = pxInstance->TransformPosition(m_xParams.m_xPosition);
	m_xParams.m_xOrientation = pxInstance->TransformMatrix(m_xParams.m_xOrientation);

	m_xParams.m_uMasterEntity = pxInstance->ConvertGuid(m_xParams.m_uMasterEntity);
	m_xParams.m_xOffsetFromMaster = m_xParams.m_xOffsetFromMaster * pxInstance->GetRelativeMatrix();
}
#endif // ASURA_TOOLS


#endif //#ifdef ASURA_TARGET_SM3

