//
// Asura_ClientEntity_Activatable.cpp
//
// Created by Richard Rice - 10/11/2004
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//


#include "Core/Asura.h"
#include "Entity/Triggers/Asura_ClientEntity_Activatable.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Network/Asura_Message.h"


////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Asura_ClientEntity_Activatable **
**--------------------------------*/

#define ASURA_CLIENTENTITY_ACTIVATABLE_VERSION_NUMBER 3


Asura_ClientEntity_Activatable::Asura_ClientEntity_Activatable(Asura_Guid uGuid,const char*& pcChunkBuffer)
:Asura_Entity_Client(uGuid)
{
	//check the version
	int iVersion;
	ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion,int);
	Asura_Assert(iVersion <= ASURA_CLIENTENTITY_ACTIVATABLE_VERSION_NUMBER,"");

	//skip Asura_Entity_Server data
	if(iVersion >= 3)
	{
		Asura_Entity_Server::SkipDataInChunkBuffer(pcChunkBuffer);
	}

	unsigned int uActive;
	ASURA_CHUNK_EXTRACT(pcChunkBuffer, uActive,unsigned int);

    m_bActive = uActive ? 1 : 0;
}

Asura_ClientEntity_Activatable::Asura_ClientEntity_Activatable(Asura_Guid uGuid)
:Asura_Entity_Client(uGuid)
{
	m_bActive = false;
}

void Asura_ClientEntity_Activatable::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	//check the version
	int iVersion;
	xStream >> iVersion;
	Asura_Assert(iVersion <= ASURA_CLIENTENTITY_ACTIVATABLE_VERSION_NUMBER,"");

	//skip Asura_Entity_Server data
	if(iVersion >= 3)
	{
		Asura_Entity_Server::SkipDataInChunkStream(xStream);
	}

	unsigned int uActive;
	xStream >> uActive;

    m_bActive = uActive ? 1 : 0;
}


bool Asura_ClientEntity_Activatable::HandleMessage(const Asura_Message* pxMessage)
{
    if( IsDestroyed() ) { return true; } // ignore all further messages?

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			default:
			{
				Asura_Assert(0,"Unhandled message!");
				break;
			}

			case ASURA_MSG_DESTRUCT:
			{
				OnDestruction();
				m_bActive = 0;
				return true;
			}
			case ASURA_MSG_ACTIVATE:
			{
				if (!m_bActive)
				{
					OnActivation();
					m_bActive = 1;

				}
				return true;
			}
			case ASURA_MSG_DEACTIVATE:
			{
				if (m_bActive)
				{
					OnDeactivation();
					m_bActive = 0;
				}
				return true;
			}
			case ASURA_MSG_TRIGGER:
			{
				if (m_bActive)
				{
					OnTrigger();
				}
				return true;
			}
			case ASURA_MSG_UNTRIGGER:
			{
				if (m_bActive)
				{
					OnUnTrigger();
				}
				return true;
			}
		}
	}

	return false;
}


void Asura_ClientEntity_Activatable::OnDestruction()
{
	SetAsDestroyed();
}



#define ASURA_SERVERENTITY_ACTIVATABLEWITHPOSITION_VERSION_NUMBER 0


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_ActivatableWithPosition::Asura_ClientEntity_ActivatableWithPosition
// Author       : Andy Tate
// Date         : 31-Aug-2005
//
// Description  : 
//
// Parameters   : 
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_ClientEntity_ActivatableWithPosition::Asura_ClientEntity_ActivatableWithPosition (Asura_Guid uGuid) :
	Asura_ClientEntity_Activatable(uGuid)
{
	// Set to default values...

	m_xPosition.Zero		();
	m_xOrientation.Identity	();
}

/// End of function 'Asura_ClientEntity_ActivatableWithPosition::Asura_ClientEntity_ActivatableWithPosition'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_ActivatableWithPosition::Asura_ClientEntity_ActivatableWithPosition
// Author       : Andy Tate
// Date         : 31-Aug-2005
//
// Description  : 
//
// Parameters   : 
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_ClientEntity_ActivatableWithPosition::Asura_ClientEntity_ActivatableWithPosition (Asura_Guid uGuid,
																						const char*& pcChunkBuffer) :
	Asura_ClientEntity_Activatable(uGuid, pcChunkBuffer)
{
	//check the version
	int iVersion;
	ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion,int);
	Asura_Assert(iVersion <= ASURA_SERVERENTITY_ACTIVATABLEWITHPOSITION_VERSION_NUMBER, "FATAL ERROR: unknown version number for Asura_ServerEntity_ActivatableWithPosition");

	ASURA_CHUNK_EXTRACT(pcChunkBuffer,m_xPosition,Asura_Vector_3);
	ASURA_CHUNK_EXTRACT(pcChunkBuffer,m_xOrientation,Asura_Matrix_3x3);
}

/// End of function 'Asura_ClientEntity_ActivatableWithPosition::Asura_ClientEntity_ActivatableWithPosition'. ///

////////////////////////////////////////////////////////////////////////////////
//
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_ActivatableWithPosition::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	Asura_ClientEntity_Activatable::ReadFromChunkStream(xStream);

	// Read the data...
	int iVersion;
	xStream >> iVersion;
	Asura_Assertf(iVersion <= ASURA_SERVERENTITY_ACTIVATABLEWITHPOSITION_VERSION_NUMBER,("Unknown chunk version %d for Asura_ClientEntity_ActivatableWithPosition",iVersion));

	xStream >> m_xPosition;
	xStream >> m_xOrientation;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ClientEntity_ActivatableWithPosition::SetPositionAndOrientation
// Author       : Andy Tate
// Date         : 31-Aug-2005
//
// Description  : 
//
// Parameters   : References to the position and orientation to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_ClientEntity_ActivatableWithPosition::SetPositionAndOrientation (const Asura_Vector_3& xPosition,
																			const Asura_Matrix_3x3& xOrientation)
{
	// Simply force the position and orientation...

	m_xPosition		= xPosition;
	m_xOrientation	= xOrientation;
}

/// End of function 'Asura_ClientEntity_ActivatableWithPosition::SetPositionAndOrientation'. ///


/// End of file. ///
