//////////////////////////////////////////////////////////////////////////////
//
// Asura_ClientEntity_SoundController.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     entities which control phonons
//
// History:
//
//     (MH  - 15.11.2002) - Created.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "Entity/Triggers/Asura_ClientEntity_SoundController.h"

#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Triggers/Asura_ServerEntity_SoundController.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Network/Asura_Message.h"
#include "Sound/Asura_Phonon.h"
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)

#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

Asura_Hashed_List<Asura_ClientEntity_SoundController> Asura_ClientEntity_SoundController::s_xList( 4 );

Asura_ClientEntity_SoundController::Asura_ClientEntity_SoundController(Asura_Guid uGuid,Asura_Guid uPhononGuid, bool bInitiallyActive /*= true*/)
:Asura_Entity_Client(uGuid)
{
	m_uPhononGuid = uPhononGuid;
	Classification = AsuraEntityClass_SoundController;

	// initialise the phonon
	
	
    if (bInitiallyActive)
    {
        ProcessActivationMessage();
    }
    else
    {
        ProcessDeactivationMessage();
    }

	s_xList.AddToList( this );
}

Asura_ClientEntity_SoundController::Asura_ClientEntity_SoundController(Asura_Chunk_Stream& xStream)
:Asura_Entity_Client()
{
    // read the server entities base class data
    bool bInitiallyActive;
    Asura_ServerEntity_SoundController::ParseChunkStream(xStream, &bInitiallyActive);

    Classification = AsuraEntityClass_SoundController;
    
	//check the version
	int iVersion;
	xStream >> iVersion;
	Asura_Assert(iVersion <= ASURA_SERVERENTITY_SOUNDCONTROLLER_VERSION_NUMBER, "unknown sound controller version number");

	//read in the rest
	xStream >> m_uPhononGuid;
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
	m_uPhononGuid	= Asura_Streaming_System::MakeCurrentMasterZoneGuid(m_uPhononGuid);
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)

	Asura_Vector_3 xPosition;
	if(iVersion >= 1)
	{
		xStream >> xPosition;
	}
    
    // initialise the phonon
	#ifdef ASURA_TOOLS
	//Need to prevent phonon initialisation if currently updating from a template , because the
	//phonon guid won't be correct yet.RAR 27/04/2004
	if(!Tools_EntityTemplate::IsUpdatingFromTemplate())
	#endif
	{
		if(iVersion >= 1)
		{
			//Restore the saved position.
			//In practice we don't really want to do this during the initial load in tools , but it shouldn't matter
			//since the phonon position will get reset back to its initial position anyway
			//RAR 12/11/2007
			SetPosition(xPosition);
		}

		if (bInitiallyActive)
		{
		    ProcessActivationMessage();
		}
		else
		{
		    ProcessDeactivationMessage();
		}
	}

	s_xList.AddToList( this );
}

Asura_ClientEntity_SoundController::~Asura_ClientEntity_SoundController( void )
{
	s_xList.RemoveFromList( this );
}

bool Asura_ClientEntity_SoundController::HandleMessage(const Asura_Message* pxMessage)
{
    switch (pxMessage->GetMsgID())
    {
    case ASURA_MSG_ACTIVATE:
        ProcessActivationMessage();
        return true;

    case ASURA_MSG_DEACTIVATE:
        ProcessDeactivationMessage();
        return true;

    case ASURA_MSG_TRIGGER:
        ProcessTriggerMessage();
        return true;

    default:
        return Asura_Entity_Client::HandleMessage(pxMessage);
    }
}


void Asura_ClientEntity_SoundController::ProcessActivationMessage()
{
    Asura_Phonon* pxPhonon = Asura_Phonon_System::GetPhonon(m_uPhononGuid);
	#ifndef ASURA_TOOLS
	//Assert changed to retail only.
	//In tools the phonon will often be missing on load as a result of files being out of sync , but
	//this will get resolved at various key points so it should be fine for exports.
	//RAR 03/04/2009
    Asura_Assert(pxPhonon, "Failed to find phonon.\nThis ought to have been resolved at the start of level export.");
	#endif
    if (pxPhonon)
    {
        pxPhonon->Activate();
    }
}

void Asura_ClientEntity_SoundController::ProcessDeactivationMessage()
{
    Asura_Phonon* pxPhonon = Asura_Phonon_System::GetPhonon(m_uPhononGuid);
	#ifndef ASURA_TOOLS
	//See comment in ProcessActivationMessage()
    Asura_Assert(pxPhonon, "Failed to find phonon.\nThis ought to have been resolved at the start of level export.");
	#endif
    if (pxPhonon)
    {
        pxPhonon->Deactivate();
    }
}

void Asura_ClientEntity_SoundController::ProcessTriggerMessage()
{
    Asura_Phonon* pxPhonon = Asura_Phonon_System::GetPhonon(m_uPhononGuid);
	#ifndef ASURA_TOOLS
	//See comment in ProcessActivationMessage()
    Asura_Assert(pxPhonon, "Failed to find phonon.\nThis ought to have been resolved at the start of level export.");
	#endif
    if (pxPhonon)
    {
        pxPhonon->Trigger();
    }
}


void Asura_ClientEntity_SoundController::Activate(bool bActive)
{
	if(bActive)
		ProcessActivationMessage();
	else
		ProcessDeactivationMessage();
}

void Asura_ClientEntity_SoundController::SetPosition(const Asura_Vector_3& xPos)
{
    Asura_Phonon* pxPhonon = Asura_Phonon_System::GetPhonon(m_uPhononGuid);
    if (pxPhonon)
    {
        pxPhonon->SetPosition(xPos);
    }

}


#ifdef ASURA_TOOLS
void Asura_ClientEntity_SoundController::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	//update the sound guid
	m_uPhononGuid = pxInstance->ConvertSoundGuid(m_uPhononGuid);

}
#endif // ASURA_TOOLS
