//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_CountedTrigger.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     trigger that sends its trigger messages only when it has been triggered
//     a certain number of times.
//
// History:
//
//     (MH  - 13.11.2002) - Created.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Triggers/Asura_ServerEntity_CountedTrigger.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_StaticMessage_System.h"
#ifdef ASURA_SUPPORT_ENTITYWATCH
#include "Entity/Asura_EntityWatch_System.h"
#endif
#include "Network/Asura_Message.h"

#ifdef ASURA_TOOLS
#include "Entity/Tools_Message_Editing.h"
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif


//////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Asura_CountedTrigger_MessageBlock **
**-----------------------------------*/

void Asura_CountedTrigger_MessageBlock::SetDefaultValues()
{
	m_uHashID = 0;
	m_iTriggerCount = 1;
	m_usMessageBlock = ASURA_STATICMESSAGE_EMPTYBLOCK;
	m_usFlags = 0;
	m_bRetriggerable = 1;
	m_bActive = 1;

	#ifdef ASURA_TOOLS
	m_xName = "";
	#endif
}

//////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Asura_ServerEntity_CountedTrigger **
**-----------------------------------*/

static const int ASURA_SERVERENTITY_COUNTEDTRIGGER_VERSION_NUMBER = 2;

#ifdef ASURA_TOOLS
Asura_Hashed_List<Asura_ServerEntity_CountedTrigger> Asura_ServerEntity_CountedTrigger::s_xList(5);
#endif

Asura_ServerEntity_CountedTrigger::Asura_ServerEntity_CountedTrigger(bool bInitiallyActive)
:Asura_ServerEntity_Trigger(bInitiallyActive)
{
	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif
    
    Classification = AsuraEntityClass_CountedTrigger;
	m_iCounter = 0;
	m_uNumberOfMessageBlocks = 0;
	m_pxMessageBlocks = 0;
}
                                                       

Asura_ServerEntity_CountedTrigger::Asura_ServerEntity_CountedTrigger(Asura_Guid uGuid)
:Asura_ServerEntity_Trigger(uGuid)
{
	#ifdef ASURA_TOOLS
	s_xList.AddToList(this);
	#endif
    
    Classification = AsuraEntityClass_CountedTrigger;
	m_iCounter = 0;
	m_uNumberOfMessageBlocks = 0;
	m_pxMessageBlocks = 0;
}

Asura_ServerEntity_CountedTrigger::~Asura_ServerEntity_CountedTrigger()
{
	#ifdef ASURA_TOOLS
	s_xList.RemoveFromList(this);
	#endif
	delete [] m_pxMessageBlocks;
}


bool Asura_ServerEntity_CountedTrigger::HandleMessage(const Asura_Message* pxMessage)
{
    if( IsDestroyed() ) { return true; } // ignore all further messages?

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			case ASURA_MSG_RESET:
			{
				m_iCounter = 0;
				return true;
			}

			case ASURA_MSG_COUNTEDTRIGGER_ENABLEMESSAGE :
			case ASURA_MSG_COUNTEDTRIGGER_DISABLEMESSAGE :
			{
				const Asura_Message_CountedTrigger_MessageHash* pxMessageData = reinterpret_cast
					<const Asura_Message_CountedTrigger_MessageHash*>(pxMessage->GetData());

				for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
				{
					if(m_pxMessageBlocks[i].m_uHashID == pxMessageData->m_uMessageHash)
					{
						m_pxMessageBlocks[i].m_bActive = (pxMessage->GetMsgID()==ASURA_MSG_COUNTEDTRIGGER_ENABLEMESSAGE);

						break;
					}
				}

				return true;
			}

			case ASURA_MSG_COUNTEDTRIGGER_SETTRIGGERCOUNT :
			{
				const Asura_Message_CountedTrigger_SetTriggerCount* pxMessageData = reinterpret_cast
					<const Asura_Message_CountedTrigger_SetTriggerCount*>(pxMessage->GetData());

				//search for a matching message
				for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
				{
					if(m_pxMessageBlocks[i].m_uHashID == pxMessageData->m_uMessageHash)
					{
						SetTriggerCount(i,pxMessageData->m_iTriggerCount);
						break;
					}
				}
				

				return true;
			}

			case ASURA_MSG_COUNTEDTRIGGER_SETCOUNTER :
			{
				const Asura_Message_CountedTrigger_SetCounter* pxMessageData = reinterpret_cast
					<const Asura_Message_CountedTrigger_SetCounter*>(pxMessage->GetData());

				//Set the new counter
				int iOldCount = m_iCounter;
				m_iCounter = pxMessageData->m_iCounter;

				//trigger messages between old and new count
				TriggerMessages(iOldCount,m_iCounter);
				return true;
				
			}

			case ASURA_MSG_COUNTEDTRIGGER_ADDCOUNTER :
			{
				const Asura_Message_CountedTrigger_AddCounter* pxMessageData = reinterpret_cast
					<const Asura_Message_CountedTrigger_AddCounter*>(pxMessage->GetData());

				//Set the new counter
				int iOldCount = m_iCounter;
				m_iCounter += pxMessageData->m_iChange;
				if(m_iCounter<0) m_iCounter = 0;

				//trigger messages between old and new count
				TriggerMessages(iOldCount,m_iCounter);
				return true;
				
			}
		}
	}

	return Asura_ServerEntity_Trigger::HandleMessage(pxMessage);
}



void Asura_ServerEntity_CountedTrigger::OnTrigger()
{
	m_iCounter++;
	TriggerMessages(m_iCounter-1,m_iCounter);
}

void Asura_ServerEntity_CountedTrigger::OnUnTrigger()
{
	m_iCounter--;
	if(m_iCounter<0) m_iCounter = 0;
}

void Asura_ServerEntity_CountedTrigger::TriggerMessages(int iOldCounter,int iNewCounter)
{
	if(!IsActive()) return;
	if(iNewCounter <= iOldCounter) return;

	//trigger any message blocks in between the old and new counts
	Asura_CountedTrigger_MessageBlock* pxMessage = m_pxMessageBlocks;
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++,pxMessage++)
	{
		if(!pxMessage->m_bActive) continue;
		if(iOldCounter >= pxMessage->m_iTriggerCount) continue;

		if(iNewCounter >= pxMessage->m_iTriggerCount)
		{
			if(!pxMessage->m_bRetriggerable) pxMessage->m_bActive = 0;
			if(pxMessage->m_usMessageBlock != ASURA_STATICMESSAGE_EMPTYBLOCK)
			{
	  			Asura_StaticMessage_System::SendMessages(pxMessage->m_usMessageBlock,Guid);
			}
		}
	}
}

void Asura_ServerEntity_CountedTrigger::SetTriggerCount(u_int uOutput,int iTriggerCount)
{
	Asura_Assert(uOutput < m_uNumberOfMessageBlocks,"");
	Asura_CountedTrigger_MessageBlock* pxMessage = &m_pxMessageBlocks[uOutput];

	//set the new trigger time
	int iOldTriggerCount = pxMessage->m_iTriggerCount;
	pxMessage->m_iTriggerCount = iTriggerCount;

	if(IsActive() && pxMessage->m_bActive)
	{
		//trigger the messages if the trigger count has been brought before the current counter
		if(iTriggerCount <= m_iCounter && m_iCounter < iOldTriggerCount)
		{

			if(!pxMessage->m_bRetriggerable) pxMessage->m_bActive = 0;
	  		if(pxMessage->m_usMessageBlock != ASURA_STATICMESSAGE_EMPTYBLOCK)
	  		{
	  			Asura_StaticMessage_System::SendMessages(pxMessage->m_usMessageBlock,Guid);
	  		}
		}
	}
}



//virtual function overides for saving
int Asura_ServerEntity_CountedTrigger::GetChunkBufferSize()
{
#if 0
	int iChunkSize = Asura_ServerEntity_Trigger::GetChunkBufferSize();
    iChunkSize += sizeof(int)// version
    			+ sizeof(int)// counter
    			+ sizeof(u_int)//num message blocks
				+ sizeof(u_int)*m_uNumberOfMessageBlocks*3//hash,counter , flags and message block
				+ sizeof(int); //tools data size

	#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())//exported levels won't want the tools data
	{
		iChunkSize += GetToolsDataSize();
	}
	#endif

    return iChunkSize;
#endif

	Asura_Assert(false, "CHUNK ERROR: calling deprecated function");
	return 0;
}
#ifdef ASURA_TOOLS
int Asura_ServerEntity_CountedTrigger::GetToolsDataSize()
{
#if 0
	int iChunkSize = 0;
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		iChunkSize += m_pxMessageBlocks[i].m_xName.GetSizeInInts()*4;
	}
	return iChunkSize;
#endif 
	Asura_Assert(false, "CHUNK ERROR: calling deprecated function");
	return 0;
}
#endif //ASURA_TOOLS


void Asura_ServerEntity_CountedTrigger::FillInChunkBuffer(char*& pcChunkBuffer)
{
#if 0
	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/

	Asura_ServerEntity_Trigger::FillInChunkBuffer(pcChunkBuffer);

	int iVersion = ASURA_SERVERENTITY_COUNTEDTRIGGER_VERSION_NUMBER;
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iVersion, int);

    ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_iCounter, int);
    ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_uNumberOfMessageBlocks, u_int);

	for(unsigned i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_pxMessageBlocks[i].m_uHashID, Asura_Hash_ID);
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_pxMessageBlocks[i].m_iTriggerCount, int);
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_pxMessageBlocks[i].m_usMessageBlock, unsigned short);
		ASURA_CHUNK_BITFIELD_FILLANDVERIFY( pcChunkBuffer, m_pxMessageBlocks[i].m_usFlags, unsigned short);
	}

	#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())//exported levels won't want the tools data
	{
		//save out data only needed by tools
		int iToolsDataSize = GetToolsDataSize();

		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, iToolsDataSize,int);

		for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
		{
			ASURA_CHUNK_FILL_STRING(pcChunkBuffer,m_pxMessageBlocks[i].m_xName);
		}
	}
	else
	#endif
	{
		//Either not in tools , or doing a level export.
		//In either case , we don't want the tools data.
		int iToolsDataSize = 0;
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, iToolsDataSize,int);
	}
#endif // 0
	Asura_Assert(false, "CHUNK ERROR: calling deprecated function");
}

#ifdef ASURA_SUPPORT_ENTITYWATCH
void Asura_ServerEntity_CountedTrigger::DisplayEntityWatchText()
{
	Asura_ServerEntity_Trigger::DisplayEntityWatchText();

	Asura_EntityWatch_System::PrintText(16,"Counter : %d",m_iCounter);
	Asura_EntityWatch_System::NextLine();

	Asura_CountedTrigger_MessageBlock* pxMessage = m_pxMessageBlocks;
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++,pxMessage++)
	{
		Asura_EntityWatch_System::PrintText(0,"Msg %d (%d) %s ",i,pxMessage->m_iTriggerCount,pxMessage->m_bActive ? "On " : "Off");
	}
}
#endif

void Asura_ServerEntity_CountedTrigger::CreateFromChunkStream(Asura_Guid uGuid, Asura_Chunk_Stream& xStream)
{
	Asura_ServerEntity_CountedTrigger* pxEntity = new Asura_ServerEntity_CountedTrigger(uGuid);
	if (pxEntity)
	{
		pxEntity->ReadFromChunkStream(xStream);
	}
}

#ifdef ASURA_TOOLS
void Asura_ServerEntity_CountedTrigger::GetCustomEventName(int iEventIndex, Asura_String& xEventName)
{
	xEventName = this->m_pxMessageBlocks[iEventIndex].m_xName;
}
#endif // ASURA_TOOLS

void Asura_ServerEntity_CountedTrigger::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	// Unfortunately, triggers don't use the best-practice approach of: our version -> parent write -> our data
	// This is because they've been converted from chunkbuffer constructors.
	// Instead it's: parent write -> our version and data.
	PARENT::ReadFromChunkStream(xStream);

	//check the version
	int iVersion;
	xStream >> iVersion;
	Asura_Assert(iVersion <= ASURA_SERVERENTITY_COUNTEDTRIGGER_VERSION_NUMBER, "unknown counted trigger version number");

	//read in the rest
	if(iVersion < 2)
	{
		//convert from old trigger to version with multiple message blocks RAR 19/09/2006
		//Create a single default message block
		m_uNumberOfMessageBlocks = 1;
		m_pxMessageBlocks = new Asura_CountedTrigger_MessageBlock[1];
		m_pxMessageBlocks[0].SetDefaultValues();
#ifdef ASURA_TOOLS
		m_pxMessageBlocks[0].m_xName = "Trigger 1";
#endif

		//copy from 'OnTrigger' message block
		m_pxMessageBlocks[0].m_usMessageBlock = MessageBlock[ASURA_TRIGGER_ENTITY_ONBEINGTRIGGERED];
		MessageBlock[ASURA_TRIGGER_ENTITY_ONBEINGTRIGGERED] = ASURA_STATICMESSAGE_EMPTYBLOCK;


		xStream >> m_pxMessageBlocks[0].m_iTriggerCount;
		int iRemainingInputs = m_pxMessageBlocks[0].m_iTriggerCount;
		if(iVersion>0)
		{
			xStream >> iRemainingInputs;
		}
		m_iCounter = m_pxMessageBlocks[0].m_iTriggerCount - iRemainingInputs;
	}
	else
	{
		xStream >> m_iCounter;
		xStream >> m_uNumberOfMessageBlocks;

		m_pxMessageBlocks = m_uNumberOfMessageBlocks ? new Asura_CountedTrigger_MessageBlock[m_uNumberOfMessageBlocks] : 0;
		for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
		{
			xStream >> m_pxMessageBlocks[i].m_uHashID;
			xStream >> m_pxMessageBlocks[i].m_iTriggerCount;
			xStream >> m_pxMessageBlocks[i].m_usMessageBlock;
			xStream >> m_pxMessageBlocks[i].m_usFlags;
		}

		int iToolsDataSize;
		xStream >> iToolsDataSize;

#ifdef ASURA_TOOLS
		if(iToolsDataSize)
		{
			for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				xStream >> m_pxMessageBlocks[i].m_xName;
			}
		}
		else
		{
			for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
			{
				m_pxMessageBlocks[i].m_xName = "";
			}

		}
#else
		//not in tools , so just skip this
		xStream.Skip(iToolsDataSize);
#endif // ASURA_TOOLS	

	}

	// Prevent us from ever updating...
	SetNeverUpdates(true);
}

void Asura_ServerEntity_CountedTrigger::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	/*
	This entity type can be used in template files , so any changes to WriteToChunkStream may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/

	// Unfortunately, triggers don't use the best-practice approach of: our version -> parent write -> our data
	// This is because they've been converted from chunkbuffer constructors.
	// Instead it's: parent write -> our version and data.
	PARENT::WriteToChunkStream(xStream);

	int iVersion = ASURA_SERVERENTITY_COUNTEDTRIGGER_VERSION_NUMBER;
	xStream << iVersion;

	xStream << m_iCounter;
	xStream << m_uNumberOfMessageBlocks;

	for(unsigned i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		xStream << m_pxMessageBlocks[i].m_uHashID;
		xStream << m_pxMessageBlocks[i].m_iTriggerCount;
		xStream << m_pxMessageBlocks[i].m_usMessageBlock;
		Asura_WriteToStream_Bitfield_Wrapper<u_short> xBitfieldWrapper(m_pxMessageBlocks[i].m_usFlags);
		xBitfieldWrapper.Add(m_pxMessageBlocks[i].m_bActive, 1);
		xBitfieldWrapper.Add(m_pxMessageBlocks[i].m_bRetriggerable, 1);
		xStream << xBitfieldWrapper;
	}

#ifdef ASURA_TOOLS
	if(!Tools_Level_Export::IsExport())//exported levels won't want the tools data
	{
		//save out data only needed by tools
		int iToolsDataSize = 0;
		for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
		{
			iToolsDataSize += m_pxMessageBlocks[i].m_xName.GetSizeInInts() * 4;
		}
		xStream << iToolsDataSize;

		for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
		{
			xStream << m_pxMessageBlocks[i].m_xName;
		}
	}
	else
#endif
	{
		//Either not in tools , or doing a level export.
		//In either case , we don't want the tools data.
		int iToolsDataSize = 0;
		xStream << iToolsDataSize;
	}
}


#ifdef ASURA_TOOLS
void Asura_ServerEntity_CountedTrigger::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	Asura_ServerEntity_Trigger::ApplyChangesForTemplateInstance(pxInstance);

	//update the message blocks
	for(u_int i=0;i<m_uNumberOfMessageBlocks;i++)
	{
		m_pxMessageBlocks[i].m_usMessageBlock = pxInstance->ConvertMessageBlock(m_pxMessageBlocks[i].m_usMessageBlock);
	}

}
#endif // ASURA_TOOLS


#ifdef ASURA_TOOLS
const char* Asura_ServerEntity_CountedTrigger::GetMessageDescription(const Asura_StaticMessage* pxMessage)
{
	static char szText[100];
	if(pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_ENABLEMESSAGE ||
		pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_DISABLEMESSAGE ||
		pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_SETTRIGGERCOUNT)
	{
		const Asura_Message_CountedTrigger_MessageHash* pxData = (const Asura_Message_CountedTrigger_MessageHash*) pxMessage->m_pData;


		//Find the entity being triggered
		Asura_ServerEntity_CountedTrigger* pxEntity = FindEntity(Tools_Message_Editing::GetFinalDestinationGuid(pxMessage));
		if(!pxEntity) return 0;

		for(u_int i=0;i<pxEntity->m_uNumberOfMessageBlocks;i++)
		{
			if(pxData->m_uMessageHash == pxEntity->m_pxMessageBlocks[i].m_uHashID)
			{
				if(pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_ENABLEMESSAGE)
					Asura_PrintfToBfr(szText,"Enable Msg(%s)",100,true,pxEntity->m_pxMessageBlocks[i].m_xName.GetString());
				else if(pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_DISABLEMESSAGE)
					Asura_PrintfToBfr(szText,"Disable Msg(%s)",100,true,pxEntity->m_pxMessageBlocks[i].m_xName.GetString());
				else
					Asura_PrintfToBfr(szText,"Set Msg Trigger Count(%s)",100,true,pxEntity->m_pxMessageBlocks[i].m_xName.GetString());
				return szText;
			}
		}
	}
	else if(pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_SETCOUNTER)
	{
		const Asura_Message_CountedTrigger_SetCounter* pxData = (const Asura_Message_CountedTrigger_SetCounter*) pxMessage->m_pData;
		Asura_PrintfToBfr(szText,"Set Counter (%d)",sizeof(szText),true,pxData->m_iCounter);
		return szText;
	}
	else if(pxMessage->m_usMsgID == ASURA_MSG_COUNTEDTRIGGER_ADDCOUNTER)
	{
		const Asura_Message_CountedTrigger_AddCounter* pxData = (const Asura_Message_CountedTrigger_AddCounter*) pxMessage->m_pData;
		Asura_PrintfToBfr(szText,"Add Counter (%+d)",sizeof(szText),true, pxData->m_iChange);
		return szText;
	}
	return 0;
}
#endif //ASURA_TOOLS
