//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_LookAtTrigger.cpp
//
// Created by Anthony Gregson - 22/04/04
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
// Description:
//
//     Trigger which is triggered by being looked at in a set way.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Triggers/Asura_ServerEntity_LookAtTrigger.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_Actor.h"
#ifdef ASURA_SUPPORT_ENTITYWATCH
#include "Entity/Asura_EntityWatch_System.h"
#endif
#include "Network/Asura_Message.h"
#include "Network/Asura_NetMsg.h"

#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

const int iASURA_LOOKAT_TRIGGER_CHUNK_VERSION = 0;

Asura_ServerEntity_LookAtTrigger::Asura_ServerEntity_LookAtTrigger(bool bActive/* = false*/)
: Asura_ServerEntity_Trigger(bActive)
, m_uLookAtFlags(0)
{
	Classification = AsuraEntityClass_LookAtTrigger;

	m_fMaxDistance = 10.0f;
	m_fMinDistance = 0.0f;
	m_bUseScreenArea = false;
	m_fCosMaxAngle = 0.866f; // Approximately Cos(Pi/6)
	m_fCosMinAngle = 1.0f; // Cos(0)
	m_fScreenEdgeDeadzone = 0.0f; // Allows the whole screen
	m_fScreenCentreDeadzone = 0.0f; // Allows the whole screen
	m_bUseLOS = false;
	m_bTriggerOnNotLooking = false;
	m_fTime = 0.0f;
	m_bResetTimeOnEarlyLookAway = true;
	m_fInwardConeCosMaxAngle = 1.0f;
	m_bAdjustForFOV = true;
	m_uTriggeringTypes = ASURA_LOOKATTRIGGER_TYPE_ALLPLAYERS;
	m_iSpecifiedPlayer = 0;
	m_iNumberRequired = 1;
	m_xPosition.Zero();
	m_xDirection = Asura_Vector_3(0.0f, 0.0f, 1.0f);

	m_fTimeRemaining = m_fTime;
	m_iNumberCurrentlyTriggering = 0;
	m_fUpdateTimer = Asura_Random::GetSeededFloat(0.0f, fLOOKAT_TRIGGER_UPDATE_TIME);
	m_iClientNumberCurrentlyTriggering = 0;
	m_bTriggered = false;
}

Asura_ServerEntity_LookAtTrigger::Asura_ServerEntity_LookAtTrigger(Asura_Guid uGuid)
: Asura_ServerEntity_Trigger(uGuid)
, m_fMaxDistance(10.0f)
, m_fMinDistance(0.0f)
, m_fCosMaxAngle(0.866f) // Approximately Cos(Pi/6)
, m_fCosMinAngle(1.0f) // Cos(0)
, m_fScreenEdgeDeadzone(0.0f) // Allows the whole screen
, m_fScreenCentreDeadzone(0.0f) // Allows the whole screen
, m_fTime(0.0f)
, m_fInwardConeCosMaxAngle(1.0f) // Cos(0)
, m_uTriggeringTypes(ASURA_LOOKATTRIGGER_TYPE_ALLPLAYERS)
, m_iSpecifiedPlayer(0)
, m_iNumberRequired(1)
, m_xPosition(Asura_Maths::ZeroVector3)
, m_xDirection(0.0f, 0.0f, 1.0f)
, m_uLookAtFlags(0)
, m_fTimeRemaining(0.0f)
, m_iNumberCurrentlyTriggering(0)
, m_iClientNumberCurrentlyTriggering(0)
{
	Classification = AsuraEntityClass_LookAtTrigger;

	// Some flags default to true.
	m_bResetTimeOnEarlyLookAway = true;
	m_bAdjustForFOV = true;

	m_fUpdateTimer = Asura_Random::GetSeededFloat(0.0f, fLOOKAT_TRIGGER_UPDATE_TIME);
}

Asura_ServerEntity_LookAtTrigger::Asura_ServerEntity_LookAtTrigger(Asura_Guid uGuid, const char*& pcChunkBuffer)
: Asura_ServerEntity_Trigger(uGuid)
, m_uLookAtFlags(0)
{
	Asura_Assert(false, "Deprecated method of construction - Look-at Trigger");

	Classification = AsuraEntityClass_LookAtTrigger;

	m_iNumberCurrentlyTriggering = 0;
	m_fUpdateTimer = Asura_Random::GetSeededFloat(0.0f, fLOOKAT_TRIGGER_UPDATE_TIME);
	m_iClientNumberCurrentlyTriggering = 0;

	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);

	ReadFromChunkStream(xStream);
}

void Asura_ServerEntity_LookAtTrigger::CreateFromChunkStream(Asura_Guid uGuid, Asura_Chunk_Stream& xStream)
{
	Asura_ServerEntity_LookAtTrigger* pxEntity = new Asura_ServerEntity_LookAtTrigger(uGuid);
	if (pxEntity)
	{
		pxEntity->ReadFromChunkStream(xStream);
	}
}

void Asura_ServerEntity_LookAtTrigger::Update()
{
	if (IsActive())
	{
		m_fUpdateTimer -= Asura_Timers::GetGameTimeSlice();
		// See if we want to update our count
		if (m_fUpdateTimer < 0.0f)
		{
			m_fUpdateTimer += fLOOKAT_TRIGGER_UPDATE_TIME; // Use += to avoid these tending to bunch up over time

			m_iNumberCurrentlyTriggering = m_iClientNumberCurrentlyTriggering;
			// Search the types we need to
			// For now, we only care about non-player actors in the server version (which can only use non-camera tests)

			// AJG - 30/10/07 - I'm pretty sure this will never work in a real project, since no-one subclasses the Asura_ServerEntity_Actor class any more.
			if ((m_uTriggeringTypes & ASURA_LOOKATTRIGGER_TYPE_ALLOTHERACTORS) && !m_bUseScreenArea)
			{
				Asura_Hashed_List_ConstIt<Asura_ServerEntity_Actor> xActorIt(Asura_ServerEntity_Actor::GetList());

				while (!xActorIt.Done())
				{
					const Asura_ServerEntity_Actor* pxActor = xActorIt.GetCurrent();
					xActorIt.Next();

					if (pxActor && 
						pxActor->GetGuid() > ASURA_GUID_CLIENT_END && // Exclude players.
						!pxActor->IsDead())
					{
						Asura_Vector_3 xActorToTriggerDirection = m_xPosition - pxActor->GetPosition();
						float fActorToTriggerDistance = xActorToTriggerDirection.Normalise();

						bool bActorIsLooking = false;
						const bool bTriggerOnNotLooking = m_bTriggerOnNotLooking; // Warning fix

						// Test distance within range
						if ((fActorToTriggerDistance <= m_fMaxDistance) && (fActorToTriggerDistance >= m_fMinDistance))
						{
							Asura_Vector_3 xActorDirection(pxActor->GetOrientation().e13, pxActor->GetOrientation().e23, pxActor->GetOrientation().e33);
							float fCosAngle = xActorDirection * xActorToTriggerDirection;
							// Test view angle within range
							if ((fCosAngle >= m_fCosMaxAngle) && (fCosAngle <= m_fCosMinAngle))
							{
								bool bContinue = (m_fInwardConeCosMaxAngle >= 1.0f); // Skip inward cone test if angle is 0
								if (!bContinue)
								{
									// Test within inward cone
									float fCosOtherAngle = m_xDirection * -xActorToTriggerDirection;
									bContinue = (fCosOtherAngle >= m_fInwardConeCosMaxAngle);
								}

								if (bContinue)
								{
									// LOS if required
									if (m_bUseLOS)
									{
										Asura_LOS_InputData xInput(pxActor->GetPosition(), xActorToTriggerDirection, ASURA_GUID_UNREGISTERED, fActorToTriggerDistance);
										const Asura_LOS_OutputData* pxOutput = Asura_LOS::TestAgainstEnvironment(&xInput);
										if (!pxOutput->m_bHitEnvironment)
										{
											bActorIsLooking = true;
										}
									}
									else // Already succeeded then
									{
										bActorIsLooking = true;
									}
								}
							}
						}

						if(bActorIsLooking != bTriggerOnNotLooking)	m_iNumberCurrentlyTriggering++;
					}
				}
			}
		}

		if (m_iNumberCurrentlyTriggering >= m_iNumberRequired)
		{
			m_fTimeRemaining -= Asura_Timers::GetGameTimeSlice();
		}
		else if (m_bResetTimeOnEarlyLookAway)
		{
			m_fTimeRemaining = m_fTime;
		}
	}

	PARENT::Update();
}

bool Asura_ServerEntity_LookAtTrigger::HandleMessage(const Asura_Message* pxMessage)
{
	switch (pxMessage->GetMsgID())
	{
	default:
		break;
	case ASURA_MSG_LOOKATTRIGGER_CLIENTCOUNTCHANGE:
		{
			const Asura_Msg_LookAtTrigger_ClientCountChange* pxData = static_cast<const Asura_Msg_LookAtTrigger_ClientCountChange*>(pxMessage->GetData());
			if (pxData)
			{
				// Only guaranteed to work correctly if messages use guaranteed delivery and don't reorder. Hopefully they won't be sent often enough for this to be a problem.
				m_iClientNumberCurrentlyTriggering += pxData->m_iCurrent - pxData->m_iPrevious;
				// The next line is not necessary here, as we can leave it to the update. Doing so means we don't actually see the effect of this change until the next time the update counter reaches 0, which is more consistent with what's happening here. A matter of preference really.
				// m_iNumberCurrentlyTriggering += pxData->m_iCurrent - pxData->m_iPrevious;
			}
		}
		return true;
	}

	return PARENT::HandleMessage(pxMessage);
}

void Asura_ServerEntity_LookAtTrigger::MakeCreateMessage(Asura_Guid uRequestingClient, bool bSendGuaranteed/* = false*/)
{
	u_short usMsgID;
	if (IsActive())
	{
		usMsgID = ASURA_MSG_CLIENTCREATE;
	}
	else
	{
#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		// Inactive clients are allowed, so send a create inactive message.
		usMsgID = ASURA_MSG_CLIENTCREATEINACTIVE;
#else //#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		// Inactive clients not allowed, don't send the create message.
		return;
#endif //#else //#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
	}
	Asura_Message* pxMessage = Asura_Network::CreateMessage(usMsgID, sizeof(Asura_Msg_LookAtTrigger_ClientCreate), GetClassification());
	if (pxMessage)
	{
		Asura_Msg_LookAtTrigger_ClientCreate* pxData = static_cast<Asura_Msg_LookAtTrigger_ClientCreate*>(pxMessage->GetData());
		if (pxData)
		{
			pxData->m_uClientLookAtFlags = 0;
			pxData->m_bAdjustForFOV = m_bAdjustForFOV;
			pxData->m_bUseLOS = m_bUseLOS;
			pxData->m_bTriggerOnNotLooking = m_bTriggerOnNotLooking;
			pxData->m_bUseScreenArea = m_bUseScreenArea;
			pxData->m_fMaxAngleOrEdgeDeadzone = m_bUseScreenArea ? m_fScreenEdgeDeadzone : m_fCosMaxAngle;
			pxData->m_fMinAngleOrCentreDeadzone = m_bUseScreenArea ? m_fScreenCentreDeadzone : m_fCosMinAngle;
			pxData->m_fInwardConeCosMaxAngle = m_fInwardConeCosMaxAngle;
			pxData->m_fMaxDistance = m_fMaxDistance;
			pxData->m_fMinDistance = m_fMinDistance;
			pxData->m_uTriggeringTypes = m_uTriggeringTypes;
			pxData->m_iSpecifiedPlayer = m_iSpecifiedPlayer;
			pxData->m_xDirection = m_xDirection;
			pxData->m_xPosition = m_xPosition;
		}

		pxMessage->SetToGuid(ASURA_GUID_MY_CLIENT_ENTITIES);
		pxMessage->SetFromGuid(GetGuid());
		if (bSendGuaranteed)
		{
			pxMessage->UseGuaranteedDelivery();
		}
		Asura_Network::SendToSpecificMachine(pxMessage, uRequestingClient);
	}
}

bool Asura_ServerEntity_LookAtTrigger::IsTriggered()
{
	if (m_bTriggered)
	{
		return false;
	}
	if ((m_iNumberCurrentlyTriggering >= m_iNumberRequired) && (m_fTimeRemaining <= 0.0f))
	{
		m_bTriggered = true;
		return true;
	}

	return false;
}

bool Asura_ServerEntity_LookAtTrigger::IsUnTriggered()
{
	if (!m_bTriggered)
	{
		return false;
	}
	if ((m_iNumberCurrentlyTriggering < m_iNumberRequired) || (m_fTimeRemaining > 0.0f))
	{
		m_bTriggered = false;
		return true;
	}

	return false;
}

void Asura_ServerEntity_LookAtTrigger::OnActivation()
{
	PARENT::OnActivation();

	// If we have a client, update its active status.
	if ((m_uTriggeringTypes == ASURA_LOOKATTRIGGER_TYPE_SPECIFICPLAYER) || (m_uTriggeringTypes & ASURA_LOOKATTRIGGER_TYPE_ALLPLAYERS))
	{
#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_ACTIVATE);
		if (pxMessage)
		{
			pxMessage->UseGuaranteedDelivery();

			Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
		}
#else //#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		MakeCreateMessage(ASURA_GUID_UNREGISTERED, true);
#endif //#else //#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
	}
}

void Asura_ServerEntity_LookAtTrigger::OnDeactivation()
{
	PARENT::OnDeactivation();

	// If we have a client, update its active status.
	if ((m_uTriggeringTypes == ASURA_LOOKATTRIGGER_TYPE_SPECIFICPLAYER) || (m_uTriggeringTypes & ASURA_LOOKATTRIGGER_TYPE_ALLPLAYERS))
	{
#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
#else //#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DESTRUCT);
#endif //#else //#ifdef ASURA_ALLOW_INACTIVE_CLIENT_ENTITIES
		if (pxMessage)
		{
			pxMessage->UseGuaranteedDelivery();

			Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
		}
	}
}

void Asura_ServerEntity_LookAtTrigger::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// Unfortunately, triggers don't use the best-practice approach of: our version -> parent write -> our data
	// Instead it's: parent write -> our version and data.
	PARENT::WriteToChunkStream(xStream);

	int iVersion = iASURA_LOOKAT_TRIGGER_CHUNK_VERSION;
	xStream << iVersion;

	// Assertions about unusual combinations
	Asura_Assert(m_fMaxDistance >= m_fMinDistance, "Max distance less than min distance, correcting...");
	Asura_Assert(m_fCosMaxAngle <= m_fCosMinAngle, "Max angle less than min angle, correcting...");
	Asura_Assert((m_fScreenEdgeDeadzone + m_fScreenCentreDeadzone) < 1.0f, "Screen deadzone totals at least 1, this leaves none of the screen live.");
	Asura_Assert(m_uTriggeringTypes, "Look-at trigger saved with nothing able to trigger it");
	Asura_Assert(!m_bUseScreenArea || (m_uTriggeringTypes != ASURA_LOOKATTRIGGER_TYPE_ALLOTHERACTORS), "Look-at trigger using camera-only tests only on non-players, this means it can never be triggered.");

	xStream << Asura_Maths::Max(m_fMaxDistance, m_fMinDistance);
	xStream << Asura_Maths::Min(m_fMaxDistance, m_fMinDistance);

	xStream << Asura_Maths::Min(m_fCosMaxAngle, m_fCosMinAngle);
	xStream << Asura_Maths::Max(m_fCosMaxAngle, m_fCosMinAngle);

	xStream << m_fScreenEdgeDeadzone;
	xStream << m_fScreenCentreDeadzone;

	xStream << m_fTime;

	xStream << m_fInwardConeCosMaxAngle;

	xStream << m_uTriggeringTypes;
	xStream << m_iSpecifiedPlayer;
	xStream << m_iNumberRequired;
	
	xStream << m_xPosition;
	xStream << m_xDirection;

	Asura_WriteToStream_Bitfield_Wrapper<u_int> xFlags(m_uLookAtFlags);
	xFlags.Add(m_bUseScreenArea, 1);
	xFlags.Add(m_bUseLOS, 1);
	xFlags.Add(m_bResetTimeOnEarlyLookAway, 1);
	xFlags.Add(m_bAdjustForFOV, 1);
	xFlags.Add(m_bTriggered, 1);
	xFlags.Add(m_bTriggerOnNotLooking, 1);
	xStream << xFlags;
}

void Asura_ServerEntity_LookAtTrigger::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	// Unfortunately, triggers don't use the best-practice approach of: our version -> parent write -> our data
	// Instead it's: parent write -> our version and data.
	PARENT::ReadFromChunkStream(xStream);

	int iVersion;
	xStream >> iVersion;
	Asura_Assert(iVersion <= iASURA_LOOKAT_TRIGGER_CHUNK_VERSION, "Invalid look-at trigger chunk");

	xStream >> m_fMaxDistance;
	xStream >> m_fMinDistance;

	xStream >> m_fCosMaxAngle;
	xStream >> m_fCosMinAngle;

	xStream >> m_fScreenEdgeDeadzone;
	xStream >> m_fScreenCentreDeadzone;

	xStream >> m_fTime;

	xStream >> m_fInwardConeCosMaxAngle;

	xStream >> m_uTriggeringTypes;
	xStream >> m_iSpecifiedPlayer;
	xStream >> m_iNumberRequired;
	
	xStream >> m_xPosition;
	xStream >> m_xDirection;

	xStream >> m_uLookAtFlags;
}

int Asura_ServerEntity_LookAtTrigger::GetChunkBufferSize()
{
	Asura_Assert(false, "Deprecated chunk function - Look-at Trigger");

	return 0;
}

void Asura_ServerEntity_LookAtTrigger::FillInChunkBuffer(char*& pcChunkBuffer)
{
	Asura_Assert(false, "Deprecated chunk function - Look-at Trigger");
}

#ifdef ASURA_SUPPORT_ENTITYWATCH
void Asura_ServerEntity_LookAtTrigger::DisplayEntityWatchText()
{
	Asura_ServerEntity_Trigger::DisplayEntityWatchText();
	Asura_EntityWatch_System::PrintText(0,"Num Triggering : %d \\ %d",m_iNumberCurrentlyTriggering,m_iNumberRequired);
	Asura_EntityWatch_System::PrintText(0,"Timer : %6.3f",m_fTimeRemaining);
	Asura_EntityWatch_System::PrintText(0,"Triggered : %s",m_bTriggered ? "Yes" : "No");
}
#endif

#ifdef ASURA_TOOLS
void Asura_ServerEntity_LookAtTrigger::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	Asura_ServerEntity_Trigger::ApplyChangesForTemplateInstance(pxInstance);

	m_xPosition = pxInstance->TransformPosition(m_xPosition);
	m_xDirection = m_xDirection * pxInstance->GetRelativeMatrix();

}
#endif // ASURA_TOOLS

// NetMsg registration
#ifdef ASURA_NETMSG_PACK_ALL_DATA
void Asura_Msg_LookAtTrigger_ClientCreate::RegisterFields(Asura_NetMsg* pxNetMsg) const
{
	Asura_Assert(pxNetMsg->GetMsgID() == ASURA_MSG_CLIENTCREATE, "unexpected message ID");
	Asura_Assertf(pxNetMsg->GetEntityClassification() == AsuraEntityClass_LookAtTrigger, 
		("unexpected entity classification (%s)", Asura_Network::GetEntityClassificationDesc(pxNetMsg->GetEntityClassification())));

	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_float, m_fMaxDistance );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_float, m_fMinDistance );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_float, m_fMaxAngleOrEdgeDeadzone );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_float, m_fMinAngleOrCentreDeadzone );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_float, m_fInwardConeCosMaxAngle );
	
	Asura_NetMsg_RegisterIntFieldSimpleBits( m_uTriggeringTypes, ASURA_LOOKATTRIGGER_NUM_BITS_FOR_TRIGGERING_TYPES );
	Asura_NetMsg_RegisterIntFieldSimpleBits( m_iSpecifiedPlayer, 1 );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_vector3, m_xPosition );
	Asura_NetMsg_RegisterFieldSimple( Asura_FieldType_vector3, m_xDirection );
	
	Asura_NetMsg_RegisterBitfieldSimple( m_uClientLookAtFlags, 5, ASURA_NETMSG_SINGLEBITS );
}
#endif

