//////////////////////////////////////////////////////////////////////////////
//
// Asura_ServerEntity_SoundController.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     entities which control phonons
//
// History:
//
//     (MH  - 15.11.2002) - Created.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/Triggers/Asura_ServerEntity_SoundController.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Network/Asura_Guid.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "Sound/Asura_Phonon.h"
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)

#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif


// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Hashed_List<Asura_ServerEntity_SoundController> Asura_ServerEntity_SoundController::s_xList;


// C L A S S   M E T H O D S /////////////////////////////////////////////////
Asura_ServerEntity_SoundController::Asura_ServerEntity_SoundController(Asura_Guid uGuid, Asura_Guid uPhononGuid, bool bInitiallyActive)
:Asura_ServerEntity_Activatable(uGuid)
{
    Classification = AsuraEntityClass_SoundController;

    m_uPhononGuid = uPhononGuid;
	SetStartsActive(bInitiallyActive);
    
    s_xList.AddToList(this);
}

Asura_ServerEntity_SoundController::Asura_ServerEntity_SoundController(Asura_Guid uGuid, float fDummy)
:Asura_ServerEntity_Activatable(uGuid)
{
    Classification = AsuraEntityClass_SoundController;

    m_uPhononGuid = ASURA_GUID_UNREGISTERED;

    s_xList.AddToList(this);
}

Asura_ServerEntity_SoundController::~Asura_ServerEntity_SoundController()
{
    s_xList.RemoveFromList(this);
}

void Asura_ServerEntity_SoundController::CreateFromChunkStream(Asura_Guid uGuid, Asura_Chunk_Stream& xStream)
{
	Asura_ServerEntity_SoundController* pxEntity = new Asura_ServerEntity_SoundController(uGuid,0.0f);
	pxEntity->ReadFromChunkStream(xStream);
}


void Asura_ServerEntity_SoundController::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
    Classification = AsuraEntityClass_SoundController;

	Asura_ServerEntity_Activatable::ReadFromChunkStream(xStream);

	//check the version
	int iVersion;
	xStream >> iVersion;
	Asura_Assert(iVersion <= ASURA_SERVERENTITY_SOUNDCONTROLLER_VERSION_NUMBER, "unknown sound controller version number");

	//read in the rest
	xStream >> m_uPhononGuid;
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
	m_uPhononGuid	= Asura_Streaming_System::MakeCurrentMasterZoneGuid(m_uPhononGuid);
#endif	// defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)

	if(iVersion >= 1)
	{
		//skip the position , since only the client needs to deal with it
		xStream.Skip(sizeof(Asura_Vector_3));
	}

	SetNeverUpdates(true);
}

void Asura_ServerEntity_SoundController::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/
	Asura_ServerEntity_Activatable::WriteToChunkStream(xStream);

	//get the current phonon position
	Asura_Vector_3 xPosition;
    Asura_Phonon* pxPhonon = Asura_Phonon_System::GetPhonon(m_uPhononGuid);
	if(pxPhonon)
	{
		//For level exports , ensure that the initial position is saved out instead , in case
		//the phonon is currently not in the correct position.
		//Saving the non-initial position in other cases so that the correct value gets saved for
		//quicksaves.
		//In theory the inital position should be saved instead when saving the .ent file ,but
		//there isn't currently a good way to distinguish this from a quicksave.
		//However it shouldn't matter since the value will be largely ignored when being loaded
		//back into tools.
		//RAR 12/11/2007

		#ifdef ASURA_TOOLS
		if(Tools_Level_Export::IsExport())
		{
			xPosition = pxPhonon->GetInitialPosition();
		}
		else
		#endif
		{
			xPosition = pxPhonon->GetPosition();
		}

		#ifdef ASURA_TOOLS
		//For recentreing during level exports :
		//(Normally this isn't done for entities , since it is handled via ApplyChangesForTemplateInstance() ,
		//but in this case it is done here since the entity doesn't store the position itself.)
		//RAR 12/11/2007
		xPosition -= Tools_Level_Export::GetExportCentre();
		#endif
	}
	else
	{
		xPosition.Zero();
	}



	int iVersion = ASURA_SERVERENTITY_SOUNDCONTROLLER_VERSION_NUMBER;
	xStream << iVersion;
	xStream << m_uPhononGuid;
	xStream << xPosition;
}
 

// (MH - 12.12.2002) override trigger handler so it also activates
bool Asura_ServerEntity_SoundController::HandleMessage(const Asura_Message* pxMessage)
{
    if( IsDestroyed() ) { return true; } // ignore all further messages?

	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			default:
			{
				return Asura_ServerEntity_Activatable::HandleMessage(pxMessage);
			}

			case ASURA_MSG_TRIGGER:
			{
                // (MH - 12.12.2002) override this message because:-
                // 1. base class doesn't do anything with trigger msg if it isn't active
                // 2. base class doesn't do anything with activate msg if already active.
				if (!IsActive())
				{
                    SetInternalActiveFlag(true);
                }
				OnTrigger();
				SetNeverUpdates(true); 
				return true;
			}
		}
    }
    return false;
}


// overrides, these just send the message to the clients. Don't call up to the base though.
// There should be nothing in there we need to do.
void Asura_ServerEntity_SoundController::OnActivation()
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_ACTIVATE);
   	Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
}

void Asura_ServerEntity_SoundController::OnDeactivation()
{
   	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DEACTIVATE);
   	Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
}

void Asura_ServerEntity_SoundController::OnTrigger()
{
   	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_TRIGGER);
   	Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
}




// static function to retrieve entity from phonon guid
Asura_ServerEntity_SoundController* 
Asura_ServerEntity_SoundController::GetControllerFromPhonon(Asura_Guid uPhononGuid)
{
	for(Asura_Hashed_List_It<Asura_ServerEntity_SoundController> xIterator(&s_xList);!xIterator.Done();xIterator.Next())
	{
		Asura_ServerEntity_SoundController* pxSoundController = xIterator.GetCurrent();
        if (pxSoundController->m_uPhononGuid == uPhononGuid)
        {
            return pxSoundController;
        }
    }
    return NULL;
}


#ifdef ASURA_TOOLS
void Asura_ServerEntity_SoundController::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	Asura_ServerEntity_Activatable::ApplyChangesForTemplateInstance(pxInstance);

	//update the sound guid
	m_uPhononGuid = pxInstance->ConvertSoundGuid(m_uPhononGuid);

}
#endif // ASURA_TOOLS
