/**
*
* \file Asura_ServerEntity_TimeTrigger.cpp
*
* Created by Kevin Lea - 22/04/2002 10:37:42
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

// This file's header.
#include "Entity/Triggers/Asura_ServerEntity_TimeTrigger.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#ifdef ASURA_SUPPORT_ENTITYWATCH
#include "Entity/Asura_EntityWatch_System.h"
#endif
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Message.h"
#include "Rendering/Asura_Render.h"

#define ASURA_SERVERENTITY_TIMETRIGGER_VERSION 4

Asura_ServerEntity_TimeTrigger::Asura_ServerEntity_TimeTrigger(float time, bool bActive) : Asura_ServerEntity_Trigger(bActive)
{
	Classification = AsuraEntityClass_TimeTrigger;
	Time = time;
	m_fTimeRemaining = Time;
	m_uTimeTriggerFlags = 0;
    
}

Asura_ServerEntity_TimeTrigger::Asura_ServerEntity_TimeTrigger(Asura_Guid uGuid)
: Asura_ServerEntity_Trigger(uGuid)
, Time(0.0f)
, m_fTimeRemaining(0.0f)
, m_uTimeTriggerFlags(0)
{
	Classification = AsuraEntityClass_TimeTrigger;
}

Asura_ServerEntity_TimeTrigger::Asura_ServerEntity_TimeTrigger(Asura_Guid guid,const char*& chunkBuffer)
:Asura_ServerEntity_Trigger(guid,chunkBuffer)
{
	Classification = AsuraEntityClass_TimeTrigger;
	//check the version
	int version;
	ASURA_CHUNK_EXTRACT( chunkBuffer, version,int);
	Asura_Assert(version<=ASURA_SERVERENTITY_TIMETRIGGER_VERSION,"Unknown chunk version");

	//read in the rest
	ASURA_CHUNK_EXTRACT( chunkBuffer, Time,float);

	if(version>0)
	{
		ASURA_CHUNK_EXTRACT( chunkBuffer, m_fTimeRemaining,float);
	}
	else
	{
   	    m_fTimeRemaining = Time;
	}

	if(version >= 4)
	{
		ASURA_CHUNK_EXTRACT(chunkBuffer, m_uTimeTriggerFlags, u_int);
	}
	else if(version == 3)
	{
		u_char ucTemp;
		ASURA_CHUNK_EXTRACT( chunkBuffer, ucTemp, u_char);
		m_uTimeTriggerFlags = ucTemp ? ASURA_TIMETRIGGERFLAG_LOOP : 0;
	}
	else
	{
		m_uTimeTriggerFlags = 0;
	}

}

Asura_ServerEntity_TimeTrigger::~Asura_ServerEntity_TimeTrigger()
{
}

void Asura_ServerEntity_TimeTrigger::CreateFromChunkBuffer(Asura_Guid uGuid, const char*& pcChunkBuffer)
{
#ifdef TEST_ENTITY_CHUNK_STREAM
	TestStream<Asura_ServerEntity_TimeTrigger>(uGuid, pcChunkBuffer);
#endif
	new Asura_ServerEntity_TimeTrigger(uGuid, pcChunkBuffer);
}

void Asura_ServerEntity_TimeTrigger::Update()
{
	const bool bPauseDuringCutscenes = ((m_uTimeTriggerFlags & ASURA_TIMETRIGGERFLAG_PAUSE_DURING_CUTSCENES) != 0);
	if(IsActive() && !(bPauseDuringCutscenes && Asura_Render::GetCutsceneHasCameraControl()))
	{
		m_fTimeRemaining -= Asura_Timers::GetGameTimeSlice();
	}
	Asura_ServerEntity_Trigger::Update();
}


bool Asura_ServerEntity_TimeTrigger::IsTriggered(void)
{
	if (m_fTimeRemaining<=0)
	{
		if(IsLooped())
		{
			if(Time > 0.0f)
			{
				m_fTimeRemaining = Time + Asura_Maths::FloatMod(m_fTimeRemaining,Time);
			}
			else
			{
				m_fTimeRemaining = 0.0f;
			}
		}
		else
		{
			SetInternalActiveFlag(false);
			m_fTimeRemaining = Time;
		}
		return true;
	}

	return false;
}

void Asura_ServerEntity_TimeTrigger::Reset(void)
{
	m_fTimeRemaining = Time;
}



int Asura_ServerEntity_TimeTrigger::GetChunkBufferSize()
{
	return Asura_ServerEntity_Trigger::GetChunkBufferSize()+sizeof(float)*2+sizeof(int)+sizeof(u_int);
}

void Asura_ServerEntity_TimeTrigger::FillInChunkBuffer(char*& chunkBuffer)
{
	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/
	//fill in base class
	Asura_ServerEntity_Trigger::FillInChunkBuffer(chunkBuffer);

	int version = ASURA_SERVERENTITY_TIMETRIGGER_VERSION;
	ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, version, int);
	ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, Time, float);
	ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, m_fTimeRemaining, float);
	ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, m_uTimeTriggerFlags, u_int);
}

void Asura_ServerEntity_TimeTrigger::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	const int	version	= ASURA_SERVERENTITY_TIMETRIGGER_VERSION;

	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/
	//fill in base class
	Asura_ServerEntity_Trigger::WriteToChunkStream	(xStream);

	xStream << version;
	xStream << Time;
	xStream << m_fTimeRemaining;
	xStream << m_uTimeTriggerFlags;
}

void Asura_ServerEntity_TimeTrigger::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/

	// call parent first as this used to go through the horrid 'chunk constructor' method
	Asura_ServerEntity_Trigger::ReadFromChunkStream( xStream );

	int iVersion = 0;
	xStream >> iVersion;

	if( iVersion < ASURA_SERVERENTITY_TIMETRIGGER_VERSION )
	{
		Asura_Assertf( false, ("Version %d is not supported by this class' ReadFromChunkStream", iVersion) );
		return;
	}

	xStream >> Time;
	xStream >> m_fTimeRemaining;
	xStream >> m_uTimeTriggerFlags;
}

bool Asura_ServerEntity_TimeTrigger::HandleMessage(const Asura_Message* pxMessage)
{
    if( IsDestroyed() ) { return true; } // ignore all further messages?
	
	if (pxMessage)
	{
		switch (pxMessage->GetMsgID())
		{
			case ASURA_MSG_RESET:
			{
				Reset();
				SetInternalActiveFlag(true);
				return true;
			}

			case ASURA_MSG_SETTIMER :
			{
				const Asura_Message_SetTimer* pxMessageData = reinterpret_cast
					<const Asura_Message_SetTimer*>(pxMessage->GetData());

				//Note that the timer is considered to be counting up as far as the UI and the messages are concerned,
				//but in fact counts down internally. RAR 15/09/2006
				m_fTimeRemaining = Time - pxMessageData->m_fTimer;
				return true;
			}

			case ASURA_MSG_TIMETRIGGER_SETTRIGGERTIME : 
			{
				const Asura_Message_TimeTrigger_SetTriggerTime* pxMessageData = reinterpret_cast
					<const Asura_Message_TimeTrigger_SetTriggerTime*>(pxMessage->GetData());

				//Note that the timer is considered to be counting up as far as the UI and the messages are concerned,
				//but in fact counts down internally. RAR 15/09/2006
				float fCurrentTime = Time - m_fTimeRemaining;
				Time = pxMessageData->m_fTriggerTime;
				m_fTimeRemaining = Time - fCurrentTime;
				return true;
			}

			case ASURA_MSG_TIMETRIGGER_ADDTOTRIGGERTIME : 
			{
				const Asura_Message_TimeTrigger_AddToTriggerTime* pxMessageData = reinterpret_cast
					<const Asura_Message_TimeTrigger_AddToTriggerTime*>(pxMessage->GetData());

				//Note that the timer is considered to be counting up as far as the UI and the messages are concerned,
				//but in fact counts down internally. RAR 15/09/2006
				float fCurrentTime = Time - m_fTimeRemaining;
				Time += pxMessageData->m_fTriggerTimeDelta;
				if(Time < 0.0f) Time = 0.0f;
				m_fTimeRemaining = Time - fCurrentTime;
				return true;
			}


			case ASURA_MSG_TIMETRIGGER_ADDTIME : 
				{
					const Asura_Message_TimeTrigger_AddTime* pxMessageData = reinterpret_cast
						<const Asura_Message_TimeTrigger_AddTime*>(pxMessage->GetData());

					// Add the time on to the remaining time, countdown length is not affected so trigger will reset properly
					m_fTimeRemaining += pxMessageData->m_fTimeDelta;
					if (m_fTimeRemaining < 0.0f) { m_fTimeRemaining = 0.0f; }
					return true;
				}

			case ASURA_MSG_PAUSETIMER :
			{
				//pause by deactivating the timer
				SetInternalActiveFlag(false);
				//Not decided whether it is desirable to trigger the messages for this. So just asserting there aren't any for now. RAR  18/09/2006
				Asura_Assert(MessageBlock[ASURA_TRIGGER_ENTITY_ONDEACTIVATION] == ASURA_STATICMESSAGE_EMPTYBLOCK,"Behaviour for pausing timers with on-deactivation messages is undefined");
				return true;
			}

			case ASURA_MSG_UNPAUSETIMER :
			{
				//unpause by deactivating the timer
				SetInternalActiveFlag(true);
				//Not decided whether it is desirable to trigger the messages for this. So just asserting there aren't any for now. RAR  18/09/2006
				Asura_Assert(MessageBlock[ASURA_TRIGGER_ENTITY_ONACTIVATION] == ASURA_STATICMESSAGE_EMPTYBLOCK,"Behaviour for unpausing timers with on-activation messages is undefined");
				return true;
			}

		}
	}
	
	return Asura_ServerEntity_Trigger::HandleMessage(pxMessage);
}


#ifdef ASURA_SUPPORT_ENTITYWATCH
void Asura_ServerEntity_TimeTrigger::DisplayEntityWatchText()
{
	Asura_ServerEntity_Trigger::DisplayEntityWatchText();
	Asura_EntityWatch_System::PrintText(0,"Timer : %.3f / %.3f",Time-m_fTimeRemaining,Time);
}
#endif //ASURA_SUPPORT_ENTITYWATCH
