/**
*
* \file Asura_ServerEntity_Trigger.cpp
*
* Created by Kevin Lea - 16/04/2002 10:23:15
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

// This file's header.
#include "Entity/Triggers/Asura_ServerEntity_Trigger.h"

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Message.h"

#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif


#define ASURA_SERVERENTITY_TRIGGER_VERSION_NUMBER 4

Asura_Hashed_List<Asura_ServerEntity_Trigger> Asura_ServerEntity_Trigger::s_xList(3);

Asura_ServerEntity_Trigger::Asura_ServerEntity_Trigger(bool initiallyActive) 
: Asura_ServerEntity_Activatable(initiallyActive)
{
	s_xList.AddToList(this);

	for (int i=0; i<ASURA_TRIGGER_ENTITY_NOOFSTATES; i++)
	{
		MessageBlock[i] = ASURA_STATICMESSAGE_EMPTYBLOCK;
	}
	SetNeverUpdateClient();
}

Asura_ServerEntity_Trigger::Asura_ServerEntity_Trigger(Asura_Guid uGuid) 
: Asura_ServerEntity_Activatable(uGuid)
{
	s_xList.AddToList(this);

	for (int i=0; i<ASURA_TRIGGER_ENTITY_NOOFSTATES; i++)
	{
		MessageBlock[i] = ASURA_STATICMESSAGE_EMPTYBLOCK;
	}
	SetNeverUpdateClient();
}

Asura_ServerEntity_Trigger::Asura_ServerEntity_Trigger(Asura_Guid guid,const char*& pcChunkBuffer)
:Asura_ServerEntity_Activatable(guid,pcChunkBuffer)
{
	s_xList.AddToList(this);

	//check the version
	int iVersion;
	ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion,int);
	Asura_Assert(iVersion <= ASURA_SERVERENTITY_TRIGGER_VERSION_NUMBER, "FATAL ERROR: unknown version number for Asura_ServerEntity_Trigger");

	// (MH - 08.01.2003) attempting to fix Adam's broken loader
    if(iVersion <= 2)
	{
		for(int i = 0; i < ASURA_TRIGGER_ENTITY_NOOFSTATES-1; i++)
		{
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, MessageBlock[i], unsigned short);
		}

		MessageBlock[static_cast<int>(ASURA_TRIGGER_ENTITY_NOOFSTATES-1)] = ASURA_STATICMESSAGE_EMPTYBLOCK;

		if(iVersion == 1)
		{
			//For this earlier version , initially active is stored in the data for this class , instead of the parent
			unsigned int uActive;
			ASURA_CHUNK_EXTRACT(pcChunkBuffer, uActive,unsigned int);

			SetInternalActiveFlag(uActive > 0);
		}
    }
	else
	{
		//read in the rest
		for(int i=0; i < ASURA_TRIGGER_ENTITY_NOOFSTATES; i++)
		{
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, MessageBlock[i],unsigned short);
		}
        
        // (MH - 08.01.2003) fix alignment
        if(iVersion >= 4)
        {
			pcChunkBuffer += sizeof(u_short);
        }
	}

	SetNeverUpdateClient();
}

Asura_ServerEntity_Trigger::~Asura_ServerEntity_Trigger()
{
	s_xList.RemoveFromList(this);
}

void Asura_ServerEntity_Trigger::Update()
{
    if( IsDestroyed() ) { return; }

	if (IsActive())
	{
		if(IsTriggered())
		{
			/// Kevin Lea - 16/04/2002 10:34:49 - hmm, technically could send message to self saying 'trigger'
			/// might make debugging easier... Issues with message generation and resolution - are messages that
			/// are made one frame acted upon in the same frame, or the one after? (Problems with recursive/explosive
			/// amounts of message handling.)
			OnTrigger();
		}
		else if (IsUnTriggered())
		{
			OnUnTrigger();
		}
	}
}


void Asura_ServerEntity_Trigger::OnActivation()
{
	Asura_ServerEntity_Activatable::OnActivation();
	SendMessages(ASURA_TRIGGER_ENTITY_ONACTIVATION);
	Reset();
}

void Asura_ServerEntity_Trigger::OnDeactivation()
{
	Asura_ServerEntity_Activatable::OnDeactivation();
	SendMessages(ASURA_TRIGGER_ENTITY_ONDEACTIVATION);
}

void Asura_ServerEntity_Trigger::OnTrigger()
{
	Asura_ServerEntity_Activatable::OnTrigger();
	SendMessages(ASURA_TRIGGER_ENTITY_ONBEINGTRIGGERED);
}

void Asura_ServerEntity_Trigger::OnDestruction()
{
	SendMessages(ASURA_TRIGGER_ENTITY_ONDESTRUCTION);
	Asura_ServerEntity_Activatable::OnDestruction();
}

void Asura_ServerEntity_Trigger::OnUnTrigger()
{
	Asura_ServerEntity_Activatable::OnUnTrigger();
	SendMessages(ASURA_TRIGGER_ENTITY_ONUNTRIGGER);
}


void Asura_ServerEntity_Trigger::SendMessages(enum ASURA_TRIGGER_ENTITY_STATE state)
{
	if (MessageBlock[state]!=ASURA_STATICMESSAGE_EMPTYBLOCK)
	{
		Asura_StaticMessage_System::SendMessages(MessageBlock[state],Guid);
	}
}


int Asura_ServerEntity_Trigger::GetChunkBufferSize()
{
	return Asura_ServerEntity_Activatable::GetChunkBufferSize() +
		   sizeof(unsigned short)*ASURA_TRIGGER_ENTITY_NOOFSTATES + 
           sizeof(unsigned short) +     // padding
           sizeof(int);                 // version
}

void Asura_ServerEntity_Trigger::FillInChunkBuffer(char*& chunkBuffer)
{
	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/
	Asura_ServerEntity_Activatable::FillInChunkBuffer(chunkBuffer);

	int version = ASURA_SERVERENTITY_TRIGGER_VERSION_NUMBER;
	ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, version, int);
	for(int i=0;i<ASURA_TRIGGER_ENTITY_NOOFSTATES;i++)
	{
		ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, MessageBlock[i], unsigned short);
	}

    // (MH - 08.01.2003) fix alignment.
    unsigned short usDummy = 0;
    ASURA_CHUNK_FILLANDVERIFY( chunkBuffer, usDummy, unsigned short);
}

void Asura_ServerEntity_Trigger::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	const int	version	= ASURA_SERVERENTITY_TRIGGER_VERSION_NUMBER;

	/*
	This entity type can be used in template files , so any changes to FillInChunkBuffer may
	require ApplyChangesForTemplateInstance() to be updated as well.
	Please see the comment at the top of Tools/EntityTemplate/Tools_EntityTemplate_System.h before 
	making any changes.
	RAR 29/04/2004
	*/
	Asura_ServerEntity_Activatable::WriteToChunkStream	(xStream);

	xStream << version;

	for(int i=0;i<ASURA_TRIGGER_ENTITY_NOOFSTATES;i++)
	{
		xStream << MessageBlock [i];
	}

	xStream.Write4BytePadding	();
}

void Asura_ServerEntity_Trigger::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	int version;

	Asura_ServerEntity_Activatable::ReadFromChunkStream(xStream);

	xStream >> version;

	for(int i=0;i<ASURA_TRIGGER_ENTITY_NOOFSTATES;i++)
	{
		xStream >> MessageBlock [i];
	}

	xStream.Skip4BytePadding();
}


#ifdef ASURA_TOOLS
void Asura_ServerEntity_Trigger::ApplyChangesForTemplateInstance(Tools_EntityTemplateInstance* pxInstance)
{
	Asura_ServerEntity_Activatable::ApplyChangesForTemplateInstance(pxInstance);

	//update the message blocks
	for(int i=0;i<ASURA_TRIGGER_ENTITY_NOOFSTATES;i++)
	{
		MessageBlock[i] = pxInstance->ConvertMessageBlock(MessageBlock[i]);
	}

}
#endif // ASURA_TOOLS
