//
// Asura_EffectVolume_EntityParameters.cpp
//
// Created by Emil Popovic - 15/12/2008
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

// core
#include "Core/Asura.h"

// header
#include "Entity/Triggers/EffectVolume/Asura_EffectVolume_EntityParameters.h"

// asura
#include "Volume/Asura_Volume_Cuboid.h"
#include "Volume/Asura_Volume_Sphere.h"
#include "Volume/Asura_Volume_Cylinder.h"

// tools


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Version	Author	Comment
// ~~~~~~~	~~~~~~	~~~~~~~
// 1		EP		Created.
// 2        EP      Added Transition volumes and removed u_int for num transition volumes.
////////////////////////////////////////////////////////////////////////////////
const u_int	Asura_EffectVolume_EntityParameters::s_uChunkVersion	= 2;



Asura_EffectVolume_EntityParameters::Asura_EffectVolume_EntityParameters()
: m_fFadeInDuration(0.0f)
, m_fFadeOutDuration(0.0f)
, m_uPriority(5)
, m_uNumEffects(0)
, m_pxEffects(NULL)
{
	m_pxVolumeGroup = new Asura_VolumeGroup();
	m_pxTransitionVolumeGroup = new Asura_VolumeGroup();
}


Asura_EffectVolume_EntityParameters::~Asura_EffectVolume_EntityParameters()
{
	if (m_pxVolumeGroup)
	{
		delete m_pxVolumeGroup;
		m_pxVolumeGroup = NULL;
	}

	if (m_pxTransitionVolumeGroup)	
	{
		delete m_pxTransitionVolumeGroup;
		m_pxTransitionVolumeGroup = NULL;
	}

	RemoveAllEffects();
}


void Asura_EffectVolume_EntityParameters::Clear()
{
	m_pxVolumeGroup->Clear();
	m_pxTransitionVolumeGroup->Clear();
	RemoveAllEffects();
}


bool Asura_EffectVolume_EntityParameters::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	Clear();	// Get rid of any data that already exists in this object. So we can start read the data to clean object.

	u_int uVersion;

	xStream >> uVersion;
	if( uVersion>s_uChunkVersion )
	{
		Asura_Assert( false, "Unexpected EffectVolume_EntityParameters chunk version");
		return false;
	}

	xStream >> m_fFadeInDuration;
	xStream >> m_fFadeOutDuration;
	xStream >> m_uPriority;

	xStream >> m_uNumEffects;

	if( m_uNumEffects )
	{
		// Create new list so we can fill it
		m_pxEffects = new Asura_EffectVolume_Effect* [ m_uNumEffects ];

		for( u_int u=0; u<m_uNumEffects; u++ )
		{
			m_pxEffects[u] = NULL;

			u_int uType;
			xStream >> uType;

			switch( uType )
			{
			case uEFFECTVOLUME_EFFECT_FOG:
				{
					m_pxEffects[u] = new Asura_EffectVolume_Effect_Fog();
					break;
				}
#if defined(ASURA_USE_FSFX)
			case uEFFECTVOLUME_EFFECT_FSFX:
				{
					m_pxEffects[u] = new Asura_EffectVolume_Effect_FSFX();
					break;
				}
#endif	//defined(ASURA_USE_FSFX)
			default:
				{
					Asura_Assert(false, "Cannot reach this code! Asura_EffectVolume_EntityParameters::ReadFromChunkStream.");
				}
			}

			if( m_pxEffects[u] )
			{
				bool bSuccess = m_pxEffects[u]->ReadFromChunkStream( xStream );
				
				// Check if the read from the effect was successful. If not, don' continue to read, but rather quit that.
				if( bSuccess==false )
				{
					return false;
				}
			}
		}
	}

	if( uVersion==1 )
	{
		u_int m_uNumTransitionVolumes;			// obsolete information
		xStream >> m_uNumTransitionVolumes;
	}

	if( uVersion>=2 )
	{
		m_pxTransitionVolumeGroup->ReadFromChunkStream(xStream);
	}

	m_pxVolumeGroup->ReadFromChunkStream(xStream);

	return true;
}


void Asura_EffectVolume_EntityParameters::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	xStream << s_uChunkVersion;

	xStream << m_fFadeInDuration;
	xStream << m_fFadeOutDuration;
	xStream << m_uPriority;
	xStream << m_uNumEffects;

	for( u_int u=0; u<m_uNumEffects; u++ )
	{
		u_int uType = m_pxEffects[u]->GetEffectTypeHash();
		xStream << uType;
		m_pxEffects[u]->WriteToChunkStream( xStream );
	}

	m_pxTransitionVolumeGroup->WriteToChunkStream(xStream);
	m_pxVolumeGroup->WriteToChunkStream(xStream);
}


#ifdef ASURA_TOOLS


void Asura_EffectVolume_EntityParameters::Tools_Copy( Asura_EffectVolume_EntityParameters& xTo, const Asura_EffectVolume_EntityParameters& xFrom )
{
	Asura_Chunk_Memory_Stream xStream;
	xStream.SetSwapEndian(false);

	xFrom.WriteToChunkStream(xStream);

	xStream.SetPosition(0);
	xTo.ReadFromChunkStream(xStream);
}


void Asura_EffectVolume_EntityParameters::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	if( m_pxVolumeGroup )
	{
		m_pxVolumeGroup->ApplyChangesForTemplateInstance(pxInstance);
	}

	if( m_pxTransitionVolumeGroup )
	{
		m_pxTransitionVolumeGroup->ApplyChangesForTemplateInstance(pxInstance);
	}

	if( m_pxEffects )
	{
		for( u_int u=0; u<m_uNumEffects; u++ )
		{
			m_pxEffects[u]->ApplyChangesForTemplateInstance(pxInstance);
		}
	}
}


#endif // ASURA_TOOLS


void Asura_EffectVolume_EntityParameters::AddNewEffect( Asura_EffectVolume_Effect* pxNewEffect )
{
	// Increase number of active elements.
	m_uNumEffects++;

	// Create new list so we can replace it with old one.
	Asura_EffectVolume_Effect**	pxNewList = new Asura_EffectVolume_Effect* [ m_uNumEffects ];

	// copy from old to new one
	for( u_int u=0; u<m_uNumEffects-1;  u++ )	// Also note that we are copying one less parameter, because old list has one less parameter
	{
		pxNewList[u] = m_pxEffects[u];
	}

	// and add last effect on at the tail.
	pxNewList[m_uNumEffects-1] = pxNewEffect;


	// Destroy old list and replace it with new list.
	if( m_pxEffects )
	{
		delete [] m_pxEffects;
	}

	m_pxEffects = pxNewList;
}


void Asura_EffectVolume_EntityParameters::RemoveEffect( u_int uNumOfEffectToRemove )
{
	if( uNumOfEffectToRemove>=m_uNumEffects )
	{
		Asura_Assert(false, "");
		return;
	}

	if( m_uNumEffects==1 )
	{
		RemoveAllEffects();
	}
	else
	{
		// Create new list so we can replace it with old one.
		Asura_EffectVolume_Effect**	pxNewList = new Asura_EffectVolume_Effect* [ m_uNumEffects-1 ];

		// copy from old list objects to new smaller list.  Also delete the one we want to delete
		u_int uCurrent = 0;
		for( u_int u=0; u<m_uNumEffects; u++ )
		{
			// Delete the one we want to delete. Do not copy it.
			if( u==uNumOfEffectToRemove )
			{
				delete m_pxEffects[u];
				m_pxEffects[u] = NULL;
			}
			else	// copy it, because this is one, we don't want to delete.
			{
				pxNewList[uCurrent] = m_pxEffects[u];
				uCurrent++;
			}
		}

		// And now replace old list with new one. Old list can be deleted.
		delete [] m_pxEffects;

		m_pxEffects = pxNewList;

		m_uNumEffects--;
	}
}


Asura_EffectVolume_Effect* Asura_EffectVolume_EntityParameters::GetEffect( u_int uNumOfEffectToGet )
{
	if( uNumOfEffectToGet>=m_uNumEffects )
	{
		Asura_Assert(false, "");
		return NULL;
	}
	else
	{
		return m_pxEffects[uNumOfEffectToGet];
	}
}


void Asura_EffectVolume_EntityParameters::RemoveAllEffects()
{
	if( m_pxEffects )
	{
		for( u_int u=0; u<m_uNumEffects; u++ )
		{
			delete m_pxEffects[u];	m_pxEffects[u] = NULL;
		}

		delete [] m_pxEffects;		m_pxEffects = NULL;
	}

	m_uNumEffects = 0;
}


float Asura_EffectVolume_EntityParameters::CalculateStrengthFromTransitionVolume( const Asura_Vector_3& xPosition  ) const
{
	float fStrength = 0.0f;

	u_int uNumOfVolumes = m_pxTransitionVolumeGroup->GetNumberOfVolumes();

	for( u_int uIndex = 0; uIndex<uNumOfVolumes; uIndex++ )
	{
		Asura_Volume* pxVolume = m_pxTransitionVolumeGroup->GetVolume( uIndex );

		switch( pxVolume->GetClassification() )
		{
		case ASURA_VOLUME_CLASSIFICATION_CUBOID:
			{
				Asura_Volume_Cuboid*	pxCuboid	= static_cast<Asura_Volume_Cuboid*> (pxVolume);

				Asura_Bounding_Box* pxBB = pxCuboid->GetVolumeBox();
				Asura_Quat* pxOri = pxCuboid->GetVolumeOrient();
				
				Asura_Vector_3 xPoint;

				//see if the orientation is being used
				if(pxOri->w == 1.0f )
				{
					xPoint = xPosition;
				}
				else
				{
					// Yes, it has been, so we have to modify the point;

					//need to get position relative to volume
					Asura_Vector_3 xCentre = pxBB->CalculateCentre();
					Asura_Vector_3 xRelative = xPosition - xCentre;

					//and rotate into the space of the volume
					Asura_Quat xOrientInverse = *pxOri;
					xOrientInverse.Invert();
					xRelative = xOrientInverse.RotateVector(xRelative);

					xPoint = xRelative+xCentre;
				}

				if( pxBB->Contains(xPoint) )
				{
					float fMaxZ = pxBB->GetMax().z;
					float fMinZ = pxBB->GetMin().z;

					float fLength = (fMaxZ - fMinZ);

					if( fLength>0.0f )
					{
						float fStr = (xPoint.z - fMinZ) / fLength;

						if( fStr<=1.0f && fStr>fStrength )
						{
							fStrength = fStr;
						}
					}
				}
		
				break;
			}
		case ASURA_VOLUME_CLASSIFICATION_SPHERE:
			{
				Asura_Volume_Sphere* pxSphere = static_cast<Asura_Volume_Sphere*> (pxVolume);

				Asura_Vector_3*	pxPosition = pxSphere->GetPositionPtr();
				float fRadius = *pxSphere->GetRadiusPtr();

				if( fRadius>0.0f )
				{
					float fRadiusSqr = fRadius*fRadius;
					float fToSqr = (*pxPosition-xPosition).MagnitudeSquared();

					float fStr = 1.0f - (fToSqr / fRadiusSqr);

					if( fStr<=1.0f && fStr>fStrength )
					{
						fStrength = fStr;
					}
				}

				break;
			}
		case ASURA_VOLUME_CLASSIFICATION_CYLINDER:
			{
				Asura_Volume_Cylinder*	pxCylinder	= static_cast<Asura_Volume_Cylinder*> (pxVolume);

				Asura_Vector_3*	pxPosition = pxCylinder->GetPositionPtr();
				float*			pfRadius = pxCylinder->GetRadiusPtr();
				float*			pfHeight = pxCylinder->GetHeightPtr();
				Asura_Quat*		pfOri	= pxCylinder->GetVolumeOrient();

				Asura_Vector_3 xLoc;

				//see if the orientation is being used
				if(pfOri->w == 1.0f)
				{
					xLoc = xPosition;
				}
				else
				{
					//need to get position relative to volume
					Asura_Vector_3 xRelative = xPosition - (*pxPosition);

					//and rotate into the space of the volume
					Asura_Quat xOrientInverse = (*pfOri);
					xOrientInverse.Invert();
					xRelative = xOrientInverse.RotateVector(xRelative);

					xLoc = xRelative + (*pxPosition);
				}


				// Check y in range.
				if (xLoc.y > pxPosition->y) return false;
				if (xLoc.y < pxPosition->y - *pfHeight) return false;

				const float fRadiusSqr = (*pfRadius) * (*pfRadius);		
				Asura_Vector_3 xToSqr  = ((*pxPosition) - xLoc);
				xToSqr.y = 0.f;
				const float fToSqr = xToSqr.MagnitudeSquared();

				if (fToSqr < fRadiusSqr && fRadiusSqr!=0.0f ) 
				{
					float fStr = 1.0f - (fToSqr / fRadiusSqr);

					if( fStr<=1.0f && fStr>fStrength )
					{
						fStrength = fStr;
					}
				}

				break;
			}
		default:
			{
				Asura_Assert( false, "Strange type of Asura_Volume" );
				break;
			}
		}
	}

	return fStrength;
}
