//
// Asura_FSFX_System.cpp
//
// Created by JGP 02/02/06
// 
// Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_FSFX

#include "FSFX/Asura_FSFX_System.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_SortAndSearch.h"
#include "FSFX/Asura_FSFX_Template.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_NetMsg.h"
#include "Rendering/Asura_Render.h"


static const u_int	g_uAsuraFSFXHandlerChunkVersion		= 1;
static const u_int	g_uAsuraFSFXHandlerItemChunkVersion	= 2;


///////////////////////////////////////////////////////////////////////////////////////
// FSFX Handler - One per viewport probably
///////////////////////////////////////////////////////////////////////////////////////

void Asura_FSFX_HandlerItem::Reset(void)
{
	// Reset the current data...
	m_fCurrentTime	= 0.0f;
	m_fCurrentValue	= 0.0f;

	m_fInitialValue	= 0.0f;
	m_fSustainValue	= 0.0f;
	m_fFinalValue	= 0.0f;

	m_fAttackEndTime	= 0.0f;
	m_fSustainEndTime	= 0.0f;
	m_fDecayEndTime		= 0.0f;
}

void Asura_FSFX_HandlerItem::Update(const float fTimeStep)
{
	if(m_fCurrentTime < m_fDecayEndTime)
	{
		// Update our timer and see what we should be doing...
		m_fCurrentTime	+= fTimeStep;

		if(m_fCurrentTime < m_fAttackEndTime)
		{
			const float	fParametric	= m_fCurrentTime / m_fAttackEndTime;

			// We're still attacking...
			m_fCurrentValue	= Asura_Maths::Lerp(m_fInitialValue, m_fSustainValue, fParametric);
		}
		else if(m_fCurrentTime < m_fSustainEndTime)
		{
			// Keeping the value constant...
			m_fCurrentValue	= m_fSustainValue;
		}
		else
		{
			const float	fDuration	= m_fDecayEndTime - m_fSustainEndTime;
			const float	fTime		= m_fCurrentTime - m_fSustainEndTime;

			// Do we have a release to use?
			if((fDuration > 0.0f) && (fTime < fDuration))
			{
				const float	fParametric	= fTime / fDuration;

				// We're still attacking...
				m_fCurrentValue	= Asura_Maths::Lerp(m_fSustainValue, m_fFinalValue, fParametric);
			}
			else
			{
				// Ensure that our values are all at the final setting...
				SetValue(m_fFinalValue);
			}
		}
	}
	else
	{
		// Ensure that our current value is correct - needed for instantaneous changes...
		m_fCurrentValue	= m_fFinalValue;
	}
}

void Asura_FSFX_HandlerItem::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	xStream >> uVersion;

	switch(uVersion)
	{
		case 1:
		{
			float	fTarget;
			float	fValue;
			float	fVelocity;

			xStream >> fTarget;
			xStream >> fValue;
			xStream >> fVelocity;

			if(fValue == fTarget)
			{
				SetValue(fTarget);
			}
			else
			{
				const float	fTime	= Asura_Maths::Abs((fTarget - fVelocity) / fVelocity);

				SetTarget(fTarget, fTime);
			}
			break;
		}

		case 2:
		{
			xStream >> m_fCurrentTime;
			xStream >> m_fCurrentValue;

			xStream >> m_fInitialValue;
			xStream >> m_fSustainValue;
			xStream >> m_fFinalValue;

			xStream >> m_fAttackEndTime;
			xStream >> m_fSustainEndTime;
			xStream >> m_fDecayEndTime;
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected version number found.");
			break;
		}
	}		// End of switch.
}

void Asura_FSFX_HandlerItem::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << g_uAsuraFSFXHandlerItemChunkVersion;

	xStream << m_fCurrentTime;
	xStream << m_fCurrentValue;

	xStream << m_fInitialValue;
	xStream << m_fSustainValue;
	xStream << m_fFinalValue;

	xStream << m_fAttackEndTime;
	xStream << m_fSustainEndTime;
	xStream << m_fDecayEndTime;
}

///////////////////////////////////////////////////////////////////////////////////////
// FSFX Handler - One per viewport probably
///////////////////////////////////////////////////////////////////////////////////////

int Asura_FSFX_Handler::CompareTemplates(const Asura_Hash_ID& xTemplate1,
										 const Asura_Hash_ID& xTemplate2)
{

	if(xTemplate1 == xTemplate2)
	{
		return(0);
	}
	else
	{
		Asura_FSFX_Template const*			pxTemplate1	= Asura_FSFX_System::FindTemplate(xTemplate1);
		Asura_FSFX_Template const*			pxTemplate2	= Asura_FSFX_System::FindTemplate(xTemplate2);

		if( (NULL == pxTemplate2) || (NULL == pxTemplate1) )
		{
			Asura_Assert( false, "Compare Template attempting to access NULL pointer!" );
			return(0);
		}

		const int	iPriority1			= pxTemplate1->GetPriority();
		const int	iPriority2			= pxTemplate2->GetPriority();
		const int	iPriorityDifference	= iPriority1 - iPriority2;

		if(iPriorityDifference == 0)
		{
			const int	iTemplate1	= reinterpret_cast<int>(pxTemplate1);
			const int	iTemplate2	= reinterpret_cast<int>(pxTemplate2);

			// As a last resort - sort by pointers...
			return(iTemplate1 - iTemplate2);

		}
		else
		{
			return(iPriorityDifference);
		}
	}
}

Asura_FSFX_Handler::Asura_FSFX_Handler( const Asura_Guid xOwnerGuid )
: m_xOwnerGuid(xOwnerGuid)
, m_xTemplateContainer(CompareTemplates)
{
}


void Asura_FSFX_Handler::Update()
{
	const float	fTime	= Asura_Timers::GetFrameTime();

	for(TemplateIterator xIterator(m_xTemplateContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_HandlerItem*	pxItem	= xIterator.GetCurrent();

		if(pxItem != NULL)
		{
			pxItem->Update(fTime);
		}
	}
}


void Asura_FSFX_Handler::Render() const
{
	if(Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_FSFX))
	{
		for(TemplateConstIterator xIterator(m_xTemplateContainer); !xIterator.Done(); xIterator.Next())
		{
			Asura_FSFX_Template const*			pxTemplate	= xIterator.GetCurrentKey() ? Asura_FSFX_System::FindTemplate(*xIterator.GetCurrentKey()): NULL;
			Asura_FSFX_HandlerItem const*		pxItem		= xIterator.GetCurrent();

			if((pxItem != NULL) && (pxTemplate != NULL))
			{
				const float	fValue	= pxItem->GetValue();

				if(fValue > 0.0f)
				{
					pxTemplate->Render(fValue);
				}
			}
		}

		
		// Some of the FSFX routines change this
		Asura_Render::AllowFullScreenRendering();
	}
}

void Asura_FSFX_Handler::ResetAllTemplates()
{
	m_xTemplateContainer.Reset();
}

void Asura_FSFX_Handler::ResetEnvironmentalTemplates()
{
	for(TemplateIterator xIterator(m_xTemplateContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Template const*			pxTemplate	= xIterator.GetCurrentKey() ? Asura_FSFX_System::FindTemplate(*xIterator.GetCurrentKey()): NULL;

		if((pxTemplate != NULL) && pxTemplate->IsEnvironmentalEffect())
		{
			Asura_FSFX_HandlerItem*	pxItem	= xIterator.GetCurrent();

			if(pxItem != NULL)
			{
				pxItem->Reset();
			}
		}
	}
}

void Asura_FSFX_Handler::ResetValue(const Asura_Hash_ID uNameHash)
{
	Asura_FSFX_HandlerItem*	pxItem	= m_xTemplateContainer.GetData(uNameHash);

	if(pxItem != NULL)
	{
		pxItem->Reset();
	}
}

void Asura_FSFX_Handler::SetValue( const Asura_Hash_ID uNameHash, const float fValue, const float fTime )
{
	const Asura_FSFX_Template*	pxTemplate	= Asura_FSFX_System::FindTemplate(uNameHash);

	if(pxTemplate == NULL)
	{
		Asura_Assert(false, "Cant find FSFX template to SetValue()");
	}
	else
	{
		Asura_FSFX_HandlerItem*	pxItem	= m_xTemplateContainer.GetData(uNameHash);

		if(pxItem == NULL)
		{
			pxItem	= m_xTemplateContainer.AddNode(uNameHash);
		}

		// Only one environmental effect can be active at any time.
		if((fValue > 0.0f) && pxTemplate->IsEnvironmentalEffect())
		{
			ResetEnvironmentalTemplates();
		}

		if(fTime == 0.0f)
		{
			pxItem->SetValue(fValue);
		}
		else
		{
			pxItem->SetTarget(fValue, fTime);
		}
	}
}

bool Asura_FSFX_Handler::GetValue( const Asura_Hash_ID uNameHash, float &fReturnValue  )
{
	Asura_FSFX_HandlerItem*	pxItem	= m_xTemplateContainer.GetData(uNameHash);

	if(pxItem )
	{
		fReturnValue =  pxItem->GetValue();
		return true;
	}

	fReturnValue = 0.0f;
	return false;
}

void Asura_FSFX_Handler::SetAttackSustainDecayValues(const Asura_Hash_ID uNameHash,
													 const float fSustainValue, const float fReleaseValue,
													 const float fAttackDuration, const float fSustainDuration, const float fDecayDuration)
{
	const Asura_FSFX_Template*	pxTemplate	= Asura_FSFX_System::FindTemplate(uNameHash);

	if(pxTemplate == NULL)
	{
		Asura_Assert(false, "Cant find FSFX template to SetAttackSustainDecayValues()");
	}
	else
	{
		Asura_FSFX_HandlerItem*	pxItem	= m_xTemplateContainer.GetData(uNameHash);

		if(pxItem == NULL)
		{
			pxItem	= m_xTemplateContainer.AddNode(uNameHash);
		}

		// Only one environmental effect can be active at any time.
		if(pxTemplate->IsEnvironmentalEffect())
		{
			ResetEnvironmentalTemplates();
		}

		pxItem->SetTargetAttackSustainDecay(fSustainValue, fReleaseValue, fAttackDuration, fSustainDuration, fDecayDuration);
	}
}

void Asura_FSFX_Handler::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	xStream >> uVersion;

	switch(uVersion)
	{
		case 1:
		{
			u_int	uNumTemplates;

			xStream >> uNumTemplates;

			for(u_int u = 0; u < uNumTemplates; u++)
			{
				Asura_Hash_ID				uHash;
				const Asura_FSFX_Template*	pxTemplate;

				xStream >> uHash;
				pxTemplate	= Asura_FSFX_System::FindTemplate(uHash);

				m_xTemplateContainer.AddNode(uHash, xStream);
			}
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected version number found.");
			break;
		}
	}		// End of switch.
}

void Asura_FSFX_Handler::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	const u_int	uNumTemplates	= m_xTemplateContainer.GetNumberOfNodes();	

	xStream << g_uAsuraFSFXHandlerChunkVersion;
	xStream << uNumTemplates;

	for(TemplateConstIterator xIterator(m_xTemplateContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Template const*			pxTemplate	= xIterator.GetCurrentKey() ? Asura_FSFX_System::FindTemplate(*xIterator.GetCurrentKey()): NULL;

		if(pxTemplate != NULL)
		{
			const Asura_FSFX_HandlerItem*	pxItem	= xIterator.GetCurrent();
			const Asura_Hash_ID				uHash	= pxTemplate->GetNameHash();

			xStream << uHash;
			pxItem->WriteToChunkStream(xStream);
		}
	}
}


///////////////////////////////////////////////////////////////////////////////////////
// System
///////////////////////////////////////////////////////////////////////////////////////

Asura_FSFX_System::TemplateContainer	Asura_FSFX_System::s_xTemplateContainer("FSFX");
Asura_FSFX_System::HandlerContainer		Asura_FSFX_System::s_xHandlerContainer;


void Asura_FSFX_System::Initialise()
{
	Platform_Initialise();
	PlatformProject_Initialise();
}


/*
	Static function to destroy
	the instance of the system
*/
void Asura_FSFX_System::Shutdown()
{
	PlatformProject_Shutdown();
	Platform_Shutdown();

	GetResourceSet().Reset();
}


void Asura_FSFX_System::Reset()
{
	PlatformProject_Reset();
	Platform_Reset();

	// Do not clear templates when we are in tools. Keep them in memory all the time. Including time in the game.
#ifndef ASURA_TOOLS
//	ClearTemplates();
#endif
}


///////////////////////////////////////////////////////////////////////////////////////
// Handlers
///////////////////////////////////////////////////////////////////////////////////////

void Asura_FSFX_System::ClearAllHandlerTemplates(void)
{
	for(HandlerIterator xIterator(s_xHandlerContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Handler**	ppxHandler	= xIterator.GetCurrent();
		Asura_FSFX_Handler*		pxHandler	= (ppxHandler == NULL) ? NULL : *ppxHandler;

		if(pxHandler != NULL)
		{
			pxHandler->ResetAllTemplates();
		}
	}
}

Asura_FSFX_Handler* Asura_FSFX_System::AddHandler( const Asura_Guid xOwnerGuid )
{
	Asura_FSFX_Handler*	pxHandler	= FindHandler(xOwnerGuid);
	
	if(pxHandler == NULL)
	{
		pxHandler	= new Asura_FSFX_Handler (xOwnerGuid);

		s_xHandlerContainer.Push_Back(pxHandler);
		return(pxHandler);
	}
	else
	{
		Asura_Assert(false, "Handler already exists");
		return(NULL);
	}
}


Asura_FSFX_Handler* Asura_FSFX_System::FindHandler( const Asura_Guid xOwnerGuid )
{
	for(HandlerIterator xIterator(s_xHandlerContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Handler**	ppxHandler	= xIterator.GetCurrent();
		Asura_FSFX_Handler*		pxHandler	= (ppxHandler == NULL) ? NULL : *ppxHandler;

		if((pxHandler != NULL) && (pxHandler->GetOwnerGuid() == xOwnerGuid) )
		{
			return(pxHandler);
		}
	}
	return NULL;
}


void Asura_FSFX_System::RemoveHandler( Asura_FSFX_Handler* const pxHandler )
{
	HandlerContainerElement*	pxFound	= s_xHandlerContainer.Find(pxHandler);

	if(pxFound == NULL)
	{
		Asura_Assert(false, "Cant find FSFX handler to remove");
	}
	else
	{
		delete(pxFound->Data());
		s_xHandlerContainer.Remove(pxFound);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FSFX_System::SetValueForAllHandlers
// Author       : Andy Tate
// Date         : 26-Jun-2008
////////////////////////////////////////////////////////////////////////////////

void Asura_FSFX_System::SetValueForAllHandlers(const Asura_Hash_ID uTemplateHash,
											   const float fValue,
											   const float fTime)
{
	for(HandlerIterator xIterator(s_xHandlerContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Handler**	ppxHandler	= xIterator.GetCurrent();
		Asura_FSFX_Handler*		pxHandler	= (ppxHandler == NULL) ? NULL : *ppxHandler;

		if(pxHandler != NULL)
		{
			pxHandler->SetValue(uTemplateHash, fValue, fTime);
		}
	}
}

/// End of function 'Asura_FSFX_System::SetValueForAllHandlers'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FSFX_System::GetFirstValueFromAllHandlers
// Author       : Emil Popovic
// Date         : 30-Jan-2009
////////////////////////////////////////////////////////////////////////////////

float Asura_FSFX_System::GetFirstValueFromAllHandlers(const Asura_Hash_ID uTemplateHash )
{
	for(HandlerIterator xIterator(s_xHandlerContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Handler**	ppxHandler	= xIterator.GetCurrent();
		Asura_FSFX_Handler*		pxHandler	= (ppxHandler == NULL) ? NULL : *ppxHandler;

		if(pxHandler != NULL)
		{
			float fReturnValue;
			bool bSuccess = pxHandler->GetValue(uTemplateHash, fReturnValue );

			if( bSuccess )
			{
				return fReturnValue;
			}
		}
	}

	return 0.0f;
}

/// End of function 'Asura_FSFX_System::SetValueForAllHandlers'. ///


///////////////////////////////////////////////////////////////////////////////////////
// Update / Render
///////////////////////////////////////////////////////////////////////////////////////


void Asura_FSFX_System::Update()
{
	for(HandlerIterator xIterator(s_xHandlerContainer); !xIterator.Done(); xIterator.Next())
	{
		Asura_FSFX_Handler**	ppxHandler	= xIterator.GetCurrent();
		Asura_FSFX_Handler*		pxHandler	= (ppxHandler == NULL) ? NULL : *ppxHandler;

		if(pxHandler != NULL)
		{
			pxHandler->Update();
		}
	}
}


void Asura_FSFX_System::Render()
{
	Platform_BeginRender();
	{
		for(HandlerIterator xIterator(s_xHandlerContainer); !xIterator.Done(); xIterator.Next())
		{
			Asura_FSFX_Handler**	ppxHandler	= xIterator.GetCurrent();
			Asura_FSFX_Handler*		pxHandler	= (ppxHandler == NULL) ? NULL : *ppxHandler;

			if(pxHandler != NULL)
			{
				pxHandler->Render();
			}
		}
	}
	Platform_EndRender();
}


///////////////////////////////////////////////////////////////////////////////////////
// Templates
///////////////////////////////////////////////////////////////////////////////////////

// void Asura_FSFX_System::ClearTemplates()
// {
// 	s_xTemplateContainer.Reset();
// }

bool Asura_FSFX_System::AddTemplate( Asura_FSFX_Template* const pxTemplate )
{
	if(pxTemplate == NULL)
	{
		Asura_Assert(false, "No template specified when adding to FSFX system.");
		return(false);
	}
	else
	{
		const Asura_Hash_ID	uTemplateHash	= pxTemplate->GetNameHash();

		//see if the template already exists
		Asura_FSFX_Template* pxExistingTemplate = s_xTemplateContainer.GetResource(uTemplateHash);
		if(pxExistingTemplate)
		{
			//Need to ditch the new template , and add a reference to the existing one then
			delete pxTemplate;
			//Return value for the function is slightly questionable in this case , since a template
			//may have been added to the current bank - just not the one that was requested
			return s_xTemplateContainer.AddResource(uTemplateHash, *pxExistingTemplate, false);
		}
		else
		{
			//add the new template
			return s_xTemplateContainer.AddResource(uTemplateHash, *pxTemplate, true);
		}

	}
}


#ifdef ASURA_TOOLS

Asura_FSFX_Template* Asura_FSFX_System::GetTemplateByIndex( const u_int uIndex )
{
	return(s_xTemplateContainer.GetResourceByIndex( uIndex ));
}

Asura_FSFX_Template* const	Asura_FSFX_System::FindTemplate( const char* const szName )
{
	Asura_FSFX_Template* const pxExistingTemplate =  FindTemplate( Asura_GetHashID( szName ) );
	return pxExistingTemplate;
}
Asura_FSFX_Template* const	Asura_FSFX_System::FindTemplateWithoutOverride( const char* const szName )
{
	Asura_FSFX_Template* const pxExistingTemplate =  FindTemplateWithoutOverride( Asura_GetHashID( szName ) );
	return pxExistingTemplate;
}

void Asura_FSFX_System::DeleteTemplate( const Asura_Hash_ID uResourceHash )
{
	ClearAllHandlerTemplates();	// Make sure that we get rid of all Handlers, because handlers have pointers to templates. So if we don't do this, one pointer will become invalid and it will cause crash of the system. So this is cheap trick, by simply killing all handler data.

	s_xTemplateContainer.DeleteResource( uResourceHash );
}

void Asura_FSFX_System::RelocateTemplate(const Asura_Hash_ID uOldResourceHash, const Asura_Hash_ID uNewResourceHash)
{
	ClearAllHandlerTemplates();	// Make sure that we get rid of all Handlers, because handlers have pointers to templates. So if we don't do this, one pointer will become invalid and it will cause crash of the system. So this is cheap trick, by simply killing all handler data.

	s_xTemplateContainer.RelocateResource( uOldResourceHash, uNewResourceHash );
}

Asura_FSFX_Template* Asura_FSFX_System::FindTemplate(const Asura_Hash_ID uTemplateHash)
{
	Asura_FSFX_Template* pxTemplate	= s_xTemplateContainer.GetResource(uTemplateHash);
	return pxTemplate;
}
Asura_FSFX_Template* Asura_FSFX_System::FindTemplateWithoutOverride(const Asura_Hash_ID uTemplateHash)
{
	Asura_FSFX_Template* pxTemplate	= s_xTemplateContainer.GetResourceWithoutOverride(uTemplateHash);
	return pxTemplate;
}


void Asura_FSFX_System::GetTemplateNames(Asura_String* &pszNameArray, int &iNameSize)
{
	pszNameArray = new Asura_String[iNameSize = s_xTemplateContainer.GetNumResources()];
	
	int i = 0;
	for(Asura_FSFX_Template_ResourceSet_ResourceIterator xIter(s_xTemplateContainer); !xIter.Done(); xIter.Next(), ++i)
	{
		pszNameArray[i] = Asura_String(xIter.GetCurrent()->GetName());
	}
}


#else	// ASURA_TOOLS

const Asura_FSFX_Template* Asura_FSFX_System::FindTemplate(const Asura_Hash_ID uTemplateHash)
{
	const Asura_FSFX_Template* pxTemplate = s_xTemplateContainer.GetResource(uTemplateHash);
	return pxTemplate;
}

#endif	// ASURA_TOOLS








///////////////////////////////////////////////////////////////////////////////////////
// Message registration
///////////////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_NETMSG_PACK_ALL_DATA
void Asura_FSFX_Message::RegisterFields(Asura_NetMsg* pxNetMsg) const
{
	Asura_Assert(pxNetMsg->GetMsgID() == ASURA_MSG_SET_FSFX_NEW, "unexpected message ID");

	Asura_NetMsg_RegisterFieldSimple(Asura_FieldType_uint, m_xID);
	Asura_NetMsg_RegisterFieldSimple(Asura_FieldType_float, m_fValue);
	Asura_NetMsg_RegisterFieldSimple(Asura_FieldType_float, m_fTime);

}
#endif // #ifdef ASURA_NETMSG_PACK_ALL_DATA
#endif // ASURA_USE_FSFX


/// End of file. ///
