//
// Asura_FSFX_Template.cpp
//
// Created by JGP 02/02/06
// 
// Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_FSFX

#include "Asura_FSFX_Template.h"
#include "Asura_FSFX_Component.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"

///////////////////////////////////////////////////////////////////////////////////////
// create functions and construct / destruct
///////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS

/*
	Copy only valid in tools build
*/
Asura_FSFX_Template* Asura_FSFX_Template::CreateCopy() const
{
	Asura_FSFX_Template* const pxNew = new Asura_FSFX_Template;
	
	// copy in data

	pxNew->m_xName = m_xName;
	pxNew->m_xNameHash = m_xNameHash;
	pxNew->m_iPriority = m_iPriority;
	pxNew->m_bIsEnvironmentalEffect = m_bIsEnvironmentalEffect;

	for ( Asura_FSFX_ComponentList* pxList = m_xComponentList.Front(); pxList; pxList = pxList->Next() )
	{
		Asura_FSFX_Component* pxComponent = pxList->Data();
		Asura_FSFX_Component* pxNewComponent = pxComponent->CreateCopy();
		pxNew->m_xComponentList.Push_Back( pxNewComponent );
	}

	return pxNew;
}

#endif // ASURA_TOOLS


Asura_FSFX_Template::Asura_FSFX_Template()
{
	m_iPriority = DEFAULT_PRIORITY;
	m_bIsEnvironmentalEffect = true;
}


Asura_FSFX_Template::~Asura_FSFX_Template()
{
	ClearComponentList();
}


///////////////////////////////////////////////////////////////////////////////////////
// Render
///////////////////////////////////////////////////////////////////////////////////////


/*
	Render each component in turn
*/
void Asura_FSFX_Template::Render( const float fIntensity ) const
{

	const Asura_FSFX_Component::Platform ePlatform = static_cast<Asura_FSFX_Component::Platform>( Platform_GetPlatformFlag() );

	for ( Asura_FSFX_ComponentList* pxList = m_xComponentList.Front(); pxList; pxList = pxList->Next() )
	{
		Asura_FSFX_Component* const pxComp = pxList->Data();
		if ( pxComp->IsPlatformEnabled( ePlatform ) )
		{
			pxComp->Platform_Render( fIntensity );
		}
	}
}


/*
	Clear out the component list
*/
void Asura_FSFX_Template::ClearComponentList()
{
	for ( Asura_FSFX_ComponentList* pxList = m_xComponentList.Front(); pxList; pxList = pxList->Next() )
	{
		delete pxList->Data();
	}
	m_xComponentList.Clear();
}


///////////////////////////////////////////////////////////////////////////////////////
// component setup / delete and name setting (All tools only functions)
///////////////////////////////////////////////////////////////////////////////////////


#ifdef ASURA_TOOLS

void Asura_FSFX_Template::SetName( const char* const szName )
{
	m_xName = szName;
	m_xNameHash = Asura_GetHashID( szName );
}


Asura_FSFX_ComponentList* Asura_FSFX_Template::GetComponentList() const
{
	return m_xComponentList.Front();
}


void Asura_FSFX_Template::AddComponent( Asura_FSFX_Component* pxComponent )
{
	m_xComponentList.Push_Back( pxComponent );
}


void Asura_FSFX_Template::DeleteComponent( Asura_FSFX_Component* pxComponent )
{
	Asura_FSFX_ComponentList* pxComponentListEntry = m_xComponentList.Find( pxComponent );
	m_xComponentList.Remove( pxComponentListEntry );
	
	// and delete the template
	delete pxComponent;
}


void Asura_FSFX_Template::SwapComponents( Asura_FSFX_Component* pxComponent1, Asura_FSFX_Component* pxComponent2 )
{

	// find both the entries

	Asura_FSFX_ComponentList* pxComponent1ListEntry = m_xComponentList.Find( pxComponent1 );
	Asura_FSFX_ComponentList* pxComponent2ListEntry = m_xComponentList.Find( pxComponent2 );

	// add new entries after both the existing entries (swapping)

	m_xComponentList.Insert_After( pxComponent1ListEntry, pxComponent2 );
	m_xComponentList.Insert_After( pxComponent2ListEntry, pxComponent1 );

	// now remove the entries

	m_xComponentList.Remove( pxComponent1ListEntry );
	m_xComponentList.Remove( pxComponent2ListEntry );

}

#endif // ASURA_TOOLS


///////////////////////////////////////////////////////////////////////////////////////
// save / load
///////////////////////////////////////////////////////////////////////////////////////


void Asura_FSFX_Template::ReadFromChunkStream( Asura_Chunk_Stream& xStream, const int iVersion )
{

	// clear out template
	ClearComponentList();

	// name hash
	xStream >> m_xNameHash;

	// get name
#ifdef ASURA_TOOLS
	// in tools we read in the name...
	xStream >> m_xName;
#else
	// otherwise we skip it (we've got the hash)
	xStream.SkipString();
#endif // ASURA_TOOLS

	// Version two added support for priority
	if ( iVersion >= 2 )
	{
		xStream >> m_iPriority;
	}

	// Version three: environmental effects are distinguished from non-environmental effects.
	if( iVersion >= 3 )
	{
		xStream >> m_bIsEnvironmentalEffect;
	}

	// get number of components
	int iNumComponents;
	xStream >> iNumComponents;

	for ( int i = 0; i < iNumComponents; i++ )
	{

		// get component type
		int iType;
		xStream >> iType;

		// create component
		Asura_FSFX_Component* const pxComponent = Asura_FSFX_Component::CreateOfType( static_cast<Asura_FSFX_Component::Type>( iType ) );

		// if component wasn't created, do a warning and skip the block
		if ( !pxComponent )
		{
			Asura_Assert( 0, "FSFX Component type not recognised. Skipping block" );
			
			// read header which is standard on each component
			// obviously if the data is so wrong that it doesn't
			// have a header we're not going to be able to recover
			// but worth a try anyway...
			int iComponentVersion, iSize, iSupportedPlatforms;
			xStream >> iComponentVersion;
			xStream >> iSize;
			xStream >> iSupportedPlatforms;

			// skip size bytes
			xStream.Skip( iSize );
		}
		// otherwise trust the component to load itself
		else
		{
			pxComponent->Deserialize( xStream );
			
			// and add it
			m_xComponentList.Push_Back( pxComponent );
		}

	}

}


#ifdef ASURA_TOOLS

void Asura_FSFX_Template::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{

	xStream << m_xNameHash;

	xStream << m_xName;

	xStream << m_iPriority;

	xStream << m_bIsEnvironmentalEffect;

	// count the number of components
	int iNumComponents = 0;
	for ( Asura_FSFX_ComponentList* pxList = m_xComponentList.Front(); pxList; pxList = pxList->Next() )
	{
		iNumComponents++;
	}

	xStream << iNumComponents;

	// now save per component

	for ( Asura_FSFX_ComponentList* pxList = m_xComponentList.Front(); pxList; pxList = pxList->Next() )
	{
		Asura_FSFX_Component* const pxComponent = pxList->Data();
	
		// save the type as an int

		int iType = static_cast<int>( pxComponent->GetType() );
		xStream << iType;

		// now save the component

		pxComponent->Serialize( xStream );

	}

}

#endif // ASURA_TOOLS


#endif // ASURA_USE_FSFX
