
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_FacialAnimation_AnimationHandler.h
// Author : Andy Tate
// Date   : 8-May-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_FACIALANIMATION_ANIMATIONHANDLER_H__
#define __ASURA_FACIALANIMATION_ANIMATIONHANDLER_H__

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_HashTree.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation;
class Asura_AnimationController_Group;
class Asura_Dialogue_Line;
class Asura_FacialAnimation_BoneDataModifier;

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_FacialAnimation_FaceAnimationInstance
// Author       : Andy Tate
// Date         : 26-Aug-2008
//
// Description  : Class used to store an instance of a face-animation - used
//				  playing the animations under the facial-animation system.
////////////////////////////////////////////////////////////////////////////////

class Asura_FacialAnimation_FaceAnimationInstance
{
public:
	// Public functions...
	Asura_FacialAnimation_FaceAnimationInstance(void)	: m_uFaceAnimationHash(ASURA_HASH_ID_UNSET), m_fTimePlayed(0.0f), m_piSoundHandle(NULL), m_uSoundHash(ASURA_HASH_ID_UNSET), m_bManualTimerUpdate(false), m_bLoopedAnim(false)	{}
	//~Asura_FacialAnimation_FaceAnimationInstance();

	// Manipulation functions...
	void	Initialise(const Asura_Hash_ID uFaceAnimationHash, bool bLooped = false);
	void	Initialise(const Asura_Hash_ID uFaceAnimationHash, Asura_Handle& iSoundHandle, const Asura_Hash_ID uSoundHash);
	bool	Update(const float fTimeStep, float* pfWeights, float* pfBalances, float* pfEyeValues);

	//This causes an instance of a facial animation to ignore fTimeStep during the updates , and instead
	//use the most recent timer set by this function
	void SetManualTimerUpdate(bool bSet, float fTimer);


	// Accessor functions...
	Asura_Hash_ID		GetFaceAnimationHash(void) const	{ return(m_uFaceAnimationHash); }
	float				GetTimePlayed(void) const			{ return(m_fTimePlayed); }
	
	bool				HasSound(void) const		{ return(m_piSoundHandle != NULL); }
	const Asura_Handle*	GetSoundHandle(void) const	{ return(m_piSoundHandle); }
	Asura_Hash_ID		GetSoundHash(void) const	{ return(m_uSoundHash); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private variables...
	Asura_Hash_ID	m_uFaceAnimationHash;	// The hash-ID of the face-animation to play.
	float			m_fTimePlayed;			// The number of seconds we've been playing this animation for.
	Asura_Handle*	m_piSoundHandle;		// A pointer to the handle used when playing a sound.
	Asura_Hash_ID	m_uSoundHash;			// The hash-ID of the sound played.
	bool			m_bManualTimerUpdate;	// Timer is set directly , so fTimerStep in Update() should be ignored
	bool			m_bLoopedAnim;

	// Private functions...
	Asura_FacialAnimation_FaceAnimationInstance(const Asura_FacialAnimation_FaceAnimationInstance& xOther);

	Asura_FacialAnimation_FaceAnimationInstance&	operator=(const Asura_FacialAnimation_FaceAnimationInstance& xOther);

};   // End of class 'Asura_FacialAnimation_FaceAnimationInstance'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_FacialAnimation_AnimationHandler
// Author       : Andy Tate
// Date         : 8-May-2008
//
// Description  : Abstract class used to interface between the facial-animation
//				  and the "Asura_Animation" it's being played on.
////////////////////////////////////////////////////////////////////////////////

class Asura_FacialAnimation_AnimationHandler
{
public:

	typedef Asura_TemplateHashTree<Asura_FacialAnimation_FaceAnimationInstance> FaceAnimationInstanceContainer;

	// Public functions...
	virtual ~Asura_FacialAnimation_AnimationHandler();

	// Abstract functions...
	virtual	void	Update(const float fTimeStep)	= 0;

	// Manipulation functions...
	bool	PlayFaceAnimation(const Asura_Hash_ID uFaceAnimationHash, bool bLooped = false);
	bool	PlayFaceAnimation(const Asura_Dialogue_Line& xDialogueLine, Asura_Handle& iSoundHandle);
	bool	StopFaceAnimation(const Asura_Hash_ID uFaceAnimationHash) {return m_xFaceAnimationInstanceContainer.DeleteNodeByHash(uFaceAnimationHash);}
	void	StopAllFaceAnimations()	{ m_xFaceAnimationInstanceContainer.Reset();}

	void	SetExternalEyeData(const float fLeftRight, const float fUpDown, const float fFocus, const float fClose, const float fSquint, const float fBalance);

	//This causes an instance of a facial animation to ignore fTimeStep during the updates , and instead
	//use the most recent timer set by this function
	void	SetManualTimerUpdate(Asura_Hash_ID uFaceAnimationHash, bool bSet, float fTimer);

	// Accessor functions...
	void	ZeroWeightsAndBalances(void);

	void	SetActive(const bool bActive);
	bool	IsActive(void) const	{ return(m_bActive); }

	Asura_Hash_ID	GetFaceHash(void) const	{ return(m_uFaceHash); }

	bool	IsPlayingFaceAnimation(void) const	{ return(m_xFaceAnimationInstanceContainer.GetNumberOfNodes() > 0); }

	u_int	GetNumPlayingFaceAnimations( void ) const { return m_xFaceAnimationInstanceContainer.GetNumberOfNodes(); }

	FaceAnimationInstanceContainer* GetFaceAnimInstanceContainer( void ) { return &m_xFaceAnimationInstanceContainer; }

#if defined(ASURA_TOOLS)
	// Tools-only accessor functions...
	void Tools_SetWeightAndBalance(const u_int uIndex, const float fWeight, const float fBalance)
	{
		if(m_pfMuscleGroupAnimationWeights != NULL)
		{
			m_pfMuscleGroupAnimationWeights[uIndex]	= fWeight;
		}

		if(m_pfMuscleGroupAnimationBalances != NULL)
		{
			m_pfMuscleGroupAnimationBalances[uIndex]	= fBalance;
		}
	}

	float	Tools_GetWeight(const u_int uIndex) const	{ return((m_pfMuscleGroupAnimationWeights == NULL) ? 0.0f : m_pfMuscleGroupAnimationWeights[uIndex]); }
	float	Tools_GetBalance(const u_int uIndex) const	{ return((m_pfMuscleGroupAnimationBalances == NULL) ? 0.0f : m_pfMuscleGroupAnimationBalances[uIndex]); }

	float*	Tools_GetWeightPointer(const u_int uIndex)	{ return((m_pfMuscleGroupAnimationWeights == NULL) ? NULL : &m_pfMuscleGroupAnimationWeights[uIndex]); }
	float*	Tools_GetBalancePointer(const u_int uIndex)	{ return((m_pfMuscleGroupAnimationBalances == NULL) ? NULL : &m_pfMuscleGroupAnimationBalances[uIndex]); }

	float*	Tools_GetEyeValuePointer(const u_int uIndex);
	float	Tools_GetEyeValue(const u_int uIndex) const;
#endif	//defined(ASURA_TOOLS)

	

protected:
	// Protected variables...

	// Protected functions...
	Asura_FacialAnimation_AnimationHandler(const Asura_Hash_ID uFaceHash, const char* szAnimationGroupName, const u_int uAnimationGroupPriority, const Asura_Hash_ID uParentAnimationGroupHash = ASURA_HASH_ID_UNSET);

	// Event functions...
	virtual	void	OnActivation(void);
	virtual void	OnDeactivation(void);

	// Manipulation functions...
	Asura_AnimationController_Group*	CreateAnimationGroup(Asura_Animation& xAnimation);
	bool								DestroyAnimationGroup(Asura_Animation& xAnimation);

	void								UpdateAnimation(const float fTimeStep, Asura_Animation& xAnimation);


private:
	// Private typedefs...
	typedef Asura_TemplateHashTree_Iterator<Asura_FacialAnimation_FaceAnimationInstance> FaceAnimationInstanceIterator;
	typedef Asura_TemplateHashTree_ConstIterator<Asura_FacialAnimation_FaceAnimationInstance> FaceAnimationInstanceConstIterator;

	// Private variables...
	Asura_Hash_ID					m_uFaceHash;						// The hash-ID of the face to use.
	Asura_FacialAnimation_BoneDataModifier*	m_pxBoneDataModifier;		// pointer to our bone data modifier( will be created in UpdateAnimation )

	const char*						m_szAnimationGroupName;				// The name of the animation-group used to stored the facial-animation data.
	Asura_Hash_ID					m_uAnimationGroupHash;				// The hash-ID of the animation-group used to stored the facial-animation data.
	u_int							m_uAnimationGroupPriority;			// The priority of the animation-group to be created to contain the facial-animation.
	Asura_Hash_ID					m_uParentAnimationGroupHash;		// The hash-ID of the parent animation-group used to create our animation within.

	u_int							m_uNumMuscleGroupAnimations;		// The number of muscle-groups within this facial-animation..
	float*							m_pfMuscleGroupAnimationWeights;	// A pointer to the array of weights to use for the muscle-groups for this facial-animation.
	float*							m_pfMuscleGroupAnimationBalances;	// A pointer to the array of parametric-data to use for the animations within the muscle-groups for this facial-animation.
	float*							m_pfAnimationEyeValues;				// Pointer to array of eye values calculate from the animations

	bool							m_bActive;							// Flag to say that this animation-handler is active.

	float							m_fExternalEyeLeftRight;
	float							m_fExternalEyeUpDown;
	float							m_fExternalEyeFocus;
	float							m_fExternalEyeClose;
	float							m_fExternalEyeSquint;
	float							m_fExternalEyeBalance;

	FaceAnimationInstanceContainer	m_xFaceAnimationInstanceContainer;	// A container holding the instances of face-animations we're currently playing.

	// Private functions...
	Asura_FacialAnimation_AnimationHandler(void);
	Asura_FacialAnimation_AnimationHandler(const Asura_FacialAnimation_AnimationHandler& xOther);

	Asura_FacialAnimation_AnimationHandler&	operator=(const Asura_FacialAnimation_AnimationHandler& xOther);

};   // End of class 'Asura_FacialAnimation_AnimationHandler'.

#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)

#endif   // ifndef __ASURA_FACIALANIMATION_ANIMATIONHANDLER_H__


/// End of file. ///
