//
// Asura_FacialAniamtion_EyeParamAnimation.cpp
//
// Created by Richard Rice - 15/10/2008
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)

#include "FacialAnimation/Asura_FacialAnimation_EyeParamAnimation.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"

const u_int	Asura_FacialAnimation_EyeParamAnimation::s_uChunkVersion = 1;


Asura_FacialAnimation_EyeParamAnimation::Asura_FacialAnimation_EyeParamAnimation(void)
: m_uNumKeyFrames(0)
, m_pxKeyFrames(NULL)
{
}

Asura_FacialAnimation_EyeParamAnimation::~Asura_FacialAnimation_EyeParamAnimation()
{
	delete[](m_pxKeyFrames);
}


#if defined(ASURA_TOOLS)


void Asura_FacialAnimation_EyeParamAnimation::Tools_AllocateKeyFrames(const u_int uNumKeyFrames)
{
	delete [] m_pxKeyFrames;
	m_uNumKeyFrames = uNumKeyFrames;
	m_pxKeyFrames = m_uNumKeyFrames ? new Asura_FacialAnimation_EyeParamAnimation_KeyFrame[m_uNumKeyFrames] : 0;
}

void Asura_FacialAnimation_EyeParamAnimation::Tools_SetKeyFrame(const u_int uFrame, const float fTime, const float fValue )
{
	Asura_Assert(uFrame < m_uNumKeyFrames, "Invalid keyframe index");
	if(uFrame >= m_uNumKeyFrames) return;

	m_pxKeyFrames[uFrame].Set(fTime, fValue);

}

void Asura_FacialAnimation_EyeParamAnimation::Tools_SetKeyFrames(const Asura_FacialAnimation_EyeParamAnimation& xAnimation)
{
	//delete any previous keyframes
	delete[](m_pxKeyFrames);

	//allocate the new keyframes
	m_uNumKeyFrames = xAnimation.m_uNumKeyFrames;
	m_pxKeyFrames = m_uNumKeyFrames ? new Asura_FacialAnimation_EyeParamAnimation_KeyFrame[m_uNumKeyFrames] : 0;

	//and copy from the tools list of keyframes

	for(u_int i=0; i<m_uNumKeyFrames; i++)
	{
		m_pxKeyFrames[i] = xAnimation.m_pxKeyFrames[i];
	}
}

float Asura_FacialAnimation_EyeParamAnimation::Tools_GetLastKeyFrameTime() const
{
	if(m_uNumKeyFrames)
	{
		return m_pxKeyFrames[m_uNumKeyFrames-1].GetTime();
	}
	else
	{
		return 0.0f;
	}
}

void Asura_FacialAnimation_EyeParamAnimation::Tools_CombineBaseFacePose( float fBaseFacePoseValue )
{
	// Method for combining BaseFacePose is: Add and clamp.
	for( u_int u=0; u<m_uNumKeyFrames; u++ )
	{
		const float fTime = m_pxKeyFrames[u].GetTime();
		const float fValue= m_pxKeyFrames[u].GetValue();

		float fNewValue = Asura_Maths::ClampToRange( fValue+fBaseFacePoseValue, -1.0f, 1.0f );

		m_pxKeyFrames[u].Set( fTime, fNewValue );
	}
}

void Asura_FacialAnimation_EyeParamAnimation::Tools_RemoveRedundantKeyFrames()
{
	const float fTolerance = 0.001f;

	u_int uToPos = 1;
	u_int uFromPos = 1;
	for(; uFromPos<m_uNumKeyFrames; uFromPos++)
	{
		
		const Asura_FacialAnimation_EyeParamAnimation_KeyFrame* pxPrevKeyFrame = &m_pxKeyFrames[uToPos-1];
		const Asura_FacialAnimation_EyeParamAnimation_KeyFrame* pxThisKeyFrame = &m_pxKeyFrames[uFromPos];
		const Asura_FacialAnimation_EyeParamAnimation_KeyFrame* pxNextKeyFrame = (uFromPos < m_uNumKeyFrames-1) ? &m_pxKeyFrames[uFromPos+1] : 0;

		bool bDeleteCurrent = false;
		if(pxPrevKeyFrame && pxThisKeyFrame && pxNextKeyFrame)
		{
			//Work out the value that would be got by interpolating between the previouos and next keyframes
			float fPrevTime = pxPrevKeyFrame->GetTime();
			float fNextTime = pxNextKeyFrame->GetTime();
			float fThisTime = pxThisKeyFrame->GetTime();

			if(fNextTime != fPrevTime)
			{
				float fLerp = (fThisTime - fPrevTime) / (fNextTime-fPrevTime);

				float fInterpValue = (1-fLerp)*pxPrevKeyFrame->GetValue() + fLerp*pxNextKeyFrame->GetValue();

				//Is the interpolated value close enough?
				float fValueDiff = pxThisKeyFrame->GetValue()-fInterpValue;
				
				if(Asura_Maths::Abs(fValueDiff)<fTolerance)
				{
					//can remove the middle frame
					bDeleteCurrent = true;
				}
			}
		}

		if(!bDeleteCurrent)
		{
			//Not deleting the keyframe , so copy it down to the current 'To' position
			if(uFromPos != uToPos)
			{
				m_pxKeyFrames[uToPos] = m_pxKeyFrames[uFromPos];
			}
			uToPos++;
		}		
	}
	m_uNumKeyFrames = uToPos;

	//If there are only 2 frames left , check to see if they are identical, if so , then remove the second
	if(m_uNumKeyFrames == 2)
	{
		const Asura_FacialAnimation_EyeParamAnimation_KeyFrame* pxFirstKeyFrame = &m_pxKeyFrames[0];	
		const Asura_FacialAnimation_EyeParamAnimation_KeyFrame* pxLastKeyFrame = &m_pxKeyFrames[1];	

		if(pxFirstKeyFrame->GetValue() == pxLastKeyFrame->GetValue())
		{
			m_uNumKeyFrames = 1;
		}
	}
}


#endif	//defined(ASURA_TOOLS)



bool Asura_FacialAnimation_EyeParamAnimation::CalculateValue(const float fTime, float& fValue) const
{
	const Asura_FacialAnimation_EyeParamAnimation_KeyFrame*	pxCurrentKeyFrame	= m_pxKeyFrames;

	if(fTime < pxCurrentKeyFrame->GetTime())
	{
		// Time is before the start of the animation...
		return(false);
	}
	else
	{
		fValue		= pxCurrentKeyFrame->GetValue();

		const Asura_FacialAnimation_EyeParamAnimation_KeyFrame*	pxPreviousKeyFrame	= NULL;

		// Find the key-frames which 
		for(u_int u = 0; u < m_uNumKeyFrames; u++, pxCurrentKeyFrame++)
		{
			const float	fKeyFrameTime	= pxCurrentKeyFrame->GetTime();

			if(fTime <= fKeyFrameTime)
			{
				if( pxPreviousKeyFrame )
				{
					const float	fPreviousTime	= pxPreviousKeyFrame->GetTime();
					const float	fParamteric		= (fTime - fPreviousTime) / (fKeyFrameTime - fPreviousTime);

					fValue		= Asura_Maths::Lerp(pxPreviousKeyFrame->GetValue(), pxCurrentKeyFrame->GetValue(), fParamteric);
				}

				return(true);
			}
			else
			{
				pxPreviousKeyFrame	= pxCurrentKeyFrame;
			}
		}

		// Time MUST be beyond the end of the animation...
		return(true);
	}
}

////////////////////////////////////////////////////////////////////////////////

void Asura_FacialAnimation_EyeParamAnimation::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	// First, read (and verify) the chunk-version.
	xStream >> uVersion;

	switch(uVersion)
	{
		case 1:
		{
			Asura_FacialAnimation_EyeParamAnimation_KeyFrame*	pxKeyFrame;

			xStream >> m_uNumKeyFrames;

			m_pxKeyFrames	= new Asura_FacialAnimation_EyeParamAnimation_KeyFrame [m_uNumKeyFrames];
			pxKeyFrame		= m_pxKeyFrames;

			for(u_int u = 0; u < m_uNumKeyFrames; u++, pxKeyFrame++)
			{
				float	fTime;
				float	fValue;

				xStream >> fTime;
				xStream >> fValue;

				pxKeyFrame->Set(fTime, fValue);
			}
			break;
		}

		default:
		{
			Asura_Assertf(false, ("Unexpected chunk-version %d", uVersion));
			break;
		}
	}		// End of switch.
}

void Asura_FacialAnimation_EyeParamAnimation::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	const Asura_FacialAnimation_EyeParamAnimation_KeyFrame*	pxKeyFrame	= m_pxKeyFrames;

	// First write the chunk-version...
	xStream << s_uChunkVersion;
	xStream << m_uNumKeyFrames;

	for(u_int u = 0; u < m_uNumKeyFrames; u++, pxKeyFrame++)
	{
		const float	fTime		= pxKeyFrame->GetTime();
		const float	fValue		= pxKeyFrame->GetValue();

		xStream << fTime;
		xStream << fValue;
	}
}


#endif //defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)