
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_FileCache.cpp
// Author : Andy Tate
// Date   : 26-Apr-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "FileAccess/Asura_FileCache.h"

// Asura independent header-files...
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

//#define ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE	1

#if 1
#define LOG_OUTPUT(v)
#else
#define LOG_OUTPUT(v)	Asura_OutputToDebugger_Retail(v)
#endif


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#if defined(ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE)
// Count used to allow us to force read failures...
static u_int	g_uRequestFailureCount	= ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE;
#endif	// defined(ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE)

static volatile u_int g_uSetMemoryCounter = 0;


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::GetSectionToRefresh
// Author       : Andy Tate
// Date         : 10-May-2006
////////////////////////////////////////////////////////////////////////////////

int Asura_FileCache::GetSectionToRefresh (void)
{
	// Have we got a file open?
	if(m_iFileHandle == ASURA_INVALID_HANDLE)
	{
		return(-1);
	}
	else
	{
		// Yes - but do we have any data left to read?
		if(m_uBytesLeftToCache == 0)
		{
			return(-1);
		}
		else if(m_bDataValid)
		{
			const u_int	uFirstValidSection	= m_uFirstValidOffset / m_uSectionSize;
			const u_int	uLastValidSection	= m_uLastValidOffset / m_uSectionSize;

			// Yes - do we have a wrap-around buffer?
			if(uFirstValidSection <= uLastValidSection)
			{
				// No - do we need to read more data?
				if(uLastValidSection < (m_uNumSections - 1))
				{
					return(uLastValidSection + 1);
				}
				else if(uFirstValidSection > 0)
				{
					return(0);
				}
				else
				{
					return(-1);
				}
			}
			else
			{
				// Yes - are there blank sections in-between what we already have cached?
				if((uFirstValidSection - uLastValidSection) > 1)
				{
					return(uLastValidSection + 1);
				}
				else
				{
					return(-1);
				}
			}
		}
		else
		{
			// Our data is invalid - start from the beginning...
			return(0);
		}
	}
}

/// End of function 'Asura_FileCache::GetSectionToRefresh'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::ReadDuringRefresh
// Author       : Andy Tate
// Date         : 17-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileCache::ReadDuringRefresh (const u_int uNumBytes,
										 u_char* pucDest)
{
	bool	bSuccess;

#if defined(ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE)
	// Is it time to force the failure?
	if(g_uRequestFailureCount == 0)
	{
		LOG_OUTPUT(("Asura_FileCache::ReadDuringRefresh(): Force read failure.\n"));
		g_uRequestFailureCount	= ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE;
		return(false);
	}
	else
	{
		g_uRequestFailureCount--;
	}
#endif	// defined(ASURA_NUM_FILE_CACHE_REQUESTS_PER_READ_FAILURE)

	// (MH - 14.08.2006) if another thread suspended the system before we got our lock we need to fail.
	if (!m_pucBuffer)
	{
		Asura_Assert(!m_uBufferSize, "");
		return false;
	}

	// Nasty one this - allow other file-cache operations while
	// the kernel is reading the file, but prevent them again
	// as soon as we're done - ART 03/08/06.
    // (MH - 14.08.2006) if this was called by anything other than the thread function
	// the mutex was locked twice so the unlock call did nothing. I've changed it so Refresh
	// assumes the mutex was locked by the calling context.
    // (MH - 15.08.2006) If Suspend/Resume are called while we release the mutex, we need to 
	// assume the read failed and will be retried if we return false. This is because any data 
	// which did manage to get copied into volatile memory on PSP before the suspend will have
	// been trashed.
	u_int uOldSetMemCount = g_uSetMemoryCounter;
	LOG_OUTPUT(("Asura_FileCache::ReadDuringRefresh(): %u bytes at %08X\n", uNumBytes, pucDest));
	m_xMutex.Unlock();
	{
		// Read the whole file in one go...
		bSuccess	= Asura_FileAccess::ReadData(m_iFileHandle, pucDest, uNumBytes);
	}
	m_xMutex.Lock();
	if (uOldSetMemCount != g_uSetMemoryCounter)
	{
		LOG_OUTPUT(("Asura_FileCache::ReadDuringRefresh(), assuming data read failed as SetMemory/ResetMemory were called\n"));
		return false;
	}
	return(bSuccess);
}

/// End of function 'Asura_FileCache::ReadDuringRefresh'. ///

#ifdef ASURA_USE_DEBUG_REFRESH
DEBUGVAR bool dbg_bDebugLock = true;
#endif

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::Refresh
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_DEBUG_REFRESH
bool Asura_FileCache::Refresh(const char* szFunction, int iLine)
#else
bool Asura_FileCache::Refresh()
#endif
{
	bool	bMoreToRefresh;

#ifdef ASURA_USE_DEBUG_REFRESH
	int iNumRefreshes=0;

	if(dbg_bDebugLock)
	{
		LOG_OUTPUT(("Refresh TESTING LOCK: Thread %d, Function %s, Line %d\n", Asura_Thread_System::GetCurrentThreadID(), szFunction, iLine));
	}
#endif

	while(m_bInRefresh)
	{
		// Already in a refresh - wait for the end of the refresh
		LOG_OUTPUT(("Refresh requested while already refreshing\n"));
		int iNumLocks =	m_xMutex.MultiUnlock();
		Asura_Thread_System::DelayCurrentThread(0.005f);

#ifdef ASURA_USE_DEBUG_REFRESH
		if(dbg_bDebugLock)
		{
			iNumRefreshes++;
			Asura_Assert(iNumRefreshes<500, "Suspiciously long refresh wait time.");
		}
#endif

		m_xMutex.MultiLock(iNumLocks);
	}

#ifdef ASURA_USE_DEBUG_REFRESH
	if(dbg_bDebugLock)
	{
		LOG_OUTPUT(("Refresh LOCKED: Thread %d, Function %s, Line %d\n", Asura_Thread_System::GetCurrentThreadID(), szFunction, iLine));
	}
#endif

	m_bInRefresh = true;

	// Wait until we've got exclusive access...
	// (MH - 14.08.2006) instead of locking here, ensure the calling code locks the mutex
	//m_xMutex.Lock();
	{
		// Have we got a file open?
		if(m_iFileHandle == ASURA_INVALID_HANDLE)
		{
			bMoreToRefresh	= false;
		}
		else
		{
			// Yes - but do we have any data left to read?
			if(m_uBytesLeftToCache == 0)
			{
				bMoreToRefresh	= false;
			}
			else if(m_bDataValid)
			{
				int	iSectionToRefresh	= GetSectionToRefresh();

				// Yes - do we have a section to read?
				if(iSectionToRefresh >= 0)
				{
					const u_int	uReadOffset		= iSectionToRefresh * m_uSectionSize;
					const u_int	uBytesToRead	= Asura_Maths::Min(m_uSectionSize, m_uBytesLeftToCache);

					// Yes - read the data...
					if(ReadDuringRefresh(uBytesToRead, &m_pucBuffer[uReadOffset]))
					{
						m_uValidSizeRemaining	+= uBytesToRead;
						m_uBytesLeftToCache		-= uBytesToRead;

						if(iSectionToRefresh == 0)
						{
							m_uLastValidOffset	= uBytesToRead - 1;
						}
						else
						{
							m_uLastValidOffset	+= uBytesToRead;
						}
					}
					else
					{
                        // (MH - 15.08.2006) copy of code below
						// (MH - 15.08.2006) This used to assert and do nothing. It seems to me
						// that we must force the data to be read again by going back to the seeking
						// state. I've also signalled the event to ensure the thread continues.
						LOG_OUTPUT(("Asura_FileCache::ReadDuringRefresh failed.\n"));
						m_eStateFollowingError	= SEEKING;
						LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eStateFollowingError = SEEKING\n", this, __FUNCTION__, __LINE__));
						if(!m_xUpdateReadyEvent.IsSignalled())
						{
							m_xUpdateReadyEvent.SetSignal();
						}
					}
				}
			}
			else
			{
				const u_int	uBytesRemaining	= m_uFileSize - m_uFilePosition;

				const u_int	uBufferSize		= m_uBufferSize;
				const u_int	uBytesToRead	= Asura_Maths::Min(uBufferSize, uBytesRemaining);

				// Time to validate the cache...
				if(ReadDuringRefresh(uBytesToRead, m_pucBuffer))
				{
					// Our cache is full (or as full as it needs to be)...
					m_bDataValid		= true;

					m_uCurrentOffset		= 0;
					m_uFirstValidOffset		= 0;
					m_uLastValidOffset		= uBytesToRead - 1;
					m_uValidSizeRemaining	= uBytesToRead;
					m_uBytesLeftToCache		= uBytesRemaining - uBytesToRead;
				}
				else
				{
					// (MH - 15.08.2006) copy of code above
					// (MH - 15.08.2006) This used to assert and do nothing. It seems to me
					// that we must force the data to be read again by going back to the seeking
					// state. I've also signalled the event to ensure the thread continues.
					LOG_OUTPUT(("Asura_FileCache::ReadDuringRefresh failed.\n"));
					m_eStateFollowingError	= SEEKING;
					LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eStateFollowingError = SEEKING\n", this, __FUNCTION__, __LINE__));
					if(!m_xUpdateReadyEvent.IsSignalled())
					{
						m_xUpdateReadyEvent.SetSignal();
					}
				}
			}

			// Do we still have more data to read?
			bMoreToRefresh	= (m_uBytesLeftToCache > 0);
		}
	}
	// (MH - 14.08.2006) instead of locking/unlocking here, ensure the calling code locks the mutex
	//m_xMutex.Unlock();
	
#ifdef ASURA_USE_DEBUG_REFRESH
	if(dbg_bDebugLock)
	{
		LOG_OUTPUT(("Refresh UNLOCKED: Thread %d, Function %s, Line %d\n", Asura_Thread_System::GetCurrentThreadID(), szFunction, iLine));
	}
#endif

	m_bInRefresh = false;

	return(bMoreToRefresh);
}

/// End of function 'Asura_FileCache::Refresh'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::OpenFile_Internal
// Author       : Andy Tate
// Date         : 27-Jul-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileCache::OpenFile_Internal (void)
{
	bool	bSuccess	= false;

	// Wait until we've got exclusive access...
	m_xMutex.Lock();
	m_iFileHandle	= ASURA_INVALID_HANDLE;
	{
		const char*	szFilename	= m_acFilename;

		// Do we have a disc?
		if(Asura_FileAccess::DiskExists())
		{
			// Nasty one this - allow other file-cache operations while
			// the kernel is reading the file, but prevent them again
			// as soon as we're done - ART 03/08/06.
			m_xMutex.Unlock();
			{
				// Attempt to open the file...
				m_iFileHandle	= Asura_FileAccess::OpenRaw(szFilename);
				LOG_OUTPUT( ( "OpenFile_Internal '%s' 0x%X\n", szFilename, m_iFileHandle ) );
			}
			m_xMutex.Lock();
		}
		else
		{
			// No - ensure that we fail gracefully...
			m_iFileHandle	= ASURA_INVALID_HANDLE;
		}

		if(m_iFileHandle == ASURA_INVALID_HANDLE)
		{
			// Failed to open this file...
			if(Asura_FileAccess::FileExists(szFilename))
			{
				m_eStateFollowingError	= OPENING;
				LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eStateFollowingError = OPENING\n", this, __FUNCTION__, __LINE__));
			}
			else
			{
				Asura_Assertf(false, ("Failed to open \"%s\" for file-cache.", szFilename));
			}
		}
		else
		{
			// Yes - work out how much we can read initially...

			m_bDataValid			= false;
			m_bCanRead				= false;
			LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = FALSE\n", this, __FUNCTION__, __LINE__));
			m_uFirstValidOffset 	= 0;
			m_uLastValidOffset		= 0;
			m_uValidSizeRemaining	= 0;
			m_uCurrentOffset		= 0;

			m_uFilePosition		= 0;

			// Nasty one this - allow other file-cache operations while
			// the kernel is reading the file, but prevent them again
			// as soon as we're done - ART 03/08/06.
			m_xMutex.Unlock();
			{
				const char*	const szFilename = m_acFilename;
				m_uFileSize = Asura_FileAccess::GetFileSize( szFilename );
			}
			m_xMutex.Lock();
			bSuccess = true;

			// If we have a non-zero start offset then seek to that position
			// Then update the file position so the file cache just thinks it has started reading for the start offset position
			if ( m_uStartOffset > 0 )
			{
				// Nasty one this - allow other file-cache operations while
				// the kernel is reading the file, but prevent them again
				// as soon as we're done - ART 03/08/06.
				m_xMutex.Unlock();
				{
					bSuccess = Asura_FileAccess::SeekFile( m_iFileHandle, m_uStartOffset );
				}
				m_xMutex.Lock();
				if ( bSuccess )
				{
					m_uFilePosition = m_uStartOffset;
				}
				else
				{
					Asura_Assertf( false, ( "Open '%s' Seek %d Size %d failed for file-cache.", 
								   m_acFilename, m_uStartOffset, m_uFileSize ) );
				}
			}

			m_uBytesLeftToCache	= m_uFileSize - m_uFilePosition;

			// Invalidate and refresh the data...
			Invalidate();
#ifdef ASURA_USE_DEBUG_REFRESH
			Refresh(__FUNCTION__, __LINE__);
#else
			Refresh();
#endif
			m_eState				= READING;
			LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = READING\n", this, __FUNCTION__, __LINE__));
		}
	}
	m_xMutex.Unlock();
	return(bSuccess);
}

/// End of function 'Asura_FileCache::OpenFile_Internal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::SeekFile_Internal
// Author       : Andy Tate
// Date         : 1-Aug-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileCache::SeekFile_Internal (void)
{
	bool	bSuccess	= false;

	// Wait until we've got exclusive access...
	m_xMutex.Lock();
	{
		// Nasty one this - allow other file-cache operations while
		// the kernel is reading the file, but prevent them again
		// as soon as we're done - ART 03/08/06.
		m_xMutex.Unlock();
		{
			bSuccess	= Asura_FileAccess::SeekFile(m_iFileHandle, m_uFilePosition);
		}
		m_xMutex.Lock();

		// Seek back to the last position we read before refreshing the cache...
		if(bSuccess)
		{
			m_uBytesLeftToCache	= m_uFileSize - m_uFilePosition;

#ifdef ASURA_USE_DEBUG_REFRESH
			Refresh(__FUNCTION__, __LINE__);
#else
			Refresh();
#endif

			m_eState			= READING;
			LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = READING\n", this, __FUNCTION__, __LINE__));
		}
		else
		{
			Asura_Assert(false, "Seek failed for file-cache.");
		}
	}
	m_xMutex.Unlock();
	return(bSuccess);
}

/// End of function 'Asura_FileCache::SeekFile_Internal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::CloseFile_Internal
// Author       : Andy Tate
// Date         : 27-Jul-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileCache::CloseFile_Internal (void)
{
	// Wait until we've got exclusive access...
	m_xMutex.Lock();
	{
		if(m_bInRefresh)
		{
			// Already in a refresh - wait for the end of the refresh
			LOG_OUTPUT(("Close requested while refreshing\n"));
			while(m_bInRefresh)
			{
				m_xMutex.Unlock();
				Asura_Thread_System::DelayCurrentThread(0.005f);
				m_xMutex.Lock();			
			}
		}

		m_bDataValid			= false;
		m_bCanRead				= false;

		// Nasty one this - allow other file-cache operations while
		// the kernel is reading the file, but prevent them again
		// as soon as we're done - ART 03/08/06.
		m_xMutex.Unlock();
		{
			// Close the file...
			Asura_FileAccess::Close(m_iFileHandle);
		}
		m_xMutex.Lock();

		m_iFileHandle	= ASURA_INVALID_HANDLE;
		m_eState		= IDLE;
		m_eStateFollowingError	= IDLE;
		LOG_OUTPUT( ("CloseFile_Internal '%s'\n", m_acFilename ) );
		LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = IDLE\n", this, __FUNCTION__, __LINE__));

		m_bDataValid			= false;
		m_bCanRead				= false;
		LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = FALSE\n", this, __FUNCTION__, __LINE__));
		m_uFirstValidOffset 	= 0;
		m_uLastValidOffset		= 0;
		m_uValidSizeRemaining	= 0;
		m_uCurrentOffset		= 0;
	}
	m_xMutex.Unlock();
	return(true);
}

/// End of function 'Asura_FileCache::CloseFile_Internal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::UpdateThread
// Author       : Andy Tate
// Date         : 10-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_FileCache::UpdateThread (const void* pUserData)
{
	void*				pNonConstUserData	= const_cast<void*>(pUserData);
	Asura_FileCache*	pxFileCache			= reinterpret_cast<Asura_FileCache*>(pNonConstUserData);

	// Do we have a valid file-cache?
	if ( pxFileCache != NULL )
	{
		// Yes - loop forever...
		for(;;)
		{
			// Wait until we're told to do something...
			if (pxFileCache->m_eStateFollowingError != IDLE)
			{
				pxFileCache->m_eState				= pxFileCache->m_eStateFollowingError;
				LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = m_eStateFollowingError\n", pxFileCache, __FUNCTION__, __LINE__));
				pxFileCache->m_eStateFollowingError	= IDLE;
				LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eStateFollowingError = IDLE\n", pxFileCache, __FUNCTION__, __LINE__));
			}
			else
			{
				pxFileCache->m_xUpdateReadyEvent.WaitForSignal();
				pxFileCache->m_xUpdateReadyEvent.ClearSignal();
			}
			// Catch the case where the file got closed on us whilst we were reading from it!
			if ( pxFileCache->IsOpen() == false )
			{
				if ( ( pxFileCache->m_eState != OPENING ) && ( pxFileCache->m_eState != IDLE ) )
				{
					// Reset some internal states
					pxFileCache->m_xMutex.Lock();
					pxFileCache->m_eState = IDLE;
					pxFileCache->m_eStateFollowingError = IDLE;
					pxFileCache->m_xMutex.Unlock();

					pxFileCache->m_xUpdateReadyEvent.WaitForSignal();
					pxFileCache->m_xUpdateReadyEvent.ClearSignal();
				}
			}

			pxFileCache->SetCanRead(false);
			LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = FALSE\n", pxFileCache, __FUNCTION__, __LINE__));

			// What state are we in?
			switch(pxFileCache->m_eState)
			{
				case OPENING:
				{
					pxFileCache->OpenFile_Internal();
					break;
				}

				case SEEKING:
				{
					pxFileCache->SeekFile_Internal();
					break;
				}

				case READING:
				{
					pxFileCache->m_xMutex.Lock();

#ifdef ASURA_USE_DEBUG_REFRESH
					pxFileCache->Refresh(__FUNCTION__, __LINE__);
#else
					pxFileCache->Refresh();
#endif

					pxFileCache->m_xMutex.Unlock();
					break;
				}

				case CLOSING:
				{
					pxFileCache->CloseFile_Internal();
					break;
				}

				default:
				{
					Asura_Assert(false, "Unexpected state for file-cache.");
					break;
				}
			}		// End of switch.

			pxFileCache->SetCanRead(true);			
			LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = TRUE\n", pxFileCache, __FUNCTION__, __LINE__));

			// Ensure that we don't get stuck in this loop...
			Asura_Thread_System::DelayCurrentThread(0.01f, false);

//			while(Asura_PSP_FileIO::IsSuspended())
//			{
//				Asura_Thread_System::DelayCurrentThread(0.02f, false);
//			}

		}
	}
}

/// End of function 'Asura_FileCache::UpdateThread'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::Asura_FileCache
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

Asura_FileCache::Asura_FileCache (const char* szName,
								  const u_int uBufferSize,
								  u_char* pucBuffer,
								  const u_int uNumSections,
								  const bool bAllowBlockingRead) :

m_xUpdateReadyEvent("File-cache Update Ready", false),
m_bAllocated(false),
m_pucBuffer(pucBuffer),
m_uBufferSize(uBufferSize),
m_uNumSections(uNumSections),
m_uSectionSize(uBufferSize / uNumSections),
m_bAllowBlockingRead(bAllowBlockingRead),
m_eState(IDLE),
m_eStateFollowingError(IDLE),
m_bInRefresh(false),
m_iFileHandle(ASURA_INVALID_HANDLE),
m_uFilePosition(0),
m_uFileSize(0),
m_bDataValid(false),
m_bCanRead(true),
m_uFirstValidOffset(0),	
m_uLastValidOffset(0),		
m_uValidSizeRemaining(0),	
m_uCurrentOffset(0),		
m_uBytesLeftToCache(0)
{
	Asura_PrintfToBfr( m_acName, "%s", uASURA_FILECACHE_MAX_NAME_LENGTH, true, szName );
	LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eStateFollowingError = IDLE\n", this, __FUNCTION__, __LINE__));
	LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = IDLE\n", this, __FUNCTION__, __LINE__));
	LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = TRUE\n", this, __FUNCTION__, __LINE__));
	// Do we need to create a thread to use?
	if(!m_bAllowBlockingRead)
	{
		Asura_Thread_System::CreateThread(szName, UpdateThread, this);
	}
}

/// End of function 'Asura_FileCache::Asura_FileCache'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::Asura_FileCache
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

Asura_FileCache::Asura_FileCache (const char* szName,
								  const u_int uBufferSize,
								  const u_int uNumSections,
								  const bool bAllowBlockingRead) :
m_xUpdateReadyEvent("File-cache Update Ready", false),
m_bAllocated(true),
m_uBufferSize(uBufferSize),
m_uNumSections(uNumSections),
m_uSectionSize(uBufferSize / uNumSections),
m_bAllowBlockingRead(bAllowBlockingRead),
m_eState(IDLE),
m_eStateFollowingError(IDLE),
m_bInRefresh(false),
m_iFileHandle(ASURA_INVALID_HANDLE),
m_uFilePosition(0),
m_uFileSize(0),
m_bDataValid(false),
m_bCanRead(true),
m_uFirstValidOffset(0),	
m_uLastValidOffset(0),		
m_uValidSizeRemaining(0),	
m_uCurrentOffset(0),		
m_uBytesLeftToCache(0)
{
	Asura_PrintfToBfr( m_acName, "%s", uASURA_FILECACHE_MAX_NAME_LENGTH, true, szName );
	LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eStateFollowingError = IDLE\n", this, __FUNCTION__, __LINE__));
	LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = IDLE\n", this, __FUNCTION__, __LINE__));
	LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = TRUE\n", this, __FUNCTION__, __LINE__));
	// Allocate the data...
	m_pucBuffer	= new u_char [m_uBufferSize];

	// Do we need to create a thread to use?
	if(!m_bAllowBlockingRead)
	{
		Asura_Thread_System::CreateThread(szName, UpdateThread, this);
	}
}

/// End of function 'Asura_FileCache::Asura_FileCache'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::~Asura_FileCache
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

Asura_FileCache::~Asura_FileCache ()
{
	// Do we have anything to destroy?
	if(m_bAllocated)
	{
		delete[](m_pucBuffer);
	}
}

/// End of function 'Asura_FileCache::~Asura_FileCache'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::ResetMemory
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_FileCache::ResetMemory (void)
{
	// Ensure we're doing nothing else here...
	m_xMutex.Lock();
	{
		++g_uSetMemoryCounter;

		// Did we allocate the memory?
		if(m_bAllocated)
		{
			delete[](m_pucBuffer);
			m_bAllocated	= false;
		}

		// Reset the variables...
		m_pucBuffer				= NULL;
		m_uBufferSize			= 0;
		m_uSectionSize			= 0;
		m_bDataValid			= false;
		m_bCanRead				= true;
		LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - CanRead = TRUE\n", this, __FUNCTION__, __LINE__));
		m_uFirstValidOffset 	= 0;
		m_uLastValidOffset		= 0;
		m_uValidSizeRemaining	= 0;
		m_uCurrentOffset		= 0;
		m_uBytesLeftToCache		= m_uFileSize;
	}
	m_xMutex.Unlock();
}

/// End of function 'Asura_FileCache::ResetMemory'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::SetMemory
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_FileCache::SetMemory (const u_int uNumBytes,
								 u_char* pucDest)
{
	// Ensure we're doing nothing else here...
	m_xMutex.Lock();
	{
		++g_uSetMemoryCounter;

		// Reset any existing memory usage...
		ResetMemory();

		// Set this block...
		m_pucBuffer		= pucDest;
		m_uBufferSize	= uNumBytes;
		m_uSectionSize	= uNumBytes / m_uNumSections;

		if(m_eState == READING)
		{
			m_eState	= SEEKING;
			LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = SEEKING\n", this, __FUNCTION__, __LINE__));
		}

		if((m_eState != IDLE) && !m_xUpdateReadyEvent.IsSignalled())
		{
			m_xUpdateReadyEvent.SetSignal();
		}
	}
	m_xMutex.Unlock();
}

/// End of function 'Asura_FileCache::SetMemory'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::OpenFile
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileCache::OpenFile ( const char* const szFilename,
								 const bool bForceBlocking,
								 const u_int uStartOffset )
{
	bool	bSuccess	= false;

	// Prevent anything else from using the cache...
	m_xMutex.Lock();
	{
		Asura_Assert( m_eState == IDLE && m_iFileHandle == ASURA_INVALID_HANDLE, "" );
		// JAT 22/06/07 - this sticking plaster code isn't required anymore because the root cause has been fixed replace with an assert
#if 0
		//MM  whole file was moved from HP branch
		if (m_eState == SEEKING && m_iFileHandle == ASURA_INVALID_HANDLE)
		{
			// MGH 30/4/2007: Try to fix a hang that occurs. Somehow the state is SEEKING, even though we have no filehandle.
			LOG_OUTPUT(("Asura_FileCache::OpenFile: converting SEEKING to IDLE state!\n"));
			m_eState = IDLE;
		}
#endif // #if 0
		// Do we already have a file open?
		if((m_eState == IDLE) && (m_iFileHandle == ASURA_INVALID_HANDLE))
		{
			// Nasty one this - allow other file-cache operations while
			// the kernel is reading the file, but prevent them again
			// as soon as we're done - ART 03/08/06.
			m_xMutex.Unlock();
			{
				bSuccess	= Asura_FileAccess::FileExists(szFilename);
			}
			m_xMutex.Lock();

			// No - does the file exist?
			if ( bSuccess )
			{
				// Yes - either open the file now, or save it for later...
				Asura_Assertf(strlen(szFilename) < uASURA_FILECACHE_MAX_FILENAME_LENGTH, ("filename %s too long for filecache", szFilename));
				Asura_PrintfToBfr( m_acFilename, "%s", uASURA_FILECACHE_MAX_FILENAME_LENGTH, true, szFilename );
				m_uStartOffset = uStartOffset;

				if(m_bAllowBlockingRead || bForceBlocking)
				{
					bSuccess	= OpenFile_Internal();
				}
				else
				{
					m_eState	= OPENING;
					LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = OPENING\n", this, __FUNCTION__, __LINE__));

					if(!m_xUpdateReadyEvent.IsSignalled())
					{
						m_xUpdateReadyEvent.SetSignal();
					}
				}
			}
		}
		else
		{
			Asura_Assertf(false, ("We already have a file (%s) open - cannot open a second one.", m_acFilename));
		}
	}
	m_xMutex.Unlock();
	return(bSuccess);
}

/// End of function 'Asura_FileCache::OpenFile'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::ReadData
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

u_int Asura_FileCache::ReadData (const u_int uNumBytes,
								 u_char* pucDest)
{
	Asura_Assert(uNumBytes, "Asura_FileCache::ReadData called with no data requested");

	u_int	uNumBytesRead;

	// Prevent anything else from using the cache...
	m_xMutex.Lock();
    
	// (MH - 15.08.2006) first check if the buffer has been cleared by suspend code (on f).
	// If it has we must wait until it comes back.
	if (!m_pucBuffer)
	{
		LOG_OUTPUT(("FileCache waiting for buffer to be allocated\n"));
	}
	while (!m_pucBuffer)
	{
		m_xMutex.Unlock();
		Asura_Thread_System::DelayCurrentThread(0.01f, false);
		m_xMutex.Lock();
	}

    // (MH - 18.08.2006) also check that we're in the Reading state. PSP suspend
	// can change the state to Seeking, but the update thread will eventually change 
	// it back. Without this step we end up with corrupt data.
	if (m_eState != READING)
	{
		LOG_OUTPUT( ( "FileCache:ReadData waiting for read state\n"));
	}
	while (m_eState != READING)
	{
		m_xMutex.Unlock();
/*
		if(!m_xUpdateReadyEvent.IsSignalled())
		{
			m_xUpdateReadyEvent.SetSignal();
		}
		else
		{
			LOG_OUTPUT(("Filecache waiting for read state, update event already signalled\n"));
		}
*/
		Asura_Thread_System::DelayCurrentThread(0.01f, false);
		m_xMutex.Lock();
	}

	{
		LOG_OUTPUT(("FileCache:ReadData got read state\n"));
		const u_int	uNumBytesRemaining	= m_uFileSize - m_uFilePosition;
		const int	iSectionToRefresh	= GetSectionToRefresh();

		// Do we need to force a refresh?
		if(iSectionToRefresh >= 0)
		{
			if(m_bAllowBlockingRead)
			{					
#ifdef ASURA_USE_DEBUG_REFRESH
				Refresh(__FUNCTION__, __LINE__);
#else
				Refresh();
#endif
			}
			else if(CanRead(uNumBytes))
			{
				if(!m_xUpdateReadyEvent.IsSignalled())
				{
					m_xUpdateReadyEvent.SetSignal();
				}
			}
			else
			{
				while (!CanRead(uNumBytes))
				{
					m_xMutex.Unlock();
					Asura_Thread_System::DelayCurrentThread(0.01f, false);
					m_xMutex.Lock();
				}				
			}
		}

		// Are we attempting to read past the end of the file?
		if(uNumBytes <= uNumBytesRemaining)
		{
			const u_int	uExtendedFirstValidOffset	= m_uFirstValidOffset;
			const u_int	uExtendedLastValidOffset	= (m_uFirstValidOffset < m_uLastValidOffset)
				? m_uLastValidOffset
				: (m_uBufferSize + m_uLastValidOffset);

			const u_int uExtendedStartOfReadOffset	= m_uCurrentOffset;
			const u_int	uExtendedEndOfReadOffset	= uExtendedStartOfReadOffset + (uNumBytes - 1);

			// Do we have enough data in our cache to perform the read?
			if((uExtendedStartOfReadOffset >= uExtendedFirstValidOffset) && (uExtendedStartOfReadOffset <= uExtendedLastValidOffset))
			{
				if((uExtendedEndOfReadOffset >= uExtendedFirstValidOffset) && (uExtendedEndOfReadOffset <= uExtendedLastValidOffset))
				{
					// Yes - do we need to split our read into 2 parts?
					if(uExtendedEndOfReadOffset >= m_uBufferSize)
					{
						const u_int	uNumBytesToEndOfBuffer		= m_uBufferSize - uExtendedStartOfReadOffset;
						const u_int	uNumBytesFromStartOfBuffer	= (uExtendedEndOfReadOffset + 1) - m_uBufferSize;

						// Yes - perform our 2 copies...
						LOG_OUTPUT(("Asura_FileCache::ReadData(): %u bytes from %08X to %08X\n", uNumBytesToEndOfBuffer, &m_pucBuffer[m_uCurrentOffset], pucDest));
						LOG_OUTPUT(("Asura_FileCache::ReadData(): %u bytes from %08X to %08X\n", uNumBytesFromStartOfBuffer, m_pucBuffer, &pucDest[uNumBytesToEndOfBuffer]));
						Asura_Mem::Copy(pucDest, &m_pucBuffer[m_uCurrentOffset], uNumBytesToEndOfBuffer);
						Asura_Mem::Copy(&pucDest[uNumBytesToEndOfBuffer], m_pucBuffer, uNumBytesFromStartOfBuffer);

						// Update the variables before returning...
						m_uCurrentOffset	 = uNumBytesFromStartOfBuffer;
						m_uFirstValidOffset	 = uNumBytesFromStartOfBuffer;
						m_uFilePosition		+= uNumBytes;

						uNumBytesRead		 = uNumBytes;
					}
					else
					{
						// No - we can simply copy the data in one go...
						LOG_OUTPUT(("Asura_FileCache::ReadData(): %u bytes from %08X to %08X\n", uNumBytes, &m_pucBuffer[m_uCurrentOffset], pucDest));
						Asura_Mem::Copy(pucDest, &m_pucBuffer[m_uCurrentOffset], uNumBytes);

						// Update the variables before returning...
						m_uCurrentOffset	+= uNumBytes;
						m_uFirstValidOffset	+= uNumBytes;
						m_uFilePosition		+= uNumBytes;

						uNumBytesRead		 = uNumBytes;
					}
				}
				else
				{
					Asura_Assert(false, "Attempting to read more data than there is in the cache.");
					uNumBytesRead	= 0;
				}
			}
			else
			{
				Asura_Assert(false, "Start offset for reading isn't contained within the cache.");
				uNumBytesRead	= 0;
			}

			// Decrease the size remaining...
			Asura_Assert((uNumBytes <= m_uValidSizeRemaining), "Incorrect size remaining in cache.");
			m_uValidSizeRemaining	-= uNumBytes;
		}
		else
		{
			Asura_Assert(false, "Attempting to read past the end of the file.");
			uNumBytesRead	= 0;
		}
	}
	m_xMutex.Unlock();

	Asura_Assert(uNumBytesRead, "Asura_FileCache::ReadData() failed");
	return(uNumBytesRead);
}

/// End of function 'Asura_FileCache::ReadData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::SkipData
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

u_int Asura_FileCache::SkipData (const u_int uNumBytes)
{
	u_int	uNumBytesSkipped;

	// Prevent anything else from using the cache...
	m_xMutex.Lock();
	{
		const u_int	uNumBytesRemaining	= m_uFileSize - m_uFilePosition;
		const int	iSectionToRefresh	= GetSectionToRefresh();

		// Do we need to force a refresh?
		if(iSectionToRefresh >= 0)
		{
			if(m_bAllowBlockingRead)
			{
#ifdef ASURA_USE_DEBUG_REFRESH
				Refresh(__FUNCTION__, __LINE__);
#else
				Refresh();
#endif
			}
			else if(CanRead(uNumBytes))
			{
				if(!m_xUpdateReadyEvent.IsSignalled())
				{
					m_xUpdateReadyEvent.SetSignal();
				}
			}
			else
			{
				while (!CanRead(uNumBytes))
				{
					m_xMutex.Unlock();
					Asura_Thread_System::DelayCurrentThread(0.01f, false);
					m_xMutex.Lock();
				}				
			}
		}

		// Are we attempting to skip past the end of the file?
		if(uNumBytes <= uNumBytesRemaining)
		{
			const u_int	uExtendedFirstValidOffset	= m_uFirstValidOffset;
			const u_int	uExtendedLastValidOffset	= (m_uFirstValidOffset < m_uLastValidOffset)
				? m_uLastValidOffset
				: (m_uBufferSize + m_uLastValidOffset);

			const u_int uExtendedStartOfSkipOffset	= m_uCurrentOffset;
			const u_int	uExtendedEndOfSkipOffset	= uExtendedStartOfSkipOffset + (uNumBytes - 1);

			// Do we have enough data in our cache to perform the skip?
			if((uExtendedStartOfSkipOffset >= uExtendedFirstValidOffset) && (uExtendedStartOfSkipOffset <= uExtendedLastValidOffset))
			{
				if((uExtendedEndOfSkipOffset >= uExtendedFirstValidOffset) && (uExtendedEndOfSkipOffset <= uExtendedLastValidOffset))
				{
					// Yes - do we need to split our skip into 2 parts?
					if(uExtendedEndOfSkipOffset >= m_uBufferSize)
					{
						const u_int	uNumBytesFromStartOfBuffer	= (uExtendedEndOfSkipOffset + 1) - m_uBufferSize;

						// Yes - update the variables before returning...
						m_uCurrentOffset	 = uNumBytesFromStartOfBuffer;
						m_uFirstValidOffset	 = uNumBytesFromStartOfBuffer;
						m_uFilePosition		+= uNumBytes;

						uNumBytesSkipped	 = uNumBytes;
					}
					else
					{
						// No - update the variables before returning...
						m_uCurrentOffset	+= uNumBytes;
						m_uFirstValidOffset	+= uNumBytes;
						m_uFilePosition		+= uNumBytes;

						uNumBytesSkipped	 = uNumBytes;
					}
				}
				else
				{
					Asura_Assert(false, "Attempting to skip more data than there is in the cache.");
					uNumBytesSkipped	= 0;
				}
			}
			else
			{
				Asura_Assert(false, "Start offset for skipping isn't contained within the cache.");
				uNumBytesSkipped	= 0;
			}

			// Decrease the size remaining...
			Asura_Assert((uNumBytes <= m_uValidSizeRemaining), "Incorrect size remaining in cache.");
			m_uValidSizeRemaining	-= uNumBytes;
		}
		else
		{
			Asura_Assert(false, "Attempting to skip past the end of the file.");
			uNumBytesSkipped	= 0;
		}
	}
	m_xMutex.Unlock();
	return(uNumBytesSkipped);
}

/// End of function 'Asura_FileCache::SkipData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_FileCache::CloseFile
// Author       : Andy Tate
// Date         : 26-Apr-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileCache::CloseFile (const bool bForceBlocking)
{
	bool	bSuccess;

	// Prevent anything else from using the cache...
	m_xMutex.Lock();
	{
		// Do we have a file open?
		if(m_iFileHandle == ASURA_INVALID_HANDLE)
		{
			Asura_Assert(false, "Attempting to close a file when no file is open in this file-cache.");
			bSuccess	= false;
		}
		else
		{
			Asura_Assert(m_eState != IDLE, "In idle state with file open?");

			while(m_eState == OPENING)
			{
				m_xMutex.Unlock();
				Asura_Thread_System::DelayCurrentThread(0.01f, false);
				m_xMutex.Lock();
			}

			switch(m_eState)
			{
				case READING:
				case SEEKING:
				{
					if(m_bAllowBlockingRead || bForceBlocking)
					{
						// Unlock the mutex here - it's relocked in the function below,
						// so this avoids the double-lock.
						m_xMutex.Unlock();
						bSuccess	= CloseFile_Internal();
						m_xMutex.Lock();
					}
						
					else
					{
						m_eState	= CLOSING;
						LOG_OUTPUT(("** Cache 0x%08X, %s (%d) - m_eState = CLOSING\n", this, __FUNCTION__, __LINE__));
						bSuccess	= true;
	
						if(!m_xUpdateReadyEvent.IsSignalled())
						{
							m_xUpdateReadyEvent.SetSignal();				
						}
					}
					break;
				}
				case CLOSING:
				{
					Asura_Assert(0, "Closing while already closing? Possibly safe");
					bSuccess = true;
					break;
				}
				default:
				{			
					Asura_Assert(0, "Unexpected state");
					bSuccess = false;
					break;
				}
			}
		}
	}
	m_xMutex.Unlock();
	return(bSuccess);
}

/// End of function 'Asura_FileCache::CloseFile'. ///


/// End of file. ///
