//
// Asura_GUIMenu_Element_Animation.cpp
//
// Created by RDM - 16 April 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

/************/
/* Includes */
/************/

// main asura header
#include "Core/Asura.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM

// This file's header.
#include "Asura_GUIMenu_Element_Animation.h"

// asura
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

#ifdef ASURA_TOOLS
	#include "ChunkSystem/Asura_Chunk_Loaders.h"
	#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
	#include "GUIMenu/Tools_Mode_GUIMenu.h"
	#include "Load&Save/LevelExport/Tools_Level_Export.h"
#endif

/********/
/* Data */
/********/

// debugvars
DEBUGVAR bool dbg_bRenderViewport = false;
DEBUGVAR float dbg_fDefaultFOV = 0.654f; // about 75 degrees

/*************/
/* Functions */
/*************/

// constructor
Asura_GUIMenu_Element_Animation::Asura_GUIMenu_Element_Animation()
: m_pxAnimation( NULL )
, m_uSkinID( ASURA_HASH_ID_UNSET )
, m_uAnimID( ASURA_HASH_ID_UNSET )
, m_uAnimFlags( AAC_FLAG_ACTIVE )
, m_fFOVAngle( dbg_fDefaultFOV ) 
, m_fZoomLevel( 1.0f ) 
, m_xCameraOffset( 0.0f, 0.0f, -1.0f )
, m_xRenderOrient( 0.0f, Asura_Maths::Deg180, 0.0f )
, m_xSkinOffset( 0.0f, 0.0f, 0.0f )
, m_xCameraOrient( 0.0f, 0.0f, 0.0f )
, m_xRotationSpeed( 0.0f, 0.0f, 0.0f )
, m_bUseVerticalFOV( false )
, m_bStretch( false )
#ifdef ASURA_TOOLS
, m_bExportSkinAndAnim( true )
#endif
{
	m_uType = ASURA_GUIMENU_TYPEID_ELEMENT_ANIMATION;

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_ANIMATION);
}

// destructor
Asura_GUIMenu_Element_Animation::~Asura_GUIMenu_Element_Animation()
{
	delete m_pxAnimation; 

	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_ANIMATION);
}



/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::Update
 *
 * Author        : Richard May
 *
 * Description   : Update function. Updates the render orientation of the anim
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::Update( float fElapsedTime )
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_ANIMATION);

	PARENT::Update( fElapsedTime );

#ifdef ASURA_TOOLS	
	LoadSkinAndAnimFiles();
#endif // ASURA_TOOLS

	// create the animation if it's not been made already
	if( !m_pxAnimation )
	{
		SetupAnimation();
	}

	// update render orientation	
	m_xRenderOrient += (m_xRotationSpeed * fElapsedTime);

	ASURA_GUIMENU_PROFILE_END_TASK();
}
/* end of function Update */



/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::Copy
 *
 * Author        : Richard May
 *
 * Description   : Copies all the animation element-specific data
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::Copy( const Asura_GUIMenu_Root* pxSource )
{
	PARENT::Copy( pxSource );

	const Asura_GUIMenu_Element_Animation* pxSourceAnim = static_cast<const Asura_GUIMenu_Element_Animation*> (pxSource);
		
	m_uSkinID			= pxSourceAnim->m_uSkinID;
	m_uAnimID			= pxSourceAnim->m_uAnimID;	
	m_uAnimFlags		= pxSourceAnim->m_uAnimFlags;
	m_fFOVAngle			= pxSourceAnim->m_fFOVAngle;
	m_fZoomLevel		= pxSourceAnim->m_fZoomLevel;
	m_xCameraOffset		= pxSourceAnim->m_xCameraOffset;
	m_xRenderOrient		= pxSourceAnim->m_xRenderOrient;
	m_xSkinOffset		= pxSourceAnim->m_xSkinOffset;
	m_xCameraOrient		= pxSourceAnim->m_xCameraOrient;
	m_xRotationSpeed	= pxSourceAnim->m_xRotationSpeed;
}
/* end of function Copy */

/*************************************************************
*
* Name          : Asura_GUIMenu_Element_Animation::DeleteAnimation
*
* Author        : Paul Cardy
*
* Description   : Delete animation, but do not reset hashes (so 
*					SetupAnimation can be called with no further action)
*
*************************************************************/
void Asura_GUIMenu_Element_Animation::DeleteAnimation()
{
	if (m_pxAnimation)
	{
		delete m_pxAnimation;
		m_pxAnimation = NULL;
	}
}
/* end of function Asura_GUIMenu_Element_Animation::DeleteAnimation */


/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::SetupAnimation
 *
 * Author        : Richard May
 *
 * Description   : Update function. Updates the render orientation of the anim
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::SetupAnimation()
{
	DeleteAnimation();

	if( m_uSkinID != ASURA_HASH_ID_UNSET )
	{
		m_pxAnimation = new Asura_Animation( m_uSkinID );

		if( m_uAnimID != ASURA_HASH_ID_UNSET )
		{				
			// create anim controller
			PlayAnim( m_uAnimID, m_uAnimFlags );
		}
	}
}
/* end of function Asura_GUIMenu_Element_Animation::SetupAnimation */


/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::Render
 *
 * Author        : Richard May
 *
 * Description   : Render function. Basically sets up the camera, renders
 *				   the anim and then resets the original camera/viewport values
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::Render( u_int uState, const Asura_Bounding_Rectangle& xBounds)
{
	if( !m_pxAnimation || IsRenderingDisabled() ) return;

	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_ANIMATION);

	//	BeginCamera();
	Asura_Render::FlushDisplay();

	Asura_Render::AllowFullScreenRendering();
	Asura_Render::ClearZBuffer();
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);

	float fX = xBounds.MinX;
	float fY = xBounds.MinY;
	float fHeight = xBounds.GetHeight();
	float fWidth  = xBounds.GetWidth();

#ifdef ASURA_TOOLS
	if(Tools_Mode_GUIMenu::ShouldUseOffsetAndZoom())
	{
		//Apply offset to coords from menu mode
		fX = (fX + Tools_Mode_GUIMenu::GetOffsetX())*Tools_Mode_GUIMenu::GetZoom();
		fY = (fY + Tools_Mode_GUIMenu::GetOffsetY())*Tools_Mode_GUIMenu::GetZoom();
		fWidth *= Tools_Mode_GUIMenu::GetZoom();
		fHeight *= Tools_Mode_GUIMenu::GetZoom();
	}
#endif

	Asura_Vector_2 xExtent( (fWidth / Asura_Viewport::GetWidth()), (fHeight / Asura_Viewport::GetHeight()) );
	Asura_Vector_2 xCentrePos( (fX / Asura_Viewport::GetWidth()), (fY / Asura_Viewport::GetHeight()) );
	xCentrePos.x += xExtent.x * 0.5f;
	xCentrePos.y += xExtent.y * 0.5f;

	const Asura_ViewingArea xView( xCentrePos.x - xExtent.x/2.0f, xCentrePos.y - xExtent.y/2.0f, xExtent.x, xExtent.y );
	const Asura_ViewingArea xViewBackup = *Asura_Camera::GetViewingArea();
	const Asura_Vector_3 xPositionBackup = *Asura_Camera::GetPosition();
	const Asura_Matrix_3x3 xOrientationBackup = *Asura_Camera::GetOrientation();

	//Camera Position
	//Camera Orientation
	Asura_Matrix_3x3 xOriMat;
	Asura_Camera::SetPosition( m_xCameraOffset );
	m_xCameraOrient.ToMatrix(xOriMat);
	Asura_Camera::SetOrientation(xOriMat);
	Asura_Camera::SetViewingArea(xView);

	if( dbg_bRenderViewport )
	{	
		Asura_Render::Overlay.SetRectXY( (xCentrePos.x - (xExtent.x * 0.5f)) * Asura_Viewport::GetWidth(), (xCentrePos.y - (xExtent.y * 0.5f)) * Asura_Viewport::GetHeight(), (xExtent.x * Asura_Viewport::GetWidth()), (xExtent.y * Asura_Viewport::GetHeight()) );
		Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.Colour = 0x3fff0000;
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();
	}

	Asura_Camera::SetViewingArea(xView);

	const float fOldFOV = Asura_Camera::GetHorizontalFieldOfView();
	const float fAspectRatio = m_bStretch ? 1.0f : (xExtent.x / xExtent.y);
	if(m_bUseVerticalFOV)
	{
		Asura_Camera::SetVerticalFieldOfView( (m_fFOVAngle / m_fZoomLevel), fAspectRatio);
	}
	else
	{
		Asura_Camera::SetHorizontalFieldOfView( (m_fFOVAngle / m_fZoomLevel), fAspectRatio);
	}


	Asura_Render::ApplyCameraSettings();

	m_xRenderOrient.ToMatrix( xOriMat );

	//ShouldCull returns the *inverse* of what we want to reinstate after this.
	bool bCulling = !(Asura_Hierarchy_System::ShouldCull());
	Asura_Hierarchy_System::SetNoCullingFlag(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);

	// Update animation and render
	const u_int uRenderFlags = (ASURA_HIERARCHY_RENDER_NO_SHADOW | ASURA_HIERARCHY_RENDER_NO_MODULELINK_TEST | ASURA_HIERARCHY_RENDER_NO_FRUSTUM_TEST);
	m_pxAnimation->Update( m_xSkinOffset, xOriMat, Asura_Timers::GetFrameTime(), uRenderFlags );
	
	// Restore old settings
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Hierarchy_System::SetNoCullingFlag(bCulling);	
	Asura_Camera::SetViewingArea(xViewBackup);
	Asura_Camera::SetFieldOfView(fOldFOV);
	Asura_Camera::SetPosition(xPositionBackup);
	Asura_Camera::SetOrientation(xOrientationBackup);

	Asura_Render::ApplyCameraSettings();
	Asura_Render::FlushDisplay();

	Asura_Render::AllowFullScreenRendering();

	ASURA_GUIMENU_PROFILE_END_TASK();
}
/* end of function Render */


#ifdef ASURA_TOOLS

/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::UpdateLocalDataFromStyle
 *
 * Author        : Richard May
 *
 * Description   : Update class-specific data from style
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::UpdateLocalDataFromStyle( void )
{
	//Asura_GUIMenu_Element_Animation* pxStyleAnim;

	PARENT::UpdateLocalDataFromStyle();

}
/* end of function UpdateLocalDataFromStyle */


/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::LoadSkinAndAnimFiles
 *
 * Author        : Richard May
 *
 * Description   : Tools-only function: loads all the files required
 *				   by this element.
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::LoadSkinAndAnimFiles( void ) const
{
	if( (m_uSkinID != ASURA_HASH_ID_UNSET) 
	 &&	!Asura_Animation_System::GetSkinByHash( m_uSkinID )
	 && !LoadAsuraFile( Asura_HashedFilename_System::GetFullFilename( m_uSkinID ) ) )
	{
		Asura_String xFullName( GetNameAsChar() );
		
		Asura_GUIMenu_Root* pxParent = GetParentObject();
		while( pxParent )
		{
			Asura_String xTemp = xFullName;
			xFullName = pxParent->GetNameAsChar();
			xFullName += "->";
			xFullName += xTemp;
		
			pxParent = pxParent->GetParentObject();
		}

		Asura_Assertf( false, ("Failed to load skin file for GUI anim element '%s' ", xFullName.GetString()) );
	}

	if( (m_uAnimID != ASURA_HASH_ID_UNSET) 
	 && !Asura_Animation_System::GetAnimByHash( m_uAnimID )
	 && !LoadAsuraFile( Asura_HashedFilename_System::GetFullFilename( m_uAnimID ) ) )
	{
		Asura_String xFullName( GetNameAsChar() );

		Asura_GUIMenu_Root* pxParent = GetParentObject();
		while( pxParent )
		{
			Asura_String xTemp = xFullName;
			xFullName = pxParent->GetNameAsChar();
			xFullName += "->";
			xFullName += xTemp;

			pxParent = pxParent->GetParentObject();
		}

		Asura_Assertf( false, ("Failed to load anim file for GUI anim element '%s' ", xFullName.GetString()) );
	}
}
/* end of function LoadSkinAndAnimFiles */



/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::AddFilesToExportList
 *
 * Author        : Richard May
 *
 * Description   : Tools-only function: adds files used by this 
 *				   element
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::AddFilesToExportList( void )
{		
	if( m_bExportSkinAndAnim )
	{
		// export skin
		if( m_uSkinID != ASURA_HASH_ID_UNSET )
		{
			Tools_Level_Export::AddFileToExportList( Asura_HashedFilename_System::GetFullFilename(m_uSkinID), true );
		}

		// export anim
		if( m_uAnimID != ASURA_HASH_ID_UNSET )
		{
			Tools_Level_Export::AddFileToExportList( Asura_HashedFilename_System::GetFullFilename(m_uAnimID), false );
		}		
	}

	PARENT::AddFilesToExportList();
}
/* end of function AddFilesToExportList */



/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::WriteToChunkStream
 *
 * Author        : Richard May
 *
 * Description   : Tools-only save code
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_ELEMENT_ANIMATION_VERSION, GetType() );

	// -- Write out parent data - Asura_GUIMenu_Element_Base --
	PARENT::WriteToChunkStream( xStream );

	// write out the skin hash ID used for this animation
	xStream << m_uSkinID;
	xStream << m_uAnimID;
	xStream << m_uAnimFlags;
	xStream << m_fFOVAngle;
	xStream << m_fZoomLevel;
	xStream << m_xCameraOffset;
	xStream << m_xRenderOrient;
	xStream << m_xSkinOffset;
	xStream << m_xCameraOrient;
	xStream << m_xRotationSpeed;
	xStream << m_bExportSkinAndAnim;
	xStream << m_bUseVerticalFOV;
	xStream << m_bStretch;

	// -- write out any children --
	WriteChildObjectData( xStream );
}
/* end of function WriteToChunkStream */

/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::ExportToXML
 *
 * Author        : Hussein Elgridly
 *
 * Description   : Tools-only export to XML code
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<Animation_data>\n" );
	pxExporter->IncTabs();
	
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<skin_id>" );
	pxExporter->AddToBufferHex( m_uSkinID );
	pxExporter->AddToBuffer( "</skin_id>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<anim_id>" );
	pxExporter->AddToBufferHex( m_uAnimID );
	pxExporter->AddToBuffer( "</anim_id>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<anim_flags>" );
	pxExporter->AddToBufferHex( m_uAnimFlags );
	pxExporter->AddToBuffer( "</anim_flags>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<fov_angle>" );
	pxExporter->AddToBuffer( m_fFOVAngle );
	pxExporter->AddToBuffer( "</fov_angle>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<zoom_level>" );
	pxExporter->AddToBuffer( m_fZoomLevel );
	pxExporter->AddToBuffer( "</zoom_level>\n" );
	
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<camera_offset x=\"" );
	pxExporter->AddToBuffer( m_xCameraOffset.GetX() );
	pxExporter->AddToBuffer( "\" y=\"" );
	pxExporter->AddToBuffer( m_xCameraOffset.GetY() );
	pxExporter->AddToBuffer( "\" z=\"" );
	pxExporter->AddToBuffer( m_xCameraOffset.GetZ() );
	pxExporter->AddToBuffer( "\" />\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<render_orient x=\"" );
	pxExporter->AddToBuffer( m_xRenderOrient.x );
	pxExporter->AddToBuffer( "\" y=\"" );
	pxExporter->AddToBuffer( m_xRenderOrient.y );
	pxExporter->AddToBuffer( "\" z=\"" );
	pxExporter->AddToBuffer( m_xRenderOrient.z );
	pxExporter->AddToBuffer( "\" />\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<rotation_speed x=\"" );
	pxExporter->AddToBuffer( m_xRotationSpeed.x );
	pxExporter->AddToBuffer( "\" y=\"" );
	pxExporter->AddToBuffer( m_xRotationSpeed.y );
	pxExporter->AddToBuffer( "\" z=\"" );
	pxExporter->AddToBuffer( m_xRotationSpeed.z );
	pxExporter->AddToBuffer( "\" />\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<export_skin_and_anim>" );
	pxExporter->AddToBuffer( m_bExportSkinAndAnim );
	pxExporter->AddToBuffer( "</export_skin_and_anim>\n" );

	pxExporter->AddToBuffer( "<use_vertical_fov>" );
	pxExporter->AddToBuffer( m_bUseVerticalFOV );
	pxExporter->AddToBuffer( "</use_vertical_fov>\n" );

	pxExporter->AddToBuffer( "<stretch>" );
	pxExporter->AddToBuffer( m_bStretch );
	pxExporter->AddToBuffer( "</stretch>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</Animation_data>\n" );

}

/* end of function ExportToXML */

#endif // ASURA_TOOLS

/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::ReadFromChunkStream
 *
 * Author        : Richard May
 *
 * Description   : Tools-only save code
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	// reset anim 
	delete m_pxAnimation;
	m_pxAnimation = NULL;

	// read in anim element specific data
	if( uChunkVersion == 0 )
	{
		xStream >> m_uSkinID;

		if( m_uSkinID != ASURA_HASH_ID_UNSET )
		{
			// read extra anim data
			xStream >> m_uAnimID;	
			xStream >> m_uAnimFlags;
		}
	}
	else
	{
		xStream >> m_uSkinID;
		xStream >> m_uAnimID;
		xStream >> m_uAnimFlags;
	}

	xStream >> m_fFOVAngle;
	xStream >> m_fZoomLevel;
	xStream >> m_xCameraOffset;
	xStream >> m_xRenderOrient;

	if( uChunkVersion >= 5 )
	{
		xStream >> m_xSkinOffset;
		xStream >> m_xCameraOrient;
	}
	else
	{
		m_xRenderOrient.y = Asura_Maths::RadiansToZeroCentredRange(m_xRenderOrient.y + Asura_Maths::Deg180);
	}

	if( uChunkVersion >= 2 )
	{
		xStream >> m_xRotationSpeed;
	}

	if( uChunkVersion >= 3 )
	{
		bool bExportData = false;
		xStream >> bExportData;
#ifdef ASURA_TOOLS
		m_bExportSkinAndAnim = bExportData;
#endif
	}
	if( uChunkVersion >= 4 ) 
	{
		xStream >> m_bUseVerticalFOV; 
		xStream >> m_bStretch;
	}

	// read in the container info
	ReadAllChildObjects( xStream, this);
}
/* end of function ReadFromChunkStream */



/*************************************************************
 *
 * Name          : Asura_GUIMenu_Element_Animation::PlayAnim
 *
 * Author        : Richard May
 *
 * Description   : Animation playing accessor. Creates an animation
 *			       controller using the given data
 *
 *************************************************************/
void Asura_GUIMenu_Element_Animation::PlayAnim( Asura_Hash_ID uAnimID, u_int uAnimFlags )
{
	if( m_pxAnimation )
	{
		Asura_AnimationController_Animation* pxAnimCtrl = m_pxAnimation->GetControllerAnimationFromHash( GetID() );
		if( pxAnimCtrl && (pxAnimCtrl->GetAnimationHash() != uAnimID) )
		{
			m_pxAnimation->DestroyController( GetID() );
			pxAnimCtrl = NULL;
		}

		if( pxAnimCtrl )
		{
			if( pxAnimCtrl->GetControllerFlags() != uAnimFlags )
			{
				pxAnimCtrl->Stop();
				pxAnimCtrl->SetControllerFlags( uAnimFlags );
				pxAnimCtrl->Play();
			}			
		}	
		else
		{
			pxAnimCtrl = m_pxAnimation->CreateControllerAnimation( GetID(), uAnimFlags, 0, uAnimID, NULL );
			if( pxAnimCtrl ) 
			{
				pxAnimCtrl->Play();
			}
		}
	}
}
/* end of function PlayAnim */

// accessors
void Asura_GUIMenu_Element_Animation::SetAmbient( const Asura_RGBA& xColour )
{
	if( m_pxAnimation )
	{
		m_pxAnimation->SetRGB( xColour.m_fR, xColour.m_fG, xColour.m_fB );
	}
}

void Asura_GUIMenu_Element_Animation::SetSkin( Asura_Hash_ID uSkinID )
{
	m_uSkinID = uSkinID;
	
	if( m_pxAnimation && (m_pxAnimation->GetSkinHashID() != uSkinID) )
	{
		m_pxAnimation->SetSkin( uSkinID );
	}
}

void Asura_GUIMenu_Element_Animation::SetAnim( Asura_Hash_ID uAnimID, u_int uAnimFlags )
{
	m_uAnimID    = uAnimID;
	m_uAnimFlags = uAnimFlags;

	if( m_pxAnimation )
	{
		PlayAnim( uAnimID, uAnimFlags );
	}
}

/***************/
/* End of File */
/***************/

#endif // ASURA_USE_GUI_MENU_SYSTEM
