// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Element_Image.cpp												//
//																						//
//	Description: Image element class.													//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "GUIMenu/Asura_GUIMenu_Element_Image.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM


#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "GUIMenu/Asura_GUIMenu_Animator.h"
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"
#include "GUIMenu/Asura_GUIMenu_Modifier.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"
#include "Rendering/Asura_Render.h"

#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "FileAccess/Asura_FileAccess.h"
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#include "GUIMenu/Tools_Mode_GUIMenu.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Misc/Tools_Timed_Messages.h"
#include "TextureTiler/Tools_TextureTiler_Export.h"
#include "TextureTiler/Tools_TextureTiler_TTL.h"
#endif // ASURA_TOOLS

template<typename T> void Asura_Swap( T& A, T& B )
{
	T tmp = A;

	A = B;
	B = tmp;
}

const int Asura_GUIMenu_Element_Image::s_iDefaultTextureFlags = AsuraTextureFlag_8BitAlpha | AsuraTextureFlag_NoMipMaps | AsuraTextureFlag_PS2Allow512 | AsuraTextureFlag_PS2DontHalveSize;

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Element_Image::Asura_GUIMenu_Element_Image(void)
: m_iTextureID( ASURA_TEXTURE_NONE)
, m_uHashTextureName( ASURA_HASH_ID_UNSET)
, m_uTileHash( ASURA_HASH_ID_UNSET)
, m_fRotation( 0.0f)
{
	m_uType = ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE;

	m_axUV[0].u			= 0.0f;
	m_axUV[0].v			= 0.0f;

	m_axUV[1].u			= 1.0f;
	m_axUV[1].v			= 1.0f;

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_IMAGE);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Element_Image::~Asura_GUIMenu_Element_Image(void)
{
	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_IMAGE);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::Copy( const Asura_GUIMenu_Root* pxSource )
{
	PARENT::Copy( pxSource );

	const Asura_GUIMenu_Element_Image* pxSourceImage = static_cast<const Asura_GUIMenu_Element_Image*> (pxSource);
	
#ifdef ASURA_GUIMENU_STORE_NAMES
	SetTexture(pxSourceImage->GetTextureName(),true);
#else
	SetTexture(pxSourceImage->GetTexture());
	SetTextureNameHash(pxSourceImage->GetTextureNameHash());
#endif
	SetUV(0,pxSourceImage->GetUV(0));
	SetUV(1,pxSourceImage->GetUV(1));
	SetRotation( pxSourceImage->GetRotation());

	SetTile( pxSourceImage->GetTileHash() );
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::SetTexture( const Asura_String& xTextureName, bool bFromASR)
{
	if( !xTextureName.IsNull() )
	{
		Asura_String xTexName;

		if( bFromASR == false)
		{	xTexName = "\\GUI\\";	}

		xTexName += xTextureName;

		m_iTextureID = Asura_Texture_Management::GetIndexFromHashID( xTexName.GetHash_IgnoreExtension() );

		if( m_iTextureID == ASURA_TEXTURE_NONE )
		{
			m_iTextureID = Asura_Texture_Management::AddTexture( xTexName);

			if (m_iTextureID != ASURA_TEXTURE_NONE)
			{
				Asura_Texture_Management::SetTextureFlags( m_iTextureID, s_iDefaultTextureFlags);
			}
		}


	#ifdef ASURA_GUIMENU_STORE_NAMES
		m_xTextureName = xTexName;
	#endif 

		m_uHashTextureName = Asura_GetHashID_IgnoreExtension(xTexName.GetString());
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::SetTile( const Asura_Hash_ID uTileHash )
{
	// attempt to retrive tile info from tile management
	const Asura_TiledTextureInfo* pxTileInfo( Asura_TiledTextureManagement::GetTileInfoByHash( uTileHash ) );

	if( pxTileInfo )
	{
		// override texture and uv settings using tile data
		SetTexture( pxTileInfo->m_iIndexOfTextureUsed );
		
		if( !GetFlag( ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_IGNORETILEUV ) )
		{
			// set u + v from tile
			SetUV( 0,pxTileInfo->m_fu1,pxTileInfo->m_fv1 );
			SetUV( 1,pxTileInfo->m_fu2,pxTileInfo->m_fv2 );

			// swap min and max u|v values if associated flag is set
			if( GetFlags() & Asura_GUIMenu_Element_Image::ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_INVERT_X )
			{
				Asura_Swap( m_axUV[0].u,m_axUV[1].u );	
			}

			if( GetFlags() & Asura_GUIMenu_Element_Image::ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_INVERT_Y )
			{
				Asura_Swap( m_axUV[0].v,m_axUV[1].v );	
			}
		}
	}

#ifdef ASURA_TOOLS
	if( pxTileInfo ) 
	{
		// update tile file name and set texture file name to blank
		const Asura_String* const pxFileName( Asura_TiledTextureManagement::GetAsuraFileNameByHash( pxTileInfo->m_uTileHashID ) );

		Asura_Assertf( pxFileName,("Cannot find file name for tiled texture referenced by GUI image element \"%s\"", GenerateFullyQualifiedName().GetString() ) );

		if( pxFileName )
		{
			SetTileFileName( StripProjectPathFromGraphicsFileName( *pxFileName ) );
		}

		SetTextureFileName( "" );
	}
#endif // ASURA_TOOLS
	
	SetTileHash(uTileHash);
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_Image::UpdateLocalDataFromStyle( void)
{
	
	Asura_String xTextureName;
	Asura_GUIMenu_UV xUV;
	u_int u;

	PARENT::UpdateLocalDataFromStyle();

	if( m_pxStyleObject && (m_pxStyleObject->GetType() == Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE) )
	{
		const Asura_GUIMenu_Element_Image* const pxStyleImage( static_cast<Asura_GUIMenu_Element_Image*>(m_pxStyleObject) );

		if( pxStyleImage )
		{
			if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_TEXTURE) == 0)
			{
				SetTexture( pxStyleImage->GetTextureName(), true);
			}

			if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_UVS) == 0)
			{
				for( u=0; u<ASURA_GUIMENU_MAX_UVLIST_SIZE; u++)
				{
					xUV = pxStyleImage->GetUV( u);
					SetUV( u, xUV);
				}
			}

			if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_TILE) == 0)
			{
				SetTile( pxStyleImage->GetTileHash());
			}

			if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_ROTATION) == 0)
			{
				SetRotation( pxStyleImage->GetRotation());
			}
		}
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::TraverseUsedTextures(void (*pfnCallback)(int iTextureID,void* pUserData),void* pUserData)
{
	if( (m_uTileHash == ASURA_HASH_ID_UNSET) && (GetToolsFlags() & ASURA_GUIMENU_TOOLS_FLAG_EXPORTRESOURCES) )
	{
		pfnCallback(m_iTextureID, pUserData);
		PARENT::TraverseUsedTextures(pfnCallback, pUserData);
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::AddFilesToExportList()
{
	if( (m_uTileHash != ASURA_HASH_ID_UNSET) && (GetToolsFlags() & ASURA_GUIMENU_TOOLS_FLAG_EXPORTRESOURCES) )
	{
		Asura_String xFileName( GetTileFileName() ); 
		
		Asura_Assertf( !xFileName.IsBlank(), ( "Could not find file name for tiled texture used in GUI Menu image \"%s\". Tiled Texture will not be included in GUI Menu export", GenerateFullyQualifiedName() ) );

		if( !xFileName.IsBlank() )
		{
			// find name of tiled texture for the target platform
			const Asura_String xPlatformExportPath( Tools_TT_CreateExportPathForPlatform( xFileName,Tools_Level_Export::s_eExportType ) );
			
			// attempt to create platform specific tiled texture if it does not already exist
			if( !Asura_FileAccess::FileExists( xPlatformExportPath.GetString() ) )
			{
				// if tiled texture file does not exist for platform then attempt to create it
				Tools_Timed_Message_System::AddMessage( 0xffffff00,"WARNING Could not find Tiled Texture file \"%s\" for %s platform. Atempting to create it ...",xPlatformExportPath.GetString(),Tools_Level_Export::GetPlatformName( Tools_Level_Export::s_eExportType ) );

				Tools_TT_CreateTiledTextureForPlatform( xFileName,Tools_Level_Export::s_eExportType );
			}

			// add tiled texture file to export list
			if( Asura_FileAccess::FileExists( xPlatformExportPath.GetString() ) )
			{
				Tools_Level_Export::AddFileToExportList( xPlatformExportPath,false );
			}
			else
			{
				Tools_Timed_Message_System::AddMessage( 0x00ff0000,"ERROR Could not find Tiled Texture file \"%s\". Tiled Texture will not be included in GUI Menu export",xPlatformExportPath.GetString() );
			}
		}
	}

	PARENT::AddFilesToExportList();
}
#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::Render( u_int uState)
{
	PARENT::Render( uState);
}

// --------------------------------------------------------------------------------------- 
void Asura_GUIMenu_Element_Image::Render( u_int uState, const Asura_Bounding_Rectangle& xBounds)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_IMAGE);

	// create local copies of render parameters
	Asura_RGBA xImageColour( GetColour());
	ASURA_TRANSLUCENCY_MODE eTranslucencyMode( GetTranslucencyMode());
	int iImageTextureIndex( GetTexture());

	Asura_GUIMenu_UV xUV0( GetUV( 0));
	Asura_GUIMenu_UV xUV1( GetUV( 1));

	float fWidth(xBounds.GetWidth());
	float fHeight(xBounds.GetHeight());
	float fX = xBounds.MinX;
	float fY = xBounds.MinY;

	float fRotation( m_fRotation);

	// save copy of dimensions before appling modifiers|animators 
	const float fPreScaledWidth( fWidth);
	const float fPreScaledHeight( fHeight);

	// apply effects of GUI modifiers and animators to render parameters 
	ApplyModifierChanges( uState, fX, fY, fWidth, fHeight, xImageColour, eTranslucencyMode, iImageTextureIndex, xUV0, xUV1, fRotation);

	// copy parameters to overlay
	Asura_Render::Overlay.Colour			= xImageColour.ToUIntARGB();
	Asura_Render::Overlay.m_bGouraudShaded	= false;
	Asura_Render::Overlay.TextureIndex		= iImageTextureIndex;			        
	Asura_Render::Overlay.TranslucencyMode	= eTranslucencyMode;	

	Asura_Render::Overlay.m_bNonRotated = (fRotation == 0.0f);//true;

	// change position to compensate for scaling of width|height
	if( GetFlag( ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_SCALETOORIGIN ) )
	{
		fX += (fPreScaledWidth - fWidth)   * m_xPosOrigin.x;
		fY += (fPreScaledHeight - fHeight) * m_xPosOrigin.y;
	}

	#ifdef ASURA_TOOLS
	if(Tools_Mode_GUIMenu::ShouldUseOffsetAndZoom())
	{
		//Apply offset to coords from menu mode
		fX = (fX + Tools_Mode_GUIMenu::GetOffsetX())*Tools_Mode_GUIMenu::GetZoom();
		fY = (fY + Tools_Mode_GUIMenu::GetOffsetY())*Tools_Mode_GUIMenu::GetZoom();
		fWidth *= Tools_Mode_GUIMenu::GetZoom();
		fHeight *= Tools_Mode_GUIMenu::GetZoom();
	}
	#endif // ASURA_TOOLS
	
	if( Asura_Render::Overlay.m_bNonRotated)
	{
		Asura_Render::Overlay.SetRectXY( fX, fY, fWidth, fHeight);
	}
	else
	{
		if( GetFlag( ASURA_GUIMENU_ELEMENT_IMAGE_LOCAL_FLAG_POST_ROTATE_SCALE ) )
		{
			Asura_Render::Overlay.SetRotatedRectXY( 0.0f, 0.0f, 1.0f, 1.0f, fRotation);
			for (u_int u = 0; u < 4; ++u)
			{
				Asura_Render::Overlay.X[u] *= fWidth;
				Asura_Render::Overlay.X[u] += fX + (fWidth * 0.5f);
				Asura_Render::Overlay.Y[u] *= fHeight;
				Asura_Render::Overlay.Y[u] += fY + (fHeight * 0.5f);
			}
		}
		else
		{
			Asura_Render::Overlay.SetRotatedRectXY( fX + ( fWidth*0.5f), fY+(fHeight*0.5f), fWidth, fHeight, fRotation);
		}
	}

	// -- top-left
	Asura_Render::Overlay.U[0] = xUV0.u;
	Asura_Render::Overlay.V[0] = xUV0.v;

	// -- top-right
	Asura_Render::Overlay.U[1] = xUV1.u;
	Asura_Render::Overlay.V[1] = xUV0.v;

	// -- bottom-right
	Asura_Render::Overlay.U[2] = xUV1.u;
	Asura_Render::Overlay.V[2] = xUV1.v;

	// -- bottom-left
	Asura_Render::Overlay.U[3] = xUV0.u;
	Asura_Render::Overlay.V[3] = xUV1.v;

	Asura_Render::Overlay.Render();

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 
// -- GetUV function group -------------------------------------------------------------

const Asura_GUIMenu_UV Asura_GUIMenu_Element_Image::GetUV( u_int uIndex) const
{
	Asura_GUIMenu_UV xUV;

	xUV.u = xUV.v = 0.0f;

	Asura_Assert( uIndex < ASURA_GUIMENU_MAX_UVLIST_SIZE, "Asura_GUIMenu_Element_Image::GetUV - uIndex >= MAX_UVLIST_SIZE");
	if( uIndex < ASURA_GUIMENU_MAX_UVLIST_SIZE)
	{
		xUV = m_axUV[uIndex];
	}

	return xUV;
}

// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// -- SetUV function group -------------------------------------------------------------

void Asura_GUIMenu_Element_Image::SetUV( u_int uIndex, float fU, float fV)
{
	Asura_Assert( uIndex < ASURA_GUIMENU_MAX_UVLIST_SIZE, "Asura_GUIMenu_Element_Image::SetUV - uIndex >= MAX_UVLIST_SIZE");
	
	m_axUV[uIndex].u = fU;
	m_axUV[uIndex].v = fV;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Image::SetUV( u_int uIndex, const Asura_GUIMenu_UV& xUV)
{
	SetUV( uIndex, xUV.u, xUV.v);
}

// --------------------------------------------------------------------------------------- 
void Asura_GUIMenu_Element_Image::SetUV( u_int uIndex, const Asura_GUIMenu_UV* pxUV)
{
	SetUV( uIndex, pxUV->u, pxUV->v);
}

// --------------------------------------------------------------------------------------- 
// -- Overide tool helper function -------------------------------------------------------

void Asura_GUIMenu_Element_Image::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	// read in image element specific data --
	xStream >> m_axUV[0].u;
	xStream >> m_axUV[0].v;
	xStream >> m_axUV[1].u;
	xStream >> m_axUV[1].v;

	xStream >> m_uHashTextureName;

	Asura_String xTextureName;
	xStream >> xTextureName;
	SetTexture( xTextureName, true);

	if( uChunkVersion >= 1 )
	{
		// process tile data 
		xStream >> m_uTileHash;
	
#ifdef ASURA_GUIMENU_STORE_NAMES
		xStream >> m_xTileFileName;
#else	
		xStream.SkipString();
#endif // ASURA_GUIMENU_STORE_NAMES

	}

	if( uChunkVersion >= 2)
	{
		xStream >> m_fRotation;
	}

	if( uChunkVersion >= 3 )
	{
		// set tile in non-tools build in version 3 or above
		SetTile( m_uTileHash );			
	}
#ifdef ASURA_TOOLS
	else
	{
		// only set tile in tools build for version < 3
		SetTile( m_uTileHash );	

		// set export resources flag true for older data to replicate original behaviour - NDB
		SetToolsFlags( GetToolsFlags() | ASURA_GUIMENU_TOOLS_FLAG_EXPORTRESOURCES );
	}
#endif // ASURA_TOOLS

	// -- read in the container info
	ReadAllChildObjects( xStream, this);
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_Image::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_ELEMENT_IMAGE_VERSION, GetType());

	// -- Write out parent data - Asura_GUIMenu_Element_Base --
	PARENT::WriteToChunkStream( xStream);

	xStream << m_axUV[0].u;
	xStream << m_axUV[0].v;
	xStream << m_axUV[1].u;
	xStream << m_axUV[1].v;

	xStream << m_uHashTextureName;
	xStream << m_xTextureName;
	xStream << m_uTileHash;
	xStream << m_xTileFileName;
	
	xStream << m_fRotation;
	
	// -- write out any children --
	WriteChildObjectData( xStream);
}

void Asura_GUIMenu_Element_Image::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<Image_data>\n" );
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer("<uv_set>\n");
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv0_u>" ); 
	pxExporter->AddToBuffer( m_axUV[0].u ); pxExporter->AddToBuffer( "</uv0_u>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv0_v>" ); 
	pxExporter->AddToBuffer( m_axUV[0].v ); pxExporter->AddToBuffer( "</uv0_v>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv1_u>" ); 
	pxExporter->AddToBuffer( m_axUV[1].u ); pxExporter->AddToBuffer( "</uv1_u>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<uv1_v>" ); 
	pxExporter->AddToBuffer( m_axUV[1].v ); pxExporter->AddToBuffer( "</uv1_v>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</uv_set>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<hash_texture_name>" );
	pxExporter->AddToBufferHex( m_uHashTextureName );
	pxExporter->AddToBuffer( "</hash_texture_name>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<texture_name>" );
	pxExporter->AddToBufferEscaped( m_xTextureName.GetString() );
	pxExporter->AddToBuffer( "</texture_name>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<rotation>" );
	pxExporter->AddToBuffer( m_fRotation);
	pxExporter->AddToBuffer( "</rotation>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</Image_data>\n" );

}

const Asura_String Asura_GUIMenu_Element_Image::StripProjectPathFromGraphicsFileName( const Asura_String& xFileName )
{
	Asura_String xRtn( xFileName );

	Asura_Assert( !xFileName.IsNull(),"NULL string recived as argument" );

	if( !xRtn.IsNull() && !xRtn.IsBlank() )
	{
		xRtn.MakeLowercase();

		int iLoc(0);

		xRtn.Contains( "\\graphics", &iLoc );
		xRtn = xRtn.SubString( iLoc,xRtn.GetLength() );
		xRtn.TidyFilename();
	}

	return xRtn;
}

#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
// -- Protected Functions ----------------------------------------------------------------


// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------

void Asura_GUIMenu_Element_Image::ApplyModifierChanges( u_int uStateToMatch, float& fPosX, float& fPosY, float& fWidth, float& fHeight, Asura_RGBA& xColour, ASURA_TRANSLUCENCY_MODE& eTranslucencyMode, int& iTextureID, Asura_GUIMenu_UV& xUV0, Asura_GUIMenu_UV& xUV1, float& fRotation)
{
	u_int u, uModifierCount;
	u_int uModifierState;

	Asura_GUIMenu_Modifier* pxMod;

	uModifierCount = GetModifierCount();

	for( u=0; u<uModifierCount; u++)
	{
		pxMod = GetModifier( u);
		Asura_Assert( pxMod, "modifier count / modifiers out of sync?" );
		if (pxMod)
		{
			uModifierState = pxMod->GetStateBeingModified();

			if( ((uModifierState&uStateToMatch) != 0) && pxMod->IsActivationConditionMet() )
			{
				switch( pxMod->GetModifyType())
				{
					case ASURA_GUIMENU_MODIFY_POSITION:
					{
						Asura_Vector_2 xDeltaPos = pxMod->GetVector2Data();

						fPosX += xDeltaPos.x;
						fPosY += xDeltaPos.y;
						break;
					}

					case ASURA_GUIMENU_MODIFY_POSITION_RELATIVE:
					{
						Asura_Vector_2 xDims(Asura_Maths::ZeroVector2);
						Asura_GUIMenu_Menu_Page* pxPage = FindOwnerMenu();

						Asura_Vector_2 xDeltaPos = pxMod->GetVector2Data();

						if( pxPage)
						{
							xDims = pxPage->GetScaledWorldBounds().GetDimension();
						}

						fPosX += xDeltaPos.x * xDims.x;
						fPosY += xDeltaPos.y * xDims.y;
						break;
					}

					case ASURA_GUIMENU_MODIFY_COLOUR:
					{
						xColour = pxMod->GetColourData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TRANSLUCENCY_MODE:
					{
						eTranslucencyMode = (ASURA_TRANSLUCENCY_MODE)pxMod->GetIntData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_WIDTH:
					{
						fWidth = pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_HEIGHT:
					{
						fHeight = pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_TEXTURE:
					{
						iTextureID = pxMod->GetTextureIndexData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_UVS:
					{
						Asura_GUIMenu_UV_Set* pxUVSet = pxMod->GetUVSetData();
						if( pxUVSet)
						{
							xUV0 = pxUVSet->xUV0;
							xUV1 = pxUVSet->xUV1;
						}

						break;
					}

					case ASURA_GUIMENU_MODIFY_ALPHA_SCALED:
					{
						xColour.m_fA *= pxMod->GetFloatData();
						break;
					}

					case ASURA_GUIMENU_MODIFY_IMAGE_ROTATE:
					{
						fRotation = pxMod->GetFloatData();
						break;
					}

					default:
					{
						// possibly just ignore with no assert?
						Asura_Assert(false, "unhandled modify type");
						break;
					}
				}// end switch( pxMod->GetModifyType())
			}
		}
	}// end for all modifiers


	// apply animators from this object and all parent objects
	const Asura_GUIMenu_Root* pxElementItr = this;
	while ( pxElementItr )
	{
		for( u=0; u<pxElementItr->GetAnimatorCount(); u++)
		{
			Asura_GUIMenu_Animator* pxAnim = pxElementItr->GetAnimator( u);
			Asura_Assert( pxAnim, "shouldn't fail");
			ApplyAnimatorChanges( pxAnim, uStateToMatch, fPosX, fPosY, fWidth, fHeight, xColour, eTranslucencyMode, iTextureID, xUV0, xUV1, fRotation );
		}// end loop on animators

		pxElementItr = pxElementItr->GetParentObject();
	}
}

void Asura_GUIMenu_Element_Image::ApplyAnimatorChanges( Asura_GUIMenu_Animator* pxAnim, u_int uStateToMatch, float& fPosX, float& fPosY, float& fWidth, float& fHeight, Asura_RGBA& xColour, ASURA_TRANSLUCENCY_MODE& eTranslucencyMode, int& iTextureID, Asura_GUIMenu_UV& xUV0, Asura_GUIMenu_UV& xUV1, float& fRotation)
{
	if(pxAnim->ShouldUseAnimator(uStateToMatch))
	{
		switch (pxAnim->GetModifyType())
		{
			case ASURA_GUIMENU_MODIFY_POSITION:
			{
				Asura_Vector_2 xDelta = pxAnim->GetVector2Data();

				fPosX += xDelta.x;
				fPosY += xDelta.y;
				break;
			}

			case ASURA_GUIMENU_MODIFY_POSITION_RELATIVE:
			{
				Asura_Vector_2 xDims(Asura_Maths::ZeroVector2);
				Asura_GUIMenu_Menu_Page* pxPage = FindOwnerMenu();

				Asura_Vector_2 xDelta = pxAnim->GetVector2Data();

				if( pxPage)
				{
					xDims = pxPage->GetScaledWorldBounds().GetDimension();
				}

				fPosX += xDelta.x * xDims.x;
				fPosY += xDelta.y * xDims.y;
				break;
			}

			case ASURA_GUIMENU_MODIFY_COLOUR:
			{

				//Ignore the alpha component for now I think , since that wlll probably be more useful.
				//May need to add a flag later to determine which components to use.
				//RAR 25/07/2007
				
				// u_int uAnimColour = pxAnim->GetRGBAData().ToUIntARGB();
				// uColour  = (uColour & 0xff000000) | (uAnimColour & 0x00ffffff);

				xColour.m_fB = pxAnim->GetRGBAData().m_fB;
				xColour.m_fG = pxAnim->GetRGBAData().m_fG;
				xColour.m_fR = pxAnim->GetRGBAData().m_fR;

				break;
			}

			case ASURA_GUIMENU_MODIFY_ALPHA_SCALED:
			{
				xColour.m_fA *= pxAnim->GetFloatData();
				break;
			}

			case ASURA_GUIMENU_MODIFY_WIDTH_SCALED:
			{
				fWidth *= pxAnim->GetFloatData();
				break;
			}

			case ASURA_GUIMENU_MODIFY_HEIGHT_SCALED:
			{
				fHeight *= pxAnim->GetFloatData();
				break;
			}

			case ASURA_GUIMENU_MODIFY_SIZE_SCALE:
			{
				Asura_Vector_2 xScale = pxAnim->GetVector2Data();
				fWidth *= xScale.x;
				fHeight *= xScale.y;
				break;
			}

			case ASURA_GUIMENU_MODIFY_UVS:
			{
				Asura_GUIMenu_UV_Set xUVSet = pxAnim->GetUVSetData();
				
				xUV0 = xUVSet.xUV0;
				xUV1 = xUVSet.xUV1;
				break;
			}

			case ASURA_GUIMENU_MODIFY_IMAGE_ROTATE:
			{
				fRotation = pxAnim->GetFloatData();
				break;
			}

			default:
			{
				//Not sure its worth having an assert here , since if an animator is applied to a widget , then
				//the child elements will get some types fed to it that can't be handled.
				//RAR 17/04/2007
				//Asura_Assert(false, "unhandled data type");
				break;
			}
		} // end switch
	}// end if (state is affected)
	else
	{
		pxAnim->Start();
	}

}

// --------------------------------------------------------------------------------------- 

#endif // ASURA_USE_GUI_MENU_SYSTEM
