// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Element_Path.cpp												//
//																						//
//	Description: element path class.  movement paths for other elements or widgets		//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "Asura_GUIMenu_Element_Path.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM

#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Graph.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"
#include "Rendering/Asura_Render.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Element_Path::Asura_GUIMenu_Element_Path(void)
: m_pxPath( NULL)
, m_fPosition( 0.0f)
{
	m_uType = ASURA_GUIMENU_TYPEID_ELEMENT_PATH;

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_PATH);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Element_Path::~Asura_GUIMenu_Element_Path(void)
{
	if( m_pxPath != NULL)
	{
		delete m_pxPath;
		m_pxPath = NULL;
	}

	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__ELEMENT_PATH);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::Update( float fElapsedTime)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_PATH);

	const u_int uContainerCount = GetContainerCount();
	for( u_int u=0; u<uContainerCount; u++)
	{
		m_ppxContainerParts[u]->Update( fElapsedTime);
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::Render( u_int uState)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__ELEMENT_PATH);

	const u_int uContainerCount = GetContainerCount();
	for( u_int u=0; u<uContainerCount; u++)
	{
		m_ppxContainerParts[u]->Render( uState | m_uStateOverride);
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 
Asura_Vector_2 Asura_GUIMenu_Element_Path::GetChildPosOffset() const
{
	return GetCurrentPathPosition(m_fPosition) - GetCurrentPathPosition(0.0f);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::Copy( const Asura_GUIMenu_Root* pxSource )
{
	PARENT::Copy( pxSource );

	const Asura_GUIMenu_Element_Path* pxSourcePath = static_cast<const Asura_GUIMenu_Element_Path*> (pxSource);

	delete m_pxPath;
	m_pxPath = new Asura_Graph<Asura_Vector_2>( *pxSourcePath->m_pxPath );
	
	m_fPosition = pxSourcePath->m_fPosition;
}

// --------------------------------------------------------------------------------------- 

Asura_Vector_2	Asura_GUIMenu_Element_Path::GetCurrentPathPosition( float fPosition) const
{
	Asura_Vector_2 xRetVec;

	if( m_pxPath != NULL)
	{
		xRetVec = m_pxPath->GetValueAtTime( fPosition);
	}
	else
	{
		xRetVec.Zero();
	}
	if (GetUnscaledLocalBounds().GetWidth() <= 0.0f
		|| GetUnscaledLocalBounds().GetHeight() <= 0.0f)
	{
		xRetVec.Zero();
	}
	else if (GetScaledWorldBounds().GetWidth() > 0.0f
		|| GetScaledWorldBounds().GetHeight() > 0.0f)
	{
		xRetVec.x *= GetScaledWorldBounds().GetWidth() / GetUnscaledLocalBounds().GetWidth();
		xRetVec.y *= GetScaledWorldBounds().GetHeight() / GetUnscaledLocalBounds().GetHeight();
	}

	return xRetVec;
}

// --------------------------------------------------------------------------------------- 
float Asura_GUIMenu_Element_Path::GetPathValueNearestPosition(const Asura_Vector_2& xPos) const
{
	if(!m_pxPath)
	{
		return 0.0f;
	}
	if (GetScaledWorldBounds().GetWidth() <= 0.0f
		|| GetScaledWorldBounds().GetHeight() <= 0.0f)
	{
		return 0.0f;
	}
	
	Asura_Vector_2 xRelativePos(xPos);
	xRelativePos -= GetScaledWorldPos();
	xRelativePos.x *= GetUnscaledLocalBounds().GetWidth() / GetScaledWorldBounds().GetWidth();
	xRelativePos.y *= GetUnscaledLocalBounds().GetHeight() / GetScaledWorldBounds().GetHeight();

	
	u_int uNumDiscreteSamples = m_pxPath->GetNumDiscreteSamples();

	if(uNumDiscreteSamples>=1 && uNumDiscreteSamples<=2)
	{
		//Dealing with a straight line , which ought to account for the majority of paths
		Asura_Vector_2 xLineStart = m_pxPath->GetValueAtTime(0.0f);
		Asura_Vector_2 xLineEnd = m_pxPath->GetValueAtTime(1.0f);

		//Find the point nearest to the line segment
		//Essentially the sam as Asura_Maths::GetNearestPointOnLineSegment() except that I want to get 'lambda' rather
		//than the point itself.

	    const Asura_Vector_2 xLine              = xLineEnd - xLineStart;
	    const float          fLineLengthSquared = xLine.MagnitudeSquared();

	    // Find the nearest point on the segment. If the edge length is zero,
	    // just take the start of the segment to be the nearest point. Otherwise
	    // work out the nearest point on an infinite line, and then clamp that
	    // to be on the segment (so it gets clamped to the start or end points).
	    float fLambda = 0.0f;
	    if ( fLineLengthSquared > 0.0f )
	    {
	        fLambda = Asura_Maths::ClampToRange( ( ( xRelativePos - xLineStart ) * xLine ) / fLineLengthSquared, 0.0f, 1.0f );
	    }

		return fLambda;
	}
	else
	{
		//Dealing with something other than a straight line.
		//Since this path could in theory be fairly complex , just go for the brute force method of sampling
		//the path at various points and getting the closest point.
		//RAR 01/08/2007

		float fBestValue = 0.0f;
		float fBestDistanceSquared = Asura_Maths::LargeFloat;
	
		for(int i=0;i<=100;i++)
		{
			float fValue = i/100.0f;
			Asura_Vector_2 xTestPos = m_pxPath->GetValueAtTime(fValue);
			
			float fDistanceSquared = (xTestPos-xRelativePos).MagnitudeSquared();
			if(fDistanceSquared < fBestDistanceSquared)
			{
				fBestValue = fValue;
				fBestDistanceSquared = fDistanceSquared;
			}
		}

		return fBestValue;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::SetCurrentPathValue( float fPosition)
{
	m_fPosition = fPosition;

	SetScaledWorldBounds(GetScaledWorldBounds());
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::SetPointArray( const Asura_Vector_2* pxPointArray, u_int uPointCount)
{
	u_int i;
	float fStep = 1.0f/(uPointCount-1);

	Asura_Assert( uPointCount >= 2, "Asura_GUIMenu_Element_Path::SetPointArray - uPointCount < 2, needs to be 2 or more points");

	if( m_pxPath != NULL)
	{	delete m_pxPath;	}

	m_pxPath = new Asura_Graph<Asura_Vector_2>( uPointCount, pxPointArray[0], pxPointArray[uPointCount-1]);

	for( i=1; i<uPointCount-1; i++)
	{
		m_pxPath->AddPoint( pxPointArray[i], fStep * (float)i);
	}
	
	m_pxPath->Update();
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Element_Path::GetNumberOfPoints() const
{
	if(!m_pxPath)
	{
		return 0;
	}
	return m_pxPath->GetNumberOfSplinePoints() - 2;

}

// --------------------------------------------------------------------------------------- 


Asura_Vector_2 Asura_GUIMenu_Element_Path::GetPoint(u_int uIndex) const
{
	Asura_Assert(m_pxPath,"No path");

	const Asura_Graph_Point<Asura_Vector_2>* pxPoint = m_pxPath->GetControlPoints();
	//skip the first point , since it is automatically generated
	pxPoint = pxPoint->m_pxNext;
	Asura_Assert(pxPoint && pxPoint->m_pxNext,"");

	for(u_int i=0;i<uIndex;i++)
	{
		pxPoint = pxPoint->m_pxNext;
		Asura_Assert(pxPoint->m_pxNext,"Reached end of point list");
	}

	return pxPoint->m_xValue;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);
	
	delete m_pxPath;
	if (uChunkVersion > 0)
	{
		m_pxPath = new Asura_Graph<Asura_Vector_2>( xStream);
	}
	else
	{
		Asura_Graph<Asura_Vector_3> xOldPath(xStream);
#if defined(ASURA_TOOLS)
		m_pxPath = new Asura_Graph<Asura_Vector_2>;
		Asura_Graph_Point<Asura_Vector_3>* pxOldPoint = xOldPath.GetControlPoints();
		while (pxOldPoint)
		{
			Asura_Vector_2 xPoint(pxOldPoint->m_xValue.x, pxOldPoint->m_xValue.y);
			m_pxPath->AddPointInternal(xPoint, pxOldPoint->m_fTime);
			pxOldPoint = pxOldPoint->m_pxNext;
		}
#else // #ifdef ASURA_TOOLS
		m_pxPath =  NULL;
		Asura_Assert(0, "this version of path not supported outside of tools - please re-export");
#endif // #ifdef ASURA_TOOLS
	}
#ifdef ASURA_TOOLS
	m_pxPath->Update();
#endif // #ifdef ASURA_TOOLS

	// -- read in the container info
	ReadAllChildObjects( xStream, this);

	if (uChunkVersion == 0)
	{
		for(u_int u = 0; u < m_usContainerPartsCount; ++u)
		{
			Asura_GUIMenu_Root* pxRootItem = m_ppxContainerParts[u];
			pxRootItem->SetUnscaledLocalPos(pxRootItem->GetUnscaledLocalPos() + GetCurrentPathPosition(0.0f));
		}
	}
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_Path::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_ELEMENT_PATH_VERSION, GetType());

	PARENT::WriteToChunkStream( xStream);

	// bSamplesOnly must be true when exporting the level, otherwise, tool data needs the full data
	m_pxPath->WriteToChunkStream( xStream, Tools_Level_Export::IsExport() );

	// -- write out any children --
	WriteChildObjectData( xStream);
}


void Asura_GUIMenu_Element_Path::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<Path_data>\n" );
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<graph min=\"" );
	pxExporter->AddToBuffer( m_pxPath->GetDomainMin() );
	pxExporter->AddToBuffer( "\" max=\"" );
	pxExporter->AddToBuffer( m_pxPath->GetDomainMax() );
	pxExporter->AddToBuffer( "\">\n" );
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<spline_points>" );
	pxExporter->AddToBuffer( m_pxPath->GetNumberOfSplinePoints() );
	pxExporter->AddToBuffer( "</spline_points>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<discrete_samples>" );
	pxExporter->AddToBuffer( m_pxPath->GetNumDiscreteSamples() );
	pxExporter->AddToBuffer( "</discrete_samples>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</graph>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</Path_data>\n" );

}
#endif
// --------------------------------------------------------------------------------------- 

Asura_Graph<Asura_Vector_2>* Asura_GUIMenu_Element_Path::GetActiveGraphPointer( void)
{
	return m_pxPath;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Element_Path::SetActiveGraphPointer( Asura_Graph<Asura_Vector_2>* pxPath)
{
	m_pxPath = pxPath;
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Element_Path::UpdateLocalDataFromStyle( void)
{
	PARENT::UpdateLocalDataFromStyle();

	// -- not doing anything with the path inherited values yet..probably will when it comes to init values
}
#endif
// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------

// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------


// --------------------------------------------------------------------------------------- 

#endif // ASURA_USE_GUI_MENU_SYSTEM
