// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Root.cpp															//
//																						//
//	Description: Root class for all widgets, elements and menus							//
//																						//
//	Created: Steve Keen - 9/11/06														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"

// This file's header
#include "Asura_GUIMenu_Root.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
// Asura
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Time_System.h"
#include "Core/Memory/Asura_ResourceAllocator.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "GUIMenu/Asura_GUIMenu_Allocators.h"
#include "GUIMenu/Asura_GUIMenu_Animator.h"
#include "GUIMenu/Asura_GUIMenu_Element_Animation.h"
#include "GUIMenu/Asura_GUIMenu_Element_Base.h"
#include "GUIMenu/Asura_GUIMenu_Element_FMV.h"
#include "GUIMenu/Asura_GUIMenu_Element_Image.h"
#include "GUIMenu/Asura_GUIMenu_Element_Path.h"
#include "GUIMenu/Asura_GUIMenu_Element_StatusText.h"
#include "GUIMenu/Asura_GUIMenu_Element_Text.h"
#include "GUIMenu/Asura_GUIMenu_Menu_Page.h"
#include "GUIMenu/Asura_GUIMenu_Message.h"
#include "GUIMenu/Asura_GUIMenu_Modifier.h"
#include "GUIMenu/Asura_GUIMenu_Root.h"
#include "GUIMenu/Asura_GUIMenu_System.h"

#include "GUIMenu/Asura_GUIMenu_Widget_AnimatingImage.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Button.h"
#include "GUIMenu/Asura_GUIMenu_Widget_CheckBox.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Group.h"
#include "GUIMenu/Asura_GUIMenu_Widget_ListBox.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Numeric.h"
#include "GUIMenu/Asura_GUIMenu_Widget_ProgressBar.h"
#include "GUIMenu/Asura_GUIMenu_Widget_Slider.h"
#include "GUIMenu/Asura_GUIMenu_Widget_TextBox.h"
#include "GUIMenu/Asura_GUIMenu_Widget_TextListBox.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"

#ifdef ASURA_TOOLS
#include "Interface/Asura_Dialog.h"
#endif
#include "Rendering/Asura_RenderStates.h"
#include "Rendering/Asura_Viewport.h"

// Tools
#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#include "GUIMenu/Tools_Mode_GUIMenu.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Tools_System.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Root::Asura_GUIMenu_Root()
: m_uType( ASURA_GUIMENU_TYPEID_NONE)
, m_ppxContainerParts( NULL)
, m_usContainerPartsCount( 0)
, m_usMaxContainerPartsCount( 0)
, m_uInternalDisplayConditionCount( 0)
, m_uHashName( ASURA_HASH_ID_UNSET)
, m_eTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL)
#ifdef ASURA_TOOLS
, m_uStyleChildEditableFlags( 0)
, m_uStyleUseLocalDataFlags( 0)
, m_uHashStyleObject( ASURA_HASH_ID_UNSET)
, m_pxStyleObject( NULL)
, m_uToolsFlags( 0)
#endif
, m_uFlags( 0)
, m_xScaledWorldBounds( 0.0f,0.0f,0.0f,0.0f )
, m_xUnscaledLocalBounds( 0.0f,0.0f,0.0f,0.0f )
, m_ppxModifiers( NULL)
, m_ucModifierCount( 0)
, m_ucMaxModifierCount( 0)
, m_ppxAnimators( NULL)
, m_ucAnimatorCount( 0)
, m_ucMaxAnimatorCount( 0)
, m_uHashID( ASURA_HASH_ID_UNSET)
, m_pxParent( NULL)
, m_uScalingFlags(0)
#ifdef ASURA_TOOLS
, m_uPlatforms(0xffffffff)
#endif
{
	m_xColour.FromUIntRGBA( 0xffffffff);	//make pure white with full alpha.
	
	m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] = POS_NORMAL;
	m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE]   = POS_NORMAL;
	m_xPosOrigin.Zero();

	for( u_int u(0); u<ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS; u++)
	{
		m_axInternalDisplayConditions[u].m_eConditionCompare = ASURA_GUIMENU_CONDITION_NONE;
		m_axInternalDisplayConditions[u].m_iDisplayConditionThreshold = 0;
	}
		
#ifdef ASURA_TOOLS
	AllocateContainerArraySpace( ASURA_GUIMENU_INIT_TOOL_CONTAINER_PARTS);
	AllocateModifierArraySpace( ASURA_GUIMENU_MAX_ALLOWED_MODIFIERS);
	AllocateAnimatorArraySpace( ASURA_GUIMENU_MAX_ALLOWED_ANIMATORS);

	m_xCommentList.ClearList();
#endif
}

// ---------------------------------------------------------------------------------------
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Root::~Asura_GUIMenu_Root()
{
	u_int i;

#ifdef ASURA_GUIMENU_STORE_NAMES
	m_xName.MakeNull();
#endif 

	if( m_ppxContainerParts != NULL)
	{
		for( i=0; i<m_usContainerPartsCount; i++)
		{
			delete m_ppxContainerParts[i];
		}

		delete []m_ppxContainerParts;
		m_ppxContainerParts = NULL;

		m_usContainerPartsCount = 0;
		m_usMaxContainerPartsCount = 0;
	}

	if( m_ppxModifiers != NULL)
	{
		for( i=0; i<m_ucModifierCount; i++)
		{
			delete m_ppxModifiers[i];
			m_ppxModifiers[i] = NULL;
		}

		delete []m_ppxModifiers;
		m_ppxModifiers = NULL;
		m_ucMaxModifierCount = 0;
		m_ucModifierCount = 0;
	}

	if( m_ppxAnimators != NULL)
	{
		for( i=0; i<m_ucAnimatorCount; i++)
		{
			delete m_ppxAnimators[i];
			m_ppxAnimators[i] = NULL;
		}

		delete []m_ppxAnimators;
		m_ppxAnimators = NULL;
		m_ucMaxAnimatorCount = 0;
		m_ucAnimatorCount = 0;
	}

#ifdef ASURA_TOOLS
	Asura_GUIMenu_ToolComment* pxComment;
	for( int j=0; j<m_xCommentList.GetNumEntries(); j++)
	{
		pxComment = m_xCommentList.GetEntryByIndex( j);
		delete pxComment;
	}

	m_xCommentList.ClearList();
#endif // #ifdef ASURA_TOOLS
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

void Asura_GUIMenu_Root::Copy( const Asura_GUIMenu_Root* pxSource)
{
	Asura_Assertf( GetType()==pxSource->GetType(),("Copy source (%u) should have the same type identifier as this (%u)",pxSource->GetType(),GetType()) );

	SetUnscaledLocalBounds(pxSource->GetUnscaledLocalBounds());
	m_xScaledWorldBounds = pxSource->GetScaledWorldBounds();
	SetFlags(pxSource->GetFlags());
	SetColour(pxSource->GetColour());
	SetTranslucencyMode(pxSource->GetTranslucencyMode());
	
	m_xDisplayConditions.Copy(&pxSource->m_xDisplayConditions);

	CopyInternalDisplayConditions( *pxSource );
	CopyModifiers( *pxSource );
	CopyAnimators( *pxSource );

	m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] = pxSource->m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL];
	m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] = pxSource->m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE];
	m_xPosOrigin = pxSource->m_xPosOrigin;

	SetScalingFlags( pxSource->GetScalingFlags() );

	#ifdef ASURA_TOOLS
	m_uPlatforms = pxSource->m_uPlatforms;
	#endif
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::ApplyDisplayConditionSettings( void )
{
	u_int i;

	bool bShouldBeHiddenByConditionVar = !m_xDisplayConditions.IsConditionMet();
	if (bShouldBeHiddenByConditionVar != IsHiddenByConditionVar())
	{
		SetHiddenByConditionVar(bShouldBeHiddenByConditionVar);
		SetRenderingDisabledFlag(bShouldBeHiddenByConditionVar);
	}

	if( bShouldBeHiddenByConditionVar == false)
	{
		if( m_ppxContainerParts)
		{
			for( i=0; i<m_usContainerPartsCount; i++)
			{
				m_ppxContainerParts[i]->ApplyDisplayConditionSettings();
			}
		}
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::Update( float fElapsedTime)
{
	// update display condition vars
	bool bShouldBeHiddenByConditionVar = !m_xDisplayConditions.IsConditionMet();
	if (bShouldBeHiddenByConditionVar != IsHiddenByConditionVar())
	{
		SetHiddenByConditionVar(bShouldBeHiddenByConditionVar);
		SetRenderingDisabledFlag(bShouldBeHiddenByConditionVar);
	}

	// update animator reset conditions
	for( u_int u(0);u<m_ucAnimatorCount;++u )
	{
		m_ppxAnimators[u]->UpdateStartCondition();
	}

	// update all children
	if( bShouldBeHiddenByConditionVar == false)
	{
		if( m_ppxContainerParts)
		{
			for( u_int u(0); u<m_usContainerPartsCount; ++u )
			{
				m_ppxContainerParts[u]->Update( fElapsedTime);
			}
		}
	}
}

// --------------------------------------------------------------------------------------- 

Asura_Vector_2 Asura_GUIMenu_Root::GetUnscaledLocalPos() const
{
	return m_xUnscaledLocalBounds.GetTopLeft();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetUnscaledLocalPos( const Asura_Vector_2& xPos )
{
	m_xUnscaledLocalBounds.Displace(xPos - m_xUnscaledLocalBounds.GetTopLeft());
}

// --------------------------------------------------------------------------------------- 

Asura_Vector_2 Asura_GUIMenu_Root::GetScaledWorldPos() const
{
	return m_xScaledWorldBounds.GetTopLeft();
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Widget_Base* Asura_GUIMenu_Root::FindSelectableWidgetAtPos(const Asura_Vector_2& xPos)
{
	#ifdef ASURA_TOOLS
	if(!IsFlaggedForCurrentPlatform()) return 0;
	#endif

 	for(int i = m_usContainerPartsCount - 1; i >= 0; --i)
	{
		Asura_GUIMenu_Widget_Base* pxWidget = m_ppxContainerParts[i]->FindSelectableWidgetAtPos(xPos);
		if(pxWidget)
		{
			#ifdef ASURA_TOOLS
			if(pxWidget->IsFlaggedForCurrentPlatform())
			#endif
			{
				return pxWidget;
			}
		}
	}
	return 0;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::ContainsPos(const Asura_Vector_2& xPos) const
{
	return m_xScaledWorldBounds.Contains(xPos);
}

bool Asura_GUIMenu_Root::GetCropRectForChildren(Asura_Bounding_Rectangle& xBounds) const
{
	if (m_pxParent)
	{
		return m_pxParent->GetCropRectForChildren(xBounds);
	}
	return false;
}


// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 
// -- GetColour function group -------------------------------------------------------------

const Asura_RGBA& Asura_GUIMenu_Root::GetColour() const
{
	return m_xColour;
}

// --------------------------------------------------------------------------------------- 
// --------------------------------------------------------------------------------------- 

// -- SetColour function group -------------------------------------------------------------
void Asura_GUIMenu_Root::SetColour( float fRed, float fGreen, float fBlue, float fAlpha)
{
	m_xColour.m_fR = fRed;
	m_xColour.m_fG = fGreen;
	m_xColour.m_fB = fBlue;
	m_xColour.m_fA = fAlpha;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetColour( const Asura_RGBA& xColour)
{
	SetColour( xColour.m_fR, xColour.m_fG, xColour.m_fB, xColour.m_fA);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetColour( const Asura_RGBA* pxColour)
{
	Asura_Assert( pxColour != NULL, "Asura_GUIMenu_Root::SetColour - Asura_RGBA* pxColour == NULL");
	SetColour( pxColour->m_fR, pxColour->m_fG, pxColour->m_fB, pxColour->m_fA);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetColour( u_int uRGBAColourValue)
{
	SetColourFromRGBA( uRGBAColourValue);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetColourFromRGBA( u_int uRGBAColourValue)
{
	m_xColour.FromUIntRGBA( uRGBAColourValue);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetColourFromARGB( u_int uARGBColourValue)
{
	m_xColour.FromUIntARGB( uARGBColourValue);
}

// --------------------------------------------------------------------------------------- 
// -- Flag Functions -----------------------------------------------------------------

// only changes local flag - this is assumed by the runtime flags
void Asura_GUIMenu_Root::ChangeFlag( u_int uNewFlag, bool bTF)
{
	if( bTF)
	{
		m_uFlags |= uNewFlag;
	}
	else
	{
		m_uFlags &= (0xffffffff - uNewFlag);
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::GetFlag( u_int uFlag) const
{
	return ((m_uFlags&uFlag) != 0);
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Root::GetFlags() const
{
	return m_uFlags;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::ToggleFlag( u_int uFlag)
{
	m_uFlags ^= uFlag;
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Menu_Page* Asura_GUIMenu_Root::FindOwnerMenu() const
{
	Asura_GUIMenu_Root* pxOwnerMenu = NULL;

	pxOwnerMenu = GetParentObject();

	if( pxOwnerMenu != NULL)
	{
		while( pxOwnerMenu->GetParentObject())
		{
			pxOwnerMenu = pxOwnerMenu->GetParentObject();
		}
	}

	if( pxOwnerMenu && (pxOwnerMenu->GetType() == ASURA_GUIMENU_TYPEID_MENUPAGE) )
	{
		return static_cast<Asura_GUIMenu_Menu_Page*>(pxOwnerMenu);
	}

	return NULL;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetName(const Asura_String& xName)
{
#ifdef ASURA_GUIMENU_STORE_NAMES
	Asura_Assert( xName.IsNull() == false, "Asura_GUIMenu_Widget_Base::SetName - strName is null");
	Asura_Assert( xName.IsBlank() == false, "Asura_GUIMenu_Widget_Base::SetName - strName is empty");
	
	m_xName = xName;
#endif
	m_uHashName = Asura_GetHashID( xName.GetString());
}

void Asura_GUIMenu_Root::SetName( const char* szName)
{
#ifdef ASURA_GUIMENU_STORE_NAMES
	Asura_Assert( szName && *szName , "Asura_GUIMenu_Widget_Base::SetName - szName is blank/null");
	
	m_xName = szName;
#endif
	m_uHashName = Asura_GetHashID( szName);
}


// --------------------------------------------------------------------------------------- 

#ifdef ASURA_GUIMENU_STORE_NAMES
void Asura_GUIMenu_Root::ForceGenerateAllSubIDs()
{
	u_int i;

	GenerateHashID();

	for( i=0; i<m_usContainerPartsCount; i++)
	{
		m_ppxContainerParts[i]->ForceGenerateAllSubIDs();	
	}
}
#endif	// ASURA_GUIMENU_STORE_NAMES

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::GenerateHashID()
{
#ifdef ASURA_GUIMENU_STORE_NAMES
	SetID( Asura_GetHashID( GenerateFullyQualifiedName().GetString() ) );
#else

	// SJL 10.6.08 I'm assuming that any hash that already exists was generated in a ASURA_GUIMENU_STORE_NAMES build (tools etc)
	// and doesn't need to be changed in response to this function call. The random hash generation here might be needed if an
	// object is created at runtime, in which case I'm expecting its hash to be unset.

	// Lets not break any indexing/linkage that might work.
	if (GetID() == ASURA_HASH_ID_UNSET)
	{
		// Feel free to remove this assert if your happy that everything is ok here, its here just to bring your attention to the comets above.
		Asura_Assert(0, "GUIMenu_Root::GenerateHashID Generating a random hash in response to a call in a non ASURA_GUIMENU_STORE_NAMES build.");

		// Generate a unique hash
		Asura_Hash_ID uID = GetID();
		do
		{
			uID = Asura_Random::GetIntU( 1, UINT_MAX-1 );
		}
		while( Asura_GUIMenu_System::FindRootObjectByID(uID) );
	
		SetID( uID );	
	}

#endif 	
}

// --------------------------------------------------------------------------------------- 


#ifdef ASURA_GUIMENU_STORE_NAMES
const Asura_String Asura_GUIMenu_Root::GenerateFullyQualifiedName() const
{
	Asura_GUIMenu_Root* pxParent = GetParentObjectForIDGeneration();
	Asura_String xActual = "";
	Asura_String xTemp;

	while( pxParent != NULL)
	{
		xTemp = xActual;
		
		xActual = pxParent->GetNameAsChar();
		xActual += "/";

		xActual += xTemp;

		pxParent = pxParent->GetParentObjectForIDGeneration();
	}

	xActual += m_xName;
	
	return xActual;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::OutputGUIMenuIDInfo( const Asura_String& xOutputFileName)
{
	u_int i;
	Asura_GUIMenu_Root* pxParent = GetParentObject();
	Asura_String xActual = "";
	Asura_String xTemp;

	while( pxParent != NULL)
	{
		xTemp = xActual;

		xActual = pxParent->GetNameAsChar();
		xActual += "/";

		xActual += xTemp;

		pxParent = pxParent->GetParentObject();
	}

	xActual += m_xName;

	Asura_SimpleLogfileOutput(( xOutputFileName.GetString(), "0x%08x - %s\n", GetID(), xActual.GetString()));

	for( i=0; i<m_usContainerPartsCount; i++)
	{
		m_ppxContainerParts[i]->OutputGUIMenuIDInfo( xOutputFileName);	
	}
}
#endif //ASURA_GUIMENU_STORE_NAMES

void Asura_GUIMenu_Root::SetID( Asura_Hash_ID uHashID )
{ 
	if( m_pxParent && (uHashID != m_uHashID) )
	{
		m_pxParent->ChildIDChanged( m_uHashID, uHashID );
	}

	m_uHashID = uHashID;	
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::TextInputEnded()
{
	Asura_Assert(false, "should be handled by derived class");
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Root* Asura_GUIMenu_Root::FindContainerObject( const Asura_Hash_ID uHashID) const
{
	u_int i;
	Asura_GUIMenu_Root* pxObject = NULL;
	Asura_GUIMenu_Root* pxTemp;
	for( i=0; i<m_usContainerPartsCount; i++)
	{
		pxTemp = NULL;
		if( m_ppxContainerParts[i]->GetID() == uHashID)
		{
			pxObject = m_ppxContainerParts[i];
			break;
		}

		pxTemp = m_ppxContainerParts[i]->FindContainerObject( uHashID);
		if( pxTemp)
		{
			pxObject = pxTemp;
			break;
		}
	}

	return pxObject;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddToThisContainer( Asura_GUIMenu_Root* pxObject)
{
	bool bRet = false;

	Asura_Assert( pxObject != NULL, "Asura_GUIMenu_Root::AddToThisContainer - Asura_GUIMenu_Root* pxObject == NULL");

//#ifdef ASURA_TOOLS
	CheckAndExpandContainerIfNecessary();
//#endif

	if( m_usContainerPartsCount < m_usMaxContainerPartsCount)
	{
		m_ppxContainerParts[m_usContainerPartsCount++] = pxObject;
		pxObject->SetParentObject( this);

		bRet = true;
	}

	return bRet;
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Root* Asura_GUIMenu_Root::GetContainerItem( u_int uIndex) const
{
	Asura_GUIMenu_Root* pxItem = NULL;
	if( uIndex < m_usContainerPartsCount)
	{
		pxItem = m_ppxContainerParts[uIndex];
	}

	return pxItem;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::RemoveFromThisContainer( Asura_GUIMenu_Root* pxObject)
{
	u_int i;
	bool bCopy = false;

	for( i=0; i<m_usContainerPartsCount; i++)
	{
		// -- found a match
		if( m_ppxContainerParts[i] == pxObject)
		{
			m_ppxContainerParts[i] = NULL;
			bCopy = true;	
			continue;
		}

		if( bCopy == true)
		{
			m_ppxContainerParts[i-1] = m_ppxContainerParts[i];
		}
	}

	if( bCopy == true)
	{	m_usContainerPartsCount--;	}

	return bCopy;
}

// --------------------------------------------------------------------------------------- 

// SK - 19.04.07
// An explaination:  This function is now defunct because the root class now has an expanding container system 
// in tools. It's still here until the menu page code is removed, so menu page widgets use the same container
// system as all widgets and elements.  Therefor, for now, this function will just check to see if container expansion
// is necessary for widgets and elements but will always return false because there is no limit as such.
// The menu page version of IsContainerFull(), for now, will still work as it did.
bool Asura_GUIMenu_Root::IsContainerFull()
{
//	return ( m_uNewContainerPartsCount >= m_uNewMaxContainerPartsCount);
	CheckAndExpandContainerIfNecessary();

	return false;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::CheckAndExpandContainerIfNecessary()
{
	if( m_usContainerPartsCount >= m_usMaxContainerPartsCount)
	{
		ExpandContainerArraySpace();
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::IsContainerEmpty()
{
	return (m_usContainerPartsCount == 0);
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Root* Asura_GUIMenu_Root::FindContainerItemByID( Asura_Hash_ID uHashID) const
{
	u_int i;
	Asura_GUIMenu_Root* pxItem = NULL;
	Asura_GUIMenu_Root* pxFoundItem = NULL;

	for( i=0; i<GetContainerCount(); i++)
	{
		pxItem = GetContainerItem( i);
		if( pxItem->GetID() == uHashID)
		{
			pxFoundItem = pxItem;
			break;
		}

		pxFoundItem = pxItem->FindContainerItemByID( uHashID);
		if( pxFoundItem != NULL)
		{	break;	}
	}

	return pxFoundItem;
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
Asura_GUIMenu_Root* Asura_GUIMenu_Root::FindContainerItemByLocalName( const Asura_String& xItemName)
{
	u_int i;
	Asura_GUIMenu_Root* pxItem = NULL;
	Asura_GUIMenu_Root* pxFoundItem = NULL;

	for( i=0; i<GetContainerCount(); i++)
	{
		pxItem = GetContainerItem( i);
		
		const Asura_String* pxName = pxItem->GetNamePointer();

		if( !pxName->Compare( xItemName))
		{
			pxFoundItem = pxItem;
			break;
		}

		pxFoundItem = pxItem->FindContainerItemByLocalName( xItemName);
		if( pxFoundItem != NULL)
		{	break;	}

	}

	return pxFoundItem;
}
#endif // ASURA_TOOLS
// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Root* Asura_GUIMenu_Root::FindContainerItemByLocalNameHash( Asura_Hash_ID uHashLocalName)
{
	u_int i;
	Asura_GUIMenu_Root* pxItem = NULL;
	Asura_GUIMenu_Root* pxFoundItem = NULL;

	for( i=0; i<GetContainerCount(); i++)
	{
		pxItem = GetContainerItem( i);
		if( pxItem->m_uHashName == uHashLocalName)
		{
			pxFoundItem = pxItem;
			break;
		}

		pxFoundItem = pxItem->FindContainerItemByLocalNameHash( uHashLocalName);
		if( pxFoundItem != NULL)
		{	break;	}
	}

	return pxFoundItem;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::SetInternalDisplayCondition( const int iThreshold, const Asura_GUIMenu_Display_Condition eCondition)
{
	bool bRtn(false);

	if( ReplaceInternalDisplayCondition( 0,iThreshold,eCondition ) )
	{
		m_uInternalDisplayConditionCount = 1;
		bRtn = true;
	}

	return bRtn;	
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddInternalDisplayCondition( const int iThreshold, const Asura_GUIMenu_Display_Condition eCondition)
{
	bool bRtn(false);

	if( ReplaceInternalDisplayCondition( m_uInternalDisplayConditionCount,iThreshold,eCondition ) )
	{
		++m_uInternalDisplayConditionCount;
		bRtn = true;
	}

	return bRtn;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::ReplaceInternalDisplayCondition( const u_int uIndex, const int iThreshold, const Asura_GUIMenu_Display_Condition eCondition)
{
	Asura_Assertf( uIndex < ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS, ("%s - No more condition slots available",__FUNCTION__) );

	bool bRtn(false);

	if( uIndex < ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS)
	{
		m_axInternalDisplayConditions[uIndex].m_iDisplayConditionThreshold = iThreshold;
		m_axInternalDisplayConditions[uIndex].m_eConditionCompare = eCondition;

		bRtn = true;
	}

	return bRtn;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::RemoveAllInternalDisplayConditions()
{
	for( u_int u(0); u<ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS; ++u )
	{
		ReplaceInternalDisplayCondition( u,0,ASURA_GUIMENU_CONDITION_NONE );
	}

	m_uInternalDisplayConditionCount = 0;
}

// --------------------------------------------------------------------------------------- 

int Asura_GUIMenu_Root::GetInternalDisplayConditionThreshold( const u_int uIndex) const
{
	return m_axInternalDisplayConditions[uIndex].m_iDisplayConditionThreshold;
}

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Display_Condition Asura_GUIMenu_Root::GetInternalDisplayConditionCompare( const u_int uIndex) const
{
	return m_axInternalDisplayConditions[uIndex].m_eConditionCompare;
}

// --------------------------------------------------------------------------------------- 

const Asura_GUIMenu_Display_Condition_Params* Asura_GUIMenu_Root::GetInternalDisplayCondition( const u_int uIndex) const
{
	return &m_axInternalDisplayConditions[uIndex];
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Root::GetInternalDisplayConditionCount() const
{ 
	return m_uInternalDisplayConditionCount;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::CopyInternalDisplayConditions( const Asura_GUIMenu_Root& xSource )
{
	RemoveAllInternalDisplayConditions();
	
	const Asura_GUIMenu_Display_Condition_Params* pxParams(NULL);

	for( u_int u(0);u<xSource.GetInternalDisplayConditionCount();++u )
	{
		pxParams = xSource.GetInternalDisplayCondition(u);

		if( pxParams )
		{
			AddInternalDisplayCondition( pxParams->m_iDisplayConditionThreshold,pxParams->m_eConditionCompare );
		}
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetUnscaledLocalBounds( const Asura_Bounding_Rectangle& xBounds )
{
	m_xUnscaledLocalBounds = xBounds;

#ifdef ASURA_TOOLS
	if ( GetType() == ASURA_GUIMENU_TYPEID_MENUPAGE )
	{
		// check size of menu page for errors
		if ( m_xUnscaledLocalBounds.MaxX > 1280.0f ||
			m_xUnscaledLocalBounds.MaxY > 720 )
		{
			// if tiled texture file does not exist for platform then attempt to create it
			if ( m_xName.GetString() != NULL)
			{
				Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR,"ERROR ***** BAD things are happening ***** Menu Page Size is increasing past 1280x720 \"%s\" ",m_xName.GetString() );
			}
		}
	}
#endif
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::Translate( const Asura_Vector_2& xOffset )
{
	static const float fEpsilon(0.01f);

	// only move element if its position is actually going to change
	if( (Asura_Maths::Abs( xOffset.x ) > fEpsilon) || (Asura_Maths::Abs( xOffset.y ) > fEpsilon) )
	{
		// reposition this object
		m_xScaledWorldBounds.Displace( xOffset );

		// update child quads as appropriate
		const u_int uChildMax( GetResizeChildCount() );

		Asura_GUIMenu_Root* pxChild(NULL); 

		for (u_int u(0); u < uChildMax; ++u)
		{	
			pxChild = GetResizeChild(u);

			Asura_Assertf( pxChild,("GUI Menu Resize(): '%s' contains NULL child object",GenerateFullyQualifiedName().GetString()) );

			if (pxChild)
			{
				pxChild->Translate( xOffset );
			}
		}
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetScaledWorldBounds( const Asura_Bounding_Rectangle& xBounds )
{
	if (Asura_GUIMenu_System::HasVirtualResolution())
	{
		Asura_Bounding_Rectangle xScaledBounds(xBounds);
		xScaledBounds.MinX /= Asura_GUIMenu_System::GetVirtualScreenScale().x;
		xScaledBounds.MaxX /= Asura_GUIMenu_System::GetVirtualScreenScale().x;
		xScaledBounds.MinY /= Asura_GUIMenu_System::GetVirtualScreenScale().y;
		xScaledBounds.MaxY /= Asura_GUIMenu_System::GetVirtualScreenScale().y;
		Resize(xScaledBounds);
	}
	else
	{
		Resize(xBounds);
	}
}

void Asura_GUIMenu_Root::Resize( const Asura_Bounding_Rectangle& xBounds )
{	
	// set new quad
	m_xScaledWorldBounds = xBounds;
	if (Asura_GUIMenu_System::HasVirtualResolution())
	{
		m_xScaledWorldBounds.MinX *= Asura_GUIMenu_System::GetVirtualScreenScale().x;
		m_xScaledWorldBounds.MaxX *= Asura_GUIMenu_System::GetVirtualScreenScale().x;
		m_xScaledWorldBounds.MinY *= Asura_GUIMenu_System::GetVirtualScreenScale().y;
		m_xScaledWorldBounds.MaxY *= Asura_GUIMenu_System::GetVirtualScreenScale().y;
	}
	bool bHasVirtualSize = false;
	const Asura_GUIMenu_Menu_Page* pxParentMenu = 0;
	if( GetType() == ASURA_GUIMENU_TYPEID_MENUPAGE)
	{
		pxParentMenu = static_cast<Asura_GUIMenu_Menu_Page*>(this);
	}
	else
	{
		pxParentMenu = FindOwnerMenu();
	}
	Asura_Vector_2 xVirtualScale(1.0f, 1.0f);
	if (pxParentMenu
		&& pxParentMenu->HasVirtualSize())
	{
		bHasVirtualSize = true;
		// NY - 20 November 2008 - don't scale when in tools edit mode
#ifdef ASURA_TOOLS
		if (!Tools_System::IsActive()
			|| Tools_Mode_GUIMenu::IsPreview())
#endif // #ifdef ASURA_TOOLS
		{
			xVirtualScale.x = pxParentMenu->GetVirtualSize().x / pxParentMenu->GetUnscaledLocalBounds().GetWidth();
			xVirtualScale.y = pxParentMenu->GetVirtualSize().y / pxParentMenu->GetUnscaledLocalBounds().GetHeight();
		}
	}

	// calculate the fractional change in the width|height of this object
	Asura_Vector_2 xParentScale( 1.0f,1.0f );

	if( m_xUnscaledLocalBounds.GetWidth() > 0.0f )
	{
		xParentScale.x = xBounds.GetWidth() / (m_xUnscaledLocalBounds.GetWidth() * xVirtualScale.x);
	}

	if( m_xUnscaledLocalBounds.GetHeight() > 0.0f )
	{
		xParentScale.y = xBounds.GetHeight() / (m_xUnscaledLocalBounds.GetHeight() * xVirtualScale.y);
	}


	// update child quads as appropriate
	const u_int uChildMax( GetResizeChildCount() );

	for (u_int u =0; u < uChildMax; ++u)
	{	
		Asura_GUIMenu_Root* pxChild = GetResizeChild(u);

		Asura_Assertf( pxChild,("GUI Menu Resize(): '%s' contains NULL child object",GenerateFullyQualifiedName().GetString()) );

		if (pxChild)
		{
			Asura_Bounding_Rectangle xChildQuad(pxChild->GetUnscaledLocalBounds());
			if (bHasVirtualSize)
			{
				xChildQuad.MinX *= xVirtualScale.x;
				xChildQuad.MaxX *= xVirtualScale.x;
				xChildQuad.MinY *= xVirtualScale.y;
				xChildQuad.MaxY *= xVirtualScale.y;
			}
			Asura_Vector_2 xAnchor = xChildQuad.GetTopLeft();
			xAnchor.x += pxChild->GetPosOrigin().x * xChildQuad.GetWidth();
			xAnchor.y += pxChild->GetPosOrigin().y * xChildQuad.GetHeight();
			xChildQuad.Displace(Asura_Maths::ZeroVector2 - xAnchor);

			for (u_int uAxis = 0; uAxis < ASURA_GUIMENU_POSITION_MODE_AXIS_MAX; ++uAxis)
			{
				switch (pxChild->GetPosMode(uAxis))
				{
					default:
					{
						Asura_Assert(0, "unhandled guimenu pos mode");
					}//fallthrough
					case POS_NORMAL:
					{
						//do nothing
						break;
					}
					case POS_CENTRE:
					{
						xAnchor[uAxis] += 0.5f * (xBounds.GetDimension()[uAxis] - m_xUnscaledLocalBounds.GetDimension()[uAxis] * xVirtualScale[uAxis]);
						break;
					}
					case POS_FAR:
					{
						xAnchor[uAxis] += xBounds.GetDimension()[uAxis] - m_xUnscaledLocalBounds.GetDimension()[uAxis] * xVirtualScale[uAxis];
						break;
					}
					case POS_RELATIVE:
					{
						xAnchor[uAxis] *= xParentScale[uAxis];
						break;
					}
					case POS_RELATIVE_ALTAXIS:
					{
						xAnchor[uAxis] *= xParentScale[1 - uAxis];
						break;
					}
				}
			}

			// find child scale factor 
			Asura_Vector_2 xChildDimensionScale = pxChild->GetRelativeDimensionScale(xParentScale);

			xChildQuad.MinX *= xChildDimensionScale.x;
			xChildQuad.MaxX *= xChildDimensionScale.x;
			xChildQuad.MinY *= xChildDimensionScale.y;
			xChildQuad.MaxY *= xChildDimensionScale.y;

			xChildQuad.Displace(xAnchor);
			Asura_Vector_2 xChildPosOffset = GetChildPosOffset();
			if (Asura_GUIMenu_System::HasVirtualResolution())
			{
				xChildPosOffset.x /= Asura_GUIMenu_System::GetVirtualScreenScale().x;
				xChildPosOffset.y /= Asura_GUIMenu_System::GetVirtualScreenScale().y;
			}
			xChildQuad.Displace(xBounds.GetTopLeft() + xChildPosOffset);

			pxChild->Resize(xChildQuad);
		}
	}

}



// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::IsChildObjectBeingScaled()
{
	u_int u;
	bool bRetValue = false;

	Asura_GUIMenu_Root* pxChildRoot = NULL;
	Asura_GUIMenu_Element_Base* pxChildElement = NULL;

	// -- check all child objects for a scaling flag being set --
	for( u=0; u<m_usContainerPartsCount; u++)
	{
		pxChildRoot = m_ppxContainerParts[u];

		bRetValue = m_ppxContainerParts[u]->IsChildObjectBeingScaled();
		if( !bRetValue)
		{
			if( (pxChildRoot->GetType() >= ASURA_GUIMENU_TYPEID_ELEMENT_BEGIN) &&
				(pxChildRoot->GetType() <= ASURA_GUIMENU_TYPEID_ELEMENT_END))
			{
				pxChildElement = static_cast<Asura_GUIMenu_Element_Base*>(pxChildRoot);
				
				if( pxChildElement->GetScalingFlag(static_cast<Asura_GUIMenu_Scaling_Flag>(0xFFFFFFFF)))//check if any flags set
				{
					bRetValue = true;
					break;
				}
			}
		}
		else
		{
			break;
		}
	}

	return bRetValue;
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
void Asura_GUIMenu_Root::RenderDebugBounds( void)
{
	u_int u;

	Asura_Render::Render2DLine( m_xScaledWorldBounds.MinX, m_xScaledWorldBounds.MinY, m_xScaledWorldBounds.MaxX, m_xScaledWorldBounds.MinY, 0xffffffff);
	Asura_Render::Render2DLine( m_xScaledWorldBounds.MaxX, m_xScaledWorldBounds.MinY, m_xScaledWorldBounds.MaxX, m_xScaledWorldBounds.MaxY, 0xffffffff);
	Asura_Render::Render2DLine( m_xScaledWorldBounds.MinX, m_xScaledWorldBounds.MaxY, m_xScaledWorldBounds.MaxX, m_xScaledWorldBounds.MaxY, 0xffffffff);
	Asura_Render::Render2DLine( m_xScaledWorldBounds.MinX, m_xScaledWorldBounds.MinY, m_xScaledWorldBounds.MinX, m_xScaledWorldBounds.MaxY, 0xffffffff);

	for( u=0; u<m_usContainerPartsCount; u++)
	{
		m_ppxContainerParts[u]->RenderDebugBounds();
	}
}
#endif // ASURA_DEBUG_VARIABLE_SYSTEM
// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::ChangeStyleUseLocalDataFlag( u_int uNewFlag, bool bTF)
{
	if( bTF)
	{
		m_uStyleUseLocalDataFlags |= uNewFlag;
	}
	else
	{
		m_uStyleUseLocalDataFlags &= (0xffffffff - uNewFlag);
	}

}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::GetStyleUseLocalDataFlag( u_int uFlag) const
{
	return ((m_uStyleUseLocalDataFlags&uFlag) != 0);
}
#endif // #ifdef ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
// -- Protected Functions ---------------------------------------------------------------- 

bool Asura_GUIMenu_Root::IsDisplayConditionMet( const int iValidate)
{
	const Asura_GUIMenu_Display_Condition_Params* pxParams;
	bool bCondition = true;

	for( u_int u(0); u<GetInternalDisplayConditionCount(); ++u )
	{
		pxParams = GetInternalDisplayCondition( u);
		
		switch( pxParams->m_eConditionCompare)
		{

			case ASURA_GUIMENU_CONDITION_EQUAL:
				bCondition = (iValidate == pxParams->m_iDisplayConditionThreshold);
				break;

			case ASURA_GUIMENU_CONDITION_NOT_EQUAL:
				bCondition = (iValidate != pxParams->m_iDisplayConditionThreshold);
				break;

			case ASURA_GUIMENU_CONDITION_GREATER:
				bCondition = (iValidate > pxParams->m_iDisplayConditionThreshold);
				break;

			case ASURA_GUIMENU_CONDITION_GREATER_EQUAL:
				bCondition = (iValidate >= pxParams->m_iDisplayConditionThreshold);
				break;

			case ASURA_GUIMENU_CONDITION_LESS:
				bCondition = (iValidate < pxParams->m_iDisplayConditionThreshold);
				break;

			case ASURA_GUIMENU_CONDITION_LESS_EQUAL:
				bCondition = (iValidate <= pxParams->m_iDisplayConditionThreshold);
				break;

			case ASURA_GUIMENU_CONDITION_NONE:
			default:
				bCondition = true;
				break;
		}

		if( bCondition == false)
		{	break;	}
	}

	return bCondition;
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::SetStyleObjectHash( Asura_Hash_ID uHashObjectID)
{
	Asura_Assertf( uHashObjectID != GetID(),("%s: Atempting to set style ID to objects own ID",GenerateFullyQualifiedName().GetString() ) );

	m_uHashStyleObject = uHashObjectID;	
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetStyleObject( Asura_GUIMenu_Root* pxStyleObject)
{
	m_pxStyleObject = pxStyleObject;
	
	SetStyleObjectHash( (pxStyleObject)? pxStyleObject->GetID() : ASURA_HASH_ID_UNSET );
}
#endif

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::UpdateLocalDataFromStyle()
{
	Asura_RGBA xColour;
	const Asura_GUIMenu_Display_Condition_Params* pxDisplayCondition;
	Asura_Vector_2	xLocalPos;

	Asura_Assertf( GetStyleObjectHash() != GetID(),("%s: Style ID matches objects own ID. Aborting style update.",GenerateFullyQualifiedName().GetString() ) );

	if( GetStyleObjectHash() == GetID() )
	{
		return;
	}

	if( (m_uHashStyleObject != ASURA_HASH_ID_UNSET) && (m_pxStyleObject == NULL))
	{
		m_pxStyleObject = Asura_GUIMenu_System::FindRootObjectByID( m_uHashStyleObject);
	}

	if( m_pxStyleObject != NULL)
	{
		m_uStyleUseLocalDataFlags = m_pxStyleObject->GetStyleChildEditableDataFlags();

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_POSITION) == 0)
		{
			xLocalPos = m_pxStyleObject->GetUnscaledLocalPos();
			SetUnscaledLocalPos( xLocalPos);
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_COLOUR) == 0)
		{
			xColour = m_pxStyleObject->GetColour();
			SetColour( xColour);
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_DISPLAY_CONDITIONS) == 0)
		{
			RemoveAllInternalDisplayConditions();

			for( u_int u(0); u<m_pxStyleObject->GetInternalDisplayConditionCount(); u++)
			{
				pxDisplayCondition = m_pxStyleObject->GetInternalDisplayCondition( u);
				AddInternalDisplayCondition( pxDisplayCondition->m_iDisplayConditionThreshold, pxDisplayCondition->m_eConditionCompare);
			}

			m_xDisplayConditions.Copy(&m_pxStyleObject->m_xDisplayConditions);
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_TRANSLUCENCY) == 0)
		{
			SetTranslucencyMode( m_pxStyleObject->GetTranslucencyMode());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_FLAGS) == 0)
		{
			SetFlags( m_pxStyleObject->GetFlags());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_ANCHOR) == 0)
		{
			SetPosMode( ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL,m_pxStyleObject->GetPosMode(ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL) );
			SetPosMode( ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE,m_pxStyleObject->GetPosMode(ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE) );
			m_xPosOrigin = m_pxStyleObject->m_xPosOrigin;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_SCALING) == 0)
		{
			SetScalingFlags( m_pxStyleObject->GetScalingFlags());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_ROOT_LOCAL_FLAG_PLATFORMS) == 0)
		{
			m_uPlatforms = m_pxStyleObject->m_uPlatforms;
		}


		//Deal with modifiers. RAR 04/04/2007
		//Delete any modifiers for data that is controlled by the style.
		for(u_int u=0; u<m_ucModifierCount;)
		{
			if(m_ppxModifiers[u]->GetCorrespondingInheritanceFlag() & ~m_uStyleUseLocalDataFlags)
			{
				DeleteModifier(u);
			}
			else
			{
				u++;
			}
		}

		//And copy new modifiers for that data from the style
		for(u_int u=0; u<m_pxStyleObject->m_ucModifierCount; u++)
		{
			if(m_pxStyleObject->m_ppxModifiers[u]->GetCorrespondingInheritanceFlag() & ~m_uStyleUseLocalDataFlags)
			{
				CopyModifier(m_pxStyleObject->m_ppxModifiers[u]);
			}
			
		}

		//Deal with animators. RAR 16/04/2007
		//Delete any animators that were generated by a style
		for(u_int u=0; u<m_ucAnimatorCount;)
		{
			if(m_ppxAnimators[u]->IsGeneratedByStyle())
			{
				DeleteAnimator(u);
			}
			else
			{
				u++;
			}
		}

		//Copy animators from the style , and flag them as coming from the style
		for(u_int u=0; u<m_pxStyleObject->m_ucAnimatorCount; u++)
		{
			Asura_GUIMenu_Animator* pxAnimator( m_pxStyleObject->m_ppxAnimators[u] );

			if( pxAnimator )
			{
				pxAnimator = AddAnimator( *pxAnimator);
				if(pxAnimator) pxAnimator->SetGeneratedByStyle(true);
			}
		}

	}

	// -- now update any child objects attached to this one --
	for( u_int u(0); u<m_usContainerPartsCount; u++)
	{
		m_ppxContainerParts[u]->UpdateLocalDataFromStyle();
	}
}
#endif
// --------------------------------------------------------------------------------------- 
// -- Overide tool helper function ------------------------------------------------------------

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::TraverseUsedTextures(void (*pfnCallback)(int iTextureID,void* pUserData),void* pUserData)
{
	if(Tools_Level_Export::IsExport() && !ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
	{
		return;
	}
	
	//Check for any modifiers that contain texture data
	for(u_int u=0; u<m_ucModifierCount;u++)
	{
		if(m_ppxModifiers[u]->GetDataType() == Asura_GUIMenu_Modifier::DATA_TYPE_TEXTUREINDEX)
		{
			pfnCallback(m_ppxModifiers[u]->GetTextureIndexData(),pUserData);
		}
	}	

	
	//recurse over children
	u_int uNumChildren = GetContainerCount();

	for( u_int i=0; i<uNumChildren; i++)
	{
		if(Tools_Level_Export::IsExport() && !GetContainerItem(i)->ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
		{
			continue;
		}
		GetContainerItem(i)->TraverseUsedTextures(pfnCallback,pUserData);
	}
}

void Asura_GUIMenu_Root::AddFilesToExportList()
{
	// Not sure what to do with modifiers at this time
	//for(u_int u=0; u<m_ucModifierCount;u++)
	//{
	//	
	//}	


	//recurse over children
	u_int uNumChildren = GetContainerCount();
	for( u_int i=0; i<uNumChildren; i++)
	{
		if(GetContainerItem(i)->ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
		{
			GetContainerItem(i)->AddFilesToExportList();
		}
	}
}

void Asura_GUIMenu_Root::ListUsedTextPages(Asura_HashTree& xTextPageTree) const
{
	if(Tools_Level_Export::IsExport() && !ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
	{
		return;
	}
	
	//recurse over children
	u_int uNumChildren = GetContainerCount();

	for( u_int i=0; i<uNumChildren; i++)
	{
		if(Tools_Level_Export::IsExport() && !GetContainerItem(i)->ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
		{
			continue;
		}
		GetContainerItem(i)->ListUsedTextPages(xTextPageTree);
	}
}


#endif
// --------------------------------------------------------------------------------------- 
// -- Chunk System -----------------------------------------------------------------------

void Asura_GUIMenu_Root::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	const u_int uLocalChunkVersion( ReadVersionNumber( xStream) );

	if( CheckVersionNumber( uLocalChunkVersion,ASURA_GUIMENU_CHUNK_TYPE_ROOT,xStream.GetFileHandle() ) )
	{
		SetParentObject(pxParentObject);

		char acName[256];
		xStream.ReadString( acName, sizeof(acName));
		SetName( acName);

		// -- read colour --
		xStream >> m_xColour.m_fR;
		xStream >> m_xColour.m_fG;
		xStream >> m_xColour.m_fB;
		xStream >> m_xColour.m_fA;

		if (uLocalChunkVersion >= 20)
		{
			xStream >> m_xUnscaledLocalBounds.MinX;
			xStream >> m_xUnscaledLocalBounds.MaxX;
			xStream >> m_xUnscaledLocalBounds.MinY;
			xStream >> m_xUnscaledLocalBounds.MaxY;
		}
		else
		{
			Asura_Vector_2 xLocalPos;
			float fTemp;
			xStream >> xLocalPos.x;
			xStream >> xLocalPos.y;
			xStream >> fTemp;
			m_xUnscaledLocalBounds.Displace(xLocalPos);
		}

		// -- read id hash
		xStream >> m_uHashID;

		// -- read flags --
		xStream >> m_uFlags;

		// -- if the root chunk version is <= 6 than check the hilight flag is set, if it is then set the edit flag to match.
		if( uLocalChunkVersion <= 6)
		{
			if( m_uFlags&ASURA_GUIMENU_ACTION_FLAG_IS_HILIGHTED )
			{
				m_uFlags |= ASURA_GUIMENU_ACTION_FLAG_IS_EDITING;
			}
		}

		int	iDisplayConditionThreshold(0);
		u_int uDisplayCondition(0);

		// -- read display conditions --
		if( uLocalChunkVersion == 0)
		{
			m_uInternalDisplayConditionCount = 1;
			xStream >> m_axInternalDisplayConditions[0].m_iDisplayConditionThreshold;
			xStream >> uDisplayCondition;
			m_axInternalDisplayConditions[0].m_eConditionCompare = (Asura_GUIMenu_Display_Condition)uDisplayCondition;
		}
		else
		{
			// -- Chunk Version 1 or greater --
			xStream >> m_uInternalDisplayConditionCount;
			Asura_Assert( m_uInternalDisplayConditionCount <= ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS, "Asura_GUIMenu_Root::ReadFromChunkStream - Display condition count is bigger than the storage array.");

			for( u_int u(0); u<ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS; ++u )
			{
				m_axInternalDisplayConditions[u].m_iDisplayConditionThreshold = 0;
				m_axInternalDisplayConditions[u].m_eConditionCompare = ASURA_GUIMENU_CONDITION_NONE;
			}

			for( u_int u(0); u<m_uInternalDisplayConditionCount; ++u )
			{
				xStream >> iDisplayConditionThreshold;
				xStream >> uDisplayCondition;
				if( u < ASURA_GUIMENU_MAX_DISPLAY_CONDITIONS)
				{
					m_axInternalDisplayConditions[u].m_iDisplayConditionThreshold = iDisplayConditionThreshold;
					m_axInternalDisplayConditions[u].m_eConditionCompare = (Asura_GUIMenu_Display_Condition)uDisplayCondition;
				}
			}
		}

		// -- read translucency mode --
		u_int uTransMode(0);

		xStream >> uTransMode;
		m_eTranslucencyMode = (ASURA_TRANSLUCENCY_MODE)uTransMode;

		// SK 21/11/07 - this is here because the additive option has been removed as a translucency choice.
		//               if an older version has the mode set, it will change it to the glowing option instead.
		//               This is the recommended choice based in the previous additive setting
		if( m_eTranslucencyMode == ASURA_TRANSLUCENCY_ADD)
		{
			m_eTranslucencyMode = ASURA_TRANSLUCENCY_GLOWING;
		}


		// SK 01.05.07 - chunk version 9 is where this data is no longer written out.
		// so from version 9 onwards this data should not be read in.
		if( (uLocalChunkVersion >= 2) && (uLocalChunkVersion < 9))
		{
			u_int uHashInheritanceID, uInheritanceFlags;
			
			xStream >> uHashInheritanceID;
			xStream >> uInheritanceFlags;

//			SetInheritedObjectHash( uHashInheritanceID);
//			SetInheritanceUseLocalDataFlags( uInheritanceFlags);
		}
		
		if( uLocalChunkVersion >= 3)
		{
			u_int uStyleChildEditableDataFlags, uToolsFlags;

			xStream >> uStyleChildEditableDataFlags;
			xStream >> uToolsFlags;


#ifdef ASURA_TOOLS
			if (uLocalChunkVersion < 10)
			{
				//Reset display condition flag to allow all instance to edit display conditions.
				//This is to prevent old data from being messed up , now that console var conditions
				//are controlled by this flag as well.
				//RAR 21/05/2007
				uStyleChildEditableDataFlags |= ASURA_GUIMENU_ROOT_LOCAL_FLAG_DISPLAY_CONDITIONS;
			}
			SetStyleChildEditableDataFlags(uStyleChildEditableDataFlags);
			SetToolsFlags(uToolsFlags);
#endif
		}

		if( uLocalChunkVersion >= 4)
		{
			u_int uStyleHashID, uStyleUseLocalFlags;

			xStream >> uStyleHashID;
			xStream >> uStyleUseLocalFlags;

#ifdef ASURA_TOOLS
			if (uLocalChunkVersion < 10)
			{
				//As above.. reset display condition flag. RAR 21/05/2007
				uStyleUseLocalFlags |= ASURA_GUIMENU_ROOT_LOCAL_FLAG_DISPLAY_CONDITIONS;
			}

			SetStyleObjectHash( uStyleHashID);
			SetStyleUseLocalDataFlags( uStyleUseLocalFlags);
#endif
		}

		if (uLocalChunkVersion >= 5)
		{
			if (uLocalChunkVersion >= 10)
			{
				m_xDisplayConditions.ReadFromChunkStream(xStream);
			}
			else
			{
				//Old chunks will always have one condition var.
				//Read it in ,and delete it if it is blank
				m_xDisplayConditions.m_uNumConditionVars = 1;
				m_xDisplayConditions.m_pxConditionVars = new Asura_GUIMenu_ConditionVar[1];
				m_xDisplayConditions.m_pxConditionVars[0].ReadFromChunkStream(xStream);
				if(m_xDisplayConditions.m_pxConditionVars[0].IsVarBlank() && m_xDisplayConditions.m_pxConditionVars[0].IsCmdBlank())
				{
					delete [] m_xDisplayConditions.m_pxConditionVars;
					m_xDisplayConditions.m_pxConditionVars = 0;
					m_xDisplayConditions.m_uNumConditionVars = 0;
				}
			}
		}

		if( uLocalChunkVersion >= 6)
		{
			u_char ucModifierCount;
			if( uLocalChunkVersion <= 17 )
			{
				u_int uTempCount;
				xStream >> uTempCount;
				ucModifierCount = static_cast<u_char> (uTempCount);
			}
			else
			{
				xStream >> ucModifierCount;
			}

			if( ucModifierCount > 0)
			{
#ifndef ASURA_TOOLS
// -- SK only need to initialise this here if it's a retail build --
				AllocateModifierArraySpace( ucModifierCount);
#endif
				for( u_char u(0); u<ucModifierCount; ++u )
				{
					m_ppxModifiers[u] = new Asura_GUIMenu_Modifier;
					m_ppxModifiers[u]->ReadFromChunkStream( xStream);
				}

				m_ucModifierCount = ucModifierCount;
			}	
		}

		if( uLocalChunkVersion >= 8)
		{
			u_char ucAnimatorCount;
			if( uLocalChunkVersion <= 17 )
			{
				u_int uTempCount;
				xStream >> uTempCount;
				ucAnimatorCount = static_cast<u_char> (uTempCount);
			}
			else
			{
				xStream >> ucAnimatorCount;	
			}
			
			if( ucAnimatorCount > 0)
			{
#ifndef ASURA_TOOLS
// -- MH only need to initialise this here if it's a retail build --
				AllocateAnimatorArraySpace( ucAnimatorCount);
#endif
				for( u_char u(0); u<ucAnimatorCount; ++u )
				{
					m_ppxAnimators[u] = new Asura_GUIMenu_Animator;
					m_ppxAnimators[u]->ReadFromChunkStream( xStream);
				}

				m_ucAnimatorCount = ucAnimatorCount;
			}	

			#ifdef ASURA_TOOLS
			if(uLocalChunkVersion <= 10)
			{
				//Fix cases where the same animator has been added as both generated by style and not generated by style
				//RAR 29/06/2007
				for( u_int uI(0); uI<m_ucAnimatorCount; ++uI )
				{
					const Asura_GUIMenu_Animator* pxAnimator = GetAnimator(uI);
					if(pxAnimator->IsGeneratedByStyle()) continue;

					bool bDelete = false;

					for( u_int uJ(uI+1); uJ<GetAnimatorCount(); ++uJ )
					{
						const Asura_GUIMenu_Animator* pxAnimator2 = GetAnimator(uJ);
						if(!pxAnimator2->IsGeneratedByStyle()) continue;

						if(!strcmp(pxAnimator->GetSequenceName(), pxAnimator2->GetSequenceName()))
						{
							bDelete = true;
						}
					}
					if(bDelete)
					{
						DeleteAnimator(uI);
						--uI;
					}
				}
			}
			#endif //ASURA_TOOLS

			char acComment[1024];

			if( uLocalChunkVersion == 12)
			{
				xStream.ReadString( acComment, sizeof(acComment));
				xStream.ReadString( acName, sizeof(acName));

#ifdef ASURA_TOOLS
				if( acComment[0] != '\0' && acName[0] != '\0')
				{
					AddComment( acName, acComment);
				}
#endif // ASURA_TOOLS
			}

			if( uLocalChunkVersion >= 13)
			{

				u_int uNumComments;
				xStream >> uNumComments;

				for( u_int u=0; u<uNumComments; u++)
				{
#ifdef ASURA_TOOLS
					Asura_GUIMenu_ToolComment* pxToolComment = new Asura_GUIMenu_ToolComment;
					pxToolComment->ReadFromChunkStream( xStream);
					m_xCommentList.AddToList( pxToolComment);
#else // #ifdef ASURA_TOOLS
					Asura_GUIMenu_ToolComment xTemp;
					xTemp.ReadFromChunkStream( xStream);
#endif // #ifdef ASURA_TOOLS
				}
			}

			if( uLocalChunkVersion == 14)
			{
				// Old flags that were reviously definined in the 
				// now depricated Asura_GUIMenu_Anchor.h file
				const u_int uAnchor_Left(1);	// bit 0
				const u_int uAnchor_Right(2);	// bit 1
				const u_int uAnchor_Top(4);		// bit 2
				const u_int uAnchor_Bottom(8);	// bit 3

				u_int uTemp;

				xStream >> uTemp;
				xStream >> uTemp;
				
				if( m_uFlags & ASURA_GUIMENU_ROOT_FLAG_USE_ANCHOR_POSITIONING )
				{
					if(uTemp & (uAnchor_Left|uAnchor_Right) )
					{
						m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] = POS_CENTRE;
					}
					else if(uTemp & uAnchor_Right)
					{
						m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] = POS_FAR;
					}

					if (uTemp & (uAnchor_Top|uAnchor_Bottom))
					{
						m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] = POS_CENTRE;
					}
					else if (uTemp & uAnchor_Bottom)
					{
						m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] = POS_FAR;
					}
				}
			}

			if (uLocalChunkVersion >= 15)
			{
				u_int uTemp;
				xStream >> uTemp;
				m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] = static_cast<Asura_GUIMenu_Anchoring_Flag>(uTemp);
				xStream >> uTemp;
				m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] = static_cast<Asura_GUIMenu_Anchoring_Flag>(uTemp);
				xStream >> m_xPosOrigin;
			}
			else
			{
				if( GetFlag( ASURA_GUIMENU_ROOT_FLAG_USE_RELATIVE_POSITIONING))
				{
					m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] = POS_RELATIVE;
					m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] = POS_RELATIVE;
				}
			}
		}
	}

	if( pxParentObject != NULL)
	{
		pxParentObject->AddToThisContainer( this);	
	}

	if (uLocalChunkVersion < 15)
	{
		ChangeFlag( ASURA_GUIMENU_ROOT_FLAG_USE_RELATIVE_POSITIONING, false);

		//// -- if this object being loaded with it's relative positions, we need to setup a pixel position based
		//// -- on its relative position and use the menu dimensions as default.
		//// -- If there is no owner menu, then it's safe to assume that we should use the Viewport Dims.
		//if( m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] == POS_RELATIVE)
		//{
		//	Asura_GUIMenu_Menu_Page* pxPage = FindOwnerMenu();
		//	if( pxPage)
		//	{
		//		Asura_GUIMenu_Dims xDims = pxPage->GetDims();

		//		m_xLocalPos.x = m_xLocalRelativePos.GetX() * xDims.width;
		//	}
		//	else
		//	{
		//		m_xLocalPos.x = m_xLocalRelativePos.GetX() * Asura_Viewport::GetWidth();
		//	}

		//	m_xLocalPos.z = 0.0f;
		//}
		//if( m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] == POS_RELATIVE)
		//{
		//	Asura_GUIMenu_Menu_Page* pxPage = FindOwnerMenu();
		//	if( pxPage)
		//	{
		//		Asura_GUIMenu_Dims xDims = pxPage->GetDims();

		//		m_xLocalPos.y = m_xLocalRelativePos.GetY() * xDims.height;
		//	}
		//	else
		//	{
		//		m_xLocalPos.y = m_xLocalRelativePos.GetY() * Asura_Viewport::GetHeight();
		//	}

		//	m_xLocalPos.z = 0.0f;
		//}
	}

	if (uLocalChunkVersion >= 16)
	{
		xStream >> m_uScalingFlags;
		if (uLocalChunkVersion < 20)
		{
#if defined(ASURA_TOOLS) || defined(_DEBUG)
			// NY - 07 August 2008 - chunk versions 17->19 in theory have seperate pos scaling flags - however a code
			// issue meant these were ignored. the following code maintains observed behaviour with previous code, rather
			// than inforcing the intended behaviour
			if (uLocalChunkVersion > 16 || GetScalingFlag(static_cast<Asura_GUIMenu_Scaling_Flag>(1<<4)))
			{
				if (GetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX))
				{
					m_aePosMode[0] = POS_RELATIVE;
				}
				else if (GetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTY))
				{
					m_aePosMode[0] = POS_RELATIVE_ALTAXIS;
				}

				if (GetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY))
				{
					m_aePosMode[1] = POS_RELATIVE;
				}
				else if (GetScalingFlag(ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTX))
				{
					m_aePosMode[1] = POS_RELATIVE_ALTAXIS;
				}
			}

#else // #ifdef ASURA_TOOLS
			Asura_Assert(0, "guimenu chunk version no longer support outside of tools - please re-export");
#endif // #ifdef ASURA_TOOLS
			SetScalingFlag( static_cast<Asura_GUIMenu_Scaling_Flag>(1<<4), false);
			SetScalingFlag( static_cast<Asura_GUIMenu_Scaling_Flag>(1<<5), false);
			SetScalingFlag( static_cast<Asura_GUIMenu_Scaling_Flag>(1<<6), false);
			SetScalingFlag( static_cast<Asura_GUIMenu_Scaling_Flag>(1<<7), false);
		}
	}

	if( uLocalChunkVersion >= 19)
	{
		#ifdef ASURA_TOOLS
		xStream >> m_uPlatforms;
		#else
		xStream.Skip(4);
		#endif
	}
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Root::ReadVersionNumber( Asura_Chunk_Stream& xStream)
{
	u_int uVersion;

	xStream >> uVersion;

	return uVersion;
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Root::ReadLocalChunkHeader( Asura_Chunk_Stream& xStream, u_int& uType)
{
	u_int uChunkVersion;

	xStream >> uChunkVersion;
	xStream >> uType;

	if( CheckVersionNumber( uChunkVersion,static_cast<Asura_GUIMenu_Object_Type>(uType),xStream.GetFileHandle() ) == false)
	{
		// -- setting this means that the chunk version is BAD!
		uChunkVersion = 0xffffffff;
		uType = ASURA_GUIMENU_TYPEID_NONE;
	}

	return uChunkVersion;
}

// ------------------------------------------------------------------------------------------------------

bool Asura_GUIMenu_Root::CheckVersionNumber( const u_int uChunkVersion, const Asura_GUIMenu_Object_Type eType,const Asura_Handle iFileHandle /* = ASURA_INVALID_HANDLE */ )
{
	bool bRet = true;
	u_int uCurrentVersion( GetChunkVersionNumber(eType) );

	if( uChunkVersion > uCurrentVersion)
	{
		bRet = false;
	}

#ifdef ASURA_ASSERT
	if( !bRet )
	{
		// if chunk version number is unrecognised then assert with offending file name
		
#ifdef ASURA_TOOLS
		const char* szFileName( Asura_FileAccess::GetFilename( iFileHandle ) );
		const char* szTypeName( GetTypeName( eType ) );

		if( szFileName && szTypeName )
		{
			Asura_Assertf( false,("Unrecognised %s chunk version '%u' in file '%s'",szTypeName,uChunkVersion,szFileName) );
		}
		else
#endif // ASURA_TOOLS
		{
			Asura_Assertf( false,("Unrecognised GUI Menu ROOT chunk version '%u'",uChunkVersion) );
		}

	}	
#endif


	return bRet;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::ReadAllChildObjects( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParent)
{
	u_int uContainerCount, u;

	xStream >> uContainerCount;
#ifndef ASURA_TOOLS
	AllocateContainerArraySpace( static_cast<u_short>(uContainerCount));
#endif
	for( u=0; u<uContainerCount; u++)
	{
		ReadChildObject( xStream, pxParent);
	}

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::ReadChildObject( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParent)
{
	u_int uType = ASURA_GUIMENU_TYPEID_NONE;
	u_int uChunkVersion;
	Asura_String xMenuName;
	Asura_RGBA xColour;
	Asura_Vector_3 xPos;
	Asura_GUIMenu_Root* pxObject = NULL;

	uChunkVersion  = ReadLocalChunkHeader( xStream, uType);

	switch( uType)
	{
	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_MENUPAGE:
		pxObject = Asura_GUIMenu_CreateMenuPage();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE:
		pxObject = Asura_GUIMenu_CreateImageElement();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_TEXT:
		pxObject = Asura_GUIMenu_CreateTextElement();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_PATH:
		pxObject = Asura_GUIMenu_CreatePathElement();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_ANIMATION:
		pxObject = Asura_GUIMenu_CreateAnimationElement();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_STATUSTEXT:
		pxObject = Asura_GUIMenu_CreateStatusTextElement();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_ELEMENT_FMV:
		pxObject = Asura_GUIMenu_CreateFMV();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BUTTON:
		pxObject = Asura_GUIMenu_CreateButtonWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX:
		pxObject = Asura_GUIMenu_CreateCheckBoxWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_SLIDER:
		pxObject = Asura_GUIMenu_CreateSliderWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_TEXTBOX:
		pxObject = Asura_GUIMenu_CreateTextBoxWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX:
		pxObject = Asura_GUIMenu_CreateTextListBoxWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_GROUP:
		pxObject = Asura_GUIMenu_CreateGroupWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE:
		pxObject = Asura_GUIMenu_CreateAnimImageWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_PROGRESSBAR:
		pxObject = Asura_GUIMenu_CreateProgressBarWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC:
		pxObject = Asura_GUIMenu_CreateNumericWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_LISTBOX:
		pxObject = Asura_GUIMenu_CreateListBoxWidget();
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_LISTBOXENTRY:
		Asura_Assert( false, "LISTBOX ENTRIES SHOULD NOT BE PART OF A CONTAINER GROUP!" );
		pxObject = new Asura_GUIMenu_Widget_ListBoxEntry;
		break;

	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_CAROUSEL:
	{
		pxObject = Asura_GUIMenu_CreateListBoxWidget();
		Asura_GUIMenu_Widget_ListBox* pxList = static_cast<Asura_GUIMenu_Widget_ListBox*>(pxObject);
		pxList->SetCarousel(true);
		break;
	}

	default:
		Asura_Assert( 0, "Asura_Chunk_GUIMenu::ReadChunkData - Unknown GUIMenu object type");
		break;
	}

	if( pxObject != NULL)
	{
		pxObject->ReadFromChunkStream( xStream, pxParent, uChunkVersion);
	}
	return true;
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	u_int i;

	bool bIsExporting = Tools_Level_Export::IsGUIMenuExport() | Tools_Level_Export::IsLevelExport();

	// -- Write out object type. Can be menu, widget or element --
	// -- it's important that we output this first, otherwise we don't know what we're checking for with versions
	// -- NOTE: the type is now written out by the WriteLocalChunkHeader function instead.
	//	xStream << pxRoot->GetType();

	WriteVersionNumber( ASURA_CHUNK_ROOT_VERSION, xStream);

	// -- output menu name
	xStream << m_xName;

	// -- output colour --
	xStream << m_xColour.m_fR;
	xStream << m_xColour.m_fG;
	xStream << m_xColour.m_fB;
	xStream << m_xColour.m_fA;

	// -- output local position --

	// Bug testing - Leaving this in as it may need to be added to catch
	// menu pages involuntary resizing while using the GUI menu tools.
	//#ifdef ASURA_TOOLS
	//if ( GetType() == ASURA_GUIMENU_TYPEID_MENUPAGE )
	//{
	//	// check size of menu page for errors
	//	if ( m_xUnscaledLocalBounds.MaxX > 1280.0f ||
	//		 m_xUnscaledLocalBounds.MaxY > 720 )
	//	{
	//		Asura_Assert(false,"Menu Page is the wrong size");
	//	}
	//}
	//#endif

	xStream << m_xUnscaledLocalBounds.MinX;
	xStream << m_xUnscaledLocalBounds.MaxX;
	xStream << m_xUnscaledLocalBounds.MinY;
	xStream << m_xUnscaledLocalBounds.MaxY;

	// -- output id hash
	xStream << m_uHashID;

	// -- output flags, masking out any that shouldn't be saved --
	u_int uSavedFlags = m_uFlags;
	uSavedFlags &= ~ASURA_GUIMENU_ROOT_FLAGS_NOTSAVED;
	xStream << uSavedFlags;

	// output display conditions
	xStream << m_uInternalDisplayConditionCount;

	for( i=0; i<m_uInternalDisplayConditionCount; ++i)
	{
		xStream << m_axInternalDisplayConditions[i].m_iDisplayConditionThreshold;
		xStream << (u_int)m_axInternalDisplayConditions[i].m_eConditionCompare;
	}

	// output translucency mode
	xStream << m_eTranslucencyMode;

	//output some tools data
	xStream << m_uStyleChildEditableFlags;
	xStream << m_uToolsFlags; 

	//output some tools data
	xStream << m_uHashStyleObject;
	xStream << m_uStyleUseLocalDataFlags;

	m_xDisplayConditions.WriteToChunkStream( xStream );

	if( ASURA_CHUNK_ROOT_VERSION >= 6)
	{
		xStream << m_ucModifierCount;

		for( i=0; i<m_ucModifierCount; ++i )
		{
			m_ppxModifiers[i]->WriteToChunkStream( xStream);
		}
	}

	xStream << m_ucAnimatorCount;
	for( i=0; i<m_ucAnimatorCount; ++i )
	{
		m_ppxAnimators[i]->WriteToChunkStream( xStream);
	}

	const Asura_GUIMenu_ToolComment* pxToolComment;
	u_int uCommentCount = m_xCommentList.GetNumEntries();

	// if the system is exporting and NOT saving, then we force the comment count to zero
	// this ensures that no comment data is written out at export.
	if( bIsExporting)
	{
		uCommentCount = 0;
	}

	xStream << uCommentCount;

	for( i=0; i<uCommentCount; ++i)
	{
		pxToolComment = m_xCommentList.GetEntryByIndex( i);
		pxToolComment->WriteToChunkStream( xStream);
	}

	u_int uTemp = m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL];
	xStream << uTemp;
	uTemp = m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE];
	xStream << uTemp;
	xStream << m_xPosOrigin;
	xStream << GetScalingFlags(); // added in chunk v16
	xStream << m_uPlatforms;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::WriteVersionNumber( u_int uVersion, Asura_Chunk_Stream& xStream) const
{
	xStream << uVersion;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::WriteLocalChunkHeader( Asura_Chunk_Stream& xStream, u_int uVersion, u_int uType) const
{	
	xStream << uVersion;
	xStream << uType;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::WriteChildObjectData( Asura_Chunk_Stream& xStream) const
{
	Asura_GUIMenu_Root* pxItem;

	u_int uContainerCount = static_cast<u_int> (GetContainerCount());
	u_int uContainerCountToWrite = uContainerCount;
	if(Tools_Level_Export::IsExport())
	{
		//determine how many children are going to be exported for this platform
		for( u_int u=0; u<uContainerCount; u++)
		{
			if(!GetContainerItem(u)->ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
			{
				uContainerCountToWrite--;
			}
		}

	}
	xStream << uContainerCountToWrite;


	for( u_int u=0; u<uContainerCount; u++)
	{
		pxItem = GetContainerItem( u);
		if(Tools_Level_Export::IsExport())
		{
			if(!pxItem->ShouldExportToPlatform(Tools_Level_Export::s_eExportType))
			{
				continue;
			}
		}

		pxItem->WriteToChunkStream( xStream);
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::WriteXMLOpenTag(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer("<");
	pxExporter->AddToBuffer(Tools_GUIMenu_XMLExporter::MapObjectTypesToName( GetType() ));
	pxExporter->AddToBuffer (" name=\"");
	pxExporter->AddToBufferEscaped( GetNameAsChar() );
	pxExporter->AddToBuffer("\" >\n");
	pxExporter->IncTabs();
}

void Asura_GUIMenu_Root::WriteXMLCloseTag(Tools_GUIMenu_XMLExporter* pxExporter) const
{ 
	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer("</");
	pxExporter->AddToBuffer(Tools_GUIMenu_XMLExporter::MapObjectTypesToName( GetType() ));
	pxExporter->AddToBuffer (">\n");
}

void Asura_GUIMenu_Root::ExportToXMLStart(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	WriteXMLOpenTag( pxExporter );
	ExportToXML( pxExporter );
	WriteChildXMLData( pxExporter ); 
	WriteXMLCloseTag( pxExporter );
}

void Asura_GUIMenu_Root::ExportToXML( Tools_GUIMenu_XMLExporter* pxExporter ) const
{
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<Root_data>\n" );
	pxExporter->IncTabs();

	// -- output colour --
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<colour r=\"" );
	pxExporter->AddToBuffer( m_xColour.m_fR );
	pxExporter->AddToBuffer( "\" g=\"" );
	pxExporter->AddToBuffer( m_xColour.m_fG );
	pxExporter->AddToBuffer( "\" b=\"" );
	pxExporter->AddToBuffer( m_xColour.m_fB );
	pxExporter->AddToBuffer( "\" a=\"" );
	pxExporter->AddToBuffer( m_xColour.m_fA );
	pxExporter->AddToBuffer("\" />\n" );

    // -- output local position --

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<unscaled_local_bounds MinX=\"" );
	pxExporter->AddToBuffer( m_xUnscaledLocalBounds.MinX );
	pxExporter->AddToBuffer( "\" MaxX=\"" );
	pxExporter->AddToBuffer( m_xUnscaledLocalBounds.MaxX );
	pxExporter->AddToBuffer( "\" MinY=\"" );
	pxExporter->AddToBuffer( m_xUnscaledLocalBounds.MinY );
	pxExporter->AddToBuffer( "\" MaxY=\"" );
	pxExporter->AddToBuffer( m_xUnscaledLocalBounds.MaxY );
	pxExporter->AddToBuffer("\" />\n" );

	// -- output id hash
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<hash_id>" );
	pxExporter->AddToBufferHex( m_uHashID );
	pxExporter->AddToBuffer( "</hash_id>\n" );

	// -- output flags, if any --
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<flags>" );
	pxExporter->AddToBufferHex( m_uFlags );
	pxExporter->AddToBuffer( "</flags>\n" );

	// output display conditions
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<display_condition num='0' >\n" );
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<threshold>" );
	pxExporter->AddToBuffer( m_axInternalDisplayConditions[0].m_iDisplayConditionThreshold ); pxExporter->AddToBuffer( "</threshold>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<compare>" );
	pxExporter->AddToBuffer( (u_int)m_axInternalDisplayConditions[0].m_eConditionCompare ); pxExporter->AddToBuffer( "</compare>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</display_condition>\n" );

	for( u_int u(0); u<m_uInternalDisplayConditionCount; ++u)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<display_condition num='" ); pxExporter->AddToBuffer( u ); pxExporter->AddToBuffer( "' >\n" );
		pxExporter->IncTabs();

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<threshold>" );
		pxExporter->AddToBuffer( m_axInternalDisplayConditions[u].m_iDisplayConditionThreshold ); pxExporter->AddToBuffer( "</threshold>\n" );

		pxExporter->AddToBuffer( pxExporter->GetTabs() ); pxExporter->AddToBuffer( "<compare>" );
		pxExporter->AddToBuffer( (u_int)m_axInternalDisplayConditions[u].m_eConditionCompare ); pxExporter->AddToBuffer( "</compare>\n" );
		
		pxExporter->DecTabs();
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "</display_condition>\n" );
	}

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<translucency>" );
	pxExporter->AddToBuffer( m_eTranslucencyMode );
	pxExporter->AddToBuffer( "</translucency>\n" );

	//output some tools data
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<child_editable_flags>" );
	pxExporter->AddToBufferHex( m_uStyleChildEditableFlags );
	pxExporter->AddToBuffer( "</child_editable_flags>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<tools_flags>" );
	pxExporter->AddToBufferHex( m_uToolsFlags );
	pxExporter->AddToBuffer( "</tools_flags>\n" );

	//output some tools data
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<hash_style_object>" );
	pxExporter->AddToBufferHex( m_uHashStyleObject );
	pxExporter->AddToBuffer( "</hash_style_object>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<style_use_local_data_flags>" );
	pxExporter->AddToBufferHex( m_uStyleUseLocalDataFlags );
	pxExporter->AddToBuffer( "</style_use_local_data_flags>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<platforms>" );
	pxExporter->AddToBuffer( m_uPlatforms );
	pxExporter->AddToBuffer( "</platforms>\n" );


	m_xDisplayConditions.ExportToXML( pxExporter );

	for( u_int u(0); u<m_ucModifierCount; ++u )
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<modifier num='" ); pxExporter->AddToBuffer( u ); pxExporter->AddToBuffer( "' >\n" );
		pxExporter->IncTabs();

		m_ppxModifiers[u]->ExportToXML( pxExporter );
		
		pxExporter->DecTabs();
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "</modifier>\n" );
		
	}

	for( u_int u(0); u<m_ucAnimatorCount; ++u )
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<animator num='" ); pxExporter->AddToBuffer( u ); pxExporter->AddToBuffer( "' >\n" );
		pxExporter->IncTabs();

		m_ppxAnimators[u]->ExportToXML( pxExporter );

		pxExporter->DecTabs();
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "</animator>\n" );
	}

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<user_comment>" );
	pxExporter->IncTabs();
	pxExporter->AddToBuffer( "<name>");
	pxExporter->AddToBuffer( m_xCommentUser.GetString());
	pxExporter->AddToBuffer( "</name>");
	pxExporter->AddToBuffer( "<text>");
	pxExporter->AddToBuffer( m_xComments.GetString());
	pxExporter->AddToBuffer( "</text>");
	pxExporter->DecTabs();
	pxExporter->AddToBuffer( "</user_comment>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<positioning_data>\n" );
	pxExporter->IncTabs();

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<width_mode>" );
	pxExporter->AddToBuffer( m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_HORIZONTAL] );
	pxExporter->AddToBuffer( "</width_mode>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<height_mode>" );
	pxExporter->AddToBuffer( m_aePosMode[ASURA_GUIMENU_POSITION_MODE_AXIS_VERTICLE] );
	pxExporter->AddToBuffer( "</height_mode>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<pos_origin_x>" );
	pxExporter->AddToBuffer( m_xPosOrigin.x );
	pxExporter->AddToBuffer( "</pos_origin_x>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<pos_origin_y>" );
	pxExporter->AddToBuffer( m_xPosOrigin.y );
	pxExporter->AddToBuffer( "</pos_origin_y>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</positioning_data>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</Root_data>\n" );
}

void Asura_GUIMenu_Root::WriteChildXMLData(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	if(GetContainerCount() > 0)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<contained_items>\n" );
		pxExporter->IncTabs();

		Asura_GUIMenu_Root* pxChild;
		for( u_int i=0; i<GetContainerCount(); i++)
		{
			pxChild = GetContainerItem( i);
			pxChild->ExportToXMLStart( pxExporter );
		}

		pxExporter->DecTabs();
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "</contained_items>\n" );
	}
}

#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------ 

u_int	Asura_GUIMenu_Root::GetChunkVersionNumber( const Asura_GUIMenu_Object_Type eType ) const
{
	u_int uCurrentVersion( 0xffffffff );

	switch( eType )
	{
	case ASURA_GUIMENU_TYPEID_MENUPAGE:
		uCurrentVersion = ASURA_CHUNK_GUIMENU_MENUPAGE_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_IMAGE_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_ELEMENT_TEXT:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_TEXT_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_ELEMENT_PATH:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_PATH_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_ELEMENT_ANIMATION:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_ANIMATION_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_ELEMENT_STATUSTEXT:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_STATUSTEXT_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_ELEMENT_FMV:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_FMV_VERSION;
		break;
	
	case Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BUTTON:
		uCurrentVersion = ASURA_CHUNK_WIDGET_BUTTON_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX:
		uCurrentVersion = ASURA_CHUNK_WIDGET_CHECKBOX_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_SLIDER:
		uCurrentVersion = ASURA_CHUNK_WIDGET_SLIDER_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_TEXTBOX:
		uCurrentVersion = ASURA_CHUNK_WIDGET_TEXTBOX_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX:
		uCurrentVersion = ASURA_CHUNK_WIDGET_TEXTLISTBOX_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_GROUP:
		uCurrentVersion = ASURA_CHUNK_WIDGET_GROUP_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE:
		uCurrentVersion = ASURA_CHUNK_WIDGET_ANIMIMAGE_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_PROGRESSBAR:
		uCurrentVersion = ASURA_CHUNK_WIDGET_PROGRESSBAR_VERSION;
		break;

	case ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC:
		uCurrentVersion = ASURA_CHUNK_WIDGET_NUMERIC_VERSION;
		break;

	case ASURA_GUIMENU_CHUNK_TYPE_ROOT:
		uCurrentVersion = ASURA_CHUNK_ROOT_VERSION;
		break;

	case ASURA_GUIMENU_CHUNK_TYPE_ELEMENT_BASE:
		uCurrentVersion = ASURA_CHUNK_ELEMENT_BASE_VERSION;
		break;

	case ASURA_GUIMENU_CHUNK_TYPE_WIDGET_BASE:
		uCurrentVersion = ASURA_CHUNK_WIDGET_BASE_VERSION;
		break;

	default:
		break;
	}

	return uCurrentVersion;
}

const char* Asura_GUIMenu_Root::GetTypeName( const Asura_GUIMenu_Object_Type eType ) const
{
	const char* pcRtn(NULL);

	switch( eType)
	{
		case ASURA_GUIMENU_TYPEID_MENUPAGE				: pcRtn = "ASURA_GUIMENU_TYPEID_MENUPAGE";				break;
		case ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE			: pcRtn = "ASURA_GUIMENU_TYPEID_ELEMENT_IMAGE";		break;
		case ASURA_GUIMENU_TYPEID_ELEMENT_TEXT			: pcRtn = "ASURA_GUIMENU_TYPEID_ELEMENT_TEXT";			break;
		case ASURA_GUIMENU_TYPEID_ELEMENT_PATH			: pcRtn = "ASURA_GUIMENU_TYPEID_ELEMENT_PATH";			break;
		case ASURA_GUIMENU_TYPEID_ELEMENT_ANIMATION		: pcRtn = "ASURA_GUIMENU_TYPEID_ELEMENT_ANIMATION";	break;
		case ASURA_GUIMENU_TYPEID_ELEMENT_STATUSTEXT	: pcRtn = "ASURA_GUIMENU_TYPEID_ELEMENT_STATUSTEXT";	break;
		case ASURA_GUIMENU_TYPEID_ELEMENT_FMV			: pcRtn = "ASURA_GUIMENU_TYPEID_ELEMENT_FMV";	break;
		case ASURA_GUIMENU_TYPEID_WIDGET_BUTTON			: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_BUTTON";		break;
		case ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX		: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_CHECKBOX";		break;
		case ASURA_GUIMENU_TYPEID_WIDGET_SLIDER			: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_SLIDER";		break;
		case ASURA_GUIMENU_TYPEID_WIDGET_TEXTBOX		: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_TEXTBOX";		break;
		case ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX	: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_TEXTLISTBOX";	break;
		case ASURA_GUIMENU_TYPEID_WIDGET_GROUP			: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_GROUP";			break;	
		case ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE	: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_ANIMATINGIMAGE";break;
		case ASURA_GUIMENU_TYPEID_WIDGET_PROGRESSBAR	: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_PROGRESSBAR";	break;
		case ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC		: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC";		break;
		case ASURA_GUIMENU_TYPEID_WIDGET_LISTBOX		: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_LISTBOX";		break;
		case ASURA_GUIMENU_TYPEID_WIDGET_LISTBOXENTRY	: pcRtn = "ASURA_GUIMENU_TYPEID_WIDGET_LISTBOXENTRY";	break;
		case ASURA_GUIMENU_CHUNK_TYPE_ROOT				: pcRtn = "ASURA_GUIMENU_CHUNK_TYPE_ROOT";				break;
		case ASURA_GUIMENU_CHUNK_TYPE_ELEMENT_BASE		: pcRtn = "ASURA_GUIMENU_CHUNK_TYPE_ELEMENT_BASE";		break;
		case ASURA_GUIMENU_CHUNK_TYPE_WIDGET_BASE		: pcRtn = "ASURA_GUIMENU_CHUNK_TYPE_WIDGET_BASE";		break;
															
	default:												
		pcRtn = "unknown";									
		break;
	}
	
	return pcRtn;
}

void Asura_GUIMenu_Root::AllocateContainerArraySpace( u_short usMaxContainerCount)
{
	u_int u;

	Asura_Assert( m_ppxContainerParts == NULL, "Asura_GUIMenu_Root::AllocateContainerArraySpace - container array is already allocated.  Ignoring new request.");

	if( (usMaxContainerCount > 0) && (m_ppxContainerParts == NULL))
	{
		m_usMaxContainerPartsCount = usMaxContainerCount;

		m_ppxContainerParts = new Asura_GUIMenu_Root*[m_usMaxContainerPartsCount];

		m_usContainerPartsCount = 0;

		for( u=0; u<m_usMaxContainerPartsCount; u++)
		{
			m_ppxContainerParts[u] = NULL;
		}
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::ExpandContainerArraySpace()
{
	if( m_usContainerPartsCount == m_usMaxContainerPartsCount)
	{
		u_short usNewSize = m_usMaxContainerPartsCount+ASURA_GUIMENU_CONTAINER_ADDITION_GRADIENT;
		// Time to grow the array
		Asura_GUIMenu_Root** ppxContainerParts = new Asura_GUIMenu_Root*[usNewSize];
		Asura_Assert(ppxContainerParts != NULL, "Out of memory!");

		if( ppxContainerParts != NULL)
		{
			for( u_int u=0; u<usNewSize; u++)
			{
				ppxContainerParts[u] = NULL;
			}

			// Copy the existing mob attractor data
			if( m_ppxContainerParts != NULL)
			{
				Asura_Mem::Copy( ppxContainerParts, m_ppxContainerParts, m_usMaxContainerPartsCount * sizeof( Asura_GUIMenu_Root*));
			
				// Get rid of the old array...
				delete []m_ppxContainerParts;
			}

			m_usMaxContainerPartsCount = usNewSize;

			// ...and use the new one
			m_ppxContainerParts = ppxContainerParts;
		}
	}
}

// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------ 

void Asura_GUIMenu_Root::AllocateModifierArraySpace( u_char ucMaxModifierCount)
{
	Asura_Assert( m_ppxModifiers == NULL, "Asura_GUIMenu_Root::AllocateModifierArraySpace - Modifier array is already allocated.  Ignoring new request.");

	if( (ucMaxModifierCount > 0) && (m_ppxModifiers == NULL))
	{
		m_ucMaxModifierCount = ucMaxModifierCount;

		m_ppxModifiers = new Asura_GUIMenu_Modifier*[m_ucMaxModifierCount];

		m_ucModifierCount = 0;

		for( u_int u=0; u<m_ucMaxModifierCount; u++)
		{
			m_ppxModifiers[u] = NULL;
		}
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::AllocateAnimatorArraySpace( u_char ucMaxAnimatorCount)
{
	u_int u;

	Asura_Assert( m_ppxAnimators == NULL, "Asura_GUIMenu_Root::AllocateAnimatorArraySpace - Animator array is already allocated.  Ignoring new request.");

	if( (ucMaxAnimatorCount > 0) && (m_ppxAnimators == NULL))
	{
		m_ucMaxAnimatorCount = ucMaxAnimatorCount;

		m_ppxAnimators = new Asura_GUIMenu_Animator*[m_ucMaxAnimatorCount];

		m_ucAnimatorCount = 0;

		for( u=0; u<m_ucMaxAnimatorCount; u++)
		{
			m_ppxAnimators[u] = NULL;
		}
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::IsActionEnabled( u_int uFlags)
{
	return GetFlag( uFlags);
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Root::GetType() const
{
	return m_uType;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::IsWidget() const
{
	u_int uType = GetType();
	return (uType >= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_WIDGET_BEGIN && 
		uType <= Asura_GUIMenu_Root::ASURA_GUIMENU_TYPEID_PROJECT_SPECIFIC_WIDGET_END);
}

// --------------------------------------------------------------------------------------- 

ASURA_TRANSLUCENCY_MODE Asura_GUIMenu_Root::GetTranslucencyMode() const
{
	return m_eTranslucencyMode;	
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetTranslucencyMode( ASURA_TRANSLUCENCY_MODE eTransMode)
{
	m_eTranslucencyMode = eTransMode;	
}

// --------------------------------------------------------------------------------------- 
// -- modifier functions -----------------------------------------------------------------

// SK - this will possibly need expanding to account for style inheritance at some point.
u_char Asura_GUIMenu_Root::GetModifierCount() const
{
	return m_ucModifierCount;
}

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::PostMenuPreviewCleanup()
{
	//Reset hidden by console variable state. RAR 08/03/2007
	SetHiddenByConditionVar(false);
	SetRenderingDisabledFlag(false);
}
#endif //ASURA_TOOLS

// --------------------------------------------------------------------------------------- 

Asura_GUIMenu_Modifier* Asura_GUIMenu_Root::GetModifier( u_int uIndex) const
{
	Asura_GUIMenu_Modifier* pxMod = NULL;

	if( uIndex < m_ucMaxModifierCount)
	{
		pxMod = m_ppxModifiers[uIndex];
	}

	return pxMod;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, u_int uData)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( uData);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, int iData)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( iData);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, float fData)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( fData);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_RGBA& xColourData)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( xColourData);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_Vector_2& xVector2)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( xVector2);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_UnicodeChar* pwszUnicodeString)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( pwszUnicodeString);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_Char* pcCharString)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( pcCharString);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::AddModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType, Asura_GUIMenu_UV_Set& xUVSet)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( uStateToModify, eModifyType);
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->SetData( xUVSet);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::CopyModifier( Asura_GUIMenu_Modifier* pxModifier)
{
	bool bSuccess = false;

	int iIndex = CreateModifier( pxModifier->GetStateBeingModified(), pxModifier->GetModifyType());
	if( iIndex != -1)
	{
		m_ppxModifiers[iIndex]->Copy( pxModifier);
		bSuccess = true;
	}

	return bSuccess;
}

// --------------------------------------------------------------------------------------- 

int Asura_GUIMenu_Root::CreateModifier( u_int uStateToModify, ASURA_GUIMENU_MODIFY_TYPE eModifyType)
{
	int iRetIndex = -1;

	if( (m_ucModifierCount < m_ucMaxModifierCount) && (m_ppxModifiers != NULL))
	{
		m_ppxModifiers[m_ucModifierCount] = new Asura_GUIMenu_Modifier;

		m_ppxModifiers[m_ucModifierCount]->SetStateToModify( uStateToModify);
		m_ppxModifiers[m_ucModifierCount]->SetModifyType( eModifyType );

		iRetIndex = m_ucModifierCount;
		m_ucModifierCount++;
	}

	return iRetIndex;
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
bool Asura_GUIMenu_Root::DeleteModifier( u_int uListIndex)
{
	bool bSuccess = false;

	if( uListIndex < m_ucModifierCount)
	{
		if( m_ppxModifiers[uListIndex] != NULL)
		{
			delete m_ppxModifiers[uListIndex];
			m_ppxModifiers[uListIndex] = m_ppxModifiers[m_ucModifierCount-1];
			m_ppxModifiers[m_ucModifierCount-1] = NULL;
			m_ucModifierCount--;

			bSuccess = true;
		}
	}

	return bSuccess;
}
#endif

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::CopyModifiers( const Asura_GUIMenu_Root& xSource )
{
	u_char ucModifierCount = xSource.GetModifierCount();
#ifndef ASURA_TOOLS
	AllocateModifierArraySpace(ucModifierCount);
#endif
	Asura_Assert( !m_ucModifierCount, "already got Modifiers - need to destroy them");
	for( u_char  u = 0; u < ucModifierCount; u++)
	{
		Asura_Assert( !m_ppxModifiers[u], "already got Modifier - memory leak");
		m_ppxModifiers[u] = new Asura_GUIMenu_Modifier;
		*m_ppxModifiers[u] = *xSource.GetModifier( u );
	}

	m_ucModifierCount = ucModifierCount;
}

// --------------------------------------------------------------------------------------- 

// functions for animators
Asura_GUIMenu_Animator* Asura_GUIMenu_Root::GetAnimator( u_int uIndex ) const
{
	Asura_Assert( uIndex < m_ucAnimatorCount, "invalid index");
	return m_ppxAnimators[uIndex];
}

const Asura_GUIMenu_Animator* Asura_GUIMenu_Root::GetAnimatorBySequenceHash(Asura_Hash_ID uHash) const
{
	for(u_int i=0; i<m_ucAnimatorCount; i++)
	{
		if(m_ppxAnimators[i]->GetSequenceHash() == uHash)
		{
			return m_ppxAnimators[i];
		}
	}
	return 0;
}

Asura_GUIMenu_Animator* Asura_GUIMenu_Root::GetAnimatorBySequenceHash(Asura_Hash_ID uHash)
{
	return const_cast<Asura_GUIMenu_Animator*>(const_cast<const Asura_GUIMenu_Root*>(this)->GetAnimatorBySequenceHash(uHash));
}


// GetLongestAnimatorDuration() is to help decide transition lengths. Ignoring looping animators, 
// find the longest duration of animators in this object matching the given state(s), and recurse 
// into the contained objects.
float Asura_GUIMenu_Root::GetLongestAnimatorDuration( u_int uStateFlags ) const
{
	float fLongestDuration = 0.0f;
	for (u_int u = 0; u < m_ucAnimatorCount; ++u)
	{
		//Only checking animators that are exlusively for the given state to avoid fade in/out times being affected
		//by sequences that are used in all states.
		//RAR 04/04/2008
		if ( (!m_ppxAnimators[u]->IsLooping()) && (m_ppxAnimators[u]->GetStateFlags() == uStateFlags) )
		{
			fLongestDuration = Asura_Maths::Max( fLongestDuration, m_ppxAnimators[u]->GetDuration()+m_ppxAnimators[u]->GetDelay() );
		}
	}

	for (u_int u = 0; u < m_usContainerPartsCount; ++u)
	{
		fLongestDuration = Asura_Maths::Max( fLongestDuration, m_ppxContainerParts[u]->GetLongestAnimatorDuration( uStateFlags ) );
	}

	return fLongestDuration;
}

void Asura_GUIMenu_Root::RecurseInitialiseAnimators(bool bInitMenuPage /* = false*/)
{
	for (u_int u = 0; u < m_ucAnimatorCount; ++u)
	{
		m_ppxAnimators[u]->Initialise(bInitMenuPage);
	}

	for (u_int u = 0; u < m_usContainerPartsCount; ++u)
	{
		m_ppxContainerParts[u]->RecurseInitialiseAnimators(bInitMenuPage);
	}
}

#ifdef ASURA_TOOLS
Asura_GUIMenu_Animator* Asura_GUIMenu_Root::AddAnimator( const char* szSequenceName )
{
	if( (m_ucAnimatorCount < m_ucMaxAnimatorCount) && (m_ppxAnimators != NULL))
	{
		Asura_GUIMenu_Animator* pxNewAnimator = new Asura_GUIMenu_Animator();

		pxNewAnimator->SetSequence(szSequenceName);

		m_ppxAnimators[m_ucAnimatorCount++] = pxNewAnimator;

		return pxNewAnimator;
	}
	return NULL;
}

void Asura_GUIMenu_Root::DeleteAnimator( u_int uIndex )
{
	Asura_Assert( uIndex < m_ucAnimatorCount, "invalid index");
	delete m_ppxAnimators[uIndex];
	m_ppxAnimators[uIndex] = m_ppxAnimators[m_ucAnimatorCount-1];
	m_ppxAnimators[m_ucAnimatorCount-1] = NULL;
	m_ucAnimatorCount--;
}
#endif

Asura_GUIMenu_Animator* Asura_GUIMenu_Root::AddAnimator( const Asura_GUIMenu_Animator& xAnimator )
{
	if( (m_ucAnimatorCount < m_ucMaxAnimatorCount) && (m_ppxAnimators != NULL))
	{
		Asura_GUIMenu_Animator* pxNewAnimator = new Asura_GUIMenu_Animator( xAnimator );

		m_ppxAnimators[m_ucAnimatorCount++] = pxNewAnimator;

		return pxNewAnimator;
	}
	return NULL;
}

void Asura_GUIMenu_Root::CopyAnimators( const Asura_GUIMenu_Root& xSource )
{
	const u_int uAnimatorCount( xSource.GetAnimatorCount());

#ifndef ASURA_TOOLS
	AllocateAnimatorArraySpace(static_cast<u_char>(uAnimatorCount));
#endif
	Asura_Assert( !m_ucAnimatorCount, "already got animators - need to destroy them");
	
	for( u_int u(0); u<uAnimatorCount; u++)
	{
		Asura_Assert( !m_ppxAnimators[u], "already got animator - memory leak");

		AddAnimator( *xSource.GetAnimator( u ) ); 
	}

	Asura_Assert( GetAnimatorCount() == uAnimatorCount,"Source animator count does not match targets count" );
}

// ------------------------------------------------------------------------------------------------------

void Asura_GUIMenu_Root::FixupBoundsFromChildren()
{
	// NY - 25 September 2008 - this is really a tools function, but is used at runtime to fix up old chunk data. No new calls to this function should
	// be added to non tools code.
	Asura_Bounding_Rectangle xBounds(GetUnscaledLocalBounds());
	Asura_Vector_2 xLocalPos = xBounds.GetTopLeft();

	for(u_int u = 0; u < m_usContainerPartsCount; ++u)
	{
		Asura_GUIMenu_Root* pxRootItem = m_ppxContainerParts[u];
		pxRootItem->FixupBoundsFromChildren();
		// NY - 25 September 2008 - only check the unclamped flag if we're in tools. Aside from the fact that the flag is on available in tools,
		// this function is only used outside of tools for fixing up old chunk data which precedes the flag, and so can't have it set; therefore we can assume the flag is
		// unset outside of tools. once the chunk code that calls this function is changed to tools only, this whole function can become tools only for clarity
#ifdef ASURA_TOOLS
		// We only want to adjust the bounds if the object is clamped to the bounds of it's parent
		if (!(pxRootItem->GetToolsFlags() & ASURA_GUIMENU_TOOLS_FLAG_ISUNCLAMPED))
#endif
		{
			Asura_Bounding_Rectangle xChildBounds = pxRootItem->GetUnscaledLocalBounds();
			xChildBounds.Displace(GetChildPosOffset() + xLocalPos);
			xBounds.Set(xChildBounds.GetTopLeft());
			xBounds.Set(xChildBounds.GetBottomRight());
		}
	}

	SetUnscaledLocalBounds(xBounds);

	if (xLocalPos != xBounds.GetTopLeft())
	{
		for(u_int u = 0; u < m_usContainerPartsCount; ++u)
		{
			Asura_GUIMenu_Root* pxRootItem = m_ppxContainerParts[u];
			Asura_Vector_2 xChildLocalPos = pxRootItem->GetUnscaledLocalPos();
			xChildLocalPos -= xBounds.GetTopLeft() - xLocalPos;
			pxRootItem->SetUnscaledLocalPos(xChildLocalPos);
		}
	}
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Root::GetScalingFlag( const Asura_GUIMenu_Scaling_Flag eFlag) const
{
	return ((m_uScalingFlags & eFlag) != 0);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::SetScalingFlag( const Asura_GUIMenu_Scaling_Flag eFlagsToSet, const bool bFlagSet)
{
	if( bFlagSet)
	{
		m_uScalingFlags |= eFlagsToSet;
	}
	else
	{
		m_uScalingFlags &=~ eFlagsToSet;
	}
}

// --------------------------------------------------------------------------------------- 

Asura_Vector_2 Asura_GUIMenu_Root::GetRelativeDimensionScale( const Asura_Vector_2& xParentScale ) const
{
	Asura_Vector_2 xScale(1.0f, 1.0f);

	// width scale
	if( GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTX) )
	{
		xScale.x = xParentScale.x;
	}
	else if( GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_X_TO_PARENTY) )
	{
		xScale.x = xParentScale.y;
	}

	// height scale
	if( GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTY) )
	{
		xScale.y = xParentScale.y;
	}
	else if( GetScalingFlag( ASURA_GUIMENU_ELEMENT_FLAG_SCALE_DIMENSION_Y_TO_PARENTX) )
	{
		xScale.y = xParentScale.x;
	}

	return xScale;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Root::ApplyScalingFlagsToChildren()
{
	// update child flags as appropriate
	const u_int uChildMax( GetResizeChildCount() );

	for (u_int u = 0; u < uChildMax; ++u)
	{	
		Asura_GUIMenu_Root* pxChild = GetResizeChild(u);

		Asura_Assertf( pxChild,("GUI Menu: '%s' contains NULL child object",GenerateFullyQualifiedName().GetString()) );

		if( pxChild )
		{
			pxChild->SetScalingFlags( GetScalingFlags() );
			pxChild->ApplyScalingFlagsToChildren();
		}
	}
}

void Asura_GUIMenu_Root::ApplyAnchoringFlagsToChildren()
{
	// update child anchoring as appropriate
	const u_int uChildMax( GetResizeChildCount() );

	for (u_int u = 0; u < uChildMax; ++u)
	{	
		Asura_GUIMenu_Root* pxChild = GetResizeChild(u);

		Asura_Assertf( pxChild,("GUI Menu: '%s' contains NULL child object",GenerateFullyQualifiedName().GetString()) );

		if( pxChild )
		{
			pxChild->SetPosMode( 0,GetPosMode(0) );
			pxChild->SetPosMode( 1,GetPosMode(1) );
			pxChild->ApplyAnchoringFlagsToChildren();
		}
	}
}


// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Root::AddComment( const Asura_Char* pszName, const Asura_Char* pszComment)
{
	Asura_GUIMenu_ToolComment* pxNewComment = new Asura_GUIMenu_ToolComment;
	if( pxNewComment && pszName && pszComment)
	{
		pxNewComment->Set( pszName, pszComment);
		m_xCommentList.AddToList( pxNewComment);
	}
}

// ------------------------------------------------------------------------------------------------------

const bool Asura_GUIMenu_Root::FindAndReplaceConsoleVarName( const Asura_UnicodeChar* wszName,const Asura_UnicodeChar* wszReplacementName )
{
	bool bMatchFound(false);

	if( wszName && wszReplacementName )
	{
		// Display conditions
		m_xDisplayConditions.FindAndReplaceConsoleVarName( wszName,wszReplacementName );
		
		Asura_GUIMenu_ConditionVarManager*	pxManager(NULL);

		// modifiers
		Asura_GUIMenu_Modifier*				pxModifier(NULL);

		for( u_int uModifier(0);uModifier<GetModifierCount();++uModifier )
		{
			pxModifier = GetModifier(uModifier);
			
			if( pxModifier )
			{
				pxManager = pxModifier->GetActivationConditions();
			}	

			if( pxModifier && pxManager )
			{
				pxManager->FindAndReplaceConsoleVarName( wszName,wszReplacementName );
			}
		}

		// animators
		Asura_GUIMenu_Animator*				pxAnimator(NULL);

		for( u_int uAnimator(0);uAnimator<GetAnimatorCount();++uAnimator )
		{
			pxAnimator = GetAnimator(uAnimator);
			
			if( pxAnimator )
			{
				pxManager = pxAnimator->GetStartConditions();
			}	

			if( pxAnimator && pxManager )
			{
				pxManager->FindAndReplaceConsoleVarName( wszName,wszReplacementName );
			}
		}


		// recursive call to this method on all children
		if( m_ppxContainerParts)
		{
			for( u_int u(0);u<m_usContainerPartsCount;++u )
			{
				m_ppxContainerParts[u]->FindAndReplaceConsoleVarName( wszName,wszReplacementName );
			}
		}
	}

	return bMatchFound;
}

#endif // ASURA_TOOLS

// ------------------------------------------------------------------------------------------------------
// 
// Asura_GUIMenu_ToolComment
// Created by SK 22.11.07
//
// ------------------------------------------------------------------------------------------------------

#ifdef ASURA_TOOLS

static const u_int uASURA_GUIMENU_TOOL_COMMENT_VERSION = 0;

void Asura_GUIMenu_ToolComment::Set( const Asura_Char* pszName, const Asura_Char* pszDate, const Asura_Char* pszComment)
{
	Asura_Assert( pszName, "Passed name string pointer is NULL.");
	Asura_Assert( pszDate, "Passed name string pointer is NULL.");
	Asura_Assert( pszComment, "Passed name string pointer is NULL.");

	SetNameString( pszName);
	SetDateString( pszDate);
	SetCommentString( pszComment);
}

// ------------------------------------------------------------------------------------------------------

void Asura_GUIMenu_ToolComment::Set( const Asura_Char* pszName, const Asura_Char* pszComment)
{
	Asura_Char acBuffer[256];
	Asura_Assert( pszName, "Passed name string pointer is NULL.");
	Asura_Assert( pszComment, "Passed name string pointer is NULL.");


	Asura_Time xTime = Asura_Time_System::GetCurrentDateAndTime();
	Asura_PrintfToBfr( &acBuffer[0], "%02d/%02d/%04d - %02d:%02d:%02d", 256, true,
									 xTime.m_ucDayOfMonth, xTime.m_ucMonth, xTime.m_usYear,
									 xTime.m_ucHour, xTime.m_ucMinute, xTime.m_ucSecond);


	Set( pszName, acBuffer, pszComment);
}

// ------------------------------------------------------------------------------------------------------

void Asura_GUIMenu_ToolComment::AddOverlayControls( Asura_Subwindow_Dialog* pxSubWindow, Asura_Vector_2& xOffset)
{
	Asura_Char	acBuffer[256];
	Asura_Dialog* pxDialog;

	Asura_PrintfToBfr( acBuffer, "User: %s - %s", 256, 1, m_xName.GetString(), m_xDate.GetString());
	pxDialog = new Asura_StringDisplay_Dialog( acBuffer, xOffset);
	pxDialog->SetHelpText( m_xComment.GetString());
	pxSubWindow->AddDialog( pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_StringDisplay_Dialog( m_xComment.GetString(), xOffset);
	pxDialog->SetHelpText( m_xComment.GetString());
	pxSubWindow->AddDialog( pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

}

// ------------------------------------------------------------------------------------------------------

void Asura_GUIMenu_ToolComment::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << uASURA_GUIMENU_TOOL_COMMENT_VERSION;

	xStream << m_xName;
	xStream << m_xDate;
	xStream << m_xComment;
}

#endif // ASURA_TOOLS

// ------------------------------------------------------------------------------------------------------

void Asura_GUIMenu_ToolComment::ReadFromChunkStream( Asura_Chunk_Stream& xStream)
{
	Asura_Char acBuffer[1024];

	u_int uVersion;
	xStream >> uVersion;

	xStream.ReadString( acBuffer, sizeof( acBuffer));
#ifdef ASURA_TOOLS
	m_xName = acBuffer;
#endif
	xStream.ReadString( acBuffer, sizeof( acBuffer));
#ifdef ASURA_TOOLS
	m_xDate = acBuffer;
#endif
	xStream.ReadString( acBuffer, sizeof( acBuffer));
#ifdef ASURA_TOOLS	
	m_xComment = acBuffer;
#endif
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
bool Asura_GUIMenu_Root::IsFlaggedForCurrentPlatform() const
{
	return ShouldExportToPlatform(Tools_Mode_GUIMenu::GetPreviewPlatform());
}

bool Asura_GUIMenu_Root::AreThisAndParentsFlaggedForPlatform(int iPlatform) const
{
	//check this
	if(!ShouldExportToPlatform(iPlatform)) return false;
	//recurse up the parents, since if any aren't for the current platform , then this won't appear either
	return m_pxParent ? m_pxParent->AreThisAndParentsFlaggedForPlatform(iPlatform) : true;

}
#endif //ASURA_TOOLS



#endif // ASURA_USE_GUI_MENU_SYSTEM
