// ------------------------------------------------------------------------------------ //
//																						//
//	File: Asura_GUIMenu_Widget_Numeric.cpp												//
//																						//
//	Description: realtime numeric feedback class										//
//																						//
//	Created: Steve Keen - 21/02/07														//
//																						//
//	Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.							//
// ------------------------------------------------------------------------------------ //

#include "Core/Asura.h"
#include "Asura_GUIMenu_Widget_Numeric.h"

#ifdef ASURA_USE_GUI_MENU_SYSTEM
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"
#include "GUIMenu/Asura_Profiling_GUIMenu.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Overlay_Quad.h"

#include "Asura_GUIMenu_Element_Base.h"
#include "Asura_GUIMenu_Message.h"
#include "Asura_GUIMenu_MessagePump.h"

#ifdef ASURA_TOOLS
#include "GUIMenu/Tools_GUIMenu_XMLExporter.h"
#endif

// --------------------------------------------------------------------------------------- 
// -- Constructors -----------------------------------------------------------------------

Asura_GUIMenu_Widget_Numeric::Asura_GUIMenu_Widget_Numeric(void)
: m_fElapsedTime( 0.0f)
, m_uButtonRepeatPhase( 0)
, m_fCurrentRepeatDelay( 0.0f)
, m_fInitialRepeatDelay( fASURA_GUIMENU_WIDGET_NUMERIC_CONTROL_INITIAL_REPEAT_DELAY)
, m_fSecondaryRepeatDelay( fASURA_GUIMENU_WIDGET_NUMERIC_CONTROL_SECONDARY_REPEAT_DELAY)
, m_fCurrentValue( 0.0f)
, m_fMinValue( 0.0f)
, m_fMaxValue( 1.0f)
, m_fStepValue( 0.0f)
, m_fDecFlashTimerCountdown( 0.0f)
, m_fIncFlashTimerCountdown( 0.0f)
, m_fFlashTimerStartVal( fASURA_GUIMENU_WIDGET_NUMERIC_FLASH_TIME_DEFAULT)
, m_pxDecImage( NULL)
, m_pxIncImage( NULL)
, m_pxFeedbackText( NULL)
, m_bDisplayAsInteger( false)
, m_uFloatingPointPrecision( ASURA_GUIMENU_WIDGET_NUMERIC_DEFAULT_FLOATING_POINT_PRECISION)
, m_uNumericFlags( ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_USE_CONSOLEVAR_MIN_AND_MAX)
, m_uClickState(0)
{
	m_uType = ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC;

	m_xFlashColour.Set( 1.0f, 1.0f, 1.0f, 1.0f);

	ASURA_GUIMENU_PROFILE_COUNTINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_NUMERIC);
}

// --------------------------------------------------------------------------------------- 
// -- Destructor -------------------------------------------------------------------------

Asura_GUIMenu_Widget_Numeric::~Asura_GUIMenu_Widget_Numeric(void)
{
	ASURA_GUIMENU_PROFILE_COUNTDESTROYINSTANCE(ASURA_GUIMENU_COUNTER__WIDGET_NUMERIC);
}

// --------------------------------------------------------------------------------------- 
// -- Public Functions ------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::Render( u_int uState)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_NUMERIC);

	u_int uDecColBackup = 0xffffffff;
	u_int uIncColBackup = 0xffffffff;


	if( m_pxDecImage)
	{
		uDecColBackup = m_pxDecImage->GetColour().ToUIntRGBA();

		if( m_fDecFlashTimerCountdown > 0.0f)
		{
			Asura_RGBA xColour = GetFlashColour();
			m_pxDecImage->SetColour( &xColour);
		}
	}

	if( m_pxIncImage)
	{
		uIncColBackup = m_pxIncImage->GetColour().ToUIntRGBA();

		if( m_fIncFlashTimerCountdown > 0.0f)
		{
			Asura_RGBA xColour = GetFlashColour();
			m_pxIncImage->SetColour( &xColour);
		}
	}



	PARENT::Render( uState);

	if( m_pxDecImage)
	{
		m_pxDecImage->SetColour( uDecColBackup);
	}

	if( m_pxIncImage)
	{
		m_pxIncImage->SetColour( uIncColBackup);
	}

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::Update( float fElapsedTime)	//, unsigned int uButtons, POINT ptMousePos)
{
	ASURA_GUIMENU_PROFILE_START_TASK(ASURA_GUIMENU_PROFILE__WIDGET_NUMERIC);

	Asura_GUIMenu_Root::Update( fElapsedTime);

	m_fElapsedTime = fElapsedTime;

	switch( GetCurrentState())
	{
		case ASURA_GUIMENU_WIDGET_STATE_DEFAULT:
			break;

		case ASURA_GUIMENU_WIDGET_STATE_ROLL_ON:

			if( IsNumericFlagSet( ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_ALWAYS_IN_EDIT_MODE_WHEN_HIGHLIGHTED))
			{
				SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_EDITING);
			}
			else
			{
				SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_HILIGHTED);
			}
			break;

		case ASURA_GUIMENU_WIDGET_STATE_HILIGHTED:
			break;

		case ASURA_GUIMENU_WIDGET_STATE_CLICKED:
			SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_HILIGHTED);
			break;

		case ASURA_GUIMENU_WIDGET_STATE_ROLL_OFF:
			SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_DEFAULT);
			break;

		case ASURA_GUIMENU_WIDGET_STATE_EDITING:
			break;

		case ASURA_GUIMENU_WIDGET_STATE_DISABLED:
		default:
			break;
	}

	const u_int uContainerCount = GetContainerCount();
	for( u_int u=0; u<uContainerCount; u++)
	{
		m_ppxContainerParts[u]->Update( fElapsedTime);
	}

	UpdateFlashingElements();
	UpdateFeedbackText();

	ASURA_GUIMENU_PROFILE_END_TASK();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::Copy( const Asura_GUIMenu_Root* pxSource)
{
	PARENT::Copy( pxSource);

	const Asura_GUIMenu_Widget_Numeric* pxSourceNumeric = static_cast<const Asura_GUIMenu_Widget_Numeric*>(pxSource);

	//update asura data
	SetMinAndMaxValue( pxSourceNumeric->m_fMinValue, pxSourceNumeric->m_fMaxValue);
	SetStepValue( pxSourceNumeric->m_fStepValue );

	if( pxSourceNumeric->IsFeedbackTextInteger() )
	{
		SetFeedbackTextAsInteger();
	}

	SetFeedbackTextFloatingPointPrecision( pxSourceNumeric->m_uFloatingPointPrecision);
	SetFlashColour( m_xFlashColour.m_fR, m_xFlashColour.m_fG, m_xFlashColour.m_fB, m_xFlashColour.m_fA);
	SetFlashTime( pxSourceNumeric->m_fFlashTimerStartVal);
	m_xConsoleVar.Copy(&pxSourceNumeric->m_xConsoleVar);

	SetNumericFlags( pxSourceNumeric->GetNumericFlags());

	// SK 28.02.07 - this will try to match and remap, the copied elements that are being used directly by the widget
	// -- feedback text --
	u_int u;
	Asura_GUIMenu_Root* pxOrgItem;
	Asura_GUIMenu_Root* pxCopyItem;
	Asura_GUIMenu_Root* pxFeedbackTextRoot = pxSourceNumeric->GetFeedbackTextElement();
	if( pxFeedbackTextRoot != NULL)
	{
		Asura_Hash_ID uHashFeedbackTextID = pxFeedbackTextRoot->GetID();

		for( u=0; u<pxSourceNumeric->GetContainerCount(); u++)
		{
			pxOrgItem = pxSourceNumeric->GetContainerItem( u);
			if( pxOrgItem->GetID() == uHashFeedbackTextID)
			{
				pxCopyItem = GetContainerItem( u);
				SetFeedbackTextElement( static_cast<Asura_GUIMenu_Element_Text*>(pxCopyItem));
			}
		}
	}
	// -- Decrement Image --
	Asura_GUIMenu_Root* pxDecImageRoot = pxSourceNumeric->GetDecrementImageElement();
	if( pxDecImageRoot != NULL)
	{
		Asura_Hash_ID uHashID = pxDecImageRoot->GetID();

		for( u=0; u<pxSourceNumeric->GetContainerCount(); u++)
		{
			pxOrgItem = pxSourceNumeric->GetContainerItem( u);
			if( pxOrgItem->GetID() == uHashID)
			{
				pxCopyItem = GetContainerItem( u);
				SetDecrementImage( static_cast<Asura_GUIMenu_Element_Image*>(pxCopyItem));
			}
		}
	}
	// -- Increment Image --
	Asura_GUIMenu_Root* pxIncImageRoot = pxSourceNumeric->GetIncrementImageElement();
	if( pxIncImageRoot != NULL)
	{
		Asura_Hash_ID uHashID = pxIncImageRoot->GetID();

		for( u=0; u<pxSourceNumeric->GetContainerCount(); u++)
		{
			pxOrgItem = pxSourceNumeric->GetContainerItem( u);
			if( pxOrgItem->GetID() == uHashID)
			{
				pxCopyItem = GetContainerItem( u);
				SetIncrementImage( static_cast<Asura_GUIMenu_Element_Image*>(pxCopyItem));
			}
		}
	}
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_Numeric::ProcessInputActivity( const Asura_GUIMenu_Input& xInput)
{
	u_int uButtonsTrapped = 0;
	const u_int uButtonPresses = xInput.m_uButtonPresses;
//	Asura_Hash_ID uHashExitID = ASURA_HASH_ID_UNSET;
	Asura_GUIMenu_Widget_State eCurrentWidgetState = GetCurrentState();

	bool bRightPressed = false;
	bool bLeftPressed = false;

	switch( eCurrentWidgetState)
	{
		case ASURA_GUIMENU_WIDGET_STATE_EDITING:
		{
			
			bLeftPressed = ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_LEFT_HELD) != 0);
			bRightPressed = ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_RIGHT_HELD) != 0);

			if( bLeftPressed)
			{
				uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_LEFT_HELD;
			}
			if( bRightPressed)
			{
				uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_RIGHT_HELD;
			}


			
			if( ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT) != 0) ||
				     ((uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_CANCEL) != 0))
			{
				if( IsNumericFlagSet( ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_ALWAYS_IN_EDIT_MODE_WHEN_HIGHLIGHTED ) == false)
				{
					SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_HILIGHTED);

					if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT) != 0)
					{
						uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_ACCEPT;
					}
					else
					{
						uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CANCEL;
					}
				}
			}

			if( IsNumericFlagSet( ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_ALWAYS_IN_EDIT_MODE_WHEN_HIGHLIGHTED ) == true)
			{
				uButtonsTrapped |= ProcessNavigationInput( uButtonPresses);
			}
		}	
		break;

		default:

			uButtonsTrapped |= ProcessNavigationInput( uButtonPresses);


			if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT) != 0)
			{
				if( eCurrentWidgetState != ASURA_GUIMENU_WIDGET_STATE_DISABLED)
				{
					SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_EDITING);
					uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_ACCEPT;
				}
			}

			break;
	}

	//check to see if increment/decrement should be attempted as a result of a click
	if(xInput.m_bUseCursor && (uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK_HELD))
	{
		if(m_uClickState)
		{
			uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK_HELD;

			bRightPressed = m_bClickOnInc;
			bLeftPressed = m_bClickOnDec;

		}
	   	else if(uButtonPresses & ASURA_GUIMENU_ACTION_BUTTON_CLICK)
		{
			//Check to see if click was on the increment or decrement elements
			if(m_pxDecImage && m_pxDecImage->ContainsPos(xInput.m_xCursorPos))
			{
				uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
				m_bClickOnDec = true;
				bLeftPressed = m_bClickOnDec;
			}
			else if(m_pxIncImage && m_pxIncImage->ContainsPos(xInput.m_xCursorPos))
			{
				uButtonsTrapped |= ASURA_GUIMENU_ACTION_BUTTON_CLICK;
				m_bClickOnInc = true;
				bRightPressed = m_bClickOnInc;
			}
		}
	}
	else
	{
		m_uClickState = 0;
	}

	
	
	//Deal with the results of any clicks/button presses
	bool bPressed = bLeftPressed | bRightPressed;

	if( HandleInputAutoRepeat( bPressed) == false)
	{
		bLeftPressed = false;
		bRightPressed = false;
	}

	if( bLeftPressed)
	{
		m_fCurrentValue -= GetStepValue();
		if( m_fCurrentValue < GetMinimumValue())
		{
			m_fCurrentValue = GetMinimumValue();
		}

		SetConsoleVar();

		StartDecrementFlash();
	}
	else if( bRightPressed)
	{
		m_fCurrentValue += GetStepValue();
		if( m_fCurrentValue > GetMaximumValue())
		{
			m_fCurrentValue = GetMaximumValue();
		}

		SetConsoleVar();

		StartIncrementFlash();
	}


	return uButtonsTrapped;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::HandleInputAutoRepeat( bool bHeldValue)
{
	bool bRetVal = bHeldValue;

	if( bHeldValue == true)
	{
		switch( m_uButtonRepeatPhase)
		{
			case 1:
				m_fCurrentRepeatDelay -= m_fElapsedTime;
				if( m_fCurrentRepeatDelay < 0.0f)
				{
					m_fCurrentRepeatDelay += m_fSecondaryRepeatDelay;
					m_uButtonRepeatPhase++;
				}
				else
				{
					bRetVal = false;
				}
				break;

			case 2:
				m_fCurrentRepeatDelay -= m_fElapsedTime;
				if( m_fCurrentRepeatDelay < 0.0f)
				{
					m_fCurrentRepeatDelay += m_fSecondaryRepeatDelay;
				}
				else
				{
						bRetVal = false;		
				}
				break;

			default:
				// -- setup case 1: details
				m_uButtonRepeatPhase++;
				m_fCurrentRepeatDelay = m_fInitialRepeatDelay;	
				break;
		}
	}
	else
	{
		m_uButtonRepeatPhase = 0;
		m_fCurrentRepeatDelay = 0.0f;
	}

	return bRetVal;
}

// --------------------------------------------------------------------------------------- 

//void Asura_GUIMenu_Widget_Numeric::ProcessInputActivity( u_int uButtonPresses)
//{
//	Asura_Hash_ID uHashExitID = ASURA_HASH_ID_UNSET;
//
//	if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_ACCEPT_HELD) != 0)
//	{
//		if( GetCurrentState() != ASURA_GUIMENU_WIDGET_STATE_DISABLED)
//		{
//			if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_LEFT) != 0)
//			{
//				m_fCurrentValue -= GetStepValue();
//				if( m_fCurrentValue < GetMinimumValue())
//				{
//					m_fCurrentValue = GetMinimumValue();
//				}
//
//				SetConsoleVar();
//
//				StartDecrementFlash();
//			}
//			else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_RIGHT) != 0)
//			{
//				m_fCurrentValue += GetStepValue();
//				if( m_fCurrentValue > GetMaximumValue())
//				{
//					m_fCurrentValue = GetMaximumValue();
//				}
//
//				SetConsoleVar();
//
//				StartIncrementFlash();
//			}
//		}
//	}
//	else
//	{
//		if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_UP) != 0)
//		{
//			uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_UP);
//		}
//		else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_DOWN) != 0)
//		{
//			uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_DOWN);		
//		}
//		else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_LEFT) != 0)
//		{
//			uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_LEFT);
//		}
//		else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_RIGHT) != 0)
//		{
//			uHashExitID = GetNavExitID( ASURA_GUIMENU_UI_NAV_EXIT_RIGHT);
//		}
//		else if( (uButtonPresses&ASURA_GUIMENU_ACTION_BUTTON_CANCEL) != 0)
//		{
//
//		}
//	}
//
//	// -- we have an exit hash
//	if( uHashExitID != ASURA_HASH_ID_UNSET)
//	{
//		// -- notify the next item to be selected
//		Asura_GUIMenu_MessagePump::PostMessage( uHashExitID, GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_IN, 0);
//
//		// -- notify the current item, to be deselected --
//		Asura_GUIMenu_MessagePump::PostMessage( GetID(), GetID(), ASURA_GUIMENU_UI_MSG_ON_ROLL_OUT, 0);
//	}
//}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::UpdateFlashingElements( void)
{
	if( m_fDecFlashTimerCountdown > 0.0f)
	{
		m_fDecFlashTimerCountdown -= m_fElapsedTime;
	}

	if( m_fIncFlashTimerCountdown > 0.0f)
	{
		m_fIncFlashTimerCountdown -= m_fElapsedTime;
	}

}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::UpdateFeedbackText( void)
{
	Asura_UnicodeChar awszText[64];
	Asura_UnicodeChar awszFormat[16];

	if( m_pxFeedbackText != NULL)
	{
		if( IsFeedbackTextInteger())
		{
			if( IsNumericFlagSet( ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_CONSOLEVAR_UPDATES))
			{
				int iVal = 0;
				m_xConsoleVar.GetIntValue( iVal);
				m_fCurrentValue = static_cast<float>(iVal);
			}

			Asura_PrintfToBfr( &awszText[0], L"%d", 64, true, Asura_Maths::FloatToInt(m_fCurrentValue));
		}
		else
		{
			if( IsNumericFlagSet( ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_CONSOLEVAR_UPDATES))
			{
				m_xConsoleVar.GetFloatValue( m_fCurrentValue);
			}

			// -- construct the format string --
			Asura_PrintfToBfr( &awszFormat[0], L"%c2.%df", 16, true, '%', GetFeedbackTextFloatingPointPrecision());

			// -- construct the final string
			Asura_PrintfToBfr( &awszText[0], awszFormat, 256, true, Asura_Maths::FtoD(m_fCurrentValue));
		}

		m_pxFeedbackText->SetText( &awszText[0]);
	}
}

// --------------------------------------------------------------------------------------- 

const Asura_RGBA Asura_GUIMenu_Widget_Numeric::GetFlashColour( void) const
{
	return m_xFlashColour;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFlashColour( float fRed, float fGreen, float fBlue, float fAlpha)
{
	m_xFlashColour.m_fR = fRed;
	m_xFlashColour.m_fG = fGreen;
	m_xFlashColour.m_fB = fBlue;
	m_xFlashColour.m_fA = fAlpha;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFlashColour( u_int uRGBAColourValue)
{
	m_xFlashColour.FromUIntRGBA( uRGBAColourValue);
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Widget_Numeric::GetInitialControlRepeatDelay( void)
{
	return m_fInitialRepeatDelay;	
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Widget_Numeric::GetSecondaryControlRepeatDelay( void)
{
	return m_fSecondaryRepeatDelay;	
}

// --------------------------------------------------------------------------------------- 
// -- UI Message handler functions -------------------------------------------------------

bool Asura_GUIMenu_Widget_Numeric::OnInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	PARENT::OnInit( pxMessage);

	InitFromConsoleVar();

	Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_REQUEST, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_REQUEST, ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC);		

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnDeInit( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	if( GetFlag( ASURA_GUIMENU_ACTION_FLAG_NOTIFY_DATA_CHANGE_ON_EXIT) == true)
	{
		SetConsoleVar();
		Asura_GUIMenu_MessagePump::PostMessage( ASURA_GUIMENU_EXTERNAL_NOTIFY, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC, (u_int)Asura_Maths::FloatToInt( m_fCurrentValue * 100.0f));
	}

	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnDataChange( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bHandled = false;

	if( pxMessage->uFromID != GetID())
	{
		// -- something external has changed the data --
		if( pxMessage->WidgetType == ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC)
		{
			switch( ASURA_GUIMENU_GET_SUB_MESSAGE( pxMessage->uActionType))
			{
				case ASURA_GUIMENU_NUMERIC_SET_INT:
				{
					m_fCurrentValue = static_cast<float>( pxMessage->uData );
					SetConsoleVar();
				}
				break;
				
				default:
					break;
			}
		}

		bHandled = true;
	}


	return bHandled;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnDataRequest( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	bool bHandled = false;

	if( pxMessage->uFromID != GetID())
	{
		Asura_GUIMenu_MessagePump::PostMessage( pxMessage->uFromID, GetID(), ASURA_GUIMENU_UI_MSG_ON_DATA_CHANGE, ASURA_GUIMENU_TYPEID_WIDGET_NUMERIC, (u_int)Asura_Maths::FloatToInt( m_fCurrentValue * 1000.0f));
		bHandled = true;
	}

	return bHandled;

}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnClicked( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnRollIn( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_ON);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnRollOut( const Asura_Message_GUIMenu_Notify* pxMessage)
{
	SetCurrentState( ASURA_GUIMENU_WIDGET_STATE_ROLL_OFF);
	return true;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::OnEnable( const Asura_Message_GUIMenu_Notify* pxMessage)
{
Asura_GUIMenu_Widget_State eNewState;

	if( pxMessage->uData != 0)
	{	
		if( IsFocusedObject())
		{	eNewState = ASURA_GUIMENU_WIDGET_STATE_HILIGHTED;	}
		else
		{	eNewState = ASURA_GUIMENU_WIDGET_STATE_DEFAULT;	}
	}
	else
	{
		eNewState = ASURA_GUIMENU_WIDGET_STATE_DISABLED;	
	}

	SetCurrentState( eNewState);

	return true;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetMaximumValue( float fMax)
{
	Asura_Assert( fMax > GetMinimumValue(), "Asura_GUIMenu_Widget_Numeric::SetMaximumValue - Max value specified is less than the minimum value.  This value will not be set.");

	if( fMax > GetMinimumValue())
	{
		m_fMaxValue = fMax;
	}

	if( m_fCurrentValue > m_fMaxValue)
	{
		m_fCurrentValue = m_fMaxValue;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetMinimumValue( float fMin)
{
	Asura_Assert( fMin < GetMaximumValue(), "Asura_GUIMenu_Widget_Numeric::SetMinimumValue - Min value specified is greater than the maximum value.  This value will not be set.");

	if( fMin < GetMaximumValue())
	{
		m_fMinValue = fMin;
	}

	if( m_fCurrentValue < m_fMinValue)
	{
		m_fCurrentValue = m_fMinValue;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetMinAndMaxValue(float fMin, float fMax)
{
	Asura_Assert(fMin<=fMax,"Asura_GUIMenu_Widget_Numeric::SetMinAndMaxValue - minimum should be less than maximum");
	m_fMinValue = Asura_Maths::Min(fMin,fMax);
	m_fMaxValue = Asura_Maths::Max(fMin,fMax);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetValue( const float fValue )
{
	m_fCurrentValue = Asura_Maths::ClampToRange( fValue,m_fMinValue,m_fMaxValue );
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetStepValue( float fStep)
{
	if (GetMaximumValue() == FLT_MAX || GetMinimumValue() == -FLT_MAX)
	{
		// MGH 22/8/2007: In this case trying to calculate the range is likely not to work properly
		// (i.e. give a floating point overflow), so we may as well just accept the step provided.
		m_fStepValue = fStep;
		return;
	}

	float fRange = GetMaximumValue() - GetMinimumValue();

	Asura_Assert( fStep <= fRange, "Asura_GUIMenu_Widget_Numeric::SetStepValue - Step specified is greater than the range given by Min & Max.");

	if( fStep <= fRange)
	{
		m_fStepValue = fStep;
	}
}
// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Widget_Numeric::GetStepValue( void) const
{
	return m_fStepValue;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::AddStep( void)
{
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SubstractStep( void)
{
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::StartDecrementFlash( void)
{
	m_fDecFlashTimerCountdown = GetFlashTime();
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::StartIncrementFlash( void)
{
	m_fIncFlashTimerCountdown = GetFlashTime();
}

// --------------------------------------------------------------------------------------- 

// --------------------------------------------------------------------------------------- 
// -- Protected Functions ------------------------------------------------------------------


// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetupDefaultElements( void)
{
}

// --------------------------------------------------------------------------------------- 
#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_Numeric::UpdateLocalDataFromStyle( void)
{
	Asura_GUIMenu_Widget_Numeric* pxStyleNumeric;

	PARENT::UpdateLocalDataFromStyle();

	if( m_pxStyleObject != NULL)
	{
		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_NUMERIC_LOCAL_FLAG_MINIMUM) == 0)
		{
			pxStyleNumeric = static_cast<Asura_GUIMenu_Widget_Numeric*>(m_pxStyleObject);
			SetMinimumValue( pxStyleNumeric->GetMinimumValue());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_NUMERIC_LOCAL_FLAG_MAXIMUM) == 0)
		{
			pxStyleNumeric = static_cast<Asura_GUIMenu_Widget_Numeric*>(m_pxStyleObject);
			SetMaximumValue( pxStyleNumeric->GetMaximumValue());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_NUMERIC_LOCAL_FLAG_STEP) == 0)
		{
			pxStyleNumeric = static_cast<Asura_GUIMenu_Widget_Numeric*>(m_pxStyleObject);
			SetStepValue( pxStyleNumeric->GetStepValue());
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_NUMERIC_LOCAL_FLAG_FEEDBACK_DETAILS) == 0)
		{
			pxStyleNumeric = static_cast<Asura_GUIMenu_Widget_Numeric*>(m_pxStyleObject);
			SetFeedbackTextFloatingPointPrecision( pxStyleNumeric->GetFeedbackTextFloatingPointPrecision());
			if( pxStyleNumeric->IsFeedbackTextFloatingPoint() == true)
			{
				SetFeedbackTextAsFloatingPoint();
			}

			// now also has a update data on console var change too.

			//Probably don't want to have the use console var min/max flag controlled by styles , and if it should be,
			//it doesn't come under the category of feedback details anyway
			//RAR 24/07/2007
			const u_int uFlagsToCopy = 0xffffffff &~ ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_USE_CONSOLEVAR_MIN_AND_MAX;
			m_uNumericFlags &= ~ uFlagsToCopy;
			m_uNumericFlags |= pxStyleNumeric->GetNumericFlags() & uFlagsToCopy;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_NUMERIC_LOCAL_FLAG_FLASH_DETAILS) == 0)
		{
			pxStyleNumeric = static_cast<Asura_GUIMenu_Widget_Numeric*>(m_pxStyleObject);

			m_fFlashTimerStartVal = pxStyleNumeric->m_fFlashTimerStartVal;
			m_xFlashColour = pxStyleNumeric->m_xFlashColour;
		}

		if( (m_uStyleUseLocalDataFlags&ASURA_GUIMENU_WIDGET_NUMERIC_LOCAL_FLAG_CONTROL_REPEAT) == 0)
		{
			pxStyleNumeric = static_cast<Asura_GUIMenu_Widget_Numeric*>(m_pxStyleObject);
			
			m_fInitialRepeatDelay = pxStyleNumeric->m_fInitialRepeatDelay;
			m_fSecondaryRepeatDelay = pxStyleNumeric->m_fSecondaryRepeatDelay;
		}

	}
}
#endif
// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetDecrementImage( Asura_GUIMenu_Element_Image* pxDecImage)
{
	u_int u;
	Asura_GUIMenu_Root* pxItem;
	bool bFound = false;

	// -- need to check that the image being set is a child object of the numeric widget --
	for( u=0; u<GetContainerCount(); u++)
	{
		pxItem = GetContainerItem( u);

		if( pxDecImage == pxItem)
		{
			bFound = true;
			break;
		}
	}

	if( bFound == true)
	{
		m_pxDecImage = pxDecImage;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetIncrementImage( Asura_GUIMenu_Element_Image* pxIncImage)
{
	u_int u;
	Asura_GUIMenu_Root* pxItem;
	bool bFound = false;

	// -- need to check that the image being set is a child object of the numeric widget --
	for( u=0; u<GetContainerCount(); u++)
	{
		pxItem = GetContainerItem( u);

		if( pxIncImage == pxItem)
		{
			bFound = true;
			break;
		}
	}

	if( bFound == true)
	{
		m_pxIncImage = pxIncImage;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFeedbackTextElement( Asura_GUIMenu_Element_Text* pxFeedbackText)
{
	u_int u;
	Asura_GUIMenu_Root* pxItem;
	bool bFound = false;

	Asura_Assert( pxFeedbackText != NULL, "Asura_GUIMenu_Widget_Numeric::SetFeedbackTextElement - Passed value is null");

	// -- need to check that the image being set is a child object of the numeric widget --
	for( u=0; u<GetContainerCount(); u++)
	{
		pxItem = GetContainerItem( u);

		if( pxFeedbackText == pxItem)
		{
			bFound = true;
			break;
		}
	}

	Asura_Assert( bFound == true, "Asura_GUIMenu_Widget_Numeric::SetFeedbackTextElement - the passed image pointer is not a child of this widget");
	if( bFound == true)
	{
		m_pxFeedbackText = pxFeedbackText;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFeedbackTextAsInteger( void)
{
	m_bDisplayAsInteger = true;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFeedbackTextAsFloatingPoint( void)
{
	m_bDisplayAsInteger = false;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::IsFeedbackTextInteger( void) const
{
	return m_bDisplayAsInteger;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFeedbackTextFloatingPointPrecision( u_int uPrecision)
{
	Asura_Assert( uPrecision < ASURA_GUIMENU_WIDGET_NUMERIC_PRECISION_MAX, "Asura_GUIMenu_Widget_Numeric::SetFeedbackTextFloatingPointPrecision - Precision passed is invalid.");
	if( uPrecision < ASURA_GUIMENU_WIDGET_NUMERIC_PRECISION_MAX)
	{
		m_uFloatingPointPrecision = uPrecision;
	}
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_Numeric::GetFeedbackTextFloatingPointPrecision( void) const
{
	return m_uFloatingPointPrecision;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetFlashTime( float fFlashTime)
{
	m_fFlashTimerStartVal = fFlashTime;
}

// --------------------------------------------------------------------------------------- 

float Asura_GUIMenu_Widget_Numeric::GetFlashTime( void) const
{
	return m_fFlashTimerStartVal;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::IsFeedbackTextFloatingPoint( void) const
{
	return (m_bDisplayAsInteger == false);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetNumericFlags( u_int uNumericFlags)
{
	m_uNumericFlags = uNumericFlags;
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetNumericFlag( u_int uNumericFlag, bool bTrueOrFalse)
{
	if( bTrueOrFalse)
	{
		m_uNumericFlags |= uNumericFlag;
	}
	else
	{
		m_uNumericFlags &= (0xffffffff - uNumericFlag);
	}
}

// --------------------------------------------------------------------------------------- 

u_int Asura_GUIMenu_Widget_Numeric::GetNumericFlags( void) const
{
	return m_uNumericFlags;
}

// --------------------------------------------------------------------------------------- 

u_int* Asura_GUIMenu_Widget_Numeric::GetNumericFlagsAsPtr( void)
{
	return &m_uNumericFlags;
}

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::IsNumericFlagSet( u_int uNumericFlag) const
{
	return ((m_uNumericFlags&uNumericFlag) != 0);
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::ReadFromChunkStream( Asura_Chunk_Stream& xStream, Asura_GUIMenu_Root* pxParentObject, u_int uChunkVersion)
{
	PARENT::ReadFromChunkStream( xStream, pxParentObject, 0xffffffff);

	ReadAllChildObjects( xStream, this);

	// read in numeric specific data --
	float fMax = 0.0f;
	xStream >> fMax;

	// MGH 22/8/2007: Fix up old corrupted data to prevent asserts. Caused by a bug elsewhere, which has been fixed.
	if (fMax == static_cast<float>(INT_MIN))
	{
		fMax = static_cast<float>(INT_MAX);
	}

	float fMin = 0.0f;
	xStream >> fMin;
	SetMinAndMaxValue(fMin, fMax);

	float fTemp = 0.0f;
	xStream >> fTemp;				
	SetStepValue( fTemp);

	// flash details
	xStream >> m_xFlashColour.m_fR;
	xStream >> m_xFlashColour.m_fG;
	xStream >> m_xFlashColour.m_fB;
	xStream >> m_xFlashColour.m_fA;

	xStream >> m_fFlashTimerStartVal;

	// Feedback display
	u_int uDisplayAsInt;
	xStream >> uDisplayAsInt;

	if( uDisplayAsInt == 1)
	{
		SetFeedbackTextAsInteger();
	}
	else
	{
		SetFeedbackTextAsFloatingPoint();
	}

	u_int uPrecision;
	xStream >> uPrecision;
	SetFeedbackTextFloatingPointPrecision( uPrecision);

	// -- need to patch in the following hashes to the correct pointers - 
	Asura_Hash_ID uHashDecImage = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uHashIncImage = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uHashTextFeedback = ASURA_HASH_ID_UNSET;

	xStream >> uHashDecImage;
	xStream >> uHashIncImage;
	xStream >> uHashTextFeedback;

	Asura_GUIMenu_Root* pxTempRoot = NULL;

	if( uHashDecImage != ASURA_HASH_ID_UNSET)
	{
		pxTempRoot = FindContainerItemByID( uHashDecImage);
		SetDecrementImage( static_cast<Asura_GUIMenu_Element_Image*>(pxTempRoot));
	}

	if( uHashIncImage != ASURA_HASH_ID_UNSET)
	{
		pxTempRoot = FindContainerItemByID( uHashIncImage);
		SetIncrementImage( static_cast<Asura_GUIMenu_Element_Image*>(pxTempRoot));
	}

	if( uHashTextFeedback != ASURA_HASH_ID_UNSET)
	{
		pxTempRoot = FindContainerItemByID( uHashTextFeedback);
		SetFeedbackTextElement( static_cast<Asura_GUIMenu_Element_Text*>(pxTempRoot));
	}

	if (uChunkVersion >= 1)
	{
		// console var
		m_xConsoleVar.ReadFromChunkStream(xStream);
	}

	if( uChunkVersion >= 2)
	{
		xStream >> m_fInitialRepeatDelay;
		xStream >> m_fSecondaryRepeatDelay;
	}

	if( uChunkVersion >= 3)
	{
		xStream >> m_uNumericFlags;
	}

	if( uChunkVersion < 4)
	{
		//Earlier versions take the min/max from the console var by default. RAR 24/07/2007
		m_uNumericFlags |= ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_USE_CONSOLEVAR_MIN_AND_MAX;
	}

	UpdateFeedbackText();
}

// --------------------------------------------------------------------------------------- 

#ifdef ASURA_TOOLS
void Asura_GUIMenu_Widget_Numeric::WriteToChunkStream( Asura_Chunk_Stream& xStream) const
{
	WriteLocalChunkHeader( xStream, ASURA_CHUNK_WIDGET_NUMERIC_VERSION, GetType());

	PARENT::WriteToChunkStream( xStream);

	// -- write out any children --
	WriteChildObjectData( xStream);

	// -- write out numeric data --
	xStream << m_fMaxValue;
	xStream << m_fMinValue;
	xStream << m_fStepValue;

	// flash details
	xStream << m_xFlashColour.m_fR;
	xStream << m_xFlashColour.m_fG;
	xStream << m_xFlashColour.m_fB;
	xStream << m_xFlashColour.m_fA;

	xStream << m_fFlashTimerStartVal;

	// -- Feedback display info
	xStream << static_cast<u_int>(m_bDisplayAsInteger);	
	xStream << m_uFloatingPointPrecision;


	Asura_Hash_ID uHashDecImage = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uHashIncImage = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uHashFeedBackText = ASURA_HASH_ID_UNSET;

	if( m_pxDecImage)
	{
		uHashDecImage = m_pxDecImage->GetID();
	}

	if( m_pxIncImage)
	{
		uHashIncImage = m_pxIncImage->GetID();
	}

	if( m_pxFeedbackText)
	{
		uHashFeedBackText = m_pxFeedbackText->GetID();
	}
	
	xStream << uHashDecImage;
	xStream << uHashIncImage;
	xStream << uHashFeedBackText;

	m_xConsoleVar.WriteToChunkStream( xStream);

	xStream << m_fInitialRepeatDelay;
	xStream << m_fSecondaryRepeatDelay;

	xStream << m_uNumericFlags;
}

void Asura_GUIMenu_Widget_Numeric::ExportToXML(Tools_GUIMenu_XMLExporter* pxExporter) const
{
	PARENT::ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<Numeric_data>\n" );
	pxExporter->IncTabs();

	// -- write out numeric data --

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<max_value>" );
	pxExporter->AddToBuffer( m_fMaxValue );
	pxExporter->AddToBuffer( "</max_value>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<min_value>" );
	pxExporter->AddToBuffer( m_fMinValue );
	pxExporter->AddToBuffer( "</min_value>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<step_value>" );
	pxExporter->AddToBuffer( m_fStepValue );
	pxExporter->AddToBuffer( "</step_value>\n" );

	// flash details
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<flash_colour r=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fR );
	pxExporter->AddToBuffer( "\" g=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fG );
	pxExporter->AddToBuffer( "\" b=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fB );
	pxExporter->AddToBuffer( "\" a=\"" );
	pxExporter->AddToBuffer( m_xFlashColour.m_fA );
	pxExporter->AddToBuffer("\" />\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<flash_timer_start>" );
	pxExporter->AddToBuffer( m_fFlashTimerStartVal );
	pxExporter->AddToBuffer( "</flash_timer_start>\n" );

	// -- Feedback display info
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<display_as_integer>" );
	pxExporter->AddToBuffer( m_bDisplayAsInteger );
	pxExporter->AddToBuffer( "</display_as_integer>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<floating_point_precision>" );
	pxExporter->AddToBuffer( m_uFloatingPointPrecision );
	pxExporter->AddToBuffer( "</floating_point_precision>\n" );

	if( m_pxDecImage)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<dec_image_hash>" );
		pxExporter->AddToBuffer( m_pxDecImage->GetID() );
		pxExporter->AddToBuffer( "</dec_image_hash>\n" );
	}

	if( m_pxIncImage)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<inc_image_hash>" );
		pxExporter->AddToBuffer( m_pxIncImage->GetID() );
		pxExporter->AddToBuffer( "</inc_image_hash>\n" );
	}

	if( m_pxFeedbackText)
	{
		pxExporter->AddToBuffer( pxExporter->GetTabs() );
		pxExporter->AddToBuffer( "<feedback_text>" );
		pxExporter->AddToBuffer( m_pxFeedbackText->GetID() );
		pxExporter->AddToBuffer( "</feedback_text>\n" );
	}

	m_xConsoleVar.ExportToXML( pxExporter );
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<numeric_flags>" );
	pxExporter->AddToBuffer( m_uNumericFlags );
	pxExporter->AddToBuffer( "</numeric_flags>\n" );


	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<initial_repeat_delay>" );
	pxExporter->AddToBuffer( m_fInitialRepeatDelay );
	pxExporter->AddToBuffer( "</initial_repeat_delay>\n" );

	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "<secondary_repeat_delay>" );
	pxExporter->AddToBuffer( m_fSecondaryRepeatDelay );
	pxExporter->AddToBuffer( "</secondary_repeat_delay>\n" );

	pxExporter->DecTabs();
	pxExporter->AddToBuffer( pxExporter->GetTabs() );
	pxExporter->AddToBuffer( "</Numeric_data>\n" );

}

#endif // ASURA_TOOLS

// --------------------------------------------------------------------------------------- 

bool Asura_GUIMenu_Widget_Numeric::RemoveFromThisContainer( Asura_GUIMenu_Root* pxObject)
{
	//Check for one of the children that we have a pointer to being removed.
	if(m_pxDecImage == pxObject) m_pxDecImage = 0;
	if(m_pxIncImage == pxObject) m_pxIncImage = 0;
	if(m_pxFeedbackText == pxObject) m_pxFeedbackText = 0;

	return PARENT::RemoveFromThisContainer(pxObject);
}


// --------------------------------------------------------------------------------------- 

// helpers to get/set the console var
void Asura_GUIMenu_Widget_Numeric::InitFromConsoleVar()
{
#ifdef ASURA_TOOLS
	// in tools we might have the command set to "*None*", as this is only 
	// changed to a blank string when saving the chunk.
	if (!m_xConsoleVar.GetName().Compare(ASURA_GUIMENU_CONSOLEVAR_NONE))
	{
		return;
	}
#endif

	if( m_xConsoleVar.GetName().IsBlank())
	{
		// in certain cases, the numerical widget doesn't require a console var.
		// this should stop the assert happening when the VarRepository cannot find a blank string.
		return;
	}

	Asura_CommandConsole_VarRepository::Var* pxVar = Asura_CommandConsole_VarRepository::Find( m_xConsoleVar.GetName().GetString());
	if( pxVar == NULL)
	{
		Asura_Assertf(false, ("failed to find console var %s", m_xConsoleVar.GetName().GetString()));
		return;
	}

	if (pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_INT)
	{
		if(IsNumericFlagSet(ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_USE_CONSOLEVAR_MIN_AND_MAX))
		{
			SetMinAndMaxValue( static_cast<float>(pxVar->m_xData.iLowerLimit), static_cast<float>(pxVar->m_xData.iUpperLimit));
		}
		int iValue;
		m_xConsoleVar.GetIntValue(iValue);
		m_fCurrentValue = static_cast<float>(iValue);
	}
	else if (pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT)
	{
		if(IsNumericFlagSet(ASURA_GUIMENU_WIDGET_NUMERIC_FLAGS_USE_CONSOLEVAR_MIN_AND_MAX))
		{
			SetMinAndMaxValue( pxVar->m_xData.fLowerLimit, pxVar->m_xData.fUpperLimit);
		}
		m_xConsoleVar.GetFloatValue( m_fCurrentValue );
	}
	else
	{
		Asura_Assertf(false, ("console var %s for numerical widget is not int or float type", m_xConsoleVar.GetName().GetString()));
		return;
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetConsoleVar() const
{
	Asura_CommandConsole_VarRepository::Var* pxVar = Asura_CommandConsole_VarRepository::Find( m_xConsoleVar.GetName().GetString());
	if( pxVar == NULL)
	{
		return;
	}

	if (pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_INT)
	{
		m_xConsoleVar.SetIntValue(static_cast<int>(m_fCurrentValue));
	}
	else if (pxVar->m_eType == Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT)
	{
		m_xConsoleVar.SetFloatValue( m_fCurrentValue );
	}
}

// --------------------------------------------------------------------------------------- 

void Asura_GUIMenu_Widget_Numeric::SetConsoleVarName( Asura_Char* pcVarName)
{
	m_xConsoleVar.SetName( pcVarName );
}

#ifdef ASURA_TOOLS
const bool Asura_GUIMenu_Widget_Numeric::FindAndReplaceConsoleVarName( const Asura_UnicodeChar* wszName,const Asura_UnicodeChar* wszReplacementName )
{
	bool bRtn(false);
	
	bRtn |= PARENT::FindAndReplaceConsoleVarName( wszName,wszReplacementName );
	bRtn |= m_xConsoleVar.FindAndReplaceConsoleVarName(wszName,wszReplacementName);

	return bRtn;
}
#endif // ASURA_TOOLS


// --------------------------------------------------------------------------------------- 




// --------------------------------------------------------------------------------------- 
// -- Private Functions ------------------------------------------------------------------


// --------------------------------------------------------------------------------------- 

#endif // ASURA_USE_GUI_MENU_SYSTEM
