
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ControlMapping_Key.cpp
// Author : Andy Tate
// Date   : 1-Dec-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Input/Asura_ControlMapping_Key.h"

// Asura independent header-files...
#include "Debugging/Asura_Assert.h"
#ifdef ASURA_USE_NEW_INPUT
#include "Input/Asura_ControlMap.h"
#include "Input/Asura_Input.h"
#endif
#include "Core/Asura_UserInput.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

const float	fDEBOUNCE_THRESHOLD	= 0.5f;

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ControlMapping_Key::Asura_ControlMapping_Key
// Author       : Andy Tate
// Date         : 1-Dec-2005
//
// Description  : Constructor function for the key control-mapping.
//
// Parameters   : The ID of the key to use and the type of hold to check for.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_ControlMapping_Key::Asura_ControlMapping_Key (const int iKeyID,
													const KeyHoldType eHoldType,
													const float fScale) :
	Asura_ControlMapping(fScale),
	m_iKeyID(iKeyID),
	m_eKeyHoldType(eHoldType),
	m_fOldValue(0.0f),
	m_fNewValue(0.0f)
{
	// Nothing more to initialise...
	Asura_Assert(iKeyID < MAX_NUMBER_OF_ASURA_KEYIDS, "invalid key ID");
}

/// End of function 'Asura_ControlMapping_Key::Asura_ControlMapping_Key'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ControlMapping_Key::UpdateValue
// Author       : Andy Tate
// Date         : 1-Dec-2005
////////////////////////////////////////////////////////////////////////////////

void Asura_ControlMapping_Key::UpdateValues(const bool bActive, float& fValue, float& fHUDValue)
{
#ifdef ASURA_USE_NEW_INPUT
	m_fOldValue		= m_fNewValue;
	if (m_bActive || !GetPauseWhileInactive())
	{
		m_fNewValue		= Asura_Input::GetKeyValue(m_pxOwner->GetControllerIndex(), m_iKeyID);
	}

	fValue			= CalculateValue(m_eKeyHoldType, m_fOldValue, m_fNewValue);
	fHUDValue		= 0.0f;
#else
	// Handle the unspecified key case...
	if(m_iKeyID == ASURA_KEYID_VOID)
	{
		fValue		= 0.0f;
		fHUDValue	= 0.0f;
	}
	else
	{
		const int	iKeyValue	= Asura_UserInput::GetKeyValue(m_iKeyID);

		// Calculate the values...
		m_fOldValue		= m_fNewValue;
		if (m_bActive || !GetPauseWhileInactive())
		{
			m_fNewValue		= static_cast<float>(iKeyValue) / 255.0f;
		}

		fValue			= CalculateValue(m_eKeyHoldType, m_fOldValue, m_fNewValue);
		fHUDValue		= 0.0f;
	}
#endif

	// Update the active flags...
	switch (m_eKeyHoldType)
	{
	default:
		if(bActive)
		{
			if(!m_bActive && (fValue == 0.0f))
			{
				m_bActive	= true;
			}
		}
		else
		{
			m_bActive	= false;
		}
		break;

	case ON_RELEASE:
	case ON_DEBOUNCED_RELEASE:
		// Don't wait on this one, just start reporting immediately.
		m_bActive = bActive;
		break;
	}
}

/// End of function 'Asura_ControlMapping_Key::UpdateValue'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ControlMapping_Key::CalculateKeyValue
// Author       : Andy Tate
// Date         : 1-Dec-2005
//
// Description  : Calculates the current key value based on the previous and
//				  current key values and the type of hold we're using.
//
// Parameters   : The old and new values of the key and the type of hold to use.
//
// Return value : The current value of the key (between 0.0f and 1.0f).
////////////////////////////////////////////////////////////////////////////////

float Asura_ControlMapping_Key::CalculateValue (const KeyHoldType eHoldType,
												const float fOldValue,
												const float fNewValue)
{
	// What sort of hold-type are we checking for?
	switch(eHoldType)
	{
		case ON_PRESS:
		{
			// Is the key pressed at all?
			return(fNewValue);
		}

		case ON_RELEASE:
		{
			// Is the key released at all?
			return(1.0f - fNewValue);
		}

		case ON_DEBOUNCED_PRESS:
		{
			// Have we just pressed the key?
			if((fOldValue < fDEBOUNCE_THRESHOLD) && (fNewValue >= fDEBOUNCE_THRESHOLD))
			{
				return(fNewValue);
			}

			else
			{
				return(0.0f);
			}
		}

		case ON_DEBOUNCED_RELEASE:
		{
			// Have we just pressed the key?
			if((fOldValue >= fDEBOUNCE_THRESHOLD) && (fNewValue < fDEBOUNCE_THRESHOLD))
			{
				return(1.0f - fNewValue);
			}

			else
			{
				return(0.0f);
			}
		}

		default:
		{
			Asura_Assert(false, "Unexpected key hold-type specified.");
			return(0.0f);
		}
	}
}

/// End of function 'Asura_ControlMapping_Key::CalculateKeyValue'. ///

// Clear the stored value.
void Asura_ControlMapping_Key::ResetValue()
{
	m_bActive	= false;
	m_fNewValue = 0.0f;
	m_fOldValue	= 0.0f;
}


/// End of file. ///
