
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_ControlMapping_TapAxis.cpp
// Author : Anthony Gregson
// Date   : 07/06/06
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if !defined(ASURA_USE_NEW_INPUT) || defined(ASURA_USE_NEW_INPUT_LEGACY_SUPPORT)

// This file's header...
#include "Input/Asura_ControlMapping_TapAxis.h"

// Asura independent header-files...
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Debugging/Asura_Assert.h"


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ControlMapping_TapAxis::Asura_ControlMapping_TapAxis
// Author       : Anthony Gregson
// Date         : 07/06/06
//
// Description  : Constructor function for the tap-axis control mapping.
//
// Parameters   : The ID of the axis to use and the maximum duration allowed
//				  to detect the tap over.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_ControlMapping_TapAxis::Asura_ControlMapping_TapAxis (const int iAxisID,
															const float fMaxPressDuration,
															const float fTapRegisterThreshold,
															const float fScale) :
	Asura_ControlMapping(fScale),
	m_iAxisID(iAxisID),
	m_uNumTaps(1),
	m_eHoldType(ON_DEBOUNCED_RELEASE),
	m_fMaxPressDuration(fMaxPressDuration),
	m_fMaxReleaseDuration(0.0f),
	m_fTapRegisterThreshold(fTapRegisterThreshold),
	m_eState(INACTIVE),
	m_fCurrentDuration(0.0f)
{
	// Nothing more to initialise...
	Asura_Assert(iAxisID < ASURA_MAX_NUMBER_OF_AXES, "invalid axis ID");
}

/// End of function 'Asura_ControlMapping_TapAxis::Asura_ControlMapping_TapKey'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ControlMapping_TapAxis::Asura_ControlMapping_TapAxis
// Author       : Anthony Gregson
// Date         : 07/06/06
//
// Description  : Constructor function for the tap-axis control mapping.
//
// Parameters   : The ID of the axis to use and the maximum duration allowed
//				  to detect the tap over.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_ControlMapping_TapAxis::Asura_ControlMapping_TapAxis (const int iAxisID,
															const u_int uNumTaps,
															const KeyHoldType eHoldType,
															const float fMaxPressDuration,
															const float fMaxReleaseDuration,
															const float fTapRegisterThreshold,
															const float fScale) :
	Asura_ControlMapping(fScale),
	m_iAxisID(iAxisID),
	m_uNumTaps(uNumTaps),
	m_eHoldType(eHoldType),
	m_fMaxPressDuration(fMaxPressDuration),
	m_fMaxReleaseDuration(fMaxReleaseDuration),
	m_fTapRegisterThreshold(fTapRegisterThreshold),
	m_eState(INACTIVE),
	m_fCurrentDuration(0.0f)
{
	// Nothing more to initialise...
}

/// End of function 'Asura_ControlMapping_TapAxis::Asura_ControlMapping_TapAxis'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_ControlMapping_TapAxis::UpdateValue
// Author       : Anthony Gregson
// Date         : 07/06/06
////////////////////////////////////////////////////////////////////////////////

void Asura_ControlMapping_TapAxis::UpdateValues(const bool bActive, float& fValue, float& fHUDValue)
{
	const float	fTimeStep = Asura_Timers::GetFrameTime();
	const float	fActualAxisValue = Asura_UserInput::GetAxisValue(m_iAxisID);
	const float fModifiedAxisValue = (Asura_Maths::Sign(fActualAxisValue) == Asura_Maths::Sign(m_fTapRegisterThreshold)) ? Asura_Maths::Abs(fActualAxisValue) : 0.0f;

	// Update the active flags...
	m_bActive	= bActive;

	if (!m_bActive && GetPauseWhileInactive())
	{
		fValue		= 0.0f;
		fHUDValue	= 0.0f;
		return;
	}

	// What state are we currently in?
	switch (m_eState)
	{
		case INACTIVE:
		{
			// Is the axis held-down?
			if(fModifiedAxisValue > Asura_Maths::Abs(m_fTapRegisterThreshold))
			{
				// Yes - begin handling the pressing...
				m_eState			= PRESSING;
				m_uCurrentNumTaps	= 1;
				m_fCurrentDuration	= fTimeStep;

				// Have we reached the limit on number of taps?
				if( (m_uCurrentNumTaps >= m_uNumTaps) &&
					((m_eHoldType == ON_PRESS) || (m_eHoldType == ON_DEBOUNCED_PRESS)) )
				{
					fValue		= 1.0f;
					fHUDValue	= 0.0f;
					return;
				}
				else
				{
					fValue		= 0.0f;
					fHUDValue	= 0.0f;
					return;
				}
			}
			else
			{
				fValue		= 0.0f;
				fHUDValue	= 0.0f;
				return;
			}
		}

		case PRESSING:
		{
			// Is the key still held-down?
			if(fModifiedAxisValue > Asura_Maths::Abs(m_fTapRegisterThreshold))
			{
				// Yes - begin handling the pressing...
				m_fCurrentDuration	+= fTimeStep;

				if( (m_uCurrentNumTaps >= m_uNumTaps) && (m_eHoldType == ON_PRESS) )
				{
					fValue		= 1.0f;
					fHUDValue	= 0.0f;
					return;
				}
				else
				{
					fValue		= 0.0f;
					fHUDValue	= 0.0f;
					return;
				}
			}
			else
			{
				if(m_fCurrentDuration <= m_fMaxPressDuration)
				{
					// Released in time...
					m_eState			= RELEASING;
					m_fCurrentDuration	= fTimeStep;

					// Have we reached the limit on number of taps?
					if( (m_uCurrentNumTaps >= m_uNumTaps) && 
						((m_eHoldType == ON_RELEASE) || (m_eHoldType == ON_DEBOUNCED_RELEASE)) )
					{
						fValue		= 1.0f;
						fHUDValue	= 0.0f;
						return;
					}
					else
					{
						fValue		= 0.0f;
						fHUDValue	= 0.0f;
						return;
					}
				}
				else
				{
					// Held for too long - time to reset everything...
					m_eState	= INACTIVE;
					fValue		= 0.0f;
					fHUDValue	= 0.0f;
					return;
				}
			}
		}

		case RELEASING:
		{
			// Is the key being held-down again?
			if(fModifiedAxisValue > Asura_Maths::Abs(m_fTapRegisterThreshold))
			{
				// Yes - have we released it for too long?
				if(m_fCurrentDuration <= m_fMaxReleaseDuration)
				{
					// No - begin handling the pressing...
					m_eState			 = PRESSING;
					m_uCurrentNumTaps	+= 1;
					m_fCurrentDuration	 = fTimeStep;
				}
				else
				{
					// Yes - begin pressing again...
					m_eState			= PRESSING;
					m_uCurrentNumTaps	= 1;
					m_fCurrentDuration	= fTimeStep;
				}

				// Have we reached the limit on number of taps?
				if( (m_uCurrentNumTaps >= m_uNumTaps) &&
					((m_eHoldType == ON_PRESS) || (m_eHoldType == ON_DEBOUNCED_PRESS)) )
				{
					fValue		= 1.0f;
					fHUDValue	= 0.0f;
					return;
				}
				else
				{
					fValue		= 0.0f;
					fHUDValue	= 0.0f;
					return;
				}
			}
			else
			{
				// Yes - begin handling the pressing...
				m_fCurrentDuration	+= fTimeStep;

				if( (m_uCurrentNumTaps >= m_uNumTaps) && (m_eHoldType == ON_RELEASE) )
				{
					fValue		= 1.0f;
					fHUDValue	= 0.0f;
					return;
				}
				else
				{
					fValue		= 0.0f;
					fHUDValue	= 0.0f;
					return;
				}
			}
		}

		default:
		{
			Asura_Assert(false, "Unexpected state found.");
			fValue		= 0.0f;
			fHUDValue	= 0.0f;
			return;
		}
	}		// End of switch.
}

/// End of function 'Asura_ControlMapping_TapAxis::UpdateValue'. ///


// Clear the stored value.
void Asura_ControlMapping_TapAxis::ResetValue()
{
	m_bActive			= false;
	m_uCurrentNumTaps	= 0;
	m_fCurrentDuration	= 0.0f;
	m_eState			= INACTIVE;
}

#endif//ndef ASURA_USE_NEW_INPUT

/// End of file. ///
