//////////////////////////////////////////////////////////////////////////////
//
// Asura_CommandConsole.cpp
//
// Copyright (c) 2001-2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

// This file's header.
#include "Interface/Asura_CommandConsole.h"

// Stdlib independent.
#include <ctype.h>
#include <string.h>
#include <wchar.h>

// Engine independent.
#ifdef ASURA_USE_ANIMATION_IK_SOLVER
#include "Animation/IK/Asura_Animation_IK_Chain_LOS.h"
#include "Animation/IK/Asura_Animation_IK_Chain_MarkerPoint.h"
#include "Animation/IK/Asura_Animation_IK_Solver.h"
#endif
#include "Core/Asura_Handle.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#ifndef ASURA_COMMAND_CONSOLE_READONLY
#include "Core/Asura_UserInput.h"
#endif
#include "Entity/Intangible/Asura_ClientEntity_ConsoleVar.h"
#include "Entity/Objective/Asura_ClientEntity_Objective.h"
#include "Input/Asura_ControlMap.h"
#include "Interface/Asura_CommandConsole_IO.h"
#include "Core/Asura_MemTracker.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_REMOTE
#include "Interface/Asura_CommandConsole_Remote.h"
#endif
#include "Network/Asura_InternetSimulation.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_NetTransport_System.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_RUN
#include "FileAccess/Asura_FileAccess.h"
#endif

#ifdef ASURA_USE_CONTROLMAP_MANAGER
#include "Input/Asura_ControlManager.h"
#endif

// M A C R O S ///////////////////////////////////////////////////////////////

// Built-in cmds. These are macros so we can concatenate at compile time.
#define CONSOLE_CMDNAME_CLEAR      L"Clear"
#define CONSOLE_CMDNAME_HELP       L"Help"
#define CONSOLE_CMDNAME_HELP_SHORT L"?"
#define CONSOLE_CMDNAME_LISTCMDS   L"ListCmds"
#define CONSOLE_CMDNAME_LISTVARS   L"ListVars"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_QUIT
#define CONSOLE_CMDNAME_QUIT       L"Quit"
#endif
#define CONSOLE_CMDNAME_FONTSCALE   L"FontScale"
#define CONSOLE_CMDNAME_BORDER		L"TextBorder"
#define CONSOLE_CMDNAME_SETFONT		L"SetFont"
#define CONSOLE_CMDNAME_TIMESCALE	L"TimeScale"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_RUN
#define CONSOLE_CMDNAME_RUN			L"Run"
#endif
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
#define CONSOLE_CMDNAME_DVSBORDER	L"DVSBorder"
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int CONSOLE_INPUT_HISTORY_MAX_LINES    = 8;
static const int   CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY = -1;

// G L O B A L S /////////////////////////////////////////////////////////////

#if _DEBUG
static bool g_bConsoleInitialised = false;
#endif
static bool g_bConsoleDown;
static bool g_bConsoleEnabled;
static void (*g_pfnOutputOverride)(const Asura_UnicodeChar*);

Asura_CommandConsole_IO* g_pxIO;

// Console buffer state.
static Asura_UnicodeChar** g_pwszConsoleContents; // Circular buffer. This and related state are only used if IO implementation doesn't archive output itself.
static u_int g_uConsoleNumLinesUsed;
static u_int g_uConsoleFirstLineUsed;             // Only valid if g_uConsoleNumLinesUsed is non-zero.
static u_int g_uConsoleLastLineUsed;              //                      "

// Input buffer state.
static Asura_UnicodeChar g_wszConsoleInputBuffer[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];
static u_int             g_uConsoleInputCursorPosition;
static Asura_UnicodeChar g_wszTabCompletionBuffer[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];

// Input history state.
static Asura_UnicodeChar g_awszConsoleInputHistory[CONSOLE_INPUT_HISTORY_MAX_LINES][ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE]; // Circular buffer.
static u_int g_uConsoleNumInputHistoryLines;
static u_int g_uConsoleFirstInputHistoryLine; // Only valid if g_uConsoleNumInputHistoryLines is non-zero.
static int   g_iConsoleCurInputHistoryEntry;

//For showing some console lines while the console is up
static u_int g_uNumTimedLines = 0;
static float g_afLineTimers[CONSOLE_NUM_LINES_TO_DISPLAY];

#if defined(ASURA_USE_CONTROLMAP_MANAGER)
static Asura_ControlMap	g_xConsoleDummyControlMap(0);
#endif	// defined(ASURA_USE_CONTROLMAP_MANAGER)

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

#ifndef ASURA_COMMAND_CONSOLE_READONLY
// Built-in console cmd functions.
static bool ConsoleCmd_Clear( bool bHelp );
static bool ConsoleCmd_Help( const Asura_UnicodeChar* wszCmd, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_ListCmds( const Asura_UnicodeChar* wszPrefix, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_ListVars( const Asura_UnicodeChar* wszPrefix, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_FontScale( float fSize, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_Timescale( float fTimeScale, bool bInterrogate, bool bHelp );
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_RUN
static bool ConsoleCmd_Run( const Asura_UnicodeChar* wszFileName, bool bInterrogate, bool bHelp );
#endif
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
static bool ConsoleCmd_OpenDVS( bool bHelp );
#endif
#endif

// Utility functions.
static void AttemptTabCompletion( bool bForceFirstCompletion = false );
#ifndef ASURA_COMMAND_CONSOLE_READONLY
static void ClearConsole( void );
#endif
static void EmptyInputBuffer( void );
static void EmptyTabCompletionBuffer( void );
static void MoveInputBufferIntoHistory( void );
static bool ParseCommandBuffer( const Asura_UnicodeChar* wszCmdBfr, bool bEcho, bool bInputFromConsole, bool* pbReturnValue, void(*pfnOutputCallback)(const Asura_UnicodeChar*) );
static void ShowConsole( void );
static void UpdateOutputBufferStateForNewLine( void );
#ifndef ASURA_COMMAND_CONSOLE_READONLY
static void UseInputHistoryEntry( INPUT_HISTORY_DELTA eDelta );
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_CommandConsole.
//

#ifndef ASURA_COMMAND_CONSOLE_READONLY

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnVoid = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_VOID, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(bool, bool, bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnBool = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_BOOL, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(int, bool, bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnInt = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_INT, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(u_int, bool, bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnUint = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_UINT, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(float, bool, bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnFloat = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_FLOAT, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(const Asura_Char*, bool, bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnString = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_STRING, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::AddCmd( const Asura_UnicodeChar* wszCmdName, bool(*pfnCmdFn)(const Asura_UnicodeChar*, bool, bool), REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
	
    Asura_CommandConsole_CmdList::CmdFn xCmdFn;
    xCmdFn.pfnWString = pfnCmdFn;
    return Asura_CommandConsole_CmdList::Add( wszCmdName, Asura_CommandConsole_CmdList::CMDTYPE_WSTRING, xCmdFn, eRemove, uRestrictions );
}

bool Asura_CommandConsole::RemoveCmd( const Asura_UnicodeChar* wszCmd )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
	return Asura_CommandConsole_CmdList::Remove( wszCmd );
}

bool Asura_CommandConsole::CmdExists(const Asura_UnicodeChar* wszCmd)
{
#if _DEBUG
	Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
	return(Asura_CommandConsole_CmdList::Find(wszCmd) != NULL);
}

bool Asura_CommandConsole::AddVar( const Asura_UnicodeChar* wszVarName, bool* pbData, REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
  
    Asura_CommandConsole_VarRepository::Var::Data xData; 
    xData.pbData        = pbData;
    xData.uRestrictions = uRestrictions;
  
    return Asura_CommandConsole_VarRepository::Add( wszVarName, Asura_CommandConsole_VarRepository::Var::TYPE_BOOL, xData, eRemove );
}

bool Asura_CommandConsole::AddVar( const Asura_UnicodeChar* wszVarName, int* piData, int iLowerLimit, int iUpperLimit, REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_VarRepository::Var::Data xData;

    xData.piData        = piData;
    xData.iLowerLimit   = iLowerLimit;
    xData.iUpperLimit   = iUpperLimit;
    xData.uRestrictions = uRestrictions;

    return Asura_CommandConsole_VarRepository::Add( wszVarName, Asura_CommandConsole_VarRepository::Var::TYPE_INT, xData, eRemove );
}

bool Asura_CommandConsole::AddVar( const Asura_UnicodeChar* wszVarName, u_int* puData, u_int uLowerLimit, u_int uUpperLimit, REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
	Asura_Assert( uLowerLimit <= INT_MAX && uUpperLimit <= INT_MAX, "unsigned int is restricted to positive int range");

    Asura_CommandConsole_VarRepository::Var::Data xData;

    xData.piData        = reinterpret_cast<int*>(puData);
    xData.iLowerLimit   = static_cast<int>(uLowerLimit);
    xData.iUpperLimit   = static_cast<int>(uUpperLimit);
    xData.uRestrictions = uRestrictions;

    return Asura_CommandConsole_VarRepository::Add( wszVarName, Asura_CommandConsole_VarRepository::Var::TYPE_INT, xData, eRemove );
}

bool Asura_CommandConsole::AddVar( const Asura_UnicodeChar* wszVarName, float* pfData, float fLowerLimit, float fUpperLimit, REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

    Asura_CommandConsole_VarRepository::Var::Data xData;

    xData.pfData        = pfData;
    xData.fLowerLimit   = fLowerLimit;
    xData.fUpperLimit   = fUpperLimit;
    xData.uRestrictions = uRestrictions;
    
    return Asura_CommandConsole_VarRepository::Add( wszVarName, Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT, xData, eRemove );
}

bool Asura_CommandConsole::AddVar( const Asura_UnicodeChar* wszVarName, Asura_UnicodeChar* wszData, u_int uBfrSize, REMOVE eRemove, u_int uRestrictions )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_Assert(wcslen(wszVarName) < 1023, "Var name is too long for the DVS");
#endif

    Asura_CommandConsole_VarRepository::Var::Data xData;

    xData.wszData       = wszData;
    xData.uBfrSize      = uBfrSize;
    xData.uRestrictions = uRestrictions;
    
    return Asura_CommandConsole_VarRepository::Add( wszVarName, Asura_CommandConsole_VarRepository::Var::TYPE_STRING, xData, eRemove );
}

bool Asura_CommandConsole::RemoveVar( const Asura_UnicodeChar* wszVar, const bool bRemoveChildren )
{
#if _DEBUG
	Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
	return Asura_CommandConsole_VarRepository::Remove( wszVar, bRemoveChildren );
}

bool Asura_CommandConsole::VarExists(const Asura_UnicodeChar* wszVar)
{
#if _DEBUG
	Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
	return(Asura_CommandConsole_VarRepository::Find(wszVar) != NULL);
}


bool Asura_CommandConsole::SetVarHelpText( const Asura_UnicodeChar* wszVar, const char* szDesc )
{
	return Asura_CommandConsole_VarRepository::SetHelpText( wszVar, szDesc );
}

bool Asura_CommandConsole::MakeVarReadOnly( const Asura_UnicodeChar* wszVar )
{
	return Asura_CommandConsole_VarRepository::MakeVarReadOnly( wszVar );
}

bool Asura_CommandConsole::SetPointerToModifiedFlag( const Asura_UnicodeChar* wszVar, bool* pbModified )
{
	return Asura_CommandConsole_VarRepository::SetPointerToModifiedFlag( wszVar, pbModified );
}
#endif

bool Asura_CommandConsole::Hide( bool bForce, bool bInstant )
{
    Asura_Assert( g_bConsoleDown, "" );
	if (!g_pxIO->CanBeHidden())
	{
		return false;
	}

    if( !Project_OnConsoleUp() )
    {
        if( !bForce ) { return false; }

        Asura_Assert( false, "Console being forced up but project code tried to suppress event." );
    }

    g_bConsoleDown = false;
    g_pxIO->SetDown( false, bInstant );

    EmptyInputBuffer();
    EmptyTabCompletionBuffer();

    // Reset cur history entry.
    g_iConsoleCurInputHistoryEntry = CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY;

#if defined(ASURA_USE_CONTROLMAP_MANAGER)
	Asura_ControlManager::PopOverrideControls(&g_xConsoleDummyControlMap);
#endif	// defined(ASURA_USE_CONTROLMAP_MANAGER)

	return true;
}

bool Asura_CommandConsole::CanBeHidden()
{
	return g_pxIO->CanBeHidden();
}

void Asura_CommandConsole::EarlyInitialise( void )
{	
#if _DEBUG
    Asura_Assert( !g_bConsoleInitialised, "Console has already been initialised." );
    g_bConsoleInitialised = true;
#endif

    g_pxIO = PlatformProject_CreateIO();
    Asura_Assert( g_pxIO, "" );

    // General state.
    g_bConsoleDown      = !g_pxIO->CanBeHidden();
    g_bConsoleEnabled   = true;
    g_pfnOutputOverride = NULL;

    // Console buffer state.
    if( !g_pxIO->CanArchiveOutput() )
    {
        g_pwszConsoleContents = new Asura_UnicodeChar*[ASURA_COMMAND_CONSOLE_MAX_LINES];
        g_pwszConsoleContents[0] = new Asura_UnicodeChar[ASURA_COMMAND_CONSOLE_MAX_LINES * ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH];
        for( u_int u = 1; u < ASURA_COMMAND_CONSOLE_MAX_LINES; ++u )
        {
            g_pwszConsoleContents[u] = g_pwszConsoleContents[0] + u * ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH;
        }

        g_uConsoleNumLinesUsed = 0;
    }

    // Input buffer state.
    EmptyInputBuffer();
    EmptyTabCompletionBuffer();

    // Input history state.
    g_uConsoleNumInputHistoryLines = 0;
    g_iConsoleCurInputHistoryEntry = CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY;

    // Initialise helper classes.
    Asura_CommandConsole_CmdList::Initialise();
    Asura_CommandConsole_VarRepository::Initialise();

#ifndef ASURA_COMMAND_CONSOLE_READONLY

    // Add default cmds.
    AddCmd( CONSOLE_CMDNAME_CLEAR,      ConsoleCmd_Clear,    REMOVE_AT_SHUTDOWN );
    AddCmd( CONSOLE_CMDNAME_HELP,       ConsoleCmd_Help,     REMOVE_AT_SHUTDOWN );
    AddCmd( CONSOLE_CMDNAME_HELP_SHORT, ConsoleCmd_Help,     REMOVE_AT_SHUTDOWN );
    AddCmd( CONSOLE_CMDNAME_LISTCMDS,   ConsoleCmd_ListCmds, REMOVE_AT_SHUTDOWN );
    AddCmd( CONSOLE_CMDNAME_LISTVARS,   ConsoleCmd_ListVars, REMOVE_AT_SHUTDOWN );
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_QUIT
    AddCmd( CONSOLE_CMDNAME_QUIT,       ConsoleCmd_Quit,     REMOVE_AT_SHUTDOWN );
#endif
	AddCmd( CONSOLE_CMDNAME_FONTSCALE,	ConsoleCmd_FontScale, REMOVE_AT_SHUTDOWN );
	AddCmd( CONSOLE_CMDNAME_BORDER,		ConsoleCmd_SetBorder, REMOVE_AT_SHUTDOWN );
	AddCmd( CONSOLE_CMDNAME_SETFONT,	ConsoleCmd_SetFont,	  REMOVE_AT_SHUTDOWN );
	AddCmd( CONSOLE_CMDNAME_TIMESCALE,	ConsoleCmd_Timescale, REMOVE_AT_SHUTDOWN );
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_RUN
    AddCmd( CONSOLE_CMDNAME_RUN,        ConsoleCmd_Run, REMOVE_AT_SHUTDOWN );
#endif
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	AddCmd( CONSOLE_CMDNAME_DVSBORDER,	ConsoleCmd_SetDVSBorder, REMOVE_AT_SHUTDOWN );
#endif

#endif // #ifndef ASURA_COMMAND_CONSOLE_READONLY

    Platform_Initialise();

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_REMOTE
    Asura_CommandConsole_Remote::Initialise();
#endif

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::EarlyInitialise();	
#endif
}


bool Asura_CommandConsole::IsConsoleDown( void )
{
    return g_bConsoleDown;
}

bool Asura_CommandConsole::IsConsoleEnabled( void )
{
    return g_bConsoleEnabled;
}

bool Asura_CommandConsole::IsHandlingInput( void )
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (Asura_DebugVariableSystem::IsOpen())
	{
		return true;
	}
#endif
	return IsConsoleDown();
}

#ifndef ASURA_COMMAND_CONSOLE_READONLY
void Asura_CommandConsole::KeyPressedCallback( ASURA_KEYID eKey )
{
#ifndef ASURA_USE_NEW_INPUT
    Asura_UnicodeChar wcTemp = static_cast<Asura_UnicodeChar>(eKey);
    Asura_UserInput::UpdateInputBuffer(g_wszConsoleInputBuffer, ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE, 
            g_uConsoleInputCursorPosition, KeyPressOverride, &wcTemp, 1);
#endif
}

bool Asura_CommandConsole::KeyPressOverride( ASURA_KEYID eKey )
{

    // return true if default key pressed behaviour isn't done

    // Will be set to true if we modify the input buffer here. If any
    // function we call modifies it then we'll assume it deals with the reset
    // itself as necessary.
    
    switch (eKey)
    {
        case ASURA_KEYID_TAB:
            AttemptTabCompletion();
            return true;
        case ASURA_KEYID_ESCAPE:
            EmptyInputBuffer();
            EmptyTabCompletionBuffer();
            return true;
        case ASURA_KEYID_ENTER:       // Intentional fallthrough.
        case ASURA_KEYID_NUMPADENTER:
            ParseCommandBuffer( g_wszConsoleInputBuffer, true, true, NULL, NULL );
            return true;
        case ASURA_KEYID_PAGEUP:
            g_pxIO->Scroll( false, false );
            return true;
        case ASURA_KEYID_PAGEDOWN:
            g_pxIO->Scroll( true, false );
            return true;
        case ASURA_KEYID_HOME:
            if( Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) )
            {
                g_pxIO->Scroll( false, true );
                return true;
            }
            break;
        case ASURA_KEYID_END:
            if( Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) )
            {
                g_pxIO->Scroll( true, true );
                return true;
            }
            break;
        case ASURA_KEYID_UP:
            UseInputHistoryEntry( INPUT_HISTORY_DELTA_BACK );
            return true;
        case ASURA_KEYID_DOWN:
            UseInputHistoryEntry( INPUT_HISTORY_DELTA_FORWARD );
            return true;
        default:;
    }    

    //if( bNeedToResetTabCompletionBuffer )
    {
        EmptyTabCompletionBuffer();
    }
    
    return false;

}

void Asura_CommandConsole::ProcessAsuraInput( void )
{
    if (Project_IsConsoleToggled())
    {
        if( g_pxIO->CanBeHidden() )
        {
            Hide( false, false );
        }
    }
    else
    {
        Asura_UserInput::UpdateInputBuffer(g_wszConsoleInputBuffer, ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE, 
		                                   g_uConsoleInputCursorPosition, KeyPressOverride);
	}
}
#endif

void Asura_CommandConsole::Render( void )
{
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif
    g_pxIO->Render();	
}

void Asura_CommandConsole::Reset( void )
{
#if defined(ASURA_USE_CONTROLMAP_MANAGER)
	if(IsConsoleDown())
	{
		Asura_ControlManager::PopOverrideControls(&g_xConsoleDummyControlMap);
	}
#endif	// defined(ASURA_USE_CONTROLMAP_MANAGER)

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_REMOTE
    Asura_CommandConsole_Remote::Reset();
#endif
    Asura_CommandConsole_VarRepository::Reset();
    Asura_CommandConsole_CmdList::Reset();
    g_pxIO->Reset();

    g_bConsoleDown      = !g_pxIO->CanBeHidden();
    g_bConsoleEnabled   = true;
    g_pfnOutputOverride = NULL;

    // Could clear the console but I think that's probably that ought to
    // be done explicitly.
}

void Asura_CommandConsole::SetConsoleEnabled( bool bEnabled )
{
    if( g_pxIO->CanBeHidden() )
    {
        g_bConsoleEnabled = bEnabled;

        // If the console was down but the system has now been disabled for
        // some reason, we need to hide it.
        if( g_bConsoleDown && !g_bConsoleEnabled )
        {
            Hide( true, true );
        }
    }
    else
    {
        Asura_Assert( bEnabled, "Can't disable console since it can't be hidden." );
    }
}

void Asura_CommandConsole::Shutdown( void )
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::Shutdown();	
#endif

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_REMOTE
    Asura_CommandConsole_Remote::Shutdown();
#endif

    // Shutdown helper classes.
    Asura_CommandConsole_VarRepository::Shutdown();
    Asura_CommandConsole_CmdList::Shutdown();

    if( g_pwszConsoleContents )
    {
        delete[] g_pwszConsoleContents[0];
        delete[] g_pwszConsoleContents;
        g_pwszConsoleContents = NULL;
    }

    delete g_pxIO;
    g_pxIO = NULL;

#if _DEBUG
    g_bConsoleInitialised = false;
#endif
}

void Asura_CommandConsole::Update( void )
{	
#if _DEBUG
    Asura_Assert( g_bConsoleInitialised, "Console hasn't been initialised." );
#endif

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_REMOTE
	Asura_CommandConsole_Remote::Update();
#endif

#ifndef ASURA_COMMAND_CONSOLE_READONLY
    if( g_bConsoleDown  )
    {
        if( g_pxIO->CanSupplyInput() )
        {
            g_pxIO->GetInput( KeyPressedCallback );
        }
        else
        {
            ProcessAsuraInput();
        }
    }
    else
#endif
    {
        if( g_bConsoleEnabled && Project_IsConsoleToggled() )
        {
            if( Project_OnConsoleDown() )
            {
#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
				Asura_DebugVariableSystem::CloseDisplayDVS();
#endif	// defined(ASURA_DEBUG_VARIABLE_SYSTEM)

#if defined(ASURA_USE_CONTROLMAP_MANAGER)
				if(!g_bConsoleDown && g_bConsoleEnabled)
				{
					Asura_ControlManager::PushOverrideControls(&g_xConsoleDummyControlMap);
				}
#endif	// defined(ASURA_USE_CONTROLMAP_MANAGER)

                ShowConsole();
				ResetLineTimers();
            }
        }
    }

	UpdateLineTimers();

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::Update();
#endif

    g_pxIO->Update();
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_CommandConsole::Printf().
// PARAMETERS   wszFormat - Format string.
//              ...      - Parameters for format string.
// RETURNS      Nothing.
// DESCRIPTION  Formats string and adds to console buffer. Only deals with
//              single lines - ie. no newlines.
// AUTHOR       MP - 23.11.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_CommandConsole::Printf( const Asura_UnicodeChar* wszFormat, ... )
{
    va_list xArgs;
    va_start( xArgs, wszFormat );

    Asura_UnicodeChar wszBfr[ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH];
    Asura_PrintfToBfr( wszBfr, wszFormat, xArgs, ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH, true );

    Asura_StringUtil::StripWhitespace( wszBfr, Asura_StringUtil::STRIP_TRAILING );

    void(*pfnOutput)(const Asura_UnicodeChar*) = DefaultExecuteOutputCallback;
	bool bOverrodeOutput = g_pfnOutputOverride != 0;
    if( g_pfnOutputOverride )
    {
        pfnOutput = g_pfnOutputOverride;
        g_pfnOutputOverride = NULL; // Prevent infinite recursion if callback uses Printf() itself.
    }
    pfnOutput( wszBfr );

    if (bOverrodeOutput)
	{
		// (MH - 08.03.2007) reinstate the callback in case further calls to Printf are made
		g_pfnOutputOverride = pfnOutput;
	}

    va_end( xArgs );

    // Move to end of console buffer.
    // (MP - 25.11.2002) TODO: May want to change this behaviour. For example,
    // only move to end of console buffer if user hasn't manually scrolled
    // up.
    g_pxIO->Scroll( true, true );
}
// End function Asura_CommandConsole::Printf().


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_CommandConsole::Printf().
// PARAMETERS   szFormat - Format string.
//              ...       - Parameters for format string.
// RETURNS      Nothing.
// DESCRIPTION  Formats string and adds to console buffer. Only deals with
//              single lines - ie. no newlines.
// AUTHOR       MP - 23.11.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_CommandConsole::Printf( const Asura_Char* szFormat, ... )
{
    va_list xArgs;
    va_start( xArgs, szFormat );

    Asura_Char szBfr[ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH];
    Asura_PrintfToBfr( szBfr, szFormat, xArgs, ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH, true );

    va_end( xArgs );

    Printf( L"%S", szBfr );
}
// End function Asura_CommandConsole::Printf().

void Asura_CommandConsole::PrintRaw( const Asura_UnicodeChar* wszString )
{
	g_pfnOutputOverride(wszString);
}

void Asura_CommandConsole::SetBorderX( float fBorderX )
{
    g_pxIO->SetBorderX( fBorderX );
}

void Asura_CommandConsole::SetTexture( int iTextureID )
{
	g_pxIO->SetTexture( iTextureID );
}

void Asura_CommandConsole::SetBackgroundColour( u_int uColour )
{
	g_pxIO->SetBackgroundColour(uColour);
}

void Asura_CommandConsole::SetFont( const char* szFontName )
{
	g_pxIO->SetFont( szFontName );
}

void Asura_CommandConsole::SetFont( Asura_Handle iHandle )
{
	g_pxIO->SetFont( iHandle );
}

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
void Asura_CommandConsole::SetDVSBorder( float fBorder )
{
	Asura_CommandConsole_IO_Default::SetDVSBorder(fBorder);
}

bool Asura_CommandConsole::ConsoleCmd_SetDVSBorder( float fBorder, bool bInterrogate, bool bHelp )
{
	if (bHelp || bInterrogate)
	{
		Printf( "sets amount of screen on each side left as border (0.0 to 0.2)");
	}
	else
	{
		if (fBorder >= 0.0f && fBorder <= 0.2f)
		{
			Asura_CommandConsole_IO_Default::SetDVSBorder(fBorder);
		}
		else
		{
			Printf( "must be from 0.0 to 0.2" );
			return false;
		}
	}
	return true;
}
#endif

bool Asura_CommandConsole::ConsoleCmd_SetBorder( float fBorder, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Printf( L"changes the horizontal border. value must be between 0.0 and 10.0" );
	}
	else if (bInterrogate)
	{
		Printf( L"cur border is %.1f.", Asura_Maths::FtoD(g_pxIO->GetBorderX()) );
	}
	else
	{
		if (fBorder >= 0.0f && fBorder <= 10.0f)
		{
			g_pxIO->SetBorderX( fBorder );
		}
		else
		{
			Printf( L"value must be between 0.0 and 10.0" );
			return false;
		}
	}
	return true;
}

bool Asura_CommandConsole::ConsoleCmd_SetFont( int iFont, bool bInterrogate, bool bHelp )
{
	if (bHelp || bInterrogate)
	{
		Printf( L"sets font handle. out-of-range values will cause default font to be used" );
	}
	else
	{
		return g_pxIO->SetFont( iFont );
	}
	return true;
}

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_QUIT
bool Asura_CommandConsole::ConsoleCmd_Quit( bool bHelp )
{
    if( bHelp )
    {
        Printf( L"Closes the application." );
    }
    else
    {
        if( Project_RequestQuit() )
        {
            Printf( L"Bye!" );
			// KOR - this will clean up control maps, etc. for a clean shutdown.
			if(IsConsoleDown())
			{
				Hide(true, true);
			}
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
			Asura_DebugVariableSystem::CloseDisplayDVS();
#endif
		}
    }
    return true;
}
#endif




void Asura_CommandConsole::SetFontScale( float fFontScale )
{
    g_pxIO->SetTextScale( fFontScale );
}

float Asura_CommandConsole::GetFontScale( void )
{
	return g_pxIO->GetTextScale();
}

void Asura_CommandConsole::GetInputBfrInfo( const Asura_UnicodeChar** pwszBfr, u_int* puCursorPos )
{
    if( pwszBfr     ) { *pwszBfr     = g_wszConsoleInputBuffer;       }
    if( puCursorPos ) { *puCursorPos = g_uConsoleInputCursorPosition; }
}

void Asura_CommandConsole::GetOutputBfrInfo( const Asura_UnicodeChar* const ** ppwszBfr, u_int* puStart, u_int* puCount, u_int* puCurPos/*=NULL*/ )
{
    if( ppwszBfr ) { *ppwszBfr = g_pwszConsoleContents;   }
    if( puStart  ) { *puStart  = g_uConsoleFirstLineUsed; }
    if( puCount  ) { *puCount  = g_uConsoleNumLinesUsed;  }
	if( puCurPos ) { if(g_pxIO){*puCurPos = g_pxIO->GetFirstLineToDisplay();} }
}

bool Asura_CommandConsole::Execute( const Asura_UnicodeChar* wszCommand, bool* pbReturnValue, void(*pfnOutputCallback)(const Asura_UnicodeChar*) )
{
    return ParseCommandBuffer( wszCommand, false, false, pbReturnValue, pfnOutputCallback );
}

void Asura_CommandConsole::DefaultExecuteOutputCallback( const Asura_UnicodeChar* wszOutput )
{
    if( g_pxIO->CanArchiveOutput() )
    {
        g_pxIO->AddOutput( wszOutput );
    }
    else
    {
        do 
        {
            const Asura_UnicodeChar* wszSplitPoint = g_pxIO->GetLineSplitPoint(wszOutput);
            int iCharsThisLine = wszSplitPoint - wszOutput + 1;
            if (iCharsThisLine < 0 || iCharsThisLine > ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH)
            {
                Asura_Assert(false, "Output buffer truncated");
                iCharsThisLine = ASURA_COMMAND_CONSOLE_MAX_LINE_LENGTH;
            }
            UpdateOutputBufferStateForNewLine();

            Asura_PrintfToBfr( g_pwszConsoleContents[g_uConsoleLastLineUsed], L"%s", iCharsThisLine, true, wszOutput );
            wszOutput = wszSplitPoint;
        }
        while (wszOutput && *wszOutput);
    }
}

//
// Asura_CommandConsole_CmdList.
//

Asura_CommandConsole_CmdList::Node* Asura_CommandConsole_CmdList::s_pxRoot;
Asura_TemplateHashTree<Asura_CommandConsole_CmdList::Node*> Asura_CommandConsole_CmdList::s_xNodeTree;

bool Asura_CommandConsole_CmdList::Add( const Asura_UnicodeChar* wszCmdName, CMDTYPE eType, const CmdFn& xCmdFn, Asura_CommandConsole::REMOVE eRemove, u_int uRestrictions )
{
    if( !wszCmdName || !*wszCmdName || !xCmdFn.pfnVoid )
    {
        Asura_Assert( false, "Invalid parameter(s) passed to AddCmd()." );
        return false;
    }

    // Verify that this command does not already exist.
	if( Find(wszCmdName) )
    {
        Asura_Assertf( false,("Command name %S already exists.",wszCmdName) );
        return false;
    }

	if ( Asura_CommandConsole_VarRepository::Find(wszCmdName) )
	{
        Asura_Assertf( false, ("Command name %S is already used for a variable.", wszCmdName) );
        return false;
	}

    // Allocate new list node.
    Node* pxNewNode( new Node(wszCmdName,eType,xCmdFn) );
    
	if( !pxNewNode )
    {
        return false;
    }

    // Setup node data.
    pxNewNode->SetRemoveType( eRemove );
    pxNewNode->SetRestrictions( uRestrictions );

    // Insert into list. We want to maintain alphabetical ordering, for ease
    // of use.
    {
        Node* pxPrev = NULL;
        Node* pxNext = s_pxRoot;

        while( pxNext && wcsicmp(pxNewNode->GetName(), pxNext->GetName()) > 0 )
        {
            pxPrev = pxNext;
            pxNext = pxNext->GetNext();
        }

        if( pxPrev )
        {
            pxPrev->SetNext( pxNewNode );
        }
        else
        {
            s_pxRoot = pxNewNode;
        }

        pxNewNode->SetNext( pxNext );
    }

	// Insert into hash tree
	char szCmdName[128];
	Asura_PrintfToBfr(szCmdName, "%S", 128, true, wszCmdName);
	Node** ppxAllocatedNode = s_xNodeTree.AddNode(Asura_GetHashID(szCmdName));
	
	Asura_Assert( ppxAllocatedNode,"Failed to allocate new node in hash tree" );

	if( ppxAllocatedNode )
	{
		*ppxAllocatedNode = pxNewNode;
	}

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::AddCmd(wszCmdName, pxNewNode);
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

    return true;
}

bool Asura_CommandConsole_CmdList::Remove( const Asura_UnicodeChar* wszCmdName )
{
	if ( !wszCmdName || !*wszCmdName )
	{
        Asura_Assert( false, "Invalid parameter(s) passed to Remove()." );
        return false;
    }

    // find the node for this command.
	Node* pxNodeToRemove = s_pxRoot;
	while (pxNodeToRemove && wcsicmp(pxNodeToRemove->GetName(), wszCmdName))
	{
		pxNodeToRemove = pxNodeToRemove->GetNext();
	}

	if (!pxNodeToRemove)
    {
        Asura_Assertf( false, ("Trying to unregister console command '%S', which doesn't exist.",
							   wszCmdName) );
        return false;
    }

	return Remove( pxNodeToRemove );
}

bool Asura_CommandConsole_CmdList::Remove( Node* const pxNodeToRemove )
{
	Asura_Assert( pxNodeToRemove,"Trying to remove NULL node" );

	bool bSuccess(false);

	if( pxNodeToRemove )
	{
		char szCmdName[128];
		Asura_PrintfToBfr(szCmdName, "%S", 128, true, pxNodeToRemove->GetName());
		
		// Remove from hash tree
		bSuccess = s_xNodeTree.DeleteNodeByHash( Asura_GetHashID(szCmdName) );
		
		Asura_Assertf( bSuccess,("Failed to remove command '%s' from hash tree",szCmdName) );

		// remove node from internal linked list iff node has been removed from hash tree
		if( bSuccess )
		{
			// special case for removing root node - NDB
			if( pxNodeToRemove == s_pxRoot )
			{
				s_pxRoot = s_pxRoot->GetNext();
				
			}
			else
			{
				// find the previous node by iterating through the list from the root
				Node* pxPrevNode = s_pxRoot;
				while (pxPrevNode && pxPrevNode->GetNext() != pxNodeToRemove)
				{
					pxPrevNode = pxPrevNode->GetNext();
				}

				Asura_Assert(pxPrevNode, "Failed to find previous node in internal linked list. Aborting half finished remove operation. Internal linked list and hash tree are now out of sync!" );

				if( pxPrevNode )
				{
					// set next of previous node to be one after the one we remove
					pxPrevNode->SetNext( pxNodeToRemove->GetNext() );
				}
				else
				{
					bSuccess = false;
				}

			}

			if( bSuccess )
			{
				#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
				Asura_DebugVariableSystem::DelCmd(pxNodeToRemove->GetName());
				#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM
				
				delete pxNodeToRemove;
			}
		}
	}

	return bSuccess;
}

bool Asura_CommandConsole_CmdList::Exec( const Asura_UnicodeChar* wszCmdName, const Asura_UnicodeChar* wszParameter, bool bHelpReq, bool* pbReturnValue )
{
    Asura_Assert( wszCmdName && *wszCmdName && wszParameter, "" );

    // Find corresponding command.
    const Node* pxCmdNode = Find( wszCmdName );
    if( !pxCmdNode )
    {
        return false;
    }
    Asura_Assert( pxCmdNode->GetCommand().pfnVoid, "" );

    // Can it be executed at the moment?
    bool bReturnValue;
	if( pxCmdNode->GetRestrictions() && !Asura_CommandConsole::Project_CheckRestrictions(pxCmdNode->GetRestrictions(), true) )
    {
        bReturnValue = false;
    }
    else
    {
        // Execute it.
        switch( pxCmdNode->GetType() )
        {
            case Asura_CommandConsole_CmdList::CMDTYPE_VOID:
                bReturnValue = pxCmdNode->GetCommand().pfnVoid( bHelpReq );
                break;

            case Asura_CommandConsole_CmdList::CMDTYPE_BOOL:
            {
                bool bParamValue = false;
                if( !*wszParameter || Asura_StringUtil::GetBoolValue(wszParameter, &bParamValue) )
                {
                    bReturnValue = pxCmdNode->GetCommand().pfnBool( bParamValue, !*wszParameter, bHelpReq );
                }
                else
                {
                    Asura_CommandConsole::Printf( L"%s: ERROR: Invalid parameter.", pxCmdNode->GetName() );
					bReturnValue = false;
                }
                break;
            }

            case Asura_CommandConsole_CmdList::CMDTYPE_INT:  // Intentional fallthrough.
            case Asura_CommandConsole_CmdList::CMDTYPE_UINT:
            {
				int iParamValue = 0;
                if( !*wszParameter || Asura_StringUtil::GetIntValue(wszParameter, &iParamValue) )
                {
                    if( Asura_CommandConsole_CmdList::CMDTYPE_UINT == pxCmdNode->GetType() )
                    {
                        if( *wszParameter && iParamValue < 0 )
                        {
                            Asura_CommandConsole::Printf( L"%s: ERROR: Parameter cannot be negative.", pxCmdNode->GetName() );
							bReturnValue = false;
                        }
                        else
                        {
                            bReturnValue = pxCmdNode->GetCommand().pfnUint( iParamValue, !*wszParameter, bHelpReq );
                        }
                    }
                    else
                    {
                        bReturnValue = pxCmdNode->GetCommand().pfnInt( iParamValue, !*wszParameter, bHelpReq );
                    }
                }
                else
                {
                    Asura_CommandConsole::Printf( L"%s: ERROR: Invalid parameter.", pxCmdNode->GetName() );
					bReturnValue = false;
                }
                break;
            }

            case Asura_CommandConsole_CmdList::CMDTYPE_FLOAT:
            {
                float fParamValue = 0.0f;
                if( !*wszParameter || Asura_StringUtil::GetFloatValue(wszParameter, &fParamValue) )
                {
                    bReturnValue = pxCmdNode->GetCommand().pfnFloat( fParamValue, !*wszParameter, bHelpReq );
                }
                else
                {
                    Asura_CommandConsole::Printf( L"%s: ERROR: Invalid parameter.", pxCmdNode->GetName() );
					bReturnValue = false;
                }
                break;
            }

            case Asura_CommandConsole_CmdList::CMDTYPE_STRING:
		    {
			    // need to convert parameter to non-unicode string
			    char szBuffer[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];
			    Asura_PrintfToBfr( szBuffer, "%S", ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE, true, wszParameter );
                bReturnValue = pxCmdNode->GetCommand().pfnString( szBuffer, !*wszParameter, bHelpReq );
                break;
		    }

            case Asura_CommandConsole_CmdList::CMDTYPE_WSTRING:
                bReturnValue = pxCmdNode->GetCommand().pfnWString( wszParameter, !*wszParameter, bHelpReq );
                break;

            default:
                Asura_Assert( false, "" );
                bReturnValue = false;
                break;
        }
    }

    if( pbReturnValue ) { *pbReturnValue = bReturnValue; }

    return true;
}

const Asura_CommandConsole_CmdList::Node* Asura_CommandConsole_CmdList::Find( const Asura_UnicodeChar* wszCmdName, bool bAcceptPrefixMatches )
{
	if (bAcceptPrefixMatches)
	{
		// Use string comparisions for the AcceptPrefixMatches case
		const Node* pxMatchingNode = NULL;

		if( wszCmdName && *wszCmdName )
		{
			u_int uLen = wcslen( wszCmdName );
			for( const Node* pxNext = s_pxRoot; pxNext; pxNext = pxNext->GetNext() )
			{
				int iDiff = ( bAcceptPrefixMatches ? wcsnicmp(wszCmdName, pxNext->GetName(), uLen) : wcsicmp(wszCmdName, pxNext->GetName() ) );
				if( iDiff <= 0 )
				{
					if( !iDiff ) { pxMatchingNode = pxNext; }
					break;
				}
			}
		}
		return pxMatchingNode;
	}
	else
	{
		// Find the node by hash
		if( wszCmdName && *wszCmdName )
		{
			char szCmdName[128];
			Asura_PrintfToBfr(szCmdName, "%S", 128, true, wszCmdName);
			Node** ppxMatchingNode = s_xNodeTree.GetDataByHash(Asura_GetHashID(szCmdName));
			if (ppxMatchingNode)
				return *ppxMatchingNode;
		}
		return NULL;
	}
}

Asura_CommandConsole_CmdList::Node::Node( const Asura_UnicodeChar* wszName,const CMDTYPE eType,const CmdFn xCmdFn )
:m_eType(eType)
,m_xCmdFn(xCmdFn)
,m_eRemove(Asura_CommandConsole::REMOVE_DEFAULT)
,m_uRestrictions(0)
,m_pxNext(NULL)
,m_wszName( NULL )
{
	if( wszName )
	{
		m_wszName = static_cast<Asura_UnicodeChar*> (Asura_Allocate( sizeof(Asura_UnicodeChar) *  (wcslen(wszName) + 1) ));
		wcscpy( m_wszName,wszName );
	}


	Asura_Assert( GetName(),"Node created with NULL name string" );
}

Asura_CommandConsole_CmdList::Node::~Node()
{
	Asura_Deallocate(m_wszName);
}


void Asura_CommandConsole_CmdList::Initialise( void )
{
    s_pxRoot = NULL;
}

void Asura_CommandConsole_CmdList::Reset( void )
{
	Node* pxNext( NULL );
	
	for( Node* pxCurrent( s_pxRoot );pxCurrent;pxCurrent = pxNext )
	{
		pxNext = pxCurrent->GetNext();

		if( Asura_CommandConsole::REMOVE_AT_RESET == pxCurrent->GetRemoveType() )
		{
			Remove(	pxCurrent );
		}
	}
}

void Asura_CommandConsole_CmdList::Shutdown( void )
{
	// Clear hash tree
	s_xNodeTree.Reset();

	// Deallocate list contents.
    for( Node* pxCur = s_pxRoot; pxCur; )
    {
        Node* pxNext = pxCur->GetNext();

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_DebugVariableSystem::DelCmd(pxCur->GetName() );
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

        delete pxCur;
        pxCur = pxNext;
    }
    s_pxRoot = NULL;
}

//
// Asura_CommandConsole_VarRepository.
//
Asura_CommandConsole_VarRepository::Var* Asura_CommandConsole_VarRepository::s_pxRoot;
Asura_TemplateHashTree<Asura_CommandConsole_VarRepository::Var*> Asura_CommandConsole_VarRepository::s_xVarTree;

// Constructor
Asura_CommandConsole_VarRepository::Var::Var(Asura_UnicodeChar* wszName, const char* szDesc, bool bReadOnly, bool* pbModified, TYPE eType, Asura_CommandConsole::REMOVE eRemove, Data xData)
: m_wszName(wszName)
, m_szDesc(szDesc)
, m_bReadOnly(bReadOnly)
, m_pbModified(pbModified)
, m_eType(eType)
, m_eRemove(eRemove)
, m_xData(xData)
{
	Var* pxPrev = NULL;
	Var* pxNext = s_pxRoot;

	// Find place to insert to keep list sorted alphabetically
	while( pxNext && wcsicmp(m_wszName, pxNext->m_wszName) > 0 )
	{
		pxPrev = pxNext;
		pxNext = pxNext->m_pxNext;
	}

	if( pxPrev )
	{
		pxPrev->m_pxNext = this;
	}
	else
	{
		s_pxRoot = this;
	}
	m_pxNext = pxNext;
}

// Destructor
Asura_CommandConsole_VarRepository::Var::~Var()
{
	delete[] m_wszName;
	delete[] m_szDesc;

	if (s_pxRoot == this)
	{
		// Set new root node
		s_pxRoot = m_pxNext;
	}
	else
	{
		// Find the previous node by iterating through the list from the root
		Var* pxPrevNode = s_pxRoot;
		while (pxPrevNode->m_pxNext != this)
		{
			pxPrevNode = pxPrevNode->m_pxNext;
		}

		// Set next of previous node to be one after the one we remove
		pxPrevNode->m_pxNext = m_pxNext;
	}
}

bool Asura_CommandConsole_VarRepository::Add( const Asura_UnicodeChar* wszVarName, 
											  Var::TYPE eType, 
											  const Var::Data& xData, 
											  Asura_CommandConsole::REMOVE eRemove,
											  bool bConsoleVarEntity /*= false*/)
{
    if( !wszVarName || !*wszVarName || !xData.uPtrVal )
    {
        Asura_Assert( false, "Invalid parameter(s) passed to AddVar()." );
        return false;
    }

    Asura_UnicodeChar* wszNameCopy;

    wszNameCopy = Asura_StringUtil::Duplicate( wszVarName );

    if( !wszNameCopy ) { return false; }

    // Verify that this variable doesn't exist already.
    if( Find(wszNameCopy) )
    {
		//TY 31/10/07: Avoid the cascade of dependancies associated with entities in UnitTests
		#ifndef ASURA_UNITTEST_BATCHMODE
		if (!bConsoleVarEntity)
		{
			// MGH 29/6/2007: We are trying to add a console variable in code, but it already exists.
			// This can happen if a ConsoleVar entity refers to a code defined variable, which isn't
			// set up until after the level has loaded. In that case transfer ownership to the code
			// version which takes priority.
			if (Asura_ClientEntity_ConsoleVar::TransferConsoleVarToCode(wszVarName, eType, xData, eRemove))
			{
				// Successfully transfered.
				delete[] wszNameCopy;
				return true;
			}
		}
		#endif

        Asura_Assertf( false, ("Variable \"%S\" already exists.", wszNameCopy) );
        delete[] wszNameCopy;
        return false;
    }
	if ( Asura_CommandConsole_CmdList::Find(wszNameCopy) )
	{
        Asura_Assertf( false, ("Variable name %S is already used for a command.", wszNameCopy) );
        delete[] wszNameCopy;
        return false;
	}
	// Allocate, insert and setup new list node.
	Var* pxNewNode = new Var( wszNameCopy,
							  NULL,
							  false,
							  NULL,
							  eType,
							  eRemove,
							  xData );
	if( !pxNewNode )
	{
		return false;
	}

	// Insert into hash tree
	char szVarName[128];
	Asura_PrintfToBfr(szVarName, "%S", 128, true, wszNameCopy);
	*s_xVarTree.AddNode(Asura_GetHashID(szVarName)) = pxNewNode;

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::AddVar(wszNameCopy, pxNewNode);
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

    return true;
}

void Asura_CommandConsole_VarRepository::UnlinkVar(Var& xVarToUnlink)
{
	const Asura_UnicodeChar*	wszVarName	= xVarToUnlink.m_wszName;

	char						szVarName[128];

	// Remove from hash tree
	Asura_PrintfToBfr(szVarName, "%S", 128, true, wszVarName);
	s_xVarTree.DeleteNodeByHash(Asura_GetHashID(szVarName));

	// Remove from list
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::DelVar(wszVarName);
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM
}

bool Asura_CommandConsole_VarRepository::Remove( const Asura_UnicodeChar* wszVarName, const bool bRemoveChildren )
{
	if( !wszVarName || !*wszVarName )
	{
		Asura_Assert( false, "Invalid parameter(s) passed to Remove()." );
		return false;
	}

	if(bRemoveChildren)
	{
		const int	iSize			= wcslen(wszVarName);
		Var*		pxNodeToRemove	= s_pxRoot;
		int			iNumRemoved		= 0;

		// find the node for this command.
		while(pxNodeToRemove != NULL)
		{
			Var*	pxNextNode	= pxNodeToRemove->m_pxNext;

			if(wcsnicmp(pxNodeToRemove->m_wszName, wszVarName, iSize) == 0)
			{
				UnlinkVar(*pxNodeToRemove);
				delete(pxNodeToRemove);
				iNumRemoved++;
			}

			pxNodeToRemove	= pxNextNode;
		}

		// Did we remove any nodes?
		return(iNumRemoved > 0);
	}
	else
	{
		Var* pxNodeToRemove = s_pxRoot;

		// find the node for this command.
		while (pxNodeToRemove && wcsicmp(pxNodeToRemove->m_wszName, wszVarName))
		{
			pxNodeToRemove = pxNodeToRemove->m_pxNext;
		}

		if(pxNodeToRemove == NULL)
		{
			Asura_Assertf(false, ("Trying to remove console var '%S', but it doesn't exist.", wszVarName));
			return(false);
		}
		else
		{
			// Remove the node...
			UnlinkVar(*pxNodeToRemove);
			delete(pxNodeToRemove);
			return(true);
		}
	}
}

bool Asura_CommandConsole_VarRepository::SetHelpText( const Asura_UnicodeChar* wszVarName, const char* szVarDesc )
{
	Asura_CommandConsole_VarRepository::Var* pxVar = Find( wszVarName );
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (!pxVar)
	{
		Asura_UnicodeChar awcBuffer[128];
		Asura_PrintfToBfr( awcBuffer, L"Unclassified.%s", 128, true, wszVarName );
		pxVar = Find(awcBuffer);
	}
#endif
	Asura_Assertf(pxVar, ("var %S not found", wszVarName));
	if (pxVar)
	{
		Asura_Assertf(!pxVar->m_szDesc, ("help text already set for var %S", wszVarName));
		if (!pxVar->m_szDesc)
		{
			pxVar->m_szDesc = Asura_StringUtil::Duplicate( szVarDesc );
			return true;
		}
	}
	return false;
}

bool Asura_CommandConsole_VarRepository::MakeVarReadOnly( const Asura_UnicodeChar* wszVarName )
{
	Asura_CommandConsole_VarRepository::Var* pxVar = Find( wszVarName );
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (!pxVar)
	{
		Asura_UnicodeChar awcBuffer[128];
		Asura_PrintfToBfr( awcBuffer, L"Unclassified.%s", 128, true, wszVarName );
		pxVar = Find(awcBuffer);
	}
#endif
	Asura_Assertf(pxVar, ("var %S not found", wszVarName));
	if (pxVar)
	{
		Asura_Assert(!pxVar->m_pbModified, "shouldn't have modified flag on read-only var");
		pxVar->m_bReadOnly = true;
		return true;
	}
	return false;
}

bool Asura_CommandConsole_VarRepository::SetPointerToModifiedFlag( const Asura_UnicodeChar* wszVarName, bool* pbModified )
{
	Asura_CommandConsole_VarRepository::Var* pxVar = Find( wszVarName );
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (!pxVar)
	{
		Asura_UnicodeChar awcBuffer[128];
		Asura_PrintfToBfr( awcBuffer, L"Unclassified.%s", 128, true, wszVarName );
		pxVar = Find(awcBuffer);
	}
#endif
	Asura_Assertf(pxVar, ("var %S not found", wszVarName));
	if (pxVar)
	{
		Asura_Assert(pbModified, "NULL pointer to modified flag");
		Asura_Assert(!pxVar->m_bReadOnly, "shouldn't have modified flag on read-only var");
		pxVar->m_pbModified = pbModified;
		return true;
	}
	return false;
}

bool Asura_CommandConsole_VarRepository::IsVarReadOnly( const Asura_UnicodeChar* wszVarName )
{
	const Asura_CommandConsole_VarRepository::Var* pxVar = Find( wszVarName );
	if (pxVar)
	{
		return pxVar->m_bReadOnly;
	}
	return false;// return false if not found - assumed by calling code.
}

Asura_CommandConsole_VarRepository::Var* Asura_CommandConsole_VarRepository::Find( const Asura_UnicodeChar* wszVarName )
{
	// Find the var by hash
	if( wszVarName && *wszVarName )
	{
		char szVarName[128];
		Asura_PrintfToBfr(szVarName, "%S", 128, true, wszVarName);
		Var** ppxMatchingVar = s_xVarTree.GetDataByHash(Asura_GetHashID(szVarName));
		if (ppxMatchingVar)
			return *ppxMatchingVar;
	}
	return NULL;	
}

Asura_CommandConsole_VarRepository::Var* Asura_CommandConsole_VarRepository::Find( const char* szVarName )
{
	Asura_UnicodeChar awcBuffer[128];
	Asura_PrintfToBfr( awcBuffer, L"%S", 128, true, szVarName );
	return Find( awcBuffer );
}

Asura_CommandConsole_VarRepository::Var* Asura_CommandConsole_VarRepository::Find( const Asura_Hash_ID uVarNameHash )
{
	if (uVarNameHash != ASURA_HASH_ID_UNSET)
	{
		Var** ppxMatchingVar = s_xVarTree.GetDataByHash(uVarNameHash);
		if (ppxMatchingVar)
		{
			return *ppxMatchingVar;
		}
	}

	return NULL;
}

void Asura_CommandConsole_VarRepository::Initialise( void )
{
	s_pxRoot = NULL;
}

// functions to return value of named variable. They assert and return 
// a default value if var not found or if it's the wrong type.
bool Asura_CommandConsole_VarRepository::GetBoolValue( const char* szVarName )
{
	//u_int uHashName = Asura_GetHashID( szVarName);
	//return GetBoolValue( uHashName);

    const Var* pxVar = Find( szVarName );
    if( !pxVar ) { return false; }

    Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_BOOL, ("var %s is not bool type", szVarName));
    if ( pxVar->m_eType == Var::TYPE_BOOL )
	{
		return *(pxVar->m_xData.pbData);
	}
	return false;
}

bool Asura_CommandConsole_VarRepository::GetBoolValue( const Asura_Hash_ID uHash )
{
	const Var* pxVar = Find( uHash );
	if( !pxVar ) { return false; }

	Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_BOOL, ("var %s is not bool type", pxVar->m_wszName));
	if ( pxVar->m_eType == Var::TYPE_BOOL )
	{
		return *(pxVar->m_xData.pbData);
	}
	return false;
}

int Asura_CommandConsole_VarRepository::GetIntValue( const char* szVarName )
{
	//u_int uHashName = Asura_GetHashID( szVarName);
	//return GetIntValue( uHashName);

    const Var* pxVar = Find( szVarName );
    if( !pxVar ) { return 0; }

    Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_INT, ("var %s is not int type", szVarName));
    if ( pxVar->m_eType == Var::TYPE_INT )
	{
		return *(pxVar->m_xData.piData);
	}
	return 0;
}

int Asura_CommandConsole_VarRepository::GetIntValue( const Asura_UnicodeChar* wszVarName )
{
	//if( wszVarName && *wszVarName )
	//{
	//	char szVarName[128];
	//	Asura_PrintfToBfr(szVarName, "%S", 128, true, wszVarName);
	//	u_int uHashName = Asura_GetHashID( szVarName);

	//	return GetIntValue( uHashName);
	//}
	//
	//return 0;
	const Var* pxVar = Find( wszVarName );
	if( !pxVar ) { return 0; }

	Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_INT, ("var %ls is not int type", wszVarName));
	if ( pxVar->m_eType == Var::TYPE_INT )
	{
		return *(pxVar->m_xData.piData);
	}
	return 0;
}

int Asura_CommandConsole_VarRepository::GetIntValue( const Asura_Hash_ID uHash )
{
	const Var* pxVar = Find( uHash );
	if( !pxVar ) { return 0; }

	Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_INT, ("var %s is not int type", pxVar->m_wszName));
	if ( pxVar->m_eType == Var::TYPE_INT )
	{
		return *(pxVar->m_xData.piData);
	}
	return 0;
}

float Asura_CommandConsole_VarRepository::GetFloatValue( const char* szVarName )
{
	//u_int uHashName = Asura_GetHashID( szVarName);
	//return GetFloatValue( uHashName);

    const Var* pxVar = Find( szVarName );
    if( !pxVar ) { return 0.0f; }

    Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_FLOAT, ("var %s is not float type", szVarName));
    if ( pxVar->m_eType == Var::TYPE_FLOAT)
	{
		return *(pxVar->m_xData.pfData);
	}
	return 0.0f;
}

float Asura_CommandConsole_VarRepository::GetFloatValue( const Asura_Hash_ID uHash )
{
	const Var* pxVar = Find( uHash );
	if( !pxVar ) { return 0.0f; }

	Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_FLOAT, ("var %s is not float type", pxVar->m_wszName));
	if ( pxVar->m_eType == Var::TYPE_FLOAT)
	{
		return *(pxVar->m_xData.pfData);
	}
	return 0.0f;
}

const Asura_UnicodeChar* Asura_CommandConsole_VarRepository::GetUnicodeStringValue( const char* szVarName )
{
	//u_int uHashName = Asura_GetHashID( szVarName);
	//return GetUnicodeStringValue( uHashName);
    const Var* pxVar = Find( szVarName );
    if( !pxVar ) { return NULL; }

    Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_STRING, ("var %s is not string type", szVarName));
    if ( pxVar->m_eType == Var::TYPE_STRING )
	{
		return pxVar->m_xData.wszData;
	}
	return NULL;
}

const Asura_UnicodeChar* Asura_CommandConsole_VarRepository::GetUnicodeStringValue( const Asura_Hash_ID uHash )
{
	const Var* pxVar = Find( uHash );
	if( !pxVar ) { return NULL; }

	Asura_Assert( pxVar->m_xData.uPtrVal, "var not set???" );
	Asura_Assertf( pxVar->m_eType == Var::TYPE_STRING, ("var %s is not string type", pxVar->m_wszName));
	if ( pxVar->m_eType == Var::TYPE_STRING )
	{
		return pxVar->m_xData.wszData;
	}
	return NULL;
}

bool Asura_CommandConsole_VarRepository::SetBoolValue( const char* szVarName, bool bValue )
{
	Asura_UnicodeChar awcBuffer[16];
	Asura_PrintfToBfr( awcBuffer, L"%d", 16, true, bValue );
	bool bSetOK = false;
	SetValue(szVarName, awcBuffer, &bSetOK);
	return bSetOK;
}

bool Asura_CommandConsole_VarRepository::SetIntValue( const char* szVarName, int iValue )
{
	Asura_UnicodeChar awcBuffer[16];
	Asura_PrintfToBfr( awcBuffer, L"%d", 16, true, iValue );
	bool bSetOK = false;
	SetValue(szVarName, awcBuffer, &bSetOK);
	return bSetOK;
}

bool Asura_CommandConsole_VarRepository::SetFloatValue( const char* szVarName, float fValue )
{
	Asura_UnicodeChar awcBuffer[16];
	Asura_PrintfToBfr( awcBuffer, L"%f", 16, true, Asura_Maths::FtoD(fValue) );
	bool bSetOK = false;
	SetValue(szVarName, awcBuffer, &bSetOK);
	return bSetOK;
}

bool Asura_CommandConsole_VarRepository::SetUnicodeStringValue( const char* szVarName, const Asura_UnicodeChar* wszValue )
{
	bool bSetOK = false;
	SetValue(szVarName, wszValue, &bSetOK);
	return bSetOK;
}

// report value prints the value to the command console
bool Asura_CommandConsole_VarRepository::ReportValue( const Asura_UnicodeChar* wszVarName )
{
    const Var* pxVar = Find( wszVarName );
    if( !pxVar ) { return false; }

    Asura_Assert( pxVar->m_xData.uPtrVal, "" );
    switch( pxVar->m_eType )
    {
		case Var::TYPE_BOOL:
			Asura_CommandConsole::Printf( L"%s: %s", pxVar->m_wszName, (*(pxVar->m_xData.pbData) ? L"true" : L"false") );
            break;

        case Var::TYPE_FLOAT:
            Asura_CommandConsole::Printf( L"%s: %.6f", pxVar->m_wszName, Asura_Maths::FtoD(*(pxVar->m_xData.pfData)) );
            break;

        case Var::TYPE_INT:
            Asura_CommandConsole::Printf( L"%s: %i", pxVar->m_wszName, *(pxVar->m_xData.piData) );
            break;

        case Var::TYPE_STRING:
            Asura_CommandConsole::Printf( L"%s: %s", pxVar->m_wszName, pxVar->m_xData.wszData );
            break;

		case Var::TYPE_INVALID:
			break;
	}

    return true;
}

bool Asura_CommandConsole_VarRepository::ReportRange( const Asura_UnicodeChar* wszVarName )
{
    const Var* pxVar = Find( wszVarName );
    if( !pxVar ) { return false; }

    Asura_Assert( pxVar->m_xData.uPtrVal, "" );
    switch( pxVar->m_eType )
    {
		case Var::TYPE_BOOL:
			Asura_CommandConsole::Printf( L"Allowed range of %s: true or false (can also use 0 or 1)", pxVar->m_wszName );
            break;

        case Var::TYPE_FLOAT:
			Asura_CommandConsole::Printf( L"Allowed range of %s: [%.1f,%.1f]", pxVar->m_wszName,
				Asura_Maths::FtoD(pxVar->m_xData.fLowerLimit), Asura_Maths::FtoD(pxVar->m_xData.fUpperLimit) );
            break;

        case Var::TYPE_INT:
			Asura_CommandConsole::Printf( L"Allowed range of %s: [%i,%i]", pxVar->m_wszName,
				pxVar->m_xData.iLowerLimit, pxVar->m_xData.iUpperLimit );
            break;

        case Var::TYPE_STRING:
            Asura_CommandConsole::Printf( L"Allowed length of %s: %d characters", pxVar->m_wszName, pxVar->m_xData.uBfrSize - 1);
            break;

			// SK 08.03.07 - added the default option so the ps2 compiler stops complaining
			// compiler message was :- warning: enumeration value 'TYPE_INVALID' not handled in switch. 
			// I'm a bit reluctant to return false because i don't know what impact it could have on projects.
		default:
			break;
	}

    return true;
}

void Asura_CommandConsole_VarRepository::Reset( void )
{
	for( Var* pxCur = s_pxRoot; pxCur; )
	{
		Var* pxNext = pxCur->m_pxNext;

		if( Asura_CommandConsole::REMOVE_AT_RESET == pxCur->m_eRemove )
		{
			// Remove from hash tree
			char szVarName[128];
			Asura_PrintfToBfr(szVarName, "%S", 128, true, pxCur->m_wszName);
			s_xVarTree.DeleteNodeByHash(Asura_GetHashID(szVarName));

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
			Asura_DebugVariableSystem::DelVar(pxCur->m_wszName);
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

			// Deallocate node and remove from list
			delete pxCur;
		}

		pxCur = pxNext;
	}
}

// set value sets the value from the command console input.
bool Asura_CommandConsole_VarRepository::SetValue( const Asura_UnicodeChar* wszVarName, const Asura_UnicodeChar* wszValue, bool* pbSetOK )
{
    Var* pxVar = Find( wszVarName );
    if( !pxVar ) { return false; }
	if (pxVar->m_bReadOnly )
	{
		Asura_Assertf(false, ("can't set value of %S as it's a read-only var", wszValue ));
		return false;
	}

    Var::Data&	xData  = pxVar->m_xData;
    bool		bError = false;
    bool		bOutOfRange = false;
	bool		bModified = false;

    if( xData.uRestrictions )
    {
		bError = !Asura_CommandConsole::Project_CheckRestrictions( xData.uRestrictions, false );
    }

    if( !bError )
    {
        const Asura_UnicodeChar* wszTypeDesc = NULL;
        const Asura_UnicodeChar* wszExactErr = L"invalid";
    
        switch( pxVar->m_eType )
        {
            case Var::TYPE_BOOL:
            {
                wszTypeDesc = L"boolean";
				bool bValue;
                if( !Asura_StringUtil::GetBoolValue(wszValue, &bValue) )
                {
                    bError = true;
                }
				else
				{
					// set value.
					if (*pxVar->m_xData.pbData != bValue)
					{
						bModified = true;
						*pxVar->m_xData.pbData = bValue;
					}
				}
                break;
            }

            case Var::TYPE_FLOAT:
            {
                bError      = true;
                wszTypeDesc = L"floating point";

                float fValue;
                if( Asura_StringUtil::GetFloatValue(wszValue, &fValue) )
                {
                    if( xData.fLowerLimit <= fValue && fValue <= xData.fUpperLimit )
                    {
                        if (*(xData.pfData) != fValue)
						{
							*(xData.pfData) = fValue;
							bModified		= true;
						}
						bError = false;
                    }
                    else
                    {
                        wszExactErr = L"outside permitted range";
						bOutOfRange		= true;
                    }
                }
                break;
            }

            case Var::TYPE_INT:
            {
                bError      = true;
                wszTypeDesc = L"integer";

                int iValue;
                if( Asura_StringUtil::GetIntValue(wszValue, &iValue) )
                {
                    if( xData.iLowerLimit <= iValue && iValue <= xData.iUpperLimit )
                    {
						if (*(xData.piData) != iValue)
						{
							*(xData.piData) = iValue;
							bModified		= true;
						}
                        bError          = false;
                    }
                    else
                    {
                        wszExactErr = L"outside permitted range";
						bOutOfRange		= true;
                    }
                }
                break;
            }

            case Var::TYPE_STRING:
            {
                u_int uLen = wcslen( wszValue );

                // If the string is enclosed in quotes, ignore them. This allows
                // the user to include leading/trailing whitespace or to set
                // variables to the empty string.
                wszValue = Asura_StringUtil::StripQuotes( wszValue, &uLen );

                // (MH - 08.02.2007) changed to truncate if too long (and then report error).
				// This is more likely to be what user expects (e.g. setting player name, server name etc.)
                if( uLen + 1 > xData.uBfrSize )
                {
					uLen = xData.uBfrSize - 1;
                    bError      = true;
                    wszTypeDesc = L"string";
                    wszExactErr = L"too long";
					bOutOfRange		= true;
                }
				bModified = wcsncmp(xData.wszData, wszValue, uLen) != 0;
				wcsncpy( xData.wszData, wszValue, uLen );
                xData.wszData[uLen] = L'\0';
                break;
            }

			case Var::TYPE_INVALID:
			{
				wszTypeDesc = L"Var::Invalid";
				wszExactErr = L"Trying to set data to an invalid Var Type.";
				bError = true;
				break;
			}
        }

        if( bError )
        {
            if (pxVar->m_eType == Var::TYPE_STRING)
			{
				Asura_CommandConsole::Printf( L"Truncated setting of %s variable '%s' because value specified was %s.", wszTypeDesc, wszVarName, wszExactErr );
			}
			else
			{
				Asura_CommandConsole::Printf( L"Could not set %s variable '%s' because value specified was %s.", wszTypeDesc, wszVarName, wszExactErr );
			}

			if (bOutOfRange)
			{
				ReportRange(wszVarName);
			}
        }
		
		// set modified bool
		if (bModified && pxVar->m_pbModified)
		{
			*pxVar->m_pbModified = true;
		}
    }

    if( pbSetOK ) { *pbSetOK = !bError; }
    return true;
}

bool Asura_CommandConsole_VarRepository::SetValue( const char* szVarName, const Asura_UnicodeChar* wszValue, bool* pbSetOK )
{
	Asura_UnicodeChar awcBuffer[128];
	Asura_PrintfToBfr( awcBuffer, L"%S", 128, true, szVarName );
	return SetValue( awcBuffer, wszValue, pbSetOK );
}

void Asura_CommandConsole_VarRepository::Shutdown( void )
{
	// Deallocate hash tree contents
	s_xVarTree.Reset();

	// Deallocate entire list contents.
	while( s_pxRoot )
	{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_DebugVariableSystem::DelVar(s_pxRoot->m_wszName);
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

		delete s_pxRoot;
	}
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

// Returns the next command or variable alphabeticaly after the last one we found via 
// tabcompletion and that starts with what is in the inputbuffer, or simply the first 
// one that fits with the inputbuffer if we havent tabbed before, returns NULL if no 
// matches are found.
// bForceFirstcompletion should be set if this is the second time looking
// bHadAnyMatches will be set to true if it finds any matching commands or variables 
// but they where alphabeticaly before the last command that was found.
// bUsingTabCompletionBuffer will be set to true if we have found matches before.
static const Asura_UnicodeChar* GetNextCommandOrVariable(const bool bForceFirstCompletion, bool& bHadAnyMatches, bool& bUsingTabCompletionBuffer, u_int& uRootLen )
{
	const Asura_UnicodeChar* szRoot;

    // Calculate length of input buffer, excluding whitespace.
    u_int uInputBufferLenExclWspace = wcslen( g_wszConsoleInputBuffer );
    for( ; uInputBufferLenExclWspace; --uInputBufferLenExclWspace )
    {
        if( !iswspace(g_wszConsoleInputBuffer[uInputBufferLenExclWspace - 1]) )
        {
            break;
        }
    }

    // Find the 'root', it will serve as the starting point in looking for the next
	// command or variable, it should always be the last thing we wrote, rather than
	// what has been found by the autocomplete.
    uRootLen = wcslen( g_wszTabCompletionBuffer );
    if( uRootLen )
    {
        szRoot = g_wszTabCompletionBuffer;
        bUsingTabCompletionBuffer = true;
    }
    else
    {
        uRootLen = uInputBufferLenExclWspace;
        if( uRootLen )
        {
            szRoot = g_wszConsoleInputBuffer;
            bUsingTabCompletionBuffer = false;
        }
        else
        {
            return NULL;
        }
    }

	const Asura_UnicodeChar* wszCmdCompletion = NULL;
	const Asura_UnicodeChar* wszVarCompletion = NULL;
    const Asura_UnicodeChar* wszCompletion = NULL;

    // First find the next command after the last one we found. (after the 'root')
    for( const Asura_CommandConsole_CmdList::Node* pxNext = Asura_CommandConsole_CmdList::GetRootNode(); !wszCmdCompletion && pxNext; pxNext = pxNext->GetNext() )
    {
		// Just ignore it if it isnt in our root (what we wrote before pressing tab)
        if( szRoot == NULL || !wcsnicmp( pxNext->GetName(), szRoot, uRootLen ) )
        {
			// Select this command as our next one if its name is after the current 
			// inputbuffer alphabeticaly, or if we are forcing completion, or if this is 
			// the first time we tab.
			if ( wcsnicmp( pxNext->GetName(), g_wszConsoleInputBuffer, uInputBufferLenExclWspace ) > 0 || bForceFirstCompletion || !bUsingTabCompletionBuffer )
			{
				wszCmdCompletion = pxNext->GetName();
			}
			bHadAnyMatches = true;
        }
    }

    // Then the next variable after the 'root'
	for( const Asura_CommandConsole_VarRepository::Var* pxNext = Asura_CommandConsole_VarRepository::GetRootNode(); !wszVarCompletion && pxNext; pxNext = pxNext->m_pxNext )
	{
		// Just ignore it if it isnt in our root (what we wrote before pressing tab)
		if( szRoot == NULL || !wcsnicmp(pxNext->m_wszName, szRoot, uRootLen) )
		{
			// Select this variable as our next one if its name is after the current 
			// inputbuffer alphabeticaly, or if we are forcing completion, or if this 
			// is the first time we tab.
			if ( wcsnicmp( pxNext->m_wszName, g_wszConsoleInputBuffer, uInputBufferLenExclWspace ) > 0 || bForceFirstCompletion || !bUsingTabCompletionBuffer )
			{
				wszVarCompletion = pxNext->m_wszName;
			}
			bHadAnyMatches = true;
		}
	}

	//If we only found a variable select that.
	if ( wszCmdCompletion == NULL && wszVarCompletion != NULL )
	{
		wszCompletion = wszVarCompletion;
	}

	//If we only found a command select that.
	if ( wszVarCompletion == NULL && wszCmdCompletion != NULL )
	{
		wszCompletion = wszCmdCompletion;
	}

	// If we found both, select variable or command based on which has the lowest alphabetical value.
	if ( wszVarCompletion != NULL && wszCmdCompletion != NULL )
	{
		if ( wcsncmp( wszVarCompletion, wszCmdCompletion, ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE ) < 0 )
		{
			wszCompletion = wszVarCompletion;
		}
		else
		{
			wszCompletion = wszCmdCompletion;
		}
	}

	return wszCompletion;
}

static void AttemptTabCompletion( bool bForceFirstCompletion /*= false*/ )
{
	// Is there already something in the tab completion buffer? (Have we clicked tab before)
    bool bUsingTabCompletionBuffer = false; 

	// Set to true if we've found a viable completion but haven't reached the previous one yet.
    bool bHadAnyMatches = false; 

	// How long is our root? (The original string that we wrote)
	u_int uRootLen; 

	const Asura_UnicodeChar* wszCompletion = GetNextCommandOrVariable( bForceFirstCompletion, bHadAnyMatches, bUsingTabCompletionBuffer, uRootLen );

    if( wszCompletion )
    {
        if( !bUsingTabCompletionBuffer )
        {
            wcscpy( g_wszTabCompletionBuffer, g_wszConsoleInputBuffer );
            g_wszTabCompletionBuffer[uRootLen] = '\0';
        }

		// Since we only want to add one category at a time, find the first dot after our root 
		// and only write up to it into the input buffer.
		u_int uIndexOfLastDot = uRootLen;
		while ( wszCompletion[uIndexOfLastDot] && wszCompletion[uIndexOfLastDot] != '.' )
		{
			uIndexOfLastDot++;
			if ( !wszCompletion[uIndexOfLastDot] )
			{ 
				// Since we got the the end without finding another . this must be a leaf, so let 
				// it think the dot is one more character away so we can add a whitespace there.
				uIndexOfLastDot++;
				break;
			}
		}

        bool bMoveCursorToEnd = ( g_uConsoleInputCursorPosition == wcslen(g_wszConsoleInputBuffer) );

        // Update input buffer. Note that we append a space as well for ease
        // of use, ie. so the user can type <prefix>,Tab,[<parameter>,]Enter
        // and not have to bother inserting the space.
		const u_int uMaxLength = Asura_Maths::Min( uIndexOfLastDot + 1, static_cast<u_int>( ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE ) );
        u_int uNewLen = Asura_PrintfToBfr( g_wszConsoleInputBuffer, L"%s ", uMaxLength, true, wszCompletion );

        if( bMoveCursorToEnd )
        {
            g_uConsoleInputCursorPosition = uNewLen;
        }
    }
    else
    {
        if( bUsingTabCompletionBuffer && bHadAnyMatches )
        {
            // The previous completion was the last one so we need to loop
            // back round to the first.
            AttemptTabCompletion( true );
            return;
        }

        // (MP - 27.11.2002) TODO: Beep at the user or something?
    }
}
#ifndef ASURA_COMMAND_CONSOLE_READONLY
static void ClearConsole( void )
{
    if( g_pxIO->CanArchiveOutput() )
    {
        g_pxIO->ClearOutput();
    }
    else
    {
        g_uConsoleNumLinesUsed = 0;
		g_pxIO->SetFirstLineToDisplay(0);
    }
}

static bool ConsoleCmd_Clear( bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_CLEAR );
        Asura_CommandConsole::Printf( L"    Clears the console." );
    }
    else
    {
        ClearConsole();
    }
    return true;
}

static bool ConsoleCmd_Help( const Asura_UnicodeChar* wszCmd, bool bInterrogate, bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_HELP );
        Asura_CommandConsole::Printf( L"    Describes how to use the console." );
    }
    else if( bInterrogate )
    {
        Asura_CommandConsole::Printf( L"How to use the console:" );
        Asura_CommandConsole::Printf( L"" );
        Asura_CommandConsole::Printf( L"    Type 'listcmds' for a list of all available commands." );
        Asura_CommandConsole::Printf( L"    Type 'listcmds <prefix>' for a list of commands beginning with <prefix>." );
        Asura_CommandConsole::Printf( L"    Type 'help <command>' or '? <command>' for help with a specific command." );
        Asura_CommandConsole::Printf( L"    Type '<command> [<parameter>]' to execute a command." );
        Asura_CommandConsole::Printf( L"" );
        Asura_CommandConsole::Printf( L"    Type 'listvars' for a list of all available variables." );
        Asura_CommandConsole::Printf( L"    Type 'listvars <prefix>' for a list of variables beginning with <prefix>." );
        Asura_CommandConsole::Printf( L"    Type '<var name>' to view the value of a variable." );
        Asura_CommandConsole::Printf( L"    Type '<var name> <value>' to set the value of a variable." );
    }
    else
    {
        if( !Asura_CommandConsole_CmdList::Exec(wszCmd, L"", true, NULL) )
        {
			// try vars
			Asura_CommandConsole_VarRepository::Var* pxVar = Asura_CommandConsole_VarRepository::Find( wszCmd );
			if (pxVar)
			{
				if (pxVar->m_szDesc)
				{
					Asura_CommandConsole::Printf( pxVar->m_szDesc );
				}
				else
				{
					Asura_CommandConsole::Printf( "no help available for var %S", wszCmd );
				}
				if (pxVar->m_bReadOnly)
				{
					Asura_CommandConsole::Printf( "var is read-only" );
				}
				else
				{
					Asura_CommandConsole_VarRepository::ReportRange( wszCmd );
				}
			}
			else
			{
				Asura_CommandConsole::Printf( L"Command '%s' not found.", wszCmd );
				return false;
			}
        }
    }
    return true;
}

static bool ConsoleCmd_ListCmds( const Asura_UnicodeChar* wszPrefix, bool bInterrogate, bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_LISTCMDS L" [<prefix>]" );
        Asura_CommandConsole::Printf( L"    Lists all available console commands. If the prefix" );
        Asura_CommandConsole::Printf( L"    parameter is specified then only commands beginning" );
        Asura_CommandConsole::Printf( L"    with that string are listed." );
    }
    else
    {
        const Asura_CommandConsole_CmdList::Node* pxNext = Asura_CommandConsole_CmdList::GetRootNode();
        if( pxNext )
        {          
            const u_int uPrefixLen = wcslen( wszPrefix );

            bool bFoundAny = false;

            Asura_CommandConsole::Printf( L"Available console commands:" );
            for( ; pxNext; pxNext = pxNext->GetNext() )
            {
                // Skip if a prefix is specified and this command's name
                // doesn't match it.
                if( uPrefixLen && wcsnicmp(pxNext->GetName(), wszPrefix, uPrefixLen) )
                {
                    continue;
                }

                Asura_CommandConsole::Printf( L"    %s", pxNext->GetName() );
                bFoundAny = true;
            }

            if( !bFoundAny )
            {
                Asura_Assert( uPrefixLen, "" );
                Asura_CommandConsole::Printf( L"    [none found matching specified prefix]" );
            }
        }
        else
        {
            Asura_CommandConsole::Printf( L"No commands available." );
        }
    }
    return true;
}

static bool ConsoleCmd_ListVars( const Asura_UnicodeChar* wszPrefix, bool bInterrogate, bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_LISTVARS L" [<prefix>]" );
        Asura_CommandConsole::Printf( L"    Lists all available variables. If the prefix" );
        Asura_CommandConsole::Printf( L"    parameter is specified then only variables beginning" );
        Asura_CommandConsole::Printf( L"    with that string are listed." );
    }
    else
    {
		const Asura_CommandConsole_VarRepository::Var* pxNext = Asura_CommandConsole_VarRepository::GetRootNode();
		if( pxNext )
		{          
			const u_int uPrefixLen = wcslen( wszPrefix );

			bool bFoundAny = false;

			Asura_CommandConsole::Printf( L"Available variables:" );
			for( ; pxNext; pxNext = pxNext->m_pxNext )
			{
				// Skip if a prefix is specified and this variables's name
				// doesn't match it.
				if( uPrefixLen && wcsnicmp(pxNext->m_wszName, wszPrefix, uPrefixLen) )
				{
					continue;
				}

				Asura_CommandConsole::Printf( L"    %s", pxNext->m_wszName );
				bFoundAny = true;
			}

			if( !bFoundAny )
			{
				Asura_Assert( uPrefixLen, "" );
				Asura_CommandConsole::Printf( L"    [none found matching specified prefix]" );
			}
		}
		else
		{
			Asura_CommandConsole::Printf( L"No variables available." );
		}
    }
    return true;
}

static bool ConsoleCmd_FontScale( float fSize, bool bInterrogate, bool bHelp )
{
	const float fMINFONTSCALE = 0.2999999f;
	const float fMAXFONTSCALE = 2.0f;

	if (bHelp)
	{
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_FONTSCALE L" <float size>" );
        Asura_CommandConsole::Printf( L"    changes the console font scale. Must be between %.1f and %.1f.", fMINFONTSCALE, fMAXFONTSCALE );
    }
	else if (bInterrogate)
	{
        Asura_CommandConsole::Printf( L"    current font scale is %.1f.", Asura_Maths::FtoD(Asura_CommandConsole::GetFontScale()) );
	}
	else
	{
		if (fSize >= fMINFONTSCALE && fSize <= fMAXFONTSCALE)
		{
			Asura_CommandConsole::SetFontScale( fSize );
		}
		else
		{
	        Asura_CommandConsole::Printf( L"value must be between %.1f and %.1f.", fMINFONTSCALE, fMAXFONTSCALE );
			return false;
		}
	}
	return true;
}

static bool ConsoleCmd_Timescale( float fTimeScale, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_TIMESCALE L": sets game time scale" );
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf( L"Timescale = %.2f", Asura_Maths::FtoD(Asura_Timers::GetGameTimeScale()) );
	}
	else
	{
		if (fTimeScale < 0.0f) fTimeScale = 0.0f;
		if (fTimeScale > 3.0f) fTimeScale = 3.0f;

		Asura_Timers::SetGameTimeScale( fTimeScale );
		Asura_CommandConsole::Printf( L"Timescale set to %.2f", Asura_Maths::FtoD(fTimeScale) );
	}
	return true;
}

#if ASURA_NETWORK_LOGGING && defined(ASURA_SUPPORT_EXTERNAL_NETWORK)
static bool ConsoleCmd_SetMsgLogLevel( u_int uLevel, bool bInterrogate, bool bHelp )
{
    if( !bHelp && !bInterrogate && uLevel < Asura_Network::MSG_LOG_LEVEL_NUM_LEVELS )
    {
        Asura_Network::SetMsgLogLevel( static_cast<Asura_Network::MSG_LOG_LEVEL>(uLevel), true, true );
        return true;
    }
    return false;
}
#endif // #if ASURA_NETWORK_LOGGING && defined(ASURA_SUPPORT_EXTERNAL_NETWORK)

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_RUN
static bool ConsoleCmd_Run( const Asura_UnicodeChar* wszFileName, bool bInterrogate, bool bHelp )
{
    if ( bHelp || bInterrogate )
    {
        Asura_CommandConsole::Printf( CONSOLE_CMDNAME_RUN );
        Asura_CommandConsole::Printf( L"    Runs the batch file with the specified file name." );
    }
    else if ( wcslen( wszFileName ) > 0 )
    {
        const int iBufferSize = ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE - 1;
        Asura_Char szFileName[ iBufferSize ];
        Asura_PrintfToBfr( szFileName, "%S", iBufferSize, true, wszFileName );

        const Asura_Handle iFileHandle = Asura_FileAccess::Open( szFileName );
        if ( iFileHandle != ASURA_INVALID_HANDLE )
        {
            Asura_Char        szBuffer[ iBufferSize ];
            Asura_UnicodeChar wszBuffer[ iBufferSize ];

            while ( Asura_FileAccess::ReadLine( iFileHandle, szBuffer, iBufferSize ) )
            {
                Asura_PrintfToBfr( wszBuffer, L"%S", iBufferSize, true, szBuffer );
                ParseCommandBuffer( wszBuffer, true, false, NULL, NULL );
            }

            Asura_FileAccess::Close( iFileHandle );
        }
        else
        {
            Asura_CommandConsole::Printf( CONSOLE_CMDNAME_RUN );
            Asura_CommandConsole::Printf( L"    Could not find the specified file." );
        }

        return true;
    }

    return false;
}
#endif

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
static bool ConsoleCmd_OpenDVS( bool bHelp )
{
    if( bHelp )
    {
        Asura_CommandConsole::Printf( L"Enter the debug variable system inteface." );
    }
    else
    {
        Asura_CommandConsole::Hide( false, false );
		Asura_DebugVariableSystem::OpenDisplayDVS();
    }
    return true;
}
#endif

#endif // #ifndef ASURA_COMMAND_CONSOLE_READONLY

static void EmptyInputBuffer( void )
{
    g_wszConsoleInputBuffer[0] = '\0';
    g_uConsoleInputCursorPosition = 0;
}

static void EmptyTabCompletionBuffer( void )
{
    g_wszTabCompletionBuffer[0] = '\0';
}

static void MoveInputBufferIntoHistory( void )
{
    if( g_uConsoleNumInputHistoryLines )
    {
        // Move first line back. If the input history is full then this will
        // just overwrite the last entry.
        if( g_uConsoleFirstInputHistoryLine )
        {
            --g_uConsoleFirstInputHistoryLine;
        }
        else
        {
            g_uConsoleFirstInputHistoryLine = CONSOLE_INPUT_HISTORY_MAX_LINES - 1;
        }

        // Update count.
        Asura_Assert( g_uConsoleNumInputHistoryLines <= CONSOLE_INPUT_HISTORY_MAX_LINES, "" );
        if( g_uConsoleNumInputHistoryLines < CONSOLE_INPUT_HISTORY_MAX_LINES )
        {
            ++g_uConsoleNumInputHistoryLines;
        }
    }
    else
    {
        g_uConsoleNumInputHistoryLines  = 1;
        g_uConsoleFirstInputHistoryLine = 0;
    }

    // Copy into correct slot in history array.
    wcscpy( g_awszConsoleInputHistory[g_uConsoleFirstInputHistoryLine], g_wszConsoleInputBuffer );

    // Reset cur history entry.
    g_iConsoleCurInputHistoryEntry = CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY;
}

static bool ParseCommandBuffer( const Asura_UnicodeChar* wszCmdBfr,
                                bool bEcho,
                                bool bInputFromConsole,
                                bool* pbReturnValue,
                                void(*pfnOutputCallback)(const Asura_UnicodeChar*) )
{
	while (*wszCmdBfr != L'\0')
	{
		if (*wszCmdBfr != L' ' && *wszCmdBfr != L'\t')
		{
			break;
		}
		wszCmdBfr++;
	}

	int ilen = wcslen(wszCmdBfr);

	if (ilen < 1)
	{
		// OR - 30/03/2005 12:43:52 - White line
		return false;
	}

	if (ilen > 2 && wszCmdBfr[0] == L'\\' && wszCmdBfr[1] == L'\\')
	{
		// OR - 30/03/2005 12:43:52 - Comment
		return false;
	}
	
    // Only deal with stuff that could conceivably have come from the console
    // itself.
    if( ilen >= ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE )
    {
        Asura_Assert( false, "" );
        return false;
    }

    g_pfnOutputOverride = pfnOutputCallback;

    Asura_UnicodeChar wszCommandName[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];
    Asura_UnicodeChar wszParameter[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];

    // Echo input to console.
    if( bEcho )
    {
        Asura_CommandConsole::Printf( ASURA_COMMAND_CONSOLE_CURSOR_PROMPT L"%s", wszCmdBfr );
    }

    // Split into command and parameter.
    {
        enum PARSESTATE
        {
            PARSESTATE_BEFORE_CMD,
            PARSESTATE_INSIDE_CMD,
            PARSESTATE_BETWEEN_CMD_AND_VALUE,
            PARSESTATE_INSIDE_PARAMETER,
        };

        PARSESTATE         eParseState = PARSESTATE_BEFORE_CMD;
        Asura_UnicodeChar* wszCmdOut   = wszCommandName;
        Asura_UnicodeChar* wszParamOut = wszParameter;

        for( u_int u = 0; u < ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE - 1 && wszCmdBfr[u]; ++u )
        {
            const Asura_UnicodeChar wNext = wszCmdBfr[u];

            switch( eParseState )
            {
                case PARSESTATE_BEFORE_CMD:
                    if( iswspace(wNext) )
                    {
                        // Just skip leading whitespace.
                    }
                    else
                    {
                        *wszCmdOut++ = wNext;
                        eParseState  = PARSESTATE_INSIDE_CMD;
                    }
                    break;

                case PARSESTATE_INSIDE_CMD:
                    // Copy into command name buffer until we hit whitespace.
                    if( iswspace(wNext) )
                    {
                        eParseState = PARSESTATE_BETWEEN_CMD_AND_VALUE;
                    }
                    else
                    {
                        *wszCmdOut++ = wNext;
                    }
                    break;

                case PARSESTATE_BETWEEN_CMD_AND_VALUE:
                    if( iswspace(wNext) )
                    {
                        // Just skip whitespace.
                    }
                    else
                    {
                        *wszParamOut++ = wNext;
                        eParseState    = PARSESTATE_INSIDE_PARAMETER;
                    }
                    break;

                case PARSESTATE_INSIDE_PARAMETER: 
                    *wszParamOut++ = wNext;
                    break;

                default:
                    Asura_Assert( false, "Invalid syntax position." );
                    break;
            }
        }
        
        *wszCmdOut   = '\0';
        *wszParamOut = '\0';
    }

    bool bValid;

    // Any command found?
    if( wszCommandName[0] )
    {
        // Trim parameter.
        Asura_StringUtil::StripWhitespace( wszParameter, Asura_StringUtil::STRIP_BOTH_ENDS );

        // Execute command.
        bValid = Asura_CommandConsole_CmdList::Exec( wszCommandName, wszParameter, false, pbReturnValue );

        // If it wasn't a valid command then maybe it's a var.
        if( !bValid )
        {
            if( wszParameter[0] )
            {
				if (Asura_CommandConsole_VarRepository::IsVarReadOnly( wszCommandName ))
				{
					bValid = true;
					Asura_CommandConsole::Printf( L"Variable '%s' is read-only.", wszCommandName );
				}
				else
				{
					bValid = Asura_CommandConsole_VarRepository::SetValue( wszCommandName, wszParameter, pbReturnValue );
				}
            }
            else
            {
                bValid = Asura_CommandConsole_VarRepository::ReportValue( wszCommandName );
                if( pbReturnValue ) { *pbReturnValue = bValid; }
            }
        }

        // Output an error message if it wasn't a var either.
        if( !bValid )
        {
            Asura_CommandConsole::Printf( L"Invalid command or variable '%s'.", wszCommandName );
        }

        if( bInputFromConsole )
        {
            MoveInputBufferIntoHistory();

            EmptyInputBuffer();
            EmptyTabCompletionBuffer();
        }
    }
    else
    {
        Asura_CommandConsole::Printf( L"Syntax error: command \"%ls\" found.", wszCmdBfr );
        bValid = false;
    }

    g_pfnOutputOverride = NULL;
    return bValid;
}

static void ShowConsole( void )
{
    Asura_Assert( !g_bConsoleDown, "" );

    g_bConsoleDown = true;
    g_pxIO->SetDown( true, false );
}

static void UpdateOutputBufferStateForNewLine( void )
{
    if( !g_uConsoleNumLinesUsed )
    {
        g_uConsoleNumLinesUsed  = 1;
        g_uConsoleFirstLineUsed = 0;
        g_uConsoleLastLineUsed  = 0;
    }
    else if( ASURA_COMMAND_CONSOLE_MAX_LINES == g_uConsoleNumLinesUsed )
    {
        // The first line should immediately follow the last line in the
        // line buffer then.
        Asura_Assert( g_uConsoleLastLineUsed + 1 == g_uConsoleFirstLineUsed || (!g_uConsoleFirstLineUsed && ASURA_COMMAND_CONSOLE_MAX_LINES - 1 == g_uConsoleLastLineUsed), "" );

        // Remove first line.
        ++g_uConsoleFirstLineUsed;
        if( ASURA_COMMAND_CONSOLE_MAX_LINES == g_uConsoleFirstLineUsed )
        {
            g_uConsoleFirstLineUsed = 0;
        }

        // Update last line used.
        ++g_uConsoleLastLineUsed;
        if( ASURA_COMMAND_CONSOLE_MAX_LINES == g_uConsoleLastLineUsed )
        {
            g_uConsoleLastLineUsed = 0;
        }

        // The first line should still immediately follow the last line in
        // the line buffer!
        Asura_Assert( g_uConsoleLastLineUsed + 1 == g_uConsoleFirstLineUsed || (!g_uConsoleFirstLineUsed && ASURA_COMMAND_CONSOLE_MAX_LINES - 1 == g_uConsoleLastLineUsed), "" );
    }
    else
    {
        // Update used line count.
        Asura_Assert( g_uConsoleNumLinesUsed < ASURA_COMMAND_CONSOLE_MAX_LINES, "" );
        ++g_uConsoleNumLinesUsed;

        // Update last line used.
        ++g_uConsoleLastLineUsed;
        if( ASURA_COMMAND_CONSOLE_MAX_LINES == g_uConsoleLastLineUsed )
        {
            g_uConsoleLastLineUsed = 0;
        }
    }
}
#ifndef ASURA_COMMAND_CONSOLE_READONLY
static void UseInputHistoryEntry( INPUT_HISTORY_DELTA eDelta )
{
    if( !g_uConsoleNumInputHistoryLines )
    {
        // (MP - 25.11.2002) TODO: Beep at user or something?
        return;
    }

    if( CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY == g_iConsoleCurInputHistoryEntry )
    {
        if( INPUT_HISTORY_DELTA_BACK == eDelta )
        {
            g_iConsoleCurInputHistoryEntry = g_uConsoleFirstInputHistoryLine;
        }
        else
        {
            g_iConsoleCurInputHistoryEntry = g_uConsoleFirstInputHistoryLine + (g_uConsoleNumInputHistoryLines - 1);
            if( static_cast<u_int>(g_iConsoleCurInputHistoryEntry) >= CONSOLE_INPUT_HISTORY_MAX_LINES )
            {
                g_iConsoleCurInputHistoryEntry -= CONSOLE_INPUT_HISTORY_MAX_LINES;
            }
        }
    }
    else
    {
        Asura_Assert( g_iConsoleCurInputHistoryEntry >= 0, "" );

        if( INPUT_HISTORY_DELTA_BACK == eDelta )
        {
            u_int uLastEntry = g_uConsoleFirstInputHistoryLine + (g_uConsoleNumInputHistoryLines - 1);
            if( uLastEntry >= CONSOLE_INPUT_HISTORY_MAX_LINES )
            {
                uLastEntry -= CONSOLE_INPUT_HISTORY_MAX_LINES;
            }

            if( static_cast<u_int>(g_iConsoleCurInputHistoryEntry) == uLastEntry )
            {
                g_iConsoleCurInputHistoryEntry = CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY;
            }
            else
            {
                ++g_iConsoleCurInputHistoryEntry;
                if( static_cast<u_int>(g_iConsoleCurInputHistoryEntry) >= CONSOLE_INPUT_HISTORY_MAX_LINES )
                {
                    g_iConsoleCurInputHistoryEntry = 0;
                }
            }
        }
        else
        {
            if( static_cast<u_int>(g_iConsoleCurInputHistoryEntry) == g_uConsoleFirstInputHistoryLine )
            {
                g_iConsoleCurInputHistoryEntry = CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY;
            }
            else
            {
                --g_iConsoleCurInputHistoryEntry;
                if( g_iConsoleCurInputHistoryEntry < 0 )
                {
                    g_iConsoleCurInputHistoryEntry = CONSOLE_INPUT_HISTORY_MAX_LINES - 1;
                }
            }
        }
    }

    if( CONSOLE_INPUT_HISTORY_NO_CUR_ENTRY == g_iConsoleCurInputHistoryEntry )
    {
        EmptyInputBuffer();
    }
    else
    {
        wcscpy( g_wszConsoleInputBuffer, g_awszConsoleInputHistory[g_iConsoleCurInputHistoryEntry] );
        g_uConsoleInputCursorPosition = wcslen( g_wszConsoleInputBuffer );
    }

    EmptyTabCompletionBuffer();
}
#endif
void Asura_CommandConsole::TimedShowOutputLine(float fDuration)
{
	if(fDuration <= 0.0f) return;
	if(IsConsoleDown()) return;//irrelevant if console is fully down

	if(g_uNumTimedLines < CONSOLE_NUM_LINES_TO_DISPLAY)
	{
		//add new timer to the end
		g_afLineTimers[g_uNumTimedLines] = fDuration;
		g_uNumTimedLines++;
		g_pxIO->SetPartiallyDown(g_uNumTimedLines,false);
	}
	else
	{
		//find the slot with the minimum time left
		float fMin = Asura_Maths::LargeFloat;
		int iBest = 0;
		for( u_int i=0;i<CONSOLE_NUM_LINES_TO_DISPLAY;i++)
		{
			if(g_afLineTimers[i] < fMin)
			{
				fMin = g_afLineTimers[i];
				iBest = i;
			}
		}

		if(fMin < fDuration)
		{
			g_afLineTimers[iBest] = fMin;
		}
	}

}

void Asura_CommandConsole::UpdateLineTimers()
{
	if(!g_uNumTimedLines) return;

	u_int uInitialLines = g_uNumTimedLines;
	float fTime = Asura_Timers::GetFrameTime();

	for(u_int u=0;u<g_uNumTimedLines;u++)
	{
		g_afLineTimers[u] -= fTime;
		if(g_afLineTimers[u]<=0)
		{
			//timer expirted
			g_afLineTimers[u] = g_afLineTimers[g_uNumTimedLines-1];
			g_uNumTimedLines--;
			u--;
		}
	}

	if(uInitialLines != g_uNumTimedLines)
	{
		//update the console
		g_pxIO->SetPartiallyDown(g_uNumTimedLines,false);
	}
}

void Asura_CommandConsole::ResetLineTimers()
{
	g_uNumTimedLines = 0;
}

float Asura_CommandConsole::GetBottomY()
{
	Asura_Assert(g_pxIO,"");
	return g_pxIO->GetBottomY();
}

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

Asura_DebugVariableSystem::DVS_Item::DVS_Item( Asura_UnicodeChar* wszName,
											  Asura_CommandConsole_VarRepository::Var* pxVariable,
											  Asura_CommandConsole_CmdList::Node* pxCommand )
: m_wszName(wszName)
, m_pxVariable(pxVariable)
, m_pxCommand(pxCommand)
{
	DVS_Item* pxPrev = NULL;
	DVS_Item* pxNext = s_pxRootDVSItem;

	// Find place to insert to keep list sorted alphabetically
	while( pxNext && wcsicmp(m_wszName, pxNext->m_wszName) > 0 )
	{
		pxPrev = pxNext;
		pxNext = pxNext->m_pxNext;
	}

	if( pxPrev )
	{
		pxPrev->m_pxNext = this;
	}
	else
	{
		s_pxRootDVSItem = this;
	}
	m_pxNext = pxNext;
}

Asura_DebugVariableSystem::DVS_Item::~DVS_Item()
{
	if (s_pxRootDVSItem == this)
	{
		// Set new root node
		s_pxRootDVSItem = m_pxNext;
	}
	else
	{
		// Find the previous node by iterating through the list from the root
		DVS_Item* pxPrevNode = s_pxRootDVSItem;
		while (pxPrevNode->m_pxNext != this)
		{
			pxPrevNode = pxPrevNode->m_pxNext;
		}

		// Set next of previous node to be one after the one we remove
		pxPrevNode->m_pxNext = m_pxNext;
	}
	delete [] m_wszName;
}

void Asura_DebugVariableSystem::AddVar(const Asura_UnicodeChar* wszName, Asura_CommandConsole_VarRepository::Var* pxVariable)
{
	Asura_UnicodeChar* wszConvertedName = ConvertName(wszName);
	Asura_DebugVariableSystem::DVS_Item* pxExisting = FindItem(wszConvertedName);
	if (pxExisting)
	{
		Asura_Assertf(pxExisting->m_pxVariable, ("DVS variable (%s) clashes with a DVS command", wszConvertedName));
		delete [] wszConvertedName;
		return;
	}
	
	DVS_Item* pxNewItem = new DVS_Item(wszConvertedName, pxVariable, 0);
	char szName[128];
	Asura_PrintfToBfr(szName, "%S", 128, true, wszConvertedName);
	*s_xItemTree.AddNode(Asura_GetHashID(szName)) = pxNewItem;
}

void Asura_DebugVariableSystem::AddCmd(const Asura_UnicodeChar* wszName, Asura_CommandConsole_CmdList::Node* pxCmd)
{
	if (pxCmd->GetType() != Asura_CommandConsole_CmdList::CMDTYPE_VOID
		&& pxCmd->GetType() != Asura_CommandConsole_CmdList::CMDTYPE_BOOL)
	{
		//NY - only supporting these for now...
		return;
	}
	if (wcscmp(wszName, L"DVS") == 0)
	{
		//NY - not only does this one make no sense from the DVS, it crashes!
		return;
	}
	Asura_UnicodeChar* wszConvertedName = ConvertName(wszName);
	Asura_DebugVariableSystem::DVS_Item* pxExisting = FindItem(wszConvertedName);
	if (pxExisting)
	{
		Asura_Assertf(pxExisting->m_pxCommand, ("DVS command (%s) clashes with a DVS variable", wszConvertedName));
		delete [] wszConvertedName;
		return;
	}

	DVS_Item* pxNewItem = new DVS_Item(wszConvertedName, 0, pxCmd);
	char szName[128];
	Asura_PrintfToBfr(szName, "%S", 128, true, wszConvertedName);
	*s_xItemTree.AddNode(Asura_GetHashID(szName)) = pxNewItem;
}

void Asura_DebugVariableSystem::DelVar(const Asura_UnicodeChar* wszName)
{
	Asura_UnicodeChar* wszConvertedName = ConvertName(wszName);
	Asura_DebugVariableSystem::DVS_Item* pxNode = FindItem(wszConvertedName);
	if (pxNode && pxNode->m_pxVariable)
	{
		// Remove from hash tree
		char szName[128];
		Asura_PrintfToBfr(szName, "%S", 128, true, wszConvertedName);
		s_xItemTree.DeleteNodeByHash(Asura_GetHashID(szName));

		delete pxNode;
	}
	delete [] wszConvertedName;
}

void Asura_DebugVariableSystem::DelCmd(const Asura_UnicodeChar* wszName)
{
	Asura_UnicodeChar* wszConvertedName = ConvertName(wszName);
	Asura_DebugVariableSystem::DVS_Item* pxNode = FindItem(wszConvertedName);
	if (pxNode && pxNode->m_pxCommand)
	{
		// Remove from hash tree
		char szName[128];
		Asura_PrintfToBfr(szName, "%S", 128, true, wszConvertedName);
		s_xItemTree.DeleteNodeByHash(Asura_GetHashID(szName));

		delete pxNode;
	}
	delete [] wszConvertedName;
}

Asura_DebugVariableSystem::DVS_Item* Asura_DebugVariableSystem::FindItem(const Asura_UnicodeChar* wszName)
{
	char szName[128];
	Asura_PrintfToBfr(szName, "%S", 128, true, wszName);

	Asura_DebugVariableSystem::DVS_Item** pxNode = s_xItemTree.GetDataByHash(Asura_GetHashID(szName));
	if (pxNode)
	{
		return *pxNode;
	}
	return 0;
}

Asura_DebugVariableSystem::DVS_Item* Asura_DebugVariableSystem::FindItem(const Asura_Hash_ID uHash)
{
	Asura_DebugVariableSystem::DVS_Item** pxNode = s_xItemTree.GetDataByHash(uHash);
	if (pxNode)
	{
		return *pxNode;
	}
	return 0;
}

Asura_UnicodeChar* Asura_DebugVariableSystem::ConvertName(const Asura_UnicodeChar* wszName)
{
	Asura_UnicodeChar* wszNameCopy;
    // modify uncategorised items so they
    // don't all messily appear at the top level.
    if( !wcschr(wszName, L'.') )
    {
        const Asura_UnicodeChar* wszPrefix = L"Unclassified.";

        u_int uNewBfrSize = wcslen(wszName) + wcslen(wszPrefix) + 1;
        wszNameCopy = new Asura_UnicodeChar[uNewBfrSize];

        if( wszNameCopy )
        {
            Asura_PrintfToBfr( wszNameCopy, L"%s%s", uNewBfrSize, true, wszPrefix, wszName );
        }
    }
    else
    {
        wszNameCopy = Asura_StringUtil::Duplicate( wszName );
    }
	return wszNameCopy;
}

void Asura_DebugVariableSystem::EarlyInitialise()
{
	Asura_CommandConsole::AddCmd(L"DVS", ConsoleCmd_OpenDVS, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
}

void Asura_DebugVariableSystem::LateInitialise()
{
	// moved from EarlyInitialise() so we can interrogate PC joystick types to stop any pad button clashes
	s_xControlMap.CreateMappings( KEYMAP_NUMMAPPINGS );
    Platform_InitialiseDefaultControlMapping();
	Asura_CommandConsole::AddVar(L"DVS.RepeatFirstDelay", &s_fRepeatDelayFirst, 0.1f, 2.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"DVS.RepeatLoopedDelay", &s_fRepeatDelay, 0.05f, 2.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
}

void Asura_DebugVariableSystem::Shutdown()
{
	s_xControlMap.DeleteMappings();
}

void Asura_DebugVariableSystem::Update( void )
{
	if( Asura_CommandConsole::IsConsoleDown() )
	{
		CloseDisplayDVS();
		return;
	}

	if( !s_bShowDVS )
	{
		if( Project_ShouldOpen() )
		{
			OpenDisplayDVS();
		}

		if( !s_bShowDVS )
		{
			return;
		}
	}
	else
	{
#ifdef ASURA_USE_CONTROLMAP_MANAGER
		if( s_xControlMap.GetValue( KEYMAP_CLOSE))
		{
			CloseDisplayDVS();
			return;
		}
#else
		if( Project_ShouldClose() )
		{
			CloseDisplayDVS();
			return;
		}
#endif
	}

#ifndef ASURA_USE_CONTROLMAP_MANAGER
	s_xControlMap.Update();
#endif

	// NY - 20 June 2008 - only support one key at once to keep repeat simple, and current dvs doesn't need combinations
	KEYMAP eCurrentKey = UpdateRepeatedKeys();
	
	switch (eCurrentKey)
	{
		default:
		{
			Asura_Assert(0, "unhandled case");
		}//fallthrough
		case KEYMAP_NUMMAPPINGS:
		{
			break;
		}
		case KEYMAP_MOVERIGHT:
		{
			MoveRight();
			break;
		}
		case KEYMAP_MOVELEFT:
		{
			MoveLeft();
			break;
		}
		case KEYMAP_MOVEUP:
		{
			MoveUp();
			break;
		}
		case KEYMAP_MOVEDOWN:
		{
			MoveDown();
			break;
		}
		case KEYMAP_INCREASEVALUE:
		{
			IncreaseValue();
			break;
		}
		case KEYMAP_DECREASEVALUE:
		{
			DecreaseValue();
			break;
		}
		case KEYMAP_INCREASEVALUEINCREMENT:
		{
			IncreaseValueIncrement();
			break;
		}
		case KEYMAP_DECREASEVALUEINCREMENT:
		{
			DecreaseValueIncrement();
			break;
		}
	}
}

Asura_DebugVariableSystem::KEYMAP Asura_DebugVariableSystem::UpdateRepeatedKeys()
{
	KEYMAP eCurrentKey = KEYMAP_NUMMAPPINGS;
	bool bKeyClash = false;
	u_int uCurrentKeys = 0;

	//loop through all repeatable keys - "close" isn't repeatable
	for (u_int u = 0; u < KEYMAP_CLOSE; ++u)
	{
		if (s_xControlMap.GetValue(u))
		{
			if (eCurrentKey != KEYMAP_NUMMAPPINGS)
			{
				bKeyClash = true;
			}
			eCurrentKey = static_cast<KEYMAP>(u);
			uCurrentKeys |= (1 << u);//bitfield of pressed keys - used for debounced checks
		}
	}
	if (bKeyClash)
	{
		//report no key pressed if more than one
		eCurrentKey = KEYMAP_NUMMAPPINGS;
	}

	if (eCurrentKey != KEYMAP_NUMMAPPINGS)
	{
		if ((s_uLastFrameKeys & (1 << eCurrentKey)) == 0)
		{
			//debounced
			s_eRepeatingKey = eCurrentKey;
			//start timer at longer first delay
			s_fRepeatTimer = s_fRepeatDelayFirst;
		}
		else if (eCurrentKey == s_eRepeatingKey)
		{
			s_fRepeatTimer -= Asura_Timers::GetFrameTime();
			if (s_fRepeatTimer <= 0.0f)
			{
				//reset timer to shorter looped delay
				s_fRepeatTimer = s_fRepeatDelay;
			}
			else
			{
				//timer still counting, so don't return a keypress
				eCurrentKey = KEYMAP_NUMMAPPINGS;
			}
		}
		else
		{
			//no longer pressing previous repeating key, stop repeating
			s_eRepeatingKey = KEYMAP_NUMMAPPINGS;
		}
	}

	s_uLastFrameKeys = uCurrentKeys;
	return eCurrentKey;
}

void Asura_DebugVariableSystem::MoveRight()
{	
	Asura_Assert(wcslen(s_wszSelectedPath) < 1023, "DVS selected path is too long.");

	int iSelectedLength = wcslen(s_wszSelectedPath);

	Asura_UnicodeChar wszNewPath[1024];

	for( const DVS_Item* pxItem = RootItem(); pxItem; pxItem = pxItem->m_pxNext )
	{
		if(wcsicmp(pxItem->m_wszName, s_wszSelectedPath) != 0 && 
			wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, iSelectedLength) == 0)
		{
			wcscpy(wszNewPath, pxItem->m_wszName);

			Asura_UnicodeChar* pwszDot = wcschr(wszNewPath + iSelectedLength + 1, '.');
			if(pwszDot)		
			{
				*pwszDot = 0;
			}

			wcscpy(s_wszSelectedPath, wszNewPath);
			break;
		}
	}
}

void Asura_DebugVariableSystem::MoveLeft()
{
	Asura_Assert(wcslen(s_wszSelectedPath) < 1023, "DVS selected path is too long.");

	Asura_UnicodeChar wszNewPath[1024] = L"";

	Asura_UnicodeChar* pwszDot = wcsrchr(s_wszSelectedPath, '.');
	if(pwszDot)
	{
		int iLength = (reinterpret_cast<int>(pwszDot) - reinterpret_cast<int>(s_wszSelectedPath)) / 2;
		wcsncpy(wszNewPath, s_wszSelectedPath, iLength);
		wszNewPath[iLength] = '\0';
	}

	wcscpy(s_wszSelectedPath, wszNewPath);	
}

void Asura_DebugVariableSystem::MoveUp()
{	
	Asura_Assert(wcslen(s_wszSelectedPath) < 1023, "DVS selected path is too long.");

	u_int uSelectedPathLength = wcslen(s_wszSelectedPath);

	//work out current depth, and length of string defining parent node
	u_int uParentLength = 0;
	u_int uDepth = 0;
	bool bFoundDepth = false;
	while (!bFoundDepth)
	{
		u_int uStart = 0;
		u_int uLength = 0;
		bFoundDepth = !Asura_StringUtil::FindStringPartByIndex(s_wszSelectedPath, L'.', uDepth + 1, uStart, uLength);
		if (!bFoundDepth)
		{
			++uDepth;
			uParentLength = Asura_Maths::Max(1u, uStart) - 1;
		}
	}

	//find next node in section
	const DVS_Item* pxPreviousNodeInSection = 0;
	const DVS_Item* pxLastNodeInSection = 0;
	bool bFoundNode = false;
	for( const DVS_Item* pxItem = RootItem(); pxItem; pxItem = pxItem->m_pxNext )
	{
		const u_int uNameLength = wcslen(pxItem->m_wszName);
		//look for nodes in same section
		if (uParentLength == 0 || (uNameLength >= uParentLength && wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, uParentLength) == 0))
		{
			pxLastNodeInSection = pxItem;
			//look for current node
			if (!bFoundNode
				&& uNameLength >= uSelectedPathLength
				&& wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, uSelectedPathLength) == 0)
			{
				bFoundNode = true;
			}
			if (!bFoundNode)
			{
				pxPreviousNodeInSection = pxItem;
			}
		}
		else if (pxLastNodeInSection)
		{
			//we've left section, no need to loop further
			break;
		}
	}

	//if we don't have a next node, but do have a first in section then use that - IE, wrap
	if (!pxPreviousNodeInSection && pxLastNodeInSection)
	{
		pxPreviousNodeInSection = pxLastNodeInSection;
	}

	// if we have a valid node, and we can find the the name at the current depth, then use it as the new current node
	if (pxPreviousNodeInSection)
	{
		u_int uStart = 0;
		u_int uLength = 0;
		if (Asura_StringUtil::FindStringPartByIndex(pxPreviousNodeInSection->m_wszName, L'.', uDepth, uStart, uLength))
		{
			wcsncpy(s_wszSelectedPath, pxPreviousNodeInSection->m_wszName, uStart + uLength);
			s_wszSelectedPath[uStart + uLength] = 0;
		}
	}
}

void Asura_DebugVariableSystem::MoveDown()
{
	Asura_Assert(wcslen(s_wszSelectedPath) < 1023, "DVS selected path is too long.");

	u_int uSelectedPathLength = wcslen(s_wszSelectedPath);

	//work out current depth, and length of string defining parent node
	u_int uParentLength = 0;
	u_int uDepth = 0;
	bool bFoundDepth = false;
	while (!bFoundDepth)
	{
		u_int uStart = 0;
		u_int uLength = 0;
		bFoundDepth = !Asura_StringUtil::FindStringPartByIndex(s_wszSelectedPath, L'.', uDepth + 1, uStart, uLength);
		if (!bFoundDepth)
		{
			++uDepth;
			uParentLength = Asura_Maths::Max(1u, uStart) - 1;
		}
	}

	//find next node in section
	const DVS_Item* pxFirstNodeInSection = 0;
	const DVS_Item* pxNextNodeInSection = 0;
	bool bFoundNode = false;
	for( const DVS_Item* pxItem = RootItem(); pxItem; pxItem = pxItem->m_pxNext )
	{
		const u_int uNameLength = wcslen(pxItem->m_wszName);
		//look for first node in section (IE, first node with same parents)
		if (!pxFirstNodeInSection
			&& (uParentLength == 0 || (uNameLength >= uParentLength && wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, uParentLength) == 0)))
		{
			pxFirstNodeInSection = pxItem;
		}
		//look for current node
		if (!bFoundNode)
		{
			if (uNameLength >= uSelectedPathLength
				&& wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, uSelectedPathLength) == 0)
			{
				bFoundNode = true;
			}
		}
		else if (!pxNextNodeInSection)
		{
			//look for next node in section (same parents, different string at this depth)
			if (uNameLength >= uParentLength
				&& wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, uParentLength) == 0)
			{
				//bit gnarly, but this checks the string is different at the current depth wiothout needing temp strings
				if (uNameLength < uSelectedPathLength
					|| (uNameLength > uSelectedPathLength && pxItem->m_wszName[uSelectedPathLength] != L'.')
					|| wcsnicmp(pxItem->m_wszName, s_wszSelectedPath, uSelectedPathLength) != 0)
				{
					pxNextNodeInSection = pxItem;
					//no need to keep looping
					break;
				}
			}
			else
			{
				//left section, no point in still looping
				break;
			}
		}
	}

	//if we don't have a next node, but do have a first in section then use that - IE, wrap
	if (!pxNextNodeInSection && pxFirstNodeInSection)
	{
		pxNextNodeInSection = pxFirstNodeInSection;
	}

	// if we have a valid node, and we can find the the name at the current depth, then use it as the new current node
	if (pxNextNodeInSection)
	{
		u_int uStart = 0;
		u_int uLength = 0;
		if (Asura_StringUtil::FindStringPartByIndex(pxNextNodeInSection->m_wszName, L'.', uDepth, uStart, uLength))
		{
			wcsncpy(s_wszSelectedPath, pxNextNodeInSection->m_wszName, uStart + uLength);
			s_wszSelectedPath[uStart + uLength] = 0;
		}
	}
}

void Asura_DebugVariableSystem::IncreaseValue()
{
	DVS_Item* pxItem = FindItem(s_wszSelectedPath);
	if (pxItem)
	{
		Asura_CommandConsole_VarRepository::Var* pxVar = pxItem->m_pxVariable;
		if(pxVar && !pxVar->m_bReadOnly)
		{
			bool bModified = false;
			switch(pxVar->m_eType)
			{		
				case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL: 
				{			
					*pxVar->m_xData.pbData = !*pxVar->m_xData.pbData; 
					bModified = true;
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT: 
				{		
					const float fValue = *pxVar->m_xData.pfData;
					*pxVar->m_xData.pfData += s_fFloatIncrement; 
					// (SB - 12/12/06) Upper and lower bound parameters were there for a reason
					if (*pxVar->m_xData.pfData > pxVar->m_xData.fUpperLimit)
					{
						*pxVar->m_xData.pfData = pxVar->m_xData.fUpperLimit;
					}
					bModified = (fValue != *pxVar->m_xData.pfData);
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_INT: 
				{		
					const int iValue = *pxVar->m_xData.piData;
					*pxVar->m_xData.piData += s_uIntIncrement; 
					// (SB - 12/12/06) Upper and lower bound parameters were there for a reason
					if (*pxVar->m_xData.piData > pxVar->m_xData.iUpperLimit)
					{
						*pxVar->m_xData.piData = pxVar->m_xData.iUpperLimit;
					}
					bModified = (iValue != *pxVar->m_xData.piData);
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_STRING: 
				case Asura_CommandConsole_VarRepository::Var::TYPE_INVALID: 
				{		
					break;
				}
			}

			if (bModified && pxVar->m_pbModified)
			{
				*pxVar->m_pbModified = true;
			}
		}
		else if (pxItem->m_pxCommand)
		{
			switch (pxItem->m_pxCommand->GetType())
			{
				case Asura_CommandConsole_CmdList::CMDTYPE_VOID:
				{
					Asura_CommandConsole_CmdList::Exec(pxItem->m_pxCommand->GetName(), L"", 0, 0);
					break;
				}
				case Asura_CommandConsole_CmdList::CMDTYPE_BOOL:
				{
					Asura_CommandConsole_CmdList::Exec(pxItem->m_pxCommand->GetName(), L"1", 0, 0);
					break;
				}
				default:
				{
					break;
				}
			}
		}
	}
}
void Asura_DebugVariableSystem::DecreaseValue()
{
	DVS_Item* pxItem = FindItem(s_wszSelectedPath);
	if (pxItem)
	{
		Asura_CommandConsole_VarRepository::Var* pxVar = pxItem->m_pxVariable;
		if(pxVar && !pxVar->m_bReadOnly)
		{
			bool bModified = false;
			switch(pxVar->m_eType)
			{
				case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL: 
				{
					*pxVar->m_xData.pbData = !*pxVar->m_xData.pbData;
					bModified = true;
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT: 
				{
					const float fValue = *pxVar->m_xData.pfData;
					*pxVar->m_xData.pfData -= s_fFloatIncrement;
					// (SB - 12/12/06) Upper and lower bound parameters were there for a reason
					if (*pxVar->m_xData.pfData < pxVar->m_xData.fLowerLimit)
					{
						*pxVar->m_xData.pfData = pxVar->m_xData.fLowerLimit;
					}
					bModified = (fValue != *pxVar->m_xData.pfData);
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_INT: 
				{
					const int iValue = *pxVar->m_xData.piData;
					*pxVar->m_xData.piData -= s_uIntIncrement;
					// (SB - 12/12/06) Upper and lower bound parameters were there for a reason
					if (*pxVar->m_xData.piData < pxVar->m_xData.iLowerLimit)
					{
						*pxVar->m_xData.piData = pxVar->m_xData.iLowerLimit;
					}
					bModified = (iValue != *pxVar->m_xData.piData);
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_STRING: 
				case Asura_CommandConsole_VarRepository::Var::TYPE_INVALID: 
				{
					break;
				}
			}

			if (bModified && pxVar->m_pbModified)
			{
				*pxVar->m_pbModified = true;
			}
		}
		else if (pxItem->m_pxCommand)
		{
			switch (pxItem->m_pxCommand->GetType())
			{
				case Asura_CommandConsole_CmdList::CMDTYPE_VOID:
				{
					Asura_CommandConsole_CmdList::Exec(pxItem->m_pxCommand->GetName(), L"", 0, 0);
					break;
				}
				case Asura_CommandConsole_CmdList::CMDTYPE_BOOL:
				{
					Asura_CommandConsole_CmdList::Exec(pxItem->m_pxCommand->GetName(), L"0", 0, 0);
					break;
				}
				default:
				{
					break;
				}
			}
		}
	}
}
void Asura_DebugVariableSystem::IncreaseValueIncrement()
{
	DVS_Item* pxItem = FindItem(s_wszSelectedPath);
	if (pxItem)
	{
		Asura_CommandConsole_VarRepository::Var* pxVar = pxItem->m_pxVariable;
		if(pxVar)
		{
			switch(pxVar->m_eType)
			{			
				case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL: 
				{
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT:
				{
 					// (SB - 12/12/06) Increment increment should change with magnitude
					//BW 14/8/08 more versatile implementation:
					float fStep = Asura_Maths::Power(10.0f, static_cast<float>(s_iFloatMagnitude));
					s_fFloatIncrement += fStep;
					if ( s_fFloatIncrement+fStep*.5f > fStep*10.0f )
					{
						s_fFloatIncrement = fStep*10.0f; //fix precision errors
						s_iFloatMagnitude++;
					}
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_INT: 
				{
					s_uIntIncrement++; 
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_STRING: 
				case Asura_CommandConsole_VarRepository::Var::TYPE_INVALID: 
				{
					break;
				}
			}
		}
	}
}
void Asura_DebugVariableSystem::DecreaseValueIncrement()
{	
	DVS_Item* pxItem = FindItem(s_wszSelectedPath);
	if (pxItem)
	{
		Asura_CommandConsole_VarRepository::Var* pxVar = pxItem->m_pxVariable;
		if(pxVar)
		{
			switch(pxVar->m_eType)
			{
				case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL: 
				{
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT: 
				{
					// (SB - 12/12/06) Increment increment should change with magnitude
					//BW 14/8/08 more versatile implementation:
					float fLastStep = Asura_Maths::Power(10.0f, static_cast<float>(s_iFloatMagnitude));
					if ( s_fFloatIncrement-fLastStep*.5f < fLastStep )
					{
						s_fFloatIncrement = fLastStep; //fix precision errors
						s_iFloatMagnitude--;
					}
					float fStep = Asura_Maths::Power(10.0f, static_cast<float>(s_iFloatMagnitude));
					s_fFloatIncrement -= fStep;
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_INT: 
				{
					if(s_uIntIncrement) s_uIntIncrement--; 
					break;
				}
				case Asura_CommandConsole_VarRepository::Var::TYPE_STRING: 
				case Asura_CommandConsole_VarRepository::Var::TYPE_INVALID: 
				{
					break;
				}
			}
		}
	}
}

// ---------------------------------------------------------------------------------------

void Asura_DebugVariableSystem::OpenDisplayDVS( void)
{
	if( s_bShowDVS == false)
	{
		s_bShowDVS = true;
#ifdef ASURA_USE_CONTROLMAP_MANAGER
		Asura_ControlManager::PushOverrideControls( &s_xControlMap );
#endif
	}
}

// ---------------------------------------------------------------------------------------

void Asura_DebugVariableSystem::CloseDisplayDVS( void)
{
	if( s_bShowDVS == true)
	{
		s_bShowDVS = false;
#ifdef ASURA_USE_CONTROLMAP_MANAGER
        Asura_ControlManager::RemoveOverrideControls( &s_xControlMap );
#endif
	}
}

// ---------------------------------------------------------------------------------------

void Asura_DebugVariableSystem::GetVariableValue(const Asura_CommandConsole_VarRepository::Var* pxVariable, Asura_UnicodeChar* wszBuffer, u_int uBufferSize)
{
	switch(pxVariable->m_eType)
	{
	case Asura_CommandConsole_VarRepository::Var::TYPE_INT:
		{					
			Asura_PrintfToBfr(wszBuffer, L" : %i", uBufferSize, true, *(pxVariable->m_xData.piData)); 
			break;		
		}

	case Asura_CommandConsole_VarRepository::Var::TYPE_BOOL: 
		{					
			Asura_PrintfToBfr(wszBuffer, L" : %s", uBufferSize, true, *(pxVariable->m_xData.pbData) ? L"true" : L"false"); 
			break;
		}

	case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT: 
		{					
			Asura_PrintfToBfr(wszBuffer, L" : %f", uBufferSize, true, Asura_Maths::FtoD(*(pxVariable->m_xData.pfData)) );
			break;
		}

	case Asura_CommandConsole_VarRepository::Var::TYPE_STRING: 
		{			
			Asura_PrintfToBfr(wszBuffer, L" : %s", uBufferSize, true, pxVariable->m_xData.wszData); 
			break;
		}

	case Asura_CommandConsole_VarRepository::Var::TYPE_INVALID:
		{
			Asura_PrintfToBfr(wszBuffer, L" : Invalid", uBufferSize, true);
			break;
		}
	}
}

// ---------------------------------------------------------------------------------------

Asura_UnicodeChar Asura_DebugVariableSystem::s_wszSelectedPath[1024] = L"";
bool Asura_DebugVariableSystem::s_bShowDVS = false;
float Asura_DebugVariableSystem::s_fFloatIncrement = 0.1f;
int Asura_DebugVariableSystem::s_iFloatMagnitude = -1;
unsigned int Asura_DebugVariableSystem::s_uIntIncrement = 1;
Asura_ControlMap Asura_DebugVariableSystem::s_xControlMap;
unsigned int Asura_DebugVariableSystem::s_auDebugVarCategoryLevelScrollY[ASURA_DEBUG_VARIABLE_SYSTEM_MAX_CATEGORY_LEVELS];//make this bigger if more debug var categories are needed
float Asura_DebugVariableSystem::s_fDebugVarScrollX = 0.0f;
Asura_DebugVariableSystem::DVS_Item* Asura_DebugVariableSystem::s_pxRootDVSItem = 0;
Asura_TemplateHashTree<Asura_DebugVariableSystem::DVS_Item*> Asura_DebugVariableSystem::s_xItemTree;
float Asura_DebugVariableSystem::s_fRepeatDelayFirst = 0.4f;
float Asura_DebugVariableSystem::s_fRepeatDelay = 0.15f;
float Asura_DebugVariableSystem::s_fRepeatTimer = 0.0f;
Asura_DebugVariableSystem::KEYMAP Asura_DebugVariableSystem::s_eRepeatingKey = Asura_DebugVariableSystem::KEYMAP_NUMMAPPINGS;
u_int Asura_DebugVariableSystem::s_uLastFrameKeys = 0;
#endif//#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

//#################################
//#################################
// DLH - Late Initialise Code
//#################################
//#################################

#ifdef ASURA_TOOLS
#include "Hierarchy/Tools_Animation_Files.h"
#endif

#ifdef ASURA_USE_BOTANICALS
#include "Botanicals/Asura_Botanicals_InstanceManager.h"
#endif // #ifdef ASURA_USE_BOTANICALS

#ifdef ASURA_USE_CARPET
#include "Carpet/Asura_Carpet_MeshManager.h"
#endif // #ifdef ASURA_USE_CARPET

#ifdef ASURA_SUPPORT_ENTITYWATCH
#include "Entity/Asura_EntityWatch_System.h"
#endif

#ifdef ASURA_USE_SCREENSHOT_SYSTEM
#include "Debugging/Asura_ScreenShot_System.h"
#endif

#ifdef ASURA_USE_DEBUG_MESSAGE_TRIGGER
#include "Entity/Triggers/Asura_ServerEntity_DebugMessageTrigger.h"
#include "Entity/Triggers/Asura_ServerEntity_DebugWaypointTrigger.h"
#endif // #ifdef ASURA_USE_DEBUG_MESSAGE_TRIGGER

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
#include "Jobs/Asura_Job.h"
#endif


#ifndef ASURA_COMMAND_CONSOLE_READONLY
// Built-in console cmd functions.
static bool ConsoleCmd_Clear( bool bHelp );
static bool ConsoleCmd_Help( const Asura_UnicodeChar* wszCmd, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_ListCmds( const Asura_UnicodeChar* wszPrefix, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_ListVars( const Asura_UnicodeChar* wszPrefix, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_FontScale( float fSize, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_Timescale( float fTimeScale, bool bInterrogate, bool bHelp );
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE_RUN
static bool ConsoleCmd_Run( const Asura_UnicodeChar* wszFileName, bool bInterrogate, bool bHelp );
#endif
#ifdef ASURA_ENABLE_INTERNET_SIMULATION
static bool ConsoleCmd_NetSim( bool bEnable, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_NetSim_Latency( u_int uLatencyMean, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_NetSim_LatencyStdDev( u_int uLatencyMean, bool bInterrogate, bool bHelp );
static bool ConsoleCmd_NetSim_PacketLoss( float fPacketLossPercent, bool bInterrogate, bool bHelp );
#endif
#if ASURA_NETWORK_LOGGING && defined(ASURA_SUPPORT_EXTERNAL_NETWORK)
static bool ConsoleCmd_SetMsgLogLevel( u_int uLevel, bool bInterrogate, bool bHelp );
#endif
#endif
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
static bool ConsoleCmd_MultiThreadJobQueues(int iValue, bool bInterrogate, bool bHelp );
#endif

#ifdef ASURA_ENABLE_INTERNET_SIMULATION
#define CONSOLE_CMDNAME_NETSIM					L"NetSim"
#define CONSOLE_CMDNAME_NETSIM_LATENCY			L"NetSim_Latency"
#define CONSOLE_CMDNAME_NETSIM_LATENCYSTDDEV	L"NetSim_LatencyStdDev"
#define CONSOLE_CMDNAME_NETSIM_PACKETLOSS		L"NetSim_PacketLoss"
#endif
#define CONSOLE_CMDNAME_NETWORK_LOGGING			L"SetNetworkMsgLogLevel"
#ifdef ASURA_TOOLS
#define CONSOLE_CMDNAME_RELOADANIMS L"ReloadAnims"
#endif
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
#define CONSOLE_CMDNAME_MULTITHREADJOBQUEUES L"MultiThreadJobQueues"
#endif

#ifdef ASURA_ENABLE_INTERNET_SIMULATION
static bool ConsoleCmd_NetSim( bool bEnable, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_NETSIM L": turns network packet simulator on/off. Also see other commands with " CONSOLE_CMDNAME_NETSIM L"prefix." );
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf( L"NetSim currently: %s", Asura_InternetSimulation::IsEnabled() ? L"on" : L"off" );
	}
	else
	{
		Asura_InternetSimulation::SetEnabled( bEnable );
		Asura_CommandConsole::Printf( L"NetSim now: %s", bEnable ? L"on" : L"off" );
		if (bEnable)
		{
			Asura_CommandConsole::Printf( L"  PacketLoss: %.1f%%, Latency Mean %u, Latency StdDev %u", 
				Asura_Maths::FtoD(Asura_InternetSimulation::GetDropPercent()), 
				Asura_InternetSimulation::GetLatencyMean(),
				Asura_InternetSimulation::GetLatencyStandardDeviation() );
		}
	}
	return true;
}

static bool ConsoleCmd_NetSim_Latency( u_int uLatencyMean, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_NETSIM_LATENCY L": sets mean latency." );
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf( L"Mean Latency currently: %u", Asura_InternetSimulation::GetLatencyMean() );
	}
	else
	{
		Asura_InternetSimulation::SetLatencyMean( uLatencyMean );
	}
	return true;
}

static bool ConsoleCmd_NetSim_LatencyStdDev( u_int uLatencyStdDev, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_NETSIM_LATENCYSTDDEV L": sets standard deviation of latency." );
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf( L"Latency standard deviation currently: %u", Asura_InternetSimulation::GetLatencyStandardDeviation() );
	}
	else
	{
		Asura_InternetSimulation::SetLatencyStandardDeviation( uLatencyStdDev );
	}
	return true;
}

static bool ConsoleCmd_NetSim_PacketLoss( float fPacketLossPercent, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_NETSIM_PACKETLOSS L": sets percentage of packets dropped." );
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf( L"Packet loss currently: %.1f%%", Asura_Maths::FtoD(Asura_InternetSimulation::GetDropPercent()) );
	}
	else
	{
		Asura_InternetSimulation::SetDropPercent( fPacketLossPercent );
	}
	return true;
}
#endif

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
static bool ConsoleCmd_MultiThreadJobQueues(int iValue, bool bInterrogate, bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_MULTITHREADJOBQUEUES L": Enables/Disables multithreading" );
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf( L"Multithreaded job queues currently %s", Asura_Job_Queue_System::GetForceSingleThreaded()?L"Disabled":L"Enabled" );
	}
	else
	{
		Asura_Job_Queue_System::SetForceSingleThreaded(!(iValue!=0)		);
	}
	return true;
}
#endif


void Asura_CommandConsole::LateInitialise( void )
{	
#ifndef ASURA_COMMAND_CONSOLE_READONLY

#ifdef ASURA_ENABLE_INTERNET_SIMULATION
	AddCmd( CONSOLE_CMDNAME_NETSIM,					ConsoleCmd_NetSim,				REMOVE_AT_SHUTDOWN );
	AddCmd( CONSOLE_CMDNAME_NETSIM_LATENCY,			ConsoleCmd_NetSim_Latency,		REMOVE_AT_SHUTDOWN );
	AddCmd( CONSOLE_CMDNAME_NETSIM_LATENCYSTDDEV,	ConsoleCmd_NetSim_LatencyStdDev,REMOVE_AT_SHUTDOWN );
	AddCmd( CONSOLE_CMDNAME_NETSIM_PACKETLOSS,		ConsoleCmd_NetSim_PacketLoss,	REMOVE_AT_SHUTDOWN );
#endif
#if ASURA_NETWORK_LOGGING && defined(ASURA_SUPPORT_EXTERNAL_NETWORK)// NB not #ifdef
	AddCmd( CONSOLE_CMDNAME_NETWORK_LOGGING,		ConsoleCmd_SetMsgLogLevel,		REMOVE_AT_SHUTDOWN );
#endif
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	Asura_NetTransport_System::AddConsoleCmdsAndVars();
#endif

#ifdef ASURA_SUPPORT_ENTITYWATCH
	Asura_EntityWatch_System::AddConsoleCommands();
#endif

#ifdef ASURA_USE_SCREENSHOT_SYSTEM
	Asura_ScreenShot_System::AddConsoleCommands();
#endif

#ifdef ASURA_USE_CARPET
	Asura_Carpet_MeshManager::AddConsoleCommands();
#endif // #ifdef ASURA_USE_CARPET

#ifdef ASURA_USE_BOTANICALS
	Asura_Botanicals_InstanceManager::AddConsoleCommands();
#endif //  #ifdef ASURA_USE_BOTANICALS

#ifdef ASURA_TOOLS
	AddCmd( CONSOLE_CMDNAME_RELOADANIMS, Tools_Animation_Files::ConsoleCmd_ReloadAllAnimations, REMOVE_AT_SHUTDOWN);
#endif // ASURA_TOOLS

#ifdef ASURA_USE_DEBUG_MESSAGE_TRIGGER
	Asura_ServerEntity_DebugMessageTrigger::AddConsoleCommands();

	#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_ServerEntity_DebugWaypointTrigger::AddConsoleCommands();
	#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

#endif // #ifdef ASURA_USE_DEBUG_MESSAGE_TRIGGER

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
	AddCmd( CONSOLE_CMDNAME_MULTITHREADJOBQUEUES,	ConsoleCmd_MultiThreadJobQueues, REMOVE_AT_SHUTDOWN );
#endif

#endif	// ASURA_COMMAND_CONSOLE_READONLY

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_DebugVariableSystem::LateInitialise();	
#endif
	Asura_ClientEntity_Objective::AddConsoleCommands();

#ifdef ASURA_ENABLE_MEMTRACKER
	Asura_MemTracker::Initialise_ConsoleCommands ();
#endif

#if defined (ASURA_USE_ANIMATION_IK_SOLVER) && defined (ASURA_DEBUG_VARIABLE_SYSTEM)
	Asura_Animation_IK_Chain::AddConsoleVars();
	Asura_Animation_IK_Chain_LOS::AddConsoleVars();
	Asura_Animation_IK_Chain_MarkerPoint::AddConsoleVars();
	Asura_Animation_IK_Solver::AddConsoleVars();
#endif
}

void Asura_CommandConsole_VarRepository::Var::SetData(bool* pbData)
{
	if(m_eType == TYPE_BOOL)
	{
		m_xData.pbData = pbData;
	}
	else 
	{
		Asura_Assert(false, "Mismatch between variable-type and data-type");
	}
}

void Asura_CommandConsole_VarRepository::Var::SetData(int* piData)
{
	if(m_eType == TYPE_INT)
	{
		m_xData.piData = piData;
	}
	else
	{
		Asura_Assert(false, "Mismatch between variable-type and data-type");
	}
}

void Asura_CommandConsole_VarRepository::Var::SetData(u_int* puData)
{
	if(m_eType == TYPE_INT)
	{
		m_xData.piData = reinterpret_cast<int*>(puData);
	}
	else
	{
		Asura_Assert(false, "Mismatch between variable-type and data-type");
	}
}

void Asura_CommandConsole_VarRepository::Var::SetData(float* pfData)
{
	if(m_eType == TYPE_FLOAT)
	{
		m_xData.pfData = pfData;
	}
	else
	{
		Asura_Assert(false, "Mismatch between variable-type and data-type");
	}
}

void Asura_CommandConsole_VarRepository::Var::SetData(Asura_UnicodeChar* wszData)
{
	if(m_eType == TYPE_STRING)
	{
		m_xData.wszData = wszData;
	}
	else
	{
		Asura_Assert(false, "Mismatch between variable-type and data-type");
	}
}

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE
