//////////////////////////////////////////////////////////////////////////////
//
// Asura_CommandConsole_IO.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

// This file's header.
#include "Interface/Asura_CommandConsole_IO.h"

// Stdlib independent.
#include <wchar.h>

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

// Appearance.
static const float CONSOLE_BORDER_X_DEFAULT           = 0.25f;
static const float CONSOLE_BORDER_Y                   = 2.0f;
static const u_int CONSOLE_DEFAULT_COLOUR_BACKGROUND  = 0xC0FFFFFFu;
static const u_int CONSOLE_COLOUR_CURSOR              = 0xFFEF2222u;
static const u_int CONSOLE_COLOUR_TEXT                = 0xFFFFFFFFu;
static const float CONSOLE_CURSOR_TIME_BETWEEN_BLINKS = 0.65f;
static const float CONSOLE_CURSOR_BLINK_DURATION      = 0.2f;
static const float CONSOLE_DEFAULT_FONT_SCALE         = 0.75f;
static const float CONSOLE_FULLY_DOWN_POS_Y           = 400.0f;
static const float CONSOLE_MOVE_DOWN_RATE             = CONSOLE_FULLY_DOWN_POS_Y / 0.5f;
static const u_int CONSOLE_SCROLL_DELTA               = 10;
static const char* CONSOLE_TEXTURE_NAME               = "\\console\\consbk.dds";
static const int   CONSOLE_TEXTURE_NOT_INITD_IDX      = ASURA_TEXTURE_NONE - 1;

Asura_UnicodeChar* Asura_CommandConsole_IO_Default::s_wszCommandOutputBuffer = 0;
u_int Asura_CommandConsole_IO_Default::s_uCommandOutputBufferSize = 0;

#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
float Asura_CommandConsole_IO_Default::s_fDVSBorder = 0.015f;
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_CommandConsole_IO_Default.
//

Asura_CommandConsole_IO_Default::Asura_CommandConsole_IO_Default( void )
{
    m_bDown               = false;
	m_bDisplayInputBuffer = false;
    m_fPosY               = 0.0f;
	m_fDestY              = 0.0f;
    m_uFirstLineToDisplay = 0;

    m_fBorderX            = CONSOLE_BORDER_X_DEFAULT;

    m_iTextureIdx         = CONSOLE_TEXTURE_NOT_INITD_IDX;
	m_uBackgroundColour   = CONSOLE_DEFAULT_COLOUR_BACKGROUND;
    m_iFontHandle		  = ASURA_INVALID_HANDLE;
    m_fFontScale          = CONSOLE_DEFAULT_FONT_SCALE;
	m_szFontName		  = NULL;
   
    m_fTimeUntilNextBlink = CONSOLE_CURSOR_TIME_BETWEEN_BLINKS;
    m_fDurationOfCurBlink = 0.0f;
}

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

DEBUGVAR u_int dbg_uDVSMenuBackgroundColour		= 0x70000000;
DEBUGVAR u_int dbg_uDVSVariableNameColour		= 0xffffffff;
DEBUGVAR u_int dbg_uDVSSelectedVariableColour	= 0xa0ffff00;
DEBUGVAR u_int dbg_uDVSVariableValueColour		= 0xff00ff00;
DEBUGVAR u_int dbg_uDVSReadOnlyVariableColour	= 0xff7f0000;

#if !defined(ASURA_DEBUG_VARIABLE_SYSTEM_FONT_SCALE)
#define ASURA_DEBUG_VARIABLE_SYSTEM_FONT_SCALE (1.0f)
#endif
//
//	RenderQuad - utility function to render background and highlight quads for the DVS
//
static void RenderQuad(float fXPos, float fYPos, float fWidth, float fHeight, u_int uColour)
{
	Asura_Render::Overlay.SetRectXY(fXPos, fYPos, fWidth, fHeight);
	Asura_Render::Overlay.Colour = uColour;
	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();
}



void Asura_CommandConsole_IO_Default::RenderDebugVariableSystem()
{
	if ( Asura_DebugVariableSystem::IsOpen() == false )
	{
		return;
	}

	// Set the font and scale
	CheckFontHandle();
	const float fOldScale = Asura_Font_System::GetFontScale( m_iFontHandle );
	Asura_Font_System::SetFont( m_iFontHandle );		
	Asura_Font_System::SetFontScale( m_iFontHandle, ASURA_DEBUG_VARIABLE_SYSTEM_FONT_SCALE );

	const Asura_UnicodeChar* wszSelectedPath = Asura_DebugVariableSystem::GetSelectedPath();	

	// ARC - 14 March 2007 11:50:11 - New DVS rendering turned on here
	Asura_Render::AllowFullScreenRendering();

	float fXPos = 0.0f;
	float fYPos = 0.0f;
	u_int uLevel = 0;

	const u_int uLevelOfSelectedPath = Asura_StringUtil::CountInstancesOfCharacter( wszSelectedPath, L'.' ) + 1;

	bool bContinue = true;
	while ( bContinue == true )
	{
		RenderDVSMenu( fXPos, fYPos, uLevel );
		if ( wcslen( wszSelectedPath ) != 0 && uLevel < uLevelOfSelectedPath )
		{
			++uLevel;
		}
		else
		{
			bContinue = false;
		}
	}

	Asura_Font_System::SetFontScale( m_iFontHandle, fOldScale );
}

//
//	RenderDVSMenu - renders each 'pane' of the menu by level
//
void Asura_CommandConsole_IO_Default::RenderDVSMenu(float& fXPos, float& fYPos, u_int uLevel)
{
	// Obtain some metrics so the display looks right
	const float fDeadAreaScale = s_fDVSBorder;
	const float fVerticalBorder = Asura_Viewport::GetHeight() * (fDeadAreaScale * 2.0f);
	const float fHorizontalBorder = Asura_Viewport::GetWidth() * fDeadAreaScale;

	if (fXPos < fHorizontalBorder)
	{
		fXPos = fHorizontalBorder;
	}

	if (fYPos < fVerticalBorder)
	{
		fYPos = fVerticalBorder;
	}

	// Font metrics
	const float fFontHeight = Asura_Font_System::GetFontHeight();
	const float fFontScale = Asura_CommandConsole::GetFontScale();
	const float fTextHeight = fFontHeight * fFontScale;
	const float fLastTextLine = static_cast<float>(Asura_Viewport::GetHeight()) - fVerticalBorder - fTextHeight;
	const u_int uMaxLines = Asura_Maths::FloatToInt_Chop((Asura_Viewport::GetHeight() - (fVerticalBorder * 2.0f)) / fTextHeight);

	Asura_UnicodeChar wszNameBuffer[256];
	memset(wszNameBuffer, 0, sizeof(wszNameBuffer));
	Asura_UnicodeChar wszValueBuffer[256];
	memset(wszValueBuffer, 0, sizeof(wszValueBuffer));

	static const Asura_UnicodeChar* wszSubMenuPrompt = L" > ";
	static const Asura_UnicodeChar* wszCommandPrompt = L"...";

	const Asura_UnicodeChar* wszSelection = Asura_DebugVariableSystem::GetSelectedPath();

	Asura_UnicodeChar wszSelectionBuffer[256];
	Asura_Assert(wcslen(wszSelection) < 256, "selected path too long");
	Asura_PrintfToBfr(wszSelectionBuffer, L"%s", 256, true, wszSelection);
	wszSelection = wszSelectionBuffer;

	u_int uPreviousLevelStart = 0;
	u_int uPreviousLevelLength = 0;
	u_int uRootLength = 0;
	if (uLevel > 0)
	{
		Asura_StringUtil::FindStringPartByIndex(wszSelection, L'.', uLevel - 1, uPreviousLevelStart, uPreviousLevelLength);
		uRootLength = uPreviousLevelStart + uPreviousLevelLength;

		// below top level, use '.' in search for similar variables in this category
		// so as to separate 'render' and 'rendering' categories for example PDC 17/5/2007
		Asura_Assert(uRootLength < 256, "variable too long");
		if (uRootLength < 256)
		{
			wszSelectionBuffer[uRootLength] = L'.';
			uRootLength++;
		}
	}
	u_int uSelectionSubstringStart = 0;
	u_int uSelectionSubstringLength = 0;
	Asura_StringUtil::FindStringPartByIndex(wszSelection, L'.', uLevel, uSelectionSubstringStart, uSelectionSubstringLength);

	float fNameWidth = 0.0f;
	float fValueWidth = 0.0f;
	u_int uEntries = 0;
	u_int uSelectedEntry = 0;

	u_int uVariableSubstringStart = 0;
	u_int uVariableSubstringLength = 0;

	for( const Asura_DebugVariableSystem::DVS_Item* pxItem = Asura_DebugVariableSystem::RootItem(); pxItem; pxItem = pxItem->m_pxNext )
	{
		const Asura_UnicodeChar* wszItem = pxItem->m_wszName;
		u_int uLevelOfVariable = Asura_StringUtil::CountInstancesOfCharacter(wszItem, L'.');

		if (Asura_StringUtil::FindStringPartByIndex(wszItem, L'.', uLevel, uVariableSubstringStart, uVariableSubstringLength) == true)
		{
			// Make sure that the string at this level hasn't already been displayed
			if (wcsnicmp(&wszItem[uVariableSubstringStart], wszNameBuffer, uVariableSubstringLength) != 0)
			{
				// Make sure this variable is referenced by the selection
				if (wcsnicmp(wszItem, wszSelection, uRootLength) == 0)
				{
					// Store the index of the selected entry
					if (uVariableSubstringStart == uSelectionSubstringStart &&
						uVariableSubstringLength == uSelectionSubstringLength &&
						wcsnicmp(&wszItem[uVariableSubstringStart], &wszSelection[uVariableSubstringStart], uVariableSubstringLength) == 0)
					{
						uSelectedEntry = uEntries;
					}

					// Account for the terminating NULL in the buffer
					++uVariableSubstringLength;

					Asura_Assert(uVariableSubstringLength < sizeof(wszNameBuffer)/sizeof(wszNameBuffer[0]), "Variable length is huge!");
					if (uVariableSubstringLength > sizeof(wszNameBuffer)/sizeof(wszNameBuffer[0]))
					{
						uVariableSubstringLength = sizeof(wszNameBuffer)/sizeof(wszNameBuffer[0]);
					}
					Asura_PrintfToBfr(wszNameBuffer, L"%s", uVariableSubstringLength, true, &wszItem[uVariableSubstringStart]);

					// Add the value of any variables that are fully qualified
					if (uLevel == uLevelOfVariable)
					{
						if (pxItem->m_pxVariable)
						{
							Asura_DebugVariableSystem::GetVariableValue(pxItem->m_pxVariable, wszValueBuffer, sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]));
						}
						else if (pxItem->m_pxCommand)
						{
							if (uSelectedEntry == uEntries)
							{
								wszValueBuffer[0] = 0;
								s_wszCommandOutputBuffer = wszValueBuffer;
								s_uCommandOutputBufferSize = sizeof(wszValueBuffer) / sizeof(wszValueBuffer[0]);

								if (pxItem->m_pxCommand->GetType() == Asura_CommandConsole_CmdList::CMDTYPE_VOID)
								{
									Asura_UnicodeChar wszHelpCommand[128];
									Asura_PrintfToBfr(wszHelpCommand, L"? %s", 128, true, pxItem->m_pxCommand->GetName() );
									Asura_CommandConsole::Execute(wszHelpCommand, 0, CatchCommandOutputCallback);
								}
								else
								{
									Asura_CommandConsole::Execute(pxItem->m_pxCommand->GetName(), 0, CatchCommandOutputCallback);
								}

								s_wszCommandOutputBuffer = 0;
								s_uCommandOutputBufferSize = 0;
							}
							else
							{
								Asura_PrintfToBfr(wszValueBuffer, wszCommandPrompt, sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true);
							}
						}
						else
						{
							wszValueBuffer[0] = 0;
						}
					}
					else
					{
						Asura_PrintfToBfr(wszValueBuffer, wszSubMenuPrompt, sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true);
					}

					// Calculate the length (in pixels) of the substring and store the maximum length so far encountered (for the width of the background quad)
					float fStringWidth = Asura_Font_System::GetWidthOfString(wszValueBuffer, fFontScale);
					if (fStringWidth > fValueWidth)
					{
						fValueWidth = fStringWidth;
					}
					fStringWidth = Asura_Font_System::GetWidthOfString(wszNameBuffer, fFontScale);
					if (fStringWidth > fNameWidth)
					{
						fNameWidth = fStringWidth;
					}

					++uEntries;
				}
			}
		}
	}

	// Calculate the height of the background quad based on the number of entries
	float fHeight = static_cast<float>(uEntries) * fTextHeight;

	// Reposition the quad (if necessary) to make sure it fits on the screen
	float fWidth = fNameWidth + fValueWidth;
	if (fXPos + fWidth > Asura_Viewport::GetWidth() - fHorizontalBorder)
	{
		fXPos = Asura_Viewport::GetWidth() - fHorizontalBorder - fWidth;

		if (fXPos < fHorizontalBorder)
		{
			fXPos = fHorizontalBorder;
		}
	}

	u_int uItemsToSkip = 0;
	if (fYPos + fHeight > Asura_Viewport::GetHeight() - fVerticalBorder)
	{
		fYPos = Asura_Viewport::GetHeight() - fVerticalBorder - fHeight;

		if (fYPos < fVerticalBorder)
		{
			// Need to chop items out of the display
			uItemsToSkip = Asura_Maths::FloatToInt_Chop( ((fVerticalBorder - fYPos) / fTextHeight) + 1.0f );
			if (uItemsToSkip > uSelectedEntry)
			{
				uItemsToSkip = uSelectedEntry;
			}

			fYPos = fVerticalBorder;
			fHeight = uMaxLines * fTextHeight;
		}
	}

	// Render the background quad
	RenderQuad(fXPos, fYPos, fWidth, fHeight, dbg_uDVSMenuBackgroundColour);

	float fSelectionYPos = 0.0f;
	memset(wszNameBuffer, 0, sizeof(wszNameBuffer));
	memset(wszValueBuffer, 0, sizeof(wszValueBuffer));
	u_int uValuesToSkip = 0;

	for( const Asura_DebugVariableSystem::DVS_Item* pxItem = Asura_DebugVariableSystem::RootItem(); pxItem && fYPos <= fLastTextLine; pxItem = pxItem->m_pxNext )
	{
		const Asura_UnicodeChar* wszItem = pxItem->m_wszName;
		u_int uLevelOfVariable = Asura_StringUtil::CountInstancesOfCharacter(wszItem, L'.');

		if (Asura_StringUtil::FindStringPartByIndex(wszItem, L'.', uLevel, uVariableSubstringStart, uVariableSubstringLength) == true)
		{
			// Make sure that the string at this level hasn't already been displayed
			if (wcsnicmp(&wszItem[uVariableSubstringStart], wszNameBuffer, uVariableSubstringLength) != 0)
			{
				// Make sure this variable is referenced by the selection
				if (wcsnicmp(wszItem, wszSelection, uRootLength) == 0)
				{
					if (uItemsToSkip > 0)
					{
						// Make sure this name is skipped in the next iteration of the loop
						Asura_PrintfToBfr(wszNameBuffer, L"%s", uVariableSubstringLength + 1, true, &wszItem[uVariableSubstringStart]);
						--uItemsToSkip;
					}
					else
					{
						bool bSelected = false;
						// Highlight this entry (if selected)
						if (uVariableSubstringStart == uSelectionSubstringStart &&
							uVariableSubstringLength == uSelectionSubstringLength &&
							wcsnicmp(&wszItem[uVariableSubstringStart], &wszSelection[uVariableSubstringStart], uVariableSubstringLength) == 0)
						{
							bSelected = true;
							fSelectionYPos = fYPos;

							RenderQuad(fXPos, fYPos, fWidth, fTextHeight, dbg_uDVSSelectedVariableColour);

							if (uLevel == uLevelOfVariable && wcsicmp(wszItem, wszSelection) == 0)
							{
								u_int uIntegerScale;
								float fFloatScale;
								Asura_DebugVariableSystem::GetIncrementValues(uIntegerScale, fFloatScale);

								wszValueBuffer[0] = 0;
								if (pxItem->m_pxVariable)
								{
									switch(pxItem->m_pxVariable->m_eType)
									{
										case Asura_CommandConsole_VarRepository::Var::TYPE_INT:
											Asura_PrintfToBfr(wszValueBuffer, L"+/- %i", sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true, uIntegerScale);
											break;
										case Asura_CommandConsole_VarRepository::Var::TYPE_FLOAT:
											Asura_PrintfToBfr( wszValueBuffer, L"+/- %0.3f", sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true, Asura_Maths::FtoD(fFloatScale) );
											break;
										default:
											break;
									}
								}
								else if (pxItem->m_pxCommand)
								{
									switch (pxItem->m_pxCommand->GetType() )
									{
										case Asura_CommandConsole_CmdList::CMDTYPE_VOID:
										{
											Asura_PrintfToBfr( wszValueBuffer, L"+/- To Execute", sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true);
											break;
										}
										case Asura_CommandConsole_CmdList::CMDTYPE_BOOL:
										{
											Asura_PrintfToBfr( wszValueBuffer, L"+ On, - Off", sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true);
											break;
										}
										default:
										{
											break;
										}
									}
								}

								if (wcslen(wszValueBuffer) > 0 && (!pxItem->m_pxVariable || pxItem->m_pxVariable->m_bReadOnly == false))
								{
									float fIncrementWidth = Asura_Font_System::GetWidthOfString(wszValueBuffer, fFontScale);
									float fIncrementXPos = static_cast<float>(Asura_Viewport::GetWidth()) - fHorizontalBorder - fIncrementWidth;
									RenderQuad(fIncrementXPos, fVerticalBorder, fIncrementWidth, fTextHeight, dbg_uDVSMenuBackgroundColour);
									Asura_Font_System::Print(fIncrementXPos, fVerticalBorder, fFontScale, dbg_uDVSVariableNameColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED, wszValueBuffer);
								}
							}
						}

						// Account for the terminating NULL in the buffer
						++uVariableSubstringLength;

						Asura_Assert(uVariableSubstringLength < sizeof(wszNameBuffer)/sizeof(wszNameBuffer[0]), "Variable length is huge!");
						if (uVariableSubstringLength > sizeof(wszNameBuffer)/sizeof(wszNameBuffer[0]))
						{
							uVariableSubstringLength = sizeof(wszNameBuffer)/sizeof(wszNameBuffer[0]);
						}
						Asura_PrintfToBfr(wszNameBuffer, L"%s", uVariableSubstringLength, true, &wszItem[uVariableSubstringStart]);

						// Print the variable
						Asura_Font_System::Print(fXPos, fYPos, fFontScale, dbg_uDVSVariableNameColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED, wszNameBuffer);

						// Print the value of any variables that are fully qualified
						if (uLevel == uLevelOfVariable)
						{
							u_int uColour = dbg_uDVSVariableValueColour;
							if (pxItem->m_pxVariable)
							{
								Asura_DebugVariableSystem::GetVariableValue(pxItem->m_pxVariable, wszValueBuffer, sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]));
								uColour = (pxItem->m_pxVariable->m_bReadOnly == true) ? dbg_uDVSReadOnlyVariableColour : dbg_uDVSVariableValueColour;
							}
							else if (pxItem->m_pxCommand)
							{
								if (bSelected)
								{
									wszValueBuffer[0] = 0;
									s_wszCommandOutputBuffer = wszValueBuffer;
									s_uCommandOutputBufferSize = sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]);

									if (pxItem->m_pxCommand->GetType() == Asura_CommandConsole_CmdList::CMDTYPE_VOID)
									{
										Asura_UnicodeChar wszHelpCommand[128];
										Asura_PrintfToBfr(wszHelpCommand, L"? %s", 128, true, pxItem->m_pxCommand->GetName() );
										Asura_CommandConsole::Execute(wszHelpCommand, 0, CatchCommandOutputCallback);
									}
									else
									{
										Asura_CommandConsole::Execute(pxItem->m_pxCommand->GetName(), 0, CatchCommandOutputCallback);
									}

									s_wszCommandOutputBuffer = 0;
									s_uCommandOutputBufferSize = 0;
								}
								else
								{
									Asura_PrintfToBfr(wszValueBuffer, wszCommandPrompt, sizeof(wszValueBuffer)/sizeof(wszValueBuffer[0]), true);
								}
							}
							else
							{
								wszValueBuffer[0] = 0;
							}
							if (uValuesToSkip)
							{
								--uValuesToSkip;
							}
							else if ( wszValueBuffer[0])
							{
								if (bSelected)
								{
									u_int uNewLines = Asura_StringUtil::CountInstancesOfCharacter(wszValueBuffer, L'\n');
									if (uNewLines)
									{
										RenderQuad(fXPos + fNameWidth, fYPos + fTextHeight, fValueWidth, fTextHeight * static_cast<float>(uNewLines), dbg_uDVSSelectedVariableColour);
										uValuesToSkip = uNewLines;
									}
								}
								Asura_Font_System::Print(fXPos + fNameWidth, fYPos, fFontScale, uColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED, wszValueBuffer);
							}
						}
						else
						{
							if (uValuesToSkip)
							{
								--uValuesToSkip;
							}
							else
							{
								Asura_Font_System::Print(fXPos + fNameWidth, fYPos, fFontScale, dbg_uDVSVariableNameColour, ASURA_TEXTFORMAT_LEFTJUSTIFIED, wszSubMenuPrompt);;
							}
						}


						fYPos += fTextHeight;
					}
				}
			}
		}
	}

	Asura_Render::FlushDisplay();

	// Set the positions up for the next level of menu
	fXPos += fWidth;
	fYPos = fSelectionYPos;
}

#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

void Asura_CommandConsole_IO_Default::CatchCommandOutputCallback( const Asura_UnicodeChar* wszOutput )
{
	if (!s_wszCommandOutputBuffer)
	{
		return;
	}
	if (s_wszCommandOutputBuffer[0])
	{
		Asura_PrintfToBfr(s_wszCommandOutputBuffer + wcslen(s_wszCommandOutputBuffer), L"\n%s", s_uCommandOutputBufferSize - wcslen(s_wszCommandOutputBuffer), true, wszOutput);
	}
	else
	{
		Asura_PrintfToBfr(s_wszCommandOutputBuffer, L"%s", s_uCommandOutputBufferSize, true, wszOutput);
	}
}

void Asura_CommandConsole_IO_Default::Render( void )
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	RenderDebugVariableSystem();
#endif

    Asura_Assert( m_fPosY >= 0.0f, "" );
    if( !m_fPosY ) { return; }

    if( CONSOLE_TEXTURE_NOT_INITD_IDX == m_iTextureIdx )
    {
        bool bWasNew;
        m_iTextureIdx = Asura_Texture_Management::AddTexture( CONSOLE_TEXTURE_NAME, &bWasNew );
        if( bWasNew )
        {
            // Ensure texture is loaded before we first render with it.
            Asura_Texture_Management::LoadTextures();
			// (MH - 04.09.2003) if no texture is present, use texture none, instead 
			// of repeatedly retrying.
			if (!Asura_Texture_Management::IsTextureLoaded(m_iTextureIdx))
			{
				Asura_Texture_Management::DestroyTexture(m_iTextureIdx);
				m_iTextureIdx = ASURA_TEXTURE_NONE;
			}
        }
    }
        
    // Draw background.
    {       
        Asura_Render::Overlay.Colour           = m_uBackgroundColour;
        Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
        Asura_Render::Overlay.TextureIndex     = m_iTextureIdx;

        Asura_Render::Overlay.SetRectXY( 0.0f, 0.0f, static_cast<float>(Asura_Viewport::GetWidth()), m_fPosY );

        const float fProportionVisible = m_fPosY / Asura_Viewport::GetHeight();
        Asura_Render::Overlay.SetRectUV( 0.0f, 1.0f - fProportionVisible, 1.0f, fProportionVisible );

        Asura_Render::Overlay.Render();
    }

    Asura_Render::FlushDisplay();

    const Asura_UnicodeChar* wszInput;
    u_int uCursorPos;
    Asura_CommandConsole::GetInputBfrInfo( &wszInput, &uCursorPos );

    const Asura_UnicodeChar* const * pwszOutput;
    u_int uFirstOutputLine, uNumOutputLines;
    Asura_CommandConsole::GetOutputBfrInfo( &pwszOutput, &uFirstOutputLine, &uNumOutputLines );

    // Render text.
    {
        CheckFontHandle();
		Asura_Font_System::SetFont( m_iFontHandle );

        const float fWidthOfSpace = Asura_Font_System::GetWidthOfChar( L' ', m_fFontScale );
        const float fFontHeight   = Asura_Font_System::GetFontHeight() * m_fFontScale;

        float fY = m_fPosY - (CONSOLE_BORDER_Y + fFontHeight);
        if( fY + fFontHeight >= 0.0f )
        {
            
			if(m_bDisplayInputBuffer)
			{
				// Output the input buffer.
				Asura_Font_System::Print( m_fBorderX * fWidthOfSpace,
										fY,
										m_fFontScale,
										CONSOLE_COLOUR_TEXT,
										ASURA_TEXTFORMAT_LEFTJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING,
										ASURA_COMMAND_CONSOLE_CURSOR_PROMPT L"%s",
										wszInput );

				// Draw cursor, if necessary.
				if( m_fDurationOfCurBlink )
				{
					// Calculate cursor pos.
					float fCursorX = m_fBorderX * fWidthOfSpace
								+ Asura_Font_System::GetWidthOfString(ASURA_COMMAND_CONSOLE_CURSOR_PROMPT, m_fFontScale)
								+ Asura_Font_System::GetWidthOfString(wszInput, m_fFontScale, uCursorPos);

					// Output.
					Asura_Font_System::Print( fCursorX, fY, m_fFontScale, CONSOLE_COLOUR_CURSOR, ASURA_TEXTFORMAT_LEFTJUSTIFIED, L"|" );
				}
			}

            // And now the console buffer.
            if( uNumOutputLines )
            {
                const u_int uNumLinesToDisplay = Asura_Maths::Min(uNumOutputLines, CONSOLE_NUM_LINES_TO_DISPLAY) - 1;

                // Start from last line to display.
                u_int uLine = uFirstOutputLine + m_uFirstLineToDisplay + uNumLinesToDisplay;
                if( uLine >= ASURA_COMMAND_CONSOLE_MAX_LINES )
                {
                    uLine -= ASURA_COMMAND_CONSOLE_MAX_LINES;
                }

                // Work up until we go offscreen or we've displayed the
                // first line. Note that we leave a one-line gap between
                // the bottom of the console buffer and the input line.
				if(m_bDisplayInputBuffer) fY -= 2.0f * fFontHeight;
				for( ; fY >= 0.0f; fY -= fFontHeight )
                {
                    Asura_Font_System::Print( m_fBorderX * fWidthOfSpace,
                                              fY,
                                              m_fFontScale,
                                              CONSOLE_COLOUR_TEXT,
                                              ASURA_TEXTFORMAT_LEFTJUSTIFIED,
                                              pwszOutput[uLine]);
                    
                    // Stop if we've just reached the front of the buffer.
                    if( uLine == uFirstOutputLine ) { break; }

                    uLine = ( uLine ? uLine - 1 : ASURA_COMMAND_CONSOLE_MAX_LINES - 1 );
                }
            }
        }
    }
    // End text render.	

    Asura_Render::FlushDisplay();
}

void Asura_CommandConsole_IO_Default::Reset( void )
{
    //TODO probably need to refresh font handle and texture idx
	if (m_iTextureIdx != ASURA_TEXTURE_NONE)
	{
		m_iTextureIdx = CONSOLE_TEXTURE_NOT_INITD_IDX;
	}

	m_iFontHandle = ASURA_INVALID_HANDLE;

	m_bDown = false;
	m_fPosY = 0.0f;
	m_fDestY = 0.0f;
}

void Asura_CommandConsole_IO_Default::Scroll( bool bDown, bool bToEnd )
{
    u_int uNumLines;
    Asura_CommandConsole::GetOutputBfrInfo( NULL, NULL, &uNumLines );

    int iDelta;
    if( bDown )
    {
        // Calculate the last line idx that we can set to be the first line
        // to be displayed.
        u_int uLastForFirst = ( uNumLines > CONSOLE_NUM_LINES_TO_DISPLAY ? uNumLines - CONSOLE_NUM_LINES_TO_DISPLAY : 0 );
        
        Asura_Assert( m_uFirstLineToDisplay <= uLastForFirst, "" );
        iDelta = Asura_Maths::Min( CONSOLE_SCROLL_DELTA, uLastForFirst - m_uFirstLineToDisplay );
    }
    else
    {
        iDelta = -(static_cast<int>(Asura_Maths::Min(CONSOLE_SCROLL_DELTA, m_uFirstLineToDisplay)));
    }

    m_uFirstLineToDisplay += iDelta;
}

void Asura_CommandConsole_IO_Default::SetDown( bool bDown, bool bInstantTransition )
{
    m_bDown = bDown;
	
	float fFullyDown = Asura_Maths::Min( CONSOLE_FULLY_DOWN_POS_Y, static_cast<float>(Asura_Viewport::GetHeight()) );
	m_fDestY = ( bDown ? fFullyDown : 0.0f ); 

    if( bInstantTransition ) 
	{ 
		m_fPosY = m_fDestY; 
	}

	if(bDown) m_bDisplayInputBuffer = true;
}

void Asura_CommandConsole_IO_Default::SetPartiallyDown(u_int uNumLines, bool bInstantTransition)
{
	const Asura_Font* pxFont = Asura_Font_System::GetFont(m_iFontHandle);
	if(!pxFont) return;

	m_bDown = false;

	if(uNumLines)
	{
		m_fDestY = pxFont->GetFontHeight()  * static_cast<float>(uNumLines) * m_fFontScale;
		m_fDestY += CONSOLE_BORDER_Y;
	}
	else
	{
		m_fDestY = 0.0f;
	}
	if(bInstantTransition)
	{
		m_fPosY = m_fDestY;
	}
	if(uNumLines) m_bDisplayInputBuffer = false;

}

void Asura_CommandConsole_IO_Default::Update( void )
{
    const float fFrameTime = Asura_Timers::GetFrameTime();

    // Update position if necessary.
	m_fDestY = Asura_Maths::Min(m_fDestY, static_cast<float>(Asura_Viewport::GetHeight()));
	if(m_fPosY > m_fDestY)
	{
		m_fPosY = Asura_Maths::Max(m_fDestY,m_fPosY - fFrameTime * CONSOLE_MOVE_DOWN_RATE);
	}
	else if(m_fPosY < m_fDestY)
	{
		m_fPosY = Asura_Maths::Min(m_fDestY,m_fPosY + fFrameTime * CONSOLE_MOVE_DOWN_RATE);
	}


    // If visible, update state.
    if( m_fPosY > 0.0f )
    {
        // Cursor blink.
        if( m_fDurationOfCurBlink )
        {
            m_fDurationOfCurBlink -= fFrameTime;
            if( m_fDurationOfCurBlink <= 0.0f )
            {
                m_fDurationOfCurBlink = 0.0f;
                m_fTimeUntilNextBlink = CONSOLE_CURSOR_TIME_BETWEEN_BLINKS;
            }
        }
        else
        {
            Asura_Assert( m_fTimeUntilNextBlink > 0.0f, "" );
            m_fTimeUntilNextBlink -= fFrameTime;

            if( m_fTimeUntilNextBlink <= 0.0f )
            {
                m_fDurationOfCurBlink = CONSOLE_CURSOR_BLINK_DURATION;
                m_fTimeUntilNextBlink = 0.0f;
            }
        }
    }
}


void Asura_CommandConsole_IO_Default::SetFont( const char* szFontName )
{
	m_szFontName = szFontName;
	// force refresh of handle
	m_iFontHandle = ASURA_INVALID_HANDLE;
}

bool Asura_CommandConsole_IO_Default::SetFont( Asura_Handle iHandle )
{
	if (Asura_Font_System::DoesFontHandleExist( iHandle ))
	{
		m_iFontHandle = iHandle; 
		return true;
	}
	
	Asura_Assert(false, "invalid font handle");
	return false;
}

const Asura_UnicodeChar* Asura_CommandConsole_IO_Default::GetLineSplitPoint(const Asura_UnicodeChar* wszLine)
{
    float fTemp;
	int iTemp;

	CheckFontHandle();
    Asura_Font_System::SetFont( m_iFontHandle );

    const float fWidthOfSpace = Asura_Font_System::GetWidthOfChar( L' ', m_fFontScale );

    return Asura_Font_System::FindSplitPoint(wszLine, 
        Asura_Viewport::GetWidth() - 2.f * m_fBorderX * fWidthOfSpace, m_fFontScale, fTemp, iTemp);
}


void Asura_CommandConsole_IO_Default::CheckFontHandle( void )
{
	if (m_iFontHandle == ASURA_INVALID_HANDLE)
	{
		if (m_szFontName)
		{
			m_iFontHandle = Asura_Font_System::GetFontHandle(m_szFontName);
			if (m_iFontHandle == ASURA_INVALID_HANDLE)
			{
				m_iFontHandle = 0;
			}
		}
		else
		{
			// use default
			m_iFontHandle = 0;
		}
	}
}

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE
