//
// Asura_Gadget.cpp
//
// Created by Kevin Lea - 12/06/2001 11:11:09
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_TOOLS

#include "Core/Asura_Windows.h"
#include "Interface/Asura_Gadget.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Rendering/Asura_Text_Output.h"
#include "Core/Asura_UserInput.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Overlay_System.h"
#include "Core/Asura_String.h"
#include "Rendering/Asura_Render.h"
#include "Debugging/Asura_Assert.h"
#include "Core/Asura_Timers.h"

//tools
#include "Interface/Tools_ContextMenu_System.h"



extern unsigned int Asura_OverlayBrightness;



//////////////////////////////////////////////////////////////////////////////////////////////////


Asura_Gadget::Asura_Gadget( void )
{
	m_pxNextGadget          = NULL;	
	m_iFocusFlags          = 0;
    m_bClickedThisFrame = false;
    m_bHighlight        = false;
    m_bIsClickable      = true;
    m_bIsEnabled        = true;
    m_bIsMouseOver      = false;
	m_bIsHidden			= false;
	m_bUseMixedGadgetColour = false;
	m_bUseSharedGadgetColour = false;
	m_pxDialog 			= 0;
	m_iRadioButtonGroup = 0;
	m_bRadioButton_AtLeastOneMustBePressed = false;
	m_bExpandWidthToFillOverlay = false;
	m_fKeyRepeatTimer = 0.0f;
	m_iRepeatKey = ASURA_KEYID_VOID;

}

Asura_Gadget::~Asura_Gadget()
{
	if(m_iFocusFlags)
	{
		Asura_Assert(!m_iFocusFlags,"Keyboard focus should have been lost before this gadget was destroyed");
		Asura_Overlay_System::ReleaseFocus();//hopefully preventing a crash from this happening.
	}

	if (m_pxNextGadget) delete m_pxNextGadget;
}

void Asura_Gadget::CheckCursorPosition( const Asura_Vector_2& xOrigin )
{
    if( IsEnabled() )
    {
		if(Asura_Overlay_System::AGadgetHasMouseFocus())
		{
			m_bIsMouseOver = Asura_Overlay_System::GetFocusGadget() == this;
		}
		else
		{
			Asura_Vector_2 xPos = xOrigin + m_xOffset;
			m_bIsMouseOver = Asura_Cursor::Inside( xPos, GetWidthForHitTest(), m_fHeight );

			if(m_bIsMouseOver && m_pxDialog->GetParentSubwindow() && !m_pxDialog->CanRenderOutsideOfParentSubwindow())
			{
				//gadget lives in a subwindow control , so make sure the mouse is within that as well
				m_bIsMouseOver = m_pxDialog->GetParentSubwindow()->IsCursorWithinChildRect();
			}
		}
    }
    else
    {
        m_bIsMouseOver = false;
    }
}

u_int Asura_Gadget::GetColour( bool bClicked ) const
{
    u_int uColour;

    if( IsEnabled() )
    {
		if( bClicked )
		{
			uColour = ASURA_GADGET_CLICKED_COLOUR;
		}
		else
		{
            uColour = ( m_bIsMouseOver || m_bHighlight ? ASURA_GADGET_HIGHLIGHT_COLOUR : ASURA_GADGET_DEFAULT_COLOUR );
		}
    }
    else
    {
        uColour = ASURA_GADGET_DISABLED_COLOUR;
    }

	if(m_bUseMixedGadgetColour && uColour == ASURA_GADGET_DEFAULT_COLOUR)
	{
		uColour = ASURA_GADGET_MIXED_COLOUR;
	}
	else if(m_bUseSharedGadgetColour && uColour == ASURA_GADGET_DEFAULT_COLOUR)
	{
		uColour = ASURA_GADGET_SHARED_COLOUR;
	}

    return uColour;
}

void Asura_Gadget::Handle( const Asura_Vector_2& xOrigin )
{
	CheckCursorPosition( xOrigin );

    m_bClickedThisFrame = false;
	Update( xOrigin );

    if( m_pxNextGadget ) { m_pxNextGadget->Handle( xOrigin ); }
}

void Asura_Gadget::SetEnabledFlag( bool bEnabled, bool bAffectList /*= true*/ )
{
    m_bIsEnabled = bEnabled;

    if( bAffectList && m_pxNextGadget )
    {
        m_pxNextGadget->SetEnabledFlag( bEnabled );
    }
}

void Asura_Gadget::RenderList( const Asura_Vector_2& xOrigin )
{
	if(m_pxNextGadget)
	{
		m_pxNextGadget->RenderList(xOrigin);
	}
	if(!m_bIsHidden)
	{
		Render(xOrigin);
	}
}

void Asura_Gadget::Render( const Asura_Vector_2& xOrigin )
{
    // We won't necessarily call Handle() each frame so we need to reset this
    // after rendering in case the user has moved the mouse off the overlay
    // before the next update.
    m_bClickedThisFrame = false;
    m_bIsMouseOver      = false;
}

bool Asura_Gadget::IsPointInGadget(const Asura_Vector_2& xRelativePos) const
{
	if( m_xOffset.x <= xRelativePos.x && m_xOffset.x + GetWidthForHitTest()>xRelativePos.x &&
		m_xOffset.y <= xRelativePos.y && m_xOffset.y + m_fHeight>xRelativePos.y)
	{
		return true;
	}
	return false;
}

char Asura_Gadget::GetAsciiKeyPressed()
{
	
	//ignore key presses while control is held
	if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL)) return 0;
	
	//scan through all the keys that have been pressed for those that we want to bother with
	//(this is far from exhaustive , and basically a rubbish approach =) )
	int iKeyPressed = 0;
	bool shift = Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT);
	int i;
	for(i=0;i<MAX_NUMBER_OF_ASURA_KEYIDS && !iKeyPressed;i++)
	{
		if(!Asura_UserInput::DebouncedKeyPressed(i)) continue;
		switch(i)
		{
			case ASURA_KEYID_SPACE :
				iKeyPressed=' ';
				break;
			
			case ASURA_KEYID_NUMPADDEL : 
				iKeyPressed='.';
				break;
			
			case ',' :
				if(shift) iKeyPressed='<';
				else iKeyPressed=',';
				break;
			
			case '.' :
				if(shift) iKeyPressed='>';
				else iKeyPressed='.';
				break;

			case ';' :
				if(shift) iKeyPressed=':';
				else iKeyPressed=';';
				break;

			case '-' :
				if(shift) iKeyPressed= '_';
				else iKeyPressed='-';
				break;

			case '=' :
				if(shift) iKeyPressed= '+';
				else iKeyPressed= '=';
				break;

			case ASURA_KEYID_NUMPADMULTIPLY :
				iKeyPressed='*';
				break;

			case ASURA_KEYID_NUMPADADD :
				iKeyPressed='+';
				break;

			case ASURA_KEYID_NUMPADSUB :
				iKeyPressed='-';
				break;


			case ASURA_KEYID_NUMPADDIVIDE :
				iKeyPressed='/';
				break;


			case '\\' :
				iKeyPressed='\\';
				break;
			
			case '[' :
				if(shift) iKeyPressed='{';
				else iKeyPressed='[';
				break;
			
			case ']' :
				if(shift) iKeyPressed='}';
				else iKeyPressed=']';
				break;

		
			default :
				{
					if(i>='0' && i<='9')
					{
						if(shift)
						{
							//assumes UK keyboard at the moment
							char szSymbols[] = ")!\"$%^&*(";
							iKeyPressed = szSymbols[i-'0'];

						}
						else						
						{
							iKeyPressed = i;
						}

					}	
					else if(i>=ASURA_KEYID_NUMPAD0 && i<=ASURA_KEYID_NUMPAD9)
					{
						iKeyPressed='0'+i-ASURA_KEYID_NUMPAD0;
					}
					else if(i>='a' && i<='z')
					{
						if(shift)
						{
							iKeyPressed = 'A'+i-'a';
						}
						else
						{
							iKeyPressed = i;
						}
					}
				}
		}
	}

	return (char) iKeyPressed;

}


void Asura_Gadget::MoveGadgetRelative(const Asura_Vector_2& xDiff)
{
	m_xOffset += xDiff;
}

Asura_Overlay* Asura_Gadget::GetOverlay() const
{
	return m_pxDialog->m_pxOverlay;
}

void Asura_Gadget::RadioButtonPressed(const class Asura_Gadget* pxPressedGadget)
{
	if(!pxPressedGadget->m_iRadioButtonGroup)
	{
		Asura_Assert(0,"Gadget didn't have a radio button group");
		return;
	}
	if(m_iRadioButtonGroup != pxPressedGadget->m_iRadioButtonGroup) return;

	if(this == pxPressedGadget)
	{
		//This is the button that was pressed
		if(m_bRadioButton_AtLeastOneMustBePressed)
		{
			//If a button must always be pressed , then make sure this one is
			SetChecked(true);
		}
		else
		{
			SetChecked(!IsChecked());
		}
	}
	else
	{
		//everything else in the group must be unpressed
		SetChecked(false);
	}
}

float Asura_Gadget::GetWidthForHitTest() const
{
	if(m_bExpandWidthToFillOverlay)
	{
		const Asura_Overlay* pxOverlay = GetOverlay();
		Asura_Assert(pxOverlay,"");

		return pxOverlay->m_fWidth - m_xOffset.x - 1.0f;
	}
	else
	{
		return m_fWidth;
	}
}

bool Asura_Gadget::IsRepeatableKeyPressed(int iKey)
{
	if(Asura_UserInput::DebouncedKeyPressed(iKey))
	{
		//debounced presses always accepted. Start timer before next press
		m_fKeyRepeatTimer = 0.3f;
		m_iRepeatKey = iKey;
		return true;
	}
	else if(iKey == m_iRepeatKey)
	{
		if(Asura_UserInput::KeyPressed(iKey))
		{
			//key still being held down ,see if it is time for a repeat
			m_fKeyRepeatTimer -= Asura_Timers::GetFrameTime();
			if(m_fKeyRepeatTimer <= 0.0f)
			{
				m_fKeyRepeatTimer = 0.08f;
				return true;	
			}
		}
		else
		{
			m_iRepeatKey = ASURA_KEYID_VOID;
		}
	}
	return false;
}



//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_BaseText_Gadget::Asura_BaseText_Gadget(const Asura_Vector_2& xOffset, int iMaxLength, bool* pbFlag,bool bDefaultFocus)
{
	// JEB - 31 August 2006 11:19:09 - Increased maximum text length
	if(iMaxLength > 255) iMaxLength = 255;

	m_xOffset = xOffset;
	m_iMaxLength = iMaxLength;
	m_iBoxLength = iMaxLength;
	m_iBoxLengthOverride = 0;
	m_iCursorPos = 0;
	m_iSelectionStart = 0;
	m_bDefaultFocus = bDefaultFocus;
	m_bCentreJustify = false;
	m_pbFlag = pbFlag;
	m_bDiscardChanges = false;
	m_bReadOnly = false;
	m_bUpdateImmediatelyWhileTyping = false;


	m_fWidth = (float)m_iBoxLength;
	m_fHeight = 1.0f;

	m_pszTextBuffer = new char[m_iMaxLength+1];
	m_pszTextBuffer[0] = 0;
}

Asura_BaseText_Gadget::~Asura_BaseText_Gadget()
{
	delete [] m_pszTextBuffer;
}

void Asura_BaseText_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 xPos = xOrigin + m_xOffset;
	if(m_bCentreJustify) xPos.x += m_iBoxLength*.5f;

	
	unsigned int uColour = GetColour(m_bClickedThisFrame);

	if(m_iFocusFlags & OverlayFocusType_Keyboard)
	{
        // Shouldn't have focus if disabled.
        Asura_Assert( IsEnabled(), "" );

		RenderString(xPos,Asura_OverlayBrightness|ASURA_GADGET_HIGHLIGHT_COLOUR,true);
	}
	else
	{
		u_int uTextColour = uColour;
		
		if(uTextColour == ASURA_GADGET_DEFAULT_COLOUR && IsMixedValue())
		{
			uTextColour = ASURA_GADGET_MIXED_COLOUR;
		}
		
		FillTextBuffer();
		RenderString(xPos,Asura_OverlayBrightness|uTextColour,false);
	}


	// JEB - 31 August 2006 11:19:33 - Increased maximum text length
	char szBoxString[255 + 3];
	int i;

	szBoxString[0] = 2;
	for(i=1; i<m_iBoxLength; i++)
	{
		szBoxString[i] = 3;
	}
	szBoxString[i++] = 4;
	szBoxString[i] = 0;

	if(uColour == ASURA_GADGET_DEFAULT_COLOUR && IsSharedGadget())
	{
		uColour = ASURA_GADGET_SHARED_COLOUR;
	}

	xPos = xOrigin + m_xOffset;
	xPos.x += m_iBoxLength*0.5f;
	Asura_Text_Output::Print(xPos.x,xPos.y,Asura_OverlayBrightness|uColour,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,szBoxString);

    Asura_Gadget::Render( xOrigin );
}

void Asura_BaseText_Gadget::RenderString(const Asura_Vector_2& xPos,u_int uColour,bool bCursor)
{
	int iLength = strlen(m_pszTextBuffer);
	int iOffset = 0;
	if(iLength > m_iBoxLength)
	{
		//string is longer than can be displayed , work out first character to show
		iOffset = iLength - m_iBoxLength;

		//make sure cursor , and some before ,is included
		iOffset = Asura_Maths::Min(iOffset,m_iCursorPos-3);
		if(iOffset<0) iOffset = 0;		
	}
	
	char cCopy = 0;
	if(iLength - iOffset > m_iBoxLength)
	{
		//remainder of string is too long , so put in a temporary terminator
		cCopy = m_pszTextBuffer[iOffset+m_iBoxLength];
		m_pszTextBuffer[iOffset+m_iBoxLength] = 0;
	}

	int iLengthShown = Asura_Maths::Min(iLength,m_iBoxLength);

	if(bCursor && m_iSelectionStart != m_iCursorPos)
	{
		//draw selection highlight

		//get range to highlight
		int m_iLeft = Asura_Maths::Min(m_iSelectionStart,m_iCursorPos);
		int m_iRight = Asura_Maths::Max(m_iSelectionStart,m_iCursorPos);

		//offset and clamp
		m_iLeft -= iOffset;
		m_iRight -= iOffset;
		if(m_iLeft<0) m_iLeft = 0;
		if(m_iRight>m_iBoxLength) m_iRight = m_iBoxLength;

		//get the screen position for these
		float fLeft= xPos.x+m_iLeft;
		float fRight = xPos.x+m_iRight;

		if(m_bCentreJustify)
		{
			fLeft -= iLengthShown/2.0f;
			fRight -= iLengthShown/2.0f;
		}

		//convert from print coords
		fLeft *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
		fRight *= ASURA_TEXT_OUTPUT_FONT_WIDTH;

		Asura_Render::Overlay.Colour = 0xafffffff;
		Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
		Asura_Render::Overlay.SetRectXY(fLeft,xPos.y*ASURA_TEXT_OUTPUT_FONT_HEIGHT,fRight-fLeft,ASURA_TEXT_OUTPUT_FONT_HEIGHT);
		Asura_Render::Overlay.Render();
		Asura_Render::FlushDisplay();


	}
	
	//render string
	int iJustifyFlag = m_bCentreJustify ? ASURA_TEXT_FORMAT_CENTREJUSTIFIED : ASURA_TEXT_FORMAT_LEFTJUSTIFIED;
	Asura_Text_Output::Print(xPos.x,xPos.y,uColour,iJustifyFlag,"%s",m_pszTextBuffer+iOffset);

	if(iLength - iOffset > m_iBoxLength)
	{
		//remove terminator
		m_pszTextBuffer[iOffset+m_iBoxLength] = cCopy;
	}

	//draw cursor
	if(bCursor)
	{
		if( Asura_Maths::FloatToInt( Asura_Timers::GetLocalUnscaledTime() * 4.0f ) % 2 )
		{
						
			float fPosition = xPos.x+(m_iCursorPos-iOffset)-0.5f;
			if(m_bCentreJustify) fPosition -= iLengthShown/2.0f;
			Asura_Text_Output::Print(fPosition,xPos.y,Asura_OverlayBrightness|0xffffff,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"|");
		}
	}
}

void Asura_BaseText_Gadget::Update( const Asura_Vector_2& xOrigin )
{	
	if(m_bDefaultFocus)
	{
		//grab keyboard focus
		Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Keyboard);
		m_bDefaultFocus = false;
	}

	if( m_bIsMouseOver && m_bIsClickable && !m_bReadOnly && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
		Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Keyboard);
        m_bClickedThisFrame = true;
	}

	if(m_iFocusFlags)
	{
		if(m_bReadOnly)
		{
			m_bDiscardChanges = true;
			Asura_Overlay_System::ReleaseFocus();
		}
		else
		{
			HandleKeyboardInput();
		}
	}
}

void Asura_BaseText_Gadget::TakeTabFocus()
{
	//grab keyboard focus
	Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Keyboard);
}

void Asura_BaseText_Gadget::FocusGained(int focusType)
{
	Asura_Gadget::FocusGained(focusType);

	
	if(focusType & OverlayFocusType_Keyboard)
	{
		FillTextBuffer();
		
		m_bDiscardChanges = false;
		m_iCursorPos = strlen(m_pszTextBuffer);
		m_iSelectionStart = 0;
	}
		
}

void Asura_BaseText_Gadget::FocusLost(int focusType)
{
	Asura_Gadget::FocusLost(focusType);

	
	if(focusType & OverlayFocusType_Keyboard)
	{
		if(!m_bDiscardChanges)
		{
			if(ReadFromTextBuffer())
			{
				if(m_pbFlag) *m_pbFlag = true;
			}
		}
	}
}


void Asura_BaseText_Gadget::HandleKeyboardInput()
{
	Asura_Assert(m_iFocusFlags,"");

	int iLength = strlen(m_pszTextBuffer);

	bool bChanged = false;
	bool bContinueCheckingInput = true;

	//check for copy and paste
	if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
	{
		if(Asura_UserInput::DebouncedKeyPressed('c'))
		{
			CopyText();
			bContinueCheckingInput = false;
		}
		if(Asura_UserInput::DebouncedKeyPressed('x'))
		{
			CopyText();
			DeleteSelection();
			bContinueCheckingInput = false;
		}
		if(Asura_UserInput::DebouncedKeyPressed('v'))
		{
			PasteText();
			bContinueCheckingInput = false;
			bChanged = true;
		}
	}
	if(bContinueCheckingInput)
	{
		if(IsRepeatableKeyPressed(ASURA_KEYID_LEFT))
		{
			//move cursor left
			m_iCursorPos--;
			if(m_iCursorPos<0)m_iCursorPos = 0;
			
			//clear selection if shift not held
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT)) m_iSelectionStart = m_iCursorPos;
		}
		else if(IsRepeatableKeyPressed(ASURA_KEYID_RIGHT))
		{
			
			//move cursor right
			m_iCursorPos++;
			if(m_iCursorPos>iLength)m_iCursorPos = iLength;

			//clear selection if shift not held
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT)) m_iSelectionStart = m_iCursorPos;
		}
		else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_HOME))
		{
			//move cursor to start of string
			m_iCursorPos = 0;
			//clear selection if shift not held
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT)) m_iSelectionStart = m_iCursorPos;
		}
		else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_END))
		{
			//move cursor to end of string
			m_iCursorPos = iLength;
			//clear selection if shift not held
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT)) m_iSelectionStart = m_iCursorPos;
		}
		else if(IsRepeatableKeyPressed(ASURA_KEYID_DEL))
		{
			if(IsTextSelected())
			{
				DeleteSelection();
			}
			else
			{
				//delete the character at the cursor
				for(int i=m_iCursorPos;i<iLength;i++)
				{
					m_pszTextBuffer[i] = m_pszTextBuffer[i+1];
				}
			}
			bChanged = true;
		}
		else if(IsRepeatableKeyPressed(ASURA_KEYID_BACKSPACE))
		{
			if(IsTextSelected())
			{
				DeleteSelection();
			}
			else
			{
				//delete the character before the cursor
				if(m_iCursorPos>0)
				{
					for(int i=m_iCursorPos-1;i<iLength;i++)
					{
						m_pszTextBuffer[i] = m_pszTextBuffer[i+1];
					}
					m_iCursorPos--;
					m_iSelectionStart = m_iCursorPos;
				}
			}
		
			bChanged = true;
		}
		else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ENTER) || Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_NUMPADENTER))
		{
			Asura_Overlay_System::ReleaseFocus();
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}
		else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ESCAPE))
		{
			m_bDiscardChanges = true;
			Asura_Overlay_System::ReleaseFocus();
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}
		else if(iLength<m_iMaxLength || IsTextSelected())
		{
			//Check the windows input buffer for the characters pressed
			u_int uInputBufferSize = Asura_Windows::GetSizeOfInputBuffer();
			const Asura_UnicodeChar* wszInputBuffer = Asura_Windows::GetInputBuffer();

			for(u_int i=0;i<uInputBufferSize;i++)
			{
				char cKeyPressed = (char)wszInputBuffer[i];

				//Ignore some characters that we don't want. Mainly to avoid backspaces from being printed here.
				if(cKeyPressed >=0 && cKeyPressed < ' ') return;

				if(CharacterAllowed(cKeyPressed))
				{
					if(IsTextSelected())
					{
						//typing a new character causes the selection to be deleted
						DeleteSelection();
					}
					for(int i=strlen(m_pszTextBuffer)+1;i>m_iCursorPos;i--)
					{
						m_pszTextBuffer[i] = m_pszTextBuffer[i-1];
					}

					m_pszTextBuffer[m_iCursorPos++] = cKeyPressed;
					m_iSelectionStart = m_iCursorPos;
					Asura_Overlay_System::s_bUsedKeypressThisFrame = true;

					bChanged = true;
				}

			}
		}
	}
	if(bChanged)
	{
		if(m_bUpdateImmediatelyWhileTyping)
		{
			if(!m_bDiscardChanges)
			{
				if(ReadFromTextBuffer())
				{
					if(m_pbFlag) *m_pbFlag = true;
				}
			}
		}
		if (!m_xTextChangeCallback.IsNull ())
		{
			m_xTextChangeCallback.GetCallback ()->Execute ();
		}
	}
}

void Asura_BaseText_Gadget::DeleteSelection()
{
	//delete all characters in the selection , and move the cursor to the start
	int iCharsToDelete = Asura_Maths::Abs(m_iSelectionStart-m_iCursorPos);
	m_iCursorPos = Asura_Maths::Min(m_iCursorPos,m_iSelectionStart);
	m_iSelectionStart = m_iCursorPos;
	
	int iLength = strlen(m_pszTextBuffer);
	for(int i=m_iCursorPos;(i+iCharsToDelete-1)<iLength;i++)
	{
		m_pszTextBuffer[i] = m_pszTextBuffer[i+iCharsToDelete];
	}	
}

bool Asura_BaseText_Gadget::IsTextSelected() const
{
	return m_iCursorPos != m_iSelectionStart;
}

void Asura_BaseText_Gadget::CopyText() const
{
	int iCopyStart = Asura_Maths::Min(m_iSelectionStart,m_iCursorPos);
	int iCharsSelected = Asura_Maths::Abs(m_iSelectionStart-m_iCursorPos);
	int iCharsToCopy = iCharsSelected;


	//copy to windows clipboard
	if( OpenClipboard(NULL) )
	{
		HGLOBAL hMem = GlobalAlloc( GMEM_MOVEABLE, iCharsToCopy + 1 );
		if( hMem )
		{
			char* szOut = reinterpret_cast<char*>( GlobalLock(hMem) );
			strncpy(szOut,m_pszTextBuffer+iCopyStart,iCharsToCopy);
			szOut[iCharsToCopy] = 0;
			GlobalUnlock( hMem );

			EmptyClipboard();
			SetClipboardData( CF_TEXT, hMem );
		}

		CloseClipboard();
	}

}

void Asura_BaseText_Gadget::PasteText()
{
	//pasting deletes the old selection
	DeleteSelection();


	//copy from the windows clipboard
	if( OpenClipboard(NULL) )
	{
		HANDLE hMem = GetClipboardData(CF_TEXT);
		if( hMem )
		{
			char* szPasteBuffer = reinterpret_cast<char*>( GlobalLock(hMem) );

			//determine the number of characters to move
			int iCharsToCopy = Asura_Maths::Min((int)strlen(szPasteBuffer),m_iMaxLength-m_iCursorPos);

			//move the characters after the cursor up to make room
			int iLength = strlen(m_pszTextBuffer);
			for(int i=iLength;i>=m_iCursorPos;i--)
			{
				if(i+iCharsToCopy>=m_iMaxLength) continue;//string will be truncated

				m_pszTextBuffer[i+iCharsToCopy] = m_pszTextBuffer[i];
			}
			if(iLength+iCharsToCopy>=m_iMaxLength)
			{
				//string was too long , need to add terminator back
				m_pszTextBuffer[m_iMaxLength] = 0;
			}

			//can now paste the text in the space left
			strncpy(m_pszTextBuffer+m_iCursorPos,szPasteBuffer,iCharsToCopy);

			//move the cursor to the end of the pasted text
			m_iCursorPos += iCharsToCopy;
			m_iSelectionStart = m_iCursorPos;

			GlobalUnlock( hMem );

		}

		CloseClipboard();
	}

}


void Asura_BaseText_Gadget::SetMaxStringLength(int iMaxLength)
{
	//allocate new storage space , and copy old string
	char* pszTextBuffer = new char[iMaxLength+1];
	Asura_PrintfToBfr( pszTextBuffer, "%s", iMaxLength + 1, true, m_pszTextBuffer );

	delete [] m_pszTextBuffer;
	m_pszTextBuffer = pszTextBuffer;
	m_iMaxLength = iMaxLength;

	//make sure cursor is in bounds
	if(m_iCursorPos>iMaxLength) m_iCursorPos = iMaxLength;

	
}

void Asura_BaseText_Gadget::SetBoxLength(int iNumChars)
{
	m_iBoxLengthOverride = iNumChars;
	if(m_iBoxLengthOverride > 0)
	{
		m_iBoxLength = m_iBoxLengthOverride;
		m_fWidth = (float) m_iBoxLength;
	}

}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Float_Gadget::Asura_Float_Gadget( const Asura_Vector_2& offset,float* value,float min,float max,const char* printFormat,bool* flag, bool bRightJustify /* = false */)
:Asura_BaseText_Gadget(offset,20,flag,false)
{
	m_xPrintFormat = printFormat;
	m_pfValue = value;
	m_pxSharedValue = 0;
	m_fMin = min;
	m_fMax = max;
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_pfValue < m_fMin)
	{
		Asura_Assertf(false,("The starting value of %f is less than the min value of %f. Setting the value to %f",*m_pfValue,m_fMin,m_fMin));
		*m_pfValue = m_fMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_pfValue > m_fMax)
	{
		Asura_Assertf(false,("The starting value of %f is greater than the max value of %f. Setting the value to %f",*m_pfValue,m_fMax,m_fMax));
		*m_pfValue = m_fMax;
	}
	m_bCentreJustify = true;
	m_bRightJustify = bRightJustify;

	UpdateBoxLength( m_fMin, m_fMax, true );

}

Asura_Float_Gadget::Asura_Float_Gadget( const Asura_Vector_2& offset,Asura_Dialog_SharedValue<float>* pxSharedValue,float min,float max,const char* printFormat,bool* flag, bool bRightJustify /* = false */)
:Asura_BaseText_Gadget(offset,20,flag,false)
{
	m_xPrintFormat = printFormat;
	m_pfValue = 0;
	m_pxSharedValue = pxSharedValue;
	m_fMin = min;
	m_fMax = max;
	float fValue;
	m_pxSharedValue->GetValue(&fValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( fValue < m_fMin)
	{
		Asura_Assertf(false,("The starting value of %f is less than the min value of %f. Setting the value to %f",fValue,m_fMin,m_fMin));
		m_pxSharedValue->SetValue(m_fMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( fValue > m_fMax)
	{
		Asura_Assertf(false,("The starting value of %f is greater than the max value of %f. Setting the value to %f",fValue,m_fMax,m_fMax));
		m_pxSharedValue->SetValue(m_fMax);
	}
	m_bCentreJustify = true;
	m_bRightJustify = bRightJustify;

	UpdateBoxLength( m_fMin, m_fMax, true );
}



void Asura_Float_Gadget::UpdateBoxLength(const float fMin, const float fMax, const bool bShrink )
{
	int maxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), Asura_Maths::FtoD(fMax));
	int minLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), Asura_Maths::FtoD(fMin));

	if (minLength>maxLength) maxLength = minLength;
	
	if(m_iBoxLengthOverride > 0)
	{
		m_iBoxLength = m_iBoxLengthOverride;
	}
	else
	{
		// Only make the box bigger if it needs to be bigger
		if ( bShrink == false )
		{
			if ( maxLength > m_iBoxLength )
			{
				m_iBoxLength = maxLength;
			}
		}
		else
		{
			// Make box smaller as well
			m_iBoxLength = maxLength;
		}
	}


	m_fWidth = (float)m_iBoxLength;
	m_fHeight = 1.0f;

	if ( m_bRightJustify )
	{
		m_xOffset.x -= m_fWidth;
	}
}

void Asura_Float_Gadget::FillTextBuffer()
{
	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, Asura_Maths::FtoD( GetValue() ) );
}

bool Asura_Float_Gadget::ReadFromTextBuffer()
{
	float fOldValue = GetValue();
	float fNewValue;
	
	if(sscanf(m_pszTextBuffer,"%f",&fNewValue) != 1) return false;

	if(fNewValue < m_fMin) fNewValue = m_fMin;
	else if(fNewValue > m_fMax) fNewValue = m_fMax;

	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, Asura_Maths::FtoD( fNewValue ) );
	sscanf(m_pszTextBuffer,"%f",&fNewValue);

	if(fNewValue != fOldValue)
	{
		SetValue(fNewValue);
	}

	return fNewValue != fOldValue;
}

bool Asura_Float_Gadget::CharacterAllowed(char cChar)
{
	if(cChar>='0' && cChar<='9') return true;
	if(cChar == '.') return true;
	if(cChar == '-') return true;
	return false;
}

float Asura_Float_Gadget::GetValue()
{
	Asura_Assert(m_pfValue || m_pxSharedValue,"");

	if(m_pfValue)
	{
		return *m_pfValue;
	}
	else
	{
		float fValue;
		m_pxSharedValue->GetValue(&fValue,0);
		return fValue;
	}
}

void Asura_Float_Gadget::SetValue(float fValue)
{
	Asura_Assert(m_pfValue || m_pxSharedValue,"");

	if(m_pfValue)
	{
		*m_pfValue = fValue;
	}
	else
	{
		m_pxSharedValue->SetValue(fValue);
	}
}

bool Asura_Float_Gadget::IsSharedGadget()
{
	return (m_pxSharedValue!=0);
}

bool Asura_Float_Gadget::IsMixedValue()
{
	if(!m_pxSharedValue) return false;
	bool bMixed;
	m_pxSharedValue->GetValue(0,&bMixed);
	return bMixed;
}

void Asura_Float_Gadget::SetMinMax( const float fMin, const float fMax )
{
	m_fMin = fMin;
	m_fMax = fMax;
	// Only make the box longer if it needs to be - don't make it shorter
	UpdateBoxLength( m_fMin, m_fMax, false );
}

void Asura_Float_Gadget::SetPrintFormat( const char* const szPrintFormat )
{
	m_xPrintFormat = szPrintFormat;
	// Only make the box longer if it needs to be - don't make it shorter
	UpdateBoxLength( m_fMin, m_fMax, false );
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Integer_Gadget::Asura_Integer_Gadget( const Asura_Vector_2& offset, int* value, int min, int max, const char* printFormat, bool* flag /*= 0*/)
:Asura_BaseText_Gadget(offset,20,flag,false)
{
	m_xPrintFormat = printFormat;
	m_piValue = value;
	m_pxSharedValue = 0;
	m_iMin = min;
	m_iMax = max;
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_piValue < m_iMin)
	{
		Asura_Assertf(false,("The starting value of %i is less than the min value of %i. Setting the value to %i",*m_piValue,m_iMin,m_iMin));
		*m_piValue = m_iMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_piValue > m_iMax)
	{
		Asura_Assertf(false,("The starting value of %i is greater than the max value of %i. Setting the value to %i",*m_piValue,m_iMax,m_iMax));
		*m_piValue = m_iMax;
	}
	m_bCentreJustify = true;

	//get the box length
	int maxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), max);
	int minLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), min);

	if (minLength>maxLength) maxLength = minLength;
	
	m_iBoxLength = maxLength;
	//if (m_iBoxLength>10) m_iBoxLength = 10;
	m_fWidth = (float)maxLength;
	m_fHeight = 1.0f;
}

Asura_Integer_Gadget::Asura_Integer_Gadget( const Asura_Vector_2& offset, Asura_Dialog_SharedValue<int>* pxSharedValue, int min, int max, const char* printFormat, bool* flag /*= 0*/)
:Asura_BaseText_Gadget(offset,20,flag,false)
{
	m_xPrintFormat = printFormat;
	m_piValue = 0;
	m_pxSharedValue = pxSharedValue;
	m_iMin = min;
	m_iMax = max;
	int iValue;
	m_pxSharedValue->GetValue(&iValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( iValue < m_iMin)
	{
		Asura_Assertf(false,("The starting value of %d is less than the min value of %d. Setting the value to %d",iValue,m_iMin,m_iMin));
		m_pxSharedValue->SetValue(m_iMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( iValue > m_iMax)
	{
		Asura_Assertf(false,("The starting value of %d is greater than the max value of %d. Setting the value to %d",iValue,m_iMax,m_iMax));
		m_pxSharedValue->SetValue(m_iMax);
	}
	m_bCentreJustify = true;

	//get the box length
	int maxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), max);
	int minLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), min);

	if (minLength>maxLength) maxLength = minLength;
	
	m_iBoxLength = maxLength;
	//if (m_iBoxLength>10) m_iBoxLength = 10;
	m_fWidth = (float)maxLength;
	m_fHeight = 1.0f;
}


void Asura_Integer_Gadget::FillTextBuffer()
{
	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, GetValue() );
}

bool Asura_Integer_Gadget::ReadFromTextBuffer()
{
	int iOldValue = GetValue();
	int iNewValue;
	
	if(m_pszTextBuffer[0] == '0' && m_pszTextBuffer[1] == 'x')
	{
		//read as hex
		if(sscanf(m_pszTextBuffer,"0x%x",&iNewValue) != 1) return false;
	}
	else
	{
		if(sscanf(m_pszTextBuffer,"%d",&iNewValue) != 1) return false;
	}
	if(iNewValue < m_iMin) iNewValue = m_iMin;
	else if(iNewValue > m_iMax) iNewValue = m_iMax;

	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, iNewValue );

	if(iOldValue != iNewValue)
	{
		SetValue(iNewValue);
	}

	return iOldValue != iNewValue;
}

bool Asura_Integer_Gadget::CharacterAllowed(char cChar)
{
	if(cChar>='0' && cChar<='9') return true;
	if(cChar == 'x') return true;
	if(cChar >= 'a' && cChar <= 'f') return true;
	if(cChar >= 'A' && cChar <= 'F') return true;
	if(cChar == '-') return true;
	return false;
}

int Asura_Integer_Gadget::GetValue()
{
	Asura_Assert(m_piValue || m_pxSharedValue,"");

	if(m_piValue)
	{
		return *m_piValue;
	}
	else
	{
		int iValue;
		m_pxSharedValue->GetValue(&iValue,0);
		return iValue;
	}
}

void Asura_Integer_Gadget::SetValue(int iValue)
{
	Asura_Assert(m_piValue || m_pxSharedValue,"");

	if(m_piValue)
	{
		*m_piValue = iValue;
	}
	else
	{
		m_pxSharedValue->SetValue(iValue);
	}
}

bool Asura_Integer_Gadget::IsSharedGadget()
{
	return (m_pxSharedValue!=0);
}

bool Asura_Integer_Gadget::IsMixedValue()
{
	if(!m_pxSharedValue) return false;
	bool bMixed;
	m_pxSharedValue->GetValue(0,&bMixed);
	return bMixed;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_UInt_Gadget::Asura_UInt_Gadget( const Asura_Vector_2& xOffset, u_int* puValue, u_int uMin, u_int uMax, const char* const szPrintFormat, bool* pbFlag /*= 0*/)
:Asura_BaseText_Gadget(xOffset,20,pbFlag,false)
{
	m_xPrintFormat = szPrintFormat;
	m_puValue = puValue;
	m_pxSharedValue = 0;
	m_uMin = uMin;
	m_uMax = uMax;
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_puValue < m_uMin)
	{
		Asura_Assertf(false,("The starting value of %u is less than the min value of %u. Setting the value to %u",*m_puValue,m_uMin,m_uMin));
		*m_puValue = m_uMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_puValue > m_uMax)
	{
		Asura_Assertf(false,("The starting value of %u is greater than the max value of %u. Setting the value to %u",*m_puValue,m_uMax,m_uMax));
		*m_puValue = m_uMax;
	}
	m_bCentreJustify = true;

	//get the box length
	int iMaxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_uMax);
	int iMinLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_uMin);

	if (iMinLength>iMaxLength) iMaxLength = iMinLength;
	
	m_iBoxLength = iMaxLength;
	//if (m_iBoxLength>10) m_iBoxLength = 10;
	m_fWidth = static_cast<float>(iMaxLength);
	m_fHeight = 1.0f;
}

Asura_UInt_Gadget::Asura_UInt_Gadget( const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<u_int>* pxSharedValue, u_int uMin, u_int uMax, const char* const szPrintFormat, bool* pbFlag /*= 0*/)
:Asura_BaseText_Gadget(xOffset,20,pbFlag,false)
{
	m_xPrintFormat = szPrintFormat;
	m_puValue = 0;
	m_pxSharedValue = pxSharedValue;
	m_uMin = uMin;
	m_uMax = uMax;
	u_int uValue;
	m_pxSharedValue->GetValue(&uValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( uValue < m_uMin)
	{
		Asura_Assertf(false,("The starting value of %u is less than the min value of %u. Setting the value to %u",uValue,m_uMin,m_uMin));
		m_pxSharedValue->SetValue(m_uMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( uValue > uMax)
	{
		Asura_Assertf(false,("The starting value of %u is greater than the max value of %u. Setting the value to %u",uValue,m_uMax,m_uMax));
		m_pxSharedValue->SetValue(m_uMax);
	}
	m_bCentreJustify = true;

	//get the box length
	int iMaxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_uMax);
	int iMinLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_uMin);

	if (iMinLength>iMaxLength) iMaxLength = iMinLength;
	
	m_iBoxLength = iMaxLength;
	//if (m_iBoxLength>10) m_iBoxLength = 10;
	m_fWidth = static_cast<float>(iMaxLength);
	m_fHeight = 1.0f;
}


void Asura_UInt_Gadget::FillTextBuffer()
{
	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, GetValue() );
}

bool Asura_UInt_Gadget::ReadFromTextBuffer()
{
	u_int uOldValue = GetValue();
	u_int uNewValue;
	
	if(m_pszTextBuffer[0] == '0' && m_pszTextBuffer[1] == 'x')
	{
		//read as hex
		if(sscanf(m_pszTextBuffer,"0x%x",&uNewValue) != 1) return false;
	}
	else
	{
		if(sscanf(m_pszTextBuffer,"%d",&uNewValue) != 1) return false;
	}

	if(uNewValue<m_uMin)uNewValue=m_uMin;
	else if(uNewValue>m_uMax)uNewValue=m_uMax;

	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, uNewValue );

	if(uNewValue != uOldValue)
	{
		SetValue(uNewValue);
	}

	return uOldValue != uNewValue;
}

bool Asura_UInt_Gadget::CharacterAllowed(char cChar)
{
	if(cChar>='0' && cChar<='9') return true;
	if(cChar == 'x') return true;
	if(cChar >= 'a' && cChar <= 'f') return true;
	if(cChar >= 'A' && cChar <= 'F') return true;
	return false;
}

u_int Asura_UInt_Gadget::GetValue()
{
	Asura_Assert(m_puValue || m_pxSharedValue,"");

	if(m_puValue)
	{
		return *m_puValue;
	}
	else
	{
		u_int uValue;
		m_pxSharedValue->GetValue(&uValue,0);
		return uValue;
	}
}

void Asura_UInt_Gadget::SetValue(u_int uValue)
{
	Asura_Assert(m_puValue || m_pxSharedValue,"");

	if(m_puValue)
	{
		*m_puValue = uValue;
	}
	else
	{
		m_pxSharedValue->SetValue(uValue);
	}
}

bool Asura_UInt_Gadget::IsSharedGadget()
{
	return (m_pxSharedValue!=0);
}

bool Asura_UInt_Gadget::IsMixedValue()
{
	if(!m_pxSharedValue) return false;
	bool bMixed;
	m_pxSharedValue->GetValue(0,&bMixed);
	return bMixed;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_UChar_Gadget::Asura_UChar_Gadget(const Asura_Vector_2& xOffset, u_char* pucValue, u_char ucMin, u_char ucMax, const char* const szPrintFormat, bool* pbFlag /* = 0 */)
:Asura_BaseText_Gadget(xOffset,20,pbFlag,false)
{
	m_xPrintFormat = szPrintFormat;
	m_pucValue = pucValue;
	m_pxSharedValue = 0;
	m_ucMin = ucMin;
	m_ucMax = ucMax;
	// If the value is outside of the range we need to tell the user and clamp the value.
	// Note the values have been cast to ints to make the assert more readable
	if ( *m_pucValue < m_ucMin)
	{
		Asura_Assertf(false,("The starting value of %d is less than the min value of %d. Setting the value to %d",static_cast<int>(*m_pucValue),static_cast<int>(m_ucMin),static_cast<int>(m_ucMin)));
		*m_pucValue = m_ucMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_pucValue > m_ucMax)
	{
		Asura_Assertf(false,("The starting value of %d is greater than the max value of %d. Setting the value to %d",static_cast<int>(*m_pucValue),static_cast<int>(m_ucMax),static_cast<int>(m_ucMax)));
		*m_pucValue = m_ucMax;
	}
	m_bCentreJustify = true;

	//get the box length
	int iMaxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_ucMax);
	int iMinLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_ucMin);

	iMaxLength = Asura_Maths::Max(iMaxLength, iMinLength);

	m_iBoxLength = iMaxLength;
	m_fWidth = static_cast<float>(iMaxLength);
	m_fHeight = 1.0f;
}

Asura_UChar_Gadget::Asura_UChar_Gadget(const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<u_char>* pxSharedValue, u_char ucMin, u_char ucMax, const char* const szPrintFormat, bool* pbFlag /* = 0 */)
:Asura_BaseText_Gadget(xOffset,20,pbFlag,false)
{
	m_xPrintFormat = szPrintFormat;
	m_pucValue = 0;
	m_pxSharedValue = pxSharedValue;
	m_ucMin = ucMin;
	m_ucMax = ucMax;
	u_char ucValue;
	m_pxSharedValue->GetValue(&ucValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	// Note the values have been cast to ints to make the assert more readable
	if ( ucValue < m_ucMin)
	{
		Asura_Assertf(false,("The starting value of %d is less than the min value of %d. Setting the value to %d",static_cast<int>(ucValue),static_cast<int>(m_ucMin),static_cast<int>(m_ucMin)));
		m_pxSharedValue->SetValue(m_ucMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( ucValue > ucMax)
	{
		Asura_Assertf(false,("The starting value of %d is greater than the max value of %d. Setting the value to %d",static_cast<int>(ucValue),static_cast<int>(m_ucMax),static_cast<int>(m_ucMax)));
		m_pxSharedValue->SetValue(m_ucMax);
	}
	m_bCentreJustify = true;

	//get the box length
	int iMaxLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_ucMax);
	int iMinLength = Asura_Text_Output::LengthOfString(m_xPrintFormat.GetString(), m_ucMin);

	iMaxLength = Asura_Maths::Max(iMaxLength, iMinLength);

	m_iBoxLength = iMaxLength;
	m_fWidth = static_cast<float>(iMaxLength);
	m_fHeight = 1.0f;
}


void Asura_UChar_Gadget::FillTextBuffer()
{
	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, GetValue() );
}

bool Asura_UChar_Gadget::ReadFromTextBuffer()
{
	const u_char ucOldValue = GetValue();
	u_int uNewValue;

	if(m_pszTextBuffer[0] == '0' && m_pszTextBuffer[1] == 'x')
	{
		//read as hex
		if(sscanf(m_pszTextBuffer,"0x%x",&uNewValue) != 1) return false;
	}
	else
	{
		if(sscanf(m_pszTextBuffer,"%d",&uNewValue) != 1) return false;
	}

	const u_char ucNewValue = static_cast<u_char>(Asura_Maths::ClampToRange(uNewValue, static_cast<u_int>(m_ucMin), static_cast<u_int>(m_ucMax)));

	Asura_PrintfToBfr( m_pszTextBuffer, m_xPrintFormat.GetString(), m_iMaxLength + 1, true, ucNewValue );

	if(ucNewValue != ucOldValue)
	{
		SetValue(ucNewValue);
	}

	return ucOldValue != ucNewValue;
}

bool Asura_UChar_Gadget::CharacterAllowed(char cChar)
{
	if(cChar>='0' && cChar<='9') return true;
	if(cChar == 'x') return true;
	if(cChar >= 'a' && cChar <= 'f') return true;
	if(cChar >= 'A' && cChar <= 'F') return true;
	return false;
}

u_char Asura_UChar_Gadget::GetValue()
{
	Asura_Assert(m_pucValue || m_pxSharedValue,"");

	if(m_pucValue)
	{
		return *m_pucValue;
	}
	else
	{
		u_char ucValue;
		m_pxSharedValue->GetValue(&ucValue,0);
		return ucValue;
	}
}

void Asura_UChar_Gadget::SetValue(u_char ucValue)
{
	Asura_Assert(m_pucValue || m_pxSharedValue,"");

	if(m_pucValue)
	{
		*m_pucValue = ucValue;
	}
	else
	{
		m_pxSharedValue->SetValue(ucValue);
	}
}

bool Asura_UChar_Gadget::IsSharedGadget()
{
	return (m_pxSharedValue!=0);
}

bool Asura_UChar_Gadget::IsMixedValue()
{
	if(!m_pxSharedValue) return false;
	bool bMixed;
	m_pxSharedValue->GetValue(0,&bMixed);
	return bMixed;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_NumberSlider_Gadget::Asura_NumberSlider_Gadget( const Asura_Vector_2& offset,float* value, float* minPtr, float* maxPtr,float slideWidth,const char* const szPrintFormat, bool* flag/* = 0*/)
:Asura_Float_Gadget(offset,value,*minPtr,*maxPtr,szPrintFormat,flag)
{
	Asura_Assert(minPtr,"");
	Asura_Assert(maxPtr,"");
	Asura_Assert(slideWidth>0,"");

	m_pfMinPtr = minPtr;
	m_pfMaxPtr = maxPtr;
	m_fSlideWidth = slideWidth;
	m_fSliderPosition = 0;

	UpdateBoxLength( m_fMin, m_fMax, true );
	m_fWidth = m_fSlideWidth;
}

void Asura_NumberSlider_Gadget::Update( const Asura_Vector_2& xOrigin )
{	
	if(m_fMin!=*m_pfMinPtr || m_fMax!=*m_pfMaxPtr)
	{
		//min or max for slider has changed
		m_fMin = *m_pfMinPtr;
		m_fMax = *m_pfMaxPtr;
		UpdateBoxLength( m_fMin, m_fMax, true );
		m_fWidth = m_fSlideWidth;
	}
	
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
 		Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
        m_bClickedThisFrame = true;
	}
		
 	if (Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
 	{
 		if(m_iFocusFlags)
 		{
			//move the slider to the position clicked on
			float cx,cy;
			Asura_Cursor::GetScreenCoords(cx,cy);
			//get coord relative to start of bar
			cx -= (xOrigin.x + m_xOffset.x);
			if(cx < 0) cx = 0;
			else if(cx > m_fSlideWidth) cx = m_fSlideWidth;

			*m_pfValue = m_fMin + (cx/m_fSlideWidth)*(m_fMax - m_fMin);
			if(m_pbFlag) *m_pbFlag = true;
		}
	}
	else if(m_iFocusFlags)
	{
		Asura_Overlay_System::ReleaseFocus(OverlayFocusType_Mouse);
	}
	
	UpdateSliderPosition();
}

void Asura_NumberSlider_Gadget::Render( const Asura_Vector_2& xOrigin )
{	
	if(m_fMin!=*m_pfMinPtr || m_fMax!=*m_pfMaxPtr)
	{
		//min or max for slider has changed
		m_fMin = *m_pfMinPtr;
		m_fMax = *m_pfMaxPtr;
		UpdateBoxLength( m_fMin, m_fMax, true );
		m_fWidth = m_fSlideWidth;
	}

	//draw the bar on which the number slider lives
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	
	unsigned int colour = 0x7f7f7f7f;//0000;

	Asura_Render::Overlay.Colour = colour;
	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

	float x=Asura_Maths::FloatToInt((xOrigin.x + m_xOffset.x)*ASURA_TEXT_OUTPUT_FONT_WIDTH)-0.5f;
	float y=Asura_Maths::FloatToInt((xOrigin.y + m_xOffset.y)*ASURA_TEXT_OUTPUT_FONT_HEIGHT)-0.5f;
	
	Asura_Render::Overlay.SetRectXY(x,y,m_fSlideWidth*ASURA_TEXT_OUTPUT_FONT_WIDTH,m_fHeight*ASURA_TEXT_OUTPUT_FONT_HEIGHT);

	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();
	
	//draw the number slider itself
	UpdateSliderPosition();
	float offsetX = m_xOffset.x;
	m_xOffset.x+=m_fSliderPosition;//adjust the offset to the start of the number box
	Asura_Float_Gadget::Render( xOrigin );
	m_xOffset.x=offsetX;//reset the offset
}

void Asura_NumberSlider_Gadget::UpdateSliderPosition()
{
	//calculate where the gadget should be along its slidy thing

	float proportion = (m_fMax > m_fMin) ? (*m_pfValue - m_fMin)/(m_fMax - m_fMin) : 0.0f;
	if(proportion > 1) proportion = 1.0f;
	else if(proportion < 0) proportion = 0.0f;

	m_fSliderPosition = (m_fSlideWidth-m_iBoxLength)*proportion;
	//clamp the position slightly so that thesides of the box will stilll get drawn if against the side of the screen
	if(m_fSliderPosition < 0.1f) m_fSliderPosition = 0.1f;
	if(m_fSliderPosition > m_fSlideWidth-0.1f) m_fSliderPosition = m_fSlideWidth-0.1f;
}

void Asura_NumberSlider_Gadget::SetPrintFormat(const char* const szPrintFormat)
{
	m_xPrintFormat = szPrintFormat;
	UpdateBoxLength( m_fMin, m_fMax, true );
	m_fWidth = m_fSlideWidth;
}

//////////////////////////////////////////////////////////////////////////////////////////////////

Asura_String_Gadget::Asura_String_Gadget( const Asura_Vector_2& offset, Asura_String* value, int maxLength, bool* flag, bool defaultFocus)
:Asura_BaseText_Gadget(offset,maxLength,flag,defaultFocus)
{
	if(!value->GetString()) *value = "";
	m_pxValue = value;
}


void Asura_String_Gadget::FillTextBuffer()
{
	Asura_Assert(m_pxValue->GetLength()<=m_iMaxLength,"");
	Asura_PrintfToBfr( m_pszTextBuffer, "%s", m_iMaxLength + 1, true, m_pxValue->GetString() );
}

bool Asura_String_Gadget::ReadFromTextBuffer()
{
	bool bChanged = m_pxValue->Compare(m_pszTextBuffer,false)!=0;

	*m_pxValue = m_pszTextBuffer;
	return bChanged;
}


//////////////////////////////////////////////////////////////////////////////////////////////////

Asura_Incrementor_Gadget::Asura_Incrementor_Gadget( const Asura_Vector_2& offset,bool sign,bool* flag)
{
	m_xOffset = offset;
	m_xOffset.x += 0.5f;
	m_pbFlag = flag;
	m_bSign = sign;

	m_fWidth = 1.0f;
	m_fHeight = 1.0f;
	m_fTimer = 0.0f;
}

void Asura_Incrementor_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	
	unsigned int colour = GetColour( m_bClickedThisFrame );

	unsigned char c;
	if (m_bSign)
	{
		c = 5;
	}
	else
	{
		c = 1;
	}

	Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",c);

    Asura_Gadget::Render( xOrigin );
}

void Asura_Incrementor_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable )
	{
		int Change = 0;

		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) || Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
		{
			Tools_ContextMenu_System::SetRightClickUsed();
			
			//increment the value when the button is first pressed
			Change = Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2) ? 10 : 1;
			//and set a short delay to the next change
			m_fTimer = 0.2f;
			//take mouse focus , so we know that the subsequent non-debounced presses are okay
			//(unless shift is held down , in which case we only want to allow debounced inputs
			//to be used)
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
			{
				Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
			}
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1) || Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON2))
		{
			Tools_ContextMenu_System::SetRightClickUsed();
		
			if(m_iFocusFlags)
			{
				//see if enough time has passed for the nxt change
				m_fTimer -= Asura_Timers::GetFrameTime();	
				if(m_fTimer<0)
				{
					//update at roughly 20 times per second
					m_fTimer += 0.05f;
					if(m_fTimer<0) m_fTimer = 0;

					Change = Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON2) ? 10 : 1;
				}
			}
		}
		else
		{
			Asura_Overlay_System::ReleaseFocus(OverlayFocusType_Mouse);
		}
			
		

		if(Change)
		{
			Increment(Change);
			if(m_pbFlag) *m_pbFlag = true;
            m_bClickedThisFrame = true;
		}
	}

}
//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorFloat_Gadget::Asura_IncrementorFloat_Gadget( const Asura_Vector_2& xOffset, float* pfValue, float fMin, float fMax, float fDelta, bool* bFlag /*= 0*/)
:Asura_Incrementor_Gadget(xOffset,fDelta>0,bFlag)
{
	m_pfValue = pfValue;
	m_fMin = fMin;
	m_fMax = fMax;
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_pfValue < m_fMin)
	{
		Asura_Assertf(false,("The starting value of %f is less than the min value of %f. Setting the value to %f",*m_pfValue,m_fMin,m_fMin));
		*m_pfValue = m_fMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_pfValue > m_fMax)
	{
		Asura_Assertf(false,("The starting value of %f is greater than the max value of %f. Setting the value to %f",*m_pfValue,m_fMax,m_fMax));
		*m_pfValue = m_fMax;
	}
	m_fDelta = fDelta;
}
void Asura_IncrementorFloat_Gadget::Increment(int iMultiplier)
{
	*m_pfValue += iMultiplier*m_fDelta;
	if(*m_pfValue<m_fMin) *m_pfValue = m_fMin;
	else if(*m_pfValue>m_fMax) *m_pfValue = m_fMax;
}

void Asura_IncrementorFloat_Gadget::SetDelta( const float fDelta )
{
	m_fDelta = fDelta;
}

void Asura_IncrementorFloat_Gadget::SetMinMax( const float fMin,  const float fMax )
{
	m_fMin = fMin;
	m_fMax = fMax;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorInt_Gadget::Asura_IncrementorInt_Gadget( const Asura_Vector_2& xOffset, int* piValue, int iMin, int iMax, int iDelta, bool* bFlag /*= 0*/)
:Asura_Incrementor_Gadget(xOffset,iDelta>0,bFlag)
{
	m_piValue = piValue;
	m_iMin = iMin;
	m_iMax = iMax;
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_piValue < m_iMin)
	{
		Asura_Assertf(false,("The starting value of %i is less than the min value of %i. Setting the value to %i",*m_piValue,m_iMin,m_iMin));
		*m_piValue = m_iMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_piValue > m_iMax)
	{
		Asura_Assertf(false,("The starting value of %i is greater than the max value of %i. Setting the value to %i",*m_piValue,m_iMax,m_iMax));
		*m_piValue = m_iMax;
	}
	m_iDelta = iDelta;
}
void Asura_IncrementorInt_Gadget::Increment(int iMultiplier)
{
	*m_piValue += iMultiplier*m_iDelta;
	if(*m_piValue<m_iMin) *m_piValue = m_iMin;
	else if(*m_piValue>m_iMax) *m_piValue = m_iMax;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorUInt_Gadget::Asura_IncrementorUInt_Gadget( const Asura_Vector_2& xOffset, u_int* puValue, u_int uMin, u_int uMax, int iDelta, bool* pbFlag /*= 0*/)
:Asura_Incrementor_Gadget(xOffset,iDelta>0,pbFlag)
{
	m_puValue = puValue;
	m_uMin = uMin;
	m_uMax = uMax;
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_puValue < m_uMin)
	{
		Asura_Assertf(false,("The starting value of %u is less than the min value of %u. Setting the value to %u",*m_puValue,m_uMin,m_uMin));
		*m_puValue = m_uMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_puValue > m_uMax)
	{
		Asura_Assertf(false,("The starting value of %u is greater than the max value of %u. Setting the value to %u",*m_puValue,m_uMax,m_uMax));
		*m_puValue = m_uMax;
	}
	m_iDelta = iDelta;
}
void Asura_IncrementorUInt_Gadget::Increment(int iMultiplier)
{
	int iDelta = iMultiplier*m_iDelta; 
	if (iDelta > 0)
	{
		*m_puValue += iDelta;
		if (*m_puValue > m_uMax) *m_puValue = m_uMax;
	}
	else
	{
		// (AW - 18.11.2004) - Be very careful about underflow.
		if (static_cast<u_int>(-iDelta) <= *m_puValue)
		{
			*m_puValue += iDelta;
		}
		else
		{
			*m_puValue = 0;
		}
		if (*m_puValue < m_uMin) *m_puValue = m_uMin;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorUChar_Gadget::Asura_IncrementorUChar_Gadget(const Asura_Vector_2& xOffset, u_char* pucValue, u_char ucMin, u_char ucMax, int iDelta, bool* pbFlag /* = 0 */)
:Asura_Incrementor_Gadget(xOffset,iDelta > 0, pbFlag)
{
	m_pucValue = pucValue;
	m_ucMin = ucMin;
	m_ucMax = ucMax;
	// If the value is outside of the range we need to tell the user and clamp the value.
	// Note the values have been cast to ints to make the assert more readable
	if ( *m_pucValue < m_ucMin)
	{
		Asura_Assertf(false,("The starting value of %d is less than the min value of %d. Setting the value to %d",static_cast<int>(*m_pucValue),static_cast<int>(m_ucMin),static_cast<int>(m_ucMin)));
		*m_pucValue = m_ucMin;
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( *m_pucValue > m_ucMax)
	{
		Asura_Assertf(false,("The starting value of %d is greater than the max value of %d. Setting the value to %d",static_cast<int>(*m_pucValue),static_cast<int>(m_ucMax),static_cast<int>(m_ucMax)));
		*m_pucValue = m_ucMax;
	}
	m_iDelta = iDelta;
}
void Asura_IncrementorUChar_Gadget::Increment(int iMultiplier)
{
	int iDelta = iMultiplier * m_iDelta; 
	if (iDelta > 0)
	{
		// Avoid wrapping
		if(static_cast<int>(*m_pucValue) + iDelta < 255)
		{
			*m_pucValue = *m_pucValue + static_cast<char>(iDelta);
		}
		else
		{
			*m_pucValue = 255;
		}
		*m_pucValue = Asura_Maths::Max(*m_pucValue, m_ucMin);
	}
	else
	{
		// (AW - 18.11.2004) - Be very careful about underflow.
		if (static_cast<u_int>(-iDelta) <= *m_pucValue)
		{
			*m_pucValue = *m_pucValue + static_cast<char>(iDelta);
		}
		else
		{
			*m_pucValue = 0;
		}
		*m_pucValue = Asura_Maths::Min(*m_pucValue, m_ucMax);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorSharedFloat_Gadget::Asura_IncrementorSharedFloat_Gadget( const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<float>* pxSharedValue, float fMin, float fMax, float fDelta)
:Asura_Incrementor_Gadget(xOffset,fDelta>0,0)
{
	m_pxSharedValue = pxSharedValue;
	m_fMin = fMin;
	m_fMax = fMax;
	float fValue;
	m_pxSharedValue->GetValue(&fValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( fValue < m_fMin)
	{
		Asura_Assertf(false,("The starting value of %f is less than the min value of %f. Setting the value to %f",fValue,m_fMin,m_fMin));
		m_pxSharedValue->SetValue(m_fMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( fValue > m_fMax)
	{
		Asura_Assertf(false,("The starting value of %f is greater than the max value of %f. Setting the value to %f",fValue,m_fMax,m_fMax));
		m_pxSharedValue->SetValue(m_fMax);
	}
	m_fDelta = fDelta;
}
void Asura_IncrementorSharedFloat_Gadget::Increment(int iMultiplier)
{
	float fValue;
	m_pxSharedValue->GetValue(&fValue,0);

	fValue += iMultiplier*m_fDelta;
	if(fValue<m_fMin) fValue = m_fMin;
	else if(fValue>m_fMax) fValue = m_fMax;

	m_pxSharedValue->SetValue(fValue);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorSharedInt_Gadget::Asura_IncrementorSharedInt_Gadget( const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<int>* pxSharedValue, int iMin, int iMax, int iDelta)
:Asura_Incrementor_Gadget(xOffset,iDelta>0,0)
{
	m_pxSharedValue = pxSharedValue;
	m_iMin = iMin;
	m_iMax = iMax;
	int iValue;
	m_pxSharedValue->GetValue(&iValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( iValue < m_iMin)
	{
		Asura_Assertf(false,("The starting value of %d is less than the min value of %d. Setting the value to %d",iValue,m_iMin,m_iMin));
		m_pxSharedValue->SetValue(m_iMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( iValue > m_iMax)
	{
		Asura_Assertf(false,("The starting value of %d is greater than the max value of %d. Setting the value to %d",iValue,m_iMax,m_iMax));
		m_pxSharedValue->SetValue(m_iMax);
	}
	m_iDelta = iDelta;
}
void Asura_IncrementorSharedInt_Gadget::Increment(int iMultiplier)
{
	int iValue;
	m_pxSharedValue->GetValue(&iValue,0);

	iValue += iMultiplier*m_iDelta;
	if(iValue<m_iMin) iValue = m_iMin;
	else if(iValue>m_iMax) iValue = m_iMax;

	m_pxSharedValue->SetValue(iValue);
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_IncrementorSharedUInt_Gadget::Asura_IncrementorSharedUInt_Gadget( const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<u_int>* pxSharedValue, u_int uMin, u_int uMax, int iDelta)
:Asura_Incrementor_Gadget(xOffset,iDelta>0,0)
{
	m_pxSharedValue = pxSharedValue;
	m_uMin = uMin;
	m_uMax = uMax;
	u_int uValue;
	m_pxSharedValue->GetValue(&uValue,0);
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( uValue < m_uMin)
	{
		Asura_Assertf(false,("The starting value of %u is less than the min value of %u. Setting the value to %u",uValue,m_uMin,m_uMin));
		m_pxSharedValue->SetValue(m_uMin);
	}
	// If the value is outside of the range we need to tell the user and clamp the value
	if ( uValue > uMax)
	{
		Asura_Assertf(false,("The starting value of %u is greater than the max value of %u. Setting the value to %u",uValue,m_uMax,m_uMax));
		m_pxSharedValue->SetValue(m_uMax);
	}
	m_iDelta = iDelta;
}
void Asura_IncrementorSharedUInt_Gadget::Increment(int iMultiplier)
{
	u_int uValue;
	m_pxSharedValue->GetValue(&uValue,0);
	
	int iDelta = iMultiplier*m_iDelta; 
	if (iDelta > 0)
	{
		uValue += iDelta;
		if (uValue > m_uMax) uValue = m_uMax;
	}
	else
	{
		// (AW - 18.11.2004) - Be very careful about underflow.
		if (static_cast<u_int>(-iDelta) <= uValue)
		{
			uValue += iDelta;
		}
		else
		{
			uValue = 0;
		}
		if (uValue < m_uMin) uValue = m_uMin;
	}

	m_pxSharedValue->SetValue(uValue);

}

//////////////////////////////////////////////////////////////////////////////////////////////////

const float Asura_BaseTickbox_Gadget::WIDTH = 3.0f;

Asura_BaseTickbox_Gadget::Asura_BaseTickbox_Gadget( const Asura_Vector_2& offset, bool* modifiedFlag)
{
	m_xOffset = offset;
	m_pbModifiedFlag = modifiedFlag;

	m_fWidth = WIDTH;
	m_fHeight = 1.0f;
}

void Asura_BaseTickbox_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	

	unsigned int colour = Asura_OverlayBrightness | GetColour(m_bClickedThisFrame);

	char tickBox[] = {2,3,4,0};
	if (IsChecked()) Asura_Text_Output::Print(pos.x+1.0f,pos.y,colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",ASURA_CHAR_TICK);
	Asura_Text_Output::Print(pos.x,pos.y,colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,tickBox);

    Asura_Gadget::Render( xOrigin );
}

void Asura_BaseTickbox_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
		if(m_iRadioButtonGroup)
		{
			GetOverlay()->RadioButtonPressed(this);
		}
		else
		{
			ToggleCheck();
		}
		if(m_pbModifiedFlag) *m_pbModifiedFlag = true;
		if(m_xOnClickedCallback.IsValid()) Asura_Overlay_System::SetCallbackFunctionToCall(m_xOnClickedCallback);
        m_bClickedThisFrame = true;
	}
}

Asura_BoolTickbox_Gadget::Asura_BoolTickbox_Gadget( const Asura_Vector_2& offset, bool* value, bool* flag)
:Asura_BaseTickbox_Gadget(offset,flag)
{
	m_pbValue = value;
}

Asura_FlagTickbox_Gadget::Asura_FlagTickbox_Gadget( const Asura_Vector_2& xOffset, int* piFlagsField, int iFlagBit, bool* pbModifiedFlag)
:Asura_BaseTickbox_Gadget(xOffset,pbModifiedFlag)
, m_piFlagsField(NULL)
, m_pcFlagsField(NULL)
, m_iFlagBit(0)
{
	Asura_Assert(piFlagsField,"");

	m_piFlagsField = piFlagsField;
	m_iFlagBit = iFlagBit;
}

Asura_FlagTickbox_Gadget::Asura_FlagTickbox_Gadget( const Asura_Vector_2& xOffset, char* pcFlagsField, char cFlagBit, bool* pbModifiedFlag)
:Asura_BaseTickbox_Gadget(xOffset,pbModifiedFlag)
, m_piFlagsField(NULL)
, m_pcFlagsField(NULL)
, m_iFlagBit(0)
{
	Asura_Assert(pcFlagsField,"");

	m_pcFlagsField = pcFlagsField;
	m_iFlagBit = cFlagBit;
}

bool Asura_FlagTickbox_Gadget::IsChecked( void ) const
{
	Asura_Assert((m_piFlagsField && !m_pcFlagsField) || (!m_piFlagsField && m_pcFlagsField), "Shouldn't have both int and char fields set");

	if(m_piFlagsField)
	{
		return (*m_piFlagsField & m_iFlagBit) != 0;
	}
	else
	{
		return (*m_pcFlagsField & m_iFlagBit) != 0;
	}
}

void Asura_FlagTickbox_Gadget::SetChecked(bool bSet)
{
	Asura_Assert((m_piFlagsField && !m_pcFlagsField) || (!m_piFlagsField && m_pcFlagsField), "Shouldn't have both int and char fields set");
	if(bSet)
	{
		if(m_piFlagsField)
		{
			*m_piFlagsField |= m_iFlagBit;
		}
		else
		{
			*m_pcFlagsField |= m_iFlagBit;
		}
	}
	else
	{
		if(m_piFlagsField)
		{
			*m_piFlagsField &=~ m_iFlagBit;
		}
		else
		{
			*m_pcFlagsField &=~ m_iFlagBit;
		}
	}
}


Asura_ListTickbox_Gadget::Asura_ListTickbox_Gadget(const Asura_Vector_2& xOffset,class Asura_List_Dialog* pxListDialog,int iIndex,bool* pbModifiedFlag)
:Asura_BaseTickbox_Gadget(xOffset,pbModifiedFlag)
{
	m_pxListDialog = pxListDialog;
	m_iIndex = iIndex;
}

bool Asura_ListTickbox_Gadget::IsChecked( void ) const 
{ 
	return m_pxListDialog->IsChecked(m_iIndex); 
}

void Asura_ListTickbox_Gadget::SetChecked(bool bSet)
{
	if(bSet != m_pxListDialog->IsChecked(m_iIndex))
	{
		m_pxListDialog->ToggleChecked(m_iIndex);
	}
}

Asura_SharedTickbox_Gadget::Asura_SharedTickbox_Gadget(const Asura_Vector_2& xOffset,Asura_Dialog_SharedValue<bool>* pxSharedValue)
:Asura_BaseTickbox_Gadget(xOffset,0)
{
	m_pxSharedValue = pxSharedValue;
}

bool Asura_SharedTickbox_Gadget::IsChecked( void ) const 
{ 
	bool bChecked;
	m_pxSharedValue->GetValue(&bChecked,0);
	return bChecked; 
}

void Asura_SharedTickbox_Gadget::SetChecked(bool bSet)
{
	m_pxSharedValue->SetValue(bSet);
}


void Asura_SharedTickbox_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	
	bool bChecked,bMixed;
	m_pxSharedValue->GetValue(&bChecked,&bMixed);

	unsigned int colour = GetColour(m_bClickedThisFrame);

	char tickBox[] = {2,3,4,0};
	if (bMixed || bChecked)
	{
		//If there is a mix of ticked and unticked , then draw tick with the mixed colour
		u_int uTickColour = colour;
		if(bMixed)
		{
			if(bChecked)
			{
				uTickColour = ASURA_GADGET_MIXED_COLOUR;
			}
			else
			{
				uTickColour = ASURA_GADGET_DARKMIXED_COLOUR;
			}
		}

		uTickColour |= Asura_OverlayBrightness;
		Asura_Text_Output::Print(pos.x+1.0f,pos.y,uTickColour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",ASURA_CHAR_TICK);
	}
	u_int uBoxColour = (colour == ASURA_GADGET_DEFAULT_COLOUR) ? ASURA_GADGET_SHARED_COLOUR : colour;
	uBoxColour |= Asura_OverlayBrightness;

	Asura_Text_Output::Print(pos.x,pos.y,uBoxColour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,tickBox);

    Asura_Gadget::Render( xOrigin );
}
//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_FunctionTickbox_Gadget::Asura_FunctionTickbox_Gadget(const Asura_Vector_2& xOffset, bool(*pfnIsChecked)(), void(*pfnSetChecked)(bool bSet))
:Asura_BaseTickbox_Gadget(xOffset,0)
{
	Asura_Assert(pfnIsChecked,"Missing callback function");
	Asura_Assert(pfnSetChecked,"Missing callback function");

	m_pData = 0;
	m_iData = 0;
	m_pfnVoidIsChecked = pfnIsChecked;
	m_pfnVoidSetChecked = pfnSetChecked;
	m_pfnPtrIsChecked = 0;
	m_pfnPtrSetChecked = 0;
	m_pfnIntIsChecked = 0;
	m_pfnIntSetChecked = 0;
}

Asura_FunctionTickbox_Gadget::Asura_FunctionTickbox_Gadget(const Asura_Vector_2& xOffset, void* pData, bool(*pfnIsChecked)(void* pUserData), void(*pfnSetChecked)(bool bSet,void* pUserData))
:Asura_BaseTickbox_Gadget(xOffset,0)
{
	Asura_Assert(pfnIsChecked,"Missing callback function");
	Asura_Assert(pfnSetChecked,"Missing callback function");

	m_pData = pData;
	m_iData = 0;
	m_pfnVoidIsChecked = 0;
	m_pfnVoidSetChecked = 0;
	m_pfnPtrIsChecked = pfnIsChecked;
	m_pfnPtrSetChecked = pfnSetChecked;
	m_pfnIntIsChecked = 0;
	m_pfnIntSetChecked = 0;
}

Asura_FunctionTickbox_Gadget::Asura_FunctionTickbox_Gadget(const Asura_Vector_2& xOffset, int iData, bool(*pfnIsChecked)(int iUserData), void(*pfnSetChecked)(bool bSet,int iUserData))
:Asura_BaseTickbox_Gadget(xOffset,0)
{
	Asura_Assert(pfnIsChecked,"Missing callback function");
	Asura_Assert(pfnSetChecked,"Missing callback function");

	m_pData = 0;
	m_iData = iData;
	m_pfnVoidIsChecked = 0;
	m_pfnVoidSetChecked = 0;
	m_pfnPtrIsChecked = 0;
	m_pfnPtrSetChecked = 0;
	m_pfnIntIsChecked = pfnIsChecked;
	m_pfnIntSetChecked = pfnSetChecked;
}


bool Asura_FunctionTickbox_Gadget::IsChecked() const
{
	if(m_pfnVoidIsChecked)
	{
		return m_pfnVoidIsChecked();
	}
	else if(m_pfnPtrIsChecked)
	{
		return m_pfnPtrIsChecked(m_pData);
	}
	else if(m_pfnIntIsChecked)
	{
		return m_pfnIntIsChecked(m_iData);
	}
	return false;
}

void Asura_FunctionTickbox_Gadget::SetChecked(bool bSet)
{
	if(m_pfnVoidSetChecked)
	{
		m_pfnVoidSetChecked(bSet);
	}
	else if(m_pfnPtrSetChecked)
	{
		m_pfnPtrSetChecked(bSet,m_pData);
	}
	else if(m_pfnIntSetChecked)
	{
		m_pfnIntSetChecked(bSet,m_iData);
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Minimise_Gadget::Asura_Minimise_Gadget( const Asura_Vector_2& offset, u_int* puMinimisedState, bool bMinimiseToBottom)
{
	m_xOffset = offset;
	m_puMinimisedState = puMinimisedState;
	m_bMinimiseToBottom = bMinimiseToBottom;
	
	m_fWidth = 1.0f;
	m_fHeight = 1.0f;
}

void Asura_Minimise_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	

	unsigned int colour = GetColour( m_bClickedThisFrame );

	u_int uTargetMiniState = m_bMinimiseToBottom ? ASURA_OVERLAY_MINIMISED_ATBOTTOM : ASURA_OVERLAY_MINIMISED_INPLACE;

	if (uTargetMiniState == *m_puMinimisedState)
	{
		Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",7);
	}
    else
	{
		Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",m_bMinimiseToBottom ? 9 : 13);
	}

    Asura_Gadget::Render( xOrigin );
}

void Asura_Minimise_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable )
	{
		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
		{
			u_int uTargetMiniState = m_bMinimiseToBottom ? ASURA_OVERLAY_MINIMISED_ATBOTTOM : ASURA_OVERLAY_MINIMISED_INPLACE;

			*m_puMinimisedState = (*m_puMinimisedState == uTargetMiniState) ? ASURA_OVERLAY_MINIMISED_NOTMINIMISED : uTargetMiniState;
    	    m_bClickedThisFrame = true;
		}
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Close_Gadget::Asura_Close_Gadget( const Asura_Vector_2& offset, bool *shouldClose)
{
	m_xOffset = offset;
	m_pbShouldClose = shouldClose;
	
	m_fWidth = 1.0f;
	m_fHeight = 1.0f;
}

void Asura_Close_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	

	unsigned int colour = GetColour( m_bClickedThisFrame );

	Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",8);

    Asura_Gadget::Render( xOrigin );
}

void Asura_Close_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
		*m_pbShouldClose = true;
        m_bClickedThisFrame = true;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_Resize_Gadget::Asura_Resize_Gadget(const Asura_Vector_2& xOffset)
{
	m_xOffset = xOffset;
	
	m_fWidth = 13.0f/ASURA_TEXT_OUTPUT_FONT_WIDTH;
	m_fHeight = 13.0f/ASURA_TEXT_OUTPUT_FONT_HEIGHT;
}

void Asura_Resize_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	unsigned int uColour = Asura_OverlayBrightness | GetColour( m_bClickedThisFrame );

	Asura_Vector_2 xPos = xOrigin + m_xOffset;
	float fX = xPos.x * ASURA_TEXT_OUTPUT_FONT_WIDTH;
	float fY = xPos.y * ASURA_TEXT_OUTPUT_FONT_HEIGHT;


//	Asura_Render::Render2DLine(fX, fY+9.0f, fX+10.0f, fY-1.0f, uColour);
//	Asura_Render::Render2DLine(fX+3.0f, fY+9.0f, fX+10.0f, fY+2.0f, uColour);
//	Asura_Render::Render2DLine(fX+6.0f, fY+9.0f, fX+10.0f, fY+5.0f, uColour);

	Asura_Render::Render2DLine(fX, 		fY+12.0f,	fX+13.0f,	fY-1.0f, uColour);
	Asura_Render::Render2DLine(fX+3.0f,	fY+12.0f,	fX+13.0f,	fY+2.0f, uColour);
	Asura_Render::Render2DLine(fX+6.0f,	fY+12.0f,	fX+13.0f,	fY+5.0f, uColour);
	Asura_Render::Render2DLine(fX+9.0f,	fY+12.0f,	fX+13.0f,	fY+8.0f, uColour);
    Asura_Gadget::Render( xOrigin );
}

void Asura_Resize_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable && m_pxDialog)
	{
		u_int uResizeFlags = (ASURA_OVERLAY_RESIZE_MAXX|ASURA_OVERLAY_RESIZE_MAXY) & m_pxDialog->GetAllowedResizeFlags();
		if(uResizeFlags)
		{
			if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				GetOverlay()->StartResize(m_pxDialog, uResizeFlags);
			}
			else if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				GetOverlay()->SetResizeCursor(uResizeFlags);
			}
		}
	}
}
//////////////////////////////////////////////////////////////////////////////////////////////////


const void* Asura_Function_Gadget::s_pLastFunction = 0;

Asura_Function_Gadget::Asura_Function_Gadget( const Asura_Vector_2& offset,  const Asura_Callback_Base<void>& xCallback, bool* flag)
{
	m_xOffset = offset;
	m_xOnClickedCallback.SetCallback(&xCallback);
	m_pbFlag = flag;
	
	m_fWidth = 1.0f;
	m_fHeight = 1.0f;

}



void Asura_Function_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;

	unsigned int colour = GetColour( m_bClickedThisFrame );


	Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",10);

    Asura_Gadget::Render( xOrigin );
}

void Asura_Function_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
		if (m_pbFlag) *m_pbFlag = true;
		if (m_xOnClickedCallback.IsValid())
		{
			Asura_Overlay_System::SetCallbackFunctionToCall(m_xOnClickedCallback);
			//Take note of the function called , for the repeat last function key.
			//Ignore any functions called while a modal overlay is up , since it is unlikely
			//that they would want to be repeated in this way. RAR 05/10/2005
			if(!Asura_Overlay_System::AreAnyOverlaysAlwaysOnTop())
			{
				s_pLastFunction = m_xOnClickedCallback.GetFunctionPointer();
			}
		}
        m_bClickedThisFrame = true;
	}
}

void Asura_Function_Gadget::RepeatLastFunction()
{
	if(!IsEnabled()) return;

	if (s_pLastFunction && m_xOnClickedCallback.IsValid() && s_pLastFunction==m_xOnClickedCallback.GetFunctionPointer())
	{
		Asura_Overlay_System::SetCallbackFunctionToCall(m_xOnClickedCallback);
       	m_bClickedThisFrame = true;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
Asura_FunctionWithKeybinding_Gadget::Asura_FunctionWithKeybinding_Gadget( const Asura_Vector_2& xOffset, const Asura_Callback_Base<void>& xCallback, const Asura_Key_Combination& xBinding, bool* pbFlag)
:Asura_Function_Gadget(xOffset, xCallback, pbFlag)
{
	m_xBinding = xBinding;
}



void Asura_FunctionWithKeybinding_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	//ignore keypresses if something else has already used a key , or if some other gadget has keyboard focus
	if(!Asura_Overlay_System::s_bUsedKeypressThisFrame && !Asura_Overlay_System::OverlaysHaveKeyboardFocus())
	{
		if(m_xBinding.DebouncedKeyPressed())
		{
			if (IsEnabled())
			{
				if (m_pbFlag) *m_pbFlag = true;
				if( m_xOnClickedCallback.IsValid())
				{
					Asura_Overlay_System::SetCallbackFunctionToCall(m_xOnClickedCallback);
				}
    	        m_bClickedThisFrame = true;
				Asura_Overlay_System::s_bUsedKeypressThisFrame = true;

			}
		}
	}
	Asura_Function_Gadget::Update( xOrigin );
}


//////////////////////////////////////////////////////////////////////////////////////////////////

Asura_TextSelect_Gadget::Asura_TextSelect_Gadget( 
	const Asura_Vector_2& xOffset,
	int* const piSelection,
	const int iSelectionValue,
	int* const piListOffset,
	const char* const* ppszList,
	const int iNumberOfLines,
	const int iNumberInList,
	void( *pfnFunction )( void ),
	const bool bOnlySelectOnClick /*= false*/ )
: m_piSelection( piSelection )
, m_iSelectionValue( iSelectionValue )
, m_piListOffset( piListOffset )
, m_ppszList( ppszList )
, m_iNumberOfLines( iNumberOfLines )
, m_iNumberInList( iNumberInList )
, m_pfnFunction( pfnFunction )
, m_bOnlySelectOnClick( bOnlySelectOnClick )
, m_ppszHelpText( NULL )
, m_pfnDisplayTooltipCallback( NULL )
, m_pxMultipleSelectionHandler( NULL )
, m_piIndexRemap(NULL)
{
	m_xOffset	= xOffset;
	m_fWidth	= 20.0f;
	m_fHeight	= 1.0f;
}

void Asura_TextSelect_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;

	unsigned int colour = GetColour( m_bClickedThisFrame );
	
	int iIndexInList = *m_piListOffset + m_iSelectionValue;
	if(m_piIndexRemap && iIndexInList<m_iNumberInList)
	{
		//get the index into the original unsorted list
		iIndexInList = m_piIndexRemap[iIndexInList];
	}

	if( ( iIndexInList < m_iNumberInList ) &&
		( m_ppszList[ iIndexInList ] ) )
	{
		if (IsSelected()) colour = 0xffffff00;

		if( ( m_pxMultipleSelectionHandler ) && 
			( m_pxMultipleSelectionHandler->IsInDragSelection( iIndexInList ) ) )
		{
			//used the 'clicked' colour for all elements in drag selection
			colour = GetColour(true);
		}
		Asura_Text_Output::Print( pos.x, pos.y, Asura_OverlayBrightness | colour, ASURA_TEXT_FORMAT_LEFTJUSTIFIED | ASURA_TEXT_FORMAT_SMALLFONT, "%s", m_ppszList[ iIndexInList ] );
	}	
    Asura_Gadget::Render( xOrigin );
}

void Asura_TextSelect_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	int iIndexInList = *m_piListOffset + m_iSelectionValue;
	if( iIndexInList >= m_iNumberInList )
	{
		return;
	}
	
	if( m_bIsMouseOver )
	{
		// bring up the help overlay (if there is one) if the user keeps the cursor still over something
		if( !Asura_Overlay_System::IsAnyHelpText() && Asura_Cursor::IsStatic() )
		{
			int iHelpStringIndex = iIndexInList;
			if(m_piIndexRemap)
			{
				//get the index into the original unsorted list
				iHelpStringIndex = m_piIndexRemap[iHelpStringIndex];
				Asura_Assert(iHelpStringIndex<m_iNumberInList, "");
			}

			//display help text for this item
			if(m_pfnDisplayTooltipCallback)
			{
				m_pfnDisplayTooltipCallback(iHelpStringIndex, m_ppszList[iHelpStringIndex]);
			}
			else if(m_ppszHelpText && m_ppszHelpText[ iHelpStringIndex ])
			{
				Asura_Overlay_System::SetHelpText( m_ppszHelpText[ iHelpStringIndex ] );
			}
		}
		

		if( m_pxMultipleSelectionHandler )
		{
			//update multiple selection.
			m_pxMultipleSelectionHandler->Update( ( *m_piListOffset + m_iSelectionValue ), m_bIsClickable );
		}

		
        if( m_bIsClickable )
        {
			if(Asura_UserInput::KeyPressed( ASURA_KEYID_DOUBLECLICK_MOUSEBUTTON1 ))
			{
		    	*m_piSelection = m_iSelectionValue;
				Asura_Overlay_System::SetCallbackFunctionToCall( m_xDoubleClickCallback );

                m_bClickedThisFrame = true;
			}
		    else if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_MOUSEBUTTON1 ))
		    {
		    	*m_piSelection = m_iSelectionValue;
				Asura_Overlay_System::SetCallbackFunctionToCall( m_xSingleClickCallback );

                m_bClickedThisFrame = true;
		    }
		    if( Asura_UserInput::DebouncedKeyPressed ( ASURA_KEYID_MOUSEBUTTON2 ) )
		    {
				Tools_ContextMenu_System::SetRightClickUsed();

				*m_piSelection = m_iSelectionValue;
			    if ( m_pfnFunction != NULL )
				{
					Asura_Overlay_System::SetCallbackFunctionToCall( m_pfnFunction );
				}
                m_bClickedThisFrame = true;
		    }
        }
	}


	if(m_bIsMouseOver)
	{
		HandleKeyboard();
	}
}

void Asura_TextSelect_Gadget::HandleKeyboard()
{
	const bool bShiftPressed		= Asura_UserInput::KeyPressed( ASURA_KEYID_SHIFT );
	const int iWheelDelta			= ( bShiftPressed ) ? m_iNumberOfLines : 3;			// holding shift does a 'page' at a time
	const int iCursorDelta			= ( bShiftPressed ) ? 5 : 1;						//
	const int iPagesDelta			= ( bShiftPressed ) ? 5 : 1;						//
	const int iHighestSelection		= Asura_Maths::Min( m_iNumberOfLines, m_iNumberInList ) - 1;
	const int iHighestOffset		= m_iNumberInList - ( iHighestSelection + 1 );

	//RAR 08/03/2006 Key presses ignored while control held down ,to allow that combination to be used by list ordering overlay
	if( ( !Asura_Overlay_System::s_bUsedKeypressThisFrame ) && 
		( !Asura_UserInput::KeyPressed( ASURA_KEYID_CTRL ) ) )
	{
		//mouse wheel scrolls through the list
		if( m_iNumberInList > m_iNumberOfLines )
		{
			const int iOffset = *m_piListOffset;
			if (Asura_UserInput::KeyPressed( ASURA_KEYID_MOUSEWHEELUP ) )
			{
				*m_piListOffset -= iWheelDelta;
				if( *m_piListOffset < 0 )
				{
					*m_piListOffset = 0;
				}
				Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
			}
			else if( Asura_UserInput::KeyPressed( ASURA_KEYID_MOUSEWHEELDOWN ) )
			{
				*m_piListOffset += iWheelDelta;
				if( *m_piListOffset > iHighestOffset )
				{
					*m_piListOffset = iHighestOffset;
				}
				Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
			}
			if( m_bOnlySelectOnClick )
			{
				*m_piSelection += ( iOffset - *m_piListOffset );
			}
		}
		
		int iDelta = 0;
		if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_UP ) )
		{
			iDelta -= iCursorDelta;
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}
		if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_DOWN ) )
		{
			iDelta += iCursorDelta;
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}

		const bool bIsOnPage = ( ( *m_piSelection >= 0 ) && ( *m_piSelection <= iHighestSelection ) );
		if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_PAGEUP ) )
		{
			//determine number of whole pages to move
			if( ( m_bOnlySelectOnClick ) && ( !bIsOnPage ) )
			{
				iDelta -= ( iPagesDelta * m_iNumberOfLines );
			}
			else
			{
				const int iPages = ( *m_piSelection == 0 ) ? iPagesDelta : iPagesDelta - 1;
				iDelta -= ( iPages * m_iNumberOfLines );
				iDelta -= *m_piSelection;
			}
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}
		if( Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_PAGEDOWN ) )
		{
			//determine number of whole pages to move
			if( ( m_bOnlySelectOnClick ) && ( !bIsOnPage ) )
			{
				iDelta += ( iPagesDelta * m_iNumberOfLines );
			}
			else
			{
				const int iPages = ( *m_piSelection == iHighestSelection ) ? iPagesDelta : iPagesDelta - 1;
				iDelta += ( iPages * m_iNumberOfLines );
				iDelta += ( iHighestSelection - *m_piSelection );
			}
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}

		if( iDelta != 0 )
		{
			*m_piSelection += iDelta;
			if( *m_piSelection < 0 )
			{
				*m_piListOffset += *m_piSelection;
				if( *m_piListOffset < 0 )
				{
					*m_piListOffset = 0;
				}
				*m_piSelection = 0;
			}
			if( *m_piSelection > iHighestSelection )
			{
				*m_piListOffset += ( *m_piSelection - iHighestSelection );
				if( *m_piListOffset > iHighestOffset )
				{
					*m_piListOffset = iHighestOffset;
				}
				*m_piSelection = iHighestSelection;
			}
		}


		if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_HOME))
		{
			//Home - goto start of list. RAR 02/04/03
			*m_piSelection	= 0;
			*m_piListOffset = 0;
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}
		if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_END))
		{
			//End - goto end of list. RAR 02/04/03
			*m_piSelection	= iHighestSelection;
			*m_piListOffset	= iHighestOffset;
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
		}
	}
}

void Asura_TextSelect_Gadget::SetSelection(int iIndex, bool bAffectMultipleSelection /*=true*/)
{
	if(m_iNumberOfLines > 0)
	{
		//force selection to be in range
		iIndex = Asura_Maths::ClampToRange( iIndex, 0, m_iNumberInList );

		if( m_pxMultipleSelectionHandler && bAffectMultipleSelection)
		{
			m_pxMultipleSelectionHandler->ClearSelection();
			m_pxMultipleSelectionHandler->SetSelected( iIndex, true );
		}

		//update Selection and ListOffset so that they add up to the right amount
		*m_piSelection = iIndex - *m_piListOffset;
		Asura_List_Dialog::AdjustListSelectionAndOffset( m_piListOffset, m_piSelection, m_iNumberOfLines, m_iNumberInList );
	}
}

bool Asura_TextSelect_Gadget::IsSelected() const
{
	if( m_pxMultipleSelectionHandler )
	{
		return( m_pxMultipleSelectionHandler->IsSelected( *m_piListOffset + m_iSelectionValue ) );
	}
	else
	{
		return( *m_piSelection == m_iSelectionValue );
	}
}

void Asura_TextSelect_Gadget::SetListHelpStrings(const char* const* ppszHelpText)
{
	m_ppszHelpText = ppszHelpText;
	m_pfnDisplayTooltipCallback = 0;
}

void Asura_TextSelect_Gadget::SetTooltipCallback( void (*pfnDisplayTooltipCallback)(int iIndex,const char* szText))
{
	m_ppszHelpText = NULL;
	m_pfnDisplayTooltipCallback = pfnDisplayTooltipCallback;
}


//////////////////////////////////////////////////////////////////////////////////////////////////


Asura_UpDown_Gadget::Asura_UpDown_Gadget( const Asura_Vector_2& offset, int* value, int min, int max, int delta)
{
	m_xOffset = offset;
	m_piValue = value;
	m_iMin = min;
	m_iMax = max;
	m_iDelta = delta;

	m_fWidth = 1.0f;
	m_fHeight = 1.0f;
}

void Asura_UpDown_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	

	unsigned int colour = GetColour( m_bClickedThisFrame );


	unsigned char c;
	if (m_iDelta>0)
	{
		c = 11;
	}
	else
	{
		c = 12;
	}

	Asura_Text_Output::Print(pos.x,pos.y,Asura_OverlayBrightness|colour,ASURA_TEXT_FORMAT_LEFTJUSTIFIED,"%c",c);

    Asura_Gadget::Render( xOrigin );
}

void Asura_UpDown_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable )
	{
		if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
		{
			*m_piValue += m_iDelta;
            m_bClickedThisFrame = true;
		}
		if (Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON2))
		{
			Tools_ContextMenu_System::SetRightClickUsed();

			*m_piValue += m_iDelta;
            m_bClickedThisFrame = true;
		}

		if (*m_piValue < m_iMin) *m_piValue = m_iMin;
		else if (*m_piValue > m_iMax) *m_piValue = m_iMax;
		
	}

}


Asura_Button_Gadget::Asura_Button_Gadget( const Asura_Vector_2& offset,int textureIndex,int textureSize,Asura_Vector_2 buttonPixelPos,Asura_Vector_2 buttonPixelSize,void(*clickedFunction)(void))
{
	m_xOffset = offset;
	m_iTextureIndex = textureIndex;
	m_iTextureSize = textureSize;
	m_xButtonPixelPos = buttonPixelPos;
	m_xButtonPixelSize = buttonPixelSize;
	m_pfnClickedFunction = clickedFunction;
	m_pfnClickedContextFunction = NULL;
	m_pContext = NULL;

	m_fWidth = (m_xButtonPixelSize.x-1)/ASURA_TEXT_OUTPUT_FONT_WIDTH;
	m_fHeight = (m_xButtonPixelSize.y-1)/ASURA_TEXT_OUTPUT_FONT_HEIGHT;
}

Asura_Button_Gadget::Asura_Button_Gadget( const Asura_Vector_2& offset,int textureIndex,int textureSize,Asura_Vector_2 buttonPixelPos,Asura_Vector_2 buttonPixelSize,void(*clickedContextFunction)(void*),void* pContext)
{
	m_xOffset = offset;
	m_iTextureIndex = textureIndex;
	m_iTextureSize = textureSize;
	m_xButtonPixelPos = buttonPixelPos;
	m_xButtonPixelSize = buttonPixelSize;
	m_pfnClickedFunction = NULL;
	m_pfnClickedContextFunction = clickedContextFunction;
	m_pContext = pContext;

	m_fWidth = (m_xButtonPixelSize.x-1)/ASURA_TEXT_OUTPUT_FONT_WIDTH;
	m_fHeight = (m_xButtonPixelSize.y-1)/ASURA_TEXT_OUTPUT_FONT_HEIGHT;
}


void Asura_Button_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 pos = xOrigin + m_xOffset;
	

	unsigned int colour = GetColour( m_bClickedThisFrame );

	Asura_Render::Overlay.Colour = Asura_OverlayBrightness|colour;
	Asura_Render::Overlay.TextureIndex = m_iTextureIndex;
	Asura_Render::Overlay.m_eTextureOp = ASURA_TEXTUREOP_VIEWTEXTURECOLOUR;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;

	float x=Asura_Maths::FloatToInt(pos.x*ASURA_TEXT_OUTPUT_FONT_WIDTH)-0.5f;
	float y=Asura_Maths::FloatToInt(pos.y*ASURA_TEXT_OUTPUT_FONT_HEIGHT)-0.5f;
	
	Asura_Render::Overlay.SetRectUV(m_xButtonPixelPos.x/m_iTextureSize,m_xButtonPixelPos.y/m_iTextureSize,m_xButtonPixelSize.x/m_iTextureSize,m_xButtonPixelSize.y/m_iTextureSize);
	Asura_Render::Overlay.SetRectXY(x,y,m_xButtonPixelSize.x,m_xButtonPixelSize.y);

	Asura_Render::Overlay.Render();

    Asura_Gadget::Render( xOrigin );
}

void Asura_Button_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
		if(m_pfnClickedFunction) m_pfnClickedFunction();
		if(m_pfnClickedContextFunction) m_pfnClickedContextFunction(m_pContext);
        m_bClickedThisFrame = true;
	}
}

//
// Asura_StringDisplay_Gadget.
//

Asura_StringDisplay_Gadget::Asura_StringDisplay_Gadget( const Asura_Vector_2& xOffset,
                                                        u_int uCharLimit, /* = 0 */
                                                        BOOL_CALLBACK pfnIsOn, /* = NULL */
                                                        VOID_CALLBACK pfnOnClick, /* = NULL */
                                                        int iCallbackArg, /* = 0 */
														int iFormatFlags /* = 0 */)
: Asura_Gadget()
{
    m_xOffset         = xOffset;
    m_fWidth          = static_cast<float>(uCharLimit); // Width is fixed iff char limit is set.
    m_fHeight         = 1.0f;
    m_szLabel      = NULL;
    m_uCharLimit   = uCharLimit;
    m_bHighlight   = false;
    m_pfnIsOn      = pfnIsOn;
    m_pfnOnClick   = pfnOnClick;
    m_iCallbackArg = iCallbackArg;
	m_iFormatFlags = iFormatFlags | ASURA_TEXT_FORMAT_LEFTJUSTIFIED;
}

Asura_StringDisplay_Gadget::~Asura_StringDisplay_Gadget( void )
{
    if( m_szLabel ) { delete[] m_szLabel; }
}

void Asura_StringDisplay_Gadget::Render( const Asura_Vector_2& xOrigin )
{
    if( m_szLabel )
    {
	    Asura_Text_Output::Print( xOrigin.x + this->m_xOffset.x,
                                  xOrigin.y + this->m_xOffset.y,
                                  Asura_OverlayBrightness | GetColour(m_bClickedThisFrame || (m_pfnIsOn && m_pfnIsOn(m_iCallbackArg))),
                                  m_iFormatFlags,
                                  m_szLabel );
    }
    
    Asura_Gadget::Render( xOrigin );
}

void Asura_StringDisplay_Gadget::SetLabel( const char* szLabel )
{
	if( !szLabel )
	{
		Asura_Assert( false, "" );
		return;
	}
	
	if( m_szLabel )
    {
        delete[] m_szLabel;
        m_szLabel = NULL;
    }

    if( !m_uCharLimit ) { m_fWidth = 0.0f; }

	u_int uLen = m_uCharLimit ? Asura_Maths::Min( m_uCharLimit, static_cast<u_int>(strlen(szLabel)) ) : static_cast<u_int>(strlen(szLabel));

    m_szLabel = new char[uLen + 1];
    if( m_szLabel )
    {
        if( uLen )
        {
            strncpy( m_szLabel, szLabel, uLen );
        }
        m_szLabel[uLen] = '\0';

        if( !m_uCharLimit ) { m_fWidth += uLen; }
    }
}

void Asura_StringDisplay_Gadget::Update( const Asura_Vector_2& xOrigin )
{
    if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
    {
        if( m_pfnOnClick )
        {
            m_pfnOnClick( m_iCallbackArg );
        }
        m_bClickedThisFrame = true;
    }
}


//////////////////////////////////////////////////////////////////////////////////////////////////

Asura_VerticalScroll_Gadget::Asura_VerticalScroll_Gadget(const Asura_Vector_2 xOffset,float fHeight,int iNumItemsShown,int* piValue,int iMin,int iMax)
{
	m_xOffset = xOffset;
	m_fWidth = 1.0f;
	m_fHeight = fHeight;

	m_piValue = piValue;
	m_iNumItemsShown = iNumItemsShown;

	m_fSliderPosition = 0.0f;

	SetMinAndMax(iMin,iMax);	

}

void Asura_VerticalScroll_Gadget::SetMinAndMax(int iMin,int iMax)
{
	m_iMin = iMin;
	m_iMax = iMax;

	Asura_Assert(m_iMin<=m_iMax,"");

	//calculate the size of the slider according to the proportion of list items shown at a time
	m_fSliderHeight = (m_iNumItemsShown / (float)(m_iNumItemsShown + iMax - iMin)) * m_fHeight;

	//don't allow it to get too small though
	if(m_fSliderHeight < 0.5f) m_fSliderHeight = 0.5f;

}

void Asura_VerticalScroll_Gadget::Update( const Asura_Vector_2& xOrigin )
{	
	
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
 		Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
        m_bClickedThisFrame = true;
	}
		
 	if (Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
 	{
 		if(m_iFocusFlags)
 		{
			//move the slider to the position clicked on
			float cx,cy;
			Asura_Cursor::GetScreenCoords(cx,cy);
			//get coord relative to start of bar
			cy -= (xOrigin.y + m_xOffset.y);
			if(cy < 0) cy = 0.0f;
			else if(cy > m_fHeight) cy = m_fHeight;

			if(m_iMax>m_iMin)
			{
				int iValue = (int)(m_iMin + (cy / m_fHeight) * (m_iMax - m_iMin));
				if(iValue < m_iMin) iValue = m_iMin;
				if(iValue > m_iMax) iValue = m_iMax;

				*m_piValue = iValue;
			}
			
			
		}
	}
	else if(m_iFocusFlags)
	{
		Asura_Overlay_System::ReleaseFocus(OverlayFocusType_Mouse);
	}
	
	UpdateSliderPosition();
}

void Asura_VerticalScroll_Gadget::Render( const Asura_Vector_2& xOrigin )
{	
	//draw the bar on which the number slider lives
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	
	unsigned int uColour = 0x7f7f7f7f;//0000;

	Asura_Render::Overlay.Colour = uColour;
	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

	float x=Asura_Maths::FloatToInt((xOrigin.x + m_xOffset.x)*ASURA_TEXT_OUTPUT_FONT_WIDTH)-0.5f;
	float y=Asura_Maths::FloatToInt((xOrigin.y + m_xOffset.y)*ASURA_TEXT_OUTPUT_FONT_HEIGHT)-0.5f;
	
	Asura_Render::Overlay.SetRectXY(x,y,m_fWidth*ASURA_TEXT_OUTPUT_FONT_WIDTH,m_fHeight*ASURA_TEXT_OUTPUT_FONT_HEIGHT);

	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();
	
	//draw the slider
	UpdateSliderPosition();

	Asura_Render::Overlay.Colour = 0xaf7f7faf;

	x=Asura_Maths::FloatToInt((xOrigin.x + m_xOffset.x)*ASURA_TEXT_OUTPUT_FONT_WIDTH)+0.5f;
	y=Asura_Maths::FloatToInt((xOrigin.y + m_xOffset.y+m_fSliderPosition)*ASURA_TEXT_OUTPUT_FONT_HEIGHT)+0.5f;

	Asura_Render::Overlay.SetRectXY(x,y,m_fWidth*ASURA_TEXT_OUTPUT_FONT_WIDTH - 2.0f,m_fSliderHeight*ASURA_TEXT_OUTPUT_FONT_HEIGHT - 2.0f);
	Asura_Render::Overlay.Render();

	Asura_Gadget::Render( xOrigin );
}

void Asura_VerticalScroll_Gadget::UpdateSliderPosition()
{
	//calculate where the gadget should be along its slidy thing
	if(m_iMin < m_iMax)
	{
		float fProportion = (*m_piValue - m_iMin) / (float)(m_iMax - m_iMin);
		if(fProportion > 1) fProportion = 1.0f;
		else if(fProportion < 0) fProportion = 0.0f;

		m_fSliderPosition = (m_fHeight - m_fSliderHeight) * fProportion;
	}
	else
	{
		m_fSliderPosition = 0;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////


const float Asura_Icon_Gadget::s_fUVWidth = 16.0f/256.0f;
const float Asura_Icon_Gadget::s_fUVHeight = 16.0f/256.0f;

const float Asura_Icon_Gadget::s_fWidth = 16/ASURA_TEXT_OUTPUT_FONT_WIDTH;
const float Asura_Icon_Gadget::s_fHeight = 16/ASURA_TEXT_OUTPUT_FONT_HEIGHT;

Asura_Icon_Gadget::Asura_Icon_Gadget(const Asura_Vector_2& xOffset,int iTextureIndex,int iIconIndex)
{
	Asura_Assert(iIconIndex>=0 && iIconIndex<256,"Icon index out of range");

	m_xOffset = xOffset;
	m_iTextureIndex = iTextureIndex;
	m_iIconX = iIconIndex % 16;
	m_iIconY = iIconIndex / 16;

	m_fWidth = s_fWidth;
	m_fHeight = s_fHeight;

	m_pbClickedFlag = NULL;

	m_pbValue = NULL;
	m_puFlagsField = NULL;
	m_uFlagBit = 0;

}


void Asura_Icon_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 xPos = xOrigin + m_xOffset;
	
	bool bEnabled = IsEnabled();

	unsigned int uColour;
	unsigned int uBrightness = bEnabled ? 0xff : 0x7f;
	if(ShowAsPressed())
	{
		if((m_bIsMouseOver || m_bHighlight) && bEnabled)
		{
			uColour = 0x00010100 * uBrightness;
		}
		else
		{
			uColour = 0x00010000 * uBrightness;
		}
	}
	else
	{
		if((m_bIsMouseOver || m_bHighlight) && bEnabled)
		{
			uColour = 0x00000100 * uBrightness;
		}
		else
		{
			uColour = 0x00010101 * uBrightness;
		}
	}

	Asura_Render::Overlay.Colour = Asura_OverlayBrightness|uColour;
	Asura_Render::Overlay.TextureIndex = m_iTextureIndex;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

	float x=(float)Asura_Maths::FloatToInt_Near(xPos.x*ASURA_TEXT_OUTPUT_FONT_WIDTH)/*-0.5f*/;
	float y=(float)Asura_Maths::FloatToInt_Near(xPos.y*ASURA_TEXT_OUTPUT_FONT_HEIGHT)/*-0.5f*/;
	
	Asura_Render::Overlay.SetRectUV(m_iIconX*s_fUVWidth,m_iIconY*s_fUVHeight,s_fUVWidth,s_fUVHeight);
	Asura_Render::Overlay.SetRectXY(x,y,17,17);

	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();


    Asura_Gadget::Render( xOrigin );
}

void Asura_Icon_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if(IsEnabled())
	{
		if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
		{
			if(m_iRadioButtonGroup)
			{
				GetOverlay()->RadioButtonPressed(this);
			}
			else
			{
				if(m_pbValue) *m_pbValue = !*m_pbValue;
				if(m_puFlagsField) *m_puFlagsField ^= m_uFlagBit;
				if(m_xSetCheckedCallback.IsValid()) m_xSetCheckedCallback.Execute(!IsChecked());
			}
			
			if (m_pbClickedFlag) *m_pbClickedFlag = true;
			if (m_xClickedCallback.IsValid()) Asura_Overlay_System::SetCallbackFunctionToCall(m_xClickedCallback);

			m_bClickedThisFrame = true;
		}
	}
}

bool Asura_Icon_Gadget::ShowAsPressed()
{
	if(m_pbValue) return *m_pbValue;
	if(m_puFlagsField) return (*m_puFlagsField & m_uFlagBit)!=0;
	if(m_xIsCheckedCallback.IsValid()) return m_xIsCheckedCallback.Execute();
	return m_bClickedThisFrame;
}

bool Asura_Icon_Gadget::IsChecked() const
{
	if(m_pbValue) return *m_pbValue;
	if(m_puFlagsField) return (*m_puFlagsField & m_uFlagBit) ? true : false;
	if(m_xIsCheckedCallback.IsValid()) return m_xIsCheckedCallback.Execute();
	return false;
}

void Asura_Icon_Gadget::SetChecked(bool bSet)
{
	if(m_pbValue) *m_pbValue = bSet;
	if(m_puFlagsField) bSet ? (*m_puFlagsField |= m_uFlagBit) : (*m_puFlagsField &=~ m_uFlagBit);
	if(m_xSetCheckedCallback.IsValid()) m_xSetCheckedCallback.Execute(bSet);
}

void Asura_Icon_Gadget::SetClickedFunction(const Asura_Callback_Base<void>& xClicked)
{
	m_xClickedCallback.SetCallback(&xClicked);
}

void Asura_Icon_Gadget::SetCheckboxFunction(const Asura_Callback_Base<bool>& xIsChecked, const Asura_Callback_Base_1Arg<void,bool>& xSetChecked)
{
	m_xIsCheckedCallback.SetCallback(&xIsChecked);
	m_xSetCheckedCallback.SetCallback(&xSetChecked);

}


/// Asura_Tab_Gadget //////////////////////////////////////////////////////////////////////////////

Asura_Tab_Gadget::Asura_Tab_Gadget(const char* szName)
: m_pxName(NULL)
, m_bSelected(false)
, m_uRow(0)
, m_bBottomRow(false)
{
	m_xOffset.Zero();
//	m_fWidth = static_cast<float>(strlen(szName));
	m_fWidth = 2.0f+Asura_Text_Output::WidthOfString(szName, ASURA_TEXT_FORMAT_SMALLFONT);
	m_fHeight = 1.0f;
	m_pxName = new Asura_String(szName);
}

Asura_Tab_Gadget::~Asura_Tab_Gadget()
{
	if(m_pxName) delete m_pxName;
}

void Asura_Tab_Gadget::AddDialog(Asura_Dialog* pxDialog)
{
	Asura_Assert(pxDialog, "No dialog");
	m_xDialogList.Push_Front(pxDialog);
}

void Asura_Tab_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	Asura_Vector_2 xPos = xOrigin + m_xOffset;

	const float fYOffset = (ASURA_TAB_GADGET_HEIGHT - 1.f) * 0.5f;
	Asura_Text_Output::Print(xPos.x + static_cast<float>(m_fWidth) * 0.5f, xPos.y + fYOffset,Asura_OverlayBrightness|0x00ffffff,ASURA_TEXT_FORMAT_CENTREJUSTIFIED|ASURA_TEXT_FORMAT_SMALLFONT,"%s",m_pxName->GetString());
	float fLeft = xPos.x * ASURA_TEXT_OUTPUT_FONT_WIDTH;
	float fRight = (xPos.x + m_fWidth) * ASURA_TEXT_OUTPUT_FONT_WIDTH - 1.f;
	float fTop = xPos.y * ASURA_TEXT_OUTPUT_FONT_HEIGHT + 2.f;
	float fBottom = (xPos.y + ASURA_TAB_GADGET_HEIGHT) * ASURA_TEXT_OUTPUT_FONT_HEIGHT;
	if(m_bSelected)
	{
		fTop -= 1.f;
	}
	else
	{
		// Bottom
		if(m_bBottomRow)
		{
			Asura_Render::Render2DLine(fLeft, fBottom, fRight + 1.f, fBottom, Asura_OverlayBrightness|0x00fffffff);
		}
		else
		{
			Asura_Render::Render2DLine(fLeft, fBottom, fRight + 1.f, fBottom, Asura_OverlayBrightness|0x00000000);
		}
	}
	// Left
	Asura_Render::Render2DLine(fLeft, fTop, fLeft, fBottom, Asura_OverlayBrightness|0x00fffffff);
	// Top
	Asura_Render::Render2DLine(fLeft, fTop, fRight, fTop, Asura_OverlayBrightness|0x00fffffff);
	// Right
	Asura_Render::Render2DLine(fRight, fTop, fRight, fBottom, Asura_OverlayBrightness|0x00000000);
	
    Asura_Gadget::Render( xOrigin );
}

void Asura_Tab_Gadget::Update( const Asura_Vector_2& xOrigin )
{
	if( m_bIsMouseOver && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
        m_bClickedThisFrame = true;
	}
}

void Asura_Tab_Gadget::Select(bool bSelected)
{
	if(m_xDialogList.Size())
	{
		Asura_ListElt<Asura_Dialog*>* pxDialogElt = m_xDialogList.Front();
		while(pxDialogElt)
		{
			pxDialogElt->Data()->SetHideFlag(ASURA_DIALOG_HIDEFLAG_TAB, !bSelected);
			pxDialogElt = pxDialogElt->Next();
		}
	}
	m_bSelected = bSelected;
}

void Asura_Tab_Gadget::SwapRow(u_int uCurrentRow, u_int uBottomRow)
{
	if(m_uRow == uCurrentRow)
	{
		// Switch to bottom row
		SetRow(uBottomRow);
		m_bBottomRow = true;
		return;
	}
	if(m_uRow == uBottomRow)
	{
		// Switch to current row
		SetRow(uCurrentRow);
		m_bBottomRow = false;
		return;
	}
}

void Asura_Tab_Gadget::SetRow(u_int uRow)
{
	const int iDiff = uRow - m_uRow;
	m_xOffset.y += iDiff * ASURA_TAB_GADGET_HEIGHT;
	m_uRow = uRow;
}

void Asura_Tab_Gadget::GetMaxCoords(Asura_Vector_2& xMax)
{
	xMax.x = 0;
	xMax.y = 0;

	if(m_xDialogList.Size())
	{
		Asura_ListElt<Asura_Dialog*>* pxDialogElt = m_xDialogList.Front();
		while(pxDialogElt)
		{
			const Asura_Dialog* pxDialog = pxDialogElt->Data();
			if(!pxDialog->GetParentSubwindow())
			{
				const Asura_Vector_2 xDialogMax = pxDialog->GetMaxCoords();
				if(xMax.x < xDialogMax.x) xMax.x = xDialogMax.x;
				if(xMax.y < xDialogMax.y) xMax.y = xDialogMax.y;
			}

			pxDialogElt = pxDialogElt->Next();
		}
	}
}

void Asura_Tab_Gadget::SetPosition(const Asura_Vector_2& xOffset,u_int uRow)
{
	m_xOffset = xOffset;
	m_uRow = uRow;
}

void Asura_Tab_Gadget::DeleteChildDialogs(Asura_Overlay* pxOverlay)
{
	//delete and remove all the dialogs from the overlay
	while(m_xDialogList.Size())
	{
		//Delete the first dialog in the list. This should also result in it being removed as a result of
		//RemoveReferencesToDialog() being called
		Asura_ListElt<Asura_Dialog*>* pxDialogElt = m_xDialogList.Front();
		pxOverlay->DeleteSingleDialog(pxDialogElt->Data());
	}
}

void Asura_Tab_Gadget::RemoveReferencesToDialog(const Asura_Dialog* pxDialog)
{
	m_xDialogList.Remove(const_cast<Asura_Dialog*>(pxDialog));
}

void Asura_Tab_Gadget::SetName(const char* szName, bool bChangeIDName)
{
	if(bChangeIDName)
	{
		//ID and display name are the same , so just clear out the ID name
		m_xIDName.MakeNull();
	}
	else if(m_xIDName.IsBlank())
	{
		//If the ID name was blank , then copy the old display name to preserve it.
		m_xIDName = *m_pxName;
	}
	*m_pxName = szName;
//	m_fWidth = static_cast<float>(strlen(szName));
	m_fWidth = 2.0f+Asura_Text_Output::WidthOfString(szName, ASURA_TEXT_FORMAT_SMALLFONT);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Asura_Scrollbar_Gadget **
**-----------------------*/


Asura_Scrollbar_Gadget::Asura_Scrollbar_Gadget(const Asura_Vector_2& xOffset,float* pfValue,float fMin,float fMax,bool bHorizontal,float fLength,const char* const szPrintFormat,bool* pbChanged, void(*pfnChanged)(void) /* = 0 */)
:m_pfValue(pfValue)
,m_fMin(fMin)
,m_fMax(fMax)
,m_xPrintFormat(szPrintFormat)
,m_pbChanged(pbChanged)
,m_bDraggingSlider(false)
,m_fMouseDragStart(0.0f)
,m_fSliderDragStart(0.0f)
,m_fRepeatTimer(0.0f)
,m_bHorizontal(bHorizontal)
,m_fSliderLength(12.0f)
,m_fScrollbarThickness(14.0f)
,m_pfnChangedFunction(pfnChanged)
{
	m_xOffset = xOffset;
	
	if(m_bHorizontal)
	{
		m_fWidth = fLength;
		m_fHeight = m_fScrollbarThickness/ASURA_TEXT_OUTPUT_FONT_HEIGHT;
		m_fScrollbarLength = fLength*ASURA_TEXT_OUTPUT_FONT_WIDTH;
	}
	else
	{
		m_fHeight = fLength;
		m_fWidth = m_fScrollbarThickness/ASURA_TEXT_OUTPUT_FONT_WIDTH;
		m_fScrollbarLength = fLength*ASURA_TEXT_OUTPUT_FONT_HEIGHT;
	}
}


void Asura_Scrollbar_Gadget::Render(const Asura_Vector_2& xOrigin)
{
	//get top left in screen coords
	float fX = (float)Asura_Maths::FloatToInt_Near((xOrigin.x + m_xOffset.x)*ASURA_TEXT_OUTPUT_FONT_WIDTH);
	float fY = (float)Asura_Maths::FloatToInt_Near((xOrigin.y + m_xOffset.y)*ASURA_TEXT_OUTPUT_FONT_HEIGHT);

	//and the width and height in pixels
	float fWidth = m_fWidth * ASURA_TEXT_OUTPUT_FONT_WIDTH;
	float fHeight = m_fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	//draw outline for rectangle
	u_int uColour = Asura_OverlayBrightness|0xffffff;
	Asura_Render::Render2DLine(fX,fY,fX+fWidth,fY,uColour);
	Asura_Render::Render2DLine(fX+fWidth,fY,fX+fWidth,fY+fHeight,uColour);
	Asura_Render::Render2DLine(fX+fWidth,fY+fHeight,fX,fY+fHeight,uColour);
	Asura_Render::Render2DLine(fX,fY+fHeight,fX,fY,uColour);

	//draw a coloured polygon for the main bar

	u_int uMinColour = 0x7f7f7f|Asura_OverlayBrightness;
	u_int uMaxColour = 0xbfbfbf|Asura_OverlayBrightness;

	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

	Asura_Render::Overlay.SetRectXY(fX+1.0f,fY+1.0f,fWidth-1.0f,fHeight-1.0f);
	if(m_bHorizontal)
		Asura_Render::Overlay.SetGouraudColour(uMinColour,uMaxColour,uMinColour,uMaxColour);
	else
		Asura_Render::Overlay.SetGouraudColour(uMinColour,uMinColour,uMaxColour,uMaxColour);

	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();

	//Draw a rectangle for the slider
	float fSliderPos = GetSliderCentre()-m_fSliderLength/2.0f;
	fSliderPos += m_bHorizontal ? fX : fY;
	
	uColour = Asura_OverlayBrightness|0x4f4f4f;
	if(m_bHorizontal)
		Asura_Render::Overlay.SetRectXY(fSliderPos,fY+1.0f,m_fSliderLength,fHeight-1.0f);
	else
		Asura_Render::Overlay.SetRectXY(fX+1.0f,fSliderPos,fWidth-1.0f,m_fSliderLength);
	Asura_Render::Overlay.SetGouraudColour(uColour,uColour,uColour,uColour);
	Asura_Render::Overlay.Render();


    Asura_Gadget::Render( xOrigin );
}

void Asura_Scrollbar_Gadget::Update(const Asura_Vector_2& xOrigin)
{
	if( m_bIsMouseOver && m_bIsClickable && Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) )
	{
 		Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
        m_bClickedThisFrame = true;
		m_fRepeatTimer = 0.0f;
	}
		
 	if (Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
 	{
 		if(m_iFocusFlags)
 		{
			//Get mouse coordinates relative to the slidebar
			float fMouseX,fMouseY;
			Asura_Cursor::GetScreenCoords(fMouseX,fMouseY);

			float fMousePos;
			if(m_bHorizontal)
			{
				fMousePos = fMouseX - (xOrigin.x + m_xOffset.x);
				fMousePos *= ASURA_TEXT_OUTPUT_FONT_WIDTH;
			}
			else
			{
				fMousePos = fMouseY - (xOrigin.y + m_xOffset.y);
				fMousePos *= ASURA_TEXT_OUTPUT_FONT_HEIGHT;
			}

			const float fSliderPos = GetSliderCentre();

			
			if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				//First click... is it on the slider or just on the bar
				if(fMousePos >= fSliderPos - m_fSliderLength/2.0f && fMousePos <= fSliderPos + m_fSliderLength/2.0f)
				{
					m_bDraggingSlider = true;
					//store the initial positions
					m_fMouseDragStart = fMousePos;
					m_fSliderDragStart = fSliderPos;
				}
				else
				{
					m_bDraggingSlider = false;
				}
			}

			float fNewPos = fSliderPos;
			if(m_bDraggingSlider)
			{
				//update the new position based on the distance moved since the drag started
				fNewPos = m_fSliderDragStart+(fMousePos-m_fMouseDragStart);
			}
			else
			{
				m_fRepeatTimer -= Asura_Timers::GetFrameTime();
				if(m_fRepeatTimer <= 0.0f)
				{
					float fMove = (GetScrollbarMax()-GetScrollbarMin())/20.0f;

					//move the slider towards the mouse cursor
					if(fNewPos > fMousePos)
					{
						fNewPos -= fMove;
						if(fNewPos < fMousePos) fNewPos = fMousePos;
					}
					else
					{
						fNewPos += fMove;
						if(fNewPos > fMousePos) fNewPos = fMousePos;
					}
					
					//set the timer till it next moves
					m_fRepeatTimer = Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) ? 0.2f : 0.05f;
				}
			}

			if(fNewPos != fSliderPos)
			{
				SetSliderPosition(fNewPos);
			}

			//show the current value
			if ( m_xPrintFormat.GetLength() > 0 )
			{
				char szText[100];
				Asura_PrintfToBfr(szText,m_xPrintFormat.GetString(),sizeof(szText),true,*m_pfValue);
				Asura_Overlay_System::SetHelpText(szText);
			}
		}
	}
	else if(m_iFocusFlags)
	{
		Asura_Overlay_System::ReleaseFocus(OverlayFocusType_Mouse);
	}

}

float Asura_Scrollbar_Gadget::GetScrollbarMin() const
{
	if(m_fScrollbarLength>m_fSliderLength+2.0f)
	{
		return m_fSliderLength/2.0f+1.0f;
	}
	else
	{
		return m_fScrollbarLength/2.0f;
	}
}

float Asura_Scrollbar_Gadget::GetScrollbarMax() const
{
	if(m_fScrollbarLength>m_fSliderLength+2.0f)
	{
		return m_fScrollbarLength-m_fSliderLength/2.0f;
	}
	else
	{
		return m_fScrollbarLength/2.0f;
	}
}


float Asura_Scrollbar_Gadget::GetSliderCentre() const
{
	float fScrollbarMin = GetScrollbarMin();
	float fScrollbarMax = GetScrollbarMax();

	//If min and max value are the same ,just pick the centre slider position
	if(m_fMin == m_fMax) return (fScrollbarMax+fScrollbarMin)/2.0f;
	
	//determine proportion along
	float fLerp = (*m_pfValue-m_fMin) / (m_fMax - m_fMin);
	fLerp = Asura_Maths::ClampToRange(fLerp,0.0f,1.0f);

	return fScrollbarMin + (fScrollbarMax-fScrollbarMin)*fLerp;
}

void Asura_Scrollbar_Gadget::SetSliderPosition(float fPos)
{
	//get proportion along bar
	float fScrollbarMin = GetScrollbarMin();
	float fScrollbarMax = GetScrollbarMax();
	if(fScrollbarMin == fScrollbarMax)
	{
		//slidebar not long enough , so forget it
		return;
	}

		
	float fLerp = (fPos-fScrollbarMin) / (fScrollbarMax - fScrollbarMin);
	fLerp = Asura_Maths::ClampToRange(fLerp,0.0f,1.0f);

	float fNewVal = m_fMin*(1-fLerp) + m_fMax*fLerp;

	//read/write value using the print format to force the correct number of decimal places
	if ( m_xPrintFormat.GetLength() > 0 )
	{
		char szTemp[40];
		Asura_PrintfToBfr(szTemp,m_xPrintFormat.GetString(), sizeof(szTemp), true, Asura_Maths::FtoD(fNewVal));
		sscanf(szTemp,"%f",&fNewVal);
	}

	//make sure the value is in range
	if(m_fMin<m_fMax)
	{
		fNewVal = Asura_Maths::ClampToRange(fNewVal,m_fMin,m_fMax);
	}
	else
	{
		fNewVal = Asura_Maths::ClampToRange(fNewVal,m_fMax,m_fMin);
	}

	//if the new value is different , update and set the changed flag
	if(fNewVal != *m_pfValue)
	{
		*m_pfValue = fNewVal;
		if(m_pbChanged) *m_pbChanged = true;
		if( m_pfnChangedFunction ) m_pfnChangedFunction();
	}
}

void Asura_Scrollbar_Gadget::SetSliderLength(float fLength)
{
	m_fSliderLength = fLength * (m_bHorizontal ? ASURA_TEXT_OUTPUT_FONT_WIDTH : ASURA_TEXT_OUTPUT_FONT_HEIGHT);
}

float Asura_Scrollbar_Gadget::GetSliderLength() const
{
	return m_fSliderLength / (m_bHorizontal ? ASURA_TEXT_OUTPUT_FONT_WIDTH : ASURA_TEXT_OUTPUT_FONT_HEIGHT);
}

void Asura_Scrollbar_Gadget::SetBarLength(float fBarLength)
{
	if(m_bHorizontal)
	{
		m_fWidth = fBarLength; 
		m_fScrollbarLength = fBarLength*ASURA_TEXT_OUTPUT_FONT_WIDTH;
	}
	else
	{
		m_fHeight = fBarLength;
		m_fScrollbarLength = fBarLength*ASURA_TEXT_OUTPUT_FONT_HEIGHT;
	}
}

float Asura_Scrollbar_Gadget::GetBarThickness() const
{
	return m_fScrollbarThickness / (m_bHorizontal ? ASURA_TEXT_OUTPUT_FONT_HEIGHT : ASURA_TEXT_OUTPUT_FONT_WIDTH);
}

float Asura_Scrollbar_Gadget::GetDefaultThickness(bool bHorizontal)
{
	return 14.0f/(bHorizontal ? ASURA_TEXT_OUTPUT_FONT_HEIGHT : ASURA_TEXT_OUTPUT_FONT_WIDTH);
}


#endif // #ifdef ASURA_TOOLS
