//
// Asura_Gadget.h
//
// Created by Kevin Lea - 12/06/2001 11:11:56
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_GADGET_H__
#define __ASURA_GADGET_H__

#include "Core/Asura_Callback.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_LinkList.h"
#include "Core/Asura_String.h"
#include "Interface/Asura_Key_Combination.h"

#define ASURA_GADGET_CLICKED_COLOUR   (0xff0000)
#define ASURA_GADGET_HIGHLIGHT_COLOUR (0x00ff00)
#define ASURA_GADGET_DEFAULT_COLOUR   (0xffffff)
#define ASURA_GADGET_DISABLED_COLOUR  (0x555555)
#define ASURA_GADGET_GREYED_COLOUR    (0xaaaaaa)

//colour to use for outline of the 'shared' gadgets
#define ASURA_GADGET_SHARED_COLOUR    (0x8f8fff)
//colour to use for text in shared gadgets if the value varies between objects
#define ASURA_GADGET_MIXED_COLOUR     (0x3f3fff)

//Darkened version of the above.
//Used as a special case for ticks in tickboxes to indicate the value is off for the primary selection
#define ASURA_GADGET_DARKMIXED_COLOUR     (0x00005f)

// Max time between clicks for a valid doubleclick
#define ASURA_GADGET_DOUBLECLICK_TIME (0.3f)

template <class T> class Asura_Dialog_SharedValue;
class Asura_Dialog;

class Asura_Gadget
{
	public:

        Asura_Gadget( void );
		virtual ~Asura_Gadget();

		void RenderList( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin ) = 0;

		Asura_Vector_2 m_xOffset;
		
		float m_fWidth,m_fHeight;
		
		int m_iFocusFlags;
		
		Asura_Gadget* m_pxNextGadget;

		virtual void CheckCursorPosition( const Asura_Vector_2& xOrigin );
		
		void Handle( const Asura_Vector_2& xOrigin );

		virtual u_int GetColour( bool bClicked ) const;
		
		virtual void FocusGained(int focusType){m_iFocusFlags |= focusType;}
		virtual void FocusLost(int focusType){m_iFocusFlags &=~ focusType;}

        inline bool IsEnabled( void ) const { return m_bIsEnabled && !m_bIsHidden; }
        virtual void SetEnabledFlag( bool bEnabled, bool bAffectList = true );
		virtual bool GetEnabledFlag() {return m_bIsEnabled;}

        virtual void SetClickable( bool bClickable ) { m_bIsClickable = bClickable; }

        virtual void SetHighlight( bool bHighlight ) { m_bHighlight = bHighlight; }
		
		bool IsClickedThisFrame() const { return m_bClickedThisFrame; }
		
		void SetHidden(bool bHidden) { m_bIsHidden = bHidden; }

		bool IsMouseOver() const {return m_bIsMouseOver;}

		bool IsPointInGadget(const Asura_Vector_2& xRelativePos) const;

		static char GetAsciiKeyPressed();

		//Indicates that this control can have focus switched to it using Tab
		virtual bool CanTakeTabFocus() const {return false;}
		//For taking input focus when control is tabbed to
		virtual void TakeTabFocus(){}

		//re-positions the gadget relative to its current position
		void MoveGadgetRelative(const Asura_Vector_2& xDiff);

		void SetParentDialog(Asura_Dialog* pxDialog){m_pxDialog = pxDialog;}

		void SetRadioButtonGroup(int iGroupID,bool bAtLeastOnePressed)
		{
			m_iRadioButtonGroup = iGroupID;
			m_bRadioButton_AtLeastOneMustBePressed = bAtLeastOnePressed;
		}

		//Used internally to deal with radio button style tickboxes
		void RadioButtonPressed(const class Asura_Gadget* pxPressedGadget);

		//Tells gadget to default to the colour for shared gadgets if it would otherwise be the default colour
		void SetUseSharedGadgetColour(bool bShared){m_bUseSharedGadgetColour=bShared;}

		//Tells gadget to use the colour that is used for shared gadgets with mixed data , if it would otherwise be the default colour
		void SetUseMixedGadgetColour(bool bMixed){m_bUseMixedGadgetColour=bMixed;}

		float GetWidthForHitTest() const;

		//Causes the width to expand to the right side of the overlay for the purpose of testing
		//whether the cursor is being clicked in the gadget
		void SetExpandWidthToFillOverlay(bool bSet) {m_bExpandWidthToFillOverlay = bSet;}
		bool GetExpandWidthToFillOverlay() const {return m_bExpandWidthToFillOverlay;}

	protected :
		virtual void Render( const Asura_Vector_2& xOrigin );

		//Get the overlay that this belongs to
		class Asura_Overlay* GetOverlay() const;

		//for tickbox-like controls
		virtual bool IsChecked() const {return false;}
		virtual void SetChecked(bool bSet) {}

		//Does a check for a debounced key , but allows for it to be repeated if held long enough
		bool IsRepeatableKeyPressed(int iKey);



        bool m_bClickedThisFrame;
        bool m_bHighlight;
        bool m_bIsClickable;
        bool m_bIsEnabled;
        bool m_bIsMouseOver;
		bool m_bIsHidden;
		bool m_bUseSharedGadgetColour;
		bool m_bUseMixedGadgetColour;
		bool m_bExpandWidthToFillOverlay;
		
		Asura_Dialog* m_pxDialog;

		int m_iRadioButtonGroup;
		bool m_bRadioButton_AtLeastOneMustBePressed;//I'm sure there must be a better name for this...

		//For simulating a key repeat rate while a key is pressed down
		float m_fKeyRepeatTimer;
		int m_iRepeatKey;

        
	private:
};




class Asura_Incrementor_Gadget : public Asura_Gadget
{
	public:
		
		Asura_Incrementor_Gadget(const Asura_Vector_2& offset,bool sign,bool* flag = 0);


		void Render( const Asura_Vector_2& xOrigin );
		void Update( const Asura_Vector_2& xOrigin );

		// Set the delta value of the gadget if one is supported 
		// by a derived class otherwise throw an assert.
		virtual void SetDelta(const float fDelta) { Asura_Assert(false,"Float Delta gadget function not supported"); }
		virtual void SetDelta(const int iDelta) { Asura_Assert(false,"Int Delta gadget function not supported"); }

	private:
		virtual void Increment(int multiplier)=0;
		bool m_bSign;
		bool* m_pbFlag;
		float m_fTimer;

};

class Asura_IncrementorFloat_Gadget : public Asura_Incrementor_Gadget
{
	public:
		
		Asura_IncrementorFloat_Gadget(const Asura_Vector_2& offset, float* value, float min, float max, float delta, bool* flag = 0);
		virtual void SetDelta( const float fDelta );
		void SetMinMax( const float fMin,  const float fMax );

	private:
		virtual void Increment(int iMultiplier);

		float* m_pfValue;
		float m_fMin;
		float m_fMax;
		float m_fDelta;
		bool  m_bRightJustify;

};

class Asura_IncrementorInt_Gadget : public Asura_Incrementor_Gadget
{
	public:
		
		Asura_IncrementorInt_Gadget(const Asura_Vector_2& offset, int* value, int min, int max, int delta, bool* flag = 0);

		virtual void SetDelta(const int iDelta) { m_iDelta = iDelta; }

	private:
		virtual void Increment(int iMultiplier);

		int* m_piValue;
		int m_iMin;
		int m_iMax;
		int m_iDelta;

};

class Asura_IncrementorUInt_Gadget : public Asura_Incrementor_Gadget
{
	public:
		
		Asura_IncrementorUInt_Gadget(const Asura_Vector_2& xOffset, u_int* puValue, u_int uMin, u_int uMax, int iDelta, bool* pbFlag = 0);

		virtual void SetDelta(const int iDelta) { m_iDelta = iDelta; }

	private:
		virtual void Increment(int iMultiplier);

		u_int *m_puValue;
		u_int m_uMin;
		u_int m_uMax;
		int m_iDelta;

};

class Asura_IncrementorUChar_Gadget : public Asura_Incrementor_Gadget
{
public:

	Asura_IncrementorUChar_Gadget(const Asura_Vector_2& xOffset, u_char* pucValue, u_char ucMin, u_char ucMax, int iDelta, bool* pbFlag = 0);

	virtual void SetDelta(const int iDelta) { m_iDelta = iDelta; }

private:
	virtual void Increment(int iMultiplier);

	u_char *m_pucValue;
	u_char m_ucMin;
	u_char m_ucMax;
	int m_iDelta;

};

class Asura_IncrementorSharedFloat_Gadget : public Asura_Incrementor_Gadget
{
	public:
		Asura_IncrementorSharedFloat_Gadget(const Asura_Vector_2& offset,Asura_Dialog_SharedValue<float>* pxSharedValue, float min, float max, float delta);

		virtual void SetDelta( const float fDelta ) { m_fDelta = fDelta; }

	private:
		virtual void Increment(int iMultiplier);

		Asura_Dialog_SharedValue<float>* m_pxSharedValue;
		float m_fMin;
		float m_fMax;
		float m_fDelta;

};

class Asura_IncrementorSharedInt_Gadget : public Asura_Incrementor_Gadget
{
	public:
		Asura_IncrementorSharedInt_Gadget(const Asura_Vector_2& offset, Asura_Dialog_SharedValue<int>* pxSharedValue, int min, int max, int delta);

		virtual void SetDelta( const int iDelta ) { m_iDelta = iDelta; }

	private:
		virtual void Increment(int iMultiplier);

		Asura_Dialog_SharedValue<int>* m_pxSharedValue;
		int m_iMin;
		int m_iMax;
		int m_iDelta;

};

class Asura_IncrementorSharedUInt_Gadget : public Asura_Incrementor_Gadget
{
	public:
		
		Asura_IncrementorSharedUInt_Gadget(const Asura_Vector_2& xOffset,Asura_Dialog_SharedValue<u_int>* pxSharedValue, u_int uMin, u_int uMax, int iDelta);

		virtual void SetDelta(const int iDelta) { m_iDelta = iDelta; }

	private:
		virtual void Increment(int iMultiplier);

		Asura_Dialog_SharedValue<u_int>* m_pxSharedValue;
		u_int m_uMin;
		u_int m_uMax;
		int m_iDelta;

};


class Asura_BaseTickbox_Gadget : public Asura_Gadget
{
	public:

        static const float WIDTH;

		Asura_BaseTickbox_Gadget(const Asura_Vector_2& offset,bool* modifiedFlag);

		void Render( const Asura_Vector_2& xOrigin );
		void Update( const Asura_Vector_2& xOrigin );

		//set callback function to be called whenever tickbox is clicked
		void SetClickedFunction(const Asura_Callback_Base<void>& xCallback) {m_xOnClickedCallback.SetCallback(&xCallback);}


	private :
		void ToggleCheck(){SetChecked(!IsChecked());}

		bool* m_pbModifiedFlag;

		Asura_Callback_Wrapper<void> m_xOnClickedCallback;

};

class Asura_BoolTickbox_Gadget : public Asura_BaseTickbox_Gadget
{
	public:
	
		Asura_BoolTickbox_Gadget(const Asura_Vector_2& offset, bool* value,bool* flag);
		
        void SetValue( bool* pbNewValue ) { m_pbValue = pbNewValue; }

	private:
        bool IsChecked( void ) const { return m_pbValue ? *m_pbValue : false; }
        void SetChecked( bool bSet) { if( m_pbValue ) *m_pbValue = bSet; }
		
		bool* m_pbValue;
};

class Asura_FlagTickbox_Gadget : public Asura_BaseTickbox_Gadget
{
	public :
		Asura_FlagTickbox_Gadget(const Asura_Vector_2& xOffset, int* piFlagsField, int iFlagBit, bool* pbModifiedFlag);
		Asura_FlagTickbox_Gadget(const Asura_Vector_2& xOffset, char* pcFlagsField, char cFlagBit, bool* pbModifiedFlag);

	private:
        bool IsChecked( void ) const;
		void SetChecked(bool bSet);
		
		// There should only ever be one of these non-null
		int* m_piFlagsField;
		char* m_pcFlagsField;
		int m_iFlagBit;
};

class Asura_ListTickbox_Gadget : public Asura_BaseTickbox_Gadget
{
	public :
		Asura_ListTickbox_Gadget(const Asura_Vector_2& xOffset,class Asura_List_Dialog* pxListDialog,int iIndex,bool* pbModifiedFlag);

	private:
        bool IsChecked( void ) const;
		void SetChecked( bool bSet);
		
		class Asura_List_Dialog* m_pxListDialog;
		int m_iIndex;
};

class Asura_SharedTickbox_Gadget : public Asura_BaseTickbox_Gadget
{
	public :
		Asura_SharedTickbox_Gadget(const Asura_Vector_2& xOffset,Asura_Dialog_SharedValue<bool>* pxSharedValue);

		void Render( const Asura_Vector_2& xOrigin );

	private:
        bool IsChecked( void ) const;
		void SetChecked(bool bSet);
		
		Asura_Dialog_SharedValue<bool>* m_pxSharedValue;
};

class Asura_FunctionTickbox_Gadget : public Asura_BaseTickbox_Gadget
{
	public :
		Asura_FunctionTickbox_Gadget(const Asura_Vector_2& xOffset, bool(*pfnIsChecked)(), void(*pfnSetChecked)(bool bSet));
		Asura_FunctionTickbox_Gadget(const Asura_Vector_2& xOffset, void* pData, bool(*pfnIsChecked)(void* pData), void(*pfnSetChecked)(bool bSet,void* pData));
		Asura_FunctionTickbox_Gadget(const Asura_Vector_2& xOffset, int iData, bool(*pfnIsChecked)(int iData), void(*pfnSetChecked)(bool bSet,int iData));


	private :
        bool IsChecked( void ) const;
		void SetChecked( bool bSet);

		void* m_pData;
		int m_iData;
		bool (*m_pfnVoidIsChecked)();
		void (*m_pfnVoidSetChecked)(bool bSet);
		bool (*m_pfnPtrIsChecked)(void* pData);
		void (*m_pfnPtrSetChecked)(bool bSet,void* pData);
		bool (*m_pfnIntIsChecked)(int iData);
		void (*m_pfnIntSetChecked)(bool bSet,int iData);

};


class Asura_BaseText_Gadget : public Asura_Gadget
{
	public:
		Asura_BaseText_Gadget(const Asura_Vector_2& xOffset, int iMaxLength, bool* pbFlag,bool bDefaultFocus);
		virtual ~Asura_BaseText_Gadget();
		
		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

		virtual void FocusGained(int focusType);
		virtual void FocusLost(int focusType);

		virtual void TakeTabFocus();
		virtual bool CanTakeTabFocus() const {return !m_bReadOnly;}


		//sets maximum length without resizing the box
		void SetMaxStringLength(int iMaxLength);

		//Forces the text box to the specified size
		void SetBoxLength(int iNumChars);

		void SetReadOnly(bool bReadOnly){m_bReadOnly = bReadOnly;}

		void SetDefaultFocus(bool bSet) {m_bDefaultFocus = true;}
		bool ShouldTakeDefaultFocus() {return m_bDefaultFocus;}

		void SetTextChangeCallback (const Asura_Callback_Base<void>& xCallback) { m_xTextChangeCallback.SetCallback (&xCallback); }

		//Causes the string to be updated immediately as the user types , instead of waiting for return to be pressed
		void SetUpdateImmediatelyWhileTyping(bool bSet){m_bUpdateImmediatelyWhileTyping = bSet;}

	protected:
		void HandleKeyboardInput();
		virtual void FillTextBuffer(){}
		virtual bool ReadFromTextBuffer(){return false;}
		virtual bool CharacterAllowed(char cChar){return true;}

		//returns true if gadget has value shared between multiple objects
		virtual bool IsSharedGadget(){return false;}
		
		//returns true if gadget has value shared between multiple objects
		//and is not the same for all of them
		virtual bool IsMixedValue(){return false;}

		//helper functions while editing text
		void DeleteSelection();
		bool IsTextSelected() const;
		void CopyText() const;
		void PasteText();


		void RenderString(const Asura_Vector_2& xPos,u_int uColour,bool bCursor);

		Asura_Callback_Wrapper <void> m_xTextChangeCallback;

		char* m_pszTextBuffer;
		int m_iMaxLength;
		int m_iBoxLength;
		int m_iBoxLengthOverride;
		int m_iCursorPos;
		int m_iSelectionStart;//marks start of selected text , which goes up to the cursor position
		bool m_bDefaultFocus;//determines whether this control should autoamtically get keyboard focus
		bool m_bCentreJustify;
		bool m_bDiscardChanges;
		bool m_bReadOnly;
		bool* m_pbFlag;
		bool m_bUpdateImmediatelyWhileTyping;

};


//gadget displaying a number that can be altered by clicking on it
class Asura_Float_Gadget : public Asura_BaseText_Gadget
{
	public:
		Asura_Float_Gadget(const Asura_Vector_2& offset, float* value,float min,float max,const char* const szPrintFormat,bool* flag, bool bRightJustify = false);
		Asura_Float_Gadget(const Asura_Vector_2& offset, Asura_Dialog_SharedValue<float>* pxSharedValue,float min,float max,const char* const szPrintFormat,bool* flag, bool bRightJustify = false);
		void SetMinMax( const float fMin, const float fMax );
		void SetPrintFormat( const char* const szPrintFormat );
		
	protected:
		void UpdateBoxLength( const float fMin, const float fMax, const bool bShrink );

		virtual void FillTextBuffer();
		virtual bool ReadFromTextBuffer();
		virtual bool CharacterAllowed(char cChar);
		virtual bool IsSharedGadget();
		virtual bool IsMixedValue();


		float GetValue();
		void SetValue(float fValue);

		
		Asura_String m_xPrintFormat;
		float* m_pfValue;
		Asura_Dialog_SharedValue<float>* m_pxSharedValue;
		float m_fMin;
		float m_fMax;
		bool m_bRightJustify;
};




//and for integers...
class Asura_Integer_Gadget : public Asura_BaseText_Gadget
{
	public :
		Asura_Integer_Gadget( const Asura_Vector_2& offset, int* value, int min, int max, const char* const szPrintFormat, bool* flag = 0);
		Asura_Integer_Gadget( const Asura_Vector_2& offset, Asura_Dialog_SharedValue<int>* pxSharedValue, int min, int max, const char* const szPrintFormat, bool* flag = 0);

	protected:
		virtual void FillTextBuffer();
		virtual bool ReadFromTextBuffer();
		virtual bool CharacterAllowed(char cChar);
		virtual bool IsSharedGadget();
		virtual bool IsMixedValue();

		int GetValue();
		void SetValue(int iValue);



		Asura_String m_xPrintFormat;
		int* m_piValue;
		Asura_Dialog_SharedValue<int>* m_pxSharedValue;
		int m_iMin;
		int m_iMax;

};

//and for u_ints...
class Asura_UInt_Gadget : public Asura_BaseText_Gadget
{
	public :
		Asura_UInt_Gadget( const Asura_Vector_2& offset, u_int* puValue, u_int uMin, u_int uMax, const char* const szPrintFormat, bool* pbFlag = 0);
		Asura_UInt_Gadget( const Asura_Vector_2& offset, Asura_Dialog_SharedValue<u_int>* pxSharedValue, u_int uMin, u_int uMax, const char* const szPrintFormat, bool* pbFlag = 0);

	protected:
		virtual void FillTextBuffer();
		virtual bool ReadFromTextBuffer();
		virtual bool CharacterAllowed(char cChar);
		virtual bool IsSharedGadget();
		virtual bool IsMixedValue();

		u_int GetValue();
		void SetValue(u_int uValue);

		Asura_String m_xPrintFormat;
		u_int* m_puValue;
		Asura_Dialog_SharedValue<u_int>* m_pxSharedValue;
		u_int m_uMin;
		u_int m_uMax;

};

class Asura_UChar_Gadget : public Asura_BaseText_Gadget
{
public :
	Asura_UChar_Gadget( const Asura_Vector_2& xOffset, u_char* pucValue, u_char ucMin, u_char ucMax, const char* const szPrintFormat, bool* pbFlag = 0);
	Asura_UChar_Gadget( const Asura_Vector_2& xOffset, Asura_Dialog_SharedValue<u_char>* pxSharedValue, u_char ucMin, u_char ucMax, const char* const szPrintFormat, bool* pbFlag = 0);

protected:
	virtual void FillTextBuffer();
	virtual bool ReadFromTextBuffer();
	virtual bool CharacterAllowed(char cChar);
	virtual bool IsSharedGadget();
	virtual bool IsMixedValue();

	u_char GetValue();
	void SetValue(u_char ucValue);

	Asura_String m_xPrintFormat;
	u_char *m_pucValue;
	Asura_Dialog_SharedValue<u_char>* m_pxSharedValue;
	u_char m_ucMin;
	u_char m_ucMax;
};

class Asura_NumberSlider_Gadget : public Asura_Float_Gadget
{
	public :
		Asura_NumberSlider_Gadget(const Asura_Vector_2& offset,float* value, float* min, float* max,float slideWidth, const char* const szPrintFormat, bool* flag = 0);
		virtual void Update( const Asura_Vector_2& xOrigin );
		virtual void Render( const Asura_Vector_2& xOrigin );
		void UpdateSliderPosition();
		void SetPrintFormat( const char* const szPrintFormat);

		virtual bool CanTakeTabFocus() const {return false;}


	private :
		float* m_pfMinPtr;
		float* m_pfMaxPtr;
		float m_fSlideWidth;
		float m_fSliderPosition;
};



class Asura_String_Gadget : public Asura_BaseText_Gadget
{
	public:
		Asura_String_Gadget(const Asura_Vector_2& offset, class Asura_String *value, int maxLength, bool* flag = 0, bool defaultFocus = false);

		void UpdateValue () { ReadFromTextBuffer (); }

	private:
		virtual void FillTextBuffer();
		virtual bool ReadFromTextBuffer();

		virtual bool CanTakeTabFocus() const {return true;}
		
		class Asura_String* m_pxValue;
};


class Asura_Minimise_Gadget : public Asura_Gadget
{
	public:
		Asura_Minimise_Gadget(const Asura_Vector_2& offset, u_int* puMinimisedState,bool bMimimiseToBottom);
		
		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

	private:
		
		u_int* m_puMinimisedState;
		bool m_bMinimiseToBottom;
};

class Asura_Close_Gadget : public Asura_Gadget
{
	public:
		Asura_Close_Gadget(const Asura_Vector_2& offset, bool* shouldClose);
		
		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

	private:
		
		bool* m_pbShouldClose;
};

class Asura_Resize_Gadget : public Asura_Gadget
{
	public :
		Asura_Resize_Gadget(const Asura_Vector_2& xOffset);

		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

	private :
		Asura_Dialog* m_pxDialogToResize;
};


class Asura_Function_Gadget : public Asura_Gadget
{
	public:
		Asura_Function_Gadget(const Asura_Vector_2& offset, const Asura_Callback_Base<void>& xCallback, bool* flag = 0);
		
		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

		//If the function from this gadget matches the most recently used one , then
		//it gets called again.
		void RepeatLastFunction();

		void SetCallback(const Asura_Callback_Base<void>& xCallback)
		{
			m_xOnClickedCallback.SetCallback(&xCallback);
		}


	protected:

		Asura_Callback_Wrapper<void> m_xOnClickedCallback;
		
		bool* m_pbFlag;
		
		//for shortcut key for repeating last function press
		static const void* s_pLastFunction;
};


class Asura_FunctionWithKeybinding_Gadget : public Asura_Function_Gadget
{
	public:
		Asura_FunctionWithKeybinding_Gadget( const Asura_Vector_2& xOffset, const Asura_Callback_Base<void>& xCallback, const Asura_Key_Combination& xBinding, bool* pbFlag = 0);
		
		void Update( const Asura_Vector_2& xOrigin );
	
	private :
		Asura_Key_Combination m_xBinding;
};


class Asura_TextSelect_Gadget : public Asura_Gadget
{
	public:
		Asura_TextSelect_Gadget( 
			const Asura_Vector_2& xOffset,
			int* const piSelection,
			const int iSelectionValue,
			int* const piListOffset,
			const char* const* ppszList,
			const int iNumberOfLines,
			const int iNumberInList,
			void( *pfnFunction )( void ),
			const bool bOnlySelectOnClick = false );

		void SetListHelpStrings( const char* const* ppszHelpText );
		void SetTooltipCallback( void (*pfnDisplayTooltipCallback)(int iIndex,const char* szText));
		void ReplaceStrings( int iNumberInList, const char* const* ppszList )	{ m_iNumberInList = iNumberInList; m_ppszList = ppszList; }

		//set the index of the current selection
		void SetSelection(int iIndex, bool bAffectMultipleSelection = true);

		void SetMultipleSelectionHandler(class Asura_List_Dialog_SelectionHandler* pxMultipleSelectionHandler) {m_pxMultipleSelectionHandler = pxMultipleSelectionHandler;}

		void SetSingleClickFunction(const Asura_Callback_Base<void>& xSingleClickCallback) {m_xSingleClickCallback.SetCallback(&xSingleClickCallback);}
		void SetDoubleClickFunction(const Asura_Callback_Base<void>& xDoubleClickCallback) {m_xDoubleClickCallback.SetCallback(&xDoubleClickCallback);}

		//Used to remap from the selected index to an index into the list of strings.
		//Used by the sorted version of the drop down list so the unsorted string list will appear sorted
		void SetIndexRemapArray(const int* piIndexRemap) {m_piIndexRemap = piIndexRemap;}



		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

		void HandleKeyboard();

		const char* const* GetList() const {return m_ppszList;}
		const char* const* GetHelpText() const {return m_ppszHelpText;}
		int* GetSelectionPointer() {return m_piSelection;}
		int* GetListOffsetPointer() {return m_piListOffset;}
		const Asura_Callback_Wrapper<void>& GetSingleClickCallback() const {return m_xSingleClickCallback;}
		const Asura_Callback_Wrapper<void>& GetDoubleClickCallback() const {return m_xDoubleClickCallback;}


	private:

		bool IsMultipleSelect() const {return m_pxMultipleSelectionHandler != 0;}
		bool IsSelected() const;

		
		int m_iSelectionValue;
		int* const m_piSelection;
		int* const m_piListOffset;
		int m_iNumberOfLines;
		int m_iNumberInList;
		const char* const* m_ppszList;
		const char* const* m_ppszHelpText;
		void(*m_pfnDisplayTooltipCallback)(int iIndex,const char* szText);
		void(*m_pfnFunction)(void);

		Asura_Callback_Wrapper<void> m_xSingleClickCallback;
		Asura_Callback_Wrapper<void> m_xDoubleClickCallback;

		class Asura_List_Dialog_SelectionHandler* m_pxMultipleSelectionHandler;

		const bool m_bOnlySelectOnClick;

		const int* m_piIndexRemap;
};

class Asura_UpDown_Gadget : public Asura_Gadget
{
	public:
		
		Asura_UpDown_Gadget(const Asura_Vector_2& offset, int* value, int min, int max, int delta);

		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

		void SetMin(int iMin){m_iMin = iMin;}
		void SetMax(int iMax){m_iMax = iMax;}

		int* GetValuePointer() {return m_piValue;}



	private:

		int* m_piValue;
		int m_iMin;
		int m_iMax;
		int m_iDelta;

};



class Asura_Button_Gadget : public Asura_Gadget
{
	public:
		Asura_Button_Gadget(const Asura_Vector_2& offset,int textureIndex,int textureSize,Asura_Vector_2 buttonPixelPos,Asura_Vector_2 buttonPixelSize,void(*clickedFunction)(void));
		Asura_Button_Gadget(const Asura_Vector_2& offset,int textureIndex,int textureSize,Asura_Vector_2 buttonPixelPos,Asura_Vector_2 buttonPixelSize,void(*clickedFunction)(void*),void* pContext);

		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );


	private :
		int m_iTextureIndex;
		int m_iTextureSize;
		Asura_Vector_2 m_xButtonPixelPos;
		Asura_Vector_2 m_xButtonPixelSize;
		void(*m_pfnClickedFunction)(void);
		void(*m_pfnClickedContextFunction)(void*);
		void *m_pContext;

	
};

class Asura_Icon_Gadget : public Asura_Gadget
{
	public :
		Asura_Icon_Gadget(const Asura_Vector_2& xOffset,int iTextureIndex,int iIconIndex);

		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

		static const float s_fUVWidth;
		static const float s_fUVHeight;
		static const float s_fWidth;
		static const float s_fHeight;

		void SetClickedFunction(const Asura_Callback_Base<void>& xClicked);


		void SetClickedFunction(void(*pfnFunction)(int iUserData), int iUserData) {SetClickedFunction(Asura_Callback_UserData<void,int>(pfnFunction, iUserData));}



		void SetClickedFlag(bool* pbClickedFlag)
		{ m_pbClickedFlag = pbClickedFlag; }

		void SetCheckboxBool(bool* pbValue)
		{m_pbValue = pbValue;}

		void SetCheckboxFlag(u_int* puFlagsField,u_int uFlagBit)
		{m_puFlagsField = puFlagsField;m_uFlagBit = uFlagBit;}

		void SetCheckboxFunction(const Asura_Callback_Base<bool>& xIsChecked, const Asura_Callback_Base_1Arg<void,bool>& xSetChecked);


	protected :
		virtual bool IsChecked() const;
		virtual void SetChecked(bool bSet);
		

	private :
		bool ShowAsPressed();

		int m_iTextureIndex;
		int m_iIconX;
		int m_iIconY;

		Asura_Callback_Wrapper<void> m_xClickedCallback;

		bool* m_pbClickedFlag;

		bool* m_pbValue;

		u_int* m_puFlagsField;
		u_int m_uFlagBit;

		Asura_Callback_Wrapper<bool> m_xIsCheckedCallback;
		Asura_Callback_Wrapper_1Arg<void,bool> m_xSetCheckedCallback;

};





// Like an Asura_String_Gadget except you can't modify its contents.
class Asura_StringDisplay_Gadget : public Asura_Gadget
{

    public:

        // T Y P E S /////////////////////////////////////////////////////////

        typedef bool (*BOOL_CALLBACK)( int );
        typedef void (*VOID_CALLBACK)( int );

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        Asura_StringDisplay_Gadget( const Asura_Vector_2& xOffset,
                                    u_int uCharLimit = 0,
                                    BOOL_CALLBACK pfnIsOn = NULL,
                                    VOID_CALLBACK pfnOnClick = NULL,
                                    int iCallbackArg = 0,
									int iFormatFlags = 0);

        virtual ~Asura_StringDisplay_Gadget( void );

        // M E T H O D S /////////////////////////////////////////////////////

        virtual void Render( const Asura_Vector_2& xOrigin );
        virtual void Update( const Asura_Vector_2& xOrigin );

        virtual void SetCallbackArg( int iCallbackArg ) { m_iCallbackArg = iCallbackArg; }
        virtual void SetLabel( const char* szLabel );

		// Changes maximum allowed characters and adjusts width to fit (assuming standard font, as with the constructor)
		void SetCharLimit(u_int uCharLimit) { m_uCharLimit = uCharLimit; m_fWidth = static_cast<float>(uCharLimit); }

    protected:
        
        // S T A T E /////////////////////////////////////////////////////////

        char*         m_szLabel;
        u_int         m_uCharLimit;
        BOOL_CALLBACK m_pfnIsOn;
        VOID_CALLBACK m_pfnOnClick;
        int           m_iCallbackArg;
		int			  m_iFormatFlags;
            
};

class Asura_VerticalScroll_Gadget : public Asura_Gadget
{
	public :
		Asura_VerticalScroll_Gadget(const Asura_Vector_2 xOffset,float fHeight,int iNumItemsShown,int* piValue,int iMin,int iMax);

		virtual void Render( const Asura_Vector_2& xOrigin );
		virtual void Update( const Asura_Vector_2& xOrigin );

		void SetMinAndMax(int iMin,int iMax);
	
	private :
		void UpdateSliderPosition();

		int* m_piValue;
		int  m_iMin;
		int  m_iMax;

		int m_iNumItemsShown;

		float m_fSliderHeight;
		float m_fSliderPosition;
	
};

#define ASURA_TAB_GADGET_HEIGHT (1.2f)
class Asura_Tab_Gadget : public Asura_Gadget
{
public:
	// Constructor & destructor
	Asura_Tab_Gadget(const char* szName);
	~Asura_Tab_Gadget();

	// Interface functions
	virtual void Render( const Asura_Vector_2& xOrigin );
	virtual void Update( const Asura_Vector_2& xOrigin );
	void AddDialog(class Asura_Dialog* pxDialog);
	void Select(bool bSelected);

	//gets the max coords of all the child dialogs
	void GetMaxCoords(Asura_Vector_2& xMax);
	// Gets the row of tabs this gadget belongs to
	u_int GetRow() { return m_uRow; }
	// If the tab is in either of these rows, switches it to the other one
	void SwapRow(u_int uCurrentRow, u_int uBottomRow);
	void SetBottomRow(bool bBottom) { m_bBottomRow = bBottom; }

	//get the display name
	const char* GetName() const {return m_pxName->GetString();}
	//get the name to use for identifying the tab when selecting by name
	const char* GetIDName() const {return m_xIDName.IsBlank() ? m_pxName->GetString() : m_xIDName.GetString();}
	void SetName(const char* szName, bool bChangeIDName);

	//used by the dialog to position the gadget
	void SetPosition(const Asura_Vector_2& xOffset,u_int uRow);

	void DeleteChildDialogs(class Asura_Overlay* pxOverlay);
	void RemoveReferencesToDialog(const Asura_Dialog* pxDialog);
private:
	void SetRow(u_int uRow);
	// Data
	Asura_LinkList<Asura_Dialog*>	m_xDialogList;
	class Asura_String*				m_pxName;
	Asura_String					m_xIDName;//If not blank then this is used for identifying the tab when selecting by name
	bool							m_bSelected;
	bool							m_bBottomRow;
	u_int							m_uRow;
};

////////////////////////////////////////////////////////////////////////////////////////////////////////
class Asura_Scrollbar_Gadget : public Asura_Gadget
{
public :
	Asura_Scrollbar_Gadget(const Asura_Vector_2& xOffset,float* pfValue,float fMin,float fMax,bool bHorizontal,float fLength,const char* const szPrintFormat,bool* pbChanged, void(*pfnChanged)(void) = 0);
	
	//Set min and max values
	void SetMin(float fMin){m_fMin = fMin;}
	void SetMax(float fMax){m_fMax = fMax;}

	float GetMin() const {return m_fMin;}
	float GetMax() const {return m_fMax;}

	//Sets the size of the slider in print coordinates
	void SetSliderLength(float fSliderLength);
	float GetSliderLength() const;
	//Sets the bar length in print coordinates
	void SetBarLength(float fBarLength);

	//Gets the bar thickness in print coordinates
	float GetBarThickness() const;

	//Gets the default thickness in print coords
	static float GetDefaultThickness(bool bHorizontal);

private :
	virtual void Render( const Asura_Vector_2& xOrigin );
	virtual void Update( const Asura_Vector_2& xOrigin );

	//Gets the centre of the slider in pixels relative to the top/left of the gadget
	float GetSliderCentre() const;
	//Sets the centre position of the slider
	void SetSliderPosition(float fPos);

	//Gets the min/max allowed position of the slider on the bar in pixels.
	float GetScrollbarMin() const;
	float GetScrollbarMax() const;

	float* m_pfValue;
	float m_fMin;
	float m_fMax;
	Asura_String m_xPrintFormat;
	bool* m_pbChanged;

	bool m_bDraggingSlider;
	float m_fMouseDragStart;
	float m_fSliderDragStart;
	float m_fRepeatTimer;

	bool m_bHorizontal;
	float m_fSliderLength;//Length of slider in pixels
	float m_fScrollbarLength;//Length of bar in pixels
	float m_fScrollbarThickness;//Thickness of bar in pixels

	void(*m_pfnChangedFunction)(void);
};

#endif//__ASURA_GADGET_H__
