//
// Asura_Menubar_System.cpp
//
// Created by Kevin Lea - 25/06/2001 14:54:52
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_TOOLS

#include "Interface/Asura_Menubar_System.h"
#include "Interface/Asura_Menubar.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Overlay_System.h"
#include "Core/Asura_UserInput.h"
#include "Rendering/Asura_Viewport.h"
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Render.h"

#define ASURA_MAX_NO_OF_MENUBAR_SECTIONS 10

void Asura_Menubar_System::Initialise(void)
{
	UsingMenus = false;

	MenubarStorage = new Asura_Menubar*[ASURA_MAX_NO_OF_MENUBAR_SECTIONS];
	MenubarHandleList = new Asura_Handle_List(ASURA_MAX_NO_OF_MENUBAR_SECTIONS);

	for (int i=0; i<ASURA_MAX_NO_OF_MENUBAR_SECTIONS; i++)
	{
		MenubarStorage[i]=0;
	}


}

void Asura_Menubar_System::Shutdown(void)
{
	for (int i=0; i<ASURA_MAX_NO_OF_MENUBAR_SECTIONS; i++)
	{
		Asura_Menubar*& menubar = MenubarStorage[i];

		if (menubar)
		{
			delete menubar;
			menubar = 0;
		}
	}

	delete MenubarHandleList;
	delete [] MenubarStorage;

}

void Asura_Menubar_System::Update(void)
{
	Asura_Cursor::GetScreenCoords(CursorX,CursorY);

	if (!UsingMenus)
	{
		ActiveTitle = 0;
		if (CursorY<1.0f)
		{ 
			//Don't activate the menus if a mouse button is currently held down. RAR 08/05/2008
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1) &&
			   !Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON2))
			{
				UsingMenus = true;
			}
		}
	}

	if (UsingMenus)
	{
		UserDidSomething = false;
		MenusDidSomething = false;

		if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1)
			||Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
		{
			UserDidSomething = true;
		}


		Asura_Vector_2 topLeft(0,0);

		for (int i=0; i<ASURA_MAX_NO_OF_MENUBAR_SECTIONS; i++)
		{
			Asura_Menubar* menubar = MenubarStorage[i];

			if (menubar)
			{
				menubar->Update(topLeft);

				Asura_Menubar_Title* title = menubar->FirstTitle;
				while(title)
				{
					topLeft.x += title->WidthOfTitle;
					title = title->NextTitle;
				}
			}
		}

		if (UserDidSomething && !MenusDidSomething && CursorY>1.0f)
		{
			UsingMenus = false;
		}

		MenusActive = true;
	}
	else
	{
		MenusActive = false;
	}
}

void Asura_Menubar_System::Render(void)
{
	if (UsingMenus)
	{
		Asura_Vector_2 topLeft(0,0);

		RenderBox(topLeft,(float)Asura_Viewport::GetWidth()/10.0f,2.0f);

		topLeft.x += 0.5f;
		topLeft.y += 0.5f;



		for (int i=0; i<ASURA_MAX_NO_OF_MENUBAR_SECTIONS; i++)
		{
			Asura_Menubar* menubar = MenubarStorage[i];

			if (menubar)
			{
				menubar->Render(topLeft);
				Asura_Menubar_Title* title = menubar->FirstTitle;
				while(title)
				{
					topLeft.x += title->WidthOfTitle;
					title = title->NextTitle;
				}
			}
		}
	}
}


void Asura_Menubar_System::ForceRecalculateWidths(void)
{
	for (int i=0; i<ASURA_MAX_NO_OF_MENUBAR_SECTIONS; i++)
	{
		Asura_Menubar* menubar = MenubarStorage[i];

		if (menubar)
		{
			Asura_Menubar_Title* pxTitle = menubar->FirstTitle;
			while(pxTitle)
			{
				pxTitle->AreWidthsCalculated = false;
				pxTitle = pxTitle->NextTitle;
			}			
		}
	}
}


bool Asura_Menubar_System::IsUsingMenus(void)
{
	return MenusActive;
}


Asura_Handle Asura_Menubar_System::CreateMenuSection(void)
{
	Asura_Handle handle;

	MenubarHandleList->GetNewHandle(handle);

	Asura_Assert(handle!=ASURA_INVALID_HANDLE,"Ran out of menu space");

	MenubarStorage[handle] = new Asura_Menubar;


	return handle;
}

void Asura_Menubar_System::DestroyMenuSection(Asura_Handle handle)
{
	if ( handle == ASURA_INVALID_HANDLE ) { return; }

	Asura_Menubar*& menubar = MenubarStorage[handle];

	if (menubar)
	{
		delete menubar;
		menubar = 0;
	}

	MenubarHandleList->ReleaseHandle(handle);
}

Asura_Menubar_Title* Asura_Menubar_System::AddTitle(Asura_Handle handle, const Asura_Char* name)
{
	if ( handle == ASURA_INVALID_HANDLE ) { return NULL; }
	Asura_Assert(handle>=0 && handle<ASURA_MAX_NO_OF_MENUBAR_SECTIONS,"");

	//get the menubar
	Asura_Menubar* menubar = MenubarStorage[handle];
	Asura_Assert(menubar,"");

	//allocate the title
	Asura_Menubar_Title* title = new Asura_Menubar_Title;
	title->Name = name;

	//find the end of the title list
	Asura_Menubar_Title** titleInsertionPoint = &menubar->FirstTitle;
	while(*titleInsertionPoint)
	{
		titleInsertionPoint = &((*titleInsertionPoint)->NextTitle);
	}

	*titleInsertionPoint = title;

	return title;
}

Asura_Menubar_Title* Asura_Menubar_System::GetTitle(Asura_Handle handle, const Asura_Char* name)
{
	if ( handle == ASURA_INVALID_HANDLE ) { return NULL; }
	Asura_Assert(handle>=0 && handle<ASURA_MAX_NO_OF_MENUBAR_SECTIONS,"");

	//get the menubar
	Asura_Menubar* menubar = MenubarStorage[handle];
	Asura_Assert(menubar,"");

	//search for the title with this name
	Asura_Menubar_Title* title = menubar->FirstTitle;
	while(title)
	{
		if(!title->Name.Compare(name)) return title;
		title = title->NextTitle;

	}
	return 0;
}



void Asura_Menubar_System::RenderBox(Asura_Vector_2 topleft, float width, float height)
{
	if( Asura_Overlay_System::UseSkinnedUI() )
	{
		Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);

		unsigned int colour = Asura_Overlay_System::s_uOverlayColour;//0000;
		colour |= 0xff000000;

		Asura_Render::Overlay.Colour = 0xffffffff;
		extern int MenuSkinTextureIndex;
		Asura_Render::Overlay.TextureIndex = MenuSkinTextureIndex;
		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

		float x=Asura_Maths::FloatToInt(topleft.x*10.0f)-0.5f;
		float y=Asura_Maths::FloatToInt(topleft.y*14.0f)-0.5f;

		float fTexWidth = 256.0f;
		float fTexHeight = 256.0f;

		// want to get section (8,0)->(48,25)
		float fUv00 = (8.00f / fTexWidth);
		float fUv01 = (2.00f / fTexHeight);
		float fUv10 = (48.00f / fTexWidth);
		float fUv11 = (16.00f / fTexHeight);

		Asura_Render::Overlay.SetRectUV(fUv00, fUv01, fUv10, fUv11);
		Asura_Render::Overlay.SetRectXY(x,y,width*10.0f,height*14.0f);

		Asura_Render::Overlay.Render();

		Asura_Render::FlushDisplay();
	}
	else
	{
		Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);

		unsigned int colour = Asura_Overlay_System::s_uOverlayColour;//0000;
		colour |= 0xff000000;

		Asura_Render::Overlay.Colour = colour;
		extern int MenuBackdropTextureIndex;
		Asura_Render::Overlay.TextureIndex = MenuBackdropTextureIndex;
		Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;

		float x=Asura_Maths::FloatToInt(topleft.x*10.0f)-0.5f;
		float y=Asura_Maths::FloatToInt(topleft.y*14.0f)-0.5f;

		Asura_Render::Overlay.SetRectUV(0.0f,0.0f,width*10.0f/128.0f,height*14.0f/128.0f);
		Asura_Render::Overlay.SetRectXY(x,y,width*10.0f,height*14.0f);

		Asura_Render::Overlay.Render();

		Asura_Render::FlushDisplay();
	}
}



bool					Asura_Menubar_System::UsingMenus;
bool					Asura_Menubar_System::UserDidSomething;
bool					Asura_Menubar_System::MenusDidSomething;
Asura_Menubar_Title*	Asura_Menubar_System::ActiveTitle;
Asura_Menubar**			Asura_Menubar_System::MenubarStorage;
Asura_Handle_List*		Asura_Menubar_System::MenubarHandleList;
bool					Asura_Menubar_System::MenusActive;
float					Asura_Menubar_System::CursorX;
float					Asura_Menubar_System::CursorY;

#endif // #ifdef ASURA_TOOLS
