//
// Asura_Tree_Dialog.cpp
//
// Created by Richard Rice - 11/09/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_TOOLS

#include "Interface/Asura_Tree_Dialog.h"

#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_UserInput.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Font_System.h"

#include "PC/FileSystem/Tools_FileSystem.h"

extern unsigned int Asura_OverlayBrightness;

const float ASURA_TREE_DIALOG_INDENT = 14.0f;//indent per level of hierarchy
const float ASURA_TREE_DIALOG_TEXTOFFSET = 30.0f;//horizontal offset to where the text starts from the indented position
const float ICON_WIDTH = 14.0f;
const float XGAP_BETWEEN_ICONS_AND_TEXT = 5.0f;

const u_int SELECTED_ITEM_MARQUEE_COLOUR		(0xffaaaaaa);

//////////////////////////////////////////////////////////////////////////////////////////////
/*-------------------**
** Asura_Tree_Dialog **
**-------------------*/

Asura_Tree_Dialog::Asura_Tree_Dialog(const Asura_Vector_2& xOffset,Asura_TreeDialog_Tree* pxTree,float fWidth,float fHeight,bool* pbSelectionChanged)
:Asura_Dialog("",xOffset)
,m_pxTree(pxTree)
,m_pbSelectionChanged(pbSelectionChanged)
,m_bUsePerItemHelp(false)
,m_fPreResizeWidth(0.0f)
,m_fPreResizeHeight(0.0f)
,m_pxRenameStringGadget(0)
,m_pxRenameItem(0)
,m_bRenameAccepted(false)
{
	Asura_Assert(pxTree,"");

	m_fHScrollPos = 0.0f;
	m_fVScrollPos = 0.0f;

	CreateGadgets(fWidth, fHeight);

	MakeSelectionVisible();
	UpdateScrollbars();
}

void Asura_Tree_Dialog::CreateGadgets(float fWidth, float fHeight)
{
	//Determine space for the main tree view
	m_fViewWidth = fWidth - Asura_Scrollbar_Gadget::GetDefaultThickness(false);
	m_fViewHeight = fHeight - Asura_Scrollbar_Gadget::GetDefaultThickness(true);
	//Make sure there is going to be at least some area
	if(m_fViewWidth < 2.0f || m_fViewHeight < 2.0f)
	{
		Asura_Assert(0,"Tree dialog too small");
		m_fViewWidth = Asura_Maths::Max(m_fViewWidth,2.0f);
		m_fViewHeight = Asura_Maths::Max(m_fViewHeight,2.0f);

		fWidth = m_fViewWidth + Asura_Scrollbar_Gadget::GetDefaultThickness(false);
		fHeight = m_fViewHeight + Asura_Scrollbar_Gadget::GetDefaultThickness(true);
	}
	
	m_pxTreeViewGadget = new Asura_TreeView_Gadget(m_xOffset,this,m_pxTree,m_fViewWidth,m_fViewHeight,m_pbSelectionChanged);
	AddGadget(m_pxTreeViewGadget);
	
	m_pxTreeViewGadget->SetDoubleClickCallback(m_xDoubleClickCallback.GetCallback());
	m_pxTreeViewGadget->SetDragDropCallback(m_xDragDropCallback.GetCallback());
	m_pxTreeViewGadget->SetIconCallback(m_xIconCallback.GetCallback());
	m_pxTreeViewGadget->SetIconClickCallback(m_xIconClickCallback.GetCallback());
	#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
	m_pxTreeViewGadget->SetAdditionalIconCallback(m_xAdditionalIconCallback.GetCallback());
	m_pxTreeViewGadget->SetAdditionalIconClickCallback(m_xAdditionalIconClickCallback.GetCallback());
	#endif
	m_pxTreeViewGadget->SetSelectionChangedCallback( m_xSelectionChangedCallback.GetCallback() );

	Asura_Vector_2 xPos;

	//Create the scrollbars
	xPos = m_xOffset;
	xPos.y += m_fViewHeight;
	m_pxHorizontalBar = new Asura_Scrollbar_Gadget(xPos,&m_fHScrollPos,0.0f,0.0f,true,m_fViewWidth,0,0);
	AddGadget(m_pxHorizontalBar);

	xPos = m_xOffset;
	xPos.x += m_fViewWidth;
	m_pxVerticalBar = new Asura_Scrollbar_Gadget(xPos,&m_fVScrollPos,0.0f,0.0f,false,m_fViewHeight,0,0);
	AddGadget(m_pxVerticalBar);

	m_xMaxCoord.x = m_xOffset.x + fWidth;
	m_xMaxCoord.y = m_xOffset.y + fHeight;
}

void Asura_Tree_Dialog::Render(const Asura_Vector_2& xOrigin)
{
	Asura_Dialog::Render(xOrigin);

	if(!IsHidden())
	{
		//draw a box around the control
		float fX = (float)Asura_Maths::FloatToInt_Near((xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH);
		float fY = (float)Asura_Maths::FloatToInt_Near((xOrigin.y + m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT);
		float fWidth = (m_xMaxCoord.x - m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH;
		float fHeight = (m_xMaxCoord.y - m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT;


		u_int uColour = Asura_OverlayBrightness|ASURA_GADGET_GREYED_COLOUR;
		Asura_Render::Render2DLine(fX,fY-1.0f,fX+fWidth+1.0f,fY-1.0f,uColour);
		Asura_Render::Render2DLine(fX+fWidth,fY,fX+fWidth,fY+fHeight,uColour);
		Asura_Render::Render2DLine(fX,fY+fHeight,fX+fWidth+1.0f,fY+fHeight,uColour);
		Asura_Render::Render2DLine(fX,fY,fX,fY+fHeight,uColour);
	}		
}

Asura_TreeDialog_Item* Asura_Tree_Dialog::GetVisibleItem(int iIndex) const
{
	return m_pxTree->GetVisibleItem(iIndex);
}

void Asura_Tree_Dialog::UpdateScrollbars()
{
	//Update the vertical scrollbar
	float fHeightRequired = (float) m_pxTree->GetNumVisibleItems();
	float fMaxScroll = Asura_Maths::Max(fHeightRequired - m_fViewHeight,0.0f);
	m_fVScrollPos = Asura_Maths::Min(m_fVScrollPos,fMaxScroll);

	float fSliderLength = m_fViewHeight;
	if(fMaxScroll > 0)
	{
		fSliderLength = (m_fViewHeight/fHeightRequired)*m_fViewHeight;
	}

	m_pxVerticalBar->SetSliderLength(fSliderLength);
	m_pxVerticalBar->SetMax(fMaxScroll);

	//Update the horizontal scrollbar
	//Going to go for a fixed maximum width for now.. may change this later if I feel energetic enough.
	float fWidthRequired = 200.0f;
	fMaxScroll = Asura_Maths::Max(fWidthRequired - m_fViewWidth,0.0f);
	m_fHScrollPos = Asura_Maths::Min(m_fHScrollPos,fMaxScroll);

	fSliderLength = m_fViewWidth;
	if(fMaxScroll > 0)
	{
		fSliderLength = (m_fViewWidth/fWidthRequired)*m_fViewWidth;
	}

	m_pxHorizontalBar->SetSliderLength(fSliderLength);
	m_pxHorizontalBar->SetMax(fMaxScroll);


}

void Asura_Tree_Dialog::PerFrameUpdate()
{
	Asura_Dialog::PerFrameUpdate();

	UpdateRename();
	UpdateScrollbars();
}

void Asura_Tree_Dialog::VerticalScroll(float fScrollAmount)
{
	m_fVScrollPos += fScrollAmount;
	m_fVScrollPos = Asura_Maths::ClampToRange(m_fVScrollPos,0.0f,m_pxVerticalBar->GetMax());
}

void Asura_Tree_Dialog::MakeSelectionVisible()
{
	Asura_TreeDialog_Item* pxSelected = m_pxTree->GetSelectedItem();
	if(!pxSelected) return;

	//make sure all parents are expanded
	pxSelected->MakeVisible();
	UpdateScrollbars();

	int iIndex = pxSelected->GetVisibleIndex();
	if(iIndex < 0) return;

	//get the range in which it would be visible
	float fMinPos = (float)iIndex-(m_fViewHeight-1.0f);
	float fMaxPos = (float)iIndex;

	//clamp the scroll position to the specified range
	m_fVScrollPos = Asura_Maths::ClampToRange(m_fVScrollPos,fMinPos,fMaxPos);
}

void Asura_Tree_Dialog::SetDoubleClickedFunction(void(*pfnOnClicked)(void))
{
	Asura_Callback_Void<void> xCallback(pfnOnClicked);
    SetDoubleClickedFunction(xCallback);
}

void Asura_Tree_Dialog::SetDoubleClickedFunction(void(*pfnOnClicked)(int), int iUserData)
{
	Asura_Callback_UserData<void, int> xCallback(pfnOnClicked, iUserData);
    SetDoubleClickedFunction(xCallback);
}

void Asura_Tree_Dialog::SetDoubleClickedFunction(const Asura_Callback_Base<void>& xCallback)
{
	m_pxTreeViewGadget->SetDoubleClickCallback(&xCallback);
	m_xDoubleClickCallback.SetCallback(&xCallback);
}

void Asura_Tree_Dialog::SetDragDropCallback(const Asura_Callback_Base_1Arg<void,Asura_TreeDialog_Item*>& xCallback)
{
	m_pxTreeViewGadget->SetDragDropCallback(&xCallback);
	m_xDragDropCallback.SetCallback(&xCallback);
}

void Asura_Tree_Dialog::SetIconCallback(const Asura_Callback_Base_1Arg<int,Asura_TreeDialog_Item*>& xCallback)
{
	m_pxTreeViewGadget->SetIconCallback(&xCallback);
	m_xIconCallback.SetCallback(&xCallback);
}

#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
void Asura_Tree_Dialog::SetAdditionalIconCallback(const Asura_Callback_Base_2Arg<int,Asura_TreeDialog_Item*,u_int>& xCallback)
{
	m_pxTreeViewGadget->SetAdditionalIconCallback(&xCallback);
	m_xAdditionalIconCallback.SetCallback(&xCallback);
}
#endif

void Asura_Tree_Dialog::SetIconClickCallback(const Asura_Callback_Base_2Arg<void,Asura_TreeDialog_Item*,bool>& xCallback)
{
	m_pxTreeViewGadget->SetIconClickCallback(&xCallback);
	m_xIconClickCallback.SetCallback(&xCallback);
}

#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
void Asura_Tree_Dialog::SetAdditionalIconClickCallback(const Asura_Callback_Base_2Arg<void,Asura_TreeDialog_Item*,bool>& xCallback)
{
	m_pxTreeViewGadget->SetAdditionalIconClickCallback(&xCallback);
	m_xAdditionalIconClickCallback.SetCallback(&xCallback);
}
#endif

void Asura_Tree_Dialog::SetSelectionChangedCallback( const Asura_Callback_Base_1Arg< void, Asura_TreeDialog_Item* >& xCallback )
{
	m_pxTreeViewGadget->SetSelectionChangedCallback( &xCallback );
	m_xSelectionChangedCallback.SetCallback( &xCallback );
}

void Asura_Tree_Dialog::SetRenameCallback(const Asura_Callback_Base_2Arg<void,Asura_TreeDialog_Item*,const char*>& xCallback)
{
	m_xRenameCallback.SetCallback(&xCallback);
}

void Asura_Tree_Dialog::StartResize(u_int uResizeFlags)
{
	m_fPreResizeWidth = m_xMaxCoord.x - m_xOffset.x;
	m_fPreResizeHeight = m_xMaxCoord.y - m_xOffset.y;
}

void Asura_Tree_Dialog::UpdateResize(u_int uResizeFlags, const Asura_Vector_2& xDelta)
{
	if(!m_bAllowResize) return;

	float fWidth,fHeight;

	if(uResizeFlags & ASURA_OVERLAY_RESIZE_INITIAL)
	{
		//initialise the size based on the last size chosen by the user
		if(!m_pxUserSetSize) return;
		if(*m_pxUserSetSize == Asura_Maths::ZeroVector2) return;

		fWidth = Asura_Maths::Max(4.0f, m_pxUserSetSize->x);
		fHeight = Asura_Maths::Max(4.0f, m_pxUserSetSize->y);
	}
	else
	{
		fWidth = Asura_Maths::Max(4.0f, m_fPreResizeWidth + xDelta.x);
		fHeight = Asura_Maths::Max(4.0f, m_fPreResizeHeight + xDelta.y);

		if(m_pxUserSetSize)
		{
			m_pxUserSetSize->x = fWidth;
			m_pxUserSetSize->y = fHeight;
		}
	}

	if(fWidth != m_xMaxCoord.x - m_xOffset.x || fHeight != m_xMaxCoord.y - m_xOffset.y)
	{
		while(m_pxGadgetList) DeleteSingleGadget(m_pxGadgetList);

		//re-make the gadgets for the new size
		CreateGadgets(fWidth, fHeight);

		UpdateScrollbars();
	}
}

u_int Asura_Tree_Dialog::GetAllowedResizeFlags() const
{
	return m_bAllowResize ? ASURA_OVERLAY_RESIZE_HORIZONTAL | ASURA_OVERLAY_RESIZE_VERTICAL : 0;
}

void Asura_Tree_Dialog::StartRename(Asura_TreeDialog_Item* pxItem)
{
	Asura_Assert(pxItem, "");
	if(m_pxRenameItem) CancelRename();

	//renaming is only enabled if a callback has been provided
	if(!m_xRenameCallback.IsValid()) return;

	m_pxRenameItem = pxItem;
	m_bRenameAccepted = false;
	m_xRenameString = pxItem->GetName();

	int iVisibleIndex = pxItem->GetVisibleIndex();
	Asura_Assert(iVisibleIndex>=0,"");

	//create a string editing gadget on the tree
	Asura_Vector_2 xPos = m_xOffset;
	#ifdef TOOLS_USE_OLD_ENTITY_MANAGER
	xPos.x += (((float)pxItem->GetDepth()-1) * ASURA_TREE_DIALOG_INDENT)/ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xPos.x += ASURA_TREE_DIALOG_TEXTOFFSET/ASURA_TEXT_OUTPUT_FONT_WIDTH;
	#else
	xPos.x += pxItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_TEXT] / ASURA_TEXT_OUTPUT_FONT_WIDTH;
	#endif
	xPos.x -= GetHorizontalScrollPos();
	xPos.y += (float) iVisibleIndex;
	xPos.y -= GetVerticalScrollPos();

	if(xPos.x < m_xOffset.x) xPos.x = m_xOffset.x;
	if(xPos.y < m_xOffset.y) xPos.y = m_xOffset.y;

	int iWidthLeft = Asura_Maths::FloatToInt_Chop(m_xMaxCoord.x - xPos.x) - 1;
	int iGadgetWidth = Asura_Maths::Max(20, iWidthLeft);

	#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
	int iLeftIndex, iRightIndex;

	m_pxRenameItem->GetRenameExtents (iLeftIndex, iRightIndex);

	if ((iLeftIndex != -1) && (iRightIndex != -1))
	{
		m_xRenameString = m_xRenameString.SubString (iLeftIndex, (iRightIndex - iLeftIndex) + 1);
	}
	#endif

	m_pxRenameStringGadget = new Asura_String_Gadget(xPos, &m_xRenameString, iGadgetWidth, &m_bRenameAccepted, true);
	m_pxRenameStringGadget->SetMaxStringLength(100);
	AddGadget(m_pxRenameStringGadget);
}

void Asura_Tree_Dialog::UpdateRename()
{
	if(!m_pxRenameItem) return;

	//make sure the item being renamed is still the selected item
	if(m_pxRenameItem != m_pxTree->GetSelectedItem())
	{
		CancelRename();
		return;
	}

	if(m_bRenameAccepted)
	{
		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		// Update the rename extends to match the new string size

		int iStartIndex, iEndIndex;
		m_pxRenameItem->GetRenameExtents(iStartIndex, iEndIndex);

		if ((iStartIndex != -1) && (iEndIndex != -1))
		{
			iEndIndex = (iStartIndex + m_xRenameString.GetLength ()) - 1;

			m_pxRenameItem->SetRenameExtents(iStartIndex, iEndIndex);
		}

		#endif

		//Use the rename callback.
		//Register the callback to use with the overlay system (so it gets called after the
		//update is finished).
		//Use an intermediate callback to deal with the extra argument
		Asura_Callback_MemberFunction<void,Asura_Tree_Dialog> xCallback(this, &Asura_Tree_Dialog::CallRenameCallback);
		Asura_Overlay_System::SetCallbackFunctionToCall(xCallback);
		return;
	}

	//abort renaming if the string gadget loses focus
	//wait till the gadget has had a chance to take focus for the first time though
	if(!m_pxRenameStringGadget->ShouldTakeDefaultFocus() && Asura_Overlay_System::GetFocusGadget() != m_pxRenameStringGadget)
	{
		CancelRename();
	}
}

void Asura_Tree_Dialog::CallRenameCallback()
{
	if(m_pxRenameItem && m_bRenameAccepted && m_xRenameCallback.IsValid())
	{
		//Take a copy of everything thats needed
		Asura_TreeDialog_Item* pxItem = m_pxRenameItem;
		Asura_String xNewName = m_xRenameString;

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		int iLeftIndex, iRightIndex;
		m_pxRenameItem->GetRenameExtents (iLeftIndex, iRightIndex);
		#endif

		//cancel renaming before doing the callback , in case the overlay gets deleted during it
		CancelRename();

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		// If we've got a rename extent set, then handle this

		if ((iLeftIndex != -1) && (iRightIndex != -1))
		{
			Asura_String xFullName = Asura_String (pxItem->GetName ());
			Asura_String xBuffer = xFullName.SubString (0, iLeftIndex);

			xBuffer += xNewName;

			if (iRightIndex < (xFullName.GetLength () - 1))
			{
				xBuffer += xFullName.SubString (iRightIndex + 1, xFullName.GetLength () - (iRightIndex + 1));
			}

			xNewName = xBuffer;
		}
		#endif

		m_xRenameCallback.Execute(pxItem, xNewName.GetString());
	}
	else
	{
		CancelRename();
	}
}

void Asura_Tree_Dialog::CancelRename()
{
	if(!m_pxRenameItem) return;

	m_pxRenameItem = 0;
	m_bRenameAccepted = false;
	m_xRenameString.MakeNull();

	if(Asura_Overlay_System::GetFocusGadget() == m_pxRenameStringGadget)
	{
		Asura_Overlay_System::ReleaseFocus();
	}
	DeleteSingleGadget(m_pxRenameStringGadget);
	m_pxRenameStringGadget = false;


}


//////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Asura_TreeView_Gadget **
**-----------------------*/

Asura_TreeView_Gadget::Asura_TreeView_Gadget(const Asura_Vector_2& xOffset,Asura_Tree_Dialog* pxTreeDialog,Asura_TreeDialog_Tree* pxTree,float fWidth,float fHeight,bool* pbSelectionChanged)
:m_pxTreeDialog(pxTreeDialog)
,m_pxTree(pxTree)
,m_pbSelectionChanged(pbSelectionChanged)
,m_pxDragStartItem(0)
,m_pxDragTarget(0)
,m_bSelectAndRenameIfNotDrag(false)
,m_bDragConfirmed(false)
,m_pxRenameItem(0)
,m_fRenameTimer(0.0f)
,m_bIconDrag(false)
,m_iPreviousIconDragIndex(0)
{
	m_xOffset = xOffset;
	m_fWidth = fWidth;
	m_fHeight = fHeight;
}

void Asura_TreeView_Gadget::Update( const Asura_Vector_2& xOrigin )
{

	if( m_bIsMouseOver && m_pxTreeDialog->UsePerItemHelp())
	{
		Asura_TreeDialog_Item* pxItem;
		Asura_Tree_LineElement eElement;
		GetItemUnderCursor(xOrigin,pxItem,&eElement);
		if(pxItem && eElement == ASURA_TREE_LINEELEMENT_NAME)
		{
			m_pxTreeDialog->SetHelpText(pxItem->GetHelpText());
		}
		else if (pxItem && eElement == ASURA_TREE_LINEELEMENT_ICON)
		{
			m_pxTreeDialog->SetHelpText(pxItem->GetIconHelpText());
		}
		else
		{
			m_pxTreeDialog->SetHelpText(NULL);
		}
	}

	if( m_bIsMouseOver && m_bIsClickable)
	{
		if(Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEWHEELUP))
		{
			if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
				m_pxTreeDialog->VerticalScroll(-m_fHeight);
			else
				m_pxTreeDialog->VerticalScroll(-3.0f);
		}
		
		if(Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEWHEELDOWN))
		{
			if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
				m_pxTreeDialog->VerticalScroll(m_fHeight);
			else
				m_pxTreeDialog->VerticalScroll(3.0f);
		}
	
		m_pxDragTarget = 0;
		if(m_pxDragStartItem)
		{
			if(Asura_Overlay_System::GetFocusGadget() == this && m_pxTree->IsSelected(m_pxDragStartItem))
			{
				//See what item the mouse was released over
				Asura_TreeDialog_Item* pxItem;
				GetItemUnderCursor(xOrigin,pxItem, NULL);

				if(pxItem && pxItem != m_pxDragStartItem)
				{
					m_bDragConfirmed = true;
				}

				if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
				{
					
					if(pxItem)
					{
						if(pxItem == m_pxDragStartItem)
						{
							//turned out not to be a drag , so just select if needed
							if(m_bSelectAndRenameIfNotDrag)
							{
								m_pxTree->SetSelectedItem(pxItem);
								//May want to rename this item.
								//Start a timer to make sure a double click isn't going to happen first.
								StartRenameTimer(pxItem);
							}
						}
						else if(m_xDragDropCallback.IsValid())
						{
							m_pxDragTarget = pxItem;
							//Register the calllback to use with the overlay system (so it gets called after the
							//update is finished).
							//Use an intermediate callback to deal with the extra argument
							Asura_Callback_MemberFunction<void,Asura_TreeView_Gadget> xCallback(this, &Asura_TreeView_Gadget::CallDragDropCallback);
							Asura_Overlay_System::SetCallbackFunctionToCall(xCallback);
						}
					}
					Asura_Overlay_System::ReleaseFocus();
					m_pxDragStartItem = 0;
					m_bSelectAndRenameIfNotDrag = false;
					m_bDragConfirmed = false;
				}
			}
			else
			{
				//something else has intefered with the drag ,so abort it
				m_pxDragStartItem = 0;
				m_bSelectAndRenameIfNotDrag = false;
				m_bDragConfirmed = false;
			}
		}
		else if(m_bIconDrag)
		{
			bool bContinueDrag = false;
			if(Asura_Overlay_System::GetFocusGadget() == this && Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				//get the index of the item under the cursor
				Asura_TreeDialog_Item* pxItem;
				GetItemUnderCursor(xOrigin,pxItem, NULL);
				if(pxItem)
				{
					int iVisibleIndex = pxItem->GetVisibleIndex();
					if(iVisibleIndex >=0)
					{
						//use the callback for each item between the previous and the current index
						while(m_iPreviousIconDragIndex != iVisibleIndex)
						{
							(m_iPreviousIconDragIndex < iVisibleIndex) ? m_iPreviousIconDragIndex++ : m_iPreviousIconDragIndex--;

							Asura_TreeDialog_Item* pxIntermediateItem = m_pxTree->GetVisibleItem(m_iPreviousIconDragIndex);
							if(pxIntermediateItem)
							{
								m_xIconClickCallback.Execute(pxIntermediateItem,true);
							}
							
						}
					}
				}
				bContinueDrag = true;	
			}
			
			if(!bContinueDrag)
			{
				//cancel icon drag
				if(Asura_Overlay_System::GetFocusGadget() == this)
				{
					Asura_Overlay_System::ReleaseFocus();
				}
				m_bIconDrag = false;
			}
			
		}
		else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) || 
			Asura_UserInput::KeyPressed(ASURA_KEYID_DOUBLECLICK_MOUSEBUTTON1) ||
			Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
		{
			Asura_TreeDialog_Item* pxItem;
			Asura_Tree_LineElement eElement;
			GetItemUnderCursor(xOrigin,pxItem,&eElement);

			CancelRenameTimer();
			
			if(pxItem)
			{
				if(eElement == ASURA_TREE_LINEELEMENT_EXPAND  )
				{
					//clicked on expand symbol ,so toggle expand icon
					pxItem->Expand(!pxItem->IsExpanded());
				}
				else if(eElement == ASURA_TREE_LINEELEMENT_ICON)
				{
					//clicked on the icon , so either use callback or toggle expand
					if(m_xIconClickCallback.IsValid())
					{
						m_xIconClickCallback.Execute(pxItem,false);
						m_iPreviousIconDragIndex = pxItem->GetVisibleIndex();
						m_bIconDrag = true;
						Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
					}
					else
					{
						pxItem->Expand(!pxItem->IsExpanded());
					}
				}
				#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
				else if (eElement == ASURA_TREE_LINEELEMENT_ADDITIONAL_ICON)
				{
					//clicked on the icon , so either use callback or toggle expand
					if(m_xAdditionalIconClickCallback.IsValid())
					{
						m_xAdditionalIconClickCallback.Execute(pxItem,false);
						m_iPreviousIconDragIndex = pxItem->GetVisibleIndex();
						m_bIconDrag = true;
						Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
					}
					else
					{
						pxItem->Expand(!pxItem->IsExpanded());
					}
				}
				#endif
				else
				{
					//clicked on name
					if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2))
					{
						if(m_pxTree->IsMultipleSelectionEnabled())
						{
							//If the right clicked item is already selected , then just add it to the selection
							//again to make it the main selection.
							//If not selected then just replace the selection
							if(m_pxTree->IsSelected(pxItem))
							{
								m_pxTree->AddToSelection(pxItem);
							}
							else
							{
								m_pxTree->SetSelectedItem(pxItem);
							}
						}
						else
						{
							m_pxTree->SetSelectedItem(pxItem);
						}
						if(m_pbSelectionChanged) *m_pbSelectionChanged = true;
						if(m_xSelectionChangedCallback.IsValid())
						{
							Asura_Overlay_System::SetCallbackFunctionToCall( Asura_Callback_Wrapper< void >( m_xSelectionChangedCallback, pxItem ) );
						}
					}
					else
					{
						//Start drag action and/or select the item
						if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) && !Asura_UserInput::KeyPressed(ASURA_KEYID_DOUBLECLICK_MOUSEBUTTON1))
						{
							m_pxDragStartItem = pxItem;
							m_bSelectAndRenameIfNotDrag = false;
							m_bDragConfirmed = false;
							Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Mouse);
						}

						if(m_pxTree->IsMultipleSelectionEnabled())
						{
							Asura_TreeDialog_Item* pxPrevSelection = m_pxTree->GetSelectedItem();
							
							if(pxPrevSelection && Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) && Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
							{
								//Add everything between this and the most recent selection to the selected list
								m_pxTree->SelectMultiple(pxPrevSelection, pxItem, true);
							}
							else if(pxPrevSelection && Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
							{
								//Replace selection with everything between this and the most recent selection
								m_pxTree->SelectMultiple(pxPrevSelection, pxItem, false);
							}
							else if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
							{
								//add/remove from the current selection
								m_pxTree->ToggleSelection(pxItem);
							}
							else
							{
								//replace the current selection
								if(m_pxDragStartItem)
								{
									if(m_pxTree->IsSelected(m_pxDragStartItem))
									{
										//clicked on an already selected item. If this turns out not to be a drag/drop
										//operation then it will need to be single selected later
										m_bSelectAndRenameIfNotDrag = true;	
									}
									else
									{
										m_pxTree->SetSelectedItem(m_pxDragStartItem);
									}
								}
								else
								{
									m_pxTree->SetSelectedItem(pxItem);
								}
							}
						}
						else
						{
							m_bSelectAndRenameIfNotDrag = m_pxDragStartItem && m_pxTree->IsSelected(m_pxDragStartItem);
							m_pxTree->SetSelectedItem(pxItem);
						}
						if(m_pbSelectionChanged) *m_pbSelectionChanged = true;
						if(m_xSelectionChangedCallback.IsValid())
						{
							Asura_Overlay_System::SetCallbackFunctionToCall( Asura_Callback_Wrapper< void >( m_xSelectionChangedCallback, pxItem ) );
						}

						if (Asura_UserInput::KeyPressed(ASURA_KEYID_DOUBLECLICK_MOUSEBUTTON1))
						{
                    	    // call the double clicked function, if there is one
							if (m_xDoubleClickCallback.IsValid()) Asura_Overlay_System::SetCallbackFunctionToCall(m_xDoubleClickCallback);
							// and make sure its children are shown
							pxItem->Expand(!pxItem->IsExpanded());
						}
					}
				}
	    	    m_bClickedThisFrame = true;
			}

		}			
	}

	if(m_pxRenameItem)
	{
		UpdateRenameTimer();		
	}

 	if (m_bIsMouseOver && !Asura_Overlay_System::s_bUsedKeypressThisFrame && !m_pxTreeDialog->GetItemBeingRenamed())
	{
		bool bUsedKey = false;
		Asura_TreeDialog_Item* pxSelected = m_pxTree->GetSelectedItem();
		if(pxSelected)
		{		
			if (IsRepeatableKeyPressed(ASURA_KEYID_UP))
			{
				bUsedKey = true;
				//Select item above
				int iIndex = pxSelected->GetVisibleIndex();
				if(iIndex > 0)
				{
					Asura_TreeDialog_Item* pxSelected = m_pxTree->GetVisibleItem(iIndex-1);
					m_pxTree->SetSelectedItem(pxSelected);
					if(m_pbSelectionChanged) *m_pbSelectionChanged = true;
					if(m_xSelectionChangedCallback.IsValid())
					{
						Asura_Overlay_System::SetCallbackFunctionToCall( Asura_Callback_Wrapper< void >( m_xSelectionChangedCallback, pxSelected ) );
					}
					m_pxTreeDialog->MakeSelectionVisible();
				}
			}
			else if(IsRepeatableKeyPressed(ASURA_KEYID_DOWN))
			{
				bUsedKey = true;
				//Select item below
				int iIndex = pxSelected->GetVisibleIndex();
				Asura_TreeDialog_Item* pxItem = m_pxTree->GetVisibleItem(iIndex+1);
				if(pxItem)
				{
					m_pxTree->SetSelectedItem(pxItem);
					if(m_pbSelectionChanged) *m_pbSelectionChanged = true;
					if(m_xSelectionChangedCallback.IsValid())
					{
						Asura_Overlay_System::SetCallbackFunctionToCall( Asura_Callback_Wrapper< void >( m_xSelectionChangedCallback, pxItem ) );
					}
					m_pxTreeDialog->MakeSelectionVisible();
				}
			}
			else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_LEFT))
			{
				bUsedKey = true;

				//If the selected item has children and is expanded , then un-expand
				if(pxSelected->IsExpanded() && pxSelected->GetNumChildren())
				{
					pxSelected->Expand(false);
				}
				//Otherwise select the parent
				else
				{
					Asura_TreeDialog_Item* pxParent = pxSelected->GetParent();
					if(pxParent && !pxParent->IsRoot())
					{
						m_pxTree->SetSelectedItem(pxParent);
						if(m_pbSelectionChanged) *m_pbSelectionChanged = true;
						if(m_xSelectionChangedCallback.IsValid())
						{
							Asura_Overlay_System::SetCallbackFunctionToCall( Asura_Callback_Wrapper< void >( m_xSelectionChangedCallback, pxParent ) );
						}
						m_pxTreeDialog->MakeSelectionVisible();
					}
				}	
			}
			else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_RIGHT))
			{
				bUsedKey = true;

				//If the selected item has children and isn't expanded ,then expand it
				if(!pxSelected->IsExpanded() && pxSelected->GetNumChildren())
				{
					pxSelected->Expand(true);
				}
				//Otherwise select the first child
				else if(pxSelected->GetNumChildren())
				{
					Asura_TreeDialog_Item* pxNewSelected = pxSelected->GetChild();
					m_pxTree->SetSelectedItem(pxNewSelected);
					if(m_pbSelectionChanged) *m_pbSelectionChanged = true;
					if(m_xSelectionChangedCallback.IsValid())
					{
						Asura_Overlay_System::SetCallbackFunctionToCall( Asura_Callback_Wrapper< void >( m_xSelectionChangedCallback, pxSelected ) );
					}
					m_pxTreeDialog->MakeSelectionVisible();
				}	
			}
			else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_HOME))
			{
				//scroll to the top
				bUsedKey = false;
				m_pxTreeDialog->VerticalScroll(-m_pxTreeDialog->GetMaxVerticalScroll());
			}
			else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_END))
			{
				//scroll to the bottom
				bUsedKey = false;
				m_pxTreeDialog->VerticalScroll(m_pxTreeDialog->GetMaxVerticalScroll());
			}
			else if(IsRepeatableKeyPressed(ASURA_KEYID_PAGEUP))
			{
				//up one page
				bUsedKey = false;
				m_pxTreeDialog->VerticalScroll(-m_pxTreeDialog->GetViewHeight());
			}
			else if(IsRepeatableKeyPressed(ASURA_KEYID_PAGEDOWN))
			{
				//down one page
				bUsedKey = false;
				m_pxTreeDialog->VerticalScroll(m_pxTreeDialog->GetViewHeight());
			}

		}
		if(bUsedKey)
		{
			Asura_Overlay_System::s_bUsedKeypressThisFrame = true;
			//take keyboard focus to prevent keys from also being used for movement
			Asura_Overlay_System::TakeFocus(this,OverlayFocusType_Keyboard);
		}
   	}

	if(m_bDragConfirmed || m_bIconDrag)
	{
		//If dragging , see if the cursor is above or below the tree , and scroll as needed
		float fCursorX, fCursorY;
		Asura_Cursor::GetScreenCoords(fCursorX, fCursorY);

		float fMinY = xOrigin.y + m_xOffset.y;
		float fMaxY = fMinY + m_fHeight;

		if(fCursorY < fMinY)
		{
			float fScrollAmount = (fCursorY - fMinY) * Asura_Timers::GetFrameTime() * 5.0f;
			m_pxTreeDialog->VerticalScroll(fScrollAmount);
		}
		else if(fCursorY > fMaxY)
		{
			float fScrollAmount = (fCursorY - fMaxY) * Asura_Timers::GetFrameTime() * 5.0f;

			m_pxTreeDialog->VerticalScroll(fScrollAmount);
		}


	}

	if(Asura_Overlay_System::GetFocusGadget() == this)
	{
		//release focus if arrow keys no longer presed
		if(!Asura_UserInput::KeyPressed(ASURA_KEYID_RIGHT) &&
			!Asura_UserInput::KeyPressed(ASURA_KEYID_LEFT) &&
			!Asura_UserInput::KeyPressed(ASURA_KEYID_UP) &&
			!Asura_UserInput::KeyPressed(ASURA_KEYID_DOWN))
		{
			Asura_Overlay_System::ReleaseFocus(OverlayFocusType_Keyboard);
		}
	}
}

void Asura_TreeView_Gadget::Render( const Asura_Vector_2& xOrigin )
{
	//Get the rectangle for rendering
	Asura_Bounding_Rectangle xRect;
	xRect.MinX = (xOrigin.x + m_xOffset.x) * ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xRect.MinY = (xOrigin.y + m_xOffset.y) * ASURA_TEXT_OUTPUT_FONT_HEIGHT;
	xRect.MaxX = xRect.MinX + m_fWidth * ASURA_TEXT_OUTPUT_FONT_WIDTH;
	xRect.MaxY = xRect.MinY + m_fHeight * ASURA_TEXT_OUTPUT_FONT_HEIGHT;

	xRect.MinX = (float)Asura_Maths::FloatToInt_Near(xRect.MinX);
	xRect.MinY = (float)Asura_Maths::FloatToInt_Near(xRect.MinY);
	xRect.MaxX = (float)Asura_Maths::FloatToInt_Near(xRect.MaxX);
	xRect.MaxY = (float)Asura_Maths::FloatToInt_Near(xRect.MaxY);

	float fPixelHeight = xRect.MaxY - xRect.MinY;

	//find the item under the cursor
	Asura_TreeDialog_Item* pxItemUnderCursor;
	Asura_Tree_LineElement eElement;
	GetItemUnderCursor(xOrigin,pxItemUnderCursor,&eElement);

	//Get the index of the first (partially) visible item
	float fVScroll = m_pxTreeDialog->GetVerticalScrollPos();
	int iIndex = Asura_Maths::FloatToInt_Chop(fVScroll);

	float fYOffset = ((float)iIndex - fVScroll) * ASURA_TEXT_OUTPUT_FONT_HEIGHT;
	fYOffset = (float) Asura_Maths::FloatToInt_Near(fYOffset);


    Asura_Font_System::SetFont(Asura_Text_Output::GetSmallFontHandle());
	u_int uColour;

	//draw until we run out of space/items
	while(fYOffset < fPixelHeight)
	{

		Asura_TreeDialog_Item* pxItem = m_pxTreeDialog->GetVisibleItem(iIndex);
		if(!pxItem) break;

		//test!
		Asura_Assert(pxItem->GetVisibleIndex() == iIndex,"");

		float fXOffset = -m_pxTreeDialog->GetHorizontalScrollPos()*ASURA_TEXT_OUTPUT_FONT_WIDTH;
		fXOffset += (float) (pxItem->GetDepth()-1)* ASURA_TREE_DIALOG_INDENT;
		fXOffset = (float) Asura_Maths::FloatToInt_Near(fXOffset);

		//render the +/- icon
		if(pxItem->GetNumChildren())
		{
			if(pxItem == pxItemUnderCursor && eElement == ASURA_TREE_LINEELEMENT_EXPAND)
			{
				uColour = GetColour(m_bClickedThisFrame);
			}
			else
			{
				uColour = IsEnabled() ? ASURA_GADGET_DEFAULT_COLOUR : ASURA_GADGET_DISABLED_COLOUR;
			}
			uColour |= Asura_OverlayBrightness;
			RenderIcon(xRect, fXOffset, fYOffset, pxItem->IsExpanded() ? ASURA_TREEDIALOG_ICON_MINUS : ASURA_TREEDIALOG_ICON_PLUS, uColour);
		}

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		fXOffset += ICON_WIDTH;

		pxItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_ICONS] = fXOffset;
		#endif

		//render the item specific icon , or open/closed folder icon
		//uColour = IsEnabled() ? ASURA_GADGET_DEFAULT_COLOUR : ASURA_GADGET_DISABLED_COLOUR;
		if(pxItem == pxItemUnderCursor && eElement == ASURA_TREE_LINEELEMENT_ICON)
		{
			uColour = GetColour(m_bClickedThisFrame);
		}
		else
		{
			uColour = IsEnabled() ? ASURA_GADGET_DEFAULT_COLOUR : ASURA_GADGET_DISABLED_COLOUR;
		}
		uColour |= Asura_OverlayBrightness;
		if(m_xIconCallback.IsValid())
		{
			int iIconIndex = m_xIconCallback.Execute(pxItem);
			if(iIconIndex >= 0)
			{
				#ifdef TOOLS_USE_OLD_ENTITY_MANAGER
				RenderIcon(xRect, fXOffset+14.0f, fYOffset, iIconIndex, uColour);
				#else
				RenderIcon(xRect, fXOffset, fYOffset, iIconIndex, uColour);
				//fXOffset += ICON_WIDTH;
				#endif
			}
		}
		else if(pxItem->GetNumChildren() || pxItem->ShouldTreatAsFolder())
		{
			#ifdef TOOLS_USE_OLD_ENTITY_MANAGER
			RenderIcon(xRect, fXOffset+14.0f, fYOffset, pxItem->IsExpanded() ? ASURA_TREEDIALOG_ICON_OPENEDFOLDER : ASURA_TREEDIALOG_ICON_CLOSEDFOLDER, uColour);
			#else
			RenderIcon(xRect, fXOffset, fYOffset, pxItem->IsExpanded() ? ASURA_TREEDIALOG_ICON_OPENEDFOLDER : ASURA_TREEDIALOG_ICON_CLOSEDFOLDER, uColour);
			//fXOffset += ICON_WIDTH;
			#endif
		}

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		fXOffset += ICON_WIDTH;

		pxItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_ADDITIONAL_ICONS] = fXOffset;
		#endif

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		// Render an additional icon (if requested)

		uColour = IsEnabled() ? ASURA_GADGET_DEFAULT_COLOUR : ASURA_GADGET_DISABLED_COLOUR;
		uColour |= Asura_OverlayBrightness;

		for (u_int n = 0; n < MAX_TREEVIEW_ADDITIONAL_ICONS; n ++)
		{
			if(m_xAdditionalIconCallback.IsValid())
			{
				int iIconIndex = m_xAdditionalIconCallback.Execute(pxItem, n);

				if(iIconIndex >= 0)
				{
					RenderIcon(xRect, fXOffset, fYOffset, iIconIndex, uColour);
					fXOffset += ICON_WIDTH;
				}
			}
		}

		#endif

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		fXOffset += XGAP_BETWEEN_ICONS_AND_TEXT;
		pxItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_TEXT] = fXOffset;
		#endif

		#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
		// If it's selected, then draw a marquee around it

		if (m_pxTree->GetSelectedItem () == pxItem)
		{
			float fLeft	= xRect.GetTopLeft ().x + fXOffset;
			float fTop	= xRect.GetTopLeft ().y + fYOffset;
			float fRight	= fLeft + Asura_Font_System::GetWidthOfString (pxItem->GetName ());
			float fBottom	= fTop + Asura_Font_System::GetFontHeight ();

			fLeft = Asura_Maths::Max (fLeft, xRect.GetTopLeft ().x);
			fRight = Asura_Maths::Min (fRight, xRect.GetBottomRight ().x);

			if (fRight > fLeft)
			{
				Asura_Render::Render2DLine (fLeft, fTop, fRight, fTop, SELECTED_ITEM_MARQUEE_COLOUR);
				Asura_Render::Render2DLine (fRight, fTop, fRight, fBottom, SELECTED_ITEM_MARQUEE_COLOUR);
				Asura_Render::Render2DLine (fRight, fBottom, fLeft, fBottom, SELECTED_ITEM_MARQUEE_COLOUR);
				Asura_Render::Render2DLine (fLeft, fBottom, fLeft, fTop, SELECTED_ITEM_MARQUEE_COLOUR);
			}
		}
		#endif

		//And print the name of the item
		if(m_pxTree->IsSelected(pxItem))
		{
			uColour = 0xffff00;
		}
		else if(pxItem == pxItemUnderCursor && eElement == ASURA_TREE_LINEELEMENT_NAME)
		{
			uColour = GetColour(m_bClickedThisFrame);
		}
		else
		{
			if(pxItem->GetColourOverride())
			{
				uColour = pxItem->GetColourOverride();
			}
			else
			{
				uColour = IsEnabled() ? ASURA_GADGET_DEFAULT_COLOUR : ASURA_GADGET_DISABLED_COLOUR;
			}
		}
		uColour |= Asura_OverlayBrightness;

		//Render the name , unless currently being renamed
		if(pxItem != m_pxTreeDialog->GetItemBeingRenamed())
		{
			#ifdef TOOLS_USE_OLD_ENTITY_MANAGER
			Asura_Font_System::Print(xRect,fXOffset+ASURA_TREE_DIALOG_TEXTOFFSET,fYOffset,0.0f,1.0f,uColour,0,pxItem->GetName());
			#else
			Asura_Font_System::Print(xRect,fXOffset/*+ASURA_TREE_DIALOG_TEXTOFFSET*/,fYOffset,0.0f,1.0f,uColour,0,pxItem->GetName());
			#endif
		}

		//move onto the next
		fYOffset += ASURA_TEXT_OUTPUT_FONT_HEIGHT;
		iIndex++;
	}

	if(m_bDragConfirmed  && m_xDragDropCallback.IsValid())
	{
		RenderDraggedItemNames();
	}

    Asura_Gadget::Render( xOrigin );
}

void Asura_TreeView_Gadget::GetItemUnderCursor(const Asura_Vector_2 xOrigin,Asura_TreeDialog_Item*& pxOutputItem, Asura_Tree_LineElement* peElement) const
{
	pxOutputItem = 0;

	if(m_bIsMouseOver)
	{
		//Get position of cursor in tree
		float fX,fY;
		Asura_Cursor::GetScreenCoords(fX,fY);

		fX -= m_xOffset.x + xOrigin.x;
		fY -= m_xOffset.y + xOrigin.y;

		if(fY < 0.0f) return;
		if(fY > m_fHeight) return;

		fX += m_pxTreeDialog->GetHorizontalScrollPos();		
		fY += m_pxTreeDialog->GetVerticalScrollPos();
		
		//which item are we over?
		int iIndex = Asura_Maths::FloatToInt_Chop(fY);

		fX *= ASURA_TEXT_OUTPUT_FONT_WIDTH;

		pxOutputItem = m_pxTreeDialog->GetVisibleItem(iIndex);
		if(pxOutputItem)
		{
			#ifdef TOOLS_USE_OLD_ENTITY_MANAGER
			//get horizontal position relative to the start of the item text
			fX -= ((float)(pxOutputItem->GetDepth()-1) * ASURA_TREE_DIALOG_INDENT);
			#endif

			if(peElement)
			{
				#ifdef TOOLS_USE_OLD_ENTITY_MANAGER

				if(fX <= ASURA_TREE_DIALOG_TEXTOFFSET/2.0f)
				{
					//On the left , so treat as being over the expand symbol
					*peElement = ASURA_TREE_LINEELEMENT_EXPAND;
				}
				else if(fX <= ASURA_TREE_DIALOG_TEXTOFFSET)
				{
					//over the folder(or whatever) icon
					*peElement = ASURA_TREE_LINEELEMENT_ICON;
				}
				else
				{
					//clicked somewhere to the right of the expand symbol ,so treat as being on the name
					*peElement = ASURA_TREE_LINEELEMENT_NAME;
				}

				#else

				if (fX < pxOutputItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_ICONS])
				{
					// On the left, so treat as being over the expand symbol
					*peElement = ASURA_TREE_LINEELEMENT_EXPAND;
				}
				else if (fX < pxOutputItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_ADDITIONAL_ICONS])
				{
					// Over the folder(or whatever) icon
					*peElement = ASURA_TREE_LINEELEMENT_ICON;
				}
				else if (fX < pxOutputItem->m_afXOffsets [Asura_TreeDialog_Item::XOFFSET_TEXT])
				{
					*peElement = ASURA_TREE_LINEELEMENT_ADDITIONAL_ICON;
				}
				else
				{
					// Clicked somewhere to the right of the expand symbol, so treat as being on the name
					*peElement = ASURA_TREE_LINEELEMENT_NAME;
				}

				#endif
			}
		}
	}
}

void Asura_TreeView_Gadget::CallDragDropCallback()
{
	if(m_pxDragTarget && m_xDragDropCallback.IsValid())
	{
		m_xDragDropCallback.Execute(m_pxDragTarget);
	}
}

void Asura_TreeView_Gadget::RenderIcon(const Asura_Bounding_Rectangle& xRect, float fXOffset, float fYOffset, int iIconIndex, u_int uColour)
{
	static int s_iTextureIndex = ASURA_TEXTURE_NONE;
	if(s_iTextureIndex == ASURA_TEXTURE_NONE)
	{
		u_int uOldResetGroup = Asura_Texture_Management::GetTextureResetGroup();
		Asura_Texture_Management::SetTextureResetGroup(ASURA_TEXTURE_LOADED_PERMANENT);
		s_iTextureIndex = Asura_Texture_Management::AddTexture("\\Tools\\ToolsTreeIcons.tga");
		Asura_Texture_Management::SetTextureFlags(s_iTextureIndex, AsuraTextureFlag_HighQualityCompression);
		Asura_Texture_Management::SetTextureResetGroup(uOldResetGroup);

		Asura_Texture_Management::LoadTextures();
	}
	

	const float fCellUVSize = 16.0f/128.0f;
	const float fIconUVSize = 12.0f/128.0f;
	const float fUVNudge = 2.5f/128.0f;
	const float fPixelUV = 1.0f/128.0f;

	float fUVX = (iIconIndex % (128/16)) * fCellUVSize + fUVNudge;
	float fUVY = (iIconIndex / (128/16)) * fCellUVSize + fUVNudge;
	float fUVWidth = fIconUVSize;
	float fUVHeight = fIconUVSize;

	float fX=(float)Asura_Maths::FloatToInt_Near(xRect.MinX + fXOffset + 1.0f);
	float fY=(float)Asura_Maths::FloatToInt_Near(xRect.MinY + fYOffset + 1.0f);
	float fWidth = 12.0f;
	float fHeight = 12.0f;

	//clip to the rectangle
	if(fX < xRect.MinX)
	{
		float fOffset = xRect.MinX-fX;
		
		fUVWidth -= fOffset*fPixelUV;
		fUVX += fOffset*fPixelUV;
		fWidth -= fOffset;
		fX += fOffset;
	}
	if(fX + fWidth > xRect.MaxX)
	{
		float fOffset = (fX + fWidth) + xRect.MaxX;
		
		fUVWidth -= fOffset*fPixelUV;
		fWidth -= fOffset;
	}
	if(fY < xRect.MinY)
	{
		float fOffset = xRect.MinY-fY;
		
		fUVHeight -= fOffset*fPixelUV;
		fUVY += fOffset*fPixelUV;
		fHeight -= fOffset;
		fY += fOffset;
	}
	if(fY + fHeight > xRect.MaxY)
	{
		float fOffset = (fY + fHeight) - xRect.MaxY;
		
		fUVHeight -= fOffset*fPixelUV;
		fHeight -= fOffset;
	} 
	if(fWidth <= 0.0f || fHeight<=0.0f) return;
	
	
	Asura_Render::Overlay.Colour = uColour;
	Asura_Render::Overlay.TextureIndex = s_iTextureIndex;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
	Asura_Render::Overlay.SetRectUV(fUVX, fUVY, fUVWidth, fUVHeight);
	Asura_Render::Overlay.SetRectXY(fX, fY, fWidth+1.0f, fHeight+1.0f);

	Asura_Render::Overlay.Render();

}

struct Asura_TreeItemAndIndex
{
	Asura_TreeDialog_Item* m_pxItem;
	int m_iIndex;
};

static int SortSelection(const void* pData1, const void* pData2)
{
	const Asura_TreeItemAndIndex* pxEntry1 = (const Asura_TreeItemAndIndex*)pData1;
	const Asura_TreeItemAndIndex* pxEntry2 = (const Asura_TreeItemAndIndex*)pData2;

	return pxEntry1->m_iIndex - pxEntry2->m_iIndex;
}

void Asura_TreeView_Gadget::RenderDraggedItemNames()
{
	const List<Asura_TreeDialog_Item*>& xSelection = m_pxTree->GetSelection();
	if(!xSelection.size()) return;

	//get a list of the selected items , and their index
	Asura_TreeItemAndIndex* pxArray = new Asura_TreeItemAndIndex[xSelection.size()];

	for(int i=0; i<xSelection.size(); i++)
	{
		pxArray[i].m_pxItem = xSelection[i];
		pxArray[i].m_iIndex = xSelection[i]->GetVisibleIndex();
		//put items in hidden branches at the end
		if(pxArray[i].m_iIndex == -1) pxArray[i].m_iIndex = 0x7fffffff;
	}

	//sort the selection
	Asura_SortAndSearch::Quicksort(pxArray, xSelection.size(), SortSelection);

	//and show all the names
	float x,y;
	Asura_Cursor::GetScreenCoords(x,y);
	x++;
	y--;

	for(int i=0; i<xSelection.size(); i++)
	{
		Asura_Text_Output::Print(x,y++,0xff00ff00,0,"%s",pxArray[i].m_pxItem->GetName());
	}

	delete [] pxArray;	
}

void Asura_TreeView_Gadget::StartRenameTimer(Asura_TreeDialog_Item* pxItem)
{
	m_pxRenameItem = pxItem;
	//Adding this frame's time on , since the timer will be decreased later on during the same frame
	m_fRenameTimer = 0.5f + Asura_Timers::GetFrameTime();
}

void Asura_TreeView_Gadget::CancelRenameTimer()
{
	m_pxRenameItem = 0;
	m_fRenameTimer = 0.0f;
}

void Asura_TreeView_Gadget::UpdateRenameTimer()
{
	if(m_pxRenameItem && m_fRenameTimer > 0.0f)
	{
		m_fRenameTimer -= Asura_Timers::GetFrameTime();
		if(m_fRenameTimer <= 0.0f)
		{
			//long enough has passed so that it shouldn't have been a double click , so start the renaming
			if(m_pxTree->IsSelected(m_pxRenameItem))
			{
				m_pxTreeDialog->StartRename(m_pxRenameItem);
				CancelRenameTimer();
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Asura_TreeDialog_Tree **
**-----------------------*/

Asura_TreeDialog_Tree::Asura_TreeDialog_Tree()
{
	m_pxRoot = new Asura_TreeDialog_Item(0,0);
	m_pxRoot->Expand(true);//root needs to be expanded to have anything show up.
	m_bAllowMultipleSelection = false;
}

Asura_TreeDialog_Tree::~Asura_TreeDialog_Tree()
{
	m_pxRoot->Delete();
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::AddItem(const char* szName,int iUserData)
{
    return AddItem(GetRoot(),szName,iUserData);
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::InsertItem(int ( *pfnCompFunc )( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo ),const char* szName,int iUserData /*= 0*/)
{
    return InsertItem( pfnCompFunc, GetRoot(), szName, iUserData );
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::AddItem(Asura_TreeDialog_Item* pxParent,const char* szName,int iUserData)
{
	Asura_Assert(pxParent,"Must have a parent for tree item");
	if(!pxParent) return 0;

	Asura_TreeDialog_Item* pxItem = new Asura_TreeDialog_Item(szName,iUserData);
	InsertItem(pxItem,pxParent);

	return pxItem;
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::InsertItem(int ( *pfnCompFunc )( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo ),Asura_TreeDialog_Item* pxParent,const char* szName,int iUserData /*= 0*/)
{
    Asura_Assert(pxParent,"Must have a parent for tree item");
    if(!pxParent) return 0;

    Asura_TreeDialog_Item* pxItem = new Asura_TreeDialog_Item(szName,iUserData);
    InsertItem(pfnCompFunc,pxItem,pxParent);

    return pxItem;
}

static int InsertFolder(const Asura_TreeDialog_Item& xNew,const Asura_TreeDialog_Item& xExisting)
{
	// Folder nodes always come before file (leaf) nodes
	if(xExisting.GetNumChildren() == 0 && !xExisting.ShouldTreatAsFolder()) return -1;

	// Comparing with another folder node
	else return _stricmp(xNew.GetName(),xExisting.GetName());
}

static bool ContainsSlash(const char* const szString,int* piPos=NULL)
{
	if(Asura_StringUtil::Contains(szString,"\\",piPos)) return true;
	else return Asura_StringUtil::Contains(szString,"/",piPos);
}

void Asura_TreeDialog_Tree::AddFolderPath(const char* szName,int iUserData /*=0*/)
{
	AddFolderPath(GetRoot(),szName,iUserData);
}

void Asura_TreeDialog_Tree::AddFolderPath(Asura_TreeDialog_Item* pxParent,const char* szName,int iUserData /*=0*/)
{
	Asura_Assert(pxParent,"Must have a parent for tree item");
	if(!pxParent) return;

	char szWorkspace[MAX_PATH];
	Asura_PrintfToBfr(szWorkspace,"%s",MAX_PATH,true,szName);

	int iSlashPos = 0;
	char* szStart = szWorkspace;

	while(ContainsSlash(szStart,&iSlashPos))
	{
		szStart[iSlashPos] = '\0';

		// Add a node for this subdir if one doesn't exist already
		bool bFound = false;

		for(Asura_TreeDialog_Item* pxCurrChild = pxParent->GetChild();
			pxCurrChild && !bFound;
			pxCurrChild = pxCurrChild->GetSibling())
		{
			if(Asura_StringUtil::CompareAgainstWildcardString(pxCurrChild->GetName(),szStart,true) && pxCurrChild->ShouldTreatAsFolder())
			{
				pxParent = pxCurrChild;
				bFound = true;
			}
		}

		if(!bFound)
		{
			pxParent = InsertItem(InsertFolder,pxParent,szStart,iUserData);
			pxParent->SetTreatAsFolder( true );
		}

		szStart += iSlashPos + 1;
	}

	// Add the final folder, if necessary.
	if ( *szStart )
	{
		bool bFound = false;

		for(Asura_TreeDialog_Item* pxCurrChild = pxParent->GetChild();
			pxCurrChild && !bFound;
			pxCurrChild = pxCurrChild->GetSibling())
		{
			if(Asura_StringUtil::CompareAgainstWildcardString(pxCurrChild->GetName(),szStart,true) && pxCurrChild->ShouldTreatAsFolder())
			{
				pxParent = pxCurrChild;
				bFound = true;
			}
		}

		if(!bFound)
		{
			pxParent = InsertItem(InsertFolder,pxParent,szStart,iUserData);
			pxParent->SetTreatAsFolder( true );
		}
	}
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::AddFilePath(const char* szName,int iFileUserData /*=0*/,int iFolderUserData /*=0*/)
{
	return AddFilePath(GetRoot(),szName,iFileUserData,iFolderUserData);
}

static int InsertFile(const Asura_TreeDialog_Item& xNew,const Asura_TreeDialog_Item& xExisting)
{
	// Folder nodes always come before file (leaf) nodes
	if(xExisting.GetNumChildren() > 0 || xExisting.ShouldTreatAsFolder()) return 1;

	// Comparing with another file node
	else return _stricmp(xNew.GetName(),xExisting.GetName());
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::AddFilePath(Asura_TreeDialog_Item* pxParent,
	const char* szName,int iFileUserData /*=0*/,int iFolderUserData /*=0*/)
{
	Asura_Assert(pxParent,"Must have a parent for tree item");
	if(!pxParent) return 0;

	char szWorkspace[MAX_PATH];
	Asura_PrintfToBfr(szWorkspace,"%s",MAX_PATH,true,szName);

	int iSlashPos = 0;
	char* szStart = szWorkspace;

	while(ContainsSlash(szStart,&iSlashPos))
	{
		szStart[iSlashPos] = '\0';

		// Add a node for this subdir if one doesn't exist already
		bool bFound = false;

		for(Asura_TreeDialog_Item* pxCurrChild = pxParent->GetChild();
			pxCurrChild && !bFound;
			pxCurrChild = pxCurrChild->GetSibling())
		{
			if(Asura_StringUtil::CompareAgainstWildcardString(pxCurrChild->GetName(),szStart,true) && pxCurrChild->ShouldTreatAsFolder())
			{
				pxParent = pxCurrChild;
				bFound = true;
			}
		}

		if(!bFound)
		{
			pxParent = InsertItem(InsertFolder,pxParent,szStart,iFolderUserData);
			pxParent->SetTreatAsFolder( true );
		}

		szStart += iSlashPos + 1;
	}

	// Add the filename as a leaf node
	return InsertItem(InsertFile,pxParent,szStart,iFileUserData);
}

void Asura_TreeDialog_Tree::DeleteItem(Asura_TreeDialog_Item* pxItemToDelete)
{
	if(!pxItemToDelete) return;
	if(pxItemToDelete->IsRoot())
	{
		Asura_Assert(!pxItemToDelete->IsRoot(),"Shuoldn't be deleting the root item like this");
		return;
	}
	if(GetRoot() != pxItemToDelete->GetRoot())
	{
		Asura_Assert(0,"Deleting item that doesn't belong to this tree");
		return;
	}

	//reset the current selection if it is being deleted
	for(LIF<Asura_TreeDialog_Item*> xSelectionLIF(&m_xSelection); !xSelectionLIF.done(); )
	{
		if(xSelectionLIF()->IsDescendentOf(pxItemToDelete))
		{
			xSelectionLIF.delete_current();
		}
		else
		{
			xSelectionLIF.next();
		}
	}


	RemoveItem(pxItemToDelete);
	pxItemToDelete->Delete();
}

void Asura_TreeDialog_Tree::MoveItem(int ( *pfnCompFunc )( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo ), Asura_TreeDialog_Item* pxItem,Asura_TreeDialog_Item* pxNewParent)
{
	if(!pxItem) return;
	if(!pxNewParent) pxNewParent = GetRoot();

	if(pxItem->IsRoot())
	{
		Asura_Assert(0,"Shouldn't be moving the root item");
		return;
	}

	if(GetRoot() != pxItem->GetRoot())
	{
		Asura_Assert(0,"Moving item that doesn't belong to this tree");
		return;
	}

	if(GetRoot() != pxNewParent->GetRoot())
	{
		Asura_Assert(0,"Attempting to move item to a different tree");
		return;
	}

	if(pxNewParent->IsDescendentOf(pxItem))
	{
		Asura_Assert(0,"Attempting to move tree item to be a child of a descendent of itself");
		return;
	}


	List<Asura_TreeDialog_Item*> xBackupSelection = m_xSelection;
	
	//remove from the tree
	RemoveItem(pxItem);

	//and re-insert
	if(pfnCompFunc)
	{
		InsertItem(pfnCompFunc,pxItem,pxNewParent);
	}
	else
	{
		InsertItem(pxItem,pxNewParent);
	}

	//The selection may have been cleared as a result of using RemoveItem() , so reselect here
	m_xSelection = xBackupSelection;
}

void Asura_TreeDialog_Tree::InsertItem(Asura_TreeDialog_Item* pxItem,Asura_TreeDialog_Item* pxParent)
{
	Asura_Assert(pxItem,"");
	Asura_Assert(pxParent,"");
	Asura_Assert(!pxItem->m_pxParent,"Already has a parent");

	pxItem->m_pxParent = pxParent;
	if(pxParent->m_pxLastChild)
	{
		pxParent->m_pxLastChild->m_pxNextSibling = pxItem;
		pxParent->m_pxLastChild = pxItem;
	}
	else
	{
		pxParent->m_pxFirstChild = pxParent->m_pxLastChild = pxItem;
	}
	pxParent->m_iNumChildren++;

	pxItem->UpdateDepth();
	//invalidate count of visible items in parents
	pxParent->InvalidateVisibleDescendents();
}

void Asura_TreeDialog_Tree::InsertItem(int ( *pfnCompFunc )( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo ),Asura_TreeDialog_Item* pxItem,Asura_TreeDialog_Item* pxParent)
{
    Asura_Assert(pxItem,"");
    Asura_Assert(pxParent,"");
    Asura_Assert(!pxItem->m_pxParent,"Already has a parent");

    pxItem->m_pxParent = pxParent;
    if(pxParent->m_pxFirstChild)
    {
        // Special cases first to make the code easier to follow.
        if ( pfnCompFunc( *pxItem, *pxParent->m_pxFirstChild ) <= 0 )
        {
            // This item becomes the first child.
            pxItem->m_pxNextSibling = pxParent->m_pxFirstChild;
            pxParent->m_pxFirstChild = pxItem;
        }
        else if ( pfnCompFunc( *pxItem, *pxParent->m_pxLastChild ) > 0 )
        {
            // This item becomes the last child.
            pxParent->m_pxLastChild->m_pxNextSibling = pxItem;
            pxParent->m_pxLastChild = pxItem;
        }
        else
        {
            // Travel along the sibling list and insert this new item in-place.
            // We can now guarantee that this item won't go at either end, and we can guarantee that there is more than one item already in the list.
            // Therefore we don't need to fix up the first and last pointers and we can skip the first item.
            Asura_TreeDialog_Item* pxPrevItem = pxParent->m_pxFirstChild;
            Asura_TreeDialog_Item* pxCurrItem = pxParent->m_pxFirstChild->GetSibling();
            while ( pxCurrItem )
            {
                if ( pfnCompFunc( *pxItem, *pxCurrItem ) <= 0 )
                {
                    // Insert before the current item.
                    pxPrevItem->m_pxNextSibling = pxItem;
                    pxItem->m_pxNextSibling = pxCurrItem;
                    break;
                }
                else
                {
                    pxPrevItem = pxCurrItem;
                    pxCurrItem = pxCurrItem->GetSibling();
                }
            }
        }
    }
    else
    {
        // No children yet, so no sorting required.
        pxParent->m_pxFirstChild = pxParent->m_pxLastChild = pxItem;
    }
    pxParent->m_iNumChildren++;

    pxItem->UpdateDepth();
    //invalidate count of visible items in parents
    pxParent->InvalidateVisibleDescendents();
}

void Asura_TreeDialog_Tree::RemoveItem(Asura_TreeDialog_Item* pxItemToRemove)
{
	Asura_TreeDialog_Item* pxParent = pxItemToRemove->m_pxParent;

	//Find the reference to the item in the parent's linked list
	Asura_TreeDialog_Item** ppxItemRef = &pxParent->m_pxFirstChild;
	Asura_TreeDialog_Item* pxPrevItem = 0;
	while(*ppxItemRef && *ppxItemRef != pxItemToRemove)
	{
		pxPrevItem = *ppxItemRef;
		ppxItemRef = &(pxPrevItem->m_pxNextSibling);
	}
	
	if(!*ppxItemRef)
	{
		Asura_Assert(0,"Failed to find tree item being deleted");
		return;
	}

	if(pxItemToRemove == pxParent->m_pxLastChild)
	{
		//deleting the last in the list , so update the last child pointer with the previous
		pxParent->m_pxLastChild = pxPrevItem;
	}
	//Replace with the next in line
	*ppxItemRef = pxItemToRemove->m_pxNextSibling;
	pxParent->m_iNumChildren--;

	//Reset the parent of the item being removed
	pxItemToRemove->m_pxParent = 0;
	pxItemToRemove->m_pxNextSibling = 0;
	pxParent->InvalidateVisibleDescendents();

	m_xSelection.delete_entry_if_exists(pxItemToRemove);
	
}

void Asura_TreeDialog_Tree::DeleteAll()
{
    m_pxRoot->Delete();
    m_pxRoot     = new Asura_TreeDialog_Item(0,0);
	m_pxRoot->Expand(true);//root needs to be expanded to have anything show up.
	m_xSelection.empty();
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::FindByHash(const Asura_Hash_ID uHash, Asura_TreeDialog_Item* pxStartItem /* = NULL */)
{
	return ((pxStartItem == NULL) ? m_pxRoot->FindByHash(uHash) : pxStartItem->FindByHash(uHash));
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::FindByName(const char* szName, Asura_TreeDialog_Item* pxStartItem /*= NULL*/)
{
	return ((pxStartItem == NULL) ? m_pxRoot->FindByName(szName) : pxStartItem->FindByName(szName));
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::FindByData(int iUserData, Asura_TreeDialog_Item* pxStartItem /* = NULL */)
{

	return ((pxStartItem == NULL) ? m_pxRoot->FindByData(iUserData) : pxStartItem->FindByData(iUserData));
}

Asura_TreeDialog_Item* Asura_TreeDialog_Tree::GetVisibleItem(int iIndex)
{
	return m_pxRoot->GetVisibleItem(iIndex);
}

int Asura_TreeDialog_Tree::GetNumVisibleItems() const
{
	return m_pxRoot->GetVisibleDescendents();
}

bool Asura_TreeDialog_Tree::BelongsToTree(Asura_TreeDialog_Item* pxItem) const
{
	if(!pxItem) return false;

	Asura_TreeDialog_Item* pxParent = pxItem->m_pxParent;
	while(!pxParent->IsRoot())
	{
		pxParent = pxParent->m_pxParent;
	}
	return (pxParent == m_pxRoot);
}

// Export content of the tree to file.
bool Asura_TreeDialog_Tree::ExportToTXTFile( Asura_String xFileName )
{
	if( m_pxRoot )
	{
		FILE* fp = fopen( xFileName.GetString(), "wb" );
		if( fp )
		{
			m_pxRoot->ExportToTXTFile( fp, 0 );
			fclose( fp );
			return true;
		}
	}

	return false;
}


void Asura_TreeDialog_Tree::SetSelectedItem(Asura_TreeDialog_Item* pxSelected)
{
	if(!pxSelected)
	{
		//deselecting
		m_xSelection.empty();
		return;
	}
	
	if(pxSelected->IsRoot())
	{
		Asura_Assert(0,"Can't select root item");
		return;
	}

	//make sure the selection is in this tree
	if(!BelongsToTree(pxSelected))
	{
		Asura_Assert(0 ,"Item isn't a member of this tree");
		return;
	}

	//expand any parents
	pxSelected->MakeVisible();

	m_xSelection.empty();
	m_xSelection.add_entry(pxSelected);
}

void Asura_TreeDialog_Tree::SetSelection(const List<Asura_TreeDialog_Item*>& xSelection)
{
	m_xSelection.empty();

	for(CLIF<Asura_TreeDialog_Item*> xSelectionLIF(&xSelection); !xSelectionLIF.done(); xSelectionLIF.next())
	{
		AddToSelection(xSelectionLIF());
	}
}


void Asura_TreeDialog_Tree::ToggleSelection(Asura_TreeDialog_Item* pxItem)
{
	if(!pxItem) return;

	if(pxItem->IsRoot())
	{
		Asura_Assert(0,"Can't select root item");
		return;
	}

	//make sure the selection is in this tree
	if(!BelongsToTree(pxItem))
	{
		Asura_Assert(0 ,"Item isn't a member of this tree");
		return;
	}

	if(!m_xSelection.delete_entry_if_exists(pxItem))
	{
		m_xSelection.add_entry(pxItem);
	}
}

void Asura_TreeDialog_Tree::AddToSelection(Asura_TreeDialog_Item* pxItem)
{
	if(!pxItem) return;

	if(pxItem->IsRoot())
	{
		Asura_Assert(0,"Can't select root item");
		return;
	}

	//make sure the selection is in this tree
	if(!BelongsToTree(pxItem))
	{
		Asura_Assert(0 ,"Item isn't a member of this tree");
		return;
	}

	m_xSelection.delete_entry_if_exists(pxItem);
	m_xSelection.add_entry(pxItem);
}

void Asura_TreeDialog_Tree::RemoveFromSelection(Asura_TreeDialog_Item* pxItem)
{
	m_xSelection.delete_entry_if_exists(pxItem);
}


void Asura_TreeDialog_Tree::SelectMultiple(Asura_TreeDialog_Item* pxFirstItem, Asura_TreeDialog_Item* pxLastItem, bool bKeepCurrentSelection)
{
	if(!pxFirstItem || !pxLastItem)
	{
		Asura_Assert(0, "Need two valid tree items for SelectMultiple()");
		return;
	}

	if(!BelongsToTree(pxFirstItem) || !BelongsToTree(pxLastItem))
	{
		Asura_Assert(0, "One of the items being used for SelectMultiple() doesn't belong to the tree");
	}

	//make sure both items are visible
	int iFirstIndex = pxFirstItem->GetVisibleIndex();
	if(iFirstIndex == -1) return;

	int iLastIndex = pxLastItem->GetVisibleIndex();
	if(iLastIndex == -1) return;


	//replacing the current selection?
	if(!bKeepCurrentSelection)
	{
		m_xSelection.empty();
	}

	//select everything between those two
	int iDirection = iFirstIndex<iLastIndex ? 1 : -1;
	for(int i=iFirstIndex; (iDirection > 0) ? i<=iLastIndex : i>=iLastIndex; i += iDirection)
	{
		Asura_TreeDialog_Item* pxItem = GetVisibleItem(i);
		if(pxItem)
		{
			AddToSelection(pxItem);
		}
	}
	

}

void Asura_TreeDialog_Tree::GetSelectionRemovingChildren(List<Asura_TreeDialog_Item*>& xOutput) const
{
	xOutput = m_xSelection;
	//Check for any cases where a descendent of a selected item is also selected.
	for(LIF<Asura_TreeDialog_Item*> xSelectionLIF(&xOutput); !xSelectionLIF.done();)
	{
		Asura_TreeDialog_Item* pxItem1 = xSelectionLIF();
		
		bool bRemoveItem1 = false;
		LIF<Asura_TreeDialog_Item*> xSelectionLIF2(xSelectionLIF);

		for(xSelectionLIF2.next(); !xSelectionLIF2.done();)
		{
			Asura_TreeDialog_Item* pxItem2 = xSelectionLIF2();

			if(pxItem2->IsDescendentOf(pxItem1))
			{
				xSelectionLIF2.delete_current();
			}
			else if(pxItem1->IsDescendentOf(pxItem2))
			{
				bRemoveItem1 = true;
				break;
			}
			else
			{
				xSelectionLIF2.next();
			}
		}

		if(bRemoveItem1)
		{
			xSelectionLIF.delete_current();
		}
		else
		{
			xSelectionLIF.next();
		}
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Asura_TreeDialog_Item **
**-----------------------*/

Asura_TreeDialog_Item::Asura_TreeDialog_Item(const char* szName,int iUserData)
{
	m_uHash = Asura_GetHashID(szName);
	m_xName = szName;
	m_iUserData = iUserData;
	m_pUserPointer = 0;

	m_pxParent = 0;
	m_pxNextSibling = 0;
	m_pxFirstChild = 0;
	m_pxLastChild = 0;
	m_bExpanded = false;
	m_bHidden = false;
	m_iNumChildren = 0;
	m_iDepth = 0;
	m_iVisibleDescendents = -1;
	m_uColourOverride = 0;
	m_bTreatAsFolder = false;

	#ifndef TOOLS_USE_OLD_ENTITY_MANAGER
	m_iRenameExtentLeftIndex = -1;
	m_iRenameExtentRightIndex = -1;
	#endif
}


void Asura_TreeDialog_Item::Delete()
{
	//Delete any children
	Asura_TreeDialog_Item* pxChild = m_pxFirstChild;
	while(pxChild)
	{
		Asura_TreeDialog_Item* pxNextChild = pxChild->m_pxNextSibling;
		pxChild->Delete();
		pxChild = pxNextChild;
	}

	delete this;
}


void Asura_TreeDialog_Item::ConcatenatePath( Asura_String& xPath ) const
{
	const Asura_TreeDialog_Item*	pxParentItem = GetParent();

	if(pxParentItem == NULL)
	{
		xPath	= GetName();
	}
	else
	{
		pxParentItem->ConcatenatePath( xPath );

		if(xPath.IsBlank())
		{
			xPath	= GetName();
		}
		else
		{
			xPath	+= "\\";
			xPath	+= GetName();
		}
	}
}

Asura_TreeDialog_Item* Asura_TreeDialog_Item::FindChildByName(const char* szName, bool bIgnoreCase)
{
	//Check each child
	for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild; pxChild; pxChild = pxChild->m_pxNextSibling)
	{
		if(!pxChild->m_xName.Compare(szName, bIgnoreCase))
		{
			return pxChild;
		}
	}
	return 0;
}

Asura_TreeDialog_Item* Asura_TreeDialog_Item::FindByHash(const Asura_Hash_ID uHash)
{
	if(!IsRoot() && (uHash == m_uHash))
	{
		return(this);
	}
	else
	{
		//Check children
		for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild; (pxChild != NULL); pxChild = pxChild->m_pxNextSibling)
		{
			Asura_TreeDialog_Item*	pxFound	= pxChild->FindByHash(uHash);

			if(pxFound != NULL)
			{
				return(pxFound);
			}
		}

		return(NULL);
	}
}

Asura_TreeDialog_Item* Asura_TreeDialog_Item::FindByName(const char* szName)
{
	if( szName )
	{
		if(!IsRoot())
		{
			//Is this the one we are after?
			if(!strcmp(szName,m_xName.GetString()))
			{
				return this;
			}
		}

		//Check children
		for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild;pxChild;pxChild = pxChild->m_pxNextSibling)
		{
			Asura_TreeDialog_Item* pxFound = pxChild->FindByName(szName);
			if(pxFound) return pxFound;
		}
	}

	return 0;
}

Asura_TreeDialog_Item* Asura_TreeDialog_Item::FindByData(int iUserData)
{
	if(!IsRoot())
	{
		//Is this the one we are after?
		if(iUserData == m_iUserData)
		{
			return this;
		}
	}

	//Check children
	for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild;pxChild;pxChild = pxChild->m_pxNextSibling)
	{
		Asura_TreeDialog_Item* pxFound = pxChild->FindByData(iUserData);
		if(pxFound) return pxFound;
	}
	return 0;
}


void Asura_TreeDialog_Item::ExportToTXTFile( FILE* fp, u_int uTabLevel )
{
	// This tabs field is pretty tricky field. It actually serve for correctly managing | characters. Ken suggested that last entry
	// shouldn't have a | character, so this field is telling, which tab level should have and which shouldn't. True represents that
	// | is drown on this level, and false represent that only empty space is drawn in this space. There is hardcoded value of 100 
	// levels.
	const static int iMaxRecursiveLevelOfTabs = 50;
	static bool s_bTabsField[iMaxRecursiveLevelOfTabs];

	// I am beggining my recursive export in root. However, root doesn't have name, so there is no need to export it.
	if( IsRoot()==false )
	{
		// write correctly | lines. this for loop will ensure that all vertical | are created correctly.
		for( u_int i=2; i<uTabLevel; i++ )
		{
			if( s_bTabsField[i] )	{	fwrite("| ",2,1,fp);	}
			else					{	fwrite("  ",2,1,fp);	}
		}

		if( uTabLevel>=2 )	{	fwrite("|___ ",5,1,fp);	}

		// And now it is time to export also the name of the element.
		fwrite( m_xName.GetString(), m_xName.GetSizeInBytes(), 1, fp );
		fwrite( "\r\n",2,1,fp );		// And don't forget to put enter at the end of the line.
	}

	int iNumOFChildren = GetNumChildren();

	for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild;pxChild;pxChild = pxChild->m_pxNextSibling)
	{
		// This is tricky if. Actually I will set true for each element that is not last one. So that means, 
		// only one last element in every level, is not going to have | character drawn.
		if( uTabLevel<iMaxRecursiveLevelOfTabs-1 )
		{
			s_bTabsField[uTabLevel+1]= (iNumOFChildren>1) ? true : false;
		}

		pxChild->ExportToTXTFile( fp, uTabLevel+1 );

		// Makes text output more readable
		if( IsRoot() )
		{
			fwrite( "\r\n",2,1,fp );
			fwrite( "\r\n",2,1,fp );
		}

		iNumOFChildren--;
	}
}


void Asura_TreeDialog_Item::UpdateVisibleDescendents() const
{
	m_iVisibleDescendents = 0;
	if(m_bExpanded)
	{
		for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild;pxChild;pxChild = pxChild->m_pxNextSibling)
		{
			//add child's count to this
			if(!pxChild->IsHidden())
			{
				m_iVisibleDescendents += 1 + pxChild->GetVisibleDescendents();
			}
		}
	}
}

void Asura_TreeDialog_Item::Expand(bool bExpand)
{
	if(m_bExpanded == bExpand) return;
	m_bExpanded = bExpand;

	//Determine new count of visible children
	int iOldCount = m_iVisibleDescendents;
	UpdateVisibleDescendents();

	//Update counts in the parents
	if(!m_bHidden)
	{
		Asura_TreeDialog_Item* pxParent = m_pxParent;
		while(pxParent)
		{
			//If the parent isn't expanded , its count will be unaffected
			if(!pxParent->m_bExpanded) break;
			//If the parent's count is invalid , then there is no need to correct it just yet
			if(!pxParent->IsVisibleDescendentCountValid()) break;

			pxParent->m_iVisibleDescendents += m_iVisibleDescendents - iOldCount;
			Asura_Assert(pxParent->m_iVisibleDescendents >= 0,"Code for tracking visible children gone screwy");
	
			pxParent = pxParent->m_pxParent;
		}
	}

	// If the alt-key is pressed, then scan through all nodes on the same level and expand these also

	if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
	{
		if(m_pxParent)
		{
			Asura_TreeDialog_Item* pxItem = m_pxParent->GetChild();

			for(; pxItem; pxItem = pxItem->GetSibling())
			{
				pxItem->Expand(bExpand);
			}
		}
	}
}

void Asura_TreeDialog_Item::SetHidden(bool bHide)
{
	if(m_bHidden == bHide) return;
	m_bHidden = bHide;

	if(m_pxParent)
	{
		m_pxParent->InvalidateVisibleDescendents();
	}
}

void Asura_TreeDialog_Item::InvalidateVisibleDescendents()
{
	Asura_TreeDialog_Item* pxItem = this;
	while(pxItem && pxItem->IsVisibleDescendentCountValid())
	{
		pxItem->m_iVisibleDescendents = -1;
		//Invalidate count in parent as well
		pxItem = pxItem->m_pxParent;
	}
}


Asura_TreeDialog_Item* Asura_TreeDialog_Item::GetVisibleItem(int iIndex) const
{
	if(IsRoot())
	{
		//Make sure the index is in range before searching further
		int iVisible = GetVisibleDescendents();
		if(iVisible <= iIndex || iIndex<0) return 0;
	}

	//check children
	for(Asura_TreeDialog_Item* pxChild = m_pxFirstChild;pxChild;pxChild = pxChild->m_pxNextSibling)
	{
		if(pxChild->IsHidden()) continue;
		if(iIndex == 0)
		{
			//immediate child is the one we want
			return pxChild;
		}
		iIndex--;

		//is the item we are after among the descendents of this child?
		int iVisible = pxChild->GetVisibleDescendents();
		if(iIndex < iVisible)
		{
			//It is
			return pxChild->GetVisibleItem(iIndex);
		}
		else
		{
			//It isn't , so skip past them all
			iIndex -= iVisible;
		}
	}
	//Should have found something by now
	Asura_Assert(0,"Probably shouldn't be getting here");
	return 0;
}

void Asura_TreeDialog_Item::MakeVisible()
{
	Asura_TreeDialog_Item* pxParent = m_pxParent;
	while(!pxParent->IsRoot())
	{
		pxParent->Expand(true);
		pxParent = pxParent->m_pxParent;
	}
}

int Asura_TreeDialog_Item::GetVisibleIndex() const
{
	//If parent isn't expanded then this isn't visible
	if(!m_pxParent->IsExpanded()) return -1;
	//And if this is hidden , it isn't visible
	if(m_bHidden) return -1;

	//get the parent's index
	int iIndex = -1;
	if(!m_pxParent->IsRoot())
	{
		iIndex = m_pxParent->GetVisibleIndex();
		if(iIndex == -1) return -1;
	}

	//determine position in parent's list
	Asura_TreeDialog_Item* pxSibling = m_pxParent->m_pxFirstChild;
	while(pxSibling)
	{
		if(!pxSibling->IsHidden())
		{
			iIndex++;
			if(pxSibling == this) break;
			//add any children of previous siblings , since they will be shown first
			iIndex += pxSibling->GetVisibleDescendents();
		}
		pxSibling = pxSibling->m_pxNextSibling;
	}
	Asura_Assert(pxSibling == this,"Tree data not happy");

	return iIndex;
}

void Asura_TreeDialog_Item::UpdateDepth()
{
	m_iDepth = m_pxParent ? m_pxParent->m_iDepth+1 : 0;

	//update children
	Asura_TreeDialog_Item* pxChild = m_pxFirstChild;
	while(pxChild)
	{
		pxChild->UpdateDepth();
		pxChild = pxChild->m_pxNextSibling;
	}
}

// Helper function that will go trough all Asura_TreeDialog_Item and will try to find one which hash with path corresponds to it.
Asura_TreeDialog_Item* Asura_TreeDialog_FileTree::FindTreeItemByHashWithPath( Asura_TreeDialog_Item* pxRootItem, Asura_Hash_ID uSearchItemHashID )
{
	// Is there is something to look for? If not, don't continue.
	if( uSearchItemHashID==ASURA_HASH_ID_UNSET )
	{
		return NULL;
	}

	if( pxRootItem )
	{
		Asura_String xPath;
		pxRootItem->ConcatenatePath( xPath );

		if( uSearchItemHashID==Asura_GetHashID( xPath.GetString() ) )
		{
			return pxRootItem;
		}
		else
		{

			Asura_TreeDialog_Item* pxChild = pxRootItem->GetChild();
			while ( pxChild )
			{
				Asura_TreeDialog_Item* pxSearchResult = FindTreeItemByHashWithPath( pxChild, uSearchItemHashID );
				if ( pxSearchResult )
				{
					return pxSearchResult;
				}

				pxChild = pxChild->GetSibling();
			}
		}
	}

	return NULL;
}


Asura_TreeDialog_Item* Asura_TreeDialog_FileTree::CreateItem(const Asura_String& xPath)
{
	if(xPath.IsBlank())
	{
		return(NULL);
	}
	else
	{
		Asura_String	xParent;
		Asura_String	xName;

		if(xPath.SplitFilename(&xParent, &xName, NULL))
		{
			const char*				szName			= xName.GetString();
			Asura_TreeDialog_Item*	pxParentItem	= CreateItem(xParent);
			Asura_TreeDialog_Item*	pxExistingItem	= FindByName(szName, pxParentItem);

			if(pxExistingItem == NULL)
			{
				Asura_TreeDialog_Item*	pxItem	= (pxParentItem == NULL)
												? PARENT::AddItem(szName, DIRECTORY_ITEM)
												: PARENT::AddItem(pxParentItem, szName, DIRECTORY_ITEM);

				return(pxItem);
			}
			else
			{
				return(pxExistingItem);
			}
		}
		else
		{
			return(NULL);
		}
	}
}

Asura_TreeDialog_Item* Asura_TreeDialog_FileTree::AddItem( const Asura_String& xPath )
{
	Asura_String	xParent;
	Asura_String	xName;

	if(xPath.SplitFilename(&xParent, &xName, NULL))
	{
		const Asura_Hash_ID		uHash			= xName.GetHash();
		Asura_TreeDialog_Item*	pxExistingItem	= FindByHash(uHash);
		Asura_TreeDialog_Item*	pxParentItem	= CreateItem(xParent);

		if(pxExistingItem == NULL)
		{
			const char*				szName	= xName.GetString();
			Asura_TreeDialog_Item*	pxItem	= (pxParentItem == NULL)
											? PARENT::AddItem(szName, FILE_ITEM)
											: PARENT::AddItem(pxParentItem, szName, FILE_ITEM);

			return(pxItem);
		}
		else
		{
			return(pxExistingItem);
		}
	}
	else
	{
		return(NULL);
	}
}


void Asura_TreeDialog_FileTree::PopulateTree( const char* szRootDirectory, const char* szFileExtension, const bool bAllowNone )
{
	const u_int			uFileFlags				= TOOLS_FILEFIND_RECURSIVE | TOOLS_FILEFIND_IGNOREBACKUPDIR;
	const int			iRootDirectoryLength	= strlen(szRootDirectory) + 1;	// +1 for the '\'.
	const int			iFileExtensionLength	= strlen(szFileExtension);

	Asura_String		xDirectoryAndPattern(szRootDirectory);
	List<Asura_String>	xList;

	// Find the appropriate files...
	xDirectoryAndPattern	+= "\\*.";
	xDirectoryAndPattern	+= szFileExtension;

	Tools_FileSystem::FindFilesLocalAndNetwork(xDirectoryAndPattern.GetString(), xList, uFileFlags);

	// Add each file in turn...
	for(int i = 0; i < xList.size(); i++)
	{
		Asura_String	xFullFilename(xList.get_entry(i));
		const int		iFilenameLength	= xFullFilename.GetLength();

		xFullFilename	= xFullFilename.SubString(iRootDirectoryLength, (iFilenameLength - (iRootDirectoryLength + iFileExtensionLength)));
		AddItem(xFullFilename);
	}

	// Add Tail. This tail will ensure that user can select [NONE] and thus effectively set Face object to UNSET. EP.12.Nov.2008
	if( bAllowNone )
	{
		PARENT::AddItem(GetRoot(), "[NONE]", UNSET_ITEM ); 
	}
}

bool Asura_TreeDialog_FileTree::IsItemADirectory(const Asura_TreeDialog_Item& xItem) const
{
	const int	iUserData	= xItem.GetUserData();

	return(iUserData == DIRECTORY_ITEM);
}

bool Asura_TreeDialog_FileTree::IsItemAFile(const Asura_TreeDialog_Item& xItem) const
{
	const int	iUserData	= xItem.GetUserData();

	return(iUserData == FILE_ITEM);
}

bool Asura_TreeDialog_FileTree::IsItemUnset(const Asura_TreeDialog_Item& xItem) const
{
	const int	iUserData	= xItem.GetUserData();

	return(iUserData == UNSET_ITEM );
}

#endif //#ifdef ASURA_TOOLS


/// End of file. ///
