//////////////////////////////////////////////////////////////////////////////
//
// Asura_Job.h
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     Interface for batching up various processing tasks (methods and data)
//     so they can be processed in parallel on platforms that benefit from
//     multiple threads.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_JOB_H___
#define ___ASURA_JOB_H___ 1

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE


// On PC and 360, we can turn multithreading on/off from the console, for profiling
class Asura_Job_Queue_System
{
public:
	static void SetForceSingleThreaded(bool bVal) {m_bForceSingleThreaded=bVal;};
	static bool GetForceSingleThreaded()		  {return m_bForceSingleThreaded;};
	
private:
	static bool m_bForceSingleThreaded;
};



// new system, written for PS3 to use libsync library's Queue implementation.
// Since that lib deals with blocks of data, C-style, Jobs must be simple data
// aggregates that don't require virtual functions.
class Asura_Job_Queue
{
public:

	Asura_Job_Queue(u_int uDepth, u_int uElementSize, bool(*pfnExecuteFunction)(void *) )
	: m_uDepth(uDepth)
	, m_pxPlatformData(NULL)
	{		
		// DMR 13/11/06 - Added an 'execute job' callback. Not sure if this will work for PS3,
		// but it'll keep things tidier on PC/360
		m_pfnExecuteJob = pfnExecuteFunction;
		Platform_Initialise(uDepth, uElementSize);
	}
	~Asura_Job_Queue()
	{
		Clear();
		Platform_Shutdown();
	}

	// Platform_Push should increment the number of unfinished jobs
	void Push( void* pData )
	{
		Platform_Push( pData );
	}

	// Pop() should block if no element is available to come off the queue.
	// Pop() needs to have a buffer of the right size (uElementSize) passed in to copy the data to
	void Pop( void* pBuffer )
	{
		Platform_Pop( pBuffer );
	}

	// TryPop() is like Pop() but doen't block, and returns whether an element was popped or not.
	bool TryPop( void* pBuffer )
	{
		return Platform_TryPop( pBuffer );
	}

	u_int GetNumEntries( void ) const
	{
		return Platform_GetNumEntries();
	}

	// After a job has been executed, the code that did it should call JobFinished()
	// Platform_JobFinished should decrement the number of unfinished jobs
	void JobFinished( void ) const
	{
		Platform_JobFinished();
	}

	// Number of unfinished jobs is maintained so code that requires all jobs to be 
	// complete can wait until they are.
	u_int GetNumUnfinishedJobs( void ) const
	{
		return Platform_GetNumUnfinishedJobs();
	}

	void Clear()
	{
		Platform_Clear();
	}
	
	// Give the thread a name for debuging and profiling
	void SetThreadName( const char* pszThreadName)
	{
		Platform_SetThreadName( pszThreadName);
	}
	
	void WaitForAllJobsToFinish()
	{
		Platform_WaitForAllJobsToFinish();
	}

private:

	// non-copiable
	Asura_Job_Queue(const Asura_Job_Queue&);
	Asura_Job_Queue& operator=(const Asura_Job_Queue&);

	// platform functions
	void Platform_Initialise( u_int uDepth, u_int uDataSize );
	void Platform_Shutdown( void );
	u_int Platform_GetNumEntries( void ) const;
	// Platform_Push should increment the number of unfinished jobs
	void Platform_Push( void* pData );
	void Platform_Pop( void* pBuffer );
	bool Platform_TryPop( void* pBuffer );
	// After a job has been executed, the code that did it should call JobFinished()
	// Platform_JobFinished should decrement the number of unfinished jobs
	void Platform_JobFinished( void ) const;
	// Number of unfinished jobs is maintained so code that requires all jobs to be 
	// complete can wait until they are.
	u_int Platform_GetNumUnfinishedJobs( void ) const;
	void Platform_Clear( void );
	// Give the thread a name for debuging and profiling
	void Platform_SetThreadName( const char* pszThreadName);
	// Wait for All the JObs to be finished. If the jobs are CPU based this thread can help complete them.
	void Platform_WaitForAllJobsToFinish();


	// member data
	const u_int m_uDepth;
	void* m_pxPlatformData;
	bool(*m_pfnExecuteJob)(void *);

public:
	bool ExecuteJob(void *pxData) 
	{
		bool bResult=false;
		if (m_pfnExecuteJob)
		{
			bResult = m_pfnExecuteJob(pxData);
		}
		return bResult;
	}
};

#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_JOB_H___
