//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Job.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// By Dave Reed (13/11/06)
// By Duncan Hopkins (01/07/06)
//
// Description:
//
// This file is a platform independent version of platform routines for the Asura_Job_Queue.
// To use this file the Asura_Job_TaskQueue and Asura_Job_Manger must be active and in use.
// If the platform (e.g. PS3) wants to do its own thing then create one of these file in the /[Platform]/Jobs/Asura_Platform_Job.cpp
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

// asura independent
#include "Core/Asura_Atomic.h"
#include "Core/Asura_MultiThread.h"
#include "FileAccess/Asura_DebugFileStream.h"
#include "Jobs/Asura_Job.h"
#include "Jobs/Asura_Job_Manager.h"
#include "Jobs/Asura_Job_TaskQueue.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_CPUProfile.h"
#include <Process.h>

// C O N S T A N T S /////////////////////////////////////////////////////////

// G L O B A L S /////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
// Asura_Job_Queue
//////////////////////////////////////////////////////////////////////////////


// Platform specific data for this system
// --------------------------------------------------------------------
class Asura_Job_Queue_Data
{
public:
	Asura_Job_Queue*		m_pxJobQueueObject;
	Asura_Job_TaskQueue*	m_pxTaskQueue;
	int						m_iNumPendingJobs;
};

//////////////////////////////////////////////////////////////////////////////

// Platform_Initialise - Init the data and create the worker threads
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_Initialise( u_int uDepth, u_int uDataSize )
{
	m_pxPlatformData = new Asura_Job_Queue_Data;
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	// Create the queue
	pxData->m_pxJobQueueObject = this;
	pxData->m_pxTaskQueue = new Asura_Job_TaskQueue( uDepth, uDataSize);

	// Init pending job count
	pxData->m_iNumPendingJobs = 0;
}


// Platform_Shutdown - Terminate the worker threads, and free resources
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_Shutdown( void )
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;
	
	if (pxData)
	{	
		// Destroy worker threads / wait for them to terminate
		while(GetNumUnfinishedJobs())
		{
			Asura_Thread_System::YieldCurrentThread();
		}
		
		if( NULL != pxData->m_pxTaskQueue)
		{
			delete pxData->m_pxTaskQueue;
			pxData->m_pxTaskQueue = NULL; 
		}

		void* pxPlatformNULL = NULL;
		Asura_Atomic::ExchangePointer( this->m_pxPlatformData, pxPlatformNULL);
		
		// Delete the queue
		delete pxData;
	}
}

// Platform_SetThreadName - Give the thread a name
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_SetThreadName( const char* pszThreadName)
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;
	if( NULL != pxData)
	{
		pxData->m_pxTaskQueue->SetName( pszThreadName);
	}
}



// Get the number of entries in the job queue
// --------------------------------------------------------------------
u_int Asura_Job_Queue::Platform_GetNumEntries( void ) const
{
	u_int uLength = 0;
	
	// Return number of jobs left in the queue
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;
	
	if (pxData)
	{
		uLength = pxData->m_pxTaskQueue->GetNumberOfQueuedTasks();
	}
	
	return( uLength);
}

// Add a job to the queue (or execute it immediately if we're running
// single-threaded)
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_Push( void* pJob )
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	if (pxData)
	{
		// Increment the number of pending jobs
		Asura_Atomic::Increment( pxData->m_iNumPendingJobs);

		// Push the job
		while( false == pxData->m_pxTaskQueue->Push( pJob))
		{
			Asura_Thread_System::YieldCurrentThread();
		}
	
		// Signal the worker threads that there's something in the queue
		Asura_Job_Manager::AddJob( this, pxData->m_pxTaskQueue);
	}
}

// Attempt to pop a job from the queue
// --------------------------------------------------------------------
bool Asura_Job_Queue::Platform_TryPop( void* pBuffer )
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	if (pxData)
	{
		return pxData->m_pxTaskQueue->Pop( pBuffer);
	}

	return false;
}


// After a job has been executed, the code that did it should call JobFinished()
// Platform_JobFinished should decrement the number of unfinished jobs
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_JobFinished( void ) const
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	if (pxData)
	{
#if defined( ASURA_ASSERT)
		LONG iValue = 
#endif
		Asura_Atomic::Decrement( pxData->m_iNumPendingJobs);
		
		Asura_Assert( 0 <= iValue, "This value is too small");
	}
}


// Number of unfinished jobs is maintained so code that requires all jobs to be 
// complete can wait until they are.
// --------------------------------------------------------------------
u_int Asura_Job_Queue::Platform_GetNumUnfinishedJobs( void ) const
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	int iCount = 0;

	if (pxData)
	{
		iCount = pxData->m_iNumPendingJobs;
	}
		
	return iCount;
}


// Clear any queued jobs
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_Clear( void )
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	if (pxData)
	{
		{
			Asura_Job_TaskQueue::Task* pxTask = NULL;
			while( true == pxData->m_pxTaskQueue->PopTask( pxTask))
			{
#if defined( ASURA_ASSERT)
				LONG iValue = 
#endif
				Asura_Atomic::Decrement( pxData->m_iNumPendingJobs);
				Asura_Assert( 0 <= iValue, "This value is too small");
				
				pxData->m_pxTaskQueue->ReturnTask( pxTask);
			}
		}
	
		
		// Not sure if we need this? But we can't set that NumPendingJobs to zero here.
		while( 0 < pxData->m_iNumPendingJobs)
		{
			Asura_Thread_System::YieldCurrentThread();
		}
		
		Asura_Assert( pxData->m_pxTaskQueue->IsValid(), "Task Queue Not Valid");
	}
}

// Process any pending Jobs
// --------------------------------------------------------------------
void Asura_Job_Queue::Platform_WaitForAllJobsToFinish()
{
	Asura_Job_Queue_Data *pxData = (Asura_Job_Queue_Data*)m_pxPlatformData;

	if (pxData)
	{
		Asura_Job_Manager::WaitForAllJobsToFinish();
	}
}



#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
//////////////////////////////////////////////////////////////////////////////
