//
// Asura_GI_Solution.h
//
// Created by KJFL - 2004-07-08 11:48:32
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_GI_SOLUTION_H__
#define __ASURA_GI_SOLUTION_H__

#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"

//In tools I'm using multiple GI handlers for storing the data for the different light sets.
//In streaming games the multiple GI handlers will correspond to different zones.
//If we ever have multiple loaded zones in tools , together with multiple light sets... then things
//may get interesting. RAR 27/03/2006
#if defined(ASURA_USE_STREAMING) || defined(ASURA_TOOLS)
#define ASURA_MULTIPLE_GI_HANDLER 1
#endif

class Asura_RGBA;

class Asura_GI_Solution_Handler
{
public:
	Asura_GI_Solution_Handler();
	~Asura_GI_Solution_Handler();

	//! called by Asura_Core::Initialise()
	void Initialise( void );

	//! called by Asura_Core::Reset(), and other functions wishing to refresh the data
	void Reset( void );
	
	//! called by Asura_Core::Shutdown()
	void Shutdown( void );

	//! Get sample of the static irradiance (i.e. the pre-processed lighting)
	void GetStaticIrradiance( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition );
		
	//! Get sample of the dynamic irradiance (i.e. the dynamic Asura_Photons)
	void GetDynamicIrradiance( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition );

	//! Get sample of the all irradiance (i.e. the pre-processed lighting + dynamic Asura_Photons)
	//! This is merely the sum of the results of the GetStaticIrradiance() and GetDynamicIrradiance() functions
	void GetTotalIrradiance( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition );

	//! This function has been made as a method of reusing the old lighting, and so should only be used
	//! if really necessary (i.e. making a demo which doesn't have the new GI lighting)
	void GetStaticIrradianceFromOldLighting( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition );

#ifdef ASURA_TOOLS
	//! called by the Tools to put the data into the correct compressed format
	void Tools_InitialiseFromPhotonMappingSystem( void );

	//Generates a single fake sample for projects that don't need the lighting
	void Tools_GenerateFakeSample();

	//! Used by tools to do a quick update when global ambience changed
	void ApplyGlobalAmbienceDelta(const Asura_RGBA& xColourDiff);

	//Returns true if the samples exists , and if they are calculated from the current lighting data.
	//(This is still treated as valid even if the most recent lighting data used is itself invalid)
	bool AreSamplesValid() { return m_bSamplesValid && m_uNumberOfSamples; }

	//Marks the samples as invalid , but doesn't cause them to be deleted.
	//The invalid samples will continue to be used until recalculated.
	static void InvalidateSamples(Asura_Hash_ID uHash,Asura_GI_Solution_Handler* pxHandler) { pxHandler->m_bSamplesValid = false; }

#endif

    // Utility method to sample the total irradiance at the specified
    // point for the given normal. Note that the alpha component of the
    // Asura_RGBA instance provided is not modified.
    void CalculateDiffuseAtPoint( Asura_RGBA& xDiffuse,
                                    const Asura_Vector_3& xPosition,
                                    const Asura_Vector_3& xNormal,
                                    bool bIncludeDynamic );

	// This second version doesn't take a normal - it just kind of calculates the maximum possible brightness from any direction
	// It's a bit dodgy mathematically because there are literally no scaling factors applied - you should probably use a normal
	// if you have one available.
	void CalculateDiffuseAtPoint( Asura_RGBA& xDiffuse, const Asura_Vector_3& xPosition );
	
	u_int GetNumberOfSamples ( void ) { return m_uNumberOfSamples; }

#if defined(ASURA_USE_DEBUGOVERLAY) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	void DebugDrawGISamples( void );
#endif

private:

	friend class Asura_Chunk_GI_Solution;

	//! storage space for irradiance samples
	u_int										m_uNumberOfSamples;
	class Asura_GI_IrradianceSample*			m_pxSamples;

	//! buffers for search
	const class Asura_GI_IrradianceSample* *	m_ppxProximalSamples;
	float*										m_fProximalSampleDistancesSquared;

#ifndef ASURA_TARGET_SM3
	//! decompression data. no longer used in next-gen
	float										m_afDecompressCoefficientMinimum[Asura_SphericalHarmonics_Order2_RGB::NUMBER_OF_COEFFICIENTS];
	float										m_afDecompressCoefficientScale[Asura_SphericalHarmonics_Order2_RGB::NUMBER_OF_COEFFICIENTS];
#endif

#ifdef ASURA_TOOLS
	bool										m_bSamplesValid;
#endif
};


class Asura_GI_Solution
{
	public:
#if defined(ASURA_USE_DEBUGOVERLAY) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
		static void DoDebugOverlay (void* const pxIgnore );
#endif

		//! called by Asura_Core::Initialise()
		static void Initialise( void );

		//! called by Asura_Core::Reset(), and other functions wishing to refresh the data
		static void Reset( void );
		
		//! called by Asura_Core::Shutdown()
		static void Shutdown( void );

		//! Get sample of the static irradiance (i.e. the pre-processed lighting)
		static void GetStaticIrradiance( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition )
		{
			GetCurrentHandler().GetStaticIrradiance(xResult, xPosition);
		}
			
		//! Get sample of the dynamic irradiance (i.e. the dynamic Asura_Photons)
		static void GetDynamicIrradiance( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition )
		{
			GetCurrentHandler().GetDynamicIrradiance(xResult, xPosition);
		}

		//! Get sample of the all irradiance (i.e. the pre-processed lighting + dynamic Asura_Photons)
		//! This is merely the sum of the results of the GetStaticIrradiance() and GetDynamicIrradiance() functions
		static void GetTotalIrradiance( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition )
		{
			GetCurrentHandler().GetTotalIrradiance(xResult, xPosition);
		}

		//! This function has been made as a method of reusing the old lighting, and so should only be used
		//! if really necessary (i.e. making a demo which doesn't have the new GI lighting)
		static void GetStaticIrradianceFromOldLighting( Asura_SphericalHarmonics_Order2_RGB& xResult, const Asura_Vector_3& xPosition )
		{
			GetCurrentHandler().GetStaticIrradianceFromOldLighting(xResult, xPosition);
		}

#ifdef ASURA_TOOLS
		//! called by the Tools to put the data into the correct compressed format
		static void Tools_InitialiseFromPhotonMappingSystem( void );

		//Generates a single fake sample for projects that don't need the lighting
		static void Tools_GenerateFakeSample();

		//! Used by tools to do a quick update when global ambience changed
		static void ApplyGlobalAmbienceDelta(const Asura_RGBA& xColourDiff)
		{
			GetCurrentHandler().ApplyGlobalAmbienceDelta(xColourDiff);
		}

		//Switches off static lighting. Needed for hierarchy mode to prevent characters picking
		//up lighting from the environment
		static void DisableStaticLighting(bool bDisable) { s_bDisableStaticLighting = bDisable; }

		//Returns true if the samples exists , and if they are calculated from the current lighting data.
		//(This is still treated as valid even if the most recent lighting data used is itself invalid)
		static bool AreSamplesValid();

		//Marks the samples as invalid , but doesn't cause them to be deleted.
		//The invalid samples will continue to be used until recalculated.
		//For simplicity , going to invalidate all hadnlers here. RAR 27/03/2006
		static void InvalidateSamples()
		{
			s_xHandlerTree.Traverse(Asura_GI_Solution_Handler::InvalidateSamples);
		}

#endif

        // Utility method to sample the total irradiance at the specified
        // point for the given normal. Note that the alpha component of the
        // Asura_RGBA instance provided is not modified.
        static void CalculateDiffuseAtPoint( Asura_RGBA& xDiffuse,
                                             const Asura_Vector_3& xPosition,
                                             const Asura_Vector_3& xNormal,
                                             bool bIncludeDynamic )
		{
			GetCurrentHandler().CalculateDiffuseAtPoint(xDiffuse, xPosition, xNormal, bIncludeDynamic);
		}

		// This second version doesn't take a normal - it just kind of calculates the maximum possible brightness from any direction
		// It's a bit dodgy mathematically because there are literally no scaling factors applied - you should probably use a normal
		// if you have one available.
		static void CalculateDiffuseAtPoint( Asura_RGBA& xDiffuse, const Asura_Vector_3& xPosition )
		{
			GetCurrentHandler().CalculateDiffuseAtPoint(xDiffuse, xPosition);
		}

#if defined(ASURA_MULTIPLE_GI_HANDLER)
		// Handler functions...
		static	Asura_GI_Solution_Handler*	CreateHandler(const Asura_Hash_ID uHash);
		static	bool						DestroyHandler(const Asura_Hash_ID uHash);

		static	void						CreateAndSetHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = CreateHandler(uHash); }
		static	void						SetCurrentHandler(Asura_GI_Solution_Handler* pxHandler) { s_pxCurrentHandler = pxHandler; }
		static	void						SetCurrentHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
		static	Asura_GI_Solution_Handler&	GetCurrentHandler(void) { return *s_pxCurrentHandler; }
		static	Asura_GI_Solution_Handler*	GetHandler(const Asura_Hash_ID uHash) { return s_xHandlerTree.GetDataByHash(uHash); }
#else
		// Handler functions...
		static	Asura_GI_Solution_Handler&	GetCurrentHandler(void) { return s_xHandler; }
#endif	// defined(ASURA_MULTIPLE_GI_HANDLER)

		// The total SH for use in rendering - the sum of static and dynamic contributions 
		
		static Asura_SphericalHarmonics_Order2_RGB s_xSHForCurrentRender;

		// Prelighting SH
		static Asura_SphericalHarmonics_Order2_RGB s_xStaticSHForCurrentRender;

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		static Asura_SphericalHarmonics_Order2_RGB s_xDynamicSHForCurrentRender;
#endif
	private:

		friend class Asura_Chunk_GI_Solution;
		friend class Asura_GI_Solution_Handler;

#ifdef ASURA_MULTIPLE_GI_HANDLER
		static void Reset_Callback(Asura_Hash_ID uHash, class Asura_GI_Solution_Handler* pxData);

		static Asura_GI_Solution_Handler*							s_pxCurrentHandler;		// Currently active handler.
		static Asura_TemplateHashTree<Asura_GI_Solution_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.
#else //#ifdef ASURA_MULTIPLE_GI_HANDLER
		static Asura_GI_Solution_Handler s_xHandler;				// Single static-instance of a zone.
#endif //#else //#ifdef ASURA_MULTIPLE_GI_HANDLER

#ifdef ASURA_TOOLS
		static bool	s_bDisableStaticLighting;
		static bool	s_bSamplesValid;
#endif
};


class Asura_GI_IrradianceSample
{	
	// KJFL - 2004-07-12 - Currently 24 bytes, which isn't bad considering.
	//
	// However, if memory becomes an issue, some possible ways of reducing structure size further:
	//
	// 1. Store positions as shorts rather than floats, for a saving of 6 bytes, reducing structure size to 18 bytes
	//    (slightly awkward number cache-wise, & would involve speed hit probably).
	//    
	// 2. Dimension reduction on coefficients e.g. principal component analysis. Save a byte per dimension cut.
	//    (Bit more arithmetic to uncompress sample, but only linear combinations; search may cache better.)
	//
	// Obviously the above would need a little extra (fixed) memory overhead & (potentially) a drop in code speed.
	// However, implementing positions as shorts and dropping two dimensions would result in a structure size of 16
	// bytes, which would be extremely cache friendly as well as memory efficient. 
	//
	// AAM - 2009-02-09 - Removed compression on next-gen. Up to 60 bytes per sample

	private:
		friend class Asura_GI_Solution;
		friend class Asura_GI_Solution_Handler;
		friend class Asura_Chunk_GI_Solution;

		u_int GetPlaneAxis( void ) const
		{
			return m_uPlaneSplitFlag & 3;
		}

		
		// Using a union so that I can access the y-coord
		// as a u_int. I'm going to lose a little accuracy
		// by storing the plane split flag in the two least
		// significant bits.
		union
		{
			float			m_afPosition[3];

			struct
			{
				u_int		m_uDummyValueX;
				u_int		m_uPlaneSplitFlag;
				u_int		m_uDummyValueZ;
			};
		};

#ifdef ASURA_TARGET_SM3
		float				m_afCoefficients[12];
#else
		u_char				m_aucCompressedCoefficients[12];
#endif
};



#endif//__ASURA_GI_SOLUTION_H__
