//
// Asura_Light.cpp
//
// Created by Kevin Lea - 12/06/2001 10:56:01
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Light/Asura_Light.h"

#include "Core/Asura_AABB_Tree.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Light/Asura_Photon.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Module.h"

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "Lighting/Tools_PhotonMapping_System.h"
#include "Lighting/Tools_EnvironmentalLightTransport.h"
#include "Misc/Tools_Render_Options.h"
#endif 

int Asura_Lighting_System::NumberOfActiveLights;
Asura_Light * Asura_Lighting_System::LightStorage = NULL;
float Asura_Lighting_System::AmbientR;
float Asura_Lighting_System::AmbientG;
float Asura_Lighting_System::AmbientB;

float Asura_Lighting_System::s_fBottomAmbientR;
float Asura_Lighting_System::s_fBottomAmbientG;
float Asura_Lighting_System::s_fBottomAmbientB;

float Asura_Lighting_System::s_fEntityAmbientR;
float Asura_Lighting_System::s_fEntityAmbientG;
float Asura_Lighting_System::s_fEntityAmbientB;
unsigned Asura_Lighting_System::s_uEntityAmbientRGB; // Precalculated RGB value.


bool Asura_Lighting_System::s_bUseHemisphereAmbient;

#ifdef ASURA_TOOLS
Asura_Light_Visibility * Asura_Lighting_System::Visibility = NULL;

int Asura_Lighting_System::NumberOfLocalLights;
Asura_Light** Asura_Lighting_System::LocalLights = NULL;

bool Asura_Lighting_System::ShowLightRadius;
bool Asura_Lighting_System::LightsHaveChanged;
bool Asura_Lighting_System::ForceLightingUpdate;


float Asura_Lighting_System::SkyRed;
float Asura_Lighting_System::SkyGreen;
float Asura_Lighting_System::SkyBlue;
bool Asura_Lighting_System::s_bRenderUsingEntityLights;
void (*Asura_Lighting_System::s_pfnLightInfluenceCallback)(int,float,float,float) = 0;

static const float g_fLightScaleForIgnoreNormal = 0.5f;

float Asura_Lighting_System::s_afPrelightScale[ASURA_NUM_LIGHT_SCALE_VALUES];
float Asura_Lighting_System::s_afRuntimeLightScale[ASURA_NUM_LIGHT_SCALE_VALUES];

#endif//#ifdef ASURA_TOOLS


void Asura_Light::SetToDefault()
{
	Position.Zero();
	Direction.x = 0.0f;
	Direction.y = 0.0f;
	Direction.z = 1.0f;
	R = 255.0f;
	G = 255.0f;
	B = 255.0f;
	Brightness = 1.0f;
	Range = 5.0f;
	m_fInnerRange = 0.0f;
	Angle = 360.0f;
	ShadowStrength = 0.0f;
	m_xBoundingBox.MinX = 0.0f;
	m_xBoundingBox.MinY = 0.0f;
	m_xBoundingBox.MinZ = 0.0f;
	m_xBoundingBox.MaxX = 0.0f;
	m_xBoundingBox.MaxY = 0.0f;
	m_xBoundingBox.MaxZ = 0.0f;
	m_uFlags = 0;
	AffectsEnvironment = true;

	BrightnessOverRange = 0.2f;
	OldPosition = Position;
	OldRange = Range;
	HasChanged = false;
	
}


void Asura_Lighting_System::Initialise(void)
{
    // no need to allocate anything here in non-Tools builds. LightStorage is set up in Asura_Chunk_Light process

#ifdef ASURA_TOOLS
	LightStorage = new Asura_Light[MAX_NO_OF_LIGHTS];
    Visibility = new Asura_Light_Visibility[MAX_NO_OF_LIGHTS];
    LocalLights = new Asura_Light*[MAX_NO_OF_LOCAL_LIGHTS];

    Asura_Assert(LocalLights && Visibility && LightStorage, "Lights init failed");

	s_bRenderUsingEntityLights = false;
	NumberOfActiveLights = 0;
	ForceLightingUpdate = true;

	AmbientR = 20.0f;
	AmbientG = 20.0f;
	AmbientB = 20.0f;

	s_fBottomAmbientR = 0.0f;
	s_fBottomAmbientG = 0.0f;
	s_fBottomAmbientB = 0.0f;

	s_fEntityAmbientR = 0.0f;
	s_fEntityAmbientG = 0.0f;
	s_fEntityAmbientB = 0.0f;
	s_uEntityAmbientRGB = 0;

	s_bUseHemisphereAmbient = false;
	
	for(int i=0;i<ASURA_NUM_LIGHT_SCALE_VALUES;i++)
	{
		s_afPrelightScale[i] = 1.0f;
		s_afRuntimeLightScale[i] = 1.0f;
	}

#endif//#ifdef ASURA_TOOLS

	Asura_Photon_System::Initialise();
}

void Asura_Lighting_System::Reset(void)
{
	int i = NumberOfActiveLights;

	while(i--)
	{
		DeallocateLight(LightStorage);
	}
	NumberOfActiveLights = 0;

#ifdef ASURA_TOOLS
	s_bRenderUsingEntityLights = false;
	ForceLightingUpdate = true;
#endif//#ifdef ASURA_TOOLS

	AmbientR = 20.0f;
	AmbientG = 20.0f;
	AmbientB = 20.0f;
	
	s_fBottomAmbientR = 0.0f;
	s_fBottomAmbientG = 0.0f;
	s_fBottomAmbientB = 0.0f;

	s_fEntityAmbientR = 0.0f;
	s_fEntityAmbientG = 0.0f;
	s_fEntityAmbientB = 0.0f;
	s_uEntityAmbientRGB = 0;

	s_bUseHemisphereAmbient = false;

	// (MH - 15.12.2002) this should delete all static photons created with the old lights
    MakeStaticPhotonsList();

	Asura_Photon_System::Reset();

#ifndef ASURA_TOOLS
	// (MH - 02.05.2004) in non-tools builds light storage is allocated when Light Chunk is loaded.
	// So destroy them here, and next time we'll reallocate.
	if (LightStorage)
	{
		delete [] LightStorage;
		LightStorage = NULL;
	}
#endif

}

void Asura_Lighting_System::Shutdown(void)
{
	int i = NumberOfActiveLights;

	while(i--)
	{
		DeallocateLight(LightStorage);
	}
	NumberOfActiveLights = 0;

	Asura_Photon_System::Shutdown();

#ifdef ASURA_TOOLS
    delete [] LocalLights;
    LocalLights = NULL;
    delete [] Visibility;
    Visibility = NULL;
#endif//#ifdef ASURA_TOOLS
    
	if (LightStorage)
	{
		delete [] LightStorage;
		LightStorage = NULL;
	}
}

// (MH - 09.01.2003) hmm, this shouldn't be in independent code
#ifdef ASURA_TOOLS
void Asura_Lighting_System::Update(void)
{
	for (int i=0; i<NumberOfActiveLights; i++)
	{
		LightStorage[i].HasChanged = false;
		
		if(!Visibility[i].DataValid) Visibility[i].Sort();
		Visibility[i].DataValid = true;
	}
	ForceLightingUpdate = false;
}
#endif//#ifdef ASURA_TOOLS

Asura_Light* Asura_Lighting_System::AllocateLight(void)
{
	Asura_Light *lightPtr = 0; /* Default to null ptr */

	if (NumberOfActiveLights != MAX_NO_OF_LIGHTS) 
	{
		lightPtr = &LightStorage[NumberOfActiveLights];
#ifdef ASURA_TOOLS
		Visibility[NumberOfActiveLights].Indices = 0;
		Visibility[NumberOfActiveLights].MakeVoid();
#endif//#ifdef ASURA_TOOLS
		NumberOfActiveLights++;
	}
	else
	{
		/* unable to allocate a Light */
		Asura_Assert(0,"Maximum number of prelights reached");
	}
	return lightPtr;
}

void Asura_Lighting_System::DeallocateLight(Asura_Light *lightPtr)
{
	Asura_Assert(lightPtr, "NULL light ptr");
	if (!lightPtr) return;
	
	/*
	The tools light list in Tools_Mode_Lighting assumes that the lights will get deallocated in
	this way (the last light being copied into the deleted light's slot). 
	So if this gets changed then Tools_Mode_Lighting::DeleteLight will need to change as well.
	-Richard.
	*/
	NumberOfActiveLights--;
	*lightPtr = LightStorage[NumberOfActiveLights];
#ifdef ASURA_TOOLS
	int n = lightPtr-LightStorage;
	Visibility[n].MakeVoid();
	Visibility[n] = Visibility[NumberOfActiveLights];


	ForceLightingUpdate = true;
#endif//#ifdef ASURA_TOOLS
}



// (MH - 09.01.2003) hmm, this shouldn't be in independent code
// KJL - 03/03/2003 18:48:05 - changed this to Tools rather than not-PS2
#ifdef ASURA_TOOLS
bool Asura_Lighting_System::MakeLightListForObject(const Asura_Bounding_Box& boundingBox)
{
	NumberOfLocalLights=0;

	Asura_Vector_3 centre = boundingBox.CalculateCentre();

	LightsHaveChanged = ForceLightingUpdate;

	for (int i=0; i<NumberOfActiveLights; i++)
	{
		if(LightStorage[i].InfiniteLight)
		{
			if (LightStorage[i].HasChanged) LightsHaveChanged = true;

  			LocalLights[NumberOfLocalLights] = &LightStorage[i];

 			NumberOfLocalLights++;
		}
		else
		{
			bool intersects = boundingBox.IntersectsSphere(LightStorage[i].Position,LightStorage[i].Range);
			if ((intersects) || (boundingBox.IntersectsSphere(LightStorage[i].OldPosition,LightStorage[i].OldRange)) )
			{
				if (LightStorage[i].HasChanged) LightsHaveChanged = true;

				if (intersects)
				{
					LocalLights[NumberOfLocalLights] = &LightStorage[i];
					NumberOfLocalLights++;
				}
			}
		}
		if (NumberOfLocalLights==MAX_NO_OF_LOCAL_LIGHTS)
		{
		//	Asura_Assert(NumberOfLocalLights<MAX_NO_OF_LOCAL_LIGHTS,"Too many lights near a module");
			break;
		}
	}

	return LightsHaveChanged;
}

static bool VolumetricLightLOSTestPolygonFilter(const Asura_Mesh* pxMesh,int iPolyIndex)
{
	//For allowing LOS tests for volumetric lights to pass through polys with the dont block light flag
	//RAR 12/09/02
	Asura_Assert(pxMesh,"");

	if(!pxMesh->PolygonFlagDataPtr) return true;
	if(pxMesh->PolygonFlagDataPtr[iPolyIndex].Flags & AsuraPolygonFlag_DontBlockLight) return false;
	return true;

}




void Asura_Lighting_System::CalculatePreLightValue(const Asura_Vector_3& position,const Asura_Vector_3& xPositionForLOS, const Asura_Vector_3& normal, unsigned int& colour,int iPolygonFlags)
{
	if(s_bRenderUsingEntityLights)
	{
		CalculatePreLightValueForEntityLights(position,normal,colour,-1,-1,iPolygonFlags);
	}
	else
	{
		CalculatePreLightValue(position,xPositionForLOS,normal,colour,-1,-1,iPolygonFlags);
	}
}

void Asura_Lighting_System::CalculatePreLightValue(const Asura_Vector_3& position,const Asura_Vector_3& xPositionForLOS ,const Asura_Vector_3& normal, unsigned int& colour, int objectIndex, int vertexIndex, int iPolygonFlags)
{

	if(iPolygonFlags & AsuraPolygonFlag_FullBright)
	{
		 colour = 0x808080;
		 return;
	}
	
	
	int index = -1;
	if(objectIndex>=0 && vertexIndex>=0)
	{
		index = Asura_Light_Visibility::GetVisibilityIndex(objectIndex,vertexIndex);
	}


	float r,g,b;
	float fRedFromThisLight,fGreenFromThisLight,fBlueFromThisLight;

	if(iPolygonFlags & AsuraPolygonFlag_IgnoreAmbientLight)
	{
		r = 0;
		g = 0;
		b = 0;
	}
	else if (s_bUseHemisphereAmbient)
	{
		float fLerp = (normal.y + 1.0f) / 2.0f;

		r = (AmbientR - s_fBottomAmbientR)*fLerp + s_fBottomAmbientR;
		g = (AmbientG - s_fBottomAmbientG)*fLerp + s_fBottomAmbientG;
		b = (AmbientB - s_fBottomAmbientB)*fLerp + s_fBottomAmbientB;
	}
	else
	{
		r = (AmbientR); 
		g = (AmbientG); 
		b = (AmbientB); 
	}

	bool bIgnoreNormal = (iPolygonFlags & AsuraPolygonFlag_IgnoreNormalForLighting) ? true : false;
	
	for (int i=0; i<NumberOfLocalLights; i++)
	{
		Asura_Light* lightPtr = LocalLights[i];
		bool bLightAffectsPoint = false;

		if (!lightPtr->AffectsEnvironment) continue;

		//check against the light's bounding box if it has one
		if(lightPtr->UseBoundingBox)
		{
			if(!lightPtr->m_xBoundingBox.Contains(position)) continue;			
		}
	
		float range = lightPtr->Range;

		Asura_Vector_3 ltov = position-lightPtr->Position;
		float d=ltov.Magnitude();
		
		if (range > 0.0f && d<range || lightPtr->InfiniteLight)
		{

			float shadowAmbience = 1.0f;
			
			if (lightPtr->IsVolumetric)
			{
				bool lit = false;

				int n = lightPtr - LightStorage;
				
				if(index>=0)
				{
				#if 01
					if (Visibility[n].NumberOfVisibleVertices)
					{
						int max = Visibility[n].NumberOfVisibleVertices;
						int min = 0;
						int current;
						
						current = (min+max)/2;
						
						do
						{
							if (Visibility[n].Indices[current] == index)
							{
								lit = true;
								break;
							}
							if (Visibility[n].Indices[current]>index)
							{
								max = current;
							}
							else
							{
								min = current;
							}
							current = (min+max)/2;
							if (Visibility[n].Indices[current] == index)
							{
								lit = true;
								break;
							}
						}
						while(current!=min);
					}
				#else
					for(int c=0; c<Visibility[n].NumberOfVisibleVertices; c++)
					if (Visibility[n].Indices[c] ==index)
					{
						lit = true;
						break;
					}
				#endif
				}
				else
				{
					//don't have visibility data for this position , so need to do the LOS checks here
					
					Asura_Vector_3 alpha;
					if(lightPtr->InfiniteLight)
					{
						//light comes from a long way off back in the direction of the light
						alpha = xPositionForLOS - lightPtr->Direction*10000.0f;
					}
					else
					{
						//light comes from the lights position
						alpha = lightPtr->Position;
					}
					Asura_Vector_3 beta = xPositionForLOS;
						
					if(Asura_Module_System::IsPointVisible(alpha,beta,VolumetricLightLOSTestPolygonFilter,true))
					{
						lit = true;
					}
					

				}

				if (!lit) shadowAmbience = lightPtr->ShadowStrength;
			}
						

			if(lightPtr->InfiniteLight)
			{
				float dot=0.0f;
				float idot = lightPtr->Brightness;
				if(bIgnoreNormal)
				{
					idot *= g_fLightScaleForIgnoreNormal;
				}
				else
				{
					idot *= -(lightPtr->Direction*normal);
				}

					
				if (idot>0) dot += idot;
				dot *= shadowAmbience;			

				fRedFromThisLight=dot*lightPtr->R;
				fGreenFromThisLight=dot*lightPtr->G;
				fBlueFromThisLight=dot*lightPtr->B;
				bLightAffectsPoint = true;
			}
			else
			{
				float fRangeScale = 1.0f;
				if(d>lightPtr->m_fInnerRange)
				{
					fRangeScale = 1.0f - ((d-lightPtr->m_fInnerRange)/(range-lightPtr->m_fInnerRange));
				}

				if (lightPtr->Angle!=360.0f)
				{
					
					float fCosHalfAngle = Asura_Maths::Cosine((Asura_Maths::Pi/360.0f)*lightPtr->Angle);
					if(fCosHalfAngle<1.0f)
					{
						float dotangle = (ltov*lightPtr->Direction/d - fCosHalfAngle);
					
						if (dotangle>0)
						{
							//scale value so that light isn't attentuated when straight ahead
							dotangle /= (1.0f-fCosHalfAngle);
							
							float dot=0.0f;
							float idot = (lightPtr->Brightness*(fRangeScale)) * dotangle;
							if(bIgnoreNormal)
							{
								idot *= g_fLightScaleForIgnoreNormal;
							}
							else
							{
								idot *= -ltov*normal/d;
							}

							
							if (idot>0) dot += idot;
							dot *= shadowAmbience;			

							fRedFromThisLight=dot*lightPtr->R;
							fGreenFromThisLight=dot*lightPtr->G;
							fBlueFromThisLight=dot*lightPtr->B;
							bLightAffectsPoint = true;
						}
					}
				}
				else
				{
					{
						float dot=0.0f;
						float idot = lightPtr->Brightness*(fRangeScale);
						if(bIgnoreNormal)
						{
							idot *= g_fLightScaleForIgnoreNormal;
						}
						else
						{
							idot *= -ltov*normal/d;
						}


						
						if (idot>0) dot += idot;
						dot *= shadowAmbience;			

						fRedFromThisLight=dot*lightPtr->R;
						fGreenFromThisLight=dot*lightPtr->G;
						fBlueFromThisLight=dot*lightPtr->B;
						bLightAffectsPoint = true;
					}
				}
			}
		}

		if(bLightAffectsPoint)
		{
			if(s_pfnLightInfluenceCallback) s_pfnLightInfluenceCallback(lightPtr-LightStorage,fRedFromThisLight,fGreenFromThisLight,fBlueFromThisLight);

			r += fRedFromThisLight;
			g += fGreenFromThisLight;
			b += fBlueFromThisLight;
		}
		
	}

	/// Kevin Lea - 19/02/2003 14:24:06 - divide by two to allow room for 
	/// over-saturate effects
	r/=2.0f;
	g/=2.0f;
	b/=2.0f;


	float max = 255.0f;

	if (r>max) r=max;
	if (g>max) g=max;
	if (b>max) b=max;

	colour = (Asura_Maths::FloatToInt(r)<<16) + (Asura_Maths::FloatToInt(g)<<8) + Asura_Maths::FloatToInt(b);		

}


void Asura_Lighting_System::CalculatePreLightValueForEntityLights(const Asura_Vector_3& position, const Asura_Vector_3& normal, unsigned int& colour, int objectIndex, int vertexIndex, int iPolygonFlags)
{
	// Infinite light and volumetric flags are ignored.
	// Light angle is assumed to be 360.
	// If these rules change, then this code must be changed accordingly.aw.
	// (AW - 26.02.2004) - Now have properly set ambient light on entities.
	float r = s_fEntityAmbientR * 255.0f;
	float g = s_fEntityAmbientG * 255.0f;
	float b = s_fEntityAmbientB * 255.0f;
	
	float fShadowVolumeScale = 1.0f;

	for (int i=0; i<NumberOfLocalLights; i++)
	{
		Asura_Light* lightPtr = LocalLights[i];

		if (!lightPtr->AffectsEntities) continue;
		
		//check against the light's bounding box if it has one
		if(lightPtr->UseBoundingBox)
		{
			if(!lightPtr->m_xBoundingBox.Contains(position)) continue;			
		}
	
		float range = lightPtr->Range;

		Asura_Vector_3 ltov = position-lightPtr->Position;
		float d=ltov.Magnitude();
		
		if (range > 0.0f && d<range)
		{
			if (lightPtr->m_bIsShadowVolume)
			{
				float fShadow = 1.0f - lightPtr->ShadowStrength * ((range-d)/range);

				if (fShadow < 0.0f) fShadow = 0.0f;

				if (fShadow < fShadowVolumeScale) fShadowVolumeScale = fShadow;
			}
			else
			{
				float oorange = 1.0f/lightPtr->Range;
				float dot=0.0f;
				float idot = lightPtr->Brightness*((range-d)*oorange);
				if(iPolygonFlags & AsuraPolygonFlag_IgnoreNormalForLighting)
				{
					idot *= g_fLightScaleForIgnoreNormal;
				}
				else
				{
					idot *= -ltov*normal/d;
				}
			
				if (idot>0) dot += idot;
				r+=dot*lightPtr->R;
				g+=dot*lightPtr->G;
				b+=dot*lightPtr->B;

				if(s_pfnLightInfluenceCallback) s_pfnLightInfluenceCallback(lightPtr-LightStorage,dot*lightPtr->R,dot*lightPtr->G,dot*lightPtr->B);
			}
		}
	}
	
	r *= fShadowVolumeScale;
	g *= fShadowVolumeScale;
	b *= fShadowVolumeScale;

	/// Kevin Lea - 19/02/2003 14:24:06 - divide by two to allow room for 
	/// over-saturate effects
	r/=2.0f;
	g/=2.0f;
	b/=2.0f;


	float max = 255.0f;
	if (r>max) r=max;
	if (g>max) g=max;
	if (b>max) b=max;
	colour = (Asura_Maths::FloatToInt(r)<<16) + (Asura_Maths::FloatToInt(g)<<8) + Asura_Maths::FloatToInt(b);		
}



static int /*__cdecl*/ IntSortSub(const void *cmp1, const void *cmp2)
{
    if (*(const int*)cmp1 > *(const int*)cmp2)
        return  1;
    if (*(const int*)cmp1 < *(const int*)cmp2)
        return -1;

    return 0;
}
void Asura_Light_Visibility::Sort(void)
{
	if (NumberOfVisibleVertices && Indices)
	{
		qsort(Indices, NumberOfVisibleVertices, sizeof(int), IntSortSub);
	}
}

void Asura_Light_Visibility::IncreaseAllocation(const Asura_Light* light)
{
	Asura_Assert(light,"");
	int iOldNumberAllocated = NumberAllocated;
	
	if(light && light->InfiniteLight)
	{
		int vertexCount = 0;
		//to save large amounts of reallocation when dealing with infinite light , allocate enough storage
		//for every vertex
		for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
		{
			Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(i);
			Asura_Assert(mesh,"");

			vertexCount += mesh->NumberOfVertices;
		}
		if(vertexCount>NumberAllocated)
		{
			NumberAllocated = vertexCount; 
		}
		else
		{
			Asura_Assert(0,"Allocated visibility data for more vertices than exist in the environment");
			NumberAllocated += 500; 
		}
	}
	else
	{
		NumberAllocated+=500;
	}

	//Indices = (int*)realloc(Indices,NumberAllocated*sizeof(int));
	int* piOldIndices = Indices;
	Indices = (int*)Asura_Allocate(NumberAllocated*sizeof(int));
 	Asura_Assert(Indices,"Allocation failed");

	Asura_Mem::Copy(Indices,piOldIndices,iOldNumberAllocated*sizeof(int));
	Asura_Deallocate(piOldIndices);

}




void Asura_Lighting_System::CalculatePrelighting(Asura_Mesh& mesh, const Asura_Vector_3& position,int iModuleIndex)
{
	if (!mesh.BoundingBoxDataValid) mesh.CalculateBoundingBox();

	Asura_Bounding_Box bb;
	bb = mesh.BoundingBox;
	bb.Displace(position);


	if (!mesh.PrelightingDataValid && mesh.PrelightingDataPtr==0)
	{
		mesh.PrelightingDataPtr = new Asura_Mesh_Prelighting[mesh.NumberOfPolygons];
	}

	if (!mesh.AreNormalsValid()) mesh.CreateNormals();

#ifdef ASURA_TOOLS
	if (!(Tools_System::IsInPhotonMappingMode() || Tools_Render_Options::IsUsingGlobalIllumination()))
#endif //#ifdef ASURA_TOOLS
	{

		Asura_Lighting_System::MakeLightListForObject(bb);
		if (s_bRenderUsingEntityLights)
		{
			for (int i=0; i<mesh.NumberOfPolygons; i++)
			{
				int polygonFlags = 0;
				if(mesh.PolygonFlagDataValid) polygonFlags = mesh.PolygonFlagDataPtr[i].Flags;
				for(int v=0; v<mesh.PolygonDataPtr[i].NumberOfVertices(); v++)
				{
					if(polygonFlags & AsuraPolygonFlag_FullBright)
					{
						mesh.PrelightingDataPtr[i].Colour[v] = 0x808080;
					}
					else
					{
						int vID = mesh.PolygonDataPtr[i].VertexID[v];
						Asura_Vector_3 p = mesh.VertexDataPtr[vID].Position+position;
						Asura_Vector_3& n = mesh.NormalsDataPtr[i].VertexNormal[v];
						Asura_Lighting_System::CalculatePreLightValueForEntityLights(p,n,mesh.PrelightingDataPtr[i].Colour[v],mesh.Index,vID,polygonFlags);
					}
				}
			}
		}
		else
		{
			//See if visibility data needs to be updated for any lights
			for (int i=0; i<NumberOfLocalLights; i++)
			{
				Asura_Light* lightPtr = LocalLights[i];
				int n = lightPtr - LightStorage;

				if(Visibility[n].DataValid) continue;
				if(!lightPtr->IsVolumetric) continue;

				//determine visibility for each vertex in this mesh
				float range = lightPtr->Range;
				for(int v=0;v<mesh.NumberOfVertices;v++)
				{
					Asura_Vector_3 p = mesh.VertexDataPtr[v].Position+position;
					Asura_Vector_3 ltov = p-lightPtr->Position;
					//in range?
					if (lightPtr->InfiniteLight || ltov.Magnitude()<range)
					{
						Asura_Vector_3 alpha;
						if(lightPtr->InfiniteLight)
						{
							//light comes from a long way off back in the direction of the light
							alpha = p - lightPtr->Direction*10000.0f;
						}
						else
						{
							//light comes from the lights position
							alpha = lightPtr->Position;
						}
						Asura_Vector_3 beta = p;
						
						if(Asura_Module_System::IsPointVisible(alpha,beta,VolumetricLightLOSTestPolygonFilter,true))
						{
							int index = Asura_Light_Visibility::GetVisibilityIndex(iModuleIndex,v);
							Visibility[n].AddIndex(index,lightPtr);
						}
					}
				}
				Visibility[n].Sort();
			}
			
			for (int i=0; i<mesh.NumberOfPolygons; i++)
			{
				int polygonFlags = 0;
				if(mesh.PolygonFlagDataValid) polygonFlags = mesh.PolygonFlagDataPtr[i].Flags;
				for(int v=0; v<mesh.PolygonDataPtr[i].NumberOfVertices(); v++)
				{
 					int vID = mesh.PolygonDataPtr[i].VertexID[v];
 					Asura_Vector_3 p = mesh.VertexDataPtr[vID].Position+position;
 					Asura_Vector_3& n = mesh.NormalsDataPtr[i].VertexNormal[v];

 					Asura_Lighting_System::CalculatePreLightValue(p,p,n,mesh.PrelightingDataPtr[i].Colour[v],iModuleIndex,vID,polygonFlags);
				}
			}
		}
	}
#ifdef ASURA_TOOLS 
	else
	{
		#if !defined(ASURA_TARGET_SM3 )
		Tools_PhotonMapping_System::CheckValidityOfGlobalIllumination();
		
		for (int i=0; i<mesh.NumberOfPolygons; i++)
		{
			int polygonFlags = 0;
			if(mesh.PolygonFlagDataValid) polygonFlags = mesh.PolygonFlagDataPtr[i].Flags;
			for(int v=0; v<mesh.PolygonDataPtr[i].NumberOfVertices(); v++)
			{
 				int vID = mesh.PolygonDataPtr[i].VertexID[v];
 				Asura_Vector_3 p = mesh.VertexDataPtr[vID].Position+position;
 				Asura_Vector_3& n = mesh.NormalsDataPtr[i].VertexNormal[v];

				const u_int uPaintedColour = mesh.m_pxPaintedColourDataPtr[ i ].Colour[ v ];

				//Tools_PhotonMapping_System::SamplePhotonMap( p, n, mesh.PrelightingDataPtr[i].Colour[v] );

				if(polygonFlags & AsuraPolygonFlag_FullBright)
				{
					mesh.PrelightingDataPtr[i].Colour[v] = 0x808080;
				}
				else if(polygonFlags & AsuraPolygonFlag_IgnoreNormalForLighting)
				{
					mesh.PrelightingDataPtr[i].Colour[v] = Tools_EnvironmentalLightTransport::GetDiffuseForVertex(p, Asura_Maths::ZeroVector3, i, v, iModuleIndex, uPaintedColour );
				}
				else
				{
					mesh.PrelightingDataPtr[i].Colour[v] = Tools_EnvironmentalLightTransport::GetDiffuseForVertex(p, n, i, v, iModuleIndex, uPaintedColour );
				}
			}
		}
		#else
		//This prelighting data isn't really used with the new renderers. However , the colour values still
		//get propagated down through several systems before being discarded.
		//Initialising to zero here to avoid problems from doing comparisons on junk values later.
		//RAR 04/05/2006
		memset(mesh.PrelightingDataPtr,0,sizeof(Asura_Mesh_Prelighting)*mesh.NumberOfPolygons);
		#endif
	}

#endif //#ifdef ASURA_TOOLS

	mesh.PrelightingDataValid = true;
}
#endif//#ifdef ASURA_TOOLS

void Asura_Lighting_System::MakeStaticPhotonsList(void)
{
	/// Kevin Lea - 26/09/2002 11:20:31 - this is used by the tools/exporting

	/// Kevin Lea - 26/09/2002 14:58:26 - wipe any previous static photons
	delete [] Asura_Photon_System::s_pxStaticPhotonStorage;
	delete Asura_Photon_System::s_pxStaticPhotonTree;

	/// Kevin Lea - 26/09/2002 14:59:04 - scan through light list and count 'Affects Entities' lights
	u_int uNumberOfStaticPhotons = 0;
	for (int i=0; i<NumberOfActiveLights; i++)
	{
		Asura_Light& xLight = LightStorage[i];
		
		if (xLight.AffectsEntities && xLight.Range > 0.0f)
		{
			uNumberOfStaticPhotons++;
		}
	}

	Asura_Photon_System::s_uNumberOfStaticPhotons = uNumberOfStaticPhotons;

	if ( uNumberOfStaticPhotons == 0 )
	{
		Asura_Photon_System::s_pxStaticPhotonStorage = 0;
		Asura_Photon_System::s_pxStaticPhotonTree = 0;

		return;
	}

	Asura_Photon_System::s_pxStaticPhotonStorage = new Asura_Photon[uNumberOfStaticPhotons];

	for (int i=0, j=0; i<NumberOfActiveLights; i++)
	{
		Asura_Light& xLight = LightStorage[i];

		if (xLight.AffectsEntities && xLight.Range > 0.0f)
		{
			Asura_Photon& xPhoton = Asura_Photon_System::s_pxStaticPhotonStorage[j];
			
			xPhoton.SetFlags(0);
			xPhoton.SetAsActive(true);
		
	

			if ( xLight.m_bIsShadowVolume )
			{
				xPhoton.SetColours( xLight.ShadowStrength, xLight.ShadowStrength, xLight.ShadowStrength );
			}
			else
			{
				/// Kevin Lea - 03/10/2002 16:23:33 - photons use a colour component that 
				/// goes between 0 to 1.0f (or possibly over that for saturation effects)
				/// whereas tools lights use 0 to 255 and a separate brightness field (so
				/// artists can select a colour in the same way as other art packages like
				/// photoshop.) 
				/// Kevin Lea - 19/02/2003 14:24:06 - divide by two to allow room for 
				/// over-saturate effects
				float fScale = (xLight.Brightness / 255.0f) / 2.0f;
				xPhoton.SetColours( xLight.R * fScale, xLight.G * fScale, xLight.B * fScale );
			}

			xPhoton.SetPosition(xLight.Position);
			xPhoton.SetRange(xLight.Range);

            xPhoton.SetUseBoundingBox( xLight.UseBoundingBox );
            if( xLight.UseBoundingBox )
            {
                xPhoton.SetBoundingBox( xLight.m_xBoundingBox );
            }
			
			xPhoton.SetShadowVolume( xLight.m_bIsShadowVolume );

			xPhoton.SetAsStatic( true );

			j++;
		}
	}

	Asura_Photon_System::s_pxStaticPhotonTree = Asura_AABB_Tree_Creator::Create( uNumberOfStaticPhotons, Asura_Photon_System::AABBTreeCreateCallback );
}

#ifdef ASURA_TOOLS
bool Asura_Lighting_System::IsUsingEntityLights()
{
	return s_bRenderUsingEntityLights;
}

void Asura_Lighting_System::ToggleUsingEntityLights()
{
	s_bRenderUsingEntityLights = !s_bRenderUsingEntityLights;
	for (int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh *pxMesh = Asura_Module_System::GetMeshForModule(i);
		pxMesh->PrelightingDataValid = false;
	}
}


#endif//#ifdef ASURA_TOOLS



#ifdef ASURA_TOOLS
void Asura_Lighting_System::DeleteModule(int iModuleIndex)
{
	for(int i=0;i<NumberOfActiveLights;i++)
	{
		Visibility[i].DeleteModule(iModuleIndex);
	}
}

void Asura_Light_Visibility::DeleteModule(int iModuleIndex)
{
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<iNumberOfModules,"");

	bool bChanged = false;
	for(int i=0;i<NumberOfVisibleVertices;i++)
	{
		int iModule,iVertex;
		GetModuleAndVertexIndex(Indices[i],iModule,iVertex);

		if(iModule == iModuleIndex)
		{
			//this visibilty belongs to the deleted module ,so delete it
			Indices[i--] = Indices[--NumberOfVisibleVertices];
			bChanged = true;
		}
		else if (iModule == iNumberOfModules-1)
		{
			//this visibility belongs to the module whose index willbe changed
			Indices[i] = GetVisibilityIndex(iModuleIndex,iVertex);
			bChanged = true;
		}
	}

	if(bChanged) Sort();
}
#endif //ASURA_TOOLS
