//
// Asura_Lighting_Solution.cpp
//
// Created by Andrew Weinkove - 24/08/04
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//
// Moved class (by Rich Rice) from Asura_Photon.cpp and updated to use SH.
//

#include "Core/Asura.h"
#include "Light/Asura_Lighting_Solution.h"

#include "Light/Asura_GI_Solution.h"

#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "Lighting/Tools_GlobalIllumination_EntitySamples.h"
#endif

Asura_Lighting_Solution::Asura_Lighting_Solution(bool bKeepStatic,bool bKeepDynamic)
{
	m_xLightingPosition.Zero();

	m_bKeepStatic = bKeepStatic;
	m_bKeepDynamic = bKeepDynamic;

	m_pxStoredStaticIrradiance = bKeepStatic ? new Asura_SphericalHarmonics_Order2_RGB : 0;
	m_pxStoredDynamicIrradiance = bKeepDynamic ? new Asura_SphericalHarmonics_Order2_RGB : 0;

	Reset();
}

Asura_Lighting_Solution::~Asura_Lighting_Solution()
{
	delete m_pxStoredStaticIrradiance;
	delete m_pxStoredDynamicIrradiance;
}

void Asura_Lighting_Solution::Reset()
{
	m_bStaticValid = false;
	m_bDynamicValid = false;
}

void Asura_Lighting_Solution::UseStoredLighting() const
{
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	if(m_bStaticValid)
	{
		Asura_GI_Solution::s_xStaticSHForCurrentRender = *m_pxStoredStaticIrradiance;
	}
	else
	{
		Asura_GI_Solution::s_xStaticSHForCurrentRender.Zero();
	}

	if(m_bDynamicValid)
	{
		Asura_GI_Solution::s_xDynamicSHForCurrentRender = *m_pxStoredDynamicIrradiance;
	}
	else
	{
		Asura_GI_Solution::s_xDynamicSHForCurrentRender.Zero();
	}
	#endif

	if(m_bStaticValid)
	{
		Asura_GI_Solution::s_xStaticSHForCurrentRender = *m_pxStoredStaticIrradiance;
		Asura_GI_Solution::s_xSHForCurrentRender = *m_pxStoredStaticIrradiance;
		if(m_bDynamicValid)
		{
			Asura_GI_Solution::s_xSHForCurrentRender += *m_pxStoredDynamicIrradiance;
		}
	}
	else if(m_bDynamicValid)
	{
		Asura_GI_Solution::s_xSHForCurrentRender = *m_pxStoredDynamicIrradiance;
	}
	else
	{
		Asura_GI_Solution::s_xSHForCurrentRender.Zero();
	}
}

void Asura_Lighting_Solution::SetPosition(const Asura_Vector_3& xPosition)
{
	if(m_bStaticValid || m_bDynamicValid)
	{
		//Only update the position if it has moved enough , so as to preserve existing
		//lighting calculations. RAR 03/10/2005
		
		const float fTolerance = 0.1f*0.1f;//10 cm
		if(xPosition.ApproximatelyEquals(m_xLightingPosition,fTolerance))
		{
			//existing position is close enough.
			return;
		}
		//changing position, so invalidate old lighting
		m_bStaticValid = false;
		m_bDynamicValid = false;
	}
	m_xLightingPosition = xPosition;
}

void Asura_Lighting_Solution::RecalculateLighting(  const bool bIncludeStaticLight,
													const bool bIncludeDynamicLight)
{
	#ifdef ASURA_TOOLS
	InvalidateIfToolsLightingChanged();
	#endif


	if (bIncludeStaticLight == (m_bStaticValid ? true : false) && 
		bIncludeDynamicLight == (m_bDynamicValid ? true : false))
	{
		//copy any partial solution into the global s_xSHForCurrentRender structure
		UseStoredLighting();
		return;
	}

	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	Asura_GI_Solution::s_xStaticSHForCurrentRender.Zero();
	Asura_GI_Solution::s_xDynamicSHForCurrentRender.Zero();
	#endif

	Asura_GI_Solution::s_xSHForCurrentRender.Zero();

	//calculate the lighting values that we need
	if(bIncludeStaticLight)
	{
		if(m_bStaticValid)
		{
			Asura_GI_Solution::s_xSHForCurrentRender += *m_pxStoredStaticIrradiance;

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			Asura_GI_Solution::s_xStaticSHForCurrentRender = *m_pxStoredStaticIrradiance;
			#endif
		}
		else
		{
			//need to calculate static irradiance then
			Asura_SphericalHarmonics_Order2_RGB xTemp;
			Asura_GI_Solution::GetStaticIrradiance(xTemp, m_xLightingPosition);

			//store if desired
			if(m_bKeepStatic)
			{
				*m_pxStoredStaticIrradiance = xTemp;
				m_bStaticValid = true;
			}

			//and add on to the total result
			Asura_GI_Solution::s_xSHForCurrentRender += xTemp;

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			Asura_GI_Solution::s_xStaticSHForCurrentRender = xTemp;
			#endif
		}
	}

	if(bIncludeDynamicLight)
	{
		if(m_bDynamicValid)
		{
			Asura_GI_Solution::s_xSHForCurrentRender += *m_pxStoredDynamicIrradiance;
			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			Asura_GI_Solution::s_xDynamicSHForCurrentRender = *m_pxStoredDynamicIrradiance;
			#endif
		}
		else
		{
			//need to calculate dynamic irradiance then
			Asura_SphericalHarmonics_Order2_RGB xTemp;
			Asura_GI_Solution::GetDynamicIrradiance(xTemp, m_xLightingPosition);

			//store if desired
			if(m_bKeepDynamic)
			{
				*m_pxStoredDynamicIrradiance = xTemp;
				m_bDynamicValid = true;
			}

			//and add on to the total result
			Asura_GI_Solution::s_xSHForCurrentRender += xTemp;

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			Asura_GI_Solution::s_xDynamicSHForCurrentRender = xTemp;
			#endif
		}
	}
}

#ifdef ASURA_TOOLS
bool Asura_Lighting_Solution::HasToolsLightingChanged() const
{
	return Tools_System::IsActive() && m_uGISamplesFrameStamp != Tools_GlobalIllumination_EntitySamples::GetFrameStamp();
}

void Asura_Lighting_Solution::InvalidateIfToolsLightingChanged()
{
	//if in tools , and the GI samples have been invalidated since the lighting solution
	//was last calculated ,we will need to reset
	if(HasToolsLightingChanged())
	{
		ResetStatic();
		m_uGISamplesFrameStamp = Tools_GlobalIllumination_EntitySamples::GetFrameStamp();
	}
}
#endif // ASURA_TOOLS
