/*************************************************************
 *
 * Description   : LoadAndSave functions for asura (project independent)
 *
 * File Name     : Asura_LoadAndSave.cpp
 *
 * Author        : Richard May
 *
 *************************************************************/

/**********************/
/* Glossary of terms. */
/**********************/


/************/
/* Includes.*/
/************/

// main asura header
#include "Core/Asura.h"

// local header
#include "LoadAndSave/Asura_LoadAndSave.h"

// asura
#if defined( ASURA_USE_BOTANICALS)
#include "Botanicals/Asura_Botanical_Game_Render_GameState.h"
#endif
#include "ChunkSystem/Asura_SaveBlock.h"
#ifdef ASURA_USE_ATTRACTORS
#include "ChunkSystem/Asura_Chunk_Attractors.h"
#endif
#ifdef ASURA_USE_STREAMING
#include "ChunkSystem/Asura_Chunk_CurrentStreamingZone.h"
#endif //#ifdef ASURA_USE_STREAMING
#ifdef ASURA_USE_DIALOGUE_SYSTEM
#include "ChunkSystem/Asura_Chunk_Dialogue.h"
#endif
#include "ChunkSystem/Asura_Chunk_DynamicMusic.h"
#include "ChunkSystem/Asura_Chunk_Entity.h"
#include "ChunkSystem/Asura_Chunk_MetaMusic.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Chunk_StreamingBackgroundSound.h"
#include "Debugging/Asura_Logfile.h"
#ifdef ASURA_LOGFILES_ENABLED
#include "Entity/Asura_Entity_Classification.h"
#endif
#include "Entity/Asura_Entity.h"
#include "Entity/Asura_StaticMessage_System.h"
#if defined(ASURA_USE_FSFX)
#include "FSFX/Asura_FSFX_System.h"
#endif	//defined(ASURA_USE_FSFX)
#include "Network/Asura_Network.h"
#ifdef ASURA_USE_DIALOGUE_SYSTEM
#include "Sound/Asura_Dialogue.h"
#include "Sound/Asura_Dialogue_System.h"
#endif
#if defined(ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM)
#include "Sound/Asura_Dynamic_Music_System.h"
#endif
#include "SpecialFX/Asura_Beam_System.h"
#include "SpecialFX/Asura_Decal_System.h"

#include "SpecialFX/pfx/Asura_PFX_System.h"

#ifdef ASURA_USE_STREAMING
#include "Streaming/Asura_Streaming_System.h"
#endif //#ifdef ASURA_USE_STREAMING


#include <string.h>

// C O N S T A N T S /////////////////////////////////////////////////////////

#ifdef ASURA_LOGFILES_ENABLED
static const char* ENTITY_SAVE_LOGFILE = "logfiles\\entity_save.txt";
static const char* SAVE_SUMMARY_LOGFILE = "logfiles\\save_summary.txt";
#endif

// G L O B A L S /////////////////////////////////////////////////////////////

DEBUGVAR bool dbg_bLogEntitySaveDetails = false;
DEBUGVAR bool dbg_bLogSaveSummary = false;

/*********************/
/* Public functions. */
/*********************/

bool Asura_LoadAndSave::s_bLoadingFromBuffer = false;

bool Asura_LoadAndSave::SaveToBuffer( Asura_SaveBlock_Base& xSaveBlock )
{
#ifdef ASURA_LOGFILES_ENABLED
	static u_short usCurrentClassification = AsuraEntityClass_NotSet;
	static u_int uClassificationCount = 0;
	static u_int uCumulativeSize = 0;
	static u_int uTotalEntitySize = 0;
	static u_int uSpaceUsed = 0;
    if( dbg_bLogEntitySaveDetails )
    {
        Asura_SimpleLogfileClear( ENTITY_SAVE_LOGFILE );
	}
	if(dbg_bLogSaveSummary)
	{
		Asura_SimpleLogfileClear( SAVE_SUMMARY_LOGFILE );
		Asura_SimpleLogfileOutput(( SAVE_SUMMARY_LOGFILE, "Class * Instances :  Bytes (name)\n"));
		usCurrentClassification = AsuraEntityClass_NotSet;
		uClassificationCount = 0;
		uCumulativeSize = 0;
		uTotalEntitySize = 0;
    }
#endif

	Project_AddHeader(xSaveBlock);

	#ifdef ASURA_TOOLS
	//save the next tools created guids
	for(int i=0; i<TOOLS_NUM_GUIDRANGE; i++)
	{
		Asura_Chunk_Entity_NextGuid::Create(i, xSaveBlock);
	}
	#endif

	//Save any delayed messages
	Asura_Chunk_StaticMessages_InGame::Create(xSaveBlock);
	
	// save all entities
	for( Asura_Classified_List_It<Asura_Entity_Server> xIterator(Asura_Entity_Server::GetList()); !xIterator.Done(); xIterator.Next() )
	{
		Asura_Entity_Server* pxEntity = xIterator.GetCurrent();
		
		// ignore entities marked for destruction
		if(pxEntity->IsDestroyed()) continue;

		// save this entity
		Asura_Chunk_Entity::Create( (Asura_Entity_Server*)pxEntity, xSaveBlock );
		
#if defined(ASURA_LOGFILES_ENABLED)
		const u_short usClassification = pxEntity->GetClassification();
		int iSize = 0;
		if(pxEntity->IsFilledFromChunkBuffer())
		{
			iSize = pxEntity->GetChunkBufferSize();
		}
		else
		{
			Asura_Chunk_Dummy_Stream xStream;
			pxEntity->WriteToChunkStream(xStream);
			iSize = xStream.GetPosition();
		}
		if(dbg_bLogSaveSummary)
		{
			if(usCurrentClassification != usClassification)
			{
				const char* szBuffer = "";
#if ASURA_NETWORK_LOGGING
				Asura_Network::GetEntityClassificationDesc(usCurrentClassification, &szBuffer);
#endif
				Asura_SimpleLogfileOutput(( SAVE_SUMMARY_LOGFILE, "%5u * %4u      : %6u (%s) \n", usCurrentClassification, uClassificationCount, uCumulativeSize, szBuffer ));

				usCurrentClassification = usClassification;
				uTotalEntitySize += uCumulativeSize;
				uCumulativeSize = 0;
				uClassificationCount = 0;
			}
			++uClassificationCount;
			uCumulativeSize += iSize;
		}
        if( dbg_bLogEntitySaveDetails )
        {
			const char* szBuffer = "";
#if ASURA_NETWORK_LOGGING
			Asura_Network::GetEntityClassificationDesc(usClassification, &szBuffer);
#endif
			Asura_SimpleLogfileOutput(( ENTITY_SAVE_LOGFILE, "Classification %s (%5hu) - size %d bytes\n", szBuffer, usClassification, iSize ));
        }
#endif	// defined(ASURA_LOGFILES_ENABLED)

		#ifdef ASURA_TOOLS
		//and its tools data
		Asura_Chunk_Entity_ToolsData::Create((Asura_Entity_Server*)pxEntity,xSaveBlock);
		#endif

		// JAT 22/02/07 streams can be single byte aligned because of things like bool's (which are stored as a byte) so not sure this assert is valid anymore
		// Asura_Assert(((long)xSaveBlock.GetCurrentPosition()&1) == 0,"odd address");
	}

#if defined(ASURA_LOGFILES_ENABLED)
	if( dbg_bLogSaveSummary )
	{
		const char* szBuffer = "";
#if ASURA_NETWORK_LOGGING
		Asura_Network::GetEntityClassificationDesc(usCurrentClassification, &szBuffer);
#endif
		Asura_SimpleLogfileOutput(( SAVE_SUMMARY_LOGFILE, "%5u * %4u      : %6u (%s) \n", usCurrentClassification, uClassificationCount, uCumulativeSize, szBuffer ));
		Asura_SimpleLogfileOutput(( SAVE_SUMMARY_LOGFILE, "Total size for all entities %d bytes\n", uTotalEntitySize));
		uSpaceUsed = xSaveBlock.GetSpaceUsed();
	}
#endif	// defined(ASURA_LOGFILES_ENABLED)
	//save meta music state
	#ifndef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
    Asura_Chunk_MetaMusic_Ingame_State::Create( xSaveBlock );
	#else
	Asura_Chunk_DynamicMusic_Ingame_State::Create( xSaveBlock );
	#endif
	//streaming background sounds
	Asura_Chunk_StreamingBackgroundSound_InGame::Create(&xSaveBlock);

	//dialogue system state
	#ifdef ASURA_USE_DIALOGUE_SYSTEM
	Asura_Chunk_Dialogue_InGame::Create( &xSaveBlock );
	#endif

	//attracot state
	#ifdef ASURA_USE_ATTRACTORS
	Asura_Chunk_Attractors_InGame::Create( &xSaveBlock );
	#endif

#if defined(ASURA_LOGFILES_ENABLED)
	if( dbg_bLogSaveSummary )
	{
		const u_int uDiff = xSaveBlock.GetSpaceUsed() - uSpaceUsed;
		Asura_SimpleLogfileOutput(( SAVE_SUMMARY_LOGFILE, "Total size for all misc chunks: %d bytes\n", uDiff));		
		uSpaceUsed = xSaveBlock.GetSpaceUsed();
	}
#endif	// defined(ASURA_LOGFILES_ENABLED)

	// now call project specific save
	if(!Project_SaveToBuffer(xSaveBlock)) return false;
	
#if defined(ASURA_LOGFILES_ENABLED)
	if( dbg_bLogSaveSummary )
	{
		const u_int uDiff = xSaveBlock.GetSpaceUsed() - uSpaceUsed;
		Asura_SimpleLogfileOutput(( SAVE_SUMMARY_LOGFILE, "Total size for all project chunks: %d bytes\n", uDiff));		
	}
#endif	// defined(ASURA_LOGFILES_ENABLED)

	// tidy up
	xSaveBlock.AddAsuraTerminator();

	return true;
}

bool Asura_LoadAndSave::LoadFromBuffer( const Asura_SaveBlock_Base& xSaveBlock, bool bResetBeforeLoad/*=true*/ )
{
    s_bLoadingFromBuffer = true;

	//Asura_Chunk_ChunkBuffer_Stream xStream(xSaveBlock.GetStartPtr());
	Asura_Chunk_SaveBlockReader_Stream xStream(&xSaveBlock);
	if(!ProcessAsuraFileHeader(xStream))
	{
		return false;
	}

	if( bResetBeforeLoad )
	{
		ResetBeforeLoad();
	}

	while(ProcessChunk(xStream)) {}

    s_bLoadingFromBuffer = false;
	return true;
}

bool Asura_LoadAndSave::ProcessAsuraFileHeader(Asura_Chunk_Stream& xStream)
{
	// make sure it's a valid Asura file
	char acFileID[ASURA_FILE_ID_NUM_CHARS];
	if( xStream.Read( ASURA_FILE_ID_NUM_CHARS, acFileID ) != ASURA_FILE_ID_NUM_CHARS || 
		strnicmp(acFileID, ASURA_UNCOMPRESSED_FILE_ID, ASURA_ALL_FILE_ID_ACTUAL_CHARS) )
	{
		Asura_Assert( false, "Not a valid Asura file." );
		s_bLoadingFromBuffer = false;
		return false;
	}
	return true;
}

bool Asura_LoadAndSave::ProcessChunk(Asura_Chunk_Stream& xStream)
{
	Asura_ChunkBuffer xChunkBuffer;
	char* pcChunkHeaderBuffer = reinterpret_cast<char*>(&xChunkBuffer);

#define CHUNK_PROCESS(chunkID, chunkClass)	\
	case (chunkID):	\
	{	\
		Asura_Assert( sizeof(chunkClass) <= uASURA_MAX_CHUNKHEADER_SIZE, "chunk header too big" ); \
		xStream.ReadChunk_GetNextHeader(pcChunkHeaderBuffer, sizeof(chunkClass)); \
		class chunkClass *chunk2 = (chunkClass *)pxChunkHeader;	\
		bChunkReadCompletely = chunk2->Process(xStream);	\
	}	\
	break

	if(xStream.ReadChunk_GetNextID(pcChunkHeaderBuffer))
	{
		// KOR - note that only the chunk ID is valid at this point
		Asura_Chunk_Header*	pxChunkHeader = reinterpret_cast<Asura_Chunk_Header*>( pcChunkHeaderBuffer );

		bool bChunkReadCompletely = false;
		switch(pxChunkHeader->ID)
		{
			CHUNK_PROCESS( ASURA_CHUNK_ENTITY,           Asura_Chunk_Entity           );
#ifdef ASURA_TOOLS
			CHUNK_PROCESS( ASURA_CHUNK_ENTITY_TOOLSDATA, Asura_Chunk_Entity_ToolsData );
			CHUNK_PROCESS( ASURA_CHUNK_ENTITY_NEXTGUID, Asura_Chunk_Entity_NextGuid );
#endif
#ifndef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
			CHUNK_PROCESS( ASURA_CHUNK_METAMUSIC_INGAME, Asura_Chunk_MetaMusic_Ingame_State );
#else
		CHUNK_PROCESS( ASURA_CHUNK_DYNAMICMUSIC_INGAME, Asura_Chunk_DynamicMusic_Ingame_State );
#endif
			CHUNK_PROCESS( ASURA_CHUNK_STREAMINGBACKGROUNDSOUND_INGAME, Asura_Chunk_StreamingBackgroundSound_InGame );

			CHUNK_PROCESS( ASURA_CHUNK_STATICMESSAGES_INGAME, Asura_Chunk_StaticMessages_InGame );
#ifdef ASURA_USE_DIALOGUE_SYSTEM
			CHUNK_PROCESS( ASURA_CHUNK_DIALOGUE_INGAME, Asura_Chunk_Dialogue_InGame );
#endif
#if defined( ASURA_USE_BOTANICALS)
#if defined( ASURA_BOTANICAL_SUPPORT_DYNAMIC_DATA) && 1 == ASURA_BOTANICAL_SUPPORT_DYNAMIC_DATA
			CHUNK_PROCESS( ASURA_CHUNK_BOTANICALS_INGAME, Asura_BGR_Chunk_GameState );
#endif // #if defined( ASURA_BOTANICAL_SUPPORT_DYNAMIC_DATA) && 1 == ASURA_BOTANICAL_SUPPORT_DYNAMIC_DATA
#endif
#ifdef ASURA_USE_ATTRACTORS
			CHUNK_PROCESS( ASURA_CHUNK_ATTRACTORS_INGAME, Asura_Chunk_Attractors_InGame );
#endif

		default: 
			{	
				// couldn't handle chunk here so call project specific loader
				if(Project_LoadFromBuffer(pxChunkHeader, xStream))
				{
					bChunkReadCompletely = true;
				}
				else
				{
					s_bLoadingFromBuffer = false;
					return false;
				}

				// successfully handled by the project load from buffer so continue
				break;
			}
		}

		if(!xStream.HasReadChunkHeader())
		{
			// KOR - read the rest of the header to get the chunk size
			xStream.ReadChunk_GetNextHeader(pcChunkHeaderBuffer, sizeof(Asura_Chunk_Header));
		}
		// KOR - finish off the current chunk. The stream performs verification and skipping if necessary
		xStream.ReadChunk_End(bChunkReadCompletely);
	}
	else
	{
		// Last chunk.
		return false;
	}
	return true;

#undef CHUNK_PROCESS
}

bool Asura_LoadAndSave::LoadFromDisk(const char* szFilename)
{
    ResetBeforeLoad();
    return LoadAsuraFile(szFilename);
}

#ifdef ASURA_USE_STREAMING
bool Asura_LoadAndSave::SaveMissionToBuffer( Asura_SaveBlock_Base& xSaveBlock )
{
	if (!Asura_Streaming_System::GetCurrentMission())
	{
		Asura_Assert(false, "No mission data to save");
		return false;
	}

	if( dbg_bLogEntitySaveDetails )
	{
		Asura_SimpleLogfileClear( ENTITY_SAVE_LOGFILE );
	}

	Project_AddHeader(xSaveBlock);

	#ifdef ASURA_TOOLS
	//save the next tools created guids
	for(int i=0; i<TOOLS_NUM_GUIDRANGE; i++)
	{
		Asura_Chunk_Entity_NextGuid::Create(i, xSaveBlock);
	}
	#endif

	Asura_Streaming_System::SaveMissionDataToBuffer(xSaveBlock);

	// For mission saves, ensure we save the zone data too.
	Asura_Chunk_CurrentStreamingZone::Create(xSaveBlock);

	// now call project specific save
	if(!Project_SaveMissionToBuffer(xSaveBlock)) return false;
	
	// tidy up
	xSaveBlock.AddAsuraTerminator();

	return true;
}

bool Asura_LoadAndSave::LoadMissionFromBuffer( const Asura_SaveBlock_Base& xSaveBlock, bool bResetBeforeLoad/*=true*/ )
{
	// TODO - This currently doesn't check that the mission is the zone that set up all the current handlers. It just checks a mission exists and then assumes the rest.
	if (!Asura_Streaming_System::GetCurrentMission()) { return false; }

	// MGH 9/1/2007: The guids will be renumbered incorrectly if the mission isn't given priority.
	const bool bMissionHasPriority = Asura_Streaming_System::GetMissionHasPriority();
	Asura_Streaming_System::SetMissionHasPriority(true);

	// Ensure the mission sets up all subsystem handlers.
	// MGH 9/1/2007: If ResetBeforeLoad is true then we will be deleting the mission entities
	// and need to flag the memory bank as allowing deletion.
	const Asura_Hash_ID uMissionHash = Asura_Streaming_System::GetCurrentMission()->GetHashID();
	Asura_Streaming_System::SetCurrentMission(uMissionHash, Asura_Streaming_Zone::HIGH_DETAIL_DATA, bResetBeforeLoad);

	Asura_Chunk_SaveBlockReader_Stream xStream(&xSaveBlock);
			
	// make sure it's a valid Asura file
	if( !ProcessAsuraFileHeader(xStream) )
	{
		Asura_Assert( false, "Not a valid Asura file." );
		return false;
	}

	if( bResetBeforeLoad )
	{
		ResetMissionBeforeLoad();
	}

	Asura_ChunkBuffer xChunkBuffer;
	char* pcChunkHeaderBuffer = reinterpret_cast<char*>(&xChunkBuffer);
	
#define CHUNK_PROCESS(chunkID, chunkClass)	\
	case (chunkID):	\
	{	\
	Asura_Assert( sizeof(chunkClass) <= uASURA_MAX_CHUNKHEADER_SIZE, "chunk header too big" ); \
	xStream.ReadChunk_GetNextHeader(pcChunkHeaderBuffer, sizeof(chunkClass)); \
	class chunkClass *chunk2 = (chunkClass *)pxChunkHeader;	\
	bChunkReadCompletely = chunk2->Process(xStream);	\
	}	\
	break

	while(xStream.ReadChunk_GetNextID(pcChunkHeaderBuffer))
	{
		// note that only the chunk ID is valid at this point
		Asura_Chunk_Header*	pxChunkHeader = reinterpret_cast<Asura_Chunk_Header*>( pcChunkHeaderBuffer );

		bool bChunkReadCompletely = false;
		switch(pxChunkHeader->ID)
		{
			CHUNK_PROCESS( ASURA_CHUNK_ENTITY,           Asura_Chunk_Entity           );
#ifdef ASURA_TOOLS
			CHUNK_PROCESS( ASURA_CHUNK_ENTITY_TOOLSDATA, Asura_Chunk_Entity_ToolsData );
			CHUNK_PROCESS( ASURA_CHUNK_ENTITY_NEXTGUID, Asura_Chunk_Entity_NextGuid );
#endif
#ifdef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
			CHUNK_PROCESS( ASURA_CHUNK_DYNAMICMUSIC_INGAME, Asura_Chunk_DynamicMusic_Ingame_State );
#else
			CHUNK_PROCESS( ASURA_CHUNK_METAMUSIC_INGAME, Asura_Chunk_MetaMusic_Ingame_State );
#endif
			CHUNK_PROCESS( ASURA_CHUNK_STATICMESSAGES_INGAME, Asura_Chunk_StaticMessages_InGame );
			CHUNK_PROCESS( ASURA_CHUNK_CURRENTSTREAMINGZONE, Asura_Chunk_CurrentStreamingZone); // This chunk is only valid in mission loads

			default: 
			{	
				// couldn't handle chunk here so call project specific loader
				if(!Project_LoadMissionFromBuffer(pxChunkHeader,xStream))
				{
					// MGH 9/1/2007: Restore the settings that we have changed.
					if (bResetBeforeLoad)
					{
						Asura_Streaming_System::SetCurrentMission(uMissionHash);
					}
					Asura_Streaming_System::SetMissionHasPriority(bMissionHasPriority);

					// Give up.
					return false;
				}
				
				// successfully handled by the project load from buffer so continue
				break;
			}
		}

		if(!xStream.HasReadChunkHeader())
		{
			// KOR - read the rest of the header to get the chunk size
			xStream.ReadChunk_GetNextHeader(pcChunkHeaderBuffer, sizeof(Asura_Chunk_Header));
		}
		// KOR - finish off the current chunk. The stream performs verification and skipping if necessary
		xStream.ReadChunk_End(bChunkReadCompletely);
	}

	#undef CHUNK_PROCESS

	// MGH 9/1/2007: Restore the settings that we have changed.
	if (bResetBeforeLoad)
	{
		Asura_Streaming_System::SetCurrentMission(uMissionHash);
	}
	Asura_Streaming_System::SetMissionHasPriority(bMissionHasPriority);

	return true;
}

void Asura_LoadAndSave::ResetMissionBeforeLoad()
{
    // reset entity system before reloading entities
	Asura_Streaming_System::GetCurrentMission()->DestroySavedEntities();

	//Partial reset on the static message system. RAR 02/01/2006.
	// AJG - 03/05/06 - Just reset the current handler (as with the mission load in general, we're assuming the mission is what last set all the current handlers)
	Asura_StaticMessage_System::GetCurrentHandler().ResetDelayedMessages();

	// AJG - 03/05/06 - I'd say it's reasonable to reset all special effects completely, the entities for particles will still exist to recreate them.
    // MGH 14/1/2008: Reset the runtime decals, but we don't want to destroy the static decals.
	Asura_Decal_System::ResetBeforeLoad();

	Asura_PFX_System::ClearAllEffects();

#ifdef ASURA_USE_DIALOGUE_SYSTEM
	Asura_Dialogue_Storage::ResetInGameState();
	Asura_Dialogue_System::StopAll();
#endif

    Project_ResetMissionBeforeLoad();
}
#endif //#ifdef ASURA_USE_STREAMING

void Asura_LoadAndSave::ResetBeforeLoad()
{
    // reset entity system before reloading entities
	Asura_Entity_System::Reset();

	//Partial reset on the static message system. RAR 02/01/2006
	Asura_StaticMessage_System::ResetDelayedMessages();
	
	// MGH 14/1/2008: Reset the runtime decals, but we don't want to destroy the static decals.
	Asura_Decal_System::ResetBeforeLoad();

	Asura_PFX_System::ClearAllEffects();
	Asura_Beam_System::Reset();

#ifdef ASURA_USE_DIALOGUE_SYSTEM
	Asura_Dialogue_Storage::ResetInGameState();
	Asura_Dialogue_System::StopAll();
#endif

	// Restoring the state of the dynamic music system may cause issues with what it's currently doing if the load is in a different thread.
#ifdef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM
	Asura_Dynamic_Music_System::Stop();
#endif

#if defined(ASURA_USE_FSFX)
	Asura_FSFX_System::ClearAllHandlerTemplates();
#endif	//defined(ASURA_USE_FSFX)

    Project_ResetBeforeLoad();
}

#if defined(_DEBUG) && defined(ASURA_SAVEBLOCK_USE_ZLIB)
void Asura_LoadAndSave::RunCompressedSaveBlockTests(Asura_CompressedSaveBlock& xSaveBlock)
{
	u_int uMinTotalMemUsed = 0xffffffff;
	u_int uBestWindowBits = 0;
	u_int uBestMemLevel = 0;
	Asura_OutputToDebugger(("Mem\tWindow\tCompressed size\tTotal mem usage\n"))
	for(u_int uWindowBits = 8; uWindowBits <= 15; ++uWindowBits )
	{
		// See zconf.h line 151
		const u_int uWindowBufferSize = 1 << (uWindowBits + 2);
		for(u_int uMemLevel = 1; uMemLevel <= 9; ++uMemLevel)
		{
			const u_int uMemBufferSize = 1 << (uMemLevel + 9);
			Asura_CompressedSaveBlock::SetWindowBits(uWindowBits);
			Asura_CompressedSaveBlock::SetMemLevel(uMemLevel);
			xSaveBlock.Clear();
			SaveToBuffer(xSaveBlock);
			const u_int uCompressedSize = xSaveBlock.GetDeflatedSize();
			const u_int uTotalMemUsed = uCompressedSize + uWindowBufferSize + uMemBufferSize;
			if(uTotalMemUsed < uMinTotalMemUsed)
			{
				uMinTotalMemUsed = uTotalMemUsed;
				uBestWindowBits = uWindowBits;
				uBestMemLevel = uMemLevel;
			}
			Asura_OutputToDebugger(("%2u\t%2u\t%3uKb\t%3uKb\n", uMemLevel, uWindowBits, uCompressedSize / 1024, uTotalMemUsed / 1024));
		}
	}
	Asura_OutputToDebugger(("Best result - mem level %u, window bits %u, total mem used %uKb\n", uBestMemLevel, uBestWindowBits, uMinTotalMemUsed / 1024));
	Asura_CompressedSaveBlock::SetWindowBits(uBestWindowBits);
	Asura_CompressedSaveBlock::SetMemLevel(uBestMemLevel);
}
#endif

/****************/
/* End of file. */
/****************/
