//////////////////////////////////////////////////////////////////////////////
//
// Asura_LanGameFinder.cpp
//
// Copyright (c) 2009 Rebellion. All rights reserved.
//
// Description:
//
//     class to deal with LAN game finding using simple broadcast. Should
//     work for any platform using standard sockets. On 360 it's used only if
//     ASURA_XBOX360_BYPASS_NETWORK_SECURITY is defined.
//     Note that it's just a group of 4 functions originally defined in each
//     platform as members of Asura_NetTransport_System. As such it's very
//     intertwined with that class and is a friend of it to allow data to be
//     directly manipulated. It would be nice to tidy & refactor at some point
//     although it's unlikely that will ever be a priority...
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// This file's header.
#include "Network/Asura_LanGameFinder.h"

// Asura headers
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_Network_Session.h"
#include "Network/Online/Asura_Lobby.h"


// I N T E R N A L   C L A S S E S ///////////////////////////////////////////
class Asura_Network_LANBroadcastServerRequest
{
public:

	u_short     m_usNBOSequenceNumber;
	u_short		m_usNBOChecksum;
	u_short		m_usNBOMagicID;		// = ASURA_NETTRANSPORT_MSG_ENUMHOSTS
	u_short		m_usNBOGameVersion;
	u_char		m_ucIsLobbySystemInitialised;
	u_char		m_acPadding[3];
	u_int		m_uNBOServerSessionID;	// used for lobby system
	u_int		m_uNBOApplicationGuid;
};

class Asura_Network_LANBroadcastServerReplyHeader
{
public:

	u_short m_usNBOSequenceNumber;
	u_short	m_usNBOChecksum;
	u_short m_usNBOMagicID;		// ASURA_NETTRANSPORT_MSG_HOSTREPLY
	u_short m_usNBOGameVersion;
	u_int m_uNBOApplicationGuid;
	u_int m_uNBOServerSessionID;

	u_short m_usNBODataSize;		// this data is followed by this many bytes of user data
	Asura_UnicodeChar m_wszNBOSessionName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
	u_char m_ucNumPlayers;
	u_char m_ucMaxPlayers;
	u_char m_ucIsLobbySystemInitialised;
	u_short m_usNBOLobbyPort;
	Asura_Lobby_Attributes m_xLobbyAttributes;
	// added information in reply packet indicating requester's address and port
	// as seen by the server. This is to help prevent servers finding themselves
	// if they do a search while hosting.
	char m_acRequestingAddress[6];	// contains up to 6 bytes of netaddress
	u_short m_usRequestingPort;
};

// C O N S T A N T S /////////////////////////////////////////////////////////
const int iNUM_PORTS_TO_SCAN = 8;
const u_short usASURA_DEFAULT_PORT = 3658;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
u_int Asura_LanGameFinder::s_uServerSessionID;

bool Asura_LanGameFinder::SendEnumHostsBroadcast()
{
	bool iResult = false;

	// create array of ports to broadcast to. This should include the platform
	// default port(s), the asura default port 3658 (if not already a platform
	// default), and the dynamic port range.
	u_short ausPorts[iNUM_PORTS_TO_SCAN];
	{
		int iPort = 0;
		ausPorts[iPort++] = ASURA_NETSOCKET_DEFAULTPORT;
#if ASURA_NETSOCKET_DEFAULTPORT != usASURA_DEFAULT_PORT
		ausPorts[iPort++] = usASURA_DEFAULT_PORT;
#endif
#ifdef ASURA_NETSOCKET_ALTERNATIVE_PORT
		ausPorts[iPort++] = ASURA_NETSOCKET_ALTERNATIVE_PORT;
#endif
		Asura_Assert(iPort < iNUM_PORTS_TO_SCAN, "array overwrite!");
		u_short usDynamicPort = ASURA_NETSOCKET_FIRST_DYNAMIC_PORT;
		for(; iPort < iNUM_PORTS_TO_SCAN; ++iPort)
		{
			ausPorts[iPort] = usDynamicPort++;
		}
	}

	for(int i = 0; i < iNUM_PORTS_TO_SCAN; ++i)
	{
#ifdef ASURA_NETADDRESS_BROADCAST
		Asura_NetAddress xBroadcastAddr(ASURA_NETADDRESS_BROADCAST);
#else
		Asura_NetAddress xBroadcastAddr(0xffffffff);
#endif

		const u_short usPort = ausPorts[i];
		Asura_Network_LANBroadcastServerRequest xBroadcastMsg;
		xBroadcastMsg.m_usNBOSequenceNumber = 0; // no need to endian swap 0
		// checksum filled out later
		xBroadcastMsg.m_usNBOGameVersion = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::Project_GetVersionID());
		xBroadcastMsg.m_usNBOMagicID = ASURA_NETTRANSPORT_MSG_ENUMHOSTS;
		ASURA_NTS_CHANGE_BYTE_ORDER(xBroadcastMsg.m_usNBOMagicID);
#ifdef ASURA_SUPPORT_LOBBY
		xBroadcastMsg.m_ucIsLobbySystemInitialised = Asura_Lobby_System::IsInitialised() ? 1 : 0;
		xBroadcastMsg.m_uNBOServerSessionID = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(s_uServerSessionID);
#else
		xBroadcastMsg.m_ucIsLobbySystemInitialised = 0;
		xBroadcastMsg.m_uNBOServerSessionID = 0;
#endif
		xBroadcastMsg.m_uNBOApplicationGuid = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::Project_GetApplicationGuid());

		// checksum on length of data after SeqNum and checksum
		xBroadcastMsg.m_usNBOChecksum = Asura_NetTransport_System::ComputeChecksum( reinterpret_cast<const char*>(&xBroadcastMsg.m_usNBOMagicID), sizeof(xBroadcastMsg) - 4 );
		ASURA_NTS_CHANGE_BYTE_ORDER(xBroadcastMsg.m_usNBOChecksum);

		Asura_Assert(Asura_NetTransport_System::s_pxSocket, "No socket");
		if(Asura_NetTransport_System::s_pxSocket)
		{
			iResult |= Asura_NetTransport_System::s_pxSocket->Send(xBroadcastAddr, usPort, reinterpret_cast<const char*>(&xBroadcastMsg), sizeof(xBroadcastMsg));
			Asura_Assert(iResult, "failed to send broadcast message");

			Asura_NetTransport_System::AddDirectSendBytes(sizeof(xBroadcastMsg));
		}
	}
	Asura_NetTransport_System::s_fLastSessionRequestBroadcastTime = Asura_Timers::GetLocalUnscaledTime();
	return iResult;
}

void Asura_LanGameFinder::ProcessEnumHostsRequest( Asura_NetAddress xAddress, u_short usPort, const char* pcData, u_int uDataLength )
{
	const Asura_Network_LANBroadcastServerRequest* pxBroadcastRequestMsg = reinterpret_cast
		<const Asura_Network_LANBroadcastServerRequest*>(pcData);

	// Ignore requests from other games or versions
	const u_int uApplicationGuid = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxBroadcastRequestMsg->m_uNBOApplicationGuid);
	if (uApplicationGuid != Asura_NetTransport_System::Project_GetApplicationGuid())
	{
		return;
	}

	const u_short usGameVersion = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxBroadcastRequestMsg->m_usNBOGameVersion);
	if (usGameVersion != Asura_NetTransport_System::Project_GetVersionID())
	{
		return;
	}

	if (uDataLength != sizeof(Asura_Network_LANBroadcastServerRequest))
	{
		Asura_Assert(false, "bad data length in enum hosts");
		return;
	}

#ifdef ASURA_SUPPORT_LOBBY
	if (pxBroadcastRequestMsg->m_ucIsLobbySystemInitialised != u_char(Asura_Lobby_System::IsInitialised()))
	{
		return;
	}
	// if game has started and join-in-progress not allowed then return.
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	if (pxLobby && pxLobby->GetGameState() != ASURA_LOBBY_GAMESTATE_INVALID)
	{
		const bool bAllowJoinInProgress = Asura_Lobby_System::GetAllowJoinInProgress() && 
										  !Asura_Lobby_System::IsUsingRankedLobbies();
		if (!bAllowJoinInProgress)
		{
			// only report if this is an attempt to find the game port by a player already in the lobby
			u_int uServerSessionID = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxBroadcastRequestMsg->m_uNBOServerSessionID);
			if (uServerSessionID != Asura_NetTransport_System::GetServerSessionID())
			{
				return;
			}
		}
	}
#endif

	// check the checksum
	const u_short usChecksum = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxBroadcastRequestMsg->m_usNBOChecksum);
	if (usChecksum != Asura_NetTransport_System::ComputeChecksum( reinterpret_cast<const char*>(&pxBroadcastRequestMsg->m_usNBOMagicID), uDataLength - 4 ))
	{
		Asura_Assert(false, "bad checksum in enum hosts");
		return;
	}

	Asura_OutputToDebugger(("Received enum hosts request from %s with %i bytes of data.\n", 
		Asura_NetTransport_System::GetStringFromAddress(xAddress, usPort), uDataLength));
	SendEnumHostsResponse( xAddress, usPort );
}

void Asura_LanGameFinder::SendEnumHostsResponse( Asura_NetAddress xAddress, u_short usPort)
{
	// (MH - 01.07.2004) ensure the session data is fresh
	Asura_NetTransport_System::Project_RefreshSessionData();

	u_int uSizeOfReplyData = sizeof(Asura_Network_LANBroadcastServerReplyHeader) + Asura_NetTransport_System::s_uSessionDataSize;
	char* const pcReplyData = new char[uSizeOfReplyData];

	Asura_Network_LANBroadcastServerReplyHeader* pxHeader = reinterpret_cast
		<Asura_Network_LANBroadcastServerReplyHeader*>(pcReplyData);
	pxHeader->m_usNBOSequenceNumber = 0; // no need to endian swap 0
	// checksum filled out later
	pxHeader->m_usNBOMagicID = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(static_cast<u_short>(ASURA_NETTRANSPORT_MSG_HOSTREPLY));;
	pxHeader->m_usNBOGameVersion = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::Project_GetVersionID());
	pxHeader->m_uNBOApplicationGuid = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::Project_GetApplicationGuid());
	pxHeader->m_uNBOServerSessionID = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::s_uServerSessionID);
	pxHeader->m_usNBODataSize = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(static_cast<u_short>(Asura_NetTransport_System::s_uSessionDataSize));
#ifndef ASURA_SUPPORT_LOBBY
	pxHeader->m_ucIsLobbySystemInitialised = 0;
	pxHeader->m_usNBOLobbyPort = 0;
#else
	pxHeader->m_ucIsLobbySystemInitialised = Asura_Lobby_System::IsInitialised() ? 1 : 0;
	pxHeader->m_usNBOLobbyPort = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_Lobby_System::GetBoundTCPPort());
	if (Asura_Lobby_System::IsInitialised() && Asura_Lobby_System::GetLobby())
	{
		const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
		pxHeader->m_xLobbyAttributes = pxLobby->GetAttributes();
		pxHeader->m_xLobbyAttributes.Project_ChangeByteOrder();
		pxHeader->m_ucNumPlayers = static_cast<u_char>(pxLobby->GetNumPlayers());
		pxHeader->m_ucMaxPlayers = static_cast<u_char>(pxLobby->GetMaxPlayers());
	}
	else
#endif
	{
		//Prefer the lobbies values, as the network will only be active if the user is in-game
		pxHeader->m_ucNumPlayers = static_cast<u_char>(Asura_Network::GetNumberOfPlayers());
		pxHeader->m_ucMaxPlayers = static_cast<u_char>(Asura_Network::GetMaxNumberOfPlayers());
	}

	// session name in NBO
	Asura_PrintfToBfr(pxHeader->m_wszNBOSessionName, L"%s", uASURA_NETWORK_MAX_SESSION_NAME + 1, true, Asura_Network::GetServerName());
	for(u_int u = 0; u < uASURA_NETWORK_MAX_SESSION_NAME; ++u)
	{
		ASURA_NTS_CHANGE_BYTE_ORDER(pxHeader->m_wszNBOSessionName[u]);
	}
	// net address of requester as seen on this machine
	Asura_Mem::Copy(pxHeader->m_acRequestingAddress, &xAddress, sizeof(pxHeader->m_acRequestingAddress));
	pxHeader->m_usRequestingPort = usPort;

	u_char* pucSessionData = reinterpret_cast<u_char*>(pxHeader + 1);
	memcpy(pucSessionData, Asura_NetTransport_System::s_pucSessionData, Asura_NetTransport_System::s_uSessionDataSize);

	// checksum on length of data after SeqNum and checksum
	pxHeader->m_usNBOChecksum = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL( Asura_NetTransport_System::ComputeChecksum( 
		reinterpret_cast<const char*>(&pxHeader->m_usNBOMagicID), uSizeOfReplyData - 4 ));

#if defined ASURA_ASSERT
	const bool bResult =
#endif
		Asura_NetTransport_System::s_pxSocket->Send(xAddress, usPort, pcReplyData, uSizeOfReplyData);

	Asura_Assert(bResult, "failed to send broadcast message");

	Asura_NetTransport_System::AddDirectSendBytes(uSizeOfReplyData);

	delete[] pcReplyData;
}

void Asura_LanGameFinder::ProcessEnumHostsResponse( Asura_NetAddress xAddress, u_short usPort, const char* pcData, int iDataSize )
{
	const Asura_Network_LANBroadcastServerReplyHeader* pxHeader = reinterpret_cast
		<const Asura_Network_LANBroadcastServerReplyHeader*>(pcData);
	const u_char* pucSessionData = reinterpret_cast<const u_char*>(pxHeader + 1);

	// check project and version
	const u_int uApplicationGuid = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxHeader->m_uNBOApplicationGuid);
	const u_short usGameVersion = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxHeader->m_usNBOGameVersion);
	if (uApplicationGuid != Asura_NetTransport_System::Project_GetApplicationGuid()) return;
	if (usGameVersion != Asura_NetTransport_System::Project_GetVersionID()) return;

	// check data length
	const u_short usDataSize = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxHeader->m_usNBODataSize);
	if (static_cast<u_int>(iDataSize) != sizeof(Asura_Network_LANBroadcastServerReplyHeader) + usDataSize)
	{
		Asura_Assert(false, "wrong data length in enum hosts response");
		return;
	}

	// check the checksum
	const u_short usChecksum = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxHeader->m_usNBOChecksum);
	if (usChecksum != Asura_NetTransport_System::ComputeChecksum( reinterpret_cast<const char*>(&pxHeader->m_usNBOMagicID), iDataSize - 4 ))
	{
		Asura_Assert(false, "bad checksum in enum hosts response");
		return;
	}

	// check this isn't a response from ourselves
	Asura_NetAddress xOriginalAddress;
	Asura_Mem::Copy(&xOriginalAddress, pxHeader->m_acRequestingAddress, sizeof(xOriginalAddress));
	if (xOriginalAddress == xAddress && pxHeader->m_usRequestingPort == usPort)
	{
		Asura_OutputToDebugger(("Ignoring enum hosts response from %s since it seems to be from ourselves.\n", 
			Asura_NetTransport_System::GetStringFromAddress(xAddress, usPort)));
		return;
	}

#ifdef ASURA_SUPPORT_LOBBY
	if (pxHeader->m_ucIsLobbySystemInitialised != u_char(Asura_Lobby_System::IsInitialised()))
	{
		return;
	}
#endif

	if (!Asura_NetTransport_System::s_ppxSessionsStorage) return;

	// work out ping based on time since message was sent. Possible problem of
	// overlapping response and request...
	const u_int uPing = Asura_Maths::FloatToInt(1000.0f * (Asura_Timers::GetLocalUnscaledTime() - Asura_NetTransport_System::s_fLastSessionRequestBroadcastTime));

	const void* pPlatformData = NULL;
	u_int uPlatformDataSize = 0;

#ifdef ASURA_SUPPORT_LOBBY
	Asura_Lobby_Attributes xLobbyAttributes;
	if (Asura_Lobby_System::IsInitialised())
	{
		xLobbyAttributes = pxHeader->m_xLobbyAttributes;
		xLobbyAttributes.Project_ChangeByteOrder();
		pPlatformData = &xLobbyAttributes;
		uPlatformDataSize = sizeof(xLobbyAttributes);

		// if looking to connect to the lobby TCP socket rather than to the game, 
		// overwrite the port with the lobby port
		if (Asura_Lobby_System::GetState() != ASURA_LOBBY_STATE_INITIALISING_GAME)
		{
			usPort = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxHeader->m_usNBOLobbyPort);
		}
	}
#endif

	Asura_UnicodeChar wszSessionName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
	Asura_Mem::Copy(wszSessionName, pxHeader->m_wszNBOSessionName, sizeof(Asura_UnicodeChar) * (uASURA_NETWORK_MAX_SESSION_NAME + 1));
	for(u_int u = 0; u < uASURA_NETWORK_MAX_SESSION_NAME; ++u)
	{
		ASURA_NTS_CHANGE_BYTE_ORDER(wszSessionName[u]);
	}
	const u_int uServerSessionID = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxHeader->m_uNBOServerSessionID);
	Asura_Network_SessionInfo* pxNewSession =  new Asura_Network_SessionInfo(xAddress, 
		usPort, uPing, uServerSessionID, pPlatformData, uPlatformDataSize, pucSessionData, usDataSize, wszSessionName, 
		pxHeader->m_ucNumPlayers, pxHeader->m_ucMaxPlayers);

	for (u_int u = 0; u < Asura_NetTransport_System::s_uNumberOfActiveSessions; ++u)
	{
		if (Asura_NetTransport_System::s_ppxSessionsStorage[u]->Platform_IsSameHost(pxNewSession))
		{
			// SNH: copied from xbox version. Replace the session details.
			delete Asura_NetTransport_System::s_ppxSessionsStorage[u];
			Asura_NetTransport_System::s_ppxSessionsStorage[u] = pxNewSession;
			return;
		}
	}

	// this is a new server
	if (Asura_NetTransport_System::s_uNumberOfActiveSessions < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS)
	{
		Asura_NetTransport_System::s_ppxSessionsStorage[Asura_NetTransport_System::s_uNumberOfActiveSessions++] = pxNewSession;
	}
	else
	{
		delete pxNewSession;
	}
}

#endif //#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
