//
// Asura_NetConnection.cpp
//
// Created by KOR - 03/11/2003 11:10:22
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// This file's header.
#include "Network/Asura_NetConnection.h"

// Asura includes
#include "Core/Asura_Hash.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_Time_System.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_InternetSimulation.h"
#include "Network/Asura_MessageHeader.h"
#include "Network/Asura_NetSocket.h"

#include <wchar.h>

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

u_int Asura_NetConnection::s_uNumSavedClientRequests = 0;

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR bool dbg_bOutputAllReceivedMessages = false;
DEBUGVAR bool dbg_bOutputReliableMessages = false;
DEBUGVAR bool dbg_bOutputSentMessages = false;

// C O N S T A N T S /////////////////////////////////////////////////////////
const u_int uASURA_NETCONNECTION_MINTIMEBETWEENPONGS = 500;

const u_int uASURA_NETCONNECTION_MAX_SAVED_CLIENT_REQUESTS = 16;	// Dont save more than this many client requests (prevents potential DDoS attacks)

// SNH - 14/10/2004 12:04:20 - Reserved sequence numbers. Messages that come from non-Asura sources (like GameSpy peer / qr2 queries
//	and NatNeg data) may use the first two bytes of the message as magic numbers. Asura messages use these
//	bytes as reliable message sequence numbers. Therefore to prevent Asura messages potentially being incorrectly
//	diverted to GameSpy functions, the SNs corresponding to those magic numbers	are reserved and not used.
const u_short g_ausReservedSN[] = 
{
	0,					// Interpreted as 'unreliable' message
	0xFEFD,				// GameSpy Peer / QR2 query
	0xFDFC,				// GameSpy NatNeg data
};

const u_int g_uNumberOfReservedSN = sizeof(g_ausReservedSN) / sizeof(u_short);

void Asura_NetConnection::LogToFile(const char* szMessage, ...)
{
#if defined ASURA_LOGFILES_ENABLED || !defined(ASURA_NO_DEBUG_OUTPUT)
	static bool doassert = false;
	static char* dbg_szLofFileName[2] = { "logfiles\\server.txt", "logfiles\\client.txt" };

	char szBuffer[256];
	Asura_PrintfToBfr(szBuffer, "Asura_NetConnection[%d] : %s\n", 256, true, m_pxMachineInfo ? Asura_NetTransport_System::GetMachineIndexFromInfo(m_pxMachineInfo) : 0xffffffff, szMessage);

	char szBuffer2[256];
	va_list list;
	va_start(list, szMessage);
    Asura_PrintfToBfr( szBuffer2, szBuffer, list, sizeof(szBuffer2), true );
	va_end(list);

	Asura_Assert(!doassert, szBuffer2);

#if defined ASURA_LOGFILES_ENABLED
	// code copied form PC console window code to output time and frame number inside a logfile.
	const char* szLogfile = dbg_szLofFileName[Asura_Network::IsServer()? 0 : 1];
	static u_int ls_uLastFrame = 0;
	u_int uCurFrame = Asura_Timers::GetFrameNumber();
	if (ls_uLastFrame != uCurFrame)
	{
		ls_uLastFrame = uCurFrame;

		const Asura_UnicodeChar* wszDateAndTime = Asura_Time_System::GetCurrentDateAndTimeString(ASURA_DATE_MINIMAL);
		static Asura_UnicodeChar ls_awcLastDateAndTime[128] = L"";
		if (wcscmp(wszDateAndTime, ls_awcLastDateAndTime))
		{
			Asura_PrintfToBfr( ls_awcLastDateAndTime, L"%s", 128, true, wszDateAndTime );
			Asura_SimpleLogfileOutput((szLogfile, "\nframe %u: %S\n", uCurFrame, wszDateAndTime));
		}
		else
		{
			Asura_SimpleLogfileOutput((szLogfile, "\nframe %u\n", uCurFrame));
		}
	}
	Asura_SimpleLogfileOutput((szLogfile, szBuffer2));
#endif

	Asura_OutputToDebugger((szBuffer2));
#endif // #if defined ASURA_LOGFILES_ENABLED || !defined(ASURA_NO_DEBUG_OUTPUT)
}

// M E M B E R S /////////////////////////////////////////////////////////
Asura_NetConnection::Asura_NetConnection()
{
	m_bActive = false;
}

Asura_NetConnection::~Asura_NetConnection()
{
}

void Asura_NetConnection::Activate(Asura_NetSocket *pxSocket, Asura_NetAddress xAddress, unsigned short uRemotePort, int iTimeOut, bool bBlocking, bool bInitiate)
{
	Asura_Assert(!IsActive(), "Net connection already active");
	m_bActive = true;
	m_xAddress = xAddress;
	m_usPort = uRemotePort;
	m_pxSocket = pxSocket;
	m_eState = ASURA_CONNECTION_CONNECTING;
	m_bInitiated = bInitiate;
	m_eConnectionResult = ASURA_NETWORK_NETWORKERROR;
	m_pxMachineInfo = NULL;
	m_usSequenceNumber = 1;
	m_uLastReceiveTime = Asura_Timers::GetMSecsSinceStart();
	m_uLastKeepAliveSendTime = 0;
	m_uLastPongSendTime = 0;
	m_uReceivedSNCount = 0;
	m_uLastAckSendTime = 0;
	m_uLastResendTime = 0;
	m_uNumberOfExpectedSNs = 0;
	m_usHighestSNReceived = 0;
	m_fClosingTimer = 0.0f;
	m_pcClientRequestData = 0;

	m_ausNBOReceivedSNs[0] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(static_cast<u_short>(ASURA_NETTRANSPORT_MSG_ACK));
	m_ausNBOReceivedSNs[1] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::Project_GetVersionID());

	bool bSuccess = true;
	if(bInitiate)
	{
		// On construction, start connection attempt by sending request for challenge key
		Asura_Network_Challenge xMessage;
		xMessage.m_uChallenge = 0;
		xMessage.m_uAdditionalDataSize = Platform_GetMessageDataSize( ASURA_NETTRANSPORT_MSG_CHALLENGE_REQUEST );
		
		// Build message data
		const u_int uMsgSize = sizeof( xMessage ) + xMessage.m_uAdditionalDataSize + 4;

		char* pcMsgData = new char[ uMsgSize ];

		if( pcMsgData )
		{
			unsigned short* pusMsgData = reinterpret_cast<unsigned short*>( pcMsgData );

			pusMsgData[0] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL( static_cast<unsigned short>( ASURA_NETTRANSPORT_MSG_CHALLENGE_REQUEST ) );
			pusMsgData[1] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL( Asura_NetTransport_System::Project_GetVersionID() );

			u_int uMsgPos = sizeof( unsigned short ) * 2;
			Asura_Mem::Copy( &pcMsgData[ uMsgPos ], &xMessage, sizeof( xMessage ) );

			uMsgPos += sizeof( xMessage );
			Platform_FillMessageData( ASURA_NETTRANSPORT_MSG_CHALLENGE_REQUEST, &pcMsgData[ uMsgPos ] );

			bSuccess = Send( pcMsgData, uMsgSize, true);

			delete []pcMsgData;
		}
		else
		{
			bSuccess = false;
		}
	}
	if(bSuccess)
	{
		m_eConnectionResult = ASURA_NETWORK_SUCCESS;
	}
	else
	{
		m_eConnectionResult = ASURA_NETWORK_NETWORKERROR;
	}
}

void Asura_NetConnection::Deactivate()
{
	Asura_Assert(IsActive(), "Net connection deactivated when not active");
	m_xOutgoingBuffer.Reset();
	m_bActive = false;
	ClearClientRequestData();
}


bool Asura_NetConnection::Accept()
{
	Asura_Assert(m_eState == ASURA_CONNECTION_CONNECTING, "Already connected or closed this connection");
	if(m_eState == ASURA_CONNECTION_CONNECTING)
	{
		m_eState = ASURA_CONNECTION_CONNECTED;
		return SendSimple(ASURA_NETTRANSPORT_MSG_CONNECTION_ACCEPT, true);
	}
	else
	{
		return false;
	}
}

bool Asura_NetConnection::Close(bool bForce, Asura_Close_Reason eReason)
{
	m_eCloseReason = eReason;
	m_eState = ASURA_CONNECTION_CLOSED;

	LogToFile("Closing connection. reason[%d].", eReason);
	
	if(m_eCloseReason == ASURA_CLOSE_LOCAL)
	{
		return SendClosed(ASURA_CLOSE_REMOTE);
	}
	else
	{
		return m_pxSocket->SendClosed(m_xAddress, m_usPort, m_eCloseReason);
	}
}

bool Asura_NetConnection::HandleAccept(const char* pcMessage, int iLength)
{
	Asura_NetTransport_System::Client_ClientConnected(this, ASURA_NETWORK_SUCCESS, NULL, 0);
	return true;
}

bool Asura_NetConnection::HandleAck(const char* pcMessage, int iLength)
{
	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	Asura_Assert(iLength >= 0, "No SNs in ACK");
	const int iNumberOfSNs = iLength / 2;
	const u_short* pusMessage = reinterpret_cast<const u_short*>(pcMessage);
	for(int i = 0; i < iNumberOfSNs; ++i)
	{
		m_xOutgoingBuffer.RemoveMessage(*(pusMessage + i));
	}

	Asura_Network_MachineInfo* pxMachineInfo = Asura_NetTransport_System::GetMachineInfo(this);
	if (pxMachineInfo)
	{
		pxMachineInfo->AddAcks(static_cast<u_int>(iNumberOfSNs));
	}
	
	return true;
}

bool Asura_NetConnection::HandleClose(const char* pcMessage, int iLength)
{
	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	Asura_Assert(iLength == sizeof(u_short), "No reason for close given");
	const u_short* pusReason = reinterpret_cast<const u_short*>(pcMessage);
	m_eCloseReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(static_cast<Asura_Close_Reason>(*pusReason));
	m_eState = ASURA_CONNECTION_CLOSED;

	LogToFile("Received close connection message. reason [%d].", m_eCloseReason);

	return true;
}

bool Asura_NetConnection::HandleReject(const char* pcMessage, int iLength)
{
	if(m_eState != ASURA_CONNECTION_CONNECTING)
	{
		return false;
	}
	
	m_eState = ASURA_CONNECTION_CLOSED;

	LogToFile("Closing connection : Received a Reject connection message");

	if(!SendClosed(ASURA_CLOSE_ERROR_COMMUNICATION))	return false;

	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	Asura_NetTransport_System::Client_ClientConnected(this, ASURA_NETWORK_REJECTED, pcMessage, iLength);
	return true;
}

bool Asura_NetConnection::HandlePing(const char* pcMessage, int iLength)
{
	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	// Should just be a time
	Asura_Assert(iLength == sizeof(unsigned int), "Incorrect message data length");
	if (iLength != sizeof(unsigned int))
	{
		return false;
	}

	// Send the message back
	u_int uCurTime = Asura_Timers::GetMSecsSinceStart();
	if (uCurTime - m_uLastPongSendTime >= uASURA_NETCONNECTION_MINTIMEBETWEENPONGS)
	{
		m_uLastPongSendTime = uCurTime;
		return SendSimple(ASURA_NETTRANSPORT_MSG_PONG, false, pcMessage, iLength);
	}
	else
	{
		return true;
	}
}

bool Asura_NetConnection::HandlePong(const char* pcMessage, int iLength)
{
	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	// possible for machine info to be NULL if not properly connected.
	if (!m_pxMachineInfo) return true;

	// Should just be a time
	Asura_Assert(iLength == sizeof(unsigned int), "Incorrect message data length");
	const u_int uCurTime = Asura_Timers::GetMSecsSinceStart();
	const u_int uPingTime = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(*reinterpret_cast<const u_int*>(pcMessage));
	m_pxMachineInfo->SetPing(uCurTime - uPingTime);
	return true;
}

bool Asura_NetConnection::HandleClientRequest(const char* pcMessage, int iLength)
{
	if (m_eState != ASURA_CONNECTION_CONNECTING)
	{
		// must be a duplicate message
		return true;
	}

	// should also have gamever in here
	const u_short usGameVer = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(*reinterpret_cast<const u_short*>(pcMessage));
	if (usGameVer != Asura_NetTransport_System::Project_GetVersionID())
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_VERSION);
		LogToFile("Rejecting connection. Reason [ASURA_NDR_VERSION]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	// We need to call either Accept or Reject in response to this, but 
	// we don't have to do it immediately if, say, we want to authenticate a 
	// CD key with another computer, or see if the user is banned.
	
	// check number of client machines
	if( Asura_Network::GetNumberOfMachines() >= Asura_Network::GetMaxNumberOfMachines() )
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_SERVERFULL);
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check data received for correct password etc.
	// first check validity of message data struct
	if ( iLength < static_cast<int>(sizeof(Asura_Network_ConnectData)) )
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_VERSION);
		LogToFile("Rejecting connection. Reason[ASURA_NDR_VERSION]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	const Asura_Network_ConnectData* pxConnectData = reinterpret_cast
		<const Asura_Network_ConnectData*>(pcMessage);
	
	LogToFile("Received a connection request from client");
	
	// check length of message including extra data.
	u_int uExtraConnectDataSize = Asura_NetTransport_System::Project_GetExtraConnectDataSize();
	if (iLength < static_cast<int>(sizeof(Asura_Network_ConnectData) + uExtraConnectDataSize))
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_VERSION);
		LogToFile("Rejecting connection. Reason [ASURA_NDR_VERSION]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check game version
	const u_short usGameVersion = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxConnectData->m_usNBOGameVersion);
	if (usGameVersion != Asura_NetTransport_System::Project_GetVersionID())
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_VERSION);
		LogToFile("Rejecting connection. Reason [ASURA_NDR_VERSION]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check not trying to connect more players than supported per machine
	const u_short usNumPlayers = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxConnectData->m_usNBONumPlayers);
	if (usNumPlayers > Asura_NetTransport_System::GetMaxPlayersPerMachine())
	{
		Asura_Assert(false, "too many players in connect request. Need to call Asura_NetTransport_System::SetMaxPlayersPerMachine()?");
		Asura_Network_ConnectionRejectedData xRejectionReason;
		// probably a version mismatch.
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_VERSION);
		LogToFile("Rejecting connection. Reason [ASURA_NDR_VERSION]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check number of players
	const u_int uNumPlayersIfAccepted = Asura_Network::GetNumberOfPlayers() + usNumPlayers;
	if (uNumPlayersIfAccepted > Asura_Network::GetMaxNumberOfPlayers())
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_SERVERFULL);
		LogToFile("Rejecting connection. Reason [ASURA_NDR_SERVERFULL]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check password
	const u_int uPasswordHash = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxConnectData->m_uNBOPasswordHash);
	if ( (Asura_NetTransport_System::GetPasswordHash() != 0) && (uPasswordHash != Asura_NetTransport_System::GetPasswordHash()) )
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_PASSWORD);
		LogToFile("Rejecting connection. Reason [ASURA_NDR_PASSWORD]");
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check ServerSessionID
// TODO: get this on NovaWorld and GameSpy.
//	if (Asura_NetTransport_System::GetServerSessionID() != pxConnectData->m_uServerSessionID)
//	{
//		Asura_Network_ConnectionRejectedData xRejectionReason;
//		xRejectionReason.m_uReason = ASURA_NDR_SERVER_RESET;
//		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
//		return true;
//	}

	// check hash of server session ID (will accept current or previous hash).
	const u_int uCurHash = Asura_GetHashID_Binary(&Asura_NetTransport_System::s_auServerChallenge[0], sizeof(u_int));
	const u_int uChallengeResponse = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(pxConnectData->m_uNBOChallengeResponse);
	if (uCurHash != uChallengeResponse)
	{
		const u_int uPrevHash = Asura_GetHashID_Binary(&Asura_NetTransport_System::s_auServerChallenge[1], sizeof(u_int));
		if (uPrevHash != uChallengeResponse)
		{
			Asura_Network_ConnectionRejectedData xRejectionReason;
			xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(ASURA_NDR_WRONG_CHALLENGE_RESPONSE);
			LogToFile("Rejecting connection. Reason [ASURA_NDR_WRONG_CHALLENGE_RESPONSE]");
			Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
			return true;
		}
	}

	// check project restrictions
	u_int uProjectRejectionReason;
	const void* pExtraConnectData = pxConnectData + 1;
	if ( !Asura_NetTransport_System::Project_CheckExtraConnectData(pExtraConnectData, uProjectRejectionReason) )
	{
		Asura_Network_ConnectionRejectedData xRejectionReason;
		xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(uProjectRejectionReason);
		LogToFile("Rejecting connection. Reason [%d]", uProjectRejectionReason);
		Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
		return true;
	}

	// check platform
	Platform_ProcessMessageData( ASURA_NETTRANSPORT_MSG_CONNECTION_REQUEST, reinterpret_cast<const char*>( pxConnectData + 1 ) + pxConnectData->m_uNBOExtraDataSize );
	
	if( Platform_HandlesClientRequests() )
	{
		if( s_uNumSavedClientRequests < uASURA_NETCONNECTION_MAX_SAVED_CLIENT_REQUESTS )
		{
			// The platform will call Accept/Reject when it has finished processing the client request
			// unfortunately this means saving the message so it can be fed into ProcessMachineConnection later
			Asura_Assert( !m_pcClientRequestData, "There is already some saved message data... I'm about to blat it!");
			ClearClientRequestData();
			m_pcClientRequestData = new char[ iLength ];
			Asura_Mem::Copy( m_pcClientRequestData, pcMessage, iLength );
			++s_uNumSavedClientRequests;
			return true;
		}
		else
		{
			Asura_Network_ConnectionRejectedData xRejectionReason;
			// technically this is too many concurrent authorisation attempts, but timeout will suffice
			xRejectionReason.m_uNBOReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL( ASURA_NDR_TIMEOUT );
			LogToFile("Rejecting connection. Reason [%d]", ASURA_NDR_TIMEOUT);
			Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
			return true;
		}
	}

	const char* szAddress = Asura_NetTransport_System::GetStringFromAddress(m_xAddress, m_usPort);
	// TODO: work latency out
	int iLatency = 1;
	if (Accept())
	{
		// connection OK
		Asura_NetTransport_System::ProcessMachineConnection(this, pxConnectData, static_cast<u_int>(iLatency), pExtraConnectData);
		LogToFile("connection from %s accepted", szAddress);
		Asura_OutputToDebugger_Retail(("connection from %s accepted.\n", szAddress, NULL));
	}
	else
	{
		// connection timed out, possibly query of CD key took too long?
		LogToFile("connection from %s failed", szAddress);
		Asura_OutputToDebugger_Retail(("connection from %s failed.\n", szAddress, NULL));
	}
	return true;
}

bool Asura_NetConnection::HandleChallengeResponse(const char* pcMessage, int iLength)
{
	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	// TY - Size check no longer possible due to dynamic size
	const Asura_Network_Challenge* pxResponse = reinterpret_cast
		<const Asura_Network_Challenge*>(pcMessage);

	Platform_ProcessMessageData( ASURA_NETTRANSPORT_MSG_CHALLENGE_RESPONSE, pxResponse + 1 );

	// store the new server challenge.
	Asura_NetTransport_System::s_auServerChallenge[0] = pxResponse->m_uChallenge;
	// retry the connection
	const u_int uPlatformDataSize = Platform_GetMessageDataSize( ASURA_NETTRANSPORT_MSG_CONNECTION_REQUEST );
	const u_int uMessageSize = Asura_NetTransport_System::GetSizeOfConnectData() + uPlatformDataSize + sizeof( unsigned short ) * 2;
	char* pcMsgData = new char[uMessageSize];
	if (pcMsgData)
	{
		unsigned short* pusMsgData = reinterpret_cast< unsigned short* >( pcMsgData );
		pusMsgData[0] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL( static_cast<unsigned short>( ASURA_NETTRANSPORT_MSG_CONNECTION_REQUEST ) );
		pusMsgData[1] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL( Asura_NetTransport_System::Project_GetVersionID() );

		char* pcMsgPointer = &pcMsgData[ sizeof( unsigned short ) * 2 ];

		Asura_NetTransport_System::FillOutConnectData( pcMsgPointer );

		Asura_Network_ConnectData* pxUserConnectData = reinterpret_cast<Asura_Network_ConnectData*>( pcMsgPointer );
		pxUserConnectData->m_uExtraPlatformDataSize = Platform_GetMessageDataSize( ASURA_NETTRANSPORT_MSG_CONNECTION_REQUEST );

		Platform_FillMessageData( ASURA_NETTRANSPORT_MSG_CONNECTION_REQUEST, pcMsgPointer + sizeof( Asura_Network_ConnectData ) + pxUserConnectData->m_uNBOExtraDataSize );

		#if defined ASURA_ASSERT
			const bool bSuccess =
		#endif
			Send( pcMsgData, static_cast<int>( uMessageSize ), true);

		Asura_Assert(bSuccess, "Connection request message failed.");
		delete[] pcMsgData;
	}
	return true;
}

bool Asura_NetConnection::HandleNetSimSettings(const char* pcMessage, int iLength)
{
#ifdef ASURA_ENABLE_INTERNET_SIMULATION
	Asura_Assert( !Asura_Network::IsServer(), "shouldn't receive net sim settings on server");
	if (Asura_Network::IsServer())
	{
		return true;
	}

	// knock off game ver
	pcMessage += 2;
	iLength -= 2;

	if (sizeof(Asura_NetSim_Settings) == iLength)
	{
		const Asura_NetSim_Settings* pxSettings = reinterpret_cast
			<const Asura_NetSim_Settings*>(pcMessage);
		Asura_NetSim_Settings xSettingsCopy = *pxSettings;
		ASURA_NTS_CHANGE_BYTE_ORDER(xSettingsCopy.m_uLatencyMean);
		ASURA_NTS_CHANGE_BYTE_ORDER(xSettingsCopy.m_uLatencyStddev);
		ASURA_NTS_CHANGE_BYTE_ORDER(xSettingsCopy.m_fDropPercent);
		if (xSettingsCopy.m_uLatencyMean)
		{
			Asura_InternetSimulation::SetEnabled(true);
			Asura_InternetSimulation::SetSettings(xSettingsCopy);
		}
		else
		{
			Asura_InternetSimulation::SetEnabled(false);
		}
	}
	return true;
#else 
	return false;
#endif
}

bool Asura_NetConnection::HandleMessage(const char* pcMessage, int iLength)
{
	Asura_Assert(iLength >= 6, "No data in message");
	if(iLength < 6)
	{
		return true;
	}

	const unsigned short usSequenceNumber	= ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(*reinterpret_cast<const unsigned short*>(pcMessage));

	// (AAM - 11.05.2005) Unused at the moment
//	const unsigned short usChecksum			= ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(*reinterpret_cast<const unsigned short*>(pcMessage + 2));
	const unsigned short usMessageID		= ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(*reinterpret_cast<const unsigned short*>(pcMessage + 4));

	bool bReliable = usSequenceNumber != 0;

	if(dbg_bOutputAllReceivedMessages)
	{
		LogMessage(usMessageID, usSequenceNumber, ANTSCON_RECEIVED);
		Asura_OutputToDebugger(("\n", usSequenceNumber));
	}
	if(bReliable)
	{
		if(dbg_bOutputReliableMessages && !dbg_bOutputAllReceivedMessages)
		{
			LogMessage(usMessageID, usSequenceNumber, ANTSCON_RECEIVED);
		}
		if(RegisterSequenceNumber(usSequenceNumber))
		{
			if(dbg_bOutputReliableMessages)	{ Asura_OutputToDebugger(("...duplicate rejected\n", usSequenceNumber)); }
			// It's a duplicate, so ignore it.
			// (MH - 20.07.2004) malformed packets are also treated as duplicates so they get 
			// rejected, but additionally mark the connection as closed.
            // (MH - 08.01.2007) if a connection was dropped by the client without sending a close
			// a subsequent connection attempt will likely have the same serial number. So try and
			// allow for that here. The connection is likely to have been idle in that case.
			if (ASURA_NETTRANSPORT_MSG_CHALLENGE_REQUEST == usMessageID)
			{
				const u_int uMaxIdleTimeMSecs = 10000;
				if (GetIdleTime() > uMaxIdleTimeMSecs)
				{
					Asura_Network_ConnectionRejectedData xRejectionReason;
					xRejectionReason.m_uNBOReason = ASURA_NDR_CLIENT_RESET;
					ASURA_NTS_CHANGE_BYTE_ORDER(xRejectionReason.m_uNBOReason);
					LogToFile("Rejecting connection. Duplicate connection request after 10 seconds inactivity. ");
					Reject(reinterpret_cast<const char*>(&xRejectionReason), sizeof(xRejectionReason));
				}
			}
			return true;
		}
		if(dbg_bOutputReliableMessages)	{ Asura_OutputToDebugger(("...accepted\n", usSequenceNumber)); }
	}

	// Save time received for timeout checking
	m_uLastReceiveTime = Asura_Timers::GetMSecsSinceStart();

	// Knock off sequence number and checksum
	pcMessage += 4;
	iLength -= 4;

	if(usMessageID < ASURA_MSG_TRANSPORT_LAYER)
	{
		// 'Normal' Asura message
#if defined(ASURA_NETSOCKET_PLATFORM_ALIGN_HEADER) || !defined(ASURA_NTS_NETWORK_BYTE_ORDER_MATCHES_PLATFORM_BYTE_ORDER)
		const Asura_MessageHeader* pxOriginalHeader = reinterpret_cast<const Asura_MessageHeader*>(pcMessage);
		Asura_MessageHeader xHeader;
		Asura_MessageHeader* pxHeader = &xHeader;
		Asura_Mem::Copy(pxHeader, pxOriginalHeader, sizeof(Asura_MessageHeader));
		pxHeader->ChangeByteOrder();
#else
		const Asura_MessageHeader* pxHeader = reinterpret_cast<const Asura_MessageHeader*>(pcMessage);
#endif
		const int iMessageSize = pxHeader->GetDataSize() + sizeof(Asura_MessageHeader);
		Asura_Assert(iMessageSize == iLength, "data sizes do not match");
		if(iMessageSize != iLength)
		{
			return false;
		}
		
		// only add messages to queue if application is active, or if delivery was guaranteed
		//if(Asura_NetTransport_System::ShouldAddMessageToQueue(pxHeader->m_usMsgID))
		if(Asura_Network::IsReadyForMessages() || pxHeader->m_usMsgID == ASURA_MSG_VALIDCLIENTGUID || bReliable)
		{
			Asura_Network::AddMessageToQueue(pxHeader, pcMessage + sizeof(Asura_MessageHeader), bReliable);
		}
	}
	else
	{
		// Knock off identifier
		pcMessage += 2;
		iLength -= 2;

		// Transport layer message
		switch(usMessageID)
		{
			// Enum hosts and host reply are handled by socket
			// (MH - 20.07.2004) ...but only when there is no connection. In this case we have
			// a connection, and we're receiving LAN broadcast messages. This means that either
			// the messages have been delayed or there is a logic problem. We can either ignore
			// the messages or we can disconnect this client. I think ignoring is the best bet.
			case ASURA_NETTRANSPORT_MSG_ENUMHOSTS:
			case ASURA_NETTRANSPORT_MSG_HOSTREPLY:
				{
					//Asura_Assert(0, "Message should have been handled by socket."); 
					LogToFile("ignoring enum hosts or reply message for active connection");
					return true;
				}

			// We need to deal with these cases
			case ASURA_NETTRANSPORT_MSG_CONNECTION_REQUEST:		return HandleClientRequest(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_CONNECTION_ACCEPT:		return HandleAccept(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_CONNECTION_REJECT:		return HandleReject(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_CONNECTION_CLOSE:		return HandleClose(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_KEEPALIVE:				return true;
			case ASURA_NETTRANSPORT_MSG_PING:					return HandlePing(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_PONG:					return HandlePong(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_ACK:					return HandleAck(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_CHALLENGE_REQUEST:		return true;	// probably resent, just ignore.
			case ASURA_NETTRANSPORT_MSG_CHALLENGE_RESPONSE:		return HandleChallengeResponse(pcMessage, iLength);
			case ASURA_NETTRANSPORT_MSG_NETSIM_SETTINGS:		return HandleNetSimSettings(pcMessage, iLength);
			default:
				if(!Asura_NetTransport_System::HandleMessage(m_xAddress, m_usPort, pcMessage - 2, iLength + 2))
				{
					Asura_Assertf(0, ("Unrecognised message ID: %u", usMessageID));
					return false;
				}
				return true;
				
		}
	}
	return true;
}

bool Asura_NetConnection::Ping()
{
	const u_int uCurrentTime = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_Timers::GetMSecsSinceStart());
	return SendSimple(ASURA_NETTRANSPORT_MSG_PING, false, (char*)&uCurrentTime, sizeof(unsigned int));
}

bool Asura_NetConnection::RegisterSequenceNumber(unsigned short usSequenceNumber)
{
	// See comment member declaration for offset reasoning
	m_ausNBOReceivedSNs[m_uReceivedSNCount + 2] = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(usSequenceNumber);
	++m_uReceivedSNCount;
	// If the buffer is full...
	if(m_uReceivedSNCount == ASURA_NETCONNECTION_ACK_SIZE)
	{
		// ...send out an ACK
		SendAck();
	}

	if(usSequenceNumber == m_usHighestSNReceived)
	{
		// Duplicate
		return true;
	}
	bool bDuplicate = true;
	// Edit expected SNs accordingly
	// If it's higher than any previous...
	if(usSequenceNumber > m_usHighestSNReceived)
	{
		// ...then add all in-between to list
		
		// (MH - 20.07.2004) check the number isn't so high that we would overflow our buffer
		u_int uNumToAdd = usSequenceNumber - m_usHighestSNReceived;
		if (m_uNumberOfExpectedSNs + uNumToAdd > ASURA_NETCONNECTION_MAX_EXPECTED_SNS)
		{
			LogToFile("Closing connection. Reason ASURA_NETCONNECTION_MAX_EXPECTED_SNS [%d - %d - %d]", m_uNumberOfExpectedSNs, m_uNumberOfExpectedSNs + uNumToAdd, ASURA_NETCONNECTION_MAX_EXPECTED_SNS);
			// flag the connection as closed and reject this packet as if it were a duplicate.
			m_eState = ASURA_CONNECTION_CLOSED;
			return true;
		}

		for(unsigned short us = static_cast<unsigned short>(m_usHighestSNReceived + 1); us < usSequenceNumber; ++us)
		{
			m_ausExpectedSNs[m_uNumberOfExpectedSNs] = us;
			m_uNumberOfExpectedSNs++;
		}

		Asura_Assert(m_uNumberOfExpectedSNs <= ASURA_NETCONNECTION_MAX_EXPECTED_SNS, "buffer overflow");

		m_usHighestSNReceived = usSequenceNumber;
		bDuplicate = false;
	}
	else
	{
		// Check if it's in the list of expected SNs
		for(unsigned int u = 0; u < m_uNumberOfExpectedSNs; ++u)
		{
			if(m_ausExpectedSNs[u] == usSequenceNumber)
			{
				// Remove it
				m_uNumberOfExpectedSNs--;
				m_ausExpectedSNs[u] = m_ausExpectedSNs[m_uNumberOfExpectedSNs];
				m_ausExpectedSNs[m_uNumberOfExpectedSNs] = 0;
				bDuplicate = false;
				break;
			}
		}
	}
	return bDuplicate;
}

bool Asura_NetConnection::Reset()
{
	return true;
}

void Asura_NetConnection::Reject(const char* pcMessage, int iLength)
{
	SendSimple(ASURA_NETTRANSPORT_MSG_CONNECTION_REJECT, false, pcMessage, iLength);
	Close(true, ASURA_CLOSE_LOCAL);
}

bool Asura_NetConnection::Send(const void *pvMessage, int iLength, bool bReliable /* = false */)
{
	bool bSuccess = false;
	// Perform checks on the message
	Asura_Assert(pvMessage, "No message.");
	Asura_Assert(iLength >= 0, "Zero or negative size message.");
	if(!pvMessage || iLength < 0)
	{
		return false;
	}

	// Allocate buffer and check it (need extra 4 bytes for sequence number and checksum)
	char *pcFullMessage = new char[iLength + 4];
	Asura_Assert(pcFullMessage, "Failed to allocate message buffer");
	if(!pcFullMessage)
	{
		// (MH - 20.07.2004) out of memory for message storage - close connection if reliable.
		if (bReliable)
		{
			LogToFile("Out of memory for reliable message storage.");
			Close(false, ASURA_CLOSE_ERROR_BUFFERFULL);
		}
		return false;
	}

	// Sequence number is first two bytes of full message
	u_short* pusSequenceNumber = reinterpret_cast<u_short*>(pcFullMessage);

	// checksum is next 2 bytes
	u_short* pusChecksum = reinterpret_cast<u_short*>(pcFullMessage + 2);
	
	*pusChecksum = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::ComputeChecksum( reinterpret_cast<const char*>(pvMessage), iLength ));

	if(bReliable)
	{
		*pusSequenceNumber = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(GetNextValidSequenceNumber());

		// Save to buffer
		if(!m_xOutgoingBuffer.AddMessage(Asura_Timers::GetMSecsSinceStart(), pcFullMessage, iLength + 4))
		{
			// (MH - 27.06.2007) message wasn't added so it won't be automatically deleted. Do it here.
			// Also added return false as we don't want to send it.
			delete[] pcFullMessage;
			// Problem queuing message. Close connection
			LogToFile("Connection closed due to outgoing buffer overflow");
			Close(false, ASURA_CLOSE_ERROR_BUFFERFULL);
			return false;
		}
	}
	else
	{
		*pusSequenceNumber = 0;
	}

	Asura_Mem::Copy(pcFullMessage + 4, pvMessage, iLength);
	if(!m_pxSocket->Send(m_xAddress, m_usPort, pcFullMessage, iLength + 4))
	{
		if(!bReliable)
		{
			delete [] pcFullMessage;
		}
		int iLastError = m_pxSocket->API_GetLastError();
		switch(iLastError)
		{
			case WSAECONNRESET:
				return Reset();
			case WSAEHOSTUNREACH:
				LogToFile("Host unreachable");
				return Close(true, ASURA_CLOSE_ERROR_HOSTUNREACHABLE);
			case WSAEMSGSIZE:
				// We can't send this message, so it's probably going to break things...close the connection
				LogToFile("Message too large.");
				Asura_Assert(0, "Asura_NetConection : Message too large.");
				return Close(false, ASURA_CLOSE_ERROR_MESSAGETOOLARGE);
			case WSAENOBUFS:
				break;
			default:
			{
				// (MH - 23.02.2005) fatal error is a bit extreme, we have been getting 0 here.
				// I think that means "timeout". That's non-fatal anyway. 
				Asura_Assertf( false, ("unhandled error from send (%d)", iLastError) );
				if (iLastError != 0)
				{
					//m_pxSocket->FatalError();
					return Close(true, ASURA_CLOSE_ERROR_SOCKETERROR);
				}
				break;
			}
		}
		return false;
	}
	else
	{
		if (dbg_bOutputSentMessages)
		{
			u_short usMessageID = *reinterpret_cast<const short*>(pvMessage);
			LogMessage(usMessageID, ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(*pusSequenceNumber), ANTSCON_SENT);
		}

		Asura_Network_MachineInfo* pxMachineInfo = Asura_NetTransport_System::GetMachineInfo(this);
		if(pxMachineInfo)
		{
			pxMachineInfo->AddSentBytes(static_cast<u_int>(iLength + 4));
		}
		bSuccess = true;
	}
	if(!bReliable)
	{
		delete [] pcFullMessage;
	}
	return bSuccess;
}

// helper function for dealing with reserved sequence numbers (see comment at top of file)
unsigned short Asura_NetConnection::GetNextValidSequenceNumber()
{
	bool bValid = false;

	while(!bValid)
	{
		// Wrap at maximum, and then skip
		if(m_usSequenceNumber == 0xffff) m_usSequenceNumber = 0;
		++m_usSequenceNumber;
		bValid = true;

		// if this is one of the reserved sequence numbers, skip it
		for(u_int uReservedNumber = 0; uReservedNumber < g_uNumberOfReservedSN; ++uReservedNumber)
		{
			if(g_ausReservedSN[uReservedNumber] == m_usSequenceNumber)
			{
				bValid = false;
			}
		}
	}

	return m_usSequenceNumber;
}

// Helper function for sending an ID + optional data
bool Asura_NetConnection::SendSimple(unsigned short usID, bool bReliable, const char* pcData /* = NULL */, int iLength /* = 0 */)
{
	if(!pcData || !iLength)
	{
		ASURA_NTS_CHANGE_BYTE_ORDER(usID);
		return Send(&usID, 2, bReliable);
	}
	else
	{
		// Build message data
		Asura_Assert(iLength + 4 <= int(ASURA_NETCONNECTION_MAX_SIMPLEMESSAGE_LENGTH), "Simple message buffer overflow.");
		if (iLength + 4 > int(ASURA_NETCONNECTION_MAX_SIMPLEMESSAGE_LENGTH))
		{
			return false;
		}
		u_short ausMessageData[ASURA_NETCONNECTION_MAX_SIMPLEMESSAGE_LENGTH/2];
		unsigned short* pusMessageData = ausMessageData;
		*pusMessageData++ = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(usID);
		*pusMessageData++ = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(Asura_NetTransport_System::Project_GetVersionID());
		Asura_Mem::Copy(ausMessageData + 2, pcData, iLength);
		bool bSuccess = Send(ausMessageData, iLength + 4, bReliable);
		return bSuccess;
	}
}

bool Asura_NetConnection::SendAck()
{
	// Don't bother unless we've actually received any guaranteed messages.
	bool bSuccess = true;
	if(m_uReceivedSNCount)
	{
#ifdef _DEBUG
		for(u_int u = 0; u < m_uReceivedSNCount; ++u)
		{
			Asura_Assert(ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(m_ausNBOReceivedSNs[u + 2]), "Invalid serial number.");
		}
#endif
		// Received SNs array includes ack identifier as first element, so it can be sent without modification
		bSuccess = Send((char*)m_ausNBOReceivedSNs, (m_uReceivedSNCount + 2) * sizeof(unsigned short));
		// Reset count
		m_uReceivedSNCount = 0;
	}
	return bSuccess;
}

bool Asura_NetConnection::SendKeepAlive()
{
	return SendSimple(ASURA_NETTRANSPORT_MSG_KEEPALIVE, false);
}

bool Asura_NetConnection::SendClosed(Asura_Close_Reason eCloseReason)
{
	Platform_CloseConnection();

	LogToFile("Sending close message. reason [%d]", eCloseReason);
	const u_short usCloseReason = ASURA_NTS_CHANGE_BYTE_ORDER_RETVAL(static_cast<u_short>(eCloseReason));
	return SendSimple(ASURA_NETTRANSPORT_MSG_CONNECTION_CLOSE, false, reinterpret_cast<const char*>(&usCloseReason), sizeof(usCloseReason));
}

bool Asura_NetConnection::SendRetries()
{
	bool bSuccess = true;
	unsigned int uCurrentTime = Asura_Timers::GetMSecsSinceStart();
	for(unsigned int u = 0; u < m_xOutgoingBuffer.GetNumberOfEntries(); ++u)
	{
		const Asura_Network_Buffer_Entry *pxEntry = m_xOutgoingBuffer.GetMessage(u);
		if(uCurrentTime - pxEntry->GetSendTime() > ASURA_NETTRANSPORT_RESEND_PERIOD)
		{
			bSuccess &= m_pxSocket->Send(m_xAddress, m_usPort, pxEntry->GetBuffer(), pxEntry->GetBufferSize());

			if (bSuccess)
			{
				// (MH - 19.07.2004) add the resent data to the bandwidth stats.
				Asura_Network_MachineInfo* pxMachineInfo = Asura_NetTransport_System::GetMachineInfo(this);
				if(pxMachineInfo)
				{
					pxMachineInfo->AddSentBytes(static_cast<u_int>(pxEntry->GetBufferSize()));
					
					pxMachineInfo->IncrementNumRetries();
				}

                // (MH - 09.11.2006) output the retry.
				if (dbg_bOutputSentMessages)
				{
					const unsigned short usSequenceNumber	= *reinterpret_cast<const unsigned short*>(pxEntry->GetBuffer());
					const unsigned short usMessageID		= *reinterpret_cast<const unsigned short*>(pxEntry->GetBuffer() + 4);
					LogMessage(usMessageID, usSequenceNumber, ANTSCON_RETRY);
				}

			}
		}
	}
	return bSuccess;
}

bool Asura_NetConnection::SetClosed(float fTime)
{
	Asura_Assert(GetState() == ASURA_CONNECTION_CONNECTED, "connection not connected, but SetClosed() called");

	m_eState = ASURA_CONNECTION_CLOSING;
	m_fClosingTimer = fTime;
	return true;
}

u_int Asura_NetConnection::GetIdleTime() const
{ 
	unsigned int uCurrentTime = Asura_Timers::GetMSecsSinceStart();

	return (uCurrentTime - m_uLastReceiveTime); 
}

bool Asura_NetConnection::HasTimedOut() const
{
	return Asura_NetTransport_System::CanConnectionsTimeOut() && (GetIdleTime() > Asura_NetTransport_System::GetConnectionsTimeOutLimit());
}

bool Asura_NetConnection::Update()
{
	Asura_Assert(IsActive(), "Net connection being used when not active");
	unsigned int uCurrentTime = Asura_Timers::GetMSecsSinceStart();
	
	if(HasTimedOut())
	{
		LogToFile("Connection Timed Out [%d secs / %d secs]", GetIdleTime() / 1000, Asura_NetTransport_System::GetConnectionsTimeOutLimit() / 1000);
		
		if(!m_bInitiated)
		{
			LogToFile("Closing connection : not initiated");
			m_eCloseReason = ASURA_CLOSE_ERROR_COMMUNICATION;
			Asura_NetTransport_System::Server_ClientDisconnected(this);
		}
		Close(false, ASURA_CLOSE_ERROR_COMMUNICATION);
	}
	if(uCurrentTime - m_uLastKeepAliveSendTime > ASURA_KEEPALIVE_PERIOD)
	{
		SendKeepAlive();
		m_uLastKeepAliveSendTime = uCurrentTime;
	}
	// Send acks twice as often as message re-sends
	if(uCurrentTime - m_uLastAckSendTime > ASURA_NETTRANSPORT_RESEND_PERIOD / 2)
	{
		SendAck();
		m_uLastAckSendTime = uCurrentTime;		
	}
	if(uCurrentTime - m_uLastResendTime > ASURA_NETTRANSPORT_RESEND_PERIOD)
	{
		SendRetries();
		m_uLastResendTime = uCurrentTime;
	}

	if(m_eState == ASURA_CONNECTION_CLOSING)
	{
		m_fClosingTimer -= Asura_Timers::GetFrameTime();
		if (m_fClosingTimer <= 0.0f)
		{
			Close( true, ASURA_CLOSE_LOCAL );
		}
	}

	return true;
}

void Asura_NetConnection::CheckBufferIntegrity()
{
	m_xOutgoingBuffer.CheckIntegrity();
}

void Asura_NetConnection::LogMessage(u_short usMessageID, u_short usSN, MsgLogType eType)
{
#if _DEBUG
	const char* szType;
	switch (eType)
	{
	case ANTSCON_SENT:
		szType = "sent";
		break;
	case ANTSCON_RECEIVED:
		szType = "received";
		break;
	case ANTSCON_RETRY:
		szType = "retry send";
		break;
	default:
		szType = "";
		break;
	}

#if ASURA_NETWORK_LOGGING
	const char* szDesc = 0;
	if(!Asura_NetTransport_System::GetMsgIDDesc(usMessageID, &szDesc))
	{
		Asura_Network::GetMsgIDDesc(usMessageID, &szDesc);
	}
	if(szDesc)
	{
		Asura_OutputToDebugger(("SN %u, ID %s (0x%x) %s\n", usSN, szDesc, usMessageID, szType));
	}
	else
#endif
	{
		Asura_OutputToDebugger(("SN %u, ID %u (0x%x) %s\n", usSN, usMessageID, usMessageID, szType));
	}
#endif // _DEBUG
}

void Asura_NetConnection::ClearClientRequestData()
{
	if( m_pcClientRequestData )
	{
		delete []m_pcClientRequestData;
		m_pcClientRequestData = 0;
		--s_uNumSavedClientRequests;
	}
}

#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
