//////////////////////////////////////////////////////////////////////////////
//
// Asura_NetMsg.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     system to help reduce amount of data sent over external network for
//     regular entity updates.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// required Asura include
#include "Core/Asura.h"

// this file's include
#include "Network/Asura_NetMsg.h"

#ifdef ASURA_SUPPORT_NETMSG_SYSTEM

// asura independent includes
#include "Core/Asura_BitPacker.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Logfile.h"
#include "Entity/Asura_Entity.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"	// for player position hack
#include "Network/Asura_Message.h"
#include "Network/Asura_MessageHeader.h"
#include "Network/Asura_Network.h"
#include "Rendering/Asura_Module.h"

// platform includes
#include "Core/Asura_PlatformBasics.h"	// for Asura_UnicodeChar
// Library includes
#include <wchar.h> // for wcslen

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uASURA_NETMSG_UPDATE_PACKAGE_MAX_SIZE = 1*1024;
// was 16384 bytes which is all very well on LAN, but not for online

// allow projects to change num bits used for field index, default is 5
#ifndef ASURA_NETMSG_BITS_FOR_FIELD_INDEX
#define ASURA_NETMSG_BITS_FOR_FIELD_INDEX 5
#endif
static const u_int uBITS_FOR_FIELD_INDEX	= ASURA_NETMSG_BITS_FOR_FIELD_INDEX;
static const u_int uFIELDID_FULLUPDATE		= (1 << uBITS_FOR_FIELD_INDEX) - 1;
static const u_int uFIELDID_ENDOFDATA		= (1 << uBITS_FOR_FIELD_INDEX) - 2;
static const u_int uMAX_FIELDID				= (1 << uBITS_FOR_FIELD_INDEX) - 3;

// defines to provide limited range of player guids which actually get used.
static const u_int uNUM_BITS_FOR_PLAYER_GUID = 5;
static const u_int uFIRST_PLAYER_GUID = ASURA_GUID_CLIENT_START;//1000
static const u_int uLAST_PLAYER_GUID = uFIRST_PLAYER_GUID + (1 << uNUM_BITS_FOR_PLAYER_GUID) - 1;//1031

// number of bits in Asura_Comp_Quat_Fast
const u_int uNUM_BITS_FOR_COMP_QUAT_FAST = 32;

// if a message with msgID ASURA_MSG_INVALID is found, this is used to indicate a time offset
static const u_short usTIME_OFFSET_MESSAGE_ID = ASURA_MSG_INVALID;
static const u_int uBITS_FOR_TIME_OFFSET = 10;
static const u_int uMAX_TIME_OFFSET = (1 << uBITS_FOR_TIME_OFFSET) - 1;

static const u_int uASURA_NETMSG_NUM_LOD_THRESHOLDS = 4;

static const u_int uMAX_BYTE_OFFSET = USHRT_MAX;

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR bool dbg_bDisableAsuraNetMsgSystem = false;
DEBUGVAR u_int dbg_uMaxInitialHeartbeatTimerValue = 2000;	// in milliseconds
DEBUGVAR u_int dbg_uMaxFullUpdateBatchSize = 200;
DEBUGVAR float dbg_fQuatApproximationThreshold = 0.001f;
DEBUGVAR float dbg_fCompQuatFastApproximationThreshold = 2.5e-6f;
DEBUGVAR float dbg_afNetMsgLodThresholds[uASURA_NETMSG_NUM_LOD_THRESHOLDS] = { 
	25.0f*25.0f,	// within 25m don't drop any packets
	200.0f*200.0f,  // within 200m drop 50% (1/2) of packets, barely noticeable
	400.0f*400.0f,	// within 400m drop 66% (2/3) of packets, noticeable
	500.0f*500.0f };// within 500m drop 75% (3/4) of packets, very choppy

// G L O B A L S /////////////////////////////////////////////////////////////
static u_int g_uInitialHeartbeatTimerValue;
static u_int g_uFullUpdateBatchSize;

// C L A S S E S /////////////////////////////////////////////////////////////

// Asura_NetMsg_UpdateData contains all the data needed to build the packet
// for each client receiving updates
class Asura_NetMsg_UpdateData
{

	// private and unimplemented copy constructor and operator= to prevent 
	// copying of internal pointer.
	Asura_NetMsg_UpdateData(const Asura_NetMsg_UpdateData&);
	Asura_NetMsg_UpdateData& operator= (const Asura_NetMsg_UpdateData&);

public:
	
	Asura_NetMsg_UpdateData()
	{
		m_pxBitPacker = new Asura_BitPacker(uASURA_NETMSG_UPDATE_PACKAGE_MAX_SIZE);
		m_bPositionValid = false;
		m_iModuleIndex = -1;
		m_uGuid = ASURA_GUID_UNREGISTERED;
	}

	~Asura_NetMsg_UpdateData()
	{
		delete m_pxBitPacker;
	}

	void Restart(Asura_Guid uPlayerGuid)
	{
		m_uNumMessages = 0;
		m_uNumTimeOffsets = 0;
		m_bNeedToAddTimeOffset = false;
		
		m_pxBitPacker->Reset();

		// skip a byte to store number of messages later
		m_pxBitPacker->SkipBytes(1);

		m_fSentTimeForCurrentPacket = Asura_Timers::GetServerAbsoluteTime();

		// find the player entity from the guid
		// bit of a hack to get the player position, relies on player entities being derived from physical object.
		const Asura_ServerEntity_PhysicalObject* pxPlayer = Asura_ServerEntity_PhysicalObject::FindEntity(uPlayerGuid);
		// this could fail if server entity hasn't been made yet.
		// (MH - 20.03.2005) also not doing this optimisation if player is dead (this handles
		// things like spectating and waiting to respawn etc.)
		if (pxPlayer && !pxPlayer->IsDead())
		{
			m_bPositionValid = true;
			m_xPosition = pxPlayer->GetPosition();
			m_iModuleIndex = Asura_Module_System::GetModuleIndexFromPosition(m_xPosition, m_iModuleIndex);
		}
		else
		{
			m_bPositionValid = false;
			m_iModuleIndex = -1;
		}

		m_uGuid = uPlayerGuid;
	}

	void NewFrame()
	{
		if (m_uNumMessages)
		{
			m_bNeedToAddTimeOffset = true;
		}
		else
		{
			// no messages using the original start time - just increase it.
			m_fSentTimeForCurrentPacket = Asura_Timers::GetServerAbsoluteTime();
		}
	}

	void Send(Asura_Guid uTo, Asura_Guid uFrom)
	{
		// if no messages were added this frame, there's nothing to do
		if (!m_uNumMessages) return;

		// num msgs goes at the start of the bit packer
		Asura_Assert(m_uNumMessages + m_uNumTimeOffsets < 256, "num messages overflowed");
		m_pxBitPacker->SetByte(0, static_cast<u_char>(m_uNumMessages + m_uNumTimeOffsets));

		// now send the data to the client
		u_int uDataSize;
		const void* pData = m_pxBitPacker->GetResult( uDataSize );
		
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_NETMSGUPDATE, uDataSize, 0, pData);
		Asura_Assert(pxMessage, "failed to create message");
		if (!pxMessage) return;

		pxMessage->To = uTo;
		pxMessage->From = uFrom;
		pxMessage->SentTime = m_fSentTimeForCurrentPacket;

		// destroy the message after sending, but don't change the sent time
		bool bDestroyMessage = true;
		bool bSetSentTime = false;
		Asura_Network::Send(pxMessage, bDestroyMessage, bSetSentTime);
	}

	void AddMessageHeader( void )
	{
		if (m_bNeedToAddTimeOffset)
		{
			AddTimeOffset();
		}
		
		u_short usMsgID = Asura_NetMsg_System::s_pxAsuraMessageToAdd->GetMsgID();
		u_short usEntityClassification = Asura_NetMsg_System::s_pxAsuraMessageToAdd->GetEntityClassification();
		Asura_Guid uGuid = Asura_NetMsg_System::s_pxAsuraMessageToAdd->From;

		// MsgID
		bool bHasMsgID = usMsgID != ASURA_MSG_CLIENTUPDATE;
		m_pxBitPacker->PackBool(bHasMsgID);
		if (bHasMsgID)
		{
			// compress the msgID
			Asura_MessagePacking::PackMsgID( usMsgID, m_pxBitPacker );
		}

		// EntityClassification 
		Asura_Assert(Asura_NetMsg_System::s_usEntityClassificationPlayer != AsuraEntityClass_NotSet, "player entity classification not set. You need to call Asura_NetMsg_System::SetPlayerEntityClassForProject()");
		bool bHasEntityClassification = usEntityClassification != Asura_NetMsg_System::s_usEntityClassificationPlayer;
		m_pxBitPacker->PackBool(bHasEntityClassification);
		if (bHasEntityClassification)
		{
			// compress the EntityClassification
			Asura_MessagePacking::PackEntityClassification( usEntityClassification, m_pxBitPacker );
		}
		
		// Guid
		// most games will have player guids in the range 1000-1031. once this limit is
		// exceeded for players bandwidth will go up. If that starts happening, you may
		// want to change the code so that earlier guids get recycled.
		bool bGuidNotInRange1000to1031 = (uGuid < uFIRST_PLAYER_GUID || uGuid > uLAST_PLAYER_GUID);
		m_pxBitPacker->PackBool(bGuidNotInRange1000to1031);
		if (bGuidNotInRange1000to1031)
		{
			Asura_MessagePacking::PackGuid( uGuid, m_pxBitPacker );
		}
		else
		{
			// pack range into 5 bits
			u_int uCompressedGuid = uGuid - uFIRST_PLAYER_GUID;
			m_pxBitPacker->Pack( uCompressedGuid, uNUM_BITS_FOR_PLAYER_GUID );
		}

		m_uNumMessages++;
	}

	void AddTimeOffset( void )
	{
		// add time offset from start. This is done by adding a special message ID,
		// followed by a set number of bits for the number of milliseconds elapsed
		float fTimeOffset = Asura_Timers::GetServerAbsoluteTime() - m_fSentTimeForCurrentPacket;
		Asura_Assert(fTimeOffset >= 0.0f, "negative time offset");
		if (fTimeOffset > 0.0f)
		{
			if (RoomForBits(1 + Asura_MessagePacking::GetBitsRequiredForMsgID(usTIME_OFFSET_MESSAGE_ID) + uBITS_FOR_TIME_OFFSET))
			{
				u_int uTimeOffset = Asura_Maths::FloatToInt(fTimeOffset * 1000.0f);
				if (uTimeOffset > uMAX_TIME_OFFSET)
				{
					Asura_OutputToDebugger(("clamping time offset for NetMsg.\n"));
					uTimeOffset = uMAX_TIME_OFFSET;
				}
				m_pxBitPacker->PackBool(true);	// bHasMsgID
				Asura_MessagePacking::PackMsgID(usTIME_OFFSET_MESSAGE_ID, m_pxBitPacker);
				m_pxBitPacker->Pack(uTimeOffset, uBITS_FOR_TIME_OFFSET);

				m_uNumTimeOffsets++;
			}
			else
			{
				Asura_Assert(false, "no room to add time offset");
			}
		}

		m_bNeedToAddTimeOffset = false;
	}


	bool RoomForBytes(u_int uNumBytes) const
	{
		return m_pxBitPacker->RoomForBytes(uNumBytes);
	}


	bool RoomForBits(u_int uNumBits) const
	{
		return m_pxBitPacker->RoomForBits(uNumBits);
	}


	void AddBytes(const u_char* pucData, u_int uDataSize)
	{
		Asura_Assert(RoomForBytes(uDataSize), "CRASH BUG: no room for data");//should have been checked already
		while (uDataSize--)
		{
			m_pxBitPacker->Pack(*pucData++, uASURA_NETMSG_BITS_PER_BYTE);
		}
	}


	void AddBits(u_int uValue, u_int uNumBits)
	{
		Asura_Assert(RoomForBits(uNumBits), "CRASH BUG: no room for data");//should have been checked already
		m_pxBitPacker->Pack(uValue, uNumBits);
	}

	void AddGuid(Asura_Guid uGuid)
	{
		Asura_Assert(RoomForBits(32), "CRASH BUG: no room for data");//should have been checked already
		Asura_MessagePacking::PackGuid( uGuid, m_pxBitPacker);
	}

	bool IsPositionValid() const { return m_bPositionValid; }
	const Asura_Vector_3& GetPosition() const { return m_xPosition; }
	int GetModuleIndex() const { return m_iModuleIndex; }
	Asura_Guid GetGuid() const { return m_uGuid; }

	// for testing
	void ResetBitPacker()
	{
		m_pxBitPacker->Reset();
	}
	void GetUnpackData( u_int& uDataSize, const void*& pData )
	{
		pData = m_pxBitPacker->GetResult( uDataSize );
	}

private:
	// data

	Asura_BitPacker* m_pxBitPacker;
	u_int m_uNumMessages;
	u_int m_uNumTimeOffsets;
	float m_fSentTimeForCurrentPacket;
	bool m_bNeedToAddTimeOffset;
	bool m_bPositionValid;
	Asura_Vector_3 m_xPosition;
	int m_iModuleIndex;
	Asura_Guid m_uGuid;
};

// Asura_NetMsg_Recipients is just a wrapper around calling AddBytes etc. for an array
// of Asura_NetMsg_UpdateData.
class Asura_NetMsg_Recipients
{
public:
	Asura_NetMsg_Recipients(Asura_NetMsg_UpdateData** pxArray, u_int uArraySize)
		: m_pxArray(pxArray)
		, m_uArraySize(uArraySize)
	{}

	void AddBits(u_int uValue, u_int uNumBits)
	{
		for (u_int u = 0; u < m_uArraySize; ++u)
		{
			m_pxArray[u]->AddBits(uValue, uNumBits);
		}
	}

	void AddGuid(Asura_Guid uGuid)
	{
		for (u_int u = 0; u < m_uArraySize; ++u)
		{
			m_pxArray[u]->AddGuid(uGuid);
		}
	}

	void ChangeArraySize( u_int uNewSize )
	{
		m_uArraySize = uNewSize;
	}

private:

	Asura_NetMsg_UpdateData** m_pxArray;
	u_int m_uArraySize;
};

class Asura_NetMsg_UprightQuatData
{
public:

	void FromQuat( const Asura_Quat& xQuat )
	{
		Asura_Assert(Asura_Maths::Abs(xQuat.x) < 1e-6f && Asura_Maths::Abs(xQuat.z) < 1e-6f, "not an upright quat");

		Asura_Euler_3 xEuler;
		xQuat.ToEuler(xEuler);
		Asura_Assert(Asura_Maths::Abs(xEuler.x) < 1e-6f && Asura_Maths::Abs(xEuler.z) < 1e-6f, "not an upright euler");
		float fEulerY = xEuler.y;

		// convert euler y to [0.0f,2pi)
		while (fEulerY < 0.0f)
		{
			fEulerY += Asura_Maths::Deg360;
		}
		while (fEulerY >= Asura_Maths::Deg360)
		{
			fEulerY -= Asura_Maths::Deg360;
		}
		// convert to [0.0f,1.0f)
		fEulerY /= Asura_Maths::Deg360;
		// convert to [0, 65535]
		fEulerY *= 32768.0f;
		Asura_Assert(Asura_Maths::FloatToInt_Chop(fEulerY) < 32768, "u_short overflow");
		m_usEulerY = static_cast<u_short>(Asura_Maths::FloatToInt_Chop(fEulerY));
		m_bQuatYNegative = xQuat.y < 0.0f;
	}

	void ToQuat( Asura_Quat& xQuat ) const
	{
		float fEulerY = static_cast<float>(m_usEulerY);
		fEulerY /= 32768.0f;
		fEulerY *= Asura_Maths::Deg360;

 		Asura_Euler_3 xEuler(0.0f, fEulerY, 0.0f);
		xQuat.FromEuler( xEuler );
		Asura_Assert(Asura_Maths::Abs(xQuat.x) < 1e-6f && Asura_Maths::Abs(xQuat.z) < 1e-6f, "not an upright quat");

		bool bQuatYNegative = xQuat.y < 0;
		if (m_bQuatYNegative != bQuatYNegative)
		{
			xQuat.y = -xQuat.y;
			xQuat.w = -xQuat.w;
		}
	}
	
	union
	{
		struct 
		{
			u_short m_bQuatYNegative:1;
			u_short m_usEulerY:15;
		};
		u_short m_usData;
	};
};

class Asura_NetMsg_CompressedPositions
{
public:
	static bool FromVector( const Asura_Vector_3& xPos, u_int& uX, u_int& uY, u_int &uZ )
	{
		bool bResult = true;
		bResult &= Compress( s_fMaxX, xPos.x, uX );
		bResult &= Compress( s_fMaxY, xPos.y, uY );
		bResult &= Compress( s_fMaxZ, xPos.z, uZ );

		return bResult;
	}

	static void ToVector( Asura_Vector_3& xPos, const u_int uX, const u_int uY, const u_int uZ )
	{
		Decompress( s_xCPS.m_uBitsX, uX, xPos.x );
		Decompress( s_xCPS.m_uBitsY, uY, xPos.y );
		Decompress( s_xCPS.m_uBitsZ, uZ, xPos.z );
	}

	static bool SetCompression( const Asura_NetMsg_CompressedPositionSettings& xSettings )
	{
		bool bResult = true;

		if ((xSettings.m_uBitsX + xSettings.m_uBitsF + s_uBitsS > 32)
			|| (xSettings.m_uBitsY + xSettings.m_uBitsF + s_uBitsS > 32)
			|| (xSettings.m_uBitsZ + xSettings.m_uBitsF + s_uBitsS > 32) )
		{
			Asura_Assert( false, "Invalid compression settings" );
			bResult = false;
		}
		else
		{
			s_xCPS = xSettings;

			s_fMaxX = CalculateMaxValue( s_xCPS.m_uBitsX );
			s_fMaxY = CalculateMaxValue( s_xCPS.m_uBitsY );
			s_fMaxZ = CalculateMaxValue( s_xCPS.m_uBitsZ );
			s_fMaxF = CalculateMaxValue( s_xCPS.m_uBitsF );
		}
		
		return bResult;
	}

	static const Asura_NetMsg_CompressedPositionSettings& GetCompression()
	{
		return s_xCPS;
	}

	static u_int GetNumBitsX() { return s_uBitsS+s_xCPS.m_uBitsX+s_xCPS.m_uBitsF; }
	static u_int GetNumBitsY() { return s_uBitsS+s_xCPS.m_uBitsY+s_xCPS.m_uBitsF; }
	static u_int GetNumBitsZ() { return s_uBitsS+s_xCPS.m_uBitsZ+s_xCPS.m_uBitsF; }

	static float GetMaxX() { return s_fMaxX; }
	static float GetMaxY() { return s_fMaxY; }
	static float GetMaxZ() { return s_fMaxZ; }
	
private:
	static bool Compress( const float fMaxInteger, const float fIn, u_int &uOut )
	{
		uOut = (fIn < 0.0f) ? 1 : 0;
		float fIntegerPart;
		float fFractionalPart = modff( fIn, &fIntegerPart );
		if (uOut) //if fIn is negative
		{
			fIntegerPart = -fIntegerPart;
			fFractionalPart = -fFractionalPart;
		}

		Asura_Assert(fFractionalPart >= 0.0f && fFractionalPart < 1.0f, "modff behaved unexpectedly");
		fFractionalPart *= s_fMaxF;
		uOut |= (Asura_Maths::FloatToInt_Chop(fFractionalPart) << s_uBitsS);

		Asura_Assertf(fIntegerPart < fMaxInteger, ("Position %f out of range for NetMsg compression", fIn));
		// clamp to within bounds. This will prevent objects wrapping around the edge of the world.
		const float fIntegerClamp = fMaxInteger - 1.0f;
		fIntegerPart = Asura_Maths::Min(fIntegerPart, fIntegerClamp);
		uOut |= (Asura_Maths::FloatToInt_Chop(fIntegerPart) << (s_uBitsS + s_xCPS.m_uBitsF));

		return fIntegerPart < fMaxInteger;
	}

	static void Decompress( const u_int uIntegerBits, const u_int uIn, float& fOut )
	{
		float fFractionalPart = static_cast<float>((uIn >> s_uBitsS) & ((1 << s_xCPS.m_uBitsF) - 1));
		fFractionalPart /= s_fMaxF;
		fOut = static_cast<float>((uIn >> (s_uBitsS + s_xCPS.m_uBitsF)) & ((1 << uIntegerBits) - 1));
		fOut += fFractionalPart;
		if (uIn & s_uBitsS)
		{
			fOut = -fOut;
		}
	}

	static float CalculateMaxValue( u_int uBits)
	{
		return static_cast<float>(1 << uBits);
	}

	static Asura_NetMsg_CompressedPositionSettings s_xCPS;
	const static u_int s_uBitsS; //Sign

	static float s_fMaxX;
	static float s_fMaxY;
	static float s_fMaxZ;
	static float s_fMaxF;
};

Asura_NetMsg_CompressedPositionSettings Asura_NetMsg_CompressedPositions::s_xCPS = {0,0,0,0};
const u_int Asura_NetMsg_CompressedPositions::s_uBitsS = 1; //Sign

float Asura_NetMsg_CompressedPositions::s_fMaxX = 0;
float Asura_NetMsg_CompressedPositions::s_fMaxY = 0;
float Asura_NetMsg_CompressedPositions::s_fMaxZ = 0;
float Asura_NetMsg_CompressedPositions::s_fMaxF = 0;

// (MH - 03.03.2004) quick test cases for compressed position
//void TestCompressedPosition()
//{
//	const u_int uNumTests = 5;
//	Asura_Vector_3 axVector[uNumTests] = { 
//		Asura_Vector_3 (1.0f, 1.0f, 1.0f),
//		Asura_Vector_3 (256.0f, 32.0f, 256.0f),
//		Asura_Vector_3 (255.9f, 31.9f, 255.9f),
//		Asura_Vector_3 (-1.0f, -31.0f, -1.0f),
//		Asura_Vector_3 (-63.296f, 17.6f, -253.5f),
//	};
//
//	for (u_int u = 0; u < uNumTests; ++u)
//	{
//		u_int uX, uY, uZ = 0;
//		const Asura_Vector_3 xVector = axVector[u];
//		Asura_NetMsg_CompressedPositions::FromVector( xVector, uX, uY, uZ );
//		Asura_Vector_3 xVector2(0,0,0);
//		Asura_NetMsg_CompressedPositions::ToVector( xVector2, uX, uY, uZ );
//		Asura_OutputToDebugger(("{%.04f, %.04f, %.04f} compressed to {%.04f, %.04f, %.04f}.\n", xVector.x, xVector.y, xVector.z, xVector2.x, xVector2.y, xVector2.z));
//	}
//}

class Asura_NetMsg_MasterInstance
{
	// private and unimplemented copy constructor and operator = to prevent copying of ptr
	Asura_NetMsg_MasterInstance(const Asura_NetMsg_MasterInstance&);
	Asura_NetMsg_MasterInstance& operator= (const Asura_NetMsg_MasterInstance&);

public:

	Asura_NetMsg_MasterInstance(Asura_Guid uGuid, const Asura_NetMsg* pxNetMsg, float fHeartbeatInterval, const void* pInitialData);
	~Asura_NetMsg_MasterInstance();

	u_char* GetData() { return m_pucCurrentData; }
	bool SendDataUpdate( const Asura_Message* pxMsg );	// returns false if not enough room to pack data

	// returns number of bytes used for the instance header, i.e. msgID, EntClass, guid, datasize, flags
	u_int GetHeaderSize();

	// required to use Asura_HashedList
	Asura_Guid GetGuid() const { return m_uGuid; }
	bool IsDestroyed() const { return false; }

	// Update timer so that we will send a full update in the next message.
	void SendFullUpdate()
	{
		m_uTimeOfNextFullUpdate = Asura_Timers::GetMSecsSinceStart();
	}

	// After creation, delay the first full update by one heartbeat.
	void AdvanceTimerByOneHeartbeat()
	{
		m_uTimeOfNextFullUpdate += m_uHeartbeat;
	}

protected:
	
	// internal functions called by SendDataUpdate
	void DoFieldUpdate( u_int uFieldIndex, bool bIsFullUpdate );
	
	// data
	const Asura_Guid m_uGuid;
	const Asura_NetMsg* const m_pxNetMsg;
	u_char* m_pucCurrentData;
	u_char* m_pucLastSentData;
	char* m_pcFieldChangedLastSend;
	u_int m_uTimeOfNextFullUpdate;	// in milliseconds
	const u_int m_uHeartbeat;		// time between full updates
	bool m_bSentAnythingLastUpdate;

	u_char* m_aucNumDropped;	// for each client machine (i.e. player), how many updates have been skipped due to distance.
};


class Asura_NetMsg_SlaveInstance
{

public:

	Asura_NetMsg_SlaveInstance(Asura_Guid uGuid, const Asura_NetMsg* pxNetMsg);
	~Asura_NetMsg_SlaveInstance();

	void ProcessReceivedData(float fSentTime, Asura_BitUnpacker& xBitUnpacker, bool bIsFirstUpdate);
	void SendMessage(const u_char* pucData, float fSentTime);

	// required to use Asura_HashedList
	Asura_Guid GetGuid() const { return m_uGuid; }
	bool IsDestroyed() const { return false; }

	u_char* GetData() { return m_pucCurrentData; }

protected:
	
	// member data
	const Asura_Guid m_uGuid;
	const Asura_NetMsg* const m_pxNetMsg;
	u_char* m_pucCurrentData;

};


// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
u_char Asura_NetMsg::s_ucMaxFields = 0;
u_int Asura_NetMsg::s_uRegistrationBaseOffset = 0;

// static class data - system-wide
bool Asura_NetMsg_System::s_bIsInitialised;
bool Asura_NetMsg_System::s_bLODUpdatesAllowed;
bool Asura_NetMsg_System::s_bIsUsingLODUpdates;
u_short Asura_NetMsg_System::s_usEntityClassificationPlayer;
u_int Asura_NetMsg_System::s_uMaxDistinctClientUpdates;
u_int Asura_NetMsg_System::s_uNumClientsForThisUpdate;
const Asura_Message* Asura_NetMsg_System::s_pxAsuraMessageToAdd;
bool Asura_NetMsg_System::s_bCanAddMessages;

// project callback for netmsg filtering
bool (*Asura_NetMsg_System::s_pfnProjectShouldSendClientMessage)(Asura_Guid, const Asura_Message*) = NULL;

// before a message is processed, this array is filled out with the clients who are interested in it.
class Asura_NetMsg_UpdateData** Asura_NetMsg_System::s_apxClientsForCurMessage;
u_int Asura_NetMsg_System::s_uNumClientsForCurMessage;

// data for each updated client
Asura_NetMsg_UpdateData* Asura_NetMsg_System::s_axUpdateData;
Asura_NetMsg_Recipients* Asura_NetMsg_System::s_pxRecipients;

// shared data for compression of whole messages
Asura_NetMsg_UpdateData* Asura_NetMsg_System::s_pxTemporaryUpdateData;

// parameter is for the hash size (2^n)
Asura_Hashed_List<Asura_NetMsg> Asura_NetMsg::s_xNetMsgList(2);

// A S U R A _ N E T M S G _ S Y S T E M /////////////////////////////////////
void Asura_NetMsg_System::Initialise()
{
	Asura_Assert(!IsInitialised(), "system already initialised");

	if (Asura_Network::IsServer() && AreLODUpdatesAllowed())
	{
		s_uMaxDistinctClientUpdates = Asura_Network::GetMaxNumberOfMachines();
		s_bIsUsingLODUpdates = true;
	}
	else
	{
		s_uMaxDistinctClientUpdates = 1;
		s_bIsUsingLODUpdates = false;
	}

	s_axUpdateData = new Asura_NetMsg_UpdateData[s_uMaxDistinctClientUpdates];
	s_pxTemporaryUpdateData = new Asura_NetMsg_UpdateData;
	s_apxClientsForCurMessage = new Asura_NetMsg_UpdateData*[s_uMaxDistinctClientUpdates];
	s_pxRecipients = new Asura_NetMsg_Recipients(s_apxClientsForCurMessage, 0);
	s_uNumClientsForCurMessage = 0;

	s_bCanAddMessages = false;

	// init internal globals (these 2 are used to spread full updates out over many frames)
	g_uInitialHeartbeatTimerValue = 0;
	g_uFullUpdateBatchSize = 0;

	Asura_NetMsg_CompressedPositions::SetCompression( Asura_NetMsg_CompressedPositionSettings_Normal );
	//TestCompressedPosition();

	// flag initialised
	s_bIsInitialised = true;
}


void Asura_NetMsg_System::Shutdown()
{
	// delete all registered messages to avoid fragmentation.
	Asura_NetMsg::UnregisterAll();

	delete[] s_apxClientsForCurMessage;
	s_apxClientsForCurMessage = NULL;
	
	delete s_pxRecipients;
	s_pxRecipients = NULL;
	
	delete[] s_axUpdateData;
	s_axUpdateData = NULL;

	delete s_pxTemporaryUpdateData;
	s_pxTemporaryUpdateData = NULL;

	s_uMaxDistinctClientUpdates = 0;

	s_bCanAddMessages = false;
	
	// flag not initialised
	s_bIsInitialised = false;
}

bool Asura_NetMsg_System::SetPositionCompression( const Asura_NetMsg_CompressedPositionSettings& xSettings )
{
	return Asura_NetMsg_CompressedPositions::SetCompression( xSettings );
}

const Asura_NetMsg_CompressedPositionSettings& Asura_NetMsg_System::GetPositionCompression()
{
	return Asura_NetMsg_CompressedPositions::GetCompression();
}

float Asura_NetMsg_System::GetCompressedPosMaxX()
{
	return Asura_NetMsg_CompressedPositions::GetMaxX();
}

float Asura_NetMsg_System::GetCompressedPosMaxY()
{
	return Asura_NetMsg_CompressedPositions::GetMaxY();
}

float Asura_NetMsg_System::GetCompressedPosMaxZ()
{
	return Asura_NetMsg_CompressedPositions::GetMaxZ();
}

// Compress the position vector and add to the bitpacker.
void Asura_NetMsg_System::CompressPosition(const Asura_Vector_3& xPosition, Asura_BitPacker& xPacker)
{
	u_int uX = 0;
	u_int uY = 0;
	u_int uZ = 0;

#ifdef ASURA_ASSERT
	const bool bOk =
#endif
	Asura_NetMsg_CompressedPositions::FromVector( xPosition, uX, uY, uZ );

	Asura_Assert(bOk, "Position vector out-of-range for compression");

	xPacker.Pack(uX, Asura_NetMsg_CompressedPositions::GetNumBitsX());
	xPacker.Pack(uY, Asura_NetMsg_CompressedPositions::GetNumBitsY());
	xPacker.Pack(uZ, Asura_NetMsg_CompressedPositions::GetNumBitsZ());
}

// Uncompress the position vector from the bitunpacker.
void Asura_NetMsg_System::UncompressPosition(Asura_Vector_3& xPosition, Asura_BitUnpacker& xUnpacker)
{
	const u_int uX = xUnpacker.Unpack(Asura_NetMsg_CompressedPositions::GetNumBitsX());
	const u_int uY = xUnpacker.Unpack(Asura_NetMsg_CompressedPositions::GetNumBitsY());
	const u_int uZ = xUnpacker.Unpack(Asura_NetMsg_CompressedPositions::GetNumBitsZ());

	Asura_NetMsg_CompressedPositions::ToVector(xPosition, uX, uY, uZ);
}

void Asura_NetMsg_System::BeginUpdate()
{
	if (!s_bCanAddMessages)
	{
		// new set of updates. They are sent every 0.1s but created over several frames.
		// store the number of clients that will be used, as it could change as players join and leave.
		// If a new player joins between now and the next set of updates, he will have to wait.
		// It's only 0.1s after all. If a player leaves in the meantime, a check below will 
		// stop processing for that player and the data will not get sent when the message is
		// sent off to Asura_Network.

		if (IsUsingLODUpdates())
		{
			// (MH - 06.07.2004) this isn't currently compatible with having more than 1 players 
			// on a machine (nor with having no players on a machine, i.e. spectators)
			Asura_Assert(Asura_Network::GetNumberOfPlayers() <= s_uMaxDistinctClientUpdates, "too many players");
			s_uNumClientsForThisUpdate = Asura_Network::GetNumberOfPlayers();
		}
		else
		{
			s_uNumClientsForThisUpdate = 1;
		}

		for (u_int u = 0; u < s_uNumClientsForThisUpdate; ++u)
		{
			Asura_Guid uPlayerGuid = Asura_Network::GetPlayerGuid(u);
			s_axUpdateData[u].Restart(uPlayerGuid);
		}

		// flag OK to add messages
		s_bCanAddMessages = true;
	}
	else
	{
		for (u_int u = 0; u < s_uNumClientsForThisUpdate; ++u)
		{
			s_axUpdateData[u].NewFrame();
		}
	}
}


bool Asura_NetMsg_System::AddAsuraMessage(const Asura_Message* pxMsg)
{
	if (dbg_bDisableAsuraNetMsgSystem) return false;
	if (!s_bCanAddMessages) return false;

	// find Asura_NetMsg corresponding to message ID and entity classification.
	u_short usMsgID = pxMsg->GetMsgID();
	u_short usEntityClassification = pxMsg->GetEntityClassification();
	const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification);
	if (!pxNetMsg)
	{
		// not found, attempt to get the entity system to register it.
		bool bOK = Asura_Entity_System::RegisterUpdateMessage(usMsgID, usEntityClassification);
		if (!bOK)
		{
			// no information on this message
			Asura_Assertf(usMsgID != ASURA_MSG_CLIENTUPDATE, ("entity class %s is not using Asura_NetMsg_System for its updates", 
				Asura_Network::GetEntityClassificationDesc(usEntityClassification)));
			return false;
		}
	}
	
	Asura_Assert(!pxMsg->IsUsingGuaranteedDelivery(), "net msg system doesn't do guaranteed messages");

	// should be able to get the registered object now
	pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification);
	Asura_Assert(pxNetMsg, "Asura_Entity_System::RegisterUpdateMessage returned true but didn't create object");

	// (MH - 20.03.2005) if the message was not correctly registered we used to send full updates.
	// This could lead to huge bandwidth wastage and messed up online tests. It's better to have a 
	// bug of objects not appearing / updating than that and it's more likely to get noticed and 
	// fixed quickly.
	Asura_Assert(pxNetMsg->GetDataSize() == pxMsg->GetDataSize(), "netmsg data size and asura message data size do not match. No updates will be sent");
	if (pxNetMsg->GetDataSize() != pxMsg->GetDataSize())
	{
		return true;
	}

	// get Asura_NetMsg_MasterInstance from pxMsg guid, creating a new one if necessary
	Asura_NetMsg_MasterInstance* pxMsgMasterInstance = pxNetMsg->FindMasterInstance(pxMsg->From);
	if (!pxMsgMasterInstance)
	{
		float fHeartbeat = Asura_Entity_System::GetUpdateHeartbeat(usMsgID, usEntityClassification);
		pxMsgMasterInstance = new Asura_NetMsg_MasterInstance(pxMsg->From, pxNetMsg, fHeartbeat, pxMsg->GetData());
	}

	// update data, and add to update package.
	// first store ptr to the asura message so we can easily get the header details if we need them.
	s_pxAsuraMessageToAdd = pxMsg;
	Asura_Mem::Copy(pxMsgMasterInstance->GetData(), pxMsg->GetData(), pxMsg->GetDataSize());
	return pxMsgMasterInstance->SendDataUpdate( pxMsg );
}

////////////////////////////////////
// InitialiseFromClientUpdateData //
////////////////////////////////////
// MGH 27/1/2006: Initialise a netmsg instance for a ASURA_MSG_CLIENTUPDATE message
// from the corresponding create message. This is just a helper function for the
// following function. Pass in pointers to the message header and the update data
// that is contained within the create message. NOTE: the data must have the
// correct data layout in memory otherwise bad things will happen.
//
// pxMsg  : points to the message header of the create message.
// pxData : points to the data within the create message that must have exactly
//          the same memory layout as for the corresponding client update message.

void Asura_NetMsg_System::InitialiseFromClientUpdateData(const Asura_Message* pxMsg,
													     const void* pxData)
{
	if (dbg_bDisableAsuraNetMsgSystem) return;
	if (!IsInitialised()) return;

	Asura_Assert(pxMsg, "Null message pointer");
	if (!pxMsg) return;

	Asura_Assert(pxData, "Null data pointer");
	if (!pxData) return;

	const u_short usEntityClassification = pxMsg->GetEntityClassification();
	const Asura_Guid uFromGuid = pxMsg->From;
	
	// Create messages should always be sent from the server to the client,
	// so that the master instances are on the server and the slave instances
	// are on the remote machines.
	const bool bMaster = Asura_Network::IsServer();

	// Initialise client update message from a create message.
	InitialiseFromData(pxData, ASURA_MSG_CLIENTUPDATE, usEntityClassification, uFromGuid, bMaster);
}

////////////////////////
// InitialiseFromData //
////////////////////////
// MGH 27/1/2006: Initialise a netmsg instance from a (possibly) different message
// type that contains the same data. For example, create messages often contain the
// same data that is contained in an update message. By initialising the master or
// slave netmsg instance from the create message rather than waiting for a full update
// to be sent, we can not only ensure that the netmsg system is initialised sooner,
// but also it saves us from having to send the same data over the network twice.
// NOTE: the data must have the correct data layout in memory otherwise bad things will happen.
//
// pxData			: points to the data that must have exactly the same memory layout as
//					  for the specified type of message.
// usMsgID			: the relevant message id.
// usClassification : the relevant entity classification.
// uFromGuid		: the guid of the entity for which this data is to be applied.
// bMaster			: initialise the master or the slave instance?

void Asura_NetMsg_System::InitialiseFromData(const void* pxData,
											 const u_short usMsgID,
											 const u_short usClassification,
											 const Asura_Guid uFromGuid,
											 const bool bMaster)
{
	if (dbg_bDisableAsuraNetMsgSystem) return;
	if (!IsInitialised()) return;

	Asura_Assert(pxData, "Null data pointer");
	if (!pxData) return;

	// find Asura_NetMsg corresponding to translated message ID and entity classification.
	const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usClassification);
	if (!pxNetMsg)
	{
		// not found, attempt to get the entity system to register it.
		const bool bOK = Asura_Entity_System::RegisterUpdateMessage(usMsgID, usClassification);
		if (!bOK) return;

		// should be able to get the registered object now
		pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usClassification);
		Asura_Assert(pxNetMsg, "Asura_Entity_System::RegisterUpdateMessage returned true but didn't create object");
		if (!pxNetMsg) return;
	}

	// Initialise master or slave instance.
	if (bMaster)
	{
		// If the master instance already exists then there is no need to initialise it.
		Asura_NetMsg_MasterInstance* pxMsgMasterInstance = pxNetMsg->FindMasterInstance(uFromGuid);
		if (!pxMsgMasterInstance)
		{
			const float fHeartbeat = Asura_Entity_System::GetUpdateHeartbeat(usMsgID, usClassification);
			pxMsgMasterInstance = new Asura_NetMsg_MasterInstance(uFromGuid, pxNetMsg, fHeartbeat, pxData);
		
			// MGH 27/1/2006: Since we have just created the master instance and sent a create
			// message we can assume that the clients won't need a full update immediately.
			if (pxMsgMasterInstance)
			{
				pxMsgMasterInstance->AdvanceTimerByOneHeartbeat();
			}
		}
	}
	else
	{
		// If the slave instance already exists then there is no need to initialise it.
		Asura_NetMsg_SlaveInstance* pxMsgSlaveInstance = pxNetMsg->FindSlaveInstance(uFromGuid);
		if (!pxMsgSlaveInstance)
		{
			pxMsgSlaveInstance = new Asura_NetMsg_SlaveInstance(uFromGuid, pxNetMsg);
			Asura_Mem::Copy(pxMsgSlaveInstance->GetData(), pxData, pxNetMsg->GetDataSize());
		}
	}
}


void Asura_NetMsg_System::SendUpdate()
{
	// no more messages until next BeginUpdate is called
	s_bCanAddMessages = false;

	if (IsUsingLODUpdates())
	{
		for (u_int u = 0; u < s_uNumClientsForThisUpdate; ++u)
		{
			Asura_Handle iMachineForPlayer = Asura_Network::GetMachineForPlayer(s_axUpdateData[u].GetGuid(), false);
			if ( (0 == iMachineForPlayer) || (ASURA_INVALID_HANDLE == iMachineForPlayer) )
			{
				// this player has disconnected, or is on the server.
				continue;
			}
			s_axUpdateData[u].Send(ASURA_GUID_PLAYERS_CLIENTMACHINE, s_axUpdateData[u].GetGuid());
		}
	}
	else
	{
		s_axUpdateData[0].Send(ASURA_GUID_ALL_OTHER_CLIENTMACHINES, ASURA_GUID_UNREGISTERED);
	}
}


void Asura_NetMsg_System::Reset()
{
	// delete all registered messages to avoid fragmentation.
	Asura_NetMsg::UnregisterAll();

	// MGH 5/5/2005: Try to fix problem with negative time offsets after an
	// MP level rotation.
	s_bCanAddMessages = false;
}


void Asura_NetMsg_System::SkipMessageData( const Asura_NetMsg& xNetMsg, class Asura_BitUnpacker& xBitUnpacker )
{
	// need a couple of dummy areas to unpack to.
	bool bSkippedLastUpdate;
	u_char* pucTempBuffer = new u_char[xNetMsg.GetDataSize()];
	xNetMsg.UnpackToInstance( pucTempBuffer, bSkippedLastUpdate, xBitUnpacker );
	delete[] pucTempBuffer;
}


void Asura_NetMsg_System::ProcessSlaveUpdatePackage(const u_char* const pucData, u_short usNumBytes, float fSentTime)
{
	Asura_Assert(pucData && usNumBytes, "no data / zero data size");
	if (!(pucData && usNumBytes)) return;

	float fSentTimeForCurrentPacket = fSentTime;
	
	Asura_Assert(usNumBytes <= uASURA_NETMSG_UPDATE_PACKAGE_MAX_SIZE, "too much data received");
	Asura_BitUnpacker xUnpacker(pucData, usNumBytes);

	// unpack 1 byte to get num messages
	u_int uNumMessages = xUnpacker.Unpack(uASURA_NETMSG_BITS_PER_BYTE);

	while (uNumMessages--)
	{
		// extract the header information
		u_short usMsgID;
		u_short usEntityClassification;
		Asura_Guid uGuid;

		// MsgID
		bool bHasMsgID = xUnpacker.UnpackBool();
		if (bHasMsgID)
		{
			usMsgID = Asura_MessagePacking::UnpackMsgID( &xUnpacker );
			
			// check for special message ID
			if (usMsgID == usTIME_OFFSET_MESSAGE_ID)
			{
				u_int uTimeOffset = xUnpacker.Unpack(uBITS_FOR_TIME_OFFSET);
				float fTimeOffset = static_cast<float>(uTimeOffset) / 1000.0f;
				fSentTime = fSentTimeForCurrentPacket + fTimeOffset;
				continue;
			}
		}
		else
		{
			usMsgID = ASURA_MSG_CLIENTUPDATE;
		}

		// EntityClassification
		bool bHasEntityClassification = xUnpacker.UnpackBool();
		if (bHasEntityClassification)
		{
			usEntityClassification = Asura_MessagePacking::UnpackEntityClassification( &xUnpacker );
		}
		else
		{
			Asura_Assert(s_usEntityClassificationPlayer != AsuraEntityClass_NotSet, "player entity classification not set. You need to call Asura_NetMsg_System::SetPlayerEntityClassForProject()");
			usEntityClassification = s_usEntityClassificationPlayer;
		}

		// Guid
		bool bGuidNotInRange1000to1031 = xUnpacker.UnpackBool();
		if (bGuidNotInRange1000to1031)
		{
			uGuid = Asura_MessagePacking::UnpackGuid( &xUnpacker );
		}
		else
		{
			uGuid = xUnpacker.Unpack( uNUM_BITS_FOR_PLAYER_GUID );
			uGuid += uFIRST_PLAYER_GUID;
		}

		// First look up the message definition from msg ID and entity classification.
		const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification);
		if (!pxNetMsg)
		{
			// not found, attempt to get the entity system to register it.
			#ifdef ASURA_ASSERT
			bool bOK = 
			#endif
			Asura_Entity_System::RegisterUpdateMessage(usMsgID, usEntityClassification);
			Asura_Assert(bOK, "failed to register message on client");
			// this was registered by the master version, so it had better be registerable now.
		}

		// should be able to get the registered object now
		pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification);
		Asura_Assert(pxNetMsg, "Asura_Entity_System::RegisterUpdateMessage returned true but didn't create object");
		if(!pxNetMsg)	return;

		// find slave message instance
		Asura_NetMsg_SlaveInstance* pxSlaveInstance = pxNetMsg->FindSlaveInstance(uGuid);
		bool bIsFirstUpdate;
		if (pxSlaveInstance)
		{
			bIsFirstUpdate = false;
		}
		else
		{
			bIsFirstUpdate = true;

			// create the entity - we can only do this if this is a full update.
			// Extract the field index from data to check this.
			u_int uFieldIndex = xUnpacker.Peek(uBITS_FOR_FIELD_INDEX);
			if (uFieldIndex != uFIELDID_FULLUPDATE)
			{
				// skip the data and move on to the next message.
				SkipMessageData( *pxNetMsg, xUnpacker );
				continue;
			}

			// get Asura_NetMsg_SlaveInstance from pxMsg guid, creating a new one if necessary
			pxSlaveInstance = new Asura_NetMsg_SlaveInstance(uGuid, pxNetMsg);
		}

		// now we definitely have a slave instance
		pxSlaveInstance->ProcessReceivedData( fSentTime, xUnpacker, bIsFirstUpdate );
	}

#ifdef ASURA_ASSERT
	u_int uNumBitsRemaining = xUnpacker.GetNumBitsRemaining();
	Asura_Assert(uNumBitsRemaining < uASURA_NETMSG_BITS_PER_BYTE, "ProcessPacketOnClient read unexpected amount of data");
	if (uNumBitsRemaining < uASURA_NETMSG_BITS_PER_BYTE)
	{
		u_int uRemainingBits = xUnpacker.Unpack(uNumBitsRemaining);
		Asura_Assert(uRemainingBits == 0, "ProcessPacketOnClient read unexpected amount of data");
	}
#endif
}


// functions to pack and unpack a complete message
bool Asura_NetMsg_System::PackMessageData( const Asura_Message& xMsg, u_int& uDataSize, const void*& pData )
{
	if (!xMsg.GetDataSize())
	{
		// nothing to do!
		return true;
	}

	const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(xMsg.GetMsgID(), xMsg.GetEntityClassification());
	if (!pxNetMsg)
	{
		pxNetMsg = Asura_Network::RegisterNetMsg(xMsg.GetMsgID(), xMsg.GetEntityClassification());
	}

	if (!pxNetMsg)
	{
		return false;
	}

	Asura_Assertf(pxNetMsg->GetDataSize() == xMsg.GetDataSize(), 
		("netmsg data size doesn't match asura message's - is msgID %s used for different payloads?", 
		Asura_Network::GetMsgIDDesc(xMsg.GetMsgID())));

	// set the system to use the temporary buffer
	s_pxTemporaryUpdateData->ResetBitPacker();
	s_apxClientsForCurMessage[0] = s_pxTemporaryUpdateData;
	s_uNumClientsForCurMessage = 1;
	s_pxRecipients->ChangeArraySize(1);

	// pack the message
	pxNetMsg->PackWholeMessage(xMsg.GetData(), *Asura_NetMsg_System::s_pxRecipients);

	// set the return data size and pointer
	s_pxTemporaryUpdateData->GetUnpackData(uDataSize, pData);
	return true;
}

bool Asura_NetMsg_System::UnpackMessageData( Asura_Message& xMsg, u_int uDataSize, const void* pData )
{
	const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(xMsg.GetMsgID(), xMsg.GetEntityClassification());
	if (!pxNetMsg)
	{
		Asura_Assert(false, "failed to find NetMsg to unpack received data");
		return false;
	}
	const bool bDataSizeMatches = xMsg.GetDataSize() == pxNetMsg->GetDataSize();
	Asura_Assert(bDataSizeMatches, "mismatched data size in Asura message and NetMsg");
	if (!bDataSizeMatches)
	{
		return false;
	}

	Asura_BitUnpacker xUnpacker(pData, uDataSize);
	pxNetMsg->UnpackWholeMessage(xMsg.GetData(), xUnpacker);
	return true;
}

// (MH - 26.02.2004) function to analyse data size for each message ID - used for logging / profiling
// function returns number of messages in packet, and writes up to uArraySize entries to ausMsgID, 
// aucDataSize and abIsFullUpdate
u_int Asura_NetMsg_System::GetMessageSummaries(const u_char* const pucData, u_short usNumBytes, u_int uArraySize, 
											   u_char* aucHeaderSize, u_short* ausMsgIDOut, u_short* ausEntityClassOut, 
											   Asura_Guid* auGuidOut, u_short* ausDataSizeOut, bool* abIsFullUpdate,
											   bool* abIsTimeOffset)
{
	Asura_Assert(pucData && usNumBytes, "no data / zero data size");
	if (!(pucData && usNumBytes)) return 0;

	Asura_BitUnpacker xUnpacker(pucData, usNumBytes);

	// unpack a byte for num messages
	const u_int uNumMessages = xUnpacker.Unpack(uASURA_NETMSG_BITS_PER_BYTE);

	u_int uNumMessagesRemaining = uNumMessages;
	u_int uNumOutput = 0;

	while (uNumMessagesRemaining--)
	{
		// remember the start position
		u_int uBitsRemainingBeforeHeader = xUnpacker.GetNumBitsRemaining();
		u_int uBitsRemainingBeforeData = 0;

		// extract the header information
		u_short usMsgID;
		u_short usEntityClassification = 0;
		Asura_Guid uGuid = 0;
		bool bIsFullUpdate = false;

		bool bIsTimeOffset = false;

		// MsgID
		bool bHasMsgID = xUnpacker.UnpackBool();
		if (bHasMsgID)
		{
			usMsgID = Asura_MessagePacking::UnpackMsgID( &xUnpacker );
			// check for special message ID
			if (usMsgID == usTIME_OFFSET_MESSAGE_ID)
			{
				u_int uTimeOffset = xUnpacker.Unpack(uBITS_FOR_TIME_OFFSET);
				bIsTimeOffset = true;
				usEntityClassification = AsuraEntityClass_NotSet;
				uGuid = uTimeOffset;	// hack - fill guid in with time offset value
				bIsFullUpdate = false;
				uBitsRemainingBeforeData = xUnpacker.GetNumBitsRemaining();
			}
		}
		else
		{
			usMsgID = ASURA_MSG_CLIENTUPDATE;
		}

		// if it's a time offset, following data is not present
		if (!bIsTimeOffset)
		{
			// EntityClassification
			bool bHasEntityClassification = xUnpacker.UnpackBool();
			if (bHasEntityClassification)
			{
				usEntityClassification = Asura_MessagePacking::UnpackEntityClassification( &xUnpacker );
			}
			else
			{
				Asura_Assert(s_usEntityClassificationPlayer != AsuraEntityClass_NotSet, "player entity classification not set. You need to call Asura_NetMsg_System::SetPlayerEntityClassForProject()");
				usEntityClassification = s_usEntityClassificationPlayer;
			}

			// Guid
			bool bGuidNotInRange1000to1031 = xUnpacker.UnpackBool();
			if (bGuidNotInRange1000to1031)
			{
				uGuid = Asura_MessagePacking::UnpackGuid( &xUnpacker );
			}
			else
			{
				uGuid = xUnpacker.Unpack( uNUM_BITS_FOR_PLAYER_GUID );
				uGuid += uFIRST_PLAYER_GUID;
			}

			u_int uFieldIndex = xUnpacker.Peek(uBITS_FOR_FIELD_INDEX);
			bIsFullUpdate = (uFieldIndex == uFIELDID_FULLUPDATE);

			uBitsRemainingBeforeData = xUnpacker.GetNumBitsRemaining();
			
			// skip the data and move on to the next message.
			const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification);
			Asura_Assert(pxNetMsg, "failed to find netmsg object when logging - it must exist if it was created.");
			if (pxNetMsg)
			{
				SkipMessageData( *pxNetMsg, xUnpacker );
			}
		}

		Asura_Assert(uNumOutput < uArraySize, "array too small to receive summary data for all messages");
		if (uNumOutput < uArraySize)
		{
			aucHeaderSize[uNumOutput] = static_cast<u_char>(uBitsRemainingBeforeHeader - uBitsRemainingBeforeData);
			ausMsgIDOut[uNumOutput] = usMsgID;
			ausEntityClassOut[uNumOutput] = usEntityClassification;
			auGuidOut[uNumOutput] = uGuid;
			ausDataSizeOut[uNumOutput] = static_cast<u_short>(uBitsRemainingBeforeData - xUnpacker.GetNumBitsRemaining());
			abIsFullUpdate[uNumOutput] = bIsFullUpdate;
			abIsTimeOffset[uNumOutput] = bIsTimeOffset;
			uNumOutput++;
		}
	}

	return uNumMessages;
}



bool Asura_NetMsg_System::BeginMessage(u_short usMsgID, u_short usEntityClassification, Asura_Guid uGuid, u_int uMaxDataSize, u_char ucMaxFields)
{
	// this slightly overestimates the amount of room required by not taking the header compression into account
	u_short usMaxSize = static_cast<u_short>(sizeof(usMsgID) + sizeof(usEntityClassification) + 
		sizeof(uGuid) + sizeof(u_char) + uMaxDataSize + 
		(sizeof(u_char) * ucMaxFields));

	for (u_int u = 0; u < s_uNumClientsForCurMessage; ++u)
	{
		if (!s_apxClientsForCurMessage[u]->RoomForBytes(usMaxSize))
		{
			return false;
		}
	}

	return true;
}


void Asura_NetMsg_System::AddMessageHeader( void )
{
	for (u_int u = 0; u < s_uNumClientsForCurMessage; ++u)
	{
		s_apxClientsForCurMessage[u]->AddMessageHeader();
	}

	// clear the current message ptr now we've added it
	s_pxAsuraMessageToAdd = NULL;
}


void Asura_NetMsg_System::AddMessageDataFieldID(u_int uFieldIndex)
{
	if (NeedToAddMessageHeader())
	{
		AddMessageHeader();
		Asura_Assert(!NeedToAddMessageHeader(), "adding message header didn't clear flag");
	}

	// add field identifier
	Asura_Assert(uFieldIndex < (1 << uBITS_FOR_FIELD_INDEX), "field index too big for range");

	AddBits(uFieldIndex, uBITS_FOR_FIELD_INDEX);
	
}


void Asura_NetMsg_System::AddBits(u_int uValue, u_int uNumBits)
{
	for (u_int u = 0; u < s_uNumClientsForCurMessage; ++u)
	{
		s_apxClientsForCurMessage[u]->AddBits(uValue, uNumBits);
	}
}


bool Asura_NetMsg_System::EndMessage(bool bAddEndOfDataField, bool bSkippedLastUpdate)
{
	if (NeedToAddMessageHeader())
	{
		// still haven't added message header, so we can't have added any data
		return false;
	}
	else
	{
		if (bAddEndOfDataField)
		{
			AddBits(uFIELDID_ENDOFDATA, uBITS_FOR_FIELD_INDEX);
		}

		// write the flag
		AddBits(bSkippedLastUpdate, 1);
		return true;
	}
}


void Asura_NetMsg_System::EntityDestroyed(Asura_Guid uGuid)
{
    Asura_Hashed_List_It<Asura_NetMsg> xIterator( &Asura_NetMsg::s_xNetMsgList );
    while ( !xIterator.Done() )
    {
		Asura_NetMsg* pxNetMsg = xIterator.GetCurrent();
		xIterator.Next();
		Asura_NetMsg_MasterInstance* pxMasterInstance = pxNetMsg->FindMasterInstance(uGuid);
		Asura_NetMsg_SlaveInstance* pxSlaveInstance = pxNetMsg->FindSlaveInstance(uGuid);
		if (pxMasterInstance)
		{
			delete pxMasterInstance;
		}
		if (pxSlaveInstance)
		{
			delete pxSlaveInstance;
		}
    }
}


bool Asura_NetMsg_System::IsInitialised( void )
{
	return s_bIsInitialised;
}


// before initialising, set whether LOD updates should be used with AllowLODUpdates()
// after initialising, they can be turned on and off for debugging using SetUseLODUpdates()
void Asura_NetMsg_System::AllowLODUpdates( bool bEnable )
{
	Asura_Assert(!IsInitialised(), "can't change whether LOD updates are allowed once initialised");
	if ( !IsInitialised() )
	{
		s_bLODUpdatesAllowed = bEnable;
	}
}

bool Asura_NetMsg_System::AreLODUpdatesAllowed( void )
{
	return s_bLODUpdatesAllowed;
}

void Asura_NetMsg_System::SetUseLODUpdates( bool bEnable )
{
	Asura_Assert(s_bLODUpdatesAllowed, "LOD updates are not allowed");
	if (!s_bLODUpdatesAllowed) return;

	if (s_bIsUsingLODUpdates != bEnable)
	{
		// check if we're mid-update, if so we need to flush the updates received so far
		// and begin the update after changing the flag.
		bool bInUpdate = s_bCanAddMessages;
		if (bInUpdate)
		{
			SendUpdate();
		}
		
		s_bIsUsingLODUpdates = bEnable;

		if (bInUpdate)
		{
			BeginUpdate();
		}
	}
}

bool Asura_NetMsg_System::IsUsingLODUpdates( void )
{
	return s_bIsUsingLODUpdates;
}

////////////////////
// SendFullUpdate //
////////////////////
// MGH 10/1/2006: Call this function on the server to send a full update in the next message
// rather than waiting for the heartbeat. This can be used to reduce the delay between a player
// being created and the first full update being sent (during which time the player can't move).
void Asura_NetMsg_System::SendFullUpdate(u_short usMsgID, u_short usEntityClassification, Asura_Guid uGuid)
{
	Asura_Assert(Asura_Network::IsServer(), "Should only call on server!");
	if (!Asura_Network::IsServer()) return;

	const Asura_NetMsg* pxNetMsg = Asura_NetMsg::FindNetMsgObject(usMsgID, usEntityClassification);
	if (pxNetMsg)
	{
		Asura_NetMsg_MasterInstance* pxMasterInstance = pxNetMsg->FindMasterInstance(uGuid);
		if (pxMasterInstance)
		{
			pxMasterInstance->SendFullUpdate();
		}
	}
}


// A S U R A _ N E T M S G ///////////////////////////////////////////////////
Asura_NetMsg::Asura_NetMsg(u_short usMsgID, u_short usEntityClassification, u_char ucNumFields, u_short usDataSize)
: m_usMsgID(usMsgID)
, m_usEntityClassification(usEntityClassification)
, m_usDataSize(usDataSize)
, m_ucNumFields(0)
{
	Asura_Assert(s_ucMaxFields == 0 && s_uRegistrationBaseOffset == 0, "registration already in progress");
	Asura_Assert(ucNumFields <= uMAX_FIELDID, "too many fields");
	m_axFields = new Asura_NetMsg_Field[ucNumFields];
	m_pxMasterInstanceList = new Asura_Hashed_List<Asura_NetMsg_MasterInstance>(4);
	m_pxSlaveInstanceList = new Asura_Hashed_List<Asura_NetMsg_SlaveInstance>(4);

	s_ucMaxFields = ucNumFields;

	s_xNetMsgList.AddToList(this);
}


Asura_NetMsg::~Asura_NetMsg()
{
    // RDM - 29/06/06 - I've had to create a separate scope around these two iterators to ensure
	// that they are deleted before the lists. It never occurred to me that you might want to delete
	// the list before an iterator has been unregistered. Once I've fixed the hashed list, this won't 
	// be necessary.
	{
		// first need to delete all master and slave instances
		Asura_Hashed_List_It<Asura_NetMsg_MasterInstance> xMasterListIterator( m_pxMasterInstanceList );
		while ( !xMasterListIterator.Done() )
		{
			Asura_NetMsg_MasterInstance* pxMasterInstance = xMasterListIterator.GetCurrent();
			xMasterListIterator.Next();
			delete pxMasterInstance;
		}
	    
		Asura_Hashed_List_It<Asura_NetMsg_SlaveInstance> xSlaveListIterator( m_pxSlaveInstanceList );
		while ( !xSlaveListIterator.Done() )
		{
			Asura_NetMsg_SlaveInstance* pxSlaveInstance = xSlaveListIterator.GetCurrent();
			xSlaveListIterator.Next();
			delete pxSlaveInstance;
		}
	}
	
	delete[] m_axFields;
	delete m_pxMasterInstanceList;
	delete m_pxSlaveInstanceList;
	s_xNetMsgList.RemoveFromList(this);
}


void Asura_NetMsg::RegisterField(const char* szDbgName,
							   Asura_NetMsg_FieldType eFieldType, 
							   u_int uByteOffset, 
							   u_int uFieldSize, 
							   u_char ucArraySize, 
							   Asura_NetMsg_ThresholdType eThresholdType, 
							   float fThresholdValue)
{
	Asura_Assert(m_ucNumFields < s_ucMaxFields, "too many fields registered");
	Asura_Assert(eFieldType < NUM_ASURA_NETMSG_FIELD_TYPES, "invalid field type");
	Asura_Assert(ucArraySize != 0, "invalid array size");
	Asura_Assert(eThresholdType < NUM_ASURA_NETMSG_THRESHOLD_TYPES, "invalid threshold type");
	Asura_Assert((uFieldSize & 7) == 0, "field size must be whole number of bytes");
	Asura_Assert(uFieldSize <= UCHAR_MAX, "field size in bits must fit in u_char");

	// check and adjust values based on field type
	switch (eFieldType)
	{
		case Asura_FieldType_char:
		case Asura_FieldType_uchar:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(char), "invalid size for field type");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue, "not supported");
			break;
		}

		case Asura_FieldType_short:
		case Asura_FieldType_ushort:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(short), "invalid size for field type");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue || eThresholdType == Asura_Threshold_NotEqual, "not supported");
			break;
		}

		case Asura_FieldType_int:
		case Asura_FieldType_uint:
		case Asura_FieldType_Guid:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(int), "invalid size for field type");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue, "not supported");
			break;
		}

		case Asura_FieldType_float:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(float), "invalid size for field type");
			break;
		}

		case Asura_FieldType_vector2:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(Asura_Vector_2), "invalid size for field type");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue || eThresholdType == Asura_Threshold_Magnitude, "not supported");

			// use magnitude squared
			fThresholdValue *= fThresholdValue;
			break;
		}

		case Asura_FieldType_vector3:
		case Asura_FieldType_CompressedPosition:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(Asura_Vector_3), "invalid size for field type");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue || eThresholdType == Asura_Threshold_Magnitude, "not supported");

			// use magnitude squared
			fThresholdValue *= fThresholdValue;
			break;
		}
		case Asura_FieldType_euler3:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(Asura_Euler_3), "invalid size for field type");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue || eThresholdType == Asura_Threshold_Magnitude, "not supported");
			break;
		}
			
		case Asura_FieldType_quat:
		case Asura_FieldType_UprightQuat:
		case Asura_FieldType_CompressedQuat:
		{
			Asura_Assert(uFieldSize == uASURA_NETMSG_BITS_PER_BYTE * sizeof(Asura_Quat), "invalid size for field type");

			// check threshold is a valid cosine
			Asura_Assert(eThresholdType != Asura_Threshold_Magnitude || (fThresholdValue > 0.0f && fThresholdValue < 1.0f), "invalid threshold for quat");
			Asura_Assert(eThresholdType == Asura_Threshold_AnyChange || eThresholdType == Asura_Threshold_ConstantValue || eThresholdType == Asura_Threshold_Magnitude, "not supported");
			break;
		}
		
		default:
		{
			Asura_Assert(false, "invalid case");
			break;
		}
	}
	
	// add on any offset for being a subclass
	uByteOffset += s_uRegistrationBaseOffset;

	// check fields are registered in-order.
	Asura_Assertf( m_ucNumFields ? (m_axFields[m_ucNumFields-1].m_usByteOffset < uByteOffset) : (!uByteOffset), 
		("field %s registered out-of-order", szDbgName ? szDbgName : "N/A"));
	

	for (u_int u = 0; u < ucArraySize; ++u)
	{
		Asura_NetMsg_Field& xField = m_axFields[m_ucNumFields++];
#if _DEBUG
		xField.m_szDbgName = szDbgName;
#endif
		xField.m_ucFieldType = static_cast<u_char>(eFieldType);
		xField.m_ucBitOffset = 0;
		Asura_Assert(uByteOffset < uMAX_BYTE_OFFSET, "out of range");
		xField.m_usByteOffset = static_cast<u_short>(uByteOffset);
		Asura_Assert(uFieldSize < UCHAR_MAX, "out of range");
		xField.m_ucFieldDataSourceSizeInBits = static_cast<u_char>(uFieldSize);
		xField.m_ucFieldDataSendSizeInBits = static_cast<u_char>(uFieldSize);
		xField.m_ucThresholdType = static_cast<u_char>(eThresholdType);
		xField.m_ucFieldSizeInBytes = static_cast<u_char>(uFieldSize);
		xField.m_fThresholdValue = fThresholdValue;
		xField.m_fMinValue = 0.0f;
		xField.m_fMaxValue = 0.0f;

		uByteOffset += (uFieldSize / uASURA_NETMSG_BITS_PER_BYTE);
	}
}


void Asura_NetMsg::RegisterIntFieldBits(const char* szDbgName,
										Asura_NetMsg_FieldType eFieldType, 
										u_int uByteOffset, 
										u_int uFieldSize, 
										u_int uBits, 
										u_char ucArraySize, 
										Asura_NetMsg_ThresholdType eThresholdType)
{
	Asura_Assert(m_ucNumFields < s_ucMaxFields, "too many fields registered");
	Asura_Assert(eFieldType < NUM_ASURA_NETMSG_FIELD_TYPES, "invalid field type");
	Asura_Assert(ucArraySize != 0, "invalid array size");
	Asura_Assert(eThresholdType < NUM_ASURA_NETMSG_THRESHOLD_TYPES, "invalid threshold type");
	Asura_Assert(uBits < 32, "invalid number of bits"); // no point using 32 since that's just a regular u_int
	Asura_Assert(uFieldSize == 8 || uFieldSize == 32, "untested case");

	Asura_Assert(eFieldType == Asura_FieldType_uint_bits, "invaild field type");

	// add on any offset for being a subclass
	uByteOffset += s_uRegistrationBaseOffset;

	// check fields are registered in-order.
	Asura_Assertf( m_ucNumFields ? (m_axFields[m_ucNumFields-1].m_usByteOffset < uByteOffset) : (!uByteOffset), 
		("field %s registered out-of-order", szDbgName ? szDbgName : "N/A"));

	for (u_int u = 0; u < ucArraySize; ++u)
	{
		Asura_NetMsg_Field& xField = m_axFields[m_ucNumFields++];
#if _DEBUG
		xField.m_szDbgName = szDbgName;
#endif
		xField.m_ucFieldType = static_cast<u_char>(eFieldType);
		xField.m_ucBitOffset = 0;
		Asura_Assert(uByteOffset < uMAX_BYTE_OFFSET, "out of range");
		xField.m_usByteOffset = static_cast<u_short>(uByteOffset);
		Asura_Assert(uFieldSize < UCHAR_MAX, "out of range");
		xField.m_ucFieldDataSourceSizeInBits = static_cast<u_char>(uFieldSize);
		xField.m_ucFieldDataSendSizeInBits = static_cast<u_char>(uBits);
		xField.m_ucThresholdType = static_cast<u_char>(eThresholdType);
		xField.m_ucFieldSizeInBytes = static_cast<u_char>(uFieldSize);
		xField.m_fThresholdValue = 0.0f;
		xField.m_fMinValue = 0.0f;
		xField.m_fMaxValue = 0.0f;

		uByteOffset += (uFieldSize / uASURA_NETMSG_BITS_PER_BYTE);
	}
}

void Asura_NetMsg::RegisterBitfieldFieldBits(const char* szDbgName,
											 Asura_NetMsg_FieldType eFieldType,
											 u_int uByteOffset,
											 u_int uFieldSize,
											 u_int uBits,
											 u_char ucFirstBit,
											 Asura_NetMsg_ThresholdType eThresholdType, 
											 const char* pcBFDesc )
{
	Asura_Assert(m_ucNumFields < s_ucMaxFields, "too many fields registered");
	Asura_Assert(eFieldType < NUM_ASURA_NETMSG_FIELD_TYPES, "invalid field type");
	Asura_Assert((uBits + ucFirstBit) <= uFieldSize, "too many bits for bitfield");
	Asura_Assert(eThresholdType < NUM_ASURA_NETMSG_THRESHOLD_TYPES, "invalid threshold type");
	Asura_Assert(uFieldSize == 32, "only expecting 32-bit bitfields");
	Asura_Assert(eFieldType == Asura_FieldType_uint_Bitfield, "invalid field type");
	// fix it anyway
	eFieldType = Asura_FieldType_uint_Bitfield;

	// add on any offset for being a subclass
	uByteOffset += s_uRegistrationBaseOffset;

	// check fields are registered in-order.
	Asura_Assertf( m_ucNumFields ? (m_axFields[m_ucNumFields-1].m_usByteOffset <= uByteOffset) : (!uByteOffset), 
		("field %s registered out-of-order", szDbgName ? szDbgName : "N/A"));

	Asura_NetMsg_Field& xField = m_axFields[m_ucNumFields++];
#if _DEBUG
	xField.m_szDbgName = szDbgName;
#endif
	xField.m_ucFieldType = static_cast<u_char>(eFieldType);
	xField.m_ucBitOffset = ucFirstBit;
	Asura_Assert(uByteOffset < uMAX_BYTE_OFFSET, "out of range");
	xField.m_usByteOffset = static_cast<u_short>(uByteOffset);
	Asura_Assert(uFieldSize < UCHAR_MAX, "out of range");
	xField.m_ucFieldDataSourceSizeInBits = static_cast<u_char>(uFieldSize);
	xField.m_ucFieldDataSendSizeInBits = static_cast<u_char>(uBits);
	xField.m_ucThresholdType = static_cast<u_char>(eThresholdType);
	xField.m_ucFieldSizeInBytes = static_cast<u_char>(uFieldSize);
	xField.m_fThresholdValue = 0.0f;
	xField.m_fMinValue = 0.0f;
	xField.m_fMaxValue = 0.0f;

	// store the number of bits used for each field
	if (pcBFDesc)
	{
		u_int uNumBitsSoFar = 0;
		for (u_int u = 0; u < uASURA_NETMSG_MAX_FIELDS_IN_BITFIELD; ++u)
		{
			Asura_Assert(pcBFDesc[u] > 0, "invalid number of bits in bitfield desc");
			xField.m_acNumBits[u] = pcBFDesc[u];
			uNumBitsSoFar += pcBFDesc[u];
			if (uNumBitsSoFar >= uBits)
			{
				Asura_Assert(uNumBitsSoFar == uBits, "too many bits specified in bitfield desc");
				break;
			}
		}
	}
	else
	{
		// use default of single bits.
		for (u_int u = 0; u < uBits; ++u)
		{
			xField.m_acNumBits[u] = 1;
		}
	}
}


void Asura_NetMsg::RegisterUnicodeStringField(	const char* szDbgName, 
												Asura_NetMsg_FieldType eFieldType, 
												u_int uByteOffset, 
												u_char ucArraySize, 
												Asura_NetMsg_ThresholdType eThresholdType)
{
	Asura_Assert(m_ucNumFields < s_ucMaxFields, "too many fields registered");
	Asura_Assert(eFieldType < NUM_ASURA_NETMSG_FIELD_TYPES, "invalid field type");
	Asura_Assert(eThresholdType == Asura_Threshold_AnyChange, "invalid threshold type");

	Asura_Assert(eFieldType == Asura_FieldType_UnicodeString, "invalid field type");

	// field size is in bytes for this one, and strings are generally too big for bits
	u_int uFieldSize = sizeof(Asura_UnicodeChar) * ucArraySize;
	Asura_Assert(uFieldSize < 256, "string too long!");

	// add on any offset for being a subclass
	uByteOffset += s_uRegistrationBaseOffset;

	// check fields are registered in-order.
	Asura_Assertf( m_ucNumFields ? (m_axFields[m_ucNumFields-1].m_usByteOffset < uByteOffset) : (!uByteOffset), 
		("field %s registered out-of-order", szDbgName ? szDbgName : "N/A"));
	
	Asura_NetMsg_Field& xField = m_axFields[m_ucNumFields++];
#if _DEBUG
	xField.m_szDbgName = szDbgName;
#endif
	xField.m_ucFieldType = static_cast<u_char>(eFieldType);
	xField.m_ucBitOffset = 0;
	Asura_Assert(uByteOffset < uMAX_BYTE_OFFSET, "out of range");
	xField.m_usByteOffset = static_cast<u_short>(uByteOffset);
	xField.m_ucFieldDataSourceSizeInBits = 0;	// n/a
	xField.m_ucFieldDataSendSizeInBits = 0;		// n/a
	xField.m_ucThresholdType = static_cast<u_char>(eThresholdType);
	Asura_Assert(uFieldSize < UCHAR_MAX, "out of range");
	xField.m_ucFieldSizeInBytes = static_cast<u_char>(uFieldSize);
	xField.m_fThresholdValue = 0.0f;
	xField.m_fMinValue = 0.0f;
	xField.m_fMaxValue = 0.0f;
}

void Asura_NetMsg::RegisterFloatFieldBits(const char* szDbgName,
										  Asura_NetMsg_FieldType eFieldType, 
										  u_int uByteOffset, 
										  u_int uFieldSize, 
										  u_int uBits, 
										  float fMinValue, 
										  float fMaxValue, 
										  u_char ucArraySize, 
										  Asura_NetMsg_ThresholdType eThresholdType, 
										  float fThresholdValue)
{
	Asura_Assert(m_ucNumFields < s_ucMaxFields, "too many fields registered");
	Asura_Assert(eFieldType < NUM_ASURA_NETMSG_FIELD_TYPES, "invalid field type");
	Asura_Assert(ucArraySize != 0, "invalid array size");
	Asura_Assert(eThresholdType < NUM_ASURA_NETMSG_THRESHOLD_TYPES, "invalid threshold type");
	Asura_Assert(uBits < 32, "invalid number of bits");
	Asura_Assert(uFieldSize == sizeof(float) * uASURA_NETMSG_BITS_PER_BYTE, "incorrect field size");

	Asura_Assert(eFieldType == Asura_FieldType_CompressedFloat, "invalid field type");
	Asura_Assert(fThresholdValue >= 0.0f && fThresholdValue < (fMaxValue - fMinValue), "invalid threshold");
	Asura_Assert(fMinValue < fMaxValue, "invalid range");

	// add on any offset for being a subclass
	uByteOffset += s_uRegistrationBaseOffset;

	// check fields are registered in-order.
	Asura_Assertf( m_ucNumFields ? (m_axFields[m_ucNumFields-1].m_usByteOffset < uByteOffset) : (!uByteOffset), 
		("field %s registered out-of-order", szDbgName ? szDbgName : "N/A"));
	
	// sort this out
	for (u_int u = 0; u < ucArraySize; ++u)
	{
		Asura_NetMsg_Field& xField = m_axFields[m_ucNumFields++];
#if _DEBUG
		xField.m_szDbgName = szDbgName;
#endif
		xField.m_ucFieldType = static_cast<u_char>(eFieldType);
		xField.m_ucBitOffset = 0;
		Asura_Assert(uByteOffset < uMAX_BYTE_OFFSET, "out of range");
		xField.m_usByteOffset = static_cast<u_short>(uByteOffset);
		Asura_Assert(uFieldSize < UCHAR_MAX, "out of range");
		xField.m_ucFieldDataSourceSizeInBits = static_cast<u_char>(uFieldSize);
		xField.m_ucFieldDataSendSizeInBits = static_cast<u_char>(uBits);
		xField.m_ucThresholdType = static_cast<u_char>(eThresholdType);
		xField.m_ucFieldSizeInBytes = static_cast<u_char>(uFieldSize);
		xField.m_fThresholdValue = fThresholdValue;
		xField.m_fMinValue = fMinValue;
		xField.m_fMaxValue = fMaxValue;

		uByteOffset += (uFieldSize / uASURA_NETMSG_BITS_PER_BYTE);
	}
}


void Asura_NetMsg::EndRegistration()
{
#ifdef ASURA_ASSERT
	Asura_Assert(s_ucMaxFields == m_ucNumFields, "didn't register all fields");
	// check data size as well, last field's byte offset, plus its size, 
	// then the next multiple of 4 bytes should equal the struct size.
	const Asura_NetMsg_Field& xField = m_axFields[m_ucNumFields-1];
	u_int uNextMultipleOf4;
	if (xField.m_ucFieldType == Asura_FieldType_UnicodeString)
	{
		uNextMultipleOf4 = ( ( (xField.m_usByteOffset + (xField.m_ucFieldSizeInBytes)) + 3) / 4) * 4;
	}
	else
	{
		uNextMultipleOf4 = ( ( (xField.m_usByteOffset + (xField.m_ucFieldDataSourceSizeInBits / uASURA_NETMSG_BITS_PER_BYTE)) + 3) / 4) * 4;
	}

	Asura_Assert( (uNextMultipleOf4 == m_usDataSize) || ( (uNextMultipleOf4 == 4) && (m_usDataSize < 4) ), 
		"data offsets and struct size do not match. didn't register all data in struct?");
#endif//ASURA_ASSERT
	
	// compute max size of sent data
	u_int uTotalBytes = 0;
	u_int uTotalBits = 0;
	for (u_int u = 0; u < m_ucNumFields; ++u)
	{
		u_int uBytes;
		u_int uBits;
		m_axFields[u].GetSendDataSize( uBytes, uBits );
		uTotalBytes += uBytes;
		uTotalBits += uBits;
	}
	uTotalBytes += (uTotalBits + 7) / 8;
	Asura_Assert(uTotalBytes <= USHRT_MAX, "total size too large");
	m_usMaxSendDataSize = static_cast<u_short>(uTotalBytes);

	s_ucMaxFields = 0;
	s_uRegistrationBaseOffset = 0;
}


void Asura_NetMsg::SetBaseOffset(u_int uOffset)
{
	s_uRegistrationBaseOffset = uOffset;
}

const Asura_NetMsg* Asura_NetMsg::FindNetMsgObject(u_short usMsgID, u_short usEntityClassification)
{
	// first check for a specialised version for this entity classification, then try default.
	const Asura_NetMsg* pxNetMsg = s_xNetMsgList.FindObject(MakeGuid(usMsgID, usEntityClassification));
	if (!pxNetMsg)
	{
		pxNetMsg = s_xNetMsgList.FindObject(MakeGuid(usMsgID, AsuraEntityClass_NotSet));
	}
	return pxNetMsg;
}


// function to unpack to an instance of the update message
u_int Asura_NetMsg::UnpackToInstance( u_char* pucInstanceData, bool& bSkippedLastUpdate, Asura_BitUnpacker& xBitUnpacker ) const
{
	u_int uNumFieldsUnpacked = 0;

	while (xBitUnpacker.GetNumBitsRemaining() > 1)
	{
		// extract field index
		u_int uFieldIndex = xBitUnpacker.Unpack(uBITS_FOR_FIELD_INDEX);

		if (uFieldIndex == uFIELDID_ENDOFDATA)
		{
			Asura_Assert(uNumFieldsUnpacked > 0, "first field is end of data?");
			break;
		}
		
		uNumFieldsUnpacked++;

		if (uFieldIndex == uFIELDID_FULLUPDATE)
		{
			Asura_Assert(uNumFieldsUnpacked == 1, "full update value found mid-data");

			// decompress all the fields
			for (u_int uField = 0; uField < GetNumFields(); ++uField)
			{
				const Asura_NetMsg_Field& xField = GetField(uField);
				xField.ExtractFieldData( pucInstanceData, xBitUnpacker );
			}
			break;
		}
		else
		{
			const Asura_NetMsg_Field& xField = GetField(uFieldIndex);
			xField.ExtractFieldData( pucInstanceData, xBitUnpacker );
		}
	}

	bSkippedLastUpdate = xBitUnpacker.UnpackBool();

	return uNumFieldsUnpacked;
}

void Asura_NetMsg::PackWholeMessage( const void* pMessage, Asura_NetMsg_Recipients& xRecipients ) const
{
	// TODO: check message data length is correct.
	for (u_int uFieldIndex = 0; uFieldIndex < GetNumFields(); ++uFieldIndex)
	{
		const Asura_NetMsg_Field& xField = GetField(uFieldIndex);
		const u_char* pucCurrentValue = reinterpret_cast<const u_char*>(pMessage) + xField.m_usByteOffset;
		xField.AddDataToMessage( pucCurrentValue, xRecipients );
	}
}

void Asura_NetMsg::UnpackWholeMessage( void* pMessage, Asura_BitUnpacker& xBitUnpacker ) const
{
	u_char* const pucMessage = reinterpret_cast<u_char*>(pMessage);
	for (u_int uField = 0; uField < GetNumFields(); ++uField)
	{
		const Asura_NetMsg_Field& xField = GetField(uField);
		xField.ExtractFieldData( pucMessage, xBitUnpacker );
	}
}

Asura_NetMsg_MasterInstance* Asura_NetMsg::FindMasterInstance(Asura_Guid uGuid) const
{
	return m_pxMasterInstanceList->FindObject(uGuid);
}


Asura_NetMsg_SlaveInstance* Asura_NetMsg::FindSlaveInstance(Asura_Guid uGuid) const
{
	return m_pxSlaveInstanceList->FindObject(uGuid);
}


void Asura_NetMsg::AddToMasterInstanceList(Asura_NetMsg_MasterInstance* pxObject) const
{
	m_pxMasterInstanceList->AddToList(pxObject);
}


void Asura_NetMsg::AddToSlaveInstanceList(Asura_NetMsg_SlaveInstance* pxObject) const
{
	m_pxSlaveInstanceList->AddToList(pxObject);
}


void Asura_NetMsg::RemoveFromMasterInstanceList(Asura_NetMsg_MasterInstance* pxObject) const
{
	m_pxMasterInstanceList->RemoveFromList(pxObject);
}


void Asura_NetMsg::RemoveFromSlaveInstanceList(Asura_NetMsg_SlaveInstance* pxObject) const
{
	m_pxSlaveInstanceList->RemoveFromList(pxObject);
}


void Asura_NetMsg::UnregisterAll()
{
    Asura_Hashed_List_It<Asura_NetMsg> xIterator( &s_xNetMsgList );
    while ( !xIterator.Done() )
    {
        Asura_NetMsg* pxMsgToDelete = xIterator.GetCurrent();
		xIterator.Next();
		delete pxMsgToDelete;
    }
}

// A S U R A _ N E T M S G _ F I E L D ///////////////////////////////////////

// add this field index and data to the message (don't add field index if full update)
void Asura_NetMsg_Field::AddDataToMessage(const u_char* pucCurValue, Asura_NetMsg_Recipients& xRecipients) const
{
	switch (m_ucFieldType)
	{
		case Asura_FieldType_char:
		case Asura_FieldType_uchar:
		{
			xRecipients.AddBits(*reinterpret_cast<const u_char*>(pucCurValue), 8);
			break;
		}
		case Asura_FieldType_short:
		case Asura_FieldType_ushort:
		{
			xRecipients.AddBits(*reinterpret_cast<const u_short*>(pucCurValue), 16);
			break;
		}
		case Asura_FieldType_int:
		case Asura_FieldType_uint:
		case Asura_FieldType_float:
		{
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue), 32);
			break;
		}
		case Asura_FieldType_vector2:
		{
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue), 32);
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue + 4), 32);
			break;
		}
		case Asura_FieldType_vector3:
		case Asura_FieldType_euler3:
		{
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue), 32);
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue + 4), 32);
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue + 8), 32);
			break;
		}
		case Asura_FieldType_quat:
		{
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue), 32);
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue + 4), 32);
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue + 8), 32);
			xRecipients.AddBits(*reinterpret_cast<const u_int*>(pucCurValue + 12), 32);
			break;
		}
		case Asura_FieldType_UprightQuat:
		{
			const Asura_Quat& xCur = *reinterpret_cast<const Asura_Quat*>(pucCurValue);
			Asura_Assert(Asura_Maths::Abs(xCur.x) < 1e-6f && Asura_Maths::Abs(xCur.z) < 1e-6f, "not an upright quat");
			
			Asura_NetMsg_UprightQuatData xData;
			xData.FromQuat(xCur);

#if _DEBUG
			Asura_Quat xQuat;
			xData.ToQuat(xQuat);
			Asura_Assert(Asura_Maths::Abs(xQuat.x) < 1e-6 && Asura_Maths::Abs(xQuat.z) < 1e-6, "output quat not upright");
			Asura_Assert(Asura_Maths::Abs(xQuat.y - xCur.y) < dbg_fQuatApproximationThreshold, "bad quat approximation");
			Asura_Assert(Asura_Maths::Abs(xQuat.w - xCur.w) < dbg_fQuatApproximationThreshold, "bad quat approximation");
#endif

			xRecipients.AddBits(xData.m_bQuatYNegative, 1);
			xRecipients.AddBits(xData.m_usEulerY, 15);
			break;
		}

		case Asura_FieldType_CompressedQuat:
		{		
			const Asura_Quat& xCur = *reinterpret_cast<const Asura_Quat*>(pucCurValue);

			Asura_Comp_Quat_Fast xData; // Use fast compressed quaternions.
			xData.FromQuat(xCur);
			xRecipients.AddBits(xData.GetBits(), uNUM_BITS_FOR_COMP_QUAT_FAST);

#if _DEBUG
			Asura_Quat xQuat;
			xData.ToQuat(xQuat);
			Asura_Assert(xQuat.ApproximatelyEquals(xCur, dbg_fCompQuatFastApproximationThreshold), "Bad quat approximation");
#endif //_DEBUG
			break;	
		}

		case Asura_FieldType_CompressedPosition:
		{
			const Asura_Vector_3& xCur = *reinterpret_cast<const Asura_Vector_3*>(pucCurValue);
			
			static u_int uCompX, uCompY, uCompZ = 0;
			Asura_NetMsg_CompressedPositions::FromVector( xCur, uCompX, uCompY, uCompZ );

			xRecipients.AddBits(uCompX, Asura_NetMsg_CompressedPositions::GetNumBitsX() );
			xRecipients.AddBits(uCompY, Asura_NetMsg_CompressedPositions::GetNumBitsY() );
			xRecipients.AddBits(uCompZ, Asura_NetMsg_CompressedPositions::GetNumBitsZ() );

			break;
		}

		case Asura_FieldType_uint_bits:
		{
			u_int uData = 0;
			if (m_ucFieldDataSourceSizeInBits == 8)
			{
				uData = *pucCurValue;
			}
			else
			{
				Asura_Assert(m_ucFieldDataSourceSizeInBits == 32, "only expecting 8 or 32-bit fields for this");
				uData = *reinterpret_cast<const u_int*>(pucCurValue);
			}
			Asura_Assertf(uData < u_int(1<<m_ucFieldDataSendSizeInBits), ("value %u (0x%x) out of range for %u bits (%s)", 
				uData, uData, m_ucFieldDataSendSizeInBits, GetName()));
			xRecipients.AddBits( uData, m_ucFieldDataSendSizeInBits );
			break;
		}

		case Asura_FieldType_uint_Bitfield:
		{
#if 1 // new way that works cross-endian
			const u_int uSourceData = *reinterpret_cast<const u_int*>(pucCurValue);
			u_int uFields = 0;
			u_int uBitOffset = m_ucBitOffset;
			const u_int uEndOffset = m_ucBitOffset + m_ucFieldDataSendSizeInBits;
			while (uBitOffset < uEndOffset)
			{
				const int iNumBits = m_acNumBits[uFields++];
				Asura_Assert(iNumBits > 0, "invalid number of bits for bitfield");
				u_int uNumBits = static_cast<u_int>(iNumBits);
				u_int uMask = 0xffffffff;
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
				uMask >>= uBitOffset;
				u_int uData = uSourceData & uMask;
				uData >>= 32 - (uNumBits + uBitOffset);
#else
				uMask >>= 32 - (uNumBits + uBitOffset);
				u_int uData = uSourceData & uMask;
				uData >>= uBitOffset;
#endif

				// Add just the bits we need
				xRecipients.AddBits( uData, uNumBits );

				// and on to the next field
				uBitOffset += uNumBits;
			}

#else

			// Create a bitmask for the number of bits to send
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
			uMask <<= 32 - m_ucFieldDataSendSizeInBits;
			uMask >>= m_ucBitOffset;
#else
			uMask >>= 32 - m_ucFieldDataSendSizeInBits;
			uMask <<= m_ucBitOffset;
#endif
			// Grab the data, mask off the bits we need and shift them to the low end of the int
			u_int uData = *reinterpret_cast<const u_int*>(pucCurValue);
			uData &= uMask;
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
			uData >>= 32 - (m_ucBitOffset + m_ucFieldDataSendSizeInBits);
#else
			uData >>= m_ucBitOffset;
#endif
			// Add just the bits we need
			xRecipients.AddBits( uData, m_ucFieldDataSendSizeInBits );
#endif
			break;
		}

		case Asura_FieldType_UnicodeString:
		{
            // (MH - 21.06.2007) using u_short instead of Asura_UnicodeChar in case it happens to be signed.
			const u_short* wszString = reinterpret_cast<const u_short*>(pucCurValue);
			const u_short* wszString2 = wszString;
			u_int uLen = 0;
			u_int uMaxBitsCode = 0;
			u_int uMaxBits = 7;
			u_int uMaxLen = m_ucFieldSizeInBytes / sizeof(u_short);
			while (*wszString)
			{
				++uLen;
				if (uLen == uMaxLen)
				{
					Asura_Assert(false, "string buffer overrun");
					break;
				}

				u_int uVal = *wszString;
				// compute max bits required for string characters. only bothering with thresholds of 0-127 (ASCII)
				// 0-255 (extended ASCII) and 0-65535 (any Unicode string)
				if ( (uVal > 255) && (uMaxBitsCode < 2) )
				{ 
					uMaxBits = 16;
					uMaxBitsCode = 2; 
				}
				else if ( (uVal > 127) && (uMaxBitsCode < 1) )
				{ 
					uMaxBits = 8;
					uMaxBitsCode = 1; 
				}
				++wszString;
			}
			// add 7 bits for length (max length is 255/sizeof(u_short) = 127)
			xRecipients.AddBits(uLen, 7);
			// add 2 bits for max bits code
			xRecipients.AddBits(uMaxBitsCode, 2);
			// add the bits for each character
			while (uLen--)
			{
				xRecipients.AddBits(*wszString2++, uMaxBits);
			}
			break;
		}

		case Asura_FieldType_CompressedFloat:
		{
			const float fCurValue = *reinterpret_cast<const float*>(pucCurValue);
			float fScaledCurValue = fCurValue;
			Asura_Assert( fScaledCurValue >= m_fMinValue && fScaledCurValue <= m_fMaxValue, "float value out of range" );
			fScaledCurValue = Asura_Maths::ClampToRange(fScaledCurValue, m_fMinValue, m_fMaxValue);
			// convert to [0,1]
			fScaledCurValue -= m_fMinValue;
			fScaledCurValue /= (m_fMaxValue - m_fMinValue);
			Asura_Assert(fScaledCurValue >= 0.0f && fScaledCurValue <= 1.0f, "float conversion to [0,1] failed");
			// convert to [0,m_uRange]
			fScaledCurValue *= static_cast<float>((1<<m_ucFieldDataSendSizeInBits) - 1);
			u_int uValueToSend = Asura_Maths::FloatToInt_Chop(fScaledCurValue);
			// (MH - 28.04.2004) ensure we don't send m_fMinValue unless original value is m_fMinValue
			if ( (uValueToSend == 0) && (fCurValue != m_fMinValue) )
			{
				uValueToSend = 1;
			}
			xRecipients.AddBits( uValueToSend, m_ucFieldDataSendSizeInBits );
			break;
		}

		case Asura_FieldType_Guid:
		{
			Asura_Guid uGuid = *reinterpret_cast<const Asura_Guid*>(pucCurValue);
			xRecipients.AddGuid( uGuid );
			break;
		}

		default:
		{
			Asura_Assert(false, "unhandled case");
			break;
		}
	}
}

void Asura_NetMsg_Field::GetSendDataSize(u_int& uBytes, u_int& uBits) const
{
	switch (m_ucFieldType)
	{
		default:
		{
			uBits = m_ucFieldDataSendSizeInBits;
			uBytes = 0;
			break;
		}

		case Asura_FieldType_UprightQuat:
		{
			Asura_NetMsg_UprightQuatData xData;
			uBits = 0;
			uBytes = sizeof(xData);
			break;
		}

		case Asura_FieldType_CompressedQuat:
		{		
			uBits = uNUM_BITS_FOR_COMP_QUAT_FAST;
			uBytes = 0;
			break;	
		}

		case Asura_FieldType_CompressedPosition:
		{
			uBits += Asura_NetMsg_CompressedPositions::GetNumBitsX() + Asura_NetMsg_CompressedPositions::GetNumBitsY() + Asura_NetMsg_CompressedPositions::GetNumBitsZ();
			uBytes = 0;
			break;
		}

		case Asura_FieldType_uint_bits:
		{
			uBits = m_ucFieldDataSendSizeInBits;
			uBytes = 0;
			break;
		}

		case Asura_FieldType_uint_Bitfield:
		{
			uBits = m_ucFieldDataSendSizeInBits;
			uBytes = 0;
			break;
		}

		case Asura_FieldType_UnicodeString:
		{
			uBits = 7 + 2;					// length and max bits code
			uBytes = m_ucFieldSizeInBytes;	// max size is entire field size
			break;
		}

		case Asura_FieldType_CompressedFloat:
		{
			uBits = m_ucFieldDataSendSizeInBits;
			uBytes = 0;
			break;
		}

		case Asura_FieldType_Guid:
		{
			uBits = 32;
			uBytes = 0;
			break;
		}
	}
}


bool Asura_NetMsg_Field::HasIntegerChanged(const u_char* pucCurrentValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	u_int uBytes = m_ucFieldDataSourceSizeInBits / uASURA_NETMSG_BITS_PER_BYTE;

	u_int uCur;
	u_int uLast;
	if (uBytes == 4)
	{
		uCur = *reinterpret_cast<const u_int*>(pucCurrentValue);
		uLast = *reinterpret_cast<u_int*>(pucLastSentValue);
	}
	else if (uBytes == 1)
	{
		uCur = *pucCurrentValue;
		uLast = *pucLastSentValue;
	}
	else if (uBytes == 2)
	{
		uCur = *reinterpret_cast<const u_short*>(pucCurrentValue);
		uLast = *reinterpret_cast<u_short*>(pucLastSentValue);
	}
	else
	{
		Asura_Assert(false, "unhandled field size");
		return false;
	}
	
	bool bChanged = false;

	if (m_ucThresholdType == Asura_Threshold_AnyChange)
	{
		if (uCur != uLast)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_NotEqual)
	{
		// could easily be changed so the float is a union
		Asura_Assert(m_fThresholdValue == 0.0f, "only supporting != 0");
		if (uCur != 0)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_ConstantValue)
	{
	}
	else
	{
		Asura_Assert(false, "unhandled threshold type");
	}
	
	if (bChanged && bReplaceLastValue)
	{
		if (uBytes == 4)
		{
			*reinterpret_cast<u_int*>(pucLastSentValue) = uCur;
		}
		else if (uBytes == 1)
		{
			*pucLastSentValue = *pucCurrentValue;
		}
		else if (uBytes == 2)
		{
			*reinterpret_cast<u_short*>(pucLastSentValue) = *reinterpret_cast<const u_short*>(pucCurrentValue);
		}
	}

	return bChanged;
}

bool Asura_NetMsg_Field::HasBitfieldChanged(const u_char* pucCurValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	// Create a bitmask for the number of bits received
	u_int uMask = 0xffffffff;
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
	uMask <<= 32 - m_ucFieldDataSendSizeInBits;
	uMask >>= m_ucBitOffset;
#else
	uMask >>= 32 - m_ucFieldDataSendSizeInBits;
	uMask <<= m_ucBitOffset;
#endif

	const u_int uMaskedData = *reinterpret_cast<const u_int*>(pucCurValue) & uMask;
	const u_int uMaskedLast = *reinterpret_cast<const u_int*>(pucLastSentValue) & uMask;

	// Shift the current data for comparison
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
	const u_int uData = uMaskedData << m_ucBitOffset;
	const u_int uLast = uMaskedLast << m_ucBitOffset;
#else
	const u_int uData = uMaskedData >> m_ucBitOffset;
	const u_int uLast = uMaskedLast >> m_ucBitOffset;
#endif

	bool bChanged = false;
	switch (m_ucThresholdType)
	{
		case Asura_Threshold_AnyChange:
		{
			if (uData != uLast)
			{
				bChanged = true;
			}
			break;
		}

		case Asura_Threshold_ConstantValue:
			break;

		default:
		{
			Asura_Assert(false, "unhandled threshold type");
			break;
		}
	}

	// If the data has changed, store it for comparison next time
	if (bChanged && bReplaceLastValue)
	{
		// Mask out the old bits and OR in the received bits
		*reinterpret_cast<u_int*>(pucLastSentValue) &= ~uMask;
		*reinterpret_cast<u_int*>(pucLastSentValue) |= uMaskedData;
	}

	return bChanged;
}

bool Asura_NetMsg_Field::HasStringChanged(const u_char* pucCurValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{

	bool bChanged = false;
	switch (m_ucThresholdType)
	{
		case Asura_Threshold_AnyChange:
		{
			const Asura_UnicodeChar* pwcCur = reinterpret_cast<const Asura_UnicodeChar*>(pucCurValue);
			const Asura_UnicodeChar* pwcPrev = reinterpret_cast<const Asura_UnicodeChar*>(pucLastSentValue);
			bChanged = wcscmp( pwcCur, pwcPrev ) != 0;
			break;
		}

		case Asura_Threshold_ConstantValue:
			break;

		default:
		{
			Asura_Assert(false, "unhandled threshold type");
			break;
		}
	}

	// If the data has changed, store it for comparison next time
	if (bChanged && bReplaceLastValue)
	{
		Asura_Mem::Copy( pucLastSentValue, pucCurValue, m_ucFieldSizeInBytes );
	}

	return bChanged;
}
bool Asura_NetMsg_Field::HasFloatChanged(const u_char* pucCurrentValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	float fCur = *reinterpret_cast<const float*>(pucCurrentValue);
	float fLast = *reinterpret_cast<const float*>(pucLastSentValue);
	
	bool bChanged = false;

	if (m_ucThresholdType == Asura_Threshold_AnyChange)
	{
		if (fCur != fLast)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_NotEqual)
	{
		if (fCur != m_fThresholdValue)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_Magnitude)
	{
		if (Asura_Maths::Abs(fCur - fLast) > m_fThresholdValue)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_ConstantValue)
	{
	}
	else
	{
		Asura_Assert(false, "unhandled threshold type");
	}

	if (bChanged && bReplaceLastValue)
	{
		*reinterpret_cast<float*>(pucLastSentValue) = fCur;
	}

	return bChanged;			
}

bool Asura_NetMsg_Field::HasVector3Changed(const u_char* pucCurrentValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	const Asura_Vector_3& xCur = *reinterpret_cast<const Asura_Vector_3*>(pucCurrentValue);
	const Asura_Vector_3& xLast = *reinterpret_cast<const Asura_Vector_3*>(pucLastSentValue);

	bool bChanged = false;
	if (m_ucThresholdType == Asura_Threshold_AnyChange)
	{
		if (xCur != xLast)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_Magnitude)
	{
		// NB this uses magnitude squared - see RegisterField
		if (!xCur.ApproximatelyEquals(xLast, m_fThresholdValue))
		{
			bChanged = true;
		}
	}
	else
	{
		Asura_Assert(false, "unhandled threshold type");
	}
	
	if (bChanged && bReplaceLastValue)
	{
		Asura_Vector_3* pxLast = reinterpret_cast<Asura_Vector_3*>(pucLastSentValue);
		*pxLast = xCur;
	}

	return bChanged;
}

bool Asura_NetMsg_Field::HasVector2Changed(const u_char* pucCurrentValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	const Asura_Vector_2& xCur = *reinterpret_cast<const Asura_Vector_2*>(pucCurrentValue);
	const Asura_Vector_2& xLast = *reinterpret_cast<const Asura_Vector_2*>(pucLastSentValue);

	bool bChanged = false;

	if (xCur != xLast)
	{
		bChanged = true;
	}
	
	if (bChanged && bReplaceLastValue)
	{
		Asura_Vector_2* pxLast = reinterpret_cast<Asura_Vector_2*>(pucLastSentValue);
		*pxLast = xCur;
	}

	return bChanged;
}

bool Asura_NetMsg_Field::HasEuler3Changed(const u_char* pucCurrentValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	const Asura_Euler_3& xCur = *reinterpret_cast<const Asura_Euler_3*>(pucCurrentValue);
	const Asura_Euler_3& xLast = *reinterpret_cast<const Asura_Euler_3*>(pucLastSentValue);
	
	bool bChanged = false;
	if (m_ucThresholdType == Asura_Threshold_AnyChange)
	{
		if (xCur.x != xLast.x || xCur.y != xLast.y || xCur.z != xLast.z)
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_Magnitude)
	{
		// not too sure what to use for euler threshold. Try any element exceeds threshold.
		float fThreshold = m_fThresholdValue;
		if ( (Asura_Maths::Abs(xCur.x - xLast.x) > fThreshold) || 
			(Asura_Maths::Abs(xCur.y - xLast.y) > fThreshold) || 
			(Asura_Maths::Abs(xCur.z - xLast.z) > fThreshold) )
		{
			bChanged = true;
		}
	}
	else
	{
		Asura_Assert(false, "unhandled threshold type");
	}
	
	if (bChanged && bReplaceLastValue)
	{
		Asura_Euler_3* pxLast = reinterpret_cast<Asura_Euler_3*>(pucLastSentValue);
		*pxLast = xCur;
	}
	return bChanged;
}

bool Asura_NetMsg_Field::HasQuatChanged(const u_char* pucCurrentValue, u_char* pucLastSentValue, bool bReplaceLastValue) const
{
	const Asura_Quat& xCur = *reinterpret_cast<const Asura_Quat*>(pucCurrentValue);
	const Asura_Quat& xLast = *reinterpret_cast<const Asura_Quat*>(pucLastSentValue);

	bool bChanged = false;
	if (m_ucThresholdType == Asura_Threshold_AnyChange)
	{
		if (!(xCur == xLast))
		{
			bChanged = true;
		}
	}
	else if (m_ucThresholdType == Asura_Threshold_Magnitude)
	{
		// NB this uses cosine omega - see RegisterField
		if (!xCur.ApproximatelyEquals(xLast, m_fThresholdValue))
		{
			bChanged = true;
		}
	}
	else
	{
		Asura_Assert(false, "unhandled threshold type");
	}
	
	if (bChanged && bReplaceLastValue)
	{
		Asura_Quat* pxLast = reinterpret_cast<Asura_Quat*>(pucLastSentValue);
		*pxLast = xCur;
	}

	return bChanged;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_NetMsg_Field::HasChanged().
// PARAMETERS   pucCurValue      - ptr to current value
//              pucLastSentValue - ptr to last value sent
// RETURNS      
// DESCRIPTION  Checks whether the field needs to be sent based on the current
//				and previous values and the threshold type
//				If it returns true and bReplaceLastValue is true, it writes 
//				the value in pucCurValue to pucLastSentValue.
// AUTHOR       Mike Healey - 01.03.2004.
//
//////////////////////////////////////////////////////////////////////////////
bool Asura_NetMsg_Field::HasChanged( const u_char* pucCurValue, u_char* pucLastSentValue, bool bReplaceLastValue ) const
{
	switch (m_ucFieldType)
	{
	case Asura_FieldType_char:
	case Asura_FieldType_uchar:
	case Asura_FieldType_short:
	case Asura_FieldType_ushort:
	case Asura_FieldType_int:
	case Asura_FieldType_uint:
	case Asura_FieldType_uint_bits:
	case Asura_FieldType_Guid:
		return HasIntegerChanged(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_uint_Bitfield:
		return HasBitfieldChanged(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_CompressedFloat:
	case Asura_FieldType_float:
		return HasFloatChanged(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_vector2:
		return HasVector2Changed(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_vector3:
	case Asura_FieldType_CompressedPosition:
		return HasVector3Changed(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_euler3:
		return HasEuler3Changed(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_quat:
	case Asura_FieldType_UprightQuat:
	case Asura_FieldType_CompressedQuat:
		return HasQuatChanged(pucCurValue, pucLastSentValue, bReplaceLastValue);
	case Asura_FieldType_UnicodeString:
		return HasStringChanged(pucCurValue, pucLastSentValue, bReplaceLastValue);
	default:
		Asura_Assert(false, "unhandled case");
		return false;
	}

}
// End function Asura_NetMsg_Field::ShouldSend().


void Asura_NetMsg_Field::ExtractFieldData( u_char* pucInstanceCurData, Asura_BitUnpacker& xBitUnpacker ) const
{
	u_char* pucFieldData = pucInstanceCurData + m_usByteOffset;
	switch (m_ucFieldType)
	{
		case Asura_FieldType_char:
		case Asura_FieldType_uchar:
		{
			*reinterpret_cast<u_char*>(pucFieldData) = static_cast<u_char>(xBitUnpacker.Unpack(8));
			break;
		}

		case Asura_FieldType_short:
		case Asura_FieldType_ushort:
		{
			*reinterpret_cast<u_short*>(pucFieldData) = static_cast<u_short>(xBitUnpacker.Unpack(16));
			break;
		}

		case Asura_FieldType_int:
		case Asura_FieldType_uint:
		case Asura_FieldType_float:
		{
			*reinterpret_cast<u_int*>(pucFieldData) = xBitUnpacker.Unpack(32);
			break;
		}

		case Asura_FieldType_vector2:
		{
			*reinterpret_cast<u_int*>(pucFieldData) = xBitUnpacker.Unpack(32);
			*reinterpret_cast<u_int*>(pucFieldData + 4) = xBitUnpacker.Unpack(32);
			break;
		}

		case Asura_FieldType_vector3:
		case Asura_FieldType_euler3:
		{
			*reinterpret_cast<u_int*>(pucFieldData) = xBitUnpacker.Unpack(32);
			*reinterpret_cast<u_int*>(pucFieldData + 4) = xBitUnpacker.Unpack(32);
			*reinterpret_cast<u_int*>(pucFieldData + 8) = xBitUnpacker.Unpack(32);
			break;
		}

		case Asura_FieldType_quat:
		{
			*reinterpret_cast<u_int*>(pucFieldData) = xBitUnpacker.Unpack(32);
			*reinterpret_cast<u_int*>(pucFieldData + 4) = xBitUnpacker.Unpack(32);
			*reinterpret_cast<u_int*>(pucFieldData + 8) = xBitUnpacker.Unpack(32);
			*reinterpret_cast<u_int*>(pucFieldData + 12) = xBitUnpacker.Unpack(32);
			break;
		}

		case Asura_FieldType_UprightQuat:
		{
			// unpack to temp location
			Asura_NetMsg_UprightQuatData xData;
			xData.m_bQuatYNegative = xBitUnpacker.Unpack(1);
			xData.m_usEulerY = xBitUnpacker.Unpack(15);
			xData.ToQuat(*reinterpret_cast<Asura_Quat*>(pucFieldData));
			break;
		}

		case Asura_FieldType_CompressedQuat:
		{
			// unpack to temp location
			Asura_Comp_Quat_Fast xData;
			xData.SetBits( xBitUnpacker.Unpack(uNUM_BITS_FOR_COMP_QUAT_FAST) );

			xData.ToQuat(*reinterpret_cast<Asura_Quat*>(pucFieldData));
			break;
		}

		case Asura_FieldType_CompressedPosition:
		{
			Asura_Vector_3 xVector3;

			u_int uX = xBitUnpacker.Unpack(Asura_NetMsg_CompressedPositions::GetNumBitsX());
			u_int uY = xBitUnpacker.Unpack(Asura_NetMsg_CompressedPositions::GetNumBitsY());
			u_int uZ = xBitUnpacker.Unpack(Asura_NetMsg_CompressedPositions::GetNumBitsZ());

			Asura_NetMsg_CompressedPositions::ToVector( *reinterpret_cast<Asura_Vector_3*>(pucFieldData), uX, uY, uZ );
			break;
		}

		case Asura_FieldType_CompressedFloat:
		{
			u_int uValue = xBitUnpacker.Unpack(m_ucFieldDataSendSizeInBits);
			float fValue = static_cast<float>(uValue);
			fValue /= static_cast<float>((1<<m_ucFieldDataSendSizeInBits) - 1);
			fValue *= (m_fMaxValue - m_fMinValue);
			fValue += m_fMinValue;
			Asura_Assert( fValue >= m_fMinValue && fValue <= m_fMaxValue, "float value out of range" );
			*reinterpret_cast<float*>(pucFieldData) = fValue;
			break;
		}
		
		case Asura_FieldType_uint_bits:
		{
			u_int uData = xBitUnpacker.Unpack(m_ucFieldDataSendSizeInBits);
			if (m_ucFieldDataSourceSizeInBits == 8)
			{
				*pucFieldData = static_cast<u_char>(uData);
			}
			else
			{
				Asura_Assert(m_ucFieldDataSourceSizeInBits == 32, "only expecting 8 or 32-bit fields for this");
				*reinterpret_cast<u_int*>(pucFieldData) = uData;
			}
			break;
		}

		case Asura_FieldType_uint_Bitfield:
		{

#if 1 // new way that works cross-endian
			u_int uReceivedData = 0;
			u_int uFields = 0;
			u_int uBitOffset = m_ucBitOffset;
			const u_int uEndOffset = m_ucFieldDataSendSizeInBits + uBitOffset;
			while (uBitOffset < uEndOffset)
			{
				const int iNumBits = m_acNumBits[uFields++];
				Asura_Assert(iNumBits > 0, "invalid number of bits for bitfield");
				u_int uNumBits = static_cast<u_int>(iNumBits);
				const u_int uData = xBitUnpacker.Unpack(uNumBits);

#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
				uReceivedData |= uData << (32 - (uNumBits + uBitOffset));
#else
				uReceivedData |= uData << uBitOffset;
#endif

				// and on to the next field
				uBitOffset += uNumBits;
			}

#else
			u_int uReceivedData = xBitUnpacker.Unpack(m_ucFieldDataSendSizeInBits);

			// Shift received bitfield back into position
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
			uReceivedData <<= 32 - (m_ucBitOffset + m_ucFieldDataSendSizeInBits);
#else
			uReceivedData <<= m_ucBitOffset;
#endif
#endif // 0

			// Create a bitmask for the number of bits received
			u_int uMask = 0xffffffff;
#ifdef ASURA_PLATFORM_IS_BIG_ENDIAN
			uMask <<= 32 - m_ucFieldDataSendSizeInBits;
			uMask >>= m_ucBitOffset;
#else
			uMask >>= 32 - m_ucFieldDataSendSizeInBits;
			uMask <<= m_ucBitOffset;
#endif
			// Mask out the old bits and OR in the received bits
			*reinterpret_cast<u_int*>(pucFieldData) &= ~uMask;
			*reinterpret_cast<u_int*>(pucFieldData) |= uReceivedData;


			break;
		}

		case Asura_FieldType_UnicodeString:
		{
			// 7 bits for length (max length is 255/sizeof(Asura_UnicodeChar) = 127)
			u_int uLen = xBitUnpacker.Unpack(7);
			Asura_Assert(uLen < (m_ucFieldSizeInBytes / sizeof(Asura_UnicodeChar)), "length too long");
			uLen = Asura_Maths::Min( uLen, (m_ucFieldSizeInBytes / sizeof(Asura_UnicodeChar)) - 1);
			// add 2 bits for max bits code
			u_int uMaxBitsCode = xBitUnpacker.Unpack(2);
			u_int uMaxBits;
			if (uMaxBitsCode == 2)
			{
				uMaxBits = 16;
			}
			else if (uMaxBitsCode == 1)
			{
				uMaxBits = 8;
			}
			else
			{
				uMaxBits = 7;
			}
			// extract the string
			Asura_UnicodeChar* wszString = reinterpret_cast<Asura_UnicodeChar*>(pucFieldData);
			while (uLen--)
			{
				*wszString++ = static_cast<Asura_UnicodeChar>(xBitUnpacker.Unpack(uMaxBits));
			}
			// NULL terminate.
			*wszString = 0;
			break;
		}

		case Asura_FieldType_Guid:
		{
			*reinterpret_cast<Asura_Guid*>(pucFieldData) = Asura_MessagePacking::UnpackGuid(&xBitUnpacker);
			break;
		}

		default:
		{
			Asura_Assert(false, "unhandled case");
			break;
		}
	}
}



// A S U R A _ N E T M S G _ M A S T E R I N S T A N C E /////////////////////
Asura_NetMsg_MasterInstance::Asura_NetMsg_MasterInstance(Asura_Guid uGuid, 
														 const Asura_NetMsg* pxNetMsg, 
														 float fHeartbeatInterval, 
														 const void* pInitialData)
: m_uGuid(uGuid)
, m_pxNetMsg(pxNetMsg)
, m_uHeartbeat(Asura_Maths::FloatToInt(1000.0f * fHeartbeatInterval))
, m_bSentAnythingLastUpdate(true)
{
	// (MH - 27.02.2004) now batching up full updates into groups of 200k to avoid
	// spreading updates out too much.
	g_uFullUpdateBatchSize += pxNetMsg->GetMaxSendDataSize() + GetHeaderSize();
	
	if (g_uFullUpdateBatchSize >= dbg_uMaxFullUpdateBatchSize)
	{
		g_uFullUpdateBatchSize = 0;

		g_uInitialHeartbeatTimerValue += uASURA_NETWORK_MSECS_BETWEEN_ENTITY_UPDATES;
		if (g_uInitialHeartbeatTimerValue > dbg_uMaxInitialHeartbeatTimerValue)
		{
			g_uInitialHeartbeatTimerValue = 0;
		}
	}
	// (MH - 04.09.2003) set the heartbeat timer's initial value 0.1s earlier than 
	// the previous entity's. This avoids all the updates happening at once. All 
	// the entities will be updated the first time anyway.
	Asura_Assert(Asura_Timers::GetMSecsSinceStart() >= g_uInitialHeartbeatTimerValue, "timer problem");	
	if( Asura_Timers::GetMSecsSinceStart() < g_uInitialHeartbeatTimerValue )
	{
		m_uTimeOfNextFullUpdate = Asura_Timers::GetMSecsSinceStart();
	}
	else
	{
		m_uTimeOfNextFullUpdate = Asura_Timers::GetMSecsSinceStart() - g_uInitialHeartbeatTimerValue;
	}

	u_int uDataSize = pxNetMsg->GetDataSize();
	m_pucCurrentData = new u_char[uDataSize];
	Asura_Mem::Copy(m_pucCurrentData, pInitialData, uDataSize);
	m_pucLastSentData = new u_char[uDataSize];
	Asura_Mem::Copy(m_pucLastSentData, pInitialData, uDataSize);
	m_pcFieldChangedLastSend = new char[pxNetMsg->GetNumFields()];
	for (u_int u = 0; u < pxNetMsg->GetNumFields(); ++u)
	{
		m_pcFieldChangedLastSend[u] = 0;
	}

	// initialise the array of how many times this entity was skipped for each player due to distance
	if (Asura_NetMsg_System::AreLODUpdatesAllowed())
	{
		m_aucNumDropped = new u_char[Asura_Network::GetMaxNumberOfMachines()];
		for (u_int u = 0; u < Asura_Network::GetMaxNumberOfMachines(); ++u)
		{
			m_aucNumDropped[u] = 0;
		}
	}
	else
	{
		m_aucNumDropped = NULL;
	}

	pxNetMsg->AddToMasterInstanceList(this);
}


Asura_NetMsg_MasterInstance::~Asura_NetMsg_MasterInstance()
{
	delete[] m_pucCurrentData;
	delete[] m_pucLastSentData;
	delete[] m_pcFieldChangedLastSend;
	delete[] m_aucNumDropped;

	m_pxNetMsg->RemoveFromMasterInstanceList(this);
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_NetMsg_MasterInstance::GetHeaderSize().
// PARAMETERS   None.
// RETURNS      
// DESCRIPTION  returns number of bytes used for the instance header, i.e. 
//				msgID, EntClass, guid, datasize, flags
// AUTHOR       Mike Healey - 27.02.2004.
//
//////////////////////////////////////////////////////////////////////////////
u_int Asura_NetMsg_MasterInstance::GetHeaderSize()
{
	// todo - calculate compreassed header size.
	// todo - the first 1 is for field index FULL_UPDATE. There is also flags which could do this job!
	return sizeof(u_char) + sizeof(u_short) + sizeof(u_short) + sizeof(Asura_Guid) + sizeof(u_char) + sizeof(u_char);
}
// End function Asura_NetMsg_MasterInstance::GetHeaderSize().


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_NetMsg_MasterInstance::DoFieldUpdate().
// PARAMETERS   uFieldIndex - index of field to update
//				bIsFullUpdate - whether field should be sent regardless of 
//				whether it changed. Used for heartbeat.
// RETURNS      Nothing.
// DESCRIPTION  check update thresholds on all fields. If any field changes, 
//				flag this fact in the array m_pcFieldChangedLastSend. If the 
//				flag was set last time, send the field again to improve 
//				stability on the client in packet-loss conditions.
// AUTHOR       Mike Healey - 01.03.2004.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_NetMsg_MasterInstance::DoFieldUpdate( u_int uFieldIndex, bool bIsFullUpdate )
{
	const Asura_NetMsg_Field& xField = m_pxNetMsg->GetField(uFieldIndex);
	if ( (!bIsFullUpdate) && (xField.m_ucThresholdType == Asura_Threshold_ConstantValue) )
	{
		return;
	}

	const u_int uByteOffset = xField.m_usByteOffset;
	const u_char* const pucCurrentValue = m_pucCurrentData + uByteOffset;
	u_char* const pucLastSentValue = m_pucLastSentData + uByteOffset;
	
	bool bSendField = m_pcFieldChangedLastSend[uFieldIndex] != 0;

	if (xField.HasChanged(pucCurrentValue, pucLastSentValue))
	{
		bSendField = true;
		m_pcFieldChangedLastSend[uFieldIndex] = 3;
	}
	else
	{
		if (--m_pcFieldChangedLastSend[uFieldIndex] < 0) m_pcFieldChangedLastSend[uFieldIndex] = 0;
	}

	if (bSendField || bIsFullUpdate)
	{
		if (!bIsFullUpdate)
		{
			Asura_NetMsg_System::AddMessageDataFieldID(uFieldIndex);
		}
		xField.AddDataToMessage( pucCurrentValue, *Asura_NetMsg_System::s_pxRecipients );
	}
}
// End function Asura_NetMsg_MasterInstance::DoFieldUpdate().


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_NetMsg_MasterInstance::SendDataUpdate().
// DESCRIPTION  creates the update data packet.
// AUTHOR       Mike Healey - 25.04.2003.
// MODIFIED		Mike Healey - 01.03.2004.
//
//////////////////////////////////////////////////////////////////////////////
bool Asura_NetMsg_MasterInstance::SendDataUpdate( const Asura_Message* pxMsg )
{
	const bool bNeedFullUpdate = m_uTimeOfNextFullUpdate <= Asura_Timers::GetMSecsSinceStart();

	// (MH - 06.07.2004) first we prepare the list of clients interested in this update
	// if a player is on the server machine (machine 0) it gets the message via the entity
	// system via Asura_Network::Send so we don't need to include it here.

	if (Asura_NetMsg_System::IsUsingLODUpdates())
	{
		Asura_NetMsg_System::s_uNumClientsForCurMessage = 0;
		
		for (u_int uPlayer = 0; uPlayer < Asura_NetMsg_System::s_uNumClientsForThisUpdate; ++uPlayer)
		{
			const Asura_NetMsg_UpdateData& xUpdateData = Asura_NetMsg_System::s_axUpdateData[uPlayer];
			// Allow projects to cull messages not already culled by the LOD mechanism
			bool bProjectInclude = (Asura_NetMsg_System::s_pfnProjectShouldSendClientMessage) ? Asura_NetMsg_System::s_pfnProjectShouldSendClientMessage(xUpdateData.GetGuid(), pxMsg) : true;

			bool bInclude;
			Asura_Handle iMachineForPlayer = Asura_Network::GetMachineForPlayer(xUpdateData.GetGuid(), false);
			if ( (0 == iMachineForPlayer) || (ASURA_INVALID_HANDLE == iMachineForPlayer) )
			{
				bInclude = false;
			}
			else if (bNeedFullUpdate)
			{
				bInclude = bProjectInclude;
			}
			else
			{

				// bit of a hack (like player position) to get object's position. - assuming it is derived from PhysicalObject
				const Asura_ServerEntity_PhysicalObject* pxEntity;
				if (xUpdateData.IsPositionValid())
				{
					pxEntity = Asura_ServerEntity_PhysicalObject::FindEntity(m_uGuid);
				}
				else
				{
					pxEntity = 0;
				}

				if (pxEntity)
				{
					Asura_Vector_3 xDiff = xUpdateData.GetPosition() - pxEntity->GetPosition();
					float fDistSquared = xDiff.MagnitudeSquared();

					u_int uThreshold;
					for (uThreshold = 0; uThreshold < uASURA_NETMSG_NUM_LOD_THRESHOLDS; ++uThreshold)
					{
						if (fDistSquared < dbg_afNetMsgLodThresholds[uThreshold])
						{
							break;
						}
					}

					// Test for being in modules that aren't linked together for visibility
					// Since we'll jump to the furthest threshold, we don't need to test that case.
					if ( uThreshold < (uASURA_NETMSG_NUM_LOD_THRESHOLDS - 1) )
					{
						int iPlayerModule = xUpdateData.GetModuleIndex();
						int iEntityModule = Asura_Module_System::GetModuleIndexFromPosition(pxEntity->GetPosition());
						if ( (iPlayerModule != -1) && (iEntityModule != -1) )
						{
							if (!Asura_Module_System::IsModuleLinked(iPlayerModule, iEntityModule))
							{
								// not linked, 
								// If the object is outside of the first threshold set to furthest threshold.
								// If it's within the first threshold, move it up one level
								if (uThreshold > 0)
								{
									uThreshold = uASURA_NETMSG_NUM_LOD_THRESHOLDS - 1;
								}
								else
								{
									uThreshold = 1;
								}
							}
						}
					}

					// drop packets based on which threshold it fell in.
					u_int uNumToDrop = uThreshold;
					if (m_aucNumDropped[uPlayer] >= uNumToDrop)
					{
						bInclude = bProjectInclude;
						m_aucNumDropped[uPlayer] = 0;
					}
					else
					{
						bInclude = false;
						m_aucNumDropped[uPlayer]++;
					}
				}
				else
				{
					bInclude = bProjectInclude;
				}
			}

			if (bInclude)
			{
				u_int uIndex = Asura_NetMsg_System::s_uNumClientsForCurMessage++;
				Asura_NetMsg_System::s_apxClientsForCurMessage[uIndex] = &Asura_NetMsg_System::s_axUpdateData[uPlayer];
			}
		}

		Asura_NetMsg_System::s_pxRecipients->ChangeArraySize(Asura_NetMsg_System::s_uNumClientsForCurMessage);
	}
	else
	{
		Asura_Assert(!Asura_NetMsg_System::s_pfnProjectShouldSendClientMessage, "not currently supported unless using LOD updates");

		// not using LOD updates
		Asura_NetMsg_System::s_apxClientsForCurMessage[0] = &Asura_NetMsg_System::s_axUpdateData[0];
		Asura_NetMsg_System::s_uNumClientsForCurMessage = 1;
		Asura_NetMsg_System::s_pxRecipients->ChangeArraySize(1);
	}

	// (MH - 31.01.2007) cope with larger update sets by sending an extra Asura message.
	if (!Asura_NetMsg_System::BeginMessage(m_pxNetMsg->GetMsgID(), m_pxNetMsg->GetEntityClassification(), m_uGuid, m_pxNetMsg->GetMaxSendDataSize(), m_pxNetMsg->GetNumFields()))
	{
		Asura_NetMsg_System::SendUpdate();
		Asura_NetMsg_System::BeginUpdate();
	
		if (!Asura_NetMsg_System::BeginMessage(m_pxNetMsg->GetMsgID(), m_pxNetMsg->GetEntityClassification(), m_uGuid, m_pxNetMsg->GetMaxSendDataSize(), m_pxNetMsg->GetNumFields()))
		{
			Asura_Assert(false, "no room for update package");
			return false;
		}
	}

	bool bNeedToAddEndOfDataField = false;
	if ( bNeedFullUpdate )
	{
		//Asura_OutputToDebugger(("%d, heartbeat for %d.\n", Asura_Timers::GetFrameNumber(), m_uGuid));
		// send full update
		Asura_NetMsg_System::AddMessageDataFieldID(uFIELDID_FULLUPDATE);
		for (u_int u = 0; u < m_pxNetMsg->GetNumFields(); ++u)
		{
			// force update = true
			DoFieldUpdate(u, true);
		}

		// MGH 22/2/2006: Calculate the next update time. Fixed a bug in the previous version
		// which caused the heartbeats to cluster.

		// This should be true so long as nobody messes with the code at the top of the function.
		Asura_Assert(Asura_Timers::GetMSecsSinceStart() >= m_uTimeOfNextFullUpdate, "Error in heartbeat timers!");
		// Add as many multiples of the heartbeat as necessary to bring it into the future.
		const u_int uMultipier = 1 + (Asura_Timers::GetMSecsSinceStart() - m_uTimeOfNextFullUpdate)/m_uHeartbeat;
		m_uTimeOfNextFullUpdate += (uMultipier * m_uHeartbeat);
		Asura_Assert(Asura_Timers::GetMSecsSinceStart() < m_uTimeOfNextFullUpdate, "Next full update should be in the future!");
	}
	else
	{
		for (u_int u = 0; u < m_pxNetMsg->GetNumFields(); ++u)
		{
			DoFieldUpdate(u, false);
		}
		bNeedToAddEndOfDataField = true;
	}

	bool bSkippedLastUpdate = !m_bSentAnythingLastUpdate;
	m_bSentAnythingLastUpdate = Asura_NetMsg_System::EndMessage(bNeedToAddEndOfDataField, bSkippedLastUpdate);
	return true;
}
// End function Asura_NetMsg_MasterInstance::SendDataUpdate().




// A S U R A _ N E T M S G _ S L A V E I N S T A N C E ///////////////////////
Asura_NetMsg_SlaveInstance::Asura_NetMsg_SlaveInstance(Asura_Guid uGuid, const Asura_NetMsg* pxNetMsg)
: m_uGuid(uGuid)
, m_pxNetMsg(pxNetMsg)
{
	m_pucCurrentData = new u_char[pxNetMsg->GetDataSize()];
	memset(m_pucCurrentData, 0, pxNetMsg->GetDataSize());

	pxNetMsg->AddToSlaveInstanceList(this);
}


Asura_NetMsg_SlaveInstance::~Asura_NetMsg_SlaveInstance()
{
	delete[] m_pucCurrentData;

	m_pxNetMsg->RemoveFromSlaveInstanceList(this);
}


void Asura_NetMsg_SlaveInstance::ProcessReceivedData(float fSentTime, Asura_BitUnpacker& xBitUnpacker, bool bIsFirstUpdate)
{
	// copy the current data in case we find we need to send it. This happens
	// if an entity started moving - the first update you get is the new position 
	// and the time of the last update may be up to 5 seconds old.
	u_int uDataSize = m_pxNetMsg->GetDataSize();
	u_char* pucLastData = new u_char[uDataSize];
	Asura_Mem::Copy(pucLastData, m_pucCurrentData, uDataSize);

	bool bSkippedLastUpdate;
	m_pxNetMsg->UnpackToInstance( m_pucCurrentData, bSkippedLastUpdate, xBitUnpacker );

	if (bSkippedLastUpdate && !bIsFirstUpdate)
	{
		// send the message that would have been sent for the previous server update.
		SendMessage(pucLastData, fSentTime - fASURA_NETWORK_TIME_BETWEEN_ENTITY_UPDATES);
	}

	SendMessage(m_pucCurrentData, fSentTime);
	delete[] pucLastData;
}


void Asura_NetMsg_SlaveInstance::SendMessage(const u_char* pucData, float fSentTime)
{
	// send message via entity system directly, not via network system.
	Asura_Message* pxMessage = Asura_Network::CreateMessage(m_pxNetMsg->GetMsgID(), 
		m_pxNetMsg->GetDataSize(), m_pxNetMsg->GetEntityClassification(), pucData);
	Asura_Assert( pxMessage, "Couldn't create message!" );
	if(pxMessage)
	{
		// fill in message header
		// need to set SentTime to what it was when the message was first sent.
		pxMessage->SentTime = fSentTime;
		pxMessage->SetFromGuid(m_uGuid);
		pxMessage->SetToGuid(ASURA_GUID_MY_CLIENT_ENTITIES);

		// deliver to client entity
		Asura_Entity_System::DeliverMessageToClientEntity(pxMessage);

		Asura_Network::DestroyMessage(pxMessage);
	}
}

//
// Asura_NetMsg_Data_CompressedNormal.
//

void Asura_NetMsg_Data_CompressedNormal::To16Bits( const Asura_Vector_3& xIn, u_short* pusOut )
{
    u_short usOut = 0;

    {
        float fZ = xIn.z;
        if( fZ < 0.0f )
        {
            usOut |= 1 << 4;
            fZ = -fZ;
        }
        Asura_Assert( fZ <= 1.01f, "" );
        fZ *= 16.0f;
        usOut |= Asura_Maths::FloatToInt_Near( Asura_Maths::Min(fZ, 15.0f) );
        usOut <<= 5;
    }

    {
        float fY = xIn.y;
        if( fY < 0.0f )
        {
            usOut |= 1 << 4;
            fY = -fY;
        }
        Asura_Assert( fY <= 1.01f, "" );
        fY *= 16.0f;
        usOut |= Asura_Maths::FloatToInt_Near( Asura_Maths::Min(fY, 15.0f) );
        usOut <<= 5;
    }

    {
        float fX = xIn.x;
        if( fX < 0.0f )
        {
            usOut |= 1 << 4;
            fX = -fX;
        }
        Asura_Assert( fX <= 1.01f, "" );
        fX *= 16.0f;
        usOut |= Asura_Maths::FloatToInt_Near( Asura_Maths::Min(fX, 15.0f) );
    }
    
    *pusOut = usOut;
}

void Asura_NetMsg_Data_CompressedNormal::From16Bits( u_short usIn, Asura_Vector_3* pxOut )
{
    u_int uValue = usIn;

    {
        int iX = uValue & 0xF;
        if( uValue & 0x10 ) { iX = -iX; }
        pxOut->x = iX / 16.0f;
    }

    {
        uValue >>= 5;
        int iY = uValue & 0xF;
        if( uValue & 0x10 ) { iY = -iY; }
        pxOut->y = iY / 16.0f;
    }

    {
        uValue >>= 5;
        int iZ = uValue & 0xF;
        if( uValue & 0x10 ) { iZ = -iZ; }
        pxOut->z = iZ / 16.0f;
    }
}

void Asura_NetMsg_Data_CompressedNormal::To24Bits( const Asura_Vector_3& xIn, u_char* pucOut )
{
    {
        u_int uX = 0;
        float fX = xIn.x;
        if( fX < 0.0f )
        {
            uX |= 0x80;
            fX = -fX;
        }
        Asura_Assert( fX <= 1.01f, "" );
        fX *= 128.0f;
        uX |= Asura_Maths::FloatToInt_Near( Asura_Maths::Min(fX, 127.0f) );
        *pucOut = static_cast<u_char>( uX );
    }

    {
        u_int uY = 0;
        float fY = xIn.y;
        if( fY < 0.0f )
        {
            uY |= 0x80;
            fY = -fY;
        }
        Asura_Assert( fY <= 1.01f, "" );
        fY *= 128.0f;
        uY |= Asura_Maths::FloatToInt_Near( Asura_Maths::Min(fY, 127.0f) );
        *(pucOut + 1) = static_cast<u_char>( uY );
    }

    {
        u_int uZ = 0;
        float fZ = xIn.z;
        if( fZ < 0.0f )
        {
            uZ |= 0x80;
            fZ = -fZ;
        }
        Asura_Assert( fZ <= 1.01f, "" );
        fZ *= 128.0f;
        uZ |= Asura_Maths::FloatToInt_Near( Asura_Maths::Min(fZ, 127.0f) );
        *(pucOut + 2) = static_cast<u_char>( uZ );
    }
}

void Asura_NetMsg_Data_CompressedNormal::From24Bits( const u_char* pucIn, Asura_Vector_3* pxOut )
{
    {
        int iX = *pucIn;
        int iSignX = iX & 0x80;
        iX &= 0x7F;
        if( iSignX ) { iX = -iX; }
        pxOut->x = iX / 128.0f;
    }

    
    {
        int iY = *(pucIn + 1);
        int iSignY = iY & 0x80;
        iY &= 0x7F;
        if( iSignY ) { iY = -iY; }
        pxOut->y = iY / 128.0f;
    }

    {
        int iZ = *(pucIn + 2);
        int iSignZ = iZ & 0x80;
        iZ &= 0x7F;
        if( iSignZ ) { iZ = -iZ; }
        pxOut->z = iZ / 128.0f;
    }
}


// unit tests - have to go in this file since they use some of the internals that 
// are only declared inside it.
#if defined(ASURA_SUPPORT_UNITTEST) || defined(_DEBUG)

#include "UnitTest/Asura_UnitTest.h"

class Asura_NetMsg_TestClass
{
public:

	Asura_NetMsg_TestClass()
	{
		Asura_Mem::Set(this, 0, sizeof(*this));
		m_xQuat.Identity();
	}

	static Asura_NetMsg* RegisterFields(u_short usMsgID, u_short usEntityClassification);
	static void RegisterFields(class Asura_NetMsg* pxNetMsg);
	static u_char GetNumFields()
	{
		return static_cast<u_char> (27);
	}

	void SetTestData();
	void ChangeAllFields();

	// data
	char m_cSignedChar;
	u_char m_ucUnsignedChar;
	short m_sSignedShort;
	u_short m_usUnsignedShort;
	int m_iSignedInt;
	u_int m_uUnsignedInt;
	float m_fFloat;
	Asura_Vector_2 m_xVector2;
	Asura_Vector_3 m_xVector3;
	Asura_Euler_3 m_xEuler3;
	Asura_Quat m_xQuat;
	u_int m_uInt27bits;
	u_int m_uInt0to4;
	bool m_bBoolAs1bit;

	// bitfields (first 2 taken from real BF3 code)
	union
	{
		struct
		{
			u_int	m_bIsControlledByAnyPlayer		  : 1,
					m_bDamaged						  : 1,
					m_bCloaked						  : 1,
					m_bTargeted						  : 1,
					m_bRemoteControllingVehicle		  : 1,
					m_bLocalShieldActive			  : 1,
					m_bCorpseAlphaOut				  : 1,
					m_bOnGamescene					  : 1,
					m_bShowHeroFX					  : 1,
					m_bHero							  : 1,
					m_bIsIndoors					  : 1,
					m_bClientPosAndOriTimeValid       : 1;
		};
		u_int m_uCUAllFlags;
	};
	union
	{
		struct 
		{
			u_int m_bActive:1;
			u_int m_bForceTargetFoot:1;
			u_int m_bForceTargetVehicle:1;
			u_int m_bForceTargetIonStrike:1;
			u_int m_bTargetable:1;
		};
		u_int m_uBF3PhysicalFlags;
	};
	union
	{
		struct 
		{
			u_int m_u2bitsCB1:2;
			u_int m_u7bitsCB1:7;
			u_int m_u1bitCB1A:1;
			u_int m_u13bitsCB1:13;
			u_int m_u1bitCB1B:1;
			u_int m_u8bitsCB1:8;
		};
		u_int m_uComplexBitfield1;
	};
	union
	{
		struct 
		{
			u_int m_u2bitsCB2:2;
			u_int m_u7bitsCB2:7;
			u_int m_u1bitCB2A:1;
			u_int m_u13bitsCB2:13;
			u_int m_u1bitCB2B:1;
			u_int m_u8bitsCB2:8;
		};
		u_int m_uComplexBitfield2;
	};

	Asura_UnicodeChar m_awcName32[32];
	Asura_UnicodeChar m_awcName32Pad1[32];
	Asura_UnicodeChar m_awcName32Pad2[32];
	Asura_Guid m_uGuid;
};

class Asura_NetMsg_TestClass_Approximations
{
public:

	Asura_NetMsg_TestClass_Approximations()
	{
		Asura_Mem::Set(this, 0, sizeof(*this));
		m_xCompressedQuat.Identity();
		m_xUprightQuat.Identity();
	}

	static Asura_NetMsg* RegisterFields(u_short usMsgID, u_short usEntityClassification);
	static void RegisterFields(class Asura_NetMsg* pxNetMsg);
	static u_char GetNumFields()
	{
		return static_cast<u_char> (5);
	}

	void SetTestData();
	void ChangeAllFields();

	Asura_Vector_3 m_xCompressedVector3;
	float m_fCompressedFloat4bits;
	float m_fCompressedFloat17bits;
	Asura_Quat m_xCompressedQuat;
	Asura_Quat m_xUprightQuat;
};

Asura_NetMsg* Asura_NetMsg_TestClass::RegisterFields( u_short usMsgID, u_short usEntityClassification )
{
	Asura_Assert( !Asura_NetMsg::FindNetMsgObject( usMsgID, usEntityClassification ), "already exists" );

	u_char ucNumFields = GetNumFields();

	Asura_NetMsg* pxNetMsg = new Asura_NetMsg( usMsgID, usEntityClassification, ucNumFields, sizeof(Asura_NetMsg_TestClass) );
	RegisterFields( pxNetMsg );

	pxNetMsg->EndRegistration();

	return pxNetMsg;
}

void Asura_NetMsg_TestClass::RegisterFields( class Asura_NetMsg* pxNetMsg )
{
	Asura_NetMsg_TestClass xDummy;

	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_char, m_cSignedChar, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_uchar, m_ucUnsignedChar, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_short, m_sSignedShort, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_ushort, m_usUnsignedShort, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_int, m_iSignedInt, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_uint, m_uUnsignedInt, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_float, m_fFloat, 1, Asura_Threshold_AnyChange, 0.0f );

	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_vector2, m_xVector2, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_vector3, m_xVector3, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_euler3, m_xEuler3, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_quat, m_xQuat, 1, Asura_Threshold_AnyChange, 0.5f );

	Asura_NetMsg_RegisterIntFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_bits, 27, m_uInt27bits, 1, Asura_Threshold_AnyChange ); 
	Asura_NetMsg_RegisterIntFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_bits, 3, m_uInt0to4, 1, Asura_Threshold_AnyChange ); // 0-4
	Asura_NetMsg_RegisterIntFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_bits, 1, m_bBoolAs1bit, 1, Asura_Threshold_AnyChange ); 

	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 12, 0, m_uCUAllFlags, Asura_Threshold_AnyChange, ASURA_NETMSG_SINGLEBITS ); 
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 5, 0, m_uBF3PhysicalFlags, Asura_Threshold_AnyChange, ASURA_NETMSG_SINGLEBITS );
	char acComplexBitfield1Desc[6] = {2,7,1,13,1,8};
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 32, 0, m_uComplexBitfield1, Asura_Threshold_AnyChange, acComplexBitfield1Desc );

	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 2, 0, m_uComplexBitfield2, Asura_Threshold_AnyChange, &acComplexBitfield1Desc[0] );
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 7, 2, m_uComplexBitfield2, Asura_Threshold_AnyChange, &acComplexBitfield1Desc[1] );
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 1, 9, m_uComplexBitfield2, Asura_Threshold_AnyChange, &acComplexBitfield1Desc[2] );
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 13, 10, m_uComplexBitfield2, Asura_Threshold_AnyChange, &acComplexBitfield1Desc[3] );
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 1, 23, m_uComplexBitfield2, Asura_Threshold_AnyChange, &acComplexBitfield1Desc[4] );
	Asura_NetMsg_RegisterBitfieldFieldBits( pxNetMsg, xDummy, Asura_FieldType_uint_Bitfield, 8, 24, m_uComplexBitfield2, Asura_Threshold_AnyChange, &acComplexBitfield1Desc[5] );

	Asura_NetMsg_RegisterUnicodeStringField( pxNetMsg, xDummy, Asura_FieldType_UnicodeString, m_awcName32, 32, Asura_Threshold_AnyChange );
	Asura_NetMsg_RegisterUnicodeStringField( pxNetMsg, xDummy, Asura_FieldType_UnicodeString, m_awcName32Pad1, 32, Asura_Threshold_AnyChange );
	Asura_NetMsg_RegisterUnicodeStringField( pxNetMsg, xDummy, Asura_FieldType_UnicodeString, m_awcName32Pad2, 32, Asura_Threshold_AnyChange );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_Guid, m_uGuid, 1, Asura_Threshold_AnyChange, 0.0f );
}

Asura_NetMsg* Asura_NetMsg_TestClass_Approximations::RegisterFields( u_short usMsgID, u_short usEntityClassification )
{
	Asura_Assert( !Asura_NetMsg::FindNetMsgObject( usMsgID, usEntityClassification ), "already exists" );

	u_char ucNumFields = GetNumFields();

	Asura_NetMsg* pxNetMsg = new Asura_NetMsg( usMsgID, usEntityClassification, ucNumFields, sizeof(Asura_NetMsg_TestClass_Approximations) );
	RegisterFields( pxNetMsg );

	pxNetMsg->EndRegistration();

	return pxNetMsg;
}

void Asura_NetMsg_TestClass_Approximations::RegisterFields( class Asura_NetMsg* pxNetMsg )
{
	Asura_NetMsg_TestClass_Approximations xDummy;

	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_CompressedPosition, m_xCompressedVector3, 1, Asura_Threshold_AnyChange, 0.0f );
	Asura_NetMsg_RegisterFloatFieldBits( pxNetMsg, xDummy, Asura_FieldType_CompressedFloat, 4, 0.0f, 256.0f, m_fCompressedFloat4bits, 1, Asura_Threshold_AnyChange, 0.0f ); 	
	Asura_NetMsg_RegisterFloatFieldBits( pxNetMsg, xDummy, Asura_FieldType_CompressedFloat, 17, 0.0f, 256.0f, m_fCompressedFloat17bits, 1, Asura_Threshold_AnyChange, 0.0f ); 	
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_CompressedQuat, m_xCompressedQuat, 1, Asura_Threshold_AnyChange, 0.5f );
	Asura_NetMsg_RegisterField( pxNetMsg, xDummy, Asura_FieldType_UprightQuat, m_xUprightQuat, 1, Asura_Threshold_AnyChange, 0.5f );
}

static Asura_RandomGenerator g_xNetMsgTestRandomiser(13031974); // arbitrary seed

static bool GetRandomBool()
{
	return g_xNetMsgTestRandomiser.GetIntU(0, 1) == 1;
}
static float GetRandomFloat()
{
	// floating point differences between PC and PS3 mean we can't use GetFloat() apparently.
	int iUnits = g_xNetMsgTestRandomiser.GetInt(-9999, 9999);
	int iDecimals = g_xNetMsgTestRandomiser.GetInt(0, 999);
	float fValue = static_cast<float>(iUnits) + static_cast<float>(iDecimals) / 1000.0f;
	return fValue;
}
static float GetRandomAngle()
{
	//return g_xNetMsgTestRandomiser.GetFloat(-Asura_Maths::Pi, Asura_Maths::Pi);
	// floating point differences between PC and PS3 mean we can't use GetFloat() apparently.
	int iUnits = g_xNetMsgTestRandomiser.GetInt(-3, 3);
	int iDecimals = g_xNetMsgTestRandomiser.GetInt(0, 1415);
	float fValue = static_cast<float>(iUnits) + static_cast<float>(iDecimals) / 10000.0f;
	return fValue;
}
static float GetRandomFloatZeroToOne()
{
	int iDecimals = g_xNetMsgTestRandomiser.GetInt(0, 9999);
	float fValue = static_cast<float>(iDecimals) / 10000.0f;
	return fValue;
}
static float FixupFloatForCrossPlatformComparison(float fInput)
{
	// use 3 decimal places
	fInput *= 1000.f;
	int iInput = Asura_Maths::FloatToInt_Chop(fInput);
	fInput = static_cast<float>(iInput);
	fInput /= 1000.f;
	return fInput;
}
static u_int GetRandomUnsignedWithBits(u_int uNumBits)
{
	return g_xNetMsgTestRandomiser.GetIntU(0, (1 << uNumBits) - 1);
}
void Asura_NetMsg_TestClass::SetTestData()
{
	m_cSignedChar = static_cast<char>(g_xNetMsgTestRandomiser.GetInt(CHAR_MIN, CHAR_MAX));
	m_ucUnsignedChar = static_cast<u_char>(g_xNetMsgTestRandomiser.GetIntU(0, UCHAR_MAX));
	m_sSignedShort = static_cast<short>(g_xNetMsgTestRandomiser.GetInt(SHRT_MIN, SHRT_MAX));
	m_usUnsignedShort = static_cast<u_short>(g_xNetMsgTestRandomiser.GetIntU(0, USHRT_MAX));
	m_iSignedInt = g_xNetMsgTestRandomiser.GetInt(INT_MIN + 1, INT_MAX); // +1 is a workaround for lack of range in ranrot
	m_uUnsignedInt = g_xNetMsgTestRandomiser.GetIntU(0, INT_MAX) + g_xNetMsgTestRandomiser.GetIntU(1, INT_MAX); // use it twice due to lack of range in ranrot
	m_fFloat = GetRandomFloat();

	m_xVector2.x = GetRandomFloat();
	m_xVector2.y = GetRandomFloat();

	m_xVector3.x = GetRandomFloat();
	m_xVector3.y = GetRandomFloat();
	m_xVector3.z = GetRandomFloat();

	m_xEuler3.x = GetRandomAngle();
	m_xEuler3.y = GetRandomAngle();
	m_xEuler3.z = GetRandomAngle();

	m_xQuat.FromEuler(m_xEuler3);
	m_xQuat.x = FixupFloatForCrossPlatformComparison(m_xQuat.x);
	m_xQuat.y = FixupFloatForCrossPlatformComparison(m_xQuat.y);
	m_xQuat.z = FixupFloatForCrossPlatformComparison(m_xQuat.z);
	m_xQuat.w = FixupFloatForCrossPlatformComparison(m_xQuat.w);

	m_uInt27bits = GetRandomUnsignedWithBits(27);
	m_uInt0to4 = g_xNetMsgTestRandomiser.GetIntU(0, 4);
	m_bBoolAs1bit = GetRandomBool();

	m_uCUAllFlags = 0;
	m_bIsControlledByAnyPlayer		  = GetRandomBool();
	m_bDamaged						  = GetRandomBool();
	m_bCloaked						  = GetRandomBool();
	m_bTargeted						  = GetRandomBool();
	m_bRemoteControllingVehicle		  = GetRandomBool();
	m_bLocalShieldActive			  = GetRandomBool();
	m_bCorpseAlphaOut				  = GetRandomBool();
	m_bOnGamescene					  = GetRandomBool();
	m_bShowHeroFX					  = GetRandomBool();
	m_bHero							  = GetRandomBool();
	m_bIsIndoors					  = GetRandomBool();
	m_bClientPosAndOriTimeValid       = GetRandomBool();

	m_uBF3PhysicalFlags = 0;
	m_bActive							= GetRandomBool();
	m_bForceTargetFoot					= GetRandomBool();
	m_bForceTargetVehicle				= GetRandomBool();
	m_bForceTargetIonStrike				= GetRandomBool();
	m_bTargetable						= GetRandomBool();

	m_uComplexBitfield1 = 0;
	m_u2bitsCB1 = GetRandomUnsignedWithBits(2);
	m_u7bitsCB1 = GetRandomUnsignedWithBits(7);
	m_u1bitCB1A = GetRandomUnsignedWithBits(1);
	m_u13bitsCB1 = GetRandomUnsignedWithBits(13);
	m_u1bitCB1B = GetRandomUnsignedWithBits(1);
	m_u8bitsCB1 = GetRandomUnsignedWithBits(8);

	m_uComplexBitfield2 = 0;
	m_u2bitsCB2 = GetRandomUnsignedWithBits(2);
	m_u7bitsCB2 = GetRandomUnsignedWithBits(7);
	m_u1bitCB2A = GetRandomUnsignedWithBits(1);
	m_u13bitsCB2 = GetRandomUnsignedWithBits(13);
	m_u1bitCB2B = GetRandomUnsignedWithBits(1);
	m_u8bitsCB2 = GetRandomUnsignedWithBits(8);

	// random name
	u_int uNameLength = g_xNetMsgTestRandomiser.GetIntU(0, 31);
	for (u_int u = 0; u < uNameLength; ++u)
	{
		m_awcName32[u] = static_cast<Asura_UnicodeChar>(L'A' + g_xNetMsgTestRandomiser.GetIntU(0, 25));
		m_awcName32Pad1[u] = static_cast<Asura_UnicodeChar>(L'A' + g_xNetMsgTestRandomiser.GetIntU(0, 25));
		m_awcName32Pad2[u] = static_cast<Asura_UnicodeChar>(L'A' + g_xNetMsgTestRandomiser.GetIntU(0, 25));
	}
	m_awcName32[uNameLength] = 0;
	m_awcName32Pad1[uNameLength] = 0;
	m_awcName32Pad2[uNameLength] = 0;

	m_uGuid = g_xNetMsgTestRandomiser.GetIntU(0, 200000);
}

void Asura_NetMsg_TestClass::ChangeAllFields()
{
	Asura_NetMsg_TestClass xOther = *this;

	// regex: find ":b{m.*}{:b=.*;}"
	// replace with "do { \1\2 } while (\1 == xOther.\1);"
	do { m_cSignedChar = static_cast<char>(g_xNetMsgTestRandomiser.GetInt(CHAR_MIN, CHAR_MAX)); } while (m_cSignedChar == xOther.m_cSignedChar);
	do { m_ucUnsignedChar = static_cast<u_char>(g_xNetMsgTestRandomiser.GetIntU(0, UCHAR_MAX)); } while (m_ucUnsignedChar == xOther.m_ucUnsignedChar);
	do { m_sSignedShort = static_cast<short>(g_xNetMsgTestRandomiser.GetInt(SHRT_MIN, SHRT_MAX)); } while (m_sSignedShort == xOther.m_sSignedShort);
	do { m_usUnsignedShort = static_cast<u_short>(g_xNetMsgTestRandomiser.GetIntU(0, USHRT_MAX)); } while (m_usUnsignedShort == xOther.m_usUnsignedShort);
	do { m_iSignedInt = g_xNetMsgTestRandomiser.GetInt(INT_MIN + 1, INT_MAX); } while (m_iSignedInt == xOther.m_iSignedInt); // +1 is a workaround for lack of range in ranrot
	do { m_uUnsignedInt = g_xNetMsgTestRandomiser.GetIntU(0, INT_MAX) + g_xNetMsgTestRandomiser.GetIntU(1, INT_MAX); } while (m_uUnsignedInt == xOther.m_uUnsignedInt); // use it twice due to lack of range in ranrot
	do { m_fFloat = GetRandomAngle(); } while (m_fFloat == xOther.m_fFloat);

	do { m_xVector2.x = GetRandomFloat(); } while (m_xVector2.x == xOther.m_xVector2.x);
	do { m_xVector2.y = GetRandomFloat(); } while (m_xVector2.y == xOther.m_xVector2.y);

	do { m_xVector3.x = GetRandomFloat(); } while (m_xVector3.x == xOther.m_xVector3.x);
	do { m_xVector3.y = GetRandomFloat(); } while (m_xVector3.y == xOther.m_xVector3.y);
	do { m_xVector3.z = GetRandomFloat(); } while (m_xVector3.z == xOther.m_xVector3.z);

	do { m_xEuler3.x = GetRandomAngle(); } while (m_xEuler3.x == xOther.m_xEuler3.x);
	do { m_xEuler3.y = GetRandomAngle(); } while (m_xEuler3.y == xOther.m_xEuler3.y);
	do { m_xEuler3.z = GetRandomAngle(); } while (m_xEuler3.z == xOther.m_xEuler3.z);
	m_xQuat.FromEuler(m_xEuler3);

	do { m_uInt27bits = GetRandomUnsignedWithBits(27); } while (m_uInt27bits == xOther.m_uInt27bits);
	do { m_uInt0to4 = g_xNetMsgTestRandomiser.GetIntU(0, 4); } while (m_uInt0to4 == xOther.m_uInt0to4);
	do { m_bBoolAs1bit = GetRandomBool(); } while (m_bBoolAs1bit == xOther.m_bBoolAs1bit);

	m_bIsControlledByAnyPlayer		  = !m_bIsControlledByAnyPlayer		 ;
	m_bDamaged						  = !m_bDamaged						 ;
	m_bCloaked						  = !m_bCloaked						 ;
	m_bTargeted						  = !m_bTargeted						 ;
	m_bRemoteControllingVehicle		  = !m_bRemoteControllingVehicle		 ;
	m_bLocalShieldActive			  = !m_bLocalShieldActive			 ;
	m_bCorpseAlphaOut				  = !m_bCorpseAlphaOut				 ;
	m_bOnGamescene					  = !m_bOnGamescene					 ;
	m_bShowHeroFX					  = !m_bShowHeroFX					 ;
	m_bHero							  = !m_bHero							 ;
	m_bIsIndoors					  = !m_bIsIndoors					 ;
	m_bClientPosAndOriTimeValid       = !m_bClientPosAndOriTimeValid      ;

	m_bActive						 = !m_bActive						;
	m_bForceTargetFoot				 = !m_bForceTargetFoot				;
	m_bForceTargetVehicle			 = !m_bForceTargetVehicle			;
	m_bForceTargetIonStrike			 = !m_bForceTargetIonStrike			;
	m_bTargetable					 = !m_bTargetable					;

	do { m_u2bitsCB1 = GetRandomUnsignedWithBits(2); } while (m_u2bitsCB1 == xOther.m_u2bitsCB1);
	do { m_u7bitsCB1 = GetRandomUnsignedWithBits(7); } while (m_u7bitsCB1 == xOther.m_u7bitsCB1);
	do { m_u1bitCB1A = GetRandomUnsignedWithBits(1); } while (m_u1bitCB1A == xOther.m_u1bitCB1A);
	do { m_u13bitsCB1 = GetRandomUnsignedWithBits(13); } while (m_u13bitsCB1 == xOther.m_u13bitsCB1);
	do { m_u1bitCB1B = GetRandomUnsignedWithBits(1); } while (m_u1bitCB1B == xOther.m_u1bitCB1B);
	do { m_u8bitsCB1 = GetRandomUnsignedWithBits(8); } while (m_u8bitsCB1 == xOther.m_u8bitsCB1);

	do { m_u2bitsCB2 = GetRandomUnsignedWithBits(2); } while (m_u2bitsCB2 == xOther.m_u2bitsCB2);
	do { m_u7bitsCB2 = GetRandomUnsignedWithBits(7); } while (m_u7bitsCB2 == xOther.m_u7bitsCB2);
	do { m_u1bitCB2A = GetRandomUnsignedWithBits(1); } while (m_u1bitCB2A == xOther.m_u1bitCB2A);
	do { m_u13bitsCB2 = GetRandomUnsignedWithBits(13); } while (m_u13bitsCB2 == xOther.m_u13bitsCB2);
	do { m_u1bitCB2B = GetRandomUnsignedWithBits(1); } while (m_u1bitCB2B == xOther.m_u1bitCB2B);
	do { m_u8bitsCB2 = GetRandomUnsignedWithBits(8); } while (m_u8bitsCB2 == xOther.m_u8bitsCB2);

	// random name
	if (m_awcName32[0])
	{
		m_awcName32[0] = 0;
	}
	else
	{
		m_awcName32[0] = L'A';
		m_awcName32[1] = 0;
	}
	if (m_awcName32Pad1[0])
	{
		m_awcName32Pad1[0] = 0;
	}
	else
	{
		m_awcName32Pad1[0] = L'A';
		m_awcName32Pad1[1] = 0;
	}
	if (m_awcName32Pad2[0])
	{
		m_awcName32Pad2[0] = 0;
	}
	else
	{
		m_awcName32Pad2[0] = L'A';
		m_awcName32Pad2[1] = 0;
	}

	do { m_uGuid = g_xNetMsgTestRandomiser.GetIntU(0, 200000); } while (m_uGuid == xOther.m_uGuid);
}

// approximations
void Asura_NetMsg_TestClass_Approximations::SetTestData()
{
	m_xCompressedVector3.x = GetRandomAngle();
	m_xCompressedVector3.y = GetRandomAngle();
	m_xCompressedVector3.z = GetRandomAngle();

	m_fCompressedFloat4bits = GetRandomFloatZeroToOne();
	m_fCompressedFloat17bits = GetRandomAngle();

	Asura_Euler_3 xTemp(m_xCompressedVector3.x, m_xCompressedVector3.y, m_xCompressedVector3.z);
	m_xCompressedQuat.FromEuler(xTemp);

	Asura_Euler_3 xUpright(0.0f, m_fCompressedFloat17bits, 0.0f);
	m_xUprightQuat.FromEuler(xUpright);
}

void Asura_NetMsg_TestClass_Approximations::ChangeAllFields()
{
	Asura_NetMsg_TestClass_Approximations xOther = *this;

	do { m_xCompressedVector3.x = GetRandomAngle(); } while (m_xCompressedVector3.x == m_xCompressedVector3.x);
	do { m_xCompressedVector3.y = GetRandomAngle(); } while (m_xCompressedVector3.y == m_xCompressedVector3.y);
	do { m_xCompressedVector3.z = GetRandomAngle(); } while (m_xCompressedVector3.z == m_xCompressedVector3.z);

	do { m_fCompressedFloat4bits = GetRandomFloatZeroToOne(); } while (m_fCompressedFloat4bits == xOther.m_fCompressedFloat4bits);
	do { m_fCompressedFloat17bits = GetRandomAngle(); } while (m_fCompressedFloat17bits == xOther.m_fCompressedFloat17bits);

	Asura_Euler_3 xTemp(m_xCompressedVector3.x, m_xCompressedVector3.y, m_xCompressedVector3.z);
	m_xCompressedQuat.FromEuler(xTemp);

	Asura_Euler_3 xUpright(0.0f, m_fCompressedFloat17bits, 0.0f);
	m_xUprightQuat.FromEuler(xUpright);
}
#endif // #if defined(ASURA_SUPPORT_UNITTEST) || defined(_DEBUG)

#if defined(ASURA_SUPPORT_UNITTEST)
ASURA_UNITTEST(NetMsg, ExactTests)
{
	Asura_NetMsg_UpdateData xUpdateData;
	Asura_NetMsg_UpdateData* apxUpdateData[1] = {&xUpdateData};
	Asura_NetMsg_Recipients xRecipients(apxUpdateData, 1);

	Asura_NetMsg* pxNetMsg = Asura_NetMsg_TestClass::RegisterFields(0, 2);// msgID = 0, classification = 2

	Asura_NetMsg_TestClass xData;
	Asura_NetMsg_TestClass xCompareData;

	for (u_int uNumTests = 0; uNumTests < 1000; ++uNumTests)
	{
		// pack into our local update data
		xUpdateData.ResetBitPacker();
		xData.SetTestData();
		pxNetMsg->PackWholeMessage(&xData, xRecipients);

		// unpack from local unpacker
		u_int uDataSize = 0;
		const void* pData = 0;
		xUpdateData.GetUnpackData(uDataSize, pData);
		Asura_BitUnpacker xUnpacker(pData, uDataSize);
		pxNetMsg->UnpackWholeMessage( reinterpret_cast<u_char*>(&xCompareData), xUnpacker );

		// check the data matches.
		ASURA_UNITTEST_CHECK(!Asura_Mem::Compare(&xData, &xCompareData, sizeof(xData)));

		// check that HasChanged() works correctly. The data is the same so HasChanged() should return 
		// false for all fields.
		for (u_int uField = 0; uField < pxNetMsg->GetNumFields(); ++uField)
		{
			const Asura_NetMsg_Field& xField = pxNetMsg->GetField(uField);
			const u_char* pucCurrentValue = reinterpret_cast<const u_char*>(&xData) + xField.m_usByteOffset;
			u_char* pucCompareValue = reinterpret_cast<u_char*>(&xCompareData) + xField.m_usByteOffset;
			ASURA_UNITTEST_CHECK(!xField.HasChanged(pucCurrentValue, pucCompareValue, false));
		}

		// change every field
		xData.ChangeAllFields();

		// every field is now different, so HasChanged() should return true for all fields.
		for (u_int uField = 0; uField < pxNetMsg->GetNumFields(); ++uField)
		{
			const Asura_NetMsg_Field& xField = pxNetMsg->GetField(uField);
			const u_char* pucCurrentValue = reinterpret_cast<const u_char*>(&xData) + xField.m_usByteOffset;
			u_char* pucCompareValue = reinterpret_cast<u_char*>(&xCompareData) + xField.m_usByteOffset;
			ASURA_UNITTEST_CHECK(xField.HasChanged(pucCurrentValue, pucCompareValue, false));
			// HasChanged should set the value to be the current value if we pass in true.
			if (xField.HasChanged(pucCurrentValue, pucCompareValue, true))
			{
				ASURA_UNITTEST_CHECK(!xField.HasChanged(pucCurrentValue, pucCompareValue, false));
			}
		}
	}

	delete pxNetMsg;
}

//ASURA_UNITTEST(NetMsg, Approximations)
//{
//	Asura_NetMsg_UpdateData xUpdateData;
//	Asura_NetMsg_Recipients xRecipients(&xUpdateData, 1);
//
//	Asura_NetMsg* pxNetMsg = Asura_NetMsg_TestClass_Approximations::RegisterFields(0, 2);// msgID = 0, classification = 2
//
//	Asura_NetMsg_TestClass_Approximations xData;
//	Asura_NetMsg_TestClass_Approximations xCompareData;
//
//	for (u_int uNumTests = 0; uNumTests < 1000; ++uNumTests)
//	{
//		// pack into our local update data
//		xUpdateData.ResetBitPacker();
//		xData.SetTestData();
//		pxNetMsg->PackWholeMessage(&xData, xRecipients);
//
//		// unpack from local unpacker
//		u_int uDataSize = 0;
//		const u_char* pucData = 0;
//		xUpdateData.GetUnpackData(uDataSize, pucData);
//		Asura_BitUnpacker xUnpacker(pucData, uDataSize);
//		pxNetMsg->UnpackWholeMessage( reinterpret_cast<u_char*>(&xCompareData), xUnpacker );
//
//		// check that HasChanged() works correctly. The data is the same so HasChanged() should return 
//		// false for all fields.
//		for (u_int uField = 0; uField < pxNetMsg->GetNumFields(); ++uField)
//		{
//			const Asura_NetMsg_Field& xField = pxNetMsg->GetField(uField);
//			const u_char* pucCurrentValue = reinterpret_cast<const u_char*>(&xData) + xField.m_usByteOffset;
//			u_char* pucCompareValue = reinterpret_cast<u_char*>(&xCompareData) + xField.m_usByteOffset;
//			ASURA_UNITTEST_CHECK(!xField.HasChanged(pucCurrentValue, pucCompareValue));
//		}
//
//		// change every field
//		xData.ChangeAllFields();
//
//		// every field is now different, so HasChanged() should return true for all fields.
//		for (u_int uField = 0; uField < pxNetMsg->GetNumFields(); ++uField)
//		{
//			const Asura_NetMsg_Field& xField = pxNetMsg->GetField(uField);
//			const u_char* pucCurrentValue = reinterpret_cast<const u_char*>(&xData) + xField.m_usByteOffset;
//			u_char* pucCompareValue = reinterpret_cast<u_char*>(&xCompareData) + xField.m_usByteOffset;
//			ASURA_UNITTEST_CHECK(xField.HasChanged(pucCurrentValue, pucCompareValue, false));
//			// HasChanged should set the value to be the current value if we pass in true.
//			if (xField.HasChanged(pucCurrentValue, pucCompareValue, true))
//			{
//				ASURA_UNITTEST_CHECK(!xField.HasChanged(pucCurrentValue, pucCompareValue, false));
//			}
//		}
//	}
//
//	delete pxNetMsg;
//}
#endif // #if defined(ASURA_SUPPORT_UNITTEST)

#ifdef _DEBUG
#include "Network/Asura_NetTransport_System.h"

void Asura_NetMsg_System::SendTestData()
{
	Asura_NetMsg_UpdateData xUpdateData;
	Asura_NetMsg_UpdateData* apxUpdateData[1] = {&xUpdateData};
	Asura_NetMsg_Recipients xRecipients(apxUpdateData, 1);

	Asura_NetMsg* pxNetMsg = Asura_NetMsg_TestClass::RegisterFields(0, 2);// msgID = 0, classification = 2

	Asura_NetMsg_TestClass xData;
	Asura_NetMsg_TestClass xCompareData;

	// save the state of the random generator.
	int iLow;
	int iHigh;
	g_xNetMsgTestRandomiser.GetCurrentState(iLow, iHigh);
	// pack the state into the data
	xUpdateData.AddBits(*reinterpret_cast<const u_int*>(&iLow), 32);
	xUpdateData.AddBits(*reinterpret_cast<const u_int*>(&iHigh), 32);

	// set the data in the message using that random state
	xData.SetTestData();
	// pack the data
	pxNetMsg->PackWholeMessage(&xData, xRecipients);

	// now send the lot
	u_int uSendSize = 0;
	const void* pSendData = 0;
	xUpdateData.GetUnpackData(uSendSize, pSendData);

	Asura_NetTransport_System::BroadcastTestNetMsg(pSendData, uSendSize);

	// finally, test that using the data we sent will result in a good comparison
	g_xNetMsgTestRandomiser.SetCurrentState(iLow, iHigh);
	xCompareData.SetTestData();
	Asura_Assert(!Asura_Mem::Compare(&xData, &xCompareData, sizeof(xData)), "netmsg data sent over network will not match");

	delete pxNetMsg;
}

void Asura_NetMsg_System::ProcessTestData( const void* pData, u_int uDataSize )
{
	Asura_NetMsg* pxNetMsg = Asura_NetMsg_TestClass::RegisterFields(0, 2);// msgID = 0, classification = 2

	Asura_NetMsg_TestClass xData;
	Asura_NetMsg_TestClass xCompareData;

	Asura_BitUnpacker xUnpacker(pData, uDataSize);

	// unpack the 2 u_ints for the RNG
	u_int uLow = xUnpacker.Unpack(32);
	u_int uHigh = xUnpacker.Unpack(32);

	pxNetMsg->UnpackWholeMessage( reinterpret_cast<u_char*>(&xData), xUnpacker );

	// restore the state of the random generator.
	int iLow = *reinterpret_cast<const u_int*>(&uLow);
	int iHigh= *reinterpret_cast<const u_int*>(&uHigh);
	g_xNetMsgTestRandomiser.SetCurrentState(iLow, iHigh);
	xCompareData.SetTestData();
	
	Asura_Assert(!Asura_Mem::Compare(&xData, &xCompareData, sizeof(xData)), "netmsg data sent over network doesn't match");

	delete pxNetMsg;

}
#endif

#endif//#ifdef ASURA_SUPPORT_NETMSG_SYSTEM
