//////////////////////////////////////////////////////////////////////////////
//
// Asura_Network_Session.cpp
//
// Copyright (c) 2002-2004 Rebellion. All rights reserved.
//
// Description:
//
//     LAN session management - for clients.
//
//////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_NETWORK_SESSION_H__
#define __ASURA_NETWORK_SESSION_H__

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// Asura headers
#include "Core/Asura_String.h"
#include "Core/Memory/Asura_Mem.h"
#include "Network/Asura_Platform_NetSocket.h"

static const u_int uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS = 16;
static const u_int uASURA_NETWORK_MAX_SESSION_NAME = 31;

class Asura_Network_Session
{
	public:

		Asura_Network_Session()
		{
			m_uCurrentNumberOfPlayers = 0;
			m_uMaxNumberOfPlayers = 0;
			m_wszSessionName[0] = 0;
			m_uRoundTripLatency = 999;
			m_uServerSessionID = 0;
			m_bValid = false;
			m_pPlatformSessionData = NULL;
			m_uPlatformSessionDataSize = 0;
			m_pProjectSessionData = NULL;
			m_uProjectSessionDataSize = 0;
		}

		~Asura_Network_Session()
		{
			delete[] m_pPlatformSessionData;
			delete[] m_pProjectSessionData;
		}

	private:
		
		// private and unimplemented copy constructor and assignment operator.
		// These prevent the default versions being implicitly called and possibly
		// ending up with stale data.
		Asura_Network_Session(const Asura_Network_Session& xFrom);
		Asura_Network_Session& operator=(const Asura_Network_Session& xRHS);
		
	public:

		inline bool IsValid(void) const
		{
			return m_bValid;
		}
		inline Asura_NetAddress GetAddress( void ) const { return m_xServerAddress; }
		inline u_short GetPort( void ) const { return m_usServerPort; }
		inline u_int GetCurrentNumberOfPlayers( void ) const { return m_uCurrentNumberOfPlayers; }
		inline u_int GetMaxNumberOfPlayers( void ) const { return m_uMaxNumberOfPlayers; }
		inline const Asura_UnicodeChar* GetSessionName( void ) const { return m_wszSessionName; }
		inline u_int GetPing( void ) const { return m_uRoundTripLatency; }
		inline u_int GetServerSessionID( void ) const { return m_uServerSessionID; }
		inline const void* GetPlatformSessionData( void ) const { return m_pPlatformSessionData; }
		inline u_int GetPlatformSessionDataSize( void ) const { return m_uPlatformSessionDataSize; }
		inline const void* GetProjectSessionData( void ) const { return m_pProjectSessionData; }
		inline u_int GetProjectSessionDataSize( void ) const { return m_uProjectSessionDataSize; }

		void Reset(void)
		{
			m_bValid = false;
		}

		void Set( Asura_NetAddress xAddress,
				  u_short usPort,
				  u_int uCurPlayers,
                  u_int uMaxPlayers,
                  const Asura_UnicodeChar* wszSessionName,
                  u_int uRoundTripLatency,
				  u_int uServerSessionID,
				  const void* pvPlatformSessionData,
				  u_int uPlatformSessionDataSize,
                  const void* pvProjectSessionData,
                  u_int uProjectSessionDataSize )
		{
			m_xServerAddress			=	xAddress;
			m_usServerPort				=	usPort;
			m_uCurrentNumberOfPlayers	=	uCurPlayers;
			m_uMaxNumberOfPlayers		=	uMaxPlayers;
			m_uRoundTripLatency			=	uRoundTripLatency;
			m_uServerSessionID			=   uServerSessionID;
			m_bValid					=	true;

			Asura_PrintfToBfr( m_wszSessionName, L"%ls", uASURA_NETWORK_MAX_SESSION_NAME, true, wszSessionName );
			m_wszSessionName[uASURA_NETWORK_MAX_SESSION_NAME] = 0;			

			if((m_pPlatformSessionData)&&(m_uPlatformSessionDataSize != uPlatformSessionDataSize))
			{
				delete[] m_pPlatformSessionData; m_pPlatformSessionData = 0;
			}

			if(!m_pPlatformSessionData)
			{
				m_pPlatformSessionData = new char[uPlatformSessionDataSize];
			}
			Asura_Mem::Copy(m_pPlatformSessionData, pvPlatformSessionData, uPlatformSessionDataSize);
			m_uPlatformSessionDataSize = uPlatformSessionDataSize;

			SetProjectSessionData(pvProjectSessionData, uProjectSessionDataSize);
		}

		void SetProjectSessionData( const void* pvProjectSessionData, u_int uProjectSessionDataSize )
		{
			// (MH - 10.11.2003) changed this to do a deep copy.
			if (m_pProjectSessionData)
			{
				if (m_uProjectSessionDataSize != uProjectSessionDataSize)
				{
					delete[] m_pProjectSessionData;
					m_pProjectSessionData =	new char [uProjectSessionDataSize];
				}
			}
			else
			{
				m_pProjectSessionData =	new char [uProjectSessionDataSize];
			}
			Asura_Mem::Copy(m_pProjectSessionData, pvProjectSessionData, uProjectSessionDataSize);
			m_uProjectSessionDataSize = uProjectSessionDataSize;
		}

		// PrepareConnection() - function to do any stuff needed before server can be connected to.
		// using this for GameSpy natneg and xbox 360 secure connection
		void PrepareConnection() const; 

		void Platform_PrepareConnection() const; 

	private:

		// AS 20/12/2007 This needs to be mutable for xbox360
		// as its only setup from the Platform_PrepareConnection call
		// if this causes problems then the PrepareConnection calls
		// and Asura_Network_Session_System::GetSession will have 
		// to become non const.
		mutable Asura_NetAddress	m_xServerAddress;
		u_short				m_usServerPort;
		u_int				m_uCurrentNumberOfPlayers;
		u_int				m_uMaxNumberOfPlayers;
		Asura_UnicodeChar	m_wszSessionName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
		u_int				m_uRoundTripLatency;
		u_int				m_uServerSessionID;
		bool				m_bValid;

		char*				m_pPlatformSessionData;
		u_int				m_uPlatformSessionDataSize;

		char*				m_pProjectSessionData;
		u_int				m_uProjectSessionDataSize;
};

class Asura_Network_Session_System
{
	
public:
	
	static void Initialise( bool bLan = true );
	static void Update( void );
	static void Shutdown( void );
	static bool IsInitialised() { return s_pxSessions != NULL; }
	
	// code called by menus
	static void ForceRefresh( void );
	static u_int GetNumberOfSessions( void ) { return s_uNumberOfSessions; }
	static const Asura_Network_Session* GetSession(u_int uSessionIndex);

	// code called by network code for LAN sessions
	static void SetNumberOfSessions( u_int uNumSessions ) { s_uNumberOfSessions = uNumSessions; }
	static void SetSession(	u_int uSessionIndex,
							Asura_NetAddress xAddress,
							u_short usPort,
							u_int uCurPlayers,
							u_int uMaxPlayers,
							const Asura_UnicodeChar* wszSessionName,
							u_int uRoundTripLatency,
							u_int uServerSessionID,
							const void* pvPlatformSessionData,
							u_int uPlatformSessionDataSize,
							const void* pvProjectSessionData,
							u_int uProjectSessionDataSize );
	static void ClearSession( u_int uSessionIndex );

private:
	
	static void Platform_Initialise( void );
	static void Platform_Update( void );
	static void Platform_Shutdown( void );

#ifdef ASURA_SUPPORT_GAMESPY
	static u_int Project_GetSessionDataSize();
	static void Project_FillSessionDataFromGameSpyServer( u_int uServerIndex, void* pData );
#endif

	// private data
	static Asura_Network_Session* s_pxSessions;
	static u_int s_uNumberOfSessions;
	static bool s_bLan;
};

#endif //#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK


#endif//__ASURA_NETWORK_SESSION_H__
