/**
*
* \file Asura_Network_StatLogging.h
*
* Created by Richard May - 20/11/2003 11:59:20
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/


#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_NETWORK_STATLOGGING_H__
#define __ASURA_NETWORK_STATLOGGING_H__



/************/
/* Includes.*/
/************/

#include <string.h>

// asura
#include "Core/Asura_HashTree.h"
#include "Network/Asura_Guid.h"

#ifdef ASURA_USE_NETWORK_STATLOGGING

/*********/
/* Enums */
/*********/

enum ASURA_STATLOG_SNAPSHOT_MODE
{
	ASURA_STATLOG_SNAPSHOTMODE_ONEND = 0,  // snapshot sent on game end
	ASURA_STATLOG_SNAPSHOTMODE_ONINTERVAL, // snapshot sent on every n seconds
	ASURA_STATLOG_SNAPSHOTMODE_ONTRIGGER,  // snapshot triggered manually
};

enum STAT_TYPE
{
	STAT_NONE = 0,
		
	// basic types
	STAT_INT,
	STAT_UINT,
	STAT_BOOL,
	STAT_FLOAT,
	
	// pointer types
	STAT_PTR_INT,
	STAT_PTR_UINT,
	STAT_PTR_BOOL,
	STAT_PTR_FLOAT,
	STAT_STRING, // special pointer type
};

/***********/
/* Defines */
/***********/

#define  ASURA_STATLOG_NOTEAM   -1
#define  ASURA_STATLOG_MAXSTATS 256
#define  ASURA_STATLOG_MAXKEYSIZE  32

// snapshot send mode flags
#define  ASURA_STATLOG_SNAPSHOTFLAG_ONEND       (1<<ASURA_STATLOG_SNAPSHOTMODE_ONEND)
#define  ASURA_STATLOG_SNAPSHOTFLAG_ONINTERVAL  (1<<ASURA_STATLOG_SNAPSHOTMODE_ONINTERVAL)
#define  ASURA_STATLOG_SNAPSHOTFLAG_ONTRIGGER   (1<<ASURA_STATLOG_SNAPSHOTMODE_ONTRIGGER)

/************/
/* Classes. */
/************/

/*************************************************************
 *
 * Class Name    : Asura_Network_StatLogging
 *
 * Author        : Richard May
 *
 * Description   : Game and player statistic logging system
 *				   for use with the gamespy and Xbox live SDKs
 *
 *************************************************************/
class Asura_Network_StatLogging
{

public:
	/* Public interface functions. */	
	static void Initialise(void);
	static void Shutdown(void);
	static void Reset(void);
	static void Update(void);

	// session management
	static bool StartSession(void);
	static void EndSession(void);
	static bool IsSessionInProgress(void);
	
	// snapshot setup & control
	static void SetSnapshotMode( ASURA_STATLOG_SNAPSHOT_MODE eMode, bool bValue );
	static void SetSnapshotMode( u_int uSnapshotFlags );
	static void SetSnapshotInterval( float fTimeBetweenSnapshots );
	static bool TriggerSnapshot(void);
	static u_int GetSnapshotModeFlags(void) { return s_uSnapshotModeFlags; }	

	// variable registering functions
	static bool RegisterVariable( const char* szName, const bool*  pbVar,  Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static bool RegisterVariable( const char* szName, const float* pfVar,  Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static bool RegisterVariable( const char* szName, const int*   piVar,  Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static bool RegisterVariable( const char* szName, const u_int* puVar,  Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static bool RegisterVariable( const char* szName, const char*  szVar, Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );

	// one-shot data adding functions
	// by guid
	static void AddData( const char* szName, bool  bValue, Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static void AddData( const char* szName, float fValue, Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static void AddData( const char* szName, int   iValue, Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	static void AddData( const char* szName, u_int uValue, Asura_Guid uPlayer=ASURA_GUID_UNREGISTERED, int iTeam=ASURA_STATLOG_NOTEAM );
	// by integer id (for setting stats in menus and the like)
	static void AddData( const char* szName, bool  bValue, int iPlayerID );
	static void AddData( const char* szName, float fValue, int iPlayerID );
	static void AddData( const char* szName, int   iValue, int iPlayerID );
	static void AddData( const char* szName, u_int uValue, int iPlayerID );


	// data accessor function
	static const class Asura_StatLogging_Stat* GetStatData(void) { return s_axVariableData; }

protected:
	/* Helper functions. */
	static bool SendSnapshot( bool bGameEnd=false );
	static void ClearAllData(void); 
	static void GenerateVariableKey( char* szNewKey, const char* szName, Asura_Guid uPlayer, int iTeam );

	// platform-specific functions (used to access the corresponding SDK-specific functions)
	static void Platform_Initialise(void);
	static void Platform_Shutdown(void);

	static bool Platform_StartSession(void);
	static void Platform_EndSession(void);

	static bool Platform_SendSnapshot( bool bGameEnd );

protected:
	/* Data. */
	static u_int s_uSnapshotModeFlags;
	static bool  s_bSessionOpen;
	static float s_fSnapshotInterval;
	static float s_fIntervalTimer;

	static class Asura_StatLogging_Stat s_axVariableData[ASURA_STATLOG_MAXSTATS];
	static Asura_HashTree s_xVariableTree;
};
/* end of class Asura_Network_StatLogging */


// simple data class for stat logging
class Asura_StatLogging_Stat
{
public:
	/* Constructor & Destructor */
	Asura_StatLogging_Stat()
	:	m_eType(STAT_NONE)
	,	m_uOwner(ASURA_GUID_UNREGISTERED)
	,	m_iTeamID(ASURA_STATLOG_NOTEAM)
	,	m_bIsGuid(false)
	,   m_bValue(false)
	,	m_iValue(0)
	,	m_uValue(0)
	,	m_fValue(0.0f)		
	,	m_szValue(NULL)
	,	m_pbValue(NULL)
	,	m_piValue(NULL)
	,	m_puValue(NULL)
	,	m_pfValue(NULL)
	{
		strcpy(m_szName,"");
	}
	
	Asura_StatLogging_Stat( STAT_TYPE eType, const char* szName )
	:	m_eType(eType)
	,	m_uOwner(ASURA_GUID_UNREGISTERED)
	,	m_iTeamID(ASURA_STATLOG_NOTEAM)
	,	m_bIsGuid(false)
	,   m_bValue(false)
	,	m_iValue(0)
	,	m_uValue(0)
	,	m_fValue(0.0f)		
	,	m_szValue(NULL)
	,	m_pbValue(NULL)
	,	m_piValue(NULL)
	,	m_puValue(NULL)
	,	m_pfValue(NULL)
	{
		m_szName[0] = 0;
		if(szName)
		{
			strcpy(m_szName,szName);
		}		
	}
	
	~Asura_StatLogging_Stat() {}
	
public: 
	/* data */
	char  m_szName[ASURA_STATLOG_MAXKEYSIZE];
	STAT_TYPE   m_eType;
	u_int		m_uOwner;
	int			m_iTeamID; 
	bool		m_bIsGuid;
	
	union 	
	{
		bool  m_bValue;
		int   m_iValue;
		u_int m_uValue;
		float m_fValue;
		const char* m_szValue;
		
		const bool*  m_pbValue;
		const int*   m_piValue;
		const u_int* m_puValue;
		const float* m_pfValue;
	};
};

#endif // ASURA_USE_NETWORK_STATLOGGING

/****************/
/* End of file. */
/****************/
#endif//__ASURA_NETWORK_STATLOGGING_H__
