///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_Qos_System.cpp
//
// Created by Martin G.Harris (9/4/2009)
// 
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
// Description: System for measuring the upload bandwidth of machines in the
//				lobby of online games. This information can then be used to
//              help determine which machine would make the best game server.
//
///////////////////////////////////////////////////////////////////////////////

// Test network bandwidth by:
// 1) sending a small packet (about 28 bytes) to start the test (zeroth packet).
// 2) delaying (about 100 microseconds).
// 3) sending 2 (i.e. uASURA_QOS_NUM_LARGE_PACKETS) large packets (1024 bytes each)
//    in quick succession (these are packets 1 to 2).
//
// Measure the delay between the zeroth packet and the arrival of the others,
// use this data to estimate the bandwidth.
// Repeat a number of times (up to 16, i.e. uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION) and average the results.


#include "Core/Asura.h"
#include "Network/Asura_Qos_System.h"

#ifdef ASURA_ENABLE_QOS_SYSTEM

// Includes
#include "Core/Asura_Atomic.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_SortAndSearch.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_NetLinkManager.h"
#include "Network/Asura_NetTransport_System.h"

#include <limits.h>

/////////////
// Defines //
/////////////

// Enable logging of all messages. Note: this will affect all the timings significantly.
// NOTE: Do *not* enable this for submission builds!
//#define ASURA_QOS_ENABLE_LOG_ALL_MESSAGES 1

// Enable message loss testing - randomly drops messages (to test for bugs).
// NOTE: Do *not* enable this for submission builds!
//#define ASURA_QOS_TEST_LOST_MESSAGES 1
const float fASURA_QOS_FRACTION_MESSAGES_TO_DROP = 0.1f;

// Default port number for server QoS listener (27015 for PC, 3658 for PS3)
#ifndef ASURA_QOS_DEFAULT_PORT_SERVER
#define ASURA_QOS_DEFAULT_PORT_SERVER 27015
#endif

// Default port number for client QoS listener (0 means use dynamic port number)
#ifndef ASURA_QOS_DEFAULT_PORT_CLIENT
#define ASURA_QOS_DEFAULT_PORT_CLIENT 0
#endif

// Password (to exclude irrelevant messages that may arrive at the port).
#define MAGIC_PASSWORD_NUMBER 0x3212abcd


///////////////
// Constants //
///////////////
const u_int uASURA_ZEROTH_PACKET_SIZE = 28;		// bytes
const u_int uASURA_LARGE_PACKET_SIZE  = 1024;	// bytes

const float fASURA_QOS_DELAY_BEFORE_TEST = 1.0f; // seconds

// MGH 7/5/2009: Note: that on fast LANs using BSD sockets, the networking is so fast that the
// timing of tests can become synchronised with the sleep time of the receiving thread.
// By adding an extra 0.0006f (i.e. 600 microseconds) we can desync them again.

// Delay between tests.
const float fASURA_QOS_DELAY_BETWEEN_TESTS_QUICK		= 0.2006f;	// Quick test (seconds).
const float fASURA_QOS_DELAY_BETWEEN_TESTS_SLOW			= 1.0006f;	// Slow test for debugging (seconds).

// Delay between packets.
const float fASURA_QOS_DELAY_BEFORE_LARGE_PACKETS  = 200.0e-6f;	// Delay between zeroth and first packet (seconds).
const float fASURA_QOS_DELAY_BETWEEN_LARGE_PACKETS = 100.0e-6f;	// Delay between large packets (seconds)

// Minimum time (in seconds) between keep alive messages.
const float fASURA_QOS_KEEP_ALIVE_TIME = 1.1f;

// If no messages received for this amount of time then we have connection problems.
const float fASURA_QOS_CONNECTION_PROBLEMS_TIME = 4.0f; // seconds

// If no messages received for this amount of time then drop the connection.
const float fASURA_QOS_DISCONNECT_TIMEOUT = 15.0f; // seconds

// If a test isn't finished in this time, then abort it (we've probably lost a message somewhere).
const float fASURA_QOS_TEST_TIMEOUT_TIME = 1.5f; // seconds

// Size of header for QoS, UDP, IP etc.
// QoS: 8 bytes for password, msg id, packet id.
const u_int uASURA_QOS_DATA_HEADER_LENGTH = sizeof(Asura_Qos_MessageHeader); // 8
const int   iASURA_QOS_DATA_HEADER_LENGTH = static_cast<int>(uASURA_QOS_DATA_HEADER_LENGTH);

// UDP: 8 bytes for source/destination ports, length, checksum.
// IPv4: 12 bytes for source/destination address, zeros, protocol, UDP length.
// Note: this may not be exactly correct for PS3/steamworks, but for 1024 byte packets
// the difference in header lengths will make negligible difference in the results.
const u_int uASURA_QOS_UDPIP_HEADER_LENGTH = 8 + 12;

// One bit is set for each packet in the test (currently there are three packets: 0,1,2).
const u_int uASURA_QOS_LARGE_PACKETS_MASK = ( 1<<(uASURA_QOS_NUM_LARGE_PACKETS+1) ) - 1;

// Clamp the maximum measured value.
const int iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC = 10000; // 10 Mbps

#ifdef ASURA_LOGFILES_ENABLED
static const char* ASURA_QOS_LOGFILE_NAME = "logfiles\\qos-summary.txt";
#endif //ASURA_LOGFILES_ENABLED

///////////
// Enums //
///////////
// Data indices for the results message
enum ASURA_QOS_RESULT_MESSAGE_ID
{
	ASURA_QOS_RESULT_MESSAGE_ID_FLAGS,
	ASURA_QOS_RESULT_MESSAGE_ID_SEND_BAND,
	ASURA_QOS_RESULT_MESSAGE_ID_RECV_BAND,
	ASURA_QOS_RESULT_MESSAGE_ID_RTT,

	ASURA_QOS_RESULT_MESSAGE_ID_NUM_VARS
};

//
// Static variables
//

// Enable verbose messages in the logging and output.
DEBUGVAR bool dbg_bQosVerbose = false;

// Enable quick testing (normal case) or slow testing (for debugging purposes).
DEBUGVAR bool dbg_bQosQuick   = true;


///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_ResultCalculator
///////////////////////////////////////////////////////////////////////////////
// Calculate the min, max, mean and median of some data.
template<u_int uMAX_SIZE>
class Asura_Qos_ResultCalculator
{
public:
	Asura_Qos_ResultCalculator()
		: m_iMin(INT_MAX)
		, m_iMax(INT_MIN)
		, m_iMean(0)
		, m_iMedian(0)
		, m_uNumDataPoints(0)
	{
	}

	int GetMin() const		{ return m_iMin; }
	int GetMax() const		{ return m_iMax; }
	int GetMean() const		{ return m_iMean; }
	int GetMedian() const	{ return m_iMedian; }

	u_int GetNumDataPoints() const { return m_uNumDataPoints; }

	void AddDataPoint(int iValue)
	{
		if (iValue <= 0) return; // Only positive values are valid data.

		Asura_Assert(m_uNumDataPoints<uMAX_SIZE, "Asura_Qos_ResultCalculator: No more space!");
		if (m_uNumDataPoints >= uMAX_SIZE) return;

		m_aiData[m_uNumDataPoints] = iValue;
		++m_uNumDataPoints;

		m_iMean += iValue;

		if (iValue > m_iMax) m_iMax = iValue;
		if (iValue < m_iMin) m_iMin = iValue;
	}

	void Calculate()
	{
		if (m_uNumDataPoints > 0)
		{
			m_iMean /= m_uNumDataPoints;
			Asura_SortAndSearch::Quicksort(m_aiData, m_uNumDataPoints, Asura_SortAndSearch::SortAscendingInt);
			m_iMedian = (m_aiData[(m_uNumDataPoints-1)/2] + m_aiData[m_uNumDataPoints/2])/2;
		}

		// Invalid results are set to be zero.
		if (m_iMin == INT_MAX) m_iMin = 0;
		if (m_iMax == INT_MIN) m_iMax = 0;
	}

private:
	int m_iMin;
	int m_iMax;
	int m_iMean;
	int m_iMedian;

	u_int m_uNumDataPoints;
	int m_aiData[uMAX_SIZE];
};



/////////////////////////
// GetNetAddressString //
/////////////////////////
#ifdef ASURA_QOS_ENABLE_LOGGING
// Convert address (NBO) and port (HBO) to a string (which is held in a static buffer, so copy or use it immediately).
// Note: this isn't thread-safe, but mostly its used for logging and debugging.
static const char* GetNetAddressString(const Asura_Qos_NetConnectionInfo& xInfo)
{
	return Asura_NetTransport_System::GetStringFromAddress(xInfo.m_xAddress, xInfo.m_usPort);
}
#endif	//ASURA_QOS_ENABLE_LOGGING

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_System
///////////////////////////////////////////////////////////////////////////////

Asura_Qos_Listener* Asura_Qos_System::s_pxListener = 0;
bool Asura_Qos_System::s_bInitialised = false;

u_int volatile Asura_Qos_System::s_uNumberOfActiveThreads = 0;
bool volatile Asura_Qos_System::s_bTerminateThreads  = false;
bool volatile Asura_Qos_System::s_bThreadShouldSleep = false;
bool volatile Asura_Qos_System::s_bThreadIdle		 = false;

float Asura_Qos_System::s_fTimeSlice = 0.0f;
Asura_TickCount Asura_Qos_System::s_xLastTickCount;
Asura_NetAddress Asura_Qos_System::s_xServerAddress;
int Asura_Qos_System::s_iNumStoredResults = 0;

////////////////
// Initialise //
////////////////
void Asura_Qos_System::Initialise()
{
	Asura_Assert(!s_bInitialised, "Already initialised, can't initialise!");
	if (!s_bInitialised)
	{
		s_bInitialised = true;
		s_bTerminateThreads = false;
		s_bThreadShouldSleep = false;
		s_bThreadIdle = false;

		Asura_Assert(s_uNumberOfActiveThreads == 0, "Number of active Qos threads should be zero!");
		s_uNumberOfActiveThreads = 0;

		s_fTimeSlice = 0.0f;
		s_iNumStoredResults = 0;

		Asura_Mem::Set(&s_xServerAddress, 0, sizeof(s_xServerAddress));

		Platform_Initialise();

#ifdef ASURA_QOS_TEST_LOST_MESSAGES
		Asura_OutputToDebugger_Retail(("Warning: QoS system will drop %.1f%% of messages.\n\n", (fASURA_QOS_FRACTION_MESSAGES_TO_DROP*100)));
#endif //ASURA_QOS_TEST_LOST_MESSAGES

#ifdef ASURA_LOGFILES_ENABLED
		Asura_SimpleLogfileClear(ASURA_QOS_LOGFILE_NAME);
#endif //ASURA_LOGFILES_ENABLED
	}
}

//////////////
// Shutdown //
//////////////
void Asura_Qos_System::Shutdown()
{
	// Note: due to ordering issues with the Steamworks shutdown, the Qos system shutdown may be called twice.
//	Asura_Assert(s_bInitialised, "Not initialised, can't shutdown!");

	if (s_bInitialised)
	{
		// Should already have destroyed the listener, but make sure.
		DestroyListener();

		s_bInitialised = false;
		Platform_Shutdown();
	}
}

//////////////////////
// SetServerAddress //
//////////////////////
// The lobby code sets the server address (NBO) when the client joins a lobby.
void Asura_Qos_System::SetServerAddress(Asura_NetAddress xServerAddress)
{
	s_xServerAddress = xServerAddress;
}

/////////////////
// CreateLobby //
/////////////////
void Asura_Qos_System::CreateLobby(bool bIsHost)
{
	if (s_pxListener)
	{
		Asura_Assert(0, "Qos listener already created!");
		return;
	}

	const u_short usLocalPort = static_cast<u_short>(bIsHost ? ASURA_QOS_DEFAULT_PORT_SERVER : ASURA_QOS_DEFAULT_PORT_CLIENT);
	const u_short usServerPort = static_cast<u_short>(ASURA_QOS_DEFAULT_PORT_SERVER);

	CreateListener(usLocalPort, bIsHost);

	if (bIsHost)
	{
		// If we are the host then clear the host address.
		Asura_Mem::Set(&s_xServerAddress, 0, sizeof(s_xServerAddress));
	}
	else
	{
		// On the client set the host address and port.
		AddServerConnection(s_xServerAddress, usServerPort);
	}

	// Create the socket
	if (s_pxListener)
	{
		s_pxListener->CreateSocket();
	}

	if ( !IsSocketValid() )
	{
		// Failed
		Asura_Assert(0, "Failed to create Qos listener socket!");
		DestroyListener();
		return;
	}
	
	// Note: we could raise the thread priority here, or set the thread affinity
	// (to stop clock glitches caused by the thread being moved to a different CPU).
	// However the code is pretty good at excluding bogus results caused, for example,
	// by the receiving thread being asleep when the test packets arrive.
	// So for now we'll just leave the thread at default priority.
	const bool bCreated = Asura_Thread_System::CreateThread( "Qos Listener Thread", UpdateThread, 0, ASURA_MULTITHREAD_STACK_SIZE_SMALL );
	if (!bCreated)
	{
		// Failed
		Asura_Assert(0, "Failed to create Qos listener thread!");
		DestroyListener();
		return;
	}
}

void Asura_Qos_System::LobbyTerminated()
{
	// Close connections gracefully, rather than just letting them timeout.
	DestroyListener(true);
}

void Asura_Qos_System::StartingGame()
{
	s_bThreadShouldSleep = true;
}

void Asura_Qos_System::EndingGame()
{
	s_bThreadShouldSleep = false;
}

void Asura_Qos_System::UpdateThread(const void* pUserData)
{
	Asura_Atomic::Increment( s_uNumberOfActiveThreads );

#ifdef ASURA_QOS_ENABLE_LOGGING
	Asura_OutputToDebugger_Retail(("QoS: Started listener thread.\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

	// Clients try to connect to the host.
	if (!s_bTerminateThreads)
	{
		if (s_pxListener && !s_pxListener->IsHost())
		{
			RequestServerConnect();
		}
	}

	while ( !s_bTerminateThreads && !IsDone() )
	{
		Update();

		if (s_bThreadShouldSleep || s_bThreadIdle)
		{
			// Wait for a short time and then check for messages again.
			Asura_Thread_System::DelayCurrentThread(fASURA_QOS_THREAD_SLEEP_TIME);
		}
	}

#ifdef ASURA_QOS_ENABLE_LOGGING
	Asura_OutputToDebugger_Retail(("QoS: Stopped listener thread.\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

	Asura_Atomic::Decrement( s_uNumberOfActiveThreads );
}



//
// Listener functions
//
void Asura_Qos_System::CreateListener(u_short usLocalPort, bool bIsHost)
{
	Asura_Assert(s_bInitialised, "Not initialised, can't create Qos listener!");
	if (!s_bInitialised) return;
	if (s_pxListener) return;

	s_bTerminateThreads = false;
	s_bThreadIdle		= false;

	s_fTimeSlice = 0.0f;
	s_xLastTickCount = Asura_Timers::GetCurTicks();

	// Create the listener
	if (bIsHost)
	{
		s_pxListener = new Asura_Qos_Listener_Server(usLocalPort);
	}
	else
	{
		s_pxListener = new Asura_Qos_Listener_Client(usLocalPort);
	}
}

void Asura_Qos_System::DestroyListener(bool bCloseConnections /*= false*/)
{
	// Tell the listener thread to finish.
	s_bTerminateThreads = true;
	
	// Wait for the thread to stop...
	while (s_uNumberOfActiveThreads != 0)
	{
		Asura_Thread_System::YieldCurrentThread();
	}

	if (s_pxListener && bCloseConnections)
	{
		s_pxListener->CloseConnections();
	}

	delete s_pxListener;
	s_pxListener = 0;

	s_bTerminateThreads  = false;
	s_bThreadShouldSleep = false;
	s_bThreadIdle		 = false;
}

bool Asura_Qos_System::IsSocketValid()
{
	return s_pxListener && s_pxListener->IsSocketValid();
}

bool Asura_Qos_System::IsDone()
{
	return !s_pxListener || s_pxListener->IsDone();
}

void Asura_Qos_System::AddServerConnection(Asura_NetAddress xServerAddress, u_short usServerPort)
{
	if (s_pxListener)
	{
		Asura_Qos_NetConnectionInfo xServerInfo(xServerAddress, usServerPort);
		s_pxListener->AddConnection(xServerInfo);
	}
}

void Asura_Qos_System::RequestServerConnect()
{
	if (s_pxListener)
	{
		s_pxListener->RequestConnect();
	}
}

void Asura_Qos_System::Update()
{
	if (s_pxListener)
	{
		UpdateTimeSlice();
		s_pxListener->Update();
	}
}

/////////////////////
// UpdateTimeSlice //
/////////////////////
// Each time that the listener thread updates, we need to calculate a 'frame' time.
void Asura_Qos_System::UpdateTimeSlice()
{
	const Asura_TickCount xNow = Asura_Timers::GetCurTicks();
	s_fTimeSlice = Asura_Timers::TicksToSeconds(xNow - s_xLastTickCount);
	s_xLastTickCount = xNow;
}


/////////////////
// StoreResult //
/////////////////
// Store the result of the testing.
void Asura_Qos_System::StoreResult(int iUploadBandwidthKbps, int iDownloadBandwidthKbps)
{
	// The Xbox360 is currently storing the results for the local computer
	// in the netlink manager, so we may as well do the same for the PC and PS3.

	// The Xbox360 SDK QoS listener seems to be clamping the measured values
	// so we'll clamp them too here. The measurements are only accurate for
	// low bandwidths anyway, and we are only really interested in excluding
	// computers with low bandwidth from being servers.
	
	// Clamp the bandwidths.
	const int iUploadClampedBandKbps   = Asura_Maths::ClampToRange(iUploadBandwidthKbps,   0, iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC);
	const int iDownloadClampedBandKbps = Asura_Maths::ClampToRange(iDownloadBandwidthKbps, 0, iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC);

	// Convert from kbps to bits per second.
	const u_int uUploadBandwidth   = static_cast<u_int>(1000 * iUploadClampedBandKbps);
	const u_int uDownloadBandwidth = static_cast<u_int>(1000 * iDownloadClampedBandKbps);

	// Previously stored results in bits per second.
	const u_int uOldUploadBandwidth   = Asura_NetLinkManager::GetApproxUploadBandwidth();
	const u_int uOldDownloadBandwidth = Asura_NetLinkManager::GetApproxDownloadBandwidth();
	
	// We are mainly interested in the upload bandwidth.
	// First check that we have a valid result.
	if (uUploadBandwidth > 0)
	{
		if (s_iNumStoredResults == 0 || uOldUploadBandwidth == 0)
		{
			// No upload bandwidth has been set yet. Just use the most recent calculation.
			Asura_NetLinkManager::SetApproxUploadBandwidth(uUploadBandwidth);
		}
		else
		{
			// On the server we get a number of upload bandwidth results (one for each client).
			// Otherwise clients get a result each time they go into a new lobby.
			// Average all of these results.
			const int iOldUploadBandwidthKbps = static_cast<int>(uOldUploadBandwidth/1000);
			
			int iAvUploadBandwidthKbps = (s_iNumStoredResults*iOldUploadBandwidthKbps + iUploadClampedBandKbps)/(s_iNumStoredResults+1);
			iAvUploadBandwidthKbps = Asura_Maths::ClampToRange(iAvUploadBandwidthKbps, 0, iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC);
			
			const u_int uAvUploadBandwidth = static_cast<u_int>(1000 * iAvUploadBandwidthKbps);

			Asura_NetLinkManager::SetApproxUploadBandwidth(uAvUploadBandwidth);
		}
		
		++s_iNumStoredResults;
	}

	// The measured download bandwidth is mainly limited by the upload bandwidths of
	// the remote computers (for ADSL anyway). So just use the maximum value that
	// has been measured in this and previous tests (maybe one of the remote
	// computers had good upload bandwidth and we'll get an accurate result).
	if (uDownloadBandwidth > uOldDownloadBandwidth)
	{
		Asura_NetLinkManager::SetApproxDownloadBandwidth(uDownloadBandwidth);
	}
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_Result
///////////////////////////////////////////////////////////////////////////////
void Asura_Qos_Result::PrintToBuffer(char* szBuffer, u_int uBufferSize) const
{
#ifdef ASURA_QOS_ENABLE_LOGGING
	if (!szBuffer || !uBufferSize) return;

	const char* szString0 = (IsDownloadFromRemote()	? "DOWN "	: "UP "); // Locally download or upload.
	const char* szString1 = (IsRejected()			? "REJECT "	: "OK ");

	const char* szString2 = (IsOutOfOrder()	      ? "(out-of-order) "        : "");
	const char* szString3 = (IsQueued()           ? "(queuing) "             : "");
	const char* szString4 = (IsZerothOutOfOrder() ? "(zeroth-out-of-order) " : "");
	const char* szString5 = (HasTimedOut()        ? "(timeout) "             : "");
	const char* szString6 = (HasLostPackets()     ? "(lost packets) "        : "");
	const char* szString7 = (IsTooFast()		  ? "(too fast) "            : "");
	const char* szString8 = (IsLate()			  ? "(late) "                : "");

	Asura_PrintfToBfr(szBuffer, "%2u: b0 =%7d kbps    b1 =%7d kbps    RTT =%6d  %s%s%s%s%s%s%s%s%s\n", uBufferSize, true,
					  m_uTestIDNumber, m_iSendBandKbps, m_iRecvBandKbps, m_iAdjustedRTT,
					  szString0, szString1, szString2, szString3, szString4, szString5, szString6, szString7, szString8);
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_TestTicks
///////////////////////////////////////////////////////////////////////////////
void Asura_Qos_TestTicks::Reset()
{
	for (u_int u=0; u<=uASURA_QOS_NUM_LARGE_PACKETS; ++u)
	{
		Asura_Mem::Set(&m_axTicks[u], 0, sizeof(Asura_TickCount));
	}
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_Test
///////////////////////////////////////////////////////////////////////////////
void Asura_Qos_Test::Reset()
{
	m_eTestState = ASURA_QOS_TESTSTATE_NULL;

	m_usPacketSeqID = 0;

	m_xCurrentTest.Reset();
	m_xPrevTest.Reset();

	m_fDelayTimer			= 0.0f;
	m_fRequestedDelay		= 0.0f;
	m_uNumTestResults		= 0;
	
	m_usNumValidUploadResults   = 0;
	m_usNumValidDownloadResults = 0;

	for (u_int u=0; u<uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION; ++u)
	{
		m_axResult[u].Reset();
	}

	m_uReceivedPacketFlags	= 0;
	m_uReceivedFlags		= 0;
	m_uReceivedSubIndex		= 0;

	m_bDownloadFromRemote = true;
	m_bHasFinished = false;
	m_bHasSummarized = false;
}

///////////////
// AddResult //
///////////////
// Add the result of a test.
// Return result if added, or null otherwise.
const Asura_Qos_Result* Asura_Qos_Test::AddResult(  u_int uTestIDNumber,
													u_int uFlags,
													int iSendBandKbps,
													int iRecvBandKbps,
													int iAdjustedRTT )
{
	if (m_uNumTestResults < uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION)
	{
		// Ignore duplicate results.
		for (u_int u = 0; u<m_uNumTestResults; ++u)
		{
			const bool bTestDownloadFromRemote = ( 0 != (uFlags & ASURA_QOS_RESULT_FLAG_DOWNLOAD_FROM_REMOTE) );
			
			// If the test has the same test ID number and same direction then it is a duplicate.
			if (m_axResult[u].m_uTestIDNumber == uTestIDNumber &&
				m_axResult[u].IsDownloadFromRemote() == bTestDownloadFromRemote)
			{
#ifdef ASURA_QOS_ENABLE_LOGGING
				Asura_OutputToDebugger(("QoS: Ignoring duplicate test result %u\n", uTestIDNumber));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

				return 0;
			}
		}

		if ( (uFlags & ASURA_QOS_RESULT_FLAG_REJECT) == 0 )
		{
			if ( uFlags & ASURA_QOS_RESULT_FLAG_DOWNLOAD_FROM_REMOTE )
			{
				++m_usNumValidDownloadResults;
			}
			else
			{
				++m_usNumValidUploadResults;
			}
		}

		m_axResult[m_uNumTestResults].m_uTestIDNumber	= uTestIDNumber;
		m_axResult[m_uNumTestResults].m_uFlags			= uFlags;
		m_axResult[m_uNumTestResults].m_iSendBandKbps	= iSendBandKbps;
		m_axResult[m_uNumTestResults].m_iRecvBandKbps	= iRecvBandKbps;
		m_axResult[m_uNumTestResults].m_iAdjustedRTT	= iAdjustedRTT;

		// Start off testing the download bandwidth from the remote machine,
		// then when we have enough results switch the direction around.
		if (m_usNumValidDownloadResults >= uASURA_QOS_NUM_TEST_RESULTS_NEEDED)
		{
			if (m_bDownloadFromRemote)
			{
				// Reverse the direction of the test.
				m_bDownloadFromRemote = false;
				m_eTestState = ASURA_QOS_TESTSTATE_NULL;
			}
		}

		// When we have enough results then finish the testing.
		if (m_usNumValidUploadResults   >= uASURA_QOS_NUM_TEST_RESULTS_NEEDED &&
			m_usNumValidDownloadResults >= uASURA_QOS_NUM_TEST_RESULTS_NEEDED)
		{
			m_bHasFinished = true;
		}

		++m_uNumTestResults;
		return &m_axResult[m_uNumTestResults-1];
	}

#ifdef ASURA_QOS_ENABLE_LOGGING
	Asura_OutputToDebugger(("QoS: No more space to store test result %u\n", uTestIDNumber));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

	return 0;
}


//////////////////////
// UpdateDelayTimer //
//////////////////////
void Asura_Qos_Test::UpdateDelayTimer()
{
	m_fDelayTimer -= Asura_Qos_System::GetTimeSlice();
	if (m_fDelayTimer < 0.0f)
	{
		m_fDelayTimer = 0.0f;
	}
}

////////////////
// UpdateTest //
////////////////
void Asura_Qos_Test::UpdateTest(Asura_Qos_NetConnection* pxConnection)
{
	switch (m_eTestState)
	{
	case ASURA_QOS_TESTSTATE_NULL:
		// Connected. Wait before starting testing.
		break;
		
	case ASURA_QOS_TESTSTATE_START:
		if (m_fDelayTimer == 0.0f && !pxConnection->HasConnectionProblems())
		{
			// Start a large packet test once the timer has run out,
			// assuming that there are no connection problems.
			SendTestLargePacketStart(pxConnection);
			SetRequestedDelay(fASURA_QOS_DELAY_BEFORE_LARGE_PACKETS);
			m_eTestState = ASURA_QOS_TESTSTATE_SENDING;
		}
		break;

	case ASURA_QOS_TESTSTATE_SENDING:
		if (m_fDelayTimer == 0.0f)
		{
			const bool bFinished = SendTestLargePacket(pxConnection);
			SetRequestedDelay(fASURA_QOS_DELAY_BETWEEN_LARGE_PACKETS);
			if (bFinished)
			{
				m_eTestState = ASURA_QOS_TESTSTATE_POST_TEST;
				SetRequestedDelay(fASURA_QOS_DELAY_BETWEEN_TESTS_SLOW);
			}
		}
		break;

	case ASURA_QOS_TESTSTATE_POST_TEST:
		// Wait for response
		if (m_fDelayTimer == 0.0f)
		{
			HandlePostTestEnd(pxConnection);
		}
		break;

	case ASURA_QOS_TESTSTATE_WAITING:
		break;

	case ASURA_QOS_TESTSTATE_TESTING:
		break;
	}
}

///////////////////////
// HandlePostTestEnd //
///////////////////////
void Asura_Qos_Test::HandlePostTestEnd(Asura_Qos_NetConnection* pxConnection)
{
	if (HasFinishedTests())
	{
		// Finished collecting data. Summarize the results and close the connection.
		pxConnection->SendMessageTo(ASURA_QOS_MSGID_MSG_CONNECTION_CLOSE);
		pxConnection->CloseConnection();
	}
	
	SetRequestedDelay(0.0f);
	m_eTestState = ASURA_QOS_TESTSTATE_NULL;	
}

//////////////////////////////
// SendTestLargePacketStart //
//////////////////////////////
void Asura_Qos_Test::SendTestLargePacketStart(Asura_Qos_NetConnection* pxConnection)
{
	// Save the previous tick values before starting a new test.
	m_xPrevTest = m_xCurrentTest;

	// l.s.byte stores the sub-packet number, m.s.byte stores the large packet test number.
	u_short usPacketID = m_usPacketSeqID;
	usPacketID &= 0xff00; // set sub-packet number to zero
	usPacketID += 0x0100; // increment the test number
	m_usPacketSeqID = usPacketID;

	// First packet is small.
	pxConnection->SendPacketMessageTo(m_usPacketSeqID, uASURA_ZEROTH_PACKET_SIZE);
	m_xCurrentTest.m_axTicks[0] = Asura_Timers::GetCurTicks();
}

/////////////////////////
// SendTestLargePacket //
/////////////////////////
// Return true if we've sent the last large packet of the set.
bool Asura_Qos_Test::SendTestLargePacket(Asura_Qos_NetConnection* pxConnection)
{
	const u_int uPrevSubNumber = m_usPacketSeqID & 0x00ff;

	// Wait the last few microseconds that are remaining.
	// It used to undershoot on the request time, now it overshoots slightly.
	const float fRequestedDelay = GetRequestedDelay();
	if (fRequestedDelay > 0.0f && uPrevSubNumber<uASURA_QOS_NUM_LARGE_PACKETS)
	{
		const Asura_TickCount& xPrevTicks = m_xCurrentTest.m_axTicks[uPrevSubNumber];
		while (Asura_Timers::TicksToSeconds(Asura_Timers::GetElapsedTicks(xPrevTicks)) < fRequestedDelay)
		{
		}
	}

	// increment the sub-packet number
	++m_usPacketSeqID;

	const u_int uSubNumber = m_usPacketSeqID & 0x00ff;

	// Following packets are bigger.
	pxConnection->SendPacketMessageTo(m_usPacketSeqID, uASURA_LARGE_PACKET_SIZE);

	if (uSubNumber <= uASURA_QOS_NUM_LARGE_PACKETS)
	{
		m_xCurrentTest.m_axTicks[uSubNumber] = Asura_Timers::GetCurTicks();
	}	

	return (uSubNumber == uASURA_QOS_NUM_LARGE_PACKETS);
}

//////////////////
// HandlePacket //
//////////////////
void Asura_Qos_Test::HandlePacket(Asura_Qos_NetConnection* pxConnection, u_short usPacketSeqID)
{
	const u_int uTestNumber = (usPacketSeqID >> 8);
	const u_int uSubNumber  = (usPacketSeqID & 0xff);

	const u_int uPrevTestNumber = (m_usPacketSeqID >> 8);
	const u_int uPrevSubNumber  = (m_usPacketSeqID & 0xff);

	bool bIgnored = false;

	if (uTestNumber > uPrevTestNumber)
	{
		// Starting a new test...
		if (m_uReceivedPacketFlags != 0)
		{
			// Old test hadn't finished. Report the lost packets.
#ifdef ASURA_QOS_ENABLE_LOGGING
			Asura_OutputToDebugger(("QoS: Lost packets: test = %u  recv mask = 0x%x\n", uPrevTestNumber, m_uReceivedPacketFlags));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

			m_uReceivedFlags = ASURA_QOS_RESULT_FLAG_LOSTPACKETS;
			pxConnection->SendRecvTimingMessage(false, m_usPacketSeqID);
		}

		if (uSubNumber == 0)
		{
			// Zeroth packet
			StartingTest();
			m_usPacketSeqID = usPacketSeqID;
			m_xCurrentTest.m_axTicks[0] = pxConnection->m_xLastMessageRecvTicks;
			m_uReceivedPacketFlags |= (1<<uSubNumber);
			m_uReceivedSubIndex = 1;
		}
		else
		{
			// Not zeroth packet.
			bIgnored = true;
			m_uReceivedFlags = (ASURA_QOS_RESULT_FLAG_ZEROTH_OUTOFORDER | ASURA_QOS_RESULT_FLAG_REJECT);
			pxConnection->SendRecvTimingMessage(true, usPacketSeqID);
			WaitForNextTest();
		}
	}
	else if (uTestNumber == uPrevTestNumber)
	{
		// Continuing the current test...
		if (m_eTestState == ASURA_QOS_TESTSTATE_WAITING)
		{
			// Test has already been stopped.
			bIgnored = true;
		}
		else
		{
			if (uSubNumber != uPrevSubNumber + 1)
			{
				// Out-of-order packet
				m_uReceivedFlags = ASURA_QOS_RESULT_FLAG_OUTOFORDER;		
			}

			// Save the packet details and ticks.
			m_usPacketSeqID = usPacketSeqID;

			if ( (m_uReceivedPacketFlags & (1<<uSubNumber)) == 0 )
			{
				m_uReceivedPacketFlags |= (1<<uSubNumber);

				if (m_uReceivedSubIndex <= uASURA_QOS_NUM_LARGE_PACKETS )
				{
					m_xCurrentTest.m_axTicks[m_uReceivedSubIndex] = pxConnection->m_xLastMessageRecvTicks;
					++m_uReceivedSubIndex;
				}
			}
		}

		// Check whether all the large packets have arrived
		if (m_eTestState == ASURA_QOS_TESTSTATE_TESTING)
		{
			if ((m_uReceivedPacketFlags & uASURA_QOS_LARGE_PACKETS_MASK) == uASURA_QOS_LARGE_PACKETS_MASK)
			{
				// Got all the packets. Send results back to the sender...
				pxConnection->SendRecvTimingMessage(false, usPacketSeqID);
				WaitForNextTest();
			}
		}
	}
	else
	{
		// Old test packet, ignore.
		bIgnored = true;
	}

#ifdef ASURA_QOS_ENABLE_LOGGING
	if (bIgnored)
	{
		Asura_OutputToDebugger(("QoS: Ignoring out-of-order packet: test: %u  sub: %u\n", uTestNumber, uSubNumber));
	}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
}

/////////////////////
// WaitForNextTest //
/////////////////////
void Asura_Qos_Test::WaitForNextTest()
{
	m_eTestState = ASURA_QOS_TESTSTATE_WAITING;
	m_uReceivedPacketFlags = 0;
	m_uReceivedFlags = 0;
	m_uReceivedSubIndex = 0;
}

//////////////////
// StartingTest //
//////////////////
void Asura_Qos_Test::StartingTest()
{
	m_eTestState = ASURA_QOS_TESTSTATE_TESTING;
	m_uReceivedPacketFlags = 0;
	m_uReceivedFlags = 0;
	m_uReceivedSubIndex = 0;
}

/////////////////////
// HandleStartTest //
/////////////////////
void Asura_Qos_Test::HandleStartTest()
{
	Asura_Assert(m_eTestState == ASURA_QOS_TESTSTATE_NULL, "Test state isn't null");
	m_eTestState = ASURA_QOS_TESTSTATE_START;
}

//////////////////////
// SummarizeResults //
//////////////////////
void Asura_Qos_Test::SummarizeResults(Asura_Qos_NetConnection* pxConnection)
{
	if (m_bHasSummarized)
	{
		// Only summarize the results once.
		return;
	}

	if ( m_usNumValidUploadResults == 0 && m_usNumValidDownloadResults == 0 )
	{
		// No results to summarize.
		return;
	}

	m_bHasSummarized = true;

	// Work out the min, max, mean and median of the received bandwidth (in kbps) and RTT (in microseconds).
	Asura_Qos_ResultCalculator<uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION> xRTT;
	Asura_Qos_ResultCalculator<uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION> xRecvBandUp;
	Asura_Qos_ResultCalculator<uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION> xRecvBandDown;

	for (u_int u=0; u<uASURA_QOS_MAX_NUM_TESTS_PER_CONNECTION; ++u)
	{
		const Asura_Qos_Result& xResult = GetResult(u);
		if (xResult.IsEmpty()) continue;

		if (xResult.IsValid())
		{
			if (xResult.IsDownloadFromRemote())
			{
				xRecvBandDown.AddDataPoint(xResult.m_iRecvBandKbps);
			}
			else
			{
				xRecvBandUp.AddDataPoint(xResult.m_iRecvBandKbps);
			}

			xRTT.AddDataPoint(xResult.m_iAdjustedRTT);
		}
	}

	xRecvBandUp.Calculate();
	xRecvBandDown.Calculate();
	xRTT.Calculate();

#ifdef ASURA_QOS_ENABLE_LOGGING
	const char* szAddress = GetNetAddressString(pxConnection->GetConnectionInfo());

	const u_int uBUFFER_SIZE = 1024;
	char szBuffer[uBUFFER_SIZE];
	szBuffer[0] = 0;

	Asura_PrintfToBfr(szBuffer,
					  "[%s] Recv UP   (kbps): mean =%7d   median =%7d   max =%7d\n"
					  "[%s] Recv DOWN (kbps): mean =%7d   median =%7d   max =%7d\n"
					  "[%s] RTT      (micro): mean =%7d   median =%7d   min =%7d\n\n",
					  uBUFFER_SIZE, true,
					  szAddress,   xRecvBandUp.GetMean(),	  xRecvBandUp.GetMedian(),	  xRecvBandUp.GetMax(),
					  szAddress, xRecvBandDown.GetMean(),	xRecvBandDown.GetMedian(),	xRecvBandDown.GetMax(),
					  szAddress,          xRTT.GetMean(),			 xRTT.GetMedian(),			 xRTT.GetMin());

	Asura_OutputToDebugger_Retail(("%s", szBuffer));
	Asura_SimpleLogfileOutput_PreFormatted(ASURA_QOS_LOGFILE_NAME, szBuffer, strlen(szBuffer));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING


	// Store the median bandwidth results.
	Asura_Qos_System::StoreResult(xRecvBandUp.GetMedian(), xRecvBandDown.GetMedian());
}

//////////////////////
// HandleRecvTiming //
//////////////////////
void Asura_Qos_Test::HandleRecvTiming(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage)
{
	// Since we've received the timing data we can leave the post test state.
	if (m_eTestState == ASURA_QOS_TESTSTATE_POST_TEST)
	{
		HandlePostTestEnd(pxConnection);
	}

	const u_short usPacketSeqID = pxMessage->GetPacketSeqID();

	// Check that the test number is correct.
	const u_short usSequence = GetPacketSeqID();
	const u_int uCurrentTestIDNumber = (usSequence>>8);
	const u_int uMessageTestIDNumber = (usPacketSeqID>>8);
	u_int uFlags = pxMessage->GetData_Int(0);
	
	if ( (uFlags & ASURA_QOS_RESULT_FLAG_ZEROTH_OUTOFORDER) != 0 )
	{
		// In this case the received data is not valid.
	}
	else if ( (uFlags & ASURA_QOS_RESULT_FLAG_LOSTPACKETS) != 0 )
	{
		// In this case some of the packets were lost.
		if (uMessageTestIDNumber+1 == uCurrentTestIDNumber)
		{
			// We've just received the data for the previous test (somewhat late).
			// Analyse it using the previously stored packets ticks.
			AnalyseData(pxConnection, pxMessage, true);
			return;
		}
	}
	else if ( uMessageTestIDNumber == uCurrentTestIDNumber )
	{
		// Analyse the result for the current test. All packets were received.
		AnalyseData(pxConnection, pxMessage, false);
		return;
	}
	else
	{
		if (uMessageTestIDNumber+1 == uCurrentTestIDNumber)
		{
			// We've just received the data for the previous test (somewhat late).
			// Analyse it using the previously stored packets ticks.
			AnalyseData(pxConnection, pxMessage, true);
			return;
		}
		else
		{
			// Result arrived back far too late. We have already overwritten the server ticks data.
			uFlags |= ASURA_QOS_RESULT_FLAG_TIMEOUT;
		}
	}

	// Reject the result.
	uFlags |= ASURA_QOS_RESULT_FLAG_REJECT;
	const Asura_Qos_Result* pxResult = AddResult(uMessageTestIDNumber, uFlags, 0, 0, 0);
	PrintResult(pxConnection, pxResult, 0, 0);
}

//////////////////
// HandleResult //
//////////////////
void Asura_Qos_Test::HandleResult(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage)
{
	const u_short usPacketSeqID = pxMessage->GetPacketSeqID();
	const u_int uTestIDNumber = (usPacketSeqID>>8);

	u_int uFlags = pxMessage->GetData_Int(ASURA_QOS_RESULT_MESSAGE_ID_FLAGS);
	uFlags |= ASURA_QOS_RESULT_FLAG_DOWNLOAD_FROM_REMOTE;

	const int iSendBandKbps = pxMessage->GetData_Int(ASURA_QOS_RESULT_MESSAGE_ID_SEND_BAND);
	const int iRecvBandKbps = pxMessage->GetData_Int(ASURA_QOS_RESULT_MESSAGE_ID_RECV_BAND);
	const int iAdjustedRTT  = pxMessage->GetData_Int(ASURA_QOS_RESULT_MESSAGE_ID_RTT);
	
	const Asura_Qos_Result* pxResult = AddResult(uTestIDNumber, uFlags, iSendBandKbps, iRecvBandKbps, iAdjustedRTT);
	PrintResult(pxConnection, pxResult, 0, 0);
}

/////////////////
// AnalyseData //
/////////////////
void Asura_Qos_Test::AnalyseData(Asura_Qos_NetConnection* pxConnection,
								 const Asura_Qos_Message* pxMessage,
								 bool bPreviousTest)
{
	const u_short usPacketSeqID = pxMessage->GetPacketSeqID();
	const u_int uTestIDNumber = (usPacketSeqID>>8);
//	const u_int uSubPacketIDNumber = (usPacketSeqID & 0xff);

	// Use the stored server ticks from the current or previous test.
	const Asura_TickCount* const pxServerTicks = GetTicks(bPreviousTest);

#ifdef ASURA_QOS_ENABLE_LOGGING
	if (dbg_bQosVerbose)
	{
		Asura_OutputToDebugger(("\n\nQoS: Received: result test=%u id=%u:\n", uTestIDNumber, (usPacketSeqID & 0xff)));
	}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

	// Time from after sending start packet to receiving result from client.
	const float fOverallRTT = Asura_Timers::TicksToSeconds(pxConnection->m_xLastMessageRecvTicks - pxServerTicks[0]);
	const int iOverallRTT = static_cast<int>(fOverallRTT*1.0e6f);

	const int iDelayBeforeFirstPacket = static_cast<int>(0.5f + 1.0e6f*fASURA_QOS_DELAY_BEFORE_LARGE_PACKETS); // microseconds

	// 15 microsecond fudge factor. Add a bit of leeway in the rejection tests to allow for timing errors etc.
	const int iMicrosecondFudgeFactor = 15;

	// Get flags from client.
	u_int uFlags = pxMessage->GetData_Int(0);
	if (bPreviousTest)
	{
		uFlags |= ASURA_QOS_RESULT_FLAG_LATE;
	}

	int aiSentMicro[uASURA_QOS_NUM_LARGE_PACKETS+1];
	int aiRecvMicro[uASURA_QOS_NUM_LARGE_PACKETS+1];

	u_int uNumValidLargePackets = 0; // This code deals with the case of losing some of the large packets too.

#ifdef ASURA_QOS_ENABLE_LOGGING
	if (dbg_bQosVerbose)
	{
		Asura_OutputToDebugger_Retail((" N:    sent    recv    delta0  delta1        b1\n"));
	}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

	for (u_int u=0; u<=uASURA_QOS_NUM_LARGE_PACKETS; ++u)
	{
		float fSentTime = Asura_Timers::TicksToSeconds(pxServerTicks[u] - pxServerTicks[0]);
		if (fSentTime < 0.0f) fSentTime = 0.0f;
		
		aiSentMicro[u] = static_cast<int>(fSentTime*1.0e6f);
		aiRecvMicro[u] = ( (u==0) ? 0 : pxMessage->GetData_Int(u) );

#ifdef ASURA_QOS_ENABLE_LOGGING
		if (dbg_bQosVerbose)
		{
			int iDelta0 = 0;
			if (u==1)
			{
				iDelta0 = aiSentMicro[u] - iDelayBeforeFirstPacket;
			}
			else if (u>1)
			{
				iDelta0 = aiSentMicro[u] - aiSentMicro[u-1];
			}

			int iDelta1 = 0;
			if (aiRecvMicro[u] == 0)
			{
			}
			else if (u==1)
			{
				iDelta1 = aiRecvMicro[u] - iDelayBeforeFirstPacket;
			}
			else if (u>1)
			{
				iDelta1 = aiRecvMicro[u] - aiRecvMicro[u-1];
			}

			int iB1 = 0; // kbps
			if (u>0)
			{
				const int iDiff1 = aiRecvMicro[u] - iDelayBeforeFirstPacket;
				if (iDiff1 > 0)
				{
					iB1 = static_cast<int>((8*u*uASURA_LARGE_PACKET_SIZE)/(iDiff1*1.0e-3f));
				}
			}

			Asura_OutputToDebugger_Retail(("%2u: %7d %7d   %7d %7d   %7d\n", u, aiSentMicro[u], aiRecvMicro[u], iDelta0, iDelta1, iB1));
		}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

		if (u == 1)
		{
			// If the delay between the start (packet 0) and packet 1 is too short then it must be due to queuing,
			// or the receiving thread being asleep. If so reject the data.
			// Allow a small fudge factor.
			if (aiRecvMicro[u] + iMicrosecondFudgeFactor < iDelayBeforeFirstPacket)
			{
				uFlags |= (ASURA_QOS_RESULT_FLAG_QUEUING | ASURA_QOS_RESULT_FLAG_REJECT);
			}
		}

		if (aiRecvMicro[u] != 0)
		{
			++uNumValidLargePackets;
		}
	}

	// Adjust the overall RTT by removing the time the client spent waiting for the large packets to arrive.
	// This should give the RTT time for a small packet that is dealt with immediately by the client.
	// Note: this doesn't work for the case of lost packets as the result is sent back much later.
	int iAdjustedRTT = iOverallRTT - aiRecvMicro[uASURA_QOS_NUM_LARGE_PACKETS];
	if (iAdjustedRTT < 0 || uNumValidLargePackets != uASURA_QOS_NUM_LARGE_PACKETS)
	{
		iAdjustedRTT = 0;
	}

	// Rough time from starting to send/receive packet 1 to finished sending/receiving packet 4 (in microseconds).
	const int iLastSent      = aiSentMicro[uASURA_QOS_NUM_LARGE_PACKETS] - iDelayBeforeFirstPacket;
	const int iLastSentValid = aiSentMicro[uNumValidLargePackets] - iDelayBeforeFirstPacket;
	const int iLastRecvValid = aiRecvMicro[uNumValidLargePackets] - iDelayBeforeFirstPacket;

	// Crude measure of overall sending/receiving bandwidth.
	const u_int uTotalNumBitsSent = 8*uASURA_QOS_NUM_LARGE_PACKETS*uASURA_LARGE_PACKET_SIZE;

	int iSendBandKbps = 0;
	if (iLastSent > 0)
	{
		iSendBandKbps = static_cast<int>((uTotalNumBitsSent)/(iLastSent*1.0e-3f));
	}

	int iRecvBandKbps = 0;
	if (iLastRecvValid > 0)
	{
		iRecvBandKbps = static_cast<int>((8*uNumValidLargePackets*uASURA_LARGE_PACKET_SIZE)/(iLastRecvValid*1.0e-3f));
	}

	// Sending seems to take slightly longer than receiving even for loopback (about 20 microseconds on my computer).
	// Include some extra fudge factors to lessen the strictness of the rejection test.
	// The overall receiving time delta should be less than the sending time delta; allow for some packet jitter though.
	const float fMaxTimeRatio = 1.05f;
	if (iLastSentValid > fMaxTimeRatio*(iLastRecvValid + iMicrosecondFudgeFactor))
	{
		uFlags |= (ASURA_QOS_RESULT_FLAG_REJECT | ASURA_QOS_RESULT_FLAG_TOOFAST);
	}

	// Strictly the receiving bandwidth should be no greater than the sending one.
	// Allow a bit of room for manoeuvre though.
	float fMaxBandRatio = 1.1f;
	const int iFastBandKbps     =  50000; //  50000 kbps
	const int iVeryFastBandKbps = 100000; // 100000 kbps
	if (iRecvBandKbps > iVeryFastBandKbps && iSendBandKbps > iVeryFastBandKbps)
	{
		// For very fast connections (LAN or loopback device) allow more flexibility.
		fMaxBandRatio = 1.35f;
	}
	else if (iRecvBandKbps > iFastBandKbps && iSendBandKbps > iFastBandKbps)
	{
		// For fast connections (LAN) allow more flexibility.
		fMaxBandRatio = 1.22f;
	}

	if (iRecvBandKbps > fMaxBandRatio*iSendBandKbps)
	{
		uFlags |= (ASURA_QOS_RESULT_FLAG_REJECT | ASURA_QOS_RESULT_FLAG_TOOFAST);
	}

	// On very fast LAN connections we can send all the data and receive the results back,
	// in less time than it would have taken the data to be downloaded on a 10Mbps connection.
	// This case is sometimes falsely rejected by the code that tests for queuing
	// (e.g. if the receiving thread goes to sleep at the wrong moment).
	if (uFlags & ASURA_QOS_RESULT_FLAG_REJECT)
	{
		const u_int uOtherFlags = (uFlags & ~(ASURA_QOS_RESULT_FLAG_TOOFAST | ASURA_QOS_RESULT_FLAG_QUEUING | ASURA_QOS_RESULT_FLAG_REJECT));
		if ( uOtherFlags == 0 && uNumValidLargePackets == uASURA_QOS_NUM_LARGE_PACKETS)
		{
			// Time in microseconds for a download of 2048 bytes at 10Mbps.
			const int iTimeForMaxClampBand = static_cast<int>((1000.0f * uTotalNumBitsSent)/iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC);

			if (iOverallRTT < iTimeForMaxClampBand &&
				iSendBandKbps >= iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC)
			{
				uFlags &= ~ASURA_QOS_RESULT_FLAG_REJECT;				// Clear the reject flag.
				iRecvBandKbps = iASURA_QOS_MAX_VALUE_KILOBITS_PER_SEC;	// Set recv bandwidth to maximum.
			}
		}
	}

	// Reject if we failed to calculate any of the bandwidths.
	if (iSendBandKbps == 0 || iRecvBandKbps == 0 || uNumValidLargePackets == 0)
	{
		uFlags |= ASURA_QOS_RESULT_FLAG_REJECT;
	}

#ifdef ASURA_QOS_ENABLE_LOGGING
	if (dbg_bQosVerbose)
	{
		Asura_OutputToDebugger_Retail(("\n"));
	}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

	const Asura_Qos_Result* pxResult = AddResult( uTestIDNumber, uFlags,
												  iSendBandKbps, iRecvBandKbps,
											      iAdjustedRTT );

	PrintResult(pxConnection, pxResult, aiSentMicro, aiRecvMicro);

	if (pxResult /*&& pxResult->IsValid()*/)
	{
		// Send results message back to the receiver.
		const u_int uNumDataVars =ASURA_QOS_RESULT_MESSAGE_ID_NUM_VARS;
		pxConnection->SetMessageData(uFlags,		ASURA_QOS_RESULT_MESSAGE_ID_FLAGS);		// 0
		pxConnection->SetMessageData(iSendBandKbps, ASURA_QOS_RESULT_MESSAGE_ID_SEND_BAND);	// 1
		pxConnection->SetMessageData(iRecvBandKbps, ASURA_QOS_RESULT_MESSAGE_ID_RECV_BAND);	// 2
		pxConnection->SetMessageData(iAdjustedRTT,	ASURA_QOS_RESULT_MESSAGE_ID_RTT);		// 3
		
		pxConnection->SendDataMessageTo(ASURA_QOS_MSGID_RESULT, usPacketSeqID, uNumDataVars);
	}
}

/////////////////
// PrintResult //
/////////////////
void Asura_Qos_Test::PrintResult(Asura_Qos_NetConnection* pxConnection,
								 const Asura_Qos_Result* pxResult,
								 const int* piSent, const int* piRecv)
{
#ifdef ASURA_QOS_ENABLE_LOGGING
	if (pxResult)
	{
		const char* szAddress = GetNetAddressString(pxConnection->GetConnectionInfo());
		
		const u_int uBUFFER_SIZE = 1024;
		char szBuffer[uBUFFER_SIZE];
		szBuffer[0] = 0;
		pxResult->PrintToBuffer(szBuffer, uBUFFER_SIZE);

		Asura_OutputToDebugger_Retail(("[%s]  %s", szAddress, szBuffer));
	
		if (piSent && piRecv)
		{
			// Write the raw data to the file too.
			Asura_SimpleLogfileOutput(( ASURA_QOS_LOGFILE_NAME, "[%s]  Sent: ", szAddress ));
			
			for (u_int u=0; u<=uASURA_QOS_NUM_LARGE_PACKETS; ++u)
			{
				Asura_SimpleLogfileOutput(( ASURA_QOS_LOGFILE_NAME, "%u ", piSent[u] ));
			}

			Asura_SimpleLogfileOutput(( ASURA_QOS_LOGFILE_NAME, "    Recv: " ));

			for (u_int u=0; u<=uASURA_QOS_NUM_LARGE_PACKETS; ++u)
			{
				Asura_SimpleLogfileOutput(( ASURA_QOS_LOGFILE_NAME, "%u ", piRecv[u] ));
			}
			Asura_SimpleLogfileOutput(( ASURA_QOS_LOGFILE_NAME, "\n" ));
		}

		// Write the results.
		Asura_SimpleLogfileOutput(( ASURA_QOS_LOGFILE_NAME, "[%s]  %s", szAddress, szBuffer));
	}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_NetConnection
///////////////////////////////////////////////////////////////////////////////

///////////
// Reset //
///////////
void Asura_Qos_NetConnection::Reset()
{
	m_xRemoteInfo.Reset();
	
	m_eState = ASURA_QOS_STATE_NULL;
	
	Asura_Mem::Set(&m_xLastMessageRecvTicks, 0, sizeof(Asura_TickCount));
	Asura_Mem::Set(&m_xLastMessageSentTicks, 0, sizeof(Asura_TickCount));

	m_xTest.Reset();
}

/////////////////////////
// SendPacketMessageTo //
/////////////////////////
// Send message to the remote machine corresponding to this connection.
// Send a specific length packet.
void Asura_Qos_NetConnection::SendPacketMessageTo( u_short usPacketSeqID,
												   u_int uDesiredPacketLength )
{
	Asura_Assert(m_pxListener, "Null listener");
	if (m_pxListener)
	{
		m_pxListener->SendPacketMessageTo(this, usPacketSeqID, uDesiredPacketLength);
	}
}

///////////////////////
// SendDataMessageTo //
///////////////////////
// Send a message with data.
void Asura_Qos_NetConnection::SendDataMessageTo( ASURA_QOS_MSGID eMsgID,
												 u_short usPacketSeqID,
												 u_int uNumDataVars )
{
	Asura_Assert(m_pxListener, "Null listener");
	if (m_pxListener)
	{
		m_pxListener->SendDataMessageTo(this, eMsgID, usPacketSeqID, uNumDataVars);
	}
}

////////////////////
// SetMessageData //
////////////////////
// Set the mesage data.
void Asura_Qos_NetConnection::SetMessageData(int iData, u_int uIndex) const
{
	if (m_pxListener)
	{
		m_pxListener->SetMessageData(iData, uIndex);
	}
}

///////////////////
// SendMessageTo //
///////////////////
// Send a message with no data.
void Asura_Qos_NetConnection::SendMessageTo(ASURA_QOS_MSGID eMsgID)
{
	Asura_Assert(m_pxListener, "Null listener");
	if (m_pxListener)
	{
		m_pxListener->SendMessageTo(this, eMsgID);
	}
}

///////////////////
// SendStartTest //
///////////////////
// Send a mesage to the client to start the test.
void Asura_Qos_NetConnection::SendStartTest()
{
	SendMessageTo(ASURA_QOS_MSGID_START_TEST);
	m_xTest.WaitForNextTest();
}

////////////
// Update //
////////////
// Send keep alive messages as necessary.
void Asura_Qos_NetConnection::Update(const Asura_TickCount& xNow)
{
	if (!IsConnected()) return;

	const float fTimeSinceLastRecv = Asura_Timers::TicksToSeconds(xNow - m_xLastMessageRecvTicks);
	const float fTimeSinceLastSend = Asura_Timers::TicksToSeconds(xNow - m_xLastMessageSentTicks);

	if (fTimeSinceLastRecv > fASURA_QOS_DISCONNECT_TIMEOUT)
	{
		// Haven't heard from the remote machine for a long time, drop the connection.
#ifdef ASURA_QOS_ENABLE_LOGGING
		Asura_OutputToDebugger_Retail(("QoS: Lost connection to %s.\n", m_pxListener->IsHost() ? "client" : "server"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

		SendMessageTo(ASURA_QOS_MSGID_MSG_CONNECTION_CLOSE);
		CloseConnection();
		return;
	}
	else if (fTimeSinceLastSend > fASURA_QOS_KEEP_ALIVE_TIME)
	{
		// Haven't sent anything to the remote machine for a while (>1 second), send a keep-alive message.
		SendMessageTo(ASURA_QOS_MSGID_MSG_CONNECTION_KEEPALIVE);
		//Asura_OutputToDebugger(("QoS: Sent keep alive message.\n"));
	}

	const bool bConnectionProblems = (fTimeSinceLastRecv > fASURA_QOS_CONNECTION_PROBLEMS_TIME);
	if (bConnectionProblems)
	{
		m_eState = ASURA_QOS_STATE_CONNECTED_PROBLEMS;
	}
	else
	{
		m_eState = ASURA_QOS_STATE_CONNECTED;
	}
}

/////////////////////
// CloseConnection //
/////////////////////
void Asura_Qos_NetConnection::CloseConnection()
{
	SetClosed();
	
	// Summarize the results if we haven't already.
	m_xTest.SummarizeResults(this);

#ifdef ASURA_QOS_ENABLE_LOGGING
	Asura_OutputToDebugger_Retail(("QoS: Closed connection to %s\n", GetNetAddressString(GetConnectionInfo())));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
}

///////////////////////////
// SendRecvTimingMessage //
///////////////////////////
void Asura_Qos_NetConnection::SendRecvTimingMessage(bool bZeroResults, u_short usPacketSeqID)
{
	SetMessageData(GetReceivedFlags(), 0);

	const Asura_TickCount* const pxTicks = GetTicks(false);

	// Set the times.
	for (u_int u=0; u<uASURA_QOS_NUM_LARGE_PACKETS; ++u)
	{
		if (bZeroResults)
		{
			SetMessageData(0, u+1);
		}
		else
		{
			const float fDeltaTime = Asura_Timers::TicksToSeconds(pxTicks[u+1] - pxTicks[0]);
			int iDeltaRecvTimeMicroseconds = static_cast<int>(fDeltaTime*1.0e6f);
			if (iDeltaRecvTimeMicroseconds < 0)
			{
				iDeltaRecvTimeMicroseconds = 0;
			}

			SetMessageData(iDeltaRecvTimeMicroseconds, u+1);
		}
	}

	SendDataMessageTo(ASURA_QOS_MSGID_RECV_TIMING, usPacketSeqID, uASURA_QOS_NUM_LARGE_PACKETS+1);
}

//////////////////////////////////
// SendRecvTimingMessagePartial //
//////////////////////////////////
void Asura_Qos_NetConnection::SendRecvTimingMessagePartial()
{
	m_xTest.SetLatePacketsFlags();
	SendRecvTimingMessage(false, m_xTest.GetPacketSeqID());
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_MessageHeader
///////////////////////////////////////////////////////////////////////////////

Asura_Qos_MessageHeader::Asura_Qos_MessageHeader() 
: m_usNBOMsgID(0)
, m_usNBOPacketSeqID(0)
{
	m_uNBOPassword = htonl(MAGIC_PASSWORD_NUMBER); // Convert to NBO
}

void Asura_Qos_MessageHeader::Set(ASURA_QOS_MSGID eMsgID, u_short usPacketSeqID)
{
	m_usNBOMsgID = htons(static_cast<u_short>(eMsgID));
	m_usNBOPacketSeqID = htons(usPacketSeqID); // Convert to NBO
}

bool Asura_Qos_MessageHeader::IsPasswordValid() const
{
	return (ntohl(m_uNBOPassword) == MAGIC_PASSWORD_NUMBER); // Convert to NBO
}

///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_Message
///////////////////////////////////////////////////////////////////////////////

Asura_Qos_Message::Asura_Qos_Message()
{
}

// Fill with dummy data.
void Asura_Qos_Message::FillData()
{
	// Fill the dummy message data with something.
	// Doesn't need to be particuarly random.
	for (u_int u = 0; u<uASURA_QOS_MAX_NUM_DATA_VARS; ++u)
	{
		m_aiData[u] = static_cast<int>(u);
	}
}

////////////////////
// GetPacketSeqID //
////////////////////
u_short Asura_Qos_Message::GetPacketSeqID() const
{
	const u_short usPacketSeqID = ntohs(m_xHeader.m_usNBOPacketSeqID); // Convert to HBO
	return usPacketSeqID;
}

//////////////
// GetMsgID //
//////////////
u_int Asura_Qos_Message::GetMsgID() const
{
	const u_int uMsgID = ntohs(m_xHeader.m_usNBOMsgID);  // Convert to HBO
	return uMsgID;
}

/////////////
// SetData //
/////////////
// Set message data and return the required packet length.
void Asura_Qos_Message::SetData(int iData, u_int uIndex)
{
	Asura_Assert(uIndex < uASURA_QOS_MAX_NUM_DATA_VARS, "Invalid array access!");
	m_aiData[uIndex] = static_cast<int>(htonl(static_cast<u_long>(iData))); // Convert to NBO
}

/////////////////
// GetData_Int //
/////////////////
// Get integer data.
int Asura_Qos_Message::GetData_Int(u_int uIndex) const
{
	Asura_Assert(uIndex < uASURA_QOS_MAX_NUM_DATA_VARS, "Invalid array access!");
	return static_cast<int>(ntohl(static_cast<u_long>(m_aiData[uIndex]))); // Convert to HBO
}


///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_Listener
///////////////////////////////////////////////////////////////////////////////

Asura_Qos_Listener::Asura_Qos_Listener(u_int uMaxNumConnections, u_short usLocalPort, bool bIsHost)
: m_uMaxNumConnections(uMaxNumConnections)
, m_usLocalPort(usLocalPort)
, m_bDone(false)
, m_bIsHost(bIsHost)
, m_pxConnection(0)
, m_pxSocket(0)
{
	Asura_Assert(m_uMaxNumConnections > 0, "Invalid number of connections");
	m_pxConnection = new Asura_Qos_NetConnection[m_uMaxNumConnections];
	m_pxSocket = Asura_Qos_System::Platform_CreateSocket(m_bIsHost);

	for (u_int u=0; u<m_uMaxNumConnections; ++u)
	{
		m_pxConnection[u].SetListener(this);
	}

	// Fill with message dummy data.
	m_xMessage.FillData();
}

Asura_Qos_Listener::~Asura_Qos_Listener()
{
	// Should have already closed the socket, but make sure.
	CloseSocket();

	delete m_pxSocket;
	m_pxSocket = 0;

	delete [] m_pxConnection;
	m_pxConnection = 0;
}

//////////////////
// CreateSocket //
//////////////////
// Return true if OK.
bool Asura_Qos_Listener::CreateSocket()
{
	if (m_pxSocket)
	{
		m_pxSocket->CreateSocket(this, m_usLocalPort);
		return m_pxSocket->HasCreated();
	}
	return false;
}

/////////////////
// CloseSocket //
/////////////////
void Asura_Qos_Listener::CloseSocket()
{
	if (m_pxSocket)
	{
		m_pxSocket->CloseSocket();
	}
}

//////////////////////
// CloseConnections //
//////////////////////
// Close all the connections.
void Asura_Qos_Listener::CloseConnections()
{
	for (u_int u=0; u<m_uMaxNumConnections; ++u)
	{
		if (m_pxConnection[u].IsConnected() && !m_pxConnection[u].HasConnectionProblems())
		{
			m_pxConnection[u].SendMessageTo(ASURA_QOS_MSGID_MSG_CONNECTION_CLOSE);
			m_pxConnection[u].CloseConnection();
		}
	}
}

////////////////////
// FindConnection //
////////////////////
// Parameters in host byte order. Find the connection.
// Return 0 if not found.
Asura_Qos_NetConnection* Asura_Qos_Listener::FindConnection(const Asura_Qos_NetConnectionInfo& xRemoteInfo)
{
	for (u_int u=0; u<m_uMaxNumConnections; ++u)
	{
		if (m_pxConnection[u].GetConnectionInfo() == xRemoteInfo)
		{
			return &m_pxConnection[u];
		}
	}

	return 0;
}

///////////////////
// AddConnection //
///////////////////
// Parameters in host byte order. Add new connection if existing one not found.
// Return 0 if couldn't add (i.e. hit maximum number of connections).
Asura_Qos_NetConnection* Asura_Qos_Listener::AddConnection(const Asura_Qos_NetConnectionInfo& xRemoteInfo)
{
	Asura_Qos_NetConnection* pxFound = FindConnection(xRemoteInfo);
	if (pxFound)
	{
		return pxFound;
	}

	for (u_int u=0; u<m_uMaxNumConnections; ++u)
	{
		if (!m_pxConnection[u].IsConnected())
		{
			m_pxConnection[u].Reset();
			m_pxConnection[u].SetConnectionInfo(xRemoteInfo);

		#ifdef ASURA_QOS_ENABLE_LOGGING
			#ifdef ASURA_PC_SUPPORT_STEAMWORKS
			Asura_OutputToDebugger(("QoS: Added connection to %s (#%u).\n", GetNetAddressString(xRemoteInfo), xRemoteInfo.m_xSocket));
			#else
			Asura_OutputToDebugger(("QoS: Added connection to %s.\n", GetNetAddressString(xRemoteInfo)));
			#endif
		#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

			return &m_pxConnection[u];
		}
	}

#ifdef ASURA_QOS_ENABLE_LOGGING
	Asura_Assertf(0, ("Too many connections. Ignoring message from %s.", GetNetAddressString(xRemoteInfo)));
#endif	//ASURA_QOS_ENABLE_LOGGING

	return 0;
}

/////////////////////////
// SendPacketMessageTo //
/////////////////////////
// Send a specific length packet.
void Asura_Qos_Listener::SendPacketMessageTo(Asura_Qos_NetConnection* pxDestination,
											 u_short usPacketSeqID,
											 u_int uDesiredPacketLength)
{
	if (uDesiredPacketLength > uASURA_QOS_MAX_PACKET_LENGTH)
	{
		uDesiredPacketLength = uASURA_QOS_MAX_PACKET_LENGTH;
	}

	int iDataLength = static_cast<int>(uDesiredPacketLength - uASURA_QOS_UDPIP_HEADER_LENGTH);
	if (iDataLength < iASURA_QOS_DATA_HEADER_LENGTH)
	{
		iDataLength = iASURA_QOS_DATA_HEADER_LENGTH;
	}

	m_xMessage.m_xHeader.Set(ASURA_QOS_MSGID_PACKET, usPacketSeqID);
	SocketSendTo(pxDestination, iDataLength);
}

///////////////////////
// SendDataMessageTo //
///////////////////////
// Send a message with data.
void Asura_Qos_Listener::SendDataMessageTo( Asura_Qos_NetConnection* pxDestination,
											ASURA_QOS_MSGID eMsgID,
											u_short usPacketSeqID,
											u_int uNumDataVars )
{
	const int iDataLength = static_cast<int>(uASURA_QOS_DATA_HEADER_LENGTH + sizeof(int)*uNumDataVars);

	m_xMessage.m_xHeader.Set(eMsgID, usPacketSeqID);
	SocketSendTo(pxDestination, iDataLength);
}

/////////////////
// SendMessage //
/////////////////
// Send message without any data, just the header and msg id.
void Asura_Qos_Listener::SendMessageTo(Asura_Qos_NetConnection* pxDestination, ASURA_QOS_MSGID eMsgID)
{
	m_xMessage.m_xHeader.Set(eMsgID, 0);
	SocketSendTo(pxDestination, iASURA_QOS_DATA_HEADER_LENGTH);
}

//////////////////
// SocketSendTo //
//////////////////
void Asura_Qos_Listener::SocketSendTo( Asura_Qos_NetConnection* pxDestination, int iDataLength)
{
	if (!pxDestination) return;

#ifdef ASURA_QOS_TEST_LOST_MESSAGES
	// Test how the code responds to lost messages.
	const float fRandom = Asura_Random::GetFloat();
	if (fRandom < fASURA_QOS_FRACTION_MESSAGES_TO_DROP)
	{
		// Drop the message.
		pxDestination->m_xLastMessageSentTicks = Asura_Timers::GetCurTicks();

		// Note: These lines are for debugging only, they will significantly affect the test results.
		#ifdef ASURA_QOS_ENABLE_LOG_ALL_MESSAGES
		Asura_OutputToDebugger(("QoS: Drop message id=%d\n", m_xMessage.GetMsgID()));
		#endif //ASURA_QOS_ENABLE_LOG_ALL_MESSAGES

		return;
	}
#endif //ASURA_QOS_TEST_LOST_MESSAGES

	const bool bOk = (m_pxSocket && m_pxSocket->SendTo(pxDestination, reinterpret_cast<const char*>(&m_xMessage), iDataLength));
	Asura_Assert(bOk, "Failed to send message");

	if ( !bOk )
	{
		// Failed to send the data, quit.
		m_bDone = true;
		return;
	}

	pxDestination->m_xLastMessageSentTicks = Asura_Timers::GetCurTicks();

	// Note: These lines are for debugging only, they will significantly affect the test results.
	#ifdef ASURA_QOS_ENABLE_LOG_ALL_MESSAGES
	Asura_OutputToDebugger(("QoS: Sent message id=%d\n", m_xMessage.GetMsgID()));
	#endif //ASURA_QOS_ENABLE_LOG_ALL_MESSAGES
}


/////////////////////
// ReceiveMessages //
/////////////////////
void Asura_Qos_Listener::ReceiveMessages(u_int uWaitTimeSeconds /*= 0*/)
{
	if (m_pxSocket && m_pxSocket->HasCreated())
	{
		const bool bOk = m_pxSocket->Receive(this, uWaitTimeSeconds);
		if (!bOk && !IsHost())
		{
			// Client just quits if there is any error.
			m_bDone = true;	
		}
	}
}


///////////////////////
// HandleMessageFrom //
///////////////////////
void Asura_Qos_Listener::HandleMessageFrom(const Asura_Qos_NetConnectionInfo& xRemoteInfo, const Asura_Qos_Message* pxMessage)
{
	Asura_Qos_NetConnection* pxConnection = AddConnection(xRemoteInfo);
	if (pxConnection)
	{
		pxConnection->m_xLastMessageRecvTicks = Asura_Timers::GetCurTicks();
		HandleMessage(pxConnection, pxMessage);
	}
	else if (IsHost())
	{
		// No more free connections, send a reject message.
		Asura_Qos_NetConnection xTempConnection;
		xTempConnection.SetConnectionInfo(xRemoteInfo);
		xTempConnection.SetListener(this);
		SendMessageTo(&xTempConnection, ASURA_QOS_MSGID_MSG_CONNECTION_REJECT);
	}
}


///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_Listener_Server
///////////////////////////////////////////////////////////////////////////////

Asura_Qos_Listener_Server::Asura_Qos_Listener_Server(u_short usLocalPort)
: Asura_Qos_Listener(uASURA_QOS_MAX_NUM_CONNECTIONS, usLocalPort, true)
, m_fTimeToNextTest(fASURA_QOS_DELAY_BEFORE_TEST)
, m_fTimeOfCurrentTest(0)
, m_uLastConnectionTested(0)
{
}

/////////////////////
// UpdateTestTimer //
/////////////////////
void Asura_Qos_Listener_Server::UpdateTestTimer(bool bTesting)
{
	if (bTesting)
	{
		m_fTimeOfCurrentTest += Asura_Qos_System::GetTimeSlice();
	}
	else
	{
		// Not testing. 
		m_fTimeOfCurrentTest = 0.0f;

		// If the system is asleep (i.e. the game is playing) then suspend testing.
		if (!Asura_Qos_System::IsAsleep())
		{
			m_fTimeToNextTest -= Asura_Qos_System::GetTimeSlice();

			if (m_fTimeToNextTest < 0.0f)
			{
				m_fTimeToNextTest = 0.0f;
			}
		}
	}
}

////////////
// Update //
////////////
void Asura_Qos_Listener_Server::Update()
{
	// Check for received messages.
	ReceiveMessages();

	bool bTesting = false;
	const Asura_TickCount xNow = Asura_Timers::GetCurTicks();

	u_int uNumConnected = 0;

	// Loop over connections doing stuff.
	for (u_int u=0; u<m_uMaxNumConnections; ++u)
	{
		// Make sure that the last connection tested is done at the end, so that we will
		// cycle through the connections testing them in turn.
		const u_int uIndex = ((u+m_uLastConnectionTested+1) % m_uMaxNumConnections);

		Asura_Qos_NetConnection& xConnection = m_pxConnection[uIndex];
		
		if (xConnection.IsConnected())
		{
			// Keep connection alive, by sending a keep-alive message where necessary.
			xConnection.Update(xNow);

			// Update the testing.
			UpdateActiveConnection(xConnection, uIndex, bTesting);

			++uNumConnected;
		}
	}

	// Update the test delay timer if no test is in progress,
	// otherwise update the current test timer.
	UpdateTestTimer(bTesting);

	// If nobody is connected, we may as well just sleep for a while.
	const bool bIdle = (uNumConnected == 0);
	Asura_Qos_System::SetIdle(bIdle);
}

////////////////////////////
// UpdateActiveConnection //
////////////////////////////
void Asura_Qos_Listener_Server::UpdateActiveConnection(Asura_Qos_NetConnection& xConnection,
													   u_int uIndex,
													   bool& bTesting)
{
	xConnection.UpdateDelayTimer();
	xConnection.UpdateTest();

	if (xConnection.HasConnectionProblems() || xConnection.HasFinishedTests())
	{
		// Either has connection problems (stop any test in-progress)
		// or has finished testing (close the connection).
		xConnection.HandlePostTestEnd();
		return;
	}

	if (!xConnection.CanStartTest())
	{
		// A test is in progress.
		bTesting = true;

		if (m_fTimeOfCurrentTest > fASURA_QOS_TEST_TIMEOUT_TIME)
		{
			// Test is taking longer than expected to complete, probably due to lost messages.
			bool bAbort = true;

			if (xConnection.IsDownloadFromRemote())
			{
				if ( xConnection.GetReceivedSubIndex() > ((1+uASURA_QOS_NUM_LARGE_PACKETS)/2) )
				{
					// We have received enough packets for it to be worth sending back partial timing results.
					bAbort = false;
				}
			}

			#ifdef ASURA_QOS_ENABLE_LOGGING
			Asura_OutputToDebugger(("QoS: %s test %u due to timeout.\n", (bAbort ? "Aborting": "Ending"), xConnection.GetTestID() ));
			#endif //ASURA_QOS_ENABLE_LOGGING

			if (!bAbort)
			{
				// Received some packets, it'll have to be enough. Bring the test to an end.
				xConnection.SendRecvTimingMessagePartial();
			}


			// End or abort the test.
			xConnection.HandlePostTestEnd();
		}

		return;
	}

	if (xConnection.HasFinishedTests())
	{
		// Finished testing so close the connection.
		xConnection.HandlePostTestEnd();
	}
	else if (m_fTimeToNextTest == 0.0f)
	{
		m_fTimeOfCurrentTest = 0.0f;

		// Normally we do the tests as quickly as possible,
		// but for debugging purposes it is convenient to slow them down.
		m_fTimeToNextTest = (dbg_bQosQuick ? fASURA_QOS_DELAY_BETWEEN_TESTS_QUICK : fASURA_QOS_DELAY_BETWEEN_TESTS_SLOW);

		m_uLastConnectionTested = uIndex;

		if (xConnection.IsDownloadFromRemote())
		{
			xConnection.SendStartTest();
		}
		else
		{
			xConnection.HandleStartTest();
		}
	}
}

///////////////////
// HandleMessage //
///////////////////
void Asura_Qos_Listener_Server::HandleMessage(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage)
{
	const u_int uMsgID = pxMessage->GetMsgID();
	const u_short usPacketSeqID = pxMessage->GetPacketSeqID();

	switch (uMsgID)
	{
	default:
		Asura_Assertf(0, ("Unexpected message ID %u", uMsgID));
		break;

	case ASURA_QOS_MSGID_MSG_CONNECTION_REQUEST:
		{
#ifdef ASURA_QOS_ENABLE_LOGGING
			Asura_OutputToDebugger_Retail(("QoS: Received connection request from %s (accepted).\n", GetNetAddressString(pxConnection->GetConnectionInfo())));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

			if (!pxConnection->IsConnected())
			{
				pxConnection->SetConnected();
			}
			SendMessageTo(pxConnection, ASURA_QOS_MSGID_MSG_CONNECTION_ACCEPT);
		}
		break;

	case ASURA_QOS_MSGID_MSG_CONNECTION_CLOSE:
		{
			if (pxConnection->IsConnected())
			{
				pxConnection->CloseConnection();
			}
			break;
		}

	case ASURA_QOS_MSGID_MSG_CONNECTION_KEEPALIVE:
		{
			// Keep alive message, the timer (m_xLastMessageRecvTicks)
			// has already been updated so there is nothing further to do.
			//Asura_OutputToDebugger(("QoS: Received keep alive message.\n"));
			break;
		}
	
	case ASURA_QOS_MSGID_RECV_TIMING:
		{
			if (pxConnection->IsConnected() && !pxConnection->IsTestStateReceiving())
			{
				pxConnection->HandleRecvTiming(pxMessage);
			}
			break;
		}

	case ASURA_QOS_MSGID_PACKET:
		{
			if (pxConnection->IsConnected() && pxConnection->IsTestStateReceiving())
			{
				pxConnection->HandlePacket(usPacketSeqID);
			}
		}
		break;

	case ASURA_QOS_MSGID_RESULT:
		{
			if (pxConnection->IsConnected() /*&& pxConnection->IsTestStateReceiving()*/)
			{
				pxConnection->HandleResult(pxMessage);
			}
		}
		break;
	}
}


///////////////////////////////////////////////////////////////////////////////
// Asura_Qos_Listener_Client
///////////////////////////////////////////////////////////////////////////////

void Asura_Qos_Listener_Client::RequestConnect()
{
	// Send connection message to server. Wait for response.
#ifdef ASURA_QOS_ENABLE_LOGGING
	Asura_OutputToDebugger_Retail(("QoS: Sending connection request...\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
	
	// Note: a short delay to allow Steamworks connection sockets time to finish initialising.
	const int iNumWaitAttempts = 8;
	int iWaitCount = 0;
	do
	{
		Asura_Thread_System::DelayCurrentThread(0.25f);
		++iWaitCount;
	}
	while (!IsSocketReady() && iWaitCount<iNumWaitAttempts);


	// For clients the zeroth connection contains the server.
	const int iNumConnectAttempts = 15;
	int iCount = 0;
	while (!m_pxConnection[0].IsConnected() && iCount<iNumConnectAttempts && !m_bDone)
	{
		// Send connection message, and wait for 1 second for acknowledgement.
		SendMessageTo(&m_pxConnection[0], ASURA_QOS_MSGID_MSG_CONNECTION_REQUEST);
		ReceiveMessages(1);
		++iCount;
	}

	if (!m_pxConnection[0].IsConnected())
	{
#ifdef ASURA_QOS_ENABLE_LOGGING
		Asura_OutputToDebugger_Retail(("QoS: Failed to connect to server.\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

		m_bDone = true; // Failed to connect.
	}
}

void Asura_Qos_Listener_Client::Update()
{
	// Check for received messages.
	ReceiveMessages();

	const Asura_TickCount xNow = Asura_Timers::GetCurTicks();

	for (u_int u=0; u<m_uMaxNumConnections; ++u)
	{
		Asura_Qos_NetConnection& xConnection = m_pxConnection[u];

		if (m_pxConnection[u].IsConnected())
		{
			// Keep connection alive, by sending a keep-alive message where necessary.
			xConnection.Update(xNow);

			xConnection.UpdateDelayTimer();
			xConnection.UpdateTest();
		}
		else
		{
			// Lost connection to server. Stop the listener thread.
			m_bDone = true;
		}
	}
}

void Asura_Qos_Listener_Client::HandleMessage(Asura_Qos_NetConnection* pxConnection, const Asura_Qos_Message* pxMessage)
{
	if (!pxConnection || !pxMessage)
	{
		Asura_Assert(0, "Null pointer");
		return;
	}

	const u_int uMsgID = pxMessage->GetMsgID();
	const u_short usPacketSeqID = pxMessage->GetPacketSeqID();

	switch (uMsgID)
	{
	default:
		Asura_Assertf(0, ("Unexpected message ID %u", uMsgID));
		break;

	case ASURA_QOS_MSGID_MSG_CONNECTION_ACCEPT:
		if (!pxConnection->IsConnected())
		{
			pxConnection->SetConnected();

#ifdef ASURA_QOS_ENABLE_LOGGING
			Asura_OutputToDebugger_Retail(("QoS: Received connection accept\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
		}
		break;

	case ASURA_QOS_MSGID_MSG_CONNECTION_REJECT:
		{
			// Give up.
#ifdef ASURA_QOS_ENABLE_LOGGING
			Asura_OutputToDebugger_Retail(("QoS: Received connection reject\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

			pxConnection->SetClosed();	
			m_bDone = true;
		}
		break;

	case ASURA_QOS_MSGID_MSG_CONNECTION_CLOSE:
		{
			// Finished.
#ifdef ASURA_QOS_ENABLE_LOGGING
			Asura_OutputToDebugger_Retail(("QoS: Received connection close\n"));
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING

			pxConnection->CloseConnection();
			m_bDone = true;
		}
		break;

	case ASURA_QOS_MSGID_MSG_CONNECTION_KEEPALIVE:
		{
			// Keep alive message, the timer (m_xLastMessageRecvTicks)
			// has already been updated so there is nothing further to do.
			//Asura_OutputToDebugger(("QoS: Received keep alive message.\n"));
		}
		break;

	case ASURA_QOS_MSGID_PACKET:
		{
			if (pxConnection->IsConnected() && !pxConnection->IsTestStateSending())
			{
				pxConnection->HandlePacket(usPacketSeqID);
			}
		}
		break;

	case ASURA_QOS_MSGID_RESULT:
		{
			if (pxConnection->IsConnected())
			{
				pxConnection->HandleResult(pxMessage);
			}
		}
		break;

	case ASURA_QOS_MSGID_START_TEST:
		{
			if (pxConnection->IsConnected() && pxConnection->IsTestStateNull())
			{
				pxConnection->HandleStartTest();
			}
		}
		break;

	case ASURA_QOS_MSGID_RECV_TIMING:
		{
			if (pxConnection->IsConnected() && !pxConnection->IsTestStateReceiving())
			{
				pxConnection->HandleRecvTiming(pxMessage);
			}
		}
		break;
	}
}

#endif //#ifdef ASURA_ENABLE_QOS_SYSTEM
