//
// Asura_TCPSocket.h
//
// Created by KOR - 10 February 2009
//
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_TCPSOCKET_H__
#define __ASURA_TCPSOCKET_H__

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

// Asura
#include "Core/Asura_Mempool.h"
#include "Core/Asura_RingBuffer.h"
#include "Network/Asura_NetSocket.h"
#include "Network/Asura_ReliableSocket.h"
#include "Network/Asura_TCPSocket_Types.h"

// forward declarations
class Asura_NetPort;

// constants
#ifdef _DEBUG
#define ASURA_TCP_LOGGING_ENABLED 1
#endif

const u_int uASURA_TCP_SOCKET_MAX_CONNECTIONS = 32;

// enums
enum ASURA_TCP_MESSAGE_ID
{
	ASURA_TCP_MESSAGE_FIRST_RESERVED_ID = 0xf000,
	ASURA_TCP_MESSAGE_CONNECT_REQUEST = ASURA_TCP_MESSAGE_FIRST_RESERVED_ID,
	ASURA_TCP_MESSAGE_CONNECT_CONFIRM,
	ASURA_TCP_MESSAGE_DISCONNECT,
};

enum ASURA_TCP_LOGGING_LEVEL
{
	ASURA_TCP_LOG_NOTHING,
	ASURA_TCP_LOG_ERRORS,
	ASURA_TCP_LOG_EVENTS,
	ASURA_TCP_LOG_EVERYTHING,
};

// typedefs
typedef void (*Asura_TCP_Message_Callback)(Asura_ReliableConnectionID xConnectionID, u_short usMsgID, u_short usMsgBodySize, const char* pcMsgBody, void* pUserData);
typedef bool (*Asura_TCP_Connection_Callback)(Asura_ReliableConnectionID xConnectionID, u_short usMsgSize, const char* pcMsgBody, void* pUserData);
typedef void (*Asura_TCP_Disconnection_Callback)(Asura_ReliableConnectionID xConnectionID, ASURA_TCP_DISCONNECT_REASON eReason, void* pUserData);

// types
struct Asura_TCP_MessageHeader
{
	u_short m_usMessageID;
	u_short m_usDataSize;
	u_int m_uCheckSum;
};

class Asura_TCPSocket
{
public:
	// static functions
#ifdef _DEBUG
	static void UnitTest();
#endif
	static ASURA_TCP_LOGGING_LEVEL GetLoggingLevel() { return s_eLoggingLevel; }
	static void SetLoggingLevel(ASURA_TCP_LOGGING_LEVEL eLevel) { s_eLoggingLevel = eLevel; }

	// public member functions
	SOCKET GetSocket() const { return m_xSocketHandle; }

protected:
	// We never create a raw socket - must be a listener or connection
	Asura_TCPSocket();
	virtual ~Asura_TCPSocket();

	bool CreateSocket();
	bool CloseSocket();

#ifdef _DEBUG
	const char* GetMsgIDDesc(u_int uMsgID);
#endif

	SOCKET m_xSocketHandle;

	// static data
	static ASURA_TCP_LOGGING_LEVEL s_eLoggingLevel;
};

class Asura_TCPSocket_Connection : public Asura_TCPSocket
{
public:
	enum CONNECTION_STATE
	{
		STATE_DISCONNECTED,
		STATE_CONNECTING,
		STATE_REQUESTING,
		STATE_AWAITING_REQUEST,
		STATE_CONNECTED,
		STATE_DISCONNECTING,
	};
	Asura_TCPSocket_Connection(Asura_TCP_Message_Callback pfnCallback, 
							   Asura_TCP_Connection_Callback pfnConnectionCallback, 
							   Asura_TCP_Disconnection_Callback pfnDisconnectionCallback);
	virtual ~Asura_TCPSocket_Connection();

	// user data is returned to each callback function to allow another object to store its
	// 'this' pointer when it owns an instance of this class. Since you can't use the 'this' 
	// pointer in a contructor initialiser list (and it generates a compiler warning) the
	// user data is set with this function.
	void SetCallbackUserData( void* pUserData ) { m_pCallbackUserData = pUserData; }

	// Control functions
	// Server accepts from a listener
	bool Accept(class Asura_TCPSocket_Listener& xListenerSocket, Asura_ReliableConnectionID xID);
	// Client calls connect with server address.
	bool Connect(u_int uAddress, const Asura_NetPort& xPort, u_short usConnectionDataSize, const void* pConnectionData);
	void Disconnect(bool bImmediate, ASURA_TCP_DISCONNECT_REASON eReason);
	void Update();
	bool SendMessage(u_short usMsgID, u_short usMsgBodySize, const void* pMsgBody);

	// Accessors
	bool IsActive() const { return STATE_DISCONNECTED != m_eState; }
	bool IsConnected() const { return STATE_CONNECTED == m_eState; }
	Asura_ReliableConnectionID GetID() const { return m_xID; }
	int GetIDAsInt() const { return m_xID.AsInt(); }

protected:
	bool Send(const void* pData, u_int uBytes);
	void UpdateIncomingData();
	void HandleMessages();
	void HandleMessage(u_short usMsgID, u_short usMsgBodySize, const char* pcMsgBody);
	u_int GetIncomingNumBytes() const { return m_xIncomingDataBuffer.GetCount(); }
	bool GetIncomingData(char* pcBuffer, u_int uNumBytes);

	Asura_ReliableConnectionID m_xID;
	CONNECTION_STATE m_eState;
	Asura_RingBuffer<char> m_xIncomingDataBuffer;
	Asura_TCP_MessageHeader m_xCurrentMessageHeader;
	bool m_bWaitingForMessageBody;
	u_short m_usConnectionDataSize;
	void* m_pConnectionData;
	Asura_TCP_Message_Callback m_pfnMessageCallback;
	Asura_TCP_Connection_Callback m_pfnConnectionCallback;
	Asura_TCP_Disconnection_Callback m_pfnDisconnectionCallback;
	void* m_pCallbackUserData;
};

class Asura_TCPSocket_Listener : public Asura_TCPSocket
{
public:
	Asura_TCPSocket_Listener(Asura_TCP_Message_Callback pfnMessageCallback, 
							 Asura_TCP_Connection_Callback pfnConnectionCallback, 
							 Asura_TCP_Disconnection_Callback pfnDisconnectionCallback);
	virtual ~Asura_TCPSocket_Listener();
	// user data is returned to each callback function to allow another object to store its
	// 'this' pointer when it owns an instance of this class. Since you can't use the 'this' 
	// pointer in a contructor initialiser list (and it generates a compiler warning) the
	// user data is set with this function.
	void SetCallbackUserData( void* pUserData );

	// Control functions
	void Close();
	bool Listen(const Asura_NetPort& xPort);
	bool Listen(u_int uAddress, const Asura_NetPort& xPort);
	// Sends to all connections
	bool SendToAll(u_short usMsgID, u_short usMsgBodySize, const void* pMsgBody);
	bool SendToAllExcept(u_short usMsgID, u_short usMsgBodySize, const void* pMsgBody, Asura_ReliableConnectionID xConnectionID);
	// Sends to a specified connection
	bool SendMessage(Asura_ReliableConnectionID xConnectionID, u_short usMsgID, u_short usMsgBodySize, const void* pMsgBody);
	void Update();
	void Disconnect(Asura_ReliableConnectionID xConnectionID, ASURA_TCP_DISCONNECT_REASON eReason);
	u_short GetListenPort() const;

	// Accessors
	u_int GetNumConnections() const { return m_uNumConnections; }
	bool IsActive() const { return m_bActive; }

protected:
	Asura_TCPSocket_Connection* GetConnection(Asura_ReliableConnectionID xID);

	Asura_TCPSocket_Connection* m_apxConnections[uASURA_TCP_SOCKET_MAX_CONNECTIONS];
	u_int m_uNumConnections;
	bool m_bActive;

	static Asura_ReliableConnectionID s_xNextConnectionID;
};

#endif // #ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

#endif//__ASURA_TCPSOCKET_H__
