/**
*
* \file Asura_GameSpy_HTTP.cpp
*
* Created by Richard May - 09/12/2003 12:17:37
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/



/************/
/* Includes.*/
/************/

// main asura header
#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_GAMESPY

// local header
#include "Network/GameSpy/Asura_GameSpy_HTTP.h"

// asura
#include "Debugging/Asura_Assert.h"

// middleware
#include "GameSpy/nonport.h"	// for gsifree


/********/
/* Data */
/********/

bool Asura_GameSpy_HTTP::s_bInitialised = false;

// data to deal with current file get.
bool Asura_GameSpy_HTTP::s_bGetInProgress = false;
float Asura_GameSpy_HTTP::s_fDownloadProgress;
GHTTPResult Asura_GameSpy_HTTP::s_eResult;
char* Asura_GameSpy_HTTP::s_pcDownloadedData;
int Asura_GameSpy_HTTP::s_iDownloadedDataSize;
GHTTPRequest Asura_GameSpy_HTTP::s_iHandle;

/**************/
/* Functions. */
/**************/

/*************************************************************
 *
 * Name          : Asura_GameSpy_HTTP::Initialise
 *
 * Author        : Richard May
 *
 * Description   : Sets up and intialises Gamespy HTTP SDK
 *
 *************************************************************/
void Asura_GameSpy_HTTP::Initialise ( void )
{
	if(!s_bInitialised)
	{
		ghttpStartup();
		s_bInitialised = true;
	}
}
/* end of function Initialise */


/*************************************************************
 *
 * Name          : Asura_GameSpy_HTTP::Shutdown
 *
 * Author        : Richard May
 *
 * Description   : Cleans up and shutsdown Gamespy HTTP SDK
 *
 *************************************************************/
void Asura_GameSpy_HTTP::Shutdown ( void )
{
	if(s_bInitialised)
	{
		ghttpCleanup();
		s_bInitialised = false;
		s_bGetInProgress = false;
	}	
}
/* end of function Shutdown */


/*************************************************************
 *
 * Name          : Asura_GameSpy_HTTP::Reset
 *
 * Author        : Richard May
 *
 * Description   : Shutsdown, cleans up and re-intialises
 *
 *************************************************************/
void Asura_GameSpy_HTTP::Reset ( void )
{
	Shutdown();
	Initialise();
}
/* end of function Reset */


/*************************************************************
 *
 * Name          : Asura_GameSpy_HTTP::Update
 *
 * Author        : Richard May
 *
 * Description   : Update function. Calls ghttpThink as necessary
 *
 *************************************************************/
void Asura_GameSpy_HTTP::Update ( void )
{
	Asura_Assert(s_bInitialised, "HTTP not initialised");
	// (MH - 19.06.2007) ghttpThink now called by Asura_GameSpy
	//ghttpThink();
}
/* end of function Update */


/*************************************************************
 *
 * Name          : Asura_GameSpy_HTTP::Get
 *
 * Author        : Richard May
 *
 * Description   : Synchronous (blocking) get function. Note that
 *				   this function will _not_ return until the 
 *				   http query has been completed. For non-blocking
 *				   behaviour, use GetAsync
 *
 * Parameters    : szUrl - web address to get
 *				   pxCompletedCallback - Function called on completion
 *										 of get 
 *				   pxData - void pointer to data to pass to callbacks
 *				   pxProgressCallback - Function called every time 
 *										the status of the get changes
 *										(defaults to NULL)
 *				   szDataBuffer - Buffer to place retrieved data in.
 *								  If null (default), then data is lost
 *								  after completion callback unless manually
 *								  stored by the user
 *				   iBufferSize - Size of data buffer (defaults to 0)	
 *
 * Return value  : GHTTPRequest - request id and appropriate data
 *
 *************************************************************/
GHTTPRequest Asura_GameSpy_HTTP::Get ( const char* szUrl, ghttpCompletedCallback pxCompletedCallback, void* pxData/*=NULL*/,
									   char* szDataBuffer/*=NULL*/, int iBufferSize/*=0*/, ghttpProgressCallback pxProgressCallback/*=NULL*/ )
{
	Asura_Assert(s_bInitialised, "HTTP not initialised");
	return ghttpGetEx( szUrl, NULL, szDataBuffer, iBufferSize, NULL, GHTTPFalse, GHTTPTrue, pxProgressCallback, pxCompletedCallback, pxData );
}
/* end of function Get */


/*************************************************************
 *
 * Name          : Asura_GameSpy_HTTP::GetAsync
 *
 * Author        : Richard May
 *
 * Description   : Asynchronous version of Get. Non-blocking 
 *				   behaviour
 *
 *************************************************************/
GHTTPRequest Asura_GameSpy_HTTP::GetAsync ( const char* szUrl, ghttpCompletedCallback pxCompletedCallback, void* pxData/*=NULL*/, 
										    char* szDataBuffer/*=NULL*/, int iBufferSize/*=0*/, ghttpProgressCallback pxProgressCallback/*=NULL*/ )
{
	Asura_Assert(s_bInitialised, "HTTP not initialised");
	return ghttpGetEx( szUrl, NULL, szDataBuffer, iBufferSize, NULL, GHTTPFalse, GHTTPFalse, pxProgressCallback, pxCompletedCallback, pxData );
}
/* end of function Get */


// wrapped get function - calling code need not know about ghttp sdk at all
bool Asura_GameSpy_HTTP::BeginDownload( const char* szUrl )
{
	Asura_Assert(s_bInitialised, "HTTP not initialised");
	Asura_Assert(!s_bGetInProgress, "download already in progress");
	if (s_bGetInProgress)
	{
		return false;
	}

	GHTTPRequest iResult = ghttpGetEx( szUrl, NULL, NULL, 0, NULL, GHTTPFalse, GHTTPFalse, ProgressCallback, CompletedCallback, NULL );
	Asura_Assert(iResult != -1, "ghttpGetEx returned an error");
	if (iResult != -1)
	{
		s_bGetInProgress = true;
		s_iHandle = iResult;
		return true;
	}
	else
	{
		// set an error condition
		s_eResult = GHTTPOutOfMemory;
		return false;
	}
}

bool Asura_GameSpy_HTTP::GetDownloadProgress( float* pfProgress )
{
	// returns true if progress is known and puts amount of file 
	// retrieved from 0.0f - 1.0f in *pfProgress. If filesize is unknown, 
	// returns false and puts 0.0f in *pfProgress.
	if (s_fDownloadProgress == -1.0f)
	{
		*pfProgress = 0.0f;
		return false;
	}
	*pfProgress = s_fDownloadProgress;
	return true;
}

bool Asura_GameSpy_HTTP::IsDownloadComplete( void )
{
	return !s_bGetInProgress;
}

bool Asura_GameSpy_HTTP::HasErrorOccurred( void )
{
	if (s_bGetInProgress)
	{
		return false;
	}
	else
	{
		return (s_eResult != GHTTPSuccess);
	}
}

void Asura_GameSpy_HTTP::CancelDownload()
{
	if (s_bGetInProgress)
	{
		ghttpCancelRequest(s_iHandle); 
		s_bGetInProgress = false;
		s_eResult = GHTTPFileIncomplete;
	}
}


char* Asura_GameSpy_HTTP::GetDownloadedData( void )
{
	Asura_Assert(!s_bGetInProgress, "get is in progress!");
	return s_pcDownloadedData;
}

int Asura_GameSpy_HTTP::GetDownloadedDataSize( void )
{ 
	Asura_Assert(!s_bGetInProgress, "get is in progress!");
	return s_iDownloadedDataSize; 
}

void Asura_GameSpy_HTTP::FreeDownloadedData( void )
{
	gsifree(s_pcDownloadedData);
	s_pcDownloadedData = NULL;
	s_iDownloadedDataSize = 0;
}

// callbacks to deal with current file get.
void Asura_GameSpy_HTTP::ProgressCallback( GHTTPRequest iRequest, GHTTPState eState, const char* pcBuffer, GHTTPByteCount iBufferLen, GHTTPByteCount iBytesReceived, GHTTPByteCount iTotalSize, void* pUserData )
{
	Asura_Assert(s_bGetInProgress, "unexpected http progress callback");

	if (iTotalSize == -1)
	{
		// filesize is unknown
		s_fDownloadProgress = -1.0f;
	}
	else
	{
		if (iTotalSize == 0)
		{
			s_fDownloadProgress = 1.0f;
		}
		else
		{
			float fBytesDone = static_cast<float>(iBytesReceived);
			float fTotalBytes = static_cast<float>(iTotalSize);
			s_fDownloadProgress = fBytesDone / fTotalBytes;
		}
	}
}

GHTTPBool Asura_GameSpy_HTTP::CompletedCallback( GHTTPRequest iRequest, GHTTPResult eResult, char* pcBuffer, GHTTPByteCount iBufferLen, void* pUserData )
{
	Asura_Assert(s_bGetInProgress, "unexpected http completed callback");
	s_bGetInProgress = false;
	s_eResult = eResult;
	if (eResult == GHTTPSuccess)
	{
		s_fDownloadProgress = 1.0f;
	}
	Asura_Assert( !s_pcDownloadedData, "buffer is already allocated");
	s_pcDownloadedData = pcBuffer;
	Asura_Assert( iBufferLen < 0x7fffffff, "buffer size overflow?" );
	s_iDownloadedDataSize = static_cast<int>(iBufferLen);
	
	// return false to indicate we want to use the buffer the sdk allocated
	// and we will free it ourselves.
	return GHTTPFalse;
}

#endif //#ifdef ASURA_SUPPORT_GAMESPY
/****************/
/* End of file. */
/****************/
