//////////////////////////////////////////////////////////////////////////////
//
// Asura_GameSpy_PersistentStorage.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     interface for GameSpy persistent storage SDK
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_GAMESPY

// This file's header.
#include "Network/GameSpy/Asura_GameSpy_PersistentStorage.h"

// asura independent
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/GameSpy/Asura_GameSpy.h"
#include "Network/GameSpy/Asura_GameSpy_GP.h"

// middleware
#include "GameSpy/gstats/gpersist.h"

// library
#include <string.h>

// C L A S S E S /////////////////////////////////////////////////////////////
class Asura_GSPS_Internal
{
	friend class Asura_GameSpy_PersistentStorage;
private:

	static void AuthCallback   ( int iLocalid, int iProfileid, int iAuthenticated, char* szErrmsg, void* pInstance );
	static void GetDataCallback( int iLocalid, int iProfileid, persisttype_t eType, int iIndex, int iSuccess, time_t xModified, char* pcData, int iLen, void* pInstance );
	static void SetDataCallback( int iLocalid, int iProfileid, persisttype_t eType, int iIndex, int iSuccess, time_t xModified, void* pInstance );

	static Asura_GSPS_DataReceivedCallback s_pfnDataReceivedCallback;
	static Asura_GSPS_DataStoredCallback s_pfnDataStoredCallback;
};


// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
ASURA_GSPS_STATE Asura_GameSpy_PersistentStorage::s_eState = ASURA_GSPS_STATE_NOT_INITIALISED;

Asura_GSPS_DataReceivedCallback Asura_GSPS_Internal::s_pfnDataReceivedCallback;
Asura_GSPS_DataStoredCallback Asura_GSPS_Internal::s_pfnDataStoredCallback;

// S T A T I C   C L A S S   F U N C T I O N S ///////////////////////////////
bool Asura_GameSpy_PersistentStorage::Initialise()
{
	Asura_Assert(s_eState == ASURA_GSPS_STATE_NOT_INITIALISED, "already initialised");

	Asura_Assert(Asura_GameSpy_GP::IsLoggedIn(), "not logged in to GameSpy");

	if (Asura_GameSpy::ShouldUseDevelopmentBackend())
	{
		strcpy( StatsServerHostname, "sdkdev.gamespy.com" );
	}
	
	// setup project gamename, etc. 
	// note that these may have already been set, but setting them again shouldn't hurt
	strcpy( gcd_gamename, Asura_GameSpy::GetGameName() );
	strcpy( gcd_secret_key, Asura_GameSpy::GetGameKey() );

	
	
	
	// connect to gamespy server.
	int iResult = InitStatsAsync(0,0);
	if( iResult == GE_NOERROR || iResult == GE_CONNECTING)
	{
		s_eState = ASURA_GSPS_STATE_CONNECTING;
	}
	else
	{
		// churn out some useful debugging info
		switch(iResult)
		{
		case GE_NODNS     : Asura_Assert(false,"Connection failed: Unable to resolve stats server DNS"); break;
		case GE_NOSOCKET  : Asura_Assert(false,"Connection failed: Unable to create data socket"); break;
		case GE_NOCONNECT : Asura_Assert(false,"Connection failed: Unable to connect to stats server"); break;
		case GE_DATAERROR :	Asura_Assert(false,"Connection failed: Unable to receive challenge from stats server, or bad challenge"); break;
		default : Asura_Assert(false,"Connection failed: Unknown error"); break;
		}		
	}

	return ( iResult == GE_NOERROR || iResult == GE_CONNECTING);
}

bool Asura_GameSpy_PersistentStorage::Update()
{
	Asura_Assert(s_eState != ASURA_GSPS_STATE_NOT_INITIALISED, "not initialised");
	if (s_eState == ASURA_GSPS_STATE_NOT_INITIALISED)
	{
		return false;
	}

	if (s_eState == ASURA_GSPS_STATE_CONNECTING)
	{
		int iResult = InitStatsThink();
		if (iResult == GE_CONNECTING)
		{
			return true;//still connecting to persistant storage
		}
		else if ( iResult == GE_NOERROR )
		{
			Asura_OutputToDebugger(("Connected to gamespy tracking server\n"));
			char szAuth[33];
			GenerateAuth( GetChallenge(NULL), const_cast<char*>(Asura_GameSpy_GP::GetCurrentProfilePassword()), szAuth );

			s_eState = ASURA_GSPS_STATE_AUTHENTICATING;

			PreAuthenticatePlayerPM( 0, Asura_GameSpy_GP::GetCurrentProfileID(), szAuth, Asura_GSPS_Internal::AuthCallback, 0 );
			//connected - fall through to normal update code.
		}
		else
		{
			// churn out some useful debugging info
			switch(iResult)
			{
			case GE_NODNS     : Asura_Assert(false,"Connection failed: Unable to resolve stats server DNS"); break;
			case GE_NOSOCKET  : Asura_Assert(false,"Connection failed: Unable to create data socket"); break;
			case GE_NOCONNECT : Asura_Assert(false,"Connection failed: Unable to connect to stats server"); break;
			case GE_DATAERROR :	Asura_Assert(false,"Connection failed: Unable to receive challenge from stats server, or bad challenge"); break;
			case GE_TIMEDOUT :	Asura_Assert(false,"Connection failed: timeout"); break;
			default : Asura_Assert(false,"Connection failed: Unknown error"); break;
			}
			s_eState = ASURA_GSPS_STATE_NOT_INITIALISED;
			return false;//connection failled
		}
	}

	int iResult = PersistThink();
	bool bConnectionOK = (iResult == 1);
	Asura_Assert(bConnectionOK, "connection to persistent storage server lost");
	if (!bConnectionOK)
	{
		s_eState = ASURA_GSPS_STATE_NOT_INITIALISED;
	}
	return bConnectionOK;
}

void Asura_GameSpy_PersistentStorage::Shutdown()
{
	CloseStatsConnection();
	s_eState = ASURA_GSPS_STATE_NOT_INITIALISED;
}

bool Asura_GameSpy_PersistentStorage::GetDataBlock(int iIndex, Asura_GSPS_DataReceivedCallback pfnCallback, ASURA_GSPS_DATATYPE eDataType, void* pInstance /* = 0 */)
{
	Asura_Assert(s_eState == ASURA_GSPS_STATE_AUTHENTICATED, "not authenticated, or busy");
	if (s_eState != ASURA_GSPS_STATE_AUTHENTICATED)
	{
		return false;
	}
	s_eState = ASURA_GSPS_STATE_GETTINGDATA;
	Asura_GSPS_Internal::s_pfnDataReceivedCallback = pfnCallback;
	persisttype_t ePersistType = persisttype_t(eDataType);
	GetPersistData( 0, Asura_GameSpy_GP::GetCurrentProfileID(), ePersistType, iIndex, Asura_GSPS_Internal::GetDataCallback, pInstance );
	return true;
}

bool Asura_GameSpy_PersistentStorage::GetDataBlockValues(int iIndex, const char* szKeys, Asura_GSPS_DataReceivedCallback pfnCallback, ASURA_GSPS_DATATYPE eDataType, void* pInstance /* = 0 */)
{
	Asura_Assert(s_eState == ASURA_GSPS_STATE_AUTHENTICATED, "not authenticated, or busy");
	if (s_eState != ASURA_GSPS_STATE_AUTHENTICATED)
	{
		return false;
	}
	s_eState = ASURA_GSPS_STATE_GETTINGDATA;
	Asura_GSPS_Internal::s_pfnDataReceivedCallback = pfnCallback;
	persisttype_t ePersistType = persisttype_t(eDataType);
	GetPersistDataValues( 0, Asura_GameSpy_GP::GetCurrentProfileID(), ePersistType, iIndex, const_cast<char*>(szKeys), Asura_GSPS_Internal::GetDataCallback, pInstance );
	return true;
}

bool Asura_GameSpy_PersistentStorage::SetDataBlock(int iIndex, char* pcData, u_int uDataLength, Asura_GSPS_DataStoredCallback pfnCallback, ASURA_GSPS_DATATYPE eDataType, void* pInstance /* = 0 */)
{
	Asura_Assert(s_eState == ASURA_GSPS_STATE_AUTHENTICATED, "not authenticated, or busy");
	if (s_eState != ASURA_GSPS_STATE_AUTHENTICATED)
	{
		return false;
	}
	s_eState = ASURA_GSPS_STATE_SETTINGDATA;
	Asura_GSPS_Internal::s_pfnDataStoredCallback = pfnCallback;
	persisttype_t ePersistType = persisttype_t(eDataType);
	SetPersistData( 0, Asura_GameSpy_GP::GetCurrentProfileID(), ePersistType, iIndex, pcData, 
		static_cast<int>(uDataLength), Asura_GSPS_Internal::SetDataCallback, pInstance );
	return true;
}

bool Asura_GameSpy_PersistentStorage::SetDataBlockValues(int iIndex, const char* szKeysValues, Asura_GSPS_DataStoredCallback pfnCallback, ASURA_GSPS_DATATYPE eDataType, void* pInstance /* = 0 */)
{
	Asura_Assert(s_eState == ASURA_GSPS_STATE_AUTHENTICATED, "not authenticated, or busy");
	if (s_eState != ASURA_GSPS_STATE_AUTHENTICATED)
	{
		return false;
	}
	s_eState = ASURA_GSPS_STATE_SETTINGDATA;
	Asura_GSPS_Internal::s_pfnDataStoredCallback = pfnCallback;
	persisttype_t ePersistType = persisttype_t(eDataType);
	SetPersistDataValues( 0, Asura_GameSpy_GP::GetCurrentProfileID(), ePersistType, iIndex, 
		const_cast<char*>(szKeysValues), Asura_GSPS_Internal::SetDataCallback, pInstance );
	return true;
}

// internal use functions
void Asura_GSPS_Internal::AuthCallback( int iLocalid, int iProfileid, int iAuthenticated, char* szErrmsg, void* pInstance )
{
	Asura_Assert(Asura_GameSpy_PersistentStorage::s_eState == ASURA_GSPS_STATE_AUTHENTICATING, "unexpected state");

	if (iAuthenticated)
	{
		Asura_OutputToDebugger_Retail(("logged in to persistent storage\n"));
		Asura_GameSpy_PersistentStorage::s_eState = ASURA_GSPS_STATE_AUTHENTICATED;
	}
	else
	{
		Asura_OutputToDebugger_Retail(("failed to log in to persistent storage, reason = %s\n", szErrmsg));
		Asura_GameSpy_PersistentStorage::s_eState = ASURA_GSPS_STATE_AUTH_FAILED;
	}
}

void Asura_GSPS_Internal::GetDataCallback( int iLocalid, int iProfileid, persisttype_t eType, int iIndex, 
										   int iSuccess, time_t xModified, char* pcData, int iLen, void* pInstance )
{
	Asura_Assert(Asura_GameSpy_PersistentStorage::s_eState == ASURA_GSPS_STATE_GETTINGDATA, "unexpected callback");
	if (Asura_GameSpy_PersistentStorage::s_eState == ASURA_GSPS_STATE_GETTINGDATA)
	{
		if (!iSuccess)
		{
			pcData = 0;
			iLen = 0;
		}
		Asura_GameSpy_PersistentStorage::s_eState = ASURA_GSPS_STATE_AUTHENTICATED;
		s_pfnDataReceivedCallback( iSuccess != 0, pcData, static_cast<u_int>(iLen), pInstance );
	}
}

void Asura_GSPS_Internal::SetDataCallback( int iLocalid, int iProfileid, persisttype_t eType, int iIndex, 
										   int iSuccess, time_t xModified, void* pInstance )
{
	Asura_Assert(Asura_GameSpy_PersistentStorage::s_eState == ASURA_GSPS_STATE_SETTINGDATA, "unexpected callback");
	if (Asura_GameSpy_PersistentStorage::s_eState == ASURA_GSPS_STATE_SETTINGDATA)
	{
		Asura_GameSpy_PersistentStorage::s_eState = ASURA_GSPS_STATE_AUTHENTICATED;
		s_pfnDataStoredCallback( iSuccess != 0, pInstance );
	}
}

#endif // #ifdef ASURA_SUPPORT_GAMESPY
