//////////////////////////////////////////////////////////////////////////////
//
// Asura_GameSpy_QR2.h
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     interface for GameSpy Query and Reporting SDK. Project independent part
//	   isn't able to do much except for Init/Shutdown and some very basic keys
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_GAMESPY_QR2_H___
#define ___ASURA_GAMESPY_QR2_H___ 1

#ifdef ASURA_SUPPORT_GAMESPY

// asura includes
#include "Core/Asura_PlatformBasics.h"
#include "Network/Asura_Guid.h"

// middleware includes
#include "GameSpy/qr2/qr2.h"

// Asura custom keys
#define ASURA_MAPID_KEY			50
#define ASURA_GAMETYPEID_KEY	51

class Asura_GameSpy_QR2
{
	friend class Asura_GameSpy_Peer;

	public:

		enum ServerState
		{
			QR2_ServerState_OpenWaiting,	// game has not yet started and players can join
			QR2_ServerState_ClosedWaiting,	// game has not yet started and players cannot join
			QR2_ServerState_ClosedPlaying,	// game is in progress, no joining allowed
			QR2_ServerState_OpenPlaying,	// game is in progress, players may still join
			QR2_ServerState_OpenStaging,	// Use to report that the game is in staging mode ... 
			QR2_ServerState_ClosedStaging,	// (should generally not be used directly - the Peer SDK handles this automatically).
			QR2_ServerState_Exiting,		// server is shutting down
		};
		
		static bool Initialise( bool bLanOnly, SOCKET xSocket, u_short usPort );
		static void Update( void );
		static void Shutdown( void );

		// function to be called when server changes state, e.g. from frontend
		// to in-game.
		static void GameModeChanged( void );
		
		// custom keys must be registered before either Asura_GameSpy_QR2 or Asura_GameSpy_ServerBrowsing are initialised
		//		(so they can be used as part of the basic key set)
		static void RegisterCustomKeys( void );

		static bool IsInitialised( void ) { return s_bInitialised; }
		// it takes a while to find out if we are able to connect to the GameSpy
		// backend - these 2 functions allow each project to deal with failures 
		// as they see fit, e.g. by waiting for HasFinishedInitialising to return 
		// true and displaying whether the server was created successfully.
		static bool HasFinishedInitialising( void );
		static bool InitialisedOK( void ) { return !s_bErrorAddingServer; }
	
		// SetGroupID is used to make a server report itself as part of a group room e.g. the English room.
		static void SetGroupID( int iGroupID ) { s_iGroupID = iGroupID; }

		//functions to retreive public IP address as seen by gamespy
		static bool HasPublicIP();//checks if IP address has been received
		static u_int GetPublicIP();
		static u_short GetPublicPort();

		static u_int GetNumberOfCustomServerKeys()			{ return Project_GetNumberOfCustomServerKeys(); }
		static u_int GetCustomServerKeyID(u_int uKeyIndex)	{ return Project_GetCustomServerKeyID(uKeyIndex); }


	private:

		static void ServerKeyCallback(int iKeyId, qr2_buffer_t xOutbuf, void* /*pUserData*/);
		static void PlayerKeyCallback(int iKeyId, int iIndex, qr2_buffer_t xOutbuf, void* /*pUserData*/);
		static void TeamKeyCallback(int iKeyId, int iIndex, qr2_buffer_t xOutbuf, void* /*pUserData*/);
		static void KeyListCallback(qr2_key_type eKeyType, qr2_keybuffer_t xKeybuffer, void* /*pUserData*/);
		static int CountCallback(qr2_key_type eKeyType, void* /*pUserData*/);
		static void AddErrorCallback(qr2_error_t eError, char *szErrmsg, void* /*pUserData*/);

		static void NatNegCallback(int iCookie, void* /*pUserData*/);
		static void ClientMessageCallback(char* pcData, int iLen, void* /*pUserData*/);
		static void PublicAddressCallback(u_int uIP, u_short usPort, void* /*pUserData*/);

		// project-supplied functions
		// (MH - 30.7.07) changed the map and gametype functions to return const char*. Localised games 
		// should use english for these and convert to the local language on the client for display.
		static const char* Project_GetMapName( void );
		static const char* Project_GetGameTypeString( void );
		static ServerState Project_GetServerState( void );
		static int Project_GetNumTeams( void );
		static bool Project_IsTeamGame( void );
		static int Project_GetFragLimit( void );
		static int Project_GetTimeLimit( void );

		static int Project_GetPlayerScore( Asura_Guid uPlayerGuid );
		static int Project_GetPlayerDeaths( Asura_Guid uPlayerGuid );
		static int Project_GetPlayerTeamNumber( Asura_Guid uPlayerGuid );

		static int Project_GetTeamScore( int iTeamIndex );
		static const Asura_UnicodeChar* Project_GetTeamName( int iTeamIndex );

		static int Project_GetMapID( void );
		static int Project_GetGameTypeID( void );

		
		//functions to allow projects to implement their own custom server keys
		//will need stubbing out if not needed - return zero or "" as appropriate should do it.

		//should return number of custom server keys the project wishes to add
		static u_int Project_GetNumberOfCustomServerKeys();
		
		//should return the KeyID for the requested custom key
		//(GS keys are <50, asura keys start ay 50. I've used 60 and up for BHD)
		//uKeyIndex ranges from 0 to the return value from Project_GetNumberOfCustomServerKeys,
		//and is used solely to enumerate the custom keys while reporting them to gamespy
		static int Project_GetCustomServerKeyID(u_int uKeyIndex);

		//should return the key name for the requested custom key
		//uKeyIndex ranges from 0 to the return value from Project_GetNumberOfCustomServerKeys,
		//and is used solely to enumerate the custom keys while reporting them to gamespy
		static const char * Project_GetCustomServerKeyName(u_int uKeyIndex);

		//should check if the passed key is handled the by the project, and return true if it's handled
		//call qr2_buffer_add(xOutbuf, "string") or
		//qr2_buffer_add_int(xOutbuf, 17) to pass the result to gamespy
		static bool Project_ServerKeyCallback(int iKeyId, qr2_buffer_t xOutbuf);
		
		// data
		static bool s_bInitialised;
		static bool s_bErrorAddingServer;
		static int s_iGroupID;

		//public IP as seen by Gamespy
		static u_int s_uPublicIP;
		static u_short s_usPublicPort;
};

#endif //#ifdef ASURA_SUPPORT_GAMESPY

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_GAMESPY_QR2_H___
