//////////////////////////////////////////////////////////////////////////////
//
// Asura_GameSpy_QR2.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     interface for GameSpy Query and Reporting SDK
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_GAMESPY

// this file's header
#include "Network/GameSpy/Asura_GameSpy_QR2.h"


// Asura platform-independent
#include "Debugging/Asura_Logfile.h"
#include "Network/GameSpy/Asura_GameSpy.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_NetTransport_System.h"

// middleware
#include "GameSpy/common/gsStringutil.h"
#include "GameSpy/qr2/qr2.h"

// G L O B A L S /////////////////////////////////////////////////////////////
static const char* g_aszServerStateString[] = 
{
	"openwaiting",	// QR2_ServerState_OpenWaiting,	// game has not yet started and players can join
	"closedwaiting",// QR2_ServerState_ClosedWaiting,	// game has not yet started and players cannot join
	"closedplaying",// QR2_ServerState_ClosedPlaying,	// game is in progress, no joining allowed
	"openplaying",	// QR2_ServerState_OpenPlaying,	// game is in progress, players may still join
	"openstaging",	// QR2_ServerState_OpenStaging,	// Use to report that the game is in staging mode ... 
	"closedstaging",// QR2_ServerState_ClosedStaging,	// (should generally not be used directly - the Peer SDK handles this automatically).
	"exiting",		// QR2_ServerState_Exiting,		// server is shutting down
};

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
bool Asura_GameSpy_QR2::s_bInitialised = false;
bool Asura_GameSpy_QR2::s_bErrorAddingServer = false;
int Asura_GameSpy_QR2::s_iGroupID = -1;
//public IP as seen by Gamespy
u_int Asura_GameSpy_QR2::s_uPublicIP = 0;
u_short Asura_GameSpy_QR2::s_usPublicPort = 0;

bool Asura_GameSpy_QR2::Initialise( bool bLanOnly, SOCKET xSocket, u_short usPort )
{
	const char* szGameName = Asura_GameSpy::GetGameName();
	const char* szGameKey = Asura_GameSpy::GetGameKey();

	RegisterCustomKeys();

	//call qr_init with the query port number and gamename, default IP address, and no user data
	bool bIsPublic = !bLanOnly;
	bool bNatNegotiate = !bLanOnly;
	// (MH - 25.11.2003) for nat negotiation to work we need the underlying socket, and that
	// is not available in DirectPlay. So if you want to use DP, you have to call qr2_init
	// and accept that players behind a NAT will probably not be able to host.
	qr2_error_t eError = qr2_init_socket(	NULL,								// qr2_t *qrec, 
											xSocket,							// SOCKET s, 
											usPort,								// int boundport, 
											szGameName,							// const char *gamename, 
											szGameKey,							// const char *secret_key,
											bIsPublic,							// int ispublic, 
											bNatNegotiate,						// int natnegotiate,
											ServerKeyCallback,					// qr2_serverkeycallback_t server_key_callback,
											PlayerKeyCallback,					// qr2_playerteamkeycallback_t player_key_callback,
											TeamKeyCallback,					// qr2_playerteamkeycallback_t team_key_callback,
											KeyListCallback,					// qr2_keylistcallback_t key_list_callback,
											CountCallback,						// qr2_countcallback_t playerteam_count_callback,
											AddErrorCallback,					// qr2_adderrorcallback_t adderror_callback,
											NULL);								// void *userdata
	
	if (eError != e_qrnoerror)
	{
		Asura_OutputToDebugger_Retail(("Error starting GameSpy query sockets\n"));
		Asura_Assertf(false, ("Error starting GameSpy query sockets. reason: %d\n", eError));
		return false;
	}

	qr2_register_clientmessage_callback(NULL, ClientMessageCallback);
	if (bNatNegotiate)
	{
		qr2_register_natneg_callback(NULL, NatNegCallback);
	}
	qr2_register_publicaddress_callback(NULL, PublicAddressCallback);

	s_bInitialised = true;
	s_bErrorAddingServer = false;
	s_uPublicIP = 0;
	s_usPublicPort = 0;
	return true;
}

void Asura_GameSpy_QR2::Shutdown()
{
	if (s_bInitialised)
	{
		qr2_shutdown(NULL);
	}
	s_bInitialised = false;
	s_uPublicIP = 0;
	s_usPublicPort = 0;
}

void Asura_GameSpy_QR2::Update()
{
	if (s_bInitialised)
	{
		qr2_think(NULL);
	}
}

void Asura_GameSpy_QR2::GameModeChanged()
{
	if (s_bInitialised)
	{
#if 0 //def ASURA_SUPPORT_COMMAND_CONSOLE
		static u_int ls_uLastCall = 0;
		if ( (Asura_Timers::GetMSecsSinceStart_Continuous() - ls_uLastCall) < 9999 )
		{
			Asura_Assert(false, "Asura_GameSpy_QR2::GameModeChanged() called too often");
			Asura_CommandConsole::Printf("Asura_GameSpy_QR2::GameModeChanged() called too often");
		}
		ls_uLastCall = Asura_Timers::GetMSecsSinceStart_Continuous();
#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE
		qr2_send_statechanged(NULL);
	}
}

void Asura_GameSpy_QR2::RegisterCustomKeys()
{
	Asura_Assert(!s_bInitialised, "RegisterCustomKeys should be called before QR2 is initialised");

	//register custom keys. Must be done before qr2_init is called (and also before the server browser refresh
	//	if using these keys as basic keys).
	qr2_register_key(ASURA_MAPID_KEY, "mapid");
	qr2_register_key(ASURA_GAMETYPEID_KEY, "gametypeid");
	//loop through project custom server keys and add them in
	u_int uKeys = Project_GetNumberOfCustomServerKeys();
	for (u_int u = 0; u < uKeys; u++)
	{
		qr2_register_key(Project_GetCustomServerKeyID(u),Project_GetCustomServerKeyName(u));
	}
} 

//functions to retreive public IP address as seen by gamespy
bool Asura_GameSpy_QR2::HasPublicIP()//checks if IP address has been received
{
	return s_uPublicIP && s_usPublicPort;//checks both IP and port have been returned
}

u_int Asura_GameSpy_QR2::GetPublicIP()
{
	Asura_Assert(s_uPublicIP && s_usPublicPort, "Public IP not set by gamespy");
	return s_uPublicIP;
}

u_short Asura_GameSpy_QR2::GetPublicPort()
{
	Asura_Assert(s_uPublicIP && s_usPublicPort, "Public port not set by gamespy");
	return s_usPublicPort;
}

void Asura_GameSpy_QR2::ServerKeyCallback(int iKeyId, qr2_buffer_t xOutbuf, void* /*userdata*/)
{
    Asura_Assert(Asura_Network::IsNetworkInitialised(), "network not initialised");
    Asura_Assert(Asura_Network::IsServer(), "gamespy server callback but not server");

	if (Project_ServerKeyCallback(iKeyId, xOutbuf))
	{
		return;
	}

	char acTemp[256];

	switch (iKeyId)
	{
		case HOSTNAME_KEY:
		{
			const UCS2String wszServerName = reinterpret_cast<const u_short*>(Asura_Network::GetServerName());
			UCS2ToUTF8String(wszServerName, acTemp);
			qr2_buffer_add(xOutbuf, acTemp);
			break;
		}
		case GAMEVER_KEY:
		{
			Asura_PrintfToBfr(acTemp, "%d", 256, true, Asura_GameSpy::GetGameVersion());
			qr2_buffer_add(xOutbuf, acTemp);
			break;
		}
// (MH - 25.11.2003) we're not supposed to report this if using shared sockets.
//		case HOSTPORT_KEY:
//		{
//			//qr2_buffer_add_int(xOutbuf, Asura_PC_DirectPlay::GetPort());
//			qr2_buffer_add_int(xOutbuf, Asura_GameSpy_GT2::GetPort());
//			break;
//		}
		case NUMPLAYERS_KEY:
		{
			qr2_buffer_add_int(xOutbuf, Asura_Network::GetNumberOfPlayers());
			break;
		}
		case MAXPLAYERS_KEY:
		{
			qr2_buffer_add_int(xOutbuf, Asura_Network::GetMaxNumberOfPlayers());
			break;
		}
		case MAPNAME_KEY:
		{
			qr2_buffer_add(xOutbuf, Project_GetMapName());
			break;
		}
		case GAMETYPE_KEY:
		{
			qr2_buffer_add(xOutbuf, Project_GetGameTypeString());
			break;
		}
		case NUMTEAMS_KEY:
		{
			qr2_buffer_add_int(xOutbuf, Project_GetNumTeams());
			break;
		}
		case GAMEMODE_KEY:
		{
			qr2_buffer_add(xOutbuf, g_aszServerStateString[Project_GetServerState()]);
			break;
		}
		case TEAMPLAY_KEY:
		{
			qr2_buffer_add_int(xOutbuf, Project_IsTeamGame());
			break;
		}
		case FRAGLIMIT_KEY:
		{	
			qr2_buffer_add_int(xOutbuf, Project_GetFragLimit());
			break;
		}
		case TIMELIMIT_KEY:
		{	
			qr2_buffer_add_int(xOutbuf, Project_GetTimeLimit());
			break;
		}
		case PASSWORD_KEY:
		{
            //bool bHasPassword = Asura_PC_DirectPlay::GetPasswordHash() != 0;
			bool bHasPassword = false;
			bHasPassword = Asura_NetTransport_System::GetPasswordHash() != 0;
			qr2_buffer_add_int(xOutbuf, bHasPassword);
			break;
		}
		case GROUPID_KEY:
		{
			qr2_buffer_add_int(xOutbuf, s_iGroupID);
			break;
		}

		// asura's custom keys
		case ASURA_MAPID_KEY:
		{
			qr2_buffer_add_int(xOutbuf, Project_GetMapID());
			break;
		}
		case ASURA_GAMETYPEID_KEY:
		{
			qr2_buffer_add_int(xOutbuf, Project_GetGameTypeID());
			break;
		}
		default:
		{
			//not handled by asura, so pass on to the project handler
			qr2_buffer_add(xOutbuf, "");
			break;
		}
	}
}

void Asura_GameSpy_QR2::PlayerKeyCallback(int iKeyId, int iIndex, qr2_buffer_t xOutbuf, void* pUserData)
{
    Asura_Assert(Asura_Network::IsNetworkInitialised(), "network not initialised");
    Asura_Assert(Asura_Network::IsServer(), "gamespy server callback but not server");

	// check for valid index
	if ( static_cast<u_int>(iIndex) >= Asura_Network::GetNumberOfPlayers() )
	{
		qr2_buffer_add(xOutbuf, "");
		return;
	}

	Asura_Guid uPlayerGuid = Asura_Network::GetPlayerGuid(static_cast<u_int>(iIndex));
	if (uPlayerGuid == ASURA_GUID_UNREGISTERED)
	{
		qr2_buffer_add(xOutbuf, "");
		return;
	}


	char acTemp[256];
	
	switch (iKeyId)
	{
	case PLAYER__KEY:
	{
		const UCS2String wszName = reinterpret_cast<const u_short*>(Asura_Network::GetPlayerName(uPlayerGuid));
		UCS2ToUTF8String(wszName, acTemp);
		qr2_buffer_add(xOutbuf, acTemp);
		break;
	}
	case SCORE__KEY:
	{	
		qr2_buffer_add_int(xOutbuf, Project_GetPlayerScore(uPlayerGuid));
		break;
	}
	case DEATHS__KEY:
	{
		qr2_buffer_add_int(xOutbuf, Project_GetPlayerDeaths(uPlayerGuid));
		break;
	}
	case PING__KEY:
	{	
		qr2_buffer_add_int(xOutbuf, Asura_Network::GetLatencyForPlayer(uPlayerGuid));//Asura_Network::GetLatencyForPlayer(uPlayerGuid));
		break;
	}
	case TEAM__KEY:
	{	qr2_buffer_add_int(xOutbuf, Project_GetPlayerTeamNumber(uPlayerGuid));
		break;
	}
//	case TIME__KEY:
//		qr2_buffer_add_int(xOutbuf, gamedata.players[index].ptime);
//		break;
	default:
		qr2_buffer_add(xOutbuf, "");
		break;		
	}
}

void Asura_GameSpy_QR2::TeamKeyCallback(int iKeyId, int iIndex, qr2_buffer_t xOutbuf, void *userdata)
{
    Asura_Assert(Asura_Network::IsNetworkInitialised(), "network not initialised");
    Asura_Assert(Asura_Network::IsServer(), "gamespy server callback but not server");

	// check for valid index
	if (iIndex >= Project_GetNumTeams())
	{
		Asura_Assert(false, "invalid team index");
		qr2_buffer_add(xOutbuf, "");
		return;
	}

	switch (iKeyId)
	{
	case TEAM_T_KEY:
	{
		char acTemp[256];
		const UCS2String wszTeamName = reinterpret_cast<const u_short*>(Project_GetTeamName(iIndex));
		UCS2ToUTF8String(wszTeamName, acTemp);
		qr2_buffer_add(xOutbuf, acTemp);
		break;
	}
	case SCORE_T_KEY:
	{
		Project_GetTeamScore(iIndex);
		break;
	}
//	case AVGPING_T_KEY:
//		qr2_buffer_add_int(xOutbuf, gamedata.teams[index].avgping);
//		break;
	default:
		qr2_buffer_add(xOutbuf, "");
		break;		
	}
}	

void Asura_GameSpy_QR2::KeyListCallback(qr2_key_type eKeyType, qr2_keybuffer_t xKeybuffer, void* /*pUserDate*/)
{
    Asura_Assert(Asura_Network::IsNetworkInitialised(), "network not initialised");
    Asura_Assert(Asura_Network::IsServer(), "gamespy server callback but not server");

	//need to add all the keys we support
	switch (eKeyType)
	{
	case key_server:
		qr2_keybuffer_add(xKeybuffer, HOSTNAME_KEY);
		qr2_keybuffer_add(xKeybuffer, GAMEVER_KEY);
		qr2_keybuffer_add(xKeybuffer, HOSTPORT_KEY);
		qr2_keybuffer_add(xKeybuffer, MAPNAME_KEY);
		qr2_keybuffer_add(xKeybuffer, GAMETYPE_KEY);
		qr2_keybuffer_add(xKeybuffer, NUMPLAYERS_KEY);
		qr2_keybuffer_add(xKeybuffer, NUMTEAMS_KEY);
		qr2_keybuffer_add(xKeybuffer, MAXPLAYERS_KEY);
		qr2_keybuffer_add(xKeybuffer, GAMEMODE_KEY);
		qr2_keybuffer_add(xKeybuffer, TEAMPLAY_KEY);
		qr2_keybuffer_add(xKeybuffer, FRAGLIMIT_KEY);
		qr2_keybuffer_add(xKeybuffer, TIMELIMIT_KEY);
		qr2_keybuffer_add(xKeybuffer, PASSWORD_KEY);
		if (s_iGroupID != -1)
		{
			qr2_keybuffer_add(xKeybuffer, GROUPID_KEY);
		}

		// asura custom keys
		qr2_keybuffer_add(xKeybuffer, ASURA_MAPID_KEY);
		qr2_keybuffer_add(xKeybuffer, ASURA_GAMETYPEID_KEY);
		{
			//loop through project custom server keys and add them in
			u_int uKeys = Project_GetNumberOfCustomServerKeys();
			for (u_int u = 0; u < uKeys; u++)
			{
				qr2_keybuffer_add(xKeybuffer, Project_GetCustomServerKeyID(u));
			}
		}

//		qr2_keybuffer_add(xKeybuffer, GRAVITY_KEY); //a custom key
//		qr2_keybuffer_add(xKeybuffer, RANKINGON_KEY); //a custom key
		break;
	case key_player:
		qr2_keybuffer_add(xKeybuffer, PLAYER__KEY);
		qr2_keybuffer_add(xKeybuffer, SCORE__KEY);
		qr2_keybuffer_add(xKeybuffer, DEATHS__KEY);
		qr2_keybuffer_add(xKeybuffer, PING__KEY);
		qr2_keybuffer_add(xKeybuffer, TEAM__KEY);
//		qr2_keybuffer_add(xKeybuffer, TIME__KEY); //a custom key
		break;
	case key_team:
		qr2_keybuffer_add(xKeybuffer, TEAM_T_KEY);
		qr2_keybuffer_add(xKeybuffer, SCORE_T_KEY);
//		qr2_keybuffer_add(xKeybuffer, AVGPING_T_KEY); //a custom key
		break;
	default:
		break;
	}
}

int Asura_GameSpy_QR2::CountCallback(qr2_key_type eKeyType, void* /*pUserdata*/)
{
	if (eKeyType == key_player)
		return Asura_Network::GetNumberOfPlayers();
	else if (eKeyType == key_team)
		return Project_GetNumTeams();
	else
		return 0;
}

void Asura_GameSpy_QR2::AddErrorCallback(qr2_error_t eError, char *szErrmsg, void* /*pUserData*/)
{
	s_bErrorAddingServer = (eError != e_qrnoerror);

	Asura_OutputToDebugger_Retail(("Error adding server: %d, %s\n", eError, szErrmsg));

	if (s_bErrorAddingServer)
	{
		Asura_Network::SetConnectionOkay( false );
		Asura_Network::SetReasonForDisconnection( ASURA_NDR_FAILED_TO_LIST_SERVER );
	}
}

void Asura_GameSpy_QR2::NatNegCallback(int iCookie, void* /*pUserdata*/)
{
	Asura_OutputToDebugger_Retail(("Got natneg cookie: %d\n", iCookie));
	Asura_NetTransport_System::ServerBeginNatNeg(iCookie);
}

void Asura_GameSpy_QR2::ClientMessageCallback(char* /*pcData*/, int iLen, void* /*pUserData*/)
{
	Asura_OutputToDebugger(("Got %d bytes from client\n", iLen));
}

void Asura_GameSpy_QR2::PublicAddressCallback(u_int uIP, u_short usPort, void* /*pUserData*/)
{
	// this gives the address and port as reported by the NAT to GameSpy's servers
	s_uPublicIP = uIP;
	s_usPublicPort = usPort;
	Asura_OutputToDebugger_Retail(("public ip: %u.%u.%u.%u, port: %u\n", 
		(uIP & 0x000000ff) >> 0,
		(uIP & 0x0000ff00) >> 8,
		(uIP & 0x00ff0000) >> 16,
		(uIP & 0xff000000) >> 24, 
		usPort));
}

#endif //#ifdef ASURA_SUPPORT_GAMESPY
