//////////////////////////////////////////////////////////////////////////////
//
// Asura_Matchmaking_System.cpp
//
// Copyright (c) 2009 Rebellion. All rights reserved.
//
// Description:
//
//     system which sits on top of Asura_Lobby_System to deal with automatic
//     skill-based matchmaking
//     See Wiki documentation:
//     http://asurawiki.rebellion.co.uk/twiki/bin/view/Programming/MatchmakingAlgorithm
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_LOBBY

// This file's header.
#include "Network/Online/Asura_Matchmaking_System.h"

// Asura includes
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Online/Asura_Lobby.h"


// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_Matchmaking_System::s_bInitialised;
ASURA_MATCHMAKING_STATE Asura_Matchmaking_System::s_eMatchmakingState;
u_int Asura_Matchmaking_System::s_uSkillThreshold;
u_int Asura_Matchmaking_System::s_uCurrentSkillthresholdIndex;
u_int Asura_Matchmaking_System::s_uBestLobby;			
u_int Asura_Matchmaking_System::s_uHostingStartTimestamp;
u_int Asura_Matchmaking_System::s_uHostingStateEnterTimestamp;
u_int Asura_Matchmaking_System::s_uSearchResultsTimestamp;
float Asura_Matchmaking_System::s_fMatchQualityThreshold;
bool* Asura_Matchmaking_System::s_pbLobbyAttempted;
bool Asura_Matchmaking_System::s_bConnectingStateHasJoinedLobby;

// the following parameters are intended to be tweakable
u_int Asura_Matchmaking_System::s_uMinPlayers = ASURA_LOBBY_MAX_PLAYERS / 2;
u_int Asura_Matchmaking_System::s_uMaxPlayers = ASURA_LOBBY_MAX_PLAYERS;
float Asura_Matchmaking_System::s_fInitialMatchQualityThreshold = 0.9f;
float Asura_Matchmaking_System::s_fMatchQualityThresholdIncrement = 0.1f;

#ifdef ASURA_USE_MERIT_SYSTEM
u_int Asura_Matchmaking_System::s_auSkillThresholds[uASURA_MATCHMAKING_NUM_SKILLTHRESHOLDS] = { 1,3,5,9999 };
#else
u_int Asura_Matchmaking_System::s_auSkillThresholds[uASURA_MATCHMAKING_NUM_SKILLTHRESHOLDS] = { 9999 };
#endif
u_int Asura_Matchmaking_System::s_uSearchResultsStaleTime = 10000;
u_int Asura_Matchmaking_System::s_uTimeToWaitForSufficientPlayers = 12000;//120000;
float Asura_Matchmaking_System::s_fMaxPlayersTime = 5.0f;//30.0f;
float Asura_Matchmaking_System::s_fMinPlayersTime = 5.0f;//30.0f;

// D E B U G G I N G   F U N C T I O N S /////////////////////////////////////
#ifdef ASURA_LOGFILES_ENABLED
#define ASURA_MATCHMAKING_TRACE Asura_OutputToDebugger_Retail(("MATCHMAKING: ")); Asura_OutputToDebugger_Retail
static const char* MatchmakingStateToString(ASURA_MATCHMAKING_STATE eState)
{
	switch (eState)
	{
		case ASURA_MATCHMAKING_STATE_NONE: return "None";
		case ASURA_MATCHMAKING_STATE_SEARCHING: return "Searching";
		case ASURA_MATCHMAKING_STATE_CONNECTING: return "Connecting";
		case ASURA_MATCHMAKING_STATE_HOSTING: return "Hosting";
		case ASURA_MATCHMAKING_STATE_CHOOSING: return "Choosing";
		case ASURA_MATCHMAKING_STATE_CREATING: return "Creating";
		case ASURA_MATCHMAKING_STATE_WAITING: return "Waiting";
		case ASURA_MATCHMAKING_STATE_FINISHED: return "Finished";
		case ASURA_MATCHMAKING_STATE_ERROR: return "Error";
		default: return "Unknown";
	}
}
#else
#define ASURA_MATCHMAKING_TRACE(x)	//  ignore
#endif

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_Matchmaking_System::Initialise()
{
	Asura_Assert(!s_bInitialised, "already initialised?");
	s_bInitialised = true;
	s_eMatchmakingState = ASURA_MATCHMAKING_STATE_NONE;
	s_uBestLobby = ASURA_LOBBY_MAX_SEARCH_RESULTS;	// invalid value
	s_uHostingStartTimestamp = 0;
	s_uHostingStateEnterTimestamp = 0;
	s_uSearchResultsTimestamp = 0;
	s_fMatchQualityThreshold = s_fInitialMatchQualityThreshold;
	Asura_Assert(!s_pbLobbyAttempted, "memory leak?");
	s_pbLobbyAttempted = new bool[ASURA_LOBBY_MAX_SEARCH_RESULTS];
	s_bConnectingStateHasJoinedLobby = false;
	
	AddConsoleCmdsAndVars();
}

void Asura_Matchmaking_System::AddConsoleCmdsAndVars()
{
	Asura_CommandConsole::AddCmd( L"Asura.Matchmaking.BeginAutomatch", ConsoleCmd_BeginAutomatch );
	Asura_CommandConsole::AddCmd( L"Asura.Matchmaking.CancelAutomatch", ConsoleCmd_CancelAutomatch );
	Asura_CommandConsole::AddCmd( L"Asura.Matchmaking.ForceAutomatchStart", ConsoleCmd_ForceAutomatchStart );
	Asura_CommandConsole::AddCmd( L"Asura.Matchmaking.PrintAutomatchStatus", ConsoleCmd_PrintAutomatchStatus );
	Asura_CommandConsole::AddCmd( L"Asura.Matchmaking.AutomatchComplete", ConsoleCmd_IsAutomatchComplete );
}

void Asura_Matchmaking_System::Shutdown(bool bLeaveLobby)
{
	if (bLeaveLobby)
	{
		Asura_Lobby_System::LeaveLobby(false);
	}

	s_eMatchmakingState = ASURA_MATCHMAKING_STATE_NONE;
	delete[] s_pbLobbyAttempted;
	s_pbLobbyAttempted = NULL;
	s_bInitialised = false;
}

void Asura_Matchmaking_System::Start()
{
	Asura_Assert(s_bInitialised && (IsIdle() || IsFinished()), "can't start");
	if (s_bInitialised && (IsIdle() || IsFinished()))
	{
		ChangeState(ASURA_MATCHMAKING_STATE_SEARCHING);
	}
}

void Asura_Matchmaking_System::Cancel()
{
	Asura_Assert(s_bInitialised, "not intialised");
	if (s_bInitialised)
	{
		Asura_Lobby_System::LeaveLobby(false);
		Asura_Lobby_System::CancelSearch();
		ChangeState(ASURA_MATCHMAKING_STATE_NONE);
	}
}

void Asura_Matchmaking_System::Update()
{
	Asura_Assert(s_bInitialised, "not intialised");

	switch (s_eMatchmakingState)
	{
		case ASURA_MATCHMAKING_STATE_NONE:
		{
			break;
		}
		case ASURA_MATCHMAKING_STATE_SEARCHING:
		{
			ExecuteState_Searching();
			break;
		}
		case ASURA_MATCHMAKING_STATE_CONNECTING:
		{
			ExecuteState_Connecting();
			break;
		}
		case ASURA_MATCHMAKING_STATE_HOSTING:
		{
			ExecuteState_Hosting();
			break;
		}
		case ASURA_MATCHMAKING_STATE_CHOOSING:
		{
			ExecuteState_Choosing();
			break;
		}
		case ASURA_MATCHMAKING_STATE_CREATING:
		{
			ExecuteState_Creating();
			break;
		}
		case ASURA_MATCHMAKING_STATE_WAITING:
		{
			ExecuteState_Waiting();
			break;
		}
		case ASURA_MATCHMAKING_STATE_FINISHED:
		case ASURA_MATCHMAKING_STATE_ERROR:
		{
			// nothing to do
			break;
		}
		default:
		{
			Asura_Assert(false, "unhandled state");
			break;
		}
	}

	// update num players being waited for.
	if (Asura_Lobby_System::IsHost())
	{
		u_int uWaitingForPlayers = 0;
		const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
		if (pxLobby)
		{
			const u_int uNumPlayers = pxLobby->GetNumPlayers();
			const u_int uSufficientPlayers = GetSufficientPlayers();
			if (uNumPlayers < uSufficientPlayers)
			{
				uWaitingForPlayers = uSufficientPlayers - uNumPlayers;
			}
		}
		Asura_Lobby_System::SetNumPlayersBeingWaitedFor(uWaitingForPlayers);
	}
}

// state machine functions
void Asura_Matchmaking_System::ChangeState(ASURA_MATCHMAKING_STATE eState)
{
	Asura_Assert(s_bInitialised, "not intialised");

	if (eState == s_eMatchmakingState)
	{
		Asura_Assert(false, "can't change to same state");
		return;
	}
	
	ASURA_MATCHMAKING_TRACE(("state = %s\n", MatchmakingStateToString(eState)));

	s_eMatchmakingState = eState;

	switch (s_eMatchmakingState)
	{
		case ASURA_MATCHMAKING_STATE_NONE:
		case ASURA_MATCHMAKING_STATE_ERROR:
		{
			break;
		}
		case ASURA_MATCHMAKING_STATE_SEARCHING:
		{
			EnterState_Searching();
			break;
		}
		case ASURA_MATCHMAKING_STATE_CONNECTING:
		{
			EnterState_Connecting();
			break;
		}
		case ASURA_MATCHMAKING_STATE_HOSTING:
		{
			EnterState_Hosting();
			break;
		}
		case ASURA_MATCHMAKING_STATE_CHOOSING:
		{
			EnterState_Choosing();
			break;
		}
		case ASURA_MATCHMAKING_STATE_CREATING:
		{
			EnterState_Creating();
			break;
		}
		case ASURA_MATCHMAKING_STATE_WAITING:
		{
			EnterState_Waiting();
			break;
		}
		case ASURA_MATCHMAKING_STATE_FINISHED:
		{
			EnterState_Finished();
			break;
		}
		default:
		{
			Asura_Assert(false, "unhandled state");
			break;
		}
	}
}

void Asura_Matchmaking_System::EnterState_Searching()
{
	Asura_Assert(s_bInitialised, "not intialised");

	// clear array of attempted lobbies
	for (u_int u = 0; u < ASURA_LOBBY_MAX_SEARCH_RESULTS; ++u)
	{
		s_pbLobbyAttempted[u] = false;
	}

	// set to use the first skill threshold
	s_uSkillThreshold = s_auSkillThresholds[0];
	s_uCurrentSkillthresholdIndex = 0;

	PerformSearch();
}

void Asura_Matchmaking_System::ExecuteState_Searching()
{
	if (Asura_Lobby_System::GetSearchState() == ASURA_LOBBY_SEARCHSTATE_SEARCH_COMPLETE)
	{
		// set timestamp for search results
		s_uSearchResultsTimestamp = Asura_Timers::GetMSecsSinceStart();

		ASURA_MATCHMAKING_TRACE(("Got %u search results\n", Asura_Lobby_System::GetNumSearchResults()));

		// if no results were returned, retry search with less restritive skill 
		// filter. This is not necessary when using TrueSkill on 360 and will not happen since 
		// uASURA_MATCHMAKING_NUM_SKILLTHRESHOLDS = 1 in that case.
		if (!Asura_Lobby_System::GetNumSearchResults() && 
			s_uCurrentSkillthresholdIndex + 1 < uASURA_MATCHMAKING_NUM_SKILLTHRESHOLDS)
		{
			s_uSkillThreshold = s_auSkillThresholds[s_uCurrentSkillthresholdIndex++];
			PerformSearch();
		}
		else
		{
			ChangeState(ASURA_MATCHMAKING_STATE_CHOOSING);
		}
	}
}

void Asura_Matchmaking_System::EnterState_Choosing()
{
	// clear the best lobby
	s_uBestLobby = ASURA_LOBBY_MAX_SEARCH_RESULTS;
}

void Asura_Matchmaking_System::ExecuteState_Choosing()
{
	if (TimeLimitExpired(s_uSearchResultsTimestamp, s_uSearchResultsStaleTime))
	{
		// refresh the search results
		ChangeState(ASURA_MATCHMAKING_STATE_SEARCHING);
	}
	else
	{
		s_uBestLobby = PickBestLobby();
		if (s_uBestLobby < ASURA_LOBBY_MAX_SEARCH_RESULTS)
		{
			// got a lobby to join
			// if we got here by giving up our own hosting session, we need
			// to end that now. TODO: shutdown a bit more gracefully?
			if (Asura_Lobby_System::IsHost())
			{
				Asura_Lobby_System::LeaveLobby(false);
			}
			ChangeState(ASURA_MATCHMAKING_STATE_CONNECTING);
		}
		else
		{
			// no lobbies - host our own
			ChangeState(ASURA_MATCHMAKING_STATE_HOSTING);
		}
	}
}

void Asura_Matchmaking_System::EnterState_Connecting()
{
	Asura_Assert(s_uBestLobby < ASURA_LOBBY_MAX_SEARCH_RESULTS, "invalid lobby");
	
	s_bConnectingStateHasJoinedLobby = false;
}

// The connecting state should tell the lobby system to join the best lobby, then wait for it to change its state and change its own state depending on the lobby's.
void Asura_Matchmaking_System::ExecuteState_Connecting()
{
	const ASURA_LOBBY_STATE eLobbyState = Asura_Lobby_System::GetState();
	if ( !s_bConnectingStateHasJoinedLobby )
	{
		// Dont join a lobby if the lobby_system is terminating, wait for it to finish that first
		if ( eLobbyState != ASURA_LOBBY_STATE_TERMINATING ) 
		{
			if (s_uBestLobby < ASURA_LOBBY_MAX_SEARCH_RESULTS)
			{
				// record this connection attempt and try to join
				s_pbLobbyAttempted[s_uBestLobby] = true;
				Asura_Lobby_System::JoinLobby(s_uBestLobby);

				// flag the lobby as automatching, so clients know to wait in the automatch state.
				// the host may change this value later.
				Asura_Lobby_System::SetAutomatching(true);

				s_bConnectingStateHasJoinedLobby = true;
			}
		}
	}
	else // ( !s_bConnectingStateHasJoinedLobby )
	{
		switch (eLobbyState)
		{
			case ASURA_LOBBY_STATE_INVALID:
			{
				// error connecting, go back and choose another one.
				ChangeState(ASURA_MATCHMAKING_STATE_CHOOSING);
				break;
			}
			
			case ASURA_LOBBY_STATE_PRE_GAME:
			{
				ChangeState(ASURA_MATCHMAKING_STATE_WAITING);
				break;
			}

			case ASURA_LOBBY_STATE_STARTING_GAME:
			case ASURA_LOBBY_STATE_PLAYING_GAME:
			{
				ChangeState(ASURA_MATCHMAKING_STATE_FINISHED);
				break;
			}

			default:
			{
				break;
			}
		}
	}
}

void Asura_Matchmaking_System::EnterState_Hosting()
{
	s_uHostingStateEnterTimestamp = Asura_Timers::GetMSecsSinceStart();
}

void Asura_Matchmaking_System::ExecuteState_Hosting()
{
	// if we went off searching for games while hosting, we
	// might not need to create a lobby
	if (!Asura_Lobby_System::IsHost())
	{
		ChangeState(ASURA_MATCHMAKING_STATE_CREATING);
	}
	else
	{
		if (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID)
		{
			ChangeState(ASURA_MATCHMAKING_STATE_ERROR);
		}
		else if (TimeLimitExpired(s_uHostingStateEnterTimestamp, s_uTimeToWaitForSufficientPlayers))
		{
			ReduceMatchQualityThreshold();
			ChangeState(ASURA_MATCHMAKING_STATE_SEARCHING);
		}
		else if (LobbyHasSufficientPlayers() || !Asura_Lobby_System::IsAutomatching())
		{
			ChangeState(ASURA_MATCHMAKING_STATE_FINISHED);
		}
	}
}

void Asura_Matchmaking_System::EnterState_Creating()
{
	Project_CreateLobby();

	// flag the lobby as automatching, so clients know to wait in the automatch state.
	Asura_Lobby_System::SetAutomatching(true);

	// record time this hosting attempt started
	s_uHostingStartTimestamp = Asura_Timers::GetMSecsSinceStart();
}

void Asura_Matchmaking_System::ExecuteState_Creating()
{
	if (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID)
	{
		ChangeState(ASURA_MATCHMAKING_STATE_ERROR);
	}
	else if (Asura_Lobby_System::GetState() != ASURA_LOBBY_STATE_CREATING)
	{
		ChangeState(ASURA_MATCHMAKING_STATE_HOSTING);
	}
}

void Asura_Matchmaking_System::EnterState_Waiting()
{
	// nothing to do
}

void Asura_Matchmaking_System::ExecuteState_Waiting()
{
	// if host has forced game to start, lobby system will no longer be automatching
	if (Asura_Lobby_System::GetLobby() && !Asura_Lobby_System::IsAutomatching())
	{
		ChangeState(ASURA_MATCHMAKING_STATE_FINISHED);
		return;
	}

	const ASURA_LOBBY_STATE eLobbyState = Asura_Lobby_System::GetState();
	switch (eLobbyState)
	{
		case ASURA_LOBBY_STATE_INVALID:
		{
			// host quit/connection lost
			// TODO: host migration.
			ReduceMatchQualityThreshold();
			ChangeState(ASURA_MATCHMAKING_STATE_SEARCHING);
			break;
		}

		case ASURA_LOBBY_STATE_STARTING_GAME:
		case ASURA_LOBBY_STATE_PLAYING_GAME:
		{
			ChangeState(ASURA_MATCHMAKING_STATE_FINISHED);
			break;
		}

		default:
		{
			break;
		}
	}
}

void Asura_Matchmaking_System::EnterState_Finished()
{
	Asura_Lobby_System::SetAutomatching(false);
}

// utility functions
void Asura_Matchmaking_System::PerformSearch()
{
	ASURA_MATCHMAKING_TRACE(("Perform Search with Skill threshold = %u\n", s_uSkillThreshold));

	// get project code to fill out the required search parameters
	Asura_Lobby_SearchCriterion axSearchFilters[ASURA_LOBBY_MAX_SEARCH_CRITERIA];
	u_int uNumSearchFilters = 0;
	Project_SetSearchFilters( s_uSkillThreshold, axSearchFilters, uNumSearchFilters );
	Asura_Assert(uNumSearchFilters > 0, "didn't specify all search filters");
	Asura_Assert(uNumSearchFilters <= ASURA_LOBBY_MAX_SEARCH_CRITERIA, "memory overrun");

	Asura_Lobby_System::SearchForLobby(axSearchFilters, uNumSearchFilters);
}

u_int Asura_Matchmaking_System::PickBestLobby()
{
	const u_int uNumSearchResults = Asura_Lobby_System::GetNumSearchResults();
	u_int uIndexToJoin = ASURA_LOBBY_MAX_SEARCH_RESULTS; // invalid value
	float fBestMatchQuality = -1.0f;
	for (u_int u = 0; u < uNumSearchResults; ++u)
	{
		// check if we've already tried and failed to join this lobby
		if (s_pbLobbyAttempted[u])
		{
			continue;
		}
		// pick the one with the best match quality, if it has some room in it
		const u_int uNumPlayers = Asura_Lobby_System::GetSearchResultByIndex(u).m_uNumPlayers;
		const u_int uMaxPlayers = Asura_Lobby_System::GetSearchResultByIndex(u).m_uMaxPlayers;
		Asura_Assert(uNumPlayers < uMaxPlayers, "why are search results returning full lobbies?");
		
		const float fMatchQuality = ComputeMatchQuality(u);
		if (uNumPlayers < uMaxPlayers && 
			fMatchQuality >= s_fMatchQualityThreshold && 
			fMatchQuality > fBestMatchQuality)
		{
			uIndexToJoin = u;
			fBestMatchQuality = fMatchQuality;
		}
	}

	ASURA_MATCHMAKING_TRACE(("PickBestLobby() from %u...\n", uNumSearchResults));
	if (uIndexToJoin == ASURA_LOBBY_MAX_SEARCH_RESULTS)
	{
		ASURA_MATCHMAKING_TRACE(("Failed\n"));
	}
	else
	{
		ASURA_MATCHMAKING_TRACE(("Found with MQ = %.02f\n", fBestMatchQuality));
	}

	return uIndexToJoin;
}

float Asura_Matchmaking_System::ComputeMatchQuality( u_int uLobbyIndex )
{
	return ComputeMatchQuality( Asura_Lobby_System::GetSearchResultByIndex( uLobbyIndex ).m_xAttributes );
}

float Asura_Matchmaking_System::ComputeMatchQuality( const Asura_Lobby_Attributes& xAttributes )
{
	// Match Quality = 1.0f - ( |(lobby average skill) - (player skill)| / (max skill) );
	const int iMySkill = Asura_Lobby_System::GetLocalPlayerSkill();
	const int iLobbySkill = Project_GetLobbySkill( xAttributes );
	const int iMaxSkill = ASURA_LOBBY_MAX_SKILL;
	Asura_Assert(iMySkill >= 0 && iMySkill <= iMaxSkill, "invalid player skill");
	Asura_Assert(iLobbySkill >= 0 && iLobbySkill <= iMaxSkill, "invalid lobby skill");
	const float fSkillDiff = Asura_Maths::Abs(static_cast<float>(iLobbySkill) - static_cast<float>(iMySkill));
	const float fMQ = 1.0f - (fSkillDiff / static_cast<float>(iMaxSkill));
	return Asura_Maths::ClampToRange(fMQ, 0.0f, 1.0f);
}

bool Asura_Matchmaking_System::TimeLimitExpired( u_int uStartTime, u_int uTimeLimit )
{
	const u_int uCurrentTime = Asura_Timers::GetMSecsSinceStart();
	return uCurrentTime > (uStartTime + uTimeLimit);
}

void Asura_Matchmaking_System::ReduceMatchQualityThreshold()
{
	s_fMatchQualityThreshold -= s_fMatchQualityThresholdIncrement;
	ASURA_MATCHMAKING_TRACE(("New MQT = %.02f\n", s_fMatchQualityThreshold));
}

bool Asura_Matchmaking_System::IsWaitingForPlayers()
{ 
	if (s_bInitialised && Asura_Lobby_System::IsHost() && !HasErrorOccurred())
	{
		return true;
	}
	else
	{
		return s_eMatchmakingState == ASURA_MATCHMAKING_STATE_HOSTING || s_eMatchmakingState == ASURA_MATCHMAKING_STATE_WAITING; 
	}
}

u_int Asura_Matchmaking_System::GetNumPlayersBeingWaitedFor()
{
	return Asura_Lobby_System::GetNumPlayersBeingWaitedFor();
}

bool Asura_Matchmaking_System::LobbyHasSufficientPlayers()
{
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	Asura_Assert(pxLobby, "no lobby");
	if (pxLobby)
	{
		const u_int uNumPlayers = pxLobby->GetNumPlayers();
		const u_int uSufficientPlayers = GetSufficientPlayers();
		return uNumPlayers >= uSufficientPlayers;
	}
	return false;
}

u_int Asura_Matchmaking_System::GetSufficientPlayers()
{
	Asura_Assert(s_uMinPlayers <= s_uMaxPlayers, "invalid min/max players");
	Asura_Assert(s_fMinPlayersTime >= 0.0f && s_fMaxPlayersTime >= 0.0f, "invalid min/max players times");

	// SufficientPlayers is defined as follows:
	// SufficientPlayers = MinPlayers + (MaxPlayers - MinPlayers) * 
	//   ((MaxTime - MinPlayersTime) - CurrentTime) / (MaxTime - (MinPlayersTime + MaxPlayersTime));
	// SufficientPlayers = ClampToRange(SufficientPlayers, MinPlayers, MaxPlayers)
	const u_int uCurrentTime = Asura_Timers::GetMSecsSinceStart();
	Asura_Assert(uCurrentTime >= s_uHostingStartTimestamp, "invalid time stamp");
	const float fElapsedTime = static_cast<float>(static_cast<int>(uCurrentTime - s_uHostingStartTimestamp)) / 1000.0f; // convert to seconds
	float fMaxTime = static_cast<float>(static_cast<int>(s_uTimeToWaitForSufficientPlayers)) / 1000.0f; // convert to seconds
	Asura_Assert(s_fMinPlayersTime + s_fMaxPlayersTime <= fMaxTime + 1.0f, "invalid min/max players times");
	fMaxTime = Asura_Maths::Max(fMaxTime, s_fMinPlayersTime + s_fMaxPlayersTime + 1.0f);//prevent divide by zero
	const float fMaxMinusMinPlayers = static_cast<float>(static_cast<int>(s_uMaxPlayers - s_uMinPlayers));
	const float fTemp = fMaxMinusMinPlayers * ((fMaxTime - s_fMinPlayersTime) - fElapsedTime) / 
		(fMaxTime - (s_fMinPlayersTime + s_fMaxPlayersTime));
	int iSufficientPlayers = static_cast<int>(s_uMinPlayers) + Asura_Maths::FloatToInt_Near(fTemp);
	iSufficientPlayers = Asura_Maths::ClampToRange(iSufficientPlayers, static_cast<int>(s_uMinPlayers), static_cast<int>(s_uMaxPlayers));
	return static_cast<u_int>(iSufficientPlayers);
}

void Asura_Matchmaking_System::ForceGameStart()
{
	Asura_Assert(Asura_Lobby_System::IsHost() && s_bInitialised, "only expected when host");
	if (Asura_Lobby_System::IsHost() && s_bInitialised)
	{
		ChangeState(ASURA_MATCHMAKING_STATE_FINISHED);
	}
}

bool Asura_Matchmaking_System::ConsoleCmd_BeginAutomatch( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Matchmaking_System::Start();
	}
	return true;
}

bool Asura_Matchmaking_System::ConsoleCmd_CancelAutomatch( bool bHelp )
{
	if (!bHelp)
	{
		Cancel();
	}
	return true;
}

bool Asura_Matchmaking_System::ConsoleCmd_ForceAutomatchStart( bool bHelp )
{
	if (!bHelp)
	{
		ForceGameStart();
	}
	return true;
}

bool Asura_Matchmaking_System::ConsoleCmd_IsAutomatchComplete( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Assert(s_bInitialised, "not intialised");
		Asura_CommandConsole::Printf( IsFinished() ? "1" : "0" );
	}
	return true;
}

bool Asura_Matchmaking_System::ConsoleCmd_PrintAutomatchStatus( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Assert(s_bInitialised, "not intialised");
		Project_PrintAutomatchStatusToConsole();
	}
	return true;
}


#endif // #ifdef ASURA_SUPPORT_LOBBY
