//
// Asura_Physics.h
//
// Created by KJFL - 2004-03-17 11:22:07
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_PHYSICS_H__
#define __ASURA_PHYSICS_H__

#include "Core/Asura_Handle.h"
#include "Core/Asura_Maths.h"
#include "Mesh/Asura_Mesh_PolygonFlags.h"

// enable physics debugging whenever platform-independent profiling is on.
// If a platforms wants profiling, but doesn't have that define, it can set
// this directly.
#ifndef ASURA_PHYSICS_ENABLE_DEBUG_INFO
#ifdef ASURA_USE_PROFILING_SYSTEM
#define ASURA_PHYSICS_ENABLE_DEBUG_INFO	(1)
#endif
#endif // ASURA_PHYSICS_ENABLE_DEBUG_INFO

class Asura_Collision_Object;
class Asura_Physics_Island;


class Asura_Physics_System
{
	public:
		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

		static void Update( void );
		static void UpdateThread( void );

#ifdef ASURA_TOOLS
        static void UpdateForWaypointing( void );
#endif

#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
		static bool GetDisplayDebugInfo( void ) 
		{
			return s_bDisplayDebugInfo;
		}

		static void SetDisplayDebugInfo( bool bStatus )
		{
			s_bDisplayDebugInfo = bStatus;
		}

		static bool GetDisplayDebugEnvCaches( void )
		{
			return s_bDisplayDebugEnvCaches;
		}

		static bool GetDisplayBroadPhaseBBs( void )
		{
			return s_bDisplayDebugBroadphaseBBs;
		}
		
		// GetNumActiveActors() - not including vehicle actors
		static u_int GetNumActiveActors( void )
		{
			return s_uNumActiveActors;
		}
		
		static u_int GetNumActiveVehicleActors( void )
		{
			return s_uNumActiveVehicleActors;
		}
#else // #ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
		static bool GetDisplayDebugInfo( void ) 
		{
			return false;
		}

		static void SetDisplayDebugInfo( bool bStatus )
		{
		}

		static bool GetDisplayDebugEnvCaches( void )
		{
			return false;
		}

		static bool GetDisplayBroadPhaseBBs( void )
		{
			return false;
		}
#endif // else not #ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO

        static float GetTimeStep( void );

		static float GetTimeSlice( void )
		{
			return s_fPhysicsTime;
		}

		static float GetPreviousTimeSlice( void )
		{
			return s_fPreviousPhysicsTime;
		}

		static float GetTimeCorrectionScale( void )
		{
			return s_fTimeCorrectionScale;
		}

		static bool IsFirstUpdateThisFrame( void )
		{
			return s_bFirstUpdateThisFrame;
		}

		static float GetTimeLeft( void );

		// The physics update count is a simple integer that is
		// incremented every time the Physics does an update pass.
		// Note that this may not be the same as the frame number
		// count, and so if we want to use 'frame stamps' in the 
		// Physics code (e.g. to lazy-evaluate things) we need
		// to use this update count instead.
		static u_int GetPhysicsUpdateCount( void )
		{
			return s_uPhysicsUpdateCount;
		}

		static float GetGravityScalar( void );
		
		static void NotifyOfChangeToStaticObjects( void )
		{
			s_bStaticObjectsChanged = true;
		}

		static void UpdateBroadphaseIfStaticObjectsHaveChanged( void );

		// "global" gravity direction all objects use by default;
		// default setting is straight down (along +ve Y axis)
		static void SetGravityDirection( const Asura_AVector3& xDirection );
		static const Asura_AVector3& GetGravityDirection( void );

		static bool DoesOBBIntersectStaticObjects( const Asura_OBB& xOBB, Asura_Handle iHandleOfObjectToIgnore );
		static bool DoesOBBIntersectDynamicObjects( const Asura_OBB& xOBB, Asura_Handle iHandleOfObjectToIgnore );

		// See if an object (defined by an Asura_Collision_Object) can be placed at a position without intersecting anything
		// Callers can pass poly flags (e.g. AsuraPolygonFlag_DontRender) which indicate which polys to ignore
		// PHYSICSTODO - should really be able to test against objects too, and supply an ObjectToIgnore handle
		// AJG - 07/07/06 - Added flags to test for, to enable skipping polys with 0 flags. Currently, leaving as 0xFFFF will behave as before (so 0-flags aren't ignored), and anything other value will use them (and so 0-flags will be ignored).
		static bool CanPlaceObjectAt( const Asura_Vector_3& xPosition, const Asura_Collision_Object& xObject,
			const u_short usFlagsOfPolysToIgnore = 0, const Asura_Handle iHandleOfObjectToIgnore = ASURA_INVALID_HANDLE,
			const u_short usFlagsOfPolysToTestFor = 0xFFFF, bool bConsiderStaticObjects = false,
			bool bConsiderDynamicObjects = false );

		// AJG - 10/07/06 - Orientated version, needed for Gun's horses (which use orientated physics objects).
		static bool CanPlaceObjectAt( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, 
			const Asura_Collision_Object& xObject, const u_short usFlagsOfPolysToIgnore = 0, 
			const Asura_Handle iHandleOfObjectToIgnore = ASURA_INVALID_HANDLE, const u_short usFlagsOfPolysToTestFor = 0xFFFF,
			bool bConsiderStaticObjects = false, bool bConsiderDynamicObjects = false );

		// AJG - 06/08/07 - Sphere stack test similar to above. At this time, only tests against the environment.
		// The sphere stack is generated from the passed-in bounding box. The penetration value allows a little leeway in situations where it doesn't matter if there's a small intersection (which is most of the time).
		static bool CanPlaceSphereStackAt(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, u_short usFlagsOfPolysToIgnore = 0, float fAllowedPenetration = 0.0f);

		static bool MoveParticle( Asura_Vector_3& xPosition, Asura_Vector_3& xVelocity, float fElasticity, u_short usFlagsOfPolysToIgnore=AsuraPolygonFlag_NoCollisions );

		static void DebuggingRender( void );

		////////////////////////////////////////////////////////////
		//  
		////////////////////////////////////////////////////////////

		static bool RayIntersectsTri(const Asura_AVector3& alpha, const Asura_AVector3& beta, const Asura_AVector3& v0, const Asura_AVector3& v1, const Asura_AVector3& v2, bool& facingToward, float& distance);
		static bool SmallestVectorBetweenTwoLines(const Asura_AVector3 &a1, const Asura_AVector3 &a2, const Asura_AVector3 &b1, const Asura_AVector3 &b2, float& out_along_a, float& out_along_b);
		
		// test for collision between a moving sphere and a stationary edge
		static bool SweptSphereVsEdge(  const float fSphereRadius, const Asura_AVector3& xSphereStart, const Asura_AVector3& xSphereDir, const Asura_AVector3& xLineStart, const Asura_AVector3& xLineEnd, Asura_AVector3& xCollisionPoint, float& fAllowedSphereMovement );

		static float SquaredDistanceBetweenLineSegments( const Asura_AVector3& xSeg0Start, const Asura_AVector3& xSeg0End, const Asura_AVector3& xSeg1Start, const Asura_AVector3& xSeg1End );


		static bool IsMultithreaded( void ) { return s_bMultithread; }

	private:
		static void UpdateIsland( const Asura_Physics_Island* pxIsland );

		#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
		static void AddConsoleCommands( void );
		#endif

		// two sub-tests which make up the SweptSphereVsEdge test
		static bool SweptSphereVsPoint( const float fSphereRadius, const Asura_AVector3& xSphereStart, const Asura_AVector3& xSphereDir, const Asura_AVector3& xPoint, Asura_AVector3& xCollisionPoint, float& fAllowedSphereMovement );
		static bool SweptSphereVsLineSegment( const float fSphereRadius, const Asura_AVector3& xSphereStart, const Asura_AVector3& xSphereDir, const Asura_AVector3& xLineStart, const Asura_AVector3& xLineEnd, Asura_AVector3& xCollisionPoint, float& fAllowedSphereMovement );

		static float	s_fPhysicsTime;
		static float	s_fPreviousPhysicsTime;
		static float	s_fTimeCorrectionScale;
		static u_int	s_uPhysicsUpdateCount;
		static bool		s_bFirstUpdateThisFrame;

		static bool		s_bStaticObjectsChanged;
		static bool		s_bMultithread;

#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO
		static bool		s_bDisplayDebugInfo;
		static bool		s_bDisplayDebugEnvCaches;
		static bool		s_bDisplayDebugBroadphaseBBs;
		static bool		s_bDisplayDebugCollisionFX;
		static u_int	s_uNumActiveActors;	// not including vehicle actors
		static u_int	s_uNumActiveVehicleActors;
#endif

		static Asura_AVector3	s_xGravityDirection;

};

#endif//__ASURA_PHYSICS_H__
