
// Asura_Physics_Broadphase.cpp
//
// Created by KJFL - 20 March 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_Broadphase.h"

#include "Physics/Asura_Physics_RigidBody.h" // only needed for brute force method; may be removable in time

#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_SpatialHash.h"

static Asura_Physics_SpatialHash g_xDynamicObjects_SpatialHash;
static Asura_Physics_SpatialHash g_xStaticObjects_SpatialHash;

void Asura_Physics_Broadphase_System::Reset( void )
{
	g_xDynamicObjects_SpatialHash.ResetBetweenLevels();
	g_xStaticObjects_SpatialHash.ResetBetweenLevels();
}

void Asura_Physics_Broadphase_System::ResetStaticBroadphase( void )
{
	g_xStaticObjects_SpatialHash.ResetBetweenLevels();
}

void Asura_Physics_Broadphase_System::ResetDynamicBroadphase( void )
{
	g_xDynamicObjects_SpatialHash.ResetBetweenLevels();
}

void Asura_Physics_Broadphase_System::AddObjectToStaticBroadphase( Asura_Physics_RigidBodyInstance* pxInstance, const bool bMakePairings, const bool bRememberObject )
{
	g_xStaticObjects_SpatialHash.AddObject( pxInstance, pxInstance->GetBroadphaseAABB(), bMakePairings, bRememberObject );
}

void Asura_Physics_Broadphase_System::AddObjectToDynamicBroadphase( Asura_Physics_RigidBodyInstance* pxInstance, const bool bMakePairings, const bool bRememberObject )
{
	g_xDynamicObjects_SpatialHash.AddObject( pxInstance, pxInstance->GetBroadphaseAABB(), bMakePairings, bRememberObject );

}

bool Asura_Physics_Broadphase_System::DoesOBBIntersectStaticObjects( const Asura_OBB& xOBB, Asura_Handle iHandleOfObjectToIgnore )
{
	Asura_Physics_System::UpdateBroadphaseIfStaticObjectsHaveChanged();

	return g_xStaticObjects_SpatialHash.DoesOBBIntersectAnyObjects( xOBB, iHandleOfObjectToIgnore );
}

bool Asura_Physics_Broadphase_System::DoesOBBIntersectDynamicObjects( const Asura_OBB& xOBB, Asura_Handle iHandleOfObjectToIgnore )
{
	// TODO: need to test if data needs refreshing because dynamic objects have been destroyed?

	return g_xDynamicObjects_SpatialHash.DoesOBBIntersectAnyObjects( xOBB, iHandleOfObjectToIgnore );
}


Asura_Physics_Broadphase_Query::Asura_Physics_Broadphase_Query()
: m_pxFirstEntry(0)
{
}

Asura_Physics_Broadphase_Query::~Asura_Physics_Broadphase_Query()
{
	Reset();
}

void Asura_Physics_Broadphase_Query::Reset( void )
{
	const Asura_Physics_Broadphase_Entry* pxEntry = m_pxFirstEntry;

	while ( pxEntry )
	{
		const Asura_Physics_Broadphase_Entry* pxNext = pxEntry->GetNextEntry();

		Asura_Physics_Broadphase_Entry::Destroy( pxEntry );

		pxEntry = pxNext;
	}

	m_pxFirstEntry = 0;
}


void Asura_Physics_Broadphase_Query::MakeQuery( const Asura_Bounding_Box& xBroadphaseAABB, bool bTestAgainstStatics, bool bTestAgainstExternalControl, bool bTestAgainstActors,  bool bTestAgainstDynamics )
{
	Reset();

	if ( bTestAgainstExternalControl )
	{
		g_xDynamicObjects_SpatialHash.BroadphaseQuery( xBroadphaseAABB, this, true );
	}
	if ( bTestAgainstStatics )
	{
		Asura_Physics_System::UpdateBroadphaseIfStaticObjectsHaveChanged();

		g_xStaticObjects_SpatialHash.BroadphaseQuery( xBroadphaseAABB, this, false );
	}

	if ( bTestAgainstDynamics )
	{
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			if ( pxInstance->IsActor() ) continue;
			if ( pxInstance->IsIgnore() ) continue;

			if ( pxInstance->GetBroadphaseAABB().Intersects( xBroadphaseAABB ) )
			{
				AddEntry( pxInstance );
			}
		}
	}

	if ( bTestAgainstActors )
	{
		// KJFL - 28 March 2007 - brute force search for Actors. The spatial hash can't be used here as it currently stands
		// since it may not be up to date (i.e. dynamic objects are moving during the course of these calls). For a game
		// with not many Actors, this brute force method might not actually be all that bad - though a separate Actor list
		// would definitely help. To be revisited as time allows / if it shows up high on profiles for those projects
		// using the feature.

		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			if ( !pxInstance->IsActor() ) continue;
			if ( pxInstance->IsIgnore() ) continue;

			if ( pxInstance->GetBroadphaseAABB().Intersects( xBroadphaseAABB ) )
			{
				AddEntry( pxInstance );
			}
		}
	}
}

void Asura_Physics_Broadphase_Query::MakeQuery_StandardActorMovement( const Asura_Bounding_Box& xBroadphaseAABB, bool bTestAgainstExternalControl, bool bTestAgainstActors,  bool bTestAgainstDynamics )
{
	Reset();

	if ( bTestAgainstExternalControl )
	{
		g_xDynamicObjects_SpatialHash.BroadphaseQuery( xBroadphaseAABB, this, true );
	}

	// always test statics
	{
		Asura_Physics_System::UpdateBroadphaseIfStaticObjectsHaveChanged();
		g_xStaticObjects_SpatialHash.BroadphaseQuery( xBroadphaseAABB, this, false );
	}
	
	// only test actors and dynamics if needed
	if ( bTestAgainstActors || bTestAgainstDynamics )
	{
		for( Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetDynamicList(); pxInstance; pxInstance = pxInstance->GetNextInstance() )
		{
			if ( pxInstance->IsIgnore() ) continue;

			if ( pxInstance->GetIgnoreDynamicObjects() ) continue;

			const bool bActor = pxInstance->IsActor();

			if ( bActor )
			{
				if ( !bTestAgainstActors ) continue;
			}
			else
			{
				if ( !bTestAgainstDynamics || !pxInstance->GetActorsDontPushMe() ) continue;
			}

			if ( pxInstance->GetBroadphaseAABB().Intersects( xBroadphaseAABB ) )
			{
				AddEntry( pxInstance );
			}
		}
	}
}



void Asura_Physics_Broadphase_Query::AddEntry( Asura_Physics_RigidBodyInstance* pxInstance )
{
	// first check to see if we already have this instance covered (we may need to optimise this if we have many entries)
	for ( const Asura_Physics_Broadphase_Entry* pxEntry = m_pxFirstEntry; pxEntry; pxEntry=pxEntry->GetNextEntry() )
	{
		if ( pxEntry->GetInstance() == pxInstance ) return;
	}

	// it's a new one; add a new entry to the head of the list
	Asura_Physics_Broadphase_Entry* pxNewEntry = Asura_Physics_Broadphase_Entry::Create( pxInstance, m_pxFirstEntry);

	m_pxFirstEntry = pxNewEntry;
}

// TODO : mempools or whatever works best here
Asura_Physics_Broadphase_Entry* Asura_Physics_Broadphase_Entry::Create( Asura_Physics_RigidBodyInstance* pxInstance, Asura_Physics_Broadphase_Entry* pxNextEntry )
{
	Asura_Physics_Broadphase_Entry* pxEntry = new Asura_Physics_Broadphase_Entry;

	pxEntry->m_pxInstance = pxInstance;
	pxEntry->m_pxNextEntry = pxNextEntry;

	return pxEntry;
}

void Asura_Physics_Broadphase_Entry::Destroy( const Asura_Physics_Broadphase_Entry* pxEntry )
{
	delete pxEntry;
}
