
// Asura_Physics_CollisionFX_Report.cpp
//
// Created by KJFL - 2 October 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_CollisionFX_Report.h"

#ifdef ASURA_USE_COLLISIONFX

#include "Core/Asura_Mempool.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Physics/Asura_Physics.h"

//////////////////////////////////////////////////////////////////////////////
// Asura_Physics_CollisionFX_Report 
//////////////////////////////////////////////////////////////////////////////
Asura_Physics_CollisionFX_Report::Asura_Physics_CollisionFX_Report()
: m_xPosition(Asura_Maths::ZeroAVector3)
, m_xRelVelocity(Asura_Maths::ZeroAVector3)
, m_xNormal(0.0f, -1.0f, 0.0f)
, m_uEntity1Guid(ASURA_GUID_UNREGISTERED)
, m_uEntity2Guid(ASURA_GUID_SERVER)
, m_uEnvironmentFaceID(0)
, m_pxNext(0)
{
}

//////////////////////////////////////////////////////////////////////////////
// Asura_Physics_CollisionFX_Report_System
//////////////////////////////////////////////////////////////////////////////

// Make in pools of 32 - which means each allocation is over 1kb.
static Asura_ResizingMempool g_xCollisionFXReportPool( 32, sizeof(Asura_Physics_CollisionFX_Report), __FILE__, __LINE__ );

Asura_Physics_CollisionFX_Report* Asura_Physics_CollisionFX_Report_System::s_pxFirstReport;

void Asura_Physics_CollisionFX_Report_System::Initialise( void )
{
	s_pxFirstReport = 0;
}

void Asura_Physics_CollisionFX_Report_System::Reset( void )
{
	ClearAllReports();
}

void Asura_Physics_CollisionFX_Report_System::Shutdown( void )
{
	ClearAllReports();
}

void Asura_Physics_CollisionFX_Report_System::DebugRender( void )
{
#ifdef ASURA_PHYSICS_ENABLE_DEBUG_INFO

#ifndef ASURA_NO_DEBUG_OUTPUT
	Asura_Physics_CollisionFX_Report_Iterator xIt;
	
	for ( xIt.Start(); xIt.NotFinished(); xIt.Next() )
	{
		const Asura_Physics_CollisionFX_Report* pxReport = xIt.GetCurrentReport();

		Asura_Vector_3 xStart( pxReport->GetPosition() );
		Asura_Vector_3 xEnd( pxReport->GetPosition() + pxReport->GetRelVelocity() );

		u_int uColour = 0x7f00ffff;

		Asura_DebugLineRender::AddArrow( xStart, xEnd, uColour, 0.125f, 0.01f );
	}
#endif // #ifndef ASURA_NO_DEBUG_OUTPUT

#endif	// ASURA_PHYSICS_ENABLE_DEBUG_INFO
}

// Add a new object-object CollisionFX report 
void Asura_Physics_CollisionFX_Report_System::AddReport( const Asura_AVector3& xPosition, const Asura_AVector3& xRelVelocity, Asura_Guid uEntity1Guid, Asura_Guid uEntity2Guid, const Asura_AVector3& xCollisionAxis  )
{
	// allocate a report
	Asura_Physics_CollisionFX_Report* pxReport = static_cast<Asura_Physics_CollisionFX_Report*>(g_xCollisionFXReportPool.Allocate()); 

	if ( !pxReport )
	{
		Asura_Assert( pxReport, "Code bug: out of memory?" );
		return;
	}

	// set the data held in the report
	pxReport->m_xPosition = xPosition;
	pxReport->m_xRelVelocity = xRelVelocity;
	pxReport->m_xNormal = xCollisionAxis;
	pxReport->m_uEntity1Guid = uEntity1Guid;
	pxReport->m_uEntity2Guid = uEntity2Guid;
    
	// clear the unused fields
	pxReport->m_uEnvironmentFaceID = 0;

	// put the report at the head of the list
	pxReport->m_pxNext = s_pxFirstReport;
	s_pxFirstReport = pxReport;
}

// Add a new object-environment CollisionFX report 
void Asura_Physics_CollisionFX_Report_System::AddReport( const Asura_AVector3& xPosition, const Asura_AVector3& xRelVelocity, const Asura_AVector3& xNormal, Asura_Guid uEntityGuid, u_int uEnvironmentFaceID )
{
	// allocate a report
	Asura_Physics_CollisionFX_Report* pxReport = static_cast<Asura_Physics_CollisionFX_Report*>(g_xCollisionFXReportPool.Allocate());

	if ( !pxReport )
	{
		Asura_Assert( pxReport, "Code bug: out of memory?" );
		return;
	}

	// set the data held in the report
	pxReport->m_xPosition = xPosition;
	pxReport->m_xRelVelocity = xRelVelocity;
	pxReport->m_xNormal = xNormal;
	pxReport->m_uEntity1Guid = uEntityGuid;
	pxReport->m_uEntity2Guid = ASURA_GUID_SERVER;
	pxReport->m_uEnvironmentFaceID = uEnvironmentFaceID;

	// put the report at the head of the list
	pxReport->m_pxNext = s_pxFirstReport;
	s_pxFirstReport = pxReport;
}

// Clear all reports in the most efficient manner (i.e. avoid single deallocations)
void Asura_Physics_CollisionFX_Report_System::ClearAllReports( void )
{
	g_xCollisionFXReportPool.ForceReset();
	s_pxFirstReport = 0;
}


// KJFL - 2 October 2007 - Report iterator methods. 
// May make these inline in the header file if we stick with the mempool implementation.

Asura_Physics_CollisionFX_Report_Iterator::Asura_Physics_CollisionFX_Report_Iterator()
{
	m_pxCurrent = 0;
}

// Jump to the first report held and make that the current one.
void Asura_Physics_CollisionFX_Report_Iterator::Start( void )
{
	m_pxCurrent = Asura_Physics_CollisionFX_Report_System::GetFirstReport();
}

// Have we finished? (Is current report pointer null?)
bool Asura_Physics_CollisionFX_Report_Iterator::NotFinished( void )
{
	return (m_pxCurrent != 0);
}

// Make the current report the next in the list.
void Asura_Physics_CollisionFX_Report_Iterator::Next( void )
{
	if ( m_pxCurrent )
	{
		m_pxCurrent = m_pxCurrent->GetNext();
	}
}

// Get pointer to current report.
const Asura_Physics_CollisionFX_Report* Asura_Physics_CollisionFX_Report_Iterator::GetCurrentReport( void )
{
	return m_pxCurrent;
}
#endif // ASURA_USE_COLLISIONFX
