//
// Asura_Physics_CollisionFX_Report.h
//
// Created by KJFL - 2 October 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_COLLISIONFX_REPORT_H__
#define __ASURA_PHYSICS_COLLISIONFX_REPORT_H__

#ifdef ASURA_USE_COLLISIONFX

#include "Core/Asura_Maths.h"
#include "Network/Asura_Guid.h"

class Asura_Physics_CollisionFX_Report;

// Report iterator. Hides the actual mechanism by which reports are stored (e.g. big array, linked list, etc).
//
// Intended use is like so:
//
// Asura_Physics_CollisionFX_Report_Iterator xIt;
//
// for ( xIt.Start(); xIt.NotFinished(); xIt.Next() )
// {
//		const Asura_Physics_CollisionFX_Report* pxReport = xIt.GetCurrentReport();
//
//		// ... use pxReport as required
// }
//
class Asura_Physics_CollisionFX_Report_Iterator
{
	public:
		Asura_Physics_CollisionFX_Report_Iterator();
		~Asura_Physics_CollisionFX_Report_Iterator() {}

		// Jump to the first report held and make that the current one.
		void Start( void );

		// Have we finished? (Is current report pointer null?)
		bool NotFinished( void );

		// Make the current report the next in the list.
		void Next( void );

		// Get pointer to current report.
		const Asura_Physics_CollisionFX_Report* GetCurrentReport( void );	

	private:

		const Asura_Physics_CollisionFX_Report* m_pxCurrent;
};


class Asura_Physics_CollisionFX_Report
{
	public:

		// Helper methods to query whether this is an object-object or an object-environment report.
		bool IsObjectObjectCollision( void ) const { return (m_uEntity2Guid != ASURA_GUID_SERVER); }
		bool IsEnvironmentCollision( void ) const { return (m_uEntity2Guid == ASURA_GUID_SERVER); }

		// Get entity guid of the first object (for environment collisions, this is the only object).
		Asura_Guid GetObject1EntityGuid( void ) const { return m_uEntity1Guid; }

		// Get entity guid of the second object (this is ASURA_GUID_SERVER for object-environment collisions).
		Asura_Guid GetObject2EntityGuid( void ) const { return m_uEntity2Guid; }

		// Get the average position of the collision (since there may be multiple contact points).
		const Asura_AVector3& GetPosition( void ) const { return m_xPosition; }

		// Get relative velocity of the collision, defined as velocity of object 1 minus velocity of object 2.
		// For environment collisions this just means the velocity of the object.
		const Asura_AVector3& GetRelVelocity( void ) const { return m_xRelVelocity; }

		// Return normal of environment face (if it's an environment collision).
		// Return default upwards vector if the normal hasn't been set (e.g. for physics object-object collisions).
        const Asura_AVector3& GetNormal( void ) const { return m_xNormal; }

		// Get ID of the environment face (if it's an environment collision). 
		// This can be used with Asura_Physics_Environment to get more info.
		u_int GetEnvironmentFaceID( void ) const { return m_uEnvironmentFaceID; }


		// MGH 9/10/2007: Systems other than the physics system need to be able to create these objects
		// and set the variables, so I've added a constructor and some setter functions.
		Asura_Physics_CollisionFX_Report();
		~Asura_Physics_CollisionFX_Report() {}

		void SetPosition(const Asura_AVector3& xPosition)  { m_xPosition    = xPosition; }
		void SetRelVelocity(const Asura_AVector3& xRelVel) { m_xRelVelocity = xRelVel; }
		void SetNormal(const Asura_AVector3& xNormal)      { m_xNormal      = xNormal; }
		
		void SetPosition(const Asura_Vector_3& xPosition)  { m_xPosition    = Asura_AVector3(xPosition); }
		void SetRelVelocity(const Asura_Vector_3& xRelVel) { m_xRelVelocity = Asura_AVector3(xRelVel); }
		void SetNormal(const Asura_Vector_3& xNormal)      { m_xNormal      = Asura_AVector3(xNormal); }

		void SetObjectObjectCollision(Asura_Guid uEntity1Guid, Asura_Guid uEntity2Guid)
		{
			m_uEntity1Guid = uEntity1Guid;
			m_uEntity2Guid = uEntity2Guid;
			m_uEnvironmentFaceID = 0;
		}

		void SetEnvironmentCollision(Asura_Guid uEntity1Guid, u_int uFaceID)
		{
			m_uEntity1Guid = uEntity1Guid;
			m_uEntity2Guid = ASURA_GUID_SERVER;
			m_uEnvironmentFaceID = uFaceID;
		}

	private: 
		
		friend class Asura_Physics_CollisionFX_Report_Iterator;
		friend class Asura_Physics_CollisionFX_Report_System;

		// Method to access next report in list - private since mechanism hidden by
		// companion iterator class.
		const Asura_Physics_CollisionFX_Report* GetNext( void ) const { return m_pxNext; }

		// Aligned members first for memory efficient layout
		Asura_AVector3	m_xPosition;
		Asura_AVector3	m_xRelVelocity;

		Asura_AVector3	m_xNormal;

		Asura_Guid		m_uEntity1Guid;
		Asura_Guid		m_uEntity2Guid;
		u_int			m_uEnvironmentFaceID;

		// next ptr, if we use mempools with singly-linked lists? 
		// Looks like 64 bytes with padding if AVectors are 16bytes?
		Asura_Physics_CollisionFX_Report*	m_pxNext;

};


// System class - should only be called by the internals of the Physics itself
class Asura_Physics_CollisionFX_Report_System
{
	public:
		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

		// Add a new object-object CollisionFX report 
		static void AddReport( const Asura_AVector3& xPosition, const Asura_AVector3& xRelVelocity, Asura_Guid uEntity1Guid, Asura_Guid uEntity2Guid, const Asura_AVector3& xCollisionAxis );
	
		// Add a new object-environment CollisionFX report 
		static void AddReport( const Asura_AVector3& xPosition, const Asura_AVector3& xRelVelocity, const Asura_AVector3& xNormal, Asura_Guid uEntityGuid, u_int uEnvironmentFaceID );

		// Clear all reports in the most efficient manner (i.e. avoid single deallocations)
		static void ClearAllReports( void );

		static void DebugRender( void );

	private:
		
		friend class Asura_Physics_CollisionFX_Report_Iterator;

		static const Asura_Physics_CollisionFX_Report* GetFirstReport( void ) { return s_pxFirstReport; }
		static Asura_Physics_CollisionFX_Report* s_pxFirstReport;
};

#endif //#ifdef ASURA_USE_COLLISIONFX

#endif//__ASURA_PHYSICS_COLLISIONFX_REPORT_H__


