//
// Asura_Physics_ContactCluster.h
//
// Created by KJFL - 22 November 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_CONTACTCLUSTER_H__
#define __ASURA_PHYSICS_CONTACTCLUSTER_H__

#include "Core/Asura_Maths.h"

static const u_int uASURA_CONTACTCLUSTER_MAXCONTACTS = 15;

class Asura_Physics_ContactCluster
{
	public:

		u_int GetNumberOfContacts( void ) const
		{
			return m_uNumberOfContacts;
		}

		const Asura_AVector3& GetContactNormal( void ) const
		{
			return m_xNormal;
		}

		const Asura_AVector3& GetContactPosition( u_int uContact ) const
		{
			Asura_Assert( uContact < m_uNumberOfContacts, "Code bug: index out of range" );
			return m_axContactPositions[ uContact ];
		}

		void GetAverageContactPosition( Asura_AVector3& xAverageContact ) const 
		{
			xAverageContact.Zero();

			if ( m_uNumberOfContacts )
			{
				for (u_int u=0; u<m_uNumberOfContacts; ++u)
				{
					xAverageContact += m_axContactPositions[u];
				}
				xAverageContact /= (float)m_uNumberOfContacts;
			}
		}

		const Asura_Physics_ContactCluster* GetNextCluster( void ) const
		{
			return m_pxNextCluster;
		}

		float GetPlanePenetrationConstant( void ) const
		{
			return m_fPlanePenetrationConstant;
		}

	private:

		friend class Asura_Physics_ContactCluster_List;

		Asura_Physics_ContactCluster(){};
		~Asura_Physics_ContactCluster(){};

		static Asura_Physics_ContactCluster* Create( void );

		static void Destroy( Asura_Physics_ContactCluster* pxCluster );

		bool AddContactPosition( const Asura_AVector3& xBodyPosition, const Asura_AVector3& xPosition, const Asura_AVector3& xNormal, float fPenetration )
		{
			if (m_uNumberOfContacts)
			{
				// TODO - may need to tweak epsilon
				if ( !xNormal.ApproximatelyEquals( m_xNormal ) )
				{
					return false;
				}

				// TODO - remove hardcoded limits if possible? (linked list?)
				if ( m_uNumberOfContacts >= uASURA_CONTACTCLUSTER_MAXCONTACTS )
				{
					// KJFL - 12 February = if we've filled this cluster it'll just make a new one
					//Asura_Assert( false, "Code bug: contact cluster limits hit");
					return false;
				}
			}
			else
			{
				// for first timers, set the normal
				m_xNormal = xNormal;
			}

			m_fPlanePenetrationConstant = Asura_Maths::Max( m_fPlanePenetrationConstant, xBodyPosition*m_xNormal + fPenetration );

			bool bNewPoint = true;
			for (u_int u=0; u<m_uNumberOfContacts; ++u)
			{
				if ( m_axContactPositions[u].ApproximatelyEquals(xPosition) )
				{
					bNewPoint = false;
					break;
				}
			}
			
			if (bNewPoint)
			{
				m_axContactPositions[ m_uNumberOfContacts++ ] = xPosition;
			}

            return true;
			
		}

		Asura_Physics_ContactCluster*	m_pxNextCluster;
		u_int							m_uNumberOfContacts;
		float							m_fPlanePenetrationConstant;

		Asura_AVector3					m_xNormal;

		Asura_AVector3					m_axContactPositions[uASURA_CONTACTCLUSTER_MAXCONTACTS];
};


class Asura_Physics_ContactCluster_List
{
	public:

		Asura_Physics_ContactCluster_List()
			: m_pxClusterHead( NULL )
		{
		}

		~Asura_Physics_ContactCluster_List()
		{
			Destroy();
		}

		void Reset( void )
		{
			Destroy();
		}

		void Destroy( void )
		{
			Asura_Physics_ContactCluster* pxCluster = m_pxClusterHead;

			while ( pxCluster )
			{
				Asura_Physics_ContactCluster* pxNextCluster = pxCluster->m_pxNextCluster;

				Asura_Physics_ContactCluster::Destroy( pxCluster );

				pxCluster = pxNextCluster;
			}

			m_pxClusterHead = 0;
		}

		const Asura_Physics_ContactCluster* GetFirstCluster( void ) const
		{
			return m_pxClusterHead;
		}

		void AddContactPosition( const Asura_AVector3& xBodyPosition, const Asura_AVector3& xPosition, const Asura_AVector3& xNormal, float fPenetration )
		{
			if (m_pxClusterHead)
			{
				// search through the existing clusters to see if we have a match for this normal
				for ( Asura_Physics_ContactCluster* pxCluster = m_pxClusterHead; pxCluster; pxCluster = pxCluster->m_pxNextCluster )
				{
					// if we successfully find a cluster we can just stop
					if ( pxCluster->AddContactPosition( xBodyPosition, xPosition, xNormal, fPenetration ) ) return;
				}

				// Must not be a valid cluster - let's make a new one
				Asura_Physics_ContactCluster* pxNewCluster = Asura_Physics_ContactCluster::Create();
				pxNewCluster->AddContactPosition( xBodyPosition,xPosition, xNormal, fPenetration );

				// Update the head and add this cluster to the list at the top
				// (so the next time this is called, we'll check the new cluster first)
				pxNewCluster->m_pxNextCluster = m_pxClusterHead;
				m_pxClusterHead = pxNewCluster;
			}
			else
			{
				// no cluster exists - make a new one and set the list head accordingly
				Asura_Physics_ContactCluster* pxNewCluster = Asura_Physics_ContactCluster::Create();
				pxNewCluster->AddContactPosition( xBodyPosition, xPosition, xNormal, fPenetration );

				m_pxClusterHead = pxNewCluster;
			}
		}


	private:

		Asura_Physics_ContactCluster* m_pxClusterHead;
};


#endif//__ASURA_PHYSICS_CONTACTCLUSTER_H__
