//
// Asura_Physics_EnvironmentCache.cpp
//
// Created by KJFL - 2004-03-24 15:04:18
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_EnvironmentCache.h"

#include "Core/Asura_Mempool.h"
#include "Physics/Asura_Physics_Environment.h"


static const u_int uENVCACHE_ELEMENTS_PER_POOL = 32;
static Asura_ResizingMempool g_xFacePool( uENVCACHE_ELEMENTS_PER_POOL, sizeof(Asura_Physics_EnvironmentCache_Face), __FILE__, __LINE__ );
static Asura_ResizingMempool g_xEdgePool( uENVCACHE_ELEMENTS_PER_POOL, sizeof(Asura_Physics_EnvironmentCache_Edge), __FILE__, __LINE__ );
static Asura_ResizingMempool g_xVertexPool( uENVCACHE_ELEMENTS_PER_POOL, sizeof(Asura_Physics_EnvironmentCache_Vertex), __FILE__, __LINE__ );

Asura_Physics_EnvironmentCache_Face* Asura_Physics_EnvironmentCache_Face::Create( void )
{
	return static_cast<Asura_Physics_EnvironmentCache_Face*>(g_xFacePool.Allocate());
}

void Asura_Physics_EnvironmentCache_Face::Destroy( Asura_Physics_EnvironmentCache_Face* pxFace )
{
	g_xFacePool.Deallocate( pxFace );
}

Asura_Physics_EnvironmentCache_Edge* Asura_Physics_EnvironmentCache_Edge::Create( void )
{
	return static_cast<Asura_Physics_EnvironmentCache_Edge*>(g_xEdgePool.Allocate());
}
void Asura_Physics_EnvironmentCache_Edge::Destroy( Asura_Physics_EnvironmentCache_Edge* pxEdge )
{
	g_xEdgePool.Deallocate( pxEdge );
}

Asura_Physics_EnvironmentCache_Vertex* Asura_Physics_EnvironmentCache_Vertex::Create( void )
{
	return static_cast<Asura_Physics_EnvironmentCache_Vertex*>(g_xVertexPool.Allocate());
}
void Asura_Physics_EnvironmentCache_Vertex::Destroy( Asura_Physics_EnvironmentCache_Vertex* pxVertex )
{
	g_xVertexPool.Deallocate( pxVertex );
}


u_int Asura_Physics_EnvironmentCache::s_uNumberOfCacheUpdates;

#if defined(ASURA_USE_STREAMING)
Asura_Physics_EnvironmentCache* Asura_Physics_EnvironmentCache::s_pxEnvironmentCacheList = 0;
#endif //defined(ASURA_USE_STREAMING)

void Asura_Physics_EnvironmentCache::Update( const Asura_Bounding_Box& xObjectBB, const float fBBExpansionMargin )
{
	if (m_xCacheBoundingBox.Contains(xObjectBB)) return;

	m_xCacheBoundingBox = xObjectBB;

	if (fBBExpansionMargin > 0.0f)
	{
		m_xCacheBoundingBox.Grow( fBBExpansionMargin );
	}
	
	PruneCache();

	Asura_Physics_Environment::FillCache( this );

	Asura_Physics_EnvironmentCache::s_uNumberOfCacheUpdates++;

}


void Asura_Physics_EnvironmentCache::Update( const Asura_Bounding_Box& xObjectBB, const float fDisp, const Asura_AVector3& xDirection )
{
	if (m_xCacheBoundingBox.Contains(xObjectBB)) return;

	Asura_Bounding_Box xProjectedBB = xObjectBB;
 
	m_xCacheBoundingBox = xProjectedBB;

	m_xCacheBoundingBox.Grow( 0.1f );

	// clamp the extra projection of the env cache bb to 0.5m.
	const float fProjectionDistance = Asura_Maths::Min( fDisp * 10.0f, 0.5f );

	const Asura_AVector3 xDisp = xDirection * (fProjectionDistance);

	const float fVelX = xDirection.GetX();
	if ( fVelX > 0.0f)
	{
		m_xCacheBoundingBox.MaxX += xDisp.GetX();
	}
	else if ( fVelX < 0.0f)
	{
		m_xCacheBoundingBox.MinX += xDisp.GetX();
	}

	const float fVelY = xDirection.GetY();
	if ( fVelY > 0.0f)
	{
		m_xCacheBoundingBox.MaxY += xDisp.GetY();
	}
	else if ( fVelY < 0.0f)
	{
		m_xCacheBoundingBox.MinY += xDisp.GetY();
	}


	const float fVelZ = xDirection.GetZ();
	if ( fVelZ > 0.0f)
	{
		m_xCacheBoundingBox.MaxZ += xDisp.GetZ();
	}
	else if ( fVelZ < 0.0f)
	{
		m_xCacheBoundingBox.MinZ += xDisp.GetZ();
	}


	PruneCache();

	Asura_Physics_Environment::FillCache( this );

	Asura_Physics_EnvironmentCache::s_uNumberOfCacheUpdates++;

}


void Asura_Physics_EnvironmentCache::Flush( void )
{
	// need to free mem
	Asura_Physics_EnvironmentCache_Face* pxCacheFace = m_pxFaceList;

	while (pxCacheFace)
	{
		Asura_Physics_EnvironmentCache_Face* pxNext = pxCacheFace->m_pxNext;

		Asura_Physics_EnvironmentCache_Face::Destroy( pxCacheFace );

		pxCacheFace = pxNext;
	}
	
	Asura_Physics_EnvironmentCache_Edge* pxCacheEdge = m_pxEdgeList;

	while (pxCacheEdge)
	{
		Asura_Physics_EnvironmentCache_Edge* pxNext = pxCacheEdge->m_pxNext;

		Asura_Physics_EnvironmentCache_Edge::Destroy( pxCacheEdge );

		pxCacheEdge = pxNext;
	}

	Asura_Physics_EnvironmentCache_Vertex* pxCacheVertex = m_pxVertexList;

	while (pxCacheVertex)
	{
		Asura_Physics_EnvironmentCache_Vertex* pxNext = pxCacheVertex->m_pxNext;

		Asura_Physics_EnvironmentCache_Vertex::Destroy( pxCacheVertex );

		pxCacheVertex = pxNext;
	}

	m_pxFaceList = 0;
	m_pxEdgeList = 0;
	m_pxVertexList = 0;

#ifdef ASURA_ENVCACHE_DEBUG
	m_iNumFaces = 0;
	m_iNumEdges = 0;
	m_iNumVerts = 0;
#endif

	m_xCacheBoundingBox.Reset();
}


void Asura_Physics_EnvironmentCache::Reset( void )
{
    Flush();

	m_bFacesOnly = false;

	// default to something useful for collisions calls 
#ifdef ASURA_TOOLS
    m_uFlagsOfPolysToIgnore = AsuraPolygonFlag_NoCollisions;
#else
    m_usFlagsOfPolysToIgnore = AsuraPolygonFlag_NoCollisions;
#endif
}

void Asura_Physics_EnvironmentCache::AddFaceToCache( u_int uFaceID )
{
	// is face already in cache?
	// KJFL - 2004-03-25 10:53:27 - obviously this is a prime candidate for optimisation!
	Asura_Physics_EnvironmentCache_Face* pxCacheFace = m_pxFaceList;
	
	while (pxCacheFace)
	{
		if (pxCacheFace->m_uFaceID == uFaceID)
		{
			// we're not interested then
			return;
		}

		pxCacheFace = pxCacheFace->m_pxNext;
	}

	// new face then... add to cache

	// will want mempool for cache faces
	Asura_Physics_EnvironmentCache_Face* pxNewCacheFace = Asura_Physics_EnvironmentCache_Face::Create();

	if ( pxNewCacheFace )
	{
		// fill in new data and add to head of list
		pxNewCacheFace->m_uFaceID = uFaceID;
		pxNewCacheFace->m_pxNext = m_pxFaceList;
		pxNewCacheFace->m_pxContact = 0;

		pxNewCacheFace->m_uLastSeparationPlane = 0;

		m_pxFaceList = pxNewCacheFace;

#ifdef ASURA_ENVCACHE_DEBUG
		m_iNumFaces++;
#endif
	}

}

void Asura_Physics_EnvironmentCache::AddEdgeToCache( u_int uVertex0Index, u_int uVertex1Index )
{
	// is edge already in cache?
	// KJFL - 2004-03-25 10:53:27 - obviously this is a prime candidate for optimisation!
	Asura_Physics_EnvironmentCache_Edge* pxCacheEdge = m_pxEdgeList;
	

	const Asura_AVector3& xEdge0 = Asura_Physics_Environment::GetVertex( uVertex0Index );
	const Asura_AVector3& xEdge1 = Asura_Physics_Environment::GetVertex( uVertex1Index );

	if (!Asura_Physics_Environment::DoesEdgeIntersectAABB( xEdge0, xEdge1, m_xCacheBoundingBox )) return;


	while (pxCacheEdge)
	{
		if (pxCacheEdge->m_uEdgeVertex0ID == uVertex0Index && pxCacheEdge->m_uEdgeVertex1ID == uVertex1Index )
		{
			// we're not interested then
			return;
		}

		if (pxCacheEdge->m_uEdgeVertex0ID == uVertex1Index && pxCacheEdge->m_uEdgeVertex1ID == uVertex0Index )
		{
			// we're not interested then
			return;
		}


		pxCacheEdge = pxCacheEdge->m_pxNext;
	}

	// new Edge then... add to cache

	// will want mempool for cache Edges
	Asura_Physics_EnvironmentCache_Edge* pxNewCacheEdge = Asura_Physics_EnvironmentCache_Edge::Create();

	if ( pxNewCacheEdge )
	{
		// fill in new data and add to head of list
		pxNewCacheEdge->m_uEdgeVertex0ID = uVertex0Index;
		pxNewCacheEdge->m_uEdgeVertex1ID = uVertex1Index;
		pxNewCacheEdge->m_pxContact = 0;
		pxNewCacheEdge->m_pxNext = m_pxEdgeList;

		m_pxEdgeList = pxNewCacheEdge;

#ifdef ASURA_ENVCACHE_DEBUG
		m_iNumEdges++;
#endif
	}
}

void Asura_Physics_EnvironmentCache::AddVertexToCache( u_int uVertexID )
{
	// is Vertex already in cache?
	// KJFL - 2004-03-25 10:53:27 - obviously this is a prime candidate for optimisation!
	Asura_Physics_EnvironmentCache_Vertex* pxCacheVertex = m_pxVertexList;
	
	while (pxCacheVertex)
	{
		if (pxCacheVertex->m_uVertexID == uVertexID)
		{
			// we're not interested then
			return;
		}

		pxCacheVertex = pxCacheVertex->m_pxNext;
	}

	// new Vertex then... add to cache

	// will want mempool for cache Vertices
	Asura_Physics_EnvironmentCache_Vertex* pxNewCacheVertex = Asura_Physics_EnvironmentCache_Vertex::Create();

	if ( pxNewCacheVertex )
	{
		// fill in new data and add to head of list
		pxNewCacheVertex->m_uVertexID = uVertexID;
		pxNewCacheVertex->m_pxContact = 0;
		pxNewCacheVertex->m_pxNext = m_pxVertexList;

		m_pxVertexList = pxNewCacheVertex;

#ifdef ASURA_ENVCACHE_DEBUG
		m_iNumVerts++;
#endif
	}
}

void Asura_Physics_EnvironmentCache::PruneCache( void )
{
	// faces - just throw them all away
/*
	Asura_Physics_EnvironmentCache_Face* pxCacheFace = m_pxFaceList;
	
	while (pxCacheFace)
	{
		Asura_Physics_EnvironmentCache_Face* pxNext = pxCacheFace->m_pxNext;

		delete pxCacheFace;

		pxCacheFace = pxNext;
	}

	m_pxFaceList = 0;
*/

	// Faces - keep ones which have a valid Contact
	// This is made more tricky because we don't have Prev pointers to save
	// space. Thus we have to keep a pointer to the last 'next' field
	// (or start of the list) which we update as we delete things

	Asura_Physics_EnvironmentCache_Face** ppxCacheFace = &m_pxFaceList;

	while (*ppxCacheFace)
	{
		Asura_Physics_EnvironmentCache_Face* pxCacheFace = *ppxCacheFace;

		// kill it if it doesn't have a Contact
		if (!pxCacheFace->m_pxContact)
		{
			// set our last 'next' pointer to the next Contact
			*ppxCacheFace = pxCacheFace->m_pxNext;

			Asura_Physics_EnvironmentCache_Face::Destroy( pxCacheFace );

#ifdef ASURA_ENVCACHE_DEBUG
			m_iNumFaces--;
#endif
		}
		else
		{
			ppxCacheFace = &(pxCacheFace->m_pxNext);
		}
	}
	// Edges - keep ones which have a valid Contact
	// This is made more tricky because we don't have Prev pointers to save
	// space. Thus we have to keep a pointer to the last 'next' field
	// (or start of the list) which we update as we delete things

	Asura_Physics_EnvironmentCache_Edge** ppxCacheEdge = &m_pxEdgeList;

	while (*ppxCacheEdge)
	{
		Asura_Physics_EnvironmentCache_Edge* pxCacheEdge = *ppxCacheEdge;

		// kill it if it doesn't have a Contact
		if (!pxCacheEdge->m_pxContact)
		{
			// set our last 'next' pointer to the next Contact
			*ppxCacheEdge = pxCacheEdge->m_pxNext;

			Asura_Physics_EnvironmentCache_Edge::Destroy( pxCacheEdge );

#ifdef ASURA_ENVCACHE_DEBUG
			m_iNumEdges--;
#endif
		}
		else
		{
			ppxCacheEdge = &(pxCacheEdge->m_pxNext);
		}
	}

	// Vertices - keep ones which have a valid Contact
	// This is made more tricky because we don't have Prev pointers to save
	// space. Thus we have to keep a pointer to the last 'next' field
	// (or start of the list) which we update as we delete things

	Asura_Physics_EnvironmentCache_Vertex** ppxCacheVertex = &m_pxVertexList;

	while (*ppxCacheVertex)
	{
		Asura_Physics_EnvironmentCache_Vertex* pxCacheVertex = *ppxCacheVertex;

		// kill it if it doesn't have a Contact
		if (!pxCacheVertex->m_pxContact)
		{
			// set our last 'next' pointer to the next Contact
			*ppxCacheVertex = pxCacheVertex->m_pxNext;

			Asura_Physics_EnvironmentCache_Vertex::Destroy( pxCacheVertex );

#ifdef ASURA_ENVCACHE_DEBUG
			m_iNumVerts--;
#endif
		}
		else
		{
			ppxCacheVertex = &(pxCacheVertex->m_pxNext);
		}
	}
}

#if defined(ASURA_USE_STREAMING)
// MGH 10/4/2006: Zone has been unloaded. Clean up any remaining references to it
// in all the environment caches.
// NOTE: You should clean up any contacts before calling this function.
void Asura_Physics_EnvironmentCache::RemoveZone(const u_int uZoneIndex)
{
	for (Asura_Physics_EnvironmentCache* pxNext = s_pxEnvironmentCacheList;
		 pxNext != 0;
		 pxNext = pxNext->m_pxNext)
	{
		pxNext->RemoveZoneFromInstance(uZoneIndex);
	}
}

void Asura_Physics_EnvironmentCache::RemoveZoneFromInstance(const u_int uZoneIndex)
{
	const u_int uZoneShifted = (uZoneIndex<<uASURA_ZONE_INDEX_SHIFT);

	// Remove faces in the specfied zone.
	Asura_Physics_EnvironmentCache_Face** ppxCacheFace = &m_pxFaceList;
	while (*ppxCacheFace)
	{
		Asura_Physics_EnvironmentCache_Face* pxCacheFace = *ppxCacheFace;

		if ((pxCacheFace->m_uFaceID & uASURA_ZONE_INDEX_MASK) == uZoneShifted)
		{
			// set our last 'next' pointer to the next Contact
			*ppxCacheFace = pxCacheFace->m_pxNext;
			Asura_Physics_EnvironmentCache_Face::Destroy( pxCacheFace );

#ifdef ASURA_ENVCACHE_DEBUG
			m_iNumFaces--;
#endif
		}
		else
		{
			ppxCacheFace = &(pxCacheFace->m_pxNext);
		}
	}

	// Remove edges in the specfied zone.
	Asura_Physics_EnvironmentCache_Edge** ppxCacheEdge = &m_pxEdgeList;
	while (*ppxCacheEdge)
	{
		Asura_Physics_EnvironmentCache_Edge* pxCacheEdge = *ppxCacheEdge;

		if ( ((pxCacheEdge->m_uEdgeVertex0ID & uASURA_ZONE_INDEX_MASK) == uZoneShifted) ||
			 ((pxCacheEdge->m_uEdgeVertex1ID & uASURA_ZONE_INDEX_MASK) == uZoneShifted) )
		{
			// kill it if it doesn't have a Contact
			Asura_Assert(!pxCacheEdge->m_pxContact, "Failed to clean up contact");
			if (!pxCacheEdge->m_pxContact)
			{
				// set our last 'next' pointer to the next Contact
				*ppxCacheEdge = pxCacheEdge->m_pxNext;
				Asura_Physics_EnvironmentCache_Edge::Destroy( pxCacheEdge );

#ifdef ASURA_ENVCACHE_DEBUG
				m_iNumEdges--;
#endif
				continue;
			}
		}
			
		ppxCacheEdge = &(pxCacheEdge->m_pxNext);
	}
	
	// Remove vertices in the specfied zone.
	Asura_Physics_EnvironmentCache_Vertex** ppxCacheVertex = &m_pxVertexList;
	while (*ppxCacheVertex)
	{
		Asura_Physics_EnvironmentCache_Vertex* pxCacheVertex = *ppxCacheVertex;

		if ((pxCacheVertex->m_uVertexID & uASURA_ZONE_INDEX_MASK) == uZoneShifted)
		{
			// kill it if it doesn't have a Contact
			Asura_Assert(!pxCacheVertex->m_pxContact, "Failed to clean up contact");
			if (!pxCacheVertex->m_pxContact)
			{
				// set our last 'next' pointer to the next Contact
				*ppxCacheVertex = pxCacheVertex->m_pxNext;
				Asura_Physics_EnvironmentCache_Vertex::Destroy( pxCacheVertex );

#ifdef ASURA_ENVCACHE_DEBUG
				m_iNumVerts--;
#endif

				continue;
			}
		}

		ppxCacheVertex = &(pxCacheVertex->m_pxNext);
	}
}
#endif //defined(ASURA_USE_STREAMING)
