//
// Asura_Physics_MovementQuery.h
//
// Created by KJFL - 2 February	2007
//
// Copyright (C) 2007 Rebellion, All Rights	Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma	once
#endif

#ifndef	__ASURA_PHYSICS_MOVEMENTQUERY_H__
#define	__ASURA_PHYSICS_MOVEMENTQUERY_H__

#include "Core/Asura_Maths.h"

#include "Mesh/Asura_Mesh_PolygonFlags.h"

// To make a Physics Query fill	in the Asura_Physics_MovementQuery class below,	and	then pass it
// in as the parameter for the Asura_Physics_MovementQuery_Results class (see bottom of	file). The
// input data is not modified, so you can tweak	individual settings	and	ask	for	multiple results, etc.
//
// e.g.	
//
//	Asura_Physics_MovementQuery	xInput(	xStartPosition,	xDisplacement, xOBB, ASURA_PMQCP_SPHERESTACK );
//
//	Asura_Physics_MovementQuery_Results	xResults( xInput );
// 
//	...	 then query	methods	like xResults.GetEndPosition();	
//
class Asura_Physics_MovementQuery
{
	public:
			
		enum ASURA_PMQ_COLLISION_PROXY
		{
			ASURA_PMQCP_OBB,
			ASURA_PMQCP_SPHERE,
			ASURA_PMQCP_SPHERESTACK
		};

		// All queries use the Asura_OBB class,	even if	you	want to	use	a collision	proxy like a sphere.
		// The Asura_OBB constuctors take all sorts	of values such as sphere radii,	AABBs, AABBs with rotations, etc.
		// A sphere	stack will automatically places	spheres	along the long axis	of the OBB etc.
		Asura_Physics_MovementQuery( const Asura_AVector3& xStartPosition, const Asura_AVector3& xDisplacement,
									 const Asura_OBB& xOBB,	enum ASURA_PMQ_COLLISION_PROXY eCollisionProxy = ASURA_PMQCP_OBB,
									 u_short usFlagsToIgnore = AsuraPolygonFlag_NoCollisions, bool bStopAtFirstCollision = true,
									 bool bCollideWithStaticObjects	= false, bool bCollideWithExternalControlObjects = false,
									 bool bCollideWithActorObjects = false,	bool bCollideWithEnvironment = true,
									 bool bCollideWithDynamicObjects = false,
									 u_int uCollisionGroups = 0xffffffff, u_int uCollisionGroupsToCollideWith = 0xffffffff );

		// Starting	position for test object 
		void SetStartPosition( const Asura_AVector3& xStartPosition	);
		const Asura_AVector3& GetStartPosition(	void ) const;

		// Displacement	is the direction and distance you wish the test	object to try to travel
		void SetDisplacement( const	Asura_AVector3&	xDisplacement );
		const Asura_AVector3& GetDisplacement( void	) const;

		// Collision OBB for test object
		void SetOBB( const Asura_OBB& xOBB );
		const Asura_OBB& GetOBB( void )	const;

		// The collision proxy shape of	the	test object
		void SetCollisionProxy(	enum ASURA_PMQ_COLLISION_PROXY eProxy )	{ m_eCollisionProxy	= eProxy; }
		enum ASURA_PMQ_COLLISION_PROXY GetCollisionProxy( void ) const { return	m_eCollisionProxy; }

		// Should test object Stop at first	collision, or carry	on moving as an	Actor would
		void SetStopAtFirstCollision( bool bStop );
		bool GetStopAtFirstCollision( void ) const;

		// Flags of	polygons test object should	ignore
		void SetFlagsToIgnore( u_short usFlags );
		u_short	GetFlagsToIgnore( void ) const;

		// Handle of physics object	test object	should ignore
		void SetHandleToIgnore(	Asura_Handle iHandle );
		Asura_Handle GetHandleToIgnore(	void ) const;

		// object collisions: vs STATIC	objects
		void SetCollideWithStaticObjects( bool bCollide	);
		bool GetCollideWithStaticObjects( void ) const;

		// object collisions: vs EXTERNALCONTROL objects
		void SetCollideWithExternalControlObjects( bool	bCollide );
		bool GetCollideWithExternalControlObjects( void	) const;

		// object collisions: vs ACTOR objects
		void SetCollideWithActorObjects( bool bCollide );
		bool GetCollideWithActorObjects( void )	const;

		// object collisions: vs environment
		void SetCollideWithEnvironment(	bool bCollide );
		bool GetCollideWithEnvironment(	void ) const;

		// object collisions: vs DYNAMIC objects
		void SetCollideWithDynamicObjects( bool bCollide );
		bool GetCollideWithDynamicObjects( void ) const;

		// Collision Groups - for two objects to collide, they must *both* want to collide with 
		// at least one collision group the other object is a member of.
		//
		// Get/Set the bitfield which says which groups the object is a member of
		void SetCollisionGroups( u_int uGroupsBitfield );
		u_int GetCollisionGroups( void ) const;

		// Get/Set the bitfield which says which groups the object should collide with
		void SetCollisionGroupsToCollideWith( u_int uGroupsBitfield );
		u_int GetCollisionGroupsToCollideWIth( void ) const;

		~Asura_Physics_MovementQuery() {}	

	private:

		// The OBB below actually represents all collision types - OBB,	Sphere,	SphereStack, AABB -
		// and is coupled with the "CollisionProxy"	value. These values	are	set	in the constructors.
		Asura_AVector3					m_xStartPosition;
		Asura_AVector3					m_xDisplacement;

		Asura_OBB						m_xOBB;
		enum ASURA_PMQ_COLLISION_PROXY	m_eCollisionProxy;

		Asura_Handle					m_iHandleToIgnore;
		u_short							m_usFlagsToIgnore;
		u_short							m_bStopAtFirstCollision	:1;
		u_short							m_bCollideWithStaticObjects	:1;
		u_short							m_bCollideWithExternalControlObjects :1;
		u_short							m_bCollideWithActorObjects :1;
		u_short							m_bCollideWithEnvironment :1;
		u_short							m_bCollideWithDynamicObjects :1;
		
		u_int							m_uCollisionGroups;
		u_int							m_uCollisionGroupsToCollideWith;
};

class Asura_Physics_MovementQuery_Results
{
	public:

		// constructor makes query based on	input data
		Asura_Physics_MovementQuery_Results( const Asura_Physics_MovementQuery&	xInput );
		
		// accessors to	output data	
		const Asura_AVector3& GetEndPosition( void ) const { return	m_xEndPosition;	}

		const Asura_AVector3& GetAllowedDisplacement( void ) const { return	m_xAllowedDisplacement;	}

		const Asura_AVector3& GetLastCollisionNormal( void ) const { return	m_xLastCollisionNormal;	}

		~Asura_Physics_MovementQuery_Results() {}
	
	private:

		Asura_AVector3	m_xAllowedDisplacement;
		
		Asura_AVector3	m_xEndPosition;

		Asura_AVector3	m_xLastCollisionNormal;

};


#endif//__ASURA_PHYSICS_MOVEMENTQUERY_H__
