//
// Asura_Physics_Object.cpp
//
// Created by KJFL - Tuesday, November 16, 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"


// This file's header.
#include "Physics/Asura_Physics_Object.h"

#include "Core/Asura_MempoolFactory_Template.h"
#include "Dynamics/Asura_Collision_Object.h"
#include "Physics/Asura_Physics_RigidBody.h"
#include "Physics/Asura_Physics_RigidBodyDefinition.h"
#include "Physics/Asura_Physics.h" // only needed to get time slice - interface change perhaps to simplify linkage?

#include "ProjectSettings/Asura_PhysicsConfiguration.h"

#ifndef ASURA_PHYSICS_OBJECT_SUSPICIOUS_VELOCITY_THRESHOLD
	#define	  ASURA_PHYSICS_OBJECT_SUSPICIOUS_VELOCITY_THRESHOLD   (250.0f)
#endif 


static u_short(*g_pfnDefaultFlagsOfPolysToIgnoreCallback)(void) = NULL;

Asura_Physics_Object_Factory	Asura_Physics_Object::s_xFactory;


void Asura_Physics_Object::Reset( void )
{
    s_xFactory.Reset();
}


void Asura_Physics_Object::SetDefaultFlagsOfPolysToIgnoreCallback( u_short(*pfnCallback)(void) )
{
    g_pfnDefaultFlagsOfPolysToIgnoreCallback = pfnCallback;
}


Asura_Physics_Object* Asura_Physics_Object::Create( enum ASURA_PHYSICS_OBJECT_TYPE eType /*= ASURA_PHYSICS_OBJECT_STATIC*/ )
{
	Asura_Handle iHandle = s_xFactory.CreateInstance();
		
	if (iHandle == ASURA_INVALID_HANDLE) return 0;

	Asura_Physics_Object* pxNewObject = s_xFactory.GetInstancePointer( iHandle );
		
    pxNewObject->m_eType = eType;
	pxNewObject->m_iHandle = iHandle;
	pxNewObject->m_pxCollisionObject = 0;

	switch (eType)
	{
		case ASURA_PHYSICS_OBJECT_ACTOR:
		{
			pxNewObject->m_iRigidBodyHandle = Asura_Physics_RigidBodyInstance::CreateDynamicInstance( );	

			Asura_Physics_RigidBodyInstance* pxInstance = Asura_Physics_RigidBodyInstance::GetInstancePointer( pxNewObject->m_iRigidBodyHandle );
			
			pxInstance->m_bAxiallyAligned = true;
			pxInstance->m_bActor = true;

			// set friction to zero, so actors don't drag objects around if they walk over them etc.
			pxInstance->m_fFriction = 0.0f;
			break;
		}
		case ASURA_PHYSICS_OBJECT_DYNAMIC:
		{
			pxNewObject->m_iRigidBodyHandle = Asura_Physics_RigidBodyInstance::CreateDynamicInstance( );	
			break;
		}
		case ASURA_PHYSICS_OBJECT_EXTERNALCONTROL:
		{
			pxNewObject->m_iRigidBodyHandle = Asura_Physics_RigidBodyInstance::CreateExternallyControlledInstance( );	
			break;
		}
		case ASURA_PHYSICS_OBJECT_STATIC:
		{
			pxNewObject->m_iRigidBodyHandle = Asura_Physics_RigidBodyInstance::CreateStaticInstance( );	
			break;
		}

		default:
		{
			pxNewObject->m_iRigidBodyHandle = ASURA_INVALID_HANDLE;
			break;
		}
	}

	if ( pxNewObject->m_iRigidBodyHandle != ASURA_INVALID_HANDLE )
	{
		pxNewObject->m_pxRigidBodyInstance = Asura_Physics_RigidBodyInstance::GetInstancePointer( pxNewObject->m_iRigidBodyHandle );

	}
	else
	{
		pxNewObject->m_pxRigidBodyInstance = 0;
	}

	Asura_Assert( pxNewObject->m_pxRigidBodyInstance, "Fatal Error: Unable to allocate a rigid body!" );

	// If we haven't got a rigid body pointer, the next assert that reads through m_pxRigidBodyInstance will crash,
	// but that's going to happen anyway as soon as anything else tries to read through it. At least we asserted on
	// it good and early. We're in serious trouble if we were unable to find any memory, so there's a limit to how
	// graceful we can be in this scenario. However, we could handle it better than it currently does.
	Asura_Assert( pxNewObject->m_iRigidBodyHandle == pxNewObject->m_pxRigidBodyInstance->GetHandle(), "Handle Mismatch!" );

	pxNewObject->SetIgnore( false );

	pxNewObject->SetPosition( Asura_Maths::ZeroVector3 );
	pxNewObject->SetOrientation( Asura_Maths::IdentityMatrix3x3 );

	// A handy little test case
#if 0
	if ( eType == ASURA_PHYSICS_OBJECT_EXTERNALCONTROL )
	{
		pxNewObject->SetCollisionProxy( COLLISION_PROXY_POLYGONAL );
	}
#endif

    if( g_pfnDefaultFlagsOfPolysToIgnoreCallback )
    {
        pxNewObject->SetFlagsOfPolysToIgnore( g_pfnDefaultFlagsOfPolysToIgnoreCallback() );
    }

	return pxNewObject;
}

void Asura_Physics_Object::ChangeType( enum ASURA_PHYSICS_OBJECT_TYPE eType )
{
	if ( m_eType == eType )
	{
		return; // no change required
	}

	bool bStaticsChanged = (m_eType == ASURA_PHYSICS_OBJECT_STATIC) || (eType == ASURA_PHYSICS_OBJECT_STATIC);

	m_eType = eType;

	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// set the type flags
	m_pxRigidBodyInstance->m_bActor = (eType == ASURA_PHYSICS_OBJECT_ACTOR);
	m_pxRigidBodyInstance->m_bStatic = (eType == ASURA_PHYSICS_OBJECT_STATIC);
	m_pxRigidBodyInstance->m_bExternallyControlled = (eType == ASURA_PHYSICS_OBJECT_EXTERNALCONTROL);

	m_pxRigidBodyInstance->FlushCachedData();
	m_pxRigidBodyInstance->m_xJumpVelocity.Zero();

	m_pxRigidBodyInstance->RemoveFromCurrentList();

	switch (eType)
	{
		default:
		case ASURA_PHYSICS_OBJECT_ACTOR:
		case ASURA_PHYSICS_OBJECT_DYNAMIC:
		{
			m_pxRigidBodyInstance->AddToList( &Asura_Physics_RigidBodyInstance::GetDynamicList() );

			// Check we don't have infinite mass because that's only valid
            // for external control and static physics objects.
			if ( m_pxRigidBodyInstance->GetInvMass() == 0.0f )
			{	
				m_pxRigidBodyInstance->SetInvMass( 1.0f );
			}

			m_pxRigidBodyInstance->WakeUp();
			break;
		}
		case ASURA_PHYSICS_OBJECT_EXTERNALCONTROL:
		{
			m_pxRigidBodyInstance->AddToList( &Asura_Physics_RigidBodyInstance::GetExternallyControlledList() );
			
			// NB: Used to force infinite mass here but that's dealt with
            // internally by the rigid body instance now, so that a sequence
            // of calls to ChangeType() don't reset the mass.
			
			m_pxRigidBodyInstance->WakeUp();
			break;
		}
		case ASURA_PHYSICS_OBJECT_STATIC:
		{
			m_pxRigidBodyInstance->AddToList( &Asura_Physics_RigidBodyInstance::GetStaticList() );

            // See comment about infinite mass for external control case.
			
			m_pxRigidBodyInstance->m_bAsleep = true;
			break;
		}
	}

	// our effective mass may have changed, and with it our inertia tensor
	m_pxRigidBodyInstance->CalculateInvInertiaTensorDiagonal();

	if (bStaticsChanged) Asura_Physics_System::NotifyOfChangeToStaticObjects();

	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

}

void Asura_Physics_Object::Destroy( Asura_Physics_Object* pxObject )
{
	if (!pxObject)
	{
		//PHYSICSTODO - need some ironclad asserts and error checking
		return;
	}

	if ( pxObject->m_iRigidBodyHandle != ASURA_INVALID_HANDLE )
	{
		Asura_Physics_RigidBodyInstance::Destroy( pxObject->m_iRigidBodyHandle );
	}

	s_xFactory.DestroyInstance( pxObject->m_iHandle );
	
}


const Asura_Vector_3& Asura_Physics_Object::GetPosition() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_xPosition = m_pxRigidBodyInstance->GetRenderingPosition();

	return m_xPosition;
}

const Asura_Matrix_3x3& Asura_Physics_Object::GetOrientation() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	if (m_pxRigidBodyInstance->m_bAxiallyAligned)
	{
		return m_xOrientation;
	}
	else
	{
		m_pxRigidBodyInstance->GetOrientation().ToMat3( m_xOrientation );
		return m_xOrientation;	
	}
}

const Asura_Vector_3& Asura_Physics_Object::GetLinImpulse() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	static Asura_Vector_3 xVel;

	xVel = m_pxRigidBodyInstance->GetCentreOfMass() - m_pxRigidBodyInstance->GetPreviousCentreOfMass();

	xVel /= Asura_Physics_System::GetTimeSlice();

	return xVel;	

}
const Asura_Bounding_Box& Asura_Physics_Object::GetBoundingBox() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->m_xBoundingBox;
}

void Asura_Physics_Object::SetPosition( const Asura_Vector_3& xPosition )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// KJFL - 01 March 2007 - anything other than Externally Controlled objects should
	// treat a change in position as a teleport
	if ( m_pxRigidBodyInstance->IsExternallyControlled() )
	{
		m_pxRigidBodyInstance->SetPosition( Asura_AVector3( xPosition ) );
	}
	else
	{
		m_pxRigidBodyInstance->TeleportTo( Asura_AVector3( xPosition ) );
	}
}

void Asura_Physics_Object::TeleportTo( const Asura_Vector_3& xPosition )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->TeleportTo( Asura_AVector3( xPosition ) );
}

void Asura_Physics_Object::SetOrientation( const Asura_Quat& xOrient )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	if (m_pxRigidBodyInstance->m_bAxiallyAligned)
	{
		xOrient.ToMatrix(m_xOrientation);
	}
	else
	{
		Asura_AMatrix3x3 xTemp;

		xOrient.ToMatrix(xTemp);
		m_pxRigidBodyInstance->SetOrientation( xTemp );
	}
}

void Asura_Physics_Object::SetOrientation( const Asura_Matrix_3x3& xOrient )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	if (m_pxRigidBodyInstance->m_bAxiallyAligned)
	{
		m_xOrientation = xOrient;
	}
	else
	{
		Asura_AMatrix3x3 xTemp;
		xTemp.FromMat3( xOrient );
		m_pxRigidBodyInstance->SetOrientation( xTemp );
	}
}

void Asura_Physics_Object::SetAxisConstraint( enum AXIS_CONSTRAINT_ID eAxisConstraint )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	switch ( eAxisConstraint )
	{
		default:
		case AXIS_CONSTRAINT_NONE:
		{
			m_pxRigidBodyInstance->m_bAxiallyAligned = false;
			m_pxRigidBodyInstance->m_bLockYAxisToBeVertical = false;
			m_pxRigidBodyInstance->m_bSelfRighting = false;
			break;
		}
		case AXIS_CONSTRAINT_Y_AXIS:
		{
			m_pxRigidBodyInstance->m_bAxiallyAligned = false;
			m_pxRigidBodyInstance->m_bLockYAxisToBeVertical = true;
			m_pxRigidBodyInstance->m_bSelfRighting = false;
			break;
		}
		case AXIS_CONSTRAINT_ALL_AXES:
		{
			m_pxRigidBodyInstance->m_bAxiallyAligned = true;
			m_pxRigidBodyInstance->m_bLockYAxisToBeVertical = false;
			m_pxRigidBodyInstance->m_bSelfRighting = false;
			break;
		}
		case AXIS_CONSTRAINT_SELF_RIGHTING:
		{
			m_pxRigidBodyInstance->m_bAxiallyAligned = false;
			m_pxRigidBodyInstance->m_bLockYAxisToBeVertical = false;
			m_pxRigidBodyInstance->m_bSelfRighting = true;
			break;
		}
	}
}

Asura_Physics_Object::AXIS_CONSTRAINT_ID Asura_Physics_Object::GetAxisConstraint( void ) const
{
	if( m_pxRigidBodyInstance )
	{
		if( m_pxRigidBodyInstance->m_bAxiallyAligned )
		{
			return AXIS_CONSTRAINT_ALL_AXES;
		}
		else if( m_pxRigidBodyInstance->m_bLockYAxisToBeVertical )
		{
			return AXIS_CONSTRAINT_Y_AXIS;
		}
		else if( m_pxRigidBodyInstance->m_bSelfRighting )
		{
			return AXIS_CONSTRAINT_SELF_RIGHTING;
		}
	}

	return AXIS_CONSTRAINT_NONE;
}

void Asura_Physics_Object::SetLinVelocity( const Asura_Vector_3& xLinVelocity )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	const float fDeltaTime = Asura_Physics_System::GetTimeSlice();

	if ( m_pxRigidBodyInstance->m_bActor )
	{
		// (MP - 13.07.2006) Obviously this will want moving or changing if we're
		// ever simulating very fast things, eg. a plane, but nothing we're doing
		// right now is going to want to go this fast.
		// KJFL - upped this to 250m/s for space ships in BFF :)
		// RDM - made this project-definable in the Asura_Physics_Configuration.h file (still defaults to 250m/s)
		Asura_Assertf( xLinVelocity.MagnitudeSquared() < (ASURA_PHYSICS_OBJECT_SUSPICIOUS_VELOCITY_THRESHOLD * ASURA_PHYSICS_OBJECT_SUSPICIOUS_VELOCITY_THRESHOLD ), 
			( "Suspiciously large velocity for entity %d.", m_pxRigidBodyInstance->GetAssociatedEntityGuid() ) ); // 250m/s is roughly 558mph.

		m_pxRigidBodyInstance->m_xForceEffect = xLinVelocity * fDeltaTime;
	}
	else
	{
		Asura_AVector3 xScaledLinearVelocity( xLinVelocity );
		xScaledLinearVelocity *= fDeltaTime;

		m_pxRigidBodyInstance->SetLinearVelocityOfBody( xScaledLinearVelocity );
	}

	m_pxRigidBodyInstance->WakeUp();
}

void Asura_Physics_Object::SetAngVelocity( const Asura_Vector_3& xAngVelocity )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	const float fDeltaTime = Asura_Physics_System::GetTimeSlice();

	Asura_AVector3 xScaledAngularVelocity( xAngVelocity );
	xScaledAngularVelocity *= fDeltaTime;

	m_pxRigidBodyInstance->SetAngularVelocityOfBody( xScaledAngularVelocity );
	m_pxRigidBodyInstance->WakeUp();
}

void Asura_Physics_Object::SetVelocities( const Asura_Vector_3& xLinVelocity, const Asura_Vector_3& xAngVelocity )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	const float fDeltaTime = Asura_Physics_System::GetTimeSlice();

	Asura_AVector3 xScaledAngularVelocity( xAngVelocity );
	Asura_AVector3 xScaledLinearVelocity( xLinVelocity );
	xScaledAngularVelocity *= fDeltaTime;
	xScaledLinearVelocity *= fDeltaTime;


	m_pxRigidBodyInstance->SetVelocitiesOfBody( xScaledLinearVelocity, xScaledAngularVelocity );
	m_pxRigidBodyInstance->WakeUp();
}

void Asura_Physics_Object::GetVelocities( Asura_Vector_3& xLinVelocity, Asura_Vector_3& xAngVelocity ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	const float fDeltaTime = Asura_Physics_System::GetTimeSlice();

	Asura_AVector3 xUnscaledAngularVelocity;
	Asura_AVector3 xUnscaledLinearVelocity;

	m_pxRigidBodyInstance->GetVelocitiesOfBody( xUnscaledLinearVelocity, xUnscaledAngularVelocity );

	xUnscaledAngularVelocity /= fDeltaTime;
	xUnscaledLinearVelocity	 /= fDeltaTime;

	xLinVelocity = xUnscaledLinearVelocity;
	xAngVelocity = xUnscaledAngularVelocity;
}


void Asura_Physics_Object::SetHandleToIgnore(Asura_Handle iHandle)
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->m_iHandleToIgnore = iHandle;
}

Asura_Handle Asura_Physics_Object::GetHandleToIgnore() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance ? m_pxRigidBodyInstance->m_iHandleToIgnore : ASURA_INVALID_HANDLE;
}

void Asura_Physics_Object::SetCollisionObject(const Asura_Collision_Object* pxCollision)
{
	// already been set to this? if so return
	if ( m_pxCollisionObject == pxCollision ) return;

	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	Asura_Assert(pxCollision->GetBoundingRadius() > 0.001f, "tiny collision object");

	m_pxCollisionObject = pxCollision; 
	
	if ( m_pxRigidBodyInstance->m_pxDefinition )
	{
		delete m_pxRigidBodyInstance->m_pxDefinition;
	}

	m_pxRigidBodyInstance->m_pxDefinition = new Asura_Physics_RigidBodyDefinition( pxCollision );
	m_pxRigidBodyInstance->InitialiseFromDefinition();
}

enum Asura_Physics_Object::COLLISION_PROXY Asura_Physics_Object::GetCollisionProxy( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// KJFL - 05 September 2006 - don't particularly like these static casts, and the assumed matching
	// of the two enums, but it means we can separate out the internal and external workings. Can always
	// change to a switch case etc.

	// JAT 6/09/06 - PS2 won't compile this code
	// return static_cast<Asura_Physics_Object::COLLISION_PROXY>(m_pxRigidBodyInstance->GetCollisionProxy());
	return static_cast<Asura_Physics_Object::COLLISION_PROXY>( static_cast<int>(m_pxRigidBodyInstance->GetCollisionProxy()) );
}

void Asura_Physics_Object::SetCollisionProxy( enum Asura_Physics_Object::COLLISION_PROXY eProxyType )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	// KJFL - 05 September 2006 - don't particularly like these static casts, and the assumed matching
	// of the two enums, but it means we can separate out the internal and external workings. Can always
	// change to a switch case etc.
	// JAT 6/09/06 - PS2 won't compile this code
	// Asura_Physics_RigidBodyInstance::COLLISION_PROXY eInternalProxy = static_cast<Asura_Physics_RigidBodyInstance::COLLISION_PROXY>(eProxyType);
	Asura_Physics_RigidBodyInstance::COLLISION_PROXY eInternalProxy = 
				static_cast<Asura_Physics_RigidBodyInstance::COLLISION_PROXY>( static_cast<int>(eProxyType) );

	if ( eInternalProxy == m_pxRigidBodyInstance->GetCollisionProxy() )
	{
		// nothing to do, already set to use this proxy
		return;
	}

	m_pxRigidBodyInstance->SetCollisionProxy( eInternalProxy );

	// Flush cached data, since they won't be valid if we change proxy.
	m_pxRigidBodyInstance->FlushCachedData();

}


void Asura_Physics_Object::SetupVehicle( const Asura_Bounding_Box& xBB, const Asura_AVector3* pxBasePositions )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	// TODO - this function is not used
}


void Asura_Physics_Object::ApplyImpulse( const Asura_Vector_3& xImpulse, bool bIgnoreMass /*=true*/ )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// assert on this?
	if ( !m_pxRigidBodyInstance->IsDynamic() ) return;

	if ( m_pxRigidBodyInstance->m_bActor )
	{
		Asura_AVector3 xScaledImpulse(xImpulse);

		if (!bIgnoreMass) xScaledImpulse *= m_pxRigidBodyInstance->GetInvMass();

		m_pxRigidBodyInstance->m_xJumpVelocity += xScaledImpulse;
	}
	else
	{
		Asura_AVector3 xScaledImpulse = Asura_AVector3(xImpulse) * Asura_Physics_System::GetTimeSlice();

		if (!bIgnoreMass) xScaledImpulse *= m_pxRigidBodyInstance->GetInvMass();

		m_pxRigidBodyInstance->ApplyVelocityChange( xScaledImpulse );
		m_pxRigidBodyInstance->WakeUp();
	}
}


void Asura_Physics_Object::ApplyImpulse( const Asura_Vector_3& xImpulse, const Asura_Vector_3& xPosition, bool bIgnoreMass /*=true*/ )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	// assert on this?
	if ( !m_pxRigidBodyInstance->IsDynamic() ) return;

	const Asura_AVector3 xScaledImpulse = Asura_AVector3(xImpulse) * Asura_Physics_System::GetTimeSlice();
	
	Asura_AVector3 xRelPosition(xPosition);

	xRelPosition -= m_pxRigidBodyInstance->GetCentreOfMass();

	if ( bIgnoreMass )
	{
        m_pxRigidBodyInstance->ApplyVelocityChangeAtRelativePosition( xRelPosition, xScaledImpulse );
	}
	else
	{
		m_pxRigidBodyInstance->ApplyImpulseAtRelativePosition( xRelPosition, xScaledImpulse );
	}

	m_pxRigidBodyInstance->WakeUp();
}

void Asura_Physics_Object::ApplyRotationalImpulse( const Asura_Euler_3& xAngularVelocity, bool bIgnoreMass /*=true*/  )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// assert on this?
	if ( !m_pxRigidBodyInstance->IsDynamic() ) return;

	Asura_AVector3 xScaledAngularVelocity( xAngularVelocity.x, xAngularVelocity.y, xAngularVelocity.z );
		
	xScaledAngularVelocity *= Asura_Physics_System::GetTimeSlice();

	if (!bIgnoreMass) xScaledAngularVelocity *= m_pxRigidBodyInstance->GetInvMass();

	m_pxRigidBodyInstance->ApplyAngularVelocityChange( xScaledAngularVelocity );
	m_pxRigidBodyInstance->WakeUp();
}

// Get the 'jump velocity' of an actor - this is a separate internal velocity due to jumping & falling
void Asura_Physics_Object::GetActorJumpVelocity( Asura_Vector_3& xJumpVelocity ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// This isn't a valid for anything other than an actor
	Asura_Assert( m_pxRigidBodyInstance->m_bActor, "Code Bug: This method only valid for ACTORS" );

	xJumpVelocity = m_pxRigidBodyInstance->m_xJumpVelocity;
}

void Asura_Physics_Object::SetActorJumpVelocity( const Asura_Vector_3& xJumpVelocity )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// This isn't a valid for anything other than an actor
	Asura_Assert( m_pxRigidBodyInstance->m_bActor, "Code Bug: This method only valid for ACTORS" );

	m_pxRigidBodyInstance->m_xJumpVelocity = xJumpVelocity;
}

void Asura_Physics_Object::ZeroMomentum( void )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->ZeroMomentum();
}

void Asura_Physics_Object::ZeroAngularMomentum( void )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->ZeroAngularMomentum();
}

void Asura_Physics_Object::ZeroLinearMomentum( void )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->ZeroLinearMomentum();
}

bool Asura_Physics_Object::IsAsleep() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->IsAsleep();
}

void Asura_Physics_Object::WakeUp( void )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->WakeUp();
}

void Asura_Physics_Object::Sleep()
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->Sleep();
}

bool Asura_Physics_Object::IsIgnore() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->IsIgnore();
}

void Asura_Physics_Object::SetIgnore(bool bIgnore)
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetIgnore( bIgnore );
}

	
bool Asura_Physics_Object::FeelsGravity() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	return m_pxRigidBodyInstance->m_bFeelsGravity;
}

void Asura_Physics_Object::SetFeelsGravity(bool bFeelsGravity)
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	
	m_pxRigidBodyInstance->m_bFeelsGravity = bFeelsGravity;

	if ( !bFeelsGravity )
	{
		m_pxRigidBodyInstance->m_xJumpVelocity.Zero();
	}
}

// KJFL - 22 February 2007 
//
// Default movement for an Actor is "super-slidy" - when an Actor hits a poly it will resolve 
// its movement direction against the obstacle and carry on moving with the same speed. E.g.
// If you ask an Actor to move at 6m/s along the ground, but it hits a slope, it will change
// direction to be going up the slope and move the same distance along the slope as it would 
// have on flat ground. Simiarly for running into walls - the Actor will slide along at normal
// speed.
//
// You can turn this behaviour off by setting the value below to false, which will make the 
// velocity be resolved against the obstacle normal inelastically. *HOWEVER* - Actors will 
// still be super-slidy on polys flagged as "PHYSICS WALKABLE SURFACE".
bool Asura_Physics_Object::GetActorIsSuperSlidy( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetActorIsSuperSlidy();
}
void Asura_Physics_Object::SetActorIsSuperSlidy( bool bSuperSlidy )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetActorIsSuperSlidy( bSuperSlidy );
}

// KJFL - 22 February 2007  
//
// Actors can be flagged to treat any polygons which are steep slopes (beyond 45 degrees
// currently) as vertical walls - so rather than going a bit up the slope and then sliding
// down again, they will slide perpendicularly as if it was a vertical wall in front of them.
// The default behaviour is not to do this (i.e. false in the value below) but this is for
// purely historical reasons to keep older projects as-is. Note this doesn't affect Actors
// when sliding down steep slopes - just when they're trying to go up them.
//
// Note: Polys flagged as "PHYSICS WALKABLE SURFACE" are always counted as floor irrespective
// of this setting, and the Actor will be able to walk up them without sliding down.
bool Asura_Physics_Object::GetActorShouldTreatSteepSlopesAsVertical( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetActorShouldTreatSteepSlopesAsVertical();
}
void Asura_Physics_Object::SetActorShouldTreatSteepSlopesAsVertical( bool bTreatAsVertical )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetActorShouldTreatSteepSlopesAsVertical( bTreatAsVertical );
}

bool Asura_Physics_Object::GetActorMoveAsVehicle( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetActorMoveAsVehicle();
}
void Asura_Physics_Object::SetActorMoveAsVehicle( bool bMoveAsVehicle )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetActorMoveAsVehicle( bMoveAsVehicle );
}

bool Asura_Physics_Object::GetActorDontPushActors( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetActorDontPushActors();
}
void Asura_Physics_Object::SetActorDontPushActors( bool bDontPushActors )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetActorDontPushActors( bDontPushActors );
}


// KJFL - 29 May 2007 - This object should be treated as super-slidy by Actors.
// Can be used to stop Actors standing on each other's heads.
bool Asura_Physics_Object::GetActorsShouldSlideOffMe( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetActorsSlideOffMe();
}

void Asura_Physics_Object::SetActorsShouldSlideOffMe( bool bSlide )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetActorsSlideOffMe( bSlide );
}

// KJFL - 21 August 2007 - Actors should treat this object as a static obstacle when they're
// moving and slide off it. However, when the objects are intersecting etc or the object runs 
// into the Actor, masses are used as normal to determine the collision response.
bool Asura_Physics_Object::GetActorsDontPushMe( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetActorsDontPushMe();
}

void Asura_Physics_Object::SetActorsDontPushMe( bool bDontPush ) 
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetActorsDontPushMe( bDontPush );
}


// if "ReportObjectCollisions" is true, any collisions with other objects will mean
// a report is generated and added to the linked list accessed by GetObjectCollisionReportList()
bool Asura_Physics_Object::ReportObjectCollisions( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetReportObjectCollisions();
}

void Asura_Physics_Object::SetReportObjectCollisions( bool bReport )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetReportObjectCollisions( bReport );
}

// returns a linked list of collision reports - see report class for more details
const Asura_Physics_ObjectCollisionReport* Asura_Physics_Object::GetObjectCollisionReportList( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetObjectCollisionReportList();
}

// if "ReportEnvironmentCollisions" is true, any collisions with the environments will mean
// a report is generated and added to the linked list accessed by GetEnvironmentCollisionReportList()
bool Asura_Physics_Object::ReportEnvironmentCollisions( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetReportEnvironmentCollisions();
}

void Asura_Physics_Object::SetReportEnvironmentCollisions( bool bReport )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetReportEnvironmentCollisions( bReport );
}


// returns a linked list of collision reports - see report class for more details
const Asura_Physics_EnvironmentCollisionReport* Asura_Physics_Object::GetEnvironmentCollisionReportList( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetEnvironmentCollisionReportList();
}

#ifdef ASURA_USE_COLLISIONFX
void Asura_Physics_Object::SetReportObjectCollisionFX( bool bReport )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetReportObjectCollisionFX( bReport );
}

bool Asura_Physics_Object::GetReportObjectCollisionFX( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetReportObjectCollisionFX();
}


void Asura_Physics_Object::SetReportEnvironmentCollisionFX( bool bReport )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetReportEnvironmentCollisionFX( bReport );
}
bool Asura_Physics_Object::GetReportEnvironmentCollisionFX( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetReportEnvironmentCollisionFX();
}
#endif

Asura_Guid Asura_Physics_Object::GetAssociatedEntityGuid( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetAssociatedEntityGuid();
}

void Asura_Physics_Object::SetAssociatedEntityGuid( Asura_Guid uGuid )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetAssociatedEntityGuid( uGuid );
}

bool Asura_Physics_Object::GetUseOwnGravityDirection( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetUseOwnGravityDirection();
}

void Asura_Physics_Object::SetUseOwnGravityDirection( bool bUseOwnGravityDirection )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetUseOwnGravityDirection( bUseOwnGravityDirection );
}

const Asura_AVector3& Asura_Physics_Object::GetOwnGravityDirection( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetOwnGravityDirection();
}

void Asura_Physics_Object::SetOwnGravityDirection( const Asura_AVector3& xDirection )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetOwnGravityDirection( xDirection );
}

const Asura_AVector3& Asura_Physics_Object::GetGravityDirection( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetGravityDirection();
}

// Get/Set the scale factor to use with gravity, e.g. 0.0f means no gravity,
// 2.0f means double gravity, -1.0f means gravity is inverted, etc.
float Asura_Physics_Object::GetGravityScale( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetGravityScale();
}

void Asura_Physics_Object::SetGravityScale( float fGravityScale )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetGravityScale( fGravityScale );
}

float Asura_Physics_Object::GetElasticity( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetElasticity();
}

void Asura_Physics_Object::SetElasticity( float fElasticity )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	fElasticity = Asura_Maths::Max( 0.0f, fElasticity );
	fElasticity = Asura_Maths::Min( 1.0f, fElasticity );

	m_pxRigidBodyInstance->SetElasticity( fElasticity );
}

float Asura_Physics_Object::GetDrag( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return 1.0f - m_pxRigidBodyInstance->GetDrag();
}

void Asura_Physics_Object::SetDrag( float fDrag )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	fDrag = 1.0f - fDrag;
	fDrag = Asura_Maths::Max( 0.0f, fDrag );
	fDrag = Asura_Maths::Min( 1.0f, fDrag );

	m_pxRigidBodyInstance->SetDrag( fDrag );
}

u_int Asura_Physics_Object::GetSupportFeature( Asura_AVector3* pxSupportPoints, const Asura_AVector3& xAxis ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetSupportFeature( pxSupportPoints, xAxis );
}

float Asura_Physics_Object::GetAngularDrag( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return 1.0f - m_pxRigidBodyInstance->GetAngularDrag();
}

void Asura_Physics_Object::SetAngularDrag( float fAngularDrag )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	fAngularDrag = 1.0f - fAngularDrag;
	fAngularDrag = Asura_Maths::Max( 0.0f, fAngularDrag );
	fAngularDrag = Asura_Maths::Min( 1.0f, fAngularDrag );

	m_pxRigidBodyInstance->SetAngularDrag( fAngularDrag );
}


float Asura_Physics_Object::GetFriction( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->GetFriction();
}

void Asura_Physics_Object::SetFriction( float fFriction )
{
    Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	fFriction = Asura_Maths::Max( 0.0f, fFriction );

	m_pxRigidBodyInstance->SetFriction( fFriction );
}

bool Asura_Physics_Object::IsInContactWithFloor( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->m_bIsInContactWithFloor;
}

bool Asura_Physics_Object::HasHitSomething( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// This isn't a valid updated flag for static object
	Asura_Assertf( !m_pxRigidBodyInstance->IsStatic(), ("Code Bug: The flag is not valid for static objects, guid '%u'", GetAssociatedEntityGuid()) );

	return m_pxRigidBodyInstance->HasHitSomething();
}

bool Asura_Physics_Object::IsSupported( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// This isn't a valid updated flag for static object
	Asura_Assertf( !m_pxRigidBodyInstance->IsStatic(), ("Code Bug: The flag is not valid for static objects, guid '%u'", GetAssociatedEntityGuid()) );

	return m_pxRigidBodyInstance->IsSupported();
}

bool Asura_Physics_Object::HasHitAnObject( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	// This isn't a valid updated flag for static object
	Asura_Assertf( !m_pxRigidBodyInstance->IsStatic(), ("Code Bug: The flag is not valid for static objects, guid '%u'", GetAssociatedEntityGuid()) );

	return m_pxRigidBodyInstance->HasHitAnObject();
}

bool Asura_Physics_Object::IsIgnoreDynamicObjects( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->m_bIgnoreDynamicObjects;
}

// Turn on/off collisions with dynamic objects 
// (meaning you can still collide with env, static & external controlled objects)
void Asura_Physics_Object::SetIgnoreDynamicObjects( bool bNewSetting )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->m_bIgnoreDynamicObjects = bNewSetting;
}

bool Asura_Physics_Object::IsIgnoreExternalControlObjects( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	return m_pxRigidBodyInstance->m_bIgnoreExternalControlObjects;
}

// Turn on/off collisions with dynamic objects 
// (meaning you can still collide with env, static & external controlled objects)
void Asura_Physics_Object::SetIgnoreExternalControlObjects( bool bNewSetting )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->m_bIgnoreExternalControlObjects = bNewSetting;
}

u_short Asura_Physics_Object::GetFlagsOfPolysToIgnore() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return(m_pxRigidBodyInstance->GetFlagsOfPolysToIgnore());
}

void Asura_Physics_Object::SetFlagsOfPolysToIgnore( const u_short usFlagsOfPolysToIgnore )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetFlagsOfPolysToIgnore(usFlagsOfPolysToIgnore);
}

#ifdef ASURA_TOOLS
u_int Asura_Physics_Object::Tools_GetFlagsOfPolysToIgnore() const
{
    Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
    return ( m_pxRigidBodyInstance->Tools_GetFlagsOfPolysToIgnore() );
}

void Asura_Physics_Object::Tools_SetFlagsOfPolysToIgnore( const u_int uFlagsOfPolysToIgnore )
{
    Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
    m_pxRigidBodyInstance->Tools_SetFlagsOfPolysToIgnore( uFlagsOfPolysToIgnore );
}
#endif

float Asura_Physics_Object::GetInvMass( bool bReturnEffectiveValue ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetInvMass( bReturnEffectiveValue );
}

void Asura_Physics_Object::SetInvMass( float fInvMass )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );

	m_pxRigidBodyInstance->SetInvMass( fInvMass );
}


void Asura_Physics_Object::SetUseAsRestFrame( bool bUseAsRestFrame )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetUseAsRestFrame( bUseAsRestFrame );
}

bool Asura_Physics_Object::HasRestFrame( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return (m_pxRigidBodyInstance->HasRestFrame());
}

bool Asura_Physics_Object::HasRestFrameFrom(const Asura_Physics_Object* pxOtherObject) const
{
	Asura_Assert(m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	Asura_Assert(pxOtherObject, "Asura_Physics_Object::HasRestFrameFrom - other object is null");
	if (pxOtherObject)
	{
		Asura_Assert(pxOtherObject->m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
		return (m_pxRigidBodyInstance->HasRestFrameFrom(pxOtherObject->m_pxRigidBodyInstance));
	}
	return false;
}

const Asura_Physics_RestFrame* Asura_Physics_Object::GetOwnedRestFrame() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetOwnedRestFrame();
}

bool Asura_Physics_Object::HasMatchingRestFrame(const Asura_Physics_RestFrame* pxRestFrame) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return (m_pxRigidBodyInstance->HasMatchingRestFrame(pxRestFrame));
}

// Get/Set the object's collision groups bitfield. It will only collide with another object
// if when their collision groups are ANDed together the results are non-zero.
u_int Asura_Physics_Object::GetCollisionGroups( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetCollisionGroups();
}

void Asura_Physics_Object::SetCollisionGroups( u_int uGroupsBitfield )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetCollisionGroups( uGroupsBitfield );
}

u_int Asura_Physics_Object::GetCollisionGroupsToCollideWith( void ) const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetCollisionGroupsToCollideWith();
}

void Asura_Physics_Object::SetCollisionGroupsToCollideWith( u_int uGroupsBitfield )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetCollisionGroupsToCollideWith( uGroupsBitfield );
}

void Asura_Physics_Object::SetAllSurfacesWalkable( bool bSet )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetAllSurfacesWalkable( bSet );
}

bool Asura_Physics_Object::GetAllSurfacesWalkable() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetAllSurfacesWalkable();
}

void Asura_Physics_Object::SetNotWalkableSurfaceFlags( u_short usFlags )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetNotWalkableSurfaceFlags( usFlags );
}

u_short Asura_Physics_Object::GetNotWalkableSurfaceFlags() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetNotWalkableSurfaceFlags();
}

void Asura_Physics_Object::SetPolyFlagsToReturnForActorCollision( u_short usFlags )
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	m_pxRigidBodyInstance->SetPolyFlagsToReturnForActorCollision( usFlags );
}

u_short Asura_Physics_Object::GetPolyFlagsToReturnForActorCollision() const
{
	Asura_Assert( m_pxRigidBodyInstance, "Fatal error: Rigid body instance not allocated!" );
	return m_pxRigidBodyInstance->GetPolyFlagsToReturnForActorCollision();
}
