//
// Asura_Physics_RigidBody.h
//
// Created by KJFL - 2004-03-17 11:23:18
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_PHYSICS_RIGIDBODY_H__
#define __ASURA_PHYSICS_RIGIDBODY_H__

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"
#include "Core/Asura_Mempool.h"
#include "Network/Asura_Guid.h"
#include "Physics/Asura_Physics_BodyState.h"
#include "Physics/Asura_Physics_ContactCluster.h"
#include "Physics/Asura_Physics_EnvironmentCache.h"
#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
#include "Physics/Constraints/Asura_Physics_Constraint.h"
#endif

#include "ProjectSettings/Asura_PhysicsConfiguration.h"


// D E F I N E S /////////////////////////////////////////////////////////////

// code will assert if any edge length of bounding box is larger than this.
#ifndef ASURA_PHYSICS_MAX_EXPECTED_OBJECT_SIZE
#define ASURA_PHYSICS_MAX_EXPECTED_OBJECT_SIZE (257.0f)
#endif

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Physics_RigidBodyDefinition;
class Asura_Physics_RigidBody_BaseLink;
class Asura_Physics_ObjectCollisionReport;
class Asura_Physics_EnvironmentCollisionReport;
class Asura_Physics_Island;
class Asura_Physics_RestFrame;

// C L A S S E S /////////////////////////////////////////////////////////////
class Asura_Physics_RigidBodyInstance
{
	public:

		//! Create a rigid body instance - memory will be allocated from mempool
		//! An invalid handle will be returned if there was an error.
		static Asura_Handle CreateDynamicInstance( void );
		static Asura_Handle CreateStaticInstance( void );
		static Asura_Handle CreateExternallyControlledInstance( void );
		
		static Asura_Handle CreateVehicle( const Asura_Bounding_Box& xBB );

		//! Destroy a rigid body instance
		static void Destroy( const Asura_Handle iHandle );

#if defined(ASURA_USE_STREAMING)
		// MGH 10/4/2006: Zone has been unloaded. Clean up any remaining references to it
		// in the environment cache and contacts list.
		static void RemoveZone( u_int uZoneIndex );
		static void RemoveZoneFromPool( u_int uZoneIndex, Asura_Physics_RigidBodyInstance* pxPoolHead );
		void RemoveZoneFromInstance( u_int uZoneIndex );
#endif //defined(ASURA_USE_STREAMING)

		void FlushCachedData( void );

		//Resets the cached data for everything. Needed for when the physics environment is recalculated. RAR 06/05/2008
		static void FlushCachedDataForAllInstances();

		//!
		//! Query methods 
		//!

		static Asura_Physics_RigidBodyInstance* GetInstancePointer( Asura_Handle iHandle )
		{
			return GetPool()->GetPointer(iHandle);
		}


		inline bool IsIgnore( void ) const { return m_bIgnore; }
		
		inline bool IsDynamic( void ) const
		{
			return ! ( IsStatic() || IsExternallyControlled() );
		}

		inline bool IsActor( void ) const
		{
			return m_bActor;
		}

		//! Is the object 'Externally Controlled' e.g. like a lift or drawbridge, which acts
		//! like an object with infinite mass when colliding with other objects
		inline bool IsExternallyControlled( void ) const
		{
			return m_bExternallyControlled;
		}

		// is the object static, i.e. it never moves 
		inline bool IsStatic( void ) const
		{
			return m_bStatic;
		}

		//! Is the object inactive i.e. it hasn't moved for a set period of time?
		inline bool IsAsleep( void ) const 
		{
			return m_bAsleep;
		}
		
		bool IsWideAwake( void ) const;

		bool HasContacts( void ) const;

		// has collision mesh & rigid body definition been created?
		bool IsMeshDefinitionInvalid( void ) 
		{
			if (!m_pxDefinition)
			{
				Asura_Assertf( m_pxDefinition, ( "Rigid body definition not yet set for object with handle 0x%x (entity guid %u). Flagging as ignore.", m_iHandle, m_uEntityGuid ) );
				m_bIgnore = true;
			}

			return !m_pxDefinition;
		}

		//! what is the displacement due to the forces acting on the object?
		inline const Asura_AVector3& GetForceEffect() const
		{
			return m_xForceEffect;
		}

		inline const Asura_AMatrix3x3& GetOrientation() const
		{
			return m_xBodyState.m_xOrientation;
		}

		inline const Asura_AVector3& GetRenderingPosition() const
		{
			CalculateRenderingPosition();

			return m_xInstancePosition;
		}

		inline const Asura_AVector3& GetCentreOfMass() const
		{
			return m_xBodyState.m_xCentreOfMass;
		}
		inline const Asura_AVector3& GetPreviousCentreOfMass() const
		{
			return m_xBodyState.m_xPreviousCentreOfMass;
		}
	
		inline const Asura_AVector3& GetCollisionMeshCentre() const
		{
			return m_xBodyState.m_xCentreOfMass;
		}

		Asura_AVector3 GetHalfExtents( void ) const;
		
		u_int GetSupportFeature( Asura_AVector3* pxSupportPoints, const Asura_AVector3& xAxis ) const;
	

		inline const Asura_Physics_RigidBodyDefinition*	GetDefinition() const
		{
			return m_pxDefinition;
		}

        inline float GetInvMass( bool bReturnEffectiveValue = true ) const { return( bReturnEffectiveValue && (m_bStatic || m_bExternallyControlled) ? 0.0f : m_fInvMass ); }
		const Asura_AMatrix3x3& GetInvInertiaTensorDiagonal( void ) const;
	
        		
		//!
		//! Modifier methods
		//!

		
		void SetOrientation( const Asura_AMatrix3x3& xNewOrientation );
		void SetPosition( const Asura_AVector3& xNewPosition );
		void TeleportTo( const Asura_AVector3& xNewPosition );
	
		// Tell an object that is asleep to wake up and become active.
		// Note, this has an effect even if the object is already awake,
		// since it resets the 'time spent motionless' counter to zero.
		inline void WakeUp( void )
		{
			// static objects are always asleep
			if ( IsStatic() ) return;

			m_bAsleep = false;
			m_fTimeMotionless = 0.0f;
		}

		inline void Sleep()
		{
			m_bAsleep = true;
		}

		void SetIgnore( bool bIgnore );

		void SetActorIsSuperSlidy( bool bSuperSlidy ) { m_bActor_SuperSlidy = bSuperSlidy; }
		bool GetActorIsSuperSlidy( void ) const { return m_bActor_SuperSlidy; }

		void SetActorShouldTreatSteepSlopesAsVertical( bool bVertical ) { m_bActor_TreatSteepSlopesAsVertical = bVertical; }
		bool GetActorShouldTreatSteepSlopesAsVertical( void ) const { return m_bActor_TreatSteepSlopesAsVertical; }

		void SetActorMoveAsVehicle( bool bMoveAsVehicle ) { m_bActor_MoveAsVehicle = bMoveAsVehicle; }
		bool GetActorMoveAsVehicle( void ) const { return m_bActor_MoveAsVehicle; }

		void SetActorDontPushActors( bool bDontPushActors ) { m_bActor_DontPushActors = bDontPushActors; }
		bool GetActorDontPushActors( void ) const { return m_bActor_DontPushActors; }

		void SetLockYAxisToBeVertical( bool bLock )	{ m_bLockYAxisToBeVertical = bLock;	}

		void SetAllSurfacesWalkable( bool bSet ) { m_bAllSurfacesWalkable = bSet; }
		bool GetAllSurfacesWalkable() const { return m_bAllSurfacesWalkable; }

		void	SetNotWalkableSurfaceFlags( u_short usFlags ) { m_usNotWalkableSurfaceFlags = usFlags; }
		u_short GetNotWalkableSurfaceFlags() const { return m_usNotWalkableSurfaceFlags; }

		void	SetPolyFlagsToReturnForActorCollision( u_short usFlags ) { m_usPolyFlagsToReturnForActorCollision = usFlags; }
		u_short	GetPolyFlagsToReturnForActorCollision() const { return m_usPolyFlagsToReturnForActorCollision; }

		// Setting the inv mass also updates the inertia tensor
		void SetInvMass( float fInvMass );

		inline void ApplyVelocityChange( const Asura_AVector3& xDelta )
		{
			m_xBodyState.m_xLinearVelocity += xDelta;
		}

		inline void ApplyAngularVelocityChange( const Asura_AVector3& xDelta )
		{
			m_xBodyState.m_xAngularVelocity += xDelta;
		}

		inline void ApplyVelocityChangeAtRelativePosition( const Asura_AVector3& xRel, const Asura_AVector3& xDelta )
		{
			m_xBodyState.m_xLinearVelocity += xDelta;

			if ( GetInvMass() != 0.0f)
			{
				m_xBodyState.m_xAngularVelocity += (( xRel ^ xDelta )* GetInvInertiaTensorDiagonal()) / GetInvMass();
			}
		}
		inline void ApplyImpulseAtRelativePosition( const Asura_AVector3& xRel, const Asura_AVector3& xImpulse )
		{
			m_xBodyState.m_xLinearVelocity += xImpulse * GetInvMass();
			m_xBodyState.m_xAngularVelocity += ( xRel ^ xImpulse )* GetInvInertiaTensorDiagonal();
		}
	
		void GetVelocitiesOfBody( Asura_AVector3& xLinearVelocity, Asura_AVector3& xAngularVelocity ) const;

		Asura_AVector3 CalculateVelocityOfPointOnBody( const Asura_AVector3& xPoint ) const;

		inline void CalculateVelocityOfRelativePointOnBody( const Asura_AVector3& xRelPoint, Asura_AVector3& xVelocity) const
		{
			xVelocity = ( m_xBodyState.m_xLinearVelocity + ( m_xBodyState.m_xAngularVelocity ^ (xRelPoint) ));
		}

		void SetVelocitiesOfBody( const Asura_AVector3& xLinearVelocity, const Asura_AVector3& xAngularVelocity );
		void SetAngularVelocityOfBody( const Asura_AVector3& xAngularVelocity );
		void SetLinearVelocityOfBody( const Asura_AVector3& xLinearVelocity );

		void ZeroMomentum( void );
		void ZeroAngularMomentum( void );
		void ZeroLinearMomentum( void );

		// Get/Set the scale factor to use with gravity, e.g. 0.0f means no gravity,
		// 2.0f means double gravity, -1.0f means gravity is inverted, etc.
		float GetGravityScale( void ) const;
		void SetGravityScale( float fGravityScale );

		// Get gravity direction - it will either be the instance's own or the local value
		const Asura_AVector3& GetGravityDirection( void ) const;

		const Asura_AVector3& GetOwnGravityDirection( void ) const;
		void SetOwnGravityDirection( const Asura_AVector3& xDirection );

		bool GetUseOwnGravityDirection( void ) const { return m_bUseOwnGravityDirection; }
		void SetUseOwnGravityDirection( bool bUseOwnGravityDirection ) { m_bUseOwnGravityDirection = bUseOwnGravityDirection; }

		inline float GetElasticity( void ) const { return m_fElasticity; }
		void SetElasticity( float fElasticity ) { m_fElasticity = fElasticity; }
	
		inline float GetFriction( void ) const { return m_fFriction; }
		void SetFriction( float fFriction ) { m_fFriction = fFriction; }

		inline float GetDrag( void ) const { return m_fDrag; }
		void SetDrag( float fDrag ) { m_fDrag = fDrag; }

		inline float GetAngularDrag( void ) const { return m_fAngularDrag; }
		void SetAngularDrag( float fAngularDrag ) { m_fAngularDrag = fAngularDrag; }

		void InitialiseFromDefinition( void );

		void MoveBase( void );
		void MoveActor( void );

		// integrate the centre of mass and orientation using the given velocities
		void Integrate( const Asura_AVector3& xLinearVelocity, const Asura_AVector3& xAngularVelocity );

		void TestForNewContacts( void );
		void TestForNewContacts_SphereStack( void );

		void TestForNewContacts_OBB( void );

		void TestForNewContacts_ObjectVertices( void );
		void TestForNewContacts_EnvironmentVertices( void );
		void TestForNewContacts_Edges( void );

#if 0 // shouldn't be needed any more, but keeping around out of paranoia
/*
		float TestForNewContacts_LeastDistance_ObjectVertices( const Asura_AVector3& xDisplacement, const float fDistance );
		float TestForNewContacts_LeastDistance_EnvironmentVertices( const Asura_AVector3& xDisplacement, const float fDistance );
		float TestForNewContacts_LeastDistance_Edges( const Asura_AVector3& xDisplacement, const float fDistance );
*/
#endif
		void CalculateCollisionResponses( const Asura_Physics_ContactCluster* pxCluster, bool bInelastic );
#ifdef ASURA_PHYSICS_RESOLUTION_EXPERIMENT
		void CalculateCollisionResponses_Exp( const Asura_Physics_ContactCluster* pxCluster, bool bInelastic );
#endif
		void RelaxPositionConstraints( void );
		void CalculateAllCollisionResponses( bool bInelastic );

		void AddContactPosition( const Asura_AVector3& xPosition, const Asura_AVector3& xNormal, float fPenetration );


		void PositionProjection( const Asura_AVector3& xDisplacement, const Asura_AVector3 axContacts[], u_int uNumberOfContacts );

		void RegisterCollision( const Asura_AVector3& xNormal );
		void ClearObstructionFlags( void );

	
		void CalculateRenderingPosition( void ) const;


		void CheckSleepStatus( void );
		
		// access to instance pool
		static Asura_TrackingHandlePool<Asura_Physics_RigidBodyInstance>* GetPool()
		{
			return &s_xInstancePool;
		}

		Asura_Physics_RigidBodyInstance* GetNextInstance( void ) const
		{
			return m_pxNextInstance;
		}
	
		void RemoveFromCurrentList( void ) 
		{
			if ( m_pxNextInstance ) 
			{
				m_pxNextInstance->m_ppxPrevInstanceNextPointer = m_ppxPrevInstanceNextPointer;
			}
			Asura_Assert( m_ppxPrevInstanceNextPointer, "Code bug: list error" );

			if ( m_ppxPrevInstanceNextPointer )
			{
				*m_ppxPrevInstanceNextPointer = m_pxNextInstance;
			}
		}

		void AddToList( Asura_Physics_RigidBodyInstance** ppxListHead )
		{
			if (*ppxListHead)
			{
				(*ppxListHead)->m_ppxPrevInstanceNextPointer = &m_pxNextInstance;
			}

			m_pxNextInstance = *ppxListHead;
			m_ppxPrevInstanceNextPointer = ppxListHead;
            
			*ppxListHead = this;
		}

		static Asura_Physics_RigidBodyInstance*& GetDynamicList( void )
		{
			return s_pxDynamicList;
		}
		static Asura_Physics_RigidBodyInstance*& GetExternallyControlledList( void )
		{
			return s_pxExternallyControlledList;
		}
		static Asura_Physics_RigidBodyInstance*& GetStaticList( void )
		{
			return s_pxStaticList;
		}


		const Asura_AVector3 GetPositionOnInstanceFace( u_int uFace , u_int uVertex = 0 );
		void GetInstanceFaceNormal( Asura_AVector3& xNormal, const u_int uFace );

		// ART - 06/06/06 - Accessors for flags of polys to be ignored by this rigid-body.
		u_short GetFlagsOfPolysToIgnore( void ) const
		{
			return(m_xEnvironmentCache.GetFlagsOfPolysToIgnore());
		}
		void SetFlagsOfPolysToIgnore( const u_short usFlagsOfPolysToIgnore )
		{
			// If we change which polys to ignore, we have to flush the cache
			// since otherwise we might have invalid polys in there. The 
			// cache should be refilled next time it's updated.
			if ( usFlagsOfPolysToIgnore != GetFlagsOfPolysToIgnore() )
			{
				m_xEnvironmentCache.SetFlagsOfPolysToIgnore(usFlagsOfPolysToIgnore);
				m_xEnvironmentCache.Flush();
			}
		}

#ifdef ASURA_TOOLS
        u_int Tools_GetFlagsOfPolysToIgnore() const
        {
            return ( m_xEnvironmentCache.Tools_GetFlagsOfPolysToIgnore() );
        }
        void Tools_SetFlagsOfPolysToIgnore( const u_int uFlagsOfPolysToIgnore )
        {
            // If we change which polys to ignore, we have to flush the cache
            // since otherwise we might have invalid polys in there. The 
            // cache should be refilled next time it's updated.
            if ( uFlagsOfPolysToIgnore != Tools_GetFlagsOfPolysToIgnore() )
            {
                m_xEnvironmentCache.Tools_SetFlagsOfPolysToIgnore( uFlagsOfPolysToIgnore );
                m_xEnvironmentCache.Flush();
            }
        }
#endif

		void GetExtentsAlongAxis( const Asura_AVector3& xAxis, float& fMin, float& fMax ) const;

		void GetCentreAndHalfExtentAlongAxis( const Asura_AVector3& xAxis, float& fCentre, float& fHalfExtent ) const;

	
		bool HasHitSomething( void ) const
		{
			// m_bHasHitSomething gets set internally by objects hitting the environment OR another 
			// object. m_bHasHitAnObject only gets set on object-object collisions.
			return m_bHasHitSomething;
		}

		bool HasHitAnObject( void ) const
		{
			return m_bHasHitAnObject;
		}

		bool IsSupported( void ) const
		{
			return m_bIsSupported;
		}

		Asura_Physics_Island* GetIsland( void )
		{
			return m_pxIsland;
		}

		void SetIsland( Asura_Physics_Island* pxIsland )
		{
			m_pxIsland = pxIsland;
		}

		Asura_Handle GetHandle( void ) const 
		{
			return m_iHandle;
		}

		void SetHandleToIgnore( Asura_Handle iHandle )
		{
			m_iHandleToIgnore = iHandle;
		}

		Asura_Handle GetHandleToIgnore( void ) const
		{
			return m_iHandleToIgnore;
		}

		void SetAssociatedEntityGuid( Asura_Guid uGuid )
		{
			m_uEntityGuid = uGuid;
		}

		Asura_Guid GetAssociatedEntityGuid( void ) const
		{
			return m_uEntityGuid;
		}


		enum COLLISION_PROXY
		{
			COLLISION_PROXY_OBB = 0,
			COLLISION_PROXY_SPHERE = 1,
			COLLISION_PROXY_SPHERESTACK = 2, 
			COLLISION_PROXY_POLYGONAL = 3,
		};

		enum COLLISION_PROXY GetCollisionProxy( void ) const
		{
			return m_eCollisionProxy;
		}

		void SetCollisionProxy( enum COLLISION_PROXY eProxyType )
		{
			m_eCollisionProxy = eProxyType;
			CalculateInvInertiaTensorDiagonal();
			SetCollisionTestAABB();
		}

		void SetUseAsRestFrame( bool bUseAsRestFrame );
		bool HasOwnRestFrame( void ) const { return m_bOwnsRestFrame; }

		void LinkToRestFrame( Asura_Physics_RestFrame* pxRestFrame );
		void UnlinkFromRestFrame( void );
		void MoveWithRestFrame( void );

		void CreateOwnRestFrame( void );
		void DestroyOwnRestFrame( void );

		bool HasRestFrame( void ) const { return !(m_pxRestFrame == NULL); }

		bool HasRestFrameFrom(const Asura_Physics_RigidBodyInstance* pxOtherInstance) const;

		// Alternative for the rest frame test - get the rest frame from one instance, and see if it's the one in the other instance.
		// So the system is similar, only return the rest frame if we own it (i.e. things rest on us, not us on them).
		const Asura_Physics_RestFrame* GetOwnedRestFrame() const;
		bool HasMatchingRestFrame(const Asura_Physics_RestFrame* pxRestFrame) const;

		void SetReportObjectCollisions( bool bReport ) { m_bReportObjectCollisions = bReport; }
		bool GetReportObjectCollisions( void ) const
		{
			return m_bReportObjectCollisions && (!IsStatic()); 
		}

		void SetReportEnvironmentCollisions( bool bReport ) { m_bReportEnvironmentCollisions = bReport; }
		bool GetReportEnvironmentCollisions( void ) const
		{
			return m_bReportEnvironmentCollisions && (!IsStatic()); 
		}
		
		#ifdef ASURA_USE_COLLISIONFX
		void SetReportObjectCollisionFX( bool bReport ) { m_bReportObjectCollisionFX = bReport; }
		bool GetReportObjectCollisionFX( void ) const
		{
			return m_bReportObjectCollisionFX && (!IsStatic()); 
		}

		void SetReportEnvironmentCollisionFX( bool bReport ) { m_bReportEnvironmentCollisionFX = bReport; }
		bool GetReportEnvironmentCollisionFX( void ) const
		{
			return m_bReportEnvironmentCollisionFX && (!IsStatic()); 
		}
		#endif

		void SetActorsSlideOffMe( bool bSlide ) { m_bActorsSlideOffMe = bSlide; }
		bool GetActorsSlideOffMe( void ) const { return m_bActorsSlideOffMe; }

		void SetActorsDontPushMe( bool bDontPush ) { m_bActorsDontPushMe = bDontPush; }
		bool GetActorsDontPushMe( void ) const { return m_bActorsDontPushMe; }

		void SetIgnoreDynamicObjects( bool bIgnore ) { m_bIgnoreDynamicObjects = bIgnore; }
		bool GetIgnoreDynamicObjects( void ) const { return m_bIgnoreDynamicObjects; }

		const Asura_Physics_ObjectCollisionReport* GetObjectCollisionReportList( void ) const
		{
			return m_pxObjectCollisionReports;
		}

		const Asura_Physics_EnvironmentCollisionReport* GetEnvironmentCollisionReportList( void ) const
		{
			return m_pxEnvironmentCollisionReports;
		}

		// clear this object's collision reports
		void ClearCollisionReports( void );

		void AddObjectCollisionReport( const Asura_Physics_RigidBodyInstance* xCollidee );

		const Asura_Bounding_Box& GetBroadphaseAABB( void ) const
		{
			return m_xBoundingBox;
		}

		void SetCollisionGroups( u_int uGroups ) { m_uCollisionGroupsBelongedTo = uGroups; }
		u_int GetCollisionGroups( void ) const { return m_uCollisionGroupsBelongedTo; }
		
		void SetCollisionGroupsToCollideWith( u_int uGroups ) { m_uCollisionGroupsToCollideWith = uGroups; }
		u_int GetCollisionGroupsToCollideWith( void ) const { return m_uCollisionGroupsToCollideWith; }

		bool CollisionGroupsAllowCollision( const Asura_Physics_RigidBodyInstance* pxOther ) const
		{
			// should I collide with his group(s)?
			if ( ( GetCollisionGroupsToCollideWith() & pxOther->GetCollisionGroups() ) == 0 ) return false;

			// and should he collide with my group(s)?
			return ( GetCollisionGroups() & pxOther->GetCollisionGroupsToCollideWith() ) != 0;
		}

#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
		void AddConstraint( Asura_Physics_Constraint* pxConstraint );
		void RemoveConstraint( Asura_Physics_Constraint* pxConstraint );
		void ResolveConstraints( void );

		void MergeIslandsUsingConstraints( void );

#endif
		
		Asura_Physics_BodyState* GetBodyState( void ) 
		{
			return &(m_xBodyState);
		}

	private:

		friend class Asura_Physics_Solver;

		void AddEnvironmentCollisionReport( u_int uFaceID, const Asura_AVector3& xNormal, const Asura_AVector3& xPosition );

        void SetCollisionTestAABB( bool bUpdateEnvCache = false );
		void HandleOrientationConstraints( void );

		void CalculateInvInertiaTensorDiagonal( void );
		void CalculateWorldInvInertiaTensor( void );

		// to improve memory packing, the large aligned data members come first...
		Asura_Physics_BodyState					m_xBodyState;
		
		// followed by all the AVectors...
		Asura_AVector3							m_xOwnGravityDirection;
		Asura_AVector3							m_xForceEffect;
	
		Asura_AVector3							m_xInvInertiaTensorDiagonal;
		Asura_AVector3							m_xJumpVelocity;

		mutable Asura_AVector3					m_xInstancePosition;
	
		Asura_Physics_RigidBodyInstance*		m_pxNextInstance;
		Asura_Physics_RigidBodyInstance**		m_ppxPrevInstanceNextPointer;

		// walkable stuff
		u_short									m_usPolyFlagsToReturnForActorCollision;
		u_short									m_usNotWalkableSurfaceFlags;

        union
        {
            struct
            {
		        u_int							m_bAllSurfacesWalkable : 1;

		        // sleep and ignore status
		        u_int							m_bAsleep :1;
		        u_int							m_bIgnore :1;
		        u_int							m_bHasAwakePartner :1;
        	
		        // behaviours
		        u_int							m_bStatic :1;
		        u_int							m_bExternallyControlled :1;
		        u_int							m_bActor :1;
		        u_int							m_bActor_MoveAsVehicle :1;
		        u_int							m_bActor_DontPushActors :1;
		        u_int							m_bOwnsRestFrame :1;
		        u_int							m_bActorsSlideOffMe : 1;
		        u_int							m_bActorsDontPushMe : 1;
		        u_int							m_bActor_SuperSlidy : 1;
		        u_int							m_bActor_TreatSteepSlopesAsVertical : 1;

		        // reports collision that will be held by the object?
		        u_int							m_bReportObjectCollisions :1;
		        u_int							m_bReportEnvironmentCollisions :1;

		        // report collisions that will go to the CollisionFX system?
		        // KJFL - NB: left outside #ifdef on purpose to keep track
		        // of total number of bitfield entries.
		        u_int							m_bReportObjectCollisionFX : 1;
		        u_int							m_bReportEnvironmentCollisionFX :1;
		        u_int							m_bReportedEnvCollisionFXThisFrame :1;
        		
		        // constraints
		        u_int							m_bAxiallyAligned :1;
		        u_int							m_bLockYAxisToBeVertical :1;
		        u_int							m_bSelfRighting :1;
		        u_int							m_bIgnoreExternalControlObjects :1;
		        u_int							m_bIgnoreDynamicObjects :1;
        	
		        // contact status
		        u_int							m_bWasInContactWithFloor :1;
		        u_int							m_bIsInContactWithFloor :1;
		        u_int							m_bHasHitSomething :1;
		        u_int							m_bHasHitAnObject :1;
		        u_int							m_bIsSupported :1;

		        // force effects
		        u_int							m_bFeelsGravity : 1;
		        u_int							m_bUseOwnGravityDirection : 1;
            };

            u_int m_uAllBitFlags;
        };

		// scale factor to apply to gravity strength for this instance
		float									m_fGravityScale;
		float									m_fElasticity;
		float									m_fFriction;
		float									m_fDrag;
		float									m_fAngularDrag;
	
		// KJFL - 24 October 2005 - trying to cut down the amount of stuff that needs access to members
		// that should really be completely private.
		friend class Asura_Physics_Object;
		friend class Asura_Physics_System;
		friend class Asura_Physics_Pairing;
		
		// needed to access private constructor/destructor
		friend class Asura_TrackingHandlePool_Node<Asura_Physics_RigidBodyInstance>;

		enum COLLISION_PROXY					m_eCollisionProxy;
		Asura_Physics_RigidBodyDefinition*		m_pxDefinition;

		Asura_AVector3*							m_pxObjectVertexPosition;
		Asura_AVector3*							m_pxObjectVertexPrevPosition;
		

		Asura_Physics_EnvironmentCache			m_xEnvironmentCache;
		
		Asura_Bounding_Box						m_xBoundingBox;

		float									m_fTimeMotionless;

		Asura_Handle							m_iHandle;
		Asura_Handle							m_iHandleToIgnore;
		Asura_Guid								m_uEntityGuid;

		Asura_Physics_ObjectCollisionReport*	m_pxObjectCollisionReports;
		Asura_Physics_EnvironmentCollisionReport* m_pxEnvironmentCollisionReports;

		// InvMass = 1.0f / Mass 
		// (so infinite mass means m_fInvMass = 0, and nothing can have zero mass)
		float									m_fInvMass; 

		Asura_Physics_Island*					m_pxIsland;

		Asura_Physics_RestFrame*				m_pxRestFrame;

		Asura_Physics_ContactCluster_List		m_xContactClusterList;
#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
		Asura_Physics_Constraint_List			m_xConstraintList;
#endif

		u_int									m_uCollisionGroupsBelongedTo;
		u_int									m_uCollisionGroupsToCollideWith;

#ifdef ASURA_PHYSICS_SUPPORT_UPDATE_FREEZE
        struct Backup;
        Backup* m_pxBackup;

        void DeallocateBackup( void );
        void StoreToBackup( void );
        void RestoreFromBackup( void );
#endif

		// Initialise, Reset and Shutdown are called solely by Asura_Physics_System. 
		// They handle the rigid body instance mempool etc
		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

		static void ResetAllCollisionReports( void );
	
		void DebuggingRender( void );

		//! private constructor/destructor so people have to go through our factory methods
		Asura_Physics_RigidBodyInstance(){};
		~Asura_Physics_RigidBodyInstance(){};

        // This must be called for all instances acquired from any of the pools.
		void InitialiseInstanceAfterCreation(Asura_Handle iHandle );

		static Asura_TrackingHandlePool<Asura_Physics_RigidBodyInstance>	s_xInstancePool;


		static Asura_Physics_RigidBodyInstance*	s_pxDynamicList;
		static Asura_Physics_RigidBodyInstance*	s_pxExternallyControlledList;
		static Asura_Physics_RigidBodyInstance*	s_pxStaticList;
};


#endif//__ASURA_PHYSICS_RIGIDBODY_H__
