//
// Asura_Physics_Solver.cpp
//
// Created by KJFL - 28 May 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Physics/Asura_Physics_Solver.h"

#include "Debugging/Asura_Assert.h"
#include "Physics/Asura_Physics_Pairing.h"
#include "Physics/Asura_Physics_Island.h"
#include "Physics/Asura_Physics_Ragdoll.h"
//#include "Debugging/Asura_CPUProfile.h"

#ifndef ASURA_PHYSICS_USE_OLD_SOLVER

//#include "Core/Asura_Multithread.h"




//static Asura_Physics_Solver_Data	g_xTask __attribute__(( aligned(16) ));


static Asura_Physics_Solver_Data*				g_pxSolverDataHeader;

static u_int									g_uNumberOfSingleBodyContacts;
static Asura_Physics_Solver_SingleBodyContact*	g_pxSingleBodyContacts;

static u_int									g_uNumberOfPairingContacts;
static Asura_Physics_Solver_PairingContact*		g_pxPairingContacts;

static u_int									g_uNumberOfConeTwistConstraints;
static Asura_Physics_Solver_ConeTwistConstraint* g_pxConeTwistConstraints;

static u_int									g_uNumberOfSolverBodies;
static Asura_Physics_Solver_Body*				g_pxBodies;

u_int g_uPeakMemUsage = 0;

void Asura_Physics_Solver::Initialise( void )
{
	Platform_Initialise();
}

void Asura_Physics_Solver::Reset()
{
	// KJFL - 11 June 2008 - not sure we actually want to do anything here - we have a fixed buffer size
	// so unless memory is getting fragmented around it, it should be safe to keep. Depends on what other
	// systems are allocating memory prior to Asura_Physics_System::Initialise().
	// Otherwise we can release/reallocate if that would be better by uncommenting the two lines below.

	// Shutdown();
	// Initialise();
}


void Asura_Physics_Solver::Shutdown( void )
{
	Platform_Shutdown();
}


void Asura_Physics_Solver::Begin( void )
{
    Platform_Begin();
}

void Asura_Physics_Solver::End( void )
{
    Platform_End();
}

Asura_Physics_Solver_Body* Asura_Physics_Solver::FindSolverBody( Asura_Physics_RigidBodyInstance* pxInstance, bool bAddIfNotFound ) 
{
	// is this a body we've used already?
	// TODO - may become a costly test for very large islands!
 
	for (u_int u=0; u<g_uNumberOfSolverBodies; ++u)
	{
		if ( g_pxBodies[u].m_pxParentBody == pxInstance->GetBodyState() )
		{
			return &g_pxBodies[u];
		}
	}

	// body not found - add it
	if ( bAddIfNotFound )
	{
		g_pxBodies[g_uNumberOfSolverBodies].Initialise( pxInstance );

		return &g_pxBodies[g_uNumberOfSolverBodies++];
	}

	Asura_Assert( false, "Code bug: body not found" );

	// Panic - return first body so we have a valid pointer. This is not ideal as the simulation
	// will be wrong, but something has gone badly wrong anyway if we're at this point.
	return &g_pxBodies[0];
}

void Asura_Physics_Solver::SolveIsland( Asura_Physics_Island* pxIsland )
{
	// sanity check - assert that we haven't somehow got a pairing list but not an object list
	Asura_Assert( !( (pxIsland->m_pxPairingNode != 0) && (pxIsland->m_pxObjectNode == 0) ), "Code bug: pairing with objects?" );
	if ( pxIsland->m_pxObjectNode == 0 ) return;

	// If there are no pairings, scan through the updatable objects and trivial reject if they are *all* asleep
	if ( pxIsland->m_pxPairingNode == 0 )
	{
		bool bAllAsleep = true;
		for( Asura_Physics_Island_ObjectNode* pxNode = pxIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
		{
			const Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxObject;

			if ( !pxInstance->IsAsleep() )
			{
				bAllAsleep = false;
				break;
			}
		}
		if (bAllAsleep) return;
	}

	
	u_int u1 = sizeof(Asura_Physics_Solver_Body);
	u_int u2 = sizeof(Asura_Physics_Solver_SingleBodyContact);
	u_int u3 = sizeof(Asura_Physics_Solver_PairingContact);

	// These alignment cases are currently for Asura_AVector3 to be aligned, but we may need a larger alignment
	// value for PS3/SPU allocations (at which point a platform layer becomes necessary). Similarly the reference
	// version of Asura_AVector3 may only be 12 bytes in size (i.e. not actually aligned as a 128bit structure).
	// TODO - a platform function for GetRequiredAlignment()?
//	Asura_Assert( (u1&15) == 0, "Code bug, Asura_Physics_Solver_Body not a 16 byte multiple in size" );
//	Asura_Assert( (u2&15) == 0, "Code bug, Asura_Physics_SingleBodyContact not a 16 byte multiple in size" );
//	Asura_Assert( (u3&15) == 0, "Code bug, Asura_Physics_PairingContact not a 16 byte multiple in size" );

	g_uNumberOfSingleBodyContacts = 0;
	g_uNumberOfPairingContacts = 0;
	g_uNumberOfConeTwistConstraints = 0;
	g_uNumberOfSolverBodies = 0;

	g_pxSolverDataHeader = (Asura_Physics_Solver_Data*)( Platform_GetSolverBuffer() );

	g_pxBodies = (Asura_Physics_Solver_Body*)((u_char*)g_pxSolverDataHeader + sizeof( Asura_Physics_Solver_Data));
//	Asura_Assert( (reinterpret_cast<u_int>(g_pxBodies)&15) == 0, "Code bug, alignment wrong" );


	// scan over the updatable objects and add them to the body list
	for( Asura_Physics_Island_ObjectNode* pxNode = pxIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
	{
		Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxObject;

		Asura_Physics_Solver_Body& xBody = g_pxBodies[g_uNumberOfSolverBodies++];

		xBody.Initialise( pxInstance );

	}

	// Scan over the pairing objects and add them to the body list if they're not already accounted for.
	// Originally we did this as we updating the pairings, but if we want one nice sequential buffer we need
	// to have all the bodies before we start dealing with constraints.
	if ( pxIsland->m_pxPairingNode )
	{
		for( Asura_Physics_Island_PairingNode* pxNode = pxIsland->m_pxPairingNode; pxNode!=0; pxNode = pxNode->m_pxNextPairingNode )
		{
			Asura_Physics_Pairing* pxPairing = pxNode->m_pxPairing;

			if ( pxPairing )
			{
				FindSolverBody( pxPairing->GetInstanceA(), true );
			 	FindSolverBody( pxPairing->GetInstanceB(), true );
			}
		}
	}

	// We should now have all the bodies we're going to reference in this solution.
	// Next up is the constraints between the bodies / world.

	{
	
		// First look for all the active pairings and get any constraints from them. This may wake up some objects,
		// which means we'll also later check for environment constraints.
		g_pxPairingContacts = (Asura_Physics_Solver_PairingContact*)( (u_char*)g_pxBodies + g_uNumberOfSolverBodies * sizeof( Asura_Physics_Solver_Body));
//		Asura_Assert( (reinterpret_cast<u_int>(g_pxPairingContacts)&15) == 0, "Code bug, alignment wrong" );

#if 01
		if ( pxIsland->m_pxPairingNode )
		{
			for( Asura_Physics_Island_PairingNode* pxNode = pxIsland->m_pxPairingNode; pxNode!=0; pxNode = pxNode->m_pxNextPairingNode )
			{
				Asura_Physics_Pairing* pxPairing = pxNode->m_pxPairing;

				// pairing ptr can be null if it's destroyed in current iterations
				if ( pxPairing )
				{
					/*bool bShouldDestroy =*/ pxPairing->Update();
				}

				u_int uBytes = u1*g_uNumberOfSolverBodies+u2*g_uNumberOfSingleBodyContacts+u3*g_uNumberOfPairingContacts;
				if ( uBytes > (Platform_GetSolverBufferSize()*3)/4 ) break;
			}
			
	
		}
#endif

		// Go through objects which aren't asleep and add any environment constraints.
		g_pxSingleBodyContacts = (Asura_Physics_Solver_SingleBodyContact*)((u_char*)g_pxPairingContacts + g_uNumberOfPairingContacts * sizeof( Asura_Physics_Solver_PairingContact));
//		Asura_Assert( (reinterpret_cast<u_int>(g_pxSingleBodyContacts)&15) == 0, "Code bug, alignment wrong" );

		u_int uBodyIndex = 0;
		for( Asura_Physics_Island_ObjectNode* pxNode = pxIsland->m_pxObjectNode; pxNode!=0; pxNode = pxNode->m_pxNextObjectNode )
		{
			Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxObject;

			if ( pxInstance->IsDynamic() && !pxInstance->IsAsleep() )
			{
				pxInstance->m_xContactClusterList.Reset();
				
				//TODO - constraints on objects would also be added here
	#ifdef ASURA_PHYSICS_SUPPORT_CONSTRAINTS
				// Resolve constraints
				pxInstance->ResolveConstraints();
	#endif

				// Collide with environment
				pxInstance->TestForNewContacts();

				if (pxInstance->IsActor()) // note - vehicles TODO
				{
					pxInstance->RelaxPositionConstraints(); 
				}
				else
				{
					for ( const Asura_Physics_ContactCluster* pxCluster = pxInstance->m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
					{
						Asura_Assert( pxInstance->GetBodyState() == g_pxBodies[uBodyIndex].m_pxParentBody, "Code bug: body mismatch");
						AddEnvironmentContacts( pxInstance,&g_pxBodies[uBodyIndex], pxCluster );
					}
				}

			}
			uBodyIndex++;

			u_int uBytes = u1*g_uNumberOfSolverBodies+u2*g_uNumberOfSingleBodyContacts+u3*g_uNumberOfPairingContacts;

			if ( uBytes > (Platform_GetSolverBufferSize()*3)/4 ) break;
		}
	}

	u_int uBytes = u1*g_uNumberOfSolverBodies+u2*g_uNumberOfSingleBodyContacts+u3*g_uNumberOfPairingContacts;

	g_uPeakMemUsage = Asura_Maths::Max( g_uPeakMemUsage, uBytes );

//	if ( g_uNumberOfSingleBodyContacts || g_uNumberOfPairingContacts )
	{
		g_pxSolverDataHeader->m_uNumberOfSolverBodies = g_uNumberOfSolverBodies;
		g_pxSolverDataHeader->m_uNumberOfPairingContacts = g_uNumberOfPairingContacts;
		g_pxSolverDataHeader->m_uNumberOfSingleBodyContacts = g_uNumberOfSingleBodyContacts;
		g_pxSolverDataHeader->m_uNumberOfConeTwistConstraints = 0;

		g_pxSolverDataHeader->m_uDataSizeInBytes = uBytes;
		g_pxSolverDataHeader->m_uDataAddress = (u_int)( (u_char*)g_pxSolverDataHeader + sizeof( Asura_Physics_Solver_Data ) );

		Asura_Assert( uBytes < Platform_GetSolverBufferSize(), "Code bug: buffer overflow" );

        if( !Platform_SolveAllContacts(g_pxSolverDataHeader) )
        {
		    Reference_SolveAllContacts( g_pxSolverDataHeader );
        }
	}
}

void Asura_Physics_Solver::SolveRagdoll( Asura_Physics_Ragdoll* pxRagdoll )
{
	u_int u1 = sizeof(Asura_Physics_Solver_Body);
	u_int u2 = sizeof(Asura_Physics_Solver_SingleBodyContact);
	u_int u3 = sizeof(Asura_Physics_Solver_PairingContact);
	u_int u4 = sizeof(Asura_Physics_Solver_ConeTwistConstraint);
 
	g_uNumberOfSingleBodyContacts = 0;
	g_uNumberOfPairingContacts = 0;
	g_uNumberOfConeTwistConstraints = 0;
	g_uNumberOfSolverBodies = 0;

	g_pxSolverDataHeader = (Asura_Physics_Solver_Data*)( Platform_GetSolverBuffer() );

	g_pxBodies = (Asura_Physics_Solver_Body*)((u_char*)g_pxSolverDataHeader + sizeof( Asura_Physics_Solver_Data));
	
	const u_int uNumberOfCollisionObjects = pxRagdoll->m_pxRagdollDesc->GetNumberOfCollisionObjects();

	for ( u_int u=0; u<uNumberOfCollisionObjects; ++u )
	{
		Asura_Physics_Solver_Body& xBody = g_pxBodies[g_uNumberOfSolverBodies++];
		xBody.Initialise( &(pxRagdoll->m_pxCollisionObjects[u]) );
	}

	{
		// First look for all the active pairings and get any constraints from them. This may wake up some objects,
		// which means we'll also later check for environment constraints.
		g_pxPairingContacts = (Asura_Physics_Solver_PairingContact*)( (u_char*)g_pxBodies + g_uNumberOfSolverBodies * sizeof( Asura_Physics_Solver_Body));

		// Go through objects which aren't asleep and add any environment constraints.
		g_pxSingleBodyContacts = (Asura_Physics_Solver_SingleBodyContact*)((u_char*)g_pxPairingContacts + g_uNumberOfPairingContacts * sizeof( Asura_Physics_Solver_PairingContact));

		for ( u_int u=0; u<uNumberOfCollisionObjects; ++u )
		{
			Asura_Physics_Ragdoll_CollisionObject& xInstance = pxRagdoll->m_pxCollisionObjects[u];

			for ( const Asura_Physics_ContactCluster* pxCluster = xInstance.m_xContactClusterList.GetFirstCluster(); pxCluster; pxCluster = pxCluster->GetNextCluster() )
			{
				// some things are constant over a whole cluster - normal, penetration distance, friction & elasticity values
				const Asura_AVector3& xNormal = pxCluster->GetContactNormal();
				const float fPenetration = pxCluster->GetPlanePenetrationConstant() - xInstance.GetPosition()*xNormal;

				const u_int uNumberOfContacts = pxCluster->GetNumberOfContacts();

				for (u_int v=0; v<uNumberOfContacts; ++v)
				{
					const Asura_AVector3 xRel = pxCluster->GetContactPosition(v) - xInstance.GetPosition();

					Asura_Physics_Solver_SingleBodyContact& xContact = g_pxSingleBodyContacts[g_uNumberOfSingleBodyContacts++];

					Asura_Physics_Solver::AddContact( xContact, &g_pxBodies[u], xRel, xNormal, fPenetration );
				}
			}
	
			u_int uBytes = u1*g_uNumberOfSolverBodies+u2*g_uNumberOfSingleBodyContacts+u3*g_uNumberOfPairingContacts;

			if ( uBytes > (Platform_GetSolverBufferSize()*3)/4 ) break;
		}
	}
	{
		g_pxConeTwistConstraints = (Asura_Physics_Solver_ConeTwistConstraint*)((u_char*)g_pxSingleBodyContacts + g_uNumberOfSingleBodyContacts * sizeof( Asura_Physics_Solver_SingleBodyContact));

		const u_int uNumberOfJoints = pxRagdoll->m_pxRagdollDesc->GetNumberOfJoints();

		for ( u_int u=0; u<uNumberOfJoints; ++u )
		{
			const Asura_Physics_Ragdoll_Desc_Joint& xJoint =  pxRagdoll->m_pxRagdollDesc->GetJoint(u);

			if (!xJoint.IsValid()) continue;

			Asura_Physics_Solver_ConeTwistConstraint& xContact = g_pxConeTwistConstraints[g_uNumberOfConeTwistConstraints++];

			pxRagdoll->SetupConeTwist( xContact, xJoint, g_pxBodies );
		}
	}

	u_int uBytes = u1*g_uNumberOfSolverBodies+u2*g_uNumberOfSingleBodyContacts+u3*g_uNumberOfPairingContacts+u4*g_uNumberOfConeTwistConstraints;

	g_uPeakMemUsage = Asura_Maths::Max( g_uPeakMemUsage, uBytes );

//	if ( g_uNumberOfSingleBodyContacts || g_uNumberOfPairingContacts )
	{
		g_pxSolverDataHeader->m_uNumberOfSolverBodies = g_uNumberOfSolverBodies;
		g_pxSolverDataHeader->m_uNumberOfPairingContacts = g_uNumberOfPairingContacts;
		g_pxSolverDataHeader->m_uNumberOfSingleBodyContacts = g_uNumberOfSingleBodyContacts;
		g_pxSolverDataHeader->m_uNumberOfConeTwistConstraints = g_uNumberOfConeTwistConstraints;

		g_pxSolverDataHeader->m_uDataSizeInBytes = uBytes;
		g_pxSolverDataHeader->m_uDataAddress = (u_int)( (u_char*)g_pxSolverDataHeader + sizeof( Asura_Physics_Solver_Data ) );

		Asura_Assert( uBytes < Platform_GetSolverBufferSize(), "Code bug: buffer overflow" );

        if( !Platform_SolveAllContacts(g_pxSolverDataHeader) )
        {
		    Reference_SolveAllContacts( g_pxSolverDataHeader );
        }
	}
}


void Asura_Physics_Solver::AddEnvironmentContacts( Asura_Physics_RigidBodyInstance* pxInstance, Asura_Physics_Solver_Body* pxBody, const Asura_Physics_ContactCluster* pxCluster )
{
	Asura_AVector3 xCentreOfMass;

	// TODO - wouldn't need to do this if could set CoM offset
#if 01
	if (pxInstance->IsActor() && pxInstance->GetActorMoveAsVehicle())
	{
		// fake CoM to be at pivot of Actor
		pxInstance->CalculateRenderingPosition();
		xCentreOfMass = pxInstance->m_xInstancePosition;
	}
	else
#endif
	{
		xCentreOfMass = pxInstance->m_xBodyState.m_xCentreOfMass;
	}

	// some things are constant over a whole cluster - normal, penetration distance, friction & elasticity values
	const Asura_AVector3& xNormal = pxCluster->GetContactNormal();
	const float fPenetration = pxCluster->GetPlanePenetrationConstant() - pxInstance->m_xBodyState.m_xCentreOfMass*xNormal;

	const u_int uNumberOfContacts = pxCluster->GetNumberOfContacts();

	for (u_int u=0; u<uNumberOfContacts; ++u)
	{
		const Asura_AVector3 xRel = pxCluster->GetContactPosition(u) - xCentreOfMass;

		Asura_Physics_Solver_SingleBodyContact& xContact = g_pxSingleBodyContacts[g_uNumberOfSingleBodyContacts++];
		AddContact( xContact, pxBody, xRel, xNormal, fPenetration );
	}
}

		
void Asura_Physics_Solver::AddContact( Asura_Physics_Solver_SingleBodyContact& xContact, Asura_Physics_Solver_Body* pxBody, const Asura_AVector3& xRel, const Asura_AVector3& xNormal, float fPenetration )
{
	fPenetration = Asura_Maths::Max( fPenetration, 0.0f );

	bool bInelastic = false;
	const float fElasticity = bInelastic ? 0.0f : pxBody->m_fElasticity;

	const Asura_AMatrix3x3& xInvInertiaTensor = pxBody->m_xInvInertiaTensor;

	xContact.m_pxBody = pxBody;

	xContact.m_xRel = xRel;
	xContact.m_xNormal = xNormal;

	Asura_AVector3 xVelocity;
	xVelocity = ( pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xRel) ));

	const float fNormalVelocity = xVelocity * xNormal;
	
	// Contact resolution- penetration and normal velocities
	xContact.m_fInvJacobianDiagonal_Normal = Asura_Physics_Solver_SingleBodyContact::CalculateJacobian( xRel, xNormal, xInvInertiaTensor, pxBody->m_fInvMass );
		
	xContact.m_fPenetration = fPenetration * 0.4f;
	xContact.m_fRestitution = Asura_Maths::Max( 0.0f, -fNormalVelocity * fElasticity );
		
	// If the restitution velocity is big enough that it would 
	// solve the penetration anyway, drop the penetration component.
	if ( xContact.m_fRestitution > xContact.m_fPenetration )
	{
		xContact.m_fPenetration = 0.0f;
	}

	// And then friction / tangential velocities

	// First find a good tangential vector on the plane of contact
	Asura_AVector3 xTangent = xVelocity - xNormal * fNormalVelocity;
	
	float fTangentialSpeed = xTangent.NormaliseReturningMagnitude();

	// TODO small epsilon needed here for stability?
	if ( fTangentialSpeed > 0.0f )
	{	
	} 
	else // pick arbitrary binormal and tangent
	{
		// Even though the tangential velocity is currently near zero,
		// we can't be sure that resolution of normal velocities (and
		// tangential velocities of other contacts) won't introduce
		// some velocity we need to damp during the iterative solver.
		// Hence the need for valid axes rather than just skipping
		// this contact.
		if (xNormal.GetZ()==0.0f && xNormal.GetX()==0.0f)
		{
			xTangent = Asura_AVector3( -xNormal.GetY(), xNormal.GetX(), 0.0f );
		}
		else
		{
			xTangent = Asura_AVector3( xNormal.GetZ(), 0.0f, -xNormal.GetX() );
		}
		xTangent.Normalise();
	}

	const Asura_AVector3 xBinormal = xTangent ^ xNormal;


	xContact.m_fInvJacobianDiagonal_Tangent = Asura_Physics_Solver_SingleBodyContact::CalculateJacobian( xRel, xTangent, xInvInertiaTensor, pxBody->m_fInvMass );
	xContact.m_xTangent = xTangent;

	xContact.m_fInvJacobianDiagonal_Binormal = Asura_Physics_Solver_SingleBodyContact::CalculateJacobian( xRel, xBinormal, xInvInertiaTensor, pxBody->m_fInvMass );
	xContact.m_xBinormal = xBinormal;
	
	// KJFL - 23 May 2008 - precalculated terms to avoid doing the same thing all the time when applying impulses
	xContact.m_xAngularComponent_Normal = ( xRel ^ xNormal ) * xInvInertiaTensor;
	xContact.m_xAngularComponent_Tangent = ( xRel ^ xTangent ) * xInvInertiaTensor;
	xContact.m_xAngularComponent_Binormal = ( xRel ^ xBinormal ) * xInvInertiaTensor;

	xContact.m_fAccumulatedImpulse_Normal = 0.0f;
	xContact.m_fAccumulatedImpulse_Tangent = 0.0f;
	xContact.m_fAccumulatedImpulse_Binormal = 0.0f;

	xContact.m_fAccumulatedSeparationImpulse_Normal = 0.0f;

}

void Asura_Physics_Solver::AddPairingContact( Asura_Physics_RigidBodyInstance* pxInstanceA, const Asura_AVector3& xRelA, Asura_Physics_RigidBodyInstance* pxInstanceB, const Asura_AVector3& xRelB, const Asura_AVector3& xNormal, float fPenetration  )
{
	fPenetration = Asura_Maths::Max( fPenetration, 0.0f );

	Asura_Physics_Solver_Body* pxBodyA = FindSolverBody( pxInstanceA );
	Asura_Physics_Solver_Body* pxBodyB = FindSolverBody( pxInstanceB );

	bool bInelastic = false;
	const float fElasticity = bInelastic ? 0.0f : pxBodyA->m_fElasticity*pxBodyB->m_fElasticity;

	const Asura_AMatrix3x3& xInvInertiaTensorA = pxBodyA->m_xInvInertiaTensor;
	const Asura_AMatrix3x3& xInvInertiaTensorB = pxBodyB->m_xInvInertiaTensor;

	Asura_Physics_Solver_PairingContact& xContact = g_pxPairingContacts[g_uNumberOfPairingContacts++];

	xContact.m_pxBodyA = pxBodyA;
	xContact.m_pxBodyB = pxBodyB;

	xContact.m_xRelA = xRelA;
	xContact.m_xRelB = xRelB;
	xContact.m_xNormal = xNormal;

	Asura_AVector3 xVelocity;
	xVelocity = ( pxBodyA->m_xLinearVelocity + ( pxBodyA->m_xAngularVelocity ^ (xRelA) ));
	xVelocity -= ( pxBodyB->m_xLinearVelocity + ( pxBodyB->m_xAngularVelocity ^ (xRelB) ));

	const float fNormalVelocity = xVelocity * xNormal;
	
	// Contact resolution- penetration and normal velocities
	xContact.m_fInvJacobianDiagonal_Normal = Asura_Physics_Solver_PairingContact::CalculateJacobian( xRelA, xRelB, xNormal, xInvInertiaTensorA, pxBodyA->m_fInvMass, xInvInertiaTensorB, pxBodyB->m_fInvMass );
		
	xContact.m_fPenetration = fPenetration * 0.4f;
	xContact.m_fRestitution = Asura_Maths::Max( 0.0f, -fNormalVelocity * fElasticity );
		
	// If the restitution velocity is big enough that it would 
	// solve the penetration anway, drop the penetration component.
	if ( xContact.m_fRestitution > xContact.m_fPenetration )
	{
		xContact.m_fPenetration = 0.0f;
	}

	// And then friction / tangential velocities

	// First find a good tangential vector on the plane of contact
	Asura_AVector3 xTangent = xVelocity - xNormal * fNormalVelocity;
	
	float fTangentialSpeed = xTangent.NormaliseReturningMagnitude();

	// TODO small epsilon needed here for stability?
	if ( fTangentialSpeed > 0.0f )
	{	
	} 
	else // pick arbitrary binormal and tangent
	{
		// Even though the tangential velocity is currently near zero,
		// we can't be sure that resolution of normal velocities (and
		// tangential velocities of other contacts) won't introduce
		// some velocity we need to damp during the iterative solver.
		// Hence the need for valid axes rather than just skipping
		// this contact.
		if (xNormal.GetZ()==0.0f && xNormal.GetX()==0.0f)
		{
			xTangent = Asura_AVector3( -xNormal.GetY(), xNormal.GetX(), 0.0f );
		}
		else
		{
			xTangent = Asura_AVector3( xNormal.GetZ(), 0.0f, -xNormal.GetX() );
		}
		xTangent.Normalise();
	}

	const Asura_AVector3 xBinormal = xTangent ^ xNormal;


	xContact.m_fInvJacobianDiagonal_Tangent = Asura_Physics_Solver_PairingContact::CalculateJacobian( xRelA, xRelB, xTangent, xInvInertiaTensorA, pxBodyA->m_fInvMass, xInvInertiaTensorB, pxBodyB->m_fInvMass );
	xContact.m_xTangent = xTangent;

	xContact.m_fInvJacobianDiagonal_Binormal = Asura_Physics_Solver_PairingContact::CalculateJacobian( xRelA, xRelB, xBinormal, xInvInertiaTensorA, pxBodyA->m_fInvMass, xInvInertiaTensorB, pxBodyB->m_fInvMass );
	xContact.m_xBinormal = xBinormal;
	
	// KJFL - 23 May 2008 - precalculated terms to avoid doing the same thing all the time when applying impulses
	xContact.m_xAngularComponentBodyA_Normal = ( xRelA ^ xNormal ) * xInvInertiaTensorA;
	xContact.m_xAngularComponentBodyA_Tangent = ( xRelA ^ xTangent ) * xInvInertiaTensorA;
	xContact.m_xAngularComponentBodyA_Binormal = ( xRelA ^ xBinormal ) * xInvInertiaTensorA;

	xContact.m_xAngularComponentBodyB_Normal = ( xRelB ^ xNormal ) * xInvInertiaTensorB;
	xContact.m_xAngularComponentBodyB_Tangent = ( xRelB ^ xTangent ) * xInvInertiaTensorB;
	xContact.m_xAngularComponentBodyB_Binormal = ( xRelB ^ xBinormal ) * xInvInertiaTensorB;


	xContact.m_fAccumulatedImpulse_Normal = 0.0f;
	xContact.m_fAccumulatedImpulse_Tangent = 0.0f;
	xContact.m_fAccumulatedImpulse_Binormal = 0.0f;

	xContact.m_fAccumulatedSeparationImpulse_Normal = 0.0f;
}

void Asura_Physics_Solver::Reference_SolveAllContacts( Asura_Physics_Solver_Data* pxSolverDataHeader )
{
	u_int uNumberOfSolverBodies = pxSolverDataHeader->m_uNumberOfSolverBodies;
	u_int uNumberOfPairingContacts = pxSolverDataHeader->m_uNumberOfPairingContacts;
	u_int uNumberOfSingleBodyContacts = pxSolverDataHeader->m_uNumberOfSingleBodyContacts;
	u_int uNumberOfConeTwistConstraints = pxSolverDataHeader->m_uNumberOfConeTwistConstraints;

	// if there is nothing to resolve, early out
    if( !uNumberOfPairingContacts && !uNumberOfSingleBodyContacts && !uNumberOfConeTwistConstraints )
    {
        return;
    }

	Asura_Physics_Solver_Body* pxBodies = (Asura_Physics_Solver_Body*)((u_char*)pxSolverDataHeader + sizeof( Asura_Physics_Solver_Data));
	Asura_Physics_Solver_PairingContact* pxPairingContacts = (Asura_Physics_Solver_PairingContact*)( (u_char*)pxBodies + uNumberOfSolverBodies * sizeof( Asura_Physics_Solver_Body));
	Asura_Physics_Solver_SingleBodyContact* pxSingleBodyContacts = (Asura_Physics_Solver_SingleBodyContact*)((u_char*)pxPairingContacts + uNumberOfPairingContacts * sizeof( Asura_Physics_Solver_PairingContact));
	Asura_Physics_Solver_ConeTwistConstraint* pxConeTwistConstraints = (Asura_Physics_Solver_ConeTwistConstraint*)((u_char*)pxSingleBodyContacts + uNumberOfSingleBodyContacts * sizeof( Asura_Physics_Solver_SingleBodyContact ) );

	// Precalculate anything we can before we do all the iterations
	for ( u_int u=0; u<uNumberOfConeTwistConstraints; ++u )
	{
		pxConeTwistConstraints[u].CalculateJacobians();
	}
	
	u_int uNumberOfIterations = 20;

	// Use the clamped contacts when the cone-twists are present (which currently is when it's a ragdoll)
	// KJFL - TODO - this is a trifle inelegant.
	bool bUseClampedContacts = (uNumberOfConeTwistConstraints > 0);

	for (u_int uIterations = 0; uIterations<uNumberOfIterations; ++uIterations)
	{
		// do all cone twist joints
		for ( u_int u=0; u<uNumberOfConeTwistConstraints; ++u )
		{
			Asura_Physics_Solver_ConeTwistConstraint::Solve( pxConeTwistConstraints[u] );
		}

		// do all pairing contacts
		for (u_int u=0; u<uNumberOfPairingContacts; ++u)
		{
			Asura_Physics_Solver_PairingContact& xContact = pxPairingContacts[u];

			Asura_Physics_Solver_PairingContact::Solve_LowPenetration( xContact );
		}
		
		if ( bUseClampedContacts )
		{
			// and then all vs environment contacts
			for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
			{
				Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];
	
				Asura_Physics_Solver_SingleBodyContact::Solve_Clamped( xContact );
			} 
		}
		else
		{
			// and then all vs environment contacts
			for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
			{
				Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];

				Asura_Physics_Solver_SingleBodyContact::Solve_LowPenetration( xContact );
			} 
		}

		// do all pairing contacts
		for (u_int u=0; u<uNumberOfPairingContacts; ++u)
		{
			Asura_Physics_Solver_PairingContact& xContact = pxPairingContacts[u];
			if (xContact.m_fPenetration >= fHIGH_PENETRATION_THRESHOLD )
			{
				Asura_Physics_Solver_PairingContact::Solve_HighPenetration( xContact );
			}

		}

		if ( !bUseClampedContacts )
		{
			// and then all vs environment contacts
			for (u_int u=0; u<uNumberOfSingleBodyContacts; ++u)
			{
				Asura_Physics_Solver_SingleBodyContact& xContact = pxSingleBodyContacts[u];
				if (xContact.m_fPenetration >= fHIGH_PENETRATION_THRESHOLD )
				{
					Asura_Physics_Solver_SingleBodyContact::Solve_HighPenetration( xContact );
				}

			} 
		}
		
	}

	// Write back the results to the original bodies
	for (u_int u=0; u<uNumberOfSolverBodies; ++u)
	{
		pxBodies[u].WriteBack();
	}

}

void Asura_Physics_Solver_Body::Initialise( Asura_Physics_RigidBodyInstance* pxBody )
{
	m_xInvInertiaTensor = pxBody->GetInvInertiaTensorDiagonal();
	
	pxBody->GetVelocitiesOfBody( m_xLinearVelocity, m_xAngularVelocity );
	
	m_xSeparationLinearVelocity.Zero();
	m_xSeparationAngularVelocity.Zero();

	m_fInvMass = pxBody->GetInvMass();
	m_fElasticity = pxBody->GetElasticity();
	m_fFriction = pxBody->GetFriction();

	m_pxParentBody = pxBody->GetBodyState();
}

void Asura_Physics_Solver_Body::Initialise( Asura_Physics_Ragdoll_CollisionObject* pxObject )
{
	m_xInvInertiaTensor = pxObject->GetWorldInvInertiaTensor();
	
	m_xLinearVelocity = pxObject->GetLinearVelocity();
	m_xAngularVelocity = pxObject->GetAngularVelocity();
	
	m_xSeparationLinearVelocity.Zero();
	m_xSeparationAngularVelocity.Zero();

	m_fInvMass = pxObject->GetInvMass();
	m_fElasticity = 0.5f;//pxObject->GetElasticity();
	m_fFriction = 1.0f;//pxObject->GetFriction();

	m_pxParentBody = pxObject->GetBodyState();
}


void Asura_Physics_Solver_Body::WriteBack( void )
{
	// skip any infinite mass objects
	if ( m_fInvMass == 0.0f ) return;

	m_pxParentBody->m_xLinearVelocity = m_xLinearVelocity;
	m_pxParentBody->m_xAngularVelocity = m_xAngularVelocity;

	m_pxParentBody->Integrate( m_xSeparationLinearVelocity, m_xSeparationAngularVelocity );
}

 	

#endif 
