//
// Asura_Physics_Solver.h
//
// Created by KJFL - 28 May 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_SOLVER_H__
#define __ASURA_PHYSICS_SOLVER_H__

#include "Core/Asura_Maths.h"

class Asura_Physics_RigidBodyInstance;
class Asura_Physics_Island;
class Asura_Physics_BodyState;
class Asura_Physics_ContactCluster;
class Asura_Physics_Solver_Body;
class Asura_Physics_Solver_Data;
class Asura_Physics_Solver_SingleBodyContact;
class Asura_Physics_Ragdoll;
class Asura_Physics_Ragdoll_CollisionObject;

class Asura_Physics_Solver
{
	public:

		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

		static void Begin( void );
		static void End( void );
		
		static void SolveIsland( Asura_Physics_Island* pxIsland );
		static void SolveRagdoll( Asura_Physics_Ragdoll* pxRagdoll );

		static void AddEnvironmentContacts( Asura_Physics_RigidBodyInstance* pxInstance, Asura_Physics_Solver_Body* pxBody, const Asura_Physics_ContactCluster* pxCluster );

		static void AddPairingContact( Asura_Physics_RigidBodyInstance* pxInstanceA, const Asura_AVector3& xRelA, Asura_Physics_RigidBodyInstance* pxInstanceB, const Asura_AVector3& xRelB, const Asura_AVector3& xNormal, float fPenetration  );

		static void Reference_SolveAllContacts( Asura_Physics_Solver_Data* pxSolverDataHeader );

		static void AddContact( Asura_Physics_Solver_SingleBodyContact& xContact, Asura_Physics_Solver_Body* pxBody, const Asura_AVector3& xRel, const Asura_AVector3& xNormal, float fPenetration );
		
	private:

		static void Platform_Initialise( void );
		static void Platform_Shutdown( void );

		static void Platform_Begin( void );
		static void Platform_End( void );

		static u_char* Platform_GetSolverBuffer( void );
		static u_int Platform_GetSolverBufferSize( void );

		static bool Platform_SolveAllContacts( Asura_Physics_Solver_Data* pxTask );
		

	


		static Asura_Physics_Solver_Body* FindSolverBody( Asura_Physics_RigidBodyInstance* pxInstance, bool bAddIfNotFound = false );

};


class Asura_Physics_Solver_Body
{
	public: // a "magic" 128 bytes currently
		
		void Initialise( Asura_Physics_RigidBodyInstance* pxBody );
		void Initialise( Asura_Physics_Ragdoll_CollisionObject* pxObject );
		void WriteBack( void );

		Asura_AMatrix3x3					m_xInvInertiaTensor;
		Asura_AVector3						m_xLinearVelocity;
		Asura_AVector3						m_xAngularVelocity;

		Asura_AVector3						m_xSeparationLinearVelocity;
		Asura_AVector3						m_xSeparationAngularVelocity;

		float								m_fInvMass;
		float								m_fElasticity;
		float								m_fFriction;

		// Reference for write-backs of velocities
		Asura_Physics_BodyState*			m_pxParentBody;

};

class Asura_Physics_Solver_ContactManifold
{
	public: // 24+(n*16) bytes where n=number of contacts
		// e.g. 248 bytes for 14 contacts
	
		Asura_AVector4	m_xNormal;
		u_int			m_uNumberOfContacts;
		float			m_fPenetration;

		// TODO - how to arrange this:
		// 1) Fixed size array of Asura_AVector4 contact positions?
		// 2) open ended "array follows this structure"?
		// 3) pointer to memory elsewhere
};


class Asura_Physics_Solver_SingleBodyContact
{
	public: // 160 bytes?

		static inline void Solve_Clamped( Asura_Physics_Solver_SingleBodyContact& xContact );
		static inline void Solve_LowPenetration( Asura_Physics_Solver_SingleBodyContact& xContact );
		static inline void Solve_HighPenetration( Asura_Physics_Solver_SingleBodyContact& xContact );
		Asura_AVector3	m_xRel;

		Asura_AVector3	m_xNormal;
		Asura_AVector3	m_xTangent;
		Asura_AVector3	m_xBinormal;

		Asura_AVector3	m_xAngularComponent_Normal;
		Asura_AVector3	m_xAngularComponent_Tangent;
		Asura_AVector3	m_xAngularComponent_Binormal;

		float			m_fAccumulatedImpulse_Normal;
		float			m_fAccumulatedSeparationImpulse_Normal;
		float			m_fInvJacobianDiagonal_Normal;
		
		float			m_fAccumulatedImpulse_Tangent;
		float			m_fInvJacobianDiagonal_Tangent;

		float			m_fAccumulatedImpulse_Binormal;
		float			m_fInvJacobianDiagonal_Binormal;
		
		float			m_fPenetration;
		float			m_fRestitution;
		
		Asura_Physics_Solver_Body* m_pxBody;

		// helper function for finding the Jacobian
		static inline float CalculateJacobian( const Asura_AVector3& xRel, const Asura_AVector3& xAxis, const Asura_AMatrix3x3& xInvInertiaTensor, float fInvMass )
		{
			Asura_AVector3 xDenom = xRel^xAxis;

			xDenom = xDenom*xInvInertiaTensor;

			xDenom = xDenom^xRel;

			float fInvJacobian = fInvMass + xDenom * xAxis;

			if (fInvJacobian != 0.0f )
			{
				fInvJacobian = 1.0f / fInvJacobian;
			}

			return fInvJacobian;
		}
};


class Asura_Physics_Solver_PairingContact
{
	public: // 224 bytes?

		static inline void Solve_LowPenetration( Asura_Physics_Solver_PairingContact& xContact );
		static inline void Solve_HighPenetration( Asura_Physics_Solver_PairingContact& xContact );

		Asura_AVector3	m_xRelA;
		Asura_AVector3	m_xRelB;

		Asura_AVector3	m_xNormal;
		Asura_AVector3	m_xTangent;
		Asura_AVector3	m_xBinormal;

		Asura_AVector3	m_xAngularComponentBodyA_Normal;
		Asura_AVector3	m_xAngularComponentBodyA_Tangent;
		Asura_AVector3	m_xAngularComponentBodyA_Binormal;

		Asura_AVector3	m_xAngularComponentBodyB_Normal;
		Asura_AVector3	m_xAngularComponentBodyB_Tangent;
		Asura_AVector3	m_xAngularComponentBodyB_Binormal;

		float			m_fAccumulatedImpulse_Normal;
		float			m_fAccumulatedSeparationImpulse_Normal;
		float			m_fInvJacobianDiagonal_Normal;

		float			m_fAccumulatedImpulse_Tangent;
		float			m_fInvJacobianDiagonal_Tangent;

		float			m_fAccumulatedImpulse_Binormal;
		float			m_fInvJacobianDiagonal_Binormal;
		
		float			m_fPenetration;
		float			m_fRestitution;

		Asura_Physics_Solver_Body* m_pxBodyA;
		Asura_Physics_Solver_Body* m_pxBodyB;

		
		// helper function for finding the Jacobian
		static inline float CalculateJacobian(	const Asura_AVector3& xRelA, const Asura_AVector3& xRelB, const Asura_AVector3& xAxis, 
												const Asura_AMatrix3x3& xInvInertiaTensorA, float fInvMassA, const Asura_AMatrix3x3& xInvInertiaTensorB, float fInvMassB )
		{
			Asura_AVector3 xDenomA = xRelA^xAxis;
			Asura_AVector3 xDenomB = xRelB^xAxis;

			xDenomA = xDenomA*xInvInertiaTensorA;
			xDenomB = xDenomB*xInvInertiaTensorB;

			Asura_AVector3 xDenom = (xDenomA ^ xRelA) + (xDenomB ^ xRelB);

			float fInvJacobian = fInvMassA + fInvMassB + ( xDenom * xAxis );

			if ( fInvJacobian != 0.0f )
			{
				fInvJacobian = 1.0f / fInvJacobian;
			}

			return fInvJacobian;
		}
	
};

class Asura_Physics_Solver_ConeTwistConstraint
{
	public:

		static inline void Solve( Asura_Physics_Solver_ConeTwistConstraint& xContact );

		void inline CalculateJacobians( void );

		Asura_AMatrix3x3 m_xAngularTransformBodyA;
		Asura_AMatrix3x3 m_xAngularTransformBodyB;

		Asura_AVector3	m_xRelA;
		Asura_AVector3	m_xRelB;
		Asura_AVector3	m_xTargetVelocity;

		Asura_AVector3	m_xSwingAxisX;
		Asura_AVector3  m_xSwingAxisY;
		Asura_AVector3	m_xTwistAxis;

		float			m_fInvJacobianX;
		float			m_fInvJacobianY;
		float			m_fInvJacobianZ;

		float			m_fSwingErrorX;
		float			m_fSwingAccumulatedImpulseX;
		float			m_fSwingInvJacobianDiagonalX;
		
		float			m_fSwingErrorY;
		float			m_fSwingAccumulatedImpulseY;
		float			m_fSwingInvJacobianDiagonalY;
	
		float			m_fTwistError;
		float			m_fAccumulatedTwistImpulse;
		float			m_fTwistInvJacobianDiagonal;

		Asura_Physics_Solver_Body* m_pxBodyA;
		Asura_Physics_Solver_Body* m_pxBodyB;

		u_int			m_bDoSwingX : 1;
		u_int			m_bDoSwingY : 1;
		u_int			m_bDoTwistZ : 1;
		
		// helper function for finding the Jacobian
		static inline float CalculateAngularJacobian( const Asura_AVector3& xAxis, const Asura_AMatrix3x3& xInvInertiaTensorA, const Asura_AMatrix3x3& xInvInertiaTensorB )
		{
			// TODO - we can optimise out a dot product if this is correct.
			// The linear one has more cross products, but this looks far more simple.
			float fDenom = xAxis * ( xAxis * xInvInertiaTensorA ) + xAxis * ( xAxis * xInvInertiaTensorB );

			if ( fDenom != 0.0f )
			{
				fDenom = 1.0f / fDenom;
			}

			return fDenom;
		}

		static inline float CalculateJacobian(	const Asura_AVector3& xRelA, const Asura_AVector3& xRelB, const Asura_AVector3& xAxis, 
												const Asura_AMatrix3x3& xInvInertiaTensorA, float fInvMassA, const Asura_AMatrix3x3& xInvInertiaTensorB, float fInvMassB )
		{
			Asura_AVector3 xDenomA = xRelA^xAxis;
			Asura_AVector3 xDenomB = xRelB^xAxis;

			xDenomA = xDenomA*xInvInertiaTensorA;
			xDenomB = xDenomB*xInvInertiaTensorB;

			Asura_AVector3 xDenom = (xDenomA ^ xRelA) + (xDenomB ^ xRelB);

			float fInvJacobian = fInvMassA + fInvMassB + ( xDenom * xAxis );

			if ( fInvJacobian != 0.0f )
			{
				fInvJacobian = 1.0f / fInvJacobian;
			}

			return fInvJacobian;
		}
};


void Asura_Physics_Solver_ConeTwistConstraint::CalculateJacobians( void )
{
	// KJFL - 11 February 2009 - although all this maths greatly simplifies the inner-loop calculations, we should be able to optimise this precalculation phase even further

	Asura_AVector3 xAxisX(1.0f,0.0f,0.0f);
	Asura_AVector3 xAxisY(0.0f,1.0f,0.0f);
	Asura_AVector3 xAxisZ(0.0f,0.0f,1.0f);

	m_fInvJacobianX = Asura_Physics_Solver_ConeTwistConstraint::CalculateJacobian( m_xRelA, m_xRelB, xAxisX, m_pxBodyA->m_xInvInertiaTensor, m_pxBodyA->m_fInvMass, m_pxBodyB->m_xInvInertiaTensor, m_pxBodyB->m_fInvMass );
	m_fInvJacobianY = Asura_Physics_Solver_ConeTwistConstraint::CalculateJacobian( m_xRelA, m_xRelB, xAxisY, m_pxBodyA->m_xInvInertiaTensor, m_pxBodyA->m_fInvMass, m_pxBodyB->m_xInvInertiaTensor, m_pxBodyB->m_fInvMass );
	m_fInvJacobianZ = Asura_Physics_Solver_ConeTwistConstraint::CalculateJacobian( m_xRelA, m_xRelB, xAxisZ, m_pxBodyA->m_xInvInertiaTensor, m_pxBodyA->m_fInvMass, m_pxBodyB->m_xInvInertiaTensor, m_pxBodyB->m_fInvMass );

	Asura_AVector3 xBodyAAngularComponentX = ( m_xRelA ^ xAxisX ) * m_pxBodyA->m_xInvInertiaTensor;
	Asura_AVector3 xBodyAAngularComponentY = ( m_xRelA ^ xAxisY ) * m_pxBodyA->m_xInvInertiaTensor;
	Asura_AVector3 xBodyAAngularComponentZ = ( m_xRelA ^ xAxisZ ) * m_pxBodyA->m_xInvInertiaTensor;
	Asura_AVector3 xBodyBAngularComponentX = ( m_xRelB ^ xAxisX ) * m_pxBodyB->m_xInvInertiaTensor;
	Asura_AVector3 xBodyBAngularComponentY = ( m_xRelB ^ xAxisY ) * m_pxBodyB->m_xInvInertiaTensor;
	Asura_AVector3 xBodyBAngularComponentZ = ( m_xRelB ^ xAxisZ ) * m_pxBodyB->m_xInvInertiaTensor;

	m_xAngularTransformBodyA = Asura_AMatrix3x3( xBodyAAngularComponentX, xBodyAAngularComponentY, xBodyAAngularComponentZ );
	m_xAngularTransformBodyB = Asura_AMatrix3x3( xBodyBAngularComponentX, xBodyBAngularComponentY, xBodyBAngularComponentZ );
}

void Asura_Physics_Solver_ConeTwistConstraint::Solve( Asura_Physics_Solver_ConeTwistConstraint& xContact )
{
	Asura_Physics_Solver_Body& xBodyA = *xContact.m_pxBodyA;
	Asura_Physics_Solver_Body& xBodyB = *xContact.m_pxBodyB;

	// Get the angular matrices for the contact - these are at the beginning of the contact class 
	// and so should help reduce future cache misses from reading contact data
	const Asura_AMatrix3x3 xMatrixA = xContact.m_xAngularTransformBodyA;
	const Asura_AMatrix3x3 xMatrixB = xContact.m_xAngularTransformBodyB;

	// Grab the velocities into local variables - hopefully these will end up in registers
	// and so we'll write them back at the end, rather than reading/writing to memory
	Asura_AVector3 xBodyA_LinearVelocity = xBodyA.m_xLinearVelocity;
	Asura_AVector3 xBodyA_AngularVelocity = xBodyA.m_xAngularVelocity;
	Asura_AVector3 xBodyB_LinearVelocity = xBodyB.m_xLinearVelocity;
	Asura_AVector3 xBodyB_AngularVelocity = xBodyB.m_xAngularVelocity;

	// Handle linear constraint
	{
		const float fInvMassA = xBodyA.m_fInvMass;
		const float fInvMassB = xBodyB.m_fInvMass;

		const Asura_AVector3 xVelocityA = xBodyA_LinearVelocity + ( xBodyA_AngularVelocity ^ (xContact.m_xRelA) );
		const Asura_AVector3 xVelocityB = xBodyB_LinearVelocity + ( xBodyB_AngularVelocity ^ (xContact.m_xRelB) );
		
		const Asura_AVector3 xVel = xContact.m_xTargetVelocity + (xVelocityA - xVelocityB);

		// KJFL - 11 February 2009 - TODO - Asura_AVector3 needs a componentwise multiply,
		/// it's the one thing stopping me from vectorising this entirely
		const float fImpulseX = xVel.GetX() * xContact.m_fInvJacobianX;
		const float fImpulseY = xVel.GetY() * xContact.m_fInvJacobianY;
		const float fImpulseZ = xVel.GetZ() * xContact.m_fInvJacobianZ;
		const Asura_AVector3 xImpulse( fImpulseX, fImpulseY, fImpulseZ );
		
		// KJFL - 11 February 2009 - a huge amount of maths has been compressed down into this!
		const Asura_AVector3 xAngularImpulseA = xImpulse*xMatrixA;
		const Asura_AVector3 xAngularImpulseB = xImpulse*xMatrixB;
		
		xBodyA_LinearVelocity -= xImpulse * fInvMassA;
		xBodyB_LinearVelocity += xImpulse * fInvMassB;		
		xBodyA_AngularVelocity -= xAngularImpulseA;
		xBodyB_AngularVelocity += xAngularImpulseB;
	}

	//if ( (xContact.m_bDoSwingX || xContact.m_bDoSwingY || xContact.m_bDoTwistZ ) )
	{
		// We'll be reusing these and they won't be changing... these are at the start of the body classes
		const Asura_AMatrix3x3 xBodyAInvInertiaTensor = xBodyA.m_xInvInertiaTensor;
		const Asura_AMatrix3x3 xBodyBInvInertiaTensor = xBodyB.m_xInvInertiaTensor;

		// Handle cone constraint - Swing X
		if ( xContact.m_bDoSwingX )
		{
			const Asura_AVector3 xRelativeAngularVelocity = xBodyA_AngularVelocity - xBodyB_AngularVelocity;
			float fRelativeSpeed = xContact.m_xSwingAxisX * xRelativeAngularVelocity*1.0f;

			float fImpulse = ( xContact.m_fSwingErrorX - fRelativeSpeed ) * xContact.m_fSwingInvJacobianDiagonalX;

			float fOldImpulse = xContact.m_fSwingAccumulatedImpulseX;
			xContact.m_fSwingAccumulatedImpulseX = Asura_Maths::Max( 0.0f, fOldImpulse + fImpulse );
			fImpulse = xContact.m_fSwingAccumulatedImpulseX - fOldImpulse;
		
			Asura_AVector3 xImpulse = xContact.m_xSwingAxisX * fImpulse;
			xBodyA_AngularVelocity += xImpulse * xBodyAInvInertiaTensor;
			xBodyB_AngularVelocity -= xImpulse * xBodyBInvInertiaTensor;
		}

		// Handle cone constraint - Swing Y
		if ( xContact.m_bDoSwingY )
		{
			const Asura_AVector3 xRelativeAngularVelocity = xBodyA_AngularVelocity - xBodyB_AngularVelocity;
			float fRelativeSpeed = xContact.m_xSwingAxisY * xRelativeAngularVelocity*1.0f;

			float fImpulse = ( xContact.m_fSwingErrorY - fRelativeSpeed ) * xContact.m_fSwingInvJacobianDiagonalY;

			float fOldImpulse = xContact.m_fSwingAccumulatedImpulseY;
			xContact.m_fSwingAccumulatedImpulseY = Asura_Maths::Max( 0.0f, fOldImpulse + fImpulse );
			fImpulse = xContact.m_fSwingAccumulatedImpulseY - fOldImpulse;
		
			Asura_AVector3 xImpulse = xContact.m_xSwingAxisY * fImpulse;
			xBodyA_AngularVelocity += xImpulse * xBodyAInvInertiaTensor;
			xBodyB_AngularVelocity -= xImpulse * xBodyBInvInertiaTensor;
		}

		// Handle twist constraint
		if ( xContact.m_bDoTwistZ )
		{
			const Asura_AVector3 xRelativeAngularVelocity = xBodyA_AngularVelocity - xBodyB_AngularVelocity;
			float fRelativeSpeed = xContact.m_xTwistAxis * xRelativeAngularVelocity*1.0f;

			float fImpulse = ( xContact.m_fTwistError - fRelativeSpeed ) * xContact.m_fTwistInvJacobianDiagonal;

			float fOldImpulse = xContact.m_fAccumulatedTwistImpulse;
			xContact.m_fAccumulatedTwistImpulse = Asura_Maths::Max( 0.0f, fOldImpulse + fImpulse );
			fImpulse = xContact.m_fAccumulatedTwistImpulse - fOldImpulse;
			
			Asura_AVector3 xImpulse = xContact.m_xTwistAxis * fImpulse;
			xBodyA_AngularVelocity += xImpulse * xBodyAInvInertiaTensor;
			xBodyB_AngularVelocity -= xImpulse * xBodyBInvInertiaTensor;
		}
	}


	xBodyA.m_xLinearVelocity = xBodyA_LinearVelocity;
	xBodyA.m_xAngularVelocity = xBodyA_AngularVelocity;

	xBodyB.m_xLinearVelocity = xBodyB_LinearVelocity;
	xBodyB.m_xAngularVelocity = xBodyB_AngularVelocity;
}


class Asura_Physics_Solver_Data
{
public:

	u_int	m_uNumberOfSolverBodies;
	u_int	m_uNumberOfPairingContacts;
	u_int	m_uNumberOfSingleBodyContacts;
	u_int	m_uNumberOfConeTwistConstraints;

	u_int	m_uDataSizeInBytes;
	u_int	m_uDataAddress;
	u_int	m_uDummy[2];
};


static const float								fHIGH_PENETRATION_THRESHOLD = 0.02f; 

void Asura_Physics_Solver_PairingContact::Solve_LowPenetration( Asura_Physics_Solver_PairingContact& xContact )
{
	Asura_Physics_Solver_Body& xBodyA = *xContact.m_pxBodyA;
	Asura_Physics_Solver_Body& xBodyB = *xContact.m_pxBodyB;

	// KJFL - 23 May 2008 - Given we're going to dot by xNormal, we could trade the cross product 
	// for a dot product using the vector triple product rule thingy. (a^b)*c = a*(b^c)
	Asura_AVector3 xImpactVelocity = xBodyA.m_xLinearVelocity + ( xBodyA.m_xAngularVelocity ^ (xContact.m_xRelA) );
	xImpactVelocity -= xBodyB.m_xLinearVelocity + ( xBodyB.m_xAngularVelocity ^ (xContact.m_xRelB) );
	float fNormalVelocity = xImpactVelocity * xContact.m_xNormal;

	float fPenetrationCorrection = xContact.m_fPenetration < fHIGH_PENETRATION_THRESHOLD ? xContact.m_fPenetration : 0.0f;
	float fVelocityCorrection = ( xContact.m_fRestitution - fNormalVelocity );

	float fNormalImpulse = ( fPenetrationCorrection + fVelocityCorrection ) * xContact.m_fInvJacobianDiagonal_Normal;
	float fOldNormalImpulse = xContact.m_fAccumulatedImpulse_Normal;

	xContact.m_fAccumulatedImpulse_Normal = Asura_Maths::Max( 0.0f, fOldNormalImpulse + fNormalImpulse );

	fNormalImpulse = xContact.m_fAccumulatedImpulse_Normal - fOldNormalImpulse;

	// KJFL - 23 May 2008 - you have to question whether a float compare and branch is worse
	// than a few vector ops...
	//if (fNormalImpulse != 0.0f)
	{

		xBodyA.m_xLinearVelocity += xContact.m_xNormal * fNormalImpulse * xBodyA.m_fInvMass;
		xBodyA.m_xAngularVelocity += xContact.m_xAngularComponentBodyA_Normal * fNormalImpulse;

		xBodyB.m_xLinearVelocity -= xContact.m_xNormal * fNormalImpulse * xBodyB.m_fInvMass;
		xBodyB.m_xAngularVelocity -= xContact.m_xAngularComponentBodyB_Normal * fNormalImpulse;
	}

	fNormalImpulse = xContact.m_fAccumulatedImpulse_Normal;

	if (fNormalImpulse > 0.0f)
	{
		float fFriction = xBodyA.m_fFriction * xBodyB.m_fFriction;
		float fFrictionLimit = fNormalImpulse * fFriction;

		{
			Asura_AVector3 xVelocity = xBodyA.m_xLinearVelocity + ( xBodyA.m_xAngularVelocity ^ (xContact.m_xRelA) );
			xVelocity -= xBodyB.m_xLinearVelocity + ( xBodyB.m_xAngularVelocity ^ (xContact.m_xRelB) );

			float fTangentialVelocity = xVelocity * xContact.m_xTangent;

			float fImpulse = -fTangentialVelocity * xContact.m_fInvJacobianDiagonal_Tangent;
			float fOldTangentialImpulse = xContact.m_fAccumulatedImpulse_Tangent;
			xContact.m_fAccumulatedImpulse_Tangent = fOldTangentialImpulse + fImpulse;

			xContact.m_fAccumulatedImpulse_Tangent = Asura_Maths::Max( xContact.m_fAccumulatedImpulse_Tangent, -fFrictionLimit );
			xContact.m_fAccumulatedImpulse_Tangent = Asura_Maths::Min( xContact.m_fAccumulatedImpulse_Tangent, +fFrictionLimit );

			fImpulse = xContact.m_fAccumulatedImpulse_Tangent - fOldTangentialImpulse;

			// KJFL - 23 May 2008 - you have to question whether a float compare and branch is worse
			// than a few vector ops...
			//	if (fImpulse != 0.0f)
			{
				//Asura_AVector3 xImpulse = xContact.m_xTangent * fImpulse;
				//ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );

				xBodyA.m_xLinearVelocity += xContact.m_xTangent * fImpulse * xBodyA.m_fInvMass;
				xBodyA.m_xAngularVelocity += xContact.m_xAngularComponentBodyA_Tangent * fImpulse;

				xBodyB.m_xLinearVelocity -= xContact.m_xTangent * fImpulse * xBodyB.m_fInvMass;
				xBodyB.m_xAngularVelocity -= xContact.m_xAngularComponentBodyB_Tangent * fImpulse;
			}
		}
		{
			Asura_AVector3 xVelocity = xBodyA.m_xLinearVelocity + ( xBodyA.m_xAngularVelocity ^ (xContact.m_xRelA) );
			xVelocity -= xBodyB.m_xLinearVelocity + ( xBodyB.m_xAngularVelocity ^ (xContact.m_xRelB) );

			float fTangentialVelocity = xVelocity * xContact.m_xBinormal;

			float fImpulse = -fTangentialVelocity * xContact.m_fInvJacobianDiagonal_Binormal;
			float fOldTangentialImpulse = xContact.m_fAccumulatedImpulse_Binormal;
			xContact.m_fAccumulatedImpulse_Binormal = fOldTangentialImpulse + fImpulse;

			xContact.m_fAccumulatedImpulse_Binormal = Asura_Maths::Max( xContact.m_fAccumulatedImpulse_Binormal, -fFrictionLimit );
			xContact.m_fAccumulatedImpulse_Binormal = Asura_Maths::Min( xContact.m_fAccumulatedImpulse_Binormal, +fFrictionLimit );

			fImpulse = xContact.m_fAccumulatedImpulse_Binormal - fOldTangentialImpulse;

			// KJFL - 23 May 2008 - you have to question whether a float compare and branch is worse
			// than a few vector ops...
			//	if (fImpulse != 0.0f)
			{
				//		Asura_AVector3 xImpulse = xContact.m_xBinormal * fImpulse;
				//		ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );
				xBodyA.m_xLinearVelocity += xContact.m_xBinormal * fImpulse * xBodyA.m_fInvMass;
				xBodyA.m_xAngularVelocity += xContact.m_xAngularComponentBodyA_Binormal * fImpulse;

				xBodyB.m_xLinearVelocity -= xContact.m_xBinormal * fImpulse * xBodyB.m_fInvMass;
				xBodyB.m_xAngularVelocity -= xContact.m_xAngularComponentBodyB_Binormal * fImpulse;
			}
		}
	}
} 

void Asura_Physics_Solver_PairingContact::Solve_HighPenetration( Asura_Physics_Solver_PairingContact& xContact )
{
	Asura_Physics_Solver_Body& xBodyA = *xContact.m_pxBodyA;
	Asura_Physics_Solver_Body& xBodyB = *xContact.m_pxBodyB;

	// KJFL - 23 May 2008 - Given we're going to dot by xNormal, we could trade the cross product 
	// for a dot product using the vector triple product rule thingy. (a^b)*c = a*(b^c)
	Asura_AVector3 xVelocity = xBodyA.m_xSeparationLinearVelocity + ( xBodyA.m_xSeparationAngularVelocity ^ (xContact.m_xRelA) );
	xVelocity -= xBodyB.m_xSeparationLinearVelocity + ( xBodyB.m_xSeparationAngularVelocity ^ (xContact.m_xRelB) );
	float fNormalVelocity = xVelocity * xContact.m_xNormal;

	//float fPenetrationCorrection = xContact.m_fPenetration ;
	//float fVelocityCorrection = ( xContact.m_fRestitution - fNormalVelocity );
	float fPenetrationCorrection = Asura_Maths::Max(xContact.m_fPenetration, xContact.m_fRestitution);
	float fVelocityCorrection = ( /*xContact.m_fRestitution*/ - fNormalVelocity );


	float fNormalImpulse = ( fPenetrationCorrection + fVelocityCorrection ) * xContact.m_fInvJacobianDiagonal_Normal;
	float fOldNormalImpulse = xContact.m_fAccumulatedSeparationImpulse_Normal;

	xContact.m_fAccumulatedSeparationImpulse_Normal = Asura_Maths::Max( 0.0f, fOldNormalImpulse + fNormalImpulse );

	fNormalImpulse = xContact.m_fAccumulatedSeparationImpulse_Normal - fOldNormalImpulse;

	xBodyA.m_xSeparationLinearVelocity += xContact.m_xNormal * fNormalImpulse * xBodyA.m_fInvMass;
	xBodyA.m_xSeparationAngularVelocity += xContact.m_xAngularComponentBodyA_Normal * fNormalImpulse;
	
	xBodyB.m_xSeparationLinearVelocity -= xContact.m_xNormal * fNormalImpulse * xBodyB.m_fInvMass;
	xBodyB.m_xSeparationAngularVelocity -= xContact.m_xAngularComponentBodyB_Normal * fNormalImpulse;
} 

void Asura_Physics_Solver_SingleBodyContact::Solve_Clamped( Asura_Physics_Solver_SingleBodyContact& xContact )
{
	Asura_Physics_Solver_Body* pxBody = xContact.m_pxBody;

	// KJFL - 23 May 2008 - Given we're going to dot by xNormal, we could trade the cross product 
	// for a dot product using the vector triple product rule thingy. (a^b)*c = a*(b^c)
	Asura_AVector3 xImpactVelocity = pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xContact.m_xRel) );
	float fNormalVelocity = xImpactVelocity * xContact.m_xNormal;
	
	float fPenetrationCorrection = Asura_Maths::Min( xContact.m_fPenetration, 0.01f );//fHIGH_PENETRATION_THRESHOLD );
	float fVelocityCorrection = ( xContact.m_fRestitution - fNormalVelocity );

	float fNormalImpulse = ( fPenetrationCorrection + fVelocityCorrection ) * xContact.m_fInvJacobianDiagonal_Normal;
	float fOldNormalImpulse = xContact.m_fAccumulatedImpulse_Normal;

	xContact.m_fAccumulatedImpulse_Normal = Asura_Maths::Max( 0.0f, fOldNormalImpulse + fNormalImpulse );

	fNormalImpulse = xContact.m_fAccumulatedImpulse_Normal - fOldNormalImpulse;

	pxBody->m_xLinearVelocity += xContact.m_xNormal * fNormalImpulse * pxBody->m_fInvMass;
	pxBody->m_xAngularVelocity += xContact.m_xAngularComponent_Normal * fNormalImpulse;

	fNormalImpulse = xContact.m_fAccumulatedImpulse_Normal;

	if (fNormalImpulse > 0.0f)
	{
		float fFriction = pxBody->m_fFriction;
		float fFrictionLimit = fNormalImpulse * fFriction;

		{
			Asura_AVector3 xVelocity = pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xContact.m_xRel) );

			float fTangentialVelocity = xVelocity * xContact.m_xTangent;

			float fImpulse = -fTangentialVelocity * xContact.m_fInvJacobianDiagonal_Tangent;
			float fOldTangentialImpulse = xContact.m_fAccumulatedImpulse_Tangent;
			xContact.m_fAccumulatedImpulse_Tangent = fOldTangentialImpulse + fImpulse;

			xContact.m_fAccumulatedImpulse_Tangent = Asura_Maths::Max( xContact.m_fAccumulatedImpulse_Tangent, -fFrictionLimit );
			xContact.m_fAccumulatedImpulse_Tangent = Asura_Maths::Min( xContact.m_fAccumulatedImpulse_Tangent, +fFrictionLimit );

			fImpulse = xContact.m_fAccumulatedImpulse_Tangent - fOldTangentialImpulse;
	
			pxBody->m_xLinearVelocity += xContact.m_xTangent * fImpulse * pxBody->m_fInvMass;
			pxBody->m_xAngularVelocity += xContact.m_xAngularComponent_Tangent * fImpulse;
		
		}
		{
			Asura_AVector3 xVelocity = pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xContact.m_xRel) );

			float fTangentialVelocity = xVelocity * xContact.m_xBinormal;
	
			float fImpulse = -fTangentialVelocity * xContact.m_fInvJacobianDiagonal_Binormal;
			float fOldTangentialImpulse = xContact.m_fAccumulatedImpulse_Binormal;
			xContact.m_fAccumulatedImpulse_Binormal = fOldTangentialImpulse + fImpulse;

			xContact.m_fAccumulatedImpulse_Binormal = Asura_Maths::Max( xContact.m_fAccumulatedImpulse_Binormal, -fFrictionLimit );
			xContact.m_fAccumulatedImpulse_Binormal = Asura_Maths::Min( xContact.m_fAccumulatedImpulse_Binormal, +fFrictionLimit );

			fImpulse = xContact.m_fAccumulatedImpulse_Binormal - fOldTangentialImpulse;
			
			pxBody->m_xLinearVelocity += xContact.m_xBinormal * fImpulse * pxBody->m_fInvMass;
			pxBody->m_xAngularVelocity += xContact.m_xAngularComponent_Binormal * fImpulse;
			
		}
	}
}

void Asura_Physics_Solver_SingleBodyContact::Solve_LowPenetration( Asura_Physics_Solver_SingleBodyContact& xContact )
{
	Asura_Physics_Solver_Body* pxBody = xContact.m_pxBody;

	// KJFL - 23 May 2008 - Given we're going to dot by xNormal, we could trade the cross product 
	// for a dot product using the vector triple product rule thingy. (a^b)*c = a*(b^c)
	Asura_AVector3 xImpactVelocity = pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xContact.m_xRel) );
	float fNormalVelocity = xImpactVelocity * xContact.m_xNormal;
	
	float fPenetrationCorrection = xContact.m_fPenetration < fHIGH_PENETRATION_THRESHOLD ? xContact.m_fPenetration : 0.0f;
	float fVelocityCorrection = ( xContact.m_fRestitution - fNormalVelocity );

	float fNormalImpulse = ( fPenetrationCorrection + fVelocityCorrection ) * xContact.m_fInvJacobianDiagonal_Normal;
	float fOldNormalImpulse = xContact.m_fAccumulatedImpulse_Normal;

	xContact.m_fAccumulatedImpulse_Normal = Asura_Maths::Max( 0.0f, fOldNormalImpulse + fNormalImpulse );

	fNormalImpulse = xContact.m_fAccumulatedImpulse_Normal - fOldNormalImpulse;

	// KJFL - 23 May 2008 - you have to question whether a float compare and branch is worse
	// than a few vector ops...
	//if (fNormalImpulse != 0.0f)
	{
	//	const Asura_AVector3 xImpulse = xNormal * fNormalImpulse;
	//	ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );
		pxBody->m_xLinearVelocity += xContact.m_xNormal * fNormalImpulse * pxBody->m_fInvMass;
		pxBody->m_xAngularVelocity += xContact.m_xAngularComponent_Normal * fNormalImpulse;
	}

	fNormalImpulse = xContact.m_fAccumulatedImpulse_Normal;

	if (fNormalImpulse > 0.0f)
	{
		float fFriction = pxBody->m_fFriction;
		float fFrictionLimit = fNormalImpulse * fFriction;

		{
			Asura_AVector3 xVelocity = pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xContact.m_xRel) );

			float fTangentialVelocity = xVelocity * xContact.m_xTangent;

			float fImpulse = -fTangentialVelocity * xContact.m_fInvJacobianDiagonal_Tangent;
			float fOldTangentialImpulse = xContact.m_fAccumulatedImpulse_Tangent;
			xContact.m_fAccumulatedImpulse_Tangent = fOldTangentialImpulse + fImpulse;

			xContact.m_fAccumulatedImpulse_Tangent = Asura_Maths::Max( xContact.m_fAccumulatedImpulse_Tangent, -fFrictionLimit );
			xContact.m_fAccumulatedImpulse_Tangent = Asura_Maths::Min( xContact.m_fAccumulatedImpulse_Tangent, +fFrictionLimit );

			fImpulse = xContact.m_fAccumulatedImpulse_Tangent - fOldTangentialImpulse;
			
			// KJFL - 23 May 2008 - you have to question whether a float compare and branch is worse
			// than a few vector ops...
		//	if (fImpulse != 0.0f)
			{
				//Asura_AVector3 xImpulse = xContact.m_xTangent * fImpulse;
				//ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );

				pxBody->m_xLinearVelocity += xContact.m_xTangent * fImpulse * pxBody->m_fInvMass;
				pxBody->m_xAngularVelocity += xContact.m_xAngularComponent_Tangent * fImpulse;
			}
		}
		{
			Asura_AVector3 xVelocity = pxBody->m_xLinearVelocity + ( pxBody->m_xAngularVelocity ^ (xContact.m_xRel) );

			float fTangentialVelocity = xVelocity * xContact.m_xBinormal;
	
			float fImpulse = -fTangentialVelocity * xContact.m_fInvJacobianDiagonal_Binormal;
			float fOldTangentialImpulse = xContact.m_fAccumulatedImpulse_Binormal;
			xContact.m_fAccumulatedImpulse_Binormal = fOldTangentialImpulse + fImpulse;

			xContact.m_fAccumulatedImpulse_Binormal = Asura_Maths::Max( xContact.m_fAccumulatedImpulse_Binormal, -fFrictionLimit );
			xContact.m_fAccumulatedImpulse_Binormal = Asura_Maths::Min( xContact.m_fAccumulatedImpulse_Binormal, +fFrictionLimit );

			fImpulse = xContact.m_fAccumulatedImpulse_Binormal - fOldTangentialImpulse;
			
			// KJFL - 23 May 2008 - you have to question whether a float compare and branch is worse
			// than a few vector ops...
		//	if (fImpulse != 0.0f)
			{
		//		Asura_AVector3 xImpulse = xContact.m_xBinormal * fImpulse;
		//		ApplyImpulseAtRelativePosition( xContact.m_xRel, xImpulse );
				pxBody->m_xLinearVelocity += xContact.m_xBinormal * fImpulse * pxBody->m_fInvMass;
				pxBody->m_xAngularVelocity += xContact.m_xAngularComponent_Binormal * fImpulse;
			}
		}
	}
}
void Asura_Physics_Solver_SingleBodyContact::Solve_HighPenetration( Asura_Physics_Solver_SingleBodyContact& xContact )
{
	Asura_Physics_Solver_Body* pxBody = xContact.m_pxBody;

	// KJFL - 23 May 2008 - Given we're going to dot by xNormal, we could trade the cross product 
	// for a dot product using the vector triple product rule thingy. (a^b)*c = a*(b^c)
	Asura_AVector3 xVelocity = pxBody->m_xSeparationLinearVelocity + ( pxBody->m_xSeparationAngularVelocity ^ (xContact.m_xRel) );
	float fNormalVelocity = xVelocity * xContact.m_xNormal;

	float fPenetrationCorrection = Asura_Maths::Max(xContact.m_fPenetration, xContact.m_fRestitution);
	float fVelocityCorrection = ( /*xContact.m_fRestitution*/ - fNormalVelocity );

	float fNormalImpulse = ( fPenetrationCorrection + fVelocityCorrection ) * xContact.m_fInvJacobianDiagonal_Normal;
	float fOldNormalImpulse = xContact.m_fAccumulatedSeparationImpulse_Normal;

	xContact.m_fAccumulatedSeparationImpulse_Normal = Asura_Maths::Max( 0.0f, fOldNormalImpulse + fNormalImpulse );

	fNormalImpulse = xContact.m_fAccumulatedSeparationImpulse_Normal - fOldNormalImpulse;

	pxBody->m_xSeparationLinearVelocity += xContact.m_xNormal * fNormalImpulse * pxBody->m_fInvMass;
	pxBody->m_xSeparationAngularVelocity += xContact.m_xAngularComponent_Normal * fNormalImpulse;
}

#endif//__ASURA_PHYSICS_SOLVER_H__

