//
// Asura_Physics_SpatialHash.h
//
// Created by KJFL - 01 September 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_SPATIALHASH_H__
#define __ASURA_PHYSICS_SPATIALHASH_H__

#include "Core/Asura_Maths.h"
#include "Physics/Asura_Physics_Broadphase.h"
#include "Physics/Asura_Physics_Pairing.h"
#include "Physics/Asura_Physics_RigidBody.h"

#include "ProjectSettings/Asura_PhysicsConfiguration.h"

#ifndef ASURA_PHYSICS_SPATIALHASHSIZE
#define ASURA_PHYSICS_SPATIALHASHSIZE (1024 * 128)
#endif

#ifndef ASURA_PHYSICS_SPATIALHASH_INITIAL_SCALE
#define ASURA_PHYSICS_SPATIALHASH_INITIAL_SCALE (0.25f)
#endif


class Asura_Physics_SpatialHash_Entry
{
	public:

		Asura_Physics_SpatialHash_Entry(){}
		~Asura_Physics_SpatialHash_Entry(){}

		Asura_Physics_RigidBodyInstance*	m_pxInstance;
		Asura_Hash_ID						m_uHash;

		// tempting to have an array index rather than a pointer
		Asura_Physics_SpatialHash_Entry*	m_pxNextEntry;

};



class Asura_Physics_SpatialHash
{
	static const u_int uHASH_ENTRIES = 31;

	public:

		Asura_Physics_SpatialHash()
		{
			for( u_int u=0; u<uHASH_ENTRIES; ++u )
			{
                m_apxFirstNodes[u] = 0;
			}

			m_pucBuffer = new u_char[ASURA_PHYSICS_SPATIALHASHSIZE];
			m_pucFreeMem = m_pucBuffer;

			m_fLargestObjectExtent = 0.0f;
			m_fBestGuessSpatialHashScale = ASURA_PHYSICS_SPATIALHASH_INITIAL_SCALE;
		}

		void Reset(void)
		{
			for( u_int u=0; u<uHASH_ENTRIES; ++u )
			{
                m_apxFirstNodes[u] = 0;
			}

			m_pucFreeMem = m_pucBuffer;

			if (m_fLargestObjectExtent)
			{
				m_fBestGuessSpatialHashScale = 1.0f / (m_fLargestObjectExtent);
	
			}
			else
			{
				m_fBestGuessSpatialHashScale = ASURA_PHYSICS_SPATIALHASH_INITIAL_SCALE;
			}
			
			m_fLargestObjectExtent = 0.0f;
		}

		void ResetBetweenLevels(void)
		{
			m_fLargestObjectExtent = 0.0f;
			Reset();
		}

		~Asura_Physics_SpatialHash()
		{
			delete [] m_pucBuffer;
		}

		

		inline void AddObjectWithHash( const Asura_Hash_ID uHash, Asura_Physics_RigidBodyInstance* pxInstance, const bool bMakePairings, const bool bAddObjectToHash )
		{
			u_int uHashEntry = uHash % uHASH_ENTRIES;
	
			Asura_Physics_SpatialHash_Entry* pxNode = m_apxFirstNodes[uHashEntry];
			
			if (!pxNode) // no node at this hash entry - make one
			{
				if ( bAddObjectToHash )
				{
					pxNode = (Asura_Physics_SpatialHash_Entry*)m_pucFreeMem;
					m_pucFreeMem += sizeof(Asura_Physics_SpatialHash_Entry);
					Asura_Assert(m_pucFreeMem < m_pucBuffer + ASURA_PHYSICS_SPATIALHASHSIZE, "spatial hash memory overrun");
					if (m_pucFreeMem >= m_pucBuffer + ASURA_PHYSICS_SPATIALHASHSIZE)
					{
						return;
					}
					
					pxNode->m_uHash = uHash;
					pxNode->m_pxInstance = pxInstance;
					pxNode->m_pxNextEntry = 0;

					m_apxFirstNodes[uHashEntry] = pxNode;
				}
				return;
			}
			else // hash entry already exists
			{
				if ( bMakePairings )
				{
					while ( pxNode )
					{
						// if we come across an entry with the same hash id, we could try making a pairing
						if ( pxNode->m_uHash == uHash )
						{
							// if we come across the same object, we've been here before and can exit out
							// may actually be worth doing this in a separate loop depending on how
							// costly the rest of this loop is.
							// New nodes are added to the head of the list, so hopefully if you end up with 
							// the same hash entry you can quick exit.
						
							if ( pxNode->m_pxInstance == pxInstance )
							{
								return;
							}

							if (pxInstance->GetHandle() < pxNode->m_pxInstance->GetHandle() )
							{
								Asura_Physics_Pairing::GetOrCreatePairing( pxInstance, pxNode->m_pxInstance );
							}
							else
							{	
								Asura_Physics_Pairing::GetOrCreatePairing( pxNode->m_pxInstance, pxInstance );
							}
						}

						pxNode = pxNode->m_pxNextEntry;
					}
				}
				else
				{
					// not interested in pairings - just exit if we've already got this object in the list
					
					while ( pxNode )
					{
						if ( pxNode->m_uHash == uHash )
						{
							if ( pxNode->m_pxInstance == pxInstance )
							{
								return;
							}
						}

						pxNode = pxNode->m_pxNextEntry;
					}
				}

				// got here? then we need to add this object to the list, for future objects to consider
				if ( bAddObjectToHash )
				{

					pxNode = (Asura_Physics_SpatialHash_Entry*)m_pucFreeMem;
					m_pucFreeMem += sizeof(Asura_Physics_SpatialHash_Entry);
					Asura_Assert(m_pucFreeMem < m_pucBuffer + ASURA_PHYSICS_SPATIALHASHSIZE, "spatial hash memory overrun");
					if (m_pucFreeMem >= m_pucBuffer + ASURA_PHYSICS_SPATIALHASHSIZE)
					{
						return;
					}
					
					pxNode->m_uHash = uHash;
					pxNode->m_pxInstance = pxInstance;
					pxNode->m_pxNextEntry = m_apxFirstNodes[uHashEntry];

					m_apxFirstNodes[uHashEntry] = pxNode;
				}

			}
		}

		void AddObject( Asura_Physics_RigidBodyInstance* pxInstance, const Asura_Bounding_Box& xBoundingBox, const bool bMakePairings = true, const bool bAddObjectToHash = true )
		{
			const float fWidth = xBoundingBox.MaxX - xBoundingBox.MinX;
			m_fLargestObjectExtent = Asura_Maths::Max( fWidth, m_fLargestObjectExtent );
			
			const float fHeight = xBoundingBox.MaxY - xBoundingBox.MinY;
			m_fLargestObjectExtent = Asura_Maths::Max( fHeight, m_fLargestObjectExtent );
			
			const float fLength = xBoundingBox.MaxZ - xBoundingBox.MinZ;
			m_fLargestObjectExtent = Asura_Maths::Max( fLength, m_fLargestObjectExtent );
			

			float fMaxDimension = Asura_Maths::Max( Asura_Maths::Max(fWidth, fHeight), fLength );

			if ( fMaxDimension > ASURA_PHYSICS_MAX_EXPECTED_OBJECT_SIZE )
			{
				Asura_Assertf( false, 
					("Very large object (max edge length %.02f metres) being added to Physics spatial hash. This is either a huge object and/or an object moving insanely fast. Entity guid is %d.",
					Asura_Maths::FtoD( fMaxDimension ), pxInstance->GetAssociatedEntityGuid() ));

				return;
			}

			// find ranges of bounding box
			int iMinX = Asura_Maths::FloatToInt(xBoundingBox.MinX * m_fBestGuessSpatialHashScale);
			int iMinY = Asura_Maths::FloatToInt(xBoundingBox.MinY * m_fBestGuessSpatialHashScale);
			int iMinZ = Asura_Maths::FloatToInt(xBoundingBox.MinZ * m_fBestGuessSpatialHashScale);

			int iMaxX = Asura_Maths::FloatToInt(xBoundingBox.MaxX * m_fBestGuessSpatialHashScale);
			int iMaxY = Asura_Maths::FloatToInt(xBoundingBox.MaxY * m_fBestGuessSpatialHashScale);
			int iMaxZ = Asura_Maths::FloatToInt(xBoundingBox.MaxZ * m_fBestGuessSpatialHashScale);

			for (int iX = iMinX; iX <= iMaxX; iX++)
			{
				for (int iY = iMinY; iY <= iMaxY; iY++)
				{
					for (int iZ = iMinZ; iZ <= iMaxZ; iZ++)
					{
						Asura_Hash_ID uHash = GetHash( iX, iY, iZ );

						AddObjectWithHash( uHash, pxInstance, bMakePairings, bAddObjectToHash );
					}
				}
			}
		}

		inline bool TestForIntersection( const Asura_Hash_ID uHash, const Asura_OBB& xOBB, Asura_Handle iHandleToIgnore )
		{
			u_int uHashEntry = uHash % uHASH_ENTRIES;

			for ( Asura_Physics_SpatialHash_Entry* pxNode = m_apxFirstNodes[uHashEntry]; pxNode!=0; pxNode = pxNode->m_pxNextEntry )
			{
				// if we come across an entry with the same hash id, we need to do an intersection test
				if ( pxNode->m_uHash == uHash )
				{
					// get the object
					const Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxInstance;

					// skip if we're meant to ignore this particular instance
					if ( iHandleToIgnore == pxInstance->GetHandle() ) continue;

					// check the obvious flags
					if ( pxInstance->IsIgnore() ) continue;

					Asura_OBB xInstanceOBB( pxInstance->GetHalfExtents(), pxInstance->GetCollisionMeshCentre(), pxInstance->GetOrientation() );

					if ( xInstanceOBB.Intersects( xOBB ) ) return true;

				}
			}

			return false;
		
		}

		bool DoesOBBIntersectAnyObjects( const Asura_OBB& xOBB, Asura_Handle iHandleToIgnore )
		{
			// find ranges of orientated bounding box
			Asura_Bounding_Box xAABB;

			xOBB.CalculateEnclosingAABB( xAABB );

			int iMinX = Asura_Maths::FloatToInt(xAABB.MinX * m_fBestGuessSpatialHashScale);
			int iMinY = Asura_Maths::FloatToInt(xAABB.MinY * m_fBestGuessSpatialHashScale);
			int iMinZ = Asura_Maths::FloatToInt(xAABB.MinZ * m_fBestGuessSpatialHashScale);

			int iMaxX = Asura_Maths::FloatToInt(xAABB.MaxX * m_fBestGuessSpatialHashScale);
			int iMaxY = Asura_Maths::FloatToInt(xAABB.MaxY * m_fBestGuessSpatialHashScale);
			int iMaxZ = Asura_Maths::FloatToInt(xAABB.MaxZ * m_fBestGuessSpatialHashScale);

			for (int iX = iMinX; iX <= iMaxX; iX++)
			{
				for (int iY = iMinY; iY <= iMaxY; iY++)
				{
					for (int iZ = iMinZ; iZ <= iMaxZ; iZ++)
					{
						Asura_Hash_ID uHash = GetHash( iX, iY, iZ );

						if (TestForIntersection( uHash, xOBB, iHandleToIgnore )) return true;
					}
				}
			}

			return false;
		}

		void BroadphaseQuery( const Asura_Bounding_Box& xAABB, Asura_Physics_Broadphase_Query* pxQuery, bool bConsiderOnlyExternalControl )
		{
			// find ranges of orientated bounding box
			int iMinX = Asura_Maths::FloatToInt(xAABB.MinX * m_fBestGuessSpatialHashScale);
			int iMinY = Asura_Maths::FloatToInt(xAABB.MinY * m_fBestGuessSpatialHashScale);
			int iMinZ = Asura_Maths::FloatToInt(xAABB.MinZ * m_fBestGuessSpatialHashScale);

			int iMaxX = Asura_Maths::FloatToInt(xAABB.MaxX * m_fBestGuessSpatialHashScale);
			int iMaxY = Asura_Maths::FloatToInt(xAABB.MaxY * m_fBestGuessSpatialHashScale);
			int iMaxZ = Asura_Maths::FloatToInt(xAABB.MaxZ * m_fBestGuessSpatialHashScale);

			for (int iX = iMinX; iX <= iMaxX; iX++)
			{
				for (int iY = iMinY; iY <= iMaxY; iY++)
				{
					for (int iZ = iMinZ; iZ <= iMaxZ; iZ++)
					{
						Asura_Hash_ID uHash = GetHash( iX, iY, iZ );

						u_int uHashEntry = uHash % uHASH_ENTRIES;

						for ( Asura_Physics_SpatialHash_Entry* pxNode = m_apxFirstNodes[uHashEntry]; pxNode!=0; pxNode = pxNode->m_pxNextEntry )
						{
							// if we come across an entry with the same hash id, we need to do an intersection test
							if ( pxNode->m_uHash == uHash )
							{
								// get the object
								Asura_Physics_RigidBodyInstance* pxInstance = pxNode->m_pxInstance;

								// check the obvious flags
								if ( pxInstance->IsIgnore() ) continue;

								if ( bConsiderOnlyExternalControl && !(pxInstance->IsExternallyControlled()) ) continue;

								if ( pxInstance->GetBroadphaseAABB().Intersects( xAABB ) )
								{
									pxQuery->AddEntry( pxInstance );
								}
							}
						}
					}
				}
			}
		}



		u_int GetMemUsed( void )
		{
			return m_pucFreeMem - m_pucBuffer;
		}

		float GetBestGuessSpatialHashScale( void )
		{
			return m_fBestGuessSpatialHashScale;
		}

		u_int GetNumberOfHashNodes( void )
		{
			return uHASH_ENTRIES;
		}
		
		u_int GetNumberOfEntries( u_int uHash )
		{
			u_int uNumber = 0;

			Asura_Physics_SpatialHash_Entry* pxNode = m_apxFirstNodes[uHash];
			
			
			while ( pxNode )
			{
				uNumber++;

				pxNode = pxNode->m_pxNextEntry;
			}

			return uNumber;
		}

	private:

		Asura_Hash_ID GetHash( int iX, int iY, int iZ )
		{
			return 7* iX + 503 * iY + 24847 * iZ;
		}

		Asura_Physics_SpatialHash_Entry*	m_apxFirstNodes[uHASH_ENTRIES];

		u_char*								m_pucFreeMem;
		u_char*								m_pucBuffer;

		float								m_fLargestObjectExtent;
		float								m_fBestGuessSpatialHashScale;
};

#endif//__ASURA_PHYSICS_SPATIALHASH_H__
