//
// Asura_Physics_Vehicle_Speedboat.h
//
// Created by Andrew Weinkove - 30/08/05
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_PHYSICS_VEHICLE_SPEEDBOAT_H__
#define __ASURA_PHYSICS_VEHICLE_SPEEDBOAT_H__

// Forward declarations.
class Asura_Physics_Object;
class Asura_Physics_Vehicle_Speedboat_Description;

// Includes
#include "Core/Asura_Maths.h"
#include "Core/Asura_MempoolFactory_Template.h"
#include "Physics/Vehicles/Asura_Physics_Vehicle.h"



//
// (AW - 31/08/05) - Code for speedboat control written for Miami Vice. The emphasis is 
// very much on providing speedboats that are fun to drive. This is only a loose 
// approximation of the real physics involved.
//
// This code has to fulfil 2 goals:
//
// 1 : It must be able to make NPC boats which are on fixed splines with fixed 
// velocities tilt and skim the water surface realistically. By taking just 
// positions and orientations it has to tweak Euler.x, Euler.z and Position.y in 
// a reasonably realistic-looking manner.
//
// 2 : Player controlled boats, the player will send control info consisting of
// throttle and rudder. It will change the velocity and Euler.y. The work described
// in goal 1 is also done and feeds back into the controls.
//
// The class Asura_Physics_Vehicle_Speedboat_External does the work specific to goal one.
// The class Asura_Physics_Vehicle_Speedboat_Controller does the work specific to goal two.
// The class Asura_Physics_Vehicle_Speedboat does common functionality.
//
// Some basic concepts to be included in the model:
//
// Boats lean back while accelerating.
// Boats lift out of the water as their velocity is higher.
// Boats lean into corners.
// Boats skim along the surface at high speeds, some skim smoothly (hydroplaning) while 
//   others bounce along. Other boats might be too heavy to lift very high.
// Bouyancy force is greater the deeper into the water a boat is.
// Bouyancy will correct boat tilt with damped harmonic motion.
// Acceleration (including turning) is affected when the boat is outside of the water.
// The centre of rotation when turning (Euler.y) is moved back when the boat is tilting upwards.
// 
//

/////////////////////////////////////
// Asura_Physics_Vehicle_Speedboat //
/////////////////////////////////////
class Asura_Physics_Vehicle_Speedboat : public Asura_Physics_Vehicle
{
public:
	enum SPEEDBOAT_COLLISION
	{
		SPEEDBOAT_COLLISION_NONE,
		SPEEDBOAT_COLLISION_SMALL,
		SPEEDBOAT_COLLISION_LARGE,
	};
	Asura_Physics_Vehicle_Speedboat(const Asura_Physics_Vehicle_Speedboat_Description& xSpeedboatDescription, Asura_Physics_Object* pxPhysicsObject)
	{
		m_fCurrentEulerXOffset = 0.0f;
		m_fCurrentEulerZOffset = 0.0f;
		m_fSpeedboatCollisionTimer = 0.0f;
		m_pxSpeedboatDescription = &xSpeedboatDescription;
		m_pxPhysicsObject = pxPhysicsObject;
		m_eSpeedboatCollision = SPEEDBOAT_COLLISION_NONE;
	}

	SPEEDBOAT_COLLISION GetSpeedboatCollision() 
	{
		SPEEDBOAT_COLLISION eRet = m_eSpeedboatCollision;
		m_eSpeedboatCollision = SPEEDBOAT_COLLISION_NONE;
		return eRet;
	}
	virtual void AdjustPositionAndOrientation(Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientation);
	virtual float GetThrottle() const = 0;
	static void SetWaterLevel(float fWaterLevel) {s_fWaterLevel = fWaterLevel;}
protected:
	float GetDepth(float fYPosition, float fEulerX) const;
	void CalculateBouyancyForce(float& fBouyancyForce, float fYPosition, float fEulerX) const;
	void CalculateTiltForce(float& fEulerXForce, float& fEulerZForce, float fYPosition, const Asura_Euler_3& xOriEuler, const Asura_Vector_3& xVelocity, const Asura_Vector_3& xAcceleration, bool bInvertTurnTilt) const;

	float m_fCurrentEulerXOffset;
	float m_fCurrentEulerZOffset;
	Asura_Physics_Object* m_pxPhysicsObject;
	const Asura_Physics_Vehicle_Speedboat_Description* m_pxSpeedboatDescription;
	SPEEDBOAT_COLLISION m_eSpeedboatCollision;
	float m_fSpeedboatCollisionTimer;
	static float s_fWaterLevel;
};

//////////////////////////////////////////////
// Asura_Physics_Vehicle_Speedboat_External //
//////////////////////////////////////////////
class Asura_Physics_Vehicle_Speedboat_External : public Asura_Physics_Vehicle_Speedboat
{
public:
	// Create an externally contolled speedboat with the given description and physics object.
	static Asura_Physics_Vehicle_Speedboat_External* Create( const Asura_Physics_Vehicle_Speedboat_Description& xSpeedboatDesc, Asura_Physics_Object* pxPhysicsObject);
	// On calling this function, you can expect the y component of position and the euler.x and
	// euler.z components of orientation to be adjusted.
	virtual void AdjustPositionAndOrientation(Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientation);
	virtual float GetThrottle() const;
	virtual void Update();
private:
	// Protected constructor to force use of create function.
	Asura_Physics_Vehicle_Speedboat_External(const Asura_Physics_Vehicle_Speedboat_Description& xSpeedboatDescription, Asura_Physics_Object* pxPhysicsObject)
	: Asura_Physics_Vehicle_Speedboat(xSpeedboatDescription, pxPhysicsObject)
	{
		m_fCurrentYOffset = 0.0f;
		m_uNumberOfPosAndOriRecords = 0;
		m_fYVelocity = 0.0f;
		m_fEulerXVelocity = 0.0f;
		m_fEulerZVelocity = 0.0f;
		m_xSmoothedAcceleration.Zero();
	}

	class PosAndOriRecord
	{
	public:
		Asura_Vector_3 m_xPosition;
		Asura_Euler_3 m_xOrientation;
		float m_fTimeStamp;
	};
	PosAndOriRecord m_axPosAndOriRecord[3];
	u_int m_uNumberOfPosAndOriRecords;
	float m_fYVelocity;
	float m_fEulerXVelocity;
	float m_fEulerZVelocity;
	float m_fCurrentYOffset;
	Asura_Vector_3 m_xSmoothedAcceleration;
};


////////////////////////////////////////////////
// Asura_Physics_Vehicle_Speedboat_Controller //
////////////////////////////////////////////////
class Asura_Physics_Vehicle_Speedboat_Controller : public Asura_Physics_Vehicle_Speedboat
{
public:
	static Asura_Physics_Vehicle_Speedboat_Controller* Create( const Asura_Physics_Vehicle_Speedboat_Description& xSpeedboatDesc, Asura_Physics_Object* pxPhysicsObject);
	void HandleControls(float fThrottle, float fRudder);
	void Update();
	virtual float GetThrottle() const;
private:
	// Protected constructor to force use of create function.
	Asura_Physics_Vehicle_Speedboat_Controller(const Asura_Physics_Vehicle_Speedboat_Description& xSpeedboatDescription, Asura_Physics_Object* pxPhysicsObject)
	: Asura_Physics_Vehicle_Speedboat(xSpeedboatDescription, pxPhysicsObject)
	{
		m_xAngularVelocity.Zero();
		m_xVelocity.Zero();
		m_fThrottle = 0.0f;
		m_fRudder = 0.0f;
	}

	void ApplyFriction(float fDepth, const Asura_Vector_3& xFacing);
	float m_fThrottle;
	float m_fRudder;
	Asura_Euler_3 m_xAngularVelocity;
	Asura_Vector_3 m_xVelocity;
};

#endif //#ifndef __ASURA_PHYSICS_VEHICLE_SPEEDBOAT_H__
