
//
// Asura_Projector_System.cpp
//
// Created by Dave Reed - 05/01/2007 
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Projector system - handles texture projectors. Quite similar to the photon system
//

#include "Core/Asura.h"

#ifdef ASURA_TARGET_SM3

#include "Core/Asura_Mempool.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_Maths.h"
#include "Projectors/Asura_Projector.h"
#include "Projectors/Asura_Projector_System.h"
#include "Camera/Asura_Camera.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"

Asura_TrackingHandlePool<Asura_Projector>*	Asura_Projector_System::s_pxProjectorList;
#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
Asura_Bounding_Box							Asura_Projector_System::s_xBBOfAllVisibleModules;
#endif
bool										Asura_Projector_System::s_bWithDarkeningShadows = false;
bool										Asura_Projector_System::s_bWithRealShadows		= false;
bool										Asura_Projector_System::s_bWithNoShadows		= false;

u_int										Asura_Projector_System::s_uNumberOfLocalProjectors;
Asura_Projector*							Asura_Projector_System::s_apxLocalProjectors[ASURA_MAX_NO_OF_LOCAL_PROJECTORS];


void Asura_Projector_System::Initialise(void)
{	
	s_pxProjectorList = new Asura_TrackingHandlePool<Asura_Projector>(16);	

	Platform_Initialise();
}


void Asura_Projector_System::Shutdown(void)
{
	delete s_pxProjectorList;
	s_pxProjectorList = NULL;

	Platform_Shutdown();
}


void Asura_Projector_System::Reset(void)
{
	Asura_Assert(s_pxProjectorList, "Reset called when not Initialised - crash imminent.");
	s_pxProjectorList->Reset();

	Platform_Reset();
}


void Asura_Projector_System::Update(void)
{

}


void Asura_Projector_System::RenderCallback(Asura_Projector* pxProjector)
{

#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	// Test against BB of all modules
	const Asura_Bounding_Box &xBB = pxProjector->GetBoundingBox();
	
	if (!xBB.Intersects(s_xBBOfAllVisibleModules))
	{
		return;
	}
#endif

	// Check this light has one of the types of shadows that we're rendering this pass
	bool bAccept = false;
	uint32 uFlags =  pxProjector->GetShadowParameters().m_uFlags;
	
	if (s_bWithRealShadows)
	{
		if ((uFlags & (ASURA_SHADOWFLAG_CASTS_OWN_SHADOWS|ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS)) && ((uFlags & ASURA_SHADOWFLAG_IS_DARKENING_PASS) == 0))
		{
			if (uFlags & ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS)
			{
				if ((uFlags & ASURA_SHADOWFLAG_CONTRIBUTES_TO_AVERAGE_SHADOW) == 0)
				{
					bAccept = true;
				}

			} else {
				bAccept = true;
			}
		}
	}

	if (s_bWithDarkeningShadows)
	{
		if (((uFlags & (ASURA_SHADOWFLAG_CASTS_OWN_SHADOWS|ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS)) && (uFlags & ASURA_SHADOWFLAG_IS_DARKENING_PASS)) || (uFlags & ASURA_SHADOWFLAG_CONTRIBUTES_TO_AVERAGE_SHADOW))
		{
			bAccept = true;
		}			
	}

	if (s_bWithNoShadows)
	{
		if (0 == (uFlags & ((ASURA_SHADOWFLAG_CASTS_OWN_SHADOWS|ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS) | ASURA_SHADOWFLAG_CONTRIBUTES_TO_AVERAGE_SHADOW)))
		{
			bAccept = true;
		}			
	}

	if (!bAccept)
	{		
		return;
	}

#if defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	if (!pxProjector->IsVisible()) return;
#endif

#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	// Test against frustum
	if (!Asura_Camera::IsAABBInFrustum(xBB))	
	{
		return;
	}

	// Test against visible modules	
	if (!Asura_Module_System::DoesBBIntersectVisibleModule(xBB, &pxProjector->m_iLastKnownModule))
	{
		 return;
	}
#endif


	// Handle UV scrolling
	if (pxProjector->m_iDetailTextureIndex!=ASURA_TEXTURE_NONE)
	{
		float fTimer = Asura_Timers::GetLocalScaledTime();
		float fStep = Asura_Timers::GetFrameTime()*Asura_Timers::GetGameTimeScale();
		if (pxProjector->m_fUScrollSpeed>0.0f)
		{
			if (pxProjector->m_fUScrollRange>0.0f)
			{
				pxProjector->m_fUScrollOfs = Asura_Maths::Sine(fTimer * pxProjector->m_fUScrollSpeed) * pxProjector->m_fUScrollRange;
			}
			else
			{
				pxProjector->m_fUScrollOfs += fStep * pxProjector->m_fUScrollSpeed;
				
				if (pxProjector->m_fUScrollOfs>1.0f)
				{
					pxProjector->m_fUScrollOfs-=1.0f;
				}
			}
		}
		else
		{
			pxProjector->m_fUScrollOfs = 0.0f;
		}

		if (pxProjector->m_fVScrollSpeed>0.0f)
		{
			if (pxProjector->m_fVScrollRange>0.0f)
			{
				pxProjector->m_fVScrollOfs = Asura_Maths::Sine(fTimer * pxProjector->m_fVScrollSpeed) * pxProjector->m_fVScrollRange;
			}
			else
			{
				pxProjector->m_fVScrollOfs += fStep * pxProjector->m_fVScrollSpeed;
				
				if (pxProjector->m_fVScrollOfs>1.0f)
				{
					pxProjector->m_fVScrollOfs-=1.0f;
				}		
			}
		}
		else
		{
			pxProjector->m_fVScrollOfs = 0.0f;
		}
	}

	// Ok - we've actually got to render it
	Platform_RenderProjector(pxProjector);	
}


void Asura_Projector_System::Render(bool bWithDarkeningShadows, bool bWithRealShadows, bool bWithNoShadows)
{
	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_PROJECTORS))
	{
		return;
	}

#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	// Get a bounding box around all visible modules
	bool bValid = false;

	s_xBBOfAllVisibleModules.MinX=s_xBBOfAllVisibleModules.MinY=s_xBBOfAllVisibleModules.MinZ = 1e12f;
	s_xBBOfAllVisibleModules.MaxX=s_xBBOfAllVisibleModules.MaxY=s_xBBOfAllVisibleModules.MaxZ = -1e12f;

	const u_int uModuleCount = Asura_Module_System::GetNumberOfModules();
	for( u_int uModuleIndex = 0; uModuleIndex < uModuleCount; uModuleIndex += 1)
	{
		if( false == Asura_Module_System::IsModuleVisible( uModuleIndex))
		{
			continue;
		}

		Asura_Bounding_Box xModuleBoundingBox = Asura_Module_System::GetBoundingBoxOfModule( uModuleIndex);
		xModuleBoundingBox.Displace( Asura_Module_System::GetPositionOfModule(uModuleIndex) );
		s_xBBOfAllVisibleModules.Set( xModuleBoundingBox);		

		bValid = true;
	}	

	if (!bValid)
	{
		// No visible modules
		s_xBBOfAllVisibleModules.MinX=s_xBBOfAllVisibleModules.MinY=s_xBBOfAllVisibleModules.MinZ = 0;
		s_xBBOfAllVisibleModules.MaxX=s_xBBOfAllVisibleModules.MaxY=s_xBBOfAllVisibleModules.MaxZ = 0;
	}
#endif// !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )

	s_bWithDarkeningShadows = bWithDarkeningShadows;
	s_bWithRealShadows		= bWithRealShadows;
	s_bWithNoShadows		= bWithNoShadows;

	Platform_BeginRender();
	{
		s_pxProjectorList->Traverse(RenderCallback);
	}
	Platform_EndRender();
}





Asura_Projector* Asura_Projector_System::CreateProjector(void)
{
	Asura_Handle iHandle = s_pxProjectorList->Allocate();

	if (iHandle == ASURA_INVALID_HANDLE)
	{
		return NULL;
	}

	Asura_Projector* pxProjector = s_pxProjectorList->GetPointer(iHandle);


	pxProjector->ResetToDefaults();

	pxProjector->m_iHandle = iHandle;

	return pxProjector;
}


void Asura_Projector_System::DestroyProjector(const Asura_Projector* pxProjector)
{
	// The delete operator, etc., tolerates NULL ptrs so it kind of makes
	// sense that we should here as well.
	if( !pxProjector ) { return; }

#ifdef ASURA_ASSERT
	Asura_Handle iHandle = pxProjector->m_iHandle;
	Asura_Assert( iHandle != ASURA_INVALID_HANDLE, "invalid handle" );
	Asura_Projector* pxProjectorTest = s_pxProjectorList->GetPointer(iHandle);
	Asura_Assert(pxProjectorTest == pxProjector, "pointer mismatch" );
#endif

	s_pxProjectorList->Deallocate( pxProjector->m_iHandle );
}


Asura_Projector* Asura_Projector_System::GetProjectorPointer(Asura_Handle iHandle)
{
	Asura_Assert( iHandle!=ASURA_INVALID_HANDLE, "Invalid handle!" );

	return s_pxProjectorList->GetPointer(iHandle);
}







void Asura_Projector_System::MakeLocalProjectorListCallback(Asura_Projector* pxProjector, void *pxDataIn)
{
	Asura_MakeLocalProjectorData *pxData = static_cast<Asura_MakeLocalProjectorData *>(pxDataIn);

	if (s_uNumberOfLocalProjectors<ASURA_MAX_NO_OF_LOCAL_PROJECTORS)
	{
		if (pxData->pxBB->Intersects(pxProjector->GetBoundingBox()))
		{		
			if (!pxData->bLightsOnly || (pxProjector->m_bDynamicLight || pxProjector->m_bDynamicSpotlight))
			{			
				s_apxLocalProjectors[s_uNumberOfLocalProjectors] = pxProjector;
				s_uNumberOfLocalProjectors++;
			}
		}
	}
	else
	{
		Asura_Assert(false,"Too many local projectors - increase ASURA_MAX_NO_OF_LOCAL_PROJECTORS");
	}
}


int Asura_Projector_System::MakeLocalProjectorList(Asura_Bounding_Box xBB)
{
	Asura_MakeLocalProjectorData xData;

	xData.bLightsOnly = true;
	xData.pxBB = &xBB;

	s_uNumberOfLocalProjectors = 0;
	s_pxProjectorList->Traverse(MakeLocalProjectorListCallback, &xData);

	return s_uNumberOfLocalProjectors;
}

Asura_Projector_Iterator Asura_Projector_System::GetIterator()
{
	return s_pxProjectorList->GetIterator();
}

#endif //#ifdef ASURA_TARGET_SM3



