//
// Asura_Fog.h
//
// Created by KJFL - 2004-03-12 12:50:38
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_FOG_H__
#define __ASURA_FOG_H__

class Asura_RGBA;
class Asura_Vector_2;
#include "Core/Asura_Graph.h"
#include "Rendering/Asura_Render.h"

//!
//! KJFL - 2004-03-12 11:08:20 - Generic Fog access/set methods
//!
//! This is a generic interface for 'simple' fog that has a single colour value and a 'near' and 'far' plane
//! defined by distances from the camera position in metres.
//!
//! Any vertices in camera space which are at the far plane or beyond will be completely fogged (i.e. will be 
//! rendered as the fog colour), and any vertices at the near plane or nearer will be left unfogged. Anything in
//! between will linearly interpolate between the two extremes using the formula 
//!
//! FogValue = (FogFar - VertexInCameraSpace.z) / (FogFar - FogNear) 
//!
//! [ with the result notionally clamped so that 0.0f < FogValue < 1.0f ]
//!
//! This style of fog is designed for the capabilities of low-end PCs to take advantage of the fixed function
//! pipeline of cards which have TnL but not vertex shader support. Note that a fog value of 1.0f means not fogged
//! in the above definition, and zero means completely fogged.
//!
//! Platforms don't have to implement the interpolation with exactly this formula (for example, you could swap the 
//! first two terms so 0.0f means not fogged if that makes things easier) - but this is what the PC TnL pipeline does,
//! so on the PC (at the time of writing) the environment renderer uses TnL and the character and object vertex shaders
//! are written to utilise the same formula.
//!
//! A simple improvement would be to have a 'maximum fogging' cut off value, e.g. rather than fogging out completely, 
//! max out at a set value. This would give artists much better control.  Whilst min specs of our PC games still
//! have cards like GeForce4MXs (very popular at the time of writing) we can only achieve consistent looking fog by
//! using the simple TnL style. A project only on the Xbox & PS2, or on top-end PCs, for example, might wish to have
//! this improvement.
//! 
//! This interface is naturally expected to expand as more fogging features are added.
//!
//! (MP - 31.03.2004) I've added a fog value at far plane, so that the output
//! fog value is then
//!
//!     output fog value = fog value at far plane * ( (fog far plane - viewspace z) / (fog far plane - fog near plane) )
//!
//! Projects that need to support lower end PCs can just fix this at 1, to
//! maintain a consistent look across all platforms.
//!
//! (AW - 20.08.2004) - Graph fog added. This turns the density\depth relationship from being 
//! linear to it being a spline with 3 control points (5 if you count the ends). At the moment this
//! is implemented only for PS2, but the graph has deliberately been kept simple enough to allow the 
//! possibility of it working for PC\XBox.

#include "Rendering\Asura_Colour.h"

//------------------------------------------------------------------------------------------------------------------------------
// EP.11.DEC.2008. This class will be moved to separate file called "Asura_Fog_Parameters.cpp/.h". But I will do it as a part of bigger update
// as I don't really want to update all projects again and again, so I will do it once with all files that needs to be updated.

class Asura_Fog_Parameters
{

public:
	
	u_int GetFogColourAsUint( void )			{	return m_uFogColourRGB;	}			// get fog colour as a 24bit u_int
	const Asura_RGBA& GetFogColour( void )		{	return m_xFogColour;	}			// get fog colour as floats, range 0.0f to 1.0f (except if using HDR.) Alpha is not used
	float GetFogNearPlane( void )				{	return m_fFogNearPlane;	}			// returns distance from camera in metres
	float GetFogFarPlane( void )				{	return m_fFogFarPlane;	}			// returns distance from camera in metres
	float GetFogValueAtFarPlane( void )			{	return m_fFogValueAtFarPlane;	}	// Returns fog value at far plane, in range [0, 1].
	float GetOneOverFogPlaneSeparation( void )	{	return m_fOneOverFogPlaneSeparation;}//returns 1.0f/(FogFar-FogNear)
	float GetSkyboxFogBlendFactor( void )		{	return m_fSkyboxFogBlendFactor;	}	// returns blend factor in range 0.0 to 1.0f
	const Asura_Vector_2& GetGraphFogPoint()	{	return m_xGraphFogExtraPoint;	}
	bool  IsGraphFogEnabled()					{	return m_bGraphFogEnabled;		}

	float GetFogDensityAtDistance(float fDistance);	// Look up a distance on the graph fog curve or linear fog line

	void SetFogColour( const Asura_RGBA& xColour );
	void SetFogColour( float fRed, float fGreen, float fBlue );
	void SetFogNearPlane( float fNearPlane );
	void SetFogFarPlane( float fFarPlane );
	void SetFogValueAtFarPlane( float fFogValueAtFarPlane );
	void SetSkyboxFogBlendFactor( float fBlendFactor );

	void CalculateLerpConstant( void );

	void EnableGraphFog( bool bEnable);								// Turns graph fog on and off
	void SetGraphFogPoint( const Asura_Vector_2 &xGraphPoint );		// Set the graph fog point.
	void DestroyGraphFogGraph();

	void BlendParameters( Asura_Fog_Parameters& xFogParametersToBlend, float fStrength );

	bool ReadFromChunkStream( Asura_Chunk_Stream& xStream );
	void WriteToChunkStream( Asura_Chunk_Stream& xStream ) const;


	// Constructor and destructor.
	Asura_Fog_Parameters();	
	~Asura_Fog_Parameters();

	Asura_Fog_Parameters& operator = (const Asura_Fog_Parameters& xFog);

private:

	Asura_RGBA		m_xFogColour;
	u_int			m_uFogColourRGB;
	float			m_fFogNearPlane;
	float			m_fFogFarPlane;
	float			m_fFogValueAtFarPlane;
	float			m_fOneOverFogPlaneSeparation;
	float			m_fSkyboxFogBlendFactor;

	// Extra bits for graph fog.
	bool				m_bGraphFogEnabled;
	Asura_Vector_2		m_xGraphFogExtraPoint;
	Asura_Graph<float>*	m_pxGraphFogGraph;

	static const u_int	s_uChunkVersion;	// The version used when reading/writing.
};

// EP.11.DEC.2008. End of class that will be moved to "Asura_Fog_Parameters.cpp/.h" file.
//------------------------------------------------------------------------------------------------------------------------------

class Asura_Fog
{
	public:

		static void	Initialise();
		static void	Reset();
		static void	Shutdown();

		static void SetFog( const Asura_Fog_Parameters& xFogParameters );
		static Asura_Fog_Parameters& GetFog()				{return s_xFogParameters; }

		static void SetInitialFog( const Asura_Fog_Parameters& xFogParameters );
		static Asura_Fog_Parameters& GetInitialFog()		{return s_xInitialFogParameters; }

		//
		// Query methods
		//
		
		static bool	IsFogEnabled()			{	return s_bFogIsEnabled;		}

		// JGP - In tools build this is the fog is enabled and not hidden. In retail
		// builds it's just is the fob enabled. So in retail builds it's the equiverlant
		// of IsFogEnabled and there is not much point using it
		static bool	IsFogRenderable( void )
		{
#ifdef ASURA_TOOLS
			return s_bFogIsEnabled && ( !s_bHideFog ) && ( GetFogValueAtFarPlane() > 0.0f );
#else
			return s_bFogIsEnabled && ( GetFogValueAtFarPlane() > 0.0f );
#endif
		}


		// Setting/Getting initial values
		static const Asura_RGBA& GetInitialFogColour()			{	return s_xInitialFogParameters.GetFogColour();			}
		static float GetInitialFogNearPlane()					{	return s_xInitialFogParameters.GetFogNearPlane();		}
		static float GetInitialFogFarPlane()					{	return s_xInitialFogParameters.GetFogFarPlane();		}
		static float GetInitialFogValueAtFarPlane()				{	return s_xInitialFogParameters.GetFogValueAtFarPlane();	}
		static float GetInitialSkyboxFogBlendFactor()			{	return s_xInitialFogParameters.GetSkyboxFogBlendFactor();}
		static const Asura_Vector_2& GetInitialGraphFogPoint()	{	return s_xInitialFogParameters.GetGraphFogPoint();		}

		static void	SetInitialFogColour( const Asura_RGBA& xColour );
		static void	SetInitialFogColour( float fRed, float fGreen, float fBlue);
		static void	SetInitialFogNearPlane(float fNearPlane);
		static void	SetInitialFogFarPlane( float fFarPlane );
		static void	SetInitialFogValueAtFarPlane( float fFogValueAtFarPlane );
		static void	SetInitialSkyboxFogBlendFactor( float fBlendFactor );
		static void SetInitialGraphFogPoint( const Asura_Vector_2 &xGraphPoint );

#ifdef ASURA_TOOLS
		static void	SetToDefaultsIfNecessary( void );
#endif
		// End of Setting/Getting initial values for tool purposes.

		//
		// Get Methods
		//
		static u_int GetFogColourAsUint()				{	return s_xFogParameters.GetFogColourAsUint();	}		// get fog colour as a 24bit u_int
		static const Asura_RGBA& GetFogColour()			{	return s_xFogParameters.GetFogColour();			}		// get fog colour as floats, range 0.0f to 1.0f (except if using HDR.) Alpha is not used
		static float GetFogNearPlane()					{	return s_xFogParameters.GetFogNearPlane();		}		// returns distance from camera in metres
		static float GetFogFarPlane() 					{	return s_xFogParameters.GetFogFarPlane();		}		// returns distance from camera in metres
		static float GetFogValueAtFarPlane()			{	return s_xFogParameters.GetFogValueAtFarPlane();}		// Returns fog value at far plane, in range [0, 1].
		static float GetOneOverFogPlaneSeparation()		{	return s_xFogParameters.GetOneOverFogPlaneSeparation();	}//returns 1.0f/(FogFar-FogNear)
		static float GetSkyboxFogBlendFactor()			{	return s_xFogParameters.GetSkyboxFogBlendFactor();	}	// returns blend factor in range 0.0 to 1.0f
		static bool  IsGraphFogEnabled()				{	return s_xFogParameters.IsGraphFogEnabled();	}
		static const Asura_Vector_2& GetGraphFogPoint()	{	return s_xFogParameters.GetGraphFogPoint();		}

		// Look up a distance on the graph fog curve or linear fog line
		static float GetFogDensityAtDistance( float fDistance );

		//
		// Set methods
		//

		//! Set fog colour; components should be in the range 0.0f to 1.0f (except if using HDR.)
		static void	SetFogColour( const Asura_RGBA& xColour )	{	SetFogColour( xColour.m_fR, xColour.m_fG, xColour.m_fB );	}
		static void	SetFogColour( float fRed, float fGreen, float fBlue );

		static void	SetFogNearPlane( float fNearPlane );				// Set fog near plane, in metres
		static void	SetFogFarPlane( float fFarPlane );					// Set fog far plane, in metres
        static void	SetFogValueAtFarPlane( float fFogValueAtFarPlane );	// Set fog value at far plane. Must be in range [0, 1].


		// Set skybox fogging level. This blends the skybox to the fogging colour (0.0f=unfogged original skybox , 1.0f=fogged completely)
		// This allows artists to tint the sky to better match the fog settings; it will be shown in the tools as a percentage.
		static void	SetSkyboxFogBlendFactor( float fBlendFactor );

		
		static void	EnableFog( bool b );								// Turns fog on and off
		static void	EnableGraphFog( bool bEnable);						// Turns graph fog on and off
		static void	SetGraphFogPoint( const Asura_Vector_2 &xGraphPoint );

		static void BlendParameters( Asura_Fog_Parameters& xFogParametersToBlend, float fStrength );



 		// JGP - Set fog hidden
		static void ToggleHideFog() { s_bHideFog = !s_bHideFog; }
		static bool IsFogHidden() { return s_bHideFog; }
		static void SetHideFog( const bool bHideFog ) { s_bHideFog = bHideFog; }

		// (AAM - 23.10.2006) Render function for next-gen platforms, where fog is a discrete full-screen pass
		static void Render()
		{
			if (Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_FOG))
			{			
				if( IsFogRenderable() )
				{
					Platform_Render();
				}
			}
		}

   private:

        // C L A S S   M E T H O D S /////////////////////////////////////////

//		static void CalculateLerpConstant( void );	// Not used here anymore, so therefore I think that it is obsolete. If not, just uncomment it. //EP.11.DEC.2008

        static void Platform_UpdateForNewParameters( void );

		static void Platform_Render( void );

        // C L A S S   S T A T E /////////////////////////////////////////////

		static Asura_Fog_Parameters	s_xInitialFogParameters;	// We need to keep original settings for fog, because every frame the fog is changed by various modifiers. Namely by EffectVolumes.
		static Asura_Fog_Parameters	s_xFogParameters;			// Current setting. It is ok to change it every frame, because we can keep the copy in StoredForParameters.
		static bool					s_bFogIsEnabled;
		static bool					s_bHideFog;					// JGP - Hide fog in editor

};


#endif//__ASURA_FOG_H__
