//
// Asura_Module.cpp
//
// Created by Kevin Lea - 12/06/2001 11:04:59
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Rendering/Asura_Module.h"

#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_AABB_Tree.h"
#include "Core/Asura_LBAABB_Tree.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#if defined( ASURA_USE_TEXTURE_STREAMING )
#include "Core/Asura_Texture_Streaming.h"
#endif
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Env_System.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Mesh/Asura_Mesh.h"
#ifdef ASURA_TOOLS
#include "Mesh/Asura_Mesh_System.h"
#endif
#include "Network/Asura_Network.h"
#if defined(ASURA_MAIN) 
#include "Physics/Asura_Physics_Environment.h"
#endif
#include "Rendering/Asura_Render.h"
#ifdef ASURA_USE_STREAMING
#include "Streaming/Asura_Streaming_Zone.h"
#include "Streaming/Asura_Streaming_System.h"
#include "Streaming/Asura_Streaming_Connection.h"
#endif

#if defined(ASURA_TOOLS) && defined(ASURA_MAIN)
#include "Misc/Tools_Adjacency.h"
#endif

#if defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
#include "Visibility/Asura_Visibility_System.h"
#if defined( ASURA_TOOLS )
#include "Visibility/Tools_Visibility_System.h"
#endif
#endif

// G L O B A L   V A R I A B L E S ///////////////////////////////////////////
char ModuleFilename[128];

bool ForceRedrawAll;


//////////////////////////////////////////////////////////////////////////////
//
// Asura_Module_Handler
//
//////////////////////////////////////////////////////////////////////////////

#if defined(ASURA_TOOLS)
u_int	Asura_Module_Handler::s_uNumberOfLinkedInPolys		= 0;
u_int	Asura_Module_Handler::s_uNumberOfLinkedInTriangles	= 0;
u_int	Asura_Module_Handler::s_uNumLinkedInRenderMeshTris	= 0;
bool	Asura_Module_Handler::s_bCollisionMeshesInvalid		= true;	//flag to indicate that one or more collision mesh needs recalculation.
#endif	// defined(ASURA_TOOLS)

Asura_Module_Handler*	Asura_Module_Handler::s_pxCallbackHandler	= NULL;

// (MH - 19.06.2002) stuff to remember LOS state between callbacks from AABBTree system        
// cur module data
int													Asura_Module_Handler::s_iLastModuleIndex			= -1;
Asura_Module_Handler::ModuleTestCallbackFunction*	Asura_Module_Handler::s_pModuleTestCallbackFunction	= NULL;
Asura_Bounding_Box									Asura_Module_Handler::s_xModuleTestBB;


// Module culling test data.
unsigned short	Asura_Module_Handler::s_ausModuleList[NUM_MODULES_TO_STORE_FOR_CLIPPING];
unsigned short	Asura_Module_Handler::s_usNumModulesInList				= 0;

unsigned short	Asura_Module_Handler::s_ausVariableLengthModuleList[ MAX_VARIABLE_LENGTH_CLIPPING_COUNT ];
unsigned short	Asura_Module_Handler::s_usNumModulesInVariableList		= 0;
unsigned short	Asura_Module_Handler::s_usMaxNumModulesInVariableList	= 0;

int				Asura_Module_Handler::s_iVisibleModule					= -1;

int				Asura_Module_Handler::s_iLinkedInModule					= -1;
bool			Asura_Module_Handler::s_bInNonExteriorModule			= false;

Asura_Module_Handler::RenderFunction* Asura_Module_Handler::s_pfnRenderFunction = NULL;


Asura_Module_Handler::Asura_Module_Handler (void) :
	m_pxModuleStorage(NULL)
	,m_pxLBAABBModuleTree(NULL)
	,m_puModuleTreeLBAABBConversionArray(NULL)
	,m_iNumberOfModules(0)
	,m_pucRenderList(NULL)
	,m_pucLinkedInList(NULL)
#ifdef ASURA_TOOLS
	,m_pucPrevRenderList(NULL)
	,m_uRenderListChangedFrameStamp(0)
#endif
	,m_bModuleStorageAllocated(false)
	,m_bModuleLinkingLoaded(false)
	,m_bInitialised(false)
	,m_bUseModLinks(true)
	,m_bUseFrustumCulling(true)
	,m_bUseCameraPositionOverride(false)
	,m_xCameraPositionOverride(Asura_Maths::ZeroVector3)
#ifdef ASURA_USE_STREAMING
	,m_uNumConnectedHandlers(0)
	,m_ppxConnectedHandlers(NULL)
	,m_uZoneHash(ASURA_HASH_ID_UNSET)
	,m_uHandlerHash(ASURA_HASH_ID_UNSET)
#endif
	,m_pxEnv(0)
{
	// Nothing more to initialise...
}

Asura_Module_Handler::~Asura_Module_Handler ()
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		Asura_ResourceDeallocateArray(m_pucRenderList);
		Asura_ResourceDeallocateArray(m_pucLinkedInList);
#ifdef ASURA_TOOLS
		Asura_ResourceDeallocateArray(m_pucPrevRenderList);
#endif

		if (m_pxLBAABBModuleTree)
		{
			m_pxLBAABBModuleTree->DeleteNodes();
			delete m_pxLBAABBModuleTree;
		}

		Asura_ResourceDeallocateArray(m_puModuleTreeLBAABBConversionArray);

		Platform_Shutdown();

		if(m_bModuleStorageAllocated)
		{
			Asura_ResourceDeallocateArray(m_pxModuleStorage);
		}
#ifdef ASURA_USE_STREAMING
		Asura_ResourceDeallocateArray(m_ppxConnectedHandlers);
#endif
	}
	Asura_ResourcePopType();

}

void Asura_Module_Handler::AllocateModuleStorage( Asura_Environment& xEnvironment )
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		if( m_bModuleStorageAllocated )
		{
			Asura_Assert( xEnvironment.m_iNumberOfModules == m_iNumberOfModules, "Module count changed." );
		}
		else
		{
			m_iNumberOfModules        = xEnvironment.m_iNumberOfModules;
			m_bModuleStorageAllocated = true;

			Asura_ResourceAllocateArray(m_pxModuleStorage, m_iNumberOfModules);
		}
	}
	Asura_ResourcePopType();
}

void Asura_Module_Handler::PreLoad (void)
{
	if(m_bInitialised) return;

	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		m_bInitialised = true;
		m_bModuleLinkingLoaded = false;
		m_pxEnv = Asura_Environment_System::GetEnvironment();

		AllocateModuleStorage( *m_pxEnv );

		Asura_ResourceAllocateArray(m_pucRenderList, m_iNumberOfModules);
		Asura_ResourceAllocateArray(m_pucLinkedInList, m_iNumberOfModules);
#ifdef ASURA_TOOLS
		Asura_ResourceAllocateArray(m_pucPrevRenderList, m_iNumberOfModules);
#endif

		// (MP - 11.10.2002) Make debugging easier by initialising m_pucRenderList to
		// all 0s. Otherwise all but the least significant 4 bits remain
		// uninitialised.
		for( int i = 0; i < m_iNumberOfModules; ++i )
		{
			m_pucRenderList[ i ]   = 0;
			m_pucLinkedInList[ i ] = 0;
#ifdef ASURA_TOOLS
			m_pucPrevRenderList[ i ] = 0;
#endif
		}

		// (MP - 26.02.2003) NB: Mesh ptr, position and possibly collision mesh
		// ptr (depending on age of data) are all now setup in the Process()
		// method of Asura_Chunk_Environment_ModuleList.

		for( int iModule = 0; iModule < m_iNumberOfModules; ++iModule )
		{
			Asura_Module& xModule = m_pxModuleStorage[iModule];

			Asura_ResourceDeallocateArray(xModule.VisibleModules);
			xModule.NumberOfVisibleModules=0;
			xModule.VisibleModulesArraySize=0;

		}

		//  RJF:  (05/10/2004 16:02:56):  Create trees if they have not been loaded from the chunk
		InitMainModuleTree();


		InitAllModuleTrees();

		ForceRedrawAll = false; // KJL - 15/04/2003 12:16:23 - this is an old hack I've set to false - we should remove it entirely soon
		s_usNumModulesInList = 0;

#ifdef ASURA_USE_STREAMING
		//Allocate array of pointers to connected zones
		const Asura_Streaming_Zone* pxZone = Asura_Streaming_System::FindZone(m_uZoneHash);
		Asura_Assert(pxZone,"Failed to find streaming zone the corresponds to this module handler");
		m_uNumConnectedHandlers = pxZone->GetNumConnection();
		if(m_uNumConnectedHandlers)
		{
			Asura_ResourceAllocateArray(m_ppxConnectedHandlers,m_uNumConnectedHandlers);
			for(u_int i=0;i<m_uNumConnectedHandlers;i++)
			{
				m_ppxConnectedHandlers[i] = 0;					
			}
		}
#endif //ASURA_USE_STREAMING

		Platform_Initialise();
	}
	Asura_ResourcePopType();

#ifdef ASURA_USE_STREAMING
	Asura_Module_System::FindConnectedHandlers();
#endif
}

void Asura_Module_Handler::Reset (void)
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
// ARC - 16 May 2006 15:49:28 - Clear this pointer (fixes nasty bug with PS2 debug info - didn't happen on the PC of course! :o( )
		m_pxEnv = NULL;

		Asura_ResourceDeallocateArray(m_pucRenderList);
		Asura_ResourceDeallocateArray(m_pucLinkedInList);
#ifdef ASURA_TOOLS
		Asura_ResourceDeallocateArray(m_pucPrevRenderList);
#endif


		if (m_pxLBAABBModuleTree)
		{
			m_pxLBAABBModuleTree->DeleteNodes();
			delete m_pxLBAABBModuleTree;
			m_pxLBAABBModuleTree	= NULL;
		}

		Asura_ResourceDeallocateArray(m_puModuleTreeLBAABBConversionArray);

		Platform_Shutdown();

		if(m_bModuleStorageAllocated)
		{
			Asura_ResourceDeallocateArray(m_pxModuleStorage);
		}

#ifdef ASURA_USE_STREAMING
		Asura_ResourceDeallocateArray(m_ppxConnectedHandlers);
		m_ppxConnectedHandlers = 0;
		m_uNumConnectedHandlers = 0;
#endif //ASURA_USE_STREAMING

		// Reset the variables...
		m_iNumberOfModules			= 0;
		m_bModuleStorageAllocated	= false;
		m_bModuleLinkingLoaded		= false;
		m_bInitialised			= false;
	}
	Asura_ResourcePopType();
}

Asura_LBAABB_Tree* Asura_Module_Handler::CreateLBAABBModuleTree (void)
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		delete(m_pxLBAABBModuleTree);
		m_pxLBAABBModuleTree	= new Asura_LBAABB_Tree;
	}
	Asura_ResourcePopType();
	return(m_pxLBAABBModuleTree);
}

u_int* Asura_Module_Handler::CreateModuleTreeLBAABBConversionArray (const u_int uNumNodes)
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		Asura_ResourceDeallocateArray(m_puModuleTreeLBAABBConversionArray);
		Asura_ResourceAllocateArray(m_puModuleTreeLBAABBConversionArray, uNumNodes);
	}
	Asura_ResourcePopType();
	return(m_puModuleTreeLBAABBConversionArray);
}




void Asura_Module_Handler::InitMainModuleTree()
{
	for (int i=0; i<m_iNumberOfModules; i++)
	{
		Asura_Module& xModule = m_pxModuleStorage[i];

#ifdef ASURA_TOOLS
		Asura_Mesh* pxMesh = xModule.Mesh;	
		if( pxMesh )
		{
			if( !pxMesh->BoundingBoxDataValid )
			{
				pxMesh->CalculateBoundingBox();
			}

			if( !xModule.m_pxCollisionMesh || xModule.m_bCollisionMeshInvalid)
			{
				xModule.UpdateCollisionMesh();
			}
		}
		else
		{
			// KJL - 14/04/2003 18:33:30 - this assert isn't actually valid anymore,
			// since a PC export may not have any meshes
			///Asura_Assertf( false, ("No mesh for module %i.", i) );
		}
#endif

		Asura_Assertf( xModule.m_pxCollisionMesh, ("No collision mesh for module %i.", i) );
		xModule.Centre = xModule.m_pxCollisionMesh->GetBoundingBox().CalculateCentre() + xModule.Position;
	}

#ifdef ASURA_TOOLS
	if (!m_pxLBAABBModuleTree)
	{
		m_pxLBAABBModuleTree = new Asura_LBAABB_Tree;
		Asura_Bounding_Box *pxModuleBoundingBoxes = new Asura_Bounding_Box[m_iNumberOfModules];
		Asura_Vector_3 *pxModuleCentres = new Asura_Vector_3[m_iNumberOfModules];
		for (int i=0;i<m_iNumberOfModules;i++)
		{
			pxModuleBoundingBoxes[i] = m_pxModuleStorage[i].m_pxCollisionMesh->GetBoundingBox();
			pxModuleBoundingBoxes[i].Displace(m_pxModuleStorage[i].Position);
			pxModuleCentres[i] = pxModuleBoundingBoxes[i].CalculateCentre();
		}
		m_puModuleTreeLBAABBConversionArray = Tools_LBAABB_Tree::CreateRunTimeTree(*m_pxLBAABBModuleTree, m_iNumberOfModules, pxModuleBoundingBoxes, pxModuleCentres);
		delete[] pxModuleBoundingBoxes;
		delete[] pxModuleCentres;
	}
#endif //#ifdef ASURA_TOOLS
}

// RJF - 04 October 2004 - This code broken out from InitAllModuleTrees so that individual modules can be re-treed as necessary
void Asura_Module_Handler::InitTreeForSingleModule(int moduleIndex)
{
	Asura_Module& xModule = m_pxModuleStorage[moduleIndex];

#ifdef ASURA_TOOLS
	Asura_Mesh* pxMesh = xModule.Mesh;	
	if( pxMesh )
	{
		if( !pxMesh->BoundingBoxDataValid )
		{
			pxMesh->CalculateBoundingBox();
		}

		if( !xModule.m_pxCollisionMesh  || xModule.m_bCollisionMeshInvalid)
		{
			xModule.UpdateCollisionMesh();
		}
	}
#endif

	Asura_Assertf( xModule.m_pxCollisionMesh, ("No collision mesh for module %i.", moduleIndex) );

	const Asura_CollisionMesh* pxCMesh = xModule.m_pxCollisionMesh;
	if (!pxCMesh) return;

	const unsigned int uNumPoly = pxCMesh->GetNumberOfPolys();

	if (uNumPoly == 0) return;


}

/// Kevin Lea - 22/05/2002 15:48:42 - eventually expect this function to be replaced by levels loading in
/// precalculated data
void Asura_Module_Handler::InitAllModuleTrees()
{
	for (int i=0; i<m_iNumberOfModules; i++)
	{
		InitTreeForSingleModule(i);	
	}
}

#ifdef ASURA_TOOLS
void Asura_Module_Handler::InvalidateAllCollisionData()
{
	for (int i=0; i<m_iNumberOfModules; ++i)
	{
		InvalidateCollisionMesh(i);
		if(m_pxModuleStorage[i].Mesh)m_pxModuleStorage[i].Mesh->BoundingBoxDataValid = false;
	}

	//If the collision meshes are invalidated ,then any adjacency information for them will be invalid as well
	//RAR 08/06/2006
	#ifdef ASURA_MAIN
	Tools_Adjacency::InvalidateAllCollisionModules();
	#endif
}
#endif

void Asura_Module_Handler::Load(void)
{
	if(m_bModuleLinkingLoaded) return;
	PreLoad();//PreLoad added here , so that the system will be properly initialised even if there isn't a module linking chunk. RAR 24/03/2006

	m_bModuleLinkingLoaded = true;

	LoadAsuraFile(ModuleFilename);
}


#ifdef _DEBUG
bool Asura_Module_Handler::HasNonSceneryModules() const
{
	if (m_pxEnv)
	{
		for (int i=0; i<GetNumberOfModules(); ++i)
		{
			if (!m_pxEnv->IsSceneryModule(i))
			{
				return true;
			}
		}
	}

	return false;
}
#endif //_DEBUG


#ifdef ASURA_MAIN

static bool g_bFirstRenderListModuleFound;
static bool g_bUpdateLinkedModules_InNonExteriorModule;

void Asura_Module_Handler::AddLinksFromModule( const u_int uModuleIdx )
{
	const u_int uCamBitFlag = Asura_Camera::GetCameraIndexBitFlag();

	if( m_pxEnv->IsSceneryModule(uModuleIdx) || m_pxEnv->IsNeverCameraModule(uModuleIdx))
	{
		//Don't bother adding module links from scenery modules. RAR 26/03/03
		//Or from modules that are never considered to contain the camera. RAR 26/02/2008
		return;
	}
	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(!m_pxEnv->IsModuleInCurrentLightSet(uModuleIdx)) return;
	#endif

	bool bNonExteriorModule = !m_pxEnv->IsExteriorModule(uModuleIdx);
	if(bNonExteriorModule != g_bUpdateLinkedModules_InNonExteriorModule)
	{
		if(bNonExteriorModule)
		{
			//Previous module links added have been for exterior modules , but we have now determined that we are in a
			//non-exterior module. So reset all the links again , and just add links for non-exterior modules from now on.
			//RAR 10/04/2007
			g_bUpdateLinkedModules_InNonExteriorModule = true;
			Asura_Module_System::ClearRenderAndLinkedInListsForCurCam();
		}
		else
		{
			//Camera position has already been found to be in a non-exterior module , so ignore links for exterior modules.
			//RAR 10/04/2007
			return;
		}
	}
	else if( g_bFirstRenderListModuleFound )
	{
		Asura_Module_System::ClearRenderAndLinkedInListsForCurCam();
	}
	g_bFirstRenderListModuleFound = false;

	//add this module
	if( !(m_pucRenderList[uModuleIdx] & uCamBitFlag) )
	{
		m_pucRenderList[uModuleIdx] |= uCamBitFlag;

#ifdef ASURA_TOOLS
		if( m_pxModuleStorage[uModuleIdx].Mesh )
		{
			s_uNumberOfLinkedInPolys     += m_pxModuleStorage[uModuleIdx].Mesh->NumberOfPolygons;
			s_uNumberOfLinkedInTriangles += m_pxModuleStorage[uModuleIdx].Mesh->NumberOfTriangles;
		}
		#ifndef ASURA_USE_STREAMING // TBD Fix count for streaming. RAR 07/04/2006
		s_uNumLinkedInRenderMeshTris += Asura_Mesh_System::Platform_GetRenderMeshTriCount( uModuleIdx );
		#endif
#endif
	}

	// (MH - 11.06.2002) add it to linked in list whatever
	m_pucLinkedInList[uModuleIdx] |= uCamBitFlag;

	m_pxEnv->AddReqdRgnsForThisFrame( uModuleIdx );

	const Asura_Module& xModule = m_pxModuleStorage[uModuleIdx];

	//and all modules visible from this module
	{
		const int  iNumVisModules = xModule.NumberOfVisibleModules;
		const int* piVisModules   = xModule.VisibleModules;

		for( int i = iNumVisModules; i; --i )
		{
			int iOtherModuleIdx = *piVisModules++;

			#ifdef ASURA_USE_STREAMING
			//module index contains a connection index in top 16 bits
			
			u_int uConnectionIndex = (u_int)(iOtherModuleIdx>>16);
			iOtherModuleIdx &= 0xffff;
			//iConnectionIndex is off by one because of 0 refering to this zone
			if(uConnectionIndex>m_uNumConnectedHandlers) continue;
			Asura_Module_Handler* pxHandler = uConnectionIndex ?  m_ppxConnectedHandlers[uConnectionIndex-1] : this;
			if(!pxHandler) continue;
			#else
			Asura_Module_Handler* pxHandler = this;
			#endif

			if(iOtherModuleIdx>=pxHandler->m_iNumberOfModules) continue;

			#ifdef ASURA_TOOLS
			//Ignore modules that aren't part of the current light set. RAR 15/02/2008
			if(pxHandler->m_pxEnv && !pxHandler->m_pxEnv->IsModuleInCurrentLightSet(iOtherModuleIdx)) continue;
			#endif

			if( !(pxHandler->m_pucRenderList[iOtherModuleIdx] & uCamBitFlag) )
			{
				pxHandler->m_pucRenderList[iOtherModuleIdx] |= uCamBitFlag;

				pxHandler->m_pxEnv->AddVisRgnsForThisFrame( iOtherModuleIdx );

#ifdef ASURA_TOOLS
				if (pxHandler->m_pxModuleStorage[iOtherModuleIdx].Mesh)
				{
					s_uNumberOfLinkedInPolys     += pxHandler->m_pxModuleStorage[iOtherModuleIdx].Mesh->NumberOfPolygons;
					s_uNumberOfLinkedInTriangles += pxHandler->m_pxModuleStorage[iOtherModuleIdx].Mesh->NumberOfTriangles;
				}
				#ifndef ASURA_USE_STREAMING // TBD Fix count for streaming. RAR 07/04/2006
				s_uNumLinkedInRenderMeshTris += Asura_Mesh_System::Platform_GetRenderMeshTriCount( iOtherModuleIdx );
				#endif
#endif
			}

			pxHandler->m_pucLinkedInList[iOtherModuleIdx] |= uCamBitFlag;
		}
	}
}

void Asura_Module_Handler::Update(const Asura_Vector_3* pxViewPosition,const int* piModuleIndex)
{
	const u_int uCurCamBitFlag = Asura_Camera::GetCameraIndexBitFlag();

#ifdef ASURA_TOOLS
	Asura_Assert(m_pucRenderList, "Asura_Module_Handler::Update called but m_pucRenderList not allocated!");
	if (NULL == m_pucRenderList)
		return;
#endif

	if ((0 == m_iNumberOfModules) || (m_pucRenderList == NULL) || (m_pucLinkedInList == NULL))	return;

	if(m_bUseModLinks)
	{
		if(!pxViewPosition)
		{
			//no position passed , so use the camera position
			pxViewPosition = Asura_Camera::GetPosition();
		}
	
		if(piModuleIndex)
		{
			Asura_Module_System::ClearRenderAndLinkedInListsForCurCam();

			Asura_Assert(*piModuleIndex>=0 && *piModuleIndex<m_iNumberOfModules,"");

			//add this module
			AddLinksFromModule(*piModuleIndex);

		}
		else
		{
			// No module specified, so use view position instead

			// (MH - 10.06.2002) note that since AABBTrees deal with bounding boxes,
			// not points, this BB is just set to a point. This isn't actually a bad
			// thing, it just means storing the same values twice.
			Asura_Bounding_Box viewBoundingBox;
			viewBoundingBox.Reset();
			viewBoundingBox.Set(*pxViewPosition);

			if( m_pxLBAABBModuleTree )
			{
				TestModuleTree(viewBoundingBox, BuildRenderListInternalCallback );
			}
		}
	}
	else
	{
		// Modlinking off, so render all modules
		for( int i = 0; i < m_iNumberOfModules; ++i )
		{	
			m_pucRenderList[i] |= uCurCamBitFlag;
			m_pucLinkedInList[i] |= uCurCamBitFlag;
		}
	}
}

#ifdef ASURA_TOOLS
void Asura_Module_Handler::CheckIfRenderListChanged( void )
{
	Asura_Assert(m_pucPrevRenderList, "m_pucPrevRenderList not allocated");
	if (Asura_Mem::Compare(m_pucPrevRenderList, m_pucRenderList, m_iNumberOfModules * sizeof(m_pucRenderList[0])))
	{
		m_uRenderListChangedFrameStamp = Asura_Timers::GetFrameNumber();
		Asura_Mem::Copy(m_pucPrevRenderList, m_pucRenderList, m_iNumberOfModules * sizeof(m_pucRenderList[0]));
	}
}
#endif

void Asura_Module_Handler::ClearRenderAndLinkedInListsForCurCam( void )
{
	if ((0 == m_iNumberOfModules) || (m_pucRenderList == NULL) || (m_pucLinkedInList == NULL))	return;

	const u_int uNotBitFlag = ~Asura_Camera::GetCameraIndexBitFlag();

	const int iNumModules = m_iNumberOfModules;
	for( int i = 0; i < iNumModules; ++i )
	{
		m_pucRenderList[i]   &= uNotBitFlag;
		m_pucLinkedInList[i] &= uNotBitFlag;
	}
}

void Asura_Module_Handler::AddAllModulesWithoutModLinkingToRenderListForCurCam()
{
	if ((0 == m_iNumberOfModules) || (m_pucRenderList == NULL) || (m_pucLinkedInList == NULL))	return;

	if(!m_bUseModLinks)
	{
		u_int uCurCamBitFlag = Asura_Camera::GetCameraIndexBitFlag(); 

		for(int i=0;i<m_iNumberOfModules;i++)
		{
			m_pucRenderList[i] |= uCurCamBitFlag;
		}
	}
}

void Asura_Module_Handler::AddAllModulesToRenderListForCurCam()
{
	if ((0 == m_iNumberOfModules) || (m_pucRenderList == NULL) || (m_pucLinkedInList == NULL))	return;

	u_int uCurCamBitFlag = Asura_Camera::GetCameraIndexBitFlag(); 

	for(int i=0;i<m_iNumberOfModules;i++)
	{
		m_pucRenderList[i] |= uCurCamBitFlag;
	}
}

void Asura_Module_Handler::AddAllModulesToLinkedInListForCurCam()
{
	if ((0 == m_iNumberOfModules) || (m_pucRenderList == NULL) || (m_pucLinkedInList == NULL))	return;

	u_int uCurCamBitFlag = Asura_Camera::GetCameraIndexBitFlag(); 

	for(int i=0;i<m_iNumberOfModules;i++)
	{
		m_pucLinkedInList[i] |= uCurCamBitFlag;
	}
}


void Asura_Module_Handler::CullRenderListForCurCam()
{
	if ((0 == m_iNumberOfModules) || (m_pucRenderList == NULL) || (m_pucLinkedInList == NULL))	return;

	u_char ucCurCamBitFlag = Asura_Camera::GetCameraIndexBitFlag(); 

	// if there's a camera position override for this module handler, set 
	// the camera position before doing the frustum culling
	Asura_Vector_3 xOldCameraPos = *Asura_Camera::GetPosition();
	if( IsUsingCameraPositionOverride() )
	{
		Asura_Camera::SetPosition( m_xCameraPositionOverride );
	}

	// Cull linked-in modules against frustum.
	for( int i = 0; i < m_iNumberOfModules; ++i )
	{	
        const u_char ucCurRenderListEntry = m_pucRenderList[i];
        if( ucCurRenderListEntry & ucCurCamBitFlag )
        {
            // This should only happen if the camera isn't in any module.
            if( !IsModuleInFrustum(i) )
            {
                m_pucRenderList[i] = ucCurRenderListEntry & ~ucCurCamBitFlag;
            }
        }
        else
        {
            if( m_pucLinkedInList[i] & ucCurCamBitFlag )
            {
                if( IsModuleInFrustum(i) )
                {
                    m_pucRenderList[i] = ucCurRenderListEntry | ucCurCamBitFlag;
                }
            }
        }
    }

	// restore previous camera position
	Asura_Camera::SetPosition( xOldCameraPos );
}

/// DWE - 05/06/2003 17:58:19 - Use this function to set all the modules
/// that are linked in, determined from server player positions
/// We use this so the server will know all the LinkedIn modules (for things
/// like AI updates)

// DWE - Most of this is stolen from the BuildRenderListCallback

void Asura_Module_Handler::BuildLinkedInListCallback(const u_int uIndex)
{
	if (!m_bInitialised) return;

	u_char ucBitFlag = ASURA_MODULE_LINKEDINFORPLAYER_BITFLAG;

	if(m_pxEnv->IsSceneryModule(uIndex))
	{
		//Don't bother adding module links from scenery modules. RAR 26/03/03
		return;
	}

	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(!m_pxEnv->IsModuleInCurrentLightSet(uIndex)) return;
	#endif


	// (MH - 11.06.2002) add it to linked in list whatever
	m_pucLinkedInList[uIndex] |= ucBitFlag;

	Asura_Module& xModule = m_pxModuleStorage[uIndex];

	//and all modules visible from this module
	for (int i=0; i<xModule.NumberOfVisibleModules; i++)
	{
		int iModuleIndex = xModule.VisibleModules[i];

		m_pucLinkedInList[iModuleIndex] |= ucBitFlag;
	}
}

#ifdef ASURA_TOOLS
void Asura_Module_Handler::ClearModuleLinks(int iModuleIndex)
{
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<m_iNumberOfModules,"Module number is out of bounds");

	m_pxModuleStorage[iModuleIndex].ClearModuleLinks();
}
#endif // ASURA_TOOLS

#if defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)
void Asura_Module_Handler::RemapModuleLinkConnectionIndices(u_int uArraySize,const int* piRemap)
{
	for(int i=0;i<m_iNumberOfModules;i++)
	{
		m_pxModuleStorage[i].RemapModuleLinkConnectionIndices(uArraySize,piRemap);
	}	
}

void Asura_Module_Handler::RemapModuleLinkModuleIndices(u_int uConnectionIndex,u_int uArraySize,const int* piRemap)
{
	for(int i=0;i<m_iNumberOfModules;i++)
	{
		m_pxModuleStorage[i].RemapModuleLinkModuleIndices(uConnectionIndex,uArraySize,piRemap);
	}	
}
#endif //defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)


#if defined(ASURA_USE_STREAMING) || defined(ASURA_TOOLS)
bool Asura_Module_Handler::IsModuleLinked(int iModuleIndex,int iLinkIndex, Asura_Hash_ID uLinkZoneHash)
{
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<m_iNumberOfModules,"Module number is out of bounds");

	//get the destination handler and connection index

	#if defined ASURA_USE_STREAMING || defined ASURA_ASSERT
		Asura_Module_Handler* pxLinkHandler = this;
	#endif

	u_int uConnectionIndex = 0;
	#ifdef ASURA_USE_STREAMING
	if(uLinkZoneHash && uLinkZoneHash != m_uZoneHash)
	{
		for(u_int i=0;i<m_uNumConnectedHandlers;i++)
		{
			if(!m_ppxConnectedHandlers[i]) continue;
			if(m_ppxConnectedHandlers[i]->m_uZoneHash == uLinkZoneHash)
			{
				pxLinkHandler = m_ppxConnectedHandlers[i];
				uConnectionIndex = i+1;
				break;
			}
		}
		if(!uConnectionIndex) return false;
	}
	#endif   //ASURA_USE_STREAMING
	Asura_Assert(iLinkIndex>=0 && iLinkIndex<pxLinkHandler->GetNumberOfModules(),"Module number is out of bounds");

	//module is always linked to itself
	if(iModuleIndex==iLinkIndex && uConnectionIndex == 0) return true;

	iLinkIndex |= (uConnectionIndex<<16);

	//search through the list of linked modules
	const Asura_Module* module = &m_pxModuleStorage[iModuleIndex];
	for(int i=0;i<module->NumberOfVisibleModules;i++)
	{
		if(iLinkIndex==module->VisibleModules[i]) return true;
	}
	return false;
}
#endif //defined(ASURA_USE_STREAMING) || defined(ASURA_TOOLS)

bool Asura_Module_Handler::IsModuleLinked(int iModuleIndex,int iLinkIndex)
{
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<m_iNumberOfModules,"Module number is out of bounds");
	Asura_Assert(iLinkIndex>=0 && iLinkIndex<m_iNumberOfModules,"Module number is out of bounds");

	//module is always linked to itself
	if(iModuleIndex==iLinkIndex) return true;

	//search through the list of linked modules
	const Asura_Module* module = &m_pxModuleStorage[iModuleIndex];
	for(int i=0;i<module->NumberOfVisibleModules;i++)
	{
		if(iLinkIndex==module->VisibleModules[i]) return true;
	}
	return false;
}


void Asura_Module_Handler::Render(void)
{
	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_ENV))
	{
		return;
	}

	ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__WORLD_RENDER, "World Render");

	if(s_pfnRenderFunction)
	{
		s_pfnRenderFunction(this);
	}
	else
	{
		Platform_BeginRender();

		Platform_Render();

		Platform_EndRender();
	}

	ASURA_CPU_PROFILE__STOP(ASURA_CPU_PROFILE__WORLD_RENDER);
}



#ifdef ASURA_TARGET_SM3
void Asura_Module_Handler::RenderLate(void)
{
	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_ENV))
	{
		return;
	}

	ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__WORLD_RENDER, "World Render");
	Platform_RenderLate();		
	ASURA_CPU_PROFILE__STOP(ASURA_CPU_PROFILE__WORLD_RENDER);
}
#endif


void Asura_Module_Handler::TestModuleTree (const Asura_Bounding_Box& xBB, Asura_Module_Handler::ModuleTestCallbackFunction* pfnOutputCallback)
{
	Asura_Module_Handler*	pxOldHandler	= s_pxCallbackHandler;

	// Store the current handler...
	s_pxCallbackHandler	= this;

#ifdef ASURA_TOOLS
	if(s_bCollisionMeshesInvalid) UpdateInvalidCollisionMeshes();
#endif
	// (AW - 18.08.2005) - Index we get back from the m_pxLBAABBModuleTree is going to
	// need converting. Go through another callback to achieve this.
	s_xModuleTestBB			= xBB;
	s_pModuleTestCallbackFunction	= pfnOutputCallback;

	if (m_pxLBAABBModuleTree)
	{
		m_pxLBAABBModuleTree->TestTreeWithBB(xBB, ModuleTestInternalCallback);
	}

	// Reset the handler...
	s_pxCallbackHandler	= pxOldHandler;
}




#ifdef ASURA_TOOLS
bool Asura_Module_Handler::IsPointVisible( const Asura_Vector_3& alpha, const Asura_Vector_3& beta, bool (pfnPolygonFilter)(const Asura_Mesh* pxMesh,int iPolyIndex) /*= 0*/,bool b2SidedPolygons /*= false*/)
{
	float lowest = ASURA_LOS_MAX_DISTANCE;

	Asura_Bounding_Box box;

	const Asura_Vector_3 xDelta = beta - alpha;

	float distance = xDelta.Magnitude();

	const Asura_Vector_3 xDeltaDir = xDelta * (1.0f / distance);

	lowest = distance;

	Asura_Vector_3 alpha2;
	for (int mn=0; mn<m_iNumberOfModules; mn++)
	{
		const Asura_Mesh * mesh = m_pxModuleStorage[mn].Mesh;
		Asura_Assert(mesh, "Error: Null mesh!");
		if (!mesh) continue;

#if 0
		{
			Asura_Vector_3 centre = m_pxModuleStorage[mn].Centre;
			float radius = m_pxModuleStorage[mn].Radius;

			//if (!m_pxModuleStorage[mn].Mesh->BoundingBox.IntersectsSphere(alpha-centre,lowest)) continue;


			if (!Asura_Maths::RayIntersectsSphere(alpha,beta,centre,radius)) continue;

			float mindist = (alpha-centre).Magnitude() - radius;

			if (lowest<mindist) continue;
		}
#endif
		{
		}

		alpha2 = alpha-m_pxModuleStorage[mn].Position;

		box.Reset();
		box.Set(alpha2);
		{
			// KJL - 19/03/2001 15:10:52 - for some reason the PS2 compiler
			// doesn't like box.Set(beta+alpha2);
			// (MP - 18.09.2002) That was because Set()'s parameter used to
			// be non-const.
			box.Set( xDelta + alpha2 );
		}
		if (!box.Intersects(m_pxModuleStorage[mn].Mesh->BoundingBox)) continue;

		for (int i=0; i<mesh->NumberOfPolygons; i++)
		{

			{
				bool facingToward;
				const Asura_Vector_3& v0 = mesh->VertexDataPtr[mesh->PolygonDataPtr[i].VertexID[0]].Position;
				const Asura_Vector_3& v1 = mesh->VertexDataPtr[mesh->PolygonDataPtr[i].VertexID[1]].Position;
				const Asura_Vector_3& v2 = mesh->VertexDataPtr[mesh->PolygonDataPtr[i].VertexID[2]].Position;

				if (v0.x>box.MaxX && v1.x>box.MaxX && v2.x>box.MaxX) continue;
				if (v0.y>box.MaxY && v1.y>box.MaxY && v2.y>box.MaxY) continue;
				if (v0.z>box.MaxZ && v1.z>box.MaxZ && v2.z>box.MaxZ) continue;

				if (v0.x<box.MinX && v1.x<box.MinX && v2.x<box.MinX) continue;
				if (v0.y<box.MinY && v1.y<box.MinY && v2.y<box.MinY) continue;
				if (v0.z<box.MinZ && v1.z<box.MinZ && v2.z<box.MinZ) continue;


				float d = Asura_Maths::RayIntersectsTri(alpha2,xDeltaDir,v0,v1,v2,facingToward);

				if(facingToward || b2SidedPolygons)
				{
					if (d>=0.0f && d<distance-0.01f)
					{
						if(pfnPolygonFilter == NULL || pfnPolygonFilter(mesh,i))
						{
							return false;
						}
					}
				}
			}

			if (mesh->PolygonDataPtr[i].IsQuad())
			{
				bool facingToward;
				const Asura_Vector_3& v0 = mesh->VertexDataPtr[mesh->PolygonDataPtr[i].VertexID[2]].Position;
				const Asura_Vector_3& v1 = mesh->VertexDataPtr[mesh->PolygonDataPtr[i].VertexID[3]].Position;
				const Asura_Vector_3& v2 = mesh->VertexDataPtr[mesh->PolygonDataPtr[i].VertexID[0]].Position;

				if (v0.x>box.MaxX && v1.x>box.MaxX && v2.x>box.MaxX) continue;
				if (v0.y>box.MaxY && v1.y>box.MaxY && v2.y>box.MaxY) continue;
				if (v0.z>box.MaxZ && v1.z>box.MaxZ && v2.z>box.MaxZ) continue;

				if (v0.x<box.MinX && v1.x<box.MinX && v2.x<box.MinX) continue;
				if (v0.y<box.MinY && v1.y<box.MinY && v2.y<box.MinY) continue;
				if (v0.z<box.MinZ && v1.z<box.MinZ && v2.z<box.MinZ) continue;


				float d = Asura_Maths::RayIntersectsTri(alpha2,xDeltaDir,v0,v1,v2,facingToward);

				if(facingToward || b2SidedPolygons)
				{
					if (d>=0.0f && d<distance-0.01f)
					{
						if(pfnPolygonFilter == NULL || pfnPolygonFilter(mesh,i))
						{
							return false;
						}
					}
				}

			}

		}

	}

	return true;
}	
#endif // #ifdef ASURA_TOOLS

void Asura_Module_Handler::CullingListModuleCallback( const u_int uModuleIdx )
{
	if (!m_bInitialised) return;

	if(m_pxEnv->IsSceneryModule(uModuleIdx))
	{
		return;//ignore scenery modules RAR 26/03/03
	}

	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(!m_pxEnv->IsModuleInCurrentLightSet(uModuleIdx)) return;
	#endif

	if(s_usNumModulesInList)
	{
		bool bExterior = m_pxEnv->IsExteriorModule(uModuleIdx);
		bool bListIsExterior = (s_ausModuleList[0] != USHRT_MAX) && m_pxEnv->IsExteriorModule(s_ausModuleList[0]);

		if(bExterior != bListIsExterior)
		{
			if(bExterior)
			{
				//Found an exterior module , but the list already contains some non-exterior modules.
				//So forget about this module.
				return;
			}
			else
			{
				//List currently contains exterior modules , and we have now found a non-exterior module.
				//So clear the existing list , and just add non-exterior modules from now on,.
				s_usNumModulesInList = 0;
			}
		}
	}

	if( NUM_MODULES_TO_STORE_FOR_CLIPPING == s_usNumModulesInList )
	{
		// Too many modules to store, make data invalid.
		s_ausModuleList[0] = USHRT_MAX;
	}
	else
	{
		s_ausModuleList[s_usNumModulesInList++] = static_cast<u_short>(uModuleIdx);
	}
}

void Asura_Module_Handler::CullingTestModuleCallback( const u_int uModuleIdx )
{
	if (!m_bInitialised) return;

	if(m_pxEnv->IsSceneryModule(uModuleIdx))
	{
		return;//ignore scenery modules RAR 26/03/03
	}

	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(!m_pxEnv->IsModuleInCurrentLightSet(uModuleIdx)) return;
	#endif

	
	bool bExterior = m_pxEnv->IsExteriorModule(uModuleIdx);
	if(!bExterior && !s_bInNonExteriorModule)
	{
		//first non-exterior module found , so ignore any modules that have been previously found. RAR 10/04/2007
		s_bInNonExteriorModule = true;
		s_iVisibleModule = -1;

	}
	
	//Check this module if it is non-exterior or if all modules found so far have been exterior
	if(!bExterior || !s_bInNonExteriorModule)
	{
		if(IsModuleVisible(uModuleIdx))
		{
			s_iVisibleModule = uModuleIdx;
		}
	}
}

void Asura_Module_Handler::LinkedInForActiveCamerasCallback( const u_int uModuleIndex )
{
	if (!m_bInitialised) return;

	if(m_pxEnv->IsSceneryModule(uModuleIndex))
	{
		return;//ignore scenery modules //DWE from RAR
	}
	
	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(!m_pxEnv->IsModuleInCurrentLightSet(uModuleIndex)) return;
	#endif

	bool bExterior = m_pxEnv->IsExteriorModule(uModuleIndex);
	if(!bExterior && !s_bInNonExteriorModule)
	{
		//first non-exterior module found , so ignore any modules that have been previously found. RAR 10/04/2007
		s_bInNonExteriorModule = true;
		s_iLinkedInModule = -1;

	}
	
	//Check this module if it is non-exterior or if all modules found so far have been exterior
	if(!bExterior || !s_bInNonExteriorModule)
	{
		if(IsModuleLinkedInForAnyActiveCamera(uModuleIndex))
		{
			s_iLinkedInModule = uModuleIndex;
		}
	}
}

void Asura_Module_Handler::FillModuleList(unsigned short *pusModuleList)
{
	// Fill in the list.
	for (int i=0;i<s_usNumModulesInList;i++)
	{
		pusModuleList[i] = s_ausModuleList[i];
	}
	// Add the terminator.
	if (s_usNumModulesInList < NUM_MODULES_TO_STORE_FOR_CLIPPING)
		pusModuleList[s_usNumModulesInList] = USHRT_MAX;
	// Reset the static data.
	s_usNumModulesInList = 0;
}

void Asura_Module_Handler::SetLengthOfVariableLengthCullingList( const u_short usListLength )
{
	Asura_Assert( usListLength <= MAX_VARIABLE_LENGTH_CLIPPING_COUNT,
		"attempt to set variable length culling list over bounds" );
	s_usMaxNumModulesInVariableList =
		( usListLength <= MAX_VARIABLE_LENGTH_CLIPPING_COUNT ) ?
usListLength :
	static_cast< unsigned short >( MAX_VARIABLE_LENGTH_CLIPPING_COUNT );

#if defined(ASURA_USE_STREAMING)
	// MGH 25/5/2006: This probably won't break anything in non-streaming builds, but
	// I'll #ifdef it out just in case. This is needed for Asura_Streaming_ModuleList
	// to work properly.
	s_usNumModulesInVariableList = 0;
	s_ausVariableLengthModuleList[0] = 0;
#endif //defined(ASURA_USE_STREAMING)
}

void Asura_Module_Handler::VariableLengthCullingListCallback( const u_int uModuleIdx )
{
	if (!m_bInitialised) return;

	if(m_pxEnv->IsSceneryModule(uModuleIdx))
	{
		return;//ignore scenery modules RAR 26/03/03
	}

	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(!m_pxEnv->IsModuleInCurrentLightSet(uModuleIdx)) return;
	#endif


	if(s_usNumModulesInVariableList)
	{
		bool bExterior = m_pxEnv->IsExteriorModule(uModuleIdx);
		bool bListIsExterior = (s_ausVariableLengthModuleList[0] != USHRT_MAX) && m_pxEnv->IsExteriorModule(s_ausVariableLengthModuleList[0]);

		if(bExterior != bListIsExterior)
		{
			if(bExterior)
			{
				//Found an exterior module , but the list already contains some non-exterior modules.
				//So forget about this module.
				return;
			}
			else
			{
				//List currently contains exterior modules , and we have now found a non-exterior module.
				//So clear the existing list , and just add non-exterior modules from now on,.
				s_usNumModulesInVariableList = 0;
			}
		}
	}
	
	if( s_usNumModulesInVariableList == s_usMaxNumModulesInVariableList )
	{
		// Too many modules to store, make data invalid.
		s_ausVariableLengthModuleList[0] = USHRT_MAX;
	}
	else
	{
		s_ausVariableLengthModuleList[s_usNumModulesInVariableList++] = static_cast<u_short>(uModuleIdx);
	}
}

bool Asura_Module_Handler::FillVariableLengthModuleList( u_short* pusModuleList )
{
	bool bValid = (s_ausVariableLengthModuleList[0] != USHRT_MAX) || s_usNumModulesInVariableList == 0;

	// Fill in the list.
	for (int i=0;i<s_usNumModulesInVariableList;i++)
	{
		pusModuleList[i] = s_ausVariableLengthModuleList[i];
	}
	// Add the terminator.
	if (s_usNumModulesInVariableList < MAX_VARIABLE_LENGTH_CLIPPING_COUNT)
		pusModuleList[s_usNumModulesInVariableList] = USHRT_MAX;
	// Reset the static data.
	s_usNumModulesInVariableList = 0;

	return bValid;
}

// MGH 25/5/2006: Fill in the array starting at the specified position.
// Return the number of modules added, or -1 if the list is invalid.
int Asura_Module_Handler::FillVariableLengthModuleListNoTerminator(u_short* pusModuleList)
{
#if defined(ASURA_USE_STREAMING)
	// In streaming builds we zero this element before doing the test.
	// Zero length lists are allowed so we need to distinguish between the two cases.
	const bool bValid = (s_ausVariableLengthModuleList[0] != USHRT_MAX);
#else
	const bool bValid = (s_ausVariableLengthModuleList[0] != USHRT_MAX) || s_usNumModulesInVariableList == 0;
#endif

	// Reset the module count for next time.
	const u_int uNumModules = s_usNumModulesInVariableList;
	s_usNumModulesInVariableList = 0;

	if (!bValid) return -1;

	// Fill in the list.
	for (u_int u=0; u<uNumModules; ++u)
	{
		pusModuleList[u] = s_ausVariableLengthModuleList[u];
	}

	return static_cast<int>(uNumModules);
}


bool Asura_Module_Handler::IsPointInVisibleModule(const Asura_Vector_3& position, int* lastKnownModule)
{
#if defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)
	
	Asura_Assert( false, "Some code needs to be changed to use the Visibility System" );
	return true;
#else	// !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

	if (!m_bInitialised)
	{
		return false;
	}

	if (lastKnownModule)
	{
		int iLastKnownModule = *lastKnownModule;
		if( -1 != iLastKnownModule )
		{
			#ifndef ASURA_TOOLS
			//Disabled this assert when in tools , since this can happen as a result of modules being deleted.	RAR 02/03/2007
			Asura_Assert( 0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules, "Module number is out of bounds" );
			#endif
			
			if(0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules)
			{
				//Make sure the last known module wasn't an exterior module.
				//If it was an exterior module then knowing whether the point is still within it won't help ,since
				//we still need to check for being in non-exterior modules.
				//RAR 10/04/2007
				if(!m_pxEnv->IsExteriorModule(iLastKnownModule))
				{
					const Asura_Vector_3 point = position-m_pxModuleStorage[iLastKnownModule].Position;

					if (IsModuleLinkedInForAnyActiveCamera(iLastKnownModule) &&
						m_pxModuleStorage[iLastKnownModule].m_pxCollisionMesh->GetBoundingBox().Contains(point))
					{
						return true;
					}
				}
			}
		}
	}

	int iFoundExteriorModule = -1;

	for (int m=0; m<m_iNumberOfModules; m++)
	{
		if (m_pxEnv && m_pxEnv->IsSceneryModule(m))
		{
			continue;//ignore scenery modules RAR 26/03/03
		}

		#ifdef ASURA_TOOLS
		//Ignore modules that aren't part of the current light set. RAR 15/02/2008
		if(m_pxEnv && !m_pxEnv->IsModuleInCurrentLightSet(m)) continue;
		#endif


		const Asura_Vector_3 point = position-m_pxModuleStorage[m].Position;

		if (IsModuleLinkedInForAnyActiveCamera(m) &&
			m_pxModuleStorage[m].m_pxCollisionMesh->GetBoundingBox().Contains(point))
		{
			if(m_pxEnv->IsExteriorModule(m))
			{
				//In an exterior module , take note of it for now until we can determine whether the position
				//is in a non-exterior module. RAR 10/04/2007
				iFoundExteriorModule = m;

			}
			else
			{
				if (lastKnownModule)
				{
					*lastKnownModule = m;	
				}
			}
			return true;
		}
	}

	if(iFoundExteriorModule != -1)
	{
		//The position was in a linked-in exterior module.
		//However we can't accept this module if the position was also in a non-exterior module.
		//So check for that now.
		Asura_Bounding_Box xBB;
		xBB.SetToPoint(position);

		if(!DoesBBIntersectNonExteriorModule(xBB))
		{
			//Point was only within the exterior module , so go with that
			if (lastKnownModule)
			{
				*lastKnownModule = iFoundExteriorModule;	
			}

			return true;
		}

	}

	return false;
#endif	// !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)
}

// (MH - 08.05.2002) addition for vis module culling of characters
bool Asura_Module_Handler::DoesBBIntersectVisibleModule( const Asura_Bounding_Box& xBB, int* piLastKnownModule /*= NULL*/ )
{
#if defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)
	Asura_Assert( false, "Some code needs to be changed to use the Visibility System" );
	return true;
#else	// defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

	if( piLastKnownModule )
	{
		int iLastKnownModule = *piLastKnownModule;
		if( -1 != iLastKnownModule )
		{
			#ifndef ASURA_TOOLS
			//Disabled this assert when in tools , since this can happen as a result of modules being deleted.	RAR 02/03/2007
			Asura_Assert( 0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules, "Module number is out of bounds" );
			#endif
			
			if(0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules)
			{
				//Make sure the last known module wasn't an exterior module.
				//If it was an exterior module then knowing whether the point is still within it won't help ,since
				//we still need to check for being in non-exterior modules.
				//RAR 10/04/2007
				if(!m_pxEnv->IsExteriorModule(iLastKnownModule))
				{
					if( IsModuleVisible(iLastKnownModule) )
					{
						if( xBB.Intersects(m_pxModuleStorage[iLastKnownModule].m_pxCollisionMesh->GetBoundingBox(), m_pxModuleStorage[iLastKnownModule].Position) )
						{
							return true;
						}
					}
				}
			}
		}
	}

	s_iVisibleModule = -1;
	s_bInNonExteriorModule = false;
	TestModuleTree( xBB, CullingTestModuleInternalCallback );

	if( piLastKnownModule )
	{
		*piLastKnownModule = s_iVisibleModule;
	}

	return( -1 != s_iVisibleModule );
#endif	// defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)
}

bool Asura_Module_Handler::DoesBBIntersectLinkedInModule(const Asura_Bounding_Box& xBB)
{
	s_iLinkedInModule = -1;
	s_bInNonExteriorModule = false;

	TestModuleTree(xBB, LinkedInForActiveCamerasInternalCallback);


	return s_iLinkedInModule != -1;
}

bool Asura_Module_Handler::DoesBBIntersectLinkedInModule(const Asura_Bounding_Box& xBB, int* piLastKnownModule )
{
	if( piLastKnownModule )
	{
		int iLastKnownModule = *piLastKnownModule;
		if( -1 != iLastKnownModule )
		{
			#ifndef ASURA_TOOLS
			//Disabled this assert when in tools , since this can happen as a result of modules being deleted.	RAR 02/03/2007
			Asura_Assert( 0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules, "Module number is out of bounds" );
			#endif
			
			if(0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules)
			{
				//Make sure the last known module wasn't an exterior module.
				//If it was an exterior module then knowing whether the point is still within it won't help ,since
				//we still need to check for being in non-exterior modules.
				//RAR 10/04/2007
				if(!m_pxEnv->IsExteriorModule(iLastKnownModule))
				{
					if(IsModuleLinkedInForAnyActiveCamera(iLastKnownModule) )
					{
						if( xBB.Intersects(m_pxModuleStorage[iLastKnownModule].m_pxCollisionMesh->GetBoundingBox(), m_pxModuleStorage[iLastKnownModule].Position) )
						{
							return true;
						}
					}
				}
			}
		}
	}


	s_iLinkedInModule = -1;
	s_bInNonExteriorModule = false;

	TestModuleTree(xBB, LinkedInForActiveCamerasInternalCallback);

	return s_iLinkedInModule != -1;
}

bool Asura_Module_Handler::IsModuleInFrustum(int moduleIndex) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");

	if( IsUseFrustumCulling() )
	{
		Asura_Bounding_Box xAABB = GetBoundingBoxOfModule( moduleIndex );
		xAABB.Displace( GetPositionOfModule( moduleIndex ) );

		return Asura_Camera::IsAABBInFrustum( xAABB ); 
			//Asura_Camera::IsSphereInFrustum( m_pxModuleStorage[moduleIndex].Centre, m_pxModuleStorage[moduleIndex].m_pxCollisionMesh->GetBoundingRadius() );
	}	
	else
	{
		return true;
	}
}

bool Asura_Module_Handler::IsModuleLinkedInForCurrentCamera(int moduleIndex) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");
	return (m_pucLinkedInList[moduleIndex]&Asura_Camera::GetCameraIndexBitFlag())!=0;
}

bool Asura_Module_Handler::IsModuleLinkedInForAnyActiveCamera(int moduleIndex) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");


	if(Asura_Network::IsExternalNetworkActive() && Asura_Network::IsServer())
	{
		return ((m_pucLinkedInList[moduleIndex]&ASURA_MODULE_LINKEDINFORPLAYER_BITFLAG) != 0);
	}
	else
	{
		for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
		{
			Asura_Camera::SetCameraIndex(i);

			if (Asura_Camera::IsCameraActive())
			{
				if((m_pucLinkedInList[moduleIndex]&Asura_Camera::GetCameraIndexBitFlag())!=0)
				{
					return true;
				}
			}	
		}

		return false;
	}
}

void Asura_Module_Handler::ModuleIndexFromPositionCallback( const u_int uIdx )
{
	if( m_pxEnv && !m_pxEnv->IsSceneryModule(uIdx) )
	{
		#ifdef ASURA_TOOLS
		//Ignore modules that aren't part of the current light set. RAR 15/02/2008
		if(!m_pxEnv->IsModuleInCurrentLightSet(uIdx)) return;
		#endif
		
		
		//Prefer finding non-exterior modules over exterior modules.
		bool bExterior = m_pxEnv && m_pxEnv->IsExteriorModule(uIdx);
		if(!bExterior || s_iLastModuleIndex == -1)
		{
			s_iLastModuleIndex = uIdx;
		}
	}
}

void Asura_Module_Handler::ModuleIndexIncludingSceneryFromPositionCallback( const u_int uIdx )
{
	#ifdef ASURA_TOOLS
	//Ignore modules that aren't part of the current light set. RAR 15/02/2008
	if(m_pxEnv && !m_pxEnv->IsModuleInCurrentLightSet(uIdx)) return;
	#endif

	//Prefer finding non-exterior modules over exterior modules.
	bool bExterior = m_pxEnv && m_pxEnv->IsExteriorModule(uIdx);
	if(!bExterior || s_iLastModuleIndex == -1)
	{
		s_iLastModuleIndex = uIdx;
	}
}

// (NJD - 25.04.2003) - altered search to use the module tree
// MH 5.4.02 -- returns index of first module found whose bounding box contains the point.
// Returns -1 if point not found in any module.
// (MH - 10.06.2002) added last_known_module_id, which is checked first, if not -1
int Asura_Module_Handler::GetModuleIndexFromPosition (const Asura_Vector_3& position,
													  int last_known_module_id,
													  const bool bCheckScenery)
{
	// if we have a last known module check to see if we're still inside it
	if( 0 <= last_known_module_id && last_known_module_id < m_iNumberOfModules )
	{
		//Make sure the last known module wasn't an exterior module.
		//If it was an exterior module then knowing whether the point is still within it won't help ,since
		//we still need to check for being in non-exterior modules.
		//RAR 10/04/2007
		if(!m_pxEnv->IsExteriorModule(last_known_module_id))
		{
			// RDM - 14/08/06 - Added collision mesh pointer test to hopefully fix crashes. There's no 
			// guarantee a module will have collision mesh anymore.
			Asura_Vector_3 local_position = position - m_pxModuleStorage[last_known_module_id].Position;
			if( m_pxModuleStorage[last_known_module_id].m_pxCollisionMesh
			 && m_pxModuleStorage[last_known_module_id].m_pxCollisionMesh->GetBoundingBox().Contains(local_position) )
			{
				return last_known_module_id;
			}
		}
	}

	// do an exhaustive check
	s_iLastModuleIndex = -1;

	if( m_pxLBAABBModuleTree )
	{
		Asura_Bounding_Box xBB;
		xBB.Reset();
		xBB.Set( position );

		if(bCheckScenery)
		{
			TestModuleTree( xBB, ModuleIndexIncludingSceneryFromPositionInternalCallback );
		}
		else
		{
			TestModuleTree( xBB, ModuleIndexFromPositionInternalCallback );
		}
	}

	return s_iLastModuleIndex;
}

void Asura_Module_Handler::CameraModuleIndexFromPositionCallback( const u_int uIdx )
{
	if( m_pxEnv && !m_pxEnv->IsSceneryModule(uIdx) && !m_pxEnv->IsNeverCameraModule(uIdx))
	{
		#ifdef ASURA_TOOLS
		//Ignore modules that aren't part of the current light set. RAR 15/02/2008
		if(!m_pxEnv->IsModuleInCurrentLightSet(uIdx)) return;
		#endif
		
		
		//Prefer finding non-exterior modules over exterior modules.
		bool bExterior = m_pxEnv && m_pxEnv->IsExteriorModule(uIdx);
		if(!bExterior || s_iLastModuleIndex == -1)
		{
			s_iLastModuleIndex = uIdx;
		}
	}
}

int Asura_Module_Handler::GetCameraModuleIndexFromPosition (const Asura_Vector_3& xPos,
													  int iLastKnownModule)
{
	// if we have a last known module check to see if we're still inside it
	if( 0 <= iLastKnownModule && iLastKnownModule < m_iNumberOfModules )
	{
		//Make sure the last known module wasn't an exterior module.
		//If it was an exterior module then knowing whether the point is still within it won't help ,since
		//we still need to check for being in non-exterior modules.
		//RAR 10/04/2007
		if(!m_pxEnv->IsExteriorModule(iLastKnownModule))
		{
			Asura_Vector_3 xLocalPos = xPos - m_pxModuleStorage[iLastKnownModule].Position;
			if( m_pxModuleStorage[iLastKnownModule].m_pxCollisionMesh
			 && m_pxModuleStorage[iLastKnownModule].m_pxCollisionMesh->GetBoundingBox().Contains(xLocalPos) )
			{
				return iLastKnownModule;
			}
		}
	}

	// do an exhaustive check
	s_iLastModuleIndex = -1;

	if( m_pxLBAABBModuleTree )
	{
		Asura_Bounding_Box xBB;
		xBB.Reset();
		xBB.Set( xPos );

		TestModuleTree( xBB, CameraModuleIndexFromPositionInternalCallback );
	}

	return s_iLastModuleIndex;
}


void Asura_Module_Handler::ModuleTestCallback(const u_int uIndex)
{
	// (AW - 18.08.2005) - The main point of this function is to convert the module 
	// index before calling the user callback function. However, since the AABB trees
	// are never perfectly accurate it's worth testing against the real BB before
	// continuing.

	Asura_Assert(m_puModuleTreeLBAABBConversionArray, "Some m_puModuleTreeLBAABBConversionArray is NULL, dunno what todo, wont do anything");
	if (!m_puModuleTreeLBAABBConversionArray) return;

	u_int uConvertedIndex = m_puModuleTreeLBAABBConversionArray[uIndex];
	if (s_xModuleTestBB.Intersects(m_pxModuleStorage[uConvertedIndex].m_pxCollisionMesh->GetBoundingBox(), m_pxModuleStorage[uConvertedIndex].Position))
	{
		s_pModuleTestCallbackFunction(uConvertedIndex);
	}
}

#endif //ASURA_MAIN

#ifdef ASURA_TOOLS
void Asura_Module_Handler::SetModuleLink(int iModuleIndex,int iLinkIndex,Asura_Hash_ID uLinkZoneHash,bool bLink)
{
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<m_iNumberOfModules,"Module number is out of bounds");

	//get the destination handler and connection index

	#if defined ASURA_USE_STREAMING || defined ASURA_ASSERT
		Asura_Module_Handler* pxLinkHandler = this;
	#endif

	u_int uConnectionIndex = 0;
	#ifdef ASURA_USE_STREAMING
	if(uLinkZoneHash && uLinkZoneHash != m_uZoneHash)
	{
		for(u_int i=0;i<m_uNumConnectedHandlers;i++)
		{
			if(!m_ppxConnectedHandlers[i]) continue;
			if(m_ppxConnectedHandlers[i]->m_uZoneHash == uLinkZoneHash)
			{
				pxLinkHandler = m_ppxConnectedHandlers[i];
				uConnectionIndex = i+1;
				break;
			}
		}
		if(!uConnectionIndex) return;
	}
	#endif   //ASURA_USE_STREAMING
	Asura_Assert(iLinkIndex>=0 && iLinkIndex<pxLinkHandler->GetNumberOfModules(),"Module number is out of bounds");

	//Don't link a module to itself
	if(iModuleIndex==iLinkIndex && uConnectionIndex == 0) return;

	m_pxModuleStorage[iModuleIndex].SetModuleLink(iLinkIndex|(uConnectionIndex<<16),bLink);
}
#endif //ASURA_TOOLS

const Asura_CollisionMesh* Asura_Module_Handler::GetCollisionMeshForModule(int iModuleIndex)
{
	Asura_Assert(iModuleIndex>=0 && iModuleIndex<m_iNumberOfModules,"Module number is out of bounds");
	#ifdef ASURA_TOOLS
	if(m_pxModuleStorage[iModuleIndex].m_bCollisionMeshInvalid || !m_pxModuleStorage[iModuleIndex].m_pxCollisionMesh) UpdateInvalidCollisionMeshes();
	#endif
	return m_pxModuleStorage[iModuleIndex].m_pxCollisionMesh;
}

void Asura_Module_Handler::GetPositionOfModule(int moduleIndex, Asura_Vector_3& position) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");
	position = m_pxModuleStorage[moduleIndex].Position;
}

const Asura_Vector_3& Asura_Module_Handler::GetCentreOfModule( int iModuleIndex ) const
{
	Asura_Assert( iModuleIndex >= 0 && iModuleIndex < m_iNumberOfModules, "Module number is out of bounds" );
	return m_pxModuleStorage[iModuleIndex].Centre;
}

const Asura_Bounding_Box& Asura_Module_Handler::GetBoundingBoxOfModule( int iModuleIndex ) const
{
	Asura_Assert( iModuleIndex >= 0 && iModuleIndex < m_iNumberOfModules, "Module number is out of bounds" );
	return m_pxModuleStorage[iModuleIndex].m_pxCollisionMesh->GetBoundingBox();
}

float Asura_Module_Handler::GetRadiusOfModule(int moduleIndex) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");
	return m_pxModuleStorage[moduleIndex].m_pxCollisionMesh->GetBoundingRadius();
}


#ifdef ASURA_TOOLS
Asura_Mesh* Asura_Module_Handler::GetMeshForModule(int moduleIndex) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");
	return m_pxModuleStorage[moduleIndex].Mesh;
}
#endif


const Asura_Module* Asura_Module_Handler::GetModuleData(int moduleIndex) const
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");

	return &m_pxModuleStorage[moduleIndex];
}

Asura_Module* Asura_Module_Handler::GetModuleData(int moduleIndex)
{
	Asura_Assert(moduleIndex>=0 && moduleIndex<m_iNumberOfModules,"Module number is out of bounds");

	return &m_pxModuleStorage[moduleIndex];
}

void Asura_Module_Handler::SetUseCameraPositionOverride( bool bUseOverride, const Asura_Vector_3* pxOverride/*=NULL*/ )
{
	Asura_Assert( (pxOverride || !bUseOverride), "Using position override without specifying a position!" );

	m_bUseCameraPositionOverride = bUseOverride;
	
	if( m_bUseCameraPositionOverride && pxOverride )
	{
		m_xCameraPositionOverride = *pxOverride;
	}
}


#ifdef ASURA_TOOLS
void Asura_Module_Handler::SetupModuleMeshForFakeEnv( int iModuleIdx, Asura_Mesh* pxMesh )
{
	Asura_Assert( 0 <= iModuleIdx && iModuleIdx < m_iNumberOfModules, "Invalid module index." );
	m_pxModuleStorage[iModuleIdx].m_pxDefaultMesh = pxMesh;
	m_pxModuleStorage[iModuleIdx].Mesh = pxMesh;
}
#endif

#ifdef ASURA_TOOLS
void Asura_Module_Handler::SetupModulePosForFakeEnv( int iModuleIdx, const Asura_Vector_3& xPos )
{
	Asura_Assert( 0 <= iModuleIdx && iModuleIdx < m_iNumberOfModules, "Invalid module index." );
	m_pxModuleStorage[iModuleIdx].Position = xPos;
}
#endif


#ifdef ASURA_TOOLS
void Asura_Module_Handler::InvalidateCollisionMesh(int iModuleIndex)
{
	Asura_Assert( 0 <= iModuleIndex && iModuleIndex < m_iNumberOfModules, "Invalid module index." );
	m_pxModuleStorage[iModuleIndex].m_bCollisionMeshInvalid = true;
	s_bCollisionMeshesInvalid = true;

	#ifdef ASURA_MAIN
	//If a collision mesh is invalidated ,then any adjacency information for it will be invalid as well
	//RAR 08/06/2006
	Tools_Adjacency::InvalidateCollisionModule(iModuleIndex);


	//Physics environment will also be invalid
	//RAR 27/09/2006
	Asura_Physics_Environment::Invalidate();

	#endif


}
#endif

#ifdef ASURA_TOOLS
void Asura_Module_Handler::UpdateInvalidCollisionMeshes()
{
	if(!s_bCollisionMeshesInvalid) return;
	s_bCollisionMeshesInvalid = false;


	//calculate the collision mesh for each module as needed
	Asura_Module* pxModule = m_pxModuleStorage;
	for(int i=0;i<m_iNumberOfModules;i++,pxModule++)
	{
		if(!pxModule->Mesh) continue;
		if(pxModule->m_pxCollisionMesh && !pxModule->m_bCollisionMeshInvalid) continue;

		if( !pxModule->Mesh->BoundingBoxDataValid )
		{
			pxModule->Mesh->CalculateBoundingBox();
		}

		//collision mesh needs to be redone for this module
		pxModule->UpdateCollisionMesh();
		pxModule->m_bCollisionMeshInvalid = false;


	}


	//delete the old module tree
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		Asura_ResourceDeallocateArray(m_puModuleTreeLBAABBConversionArray);
		m_puModuleTreeLBAABBConversionArray = NULL;
	}
	Asura_ResourcePopType();

	if (m_pxLBAABBModuleTree)
	{
		m_pxLBAABBModuleTree->DeleteNodes();
		delete m_pxLBAABBModuleTree;
		m_pxLBAABBModuleTree = NULL;
	}

	//and regenerate them
	InitMainModuleTree();
	InitAllModuleTrees();
}
#endif

#ifdef ASURA_TOOLS
void Asura_Module_Handler::DeleteModule(int iModuleIndex)
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		Asura_Assert(iModuleIndex>=0 && iModuleIndex< m_iNumberOfModules,"");

		if(m_pucRenderList)
		{
			m_pucRenderList[iModuleIndex] = m_pucRenderList[m_iNumberOfModules-1];
		}
		if(m_pucLinkedInList)
		{
			m_pucLinkedInList[iModuleIndex] = m_pucLinkedInList[m_iNumberOfModules-1];
		}


		if(m_bModuleStorageAllocated)
		{
			for(int i=0;i<m_iNumberOfModules;i++)
			{
				m_pxModuleStorage[i].DeleteModule(iModuleIndex);
			}

			Asura_ResourceDeallocateArray(m_pxModuleStorage[iModuleIndex].VisibleModules);
			delete m_pxModuleStorage[iModuleIndex].m_pxCollisionMesh;
			m_pxModuleStorage[iModuleIndex] = m_pxModuleStorage[m_iNumberOfModules-1];
			m_pxModuleStorage[m_iNumberOfModules-1].VisibleModules = 0;
			m_pxModuleStorage[m_iNumberOfModules-1].m_pxCollisionMesh = 0;
		}

		m_iNumberOfModules--;

		Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
		{
			Asura_ResourceDeallocateArray(m_puModuleTreeLBAABBConversionArray);
			m_puModuleTreeLBAABBConversionArray = NULL;
		}
		Asura_ResourcePopType();

		if (m_pxLBAABBModuleTree)
		{
			m_pxLBAABBModuleTree->DeleteNodes();
			delete m_pxLBAABBModuleTree;
			m_pxLBAABBModuleTree = NULL;
		}

		InitMainModuleTree();
	}
	Asura_ResourcePopType();
}
#endif // ASURA_TOOLS




#ifdef ASURA_USE_STREAMING
void Asura_Module_Handler::FindConnectedHandlers()
{
	//skip this if it hasn't been initialised yet
	if(!m_bInitialised || (m_uNumConnectedHandlers == 0)) return;

	const Asura_Streaming_Zone* pxZone = Asura_Streaming_System::FindZone(m_uZoneHash);
	Asura_Assert(pxZone,"Failed to find streaming zone the corresponds to this module handler");

	#ifdef ASURA_TOOLS
	if(pxZone->GetNumConnection() != m_uNumConnectedHandlers)
	{
		//need to reallocate array
		Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
		
		Asura_ResourceDeallocateArray(m_ppxConnectedHandlers);
		m_uNumConnectedHandlers = pxZone->GetNumConnection();
		if(m_uNumConnectedHandlers)
		{
			Asura_ResourceAllocateArray(m_ppxConnectedHandlers,m_uNumConnectedHandlers);
		}
		else
		{
			m_ppxConnectedHandlers = NULL;
		}
	
		Asura_ResourcePopType();
	}
	#endif

	//Find the handler for each connected zone. Some may not exist
	for(u_int i=0;i<m_uNumConnectedHandlers;i++)
	{
		const Asura_Streaming_Connection* pxConnection = (i < pxZone->GetNumConnection())
														? pxZone->GetConnection(i)
														: NULL;
		if(pxConnection == NULL)
		{
			m_ppxConnectedHandlers[i] = NULL;
		}
		else
		{
			Asura_Hash_ID uConnectionHash = pxConnection->GetToZoneHashID();

			Asura_Module_Handler* pxConnectedHandler = Asura_Module_System::GetHandler(uConnectionHash);
			//Don't want links to handlers that hanve't been initialised
			if(pxConnectedHandler && !pxConnectedHandler->m_bInitialised) pxConnectedHandler = 0;
			m_ppxConnectedHandlers[i] = pxConnectedHandler;
		}
	}
}
#endif //ASURA_USE_STREAMING


#ifdef ASURA_TOOLS
bool Asura_Module_Handler::SwitchPlatformMesh(int iModuleIndex, u_int uPlatformBit)
{
	Asura_Assert( 0 <= iModuleIndex && iModuleIndex < m_iNumberOfModules, "Invalid module index." );

	Asura_Module* pxModule = &m_pxModuleStorage[iModuleIndex];

	//See if there is any mesh for this platform
	Asura_Mesh* pxMeshToUse = pxModule->m_pxDefaultMesh;

	while(pxMeshToUse && !(pxMeshToUse->m_uPlatforms & uPlatformBit))
	{
		pxMeshToUse = pxMeshToUse->m_pxNextPlatformMesh;
	}
	//If none was used , go with the default mesh
	if(!pxMeshToUse) pxMeshToUse = pxModule->m_pxDefaultMesh;

	//already using this mesh?
	if(pxMeshToUse == pxModule->Mesh) return false;

	//Change to the new mesh. Also need to invalidate the collision mesh , since it will have been based on the old mesh.
	pxModule->Mesh = pxMeshToUse;
	InvalidateCollisionMesh(iModuleIndex);

	return true;
	
}
#endif //ASURA_TOOLS


bool Asura_Module_Handler::DoesBBIntersectNonExteriorModule(const Asura_Bounding_Box& xBB)
{
	s_bInNonExteriorModule = false;
	TestModuleTree(xBB,NonExteriorModuleInternalCallback);

	return s_bInNonExteriorModule;
}

void Asura_Module_Handler::NonExteriorModuleCallback( const u_int uModuleIdx )
{
	//ignore scenery and exterior modules
	if(m_pxEnv)
	{
		if(m_pxEnv->GetModuleFlags(uModuleIdx) & (ASURA_MODULE_FLAG_SCENERY | ASURA_MODULE_FLAG_EXTERIOR)) return;
	}

	s_bInNonExteriorModule = true;
}

bool Asura_Module_Handler::IsExteriorModule( u_int uModuleIndex) const
{
	return m_pxEnv && m_pxEnv->IsExteriorModule(uModuleIndex);
}


//////////////////////////////////////////////////////////////////////////////
//
// Asura_Module_System
//
//////////////////////////////////////////////////////////////////////////////

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool											Asura_Module_System::UseModLinks				= true;
#if defined(ASURA_TOOLS)
bool											Asura_Module_System::s_bRenderAllWhenInNoModule	= true;
#endif	// defined(ASURA_TOOLS)

#if defined(ASURA_USE_STREAMING)
Asura_TemplateHashTree<Asura_Module_Handler>	Asura_Module_System::s_xHandlerTree;
Asura_Module_Handler*							Asura_Module_System::s_pxCurrentHandler	= NULL;
#else
Asura_Module_Handler							Asura_Module_System::s_xHandler;
#endif	// defined(ASURA_USE_STREAMING)


// M E M B E R   F U N C T I O N S ///////////////////////////////////////////
void Asura_Module_System::Initialise(void)
{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && !defined(ASURA_COMMAND_CONSOLE_READONLY)
	Asura_CommandConsole::AddVar( L"Render.UseModLinks", &UseModLinks, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
}


void Asura_Module_System::Shutdown(void)
{
	Reset();
}


void Asura_Module_System::Reset (void)
{
#if defined(ASURA_USE_STREAMING)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	Asura_Module_Handler&	xHandler	= GetCurrentHandler();

	// Reset the current instance...
	xHandler.Reset();
#endif	// defined(ASURA_USE_STREAMING)
}


#if defined(ASURA_USE_STREAMING)
Asura_Module_Handler* Asura_Module_System::CreateHandler (const Asura_Hash_ID uHandlerHash,
														  const Asura_Hash_ID uZoneHash)
{
	Asura_Module_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHandlerHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHandlerHash);

		//store hash ID in handler
		if (pxHandler)
		{
			pxHandler->m_uZoneHash    = uZoneHash;
			pxHandler->m_uHandlerHash = uHandlerHash;
		}
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an animation handler with this hash-ID.");
	}

	// Return the handler...
	return(pxHandler);
}

bool Asura_Module_System::DestroyHandler (const Asura_Hash_ID uHandlerHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHandlerHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Call the delete function...
	bool bSuccess = s_xHandlerTree.DeleteNodeByHash(uHandlerHash);

	//Make sure that the other handlers don't have pointers to the deleted one any more. RAR 07/04/2006
	if(bSuccess)
	{
		FindConnectedHandlers();
	}
	return bSuccess;
}
#endif	// defined(ASURA_USE_STREAMING)


void Asura_Module_System::Render (void)
{
	// ARTTODO - go through each module system here...
	GetCurrentHandler().Render();
}


#ifdef ASURA_TARGET_SM3
void Asura_Module_System::RenderLate(void)
{
	GetCurrentHandler().RenderLate();
}
#endif


#ifdef ASURA_MAIN
void Asura_Module_System::Update (const Asura_Vector_3* pxViewPosition, const int* piModuleIndex)
{
	// KJFL - 20 February 2008 -  Asura_Module_System::Update() called from lots of places, so
	// this makes it a bit easier to catch them all. Really I need to change things like project
	// code to point to the Visibility system - once I've tracked down and changed all the engine
	// calls.
	#if ( defined(ASURA_USE_NEW_VISIBILITY_SYSTEM) )
	Asura_Visibility_System::Update();

	// fill in the module visibilities to provide a bridge to older systems - will eventually be
	// redundant.
	{
		Asura_Module_Handler&	xHandler	= GetCurrentHandler();
	
		Asura_Module_System::ClearRenderAndLinkedInListsForCurCam();
		
		#ifdef ASURA_TOOLS
		// KJFL - 23 April 2008 - as part of factoring out modules, this now says they are visible all the time if
		// we're in Tools. This fixes issues with collision mesh modules not showing up in Tools.

		if( Tools_System::IsActive() )
		{
			for ( int iModule=0; iModule<xHandler.GetNumberOfModules(); ++iModule )
			{
				if ( IsModuleInFrustum( iModule ) )
				{
					xHandler.GetRenderList()[iModule] = Asura_Camera::GetCameraIndexBitFlag();
				}
			}
			return;
		}
		#endif

		Asura_Visibility_System::GetVisibleFeaturesFromCamera( xHandler.GetRenderList(), GetNumberOfModules() );
	}

	#else // not using ASURA_USE_NEW_VISIBILITY_SYSTEM

#if defined(ASURA_TOOLS)
	// Initialise the counts...
	Asura_Module_Handler::s_uNumberOfLinkedInPolys     = 0;
	Asura_Module_Handler::s_uNumberOfLinkedInTriangles = 0;
	Asura_Module_Handler::s_uNumLinkedInRenderMeshTris = 0;
#endif	// defined(ASURA_TOOLS)

	g_bFirstRenderListModuleFound = true;
	g_bUpdateLinkedModules_InNonExteriorModule = false;

	Asura_Module_Handler&	xHandler	= GetCurrentHandler();
	if(Asura_Module_System::UseModLinks)
	{

		#ifdef ASURA_USE_STREAMING
		if(piModuleIndex)
		{
			//If a module index has been passed , assume it is for the current handler
			xHandler.Update(pxViewPosition, piModuleIndex);
		}
		else
		{
			#ifdef ASURA_TOOLS
			//Otherwise , check them all
			xHandler.Update(pxViewPosition, NULL);
			#else
			// Traverse the handlers, until we find a matching module.
			// MGH: 29/1/2007: This should fix module linking problems that occur when the camera is
			// in a different zone from the player.
			s_xHandlerTree.Traverse(UpdateCallback, (void*)pxViewPosition);
			#endif

			AddAllModulesWithoutModLinkingToRenderListForCurCam();
		}
		#else
		xHandler.Update(pxViewPosition, piModuleIndex);
		#endif

		#ifdef ASURA_TOOLS
		// Render all in no module if option selected
		if( !Asura_Module_Handler::s_uNumberOfLinkedInPolys && Asura_Module_System::GetRenderAllWhenInNoModule())
		{
			#if ( !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM) )
			AddAllModulesToRenderListForCurCam();
			#endif
		}
		#endif
	}
	else
	{
		//not using module links , so add everything.
		AddAllModulesToRenderListForCurCam();
		AddAllModulesToLinkedInListForCurCam();
	}

	// Cull linked-in modules against frustum.
	CullRenderListForCurCam();

#ifdef ASURA_TOOLS
	xHandler.CheckIfRenderListChanged();
#endif
	#endif // not using ASURA_USE_NEW_VISIBILITY_SYSTEM

	#if defined (ASURA_USE_TEXTURE_STREAMING) && !defined (ASURA_TOOLS)
	Asura_Texture_Streaming_System::Update();
	#endif
}
#endif //ASURA_MAIN

#ifdef ASURA_USE_STREAMING
bool Asura_Module_System::UpdateCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler,void* pPosition)
{
	pxHandler->Update((const Asura_Vector_3*) pPosition,0);

	// Stop the traversal as soon as we find a module containing the position.
	return g_bFirstRenderListModuleFound;
}
#endif //ASURA_USE_STREAMING


//////////////////
// Asura_Module //
//////////////////
Asura_Module::Asura_Module()
{
	VisibleModules = 0;
	VisibleModulesArraySize = 0;
	NumberOfVisibleModules = 0;
	m_pxCollisionMesh = 0;
	#ifdef ASURA_TOOLS
	m_pxDefaultMesh = 0;
	Mesh = 0;
	m_bCollisionMeshInvalid = false;
	#endif
}

Asura_Module::~Asura_Module()
{
	Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
	{
		Asura_ResourceDeallocateArray(VisibleModules);
		delete m_pxCollisionMesh;
	}
	Asura_ResourcePopType();
}

#ifdef ASURA_TOOLS
void Asura_Module::SetModuleLink(int linkIndex,bool link)
{
	if(link)
	{
		//make sure the module isn't already linked in
		for(int i=0;i<NumberOfVisibleModules;i++)
		{
			if(linkIndex==VisibleModules[i]) return;
		}
		//increase memory allocation if need be
		if(NumberOfVisibleModules==VisibleModulesArraySize)
		{
			Asura_ResourcePushType(ASURA_RESOURCE_MODULE);
			{
				int* newArray;
				Asura_ResourceAllocateArray(newArray, VisibleModulesArraySize+8);
				if(VisibleModules)
				{
					Asura_Mem::Copy(newArray,VisibleModules,(VisibleModulesArraySize)*sizeof(int));
					Asura_ResourceDeallocateArray(VisibleModules);
				}
				VisibleModules = newArray;
				VisibleModulesArraySize+=8;
			}
			Asura_ResourcePopType();
		}
		//add this link
		VisibleModules[NumberOfVisibleModules] = linkIndex;
		NumberOfVisibleModules++;
	}
	else
	{
		//search for the entry in the array of visible modules
		for(int i=0;i<NumberOfVisibleModules;i++)
		{
			if(linkIndex==VisibleModules[i])
			{
				NumberOfVisibleModules--;
				//copy the last entry into this slot
				VisibleModules[i] = VisibleModules[NumberOfVisibleModules];
				break;
			}
		}
	}
}

void Asura_Module::ClearModuleLinks()
{
	delete [] VisibleModules;
	VisibleModules = 0;
	NumberOfVisibleModules = 0;
	VisibleModulesArraySize = 0;

}

void Asura_Module::DeleteModule(int iModuleIndex)
{
	//remove the deleted module from the visibilities
	int iNumberOfModules= Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<NumberOfVisibleModules;i++)
	{
		if(VisibleModules[i] == iModuleIndex)
		{
			VisibleModules[i--] = VisibleModules[--NumberOfVisibleModules];
		}
		else if(VisibleModules[i] == iNumberOfModules-1)
		{
			VisibleModules[i] = iModuleIndex;
		}
	}
	
}

void Asura_Module::UpdateCollisionMesh()
{
	if(!Mesh) return;
	if(m_pxCollisionMesh)
	{
		if(Mesh->m_pxCollisionMesh)
		{
			//copy collision mesh from the collision mesh stored in the Asura_Mesh
			m_pxCollisionMesh->Copy(Mesh->m_pxCollisionMesh);
		}
		else
		{
			//recalculate collision mesh from Mesh
			m_pxCollisionMesh->UpdateFromMesh(Mesh);
		}
	}
	else
	{
		if(Mesh->m_pxCollisionMesh)
		{
			//create collision mesh from collision mesh stored in the Asura_Mesh
			m_pxCollisionMesh = new Asura_CollisionMesh(Mesh->m_pxCollisionMesh);
		}
		else
		{
			//create collision mesh from Asura_Mesh
			m_pxCollisionMesh = new Asura_CollisionMesh(Mesh);
		}
	}
}
#endif // #ifdef ASURA_TOOLS

#if defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)
void Asura_Module::RemapModuleLinkConnectionIndices(u_int uArraySize,const int* piRemap)
{
	int iNewNumberOfVisibleModules = 0;
	for(int i=0;i<NumberOfVisibleModules;i++)
	{
		int iConnectionIndex = (VisibleModules[i] >> 16)-1;
		int iModuleIndex = VisibleModules[i] & 0xffff;

		if(iConnectionIndex >= 0)
		{
			if(iConnectionIndex>=(int)uArraySize || piRemap[iConnectionIndex] == -1)
			{
				//connection index no longer exists
				continue;
			}

			iConnectionIndex = piRemap[iConnectionIndex];
		}

		VisibleModules[iNewNumberOfVisibleModules++] = ((iConnectionIndex+1)<<16) | iModuleIndex;
	}
	NumberOfVisibleModules = iNewNumberOfVisibleModules;
}
#endif //defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)

#if defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)
void Asura_Module::RemapModuleLinkModuleIndices(u_int uConnectionIndexToChange,u_int uArraySize,const int* piRemap)
{
	int iNewNumberOfVisibleModules = 0;
	for(int i=0;i<NumberOfVisibleModules;i++)
	{
		int iConnectionIndex = (VisibleModules[i] >> 16)-1;
		int iModuleIndex = VisibleModules[i] & 0xffff;

		if(iConnectionIndex == (int) uConnectionIndexToChange)
		{
			if(iModuleIndex >= (int)uArraySize || piRemap[iModuleIndex] == -1)
			{
				//module no longer exists
				continue;
			}

			iModuleIndex = piRemap[iModuleIndex];
		}

		VisibleModules[iNewNumberOfVisibleModules++] = ((iConnectionIndex+1)<<16) | iModuleIndex;
	}
	NumberOfVisibleModules = iNewNumberOfVisibleModules;
}

#endif //defined(ASURA_TOOLS) && defined(ASURA_USE_STREAMING)

#ifdef ASURA_USE_STREAMING
void Asura_Module_System::FindConnectedHandlers()
{
	s_xHandlerTree.Traverse(FindConnectedHandlersCallback);
}

void Asura_Module_System::FindConnectedHandlersCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler)
{
	pxHandler->FindConnectedHandlers();
}
#endif //ASURA_USE_STREAMING

#ifdef ASURA_MAIN
void Asura_Module_System::ClearRenderAndLinkedInListsForCurCam()
{
	#ifdef ASURA_USE_STREAMING
	s_xHandlerTree.Traverse(ClearRenderAndLinkedInListsForCurCamCallback);
	#else
	GetCurrentHandler().ClearRenderAndLinkedInListsForCurCam();
	#endif
}
#endif // ASURA_MAIN

#ifdef ASURA_USE_STREAMING
void Asura_Module_System::ClearRenderAndLinkedInListsForCurCamCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler)
{
	pxHandler->ClearRenderAndLinkedInListsForCurCam();
}
#endif //ASURA_USE_STREAMING

#ifdef ASURA_MAIN
void Asura_Module_System::AddAllModulesToRenderListForCurCam()
{
	#ifdef ASURA_USE_STREAMING
	s_xHandlerTree.Traverse(AddAllModulesToRenderListForCurCamCallback);
	#else
	GetCurrentHandler().AddAllModulesToRenderListForCurCam();
	#endif
}

void Asura_Module_System::AddAllModulesToLinkedInListForCurCam()
{
	#ifdef ASURA_USE_STREAMING
	s_xHandlerTree.Traverse(AddAllModulesToLinkedInListForCurCamCallback);
	#else
	GetCurrentHandler().AddAllModulesToLinkedInListForCurCam();
	#endif
}

void Asura_Module_System::AddAllModulesWithoutModLinkingToRenderListForCurCam()
{
#ifdef ASURA_USE_STREAMING
	s_xHandlerTree.Traverse(AddAllModulesWithoutModLinkingToRenderListForCurCamCallback);
#else
	GetCurrentHandler().AddAllModulesWithoutModLinkingToRenderListForCurCam();
#endif
}
#endif // ASURA_MAIN

#ifdef ASURA_USE_STREAMING
void Asura_Module_System::AddAllModulesToRenderListForCurCamCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler)
{
	pxHandler->AddAllModulesToRenderListForCurCam();
}

void Asura_Module_System::AddAllModulesToLinkedInListForCurCamCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler)
{
	pxHandler->AddAllModulesToLinkedInListForCurCam();
}


void Asura_Module_System::AddAllModulesWithoutModLinkingToRenderListForCurCamCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler)
{
	pxHandler->AddAllModulesWithoutModLinkingToRenderListForCurCam();
}
#endif //ASURA_USE_STREAMING

#ifdef ASURA_MAIN
void Asura_Module_System::CullRenderListForCurCam()
{
	#ifdef ASURA_USE_STREAMING
	s_xHandlerTree.Traverse(CullRenderListForCurCamCallback);
	#else
	GetCurrentHandler().CullRenderListForCurCam();
	#endif
}
#endif // ASURA_MAIN

#ifdef ASURA_USE_STREAMING
void Asura_Module_System::CullRenderListForCurCamCallback(Asura_Hash_ID uHandlerHash,Asura_Module_Handler* pxHandler)
{
	pxHandler->CullRenderListForCurCam();
}
#endif //ASURA_USE_STREAMING

//////////////////////////////////
// DoesBBIntersectVisibleModule //
//////////////////////////////////
// MGH 26/5/2006: When the zone is unknown or may change then this is a better bet than the
// other versions of this function. Check the last known module/zone first otherwise do a full test.
#ifdef ASURA_USE_STREAMING
struct Asura_BBIntersectVisibleModule_UserData
{
	Asura_BBIntersectVisibleModule_UserData()
		: m_pxBB(0), m_piLastKnownModule(0), m_puLastKnownZone(0), m_bResult(false), m_bFoundExteriorModule(false)
	{}

	const Asura_Bounding_Box* m_pxBB;
	int* m_piLastKnownModule;
	Asura_Hash_ID* m_puLastKnownZone;
	bool m_bResult;
	bool m_bFoundExteriorModule;
};

struct Asura_BBIntersectNonExteriorModule_UserData
{
	Asura_BBIntersectNonExteriorModule_UserData()
		: m_pxBB(0), m_bResult(false)
	{}

	const Asura_Bounding_Box* m_pxBB;
	bool m_bResult;
};


bool Asura_Module_System::DoesBBIntersectVisibleModule(const Asura_Bounding_Box& xBB, int* piLastKnownModule, Asura_Hash_ID* puLastKnownZone)
{
	if (piLastKnownModule && *piLastKnownModule != -1 &&
		puLastKnownZone   && *puLastKnownZone != ASURA_HASH_ID_UNSET)
	{
		// Check the specified module and zone first of all.
		const Asura_Module_Handler* pxHandler = s_pxCurrentHandler;
		if (!pxHandler || pxHandler->GetHandlerHashID() != *puLastKnownZone)
		{
			pxHandler = GetHandler(*puLastKnownZone);
		}
		if (pxHandler)
		{
			const int iModule = *piLastKnownModule;
			if (0 <= iModule && iModule < pxHandler->GetNumberOfModules())
			{
				//Make sure the last known module wasn't an exterior module.
				//If it was an exterior module then knowing whether the point is still within it won't help ,since
				//we still need to check for being in non-exterior modules.
				//RAR 10/04/2007
				if(!pxHandler->IsExteriorModule(iModule))
				{
					if (pxHandler->IsModuleVisible(iModule))
					{
						const Asura_Module* pxModuleStorage = pxHandler->GetModuleData(iModule);
						if (pxModuleStorage && pxModuleStorage->m_pxCollisionMesh)
						{
							if (xBB.Intersects(pxModuleStorage->m_pxCollisionMesh->GetBoundingBox(), pxModuleStorage->Position))
							{
								// The last known module is still visible.
								return true;
							}
						}
					}
				}
			}
		}
	}

	// Do a full lookup.
	Asura_BBIntersectVisibleModule_UserData xUserData;
	xUserData.m_pxBB              = &xBB;
	xUserData.m_piLastKnownModule = piLastKnownModule;
	xUserData.m_puLastKnownZone   = puLastKnownZone;

	s_xHandlerTree.Traverse(DoesBBIntersectVisibleModule_Callback, &xUserData);

	if(xUserData.m_bResult && xUserData.m_bFoundExteriorModule)
	{
		//The only visible module found was exterior.
		//Need to do another check to make sure the box didn't intersect any non-exterior modules (visible or otherwise)
		Asura_BBIntersectNonExteriorModule_UserData xNonExteriorUserData;
		xNonExteriorUserData.m_pxBB = &xBB;
		
		s_xHandlerTree.Traverse(DoesBBIntersectNonExteriorModule_Callback, &xNonExteriorUserData);
		if(xNonExteriorUserData.m_bResult)
		{
			//did intersect a non exterior module , so don't accept the results of the previous test
			return false;
		}
	}

	return xUserData.m_bResult;
}
#endif //ASURA_USE_STREAMING

#ifdef ASURA_USE_STREAMING
// Return false to stop traversal.
bool Asura_Module_System::DoesBBIntersectVisibleModule_Callback(Asura_Hash_ID uHandlerHash, Asura_Module_Handler* pxHandler, void* pUserData)
{
	Asura_BBIntersectVisibleModule_UserData* pxUserData = static_cast<Asura_BBIntersectVisibleModule_UserData*>(pUserData);
	if (!pxUserData || !pxHandler) return false; // Give up.

	int iLastKnownModule = -1;
	if (pxHandler->DoesBBIntersectVisibleModule(*pxUserData->m_pxBB, &iLastKnownModule))
	{
		pxUserData->m_bResult = true;
		*pxUserData->m_piLastKnownModule = iLastKnownModule;
		*pxUserData->m_puLastKnownZone   = uHandlerHash;
		pxUserData->m_bFoundExteriorModule = pxHandler->IsExteriorModule(iLastKnownModule);

		// Found an intersecting visible module. Stop the traversal , unless it was exterior module , 
		// in which case we need to continue the search for non-exterior modules.
		// RAR 10/04/2007

		return pxUserData->m_bFoundExteriorModule;
	}
	
	// Carry on with traversal. Try the next zone.
	return true;
}
#endif //ASURA_USE_STREAMING

#ifdef ASURA_USE_STREAMING
// Return false to stop traversal.
bool Asura_Module_System::DoesBBIntersectNonExteriorModule_Callback(Asura_Hash_ID uHandlerHash, Asura_Module_Handler* pxHandler, void* pUserData)
{
	Asura_BBIntersectNonExteriorModule_UserData* pxUserData = static_cast<Asura_BBIntersectNonExteriorModule_UserData*>(pUserData);
	if (!pxUserData || !pxHandler) return false; // Give up.

	if (pxHandler->DoesBBIntersectNonExteriorModule(*pxUserData->m_pxBB))
	{
		pxUserData->m_bResult = true;

		//No need to continue searching once a non-exterior module has been found.		
		return false;
	}
	
	// Carry on with traversal. Try the next zone.
	return true;
}
#endif //ASURA_USE_STREAMING


/// End of file. ///
