///////////////////////////////////////////////////////////////////////////
//
// Asura_Primitive_Render.cpp
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if ASURA_USE_PRIMITIVE_RENDER == 1

// Default is a 32k buffer - this used to be 1MB!
#ifndef ASURA_PRIMITIVE_RENDER_MEMORY_SIZE
#define ASURA_PRIMITIVE_RENDER_MEMORY_SIZE					(32*1024)
#endif // #ifndef ASURA_PRIMITIVE_RENDER_MEMORY_SIZE

#define ASURA_PRIMITIVE_RENDER_STATIC_MEMORY_SIZE			(4*1024)

#define CONSOLE_COMMAND_PREFIX		("Debug.PrimitiveRender.")

// This classes header file
#include "Rendering/Asura_Primitive_Render.h"

// Asura files
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_StringBuffer.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_DebugOverlay_Alphabet.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Frustum.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

////////////////////////////////////////////////////////////////////////////////////////////////
//
// Put internal data and functions into a namespace to keep the interface clean in the header file (and not pollute the global namespace)
//
////////////////////////////////////////////////////////////////////////////////////////////////

namespace Asura_Primitive_Render_Data
{
	enum ASURA_DO_COMMAND
	{
		ASURA_DO_SPHERE = 0,
		ASURA_DO_LINE ,
		ASURA_DO_ARROW,
		ASURA_DO_CROSS,
		ASURA_DO_OBB,
		ASURA_DO_AXIS,
		ASURA_DO_BOUNDING_BOX,
		ASURA_DO_FRUSTUM,
		ASURA_DO_CYLINDER,
		ASURA_DO_CAPSULE,
		ASURA_DO_CONSOLE_STRING,
		ASURA_DO_DEBUG_STRING,
		ASURA_DO_NUM_PRIMS
	};
	const int ASURA_DO_HEADER_SIZE = sizeof(u_int) + sizeof(u_int) + sizeof(u_int);
	const int ASURA_DO_FLOAT_SIZE = sizeof(float);
	const int ASURA_DO_VECTOR_SIZE = 3 * ASURA_DO_FLOAT_SIZE;
	const int ASURA_DO_QUAT_SIZE = 4 * ASURA_DO_FLOAT_SIZE;
	const int ASURA_DO_UINT_SIZE = sizeof(u_int);
	const int ASURA_DO_ARGB_SIZE = sizeof(u_int);

	const int ASURA_DO_COMMAND_BIT_MASK = 0xFFFF;
	const int ASURA_DO_WORLD_SPACE_BIT_MASK = (1 << 16);
	const int ASURA_DO_IGNORE_Z_BUFFER_BIT_MASK = (1 << 17);

	static u_char* s_pucStaticListStart = NULL;
	static u_char* s_pucStaticListEnd = NULL;
	static u_char* s_pucStaticListNext = NULL;

	static u_char* s_pucListStart = NULL;
	static u_char* s_pucListEnd = NULL;
	static u_char* s_pucListNext = NULL;
	static u_int s_uBufferUsedLastFrame = 0;
	static u_int s_uPeakBufferUsed = 0;
	static u_int s_auPrimMemSize[ASURA_DO_NUM_PRIMS] = 
	{
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_ARGB_SIZE + ASURA_DO_UINT_SIZE + ASURA_DO_UINT_SIZE,		// sphere
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,												// line
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_FLOAT_SIZE,	// arrow
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE + ASURA_DO_FLOAT_SIZE,													// cross
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,							// OBB
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_FLOAT_SIZE,							// axis
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_ARGB_SIZE,	// bounding box
		ASURA_DO_HEADER_SIZE + ASURA_DO_FLOAT_SIZE * 4 + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_ARGB_SIZE,						// frustum
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_FLOAT_SIZE + ASURA_DO_UINT_SIZE + ASURA_DO_ARGB_SIZE,	// cylinder
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_QUAT_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,							// capsule
		ASURA_DO_HEADER_SIZE + ASURA_DO_ARGB_SIZE,																								// console string - not a fixed size
		ASURA_DO_HEADER_SIZE + ASURA_DO_VECTOR_SIZE + ASURA_DO_ARGB_SIZE,																		// debug string - not a fixed size
	};

	static Asura_Vector_2 s_xConsoleStartPos(0.05f, 0.21f);
	static Asura_Vector_2 s_xConsoleLineSize(0.0f, 0.025f);
	static Asura_Vector_2 s_xConsoleCurrentPos(0.0f, 0.0f);

	static u_int s_uDefaultConsoleColour;

	bool Initialise(const int iMemorySize); 
	bool Shutdown();
	void Reset(); 
	void Render();
	void Update();

	bool CanAdd(const enum ASURA_DO_COMMAND eCommand, bool bAddToStaticData);
	bool CanAdd(u_int uNumBytes, bool bAddToStaticData);
	bool AddData(const void* const pData, const size_t uNumBytes, bool bAddToStaticData);
	bool AddFloat(const float fValue, bool bAddToStaticData);
	bool AddUint(const u_int uValue, bool bAddToStaticData);
	bool AddVector(const Asura_Vector_3& xVector, bool bAddToStaticData);
	bool AddQuat(const Asura_Quat& xQuat, bool bAddToStaticData);
	bool AddRGBA(const Asura_RGBA& xColour, bool bAddToStaticData);
	bool AddARGB(const u_int uARGB, bool bAddToStaticData);
	bool AddString(const char* const szString, const u_int uNumBytes, bool bAddToStaticData);

	// For simple fixed sized primitives
	bool AddHeader(const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer, float fTime, bool bAddToStaticData);
	// For primitives with variable size e.g. text based ones
	bool AddHeader(const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes, float fTime, bool bAddToStaticData);

	bool ReadData(void* const pData, const size_t uNumBytes, bool bReadFromStaticData);
	bool ReadFloat(float& fValue, bool bReadFromStaticData = false);
	bool ReadUint(u_int& uValue, bool bReadFromStaticData = false);
	bool ReadVector(Asura_Vector_3& xVector, bool bReadFromStaticData = false);
	bool ReadQuat(Asura_Quat& xQuat, bool bReadFromStaticData = false);
	bool ReadRGBA(Asura_RGBA& xColour, bool bReadFromStaticData = false);
	bool ReadARGB(u_int& uARGB, bool bReadFromStaticData = false);
	bool ReadString(const char*& szString, const u_int uNumBytes, bool bReadFromStaticData = false);

	bool ReadHeader(u_int& uNumBytes, enum ASURA_DO_COMMAND& eCommand, bool& bWorldSpace, bool& bIgnoreZBuffer, u_int& uEndTime, bool bReadFromStaticData);

	void ResetCommandList();
	void RenderCommandList(const u_char* const pucListEnd, const bool bIgnoreZBuffer);

	// Render functions for each command
	bool RenderSphere(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderFlatLine(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderArrow(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderCross(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderOBB(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderAxis(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderBoundingBox(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderFrustum(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderCylinder(const bool bWorldSpace, const bool bIgnoreZBuffer);
	bool RenderCapsule(const bool bWorldSpace, const bool bIgnoreZBuffer);

	bool RenderDebugString(const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes);
	bool RenderConsoleString(const u_int uNumBytes);

	void RenderStringScreenInternal(const char* const szString, const u_int uARGB, const Asura_Vector_2& xPos);
	void RenderStringWorldInternal(const char* const szString, const u_int uARGB, const Asura_Vector_3& xPos);

	void RenderStringInternal(const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const char* const szString, const u_int uARGB);
	void RenderLineStripInternal(const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const float* const pfCoords, const u_int uNumPoints, const u_int uARGB);
};

////////////////////////////////////////////////////////////////////////////////////////////////
//
// The functions themselves
//
////////////////////////////////////////////////////////////////////////////////////////////////

// Initialise internals - allocate memory buffer
bool Asura_Primitive_Render_Data::Initialise(const int iMemorySize)
{
	s_pucListEnd = NULL;
	s_pucListNext = NULL;
	s_uBufferUsedLastFrame = 0;
	s_uPeakBufferUsed = 0;
	s_pucListStart = new u_char[iMemorySize];
	if (s_pucListStart == NULL)
	{
		return false;
	}
	s_pucStaticListStart = new u_char[ASURA_PRIMITIVE_RENDER_STATIC_MEMORY_SIZE];
	if(s_pucStaticListStart == NULL)
	{
		delete [] s_pucListStart;
		s_pucListStart = NULL;
		return false;
	}
	s_pucListEnd = s_pucListStart + iMemorySize;
	s_pucListNext = s_pucListStart;
	s_pucStaticListEnd = s_pucStaticListStart + ASURA_PRIMITIVE_RENDER_STATIC_MEMORY_SIZE;
	s_pucStaticListNext = s_pucStaticListStart;

	const Asura_RGBA xDefaultConsoleColour(150.0f/255.0f, 1.0f, 180.0f/255.0f, 1.0f);
	s_uDefaultConsoleColour = xDefaultConsoleColour.ToUIntARGB();

	return true;
}

// Free memory
bool Asura_Primitive_Render_Data::Shutdown()
{
	delete[] s_pucListStart;
	delete[] s_pucStaticListStart;
	s_pucListStart = NULL;
	s_pucListEnd = NULL;
	s_pucListNext = NULL;
	s_pucStaticListStart = NULL;
	s_pucStaticListEnd = NULL;
	s_pucStaticListNext = NULL;
	return true;
}
	
void Asura_Primitive_Render_Data::ResetCommandList()
{
	// store current usage, reset the list back to its beginning and the console render pos to the start
	s_uBufferUsedLastFrame = s_pucListNext - s_pucListStart;
	s_uPeakBufferUsed = Asura_Maths::Max(s_uBufferUsedLastFrame, s_uPeakBufferUsed);

	s_pucListNext = s_pucListStart;
	s_xConsoleCurrentPos = s_xConsoleStartPos;
}

// Reset internal state
void Asura_Primitive_Render_Data::Reset()
{
	ResetCommandList();
}

void Asura_Primitive_Render_Data::RenderCommandList(const u_char* const pucListEnd, const bool bDrawIgnoreZBuffer)
{
	while(s_pucListNext < pucListEnd)
	{
		// Read the next header from the list
		u_int uNumBytes;
		ASURA_DO_COMMAND eCommand;
		bool bWorldSpace;
		bool bIgnoreZBuffer;
		u_int uEndTime;
		if (ReadHeader(uNumBytes, eCommand, bWorldSpace, bIgnoreZBuffer, uEndTime, false) == false)
		{
			Asura_Assert(false, "RenderCommandList ReadHeader failed");
			return;
		}

		if (bDrawIgnoreZBuffer != bIgnoreZBuffer)
		{
			// We have already read the header
			s_pucListNext += (uNumBytes - ASURA_DO_HEADER_SIZE);
			continue;
		}
		bool bOK = false;

		switch (eCommand)
		{
			case ASURA_DO_SPHERE:
			{
				bOK = RenderSphere(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_LINE:
			{
				bOK = RenderFlatLine(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_ARROW:
			{
				bOK = RenderArrow(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_CROSS:
			{
				bOK = RenderCross(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_OBB:
			{
				bOK = RenderOBB(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_AXIS:
			{
				bOK = RenderAxis(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_BOUNDING_BOX:
			{
				bOK = RenderBoundingBox(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_FRUSTUM:
			{
				bOK = RenderFrustum(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_CYLINDER:
			{
				bOK = RenderCylinder(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_CAPSULE:
			{
				bOK = RenderCapsule(bWorldSpace, bIgnoreZBuffer);
				break;
			}
			case ASURA_DO_CONSOLE_STRING:
			{
				bOK = RenderConsoleString(uNumBytes);
				break;
			}
			case ASURA_DO_DEBUG_STRING:
			{	
				bOK = RenderDebugString(bWorldSpace, bIgnoreZBuffer, uNumBytes);
				break;
			}
#if 0
			case GOURAUD_LINE:
			{
				RenderGouraudLine(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case TRIANGLE:
			{
				RenderTriangle(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case QUAD:
			{
				RenderQuad(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case CIRCLE:
			{
				RenderCircle(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case SPHERE_MATRIX:
			{
				RenderSphereMatrix(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case CYLINDER:
			{
				RenderCylinder(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case SAUSAGE:
			{
				RenderSausage(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case CONE:
			{
				RenderCone(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case CONE_MATRIX:
			{
				RenderConeMatrix(oColour, eDrawType, io_graphicsDevice);
				break;
			}
			case CUBE:
			{
				RenderCube(oColour, eDrawType, io_graphicsDevice);
				break;
			}
#endif // #if 0
			default:
			{
				Asura_Assertf(false, ("RenderCommandList unknown eCommand %d\n", eCommand));
				break;
			}
		}
		if (bOK == false)
		{
			// If the render fails stop reading the command list
			// Yuck having to use a const_cast!
			s_pucListNext = const_cast<u_char*>(pucListEnd);
		}
	}
}

void Asura_Primitive_Render_Data::Update()
{
	// Update the static list

	u_char* pucEnd = s_pucStaticListNext;

	s_pucStaticListNext = s_pucStaticListStart;

	while (s_pucStaticListNext < pucEnd)
	{
		// Read the header

		u_int uNumBytes;
		ASURA_DO_COMMAND eCommand;
		bool bWorkspace;
		bool bIgnoreZBuffer;
		u_int uEndTime;

		u_char* pucPrevPos = s_pucStaticListNext;

		if(!ReadHeader(uNumBytes, eCommand, bWorkspace, bIgnoreZBuffer, uEndTime, true))
		{
			Asura_Assert(false, "ReadHeader failed reading command size");
			return;
		}

		s_pucStaticListNext = pucPrevPos;

		if(Asura_Timers::GetMSecsSinceStart() > uEndTime)
		{
			// Kill this timed object by removing this data from the static buffer

			u_int uBytesToShift = pucEnd - (s_pucStaticListNext + uNumBytes);

			if(uBytesToShift > 0)
			{
				Asura_Mem::Move(s_pucStaticListNext, s_pucStaticListNext + uNumBytes, uBytesToShift);
			}

			pucEnd -= uNumBytes;
		}
		else
		{
			// Copy this command to the main command buffer

			if(!AddData(s_pucStaticListNext, uNumBytes, false))
			{
				Asura_Assert(false, "AddData failed copying data from static command list to main command list");
			}

			s_pucStaticListNext += uNumBytes;
		}
	}
}

// JAT 21/10/07
// What should we do about split-screens?
// In CoreEngine we rendered the 3D debug data once per screen and the 2D data in full-screen part of the render loop

// Render the debug command list and info
void Asura_Primitive_Render_Data::Render()
{
	s_xConsoleCurrentPos = s_xConsoleStartPos;
	const float fFontHeight = Asura_Primitive_Render::GetUseVectorFont() ? 
		Asura_Primitive_Render::GetVectorFontWidth() / Asura_Primitive_Render::GetVectorFontWHRatio() : 
		ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
	s_xConsoleLineSize.y = fFontHeight / static_cast<float>(Asura_Viewport::GetHeight());

	const Asura_Vector_2 xConsoleStartPos = s_xConsoleCurrentPos;

	u_char* pucListEnd = s_pucListNext;

	// Need to decode the loop multiple times so as to draw things in the correct order

	// Render the stuff which needs z-buffering
	s_xConsoleCurrentPos = xConsoleStartPos;
	s_pucListNext = s_pucListStart;

	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	RenderCommandList(pucListEnd, false);

	// Render the stuff which doesn't need z-buffering
	s_xConsoleCurrentPos = xConsoleStartPos;
	s_pucListNext = s_pucListStart;

	Asura_Render::CurrentStates.RequestZBufferEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	RenderCommandList(pucListEnd, true);

	// Put back z-test and z-write
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////
bool Asura_Primitive_Render_Data::RenderSphere(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xCentre;
	if (ReadVector(xCentre) == false)
	{
		Asura_Assert(false, "RenderSphere failed to read centre vector");
		return false;
	}
	float fRadius;
	if (ReadFloat(fRadius) == false)
	{
		Asura_Assert(false, "RenderSphere failed to read radius value");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderSphere failed to read ARGB value");
		return false;
	}
	u_int uNumSegments;
	if (ReadUint(uNumSegments) == false)
	{
		Asura_Assert(false, "RenderSphere failed to read num segments value");
		return false;
	}
	u_int uNumSlices;
	if (ReadUint(uNumSlices) == false)
	{
		Asura_Assert(false, "RenderSphere failed to read num slices value");
		return false;
	}
	if ((uNumSegments == 0) || (uNumSlices == 0))
	{
		return true;
	}

	// Add the slices to be rendered...
	// Don't do slice 0 or slice = uNumSlices because their radius is 0
	for (u_int uSlice = 1; uSlice < uNumSlices; uSlice++)
	{
		const Asura_Angle xSliceAngle = (Asura_Maths::Deg180 * static_cast<float>(uSlice)) / uNumSlices;
		const float	fSliceRadius = fRadius * Asura_Maths::Sine(xSliceAngle);
		const float	fY = xCentre.y + (fRadius * Asura_Maths::Cosine(xSliceAngle));

		Asura_Vector_3 xLastVertex;
		for (u_int uSegment = 0; uSegment <= uNumSegments; uSegment++)
		{
			const Asura_Angle xAngle = (Asura_Maths::Deg360 * static_cast<float>(uSegment)) / uNumSegments;
			const float	fX = xCentre.x + (fSliceRadius * Asura_Maths::Sine(xAngle));
			const float	fZ = xCentre.z + (fSliceRadius * Asura_Maths::Cosine(xAngle));
			Asura_Vector_3 xThisVertex(fX, fY, fZ);

			if (uSegment > 0)
			{
				Asura_Render::RenderLine(xLastVertex, xThisVertex, uARGB);
			}
			xLastVertex = xThisVertex;
		}
	}

	// Add the segments to be rendered...
	for (u_int uSegment = 0; uSegment < uNumSegments; uSegment++)
	{
		Asura_Vector_3 xLastVertex;
		for (u_int uSlice = 0; uSlice <= uNumSlices; uSlice++)
		{
			const Asura_Angle xSliceAngle = (Asura_Maths::Deg180 * static_cast<float>(uSlice)) / uNumSlices;
			const float	fSliceRadius = fRadius * Asura_Maths::Sine(xSliceAngle);
			const Asura_Angle xAngle = (Asura_Maths::Deg360 * static_cast<float>(uSegment)) / uNumSegments;

			const float	fX = xCentre.x + (fSliceRadius * Asura_Maths::Sine(xAngle));
			const float	fY = xCentre.y + (fRadius * Asura_Maths::Cosine(xSliceAngle));
			const float	fZ = xCentre.z + (fSliceRadius * Asura_Maths::Cosine(xAngle));
			Asura_Vector_3 xThisVertex(fX, fY, fZ);

			if (uSlice > 0)
			{
				Asura_Render::RenderLine(xLastVertex, xThisVertex, uARGB);
			}
			xLastVertex = xThisVertex;
		}
	}

	return true;
}

bool Asura_Primitive_Render_Data::RenderFlatLine(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xStart;
	if (ReadVector(xStart) == false)
	{
		Asura_Assert(false, "RenderLine failed to read start vector");
		return false;
	}
	Asura_Vector_3 xEnd;
	if (ReadVector(xEnd) == false)
	{
		Asura_Assert(false, "RenderLine failed to read end vector");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderLine failed to read ARGB value");
		return false;
	}
	Asura_Render::RenderLine(xStart, xEnd, uARGB);

	return true;
}

bool Asura_Primitive_Render_Data::RenderArrow(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xStart;
	if (ReadVector(xStart) == false)
	{
		Asura_Assert(false, "RenderArrow failed to read start vector");
		return false;
	}
	Asura_Vector_3 xEnd;
	if (ReadVector(xEnd) == false)
	{
		Asura_Assert(false, "RenderArrow failed to read end vector");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderArrow failed to read ARGB value");
		return false;
	}
	float fHeadProportion;
	if (ReadFloat(fHeadProportion) == false)
	{
		Asura_Assert(false, "RenderArrow failed to read head proportion value");
		return false;
	}
	float fCrossSize;
	if (ReadFloat(fCrossSize) == false)
	{
		Asura_Assert(false, "RenderArrow failed to read cross size value");
		return false;
	}

	Asura_Vector_3 xDirection(xEnd - xStart);
	const float fLength	= xDirection.Normalise();

	Asura_Vector_3 xHeadStart;
	Asura_Vector_3 xXAxis;
	Asura_Vector_3 xYAxis;

	// Calculate which axis our arrow-head is on...
	xHeadStart.Lerp(xEnd, xStart, fHeadProportion);

	if (Asura_Maths::Abs(xDirection.GetY()) > 0.707f)
	{
		xXAxis	= xDirection ^ Asura_Vector_3(0.0f, 0.0f, 1.0f);
		xYAxis	= xXAxis ^ xDirection;
	}
	else
	{
		xXAxis	= xDirection ^ Asura_Vector_3(0.0f, 1.0f, 0.0f);
		xYAxis	= xXAxis ^ xDirection;
	}

	// Now add the lines to be rendered...
	if (fCrossSize > 0.0f)
	{
		Asura_Render::RenderLine(xStart - xDirection * fCrossSize, xEnd, uARGB);

		const Asura_Vector_3 xXHalfCross = xXAxis * fCrossSize;
		Asura_Render::RenderLine(xStart - xXHalfCross, xStart + xXHalfCross, uARGB);
		const Asura_Vector_3 xYHalfCross = xYAxis * fCrossSize;
		Asura_Render::RenderLine(xStart - xYHalfCross, xStart + xYHalfCross, uARGB);
	}
	else
	{
		Asura_Render::RenderLine(xStart, xEnd, uARGB);
	}

	const float fHeadSize = fLength * fHeadProportion;
	const Asura_Vector_3 xXHalfHead = xXAxis * fHeadSize;
	Asura_Render::RenderLine(xHeadStart + xXHalfHead, xEnd, uARGB);
	Asura_Render::RenderLine(xHeadStart - xXHalfHead, xEnd, uARGB);

	const Asura_Vector_3 xYHalfHead = xXAxis * fHeadSize;
	Asura_Render::RenderLine(xHeadStart + xYHalfHead, xEnd, uARGB);
	Asura_Render::RenderLine(xHeadStart - xYHalfHead, xEnd, uARGB);

	return true;
}

bool Asura_Primitive_Render_Data::RenderCross(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xPosition;
	if (ReadVector(xPosition) == false)
	{
		Asura_Assert(false, "RenderCross failed to read position vector");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderCross failed to read ARGB value");
		return false;
	}
	float fSize;
	if (ReadFloat(fSize) == false)
	{
		Asura_Assert(false, "RenderCross failed to read size value");
		return false;
	}
	const Asura_Vector_3 xXAxis(fSize, 0.0f,  0.0f);
	const Asura_Vector_3 xYAxis(0.0f,  fSize, 0.0f);
	const Asura_Vector_3 xZAxis(0.0f,  0.0f,  fSize);

	// Now add the lines to be rendered...
	Asura_Render::RenderLine(Asura_Vector_3(xPosition + xXAxis), Asura_Vector_3(xPosition - xXAxis), uARGB);
	Asura_Render::RenderLine(Asura_Vector_3(xPosition + xYAxis), Asura_Vector_3(xPosition - xYAxis), uARGB);
	Asura_Render::RenderLine(Asura_Vector_3(xPosition + xZAxis), Asura_Vector_3(xPosition - xZAxis), uARGB);

	return true;
}

bool Asura_Primitive_Render_Data::RenderOBB(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xCentre;
	if (ReadVector(xCentre) == false)
	{
		Asura_Assert(false, "RenderOBB failed to read centre vector");
		return false;
	}
	Asura_Quat xQuat;
	if (ReadQuat(xQuat) == false)
	{
		Asura_Assert(false, "RenderOBB failed to read orient quaternion");
		return false;
	}
	Asura_Vector_3 xHalfExtents;
	if (ReadVector(xHalfExtents) == false)
	{
		Asura_Assert(false, "RenderOBB failed to read half extents vector");
		return false;
	}

	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderOBB failed to read ARGB value");
		return false;
	}

	const float	fX = xHalfExtents.x;
	const float	fY = xHalfExtents.y;
	const float	fZ = xHalfExtents.z;
	Asura_AMatrix3x3 xOrientation;
	xQuat.ToMatrix(xOrientation);
	const Asura_AVector3 xCentreA(xCentre);

	Asura_AVector3 axVertices[8];

	// Generate the vertices...
	for(u_int u = 0; u < 8; u++)
	{
		const Asura_AVector3 xVertex((((u & 1) == 0) ? fX : -fX), (((u & 2) == 0) ? fY : -fY), (((u & 4) == 0) ? fZ : -fZ));
		axVertices[u] = (xVertex * xOrientation) + xCentreA;
	}

	// Now add the lines to be rendered...
	Asura_Render::RenderLine(axVertices [0], axVertices [1], uARGB);
	Asura_Render::RenderLine(axVertices [1], axVertices [3], uARGB);
	Asura_Render::RenderLine(axVertices [3], axVertices [2], uARGB);
	Asura_Render::RenderLine(axVertices [2], axVertices [0], uARGB);

	Asura_Render::RenderLine(axVertices [4], axVertices [5], uARGB);
	Asura_Render::RenderLine(axVertices [5], axVertices [7], uARGB);
	Asura_Render::RenderLine(axVertices [7], axVertices [6], uARGB);
	Asura_Render::RenderLine(axVertices [6], axVertices [4], uARGB);

	Asura_Render::RenderLine(axVertices [0], axVertices [4], uARGB);
	Asura_Render::RenderLine(axVertices [1], axVertices [5], uARGB);
	Asura_Render::RenderLine(axVertices [2], axVertices [6], uARGB);
	Asura_Render::RenderLine(axVertices [3], axVertices [7], uARGB);

	return true;
}

bool Asura_Primitive_Render_Data::RenderAxis(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xPosition;
	if (ReadVector(xPosition) == false)
	{
		Asura_Assert(false, "RenderAxis failed to read position vector");
		return false;
	}
	Asura_Quat xQuat;
	if (ReadQuat(xQuat) == false)
	{
		Asura_Assert(false, "RenderAxis failed to read orient quaternion");
		return false;
	}
	float fSize;
	if (ReadFloat(fSize) == false)
	{
		Asura_Assert(false, "RenderAxis failed to read size value");
		return false;
	}
	float fAlpha;
	if (ReadFloat(fAlpha) == false)
	{
		Asura_Assert(false, "RenderAxis failed to read alpha value");
		return false;
	}
	Asura_Matrix_3x3 xOrient;
	xQuat.ToMatrix(xOrient);
	Asura_Render::RenderAxes(xPosition, xOrient, fSize, fAlpha);

	return true;
}

bool Asura_Primitive_Render_Data::RenderBoundingBox(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xMin;
	if (ReadVector(xMin) == false)
	{
		Asura_Assert(false, "RenderBoundingBox failed to read min vector");
		return false;
	}
	Asura_Vector_3 xMax;
	if (ReadVector(xMax) == false)
	{
		Asura_Assert(false, "RenderBoundingBox failed to read max vector");
		return false;
	}
	Asura_Vector_3 xPosition;
	if (ReadVector(xPosition) == false)
	{
		Asura_Assert(false, "RenderBoundingBox failed to read position vector");
		return false;
	}
	Asura_Quat xQuat;
	if (ReadQuat(xQuat) == false)
	{
		Asura_Assert(false, "RenderBoundingBox failed to read orient quaternion");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderBoundingBox failed to read ARGB value");
		return false;
	}

	const u_short ausIndices[24] =
	{
		0, 2, 2, 3, 3, 1, 1, 0,
		0, 4, 2, 6, 3, 7, 1, 5,
		4, 6, 6, 7, 7, 5, 5, 4,
	};

	Asura_Vector_3 axVertices[8];

	Asura_Bounding_Box xBB;
	xBB.MinX = xMin.x;
	xBB.MinY = xMin.y;
	xBB.MinZ = xMin.z;
	xBB.MaxX = xMax.x;
	xBB.MaxY = xMax.y;
	xBB.MaxZ = xMax.z;

	Asura_Matrix_3x3 xOrientation;
	xQuat.ToMatrix(xOrientation);

	// Calculate the vertices...
	for (u_int u = 0; u < 8; u++)
	{
		axVertices[u][0] = ((u & 1) == 0) ? xBB.MinX : xBB.MaxX;
		axVertices[u][1] = ((u & 2) == 0) ? xBB.MinY : xBB.MaxY;
		axVertices[u][2] = ((u & 4) == 0) ? xBB.MinZ : xBB.MaxZ;

		axVertices[u] = (axVertices[u] * xOrientation) + xPosition;
	}

	// Add them to be rendered...
	for (u_int u = 0; u < 24; u += 2)
	{
		const Asura_Vector_3& xStart = axVertices[ausIndices[u]];
		const Asura_Vector_3& xEnd = axVertices[ausIndices[u + 1]];

		Asura_Render::RenderLine(xStart, xEnd, uARGB);
	}

	return true;
}

bool Asura_Primitive_Render_Data::RenderFrustum(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	float fNearPlane;
	if (ReadFloat(fNearPlane) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read near plane value");
		return false;
	}
	float fFarPlane;
	if (ReadFloat(fFarPlane) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read far plane value");
		return false;
	}
	float fTanHorizontalFOV;
	if (ReadFloat(fTanHorizontalFOV) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read tan horizontal FOV");
		return false;
	}
	float fTanVerticalFOV;
	if (ReadFloat(fTanVerticalFOV) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read tan vertical FOV");
		return false;
	}
	Asura_Frustum xFrustum;
	xFrustum.SetForFOV(fNearPlane, fFarPlane, fTanHorizontalFOV, fTanVerticalFOV);

	Asura_Vector_3 xPosition;
	if (ReadVector(xPosition) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read position vector");
		return false;
	}
	Asura_Quat xQuat;
	if (ReadQuat(xQuat) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read orient quaternion");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderFrustum failed to read ARGB value");
		return false;
	}

	Asura_Vector_3 axVertices[Asura_Frustum::MAX_VERTICES];

	// Generate the vertices...
	xFrustum.GenerateVertices(axVertices, xPosition, xQuat);

	// Now add the lines to be rendered...
	Asura_Render::RenderLine(axVertices[0], axVertices[1], uARGB);
	Asura_Render::RenderLine(axVertices[1], axVertices[3], uARGB);
	Asura_Render::RenderLine(axVertices[3], axVertices[2], uARGB);
	Asura_Render::RenderLine(axVertices[2], axVertices[0], uARGB);

	Asura_Render::RenderLine(axVertices[4], axVertices[5], uARGB);
	Asura_Render::RenderLine(axVertices[5], axVertices[7], uARGB);
	Asura_Render::RenderLine(axVertices[7], axVertices[6], uARGB);
	Asura_Render::RenderLine(axVertices[6], axVertices[4], uARGB);

	Asura_Render::RenderLine(axVertices[0], axVertices[4], uARGB);
	Asura_Render::RenderLine(axVertices[1], axVertices[5], uARGB);
	Asura_Render::RenderLine(axVertices[2], axVertices[6], uARGB);
	Asura_Render::RenderLine(axVertices[3], axVertices[7], uARGB);

	return true;
}

bool Asura_Primitive_Render_Data::RenderCylinder(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xPosition;
	if (ReadVector(xPosition) == false)
	{
		Asura_Assert(false, "RenderCylinder failed to read position vector");
		return false;
	}
	Asura_Quat xQuat;
	if (ReadQuat(xQuat) == false)
	{
		Asura_Assert(false, "RenderCylinder failed to read orient quaternion");
		return false;
	}
	float fHeight;
	if (ReadFloat(fHeight) == false)
	{
		Asura_Assert(false, "RenderCylinder failed to read height float");
		return false;
	}
	float fRadius;
	if (ReadFloat(fRadius) == false)
	{
		Asura_Assert(false, "RenderCylinder failed to read radius float");
		return false;
	}
	u_int uNumSegments = 16;
	if (ReadUint(uNumSegments) == false)
	{
		Asura_Assert(false, "RenderCylinder failed to read num segments u_int");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderCylinder failed to read ARGB value");
		return false;
	}

	const Asura_AVector3	xOldBottomVertex(0.0f, 0.0f, fRadius);
	const Asura_AVector3	xOldTopVertex(0.0f, fHeight, fRadius);

	Asura_AVector3			xTransformedOldBottomVertex;
	Asura_AVector3			xTransformedOldTopVertex;

	// Generate the initial vertices...
	xTransformedOldBottomVertex = xQuat.RotateVector(xOldBottomVertex);
	xTransformedOldTopVertex = xQuat.RotateVector(xOldTopVertex);

	xTransformedOldBottomVertex	+= Asura_AVector3(xPosition);
	xTransformedOldTopVertex	+= Asura_AVector3(xPosition);

	for (u_int uSegment = 1; uSegment <= uNumSegments; uSegment++)
	{
		const Asura_Angle		xAngle	= (Asura_Maths::Deg360 * static_cast<float>(uSegment)) / uNumSegments;
		const float				fX		= fRadius * Asura_Maths::Sine(xAngle);
		const float				fZ		= fRadius * Asura_Maths::Cosine(xAngle);

		const Asura_AVector3	xNewBottomVertex(fX, 0.0f, fZ);
		const Asura_AVector3	xNewTopVertex(fX, fHeight, fZ);

		Asura_AVector3			xTransformedNewBottomVertex;
		Asura_AVector3			xTransformedNewTopVertex;

		xTransformedNewBottomVertex = xQuat.RotateVector(xNewBottomVertex);
		xTransformedNewTopVertex = xQuat.RotateVector(xNewTopVertex);

		xTransformedNewBottomVertex	+= Asura_AVector3(xPosition);
		xTransformedNewTopVertex	+= Asura_AVector3(xPosition);

		Asura_Render::RenderLine(xTransformedOldBottomVertex, xTransformedNewBottomVertex, uARGB);
		Asura_Render::RenderLine(xTransformedNewBottomVertex, xTransformedNewTopVertex, uARGB);
		Asura_Render::RenderLine(xTransformedNewTopVertex, xTransformedOldTopVertex, uARGB);

		// Step onto the next vertex...
		xTransformedOldBottomVertex	= xTransformedNewBottomVertex;
		xTransformedOldTopVertex	= xTransformedNewTopVertex;
	}

	return true;
}

bool Asura_Primitive_Render_Data::RenderCapsule(const bool bWorldSpace, const bool bIgnoreZBuffer)
{
	Asura_Vector_3 xPosition;
	if (ReadVector(xPosition) == false)
	{
		Asura_Assert(false, "RenderCapsule failed to read position vector");
		return false;
	}
	Asura_Quat xQuat;
	if (ReadQuat(xQuat) == false)
	{
		Asura_Assert(false, "RenderCapsule failed to read orient quaternion");
		return false;
	}
	Asura_Vector_3 xExtents;
	if (ReadVector(xExtents) == false)
	{
		Asura_Assert(false, "RenderCapsule failed to read extents vector");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderCapsule failed to read ARGB value");
		return false;
	}

	const u_int uNumSlices = 4;

	const u_int uNumSegments = 8;
	const float fRadius = xExtents.x;
	const float fHalfCylinderLength = Asura_Maths::Max( 0.0f, xExtents.z - fRadius );

	const Asura_AVector3	xOldBottomVertex(0.0f, fRadius, -fHalfCylinderLength);
	const Asura_AVector3	xOldTopVertex(0.0f, fRadius, fHalfCylinderLength);

	Asura_AVector3			xTransformedOldBottomVertex;
	Asura_AVector3			xTransformedOldTopVertex;

	//
	// Draw the cylinder part
	//

	xTransformedOldBottomVertex = xQuat.RotateVector(xOldBottomVertex);
	xTransformedOldTopVertex = xQuat.RotateVector(xOldTopVertex);

	xTransformedOldBottomVertex	+= Asura_AVector3(xPosition);
	xTransformedOldTopVertex	+= Asura_AVector3(xPosition);

	for (u_int uSegment = 1; uSegment <= uNumSegments; uSegment++)
	{
		const Asura_Angle		xAngle	= (Asura_Maths::Deg360 * static_cast<float>(uSegment)) / uNumSegments;
		const float				fX		= fRadius * Asura_Maths::Sine(xAngle);
		const float				fY		= fRadius * Asura_Maths::Cosine(xAngle);

		const Asura_AVector3	xNewBottomVertex(fX, fY, -fHalfCylinderLength);
		const Asura_AVector3	xNewTopVertex(fX, fY, fHalfCylinderLength);

		Asura_AVector3			xTransformedNewBottomVertex;
		Asura_AVector3			xTransformedNewTopVertex;

		xTransformedNewBottomVertex = xQuat.RotateVector(xNewBottomVertex);
		xTransformedNewTopVertex = xQuat.RotateVector(xNewTopVertex);

		xTransformedNewBottomVertex	+= Asura_AVector3(xPosition);
		xTransformedNewTopVertex	+= Asura_AVector3(xPosition);

		Asura_Render::RenderLine(xTransformedOldBottomVertex, xTransformedNewBottomVertex, uARGB);
		Asura_Render::RenderLine(xTransformedNewBottomVertex, xTransformedNewTopVertex, uARGB);
		Asura_Render::RenderLine(xTransformedNewTopVertex, xTransformedOldTopVertex, uARGB);

		// Step onto the next vertex...
		xTransformedOldBottomVertex	= xTransformedNewBottomVertex;
		xTransformedOldTopVertex	= xTransformedNewTopVertex;
	}

	//
	// Draw the end hemisphere parts
	//

	// Add the slices to be rendered...
	// Don't do slice 0 or slice = uNumSlices because their radius is 0
	for (u_int uSlice = 1; uSlice < uNumSlices; uSlice++)
	{
		const Asura_Angle xSliceAngle = (Asura_Maths::Deg90 * static_cast<float>(uSlice)) / uNumSlices;
		const float	fSliceRadius = fRadius * Asura_Maths::Sine(xSliceAngle);
		const float	fZ = fHalfCylinderLength + (fRadius * Asura_Maths::Cosine(xSliceAngle));
		const float	fZ2 = (-fHalfCylinderLength) - (fRadius * Asura_Maths::Cosine(xSliceAngle));

		Asura_Vector_3 xLastVertex[2];
		for (u_int uSegment = 0; uSegment <= uNumSegments; uSegment++)
		{
			const Asura_Angle xAngle = (Asura_Maths::Deg360 * static_cast<float>(uSegment)) / uNumSegments;
			const float	fX = (fSliceRadius * Asura_Maths::Sine(xAngle));
			const float	fY = (fSliceRadius * Asura_Maths::Cosine(xAngle));

			for(u_int n = 0; n < 2; n++)
			{
				Asura_Vector_3 xThisVertex(fX, fY, n == 0 ? fZ : fZ2);

				if (uSegment > 0)
				{
					Asura_AVector3 xTransformed[2];

					xTransformed[0] = xQuat.RotateVector(xLastVertex[n]);
					xTransformed[1] = xQuat.RotateVector(xThisVertex);

					xTransformed[0] += Asura_AVector3(xPosition);
					xTransformed[1] += Asura_AVector3(xPosition);

					Asura_Render::RenderLine(xTransformed[0], xTransformed[1], uARGB);
				}
				xLastVertex[n] = xThisVertex;
			}
		}
	}

	// Add the segments to be rendered...
	for (u_int uSegment = 0; uSegment < uNumSegments; uSegment++)
	{
		Asura_Vector_3 xLastVertex[2];
		for (u_int uSlice = 0; uSlice <= uNumSlices; uSlice++)
		{
			const Asura_Angle xSliceAngle = (Asura_Maths::Deg90 * static_cast<float>(uSlice)) / uNumSlices;
			const float	fSliceRadius = fRadius * Asura_Maths::Sine(xSliceAngle);
			const Asura_Angle xAngle = (Asura_Maths::Deg360 * static_cast<float>(uSegment)) / uNumSegments;

			const float	fX = (fSliceRadius * Asura_Maths::Sine(xAngle));
			const float	fY = (fSliceRadius * Asura_Maths::Cosine(xAngle));

			for(u_int n = 0; n < 2; n++)
			{
				float fZ;
				
				if(n == 0)
				{
					fZ = fHalfCylinderLength + (fRadius * Asura_Maths::Cosine(xSliceAngle));
				}
				else
				{
					fZ = (-fHalfCylinderLength) - (fRadius * Asura_Maths::Cosine(xSliceAngle));
				}

				Asura_Vector_3 xThisVertex(fX, fY, fZ);

				if (uSlice > 0)
				{
					Asura_AVector3 xTransformed[2];

					xTransformed[0] = xQuat.RotateVector(xLastVertex[n]);
					xTransformed[1] = xQuat.RotateVector(xThisVertex);

					xTransformed[0] += Asura_AVector3(xPosition);
					xTransformed[1] += Asura_AVector3(xPosition);

					Asura_Render::RenderLine(xTransformed[0], xTransformed[1], uARGB);
				}
				xLastVertex[n] = xThisVertex;
			}
		}
	}

	return true;
}

bool Asura_Primitive_Render_Data::RenderConsoleString(const u_int uNumBytes)
{
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderConsoleString failed to read ARGB value");
		return false;
	}
	const u_int uStringLen = uNumBytes - s_auPrimMemSize[ASURA_DO_CONSOLE_STRING];
	const char* szString = NULL;
	if (ReadString(szString, uStringLen) == false)
	{
		Asura_Assert(false, "RenderConsoleString failed to read string data");
		return false;
	}

	RenderStringScreenInternal(szString, uARGB, s_xConsoleCurrentPos);
	s_xConsoleCurrentPos += s_xConsoleLineSize;

	return true;
}

bool Asura_Primitive_Render_Data::RenderDebugString(const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes)
{
	Asura_Vector_3 xPos3D;
	if (ReadVector(xPos3D) == false)
	{
		Asura_Assert(false, "RenderDebugString failed to read pos value");
		return false;
	}
	u_int uARGB;
	if (ReadARGB(uARGB) == false)
	{
		Asura_Assert(false, "RenderDebugString failed to read ARGB value");
		return false;
	}
	const u_int uStringLen = uNumBytes - s_auPrimMemSize[ASURA_DO_DEBUG_STRING];
	const char* szString = NULL;
	if (ReadString(szString, uStringLen) == false)
	{
		Asura_Assert(false, "RenderDebugString failed to read string data");
		return false;
	}

	if (bWorldSpace == false)
	{
		const Asura_Vector_2 xPos2D(xPos3D.x, xPos3D.y);
		RenderStringScreenInternal(szString, uARGB, xPos2D);
	}
	else
	{
		RenderStringWorldInternal(szString, uARGB, xPos3D);
	}

	return true;
}

// Render text using the vector font data stored in Asura_DebugOverlay_Alphabet.h
void Asura_Primitive_Render_Data::RenderStringScreenInternal(const char* const szString, const u_int uARGB, const Asura_Vector_2& xPos)
{
	if (szString == NULL)
	{
		return;
	}

	const float fFontWidth = Asura_Primitive_Render::GetUseVectorFont() ? 
		(Asura_Primitive_Render::GetVectorFontWidth() / Asura_Viewport::GetWidth()) : ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH;
	const float fFontHeight = Asura_Primitive_Render::GetUseVectorFont() ? 
		(fFontWidth / Asura_Primitive_Render::GetVectorFontWHRatio()) : ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;
	const Asura_Vector_2 xScale(fFontWidth, fFontHeight);

	RenderStringInternal(xPos, xScale, szString, uARGB);
}


// Render text using the vector font data stored in Asura_DebugOverlay_Alphabet.h
void Asura_Primitive_Render_Data::RenderStringWorldInternal(const char* const szString, const u_int uARGB, const Asura_Vector_3& xPos)
{
	if (szString == NULL)
	{
		return;
	}

	const float fFontWidth = Asura_Primitive_Render::GetUseVectorFont() ? 
		(Asura_Primitive_Render::GetVectorFontWidth() / Asura_Viewport::GetWidth()) : ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH;
	const float fFontHeight = Asura_Primitive_Render::GetUseVectorFont() ? 
		(fFontWidth / Asura_Primitive_Render::GetVectorFontWHRatio()) : ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT;

	// Made a world vector to represent the screen x-vector
	Asura_Vector_3 xScreenPos = xPos;
	Asura_Camera::ProjectWorldCoordsToScreenSpace(xScreenPos);
	if (xScreenPos.z <= 0.0f) // behind the camera
	{
		return;
	}

	const float fDelta = 1.0e-3f;

	const Asura_Vector_3 xWorldX = xPos + Asura_Vector_3(fDelta, 0.0f, 0.0f);
	Asura_Vector_3 xScreenX = xWorldX;
	Asura_Camera::ProjectWorldCoordsToScreenSpace(xScreenX);
	xScreenX -= xScreenPos;

	const Asura_Vector_3 xWorldY = xPos + Asura_Vector_3(0.0f, fDelta, 0.0f);
	Asura_Vector_3 xScreenY = xWorldY;
	Asura_Camera::ProjectWorldCoordsToScreenSpace(xScreenY);
	xScreenY -= xScreenPos;

	const Asura_Vector_3 xWorldZ = xPos + Asura_Vector_3(0.0f, 0.0f, fDelta);
	Asura_Vector_3 xScreenZ = xWorldZ;
	Asura_Camera::ProjectWorldCoordsToScreenSpace(xScreenZ);
	xScreenZ -= xScreenPos;

	const Asura_Vector_3 xScreenXdir(xScreenX.GetX(), xScreenY.GetX(), xScreenZ.GetX());
	const Asura_Vector_3 xScreenYdir(xScreenX.GetY(), xScreenY.GetY(), xScreenZ.GetY());

	// This scales the text
	const float fScale = 6.0f;
	const float fXscale = fScale * (xScreenXdir.Magnitude() / fDelta);
	const float fYscale = fScale * (xScreenYdir.Magnitude() / fDelta);

	const Asura_Vector_2 xScale(fFontWidth * fXscale, fFontHeight * fYscale);

	const Asura_Vector_2 xCurrentPos(xScreenPos.x, xScreenPos.y);
	RenderStringInternal(xCurrentPos, xScale, szString, uARGB);
}

void Asura_Primitive_Render_Data::RenderStringInternal(const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const char* const szString, const u_int uARGB)
{
	if (Asura_Primitive_Render::GetUseVectorFont())
	{
		const u_int uLen = strlen(szString);
		Asura_Vector_2 xCurrentPos(xPos.x, xPos.y);

		for (u_int u = 0; u < uLen; u++)
		{
			unsigned char ucChar = szString[u];
			if (ucChar > 127) 
			{
				ucChar = 0;
			}

			const Asura_DebugOverlay_Alphabet::Asura_AlphaCharacter& xChar = Asura_DebugOverlay_Alphabet::pxChars[ucChar];
			if (xChar.m_uNumCoords1 > 0)
			{
				RenderLineStripInternal(xCurrentPos, xScale, xChar.m_pfCoords1, xChar.m_uNumCoords1, uARGB);
			}
			if (xChar.m_uNumCoords2 > 0)
			{
				RenderLineStripInternal(xCurrentPos, xScale, xChar.m_pfCoords2, xChar.m_uNumCoords2, uARGB);
			}

			xCurrentPos.x += xScale.x;
		}
	}
	else
	{
		const u_int uScreenWidth	= Asura_Viewport::GetWidth();
		const u_int uScreenHeight	= Asura_Viewport::GetHeight();
		const float	fScreenWidth	= static_cast<float>(uScreenWidth);
		const float	fScreenHeight	= static_cast<float>(uScreenHeight);
		const int iX = Asura_Maths::FloatToInt(fScreenWidth * xPos.x / ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
		const int iY = Asura_Maths::FloatToInt(fScreenHeight * xPos.y / ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);

		Asura_DebugText::AddTextBlock(iX, iY, uARGB, 0, szString);
	}
}

void Asura_Primitive_Render_Data::RenderLineStripInternal(const Asura_Vector_2& xPos, const Asura_Vector_2& xScale, const float* const pfCoords, const u_int uNumPoints, const u_int uARGB)
{
	const u_int uScreenWidth	= Asura_Viewport::GetWidth();
	const u_int uScreenHeight	= Asura_Viewport::GetHeight();
	const float	fScreenWidth	= static_cast<float>(uScreenWidth);
	const float	fScreenHeight	= static_cast<float>(uScreenHeight);

	// The input position is in 0->1 screen dimensions
	// The input scale is in 0->1 screen dimensions
	const float* pfXValues = pfCoords;
	const float* pfYValues = pfCoords + 1;
	const u_int uNumLines = uNumPoints - 1;

	const Asura_Vector_2 xScreenScale(fScreenWidth, fScreenHeight);
	const Asura_Vector_2 xPointScale(xScreenScale.x * xScale.x, xScreenScale.y * xScale.y);

	const Asura_Vector_2 xOrigin(xPos.x * xScreenScale.x, xPos.y * xScreenScale.y);
	const Asura_Vector_2 xLineStartPoint(pfXValues[0], pfYValues[0]);
	pfXValues += 2;
	pfYValues += 2;

	Asura_Vector_2 xFrom = xOrigin + Asura_Vector_2(xLineStartPoint.x * xPointScale.x, xLineStartPoint.y * xPointScale.y);

	for (u_int u = 0; u < uNumLines; u++)
	{
		const Asura_Vector_2 xPoint(pfXValues[0], pfYValues[0]);
		pfXValues += 2;
		pfYValues += 2;

		const Asura_Vector_2 xTo = xOrigin + Asura_Vector_2(xPoint.x * xPointScale.x, xPoint.y * xPointScale.y);

		Asura_Render::Render2DLine(xFrom, xTo, uARGB);
		xFrom = xTo;
	}
}

// is there enough memory free in the list to add this type of primitive
bool Asura_Primitive_Render_Data::CanAdd(const enum ASURA_DO_COMMAND eCommand, bool bAddToStaticData)
{
	const u_char* pucListNext = bAddToStaticData ? s_pucStaticListNext : s_pucListNext;
	const u_char* pucListEnd = bAddToStaticData ? s_pucStaticListEnd : s_pucListEnd;

	const int iMemSize = s_auPrimMemSize[ eCommand ];
	if (pucListNext + iMemSize > pucListEnd)
	{
		return false;
	}
	return true;
}

// is there enough memory free in the list to add this type of primitive
bool Asura_Primitive_Render_Data::CanAdd(u_int uNumBytes, bool bAddToStaticData)
{
	const u_char* pucListNext = bAddToStaticData ? s_pucStaticListNext : s_pucListNext;
	const u_char* pucListEnd = bAddToStaticData ? s_pucStaticListEnd : s_pucListEnd;

	if (pucListNext + uNumBytes > pucListEnd)
	{
		return false;
	}
	return true;
}

// Use memory copy to avoid any alignment issues
bool Asura_Primitive_Render_Data::AddData(const void* const pData, const size_t uNumBytes, bool bAddToStaticData)
{
	u_char** pucListNext = bAddToStaticData ? &s_pucStaticListNext : &s_pucListNext;
	u_char** pucListEnd = bAddToStaticData ? &s_pucStaticListEnd : &s_pucListEnd;

	const u_char* const pucNewListNext = *pucListNext + uNumBytes;
	if (pucNewListNext > *pucListEnd)
	{
		return false;
	}	
	Asura_Mem::Copy(*pucListNext, pData, uNumBytes);
	*pucListNext += uNumBytes;
	return true;
}
bool Asura_Primitive_Render_Data::AddFloat(const float fValue, bool bAddToStaticData)
{
	return AddData(&fValue, sizeof(float), bAddToStaticData);
}

bool Asura_Primitive_Render_Data::AddUint(const u_int uValue, bool bAddToStaticData)
{
	return AddData(&uValue, sizeof(u_int), bAddToStaticData);
}

bool Asura_Primitive_Render_Data::AddVector(const Asura_Vector_3& xVector, bool bAddToStaticData)
{
	if (AddFloat(xVector.x, bAddToStaticData) == false) 
	{
		return false;
	}
	if (AddFloat(xVector.y, bAddToStaticData) == false)
	{
		return false;
	}
	if (AddFloat(xVector.z, bAddToStaticData) == false)
	{
		return false;
	}
	return true;
}

bool Asura_Primitive_Render_Data::AddQuat(const Asura_Quat& xQuat, bool bAddToStaticData)
{
	if (AddFloat(xQuat.x, bAddToStaticData) == false) 
	{
		return false;
	}
	if (AddFloat(xQuat.y, bAddToStaticData) == false)
	{
		return false;
	}
	if (AddFloat(xQuat.z, bAddToStaticData) == false)
	{
		return false;
	}
	if (AddFloat(xQuat.w, bAddToStaticData) == false)
	{
		return false;
	}
	return true;
}

bool Asura_Primitive_Render_Data::AddRGBA(const Asura_RGBA& xColour, bool bAddToStaticData)
{
	const u_int uColour = xColour.ToUIntARGB(true);
	return AddARGB(uColour, bAddToStaticData);
}

bool Asura_Primitive_Render_Data::AddARGB(const u_int uARGB, bool bAddToStaticData)
{
	return AddData(&uARGB, sizeof(u_int), bAddToStaticData);
}

bool Asura_Primitive_Render_Data::AddString(const char* const szString, const u_int uNumBytes, bool bAddToStaticData)
{
	return AddData(szString, uNumBytes, bAddToStaticData);
}

// For simple fixed sized primitives
bool Asura_Primitive_Render_Data::AddHeader(const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer, float fTime, bool bAddToStaticData)
{
	// The ASURA_DO_CONSOLE_STRING & ASURA_DO_DEBUG_STRING commands are variable length so have to use the other AddHeader function
	Asura_Assertf(((eCommand != ASURA_DO_CONSOLE_STRING) && (eCommand != ASURA_DO_DEBUG_STRING)), ("AddHeader wrong eCommand %d", eCommand));
	const u_int uNumBytes = s_auPrimMemSize[eCommand];
	return AddHeader(eCommand, bWorldSpace, bIgnoreZBuffer, uNumBytes, fTime, bAddToStaticData);
}

// For primitives with variable size e.g. text based ones
bool Asura_Primitive_Render_Data::AddHeader(const enum ASURA_DO_COMMAND eCommand, const bool bWorldSpace, const bool bIgnoreZBuffer, const u_int uNumBytes, float fTime, bool bAddToStaticData)
{
	const u_char* const* pucListNext = bAddToStaticData ? &s_pucStaticListNext : &s_pucListNext;
	const u_char* const* pucListEnd = bAddToStaticData ? &s_pucStaticListEnd : &s_pucListEnd;

	const u_char* const pucNewListNext = *pucListNext + uNumBytes;
	if (pucNewListNext > *pucListEnd)
	{
		Asura_Assertf(false, ("AddHeader not enough memory adding eCommand %d uNumBytes %d", eCommand, uNumBytes));
		return false;
	}	
	// includes header size
	if (AddUint(uNumBytes, bAddToStaticData) == false)
	{
		Asura_Assertf(false, ("AddHeader failed to add command size eCommand %d uNumBytes %d", eCommand, uNumBytes));
		return false;
	}
	u_int uHeader = (static_cast<u_int>(eCommand) & ASURA_DO_COMMAND_BIT_MASK);
	uHeader |= (bWorldSpace ? ASURA_DO_WORLD_SPACE_BIT_MASK : 0);
	uHeader |= (bIgnoreZBuffer ? ASURA_DO_IGNORE_Z_BUFFER_BIT_MASK : 0);
	if (AddUint(uHeader, bAddToStaticData) == false)
	{
		Asura_Assertf(false, ("AddHeader failed to add header eCommand %d uNumBytes %d", eCommand, uNumBytes));
		return false;
	}

	u_int uEndTime = Asura_Timers::GetMSecsSinceStart() + static_cast<u_int> (fTime * 1000.0f);

	if (AddUint(uEndTime, bAddToStaticData) == false)
	{
		Asura_Assertf(false, ("AddHeader failed to add end time to header"));
		return false;
	}
	return true;
}

// Use memory copy to avoid any alignment issues
bool Asura_Primitive_Render_Data::ReadData(void* const pData, const size_t uNumBytes, bool bReadFromStaticData)
{
	u_char** pucListNext = bReadFromStaticData ? &s_pucStaticListNext : &s_pucListNext;
	u_char** pucListEnd = bReadFromStaticData ? &s_pucStaticListEnd : &s_pucListEnd;

	const u_char* const pucNewListNext = *pucListNext + uNumBytes;
	if(pucNewListNext > *pucListEnd)
	{
		return false;
	}	
	Asura_Mem::Copy(pData, *pucListNext, uNumBytes);
	*pucListNext += uNumBytes;
	return true;
}

bool Asura_Primitive_Render_Data::ReadFloat(float& fValue, bool bReadFromStaticData)
{
	return ReadData(&fValue, sizeof(float), bReadFromStaticData);
}

bool Asura_Primitive_Render_Data::ReadUint(u_int& uValue, bool bReadFromStaticData)
{
	return ReadData(&uValue, sizeof(u_int), bReadFromStaticData);
}

bool Asura_Primitive_Render_Data::ReadVector(Asura_Vector_3& xVector, bool bReadFromStaticData)
{
	if (ReadFloat(xVector.x, bReadFromStaticData) == false) 
	{
		return false;
	}
	if (ReadFloat(xVector.y, bReadFromStaticData) == false)
	{
		return false;
	}
	if (ReadFloat(xVector.z, bReadFromStaticData) == false)
	{
		return false;
	}
	return true;
}

bool Asura_Primitive_Render_Data::ReadQuat(Asura_Quat& xQuat, bool bReadFromStaticData)
{
	if (ReadFloat(xQuat.x, bReadFromStaticData) == false) 
	{
		return false;
	}
	if (ReadFloat(xQuat.y, bReadFromStaticData) == false)
	{
		return false;
	}
	if (ReadFloat(xQuat.z, bReadFromStaticData) == false)
	{
		return false;
	}
	if (ReadFloat(xQuat.w, bReadFromStaticData) == false)
	{
		return false;
	}
	return true;
}

bool Asura_Primitive_Render_Data::ReadRGBA(Asura_RGBA& xColour, bool bReadFromStaticData)
{
	u_int uColour = 0;
	if (ReadARGB(uColour, bReadFromStaticData) == false)
	{
		return false;
	}

	xColour.FromUIntARGB(uColour);
	return true;
}

bool Asura_Primitive_Render_Data::ReadARGB(u_int& uARGB, bool bReadFromStaticData)
{
	return ReadData(&uARGB, sizeof(u_int), bReadFromStaticData);
}

bool Asura_Primitive_Render_Data::ReadString(const char*& szString, const u_int uNumBytes, bool bReadFromStaticData)
{
	u_char** pucListNext = bReadFromStaticData ? &s_pucStaticListNext : &s_pucListNext;
	u_char** pucListEnd = bReadFromStaticData ? &s_pucStaticListEnd : &s_pucListEnd;

	const u_char* const pucNewListNext = *pucListNext + uNumBytes;
	if (pucNewListNext > *pucListEnd)
	{
		return false;
	}	

	szString = reinterpret_cast<char*>(*pucListNext);
	*pucListNext += uNumBytes;
	return true;
}

// Extract data from the command list
bool Asura_Primitive_Render_Data::ReadHeader(u_int& uNumBytes, enum ASURA_DO_COMMAND& eCommand, bool& bWorldSpace, bool& bIgnoreZBuffer, u_int& uEndTime, bool bReadFromStaticData)
{
	if (ReadUint(uNumBytes, bReadFromStaticData) == false)
	{
		Asura_Assert(false, "ReadHeader failed reading command size");
		return false;
	}
	u_int uHeader = 0;
	if (ReadUint(uHeader, bReadFromStaticData) == false)
	{
		Asura_Assert(false, "ReadHeader failed reading header");
		return false;
	}
	eCommand = static_cast<ASURA_DO_COMMAND>(uHeader & ASURA_DO_COMMAND_BIT_MASK);
	bWorldSpace = (uHeader & ASURA_DO_WORLD_SPACE_BIT_MASK) ? true : false;
	bIgnoreZBuffer = (uHeader & ASURA_DO_IGNORE_Z_BUFFER_BIT_MASK) ? true : false;
	if (ReadUint(uEndTime, bReadFromStaticData) == false)
	{
		Asura_Assert(false, "ReadHeader failed reading time");
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Command console functions
//
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#define CONSOLE_CMDNAME_ACTIVATE			L"Debug.PrimitiveRender.Activate"
#define CONSOLE_VARNAME_USE_VECTOR_FONT 	L"Debug.PrimitiveRender.UseVectorFont"
#define CONSOLE_VARNAME_VECTOR_FONT_WIDTH	L"Debug.PrimitiveRender.VectorFontWidth"
#define CONSOLE_VARNAME_VECTOR_FONT_WHRATIO	L"Debug.PrimitiveRender.VectorFontWHRatio"

static bool ConsoleCmd_Activate(bool bHelp)
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf(CONSOLE_CMDNAME_ACTIVATE L": activate/de-activate the primitive render system");
	}
	else
	{
		Asura_Primitive_Render::SetActive(Asura_Primitive_Render::IsActive() ^ true);
	}
	return true;
}

#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE

////////////////////////////////////////////////////////////////////////////////////////////////
//
// Internal data & functions for Asura_Primitive_Render
//
////////////////////////////////////////////////////////////////////////////////////////////////

bool Asura_Primitive_Render::s_bActive = false;
bool Asura_Primitive_Render::s_bUseVectorFont = false;
float Asura_Primitive_Render::s_fVectorFontWidth = 12.0f;
float Asura_Primitive_Render::s_fVectorFontWHRatio = 0.5f;
Asura_TemplateHashTree<bool> Asura_Primitive_Render::s_xCategoryTree;
Asura_Hash_ID Asura_Primitive_Render::s_uCurrentCategoryHash = ASURA_HASH_ID_UNSET;
float Asura_Primitive_Render::s_fTime = 0.0f;
bool Asura_Primitive_Render::s_bUseZBuffer = true;

// Initialise internals - allocate memory buffer
void Asura_Primitive_Render::Initialise()
{
	const int iMemorySize = ASURA_PRIMITIVE_RENDER_MEMORY_SIZE;
	if (Asura_Primitive_Render_Data::Initialise(iMemorySize) == false)
	{
		Asura_Assertf(false, ("Asura_Primitive_Render failed to initialise with memory size of %dKB", iMemorySize/1024));
		return;
	}

	#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	Asura_CommandConsole::AddCmd(CONSOLE_CMDNAME_ACTIVATE, ConsoleCmd_Activate, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(CONSOLE_VARNAME_USE_VECTOR_FONT, &s_bUseVectorFont, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(CONSOLE_VARNAME_VECTOR_FONT_WIDTH, &s_fVectorFontWidth, 1.0f, 32.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(CONSOLE_VARNAME_VECTOR_FONT_WHRATIO, &s_fVectorFontWHRatio, 0.1f, 2.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	#endif

	SetRenderParams();

	Reset();
}

// Free memory
void Asura_Primitive_Render::Shutdown()
{
	Reset();
	Asura_Primitive_Render_Data::Shutdown();
	s_xCategoryTree.Reset();
	return;
}

// Reset internal state
void Asura_Primitive_Render::Reset()
{
	Asura_Primitive_Render_Data::Reset();
	return;
}

// Register a category
Asura_Hash_ID Asura_Primitive_Render::RegisterCategory(const char* szCategoryName)
{
	Asura_Hash_ID uHashId = Asura_GetHashID(szCategoryName);

	bool* pbCategoryEnabled = s_xCategoryTree.AddNode(uHashId);

	if(pbCategoryEnabled != NULL)
	{
		*pbCategoryEnabled = true;

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
		Asura_Char szCommandConsoleString[256];
		Asura_UnicodeChar wszCommandConsoleString[256];
		Asura_PrintfToBfr(szCommandConsoleString, "%s%s", 256, true, CONSOLE_COMMAND_PREFIX, szCategoryName);
		Asura_UTF8ToUnicode(szCommandConsoleString, wszCommandConsoleString);
		Asura_CommandConsole::AddVar(wszCommandConsoleString, pbCategoryEnabled, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif
	}

	return uHashId;
}

// Enable a category
void Asura_Primitive_Render::EnableCategory(Asura_Hash_ID uHashId, bool bFlag)
{
	bool *pbCategoryEnabled = s_xCategoryTree.GetDataByHash(uHashId);

	Asura_Assert(pbCategoryEnabled, "Category hash not found in tree.");

	if(pbCategoryEnabled)
	{
		*pbCategoryEnabled = bFlag;
	}
}

bool Asura_Primitive_Render::IsCategoryActive(Asura_Hash_ID uHashId)
{
	bool *pbCategoryEnabled = s_xCategoryTree.GetDataByHash(uHashId);

	return  *pbCategoryEnabled;
}

// Render the debug data - world space data, before the z-buffer clear
void Asura_Primitive_Render::Render()
{
	if (s_bActive == true)
	{
		Asura_Primitive_Render_Data::Render();
	}

	// All done reset the list ready for the next cycle
	Asura_Primitive_Render_Data::ResetCommandList();

	return;
}

// Update the Asura_Debug system - calls the correct debug callback
void Asura_Primitive_Render::Update()
{
	// Do nothing if we aren't active
	if (s_bActive == false)
	{
		return;
	}

	Asura_Primitive_Render_Data::Update();
	return;
}

// Draw text in screenspace (not z-buffered)
void Asura_Primitive_Render::DrawTextScreen(const Asura_Vector_2& xPos, const u_int uARGB, const char* const szFmtText, ...)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	Asura_String2048 xBuffer;
	va_list xArgs;
	va_start(xArgs, szFmtText);
	xBuffer.VPrint(szFmtText, xArgs);

	const bool bWorldSpace = false;
	const bool bIgnoreZBuffer = true;
	const Asura_Vector_3 xPos3D(xPos.x, xPos.y, 0.0f);
	AddDebugString(bWorldSpace, bIgnoreZBuffer, xBuffer.GetString(), xPos3D, uARGB);
}

// Draw text in world space (z-buffered)
void Asura_Primitive_Render::DrawTextWorld(const Asura_Vector_3& xPos, const u_int uARGB, const char* const szFmtText, ...)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	Asura_String2048 xBuffer;
	va_list xArgs;
	va_start(xArgs, szFmtText);
	xBuffer.VPrint(szFmtText, xArgs);

	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = true;
	AddDebugString(bWorldSpace, bIgnoreZBuffer, xBuffer.GetString(), xPos, uARGB);
}

void Asura_Primitive_Render::DrawConsoleText(const char* const szFmtText, ...)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	Asura_String2048 xBuffer;
	va_list xArgs;
	va_start(xArgs, szFmtText);
	xBuffer.VPrint(szFmtText, xArgs);

	AddConsoleString(xBuffer.GetString(), Asura_Primitive_Render_Data::s_uDefaultConsoleColour);
}

void Asura_Primitive_Render::DrawConsoleText(const u_int uARGB, const char* const szFmtText, ...)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	Asura_String2048 xBuffer;

	va_list xArgs;
	va_start(xArgs, szFmtText);
	xBuffer.VPrint(szFmtText, xArgs);

	AddConsoleString(xBuffer.GetString(), uARGB);
}

void Asura_Primitive_Render::Printf(const char* const szFmtText, ...)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}

	Asura_String2048 xBuffer;

	va_list xArgs;
	va_start(xArgs, szFmtText);
	xBuffer.VPrint(szFmtText, xArgs);

	Asura_OutputToDebugger_Retail(("%s\n", xBuffer.GetString()));
}

// Add a sphere to the list
void Asura_Primitive_Render::AddSphere(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a sphere
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_SPHERE, bUseStaticData) == false)
	{
		return;
	}
	
	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_SPHERE, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddSphere failed to add the header");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xCentre, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddSphere failed to add centre");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fRadius, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddSphere failed to add radius");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddSphere failed to add ARGB");
		return;
	}
	if (Asura_Primitive_Render_Data::AddUint(uNumSegments, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddSphere failed to add num segments");
		return;
	}
	if (Asura_Primitive_Render_Data::AddUint(uNumSlices, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddSphere failed to add num slices");
		return;
	}

	return;
}

// Add line to the list
void Asura_Primitive_Render::AddLine(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a line
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_LINE, bUseStaticData) == false)
	{
		return;
	}
	
	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_LINE, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddLine failed to add header");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xStart, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddLine failed to add start");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xEnd, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddLine failed to add end");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddLine failed to add ARGB");
		return;
	}

	return;
}

// Add arrow to the list
void Asura_Primitive_Render::AddArrow(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add an arrow
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_ARROW, bUseStaticData) == false)
	{
		return;
	}
	
	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_ARROW, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddArrow failed to add header");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xStart, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddArrow failed to add start");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xEnd, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddArrow failed to add end");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddArrow failed to add ARGB");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fHeadProportion, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddArrow failed to add head proportion");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fCrossSize, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddArrow failed to add head cross size");
		return;
	}
}

void Asura_Primitive_Render::AddCross(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a cross
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_CROSS, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_CROSS, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCross failed to add header");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xPosition, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCross failed to add position");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddLine failed to add ARGB");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fSize, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddLine failed to add size");
		return;
	}

	return;
}

void Asura_Primitive_Render::AddOBB(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_OBB& xOBB, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add an OBB
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_OBB, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_OBB, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddOBB failed to add header");
		return;
	}
	const Asura_AVector3 xCentreA = xOBB.GetCentre();
	const Asura_AMatrix3x3 xOrientA = xOBB.GetOrientation();
	const Asura_AVector3 xHalfExtentsA = xOBB.GetHalfExtents();

	const Asura_Vector_3 xCentre(xCentreA);
	if (Asura_Primitive_Render_Data::AddVector(xCentre, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddOBB failed to add centre");
		return;
	}
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientA);
	if (Asura_Primitive_Render_Data::AddQuat(xQuat, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddOBB failed to add orientation");
		return;
	}
	const Asura_Vector_3 xHalfExtents(xHalfExtentsA);
	if (Asura_Primitive_Render_Data::AddVector(xHalfExtents, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddOBB failed to add half extents");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddOBB failed to add ARGB");
		return;
	}

	return;
}

void Asura_Primitive_Render::AddAxis(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add an axis
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_AXIS, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_AXIS, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddAxis failed to add header");
	}
	if (Asura_Primitive_Render_Data::AddVector(xPosition, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddAxis failed to add position");
		return;
	}
	if (Asura_Primitive_Render_Data::AddQuat(xOrientation, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddAxis failed to add orientation");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fSize, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddAxis failed to add size");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fAlpha, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddAxis failed to add alpha");
		return;
	}

	return;
}

void Asura_Primitive_Render::AddBoundingBox(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a bounding box
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_BOUNDING_BOX, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_BOUNDING_BOX, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddBoundingBox failed to add header");
	}
	const Asura_Vector_3 xMin = xBB.GetMin();
	if (Asura_Primitive_Render_Data::AddVector(xMin, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddBoundingBox failed to add min");
		return;
	}
	const Asura_Vector_3 xMax = xBB.GetMax();
	if (Asura_Primitive_Render_Data::AddVector(xMax, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddBoundingBox failed to add max");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xPosition, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddBoundingBox failed to add position");
		return;
	}
	if (Asura_Primitive_Render_Data::AddQuat(xOrientation, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddBoundingBox failed to add orientation");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddBoundingBox failed to add ARGB");
		return;
	}

	return;
}

void Asura_Primitive_Render::AddFrustum(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a frustum
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_FRUSTUM, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_FRUSTUM, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add header");
	}
	const float fNearPlane = xFrustum.GetNearPlane();
	if (Asura_Primitive_Render_Data::AddFloat(fNearPlane, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add near plane");
		return;
	}
	const float fFarPlane = xFrustum.GetFarPlane();
	if (Asura_Primitive_Render_Data::AddFloat(fFarPlane, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add far plane");
		return;
	}
	const float fTanHorizontalFOV = xFrustum.GetTanHorizontalFOV();
	if (Asura_Primitive_Render_Data::AddFloat(fTanHorizontalFOV, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add tan horizontal FOV");
		return;
	}
	const float fTanVerticalFOV = xFrustum.GetTanVerticalFOV();
	if (Asura_Primitive_Render_Data::AddFloat(fTanVerticalFOV, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add tan vertical FOV");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xPosition, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add position");
		return;
	}
	if (Asura_Primitive_Render_Data::AddQuat(xOrientation, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add orientation");
		return;
	}

	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddFrustum failed to add ARGB");
		return;
	}
	return;
}

void Asura_Primitive_Render::AddCylinder(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a cylinder?
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_CYLINDER, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_CYLINDER, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add header");
	}
	if (Asura_Primitive_Render_Data::AddVector(xPosition, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add position");
		return;
	}
	if (Asura_Primitive_Render_Data::AddQuat(xOrientation, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add orientation");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fHeight, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add orientation");
		return;
	}
	if (Asura_Primitive_Render_Data::AddFloat(fRadius, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add radius");
		return;
	}
	if (Asura_Primitive_Render_Data::AddUint(uNumSegments, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add num segments");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCylinder failed to add ARGB");
		return;
	}
	return;
}

void Asura_Primitive_Render::AddCapsule(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const Asura_Vector_3& xExtents, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Is there enough memory to add a capsule?
	if (Asura_Primitive_Render_Data::CanAdd(Asura_Primitive_Render_Data::ASURA_DO_CAPSULE, bUseStaticData) == false)
	{
		return;
	}

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_CAPSULE, bWorldSpace, bIgnoreZBuffer, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCapsule failed to add header");
	}
	if (Asura_Primitive_Render_Data::AddVector(xPosition, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCapsule failed to add position");
		return;
	}
	if (Asura_Primitive_Render_Data::AddQuat(xOrientation, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCapsule failed to add orientation");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xExtents, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCapsule failed to add extents");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddCapsule failed to add ARGB");
		return;
	}
	return;
}

void Asura_Primitive_Render::AddConsoleString(const char* const szText, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Include the null terminator
	const int iStringLen = strlen(szText) + 1;
	const u_int uNumBytes = Asura_Primitive_Render_Data::s_auPrimMemSize[Asura_Primitive_Render_Data::ASURA_DO_CONSOLE_STRING] + iStringLen;

	// Is there enough memory to add the text
	//if (Asura_Primitive_Render_Data::s_pucListNext + uNumBytes > Asura_Primitive_Render_Data::s_pucListEnd)
	if(!Asura_Primitive_Render_Data::CanAdd(uNumBytes, bUseStaticData))
	{
		return;
	}
	const bool bWorldSpace = false;
	const bool bIgnoreZBuffer = true;

	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_CONSOLE_STRING, bWorldSpace, bIgnoreZBuffer, uNumBytes, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddConsoleString failed to add header");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddConsoleString failed to add colour");
		return;
	}
	if (Asura_Primitive_Render_Data::AddString(szText, iStringLen, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddConsoleString failed to add string");
		return;
	}
}

void Asura_Primitive_Render::AddDebugString(const bool bWorldSpace, const bool bIgnoreZBuffer, const char* const szText, const Asura_Vector_3& xPos, const u_int uARGB)
{
	bool bUseStaticData = (s_fTime > 0.0f);

	// Include the null terminator
	const int iStringLen = strlen(szText) + 1;
	const u_int uNumBytes = Asura_Primitive_Render_Data::s_auPrimMemSize[Asura_Primitive_Render_Data::ASURA_DO_DEBUG_STRING] + iStringLen;

	// Is there enough memory to add the text
	//if (Asura_Primitive_Render_Data::s_pucListNext + uNumBytes > Asura_Primitive_Render_Data::s_pucListEnd)
	if(!Asura_Primitive_Render_Data::CanAdd(uNumBytes, bUseStaticData))
	{
		return;
	}
	if (Asura_Primitive_Render_Data::AddHeader(Asura_Primitive_Render_Data::ASURA_DO_DEBUG_STRING, bWorldSpace, bIgnoreZBuffer, uNumBytes, s_fTime, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddDebugString failed to add header");
		return;
	}
	if (Asura_Primitive_Render_Data::AddVector(xPos, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddDebugString failed to add pos");
		return;
	}
	if (Asura_Primitive_Render_Data::AddARGB(uARGB, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddDebugString failed to add colour");
		return;
	}
	if (Asura_Primitive_Render_Data::AddString(szText, iStringLen, bUseStaticData) == false)
	{
		Asura_Assert(false, "AddDebugString failed to add string");
		return;
	}
}

#endif // #if ASURA_USE_PRIMITIVE_RENDER == 1
