///////////////////////////////////////////////////////////////////////////
//
// Asura_Primitive_Render.h
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////

#ifndef ASURA_PRIMITIVE_RENDER_HH
#define ASURA_PRIMITIVE_RENDER_HH

#include "Core/Asura_Maths.h"
#include "Core/Asura_HashTree.h"

// Forward declarations
class Asura_RGBA;
class Asura_Frustum;

// Static singleton class
class Asura_Primitive_Render
{
public:
	// Initialise internals - allocate memory buffer
	static void Initialise();
	// Free memory
	static void Shutdown();
 
	// Reset internal state
	static void Reset();

	// Register a category
	static Asura_Hash_ID RegisterCategory(const char* szCategoryName);

	// Register a category
	static void EnableCategory(Asura_Hash_ID uHashId, bool bFlag);

	static bool IsCategoryActive(Asura_Hash_ID uHashId);

	// Render the debug data -  does world space data, then screen space data
	static void Render();

	// Update the Asura_Debug system - calls the correct debug callback
	static void Update();

	// Primarily for the Asura_Primitive_Render_Data namespace :(
	static void SetActive(const bool bActive);
	static bool IsActive();

	#ifdef ASURA_USE_PRIMITIVE_RENDER
	static bool GetUseVectorFont() { return s_bUseVectorFont; }
	static float GetVectorFontWidth() { return s_fVectorFontWidth; }
	static float GetVectorFontWHRatio() { return s_fVectorFontWHRatio; }
	#endif

	static void SetRenderParams(float fTime = 0.0f, Asura_Hash_ID uCategory = ASURA_HASH_ID_UNSET, bool bUseZbuffer = true);

	static void UseDefaultRenderParams() { SetRenderParams(); }

	// Draw a sphere in world space
	static void DrawSphereWorld(const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments = 16, const u_int uNumSlices = 8);

	// Draw a line in world space
	static void DrawLineWorld(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB);
	static void DrawLineWorld(const Asura_AVector3& xStart, const Asura_AVector3& xEnd, const u_int uARGB);

	// Draw an arrow in world space
	static void DrawArrowWorld(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion = 0.125f, const float fCrossSize = 0.0f);

	// Draw a cross (XYZ world axis set) in world space
	static void DrawCrossWorld(const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize = 1.0f);

	// Draw an OBB in world space
	static void DrawOBBWorld(const Asura_OBB& xOBB, const u_int uARGB);

	// Draw text in world space (z-buffered)
	static void DrawTextWorld(const Asura_Vector_3& xPosition, const u_int uARGB, const char* const szFmtText, ...);

	// Quaternion functions...
	static void DrawAxisWorld(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize = 1.0f, const float fAlpha = 1.0f);
	static void DrawBoundingBoxWorld(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB);
	static void DrawFrustumWorld(const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB);
	static void DrawCylinderWorld(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB);
	static void DrawCapsuleWorld(const Asura_AVector3& xPosition, const Asura_Quat& xOrientation, const Asura_AVector3& xExtents, const u_int uARGB);

	// 3x3 matrix functions...
	static void DrawAxisWorld(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const float fSize = 1.0f, const float fAlpha = 1.0f);
	static void DrawBoundingBoxWorld(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB);
	static void DrawFrustumWorld(const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB);
	static void DrawCylinderWorld(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB);
	static void DrawCapsuleWorld(const Asura_AVector3& xPosition, const Asura_AMatrix3x3& xOrientation, const Asura_AVector3& xExtents, const u_int uARGB);

	// Draw text in normalised screenspace [top-left is 0,0, bottom-right 1,1] (not z-buffered)
	static void DrawTextScreen(const Asura_Vector_2& xPosition, const u_int uARGB, const char* const szFmtText, ...);

	// Output a console message using the default console colour
	static void DrawConsoleText(const char* const szFmtText, ...);
	// Console message with a specific colour
	static void DrawConsoleText(const u_int uARGB, const char* const szFmtText, ...);

	// Printf debugging but filtered by the mode and if the primitive render system is active
	static void Printf(const char* const szFmtText, ...);

private:

	#ifdef ASURA_USE_PRIMITIVE_RENDER
	static bool s_bActive;
	static bool s_bUseVectorFont;
	static float s_fVectorFontWidth;
	static float s_fVectorFontWHRatio;
	static Asura_TemplateHashTree<bool> s_xCategoryTree;
	static Asura_Hash_ID s_uCurrentCategoryHash;
	static float s_fTime;
	static bool s_bUseZBuffer;
	#endif

	// Internal functions to add the commands to the command list
	static void AddSphere(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices);
	static void AddLine(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB);
	static void AddArrow(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize);
	static void AddCross(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize); 
	static void AddOBB(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_OBB& xOBB, const u_int uARGB); 
	static void AddAxis(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha);
	static void AddBoundingBox(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB);
	static void AddFrustum(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB);
	static void AddCylinder(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB);
	static void AddCapsule(const bool bWorldSpace, const bool bIgnoreZBuffer, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const Asura_Vector_3& xExtents, const u_int uARGB);

	static void AddConsoleString(const char* const szText, const u_int uARGB);
	static void AddDebugString(const bool bWorldSpace, const bool bIgnoreZBuffer, const char* const szText, const Asura_Vector_3& xPos, const u_int uARGB);
};

#ifdef ASURA_USE_PRIMITIVE_RENDER

inline void Asura_Primitive_Render::SetRenderParams(float fTime, Asura_Hash_ID uCategory, bool bUseZbuffer)
{
	s_fTime = fTime; s_uCurrentCategoryHash = uCategory; s_bUseZBuffer = bUseZbuffer;
}

inline void Asura_Primitive_Render::SetActive(const bool bActive) { s_bActive = bActive; }

// Primarily for the Asura_Primitive_Render_Data namespace :(
inline bool Asura_Primitive_Render::IsActive() { return s_bActive; }

// If worried about code size and potential i-cache hits then don't inline these functions

// Draw a sphere in world space
inline void Asura_Primitive_Render::DrawSphereWorld(const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}

	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddSphere(bWorldSpace, bIgnoreZBuffer, xCentre, fRadius, uARGB, uNumSegments, uNumSlices);
}

// Draw a line in world space
inline void Asura_Primitive_Render::DrawLineWorld(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddLine(bWorldSpace, bIgnoreZBuffer, xStart, xEnd, uARGB);
}

inline void Asura_Primitive_Render::DrawLineWorld(const Asura_AVector3& xStart, const Asura_AVector3& xEnd, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const Asura_Vector_3 xStart3(xStart);
	const Asura_Vector_3 xEnd3(xEnd);
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddLine(bWorldSpace, bIgnoreZBuffer, xStart3, xEnd3, uARGB);
}

// Draw a arrow in world space
inline void Asura_Primitive_Render::DrawArrowWorld(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddArrow(bWorldSpace, bIgnoreZBuffer, xStart, xEnd, uARGB, fHeadProportion, fCrossSize);
}

// Draw a cross in world space
inline void Asura_Primitive_Render::DrawCrossWorld(const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddCross(bWorldSpace, bIgnoreZBuffer, xPosition, uARGB, fSize);
}

// Draw an OBB in world space
inline void Asura_Primitive_Render::DrawOBBWorld(const Asura_OBB& xOBB, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddOBB(bWorldSpace, bIgnoreZBuffer, xOBB, uARGB);
}

// Quaternion functions...
inline void Asura_Primitive_Render::DrawAxisWorld(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddAxis(bWorldSpace, bIgnoreZBuffer, xPosition, xOrientation, fSize, fAlpha);
}

inline void Asura_Primitive_Render::DrawBoundingBoxWorld(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddBoundingBox(bWorldSpace, bIgnoreZBuffer, xBB, xPosition, xOrientation, uARGB);
}

inline void Asura_Primitive_Render::DrawFrustumWorld(const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddFrustum(bWorldSpace, bIgnoreZBuffer, xFrustum, xPosition, xOrientation, uARGB);
}

inline void Asura_Primitive_Render::DrawCylinderWorld(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddCylinder(bWorldSpace, bIgnoreZBuffer, xPosition, xOrientation, fHeight, fRadius, uNumSegments, uARGB);
}

inline void Asura_Primitive_Render::DrawCapsuleWorld(const Asura_AVector3& xPosition, const Asura_Quat& xOrientation, const Asura_AVector3& xExtents, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const Asura_Vector_3 xPos3(xPosition);
	const Asura_Vector_3 xExtents3(xExtents);
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	AddCapsule(bWorldSpace, bIgnoreZBuffer, xPos3, xOrientation, xExtents3, uARGB);
}

// 3x3 matrix functions...
inline void Asura_Primitive_Render::DrawAxisWorld(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const float fSize, const float fAlpha)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientation);
	AddAxis(bWorldSpace, bIgnoreZBuffer, xPosition, xQuat, fSize, fAlpha);
}

inline void Asura_Primitive_Render::DrawBoundingBoxWorld(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientation);
	AddBoundingBox(bWorldSpace, bIgnoreZBuffer, xBB, xPosition, xQuat, uARGB);
}

inline void Asura_Primitive_Render::DrawFrustumWorld(const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientation);
	AddFrustum(bWorldSpace, bIgnoreZBuffer, xFrustum, xPosition, xQuat, uARGB);
}

inline void Asura_Primitive_Render::DrawCylinderWorld(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientation);
	AddCylinder(bWorldSpace, bIgnoreZBuffer, xPosition, xQuat, fHeight, fRadius, uNumSegments, uARGB);
}

inline void Asura_Primitive_Render::DrawCapsuleWorld(const Asura_AVector3& xPosition, const Asura_AMatrix3x3& xOrientation, const Asura_AVector3& xExtents, const u_int uARGB)
{
	bool* pbCategoryEnabled = s_xCategoryTree.GetDataByHash(s_uCurrentCategoryHash);

	if ((s_bActive == false) || ((pbCategoryEnabled != NULL) && (*pbCategoryEnabled == false)))
	{
		return;
	}
	const Asura_Vector_3 xPos3(xPosition);
	const Asura_Vector_3 xExtents3(xExtents);
	const bool bWorldSpace = true;
	const bool bIgnoreZBuffer = !s_bUseZBuffer;
	Asura_Quat xQuat;
	xQuat.FromMatrix(xOrientation);
	AddCapsule(bWorldSpace, bIgnoreZBuffer, xPos3, xQuat, xExtents3, uARGB);
}

#else // ASURA_USE_PRIMITIVE_RENDER

inline void Asura_Primitive_Render::SetRenderParams(float fTime, Asura_Hash_ID uCategory, bool bUseZbuffer) {}

// Just empty inline functions to effectively remove the code when the #define isn't defined
// Initialise internals - allocate memory buffer
inline void Asura_Primitive_Render::Initialise() {}

// Free memory
inline void Asura_Primitive_Render::Shutdown() {}
 
// Reset internal state
inline void Asura_Primitive_Render::Reset()  {}

// Render the debug data -  does world space data, then screen space data
inline void Asura_Primitive_Render::Render(void) {}

// Update the Asura_Debug system - calls the correct debug callback
inline void Asura_Primitive_Render::Update(void) {}

inline void Asura_Primitive_Render::SetActive(const bool bActive) { (void)bActive; }

// Primarily for the Asura_Primitive_Render_Data namespace :(
inline bool Asura_Primitive_Render::IsActive(void) { return false; }

inline Asura_Hash_ID Asura_Primitive_Render::RegisterCategory(const char* szCategoryName) { (void)szCategoryName; return 0; }
inline void Asura_Primitive_Render::EnableCategory(Asura_Hash_ID uHashId, bool bFlag) { (void)uHashId; (void)bFlag; }
inline bool Asura_Primitive_Render::IsCategoryActive(Asura_Hash_ID uHashId) { (void)uHashId; return false; }

inline void Asura_Primitive_Render::DrawSphereWorld(const Asura_Vector_3& xCentre, const float fRadius, const u_int uARGB, const u_int uNumSegments, const u_int uNumSlices) { (void)xCentre; (void)fRadius; (void)uARGB; (void)uNumSegments; (void)uNumSlices; }
inline void Asura_Primitive_Render::DrawLineWorld(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB) { (void)xStart; (void)xEnd; (void)uARGB; }
inline void Asura_Primitive_Render::DrawLineWorld(const Asura_AVector3& xStart, const Asura_AVector3& xEnd, const u_int uARGB) { (void)xStart; (void)xEnd; (void)uARGB; }
inline void Asura_Primitive_Render::DrawArrowWorld(const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd, const u_int uARGB, const float fHeadProportion, const float fCrossSize) { (void)xStart; (void)xEnd; (void)uARGB; (void)fHeadProportion; (void)fCrossSize; }
inline void Asura_Primitive_Render::DrawTextWorld(const Asura_Vector_3& xPosition, const u_int uARGB, const char* const szFmtText, ...) { (void)xPosition; (void)uARGB; (void)szFmtText; }
inline void Asura_Primitive_Render::DrawCrossWorld(const Asura_Vector_3& xPosition, const u_int uARGB, const float fSize) { (void)xPosition; (void)uARGB; (void)fSize; }
inline void Asura_Primitive_Render::DrawOBBWorld(const Asura_OBB& xOBB, const u_int uARGB) { (void)xOBB; (void)uARGB; }

inline void Asura_Primitive_Render::DrawAxisWorld(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const float fSize, const float fAlpha) { (void)xPosition; (void)xOrientation; (void)fSize; (void)fAlpha; }
inline void Asura_Primitive_Render::DrawBoundingBoxWorld(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB) { (void)xBB; (void)xPosition; (void)xOrientation; (void)uARGB; }
inline void Asura_Primitive_Render::DrawFrustumWorld(const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const u_int uARGB) { (void)xPosition; (void)xOrientation; (void)uARGB; }

inline void Asura_Primitive_Render::DrawAxisWorld(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const float fSize, const float fAlpha) { (void)xPosition; (void)xOrientation; (void)fSize; (void)fAlpha; }
inline void Asura_Primitive_Render::DrawBoundingBoxWorld(const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB) { (void)xBB; (void)xPosition; (void)xOrientation; (void)uARGB; }
inline void Asura_Primitive_Render::DrawFrustumWorld(const Asura_Frustum& xFrustum, const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, const u_int uARGB) { (void)xPosition; (void)xOrientation; (void)uARGB; }
inline void Asura_Primitive_Render::DrawCylinderWorld(const Asura_Vector_3& xPosition, const Asura_Quat& xOrientation, float fHeight, float fRadius, u_int uNumSegments, const u_int uARGB) { (void)xPosition; (void)xOrientation; (void)fHeight; (void)fRadius; (void)uNumSegments; (void)uARGB;  }
inline void Asura_Primitive_Render::DrawCapsuleWorld(const Asura_AVector3& xPosition, const Asura_Quat& xOrientation, const Asura_AVector3& xExtents, const u_int uARGB) { (void)xPosition; (void)xOrientation; (void)xExtents; (void)uARGB; }
inline void Asura_Primitive_Render::DrawCapsuleWorld(const Asura_AVector3& xPosition, const Asura_AMatrix3x3& xOrientation, const Asura_AVector3& xExtents, const u_int uARGB) { (void)xPosition; (void)xOrientation; (void)xExtents; (void)uARGB; }

inline void Asura_Primitive_Render::DrawTextScreen(const Asura_Vector_2& xPosition, const u_int uARGB, const char* const szFmtText, ...) { (void)xPosition; (void)uARGB; (void)szFmtText; }

inline void Asura_Primitive_Render::DrawConsoleText(const char* const szFmtText, ...) { (void)szFmtText; }
inline void Asura_Primitive_Render::DrawConsoleText(const u_int uARGB, const char* const szFmtText, ...) { (void)uARGB, (void)szFmtText; }

inline void Asura_Primitive_Render::Printf(const char* const szFmtText, ...) { (void)szFmtText; }

#endif // #ifdef ASURA_USE_PRIMITIVE_RENDER

#endif //ASURA_PRIMITIVE_RENDER_HH
