//
// Asura_SortedRender.cpp
//
// Created by Richard Rice - 16/12/2002
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Rendering/Asura_SortedRender.h"

// Independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#include "Core/Asura_MaterialResponse_System.h"
#endif // #if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#include "Core/Asura_TextureManagement.h"
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
#include "Core/Handlers/Asura_Handler_System.h"
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
#include "Debugging/Asura_CPUProfile.h"
#include "Entity/Glass/Asura_ClientEntity_Glass.h"
#include "Shapes/Asura_Shapes.h"


// Platform Specific
#include "Debugging/Asura_Assert.h"

// Tools
#ifdef ASURA_TOOLS
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_PlatformRenderObject.h"

#include "Misc/Tools_Render.h"
#endif

#include <stdlib.h> // for qsort

// G L O B A L S /////////////////////////////////////////////////////////////

#if _DEBUG
static u_int g_uMaxSortedRenderStorageUsed = 0;
static int   g_iMaxSortedRenderQueueSize   = 0;
#endif

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
static Asura_Hash_ID g_uContainingZoneHash = ASURA_HASH_ID_UNSET;
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

bool					Asura_SortedRender::s_bActive = false;
int						Asura_SortedRender::s_iQueueSize = 0;
Asura_SortedRenderItem*	Asura_SortedRender::s_apxRenderQueue[ASURA_SORTEDRENDER_QUEUESIZE];
unsigned				Asura_SortedRender::s_uStorageUsed = 0;
char					Asura_SortedRender::s_acRenderItemStorage[ASURA_SORTEDRENDER_STORAGESIZE];

#if defined ASURA_TARGET_SM3
	int					Asura_SortedRender::s_iFirstTransparentItem = -1;
	bool				Asura_SortedRender::s_bTransparentPass = false;
#endif

void Asura_SortedRender::Start()
{
	Asura_Assert(!s_bActive,"");

	s_bActive = true;
	s_iQueueSize = 0;
	s_uStorageUsed = 0;

	#if defined ASURA_TARGET_SM3
		s_iFirstTransparentItem = -1;
		s_bTransparentPass = false;
	#endif
}

void Asura_SortedRender::End()
{
	Asura_Assert(s_bActive,"");
	if(!s_bActive) return;

	//set active to false now , otherwise all the attempts at rendering will just get sent back here again...
	s_bActive = false;

	if(!s_iQueueSize) return;

	qsort( s_apxRenderQueue, s_iQueueSize, sizeof( Asura_SortedRenderItem* ), CompareItems );

	#if defined ASURA_TARGET_SM3
		s_bTransparentPass = false;
		s_iFirstTransparentItem = -1;
	#endif

	Platform_BeginRender();
	{
		for( int i = 0; i < s_iQueueSize; ++i )
		{
#if defined ASURA_TARGET_SM3

			// Record when we hit the first transparent item
			const bool bTransparent = ((s_apxRenderQueue[i]->m_uSortKey & 0x80000000) != 0);
			if( bTransparent && s_iFirstTransparentItem < 0 )
			{
				s_iFirstTransparentItem = i;
			}

#endif // defined ASURA_TARGET_SM3

			s_apxRenderQueue[i]->RenderItem();
		}
	}
	Platform_EndRender();
}

int Asura_SortedRender::CompareItems( const void* pItem1, const void* pItem2 )
{
	const Asura_SortedRenderItem* const pxItem1
		= *static_cast< const Asura_SortedRenderItem* const* >( pItem1 );

	const Asura_SortedRenderItem* const pxItem2
		= *static_cast< const Asura_SortedRenderItem* const* >( pItem2 );

	// Return -ve if item1 < item2, 0 if =, +ve if item1 > item2
	// KOR - we can't get away with a simple cast to signed and subtraction here,
	// because we're using the full unsigned range, so it can wrap around to negative.
	const u_int uSortKey1 = pxItem1->m_uSortKey;
	const u_int uSortKey2 = pxItem2->m_uSortKey;
	return (uSortKey1 < uSortKey2) ? (-1) : ((uSortKey1 == uSortKey2) ? 0 : 1);
}

#if defined ASURA_TARGET_SM3

void Asura_SortedRender::RenderTransparentItems()
{
	// Can't check s_bActive here - it's already been set to false in order for the
	// opaque items to actually render
	if( s_iQueueSize <= 0 || s_iFirstTransparentItem < 0 ) return;

	// This is called from outside Asura_Entity_System, so best do the profiling bracket ourselves
	ASURA_CPU_PROFILE__START_LABEL( ASURA_CPU_PROFILE__ENTITY_RENDER, "Entity Render" );

	s_bTransparentPass = true;

	Platform_BeginRender();
	{
		for( int i = s_iFirstTransparentItem; i < s_iQueueSize; ++i )
		{
			s_apxRenderQueue[i]->RenderItem();
		}
	}
	Platform_EndRender();

	s_bTransparentPass = false;
	s_iFirstTransparentItem = -1;

	ASURA_CPU_PROFILE__STOP( ASURA_CPU_PROFILE__ENTITY_RENDER );
}

#endif // defined ASURA_TARGET_SM3

void* Asura_SortedRender::AllocateSpace(unsigned uSize)
{
	Asura_Assert(uSize>=sizeof(Asura_SortedRenderItem) && uSize < ASURA_SORTEDRENDER_STORAGESIZE,"");

	if((uSize+s_uStorageUsed > ASURA_SORTEDRENDER_STORAGESIZE) || (s_iQueueSize >= ASURA_SORTEDRENDER_QUEUESIZE))
	{
		//out of space , so best render what we have so far , and start queuing more
		Asura_Assert(uSize+s_uStorageUsed <= ASURA_SORTEDRENDER_STORAGESIZE,"Not enough space to store details of all items that need to be sorted.");
		Asura_Assert(s_iQueueSize < ASURA_SORTEDRENDER_QUEUESIZE,"Too many items to sort this frame");

		End();
		Start();
	}

	Asura_SortedRenderItem* pxReturn = (Asura_SortedRenderItem*)&s_acRenderItemStorage[s_uStorageUsed];

	s_uStorageUsed += uSize;
#if _DEBUG
    g_uMaxSortedRenderStorageUsed = Asura_Maths::Max( g_uMaxSortedRenderStorageUsed, s_uStorageUsed );
    g_iMaxSortedRenderQueueSize   = Asura_Maths::Max( g_iMaxSortedRenderQueueSize,   s_iQueueSize   );
#endif
	s_apxRenderQueue[s_iQueueSize++] = pxReturn;

	return pxReturn;
}

///////////////////////////////////
// SetSortedRenderItemZoneHashID //
///////////////////////////////////
// Helper function to set zone hash ID
inline void SetSortedRenderItemZoneHashID(Asura_SortedRenderItem* pxItem)
{
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// MGH 15/5/2006: SetAllZoneData should have been called before the call to Render that got us here.
	pxItem->m_uZoneHashID = Asura_Handler_System::GetLastSetAllZoneDataHash();
	// MGH 2/6/2006: We also need the containing zone for GI lighting.
	pxItem->m_uContainingZoneHashID = g_uContainingZoneHash;
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
}

void Asura_SortedRender::AddShape( Asura_Shape_Instance* pxSI, u_int uCamVisFlags )
{
	Asura_Assert(s_bActive,"");
	if(!s_bActive) return;

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	g_uContainingZoneHash = pxSI->GetContainingZoneFromModuleTest();
	Asura_Assert(g_uContainingZoneHash != ASURA_HASH_ID_UNSET, "Shape containing zone is unset");
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)

    const Asura_Shape* pxShape = pxSI->GetShape();

	const bool bAlpha = ( 1.0f != pxSI->GetAlpha() || !pxShape->IsOpaque() );
	   
#ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	if(bAlpha)
	{
		//For alpha objects we need to add a seperate entry per camera that can see it
		u_int uCamBit = 1 << (Asura_Camera::GetNumberOfCameras() - 1);
		for( int iCam = Asura_Camera::GetNumberOfCameras(); iCam--; uCamBit>>=1 )
		{
			// if the current camera can't see this, ignore
			if(!(uCamBit & uCamVisFlags)) continue;

		    Asura_SortedRenderItem_Shape* pxShapeItem = reinterpret_cast<Asura_SortedRenderItem_Shape*>( AllocateSpace(sizeof(Asura_SortedRenderItem_Shape)) );
			Asura_Assert(pxShapeItem,"");

            pxShapeItem->m_eType        = ASURA_SORTEDRENDERITEM_TYPE_SHAPE;
			pxShapeItem->m_pxSI         = pxSI;
			pxShapeItem->m_uCamVisFlags = uCamBit;

			SetSortedRenderItemZoneHashID(pxShapeItem);

            pxShapeItem->SetupAlphaSortKey( iCam, pxSI->GetPosition() );
		}
	}
    else
#endif // #ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	{
		Asura_SortedRenderItem_Shape* pxShapeItem = reinterpret_cast<Asura_SortedRenderItem_Shape*>( AllocateSpace(sizeof(Asura_SortedRenderItem_Shape)) );
		Asura_Assert(pxShapeItem,"");

        pxShapeItem->m_eType        = ASURA_SORTEDRENDERITEM_TYPE_SHAPE;
		pxShapeItem->m_pxSI         = pxSI;
		pxShapeItem->m_uCamVisFlags = uCamVisFlags;

		SetSortedRenderItemZoneHashID(pxShapeItem);
		
		const int iMatIDx = pxShape->GetMaterialIdx();
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		const Asura_MaterialResponse* const pxMatResponse = Asura_MaterialResponse_System::GetMaterialByHashID( iMatIDx );
		const int iMaterialResponseIndex = (NULL != pxMatResponse) ? pxMatResponse->GetMaterialIndex() : 0;
		const int iSortKey = iMaterialResponseIndex;
#else // #if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
        const int iSortKey = Asura_Material_System::GetTextureID( iMatIDx );
#endif // #if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
        pxShapeItem->SetupSortKey( iSortKey, bAlpha );
	}
}

void Asura_SortedRender::AddAnimation( Asura_Animation* pxAnimation,
                                       const Asura_Vector_3& _xPosition,
                                       const Asura_Matrix_3x3& _xOrient,
                                       u_int uCamVisFlags,
                                       u_int uRenderFlags,
                                       const Asura_Vector_3& _xBSCentre,
                                       float fBSRadius )
{
	Asura_Assert(s_bActive,"");
	if(!s_bActive) return;

    const Asura_Hierarchy_Skin* pxSkin = pxAnimation->GetSkin();
	Asura_Assert( pxSkin, "" );

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	g_uContainingZoneHash = pxAnimation->GetContainingZoneFromModuleTest();
	Asura_Assert(g_uContainingZoneHash != ASURA_HASH_ID_UNSET, "Animation containing zone is unset");
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	
	//Need to take a copy of the position and orientation now , since they are stored in globals
	//that may get changed if there are too many objects and AllocateSpace is forced to render
	//the existing objects now. RAR 17/04/03
	Asura_Vector_3 xPosition = _xPosition;
	Asura_Matrix_3x3 xOrient = _xOrient;
    Asura_Vector_3 xBSCentre = _xBSCentre;

	const bool bAlpha = ( 1.0f != pxAnimation->GetAlpha() || !pxSkin->Platform_IsOpaque() );

    const Asura_SortedRenderItem_Type eType = ( pxSkin->IsSkinnedModel()
                                                    ? ASURA_SORTEDRENDERITEM_TYPE_ANIMATION_SMOOTH_SKIN
                                                    : ASURA_SORTEDRENDERITEM_TYPE_ANIMATION_HIERARCHICAL );

#ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	if(bAlpha)
	{
		//For alpha objects we need to add a seperate entry per camera that can see it
		u_int uCamBit = 1 << (Asura_Camera::GetNumberOfCameras() - 1);
		for( int iCam = Asura_Camera::GetNumberOfCameras(); iCam--; uCamBit>>=1 )
		{
			// if the current camera can't see this, ignore
			if(!(uCamBit & uCamVisFlags)) continue;
			
			// if we're just rendering for the current camera, ignore.
			if( Asura_Hierarchy_System::ShouldJustRenderForCurCam() 
			 && ((u_int)iCam != Asura_Camera::GetCameraIndex()) )
			{
				continue;
			}			

			Asura_SortedRenderItem_Animation* pxAnimItem = (Asura_SortedRenderItem_Animation*) AllocateSpace(sizeof(Asura_SortedRenderItem_Animation));
			Asura_Assert(pxAnimItem,"");
	
            pxAnimItem->m_eType        = eType;
			pxAnimItem->m_pxAnimation  = pxAnimation;
			pxAnimItem->m_xPosition    = xPosition;
			pxAnimItem->m_xOrient      = xOrient;
			pxAnimItem->m_uCamVisFlags = uCamBit;
			pxAnimItem->m_uRenderFlags = uRenderFlags;
    		pxAnimItem->m_xBSCentre    = xBSCentre;
    		pxAnimItem->m_fBSRadius    = fBSRadius;

			SetSortedRenderItemZoneHashID(pxAnimItem);

            pxAnimItem->SetupAlphaSortKey( iCam, xPosition );
		}
	}
    else
#endif //ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	{
		Asura_SortedRenderItem_Animation* pxAnimItem = (Asura_SortedRenderItem_Animation*) AllocateSpace(sizeof(Asura_SortedRenderItem_Animation));
		Asura_Assert(pxAnimItem,"");
	
        pxAnimItem->m_eType        = eType;
		pxAnimItem->m_pxAnimation  = pxAnimation;
		pxAnimItem->m_xPosition    = xPosition;
		pxAnimItem->m_xOrient      = xOrient;
		pxAnimItem->m_uCamVisFlags = uCamVisFlags;
		pxAnimItem->m_uRenderFlags = uRenderFlags;
    	pxAnimItem->m_xBSCentre    = xBSCentre;
    	pxAnimItem->m_fBSRadius    = fBSRadius;

		SetSortedRenderItemZoneHashID(pxAnimItem);

		const int iSortKey = pxSkin->Platform_GetTextureIndex();
		pxAnimItem->SetupSortKey( iSortKey, bAlpha );
	}
}

void Asura_SortedRender::AddGlass( Asura_ClientEntity_Glass* pxGlass, u_int uCamVisFlags )
{
	Asura_Assert(s_bActive,"");
	if(!s_bActive) return;

	Asura_Assert(pxGlass,"");

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	g_uContainingZoneHash = pxGlass->GetContainingZoneFromModuleTest();
	Asura_Assert(g_uContainingZoneHash != ASURA_HASH_ID_UNSET, "Glass containing zone is unset");
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)

#ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	//For alpha objects we need to add a seperate entry per camera that can see it
	u_int uCamBit = 1 << (Asura_Camera::GetNumberOfCameras() - 1);
	for( int iCam = Asura_Camera::GetNumberOfCameras(); iCam--; uCamBit>>=1 )
	{
		// if the current camera can't see this, ignore
		if(!(uCamBit & uCamVisFlags)) continue;

		Asura_SortedRenderItem_Glass* pxGlassItem = (Asura_SortedRenderItem_Glass*) AllocateSpace(sizeof(Asura_SortedRenderItem_Glass));
		Asura_Assert(pxGlassItem,"");
		
        pxGlassItem->m_eType        = ASURA_SORTEDRENDERITEM_TYPE_GLASS;
		pxGlassItem->m_pxGlass      = pxGlass;
		pxGlassItem->m_uCamVisFlags = uCamBit;

		SetSortedRenderItemZoneHashID(pxGlassItem);

        pxGlassItem->SetupAlphaSortKey( iCam, pxGlass->m_xCentre );
	}
#else // not def ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
    Asura_SortedRenderItem_Glass* pxGlassItem = (Asura_SortedRenderItem_Glass*) AllocateSpace(sizeof(Asura_SortedRenderItem_Glass));
	Asura_Assert(pxGlassItem,"");
	
	pxGlassItem->m_eType        = ASURA_SORTEDRENDERITEM_TYPE_GLASS;
	pxGlassItem->m_pxGlass      = pxGlass;
	pxGlassItem->m_uCamVisFlags = uCamVisFlags;

	SetSortedRenderItemZoneHashID(pxGlassItem);

    pxGlassItem->SetupSortKey( pxGlass->GetTextureIndex(), true );
#endif // else not def ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
}

#ifdef ASURA_TOOLS
void Asura_SortedRender::AddToolsMesh(class Asura_Mesh* pxMesh,
										const Asura_Vector_3& xPosition,
										const Asura_Matrix_3x3& xOrientation,
										bool bWithLighting,
										float fAlpha,
										const Asura_Vector_3& xScale,
										int iTextureID)
{
	Asura_Assert(s_bActive,"");
	if(!s_bActive) return;

	const bool bAlpha = true;//RAR 24/07/2006 currently only sending alpha stuff to this function anyway
	   
	
	Asura_SortedRenderItem_ToolsMesh* pxToolsMeshItem = reinterpret_cast<Asura_SortedRenderItem_ToolsMesh*>( AllocateSpace(sizeof(Asura_SortedRenderItem_ToolsMesh)) );
	Asura_Assert(pxToolsMeshItem,"");


	pxToolsMeshItem->m_eType = ASURA_SORTEDRENDERITEM_TYPE_TOOLSMESH;
	pxToolsMeshItem->m_pxMesh = pxMesh;
	pxToolsMeshItem->m_xPosition = xPosition;
	pxToolsMeshItem->m_xOrientation = xOrientation;
	pxToolsMeshItem->m_bWithLighting = bWithLighting;
	pxToolsMeshItem->m_fAlpha = fAlpha;
	pxToolsMeshItem->m_xScale = xScale;
	pxToolsMeshItem->m_iTextureID = iTextureID;

	SetSortedRenderItemZoneHashID(pxToolsMeshItem);

	#ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	if(bAlpha)
	{
		pxToolsMeshItem->SetupAlphaSortKey(0,xPosition);
	}
	else
	#endif
	{
		pxToolsMeshItem->SetupSortKey( Asura_Material_System::GetTextureID(pxMesh->TextureDataPtr[0].m_iMaterialID), bAlpha );
	}
}
#endif //ASURA_TOOLS

#ifdef ASURA_TOOLS
void Asura_SortedRender::AddToolsCuboid(const Asura_Bounding_Box& xBox,
								const Asura_Vector_3& xPosition,
								const Asura_Matrix_3x3& xOrient,
								unsigned int iColour, 
								bool bWireframe,
								int iHighlightSide,
								int iHighlightColour)
{
	Asura_Assert(s_bActive,"");
	if(!s_bActive) return;

	
	Asura_SortedRenderItem_ToolsCuboid* pxToolsCuboidItem = reinterpret_cast<Asura_SortedRenderItem_ToolsCuboid*>( AllocateSpace(sizeof(Asura_SortedRenderItem_ToolsCuboid)) );
	Asura_Assert(pxToolsCuboidItem,"");


	pxToolsCuboidItem->m_eType = ASURA_SORTEDRENDERITEM_TYPE_TOOLSCUBOID;
	pxToolsCuboidItem->m_xBox = xBox;
	pxToolsCuboidItem->m_xPosition = xPosition;
	pxToolsCuboidItem->m_xOrient = xOrient;
	pxToolsCuboidItem->m_iColour = iColour;
	pxToolsCuboidItem->m_bWireframe = bWireframe;
	pxToolsCuboidItem->m_iHighlightSide = iHighlightSide;
	pxToolsCuboidItem->m_iHighlightColour = iHighlightColour;


	SetSortedRenderItemZoneHashID(pxToolsCuboidItem);

	#ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
	//Cuboid rendering doesn't write to z buffer ,so treat them all as being alphaed
	pxToolsCuboidItem->SetupAlphaSortKey(0,xPosition+xBox.CalculateCentre());
	#else
	pxToolsCuboidItem->SetupSortKey( ASURA_TEXTURE_NONE, true );
	#endif
}
#endif //ASURA_TOOLS




#ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE
void Asura_SortedRenderItem::SetupAlphaSortKey( int iCam, const Asura_Vector_3& xPos )
{
	Asura_Camera::SetCameraIndex( iCam );

	float fDistance = (xPos - *Asura_Camera::GetPosition()).Magnitude();
	if(fDistance> 2000000.0f) fDistance = 2000000.0f;

	//fDistance is capped at 2 million , so fDistance*1000 should be less than 0x7fffffff
	//which will mean that the top bit will be set for all alpha objects , in the same way
	//as for GetSortKey
    m_uSortKey = 0xffffffffu - static_cast<u_int>(Asura_Maths::FloatToInt(fDistance * 1000.0f));
}
#endif // #ifdef ASURA_SORTEDRENDER_SORTALPHAONDISTANCE

void Asura_SortedRenderItem::SetupSortKey( int iTexID, bool bAlpha )
{
    // Keep everything unsigned by adding 1. Of course, this assumes the
    // value of ASURA_TEXTURE_NONE to be -1 and that texture IDs are
    // otherwise unsigned.
    Asura_Assert( -1 == ASURA_TEXTURE_NONE && ASURA_TEXTURE_NONE <= iTexID, "" );
    u_int uSortKey = iTexID + 1;

    // Now OR on the item type into the upper 16 bits. Most platforms will
    // have separate render type stuff for the different types of item so it
    // makes sense to minimise changes in that respect as well. Of course,
    // this means duplicating texture uploads if they're used on multiple
    // item types but I think that's unlikely to occur very often in general.
    uSortKey |= Platform_GetSortBias() << 16;

    // Finally, set the top bit if the item is transparent or translucent, to 
    // force it to draw after everything else. This means going through all
    // the renderers again but it's necessary to get things drawing anywhere
    // near correctly.
    if( bAlpha )
	{
		// Alpha textures need to appear after everything else.
        uSortKey |= 0x80000000u;
	}

	m_uSortKey = uSortKey;	
}

void Asura_SortedRenderItem::RenderItem( void ) const
{
	switch(m_eType)
	{
		case ASURA_SORTEDRENDERITEM_TYPE_SHAPE:
			(reinterpret_cast<const Asura_SortedRenderItem_Shape*>(this))->Render();
			break;

		case ASURA_SORTEDRENDERITEM_TYPE_ANIMATION_SMOOTH_SKIN:
        case ASURA_SORTEDRENDERITEM_TYPE_ANIMATION_HIERARCHICAL:
			(reinterpret_cast<const Asura_SortedRenderItem_Animation*>(this))->Render();
			break;

		case ASURA_SORTEDRENDERITEM_TYPE_GLASS:
			(reinterpret_cast<const Asura_SortedRenderItem_Glass*>(this))->Render();
			break;

		#ifdef ASURA_TOOLS
		case ASURA_SORTEDRENDERITEM_TYPE_TOOLSMESH :
			(reinterpret_cast<const Asura_SortedRenderItem_ToolsMesh*>(this))->Render();
			break;

		case ASURA_SORTEDRENDERITEM_TYPE_TOOLSCUBOID :
			(reinterpret_cast<const Asura_SortedRenderItem_ToolsCuboid*>(this))->Render();
			break;
		#endif

		default :
			Asura_Assert( false, "Invalid item type." );
			break;
	}
}


//////////////////////////////////
// SetHandlersForSortedRenderer //
//////////////////////////////////
// Helper function to set handlers for sorted renderer.
inline void SetHandlersForSortedRenderer(const Asura_SortedRenderItem* pxItem)
{
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// MGH 15/5/2006: DataIndex is currently unused.
	Asura_Handler_System::SetHandlersForSortedRenderer(pxItem->m_uZoneHashID, 0, pxItem->m_uContainingZoneHashID);
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
}


void Asura_SortedRenderItem_Shape::Render( void ) const
{
	SetHandlersForSortedRenderer(this);
	m_pxSI->Render( m_uCamVisFlags );
}

void Asura_SortedRenderItem_Animation::Render( void ) const
{
	SetHandlersForSortedRenderer(this);
	Asura_Hierarchy_System::Render( m_pxAnimation, m_xPosition, m_xOrient, m_uCamVisFlags, m_uRenderFlags, m_xBSCentre, m_fBSRadius, true );
}

void Asura_SortedRenderItem_Glass::Render( void ) const
{
	SetHandlersForSortedRenderer(this);
	m_pxGlass->Platform_Render( m_uCamVisFlags );
}


#ifdef ASURA_TOOLS
void Asura_SortedRenderItem_ToolsMesh::Render() const
{
	SetHandlersForSortedRenderer(this);
	Asura_Direct3D_ObjectRenderer::ToolsRender(m_pxMesh,m_xPosition,m_xOrientation,m_bWithLighting,m_fAlpha,m_xScale,m_iTextureID);
}
#endif //ASURA_TOOLS

#ifdef ASURA_TOOLS
void Asura_SortedRenderItem_ToolsCuboid::Render() const
{
	SetHandlersForSortedRenderer(this);
	Tools_Render::RenderCube(m_xBox,m_xPosition,m_xOrient,m_iColour,m_bWireframe,m_iHighlightSide,m_iHighlightColour);
}
#endif //ASURA_TOOLS
