/**
*
* \file Asura_Water.cpp
*
* Created by Robert Dibley - 29/11/2004 14:48:56
*
* Copyright (C) 2004 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"
#include "Rendering/Asura_Water.h"
#include "Debugging/Asura_Assert.h"

#include "Camera/Asura_Camera.h"
#include "Core/Asura_Mempool.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"

#include "Rendering/Asura_PlatformRenderWater.h"

#ifdef ASURA_TOOLS

#include "ChunkSystem/Asura_Chunk_Water.h"
#include "Load&Save/Tools_Level_Export.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_PatchSubdivision.h"
#include "Rendering/Asura_Surface.h"
#include "SpecialFX/Asura_Weather_System.h"

#include "MeshEdit/Tools_Mode_MeshEdit.h"
#include "MeshProcessing/Tools_SubdividedMesh.h"
#include "Misc/Tools_Timed_Messages.h"

#endif

float Asura_Water_System::s_fWaterCycleTime1;
float Asura_Water_System::s_fWaterCycleTime2;
float Asura_Water_System::s_fWaterCycleTime3;

float Asura_Water_System::s_afBasePassUVOffset[2];

Asura_Water_System::Water_Controls* Asura_Water_System::s_pxController;

Asura_TrackingHandlePool<Asura_Water_Packet>* Asura_Water_System::s_pxWaterPacketList;
u_int Asura_Water_System::s_uNumberOfWaterPackets = 0;

Asura_Water_System::Env_Water_Controls Asura_Water_System::s_xEnvController;

DEBUGVAR bool dbg_bWaterSystemEnabled = true;

void Asura_Water_System::Initialise()
{
    Project_SetDefaults( &s_xEnvController );

	//initialise using this controller
	Initialise( &s_xEnvController );
}   

void Asura_Water_System::Initialise( Water_Controls* pxWaveControls )
{
	// set the speed values
	s_pxController = pxWaveControls;

	if(!s_pxWaterPacketList)
	{
		s_pxWaterPacketList = new Asura_TrackingHandlePool<Asura_Water_Packet>(8);
		s_uNumberOfWaterPackets = 0;
	}

    s_afBasePassUVOffset[0] = s_afBasePassUVOffset[1] = 0.0f;

	// initialise version 3 data so that we won't get problems using version 2 saved data
	pxWaveControls->m_fSunYScale = 1.0f;
	pxWaveControls->m_fHorizonFresnelScale = 1.0f;

	Platform_Initialise();
}

void Asura_Water_System::Shutdown()
{
	Platform_Shutdown();

	delete s_pxWaterPacketList;
	s_pxWaterPacketList = 0;
	s_uNumberOfWaterPackets = 0;

}

void Asura_Water_System::Reset()
{
	Platform_Reset();
	
	s_pxWaterPacketList->Reset();
	s_uNumberOfWaterPackets = 0;

    s_afBasePassUVOffset[0] = s_afBasePassUVOffset[1] = 0.0f;
}

void Asura_Water_System::Update( float fFrameTime )
{
	if(!dbg_bWaterSystemEnabled) return;

	// we should never have frame times and cycle scales large enough to let something go a whole loop in one step !
	Asura_Assert( fFrameTime * s_pxController->m_fWaterCycleScale1 < 1.0f, "" );
	Asura_Assert( fFrameTime * s_pxController->m_fWaterCycleScale2 < 1.0f, "" );
	Asura_Assert( fFrameTime * s_pxController->m_fWaterCycleScale3 < 1.0f, "" );
	
	// Hopefully this can be shared across platforms
	s_fWaterCycleTime1 += fFrameTime * s_pxController->m_fWaterCycleScale1;
	if (s_fWaterCycleTime1 > 1.0f) s_fWaterCycleTime1 -= 1.0f;

	s_fWaterCycleTime2 += fFrameTime * s_pxController->m_fWaterCycleScale2;
	if (s_fWaterCycleTime2 > 1.0f) s_fWaterCycleTime2 -= 1.0f;

	s_fWaterCycleTime3 += fFrameTime * s_pxController->m_fWaterCycleScale3;
	if (s_fWaterCycleTime3 > 1.0f) s_fWaterCycleTime3 -= 1.0f;

    for( u_int u = 0; u < 2; ++u )
    {
        s_afBasePassUVOffset[u] += s_pxController->m_afUVSpeeds[u] * fFrameTime;

        // Clamp to (-1, 1).
        s_afBasePassUVOffset[u] -= Asura_Maths::FloatToInt_Chop( s_afBasePassUVOffset[u] );
    }

	float fMaxWaveHeight = ( Asura_Maths::Abs(s_pxController->m_fYScale1) + Asura_Maths::Abs(s_pxController->m_fYScale2) ) * 0.04811f;	// NOTE : this is the peak solution to the polynomial used for the water
	float fMaxLightMult = ( s_pxController->m_fLightOffset  + Asura_Maths::Abs(s_pxController->m_fLightScale) * fMaxWaveHeight );
	s_pxController->m_fLightRescale = Asura_Maths::Min( 1.0f, 1.0f / ( fMaxLightMult )); 
}

void Asura_Water_System::Render()
{
	if(!dbg_bWaterSystemEnabled) return;

	if(!s_pxWaterPacketList)
	{
		Asura_Assert(0,"Not initialised");
		return;
	}
	if(!s_uNumberOfWaterPackets)
	{
		return;
	}

	Platform_BeginRender();

	for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
	{
		Asura_Camera::SetCameraIndex(i);
		Asura_Render::ApplyCameraSettings();

		s_pxWaterPacketList->Traverse(RenderPacket);
	}

	Platform_EndRender();
}

void Asura_Water_System::RenderPacket(Asura_Water_Packet* pxPacket)
{
	if(Asura_Module_System::IsModuleVisible(pxPacket->m_iModuleIndex))
	{
		Platform_Render(pxPacket->m_pxPlatformPacket);
	}	
}

Asura_Water_Packet* Asura_Water_System::AllocateWaterPacket()
{
	if(!s_pxWaterPacketList)
	{
		Asura_Assert(0,"Not initialised");
		return 0;
	}
	Asura_Handle iHandle = s_pxWaterPacketList->Allocate();
	if(iHandle == ASURA_INVALID_HANDLE) return 0;

	s_uNumberOfWaterPackets++;

	return s_pxWaterPacketList->GetPointer(iHandle);
}

float Asura_Water_System::GetWaterCycleTime1(void) 
{
	return(s_fWaterCycleTime1);
}

float Asura_Water_System::GetWaterCycleTime2(void) 
{ 
	return(s_fWaterCycleTime2);
}

float Asura_Water_System::GetWaterCycleTime3(void) 
{ 
	return(s_fWaterCycleTime3);
}

bool Asura_Water_System::IsAnyWaterVisible()
{
	if(!IsAnyWaterPresent())
	{
		return false;
	}

	Asura_TrackingHandlePool_Iterator<Asura_Water_Packet> xIterator = s_pxWaterPacketList->GetIterator();
	for(xIterator.Begin(); !xIterator.End(); xIterator.Next())
	{
		const Asura_Water_Packet* pxPacket = xIterator.GetCurrentPointer();
		if(Asura_Module_System::IsModuleVisible(pxPacket->m_iModuleIndex))
		{
			return true;
		}
	}
	return false;
}

float Asura_Water_System::GetVisibleWaterYPosition()
{
	Asura_TrackingHandlePool_Iterator<Asura_Water_Packet> xIterator = s_pxWaterPacketList->GetIterator();
	for(xIterator.Begin(); !xIterator.End(); xIterator.Next())
	{
		const Asura_Water_Packet* pxPacket = xIterator.GetCurrentPointer();
		if(Asura_Module_System::IsModuleVisible(pxPacket->m_iModuleIndex))
		{
			// return the height of the first packet visible
			const float fYPos = pxPacket->m_fYPosition;
			Asura_Assert(fYPos != Asura_Maths::LargeFloat, "Y position uninitialised");
			return fYPos;
		}
	}
	Asura_Assert(0, "No water visible - should not be calling this function");
	return 0.f;
}

#ifdef ASURA_TOOLS

static bool ExtractWaterPolys(Asura_Mesh* pxMesh, int iIndex)
{
	int iFlags = pxMesh->PolygonFlagDataPtr[iIndex].Flags;
	if(iFlags & AsuraPolygonFlag_DontExport) return false;
	return ((iFlags & AsuraPolygonFlag_Water) != 0);
}

static bool ExtractQuads(Asura_Mesh* pxMesh, int iIndex)
{
	return pxMesh->PolygonDataPtr[iIndex].IsQuad();
}

static bool ExtractTriangles(Asura_Mesh* pxMesh, int iIndex)
{
	return !pxMesh->PolygonDataPtr[iIndex].IsQuad();
}

void Asura_Water_System::BuildWaterPackets(Asura_SaveBlock& xSaveBlock, bool bShouldSubdivide)
{

	//get the sun direction to help with deciding how to split quads
	const Env_Water_Controls& xControl = GetEnvController();
	Asura_Vector_3 xSunDir;
	if(xControl.m_bUseLevelSun)
	{
		xSunDir = Asura_Weather_System::GetSunPos();
		xSunDir.y = 0.0f;
	}
	else
	{
		xSunDir.x = xControl.m_afSunDir[0];
		xSunDir.y = 0.0f;
		xSunDir.z = xControl.m_afSunDir[2];
	}

	if(bShouldSubdivide)
	{
		Asura_PatchSubdivision::SubdivideNearWaypoints();
		Asura_PatchSubdivision::SetSubdivideFlat(true);
		Asura_PatchSubdivision::SetAvoidTriangleSlivers(false);
		Asura_PatchSubdivision::SetPointSearchInitialDistance(1000.f);
		Asura_PatchSubdivision::SetPointSearchMinimumDistance(0.5f);
		Asura_PatchSubdivision::SetFlatSubdivideSize(0.0125f);
		Asura_PatchSubdivision::SetCurveSubdivideSize(10000.f);
		Asura_PatchSubdivision::SetDistancePower(1.5f);
	}

	Asura_Mesh *pxWaterMesh = Asura_Mesh_System::GetMesh("tempWaterExportMesh", true);
	Asura_Mesh* pxQuadMesh = Asura_Mesh_System::GetMesh("tempWaterExportMeshQuads", true);

	for (int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh *pxMesh = Asura_Module_System::GetMeshForModule(i);
		if(!pxMesh->PolygonFlagDataValid) continue;
		if(!pxMesh->VertexDataValid) continue;
		if(!pxMesh->TextureDataValid) continue;

		// extract the water polygons from this module into the temporary mesh
		Tools_Mode_MeshEdit::CopyMesh(pxMesh, pxWaterMesh, ExtractWaterPolys, Tools_Mode_MeshEdit::CMNoPolyFlags 
																			| Tools_Mode_MeshEdit::CMNoPolyGuids 
																			| Tools_Mode_MeshEdit::CMNoPolyLights);
		
		if(!pxWaterMesh->NumberOfPolygons) continue;

		if(bShouldSubdivide)
		{
			Asura_Assertf( true == pxWaterMesh->IsMeshWellFormed(), ("Mesh '%s' is not well formed. Added mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
		
			Tools_Mode_MeshEdit::WeldVertices(pxWaterMesh);

			Asura_Assertf( true == pxWaterMesh->IsMeshWellFormed(), ("Mesh '%s' is not well formed. Added mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
		
			Asura_Vector_3 xModulePosition;
			Asura_Module_System::GetPositionOfModule(i, xModulePosition);
			Asura_PatchSubdivision::FindPointsNearModule(pxWaterMesh, xModulePosition);

			// subdivision's going to require smoothing group data
			// generate some
			if(!pxWaterMesh->SmoothingGroupDataValid)
			{
				pxWaterMesh->SmoothingGroupDataPtr = new int[pxWaterMesh->NumberOfPolygons];
				// set the top bit on each poly to indicate 'not curved'
				for(int u = 0; u < pxWaterMesh->NumberOfPolygons; ++u)
				{
					//pxWaterMesh->SmoothingGroupDataPtr[u] = 0x80000000;
					pxWaterMesh->SmoothingGroupDataPtr[u] = 0x00000000;
				}
				pxWaterMesh->SmoothingGroupDataValid = true;
			}
			Tools_SubdividedMesh* pxSubdividedMesh = new Tools_SubdividedMesh(pxWaterMesh, xModulePosition);
			pxSubdividedMesh->ConvertToMesh(pxWaterMesh, false);
			delete pxSubdividedMesh;

			Asura_Surface_Mesh_System::GetSurfaceMesh(pxWaterMesh->Index)->Invalidate();

			Asura_Assertf( true == pxWaterMesh->IsMeshWellFormed(), ("Mesh '%s' is not well formed. Added mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
			Tools_Mode_MeshEdit::WeldVertices(pxWaterMesh);
//			Tools_Mode_MeshEdit::DeleteDegenerateTriangles( pxWaterMesh);

			Asura_Assertf( true == pxWaterMesh->IsMeshWellFormed(), ("Mesh '%s' is not well formed. Added mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
		
			// merge polys back into quads
			List<int> l;
			for(int u = 0; u < pxWaterMesh->NumberOfPolygons; ++u)
			{
				l.add_entry(u);	
			}

			// merge requires smoothing groups... again...
			if(!pxWaterMesh->SmoothingGroupDataValid)
			{
				pxWaterMesh->SmoothingGroupDataPtr = new int[pxWaterMesh->NumberOfPolygons];
				// set the top bit on each poly to indicate 'not curved'
				for(int u = 0; u < pxWaterMesh->NumberOfPolygons; ++u)
				{
					//pxWaterMesh->SmoothingGroupDataPtr[u] = 0x80000000;
					pxWaterMesh->SmoothingGroupDataPtr[u] = 0x00000000;
				}
				pxWaterMesh->SmoothingGroupDataValid = true;
			}

			int* piPolyConv;
			Tools_Mode_MeshEdit::MergePolygons(pxWaterMesh, l, &piPolyConv);
			delete[] piPolyConv;

		Asura_Assertf( true == pxWaterMesh->IsMeshWellFormed(), ("Mesh '%s' is not well formed. Added mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
			/*if(!pxWaterMesh->PrelightingDataValid)
			{
				Asura_Lighting_System::CalculatePrelighting(*pxWaterMesh, xModulePosition, i);
			}*/
		}

		Asura_Assertf( true == pxWaterMesh->IsMeshWellFormed(), ("Mesh '%s' is not well formed. Added mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
		
		// Extract any quads from the water mesh
		Tools_Mode_MeshEdit::CopyMesh(pxWaterMesh, pxQuadMesh, ExtractQuads, Tools_Mode_MeshEdit::CMNoSmoothingGroups);

		// rewrite the polygons in this quad mesh to split all the quads aligned correctly against the sun
		if(pxQuadMesh->NumberOfPolygons)
		{
			Asura_Mesh_Polygon* pxNewPolyData = new Asura_Mesh_Polygon[pxQuadMesh->NumberOfPolygons * 2];
			Asura_Mesh_TextureInfo* pxNewTexData = new Asura_Mesh_TextureInfo[pxQuadMesh->NumberOfPolygons * 2];
			Asura_Mesh_Prelighting* pxNewPrelightingData = new Asura_Mesh_Prelighting[pxQuadMesh->NumberOfPolygons * 2];
			Asura_Mesh_PaintedColours* pxNewPaintedColourData = new Asura_Mesh_PaintedColours[ pxQuadMesh->NumberOfPolygons * 2 ];

			for(int u = 0; u < pxQuadMesh->NumberOfPolygons; ++u)
			{
				Asura_Mesh_Polygon* pxPoly = &pxQuadMesh->PolygonDataPtr[u];

				//need to pick the split direction ,so that its as close to perpendicular
				//to the sun as possible
				Asura_Vector_3 xDir1 = pxQuadMesh->VertexDataPtr[pxPoly->VertexID[2]].Position
									 - pxQuadMesh->VertexDataPtr[pxPoly->VertexID[0]].Position;
	
				Asura_Vector_3 xDir2 = pxQuadMesh->VertexDataPtr[pxPoly->VertexID[3]].Position
									 - pxQuadMesh->VertexDataPtr[pxPoly->VertexID[1]].Position;
	
				float fDot1 = Asura_Maths::Abs(xSunDir*xDir1);
				float fDot2 = Asura_Maths::Abs(xSunDir*xDir2);

				int iIndices[6];
				if( fDot1 < fDot2 )
				{
					// split from 0 to 2
					iIndices[0] = 0;
					iIndices[1] = 1;
					iIndices[2] = 2;
					iIndices[3] = 0;
					iIndices[4] = 2;
					iIndices[5] = 3;
				}
				else
				{
					// split from 1 to 3
					iIndices[0] = 0;
					iIndices[1] = 1;
					iIndices[2] = 3;
					iIndices[3] = 1;
					iIndices[4] = 2;
					iIndices[5] = 3;
				}
				
				// write out the new triangles

				pxNewPolyData[u*2].VertexID[0] = pxPoly->VertexID[iIndices[0]];
				pxNewPolyData[u*2].VertexID[1] = pxPoly->VertexID[iIndices[1]];
				pxNewPolyData[u*2].VertexID[2] = pxPoly->VertexID[iIndices[2]];
				pxNewPolyData[u*2].VertexID[3] = 0xFFFF;

				pxNewPolyData[u*2+1].VertexID[0] = pxPoly->VertexID[iIndices[3]];
				pxNewPolyData[u*2+1].VertexID[1] = pxPoly->VertexID[iIndices[4]];
				pxNewPolyData[u*2+1].VertexID[2] = pxPoly->VertexID[iIndices[5]];
				pxNewPolyData[u*2+1].VertexID[3] = 0xFFFF;
				
				// texturing
				pxNewTexData[u*2].m_iMaterialID = pxQuadMesh->TextureDataPtr[u].m_iMaterialID;
				pxNewTexData[u*2].m_afAlpha[0] = pxQuadMesh->TextureDataPtr[u].m_afAlpha[iIndices[0]];
				pxNewTexData[u*2].m_afAlpha[1] = pxQuadMesh->TextureDataPtr[u].m_afAlpha[iIndices[1]];
				pxNewTexData[u*2].m_afAlpha[2] = pxQuadMesh->TextureDataPtr[u].m_afAlpha[iIndices[2]];
				for(u_int uTex = 0; uTex < ASURA_MATERIAL_MAXTEXTURES; ++uTex)
				{
					pxNewTexData[u*2].U[uTex][0] = pxQuadMesh->TextureDataPtr[u].U[uTex][iIndices[0]];
					pxNewTexData[u*2].U[uTex][1] = pxQuadMesh->TextureDataPtr[u].U[uTex][iIndices[1]];
					pxNewTexData[u*2].U[uTex][2] = pxQuadMesh->TextureDataPtr[u].U[uTex][iIndices[2]];

					pxNewTexData[u*2].V[uTex][0] = pxQuadMesh->TextureDataPtr[u].V[uTex][iIndices[0]];
					pxNewTexData[u*2].V[uTex][1] = pxQuadMesh->TextureDataPtr[u].V[uTex][iIndices[1]];
					pxNewTexData[u*2].V[uTex][2] = pxQuadMesh->TextureDataPtr[u].V[uTex][iIndices[2]];
				}

				pxNewTexData[u*2+1].m_iMaterialID = pxQuadMesh->TextureDataPtr[u].m_iMaterialID;
				pxNewTexData[u*2+1].m_afAlpha[0] = pxQuadMesh->TextureDataPtr[u].m_afAlpha[iIndices[3]];
				pxNewTexData[u*2+1].m_afAlpha[1] = pxQuadMesh->TextureDataPtr[u].m_afAlpha[iIndices[4]];
				pxNewTexData[u*2+1].m_afAlpha[2] = pxQuadMesh->TextureDataPtr[u].m_afAlpha[iIndices[5]];
				for(u_int uTex = 0; uTex < ASURA_MATERIAL_MAXTEXTURES; ++uTex)
				{
					pxNewTexData[u*2+1].U[uTex][0] = pxQuadMesh->TextureDataPtr[u].U[uTex][iIndices[3]];
					pxNewTexData[u*2+1].U[uTex][1] = pxQuadMesh->TextureDataPtr[u].U[uTex][iIndices[4]];
					pxNewTexData[u*2+1].U[uTex][2] = pxQuadMesh->TextureDataPtr[u].U[uTex][iIndices[5]];

					pxNewTexData[u*2+1].V[uTex][0] = pxQuadMesh->TextureDataPtr[u].V[uTex][iIndices[3]];
					pxNewTexData[u*2+1].V[uTex][1] = pxQuadMesh->TextureDataPtr[u].V[uTex][iIndices[4]];
					pxNewTexData[u*2+1].V[uTex][2] = pxQuadMesh->TextureDataPtr[u].V[uTex][iIndices[5]];
				}

				if( NULL != pxNewPrelightingData && NULL != pxQuadMesh->PrelightingDataPtr)
				{
					// prelighting
					pxNewPrelightingData[u*2].Colour[0] = pxQuadMesh->PrelightingDataPtr[u].Colour[iIndices[0]];
					pxNewPrelightingData[u*2].Colour[1] = pxQuadMesh->PrelightingDataPtr[u].Colour[iIndices[1]];
					pxNewPrelightingData[u*2].Colour[2] = pxQuadMesh->PrelightingDataPtr[u].Colour[iIndices[2]];

					pxNewPrelightingData[u*2+1].Colour[0] = pxQuadMesh->PrelightingDataPtr[u].Colour[iIndices[3]];
					pxNewPrelightingData[u*2+1].Colour[1] = pxQuadMesh->PrelightingDataPtr[u].Colour[iIndices[4]];
					pxNewPrelightingData[u*2+1].Colour[2] = pxQuadMesh->PrelightingDataPtr[u].Colour[iIndices[5]];
				}

				{
					// painted data
					pxNewPaintedColourData[ u * 2 ].Colour[ 0 ] = pxQuadMesh->m_pxPaintedColourDataPtr[ u ].Colour[ iIndices[ 0 ] ];
					pxNewPaintedColourData[ u * 2 ].Colour[ 1 ] = pxQuadMesh->m_pxPaintedColourDataPtr[ u ].Colour[ iIndices[ 1 ] ];
					pxNewPaintedColourData[ u * 2 ].Colour[ 2 ] = pxQuadMesh->m_pxPaintedColourDataPtr[ u ].Colour[ iIndices[ 2 ] ];
 
					pxNewPaintedColourData[ u * 2 + 1 ].Colour[ 0 ] = pxQuadMesh->m_pxPaintedColourDataPtr[ u ].Colour[ iIndices[ 3 ] ];
					pxNewPaintedColourData[ u * 2 + 1 ].Colour[ 1 ] = pxQuadMesh->m_pxPaintedColourDataPtr[ u ].Colour[ iIndices[ 4 ] ];
					pxNewPaintedColourData[ u * 2 + 1 ].Colour[ 2 ] = pxQuadMesh->m_pxPaintedColourDataPtr[ u ].Colour[ iIndices[ 5 ] ];
				}
			}

			delete[] pxQuadMesh->PolygonDataPtr;			pxQuadMesh->PolygonDataPtr = pxNewPolyData;
			delete[] pxQuadMesh->TextureDataPtr;			pxQuadMesh->TextureDataPtr = pxNewTexData;
			delete[] pxQuadMesh->PrelightingDataPtr;		pxQuadMesh->PrelightingDataPtr = pxNewPrelightingData;
			delete[] pxQuadMesh->m_pxPaintedColourDataPtr;		pxQuadMesh->m_pxPaintedColourDataPtr = pxNewPaintedColourData;

			pxQuadMesh->NumberOfPolygons *= 2;
			Asura_Assert(pxQuadMesh->NumberOfPolygons == pxQuadMesh->NumberOfTriangles, "Didn't split all quads");

			// delete all but quads from the original watermesh
			Tools_Mode_MeshEdit::CopyMesh(pxWaterMesh, pxWaterMesh, ExtractTriangles, Tools_Mode_MeshEdit::CMNoSmoothingGroups
																					| Tools_Mode_MeshEdit::CMNoPolyFlags 
																					| Tools_Mode_MeshEdit::CMNoPolyGuids 
																					| Tools_Mode_MeshEdit::CMNoPolyLights );

			// merge these new triangles back into the watermesh
			Tools_Mode_MeshEdit::MergeMeshes(pxWaterMesh, pxWaterMesh, pxQuadMesh, Tools_Mode_MeshEdit::CMNoSmoothingGroups
																					| Tools_Mode_MeshEdit::CMNoPolyFlags 
																					| Tools_Mode_MeshEdit::CMNoPolyGuids 
																					| Tools_Mode_MeshEdit::CMNoPolyLights );

			Tools_Mode_MeshEdit::WeldVertices(pxWaterMesh);
		}

		// Copy the triangles through

		{
			int iNumVerts = (pxWaterMesh->NumberOfTriangles)*3;
			Asura_Assertf( iNumVerts <= 0xFFFF, ("There's an awful lot of vertices (%d) in this packet... Mesh '%s' after adding Mesh '%s'", iNumVerts, pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));
			// allocate temporary storage for the verts
			Asura_Chunk_WaterPacket_Vertex* pxVertArray = new Asura_Chunk_WaterPacket_Vertex[iNumVerts];

			bool bWarnedAboutUntextured = false;
			bool bWarnedAboutUnmatchingTexture = false;

			// copy out the polys as a triangle list
			Asura_Chunk_WaterPacket_Vertex* pxVert = pxVertArray;
			for(int j = 0; j < pxWaterMesh->NumberOfPolygons; j++)
			{
				const Asura_Mesh_Polygon* pxPoly = &pxWaterMesh->PolygonDataPtr[j];
	
				int iVertsToOutput;
				int aiVert[6];
	
				if(pxPoly->IsQuad())	 
				{
					Asura_Assertf(false, ("Should never happen. Mesh '%s' after adding Mesh '%s'", pxWaterMesh->Name.GetString(), pxMesh->Name.GetString()));

					/*
					//need to pick the split direction ,so that its as close to perpendicular
					//to the sun as possible
					Asura_Vector_3 xDir1 = pxWaterMesh->VertexDataPtr[pxPoly->VertexID[2]].Position
										 - pxWaterMesh->VertexDataPtr[pxPoly->VertexID[0]].Position;
	
					Asura_Vector_3 xDir2 = pxWaterMesh->VertexDataPtr[pxPoly->VertexID[3]].Position
										 - pxWaterMesh->VertexDataPtr[pxPoly->VertexID[1]].Position;
	
					float fDot1 = Asura_Maths::Abs(xSunDir*xDir1);
					float fDot2 = Asura_Maths::Abs(xSunDir*xDir2);
	
					
					iVertsToOutput = 6;
					if(fDot1<fDot2)
					{
						//split from vert 0 to 2
						aiVert[0] = 0;
						aiVert[1] = 1;
						aiVert[2] = 2;
						aiVert[3] = 0;
						aiVert[4] = 2;
						aiVert[5] = 3;
					}
					else
					{
						//split from vert 1 to 3
						aiVert[0] = 0;
						aiVert[1] = 1;
						aiVert[2] = 3;
						aiVert[3] = 1;
						aiVert[4] = 2;
						aiVert[5] = 3;
					}*/

				}
				else
				{
					iVertsToOutput = 3;
					aiVert[0] = 0;
					aiVert[1] = 1;
					aiVert[2] = 2;
				}

				for(int k = 0; k < iVertsToOutput; k++, pxVert++)
				{
					int iVert = aiVert[k];
					
					//position
					Asura_Module_System::GetPositionOfModule(i, pxVert->m_xPosition);
					pxVert->m_xPosition += pxWaterMesh->VertexDataPtr[pxPoly->VertexID[iVert]].Position;
					
					// colour
	                // [0,255] range, with values > 128 being overbrightening
					if(pxWaterMesh->PrelightingDataValid)
					{
						u_int uColour = pxWaterMesh->PrelightingDataPtr[j].Colour[iVert];
	
						u_int uRed   = (uColour >> 16) & 0xff;
						u_int uGreen = (uColour >>  8) & 0xff;
						u_int uBlue  = (uColour >>  0) & 0xff;
	
						pxVert->m_fRed   = static_cast<float>( uRed   );
						pxVert->m_fGreen = static_cast<float>( uGreen );
						pxVert->m_fBlue  = static_cast<float>( uBlue  );
					}
					else
					{
						pxVert->m_fRed   = 128.0f;
						pxVert->m_fGreen = 128.0f;
						pxVert->m_fBlue  = 128.0f;
					}
	
					//uvs
					Asura_Mesh_TextureInfo* pxTex = &pxWaterMesh->TextureDataPtr[j];
	
					if (pxTex->m_iMaterialID == ASURA_MATERIAL_NONE && !bWarnedAboutUntextured)
					{
						Tools_Level_Export::AddMessage( Tools_Level_Export::MT_WARNING, "Untextured water polygons found in %s. These won't have correct UVs",pxMesh->Name.GetString());
						bWarnedAboutUntextured = true;
					}
					else if (pxTex->m_iMaterialID != ASURA_MATERIAL_NONE &&
							Asura_Material_System::GetTextureID(pxTex->m_iMaterialID) != xControl.m_aiTextureID[0] &&
							!bWarnedAboutUnmatchingTexture)
					{
						bWarnedAboutUnmatchingTexture = true;
						Tools_Level_Export::AddMessage( Tools_Level_Export::MT_WARNING, "Water polygons in %s don't match surface texture. UV coordinates might get screwed up.", pxMesh->Name.GetString() );
					}
	
					pxVert->m_fU = pxTex->U[0][iVert];
					pxVert->m_fV = pxTex->V[0][iVert];
				}
			}

			Asura_Assertf(pxVertArray+iNumVerts == pxVert, ("Wrong number of verts output. Mesh '%s'", pxMesh->Name.GetString()));

			//create chunk for this module's water polygons
			Asura_Chunk_WaterPacket::Create(&xSaveBlock,i,iNumVerts,pxVertArray);
			delete [] pxVertArray;
		}
	}

	Asura_Mesh_System::DeleteMesh(pxWaterMesh);
	Asura_Mesh_System::DeleteMesh(pxQuadMesh);

	if(bShouldSubdivide)
	{
		Asura_PatchSubdivision::SetPointSearchInitialDistance(100.f);
		Asura_PatchSubdivision::SetPointSearchMinimumDistance(5.f);
		Asura_PatchSubdivision::SetDistancePower(1.f);
		Asura_PatchSubdivision::SetAvoidTriangleSlivers(true);
	}
}

#endif

/////////////////////////////////////////////////////////////////////////////////////

Asura_Water_Packet::Asura_Water_Packet()
: m_iModuleIndex(-1)
, m_fYPosition(Asura_Maths::LargeFloat)
, m_pxPlatformPacket(NULL)
{
}

Asura_Water_Packet::~Asura_Water_Packet()
{
	delete m_pxPlatformPacket;
}
