//////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_CSM

#include "Shadows/Asura_CSM.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Frustum.h"
#include "Rendering/Asura_Render.h"

#if defined ASURA_TOOLS
	#include "Interface/Asura_Dialog.h"
	#include "Interface/Asura_Overlay.h"
#endif

/******************************************************************************/
/* Asura_CSM Static Class Data                                                */
/******************************************************************************/

bool Asura_CSM::s_bEnabled = false;

// Push the CSM frustum near plane back by this much to catch shadow casters behind the camera
float Asura_CSM::s_fExtraFrustumDepth = 500.0f;

Asura_Matrix_3x3 Asura_CSM::s_axShadowViewMatrixOrients[ ASURA_CSM_MAX_MAPS ];
Asura_Vector_3 Asura_CSM::s_axShadowViewMatrixTrans[ ASURA_CSM_MAX_MAPS ];

u_int Asura_CSM::s_uNumMaps = ASURA_CSM_MAX_MAPS;

Asura_FixedArray< float, ASURA_CSM_MAX_MAPS + 1 > Asura_CSM::s_afSplits;
Asura_FixedArray< float, ASURA_CSM_MAX_MAPS > Asura_CSM::s_afStrengths;
Asura_FixedArray< float, ASURA_CSM_MAX_MAPS > Asura_CSM::s_afBSRadii;
Asura_FixedArray< Asura_Vector_3, ASURA_CSM_MAX_MAPS > Asura_CSM::s_axBSCentres;

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

// These don't need to be exposed to the platform

static Asura_Vector_3 g_xLightDir( .7f, .7f, .7f );

DVSVAR bool g_bFreezeUpdates = false;
DVSVAR bool g_bShowFrusta = false;

static float g_fSplitLambda = 0.95f;

/******************************************************************************/
/* Asura_CSM Static Class Functions                                           */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Initialise()
//
// Initialises the CSM system. Called at program startup
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Initialise()
{
	ResetSplits();

	for( u_int u = 0; u < ASURA_CSM_MAX_MAPS; ++u )
	{
		s_afStrengths[u] = 0.5f;
	}

	#if defined ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_CommandConsole::AddVar( L"Render.CSM.NumMaps", &s_uNumMaps, 1, ASURA_CSM_MAX_MAPS,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.FreezeUpdates", &g_bFreezeUpdates,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.ShowFrusta", &g_bShowFrusta,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		Asura_CommandConsole::AddVar( L"Render.CSM.ExtraFrustumDepth", &s_fExtraFrustumDepth, 0.0f,
			FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

		// NB one more split than num maps, but don't let them modify split 0; that's the near plane
		for( u_int u = 1; u <= ASURA_CSM_MAX_MAPS; ++u )
		{
			Asura_UnicodeChar wszName[24];
			Asura_PrintfToBfr( wszName, L"Render.CSM.Split%u", 24, true, u );

			Asura_CommandConsole::AddVar( wszName, &s_afSplits[u], 0.0f, FLT_MAX,
				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		}

		for( u_int u = 0; u < ASURA_CSM_MAX_MAPS; ++u )
		{
			Asura_UnicodeChar wszName[24];
			Asura_PrintfToBfr( wszName, L"Render.CSM.Strength%u", 24, true, u );

			Asura_CommandConsole::AddVar( wszName, &s_afStrengths[u], 0.0f, FLT_MAX,
				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		}
	#endif

	Platform_Initialise();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Reset()
//
// Resets the CSM system. Called between levels
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Reset()
{
	Platform_Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Shutdown()
//
// Shuts down the CSM system. Called at program termination
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Shutdown()
{
	Platform_Shutdown();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::SetShadowDir()
//
// Sets the current shadow direction for CSM
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::SetShadowDir( const Asura_Vector_3& xDir )
{
	g_xLightDir = xDir;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::GetShadowDir()
//
// Retrieves the current CSM sunlight direction
//
////////////////////////////////////////////////////////////////////////////////

const Asura_Vector_3& Asura_CSM::GetShadowDir()
{
	return g_xLightDir;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::Update()
//
// Updates the CSM system. Called once per frame
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::Update()
{
	Asura_Camera::SetCameraIndex(0);

	static Asura_Vector_3 xCamPos = *Asura_Camera::GetPosition();
	static Asura_Matrix_3x3 xCamOri = *Asura_Camera::GetOrientation();

	if( !g_bFreezeUpdates )
	{
		xCamPos = *Asura_Camera::GetPosition();
		xCamOri = *Asura_Camera::GetOrientation();
	}

	Asura_Frustum axFrusta[ ASURA_CSM_MAX_MAPS ];

	// Chop the camera view frustum up into portions
	for( u_int u = 0; u < s_uNumMaps; ++u )
	{
		const float fTanHalfHorzFOV = 1.0f / Asura_Camera::GetOneOverTanHalfFovX();
		const float fTanHalfVertFOV = 1.0f / Asura_Camera::GetOneOverTanHalfFovY();

		axFrusta[u].SetForFOV( s_afSplits[u], s_afSplits[u+1], fTanHalfHorzFOV, fTanHalfVertFOV );

		s_afBSRadii[u] = axFrusta[u].GetBoundingSphere( s_axBSCentres[u], xCamOri, xCamPos );

		// Construct shadow view matrix
		{
			g_xLightDir.Normalise();

			s_axShadowViewMatrixTrans[u] = g_xLightDir;
			s_axShadowViewMatrixTrans[u] *= s_afBSRadii[u];
			s_axShadowViewMatrixTrans[u] -= s_axBSCentres[u];

			Asura_Maths::OrientationFromDirection( s_axShadowViewMatrixOrients[u], g_xLightDir );
		}
	}

	#if !defined ASURA_NO_DEBUG_OUTPUT

		if( g_bShowFrusta )
		{
			for( u_int u = 0; u < s_uNumMaps; ++u )
			{
				Asura_RGBA xColour( 0xFF8080, 0 );
				xColour.m_fA = 1.0f - static_cast< float >(u+1) / (s_uNumMaps+1);

				Asura_DebugLineRender::AddFrustum( axFrusta[u], xCamPos, xCamOri,
					xColour.ToUIntARGB() );

				xColour.m_fR = 0.5f;
				xColour.m_fB = 1.0f;

				Asura_DebugLineRender::AddSphere( s_axBSCentres[u], s_afBSRadii[u],
					xColour.ToUIntARGB() );
			}
		}

	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::CreateShadowMaps()
//
// Called early in the frame to construct the cascaded shadow maps
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::CreateShadowMaps()
{
	if( IsEnabled() && Asura_Render::IsSystemEnabled( ASURA_RENDER_SYSTEM_CSM ) )
	{
		Platform_CreateShadowMaps();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::ApplyToScene()
//
// Called once per frame to apply the CSMs onto the scene. This may change so
// it's incorporated directly into the prelighting or new "sunlighting" pass
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::ApplyToScene()
{
	if( IsEnabled() && Asura_Render::IsSystemEnabled( ASURA_RENDER_SYSTEM_CSM ) )
	{
		Platform_ApplyToScene();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::ResetSplits()
//
// Resets split values based on Zhang's 'practical split algorithm'. See GPU
// Gems 3 PSSM article etc. for details
//
////////////////////////////////////////////////////////////////////////////////

void Asura_CSM::ResetSplits()
{
	const float fNear = Asura_Camera::GetNearClip();
	const float fFar = Asura_Camera::GetFarClip();

	for( u_int u = 0; u <= s_uNumMaps; ++u )
	{
		const float fMap =  static_cast< float >(u) / s_uNumMaps;

		const float fA = fNear + fMap * (fFar - fNear);
		const float fB = fNear * Asura_Maths::Power( fFar / fNear, fMap );

		s_afSplits[u] = Asura_Maths::Lerp( fA, fB, g_fSplitLambda );
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_CSM::AddOverlayControls()
//
// Adds controls for CSM to the Weather mode overlay in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

#if defined ASURA_TOOLS

void Asura_CSM::AddOverlayControls( Asura_Overlay* const pxOverlay, Asura_Vector_2& xOffset,
	bool& bModified )
{
	Asura_Dialog* pxDialog;
	const Asura_Vector_2 xStartOffset = xOffset;

	// 'Enable' tickbox
	pxDialog = new Asura_Tickbox_Dialog( "Enable", xOffset, &s_bEnabled, &bModified );
	pxDialog->SetHelpText( "Enables cascaded shadow mapping." );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Tickbox_Dialog::DefaultSpacing();

	// Reset splits button
	pxDialog = new Asura_Function_Dialog( "Reset splits", xOffset, ResetSplits, NULL, IsEnabled );
	pxDialog->SetHelpText( "Automatically resets split distances based on current camera far clip plane and 'lambda' tweak factor." );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Lambda tweak value
	pxDialog = new Asura_Numerical_Dialog( "Lambda", xOffset, &g_fSplitLambda, 0.0f, 1.0f, 0.01f,
		"%.2f" );
	pxDialog->SetHelpText( "A tweak factor used when automatically choosing split points." );
	pxDialog->SetIsEnabledFunction( IsEnabled );
	pxOverlay->AddDialog( pxDialog );

	xOffset.x = 16.0f;
	xOffset.y = xStartOffset.y;

	// Split 0 is the near plane, don't allow them to modify that

	for( u_int u = 1; u <= ASURA_CSM_MAX_MAPS; ++u )
	{
		char szLabel[10];
		Asura_PrintfToBfr( szLabel, "Split %u", 10, true, u );

		// Split distance
		pxDialog = new Asura_Numerical_Dialog( szLabel, xOffset, &s_afSplits[u], 0.0f, 20000.0f,
			0.1f, "%.1f", &bModified );
		pxDialog->SetHelpText( "Sets the distance in metres to this split point." );
		pxDialog->SetIsEnabledFunction( IsEnabled );
		pxOverlay->AddDialog( pxDialog );

		const float fOldX = xOffset.x;
		xOffset.x = pxDialog->m_xMaxCoord.x + 3.0f;

		// Map strength
		pxDialog = new Asura_Numerical_Dialog( "Strength", xOffset, &s_afStrengths[ u - 1 ], 0.0f,
			10.0f, 0.01f, "%.2f", &bModified );
		pxDialog->SetHelpText( "Sets the strength of the cascaded shadows in this split." );
		pxDialog->SetIsEnabledFunction( IsEnabled );
		pxOverlay->AddDialog( pxDialog );

		xOffset.x = fOldX;
		xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
	}

	xOffset.x = xStartOffset.x;
}

#endif // defined ASURA_TOOLS

#endif // defined ASURA_USE_CSM
