
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Shadow_Instance.h
// Author : Andy Tate
// Date   : 28-Nov-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_SHADOW_INSTANCE_H__
#define __ASURA_SHADOW_INSTANCE_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Frustum.h"
//#include "Shadows/Asura_Shadow_System.h"

////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_Animation;
class Asura_Shadow;
class Asura_Shape_Instance;

////////////////////////////////////////////////////////////////////////////////
// Typedefs/Enums
////////////////////////////////////////////////////////////////////////////////

// We now want shadows that come from specific lights. Each shadow instance will be one of these:
enum ASURA_SHADOW_LIGHTSOURCE
{
	ASURA_SHADOW_LIGHTSOURCE_AVERAGED = 0, // Shadow is from an averaged direction, based on nearby lightsources
	ASURA_SHADOW_LIGHTSOURCE_PHOTON,       // Shadow comes from a photon (point light). Shadow masks out the light
	ASURA_SHADOW_LIGHTSOURCE_PROJECTOR     // Shadow comes from a projector. Shadow masks out the light
};


// -----------------------------------------------------------------------------------------
// DMR 11/04/07 - This is now used for the shadow instance creation callbacks, to replace
//                the ever-increasing list of parameters. Should be easier to extend now
// -----------------------------------------------------------------------------------------

class Asura_ShadowInstance_CreationParams
{

public:

	Asura_ShadowInstance_CreationParams() {}
	Asura_ShadowInstance_CreationParams(ASURA_SHADOW_LIGHTSOURCE eSource, uint32 uHandle, const Asura_Vector_3& xLightPos, float fLightRange, 
										float fStrength, float fSoftness, const Asura_Vector_3& xLightDirection, uint32 uFlags)
	{
		m_eSource			= eSource;
		m_uHandle			= uHandle;
		m_xLightPos			= xLightPos;
		m_xLightDirection	= xLightDirection;
		m_fLightRange		= fLightRange;
		m_fStrength			= Asura_Maths::Min(fStrength, 1.0f);
		m_fSoftness			= fSoftness;
		m_uShadowFlags		= uFlags;
		m_bCubeMap = false;
	}

	ASURA_SHADOW_LIGHTSOURCE m_eSource;						// The type of light source
	uint32					 m_uHandle;						// Handle of the light source (photon/projector)
	Asura_Vector_3			 m_xLightPos;					// Position of light source (if photon/projector)
	Asura_Vector_3			 m_xLightDirection;				// Shadow Direction (if it's an averaged shadow, where there's no position)
	float					 m_fLightRange;					// Range of the light casting this shadow
	float					 m_fStrength;					// Strength scale for the shadow
	float					 m_fSoftness;					// Softness scale for the shadow
	uint32					 m_uShadowFlags;				// Shadow flags (inc. ASURA_SHADOWFLAG_IS_DARKENING_PASS)
	bool					 m_bCubeMap;
};

#if defined(ASURA_TARGET_SM3)
class Asura_ShadowInstanceEnvironment_CreationParams : public Asura_ShadowInstance_CreationParams
{

public:
	Asura_ShadowInstanceEnvironment_CreationParams() {}

	float m_fFov;					// FOV of shadow frustrum
	float m_fNearClip;				// Near Clipping Distance
	float m_fFarClip;				// Fat clipping Distance
	Asura_Matrix_3x3 m_xOrient;		// Orientation
};
#endif	//defined(ASURA_TARGET_SM3)


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Shadow_Instance
// Author       : Andy Tate
// Date         : 28-Nov-2006
//
// Description  : Class to handle the description and state of an individual 
//				  shadow request (the shadow may or may not be rendered).
////////////////////////////////////////////////////////////////////////////////

class Asura_Shadow_Instance
{
public:
	// Public functions...
	Asura_Shadow_Instance          (const Asura_Vector_3& xLightDirection, bool bCubeMap = false, bool bEnvironment = false );
	Asura_Shadow_Instance          (const Asura_Shadow& xShadow);
	virtual ~Asura_Shadow_Instance ()	{}

	void	GetBoundingBox	(Asura_Bounding_Box& xBB) const	{ m_xFrustum.GetBoundingBox(xBB, m_xLightOrientation, m_xLightPosition); }

	// Abstract functions...
	virtual	void	CalculateDistance	(void)	= 0;
	virtual	void	Render				(Asura_Shadow& xShadow) const	= 0;
	virtual	void	Render				(Asura_Shadow& xShadow, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation) const	= 0;

	// Camera initialisation function...
	static void	SetupCamera	(const u_int uCameraIndex, const float fCutoffDistance);

	// Accessor functions...
	void					InvalidateDistance	(void)									{ m_fDistance = Asura_Maths::LargeFloat; }

	float					GetDistance			(void) const							{ return(m_fDistance); }
	const Asura_Vector_3&	GetLightDirection	(void) const							{ return(m_xLightDirection); }
	const Asura_Vector_3&	GetLightPosition	(void) const							{ return(m_xLightPosition); }
	void					GetLightOrientation	(Asura_Quat& xOrientation) const		{ xOrientation = m_xLightOrientation; }
	void					GetLightOrientation	(Asura_Matrix_3x3& xOrientation) const	{ m_xLightOrientation.ToMatrix(xOrientation); }
	const Asura_Frustum&	GetFrustum			(void) const							{ return(m_xFrustum); }
	float					GetStrength			(void) const							{ return(m_fStrength); }
	float					GetSoftness			(void) const							{ return(m_fLightSourceShadowSoftness); }
	
	ASURA_SHADOW_LIGHTSOURCE  GetLightSourceType(void) const							{ return(m_eLightSourceType); }
	u_int					  GetLightSourceHandle(void) const							{ return(m_uLightSourceHandle); }
	u_int					  GetShadowFlags(void) const								{ return(m_uLightSourceShadowFlags); }

	bool					IsCubeMapped(void) const								{ return (m_uFlags & IS_CUBE_MAP) != 0;}
	bool					IsEnvironment(void) const								{ return (m_uFlags & IS_ENVIRONMENT) != 0;}
	bool					IsParallelProjection(void) const						{ return (m_uFlags & IS_PARALLEL_PROJECTED) != 0;}

protected:
	// Protected variables...
	float						m_fDistance;					// The priority value generated for use by the sorting algorithm.
	float						m_fStrength;					// Shadow strength at the near plane (is always 0 at the far plane)
	Asura_Vector_3				m_xLightDirection;				// The direction the shadow is coming from.
	Asura_Vector_3				m_xLightPosition;				// The position of the light providing the shadow.
	Asura_Quat					m_xLightOrientation;			// The orientation of the light providing the shadow.
	Asura_Frustum				m_xFrustum;						// The frustum generated for the shadow to be rendered within.
	enum Flags
	{
		IS_CUBE_MAP = 1,
		IS_ENVIRONMENT = 2,
		IS_PARALLEL_PROJECTED = 4
	};
	u_int						m_uFlags;

	// DMR's light source parameters
	ASURA_SHADOW_LIGHTSOURCE    m_eLightSourceType;				// The light source casting this shadow (photon or projector)			
	u_int						m_uLightSourceHandle;			// Handle of the photon or projector casting this shadow
	float						m_fLightSourceRange;			// Range, in metres, of the light source	
	Asura_Vector_3				m_xLightSourcePosition;			// Position of the light source
	float						m_fLightSourceShadowStrength;	// Shadow strength - from the light source's Asura_ShadowParameters
	float						m_fLightSourceShadowSoftness;	// Shadow softness - from the light source's Asura_ShadowParameters
	uint32						m_uLightSourceShadowFlags;		// Shadow flags from the light source

	// Static variables...
	static Asura_Vector_3		s_xCameraPosition;				// The position of the current camera.
	static Asura_Quat			s_xCameraOrientation;			// The orientation of the current camera.
	static Asura_Quat			s_xCameraInverseOrientation;	// The inverse-orientation of the current camera.
	static Asura_Frustum		s_xCameraFrustum;				// The camera frustum to test for the shadows.

	// Protected functions...
#if !defined(ASURA_NO_DEBUG_OUTPUT)
	virtual	void	RenderDebugInfo	(const Asura_Vector_3& xCentre, const float fRadius, const u_int uLevel) const;
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)

	// Calculation functions...
	bool	CalculateFrustum	(const Asura_Vector_3& xCentre, const float fRadius,
								 const Asura_Vector_3& xLightPosition, const float fLightRange);

	void	CalculateDistance	(const Asura_Vector_3& xCentre, const float fRadius,
								 const Asura_Vector_3& xLightPosition, const float fLightRange);

	// Accessor functions...
	void	SetLightOrientation	(const Asura_Quat& xOrientation)		{ m_xLightOrientation = xOrientation; }
	void	SetLightOrientation	(const Asura_Matrix_3x3& xOrientation)	{ m_xLightOrientation.FromMatrix(xOrientation); }


private:
	// Private variables...

	// Private functions...
	Asura_Shadow_Instance  (const Asura_Shadow_Instance& xOther);

	Asura_Shadow_Instance&	operator=	(const Asura_Shadow_Instance& xOther);

};   // End of class 'Asura_Shadow_Instance'.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Shadow_Instance_Animation
// Author       : Andy Tate
// Date         : 28-Nov-2006
//
// Description  : Class to handle the description of a shadow to be potentially
//				  rendered by shadow system.
////////////////////////////////////////////////////////////////////////////////

class Asura_Shadow_Instance_Animation : public Asura_Shadow_Instance
{
public:
	// Public functions...
	Asura_Shadow_Instance_Animation (const Asura_Shadow& xShadow, const Asura_Animation& xAnimation, const u_int uRenderFlags, const bool bUseMovementOffset)
		: Asura_Shadow_Instance(xShadow)
		, m_xAnimation(xAnimation)
		, m_uRenderFlags(uRenderFlags)
		, m_bUseMovementOffset(bUseMovementOffset)		
	{
		m_eLightSourceType			 = ASURA_SHADOW_LIGHTSOURCE_AVERAGED;	
		m_fLightSourceShadowStrength = 1.0f;
		m_fLightSourceShadowSoftness = 1.0f;	
		m_uLightSourceShadowFlags	 = 0;
	}

	Asura_Shadow_Instance_Animation (const Asura_ShadowInstance_CreationParams& xParams, const Asura_Animation& xAnimation, const u_int uRenderFlags, bool bUseMovementOffset)
		     : Asura_Shadow_Instance(xParams.m_xLightDirection, xParams.m_bCubeMap)
			 , m_xAnimation(xAnimation)
			 , m_uRenderFlags(uRenderFlags)
			 , m_bUseMovementOffset(bUseMovementOffset)
	{
		m_eLightSourceType				= xParams.m_eSource;
		m_uLightSourceHandle			= xParams.m_uHandle;
		m_xLightSourcePosition			= xParams.m_xLightPos;
		m_xLightDirection				= xParams.m_xLightDirection;
		m_fLightSourceRange				= xParams.m_fLightRange;
		m_fLightSourceShadowSoftness	= xParams.m_fSoftness;
		m_fLightSourceShadowStrength	= xParams.m_fStrength;
		m_uLightSourceShadowFlags       = xParams.m_uShadowFlags;		
	}

	// Overridden function...
	virtual	void	CalculateDistance	(void);
	virtual	void	Render				(Asura_Shadow& xShadow) const;
	virtual	void	Render				(Asura_Shadow& xShadow, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation) const;


protected:
	// Protected variables...
	const Asura_Animation&		m_xAnimation;				// A reference to the animation to render.
	u_int						m_uRenderFlags;				// The flags use when rendering this animation.
	bool						m_bUseMovementOffset;		// Flag to say whether we're to calculate the movement offset before rendering.


	// Protected functions...
#if !defined(ASURA_NO_DEBUG_OUTPUT)
	virtual	void	RenderDebugInfo	(const Asura_Vector_3& xCentre, const float fRadius, const u_int uLevel) const;
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)


private:
	// Private typedefs...
	typedef Asura_Shadow_Instance PARENT;

	// Private variables...

	// Private functions...
	Asura_Shadow_Instance_Animation  (void);
	Asura_Shadow_Instance_Animation  (const Asura_Shadow_Instance_Animation& xOther);

	Asura_Shadow_Instance_Animation&	operator=	(const Asura_Shadow_Instance_Animation& xOther);

};   // End of class 'Asura_Shadow_Instance_Animation'.


#if defined(ASURA_TARGET_SM3)

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Shadow_Instance_Environment
// Author       : Andy Southgate
// Date         : 30-Jan-"008
//
// Description  : Class to handle the description of a shadow to be potentially
//				  rendered by shadow system.
////////////////////////////////////////////////////////////////////////////////

class Asura_Shadow_Instance_Environment : public Asura_Shadow_Instance
{
public:
	// Public functions...
	Asura_Shadow_Instance_Environment (const Asura_ShadowInstanceEnvironment_CreationParams& xParams)
		: Asura_Shadow_Instance(xParams.m_xLightDirection, xParams.m_bCubeMap, true)
	{
		m_eLightSourceType				= xParams.m_eSource;
		m_uLightSourceHandle			= xParams.m_uHandle;
		m_xLightSourcePosition			= xParams.m_xLightPos;
		m_xLightPosition				= xParams.m_xLightPos;
		m_xLightDirection				= xParams.m_xLightDirection;
		m_fLightSourceRange				= xParams.m_fLightRange;
		m_fLightSourceShadowSoftness	= xParams.m_fSoftness;
		m_fStrength = m_fLightSourceShadowStrength	= xParams.m_fStrength;
		m_uLightSourceShadowFlags       = xParams.m_uShadowFlags;	
		
		if (xParams.m_bCubeMap)
		{
			m_fDistance = -1.0f;
			m_xFrustum.SetForDistance(xParams.m_fNearClip, xParams.m_fFarClip, 1.0f, 1.0f, 1.0f);
		} else {
			float fTan = Asura_Maths::Tangent(xParams.m_fFov*0.5f);
			m_fDistance = -1.0f;
			m_xFrustum.SetForFOV(xParams.m_fNearClip, xParams.m_fFarClip,fTan, fTan);
		}

		SetLightOrientation(xParams.m_xOrient);
	}

	// Overridden function...
	virtual	void	CalculateDistance	(void);
	virtual	void	Render				(Asura_Shadow& xShadow) const;
	virtual	void	Render				(Asura_Shadow& xShadow, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation) const
	{
		Asura_Assert(0, "Not Implemented");
	};


protected: 
	bool	CalculateFrustum	(float fTanHalfHFov,float fTanHalfVFov, float fNearDist, float fFarDist);


private:
	// Private typedefs...
	typedef Asura_Shadow_Instance PARENT;

	// Private variables...
  
	// Private functions...
	Asura_Shadow_Instance_Environment  (void);
	Asura_Shadow_Instance_Environment  (const Asura_Shadow_Instance_Environment& xOther);

	Asura_Shadow_Instance_Environment&	operator=	(const Asura_Shadow_Instance_Environment& xOther);

};   // End of class 'Asura_Shadow_Instance_Animation'.

#endif	//defined(ASURA_TARGET_SM3)
 

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Shadow_Instance_Shape
// Author       : Andy Tate
// Date         : 28-Nov-2006
//
// Description  : Class to handle the description of a shadow to be potentially
//				  rendered by shadow system.
////////////////////////////////////////////////////////////////////////////////

class Asura_Shadow_Instance_Shape : public Asura_Shadow_Instance
{
public:
	// Public functions...
	Asura_Shadow_Instance_Shape (const Asura_Shadow& xShadow, const Asura_Shape_Instance& xShapeInstance, const u_int uCamVisFlags)
		: Asura_Shadow_Instance(xShadow)
		, m_xShapeInstance(xShapeInstance)
		, m_uCamVisFlags(uCamVisFlags)
	{
		m_eLightSourceType = ASURA_SHADOW_LIGHTSOURCE_AVERAGED;
		m_fLightSourceShadowStrength = 1.0f;
		m_fLightSourceShadowSoftness = 1.0f;
		m_uLightSourceShadowFlags	 = 0;
	}

	Asura_Shadow_Instance_Shape (const  Asura_ShadowInstance_CreationParams& xParams, const Asura_Shape_Instance& xShapeInstance, const u_int uCamVisFlags)
		: Asura_Shadow_Instance(xParams.m_xLightDirection, xParams.m_bCubeMap)
		, m_xShapeInstance(xShapeInstance)
		, m_uCamVisFlags(uCamVisFlags)	
	{
		m_eLightSourceType				= xParams.m_eSource;
		m_uLightSourceHandle			= xParams.m_uHandle;
		m_xLightSourcePosition			= xParams.m_xLightPos;
		m_xLightDirection				= xParams.m_xLightDirection;
		m_fLightSourceRange				= xParams.m_fLightRange;
		m_fLightSourceShadowSoftness	= xParams.m_fSoftness;
		m_fLightSourceShadowStrength	= xParams.m_fStrength;
		m_uLightSourceShadowFlags       = xParams.m_uShadowFlags;		
	}


//	virtual ~Asura_Shadow_Instance_Shape ();

	// Overridden calculation function...
	virtual	void	CalculateDistance	(void);
	virtual	void	Render				(Asura_Shadow& xShadow) const;
	virtual	void	Render				(Asura_Shadow& xShadow, const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation) const;


protected:
	// Protected variables...
	const Asura_Shape_Instance&	m_xShapeInstance;			// A reference to the shape-instance to render.
	u_int						m_uCamVisFlags;				// The camera flags for the shape-instance to be rendered.


	// Protected functions...
#if !defined(ASURA_NO_DEBUG_OUTPUT)
	virtual	void	RenderDebugInfo	(const Asura_Vector_3& xCentre, const float fRadius, const u_int uLevel) const;
#endif	// !defined(ASURA_NO_DEBUG_OUTPUT)


private:
	// Private typedefs...
	typedef Asura_Shadow_Instance PARENT;

	// Private variables...

	// Private functions...
	Asura_Shadow_Instance_Shape  (void);
	Asura_Shadow_Instance_Shape  (const Asura_Shadow_Instance_Shape& xOther);

	Asura_Shadow_Instance_Shape&	operator=	(const Asura_Shadow_Instance_Shape& xOther);

};   // End of class 'Asura_Shadow_Instance_Shape'.


#endif   // ifndef __ASURA_SHADOW_INSTANCE_H__


/// End of file. ///
