//
// Asura_Shadow_System.h
//
// Created by KJFL - 18 August 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_SHADOW_SYSTEM_H__
#define __ASURA_SHADOW_SYSTEM_H__

//////////////////////////////////////////////////////////////////////////
// Include files...
//////////////////////////////////////////////////////////////////////////
#include "Core/Asura_Hash.h"
#include "Shadows/Asura_Shadow.h"

//////////////////////////////////////////////////////////////////////////
// Defines...
//////////////////////////////////////////////////////////////////////////
#if !defined(ASURA_NUM_SHADOW_TEXTURE_SIZES)
#define ASURA_NUM_SHADOW_TEXTURE_SIZES	1
#endif	// !defined(ASURA_NUM_SHADOW_TEXTURE_SIZES)

#if !defined(ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE)
#define ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE	10
#endif	// !defined(ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE)

#if !defined(ASURA_MAX_SHADOW_TEXTURE_SIZE)
#if defined(ASURA_TARGET_SM3)
#define ASURA_MAX_SHADOW_TEXTURE_SIZE	512
#else
#define ASURA_MAX_SHADOW_TEXTURE_SIZE	256
#endif	//defined(ASURA_TARGET_SM3)
#endif

#if !defined(ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS)
#define ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS		128 
#endif

#if !defined(ASURA_MAX_NUMBER_OF_SHADOW_INSTANCES)
#define ASURA_MAX_NUMBER_OF_SHADOW_INSTANCES	128
#endif	// !defined(ASURA_MAX_NUMBER_OF_SHADOW_INSTANCES)

#if defined(ASURA_USE_SCREENSHOT_SYSTEM)
// Specific defines used for taking hi-res screenshots...
#if !defined(ASURA_NUM_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZES)
#define ASURA_NUM_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZES		1
#endif	//!defined(ASURA_NUM_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZES)

#if !defined(ASURA_NUM_HIRES_SCREENSHOT_SHADOWS_PER_TEXTURE_SIZE)
#define ASURA_NUM_HIRES_SCREENSHOT_SHADOWS_PER_TEXTURE_SIZE	128
#endif	//!defined(ASURA_NUM_HIRES_SCREENSHOT_SHADOWS_PER_TEXTURE_SIZE)

#if !defined(ASURA_MAX_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZE)
#define ASURA_MAX_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZE		(ASURA_MAX_SHADOW_TEXTURE_SIZE * 2)
#endif	//!defined(ASURA_MAX_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZE)
#endif	//defined(ASURA_USE_SCREENSHOT_SYSTEM)

//////////////////////////////////////////////////////////////////////////
// Forward declarations...
//////////////////////////////////////////////////////////////////////////
class Asura_Cutscene;
class Asura_Shadow_Instance;


//////////////////////////////////////////////////////////////////////////
// Class declarations...
//////////////////////////////////////////////////////////////////////////


// -----------------------------------------------------------------------------------------
// DMR 29/03/07 - New struct for shadow parameters - these go inside photons
//                and projectors, so we know which ones cast shadows, and which
//                contribute to averaged shadows
// -----------------------------------------------------------------------------------------
enum
{
	// Shadow flags
	ASURA_SHADOWFLAG_CONTRIBUTES_TO_AVERAGE_SHADOW = 0x01,
	ASURA_SHADOWFLAG_CASTS_OWN_SHADOWS			   = 0x02,
	ASURA_SHADOWFLAG_IS_DARKENING_PASS			   = 0x04,
	ASURA_SHADOWFLAG_SCALE_STRENGTH_BY_BRIGHTNESS  = 0x08,
	ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS	   = 0x10,
	ASURA_SHADOWFLAG_USE_NOISE					   = 0x20,
};


class Asura_ShadowParameters
{	
public:
		
	Asura_ShadowParameters()
	{		
		m_uFlags			= ASURA_SHADOWFLAG_CONTRIBUTES_TO_AVERAGE_SHADOW | ASURA_SHADOWFLAG_IS_DARKENING_PASS;
		m_fShadowStrength	= 1.0f;
		m_fVerticalOffset   = 0.0f;
		m_fShadowSoftness   = 1.0f;
	}

	Asura_ShadowParameters(u_int uFlags, float fShadowStrength = 1.0f, float fVerticalOffset = 0.0f, float fShadowSoftness = 1.0f)
	{		
		m_uFlags			= uFlags;
		m_fShadowStrength	= fShadowStrength;
		m_fVerticalOffset   = fVerticalOffset;
		m_fShadowSoftness   = fShadowSoftness;
	}

	void WriteToChunkStream  ( class Asura_Chunk_Stream& xStream )  const;
	void ReadFromChunkStream ( class Asura_Chunk_Stream& xStream );

	enum ChunkData
	{
		CURRENT_VERSION = 1
	};

	u_int			m_uFlags;								// ASURA_SHADOWFLAG_ values
	float			m_fShadowStrength;						// 0-1, strength of darkening pass shadows from this light		
	float           m_fVerticalOffset;						// Y-offset from actual photon position for casting shadows	
	float			m_fShadowSoftness;						// Scales the filter kernel
};


// -----------------------------------------------------------------------------------------
// DMR 30/03/07 - Configuration for the shadow system, so we can have some per-project control
//				  over shadow appearance, instead of 'magic numbers'
// -----------------------------------------------------------------------------------------

class Asura_Shadow_System_Config
{
public:
	int	  m_iDebugLevel;									// Level of debug output
	float m_fGIShadowStrength;								// Global scale for the strength of shadows from GI samples
	float m_fGIShadowCutoff;								// Global cutoff for shadows from the GI samples - to remove unwanted feint shadows in dark areas
	float m_fGIShadowSoftness;								// Softness of averaged shadows from the GI samples
	float m_fMaxDistanceForShadows;							// Maximum distance from an object to a photon for it to be considered for casting shadows
	float m_fMaxDistanceForShadowFiltering;					// Max distance from the camera to use shadow filtering
	bool  m_bShowStats;
	

	Asura_Shadow_System_Config()
	{
		// Some defaults that should work nicely in most cases
		m_bShowStats					 = false;
		m_iDebugLevel					 = 0;
		m_fGIShadowStrength				 = 1.0f;
		m_fGIShadowCutoff				 = 0.2f;
		m_fGIShadowSoftness				 = 1.0f;
		m_fMaxDistanceForShadows		 = 25.0f;
		m_fMaxDistanceForShadowFiltering = 150.0f;
	}
};






class Asura_Shadow_System
{
	public:
		// Public functions...
		static void Initialise( void );
		static void Reset( void );
		static void Shutdown( void );

#ifdef _DEBUG
		static void ShowDebugShadowMaps();
#endif

		static bool AddShadowInstance(Asura_Shadow_Instance* pxInstance);
 		
#if !defined(ASURA_TARGET_SM3)
		// A single render call if running without deferred shading (shadows are a darkening pass)
		static void Render(void);
#endif

		// Split into 2 calls for the deferred version, where shadows can block light
		static void RenderInstances(void);
		static void RenderOnToScene(void);				
		static void FinishedRenderingShadows(void);

		// Called from within the deferred photon/projector rendering code, to project the shadows,
		// creating a mask in the destination alpha channel used to mask out light. (It does a
		// RenderOntoScene with a couple of different renderstates, for all shadows from a specific
		// light source)
		static void BuildShadowMask(ASURA_SHADOW_LIGHTSOURCE eLightSource, uint32 uHandle);

		// Render some shadow stats
		static void RenderStats(void);					

		// Accessors...
		void	SetShadowCutoffDistance	(const float fDistance)	{ s_fShadowCutoffDistance = fDistance; }
		float	GetShadowCutoffDistance	(void) const			{ return(s_fShadowCutoffDistance); }		

#if defined(ASURA_USE_SCREENSHOT_SYSTEM)
		static void	SetHiResShadowMode(bool bValue);
#endif


		typedef void(*AddShadowInstanceCallback)(const Asura_ShadowInstance_CreationParams& xParams, const void* pxUserData);


		static void AddShadowInstances(const Asura_Vector_3 &xPos, void *pxUserData, AddShadowInstanceCallback pfnCallback);

		static Asura_Shadow_System_Config& GetConfig(void) {return s_xConfig;}

		static bool LightHasActiveShadows(ASURA_SHADOW_LIGHTSOURCE eLightSource, uint32 uLightSourceHandle);

		// (SB - 21/02/2008) New PC graphics settings system needs an accessor to disable shadows
		static bool AreShadowsEnabled( void )
		{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
			return s_bRenderEnable;
#endif
		}
		static void EnableShadows( bool bEnable = true )
		{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
			s_bRenderEnable = bEnable;
#endif
		}

#if defined(ASURA_TARGET_SM3)
		static void RenderEnvironmentShadowMap( Asura_Shadow &xShadow)
		{
			Platform_RenderEnvironmentShadowMap( xShadow);
		}
#endif	//defined(ASURA_TARGET_SM3)

		// Private variables...
		static Asura_Shadow_System_Config s_xConfig;

		static u_int					s_uNumShadows;
		static u_int					s_uNumShadowInstances;
		static Asura_Shadow_Instance*	s_apxShadowInstances[ASURA_MAX_NUMBER_OF_SHADOW_INSTANCES];

		static u_int					s_uNumShadowTextureSizes;
		static u_int					s_uNumShadowsPerTextureSize;
		static u_int					s_uMaxShadowTextureSize;

		static u_int					s_uMaxNonCubemapShadows;
		static u_int					s_uMaxShadows;
		static Asura_Shadow**			s_apxShadows;

		static float					s_fShadowCutoffDistance;


private:
		
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
		static bool						s_bRenderEnable;
#endif

		static bool						s_bHiResShadowMode;

		// Stats counters
		static uint32					s_uNumShadowCastingObjects;
		static uint32					s_uNumShadowInstancesAdded;
		static uint32					s_uNumShadowInstancesRendered;

		// Private functions...
		static void	DestroyShadowInstances	(void);

		// Callback functions...
		static void	Traverse_RenderCutsceneShadows (Asura_Hash_ID uHashID, Asura_Cutscene** ppxCutscene);
		static int	SortInstancesByDistance(const void* pData1, const void* pData2);

		// Platform specific functions...
		static void Platform_Initialise( void );
		static void Platform_Reset( void );
		static void Platform_Shutdown( void );

		static void Platform_BeginCreateShadows( void );
		static void Platform_EndCreateShadows( void );
		static void Platform_RenderOnToScene( ASURA_SHADOW_LIGHTSOURCE eLightSource, uint32 uLightSourceHandle );
		
#if defined(ASURA_TARGET_SM3)
		static void Platform_RenderEnvironmentShadowMap( Asura_Shadow &xShadow);
#endif	//defined(ASURA_TARGET_SM3)

#if defined(ASURA_TOOLS) && defined(ASURA_USE_SCREENSHOT_SYSTEM)
		static void Platform_SetHiResShadowMode(bool bValue);
#endif


		static void Platform_RenderCubicEnvOnToScene( class Asura_ClientEntity_DynamicLight* pxLight);
};


#endif//__ASURA_SHADOW_SYSTEM_H__
