
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Shape_System.cpp
// Author : Andy Tate
// Date   : 13-Mar-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Core/Asura_Stream.h"

// This file's header...
#include "Shapes/Asura_Shape_System.h"

// Asura independent header-files...
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"

// Asura platform header-files...

// Game independent header-files...

// Game platform header-files...

// Tools header-files...


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
Asura_Shape_Handler*						Asura_Shape_System::s_pxCurrentHandler	= NULL;		// Currently active handler.
Asura_TemplateHashTree<Asura_Shape_Handler>	Asura_Shape_System::s_xHandlerTree;					// Tree of all of the handlers available.
#else
Asura_Shape_Handler							Asura_Shape_System::s_xHandler;						// Single static-instance of a zone.
#endif


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::LoadShape
// Author       : Andy Tate
// Date         : 13-Mar-2006
//
// Description  : Loads the mesh for the specified shape.
//
// Parameters   : The hash-ID of the shape and a pointer to the shape itself.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_Handler::LoadShape (Asura_Hash_ID uHash, Asura_Shape* pxShape)
{
	// Load this shape...
	pxShape->LoadShape();
}

/// End of function 'Asura_Shape_Handler::LoadShape'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::LoadAllShapes
// Author       : Andy Tate
// Date         : 13-Mar-2006
//
// Description  : Loops over all of the shapes in the shape-tree. 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_Handler::LoadAllShapes (void)
{
	// Traverse the tree, loading the mesh for each shape in-turn...
	m_xShapeTree.Traverse(LoadShape);

	// MGH 24/4/2003: Unfortunately doing the above can change the flags on the shapes,
	// so that they could now be using a custom BB for dynamics. This would be fine except
	// for the fact that we've already set up the server entities' collision objects,
	// so we need to do it again.
	Asura_ServerEntity_PhysicalObject::ResetAllCollisionObjects();
}

/// End of function 'Asura_Shape_Handler::LoadAllShapes'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::SetShapeData
// Author       : Andy Tate
// Date         : 13-Mar-2006
//
// Description  : Sets the specified data for a shape with the specified ID.
//
// Parameters   : The hash-ID of the shape to find, the flags and bounding-box
//				  to be set for this shape.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_Handler::SetShapeData (Asura_Hash_ID uHash,
										unsigned uFlags,
										Asura_Bounding_Box *pxBB)
{
	Asura_Shape*	pxShape	= GetShape(uHash);

	// Do we have a shape?
	if(pxShape != NULL)
	{
		pxShape->SetShapeData(uFlags, pxBB);
	}
}

/// End of function 'Asura_Shape_Handler::SetShapeData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::NewShape
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Function to create a shape for this handler.
//
// Parameters   : The hash-ID of the shape to create.
//
// Return value : A pointer to the shape created.
////////////////////////////////////////////////////////////////////////////////

Asura_Shape* Asura_Shape_Handler::NewShape (Asura_Hash_ID uHash)
{
	// Attempt to add the shape to the hash-tree...
	bool bDidExist = true;
	Asura_Shape* pxShape = m_xShapeTree.AddOrFindNode(uHash, &bDidExist, true);
	if(pxShape && !bDidExist)
	{
		pxShape	= m_xShapeTree.GetDataByHash(uHash);

		// Do we have a shape?
		if(pxShape)
		{
			pxShape->m_uHashID	= uHash;
		}

		// Return the shape...
		return(pxShape);
	}
	else
	{
		// Couldn't add for some reason...
		Asura_Assertf(!bDidExist,("Not adding duplicate shape 0x08%x",uHash));
		return(NULL);
	}
}

/// End of function 'Asura_Shape_Handler::NewShape'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::NewShape
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Function to create a shape for this handler.
//
// Parameters   : The hash-ID of the shape to create and a reference of the shape
//				  to copy our data from.
//
// Return value : A pointer to the shape created.
////////////////////////////////////////////////////////////////////////////////

Asura_Shape* Asura_Shape_Handler::NewShape (Asura_Hash_ID uHash,
											const Asura_Shape& xOther)
{
	// Attempt to add the shape to the hash-tree...
	bool bDidExist = (m_xShapeTree.GetDataByHash(uHash,true) != NULL);
	if(!bDidExist && m_xShapeTree.AddNode(uHash, xOther))
	{
		Asura_Shape*	pxShape	= m_xShapeTree.GetDataByHash(uHash);

		// Do we have a shape?
		if(pxShape != NULL)
		{
			pxShape->m_uHashID	= uHash;
		}

		// Return the shape...
		return(pxShape);
	}
	else
	{
		// Couldn't add for some reason...
		Asura_Assertf(!bDidExist,("Not adding duplicate shape 0x08%x",uHash));
		return(NULL);
	}
}

/// End of function 'Asura_Shape_Handler::NewShape'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::NewShape
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Function to create a shape for this handler.
//
// Parameters   : The hash-ID of the shape to create and a reference to the 
//				  chunk-stream to read the data from.
//
// Return value : A pointer to the shape created.
////////////////////////////////////////////////////////////////////////////////

Asura_Shape* Asura_Shape_Handler::NewShape (Asura_Hash_ID uHash,
											Asura_Chunk_Stream& xStream)
{
	// Attempt to add the shape to the hash-tree...
	bool bDidExist = (m_xShapeTree.GetDataByHash(uHash,true) != NULL);
	if(!bDidExist && m_xShapeTree.AddNode(uHash, xStream))
	{
		Asura_Shape*	pxShape	= m_xShapeTree.GetDataByHash(uHash);

		// Do we have a shape?
		if(pxShape != NULL)
		{
			pxShape->m_uHashID	= uHash;
		}

		// Return the shape...
		return(pxShape);
	}
	else
	{
		// Couldn't add for some reason...
		Asura_Assertf(!bDidExist,("Not adding duplicate shape 0x08%x",uHash));
		return(NULL);
	}
}

/// End of function 'Asura_Shape_Handler::NewShape'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Handler::DeleteShape
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Deletes the specified shape from this hash-tree.
//
// Parameters   : A pointer to the shape to be deleted.
//
// Return value : "true" if deleted successfully, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_Shape_Handler::DeleteShape (Asura_Shape* pxShape)
{
	// Is there a shape to delete?
	if(pxShape == NULL)
	{
		return(false);
	}
	else
	{
		const Asura_Hash_ID	uHash	= pxShape->GetShapeHashID();

		// Attempt the deletion...
		return(m_xShapeTree.DeleteNodeByHash(uHash));
	}
}

/// End of function 'Asura_Shape_Handler::DeleteShape'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_System::Initialise
// Author       : Andy Tate
// Date         : 13-Mar-2006
//
// Description  : Initialises the shape-system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_System::Initialise (void)
{
	// Nothing to initialise...
}

/// End of function 'Asura_Shape_System::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_System::Shutdown
// Author       : Andy Tate
// Date         : 13-Mar-2006
//
// Description  : Handles the shutdown of the shape-system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_System::Shutdown (void)
{
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Reset the entire handler tree...
	s_xHandlerTree.Reset();

	// Remove the current handler...
	s_pxCurrentHandler	= NULL;
#else
	Asura_Shape_Handler&	xHandler	= GetCurrentHandler();

	// Reset the current instance...
	xHandler.Reset();
#endif
}

/// End of function 'Asura_Shape_System::Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_System::Reset
// Author       : Andy Tate
// Date         : 13-Mar-2006
//
// Description  : Handles the resetting of the shape-system...
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_System::Reset (void)
{
	// Simply shutdown the system, then re-initialise it...
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	GetCurrentHandler().Reset();
#else
	Shutdown();
	Initialise();
#endif
}

/// End of function 'Asura_Shape_System::Reset'. ///


#if defined(ASURA_USE_MULTIPLE_HANDLERS)

void Asura_Shape_System::ReparentHandler(Asura_Shape_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash)
{
	Asura_Shape_Handler*	pxParentHandler	= s_xHandlerTree.GetDataByHash(uNewParentHash);

	if(pxParentHandler != NULL)
	{
		xSourceHandler.m_xShapeTree.SetParent(&(pxParentHandler->m_xShapeTree));
	}
}

void Asura_Shape_System::ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash)
{
	Asura_Shape_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uSourceHash);

	if(pxHandler != NULL)
	{
		ReparentHandler(*pxHandler, uNewParentHash);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_System::CreateHandler
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Attempts to create a shape-system handler with the specified
//				  hash-ID.
//
// Parameters   : The hash-ID of the handler to create.
//
// Return value : A pointer to the handler created.
////////////////////////////////////////////////////////////////////////////////

Asura_Shape_Handler* Asura_Shape_System::CreateHandler (const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)
{
	Asura_Shape_Handler*	pxHandler	= s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler	= s_xHandlerTree.AddNode(uHash);

		if(pxHandler != NULL)
		{
			ReparentHandler(*pxHandler, uParentHandlerHash);
		}
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an shape-handler with this hash-ID.");
	}

	// Return the handler...
	return(pxHandler);
}

/// End of function 'Asura_Shape_System::CreateHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_System::DestroyHandler
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Attempts to destroy the shape-handler with the specified
//				  unique hash-ID.
//
// Parameters   : The hash-ID of the handler to destroy.
//
// Return value : "true" if destroyed, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_Shape_System::DestroyHandler (const Asura_Hash_ID uHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

/// End of function 'Asura_Shape_System::DestroyHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_System::WriteHandlerLog
// Author       : Paul Cardy
// Date         : 20-12-08
//
// Description  : Writes a log of the contents of the supplied handler
//
// Parameters   : The hash-ID of the handler to log and the filename to log to
//
// Return value : None
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_System::WriteHandlerReport(u_int uHandler, Asura_Stream& xLog)
{
	const u_int uLOGBUFFERSIZE = 64;
	char szLogBuffer[uLOGBUFFERSIZE];

	Asura_Shape_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHandler);
	if (pxHandler)
	{
		if (!pxHandler->m_xShapeTree.GetNumberOfNodes())
		{
			xLog << "Shape:\t\t\t\tNone\n";
		}
		else
		{
			Asura_TemplateHashTree_Iterator<Asura_Shape> xIt(pxHandler->m_xShapeTree);
			for ( ; !xIt.Done(); xIt.Next() )
			{
				Asura_Shape* pxShape = xIt.GetCurrent();
				Asura_PrintfToBfr(szLogBuffer, "Shape:\t\t(0x%08x)\n", uLOGBUFFERSIZE, true, pxShape->GetShapeHashID());
				xLog << szLogBuffer;
			}		
		}
	}
	else
	{
		xLog << "Shape:\t\t\t\tNo Handler\n";
	}
}

/// End of function 'Asura_Shape_System::WriteHandlerLog'. ///


#endif	// defined(ASURA_USE_MULTIPLE_HANDLERS)


/// End of file. ///
