/////////////////////////////////////////////////////////////////////////////////////////
//
//	FILE:			Asura_Shapes.cpp
//
//	DESCRIPTION:	
//
//	AUTHOR:			Richard M. Smith (5 August 2002)
//
/////////////////////////////////////////////////////////////////////////////////////////



//////////////
// INCLUDES //
//////////////

#include "Core/Asura.h"

#include "Shapes/Asura_Shapes.h"

#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#ifdef ASURA_TOOLS
#include "ChunkSystem/Asura_Chunk_Shape.h"
#include "ChunkSystem/Asura_Chunk_ShapeData.h"
#endif
#include "Core/Asura_Hash.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_ScreenShot_System.h"
#include "Light/Asura_Light.h"
#include "Light/Asura_Lighting_Solution.h"
#include "Light/Asura_GI_Solution.h"
#include "Mesh/Asura_CollisionMesh.h"
#ifdef ASURA_TOOLS
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#endif
#include "Rendering/Asura_Colour.h"
#if defined( ASURA_TARGET_SM3)
#include "Rendering/Asura_DeferredShading.h"
#endif // #if defined( ASURA_TARGET_SM3)
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
#include "Shadows/Asura_Shadow_Instance.h"
#include "Shadows/Asura_Shadow_System.h"
#endif	// defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
#include "Shapes/Asura_Shape_System.h"
#include "Shapes/Asura_Platform_Shape.h"
#include "Debugging/Asura_DebugLineRender.h"

#ifdef ASURA_TOOLS
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Render.h"
#include "Shapes/Tools_Shapes.h"
#endif //ASURA_TOOLS

#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif //defined(ASURA_USE_STREAMING)

#if defined (ASURA_USE_NEW_VISIBILITY_SYSTEM)
#include "Visibility/Asura_Visibility_System.h"
#endif

///////////////
// DEBUGVARs //
///////////////

#if _DEBUG || defined(ASURA_TOOLS)
DEBUGVAR bool dbg_bShowShapeLOSBBs = false;
#endif


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape::Asura_Shape
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Default constructor for the shape class.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

Asura_Shape::Asura_Shape (void)
:m_uFlags(0)
,m_uHashID( ASURA_HASH_ID_UNSET )
,m_uFragmentHashID(ASURA_HASH_ID_UNSET)
,m_pxLOSCollisionData( NULL )
,m_pxDynamicsCollisionMesh( NULL )
,m_pxDynamicsCollisionData( NULL )
,m_pxRenderData( NULL )
{
}

/// End of function 'Asura_Shape::Asura_Shape'. ///



/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::Asura_Shape (Constructor)
// PARAMETERS	Hash ID for this new shape
// RETURNS		Default (Constructor)
// DESCRIPTION	Constructor
// AUTHOR		Richard M. Smith (8 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
Asura_Shape::Asura_Shape( Asura_Hash_ID uHashID )
:m_uFlags(0)
,m_uHashID( uHashID )
,m_uFragmentHashID(ASURA_HASH_ID_UNSET)
,m_pxLOSCollisionData( NULL )
,m_pxDynamicsCollisionMesh( NULL )
,m_pxDynamicsCollisionData( NULL )
,m_pxRenderData( NULL )
{
} // End Function 'Asura_Shape::Asura_Shape (Constructor)'



/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::Asura_Shape (Constructor)
// PARAMETERS	Chunk buffer pointer
// RETURNS		Default (Constructor)
// DESCRIPTION	Constructor
// AUTHOR		Richard M. Smith (8 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
Asura_Shape::Asura_Shape(Asura_Hash_ID uHash, Asura_Chunk_Stream& xStream)
:m_uFlags(0)
,m_uHashID( uHash )
,m_uFragmentHashID(ASURA_HASH_ID_UNSET)
,m_pxLOSCollisionData( NULL )
,m_pxDynamicsCollisionMesh( NULL )
,m_pxDynamicsCollisionData( NULL )
,m_pxRenderData( NULL )
{
	// ART - 14/03/06 - This is nasty, but I haven't the time to write chunk-stream functions
	// for all of the classes held here (particularly "Asura_Platform_Shape_RenderData").
	// This does need changing though. (ARTTODO)
	Asura_Chunk_Memory_Stream*	pxMemoryStream	= static_cast<Asura_Chunk_Memory_Stream*>(&xStream);
	const char*					pcChunkBuffer	= static_cast<char*>(pxMemoryStream->GetAddress());

	// Read back the flags to indicate what follows
	unsigned int uFlags = 0;

	ASURA_CHUNK_EXTRACT( pcChunkBuffer, uFlags, unsigned int );

	if( uFlags & ASURA_SHAPE_FLAG_HAS_COLLISION_DATA )
	{
		// Read the collision data
		m_pxLOSCollisionData = new Asura_CollisionMesh( pcChunkBuffer );
		Asura_Assert( NULL != m_pxLOSCollisionData, "(RMS) Out of memory" );
	}

	if( uFlags & ASURA_SHAPE_FLAG_HAS_RENDER_DATA )
	{
		// Read the render data
		m_pxRenderData = new Asura_Platform_Shape_RenderData( pcChunkBuffer );
		Asura_Assert( NULL != m_pxRenderData, "(RMS) Out of memory" );
	}

	if( uFlags & ASURA_SHAPE_FLAG_HAS_FRAGMENT_DATA )
	{
		// Read the fragment hash id
		ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_uFragmentHashID, Asura_Hash_ID);
	}

	if( uFlags & ASURA_SHAPE_FLAG_HAS_DYNAMICS_COLLISIONMESH)
	{
		//read the dynamics collision mesh
		m_pxDynamicsCollisionMesh = new Asura_CollisionMesh(pcChunkBuffer);
	}

    // We should have read up until the end of the memory stream.
    Asura_Assert( pcChunkBuffer == pxMemoryStream->GetEndAddress(), "" );
}



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape::Asura_Shape
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : 
//
// Parameters   : A reference to the shape to copy.
//
// Return value : 
////////////////////////////////////////////////////////////////////////////////

Asura_Shape::Asura_Shape (const Asura_Shape& xOther)
:m_uFlags(xOther.m_uFlags)
,m_uHashID(xOther.m_uHashID)
,m_uFragmentHashID(xOther.m_uFragmentHashID)
,m_pxLOSCollisionData( NULL )
,m_pxDynamicsCollisionMesh( NULL )
,m_pxDynamicsCollisionData( NULL )
,m_pxRenderData( NULL )
{
	// Shouldn't ever call this...
	Asura_Assert(false, "Shouldn't be calling the copy constructor for \"Asura_Shape\".");
}

/// End of function 'Asura_Shape::Asura_Shape'. ///



/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::~Asura_Shape (Destructor)
// PARAMETERS	Hash ID for this new shape
// RETURNS		Default (Destructor)
// DESCRIPTION	Destructor
// AUTHOR		Richard M. Smith (8 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
Asura_Shape::~Asura_Shape()
{
	delete(m_pxLOSCollisionData);
	delete(m_pxDynamicsCollisionMesh);
	delete(m_pxDynamicsCollisionData);
	delete(m_pxRenderData);
} // End Function 'Asura_Shape::~Asura_Shape (Destructor)'



/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::LoadShape
// PARAMETERS	None
// RETURNS		None
// DESCRIPTION	Loads the specified Asura file and obtains the shape within
// AUTHOR		Richard M. Smith (8 August 2002)
// MODIFIED     Mike Healey (30 August 2002) - split off the tools part
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape::LoadShape()
{
	Asura_Assert( NULL != m_pxRenderData, "(RMS) No render data!" );
	m_pxRenderData->Load(m_uHashID);

	//refresh the fragment id
	m_pxRenderData->GetFragmentHashFromMesh(m_uFragmentHashID);

	/*
    //Should no longer be needed , since the shape table is always rebuilt from scratch in tools
	//RAR 26/02/2004

	#ifdef ASURA_TOOLS 
	//Recalculate the collision mesh, in case the mesh in the shape asura file has been changed/
	//(Probably need a way of detecting when the file has been modified).
	//RAR 25/03/03
	Asura_Mesh* pxMesh = m_pxRenderData->GetMesh();
	if(pxMesh && m_pxLOSCollisionData)
	{
	    m_pxLOSCollisionData->UpdateFromMesh( pxMesh );
	}

	if(pxMesh && pxMesh->m_pxCollisionMesh)
	{
		//update the dynamics collision mesh
		if(m_pxDynamicsCollisionMesh)
		{
			m_pxDynamicsCollisionMesh->Copy(pxMesh->m_pxCollisionMesh);
		}
		else
		{
			m_pxDynamicsCollisionMesh = new Asura_CollisionMesh(pxMesh->m_pxCollisionMesh);
		}
	}
	else
	{
		delete m_pxDynamicsCollisionMesh;
		m_pxDynamicsCollisionMesh = 0;
	}
	#endif
	*/

    if( IsNeverRendered() )
    {
        // (MP - 16.05.2005) Currently we're not exporting render data for
        // shapes that are never rendered. Therefore attempting to retrieve
        // the material index below is unlikely to be terribly successful.
        // While a bit rubbish, this fix is sufficient for Rogue for now - it
        // prevents an assert and is all that would have happened anyway. At
        // some point we should probably sort out the saving of shapes'
        // collision mesh data so that the materials are correct in there in
        // the first place and this strange method of going and fixing them
        // up using render info is no longer required.
        m_pxLOSCollisionData->SetAllMaterialIDs( ASURA_MATERIAL_NONE );
		m_pxLOSCollisionData->SetAllMaterialResponseHashIDs( ASURA_HASH_ID_UNSET );
    }
    else
    {
		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			m_pxLOSCollisionData->SetAllMaterialIDs( ASURA_MATERIAL_NONE );
			m_pxLOSCollisionData->SetAllMaterialResponseHashIDs( Platform_GetMaterialIdx() );
		#else
	        m_pxLOSCollisionData->SetAllMaterialIDs( static_cast<short>(Platform_GetMaterialIdx()) );
			m_pxLOSCollisionData->SetAllMaterialResponseHashIDs( ASURA_HASH_ID_UNSET );
		#endif
    }
}
// End Function 'Asura_Shape::LoadShape'




#ifdef ASURA_TOOLS
/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::Initialise
// PARAMETERS	AsurA_Mesh pointer
// RETURNS		Default (Constructor)
// DESCRIPTION	Constructor
// AUTHOR		Richard Rice
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape::Initialise( Asura_Mesh* pxMesh )
{
	Asura_Assert(pxMesh,"");

	// Remove any existing data...
	delete(m_pxRenderData);
	delete(m_pxLOSCollisionData);
	delete(m_pxDynamicsCollisionMesh);
	delete(m_pxDynamicsCollisionData);

	//Create the render data
	m_pxRenderData				= new Asura_Platform_Shape_RenderData( pxMesh );
	m_pxLOSCollisionData		= NULL;
	m_pxDynamicsCollisionMesh	= NULL;
	m_pxDynamicsCollisionData	= NULL;

	if( !pxMesh->BoundingBoxDataValid )
	{
		pxMesh->CalculateBoundingBox();
	}

	if( !pxMesh->AreNormalsValid() )
	{
		pxMesh->CreateNormals();
	}

#ifdef ASURA_TOOLS
	m_pxLOSCollisionData = new Asura_CollisionMesh( pxMesh );
	Asura_Assert( NULL != m_pxLOSCollisionData, "(MH) Error - Out of memory" );

	if(pxMesh->m_pxCollisionMesh)
	{
		//copy the low poly dynamics collision mesh
		m_pxDynamicsCollisionMesh = new Asura_CollisionMesh(pxMesh->m_pxCollisionMesh);
	}
#endif

	m_uFragmentHashID = pxMesh->m_uFragmentHash;

} // End Function 'Asura_Shape::Asura_Shape (Constructor)'


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::CreateShape
// PARAMETERS	Filename of an Asura file to load the shape from
//				Name of the mesh (normally just the stripped down filename)
// RETURNS		TRUE on success; FALSE otherwise
// DESCRIPTION	Loads the specified Asura file and obtains the shape within
// AUTHOR		Richard M. Smith (8 August 2002)
// MODIFIED     Mike Healey (30 August 2002) - tools bit seperate. create collision mesh.
/////////////////////////////////////////////////////////////////////////////////////////
bool Asura_Shape::CreateShape( const char* szShapeFileName, const char* szMeshName )
{
    // Get the render data for our *TEMPORARY* format!
	m_pxRenderData = new Asura_Platform_Shape_RenderData( szShapeFileName, szMeshName );
	Asura_Assert( NULL != m_pxRenderData, "(RMS) Error - Out of memory" );

	Asura_Mesh*	pxTheMesh = m_pxRenderData->GetMesh();

	if( NULL == pxTheMesh )
	{
		Asura_Assertf(false, ("Filename '%s' does not match internal mesh name - Shape will NOT be added", szShapeFileName));
/// Richard May - 11/11/2002 11:54:47 - commented this tools call out as this function is used _outside_ tools
/// and attempting to render tools overlays was causing the system to crash. Replace with the above assert
//		Tools_ShapeTable_WarningDlg	xWarning("Filename does not match internal mesh name - Shape will NOT be added");
//		xWarning.DoModal();

		delete m_pxRenderData;
		m_pxRenderData = NULL;

		return false;
	}

    if( !pxTheMesh->BoundingBoxDataValid )
	{
		pxTheMesh->CalculateBoundingBox();
    }
	if( !pxTheMesh->AreNormalsValid() )
    {
	    pxTheMesh->CreateNormals();
	}

    m_pxLOSCollisionData = new Asura_CollisionMesh( pxTheMesh );
	Asura_Assert( NULL != m_pxLOSCollisionData, "(MH) Error - Out of memory" );
	
	if(pxTheMesh->m_pxCollisionMesh)
	{
		//copy the low poly dynamics collision mesh
		m_pxDynamicsCollisionMesh = new Asura_CollisionMesh(pxTheMesh->m_pxCollisionMesh);
	}

	m_uFragmentHashID = pxTheMesh->m_uFragmentHash;

	return true;

} // End Function 'Asura_Shape::CreateShape'

#endif//ASURA_TOOLS

#ifdef ASURA_TOOLS
/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::GetChunkBufferSize
// PARAMETERS	None
// RETURNS		Returns chunk size in bytes required to save this shape
// DESCRIPTION	See returns
// AUTHOR		Richard M. Smith (8 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
int Asura_Shape::GetChunkBufferSize( void )
{
	int	iSizeInBytes;

// Hash value
	iSizeInBytes = sizeof( Asura_Hash_ID );

// Flags
	iSizeInBytes += sizeof( unsigned int );

	if( NULL != m_pxLOSCollisionData )
	{
		iSizeInBytes += m_pxLOSCollisionData->GetChunkBufferSize();
	}

	if( NULL != m_pxRenderData )
	{
		iSizeInBytes += m_pxRenderData->GetChunkBufferSize();
	}

	if(m_uFragmentHashID != ASURA_HASH_ID_UNSET)
	{
		iSizeInBytes += sizeof( Asura_Hash_ID );
	}

	if(m_pxDynamicsCollisionMesh)
	{
		iSizeInBytes += m_pxDynamicsCollisionMesh->GetChunkBufferSize();
	}

	return iSizeInBytes;
	
} // End Function 'Asura_Shape::GetChunkBufferSize'
#endif//ASURA_TOOLS





#ifdef ASURA_TOOLS
/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape::FillInChunkBuffer
// PARAMETERS	Pointer to the chunk buffer
// RETURNS		None
// DESCRIPTION	
// AUTHOR		Richard M. Smith (8 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape::FillInChunkBuffer( char*& pcChunkBuffer, const List<Asura_Mesh*>& xMeshList )
{
// Write out the hash ID
	ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_uHashID, Asura_Hash_ID );

// Write out the flags to indicate what follows
	unsigned int uFlags = 0;

	if( NULL != m_pxLOSCollisionData )
	{
		uFlags |= ASURA_SHAPE_FLAG_HAS_COLLISION_DATA;
	}

	if( NULL != m_pxRenderData )
	{
		uFlags |= ASURA_SHAPE_FLAG_HAS_RENDER_DATA;
	}
	
	if(m_uFragmentHashID != ASURA_HASH_ID_UNSET)
	{
		uFlags |= ASURA_SHAPE_FLAG_HAS_FRAGMENT_DATA;
	}

	if(m_pxDynamicsCollisionMesh)
	{
		uFlags |= ASURA_SHAPE_FLAG_HAS_DYNAMICS_COLLISIONMESH;
	}

	ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, uFlags, unsigned int );

    int* piMaterialIDToIndexConversion = new int[ASURA_MAX_NO_OF_MATERIALS];
    Tools_SaveAsura::PrepareMeshMaterialConversionLists( xMeshList, piMaterialIDToIndexConversion, 0 );

// Write out the (optional) collision data
	if( NULL != m_pxLOSCollisionData )
	{
        m_pxLOSCollisionData->FillInChunkBuffer( pcChunkBuffer, piMaterialIDToIndexConversion);
	}

// Write out the (optional) render data
	if( NULL != m_pxRenderData )
	{
		m_pxRenderData->FillInChunkBuffer( pcChunkBuffer );
	}

//Write out the (optional) fragment id
	if(m_uFragmentHashID != ASURA_HASH_ID_UNSET)
	{
		ASURA_CHUNK_FILLANDVERIFY( pcChunkBuffer, m_uFragmentHashID, Asura_Hash_ID );
	}

//and the (optional) dynamics collision mesh
	if(m_pxDynamicsCollisionMesh)
	{
        m_pxDynamicsCollisionMesh->FillInChunkBuffer(pcChunkBuffer, piMaterialIDToIndexConversion);	
	}

    delete[] piMaterialIDToIndexConversion;

} // End Function 'Asura_Shape::FillInChunkBuffer'
#endif//ASURA_TOOLS

void Asura_Shape::SetShapeData(unsigned uFlags, Asura_Bounding_Box *pxBB)
{
	m_uFlags = uFlags;
	if (pxBB)
	{
		Asura_Assert(m_uFlags & ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB, "Sanity check");
		if (!m_pxDynamicsCollisionData)
		{
			m_pxDynamicsCollisionData = new Asura_Collision_Box(*pxBB);
		}
		else
		{
			// MGH 29/4/03: Make sure the BB is up to date.
			m_pxDynamicsCollisionData->ChangeBB(*pxBB);
		}
	}
}

int Asura_Shape::GetTextureIndex() const
{
    return Asura_Material_System::GetTextureID( Platform_GetMaterialIdx() );
}




/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::Asura_Shape_Instance (Constructor)
// PARAMETERS	None
// RETURNS		Default (Constructor)
// DESCRIPTION	Constructor
// AUTHOR		Richard M. Smith (9 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
Asura_Shape_Instance::Asura_Shape_Instance( void )
:m_iLastKnownModule(-1)
,m_uIgnoreForThisPlayer(ASURA_GUID_UNREGISTERED)
,m_uRenderOnlyForThisPlayer(ASURA_GUID_UNREGISTERED)
,m_xLightingSolution(true,false)

{
    m_uFlags = 0;
	Platform_SetDefaultAmbientRGB();

	m_uShapeHashID	= ASURA_HASH_ID_UNSET;
	m_pxTheShape	= NULL;
    m_pxAnimation   = NULL;
	m_pxParentAnimation = NULL;

#if !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

#if defined(ASURA_USE_STREAMING)
	m_uFrameNumberCalculatedModules = 0;
	m_uLastKnownZone = ASURA_HASH_ID_UNSET;
#else
	m_ausModuleList[0] = USHRT_MAX;
#endif

#endif // !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

	m_xRotatedDisplacedBB.Reset();
	m_xPos.x = Asura_Maths::LargeFloat;
	m_xPos.y = Asura_Maths::LargeFloat;
	m_xPos.z = Asura_Maths::LargeFloat;
	m_xOri.Identity();
	
    m_fAlpha = 1.0f;
	m_fAnimTimestep = ANIMTIME_USEGAMETIME;

} // End Function 'Asura_Shape_Instance::Asura_Shape_Instance (Constructor)'





/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::Asura_Shape_Instance (Constructor)
// PARAMETERS	Hash ID for this new shape
// RETURNS		Default (Constructor)
// DESCRIPTION	Constructor
// AUTHOR		Richard M. Smith (9 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
Asura_Shape_Instance::Asura_Shape_Instance( const Asura_Hash_ID uShapeHashID )
:m_iLastKnownModule(-1)
,m_uIgnoreForThisPlayer(ASURA_GUID_UNREGISTERED)
,m_uRenderOnlyForThisPlayer(ASURA_GUID_UNREGISTERED)
,m_xLightingSolution(true,false)
{
    m_uFlags = 0;
	Platform_SetDefaultAmbientRGB();

	m_uShapeHashID	= uShapeHashID;
	m_pxTheShape	= Asura_Shape_System::GetShape( uShapeHashID );
    m_pxAnimation   = NULL;
	m_pxParentAnimation = NULL;

#if !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

#if defined(ASURA_USE_STREAMING)
	m_uFrameNumberCalculatedModules = 0;
	m_uLastKnownZone = ASURA_HASH_ID_UNSET;
#else
	m_ausModuleList[0] = USHRT_MAX;
#endif

#endif // !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)

	m_xRotatedDisplacedBB.Reset();
	m_xPos.x = Asura_Maths::LargeFloat;
	m_xPos.y = Asura_Maths::LargeFloat;
	m_xPos.z = Asura_Maths::LargeFloat;
	m_xOri.Identity();
    
	m_fAlpha = 1.0f;
	m_fAnimTimestep = ANIMTIME_USEGAMETIME;

} // End Function 'Asura_Shape_Instance::Asura_Shape_Instance (Constructor)'


Asura_Shape_Instance::~Asura_Shape_Instance()
{
    if (m_pxAnimation)
    {
		//reset the parent animation to be safe... probably shouldn't matter , but
		//just in case it suceeds at unlinking the attachment. RAR 12/05/2005
		m_pxAnimation->SetParent(0);

        delete m_pxAnimation;
    }
}


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::SetShape
// PARAMETERS	Hash ID of the shape
// RETURNS		None
// DESCRIPTION	Sets the shape that this shape instance uses
// AUTHOR		Richard M. Smith (9 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::SetShape( const Asura_Hash_ID uShapeHashID )
{
    if( m_pxAnimation )
	{
		//reset the parent animation to be safe... probably shouldn't matter , but
		//just in case it suceeds at unlinking the attachment. RAR 12/05/2005
		m_pxAnimation->SetParent(0);
		
		delete m_pxAnimation;
		m_pxAnimation = NULL;
	}

	if( m_uShapeHashID != uShapeHashID )
	{
		m_uShapeHashID = uShapeHashID;
		if(uShapeHashID == ASURA_HASH_ID_UNSET)
		{
			m_pxTheShape = 0;
		}
		else
		{
			m_pxTheShape = Asura_Shape_System::GetShape( uShapeHashID );
		}
	}
} // End Function 'Asura_Shape_Instance::SetShape'


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::SetAnimationPointer
// PARAMETERS	Pointer to an animation
// RETURNS		None
// DESCRIPTION	Sets the internal animation pointer to an arbitrary animation created elsewhere.
//				Useful for using posable anims and the like with the Shape instances management functionality.
//				Deletes any existing animation.
// AUTHOR		Richard May, 30/01/06
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::SetAnimationPointer( Asura_Animation* pxAnimation )
{
	if( m_pxAnimation != pxAnimation )
	{
		if( m_pxAnimation )
		{
			m_pxAnimation->SetParent(NULL); // hopefully prevents anything nasty happening on deletion
			delete m_pxAnimation;
		}
	}

	m_pxAnimation = pxAnimation;
} // End Function 'Asura_Shape_Instance::SetAnimationPointer'


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::SetSkinAndAnimID
// PARAMETERS	Hash ID of the shape
// RETURNS		None
// DESCRIPTION	Sets the skin and animation of the shape instance. 
//			    This will over-write any existing shape the instance may have.
// AUTHOR		
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::SetSkinAndAnimID( const Asura_Hash_ID uSkinID, const Asura_Hash_ID uAnimID, const int iAnimFlags,float fInitialAnimTime /*=0*/ )
{
    Asura_Assert(uSkinID != ASURA_HASH_ID_UNSET, "");
    if(m_uShapeHashID != ASURA_HASH_ID_UNSET)
	{
		m_uShapeHashID = ASURA_HASH_ID_UNSET;
		m_pxTheShape = NULL;
	}

    if (m_pxAnimation)
    {
		Asura_Hash_ID uCurrentSkinID = ASURA_HASH_ID_UNSET;

        const Asura_Hierarchy_Skin* pxSkin = m_pxAnimation->GetSkin();
        Asura_Assert(pxSkin, "");
        if( pxSkin )
        {
			//uCurrentSkinID = Asura_GetHashID( pxSkin->Name.GetString() );
			uCurrentSkinID = pxSkin->m_uNameHashID; 
		}
        
        if( uCurrentSkinID != uSkinID )
        {
            m_pxAnimation->SetSkin(uSkinID);

			// If the skin doesn't create correctly, don't create an anim.
	        const Asura_Hierarchy_Skin* pxNewSkin = m_pxAnimation->GetSkin();
            if( pxNewSkin )
			{
				m_pxAnimation->m_xPrimaryWrapper.StartAnim2(uAnimID, iAnimFlags, 0.0f,-fInitialAnimTime);				
			}
			else
			{
				Asura_Assert(0, "Failed to create skin");
			}
        }
        else if( m_pxAnimation->m_xPrimaryWrapper.GetPrimaryAnimHashID() != uAnimID )
        {
            /// Richard May - 24/06/2003 17:24:44 - This is actually necessary sometimes 
			/// (like if you want a client triggering its animations, rather than the server 
			/// managing everything) and doesn't actually hurt, so I've removed the assert.
			// must have missed a change anim message. This shouldn't happen.
            //Asura_Assert(false, "missed a change anim message");
            m_pxAnimation->m_xPrimaryWrapper.StartAnim2(uAnimID, iAnimFlags,DEFAULT_TWEEN_TIME,-fInitialAnimTime);
        }
    }
    else
    {
        m_pxAnimation = new Asura_Animation(uSkinID);
        if (m_pxAnimation)
        {
            m_pxAnimation->m_xPrimaryWrapper.StartAnim2(uAnimID, iAnimFlags, 0.0f,-fInitialAnimTime);

			//If being used as part of an attachment , inform the animation of this.
			if(m_pxParentAnimation)
			{
				m_pxAnimation->SetParent(m_pxParentAnimation);
			}
        }
    }
}  // End Function 'Asura_Shape_Instance::SetSkinAndAnimID'



//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_Shape_Instance::ChangeAnimation().
// PARAMETERS   uAnimID        - hashed anim name
//              iAnimFlags     - flags, e.g. AsuraAnimationFlag_Loop
//              fTweenDuration - used if tween flag set
// RETURNS      None.
// DESCRIPTION  changes the primary animation of the skin/hierarchy
// AUTHOR       Mike Healey - 19.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::ChangeAnimation(const Asura_Hash_ID uAnimID, const int iAnimFlags, float fTweenDuration)
{
    Asura_Assert(m_pxAnimation, "no skin/hierarchy");
    if (m_pxAnimation)
    {
        m_pxAnimation->m_xPrimaryWrapper.StartAnim2(uAnimID, iAnimFlags, fTweenDuration);
    }
}


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_Shape_Instance::ChangeAnimation().
// PARAMETERS   uAnimID        - hashed anim name
//              iAnimFlags     - flags, e.g. AsuraAnimationFlag_Loop
//              fTweenDuration - used if tween flag set
//				fStartPos      - position in animation to play from
//				fEndPos        - position in animation to play to
// RETURNS      None.
// DESCRIPTION  changes the primary animation of the skin/hierarchy with extra control
// AUTHOR       Richard May    - 15/06/2004
//
//////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::ChangeAnimation(const Asura_Hash_ID uAnimID, const int iAnimFlags, float fTweenDuration, float fStartPos, float fEndPos/*=1.0f*/ )
{
    Asura_Assert(m_pxAnimation, "no skin/hierarchy");
    if (m_pxAnimation)
    {
        m_pxAnimation->m_xPrimaryWrapper.PlayAnimToPos( uAnimID, fStartPos, fEndPos, iAnimFlags, fTweenDuration );
    }
}


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::IsShape
// PARAMETERS	None.
// RETURNS		Boolean to say if it is an shape (as opposed to an animation)
// DESCRIPTION	Means to find out whether this is a shape or an animation.
// AUTHOR		Andrew Weinkove (25 October 2002)
/////////////////////////////////////////////////////////////////////////////////////////
bool Asura_Shape_Instance::IsShape() const
{
//	Asura_Assert(m_pxTheShape || m_pxAnimation, "Sanity check");
	Asura_Assert(!(m_pxTheShape && m_pxAnimation), "Sanity check");
	return (m_pxTheShape != NULL);
}


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::SetNewPosAndOri
// PARAMETERS	Position and Orientation.
// RETURNS		None
// DESCRIPTION	Sets the position and orientation of the shape.
//				Also, creates a BB which accounts for the current position
//				and orientation. This BB will be used for culling against
//				visible modules.
// AUTHOR		Andrew Weinkove (3 October 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::SetNewPosAndOri(const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri )
{
	m_xPos = xPos;
	m_xOri = xOri;
    if (m_pxTheShape)
    {
        Asura_Assert(m_pxTheShape->GetLOSCollisionData(), "");
	    m_xRotatedDisplacedBB.FromOBB( m_pxTheShape->GetLOSCollisionData()->GetBoundingBox(), xOri );
	    m_xRotatedDisplacedBB.Displace(m_xPos);
    }
}
// End function Asura_Shape_Instance::SetNewPosAndOri().


#if  !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM)
/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::CalculateModuleList
// PARAMETERS	None.
// RETURNS		None
// DESCRIPTION	Calculates which modules this shape is currently in.
//				This will be used for quick culling.
// AUTHOR		Andrew Weinkove (25 October 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::CalculateModuleList()
{
	Asura_Assert(m_pxTheShape, "This function should only be called on instances with valid shape data");
    Asura_Assert(m_xRotatedDisplacedBB.MinX != Asura_Maths::LargeFloat, "bounding box is not initialised");

#if defined(ASURA_USE_STREAMING)
	m_xModuleList.CalculateModuleList(m_xRotatedDisplacedBB);
	m_uFrameNumberCalculatedModules = Asura_Timers::GetFrameNumber();
#else
	Asura_Module_System::TestModuleTree(m_xRotatedDisplacedBB, Asura_Module_System::CullingListModuleCallback);
	Asura_Module_System::FillModuleList(m_ausModuleList);
#endif

}


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::InvalidateModuleList
// PARAMETERS	None.
// RETURNS		None
// DESCRIPTION	Makes the module list invalid, so that it won't be used for culling.
// AUTHOR		Andrew Weinkove (25 October 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::InvalidateModuleList()
{
#if defined(ASURA_USE_STREAMING)
	m_xModuleList.MakeInvalid();
#else
	m_ausModuleList[0] = USHRT_MAX;
#endif
}
#endif

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::ResetLastKnownModuleID
// PARAMETERS	None.
// RETURNS		None
// DESCRIPTION	Resets the last known module ID - useful for 3D decals.
// AUTHOR		Andrew Weinkove (26 August 2004)
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::ResetLastKnownModuleID()
{
	m_iLastKnownModule = -1;
#if defined(ASURA_USE_STREAMING)
	m_uLastKnownZone = ASURA_HASH_ID_UNSET;
#endif
}
#if !defined ( ASURA_USE_NEW_VISIBILITY_SYSTEM )
/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::TestModuleVisibility
// PARAMETERS	None.
// RETURNS		Returns false if we needn't draw.
// DESCRIPTION	Does quick culling based on the module list.
// AUTHOR		Andrew Weinkove (25 October 2002)
/////////////////////////////////////////////////////////////////////////////////////////
bool Asura_Shape_Instance::TestModuleVisibility()
{
	Asura_Assert(m_pxTheShape, "This function should only be called on instances with valid shape data");
	Asura_Assert(Asura_Module_System::GetNumberOfModules() < USHRT_MAX, "Assumption in clipping code about number of modules is incorrect");
	if (m_uFlags & ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING)
	{
		return true;
	}

#if defined(ASURA_USE_STREAMING)
	return m_xModuleList.IsVisible();
#else
	bool bAccept = false;
	// Do some clipping here, based on module visibility.
	for (int i=0;i<NUM_MODULES_TO_STORE_FOR_CLIPPING;i++)
	{
		if (m_ausModuleList[i] == USHRT_MAX)
			break;
		else if (Asura_Module_System::IsModuleVisible(m_ausModuleList[i]))
		{
			bAccept = true;
			break;
		}
	}
	return bAccept;
#endif //defined(ASURA_USE_STREAMING)
}
#endif// !defined ( ASURA_USE_NEW_VISIBILITY_SYSTEM )

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::GetContainingZoneFromModuleTest
// PARAMETERS	None
// RETURNS		Hash of last known containing zone.
// DESCRIPTION	Return last known containing zone, based on module visibility tests. 
// AUTHOR		Martin G. Harris (5 June 2006)
/////////////////////////////////////////////////////////////////////////////////////////
#if defined(ASURA_USE_STREAMING)
Asura_Hash_ID Asura_Shape_Instance::GetContainingZoneFromModuleTest() const
{
	Asura_Hash_ID uHash = ASURA_HASH_ID_UNSET;
	if (m_pxAnimation)
	{
		uHash = m_pxAnimation->GetContainingZoneFromModuleTest();
	}
	else
	{
		uHash = m_xModuleList.GetContainingZoneHash();
	}
		
	return (uHash == ASURA_HASH_ID_UNSET) ? m_uLastKnownZone : uHash;
}
#endif //defined(ASURA_USE_STREAMING)

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::IsSphereInFrustum
// PARAMETERS	None
// RETURNS		None
// DESCRIPTION	Does sphere in frustum test.
// AUTHOR		Andrew Weinkove (25 October 2002)
/////////////////////////////////////////////////////////////////////////////////////////
bool Asura_Shape_Instance::IsSphereInFrustum()
{
  	Asura_Assert(m_pxTheShape, "This function should only be called on instances with valid shape data");
	Asura_Assert(m_pxTheShape->GetLOSCollisionData(), "");
	const float fRadius = m_pxTheShape->GetLOSCollisionData()->GetBoundingRadius();
	const Asura_Bounding_Box &xBB = m_pxTheShape->GetLOSCollisionData()->GetBoundingBox();
	const Asura_Vector_3 xCentre(m_xPos + (xBB.CalculateCentre() * m_xOri));
	return Asura_Camera::IsSphereInFrustum(xCentre,fRadius);
}

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::Render
// PARAMETERS	Position to render the shape at
//				Orientation to render the shape in
// RETURNS		None
// DESCRIPTION	Renders the shape associated with this shape instance
// AUTHOR		Richard M. Smith (9 August 2002)
// MODIFIED     (MH - 11.10.2002) added Asura_Animation*
/////////////////////////////////////////////////////////////////////////////////////////
#if _DEBUG
DEBUGVAR float dbg_fForceShapeInstanceAlpha = -1.0f;
#endif
void Asura_Shape_Instance::Render(unsigned uCamVisFlags /*= 0*/)
{
	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_OBJECTS))
	{
		return;
	}

#if _DEBUG
    if (dbg_fForceShapeInstanceAlpha >= 0.0f)
    {
        m_fAlpha = dbg_fForceShapeInstanceAlpha;
    }
#endif

    const Asura_Shape* const pxShape = m_pxTheShape;
    Asura_Animation* const pxAnim = m_pxAnimation;

    Asura_Assert( !(pxShape && pxAnim), "" );
    if( !pxShape && !pxAnim )
    {
        return;
    }

	Asura_Assert(m_xPos.x != Asura_Maths::LargeFloat, "Shape instance position never set, ensure SetNewPosAndOri is called at least once before rendering.");
	if (m_xPos.x == Asura_Maths::LargeFloat)
	{
		return;
	}

	if( pxShape )
    {
		Asura_Assert( pxShape->m_pxRenderData, "" );

		if( pxShape->IsNeverRendered() )
		{
			Asura_Assert(0,"Attempting to render shape flagged as never rendered");
			return;
		}
				
		if(uCamVisFlags == 0)
		{
			// Now test to see if this object is visible in any camera.aw.
			for( u_int uCam = Asura_Camera::GetNumberOfCameras(); uCam--; )
			{
				uCamVisFlags <<= 1;
                
                const bool bIsAttachedToAnim = ( NULL != m_pxParentAnimation );

                if( bIsAttachedToAnim && Asura_Hierarchy_System::ShouldJustRenderForCurCam() && uCam != Asura_Camera::GetCameraIndex() )
                {
                    continue;
                }
                else
                {
				    Asura_Camera::SetCameraIndex( uCam );
                }

                if( !Asura_Camera::IsCameraActive() ) { continue; }

                const Asura_Guid uEntityToIgnore = Asura_Camera::GetEntityToIgnore();

				/// Richard May - 11/11/2002 17:47:46 - skip this camera if this is a 'render only for a single camera'
				/// type shape and this _isn't_ the camera to render it for
				if( (m_uRenderOnlyForThisPlayer != ASURA_GUID_UNREGISTERED) && (uEntityToIgnore != m_uRenderOnlyForThisPlayer) )
				{ 
					continue;
				}

				// RJF - 02 August 2005 - Skip rendering if this shape instance is attached to an anim which is ignored by the current cam
				if( bIsAttachedToAnim && m_pxParentAnimation->GetRootParentEntityGUID() == uEntityToIgnore )
				{
					continue;
				}

				if( uEntityToIgnore == m_uIgnoreForThisPlayer )
                {
                    continue;
                }

#if defined ( ASURA_USE_NEW_VISIBILITY_SYSTEM )

				// KJFL - hmm, currently the visibility system handles the frustum culling too automatically
				// we may want to pass in "ignore frustum culling" as a bool to the Asura_Visibility_Query.
				// KJFL - 18 February 2008 - also we could check for all cameras in one go?
				if ((m_uFlags & ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING) == 0)
				{
					const Asura_Bounding_Box& xAABB = m_xRotatedDisplacedBB;
					if ( !m_xVisibilityQuery.IsAABBVisibleFromCurrentCamera(xAABB) ) continue;
				}

#else // pre Visibility system

#if defined(ASURA_USE_STREAMING)
				if (m_xModuleList.IsValid())
				{
					// MGH 30/5/2006: If the module list is empty then it may just be that
					// the zones were unloaded when we did the module test.
					if (m_xModuleList.IsEmpty())
					{
						// Do the test again, but only if we have loaded a zone since the last test.
						if (Asura_Streaming_System::GetFrameNumberLastZoneLoad() > m_uFrameNumberCalculatedModules)
						{
							CalculateModuleList();
						}
					}
#else
				if (m_ausModuleList[0] != USHRT_MAX)
				{
#endif //defined(ASURA_USE_STREAMING)
					// Module visibility data is valid, do quick rejection test.
					if (((m_uFlags & ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING) == 0) && !TestModuleVisibility())
						continue;
					if (((m_uFlags & ASURA_SHAPE_INSTANCE_NO_FRUSTUM_TEST) == 0) && !IsSphereInFrustum())
						continue;
				}
				else
				{
					// Module visibility data is invalid.
					if (((m_uFlags & ASURA_SHAPE_INSTANCE_NO_FRUSTUM_TEST) == 0) && !IsSphereInFrustum())
						continue;
					// Now, do the slow culling test since our module visibility data is invalid.
                    Asura_Assert(m_xRotatedDisplacedBB.MinX != Asura_Maths::LargeFloat, "bounding box is not initialised");
					if ((m_uFlags & ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING) == 0 &&
#if defined(ASURA_USE_STREAMING)
						!Asura_Module_System::DoesBBIntersectVisibleModule(m_xRotatedDisplacedBB, &m_iLastKnownModule, &m_uLastKnownZone))
#else
						!Asura_Module_System::DoesBBIntersectVisibleModule(m_xRotatedDisplacedBB, &m_iLastKnownModule))
#endif //defined(ASURA_USE_STREAMING)
					{
						continue;
					}
				}

#endif // !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )

				uCamVisFlags |= 1;
			}
		}
        
		if( uCamVisFlags )
		{
		    if( Asura_SortedRender::IsActive() )
		    {
			    //add this to a list of items to be sorted by texture and rendered later
			    Asura_SortedRender::AddShape(this,uCamVisFlags);
		    }
		    else
		    {
                u_int uARGB = Asura_Maths::FloatToInt(255.0f * m_fAlpha);
                uARGB <<= 24;
				Asura_Assert((Asura_Lighting_System::s_uEntityAmbientRGB & 0xff000000) == 0, "Alpha componenet of s_uEntityAmbientRGB should always be zero");
				if(m_uFlags & ASURA_SHAPE_INSTANCE_FULLBRIGHT)
					uARGB |= 0x00808080;
				else
	                uARGB |= (m_uFlags & ASURA_SHAPE_INSTANCE_USES_DEFAULT_AMBIENT) ? Asura_Lighting_System::s_uEntityAmbientRGB : (m_uAmbientARGB & 0x00ffffff);
				pxShape->Project_Render( this, uARGB, uCamVisFlags );

#if _DEBUG || defined(ASURA_TOOLS)
                if( dbg_bShowShapeLOSBBs )
                {
                    if( pxShape->m_pxLOSCollisionData )
                    {
                        Asura_Render::RenderBB( m_pxTheShape->m_pxLOSCollisionData->GetBoundingBox(), m_xPos, m_xOri );
                    }
                }
#endif // #if _DEBUG || defined(ASURA_TOOLS)
		    }
        }
	}
    else
    {
        // (MH - 28.11.2002) handle ambient-only flag (for flashing objects)
        u_int uRenderFlags = 0;
        if (m_uFlags & (ASURA_SHAPE_INSTANCE_AMBIENT_ONLY|ASURA_SHAPE_INSTANCE_FULLBRIGHT))
        {
            uRenderFlags |= ASURA_HIERARCHY_RENDER_FORCE_AMBIENT_COLOUR_ONLY;

            // (MP - 01.06.2006) Hmm, surely we could pass this information
            // on only when the shape flags change, rather than every frame?
			if( m_uFlags & ASURA_SHAPE_INSTANCE_FULLBRIGHT )
            {
                pxAnim->SetRGB( 0.5f, 0.5f, 0.5f );
            }
			else
            {
                u_int uRGB = ( (m_uFlags & ASURA_SHAPE_INSTANCE_USES_DEFAULT_AMBIENT) ? Asura_Lighting_System::s_uEntityAmbientRGB : m_uAmbientARGB );
                pxAnim->SetRGB( ((uRGB >> 16) & 0xFF) / 255.0f, ((uRGB >> 8) & 0xFF) / 255.0f, (uRGB & 0xFF) / 255.0f );
            }
        }
		if(m_uFlags & ASURA_SHAPE_INSTANCE_NO_SHADOW)
		{
			uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_SHADOW;
		}
		if (m_uFlags & ASURA_SHAPE_INSTANCE_IGNORE_MOD_LINKING)
		{
			uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_MODULELINK_TEST;
		}
		if(m_uFlags & ASURA_SHAPE_INSTANCE_BONES_IN_WORLDSPACE)
		{
			uRenderFlags |= ASURA_HIERARCHY_RENDER_BONES_IN_WORLDSPACE;
		}
		if(m_uFlags & ASURA_SHAPE_INSTANCE_NO_FRUSTUM_TEST)
		{
			uRenderFlags |= ASURA_HIERARCHY_RENDER_NO_FRUSTUM_TEST;
		}
       
#ifdef ASURA_USE_SCREENSHOT_SYSTEM
        // (MH - 30.03.2007) if rendering hi-res screenshot must use 0 for anim timestep.
		float fOldAnimTimestep = m_fAnimTimestep;
		if (Asura_ScreenShot_System::ShouldFreezeFrame())
		{
			m_fAnimTimestep = 0.0f;
		}
#endif // #ifdef ASURA_USE_SCREENSHOT_SYSTEM

        // (MH - 11.10.2002) *FIXME* need to separate Update() and Render()
		pxAnim->SetAlpha(m_fAlpha,true);
		if(m_uFlags & ASURA_SHAPE_INSTANCE_USE_ANIM_MOVEMENT)
		{
			//Apply offset from the animation. RAR 28/06/2006
			Asura_Vector_3 xPos = m_xPos;
			Asura_Matrix_3x3 xMat = m_xOri;

			Asura_Vector_3 xPosOffset;
			Asura_Quat xQuatOffset;
			if(pxAnim->GetMovementOffset(&xPosOffset,&xQuatOffset))
			{
				Asura_Matrix_3x3 xMatOffset;
				xQuatOffset.ToMatrix(xMatOffset);

				xPos = xPos + xPosOffset*xMat;
				xMat = xMat * xMatOffset;
			}
        	pxAnim->Update(xPos, xMat, m_fAnimTimestep, uRenderFlags);
						
		}
		else
		{
        	pxAnim->Update(m_xPos, m_xOri, m_fAnimTimestep, uRenderFlags);
		}

#ifdef ASURA_USE_SCREENSHOT_SYSTEM
        // (MH - 30.03.2007) if rendering hi-res screenshot restore the anim timestep.
		if (Asura_ScreenShot_System::ShouldFreezeFrame())
		{
			m_fAnimTimestep = fOldAnimTimestep;
		}
#endif // #ifdef ASURA_USE_SCREENSHOT_SYSTEM

	}

} // End Function 'Asura_Shape_Instance::Render'





#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Instance::AddShadowCallback_Anim
// Author       : Dave Reed
// Date         : 28-Mar-2007
////////////////////////////////////////////////////////////////////////////////

struct Asura_ShapeInstance_ShadowData
{
	Asura_Animation		 *pxAnim;
	Asura_Shape_Instance *pxShapeInst;
	bool				  bUseMovementOffset;
	uint32				  uCamVisFlags;
};

void Asura_Shape_Instance::AddShadowCallback_Anim( const Asura_ShadowInstance_CreationParams& xParams, const void *pxUserData )
{
	const Asura_ShapeInstance_ShadowData *pxData = static_cast<const Asura_ShapeInstance_ShadowData*>(pxUserData);
	Asura_Shadow_Instance_Animation* pxNewShadowInstance = new Asura_Shadow_Instance_Animation (xParams, *pxData->pxAnim, 0, pxData->bUseMovementOffset);
	if ( pxNewShadowInstance )
	{
		if ( Asura_Shadow_System::AddShadowInstance( pxNewShadowInstance ) )
		{
			pxNewShadowInstance->CalculateDistance();
			Asura_Matrix_3x3 xOri;
			pxNewShadowInstance->GetLightOrientation( xOri );
			Asura_Vector_3 xShadowBSCentre;
			float fShadowBSRad = pxNewShadowInstance->GetFrustum().GetBoundingSphere( xShadowBSCentre, xOri, pxNewShadowInstance->GetLightPosition() );
			pxData->pxAnim->AddShadowBSToCoarseBS( fShadowBSRad, xShadowBSCentre );
		}
	}
}


void Asura_Shape_Instance::AddShadowCallback_Shape( const Asura_ShadowInstance_CreationParams& xParams, const void *pxUserData )
{
	const Asura_ShapeInstance_ShadowData *pxData = static_cast<const Asura_ShapeInstance_ShadowData*>(pxUserData);	
	Asura_Shadow_System::AddShadowInstance(new Asura_Shadow_Instance_Shape (xParams, *pxData->pxShapeInst, pxData->uCamVisFlags));
}

#endif

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shape_Instance::RenderShadow
// Author       : Andy Tate
// Date         : 28-Nov-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_Shape_Instance::RenderShadow (const u_int uCamVisFlags)
{
	if((m_uFlags & ASURA_SHAPE_INSTANCE_NO_SHADOW) == 0)
	{
#if defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
		Asura_Vector_3		xPos;

		if(m_pxAnimation != NULL)
		{
			xPos = m_pxAnimation->GetPosition();
		}
		else
		{
			xPos = GetPosition();
		}

		UpdateLighting();

// WGP 02/03/2009 -	Don't do a visibility check here as we need to add the shadow to see whether the shadow is visible,
//					to see if we should update the bone data on the animation so that we can get an accurate shadow.
//					Otherwise shadows bink out when the casting entity goes out of view.
//					If it's not visible then nothing will be rendered for it anyway so it doesn't cost any extra.
//					There will probably be a frame where the shadow gets rendered with old bone data as it's a bit late being calculated here,
//					but I doubt anyone will notice, and it's better than the whole thing binking in.

		if(m_pxTheShape == NULL)
		{
			if(m_pxAnimation == NULL)
			{
				Asura_Assert(false, "No shape or animation to use when adding a shadow for this shape-instance.");
			}
			else
			{
				const bool	bUseMovementOffset	= ((m_uFlags & ASURA_SHAPE_INSTANCE_USE_ANIM_MOVEMENT) != 0);

				Asura_ShapeInstance_ShadowData xData;
				xData.pxAnim = m_pxAnimation;
				xData.bUseMovementOffset = bUseMovementOffset;

				Asura_Shadow_System::AddShadowInstances(xPos, &xData, AddShadowCallback_Anim);
			}
		}
		else
		{
			Asura_ShapeInstance_ShadowData xData;
			xData.pxShapeInst = this;
			xData.uCamVisFlags = uCamVisFlags;			
			Asura_Shadow_System::AddShadowInstances(xPos, &xData, AddShadowCallback_Shape);
		}
#endif	// defined(ASURA_USE_INDEPENDENT_SHADOW_SYSTEM)
	}
}

/// End of function 'Asura_Shape_Instance::RenderShadow'. ///


/////////////////////////////////////////////////////////////////////////////////////////
// Asura_Shape_Instance::LOSTest
//
// Input: LOS data,
//        Position, orientation
//
// Returns: true iff it is a hit and is better than previous hits (ie closer),
//          (if it returns true then the LOS output data is updated).
/////////////////////////////////////////////////////////////////////////////////////////
bool Asura_Shape_Instance::LOSTest(const Asura_Vector_3& xPosition, 
								   const Asura_Matrix_3x3& xOrient,
								   const Asura_LOS_InputData* pxInput,
								   Asura_LOS_OutputData* pxOutput)
{
	Asura_Assert(pxInput && pxOutput, "NULL pointer");
	if (m_pxTheShape && m_pxTheShape->m_pxLOSCollisionData)
	{
		return m_pxTheShape->m_pxLOSCollisionData->LOSTest(xPosition, xOrient, pxInput, pxOutput);
	}
    else if (m_pxAnimation)
    {
		if(m_uFlags & ASURA_SHAPE_INSTANCE_USE_ANIM_MOVEMENT)
		{
			//Apply offset from the animation. RAR 28/06/2006
			Asura_Vector_3 xPos = xPosition;
			Asura_Matrix_3x3 xMat = xOrient;

			Asura_Vector_3 xPosOffset;
			Asura_Quat xQuatOffset;
			if(m_pxAnimation->GetMovementOffset(&xPosOffset,&xQuatOffset))
			{
				Asura_Matrix_3x3 xMatOffset;
				xQuatOffset.ToMatrix(xMatOffset);

				xPos = xPos + xPosOffset*xMat;
				xMat = xMat * xMatOffset;
			}
        	return m_pxAnimation->LOSTest(xPos, xMat, pxInput, pxOutput);
		}
		else
		{
        	return m_pxAnimation->LOSTest(xPosition, xOrient, pxInput, pxOutput);
		}
    }

	return false;
}

int Asura_Shape_Instance::GetTextureIndex() const
{
	if(m_pxTheShape) return m_pxTheShape->GetTextureIndex();
	if(m_pxAnimation)
	{
		const Asura_Hierarchy_Skin* pxSkin = m_pxAnimation->GetSkin();
		if(pxSkin) return pxSkin->Platform_GetTextureIndex();
	}

	return ASURA_TEXTURE_NONE;
}


Asura_Hash_ID Asura_Shape_Instance::GetShapeOrSkinHashID() const
{
	if(m_pxTheShape)
	{
		return m_uShapeHashID;
	}
	if(m_pxAnimation)
	{
        //get the skin , and generate the hash id from the name
        const Asura_Hierarchy_Skin* pxSkin = m_pxAnimation->GetSkin();
        Asura_Assert(pxSkin, "");
        if (pxSkin)
        {
            //return Asura_GetHashID(pxSkin->Name.GetString());
			return pxSkin->m_uNameHashID; 
		}
	}
	return ASURA_HASH_ID_UNSET;
}



bool Asura_Shape_Instance::GetMarkerLocationFromHash( Asura_Hash_ID uMarkerHash, Asura_Vector_3* pxOutputPosition, Asura_Matrix_3x3* pxOutputOrient, bool bWorldSpace/*=true*/, int iAttachID/*=Asura_AnimationAttachment::INVALID_ID*/ ) const
{
	if( !m_pxAnimation ) return false;

	if( !m_pxAnimation->GetMarkerLocationFromHash( uMarkerHash, pxOutputPosition, pxOutputOrient, bWorldSpace, iAttachID ) ) return false;

	return true;
}


#ifdef ASURA_TOOLS
void Asura_Shape::GetShapeFilename( char* szString, unsigned int uMaxLen )
{
	if( m_pxRenderData )
	{
		m_pxRenderData->GetShapeFilename( szString, uMaxLen );
	}
	else
	{
		szString[0] = '\0';
	}
}

void Asura_Shape::GetShapeName( char* szString, unsigned int uMaxLen )
{
	if( m_pxRenderData )
	{
		m_pxRenderData->GetShapeName( szString, uMaxLen );
	}
	else
	{
		szString[0] = '\0';
	}
}
#endif //ASURA_TOOLS




void Asura_Shape_Instance::UpdateLighting() const
{
	if (m_uFlags & (ASURA_SHAPE_INSTANCE_AMBIENT_ONLY|ASURA_SHAPE_INSTANCE_FULLBRIGHT))
	{
		Asura_RGBA xColour;
		if(m_uFlags & ASURA_SHAPE_INSTANCE_FULLBRIGHT)
			xColour.FromUIntARGB(0x00808080);
		else
			xColour.FromUIntARGB(GetAmbientARGB());

		Asura_GI_Solution::s_xSHForCurrentRender.Zero();
		Asura_GI_Solution::s_xSHForCurrentRender.m_afRedCoefficients[0] = xColour.m_fR;
		Asura_GI_Solution::s_xSHForCurrentRender.m_afGreenCoefficients[0] = xColour.m_fG;
		Asura_GI_Solution::s_xSHForCurrentRender.m_afBlueCoefficients[0] = xColour.m_fB;
	}
	else if(m_pxParentAnimation)
	{
		//use the root animation's lighting
		m_pxParentAnimation->UpdateLighting();
	}
	else
	{
        const bool bJustLightOnce = ( (m_uFlags & ASURA_SHAPE_INSTANCE_JUST_LIGHT_ONCE) != 0 );       
        if( !bJustLightOnce || !m_xLightingSolution.IsStaticValid() )
        {
		    m_xLightingSolution.SetPosition( m_xPos );
        }

		bool bIncludeDynamicLights = !bJustLightOnce;
#if defined( ASURA_TARGET_SM3)
		if( !Asura_SortedRender::IsTransparentPass() )
		{
			bIncludeDynamicLights = false;
		}
#endif
		m_xLightingSolution.RecalculateLighting( true, bIncludeDynamicLights );
	}
}

void Asura_Shape_Instance::InvalidateLighting()
{
	m_xLightingSolution.Reset();
}

void Asura_Shape_Instance::SetParentAnimation(Asura_Animation* pxParentAnimation)
{
	if(m_pxAnimation)
	{
		//update parent animation in child anim as well
		Asura_Assert(m_pxAnimation->GetParent() == m_pxParentAnimation,"Something else changing the parent animation??");
		m_pxAnimation->SetParent(pxParentAnimation);
	}
	m_pxParentAnimation = pxParentAnimation;
}

void Asura_Shape_Instance::SetTextureAnimSequence(u_int uSequence)
{
	if(!m_pxAnimation)
	{
		Asura_Assert(0,"SetTextureAnimSequence only supported by animating shape instances");
		return;
	}
	m_pxAnimation->SetTextureAnimSequence(uSequence);
}

void Asura_Shape_Instance::SetTextureAnimSpeed(float fSpeed)
{
	if(!m_pxAnimation)
	{
		Asura_Assert(0,"SetTextureAnimSpeed only supported by animating shape instances");
		return;
	}
	m_pxAnimation->SetTextureAnimSpeed(fSpeed);
}

#ifdef ASURA_TOOLS
void Asura_Shape::SaveShapeChunks( Asura_SaveBlock* pxSaveBlock, const List<Asura_Mesh*>& xMeshList )
{
	//add the chunks to a save block
	Asura_Chunk_Shape::Create( this, *pxSaveBlock, xMeshList );
	Asura_Bounding_Box *pxBB = NULL;
	if (GetFlags() & ASURA_SHAPE_DATA_FLAGS_HAS_CUSTOM_BB)
	{
		pxBB = new Asura_Bounding_Box;
		*pxBB = GetDynamicsCollisionData()->GetBoundingBox();
	}
	Asura_Chunk_ShapeData::Create( GetFlags(), pxBB, GetShapeHashID(),*pxSaveBlock);
	if (pxBB)
	{
		delete pxBB;
	}
		
}

#endif//ASURA_TOOLS
////////////////////////////////////// End of file //////////////////////////////////////
