// Asura_Audio_DSP.cpp
// Thomas Young - 17.03.09
// Copyright (C) 2009 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_AUDIO_SYSTEM

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_Audio_DSP.h"

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

//_______________________________ Asura_Audio_DSPChain ____________________________________

Asura_Audio_DSPChain::Asura_Audio_DSPChain	( )
{
	m_uNumEffects = 0;
	m_ppxEffects = 0;
}


Asura_Audio_DSPChain::~Asura_Audio_DSPChain	( )
{
	Clear();
}


void Asura_Audio_DSPChain::AddEffect ( const Asura_Audio_DSPType eType )
{
	// Stick a new item on the end of the array of effects

	m_uNumEffects += 1;

	Asura_DSP_Params** ppxNewEffects = new Asura_DSP_Params*[ m_uNumEffects ];

	for( u_int uEffect = 0; uEffect < ( m_uNumEffects - 1 ); ++uEffect )
	{
		ppxNewEffects[ uEffect ] = m_ppxEffects[ uEffect ];
	}

	ppxNewEffects[ m_uNumEffects - 1 ] = Asura_DSP_Params::CreateEffect( eType );

	if( m_ppxEffects )
	{
		delete []m_ppxEffects;
	}
	m_ppxEffects = ppxNewEffects;
}


void Asura_Audio_DSPChain::DeleteEffect ( const u_int uIndex )
{
	if( uIndex >= m_uNumEffects )
	{
		Asura_Assert( false, "Invalid index passed to Asura_Audio_DSPChain::DeleteEffect" );
		return;
	}

	if( m_uNumEffects <= 1 )
	{
		Clear();
		return;
	}

	// Remove an effect from the array

	Asura_DSP_Params** ppxNewEffects = new Asura_DSP_Params*[ m_uNumEffects - 1 ];

	u_int uNewIndex = 0;
	for( u_int uEffect = 0; uEffect < m_uNumEffects; ++uEffect )
	{
		if( uEffect != uIndex )
		{
			ppxNewEffects[ uNewIndex ] = m_ppxEffects[ uEffect ];
			++uNewIndex;
		}
	}

	delete []m_ppxEffects;
	m_ppxEffects = ppxNewEffects;

	--m_uNumEffects;
}


void Asura_Audio_DSPChain::Clear ()
{
	for( u_int uEffect = 0; uEffect < m_uNumEffects; ++uEffect )
	{
		delete m_ppxEffects[ uEffect ];
	}
	delete []m_ppxEffects;
	m_ppxEffects = 0;

	m_uNumEffects = 0;
}


Asura_Audio_DSPChain& Asura_Audio_DSPChain::operator= ( const Asura_Audio_DSPChain& xRhs )
{
	Clear();

	// Deep copy our effects

	m_uNumEffects = xRhs.m_uNumEffects;
	
	if( m_uNumEffects > 0 )
	{
		m_ppxEffects = new Asura_DSP_Params*[ m_uNumEffects ];

		for( u_int uEffect = 0; uEffect < m_uNumEffects; ++uEffect )
		{
			m_ppxEffects[ uEffect ] = Asura_DSP_Params::CreateEffect( xRhs.m_ppxEffects[ uEffect ]->GetType() );
			*m_ppxEffects[ uEffect ] = *xRhs.m_ppxEffects[ uEffect ];
		}
	}

	m_xName = xRhs.m_xName;

	return *this;
}


void Asura_Audio_DSPChain::MoveEffectUp ( const u_int uIndex )
{
	if( uIndex >= m_uNumEffects )
	{
		Asura_Assert( false, "Invalid index passed to Asura_Audio_DSPChain::DeleteEffect" );
		return;
	}

	if( uIndex == 0 )
	{
		// Can't move up
		return;
	}

	Asura_DSP_Params* pxEffect = m_ppxEffects[uIndex];

	m_ppxEffects[ uIndex ] = m_ppxEffects[ uIndex - 1 ];
	m_ppxEffects[ uIndex - 1 ] = pxEffect;
}

//_______________________________ Asura_Audio_DSPEffectParams ____________________________________

Asura_DSP_Params* Asura_DSP_Params::CreateEffect ( const Asura_Audio_DSPType eType )
{
	Asura_DSP_Params* pxNewEffect = 0;

	switch( eType )
	{
		case ASURA_DSP_SIMPLE_FILTER:	pxNewEffect = new Asura_DSP_Params_SimpleFilter;		break;
		case ASURA_DSP_ADD_NOISE:		pxNewEffect = new Asura_DSP_Params_AddNoise;			break;
		case ASURA_DSP_FLANGER:			pxNewEffect = new Asura_DSP_Params_Flanger;				break;
		case ASURA_DSP_CHORUS:			pxNewEffect = new Asura_DSP_Params_Chorus;				break;

		case ASURA_DSP_NUM_TYPES:		
		default:						Asura_Assert( false, "Invalid effect enum" );
	}

	return pxNewEffect;
}

//__________________________ Asura_Audio_DSPEffect_Simple_Filter _______________________________


Asura_DSP_Params_SimpleFilter::Asura_DSP_Params_SimpleFilter()
{
	m_xParams.m_uType			= FILTER_TYPE_LOWPASS;
	m_xParams.m_fFrequency		= 120.0f;
	m_xParams.m_fFrequency2		= 0.0f;
	m_xParams.m_fGain			= 1.0f;
	m_xParams.m_fAttenuation	= 0.0f;
	m_xParams.m_bValid			= true;

	Asura_Mem::Set( m_xParams.m_afCoefficients, 0, uFILTER_NUM_COEFFICIENTS * sizeof( float ) );
}


void Asura_DSP_Params_SimpleFilter::WriteToChunkStream ( Asura_Chunk_Stream &xStream ) const
{
	u_int uVersion = uCHUNK_VERSION;
	
	xStream << uVersion;

	xStream << m_xParams.m_uType;
	xStream << m_xParams.m_fFrequency;
	xStream << m_xParams.m_fFrequency2;
	xStream << m_xParams.m_fGain;
	xStream << m_xParams.m_fAttenuation;
	xStream << m_xParams.m_bValid;

	u_int uCoeffs = uFILTER_NUM_COEFFICIENTS;

	xStream << uCoeffs;

	for( u_int uCoeff = 0; uCoeff < uCoeffs; ++uCoeff )
	{
		xStream << m_xParams.m_afCoefficients[ uCoeff ];
	}
}


void Asura_DSP_Params_SimpleFilter::ReadFromChunkStream ( Asura_Chunk_Stream &xStream )
{
	u_int uVersion;

	xStream >> uVersion;

	if( uVersion > uCHUNK_VERSION )
	{
		Asura_Assert( false, "Simple filter DSP chunk version mismatch (out of date code?)");
	}

	xStream >> m_xParams.m_uType;
	xStream >> m_xParams.m_fFrequency;
	xStream >> m_xParams.m_fFrequency2;
	xStream >> m_xParams.m_fGain;
	xStream >> m_xParams.m_fAttenuation;
	xStream >> m_xParams.m_bValid;

	u_int uCoeffs = 0;

	xStream >> uCoeffs;

	Asura_Mem::Set( m_xParams.m_afCoefficients, 0, sizeof( float ) * uFILTER_NUM_COEFFICIENTS );

	if( uCoeffs > uFILTER_NUM_COEFFICIENTS )
	{
		Asura_Assert( false, "Wrong number of coefficients in simple filter chunk, you need to reexport sound data");
	}
	else
	{
		for( u_int uCoeff = 0; uCoeff < uCoeffs; ++uCoeff )
		{
			xStream >> m_xParams.m_afCoefficients[ uCoeff ];
		}
	}
}

#ifdef ASURA_TOOLS

void Asura_DSP_Params_SimpleFilter::ReCalcCoefficients ()
{
	double dCoeffs		[ uFILTER_NUM_COEFFICIENTS ];	// double precision coefficients, converted to float later
	double dBandEdges	[ 6 ];								// Max 6 band edges for bandpass/stop, 4 for low/highpass
	double dGain		[ 3 ];								// Max 3 bands for bandpass/stop, 2 for low/highpass
	double dWeights		[ 3 ];								// as above

	switch( m_xParams.m_uType )
	{
		case FILTER_TYPE_LOWPASS:
		{
			Asura_Assert( m_xParams.m_fFrequency <= 22050.0f, "Frequency should be below the nyquist" );

			dBandEdges[0] = 0.0f;
			dBandEdges[1] = m_xParams.m_fFrequency / 44100.0f;
			dBandEdges[2] = dBandEdges[1] + 0.05f;
			dBandEdges[3] = 0.5f;

			dGain[0] = 1.0f;
			dGain[1] = m_xParams.m_fAttenuation;

			dWeights[0] = 1.0f;
			dWeights[1] = 20.0f;

			m_xParams.m_bValid = RemezExchange( &dCoeffs[0], uFILTER_NUM_COEFFICIENTS, 2, dBandEdges, dGain, dWeights );
		}
		break;

		case FILTER_TYPE_HIGHPASS:
		{
			Asura_Assert( m_xParams.m_fFrequency <= 22050.0f, "Frequency should be below the nyquist" );

			dBandEdges[0] = 0.0f;
			dBandEdges[1] = m_xParams.m_fFrequency / 44100.0f;
			dBandEdges[2] = dBandEdges[1] + 0.05f;
			dBandEdges[3] = 0.5f;

			dGain[0] = m_xParams.m_fAttenuation;
			dGain[1] = 1.0f;

			dWeights[0] = 1.0f;
			dWeights[1] = 20.0f;

			m_xParams.m_bValid = RemezExchange( &dCoeffs[0], uFILTER_NUM_COEFFICIENTS, 2, dBandEdges, dGain, dWeights );
		}
		break;

		case FILTER_TYPE_BANDPASS:
		case FILTER_TYPE_BANDSTOP:
		{
			Asura_Assert( m_xParams.m_fFrequency <= 22050.0f, "Frequency should be below the nyquist" );
			Asura_Assert( m_xParams.m_fFrequency2 <= 22050.0f, "Frequency should be below the nyquist" );
			
			if( m_xParams.m_fFrequency + 1000.0f > m_xParams.m_fFrequency2 )
			{
				m_xParams.m_fFrequency2 = m_xParams.m_fFrequency + 1000.0f;
			}

			const float fNormFreq1 = m_xParams.m_fFrequency / 44100.0f;
			const float fNormFreq2 = m_xParams.m_fFrequency2 / 44100.0f;

			dBandEdges[0] = 0.0f;
			dBandEdges[1] = fNormFreq1;
			dBandEdges[2] = fNormFreq1 + 0.01f;
			dBandEdges[3] = fNormFreq2;
			dBandEdges[4] = fNormFreq2 + 0.01f;
			dBandEdges[5] = 0.5f;

			if( m_xParams.m_uType == FILTER_TYPE_BANDPASS )
			{
				dGain[0] = m_xParams.m_fAttenuation;
				dGain[1] = 1.0f;
				dGain[2] = m_xParams.m_fAttenuation;

				dWeights[0] = 10.0f;
				dWeights[1] = 1.0f;
				dWeights[2] = 10.0f;
			}
			else	// Band stop
			{
				dGain[0] = 1.0f;
				dGain[1] = m_xParams.m_fAttenuation;
				dGain[2] = 1.0f;

				dWeights[0] = 1.0f;
				dWeights[1] = 10.0f;
				dWeights[2] = 1.0f;
			}

			m_xParams.m_bValid = RemezExchange( &dCoeffs[0], uFILTER_NUM_COEFFICIENTS, 3, dBandEdges, dGain, dWeights );
		}
		break;
	}

	if( m_xParams.m_bValid )
	{
		for( int iCoeff = 0; iCoeff < uFILTER_NUM_COEFFICIENTS; ++iCoeff )
		{
			m_xParams.m_afCoefficients[ uFILTER_NUM_COEFFICIENTS - ( iCoeff + 1 ) ] = static_cast<float>( dCoeffs[ iCoeff ] );
		}
	}
}


bool Asura_DSP_Params_SimpleFilter::RemezExchange( double dCoeffs[], int iNumTaps, int iNumBands, double adFreqs[], double adGains[], double adWeights[] )
{
	// TY 01.04.09 - This code for the filter design is adapted from a C port of the Parks-McClellan FORTRAN program (McClellan et al. 1973)
	// NOTE: I have done a little housekeeping but this code is still a bit ugly, especially when it comes to parameter validation

	double *pdGrid, *pdScratchW, *pdScratchD, *pdScratchE, *pdTaps;
	int    i, iIter, iGridSize, iNumExtrema;
	int* piExtrema;
	double dCos;
	double *pdParamX, *pdParamY, *pdParamAD;
	
	iNumExtrema = iNumTaps / 2;		// number of extrema
	if ( iNumTaps % 2 )
	{
		iNumExtrema++;
	}

	// Predict dense grid size in advance for memory allocation, .5 is so we round up, not truncate
	
	iGridSize = 0;
	for ( i = 0; i < iNumBands; i++ )
	{
		iGridSize += static_cast<int>( 2 * iNumExtrema * uREMEZ_GRID_DENSITY * ( adFreqs[ 2 * i + 1 ] - adFreqs[ 2 * i ] ) + 0.5 );
	}

	// Dynamically allocate memory for arrays with proper sizes

	pdGrid		= reinterpret_cast<double *>( Asura_Allocate ( iGridSize * sizeof( double ) ) );
	pdScratchD	= reinterpret_cast<double *>( Asura_Allocate ( iGridSize * sizeof( double ) ) );
	pdScratchW	= reinterpret_cast<double *>( Asura_Allocate ( iGridSize * sizeof( double ) ) );
	pdScratchE	= reinterpret_cast<double *>( Asura_Allocate ( iGridSize * sizeof( double ) ) );
	piExtrema	= reinterpret_cast<int *>	( Asura_Allocate ( ( iNumExtrema + 1 ) * sizeof( int ) ) );
	pdTaps		= reinterpret_cast<double *>( Asura_Allocate ( ( iNumExtrema + 1 ) * sizeof( double ) ) );
	pdParamX	= reinterpret_cast<double *>( Asura_Allocate ( ( iNumExtrema + 1 ) * sizeof( double ) ) );
	pdParamY	= reinterpret_cast<double *>( Asura_Allocate ( ( iNumExtrema + 1 ) * sizeof( double ) ) );
	pdParamAD	= reinterpret_cast<double *>( Asura_Allocate ( ( iNumExtrema + 1 ) * sizeof( double ) ) );

	// Create dense frequency grid

	Remez_CreateDenseGrid( iNumExtrema, iNumTaps, iNumBands, adFreqs, adGains, adWeights, &iGridSize, pdGrid, pdScratchD, pdScratchW );
	Remez_InitialGuess( iNumExtrema, piExtrema, iGridSize );

	if (iNumTaps % 2 == 0)
	{
		for (i=0; i<iGridSize; i++)
		{
			dCos = cos( Asura_Maths::Pi * pdGrid[i] );
			pdScratchD[i] /= dCos;
			pdScratchW[i] *= dCos; 
		}
	}

	// Perform the Remez Exchange algorithm
	for (iIter = 0; iIter < uREMEZ_MAX_ITERATIONS; iIter++)
	{
		Remez_CalcParams(iNumExtrema, piExtrema, pdGrid, pdScratchD, pdScratchW, pdParamAD, pdParamX, pdParamY);
		Remez_CalcError(iNumExtrema, pdParamAD, pdParamX, pdParamY, iGridSize, pdGrid, pdScratchD, pdScratchW, pdScratchE);
		Remez_Search(iNumExtrema, piExtrema, iGridSize, pdScratchE);
		if ( Remez_IsDone( iNumExtrema, piExtrema, pdScratchE ) )
		{
			break;
		}
	}
	if ( iIter == uREMEZ_MAX_ITERATIONS )
	{
		// Failed to converge on a solution :(
		return false;
	}

	Remez_CalcParams(iNumExtrema, piExtrema, pdGrid, pdScratchD, pdScratchW, pdParamAD, pdParamX, pdParamY);

	// Find the 'taps' of the filter for use with Frequency Sampling.  If odd or Negative symmetry, fix the taps
	// according to Parks McClellan
	
	for ( i = 0; i <= iNumTaps / 2; i++)
	{
		if ( iNumTaps % 2 )
		{
			dCos = 1;
		}
		else
		{
			dCos = cos( Asura_Maths::Pi * (double)i/iNumTaps);
		}
		pdTaps[i] = Remez_ComputeA( ( double )i/iNumTaps, iNumExtrema, pdParamAD, pdParamX, pdParamY)*dCos;
	}

	// Frequency sampling design with calculated taps
	Remez_FreqSample( iNumTaps, pdTaps, dCoeffs );

	Asura_Deallocate( pdGrid );
	Asura_Deallocate( pdScratchW );
	Asura_Deallocate( pdScratchD );
	Asura_Deallocate( pdScratchE );
	Asura_Deallocate( piExtrema );
	Asura_Deallocate( pdTaps );
	Asura_Deallocate( pdParamX );
	Asura_Deallocate( pdParamY );
	Asura_Deallocate( pdParamAD );

	return true;
}


bool Asura_DSP_Params_SimpleFilter::Remez_IsDone( int iNumExtrema, int piExtrema[], double pdExtrema[] )
{
	int i;
	double dMin, dMax, dCurrent;

	dMin = dMax = fabs( pdExtrema[piExtrema[0]] );
	for (i = 1; i <= iNumExtrema; i++ )
	{
		dCurrent = fabs(pdExtrema[piExtrema[i]]);
		if (dCurrent < dMin)
		{
			dMin = dCurrent;
		}
		if (dCurrent > dMax)
		{
			dMax = dCurrent;
		}
	}
	if ( ( ( dMax - dMin ) / dMax ) < 0.0001 )
	{
		return true;
	}
	return false;
}


void Asura_DSP_Params_SimpleFilter::Remez_FreqSample( int iNumTaps, double pdTaps[], double pdCoeffs[] )
{
	double x, M;

	M = ( iNumTaps - 1.0 ) / 2.0;
	
	double dLimit = ( iNumTaps / 2 - 1 );
	if ( iNumTaps % 2 )
	{
		dLimit = M;
	}
	for ( int n = 0; n < iNumTaps; n++ )
	{
		double dTap = pdTaps[0];
		x = Asura_Maths::Pi * 2 * (n - M)/iNumTaps;
		for ( int k = 1; k <= dLimit; k++ )
		{
			dTap += 2.0 * pdTaps[k] * cos( x * k );
		}
		pdCoeffs[n] = dTap / iNumTaps;
	}
}


void Asura_DSP_Params_SimpleFilter::Remez_Search( int iNumExtrema, int piExtrema[], int iGridSize, double pdExtrema[] )
{
	int i, j, k, l;		// Counters ( TY - lowercase 'l' variable? really?! )
	int iUp, iAlt;
	int *piFoundExt;	// Array of found extremals

	// Allocate enough space for found extremals.
	piFoundExt = reinterpret_cast<int *>( Asura_Allocate( 10 * iNumExtrema * sizeof( int ) ) );
	k = 0;

	for( i = 0; i < 10 * iNumExtrema; ++i )
	{
		// TY - Bugfix: Mark these 'found extrema' as not valid
		piFoundExt[ i ] = -1;
	}

	// Check for extremum at 0.
	if	( ( ( pdExtrema[0] > 0.0 ) && ( pdExtrema[0] > pdExtrema[1] ) ) ||
		  ( ( pdExtrema[0] < 0.0 ) && ( pdExtrema[0] < pdExtrema[1] ) ) )
	{
		piFoundExt[k++] = 0;
	}

	// Check for extrema inside dense grid
	for (i=1; i<iGridSize-1; i++)
	{
		if	( ( ( pdExtrema[i] >= pdExtrema[i-1] ) && ( pdExtrema[i] > pdExtrema[i+1] ) && ( pdExtrema[i] > 0.0 ) ) ||
			  ( ( pdExtrema[i] <= pdExtrema[i-1] ) && ( pdExtrema[i] < pdExtrema[i+1] ) && ( pdExtrema[i] < 0.0 ) ) )
		{
			piFoundExt[k++] = i;
		}
	}

	// Check for extremum at 0.5
	j = iGridSize - 1;
	if ( ( ( pdExtrema[j] > 0.0 ) && ( pdExtrema[j] > pdExtrema[j-1] ) ) ||
		 ( ( pdExtrema[j] < 0.0 ) && ( pdExtrema[j] < pdExtrema[j-1] ) ) )
	{
		piFoundExt[k++] = j;
	}

	// Remove extra extremals
	int iExtra = k - ( iNumExtrema + 1 );

	while ( iExtra > 0 )
	{
		if (pdExtrema[piFoundExt[0]] > 0.0)
		{
			iUp = 1;	// first one is a maxima
		}
		else
		{
			iUp = 0; // first one is a minima
		}

		l = 0;
		iAlt = 1;
		for ( j = 1; j < k; j++ )
		{
			if (fabs(pdExtrema[piFoundExt[j]]) < fabs(pdExtrema[piFoundExt[l]]))
			{
				l = j;	// new smallest error.
			}
			if ( ( iUp ) && ( pdExtrema[ piFoundExt[j] ] < 0.0 ) )
			{
				iUp = 0;		// switch to a minima
			}
			else if ( ( !iUp ) && ( pdExtrema[ piFoundExt[j] ] > 0.0 ) )
			{
				iUp = 1;		// switch to a maxima
			}
			else
			{ 
				iAlt = 0;
				break;		// Ooops, found two non-alternating extrema.  Delete smallest of them
			}
		} 
		// if the loop finishes, all extrema are alternating

		// If there's only one extremal and all are alternating, delete the smallest of the first/last extremals.
		if ( ( iAlt ) && ( iExtra == 1 ) )
		{
			if (fabs(pdExtrema[piFoundExt[k-1]]) < fabs(pdExtrema[piFoundExt[0]]))
			{
				l = piFoundExt[k-1];	// Delete last extremal
			}
			else
			{
				l = piFoundExt[0];	// Delete first extremal
			}
		}

		for ( j = l; j < k; j++)		// Loop that does the deletion
		{
			piFoundExt[j] = piFoundExt[j+1];
		}
		k--;
		iExtra--;
	}

	for (i=0; i<=iNumExtrema; i++)
	{
		if( piFoundExt[i] != -1 )
		{
			piExtrema[i] = piFoundExt[i];	// Copy found extremals to Ext[] if valid
		}
	}

	Asura_Deallocate( piFoundExt );
}


void Asura_DSP_Params_SimpleFilter::Remez_CalcError	( int iNumExtrema
													, double pdParamD[]
													, double pdParamX[]
													, double pdParamY[]
													, int iGridSize
													, double pdGrid[]
													, double pdScratchD[]
													, double pdScratchW[]
													, double pdScratchE[] )
{
	for ( int i = 0; i < iGridSize; i++ )
	{
		double dResult = Remez_ComputeA( pdGrid[i], iNumExtrema, pdParamD, pdParamX, pdParamY );
		pdScratchE[i] = pdScratchW[i] * ( pdScratchD[i] - dResult );
	}
}


double Asura_DSP_Params_SimpleFilter::Remez_ComputeA( double dFreq, int iNumExtrema, double pdParamD[], double pdParamX[], double pdParamY[] )
{
	int i;
	double dCosX, c, dDenom, dNumer;

	dDenom = dNumer = 0;
	dCosX = cos(Asura_Maths::Pi * 2 * dFreq);
	for (i=0; i<=iNumExtrema; i++)
	{
		c = dCosX - pdParamX[i];
		if ( fabs(c) < 1.0e-7 )
		{
			dNumer = pdParamY[i];
			dDenom = 1;
			break;
		}
		c = pdParamD[i]/c;
		dDenom += c;
		dNumer += c * pdParamY[i];
	}
	return dNumer/dDenom;
}


void Asura_DSP_Params_SimpleFilter::Remez_CalcParams	( int iNumExtrema
														, int piExtrema[]
														, double pdGrid[]
														, double pdParamD[]
														, double pdParamW[]
														, double pdParamAD[]
														, double pdParamX[]
														, double pdParamY[] )
{
	int i, j, k, ld;
	double dSign, xi, dDelta, dDenom, dNumer;

	// Find x[]
	for (i = 0; i <= iNumExtrema; i++)
	{
		pdParamX[i] = cos( Asura_Maths::Pi * 2 * pdGrid[ piExtrema[i] ] );
	}

	// Calculate ad[]  - Oppenheim & Schafer eq 7.132
	ld = ( iNumExtrema-1 ) / 15 + 1;	// Skips around to avoid round errors
	for ( i = 0; i <= iNumExtrema; i++ )
	{
		dDenom = 1.0;
		xi = pdParamX[i];
		for (j=0; j<ld; j++)
		{
			for (k = j; k <= iNumExtrema; k += ld )
			{
				if (k != i)
				{
					dDenom *= 2.0 * ( xi - pdParamX[k] );
				}
			}
		}
		if ( fabs(dDenom) < 0.00001 )
		{
			dDenom = 0.00001;
		}
		pdParamAD[i] = 1.0 / dDenom;
	}

	// Calculate delta  - Oppenheim & Schafer eq 7.131
	dNumer = dDenom = 0;
	dSign = 1;
	for ( i = 0; i <= iNumExtrema; i++ )
	{
		dNumer += pdParamAD[i] * pdParamD[piExtrema[i]];
		dDenom += dSign * pdParamAD[i] / pdParamW[piExtrema[i]];
		dSign = -dSign;
	}
	dDelta = dNumer / dDenom;
	dSign = 1;

	// Calculate y[]  - Oppenheim & Schafer eq 7.133b
	for ( i = 0; i <= iNumExtrema; i++ )
	{
		pdParamY[i] = pdParamD[piExtrema[i]] - dSign * dDelta / pdParamW[piExtrema[i]];
		dSign = -dSign;
	}
}


void Asura_DSP_Params_SimpleFilter::Remez_InitialGuess( int iNumExtrema, int piExtrema[], int iGridsize )
{
	for ( int i = 0; i <= iNumExtrema; i++ )
	{
		piExtrema[i] = i * ( iGridsize - 1 ) / iNumExtrema;
	}
}


void Asura_DSP_Params_SimpleFilter::Remez_CreateDenseGrid	( int iNumExtrema
															, int iNumTaps
															, int iNumBands
															, double dBands[]
															, double dDesired[]
															, double dWeights[]
															, int *piGridSize
															, double dGrid[]
															, double pScratchD[]
															, double pScratchW[] )
{
	int i, j, k, iBand;
	double dDelF, dLowF, dHighF;

	dDelF = 0.5 / ( uREMEZ_GRID_DENSITY * iNumExtrema );

	j=0;
	for ( iBand=0; iBand < iNumBands; iBand++ )
	{
		dGrid[j]	= dBands[ 2 * iBand ];
		dLowF		= dBands[ 2 * iBand ];
		dHighF		= dBands[ 2 * iBand + 1];
		
		k = static_cast<int>( ( dHighF - dLowF ) / dDelF + 0.5 );	// .5 for rounding
		
		for ( i = 0; i < k; i++ )
		{
			pScratchD[j] = dDesired[iBand];
			pScratchW[j] = dWeights[iBand];
			dGrid[j] = dLowF;
			dLowF += dDelF;
			j++;
		}
		dGrid[ j-1 ] = dHighF;
	}
}

#endif


//__________________________ Asura_DSP_Params_AddNoise _______________________________

Asura_DSP_Params_AddNoise::Asura_DSP_Params_AddNoise()
{
	m_xParams.m_fNoiseAmount = 0.0f;
}

void Asura_DSP_Params_AddNoise::WriteToChunkStream ( Asura_Chunk_Stream &xStream ) const
{
	u_int uVersion = uCHUNK_VERSION;

	xStream << uVersion;

	xStream << m_xParams.m_fNoiseAmount;
}

void Asura_DSP_Params_AddNoise::ReadFromChunkStream ( Asura_Chunk_Stream &xStream )
{
	u_int uVersion = 0;

	xStream >> uVersion;

	if( uVersion > uCHUNK_VERSION )
	{
		Asura_Assert( false, "Add Noise DSP chunk version mismatch (need new code?)");
		return;
	}

	xStream >> m_xParams.m_fNoiseAmount;
}

//__________________________ Asura_DSP_Params_Flanger _______________________________

Asura_DSP_Params_Flanger::Asura_DSP_Params_Flanger()
{
	m_xParams.m_fWetDry = 1.0f;
	m_xParams.m_fRange = 1.0f;
	m_xParams.m_fGain = 1.0f;
}


void Asura_DSP_Params_Flanger::WriteToChunkStream ( Asura_Chunk_Stream &xStream ) const
{
	u_int uVersion = uCHUNK_VERSION;

	xStream << uVersion;
	xStream << m_xParams.m_fRange;
	xStream << m_xParams.m_fWetDry;
	xStream << m_xParams.m_fGain;
}


void Asura_DSP_Params_Flanger::ReadFromChunkStream ( Asura_Chunk_Stream &xStream )
{
	u_int uVersion = 0;

	xStream >> uVersion;

	if( uVersion > uCHUNK_VERSION )
	{
		Asura_Assert( false, "Flanger DSP chunk version mismatch (need new code?)");
		return;
	}

	xStream >> m_xParams.m_fRange;
	xStream >> m_xParams.m_fWetDry;
	xStream >> m_xParams.m_fGain;
}

//__________________________ Asura_DSP_Params_Chorus _______________________________

Asura_DSP_Params_Chorus::Asura_DSP_Params_Chorus()
{
	m_xParams.m_fDelay			= 20.0f;
	m_xParams.m_fSweepDepth		= 5.0f;
	m_xParams.m_fWetDry			= 1.0f;
	m_xParams.m_fGain			= 1.0f;
}


void Asura_DSP_Params_Chorus::WriteToChunkStream ( Asura_Chunk_Stream &xStream ) const
{
	u_int uVersion = uCHUNK_VERSION;

	xStream << uVersion;
	xStream << m_xParams.m_fDelay;
	xStream << m_xParams.m_fSweepDepth;
	xStream << m_xParams.m_fWetDry;
	xStream << m_xParams.m_fGain;

}


void Asura_DSP_Params_Chorus::ReadFromChunkStream ( Asura_Chunk_Stream &xStream )
{
	u_int uVersion = 0;

	xStream >> uVersion;

	if( uVersion > uCHUNK_VERSION )
	{
		Asura_Assert( false, "Chorus DSP chunk version mismatch (need new code?)");
		return;
	}

	xStream >> m_xParams.m_fDelay;
	xStream >> m_xParams.m_fSweepDepth;
	xStream >> m_xParams.m_fWetDry;
	xStream >> m_xParams.m_fGain;
}


#endif //ASURA_USE_AUDIO_SYSTEM

//eof
