//////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

#include "Sound/Asura_Audio_System.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_String.h"
#include "Core/Resources/Asura_ResourceSet.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_Debug_Timers.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Guid.h"
#include "Sound/Asura_Audio_Voice.h"
#include "Sound/Asura_Audio_DSP.h"

/******************************************************************************/
/* Global variables                                                           */
/******************************************************************************/

static const float fASURA_AUDIO_REVERB_TRANSITION_UNSET = -1.0f;
static const float fASURA_AUDIO_REVERB_TRANSITION_SPEED = 0.5f;

//we can make this an array of Asura_Vector_3 if we want to support a 3D sound field
static const Asura_Vector_2 g_axSPEAKER_ANGLES[ Asura_Audio_System::ASURA_NUM_SPEAKERS ] = {

		Asura_Vector_2(	-0.7071f,	0.7071f	)	,	//ASURA_SPEAKER_FRONT_LEFT
		Asura_Vector_2(	0.7071f	,	0.7071f	)	,	//ASURA_SPEAKER_FRONT_RIGHT
		Asura_Vector_2(	0.0f	,	1.0f	)	,	//ASURA_SPEAKER_FRONT_CENTRE
		Asura_Vector_2(	0.0f	,	0.0f	)	,	//ASURA_SPEAKER_LOW_FREQ
		Asura_Vector_2(	-0.7071f, -0.7071f	)	,	//ASURA_SPEAKER_REAR_LEFT
		Asura_Vector_2(	0.7071f	, -0.7071f	)	,	//ASURA_SPEAKER_REAR_RIGHT
		Asura_Vector_2(	0.0f	, -1.0f		)	,	//ASURA_SPEAKER_REAR_CENTRE
		Asura_Vector_2(	-1.0f	, 0.0f		)	,	//ASURA_SPEAKER_SIDE_LEFT
		Asura_Vector_2(	1.0f	, 0.0f		)	,	//ASURA_SPEAKER_SIDE_RIGHT
		Asura_Vector_2(	-0.87f	, 0.5f		)	,	//ASURA_SPEAKER_FRONT_CENTRE_LEFT		made this one up
		Asura_Vector_2(	0.87f	, 0.5f		)	};	//ASURA_SPEAKER_FRONT_CENTRE_RIGHT		...and this one

DVSVAR float g_fDistanceFilterRangeMax	= 150.0f;
DVSVAR float g_fDistanceFilterRangeMin	= 50.0f;

/******************************************************************************/
/* DVS Var definitions                                                        */
/******************************************************************************/

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

#define CONSOLE_VARNAME_DISTANCE_FILTER_RANGE_MIN	L"SoundSystem.Distance Filter.MinRange"
#define CONSOLE_VARNAME_DISTANCE_FILTER_RANGE_MAX	L"SoundSystem.Distance Filter.MaxRange"

#endif

/******************************************************************************/
/* Asura_Audio_System Static Variable Definitions                             */
/******************************************************************************/

u_int Asura_Audio_System::s_uActiveSpeakers = 0;
int Asura_Audio_System::s_iNumActiveSpeakers = 0;
float Asura_Audio_System::s_fMaxAngleBetweenSpeakers = 0.0f;

float Asura_Audio_System::s_fOcclusionCPUTimePerFrame = 0.001f;	// 1ms by default
bool Asura_Audio_System::s_bShouldUpdateOcclusion = true;
Asura_FixedArray<Asura_Guid, ASURA_MAX_NO_OF_LISTENERS> Asura_Audio_System::s_auOcclusionGUIDsToIgnore;
Asura_Audio_OcclusionFactor_Callback Asura_Audio_System::s_pfnGetEntityOcclusionFactor = 0;
Asura_Audio_OcclusionFactor_Callback Asura_Audio_System::s_pfnGetEnvironmentOcclusionFactor = 0;

Asura_TemplateHashTree< Asura_Audio_DSPChain > Asura_Audio_System::s_xDSPChains;
Asura_TemplateHashTree< Asura_Audio_StreamInfo > Asura_Audio_System::s_xStreamInfoTree;
Asura_ResourceSet< Asura_Audio_Reverb > Asura_Audio_System::s_xReverbs( "AudioReverb" );

Asura_Audio_Reverb Asura_Audio_System::s_xCurrReverb;
float Asura_Audio_System::s_fCurrReverbAmount = 0.0f;
Asura_Audio_Reverb Asura_Audio_System::s_xReverbBuffers[2];
float Asura_Audio_System::s_fReverbAmountBuffers[2];
float Asura_Audio_System::s_fReverbTransition = fASURA_AUDIO_REVERB_TRANSITION_UNSET;
int Asura_Audio_System::s_iCurrentReverbBuffer = 0;

u_int Asura_Audio_System::s_uNumListeners = 1;
bool Asura_Audio_System::s_abListenerActive[ ASURA_MAX_NO_OF_LISTENERS ] = { true, false, false, false };
bool Asura_Audio_System::s_abListenerUsesCurrentCamera[ ASURA_MAX_NO_OF_LISTENERS ] = { true, false, false, false };
int Asura_Audio_System::s_aiListenerCameraIndex[ ASURA_MAX_NO_OF_LISTENERS ] = { -1, -1, -1, -1 };
Asura_Vector_3 Asura_Audio_System::s_axListenerPosition[ ASURA_MAX_NO_OF_LISTENERS ];
Asura_Matrix_3x3 Asura_Audio_System::s_axListenerOrientation[ ASURA_MAX_NO_OF_LISTENERS ];

/******************************************************************************/
/* Asura_Audio_System Static Class Functions                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::Initialise()
//
// Initialises the audio system
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::Initialise()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Audio_System::Initialise" );
	LoadAsuraFile( szASURA_AUDIO_STREAMING_FILENAME );
	#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_DISTANCE_FILTER_RANGE_MIN	, &g_fDistanceFilterRangeMin, 0.0f, 200.0f	, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_DISTANCE_FILTER_RANGE_MAX	, &g_fDistanceFilterRangeMax, 0.0f, 400.0f	, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif

	for( int iCam = 0; iCam < ASURA_MAX_NO_OF_LISTENERS; ++iCam )
	{
		s_auOcclusionGUIDsToIgnore[ iCam ] = ASURA_GUID_SERVER;
	}

	Platform_Initialise();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::Reset()
//
// Resets the audio system
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::Reset()
{
	Platform_Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::Shutdown()
//
// Shuts down the audio system
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::Shutdown()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Audio_System::Shutdown" );
	Platform_Shutdown();
	s_xReverbs.Reset();
	s_xDSPChains.Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::Update()
//
// Updates the audio system each frame
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::Update()
{
	Asura_Audio_Voice::PreUpdate();

	// Update our reverb
	if( s_fReverbTransition != fASURA_AUDIO_REVERB_TRANSITION_UNSET )
	{
		s_fReverbTransition += fASURA_AUDIO_REVERB_TRANSITION_SPEED * Asura_Timers::GetFrameTime();

		if( s_fReverbTransition >= 1.0f )
		{
			// Finished transitioning
			s_xCurrReverb = s_xReverbBuffers[ s_iCurrentReverbBuffer ];
			s_fCurrReverbAmount = s_fReverbAmountBuffers[ s_iCurrentReverbBuffer ];
			s_fReverbTransition = fASURA_AUDIO_REVERB_TRANSITION_UNSET;
		}
		else
		{
			// Update the current reverb by lerping between our reverb buffers
			Asura_Audio_Reverb* pxDestReverb	= &s_xReverbBuffers[ s_iCurrentReverbBuffer ];
			Asura_Audio_Reverb* pxSourceReverb	= &s_xReverbBuffers[ !s_iCurrentReverbBuffer ];

			s_xCurrReverb = Asura_Audio_Reverb( pxSourceReverb, pxDestReverb, s_fReverbTransition );
			s_fCurrReverbAmount = Asura_Maths::Lerp( s_fReverbAmountBuffers[ !s_iCurrentReverbBuffer ], s_fReverbAmountBuffers[ s_iCurrentReverbBuffer ], s_fReverbTransition );
		}

		/* // Useful for debugging reverb
		#ifdef ASURA_USE_DEBUG_TEXT
		Asura_DebugText::AddTextBlock( 15, 15, 0xffffff00, 0, "Room         : %d "	, s_xCurrReverb.m_iRoom );
		Asura_DebugText::AddTextBlock( 15, 16, 0xffffff00, 0, "Room HF      : %d "	, s_xCurrReverb.m_iRoomHF );
		Asura_DebugText::AddTextBlock( 15, 17, 0xffffff00, 0, "Room Rolloff : %.2f ", s_xCurrReverb.m_fRoomRolloffFactor );
		Asura_DebugText::AddTextBlock( 15, 18, 0xffffff00, 0, "Decay Time   : %.2f ", s_xCurrReverb.m_fDecayTime );
		Asura_DebugText::AddTextBlock( 15, 19, 0xffffff00, 0, "Decay HF R   : %.2f ", s_xCurrReverb.m_fDecayHFRatio );
		Asura_DebugText::AddTextBlock( 15, 20, 0xffffff00, 0, "Reflections  : %d "	, s_xCurrReverb.m_iReflections );
		Asura_DebugText::AddTextBlock( 15, 21, 0xffffff00, 0, "Reflections D: %.2f ", s_xCurrReverb.m_fReflectionsDelay );
		Asura_DebugText::AddTextBlock( 15, 22, 0xffffff00, 0, "Reverb       : %d "	, s_xCurrReverb.m_iReverb );
		Asura_DebugText::AddTextBlock( 15, 23, 0xffffff00, 0, "Reverb Delay : %.2f ", s_xCurrReverb.m_fReverbDelay );
		Asura_DebugText::AddTextBlock( 15, 24, 0xffffff00, 0, "Diffusion    : %.2f ", s_xCurrReverb.m_fDiffusion );
		Asura_DebugText::AddTextBlock( 15, 25, 0xffffff00, 0, "Density      : %.2f ", s_xCurrReverb.m_fDensity );
		Asura_DebugText::AddTextBlock( 15, 26, 0xffffff00, 0, "HF Ref       : %.2f ", s_xCurrReverb.m_fHFReference );
		#endif
		*/

		Asura_Assertf( ( s_xCurrReverb.m_iRoom				>= -10000	) && ( s_xCurrReverb.m_iRoom				<= 0		), ("Reverb parameter 'm_iRoom' is invalid (%d)", s_xCurrReverb.m_iRoom ));
		Asura_Assertf( ( s_xCurrReverb.m_iRoomHF			>= -10000	) && ( s_xCurrReverb.m_iRoomHF				<= 0		), ("Reverb parameter 'm_iRoomHF' is invalid (%d)", s_xCurrReverb.m_iRoomHF ));
		Asura_Assertf( ( s_xCurrReverb.m_fRoomRolloffFactor >= 0.0f		) && ( s_xCurrReverb.m_fRoomRolloffFactor	<= 10.0f	), ("Reverb parameter 'm_fRoomRolloffFactor' is invalid (%f)", s_xCurrReverb.m_fRoomRolloffFactor ));
		Asura_Assertf( ( s_xCurrReverb.m_fDecayTime			>= 0.1f		) && ( s_xCurrReverb.m_fDecayTime			<= 20.0f	), ("Reverb parameter 'm_fDecayTime' is invalid (%f)", s_xCurrReverb.m_fDecayTime ));
		Asura_Assertf( ( s_xCurrReverb.m_fDecayHFRatio		>= 0.1f		) && ( s_xCurrReverb.m_fDecayHFRatio		<= 2.0f		), ("Reverb parameter 'm_fDecayHFRatio' is invalid (%f)", s_xCurrReverb.m_fDecayHFRatio ));
		Asura_Assertf( ( s_xCurrReverb.m_iReflections		>= -10000	) && ( s_xCurrReverb.m_iReflections			<= 1000		), ("Reverb parameter 'm_iReflections' is invalid (%d)", s_xCurrReverb.m_iReflections ));
		Asura_Assertf( ( s_xCurrReverb.m_fReflectionsDelay	>= 0.0f		) && ( s_xCurrReverb.m_fReflectionsDelay	<= 0.3f		), ("Reverb parameter 'm_fReflectionsDelay' is invalid (%f)", s_xCurrReverb.m_fReflectionsDelay ));
		Asura_Assertf( ( s_xCurrReverb.m_iReverb			>= -10000	) && ( s_xCurrReverb.m_iReverb				<= 2000		), ("Reverb parameter 'm_iReverb' is invalid (%d)", s_xCurrReverb.m_iReverb ));
		Asura_Assertf( ( s_xCurrReverb.m_fReverbDelay		>= 0.0f		) && ( s_xCurrReverb.m_fReverbDelay			<= 0.1f		), ("Reverb parameter 'm_fReverbDelay' is invalid (%f)", s_xCurrReverb.m_fReverbDelay ));
		Asura_Assertf( ( s_xCurrReverb.m_fDiffusion			>= 0.0f		) && ( s_xCurrReverb.m_fDiffusion			<= 100.0f	), ("Reverb parameter 'm_fDiffusion' is invalid (%f)", s_xCurrReverb.m_fDiffusion ));
		Asura_Assertf( ( s_xCurrReverb.m_fDensity			>= 0.0f		) && ( s_xCurrReverb.m_fDensity				<= 100.0f	), ("Reverb parameter 'm_fDensity' is invalid (%f)", s_xCurrReverb.m_fDensity ));
		Asura_Assertf( ( s_xCurrReverb.m_fHFReference		>= 20.0f	) && ( s_xCurrReverb.m_fHFReference			<= 20000.0f	), ("Reverb parameter 'm_fHFReference' is invalid (%f)", s_xCurrReverb.m_fHFReference ));
		
	}

	Platform_Update();
	Asura_Audio_Voice::PostUpdate();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::CalcVolFalloff()
//
// Calculates the volume falloff factor linearly between inner & outer ranges
//
////////////////////////////////////////////////////////////////////////////////

float Asura_Audio_System::CalcVolFalloff( const float fInnerRadius, const float fOuterRadius,
	const float fDistance )
{
	const float fRange = fOuterRadius - fInnerRadius;
	const float fProportion = (fOuterRadius - fDistance) / fRange;
	return Asura_Maths::ClampToRange( fProportion, 0.0f, 1.0f );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::CalcVolFalloff()
//
// Calculates the volume falloff for inner & outer cuboids. Position passed
// relative to the sound source
//
////////////////////////////////////////////////////////////////////////////////

float Asura_Audio_System::CalcVolFalloff( const Asura_Vector_3& xInnerRange,
	const Asura_Vector_3& xOuterRange, const Asura_Vector_3& xPos )
{
	float afFalloff[3];

	for( u_int u = 0; u < 3; ++u )
	{
		const float fInner = xInnerRange.GetElement(u);
		const float fOuter = xOuterRange.GetElement(u);

		const float fAbsDist = Asura_Maths::Abs( xPos.GetElement(u) );
		if( fAbsDist > fOuter ) return 0.0f;

		afFalloff[u] = CalcVolFalloff( fInner, fOuter, fAbsDist );
	}

	return Asura_Maths::Min( Asura_Maths::Min( afFalloff[0], afFalloff[1] ), afFalloff[2] );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::LoadWave()
//
// Loads an audio wave from a chunk stream
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Audio_System::LoadWave( Asura_Chunk_Stream& xStream, const u_int uFileSize,
	const u_int uSubType, const u_int uFlags, const char* const szFilename )
{
	return Platform_LoadWave( xStream, uFileSize, uSubType, uFlags, szFilename );
}

#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::LoadWave()
//
// Loads an audio wave from a file. Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Audio_System::LoadWave( const char* const szFilename, const bool bLooping )
{
	return Platform_LoadWave( szFilename, bLooping );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::ForceWaveToBeLoaded()
//
// Forces a delayed wave to be loaded now. Only available in Tool builds
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::ForceWaveToBeLoaded( const Asura_Hash_ID uWave )
{
	Platform_ForceWaveToBeLoaded( uWave );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::ReloadWave()
//
// Reloads a wave file. Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::ReloadWave( const Asura_Hash_ID uWave, const bool bForceReloadNow )
{
	Platform_ReloadWave( uWave, bForceReloadNow );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::SetWaveLooping()
//
// Dynamically changes a wave's looping property. Only needed in Tools builds,
// since retail configs don't change it after loading from chunks
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::SetWaveLooping( const Asura_Hash_ID uWave, const bool bLoop )
{
	if( uWave != ASURA_HASH_ID_UNSET )
	{
		Platform_SetWaveLooping( uWave, bLoop );
	}
}

#endif // defined ASURA_TOOLS

#if defined ASURA_AUDIO_STORE_FILENAMES

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetWaveFilename()
//
// Retrieves the filename of the given audio wave
//
////////////////////////////////////////////////////////////////////////////////

const char* Asura_Audio_System::GetWaveFilename( const Asura_Hash_ID uWave )
{
	return Platform_GetWaveFilename( uWave );
}

#endif // defined ASURA_AUDIO_STORE_FILENAMES

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::CreateDSPChain()
//
// Creates a DSP chain, initialised from the given essence
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::AddDSPChain( const Asura_Hash_ID uHash, const Asura_Audio_DSPChain* pxChain )
{
	Asura_Audio_DSPChain* pxNewChain = s_xDSPChains.AddNode( uHash );

	*pxNewChain = *pxChain;

	//Platform_CreateDSPChain( uHash, pxChain );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::RemoveDSPChain()
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::RemoveDSPChain ( const Asura_Hash_ID uHash )
{
	s_xDSPChains.DeleteNodeByHash( uHash );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetNumDSPChains()
//
////////////////////////////////////////////////////////////////////////////////

int Asura_Audio_System::GetNumDSPChains ( )
{
	return s_xDSPChains.GetNumberOfNodes( );
}

#ifdef ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetDSPChainByIndex()
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_DSPChain* Asura_Audio_System::GetDSPChainByIndex ( const int iIndex )
{
	return s_xDSPChains.GetDataByIndex( iIndex );
}

#endif

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetDSPChainByHash()
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_DSPChain* Asura_Audio_System::GetDSPChainByHash ( const Asura_Hash_ID uHash )
{
	return s_xDSPChains.GetDataByHash( uHash );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::SetMasterDSP()
//
// Sets the master DSP chain to apply to voices which use DSP
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::SetMasterDSP( const Asura_Hash_ID uDSPChain, const float fFadeTime )
{
	Platform_SetMasterDSP( uDSPChain, fFadeTime );
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::CreateVoice()
//
// Creates an audio voice, initialised to play the given wave
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_Audio_System::CreateVoice( const Asura_Hash_ID uWave, const bool bFilter,
	const bool bMusic, const Asura_Hash_ID uDSPChain /*= ASURA_HASH_ID_UNSET*/ )
{
	return Platform_CreateVoice( uWave, bFilter, bMusic, uDSPChain );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::CreateStreamingVoice()
//
// Creates an audio voice, initialised to play the given stream from disc.
// Optionally provide a time offset from the beginning of the sound, in seconds
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_Audio_System::CreateStreamingVoice( const Asura_Hash_ID uStreamHash, const bool bFilter,
	const bool bMusic, const float fTimeOffset /*= 0.0f*/, const Asura_Hash_ID uDSPChain /*= ASURA_HASH_ID_UNSET*/ )
{
	return Platform_CreateStreamingVoice( uStreamHash, bFilter, bMusic, fTimeOffset, uDSPChain );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::CreateSubmixVoice()
//
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_Audio_System::CreateSubmixVoice( const bool bFilter, const bool bMusic, const Asura_Hash_ID uDSPChain )
{
	return Platform_CreateSubmixVoice( bFilter, bMusic, uDSPChain );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::DestroyVoice()
//
// Destroys an audio voice and sets its handle to ASURA_INVALID_HANDLE
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::DestroyVoice( Asura_Handle& iVoiceHandle )
{
	// The old sound system didn't assert here so we won't either
	if( iVoiceHandle == ASURA_INVALID_HANDLE ) return;

	Platform_DestroyVoice( iVoiceHandle );

	// Again, follow the old sound system
	iVoiceHandle = ASURA_INVALID_HANDLE;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetVoice()
//
// Retrieves an audio playback voice by handle
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_Voice* Asura_Audio_System::GetVoice( const Asura_Handle iVoiceHandle )
{
	return (iVoiceHandle == ASURA_INVALID_HANDLE ? NULL : Platform_GetVoice( iVoiceHandle ));
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::SetSpeakerActive()
//
// Set a speaker as active for output
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::SetSpeakerActive ( Speakers eSpeaker )
{
	Asura_Assert( eSpeaker != ASURA_NUM_SPEAKERS, "Invalid speaker" );

	s_uActiveSpeakers |= ( 1 << static_cast<u_int>( eSpeaker ) );

	//Update the speaker bleed (this is defined by the maximum angle between 2 speakers)
	
	//Make an array of speakers in the order they would appear on a circle around the user
	Speakers aeSpeakerOrder[10] =	{	ASURA_SPEAKER_FRONT_CENTRE
									,	ASURA_SPEAKER_FRONT_CENTRE_RIGHT
									,	ASURA_SPEAKER_FRONT_RIGHT
									,	ASURA_SPEAKER_SIDE_RIGHT
									,	ASURA_SPEAKER_REAR_RIGHT
									,	ASURA_SPEAKER_REAR_CENTRE
									,	ASURA_SPEAKER_REAR_LEFT
									,	ASURA_SPEAKER_SIDE_LEFT
									,	ASURA_SPEAKER_FRONT_LEFT
									,	ASURA_SPEAKER_FRONT_CENTRE_LEFT };

	//Find the first active speaker
	int iFirstSpeaker	= -1;
	int iPrevSpeaker	= -1;

	s_iNumActiveSpeakers = 0;
	float fMaxAngle = 0.0f;

	for( int iSpeaker = 0; iSpeaker < 10; ++iSpeaker )
	{
		const int iCurrentSpeaker = aeSpeakerOrder[ iSpeaker ];
		if( IsSpeakerActive( static_cast<Speakers>( iCurrentSpeaker ) ) )
		{
			if( iFirstSpeaker == -1 )
			{
				iFirstSpeaker	= iCurrentSpeaker;
				iPrevSpeaker	= iCurrentSpeaker;
			}
			else
			{
				//Get the angle between this and the previous speaker
				Asura_Vector_2 xSpeakerPos1 = GetSpeakerPosition( static_cast<Speakers>( iPrevSpeaker	) );
				Asura_Vector_2 xSpeakerPos2 = GetSpeakerPosition( static_cast<Speakers>( iCurrentSpeaker) );

				const float fDot = xSpeakerPos1 * xSpeakerPos2;
				float fAngle = Asura_Maths::ArcCosine( fDot );
				
				//we need to take the cross as well to know if it is acute (optimised a bit)
				const float fNormal =	- xSpeakerPos1.x * ( xSpeakerPos2.y - xSpeakerPos1.y )
										+ xSpeakerPos1.y * ( xSpeakerPos2.x - xSpeakerPos1.x );

				if( fNormal < 0.0f )
				{
					fAngle += 3.14159f;
				}

				fMaxAngle = Asura_Maths::Max( fMaxAngle, fAngle );

				iPrevSpeaker = iCurrentSpeaker;
			}
		
			++s_iNumActiveSpeakers;
		}
	}

	//Also account for the seperation of the final speaker with the first one
	Asura_Vector_2 xSpeakerPos1 = GetSpeakerPosition( static_cast<Speakers>( iPrevSpeaker	) );
	Asura_Vector_2 xSpeakerPos2 = GetSpeakerPosition( static_cast<Speakers>( iFirstSpeaker	) );

	const float fDot = xSpeakerPos1 * xSpeakerPos2;
	float fAngle = Asura_Maths::ArcCosine( fDot );
	
	const float fNormal =	- xSpeakerPos1.x * ( xSpeakerPos2.y - xSpeakerPos1.y )
							+ xSpeakerPos1.y * ( xSpeakerPos2.x - xSpeakerPos1.x );

	if( fNormal < 0.0f )
	{
		fAngle += 3.14159f;
	}

	fMaxAngle = Asura_Maths::Max( fMaxAngle, fAngle );
	
	s_fMaxAngleBetweenSpeakers = fMaxAngle;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetSpeakerPosition()
//
// Get the position of a speaker as a vector from the listener
//
////////////////////////////////////////////////////////////////////////////////

Asura_Vector_2	Asura_Audio_System::GetSpeakerPosition ( Speakers eSpeaker )
{
	Asura_Assert( eSpeaker != ASURA_NUM_SPEAKERS, "Invalid speaker" );

	return g_axSPEAKER_ANGLES[ static_cast<u_int>( eSpeaker ) ];
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetSpeakerPosition()
//
// Get the position of a speaker as a vector from the listener
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Audio_System::IsSpeakerActive ( Speakers eSpeaker )
{
	Asura_Assert( eSpeaker != ASURA_NUM_SPEAKERS, "Invalid speaker" );

	return ( ( s_uActiveSpeakers & ( 1 << static_cast<u_int>( eSpeaker ) ) ) != 0 );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::AddStreamInfo()
//
// Registers a new streaming sound with the audio system
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_StreamInfo* Asura_Audio_System::AddStreamInfo( const char* const szFilename,
	const char* const szPackage, const u_int uLocation, const bool bLooping )
{
	Asura_Audio_StreamInfo* const pxStreamInfo =
		s_xStreamInfoTree.AddNode( Asura_GetHashID( szFilename ) );

	Asura_Assertf( pxStreamInfo, ("Failed to allocate stream info node for '%s'.", szFilename) );
	if( !pxStreamInfo ) return NULL;

	#if defined ASURA_AUDIO_STORE_FILENAMES
		pxStreamInfo->m_szFilename = Asura_StringUtil::Duplicate( szFilename );
	#endif

	pxStreamInfo->m_szPackage = Asura_StringUtil::Duplicate( szPackage );
	pxStreamInfo->m_uLocation = uLocation;
	pxStreamInfo->m_bLooping = bLooping;

	return pxStreamInfo;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::DeleteStreamInfo()
//
// Unregisters a streaming sound from the audio system
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::DeleteStreamInfo( const Asura_Hash_ID uHash )
{
	s_xStreamInfoTree.DeleteNodeByHash( uHash );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetStreamInfoTree()
//
// Gets access to the streaming sound info tree, for iteration purposes
//
////////////////////////////////////////////////////////////////////////////////

Asura_TemplateHashTree< Asura_Audio_StreamInfo >& Asura_Audio_System::GetStreamInfoTree()
{
	return s_xStreamInfoTree;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetStreamInfo()
//
// Gets info about a streaming sound. Allows modification since needed by Tools
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_StreamInfo* Asura_Audio_System::GetStreamInfo( const Asura_Hash_ID uHash )
{
	Asura_Audio_StreamInfo* const pxStreamInfo = s_xStreamInfoTree.GetDataByHash( uHash );
	Asura_Assertf( pxStreamInfo, ("Failed to get stream info for hash 0x%X.", uHash) );

	return pxStreamInfo;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetDistanceFilterMinRange()
//
////////////////////////////////////////////////////////////////////////////////

float Asura_Audio_System::GetDistanceFilterMinRange()
{
	return g_fDistanceFilterRangeMin;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetDistanceFilterMaxRange()
//
////////////////////////////////////////////////////////////////////////////////

float Asura_Audio_System::GetDistanceFilterMaxRange()
{
	return g_fDistanceFilterRangeMax;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System Occlusion related functions
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::SetGUIDToIgnoreForOcclusion ( const Asura_Guid uIgnore, const u_int uListener )
{
	if ( uListener < s_uNumListeners )
	{
		s_auOcclusionGUIDsToIgnore[ uListener ] = uIgnore;
	}
}


Asura_Guid Asura_Audio_System::GetGUIDToIgnoreForOcclusion ( const u_int uListener )
{
	if ( uListener < s_uNumListeners )
	{
		return s_auOcclusionGUIDsToIgnore[ uListener ];
	}
	return ASURA_GUID_SERVER;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::AddReverb()
//
// Creates & registers a new reverb with the given name
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_Reverb* Asura_Audio_System::AddReverb( const char* szName, bool& bAlreadyExists )
{
	Asura_Hash_ID uNameHash = Asura_GetHashID(szName);

	//see if this reverb already exists
	Asura_Audio_Reverb* pxReverb = s_xReverbs.GetResource(uNameHash);
	if(pxReverb)
	{
		bAlreadyExists = true;
		//add a reference for the current resource bank
		s_xReverbs.AddResource(uNameHash, *pxReverb, false);
	}
	else
	{
		bAlreadyExists = false;

		//create a new reverb and add it
		pxReverb = new Asura_Audio_Reverb;

		Asura_Assertf( pxReverb, ("Failed to allocate memory for reverb '%s'.", szName) );
		if( !pxReverb ) return NULL;

		#if defined ASURA_AUDIO_STORE_FILENAMES
			pxReverb->m_xName = szName;
		#endif

		s_xReverbs.AddResource(uNameHash, *pxReverb, true );

	}

	return pxReverb;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetReverb()
//
// Retrieves the given reverb, with the option to silently ignore failures
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_Reverb* Asura_Audio_System::GetReverb( const Asura_Hash_ID uHash,
	const bool bAssertOnFail /*= true*/ )
{
	Asura_Audio_Reverb* const pxReverb = s_xReverbs.GetResource( uHash );
	Asura_Assertf( pxReverb || !bAssertOnFail, ("Invalid reverb hash 0x%X.", uHash) );

	return pxReverb;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::SetCurrentReverb()
//
// Sets the current reverb to be applied
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::SetCurrentReverb( const Asura_Hash_ID uHash, const float fAmount )
{
	float fDesiredReverbAmount = 0.0f;
	Asura_Audio_Reverb xDesiredReverb;

	if( uHash != ASURA_HASH_ID_UNSET )
	{
		const Asura_Audio_Reverb* const pxReverb = GetReverb( uHash );
		if( pxReverb )
		{
			xDesiredReverb = *pxReverb;
			fDesiredReverbAmount = fAmount;
		}
	}

	if(		( xDesiredReverb		== s_xReverbBuffers			[ s_iCurrentReverbBuffer ] ) 
		&&	( fDesiredReverbAmount	== s_fReverbAmountBuffers	[ s_iCurrentReverbBuffer ] ) )
	{
		// The desired reverb is the current one, done
		return;
	}

	if( s_fReverbTransition != fASURA_AUDIO_REVERB_TRANSITION_UNSET )
	{
		// We are currently transitioning between reverbs...
		
		if(		( xDesiredReverb		== s_xReverbBuffers			[ !s_iCurrentReverbBuffer ] ) 
			&&	( fDesiredReverbAmount	== s_fReverbAmountBuffers	[ !s_iCurrentReverbBuffer ] ) )
		{
			// The reverb we are going to is the one we just came from, this should be the usual reason
			// for getting here, and we can do a little trickery to just transition back
			s_iCurrentReverbBuffer = !s_iCurrentReverbBuffer;
			s_fReverbTransition = 1.0f - s_fReverbTransition;
		}
		else
		{
			// we are transitioning to yet another reverb, transition from whatever the current reverb
			// is to the new one, not a perfect solution, but should sound okay
			s_xReverbBuffers		[ s_iCurrentReverbBuffer ] = xDesiredReverb;
			s_fReverbAmountBuffers	[ s_iCurrentReverbBuffer ] = fDesiredReverbAmount;

			s_xReverbBuffers		[ !s_iCurrentReverbBuffer ] = s_xCurrReverb;
			s_fReverbAmountBuffers	[ !s_iCurrentReverbBuffer ] = s_fCurrReverbAmount;

			s_fReverbTransition = 0.0f;
		}
	}
	else
	{
		// Set our active reverb and kick off the transition
		s_iCurrentReverbBuffer = !s_iCurrentReverbBuffer;
		
		s_xReverbBuffers		[ s_iCurrentReverbBuffer ] = xDesiredReverb;
		s_fReverbAmountBuffers	[ s_iCurrentReverbBuffer ] = fDesiredReverbAmount;
		
		s_fReverbTransition = 0.0f;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetCurrentReverb()
//
// Called by the platform to query the current reverb values, including any
// interpolation that's happening as presets are faded in & out
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::GetCurrentReverb( Asura_Audio_Reverb& xReverb, float& fAmount )
{
	xReverb = s_xCurrReverb;
	fAmount = s_fCurrReverbAmount;
}

#if defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::DeleteReverb()
//
// Removes the given reverb. Only available in Tools builds. Retail builds
// should clear out their resource banks when appropriate
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::DeleteReverb( const Asura_Hash_ID uHash )
{
	#if defined ASURA_ASSERT
		const bool bSuccess =
	#endif
		s_xReverbs.DeleteResource( uHash );

	Asura_Assertf( bSuccess, ("Failed to delete reverb with hash 0x%X.", uHash) );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetNumReverbs()
//
// Queries the number of reverbs registered. Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

u_int Asura_Audio_System::GetNumReverbs()
{
	return s_xReverbs.GetNumResources();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetReverbByIndex()
//
// Retrieves a reverb by index. Only available in Tools builds
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_Reverb* Asura_Audio_System::GetReverbByIndex( const u_int uIndex )
{
	return s_xReverbs.GetResourceByIndex( uIndex );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_System::GetReverbNames()
//
// Produces a name list of registered reverbs, sorted alphabetically. The caller
// must ensure the destination array is large enough. Only available in Tools
// builds
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_System::GetReverbNames( const char** const pszNames )
{
	u_int u = 0;

	for( Asura_ResourceSet_ConstIterator< Asura_Audio_Reverb > xIter( s_xReverbs );
		!xIter.Done(); xIter.Next() )
	{
		const Asura_Audio_Reverb* const pxReverb = xIter.GetCurrent();
		pszNames[u++] = pxReverb->m_xName.GetString();
	}

	Asura_SortAndSearch::Quicksort( pszNames, GetNumReverbs(),
		Asura_SortAndSearch::SortStringsCaseInsensitive );
}

#endif // defined ASURA_TOOLS

void Asura_Audio_System::SetNumberOfListeners( const u_int uNum )
{
	if ( uNum > ASURA_MAX_NO_OF_LISTENERS )
	{
		Asura_Assertf( false, ( "Invalid number of listeners set (%d), maximum is %d", uNum, ASURA_MAX_NO_OF_LISTENERS ) );
		return;
	}

	s_uNumListeners = uNum;
}

void Asura_Audio_System::SetListenerActive( const bool bActive, const u_int uListenerIdx /*= 0*/ )
{
	if ( uListenerIdx >= s_uNumListeners )
	{
		Asura_Assertf( false, ( "Listener index %d out of range", uListenerIdx ) );
		return;
	}

	s_abListenerActive[ uListenerIdx ] = bActive;
}

void Asura_Audio_System::SetListenerUsesCurrentCamera( const u_int uListenerIdx /*= 0 */ )
{
	if ( uListenerIdx >= s_uNumListeners )
	{
		Asura_Assertf( false, ( "Listener index %d out of range", uListenerIdx ) );
		return;
	}

	s_abListenerUsesCurrentCamera[ uListenerIdx ] = true;
}

void Asura_Audio_System::SetListenerUsesCamera( const int iCamera, const u_int uListenerIdx /*= 0 */ )
{
	if ( uListenerIdx >= s_uNumListeners )
	{
		Asura_Assertf( false, ( "Listener index %d out of range", uListenerIdx ) );
		return;
	}

	s_abListenerUsesCurrentCamera[ uListenerIdx ] = false;
	s_aiListenerCameraIndex[ uListenerIdx ] = iCamera;
}

void Asura_Audio_System::SetListenerPosition( const Asura_Vector_3 xPos, const Asura_Matrix_3x3 xOrient, const u_int uListenerIdx /*= 0 */ )
{
	if ( uListenerIdx >= s_uNumListeners )
	{
		Asura_Assertf( false, ( "Listener index %d out of range", uListenerIdx ) );
		return;
	}

	s_abListenerActive[ uListenerIdx ] = true;
	s_abListenerUsesCurrentCamera[ uListenerIdx ] = false;
	s_aiListenerCameraIndex[ uListenerIdx ] = -1;
	s_axListenerPosition[ uListenerIdx ] = xPos;
	s_axListenerOrientation[ uListenerIdx ] = xOrient;
}

bool Asura_Audio_System::IsListenerActive( const u_int uListenerIdx /*= 0*/ )
{
	if ( uListenerIdx >= s_uNumListeners )
	{
		Asura_Assertf( false, ( "Listener index %d out of range", uListenerIdx ) );
		return false;
	}

	if ( s_abListenerUsesCurrentCamera[ uListenerIdx ] )
	{
		s_aiListenerCameraIndex[ uListenerIdx ] = Asura_Camera::GetCameraIndex();
	}
	if ( s_aiListenerCameraIndex[ uListenerIdx ] >= 0 && s_aiListenerCameraIndex[ uListenerIdx ] < Asura_Camera::GetNumberOfCameras() )
	{
		s_abListenerActive[ uListenerIdx ] = Asura_Camera::IsCameraActive( s_aiListenerCameraIndex[ uListenerIdx ] );
	}
	return s_abListenerActive[ uListenerIdx ];
}

void Asura_Audio_System::GetListenerPosition( Asura_Vector_3* pxPos, Asura_Matrix_3x3* pxOrient, const u_int uListenerIdx /*= 0*/ )
{
	if ( uListenerIdx >= s_uNumListeners )
	{
		Asura_Assertf( false, ( "Listener index %d out of range", uListenerIdx ) );
		return;
	}

	if ( s_abListenerUsesCurrentCamera[ uListenerIdx ] )
	{
		s_aiListenerCameraIndex[ uListenerIdx ] = Asura_Camera::GetCameraIndex();
	}
	if ( s_aiListenerCameraIndex[ uListenerIdx ] >= 0 && s_aiListenerCameraIndex[ uListenerIdx ] < Asura_Camera::GetNumberOfCameras() )
	{
		s_axListenerPosition[ uListenerIdx ] = Asura_Camera::GetCameraData( s_aiListenerCameraIndex[ uListenerIdx ] ).m_xPosition;
		s_axListenerOrientation[ uListenerIdx ] = Asura_Camera::GetCameraData( s_aiListenerCameraIndex[ uListenerIdx ] ).m_xOrientation;
	}

	if ( pxPos )
	{
		*pxPos = s_axListenerPosition[ uListenerIdx ];
	}
	if ( pxOrient )
	{
		*pxOrient = s_axListenerOrientation[ uListenerIdx ];
	}
}


/******************************************************************************/
/* Asura_Audio_StreamInfo Class Functions                                     */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_StreamInfo::Asura_Audio_StreamInfo()
//
// Constructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_StreamInfo::Asura_Audio_StreamInfo()
	: m_szPackage( NULL )
	, m_uLocation(0)
	, m_bLooping( false )
{
	#if defined ASURA_AUDIO_STORE_FILENAMES
		m_szFilename = NULL;
	#endif

	#if defined ASURA_TOOLS
		m_bDialogue = false;
	#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_StreamInfo::~Asura_Audio_StreamInfo()
//
// Destructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_StreamInfo::~Asura_Audio_StreamInfo()
{
	#if defined ASURA_AUDIO_STORE_FILENAMES
		delete[] m_szFilename;
	#endif

	delete[] m_szPackage;
}

/******************************************************************************/
/* Asura_Audio_Reverb Class Functions                                         */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_Reverb::Asura_Audio_Reverb()
//
// Constructor
//
////////////////////////////////////////////////////////////////////////////////

Asura_Audio_Reverb::Asura_Audio_Reverb()
	// Defaults taken from the 'default' I3DL2 preset, see XAUDIO2FX_I3DL2_PRESET_DEFAULT
	: m_iRoom( -10000 )
	, m_iRoomHF( 0 )
	, m_fRoomRolloffFactor( 0.0f )
	, m_fDecayTime( 1.0f )
	, m_fDecayHFRatio( 0.5f )
	, m_iReflections( -10000 )
	, m_fReflectionsDelay( 0.02f )
	, m_iReverb( -10000 )
	, m_fReverbDelay( 0.04f )
	, m_fDiffusion( 100.0f )
	, m_fDensity( 100.0f )
	, m_fHFReference( 5000.0f )
{
}

static int InterpolateLog( const int iStart, const int iEnd, const float fFactor )
{
	if( iStart == iEnd )
	{
		return iStart;
	}

	// Interpolate along a Log10 curve

	float fStart		= 0.0f;
	float fEnd			= 0.0f;
	float fActualFactor	= fFactor;
	bool bFlipSign = false;
	
	if( iEnd < iStart )
	{
		// For simplicity always make start less than end

		fStart			= static_cast<float>( iEnd );
		fEnd			= static_cast<float>( iStart );
		fActualFactor	= 1.0f - fFactor;
	}
	else
	{
		fStart	= static_cast<float>( iStart );
		fEnd	= static_cast<float>( iEnd );
	}

	if( ( fStart < 0.0f ) && ( fEnd > 0.0f ) )
	{
		// This case is a pain, because we actually just want to interpolate on a positive log scale (actually from 1)
		// and retain the sign of the result. So what we actually do is determine if we are in the positive or negative
		// bit and interpolate within that
		
		// Where is the zero line
		const float fRelativeZero = -fStart / ( fEnd - fStart );

		if( fActualFactor > fRelativeZero )
		{
			// We are in the positive part, change our start point to be zero, and modify the factor appropriately
						
			// Is there a quicker way to adjust the factor? this seems clumsy
			const float fPrevLinearPoint = fStart + (fEnd - fStart) * fActualFactor;
			fActualFactor = fPrevLinearPoint / fEnd;

			fStart = 0.0f;
		}
		else
		{
			// We are in the negative part
			const float fPrevLinearPoint = fStart + (fEnd - fStart) * fActualFactor;
			fActualFactor = 1.0f - fPrevLinearPoint / fStart;

			fEnd = 0.0f;
		}
	}

	if( ( fStart <= 0.0f ) && ( fEnd <= 0.0f ) )
	{
		// Both numbers are negative, interpolate as if they were positive but switch signs later
		const float fTemp = fStart;
		fStart = -fEnd;
		fEnd = -fTemp;
		fActualFactor = 1.0f - fActualFactor;
		bFlipSign = true;
	}
	// Increment so that our minimum value is 1 ( log10(1) = 0 )
	++ fStart;
	++ fEnd;

	const float fLogStart	= log10f( static_cast<float>( fStart ));
	const float fLogEnd		= log10f( static_cast<float>( fEnd ));

	// linearly interpolate in the log10 domain
	const float fLerped = Asura_Maths::Lerp( fLogStart, fLogEnd, fActualFactor );

	// Convert back
	int iResult = static_cast<int>( Asura_Maths::Power( 10.0f, fLerped ) );

	// Take the 1 back off
	-- iResult;
	
	if( bFlipSign )
	{
		// We had both negative start and end points, flip our result
		iResult = -iResult;
	}

	return iResult;
}

Asura_Audio_Reverb::Asura_Audio_Reverb( const Asura_Audio_Reverb* pxReverb1, const Asura_Audio_Reverb* pxReverb2, const float fLerpFactor )
{
	// Fill our properties with the lerped values from the provided reverbs
	if( !pxReverb1 || !pxReverb2 )
	{
		Asura_Assert( false, "Did not provide required values to reverb constructor" );
		return;
	}
	
	m_iRoom					= InterpolateLog	( pxReverb1->m_iRoom				, pxReverb2->m_iRoom				, fLerpFactor );
	m_iRoomHF				= InterpolateLog	( pxReverb1->m_iRoomHF				, pxReverb2->m_iRoomHF				, fLerpFactor );
	m_fRoomRolloffFactor	= Asura_Maths::Lerp	( pxReverb1->m_fRoomRolloffFactor	, pxReverb2->m_fRoomRolloffFactor	, fLerpFactor );
	m_fDecayTime			= Asura_Maths::Lerp	( pxReverb1->m_fDecayTime			, pxReverb2->m_fDecayTime			, fLerpFactor );
	m_fDecayHFRatio			= Asura_Maths::Lerp	( pxReverb1->m_fDecayHFRatio		, pxReverb2->m_fDecayHFRatio		, fLerpFactor );
	m_iReflections			= InterpolateLog	( pxReverb1->m_iReflections			, pxReverb2->m_iReflections			, fLerpFactor );
	m_fReflectionsDelay		= Asura_Maths::Lerp	( pxReverb1->m_fReflectionsDelay	, pxReverb2->m_fReflectionsDelay	, fLerpFactor );
	m_iReverb				= Asura_Maths::Lerp	( pxReverb1->m_iReverb				, pxReverb2->m_iReverb				, fLerpFactor );	// Sounds better lerped
	m_fReverbDelay			= Asura_Maths::Lerp	( pxReverb1->m_fReverbDelay			, pxReverb2->m_fReverbDelay			, fLerpFactor );
	m_fDiffusion			= Asura_Maths::Lerp	( pxReverb1->m_fDiffusion			, pxReverb2->m_fDiffusion			, fLerpFactor );
    m_fDensity				= Asura_Maths::Lerp	( pxReverb1->m_fDensity				, pxReverb2->m_fDensity				, fLerpFactor );
    m_fHFReference			= Asura_Maths::Lerp	( pxReverb1->m_fHFReference			, pxReverb2->m_fHFReference			, fLerpFactor );
}

bool Asura_Audio_Reverb::operator==(const Asura_Audio_Reverb& xReverb)
{
	if( this->m_iRoom				!= xReverb.m_iRoom				)	return false;
	if( this->m_iRoomHF				!= xReverb.m_iRoomHF			)	return false;
	if( this->m_fRoomRolloffFactor	!= xReverb.m_fRoomRolloffFactor	)	return false;
	if( this->m_fDecayTime			!= xReverb.m_fDecayTime			)	return false;
	if( this->m_fDecayHFRatio		!= xReverb.m_fDecayHFRatio		)	return false;
	if( this->m_iReflections		!= xReverb.m_iReflections		)	return false;
	if( this->m_fReflectionsDelay	!= xReverb.m_fReflectionsDelay	)	return false;
	if( this->m_iReverb				!= xReverb.m_iReverb			)	return false;
	if( this->m_fReverbDelay		!= xReverb.m_fReverbDelay		)	return false;
	if( this->m_fDiffusion			!= xReverb.m_fDiffusion			)	return false;
	if( this->m_fDensity			!= xReverb.m_fDensity			)	return false;
	if( this->m_fHFReference		!= xReverb.m_fHFReference		)	return false;

	return true;
}

#ifdef ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Audio_Reverb::WriteToChunkStream()
//
// Writes the properties of a reverb preset to a chunk stream
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Audio_Reverb::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{

	xStream << m_xName;

	xStream << m_iRoom;
	xStream << m_iRoomHF;
	xStream << m_fRoomRolloffFactor;
	xStream << m_fDecayTime;
	xStream << m_fDecayHFRatio;
	xStream << m_iReflections;
	xStream << m_fReflectionsDelay;
	xStream << m_iReverb;
	xStream << m_fDiffusion;
	xStream << m_fDensity;
	xStream << m_fHFReference;
}

#endif	// ASURA_TOOLS

#endif // defined ASURA_USE_AUDIO_SYSTEM

//eof
