//////////////////////////////////////////////////////////////////////////////
//
// Asura_Dialogue.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_DIALOGUE_SYSTEM

// This file's header.
#include "Sound/Asura_Dialogue.h"

// Asura includes
#include "Core/Asura_HashTree.h"
#include "Core/Asura_PrimeSearch.h"
#include "Core/Asura_Stream.h"
#include "Core/Resources/Asura_ResourceSet.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Network/Asura_Network.h"
#include "Sound/Asura_Sound.h"
#include "Sound/Asura_Dialogue_Playback.h"
#include "Sound/Asura_Dialogue_System.h"
#include "Sound/Asura_Sound_Mixer.h"
#include "Sound/Asura_Sound_Event.h"

// Tools includes
#ifdef ASURA_TOOLS
#include "Tools_System.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Sound/Tools_Dialogue_System.h"
#include "Sound/Tools_PickSound.h"
#ifdef ASURA_USE_STREAMING
#include "Streaming/Asura_Streaming_System.h"
#endif //#ifdef ASURA_USE_STREAMING
#endif //#ifdef ASURA_TOOLS

// G L O B A L S /////////////////////////////////////////////////////////////

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

/*------------------------**
** Asura_Dialogue_Storage **
**------------------------*/

Asura_ResourceSet<Asura_Dialogue_Event> Asura_Dialogue_Storage::s_xEventContainer("Dialogue Events");
Asura_ResourceSet<Asura_Dialogue_Line> Asura_Dialogue_Storage::s_xLineContainer("Dialogue Lines");
Asura_ResourceSet<Asura_Dialogue_Line_Template> Asura_Dialogue_Storage::s_xLineTemplateContainer("Dialogue Line Templates");


void Asura_Dialogue_Storage::Shutdown()
{
	s_xEventContainer.Reset();
	s_xLineContainer.Reset();
	s_xLineTemplateContainer.Reset();
}

void Asura_Dialogue_Storage::Reset()
{
	ResetInGameState();
}



Asura_Dialogue_Event* Asura_Dialogue_Storage::CreateEvent(const char* szName, bool& bAlreadyExists)
{
	Asura_Hash_ID uHash = Asura_GetHashID(szName);

	//search for an existing event
	Asura_Dialogue_Event* pxEvent = s_xEventContainer.GetResource(uHash);
	if(pxEvent)
	{
		bAlreadyExists = true;
		//add a reference for the current resource bank
		s_xEventContainer.AddResource(uHash, *pxEvent, false);
	}
	else
	{
		//create a new event and add it
		bAlreadyExists = false;
		pxEvent = new Asura_Dialogue_Event(szName);
		s_xEventContainer.AddResource(uHash, *pxEvent, true);
	}
	return pxEvent;
}

Asura_Dialogue_Line* Asura_Dialogue_Storage::CreateLine(const char* szName, bool& bAlreadyExists)
{
	Asura_Hash_ID uHash = Asura_GetHashID(szName);

	//search for an existing line
	Asura_Dialogue_Line* pxLine = s_xLineContainer.GetResource(uHash);
	if(pxLine)
	{
		bAlreadyExists = true;
		//add a reference for the current resource bank
		s_xLineContainer.AddResource(uHash, *pxLine, false);
	}
	else
	{
		//create a new line and add it
		bAlreadyExists = false;
		pxLine = new Asura_Dialogue_Line(szName);
		s_xLineContainer.AddResource(uHash, *pxLine, true);
	}
	return pxLine;
}

Asura_Dialogue_Line_Template* Asura_Dialogue_Storage::CreateLineTemplate(const char* szName, bool& bAlreadyExists)
{
	Asura_Hash_ID uHash = Asura_GetHashID(szName);

	//search for an existing template
	Asura_Dialogue_Line_Template* pxLineTemplate = s_xLineTemplateContainer.GetResource(uHash);
	if(pxLineTemplate)
	{
		bAlreadyExists = true;
		//add a reference for the current resource bank
		s_xLineTemplateContainer.AddResource(uHash, *pxLineTemplate, false);
	}
	else
	{
		//create a new template and add it
		bAlreadyExists = false;
		pxLineTemplate = new Asura_Dialogue_Line_Template(szName);
		s_xLineTemplateContainer.AddResource(uHash, *pxLineTemplate, true);
	}
	return pxLineTemplate;
}



const Asura_Dialogue_Event* Asura_Dialogue_Storage::FindEvent(Asura_Hash_ID uHashID)
{
	return s_xEventContainer.GetResource(uHashID);
}


const Asura_Dialogue_Line* Asura_Dialogue_Storage::FindLine(Asura_Hash_ID uHashID)
{
	return s_xLineContainer.GetResource(uHashID);
}

const Asura_Dialogue_Line_Template* Asura_Dialogue_Storage::FindLineTemplate(Asura_Hash_ID uHashID)
{
	return s_xLineTemplateContainer.GetResource(uHashID);
}

void Asura_Dialogue_Storage::ResetInGameState()
{
	for(Asura_ResourceSet_Iterator<Asura_Dialogue_Event> xIt(s_xEventContainer); !xIt.Done(); xIt.Next())
	{
		xIt.GetCurrent()->ResetInGameState();
	}
}

void Asura_Dialogue_Storage::ResetInGameStateCallback(Asura_Hash_ID uHash,Asura_Dialogue_Event* pxEvent)
{
	pxEvent->ResetInGameState();
}


#ifdef ASURA_TOOLS
u_int Asura_Dialogue_Storage::GetNumberOfEvents()
{
	return s_xEventContainer.GetNumResources();
}

Asura_Dialogue_Event* Asura_Dialogue_Storage::GetEventByHash(u_int uHashID)
{
	return s_xEventContainer.GetResource(uHashID);
}

Asura_Dialogue_Event* Asura_Dialogue_Storage::GetEventByIndex(u_int uIndex)
{
	return s_xEventContainer.GetResourceByIndex(uIndex);

}

u_int Asura_Dialogue_Storage::GetNumberOfLines()
{
	return s_xLineContainer.GetNumResources();
}

Asura_Dialogue_Line* Asura_Dialogue_Storage::GetLineByHash(u_int uHashID)
{
	return s_xLineContainer.GetResource(uHashID);
}

Asura_Dialogue_Line* Asura_Dialogue_Storage::GetLineByIndex(u_int uIndex)
{
	return s_xLineContainer.GetResourceByIndex(uIndex);
}

u_int Asura_Dialogue_Storage::GetNumberOfLineTemplates()
{
	return s_xLineTemplateContainer.GetNumResources();
}

Asura_Dialogue_Line_Template* Asura_Dialogue_Storage::GetLineTemplateByHash(u_int uHashID)
{
	return s_xLineTemplateContainer.GetResource(uHashID);
}

Asura_Dialogue_Line_Template* Asura_Dialogue_Storage::GetLineTemplateByIndex(u_int uIndex)
{
	return s_xLineTemplateContainer.GetResourceByIndex(uIndex);
}


bool Asura_Dialogue_Storage::RenameEvent(const char* szOldName,const char* szNewName)
{
	Asura_Hash_ID uOldHash = Asura_GetHashID(szOldName);
	Asura_Hash_ID uNewHash = Asura_GetHashID(szNewName);

	//make sure the event exists
	Asura_Dialogue_Event* pxEvent = GetEventByHash(uOldHash);
	if(!pxEvent) return false;

	//see if the hash replacement is okay
	if(uOldHash != uNewHash)
	{
		if(!s_xEventContainer.RelocateResource(uOldHash,uNewHash)) return false;
	}

	pxEvent->m_xName = szNewName;
	pxEvent->m_uHashID = uNewHash;

	return true;
}

bool Asura_Dialogue_Storage::RenameLineTemplate(const char* szOldName,const char* szNewName)
{
	Asura_Hash_ID uOldHash = Asura_GetHashID(szOldName);
	Asura_Hash_ID uNewHash = Asura_GetHashID(szNewName);

	//make sure the template exists
	Asura_Dialogue_Line_Template* pxTemplate = GetLineTemplateByHash(uOldHash);
	if(!pxTemplate) return false;

	//see if the hash replacement is okay
	if(uOldHash != uNewHash)
	{
		if(!s_xLineTemplateContainer.RelocateResource(uOldHash,uNewHash)) return false;
	}

	pxTemplate->m_xName = szNewName;
	pxTemplate->m_uHashID = uNewHash;

	return true;
}


void Asura_Dialogue_Storage::DeleteEvent(Asura_Dialogue_Event* pxEvent)
{
	if(!pxEvent) return;

	//delete the event  from the resource sset 
	#ifdef ASURA_ASSERT
		bool bOK = 
	#endif
	s_xEventContainer.DeleteResource( pxEvent->GetHashID() );
	Asura_Assert(bOK, "failed to delete event from resource set");
}

void Asura_Dialogue_Storage::DeleteLineTemplate(Asura_Dialogue_Line_Template* pxTemplate)
{
	if(!pxTemplate) return;

	//delete the line template  from the resource sset 
	#ifdef ASURA_ASSERT
		bool bOK = 
	#endif
	s_xLineTemplateContainer.DeleteResource( pxTemplate->m_uHashID );
	Asura_Assert(bOK, "failed to delete line template from resource set");
}

void Asura_Dialogue_Storage::ResetLines()
{
	s_xLineContainer.Reset();
}


#endif // ASURA_TOOLS



/*-------------------------**
** Asura_Dialogue_Playable **
**-------------------------*/
Asura_Dialogue_Playable::Asura_Dialogue_Playable(const char* szName)
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
: m_xName(szName)
#endif
{
	m_uHashID = Asura_GetHashID(szName);
}

/*----------------------**
** Asura_Dialogue_Event **
**----------------------*/

Asura_Dialogue_Event::Asura_Dialogue_Event(const char* szName)
: Asura_Dialogue_Playable(szName)
, m_ePlaybackType(ASURA_DPT_CDPLAYERRANDOM)
, m_uInitialMessage(0)
, m_uFinalMessage(0)
, m_fReplayDelayTime(0.0f)
, m_ucPlayProbability(100)
, m_cPriority(100)
, m_bGlobalEvent(false)
, m_fLastPlayTime(0.0f)
, m_uLastPlayedItem((u_int)-1)//initialise to '-1' so that the first track chosen when using 'one by one' is the first in the list. RAR 09/01/2007
, m_pxPrimeSearch(NULL)
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
, m_xFilename("default")
#endif
{
}

Asura_Dialogue_Event::~Asura_Dialogue_Event()
{

	for(u_int i=0;i<m_xMessageStore.Size();i++)
	{
		delete m_xMessageStore[i];
	}
	
	delete m_pxPrimeSearch;
}

// GetPlayableIndex() returns an index of an entry from the list as appropriate and sets m_uLastPlayedItem.
// Returns -1 if none are available.
int Asura_Dialogue_Event::GetPlayableIndex() const
{
	u_int uNumPlayables = m_xPlayableList.Size();

#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	Asura_Assertf(uNumPlayables, ( "No items to play in dialogue event %s", m_xName.GetString() ) );
#else
	Asura_Assertf(uNumPlayables, ( "No items to play in dialogue event 0x%08X", m_uHashID ) );
#endif

	if (!uNumPlayables)
	{
		return -1;
	}

	switch (m_ePlaybackType)
	{
		case ASURA_DPT_ALLSEQUENTIAL:
		{
			// one at a time triggering next one when finished
			m_uLastPlayedItem = 0;
			break;
		}
		case ASURA_DPT_ONEBYONE:
		{
			// current one in list, stopping and advancing in list when finished 
			m_uLastPlayedItem = (m_uLastPlayedItem + 1) % uNumPlayables;
			break;
		}
		case ASURA_DPT_RANDOM:
		{
			// pick at random

			//get the total weight
			u_int uTotalWeight = 0;
			for(u_int i=0;i<uNumPlayables;i++)
			{
				uTotalWeight += GetWeight(i);
			}
			
			if(!uTotalWeight)
			{
				//err , all weighted at zero , so just pick with equal weighting I guess
				m_uLastPlayedItem = Asura_Random::GetIntU(0, uNumPlayables - 1);
			}
			else
			{
				int iRandom = (int)Asura_Random::GetIntU(0,uTotalWeight-1);

				//subtract the weightings until we get a negative value
				for(u_int i=0;i<uNumPlayables;i++)
				{
					iRandom -= GetWeight(i);

					if(iRandom < 0)
					{
						m_uLastPlayedItem = i;
						break;
					}
				}
				Asura_Assert(iRandom<0,"Error picking random playable");
			}

			break;
		}
		case ASURA_DPT_CDPLAYERRANDOM:
		{
			// like a CD random function (plays all in random order)
			Asura_Assert(m_pxPrimeSearch, "no object to choose cdplayer-style random item");
			if(uNumPlayables == 1)
			{
				//If there is only one playable , just go for it.
				//(Prevents crash caused by attempting to find a playable different from the previous one).
				//RAR 16/10/2006
				m_uLastPlayedItem = 0;
			}
			else
			{
				if (m_pxPrimeSearch->Done())
				{
					m_pxPrimeSearch->Regenerate();
				}
				u_int uNextItem = m_pxPrimeSearch->GetNext();
				if (m_uLastPlayedItem == uNextItem)
				{
					// this can happen after regenerating.
					uNextItem = m_pxPrimeSearch->GetNext();
				}
				m_uLastPlayedItem = uNextItem;
			}
			break;
		}
		default:
		{
			Asura_Assert( false, "function not valid for this playback type");
			m_uLastPlayedItem = 0;
			break;
		}
	}

	return m_uLastPlayedItem;
}

// this overloaded version of GetPlayable() is simply an accessor and 
// allows sequential playback.
const Asura_Dialogue_Playable* Asura_Dialogue_Event::GetPlayable(u_int uIndex) const
{
	u_int uNumPlayables = m_xPlayableList.Size();
	Asura_Assert(uIndex < uNumPlayables, "invalid playable sound index");
	if (uIndex >= uNumPlayables)
	{
		return NULL;
	}
	return m_xPlayableList[uIndex];
}

const Asura_Dialogue_Message* Asura_Dialogue_Event::GetMessageFromID(Asura_Hash_ID uHashID) const
{
	if(uHashID)
	{
		for(u_int j=0;j<m_xMessageStore.Size();j++)
		{
			if(uHashID == m_xMessageStore[j]->m_uHashID)
			{
				return m_xMessageStore[j];
			}
		}
	}
	return 0;
}

Asura_Dialogue_Message* Asura_Dialogue_Event::GetMessageFromID(Asura_Hash_ID uHashID)
{
	if(uHashID)
	{
		for(u_int j=0;j<m_xMessageStore.Size();j++)
		{
			if(uHashID == m_xMessageStore[j]->m_uHashID)
			{
				return m_xMessageStore[j];
			}
		}
	}
	return 0;
}


void Asura_Dialogue_Event::TriggerMessages(Asura_Hash_ID uMessageID,Asura_Guid uSpeakerGuid,u_short usMessageBlockFromInstance,Asura_Hash_ID uStaticMessageHandler) const
{
	if(!uMessageID) return;
	//don't want to trigger messages in this way while previewing in tools. RAR 21/11/2006
	#ifdef ASURA_TOOLS
	if(Tools_System::IsActive()) return;
	#endif

	//Dialogue playback takes place on the clients ,but message sending needs to take place on the server.
	//So only trigger if this is taking place on the server machine.
	//(Somewhat dodgy if this is wanted with dedicated servers).
	//RAR 22/11/2006
	if(!Asura_Network::IsServer()) return;

	const Asura_Dialogue_Message* pxMessageBlock = GetMessageFromID(uMessageID);
	if(!pxMessageBlock)
	{
		Asura_Assert(0,"Failed to find message block being triggered by dialogue event.\nPlease use the dialogue system Verify function to check for errors.");
		return;
	}

	//Get the static message handler that should own the message block from the instance
	//The handler hash ID is taken from the dialogue storage , but hopefully it should be correct for this as well
	Asura_StaticMessage_Handler* pxHandler = 0;
#if defined(ASURA_USE_STREAMING)
	pxHandler = Asura_StaticMessage_System::GetHandler(uStaticMessageHandler);
#else
	pxHandler = &Asura_StaticMessage_System::GetCurrentHandler();
#endif	// defined(ASURA_USE_STREAMING)

	pxMessageBlock->m_xMessageBlock.SendMessagesForDialogueSystem(uSpeakerGuid,usMessageBlockFromInstance,pxHandler);
}


const Asura_Dialogue_Message* Asura_Dialogue_Event::GetMessageFromStoreByHash(const Asura_Hash_ID uHash) const
{
	for( u_int uMsg = 0; uMsg < m_xMessageStore.Size(); ++uMsg )
	{
		if( m_xMessageStore[uMsg] && ( m_xMessageStore[uMsg]->m_uHashID == uHash ) )
		{
			return m_xMessageStore[uMsg];
		}
	}
	return 0;
}


#ifdef ASURA_TOOLS
void Asura_Dialogue_Event::SetPlaybackType( ASURA_DIALOGUE_PLAYBACK_TYPE eType )
{
	if (m_ePlaybackType == eType)
	{
		// nothing to do
		return;
	}
	if (m_ePlaybackType == ASURA_DPT_CDPLAYERRANDOM)
	{
		// free the random search object
		delete m_pxPrimeSearch;
		m_pxPrimeSearch = NULL;
	}
	if (eType == ASURA_DPT_CDPLAYERRANDOM)
	{
		// allocate a random search object
		Asura_Assert(!m_pxPrimeSearch, "overwriting random search object");
		if(GetNumPlayables())
		{
			m_pxPrimeSearch = new Asura_PrimeSearch(GetNumPlayables());
		}
	}
	else if (eType == ASURA_DPT_ONEBYONE)
	{
		//Set last played item so we start from the beginning of the list.
		//RAR 09/01/2007
		m_uLastPlayedItem = (u_int)-1;

	}
	m_ePlaybackType = eType;
}

void Asura_Dialogue_Event::SetPlayableList(  const Asura_Dialogue_Playable* const* apxPlayables, u_int uNumPlayables )
{
	// overwrite the existing array with this new one
	u_int uCurrentSize = m_xPlayableList.Size();
	if (uNumPlayables > uCurrentSize)
	{
		m_xPlayableList.Clear();
		uCurrentSize = 0;
	}
	if (!uCurrentSize)
	{
		m_xPlayableList.SetSize(uNumPlayables);
	}
	else
	{
		m_xPlayableList.Shrink(uNumPlayables);
	}

	for (u_int u = 0; u < uNumPlayables; ++u)
	{
		m_xPlayableList[u] = apxPlayables[u];
	}

	// ensure the random playback object has the right number of elements
	if (m_ePlaybackType == ASURA_DPT_CDPLAYERRANDOM)
	{
		// changing playback type will realloc the random thingy
		SetPlaybackType(ASURA_DPT_RANDOM);
		SetPlaybackType(ASURA_DPT_CDPLAYERRANDOM);
	}

	// TODO: fix this
	// set the size of the delays list
	u_int uCurDelaysSize = m_xDelayList.Size();
	if (uNumPlayables > uCurDelaysSize)
	{
		m_xDelayList.Clear();
		uCurDelaysSize = 0;
	}
	if (!uCurDelaysSize)
	{
		m_xDelayList.SetSize(uNumPlayables);
	}
	else
	{
		m_xDelayList.Shrink(uNumPlayables);
	}
	
	// set the size of the weights list
	u_int uCurWeightsSize = m_xWeightsList.Size();
	if (uNumPlayables > uCurWeightsSize)
	{
		m_xWeightsList.Clear();
		uCurWeightsSize = 0;
	}
	if (!uCurWeightsSize)
	{
		m_xWeightsList.SetSize(uNumPlayables);
	}
	else
	{
		m_xWeightsList.Shrink(uNumPlayables);
	}

	// set the size of the message lists
	u_int uCurMessagesSize = m_xStartMessageList.Size();
	Asura_Assert(m_xEndMessageList.Size() == uCurMessagesSize,"");

	if (uNumPlayables > uCurMessagesSize)
	{
		m_xStartMessageList.Clear();
		m_xEndMessageList.Clear();
		uCurMessagesSize = 0;
	}
	if (!uCurMessagesSize)
	{
		m_xStartMessageList.SetSize(uNumPlayables);
		m_xEndMessageList.SetSize(uNumPlayables);
	}
	else
	{
		m_xStartMessageList.Shrink(uNumPlayables);
		m_xEndMessageList.Shrink(uNumPlayables);
	}

}

void Asura_Dialogue_Event::SetDelay( u_int uItem, float fDelay )
{
	Asura_Assert(uItem < m_xDelayList.Size(),"");
	m_xDelayList[uItem] = fDelay;
}

void Asura_Dialogue_Event::SetWeight( u_int uItem, u_char ucWeight )
{
	Asura_Assert(uItem < m_xWeightsList.Size(),"");
	m_xWeightsList[uItem] = ucWeight;
}


const Asura_Dialogue_PlayableIDListItem& Asura_Dialogue_Event::GetPlayableID(u_int uIndex) const
{
	u_int uNumPlayables = m_xPlayableIDList.Size();
	Asura_Assert(uIndex < uNumPlayables, "invalid playable sound index");

	return m_xPlayableIDList[uIndex];
}

Asura_Dialogue_PlayableIDListItem* Asura_Dialogue_Event::GetPlayableIDPointer(u_int uIndex)
{
	u_int uNumPlayables = m_xPlayableIDList.Size();
	Asura_Assert(uIndex < uNumPlayables, "invalid playable sound index");

	return &(m_xPlayableIDList[uIndex]);
}

bool Asura_Dialogue_Event::SetPlayableIDList( const Asura_Dialogue_PlayableIDListItem* pxPlayables, u_int uNumPlayables)
{
	if(uNumPlayables == m_xPlayableIDList.Size())
	{
		//If nothing has changed ,then we can return early
		//RAR 05/07/2006
		bool bDiff = false;
		for(u_int i=0;i<uNumPlayables;i++)
		{
			if(pxPlayables[i] != m_xPlayableIDList[i])
			{
				bDiff = true;
				break;
			}
		}
		if(!bDiff) return false;
	}


	// overwrite the existing array with this new one
	u_int uCurrentSize = m_xPlayableIDList.Size();
	if (uNumPlayables > uCurrentSize)
	{
		m_xPlayableIDList.Clear();
		uCurrentSize = 0;
	}
	if (!uCurrentSize)
	{
		m_xPlayableIDList.SetSize(uNumPlayables);
	}
	else
	{
		m_xPlayableIDList.Shrink(uNumPlayables);
	}

	for (u_int u = 0; u < uNumPlayables; ++u)
	{
		m_xPlayableIDList[u] = pxPlayables[u];
	}

	UpdatePlayableListFromIDList();
	return true;
}

bool Asura_Dialogue_Event::RemovePlayableID(const Asura_Dialogue_PlayableID& xID)
{
	u_int uDeleted = 0;
	//Remove any references to the playable ID
	for(u_int i=0;i<m_xPlayableIDList.Size();i++)
	{
		if(m_xPlayableIDList[i].m_xID != xID)
		{
			if(uDeleted)
			{
				m_xPlayableIDList[i-uDeleted] = m_xPlayableIDList[i];
			}
		}
		else
		{
			uDeleted++;
		}
	}
	if(!uDeleted) return false;

	//Some have been deleted , so update the playable pointers as well
	m_xPlayableIDList.Shrink(m_xPlayableIDList.Size()-uDeleted);
	UpdatePlayableListFromIDList();
	return true;
}

bool Asura_Dialogue_Event::ReplacePlayableID(const Asura_Dialogue_PlayableID& xOldID,const Asura_Dialogue_PlayableID& xNewID)
{
	u_int uReplaced = 0;
	//Replace any references to the playable ID
	for(u_int i=0;i<m_xPlayableIDList.Size();i++)
	{
		if(m_xPlayableIDList[i].m_xID == xOldID)
		{
			m_xPlayableIDList[i].m_xID = xNewID;
			uReplaced++;
		}
	}
	if(!uReplaced) return false;

	//Some have been replaced , so update the playable pointers as well
	UpdatePlayableListFromIDList();
	return true;
}

void Asura_Dialogue_Event::UpdatePlayableListFromIDList()
{
	//build array of pointers to the playables for the non-tools data
	const Asura_Dialogue_Playable** ppxPlayables = m_xPlayableIDList.Size() ? new const Asura_Dialogue_Playable*[m_xPlayableIDList.Size()] : 0;
	int* piIndexConv = m_xPlayableIDList.Size() ? new int[m_xPlayableIDList.Size()] : 0;
	u_int uNumFound = 0;
	for(u_int i=0;i<m_xPlayableIDList.Size();i++)
	{
		if(m_xPlayableIDList[i].m_xID.m_eType == ASURA_DIALOGUE_PLAYABLE_EVENT)
		{
			ppxPlayables[uNumFound] = Asura_Dialogue_Storage::GetEventByHash(m_xPlayableIDList[i].m_xID.m_uHashID);
		}
		else
		{
			Asura_Assert(m_xPlayableIDList[i].m_xID.m_eType == ASURA_DIALOGUE_PLAYABLE_LINE,"");
			ppxPlayables[uNumFound] = Asura_Dialogue_Storage::GetLineByHash(m_xPlayableIDList[i].m_xID.m_uHashID);
		}
		if(ppxPlayables[uNumFound])
		{
			//update the stored name in the ID list. RAR 27/11/2006
			m_xPlayableIDList[i].m_xName = ppxPlayables[uNumFound]->GetName();

			piIndexConv[i] = uNumFound;		
			uNumFound++;
		}
		else
		{
			piIndexConv[i] = -1;
		}
	}

	SetPlayableList(ppxPlayables,uNumFound);

	//set the weights and delays
	for(u_int i=0;i<m_xPlayableIDList.Size();i++)
	{
		int iIndex = piIndexConv[i];
		if(iIndex==-1) continue;
        		
		SetWeight(iIndex,m_xPlayableIDList[i].m_ucWeight);
		SetDelay(iIndex,m_xPlayableIDList[i].m_fDelay);
	}

	//set the messages
	for(u_int i=0;i<m_xPlayableIDList.Size();i++)
	{
		int iIndex = piIndexConv[i];
		if(iIndex==-1) continue;

		m_xStartMessageList[iIndex] = m_xPlayableIDList[i].m_uStartMessageID;
		m_xEndMessageList[iIndex] = m_xPlayableIDList[i].m_uEndMessageID;

	}
    	
	delete [] piIndexConv;
	delete [] ppxPlayables;
}

void Asura_Dialogue_Event::SetMessageStore(const Asura_Dialogue_Message* const* ppxMessages,u_int uNumMessages)
{

	//delete the existing messages
	for(u_int i=0;i<m_xMessageStore.Size();i++)
	{
		delete m_xMessageStore[i];
	}

	//Resize the list
	u_int uCurrentSize = m_xMessageStore.Size();
	if (uNumMessages > uCurrentSize)
	{
		m_xMessageStore.Clear();
		uCurrentSize = 0;
	}
	if (!uCurrentSize)
	{
		m_xMessageStore.SetSize(uNumMessages);
	}
	else
	{
		m_xMessageStore.Shrink(uNumMessages);
	}

	//allocate copies of the messages
	for(u_int i=0;i<uNumMessages;i++)
	{
		m_xMessageStore[i] = new Asura_Dialogue_Message;
		*m_xMessageStore[i] = *ppxMessages[i];
	}
}

void Asura_Dialogue_Event::RemoveUnusedMessageBlocks ()
{
	// This function doesnt handle events in events but the places which call it dont care
	for(u_int uMsg = m_xMessageStore.Size() - 1; uMsg != u_int(-1) ; uMsg--)
	{
		if( m_uInitialMessage == m_xMessageStore[uMsg]->m_uHashID )
		{
			continue;
		}

		if( m_uFinalMessage == m_xMessageStore[uMsg]->m_uHashID )
		{
			continue;
		}

		// Check if any playables use this message block
		bool bInUse = false;
		
		for( u_int uPlayable = 0; uPlayable < m_xPlayableList.Size(); ++uPlayable )
		{
			if(		( m_xPlayableIDList[uPlayable].m_uStartMessageID == m_xMessageStore[uMsg]->m_uHashID )
				||	( m_xPlayableIDList[uPlayable].m_uEndMessageID == m_xMessageStore[uMsg]->m_uHashID ) )
			{
				bInUse = true;
				break;
			}
		}

		if( !bInUse )
		{
			delete m_xMessageStore[uMsg];
			m_xMessageStore.RemoveElement( uMsg );
		}
	}
	UpdatePlayableListFromIDList();
}

#endif // #ifdef ASURA_TOOLS

Asura_Dialogue_Line::Asura_Dialogue_Line(const char* szName)
: Asura_Dialogue_Playable(szName)
, m_xCharacterID(ASURA_HASH_ID_UNSET)
, m_wszText(NULL)
, m_pxTemplate(NULL)
, m_uSoundHash(ASURA_HASH_ID_UNSET)
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
, m_uFaceAnimationHash(ASURA_HASH_ID_UNSET)
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
, m_uAnimation(ASURA_HASH_ID_UNSET)
, m_fSoundLength(0.0f)
#ifdef ASURA_USE_ARTICULATION_SYSTEM
, m_pxArticulation(NULL)
#endif
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
, m_xCharacterName("none")
, m_xTemplateName("default")
#endif
#ifdef ASURA_TOOLS
, m_bNeedToGenerateSoundID(false)
, m_xSourceDialogueFilename( "" )
#endif
{
	{
		Asura_String xTemp("sounds/dialogue/english/");
		xTemp += szName;
		xTemp += ".wav";
		m_uSoundHash = Asura_GetHashID(xTemp.GetString());
	}
#ifdef ASURA_USE_ARTICULATION_SYSTEM
	m_pxArticulation = const_cast<Asura_LipSync_Articulation*>( Asura_LipSync_Manager::GetArticulationByHash( GetHashID() ) );
#endif
	
};

Asura_Dialogue_Line::~Asura_Dialogue_Line()
{
	// MGH 12/2/2007: Fix crash caused by dangling pointer. Typically happens when a zone unloads,
	// whilst the dialogue is still playing.
	Asura_Dialogue_System::RemoveCurrentDialogueLine(this);

	#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
	if(m_xLocalisedText.Size())
	{
		for(u_int i=0;i<m_xLocalisedText.Size();i++)
		{
			delete [] m_xLocalisedText[i].m_wszText;
		}
	}
	else
	#endif
	{
		delete [] m_wszText;
	}
}

#ifndef ASURA_USE_AUDIO_SYSTEM

int Asura_Dialogue_Line::GetSoundID() const
{
	#ifdef ASURA_TOOLS
	if(m_bNeedToGenerateSoundID)
	{
		if(!m_xSoundFilename.IsBlank())
		{
			//Generate a sound ID for the filename , if the file exists
			if(Tools_FileSystem::FileExists_CheckNetwork(m_xSoundFilename.GetString()))
			{
				Tools_PickSound::GetSoundIDForSoundEventFilename(m_xSoundFilename.GetString());
			}
		}
		m_bNeedToGenerateSoundID = false;
	}
	#endif //ASURA_TOOLS

	return Asura_Sound_System::GetSoundIDFromHash(m_uSoundHash);
}

#endif // ndef ASURA_USE_AUDIO_SYSTEM

#ifdef ASURA_TOOLS
void Asura_Dialogue_Line::SetCharacter(const char* szCharacter)
{
	m_xCharacterName = szCharacter;
	m_xCharacterID = Asura_GetHashID(szCharacter);
}

void Asura_Dialogue_Line::SetTemplate(const char* szTemplateName)
{
	m_xTemplateName = szTemplateName;
	FindTemplate();
}

void Asura_Dialogue_Line::FindTemplate()
{
	Asura_Hash_ID uHash = m_xTemplateName.IsBlank() ? 0 : Asura_GetHashID(m_xTemplateName.GetString());
	m_pxTemplate = Asura_Dialogue_Storage::GetLineTemplateByHash(uHash);
}

void Asura_Dialogue_Line::SetText(const Asura_Dialogue_Text* pxText,u_int uNumLanguages)
{
	//clear out any old strings
	for(u_int i=0;i<m_xLocalisedText.Size();i++)
	{
		delete [] m_xLocalisedText[i].m_wszText;
	}

	//resize localised text array
	u_int uCurrentSize = m_xLocalisedText.Size();
	if (uNumLanguages > uCurrentSize)
	{
		m_xLocalisedText.Clear();
		uCurrentSize = 0;
	}
	if (!uCurrentSize)
	{
		m_xLocalisedText.SetSize(uNumLanguages);
	}
	else
	{
		m_xLocalisedText.Shrink(uNumLanguages);
	}

	//and fill it in
	for(u_int i=0;i<uNumLanguages;i++)
	{
		m_xLocalisedText[i].m_iLanguage = pxText[i].m_iLanguage;
		if(pxText[i].m_wszText)
		{
			Asura_UnicodeChar* wszText = new Asura_UnicodeChar[wcslen(pxText[i].m_wszText)+1];
			wcscpy(wszText,pxText[i].m_wszText);
			m_xLocalisedText[i].m_wszText = wszText;

		}
		else
		{
			m_xLocalisedText[i].m_wszText = 0;
		}
	}

	//put the first string (english) as the current language text
	m_wszText = m_xLocalisedText[0].m_wszText;

}

const Asura_UnicodeChar* Asura_Dialogue_Line::GetTextForLanguage(int iLanguage) const
{
	u_int uNumLanguages = m_xLocalisedText.Size();

	for(u_int i=0;i<uNumLanguages;i++)
	{
		if(m_xLocalisedText[i].m_iLanguage == iLanguage)
		{					
			return m_xLocalisedText[i].m_wszText;
		}
	}
	return 0;
}

void Asura_Dialogue_Line::SetSoundFilename(const char* szFilename)
{
	m_xSoundFilename = szFilename;
	m_uSoundHash = Asura_GetHashID(szFilename);
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	m_uFaceAnimationHash	= ASURA_HASH_ID_UNSET;
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	m_bNeedToGenerateSoundID = true;
}

void Asura_Dialogue_Line::SetSourceDialogueFilename( const char* szFilename )
{
	m_xSourceDialogueFilename = szFilename;
}

#endif //ASURA_TOOLS

Asura_Dialogue_Line_Template::Asura_Dialogue_Line_Template(const char* szName)
: m_fGain(1.0f)
, m_fInnerRange(5.0f)
, m_fOuterRange(30.0f)
, m_fMinGain(0.0f)
, m_bVoiceover(false)
, m_bNoWavForPSP(false)
, m_bNoSubtitle(false)
, m_uMixerPreset(ASURA_HASH_ID_UNSET)
, m_pxMixingChannel(NULL)
, m_uMixerHash(0)
, m_bStreamed(true)
#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
, m_xName(szName)
, m_usSampleRateForPS2(0)
, m_usSampleRateForPSP(0)
, m_xMixingChannelName("Speech")
,m_usCompressionQualityPS3(TSS_QUALITY_MEDIUM)
,m_usCompressionQuality360(TSS_QUALITY_MEDIUM)
#endif
{
	m_uHashID = Asura_GetHashID(szName);

	#ifdef ASURA_TOOLS
	//Record the fact that an audio template has existed , so the file should get saved out if needed later , even
	//if it would be empty. RAR 06/11/2006
	Tools_Dialogue_System::SetEverHadAudioTemplates();
	#endif
};

Asura_Dialogue_Line_Template::~Asura_Dialogue_Line_Template()
{
}


const Asura_Sound_Mixer_Channel* Asura_Dialogue_Line_Template::GetMixerChannel() const
{
	if(m_pxMixingChannel) return m_pxMixingChannel;

	//Search for mixer if we haven't done so yet
	if(m_uMixerHash)
	{
		m_pxMixingChannel = Asura_Sound_Mixer_System::FindChannel(m_uMixerHash);
		if(!m_pxMixingChannel)
		{
			#ifdef ASURA_DIALOGUE_INCLUDE_TOOLS_DATA
			Asura_Assertf(m_pxMixingChannel,("Failed to get mixer %s for dialogue template %s",m_xMixingChannelName.GetString(),m_xName.GetString()));
			#else
			Asura_Assert(m_pxMixingChannel,"Failed to get mixer for dialogue template");
			#endif
			//reset hash so we don't try searching again
			m_uMixerHash = 0;
		}
	}

	return m_pxMixingChannel;
}

#endif // #ifdef ASURA_USE_DIALOGUE_SYSTEM

