// Asura_DynamicMusic_System.cpp
// Thomas Young 12.08.08
// Copyright (C) 2008 Rebellion, All Rights Reserved.

#include "Core/Asura.h"							//Main Asura header

#ifdef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Sound/Asura_Dynamic_Music_System.h"	//My header

#include "Core/Asura_Timers.h"					//Other Asura headers
#include "Core/Asura_Multithread.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_Dynamic_Music_System_Triggers.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Audio_Voice.h"
#include "Sound/Asura_Sound_Mixer.h"
#include "Entity/Triggers/Asura_ServerEntity_DynamicMusicTrigger.h"

#ifdef ASURA_USE_DEBUGOVERLAY
#include "Debugging/Asura_DebugOverlay.h"
#ifdef ASURA_TOOLS
#include "Entity/Tools_Entity_DynamicMusicTrigger_Object.h"
#include "Entity/Tools_Entity_Data.h"
#endif
#endif

// D E F I N E S /////////////////////////////////////////////////////////////

static const float fASURA_DMS_SEEK_PREPARE_TIME = 4.0f;

// C L A S S   D E C L A R A T I O N S //////////////////////////////////////

class Asura_Dynamic_Music_ActiveTrigger
{
	public:
		
		Asura_Hash_ID	m_uTriggerHash;	// Hash of this trigger (to lookup in the full trigger list)
		float			m_fTriggerTime;	// This is the time that has passed for this trigger

		void Clear ()
		{
			m_uTriggerHash	= ASURA_HASH_ID_UNSET;
			m_fTriggerTime	= -1.0f;
		}
};

class Asura_Dynamic_Music_ActiveTrack
{
	public:

		// -------------- Enumerations ------------------

		enum StartOffset
		{
			START_OFFSET_BODY,
			START_OFFSET_OUTRO,
			START_OFFSET_NONE
		};

		// --------------- Functions --------------------

		Asura_Dynamic_Music_ActiveTrack ();
		~Asura_Dynamic_Music_ActiveTrack ();

		void	SetPlaybackStartTime		( const float fStartTime );

		void	SetTrackData				( Asura_Dynamic_Music_TrackData* pxTrackData );
		
		void	BeginPlayback				( );
		void	Update						( );
		void	Clear						( );

		bool	IsPreparing					( );

		void	SetShouldLoop				( const bool bLoop );
		void	SetLayerVolume				( const int iLayer, const float fVolume );
		void	SetAllLayerVolumes			( const float fVolume );
		void	SetAllLayerMaxVolume		( const float fVolume );

		void	SetStartOffset				( const StartOffset eStartOffset );

		void	PrepareToSeekToPosition		( const float fSeekPosition );
		void	ActivateSeek				( );

		bool	IsSeekState_None			( );
		bool	IsSeekState_Preparing		( );
		bool	IsSeekState_Ready			( );
		bool	IsSeekingToLoop				( );
		
		float	GetPlayTime					( ) const;
		float	GetIntroTime				( ) const;
		float	GetOutroTime				( ) const;
		float	GetBodyDuration				( ) const;
		float	GetBarDuration				( ) const;
		float	GetNextBarBoundry			( ) const;
		bool	HitBarBoundary				( ) const	{ return m_bHitBarBoundaryThisFrame; }

		bool	IsPlaying					( ) const	{ return m_bPlaying; }
		bool	ShouldPlayAfterStateLoad	( ) const	{ return m_bPlayingSavedState; }

		void	Pause						( );
		void	Unpause						( );

		Asura_Hash_ID	GetNameHash			( ) const;

		void	WriteStateToChunkStream		( Asura_Chunk_Stream& xStream ) const;
		void	ReadStateFromChunkStream	( Asura_Chunk_Stream& xStream );

		#ifdef ASURA_USE_DEBUGOVERLAY
		void DoActiveTrackDebugOverlay		( ) const;
		#endif
		
	private:
	
		// -------------- Enumerations ------------------

		enum { ACTIVE_TRACK_LOADSAVE_STATE_VERSION = 1 };

		enum SeekState
		{
			SEEK_STATE_NONE,
			SEEK_STATE_PREPARING,
			SEEK_STATE_READY
		};

		// --------------- Functions --------------------

		Asura_Audio_Voice*	GetLayerVoice				( const int iLayer ) const;
		Asura_Audio_Voice*	GetLayerSeekVoice			( const int iLayer ) const;

		// ---------------- Variables -------------------

		Asura_Dynamic_Music_TrackData*		m_pxTrackData;
		Asura_Handle						m_iLayerHandles		[ ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ];
		Asura_Handle						m_iLayerSeekHandles	[ ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ];

		bool								m_bShouldLoop;

		bool								m_bPlaying;
		bool								m_bPlayingSavedState;

		float								m_fLastBarBoundary;
		bool								m_bHitBarBoundaryThisFrame;

		SeekState							m_eSeekState;
		bool								m_bPendingSeekIsLoop;
		float								m_fSeekPos;
		
		float								m_fLayerVolumes [ ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ];

		StartOffset							m_eStartOffset;
};

// Asura_Dynamic_Music_ActiveTrack member functions //////////////////////////////////////

Asura_Dynamic_Music_ActiveTrack::Asura_Dynamic_Music_ActiveTrack ()
{
	m_pxTrackData = 0;
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		m_iLayerHandles		[ iLayer ] = ASURA_INVALID_HANDLE;
		m_iLayerSeekHandles	[ iLayer ] = ASURA_INVALID_HANDLE;
		m_fLayerVolumes		[ iLayer ] = 1.0f;
	}
	m_bShouldLoop					= true;
	m_bPlaying						= false;
	m_eSeekState					= SEEK_STATE_NONE;
	m_fSeekPos						= 0.0f;
	m_bPendingSeekIsLoop			= false;
	m_eStartOffset					= START_OFFSET_NONE;
	m_bHitBarBoundaryThisFrame		= false;
	m_bPlayingSavedState			= false;
	m_fLastBarBoundary				= -1.0f;
}


Asura_Dynamic_Music_ActiveTrack::~Asura_Dynamic_Music_ActiveTrack ()
{
	Clear ();
}


Asura_Audio_Voice* Asura_Dynamic_Music_ActiveTrack::GetLayerVoice ( const int iLayer ) const
{
	if( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE )
	{
		return Asura_Audio_System::GetVoice( m_iLayerHandles[ iLayer ] );
	}
	return 0;
}

Asura_Audio_Voice* Asura_Dynamic_Music_ActiveTrack::GetLayerSeekVoice ( const int iLayer ) const
{
	if( m_iLayerSeekHandles[ iLayer ] != ASURA_INVALID_HANDLE )
	{
		return Asura_Audio_System::GetVoice( m_iLayerSeekHandles[ iLayer ] );
	}
	return 0;
}


void Asura_Dynamic_Music_ActiveTrack::SetTrackData ( Asura_Dynamic_Music_TrackData* pxTrackData )
{
	m_pxTrackData = pxTrackData;

	if( m_pxTrackData )
	{
		if( m_pxTrackData->GetBarDuration() == 0.0f )
		{
			Asura_Assertf( false, ("Track '%s' does not have a valid bar length", pxTrackData->m_xName.GetString() ) );
			m_pxTrackData = 0;
			return;
		}

		//Create streaming sounds 
		for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
		{
			if( !m_pxTrackData->m_axLayerFiles[ iLayer ].IsBlank() )
			{
				float fStartOffset = 0.0f;
				switch( m_eStartOffset )
				{
					case START_OFFSET_NONE:		{	fStartOffset = 0.0f;							break;	}
					case START_OFFSET_BODY:		{	fStartOffset = m_pxTrackData->m_fBodyOffset;	break;	}
					case START_OFFSET_OUTRO:	{	fStartOffset = m_pxTrackData->m_fOutroOffset;	break;	}
				}

				m_iLayerHandles[ iLayer ] = Asura_Audio_System::CreateStreamingVoice( Asura_GetHashID( m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ), false, true, fStartOffset );
				Asura_Assertf( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE, ("Failed to create streaming sound from file %s", m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ) );
			}
		}
	}
}

void	Asura_Dynamic_Music_ActiveTrack::BeginPlayback ( )
{
	if( m_bPlaying )
	{
		return;
	}

	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		Asura_Audio_Voice* pxVoice = GetLayerVoice( iLayer );
		if( pxVoice )
		{
			const float fVolume = m_fLayerVolumes[ iLayer ] * m_pxTrackData->m_fVolume * Asura_Dynamic_Music_System::GetMusicMixerVolume();
			pxVoice->SetVolume( fVolume );
			pxVoice->SetPitchUsesTimescale( false );
			pxVoice->Start ();
		}
	}
	m_fLastBarBoundary = GetPlayTime ();
	//The track has already started (so the time might be a bit off), round this to the nearest bar boundry
	m_fLastBarBoundary = m_pxTrackData->RoundTimeToBars( m_fLastBarBoundary );
	m_bPlaying = true;
}

void	Asura_Dynamic_Music_ActiveTrack::Update ()
{
	if( !m_pxTrackData || IsPreparing() )
	{
		return;
	}

	// Check if our seeks have completed
	if( IsSeekState_Preparing() )
	{
		bool bFinishedSeek = true;
		for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
		{
			if( ( m_iLayerSeekHandles[ iLayer ] != ASURA_INVALID_HANDLE ) )
			{
				Asura_Audio_Voice* pxVoice = GetLayerSeekVoice( iLayer );
				if( pxVoice && pxVoice->IsPreparing() )
				{
					bFinishedSeek = false;
					break;
				}
			}
		}

		if( bFinishedSeek )
		{
			m_eSeekState = SEEK_STATE_READY;
		}
	}

	if( !m_bPlaying )
	{
		//If we are not playing, or we are waiting for state to be restored we are done updating here
		return;
	}

	//Update the volume
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( ( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE ) && m_pxTrackData )
		{
			const float fVolume = m_fLayerVolumes[ iLayer ] * m_pxTrackData->m_fVolume * Asura_Dynamic_Music_System::GetMusicMixerVolume();
			Asura_Audio_Voice* pxVoice = Asura_Audio_System::GetVoice( m_iLayerHandles[ iLayer ] );
			if( pxVoice )
			{
				pxVoice->SetVolume( fVolume );
			}
		}
	}

	//Update the play position
		
	const float fBarDuration	= m_pxTrackData->GetBarDuration();
	const float fPlayTime		= GetPlayTime ();

	m_bHitBarBoundaryThisFrame = false;
	if( fPlayTime > ( m_fLastBarBoundary + fBarDuration ) )
	{
		m_bHitBarBoundaryThisFrame = true;
		m_fLastBarBoundary = m_fLastBarBoundary + fBarDuration;
	}

	// check if we need to loop back to the start of the body section soon
	if( ( m_eSeekState == SEEK_STATE_NONE ) && m_bShouldLoop && ( ( fPlayTime + fASURA_DMS_SEEK_PREPARE_TIME ) >= m_pxTrackData->m_fOutroOffset ) )
	{
		PrepareToSeekToPosition( m_pxTrackData->m_fBodyOffset );
		// remember that this seek is to loop, so that we know at what point to set it off
		m_bPendingSeekIsLoop = true;
	}

	// Check if layers have stopped
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		Asura_Audio_Voice* pxVoice = GetLayerVoice( iLayer );
		if( pxVoice && !pxVoice->IsPlaying() )
		{
			m_bPlaying = false;
		}
	}

	// Check if we are at the point where we should jump back to the start of the body
	if(		( !IsPlaying() && IsSeekingToLoop() ) 
		||	( HitBarBoundary() && IsSeekingToLoop() && IsSeekState_Ready() && ( fPlayTime >= m_pxTrackData->m_fOutroOffset ) ) )
	{
		ActivateSeek();
		m_bPendingSeekIsLoop = false;
		m_bPlaying = true;
	}
}

void	Asura_Dynamic_Music_ActiveTrack::Clear						()
{
	m_pxTrackData = 0;
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_System::DestroyVoice( m_iLayerHandles[ iLayer ] );
			m_iLayerHandles[ iLayer ] = ASURA_INVALID_HANDLE;
		}
		if( m_iLayerSeekHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_System::DestroyVoice( m_iLayerSeekHandles[ iLayer ] );
			m_iLayerSeekHandles[ iLayer ] = ASURA_INVALID_HANDLE;
		}
	}
	m_bShouldLoop				= true;
	m_bPlaying					= false;
	m_eStartOffset				= START_OFFSET_NONE;
	m_bHitBarBoundaryThisFrame	= false;
	m_eSeekState				= SEEK_STATE_NONE;
	m_fSeekPos					= 0.0f;
	m_bPendingSeekIsLoop		= false;
	m_fLastBarBoundary			= -1.0f;
	m_bPlayingSavedState		= false;
}

bool Asura_Dynamic_Music_ActiveTrack::IsPreparing ( )
{
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_Voice* pxVoice = Asura_Audio_System::GetVoice( m_iLayerHandles[ iLayer ] );
			
			if( pxVoice && pxVoice->IsPreparing() )
			{
				return true;
			}
		}
	}
	return false;
}

void Asura_Dynamic_Music_ActiveTrack::SetShouldLoop				( const bool bLoop )
{
	m_bShouldLoop = bLoop;
}

void	Asura_Dynamic_Music_ActiveTrack::SetLayerVolume				( const int iLayer, const float fVolume )
{
	Asura_Assert( ( iLayer >= 0 ) && ( iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS), "Invalid layer" );
	m_fLayerVolumes[ iLayer ] = fVolume;
}

void	Asura_Dynamic_Music_ActiveTrack::SetAllLayerVolumes ( const float fVolume )
{
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		m_fLayerVolumes[ iLayer ] = fVolume;
	}
}

void Asura_Dynamic_Music_ActiveTrack::SetAllLayerMaxVolume ( const float fVolume )
{
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		m_fLayerVolumes[ iLayer ] = Asura_Maths::Min( fVolume, m_fLayerVolumes[ iLayer ] );
	}
}

void Asura_Dynamic_Music_ActiveTrack::SetStartOffset ( const StartOffset eStartOffset )
{
	if( IsPlaying() )
	{
		Asura_Assert( false, "Cannot set start offset, voice is already playing" );
		return;
	}

	m_eStartOffset = eStartOffset;
}


float	Asura_Dynamic_Music_ActiveTrack::GetPlayTime () const
{
	float fCurrentTrackTime = 0.0f;
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		Asura_Audio_Voice* pxVoice = GetLayerVoice( iLayer );
		if( pxVoice )
		{
			fCurrentTrackTime = pxVoice->GetPlayPosition();
			break;
		}
	}
	return fCurrentTrackTime;
}


float Asura_Dynamic_Music_ActiveTrack::GetIntroTime () const
{
	if(m_pxTrackData)
	{
		return m_pxTrackData->m_fBodyOffset;
	}
	return 0.0f;
}


float Asura_Dynamic_Music_ActiveTrack::GetOutroTime	() const
{
	if(m_pxTrackData)
	{
		return m_pxTrackData->m_fOutroOffset;
	}
	return 0.0f;
}


float Asura_Dynamic_Music_ActiveTrack::GetBodyDuration () const
{
	if(m_pxTrackData)
	{
		return (m_pxTrackData->m_fOutroOffset - m_pxTrackData->m_fBodyOffset);
	}
	return 0.0f;
}


float Asura_Dynamic_Music_ActiveTrack::GetBarDuration () const
{
	if( m_pxTrackData ) 
	{
		return m_pxTrackData->GetBarDuration();
	}
	return 0.0f;
}

float Asura_Dynamic_Music_ActiveTrack::GetNextBarBoundry () const
{
	if( !m_bPlaying )
	{
		// Not playing so the next bar boundry is pretty much now
		return 0.0f;
	}
	if( m_pxTrackData ) 
	{
		return m_pxTrackData->RoundTimeUpToBars( GetPlayTime() );
	}
	return 0.0f;
}


Asura_Hash_ID Asura_Dynamic_Music_ActiveTrack::GetNameHash() const
{ 
	if(m_pxTrackData)
	{
		return m_pxTrackData->m_xName.GetHash();
	}
	return ASURA_HASH_ID_UNSET;
}


void Asura_Dynamic_Music_ActiveTrack::PrepareToSeekToPosition ( const float fSeekPosition )
{
	if( m_eSeekState != SEEK_STATE_NONE )
	{
		Asura_Assert( false, "Already seeking dynamic music track, cannot seek again" );
		return;
	}
	
	m_eSeekState = SEEK_STATE_PREPARING;
	m_fSeekPos = fSeekPosition;				// We just need this for state saving

	// create streaming sounds at the appropriate position, ready to play
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( !m_pxTrackData->m_axLayerFiles[ iLayer ].IsBlank() )
		{
			m_iLayerSeekHandles[ iLayer ] = Asura_Audio_System::CreateStreamingVoice( Asura_GetHashID( m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ), false, true, fSeekPosition );
			Asura_Assertf( m_iLayerSeekHandles[ iLayer ] != ASURA_INVALID_HANDLE, ("Failed to create streaming sound from file %s", m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ) );
		}
	}	
}

bool Asura_Dynamic_Music_ActiveTrack::IsSeekState_None ( )
{
	return ( m_eSeekState == SEEK_STATE_NONE );
}

bool Asura_Dynamic_Music_ActiveTrack::IsSeekState_Preparing ( )
{
	return ( m_eSeekState == SEEK_STATE_PREPARING );
}


bool Asura_Dynamic_Music_ActiveTrack::IsSeekState_Ready ( )
{
	return ( m_eSeekState == SEEK_STATE_READY );
}


bool Asura_Dynamic_Music_ActiveTrack::IsSeekingToLoop ( )
{
	return ( ( m_eSeekState != SEEK_STATE_NONE ) && m_bPendingSeekIsLoop );
}


void Asura_Dynamic_Music_ActiveTrack::ActivateSeek ( )
{
	if( m_eSeekState != SEEK_STATE_READY )
	{
		Asura_Assert( false, "Cannot activate dynamic music seek, still pending" );
		return;
	}

	// Start playback of our seek voices and swap the handles
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_Voice* pxVoice = GetLayerVoice( iLayer );
			if( pxVoice )
			{
				Asura_Audio_System::DestroyVoice( m_iLayerHandles[ iLayer ] );
			}
		}

		if( m_iLayerSeekHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_Voice* pxVoice = GetLayerSeekVoice( iLayer );
			if( pxVoice )
			{
				const float fVolume = m_fLayerVolumes[ iLayer ] * m_pxTrackData->m_fVolume * Asura_Dynamic_Music_System::GetMusicMixerVolume();
				pxVoice->SetVolume( fVolume );
				pxVoice->SetPitchUsesTimescale( false );
				pxVoice->Start();
			}
		}

		// Swap handles
		m_iLayerHandles[ iLayer ] = m_iLayerSeekHandles[ iLayer ];
		m_iLayerSeekHandles[ iLayer ] = ASURA_INVALID_HANDLE;
	}
	
	m_eSeekState = SEEK_STATE_NONE;
	m_fSeekPos = 0.0f;

	m_fLastBarBoundary = GetPlayTime ();
	//The track has already started (so the time might be a bit off), round this to the nearest bar boundry
	m_fLastBarBoundary = m_pxTrackData->RoundTimeToBars( m_fLastBarBoundary );

	if( m_fLastBarBoundary >= ( GetOutroTime() - 0.5f ) )
	{
		// If we seeked to the outro or further, presume we no longer want to loop
		SetShouldLoop( false );
	}
}


void Asura_Dynamic_Music_ActiveTrack::WriteStateToChunkStream ( Asura_Chunk_Stream& xStream ) const
{
	unsigned char ucVersion = ACTIVE_TRACK_LOADSAVE_STATE_VERSION;

	xStream << ucVersion;

	float fPlayPos = GetPlayTime();
	xStream << fPlayPos;

	bool bHasTrackData = ( m_pxTrackData != 0 );
	xStream << bHasTrackData;
	
	if( bHasTrackData )
	{
		xStream << m_pxTrackData->m_xName.GetHash();
	}
	
	xStream << m_bShouldLoop;
	xStream << m_bPlaying;

	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		xStream << m_fLayerVolumes[ iLayer ];
	}

	// Are we in the process of seeking
	const bool bSeeking = ( m_eSeekState != SEEK_STATE_NONE );
	xStream << bSeeking;

	if( bSeeking )
	{
		// Save the time we are seeking to
		xStream << m_fSeekPos;
		xStream << m_bPendingSeekIsLoop;
	}
}


void Asura_Dynamic_Music_ActiveTrack::ReadStateFromChunkStream ( Asura_Chunk_Stream& xStream )
{
	Clear();

	unsigned char ucVersion;

	xStream >> ucVersion;

	if( ucVersion != ACTIVE_TRACK_LOADSAVE_STATE_VERSION )
	{
		Asura_Assert( false, "Wrong in-game state version for DM active track, ignoring" );
		return;
	}

	float fPlayPos;
	xStream >> fPlayPos;

	// Read track data
	Asura_Hash_ID xTrackHash = ASURA_HASH_ID_UNSET;
	bool bHasTrackData;
	
	xStream >> bHasTrackData;
	
	if( bHasTrackData )
	{
		xStream >> xTrackHash;
		m_pxTrackData = Asura_Dynamic_Music_System::GetTrackByHash( xTrackHash );
	}
	
	xStream >> m_bShouldLoop;
	xStream >> m_bPlayingSavedState;

	Asura_Assert( m_pxTrackData || !m_bPlaying, "Loaded music with no track data, but is playing?" );
	
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		xStream >> m_fLayerVolumes[ iLayer ];
	}

	// Are we in the process of seeking
	bool bSeeking;
	xStream >> bSeeking;

	if( bSeeking )
	{
		m_eSeekState = SEEK_STATE_PREPARING;
		// read the time we are seeking to
		xStream >> m_fSeekPos;
		xStream >> m_bPendingSeekIsLoop;
	}

	// Now we have all our data, create the voices
	Asura_Assert( m_pxTrackData, "Failed to find track data" );
	if( m_pxTrackData )
	{
		//Create streaming sounds 
		for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
		{
			if( !m_pxTrackData->m_axLayerFiles[ iLayer ].IsBlank() )
			{
				m_iLayerHandles[ iLayer ] = Asura_Audio_System::CreateStreamingVoice( Asura_GetHashID( m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ), false, true, fPlayPos );
				Asura_Assertf( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE, ("Failed to create streaming sound from file %s", m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ) );

				if( bSeeking )
				{
					m_iLayerSeekHandles[ iLayer ] = Asura_Audio_System::CreateStreamingVoice( Asura_GetHashID( m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ), false, true, m_fSeekPos );
					Asura_Assertf( m_iLayerSeekHandles[ iLayer ] != ASURA_INVALID_HANDLE, ("Failed to create streaming sound from file %s", m_pxTrackData->m_axLayerFiles[ iLayer ].GetString() ) );
				}
			}
		}
	}
	else
	{
		m_eSeekState = SEEK_STATE_NONE;
	}
}


void Asura_Dynamic_Music_ActiveTrack::Pause ( )
{
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_Voice* pxVoice = GetLayerVoice( iLayer );
			if( pxVoice )
			{
				pxVoice->SetPitch( 0.0f );
			}
		}
	}
}


void Asura_Dynamic_Music_ActiveTrack::Unpause ( )
{
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( m_iLayerHandles[ iLayer ] != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_Voice* pxVoice = GetLayerVoice( iLayer );
			if( pxVoice )
			{
				pxVoice->SetPitch( 1.0f );
			}
		}
	}
}


#ifdef ASURA_USE_DEBUGOVERLAY
void Asura_Dynamic_Music_ActiveTrack::DoActiveTrackDebugOverlay () const
{
	if( !m_pxTrackData )
	{
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "This track has not been prepared yet" );
		return;
	}

	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, m_pxTrackData->m_xName.GetString() );
	if( !IsPlaying() )
	{
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "This track is not playing yet" );
		return;
	}

	const int iBodyMinutes = static_cast<int>( GetIntroTime() / 60.0f);
	const int iBodySeconds = static_cast<int>( GetIntroTime() ) % 60;

	const int iOutroMinutes = static_cast<int>( GetOutroTime() / 60.0f);
	const int iOutroSeconds = static_cast<int>( GetOutroTime() ) % 60;
	
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Loop start/end : %02d:%02d / %02d:%02d ", iBodyMinutes, iBodySeconds, iOutroMinutes, iOutroSeconds );	
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "BPM/BPB/Volume : %d / %d / %.2f", m_pxTrackData->m_iBPM, m_pxTrackData->m_iBeatsPerBar, m_pxTrackData->m_fVolume );
	
	const float fPlayTime = GetPlayTime();
	const int iPlayMinutes = static_cast<int>( fPlayTime / 60.0f);
	const int iPlaySeconds = static_cast<int>( fPlayTime ) % 60;
	
	char szProgress[16];
	szProgress[0] = '[';
	szProgress[11] = ']';
	szProgress[12] = 0;

	int iProgressPos = static_cast<int>( ( fPlayTime - GetIntroTime() ) / GetBodyDuration() * 10.0f );
	for(int iTick = 1; iTick < 11; ++iTick)
	{
		if( iTick-1 <= iProgressPos )
		{
			szProgress[iTick] = '-';
		}
		else
		{
			szProgress[iTick] = ' ';
		}
	}

	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Play Pos       : %02d:%02d   %s", iPlayMinutes, iPlaySeconds, szProgress );
	
	iProgressPos = static_cast<int>( ( GetNextBarBoundry() - fPlayTime ) / GetBarDuration() * 10.0f );
	for(int iTick = 1; iTick < 11; ++iTick)
	{
		if( iTick-1 <= iProgressPos )
		{
			szProgress[iTick] = '-';
		}
		else
		{
			szProgress[iTick] = ' ';
		}
	}

	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Next bar       : %1.1f     %s", GetNextBarBoundry() - fPlayTime, szProgress );

	ASURA_STATIC_ASSERT( ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS == 3 );	// Following line will need to be changed if ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS != 3
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Layer volumes  : (1) %.2f (2) %.2f (3) %.2f ", m_fLayerVolumes[0], m_fLayerVolumes[1], m_fLayerVolumes[2] );
}
#endif // ASURA_USE_DEBUGOVERLAY

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

Asura_Dynamic_Music_TrackData*						Asura_Dynamic_Music_System::s_axTracks				= 0;
int													Asura_Dynamic_Music_System::s_iNumTracks			= 0;

Asura_String										Asura_Dynamic_Music_System::s_axLayerNames			[ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS];

Asura_List<Asura_Dynamic_Music_ActiveTrigger>		Asura_Dynamic_Music_System::s_xActiveTriggerList;
Asura_Dynamic_Music_ActiveTrigger*					Asura_Dynamic_Music_System::s_pxCurrentTrigger		= 0;

Asura_Dynamic_Music_ActiveTrack*					Asura_Dynamic_Music_System::s_pxActiveTracks		[ Asura_Dynamic_Music_System::NUM_ACTIVE_TRACKS ];

Asura_List<Asura_Dynamic_Music_Trigger_Data>		Asura_Dynamic_Music_System::s_xTriggerDataList;

float												Asura_Dynamic_Music_System::s_fMusicMixerVolume		= 0.0f;

volatile Asura_Dynamic_Music_System::ResumeState	Asura_Dynamic_Music_System::s_eResumeState			= Asura_Dynamic_Music_System::RESUME_STATE_NONE;

bool												Asura_Dynamic_Music_System::s_bPaused				= false;

bool												Asura_Dynamic_Music_System::s_bInitialised			= false;

// Asura_Dynamic_Music_System member functions ////////////////////////////////

void Asura_Dynamic_Music_System::Initialise ()
{
	for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
	{
		char szBuff [ASURA_DYNAMIC_MUSIC_SYSTEM_LAYER_NAME_MAX_LENGTH];
		Asura_PrintfToBfr( szBuff, "Layer %d", ASURA_DYNAMIC_MUSIC_SYSTEM_LAYER_NAME_MAX_LENGTH, true, i + 1 );

		s_axLayerNames[ i ] = szBuff;
	}

	for( int i = 0; i < NUM_ACTIVE_TRACKS; ++i )
	{
		s_pxActiveTracks[ i ] = 0;
	}

	s_fMusicMixerVolume = Asura_Sound_Mixer_System::GetMixerVolume( Asura_Sound_Mixer_System::FindChannel( uMIXER_HASHID_MUSIC ) );

	#ifdef ASURA_USE_DEBUGOVERLAY
	if( !Asura_DebugOverlay::ModeIsRegistered( ASURA_DO_DYNAMICMUSIC ) )
	{
		Asura_DebugOverlay::RegisterMode( "DynamicMusic", ASURA_DO_DYNAMICMUSIC, DoDynamicMusicDebugOverlay, NULL );
	}	
	#endif

	s_bInitialised = true;
}


void Asura_Dynamic_Music_System::Update ()
{
	if( s_eResumeState == RESUME_STATE_LOADING )
	{
		// we are currently loading our state (from another thread), do not update
		return;
	}

	if( !s_bInitialised || s_bPaused )
	{
		return;
	}

	s_fMusicMixerVolume = Asura_Sound_Mixer_System::GetMixerVolume( Asura_Sound_Mixer_System::FindChannel( uMIXER_HASHID_MUSIC ) );

	//Do we have any triggers pending?
	if( !s_pxCurrentTrigger && !s_xActiveTriggerList.IsEmpty() )
	{
		s_pxCurrentTrigger = s_xActiveTriggerList.PopHeadObject();
	}

	for( int iActiveTrack = 0; iActiveTrack < NUM_ACTIVE_TRACKS; ++iActiveTrack )
	{
		if( s_pxActiveTracks[ iActiveTrack ] )
		{
			s_pxActiveTracks[ iActiveTrack ]->Update ();
		}
	}

	if( s_eResumeState == RESUME_STATE_PREPARING )
	{
		// we are waiting for all streaming sounds to be prepared after loading, check if they are finished
		bool bStillResuming = false;
		
		for( int iActiveTrack = 0; iActiveTrack < NUM_ACTIVE_TRACKS; ++iActiveTrack )
		{
			if(		s_pxActiveTracks[ iActiveTrack ] 
				&&	( s_pxActiveTracks[ iActiveTrack ]->IsPreparing() 
				||	s_pxActiveTracks[ iActiveTrack ]->IsSeekState_Preparing() ) )
			{
				bStillResuming = true;
			}
		}

		if( bStillResuming )
		{
			return;
		}
		else
		{
			s_eResumeState = RESUME_STATE_NONE;

			for( int iActiveTrack = 0; iActiveTrack < NUM_ACTIVE_TRACKS; ++iActiveTrack )
			{
				if(	s_pxActiveTracks[ iActiveTrack ] && s_pxActiveTracks[ iActiveTrack ]->ShouldPlayAfterStateLoad() )
				{
					s_pxActiveTracks[ iActiveTrack ]->BeginPlayback();
				}
			}
		}
	}

	if( s_pxCurrentTrigger )
	{
		//Deal with the current trigger
		Asura_Dynamic_Music_Trigger_Data* pxData = GetTriggerDataByHash( s_pxCurrentTrigger->m_uTriggerHash );
		if( pxData )
		{
			switch( pxData->m_eType )
			{
				case DYNAMIC_MUSIC_TRIGGER_TRANSITION	: Update_LayerTransTrigger	( pxData->TriggerData.m_pxLayerTransition	); break;
				case DYNAMIC_MUSIC_TRIGGER_PLAY			: Update_PlayTrigger		( pxData->TriggerData.m_pxPlay				); break;
				case DYNAMIC_MUSIC_TRIGGER_STOP			: Update_StopTrigger		( pxData->TriggerData.m_pxStop				); break;
				case DYNAMIC_MUSIC_TRIGGER_JOINRELEASE	: Update_JoinReleaseTrigger	( pxData->TriggerData.m_pxJoinRelease		); break;
				
				case DYNAMIC_MUSIC_NUM_TRIGGER_TYPES:
				default:
				{
					Asura_Assert( false, "Unrecognised trigger type" );
				}
				break;
			}
		}
		else
		{
			Asura_Assertf( false, ( "Failed to find Dynamic Music trigger data for trigger HASH %u (you may need to update your snd file)", s_pxCurrentTrigger->m_uTriggerHash ) );
			FinishedWithCurrentTrigger();
		}
	}
}

void Asura_Dynamic_Music_System::Update_LayerTransTrigger	( const Asura_Dynamic_Music_Layer_Transition_Trigger_Data* pxData )
{
	if( !pxData || !s_pxCurrentTrigger )
	{
		return;
	}

	
	if( ( pxData->m_uTrack != ASURA_HASH_ID_UNSET ) && !s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] )
	{
		//The secondary track has not been created, now would be a good time to create it
		s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] = new Asura_Dynamic_Music_ActiveTrack ();
		if( s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] )
		{
			s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->Clear ();

			Asura_Dynamic_Music_TrackData* pxTrackData = GetTrackByHash( pxData->m_uTrack );
			Asura_Assert( pxTrackData, "Failed to find track" );

			if( !pxTrackData )
			{
				delete s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ];
				s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] = 0;
				return;
			}

			if( !pxData->m_bUseIntro )
			{
				s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->SetStartOffset( Asura_Dynamic_Music_ActiveTrack::START_OFFSET_BODY );
			}

			s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->SetTrackData ( pxTrackData );

			for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
			{
				//It would be good to optimise this so that the stream does not actually exist until necessary
				//using some sort of heuristic to determine when that is
				s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->SetLayerVolume ( i, 0.0f );
			}
		}
	}
	
	float fOldTriggerTime = s_pxCurrentTrigger->m_fTriggerTime;

	if( s_pxCurrentTrigger->m_fTriggerTime < 0.0f )
	{
		// If we need to play the next track within a very short time, wait to be prepared
		if( s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] )
		{
			const float fNextStart = pxData->m_bUseIntro ? pxData->m_fNextIntroStartOffset : pxData->m_fNextIntroEndOffset;
			if( ( fNextStart < fASURA_DMS_SEEK_PREPARE_TIME ) && s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->IsPreparing() )
			{
				// Wait to be prepared
				return;
			}
		}

		// See if we are at a bar boundry and can activate the trigger (or there is no current track)
		if ( !s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] )
		{ 
			s_pxCurrentTrigger->m_fTriggerTime = 0.0f;
		}
		else if ( !s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsSeekingToLoop() && s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->HitBarBoundary() )
		{
			if( pxData->NeedToPlayOutro() )
			{
				// We need to check that when the outro is due to play, we wont be seeking for looping (cant overlap seeks)
				float fPlayTimeWhenWeSeek = s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->GetPlayTime() + pxData->m_fPrevOutroStartOffset - fASURA_DMS_SEEK_PREPARE_TIME;
				//The track will loop seek the bar before the outro starts, so check the time we will seek is before that
				if( fPlayTimeWhenWeSeek < ( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->GetOutroTime() - fASURA_DMS_SEEK_PREPARE_TIME ) )
				{
					s_pxCurrentTrigger->m_fTriggerTime = 0.0f;
				}
			}
			else
			{
				s_pxCurrentTrigger->m_fTriggerTime = 0.0f;
			}
		}
	}
	else
	{
		//Hopefully this is accurate enough (may need to use the track playtime instead)
		s_pxCurrentTrigger->m_fTriggerTime += Asura_Timers::GetFrameTime ();
	}
	
	if( s_pxCurrentTrigger->m_fTriggerTime < 0.0f )
	{
		//We haven't activated the trigger yet so return
		return;
	}

	//The trigger is activated so apply the volume at this point
	if( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] )
	{
		const float fVolume = pxData->GetVolumeAtTime ( s_pxCurrentTrigger->m_fTriggerTime, Asura_Dynamic_Music_Layer_Transition_Trigger_Data::TRACK_PREVIOUS );
		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetAllLayerMaxVolume( fVolume );
		
		//Are we close to the outro, need to prepare to seek to it
		if( pxData->NeedToPlayOutro() )
		{
			const float fTimeToFind = pxData->m_fPrevOutroStartOffset - fASURA_DMS_SEEK_PREPARE_TIME;
			if( ( s_pxCurrentTrigger->m_fTriggerTime >= fTimeToFind ) && ( fOldTriggerTime < fTimeToFind ) )
			{
				//Asura_OutputToDebugger(("[DMS] Jumping to outro, track playtime is %.2f, trigger time is %.2f, server time is %.2f\n", s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->GetPlayTime(), s_pxCurrentTrigger->m_fTriggerTime, Asura_Timers::GetServerAbsoluteTime()));
				s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->PrepareToSeekToPosition( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->GetOutroTime() );
			}
		}

		// Are we waiting to seek to the outro?
		if(		( s_pxCurrentTrigger->m_fTriggerTime >= pxData->m_fPrevOutroStartOffset )	// passed the point where we need to go to the outro
			&&	s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsSeekState_Ready()				// seek is ready to activate
			&&	!s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsSeekingToLoop() )				// the seek isnt an internal loop by the active track
		{
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->ActivateSeek();
		}
	}

	if( s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] )
	{
		//Is it time to start this track playing?
		if( !s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->IsPlaying() )
		{
			float fTimeToFind = 0.0f;
			if( pxData->m_bUseIntro )
			{
				fTimeToFind = pxData->m_fNextIntroEndOffset - s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->GetIntroTime();
			}
			else
			{
				fTimeToFind = pxData->m_fNextIntroEndOffset;
			}

			Asura_Assert( ( fTimeToFind >= 0.0f ), "Not enough warning given to start track" );

			if( ( s_pxCurrentTrigger->m_fTriggerTime >= fTimeToFind ) && ( fOldTriggerTime < fTimeToFind ) )	//or we just passed the start time
			{
				Asura_Assert( !s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->IsPreparing(), "Track should be prepared by now" );
				//TODO: in reality we passed the playback time be a little bit, and should start at (s_pxCurrentTrigger->m_fTriggerTime - fTimeToFind)
				s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->BeginPlayback( );
			}
		}
				
		//The secondary track has been created, set its volume
		if( s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->IsPlaying() )
		{
			const float fVolume = pxData->GetVolumeAtTime ( s_pxCurrentTrigger->m_fTriggerTime, Asura_Dynamic_Music_Layer_Transition_Trigger_Data::TRACK_NEXT );
			for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
			{
				if( !pxData->m_bMuteLayers[ i ] )
				{
					s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ]->SetLayerVolume ( i, fVolume );
				}
			}
		}
	}
	
	if( s_pxCurrentTrigger->m_fTriggerTime >= pxData->GetFinishTime() )
	{
		//Trigger is finished, clear the old primary track and set the secondary to be primary
		if( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] )
		{
			delete s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ];
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] = 0; 
		}
		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] = s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ];
		s_pxActiveTracks[ ACTIVE_TRACK_SECONDARY ] = 0;
		
		FinishedWithCurrentTrigger();
	}
}


void Asura_Dynamic_Music_System::Update_PlayTrigger			( const Asura_Dynamic_Music_Play_Trigger_Data* pxData )
{
	if( pxData->m_uTrack == ASURA_HASH_ID_UNSET)
	{
		FinishedWithCurrentTrigger();
		return;
	}

	if(		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] 
		&&	s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->GetNameHash() == pxData->m_uTrack )
	{
		// The track has already been set, we are either waiting for it to be prepared or it was already playing
		if( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsPreparing() )
		{
			// Still preparing
			return;
		}
		if( !s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsPlaying() )
		{
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->BeginPlayback();
		}
		//and we're done...
		FinishedWithCurrentTrigger();
		return;
	}

	//the active track is not the desired one... we need to stop the current tracks
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		if( s_pxActiveTracks[ iTrack ] )
		{
			s_pxActiveTracks[ iTrack ]->Clear();
			delete s_pxActiveTracks[ iTrack ];
			s_pxActiveTracks[ iTrack ] = 0;
		}
	}

	//now play the next track
	if( pxData->m_uTrack != ASURA_HASH_ID_UNSET )
	{
		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] = new Asura_Dynamic_Music_ActiveTrack ();
		if( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] )
		{
			Asura_Dynamic_Music_TrackData* pxTrackData = GetTrackByHash( pxData->m_uTrack );
			Asura_Assert( pxTrackData, "Failed to find track" );
			if( !pxTrackData )
			{
				delete s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ];
				s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] = 0;
				return;
			}
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetTrackData ( pxTrackData );

			for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
			{
				if( pxData->m_bMuteLayers[ i ] )
				{
					//It would be good to optimise this so that the stream does not actually exist until necessary
					//using some sort of heuristic to determine when that is
					s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetLayerVolume ( i, 0.0f );
				}
			}
		}
	}

	if( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsPreparing() )
	{
		// Wait until we are prepared
		return;
	}
	s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->BeginPlayback();

	//and we're done...
	FinishedWithCurrentTrigger();
}


void Asura_Dynamic_Music_System::Update_StopTrigger			( const Asura_Dynamic_Music_Stop_Trigger_Data* pxData )
{
	if( !pxData || !s_pxCurrentTrigger )
	{
		return;
	}

	Asura_Dynamic_Music_ActiveTrack* &pxPrimaryTrack = s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ];
	float fOldTriggerTime = s_pxCurrentTrigger->m_fTriggerTime;

	if( s_pxCurrentTrigger->m_fTriggerTime < 0.0f )
	{
		if ( !pxPrimaryTrack )
		{ 
			//There isnt a primary track, we cant do anything
			FinishedWithCurrentTrigger();
			return;
		}

		// Dont do anything while we are seeking to loop
		if( !pxPrimaryTrack->IsSeekingToLoop() )
		{
			if(		pxPrimaryTrack->IsSeekState_None()
				&&	pxData->NeedToPlayOutro()
				&& ( pxPrimaryTrack->GetPlayTime() < ( pxPrimaryTrack->GetOutroTime() - fASURA_DMS_SEEK_PREPARE_TIME ) ) )
			{
				// Seek to the outro
				pxPrimaryTrack->PrepareToSeekToPosition( pxPrimaryTrack->GetOutroTime() );
			}
			else if( pxPrimaryTrack->IsSeekState_Ready() || !pxData->NeedToPlayOutro() )
			{
				// Should we activate?
				if( pxPrimaryTrack->HitBarBoundary() || !pxData->m_bActivateOnBarBoundry )
				{
					if( !pxData->m_bJustOutro && pxData->NeedToPlayOutro() )
					{
						// We need to check that when the outro is due to play, we wont be seeking for looping (cant overlap seeks)
						float fPlayTimeWhenWeSeek = pxPrimaryTrack->GetPlayTime() + pxData->m_fBodyEndTime;
						//The track will loop seek fASURA_DMS_SEEK_PREPARE_TIME before the outro starts, so check the time we will seek is before that
						if( fPlayTimeWhenWeSeek < ( pxPrimaryTrack->GetOutroTime()- fASURA_DMS_SEEK_PREPARE_TIME ) )
						{
							s_pxCurrentTrigger->m_fTriggerTime = 0.0f;
						}
					}
					else
					{
						s_pxCurrentTrigger->m_fTriggerTime = 0.0f;
					}
				}
			}
		}
		//Trigger is not active and no further action is required
		return;
	}
	else
	{
		//Hopefully this is accurate enough
		s_pxCurrentTrigger->m_fTriggerTime += Asura_Timers::GetFrameTime ();
	}

	Asura_Assert(pxPrimaryTrack, "Updating stop trigger without a primary track - this shouldn't be possible.");
	if (pxPrimaryTrack)
	{
		//Update the volumes based on our trigger data
		const float fVolume = pxData->GetVolumeAtTime( s_pxCurrentTrigger->m_fTriggerTime );
		pxPrimaryTrack->SetAllLayerMaxVolume( fVolume );

		//Are we less than one bar away from the outro? If so we need to prepare to jump to it
		if( !pxData->m_bJustOutro && pxData->NeedToPlayOutro() )
		{
			if( ( s_pxCurrentTrigger->m_fTriggerTime > pxData->m_fBodyEndTime ) && ( fOldTriggerTime < pxData->m_fBodyEndTime ) )
			{
				pxPrimaryTrack->ActivateSeek ();
			}
		}
	}

	if( s_pxCurrentTrigger->m_fTriggerTime >= pxData->GetFinishTime() )
	{
		FinishedWithCurrentTrigger();
		delete pxPrimaryTrack;
		pxPrimaryTrack = 0;
	}
}


void Asura_Dynamic_Music_System::Update_JoinReleaseTrigger	( const Asura_Dynamic_Music_JoinRelease_Trigger_Data* pxData )
{
	//In the future this logic will probably change quite a bit since we dont really want streams running unless
	//necessary. So future work for this function: start the appropriate layers playing and sync to the existing play time
	
	if( !pxData || !s_pxCurrentTrigger )
	{
		return;
	}

	if( !pxData )
	{
		return;
	}

	if( s_pxCurrentTrigger->m_fTriggerTime < 0.0f )
	{
		//See if we are at a bar boundry and can activate the trigger (or we aren't activating on bar boundries)
		if ( !s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] )
		{ 
			//There isnt a primary track, we cant do anything
			FinishedWithCurrentTrigger();
			return;
		}
		else if (		( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] && s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->HitBarBoundary() )
					||	!pxData->m_bActivateOnBarBoundry )
		{
			//Time to activate
			s_pxCurrentTrigger->m_fTriggerTime = 0.0f;
		}
		else
		{
			//Trigger isnt ready to be activated
			return;
		}
	}
	else
	{
		//Hopefully this is accurate enough
		s_pxCurrentTrigger->m_fTriggerTime += Asura_Timers::GetFrameTime ();
	}

	//Update the volumes based on our trigger data
	for( int iLayer = 0; iLayer < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++iLayer )
	{
		if( pxData->m_xLayerChanges[iLayer].m_eAction != Asura_Dynamic_Music_JoinRelease_Trigger_Data::ACTION_NONE )
		{
			const float fVolume = pxData->GetVolumeAtTime( iLayer, s_pxCurrentTrigger->m_fTriggerTime );
			
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetLayerVolume( iLayer, fVolume );
		}
	}
	

	if( s_pxCurrentTrigger->m_fTriggerTime >= pxData->GetFinishTime() )
	{
		FinishedWithCurrentTrigger();
	}
}


void Asura_Dynamic_Music_System::FinishedWithCurrentTrigger ()
{
	delete s_pxCurrentTrigger;
	s_pxCurrentTrigger = 0;
}


void Asura_Dynamic_Music_System::Shutdown ()
{
	if( !s_bInitialised )
	{
		Asura_Assert( false, "Dynamic music system not initialised" );
		return;
	}

	s_eResumeState = RESUME_STATE_NONE;

	//Clean up all active stuff (just call stop)
	Stop();

	//clean up tracks
	delete []s_axTracks;
	s_axTracks = 0;

	//clean up triggers
	while( !s_xTriggerDataList.IsEmpty() )
	{
		Asura_Dynamic_Music_Trigger_Data* pxData = s_xTriggerDataList.PopHeadObject();
		delete pxData;
	}
}


void Asura_Dynamic_Music_System::Reset ()
{
	Shutdown	();
	Initialise	();
}

void Asura_Dynamic_Music_System::WriteStateToChunkStream ( Asura_Chunk_Stream& xStream )
{
	//TODO: It might be wise to pause the DMS here since we dont want its state to be changing while we are reading it

	unsigned char ucVersion = DYNAMIC_MUSIC_LOADSAVE_STATE_VERSION;

	xStream << ucVersion;

	// Write out the state of the active tracks
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		const bool bTrackExists = ( s_pxActiveTracks[ iTrack ] != 0 );
		xStream << bTrackExists;

		if( bTrackExists )
		{			
			s_pxActiveTracks[ iTrack ]->WriteStateToChunkStream( xStream );
		}
	}

	// Write out the state of the active trigger
	const bool bActiveTriggerExists = ( s_pxCurrentTrigger != 0 );
	xStream << bActiveTriggerExists;

	if( bActiveTriggerExists )
	{
		xStream << s_pxCurrentTrigger->m_uTriggerHash;
		xStream << s_pxCurrentTrigger->m_fTriggerTime;
	}

	// Write out any queued triggers
	const u_int uNumQueuedTrigs = s_xActiveTriggerList.GetNumEntries();

	xStream << uNumQueuedTrigs;

	for( u_int uTrig = 0; uTrig < uNumQueuedTrigs; ++uTrig )
	{
		// We just need the trigger guid's from these
		Asura_Dynamic_Music_ActiveTrigger* pxTrig = s_xActiveTriggerList.GetEntryByIndex( uTrig );
		xStream << pxTrig->m_uTriggerHash;
	}
}


void Asura_Dynamic_Music_System::ReadStateFromChunkStream ( Asura_Chunk_Stream& xStream )
{
	// Flag ourselves as resuming
	s_eResumeState = RESUME_STATE_LOADING;

	Stop();

	unsigned char ucVersion = 0;

	xStream >> ucVersion;

	if( ucVersion > DYNAMIC_MUSIC_LOADSAVE_STATE_VERSION )
	{
		Asura_Assert( false, "Dynamic music in game state being loaded is the wrong version, ignoring" );
		return;
	}

	// Read the state of the active tracks
	ASURA_STATIC_ASSERT( Asura_Dynamic_Music_System::NUM_ACTIVE_TRACKS == 2 );

	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		bool bTrackExists;
		xStream >> bTrackExists;

		if( bTrackExists )
		{
			s_pxActiveTracks[ iTrack ] = new Asura_Dynamic_Music_ActiveTrack();
			s_pxActiveTracks[ iTrack ]->ReadStateFromChunkStream( xStream );
		}
	}

	// Read the state of the active trigger
	bool bActiveTriggerExists;
	xStream >> bActiveTriggerExists;

	if( bActiveTriggerExists )
	{
		s_pxCurrentTrigger = new Asura_Dynamic_Music_ActiveTrigger;

		// < version 2 is a GUID, which is still a unique unsigned int - so we're ok
		xStream >> s_pxCurrentTrigger->m_uTriggerHash;
		xStream >> s_pxCurrentTrigger->m_fTriggerTime;
	}

	// Read any queued triggers
	u_int uNumQueuedTrigs;
	xStream >> uNumQueuedTrigs;

	while( uNumQueuedTrigs )
	{
		Asura_Dynamic_Music_ActiveTrigger* pxTrig = new Asura_Dynamic_Music_ActiveTrigger;

		// < version 2 is a GUID, which is still a unique unsigned int - so we're ok
		xStream >> pxTrig->m_uTriggerHash;

		s_xActiveTriggerList.AddToList( pxTrig );

		--uNumQueuedTrigs;
	}

	s_eResumeState = RESUME_STATE_PREPARING;
}


bool Asura_Dynamic_Music_System::IsRestoringState ()
{
	return ( s_eResumeState != RESUME_STATE_NONE );
}


void Asura_Dynamic_Music_System::Pause ()
{
	if( s_bPaused )
	{
		return;
	}

	for( int iActiveTrack = 0; iActiveTrack < NUM_ACTIVE_TRACKS; ++iActiveTrack )
	{
		if( s_pxActiveTracks[ iActiveTrack ] )
		{
			s_pxActiveTracks[ iActiveTrack ]->Pause ();
		}
	}

	s_bPaused = true;
}


void Asura_Dynamic_Music_System::Unpause ()
{
	if( !s_bPaused )
	{
		return;
	}

	for( int iActiveTrack = 0; iActiveTrack < NUM_ACTIVE_TRACKS; ++iActiveTrack )
	{
		if( s_pxActiveTracks[ iActiveTrack ] )
		{
			s_pxActiveTracks[ iActiveTrack ]->Unpause ();
		}
	}

	s_bPaused = false;
}


bool Asura_Dynamic_Music_System::IsPaused ()
{
	return s_bPaused;
}


void Asura_Dynamic_Music_System::Stop ()
{
	if( !s_bInitialised )
	{
		Asura_Assert( false, "Dynamic music system not initialised" );
		return;
	}

	//Clear active tracks
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		if( s_pxActiveTracks[ iTrack ] )
		{
			s_pxActiveTracks[ iTrack ]->Clear();
			delete s_pxActiveTracks[ iTrack ];
			s_pxActiveTracks[ iTrack ] = 0;
		}
	}

	//Clear active triggers
	while( !s_xActiveTriggerList.IsEmpty() )
	{
		Asura_Dynamic_Music_ActiveTrigger* pxTrigger = s_xActiveTriggerList.PopHeadObject ();
		delete pxTrigger;
	}
	delete s_pxCurrentTrigger;
	s_pxCurrentTrigger = 0;

	// Stopping has the effect of unpausing
	s_bPaused = false;
}

void Asura_Dynamic_Music_System::PlayTrack ( const Asura_Hash_ID uTrackHash, const float fFadeOutTime, const float fFadeInTime )
{
	if( !s_bInitialised )
	{
		Asura_Assert( false, "Dynamic music system not initialised" );
		return;
	}

	// Temporary function to allow music tracks to be triggered in code

	const Asura_Hash_ID uTempHash = 0x76507E51;	// string "temporary"

	// BF3 seems to want to call playtrack every frame, so check we aren't already playing this track
	if(		( s_pxActiveTracks[ACTIVE_TRACK_PRIMARY]	&& ( s_pxActiveTracks[ACTIVE_TRACK_PRIMARY]->GetNameHash() == uTrackHash ) )
		||	( s_pxActiveTracks[ACTIVE_TRACK_SECONDARY]	&& ( s_pxActiveTracks[ACTIVE_TRACK_SECONDARY]->GetNameHash() == uTrackHash ) ) )
	{
		return;
	}

	// Grab the temporary trigger, create it if it does not exist

	if( IsTriggerActive( uTempHash ) )
	{
		Asura_Assert( false, "Already playing a music track triggered in code" );
		return;
	}

	Asura_Dynamic_Music_Trigger_Data* pxTrigger = GetTriggerDataByHash( uTempHash );
	
	if( !pxTrigger )
	{
		pxTrigger = CreateNewTriggerData();
	}

	pxTrigger->m_uHash = uTempHash;
	pxTrigger->SetTriggerType( DYNAMIC_MUSIC_TRIGGER_TRANSITION );

	Asura_Dynamic_Music_Layer_Transition_Trigger_Data* pxData = pxTrigger->TriggerData.m_pxLayerTransition;

	pxData->m_uTrack = uTrackHash;

	pxData->m_fPrevOutroStartOffset = 1000.0f;	// We dont intend to play the outro
	pxData->m_fPrevOutroEndOffset	= 1001.0f;
	
	// Find out how long the intro is
	Asura_Dynamic_Music_TrackData* pxTrackData = GetTrackByHash( uTrackHash );
	if( !pxTrackData )
	{
		Asura_Assertf( false, ( "Failed to find track data with hash %u", uTrackHash ) );
		return;
	}

	pxData->m_fNextIntroStartOffset	= 0.0f;
	pxData->m_fNextIntroEndOffset	= pxTrackData->m_fBodyOffset;

	pxData->m_bUseIntro = true;

	Asura_Mem::Set( pxData->m_bMuteLayers, 0, sizeof( bool ) * ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS );

	//These points define the linear volume changes to fade in/out the tracks
	ASURA_STATIC_ASSERT( ASURA_DYNAMIC_MUSIC_SYSTEM_FADE_RESOLUTION == 3 );

	pxData->m_xPrevTrackAmplitude[0].m_fTime = 0.0f;
	pxData->m_xPrevTrackAmplitude[0].m_fAmplitude = 1.0f;

	pxData->m_xPrevTrackAmplitude[1].m_fTime = fFadeOutTime;
	pxData->m_xPrevTrackAmplitude[1].m_fAmplitude = 0.0f;

	pxData->m_xPrevTrackAmplitude[2].m_fTime = 1000.0f;
	pxData->m_xPrevTrackAmplitude[2].m_fAmplitude = 0.0f;

	pxData->m_xNewTrackAmplitude[0].m_fTime = 0.0f;
	pxData->m_xNewTrackAmplitude[0].m_fAmplitude = 0.0f;

	pxData->m_xNewTrackAmplitude[1].m_fTime = fFadeInTime;
	pxData->m_xNewTrackAmplitude[1].m_fAmplitude = 1.0f;

	pxData->m_xNewTrackAmplitude[2].m_fTime = 1000.0f;
	pxData->m_xNewTrackAmplitude[2].m_fAmplitude = 1.0f;

	// Activate this trigger
	ActivateTrigger( uTempHash );
}

#ifdef ASURA_TOOLS

void Asura_Dynamic_Music_System::PlayTrack ( Asura_Hash_ID uNameHash, int iPosition )
{
	if( !s_bInitialised )
	{
		Asura_Assert( false, "Dynamic music system not initialised" );
		return;
	}

	//This is just for tools authoring

	Stop ();

	s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] = new Asura_Dynamic_Music_ActiveTrack ();
	if( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] )
	{
		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->Clear ();

		Asura_Dynamic_Music_TrackData* pxTrackData = GetTrackByHash( uNameHash );
		if( !pxTrackData )
		{
			Asura_Assertf( false, ( "Failed to find track by hash %u", uNameHash ) );
			delete s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ];
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ] = 0;
			return;
		}
		
		if( iPosition == TRACK_PART_BODY )
		{
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetStartOffset( Asura_Dynamic_Music_ActiveTrack::START_OFFSET_BODY );
		}
		else if ( iPosition == TRACK_PART_OUTRO )
		{
			s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetStartOffset( Asura_Dynamic_Music_ActiveTrack::START_OFFSET_OUTRO );
		}
		
		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->SetTrackData ( pxTrackData );

		// Pretty horrible, busy loop while preparing, its for tools authoring tho, so we can live with it
		while( s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->IsPreparing() )
		{
			Asura_Thread_System::DelayCurrentThread( 0.05f );
		}

		s_pxActiveTracks[ ACTIVE_TRACK_PRIMARY ]->BeginPlayback();
	}
}

float Asura_Dynamic_Music_System::GetTriggerPlayTime ( Asura_Hash_ID uTriggerHash )
{
	//This is just for tools authoring
	if( s_pxCurrentTrigger && ( s_pxCurrentTrigger->m_uTriggerHash == uTriggerHash ) )
	{
		return s_pxCurrentTrigger->m_fTriggerTime;
	}
	return 0.0f;
}

#endif

#ifdef ASURA_USE_DEBUGOVERLAY
void Asura_Dynamic_Music_System::DoDynamicMusicDebugOverlay( void* const pxIgnore )
{
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		if( s_pxActiveTracks[ iTrack ] )
		{
			Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, ">>>>>>>> Active Track %d <<<<<<<<", (iTrack + 1 ) );
			s_pxActiveTracks[ iTrack ]->DoActiveTrackDebugOverlay();
		}
	}

	if( s_pxCurrentTrigger )
	{
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, ">>>>>>>> Current trigger <<<<<<<<" );
		DoActiveTriggerDebugOverlay( s_pxCurrentTrigger );
	}
	

	if( !s_xActiveTriggerList.IsEmpty() )
	{
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, ">>>>>>>> Trigger Queue <<<<<<<<" );

		Asura_List_It<Asura_Dynamic_Music_ActiveTrigger> xIter(s_xActiveTriggerList);

		while( !xIter.Done() )
		{
			Asura_Dynamic_Music_ActiveTrigger* pxTrigger = xIter.GetCurrent();

			if( pxTrigger )
			{
				DoActiveTriggerDebugOverlay( pxTrigger );
				Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "----------------------------" );
			}
			
			xIter.Next ();
		}
	}
}

void Asura_Dynamic_Music_System::DoActiveTriggerDebugOverlay( const Asura_Dynamic_Music_ActiveTrigger* pxTrigger )
{
	if( !pxTrigger )
	{
		return;
	}

	Asura_ServerEntity_DynamicMusicTrigger* pxEntity = Asura_ServerEntity_DynamicMusicTrigger::FindEntityByHash(pxTrigger->m_uTriggerHash);

	Asura_Assert(pxEntity, "Cannot find active trigger entity");

	#ifdef ASURA_TOOLS

	Tools_Entity_DynamicMusicTrigger_Object* pxToolsObject = 0;
	if( pxEntity )
	{
		Tools_Entity_Data* pxToolsEntityData = pxEntity->ToolsData;
		if( pxToolsEntityData )
		{
			pxToolsObject = reinterpret_cast<Tools_Entity_DynamicMusicTrigger_Object*>( pxToolsEntityData->ToolsObject );
		}
	}
	if( pxToolsObject )
	{
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Name: %s (GUID: %u)", pxToolsObject->GetName(), pxEntity->Guid);
	}
	else
	{
		if (pxEntity)
		{
			Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "GUID: %u", pxEntity->Guid );
		}
	}
	#else
	if (pxEntity)
	{
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "GUID: %u", pxEntity->Guid );
	}	
	#endif

	Asura_Dynamic_Music_Trigger_Data* pxData = GetTriggerDataByHash( pxTrigger->m_uTriggerHash );
	if( pxData )
	{
		switch( pxData->m_eType )
		{
			case DYNAMIC_MUSIC_TRIGGER_TRANSITION	: Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Type: Transition" );		break;
			case DYNAMIC_MUSIC_TRIGGER_PLAY			: Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Type: Play" );			break;
			case DYNAMIC_MUSIC_TRIGGER_STOP			: Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Type: Stop" );			break;
			case DYNAMIC_MUSIC_TRIGGER_JOINRELEASE	: Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Type: Join/Release" );	break;

			case DYNAMIC_MUSIC_NUM_TRIGGER_TYPES	: Asura_Assert( false, "Unexpected trigger type" );
		}
	}

	if( pxTrigger == s_pxCurrentTrigger )
	{
		// If this is debug for the current trigger, show how much time has passed for it
		Asura_DebugOverlay::DrawConsoleText( ASURA_DO_DYNAMICMUSIC, "Time: %.1f Secs", pxTrigger->m_fTriggerTime );
	}
}

#endif // ASURA_USE_DEBUGOVERLAY

void Asura_Dynamic_Music_System::HandleMessage ( const Asura_Message* pxMessage )
{
	switch(pxMessage->GetMsgID())
	{
		case ASURA_MSG_DYNAMIC_MUSIC_TRIGGER:
		{
			const Asura_MsgData_DynamicMusic_Trigger* pxData = static_cast<const Asura_MsgData_DynamicMusic_Trigger*>( pxMessage->GetData() );

			//Look up the trigger and activate it
			ActivateTrigger( pxData->m_uTriggerHash );
		}
	}
}

int Asura_Dynamic_Music_System::GetNumberOfTracks ()
{
	return s_iNumTracks;
}


Asura_Dynamic_Music_TrackData* Asura_Dynamic_Music_System::GetTrackByIndex ( const int iIndex )
{
	if( ( iIndex < 0 ) || ( iIndex >= s_iNumTracks) )
	{
		Asura_Assert ( false, "Trying to retrieve an invalid track from the dynamic music system" );
		return NULL;
	}
	return &s_axTracks[ iIndex ];
}


Asura_Dynamic_Music_TrackData* Asura_Dynamic_Music_System::GetTrackByHash ( Asura_Hash_ID uNameHash )
{
	if( uNameHash == ASURA_HASH_ID_UNSET )
	{
		return NULL;
	}
	for ( int iSeq = 0; iSeq < s_iNumTracks; ++iSeq )
	{
		if( s_axTracks[ iSeq ].m_xName.GetHash() == uNameHash )
		{
			return &s_axTracks[ iSeq ];
		}
	}
	return NULL;
}


bool Asura_Dynamic_Music_System::TrackExists ( Asura_Hash_ID uNameHash )
{
	for ( int iSeq = 0; iSeq < s_iNumTracks; ++iSeq )
	{
		if( s_axTracks[ iSeq ].m_xName.GetHash() == uNameHash )
		{
			return true;
		}
	}
	return false;
}


Asura_Dynamic_Music_TrackData* Asura_Dynamic_Music_System::CreateNewTrack ()
{
	//Create a new track data object by appending it to the end of the list of tracks
	if(s_iNumTracks == 0)
	{
		//easy case
		s_axTracks = new Asura_Dynamic_Music_TrackData[1];
		s_iNumTracks = 1;
	}
	else
	{
		Asura_Dynamic_Music_TrackData* pxNewArray = new Asura_Dynamic_Music_TrackData [ s_iNumTracks + 1 ];
		//manually assign all of the elements of the old array to the new array, bit slow but we need a deep copy
		for( int i = 0; i < s_iNumTracks; ++i )
		{
			pxNewArray[ i ] = s_axTracks[ i ];
		}
		delete []s_axTracks;
		s_axTracks = pxNewArray;
		++s_iNumTracks;
	}
	return &s_axTracks[ s_iNumTracks - 1 ];
}


void Asura_Dynamic_Music_System::DeleteTrack ( Asura_Hash_ID uNameHash )
{
	if( s_iNumTracks == 1 )
	{
		//easy case
		if( s_axTracks[0].m_xName.GetHash() == uNameHash )
		{
			delete []s_axTracks;
			s_axTracks = 0;
			s_iNumTracks = 0;
		}
	}
	else
	{
		//Make a new array and assign all the old element except the deleted one into it
		Asura_Dynamic_Music_TrackData* pxNewArray = new Asura_Dynamic_Music_TrackData [ s_iNumTracks - 1 ];
		int iNewIndex = 0;
		for( int i = 0; i < s_iNumTracks; ++i )
		{
			if( s_axTracks[i].m_xName.GetHash() != uNameHash )
			{
				pxNewArray[ iNewIndex ] = s_axTracks[ i ];
				++iNewIndex;
			}
		}
		delete []s_axTracks;
		s_axTracks = pxNewArray;
		--s_iNumTracks;
	}
}


void Asura_Dynamic_Music_System::ClearTracks ()
{
	Stop ();

	delete []s_axTracks;
	s_axTracks = 0;
	s_iNumTracks = 0;
}


bool Asura_Dynamic_Music_System::TrackIsPlaying ( Asura_Hash_ID uNameHash )
{
	//Clear active tracks
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		if( s_pxActiveTracks[ iTrack ] && ( s_pxActiveTracks[ iTrack ]->GetNameHash() == uNameHash ) )
		{
			return s_pxActiveTracks[ iTrack ]->IsPlaying();
		}
	}
	return false;
}


float Asura_Dynamic_Music_System::GetTrackPlayTime ( Asura_Hash_ID uNameHash )
{
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		if( s_pxActiveTracks[ iTrack ] && ( s_pxActiveTracks[ iTrack ]->GetNameHash() == uNameHash ) )
		{
			if( s_pxActiveTracks[ iTrack ]->IsPlaying() )
			{
				return s_pxActiveTracks[ iTrack ]->GetPlayTime();
			}
		}
	}
	return 0.0f;
}


char* Asura_Dynamic_Music_System::GetLayerName ( const int iIndex )
{
	if( ( iIndex >= 0 ) && ( iIndex < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ) )
	{
		return s_axLayerNames[iIndex].GetString();
	}
	return 0;
}


void Asura_Dynamic_Music_System::SetLayerName ( const int iIndex, Asura_String xString )
{
	if( ( iIndex >= 0 ) && ( iIndex < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ) )
	{
		s_axLayerNames[iIndex] = xString;
	}
}


bool Asura_Dynamic_Music_System::ActivateTrigger ( Asura_Hash_ID uTriggerHash )
{
	if( !s_bInitialised )
	{
		Asura_Assert( false, "Dynamic music system not initialised" );
		return false;
	}

	//Check if this is a stop trigger, if so we may want to clear the trigger queue
	Asura_Dynamic_Music_Trigger_Data* pxTriggerData = GetTriggerDataByHash( uTriggerHash );
	if( pxTriggerData && ( pxTriggerData->m_eType == DYNAMIC_MUSIC_TRIGGER_STOP ) )
	{
		Asura_Dynamic_Music_Stop_Trigger_Data* pxStopTrigger = pxTriggerData->TriggerData.m_pxStop;
		if( pxStopTrigger->m_bUnqueueAllTriggers )
		{
			//Clear all queued triggers
			s_xActiveTriggerList.ClearList();
		}
	}

	//Queue this trigger
	Asura_Dynamic_Music_ActiveTrigger* pxActiveTrigger = new Asura_Dynamic_Music_ActiveTrigger;

	if( pxActiveTrigger )
	{
		pxActiveTrigger->Clear();
		pxActiveTrigger->m_uTriggerHash = uTriggerHash;

		s_xActiveTriggerList.PushTail( pxActiveTrigger );
		return true;
	}
	return false;
}


bool Asura_Dynamic_Music_System::IsTriggerActive ( Asura_Hash_ID uTriggerHash )
{
	if( s_pxCurrentTrigger && ( s_pxCurrentTrigger->m_uTriggerHash == uTriggerHash ) )
	{
		return true;
	}

	Asura_List_It<Asura_Dynamic_Music_ActiveTrigger> xIter(s_xActiveTriggerList);
	
	while( !xIter.Done() )
	{
		Asura_Dynamic_Music_ActiveTrigger* pxTrigger = xIter.GetCurrent();
		if(pxTrigger && ( pxTrigger->m_uTriggerHash == uTriggerHash ) )
		{
			return true;
		}
		xIter.Next ();
	}
	return false;
}


Asura_Dynamic_Music_Trigger_Data* Asura_Dynamic_Music_System::CreateNewTriggerData ( bool bCodeTrigger /*= false*/ )
{
	Asura_Dynamic_Music_Trigger_Data* pxNewData = new Asura_Dynamic_Music_Trigger_Data;
	pxNewData->m_bCodeTrigger = bCodeTrigger;
	s_xTriggerDataList.AddToList( pxNewData );
	return pxNewData;
}


Asura_Dynamic_Music_Trigger_Data* Asura_Dynamic_Music_System::GetTriggerDataByHash ( const Asura_Hash_ID uTriggerHash )
{
	Asura_List_It<Asura_Dynamic_Music_Trigger_Data> xIter( s_xTriggerDataList );

	while ( !xIter.Done() )
	{
		Asura_Dynamic_Music_Trigger_Data* pxData = xIter.GetCurrent();
		if( pxData && ( pxData->m_uHash == uTriggerHash ) )
		{
			return pxData;
		}
		xIter.Next();
	}

	return NULL;
}


int Asura_Dynamic_Music_System::GetNumberOfTriggerData ()
{
	return s_xTriggerDataList.GetNumEntries();
}


Asura_Dynamic_Music_Trigger_Data* Asura_Dynamic_Music_System::GetTriggerDatabyIndex ( const int iIndex )
{
	return s_xTriggerDataList.GetEntryByIndex( iIndex );
}

void Asura_Dynamic_Music_System::DeleteTriggerData ( Asura_Hash_ID uTriggerHash )
{
	Asura_Dynamic_Music_Trigger_Data* pxData = GetTriggerDataByHash( uTriggerHash );
	if( pxData )
	{
		s_xTriggerDataList.RemoveFromList( pxData );
		delete pxData;
	}
}

bool Asura_Dynamic_Music_System::IsPlaying ()
{
	for( int iTrack = 0; iTrack < NUM_ACTIVE_TRACKS; ++iTrack )
	{
		if( s_pxActiveTracks[ iTrack ] )
		{
			return true;
		}
	}
	return false;
}

#endif //ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM 

//eof
