// Asura_DynamicMusic_System.h
// Thomas Young 12.08.08
// Copyright (C) 2008 Rebellion, All Rights Reserved.

// F I L E   W R A P P E R ///////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __ASURA_DYNAMIC_MUSIC_SYSTEM__
#define __ASURA_DYNAMIC_MUSIC_SYSTEM__

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM

#ifndef ASURA_USE_AUDIO_SYSTEM
#error The dynamic music system requires the audio system to be enabled
#endif

// D E F I N E S /////////////////////////////////////////////////////////////

#define ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS				3
#define ASURA_DYNAMIC_MUSIC_SYSTEM_LAYER_NAME_MAX_LENGTH	16

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_List.h"
#include "Network/Asura_Message.h"
#include "Sound/Asura_Dynamic_Music_System_Triggers.h"

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

class Asura_Dynamic_Music_ActiveTrigger;
class Asura_Dynamic_Music_ActiveTrack;
class Asura_Dynamic_Music_TrackData;

class Asura_Dynamic_Music_Trigger_Data;
class Asura_Dynamic_Music_Layer_Transition_Trigger_Data;
class Asura_Dynamic_Music_Play_Trigger_Data;
class Asura_Dynamic_Music_Stop_Trigger_Data;
class Asura_Dynamic_Music_JoinRelease_Trigger_Data;

// C L A S S   D E C L A R A T I O N S //////////////////////////////////////

// Dynamic music system messages

class Asura_MsgData_DynamicMusic_Trigger
{
	public:
		
		Asura_Hash_ID m_uTriggerHash;
};

// Dynamic music system class

class Asura_Dynamic_Music_System
{
	friend class Asura_Chunk_DynamicMusic;

	public:

		// --------------- public enumerations ----------------------

		// version 2 - now saving hash ID instead of GUID in server entity - JCT
		enum { DYNAMIC_MUSIC_CURRENT_TRIGGER_ENTITY_VERSION = 2 };	//This is used by the server entities

		enum TrackParts
		{
			TRACK_PART_INTRO	= 0,
			TRACK_PART_BODY		,
			TRACK_PART_OUTRO	,
		};

		// ---------------- public functions ------------------------

		Asura_Dynamic_Music_System	()	{};
		~Asura_Dynamic_Music_System	()	{};

		static void 								Initialise 					( );
		static void 								Update						( );
		static void 								Reset						( );
		static void 								Shutdown					( );

		static void									PlayTrack					( const Asura_Hash_ID uTrackHash, const float fFadeOutTime, const float fFadeInTime );
		static void									Stop						( );
		static bool									IsPlaying					( );

		static void									Pause						( );
		static void									Unpause						( );
		static bool									IsPaused					( );
		
		static void 								HandleMessage				( const Asura_Message* pxMessage );
		
		static char*								GetLayerName				( const int iIndex );
		static void									SetLayerName				( const int iIndex, Asura_String xString );
		
		// Load / Save state functions

		static void									WriteStateToChunkStream		( Asura_Chunk_Stream& xStream );
		static void									ReadStateFromChunkStream	( Asura_Chunk_Stream& xStream );

		static bool									IsRestoringState			( );

		// Track data functions

		static int									GetNumberOfTracks			( );
		static Asura_Dynamic_Music_TrackData*		GetTrackByIndex				( int iIndex );
		static Asura_Dynamic_Music_TrackData*		GetTrackByHash				( Asura_Hash_ID uNameHash );
		static bool									TrackExists					( Asura_Hash_ID uNameHash );
		static Asura_Dynamic_Music_TrackData*		CreateNewTrack				( );
		static void									DeleteTrack					( Asura_Hash_ID uNameHash );
		static void									ClearTracks					( );
		static bool									TrackIsPlaying				( Asura_Hash_ID uNameHash );
		static float								GetTrackPlayTime			( Asura_Hash_ID uNameHash );

		// Trigger functions

		static bool									ActivateTrigger				( Asura_Hash_ID uTriggerHash );
		static bool									IsTriggerActive				( Asura_Hash_ID uTriggerHash );

		// Trigger *data* functions

		static Asura_Dynamic_Music_Trigger_Data*	CreateNewTriggerData		( bool bCodeTrigger = false );
		static int									GetNumberOfTriggerData		( );
		static Asura_Dynamic_Music_Trigger_Data*	GetTriggerDataByHash		( const Asura_Hash_ID uTriggerHash );
		static Asura_Dynamic_Music_Trigger_Data*	GetTriggerDatabyIndex		( const int iIndex );
		static void									DeleteTriggerData			( Asura_Hash_ID uTriggerHash );

		//Other functions

		static float								GetMusicMixerVolume			( )	{ return s_fMusicMixerVolume; }

		#ifdef ASURA_TOOLS
		static void									PlayTrack					( Asura_Hash_ID uNameHash, int iPosition );
		static float								GetTriggerPlayTime			( Asura_Hash_ID uTriggerHash );
		#endif

		#ifdef ASURA_USE_DEBUGOVERLAY
		static void DoDynamicMusicDebugOverlay									( void* const pxIgnore );
		static void DoActiveTriggerDebugOverlay									( const Asura_Dynamic_Music_ActiveTrigger* pxTrigger );
		#endif

	private:

		// ---------------- private enumerations ------------------------

		// version 2 - now saving hash ID instead of GUID in Asura_Dynamic_Music_ActiveTrigger - JCT
		enum { DYNAMIC_MUSIC_LOADSAVE_STATE_VERSION = 2 };

		enum ActiveTracks
		{
			ACTIVE_TRACK_PRIMARY	= 0,
			ACTIVE_TRACK_SECONDARY = 1,

			NUM_ACTIVE_TRACKS
		};

		enum ResumeState
		{
			RESUME_STATE_NONE,
			RESUME_STATE_LOADING,
			RESUME_STATE_PREPARING
		};

		// ---------------- private functions ------------------------
		
		static void		Update_LayerTransTrigger		( const Asura_Dynamic_Music_Layer_Transition_Trigger_Data*	pxData );
		static void		Update_PlayTrigger				( const Asura_Dynamic_Music_Play_Trigger_Data*				pxData );
		static void		Update_StopTrigger				( const Asura_Dynamic_Music_Stop_Trigger_Data*				pxData );
		static void		Update_JoinReleaseTrigger		( const Asura_Dynamic_Music_JoinRelease_Trigger_Data*		pxData );

		static void		FinishedWithCurrentTrigger		();

		// ------------------ private data ------------------------

		static Asura_Dynamic_Music_TrackData*					s_axTracks;
		static int												s_iNumTracks;

		static Asura_String										s_axLayerNames[ ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ];

		static Asura_List<Asura_Dynamic_Music_ActiveTrigger>	s_xActiveTriggerList;
		static Asura_Dynamic_Music_ActiveTrigger*				s_pxCurrentTrigger;

		static Asura_Dynamic_Music_ActiveTrack*					s_pxActiveTracks[ NUM_ACTIVE_TRACKS ];

		static Asura_List<Asura_Dynamic_Music_Trigger_Data>		s_xTriggerDataList;

		static float											s_fMusicMixerVolume;

		static volatile ResumeState								s_eResumeState;

		static bool												s_bPaused;

		static bool												s_bInitialised;
};

// Track data class

class Asura_Dynamic_Music_TrackData
{
	public:

		enum { CURRENT_DATA_VERSION = 1 };

		Asura_Dynamic_Music_TrackData	()	{ SetDefaults(); }
		~Asura_Dynamic_Music_TrackData	()	{}

		Asura_String	m_xName;
		Asura_String	m_axLayerFiles	[ ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS ];
		
		float			m_fBodyOffset;		//Start of the body  in secs
		float			m_fOutroOffset;		//Start of the outro in secs
		
		float			m_fVolume;
		int				m_iBPM;
		int				m_iBeatsPerBar;

		void SetDefaults ()
		{
			m_xName = "";
			for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
			{
				m_axLayerFiles	[i] = "";
			}
			
			m_fBodyOffset = 0.0f;
			m_fOutroOffset = 0.0f;

			m_fVolume		= 1.0f;
			m_iBPM			= 60;
			m_iBeatsPerBar	= 4;
		}

		Asura_Dynamic_Music_TrackData& operator =( const Asura_Dynamic_Music_TrackData& xSource )
		{
			if(&xSource == this)
			{
				return (*this);
			}
			m_xName = xSource.m_xName;
			for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
			{
				m_axLayerFiles[ i ] = xSource.m_axLayerFiles[ i ];
			}
			
			m_fBodyOffset = xSource.m_fBodyOffset;
			m_fOutroOffset = xSource.m_fOutroOffset;

			m_fVolume		= xSource.m_fVolume;
			m_iBPM			= xSource.m_iBPM;
			m_iBeatsPerBar	= xSource.m_iBeatsPerBar;
			return (*this);
		}

		void ReadFromChunkStream ( Asura_Chunk_Stream& xStream )
		{
			int iVersion;
			xStream >> iVersion;
			Asura_Assert(iVersion <= CURRENT_DATA_VERSION, "Unknown version");

			xStream >> m_xName;
			int iNumLayers = 0;
			xStream >> iNumLayers;
			for( int i = 0; i < iNumLayers; ++i )
			{
				if( i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS )
				{
					xStream >> m_axLayerFiles[i];
				}
				else
				{
					//Throw away this extra information if need be
					Asura_String xDummy;
					xStream >> xDummy;
				}
				
			}
			xStream >> m_fBodyOffset;
			xStream >> m_fOutroOffset;

			xStream >> m_fVolume;
			xStream >> m_iBPM;
			xStream >> m_iBeatsPerBar;
		}

		void WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
		{
			xStream << CURRENT_DATA_VERSION;

			xStream << m_xName;
			xStream << ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS;
			for( int i = 0; i < ASURA_DYNAMIC_MUSIC_SYSTEM_NUM_LAYERS; ++i )
			{
				xStream << m_axLayerFiles[i];
			}
			xStream << m_fBodyOffset;
			xStream << m_fOutroOffset;

			xStream << m_fVolume;
			xStream << m_iBPM;
			xStream << m_iBeatsPerBar;
		}

		float GetBarDuration () const
		{
			if( m_iBPM == 0 )
			{
				return 0.0f;
			}
			return ( (60.0f / static_cast<float>( m_iBPM ) ) * static_cast<float>( m_iBeatsPerBar ) );
		}

		float RoundTimeToBars ( float fTime )
		{
			//Useful helper function

			const float fBarDuration = GetBarDuration();

			if(fBarDuration == 0.0f)
			{
				return fTime;
			}

			fTime += fBarDuration * 0.5f;

			const int iNumBars = static_cast<int>(fTime / fBarDuration);
			return ( fBarDuration * static_cast<float>( iNumBars ) );
		}

		float RoundTimeUpToBars ( float fTime )
		{
			//Useful helper function

			const float fBarDuration = GetBarDuration();
			
			if(fBarDuration == 0.0f)
			{
				return fTime;
			}

			const int iNumBars = static_cast<int>(fTime / fBarDuration) + 1;
			return ( fBarDuration * static_cast<float>( iNumBars ) );
		}

		float RoundTimeDownToBars ( float fTime )
		{
			//Useful helper function

			const float fBarDuration = GetBarDuration();

			if(fBarDuration == 0.0f)
			{
				return fTime;
			}

			const int iNumBars = static_cast<int>(fTime / fBarDuration);
			return ( fBarDuration * static_cast<float>( iNumBars ) );
		}
};

#endif //ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM

#endif //__ASURA_DYNAMIC_MUSIC_SYSTEM__

//eof
