//////////////////////////////////////////////////////
//
// FILE : Asura_Engine_Sound.h
//
// Created by Andrew Weinkove - 07/09/05
//
// Class for handling engine sounds.
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
/////////////////////////////////////////////////////


#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_ENGINE_SOUND_H__
#define __ASURA_ENGINE_SOUND_H__

#ifndef ASURA_USE_AUDIO_SYSTEM

// Includes
#include "Core/Asura_Handle.h"
#include "Sound/Asura_ProjectSoundIDs.h"

// Forward Declarations
class Asura_Vector_3;

//
// (AW - 08/09/05) - The purpose of this class is to enable smooth blending between 
// multiple pitch-shifted looping sounds.
//
// The simple way to model sound depending on engine revs is to take a looping sound
// and pitch-shift it. However, pitch-shifting a sound over too large a range can sound
// wrong - also as the rpms of an engine change, more than just the pitch of the sound changes.
//
// This class allows a solution where you have a number of different engine sounds (perhaps
// idling, low rpm, med rpm and high rpm). At any one time a maximum of 2 of the loops 
// will play (with volumes balanced appropriately) both of which will be pitch shifted in
// an attempt to sound the same.
//
// This code can be (ab)used for situations other than engine sounds. For example
// in MV we use it to blend between our stationary/slow/fast boat moving sounds (and don't do
// any pitch-shifting). 
//
// I don't recommend using this code to pitch-shift a single sound - it is overkill
// to use this class. Also, in determining the sound(s) to play there is a sequential search,
// you should consider optimising this if you're going to start using 10+ different sounds in your 
// range.
//


// This structure represents one of the individual sounds that has been recorded.
class Asura_Engine_Sound_Point
{
public:
	Asura_Sound_ID m_eSoundID;		// The sound ID.
	float m_fCorrectRevs;			// The correct revs for this sound (where it will play at volume = 1, pitch = 1)
	float m_fPitchChange;			// How much to change the pitch : pitch = 1 + ((fRevs - m_fCorrectRevs) * m_fPitchChange)
	float m_fCrossfadeStart;		// How many revs before this sound starts to cross-fade into the next one (counted from when this sound is fully on)
	float m_fCrossfadeLength;		// How many revs the cross-fade lasts.
};

class Asura_Engine_Sound
{
public:
	Asura_Engine_Sound(const Asura_Engine_Sound_Point* pxSoundPointArray, u_int uNumberSoundPoints, float fSoundInnerRadius, float fSoundOuterRadius);
	~Asura_Engine_Sound();

	void Update(const Asura_Vector_3& xPosition, float fRevs, float fVolumeScale = 1.0f);
private:
	void CreateSingleSound(u_int uIndex, const Asura_Vector_3& xPosition);
	void CreateDoubleSound(u_int uLowerIndex, const Asura_Vector_3& xPosition);
	void SetVolumeBlend(float fBlend, float fVolumeScale);
	void ReleaseResources();

	const Asura_Engine_Sound_Point* m_pxSoundPointArray;  // Array of sound points.
	u_int m_uNumberSoundPoints;						// Number of sound points in our array.
	Asura_Handle m_iLowSoundHandle;					// Sound handle of current playing lower sound point
	Asura_Handle m_iHighSoundHandle;				// Sound handle of current playing higher sound point
	float m_fSoundInnerRadius;
	float m_fSoundOuterRadius;
};	

#endif //__ASURA_ENGINE_SOUND_H__

#endif // #ifndef ASURA_USE_AUDIO_SYSTEM

//eof
