///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_MetaMusic_System.h
//
// Created by Richard Rice - 16/01/04
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_METAMUSIC_SYSTEM_H__
#define __ASURA_METAMUSIC_SYSTEM_H__

#if !defined(ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM)

// Includes
#include "Core/Asura_Hash.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_String.h"

// Forward declarations
class Asura_Chunk_MetaMusic;
class Asura_Chunk_Stream;
class Asura_MetaMusic;
class Asura_Zone_Controller;
class Tools_Mode_MetaMusic;
struct Asura_MetaMusic_File;
class Asura_Message;
class Asura_Stream;

//////////////////////////
// Asura_MetaMusicState //
//////////////////////////
enum Asura_MetaMusicState
{
	ASURA_METAMUSIC_NOSTATE = -1,
	ASURA_METAMUSIC_INTRO = 0,
	ASURA_METAMUSIC_AMBIENTOUTRO,
	ASURA_METAMUSIC_TENSIONOUTRO,
	ASURA_METAMUSIC_ACTIONOUTRO,
	ASURA_METAMUSIC_AMBIENT,
	ASURA_METAMUSIC_AMBIENT_TO_TENSION,
	ASURA_METAMUSIC_AMBIENT_TO_ACTION,
	ASURA_METAMUSIC_TENSION,
	ASURA_METAMUSIC_TENSION_TO_AMBIENT,
	ASURA_METAMUSIC_TENSION_TO_ACTION,
	ASURA_METAMUSIC_ACTION,
	ASURA_METAMUSIC_ACTION_TO_AMBIENT,
	ASURA_METAMUSIC_ACTION_TO_TENSION,
	ASURA_METAMUSIC_NUMSTATES,

	ASURA_METAMUSIC_FORCETOINT = 0x7fffffff
};

#define ASURA_METAMUSIC_MAXQUEUE 8

////////////////////////////////
// Asura_MetaMusic_TrackIndex //
////////////////////////////////
// MGH 14/6/2006: For non-streaming builds the track index is just an integer.
// However for streaming builds we also need the zone of the relevant handler.
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
class Asura_MetaMusic_TrackIndex
{
public:
	Asura_MetaMusic_TrackIndex(int iTrack, Asura_Hash_ID uHandlerHash)
		: m_iTrack(iTrack)
		, m_uHandlerHash(uHandlerHash)
	{}

	Asura_MetaMusic_TrackIndex() { Reset(); }

	Asura_MetaMusic_TrackIndex(const Asura_MetaMusic_TrackIndex& xOther)
		: m_iTrack(xOther.m_iTrack)
		, m_uHandlerHash(xOther.m_uHandlerHash)
	{}

	void Reset()
	{
		m_iTrack = -1;
		m_uHandlerHash = ASURA_HASH_ID_UNSET;
	}

	const int GetTrack() const { return m_iTrack; }
	const Asura_Hash_ID GetHandlerHash() const { return m_uHandlerHash; }

	const bool IsNull() const
	{
		return (m_iTrack<0 || m_uHandlerHash == ASURA_HASH_ID_UNSET);
	}

	bool operator==(const Asura_MetaMusic_TrackIndex& xOther) const
	{
		return (m_iTrack == xOther.m_iTrack) && (m_uHandlerHash == xOther.m_uHandlerHash);
	}

	bool operator!=(const Asura_MetaMusic_TrackIndex& xOther) const
	{
		return !(*this == xOther);
	}

private:
	int m_iTrack;
	Asura_Hash_ID m_uHandlerHash;
};
#else
class Asura_MetaMusic_TrackIndex
{
public:
	Asura_MetaMusic_TrackIndex(int iTrack, Asura_Hash_ID uHandlerHash) : m_iTrack(iTrack) {}
	Asura_MetaMusic_TrackIndex() { Reset(); }
	Asura_MetaMusic_TrackIndex(const Asura_MetaMusic_TrackIndex& xOther) : m_iTrack(xOther.m_iTrack) {}

	void Reset() { m_iTrack = -1; }

	int GetTrack() const { return m_iTrack; }
	Asura_Hash_ID GetHandlerHash() const { return ASURA_HASH_ID_UNSET; }
	bool IsNull() const { return (m_iTrack<0); }

	bool operator==(const Asura_MetaMusic_TrackIndex& xOther) const
	{
		return (m_iTrack == xOther.m_iTrack);
	}

	bool operator!=(const Asura_MetaMusic_TrackIndex& xOther) const
	{
		return !(*this == xOther);
	}

	operator int() const { return m_iTrack; }

private:
	int m_iTrack;
};
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)


////////////////////////////////
// Asura_MetaMusic_QueueEntry //
////////////////////////////////
struct Asura_MetaMusic_QueueEntry
{
	Asura_MetaMusic_TrackIndex m_xTrackIndex;
	Asura_MetaMusicState m_eState;
	bool m_bImmediate;
};


/////////////////////////////////////
// Asura_MetaMusic_StateDefinition //
/////////////////////////////////////
struct Asura_MetaMusic_StateDefinition
{
	//index of first file that belongs to this state
	int m_iFirstIndex;
	//number of music files that belong to this state
	int m_iNumberOfFiles;
	//Fade in time when going to this state from a looping state
	//For transition states this is also used as the crossfade time if no track exists
	float m_fFadeInTime;
	//Fade out time when going from this state (not applicable to looping states)
	float m_fFadeOutTime;
};

#ifdef ASURA_TOOLS
#define ASURA_METAMUSIC_PLAYAMBIENTTOENDOFLOOP		0x00000001
#define ASURA_METAMUSIC_PLAYTENSIONTOENDOFLOOP		0x00000002
#define ASURA_METAMUSIC_PLAYACTIONTOENDOFLOOP		0x00000004
#define ASURA_METAMUSIC_CHOOSENEWFILEAFTEREACHLOOP	0x00000008
#endif //ASURA_TOOLS

/////////////////////
// Asura_MetaMusic //
/////////////////////
class Asura_MetaMusic
{
public :
	~Asura_MetaMusic();

#ifdef ASURA_TOOLS
	void Initialise(const char* pszName);
#endif

	//indicates whether there are any pieces for this state
	bool HasState(Asura_MetaMusicState eState) const;
	//get the name of a sound file in the stae
	const Asura_MetaMusic_File* GetFileInState(Asura_MetaMusicState eState) const;
	//Gets the intervals at which transitions from the pieces in this meta music track can be done
	float GetBeatLoopTime(Asura_MetaMusicState eCurrentState) const;

	//Disabled because of issues implementing on XBox. RAR 06/02/2005
	bool ShouldLoopingStatePlayToEndOfLoop(Asura_MetaMusicState eState) const {return false;}

	//Determine whether a state should be played as looping
	bool ShouldStateBeLooped(Asura_MetaMusicState eState) const;

	Asura_String m_xName;
	Asura_Hash_ID m_uNameHash;

	//Time in seconds at which to end piece if no messages are recieved
	float m_fDuration;

	//These two are possibly going to be used to determine a good point at which to start a transition
	float m_fAmbientBeatsPerMinute;
	float m_fAmbientLoopBeats;
	float m_fTensionBeatsPerMinute;
	float m_fTensionLoopBeats;
	float m_fActionBeatsPerMinute;
	float m_fActionLoopBeats;

	//volume multiplier for this piece of meta music as a whole
	float m_fVolume;

	Asura_MetaMusic_StateDefinition m_axStates[ASURA_METAMUSIC_NUMSTATES];

	//Number of music files that belong to this track
	int m_iNumberOfFiles;
	Asura_MetaMusic_File* m_pxFiles;

	//determine whether intro counts as ambient,tension or action
	Asura_MetaMusicState m_eIntroLoopState;

	union
	{
		u_int m_uFlags;
		// NOTE: These bit's are written out, adding any new bits means
		// ~~~~~ that you will need to add those bits to the bitfield
		//       preparation code in "WriteToChunkStream".
		struct
		{
			u_int	m_bPlayAmbientToEndOfLoop : 1,
					m_bPlayTensionToEndOfLoop : 1,
					m_bPlayActionToEndOfLoop : 1,
						   //Causes the looping states to be played as non-looping , with a new
						   //piece being chosen after each loop
					m_bChooseNewFileAfterEachLoop : 1;
		};
	};

};


/////////////////////////////
// Asura_MetaMusic_Handler //
/////////////////////////////
class Asura_MetaMusic_Handler
{
public:
	// Public functions...
	Asura_MetaMusic_Handler();
	~Asura_MetaMusic_Handler();

	// Initialisation/reset functions...
	Asura_MetaMusic* Initialise(const int iNumberOfTracks);
	void Reset();

	// Get zone hash of the handler, for streaming builds.
	Asura_Hash_ID GetZoneHashID() const { return m_uZoneHashID; }
	void SetZoneHashID(Asura_Hash_ID uZoneHashID) { m_uZoneHashID = uZoneHashID; }

#if defined(ASURA_TOOLS)
	void SetTracks(Asura_MetaMusic* pxTracks, const int iNumberOfTracks);
	Asura_MetaMusic* GetTrackData(int iIndex) { return &m_pxMusicTracks[iIndex]; }
#endif	// defined(ASURA_TOOLS)

	// Read/write functions...
	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);

	//gets track index number, or -1 if track not found
	Asura_MetaMusic_TrackIndex GetTrackIndexFromHash(Asura_Hash_ID uHash) const;
	
#if defined(ASURA_TOOLS)
	const char*   GetTrackName(int iIndex) const;
	Asura_Hash_ID GetTrackHash(int iIndex) const;
#endif //defined(ASURA_TOOLS)

	const char*   GetTrackName(const Asura_MetaMusic_TrackIndex& xIndex) const;
	Asura_Hash_ID GetTrackHash(const Asura_MetaMusic_TrackIndex& xIndex) const;

	int GetNumberOfTracks() const { return m_iNumberOfMusicTracks; }

	const Asura_MetaMusic* GetTrack(const Asura_MetaMusic_TrackIndex& xIndex) const;

#if defined(ASURA_USE_METAMUSICZONES)
	Asura_Zone_Controller* GetZoneController() const { return m_pxZoneController; }
#endif	// defined(ASURA_USE_METAMUSICZONES)


protected:
	// Protected variables...
	int m_iNumberOfMusicTracks;
	Asura_MetaMusic* m_pxMusicTracks;

#if defined(ASURA_USE_METAMUSICZONES)
	Asura_Zone_Controller* m_pxZoneController;
#endif	// defined(ASURA_USE_METAMUSICZONES)

	Asura_Hash_ID m_uZoneHashID;
};

////////////////////////////
// Asura_MetaMusic_System //
////////////////////////////
class Asura_MetaMusic_System
{
	friend class Asura_Chunk_MetaMusic;
	friend class Asura_MetaMusic_Handler;
	friend class Tools_Mode_MetaMusic;

public :
	static void Initialise();
	static void Shutdown();
	static void Reset();
	//Pauses the sound tracks being played , and prevents any updates
	//until unpaused as many times as has been paused
	static void Pause(bool bImmediate = true);
	//Stops the sound being played (using fade if not bImmediate) allowing the stream handles
	//to be used elsewhere. When unpaused it will start up again from the next intro/looping
	//state in the queue.
	static void PauseAndReleaseSound(bool bImmediate);
	static void Unpause();

	static void ClearQueue(int iFrom = 0);
	static void Update();

	static void PlayTrack(Asura_Hash_ID uHash, bool bHighPriority);
	static void PlayTrackByIndex(const Asura_MetaMusic_TrackIndex& xTrackIndex, bool bHighPriority);
	static void StopTrack(bool bHighPriority);

	static void ChangeToAmbient() { ChangeState(ASURA_METAMUSIC_AMBIENT); }
	static void ChangeToTension() { ChangeState(ASURA_METAMUSIC_TENSION); }
	static void ChangeToAction()  { ChangeState(ASURA_METAMUSIC_ACTION);  }

	static void ChangeState(Asura_MetaMusicState eState);

	static void RestartLastTrack();

#if defined(ASURA_TOOLS)
	static void SetTracks(Asura_MetaMusic* pxTracks, const int iNumberOfTracks)	{ GetCurrentHandler().SetTracks(pxTracks, iNumberOfTracks); }
#endif	// defined(ASURA_TOOLS)

	// Gets track index, or -1 (ie null track index) if track not found.
	static Asura_MetaMusic_TrackIndex GetTrackIndexFromHash(Asura_Hash_ID uHash);

	static int GetNumberOfTracks() { return GetCurrentHandler().GetNumberOfTracks(); }

#if defined(ASURA_TOOLS)
	static const char* GetTrackName(int iIndex)				{ return GetCurrentHandler().GetTrackName(iIndex); }
	static Asura_Hash_ID GetTrackHash(int iIndex)			{ return GetCurrentHandler().GetTrackHash(iIndex); }
#endif //defined(ASURA_TOOLS)

	static const char* GetTrackName(const Asura_MetaMusic_TrackIndex& xIndex)
	{
		const Asura_MetaMusic_Handler* const pxHandler = GetHandlerForTrack(xIndex);
		return pxHandler ? pxHandler->GetTrackName(xIndex) : 0;
	}

	static Asura_Hash_ID GetTrackHash(const Asura_MetaMusic_TrackIndex& xIndex)
	{
		const Asura_MetaMusic_Handler* const pxHandler = GetHandlerForTrack(xIndex);
		return pxHandler ? pxHandler->GetTrackHash(xIndex) : ASURA_HASH_ID_UNSET;
	}

	static void HandleMessage(const Asura_Message* pxMessage);

	static bool IsLoopingState(Asura_MetaMusicState eState);
	static bool IsOutroState(Asura_MetaMusicState eState);
	static bool IsTransitionState(Asura_MetaMusicState eState);

	// Gets the beat number in the loop for the playing track. Useful mostly for debugging.
	static float GetBeatTimer();

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// Handler functions...
	static	Asura_MetaMusic_Handler*	CreateHandler		(const Asura_Hash_ID uHash);
	static	bool						DestroyHandler		(const Asura_Hash_ID uHash);
	static	void						WriteHandlerReport	(u_int uHandler, Asura_Stream& xLog);

	static	void						CreateAndSetHandler	(const Asura_Hash_ID uHash)				{ s_pxCurrentHandler = CreateHandler(uHash); }
	static	void						SetCurrentHandler	(Asura_MetaMusic_Handler* pxHandler)	{ s_pxCurrentHandler = pxHandler; }
	static	void						SetCurrentHandler	(const Asura_Hash_ID uHash)				{ s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }

	static	Asura_MetaMusic_Handler&	GetCurrentHandler()	{ return *s_pxCurrentHandler; }
    
    static  Asura_MetaMusic_Handler*	GetHandler (const Asura_Hash_ID uHandlerHash) { return(s_xHandlerTree.GetDataByHash(uHandlerHash)); }
#else
	// Handler functions...
	static	Asura_MetaMusic_Handler&	GetCurrentHandler()	{ return s_xHandler; }
#endif


protected :
#ifdef ASURA_USE_METAMUSICZONES
	static int Project_GetNumberOfAttackingNPCs();

	//per-frame update of music zones
	static void UpdateMusicZones(Asura_MetaMusic_Handler& xHandler);

	//storing/restoring state for when the music zones take/release control
	static void MusicZone_StoreState();
	static void MusicZone_RestoreState();
#endif

#ifdef ASURA_USE_PROJECTMETAMUSIC
	//returns true if the project has dealt with the message
	static bool Project_HandleMessage(const Asura_Message* pxMessage);
	//project initilise called from the main initialise function
	static void Project_Initialise();
	//project updating called from the main update function
	static void Project_Update();
	//project shutdown called from the main shutdown function
	static void Project_Shutdown();
#endif

	static void UpdateForHandler(Asura_Hash_ID uHash, Asura_MetaMusic_Handler* pxHandler);

	static void PlayNextInQueue();
	static void AddToQueue(const Asura_MetaMusic_TrackIndex& xTrackIndex, Asura_MetaMusicState eState, bool bImmediate);
	static void AddToStartOfQueue(const Asura_MetaMusic_TrackIndex& xTrackIndex, Asura_MetaMusicState eState, bool bImmediate);
	static Asura_MetaMusic_QueueEntry* GetFromQueue(int iIndex)
	{
		return &s_axQueue[(s_iQueuePosition+iIndex)%ASURA_METAMUSIC_MAXQUEUE];
	}

	// this takes the current track+state and the queue and calculates what track+state the meta music
	// system should be using once the queue has been acted upon
	static void CalculateFinalDummyState( Asura_Hash_ID& uLastTrackHashID,
										  Asura_Hash_ID& uTrackHashID,
										  Asura_MetaMusicState& eState );

	// this updates this dummy state to what it would be if the message were acted upon
	static void UpdateDummyState( Asura_Hash_ID& uLastTrackHashID,
								  Asura_Hash_ID& uTrackHashID,
								  Asura_MetaMusicState& eState,
								  const Asura_Message* pxMessage );

	// sets the system back to a state using the supplied variables
	static void RestoreSystemState( const Asura_Hash_ID uLastTrackHashID,
									const Asura_Hash_ID uTrackHashID,
									const Asura_MetaMusicState eState );

	//Get the outro state to use from the current level
	static Asura_MetaMusicState GetOutroState(const Asura_MetaMusic_TrackIndex& xTrack, Asura_MetaMusicState eCurrentState);

	//Get the transition state between 2 looping states
	static Asura_MetaMusicState GetTransitionState(Asura_MetaMusicState eCurrentState, Asura_MetaMusicState eNextState);

	//Get the loop state that the intro is counted as
	static Asura_MetaMusicState GetIntroLoopState(const Asura_MetaMusic_TrackIndex& xTrack);
	static Asura_MetaMusicState GetIntroLoopStateFromHash(Asura_Hash_ID uTrackHash);

	//Determine whether a state should be played as looping for a given track
	static bool ShouldStateBeLooped(const Asura_MetaMusic_TrackIndex& xTrack, Asura_MetaMusicState eState);

	//Get the fade time required for a state change
	static float GetFadeTime(const Asura_MetaMusic_TrackIndex& xCurrentTrack, Asura_MetaMusicState eCurrentState, const Asura_MetaMusic_TrackIndex& xNextTrack, Asura_MetaMusicState eNextState, bool bImmediate);

	//Checks to see if there is enough time left in current track to do desired fade
	static bool IsTimeLeftForFade(const Asura_MetaMusic_TrackIndex& xTrack, Asura_MetaMusicState eState, bool bImmediate);

	//Searches the queue for the next entry that will actually get played (that is,
	//it has some files associated with it)
	static const Asura_MetaMusic_QueueEntry* GetNextInQueueWithFiles();

	static bool TrackHasState(const Asura_MetaMusic_TrackIndex& xTrack, Asura_MetaMusicState eState);

	static void ClearBestGuessFile ();

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	static bool GetTrackIndexFromHash_Callback(Asura_Hash_ID uHash, const Asura_MetaMusic_Handler* pxHandler, void* pUserData);

	// Protected variables...
	static	Asura_MetaMusic_Handler* s_pxCurrentHandler;	// Currently active handler.
	static	Asura_TemplateHashTree<Asura_MetaMusic_Handler>	s_xHandlerTree;	// Tree of all of the handlers available.
#else
	// Protected variables...
	static	Asura_MetaMusic_Handler	s_xHandler;	// Single static-instance of a zone.
#endif	//defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)

	static Asura_MetaMusic_TrackIndex s_xCurrentTrack; // Current meta music track being played.
	static Asura_MetaMusic_TrackIndex s_xLastTrackRequested;

	static Asura_MetaMusicState s_eCurrentState; // current state of track being played
	static Asura_Hash_ID s_uCurrentFilenameHash; //hashed filename of file being played by music system

	static float s_fIdleTimer; //How long current track has being playing without changes in state
	static float s_fTransitionPosition;//Position in current music track at which transition to next queue entry should start
	static float s_fPreviousPosition;

	static int s_iQueuePosition;
	static int s_iQueueLength;	
	static Asura_MetaMusic_QueueEntry s_axQueue[ASURA_METAMUSIC_MAXQUEUE];
	static int s_iPauseCount;
	static bool s_bPausedMusicSystem;//indicates whether the music system was paused or stopped when meta music was paused

	static const Asura_MetaMusic_File* s_pxBestGuessFile;
	static Asura_MetaMusicState s_xBestGuessState;

#ifdef ASURA_USE_METAMUSICZONES
	static u_int s_uZone_TriggerLevel;
	static u_int s_uZone_CurrentID;
	static u_int s_uZone_CurrentHandler;
	static float s_fZone_Timer;

	//for storing the state from messages while music is under control of the zones
	static bool s_bZone_HasStoredState;
	static u_int s_uZone_StoredLastTrackHash;
	static u_int s_uZone_StoredTrackHash;
	static Asura_MetaMusicState s_eZone_StoredTrackState;
#endif //ASURA_USE_METAMUSICZONES

	// Protected functions...
#ifdef ASURA_TOOLS
	// Direct access to the array for the current handler.
	static Asura_MetaMusic* GetTrack(const int iIndex) { return GetCurrentHandler().GetTrackData(iIndex); }
#endif //ASURA_TOOLS

	static const Asura_MetaMusic* GetTrack(const Asura_MetaMusic_TrackIndex& xIndex)
	{
		const Asura_MetaMusic_Handler* const pxHandler = GetHandlerForTrack(xIndex);
		return pxHandler ? pxHandler->GetTrack(xIndex) : 0;
	}

	static const Asura_MetaMusic* GetCurrentTrack() { return GetTrack(s_xCurrentTrack); }

	static Asura_MetaMusic_Handler* GetHandlerForTrack(const Asura_MetaMusic_TrackIndex& xIndex);

#if defined(ASURA_USE_METAMUSICZONES)
	static	Asura_Zone_Controller* GetZoneController() { return GetCurrentHandler().GetZoneController(); }
#endif	// defined(ASURA_USE_METAMUSICZONES)

	#if defined( ASURA_USE_DEBUGOVERLAY )
	static void DoMetaMusicDebugOverlay( void* const pxIgnore );
	#endif
};

//////////////////////////
// Asura_MetaMusic_File //
//////////////////////////
struct Asura_MetaMusic_File
{
	Asura_String m_xFilename;
	float m_fVolume;//volume multiplier to use for this file
};

//////////////
// Messages //
//////////////

//message sent by music trigger to start track.
class Asura_MsgData_MetaMusic_Play
{
public :
	Asura_Hash_ID m_uTrackID;
	unsigned char m_ucHighPriority;
};

//message sent by music trigger to change state.
class Asura_MsgData_MetaMusic_ChangeState
{
public :
	unsigned char m_ucState;
	unsigned char m_ucHighPriority;
};

// message sent by music trigger to restart current track
class Asura_MsgData_MetaMusic_Restart
{
public:
	unsigned char m_ucHighPriority;
};


#ifdef ASURA_USE_METAMUSICZONES
//message data used by the activate/deactivate zone messages
//(this one gets saved out to static messages)
struct Asura_MsgData_MetaMusic_Zone
{
	u_int m_uZoneID;
};
#endif //ASURA_USE_METAMUSICZONES

#endif	//!defined(ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM)

#endif //__ASURA_METAMUSIC_SYSTEM_H__
