//
// Asura_OpenAL.cpp
//
// Created by Andy Patrick - 07-08-2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Asura_OpenAL.h"
#if defined( ASURA_USE_OPENAL ) && ( ASURA_USE_OPENAL == 1 ) && !defined(ASURA_USE_AUDIO_SYSTEM)
#include "OpenAL/OpenAL1-1SDK/include/efx-creative.h"
#include "OpenAL/OpenAL1-1SDK/include/efx-util.h"
#include "Core/Asura_Timers.h"

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
#include "Camera/Asura_Camera.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Text_Output.h"

static float g_fDistanceFilterScale			= 1.0f;
static float g_fOcclusionFilterScale		= 1.0f;
#define CONSOLE_VARNAME_AUDIO_DISTANCE_SCALE	L"Audio.Distance.LowPassFilterScale"
#define CONSOLE_VARNAME_AUDIO_OCCLUSION_SCALE	L"Audio.Occlusion.LowPassFilterScale"

#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

//Force win32 to use the generic software device (will otherwise not use windows mixer settings)
#if defined(WIN32) && !defined(ASURA_OPENAL_FORCE_SOFTWARE_DEVICE)
#define ASURA_OPENAL_FORCE_SOFTWARE_DEVICE
#endif

/* ------------------------------------------------------------------------- */

#define ENABLE_AUXILIARY_SEND_FILTER	(1)	// XXX this causes leaks in OpenAL
#define ENABLE_PER_VOICE_FILTER			(1)

/* ------------------------------------------------------------------------- */

// Cached OpenAL EFX functions, to avoid an expensive look-up every time they are called.

LPALGENEFFECTS					alGenEffects = NULL;
LPALDELETEEFFECTS				alDeleteEffects = NULL;
LPALISEFFECT					alIsEffect = NULL;
LPALEFFECTI						alEffecti = NULL;
LPALEFFECTIV					alEffectiv = NULL;
LPALEFFECTF						alEffectf = NULL;
LPALEFFECTFV					alEffectfv = NULL;
LPALGETEFFECTI					alGetEffecti = NULL;
LPALGETEFFECTIV					alGetEffectiv = NULL;
LPALGETEFFECTF					alGetEffectf = NULL;
LPALGETEFFECTFV					alGetEffectfv = NULL;

LPALGENFILTERS					alGenFilters = NULL;
LPALDELETEFILTERS				alDeleteFilters = NULL;
LPALISFILTER					alIsFilter = NULL;
LPALFILTERI						alFilteri = NULL;
LPALFILTERIV					alFilteriv = NULL;
LPALFILTERF						alFilterf = NULL;
LPALFILTERFV					alFilterfv = NULL;
LPALGETFILTERI					alGetFilteri = NULL;
LPALGETFILTERIV					alGetFilteriv = NULL;
LPALGETFILTERF					alGetFilterf = NULL;
LPALGETFILTERFV					alGetFilterfv = NULL;

LPALGENAUXILIARYEFFECTSLOTS		alGenAuxiliaryEffectSlots = NULL;
LPALDELETEAUXILIARYEFFECTSLOTS	alDeleteAuxiliaryEffectSlots = NULL;
LPALISAUXILIARYEFFECTSLOT		alIsAuxiliaryEffectSlot = NULL;
LPALAUXILIARYEFFECTSLOTI		alAuxiliaryEffectSloti = NULL;
LPALAUXILIARYEFFECTSLOTIV		alAuxiliaryEffectSlotiv = NULL;
LPALAUXILIARYEFFECTSLOTF		alAuxiliaryEffectSlotf = NULL;
LPALAUXILIARYEFFECTSLOTFV		alAuxiliaryEffectSlotfv = NULL;
LPALGETAUXILIARYEFFECTSLOTI		alGetAuxiliaryEffectSloti = NULL;
LPALGETAUXILIARYEFFECTSLOTIV	alGetAuxiliaryEffectSlotiv = NULL;
LPALGETAUXILIARYEFFECTSLOTF		alGetAuxiliaryEffectSlotf = NULL;
LPALGETAUXILIARYEFFECTSLOTFV	alGetAuxiliaryEffectSlotfv = NULL;

/* ------------------------------------------------------------------------- */

struct Asura_Sound_ReverbPreset
{
	Asura_UnicodeChar*			m_szName;
	EFXEAXREVERBPROPERTIES		m_xReverbProperties;
};

// This table was automatically generated by calling the ConvertReverbParameters
// function on all the reverb presets in EAX-util.h in turn. This saves calling
// that function at runtime, and also means the presets can be used on platforms
// where that function is not available eg. Xbox 360. The indices into the table
// are the enums in Asura_Sound_ReverbID.
// TODO this will need to be in Asura_Sound to be shared with non-OpenAL implementations.
static const Asura_Sound_ReverbPreset s_axReverbPresets[] =
{
	// NOTE how "NONE" has gain set to 0.0, this silences the reverb and makes the rest of the values for that preset irrelevant
	{ L"NONE",						{ 1.000000f, 1.000000f, 0.000000f, 1.000000f, 1.000000f, 1.000000f, 0.100000f, 1.000000f, 1.000000f, 0.001000f, { 0.000000f, 0.000000f, 0.000000f }, 1.000000f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 0.100000f, 0.000000f, 0.100000f, 0.000000f, 1.000000f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"GENERIC",					{ 1.000000f, 1.000000f, 0.316228f, 0.891251f, 1.000000f, 1.490000f, 0.830000f, 1.000000f, 0.050003f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"PADDEDCELL",				{ 0.171500f, 1.000000f, 0.316228f, 0.001000f, 1.000000f, 0.170000f, 0.100000f, 1.000000f, 0.250035f, 0.001000f, { 0.000000f, 0.000000f, 0.000000f }, 1.269112f, 0.002000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"ROOM",						{ 0.428687f, 1.000000f, 0.316228f, 0.592925f, 1.000000f, 0.400000f, 0.830000f, 1.000000f, 0.150314f, 0.002000f, { 0.000000f, 0.000000f, 0.000000f }, 1.062919f, 0.003000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"BATHROOM",					{ 0.171500f, 1.000000f, 0.316228f, 0.251189f, 1.000000f, 1.490000f, 0.540000f, 1.000000f, 0.653131f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 3.273407f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"LIVINGROOM",				{ 0.976563f, 1.000000f, 0.316228f, 0.001000f, 1.000000f, 0.500000f, 0.100000f, 1.000000f, 0.205116f, 0.003000f, { 0.000000f, 0.000000f, 0.000000f }, 0.280543f, 0.004000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"STONEROOM",					{ 1.000000f, 1.000000f, 0.316228f, 0.707946f, 1.000000f, 2.310000f, 0.640000f, 1.000000f, 0.441062f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.100272f, 0.017000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"AUDITORIUM",				{ 1.000000f, 1.000000f, 0.316228f, 0.578096f, 1.000000f, 4.320000f, 0.590000f, 1.000000f, 0.403181f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.716968f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CONCERTHALL",				{ 1.000000f, 1.000000f, 0.316228f, 0.562341f, 1.000000f, 3.920000f, 0.700000f, 1.000000f, 0.242661f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.997700f, 0.029000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CAVE",						{ 1.000000f, 1.000000f, 0.316228f, 1.000000f, 1.000000f, 2.910000f, 1.300000f, 1.000000f, 0.500035f, 0.015000f, { 0.000000f, 0.000000f, 0.000000f }, 0.706318f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"ARENA",						{ 1.000000f, 1.000000f, 0.316228f, 0.447713f, 1.000000f, 7.240000f, 0.330000f, 1.000000f, 0.261216f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 1.018591f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"HANGAR",					{ 1.000000f, 1.000000f, 0.316228f, 0.316228f, 1.000000f, 10.050000f, 0.230000f, 1.000000f, 0.500035f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 1.256030f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CARPETTEDHALLWAY",			{ 0.428687f, 1.000000f, 0.316228f, 0.010000f, 1.000000f, 0.300000f, 0.100000f, 1.000000f, 0.121479f, 0.002000f, { 0.000000f, 0.000000f, 0.000000f }, 0.153109f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"HALLWAY",					{ 0.364500f, 1.000000f, 0.316228f, 0.707946f, 1.000000f, 1.490000f, 0.590000f, 1.000000f, 0.245754f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.661499f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"STONECORRIDOR",				{ 1.000000f, 1.000000f, 0.316228f, 0.761202f, 1.000000f, 2.700000f, 0.790000f, 1.000000f, 0.247172f, 0.013000f, { 0.000000f, 0.000000f, 0.000000f }, 1.575796f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"ALLEY",						{ 1.000000f, 0.300000f, 0.316228f, 0.732825f, 1.000000f, 1.490000f, 0.860000f, 1.000000f, 0.250035f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 0.995405f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.125000f, 0.950000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"FOREST",					{ 1.000000f, 0.300000f, 0.316228f, 0.022387f, 1.000000f, 1.490000f, 0.540000f, 1.000000f, 0.052481f, 0.162000f, { 0.000000f, 0.000000f, 0.000000f }, 0.768245f, 0.088000f, { 0.000000f, 0.000000f, 0.000000f }, 0.125000f, 1.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CITY",						{ 1.000000f, 0.500000f, 0.316228f, 0.398107f, 1.000000f, 1.490000f, 0.670000f, 1.000000f, 0.073030f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 0.142725f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"MOUNTAINS",					{ 1.000000f, 0.270000f, 0.316228f, 0.056234f, 1.000000f, 1.490000f, 0.210000f, 1.000000f, 0.040738f, 0.300000f, { 0.000000f, 0.000000f, 0.000000f }, 0.191867f, 0.100000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 1.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"QUARRY",					{ 1.000000f, 1.000000f, 0.316228f, 0.316228f, 1.000000f, 1.490000f, 0.830000f, 1.000000f, 0.000000f, 0.061000f, { 0.000000f, 0.000000f, 0.000000f }, 1.778279f, 0.025000f, { 0.000000f, 0.000000f, 0.000000f }, 0.125000f, 0.700000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"PLAIN",						{ 1.000000f, 0.210000f, 0.316228f, 0.100000f, 1.000000f, 1.490000f, 0.500000f, 1.000000f, 0.058479f, 0.179000f, { 0.000000f, 0.000000f, 0.000000f }, 0.108893f, 0.100000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 1.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"PARKINGLOT",				{ 1.000000f, 1.000000f, 0.316228f, 1.000000f, 1.000000f, 1.650000f, 1.500000f, 1.000000f, 0.208209f, 0.008000f, { 0.000000f, 0.000000f, 0.000000f }, 0.265155f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"SEWERPIPE",					{ 0.307063f, 0.800000f, 0.316228f, 0.316228f, 1.000000f, 2.810000f, 0.140000f, 1.000000f, 1.638702f, 0.014000f, { 0.000000f, 0.000000f, 0.000000f }, 3.247133f, 0.021000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"UNDERWATER",				{ 0.364500f, 1.000000f, 0.316228f, 0.010000f, 1.000000f, 1.490000f, 0.100000f, 1.000000f, 0.596348f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 7.079458f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 1.180000f, 0.348000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"DRUGGED",					{ 0.428687f, 0.500000f, 0.316228f, 1.000000f, 1.000000f, 8.390000f, 1.390000f, 1.000000f, 0.875992f, 0.002000f, { 0.000000f, 0.000000f, 0.000000f }, 3.108136f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 1.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"DIZZY",						{ 0.364500f, 0.600000f, 0.316228f, 0.630957f, 1.000000f, 17.230000f, 0.560000f, 1.000000f, 0.139155f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.493742f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 1.000000f, 0.810000f, 0.310000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"PSYCHOTIC",					{ 0.062500f, 0.500000f, 0.316228f, 0.840427f, 1.000000f, 7.560000f, 0.910000f, 1.000000f, 0.486407f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 2.437811f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 4.000000f, 1.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"CASTLE_SMALLROOM",			{ 1.000000f, 0.890000f, 0.316228f, 0.398107f, 0.100000f, 1.220000f, 0.830000f, 0.310000f, 0.891251f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 1.995262f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.138000f, 0.080000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_SHORTPASSAGE",		{ 1.000000f, 0.890000f, 0.316228f, 0.316228f, 0.100000f, 2.320000f, 0.830000f, 0.310000f, 0.891251f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.023000f, { 0.000000f, 0.000000f, 0.000000f }, 0.138000f, 0.080000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_MEDIUMROOM",			{ 1.000000f, 0.930000f, 0.316228f, 0.281838f, 0.100000f, 2.040000f, 0.830000f, 0.460000f, 0.630957f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 1.584893f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.155000f, 0.030000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_LONGPASSAGE",		{ 1.000000f, 0.890000f, 0.316228f, 0.398107f, 0.100000f, 3.420000f, 0.830000f, 0.310000f, 0.891251f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.023000f, { 0.000000f, 0.000000f, 0.000000f }, 0.138000f, 0.080000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_LARGEROOM",			{ 1.000000f, 0.820000f, 0.316228f, 0.281838f, 0.125893f, 2.530000f, 0.830000f, 0.500000f, 0.446684f, 0.034000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.016000f, { 0.000000f, 0.000000f, 0.000000f }, 0.185000f, 0.070000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_HALL",				{ 1.000000f, 0.810000f, 0.316228f, 0.281838f, 0.177828f, 3.140000f, 0.790000f, 0.620000f, 0.177828f, 0.056000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.024000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_CUPBOARD",			{ 1.000000f, 0.890000f, 0.316228f, 0.281838f, 0.100000f, 0.670000f, 0.870000f, 0.310000f, 1.412538f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 3.548134f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 0.138000f, 0.080000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"CASTLE_COURTYARD",			{ 1.000000f, 0.420000f, 0.316228f, 0.446684f, 0.199526f, 2.130000f, 0.610000f, 0.230000f, 0.223872f, 0.160000f, { 0.000000f, 0.000000f, 0.000000f }, 0.707946f, 0.036000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.370000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"CASTLE_ALCOVE",				{ 1.000000f, 0.890000f, 0.316228f, 0.501187f, 0.100000f, 1.640000f, 0.870000f, 0.310000f, 1.000000f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.034000f, { 0.000000f, 0.000000f, 0.000000f }, 0.138000f, 0.080000f, 0.250000f, 0.000000f, 0.994260f, 5168.600098f, 139.500000f, 0.000000f, 1 } },
	{ L"FACTORY_ALCOVE",			{ 0.364500f, 0.590000f, 0.251189f, 0.794328f, 0.501187f, 3.140000f, 0.650000f, 1.310000f, 1.412538f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.000000f, 0.038000f, { 0.000000f, 0.000000f, 0.000000f }, 0.114000f, 0.100000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_SHORTPASSAGE",		{ 0.364500f, 0.640000f, 0.251189f, 0.794328f, 0.501187f, 2.530000f, 0.650000f, 1.310000f, 1.000000f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.038000f, { 0.000000f, 0.000000f, 0.000000f }, 0.135000f, 0.230000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_MEDIUMROOM",		{ 0.428687f, 0.820000f, 0.251189f, 0.794328f, 0.501187f, 2.760000f, 0.650000f, 1.310000f, 0.281838f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.023000f, { 0.000000f, 0.000000f, 0.000000f }, 0.174000f, 0.070000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_LONGPASSAGE",		{ 0.364500f, 0.640000f, 0.251189f, 0.794328f, 0.501187f, 4.060000f, 0.650000f, 1.310000f, 1.000000f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.037000f, { 0.000000f, 0.000000f, 0.000000f }, 0.135000f, 0.230000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_LARGEROOM",			{ 0.428687f, 0.750000f, 0.251189f, 0.707946f, 0.630957f, 4.240000f, 0.510000f, 1.310000f, 0.177828f, 0.039000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.023000f, { 0.000000f, 0.000000f, 0.000000f }, 0.231000f, 0.070000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_HALL",				{ 0.428687f, 0.750000f, 0.316228f, 0.707946f, 0.630957f, 7.430000f, 0.510000f, 1.310000f, 0.063096f, 0.073000f, { 0.000000f, 0.000000f, 0.000000f }, 0.891251f, 0.027000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.070000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_CUPBOARD",			{ 0.307063f, 0.630000f, 0.251189f, 0.794328f, 0.501187f, 0.490000f, 0.650000f, 1.310000f, 1.258925f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.995262f, 0.032000f, { 0.000000f, 0.000000f, 0.000000f }, 0.107000f, 0.070000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_COURTYARD",			{ 0.307063f, 0.570000f, 0.316228f, 0.316228f, 0.630957f, 2.320000f, 0.290000f, 0.560000f, 0.223872f, 0.140000f, { 0.000000f, 0.000000f, 0.000000f }, 0.398107f, 0.039000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.290000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"FACTORY_SMALLROOM",			{ 0.364500f, 0.820000f, 0.316228f, 0.794328f, 0.501187f, 1.720000f, 0.650000f, 1.310000f, 0.707946f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.778279f, 0.024000f, { 0.000000f, 0.000000f, 0.000000f }, 0.119000f, 0.070000f, 0.250000f, 0.000000f, 0.994260f, 3762.600098f, 362.500000f, 0.000000f, 1 } },
	{ L"ICEPALACE_ALCOVE",			{ 1.000000f, 0.840000f, 0.316228f, 0.562341f, 0.281838f, 2.760000f, 1.460000f, 0.280000f, 1.122018f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 0.891251f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.161000f, 0.090000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_SHORTPASSAGE",	{ 1.000000f, 0.750000f, 0.316228f, 0.562341f, 0.281838f, 1.790000f, 1.460000f, 0.280000f, 0.501187f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.019000f, { 0.000000f, 0.000000f, 0.000000f }, 0.177000f, 0.090000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_MEDIUMROOM",		{ 1.000000f, 0.870000f, 0.316228f, 0.562341f, 0.446684f, 2.220000f, 1.530000f, 0.320000f, 0.398107f, 0.039000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.027000f, { 0.000000f, 0.000000f, 0.000000f }, 0.186000f, 0.120000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_LONGPASSAGE",		{ 1.000000f, 0.770000f, 0.316228f, 0.562341f, 0.398107f, 3.010000f, 1.460000f, 0.280000f, 0.794328f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.025000f, { 0.000000f, 0.000000f, 0.000000f }, 0.186000f, 0.040000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_LARGEROOM",		{ 1.000000f, 0.810000f, 0.316228f, 0.562341f, 0.446684f, 3.140000f, 1.530000f, 0.320000f, 0.251189f, 0.039000f, { 0.000000f, 0.000000f, 0.000000f }, 1.000000f, 0.027000f, { 0.000000f, 0.000000f, 0.000000f }, 0.214000f, 0.110000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_HALL",			{ 1.000000f, 0.760000f, 0.316228f, 0.446684f, 0.562341f, 5.490000f, 1.530000f, 0.380000f, 0.112202f, 0.054000f, { 0.000000f, 0.000000f, 0.000000f }, 0.630957f, 0.052000f, { 0.000000f, 0.000000f, 0.000000f }, 0.226000f, 0.110000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_CUPBOARD",		{ 1.000000f, 0.830000f, 0.316228f, 0.501187f, 0.223872f, 0.760000f, 1.530000f, 0.260000f, 1.122018f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.995262f, 0.016000f, { 0.000000f, 0.000000f, 0.000000f }, 0.143000f, 0.080000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_COURTYARD",		{ 1.000000f, 0.590000f, 0.316228f, 0.281838f, 0.316228f, 2.040000f, 1.200000f, 0.380000f, 0.316228f, 0.173000f, { 0.000000f, 0.000000f, 0.000000f }, 0.316228f, 0.043000f, { 0.000000f, 0.000000f, 0.000000f }, 0.235000f, 0.480000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"ICEPALACE_SMALLROOM",		{ 1.000000f, 0.840000f, 0.316228f, 0.562341f, 0.281838f, 1.510000f, 1.530000f, 0.270000f, 0.891251f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.164000f, 0.140000f, 0.250000f, 0.000000f, 0.994260f, 12428.500000f, 99.599998f, 0.000000f, 1 } },
	{ L"SPACESTATION_ALCOVE",		{ 0.210938f, 0.780000f, 0.316228f, 0.707946f, 0.891251f, 1.160000f, 0.810000f, 0.550000f, 1.412538f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.000000f, 0.018000f, { 0.000000f, 0.000000f, 0.000000f }, 0.192000f, 0.210000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_MEDIUMROOM",	{ 0.210938f, 0.750000f, 0.316228f, 0.630957f, 0.891251f, 3.010000f, 0.500000f, 0.550000f, 0.398107f, 0.034000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.035000f, { 0.000000f, 0.000000f, 0.000000f }, 0.209000f, 0.310000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_SHORTPASSAGE",	{ 0.210938f, 0.870000f, 0.316228f, 0.630957f, 0.891251f, 3.570000f, 0.500000f, 0.550000f, 1.000000f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.016000f, { 0.000000f, 0.000000f, 0.000000f }, 0.172000f, 0.200000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_LONGPASSAGE",	{ 0.428687f, 0.820000f, 0.316228f, 0.630957f, 0.891251f, 4.620000f, 0.620000f, 0.550000f, 1.000000f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.031000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.230000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_LARGEROOM",	{ 0.364500f, 0.810000f, 0.316228f, 0.630957f, 0.891251f, 3.890000f, 0.380000f, 0.610000f, 0.316228f, 0.056000f, { 0.000000f, 0.000000f, 0.000000f }, 0.891251f, 0.035000f, { 0.000000f, 0.000000f, 0.000000f }, 0.233000f, 0.280000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_HALL",			{ 0.428687f, 0.870000f, 0.316228f, 0.630957f, 0.891251f, 7.110000f, 0.380000f, 0.610000f, 0.177828f, 0.100000f, { 0.000000f, 0.000000f, 0.000000f }, 0.630957f, 0.047000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.250000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_CUPBOARD",		{ 0.171500f, 0.560000f, 0.316228f, 0.707946f, 0.891251f, 0.790000f, 0.810000f, 0.550000f, 1.412538f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.778279f, 0.018000f, { 0.000000f, 0.000000f, 0.000000f }, 0.181000f, 0.310000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"SPACESTATION_SMALLROOM",	{ 0.210938f, 0.700000f, 0.316228f, 0.707946f, 0.891251f, 1.720000f, 0.820000f, 0.550000f, 0.794328f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.013000f, { 0.000000f, 0.000000f, 0.000000f }, 0.188000f, 0.260000f, 0.250000f, 0.000000f, 0.994260f, 3316.100098f, 458.200012f, 0.000000f, 1 } },
	{ L"WOODEN_ALCOVE",				{ 1.000000f, 1.000000f, 0.316228f, 0.125893f, 0.316228f, 1.220000f, 0.620000f, 0.910000f, 1.122018f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 0.707946f, 0.024000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_SHORTPASSAGE",		{ 1.000000f, 1.000000f, 0.316228f, 0.125893f, 0.316228f, 1.750000f, 0.500000f, 0.870000f, 0.891251f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 0.630957f, 0.024000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_MEDIUMROOM",			{ 1.000000f, 1.000000f, 0.316228f, 0.100000f, 0.281838f, 1.470000f, 0.420000f, 0.820000f, 0.891251f, 0.049000f, { 0.000000f, 0.000000f, 0.000000f }, 0.891251f, 0.029000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_LONGPASSAGE",		{ 1.000000f, 1.000000f, 0.316228f, 0.100000f, 0.316228f, 1.990000f, 0.400000f, 0.790000f, 1.000000f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.446684f, 0.036000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_LARGEROOM",			{ 1.000000f, 1.000000f, 0.316228f, 0.089125f, 0.281838f, 2.650000f, 0.330000f, 0.820000f, 0.891251f, 0.066000f, { 0.000000f, 0.000000f, 0.000000f }, 0.794328f, 0.049000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_HALL",				{ 1.000000f, 1.000000f, 0.316228f, 0.079433f, 0.281838f, 3.450000f, 0.300000f, 0.820000f, 0.891251f, 0.088000f, { 0.000000f, 0.000000f, 0.000000f }, 0.794328f, 0.063000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_CUPBOARD",			{ 1.000000f, 1.000000f, 0.316228f, 0.141254f, 0.316228f, 0.560000f, 0.460000f, 0.910000f, 1.122018f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.028000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_SMALLROOM",			{ 1.000000f, 1.000000f, 0.316228f, 0.112202f, 0.316228f, 0.790000f, 0.320000f, 0.870000f, 1.000000f, 0.032000f, { 0.000000f, 0.000000f, 0.000000f }, 0.891251f, 0.029000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"WOODEN_COURTYARD",			{ 1.000000f, 0.650000f, 0.316228f, 0.079433f, 0.316228f, 1.790000f, 0.350000f, 0.790000f, 0.562341f, 0.123000f, { 0.000000f, 0.000000f, 0.000000f }, 0.100000f, 0.032000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 4705.000000f, 99.599998f, 0.000000f, 1 } },
	{ L"SPORT_EMPTYSTADIUM",		{ 1.000000f, 1.000000f, 0.316228f, 0.446684f, 0.794328f, 6.260000f, 0.510000f, 1.100000f, 0.063096f, 0.183000f, { 0.000000f, 0.000000f, 0.000000f }, 0.398107f, 0.038000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"SPORT_SQUASHCOURT",			{ 1.000000f, 0.750000f, 0.316228f, 0.316228f, 0.794328f, 2.220000f, 0.910000f, 1.160000f, 0.446684f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 0.794328f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.126000f, 0.190000f, 0.250000f, 0.000000f, 0.994260f, 7176.899902f, 211.199997f, 0.000000f, 1 } },
	{ L"SPORT_SMALLSWIMMINGPOOL",	{ 1.000000f, 0.700000f, 0.316228f, 0.794328f, 0.891251f, 2.760000f, 1.250000f, 1.140000f, 0.630957f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.794328f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.179000f, 0.150000f, 0.895000f, 0.190000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"SPORT_LARGESWIMMINGPOOL",	{ 1.000000f, 0.820000f, 0.316228f, 0.794328f, 1.000000f, 5.490000f, 1.310000f, 1.140000f, 0.446684f, 0.039000f, { 0.000000f, 0.000000f, 0.000000f }, 0.501187f, 0.049000f, { 0.000000f, 0.000000f, 0.000000f }, 0.222000f, 0.550000f, 1.159000f, 0.210000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"SPORT_GYMNASIUM",			{ 1.000000f, 0.810000f, 0.316228f, 0.446684f, 0.891251f, 3.140000f, 1.060000f, 1.350000f, 0.398107f, 0.029000f, { 0.000000f, 0.000000f, 0.000000f }, 0.562341f, 0.045000f, { 0.000000f, 0.000000f, 0.000000f }, 0.146000f, 0.140000f, 0.250000f, 0.000000f, 0.994260f, 7176.899902f, 211.199997f, 0.000000f, 1 } },
	{ L"SPORT_FULLSTADIUM",			{ 1.000000f, 1.000000f, 0.316228f, 0.070795f, 0.794328f, 5.250000f, 0.170000f, 0.800000f, 0.100000f, 0.188000f, { 0.000000f, 0.000000f, 0.000000f }, 0.281838f, 0.038000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"SPORT_STADIUMTANNOY",		{ 1.000000f, 0.780000f, 0.316228f, 0.562341f, 0.501187f, 2.530000f, 0.880000f, 0.680000f, 0.281838f, 0.230000f, { 0.000000f, 0.000000f, 0.000000f }, 0.501187f, 0.063000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.200000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"PREFAB_WORKSHOP",			{ 0.428687f, 1.000000f, 0.316228f, 0.141254f, 0.398107f, 0.760000f, 1.000000f, 1.000000f, 1.000000f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"PREFAB_SCHOOLROOM",			{ 0.402178f, 0.690000f, 0.316228f, 0.630957f, 0.501187f, 0.980000f, 0.450000f, 0.180000f, 1.412538f, 0.017000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.015000f, { 0.000000f, 0.000000f, 0.000000f }, 0.095000f, 0.140000f, 0.250000f, 0.000000f, 0.994260f, 7176.899902f, 211.199997f, 0.000000f, 1 } },
	{ L"PREFAB_PRACTISEROOM",		{ 0.402178f, 0.870000f, 0.316228f, 0.398107f, 0.501187f, 1.120000f, 0.560000f, 0.180000f, 1.258925f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.011000f, { 0.000000f, 0.000000f, 0.000000f }, 0.095000f, 0.140000f, 0.250000f, 0.000000f, 0.994260f, 7176.899902f, 211.199997f, 0.000000f, 1 } },
	{ L"PREFAB_OUTHOUSE",			{ 1.000000f, 0.820000f, 0.316228f, 0.112202f, 0.158489f, 1.380000f, 0.380000f, 0.350000f, 0.891251f, 0.024000f, { 0.000000f, 0.000000f, -0.000000f }, 0.630957f, 0.044000f, { 0.000000f, 0.000000f, 0.000000f }, 0.121000f, 0.170000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 107.500000f, 0.000000f, 0 } },
	{ L"PREFAB_CARAVAN",			{ 1.000000f, 1.000000f, 0.316228f, 0.089125f, 0.125893f, 0.430000f, 1.500000f, 1.000000f, 1.000000f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 1.995262f, 0.012000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"DOME_TOMB",					{ 1.000000f, 0.790000f, 0.316228f, 0.354813f, 0.223872f, 4.180000f, 0.210000f, 0.100000f, 0.386812f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 1.678804f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 0.177000f, 0.190000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 20.000000f, 0.000000f, 0 } },
	{ L"DOME_SAINTPAULS",			{ 1.000000f, 0.870000f, 0.316228f, 0.354813f, 0.223872f, 10.480000f, 0.190000f, 0.100000f, 0.177828f, 0.090000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.042000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.120000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 20.000000f, 0.000000f, 1 } },
	{ L"PIPE_SMALL",				{ 1.000000f, 1.000000f, 0.316228f, 0.354813f, 0.223872f, 5.040000f, 0.100000f, 0.100000f, 0.501187f, 0.032000f, { 0.000000f, 0.000000f, 0.000000f }, 2.511886f, 0.015000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 20.000000f, 0.000000f, 1 } },
	{ L"PIPE_LONGTHIN",				{ 0.256000f, 0.910000f, 0.316228f, 0.446684f, 0.281838f, 9.210000f, 0.180000f, 0.100000f, 0.707946f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 0.707946f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 20.000000f, 0.000000f, 0 } },
	{ L"PIPE_LARGE",				{ 1.000000f, 1.000000f, 0.316228f, 0.354813f, 0.223872f, 8.450000f, 0.100000f, 0.100000f, 0.398107f, 0.046000f, { 0.000000f, 0.000000f, 0.000000f }, 1.584893f, 0.032000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 20.000000f, 0.000000f, 1 } },
	{ L"PIPE_RESONANT",				{ 0.137312f, 0.910000f, 0.316228f, 0.446684f, 0.281838f, 6.810000f, 0.180000f, 0.100000f, 0.707946f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.000000f, 0.022000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 20.000000f, 0.000000f, 0 } },
	{ L"OUTDOORS_BACKYARD",			{ 1.000000f, 0.450000f, 0.316228f, 0.251189f, 0.501187f, 1.120000f, 0.340000f, 0.460000f, 0.446684f, 0.069000f, { 0.000000f, 0.000000f, -0.000000f }, 0.707946f, 0.023000f, { 0.000000f, 0.000000f, 0.000000f }, 0.218000f, 0.340000f, 0.250000f, 0.000000f, 0.994260f, 4399.100098f, 242.899994f, 0.000000f, 0 } },
	{ L"OUTDOORS_ROLLINGPLAINS",	{ 1.000000f, 0.000000f, 0.316228f, 0.011220f, 0.630957f, 2.130000f, 0.210000f, 0.460000f, 0.177828f, 0.300000f, { 0.000000f, 0.000000f, -0.000000f }, 0.446684f, 0.019000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 1.000000f, 0.250000f, 0.000000f, 0.994260f, 4399.100098f, 242.899994f, 0.000000f, 0 } },
	{ L"OUTDOORS_DEEPCANYON",		{ 1.000000f, 0.740000f, 0.316228f, 0.177828f, 0.630957f, 3.890000f, 0.210000f, 0.460000f, 0.316228f, 0.223000f, { 0.000000f, 0.000000f, -0.000000f }, 0.354813f, 0.019000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 1.000000f, 0.250000f, 0.000000f, 0.994260f, 4399.100098f, 242.899994f, 0.000000f, 0 } },
	{ L"OUTDOORS_CREEK",			{ 1.000000f, 0.350000f, 0.316228f, 0.177828f, 0.501187f, 2.130000f, 0.210000f, 0.460000f, 0.398107f, 0.115000f, { 0.000000f, 0.000000f, -0.000000f }, 0.199526f, 0.031000f, { 0.000000f, 0.000000f, 0.000000f }, 0.218000f, 0.340000f, 0.250000f, 0.000000f, 0.994260f, 4399.100098f, 242.899994f, 0.000000f, 0 } },
	{ L"OUTDOORS_VALLEY",			{ 1.000000f, 0.280000f, 0.316228f, 0.028184f, 0.158489f, 2.880000f, 0.260000f, 0.350000f, 0.141254f, 0.263000f, { 0.000000f, 0.000000f, -0.000000f }, 0.398107f, 0.100000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.340000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 107.500000f, 0.000000f, 0 } },
	{ L"MOOD_HEAVEN",				{ 1.000000f, 0.940000f, 0.316228f, 0.794328f, 0.446684f, 5.040000f, 1.120000f, 0.560000f, 0.242661f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.029000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.080000f, 2.742000f, 0.050000f, 0.997700f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"MOOD_HELL",					{ 1.000000f, 0.570000f, 0.316228f, 0.354813f, 0.446684f, 3.570000f, 0.490000f, 2.000000f, 0.000000f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.110000f, 0.040000f, 2.109000f, 0.520000f, 0.994260f, 5000.000000f, 139.500000f, 0.000000f, 0 } },
	{ L"MOOD_MEMORY",				{ 1.000000f, 0.850000f, 0.316228f, 0.630957f, 0.354813f, 4.060000f, 0.820000f, 0.560000f, 0.039811f, 0.000000f, { 0.000000f, 0.000000f, 0.000000f }, 1.122018f, 0.000000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.474000f, 0.450000f, 0.988553f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"DRIVING_COMMENTATOR",		{ 1.000000f, 0.000000f, 1.000000f, 0.562341f, 0.501187f, 2.420000f, 0.880000f, 0.680000f, 0.199526f, 0.093000f, { 0.000000f, 0.000000f, 0.000000f }, 0.251189f, 0.017000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 1.000000f, 0.250000f, 0.000000f, 0.988553f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"DRIVING_PITGARAGE",			{ 0.428687f, 0.590000f, 0.316228f, 0.707946f, 0.562341f, 1.720000f, 0.930000f, 0.870000f, 0.562341f, 0.000000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.016000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.110000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
	{ L"DRIVING_INCAR_RACER",		{ 0.083188f, 0.800000f, 0.316228f, 1.000000f, 0.794328f, 0.170000f, 2.000000f, 0.410000f, 1.778279f, 0.007000f, { 0.000000f, 0.000000f, 0.000000f }, 0.707946f, 0.015000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 10268.200195f, 251.000000f, 0.000000f, 1 } },
	{ L"DRIVING_INCAR_SPORTS",		{ 0.083188f, 0.800000f, 0.316228f, 0.630957f, 1.000000f, 0.170000f, 0.750000f, 0.410000f, 1.000000f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 0.562341f, 0.000000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 10268.200195f, 251.000000f, 0.000000f, 1 } },
	{ L"DRIVING_INCAR_LUXURY",		{ 0.256000f, 1.000000f, 0.316228f, 0.100000f, 0.501187f, 0.130000f, 0.410000f, 0.460000f, 0.794328f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 1.584893f, 0.010000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 10268.200195f, 251.000000f, 0.000000f, 1 } },
	{ L"DRIVING_FULLGRANDSTAND",	{ 1.000000f, 1.000000f, 0.316228f, 0.281838f, 0.630957f, 3.010000f, 1.370000f, 1.280000f, 0.354813f, 0.090000f, { 0.000000f, 0.000000f, 0.000000f }, 0.177828f, 0.049000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 10420.200195f, 250.000000f, 0.000000f, 0 } },
	{ L"DRIVING_EMPTYGRANDSTAND",	{ 1.000000f, 1.000000f, 0.316228f, 1.000000f, 0.794328f, 4.620000f, 1.750000f, 1.400000f, 0.208209f, 0.090000f, { 0.000000f, 0.000000f, 0.000000f }, 0.251189f, 0.049000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.000000f, 0.994260f, 10420.200195f, 250.000000f, 0.000000f, 0 } },
	{ L"DRIVING_TUNNEL",			{ 1.000000f, 0.810000f, 0.316228f, 0.398107f, 0.891251f, 3.420000f, 0.940000f, 1.310000f, 0.707946f, 0.051000f, { 0.000000f, 0.000000f, 0.000000f }, 0.707946f, 0.047000f, { 0.000000f, 0.000000f, 0.000000f }, 0.214000f, 0.050000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 155.300003f, 0.000000f, 1 } },
	{ L"CITY_STREETS",				{ 1.000000f, 0.780000f, 0.316228f, 0.707946f, 0.891251f, 1.790000f, 1.120000f, 0.910000f, 0.281838f, 0.046000f, { 0.000000f, 0.000000f, 0.000000f }, 0.199526f, 0.028000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.200000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CITY_SUBWAY",				{ 1.000000f, 0.740000f, 0.316228f, 0.707946f, 0.891251f, 3.010000f, 1.230000f, 0.910000f, 0.707946f, 0.046000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.028000f, { 0.000000f, 0.000000f, 0.000000f }, 0.125000f, 0.210000f, 0.250000f, 0.000000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CITY_MUSEUM",				{ 1.000000f, 0.820000f, 0.316228f, 0.177828f, 0.177828f, 3.280000f, 1.400000f, 0.570000f, 0.251189f, 0.039000f, { 0.000000f, 0.000000f, -0.000000f }, 0.891251f, 0.034000f, { 0.000000f, 0.000000f, 0.000000f }, 0.130000f, 0.170000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 107.500000f, 0.000000f, 0 } },
	{ L"CITY_LIBRARY",				{ 1.000000f, 0.820000f, 0.316228f, 0.281838f, 0.089125f, 2.760000f, 0.890000f, 0.410000f, 0.354813f, 0.029000f, { 0.000000f, 0.000000f, -0.000000f }, 0.891251f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 0.130000f, 0.170000f, 0.250000f, 0.000000f, 0.994260f, 2854.399902f, 107.500000f, 0.000000f, 0 } },
	{ L"CITY_UNDERPASS",			{ 1.000000f, 0.820000f, 0.316228f, 0.446684f, 0.891251f, 3.570000f, 1.120000f, 0.910000f, 0.398107f, 0.059000f, { 0.000000f, 0.000000f, 0.000000f }, 0.891251f, 0.037000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.140000f, 0.250000f, 0.000000f, 0.991973f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"CITY_ABANDONED",			{ 1.000000f, 0.690000f, 0.316228f, 0.794328f, 0.891251f, 3.280000f, 1.170000f, 0.910000f, 0.446684f, 0.044000f, { 0.000000f, 0.000000f, 0.000000f }, 0.281838f, 0.024000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.200000f, 0.250000f, 0.000000f, 0.996552f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"DUSTYROOM",					{ 0.364500f, 0.560000f, 0.316228f, 0.794328f, 0.707946f, 1.790000f, 0.380000f, 0.210000f, 0.501187f, 0.002000f, { 0.000000f, 0.000000f, 0.000000f }, 1.258925f, 0.006000f, { 0.000000f, 0.000000f, 0.000000f }, 0.202000f, 0.050000f, 0.250000f, 0.000000f, 0.988553f, 13046.000000f, 163.300003f, 0.000000f, 1 } },
	{ L"CHAPEL",					{ 1.000000f, 0.840000f, 0.316228f, 0.562341f, 1.000000f, 4.620000f, 0.640000f, 1.230000f, 0.446684f, 0.032000f, { 0.000000f, 0.000000f, 0.000000f }, 0.794328f, 0.049000f, { 0.000000f, 0.000000f, 0.000000f }, 0.250000f, 0.000000f, 0.250000f, 0.110000f, 0.994260f, 5000.000000f, 250.000000f, 0.000000f, 1 } },
	{ L"SMALLWATERROOM",			{ 1.000000f, 0.700000f, 0.316228f, 0.447713f, 1.000000f, 1.510000f, 1.250000f, 1.140000f, 0.891251f, 0.020000f, { 0.000000f, 0.000000f, 0.000000f }, 1.412538f, 0.030000f, { 0.000000f, 0.000000f, 0.000000f }, 0.179000f, 0.150000f, 0.895000f, 0.190000f, 0.991973f, 5000.000000f, 250.000000f, 0.000000f, 0 } },
};

const float fASURA_OPENAL_INVALID_FILTER_GAIN = 1.1f;	//an invalid gain value (within our context) Should be > 1.0f
const u_int uOPENAL_DISTANCE_FILTER_GAIN_POWER = 3;		//Empirically this sounds to me like its closest to 360 and gives a pleasing falloff

/* ------------------------------------------------------------------------- */

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

	// TODO these variables should be stored with the rest of the DVS stuff
	static bool s_bPrintEvents				= false;
	static int s_iAirAbsorptionFactor		= 0;
	#define CONSOLE_VARNAME_PRINT_EVENTS	L"Audio.Debug.PrintOpenALEvents"
	#define CONSOLE_VARNAME_AIR_ABSORPTION	L"Audio.Settings.AirAbsorption"

	#define PRINT_SOUND_EVENT( params )		{ if ( s_bPrintEvents ){ Asura_OutputToDebugger_Retail( params ); } }

#else // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

	#define PRINT_SOUND_EVENT( params )

	static const int s_iAirAbsorptionFactor	= 0;

#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

/* ------------------------------------------------------------------------- */

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
// TODO move this into Asura_Sound_System
#define CONSOLE_CMDNAME_SETREVERB	L"SetReverb"
static bool ConsoleCmd_SetReverb( int iReverbID, bool bInterrogate, bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_SETREVERB L": override reverb setting - specify a number corresponding to reverb ID. -1 to return reverb control to game." );
		ConsoleCmd_SetReverb( Asura_OpenAL_Sound_API::Platform_GetAPI().GetForcedReverb(), true, false );
	}
	else if ( bInterrogate )
	{
		int currentIndex = Asura_OpenAL_Sound_API::Platform_GetAPI().GetForcedReverb();
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_SETREVERB L" %d (%s)", currentIndex,
			currentIndex >= 0 ? s_axReverbPresets[currentIndex].m_szName : L"under game control" );

		if ( !Asura_OpenAL_Sound_API::Platform_GetAPI().IsEFXAvailable() )
		{
			Asura_CommandConsole::Printf( L"(EFX not available - install the latest OpenAL and update your sound drivers)" );
		}
	}
	else
	{
		Asura_OpenAL_Sound_API::Platform_GetAPI().ForceReverb( iReverbID );
		ConsoleCmd_SetReverb( iReverbID, true, false );
	}

	return true;
}
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM

/* ------------------------------------------------------------------------- */

#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )

#define CONSOLE_CMDNAME_AUDIO_DIAGNOSTICS	L"AudioDiagnostics"
bool ConsoleCmd_AudioDiagnostics( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_AUDIO_DIAGNOSTICS L": prints audio system info to aid with debugging." );
	}
	else
	{
		ALint majorValue = 0;
		ALint minorValue = 0;
		const ALchar* deviceName = NULL;
		
		alcGetIntegerv( Asura_OpenAL_Sound_API::Platform_GetAPI().m_pxDevice, ALC_MAJOR_VERSION, 1, &majorValue );
		alcGetIntegerv( Asura_OpenAL_Sound_API::Platform_GetAPI().m_pxDevice, ALC_MAJOR_VERSION, 1, &minorValue );
		deviceName = alcGetString( Asura_OpenAL_Sound_API::Platform_GetAPI().m_pxDevice, ALC_DEVICE_SPECIFIER );
		
		Asura_CommandConsole::Printf( L"OpenAL %d.%d", majorValue, minorValue );
		Asura_CommandConsole::Printf( L"Device: 0x%08x, Context: 0x%08x, Device Name: %S", 
			Asura_OpenAL_Sound_API::Platform_GetAPI().m_pxDevice, Asura_OpenAL_Sound_API::Platform_GetAPI().m_pxContext, deviceName ? deviceName : "unknown" );

		int theoreticalMaxAuxSends = 0;
		alcGetIntegerv( Asura_OpenAL_Sound_API::Platform_GetAPI().m_pxDevice, ALC_MAX_AUXILIARY_SENDS, 1, &theoreticalMaxAuxSends );
		Asura_CommandConsole::Printf( L"EFX are%s available, num aux sends: %d (theoretically %d)",
			Asura_OpenAL_Sound_API::Platform_GetAPI().m_bEFXAvailable ? "" : " not", Asura_OpenAL_Sound_API::Platform_GetAPI().m_iAuxSendsPerSource, theoreticalMaxAuxSends );

		Asura_CommandConsole::Printf( L"Listener Position: %.2f, %.2f, %.2f",
			Asura_OpenAL_Sound_API::Platform_GetAPI().m_xListenerPos.x, Asura_OpenAL_Sound_API::Platform_GetAPI().m_xListenerPos.y, Asura_OpenAL_Sound_API::Platform_GetAPI().m_xListenerPos.z );
		Asura_CommandConsole::Printf( L"Current Reverb: %d (%s)", Asura_OpenAL_Sound_API::Platform_GetAPI().m_eCurrentReverb,
			Asura_OpenAL_Sound_API::Platform_GetAPI().m_eCurrentReverb >= 0 ? s_axReverbPresets[Asura_OpenAL_Sound_API::Platform_GetAPI().m_eCurrentReverb].m_szName : L"not set" );
	}

	return true;
}
#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )

/* ------------------------------------------------------------------------- */

static bool CacheEFXFunctions()
{
	PRINT_SOUND_EVENT(( __FUNCTION__ "\n" ));

	alGenEffects = static_cast<LPALGENEFFECTS>( alGetProcAddress( "alGenEffects" ) );
	alDeleteEffects = static_cast<LPALDELETEEFFECTS>( alGetProcAddress( "alDeleteEffects" ) );
	alIsEffect = static_cast<LPALISEFFECT>( alGetProcAddress( "alIsEffect" ) );
	alEffecti = static_cast<LPALEFFECTI>( alGetProcAddress( "alEffecti" ) );
	alEffectiv = static_cast<LPALEFFECTIV>( alGetProcAddress( "alEffectiv" ) );
	alEffectf = static_cast<LPALEFFECTF>( alGetProcAddress( "alEffectf" ) );
	alEffectfv = static_cast<LPALEFFECTFV>( alGetProcAddress( "alEffectfv" ) );
	alGetEffecti = static_cast<LPALGETEFFECTI>( alGetProcAddress( "alGetEffecti" ) );
	alGetEffectiv = static_cast<LPALGETEFFECTIV>( alGetProcAddress( "alGetEffectiv" ) );
	alGetEffectf = static_cast<LPALGETEFFECTF>( alGetProcAddress( "alGetEffectf" ) );
	alGetEffectfv = static_cast<LPALGETEFFECTFV>( alGetProcAddress( "alGetEffectfv" ) );
	alGenFilters = static_cast<LPALGENFILTERS>( alGetProcAddress( "alGenFilters" ) );
	alDeleteFilters = static_cast<LPALDELETEFILTERS>( alGetProcAddress( "alDeleteFilters" ) );
	alIsFilter = static_cast<LPALISFILTER>( alGetProcAddress( "alIsFilter" ) );
	alFilteri = static_cast<LPALFILTERI>( alGetProcAddress( "alFilteri" ) );
	alFilteriv = static_cast<LPALFILTERIV>( alGetProcAddress( "alFilteriv" ) );
	alFilterf = static_cast<LPALFILTERF>( alGetProcAddress( "alFilterf" ) );
	alFilterfv = static_cast<LPALFILTERFV>( alGetProcAddress( "alFilterfv" ) );
	alGetFilteri = static_cast<LPALGETFILTERI>( alGetProcAddress( "alGetFilteri" ) );
	alGetFilteriv = static_cast<LPALGETFILTERIV>( alGetProcAddress( "alGetFilteriv" ) );
	alGetFilterf = static_cast<LPALGETFILTERF>( alGetProcAddress( "alGetFilterf" ) );
	alGetFilterfv = static_cast<LPALGETFILTERFV>( alGetProcAddress( "alGetFilterfv" ) );
	alGenAuxiliaryEffectSlots = static_cast<LPALGENAUXILIARYEFFECTSLOTS>( alGetProcAddress( "alGenAuxiliaryEffectSlots" ) );
	alDeleteAuxiliaryEffectSlots = static_cast<LPALDELETEAUXILIARYEFFECTSLOTS>( alGetProcAddress( "alDeleteAuxiliaryEffectSlots" ) );
	alIsAuxiliaryEffectSlot = static_cast<LPALISAUXILIARYEFFECTSLOT>( alGetProcAddress( "alIsAuxiliaryEffectSlot" ) );
	alAuxiliaryEffectSloti = static_cast<LPALAUXILIARYEFFECTSLOTI>( alGetProcAddress( "alAuxiliaryEffectSloti" ) );
	alAuxiliaryEffectSlotiv = static_cast<LPALAUXILIARYEFFECTSLOTIV>( alGetProcAddress( "alAuxiliaryEffectSlotiv" ) );
	alAuxiliaryEffectSlotf = static_cast<LPALAUXILIARYEFFECTSLOTF>( alGetProcAddress( "alAuxiliaryEffectSlotf" ) );
	alAuxiliaryEffectSlotfv = static_cast<LPALAUXILIARYEFFECTSLOTFV>( alGetProcAddress( "alAuxiliaryEffectSlotfv" ) );
	alGetAuxiliaryEffectSloti = static_cast<LPALGETAUXILIARYEFFECTSLOTI>( alGetProcAddress( "alGetAuxiliaryEffectSloti" ) );
	alGetAuxiliaryEffectSlotiv = static_cast<LPALGETAUXILIARYEFFECTSLOTIV>( alGetProcAddress( "alGetAuxiliaryEffectSlotiv" ) );
	alGetAuxiliaryEffectSlotf = static_cast<LPALGETAUXILIARYEFFECTSLOTF>( alGetProcAddress( "alGetAuxiliaryEffectSlotf" ) );
	alGetAuxiliaryEffectSlotfv = static_cast<LPALGETAUXILIARYEFFECTSLOTFV>( alGetProcAddress( "alGetAuxiliaryEffectSlotfv" ) );
	
	// Check all functions are available (non-NULL). If they're not, we can't use EFX
	// (no sense in trying to work out if we can use subsets of it).

	bool allFuncsPresent =
		( alGenEffects
		&& alDeleteEffects
		&& alIsEffect
		&& alEffecti
		&& alEffectiv
		&& alEffectf
		&& alEffectfv
		&& alGetEffecti
		&& alGetEffectiv
		&& alGetEffectf
		&& alGetEffectfv
		&& alGenFilters
		&& alDeleteFilters
		&& alIsFilter
		&& alFilteri
		&& alFilteriv
		&& alFilterf
		&& alFilterfv
		&& alGetFilteri
		&& alGetFilteriv
		&& alGetFilterf
		&& alGetFilterfv
		&& alGenAuxiliaryEffectSlots
		&& alDeleteAuxiliaryEffectSlots
		&& alIsAuxiliaryEffectSlot
		&& alAuxiliaryEffectSloti
		&& alAuxiliaryEffectSlotiv
		&& alAuxiliaryEffectSlotf
		&& alAuxiliaryEffectSlotfv
		&& alGetAuxiliaryEffectSloti
		&& alGetAuxiliaryEffectSlotiv
		&& alGetAuxiliaryEffectSlotf
		&& alGetAuxiliaryEffectSlotfv );
	
	Asura_Assert( allFuncsPresent, "Out of date version of the OpenAL DLLs detected. Run the latest OpenAL installer." );
	return allFuncsPresent;
}

/* ------------------------------------------------------------------------- */

Asura_OpenAL_ActiveSound_API::Asura_OpenAL_ActiveSound_API() :
	m_uBufferID					( 0 ),
	m_uSourceID					( 0 ),
	m_bIsSpatial				( false ),
	m_fInnerRadius				( 0.0f ),
	m_fOuterRadius				( 0.0f ),
	m_bUseReverb				( true ),
	m_fDryVolume				( 1.0f ),
	m_fWetVolume				( 1.0f ),
	m_fOcclusion				( 0.0f ),
	m_fOcclusionFilterGain		( 1.0f ),
	m_bNeedUpdateReverbParams	( true ),
	m_fDistanceFilterGain		( 0.0f ),
	m_fLastDistanceFilterGain	(fASURA_OPENAL_INVALID_FILTER_GAIN )
{
	for(u_int u = 0; u < ASURA_OPENAL_FILTER_ID_NUM_ITEMS; ++ u)
	{
		m_auFilterID[u] = AL_FILTER_NULL;
	}
	m_xPosition.Zero();
	m_xPreviousPosition.Zero();
}

/* ------------------------------------------------------------------------- */

Asura_OpenAL_ActiveSound_API::~Asura_OpenAL_ActiveSound_API()
{

}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::Initialise( bool bUseHardware3dPan, u_int uSize, const void* pvFormat, const u_char* pucData )
{
	ALenum iError = alGetError();	// Clear error code

	alGenBuffers( 1, &m_uBufferID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to gen buffer, error %d", iError ) );
		return false;
	}

	if ( !Platform_FillBuffer( m_uBufferID, pvFormat, pucData, uSize ) )
	{
		Destroy();
		return false;
	}

	alGenSources( 1, &m_uSourceID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to gen sources, error %d", iError ) );
		Destroy();
		return false;
	}

	alSourcei( m_uSourceID, AL_BUFFER, m_uBufferID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to attach buffer to source, error %d", iError ) );
		Destroy();
		return false;
	}

	// Per-voice filters
	if ( Asura_OpenAL_Sound_API::Platform_GetAPI().IsEFXAvailable() )
	{
		#ifdef ENABLE_PER_VOICE_FILTER
		alGenFilters( ASURA_OPENAL_FILTER_ID_NUM_ITEMS, m_auFilterID );

		if ( ( iError = alGetError() ) == AL_NO_ERROR )
		{
			for(u_int u = 0; u < ASURA_OPENAL_FILTER_ID_NUM_ITEMS; ++u)
			{
				Asura_Assert( ( m_auFilterID[u]  != AL_FILTER_NULL ) && alIsFilter( m_auFilterID[u]  ), "OpenAL failed to create filters correctly" );

				alFilteri( m_auFilterID[u], AL_FILTER_TYPE, AL_FILTER_LOWPASS );

				if ( ( iError = alGetError() ) == AL_NO_ERROR )
				{
					alFilterf( m_auFilterID[u], AL_LOWPASS_GAIN, 1.0f );
					alFilterf( m_auFilterID[u], AL_LOWPASS_GAINHF, 1.0f );
				}
				else
				{
					Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter values, error %d", iError ) );
				}
			}
		}
		else
		{
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to gen filters, error %d", iError ) );
		}

		//TY 18.12.07 - dont attach a direct filter at this stage, we need to wait to determine which we will use
		#else 
		for(u_int u = 0; u < ASURA_OPENAL_FILTER_ID_NUM_ITEMS; ++ u)
		{
			m_auFilterID[u] = AL_FILTER_NULL;
		}
		#endif
		
		// Air absorption. See p83 of the Effects Extension Guide.
		alSourcei( m_uSourceID, AL_AIR_ABSORPTION_FACTOR, s_iAirAbsorptionFactor );
		#ifdef ASURA_ASSERT
		iError = alGetError();
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set air absorption factor, error %d", iError ) );
		#endif
	}

	PRINT_SOUND_EVENT(( "Initialised Active Sound %x\n", m_uSourceID ));

	return true;
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_ActiveSound_API::Update()
{
	#ifdef ASURA_ASSERT
	ALenum iError = alGetError();	//Reset error code
	#endif

	if ( Asura_OpenAL_Sound_API::Platform_GetAPI().IsEFXAvailable() )
	{
		//Remember the highest gain so that we can select the appropriate filter
		float fDistFilterGain = fASURA_OPENAL_INVALID_FILTER_GAIN;
					
		//Update the distance filter
		if (	(m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] != AL_FILTER_NULL)	&&	//Have a distance filter
				(s_iAirAbsorptionFactor == 0) &&										//arent using built in air absorbtion 
				(m_bIsSpatial))															//and is spacial (can be false!)
		{
			fDistFilterGain = m_fDistanceFilterGain;
		}

		//Pick which filter to apply
		if( ( fDistFilterGain < m_fOcclusionFilterGain ) && ( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] != AL_FILTER_NULL ) )
		{
			//Modify the low frequency sound gain (this is the dry volume)
			alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE], AL_LOWPASS_GAIN, m_fDryVolume); 
			#ifdef ASURA_ASSERT
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter gain HF, error %x", iError ) );
			#endif

			//Update distance filter
			alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE], AL_LOWPASS_GAINHF, fDistFilterGain );
			#ifdef ASURA_ASSERT
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter gain HF, error %x", iError ) );
			#endif

			//Attach distance filter
			alSourcei( m_uSourceID, AL_DIRECT_FILTER, m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] );
		}
		else if ( m_auFilterID[ASURA_OPENAL_FILTER_ID_OCCLUSION] != AL_FILTER_NULL )
		{
			//Modify the low frequency sound gain (this is the dry volume)
			alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_OCCLUSION], AL_LOWPASS_GAIN, m_fDryVolume); 
			#ifdef ASURA_ASSERT
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter gain HF, error %x", iError ) );
			#endif

			//Update occlusion filter
			alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_OCCLUSION], AL_LOWPASS_GAINHF, m_fOcclusionFilterGain ); 
			#ifdef ASURA_ASSERT
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter gain HF, error %x", iError ) );
			#endif

			//attach occlusion filter
			alSourcei( m_uSourceID, AL_DIRECT_FILTER, m_auFilterID[ASURA_OPENAL_FILTER_ID_OCCLUSION] );
		}

		#ifdef ASURA_ASSERT
		iError = alGetError();
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter, error %x", iError ) );
		#endif

		if(m_bNeedUpdateReverbParams)
		{
			// Set reverb regardless of whether or not a filter exists for this sound
			//TY 20.12.07 - This is a very expensive operation so only do it if necessary
			SetReverbParams( Asura_OpenAL_Sound_API::Platform_GetAPI().GetReverbParams() );
			m_bNeedUpdateReverbParams = false;
		}
		
	}
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_ActiveSound_API::Destroy()
{
	PRINT_SOUND_EVENT(( "Destroying sound event %x\n", m_uSourceID ));

	for ( int i = 0; i < Asura_OpenAL_Sound_API::Platform_GetAPI().GetNumAuxSendsPerSource(); ++i )
	{
		alSource3i( m_uSourceID, AL_AUXILIARY_SEND_FILTER, AL_EFFECTSLOT_NULL, i, AL_FILTER_NULL );
	}

	alSourcei( m_uSourceID, AL_DIRECT_FILTER, AL_FILTER_NULL );

	#ifdef ASURA_ASSERT
	ALenum iError = alGetError(); //Reset error code
	#endif

	if ( Asura_OpenAL_Sound_API::Platform_GetAPI().IsEFXAvailable() )
	{
		for(u_int u = 0; u < ASURA_OPENAL_FILTER_ID_NUM_ITEMS; ++ u)
		{
			if( ( m_auFilterID[u] != AL_FILTER_NULL ) && alIsFilter( m_auFilterID[u] ) )
			{
				alDeleteFilters( 1, &m_auFilterID[u] );
				
				#ifdef ASURA_ASSERT
				iError = alGetError();
				Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to delete filter, error %x", iError ) );
				#endif
			}
			m_auFilterID[u] = AL_FILTER_NULL;
		}
	}

	alDeleteSources( 1, &m_uSourceID );
	m_uSourceID = 0;

	#ifdef ASURA_ASSERT
	iError = alGetError();
	Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to delete source, error %x", iError ) );
	#endif

	alDeleteBuffers( 1, &m_uBufferID );
	m_uBufferID = 0;

	#ifdef ASURA_ASSERT
	iError = alGetError();
	Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to delete buffer, error %x", iError ) );
	#endif

	m_xPosition.Zero();
	m_xPreviousPosition.Zero();

	m_fLastDistanceFilterGain = fASURA_OPENAL_INVALID_FILTER_GAIN;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::IsValid() const
{
	return ( ( m_uBufferID != 0 ) && ( m_uSourceID != 0 ) );
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::IsPlaying() const
{
	ALenum iError = alGetError();	// Clear error state

	ALint state;
	alGetSourcei( m_uSourceID, AL_SOURCE_STATE, &state );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to get source state, error %x", iError ) );
		return false;
	}

	return state == AL_PLAYING;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::IsPaused() const
{
	ALenum iError = alGetError();	// Clear error state

	ALint state;
	alGetSourcei( m_uSourceID, AL_SOURCE_STATE, &state );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to get source state, error %x", iError ) );
		return false;
	}

	return state == AL_PAUSED;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::Play( bool bIsLooping )
{
	ALenum iError = alGetError();	// Clear error state

	alSourcei( m_uSourceID, AL_LOOPING, bIsLooping );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set source as looping, error %x", iError ) );
		return false;
	}

	Update();

	alSourcePlay( m_uSourceID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to play source, error %x", iError ) );
		return false;
	}

	PRINT_SOUND_EVENT(( "Played sound %x\n", m_uSourceID ));

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::Pause()
{
	ALenum iError = alGetError();	// Clear error state

	alSourcePause( m_uSourceID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to pause source, error %x", iError ) );
		return false;
	}

	PRINT_SOUND_EVENT(( "Paused sound %x\n", m_uSourceID ));

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::Unpause( bool bIsLooping )
{
	ALenum iError = alGetError();	// Clear error state

	alSourcei( m_uSourceID, AL_LOOPING, bIsLooping );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set source as looping, error %x", iError ) );
		return false;
	}

	alSourcePlay( m_uSourceID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to unpause source, error %x", iError ) );
		return false;
	}

	PRINT_SOUND_EVENT(( "Unpaused sound %x\n", m_uSourceID ));

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::Stop()
{
	ALenum iError = alGetError();	// Clear error state

	alSourceStop( m_uSourceID );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to stop source, error %x", iError ) );
		return false;
	}

	PRINT_SOUND_EVENT(( "Stopped sound %x\n", m_uSourceID ));

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetVolume( float fVolume )
{
	ALenum iError = alGetError();	// Clear error state
	Asura_Assert( fVolume >= 0.0f && fVolume <= 1.0f, "Volume out of range, should already have been clamped?" );
	fVolume = Asura_Maths::ClampToRange( fVolume, 0.0f, 1.0f );

	alSourcef( m_uSourceID, AL_GAIN, fVolume);

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set volume, error %x", iError ) );
		return false;
	}

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetSpatial( bool bIsSpatial )
{
	ALenum iError = alGetError();	// Clear error state

	alSourcei( m_uSourceID, AL_SOURCE_RELATIVE, bIsSpatial ? AL_FALSE : AL_TRUE );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set sound as source relative, error %x", iError ) );
		return false;
	}

	m_bIsSpatial = bIsSpatial;

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetPosition( const Asura_Vector_3& xPosition )
{
	ALenum iError = alGetError();	// Clear error state

	if ( SetSpatial( true ) )
	{
		m_xPreviousPosition = m_xPosition;
		m_xPosition = xPosition;

		if(m_bUseReverb && (m_xPosition != m_xPreviousPosition))
		{
			alSource3f( m_uSourceID, AL_POSITION, m_xPosition.x, m_xPosition.y, m_xPosition.z );
			m_bNeedUpdateReverbParams = true;
		}
		
		if ( ( iError = alGetError() ) != AL_NO_ERROR )
		{
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set position, error %x", iError ) );
			return false;
		}

		return true;
	}

	return false;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetRadii( float fInner, float fOuter )
{
	// NOTE distance volume attenuation is done by the game. These values are used
	// only to modify the distance filter

	m_fInnerRadius = fInner;
	m_fOuterRadius = fOuter;

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetPan( float fPan )
{
	ALenum iError = alGetError();	// Clear error state

	if ( SetSpatial( false ) )
	{
		// This is a little experimental, I don't know how well it will work,
		// might have to come back later and reconsider this
		float posX = ( fPan - 0.5f ) * 2.0f;
		alSource3f( m_uSourceID, AL_POSITION, posX, 0.0f, 0.0f );

		if ( ( iError = alGetError() ) != AL_NO_ERROR )
		{
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set position, error %x", iError ) );
			return false;
		}

		return true;
	}

	return false;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetForcedPanAndFade( float fPan, float fFade )
{
	ALenum iError = alGetError();	// Clear error state

	if ( SetSpatial( false ) )
	{
		// This is a little experimental, I don't know how well it will work,
		// might have to come back later and reconsider this
		const float posX = ( fPan - 0.5f ) * 2.0f;
		// Need to test this in earnest on surround sound setup - not sure about posZ
		const float posZ = -( fFade - 0.5f ) * 2.0f;
		alSource3f( m_uSourceID, AL_POSITION, posX, 0.0f, posZ );

		if ( ( iError = alGetError() ) != AL_NO_ERROR )
		{
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set position, error %x", iError ) );
			return false;
		}

		return true;
	}

	return false;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetPitch( float fPitch )
{
	ALenum iError = alGetError();	// Clear error state

	alSourcef( m_uSourceID, AL_PITCH, fPitch );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set pitch, error %x", iError ) );
		return false;
	}

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetCone( const Asura_Vector_3& xDirection, float fInnerAngle, float fOuterAngle, float fVolumeOutsideCone )
{
	ALenum iError = alGetError();	// Clear error state

	alSourcef( m_uSourceID, AL_CONE_INNER_ANGLE, fInnerAngle );	// NOTE degrees

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set inner cone angle, error %x", iError ) );
		return false;
	}

	alSourcef( m_uSourceID, AL_CONE_OUTER_ANGLE, fOuterAngle );	// NOTE degrees

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set outer cone angle, error %x", iError ) );
		return false;
	}

	alSourcef( m_uSourceID, AL_CONE_OUTER_GAIN, fVolumeOutsideCone );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set outer cone volume, error %x", iError ) );
		return false;
	}

	return true;
}

/* ------------------------------------------------------------------------- */

#if _DEBUG
void Asura_OpenAL_ActiveSound_API::CheckStatus() const
{
	Asura_Assert( IsValid(), "Sound does not appear to be valid?" );
}
#endif

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_ActiveSound_API::SetReverbParams( const Asura_OpenAL_ReverbParameters& xReverbParams )
{
	//This function attaches a source to an effect slot, the effect slots are not dedicated to sources

	if(Asura_OpenAL_Sound_API::Platform_GetAPI().IsEFXAvailable())
	{
		#ifdef ASURA_ASSERT
			ALenum iError = alGetError();
		#endif

		if ( m_bUseReverb )
		{
			//Get the distance filter value (if we have a distance filer) and adjust it with our current wet path volume
			if( (m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] != AL_FILTER_NULL) && m_bIsSpatial)
			{
				float fDistFilterGain = 0.0f;
				Asura_Assert(alIsFilter( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE]), "Invalid filter");
				
				alGetFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE], AL_LOWPASS_GAINHF, &fDistFilterGain );
				#ifdef ASURA_ASSERT
					iError = alGetError();
					Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to get filter properties, error %x", iError ) );
				#endif

				fDistFilterGain *= (m_fWetVolume * m_fWetVolume); //use square of volume

				alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_WET], AL_LOWPASS_GAINHF, fDistFilterGain );
				#ifdef ASURA_ASSERT
					iError = alGetError();
					Asura_Assertf( iError == AL_NO_ERROR, ( "Setting filter properties failed, error %x", iError ) );
				#endif
			}

			if ( Asura_OpenAL_Sound_API::Platform_GetAPI().GetNumAuxSendsPerSource() == 1 )
			{
				Asura_Assert(alIsAuxiliaryEffectSlot(xReverbParams.uEffectSlotCurrent), "Passed an invalid effect slot");
				Asura_Assert(alIsFilter( m_auFilterID[ASURA_OPENAL_FILTER_ID_WET]), "Invalid filter");
				Asura_Assert(alIsSource(m_uSourceID), "Invalid source" );
				// There is only one reverb send available, therefore, we can't crossfade when it
				// changes - it might click but there's nothing we can do to stop that, save perhaps
				// fade to zero reverb and back, which might be unnecessary. (Interpolating the values
				// of the reverb definitely won't help, don't even bother trying).
				#ifdef ENABLE_AUXILIARY_SEND_FILTER
					alSource3i( m_uSourceID, AL_AUXILIARY_SEND_FILTER, xReverbParams.uEffectSlotCurrent, 0, m_auFilterID[ASURA_OPENAL_FILTER_ID_WET] );
					#ifdef ASURA_ASSERT
						iError = alGetError();
						Asura_Assertf( iError == AL_NO_ERROR, ( "Setting filter send failed, error %x", iError ) );
					#endif
				#endif

			}
			else if ( Asura_OpenAL_Sound_API::Platform_GetAPI().GetNumAuxSendsPerSource() >= 1 )
			{
				Asura_Assert(alIsAuxiliaryEffectSlot(xReverbParams.uEffectSlotCurrent), "Passed an invalid effect slot (current)");
				Asura_Assert(alIsAuxiliaryEffectSlot(xReverbParams.uEffectSlotPrevious), "Passed an invalid effect slot (previous)");
				Asura_Assert(alIsFilter( m_auFilterID[ASURA_OPENAL_FILTER_ID_WET] ), "Invalid filter");
				Asura_Assert(alIsSource(m_uSourceID), "Invalid source" );
				// With more than one reverb send available, we can crossfade between them to avoid clicks.
				// The actual fading is controlled by the auxiliary slot manager, we just send to both (even
				// though they'll only both be active during a fade).
				#ifdef ENABLE_AUXILIARY_SEND_FILTER
					alSource3i( m_uSourceID, AL_AUXILIARY_SEND_FILTER, xReverbParams.uEffectSlotCurrent, 0, m_auFilterID[ASURA_OPENAL_FILTER_ID_WET] );
					#ifdef ASURA_ASSERT
						iError = alGetError();
						Asura_Assertf( iError == AL_NO_ERROR, ( "Setting filter send (current) failed, error %x", iError ) );
					#endif
					alSource3i( m_uSourceID, AL_AUXILIARY_SEND_FILTER, xReverbParams.uEffectSlotPrevious, 1, m_auFilterID[ASURA_OPENAL_FILTER_ID_WET] );
					#ifdef ASURA_ASSERT
						iError = alGetError();
						Asura_Assertf( iError == AL_NO_ERROR, ( "Setting filter send (previous) failed, error %x", iError ) );
					#endif
				#endif
			}
		}
		else
		{
			// Disable reverb
			for ( int i = 0; i < Asura_OpenAL_Sound_API::Platform_GetAPI().GetNumAuxSendsPerSource(); ++i )
			{
				alSource3i( m_uSourceID, AL_AUXILIARY_SEND_FILTER, AL_EFFECTSLOT_NULL, i, AL_FILTER_NULL );
				#ifdef ASURA_ASSERT
					iError = alGetError();
					Asura_Assertf( iError == AL_NO_ERROR, ( "Disabling filter send failed, error %x", iError ) );
				#endif
			}
		}
	}
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetUseDistanceFilter( bool bUseDistanceFilter )
{
	if ( Asura_OpenAL_Sound_API::Platform_GetAPI().IsEFXAvailable() )
	{
		ALenum iError = alGetError();

		if ( bUseDistanceFilter )
		{
			if ( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] == AL_FILTER_NULL )
			{
				alGenFilters( 1, &m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] );

				if ( ( iError = alGetError() ) == AL_NO_ERROR )
				{
					Asura_Assert( ( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] != AL_FILTER_NULL ) && alIsFilter( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] ) , "You have to be kidding me..." );
					alFilteri( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE], AL_FILTER_TYPE, AL_FILTER_LOWPASS );

					if ( ( iError = alGetError() ) == AL_NO_ERROR )
					{
						alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE], AL_LOWPASS_GAIN, 1.0f );
						alFilterf( m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE], AL_LOWPASS_GAINHF, 1.0f );
					}
					else
					{
						Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set filter values, error %x", iError ) );
					}
				}
				else
				{
					Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to gen filters, error %x", iError ) );
					return false;
				}
			}
		}
		else
		{
			// Disable any distance filter that is already set
			alDeleteFilters( 1, &m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] );
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to delete filters, error %x", iError ) );
			m_auFilterID[ASURA_OPENAL_FILTER_ID_DISTANCE] = AL_FILTER_NULL;
		}

		//Dont reattach the filter here, it will be done in Update
		m_bNeedUpdateReverbParams = true;
	}

	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetOcclusionLPF( const float fOcclusion )
{
	Asura_Assertf( ( fOcclusion >= 0.0f ) && ( fOcclusion <= 1.0f ), ("Occlusion factor is invalid %.2f", fOcclusion) );
	if ( m_auFilterID[ASURA_OPENAL_FILTER_ID_OCCLUSION] != AL_FILTER_NULL) //filter is valid
	{
		m_fOcclusion = fOcclusion;
		
		//OpenAL filters have a particularily aggressive falloff, so take the square
		float fGain = 1.0f - fOcclusion;
		fGain *= fGain;
		//scale with the dry volume (linearly?)
		fGain *= m_fDryVolume;

		#ifdef CONSOLE_VARNAME_AUDIO_OCCLUSION_SCALE
		//Use the scale to reduce the effect of the filter (i.e. move it closer to 1)
		fGain = 1.0f - ((1.0f - fGain) * g_fOcclusionFilterScale);
		#endif
		
		m_fOcclusionFilterGain = fGain;

		//We dont create or apply the filter at this stage, it will be applied during the sound update
	}
	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetWetDryVolume(const float fWetVolume, const float fDryVolume)
{
	if(m_bUseReverb && (m_fWetVolume != fWetVolume))
	{
		m_bNeedUpdateReverbParams = true;
	}

	if(fWetVolume != m_fWetVolume)
	{
		m_bNeedUpdateReverbParams = true;
		m_fWetVolume = fWetVolume;
	}
	m_fDryVolume = fDryVolume;
	
	//Unfortunately we cant modify the dry path directly, since we can only change the volume of
	//the source before it is sent to the aux effect slot, and this would affect the wet volume
	//Instead we use the dry volume to modify the gain of the occlusion LPF
	SetOcclusionLPF(m_fOcclusion);	//Call this to recalc the filter gain

	//The wet path volume will automatically affect the wet path filter to the aux effect slots
	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::SetUseReverb( bool bUseReverb )
{
	m_bUseReverb = bUseReverb;
	m_bNeedUpdateReverbParams = true;
	return true;
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_ActiveSound_API::SetDistanceFilterGain (const float fGain)
{
	if(fGain != m_fLastDistanceFilterGain)
	{
		#ifdef CONSOLE_VARNAME_AUDIO_DISTANCE_SCALE
		m_fDistanceFilterGain = (1.0f - (fGain *  g_fDistanceFilterScale));
		#else
		m_fDistanceFilterGain = (1.0f - fGain);
		#endif
		//Compensate for the slightly strange gain scale of openAL filters
		for(u_int u = 1; u < uOPENAL_DISTANCE_FILTER_GAIN_POWER; ++u)
		{
			m_fDistanceFilterGain *= m_fDistanceFilterGain;
		}
		m_bNeedUpdateReverbParams = true;
		m_fLastDistanceFilterGain = fGain;
	}
}

/* ------------------------------------------------------------------------- */
// Asura_OpenAL_Sound_API
/* ------------------------------------------------------------------------- */

const float Asura_OpenAL_Sound_API::ASURA_OPENAL_REVERB_CROSSFADE_TIME = 0.5f;

Asura_OpenAL_Sound_API::Asura_OpenAL_Sound_API() :
	m_pxDevice( NULL ), 
	m_pxContext( NULL ),
	m_bEFXAvailable( false ),
	m_iAuxSendsPerSource( 0 ),
	m_fReverbCrossfadeTime( 0.0f ),
	m_fReverbCrossfadeProportion( 0.0f ),
	m_fCurrentReverbAmount( 1.0f ),
	m_eCurrentReverb( Asura_Sound_System::ASURA_REVERB_ID_MAX ),
	m_iEffectSlotIndexCurrent( 0 ),
	m_iEffectSlotIndexPrevious( 1 )
	#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
		, m_bManualReverbControl( false )
	#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
{
	m_xReverbParameters.uEffectSlotPrevious = AL_EFFECTSLOT_NULL;
	m_xReverbParameters.uEffectSlotCurrent = AL_EFFECTSLOT_NULL;
	m_xListenerPos.Zero();
}

/* ------------------------------------------------------------------------- */

Asura_OpenAL_Sound_API::~Asura_OpenAL_Sound_API()
{
	Asura_Assert( !m_pxContext, "Context not destroyed!" );
	Asura_Assert( !m_pxDevice, "Device not closed!" );
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_Sound_API::Initialise()
{
#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_AUDIO_DISTANCE_SCALE,  &g_fDistanceFilterScale,  0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( CONSOLE_VARNAME_AUDIO_OCCLUSION_SCALE, &g_fOcclusionFilterScale, 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif	

#define ASURA_OPENAL_DEBUG_INIT 0

	Platform_Initialise();
	
#if defined(ASURA_OPENAL_FORCE_SOFTWARE_DEVICE)
	const ALCchar* szDeviceName	= "Generic Software";
#else
	const ALCchar* szDeviceName = NULL;

	// Discover what devices are available
	int iNumDevices = 0;
	if ( alcIsExtensionPresent( NULL, "ALC_ENUMERATION_EXT" ) == AL_TRUE )
	{
		const ALCchar* pcDevices = alcGetString( NULL, ALC_DEVICE_SPECIFIER );
		// devices contains the device names, separated by NULL and terminated by two consecutive NULLs

		if ( pcDevices )
		{
			// Loop through the list, just so we can step through in the debugger and see what's there
			const ALCchar* pcNextDevice = pcDevices;

			while ( *pcNextDevice )
			{
				#if ASURA_OPENAL_DEBUG_INIT == 1
					Asura_Assertf( false, ( "OpenAL device %d '%s'\n", iNumDevices, pcNextDevice ) );
				#endif // #if ASURA_OPENAL_DEBUG_INIT == 1
				iNumDevices++;
				PRINT_SOUND_EVENT(( "Sound device found: %s\n", pcNextDevice ));
				pcNextDevice += strlen( pcNextDevice ) + 1;
			}

			// TODO offer the user the ability to choose a non-default sound device from the list we
			// just enumerated. Then, you'd assign szDeviceName to the name of the chosen device, here.
		}
	}

	#if ASURA_OPENAL_DEBUG_INIT == 1
		Asura_Assertf( false, ( "OpenAL numDevices %d\n", iNumDevices ) );
	#endif // #if ASURA_OPENAL_DEBUG_INIT == 1
#endif	// defined(ASURA_OPENAL_FORCE_SOFTWARE_DEVICE)

	m_pxDevice = alcOpenDevice( szDeviceName );

	if ( !m_pxDevice )
	{
		Asura_Assert( m_pxDevice, "Failed to get OpenAL device!" );
		return false;
	}

	#if defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && !defined( NDEBUG )
		const ALCchar* szChosenDevice = alcGetString( m_pxDevice, ALC_DEVICE_SPECIFIER );
		PRINT_SOUND_EVENT(( "Using sound device %s\n", szChosenDevice ? szChosenDevice : "???" ));
	#endif

	#if ASURA_OPENAL_DEBUG_INIT == 1
		const ALCchar* szChosenDevice1 = alcGetString( m_pxDevice, ALC_DEVICE_SPECIFIER );
		Asura_Assertf( false, ( "Chosen Device:%s", szChosenDevice1 ) );
	#endif // #if ASURA_OPENAL_DEBUG_INIT == 1

	ALint aiAttribs[4] = { 0 };

	// Check for EFX extensions; not a fatal failure if not available
	if ( alcIsExtensionPresent( m_pxDevice, "ALC_EXT_EFX" ) != AL_FALSE )
	{
		aiAttribs[0] = ALC_MAX_AUXILIARY_SENDS;
		aiAttribs[1] = ASURA_OPENAL_REQUEST_AUX_SENDS_NUM;
		m_bEFXAvailable = true;
	}
	
	m_pxContext = alcCreateContext( m_pxDevice, m_bEFXAvailable ? aiAttribs : NULL );

	if ( !m_pxContext )
	{
		Asura_Assert( m_pxContext, "Failed to create OpenAL context!" );
		alcCloseDevice( m_pxDevice );
		m_pxDevice = NULL;
		return false;
	}
	
	alcMakeContextCurrent( m_pxContext );

	ALenum xError = alGetError();	// Do not call alGetError before this, it will always report an error
	Asura_Assertf( xError == AL_NO_ERROR, ( "Error previous to alcMakeContextCurrent, xError %x", xError ) );

	// Confirm EFX is genuinely available and get what we need to use it
	if ( m_bEFXAvailable )
	{
		m_bEFXAvailable = CacheEFXFunctions();

		// Get the actual number of auxiliary sends available
		// JAT 18/10/07 let the h/w platform specify this value instead of OpenAL
		m_iAuxSendsPerSource = Platform_GetNumAuxSendsPerSource();
	}
	
	// Our game does all the distance attenuation and doppler calculations, we don't need OpenAL to do it
	alDistanceModel( AL_NONE );
	alDopplerFactor( 0.0f );

	if ( ( xError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( false, ( "Failed to set distance model! %d", xError ) );
		return false;
	}

	if ( m_bEFXAvailable )
	{
		for ( int i = 0; i < m_iAuxSendsPerSource; ++i )
		{
			alGenAuxiliaryEffectSlots( 1, &m_uEffectSlots[i] );

			if ( ( xError = alGetError() ) != AL_NO_ERROR )
			{
				Asura_Assert( xError == AL_NO_ERROR, "Failed to generate an auxiliary effect slot that should have been possible!" );
				m_bEFXAvailable = false;
				break;
			}
		}

		if ( m_bEFXAvailable )
		{	
			for ( int i = 0; i < m_iAuxSendsPerSource; ++i )
			{
				alGenEffects( 1, &m_uEffects[i] );

				if ( ( xError = alGetError() ) == AL_NO_ERROR )
				{
					SetReverbProperties( i, Asura_Sound_System::ASURA_REVERB_ID_NONE, 1.0f );
				}
				else
				{
					Asura_Assert( xError == AL_NO_ERROR, "Failed to generate an effect that should have been possible!" );
					m_bEFXAvailable = false;
					break;
				}
			}

			SwitchToReverb( Asura_Sound_System::ASURA_REVERB_ID_NONE, 1.0f );
		}
	}

	#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
		Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_SETREVERB, ConsoleCmd_SetReverb, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_AUDIO_DIAGNOSTICS, ConsoleCmd_AudioDiagnostics, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( CONSOLE_VARNAME_PRINT_EVENTS, &s_bPrintEvents, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( CONSOLE_VARNAME_AIR_ABSORPTION, &s_iAirAbsorptionFactor, 0, 10, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )

	PRINT_SOUND_EVENT(( "Sound System (OpenAL) Initialised\n" ));
	PRINT_SOUND_EVENT(( "EFX (filter, reverb etc) are%s available\n", m_bEFXAvailable ? "" : " NOT" ));

	// Immediately update so that reverb settings etc. are committed from the beginning
	Update();
	
	return true;
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_Sound_API::Update()
{
	#if ASURA_ASSERT
		ALenum xError = alGetError();
	#endif
		
	if ( m_bEFXAvailable )
	{
		// Update reverb crossfades
		if ( m_fReverbCrossfadeTime > 0.0f )
		{
			m_fReverbCrossfadeProportion = m_fReverbCrossfadeTime / ASURA_OPENAL_REVERB_CROSSFADE_TIME;
			m_fReverbCrossfadeProportion = Asura_Maths::ClampToRange( m_fReverbCrossfadeProportion, 0.0f, 1.0f );
			
			alAuxiliaryEffectSlotf( m_uEffectSlots[m_iEffectSlotIndexPrevious], AL_EFFECTSLOT_GAIN, m_fReverbCrossfadeProportion );
			alAuxiliaryEffectSlotf( m_uEffectSlots[m_iEffectSlotIndexCurrent], AL_EFFECTSLOT_GAIN, ( 1.0f - m_fReverbCrossfadeProportion ) );
			
			// Countdown the timer
			m_fReverbCrossfadeTime -= Asura_Timers::GetPreviousFrameTime();
		}
		else
		{
			// No crossfade to worry about, just ensure the volumes are as expected
			if ( m_uEffectSlots[m_iEffectSlotIndexPrevious] != AL_EFFECTSLOT_NULL )
			{
				alAuxiliaryEffectSlotf( m_uEffectSlots[m_iEffectSlotIndexPrevious], AL_EFFECTSLOT_GAIN, 0.0f );
			}

			alAuxiliaryEffectSlotf( m_uEffectSlots[m_iEffectSlotIndexCurrent], AL_EFFECTSLOT_GAIN, 1.0f );
		}

		Asura_Assertf( ( xError = alGetError() )== AL_NO_ERROR, ( "Failed to set auxiliary effect slot, xError %x", xError ) );

		m_xReverbParameters.uEffectSlotPrevious = m_uEffectSlots[m_iEffectSlotIndexPrevious];
		m_xReverbParameters.uEffectSlotCurrent = m_uEffectSlots[m_iEffectSlotIndexCurrent];
	}
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_Sound_API::Destroy()
{
	ALenum iError = alGetError();

	if(m_bEFXAvailable)
	{
		for ( int i = 0; i < m_iAuxSendsPerSource; ++i )
		{
			alDeleteEffects( 1, &m_uEffects[i] );
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to delete an effect, error %x", iError ) );
		}
		
		for ( int i = 0; i < m_iAuxSendsPerSource; ++i )
		{
			alDeleteAuxiliaryEffectSlots( 1, &m_uEffectSlots[i] );
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to delete an auxiliary effect slot, error %x", iError ) );
		}
	}

	alcMakeContextCurrent( NULL );
	alcDestroyContext( m_pxContext );
	m_pxContext = NULL;
	alcCloseDevice( m_pxDevice );
	m_pxDevice = NULL;

	PRINT_SOUND_EVENT(( "Sound System (OpenAL) Shut Down\n" ));
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_Sound_API::IsValid() const
{
	return ( m_pxDevice && m_pxContext );
}

/* ------------------------------------------------------------------------- */
	
void Asura_OpenAL_Sound_API::SetListener( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation )
{
	ALenum iError = alGetError();

	m_xListenerPos = xPosition;

	ALfloat listenerPos[] = { xPosition.x, xPosition.y, xPosition.z };
	ALfloat listenerOrientation[] = {	xOrientation.e13, xOrientation.e23, xOrientation.e33,			// Forward
										-xOrientation.e12, -xOrientation.e22, -xOrientation.e32 };		// Up

	alListenerfv( AL_POSITION, listenerPos );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( false, ( "Failed to set listener position! %d", iError ) );
		return;
	}

	alListenerfv( AL_ORIENTATION, listenerOrientation );

	if ( ( iError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( false, ( "Failed to set listener orientation! %d", iError ) );
		return;
	}
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_Sound_API::ForceListenerUpdate()
{
	// Not required in OpenAL
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_Sound_API::SetReverb( Asura_Sound_System::ReverbType eType, float fReverbAmount )
{
	#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
	if ( !m_bManualReverbControl )
	#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
	{
		SwitchToReverb( eType, fReverbAmount );
	}
	#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
	else
	{
		m_eOldReverb = eType;
		m_fOldReverbAmount = fReverbAmount;
	}
	#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
}

/* ------------------------------------------------------------------------- */

#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
void Asura_OpenAL_Sound_API::ForceReverb( int iReverbID )
{
	if ( iReverbID == -1 )
	{
		m_bManualReverbControl = false;
		SwitchToReverb( m_eOldReverb, m_fOldReverbAmount );
	}
	else if ( iReverbID >= Asura_Sound_System::ASURA_REVERB_ID_NONE && iReverbID < Asura_Sound_System::ASURA_REVERB_ID_MAX )
	{
		if ( !m_bManualReverbControl )	// Otherwise repeatedly forcing reverb forgets the old reverb
		{
			m_eOldReverb = m_eCurrentReverb;
			m_fOldReverbAmount = m_fCurrentReverbAmount;
		}

		m_bManualReverbControl = true;
		m_eForcedReverb = static_cast<Asura_Sound_System::ReverbType>( iReverbID );
		SwitchToReverb( m_eForcedReverb, 1.0f );
	}
}
#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_Sound_API::SwitchToReverb( Asura_Sound_System::ReverbType eReverbID, float fReverbAmount )
{
	// Would rather have this next to the table but it doesn't seem to like living outside a function?
	ASURA_STATIC_ASSERT( ( sizeof(s_axReverbPresets) / sizeof(s_axReverbPresets[0]) ) == Asura_Sound_System::ASURA_REVERB_ID_MAX );

	const float fAmountDiff = Asura_Maths::Abs( m_fCurrentReverbAmount - fReverbAmount );

	if ( m_eCurrentReverb != eReverbID || fAmountDiff > 0.01f )
	{
		// When possible, we will crossfade between reverbs
		if ( m_iAuxSendsPerSource > 1 )
		{
			// TODO deal with the case where we are already crossfading
			// - if we're fading back to the same kind of reverb as we've just come from,
			//	 reverse the fade and reset the time only to what has elapsed (not a full reset)
			// - otherwise, might just have to accept there may be clicks...

			m_fReverbCrossfadeTime = ASURA_OPENAL_REVERB_CROSSFADE_TIME;
			int temp = m_iEffectSlotIndexCurrent;
			m_iEffectSlotIndexCurrent = m_iEffectSlotIndexPrevious;
			m_iEffectSlotIndexPrevious = temp;
		}

		m_eCurrentReverb = eReverbID;
		m_fCurrentReverbAmount = fReverbAmount;
		return SetReverbProperties( m_iEffectSlotIndexCurrent, eReverbID, fReverbAmount );
	}
	return true;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_Sound_API::SetReverbProperties( int index, Asura_Sound_System::ReverbType eReverbID, float fReverbAmount )
{
	// XXX assumes effects and effects slots have a 1:1 relationship, this 
	// may not always be true in future...
	ALuint uEffectSlotID = m_uEffectSlots[index];
	if ( m_bEFXAvailable )
	{
		ALenum iError = alGetError();

		if ( eReverbID != Asura_Sound_System::ASURA_REVERB_ID_NONE )
		{
			// Get the values we need for the OpenAL view of the reverb
			const EFXEAXREVERBPROPERTIES& efxProperties = s_axReverbPresets[eReverbID].m_xReverbProperties;
			ALuint uEffectID = m_uEffects[index];

			// Attempt to use an EAX reverb first as it is more advanced
			alEffecti( uEffectID, AL_EFFECT_TYPE, AL_EFFECT_EAXREVERB );

			if (( iError = alGetError() ) == AL_NO_ERROR )
			{
				alEffectf( uEffectID, AL_EAXREVERB_DENSITY, efxProperties.flDensity );
				alEffectf( uEffectID, AL_EAXREVERB_DIFFUSION, efxProperties.flDiffusion );
				alEffectf( uEffectID, AL_EAXREVERB_GAIN, efxProperties.flGain * fReverbAmount );
				alEffectf( uEffectID, AL_EAXREVERB_GAINHF, efxProperties.flGainHF );
				alEffectf( uEffectID, AL_EAXREVERB_GAINLF, efxProperties.flGainLF );
				alEffectf( uEffectID, AL_EAXREVERB_DECAY_TIME, efxProperties.flDecayTime );
				alEffectf( uEffectID, AL_EAXREVERB_DECAY_HFRATIO, efxProperties.flDecayHFRatio );
				alEffectf( uEffectID, AL_EAXREVERB_DECAY_LFRATIO, efxProperties.flDecayLFRatio );
				alEffectf( uEffectID, AL_EAXREVERB_REFLECTIONS_GAIN, efxProperties.flReflectionsGain );
				alEffectf( uEffectID, AL_EAXREVERB_REFLECTIONS_DELAY, efxProperties.flReflectionsDelay );
				alEffectfv( uEffectID, AL_EAXREVERB_REFLECTIONS_PAN, const_cast<ALfloat*>( efxProperties.flReflectionsPan ) );
				alEffectf( uEffectID, AL_EAXREVERB_LATE_REVERB_GAIN, efxProperties.flLateReverbGain );
				alEffectf( uEffectID, AL_EAXREVERB_LATE_REVERB_DELAY, efxProperties.flLateReverbDelay );
				alEffectfv( uEffectID, AL_EAXREVERB_LATE_REVERB_PAN, const_cast<ALfloat*>( efxProperties.flLateReverbPan ) );
				alEffectf( uEffectID, AL_EAXREVERB_ECHO_TIME, efxProperties.flEchoTime );
				alEffectf( uEffectID, AL_EAXREVERB_ECHO_DEPTH, efxProperties.flEchoDepth );
				alEffectf( uEffectID, AL_EAXREVERB_MODULATION_TIME, efxProperties.flModulationTime );
				alEffectf( uEffectID, AL_EAXREVERB_MODULATION_DEPTH, efxProperties.flModulationDepth );
				alEffectf( uEffectID, AL_EAXREVERB_AIR_ABSORPTION_GAINHF, efxProperties.flAirAbsorptionGainHF );
				alEffectf( uEffectID, AL_EAXREVERB_HFREFERENCE, efxProperties.flHFReference );
				alEffectf( uEffectID, AL_EAXREVERB_LFREFERENCE, efxProperties.flLFReference );
				alEffectf( uEffectID, AL_EAXREVERB_ROOM_ROLLOFF_FACTOR, efxProperties.flRoomRolloffFactor );
				alEffecti( uEffectID, AL_EAXREVERB_DECAY_HFLIMIT, efxProperties.iDecayHFLimit );
				#ifdef ASURA_ASSERT
				iError = alGetError();
				Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set EAX EFX reverb properties, error %x", iError ) );
				#endif
			}
			else
			{
				// EAX reverbs not supported - try a standard reverb instead
				alEffecti( uEffectID, AL_EFFECT_TYPE, AL_EFFECT_REVERB );

				if ( ( iError = alGetError() ) == AL_NO_ERROR )
				{
					alEffectf( uEffectID, AL_REVERB_DENSITY, efxProperties.flDensity );
					alEffectf( uEffectID, AL_REVERB_DIFFUSION, efxProperties.flDiffusion);
					alEffectf( uEffectID, AL_REVERB_GAIN, efxProperties.flGain * fReverbAmount );
					alEffectf( uEffectID, AL_REVERB_GAINHF, efxProperties.flGainHF );
					alEffectf( uEffectID, AL_REVERB_DECAY_HFRATIO, efxProperties.flDecayHFRatio );
					alEffectf( uEffectID, AL_REVERB_REFLECTIONS_GAIN, efxProperties.flReflectionsGain );
					alEffectf( uEffectID, AL_REVERB_REFLECTIONS_DELAY, efxProperties.flReflectionsDelay );
					alEffectf( uEffectID, AL_REVERB_LATE_REVERB_GAIN, efxProperties.flLateReverbGain );
					alEffectf( uEffectID, AL_REVERB_LATE_REVERB_DELAY, efxProperties.flLateReverbDelay );
					alEffectf( uEffectID, AL_REVERB_AIR_ABSORPTION_GAINHF, efxProperties.flAirAbsorptionGainHF );
					alEffectf( uEffectID, AL_REVERB_ROOM_ROLLOFF_FACTOR, efxProperties.flRoomRolloffFactor );
					alEffecti( uEffectID, AL_REVERB_DECAY_HFLIMIT, efxProperties.iDecayHFLimit );
					#ifdef ASURA_ASSERT
					iError = alGetError();
					Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set EAX reverb properties, error %x", iError ) );
					#endif
				}
				else
				{
					return false;
				}
			}

			alAuxiliaryEffectSloti( uEffectSlotID, AL_EFFECTSLOT_EFFECT, uEffectID );
			#ifdef ASURA_ASSERT
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set effect slot, error %x", iError ) );
			#endif
		}
		else
		{
			// Clears any set reverb
			alAuxiliaryEffectSloti( uEffectSlotID, AL_EFFECTSLOT_EFFECT, AL_EFFECT_NULL );
			#ifdef ASURA_ASSERT
			iError = alGetError();
			Asura_Assertf( iError == AL_NO_ERROR, ( "Failed to set effect slot to NULL, error %x", iError ) );
			#endif
		}
	}
	else
	{
		return false;
	}

	PRINT_SOUND_EVENT(( "Sound System Reverb set to %d\n", eReverbID ));

	return true;
}

/* ------------------------------------------------------------------------- */

#endif	//defined( ASURA_USE_OPENAL ) && ( ASURA_USE_OPENAL == 1 ) && !defined(ASURA_USE_AUDIO_SYSTEM)

