//
// Asura_Phonon.h
//
// Created by Kevin Lea - 12/06/2001 11:12:57
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_PHONON_H__
#define __ASURA_PHONON_H__

#include "Core/Asura_Handle.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_OrientedVolume.h"
#ifdef ASURA_TOOLS
#include "Core/Asura_String.h" //for storing tools phonon name
#endif
#include "Network/Asura_Guid.h"
#include "Sound/Asura_Sound_Event.h"

#if !defined(ASURA_USE_AUDIO_SYSTEM)
#include "Sound/Asura_ProjectSoundIDs.h"
#endif	//!defined(ASURA_USE_AUDIO_SYSTEM)

#ifdef ASURA_TOOLS
#include "Core/Asura_String.h" //for storing tools phonon name
#endif

class Asura_Stream;
class Asura_VolumeGroup;

//Defines for phonon guid ranges.
const Asura_Guid ASURA_PHONONGUID_COMMONFILE_START = ASURA_GUID_TOOLSCREATEDENTITY_START;
const Asura_Guid ASURA_PHONONGUID_MISSIONFILE_START = ASURA_PHONONGUID_COMMONFILE_START+100000;

// (MH - 27.05.2002) new for version 2: Inner radius, m_eSoundID
//                   removed for version 2: WolfName, WolfVolume
// (MH - 12.06.2002) added cuboid falloff and occlusion.
// (MH - 26.06.2006) Sound System Upgrade: old phonon chunks "PHON" need to be
// converted on load by creating an appropriate event. Chunk data will be 
// saved for the new system as "PHOE". 
// (RAR 11/07/2006 Using "SDPH" to 'match' other new chunks)
// Asura_Phonon_System::SetPhononVolumeScale() should be deprecated (it's only 
// used in MiamiVice) as its functionality can be achieved using the mixer.
// NB class has references to Asura_Sound_Event*
class Asura_Phonon
{
    friend class Asura_Phonon_Handler;
    friend class Asura_Chunk_Phonons;
	friend class Asura_Chunk_SoundEvent_Phonon;
    friend class Tools_Mode_Sound;
    friend class Tools_Phonon_Object;

public:
    
    Asura_Phonon() 
	{
		m_pxEvent = NULL;
		m_uGuid = ASURA_GUID_UNREGISTERED;
	#ifdef ASURA_TOOLS
		m_uTemplateGuid = ASURA_GUID_UNREGISTERED;
		m_uLightSets = 0xffffffff;
		m_uPlatforms = 0xffffffff;
	#endif
		m_pxClipVolume = 0;
	}

	~Asura_Phonon()
	{
		Shutdown();
	}

	const Asura_Phonon& operator = ( const Asura_Phonon& xPhonon);


	//Deals with some allocation/deallocation that can't be done in the constructor/destructor because of the
	//way in which the phonons are allocated
	void Initialise();
	void Shutdown();

    // functions for sending messages
    void Activate();
    void Deactivate();
    void Trigger();

    // functions required in order to use Asura_Hashed_List
    unsigned int GetGuid() const { return m_uGuid; }
    bool IsDestroyed() const { return false; }

    // access functions
	const Asura_Sound_Event* GetEvent() const { return m_pxEvent; }
    bool  IsPlaying() const;

   	void SetPosition(const Asura_Vector_3& xPos);


#ifdef ASURA_TOOLS
	//resets phonons back to its original positions
	void ResetPosition() 
	{
		if (m_xPosition != m_xInitialPosition) SetPosition(m_xInitialPosition);
	}
#endif



	// Accessor functions
	float GetInnerRadius() const { return m_fInnerRadius; }
	float GetOuterRadius() const { return m_fOuterRadius; }
	Asura_Vector_3 GetPosition() const { return m_xPosition; }	// used by deprecated Audio system
	const Asura_Vector_3* GetPositionPtr() const { return &m_xPosition; }
	Asura_Quat GetOrientation() const { return m_xOrient; }		// used by deprecated Audio system
	bool GetUseCuboidFalloff() const { return m_bUseCuboidFalloff; }
	const Asura_Vector_3& GetInnerCuboidRadius() const { return m_xInnerCuboidRadius; } 
	const Asura_Vector_3& GetOuterCuboidRadius() const { return m_xOuterCuboidRadius; } 
	bool GetUseClipVolume() const { return m_bUseClipVolume; }
	bool GetInvertClipVolume() const { return m_bInvertClipVolume; }
	const Asura_VolumeGroup* GetClipVolume() const { return m_pxClipVolume; }
	bool GetUseOrient() const { return m_bUseOrient; }
	const Asura_Quat& GetOrient() const { return m_xOrient; }

#ifdef ASURA_TOOLS
	//Used after a phonon has been created from the template's chunk buffers to adjust positions etc
	//as needed. RAR 27/04/2004
	void ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance);

	bool IsMemberOfTemplateInstance() const {return m_uTemplateGuid != ASURA_GUID_UNREGISTERED;}
	Asura_Guid GetTemplateGuid() const {return m_uTemplateGuid;}
	void SetTemplateGuid(Asura_Guid uGuid) {m_uTemplateGuid = uGuid;}
	bool IsMemberOfCurrentLightSet() const;
	void SetLightSets(u_int uLightSets) {m_uLightSets = uLightSets;}
	u_int GetLightSets() const {return m_uLightSets;}
	bool IsEnabledForPlatform(int iPlatform) const {return (m_uPlatforms & (1<<iPlatform)) ? true : false;}
	const Asura_String* GetName() const {return &m_xName;}
	const Asura_Vector_3& GetInitialPosition() const{return m_xInitialPosition;}

	bool BelongsToMissionFile() const {return m_uGuid >= ASURA_PHONONGUID_MISSIONFILE_START;}
#endif // #ifdef ASURA_TOOLS

private:
    
    // methods
    void Set(const Asura_Phonon& xNewSettings);

	void Restart(void);
	
	inline bool operator == ( const Asura_Phonon& l) const
	{
		//Equivalence now checks the initial position instead of the current position , so that
		//the phonon object won't think anything has changed when the phonon gets temporarily moved
		//by cutscenes. RAR 13/11/2007
		//Not bothering to check m_pxClipVolume since Asura_VolumeGroup doesn't have an == , and the only
		//case that uses this function is just doing this as a backup means of checking for changes , which
		//shouldn't be required for the volume edited.
		//RAR 24/01/2008
		return 
			(m_pxEvent==l.m_pxEvent)  
#ifdef ASURA_TOOLS
			&& (m_xInitialPosition == l.m_xInitialPosition)
#endif
			&& (m_fInnerRadius==l.m_fInnerRadius)
            && (m_fOuterRadius==l.m_fOuterRadius)
            && (m_xInnerCuboidRadius==l.m_xInnerCuboidRadius)
            && (m_xOuterCuboidRadius==l.m_xOuterCuboidRadius)
            && (m_xOrient==l.m_xOrient)
			&& (m_bUseCuboidFalloff==l.m_bUseCuboidFalloff)
            && (m_bStartsActive==l.m_bStartsActive)
            && (m_bUseClipVolume==l.m_bUseClipVolume)
            && (m_bInvertClipVolume==l.m_bInvertClipVolume)
            && (m_bUseOrient==l.m_bUseOrient)
		#ifdef ASURA_TOOLS
            && (m_uLightSets==l.m_uLightSets)
            && (m_uPlatforms==l.m_uPlatforms)
		#endif
			&& (m_bRetriggerable == l.m_bRetriggerable) 
#if defined(ASURA_TOOLS)
			&& (m_uEventHashID==l.m_uEventHashID) 
#endif
			&& (m_bUseRetriggerRadius == l.m_bUseRetriggerRadius)
			&& (m_bUseRetriggerBoundingBox == l.m_bUseRetriggerBoundingBox)
			&& (m_fRetriggerRadius == l.m_fRetriggerRadius)
			&& (m_xRetriggerBoundingBox == l.m_xRetriggerBoundingBox)
			;
	}


	void CalculateLargestRadius();

	bool CurrentlyPlaying() const;
	bool HasValidSoundHandle() const;
	void CreateSound();
	void DestroySound();

private:

    unsigned int m_uGuid;
	const Asura_Sound_Event*	m_pxEvent;

	Asura_Vector_3 m_xPosition;
	float m_fInnerRadius;
	float m_fOuterRadius;
    Asura_Vector_3 m_xInnerCuboidRadius;
    Asura_Vector_3 m_xOuterCuboidRadius;
    float m_fLargestRadius;
	//Asura_OrientedVolume m_xBoundingBox;//optional bounding box to clip sound to
	Asura_VolumeGroup* m_pxClipVolume;//optional volumes to clip sound to
	Asura_Quat m_xOrient;//orientation to use with cuboid radius


	bool m_bUseCuboidFalloff;
    bool m_bStartsActive;
	bool m_bUseClipVolume;
	bool m_bInvertClipVolume;
	bool m_bUseOrient;
	bool m_bRetriggerable; // If looping, this is ignored
	
	// new functionality as part of Sound System Upgrade
	bool				m_bUseRetriggerRadius;		// if neither is set...
	bool				m_bUseRetriggerBoundingBox;	// retrigger is done based on outer range.
	float				m_fRetriggerRadius;			//optional radius for retriggering. 
	Asura_Bounding_Box	m_xRetriggerBoundingBox;	//optional bounding box for retriggering. 

	// internal data		
	Asura_Handle m_iSoundEventHandle;

	bool m_bShouldTrigger;
	bool m_bHasChanged;
    bool m_bIsActive;

#ifdef ASURA_TOOLS
	Asura_Guid m_uTemplateGuid;//guid of template instance that owns this
	Asura_String m_xName;//name to be used by tools object
	u_int m_uLightSets;//bitfield of of the increasingly inappropriately named lightsets that this is included in
	u_int m_uPlatforms;//bitfield of platforms that this should be exported for
#endif

#if defined(ASURA_TOOLS)
	Asura_Hash_ID m_uEventHashID;
#endif

#ifdef ASURA_TOOLS
	//backup of initial position ,to allow the phonons to be moved back in tools after
	//being moved by cutscenes. RAR 12/11/2007
	Asura_Vector_3 m_xInitialPosition;
#endif

};


class Asura_Phonon_Handler
{
	friend class Asura_Phonon_System;

public:
	// Public functions...
	Asura_Phonon_Handler  (void)	{}
	~Asura_Phonon_Handler ()
	{
		//RAR 03/04/2006 Make sure sounds are stopped before we go
		Silence();
	}

	// Reset function...
	void			Reset						(void)				{ Silence(); m_xPhononTree.Reset(); }

	void 			Silence						(void)				{ m_xPhononTree.Traverse(SilencePhonon); }
	void 			Update						(void)				{ m_xPhononTree.Traverse(UpdatePhonon); }

	// Traversal/callback function...
	void 			Traverse					(void (*pfnFunc)(Asura_Hash_ID uHash,Asura_Phonon* pxPhonon))	{ m_xPhononTree.Traverse(pfnFunc); }

	// Allocation/deallocation functions...
	Asura_Phonon*	AllocatePhonon				(Asura_Guid uGuid);
	Asura_Phonon*	AllocatePhonon				(bool bMissionFile);
	void			DeallocatePhonon			(Asura_Phonon *pxPhonon);

	// Accessor functions...
	Asura_Phonon*	GetPhonon					(Asura_Guid uGuid)	{ return(m_xPhononTree.GetDataByHash(uGuid)); }
#if defined(ASURA_TOOLS)
	Asura_Phonon*	GetPhononByIndex			(int iIndex)		{ return(m_xPhononTree.GetDataByIndex(iIndex)); }
	int				GetNumberOfPhonons			(void)				{ return(m_xPhononTree.GetNumberOfNodes()); }
#endif	// defined(ASURA_TOOLS)

#if defined(ASURA_TOOLS)
	void SoundEventDeleted(const Asura_Sound_Event* pxEvent);
	static bool SoundEventDeletedCallback(Asura_Hash_ID uHash,Asura_Phonon* pxPhonon,void* pUserData);

	void SoundEventRenamed(Asura_Hash_ID uOldHash,Asura_Hash_ID uNewHash);
	static bool SoundEventRenamedCallback(Asura_Hash_ID uHash,Asura_Phonon* pxPhonon,void* pUserData);

#endif

#ifdef ASURA_TOOLS
	//resets all phonons back to their original positions
	void ResetPositions() {m_xPhononTree.Traverse(ResetPosition);}
#endif



private:
	// Private variables...
	Asura_TemplateHashTree<Asura_Phonon>	m_xPhononTree;

	// Private functions...
	static	void	SilencePhonon				(Asura_Hash_ID uHash, Asura_Phonon* pxPhonon);
	static	void	UpdatePhonon				(Asura_Hash_ID uHash, Asura_Phonon* pxPhonon);

#ifdef ASURA_TOOLS
	static	void	ResetPosition				(Asura_Hash_ID uHash, Asura_Phonon* pxPhonon) {pxPhonon->ResetPosition();}
#endif
};


class Asura_Phonon_System
{
	friend class Asura_Phonon_Handler;

public:
	// Public functions...
	static 	void 					Initialise					(void);
	static  void					Update						(void);
	static 	void 					Shutdown					(void);
	static 	void 					Reset						(void);

	static 	void 					Silence						(void);
	static 	void 					Traverse					(void (*pfnFunc)(Asura_Hash_ID uHash,Asura_Phonon* pxPhonon)) { GetCurrentHandler().Traverse(pfnFunc); }

	static 	Asura_Phonon* 			GetPhonon					(Asura_Guid uGuid)								{ return(GetCurrentHandler().GetPhonon(uGuid)); }
#if defined(ASURA_TOOLS)
	static 	Asura_Phonon* 			GetPhononByIndex			(int iIndex)									{ return(GetCurrentHandler().GetPhononByIndex(iIndex)); }
	static	int						GetNumberOfPhonons			(void)											{ return(GetCurrentHandler().GetNumberOfPhonons()); }
#endif	// defined(ASURA_TOOLS)

	static 	Asura_Phonon*			AllocatePhonon				(Asura_Guid uGuid)								{ return(GetCurrentHandler().AllocatePhonon(uGuid)); }
	static 	Asura_Phonon*			AllocatePhonon				(bool bMissionFile)								{ return(GetCurrentHandler().AllocatePhonon(bMissionFile)); }
	static 	void					DeallocatePhonon			(Asura_Phonon *pxPhonon)						{ GetCurrentHandler().DeallocatePhonon(pxPhonon); }

	static 	float					GetPhononVolumeScale		(void)											{ return 1.0f; }

	static 	void 					DisablePhonons				(void)											{ s_bDisablePhonons = true; }
	static 	void 					EnablePhonons				(void)											{ s_bDisablePhonons = false; }
	static 	bool 					ArePhononsDisabled			(void)											{ return(s_bDisablePhonons); }

	static 	void 					PausePhononUpdates			(void)											{ s_bPausePhononUpdates = true; }
	static 	void 					ResumePhononUpdates			(void)											{ s_bPausePhononUpdates = false; }
	static 	bool 					AreUpdatesPaused			(void)											{ return(s_bPausePhononUpdates); }
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Handler functions...
	static	void					ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash);
	static	void					ReparentHandler(Asura_Phonon_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash);

	static	Asura_Phonon_Handler*	CreateHandler				(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash);
	static	bool					DestroyHandler				(const Asura_Hash_ID uHash);
	static	void					WriteHandlerReport			(u_int uHandler, Asura_Stream& xLog);

	static	void					CreateAndSetHandler			(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)			{ s_pxCurrentHandler = CreateHandler(uHash, uParentHandlerHash); }
	static	void					SetCurrentHandler			(const Asura_Hash_ID uHash)						{ s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static	void					SetCurrentHandler			(Asura_Phonon_Handler* pxHandler)				{ s_pxCurrentHandler = pxHandler; }
	static	Asura_Phonon_Handler&	GetCurrentHandler			(void)											{ return(*s_pxCurrentHandler); }
	static	Asura_Phonon_Handler*	GetHandler					(const Asura_Hash_ID uHash)						{ return(s_xHandlerTree.GetDataByHash(uHash)); }
#else
	// Handler functions...
	static	Asura_Phonon_Handler&	GetCurrentHandler			(void)											{ return(s_xHandler); }
#endif

#if defined(ASURA_TOOLS)
	static void SoundEventDeleted(const Asura_Sound_Event* pxEvent) {GetCurrentHandler().SoundEventDeleted(pxEvent);}
	static void SoundEventRenamed(Asura_Hash_ID uOldHash,Asura_Hash_ID uNewHash) {GetCurrentHandler().SoundEventRenamed(uOldHash,uNewHash);}
#endif

	static void ResetNumPlayingPhonons() { s_uNumPlayingPhonons = 0; }
	static u_int GetNumPlayingPhonons() { return s_uNumPlayingPhonons; }

#ifdef ASURA_TOOLS
	//resets all phonons back to their original positions
	static void ResetPositions() {GetCurrentHandler().ResetPositions();}
#endif

private:
	// Private variables...
	static 	float											s_fVolumeScale;
	static 	bool											s_bDisablePhonons;
	static	bool											s_bPausePhononUpdates;

	static 	Asura_Guid										s_uNextCommonGuid;
	static 	Asura_Guid										s_uNextMissionGuid;

	static	u_int											s_uNumPlayingPhonons;

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	static	Asura_Phonon_Handler*							s_pxCurrentHandler;		// Currently active handler.
	static	Asura_TemplateHashTree<Asura_Phonon_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.

	// Private functions...
	static	void					SilenceHandler				(Asura_Hash_ID uHash, Asura_Phonon_Handler* pxHandler);
	static	void					UpdateHandler				(Asura_Hash_ID uHash, Asura_Phonon_Handler* pxHandler);
#else
	static	Asura_Phonon_Handler							s_xHandler;				// Single static-instance of a zone.
#endif
};

#endif//__ASURA_PHONON_H__
