//
// Asura_Sound.h
//
// Created by Kevin Lea - 12/06/2001 11:13:07
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_SOUND_H__
#define __ASURA_SOUND_H__

#ifndef ASURA_USE_AUDIO_SYSTEM

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Camera/Asura_Camera.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_Platform_Timers.h"
#include "Core/Asura_HandleTreeTemplate.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_OrientedVolume.h"
#include "Core/Asura_Maths.h"

#include "Sound/Asura_ProjectSoundIDs.h"  // project-supplied file

// D E F I N E S /////////////////////////////////////////////////////////////

#if _DEBUG || defined(ASURA_TOOLS)
    #define ASURA_SOUND_STORE_FILENAMES (1)
#else
    #define ASURA_SOUND_STORE_FILENAMES (0)
#endif

// macro to shorten code using this template instance
template <class T> class Asura_HandleTree;
#define Asura_Sound_Handle_Tree Asura_HandleTree<Asura_Sound_ID>

template <class T> class Asura_TemplateHashTree;

// (NJD - 31.03.2003) - this has now been moved to Asura_ProjectSoundIDs.h
//#define ASURA_MAX_NUMBER_OF_LOADED_SOUNDS 256   // same for all platforms, 
// note that max playing sounds is platform-specific. It is determined by 
// independent code at initialisation time.
#ifdef ASURA_MAX_NUMBER_OF_LOADED_SOUNDS
	//Max loaded sounds needs to be much higher while in tools , so bump it up if it is lower
	//RAR 23/05/2007
	#if defined(ASURA_TOOLS) && defined(ASURA_MAIN)
		#if (ASURA_MAX_NUMBER_OF_LOADED_SOUNDS < 8192)
			#undef ASURA_MAX_NUMBER_OF_LOADED_SOUNDS
			#define ASURA_MAX_NUMBER_OF_LOADED_SOUNDS (8192)
		#endif
	#endif
#else
#error You need to define ASURA_MAX_NUMBER_OF_LOADED_SOUNDS in your Asura_ProjectSoundIDs.h file!
#endif

#define ASURA_MAX_NUMBER_OF_ACTIVE_SOUNDS 128   // same for all platforms - 
// active sounds are not necessarily playing but are considered for 3d volume
// etc. and stopped / restarted as necessary.

#define ASURA_INVALID_LOADED_SOUND_INDEX (-1)

#define ASURA_SOUND_MAX_SPLIT_SCREENS 4

//value to pas for inner/outer radius to get sound system to use the default for that sound
#define ASURA_SOUND_USE_DEFAULT_RADIUS -1.0f

#ifndef ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS // this can be overridden in project settings
#define ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS (2)
#endif

//All sounds are being recorded so that a volume of 0.75 should correspond to the maximum
//normal volume. 
//Therefore we allow the various volume multipliers to go above 1 and then multiply
//the result by 0.75 and cap at 1.
//This multiplier and cap is applied before the volume gets sent to platform code , and
//therefore will be done before the user sound and music volume multipliers get applied.
//RAR 16/12/2004
//This can now be set on a per-project basis
//RAR 13/06/2006
#ifndef ASURA_SOUND_GLOBAL_VOLUME_SCALE
#define ASURA_SOUND_GLOBAL_VOLUME_SCALE (0.75f)
#endif

#if defined(ASURA_USE_STREAMING)
#define ASURA_SOUND_ID_STREAMING_START  0
#define ASURA_SOUND_ID_STREAMING_ZONE_MULTIPLE 1000000
#endif //defined(ASURA_USE_STREAMING)

//Special sound id for use in resource chunks that indicates that a sound ID should be assigned
//dynamically upon load. RAR 13/07/2006
#define ASURA_SID_DYNAMIC ((Asura_Sound_ID)0x7fffffff)
//Range in which dynamic sound ids will be allocated. These should never be saved out
//so it should be okay to change them if needed. RAR 13/07/2006
#define ASURA_SID_FIRST_DYNAMIC ((Asura_Sound_ID)0x10000000)
#define ASURA_SID_NUM_DYNAMIC (65536) //If this is changed then GetDynamicSoundID() will need to be changed
#define ASURA_SID_LAST_DYNAMIC  ((Asura_Sound_ID)(ASURA_SID_FIRST_DYNAMIC+ASURA_SID_NUM_DYNAMIC-1))

// C O N S T A N T S /////////////////////////////////////////////////////////

static const float fASURA_SOUND_DEFAULT_INNER_RADIUS = 1.0f;
static const float fASURA_SOUND_DEFAULT_OUTER_RADIUS = 20.0f;

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

#if ASURA_SOUND_STORE_FILENAMES
class Asura_String;
#endif
class Asura_ActiveSoundSample;
class Asura_SoundProperties;
class Asura_LOS_OutputData;
class Asura_Stream;
class Asura_VolumeGroup;

// T Y P E D E F S ///////////////////////////////////////////////////////////

typedef float(*Asura_Sound_OcclusionFactor_Callback)		(const Asura_LOS_OutputData* pxLoSData);

// C L A S S E S /////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Sound_Handler
// Author       : Andy Tate
// Date         : 23-Mar-2006
//
// Description  : Handles individual sets of sound-system data.
////////////////////////////////////////////////////////////////////////////////

class Asura_Sound_Handler
{
	friend class Asura_Sound_System;

public:
	// Public functions...
	Asura_Sound_Handler(void);
	~Asura_Sound_Handler();

	// Reset function...
	void Reset(void);

#if !defined(ASURA_TOOLS)
	// Manipulation functions...
	bool Add(const Asura_Sound_ID eSoundID, int& iIndex);
	bool Delete(const Asura_Sound_ID eSoundID);

	// Accessor functions...
	bool FindIndex(const Asura_Sound_ID eSoundID, int& iIndex) const { int* piIndex = m_xSoundIDToLoadedIndexTree.GetDataByHash(eSoundID); if(piIndex == NULL) { return(false); } else { iIndex = *piIndex; return(true); } }
	bool IsPresent(const Asura_Sound_ID eSoundID) const { return(m_xSoundIDToLoadedIndexTree.GetDataByHash(eSoundID) != NULL); }

	static u_int GetNumLoadedSounds(void) { return(s_xLoadedIndexList.GetNumberOfActiveHandles()); }
#endif	// !defined(ASURA_TOOLS)

	bool AddToHashToIDTree(Asura_Hash_ID uHash, int iSoundID);
	bool RemoveFromHashToIDTree(Asura_Hash_ID uHash) { return m_xHashToIDTree.DeleteNodeByHash(uHash); }

	const int* GetIDByHash(Asura_Hash_ID uHash) { return m_xHashToIDTree.GetDataByHash(uHash); }


private:
	// Private variables...
#if !defined(ASURA_TOOLS)
	// static class data 
	static Asura_Handle_List s_xLoadedIndexList;				// List of loaded sound indices.

	// member data
	Asura_TemplateHashTree<int>	m_xSoundIDToLoadedIndexTree;	// Tree of mappings from sound-IDs to loaded sound indices.
#endif	// !defined(ASURA_TOOLS)

	//hash tree for converting from hashed filenames to loaded sound ids
	//(only stores the id of one instance of any given wav file , which may be an issue
	//if the sounds aren't delete at the same time) RAR 09/12/2003
	Asura_TemplateHashTree<int> m_xHashToIDTree;

	// Private functions...
	Asura_Sound_Handler(const Asura_Sound_Handler& xOther);
	Asura_Sound_Handler& operator= (const Asura_Sound_Handler& xOther);

};   // End of class 'Asura_Sound_Handler'.



class Asura_Sound_System
{
 	friend class Asura_Sound_Handler;

	public:


        enum SoundOutputType
        {
            SoundOutputStereo,
            SoundOutputMono,
            SoundOutputHeadphones,
			SoundOutputSurround,
        };

        // (MH - 06.12.2002) preliminary support for sound groups so you can set 
        // volumes for speech, ambient sound, weapons etc. independently.
        // SoundFX are currently all played with the first group.
        // PS2 will only support 4 groups. To support more groups it would have to be
        // done in independent code.
        // Streaming sounds can choose whether they are music or not.
        // You could define meaningful names to these for your project.

		//(AFM 05/11/08) - Sound Groups are Obsolete.
		// The functionality can be replicated by feeding all mixer channels through code controlled mixer channels.
		//enum SoundGroup
		//{
        //    SoundGroup_SoundFX,
        //    SoundGroup_Music,
        //    NumSoundGroups
        //};

		// reverb types are basically presets which are implemented on each platform.
		// Not too sure what to do for split screen, probably use first one which isn't
		// "none".
		enum ReverbType
		{
			ASURA_REVERB_ID_NONE,
			ASURA_REVERB_ID_GENERIC,
			ASURA_REVERB_ID_PADDEDCELL,
			ASURA_REVERB_ID_ROOM,
			ASURA_REVERB_ID_BATHROOM,
			ASURA_REVERB_ID_LIVINGROOM,
			ASURA_REVERB_ID_STONEROOM,
			ASURA_REVERB_ID_AUDITORIUM,
			ASURA_REVERB_ID_CONCERTHALL,
			ASURA_REVERB_ID_CAVE,
			ASURA_REVERB_ID_ARENA,
			ASURA_REVERB_ID_HANGAR,
			ASURA_REVERB_ID_CARPETTEDHALLWAY,
			ASURA_REVERB_ID_HALLWAY,
			ASURA_REVERB_ID_STONECORRIDOR,
			ASURA_REVERB_ID_ALLEY,
			ASURA_REVERB_ID_FOREST,
			ASURA_REVERB_ID_CITY,
			ASURA_REVERB_ID_MOUNTAINS,
			ASURA_REVERB_ID_QUARRY,
			ASURA_REVERB_ID_PLAIN,
			ASURA_REVERB_ID_PARKINGLOT,
			ASURA_REVERB_ID_SEWERPIPE,
			ASURA_REVERB_ID_UNDERWATER,
			ASURA_REVERB_ID_DRUGGED,
			ASURA_REVERB_ID_DIZZY,
			ASURA_REVERB_ID_PSYCHOTIC,
			ASURA_REVERB_ID_CASTLE_SMALLROOM,
			ASURA_REVERB_ID_CASTLE_SHORTPASSAGE,
			ASURA_REVERB_ID_CASTLE_MEDIUMROOM,
			ASURA_REVERB_ID_CASTLE_LONGPASSAGE,
			ASURA_REVERB_ID_CASTLE_LARGEROOM,
			ASURA_REVERB_ID_CASTLE_HALL,
			ASURA_REVERB_ID_CASTLE_CUPBOARD,
			ASURA_REVERB_ID_CASTLE_COURTYARD,
			ASURA_REVERB_ID_CASTLE_ALCOVE,
			ASURA_REVERB_ID_FACTORY_ALCOVE,
			ASURA_REVERB_ID_FACTORY_SHORTPASSAGE,
			ASURA_REVERB_ID_FACTORY_MEDIUMROOM,
			ASURA_REVERB_ID_FACTORY_LONGPASSAGE,
			ASURA_REVERB_ID_FACTORY_LARGEROOM,
			ASURA_REVERB_ID_FACTORY_HALL,
			ASURA_REVERB_ID_FACTORY_CUPBOARD,
			ASURA_REVERB_ID_FACTORY_COURTYARD,
			ASURA_REVERB_ID_FACTORY_SMALLROOM,
			ASURA_REVERB_ID_ICEPALACE_ALCOVE,
			ASURA_REVERB_ID_ICEPALACE_SHORTPASSAGE,
			ASURA_REVERB_ID_ICEPALACE_MEDIUMROOM,
			ASURA_REVERB_ID_ICEPALACE_LONGPASSAGE,
			ASURA_REVERB_ID_ICEPALACE_LARGEROOM,
			ASURA_REVERB_ID_ICEPALACE_HALL,
			ASURA_REVERB_ID_ICEPALACE_CUPBOARD,
			ASURA_REVERB_ID_ICEPALACE_COURTYARD,
			ASURA_REVERB_ID_ICEPALACE_SMALLROOM,
			ASURA_REVERB_ID_SPACESTATION_ALCOVE,
			ASURA_REVERB_ID_SPACESTATION_MEDIUMROOM,
			ASURA_REVERB_ID_SPACESTATION_SHORTPASSAGE,
			ASURA_REVERB_ID_SPACESTATION_LONGPASSAGE,
			ASURA_REVERB_ID_SPACESTATION_LARGEROOM,
			ASURA_REVERB_ID_SPACESTATION_HALL,
			ASURA_REVERB_ID_SPACESTATION_CUPBOARD,
			ASURA_REVERB_ID_SPACESTATION_SMALLROOM,
			ASURA_REVERB_ID_WOODEN_ALCOVE,
			ASURA_REVERB_ID_WOODEN_SHORTPASSAGE,
			ASURA_REVERB_ID_WOODEN_MEDIUMROOM,
			ASURA_REVERB_ID_WOODEN_LONGPASSAGE,
			ASURA_REVERB_ID_WOODEN_LARGEROOM,
			ASURA_REVERB_ID_WOODEN_HALL,
			ASURA_REVERB_ID_WOODEN_CUPBOARD,
			ASURA_REVERB_ID_WOODEN_SMALLROOM,
			ASURA_REVERB_ID_WOODEN_COURTYARD,
			ASURA_REVERB_ID_SPORT_EMPTYSTADIUM,
			ASURA_REVERB_ID_SPORT_SQUASHCOURT,
			ASURA_REVERB_ID_SPORT_SMALLSWIMMINGPOOL,
			ASURA_REVERB_ID_SPORT_LARGESWIMMINGPOOL,
			ASURA_REVERB_ID_SPORT_GYMNASIUM,
			ASURA_REVERB_ID_SPORT_FULLSTADIUM,
			ASURA_REVERB_ID_SPORT_STADIUMTANNOY,
			ASURA_REVERB_ID_PREFAB_WORKSHOP,
			ASURA_REVERB_ID_PREFAB_SCHOOLROOM,
			ASURA_REVERB_ID_PREFAB_PRACTISEROOM,
			ASURA_REVERB_ID_PREFAB_OUTHOUSE,
			ASURA_REVERB_ID_PREFAB_CARAVAN,
			ASURA_REVERB_ID_DOME_TOMB,
			ASURA_REVERB_ID_DOME_SAINTPAULS,
			ASURA_REVERB_ID_PIPE_SMALL,
			ASURA_REVERB_ID_PIPE_LONGTHIN,
			ASURA_REVERB_ID_PIPE_LARGE,
			ASURA_REVERB_ID_PIPE_RESONANT,
			ASURA_REVERB_ID_OUTDOORS_BACKYARD,
			ASURA_REVERB_ID_OUTDOORS_ROLLINGPLAINS,
			ASURA_REVERB_ID_OUTDOORS_DEEPCANYON,
			ASURA_REVERB_ID_OUTDOORS_CREEK,
			ASURA_REVERB_ID_OUTDOORS_VALLEY,
			ASURA_REVERB_ID_MOOD_HEAVEN,
			ASURA_REVERB_ID_MOOD_HELL,
			ASURA_REVERB_ID_MOOD_MEMORY,
			ASURA_REVERB_ID_DRIVING_COMMENTATOR,
			ASURA_REVERB_ID_DRIVING_PITGARAGE,
			ASURA_REVERB_ID_DRIVING_INCAR_RACER,
			ASURA_REVERB_ID_DRIVING_INCAR_SPORTS,
			ASURA_REVERB_ID_DRIVING_INCAR_LUXURY,
			ASURA_REVERB_ID_DRIVING_FULLGRANDSTAND,
			ASURA_REVERB_ID_DRIVING_EMPTYGRANDSTAND,
			ASURA_REVERB_ID_DRIVING_TUNNEL,
			ASURA_REVERB_ID_CITY_STREETS,
			ASURA_REVERB_ID_CITY_SUBWAY,
			ASURA_REVERB_ID_CITY_MUSEUM,
			ASURA_REVERB_ID_CITY_LIBRARY,
			ASURA_REVERB_ID_CITY_UNDERPASS,
			ASURA_REVERB_ID_CITY_ABANDONED,
			ASURA_REVERB_ID_DUSTYROOM,
			ASURA_REVERB_ID_CHAPEL,
			ASURA_REVERB_ID_SMALLWATERROOM,

			ASURA_REVERB_ID_MAX
		};

		#ifdef ASURA_TOOLS
			static const char* s_aszReverbNames[ASURA_REVERB_ID_MAX];
		#endif

		// EQ types
		enum EQType
		{
			EQ_OFF,
			EQ_UNDERWATER,
			EQ_DEAFENED,
			NUM_EQ_TYPES
		};

		static void Initialise(void);
		static void Shutdown(void);
		static void Reset(void);

        // (MH - 30.05.2002) project-supplied function
        static void LoadCoreProjectSounds();

        static SoundOutputType GetSoundOutput(void) { return s_eSoundOutputType; }
        static void SetSoundOutput(SoundOutputType eOutputType);

		
		// (MH - 18.03.2004) 2 independent mutes - soundfx and music. these are intended for console mute commands.
		static void SetMuteSoundFX(bool bMute) { s_bMuteSoundFX = bMute; }
		static bool GetMuteSoundFX() { return s_bMuteSoundFX; }
		static void SetMuteMusic(bool bMute) { s_bMuteMusic = bMute; }
		static bool GetMuteMusic() { return s_bMuteMusic; }

        static void SetReverb(enum ReverbType eType, float fReverbAmount);
		static Asura_Handle GetEQHandle();
		static void ReleaseEQHandle(Asura_Handle& iHandle);
		// SetEQType() - set EQ type for EQ handle which is applied to all sounds with EQ on. Effect is maximum at 1.0f and off at 0.0f
		static void SetEQType(Asura_Handle iHandle, EQType eType, float fLevel);
		static EQType GetEQType(Asura_Handle iHandle, float* pfLevel = NULL);
		// Set3dSoundEQDefault() - should 3d sounds have EQ on by default?
		static void Set3dSoundEQDefault( bool bEnable ) { s_b3dSoundEQDefault = bEnable; }

		static bool LoadSoundFromResourceFile( Asura_Handle iFileHandle, int iFileSize, u_int uSoundID, const char* szSoundFilename );

        // (MH - 05.03.2003) - function to allow platform code to load sounds from 
        // soundIDs and get them a loadedSoundIndex. 
		// (MH - 25.10.2005) Also for checking things, can now get loaded index for
		// a given sound ID.
        static int SoundLoadedByPlatformCode(int iSoundID, Asura_Hash_ID uSoundFilenameHashID = ASURA_HASH_ID_UNSET);
		static int GetSoundLoadedIndex(int iSoundID);
        // (MH - 29.05.2002) returns whether a sound is loaded for the specified sound id
        static bool SoundIsLoaded(const Asura_Sound_ID eSoundID);

#ifdef ASURA_TOOLS
		//MKR 18/09/06: Balances the sound id handle tree
		static void BalanceSoundTree() { s_xSoundIDToIndexTree.Balance(); }

        static bool LoadSound( const char* szFilename, Asura_Sound_ID eSoundID);
		//Forces sound to be loaded now , instead of waiting until it is played
		static void ForceSoundToBeLoaded(Asura_Sound_ID eSoundID);
		//Flags sound so that it will be reloaded before next being used
		static void ReloadSound(Asura_Sound_ID eSoundID);

        // for creating lists in tools of loaded sounds
        static int GetNumberOfNamedSounds(Asura_Sound_ID eMinID,Asura_Sound_ID eMaxID);
        static void MakeNameList( const Asura_String** nameList, Asura_Sound_ID eMinID, Asura_Sound_ID eMaxID );
		static Asura_Sound_ID GetSoundIDFromFileName(const char* szFilename,Asura_Sound_ID eMinID,Asura_Sound_ID eMaxID);

		//Gets length of sound in seconds, returning 0 if unavailable.
		//Tools only for now since its only implemented for PC.
		static float GetSoundLength(Asura_Sound_ID eSoundID);
#endif

#if ASURA_SOUND_STORE_FILENAMES
        static Asura_Sound_ID GetSoundIDFromFileName(const char* szFilename);
		static const char* GetFileNameFromSoundID(Asura_Sound_ID eSoundID);
        static const char* GetFileNameFromSoundIndex( int iLoadedSoundIdx );

		static void SetSoundFilename(int iLoadedSoundIndex,Asura_Sound_ID eSoundID,const char* pszFilename);
#endif

		static Asura_Sound_ID GetSoundIDFromHash(Asura_Hash_ID uSoundHash);

		//Creates a sound with eNewSoundID that shares the same 'wav' file data
		//as eExistingSoundID. RAR 25/04/03
		static void LoadSoundIDMapping(Asura_Sound_ID eNewSoundID,Asura_Sound_ID eExistingSoundID,Asura_Hash_ID uExistingSoundHash);

        // (MH - 26.02.2003) ReplaceSound is designed to work around the problem 
        // with UnloadSound. If some info about where to get a sound on the disk 
        // has been given earlier, and the sound being replaced is no smaller in 
        // size than the new one, we can overwrite the sound RAM area. This should 
        // be handled by Platform_ReplaceSound(). The function returns false if 
        // the sound was not replaced (e.g. not enough sound RAM, no data to get 
        // new sound on disk)
        // (MH - 05.03.2003) LoadReplaceableSound() is used to get the sound in 
        // there in the first place.
        static bool LoadReplaceableSound(const Asura_Sound_ID eSoundID);
        static bool ReplaceSound(const Asura_Sound_ID eExistingSoundID, const Asura_Sound_ID eNewSoundID);
		
		// (MH - 07.04.2003) on PS2 replaceable sounds happen to all be partially streamed.
		// At some point we should change this as there is no reason other than memory 
		// fragmentation that we can't dynamically load variable-sized sounds in and out.
		// However for now you can't guarantee that more than 2 such sounds can be playing
		// at once on PS2.
		static int GetMaxActiveReplaceableSounds() { return Platform_GetMaxActiveReplaceableSounds(); }

		static void Update(void);
        static void PauseAll();
        static void UnpauseAll();
		// If bReleaseStreams is set to true, all streaming sounds will be stopped.
		// To be used before things like PS2 FMVs which will stomp on streams.
		static void PauseAllSystems(bool bReleaseStreams = false);
		static void UnpauseAllSystems();

		static bool GetUseAutomaticSoundCentre() { return s_bUseAutomaticSoundCentre; }
        static void SetUseAutomaticSoundCentre(bool bEnable);	// enabled by default. 
        // If enabled, sound centre is taken from the cameras

        // these 2 functions assert if automatic sound centre is enabled.
        static void SetNumSplitScreens(int iNumScreens);
        static void SetSoundCentre(int iSplitScreenID, 
            const Asura_Vector_3& xCentre, const Asura_Matrix_3x3& xOrientation);
        // orientation should probably be the camera's orientation rather than the 3rd person player's

        // enabled by default, allow to speed up game without speeding up sound
        static void SetPitchUsesGameTimeScale(bool bEnable, float fTimeScaleToUse = 1.0f);
		// access current setting
		static bool GetPitchUsesGameTimeScale( void ) { return s_bPitchUsesGameTimeScale; }
            
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
		static Asura_Handle CreateSound(const Asura_Sound_ID eSoundID, Asura_Sound_Handler* pxHandler = 0);
#else
		static Asura_Handle CreateSound(const Asura_Sound_ID eSoundID);
#endif
		// (MH - 27.05.2002) DestroySound copes with iSoundHandle == ASURA_INVALID_HANDLE, to ease
        // shutting down of objects
        // (MH - 17.06.2002) now it takes a reference and sets the handle to invalid
        static void DestroySound(Asura_Handle& iSoundHandle);
		
		static bool PlaySound(Asura_Handle iSoundHandle, bool bLooping);
		// (MH - 17.06.2002) note that stop sound DOES NOT free up the sound resources
        // use destroy sound when a sound is finished with. I am renaming this to pause 
        // sound to avoid confusion.
        //static bool StopSound(Asura_Handle iSoundHandle);
        static void PauseSound(Asura_Handle iSoundHandle);
        static void UnpauseSound(Asura_Handle iSoundHandle);

		static void SetSoundVolume(Asura_Handle iSoundHandle, float fVolume);
		// by default, any sound with a position will be affected by EQ, but the caller of SetSoundPosition can disable that.
		// You can change the default for all new sounds by calling Set3dSoundEQDefault
		static void SetSoundPosition(Asura_Handle iSoundHandle, const Asura_Vector_3& xPosition, bool bEnableEQ = s_b3dSoundEQDefault);
		static void SetSoundPitch(Asura_Handle iSoundHandle, float fPitch);
		static void SetSoundRadius(Asura_Handle iSoundHandle, float fInnerRange, float fOuterRange);
		static void SetSoundSpatial(Asura_Handle iSoundHandle, bool bIsSpatial);
        static void SetSoundHardPan(Asura_Handle iSoundHandle, unsigned int iCameraID);
		// For historical reasons, pan and fade are from 0.0 to 1.0 (fade is 0.0 = back, 1.0 = front)
		static void SetSoundForcedPanAndFade(Asura_Handle iSoundHandle, float fPan, float fFade);
        static void SetSoundCuboidParameters(Asura_Handle iSoundHandle, 
                                             const Asura_Vector_3& fInner, 
                                             const Asura_Vector_3& fOuter);
        static void SetSoundUseCuboidFalloff(Asura_Handle iSoundHandle, bool bUseCuboid);
		//Sets a volume group to use for clipping the sound against. This keeps the pointer rather than copying the contents,
		//to avoid having to copy from the phonon system each time a placed phonon plays
		static void SetSoundClipVolume(Asura_Handle iSoundHandle, const Asura_VolumeGroup* pxClipVolume);
		static void SetSoundInvertClipVolume(Asura_Handle iSoundHandle, bool bInvertClipVolume);
		static void SetSoundOrient(Asura_Handle iSoundHandle, const Asura_Quat& xOrient);
		static void SetSoundUseOrient(Asura_Handle iSoundHandle, bool bUseOrient);
		static void SetSoundMinVolume(Asura_Handle iSoundHandle, float fMinVolume);
		static void SetSoundUseEQ(Asura_Handle iSoundHandle, bool bEnable);// is EQ enabled for this sound? set by default for 3d sounds.
		static void SetSoundPitchUsesTimeScale(Asura_Handle iSoundHandle, bool bUseTimeScale, float fTimeScaleToUse = 1.0f); // Final parameter ignored if bUseTimeScale is true
		static void SetSoundUsesReverb(Asura_Handle iSoundHandle, bool bUseReverb);
		static void SetSoundUsesDistanceFilter(Asura_Handle iSoundHandle, bool bUseDistanceFilter);
		static void SetSoundUsesDoppler(Asura_Handle iSoundHandle, bool bUseDoppler);
		static void SetSoundUsesOcclusion(Asura_Handle iSoundHandle, bool bUseOcclusion);
		static void GetSoundPosition(Asura_Handle iSoundHandle, Asura_Vector_3 &xPosition);
		static float GetSoundVolume(Asura_Handle iSoundHandle);

		static bool SoundIsPlaying(Asura_Handle iSoundHandle);
		static bool SoundIsPreparing(Asura_Handle iSoundHandle);		//Is a sound loading from disc
		static bool SoundHasVoiceAllocated(Asura_Handle iSoundHandle);
		static bool SoundIsLooping(Asura_Handle iSoundHandle);
		static Asura_Sound_ID GetSoundID(Asura_Handle iSoundHandle);
        // use IsPositionInRange to see if a 3d sound would actually be heard if played
        // function returns the volume that a sound would be played at (0.0f means out of range)
        static float IsPositionInRange( const Asura_Vector_3& xPosition,
                                        const float fOuterRadius = fASURA_SOUND_DEFAULT_OUTER_RADIUS,
                                        const float fInnerRadius = fASURA_SOUND_DEFAULT_INNER_RADIUS );
        // returns the outer radius required the make the sound be played at the required volume
        static float GetRequiredOuterRadius( const Asura_Vector_3& xPosition,
                                             const float fVolumeRequired,
                                             const float fInnerRadius = fASURA_SOUND_DEFAULT_INNER_RADIUS );
        // IsSphereInRange() and IsBBInRange() return true is the volume intersects any listener.
		static bool IsSphereInRange( const Asura_Vector_3& xPosition, float fRadius );
        static bool IsBBInRange( const Asura_Bounding_Box& xBB );
		//Similar test for the cuboid volumes theat can be used by phonons
		static bool IsCuboidVolumeInRange(const Asura_Vector_3& xPosition, const Asura_Vector_3&  xExtents, const Asura_Quat* pxOrient);

		static float GetDefaultOuterRadius( Asura_Sound_ID eSoundID );
		static float GetDefaultInnerRadius( Asura_Sound_ID eSoundID );

        // (MH - 22.05.2002) automatic return of sound handle when sound has finished playing
        static void FinishedWithSoundHandle(Asura_Handle& iSoundHandle);

        // (MH - 05.06.2002) some functions to make sound play calls easier
        static Asura_Handle Create3dSound(Asura_Sound_ID eSoundID, const Asura_Vector_3* pxPosition = NULL, float fOuterRadius = ASURA_SOUND_USE_DEFAULT_RADIUS, float fInnerRadius = ASURA_SOUND_USE_DEFAULT_RADIUS );
        static bool PlayOneShotSound(Asura_Sound_ID eSoundID, const Asura_Vector_3* pxPosition = NULL, float fOuterRadius = ASURA_SOUND_USE_DEFAULT_RADIUS, float fInnerRadius = ASURA_SOUND_USE_DEFAULT_RADIUS, float fVolume = 1.0f, float fPitch = 1.0f );
		// Play a 2d sound with an id and a volume - can replace PlayOneShotSound in many places
		static bool Play2DSound(Asura_Sound_ID eSoundID, float fVolume = 1.0f);
        static Asura_Handle PlayLoopingSound(Asura_Sound_ID eSoundID, const Asura_Vector_3* pxPosition = NULL, float fOuterRadius = ASURA_SOUND_USE_DEFAULT_RADIUS, float fInnerRadius = ASURA_SOUND_USE_DEFAULT_RADIUS);

		/// Kevin Lea - 25/04/2002 10:57:47 - streaming sounds
		static Asura_Handle CreateStreamingSound( const char* szFilename);
		static void PlayStreamingSound(Asura_Handle iHandle, bool bLooping, bool bPreload = false);
		static void StopStreamingSound(Asura_Handle iHandle);
		static void PauseStreamingSound(Asura_Handle iHandle);
		static void ResumeStreamingSound(Asura_Handle iHandle);
		static void ReleaseStreamingSound(Asura_Handle iHandle);
        static bool StreamingSoundIsReady(Asura_Handle iHandle);
		static void StartPreloadedStreamingSound(Asura_Handle iHandle);
        static Asura_Sound_ID GetSoundIDFromSoundHandle( Asura_Handle iHandle );
        static void EnsureAllStreamingSoundsAreStopped();
        static void StreamingSoundSetVolume(Asura_Handle iHandle, float fVolume);
		static bool StreamingSoundIsPlaying(Asura_Handle iHandle);
		static int GetNumberActiveStreamingSounds() {return s_pxStreamingSoundList->GetNumberOfActiveHandles();}
		static int GetMaxStreamingSounds() { return Platform_GetMaxStreamingSounds(); }

		static bool StreamingSoundIsFinished(Asura_Handle iHandle);
		static float GetStreamingSoundPosition(Asura_Handle iHandle);
		static float GetStreamingSoundLength(Asura_Handle iHandle);
		#if ( defined ASURA_TOOLS ) || ( defined ASURA_SUPPORT_DYNAMIC_MUSIC_SYSTEM )
		static void SetStreamingSoundPosition(Asura_Handle iHandle,float fTime);
		#endif

		//Getting sound properties.
		static Asura_SoundProperties* GetSoundPropertiesByIndex(int iLoadedSoundIndex);
		static Asura_SoundProperties* GetSoundPropertiesByID(const Asura_Sound_ID eSoundID);

		static int GetNumberOfPlayingSounds()
		{
			if (s_pxPlayingSoundHandleListWithEQ)
			{
				return s_pxPlayingSoundHandleListWithEQ->GetNumberOfActiveHandles() + s_pxPlayingSoundHandleListWithoutEQ->GetNumberOfActiveHandles();
			}
			return 0;
		}
		static int GetMaxNumberOfPlayingSounds()	{ return s_iMaxPlayingSounds; }
		static int GetNumberOfFreeSounds()			{ return s_iMaxPlayingSounds - GetNumberOfPlayingSounds(); }

		// Handler functions...
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
		static	void					ReparentHandler(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash);
		static	void					ReparentHandler(Asura_Sound_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash);

		static	Asura_Sound_Handler*	CreateHandler		(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash);
		static	bool					DestroyHandler		(const Asura_Hash_ID uHash);
		static	void					WriteHandlerReport	(u_int uHandler, Asura_Stream& xLog);

		static	void					CreateAndSetHandler	(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash);
		static	void					SetCurrentHandler	(const Asura_Hash_ID uHash);
		static	void					SetCurrentHandler	(Asura_Sound_Handler* pxHandler)	{ s_pxCurrentHandler = pxHandler; }
		static	Asura_Sound_Handler*	GetHandler			(const Asura_Hash_ID uHash)			{ return(s_xHandlerTree.GetDataByHash(uHash)); }
#else
		static void SetResetPoint(void);

		static void DisableResetPointDuringInitialisation(void)	{ s_bSetResetPointDuringInitialisation = false; }
		static void EnableResetPointDuringInitialisation(void)	{ s_bSetResetPointDuringInitialisation = true; }
#endif
		static	Asura_Sound_Handler&	GetCurrentHandler	(void)								{ return(*s_pxCurrentHandler); }

		static float AmplitudeToDecibels( float fAmplitude );
		static float DecibelsToAmplitude( float fDecibels );

		//Public occlusion related functions
		static void SetEntityOcclusionFactorCallback		(Asura_Sound_OcclusionFactor_Callback pfnCallback);
		static void SetEnvironmentOcclusionFactorCallback	(Asura_Sound_OcclusionFactor_Callback pfnCallback);

		static void SetOcclusionCPUTimePerFrame (float fSeconds);

		static void SetGUIDToIgnoreForOcclusion		(Asura_Guid uIgnore, u_int uCameraIndex = 0);
		
private:
		// Private variables...
		static	Asura_Sound_Handler* s_pxCurrentHandler;		// Currently active handler.
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
		static	Asura_TemplateHashTree<Asura_Sound_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.

		// ART - 31/07/2006 - Need to ensure that the sound handler is correct when we update
		// phonons otherwise they will attempt to play with the wrong data...
		static void TraverseUpdatePhonons(Asura_Hash_ID uHash, Asura_Sound_Handler* pxHandler);
#else
		static	Asura_Sound_Handler	s_xPermanentHandler;	// handler for permanent sounds in non-streaming builds.
		static	Asura_Sound_Handler	s_xResettableHandler;	// handler for resettable sounds non-streaming builds.

		static bool					s_bSetResetPointDuringInitialisation;	// Flag to say whether to allow "Asura_Sound_System::Initialise()" to set the reset-point.
#endif

        // (MH - 23.05.2002) returns whether the sound was successfully unloaded.
        // (MH - 26.02.2003) note that on some platforms, e.g. PS2, UnloadSound 
        // does not actually free up any sound RAM. This is only achieved by the 
        // Reset function.
#if defined(ASURA_TOOLS)
		static bool TraverseUnloadSound(const Asura_Sound_ID eSoundID);
#else
        static void TraverseUnloadSound(Asura_Hash_ID eSoundID, int* piLoadedSoundIndex);
#endif	// defined(ASURA_TOOLS)

#if defined( ASURA_USE_DEBUGOVERLAY )
		static void DoDebugOverlay						( void* const pxIgnore );
		static void DoDebugOverlayActiveSoundsOcclusion	( void* const pxIgnore );

		static void TraversActiveSoundsOcclusion_DebugOverlay (Asura_Handle iSoundHandle);
#endif // #if defined( ASURA_USE_DEBUGOVERLAY )

		static void TraverseDestroySound(Asura_Handle iSoundHandle);
		static void TraverseDestroyUnloadedSound(Asura_Handle iSoundHandle);
        
        static void  StopAll();// destroys all active sounds without external handles. asserts if any remain.
        static void  StopAllCallback(Asura_Handle iSoundHandle);
        static void  AutoUpdateSoundCentres();
		static void  UpdateModuleReverb();
        static void  UpdateSound(Asura_Handle handle);
        static void  UpdateGlobalPitchPerSound(Asura_Handle handle);
        static bool  PlaySoundInternal(Asura_Handle iSoundHandle, float fVolume, float fPan, float fFade);
		static bool  Calculate3dVolumePanAndFade(Asura_Handle iSoundHandle, float* pfVolume, float* pfPan, float* pfFade);
		static bool	 Calculate3dDoppler(Asura_Handle iSoundHandle);
        static float GetHardPan(unsigned int iCamID);
        static float GetLinearVolumeFalloff(float inner_range, float outer_range, float distance);
        static float GetCuboidVolumeFalloff(const struct Asura_Cuboid3dSoundData * cuboid_data, const Asura_Vector_3 * relativePosition);
		static float CalculateDistanceFilterGain (Asura_Handle iSoundHandle);

        static bool SetPlayingSoundVolumePanAndFade(Asura_Handle iPlayingSoundHandle, float fVolume, float fPan, float fFade);
		static bool SetPlayingSoundVolume(Asura_Handle iPlayingSoundHandle, float fVolume);
		static void SetSoundWetDryVolumeInternal(Asura_Handle iSoundHandle, float fWetVolume, float fDryVolume);

		//occlusion related functions		
		static float AdjustVolumeByOcclusion (const float fVolume, const float fOcclusion);
		static void UpdateOcclusionData (int iSoundHandle);

        // project specific functions
        static void  Project_Update();
        static bool  Project_Adjust3dVolumePanAndFade( const Asura_Handle iSoundHandle, float& fVolume, float& fPan, float& fFade );
        static bool  Project_Adjust2dVolume( const Asura_Handle iSoundHandle, float& fVolume );
		static bool  Project_ShouldLoadSound(Asura_Sound_ID eSoundID);
		
		// platform specific functions
		static void PlatformUpdate();

		static int PlatformInitialise(void);    // returns max active sounds
		static int Platform_GetMaxSoundsWithEQ(void);	// returns number of those sounds that can have EQ.
		static void PlatformShutdown(void);
        static void PlatformReset(void);        // platforms that handle unloading may want to
												// check that memory available at PlatformReset
        static void PlatformSetResetPoint(void);// is the same as when SetResetPoint was called.
                

        
        static void Platform_SetReverb(enum ReverbType eType, float fReverbAmount);
		static void Platform_SetEQType(Asura_Handle iHandle, EQType eType, float fLevel);
        
		static bool PlatformLoadSound( const char* filename, const int loadedSoundIndex);
		#ifdef ASURA_TOOLS
		static void PlatformForceSoundToBeLoaded(int iLoadedSoundIndex);
		static void PlatformReloadSound(int iLoadedSoundIndex);
		#endif
		static bool PlatformLoadSoundFromResource(Asura_Handle fileHandle, int fileSize, int loadedSoundIndex);
		static bool PlatformUnloadSound(const int loadedSoundIndex);
        static bool PlatformReplaceSound(const int iExistingSoundLoadedID, const int iNewSoundLoadedID);
        static bool PlatformLoadReplaceableSound(const int iLoadedSoundID);
		static int Platform_GetMaxActiveReplaceableSounds();
        static int Platform_GetReplaceableSoundSize(const int iLoadedSoundID);
			
        static bool PlatformCreateSound(int loadedSoundIndex, Asura_Handle activeSoundIndex, bool bUseEQ, bool bUseHardware3dPan);

		static bool PlatformPlaySound(Asura_Handle iSoundHandle, bool bLooping);
		static bool PlatformStopSound(Asura_Handle iSoundHandle);
        static bool PlatformDestroySound(Asura_Handle iSoundHandle);

		static bool PlatformSoundIsPlaying(Asura_Handle iSoundHandle);
		static bool PlatformSoundIsPreparing(Asura_Handle iSoundHandle);

		// volume, pan and fade are from 0.0 to 1.0 (fade is 0.0 = back, 1.0 = front)
		// fade is ignored on platforms which have no surround sound or if using HardwarePan (on PC)
        static bool PlatformSetSoundVolumePanAndFade(Asura_Handle iSoundHandle, float fVolume, float fPan, float fFade);

		//Set the distance filter gain for a sound (normalised where 0.0 is no filtering, 1.0 is max filtering)
		static bool Platform_SetSoundDistanceFilterGain (Asura_Handle iPlayingSoundHandle, float fNormalisedGain);
		
		//Wet and dry volumes are enviromental reverb on individual sounds. They scale the master volume and should be between 0.0 and 1.0
		static bool Platform_SetSoundWetDryVolume(Asura_Handle iPlayingSoundHandle, float fWetVolume, float fDryVolume);

		//Update the low pass filter for occluded sounds, Occlusion value ranges from 0 - 1
		static bool Platform_SetOcclusionLPF (const Asura_Handle iPlayingSoundHandle, const float fOcclusion);

		// Hardware panning is currently PC-only (it's the only way under DirectSound to access surround speakers)
		// All other platforms allow us to specify speaker volumes directly (or if using Dolby PL2 have some concept of front and back as well as pan).
		static bool Platform_HardwarePanSetSoundVolume(Asura_Handle iSoundHandle, float fVolume);
		static bool Platform_HardwarePanSetSoundPosition(Asura_Handle iSoundHandle, const Asura_Vector_3& xPosition);
		static bool Platform_HardwarePanSetSoundSpatial(Asura_Handle iSoundHandle, bool bIsSpatial);
		static bool Platform_HardwarePanSetSoundRadii(Asura_Handle iSoundHandle, float fInner, float fOuter);
		static bool Platform_HardwarePanSetSoundForcedPanAndFade(Asura_Handle iSoundHandle, float fPan, float fFade);
		
        static bool PlatformSetSoundPitch(Asura_Handle iSoundHandle, float fPitch);
		static bool Platform_SetSoundUsesReverb(Asura_Handle iSoundHandle, bool bUseReverb);
		static bool Platform_SetSoundUsesDistanceFilter(Asura_Handle iSoundHandle, bool bUseReverb);

        static bool PlatformPauseSound(Asura_Handle iSoundHandle);
        static bool PlatformUnpauseSound(Asura_Handle iSoundHandle);

		#ifdef ASURA_TOOLS
		static float PlatformGetSoundLength(int iLoadedSoundIndex);
		#endif

		// internal functions to manage 2 lists of playing sound handles: ones that need to have
		// EQ enabled and ones that don't care.
		static bool GetPlayingSoundHandle(Asura_Handle& iHandle, bool bNeedEQ);
		static void ReleasePlayingSoundHandle(Asura_Handle iHandle);
		static void TraversePlayingSoundHandles(void (*function)(Asura_Handle handle));
		static bool IsValidPlayingSoundHandle(Asura_Handle iHandle);

		// (AAM - 24.01.2007) Retrieve the playing sound handle used by an active sound
		static Asura_Handle GetPlayingHandleForActiveSound( const Asura_Handle iActiveSoundHandle );

		// Conversion functions from "Asura_Sound_ID" to array-index...
#if defined(ASURA_TOOLS)
		static bool AddSoundIDToHandler(const Asura_Sound_ID eSoundID, int& iIndex)		{ return(!s_xSoundIDToIndexTree.FindNodeIndex(eSoundID, true, &iIndex)); }
		static bool DeleteSoundIDFromHandler(const Asura_Sound_ID eSoundID)				{ return(s_xSoundIDToIndexTree.DeleteNode(eSoundID)); }
		static bool FindSoundIDInHandler(const Asura_Sound_ID eSoundID, int& iIndex)	{ return(s_xSoundIDToIndexTree.FindNodeIndex(eSoundID, false, &iIndex)); }
		static bool IsSoundIDInHandler(const Asura_Sound_ID eSoundID)					{ return(s_xSoundIDToIndexTree.FindNodeIndex(eSoundID, false, NULL)); }
		static u_int GetNumLoadedSounds(void)											{ return(s_xSoundIDToIndexTree.GetNumberOfNodes()); }
		static bool IsHandlerFull(void)													{ return(s_xSoundIDToIndexTree.GetNumberOfNodes() >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS); }
#else
		static bool AddSoundIDToHandler(const Asura_Sound_ID eSoundID, int& iIndex)		{ if(s_pxCurrentHandler == NULL) { return(false); } else { return(s_pxCurrentHandler->Add(eSoundID, iIndex)); } }
		static bool DeleteSoundIDFromHandler(const Asura_Sound_ID eSoundID)				{ if(s_pxCurrentHandler == NULL) { return(false); } else { return(s_pxCurrentHandler->Delete(eSoundID)); } }
		static bool FindSoundIDInHandler(const Asura_Sound_ID eSoundID, int& iIndex)	{ if(s_pxCurrentHandler == NULL) { return(false); } else { return(s_pxCurrentHandler->FindIndex(eSoundID, iIndex)); } }
		static bool IsSoundIDInHandler(const Asura_Sound_ID eSoundID)					{ if(s_pxCurrentHandler == NULL) { return(false); } else { return(s_pxCurrentHandler->IsPresent(eSoundID)); } }
		static u_int GetNumLoadedSounds(void)											{ return(Asura_Sound_Handler::GetNumLoadedSounds()); }
		static bool IsHandlerFull(void)													{ return(Asura_Sound_Handler::GetNumLoadedSounds() >= ASURA_MAX_NUMBER_OF_LOADED_SOUNDS); }
#endif	// defined(ASURA_TOOLS)

		//Gets the next dynamic sound ID to use.
		static Asura_Sound_ID GetDynamicSoundID();

#if defined(ASURA_TOOLS)
		static Asura_Sound_Handle_Tree s_xSoundIDToIndexTree;
#endif

        static bool s_bUseAutomaticSoundCentre;
        static int s_iNumSplitScreens;
        static Asura_Vector_3 s_axSoundCentre[ASURA_SOUND_MAX_SPLIT_SCREENS];
        static Asura_Matrix_3x3 s_axSoundOrientation[ASURA_SOUND_MAX_SPLIT_SCREENS];
		static Asura_Vector_3 s_axSoundVelocity[ASURA_SOUND_MAX_SPLIT_SCREENS];

        static class Asura_ActiveSoundSample* s_pxActiveSoundList;
        static Asura_Tracking_Handle_List* s_pxActiveSoundHandleList;
        static Asura_Tracking_Handle_List* s_pxPlayingSoundHandleListWithEQ;
		static Asura_Tracking_Handle_List* s_pxPlayingSoundHandleListWithoutEQ;
        static int s_iMaxPlayingSounds;
		static int s_iMaxPlayingSoundsWithEQ;
        static SoundOutputType s_eSoundOutputType;
		static ReverbType s_eCurrentReverbType;
		static float s_fCurrentReverbAmount;
		static EQType s_aeEQType[ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS];
		static float s_afEQLevel[ASURA_SOUND_MAX_NUMBER_OF_EQ_FILTERS];
		static bool s_bUseHardware3dPan;



		//array of sound properties corresponding to the indeces of loaded sounds.
		static Asura_SoundProperties* s_pxSoundProperties;
		static Asura_Hash_ID* s_pxSoundHashIDs;

#if ASURA_SOUND_STORE_FILENAMES
		static Asura_String* s_pxSoundFilenames;
        static Asura_Sound_ID* s_peNamedSoundIDArray;
#endif

        static float s_fPitchScale;
        static bool s_bPitchUsesGameTimeScale;

		static bool s_bMuteSoundFX;
		static bool s_bMuteMusic;
		static bool s_b3dSoundEQDefault;
		static Asura_Tracking_Handle_List* s_pxEQHandleList;

		/// Kevin Lea - 25/04/2002 11:06:18 - streaming sound platform methods
		/// initialise/shutdown should be handled in PlatformInitialise/PlatformShutdown
        // (MH - 14.01.2003) iHandle should be between 0 and Platform_GetMaxStreamingSounds() - 1
        // for each platform.
        static int  Platform_GetMaxStreamingSounds();
		static bool Platform_CreateStreamingSound(Asura_Handle iHandle, const char* szFilename);
		static void Platform_PlayStreamingSound(Asura_Handle iHandle, bool bLooping, bool bPreload);
		static void Platform_StopStreamingSound(Asura_Handle iHandle);
		static void Platform_PauseStreamingSound(Asura_Handle iHandle);
		static void Platform_ResumeStreamingSound(Asura_Handle iHandle);
		static void Platform_ReleaseStreamingSound(Asura_Handle iHandle);
        static bool Platform_StreamingSoundIsReady(Asura_Handle iHandle);
        static void Platform_StartPreloadedStreamingSound(Asura_Handle iHandle);
        static void Platform_SetStreamingSoundOutput(Asura_Handle iHandle, SoundOutputType eOutputType);
        static void Platform_EnsureAllStreamingSoundsAreStopped();
        static void Platform_SetStreamingSoundVolume(Asura_Handle iHandle, float fVolume);
		static bool Platform_StreamingSoundIsFinished(Asura_Handle iHandle);
		static float Platform_GetStreamingSoundPosition(Asura_Handle iHandle);
		static float Platform_GetStreamingSoundLength(Asura_Handle iHandle);
		#ifdef ASURA_TOOLS
		static void Platform_SetStreamingSoundPosition(Asura_Handle iHandle,float fTime);
		#endif

		static Asura_Tracking_Handle_List* s_pxStreamingSoundList;
		static int s_iNumberOfStreamingSounds;

		static u_int s_uNextDynamicSIDIndex;
		static u_int s_uNumSystemPauses;

		//occlusions
		static Asura_Sound_OcclusionFactor_Callback s_pfnGetEntityOcclusionFactor;
		static Asura_Sound_OcclusionFactor_Callback s_pfnGetEnvironmentOcclusionFactor;

		static Asura_TickCount s_xOcclusionCPUTicksPerFrame;

		static Asura_FixedArray<Asura_Guid, ASURA_MAX_NO_OF_CAMERAS> s_auOcclusionGUIDsToIgnore;
};

class Asura_SoundProperties
{
public :
	void Initialise()
	{
		m_iLoadedSoundIndexMapping = ASURA_INVALID_LOADED_SOUND_INDEX;
		m_uHashID = ASURA_HASH_ID_UNSET;
		m_fPitchScale = 1.0f;
		m_fVolumeScale = 1.0f;
		m_fDefaultInnerRadius = fASURA_SOUND_DEFAULT_INNER_RADIUS;
		m_fDefaultOuterRadius = fASURA_SOUND_DEFAULT_OUTER_RADIUS;

	}

	//Remapping for loaded sound index for sounds that share the same 'wav' file as
	//a previously loaded one.
	int m_iLoadedSoundIndexMapping;

	//hash value from sound filename
	Asura_Hash_ID m_uHashID;
	
	float m_fPitchScale;
	float m_fVolumeScale;
	float m_fDefaultInnerRadius;
	float m_fDefaultOuterRadius;
};

class Asura_Message_SoundReverb
{
public:
	int	m_iReverbID;
};

#endif// ASURA_USE_AUDIO_SYSTEM

#endif//__ASURA_SOUND_H__
