//
// Asura_Sound_DialogueController.cpp
//
// Created by Richard Rice - 11/11/2004
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

//////////////////////////////////////////////////////////////////////////////
//
// includes
//
//////////////////////////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"
#include "Debugging/Asura_Logfile.h"

// header
#include "Sound/Asura_Sound_DialogueController.h"

// asura
#include "Entity/Sound/Asura_ClientEntity_Tannoy.h"
#include "Sound/Asura_Sound.h"

//////////////////////////////////////////////////////////////////////////////
//
// implementation of support classes for the dialogue controller
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// DialogueMemory
//
//////////////////////////////////////////////////////////////////////////////

void DialogueMemory::DestroyAndReplaceDialogue()
{
    const Asura_Sound_ID eCurrentSound = Asura_Sound_System::GetSoundIDFromSoundHandle( m_iSoundHandle );
    const Asura_Sound_ID eSoundToLoad  = GetSoundToLoad();
    Asura_Assert( eCurrentSound != ASURA_SID_NO_SOUND, "" );
    Asura_Assert( eSoundToLoad != ASURA_SID_NO_SOUND, "" );

    Asura_Sound_System::DestroySound( m_iSoundHandle );

    bool bShuffle = true;
    if ( eSoundToLoad != eCurrentSound && !Asura_Sound_System::ReplaceSound( eCurrentSound, eSoundToLoad ) )
    {
        bShuffle = false;
    }
    if ( bShuffle )
    {
        ShuffleSoundIndexData();
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// TannoyDialogueMemory
//
//////////////////////////////////////////////////////////////////////////////

TannoyDialogueMemory TannoyDialogueMemory::s_axDialogueMemory[ DSC_PER_MEMORY_TYPE_MEMORY_SIZE ];

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

TannoyDialogueMemory::TannoyDialogueMemory()
: m_eSoundID( ASURA_SID_NO_SOUND )
, m_bPositionValid( false )
, m_xPosition( Asura_Maths::ZeroVector3 )
, m_fInnerRadius( 0.0f )
, m_fOuterRadius( 0.0f )
{
}

TannoyDialogueMemory::~TannoyDialogueMemory()
{
}

void TannoyDialogueMemory::Clear()
{
    DialogueMemory::Clear();

    m_eSoundID       = ASURA_SID_NO_SOUND;
    m_bPositionValid = false;
    m_xPosition      = Asura_Maths::ZeroVector3;
    m_fInnerRadius   = 0.0f;
    m_fOuterRadius   = 0.0f;
    m_uProjectFlags  = 0;
}

Asura_Sound_ID TannoyDialogueMemory::GetSoundToPlay() const
{
    return m_eSoundID;
}

Asura_Sound_ID TannoyDialogueMemory::GetSoundToLoad() const
{
    return m_eSoundID;
}

unsigned int TannoyDialogueMemory::GetTextIdToShow( bool& bTextDataSet ) const
{
    bTextDataSet = true;
    return static_cast< unsigned int >( m_eSoundID - ( ASURA_SID_TANNOY_START + 1 ) );
}

Asura_Hash_ID TannoyDialogueMemory::GetTextPageHashId() const
{
    return 0xCB7CBB97; //Asura_GetHashID( "tannoy" );
}

Asura_Handle TannoyDialogueMemory::CreateSound( const Asura_Sound_ID eSoundToPlay ) const
{
    Asura_Handle iSoundHandle;
    if ( m_bPositionValid )
    {
        iSoundHandle = Asura_Sound_System::Create3dSound( eSoundToPlay,
                                                          &m_xPosition,
                                                          m_fOuterRadius,
                                                          m_fInnerRadius );
    }
    else
    {
        iSoundHandle = Asura_Sound_System::CreateSound( eSoundToPlay );
        if ( iSoundHandle != ASURA_INVALID_HANDLE )
        {
            //## (NJD - 08.04.2003) - HELP! How do we know which player to play this at?
            Asura_Sound_System::SetSoundHardPan( iSoundHandle, 0 );
        }
    }
    return iSoundHandle;
}

TannoyDialogueMemory* TannoyDialogueMemory::GetFreePointer()
{
    for ( int i = 0; i < DSC_PER_MEMORY_TYPE_MEMORY_SIZE; ++i )
    {
        if ( s_axDialogueMemory[ i ].IsFree() )
        {
            return &s_axDialogueMemory[ i ];
        }
    }
    return 0;
}

bool TannoyDialogueMemory::AnyPlaying()
{
    for ( int i = 0; i < DSC_PER_MEMORY_TYPE_MEMORY_SIZE; ++i )
    {
        if ( !s_axDialogueMemory[ i ].IsFree() )
        {
            return true;
        }
    }
    return false;
}

bool TannoyDialogueMemory::AnyPlayingWithFlag( const u_int uFlags )
{
    for ( int i = 0; i < DSC_PER_MEMORY_TYPE_MEMORY_SIZE; ++i )
    {
        if ( s_axDialogueMemory[ i ].IsPlaying() )
        {
            if ( s_axDialogueMemory[ i ].m_uProjectFlags & uFlags )
            {
                return true;
            }
        }
    }

    return false;
}

bool TannoyDialogueMemory::AnyPlayingWithHandle( u_int& uFlags, const Asura_Handle iAsuraSoundHandle )
{
    for ( int i = 0; i < DSC_PER_MEMORY_TYPE_MEMORY_SIZE; ++i )
    {
        if ( s_axDialogueMemory[ i ].IsPlaying() &&
             s_axDialogueMemory[ i ].m_iSoundHandle == iAsuraSoundHandle )
        {
            uFlags = s_axDialogueMemory[ i ].m_uProjectFlags;
            return true;
        }
    }

    return false;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_DialogueController
//
//////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// static data
//
//////////////////////////////////////////////////////////////////////////////

int             Asura_Sound_DialogueController::s_iPlayingSounds = 0;
unsigned char   Asura_Sound_DialogueController::s_ucMemorySize   = 0;
DialogueMemory* Asura_Sound_DialogueController::s_apxDialogueMemory[ DSC_MEMORY_SIZE ];
unsigned char   Asura_Sound_DialogueController::s_ucMemoryIndex  = 0;
bool Asura_Sound_DialogueController::s_bPitchUsesTimeScale = true;
float Asura_Sound_DialogueController::s_fTimeScaleToUse = 1.0f;

//////////////////////////////////////////////////////////////////////////////
//
// public methods
//
//////////////////////////////////////////////////////////////////////////////

void Asura_Sound_DialogueController::Initialise()
{
	Asura_Assert(s_ucMemorySize == 0, "dialogue controller wasn't shut down or handles not released?");

    LoadTannoySounds();

    s_iPlayingSounds = 0;

    // check we've got enough space in memory
    // there should be enough space for all streams to be of the same type (to handle the worst case scenario)
    Asura_Assert( Asura_Sound_System::GetMaxActiveReplaceableSounds() <= DSC_PER_MEMORY_TYPE_MEMORY_SIZE,
                  "Asura_Sound_DialogueController::Initialise - insufficient memory space for worst case!" );

    for ( int iMemoryIndex = 0; iMemoryIndex < DSC_MEMORY_SIZE; ++iMemoryIndex )
    {
        s_apxDialogueMemory[ iMemoryIndex ] = 0;
    }
    s_ucMemorySize  = 0;
    s_ucMemoryIndex = 0;
}

void Asura_Sound_DialogueController::Shutdown()
{
    // (MH - 03.08.2006) clean up any sounds which are ready for deletion
    for ( int i = 0; i < DSC_MEMORY_SIZE; ++i )
    {
        if ( s_apxDialogueMemory[ i ] )
        {
			if ( s_apxDialogueMemory[ i ]->IsReadyForDeletion() )
            {
                // our handle user now knows that the handle is invalid so it's safe for us to get rid of it
                s_apxDialogueMemory[ i ]->ReturnThis();
                s_apxDialogueMemory[ i ] = 0;
                --s_ucMemorySize;
            }
        }
    }

	Asura_Assert(s_ucMemorySize == 0, "dialogue controller handles not released?");
}

void Asura_Sound_DialogueController::Update()
{
    // go through the sound memory and check if any sounds have finished playing
    for ( int i = 0; i < DSC_MEMORY_SIZE; ++i )
    {
        if ( s_apxDialogueMemory[ i ] )
        {
            if ( s_apxDialogueMemory[ i ]->IsPlaying() )
            {
                Asura_Assert( s_apxDialogueMemory[ i ]->m_iSoundHandle != ASURA_INVALID_HANDLE, "" );

                if ( !Asura_Sound_System::SoundIsPlaying( s_apxDialogueMemory[ i ]->m_iSoundHandle ) )
                {
                    // sound has finished - delete the old sound and load the new in Asura
                    s_apxDialogueMemory[ i ]->DestroyAndReplaceDialogue();
                    s_apxDialogueMemory[ i ]->SoundFinished();
                    --s_iPlayingSounds;
                    Asura_Assert( s_iPlayingSounds >= 0, "" );
                }
                else
                {
                    // sound is still playing
                    s_apxDialogueMemory[ i ]->UpdateSoundWorthiness();
                }
            }
            else if ( s_apxDialogueMemory[ i ]->IsReadyForDeletion() )
            {
                // our handle user now knows that the handle is invalid so it's safe for us to get rid of it
                s_apxDialogueMemory[ i ]->ReturnThis();
                s_apxDialogueMemory[ i ] = 0;
                --s_ucMemorySize;
            }
        }
    }
}

Asura_Handle Asura_Sound_DialogueController::PlayTannoyDialogue( const Asura_Sound_ID& eSoundID,
                                                                 const float& fInnerRadius,
                                                                 const float& fOuterRadius,
                                                                 const float& fVolume,
                                                                 const u_int& uProjectFlags,
                                                                 const Asura_Vector_3* pxPosition )
{
    if ( eSoundID != ASURA_SID_NO_SOUND )
    {
        TannoyDialogueMemory* pxTannoyDialogueMemory = TannoyDialogueMemory::GetFreePointer();
        if ( pxTannoyDialogueMemory != NULL )
        {
            pxTannoyDialogueMemory->m_eSoundID = eSoundID;
            if ( pxPosition != NULL )
            {
                pxTannoyDialogueMemory->m_bPositionValid = true;
                pxTannoyDialogueMemory->m_xPosition      = *pxPosition;
            }
            else
            {
                pxTannoyDialogueMemory->m_bPositionValid = false;
            }
            pxTannoyDialogueMemory->m_fWorthiness   = 0.98f;
            pxTannoyDialogueMemory->m_fInnerRadius  = fInnerRadius;
            pxTannoyDialogueMemory->m_fOuterRadius  = fOuterRadius;
            pxTannoyDialogueMemory->m_uProjectFlags = uProjectFlags;

            bool bCancelPlaying = false;
            
            // if we shouldn't play the sound then cancel
            if ( !pxTannoyDialogueMemory->Project_ShouldPlayThisSound() )
            {
                bCancelPlaying = true;
            }
            
            // if we've run out of room then cancel
            if ( PlatformSoundLimitReached() )
            {
                bCancelPlaying = true;
            }

            pxTannoyDialogueMemory->SoundInitialising();

            // finally, try to play the requested dialogue
            Asura_Handle iAsuraHandle = ASURA_INVALID_HANDLE;
            if ( !bCancelPlaying )
            {
                iAsuraHandle = PlaySound( pxTannoyDialogueMemory, fVolume );
            }
			
			#if !ASURA_TANNOY_USEGAMETIMESCALE
            if ( iAsuraHandle != ASURA_INVALID_HANDLE )
			{
				Asura_Sound_System::SetSoundPitchUsesTimeScale(iAsuraHandle,false);
			}
			#endif

            // if we haven't managed to play the dialogue for whatever reason then free the resources
            if ( iAsuraHandle == ASURA_INVALID_HANDLE )
            {
                pxTannoyDialogueMemory->ReturnThis();
            }
            return iAsuraHandle;
        }
        else
        {
            Asura_Assert( s_iPlayingSounds == Asura_Sound_System::GetMaxActiveReplaceableSounds(),
                          "Asura_Sound_DialogueController::PlayDialogue - "
                          "run out of tannoy dialogue memories before running out of streaming sounds" );
            return ASURA_INVALID_HANDLE;
        }
    }
    else
    {
        return ASURA_INVALID_HANDLE;
    }
}

void Asura_Sound_DialogueController::UpdateDialoguePosition( const Asura_Handle& iDialogueHandle,
                                                             const Asura_Vector_3& xPosition )
{
    int        iAssociatedIndex = 0;
    const bool bValidIndex      = GetIndexFromHandle( iDialogueHandle, iAssociatedIndex );
    if ( bValidIndex && s_apxDialogueMemory[ iAssociatedIndex ] && s_apxDialogueMemory[ iAssociatedIndex ]->IsPlaying() )
    {
        Asura_Assert( s_apxDialogueMemory[ iAssociatedIndex ]->m_iSoundHandle != ASURA_INVALID_HANDLE, "" );
        s_apxDialogueMemory[ iAssociatedIndex ]->UpdatePosition( xPosition );
    }
}

void Asura_Sound_DialogueController::UpdateDialogueHandle( Asura_Handle& iDialogueHandle )
{
    // check to see if the sound has finished playing. if it has then set the handle as invalid
    if ( iDialogueHandle != ASURA_INVALID_HANDLE )
    {
        int        iAssociatedIndex = 0;
        const bool bValidIndex      = GetIndexFromHandle( iDialogueHandle, iAssociatedIndex );
        if ( bValidIndex && s_apxDialogueMemory[ iAssociatedIndex ] && s_apxDialogueMemory[ iAssociatedIndex ]->IsFinished() )
        {
            iDialogueHandle = ASURA_INVALID_HANDLE;
            s_apxDialogueMemory[ iAssociatedIndex ]->SoundReadyForDeletion();
        }
    }
}

void Asura_Sound_DialogueController::DestroyDialogue( Asura_Handle& iDialogueHandle )
{
    if( iDialogueHandle == ASURA_INVALID_HANDLE ) { return; }

    // look up the Asura_Sound_ID from the Dialogue_ID and destroy the sound in Asura_Sound_System
    int iAssociatedIndex = 0;
    if ( GetIndexFromHandle( iDialogueHandle, iAssociatedIndex ) )
    {
        Asura_Assert( s_apxDialogueMemory[ iAssociatedIndex ], "" );
        if ( !s_apxDialogueMemory[ iAssociatedIndex ] )
        {
            return;
        }

        if ( s_apxDialogueMemory[ iAssociatedIndex ]->IsPlaying() )
        {
            // delete the old sound and load the new in Asura
            s_apxDialogueMemory[ iAssociatedIndex ]->DestroyAndReplaceDialogue();
            --s_iPlayingSounds;
            Asura_Assert( s_iPlayingSounds >= 0, "" );
        }

        iDialogueHandle = ASURA_INVALID_HANDLE;
        s_apxDialogueMemory[ iAssociatedIndex ]->SoundReadyForDeletion();
    }
}

void Asura_Sound_DialogueController::DestroyAllDialogue()
{
    for ( int iMemoryIndex = 0; iMemoryIndex < DSC_MEMORY_SIZE; ++iMemoryIndex )
    {
        if ( s_apxDialogueMemory[ iMemoryIndex ] && s_apxDialogueMemory[ iMemoryIndex ]->IsPlaying() )
        {
            // delete the old sound and load the new in Asura
            s_apxDialogueMemory[ iMemoryIndex ]->DestroyAndReplaceDialogue();
            s_apxDialogueMemory[ iMemoryIndex ]->SoundFinished();
            --s_iPlayingSounds;
            Asura_Assert( s_iPlayingSounds >= 0, "" );
        }
    }
}

Asura_Handle Asura_Sound_DialogueController::GetNextAsuraSoundHandle( const bool bReset )
{
    if ( bReset )
    {
        s_ucMemoryIndex = 0;
    }
    else if ( s_ucMemoryIndex <= s_ucMemorySize )
    {
        ++s_ucMemoryIndex;
    }

    if ( s_ucMemoryIndex < s_ucMemorySize && s_apxDialogueMemory[ s_ucMemoryIndex ] != 0 )
    {
        return s_apxDialogueMemory[ s_ucMemoryIndex ]->m_iSoundHandle;
    }
    else
    {
        return ASURA_INVALID_HANDLE;
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// private methods
//
//////////////////////////////////////////////////////////////////////////////

bool Asura_Sound_DialogueController::PlatformSoundLimitReached()
{
    Asura_Assert( 0 <= s_iPlayingSounds && s_iPlayingSounds <= Asura_Sound_System::GetMaxActiveReplaceableSounds(), "RT_Sound_DialogueController - s_iPlayingSounds is invalid size" );
    return ( s_iPlayingSounds == Asura_Sound_System::GetMaxActiveReplaceableSounds() );
}

bool Asura_Sound_DialogueController::ExpelWorseSound( const float fWorthiness )
{
    int   iWorstIndex  = -1;
    float fLeastWorthy = 1.1f;

    for ( int i = 0; i < DSC_MEMORY_SIZE; ++i )
    {
        if ( s_apxDialogueMemory[ i ] &&
            s_apxDialogueMemory[ i ]->IsPlaying() &&
            s_apxDialogueMemory[ i ]->m_fWorthiness < fWorthiness &&
            s_apxDialogueMemory[ i ]->m_fWorthiness < fLeastWorthy )
        {
            fLeastWorthy = s_apxDialogueMemory[ i ]->m_fWorthiness;
            iWorstIndex  = i;
        }
    }

    if ( iWorstIndex != -1 )
    {
        // delete the old sound and load the new in Asura
        s_apxDialogueMemory[ iWorstIndex ]->DestroyAndReplaceDialogue();
        s_apxDialogueMemory[ iWorstIndex ]->SoundFinished();
        --s_iPlayingSounds;
        Asura_Assert( s_iPlayingSounds >= 0, "" );
        return true;
    }
    else
    {
        // we expect this to happen every so often
        return false;
    }
}

Asura_Handle Asura_Sound_DialogueController::PlaySound( DialogueMemory* pxDialogueMemory, float fVolume /*=1.0f*/)
{
    const Asura_Sound_ID eSoundToPlay = pxDialogueMemory->GetSoundToPlay();
    const Asura_Hash_ID  uHashId      = pxDialogueMemory->GetTextPageHashId();
    bool                 bTextDataSet = false;
    const unsigned int   uTextId      = pxDialogueMemory->GetTextIdToShow( bTextDataSet );
    bTextDataSet &= ( uHashId == ASURA_HASH_ID_UNSET ) ? false : true;

    if ( eSoundToPlay == ASURA_SID_NO_SOUND ) { return ASURA_INVALID_HANDLE; }
    if ( !Asura_Sound_System::SoundIsLoaded( eSoundToPlay ) ) { return ASURA_INVALID_HANDLE; }

    if ( RoomInMemory() )
    {
        Asura_Handle iSoundHandle = pxDialogueMemory->CreateSound( eSoundToPlay );
        if ( iSoundHandle != ASURA_INVALID_HANDLE )
        {
            if ( Asura_Sound_System::PlaySound( iSoundHandle, false ) )
            {
                if ( fVolume != 1.0f )
                {
                    Asura_Sound_System::SetSoundVolume( iSoundHandle, fVolume );
                }
                
                ++s_iPlayingSounds;

                if ( bTextDataSet )
                {
                    Project_SendShowTextMessage( uHashId, uTextId, pxDialogueMemory->m_uProjectFlags );
                }

				Asura_Sound_System::SetSoundPitchUsesTimeScale(iSoundHandle, s_bPitchUsesTimeScale, s_fTimeScaleToUse);

                return AddSoundToMemory( iSoundHandle, pxDialogueMemory );
            }
            else
            {
                Asura_Sound_System::FinishedWithSoundHandle( iSoundHandle );
                return ASURA_INVALID_HANDLE;
            }
        }
        else
        {
            return ASURA_INVALID_HANDLE;
        }
    }
    else
    {
        Asura_Assert( 0, "Asura_Sound_DialogueController::PlaySound - run out of memory" );
        return ASURA_INVALID_HANDLE;
    }
}

Asura_Handle Asura_Sound_DialogueController::AddSoundToMemory( const Asura_Handle& iSoundHandle, DialogueMemory* pxDialogueMemory )
{
    for ( int i = 0; i < DSC_MEMORY_SIZE; ++i )
    {
        if ( s_apxDialogueMemory[ i ] == 0 )
        {
            s_apxDialogueMemory[ i ] = pxDialogueMemory;
            s_apxDialogueMemory[ i ]->m_iSoundHandle = iSoundHandle;
            s_apxDialogueMemory[ i ]->SoundPlaying();

            ++s_ucMemorySize;
            return i;
        }
    }

    // (NJD - 17.03.2003) - we should never get to this because we've already checked for space
    Asura_Assert( 0, "Asura_Sound_DialogueController::AddSoundToMemory - run out of memory - THIS ASSERT SHOULD NEVER HAPPEN!!" );
    return ASURA_INVALID_HANDLE;
}

void Asura_Sound_DialogueController::LoadTannoySounds()
{
    Asura_ClientEntity_Tannoy::LoadTannoySounds();
}

void Asura_Sound_DialogueController::SetDialoguePitchUsesTimeScale(bool bUseTimeScale, float fTimeScaleToUse/* = 1.0f*/)
{
	if ((s_bPitchUsesTimeScale != bUseTimeScale) || (!bUseTimeScale && (s_fTimeScaleToUse != fTimeScaleToUse)))
	{
		for ( int i = 0; i < DSC_MEMORY_SIZE; ++i )
		{
			if ( s_apxDialogueMemory[ i ] )
			{
				Asura_Sound_System::SetSoundPitchUsesTimeScale(s_apxDialogueMemory[ i ]->m_iSoundHandle, bUseTimeScale, fTimeScaleToUse);
			}
		}
		s_bPitchUsesTimeScale = bUseTimeScale;
		s_fTimeScaleToUse = fTimeScaleToUse;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
