//////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_Event.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     internal header for sound event implementation
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_SOUND_EVENT_H___
#define ___ASURA_SOUND_EVENT_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_HeapArray.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#if !defined(ASURA_USE_AUDIO_SYSTEM)
#include "Sound/Asura_ProjectSoundIDs.h"	// TODO get rid of this
#endif	//!defined(ASURA_USE_AUDIO_SYSTEM)

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Sound_Mixer_Channel;

// D E F I N E S /////////////////////////////////////////////////////////////
#ifndef ASURA_SOUND_INCLUDE_TOOLS_DATA
#if defined(ASURA_TOOLS) || defined(_DEBUG)
#define ASURA_SOUND_INCLUDE_TOOLS_DATA	1	// text names, sound categories etc.
#endif
#endif

// JAT 12/09/06 allow projects to control the max for simultaneous play value
#ifndef ASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY
#define ASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY 4
#endif // #ifndef ASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY

// C O N S T A N T S /////////////////////////////////////////////////////////
static const float fASURA_SOUND_EVENT_VOLUME_MAX = 1.333f;
static const float fASURA_SOUND_EVENT_PRIORITY_MIN = 0.0f;
static const float fASURA_SOUND_EVENT_PRIORITY_MAX = 2.0f;
static const u_int uASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY = ASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY;
static const u_int uASURA_SOUND_EVENT_MAX_CONTROLLERS = 128;

// E N U M S /////////////////////////////////////////////////////////////////
enum ASURA_SOUND_PLAYBACK_TYPE
{
	ASURA_PT_SIMULTANEOUS,		// all at once
	ASURA_PT_ALLSEQUENTIAL,		// one at a time triggering next one when finished
	ASURA_PT_ONEBYONE,			// current one in list, stopping and advancing in list when finished 
	ASURA_PT_RANDOM,			// pick at random
	ASURA_PT_CDPLAYERRANDOM,	// like a CD random function (plays all in random order)
	ASURA_PT_VARIABLE,			// volume/pitch of samples is randomly varied during playback. Otherwise treated the same as simultaneous.
};

enum ASURA_3DSOUND_TYPE
{
	ASURA_SOUND_POINT,			// pan relative to point position, 3d volume
	ASURA_SOUND_SOLID,			// no pan inside inner range, 3d volume
	ASURA_SOUND_NOPAN,			// no pan, 3d volume
	ASURA_SOUND_DIRECT,			// no pan or volume modifiers (take from event)
	ASURA_SOUND_RANDOMPOS,		// like "direct" but with random "head-relative" pan
};

enum TOOLS_SOUND_SAMPLE_QUALITY
{
	TSS_QUALITY_USE_DEFAULT	= -1, //If set in a sample then that sample uses the default quality for the platform

	TSS_QUALITY_PERFECT 	= 0,
	TSS_QUALITY_HIGH		= 1,
	TSS_QUALITY_MEDIUM		= 2,
	TSS_QUALITY_LOW			= 3,
	TSS_QUALITY_VERY_LOW	= 4,

	TSS_QUALITY_NUM_ENTRIES
};

//playable class types for use in the playable ID class
enum ASURA_SOUND_PLAYABLE_TYPE
{
	ASURA_SOUND_PLAYABLE_SAMPLE,
	ASURA_SOUND_PLAYABLE_EVENT,
};

// C L A S S E S /////////////////////////////////////////////////////////////

// positioning in surround sound field
class Asura_SurroundPan
{
public:

	Asura_SurroundPan()
	{
		m_fPan = 0.5f;	// centre
		m_fFade = 1.0f;	// front
	}

	Asura_SurroundPan& operator += (const Asura_SurroundPan& xRHS)
	{
		m_fPan = (m_fPan + xRHS.m_fPan) * 0.5f;
		m_fFade = (m_fFade + xRHS.m_fFade) * 0.5f;
		return *this;
	}

	// public data
	float m_fPan;	// left-to-right
	float m_fFade;	// back-to-front
};


// parameters required to actually play a sample
class Asura_Sound_PlayParams
{
public:

	void Initialise()
	{
		m_fVolume = 1.0f;
		m_fPitch = 1.0f;
		m_fDelay = 0.0f;
		m_bForcedPan = false;
		m_bUseOcclusion = true;
		m_bUseDistanceFilter = true;
		m_bUseDoppler = true;
		m_bUseEnvReverb = true;
		m_bUseMasterDSP = true;
		m_pxMixingChannel = NULL;
		m_uDSPChain = ASURA_HASH_ID_UNSET;
	}

	// public data
	float								m_fVolume;				// multiply values in sample and events
	float								m_fPitch;				// multiply values in sample and events
	float								m_fDelay;				// add values in events
	Asura_SurroundPan					m_xSurroundPan;			// if event forced pan is not set, leave untouched
																// else (if m_bForcedPan already set add, else set from event)
	bool								m_bForcedPan;			// OR of values in events
	bool								m_bUseOcclusion;		// AND of values in events
	bool								m_bUseDistanceFilter;	// AND of values in events
	bool								m_bUseDoppler;			// AND of values in events
	bool								m_bUseEnvReverb;		// AND of values in events
	bool								m_bUseMasterDSP;		// AND of values in events
	const Asura_Sound_Mixer_Channel*	m_pxMixingChannel;		// overwrite if not NULL
	Asura_Hash_ID						m_uDSPChain;
};

// parameters for sound events with varying pitch/volume
class Asura_Sound_VariableParams
{
public :
	void Initialise()
	{
		m_fMinInterval = 1.0f;
		m_fMaxInterval = 1.0f;
		m_fMinVolumeChangeTime = 1.0f;
		m_fMaxVolumeChangeTime = 1.0f;
		m_fMinPitchChangeTime = 1.0f;
		m_fMaxPitchChangeTime = 1.0f;
	}

	float	m_fMinInterval;//Minimum time until next volume/pitch is chosen
	float	m_fMaxInterval;//Maximum time until next volume/pitch is chosen
	float	m_fMinVolumeChangeTime;//Minimum time over which volume should interpolate to next value (as proportion of the interval time)
	float	m_fMaxVolumeChangeTime;//Maximum time over which volume should interpolate to next value (as proportion of the interval time)
	float	m_fMinPitchChangeTime;//Minimum time over which pitch should interpolate to next value (as proportion of the interval time)
	float	m_fMaxPitchChangeTime;//Maximum time over which pitch should interpolate to next value (as proportion of the interval time)

};


// abstract base class for playable objects
class Asura_Sound_Playable
{
public:

	Asura_Sound_Playable() {}
	virtual ~Asura_Sound_Playable() {}

	// FillInPlayParams() fills out the play parameters structure
	virtual void FillInPlayParams( Asura_Sound_PlayParams& xParams ) const = 0;

	virtual bool IsMetaEvent() const = 0;
	virtual bool IsSample() const = 0;
	bool IsEvent() const { return !IsSample(); }

private:

	// non-copyable
	Asura_Sound_Playable(const Asura_Sound_Playable&);
	Asura_Sound_Playable& operator=(const Asura_Sound_Playable&);
};

//platform specific settings for sound samples. RAR 11/08/2006
class Asura_Sound_Sample_PlatformSettings
{
public :
	void Initialise(int iPlatform)
	{
		m_iPlatform = iPlatform;
		m_iSampleRate = 0;
	}

	int m_iPlatform;//Platform this is for. Matches enumeration in Tools_Level_Export/Tools_Platform.
	int m_iSampleRate;//Sample rate for wav conversion. 0 means use default values, also used for compression quality
};

// basic sample data
class Asura_Sound_Sample : public Asura_Sound_Playable
{
	friend class Asura_Chunk_SoundEvent_Sample;
	friend class Asura_Sound_Event_Storage;
private :
	Asura_Sound_Sample(const char* szName);//needs to be allocated by Asura_Sound_Event_Storage::CreateSample()

public:
	// destructor needs to check no events still reference this
	~Asura_Sound_Sample();

	// virtual function overrides
	// FillInPlayParams() should modify the play params by this sample's params.
	virtual void FillInPlayParams( Asura_Sound_PlayParams& xParams ) const
	{
		xParams.m_fVolume *= m_fVolume;
		xParams.m_fPitch *= m_fPitch;
	}

	virtual bool IsMetaEvent() const { return false; }
	virtual bool IsSample() const { return true; }

	// non virtual functions
#ifdef ASURA_USE_AUDIO_SYSTEM
	Asura_Hash_ID GetWaveHash() const { return m_uWave; }
#else
	Asura_Sound_ID GetSoundID() const { return m_eSoundID; }
#endif

	bool GetLooping() const { return m_bLooping; }

	Asura_Hash_ID GetHashID() const { return m_uHashID; }

#ifdef ASURA_TOOLS	// tools-only data
	const char* GetName() const { return m_xSampleName.GetString(); }
	const char* GetSoundType() const { return m_xSoundType.GetString(); }
	const char* GetSoundSubType() const { return m_xSoundSubType.GetString(); }
	const char* GetFilename() const { return m_xFilename.GetString(); }
	int GetSampleRate(int iPlatform) const;
	int GetCompressionQuality( const int iPlatform ) const;

	void SetSoundType( const char* szSoundType ) { m_xSoundType = Asura_String(szSoundType); }
	void SetSoundSubType( const char* szSoundSubType ) { m_xSoundSubType = Asura_String(szSoundSubType); }
	void SetFilename(const char* szFilename) { m_xFilename = szFilename; }
	void SetSampleRate(int iPlatform,int iSampleRate);
	void SetCompressionQuality( const int iPlatform, const int iCompressionQuality );
#endif

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	//Gets the handler for the zone that this sample was created in , so it can look in the right place for the sound. RAR 20/10/2006
	class Asura_Sound_Handler* GetSoundHandler() const {return m_pxSoundHandler;}
#endif

public:

	// public data
#ifdef ASURA_USE_AUDIO_SYSTEM
	Asura_Hash_ID	m_uWave;
#else
	Asura_Sound_ID	m_eSoundID;
#endif

	float			m_fVolume;
	float			m_fPitch;
	float			m_fDuration;
	bool			m_bLooping;
#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA
	Asura_String	m_xSampleName;	// must be unique, converted to hashID
	Asura_String	m_xSoundType;	// e.g. weapons, footsteps, environmental
	Asura_String	m_xSoundSubType;
	Asura_String	m_xFilename;   //File that this sample gets saved to.
#endif
	Asura_Hash_ID	m_uHashID;


private :
#ifdef ASURA_TOOLS
	const Asura_Sound_Sample_PlatformSettings* GetPlatformSettings(int iPlatform) const;
	Asura_Sound_Sample_PlatformSettings* GetPlatformSettings(int iPlatform,bool bCreate);

	int m_iNumPlatforms;
	Asura_Sound_Sample_PlatformSettings* m_pxPlatformSettings;
#endif
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	class Asura_Sound_Handler* m_pxSoundHandler;
#endif
};

//playable ID is used to identify the sounds/events during loading and saving. RAR 10/07/2006
class Asura_Sound_PlayableID
{
public :
	ASURA_SOUND_PLAYABLE_TYPE m_eType;
	Asura_Hash_ID m_uHashID; 

	inline bool operator == (const Asura_Sound_PlayableID& xID) const
	{
		return m_eType == xID.m_eType && m_uHashID == xID.m_uHashID;
	}	
	inline bool operator != (const Asura_Sound_PlayableID& xID) const
	{
		return !(*this == xID);
	}
};


// event - references to Asura_Sound_Playable* and Asura_Sound_Mixer_Channel*
class Asura_Sound_Event : public Asura_Sound_Playable
{
	friend class Asura_Chunk_SoundEvent_Event;
	friend class Asura_Chunk_SoundEvent_Mapping;
	friend class Asura_Sound_Event_Mapping;
	friend class Asura_Sound_Event_Storage;//for renaming events and for checking for permanent sounds
private :

	//needs to be allocated by Asura_Sound_Event_Storage::CreateEvent()
	Asura_Sound_Event(const char* szName);	// name must be unique
	//Event deletion made private since the storage system needs to handle it. RAR 30/08/2006
	~Asura_Sound_Event();

public :

	// virtual function overrides
	virtual void FillInPlayParams( Asura_Sound_PlayParams& xParams ) const;
	virtual bool IsMetaEvent() const { return m_bIsMetaEvent; }
	virtual bool IsSample() const { return false; }

	// not valid for ASURA_PT_SIMULTANEOUS, GetPlayableSound() returns an entry
	// from the list as appropriate and sets m_uLastPlayedItem.
	const Asura_Sound_Playable* GetPlayableSound() const;

	// this overloaded version of GetPlayableSound() is simply an accessor and 
	// allows sequential playback.
	const Asura_Sound_Playable* GetPlayableSound(u_int uIndex) const;
	u_int GetNumPlayables() const { return m_xPlayableList.Size(); }

	// PlayEntry() is for "all sequential" playback. It assumes the params
	// are set correctly already for this event.
	void PlayEntry( Asura_Sound_PlayParams& xParams, u_int uEntry ) const;

	//Returns false if the event fails its die roll for determining whether to play.
	bool CheckPlayProbability() const {return (m_ucPlayProbability==100) ? true : (Asura_Random::GetInt(1,100) <= m_ucPlayProbability);}


	// Get() accessors
#ifdef ASURA_TOOLS
	bool IsEndlessPlayback() const;	// in tools this updates the value first.
	float GetMaxOuterRadius() const;
#else
	bool IsEndlessPlayback() const { return m_bEndlessPlayback; }
	float GetMaxOuterRadius() const {return m_fMaxOuterRadius;}
#endif
	ASURA_SOUND_PLAYBACK_TYPE GetPlaybackType() const { return m_ePlaybackType; }
	bool UsesOcclusion() const { return m_bUseOcclusion; }
	bool UsesDistanceFilter() const { return m_bUseDistanceFilter; }
	bool UsesDoppler() const { return m_bUseDoppler; }
	bool UsesEnvironmentalReverb() const { return m_bUseEnvironmentalReverb; }
	bool UsesMasterDSP() const { return m_bUseMasterDSP; }
	float GetMinVolume() const { return m_fMinVol; }
	float GetMaxVolume() const { return m_fMaxVol; }
	float GetMinPitch() const { return m_fMinPitch; }
	float GetMaxPitch() const { return m_fMaxPitch;}
	float GetMinDelay() const { return m_fMinDelay; }
	float GetMaxDelay() const { return m_fMaxDelay;}
	float GetMinDuration() const { return m_fMinDuration;}
	float GetMaxDuration() const { return m_fMaxDuration;}
	float GetRandomDuration() const { return Asura_Random::GetFloat(m_fMinDuration, m_fMaxDuration);}
	bool HasDuration() const {return m_fMaxDuration != 0.0f;}
	const Asura_SurroundPan& GetSurroundPan() const { return m_xSurroundPan; }
	float GetPriority() const { return m_fPriority; }
	ASURA_3DSOUND_TYPE Get3dSoundType() const { return m_e3dSoundType; }
	float GetInnerRadius() const { return m_fInnerRadius; }
	float GetOuterRadius() const { return m_fOuterRadius; }
	bool IsLooping() const { return m_bEventLooping; }
	bool IsPanOverride() const { return m_bPanOverride; }
	const Asura_Sound_VariableParams* GetVariableParams() const;
	u_int GetMaxInstances() const {return m_uMaxInstances;}
	u_int GetMinFreeVoices() const {return m_uMinFreeVoices;}
	Asura_Hash_ID GetInstanceGroupID() const {return m_uInstanceGroupID ? m_uInstanceGroupID : m_uHashID;}
	u_char GetPlayProbability() const { return m_ucPlayProbability; }
	u_int GetSoundType() const { return m_uSoundType; }
	Asura_Hash_ID GetDSPChainHash() const { return m_uDSPChainHash; }

#ifdef ASURA_TOOLS
	// Set() accessors
	void SetVolume( float fMin, float fMax );
	void SetPitch( float fMin, float fMax );
	void SetDelay( float fMin, float fMax );
	void SetDuration(float fMin, float fMax);
	void SetSurroundPan( const Asura_SurroundPan& xPan );
	void SetPriority( float fPriority );
	void SetLooping( bool bLooping ) { m_bEventLooping = bLooping; }
	void SetOcclusion( bool bOcclusion ) { m_bUseOcclusion = bOcclusion; }
	void SetDistanceFilter( bool bDistanceFilter ) { m_bUseDistanceFilter = bDistanceFilter; }
	void SetDoppler( bool bDoppler ) { m_bUseDoppler = bDoppler; }
	void SetEnvironmentalReverb( bool bEnvReverb ) { m_bUseEnvironmentalReverb = bEnvReverb; }
	void SetMasterDSP( bool bMasterDSP ) { m_bUseMasterDSP = bMasterDSP; }
	void SetPanOverride( bool bPanOverride ) { m_bPanOverride = bPanOverride; }
	void Set3dSoundType( ASURA_3DSOUND_TYPE eType ) { m_e3dSoundType = eType; }
	void SetRadii( float fInner, float fOuter );
	void SetPlayableList(  const Asura_Sound_Playable* const* apxPlayables, u_int uNumPlayables );
	void SetVariableParams(const Asura_Sound_VariableParams& xParams);
	void SetPlayProbability( u_char ucProbability ) { m_ucPlayProbability = ucProbability; }

	void SetMaxInstances(u_int uMaxInstances) {m_uMaxInstances = uMaxInstances;}
	void SetMinFreeVoices(u_int uMinFreeVoices) {m_uMinFreeVoices = uMinFreeVoices;}
#endif
	void SetPlaybackType( ASURA_SOUND_PLAYBACK_TYPE eType );

#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA	// tools-only data (also in debug)
	const char* GetName() const { return m_xEventName.GetString(); }
	const char* GetEventType() const { return m_xEventType.GetString(); }
	const char* GetEventSubType() const { return m_xEventSubType.GetString(); }
	const char* GetMixerChannelName() const { return m_xMixerChannelName.GetString(); }
	const char* GetFilename() const { return m_xFilename.GetString(); }
	const char* GetInstanceGroupName() const {return m_xInstanceGroupName.GetString();}
	bool IsGlobalSound() const {return m_bGlobalSound; }
#endif

#ifdef ASURA_TOOLS	// tools-only functions
	void SetEventType( const char* szEventType ) { m_xEventType = Asura_String(szEventType); }
	void SetEventSubType( const char* szEventSubType ) { m_xEventSubType = Asura_String(szEventSubType); }
	void SetMixerChannelName(const char* szMixerChannelName);
	void SetFilename(const char* szFilename) { m_xFilename = szFilename; }
	void SetGlobalSound(bool bSet) {m_bGlobalSound = bSet;}
	void SetInstanceGroupName(const char* szInstanceGroup);
	void SetSoundType(const u_int uSoundType) { m_uSoundType = uSoundType; }
	void SetDSPChainHash(const Asura_Hash_ID uHash) { m_uDSPChainHash = uHash; }

	u_int GetNumPlayableID() const { return m_xPlayableIDList.Size(); }
	const Asura_Sound_PlayableID&  GetPlayableID(u_int uIndex) const;
	//Updates list of playables , returns true if there was any change
	bool SetPlayableIDList( const Asura_Sound_PlayableID* pxPlayables, u_int uNumPlayables);
	//Removes playable if the event uses it. Returns true if anything was removed.
	bool RemovePlayableID(const Asura_Sound_PlayableID& xID);
	//Replaces/renames playable if the event uses it. Returns true if any were replaced
	bool ReplacePlayableID(const Asura_Sound_PlayableID& xOldID,const Asura_Sound_PlayableID& xNewID);

	//updates the list of playable pointers from the tools list of playable IDs
	void UpdatePlayableListFromIDList();

	bool IsEventMapping() const {return m_bEventMapping;}
	void SetEventMapping(bool bSet) {m_bEventMapping = bSet;}

	//Finds the event at the end of the chain of event mappings
	const Asura_Sound_Event* GetRemappedEvent() const
	{
		if(!m_bEventMapping) return this;
		if(m_xPlayableList.Size() && m_xPlayableList[0]->IsEvent())
		{
			return ((const Asura_Sound_Event*)(m_xPlayableList[0]))->GetRemappedEvent();
		}
		else
		{
			Asura_Assertf( 0, ( "Sound event '%s' has the event mapping flag but doesn't have an event in its playable list.", GetName() ) );
			return this;
		}
	}
	 
#endif
	Asura_Hash_ID GetHashID() const { return m_uHashID; }

private:

	// private methods

#ifdef ASURA_TOOLS
	// UpdateEndlessPlayback() goes through every possibility of event playback to establish
	// whether this event plays forever. If some paths do and some don't the function returns
	// false. It should be called whenever the playable list changes.
	bool UpdateEndlessPlayback() const;

	//Checks for the maximum outer radius of all child events. This is set to Asura_Maths::LargeFloat if
	//there are any events that ignore the outer radius
	void UpdateMaxOuterRadius() const;
#endif

	// private data
	
	// list of samples/other events
	Asura_HeapArray<const Asura_Sound_Playable*>	m_xPlayableList;
	bool											m_bIsMetaEvent;		// whether the list is events or samples
	mutable bool									m_bEndlessPlayback;	// does playing this event ever stop? i.e. does this event, or its playables or their playables etc. loop

	// runtime playback data shared between all event references
	mutable u_int						m_uLastPlayedItem;	// initialised to m_uNumPlayables. Used for sequential and one-by-one playback
	mutable class Asura_PrimeSearch*	m_pxPrimeSearch;	// for ASURA_PT_CDPLAYERRANDOM

	// properties
	float 								m_fMinVol;		// pick random vol in range [min,max]
	float 								m_fMaxVol;
	float 								m_fMinPitch;	// pick random pitch in range [min,max]
	float 								m_fMaxPitch;
	float 								m_fMinDelay;	// pick random start delay in range [min,max]
	float 								m_fMaxDelay;
	float								m_fMinDuration;	// If non-zero then these are used to generate a random duration for the event
	float								m_fMaxDuration;	// (the event may still stop before that duration is up for other reasons though)
	Asura_SurroundPan					m_xSurroundPan;
	float								m_fPriority;	// 0.0f to 2.0f, 1.0f is default
	ASURA_SOUND_PLAYBACK_TYPE			m_ePlaybackType;
	const Asura_Sound_Mixer_Channel*	m_pxMixingChannel;	// if NULL, use master
	Asura_Sound_VariableParams*			m_pxVariableParams; //paramaters for variable volume/pitch if using ASURA_PT_VARIABLE
	
	union
	{
		u_int							m_uFlags;
		struct
		{
			u_int						m_bEventLooping:1;	// event restarts after playback finished (possibly not seamlessly)
			u_int						m_bUseOcclusion:1;
			u_int						m_bUseDistanceFilter:1;
			u_int						m_bUseDoppler:1;
			u_int						m_bUseEnvironmentalReverb:1;
			u_int						m_bPanOverride:1;
			u_int						m_bUseMasterDSP:1;
		};
	};
	u_char								m_ucPlayProbability;	// the probability that this event will play at all when triggered. (0-100, defaults to 100).

	
	// 3d data
	ASURA_3DSOUND_TYPE					m_e3dSoundType;
	float 								m_fInnerRadius;
	float 								m_fOuterRadius;

	u_int								m_uMaxInstances;//max instances of this event or group of events to play. If zero , then the global value will be used.
	u_int								m_uMinFreeVoices;//minimum free voices that should be available before attempting to play this sound
	Asura_Hash_ID						m_uInstanceGroupID;//all sounds will the same group ID are bundled together for determining the number of playing instances. If 0 then the event hash ID is used instead

	mutable float						m_fMaxOuterRadius; //Max outer radius out of this and all child events. This will be Asura_Maths::LargeFloat if there are any events that it doesn't apply to

#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA
	Asura_String						m_xEventName;	// must be unique, converted to hashID
	Asura_String						m_xEventType;	// e.g. weapons, footsteps, environmental
	Asura_String						m_xEventSubType;
	Asura_String						m_xMixerChannelName;// defaults to "master"
	Asura_String						m_xFilename; //File that this event gets saved to.
	Asura_String						m_xInstanceGroupName;// Used for generating m_uInstanceGroupID

	Asura_HeapArray<Asura_Sound_PlayableID> m_xPlayableIDList;//list of the hash IDs and types of all the samples/events. May be larger than m_xPlayableList if some are unavailable

	bool								m_bGlobalSound;//true if sound event should be included in the global sound asr file.
#endif

#ifdef ASURA_TOOLS
	bool								m_bEventMapping;//This event is just a mapping to another event , and so the majority of the settings are ignored
#endif

	Asura_Hash_ID						m_uHashID;

	u_int								m_uMappingCount;//The number of entries this sound event has in the storage system tree

	u_int								m_uSoundType; // The sound type for sound registration;

	Asura_Hash_ID						m_uDSPChainHash; // The DSP effect chain to apply to this event
};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_SOUND_EVENT_H___
