//////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_Event_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     game api for sound events system
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Sound/Asura_Sound_Event_System.h"

// Asura includes
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Mempool.h"
#include "Core/Asura_SortAndSearch.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_DebugOverlay.h"
#include "Debugging/Asura_Debug_Timers.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Viewport.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Phonon.h"
#include "Sound/Asura_Sound.h"
#include "Sound/Asura_Sound_Event.h"
#include "Sound/Asura_Sound_Mixer.h"
#include "Sound/Asura_Sound_Event_Playback.h"
#include "Sound/Asura_Sound_Event_Storage.h"

// Tools includes
#ifdef ASURA_TOOLS
#include "Sound/Tools_SoundEvent_System.h"
#endif

// D E F I N E S /////////////////////////////////////////////////////////////
#ifndef ASURA_SOUND_MAX_ACTIVE_EVENTS
#define ASURA_SOUND_MAX_ACTIVE_EVENTS 128
#endif

#define ASURA_SOUND_INSTANCE_COUNT_TABLE_SIZE 32 //should be power of 2

// G L O B A L S /////////////////////////////////////////////////////////////
static bool g_bMixerChanged;

void ( *Asura_Sound_Event_System::s_pfnSoundRegistryFn )(const Asura_Guid uSourceGuid, const Asura_Vector_3* pxPosition, const u_int uSoundType, const float fSoundRange ) = NULL;

#if defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)
static const Asura_Sound_Event* g_pxEventDeleted;
#endif	//defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)

#if defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )
bool Asura_Sound_Event_System::s_bPrintSoundEvents = false;
#define CONSOLE_VARNAME_PRINT_SOUND_EVENTS	L"Audio.Debug.PrintSoundEvents"
#endif	//defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Sound_Event_Instance* Asura_Sound_Event_System::s_pxActiveEventList = NULL;
Asura_Tracking_Handle_List* Asura_Sound_Event_System::s_pxActiveEventHandleList = NULL;
u_int Asura_Sound_Event_System::s_uDefaultMaxInstances = 0;

Asura_Mempool* Asura_Sound_Event_System::s_pxInstanceCountMempool;
Asura_SoundEvent_InstanceCount** Asura_Sound_Event_System::s_ppxInstanceCountTable;

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

float Asura_Sound_Event_System::AmplitudeToDecibels( float fAmplitude )
{
	if ( fAmplitude <= 0.0f )
	{
		return -96.0f;	// Effectively silence
	}
	else
	{
		return 20.0f * log10f( fAmplitude );
	}
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

#ifdef ASURA_TOOLS

Asura_Sound_Event_Lookup::Asura_Sound_Event_Lookup( Asura_Hash_ID uNameHash, const char* szName )
{
	if( szName )
	{
		Asura_Assert( Asura_GetHashID(szName) == uNameHash, "sound event hash and name do not match" );

		Tools_SoundEvent_System::RegisterCodeEvent(szName);
	}
}


const Asura_Sound_Event* Asura_Sound_Event_Lookup::FindEvent_CalledByMacro( Asura_Hash_ID uNameHash, const char* szName, bool bMissingEventHandledExternally )
{
	if( szName ) 
	{
		Asura_Assertf( Asura_GetHashID(szName) == uNameHash, ("Sound event hash and name '%s' do not match", szName) );
	}

	#ifdef ASURA_TOOLS
	Asura_Assertf(Tools_SoundEvent_System::IsCodeEvent(uNameHash),("Sound event '%s' with hash 0x%x hasn't been registered as a code event using the GLOBALINIT macro", (szName ? szName : "(Name not supplied)"), uNameHash));
	#endif
	
    const Asura_Sound_Event* const pxEvent = Asura_Sound_Event_Storage::FindEvent( uNameHash );

    if( !bMissingEventHandledExternally )
    {
		Asura_Assertf( pxEvent, ("Failed to find sound event '%s'", (szName ? szName : "<UNKNOWN>") ) );
    }

	return pxEvent;
}

#else

const Asura_Sound_Event* Asura_Sound_Event_Lookup::FindEvent_CalledByMacro( Asura_Hash_ID uNameHash, bool bMissingEventHandledExternally )
{
    const Asura_Sound_Event* const pxEvent = Asura_Sound_Event_Storage::FindEvent( uNameHash );

    if( !bMissingEventHandledExternally )
    {
	    Asura_Assertf( pxEvent, ("Failed to find sound event with hash ID 0x%08X", uNameHash) );
    }

	return pxEvent;
}

#endif // ASURA_TOOLS


// Asura_Sound_Event_System
void Asura_Sound_Event_System::Initialise()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Sound_Event_System::Initialise()");

	Asura_Assert(!s_pxActiveEventList, "already initialised");
	Asura_Assert(!s_pxActiveEventHandleList, "already initialised");
	s_pxActiveEventList = new Asura_Sound_Event_Instance[ASURA_SOUND_MAX_ACTIVE_EVENTS];
	s_pxActiveEventHandleList = new Asura_Tracking_Handle_List(ASURA_SOUND_MAX_ACTIVE_EVENTS);

	s_pxInstanceCountMempool = new Asura_Mempool(sizeof(Asura_SoundEvent_InstanceCount), 64);
	s_ppxInstanceCountTable = new Asura_SoundEvent_InstanceCount*[ASURA_SOUND_INSTANCE_COUNT_TABLE_SIZE];
	Asura_Mem::Set(s_ppxInstanceCountTable, 0, sizeof(Asura_SoundEvent_InstanceCount*) * ASURA_SOUND_INSTANCE_COUNT_TABLE_SIZE);

	#if defined(ASURA_TOOLS) && !defined(ASURA_USE_MULTIPLE_HANDLERS)
	//(In streaming builds the intialisation needs to wait until the handlers are set up)
	//RAR 29/09/2006
	Tools_SoundEvent_System::Initialise();
	#endif

#if defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )
		Asura_CommandConsole::AddVar( CONSOLE_VARNAME_PRINT_SOUND_EVENTS, &s_bPrintSoundEvents, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif	//defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )

	#if defined( ASURA_USE_DEBUGOVERLAY )
		Asura_DebugOverlay::RegisterMode( "PlayingSoundsList", ASURA_DO_SOUND_LIST, DoDebugOverlayListSounds, NULL );
		Asura_DebugOverlay::RegisterMode( "PlayingSoundsDraw", ASURA_DO_SOUND_DRAW, DoDebugOverlayDrawSounds, NULL );
	#endif // #if defined( ASURA_USE_DEBUGOVERLAY )
}

void Asura_Sound_Event_System::Reset()
{
	// stop all active instances with no external handle.
	if( s_pxActiveEventHandleList )
	{
		s_pxActiveEventHandleList->Traverse(StopAllWithNoExternalHandleCallback);
		Asura_Assert(!s_pxActiveEventHandleList->GetNumberOfActiveHandles(), "external sound event handles remain at Reset()");
	}

	Asura_Sound_Event_Storage::Reset();
}

void Asura_Sound_Event_System::Shutdown()
{	
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_Sound_Event_System::Shutdown()" );

	// stop all active instances with no external handle.
	if( s_pxActiveEventHandleList )
	{
		s_pxActiveEventHandleList->Traverse(StopAllWithNoExternalHandleCallback);
		Asura_Assert(!s_pxActiveEventHandleList->GetNumberOfActiveHandles(), "external sound event handles remain at Shutdown()");
	}

	delete[] s_pxActiveEventList;
	s_pxActiveEventList = NULL;
	delete s_pxActiveEventHandleList;
	s_pxActiveEventHandleList = NULL;

	Asura_Assert(!s_pxInstanceCountMempool->GetNumAllocated(), "Sound event instance counts should have all been zeroed by now.");

	delete s_pxInstanceCountMempool;
	s_pxInstanceCountMempool = NULL;
	delete [] s_ppxInstanceCountTable;
	s_ppxInstanceCountTable = NULL;

	#ifdef ASURA_TOOLS
	Tools_SoundEvent_System::Shutdown();
	#endif

	Asura_Sound_Event_Storage::Shutdown();
}

// Update() will traverse the active event instance list checking sound status, updating sound parameters etc.
void Asura_Sound_Event_System::Update()
{
	g_bMixerChanged = Asura_Sound_Mixer_System::MixerChangedThisFrame();

	if( s_pxActiveEventHandleList )
	{
		s_pxActiveEventHandleList->Traverse( UpdateEventInstance );
	}
		
	g_bMixerChanged = false;
}

#if defined( ASURA_USE_DEBUGOVERLAY )
// helper for sorting events
static int CompareEventInstances( const void* pEventInstance1, const void* pEventInstance2 )
{
	const Asura_Sound_Event_Instance* const * ppxInstance1 = reinterpret_cast<const Asura_Sound_Event_Instance* const *>(pEventInstance1);
	const Asura_Sound_Event_Instance* const * ppxInstance2 = reinterpret_cast<const Asura_Sound_Event_Instance* const *>(pEventInstance2);
	const Asura_Sound_Event* pxEvent1 = (*ppxInstance1)->GetEvent();
	const Asura_Sound_Event* pxEvent2 = (*ppxInstance2)->GetEvent();
	const Asura_Hash_ID uHash1 = pxEvent1 ? pxEvent1->GetHashID() : ASURA_HASH_ID_UNSET;
	const Asura_Hash_ID uHash2 = pxEvent2 ? pxEvent2->GetHashID() : ASURA_HASH_ID_UNSET;
	if (uHash1 > uHash2)
	{
		return 1;
	}
	else if (uHash1 < uHash2)
	{
		return -1;
	}
	return 0;
}

void Asura_Sound_Event_System::DoDebugOverlayListSounds( void* const pxIgnore )
{
	(void)pxIgnore;

	// print num active
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_SOUND_LIST, "%u/%u active sound events", 
		s_pxActiveEventHandleList->GetNumberOfActiveHandles(), ASURA_SOUND_MAX_ACTIVE_EVENTS );

	// build sorted array of active event instance pointers
	const Asura_Sound_Event_Instance* apxSortedEvents[ASURA_SOUND_MAX_ACTIVE_EVENTS];
	u_int uNumEvents = 0;

	for (u_int uCurEvent = 0; uCurEvent < ASURA_SOUND_MAX_ACTIVE_EVENTS; ++uCurEvent)
	{
		const Asura_Sound_Event_Instance& xEventInstance = s_pxActiveEventList[uCurEvent];
		if (xEventInstance.IsPlaying() || !xEventInstance.NoExternalHandle())
		{
			apxSortedEvents[uNumEvents++] = &xEventInstance;
		}
	}
	Asura_SortAndSearch::Quicksort(apxSortedEvents, uNumEvents, CompareEventInstances);

	u_int uTotalControllers = 0;
	Asura_Hash_ID uPrevEventHash = ASURA_HASH_ID_UNSET;
	for (u_int uCurEvent = 0; uCurEvent < uNumEvents; ++uCurEvent)
	{
		const Asura_Sound_Event_Instance& xEventInstance = *apxSortedEvents[uCurEvent];
		const Asura_Sound_Event* pxEvent = xEventInstance.GetEvent();

		const float fVolume = xEventInstance.GetVolumeScale();

		Asura_Assert(pxEvent, "no event for active event instance");
		if ( pxEvent )
		{
			u_int uNumControllers = xEventInstance.GetNumControllers();
			uTotalControllers += uNumControllers;

			// check if we've already printed info about this event.
			Asura_Hash_ID uHashID = pxEvent->GetHashID();
			if (uPrevEventHash == uHashID)
			{
				continue;
			}
			uPrevEventHash = uHashID;

			// peek at other instances of this event ahead of us.
			u_int uNumVoices = xEventInstance.GetNumVoices();
			u_int uNumInstances = 1;
			for (u_int uTemp = uCurEvent + 1; uTemp < uNumEvents; ++uTemp)
			{
				const Asura_Sound_Event* pxNextEvent = (apxSortedEvents[uTemp]->GetEvent());
				if (!pxNextEvent || (pxNextEvent->GetHashID() != uHashID))
				{
					break;
				}
				++uNumInstances;
				uNumVoices += apxSortedEvents[uTemp]->GetNumVoices();
				uNumControllers += apxSortedEvents[uTemp]->GetNumControllers();
			}

			const Asura_Char* szName = "";
			const Asura_Char* szMixerName = "";
#if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )
			szName = pxEvent->GetName();
			szMixerName = pxEvent->GetMixerChannelName();
#else
			char acNameBuffer[11];
			Asura_PrintfToBfr(acNameBuffer, "0x%x", 11, true, pxEvent->GetHashID());
			szName = acNameBuffer;
#endif
			const Asura_Hash_ID uGroupID = pxEvent->GetInstanceGroupID();
			u_int uNumGroupInstances = GetNumActiveInstances(uGroupID);
			u_int uMaxGroupInstances = pxEvent->GetMaxInstances() ? pxEvent->GetMaxInstances() : GetDefaultMaxInstances();
			if (uMaxGroupInstances)
			{
				Asura_DebugOverlay::DrawConsoleText( ASURA_DO_SOUND_LIST, 
					"%u of %s (group 0x%x %u/%u), (%u ctrl %u voices) vol %.2f mixer: %s", 
					uNumInstances, szName,  
					pxEvent->GetInstanceGroupID(), uNumGroupInstances, uMaxGroupInstances, 
					uNumControllers, uNumVoices, Asura_Maths::FtoD(fVolume), szMixerName );
			}
			else
			{
				Asura_DebugOverlay::DrawConsoleText( ASURA_DO_SOUND_LIST, 
					"%u of %s (%u ctrl %u voices) vol %.2f mixer: %s", 
					uNumInstances, szName,  
					uNumControllers, uNumVoices, Asura_Maths::FtoD(fVolume), szMixerName );
			}

			// Display any child event information

			u_int uNumPlayables = pxEvent->GetNumPlayables();

			for(u_int n = 0; n < uNumPlayables; n++)
			{
				const Asura_Sound_Playable* pxPlayable = pxEvent->GetPlayableSound(n);

				if(pxPlayable->IsEvent())
				{
					const Asura_Sound_Event* pxChildEvent = static_cast<const Asura_Sound_Event*>(pxPlayable);

					Asura_Assert(pxChildEvent, "Could not find playable event");

					if(pxChildEvent)
					{
						szMixerName = "";

						#if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )
						szName = pxChildEvent->GetName();
						szMixerName = pxChildEvent->GetMixerChannelName();
						#else
						Asura_PrintfToBfr(acNameBuffer, "0x%x", 11, true, pxEvent->GetHashID());
						szName = acNameBuffer;
						#endif

						Asura_DebugOverlay::DrawConsoleText( ASURA_DO_SOUND_LIST, ">> %s - mixer: %s", szName, szMixerName);
					}
				}
			}
		}
	}

	// print num play controllers used, and number of sound system voices

#if defined( ASURA_USE_AUDIO_SYSTEM )
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_SOUND_LIST, "%u/%u event controllers used", 
		uTotalControllers, uASURA_SOUND_EVENT_MAX_CONTROLLERS);
#else
	Asura_DebugOverlay::DrawConsoleText( ASURA_DO_SOUND_LIST, "%u/%u event controllers used. %u/%u voices used", 
		uTotalControllers, uASURA_SOUND_EVENT_MAX_CONTROLLERS,
		Asura_Sound_System::GetNumberOfPlayingSounds(), Asura_Sound_System::GetMaxNumberOfPlayingSounds());
#endif
}

#endif // #if defined( ASURA_USE_DEBUGOVERLAY )


#if defined( ASURA_USE_DEBUGOVERLAY )
void Asura_Sound_Event_System::DoDebugOverlayDrawSounds( void* const pxIgnore )
{
	(void)pxIgnore;

	if( s_pxActiveEventHandleList )
	{
		s_pxActiveEventHandleList->Traverse( DoDebugOverlayDrawSound );
	}
}

void Asura_Sound_Event_System::DoDebugOverlayDrawSound( Asura_Handle iHandle )
{
	Asura_Assert( s_pxActiveEventHandleList->IsActiveHandle( iHandle ), "unexpected handle" );	
	const Asura_Sound_Event_Instance& xEventInstance = s_pxActiveEventList[iHandle];

	if ( xEventInstance.GetUsePosition() )
	{
		// DrawTextWorld makes the text unreadably big if you're close to it, so we'll do
		// the world to screen projection ourselves
		Asura_Vector_3 xPositionWorld = xEventInstance.GetPosition();

		if ( Asura_Camera::ProjectWorldCoordsToScreenSpace( xPositionWorld, true ) && xPositionWorld.z > 0.0f )
		{
			Asura_Vector_2 xPositionScreen( xPositionWorld.x / Asura_Viewport::GetWidth(), xPositionWorld.y / Asura_Viewport::GetHeight() );

			const Asura_Sound_Event* pxEvent = xEventInstance.GetEvent();
			const float fVolume = xEventInstance.GetVolumeScale();

			if ( pxEvent )
			{
				#if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )
					const Asura_Char* name = pxEvent->GetName();
					Asura_DebugOverlay::DrawTextScreen( ASURA_DO_SOUND_DRAW, xPositionScreen, 0xFF00FF00,
						"%s (%.2f / %.2fdB)", name ? name : "???", Asura_Maths::FtoD(fVolume), Asura_Maths::FtoD(AmplitudeToDecibels( fVolume )) );
				#else
					Asura_DebugOverlay::DrawTextScreen( ASURA_DO_SOUND_DRAW, xPositionScreen, 0xFF00FF00,
						"0x%08x (%.2f / %.2fdB)", pxEvent->GetHashID(), Asura_Maths::FtoD(fVolume), Asura_Maths::FtoD(AmplitudeToDecibels( fVolume )) );
				#endif
			}
			else
			{
				Asura_DebugOverlay::DrawTextScreen( ASURA_DO_SOUND_DRAW, xPositionScreen, 0xFFFF0000,
					"??? (%.2f / %.2fdB)", Asura_Maths::FtoD(fVolume), Asura_Maths::FtoD(AmplitudeToDecibels( fVolume )) );
			}

			// Draw the sound range
			Asura_DebugOverlay::DrawSphereWorld( ASURA_DO_SOUND_DRAW, xEventInstance.GetPosition(),
				xEventInstance.GetOuterRadius(), 0xFF00FF00 );
			Asura_DebugOverlay::DrawSphereWorld( ASURA_DO_SOUND_DRAW, xEventInstance.GetPosition(),
				xEventInstance.GetInnerRadius(), 0xFFFF0000 );
		}
	}
}
#endif

// PlayEvent()
// A position may be passed in and is ignored for "direct" sounds, but must be passed in for other 3d
// sound types. If not, then assert and play as 2d.
void Asura_Sound_Event_System::PlayEvent( const Asura_Sound_Event* pxEvent, 
										Asura_Handle* piHandleOut, 
										const Asura_Vector_3* pxPos /* = NULL */,
										float fVolumeScale /* = 1.0f */,
										Asura_Guid uSourceGuid /*= ASURA_GUID_UNREGISTERED*/)
{
	
	// parameter checking
	if (piHandleOut)
	{
		Asura_Assert(*piHandleOut == ASURA_INVALID_HANDLE, "handle is already set, possible leak");
	}

	Asura_Assert(pxEvent, "no event specified");
	if (!pxEvent)
	{
		if (piHandleOut)
		{
			*piHandleOut = ASURA_INVALID_HANDLE;
		}
		return;
	}
	#ifdef ASURA_TOOLS
	//resolve event mapping. RAR 08/08/2007
	pxEvent = pxEvent->GetRemappedEvent();
	#endif

	const bool bEndlessPlayback = pxEvent->IsEndlessPlayback();
	
	if (bEndlessPlayback && !piHandleOut)
	{
		Asura_Assert(false, "can't play looping events without keeping the handle");
		return;
	}

	bool bCull = false;
	if(!pxEvent->CheckPlayProbability())
	{
		//Failed roll to see if the event should play
		bCull = true;
	}

	// call our Sound Registry Callback if it has been set.
	if ( s_pfnSoundRegistryFn != NULL && uSourceGuid != ASURA_GUID_UNREGISTERED )
	{
		s_pfnSoundRegistryFn( uSourceGuid, pxPos, pxEvent->GetSoundType(), pxEvent->GetOuterRadius() );
	}

	// AUDIOTODO: The audio system probably needs the concept of a maximum number of sounds
	#ifndef ASURA_USE_AUDIO_SYSTEM
	if(!bCull && ((int)pxEvent->GetMinFreeVoices() > Asura_Sound_System::GetNumberOfFreeSounds()))
	{
		//Too many sounds playing for this event to be allowed.
		bCull = true;
	}
	#endif

	//For no looping sound events , check to see if they are going to be in range
	if(!bCull && !bEndlessPlayback && pxPos)
	{
		const float fOuterRadius = pxEvent->GetMaxOuterRadius();
#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA
		Asura_Assertf(fOuterRadius > 0.f, ("Sound event %s has 0 or negative outer radius", pxEvent->GetName()) );
#else
		Asura_Assertf(fOuterRadius > 0.f, ("Sound event 0x%x  has 0 or negative outer radius", pxEvent->GetHashID()) );
#endif
		
		if(fOuterRadius < Asura_Maths::LargeFloat)
		{
			if( !IsSphereInRange(*pxPos,fOuterRadius) )
			{
				bCull = true;
			}
		}
	}

	if(!bCull)
	{
		//Check to make sure that too many instances of this sound event aren't already playing
		u_int uMaxAllowedInstances = pxEvent->GetMaxInstances();
		if(!uMaxAllowedInstances) uMaxAllowedInstances = GetDefaultMaxInstances();
		if(uMaxAllowedInstances)
		{
			const Asura_Hash_ID uGroupID = pxEvent->GetInstanceGroupID();
			if(GetNumActiveInstances(uGroupID) >= uMaxAllowedInstances)
			{
				bCull = true;
			}
		}
	}

	if(bCull)
	{
		if (piHandleOut)
		{
			*piHandleOut = ASURA_INVALID_HANDLE;
		}
		return;
	}

	// allocate an event instance
	Asura_Handle iHandle = ASURA_INVALID_HANDLE;
	bool bOK = s_pxActiveEventHandleList ? s_pxActiveEventHandleList->GetNewHandle(iHandle) : false;

	if (piHandleOut)
	{
		*piHandleOut = iHandle;
	}
	if (!bOK)
	{
		Asura_Assert(false, "ran out of event instances to allocate");
		return;
	}

	// deal with playback
	// initialise the active event and play the sound
	if( s_pxActiveEventList )
	{
		s_pxActiveEventList[iHandle].Play(pxEvent, pxPos, fVolumeScale);
		s_pxActiveEventList[iHandle].SetHasExternalHandle(piHandleOut != NULL);
	}
}

Asura_Handle Asura_Sound_Event_System::CreatePhononInstance( const Asura_Phonon* pxPhonon )
{
	// parameter checking
	Asura_Assert(pxPhonon, "no phonon" );
	if ( !pxPhonon)
	{
		return ASURA_INVALID_HANDLE;
	}
	
	Asura_Assertf( pxPhonon->GetEvent(), ("no event specified for phonon %d", pxPhonon->GetGuid() ));
	if (!pxPhonon->GetEvent())
	{
		return ASURA_INVALID_HANDLE;
	}

	// allocate an event instance
	Asura_Handle iHandle = ASURA_INVALID_HANDLE;
	if( s_pxActiveEventHandleList && s_pxActiveEventList )
	{
		const bool bOK = s_pxActiveEventHandleList->GetNewHandle(iHandle);
		if( bOK )
		{
			s_pxActiveEventList[iHandle].SetHasExternalHandle(true);
		}
		else
		{
			Asura_Assert(false, "ran out of event instances to allocate");
		}
	}

	return iHandle;
}

void Asura_Sound_Event_System::UpdatePhononInstanceParameters( Asura_Handle iHandle, const Asura_Phonon* pxPhonon )
{
	if( !s_pxActiveEventHandleList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return; }
	Asura_Assert(false, "UpdatePhononInstanceParameters() not implemented yet");
}

void Asura_Sound_Event_System::TriggerPhononInstanceEventPlay( Asura_Handle iHandle, const Asura_Phonon* pxPhonon )
{
	if( !s_pxActiveEventHandleList ) return;
	
	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return; }
    Asura_Assert(!s_pxActiveEventList[iHandle].IsPlaying(),"Phonon already playing?");

	const Asura_Sound_Event* pxEvent = pxPhonon->GetEvent();
	Asura_Assert(pxEvent,"");

	#ifdef ASURA_TOOLS
	//resolve event mapping. RAR 08/08/2007
	pxEvent = pxEvent->GetRemappedEvent();
	#endif

	if(!pxEvent->CheckPlayProbability())
	{
		//Failed roll to see if the event should play
		return;
	}

	// AUDIOTODO: Audio system needs the idea of a max number of sounds
	#ifndef ASURA_USE_AUDIO_SYSTEM
	if((int)pxEvent->GetMinFreeVoices() > Asura_Sound_System::GetNumberOfFreeSounds())
	{
		//Too many sounds playing for this event to be allowed
		return;
	}
	#endif

	//Check to make sure that too many instances of this sound event aren't already playing
	u_int uMaxAllowedInstances = pxEvent->GetMaxInstances();
	if(!uMaxAllowedInstances) uMaxAllowedInstances = GetDefaultMaxInstances();
	if(uMaxAllowedInstances)
	{
		const Asura_Hash_ID uGroupID = pxEvent->GetInstanceGroupID();
		if(GetNumActiveInstances(uGroupID) >= uMaxAllowedInstances)
		{
			return;
		}
	}


	//Stop the instance to prevent controller leaks if this is a phonon that has stopped playing and is being re-triggered
	//RAR 12/06/2007
	s_pxActiveEventList[iHandle].Stop();
	s_pxActiveEventList[iHandle].Play(pxPhonon);
}


// Stop() destroys the sound and releases the event instance handle, so it uses the reference
// to set the handle to be invalid automatically.
void Asura_Sound_Event_System::Stop( Asura_Handle& iHandleRef, float fFadeOut /*= 0.0f */ )
{
	// allow invalid handle with no asserts for easier cleaning up etc.
	if (iHandleRef == ASURA_INVALID_HANDLE)
    {
        return;
    }

	if( s_pxActiveEventHandleList && s_pxActiveEventList )
	{
		Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandleRef), "unexpected handle");
		s_pxActiveEventList[iHandleRef].SetHasExternalHandle(false);
		if (fFadeOut > 0.0f)
		{
			s_pxActiveEventList[iHandleRef].FadeOut(fFadeOut);
		}
		else
		{
			s_pxActiveEventList[iHandleRef].Stop();
			s_pxActiveEventHandleList->ReleaseHandle(iHandleRef);
		}
	}

	iHandleRef = ASURA_INVALID_HANDLE;
}


// FinishedWithHandle() allows the event sound system code to free up the event handle when 
// the event has finished playing. If the event is looping it should assert and stop it
// immediately. The reference is used to set the handle to be invalid automatically.
void Asura_Sound_Event_System::FinishedWithHandle( Asura_Handle& iHandleRef )
{
	// RDM - 23/11/08 - this used to just return if either list was null but that might mean the handle was not
	// reset correctly and callers relying on this could have invalidated handles.
	if( s_pxActiveEventHandleList && s_pxActiveEventList )
	{			
	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandleRef), "unexpected handle");
    if ( iHandleRef == ASURA_INVALID_HANDLE ) { return; }
	Asura_Assert(!s_pxActiveEventList[iHandleRef].IsLooping(), "shouldn't finish with handle of looping event");
	if (s_pxActiveEventList[iHandleRef].IsLooping())
	{
		s_pxActiveEventList[iHandleRef].Stop();
	}

	s_pxActiveEventList[iHandleRef].SetHasExternalHandle(false);
	}

	iHandleRef = ASURA_INVALID_HANDLE;
}

// IsPlaying() returns whether the given event instance is playing
bool Asura_Sound_Event_System::IsPlaying( Asura_Handle iHandle )
{
	if ( !s_pxActiveEventHandleList || !s_pxActiveEventList )	return false;
	if ( iHandle == ASURA_INVALID_HANDLE )						return false;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
	return s_pxActiveEventList[iHandle].IsPlaying();
}

// IsLooping() returns true if the event instance loops (either in full or in part)
bool Asura_Sound_Event_System::IsLooping( Asura_Handle iHandle )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return false;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return false; }
	return s_pxActiveEventList[iHandle].IsLooping();
}


// set event instance position
void Asura_Sound_Event_System::UpdatePosition( Asura_Handle iHandle, const Asura_Vector_3& xPos )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return; }
    return s_pxActiveEventList[iHandle].SetPosition(xPos);
}

// scale the volume (e.g. to make footsteps louder when moving quickly)
void Asura_Sound_Event_System::SetVolumeScale( Asura_Handle iHandle, float fVolumeScale )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return; }
	s_pxActiveEventList[iHandle].SetVolumeScale(fVolumeScale);
}

float Asura_Sound_Event_System::GetVolumeScale( Asura_Handle iHandle )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return 1.0f;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return 1.0f; }
	return s_pxActiveEventList[iHandle].GetVolumeScale();
}

// scale the pitch (e.g. to make an engine sound loop pitch up/down with speed)
void Asura_Sound_Event_System::SetPitchScale( Asura_Handle iHandle, float fPitchScale )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return; }
	s_pxActiveEventList[iHandle].SetPitchScale(fPitchScale);
}

float Asura_Sound_Event_System::GetPitchScale( Asura_Handle iHandle )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return 1.0f;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return 1.0f; }
	return s_pxActiveEventList[iHandle].GetPitchScale();
}

// get outer radius of sound event
float Asura_Sound_Event_System::GetOuterRadius( Asura_Handle iHandle )
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return 0.0f;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
    if ( iHandle == ASURA_INVALID_HANDLE ) { return 0.0f; }
    return s_pxActiveEventList[iHandle].GetOuterRadius();
}

void Asura_Sound_Event_System::UpdateEventInstance(Asura_Handle iHandle)
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Sound_Event_Instance& xEventInstance = s_pxActiveEventList[iHandle];
	
	xEventInstance.Update(g_bMixerChanged);

	if (!xEventInstance.IsPlaying() && xEventInstance.NoExternalHandle())
	{
		xEventInstance.Stop();
		s_pxActiveEventHandleList->ReleaseHandle(iHandle);
	}
}

void Asura_Sound_Event_System::StopAllWithNoExternalHandleCallback(Asura_Handle iHandle)
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Sound_Event_Instance& xEventInstance = s_pxActiveEventList[iHandle];
	if (xEventInstance.NoExternalHandle())
	{
		Stop(iHandle);
	}
#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA	// tools-only data (also in debug)
	else
	{
#ifdef ASURA_ASSERT
		const Asura_Sound_Event* pxEvent = xEventInstance.GetEvent();
		const char* szEvent = pxEvent ? pxEvent->GetName() : "unknown";
		Asura_Assertf(false, ("event %s still has outstanding handle\n", szEvent));
#endif
	}
#endif
}

#if defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)
void Asura_Sound_Event_System::SoundEventDeleted(const Asura_Sound_Event* pxEvent)
{
	if( !s_pxActiveEventHandleList ) return;	

	g_pxEventDeleted = pxEvent;
	s_pxActiveEventHandleList->Traverse(SoundEventDeletedCallback);
	g_pxEventDeleted = NULL;
}

void Asura_Sound_Event_System::SoundEventDeletedCallback(Asura_Handle iHandle)
{
	if( !s_pxActiveEventHandleList || !s_pxActiveEventList ) return;

	Asura_Assert(s_pxActiveEventHandleList->IsActiveHandle(iHandle), "unexpected handle");
	Asura_Sound_Event_Instance& xEventInstance = s_pxActiveEventList[iHandle];
	xEventInstance.SoundEventDeleted(g_pxEventDeleted);
}
#endif	//defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)

u_int Asura_Sound_Event_System::GetNumActiveInstances(u_int uGroupID)
{
	Asura_Assert(s_pxInstanceCountMempool && s_ppxInstanceCountTable,"");
	if( !s_pxInstanceCountMempool || !s_ppxInstanceCountTable ) return 0;
	
	u_int uIndex = uGroupID & (ASURA_SOUND_INSTANCE_COUNT_TABLE_SIZE-1);
	
	//search for an entry for this ID
	Asura_SoundEvent_InstanceCount* pxEntry = s_ppxInstanceCountTable[uIndex];

	while(pxEntry)
	{
		if(pxEntry->m_uGroupID == uGroupID)
		{
			return pxEntry->m_uCount;
		}
		else
		{
			pxEntry = pxEntry->m_pxNext;
		}
	}
	return 0;
}

void Asura_Sound_Event_System::AddToActiveInstanceCount(Asura_Hash_ID uGroupID)
{
	Asura_Assert(s_pxInstanceCountMempool && s_ppxInstanceCountTable,"");
	if( !s_pxInstanceCountMempool || !s_ppxInstanceCountTable ) return;
	
	u_int uIndex = uGroupID & (ASURA_SOUND_INSTANCE_COUNT_TABLE_SIZE-1);

	//search for an existing entry
	Asura_SoundEvent_InstanceCount* pxEntry = s_ppxInstanceCountTable[uIndex];

	while(pxEntry)
	{
		if(pxEntry->m_uGroupID == uGroupID)
		{
			pxEntry->m_uCount++;
			return;
		}
		else
		{
			pxEntry = pxEntry->m_pxNext;
		}
	}
	//add a new entry for this ID
	pxEntry = (Asura_SoundEvent_InstanceCount*) s_pxInstanceCountMempool->Allocate();
	pxEntry->m_uGroupID = uGroupID;
	pxEntry->m_uCount = 1;
	pxEntry->m_pxNext = s_ppxInstanceCountTable[uIndex];
	s_ppxInstanceCountTable[uIndex] = pxEntry;
}

void Asura_Sound_Event_System::RemoveFromActiveInstanceCount(Asura_Hash_ID uGroupID)
{
	Asura_Assert(s_pxInstanceCountMempool && s_ppxInstanceCountTable,"");
	if( !s_pxInstanceCountMempool || !s_ppxInstanceCountTable ) return;

	u_int uIndex = uGroupID & (ASURA_SOUND_INSTANCE_COUNT_TABLE_SIZE-1);

	//search for the (hopefully) existing entry
	Asura_SoundEvent_InstanceCount** ppxEntry = &s_ppxInstanceCountTable[uIndex];

	while(*ppxEntry)
	{
		Asura_SoundEvent_InstanceCount* pxEntry = *ppxEntry;
		if(pxEntry->m_uGroupID == uGroupID)
		{
			//decrease the count and remove the entry if needed
			pxEntry->m_uCount--;
			if(!pxEntry->m_uCount)
			{
				*ppxEntry = pxEntry->m_pxNext;
				s_pxInstanceCountMempool->Deallocate(pxEntry);
			}
			return;
		}
		else
		{
			ppxEntry = &((*ppxEntry)->m_pxNext);
		}
	}

	Asura_Assert(0, "Failed to find an entry for this ID");
}

bool Asura_Sound_Event_System::IsSphereInRange( const Asura_Vector_3& xPosition, float fRadius )
{
	// Check this sphere is in range of any camera
    const float fRSqd = fRadius * fRadius;
#if defined ASURA_USE_AUDIO_SYSTEM
	for (u_int uListenerIdx = 0; uListenerIdx < Asura_Audio_System::GetNumberOfListeners(); ++uListenerIdx)
	{
		if (!Asura_Audio_System::IsListenerActive( uListenerIdx ))
		{
			continue;
		}

		Asura_Vector_3 xListenerPosition;
		Asura_Audio_System::GetListenerPosition( &xListenerPosition, NULL, uListenerIdx );
	    // Is the camera inside this sphere
        if( (xPosition - xListenerPosition).MagnitudeSquared() < fRSqd )
        {
		    return true;
	    }
    }
#else
	for (int iCamID = 0; iCamID < Asura_Camera::GetNumberOfCameras(); ++iCamID)
	{
        const Asura_CameraData& xCamData = Asura_Camera::GetCameraData( iCamID );
        if( xCamData.m_bIsActive )
        {
		    // Is the camera inside this sphere
            if( (xPosition - xCamData.m_xPosition).MagnitudeSquared() < fRSqd )
            {
			    return true;
		    }
        }
	}
#endif

	return false;
}

//eof
