 //////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_Event_System.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     game api for sound events system
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_SOUND_EVENT_SYSTEM_H___
#define ___ASURA_SOUND_EVENT_SYSTEM_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Debugging/Asura_DebugOverlay.h"
#include "Sound/Asura_Sound_Event.h"
#include "Network/Asura_Guid.h"

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_Phonon;
class Asura_Sound_Event_Instance;
class Asura_Sound_Event_Lookup;
class Asura_Tracking_Handle_List;
class Asura_Vector_3;
struct Asura_SoundEvent_InstanceCount;
class Asura_Mempool;


// M A C R O S ///////////////////////////////////////////////////////////////
/*
These macros are intended for finding/registering sound events that have hard-coded names/IDs.
The idea is that any such hard-coded events will get added to a list by the global init macro , so
that they are presented to the audio people as events that need creating/filling in.

If you need to find a non hard-coded sound event (such as one set by the artists in blueprints or
animations) , then you should use Asura_Sound_Event_Storage::FindEvent() instead.
RAR 12/03/2008
*/

#ifdef ASURA_TOOLS

    // macro which in tools builds adds the lookup to a list using global ctors.
    #define ASURA_SOUND_EVENT_LOOKUP_GLOBALINIT( uHASHID, szNAME ) \
        static Asura_Sound_Event_Lookup g_xSoundEventLookup##uHASHID( uHASHID, szNAME )

    #define ASURA_SOUND_EVENT_LOOKUP( uHASHID, szNAME ) \
	    Asura_Sound_Event_Lookup::FindEvent_CalledByMacro( uHASHID, szNAME, false )

    // Version of the macro that allows calling code to handle missing events
    // rather than asserting about it.
    #define ASURA_SOUND_EVENT_LOOKUP_IF_LOADED( uHASHID, szNAME ) \
	    Asura_Sound_Event_Lookup::FindEvent_CalledByMacro( uHASHID, szNAME, true )

#else // not #ifdef ASURA_TOOLS

    // non-tools builds ignore everything except the hashID for the lookup call.

    #define ASURA_SOUND_EVENT_LOOKUP_GLOBALINIT( uHASHID, szNAME ) // ignore

    #define ASURA_SOUND_EVENT_LOOKUP( uHASHID, szNAME ) \
	    Asura_Sound_Event_Lookup::FindEvent_CalledByMacro( uHASHID, false )

    // Version of the macro that allows calling code to handle missing events
    // rather than asserting about it.
    #define ASURA_SOUND_EVENT_LOOKUP_IF_LOADED( uHASHID, szNAME ) \
        Asura_Sound_Event_Lookup::FindEvent_CalledByMacro( uHASHID, true )

#endif // else not ASURA_TOOLS


// C L A S S E S /////////////////////////////////////////////////////////////

// interface class for game sound playback calls, equivalent to 
// Asura_Sound_System for those calls.
class Asura_Sound_Event_System
{
	friend class Asura_Sound_Event_Lookup;

public:

	static void Initialise();
	static void Reset();
	static void Shutdown();

	// Update() will traverse the active event instance list checking sound status, updating sound parameters etc.
	static void Update();

	// PlayEvent()
	// if pointer to Asura_Handle is passed in, the external handle is returned and must 
	// be released when the calling code is finised with it (by calling StopEvent() or
	// FinishedWithHandle()
	static void PlayEvent( const Asura_Sound_Event* pxEvent, Asura_Handle* piHandleOut, 
							const Asura_Vector_3* pxPos = NULL, float fVolumeScale = 1.0f,
							Asura_Guid uSourceGuid = ASURA_GUID_UNREGISTERED);

	// Phonons keep an active instance and update it if/when they change, and trigger the
	// event to play based on their own internal update
	static Asura_Handle CreatePhononInstance( const Asura_Phonon* pxPhonon );
	static void UpdatePhononInstanceParameters( Asura_Handle iHandle, const Asura_Phonon* pxPhonon );
	static void TriggerPhononInstanceEventPlay( Asura_Handle iHandle, const Asura_Phonon* pxPhonon );

	// Stop() destroys the sound and releases the event instance handle, so it uses the reference
	// to set the handle to be invalid automatically. If fadeout parameter is set the sound's volume
	// is reduced a little each frame until it reaches 0 before actually stopping and destroying it.
	static void Stop( Asura_Handle& iHandleRef, float fFadeOut = 0.0f );


	// FinishedWithHandle() allows the event sound system code to free up the event handle when 
	// the event has finished playing. If the event is looping it should assert and stop it
	// immediately. The reference is used to set the handle to be invalid automatically.
	static void FinishedWithHandle( Asura_Handle& iHandleRef );

	// IsPlaying() returns whether the given event instance is playing
	static bool IsPlaying( Asura_Handle iHandle );

	/// IsLooping() returns true if the event instance loops (either in full or in part)
	static bool IsLooping( Asura_Handle iHandle );

	// set event instance position
	static void UpdatePosition( Asura_Handle iHandle, const Asura_Vector_3& xPos );

	// scale the volume (e.g. to make footsteps louder when moving quickly)
	static void SetVolumeScale( Asura_Handle iHandle, float fVolumeScale );
	static float GetVolumeScale( Asura_Handle iHandle );

	// scale the pitch (e.g. to make an engine sound loop pitch up/down with speed)
	static void SetPitchScale( Asura_Handle iHandle, float fPitchScale );
	static float GetPitchScale( Asura_Handle iHandle );

	// get outer radius of sound event
	static float GetOuterRadius( Asura_Handle iHandle );

	//Gets the maximum instances allowed of each sound event type , for those events that haven't
	//got a value set. If zero then there is no maximum.
	static u_int GetDefaultMaxInstances() {return s_uDefaultMaxInstances;}
	static void SetDefaultMaxInstances(u_int uMax) {s_uDefaultMaxInstances = uMax;}

	//used by playback code to track the number of instances of each sound event
	static void AddToActiveInstanceCount(Asura_Hash_ID uGroupID);
	static void RemoveFromActiveInstanceCount(Asura_Hash_ID uGroupID);
	static u_int GetNumActiveInstances(Asura_Hash_ID uInstanceGroupID);

#if defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)
	//Removes any references to the deleted event from any playing instances.
	static void SoundEventDeleted(const Asura_Sound_Event* pxEvent);
#endif	//defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)

	//----------------------------------
	// SOUND REGISTRY FUCNTION POINTERS
	//----------------------------------
	// Set these up project side if you want to use the Sound Registry function You must define all or none of these functions.
	static void SetSoundRegistryFunction( void ( *pfnFunction )( const Asura_Guid uSourceGuid, const Asura_Vector_3* pxPosition, const u_int uSoundType, const float fSoundRange ) ) { s_pfnSoundRegistryFn = pfnFunction; }
	static bool GetSoundRegistryFunctionDefined( void ) { return s_pfnSoundRegistryFn == NULL ? false : true; }

	static bool IsSphereInRange( const Asura_Vector_3& xPosition, float fRadius );

private:

	static float AmplitudeToDecibels( float fAmplitude );

	static void UpdateEventInstance(Asura_Handle iHandle);
	static void StopAllWithNoExternalHandleCallback(Asura_Handle iHandle);

	static void CountActiveInstancesCallback(Asura_Handle iHandle);

#if defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)
	static void SoundEventDeletedCallback(Asura_Handle iHandle);
#endif	//defined(ASURA_TOOLS) || defined(ASURA_USE_MULTIPLE_HANDLERS)

    static Asura_Sound_Event_Instance* s_pxActiveEventList;
    static Asura_Tracking_Handle_List* s_pxActiveEventHandleList;

	//For tracking the number of instances of each sound event or sound event group that are playing
	static Asura_Mempool* s_pxInstanceCountMempool;
	static Asura_SoundEvent_InstanceCount** s_ppxInstanceCountTable;


	static u_int s_uDefaultMaxInstances;

#if defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )
public:
	static bool s_bPrintSoundEvents;
#endif	//defined( ASURA_DEBUG_VARIABLE_SYSTEM ) && defined( ASURA_SOUND_INCLUDE_TOOLS_DATA )

#if defined( ASURA_USE_DEBUGOVERLAY )
private: 
	static void DoDebugOverlayListSounds( void* const pxIgnore );
	static void DoDebugOverlayListSound( Asura_Handle iHandle );
	static void DoDebugOverlayDrawSounds( void* const pxIgnore );
	static void DoDebugOverlayDrawSound( Asura_Handle iHandle );
#endif

	// Called when a sound is going to successfully play. Starts as NULL override at project level if you want to use it.
	// Remember to define every all three of these functions!
	static void ( *s_pfnSoundRegistryFn ) ( const Asura_Guid uSourceGuid, const Asura_Vector_3* pxPosition, const u_int uSoundType, const float fSoundRange );
};

// class to deal with lookup of events by hashID. Not to be used directly (see
// the macros at the top of this file).
class Asura_Sound_Event_Lookup
{

public:

#ifdef ASURA_TOOLS

	Asura_Sound_Event_Lookup( Asura_Hash_ID uNameHash, const char* szName );

	static const Asura_Sound_Event* FindEvent_CalledByMacro( Asura_Hash_ID uNameHash, const char* szName, bool bMissingEventHandledExternally );

#else

	static const Asura_Sound_Event* FindEvent_CalledByMacro( Asura_Hash_ID uNameHash, bool bMissingEventHandledExternally );

#endif

};

struct Asura_SoundEvent_InstanceCount
{
	Asura_Hash_ID m_uGroupID;
	u_int m_uCount;
	Asura_SoundEvent_InstanceCount* m_pxNext;
};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_SOUND_EVENT_SYSTEM_H___
