//////////////////////////////////////////////////////////////////////////////
//
// Asura_Sound_Mixer.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Sound/Asura_Sound_Mixer.h"

// Asura includes
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Timers.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif //#ifdef __#ifdef ASURA_SUPPORT_COMMAND_CONSOLE__

#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Viewport.h"
#endif //__#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__

#include "Sound/Asura_Audio_System.h"

// G L O B A L S /////////////////////////////////////////////////////////////
static Asura_TemplateHashTree<Asura_Sound_Mixer_Channel*> g_xMixerChannelTree;
static Asura_TemplateHashTree<Asura_Sound_Mixer_Preset*> g_xMixerPresetTree;

// D E F I N E S /////////////////////////////////////////////////////////////

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE              
#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
#define ASURA_SOUND_MIXER_CONVAR_DEBUGRENDER	L"Audio.Mixer.ShowDebug"
#endif //__#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__
#define ASURA_SOUND_MIXER_CONCMD_QUEUE			L"Audio.Mixer.Queue"
#define ASURA_SOUND_MIXER_CONCMD_UNQUEUE		L"Audio.Mixer.Unqueue"
#endif //__#ifdef ASURA_SUPPORT_COMMAND_CONSOLE__

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_Sound_Mixer* Asura_Sound_Mixer_System::s_pxInstance;
#ifdef ASURA_TOOLS
bool Asura_Sound_Mixer_System::s_bIsLoading = false;
#endif
static bool g_bShuttingDown = false;

ASURA_SOUND_MIXER_PRESET_SYSTEM_TYPE Asura_Sound_Mixer_System::s_eMixerPresetSystem = ASURA_SOUND_MIXER_PRESET_SYSTEM;

Asura_Sound_Mixer_System::PresetListItem Asura_Sound_Mixer_System::s_axPresetList[ASURA_SOUND_MIXER_LIST_SIZE];
Asura_Sound_Mixer_System::PresetListItem* Asura_Sound_Mixer_System::s_pxTopPreset = Asura_Sound_Mixer_System::s_axPresetList;

#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
bool Asura_Sound_Mixer_System::s_bRenderDebug = false;
Asura_Sound_Mixer_Channel** Asura_Sound_Mixer_System::s_ppxChannelList = NULL;
u_int Asura_Sound_Mixer_System::s_uChannelListSize = 0;
#endif //__#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__

bool Asura_Sound_Mixer_System::s_bMuteMaster = false;
bool Asura_Sound_Mixer_System::s_bMixerChangedThisFrame = false;

// F U N C T I O N S /////////////////////////////////////////////////////////

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
static bool ConsoleCmd_MixerQueuePreset(const Asura_Char* szPreset, bool bInterrogate , bool bHelp);
static bool ConsoleCmd_MixerUnqueuePreset(const Asura_Char* szPreset, bool bInterrogate , bool bHelp);
#endif //__#ifdef ASURA_SUPPORT_COMMAND_CONSOLE__

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

bool ConsoleCmd_MixerQueuePreset(const Asura_Char* szPreset, bool bInterrogate , bool bHelp)
{
	bool bResult = true;

	if (bHelp || bInterrogate)
	{
		Asura_CommandConsole::Printf("Usage: %S [PresetName]",ASURA_SOUND_MIXER_CONCMD_QUEUE);
		Asura_CommandConsole::Printf("Adds the supplied mixer preset to the priority list");
	}
	else
	{
		Asura_Hash_ID uHashID = Asura_GetHashID( szPreset );
		bResult = Asura_Sound_Mixer_System::QueuePreset(uHashID);
		if (!bResult) Asura_CommandConsole::Printf("Error queuing preset, see assert for details");
	}

	return bResult;
}

bool ConsoleCmd_MixerUnqueuePreset(const Asura_Char* szPreset, bool bInterrogate , bool bHelp)
{
	bool bResult = true;

	if (bHelp || bInterrogate)
	{
		Asura_CommandConsole::Printf("Usage: %S [PresetName]",ASURA_SOUND_MIXER_CONCMD_UNQUEUE);
		Asura_CommandConsole::Printf("Removes the supplied mixer preset from the priority list");
	}
	else
	{
		Asura_Hash_ID uHashID = Asura_GetHashID( szPreset );
		bResult = Asura_Sound_Mixer_System::UnqueuePreset(uHashID);
		if (!bResult) Asura_CommandConsole::Printf("Error unqueuing preset, see assert for details");
	}

	return bResult;
}

#ifdef ASURA_USE_AUDIO_SYSTEM
#define CONSOLE_CMDNAME_MUTE		L"Mute"
static bool ConsoleCmd_Mute( bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_MUTE L": mutes / unmutes all sound and music, independently of game options settings" );
	}
	else
	{
		if( Asura_Sound_Mixer_System::IsMasterMuted() )
		{
			Asura_Sound_Mixer_System::UnmuteMaster();
		}
		else
		{
			Asura_Sound_Mixer_System::MuteMaster();
		}
		Asura_CommandConsole::Printf( CONSOLE_CMDNAME_MUTE L" %s", Asura_Sound_Mixer_System::IsMasterMuted() ? L"on" : L"off" );
	}
	return true;
}
#endif // ASURA_USE_AUDIO_SYSTEM

#endif //__#ifdef ASURA_SUPPORT_COMMAND_CONSOLE__

// A S U R A _ S O U N D _ M I X E R _ S Y S T E M ///////////////////////////

void Asura_Sound_Mixer_System::Initialise()
{
	Asura_Assert(!s_pxInstance, "already initialised");
	if (!s_pxInstance)
	{
		s_pxInstance = new Asura_Sound_Mixer;
		
		Reset();

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
		Asura_CommandConsole::AddCmd(ASURA_SOUND_MIXER_CONCMD_QUEUE, ConsoleCmd_MixerQueuePreset, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
		Asura_CommandConsole::AddCmd(ASURA_SOUND_MIXER_CONCMD_UNQUEUE, ConsoleCmd_MixerUnqueuePreset, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
		Asura_CommandConsole::AddVar(ASURA_SOUND_MIXER_CONVAR_DEBUGRENDER, &s_bRenderDebug, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif //__#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__
#ifdef ASURA_USE_AUDIO_SYSTEM 
		Asura_CommandConsole::AddCmd( CONSOLE_CMDNAME_MUTE,	ConsoleCmd_Mute, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif

#endif //__#ifdef ASURA_SUPPORT_COMMAND_CONSOLE__
	}	
}

void Asura_Sound_Mixer_System::Update()
{
	Asura_Assert(s_pxInstance, "not initialised");
	
	s_bMixerChangedThisFrame = false;

	if (s_pxInstance)
	{
		// Step through non-default presets looking for those which have expired
		for (u_int u = 1; u < ASURA_SOUND_MIXER_LIST_SIZE; u++)
		{
			if ( (s_axPresetList[u].m_fExpiryTime != 0.0f) && (s_axPresetList[u].m_fExpiryTime < Asura_Timers::GetLocalScaledTime()) )
			{
				UnqueuePreset(s_axPresetList[u].m_uHashID);
			}
		}

		s_pxInstance->Update();

		s_bMixerChangedThisFrame = s_pxInstance->MixerChangedThisFrame();
	}
}

void Asura_Sound_Mixer_System::Reset()
{
	//Mixer channels and presets appear in the global sound file , and so don't get reset in between levels
	//But we do need to reset the channel settings to their defaults , if there is a default preset
	//RAR 26/07/2006

	// (AAM - 12.11.2007) Used to assert here if there were presets left in the queue, but that's
	// not necessary - we want to just clear everything out

	// Reset preset list
	for (int i = 0; i < ASURA_SOUND_MIXER_LIST_SIZE; i++)
	{
		s_axPresetList[i].Reset();
	}
	s_axPresetList[0].m_uHashID =  uMIXER_HASHID_DEFAULT;
	s_pxTopPreset = s_axPresetList;

	// Set initial preset
	Asura_Sound_Mixer_System::SetPresetInternal(s_pxTopPreset->m_uHashID,0.0f);

	s_eMixerPresetSystem = ASURA_SOUND_MIXER_PRESET_SYSTEM;

#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
	delete[] s_ppxChannelList;
	s_ppxChannelList = NULL;
	s_uChannelListSize = 0;
#endif //__#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__
}

void Asura_Sound_Mixer_System::Shutdown()
{
	Asura_Assert(s_pxInstance, "not initialised");

	if (s_pxInstance)
	{
		g_bShuttingDown = true;
		g_xMixerPresetTree.Traverse(DeletePreset);
		g_xMixerPresetTree.Reset();
		g_xMixerChannelTree.Traverse(DeleteChannel);
		g_xMixerChannelTree.Reset();
		g_bShuttingDown = false;

		delete s_pxInstance;
		s_pxInstance = NULL;
	
#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
		if (s_ppxChannelList)
		{
			delete[] s_ppxChannelList;
			s_ppxChannelList = NULL;
		}
		s_uChannelListSize = 0;
#endif

		Reset();

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
		Asura_CommandConsole::RemoveCmd(ASURA_SOUND_MIXER_CONCMD_QUEUE);
		Asura_CommandConsole::RemoveCmd(ASURA_SOUND_MIXER_CONCMD_UNQUEUE);
#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG
		Asura_CommandConsole::RemoveVar(ASURA_SOUND_MIXER_CONVAR_DEBUGRENDER);
#endif //__#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__
#endif //__#ifdef ASURA_SUPPORT_COMMAND_CONSOLE__
	}
}

void Asura_Sound_Mixer_System::DeleteChannel(Asura_Hash_ID uHash,Asura_Sound_Mixer_Channel** ppxChannel)
{
	delete *ppxChannel;
}

void Asura_Sound_Mixer_System::DeletePreset(Asura_Hash_ID uHash, Asura_Sound_Mixer_Preset** ppxPreset)
{
	delete *ppxPreset;
}

Asura_Sound_Mixer_Channel* Asura_Sound_Mixer_System::FindChannelInternal(Asura_Hash_ID uHashID)
{
	Asura_Sound_Mixer_Channel** pxResult = g_xMixerChannelTree.GetDataByHash(uHashID);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}

Asura_Sound_Mixer_Preset* Asura_Sound_Mixer_System::FindPresetInternal(Asura_Hash_ID uHashID)
{
	Asura_Sound_Mixer_Preset** pxResult = g_xMixerPresetTree.GetDataByHash(uHashID);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}

bool Asura_Sound_Mixer_System::SetPreset(Asura_Hash_ID uHashID, float fFadeTime /* = 0.0f */)
{
#ifdef ASURA_ASSERT
	CheckSystem(ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD);
#endif //__#ifdef ASURA_ASSERT__

	return SetPresetInternal(uHashID, fFadeTime);
}

bool Asura_Sound_Mixer_System::SetPresetInternal(Asura_Hash_ID uHashID, float fFadeTime)
{
	Asura_Sound_Mixer_Preset* pxPreset = FindPresetInternal(uHashID);
	if (pxPreset)
	{
		for (u_int u = 0; u < pxPreset->GetNumChannels(); ++u)
		{
			// find the tools mixer channel which matches
			const Asura_Sound_Mixer_ChannelPreset* pxChannelPreset = pxPreset->GetChannel(u);
			Asura_Sound_Mixer_Channel* pxChannel = FindChannelInternal(pxChannelPreset->GetChannelHashID());
			// skip code mixers (which shouldn't really end up in the presets)
			if (pxChannel && !pxChannel->IsCodeMixer())
			{
				pxChannel->Set(pxChannelPreset->GetSettings(),fFadeTime);
			}
		}

		#ifdef ASURA_USE_AUDIO_SYSTEM
			Asura_Audio_System::SetMasterDSP( pxPreset->GetDSPChain(), fFadeTime );
		#endif

		return true;
	}
	return false;
}

// JAT 26/02/07 IsIgnoreGameSpeedPitch() returns ignore game speed pitch flag (false if channel is NULL)
bool Asura_Sound_Mixer_System::IsIgnoreGameSpeedPitch( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert( s_pxInstance, "not initialised" );
	return s_pxInstance->IsIgnoreGameSpeedPitch( pxChannel );
}

bool Asura_Sound_Mixer_System::GetMixerUseReverb( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert( s_pxInstance, "not initialised" );
	return s_pxInstance->GetUseReverb( pxChannel );
}

bool Asura_Sound_Mixer_System::GetMixerUseDistanceFilter( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert( s_pxInstance, "not initialised" );
	return s_pxInstance->GetUseDistanceFilter( pxChannel );
}

bool Asura_Sound_Mixer_System::GetMixerUseDoppler( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert( s_pxInstance, "not initialised" );
	return s_pxInstance->GetUseDoppler( pxChannel );
}

bool Asura_Sound_Mixer_System::GetMixerUseOcclusion( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert( s_pxInstance, "not initialised" );
	return s_pxInstance->GetUseOcclusion( pxChannel );
}

bool Asura_Sound_Mixer_System::GetMixerUseMasterDSP( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert( s_pxInstance, "not initialised" );
	return s_pxInstance->GetUseMasterDSP( pxChannel );
}

// GetMixerVolume() returns master volume if channel is NULL
float Asura_Sound_Mixer_System::GetMixerVolume( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	Asura_Assert(s_pxInstance, "not initialised");
	return s_pxInstance->GetMixerVolume(pxChannel);
}

void Asura_Sound_Mixer_System::WriteToStream (Asura_Chunk_Stream& xStream)
{
	for(u_int u = 0; u < ASURA_SOUND_MIXER_LIST_SIZE; ++u)
	{
		Asura_Hash_ID uMixer = s_axPresetList[u].m_uHashID;
		xStream << uMixer;
	}
}

void Asura_Sound_Mixer_System::ReadFromStream (Asura_Chunk_Stream& xStream)
{
	//Read the hash ID's of the mixer presets and queue them 
	//(we have to reset the system first on the assumption that this is a checkpoint load or something similar)
	Asura_Sound_Mixer_System::Reset();
	for(u_int u = 0; u < ASURA_SOUND_MIXER_LIST_SIZE; ++u)
	{
		Asura_Hash_ID uMixer = ASURA_HASH_ID_UNSET;
		xStream >> uMixer;

		if( (uMixer != ASURA_HASH_ID_UNSET) && (uMixer != uMIXER_HASHID_DEFAULT) )
		{
			Asura_Sound_Mixer_System::QueuePreset(uMixer);
		}
	}
}

void Asura_Sound_Mixer_System::AddSolo()
{
	Asura_Assert(s_pxInstance, "not initialised");
	s_pxInstance->AddSolo();
}

void Asura_Sound_Mixer_System::RemoveSolo()
{
	Asura_Assert(s_pxInstance, "not initialised");
	s_pxInstance->RemoveSolo();
}

void Asura_Sound_Mixer_System::AddChannelToTree( Asura_Sound_Mixer_Channel& xChannel )
{
	Asura_Assert(!FindChannel(xChannel.GetHashID()), "mixer channel already exists in tree");
	g_xMixerChannelTree.AddNode(xChannel.GetHashID(), &xChannel);

#ifdef ASURA_TOOLS
	if(!s_bIsLoading)//while loading ,the channels get added by the chunk process code
	{
		Asura_Assert(s_pxInstance, "not initialised");
		s_pxInstance->AddChannel(&xChannel);
	}
#endif
}

void Asura_Sound_Mixer_System::RemoveChannelFromTree( Asura_Sound_Mixer_Channel& xChannel )
{
	//Don't bother removing entries from the tree while shutting down , since the tree will be fully reset afterwards.
	//RAR 12/07/2006
	if(g_bShuttingDown) return;
#ifdef ASURA_ASSERT
	bool bOK = 
#endif
	g_xMixerChannelTree.DeleteNodeByHash( xChannel.GetHashID() );
	Asura_Assert(bOK, "failed to delete node from hash tree");

#ifdef ASURA_TOOLS
	Asura_Assert(s_pxInstance, "not initialised");
	s_pxInstance->RemoveChannel(&xChannel);
#endif
}

void Asura_Sound_Mixer_System::SetChanged()
{
	Asura_Assert(s_pxInstance, "not initialised");
	s_pxInstance->SetChanged();
}

#ifdef ASURA_TOOLS
u_int Asura_Sound_Mixer_System::GetNumberOfChannels()
{
	return g_xMixerChannelTree.GetNumberOfNodes();
}

Asura_Sound_Mixer_Channel* Asura_Sound_Mixer_System::GetChannelByIndex(u_int uIndex)
{
	Asura_Sound_Mixer_Channel** pxResult = g_xMixerChannelTree.GetDataByIndex(uIndex);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}

u_int Asura_Sound_Mixer_System::GetNumberOfPresets()
{
	return g_xMixerPresetTree.GetNumberOfNodes();
}

Asura_Sound_Mixer_Preset* Asura_Sound_Mixer_System::GetPresetByIndex(u_int uIndex)
{
	Asura_Sound_Mixer_Preset** pxResult = g_xMixerPresetTree.GetDataByIndex(uIndex);
	if (pxResult)
	{
		return *pxResult;
	}
	return NULL;
}

bool Asura_Sound_Mixer_System::RenameChannel(const char* szOldName,const char* szNewName)
{
	Asura_Hash_ID uOldHash = Asura_GetHashID(szOldName);
	Asura_Hash_ID uNewHash = Asura_GetHashID(szNewName);

	//make sure the mixer channel exists
	Asura_Sound_Mixer_Channel* pxChannel = GetChannelByHash(uOldHash);
	if(!pxChannel) return false;

	//see if the hash replacement is okay
	if(uOldHash != uNewHash)
	{
		if(!g_xMixerChannelTree.ReplaceHash(uOldHash,uNewHash)) return false;
	}

	pxChannel->m_xName = szNewName;
	pxChannel->m_uHashID = uNewHash;

	return true;
}

void Asura_Sound_Mixer_System::SetToolsMute(Asura_Hash_ID uHashID, bool bMute)
{
	Asura_Assert(s_pxInstance, "not initialised");

	if(uHashID == uMIXER_HASHID_MASTER)
	{
		s_pxInstance->SetToolsMute(bMute);
	}
	else
	{
		Asura_Sound_Mixer_Channel* pxChannel = Asura_Sound_Mixer_System::FindChannelInternal(uHashID);
		if(pxChannel)
		{
			pxChannel->SetToolsMute(bMute);
		}
	}
}

#endif //ASURA_TOOLS


void Asura_Sound_Mixer_System::MuteMaster()
{
	Asura_Assert(s_pxInstance, "not initialised");
	s_bMuteMaster = true;
	s_pxInstance->SetChanged();
}


void Asura_Sound_Mixer_System::UnmuteMaster()
{
	Asura_Assert(s_pxInstance, "not initialised");
	s_bMuteMaster = false;
	s_pxInstance->SetChanged();
}


bool Asura_Sound_Mixer_System::QueuePreset( Asura_Hash_ID uPreset, float fFadeInTimeOverride /* = -1.0f */)
{
#ifdef ASURA_ASSERT
	CheckSystem(ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW);
#endif //__#ifdef ASURA_ASSERT__

	if (!s_pxInstance) 
	{
		Asura_Assertf(s_pxInstance,("Preset 0x%x activate requested while SS2_Mixer inactive",uPreset));
		return false;
	}

	// Check preset is not 'default'
	if (uPreset == uMIXER_HASHID_DEFAULT)
	{
		Asura_Assert(uPreset != uMIXER_HASHID_DEFAULT, "Trying to queue default preset in mixer - illegal");
		return false;
	}

	// Get priority
	const Asura_Sound_Mixer_Preset* pxNewPreset = FindPresetInternal(uPreset);
	if (!pxNewPreset)
	{
		Asura_Assertf(pxNewPreset,("Couldn't find preset 0x%x in mixer system to queue",uPreset));
		return false;
	}

	// Store current top preset
	Asura_Hash_ID uOldTopPresetHash = s_pxTopPreset->m_uHashID;

	// Check for space
	bool bSpaceInList = &s_axPresetList[ASURA_SOUND_MIXER_LIST_SIZE - 1] != s_pxTopPreset;

	// Find a slot for the preset
	bool bReplace = false;
	u_int uInsertSpace = 0;
	while (uInsertSpace < ASURA_SOUND_MIXER_LIST_SIZE)
	{
		// We've got to an empty space
		if (s_axPresetList[uInsertSpace].m_uHashID == ASURA_HASH_ID_UNSET) break;
		const Asura_Sound_Mixer_Preset* pxPreset = FindPresetInternal(s_axPresetList[uInsertSpace].m_uHashID);
		if (pxPreset)
		{
			// place our new preset in the spot occupied with the first current preset with higher priority
			// if there's space
			if ((pxPreset->GetPriority() > pxNewPreset->GetPriority()) && bSpaceInList) break;

			// or if we're timed, replace any other instances of ourselves
			if ((pxPreset->GetLifeTime() != 0.0f)&&(pxNewPreset->GetLifeTime() != 0.0f)) 
			{
				bReplace = true;
				break;
			}
		}
		else
		{
			Asura_Assertf(false,("Couldn't find queued preset 0x%x in sound mixer, deleting reference from queue",s_axPresetList[uInsertSpace].m_uHashID));
			UnqueuePreset(s_axPresetList[uInsertSpace].m_uHashID,0.0f);
		}

		uInsertSpace++;

		// Run out of space
		if (uInsertSpace >= ASURA_SOUND_MIXER_LIST_SIZE)
		{
			Asura_Assertf(false,("Out of mixer stack space! Preset 0x%x will not be applied. Consider increasing ASURA_SOUND_MIXER_LIST_SIZE.",uPreset));
			return false;
		}
	}

	// Find top preset
	u_int uTopPresetPosition = (ASURA_SOUND_MIXER_LIST_SIZE - 1);
	while (s_axPresetList[uTopPresetPosition].m_uHashID == ASURA_HASH_ID_UNSET) { uTopPresetPosition--; }

	// Shift current presets up queue
	if (!bReplace)
	{
		for (int i = static_cast<int>(uTopPresetPosition); i >= static_cast<int>(uInsertSpace); i--)
		{
			if (i < (ASURA_SOUND_MIXER_LIST_SIZE - 1))
			{
				s_axPresetList[i+1] = s_axPresetList[i];
			}
		}

		//Account for new preset
		uTopPresetPosition++;
	}

	// Add new preset
	s_axPresetList[uInsertSpace].m_uHashID = uPreset;
	s_axPresetList[uInsertSpace].m_fExpiryTime = pxNewPreset->GetLifeTime();
	if (s_axPresetList[uInsertSpace].m_fExpiryTime > 0.0f) s_axPresetList[uInsertSpace].m_fExpiryTime += Asura_Timers::GetLocalScaledTime();

	// Update top preset
	s_pxTopPreset = &s_axPresetList[uTopPresetPosition];

	// Attempt to apply new preset (if required)
	if (uOldTopPresetHash != s_pxTopPreset->m_uHashID)
	{
		float fFadeInTime = (fFadeInTimeOverride >= 0.0f) ? fFadeInTimeOverride : pxNewPreset->GetFadeInTime();
		bool bResult = Asura_Sound_Mixer_System::SetPresetInternal( uPreset, fFadeInTime );
		if (!bResult)
		{
			Asura_Assert(false,"Couldn't apply preset or preset doesn't exist");
			return false;
		}
	}

	return true;
}

bool Asura_Sound_Mixer_System::UnqueuePreset( Asura_Hash_ID uPreset, float fFadeOutTimeOverride /* = -1.0f */ )
{
#ifdef ASURA_ASSERT
	CheckSystem(ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW);
#endif //__#ifdef ASURA_ASSERT__

	if (!s_pxInstance) 
	{
		Asura_Assertf(s_pxInstance,("Attempt to unqueue preset 0x%x while Asura_Sound_Mixer_System inactive",uPreset));
		return false;
	}

	// Check preset is not 'default'
	if (uPreset == uMIXER_HASHID_DEFAULT)
	{
		Asura_Assert(uPreset == uMIXER_HASHID_DEFAULT,"Trying to unqueue default preset in mixer - illegal");
		return false;
	}

	// Check the preset exists
	Asura_Sound_Mixer_Preset* pxPreset = FindPresetInternal(uPreset);
	Asura_Assertf(pxPreset,("Can't find preset 0x%x in mixer system to unqueue",uPreset));
	if (!pxPreset) return false;

	// Store current top preset
	Asura_Hash_ID uOldTopPresetHash = s_pxTopPreset->m_uHashID;

	// Find the preset in the queue
	bool bFoundUnexpiredPreset = false;
	u_int uRemovalLoop = 0;
	while (uRemovalLoop < ASURA_SOUND_MIXER_LIST_SIZE)
	{
		if (s_axPresetList[uRemovalLoop].m_uHashID == uPreset)
		{
			// Check the preset is valid for removal
			if (s_axPresetList[uRemovalLoop].m_fExpiryTime > Asura_Timers::GetLocalScaledTime())
			{
				bFoundUnexpiredPreset = true;
			}
			else
			{
				// Remove the preset
				s_axPresetList[uRemovalLoop].Reset();
				break;
			}
		}

		uRemovalLoop++;
		if (uRemovalLoop >= ASURA_SOUND_MIXER_LIST_SIZE)
		{
			Asura_Assertf(bFoundUnexpiredPreset,("Trying to unqueue preset 0x%x which doesn't exist in queue",uPreset));
			Asura_Assertf(!bFoundUnexpiredPreset,("Trying to unqueue timed preset - illegal",uPreset));
			return false;
		}
	}

	// Shift the other presets (if we're not at the end and there are actually presets to move)
	while (((uRemovalLoop+1) < ASURA_SOUND_MIXER_LIST_SIZE) && (s_axPresetList[uRemovalLoop+1].m_uHashID != ASURA_HASH_ID_UNSET))
	{
		s_axPresetList[uRemovalLoop] = s_axPresetList[uRemovalLoop+1];
		s_axPresetList[uRemovalLoop+1].Reset();
		uRemovalLoop++;
	}

	// Update top preset
	s_pxTopPreset = &s_axPresetList[uRemovalLoop-1];

	// Apply new top preset (where it has changed)
	if (s_pxTopPreset->m_uHashID != uOldTopPresetHash)
	{
		float fFadeOutTime = (fFadeOutTimeOverride >= 0.0f) ? fFadeOutTimeOverride : 1.0f;

		if (fFadeOutTimeOverride < 0.0f)
		{
			// Find fade out time for old preset
			Asura_Sound_Mixer_Preset* pxOldPreset = FindPresetInternal(uOldTopPresetHash);
			Asura_Assert(pxOldPreset,"Couldn't find old top preset");
			if (pxOldPreset) fFadeOutTime = pxOldPreset->GetFadeOutTime();
		}		

		// apply all presets in order to ensure the correct value is set when the top preset doesn't include it.
		u_int uPresetToApply = 0;
		while ((uPresetToApply < ASURA_SOUND_MIXER_LIST_SIZE) && (s_axPresetList[uPresetToApply].m_uHashID != ASURA_HASH_ID_UNSET))
		{
#ifdef ASURA_ASSERT
			bool bResult = 
#endif //__#ifdef ASURA_ASSERT__
			Asura_Sound_Mixer_System::SetPresetInternal( s_axPresetList[uPresetToApply].m_uHashID, fFadeOutTime );
			Asura_Assert(bResult,"Preset applied in DeactivatePreset not found, this should be impossible");
			++uPresetToApply;
		}
	}

	return true;
}

#ifdef ASURA_SOUND_MIXER_RENDER_DEBUG

// Stack Text Variables
DEBUGVAR Asura_Char* dbg_szStackTitle = "Preset Queue:   ";
DEBUGVAR Asura_Char* dbg_szStackInfoTitle = "Queue Status:   ";
DEBUGVAR u_int dbg_uStackColorNormal = 0xFFFFFFFF;
DEBUGVAR u_int dbg_uStackColorTopmost = 0xFF00FF00;

// Mixer Text Variables
DEBUGVAR Asura_Char* dbg_szMixerTitle = "Mixer Channels: ";
DEBUGVAR u_int dbg_uMixerColor = 0xFFFFFFFF;

// Mixer Bar Variables
DEBUGVAR u_int dbg_uBarColour = 0xFF999999;
DEBUGVAR u_int dbg_uBarColourUnselected = 0xFF992222;

// Shared Variables
DEBUGVAR float dbg_fAllVertPadding = ASURA_SOUND_MIXER_RENDER_TOP;
DEBUGVAR float dbg_fAllWidth = ASURA_SOUND_MIXER_RENDER_WIDTH;
//DEBUGVAR float dbg_fAllTextScale = ASURA_SOUND_MIXER_FONT_SCALE;
DEBUGVAR u_int dbg_uAllTextFlags = ASURA_TEXTFORMAT_LEFTJUSTIFIED;        

Asura_Handle	Asura_Sound_Mixer_System::m_uDebugFontHandle = ASURA_INVALID_HANDLE;
float			Asura_Sound_Mixer_System::m_fDebugFontScale = -1.0f;

bool Asura_Sound_Mixer_System::ChannelTreeCallback(Asura_Hash_ID uHash, Asura_Sound_Mixer_Channel** ppxChannel, void* pUserData)
{
	u_int* uUserData = static_cast<u_int*>(pUserData);
	Asura_Assert(*uUserData < s_uChannelListSize,"Too many channels in tree");
	s_ppxChannelList[*uUserData] = *ppxChannel;
	(*uUserData)++;

	return true;
}

void Asura_Sound_Mixer_System::DebuggingRender()
{
	if (!s_pxInstance || !s_bRenderDebug) return;

	if (s_eMixerPresetSystem == ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD)
	{
		Asura_Assert(s_eMixerPresetSystem != ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD,"Attempting to use sound mixer preset system debug rendering when old system is active");
		return;
	}

	if(m_uDebugFontHandle == ASURA_INVALID_HANDLE)
	{
		m_uDebugFontHandle = Asura_Font_System::GetFontHandle(ASURA_SOUND_MIXER_FONT_FACE);
	}
	if (m_uDebugFontHandle != ASURA_INVALID_HANDLE) Asura_Font_System::SetFont(m_uDebugFontHandle);

	if(m_fDebugFontScale < 0)
	{
		m_fDebugFontScale = ASURA_SOUND_MIXER_FONT_SCALE;
	}

	float fTextHeight = Asura_Font_System::GetFontHeight() * m_fDebugFontScale;
	float fTextOffset = Asura_Font_System::GetWidthOfString("(XX)",m_fDebugFontScale);
	float fBarOffset = Asura_Font_System::GetWidthOfString("100 ",m_fDebugFontScale);
	float fMeatOffset = Asura_Font_System::GetWidthOfString(dbg_szMixerTitle,m_fDebugFontScale) * 1.1f;

	float fBarWidth = dbg_fAllWidth / 3.0f;

	float fScreenWidth = (float)Asura_Viewport::GetWidth();
	float fScreenHeight = (float)Asura_Viewport::GetHeight();
	static Asura_Bounding_Rectangle xStackRect;
	static Asura_Bounding_Rectangle xMixerRect;

	xStackRect.MinY = dbg_fAllVertPadding;
	xStackRect.MaxY = dbg_fAllVertPadding + (fTextHeight * (ASURA_SOUND_MIXER_LIST_SIZE+1));
	xStackRect.MinX = fScreenWidth - dbg_fAllWidth;
	xStackRect.MaxX = fScreenWidth;

	int iStackNameMaxCharacters = Asura_Maths::FloatToInt_Near(((dbg_fAllWidth - fMeatOffset) - (fTextOffset)) / Asura_Font_System::GetWidthOfChar('X',m_fDebugFontScale));
	
	// Debug Rendering: Queue Information

	Asura_Font_System::Print( xStackRect, 0.0f, 0.0f, 1.0f, m_fDebugFontScale, dbg_uStackColorNormal, dbg_uAllTextFlags, dbg_szStackInfoTitle );
	
	int iFull = (int)( ASURA_SOUND_MIXER_LIST_SIZE - ((s_pxTopPreset - s_axPresetList) / (&s_axPresetList[1] - s_axPresetList)) - 1 );		// PAZ - fix for PC.x64 build ** POINTER ARITHMETIC **
		
	static Asura_Char acQueueStatus[20];
	Asura_PrintfToBfr(acQueueStatus, "%d of %d slots free", 20, true, iFull, ASURA_SOUND_MIXER_LIST_SIZE );
	Asura_Font_System::Print( xStackRect, fMeatOffset, 0.0f, 1.0f, m_fDebugFontScale, dbg_uStackColorNormal, dbg_uAllTextFlags, acQueueStatus );

	// Debug Rendering: Queue
	u_int uCount = 2;
	u_int uDupCount = 1;
	
	Asura_Font_System::Print( xStackRect, 0.0f, (uCount*fTextHeight), 1.0f, m_fDebugFontScale, dbg_uStackColorNormal, dbg_uAllTextFlags, dbg_szStackTitle );
		
	for (int i = (ASURA_SOUND_MIXER_LIST_SIZE - 1); i >= 0; i--)
	{
		if (s_axPresetList[i].m_uHashID != ASURA_HASH_ID_UNSET)
		{
			const u_int dbg_uTextColor = (uCount > 2) ? dbg_uStackColorNormal : dbg_uStackColorTopmost;
			const Asura_Sound_Mixer_Preset* pxPreset = FindPresetInternal(s_axPresetList[i].m_uHashID);
			if (pxPreset)
			{
				if (pxPreset->GetLifeTime() > 0.0f)
				{
					static Asura_Char acPresetTime[5];
					Asura_PrintfToBfr(acPresetTime, "%.1fs", 5, true, Asura_Maths::FtoD(s_axPresetList[i].m_fExpiryTime - Asura_Timers::GetLocalScaledTime()) );
					Asura_Font_System::Print( xStackRect, xStackRect.GetWidth() - (fTextOffset * 1.5f), (uCount*fTextHeight), 1.0f, m_fDebugFontScale, dbg_uTextColor, dbg_uAllTextFlags, acPresetTime );
				}
				else if ((i > 0) && (s_axPresetList[i].m_uHashID == s_axPresetList[i-1].m_uHashID))
				{
					uDupCount++;
					continue; //don't render this item
				}
				else if (uDupCount > 1)
				{
					static Asura_Char acPresetDupCount[5];
					Asura_PrintfToBfr(acPresetDupCount, "x%d", 5, true, uDupCount );
					Asura_Font_System::Print( xStackRect, xStackRect.GetWidth() - (fTextOffset * 1.5f), ((uCount)*fTextHeight), 1.0f, m_fDebugFontScale, dbg_uTextColor, dbg_uAllTextFlags, acPresetDupCount );
					uDupCount = 1;
				}

				static Asura_Char acPresetPriority[5];
				Asura_PrintfToBfr(acPresetPriority, "(%d)", 5, true, pxPreset->GetPriority() );
				Asura_Font_System::Print( xStackRect, fMeatOffset, (uCount*fTextHeight), 1.0f, m_fDebugFontScale, dbg_uTextColor, dbg_uAllTextFlags, acPresetPriority );
				
				static Asura_Char acPresetName[30];
				Asura_PrintfToBfr(acPresetName, "%s", Asura_Maths::Min(30,iStackNameMaxCharacters), true, pxPreset->GetName());
				int iNameLength = static_cast<int>(strlen(pxPreset->GetName()));
				if (iNameLength >= iStackNameMaxCharacters)
				{
					Asura_PrintfToBfr(acPresetName, "%s..", 30, true, acPresetName );
				}
				Asura_Font_System::Print( xStackRect, fMeatOffset + fTextOffset, (uCount*fTextHeight), 1.0f, m_fDebugFontScale, dbg_uTextColor, dbg_uAllTextFlags, acPresetName );
				
				uCount++;
			}
		}		
	}

	// Debug Rendering: Mixer
	uCount++;

	xMixerRect.MinY = xStackRect.MinY + Asura_Maths::Max(6.0f, static_cast<float>(uCount)) * fTextHeight;
	xMixerRect.MaxY = fScreenHeight;
	xMixerRect.MinX = fScreenWidth - dbg_fAllWidth;
	xMixerRect.MaxX = fScreenWidth;

	Asura_Font_System::Print( xMixerRect, 0.0f, 0.0f, 1.0f, m_fDebugFontScale, dbg_uMixerColor, dbg_uAllTextFlags, dbg_szMixerTitle );
	const Asura_Sound_Mixer_Preset* pxPreset = FindPresetInternal( s_pxTopPreset->m_uHashID );
	
	if (s_uChannelListSize != static_cast<u_int>(g_xMixerChannelTree.GetNumberOfNodes()))
	{
		if (s_ppxChannelList) 
		{
			delete[] s_ppxChannelList;
			s_ppxChannelList = NULL;
		}

		s_uChannelListSize = g_xMixerChannelTree.GetNumberOfNodes();
		if (s_uChannelListSize > 0) s_ppxChannelList = new Asura_Sound_Mixer_Channel* [s_uChannelListSize];
	}

	u_int uChannelCount = 0;
	g_xMixerChannelTree.Traverse( ChannelTreeCallback, &uChannelCount );
	
	for( u_int u=0; u < uChannelCount; u++ )
	{
		const Asura_Sound_Mixer_Channel* pxChannel = s_ppxChannelList[u];
		if (pxChannel)
		{
			u_int uBarColour = dbg_uBarColourUnselected;
			for (u_int m = 0; m < pxPreset->GetNumChannels(); m++)
			{
				if (pxPreset->GetChannel(m) && (pxPreset->GetChannel(m)->GetChannelHashID() == pxChannel->GetHashID()))
				{
					uBarColour = dbg_uBarColour;
					break;
				}
			}
			
			// Render Mixer Text
			static Asura_Char acChannelValue[5];
			Asura_PrintfToBfr(acChannelValue, "%.f", 5, true, Asura_Maths::FtoD(pxChannel->GetCurrentVolume() * 100.0f) );
			Asura_Font_System::Print( xMixerRect, fMeatOffset, (u*fTextHeight*2.0f), 1.0f, m_fDebugFontScale, dbg_uMixerColor, dbg_uAllTextFlags, pxChannel->GetName() );
			Asura_Font_System::Print( xMixerRect, fMeatOffset, (u*fTextHeight*2.0f) + fTextHeight, 1.0f, m_fDebugFontScale, uBarColour , dbg_uAllTextFlags, acChannelValue );

			// Render Mixer Bar
			float fBarMinX = xMixerRect.MinX + fMeatOffset + fBarOffset;
			float fBarMinY = xMixerRect.MinY + (2.0f * m_fDebugFontScale);
			Asura_Render::Overlay.TranslucencyMode	= ASURA_TRANSLUCENCY_NORMAL;
			Asura_Render::Overlay.Colour			= uBarColour;
			Asura_Render::Overlay.TextureIndex		= ASURA_TEXTURE_NONE;
			Asura_Render::Overlay.SetRectXY( fBarMinX, fBarMinY + (u*fTextHeight*2.0f) + (fTextHeight * 1.1f), Asura_Maths::Max( 2.0f, fBarWidth * pxChannel->GetCurrentVolume()), (fTextHeight*0.6f) );
			Asura_Render::Overlay.SetRectUV( 0.0f, 0.0f, 1.0f , 1.0f );
			Asura_Render::Overlay.Render();

			// attempt to scale to fit everything on screen
			if (Asura_Render::Overlay.Y[3] > Asura_Viewport::GetHeight())
			{
				m_fDebugFontScale *= .99f;
				break;
			}
		}
	}
}

#endif //__ifdef ASURA_SOUND_MIXER_RENDER_DEBUG__

#ifdef ASURA_ASSERT
bool Asura_Sound_Mixer_System::CheckSystem( ASURA_SOUND_MIXER_PRESET_SYSTEM_TYPE eRequiredSystem )
{
	switch (eRequiredSystem)
	{
		case ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW:
		{
			if (s_eMixerPresetSystem == ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD)
			{
				Asura_Assertf(s_eMixerPresetSystem != ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD,	("%s%s%s%s",
					"Attempting to use new sound mixer preset system when the old system is active\n\n",
					((ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD) 
						?			"The two systems are incompatible and the old system has been defined in the project engine configuration as being the active system\n\n"
						:			"The two systems are incompatible, a system hasn't been defined in the project engine configuration, but the old system has already been used\n\n"),

					((ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET)
						? "Possible Causes if the project is intending to use the new system:\n1/ A \'Set Preset\' Static Message was *previously* sent to the \'All Clients\' object in tools\n2/ The Asura_Sound_Mixer_System::SetPreset() method was *previously* used\n\n"
						: ""),

					(((ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET) || (ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD))
						? "Possible Causes if project is intending to use the old system:\n1/ A \'Queue Preset\' or \'Unqueue Preset\' Static Message was sent to the \'All Clients\' object in tools\n2/ A GUIMenu item called the Audio.Mixer.QueuePreset or Audio.Mixer.UnqueuePreset console commands\n3/ The Asura_Sound_Mixer_System::QueuePreset() or Asura_Sound_Mixer_System::UnqueuePreset() methods were used"
						: ""),
				""));
				
				return false;
			}
			else if (s_eMixerPresetSystem == ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET)
			{
				s_eMixerPresetSystem = ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW;
			}
			break;
		}
		case ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD:
		{
			if (s_eMixerPresetSystem == ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW)
			{
				Asura_Assertf(s_eMixerPresetSystem != ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW,	("%s%s%s%s",
					"Attempting to use old sound mixer preset system when the new system is active\n\n",
					((ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_NEW) 
					?			"The two systems are incompatible and the new system has been defined in the project engine configuration as being the active system\n\n"
					:			"The two systems are incompatible, a system hasn't been defined in the project engine configuration, but the new system has already been used\n\n"),

					((ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET)
					? "Possible Causes if the project is intending to use the new system:\n1/ A \'Set Preset\' Static Message was sent to the \'All Clients\' object in tools\n2/ The Asura_Sound_Mixer_System::SetPreset() method was used\n\n"
					: ""),

					(((ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET) || (ASURA_SOUND_MIXER_PRESET_SYSTEM == ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD))
					? "Possible Causes if project is intending to use the old system:\n1/ A \'Queue Preset\' or \'Unqueue Preset\' Static Message was *previously* sent to the \'All Clients\' object in tools\n2/ A GUIMenu item *previously* called the Audio.Mixer.QueuePreset or Audio.Mixer.UnqueuePreset console commands\n3/ The Asura_Sound_Mixer_System::QueuePreset() or Asura_Sound_Mixer_System::UnqueuePreset() methods were *previously* used"
					: ""),
					""));

				return false;
			}
			else if (s_eMixerPresetSystem == ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET)
			{
				s_eMixerPresetSystem = ASURA_SOUND_MIXER_PRESET_SYSTEM_OLD;
			}
			break;
		}
		case ASURA_SOUND_MIXER_PRESET_SYSTEM_UNSET:
		default:
			break;
	}

	return true;
}
#endif //__#ifdef ASURA_ASSERT__

// A S U R A _ S O U N D _ M I X E R /////////////////////////////////////////
Asura_Sound_Mixer::Asura_Sound_Mixer()
 : m_fMasterVolume(1.0f)
 , m_uNumSolo(0)
 , m_uUpdateFrameStamp(0)
 , m_bChanged(true)
#ifdef ASURA_TOOLS
 , m_bToolsMute(false)
 , m_bMixerUpdatedThisFrame(false)
#endif
{
}

void Asura_Sound_Mixer::Update()
{
	m_bMixerUpdatedThisFrame = false;

	if (!m_bChanged)
	{
		return;
	}
	m_bChanged = false;//Set changed to false now , since it may be set to true again by fade updates

	//Update fades , and determine the volume scale for non-solo channels
	float fNonSoloScale = 1.0f;
	for (u_int u = 0; u < m_xChannelsArray.Size(); ++u)
	{
		m_xChannelsArray[u]->UpdateFade();
		if(m_xChannelsArray[u]->m_fSoloFadeTimer < fNonSoloScale)
		{
			fNonSoloScale = m_xChannelsArray[u]->m_fSoloFadeTimer;
		}
	}


	++m_uUpdateFrameStamp;
	for (u_int u = 0; u < m_xChannelsArray.Size(); ++u)
	{
		m_xChannelsArray[u]->Update( fNonSoloScale, m_uUpdateFrameStamp );
		if( Asura_Sound_Mixer_System::IsMasterMuted() )
		{
			m_xChannelsArray[u]->m_fCurrentVolume = 0.0f;
		}
		else
		{
			m_xChannelsArray[u]->m_fCurrentVolume *= m_fMasterVolume;
		}
		#ifdef ASURA_TOOLS
		if(m_bToolsMute) m_xChannelsArray[u]->m_fCurrentVolume = 0.0f;
		#endif
	}

	m_bMixerUpdatedThisFrame = true;
}

void Asura_Sound_Mixer::Reset()
{}

void Asura_Sound_Mixer::AddSolo()
{
	++m_uNumSolo;
	m_bChanged = true;
}

void Asura_Sound_Mixer::RemoveSolo()
{
	Asura_Assert(m_uNumSolo, "error: no channels are solo");
	--m_uNumSolo;
	m_bChanged = true;
}

// JAT 26/02/07 returns if the ignore gamespeed pitch flag is set
// false if channel is NULL
bool Asura_Sound_Mixer::IsIgnoreGameSpeedPitch( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if ( pxChannel )
	{
		return pxChannel->IsIgnoreGameSpeedPitch();
	}
	return false;
}

bool Asura_Sound_Mixer::GetUseDistanceFilter( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if ( pxChannel )
	{
		return pxChannel->GetUseDistanceFilter();
	}
	return true;
}

bool Asura_Sound_Mixer::GetUseReverb( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if ( pxChannel )
	{
		return pxChannel->GetUseReverb();
	}
	return true;
}

bool Asura_Sound_Mixer::GetUseDoppler( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if ( pxChannel )
	{
		return pxChannel->GetUseDoppler();
	}
	return true;
}

bool Asura_Sound_Mixer::GetUseOcclusion( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if ( pxChannel )
	{
		return pxChannel->GetUseOcclusion();
	}
	return true;
}

bool Asura_Sound_Mixer::GetUseMasterDSP( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if ( pxChannel )
	{
		return pxChannel->GetUseMasterDSP();
	}
	return true;
}

// GetMixerVolume() returns master volume if channel is NULL
float Asura_Sound_Mixer::GetMixerVolume( const Asura_Sound_Mixer_Channel* const pxChannel )
{
	if (pxChannel)
	{
		return pxChannel->GetCurrentVolume();
	}
	#ifdef ASURA_TOOLS
	if(m_bToolsMute) return 0.0f;
	#endif
	if( Asura_Sound_Mixer_System::IsMasterMuted() ) return 0.0f;

	return m_fMasterVolume;
}

#ifdef ASURA_TOOLS
void Asura_Sound_Mixer::AddChannel(Asura_Sound_Mixer_Channel* pxChannel)
{
	u_int uCurNumChannels = GetNumChannels();
#ifdef ASURA_ASSERT
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		Asura_Assert(m_xChannelsArray[u] != pxChannel, "already added");
	}
#endif
	// make copy of channels
	Asura_HeapArray<Asura_Sound_Mixer_Channel*> xTempChannelsArray;
	xTempChannelsArray.SetSize(uCurNumChannels);
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		xTempChannelsArray[u] = m_xChannelsArray[u];
	}
	m_xChannelsArray.Clear();
	m_xChannelsArray.SetSize(uCurNumChannels + 1);
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		m_xChannelsArray[u] = xTempChannelsArray[u];
	}
	m_xChannelsArray[uCurNumChannels] = pxChannel;
}

void Asura_Sound_Mixer::RemoveChannel(Asura_Sound_Mixer_Channel* pxChannel)
{
	u_int uCurNumChannels = GetNumChannels();

	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		if (m_xChannelsArray[u] == pxChannel)
		{
			// swap with last and resize
			m_xChannelsArray[u] = m_xChannelsArray[uCurNumChannels-1];
			m_xChannelsArray.Shrink(uCurNumChannels - 1);
			return;
		}
	}

	Asura_Assert(false, "channel not found");
}

void Asura_Sound_Mixer::SetToolsMute(bool bMute)
{
	m_bToolsMute = bMute;
	m_bChanged = true;
}
#endif

// A S U R A _ S O U N D _ M I X E R _ C H A N N E L _ S E T T I N G S ///////
Asura_Sound_Mixer_Channel_Settings::Asura_Sound_Mixer_Channel_Settings()
: m_fVolume(1.0f)
, m_bPanOverride(false)
, m_bMute(false)
, m_bSolo(false)
// JAT 26/02/07 - ignore gamesped pitch variations
, m_bIgnoreGameSpeedPitch( false )
// AP 17/08/07 - reverb and filter
, m_bUseReverb( true )
, m_bUseDistanceFilter( true )
, m_bUseDoppler( true )
, m_bUseOcclusion( true )
, m_bUseMasterDSP( true )
{}

// A S U R A _ S O U N D _ M I X E R _ C H A N N E L /////////////////////////
Asura_Sound_Mixer_Channel::Asura_Sound_Mixer_Channel(const char* szName)
: m_pxOutput(NULL)
, m_bCodeMixer(false)
, m_fCurrentVolume(1.0f)
, m_uUpdateFrameStamp(0)
, m_bFadeToSolo(false)
, m_bFadeToMute(false)
, m_fSoloFadeTimer(1.0f)
, m_fMuteFadeTimer(1.0f)
, m_fFadeStartVolume(1.0f)
, m_fFadeEndVolume(1.0f)
, m_fVolumeFadeTimer(1.0f)
, m_fFadeDuration(0.0f)
#ifdef ASURA_TOOLS
, m_bToolsMute(false)
#endif
#if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA ) || defined( ASURA_SOUND_MIXER_RENDER_DEBUG )
, m_xName(szName)
, m_xOutputMixer("Master")
#endif
{
	m_uHashID = Asura_GetHashID(szName);

	Asura_Sound_Mixer_System::AddChannelToTree(*this);
}


Asura_Sound_Mixer_Channel::~Asura_Sound_Mixer_Channel()
{
	Asura_Sound_Mixer_System::RemoveChannelFromTree(*this);
}

void Asura_Sound_Mixer_Channel::UpdateFade()
{
	if(!m_fFadeDuration) return;
	
	//determine the amount of fade change this frame
	float fTime = Asura_Timers::GetFrameTime()/m_fFadeDuration;
		
	bool bChanged = false;
	
	if(m_bFadeToSolo)
	{
		if(m_fSoloFadeTimer != 0.0f)
		{
			//fading towards solo
			m_fSoloFadeTimer -= fTime;
			if(m_fSoloFadeTimer <= 0.0f)
			{
				//have reached solo
				m_fSoloFadeTimer = 0.0f;
				if(!m_xSettings.m_bSolo)
				{
					m_xSettings.m_bSolo = true;
					Asura_Sound_Mixer_System::AddSolo();
				}
			}
			bChanged = true;
		}
	}
	else
	{
		if(m_fSoloFadeTimer != 1.0f)
		{
			//fading from solo
			m_fSoloFadeTimer += fTime;
			if(m_fSoloFadeTimer > 1.0f) m_fSoloFadeTimer = 1.0f;

			//leave solo as soon as the timer becomes non-zero
			if(m_xSettings.m_bSolo && m_fSoloFadeTimer > 0.0f)
			{
				m_xSettings.m_bSolo = false;
				Asura_Sound_Mixer_System::RemoveSolo();
			}
			bChanged = true;

		}
	}

	if(m_bFadeToMute)
	{
		if(m_fMuteFadeTimer != 0.0f)
		{
			//fading towards mute
			m_fMuteFadeTimer -= fTime;
			if(m_fMuteFadeTimer <= 0.0f)
			{
				//have reached solo
				m_fMuteFadeTimer = 0.0f;
				if(!m_xSettings.m_bMute)
				{
					m_xSettings.m_bMute = true;
				}
			}
			bChanged = true;
		}
	}
	else
	{
		if(m_fMuteFadeTimer != 1.0f)
		{
			//fading from mute
			m_fMuteFadeTimer += fTime;
			if(m_fMuteFadeTimer > 1.0f) m_fMuteFadeTimer = 1.0f;

			//leave mute as soon as the timer becomes non-zero
			if(m_xSettings.m_bMute && m_fMuteFadeTimer > 0.0f)
			{
				m_xSettings.m_bMute = false;
			}
			bChanged = true;

		}
	}

	if(m_fVolumeFadeTimer != 1.0f)
	{
		m_fVolumeFadeTimer += fTime;
		if(m_fVolumeFadeTimer > 1.0f) m_fVolumeFadeTimer = 1.0f;

		//update the current colume in the settings
		m_xSettings.m_fVolume = m_fFadeStartVolume*(1.0f-m_fVolumeFadeTimer) + m_fFadeEndVolume*m_fVolumeFadeTimer;
		bChanged = true;
	}

	if(bChanged)
	{
		Asura_Sound_Mixer_System::SetChanged();
	}
	else
	{
		//set fade duration to zero since fading has finished
		m_fFadeDuration = 0.0f;
	}
}

void Asura_Sound_Mixer_Channel::Update(float fNonSoloScale, u_int uFrameStamp)
{
	if (m_uUpdateFrameStamp == uFrameStamp)
	{
		return;
	}
	m_uUpdateFrameStamp = uFrameStamp;

	if (IsMute())
	{
		m_fCurrentVolume = 0.0f;
	}
	else
	{
		// if any are solo, then volume is 0 unless this channel is solo
		m_fCurrentVolume = m_xSettings.m_fVolume;
		if (m_pxOutput)
		{
			m_pxOutput->Update(fNonSoloScale, uFrameStamp);
			m_fCurrentVolume *= m_pxOutput->m_fCurrentVolume;
		}

		//Apply scale for fading mute
		m_fCurrentVolume *= m_fMuteFadeTimer;

		//And for fading solo
		//fNonSoloScale is the component of volume allowed because of other channels fading to/from solo
		//1.0f-m_fSoloFadeTimer is the component of volume allowed because this channel is fading to/from solo
		float fSoloScale = fNonSoloScale + (1.0f - m_fSoloFadeTimer);
		m_fCurrentVolume *= fSoloScale;
	}
}

void Asura_Sound_Mixer_Channel::Set(const Asura_Sound_Mixer_Channel_Settings& xSettings,float fFadeTime)
{
	//Pan settings aren't faded , so set those directly
	m_xSettings.m_xSurroundPan = xSettings.m_xSurroundPan;
	m_xSettings.m_bPanOverride = xSettings.m_bPanOverride;

	// JAT 26/02/07 - ignore gamesped pitch variations
	m_xSettings.m_bIgnoreGameSpeedPitch = xSettings.m_bIgnoreGameSpeedPitch;
	m_xSettings.m_bUseDistanceFilter = xSettings.m_bUseDistanceFilter;
	m_xSettings.m_bUseReverb = xSettings.m_bUseReverb;
	m_xSettings.m_bUseDoppler = xSettings.m_bUseDoppler;
	m_xSettings.m_bUseOcclusion = xSettings.m_bUseOcclusion;
	m_xSettings.m_bUseMasterDSP = xSettings.m_bUseMasterDSP;

	//Set fade targets for volue,mute and solo
	m_fFadeDuration = fFadeTime;

	if(fFadeTime)
	{
		//Set the targets for the fade
		m_bFadeToSolo = xSettings.m_bSolo;
		m_bFadeToMute = xSettings.m_bMute;
		m_fFadeStartVolume = m_xSettings.m_fVolume;
		m_fFadeEndVolume = xSettings.m_fVolume;
		m_fVolumeFadeTimer = 0.0f;
	}
	else
	{
		//No fade , so change to the new values immediately
		SetSolo(xSettings.m_bSolo);
		SetMute(xSettings.m_bMute);
		SetVolume(xSettings.m_fVolume);
	}

	Asura_Sound_Mixer_System::SetChanged();

}

void Asura_Sound_Mixer_Channel::SetSolo( bool bIsSolo )
{
	if (m_xSettings.m_bSolo == bIsSolo)
	{
		return;
	}

	m_xSettings.m_bSolo = bIsSolo; 
	if (bIsSolo)
	{
		Asura_Sound_Mixer_System::AddSolo();
	}
	else
	{
		Asura_Sound_Mixer_System::RemoveSolo();
	}
	m_bFadeToSolo = bIsSolo;
	m_fSoloFadeTimer = bIsSolo ? 0.0f : 1.0f;

}

void Asura_Sound_Mixer_Channel::SetOutputMixerName(const char* szOutputMixer)
{
	Asura_Hash_ID uHash = Asura_GetHashID(szOutputMixer);
	m_pxOutput = Asura_Sound_Mixer_System::FindChannelInternal(uHash);
	Asura_Sound_Mixer_System::SetChanged();

	Asura_Assertf(m_pxOutput || uHash == uMIXER_HASHID_MASTER,("Failed to found output mixer channel '%s'",szOutputMixer));

	#if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA ) || defined( ASURA_SOUND_MIXER_RENDER_DEBUG )
	m_xOutputMixer = szOutputMixer;
	#endif
}


// A S U R A _ S O U N D _ M I X E R _ C H A N N E L P R E S E T /////////////
#ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA
Asura_Sound_Mixer_ChannelPreset::Asura_Sound_Mixer_ChannelPreset( 
	const Asura_Sound_Mixer_Channel_Settings& xSettings, const char* szChannelName )
: m_xChannelName(szChannelName)
{
	m_xSettings = xSettings;
	m_uChannelHashID = Asura_GetHashID(szChannelName);
}
#else
Asura_Sound_Mixer_ChannelPreset::Asura_Sound_Mixer_ChannelPreset( 
	const Asura_Sound_Mixer_Channel_Settings& xSettings, Asura_Hash_ID uChannelHashID )
{
	m_xSettings = xSettings;
	m_uChannelHashID = uChannelHashID;
}
#endif // end else not #ifdef ASURA_SOUND_INCLUDE_TOOLS_DATA

#ifdef ASURA_TOOLS
void Asura_Sound_Mixer_ChannelPreset::SetChannelName(const char* szName)
{
	m_xChannelName = szName;
	m_uChannelHashID = Asura_GetHashID(szName);
}
#endif

// A S U R A _ S O U N D _ M I X E R _ P R E S E T ///////////////////////////
Asura_Sound_Mixer_Preset::Asura_Sound_Mixer_Preset(const char* szName)
: m_fMasterVolume(0.0f)
#if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA ) || defined( ASURA_SOUND_MIXER_RENDER_DEBUG )
, m_xName(szName)
#endif // #if defined( ASURA_SOUND_INCLUDE_TOOLS_DATA ) || defined( ASURA_SOUND_MIXER_RENDER_DEBUG )
, m_fFadeInTime(1.0f)
, m_fFadeOutTime(1.0f)
, m_fLifeTime(0.0f)
, m_uDSPChain( ASURA_HASH_ID_UNSET )

{
	m_uHashID = Asura_GetHashID(szName);
	Asura_Assert(!g_xMixerPresetTree.GetDataByHash(m_uHashID), "duplicate mixer preset entry");
	g_xMixerPresetTree.AddNode( m_uHashID, this );
	m_ucPriority = (m_uHashID == uMIXER_HASHID_DEFAULT) ? 0 : 1;
}

Asura_Sound_Mixer_Preset::~Asura_Sound_Mixer_Preset()
{
	// empty out the channels
	Reset();

	// check and remove from tree of presets
	Asura_Assert(g_xMixerPresetTree.GetDataByHash(m_uHashID), "mixer preset not in tree");

	//Don't bother removing entries from the tree while shutting down , since the tree will be fully reset afterwards.
	//RAR 12/07/2006
	if(g_bShuttingDown) return;
#ifdef ASURA_ASSERT
	bool bOK = 
#endif
	g_xMixerPresetTree.DeleteNodeByHash( m_uHashID );
	Asura_Assert(bOK, "failed to delete node from hash tree");
}

// Reset() clears out all the channel presets
void Asura_Sound_Mixer_Preset::Reset()
{
	for (u_int u = 0; u < m_xChannelsArray.Size(); ++u)
	{
		delete m_xChannelsArray[u];
	}
	m_xChannelsArray.Clear();
}

#ifdef ASURA_TOOLS
void Asura_Sound_Mixer_Preset::AddChannel( const Asura_Sound_Mixer_Channel_Settings& xSettings, const char* szChannelName )
{
	Asura_Hash_ID uHashID = Asura_GetHashID(szChannelName);
	u_int uCurNumChannels = GetNumChannels();
#ifdef ASURA_ASSERT
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		Asura_Assert(m_xChannelsArray[u]->GetChannelHashID() != uHashID, "already added");
	}
#endif

	// make copy of current array of pointers
	Asura_HeapArray<Asura_Sound_Mixer_ChannelPreset*> xTempChannelsArray;
	xTempChannelsArray.SetSize(uCurNumChannels);
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		xTempChannelsArray[u] = m_xChannelsArray[u];
	}
	m_xChannelsArray.Clear();
	m_xChannelsArray.SetSize(uCurNumChannels + 1);
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		m_xChannelsArray[u] = xTempChannelsArray[u];
	}

	// add new channel to the end
	m_xChannelsArray[uCurNumChannels] = new Asura_Sound_Mixer_ChannelPreset(xSettings, szChannelName);
}

void Asura_Sound_Mixer_Preset::RemoveChannel( Asura_Hash_ID uHashID )
{
	u_int uCurNumChannels = GetNumChannels();
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		if (m_xChannelsArray[u]->GetChannelHashID() == uHashID)
		{
			// delete and put last one in its place
			delete m_xChannelsArray[u];
			m_xChannelsArray[u] = m_xChannelsArray[uCurNumChannels - 1];
			m_xChannelsArray.Shrink(uCurNumChannels - 1);
			return;
		}
	}
	//RAR 26/07/2006 removed assert for missing channel ,so the delete mixer channel code can safely
	//call this on every preset
	//Asura_Assert(false, "channel to remove not found");
}

void Asura_Sound_Mixer_Preset::ReplaceChannel(const char* szOldName, const char* szNewName)
{
	Asura_Hash_ID uOldHash = Asura_GetHashID(szOldName);

	u_int uCurNumChannels = GetNumChannels();
	for (u_int u = 0; u < uCurNumChannels; ++u)
	{
		if (m_xChannelsArray[u]->GetChannelHashID() == uOldHash)
		{
			m_xChannelsArray[u]->SetChannelName(szNewName);
			return;
		}
	}
}

Asura_Sound_Mixer_ChannelPreset* Asura_Sound_Mixer_Preset::FindChannel( const Asura_Hash_ID uHash ) const
{
	u_int uNumChannels = GetNumChannels();
	for (u_int u = 0; u < uNumChannels; ++u)
	{
		if (m_xChannelsArray[u]->GetChannelHashID() == uHash)
		{
			return m_xChannelsArray[u];
		}
	}
	return 0;
}

#endif // #ifdef ASURA_TOOLS

//eof
