// Asura_XAPO_DSP_Chorus.cpp
// Thomas Young - 28.04.09
// Copyright (C) 2009 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_AUDIO_SYSTEM

#include "Sound/XAPO_DSP/Asura_XAPO_DSP_Chorus.h"

#include "Sound/Asura_Audio_DSP.h"
#include "Sound/Asura_XAudio2.h"

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

Asura_XAPO_DSP_Chorus::Asura_XAPO_DSP_Chorus( ) : Asura_XAPOBase< Asura_XAPO_DSP_Chorus, Asura_DSP_Params_Chorus::Params >( )
{
	m_pfSignal = 0;
	m_uExpectedChans = 0;
	m_uExpectedFrames = 0;

	m_fCounter = 0.0f;
	m_bReverseCounter = false;
	m_fXaudioSampleRate = static_cast<float>( Asura_XAudio2::GetOutputSampleRate() );

}


Asura_XAPO_DSP_Chorus::~Asura_XAPO_DSP_Chorus()
{
	if( m_pfSignal )
	{
		Asura_DeallocateAlign( m_pfSignal );
	}
}


void Asura_XAPO_DSP_Chorus::DoProcess ( float* pfInputBuffer, u_int uFrames, u_int uChannels, const Asura_DSP_Params_Chorus::Params& xParams )
{
	if( !m_pfSignal )
	{
		// We need to create a temp buffer, so we make some assumptions about the buffer format based on this first call
		m_uExpectedChans = uChannels;
		m_uExpectedFrames = uFrames;

		// Allocate a buffer to store the previous signal
		const u_int uAlocSize = ( uChannels * uFrames * 4 ) * sizeof( float );

		m_pfSignal = reinterpret_cast<float*>( Asura_AllocateAlign( uAlocSize, ASURA_ALIGN_128_BIT ) );

		// Zero it since we will be using it immediately
		Asura_Mem::Set( m_pfSignal, 0, uAlocSize );
	}

	if( ( m_uExpectedChans != uChannels ) || ( m_uExpectedFrames != uFrames ) )
	{
		return;
	}

	const float fSweepTimeInSamples = ( xParams.m_fSweepDepth / 1000.0f ) * m_fXaudioSampleRate;
	const float fDelayTimeInSamples = ( xParams.m_fDelay / 1000.0f ) * m_fXaudioSampleRate;
	const float fXAudioSampleRateRecip = 1.0f / m_fXaudioSampleRate;

	// Update our signal buffer, it consists of the signal from... | 3 calls ago | 2 calls ago | 1 call ago | This call |
	// Move the last signals further down the buffer

	Asura_Mem::Move( m_pfSignal, &m_pfSignal[ uChannels * uFrames ], ( uChannels * uFrames * 3 ) * sizeof( float ) );

	// Copy the current signal into the end of the temp buffer

	Asura_Mem::Copy( &m_pfSignal[ uChannels * uFrames * 3 ], pfInputBuffer, ( uChannels * uFrames ) * sizeof( float ) );

	// Chorus

	// Algorithm:	Loop over all samples mulutplying them by an offset version on the signal.
	//				Said offset changes over time

	const float fSampleContrib1 = ( xParams.m_fWetDry * 0.5f );				// Previous sample contribution
	const float fSampleContrib2 = ( 1.0f - xParams.m_fWetDry * 0.5f );		// Current sample contribution

	// Loop over every sample
	for( int iResultPos = 0; iResultPos < static_cast<int>( uFrames ); ++iResultPos )
	{
		// This bit of code is our LFO (using a triangle wave), our period is 2 seconds
		// we could let the user specify the period but its unlikely to be useful

		if( !m_bReverseCounter )
		{
			m_fCounter += fXAudioSampleRateRecip;
			
			if( m_fCounter >= 1.0f )
			{
				m_fCounter = 1.0f;
				m_bReverseCounter = true;
			}
		}
		else
		{
			m_fCounter -= fXAudioSampleRateRecip;

			if( m_fCounter <= 0.0f )
			{
				m_fCounter = 0.0f;
				m_bReverseCounter = false;
			}
		}
		
		float fModulatedSample = m_fCounter * ( fSweepTimeInSamples );
		fModulatedSample += fDelayTimeInSamples;

		// Do a litte range checking just to be safe
		if( fModulatedSample >= ( 3.0f * static_cast<float>( uFrames ) ) )
		{
			fModulatedSample = ( 3.0f * static_cast<float>( uFrames ) ) - 1.0f;
		}
		else if( fModulatedSample <= 0.0f )
		{
			fModulatedSample = 0.0f;
		}
	
		// The rest is constructing the signal Pos1a/b is the old signal, Pos2 is the current signal

		const int iSourcePos1a = ( iResultPos - static_cast<int>( fModulatedSample ) );				// We have a floating point sample index
		const int iSourcePos1b = ( iResultPos - static_cast<int>( fModulatedSample + 0.5f ) );		// which we use to interpolate between samples
		const int iSourcePos2 = ( iResultPos );

		const int iStartOfCurrentSignal = static_cast<int>( uFrames * uChannels * 3 );

		const int iSourceFramePos1a = iStartOfCurrentSignal + iSourcePos1a * static_cast<int>( uChannels );
		const int iSourceFramePos1b = iStartOfCurrentSignal + iSourcePos1b * static_cast<int>( uChannels );
		const int iSourceFramePos2 = iStartOfCurrentSignal + iSourcePos2 * static_cast<int>( uChannels );
		
		const int iResultFramePos = iResultPos * static_cast<int>( uChannels );

		// for all channels in this frame
		for( int iChan = 0; iChan < static_cast<int>( uChannels ); ++iChan )
		{

			const int iSourceSamplePos1a = iSourceFramePos1a + iChan;
			const int iSourceSamplePos1b = iSourceFramePos1b + iChan;
			const int iSourceSamplePos2 = iSourceFramePos2 + iChan;

			const int iResultSamplePos = iResultFramePos + iChan;

			float fPos1SampleVal = Asura_Maths::Lerp( m_pfSignal[ iSourceSamplePos1a ], m_pfSignal[ iSourceSamplePos1b ], fModulatedSample - floorf( fModulatedSample ) );

			pfInputBuffer[ iResultSamplePos ]	= ( fPos1SampleVal * fSampleContrib1 + m_pfSignal[ iSourceSamplePos2 ] * fSampleContrib2 ) * xParams.m_fGain;
		}
	}
}


#endif

//eof
