//////////////////////////////////////////////////////////////////////////////
//
// Asura_BillboardPlant_Batch.h
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Author: Andrew Weinkove - 11/12/02
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura.h"

#ifdef ASURA_USE_BILLBOARD_PLANTS

#include "SpecialFX/Asura_BillboardPlant_Batch.h"

#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_Entity_Classification.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_Module.h"
#include "SpecialFX/Asura_BillboardPlant_TemplateStore.h"


const float Asura_BillboardPlant_Batch::s_fSizeVariation = 0.4f;


#ifdef ASURA_USE_WIND_SYSTEM
void Asura_BillboardPlant::Update(const float fTimeslice, const Asura_Wind_Data &xWindData, const float fStiffness, const float fMaxDistortion)
{
	const Asura_Vector_3 &xPush = xWindData.m_xWind;
#else
void Asura_BillboardPlant::Update(const float fTimeslice, const Asura_Vector_3 &xPush, const float fStiffness, const float fMaxDistortion)
{
#endif //ASURA_USE_WIND_SYSTEM
	float fWindResistance = 7.5f;
	float fWindMagnitude = xPush.Magnitude();
	Asura_Vector_3 xWindDirection;
	if (fWindMagnitude <= 0.0f)
	{
		xWindDirection.Zero();
	}
	else
	{
		xWindDirection = xPush / fWindMagnitude;
		fWindMagnitude = Asura_Maths::Max(0.0f, fWindMagnitude - fWindResistance);
		xWindDirection *= fWindMagnitude;
	}
	// Update the offset.
	m_xOffset += m_xVelocities * fTimeslice;
	// Update the velocity according to wind.
	m_xVelocities.x += xWindDirection.x * fTimeslice * (1.0f - fStiffness);
	m_xVelocities.y += xWindDirection.z * fTimeslice * (1.0f - fStiffness);
	// Test for max
	if (Asura_Maths::Abs(m_xOffset.x) > fMaxDistortion)
	{
		m_xVelocities.x = m_xVelocities.x * -0.5f;	// Bounce back if we hit the max.
	}
	// Test for max
	if (Asura_Maths::Abs(m_xOffset.y) > fMaxDistortion)
	{
		m_xVelocities.y = m_xVelocities.y * -0.5f;	// Bounce back if we hit the max.
	}
	// Update the velocity according to stiffness.
	m_xVelocities.x -= m_xOffset.x * fTimeslice * fStiffness;
	m_xVelocities.y -= m_xOffset.y * fTimeslice * fStiffness;
	// Damp the velocity.
	float fVelocityMultiplier = (1.0f - (fTimeslice * fStiffness * 0.4f));
	m_xVelocities *= fVelocityMultiplier;
}

Asura_BillboardPlant_Batch::Asura_BillboardPlant_Batch(const unsigned uPlantType,
																	const unsigned uNumberOfPlants,
																	const Asura_Vector_3 *pxPositions,
																	const int iModuleNumber)
{
	m_uPlantType = uPlantType;
	m_iPlantIndex = -1;
	m_uNumberOfPlants = uNumberOfPlants;
	Asura_Assert(m_uNumberOfPlants, "Trying to create plant batch with no plants!");
	m_iModuleNumber = iModuleNumber;
	// Copy the array of positions, bad for memory fragmentation, but this is tools.
	m_pxPlants = new Asura_BillboardPlant[m_uNumberOfPlants];
	for (unsigned u=0;u<m_uNumberOfPlants;u++)
	{
		m_pxPlants[u].m_xPosition = pxPositions[u];
		m_pxPlants[u].m_xOffset = Asura_Maths::ZeroVector2;
		m_pxPlants[u].m_xVelocities = Asura_Maths::ZeroVector2;
		m_pxPlants[u].m_xScales.x = 1.0f + ((Asura_Random::GetFloat() * s_fSizeVariation) - (s_fSizeVariation * 0.5f));
		m_pxPlants[u].m_xScales.y = 1.0f + ((Asura_Random::GetFloat() * s_fSizeVariation) - (s_fSizeVariation * 0.5f));
	}
	m_uCamVisFlags = 0xffffffff;
	SetupPlantData();
}

Asura_BillboardPlant_Batch::Asura_BillboardPlant_Batch(Asura_Handle fileHandle)
{
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uPlantType, unsigned);
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_iModuleNumber, int);
	ASURA_CHUNK_FILEEXTRACT(fileHandle, m_uNumberOfPlants, unsigned);
	Asura_Assert(m_uNumberOfPlants, "Trying to create plant batch with no plants!");
	m_pxPlants = new Asura_BillboardPlant[m_uNumberOfPlants];
	for (unsigned i=0;i<m_uNumberOfPlants;i++)
	{
		ASURA_CHUNK_FILEEXTRACT(fileHandle, m_pxPlants[i].m_xPosition, Asura_Vector_3);
		m_pxPlants[i].m_xOffset = Asura_Maths::ZeroVector2;
		m_pxPlants[i].m_xVelocities = Asura_Maths::ZeroVector2;
		m_pxPlants[i].m_xScales.x = 1.0f + ((Asura_Random::GetFloat() * s_fSizeVariation) - (s_fSizeVariation * 0.5f));
		m_pxPlants[i].m_xScales.y = 1.0f + ((Asura_Random::GetFloat() * s_fSizeVariation) - (s_fSizeVariation * 0.5f));
	}
	m_uCamVisFlags = 0xffffffff;
	SetupPlantData();
}

Asura_BillboardPlant_Batch::~Asura_BillboardPlant_Batch()
{
	delete[] m_pxPlants;
}

void Asura_BillboardPlant_Batch::Update(void)
{
	UpdateClipping();
	if (!m_uCamVisFlags ||
		m_iPlantIndex == -1)	// Don't update if the plants aren't visible.
		return;
	float fTimeSlice = Asura_Maths::Min(0.1f, Asura_Timers::GetGameTimeSlice());

	Asura_BillboardPlant_Template *pxTemplate = Asura_BillboardPlant_TemplateStore::GetTemplate(m_iPlantIndex);
	Asura_Assert(pxTemplate, "");
	if (!pxTemplate)
		return;
	float fStiffness = pxTemplate->GetStiffness();
	float fMaxDistortion = pxTemplate->GetMaxDistortion();
	for (unsigned i=0;i<m_uNumberOfPlants;i++)
	{
		float fAdjustedStiffness = Asura_Maths::Min(1.0f, fStiffness + (float((i % 20) - 10) * 0.002f));
#ifdef ASURA_USE_WIND_SYSTEM
		Asura_Wind_Data xWindData;
		Asura_Wind_System::GetWindAtPoint(m_pxPlants[i].m_xPosition, xWindData);
		m_pxPlants[i].Update(fTimeSlice, xWindData, fStiffness, fMaxDistortion);
#else
		Asura_Vector_3 xPlantPush;
		Project_GetPlantPush(xPlantPush);
		m_pxPlants[i].Update(fTimeSlice, xPlantPush, fStiffness, fMaxDistortion);
#endif //#ifdef ASURA_USE_WIND_SYSTEM
	}
}	

void Asura_BillboardPlant_Batch::Render(void)
{
	if (!m_uCamVisFlags)
		return;
	Platform_Render(m_uCamVisFlags);
}

void Asura_BillboardPlant_Batch::SetupPlantData()
{
	m_iPlantIndex = Asura_BillboardPlant_TemplateStore::GetIndexFromUniqueID(m_uPlantType);
	m_xBB.Reset();
	for (unsigned i=0;i<m_uNumberOfPlants;i++)
	{
		m_xBB.Set(m_pxPlants[i].m_xPosition);
	}
	if (m_iPlantIndex == -1)
	{
		return;
	}
	Asura_BillboardPlant_TemplateStore::SetTextureID(m_iPlantIndex);
	Asura_BillboardPlant_Template *pxTemplate = Asura_BillboardPlant_TemplateStore::GetTemplate(m_iPlantIndex);
	m_xBB.MinX -= pxTemplate->m_fWidth * 0.5f;
	m_xBB.MinZ -= pxTemplate->m_fWidth * 0.5f;
	m_xBB.MaxX += pxTemplate->m_fWidth * 0.5f;
	m_xBB.MaxZ += pxTemplate->m_fWidth * 0.5f;
	m_xBB.MaxY += pxTemplate->m_fHeight;
	m_xBB.MinY -= pxTemplate->m_fHeight;
	m_xCentre = m_xBB.CalculateCentre();
	m_fRadius = (Asura_Vector_3(m_xBB.MinX, m_xBB.MinY, m_xBB.MinZ) - m_xCentre).Magnitude();

	Asura_GI_Solution::CalculateDiffuseAtPoint(m_xLighting, m_xCentre);
}

void Asura_BillboardPlant_Batch::FillVertices(Asura_Vector_3 *pxVertices, const int iPlantIndex)
{
	if (m_iPlantIndex == -1) return;
	Asura_BillboardPlant_Template *pxTemplate = Asura_BillboardPlant_TemplateStore::GetTemplate(m_iPlantIndex);
	Asura_Assert(pxTemplate, "");
	Asura_Vector_3 xWidthVector(pxTemplate->m_fWidth, 0.0f, 0.0f);
	xWidthVector.x *= 0.5f * m_pxPlants[iPlantIndex].m_xScales.x;
	Asura_Vector_3 xHeightVector(0.0f, pxTemplate->m_fHeight * m_pxPlants[iPlantIndex].m_xScales.y, 0.0f);
	Asura_Vector_3 xTopRowOffset = Asura_Vector_3(m_pxPlants[iPlantIndex].m_xOffset.x, 0.0f, m_pxPlants[iPlantIndex].m_xOffset.y);
	Asura_Matrix_3x3 xMatrix;
	if (pxTemplate->m_bDontTurnToFaceCamera)
	{
		xMatrix.Identity();
	}
	else
	{
		Asura_Vector_3 xCamDirection(Asura_Camera::GetOrientation()->e13, 0, Asura_Camera::GetOrientation()->e33);
		Asura_Maths::OrientationFromDirection(xMatrix, xCamDirection);
	}
	xTopRowOffset = xTopRowOffset * xMatrix;
	if (pxTemplate->m_bUseRandomOffset)
	{
		// AW - 14/03/03 - This isn't really a random, just an arbitrary offset.
		float fAngle = Asura_Maths::Pi * (static_cast<float>(iPlantIndex) / static_cast<float>(m_uNumberOfPlants));
		float fSinAngle = Asura_Maths::Sine(fAngle);
		float fCosAngle = Asura_Maths::Cosine(fAngle);
		Asura_Matrix_3x3 xRotationMatrix(fCosAngle, 0.0f, fSinAngle, 
										0.0f, 1.0f, 0.0f,
										fSinAngle, 0.0f, fCosAngle);
/*		Asura_Euler_3 xEuler(0.0f, fAngle, 0.0f);
		Asura_Matrix_3x3 xRotationMatrix;
		xEuler.ToMatrix(xRotationMatrix);*/

		xMatrix = xMatrix * xRotationMatrix;
	}
	
	// Bottom left vertex.
	pxVertices[2] = m_pxPlants[iPlantIndex].m_xPosition - (xWidthVector * xMatrix);
	// Bottom right vertex.
	pxVertices[3] = m_pxPlants[iPlantIndex].m_xPosition + (xWidthVector * xMatrix);
	xHeightVector *= 1.0f - (xTopRowOffset * xTopRowOffset);
	// Top left vertex.
	pxVertices[0] = m_pxPlants[iPlantIndex].m_xPosition - (xWidthVector * xMatrix) - xHeightVector;
	pxVertices[0] += xTopRowOffset;
	// Top right vertex.
	pxVertices[1] = m_pxPlants[iPlantIndex].m_xPosition + (xWidthVector * xMatrix) - xHeightVector;
	pxVertices[1] += xTopRowOffset;
}


int Asura_BillboardPlant_Batch::GetChunkBufferSize()
{
	int iChunkSize = 0;
	iChunkSize += sizeof(unsigned);	//m_uPlantType
	iChunkSize += sizeof(int);		// m_iModuleNumber
	iChunkSize += sizeof(unsigned); //m_uNumberOfPlants
	iChunkSize += sizeof(Asura_Vector_3) * m_uNumberOfPlants; //m_pxPositions
	return iChunkSize;
}

void Asura_BillboardPlant_Batch::FillInChunkBuffer(char*& pcChunkBuffer)
{
	ASURA_CHUNK_FILL(pcChunkBuffer, m_uPlantType, unsigned);
	ASURA_CHUNK_FILL(pcChunkBuffer, m_iModuleNumber, int);
	ASURA_CHUNK_FILL(pcChunkBuffer, m_uNumberOfPlants, unsigned);
	for (unsigned i=0;i<m_uNumberOfPlants;i++)
	{
		ASURA_CHUNK_FILL(pcChunkBuffer, m_pxPlants[i].m_xPosition, Asura_Vector_3);
	}
}

void Asura_BillboardPlant_Batch::UpdateClipping()
{
	m_uCamVisFlags = 0;
	for (int i=0;i<Asura_Camera::GetNumberOfCameras();i++)
	{
		if (!Asura_Module_System::IsModuleVisible(m_iModuleNumber, (1 << i)))
			continue;
		Asura_Camera::SetCameraIndex(i);
		if (!Asura_Camera::IsSphereInFrustum(m_xCentre, m_fRadius))
			continue;
		m_uCamVisFlags |= (1 << i);
	}
}

#endif // #ifdef ASURA_USE_BILLBOARD_PLANTS
