///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_CollisionFX_System.h
//
// Created by Martin G.Harris (28/09/2007)
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Play sound events and/or PFX according to collision information from the
// physics system. Both impact and sliding collisions are supported.
//
///////////////////////////////////////////////////////////////////////////////
//
// NOTE: to use this class projects should:
//
// 1) Define ASURA_USE_COLLISIONFX
//
// 2) Call Asura_CollisionFX_System::Update() from the main Project_Game_Core
//    update loop.
//
// 3) Call Asura_CollisionFX_System::Reset() just before doing a checkpoint load.
//
// 4) Set a callback using Asura_CollisionFX_System::SetCallback().
//
// 5) Optionally, call Asura_CollisionFX_System::SetEnableSlidingCollisions(true)
//    to enable effects for sliding collisions. (But only if the callback
//    function is set up to handle them).
//
// 6) Create a Project_CollisionFX_System to handle the callback, project-specific
//    surface types, look up tables for effects, mapping from entity guids to
//    surface types etc.
//
// 7) Turn on generation of physics collision FX reports for relevant entities
//    (e.g. physics objects), by calling SetReportObjectCollisionFX() and/or
//    SetReportEnvironmentCollisionFX(). Associated entity guids need to be set
//    for the Asura_Physics_Objects.
//
// 8) Optionally, implement a project-specific effects class derived from
//    Asura_CFX_Effect to fine tune the behaviour of the effects.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_COLLISION_FX_SYSTEM_H__
#define __ASURA_COLLISION_FX_SYSTEM_H__

// Only compile this code if the symbol is defined.
#ifdef ASURA_USE_COLLISIONFX

// Includes
#include "Core/Asura_HashTable.h"

#include "Core/Asura_Mempool.h"
#include "Core/Asura_Maths.h"
#include "Network/Asura_Guid.h"

// Forward declarations
class Asura_Physics_CollisionFX_Report;
class Asura_LOS_InputData;
class Asura_LOS_OutputData;
class Asura_CFX_Effect;
class Asura_Message;
class Asura_Message_CFX_ClientUpdate_Entry;
class Asura_CFX_Effect_PFX_SE_RemoteClient;
class Asura_CFX_Decal_Handle;

///////////////
// Constants //
///////////////
// Collision type for physics collisions.
const int iASURA_CFX_TYPE_PHYSICS_IMPACT = 0; // Physics impact collision.
const int iASURA_CFX_TYPE_PHYSICS_SLIDE  = 1; // Physics sliding collision.

const int iASURA_CFX_TYPE_PROJECT_START  = 2; // Start of project-specific collision types.


//////////////
// Typedefs //
//////////////
// Callback typedef
typedef Asura_CFX_Effect*( Asura_CollisionFX_CreateEffectFuncPtr( const Asura_Physics_CollisionFX_Report&, int ) );


/////////////////////////////
// Asura_CollisionFX_Entry //
/////////////////////////////
// Entry in the table of on-going collisions.
class Asura_CollisionFX_Entry
{
public:
	Asura_CollisionFX_Entry(Asura_Guid uGuid1, Asura_Guid uGuid2, int iType);
	Asura_CollisionFX_Entry(const Asura_CollisionFX_Entry& xOther);

	Asura_CollisionFX_Entry& operator=(const Asura_CollisionFX_Entry& xRight);

	bool IsObjectObjectCollision() const { return (m_uGuid2 != ASURA_GUID_SERVER); }
	bool IsEnvironmentCollision()  const { return (m_uGuid2 == ASURA_GUID_SERVER); }

	bool IsPhysicsCollision() const
	{
		return (m_iType == iASURA_CFX_TYPE_PHYSICS_IMPACT) || (m_iType == iASURA_CFX_TYPE_PHYSICS_SLIDE);
	}

	// Equality, if and only if the colliding objects are the same and it is the same type of collision.
	bool operator==(const Asura_CollisionFX_Entry& xRight) const
	{
		return (m_uGuid1 == xRight.m_uGuid1) &&
			   (m_uGuid2 == xRight.m_uGuid2) &&
			   (m_iType == xRight.m_iType);
	}

	// Return a short key for the hash table.
	Asura_Hash_ID GetHashID() const { return m_uGuid1 + (m_uGuid2<<2);}

	// Accessors
	Asura_Guid	GetGuid1() const { return m_uGuid1; }
	Asura_Guid	GetGuid2() const { return m_uGuid2; }
	int			GetType()  const { return m_iType; }

	float GetStartTime()          const { return m_fStartTime; }
	float GetLastCollisionTime()  const { return m_fLastCollisionTime; }
	float GetSlideStartTime()     const { return m_fSlideStartTime; }
	float GetDestroyTime()        const { return m_fDestroyTime; }

	const Asura_CFX_Effect* GetEffect() const { return m_pxEffect; }

private:
	void UpdateDestroyTime(float fDelay);
	void SetEffect(Asura_CFX_Effect* pxEffect);

private:
	// Colliding entities.
	Asura_Guid m_uGuid1;
	Asura_Guid m_uGuid2;	// ASURA_GUID_SERVER for environment.
	
	// Type of collision: iASURA_CFX_TYPE_PHYSICS_IMPACT (=0) for physics impact,
	// iASURA_CFX_TYPE_PHYSICS_SLIDE (=1) for physics sliding collision,
	// other values are project-specific (e.g. weapon hitting something based on LOS test).
	int m_iType;

	// Time of first collision (i.e. when this entry was created).
	float m_fStartTime;

	// Time of last (i.e. most recent) collision.
	float m_fLastCollisionTime;

	// Start time for a series of slide collisions (only relevant for iASURA_CFX_TYPE_PHYSICS_SLIDE).
	float m_fSlideStartTime;

	// Time to destroy this entry.
	float m_fDestroyTime;

	// Contains the code to start/stop/update the effect.
	Asura_CFX_Effect* m_pxEffect;

	// Friends
	friend class Asura_CollisionFX_System;
};


////////////////////////////
// Asura_CollisionFX_List //
////////////////////////////
// Table of on-going collisions.
class Asura_CollisionFX_List : public Asura_HashTable<Asura_CollisionFX_Entry>
{
public:
	typedef Asura_HashTable_It< Asura_CollisionFX_Entry > Iterator;

	Asura_CollisionFX_List();

	// Return an interator.
	Iterator GetIterator() { return Iterator(this); }
};

/////////////////////////////
// Asura_CollisionFX_Param //
/////////////////////////////
// Parameters that can be set by the project.
class Asura_CollisionFX_Param
{
public:
	// Construct with default values.
	Asura_CollisionFX_Param();

	// Speed below which physics collisions are ignored (in m/s).
	float m_fThresholdMinSpeed;

	// Cosine of angle between velocity and normal below which
	// the collision is taken to be sliding over the environment.
	float m_fThresholdMinDotProduct;

	// Waiting time before playing a slide effect (in seconds).
	float m_fSlideDelayBeforePlaying;

	// In the sliding waiting phase, restart the timer if there is a
	// gap between collisions of more than this time (in seconds).
	float m_fSlideDelayStopWaiting;

	// Stop playing the sliding effect if we have had
	// no collisions for this period of time (in seconds).
	float m_fSlideDelayStopPlaying;

	// Fade out time when stopping sound events.
	float m_fSoundEventFadeOutTime;

	// Enable sliding collisions, which are ignored by default.
	bool m_bEnableSlidingCollisions;
};


//////////////////////////////
// Asura_CollisionFX_System //
//////////////////////////////
class Asura_CollisionFX_System
{
public:
	static void Initialise();
	static void Reset();
	static void Shutdown();

	static void Update();

	// Stops all the effects without actually deleting any of the ongoing collision entries.
	static void StopAllEffects(bool bImmediate);

	// Register collision from physics or other source (e.g. using data from Line-Of-Sight tests).
	// iTypeID is 0 or 1 for collisions from the physics system, other than that it is entirely project-specific.
	// Project code can call this with bForceSlide true to create sliding rather than impact events.
	static void RegisterCollision(const Asura_Physics_CollisionFX_Report& xCollInfo,
								  int iTypeID,
								  bool bForceSlide = false);
	
	// Helper function to register a collision based on LOS data.
	static void RegisterCollision(const Asura_LOS_InputData& xLOSInput,
								  const Asura_LOS_OutputData& xLOSOutput,
								  Asura_Guid uEntity1Guid,
								  const Asura_Vector_3& xRelVelocity,
								  int iTypeID,
								  bool bForceSlide = false);

	// Remove collision manually. Return true if an entry was removed.
	static bool RemoveCollision(Asura_Guid uGuid1, Asura_Guid uGuid2, int iTypeID, bool bImmediate);

	// Remove all collisions associated with this entity.
	static void RemoveAllCollisions(Asura_Guid uGuid, bool bImmediate);

	// Number of currently active collision FX.
	static int GetNumCollisionFX() { return s_xCollisionList.GetNumEntries(); }

	// Set the callback for project-specific collision FX effects.
	static void SetCallback(Asura_CollisionFX_CreateEffectFuncPtr* pfnCreateEffectFuncPtr)
	{
		s_pfnCreateEffectFuncPtr = pfnCreateEffectFuncPtr;
	}

	// Set the threshold minimum speed, below which physics collisions are ignored.
	static void SetThresholdMinSpeed(float fSpeed)
	{
		s_xParam.m_fThresholdMinSpeed = fSpeed;
	}

	// By default sliding collisions are ignored, but they can be turned on if the
	// project CreateEffect callback is set up to generate suitable effects for such collisions.
	static void SetEnableSlidingCollisions(bool bEnable)
	{
		s_xParam.m_bEnableSlidingCollisions = bEnable;
	}

	// Set the parameters to be used.
	static void SetParameters(const Asura_CollisionFX_Param& xParam) { s_xParam = xParam; }

	// Get the current parameters.
	static const Asura_CollisionFX_Param& GetParameters() { return s_xParam; }

#ifdef _DEBUG
	// Dump a copy of the table to the debugger.
	static void DebugDumpTable();
#endif // _DEBUG

	// Render some debugging information.
	static void RenderDebugInfo();

	// Functions for messages sent from server to client.
	static bool ShouldSendMessageToClients();
	static void SendClientUpdateMessage();
	static void HandleClientUpdateMessage(const Asura_Message* pxMessage);
	static void HandleClientCreateDecalMessage(const Asura_Message* pxMessage);

	static void SendClientCreateDecalMessage(const Asura_CFX_Decal_Handle& xHandle,
											 const Asura_Vector_3& xPosition,
											 const Asura_Vector_3& xNormal,
											 const Asura_Vector_3* pxTopDirection);

	// Enable/disable the client update mesages that are sent to remote client machines.
	static void SetEnableClientUpdateMessages(bool bEnable) { s_bEnableCFXClientUpdateMessages = bEnable; }

protected:
	static void HandleClientUpdateMessageData(const Asura_Message_CFX_ClientUpdate_Entry& xData, bool bOldMessage);

	static Asura_CollisionFX_Entry* AddEntry(const Asura_CollisionFX_Entry& xEntryToAdd);

	static void StopEffect(Asura_CollisionFX_Entry* pxEntry, bool bImmediate);

    static void* AllocateFromListNodeMempool()                { return s_xListNodeMempool.Allocate(); }
    static void  DeallocateToListNodeMempool( void* pMemory ) { s_xListNodeMempool.Deallocate( pMemory ); }

	// Create a CFX effect.
	static Asura_CFX_Effect* CreateEffect(const Asura_Physics_CollisionFX_Report& xCollInfo, int iType);

	// Delete CFX effect and set the pointer to null.
	static void DeleteEffect(Asura_CFX_Effect*& pxEffect);

	// Stop effect, delete CFX, deallocate entry. Doesn't delete the list node though.
	static void DeleteEntry(Asura_CollisionFX_Entry* pxEntry, bool bImmediate);


	// Register a sliding collision.
	static void RegisterCollision_Sliding(const Asura_CollisionFX_Entry& xTestEntry,
										  Asura_CollisionFX_Entry* pxEntry,
										  const Asura_Physics_CollisionFX_Report& xCollInfo,
										  int iTypeID);

	// Update a sliding collision, to stop the effect if it has timed out.
	static void UpdateCollision_Sliding(Asura_CollisionFX_Entry* pxEntry);

	// Update the cached value of the next time that an entry needs to be removed or stopped.
	static void UpdateDestroyedTime(float fTime)
	{
		if (fTime < s_fNextDestroyTime)
		{
			s_fNextDestroyTime = fTime;
		}
	}

	static void UpdateProcessCollisionReports();
	static void UpdateDestroyOldEntries();

	// Create an effect for a remote client machine from the message data provided by the server.
	static Asura_CFX_Effect_PFX_SE_RemoteClient* CreateRemoteClientEffect(const Asura_Message_CFX_ClientUpdate_Entry& xData);

private:
	// Hash table of all the collision FX entries.
	static Asura_CollisionFX_List s_xCollisionList;

	// Memory pools for the hash table.
	static Asura_ResizingMempool s_xListNodeMempool;
    static Asura_ResizingMempool s_xEntryMempool;

	// Project callback for creating effects.
	static Asura_CollisionFX_CreateEffectFuncPtr* s_pfnCreateEffectFuncPtr;

	// Cache the next destroy time so that we don't have to keep iterating over the list
	// just in case a collision entry might have timed out.
	static float s_fNextDestroyTime;

	// Timestamp for last received client update message
	static float s_fLastClientUpdateTimestamp;

	// Parameters that can be set by the project.
	static Asura_CollisionFX_Param s_xParam;

	// Enable the client update messages.
	static bool s_bEnableCFXClientUpdateMessages;

	// Friends
	friend class Asura_CollisionFX_List;
	friend class Asura_CollisionFX_Entry;
};

#endif //ASURA_USE_COLLISIONFX

#endif //__ASURA_COLLISION_FX_SYSTEM_H__
