/**
*
* \file Asura_Corona_System.cpp
*
* Created by Kevin Lea - 27/01/2003 09:22:11
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

// This file's header.
#include "SpecialFX/Asura_Corona_System.h"

#include <limits.h>             // USHRT_MAX

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#ifdef ASURA_USE_STREAMING
#include "Streaming/Asura_Streaming_System.h"
#endif

#ifndef ASURA_CORONA_DEFAULT_VOLUMETRICDEPTHSCALE
#define ASURA_CORONA_DEFAULT_VOLUMETRICDEPTHSCALE (0.0f)
#endif

DEBUGVAR bool dbg_bCoronaRenderingEnabled = true;

/////////////////////////
// Asura_Corona_System //
/////////////////////////

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
Asura_Corona_Handler*							Asura_Corona_System::s_pxCurrentHandler;
Asura_TemplateHashTree<Asura_Corona_Handler>	Asura_Corona_System::s_xHandlerTree;
#else
Asura_Corona_Handler Asura_Corona_System::s_xHandler;
#endif

#ifdef ASURA_TOOLS
bool Asura_Corona_System::s_bDisabled = false;
#endif

void Asura_Corona_System::Initialise()
{
#if !defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	s_xHandler.Initialise();
#endif 

	Platform_Initialise();
}

void Asura_Corona_System::Shutdown()
{
	Platform_Shutdown();

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// Just free up the whole tree
	s_xHandlerTree.Reset();
#else
	s_xHandler.Shutdown();
#endif
}

void Asura_Corona_System::Reset()
{
#ifdef ASURA_USE_STREAMING
	s_xHandlerTree.Traverse(Reset_Callback);
#else
	GetCurrentHandler().Reset();
#endif
}


void Asura_Corona_System::Update()
{
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// Loop over all the handlers rendering them.
	s_xHandlerTree.Traverse(Update_Callback);
#else
	GetCurrentHandler().Update();
#endif
}

void Asura_Corona_System::Render()
{
	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_CORONAS))
	{
		return;
	}

#ifdef ASURA_TOOLS
	if(s_bDisabled) return;
#endif

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// Loop over all the handlers rendering them.
	s_xHandlerTree.Traverse(Render_Callback);
#else
	GetCurrentHandler().Render();
#endif
}


float Asura_Corona_System::GetDefaultVolumetricDepthScale(void)
{
	return ASURA_CORONA_DEFAULT_VOLUMETRICDEPTHSCALE;
}


void Asura_Corona_System::RenderCorona( Asura_Corona* pxCorona )
{
	u_int uCamVisFlags = 0;

	if (pxCorona->m_bOnlyDrawIfPositionSetPerFrame && !pxCorona->m_bPositionUpdatedThisFrame)
	{
		return;
	}

	// reset flag
	pxCorona->m_bPositionUpdatedThisFrame = false;

	if(pxCorona->m_bIsHidden)
	{
		return;
	}

	for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
	{
		Asura_Camera::SetCameraIndex(i);

        // Distance culling.
        if( !pxCorona->m_bDoNotFadeOutWithDistance )
        {
            Asura_Vector_3 xToCam( *Asura_Camera::GetPosition() - pxCorona->m_xPosition);

            if( xToCam * xToCam >= ASURA_CORONA_FADE_DIST_END * ASURA_CORONA_FADE_DIST_END )
            {
                continue;
            }
			// [18/9/2008 AS] Additional check for coronas being behind the camera
			Asura_Vector_3 xCamFwd;
			Asura_Camera::GetForwardVector(xCamFwd);
			if ((xToCam * xCamFwd) > 0.0f)
			{
				continue;
        }

        }

		if(pxCorona->m_bUseModuleLinking)
		{
#if defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
			Asura_Bounding_Box xBB;
			xBB.SetToPoint( pxCorona->m_xPosition);
			xBB.Grow( pxCorona->m_fRadius );
			if ( !(pxCorona->m_xVisibilityQuery.IsAABBVisibleFromCurrentCamera( xBB )) ) continue;
#else
			if(!pxCorona->IsInVisibleModule()) continue;
#endif
		}

		if (Asura_Camera::IsSphereInFrustum(pxCorona->m_xPosition, pxCorona->m_fRadius))
		{
			/// Kevin Lea - 27/01/2003 14:48:52 - possibly check here for module visibility
			/// (which will require another per corona flag CheckModuleVisibility etc).

			uCamVisFlags |= (1<<i);
		}
	}

	if (!uCamVisFlags) return;

	Platform_RenderCorona(pxCorona, uCamVisFlags);
}

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
void Asura_Corona_System::Reset_Callback(Asura_Hash_ID uHash, Asura_Corona_Handler* pxData)
{
	// Don't need the hash, just reset the data in the pointer
	if (pxData)
	{
		pxData->Reset();
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Corona_System::CreateHandler
// Author       : Anthony Gregson
// Date         : 24-Mar-2006
//
// Description  : Attempts to create a corona system handler with the specified
//				  hash-ID.
//
// Parameters   : The hash-ID of the handler to create.
//
// Return value : A pointer to the handler created.
////////////////////////////////////////////////////////////////////////////////

Asura_Corona_Handler* Asura_Corona_System::CreateHandler(const Asura_Hash_ID uHash)
{
	Asura_Corona_Handler* pxHandler = s_xHandlerTree.GetDataByHash(uHash);

	// Do we already have this handler?
	if(pxHandler == NULL)
	{
		// No - create one...
		pxHandler = s_xHandlerTree.AddNode(uHash);
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have a corona handler with this hash ID.");
	}

	// Return the handler...
	return(pxHandler);
}

/// End of function 'Asura_Corona_System::CreateHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Corona_System::DestroyHandler
// Author       : Anthony Gregson
// Date         : 24-Mar-2006
//
// Description  : Attempts to destroy the corona system handler with the specified
//				  unique hash-ID.
//
// Parameters   : The hash-ID of the handler to destroy.
//
// Return value : "true" if destroyed, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_Corona_System::DestroyHandler(const Asura_Hash_ID uHash)
{
	// Is this the current handler?
	if(s_pxCurrentHandler == s_xHandlerTree.GetDataByHash(uHash))
	{
		s_pxCurrentHandler	= NULL;
	}

	// Call the delete function...
	return(s_xHandlerTree.DeleteNodeByHash(uHash));
}

/// End of function 'Asura_Corona_System::DestroyHandler'. ///
#endif //#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)

//////////////////////////
// Asura_Corona_Handler //
//////////////////////////

Asura_Corona_Handler::Asura_Corona_Handler()
{
	// AJG - 27/03/06 - Non-streaming one doesn't need to initialise at this point
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	Initialise();
#endif
}

Asura_Corona_Handler::~Asura_Corona_Handler()
{
	// AJG - 27/03/06 - Non-streaming one doesn't need to shutdown at this point
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	Shutdown();
#endif
}

void Asura_Corona_Handler::Initialise(void)
{
	// TODO - The pool probably won't work properly with streaming, needs changing.
	m_pxCoronaPool = new Asura_TrackingHandlePool<Asura_Corona>(16);
}

void Asura_Corona_Handler::Reset(void)
{
	m_pxCoronaPool->Reset();
}

void Asura_Corona_Handler::Shutdown(void)
{
	delete m_pxCoronaPool;
	m_pxCoronaPool = NULL;
}

Asura_Handle Asura_Corona_Handler::CreateCorona(void)
{
	/// Kevin Lea - 27/01/2003 11:16:49 - pool Allocate will return ASURA_INVALID_HANDLE
	/// for us if memory operation fails.
	Asura_Handle iHandle = m_pxCoronaPool->Allocate();

	if (iHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Corona* pxCorona = m_pxCoronaPool->GetPointer(iHandle);

		if (pxCorona)
		{
			Asura_Mem::Set(pxCorona, 0, sizeof(Asura_Corona));
			#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
			pxCorona->InvalidateModuleList();
			pxCorona->m_iLastKnownModule = -1;
			#endif
			pxCorona->m_fVolumetricDepthScale = Asura_Corona_System::GetDefaultVolumetricDepthScale();
		}
	}

	return iHandle;
}

Asura_Handle Asura_Corona_Handler::CopyCorona( Asura_Handle iSrcHandle )
{
	Asura_Handle iDestHandle = m_pxCoronaPool->Allocate();
	if( iDestHandle != ASURA_INVALID_HANDLE )
	{
		Asura_Corona* pxDestCorona = m_pxCoronaPool->GetPointer(iDestHandle);
		Asura_Corona* pxSrcCorona  = m_pxCoronaPool->GetPointer(iSrcHandle);

		if( pxDestCorona && pxSrcCorona )
		{
			pxDestCorona->m_xPosition	  = pxSrcCorona->m_xPosition;
			pxDestCorona->m_xDirection	  = pxSrcCorona->m_xDirection;

			pxDestCorona->m_fRadius		  = pxSrcCorona->m_fRadius;
			pxDestCorona->m_iTextureIndex = pxSrcCorona->m_iTextureIndex;

			pxDestCorona->m_fVolumetricDepthScale	= pxSrcCorona->m_fVolumetricDepthScale;

			pxDestCorona->m_fRed		  = pxSrcCorona->m_fRed;
			pxDestCorona->m_fGreen		  = pxSrcCorona->m_fGreen;
			pxDestCorona->m_fBlue		  = pxSrcCorona->m_fBlue;
			
			pxDestCorona->m_uFlags		  = pxSrcCorona->m_uFlags;
#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
			pxDestCorona->CalculateModuleList();
#endif
		}
	}

	return iDestHandle;
}

void Asura_Corona_Handler::DestroyCorona(Asura_Handle& iHandle)
{
	// (MH - 15.04.2004) iHandle is now passed by reference and automatically 
	// set to ASURA_INVALID_HANDLE. Also checking it's not invalid first.
	if (iHandle != ASURA_INVALID_HANDLE)
	{
		m_pxCoronaPool->Deallocate(iHandle);
		iHandle = ASURA_INVALID_HANDLE;
	}
}


void Asura_Corona_Handler::Render(void)
{
	if(!dbg_bCoronaRenderingEnabled) return;

	// TODO - once we're going through all handlers, it may be possible to move the start and end out to the corona system's render function.
	Asura_Corona_System::Platform_RenderStart();

	m_pxCoronaPool->Traverse(Asura_Corona_System::RenderCorona);

	Asura_Corona_System::Platform_RenderEnd();
}



/////////////////////////////////////////////////////////////////////////////////////////////
/*--------------**
** Asura_Corona **
**--------------*/
#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
bool Asura_Corona::IsInVisibleModule()
{
	#if defined(ASURA_USE_STREAMING)
	if (m_xModuleList.IsValid())
	{
		// MGH 30/5/2006: If the module list is empty then it may just be that
		// the zones were unloaded when we did the module test.
		if (m_xModuleList.IsEmpty())
		{
			// Do the test again, but only if we have loaded a zone since the last test.
			if (Asura_Streaming_System::GetFrameNumberLastZoneLoad() > m_uFrameNumberCalculatedModules)
			{
				CalculateModuleList();
			}
		}
		return TestModuleVisibility();
	}
	else
	{
		// Now, do the slow culling test since our module visibility data is invalid.
		Asura_Bounding_Box xBB;
		xBB.SetToPoint(m_xPosition);
		return Asura_Module_System::DoesBBIntersectVisibleModule(xBB, &m_iLastKnownModule, &m_uLastKnownZone);
	}
	#else
	if (m_ausModuleList[0] != USHRT_MAX)
	{
		// Module visibility data is valid, do quick rejection test.
		return TestModuleVisibility();
	}
	else
	{
		// Now, do the slow culling test since our module visibility data is invalid.
		Asura_Bounding_Box xBB;
		xBB.SetToPoint(m_xPosition);
		return Asura_Module_System::DoesBBIntersectVisibleModule(xBB, &m_iLastKnownModule);
	}
	#endif //defined(ASURA_USE_STREAMING)
}


bool Asura_Corona::TestModuleVisibility()
{
	Asura_Assert(Asura_Module_System::GetNumberOfModules() < USHRT_MAX, "Assumption in clipping code about number of modules is incorrect");

	#if defined(ASURA_USE_STREAMING)
	return m_xModuleList.IsVisible();
	#else
	bool bAccept = false;
	// Do some clipping here, based on module visibility.
	for (int i=0;i<NUM_MODULES_TO_STORE_FOR_CLIPPING;i++)
	{
		if (m_ausModuleList[i] == USHRT_MAX)
			break;
		else if (Asura_Module_System::IsModuleVisible(m_ausModuleList[i]))
		{
			bAccept = true;
			break;
		}
	}
	return bAccept;
	#endif //defined(ASURA_USE_STREAMING)
}

void Asura_Corona::CalculateModuleList()
{

	Asura_Bounding_Box xBB;
	xBB.SetToPoint(m_xPosition);
	
	#if defined(ASURA_USE_STREAMING)
	m_xModuleList.CalculateModuleList(xBB);
	m_uFrameNumberCalculatedModules = Asura_Timers::GetFrameNumber();
	#else
	Asura_Module_System::TestModuleTree(xBB, Asura_Module_System::CullingListModuleCallback);
	Asura_Module_System::FillModuleList(m_ausModuleList);
	#endif

}
void Asura_Corona::InvalidateModuleList()
{
	#if defined(ASURA_USE_STREAMING)
	m_xModuleList.MakeInvalid();
	#else
	m_ausModuleList[0] = USHRT_MAX;
	#endif
}

#endif //#if !defined (ASURA_USE_NEW_VISIBILITY_SYSTEM )
