/**
*
* \file Asura_Decal_System.h
*
* Created by Kevin Lea - 11/11/2002 12:50:52
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/
#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_DECAL_SYSTEM_H__
#define __ASURA_DECAL_SYSTEM_H__

// Includes
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"

// Forward declarations
class Asura_Decal_Template;
class Asura_RGBA;

////////////////////////
// Asura_Decal_Vertex //
////////////////////////
class Asura_Decal_Vertex
{
public:
	Asura_Vector_3	m_xPosition;
	float			m_fU;
	float			m_fV;
};

////////////////////////////////
// Asura_Platform_DecalBuffer //
////////////////////////////////
// Platform specific decal buffers inherit off this base class.
// These buffers form a linked list, each pointing to the next element in the list.
class Asura_DecalBuffer
{
protected:
	Asura_DecalBuffer() {}

public:
	// Derived classes should delete the next element in the linked list.
	virtual ~Asura_DecalBuffer() {}
};

/////////////////////////
// Asura_Decal_Handler //
/////////////////////////
// Decal handler which keeps track of all the decal buffers in a zone.
// It has a pointer to the first buffer in the linked list.
class Asura_Decal_Handler
{
public:
	Asura_Decal_Handler()
		: m_pxFirstDecalBuffer(0) 
	{
	}
	
	virtual ~Asura_Decal_Handler()
	{
		Reset(); // Delete the linked list of buffers.
	}

	void Reset()
	{
		// Delete the linked list.
		delete m_pxFirstDecalBuffer;
		m_pxFirstDecalBuffer = 0;
	}

	// Platform functions
	virtual void Platform_Render();
	void Platform_AddStrip( const Asura_Decal_Vertex* pxStrip,
                            u_int uNumVertices,
                            const Asura_RGBA& xDiffuseRGBA,
                            const Asura_Decal_Template* pxDecalTemplate );

protected:
	Asura_DecalBuffer* m_pxFirstDecalBuffer; // Head of linked list.
};


////////////////////////
// Asura_Decal_System //
////////////////////////
class Asura_Decal_System
{
public:
	static void Initialise();
	static void Reset();			// Reset both static and runtime decals.
	static void ResetBeforeLoad();	// Reset only runtime decals.
	static void Shutdown();

	// Loops over all the zones rendering each in turn.
	static void Render();

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
	static void RenderStaticDecals();
#endif

	// Set the handler for the correct zone before calling these functions.
	static void AddDecal(const Asura_Vector_3& xPosition, const Asura_Vector_3& xSurfaceNormal, const u_int uDecalTemplateID, float fScale = 1.0f, float fRotation = 0.0f);
	
	// Method that allows you to directly set orientation.
	static void AddDecal(const Asura_Vector_3& xPosition, const Asura_Vector_3& xSurfaceNormal,
						 const Asura_Vector_3& xDecalTopDirection, const u_int uDecalTemplateID,
						 float fScale = 1.0f, float fRotation = 0.0f);
	
	static inline const Asura_Decal_Template* GetDecalTemplate( u_int uTemplateID ) { return Project_GetDecalTemplate( uTemplateID ); }

private:
	static void AddDecal(const Asura_Vector_3& xPosition, const Asura_Vector_3& xSurfaceNormal, const Asura_Decal_Template* pxDecalTemplate, const Asura_Matrix_3x3& xOrient, float fScale );

	static u_int ClipPolyToPlane(Asura_Decal_Vertex* pxInVertices, u_int uNumberOfVertices, Asura_Decal_Vertex* pxOutVertices, Asura_Vector_3& xPlaneNormal, float fPlaneConstant);

	static void Project_SetupTemplates();

	static const class Asura_Decal_Template* Project_GetDecalTemplate(u_int uDecalTemplateID);

	static	void					ResetHandlerCallback(Asura_Hash_ID uHash, Asura_Decal_Handler* pxHandler);

	// Platform functions
	static void Platform_Initialise( void );
	static void Platform_Reset( void );
	static void Platform_Shutdown( void );

#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
	static void Platform_RenderStaticDecals( void );
#endif

public:
	
	static void CalculateTangentAndBinormal( const Asura_Vector_3* pxPositions, const Asura_Vector_2* pxTextureCoords, Asura_Vector_3& rxTangent, Asura_Vector_3& rxBinormal );

#ifdef ASURA_TARGET_SM3
	static const class Asura_Decal_Material* Project_GetDecalMaterial(u_int uDecalMaterialID);
	
	static Asura_Vector_3 s_xCurrentDecalNormal;
	static Asura_Vector_3 &GetCurrentDecalNormal(void) {return s_xCurrentDecalNormal;};
#endif

#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	// Handler functions...
	static  Asura_Decal_Handler*	CreateHandler(const Asura_Hash_ID uHash);
	static  bool					DestroyHandler(const Asura_Hash_ID uHash);

	static  void                    CreateAndSetHandler(const Asura_Hash_ID uHash)    { s_pxCurrentHandler = CreateHandler(uHash); }
	static  void					SetCurrentHandler(Asura_Decal_Handler* pxHandler) { s_pxCurrentHandler = pxHandler; }
	static	void					SetCurrentHandler(const Asura_Hash_ID uHash)      { s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static  Asura_Decal_Handler&	GetCurrentHandler()								  { return (*s_pxCurrentHandler); }
	static  Asura_Decal_Handler*	GetHandler(const Asura_Hash_ID uHash)			  { return s_xHandlerTree.GetDataByHash(uHash); }
#else
	// Handler functions...
	static	Asura_Decal_Handler&	GetCurrentHandler()	{ return (*s_pxHandler); }
#endif

#ifdef ASURA_TOOLS
#ifdef ASURA_USE_STATIC_DECAL_SYSTEM
	static void BeginStaticDecalCreation(bool bAllowBackFaceProjection = true);
	static void EndStaticDecalCreation();
	static void AddDecal(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrient, float fScale, const u_int uDecalTemplateID);
	static bool s_bCreatingStaticDecals;
	static bool s_bAllowBackFaceProjection;
#endif
#endif

private:
#if defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
	static	Asura_Decal_Handler* s_pxCurrentHandler; // Currently active handler.
	static	Asura_TemplateHashTree<Asura_Decal_Handler> s_xHandlerTree; // Tree of all of the handlers available.
#else	
	static	Asura_Decal_Handler* s_pxHandler; // Single static-instance of a zone.
#endif
};

#endif//__ASURA_DECAL_SYSTEM_H__
