/**
*
* \file Asura_SFX_GlassSmash.cpp
*
* Created by Andrew Weinkove - 23/07/2002
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "SpecialFX/Asura_SFX_GlassSmash.h"

#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Glass/Asura_ClientEntity_Glass.h"
#include "SpecialFX/Asura_Shard_System.h"

// First an overview of the method.aw.

// We split the glass into a uniform grid, the grid dimensions depend on the size of 
// the pane we are breaking, subject to minimums and maximums.

// We then move the grid lines so that the grid rectangles near to the point of damage (or the 
// centre in the case of blast damage) are smaller than ones further away.

// We now have a grid overlaid onto our glass. Every point where the grid lines cross 
// is one of our vertices. Now we move these vertices randomly, but we must make sure
// to maintain the integrity of the shape. Therefore, we cannot move any vertex that 
// is on the edge of the glass. We must also constrain the movement of the vertices so
// that they can only move half (or less) of the distance towards the next grid line in that direction.
// This randomises the shapes without creating holes.

// Now we split these quads into triangles, randomly deciding which direction to do this in.

// We turn the triangles into shards and give each vertex a shove (velocity) according to 
// their distance from the impact and the power of the impact. We may also give the shards a 
// delay time, based on their distance from the epicentre.

// Now we watch!


// This is inversed, it represents a grid square size of 0.5 metres.
#define ASURA_SFX_GLASSSMASH_AVERAGE_INVERSE_GRID_SIZE 2.0f	
// This is the minimum number of grid lines. 3 is the working minimum. You may want to increase it for aesthetic purposes though.
#define ASURA_SFX_GLASSSMASH_MIN_GRID_NO 3
// This confines the grid size. A 5x5 grid creates 25 quads, which translates to 50 shards.
#define ASURA_SFX_GLASSSMASH_MAX_GRID_NO_SHARDS 50 
// This affects the way grid lines are 'attracted' to the epicentre. I am using this to dampen the effect.
#define ASURA_SFX_GLASSSMASH_FOCUS_CONSTANT 0.5f
// This must be between 0 and 0.5 to maintain integrity of the glass. I keep it a bit below 0.5 to avoid very thin triangles.
#define ASURA_SFX_GLASSSMASH_VERTEX_RANDOM_CONSTANT 0.45f

// This multiplies the velocities.
const float Asura_SFX_GlassSmash::s_fVelocityMultiplier = 12.5f;
// This is the time to spend alpha'ing out the shards.
const float Asura_SFX_GlassSmash::s_fAlphaOutTime = 0.75f;

void Asura_SFX_GlassSmash::Create(const Asura_ClientEntity_Glass& xGlass, const float fPower, const Asura_Vector_2& xHitCoordinates, const Asura_Vector_3& xDirection)
{
	Create(xGlass, fPower, xHitCoordinates, xDirection, false);
}

void Asura_SFX_GlassSmash::Create(const Asura_ClientEntity_Glass& xGlass, const float fPower, const bool bSmashInPositiveZDir)
{
	Asura_Vector_3 xDirection = xGlass.GetBlastDirection(bSmashInPositiveZDir);
	Create(xGlass, fPower, Asura_Vector_2(0.5f, 0.5f), xDirection, true);
}

void Asura_SFX_GlassSmash::Create(const Asura_ClientEntity_Glass& xGlass, const float fPower, const Asura_Vector_2& xHitCoordinates, const Asura_Vector_3& xDirection, bool bBlast)
{
	Asura_Assert(ASURA_SFX_GLASSSMASH_MIN_GRID_NO * ASURA_SFX_GLASSSMASH_MIN_GRID_NO * 2 < ASURA_SFX_GLASSSMASH_MAX_GRID_NO_SHARDS, "Constants set so that glass will never work.");

	// Calculate grid dimensions.
	const Asura_WorldSpace_Rectangle& xPane = xGlass.m_xPane;
	float fAvgInvGridSquareSize = ASURA_SFX_GLASSSMASH_AVERAGE_INVERSE_GRID_SIZE;
	bool bCreatedGrid = false;
	int iGridSizeX = ASURA_SFX_GLASSSMASH_MIN_GRID_NO, iGridSizeY = ASURA_SFX_GLASSSMASH_MIN_GRID_NO;

	while (!bCreatedGrid)
	{
		iGridSizeX = int(xPane.Dimensions.x * fAvgInvGridSquareSize);
		iGridSizeY = int(xPane.Dimensions.y * fAvgInvGridSquareSize);
		if (iGridSizeX < ASURA_SFX_GLASSSMASH_MIN_GRID_NO)
			iGridSizeX = ASURA_SFX_GLASSSMASH_MIN_GRID_NO;
		if (iGridSizeY < ASURA_SFX_GLASSSMASH_MIN_GRID_NO)
			iGridSizeY = ASURA_SFX_GLASSSMASH_MIN_GRID_NO;
		if (iGridSizeX * iGridSizeY * 2 > ASURA_SFX_GLASSSMASH_MAX_GRID_NO_SHARDS)
		{
			// Grid too big. Shrink it by a quarter.
			fAvgInvGridSquareSize *= 0.75;
		}
		else
		{
			bCreatedGrid = true;
		}
	}

	// Create the shards now - just in case we have to exit early.
	int iNumberOfShards = iGridSizeX * iGridSizeY * 2;
	Asura_Shard **ppxShards = new Asura_Shard*[iNumberOfShards];

	if (!ppxShards)
	{
		return;
	}

	// Calculate grid line positions.
	// Would allocating here cause fragmentation? If so, then allocate a buffer.
	float *pfGridXPositions = new float[iGridSizeX + 1];
	float *pfGridYPositions = new float[iGridSizeY + 1];

	float fInvGridSize = 1.0f / float(iGridSizeX);// Precalculated for speed.
	for (int x=0;x<iGridSizeX + 1;x++)
	{
		pfGridXPositions[x] = fInvGridSize * x;
	}
	pfGridXPositions[iGridSizeX] = 1.0f;	// Make sure of this.

	fInvGridSize = 1.0f / float(iGridSizeY);// Precalculated for speed.
	for (int y=0;y<iGridSizeY + 1;y++)
	{
		pfGridYPositions[y] = fInvGridSize * y;
	}
	pfGridYPositions[iGridSizeY] = 1.0f;	// Make sure of this.

	// Now we alter the grid lines to concentrate on the epicentre.
	// We will use the following formula.

	// p += (e - p) * (1 - |e - p|) * a.
	//
	// where	p = Position of line.
	//			e = Position of epicentre.
	//			a = Coefficient to dampen (or amplify) movement.

	for (int x=1;x<iGridSizeX;x++)
	{
		pfGridXPositions[x] += (xHitCoordinates.x - pfGridXPositions[x]) * (1.0f - Asura_Maths::Abs(xHitCoordinates.x - pfGridXPositions[x])) * ASURA_SFX_GLASSSMASH_FOCUS_CONSTANT;
	}
	for (int y=1;y<iGridSizeY;y++)
	{
		pfGridYPositions[y] += (xHitCoordinates.y - pfGridYPositions[y]) * (1.0f - Asura_Maths::Abs(xHitCoordinates.y - pfGridYPositions[y])) * ASURA_SFX_GLASSSMASH_FOCUS_CONSTANT;
	}

	// Now we create and randomise our vertices. In 2D for now.
	Asura_Vector_2 *pxVertices;
	pxVertices = new Asura_Vector_2[(iGridSizeX + 1) * (iGridSizeY + 1)];

	for (int x=0;x<iGridSizeX + 1;x++)
	{
		for (int y=0;y<iGridSizeY + 1;y++)
		{
			int iIndex = ((iGridSizeX + 1) * y) + x;
			pxVertices[iIndex] = Asura_Vector_2(pfGridXPositions[x], pfGridYPositions[y]);
			if (x != 0 &&
				x != iGridSizeX)
			{
				float fMinX, fMaxX;
				fMinX = (pfGridXPositions[x-1] - pfGridXPositions[x]) * ASURA_SFX_GLASSSMASH_VERTEX_RANDOM_CONSTANT;
				fMaxX = (pfGridXPositions[x+1] - pfGridXPositions[x]) * ASURA_SFX_GLASSSMASH_VERTEX_RANDOM_CONSTANT;
				Asura_Assert(fMaxX > fMinX, "");
				pxVertices[iIndex].x += fMinX + (Asura_Random::GetFloat() * (fMaxX - fMinX));
			}
			if (y != 0 &&
				y != iGridSizeY)
			{
				float fMinY, fMaxY;
				fMinY = (pfGridYPositions[y-1] - pfGridYPositions[y]) * ASURA_SFX_GLASSSMASH_VERTEX_RANDOM_CONSTANT;
				fMaxY = (pfGridYPositions[y+1] - pfGridYPositions[y]) * ASURA_SFX_GLASSSMASH_VERTEX_RANDOM_CONSTANT;
				Asura_Assert(fMaxY > fMinY, "");
				pxVertices[iIndex ].y += fMinY + (Asura_Random::GetFloat() * (fMaxY - fMinY));
			}
		}
	}

	// Now we create the shards. Randomising the direction of the split.
	for (int x=0;x<iGridSizeX;x++)
	{
		for (int y=0;y<iGridSizeY;y++)
		{
			bool bSplitAlong0And3 = (Asura_Random::GetIntU() % 2) != 0;
			Asura_Vector_2 axQuad[4];
			axQuad[0] = pxVertices[((iGridSizeX + 1) * y) + x];			// Top left.
			axQuad[1] = pxVertices[((iGridSizeX + 1) * (y + 1)) + x];	// Bottom left.
			axQuad[2] = pxVertices[((iGridSizeX + 1) * y) + (x + 1)];	// Top Right.
			axQuad[3] = pxVertices[((iGridSizeX + 1) * (y + 1)) + (x + 1)];	// Bottom right.

			for (int t=0;t<2;t++)
			{
				int i = (((x * iGridSizeY) + y) * 2) + t;	// Precalculate for speed and readability.
				ppxShards[i] = Asura_Shard_System::AllocateShard();
				Asura_Assert(ppxShards[i], "Failed to allocate shard, this shouldn't happen because shards use resizing mempools");

				ppxShards[i]->m_fLifeTimeRemaining = 5.0f;
				ppxShards[i]->m_fTimeToStayStill = 0.0f;
				int aiIndex[3] = {0, 3, 1};
				if (bSplitAlong0And3)
				{
					if (t != 0)
					{
						aiIndex[1] = 2;
						aiIndex[2] = 3;
					}
				}
				else
				{
					if (t == 0)
						aiIndex[1] = 2;
					else
						aiIndex[0] = 2;
				}
				float fWaitTime = Asura_Maths::LargeFloat;
				for (int v=0; v<3; v++)
				{
					ppxShards[i]->m_afU[v] = xGlass.m_afU[0] + (axQuad[aiIndex[v]].x * (xGlass.m_afU[3] - xGlass.m_afU[0]));
					ppxShards[i]->m_afV[v] = xGlass.m_afV[0] + (axQuad[aiIndex[v]].y * (xGlass.m_afV[3] - xGlass.m_afV[0]));

					ppxShards[i]->m_axVertices[v] = xPane.CornerPosition + (Asura_Vector_3(axQuad[aiIndex[v]].x * xPane.Dimensions.x, axQuad[aiIndex[v]].y * xPane.Dimensions.y, 0.0f) * xPane.Orientation);
					if (bBlast)
					{
						ppxShards[i]->m_axVelocities[v] = CalculateInitialVelocityFromBlast(axQuad[aiIndex[v]], fPower, xHitCoordinates, xDirection) * xPane.Orientation;
					}
					else
					{
						Asura_Matrix_3x3 xLocalMat = xPane.Orientation;
						xLocalMat.Transpose();
						ppxShards[i]->m_axVelocities[v] = CalculateInitialVelocityFromShot(axQuad[aiIndex[v]], fPower, xHitCoordinates, xDirection * xLocalMat, &fWaitTime) * xPane.Orientation;
					}
				}
				if (fWaitTime != 0.0f &&
					fWaitTime != Asura_Maths::LargeFloat)
				{
					ppxShards[i]->m_fTimeToStayStill = fWaitTime + ((Asura_Random::GetFloat() * 0.1f) - 0.05f);
				}
				ppxShards[i]->CalculateEdgeLengths();
			}
		}
	}
	Asura_Shard_Effect *pxEffect = Asura_Shard_System::CreateShardEffect();
	unsigned uFlags = 0;
	if (xGlass.m_bUseEnvironmentMap)
		uFlags |= SHARD_EFFECT_FLAGS_USE_ENVIRONMENT_MAP;
	pxEffect->Initialise(ppxShards, iNumberOfShards, s_fAlphaOutTime, uFlags, xGlass.m_uColour, xGlass.m_iTextureIndex);

	delete[] ppxShards;
	delete[] pfGridXPositions;
	delete[] pfGridYPositions;
	delete[] pxVertices;
}

Asura_Vector_3 Asura_SFX_GlassSmash::CalculateInitialVelocityFromBlast(const Asura_Vector_2& xVertexPosition, const float fPower, const Asura_Vector_2& xHitCoordinates, const Asura_Vector_3& xDirection)
{
	Asura_Assert(xHitCoordinates.y == 0.5f, "");
	Asura_Assert(xHitCoordinates.x == 0.5f, "");
	Asura_Assert(xVertexPosition.x >= 0.0f, "");
	Asura_Assert(xVertexPosition.x <= 1.0f, "");
	Asura_Assert(xVertexPosition.y >= 0.0f, "");
	Asura_Assert(xVertexPosition.y <= 1.0f, "");
	float fMagnitude = (-2.0f * (xVertexPosition.x * xVertexPosition.x)) + (2.0f * xVertexPosition.x);
	fMagnitude += (-2.0f * (xVertexPosition.y * xVertexPosition.y)) + (2.0f * xVertexPosition.y);
	fMagnitude *= fPower * s_fVelocityMultiplier;
	
	fMagnitude *= 1.0f - ((Asura_Random::GetFloat() * 0.5f)- 0.25f);

	Asura_Vector_3 xRet(0.0f, 0.0f, 0.0f);
	xRet.x = xVertexPosition.x - xHitCoordinates.x;
	xRet.y = xVertexPosition.y - xHitCoordinates.y;
	xRet.z = xDirection.z * ((1.25f - xRet.x) - xRet.y);
	xRet.Normalise();
	// Randomise direction?
	xRet *= fMagnitude;
	return xRet;
}

Asura_Vector_3 Asura_SFX_GlassSmash::CalculateInitialVelocityFromShot(const Asura_Vector_2& xVertexPosition, const float fPower, const Asura_Vector_2& xHitCoordinates, const Asura_Vector_3& xDirection, float *pfWaitTime)
{
	Asura_Assert(xVertexPosition.x >= 0.0f, "");
	Asura_Assert(xVertexPosition.x <= 1.0f, "");
	Asura_Assert(xVertexPosition.y >= 0.0f, "");
	Asura_Assert(xVertexPosition.y <= 1.0f, "");
	Asura_Assert(xHitCoordinates.x >= 0.0f, "");
	Asura_Assert(xHitCoordinates.x <= 1.0f, "");
	Asura_Assert(xHitCoordinates.y >= 0.0f, "");
	Asura_Assert(xHitCoordinates.y <= 1.0f, "");

	float fDistFromEpicentre2 = (xHitCoordinates - xVertexPosition) * (xHitCoordinates - xVertexPosition);
	float fMagnitude = (-222.222f * fDistFromEpicentre2) + 5;
	if (fMagnitude > 0.25f)
	{
		// This vertex is close to the shot.
		Asura_Assert(fDistFromEpicentre2 < 0.03f, "Hmm.. My maths went wrong somewhere");
		Asura_Vector_3 xRet = xDirection;
		fMagnitude *= 1.0f - ((Asura_Random::GetFloat() * 0.5f)- 0.25f);
		xRet.Normalise();
		xRet *= (fMagnitude * fPower * s_fVelocityMultiplier);
		*pfWaitTime = 0.0f;
		return xRet;
	}
	else
	{
		// This vertex is far from the shot.
		Asura_Assert(fDistFromEpicentre2 >= 0.0f, "");
		float fDistFromEpicentre = Asura_Maths::Sqrt(fDistFromEpicentre2);
		Asura_Assert(fDistFromEpicentre > 0.1, "Hmm.. My maths went wrong somewhere");
		fMagnitude = 0.357f - (0.714f * fDistFromEpicentre);
		fMagnitude *= 1.0f - ((Asura_Random::GetFloat() * 0.5f)- 0.25f);
		if (fMagnitude < 0.0f)
			fMagnitude = 0.0f;
		float fWaitTime = fDistFromEpicentre;
		if (*pfWaitTime > fWaitTime)
			*pfWaitTime = fWaitTime;

		Asura_Vector_3 xRet(0.0f, 0.0f, 0.0f);
		xRet.z = xDirection.z > 0.0f ? 1.0f : -1.0f;
		xRet.Normalise();
		xRet *= (fMagnitude * fPower * s_fVelocityMultiplier);
		return xRet;
	}
}
