//
// Asura_SimplexNoise_System.cpp
//
// Created by KOR - 16/10/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"


// This file's header.
#include "SpecialFX/Asura_SimplexNoise_System.h"

// Asura
#include "Core/Asura_Maths.h"

#define ASURA_SIMPLEX_NOISE_USE_LOOKUPTABLES (1)

// Static data
#ifdef ASURA_SIMPLEX_NOISE_USE_LOOKUPTABLES
//const u_int uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE = 256;
const u_int uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE = 16;
int	g_aiPermutation[uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE * 2];

// The recommended set of gradients is the midpoints of each of the 12 edges of a cube centred on the origin
const u_int uASURA_SIMPLEX_GRADIENT_TABLE_SIZE = 12;
Asura_Vector_3 g_axGradients[uASURA_SIMPLEX_GRADIENT_TABLE_SIZE] =
{
	Asura_Vector_3( 0.f, 1.f, 1.f),
	Asura_Vector_3( 0.f, 1.f,-1.f),
	Asura_Vector_3( 0.f,-1.f, 1.f),
	Asura_Vector_3( 0.f,-1.f,-1.f), 
	Asura_Vector_3( 1.f, 0.f, 1.f),
	Asura_Vector_3( 1.f, 0.f,-1.f),
	Asura_Vector_3(-1.f, 0.f, 1.f),
	Asura_Vector_3(-1.f, 0.f,-1.f),
	Asura_Vector_3( 1.f, 1.f, 0.f),
	Asura_Vector_3( 1.f,-1.f, 0.f),
	Asura_Vector_3(-1.f, 1.f, 0.f),
	Asura_Vector_3(-1.f,-1.f, 0.f),
};
#endif

bool Asura_SimplexNoise_System::s_bNoiseInitialised = false;

// Helper functions
#ifndef ASURA_SIMPLEX_NOISE_USE_LOOKUPTABLES
static u_char g_aucBitPatterns[] =
{
	0x15, 0x38, 0x32, 0x2c, 0x0d, 0x13, 0x07, 0x2a,
};

inline int GetBit(int uValue, u_int uBit)
{
	return (uValue >> uBit) & 0x01;
}

inline u_char GetBitPattern(int iX, int iY, int iZ, u_int uBit)
{
	const u_int uPatternIndex = (GetBit(iX, uBit) << 2) + (GetBit(iY, uBit) << 1) + GetBit(iZ, uBit);
	return g_aucBitPatterns[uPatternIndex];
}

inline u_char GetGradientBits(int iX, int iY, int iZ)
{
	return GetBitPattern(iX, iY, iZ, 0)
		 + GetBitPattern(iY, iZ, iX, 1)
		 + GetBitPattern(iZ, iX, iY, 2)
		 + GetBitPattern(iX, iY, iZ, 3)
		 + GetBitPattern(iY, iZ, iX, 4)
		 + GetBitPattern(iZ, iX, iY, 5)
		 + GetBitPattern(iX, iY, iZ, 6)
		 + GetBitPattern(iY, iZ, iX, 7);
}
#endif//ndef ASURA_SIMPLEX_NOISE_USE_LOOKUPTABLES

void Asura_SimplexNoise_System::Initialise()
{
	Asura_Assert(!s_bNoiseInitialised, "Noise system already initialised");
	if (s_bNoiseInitialised) return;
	s_bNoiseInitialised = true;
#ifdef ASURA_SIMPLEX_NOISE_USE_LOOKUPTABLES
	// Generate permutation table
	for (u_int u = 0 ; u < uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE ; ++u)
	{
		g_aiPermutation[u] = u;
	}
	int iCounter = uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE;
	while (--iCounter)
	{
		const int iBackup = g_aiPermutation[iCounter];
		// Swap with a random entry
		const int iRandom = Asura_Random::GetIntU(0, uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE - 1);
		g_aiPermutation[iCounter] = g_aiPermutation[iRandom];
		g_aiPermutation[iRandom] = iBackup;
	}
	// Extend tables
	for (u_int u = 0 ; u < uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE; ++u)
	{
		g_aiPermutation[uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE + u] = g_aiPermutation[u];
	}
#endif
}

float Asura_SimplexNoise_System::GetNoise(const Asura_Vector_2& xValue)
{
	Asura_Assert(0.f, "2D simplex noise not yet implemented");
	return 0.f;
}

float GetContributionFromVertex(const Asura_Vector_3& xOffset, const int iX, const int iY, const int iZ)
{
	const float fT = 0.6f - xOffset.MagnitudeSquared();
	if(fT > 0.f)
	{
		const float fTSq = fT * fT;
		const float fScale = 8.f * fTSq * fTSq;
#ifdef ASURA_SIMPLEX_NOISE_USE_LOOKUPTABLES
		const int iXMod = iX & (uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE - 1);
		const int iYMod = iY & (uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE - 1);
		const int iZMod = iZ & (uASURA_SIMPLEX_PERMUTATION_TABLE_SIZE - 1);
		const int iGradientIndex = g_aiPermutation[iXMod + g_aiPermutation[iYMod + g_aiPermutation[iZMod] ] ];
		const Asura_Vector_3& xGradient = g_axGradients[iGradientIndex % uASURA_SIMPLEX_GRADIENT_TABLE_SIZE];
		return fScale * (xGradient * xOffset);
#else
		const u_int uGradientBits = GetGradientBits(iX, iY, iZ);
		const u_int uBit5 = GetBit(uGradientBits, 5);
		const u_int uBit4 = GetBit(uGradientBits, 4);
		const u_int uBit3 = GetBit(uGradientBits, 3);
		const u_int uBit2 = GetBit(uGradientBits, 2);
		const u_int uBits10 = (uGradientBits & 3);
		float fP = (uBits10 == 1) ? xOffset.y : ((uBits10 == 2) ? xOffset.z : xOffset.x);
		float fQ = (uBits10 == 1) ? xOffset.z : ((uBits10 == 2) ? xOffset.x : xOffset.y);
		float fR = (uBits10 == 1) ? xOffset.x : ((uBits10 == 2) ? xOffset.y : xOffset.z);
		fP = (uBit3 == uBit5) ? -fP : fP;
		fQ = (uBit4 == uBit5) ? -fQ : fQ;
		fR = ( (uBit3 ^ uBit4) == uBit5) ? fR : -fR;
		return fScale * (fP + (uBits10 == 0) ? (fQ + fR) : ( (uBit2 == 0) ? fQ : fR));
#endif
	}
	return 0.f;
}

float Asura_SimplexNoise_System::GetNoise(const Asura_Vector_3& xValue)
{
	float fSum = 0.f;
	// Convert to skewed space
	const float fSkewFactor = 1.f / 3.f;
	const float fSkewOffset = (xValue.x + xValue.y + xValue.z) * fSkewFactor;

	const int iCellX = Asura_Maths::FloatToInt_Chop(xValue.x + fSkewOffset);
	const int iCellY = Asura_Maths::FloatToInt_Chop(xValue.y + fSkewOffset);
	const int iCellZ = Asura_Maths::FloatToInt_Chop(xValue.z + fSkewOffset);

	const float fUnskewFactor = 1.f / 6.f;
	const Asura_Vector_3 xUnskewFactor(fUnskewFactor, fUnskewFactor, fUnskewFactor);

	const float fUnskewOffset = static_cast<float>(iCellX + iCellY + iCellZ) * fUnskewFactor;
	const Asura_Vector_3 xUnskewOffset(fUnskewOffset, fUnskewOffset, fUnskewOffset);
	const Asura_Vector_3 xCellCoords(static_cast<float>(iCellX), static_cast<float>(iCellY), static_cast<float>(iCellZ));
	const Asura_Vector_3 xOffset0 = xValue - xCellCoords + xUnskewOffset;
	fSum += GetContributionFromVertex(xOffset0, iCellX, iCellY, iCellZ);

	// Work out the order that we are going to traverse the simplex.
	const u_int uHighest = (xOffset0.x >= xOffset0.z) ? ((xOffset0.x >= xOffset0.y) ? 0 : 1) : ((xOffset0.y >= xOffset0.z) ? 1 : 2);
	const u_int uLowest =  (xOffset0.x <  xOffset0.z) ? ((xOffset0.x <  xOffset0.y) ? 0 : 1) : ((xOffset0.y <  xOffset0.z) ? 1 : 2);

	u_int auSimplexVertex[3] = { 0, 0, 0 };

	// Calculate contribution from vertex 1
	++auSimplexVertex[uHighest];
	const Asura_Vector_3 xVertex1Coords(static_cast<float>(auSimplexVertex[0]), static_cast<float>(auSimplexVertex[1]), static_cast<float>(auSimplexVertex[2]));
	const Asura_Vector_3 xOffset1 = xOffset0 - xVertex1Coords + xUnskewFactor;
	fSum += GetContributionFromVertex(xOffset1, iCellX + auSimplexVertex[0], iCellY + auSimplexVertex[1], iCellZ + auSimplexVertex[2]);

	// Calculate contribution from vertex 2
	++auSimplexVertex[3 - uHighest - uLowest];
	const Asura_Vector_3 xVertex2Coords(static_cast<float>(auSimplexVertex[0]), static_cast<float>(auSimplexVertex[1]), static_cast<float>(auSimplexVertex[2]));
	const Asura_Vector_3 xOffset2 = xOffset0 - xVertex2Coords + xUnskewFactor * 2.f;
	fSum += GetContributionFromVertex(xOffset2, iCellX + auSimplexVertex[0], iCellY + auSimplexVertex[1], iCellZ + auSimplexVertex[2]);

	// Calculate contribution from vertex 3
	++auSimplexVertex[uLowest];
	const Asura_Vector_3 xVertex3Coords(static_cast<float>(auSimplexVertex[0]), static_cast<float>(auSimplexVertex[1]), static_cast<float>(auSimplexVertex[2]));
	const Asura_Vector_3 xOffset3 = xOffset0 - xVertex3Coords + xUnskewFactor * 2.f;
	fSum += GetContributionFromVertex(xOffset3, iCellX + auSimplexVertex[0], iCellY + auSimplexVertex[1], iCellZ + auSimplexVertex[2]);
	
	// Extra scaling factor to get it into the same range as our Perlin noise implementation
	return 2.f * fSum;
}

void Asura_SimplexNoise_System::CalculateNoiseVector (const Asura_Vector_3& xPosition,
													 Asura_Vector_3& xVector)
{
	xVector[0]	= GetNoise(Asura_Vector_2(xPosition[1], xPosition[2]));
	xVector[1]	= GetNoise(Asura_Vector_2(xPosition[2], xPosition[0]));
	xVector[2]	= GetNoise(Asura_Vector_2(xPosition[0], xPosition[1]));
}

void Asura_SimplexNoise_System::CalculateNoiseVector (const Asura_AVector3& xPosition,
													 Asura_AVector3& xVector)
{
	xVector.SetX(GetNoise(Asura_Vector_2(xPosition.GetY(), xPosition.GetZ())));
	xVector.SetY(GetNoise(Asura_Vector_2(xPosition.GetZ(), xPosition.GetX())));
	xVector.SetZ(GetNoise(Asura_Vector_2(xPosition.GetX(), xPosition.GetY())));
}

void Asura_SimplexNoise_System::CalculateNoiseVector (const Asura_Vector_3& xPosition,
													 const float fTime,
													 Asura_Vector_3& xVector)
{
	xVector[0]	= GetNoise(Asura_Vector_3(xPosition[1], xPosition[2], fTime));
	xVector[1]	= GetNoise(Asura_Vector_3(xPosition[2], xPosition[0], fTime));
	xVector[2]	= GetNoise(Asura_Vector_3(xPosition[0], xPosition[1], fTime));
}

void Asura_SimplexNoise_System::CalculateNoiseVector (const Asura_AVector3& xPosition,
													 const float fTime,
													 Asura_AVector3& xVector)
{
	xVector.SetX(GetNoise(Asura_Vector_3(xPosition.GetY(), xPosition.GetZ(), fTime)));
	xVector.SetY(GetNoise(Asura_Vector_3(xPosition.GetZ(), xPosition.GetX(), fTime)));
	xVector.SetZ(GetNoise(Asura_Vector_3(xPosition.GetX(), xPosition.GetY(), fTime)));
}
