//
// Asura_StaticCorona_System.h
//
// Created by Richard Rice - 21/02/2003
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _ASURA_STATICCORONA_SYSTEM_H_
#define _ASURA_STATICCORONA_SYSTEM_H_

#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_NameTree.h"
#ifdef ASURA_TOOLS
#include "Network/Asura_Network.h"
#endif

//Defines for accessing flags in m_uFlags , for use by the tickboxes
//Must match up with entries in the bitfield
#define AsuraStaticCoronaFlag_UseDirectionEffect	 0x00000001
#define AsuraStaticCoronaFlag_DontFadeWithDistance	 0x00000002
#define AsuraStaticCoronaFlag_DontFadeNear			 0x00000004
#define AsuraStaticCoronaFlag_UseModuleLinking       0x00000008


//Value ORed in with the light IDs for coronas that belong to template files
//(to distinguish from those that are just part of the test environment)
#define TOOLS_CORONA_TEMPLATECORONA 0x80000000

//For marking coronas that are to be saved out to the mission .lit file , instead
//of the common .lit file.
#define TOOLS_CORONA_MISSIONCORONA	0x40000000

#define TOOLS_CORONA_IDMASK (~(TOOLS_CORONA_TEMPLATECORONA|TOOLS_CORONA_MISSIONCORONA))

class Asura_StaticCorona
{
public :

	void Initialise();//reset everything to some default values
	void UpdateCorona();//update the corona in the corona system;
	void DeleteCorona();//remove the corona from the main corona system;

	#ifdef ASURA_TOOLS
	//Used after a corona has been created from the template's chunk buffers to adjust positions etc
	//as needed. RAR 27/04/2004
	void ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance);
	#endif

	#ifdef ASURA_TOOLS
	//Returns true if this is part of the active light set.
	bool IsMemberOfCurrentLightSet() const;

	bool IsMissionCorona() const {return (m_uID & TOOLS_CORONA_MISSIONCORONA)!=0;}
	#endif

	const Asura_StaticCorona& operator=(const Asura_StaticCorona& xCorona);


	float m_fRed;
	float m_fGreen;
	float m_fBlue;

	float m_fRadius;

	int   m_iTextureID;

	float m_fVolumetricDepthScale;

	Asura_Vector_3	m_xPosition;
	Asura_Vector_3	m_xDirection;

	union
	{
		u_int m_uFlags;
		struct
		{
			u_int	m_bUseDirectionEffect :1;
			u_int	m_bDontFadeWithDistance :1;
			u_int	m_bDontFadeNear :1;
			u_int   m_bUseModuleLinking:1;
			u_int   m_uSpareBits : 27;


			u_int	m_bNotAllocated : 1;//used to mark invalid entries in corona array
		};
	};

	#ifdef ASURA_TOOLS
	int m_iTextureIDForSave;
	void BuildTextureTree(Asura_NameTree* pxNameTree);
	#endif

	#ifdef ASURA_TOOLS
	u_int m_uID;//unique ID for this corona , so it can be identified by templates etc
	Asura_Guid m_uTemplateGuid; //guid of template that owns this , if any
	int m_iTemplateVersion;

	Asura_String m_xName;

	u_int m_uLightSets;//bitfield indicating which light sets this corona belongs to


	void GenerateID()
	{
		m_uID = s_uNextID++;
	}

	static u_int s_uNextID;
	#endif


private :
	Asura_Handle m_iCoronaHandle;

	friend class Asura_StaticCorona_Handler;
	friend class Asura_StaticCorona_System;
};

class Asura_StaticCorona_Handler
{
public :
	Asura_StaticCorona_Handler()
	{
	// AJG - 27/03/06 - Non-streaming one doesn't need to initialise at this point
#ifdef ASURA_USE_STREAMING
		Initialise();
#endif //#ifdef ASURA_USE_STREAMING
	}
	~Asura_StaticCorona_Handler()
	{
	// AJG - 27/03/06 - Non-streaming one doesn't need to shutdown at this point
#ifdef ASURA_USE_STREAMING
		Shutdown();
#endif //#ifdef ASURA_USE_STREAMING
	}

	void Initialise();
	void Shutdown();
	void Reset();

	int AllocateOnLoad(int iNumberOfCoronas);
	Asura_StaticCorona* GetCorona(int iIndex);

	int GetNumberOfCoronas() { return m_iNumberOfCoronas; }


	#ifdef ASURA_TOOLS
	int AllocateCorona(); //allocate a new corona , returning its index. -1 if it fails
	void DeallocateCorona(int iIndex);

	int GetArraySize() { return m_iArraySize; }
	bool IsValidCorona(int iIndex);

	void BuildTextureTree(Asura_NameTree* pxNameTree);
#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	u_int GetTextureUsage(u_int uTextureHash);
#endif
	int GetCoronaIndexFromID(u_int uID);

	void UpdateAllCoronas();
	#endif

	bool IncreaseAllocation(int iAmount);
private :

	int m_iArraySize;
	int m_iNumberOfCoronas;
	Asura_StaticCorona* m_pxCoronas;
};


class Asura_StaticCorona_System
{
public :
	static void Initialise();
	static void Shutdown();
	static void Reset();

	static int AllocateOnLoad(int iNumberOfCoronas) { return GetCurrentHandler().AllocateOnLoad(iNumberOfCoronas); }
	static Asura_StaticCorona* GetCorona(int iIndex) { return GetCurrentHandler().GetCorona(iIndex); }

	static int GetNumberOfCoronas() { return GetCurrentHandler().GetNumberOfCoronas(); }


	#ifdef ASURA_TOOLS
	static int AllocateCorona() //allocate a new corona , returning its index. -1 if it fails
	{
		return GetCurrentHandler().AllocateCorona();
	}
	static void DeallocateCorona(int iIndex)
	{
		GetCurrentHandler().DeallocateCorona(iIndex);
	}

	static int GetArraySize() { return GetCurrentHandler().GetArraySize(); }
	static bool IsValidCorona(int iIndex) { return GetCurrentHandler().IsValidCorona(iIndex); }

	static void BuildTextureTree(Asura_NameTree* pxNameTree)
	{
		GetCurrentHandler().BuildTextureTree(pxNameTree); 
	}
#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	static u_int GetTextureUsage(u_int uTextureHash)
	{
		return GetCurrentHandler().GetTextureUsage(uTextureHash); 
	}
#endif

	static int GetCoronaIndexFromID(u_int uID)
	{
		return GetCurrentHandler().GetCoronaIndexFromID(uID);
	}

	static void UpdateAllCoronas()
	{
		GetCurrentHandler().UpdateAllCoronas();
	}
	#endif

#if defined(ASURA_USE_STREAMING)
	// Handler functions...
	static	Asura_StaticCorona_Handler*	CreateHandler(const Asura_Hash_ID uHash);
	static	bool						DestroyHandler(const Asura_Hash_ID uHash);

	static	void						CreateAndSetHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = CreateHandler(uHash); }
	static	void						SetCurrentHandler(Asura_StaticCorona_Handler* pxHandler) { s_pxCurrentHandler = pxHandler; }
	static	void						SetCurrentHandler(const Asura_Hash_ID uHash) { s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static	Asura_StaticCorona_Handler&	GetCurrentHandler(void) { return *s_pxCurrentHandler; }
	static	Asura_StaticCorona_Handler*	GetHandler(const Asura_Hash_ID uHash) { return s_xHandlerTree.GetDataByHash(uHash); }
#else
	// Handler functions...
	static	Asura_StaticCorona_Handler&	GetCurrentHandler(void) { return s_xHandler; }
#endif	// defined(ASURA_USE_STREAMING)

private :
	static bool IncreaseAllocation(int iAmount)
	{
		return GetCurrentHandler().IncreaseAllocation(iAmount);
	}

#ifdef ASURA_USE_STREAMING
	static void Reset_Callback(Asura_Hash_ID uHash, class Asura_StaticCorona_Handler* pxData);

	static Asura_StaticCorona_Handler*							s_pxCurrentHandler;		// Currently active handler.
	static Asura_TemplateHashTree<Asura_StaticCorona_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.
#else //#ifdef ASURA_USE_STREAMING
	static Asura_StaticCorona_Handler s_xHandler;				// Single static-instance of a zone.
#endif //#else //#ifdef ASURA_USE_STREAMING
};



#endif //_ASURA_STATICCORONA_SYSTEM_H_
