//
// Asura_PFX_Effect.h
//
// Created by KOR - 15/07/2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __ASURA_PFX_EFFECT_H__
#define __ASURA_PFX_EFFECT_H__

#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Module_Defines.h"
#include "SpecialFX/PFX/Asura_PFX_Enums.h"

#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_ModuleList.h"
#endif //defined(ASURA_USE_STREAMING)

#if defined (ASURA_USE_NEW_VISIBILITY_SYSTEM)
#include "Visibility/Asura_Visibility_System.h"
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////
#define MAX_CAMERAS_FOR_PFX_EFFECT (4)

// For streaming builds, this is the maximum number of zones that are taken into account
// when building up module lists for PFX effects. (Try increasing to 3 or 4 if this isn't enough).
#define MAX_ZONES_FOR_PFX_CLIPPING 2

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////
class Asura_PFX_Particle;
class Asura_PFX_Particle_Template;
class Asura_RGBA;
class Asura_PFX_Source_Template;
class Asura_PFX_Source;
template <typename T> class Asura_Spline;

/// Asura_PFX_Effect //////////////////////////////////////////////////////////////////////////////
// An effect instance. Inherit from this to make a code effect
///////////////////////////////////////////////////////////////////////////////////////////////////

class Asura_PFX_Effect
{
	friend class Asura_PFX_System;
	friend class Asura_PFX_Handler;
public:
	// Constructor and destructor
	Asura_PFX_Effect(u_short uPFXFlags = ASURA_PFX_EFFECT_CAMERA_FLAGS);
	virtual ~Asura_PFX_Effect();
	
	virtual void Render() = 0;
	virtual void Update() = 0;

	// Get methods
	// For simple dynamics
	virtual bool					GetCollisionPlane(Asura_Vector_3* pxPosition, Asura_Vector_3* pxDirection) const { return false; }
	// For point gravity
	virtual bool					GetGravityPosition(Asura_Vector_3* pxPosition) const { return false; }
	// For oriented particles
	virtual const Asura_Matrix_3x3* GetOrientation()	const { return NULL; }
	bool							GetIsHUDEffect()	const { return (m_usPFXFlags & ASURA_PFX_EFFECT_HUD_EFFECT) != 0; }
	bool							IsEntityEffect()	const { return (m_usPFXFlags & ASURA_PFX_EFFECT_ENTITY_EFFECT) != 0; }
	ASURA_PFX_RENDER_TYPE			GetRenderType()		const;
	unsigned int					GetID()				const { return m_uEffectID; }
	ASURA_PFX_EFFECT_STATE			GetState()			const { return m_eState; }
	const Asura_Vector_3&			GetDelta()			const { return m_xDelta; }
	float							GetSpeedScale()		const { return m_fSpeedScale; }
	float							GetSizeScale()		const { return m_fSizeScale; }
	float							GetReleaseRateScale() const {return m_fReleaseRateScale;}
	//get culling bounding box
	const Asura_Bounding_Box&		GetBoundingBox()	const { return m_xBB; }
	//returns bitfield indicating which cameras can see the effect ,
	//updating them if not calculated yet this frame
	u_int							GetCameraFlags(bool bFrustumTest);

	// State queries
	bool IsRunning() const { return GetState() == ASURA_PFX_EFFECT_RUNNING; }
	bool IsStopping() const { return GetState() == ASURA_PFX_EFFECT_STOPPING; }
	bool IsStandingBy() const { return GetState() == ASURA_PFX_EFFECT_STANDBY; }
	bool IsDestroyed() const { return GetState() == ASURA_PFX_EFFECT_DESTROYED; }

	virtual const Asura_PFX_Source*	GetSource(u_int uSourceNum) const { return NULL; }

	// Performs independent culling, etc.
	void RenderParticles( const Asura_PFX_Particle* const pxParticles, u_int uNumParticles, 
		const Asura_PFX_Particle_Template* pxTemplate, const Asura_RGBA& xLightColour, 
		void* pPlatformRenderData );

	// Moves particles by rules defined by source template
	static void UpdateParticles(Asura_PFX_Particle* pxParticles, u_int uNumParticles, const Asura_PFX_Source_Template* pxSourceTemplate, 
					bool bUseCollisionPlane, const Asura_Vector_3& xCollisionPosition, const Asura_Vector_3& xCollisionDirection, 
					bool bUsePointGravity, const Asura_Vector_3& xGravityPosition, float fSizeScale, float fTimeSlice, float fTurbulencePhase,
					const Asura_PFX_Particle_Template* pxPcleTemp, const Asura_Spline<Asura_Vector_3>* pxSpline = NULL );
	void UpdateParticles( Asura_PFX_Particle* pxParticles, u_int uNumParticles,
                          const Asura_PFX_Source_Template* pxSourceTemplate,
						  const Asura_PFX_Particle_Template* pxParticleTemplate,
						  const Asura_Spline<Asura_Vector_3>* pxSpline = NULL) const;

	// Platform implementation of renderer
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
	void Platform_RenderParticles( const Asura_PFX_Particle* const pxParticles, u_int uNumParticles, 
		const Asura_PFX_Particle_Template* pxTemplate, const Asura_RGBA& xLightColour, 
		void* pPlatformRenderData ) const;
#else
	void Platform_RenderParticles( const Asura_PFX_Particle* const pxParticles, u_int uNumParticles, 
		const Asura_PFX_Particle_Template* pxTemplate, const Asura_RGBA& xLightColour ) const;
#endif

	void Destroy() { m_eState = ASURA_PFX_EFFECT_DESTROYED; }

	void SetIgnoreModuleLinking(bool bIgnore);
	void SetSkyboxEffect(bool bSkybox);
	void SetHUDEffect(bool bHUD);
	void SetWaterDistorterEffect(bool bSkybox);
	void SetForceRenderOnMainThread(bool bSet);
	void SetSpeedScale( float fScale ) { m_fSpeedScale = fScale; }
	void SetSizeScale( float fScale ) { m_fSizeScale = fScale; }
	//Scales the release rate that has been set in the template.
	//Scale must be between 0 and 1.
	void SetReleaseRateScale( float fScale);

	//Set culling bounding box
	void SetBoundingBox(const Asura_Bounding_Box& xBB);

	void SetDesiredCameraFlags( u_char ucFlags );

	float GetBoundingRadius(void) const {return m_fBoundingRadius;}

#ifdef ASURA_TOOLS
	void RenderVisibility();
#endif

#if _DEBUG
    virtual void DumpStats( void ) { }
#endif

#ifdef ASURA_TOOLS
	//Sets random number generator to use for cutscene recording purposes.	RAR 22/05/2006
	virtual void SetRNG(Asura_RandomGenerator* pxRNG){m_pxRNG = pxRNG;}
	Asura_RandomGenerator* GetRNG() const {return m_pxRNG ? m_pxRNG : Asura_Random::GetSharedGenerator();}
#else
	Asura_RandomGenerator* GetRNG() const {return Asura_Random::GetSharedGenerator();}
#endif

#ifdef ASURA_TOOLS
	//Flags effect as selected for determining whether to include it in the particle count. RAR 22/01/2007
	void SetSelectedInTools(bool bSet) {m_bSelectedInTools = bSet;}
#endif


private :
#if !defined(ASURA_USE_NEW_VISIBILITY_SYSTEM )
	void InvalidateModuleList();
	bool ModuleListIsValid() const;
	bool ModuleListIsInvalid() const;
	void CalculateModuleList();
	bool ModuleListIsVisible() const;
#endif
	

protected:
	ASURA_PFX_EFFECT_STATE		m_eState;
	Asura_Vector_3				m_xDelta; // for moving code effects/cutscene effects, so that sources can create particles back along the path
	unsigned short				m_usPFXFlags; // Flags for various settings (see enum above)

#ifdef ASURA_TOOLS
	//Effect specific RNG for cutscene AVI recording. RAR 22/05/2006
	Asura_RandomGenerator*		m_pxRNG;
#endif
private:
	unsigned char				m_ucCameraFlags; // Bitfield of which cameras can see this effect 
	unsigned char				m_ucCameraFlagsIgnoringFrustum;//Similarly but ignoring the frustum test

	unsigned int				m_uEffectID; // Unique, even after reset  


	Asura_Bounding_Box			m_xBB;//bounding box for culling
	float						m_fBoundingRadius;
	float						m_fSpeedScale;
	float						m_fSizeScale;
	float						m_fReleaseRateScale;

#if !defined( ASURA_USE_NEW_VISIBILITY_SYSTEM )
	int					m_iLastKnownModule[ MAX_CAMERAS_FOR_PFX_EFFECT ];

#if defined(ASURA_USE_STREAMING)
	Asura_Hash_ID		m_uLastKnownZone[ MAX_CAMERAS_FOR_PFX_EFFECT ];
	Asura_Streaming_ModuleList<MAX_VARIABLE_LENGTH_CLIPPING_COUNT, MAX_ZONES_FOR_PFX_CLIPPING> m_xModuleList;
#else
	unsigned short		m_ausModuleList[ MAX_VARIABLE_LENGTH_CLIPPING_COUNT ];
#endif


	//Various frame stamps etc for keeping track of module visibility
	u_int				m_uCameraFlagsFrameStamp;
	u_int               m_uRestartFrameNumber;
	u_int               m_uPrevCalcFrameNumber;
	u_int               m_uCalcInSuccession;


	u_int						m_bBoundingBoxValid:1,
								m_bAbandonModuleList:1,
								m_bModuleListValid:1;
#else
	u_int						m_uCameraFlagsFrameStamp;
	Asura_Visibility_Query		m_xVisibilityQuery;
	u_int						m_bBoundingBoxValid:1;
#endif

#ifdef ASURA_TOOLS
	bool				m_bSelectedInTools;
#endif

};

#endif//__ASURA_PFX_EFFECT_H__
