//////////////////////////////////////////////////////////////////////////////
//
// Asura_PFX_Effect_UpdateParticles.cpp
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Description:
//
//     main UpdateParticles function which can be called from main CPU or
//     from other threads.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"


#include "Core/Asura_Spline.h"
#include "SpecialFX/Asura_PerlinNoise_System.h"
#include "SpecialFX/Asura_SimplexNoise_System.h"
#include "SpecialFX/PFX/Asura_PFX_Classes.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Effect.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

static bool g_bUseSimplexNoise = false;

// static function
void Asura_PFX_Effect::UpdateParticles(Asura_PFX_Particle* pxParticles, u_int uNumParticles, const Asura_PFX_Source_Template* pxSourceTemplate, 
				bool bUseCollisionPlane, const Asura_Vector_3& xCollisionPosition, const Asura_Vector_3& xCollisionDirection, 
				bool bUsePointGravity, const Asura_Vector_3& xGravityPosition, float fSizeScale, float fTimeSlice, float fTurbulencePhase,
				const Asura_PFX_Particle_Template* pxPcleTemp, const Asura_Spline<Asura_Vector_3>* pxSpline /* = NULL */ )
{
	if(!uNumParticles) return;

	// Looped particles are not allowed to move - otherwise they'd end up travelling forever.
	if(pxPcleTemp && pxPcleTemp->m_bLooped) { return; }

	Asura_Assert(pxParticles, "No particles");
	Asura_Assert(pxSourceTemplate, "No source template");

	if(pxSourceTemplate->m_bUseTimeScale)
	{
		fTimeSlice *= pxSourceTemplate->m_fTimeScale;
	}

	if(pxSpline)
	{
		if(fTimeSlice > 0.f)
		{
			// Splines over-ride normal movement
			for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
			{
				// Get particle
				Asura_PFX_Particle& xParticle = pxParticles[uParticle];
				const float fSpeed = xParticle.m_xVelocity.Magnitude();
				const float fAge = xParticle.m_fAge * pxPcleTemp->GetLifetimeFromSeed(xParticle.m_fSeed);
				const float fDistance = fAge * fSpeed;
				const Asura_Vector_3 xOldPosition = xParticle.m_xPosition;
				pxSpline->GetPositionFromDistance(fDistance, xParticle.m_xPosition);
				const Asura_Vector_3 xDiff = xParticle.m_xPosition - xOldPosition;
				if(xDiff.MagnitudeSquared() > 0.f)
				{
					xParticle.m_xVelocity = xDiff;
					xParticle.m_xVelocity.Normalise();
					xParticle.m_xVelocity *= fSpeed;
				}
				else
				{
					// Kill the particle when it reaches the end of the spline
					xParticle.m_fAge = 1.f;
				}
			}
		}
		return;
	}

	// update position
	for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
	{
		// Get particle
		Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];

		pxParticle->m_xPosition += pxParticle->m_xVelocity * fTimeSlice;
	}

	// do collision
	if(bUseCollisionPlane)
	{
		for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
		{
			// Get particle
			Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];
			const float fPenetration = (pxParticle->m_xPosition - xCollisionPosition) * xCollisionDirection;
			if(fPenetration < 0.f)
			{
				// We have contacted/penetrated the floor

				// Change position to be on opposite side of plane
				pxParticle->m_xPosition -= xCollisionDirection * (2.f * fPenetration);
				// Change velocity to go in opposite direction to plane
				const float fDotProd = (pxParticle->m_xVelocity * xCollisionDirection);
				if(fDotProd < 0.f)
				{
					const float fElasticity = pxSourceTemplate->m_fParticleBounce + 1.f;
					pxParticle->m_xVelocity -= xCollisionDirection * (fElasticity * fDotProd);
				}
				// Apply friction
				const Asura_Vector_3 xDirectionOnFloor = pxParticle->m_xVelocity - xCollisionDirection * fDotProd;
				pxParticle->m_xVelocity -= xDirectionOnFloor * pxSourceTemplate->m_fParticleFriction * fTimeSlice;
			}
		}
	}

	// Apply gravity
	if(bUsePointGravity)
	{
		for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
		{
			// Get particle
			Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];
			const Asura_Vector_3 xDirection(xGravityPosition - pxParticle->m_xPosition);
			const float fDistance = xDirection.Magnitude();
			const float fMag = (fDistance > 0.f) ? (fTimeSlice / fDistance) : 0.f;
			pxParticle->m_xVelocity += xDirection * (fMag * pxSourceTemplate->m_xGravityStrength.GetValueAtTime(pxParticle->m_fAge));
		}
	}
	else if(pxSourceTemplate->m_bDirectionalGravity)
	{
		for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
		{
			// Get particle
			Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];
			pxParticle->m_xVelocity += pxSourceTemplate->m_xGravityVector * fTimeSlice * pxSourceTemplate->m_xGravityStrength.GetValueAtTime(pxParticle->m_fAge);
		}
	}
	// Turbulence
	if(pxSourceTemplate->m_bUseTurbulence)
	{
		for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
		{
			// Get particle
			Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];
			const float fAmplitude = pxSourceTemplate->m_xTurbulenceStrength.GetValueAtTime(pxParticle->m_fAge) * fTimeSlice;
			const Asura_Vector_3 xNoiseInput = pxParticle->m_xPosition * pxSourceTemplate->m_fTurbulenceFrequency;

			Asura_Vector_3 xTurb;
			if(g_bUseSimplexNoise)
			{
				xTurb.x = Asura_SimplexNoise_System::GetNoise(Asura_Vector_3(xNoiseInput.y, xNoiseInput.z, fTurbulencePhase));
				xTurb.y = Asura_SimplexNoise_System::GetNoise(Asura_Vector_3(xNoiseInput.z, xNoiseInput.x, fTurbulencePhase));
				xTurb.z = Asura_SimplexNoise_System::GetNoise(Asura_Vector_3(xNoiseInput.x, xNoiseInput.y, fTurbulencePhase));				
			}
			else
			{
				Asura_PerlinNoise_System::CalculateNoiseVector( xNoiseInput, fTurbulencePhase, xTurb ); 
			}
			pxParticle->m_xVelocity += xTurb * fAmplitude;
		}
	}
	// Apply air resistance
  	if(pxSourceTemplate->m_bARPTS)
  	{
  		Asura_Assert(pxPcleTemp, "No particle template");
		if (pxPcleTemp)
		{
  			for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
  			{
  				// Get particle
  				Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];
  				float fSize = pxPcleTemp->m_xSize.GetValueAtTime(pxParticle->m_fAge);
  				if( pxPcleTemp->m_bUseSizeScale )
  				{
  					fSize *= fSizeScale;
  				}

  				pxParticle->m_xVelocity *= Asura_Maths::Max(0.0f, (1.f - (fSize * (pxSourceTemplate->m_fAirResistance * fTimeSlice))));
  			}
		}
  	}
  	else if(pxSourceTemplate->m_bFixedAirResistance)
	{
		for(u_int uParticle = 0; uParticle < uNumParticles; ++uParticle)
		{
			// Get particle
			Asura_PFX_Particle* pxParticle = &pxParticles[uParticle];
			pxParticle->m_xVelocity *= Asura_Maths::Max(0.0f, (1.f - (pxSourceTemplate->m_fAirResistance * fTimeSlice)));
		}
	}
}



/// End of file. ///
