
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Streaming_System.h
// Author : Andy Tate
// Date   : 2-Feb-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

// RJT NOTE: Required for GCC builds as the file list generation doesn't account for VC project exclusion from build
#ifdef ASURA_USE_STREAMING

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_STREAMING_SYSTEM_H__
#define __ASURA_STREAMING_SYSTEM_H__

////////////////////////////////////////////////////////////////////////////////
// Define checking before any other defines are created...
////////////////////////////////////////////////////////////////////////////////

#if !defined(ASURA_USE_MULTIPLE_HANDLERS)
#error "ASURA_USE_MULTIPLE_HANDLERS" must be defined for streaming builds.
#endif	//!defined(ASURA_USE_MULTIPLE_HANDLERS)

#if !defined(ASURA_USE_MULTIPLE_ZONE_HANDLERS)
#error "ASURA_USE_MULTIPLE_ZONE_HANDLERS" must be defined for streaming builds.
#endif	//!defined(ASURA_USE_MULTIPLE_HANDLERS)


////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "ProjectSettings/Asura_StreamingConfiguration.h"

#include "Network/Asura_Guid.h"
#include "Streaming/Asura_Streaming_ZoneHandler.h"
#include "Streaming/Jobs/Asura_Streaming_JobQueue.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_FileCache;
class Asura_MemoryBank;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_Streaming_System
// Author       : Andy Tate
// Date         : 2-Feb-2006
//
// Description  : Handles the management of the streaming system.
//
//				  NOTE: The streaming system is only called when the define
//				  ~~~~~ "ASURA_USE_STREAMING" is present.
////////////////////////////////////////////////////////////////////////////////

class Asura_Streaming_System
{
public:
	// Typedefs
	typedef void TraverseZoneCallback (Asura_Hash_ID uHash, Asura_Streaming_Zone* pxZone);

	// Public enumerated types...
	enum DataType
	{
		PERMANENT_DATA,
		COMMON_DATA,
		SKYBOX_DATA,
		ZONE_DATA,
		MISSION_DATA,

		MAX_DATA_TYPES
	};

#if defined(ASURA_RESOURCE_ALLOCATOR_USES_MEMORY_BANK)
	enum MemoryBankType
	{
		PERMANENT_MEMORY_BANK,
		COMMON_MEMORY_BANK,
		SKYBOX_MEMORY_BANK,

		MAX_MEMORY_BANK_TYPES
	};
#endif	// defined(ASURA_RESOURCE_ALLOCATOR_USES_MEMORY_BANK)

	// Public functions...

	// Initialisation/shutdown functions...
	static	void						Initialise					(void);
	static	void						Shutdown					(void);

	// Creation/destruction functions...
	static	void						AllocateZones				(const u_int* puNumZoneDataIndices, const u_int* puNumMissionDataIndices);
	static	void						DeallocateZones				(void);

	static	bool						CreateData					(const DataType eDataType);
	static	bool						DestroyData					(const DataType eDataType);

	static	Asura_Streaming_Zone*		CreateZone					(const DataType eDataType, const Asura_String& xFilename)	{ return(CreateZone(eDataType, xFilename.GetString())); }
	static	Asura_Streaming_Zone*		CreateZone					(const DataType eDataType, const char* szFilename);
	static	bool						DestroyZone					(const DataType eDataType, const Asura_Hash_ID uHash);
	static	void						DestroyAllZones				(const DataType eDataType);

	static	bool						ReloadZone					(const DataType eDataType, const Asura_Hash_ID uHash, const bool bBlocking);

	static	void						DestroySavedMissionEntities	(void);

	// MGH 15/11/2006: Function to allow projects to pre-fill the unique index hash tree.
	// This allows them to ensure that zones always have the same unique index
	// (which is used to modify GUIDs). Hence allowing such things as messaging between zones.
	static int AddToZoneUniqueIndexHashTree(const Asura_Hash_ID uHash);
	// Look up the unique indices in the tree for the zone handler.
	static int GetZoneHandlerUniqueIndexForZone(const Asura_Hash_ID uHash);

	// Loading functions...
	static	bool						CreateEmptyZone				(const DataType eDataType, const Asura_Hash_ID uHash);
	static	bool						LoadZone					(const DataType eDataType, const Asura_Hash_ID uHash, const bool bRecurse);

#if !defined(ASURA_TOOLS)
	// Guid calculation functions...
	static	Asura_Guid					MakeCurrentMasterZoneGuid	(const Asura_Guid uGuid);
#endif	// !defined(ASURA_TOOLS)

	// Update functions...
	static	void						Update						(const bool bRecurse);
	static	void						UpdateEntities				(void);

	static	void						PreRenderUpdate				(void);

	static	void						EndOfFrameUpdate			(const float fTimeAllowed, const float fFrameDuration);

	// Suspend/resume functions...
	static	void						Suspend						(void);
	static	void						Resume						(void);

	// Render functions...
	static	void						RenderEntities				(void);
	static	void						RenderEnvironment			(void);
	static	void						RenderSkybox				(void);
	static	void						RenderSpecialFX				(void);

	// Client update functions
	static	void ResetPeriodicClientUpdateList();
	static	void FinishPeriodicClientUpdates();
	static	void DoSomePeriodicClientUpdates();

	static	void ResetConstantClientUpdateList();
	static	void DoConstantClientUpdates();


	// Query functions...
	static	Asura_Streaming_Zone**		MakeArrayOfZones			(const DataType eDataType, int& iNumZones);

	static	Asura_Streaming_Zone*		FindMission					(const Asura_Hash_ID uHash)		{ return(s_pxMissionHandler->FindZone(uHash)); }
	static	Asura_Streaming_Zone*		FindMission					(const char* szName)			{ return(FindMission(Asura_GetHashID(szName))); }

	static	Asura_Streaming_Zone*		FindZone					(const Asura_Hash_ID uHash)		{ return(s_pxZoneHandler->FindZone(uHash)); }
	static	Asura_Streaming_Zone*		FindZone					(const char* szName)			{ return(FindZone(Asura_GetHashID(szName))); }	

	static	Asura_Streaming_Zone*		FindMasterZone				(const Asura_Hash_ID uHash);
	static	Asura_Streaming_Zone*		FindMasterZone				(const char* szName)			{ return(FindMasterZone(Asura_GetHashID(szName))); }

	//Get the offset that needs to be added to transform from a position relative to iZoneIndexFrom
	//to a position relative to iZoneIndexTo.
	//TBD ... Actually implement it.
	//RAR 27/02/2006 
	static	Asura_Vector_3				GetOffsetBetweenZones		(int iZoneIndexFrom,int iZoneIndexTo) {return Asura_Maths::ZeroVector3;}

#if defined(ASURA_RESOURCE_ALLOCATOR_USES_MEMORY_BANK)
	// Memory-bank functions...
	static	void						AllocateMemoryBanks			(const DataType eDataType, Asura_Streaming_ZoneHandler::MemoryBankAllocationFunction* pfnCreateMemoryBank);
	static	void						DeallocateMemoryBanks		(const DataType eDataType);

	static	void						ResetCurrentMemoryBank		(void);
	static	Asura_MemoryBank*			GetMemoryBankFromDataIndex	(const u_int uDataIndex);
#endif	// defined(ASURA_RESOURCE_ALLOCATOR_USES_MEMORY_BANK)

	// Calculates offsets to and from the zone...
	static  void						CalculateOffsets			(Asura_Streaming_Zone* pxZone)	{ s_pxZoneHandler->CalculateOffsets(pxZone); }

	// Reset index functions...
	static	u_int						GetResetGroupFromDataIndex	(const u_int uDataIndex);

	// Accessor functions...
	static	void						SetAllowLODs				(const bool bSet)				{ s_bAllowLODs = bSet; }
	static	bool						GetAllowLODs				(void)							{ return(s_bAllowLODs); }

	static  void						SetShouldQueueMessages		(const bool bSet)				{ s_bShouldQueueMessages = bSet; }
	static  bool						GetShouldQueueMessages		()								{ return s_bShouldQueueMessages; }

	static	void						SetForceBlockingUpdate		(const bool bSet)				{ s_bForceBlockingUpdate = bSet; }
	static	bool						GetForceBlockingUpdate		(void)							{ return(s_bForceBlockingUpdate); }

	static	bool						IsInLowDetailZone			(void)							{ return(s_bInLowDetailZone); }

	static	void						ResetCurrentMission			(void)							{ s_pxMissionHandler->ResetCurrentZone(); }
	static	void						SetCurrentMission			(const Asura_Hash_ID uHash)		{ s_pxMissionHandler->SetCurrentZone(uHash); }
	static	void						SetCurrentMission			(const Asura_Hash_ID uHash, const Asura_Streaming_Zone::DataState eDataState, const bool bAllowDeletion)	{ s_pxMissionHandler->SetCurrentZone(uHash, eDataState, bAllowDeletion); }
	static	Asura_Streaming_Zone*		GetCurrentMission			(void)							{ return((s_pxMissionHandler == NULL) ? NULL : s_pxMissionHandler->GetCurrentZone()); }

	static	void						ResetCurrentZone			(void)							{ s_pxZoneHandler->ResetCurrentZone(); }
	static	void						SetCurrentZone				(const Asura_Hash_ID uHash, const Asura_Streaming_Zone::DataState eDataState, const bool bAllowDeletion)	{ s_pxZoneHandler->SetCurrentZone(uHash, eDataState, bAllowDeletion); }
	static	void						SetCurrentZone				(const Asura_Hash_ID uHash)		{ s_pxZoneHandler->SetCurrentZone(uHash); }
	static	Asura_Streaming_Zone*		GetCurrentZone				(void)							{ return((s_pxZoneHandler == NULL) ? NULL : s_pxZoneHandler->GetCurrentZone()); }
	static	u_int						GetNumZones					(const DataType eDataType);

	static	Asura_Streaming_Zone*		GetZoneToBeCreatedIn		(void);
	// MGH 22/5/2006: Return the hash for the current zone, but only if it isn't a mission.
	static	Asura_Hash_ID				GetContainingZoneHash(const Asura_Hash_ID uHash);

	static	u_int						GetNumDataIndices			(void)							{ return(s_uNumDataIndices); }

	static	void						ResetControllerEntityGuid	(void)							{ s_uControllerEntityGuid = ASURA_GUID_UNREGISTERED; }
	static	void						SetControllerEntityGuid		(const Asura_Guid uGuid)		{ s_uControllerEntityGuid = uGuid; }
	static	Asura_Guid					GetControllerEntityGuid		(void)							{ return(s_uControllerEntityGuid); }

	static void SaveMissionDataToBuffer(class Asura_SaveBlock_Base& xSaveBlock);
	static	Asura_Streaming_JobQueue&	GetJobQueue					(void)							{ return(s_xJobQueue); }

	// Return the frame number of the completion of the most recent zone load.
	// This is to allow systems such as Shape Instances to redo module tests if necessary.
	static u_int GetFrameNumberLastZoneLoad() { return s_uFrameNumberLastZoneLoad; }
	static void MarkZoneLoaded();

	// Get and set for priority flag. When true, new entities will be created in the current mission unless a load is in progress, and if false they'll be in the current zone.
	// Default (and the behaviour in Gun) is with this flag set to true.
	static void SetMissionHasPriority(bool bMissionHasPriority)		{ s_bMissionHasPriority = bMissionHasPriority; }
	static bool GetMissionHasPriority()								{ return s_bMissionHasPriority; }

	// Traverse all of the zones in the zonehandler.
	static void TraverseZones(TraverseZoneCallback* pfnCallback)
	{
		if (s_pxZoneHandler)
		{
			s_pxZoneHandler->TraverseZones(pfnCallback);
		}
	}


#ifdef ASURA_TOOLS
	// Tools needs to access the special data hash ids, so it can tell if the zone being loaded is the actual level or not.
	static Asura_Hash_ID GetPermanentDataHash() { return s_uPermanentDataHash; }
	static Asura_Hash_ID GetCommonDataHash() { return s_uCommonDataHash; }
	static Asura_Hash_ID GetSkyboxDataHash() { return s_uSkyboxDataHash; }
#endif //#ifdef ASURA_TOOLS

protected:
	// Protected variables...


	// Protected functions...


private:
	// Private variables...

	// Flags...
	static	bool volatile					s_bSuspended;				// Flag to say whether we're currently suspended.
	static	bool volatile					s_bUpdating;				// Flag to say whether we're currently updating any jobs
	static	bool							s_bAllowLODs;				// Flag to say whether we bother with LODs.
	static  bool							s_bShouldQueueMessages;		// If true we will try to queue all messages to server entities.
	static	bool							s_bForceBlockingUpdate;		// Flag to force all updates to be performed blocking.
	static	bool							s_bInLowDetailZone;			// Flag to say whether the player is in a low-detail zone.
	static	bool							s_bMissionHasPriority;		// Flag for where to put new entities if no loading is in progress, also affecting whether zones or missions update first. Set to true to behave as Gun did.

	static  u_int							s_uFrameNumberLastZoneLoad; // Frame number of completion of last zone load.

	// Storage data...
	static	Asura_Streaming_ZoneHandler*	s_pxMissionHandler;			// Handler for the missions.
	static	Asura_Streaming_ZoneHandler*	s_pxZoneHandler;			// Handler for the zones.

	// Current data...
	static	Asura_Guid						s_uControllerEntityGuid;	// The Guid to take the current position from.
	static	Asura_Vector_3					s_xCurrentPosition;			// The position which the streaming decisions are made from.

	// Memory-bank indices...
	static	u_int							s_uPermanentDataIndex;
	static	u_int							s_uCommonDataIndex;
	static	u_int							s_uSkyboxDataIndex;

	static	u_int							s_uNumDataIndices;

#if defined(ASURA_RESOURCE_ALLOCATOR_USES_MEMORY_BANK)
	// Memory-banks...
	static	Asura_MemoryBank*				s_pxPermanentMemoryBank;
	static	Asura_MemoryBank*				s_pxCommonMemoryBank;
	static	Asura_MemoryBank*				s_pxSkyboxMemoryBank;
#endif	// defined(ASURA_RESOURCE_ALLOCATOR_USES_MEMORY_BANK)

	// Loading handlers...
	static	u_char*							s_pucFileCacheMemory;		// The block of memory used by the file-cache.
	static	Asura_FileCache*				s_pxFileCache;				// The file-cache to use when streaming.

	static	float							s_fJobQueueUpdateTime;		// The cumulative time we've spent handling the job-queue.
	static	float							s_fJobQueueFrameTime;		// The total time we've spent for the job-queue updates.
	static	Asura_Streaming_JobQueue		s_xJobQueue;				// The queue of jobs to be processed by the streaming system.

	// Private functions...

	Asura_Streaming_System  (void);
	Asura_Streaming_System  (const Asura_Streaming_System& xOther);
	~Asura_Streaming_System ();

			Asura_Streaming_System&	operator=							(const Asura_Streaming_System& xOther);

	// Internal update functions...
	static	void					UpdateConnections					(const Asura_Vector_3* pxCurrentPosition, const bool bBlocking);
	static	bool					CheckInLowDetailZone				(const Asura_Streaming_Zone& xCurrentZone, const Asura_Vector_3& xCurrentPosition);

	// Separated update stages, so we can swap the order round according to priority easily.
	static void UpdateZoneHandler(bool bRecurse);
	static void UpdateMissionHandler();

	// Platform specific file-cache functions...
	static	u_char*					Platform_AllocateFileCacheMemory	(const u_int uSize);
	static	void					Platform_DeallocateFileCacheMemory	(u_char* pucData);

	static const char*			s_szPermanentDataName;
	static const Asura_Hash_ID	s_uPermanentDataHash;

	static const char*			s_szCommonDataName;
	static const Asura_Hash_ID	s_uCommonDataHash;

	static const char*			s_szSkyboxDataName;
	static const Asura_Hash_ID	s_uSkyboxDataHash;
};   // End of class 'Asura_Streaming_System'.


#endif   // ifndef __ASURA_STREAMING_SYSTEM_H__

#endif	// ASURA_USE_STREAMING

/// End of file. ///
