// Asura_UnitTest_Memheap.cpp
// TY 31/10/07
// Copyright (c) 2007 Rebellion. All rights reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_UNITTEST

#include "UnitTest/Asura_UnitTest.h"

#include "Core/Asura_Maths.h"
#include "Core/Asura_Memheap.h"
#include "Core/Asura_MemoryManagement.h"
#include "Debugging/Asura_Assert.h"

#if defined(_WIN64) || defined(_WIN32)
#include <basetsd.h>						//(for UINT_PTR)
#endif

//////////////////////////////////////////////////////////////////////////////

//Local class for providing contiguous memory allocation
class Asura_Memory_Page
{
public:

	Asura_Memory_Page (size_t uPageSize)
	{
		m_pcBaseMem		= 0;
		m_uMemPointer	= 0;
		m_uMemSize		= 0;

		m_pcBaseMem = new char[uPageSize];
		if(m_pcBaseMem)
		{
			m_uMemSize = uPageSize;
		}
	}
	~Asura_Memory_Page()
	{
		delete []m_pcBaseMem;
	}

	void* GetMemory  (size_t uSize)
	{
		void* pRet;
		if(uSize+m_uMemPointer <= m_uMemSize)
		{
			pRet = static_cast<void*>(&m_pcBaseMem[m_uMemPointer]);
			m_uMemPointer+=uSize;
			return pRet;
		}
		Asura_Assert(false, "Out of memory");
		return 0;
	}
	void* TrimMemory (size_t uSize)
	{
		if(m_uMemPointer >= uSize)
		{
			m_uMemPointer -= uSize;
			return static_cast<void*>(&m_pcBaseMem[m_uMemPointer]);
		}
		Asura_Assert(false, "Trimming too much");
		return 0;
	}
	void  ClearMemory()
	{
		m_uMemPointer = 0;
	}

private:

	char* m_pcBaseMem;
	size_t m_uMemPointer;
	size_t m_uMemSize;
};


//////////////////////////////////////////////////////////////////////////////

const unsigned int uASURA_MEMHEAP_TEST_NUM_ALLOCS		= 150;	//Do this many allocations...
const unsigned int uASURA_MEMHEAP_TEST_MAX_SAMPLE_SIZE	= 4096;	//each up to this size (*4 because it will allocate ints)
const unsigned int uASURA_MEMHEAP_TEST_MAX_SAMPLE_NUM	= 32768;
const          int iASURA_MEMHEAP_TEST_RAND_SEED		= 0xDEADBEEF;

#if defined(_WIN64) || defined(_WIN32)	//Windows
typedef UINT_PTR ptr_uint;
#else									//Other platforms (which are all 32bit)
typedef u_int ptr_uint;
#endif

//////////////////////////////////////////////////////////////////////////////

static Asura_Memory_Page* g_pxMemPage = 0;

//////////////////////////////////////////////////////////////////////////////

enum ASURA_MEMHEAP_TEST_STAGES
{
	ASURA_MEMHEAP_TEST_STAGE1,
	ASURA_MEMHEAP_TEST_STAGE2,
	ASURA_MEMHEAP_TEST_STAGE3,
//	ASURA_MEMHEAP_TEST_STAGE4,
	ASURA_MEMHEAP_TEST_STAGE5,

	ASURA_MEMHEAP_TEST_NUM_STAGES
};

//////////////////////////////////////////////////////////////////////////////

static void* Asura_MemHeap_Test_Alloc(size_t uSize)
{
	if(g_pxMemPage)
	{
		return g_pxMemPage->GetMemory(uSize);
	}
	return 0;
}

static void* Asura_MemHeap_Test_Trim(size_t uSize)
{
	if(g_pxMemPage)
	{
		return g_pxMemPage->TrimMemory(uSize);
	}
	return 0;
}

ASURA_UNITTEST(GeneralTests, MemHeap)
{
	//randomise
	Asura_Random::Initialise();
	Asura_Random::SetSeed(iASURA_MEMHEAP_TEST_RAND_SEED);
	
	for(u_int uTestNum=0;uTestNum<ASURA_MEMHEAP_TEST_NUM_STAGES;++uTestNum)
	{
	    Asura_MemHeap*	pxMemHeap			= 0;	
//	    Asura_MemHeap*	pxParentMemHeap		= 0;
	    u_int			uTotal				= 0;
//	    char*			pcTemp				= 0;
	    char*			pcFixedMem			= 0;

    	Asura_Memheap_Core_Details	xDetails;

        // Allocate the mem page every time since most tests use it.
		g_pxMemPage = new Asura_Memory_Page(2*1024*1024);	//use 2Mb of contiguous memory
		ASURA_UNITTEST_CHECK(g_pxMemPage);
		if(!g_pxMemPage)
		{
			Asura_Assert(false, "Failed to create memory page");
			return;
		}

		//Behave differently for the different stages of the test
		switch(uTestNum)
		{
			case ASURA_MEMHEAP_TEST_STAGE1:
				// Use the default page size, contiguous allocator.

				xDetails.m_pfnAllocator	= &Asura_MemHeap_Test_Alloc;
				xDetails.m_pfnTrimmer	= &Asura_MemHeap_Test_Trim;

				pxMemHeap = new Asura_MemHeap( xDetails );
				ASURA_UNITTEST_CHECK(pxMemHeap);
				if(!pxMemHeap)
				{
					Asura_Assert(false, "Failed to create memheap");
					delete g_pxMemPage;
					return;
				}

				break;
			
			case ASURA_MEMHEAP_TEST_STAGE2:				
				// Use larger page size, contiguous allocator, padding,
                // thread safety.

				xDetails.m_pfnAllocator	= &Asura_MemHeap_Test_Alloc;
				xDetails.m_pfnTrimmer	= &Asura_MemHeap_Test_Trim;
				xDetails.m_uPageSize    = 16 * 1024;
				xDetails.m_uPad		    = 1024;

				pxMemHeap = new Asura_MemHeap( xDetails );
				ASURA_UNITTEST_CHECK(pxMemHeap);	
				if(!pxMemHeap)
				{
					Asura_Assert(false, "Failed to create memheap");
					delete g_pxMemPage;
					return;
				}

				pxMemHeap->EnableThreadSafety();
				break;

			case ASURA_MEMHEAP_TEST_STAGE3:
				// Use smaller page size, contiguous allocator, smaller
                // trim threshold

				xDetails.m_pfnAllocator	= &Asura_MemHeap_Test_Alloc;
				xDetails.m_pfnTrimmer	= &Asura_MemHeap_Test_Trim;
				xDetails.m_uPageSize	= 2 * 1024;
                xDetails.m_uTrim        = 4 * 1024;

				pxMemHeap = new Asura_MemHeap(xDetails);
				ASURA_UNITTEST_CHECK(pxMemHeap);
				if(!pxMemHeap)
				{
					delete g_pxMemPage;
					Asura_Assert(false, "Failed to create memheap");
					return;
				}		
				break;
				
//			case ASURA_MEMHEAP_TEST_STAGE4:
//				// Use the default page size, contiguous allocator, sub heap.
//				xDetails.m_pfnAllocator	= &Asura_MemHeap_Test_Alloc;
//				xDetails.m_pfnTrimmer	= &Asura_MemHeap_Test_Trim;
//
//				pxParentMemHeap = new Asura_MemHeap( xDetails );
//				ASURA_UNITTEST_CHECK(pxParentMemHeap);
//				if(!pxParentMemHeap)
//				{					
//					Asura_Assert(false, "Failed to create memheap");
//					return;
//				}		
//
//				pxMemHeap = new Asura_MemHeap(pxParentMemHeap);
//				ASURA_UNITTEST_CHECK(pxMemHeap);
//				if(!pxMemHeap)
//				{					
//					Asura_Assert(false, "Failed to create memheap");
//					return;
//				}
//
//              //Use up some of the parent heap
//				pcTemp = reinterpret_cast<char*>(pxParentMemHeap->Alloc(1024));
//
//				break;

			case ASURA_MEMHEAP_TEST_STAGE5:

				//Use a fixed memory page
				pcFixedMem = new char[3*1024*1024]; //3Mb of fixed mem
				pxMemHeap = new Asura_MemHeap(pcFixedMem, 3*1024*1024);
				ASURA_UNITTEST_CHECK(pxMemHeap);
				if(!pxMemHeap)
				{
					Asura_Assert(false, "Failed to create memheap");
					return;
				}
				break;
		}

		// --------- Tests --------------


		//Do some allocations		
		u_int*	apuInt	[uASURA_MEMHEAP_TEST_NUM_ALLOCS];	//Allocations
		char*   pcChar;	//Aligned Allocations
		int		aiSizes	[uASURA_MEMHEAP_TEST_NUM_ALLOCS];	//Sizes of the allocations

		uTotal = 0;
		for(int i=0;i<int(uASURA_MEMHEAP_TEST_NUM_ALLOCS);++i)
		{
			aiSizes[i] = static_cast<int>(2 + static_cast<float>(uASURA_MEMHEAP_TEST_MAX_SAMPLE_SIZE-2)/static_cast<float>(uASURA_MEMHEAP_TEST_NUM_ALLOCS)*i);
			apuInt[i] = (u_int*)pxMemHeap->Alloc(sizeof(u_int)*aiSizes[i]);
			ASURA_UNITTEST_CHECK(apuInt[i]);	//Check allocation succeeded			
			for(int j=0;j<aiSizes[i];++j)
			{
				apuInt[i][j] = j % uASURA_MEMHEAP_TEST_MAX_SAMPLE_NUM;
			}
			//this line is here primarily so you can judge approximately how much memory is being requested
			//(about 2Mb with the current settings: 250 allocations up to 16384 bytes)
			uTotal += aiSizes[i]*sizeof(int);	
		}
		
		//Check we didnt have integrity failures
		#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
		ASURA_UNITTEST_CHECK(!pxMemHeap->GetNumIntegrityFailures());
		#endif

		//loop over checking our data and freeing
		int iIndex = 0;

		bool bDone = false;
		while(!bDone)
		{	
			//Free a random allocation
			iIndex = Asura_Random::GetSeededInt() % uASURA_MEMHEAP_TEST_NUM_ALLOCS;
			if(!apuInt[iIndex])
			{
				bDone = true;
				for(int i = 0; i < int(uASURA_MEMHEAP_TEST_NUM_ALLOCS); ++i)
				{
					if(apuInt[i])
					{
						iIndex = i;
						bDone = false;
						break;
					}		
				}
			}
			if(!bDone)
			{
				for(int i=0;i<aiSizes[iIndex];++i)
				{
					ASURA_UNITTEST_CHECK((apuInt[iIndex][i] == i % uASURA_MEMHEAP_TEST_MAX_SAMPLE_NUM));	//check our data is ok
				}
				pxMemHeap->Free(apuInt[iIndex]);
				apuInt[iIndex] = 0;			
			}		

			//Do an aligned allocation
			u_int uAlign = static_cast<u_int>(Asura_Maths::Power(2.0 ,static_cast<float>(Asura_Random::GetSeededInt() % 5 + 4))); //alignment from 16 to 256;
			u_int uBytes = static_cast<u_int>(Asura_Random::GetSeededInt()%uASURA_MEMHEAP_TEST_MAX_SAMPLE_SIZE+2);

			pcChar = (char*)pxMemHeap->Alloc(uBytes,uAlign);
			ASURA_UNITTEST_CHECK(reinterpret_cast<ptr_uint>(pcChar) % uAlign == 0); //check we got some memory aligned to our requirements
			pxMemHeap->Free(pcChar);

			//Check we didn't have integrity failures
			#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
			ASURA_UNITTEST_CHECK(!pxMemHeap->GetNumIntegrityFailures());
			#endif
		}

		//Test realloc
		for(int i=0;i<int(uASURA_MEMHEAP_TEST_NUM_ALLOCS);++i)
		{
			//pick a size which is between 50% and 150% of the size we desire, then realloc it
			int iInitialSize = static_cast<int>(static_cast<float>(aiSizes[i]) * Asura_Random::GetFloat(0.5f,1.5f));
			iInitialSize = iInitialSize;

			apuInt[i] = (u_int*)pxMemHeap->Alloc(iInitialSize * sizeof(int));
			ASURA_UNITTEST_CHECK(apuInt[i]);
			for(int j=0;j<iInitialSize;++j)
			{
				apuInt[i][j] = j % uASURA_MEMHEAP_TEST_MAX_SAMPLE_NUM;
			}

			//Realloc it to the size we desire
			apuInt[i] = (u_int*)pxMemHeap->Realloc(apuInt[i],sizeof(u_int) * aiSizes[i]);
			ASURA_UNITTEST_CHECK(apuInt[i]);

			//Check our original data is ok
			for(int j=0;j< Asura_Maths::Min(iInitialSize,aiSizes[i]);++j)
			{
				ASURA_UNITTEST_CHECK((apuInt[i][j] == j % uASURA_MEMHEAP_TEST_MAX_SAMPLE_NUM));	//check our data is ok
			}
			
			//check we can write to any more memory that has been allocated for us
			if(aiSizes[i] > iInitialSize)
			{
				for(int j=iInitialSize;j<aiSizes[i];++j)
				{
					apuInt[i][j] = j % uASURA_MEMHEAP_TEST_MAX_SAMPLE_NUM;
				}
			}
			//Check we didn't have integrity failures
			#ifdef ASURA_MEMHEAP_INTEGRITY_CHECKS
			ASURA_UNITTEST_CHECK(!pxMemHeap->GetNumIntegrityFailures());
			#endif
			
			//and finally free...
			pxMemHeap->Free(apuInt[i]);
		}

		// ---------- free as appropriate ------------
		switch(uTestNum)
		{
			case ASURA_MEMHEAP_TEST_STAGE1:
				delete pxMemHeap;
				break;

			case ASURA_MEMHEAP_TEST_STAGE2:
				delete pxMemHeap;
				break;

			case ASURA_MEMHEAP_TEST_STAGE3:
				delete pxMemHeap;
				break;

//			case ASURA_MEMHEAP_TEST_STAGE4:
//				pxParentMemHeap->Free(pcTemp);
//				delete pxParentMemHeap;
//				delete pxMemHeap;
//				break;

			case ASURA_MEMHEAP_TEST_STAGE5:
				delete pxMemHeap;
				delete[] pcFixedMem;
				break;
		}

        delete g_pxMemPage;
        g_pxMemPage = NULL;
	}
}

#endif //#ifdef ASURA_SUPPORT_UNITTEST

//eof
