//
// Asura_Visibility_System.cpp
//
// Created by KJFL - 06 May 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"


#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM

// This file's header.
#include "Visibility/Asura_Visibility_System.h"

#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Rendering/Asura_Render.h"
#include "Visibility/Asura_Visibility_Node.h"

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#endif

#ifdef ASURA_TOOLS
#include "Visibility/Tools_Visibility_System.h"
#endif

u_int					Asura_Visibility_System::s_uNumberOfNodes;
Asura_Visibility_Node*	Asura_Visibility_System::s_pxVisibilityNodes;	

// KJFL - 18 February 2008 - Extra bitfield per node for hierarchical frustum tests
// - uses a single bit per camera frustum
// - better memory access if a separate array, and can be more easily zeroed etc
// - dynamic so no need to load/save
Asura_Visibility_NodeFlags*	Asura_Visibility_System::s_pxNodeFlags;


u_int					Asura_Visibility_System::s_uNumberOfLeaves;
Asura_Visibility_Leaf*	Asura_Visibility_System::s_pxVisibilityLeaves;

u_int					Asura_Visibility_System::s_uNumberOfVisibilityFeatures;
u_char*					Asura_Visibility_System::s_pucVisibleFeatures;

u_int					Asura_Visibility_System::s_auLastKnownCameraNode[];
Asura_Visibility_Leaf*	Asura_Visibility_System::s_apxCameraLeaf[];

bool					Asura_Visibility_System::s_bForceVisRecalcForCurrentCameraNode;
bool					Asura_Visibility_System::s_bRenderEverything;
bool					Asura_Visibility_System::s_bFreezeCameraNode;
bool					Asura_Visibility_System::s_bXRayMode;

void Asura_Visibility_System::Initialise( void )
{
	s_uNumberOfNodes = 0;
	s_pxVisibilityNodes = 0;
	s_pxNodeFlags = 0;

	s_uNumberOfLeaves = 0;
	s_pxVisibilityLeaves = 0;

	s_uNumberOfVisibilityFeatures = 0;
	s_pucVisibleFeatures = 0;

	s_bForceVisRecalcForCurrentCameraNode = true;

	s_bRenderEverything = false;
	s_bFreezeCameraNode = false;
	s_bXRayMode = false;

	for ( u_int u=0; u<ASURA_MAX_NO_OF_CAMERAS; ++u )
	{
		s_apxCameraLeaf[u] = 0;
		s_auLastKnownCameraNode[u] = 0;
	}

	#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	AddConsoleCommands();
	#endif
}

void Asura_Visibility_System::Reset( void )
{
	delete [] s_pxVisibilityNodes;
	delete [] s_pxNodeFlags;
	s_pxVisibilityNodes = 0;
	s_pxNodeFlags = 0;
	s_uNumberOfNodes = 0;

	delete [] s_pxVisibilityLeaves;
	s_pxVisibilityLeaves = 0;
	s_uNumberOfLeaves = 0;

	s_uNumberOfVisibilityFeatures = 0;
	delete [] s_pucVisibleFeatures;
	s_pucVisibleFeatures = 0;

	for ( u_int u=0; u<ASURA_MAX_NO_OF_CAMERAS; ++u )
	{
		s_apxCameraLeaf[u] = 0;
		s_auLastKnownCameraNode[u] = 0;
	}

	s_bForceVisRecalcForCurrentCameraNode = true;
	s_bFreezeCameraNode = false;
}

void Asura_Visibility_System::Shutdown( void )
{
	delete [] s_pxVisibilityNodes;
	delete [] s_pxNodeFlags;
	s_pxVisibilityNodes = 0;
	s_pxNodeFlags = 0;
	s_uNumberOfNodes = 0;


	delete [] s_pxVisibilityLeaves;
	s_pxVisibilityLeaves = 0;
	s_uNumberOfLeaves = 0;

	s_uNumberOfVisibilityFeatures = 0;
	delete [] s_pucVisibleFeatures;
	s_pucVisibleFeatures = 0;
}

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
void Asura_Visibility_System::AddConsoleCommands( void )
{
	Asura_CommandConsole::AddVar( L"Visibility.RenderEverything", &(s_bRenderEverything), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Visibility.FreezeCameraNode", &(s_bFreezeCameraNode), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Visibility.XRayMode", &(s_bXRayMode), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}
#endif // #ifdef ASURA_SUPPORT_COMMAND_CONSOLE

void Asura_Visibility_System::Update( void )
{
	if ( !s_pxVisibilityNodes ) return;

	ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__VISIBILITY_UPDATE, "Visibility Update");

	// First we find which node(s) the camera(s) are in. Unless it's outside the environment,
	// since they are represented by a point each camera should be inside a leaf node.

	// TODO: should have for loop over active cameras
	u_int uCameraIndex = Asura_Camera::GetCameraIndex();
	u_int uNode = s_auLastKnownCameraNode[ uCameraIndex ];

	if ( !s_bFreezeCameraNode )
	{
		FindNodeContainingPoint( *Asura_Camera::GetPosition(), uNode );
	}

	if( uNode != s_auLastKnownCameraNode[ uCameraIndex ] || s_bForceVisRecalcForCurrentCameraNode )
	{
		s_auLastKnownCameraNode[ uCameraIndex ] = uNode;
		s_bForceVisRecalcForCurrentCameraNode = false;
		
		Asura_Assert( s_auLastKnownCameraNode[ uCameraIndex ] < s_uNumberOfNodes, "Code bug: node out of range" );

		const Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ s_auLastKnownCameraNode[ uCameraIndex ] ];


		if ( xNode.GetLeftChild() )
		{
			// We didn't find a leaf node - either we have a bug or we may simply be outside
			// the environment. All we can do in this situation is make everything visible.
			
			s_auLastKnownCameraNode[ uCameraIndex ] = 0;
			s_apxCameraLeaf[ uCameraIndex ] = 0;
		}
		else
		{
			Asura_Assert( xNode.GetLeafIndex() < s_uNumberOfLeaves, "Code bug: leaf index out of range" );

			s_apxCameraLeaf[ uCameraIndex ] = &s_pxVisibilityLeaves[ xNode.GetLeafIndex() ];
		}
		
		UpdateViewIndependentVisibilties();
	}

	UpdateCameraVisibilities();

	ASURA_CPU_PROFILE__END(ASURA_CPU_PROFILE__VISIBILITY_UPDATE);
}


void Asura_Visibility_System::CreateNodeVisibilityFlags( void )
{
	Asura_Assert( s_pxNodeFlags == 0, "Already Initialised" );

	s_pxNodeFlags = new Asura_Visibility_NodeFlags [s_uNumberOfNodes];

	// No real need to initialise the data in the array, as it's dynamically created

	s_pucVisibleFeatures = new u_char [s_uNumberOfVisibilityFeatures];
}

void Asura_Visibility_System::UpdateCameraVisibilities( void )
{
	// KJFL - 19 February 2008 - 
	// Go through what the visibility system thinks could be visible from this point
	// and test it against the camera frustum. This is done in a hierachical manner
	// to reduce the number of calls to Asura_Camera::IsAABBInFrustum().

	// Blank all the ORed flags before we begin.
	for( u_int u=0; u<s_uNumberOfNodes; ++u)
	{
		s_pxNodeFlags[u].SetORedVisibilityFlags(0);
	}

	{
		u_int apxViewCellStack[32];
		u_int auClipBitfield[32];
		u_int uStack = 0;

		// push root onto stack
		apxViewCellStack[ uStack ] = 0;
		auClipBitfield[ uStack++ ] = 31;

		while ( uStack )
		{
			// pop something off the stack to work on
			const u_int uNodeIndex = apxViewCellStack[ --uStack ];
			u_int uClip = auClipBitfield[ uStack ];

			if ( s_pxNodeFlags[uNodeIndex].GetORedViewIndependentFlags()==0) continue;

			bool bInside;

			const Asura_Visibility_Node& xNode = Asura_Visibility_System::s_pxVisibilityNodes[ uNodeIndex ];
		
			if ( uClip )
			{
				bInside = Asura_Camera::IsAABBInFrustum( xNode.GetAABB(), uClip, uClip );
			}
			else
			{
				bInside = true;
			}
										
			if ( bInside )
			{
				s_pxNodeFlags[ uNodeIndex].SetORedVisibilityFlags( 1 );

				if ( xNode.GetLeftChild() )
				{
					auClipBitfield[ uStack ] = uClip;
					apxViewCellStack[ uStack++ ] = xNode.GetLeftChild();
					auClipBitfield[ uStack ] = uClip;
					apxViewCellStack[ uStack++ ] = xNode.GetRightChild();
				}
			}

		}

	}

	// We should now have correct ORed flags - at least for leaves. We need to recurse back up
	// the tree to check for validity and set the ANDed flags as we go too. Rather than using a tree
	// walk we can scan through the tree linearly, but backwards. This means we'll meet children
	// before parents, which is what we're after.
	Asura_Mem::PreFetchData(&s_pxVisibilityNodes[0], sizeof(s_pxVisibilityNodes[0]));

	for ( u_int u=s_uNumberOfNodes; u-->0; )
	{
		Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ u ];

		Asura_Mem::PreFetchData(&s_pxVisibilityNodes[u + 1], sizeof(s_pxVisibilityNodes[0]));

		if ( xNode.GetLeftChild() )
		{
			// We're not a leaf - we want to AND and OR together the camera flags of our children
			const u_int uLeftChild = xNode.GetLeftChild();
			const u_int uRightChild = uLeftChild + 1;

			s_pxNodeFlags[u].SetORedVisibilityFlags( s_pxNodeFlags[ uLeftChild ].GetORedVisibilityFlags() | s_pxNodeFlags[ uRightChild ].GetORedVisibilityFlags() );
			s_pxNodeFlags[u].SetANDedVisibilityFlags( s_pxNodeFlags[ uLeftChild ].GetANDedVisibilityFlags() & s_pxNodeFlags[ uRightChild ].GetANDedVisibilityFlags() );
		}
		else
		{
			// for a leaf, the ORed flags are already correct - and the ANDed flags should be made to match them
			s_pxNodeFlags[u].SetANDedVisibilityFlags( s_pxNodeFlags[u].GetORedVisibilityFlags() );
		}
	}

		
}

void Asura_Visibility_System::UpdateViewIndependentVisibilties( void )
{

	// Scan through the tree linearly, but backwards. This means we'll meet children before parents.

	for ( u_int u=s_uNumberOfNodes; u-->0; )
	{
		Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ u ];

		if ( xNode.GetLeftChild() )
		{
			// We're not a leaf - we want to AND and OR together the camera flags of our children
			const u_int uLeftChild = xNode.GetLeftChild();
			const u_int uRightChild = uLeftChild + 1;

			s_pxNodeFlags[u].SetORedViewIndependentFlags( s_pxNodeFlags[ uLeftChild ].GetORedViewIndependentFlags() | s_pxNodeFlags[ uRightChild ].GetORedViewIndependentFlags() );
			s_pxNodeFlags[u].SetANDedViewIndependentFlags( s_pxNodeFlags[ uLeftChild ].GetANDedViewIndependentFlags() & s_pxNodeFlags[ uRightChild ].GetANDedViewIndependentFlags() );
		}
		else
		{
			// We're a leaf - get our visibilities straight from the leaf the camera is in

			// TODO: should have for loop over active cameras
			const u_int uCameraIndex = Asura_Camera::GetCameraIndex();
			const u_char uCameraBitFlag = Asura_Camera::GetCameraIndexBitFlag();

			const Asura_Visibility_Leaf* pxLeaf = s_apxCameraLeaf[ uCameraIndex ];

			u_char ucFlags = 0x00; 
			
			if ( !pxLeaf )
			{
				// if we're outside the world, assume visible
				ucFlags = uCameraBitFlag;
			}
			else if ( pxLeaf->IsLeafVisible( xNode.GetLeafIndex() ) )
			{
				ucFlags = uCameraBitFlag;
			}

			s_pxNodeFlags[u].SetORedViewIndependentFlags( ucFlags );
			s_pxNodeFlags[u].SetANDedViewIndependentFlags( ucFlags );

		}
	}

	// now extract the feature visibilities 
	{
		// TODO: should have for loop over active cameras
		const u_int uCameraIndex = Asura_Camera::GetCameraIndex();
		const u_char uCameraBitFlag = Asura_Camera::GetCameraIndexBitFlag();
		const Asura_Visibility_Leaf* pxLeaf = s_apxCameraLeaf[ uCameraIndex ];
			
		for ( u_int u=0; u<s_uNumberOfVisibilityFeatures; ++u )
		{
			u_char uBitfield = 0;
	
			if ( !pxLeaf )
			{
				uBitfield |= uCameraBitFlag;
			}
			else if ( pxLeaf->IsFeatureVisible(u) )
			{
				uBitfield |= uCameraBitFlag;
			}

			s_pucVisibleFeatures[u] = uBitfield;
		}

	}



}

bool Asura_Visibility_System::IsAABBVisibleFromCurrentCamera( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode )
{
	return IsAABBVisible( xAABB, uLastKnownNode, Asura_Camera::GetCameraIndexBitFlag() );
}

bool Asura_Visibility_System::IsAABBVisible( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode, u_char ucCameraFlags, bool bUpdateNode )
{
	if (s_bRenderEverything) return true;

	// if we have no data, everything is visible
	if ( !s_pxVisibilityNodes ) return true;

	// Check node in range before using - may have recalc'd viewcells (or simply have dodgy data)
	if ( uLastKnownNode > s_uNumberOfNodes )
	{
		uLastKnownNode = 0;
		bUpdateNode = true;
	}
	
	bool bInsideEnvironment;
	
	if ( bUpdateNode )
	{
		bInsideEnvironment = FindNodeContainingAABB( xAABB, uLastKnownNode );
	}
	else 
	{
		// if the last known node was the root, flag it as outside the environment
		// which means we'll render it if it's in the view frustum
		bInsideEnvironment = ( uLastKnownNode != 0 );
	}

	if ( !bInsideEnvironment )
	{
		// AABB is not completely contained by the environment.
		// Currently we wish to say that's it's visible by default -
		// this copes with scenery mesh outside the world (it will be drawn
		// if its module is linked in). 
		return Asura_Camera::IsAABBInFrustum( xAABB );
	}
	else if ( s_pxNodeFlags[ uLastKnownNode ].GetANDedVisibilityFlags() & ucCameraFlags )
	{
		// definitely visible if in frustum
		return Asura_Camera::IsAABBInFrustum( xAABB );
	}
	else if ( s_pxNodeFlags[ uLastKnownNode ].GetORedVisibilityFlags() & ucCameraFlags )
	{
		// Could be visible - needs more checks to see if box does actually
		// overlap a completely visible node (i.e. one with ANDed flags set).

		// need to do a stack based search and visit all the node's children
		u_int auStack[32];
		u_int uStack = 0;

		// push root onto stack
		auStack[ uStack ] = uLastKnownNode;
		++uStack;

		while ( uStack )
		{
			// Pop something off the stack to work on. No need to check stack
			// size because creation routines won't allow invalid tree depths.
			const u_int uNodeIndex = auStack[ --uStack ];
			const Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ uNodeIndex ];

			if ( xNode.GetAABB().Intersects( xAABB ) )
			{
				// if this node is definitely visible, just check frustum
				if ( s_pxNodeFlags[uNodeIndex].GetANDedVisibilityFlags() & ucCameraFlags )
				{
					// definitely visible if in frustum
					return Asura_Camera::IsAABBInFrustum( xAABB );
				}
				else if ( s_pxNodeFlags[uNodeIndex].GetORedVisibilityFlags() & ucCameraFlags )
				{
					// otherwise let's keep checking the children
					if ( xNode.GetLeftChild() )
					{
						auStack[ uStack++ ] = xNode.GetLeftChild();
						auStack[ uStack++ ] = xNode.GetRightChild();
					}
				}
			}
		}

		return false;
	}
	else
	{
		// definitely not visible
		return false;
	}
}

bool Asura_Visibility_System::IsAABBVisibleIgnoringCameraDirection( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode, u_char ucCameraFlags, bool bUpdateNode )
{
	if (s_bRenderEverything) return true;

	// if we have no data, everything is visible
	if ( !s_pxVisibilityNodes ) return true;

	// Check node in range before using - may have recalc'd viewcells (or simply have dodgy data)
	if ( uLastKnownNode > s_uNumberOfNodes )
	{
		uLastKnownNode = 0;
		bUpdateNode = true;
	}
	
	bool bInsideEnvironment;
	
	if ( bUpdateNode )
	{
		bInsideEnvironment = FindNodeContainingAABB( xAABB, uLastKnownNode );
	}
	else 
	{
		// if the last known node was the root, flag it as outside the environment
		// which means we'll render it if it's in the view frustum
		bInsideEnvironment = ( uLastKnownNode != 0 );
	}

	if ( !bInsideEnvironment )
	{
		// AABB is not completely contained by the environment.
		// Currently we wish to say that's it's visible by default -
		return true;
	}
	else if ( s_pxNodeFlags[ uLastKnownNode ].GetANDedViewIndependentFlags() & ucCameraFlags )
	{
		// definitely visible if we're not testing frusta
		return true;
	}
	else if ( s_pxNodeFlags[ uLastKnownNode ].GetORedViewIndependentFlags() & ucCameraFlags )
	{
		// Could be visible - needs more checks to see if box does actually
		// overlap a completely visible node (i.e. one with ANDed flags set).

		// need to do a stack based search and visit all the node's children
		u_int auStack[32];
		u_int uStack = 0;

		// push root onto stack
		auStack[ uStack ] = uLastKnownNode;
		++uStack;

		while ( uStack )
		{
			// Pop something off the stack to work on. No need to check stack
			// size because creation routines won't allow invalid tree depths.
			const u_int uNodeIndex = auStack[ --uStack ];
			const Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ uNodeIndex ];

			if ( xNode.GetAABB().Intersects( xAABB ) )
			{
				// if this node is definitely visible, just check frustum
				if ( s_pxNodeFlags[uNodeIndex].GetANDedViewIndependentFlags() & ucCameraFlags )
				{
					// definitely visible
					return true;
				}
				else if ( s_pxNodeFlags[uNodeIndex].GetORedViewIndependentFlags() & ucCameraFlags )
				{
					// otherwise let's keep checking the children
					if ( xNode.GetLeftChild() )
					{
						auStack[ uStack++ ] = xNode.GetLeftChild();
						auStack[ uStack++ ] = xNode.GetRightChild();
					}
				}
			}
		}

		return false;
	}
	else
	{
		// definitely not visible
		return false;
	}
}

bool Asura_Visibility_System::IsModuleLinkedInFromCurrentCamera( u_int uModuleIndex )
{
	if ( uModuleIndex >= s_uNumberOfVisibilityFeatures - 1 )
	{
		Asura_Assert( false, "Code bug: invalid module index" );
		return false;
	}

	if (s_bRenderEverything) return true;

	// if we have no data, everything is visible
	if ( !s_pxVisibilityNodes ) return true;

	const u_char uCameraBitFlag = Asura_Camera::GetCameraIndexBitFlag();
	
	return (s_pucVisibleFeatures[ uModuleIndex ] & uCameraBitFlag) != 0;
}

bool Asura_Visibility_System::IsSkyboxVisibleFromCurrentCamera( void )
{
	if (s_bRenderEverything) return true;

	// if we have no data, everything is visible
	if ( !s_pxVisibilityNodes ) return true;

	Asura_Visibility_Leaf* pxLeaf = s_apxCameraLeaf[ Asura_Camera::GetCameraIndex() ];

	// outside the world - again it's visible
	if ( !pxLeaf ) return true;

	return pxLeaf->IsLeafVisible( s_uNumberOfLeaves - 1 );
}

void Asura_Visibility_System::FindNodeContainingPoint( const Asura_Vector_3& xPoint, u_int& uLastKnownNode )
{
	// Check node in range before using - may have recalc'd viewcells (or simply have dodgy data)
	if ( uLastKnownNode > s_uNumberOfNodes )
	{
		uLastKnownNode = 0;
	}

	// keep track of the last visited node to reduce duplicated tests
	u_int uLastVisitedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;

	do
	{
		const Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ uLastKnownNode ];

		// does this node contain the point in question?
		if ( xNode.GetAABB().Contains( xPoint ) )
		{
			// If we have children, see if they are a better match,
			// but don't bother tested a node we've already visited
			u_int uLeftNode = xNode.GetLeftChild();

			if ( uLeftNode )
			{
				u_int uRightNode = uLeftNode + 1;

				if ( (uLeftNode != uLastVisitedNode) && s_pxVisibilityNodes[ uLeftNode ].GetAABB().Contains( xPoint ) )
				{
					uLastVisitedNode = uLastKnownNode;
					uLastKnownNode = uLeftNode;
				}
				else if ( (uRightNode != uLastVisitedNode) && s_pxVisibilityNodes[ uRightNode ].GetAABB().Contains( xPoint ) )
				{
					uLastVisitedNode = uLastKnownNode;
					uLastKnownNode = uRightNode;
				}
				else 
				{
					Asura_Assert( false,"Code bug: bounding boxes invalid" );
					// Returning here means we at least have a sensible value
					// for uLastKnownNode (even if it's 0)
					return;
				}
			}
			else // we're inside a leaf node - job done
			{
				return;
			}
		}
		else // we have to try the parent
		{
			// if we're already at the root, there's nothing we can do
			if ( uLastKnownNode == 0 ) return;

			uLastVisitedNode = uLastKnownNode;
			uLastKnownNode = xNode.GetParent();
		}
	}
	while(true);
}

// returns true if AABB is wholly contained by root viewcell 
bool Asura_Visibility_System::FindNodeContainingAABB( const Asura_Bounding_Box& xAABB, u_int& uLastKnownNode )
{
    if( !s_pxVisibilityNodes )
    {
        return false;
    }

	// Check node in range before using - may have recalc'd viewcells (or simply have dodgy data)
	if ( uLastKnownNode > s_uNumberOfNodes )
	{
		uLastKnownNode = 0;
	}

	// keep track of the last visited node to reduce duplicated tests
	u_int uLastVisitedNode = uASURA_VISIBILITY_INVALID_NODE_INDEX;

	do
	{
		const Asura_Visibility_Node& xNode = s_pxVisibilityNodes[ uLastKnownNode ];

		// does this node contain the point in question?
		if ( xNode.GetAABB().Contains( xAABB ) )
		{
			// If we have children, see if they are a better match,
			// but don't bother tested a node we've already visited
			u_int uLeftNode = xNode.GetLeftChild();

			if ( uLeftNode )
			{
				u_int uRightNode = uLeftNode + 1;

				if ( (uLeftNode != uLastVisitedNode) && s_pxVisibilityNodes[ uLeftNode ].GetAABB().Contains( xAABB ) )
				{
					uLastVisitedNode = uLastKnownNode;
					uLastKnownNode = uLeftNode;
				}
				else if ( (uRightNode != uLastVisitedNode) && s_pxVisibilityNodes[ uRightNode ].GetAABB().Contains( xAABB ) )
				{
					uLastVisitedNode = uLastKnownNode;
					uLastKnownNode = uRightNode;
				}
				else 
				{
					// Unlike the point algo above, this is a valid situation where 
					// a bounding box is too big to fit into this node's children,
					// so we correctly return this node.
					return true;
				}
			}
			else // we're inside a leaf node - job done
			{
				return true;
			}
		}
		else // we have to try the parent
		{
			// if we're already at the root, there's nothing we can do
			// return false to indicate AABB sticks outside world
			if ( uLastKnownNode == 0 ) return false;

			uLastVisitedNode = uLastKnownNode;
			uLastKnownNode = xNode.GetParent();
		}
	}
	while(true);
}


bool Asura_Visibility_System::FindNodeContainingSphere( const Asura_Vector_3& xCentre, float fRadius, u_int& uLastKnownNode )
{
    Asura_Bounding_Box xAABB;
    xAABB.SetToPoint( xCentre );
    xAABB.Grow( fRadius );

    return FindNodeContainingAABB( xAABB, uLastKnownNode );
}


// This method provides a bridge to the old-style module-system setting of a RenderList array.
// Returns false if there was no data available.
void Asura_Visibility_System::GetVisibleFeaturesFromCamera( u_char* pucVisibleFeaturesArray, u_int uNumberOfFeatures )
{
	u_int uCameraIndex = Asura_Camera::GetCameraIndex();
	u_char ucCameraBitfield = Asura_Camera::GetCameraIndexBitFlag();

	if ( s_apxCameraLeaf[ uCameraIndex ] && !s_bRenderEverything )
	{
		s_apxCameraLeaf[ uCameraIndex ]->SetVisibleFeaturesArray( pucVisibleFeaturesArray, ucCameraBitfield, uNumberOfFeatures );
	}
	else
	{
		// outside the world? everything is visible

		Asura_Mem::Set( pucVisibleFeaturesArray, ucCameraBitfield, uNumberOfFeatures );
	}
}

#ifdef ASURA_TOOLS
void Asura_Visibility_System::Tools_DeleteModule( u_int uModuleIndex )
{
	// KJFL - 11 January 2008 - Standard behaviour of all the other
	// engine systems is to replace the module "uModuleIndex" with
	// the last module, and then decrement the number of modules.
	//
	// For this system we complicate matters slightly by storing
	// the skybox as our last feature, i.e. we have 1 more features 
	// than there are modules. However, the number of leaves and
	// nodes won't actually change, which makes things easier.

	if ( uModuleIndex >= s_uNumberOfVisibilityFeatures - 1 )
	{
		Asura_Assert( false, "Code bug: invalid module index when deleting a module" );
		return;
	}

	for ( u_int u=0; u<s_uNumberOfLeaves; ++u )
	{
		s_pxVisibilityLeaves[u].Tools_DeleteModule( uModuleIndex );
	}

	--s_uNumberOfVisibilityFeatures;
}
#endif

#endif // #ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM


#if defined(ASURA_USE_VISIBILITY_SYSTEM)

// This file's header.
#include "Visibility/Asura_Visibility_System.h"

#include "Camera/Asura_Camera.h"
#include "Visibility/Asura_Visibility_Node.h"
#include "Debugging/Asura_DebugText.h"

#ifdef ASURA_TOOLS
#include "Misc/Tools_Large_AABB_Tree.h"
#endif

// KJFL - 07 June 2006 - temporary measure of having a bool array of visible strips, to simplify
// queries using older IsEnvironmentStripVisible() method
static bool* g_pbVisible = 0;

static float GetSqdMaxDistFromPoint( const Asura_Bounding_Box& xBB, const Asura_Vector_3& xPoint ) 
{
    float fSqdDistance = 0.0f;
    float fDelta;

    if( xPoint.x < xBB.MinX )
    {
        fDelta = xPoint.x - xBB.MaxX;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xPoint.x > xBB.MaxX )
    {
        fDelta = xPoint.x - xBB.MinX;
        fSqdDistance += fDelta * fDelta;
    }

    if( xPoint.y < xBB.MinY )
    {
        fDelta = xPoint.y - xBB.MaxY;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xPoint.y > xBB.MaxY )
    {
        fDelta = xPoint.y - xBB.MinY;
        fSqdDistance += fDelta * fDelta;
    }

    if( xPoint.z < xBB.MinZ )
    {
        fDelta = xPoint.z - xBB.MaxZ;
        fSqdDistance += fDelta * fDelta;
    }
    else if( xPoint.z > xBB.MaxZ )
    {
        fDelta = xPoint.z - xBB.MinZ;
        fSqdDistance += fDelta * fDelta;
    }

    return fSqdDistance;
}

static const u_int uSTACK_SIZE = 2048;

static u_int g_auStack[uSTACK_SIZE];


u_int					Asura_Visibility_System::s_uNumberOfNodes;
Asura_Visibility_Node*	Asura_Visibility_System::s_pxNodes;

u_int					Asura_Visibility_System::s_uLastFrameStamp;
u_int					Asura_Visibility_System::s_uFrameStamp;

u_int					Asura_Visibility_System::s_uNumberOfStrips;
u_int					Asura_Visibility_System::s_uNumberOfVisibleStrips;
u_int*					Asura_Visibility_System::s_puVisibleStrips;


static u_int g_uNumberOfAABBTests;
static u_int g_uNumberOfStripTests;
static u_int g_uNumberOfFlushes;

void Asura_Visibility_System::Initialise( void )
{
	s_pxNodes = 0;

	s_uFrameStamp = 0;

	s_uNumberOfStrips = 0;
	s_uNumberOfVisibleStrips = 0;
	s_puVisibleStrips = 0;

	g_pbVisible = 0;
}

void Asura_Visibility_System::Reset( void )
{
	Shutdown();
	Initialise();
}

void Asura_Visibility_System::Shutdown( void )
{
	delete [] s_pxNodes;
	delete [] s_puVisibleStrips;

	delete [] g_pbVisible;
}


void Asura_Visibility_System::SortVisibleStripsOnCameraDistance( void )
{
	qsort( s_puVisibleStrips, s_uNumberOfVisibleStrips, sizeof(u_int), NodeSortOnCameraDistance );
}

void Asura_Visibility_System::SortVisibleStripsOnIndexOrder( void )
{
	qsort( s_puVisibleStrips, s_uNumberOfVisibleStrips, sizeof(u_int), NodeSortOnIndexOrder );
}

int Asura_Visibility_System::NodeSortOnCameraDistance( const void *pNode1, const void *pNode2 )
{
	// sort nodes so nearest strip to camera comes first

	const u_int* puNode1 = (const u_int*)pNode1;
	const u_int* puNode2 = (const u_int*)pNode2;

	const Asura_Visibility_Node& xVisibilityNode1 = s_pxNodes[*puNode1];
	const Asura_Visibility_Node& xVisibilityNode2 = s_pxNodes[*puNode2];

	// can probably cache this nicely
	const Asura_Vector_3 xCameraPosition = *Asura_Camera::GetPosition();
	float fDistance1 = Asura_Maths::Sqrt(xVisibilityNode1.m_xBB.GetSqdMinDistFromPoint( xCameraPosition ));
	float fDistance2 = Asura_Maths::Sqrt(xVisibilityNode2.m_xBB.GetSqdMinDistFromPoint( xCameraPosition ));

	if (fDistance1 == fDistance2)
	{

		fDistance1 = Asura_Maths::Sqrt( GetSqdMaxDistFromPoint ( xVisibilityNode1.m_xBB, xCameraPosition ));
		fDistance2 = Asura_Maths::Sqrt( GetSqdMaxDistFromPoint ( xVisibilityNode2.m_xBB,  xCameraPosition ));


	}


	if (fDistance1 > fDistance2)
		return  1;
	if (fDistance1 < fDistance2)
		return -1;

	return 0;
}

int Asura_Visibility_System::NodeSortOnIndexOrder( const void *pNode1, const void *pNode2 )
{
	// sort nodes so lowest index comes first

	const u_int* puNode1 = (const u_int*)pNode1;
	const u_int* puNode2 = (const u_int*)pNode2;

	const Asura_Visibility_Node& xVisibilityNode1 = s_pxNodes[*puNode1];
	const Asura_Visibility_Node& xVisibilityNode2 = s_pxNodes[*puNode2];

	if (xVisibilityNode1.m_uStrip > xVisibilityNode2.m_uStrip)
		return  1;
	if (xVisibilityNode1.m_uStrip < xVisibilityNode2.m_uStrip)
		return -1;

	return 0;
}

void Asura_Visibility_System::Update( void ) 
{
	if ( s_uNumberOfStrips == 0 || s_uNumberOfNodes == 0)
	{
		Asura_Assert( false, "No Visibility data present." );
		return;
	}

	if( !g_pbVisible ) g_pbVisible = new bool [ s_uNumberOfStrips ];

    g_uNumberOfAABBTests = 0;
	g_uNumberOfStripTests = 0;
	g_uNumberOfFlushes = 0;

	s_uLastFrameStamp = s_uFrameStamp++;

	const Asura_Vector_3 xCameraPosition = *Asura_Camera::GetPosition();
	
	// first off, stick down a z pass of what was previously visible
#if 1
	SortVisibleStripsOnCameraDistance();
	for (u_int u=0; u<s_uNumberOfVisibleStrips; u++)
	{
		const u_int uNodeIndex = s_puVisibleStrips[u];
		Asura_Visibility_Node& xVisibilityNode = s_pxNodes[uNodeIndex];


		Platform_RenderStripZOnly( xVisibilityNode.m_uStrip );
	}
#else
	if ( s_uNumberOfVisibleStrips )
	{
		Platform_RenderZPassOfVisibleStrips();
	}
#endif

	s_uNumberOfVisibleStrips = 0;
	
	u_int* puStack = &g_auStack[0];

	*puStack++ = 0;

	while( puStack != &g_auStack[0] || Platform_GetNumberOfOutstandingOcclusionQueries() )
	{
		Asura_Assert( static_cast<u_int>(puStack - g_auStack) < uSTACK_SIZE, "Stack overflow danger" );

		bool bForceFlush = false;

		if (puStack == &g_auStack[0])
		{
			// if the stack is empty, we have no new nodes to consider
			// if we're still waiting on occlusion query results, we'll have to flush them out
			bForceFlush = true;

			g_uNumberOfFlushes++;
		}
		else
		{
			// pop a node off the stack, so we can consider it
			const u_int uNodeIndex = *--puStack;

			Asura_Visibility_Node& xVisibilityNode = s_pxNodes[uNodeIndex];

			bool bPreviouslyVisible = (xVisibilityNode.m_bVisible);// && (xVisibilityNode.m_uFrameStamp == s_uLastFrameStamp);

			bool bIsLeaf = xVisibilityNode.m_uStrip != 0xffffffff;
			
			// Have we considered this node already this frame? If so, skip it.
			if ( s_uFrameStamp == xVisibilityNode.m_uFrameStamp )
			{
			//	bPreviouslyVisible = true;
			}
			// Otherwise, if this node was not previously visible, OR it's been considered visible for a period of time, assume
			// it's invisible and fire off a query to test it.
			//
			// KJFL - 17 November 2005 - Using a 'random' time period for retests (e.g. constant + the node index number modulo
			// constant), so the retests are hopefully spread out more for a given scene, causes flickering. Haven't figured out
			// why yet... simply keeping the strip visible for say 30 frames and then retesting gives a good speed increase 
			// though, so I'm using that for now. 
			else if ( (!bPreviouslyVisible) || ( (s_uFrameStamp - xVisibilityNode.m_uFrameStamp) >= (20+(uNodeIndex&31)) ))
			{
				// assume invis until proven otherwise
				xVisibilityNode.m_bVisible = false;
				xVisibilityNode.m_uFrameStamp = s_uFrameStamp;
				xVisibilityNode.m_bPreviouslyVisible = bPreviouslyVisible;

				// if it's a leaf, we need to query using the strip triangles
				if ( bIsLeaf )
				{
					++g_uNumberOfStripTests;
					Platform_IssueOcclusionQuery( uNodeIndex, xVisibilityNode.m_uStrip );
				}
				// If it's a not a leaf node, and it was previously invisible, we need
				// to issue a query (unless the camera is inside the node, in which case
				// we'll take it as visible straight off).
				else //if ( !bPreviouslyVisible )
				{
					// If we're positioned inside the bb of the node (or it's the root node)
					// then assume it is visible.
					if ( xVisibilityNode.m_xBB.Contains( xCameraPosition ) || uNodeIndex == 0 )
					{
						bPreviouslyVisible = true;
						xVisibilityNode.m_bVisible = true;
					}
					// otherwise if it's in the frustum we'll need a occlusion query
					else if ( Asura_Camera::IsAABBInFrustum( xVisibilityNode.m_xBB ))
					{
						++g_uNumberOfAABBTests;
						Platform_IssueOcclusionQuery( uNodeIndex, xVisibilityNode.m_xBB );
						bPreviouslyVisible = false;
					}
				}
			}
			else if ( bIsLeaf && bPreviouslyVisible )
			{
			//	Asura_Direct3D_EnvironmentRenderer::RenderStripZOnly ( xVisibilityNode.m_uStrip );
				s_puVisibleStrips[ s_uNumberOfVisibleStrips++ ]  = uNodeIndex;
			}
		

			// non leaf node, consider the children
			if ( !bIsLeaf && bPreviouslyVisible ) 
			{
				const Asura_Visibility_Node& xLeftNode = s_pxNodes[ xVisibilityNode.m_uLeftChild ];
				const Asura_Visibility_Node& xRightNode = s_pxNodes[ xVisibilityNode.m_uRightChild ];

				const Asura_Bounding_Box& xLeftBB = xLeftNode.m_xBB;
				const Asura_Bounding_Box& xRightBB = xRightNode.m_xBB;
		
				float fLeftDistance = Asura_Maths::Sqrt(xLeftBB.GetSqdMinDistFromPoint( xCameraPosition ));
				bool bLeftHit = Asura_Camera::IsAABBInFrustum(  xLeftBB ) || xLeftBB.Contains (xCameraPosition);

				float fRightDistance = Asura_Maths::Sqrt(xRightBB.GetSqdMinDistFromPoint( xCameraPosition ));
				bool bRightHit = Asura_Camera::IsAABBInFrustum( xRightBB ) || xRightBB.Contains( xCameraPosition);

				if (fLeftDistance == fRightDistance)
				{
	#if 01
					//				fLeftDistance = GetSqdMaxDistFromPoint( xLeftBB, xCameraPosition );
	//				fRightDistance = GetSqdMaxDistFromPoint( xRightBB, xCameraPosition );

				//	if (fLeftDistance != 0.0f)
					{
						Asura_Vector_3 xForwards;

						Asura_Camera::GetForwardVector( xForwards );

						Asura_Vector_3 xLeft, xRight;

						if (xForwards.x > 0)
						{
							xLeft.x  = xLeftBB.MaxX  - xCameraPosition.x;
							xRight.x = xRightBB.MaxX - xCameraPosition.x;
						}
						else
						{
							xLeft.x  = xLeftBB.MinX  - xCameraPosition.x;
							xRight.x = xRightBB.MinX - xCameraPosition.x;
						}
						if (xForwards.y > 0)
						{
							xLeft.y  = xLeftBB.MaxY  - xCameraPosition.y;
							xRight.y = xRightBB.MaxY - xCameraPosition.y;
						}
						else
						{
							xLeft.y  = xLeftBB.MinY  - xCameraPosition.y;
							xRight.y = xRightBB.MinY - xCameraPosition.y;
						}
						if (xForwards.z > 0)
						{
							xLeft.z  = xLeftBB.MaxZ  - xCameraPosition.z;
							xRight.z = xRightBB.MaxZ - xCameraPosition.z;
						}
						else
						{
							xLeft.z  = xLeftBB.MinZ  - xCameraPosition.z;
							xRight.z = xRightBB.MinZ - xCameraPosition.z;
						}

						fLeftDistance = xForwards * xLeft;
						fRightDistance = xForwards * xRight;
					}

			#endif		
	//				
				}

				const bool bLeftFirst = (bLeftHit && bRightHit) && (fLeftDistance<fRightDistance);

				//  left node first?
				if (bLeftHit && !bLeftFirst)
				{
					*puStack++ =  xVisibilityNode.m_uLeftChild ;
				}
			
				// if we're still considering right node, add it to the stack
				if (bRightHit)
				{
					*puStack++ =  xVisibilityNode.m_uRightChild ;
				}

				// and similarly for the left node
				if (bLeftHit && bLeftFirst)
				{
					*puStack++ =  xVisibilityNode.m_uLeftChild ;
				}
			}
		}

		if ( Platform_GetNumberOfOutstandingOcclusionQueries() )
		{
			u_int uNodeIndex;

			while( Platform_GetOcclusionResults( uNodeIndex, bForceFlush ) )
			{
				bForceFlush = false;

				bool bIsLeaf = s_pxNodes[uNodeIndex].m_uStrip != 0xffffffff;

				u_int uNode = uNodeIndex;

				// scan upwards, setting all the node's parents to be also visible
				// (logically, non of the parent nodes will be leaves)
				while( uNode!=0 && !s_pxNodes[uNode].m_bVisible )
				{
					s_pxNodes[uNode].m_bVisible = true;
					uNode = s_pxNodes[uNode].m_uParent;
				}				
			
				// if this node is a leaf, add it to the render list
				if ( bIsLeaf )
				{
					s_puVisibleStrips[ s_uNumberOfVisibleStrips++ ]  = uNodeIndex;
				}
				// otherwise add the node to the stack, so it's children can be considered
				else //if (!s_pxNodes[uNodeIndex].m_bPreviouslyVisible)
				{
					*puStack++ = uNodeIndex;
				}
		
			}
		}

	}
	SortVisibleStripsOnIndexOrder();

	for (u_int u=0; u<s_uNumberOfStrips; u++)
	{
		g_pbVisible[ u ] = false;
	}
	for (u_int u=0; u<s_uNumberOfVisibleStrips; u++)
	{
		g_pbVisible[ s_pxNodes[ s_puVisibleStrips[u] ].m_uStrip ] = true;
	}


#if 0

	Asura_DebugText::AddTextBlock( 5,3, 0xffffffff, 0, "TotalNumberOfStrips: %d", s_uNumberOfStrips );

	Asura_DebugText::AddTextBlock( 5,5, 0xffffffff, 0, "NumberOfAABBTests %d", g_uNumberOfAABBTests );
	Asura_DebugText::AddTextBlock( 5,6, 0xffffffff, 0, "NumberOfStripTests %d", g_uNumberOfStripTests );
	Asura_DebugText::AddTextBlock( 5,7, 0xffffffff, 0, "NumberOfRenderedStrips %d", s_uNumberOfVisibleStrips );
	Asura_DebugText::AddTextBlock( 5,8, 0xffffffff, 0, "NumberOfFlushes %d", g_uNumberOfFlushes );

	
#endif
}

bool Asura_Visibility_System::IsEnvironmentStripVisible( u_int uStripID )
{
	if ( !g_pbVisible ) return false;

	return g_pbVisible[uStripID];
}

const Asura_Bounding_Box& Asura_Visibility_System::GetStripBB( u_int uStripID )
{
	Asura_Visibility_Node& xNode = s_pxNodes[ (( s_uNumberOfNodes - 1 ) / 2) + uStripID];

	Asura_Assert( xNode.m_uStrip == uStripID, "Error, strip mismatch" );

	return xNode.m_xBB;
}

u_int Asura_Visibility_System::GetVisibleStrip( const u_int uVisibleStripIndex )
{
	// Return the index of i'th visible environment strip

	Asura_Assert( uVisibleStripIndex < s_uNumberOfNodes, "Invalid parameter passed to Asura_Visibility_System::GetVisibleStrip()." );

	const u_int uNodeIndex = s_puVisibleStrips[ uVisibleStripIndex ];
	Asura_Assert( uNodeIndex < s_uNumberOfNodes, "Invalid node index in visible strip set." );

	const u_int uStripIndex = s_pxNodes[ uNodeIndex ].m_uStrip;
	Asura_Assert( uStripIndex != 0xFFFFFFFF, "Non-leaf found in visible strip set." );

	return uStripIndex;
}

#ifdef ASURA_TOOLS
void Asura_Visibility_System::Tools_Create( const u_int uNumberOfStrips, Asura_Bounding_Box* pxStripBBs, const Asura_Vector_3* pxStripCentres )
{
	Reset();

	s_uNumberOfStrips = uNumberOfStrips;
	if(!s_uNumberOfStrips)
	{
		//If there are no strips , then give up now , since the code below will get upset about this. RAR 31/05/2006
		s_uNumberOfNodes = 0;
		return;
	}

	for ( u_int u=0; u<uNumberOfStrips; ++u )
	{
		pxStripBBs[u].Grow(0.01f);
	}


	Tools_Large_AABB_Tree* pxTree = Tools_Large_AABB_Tree_Creator::Create( uNumberOfStrips, pxStripBBs, pxStripCentres );


	s_uNumberOfNodes = uNumberOfStrips * 2 - 1;

	s_pxNodes = new Asura_Visibility_Node[ s_uNumberOfNodes ];

	s_puVisibleStrips = new u_int[ s_uNumberOfNodes ];

	s_pxNodes[0].m_uParent = 0;
	s_pxNodes[0].m_xBB = pxTree->RootBoundingBox;

	for (u_int u=0; u<uNumberOfStrips - 1; u++)
	{

		Tools_Large_AABB_Tree_Node& xTreeNode = pxTree->Nodes[u];

		s_pxNodes[u].m_bVisible = true;
		s_pxNodes[u].m_uFrameStamp = 0;
		s_pxNodes[u].m_uStrip = 0xffffffff;

		s_pxNodes[u].m_uLeftChild = xTreeNode.LeftIndex;
		s_pxNodes[u].m_uRightChild = xTreeNode.RightIndex;

		if (xTreeNode.LeftIsLeaf)
		{
			s_pxNodes[u].m_uLeftChild += uNumberOfStrips - 1;
		}
		if (xTreeNode.RightIsLeaf)
		{
			s_pxNodes[u].m_uRightChild += uNumberOfStrips - 1;
		}

		s_pxNodes[ s_pxNodes[u].m_uLeftChild ].m_uParent = u;
		s_pxNodes[ s_pxNodes[u].m_uRightChild ].m_uParent = u;

		s_pxNodes[ u ].m_xBB.Reset();
	}

	for (u_int u=0; u<uNumberOfStrips; u++)
	{
		u_int uVisNode = u + uNumberOfStrips - 1;

		s_pxNodes[ uVisNode ].m_bVisible = false;
		s_pxNodes[ uVisNode ].m_uFrameStamp = 0;
		s_pxNodes[ uVisNode ].m_uStrip = u;

		s_pxNodes[ uVisNode ].m_xBB = pxStripBBs[ u ];

	}

	for (u_int u=0; u<uNumberOfStrips - 1; u++)
	{
		u_int* puStack = &g_auStack[0];

		*puStack++ = u;
		
		while ( puStack != &g_auStack[0])
		{
			u_int uNode = *--puStack;

			if (s_pxNodes[ uNode ].m_uStrip != 0xffffffff)
			{
				s_pxNodes[u].m_xBB.Set( s_pxNodes[ uNode ].m_xBB );
			}
			else
			{
				*puStack++ = s_pxNodes[uNode].m_uLeftChild;
				*puStack++ = s_pxNodes[uNode].m_uRightChild;
			}
		}
	}

	delete pxTree;
	
}
#endif //#ifdef ASURA_TOOLS


#endif	// defined(ASURA_USE_VISIBILITY_SYSTEM)


/// Emd of file. ///
