//////////////////////////////////////////////////////////////////////
// Asura_Volume_Cuboid												//
// SJL 6/11/07														//
// Copyright (C) 2007 Rebellion, All Rights Reserved.			    //
//																	//
// An oriented cuboid volume class								    //
//////////////////////////////////////////////////////////////////////

// Main asura file
#include "Core/Asura.h"

// This file's header.
#include "Volume/Asura_Volume_Cuboid.h"

// Asura headers
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_OrientedVolume.h"
#if !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Debugging/Asura_DebugLineRender.h"
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)

// Tools headers
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

// Static class members
const u_int Asura_Volume_Cuboid::s_uCHUNKVERSION = 0;

Asura_Volume_Cuboid::Asura_Volume_Cuboid()
{
	m_iClassification = ASURA_VOLUME_CLASSIFICATION_CUBOID;

	m_xBox.Zero();
	m_xBox.Grow(1.5f);
	m_xOrient.Identity();
}

Asura_Volume_Cuboid::Asura_Volume_Cuboid(const Asura_Bounding_Box& xBox, const Asura_Quat& xOrient)
: m_xOrient(xOrient)
, m_xBox(xBox)
{
	m_iClassification = ASURA_VOLUME_CLASSIFICATION_CUBOID;
}


Asura_Volume_Cuboid::~Asura_Volume_Cuboid()
{
	
}


void Asura_Volume_Cuboid::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	// Classification should be read out already by the container class / whatever code created this object

	// Read and check version number
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert(uVersion == s_uCHUNKVERSION, "Unknown chunk version loading Asura_Volume_Cuboid");
	if (uVersion != s_uCHUNKVERSION ) return;

	// Read data
	xStream >> m_xBox;
	xStream >> m_xOrient;
}

void Asura_Volume_Cuboid::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	// Write classification
	xStream << m_iClassification;

	// Write version number
	xStream << s_uCHUNKVERSION;

	// Write data
	xStream << m_xBox;
	xStream << m_xOrient;
}

bool Asura_Volume_Cuboid::Contains(const Asura_Vector_3& xPosition) const
{
	//see if the orientation is being used
	if(m_xOrient.w == 1.0f)
	{
		//nope , can just do simple test
		return m_xBox.Contains(xPosition);
	}
	else
	{
		//need to get position relative to volume
		Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
		Asura_Vector_3 xRelative = xPosition - xCentre;

		//and rotate into the space of the volume
		Asura_Quat xOrientInverse = m_xOrient;
		xOrientInverse.Invert();
		xRelative = xOrientInverse.RotateVector(xRelative);

		return m_xBox.Contains(xRelative+xCentre);
	}
}

bool Asura_Volume_Cuboid::Overlaps(const Asura_Bounding_Box& xBB) const
{
	return Asura_OrientedVolume::Overlaps(m_xOrient, m_xBox, xBB);
}

float Asura_Volume_Cuboid::GetSqdMinDistFromPoint(const Asura_Vector_3& xPosition) const
{
	if (m_xOrient.w == 1.0f)
	{
		return m_xBox.GetSqdMinDistFromPoint(xPosition);
	}
	else
	{
		//need to get position relative to volume
		Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
		Asura_Vector_3 xRelative = xPosition - xCentre;

		//and rotate into the space of the volume
		Asura_Quat xOrientInverse = m_xOrient;
		xOrientInverse.Invert();
		xRelative = xOrientInverse.RotateVector(xRelative);

		return m_xBox.GetSqdMinDistFromPoint(xRelative+xCentre);
	}
}

void Asura_Volume_Cuboid::CalcMinVectorIntoVolume( const Asura_Vector_3& xPosition, Asura_Vector_3& xVector ) const
{
	if ( m_xOrient.w == 1.0f )
	{
		m_xBox.CalcMinVectorIntoBox( xPosition, xVector );
	}
	else
	{
		// need to get position relative to volume
		Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
		Asura_Vector_3 xRelative = xPosition - xCentre;

		// and rotate into the space of the volume
		Asura_Quat xOrientInverse = m_xOrient;
		xOrientInverse.Invert();
		xRelative = xOrientInverse.RotateVector(xRelative);

		m_xBox.CalcMinVectorIntoBox( xRelative+xCentre, xVector );

		// and finally rotate the result back into world space
		xVector = m_xOrient.RotateVector( xVector);
	}
}

Asura_Volume* Asura_Volume_Cuboid::Duplicate() const
{
	Asura_Volume_Cuboid* pxNew = new Asura_Volume_Cuboid();
	pxNew->Copy(this);
	return static_cast<Asura_Volume*>(pxNew);
}

void Asura_Volume_Cuboid::Copy(const Asura_Volume_Cuboid* pxFrom)
{
    m_xBox	  = pxFrom->m_xBox;
	m_xOrient = pxFrom->m_xOrient;
}

void Asura_Volume_Cuboid::GetAABB(Asura_Bounding_Box& xBB) const
{
	// Re-centre the box about the origin.
	const Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
	Asura_Bounding_Box xCentredBB = m_xBox;
	xCentredBB.Displace(-xCentre);
	// Rotate it.
	Asura_Matrix_3x3 xOrient;
	m_xOrient.ToMatrix(xOrient);
	xBB.FromOBB(xCentredBB, xOrient);
	// Displace back to the original position.
	xBB.Displace(xCentre);
}

#if !defined(ASURA_NO_DEBUG_OUTPUT)
void Asura_Volume_Cuboid::AddToLineSet(const u_int uARGB,
									   Asura_DebugLineRender_LineSet& xLineSet) const
{
	xLineSet.AddBoundingBox(m_xBox, Asura_Maths::ZeroVector3, m_xOrient, uARGB);
}
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)

#ifdef ASURA_TOOLS
void Asura_Volume_Cuboid::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
    Asura_Vector_3 xCentre = m_xBox.CalculateCentre();
    Asura_Vector_3 xNewCentre = pxInstance->TransformPosition( xCentre );
    m_xBox.Displace( xNewCentre - xCentre );

    m_xOrient = pxInstance->TransformQuat( m_xOrient );

}
#endif //ASURA_TOOLS

void Asura_Volume_Cuboid::RotateAbout(const Asura_Vector_3& xCentre, const Asura_Matrix_3x3& xOri)
{
	Asura_Matrix_3x3 xCur;
	m_xOrient.ToMatrix(xCur);
	xCur = xOri * xCur;
	m_xOrient.FromMatrix(xCur);

	Asura_Vector_3 xOldCentre = m_xBox.CalculateCentre();
	Asura_Vector_3 xPos = xOldCentre - xCentre;
	xPos = xPos * xOri;
	xPos += xCentre;

	Asura_Vector_3 xDisp = xPos - xOldCentre;
	m_xBox.Displace( xDisp );

}

