//////////////////////////////////////////////////////////////////////
// Asura_Volume_Sphere												//
// SJL 19/11/07														//
// Copyright (C) 2007 Rebellion, All Rights Reserved.			    //
//																	//
// An oriented sphere volume class								    //
//////////////////////////////////////////////////////////////////////

// Main asura file
#include "Core/Asura.h"

// This file's header.
#include "Volume/Asura_Volume_Sphere.h"

// Asura headers
#include "ChunkSystem/Asura_Chunk_Stream.h"
#if !defined(ASURA_NO_DEBUG_OUTPUT)
#include "Debugging/Asura_DebugLineRender.h"
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)

// Tools headers
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif


// Static class members
const u_int Asura_Volume_Sphere::s_uCHUNKVERSION = 0;

Asura_Volume_Sphere::Asura_Volume_Sphere()
{
	m_iClassification = ASURA_VOLUME_CLASSIFICATION_SPHERE;
	m_fRadius = 1.0f;

	m_xOrient.Identity();
}

Asura_Volume_Sphere::~Asura_Volume_Sphere()
{

}

void Asura_Volume_Sphere::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	// Classification should be read out already by the container class / whatever code created this object

	// Read and check version number
	u_int uVersion;
	xStream >> uVersion;
	Asura_Assert(uVersion == s_uCHUNKVERSION, "Unknown chunk version loading Asura_Volume_Sphere");
	if (uVersion != s_uCHUNKVERSION ) return;

	// Read data
	xStream >> m_xPosition;
	xStream >> m_fRadius;
	xStream >> m_xOrient;
}

void Asura_Volume_Sphere::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	// Write classification
	xStream << m_iClassification;

	// Write version number
	xStream << s_uCHUNKVERSION;

	// Write data
	xStream << m_xPosition;
	xStream << m_fRadius;
	xStream << m_xOrient;
}

bool Asura_Volume_Sphere::Contains(const Asura_Vector_3& xPosition) const
{
	float fRadiusSqr = m_fRadius*m_fRadius;
	float fToSqr = (m_xPosition-xPosition).MagnitudeSquared();

	if (fToSqr < fRadiusSqr) 
	{
		return true;
	}
	else
	{
		return false;
	}
}

bool Asura_Volume_Sphere::Overlaps(const Asura_Bounding_Box& xBB) const
{
	float fMinDistSquared = 0;

	if( m_xPosition.x < xBB.MinX)
	{
		float fDiff = m_xPosition.x-xBB.MinX;
		fMinDistSquared += fDiff*fDiff;
	}
	else if( m_xPosition.x > xBB.MaxX)
	{
		float fDiff = m_xPosition.x-xBB.MaxX;
		fMinDistSquared += fDiff*fDiff;
	}

	if( m_xPosition.y < xBB.MinY)
	{
		float fDiff = m_xPosition.y-xBB.MinY;
		fMinDistSquared += fDiff*fDiff;
	}
	else if( m_xPosition.y > xBB.MaxY)
	{
		float fDiff = m_xPosition.y-xBB.MaxY;
		fMinDistSquared += fDiff*fDiff;
	}

	if( m_xPosition.z < xBB.MinZ)
	{
		float fDiff = m_xPosition.z-xBB.MinZ;
		fMinDistSquared += fDiff*fDiff;
	}
	else if( m_xPosition.z > xBB.MaxZ)
	{
		float fDiff = m_xPosition.z-xBB.MaxZ;
		fMinDistSquared += fDiff*fDiff;
	}

	return fMinDistSquared <= m_fRadius*m_fRadius;

}

Asura_Volume* Asura_Volume_Sphere::Duplicate() const
{
	Asura_Volume_Sphere* pxNew = new Asura_Volume_Sphere();
	pxNew->Copy(this);
	return static_cast<Asura_Volume*>(pxNew);
}

void Asura_Volume_Sphere::Copy(const Asura_Volume_Sphere* pxFrom)
{
   m_xPosition = pxFrom->m_xPosition;
   m_fRadius = pxFrom->m_fRadius;
   m_xOrient = pxFrom->m_xOrient;
}

void Asura_Volume_Sphere::GetAABB(Asura_Bounding_Box& xBB) const
{
	xBB.Zero();
	xBB.Grow(m_fRadius);
	xBB.Displace(m_xPosition);
}

#if !defined(ASURA_NO_DEBUG_OUTPUT)
void Asura_Volume_Sphere::AddToLineSet(const u_int uARGB,
									   Asura_DebugLineRender_LineSet& xLineSet) const
{
	xLineSet.AddSphere(m_xPosition, m_fRadius, uARGB);
}
#endif	//!defined(ASURA_NO_DEBUG_OUTPUT)

#ifdef ASURA_TOOLS
void Asura_Volume_Sphere::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
   	m_xPosition = pxInstance->TransformPosition( m_xPosition );
    m_xOrient = pxInstance->TransformQuat( m_xOrient );
}
#endif //ASURA_TOOLS

void Asura_Volume_Sphere::RotateAbout(const Asura_Vector_3& xCentre, const Asura_Matrix_3x3& xOri)
{
	Asura_Matrix_3x3 xCur;
	m_xOrient.ToMatrix(xCur);
	xCur = xOri * xCur;
	m_xOrient.FromMatrix(xCur);

	m_xPosition -= xCentre;
	m_xPosition = m_xPosition * xOri;
	m_xPosition += xCentre;
}
