//
// Asura_Zone_Controller.cpp
//
// Created by Richard Rice - 25/01/05
// 
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Zone/Asura_Zone_Controller.h"

#include "Core/Asura_AABB_Tree.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Zone/Asura_Zone.h"


#define ASURA_ZONE_CONTROLLER_VERSION 1

//for use in a couple of callback functions
static const Asura_Zone_Controller* g_pxCurrentZoneController = 0;
static Asura_Vector_3 g_xTestPosition;
static int g_iFoundZoneIndex;
static float g_fFoundZonePriority;


Asura_Zone_Controller::Asura_Zone_Controller()
{
	Initialise();
}

void Asura_Zone_Controller::Initialise()
{
	m_iNumberOfZones = 0;
	m_ppxZones = 0;
	m_iPreviousZoneIndex = -1;
	m_pxVolumeTree = 0;
	m_iNumberOfVolumes = 0;
	#ifdef ASURA_TOOLS
	m_bVolumeTreeValid = false;
	#endif
}

Asura_Zone_Controller::~Asura_Zone_Controller()
{
	Shutdown();
}

void Asura_Zone_Controller::Shutdown()
{
	for(int i=0;i<m_iNumberOfZones;i++)
	{
		delete m_ppxZones[i];
	}
	delete [] m_ppxZones;
	delete m_pxVolumeTree;
}

void Asura_Zone_Controller::Reset()
{
	Shutdown();
	Initialise();
}

const Asura_Zone* Asura_Zone_Controller::GetZoneFromPosition(const Asura_Vector_3& xPos) const
{
	if(UseMostRecentZoneByDefault())
	{
		if(m_iPreviousZoneIndex>=0 && m_iPreviousZoneIndex<m_iNumberOfZones)
		{
			//check the most recent zone
			if(m_ppxZones[m_iPreviousZoneIndex]->Contains(xPos))
			{
				return m_ppxZones[m_iPreviousZoneIndex];
			}
		}
	}

	#ifdef ASURA_TOOLS
	if(!m_bVolumeTreeValid) UpdateVolumeTree();
	#endif

	//Check the aabb tree for a suitable zone
	if(m_pxVolumeTree)
	{
		g_xTestPosition = xPos;
		g_pxCurrentZoneController = this;
		g_iFoundZoneIndex = -1;
		g_fFoundZonePriority = 0.0f;

		Asura_Bounding_Box xTestBox;
		xTestBox.MinX = xTestBox.MaxX = g_xTestPosition.x;
		xTestBox.MinY = xTestBox.MaxY = g_xTestPosition.y;
		xTestBox.MinZ = xTestBox.MaxZ = g_xTestPosition.z;


		m_pxVolumeTree->TestTree(xTestBox,VolumeTreeCallback);
		g_pxCurrentZoneController = 0;

		if(g_iFoundZoneIndex>=0)
		{
			m_iPreviousZoneIndex = g_iFoundZoneIndex;
			return m_ppxZones[m_iPreviousZoneIndex];
		}
	}


	//not in any zone.. is the first one a default zone?
	if(m_iNumberOfZones && m_ppxZones[0]->IsDefaultZone())
	{
		return m_ppxZones[0];
	}
	return 0;
}

const Asura_Zone* Asura_Zone_Controller::GetZoneByID( const u_int uZoneID ) const
{
    for ( int i = 0; i < m_iNumberOfZones; ++i )
    {
        if ( m_ppxZones[ i ]->GetZoneID() == uZoneID )
        {
            return m_ppxZones[ i ];
        }
    }

    return NULL;
}

Asura_Zone* Asura_Zone_Controller::GetZoneByID( const u_int uZoneID )
{
    for ( int i = 0; i < m_iNumberOfZones; ++i )
    {
        if ( m_ppxZones[ i ]->GetZoneID() == uZoneID )
        {
            return m_ppxZones[ i ];
        }
    }

    return NULL;
}


const Asura_Zone* Asura_Zone_Controller::GetZoneByIndex(int iIndex) const
{
	Asura_Assert(iIndex>=0 && iIndex<m_iNumberOfZones,"Invalid zone index");
	return m_ppxZones[iIndex];
}

Asura_Zone* Asura_Zone_Controller::GetZoneByIndex(int iIndex)
{
	Asura_Assert(iIndex>=0 && iIndex<m_iNumberOfZones,"Invalid zone index");
	return m_ppxZones[iIndex];
}


void Asura_Zone_Controller::InitialiseFromChunkBuffer(const char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	Asura_Zone_Controller::ReadFromChunkStream(xStream);
}

void Asura_Zone_Controller::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	int iVersion;
	xStream >> iVersion;

	Asura_Assertf(iVersion<=ASURA_ZONE_CONTROLLER_VERSION,("Unknown chunk version %d in Asura_Zone_Controller",iVersion));

	int iNumberOfZones;
	xStream >> iNumberOfZones;

	if(iNumberOfZones)
	{
		m_ppxZones = new Asura_Zone*[iNumberOfZones];

		for(int i=0;i<iNumberOfZones;i++)
		{
			m_ppxZones[i] = AllocateZone( true );
			Asura_Assert(m_ppxZones[i],"");

			m_ppxZones[i]->ReadFromChunkStream(xStream);	
		}
	}
	m_iNumberOfZones = iNumberOfZones;

	if(iVersion >= 1)
	{
		int iNumberOfVolumes;
		xStream >> iNumberOfVolumes;

		//set up the volume tree
		if(iNumberOfVolumes)
		{
			Asura_AABB_Tree* pxTree = m_pxVolumeTree = new Asura_AABB_Tree;
		
			if(iNumberOfVolumes > 1)
			{
				Asura_ResourceAllocateArray(pxTree->Nodes, iNumberOfVolumes-1);
			}
			else
			{
				pxTree->Nodes = 0;
			}

			xStream >> pxTree->RootBoundingBox;

			Asura_AABB_Tree_Node* pxNode = pxTree->Nodes;
			for(int i=0;i<iNumberOfVolumes-1;i++,pxNode++)
			{
				pxNode->ReadFromChunkStream(xStream);
			}
		}
		m_iNumberOfVolumes = iNumberOfVolumes;
	}
	#ifndef ASURA_TOOLS
	else
	{
		Asura_Assert(0,"Loading an old version of an Asura_Zone_Controller. The level will need to be re-exported");
	}
	#endif
}


#ifdef ASURA_TOOLS

void Asura_Zone_Controller::FillInChunkBuffer(char*& pcChunkBuffer)
{
	Asura_Chunk_ChunkBuffer_Stream xStream(pcChunkBuffer);
	Asura_Zone_Controller::WriteToChunkStream(xStream);
}

int Asura_Zone_Controller::GetChunkBufferSize()
{
	Asura_Chunk_Dummy_Stream xStream;
	Asura_Zone_Controller::WriteToChunkStream(xStream);

	return xStream.GetPosition();
}

void Asura_Zone_Controller::WriteToChunkStream(Asura_Chunk_Stream& xStream)
{
	xStream << ASURA_ZONE_CONTROLLER_VERSION;
	xStream << m_iNumberOfZones;
	
	for(int i=0;i<m_iNumberOfZones;i++)
	{
		m_ppxZones[i]->WriteToChunkStream(xStream);
	}

	//update aabb tree if needed , and save it out
	if(!m_bVolumeTreeValid) UpdateVolumeTree();

	xStream << m_iNumberOfVolumes;
	if(m_pxVolumeTree)
	{
		Asura_AABB_Tree::SaveTree(1,m_pxVolumeTree,m_iNumberOfVolumes,xStream);
	}

}
#endif // ASURA_TOOLS


#ifdef ASURA_TOOLS

void Asura_Zone_Controller::UpdateVolumeTree()
{
	//delete the old tree
	delete m_pxVolumeTree;
	m_pxVolumeTree = 0;

	//update the count of the number of volumes
	m_iNumberOfVolumes = 0;
	for(int i=0;i<m_iNumberOfZones;i++)
	{
		m_iNumberOfVolumes += m_ppxZones[i]->GetNumberOfVolumes();
	}
	m_bVolumeTreeValid = true;

	if(!m_iNumberOfVolumes) return;

	//and create a new one
	g_pxCurrentZoneController = this;
	m_pxVolumeTree = Asura_AABB_Tree_Creator::Create(m_iNumberOfVolumes,FillVolumeTreeCallback);
	g_pxCurrentZoneController = 0;
}

bool Asura_Zone_Controller::FillVolumeTreeCallback(u_int uNumObjects,Asura_Bounding_Box* pxBoxes,Asura_Vector_3* pxCentres)
{
	int iDestIndex = 0;


	for(int i=0; i<g_pxCurrentZoneController->m_iNumberOfZones; i++)
	{
		int iNumVolumes = g_pxCurrentZoneController->m_ppxZones[i]->GetNumberOfVolumes();

		for(int j=0; j<iNumVolumes; j++)
		{

			g_pxCurrentZoneController->m_ppxZones[i]->GetAABBForVolume(j,pxBoxes[iDestIndex]);
					
			//add a small amount onto the volumes in order to make sure that the volumes in the AABB
			//tree wil contain the actual volumes
			pxBoxes[iDestIndex].MaxX += 0.5f;
			pxBoxes[iDestIndex].MaxY += 0.5f;
			pxBoxes[iDestIndex].MaxZ += 0.5f;
			pxBoxes[iDestIndex].MinX -= 0.5f;
			pxBoxes[iDestIndex].MinY -= 0.5f;
			pxBoxes[iDestIndex].MinZ -= 0.5f;

			pxCentres[iDestIndex] = pxBoxes[iDestIndex].CalculateCentre();

			iDestIndex++;
		}
	}
	
	Asura_Assert(iDestIndex == (int)uNumObjects,"Error building tree for Asura_Zone_Controller.");
	Asura_Assert(iDestIndex == g_pxCurrentZoneController->m_iNumberOfVolumes,"Error building tree for Asura_Zone_Controller.");
	return true;
}
#endif //ASURA_TOOLS

bool Asura_Zone_Controller::VolumeTreeCallback(u_short usIndex)
{
	//Find the zone and volume index from this index
	int iVolumeIndex = usIndex;
	for(int i=0;i<g_pxCurrentZoneController->m_iNumberOfZones; i++)
	{
		const Asura_Zone* pxZone = g_pxCurrentZoneController->m_ppxZones[i];
		
		if(iVolumeIndex<pxZone->GetNumberOfVolumes())
		{
			//volume belongs to this zone , so do a second check to see whether the volume
			//contains the test point
			if(pxZone->Contains(g_xTestPosition, iVolumeIndex))
			{
				if(g_pxCurrentZoneController->UsePriorityForOverlaps())
				{
					float fPriority = pxZone->GetPriority(iVolumeIndex);
					if(g_iFoundZoneIndex == -1 || (fPriority > g_fFoundZonePriority))
					{
						g_iFoundZoneIndex = i;
						g_fFoundZonePriority = fPriority;	
					}
					//Need to continue searching to see if a zone with a higher priority is found
					return true;
				}
				else
				{
					//we are accepting the first zone that contains the position , and this is it
					//so stop searching
					g_iFoundZoneIndex = i;
					return false;
				}

			}
		}		
		else
		{
			iVolumeIndex -= pxZone->GetNumberOfVolumes();
			if(iVolumeIndex < 0)
			{
				Asura_Assert(0,"Invalid index when testing volume tree for Asura_Zone_Controller");
				break;
			}
		}
	}

	return true;

}
