///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PC-CLI_Manager.cpp
// Author : Johan Lerstrom
// Date   : 19-Mar-2009
//
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

// INCLUDES

#include "Core/Asura.h"

#include "Debugging/Asura_Version.h"

// This file's header.
#include "Main/Asura_PC-CLI_Manager.h"

// Asura independant
#include "Core/Asura_Core.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CfgFileReader.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_NetTransport_System.h"
#include "Rendering/Asura_Viewport.h"

// Asura PC-CLI
#include "Core/Asura_Windows.h"

// Project
#include "Main/Asura_Project_Main.h"

#ifdef ASURA_SUPPORT_GAMESPY
#include "GameSpy/common/gsAvailable.h"
#include "GameSpy/gstats/gstats.h"
#endif

#include <process.h>
#include <errno.h>

int Asura_PCCLI_Manager::s_iBotsToSpawn = 0;
int Asura_PCCLI_Manager::s_iAutomatchTestingBotsToSpawn = 0;
u_int Asura_PCCLI_Manager::s_uNextBotSpawnTime = 0;
bool Asura_PCCLI_Manager::s_bPlayingOnline = false;
bool Asura_PCCLI_Manager::s_bHostRequested = false;
bool Asura_PCCLI_Manager::s_bJoinRequested = false;

// Constants

const u_int uBOTSPAWN_COMMANDLINE_MAX_LEN = 256;
// OR - 16/09/2004 15:48:31 - Commands that have to be executed asap
const Asura_UnicodeChar* wszPriorityCommands[] = 
{ 
	L"NOASSERT",
	L"PORT" 
	L"LANONLY"
};
const u_int uFAKE_VIEWPORT_WIDTH = 800;
const u_int uFAKE_VIEWPORT_HEIGHT = 600;
const int iSPAWN_COOLDOWN_MSECS = 2000;
const int iMAX_BOTS_TO_SPAWN = 32;

void Asura_PCCLI_Manager::InitCommandLineArgs(LPSTR& lpCmdLine)
{
	Asura_CommandConsole::EarlyInitialise();
	Asura_CommandConsole::LateInitialise();
	Asura_CommandConsole::AddCmd( L"Port", ConsoleCmd_Port, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"LANONLY", ConsoleCmd_LanOnly, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	ProcessCommandLine(lpCmdLine, true);
	Asura_CommandConsole::Shutdown();
}

bool Asura_PCCLI_Manager::Initialise(LPSTR& lpCmdLine)
{
	Asura_Core::EarlyInitialise();
	Asura_Project::EarlyInitialise();

	//Setup a fake viewport to avoid an assert in the GUIMenu system.
	Asura_Viewport::Set( uFAKE_VIEWPORT_WIDTH, uFAKE_VIEWPORT_HEIGHT );

	Asura_Core::LateInitialise();
	Asura_Project::LateInitialise();

#ifdef ASURA_SUPPORT_GAMESPY
	Asura_CommandConsole::Printf(L"Starting GameSpy availability check");
	Asura_GameSpy::Initialise();

	do 
	{
		Sleep(0);
		Asura_GameSpy::Update();

	} while(Asura_GameSpy::GetState() == ASURA_GAMESPY_STATE_AVAILABLE_CHECK);

	if (Asura_GameSpy::GetState() != ASURA_GAMESPY_STATE_INITIALISED)
	{
		Asura_CommandConsole::Printf(L"GameSpy availability check failed (%d)", Asura_GameSpy::GetState());
		Asura_Project::Shutdown();
		Asura_Core::Shutdown();
		Asura_OutputToDebugger_Retail(("return -1\n"));
		return false;
	}

	Asura_CommandConsole::Printf(L"GameSpy availability check success");
#endif // #ifdef ASURA_SUPPORT_GAMESPY

	// WSAStartup needs to have been called to initialise Asura_NetTransport_System, 
	// and we need that to get the IP Address to print to the console.
	WSADATA xData;
	WSAStartup(0x0101, &xData);
	Asura_NetTransport_System::Initialise();

	Asura_CommandConsole::Printf(L"-----------<<< ASURA PC-CLI >>>-----------");
	Asura_CommandConsole::Printf(L"- IP Address : [%s]", Asura_NetTransport_System::GetNetAdapterIPWStr(Asura_NetTransport_System::GetNetAdapter()));
	Asura_CommandConsole::Printf(L"- Build      : [%S]", Asura_BuildTimeAndDate);
	Asura_CommandConsole::Printf(L"- Version    : [%d]", Asura_NetTransport_System::Project_GetVersionID());
	Asura_CommandConsole::Printf(L"- Parameters : [\"%S\"]", lpCmdLine);
	Asura_CommandConsole::Printf(L"------------------------------------------");

	Asura_NetTransport_System::Shutdown();
	WSACleanup();

	Asura_CommandConsole::AddCmd( L"Exec", ConsoleCmd_ExecConfig, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AddBot", ConsoleCmd_SpawnBots, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#ifdef ASURA_SUPPORT_GAMESPY
	Asura_CommandConsole::AddCmd( L"RefreshGameSpyListing", ConsoleCmd_RefreshGameSpyQR2, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
	Asura_CommandConsole::AddCmd( L"Port", ConsoleCmd_Port, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"LANONLY", ConsoleCmd_LanOnly, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Host", Project_ConsoleCmd_RequestHost, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Join", Project_ConsoleCmd_RequestJoin, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"Map", Project_ConsoleCmd_RequestMap, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( ASURA_CONSOLECMD_LISTPLAYERS, Project_ConsoleCmd_ListPlayers, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	Asura_CommandConsole::AddCmd( L"AutomatchTesting.AddBot", ConsoleCmd_SpawnAutomatchTestingBots, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	if (!ProcessCommandLine(lpCmdLine, false))
	{
		if (!ProcessCommandLine("-exec default.cfg", false))
		{
			Asura_CommandConsole::Printf(L"config file \"%S\" not found.", "default.cfg");
		}
	}
	
	Project_Initialise();

	return true;
}

void Asura_PCCLI_Manager::InitLogFiles( LPSTR& lpCmdLine )
{
	// (MH - 18.04.2005) init the logfiles dir. Need to find whether the logfiles dir is
	// overridden or not.
	const char* szLogDirCommand = "-logdir ";
	const char* szLogDir = strstr( lpCmdLine, szLogDirCommand );
	if (szLogDir)
	{
		char acTemp[MAX_PATH];
		szLogDir += strlen(szLogDirCommand);
		Asura_PrintfToBfr( acTemp, "%s", MAX_PATH, true, szLogDir );
		char* szSpace = strstr( acTemp, " " );
		if (szSpace)
		{
			*szSpace = 0;
		}
		Asura_OverrideLogfilesDir( acTemp, true );
	}
	else
	{
		Asura_OverrideLogfilesDir( "logfiles_dedicatedserver", false );
	}
}

void Asura_PCCLI_Manager::Shutdown(void)
{
#ifdef ASURA_SUPPORT_GAMESPY
	// close connection to GameSpy stats server
	CloseStatsConnection();
	// shutdown other systems
	Asura_GameSpy::Shutdown();
#endif
	Asura_Project::Shutdown();
	Asura_Core::Shutdown();
	Asura_Windows::Shutdown();
	Asura_OutputToDebugger_Retail(("return 0\n"));
}


void Asura_PCCLI_Manager::Update( void )
{
	float fMaxFrameRate = (Asura_Network::IsNetworkInitialised() && Asura_Network::IsServer())? 60.0f : 20.0f;
	Asura_Timers::SetMaxFrameRate(fMaxFrameRate);

	// force LAN reporting to work
	if (Asura_Network::IsServer() && !s_bPlayingOnline)
	{
		Asura_NetTransport_System::SetServerLANReporting(true);
	}

	Project_Update();

	// spawn a bot application every 4 seconds until we've done them all
	if (s_iBotsToSpawn && 
		Asura_Network::IsNetworkInitialised() && 
		Asura_Network::IsServer() && 
		Asura_Timers::GetMSecsSinceStart() > s_uNextBotSpawnTime)
	{
		s_uNextBotSpawnTime = Asura_Timers::GetMSecsSinceStart() + iSPAWN_COOLDOWN_MSECS;
		SpawnBot();
		--s_iBotsToSpawn;
	}

	if (s_iAutomatchTestingBotsToSpawn && Asura_Timers::GetMSecsSinceStart() > s_uNextBotSpawnTime)
	{
		s_uNextBotSpawnTime = Asura_Timers::GetMSecsSinceStart() + iSPAWN_COOLDOWN_MSECS;
		SpawnAutomatchTestingBot();
		--s_iAutomatchTestingBotsToSpawn;
	}
}

void Asura_PCCLI_Manager::UpdateFrontEnd( void )
{
	Asura_CommandConsole::Update();
}

bool Asura_PCCLI_Manager::ProcessCommandLine(LPSTR lpCmdLine, bool bPriorityCommands)
{
	if (!lpCmdLine || lpCmdLine[0] == 0)
	{
		Asura_CommandConsole::Printf(L"No command line parameters.");
		return false;
	} 
	
	bool bReturnValue = true;

	Asura_UnicodeChar* wszSeps = L"-";

	//Make a copy of the command line, since it will be altered by strtok. 
	int length = strlen(lpCmdLine) + 1;
	
	Asura_UnicodeChar* wszCommandLine = new Asura_UnicodeChar[length];

	Asura_PrintfToBfr(wszCommandLine, L"%S", length, true, lpCmdLine);

	_wcsupr(wszCommandLine);

	Asura_UnicodeChar* wszToken = wcstok(wszCommandLine, wszSeps);
	
	while(wszToken)
	{
		bool bIsPriorityCommand = false;

		for(int j = 0; j < sizeof(wszPriorityCommands) / sizeof(*wszPriorityCommands); j ++)
		{
			Asura_UnicodeChar wszUpperCommand[64];

			wcscpy(wszUpperCommand, wszPriorityCommands[j]);
			_wcsupr(wszUpperCommand);
	
			bIsPriorityCommand = (wcsstr(wszToken, wszUpperCommand) != NULL);

			if (bIsPriorityCommand)
				break;
		}

		if (!(bIsPriorityCommand ^ bPriorityCommands))
		{
			// (MH - 19.04.2005) hack - exclude commands which are processed right at the start
			bool bExclude = (0 == _wcsnicmp(wszToken, L"logdir", wcslen(L"logdir")));
			if (!bExclude)
			{
				Asura_CommandConsole::Execute(wszToken, &bReturnValue);
			}
		}
		
		wszToken = wcstok(NULL, wszSeps);
	}
	delete[] wszCommandLine;

	return bReturnValue;
}


bool Asura_PCCLI_Manager::ProcessCommandLine(int iNumArgs, const char** aszArgs, bool bPriorityCommands)
{
	if (iNumArgs < 1)
		return false;

	int iLen=0;
	for(int i = 1; i < iNumArgs; i ++)
	{
		iLen += strlen(aszArgs[i]);
	}
	char *szCmdLine = new char[iLen + 10];

	szCmdLine[0] = 0;
	
	for(int i = 1; i < iNumArgs; i ++)
	{
		strcat(szCmdLine, aszArgs[i]);
	}
	
	bool bRet = ProcessCommandLine(szCmdLine, bPriorityCommands);

	delete[] szCmdLine;

	return bRet;
}

bool Asura_PCCLI_Manager::ConsoleCmd_SpawnBots(int iLimit, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		Asura_CommandConsole::Printf(L"Spawn fake client applications.");
	}
	else
	{
		if (iLimit <= iMAX_BOTS_TO_SPAWN)
		{
			s_iBotsToSpawn = iLimit;
		}
		else
		{
			Asura_CommandConsole::Printf(L"can't spawn that many bots.");
			return false;
		}
	}
	return true;
}

bool Asura_PCCLI_Manager::ConsoleCmd_SpawnAutomatchTestingBots( int iLimit, bool bInterrogate, bool bHelp )
{
	if (bHelp || bInterrogate)
	{
		Asura_CommandConsole::Printf(L"Spawn automatch testing client applications.");
	}
	else
	{
		if (iLimit <= iMAX_BOTS_TO_SPAWN)
		{
			s_iAutomatchTestingBotsToSpawn = iLimit;
		}
		else
		{
			Asura_CommandConsole::Printf(L"can't spawn that many bots.");
			return false;
		}
	}
	return true;
}

bool Asura_PCCLI_Manager::ConsoleCmd_LanOnly(bool bHelp)
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf(L"Only used at startup - sets server to be LAN-only");
	}
	else
	{
		s_bPlayingOnline = false;
	}
	return true;
}

bool Asura_PCCLI_Manager::ConsoleCmd_ExecConfig(const Asura_UnicodeChar* wszFile, bool bInterrogate, bool bHelp)
{
	if (bInterrogate || bHelp)
	{
		Asura_CommandConsole::Printf(L"Exec a config file.");
	}
	else
	{
		return Asura_CfgFileReader::ExecuteFile(wszFile);
	}

	return true;
}

bool Asura_PCCLI_Manager::ConsoleCmd_Port(int iPort, bool bInterrogate, bool bHelp)
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf(L"Sets network port to use.");
	}
	else if (bInterrogate)
	{
		Asura_CommandConsole::Printf(L"Port : [%d]", Asura_NetTransport_System::GetPort());
	}
	else
	{
		Asura_NetTransport_System::SetPort(static_cast<u_short>(iPort));
	}
	return true;
}

// Gets a command line from a project specific function, then spawns a new
// Application based on the current one with the retrieved command line
void Asura_PCCLI_Manager::SpawnBot( void )
{
	static u_int ls_uNumSpawned = 0;
	char szCmdLine[uBOTSPAWN_COMMANDLINE_MAX_LEN];
	char acFullPath[MAX_PATH];
	GetModuleFileName( NULL, acFullPath, MAX_PATH  );
	acFullPath[MAX_PATH - 1] = 0;

	Project_GetDefaultSpawnBotCommandline(szCmdLine, uBOTSPAWN_COMMANDLINE_MAX_LEN, ls_uNumSpawned );

	int err = _spawnl(_P_NOWAIT, acFullPath, acFullPath, szCmdLine, NULL);

	if (err == -1)
	{
		switch(errno)
		{
		case E2BIG:		Asura_CommandConsole::Printf(L"Spawn failed : argument list greater than 128 bytes."); break;
		case EINVAL:	Asura_CommandConsole::Printf(L"Spawn failed : mode argument is invalid."); break;
		case ENOMEM:	Asura_CommandConsole::Printf(L"Spawn failed : Not enough memory is available to execute new process"); break;
		case ENOENT:	Asura_CommandConsole::Printf(L"Spawn failed : path to %S not found.", acFullPath); break;
		case ENOEXEC:	Asura_CommandConsole::Printf(L"Spawn failed : %S is an invalid executable.", acFullPath); break;
		default:		Asura_CommandConsole::Printf(L"Spawn Failed : Unhandled errno = %d", err); break;
		}
	}
	else
	{
		++ls_uNumSpawned;
	}
}

// Gets a command line from a project specific function, then spawns a new
// Application based on the current one with the retrieved command line
void Asura_PCCLI_Manager::SpawnAutomatchTestingBot( void )
{
	static u_int ls_uAutomatchTestingBotsSpawned = 0;
	char szCmdLine[uBOTSPAWN_COMMANDLINE_MAX_LEN];
	char acFullPath[MAX_PATH];
	GetModuleFileName( NULL, acFullPath, MAX_PATH  );
	acFullPath[MAX_PATH - 1] = 0;

	Project_GetAutomatchTestingSpawnBotCommandline(szCmdLine, uBOTSPAWN_COMMANDLINE_MAX_LEN, ls_uAutomatchTestingBotsSpawned );

	int err = _spawnl(_P_NOWAIT, acFullPath, acFullPath, szCmdLine, NULL);

	if (err == -1)
	{
		switch(errno)
		{
		case E2BIG:		Asura_CommandConsole::Printf(L"Spawn failed : argument list greater than 128 bytes."); break;
		case EINVAL:	Asura_CommandConsole::Printf(L"Spawn failed : mode argument is invalid."); break;
		case ENOMEM:	Asura_CommandConsole::Printf(L"Spawn failed : Not enough memory is available to execute new process"); break;
		case ENOENT:	Asura_CommandConsole::Printf(L"Spawn failed : path to %S not found.", acFullPath); break;
		case ENOEXEC:	Asura_CommandConsole::Printf(L"Spawn failed : %S is an invalid executable.", acFullPath); break;
		default:		Asura_CommandConsole::Printf(L"Spawn Failed : Unhandled errno = %d", err); break;
		}
	}
	else
	{
		++ls_uAutomatchTestingBotsSpawned;
	}
}


#ifdef ASURA_SUPPORT_GAMESPY
bool Asura_PCCLI_Manager::ConsoleCmd_RefreshGameSpyQR2( bool bHelp )
{
	if (bHelp)
	{
		Asura_CommandConsole::Printf("Force update of GameSpy server listing (for development purposes)");
	}
	else
	{
		static u_int ls_uLastCall = 0;
		if ( (Asura_Timers::GetMSecsSinceStart_Continuous() - ls_uLastCall) < 9999 )
		{
			Asura_CommandConsole::Printf("RefreshGameSpyListing can only be used once every 10 seconds");
			return false;
		}
		else
		{
			Asura_GameSpy_QR2::GameModeChanged();
			ls_uLastCall = Asura_Timers::GetMSecsSinceStart_Continuous();
		}
	}
	return true;
}
#endif // #ifdef ASURA_SUPPORT_GAMESPY
