//*******************************************************************
//
// File: Asura_Direct3D_Carpet_Game_Render.cpp
//
// Created: 19 Jan 2006
//
// Description:
//
// Shader Constants
//		Camera
//		Billboard Right
//		Billboard Up
//		Array[4] Corner Offsets (-1.0 to 1.0) or (0.0 to 1.0)
//		Array[16] Wave Table XYZ
//		Array[Dynamic] Billboard Styles
//			Size Width, Height
//			UV	TL, TR, BL, BR
//
// Shader Vertex Format
//		Position (float3)
//		Corner Index (uint8)
//		Wave Index (uint8)
//		Style Index (uint8)
//		EdgeDistance A, B, (C)? (2*uint8)
//		Scale Width, Height (2*uint8)
//		Colour (4*uint8)
//		
//
//
//*******************************************************************

// system libraries
#include "Core/Asura_MemoryManagement_DebugOff.h"
#include <vector>
#include "Core/Asura_MemoryManagement_DebugOn.h"

#include "Core/Asura.h"

#if defined( ASURA_USE_CARPET)

// other libraries
#include "Core/Asura_Mempool.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Camera/Asura_Camera.h"
#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
#include "Light/Asura_Photon.h"
#endif

#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Direct3d.h"
#include "Rendering/Asura_DIPCache.h"
#if defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	#include "Rendering/Asura_Direct3D_Effects_System.h"
#endif
#include "Rendering/Asura_Direct3D_Shaders.h"
#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
#include "Rendering/Asura_Fog.h"
#endif

#include "Carpet/Asura_Carpet_MeshManager.h"

#if defined( ASURA_TOOLS)
#include "Carpet/Tools_Carpet_Statistics.h"
#include "Carpet/Tools_Carpet_Mesh.h"
#include "Carpet/Asura_Carpet_StyleManager.h"
#endif

#include "Debugging/Asura_Profiling_System.h"


#include "Light/Asura_GI_Solution.h"
#include "Light/Asura_Lighting_Solution.h"

// main header
#include "Carpet/Asura_Direct3D_Carpet_Game_Render.h"

// other libraries

// internal dependencies

//*******************************************************************
// Constants
//*******************************************************************

#if !defined( ASURA_CGR_MESH_VB_SIZE)
	#define ASURA_CGR_MESH_VB_SIZE 0x20000
#endif

#if defined( ASURA_TOOLS)
#if ASURA_DIRECT3D_CGR_POLYGONSTYLE_LOOKUP_LENGTH != ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH
	#error ASURA_DIRECT3D_CGR_POLYGONSTYLE_LOOKUP_LENGTH != ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH
#endif
#endif

#define MANGLED_SB_TEXTURE_COUNT 0x1000
#define MANGLED_SB_STYLE_COUNT 0x1000
#define MANGLED_SB_BILLBOARD_COUNT 0x1000

#define MANGLED_SB_TEXTURE_OFFSET 1
#define MANGLED_SB_STYLE_OFFSET  (MANGLED_SB_TEXTURE_COUNT * MANGLED_SB_TEXTURE_OFFSET)
#define MANGLED_SB_BILLBOARD_OFFSET (MANGLED_SB_STYLE_COUNT * MANGLED_SB_STYLE_OFFSET)


//*******************************************************************
// Local Classes
//*******************************************************************

static bool Asura_DIPCache_Render( const Asura_DIPCache_DIPData& xDIPData, const uint32 uCallbackData);

//*******************************************************************
// Local Data
//*******************************************************************
const uint32 Asura_Direct3D_CGR_PolygonStyle::uiInvalidStyleIndex = uint32( -1);

uint32 Asura_Direct3D_Carpet_Game_Render::s_uMeshesCount = 0;
Asura_Direct3D_CGR_Mesh* Asura_Direct3D_Carpet_Game_Render::s_pxMeshes = NULL;

uint32 Asura_Direct3D_Carpet_Game_Render::s_uTimeStampLast = 0;
Asura_Direct3D_Carpet_Game_Render::VertexMode Asura_Direct3D_Carpet_Game_Render::s_eVertexMode = Asura_Direct3D_Carpet_Game_Render::SHADER_SIMPLE;
Direct3D_VertexDeclaration* Asura_Direct3D_Carpet_Game_Render::s_pxVertexDeclarationSimple = NULL;

int Asura_Direct3D_Carpet_Game_Render::s_iDrawCacheTextureID = -1;
int Asura_Direct3D_Carpet_Game_Render::s_iAttenuationMap = ASURA_TEXTURE_INVALID;

uint32 Asura_Direct3D_Carpet_Game_Render::s_uBatchIndexBufferSize = 6 * (0x10000 / 4);
Direct3D_IndexBuffer* Asura_Direct3D_Carpet_Game_Render::s_pxBatchIndexBuffer = NULL;
//uint32 Asura_Direct3D_Carpet_Game_Render::s_uBatchVertexBufferSize = 0x10000;
//Direct3D_VertexBuffer* Asura_Direct3D_Carpet_Game_Render::s_pxBatchVertexBuffer = NULL;
#undef TOOLS_CARPET_MESH_DRAW_CACHE_QUADS

Asura_DIPCache_Static< 100> s_xDIPCache( Asura_DIPCache_Render, 0);

//*******************************************************************
// Function Definitions
//*******************************************************************

static bool Asura_DIPCache_Render( const Asura_DIPCache_DIPData& xDIPData, const uint32 uCallbackData)
{
	const static uint32 uIndexMaxCount = Asura_Direct3D_Carpet_Game_Render::GetBatchIndexBufferSize();
	
	HRESULT xResult;

	uint32 uIndexTotalCount = xDIPData.GetIndexBufferStop() - xDIPData.GetIndexBufferStart();
	uint32 uVertexBufferOffset = 0;
	while( 0 < uIndexTotalCount)
	{
		const uint32 uIndexCount = Asura_Maths::Min( uIndexMaxCount, uIndexTotalCount);
		const uint32 uiCountPrimitives = uIndexCount / 3;
		const uint32 uVertexCount = 2 * uiCountPrimitives;
		xResult = Asura_Direct3D::D3DDevice->DrawIndexedPrimitive( 
			D3DPT_TRIANGLELIST, 
			xDIPData.GetVertexBufferStart() + uVertexBufferOffset, 0, uVertexCount, 
			0, uiCountPrimitives);
		if( ASURA_D3D_FAILED( xResult))
		{
			return( false);
		}
		
		#if defined( ASURA_TOOLS)
			TOOLS_CARPET_STATISTICS_ADD( s_uDIPCount, 1);
			TOOLS_CARPET_STATISTICS_ADD( s_uDIPIndexCount, uIndexCount);
			TOOLS_CARPET_STATISTICS_ADD( s_uDIPVertexCount, uVertexCount);
			TOOLS_CARPET_STATISTICS_ADD( s_uDIPPrimitiveCount, uiCountPrimitives);
		#endif
		
		uIndexTotalCount -= uIndexCount;
		uVertexBufferOffset += uVertexCount;
	}
	
	return( true);
}

bool Asura_Direct3D_CGR_Camera::GetCurrentCamera( const float fFarPlane)
{
	const Asura_CameraData& xCurrentCamera = Asura_Camera::GetCameraData( 0 );
	
	this->m_fFarPlane = fFarPlane;
	this->m_xOrientation = xCurrentCamera.m_xOrientation;
	this->m_xPosition = xCurrentCamera.m_xPosition;
	this->m_fOneOverCosHalfFovX = xCurrentCamera.m_fOneOverCosHalfFovX;
	this->m_fOneOverCosHalfFovY = xCurrentCamera.m_fOneOverCosHalfFovY;
	this->m_fTanHalfFovX = xCurrentCamera.m_fTanHalfFovX;
	this->m_fYOverXAspectRatio = xCurrentCamera.m_fYOverXAspectRatio;
	
	this->m_fFarPlane = fFarPlane;
	
	return( true);
}

Asura_Direct3D_CGR_RangeManager::Asura_Direct3D_CGR_RangeManager()
{
	this->m_uRangeMaxCount = 0;
	this->m_uRangeUsedCount = 0;
	this->m_uBlockUsedCount = 0;
	//this->m_uBlockMaxCount = 0;
}

void Asura_Direct3D_CGR_RangeManager::Clear()
{
	this->m_uRangeUsedCount = 0;
	this->m_uBlockUsedCount = 0;
}

void Asura_Direct3D_CGR_RangeManager::SetMaxRange( const uint32 uMaxRange)
{
	this->m_uRangeMaxCount = uMaxRange;
}

Asura_Direct3D_CGR_RTreeNode::Asura_Direct3D_CGR_RTreeNode()
{
	this->m_xBoundingBox.Zero();
	this->m_pxChild = NULL;
	this->m_pxSibling = NULL;
	this->m_uData = 0;
}

Asura_Direct3D_CGR_Vertex::Asura_Direct3D_CGR_Vertex()
{
	//this->m_xPosition;
	//this->m_aucColourARGB[ 4];
};

Asura_Direct3D_CGR_Polygon::Asura_Direct3D_CGR_Polygon()
{
	this->m_fLodFadeFar = 0;
	this->m_fDenisty = 0;
	this->m_uInstanceCount = 0;
	this->m_uRandomSeed = 0;
	
	this->m_uVertexIndex[0] = 0;
	this->m_uVertexIndex[1] = 0;
	this->m_uVertexIndex[2] = 0;
};

Asura_Direct3D_CGR_PolygonBatch::Asura_Direct3D_CGR_PolygonBatch()
{
	this->m_uDataOffset = 0;
	this->m_uDataSize = 0;
	this->m_uVertexBufferOffset = 0;
	this->m_uVertexBufferSize = 0;
	this->m_uTimeStampLastUsed = 0;
}

Asura_Direct3D_CGR_PolygonStyle::Asura_Direct3D_CGR_PolygonStyle()
{
#if defined( ASURA_TOOLS)
	this->m_uStyle = 0;
	this->m_fLodFadeFar = 0;
#endif
};

Asura_Direct3D_CGR_Style::Asura_Direct3D_CGR_Style()
{
	this->m_fWidth = 0;
	this->m_fHeight = 0;
	this->m_afU[0] = 0;
	this->m_afV[0] = 0;
	this->m_afU[1] = 0;
	this->m_afV[1] = 0;
	this->m_fScaleMin = 0; // DLH - 0-255 encoded 0.0f-1.0f
	this->m_fLodFadeFar = 0;
	this->m_fLodFadeNear = 0;
	this->m_fLodFadeDelta = 1e-5f;
	
#if defined( ASURA_TOOLS)
	this->m_uTriangleCount = 0;
#endif
}

Asura_Direct3D_CGR_Mesh::Asura_Direct3D_CGR_Mesh()
{
	this->m_iTexture = 0;
	
	this->m_uVertexCount = 0;
	this->m_uPolygonCount = 0;
	this->m_uPolygonStyleCount = 0;
	this->m_uStyleCount = 0;
	this->m_uRTreeCount = 0;
	this->m_uPolygonBatchCount = 0;
	
	//this->m_BoundingBox;
	
	this->m_pxVertices = NULL;
	this->m_pxPolygons = NULL;
	this->m_pxPolygonStyles = NULL;
	this->m_pxStyles = NULL;
	this->m_pxRTree = NULL;
	this->m_pxPolygonBatches = NULL;
	
	this->m_uBatchVertexBufferSize = ASURA_CGR_MESH_VB_SIZE;
	this->m_pxBatchVertexBuffer = NULL;
}

Asura_Direct3D_CGR_Mesh::~Asura_Direct3D_CGR_Mesh()
{
	SAFE_RELEASE(m_pxBatchVertexBuffer);
	
	delete[] m_pxVertices;
	delete[] m_pxPolygons;
	delete[] m_pxPolygonStyles;
	delete[] m_pxStyles;
	delete[] m_pxRTree;
	delete[] m_pxPolygonBatches;
	
	m_uVertexCount = 0;
	m_uPolygonCount = 0;
	m_uPolygonStyleCount = 0;
	m_uStyleCount = 0;
	m_uRTreeCount = 0;
	m_uPolygonBatchCount = 0;
	m_uBatchVertexBufferSize = 0;
//	m_uBatchVertexBufferPresentSize = 0;
}

//-------------------------------------------------------------------
// Function:    Asura_Direct3D_Carpet_Game_Render::GetCullStateOfSphereInFrustum
//
// Description: 
//
// Parameters:  xCentre         - 
//              fRadius         - 
//
// Returns:     Inside, outside or Touching the view frustum extents
//-------------------------------------------------------------------
Asura_Direct3D_Carpet_Game_Render::CullState Asura_Direct3D_Carpet_Game_Render::GetCullStateOfSphereInFrustum( const Asura_Direct3D_CGR_Camera& xCamera, const Asura_Vector_3& xCentre, const float fRadius)
{			
	//const Asura_CameraData* pxCurrentCamera = Asura_Camera::GetCameraData();
	const Asura_Matrix_3x3& xOrientation = xCamera.m_xOrientation;

	const Asura_Vector_3 xViewSpacePos = xOrientation.ApplyTranspose( xCentre - xCamera.m_xPosition );

	const float fZFar = xViewSpacePos.z + fRadius;
	// is the sphere outside of the view frustum?
	if( fZFar < 0)
	{
		return( CULL_OUTSIDE);
	}
	
	const float fZNear = xViewSpacePos.z - fRadius;
	if( xCamera.m_fFarPlane < fZNear)
	{
		return( CULL_OUTSIDE);
	}
	
	const float fRX = fRadius * xCamera.m_fOneOverCosHalfFovX;
	const float fZX = xViewSpacePos.z * xCamera.m_fTanHalfFovX;

	// is the sphere outside of the left and right planes?
	if( fRX < (xViewSpacePos.x - fZX) || fRX < (-xViewSpacePos.x - fZX))
	{
		return( CULL_OUTSIDE);
	}
	
	const float fZY = fZX * xCamera.m_fYOverXAspectRatio;
	const float fRY = fRadius * xCamera.m_fOneOverCosHalfFovY;

	// is the sphere outside of the top and bottom planes?
	if ( fRY < (xViewSpacePos.y-fZY) || fRY < (-xViewSpacePos.y-fZY))
	{
		return( CULL_OUTSIDE);
	}
	
	
	// is the sphere inside of all the planes?
	if( 
			0 <= fZNear
		&&	
			fRX <= (fZX - xViewSpacePos.x)
		&& 
			fRX <= (xViewSpacePos.x - (-fZX))
		&&	
			fRY <= (fZY - xViewSpacePos.y)
		&& 
			fRY <= (xViewSpacePos.y - (-fZY))
		&&
			fZFar < xCamera.m_fFarPlane
		)
	{
		return( CULL_INSIDE);
	}
	
	
	return( CULL_TOUCHING);
}

Asura_Direct3D_Carpet_Game_Render::CullState Asura_Direct3D_Carpet_Game_Render::GetCullStateOfAABBInFrustum( const Asura_Direct3D_CGR_Camera& xCamera, const Asura_Bounding_Box& xBB )
{
	// KJFL - 09 May 2005 - overly conservative placeholder test 
	return( GetCullStateOfSphereInFrustum( xCamera, xBB.CalculateCentre(), xBB.CalculateRadius() ));
}

Asura_Direct3D_Carpet_Game_Render::CullState Asura_Direct3D_Carpet_Game_Render::GetCullStateOfAABBInFrustum( const Asura_Direct3D_CGR_Camera_And_Box& xCamera, const Asura_Bounding_Box& xBB )
{
	const bool bBoxIntersects = xCamera.m_xBoundingBox.Intersects( xBB);
	if( false == bBoxIntersects)
	{
		return( CULL_OUTSIDE);
	}
	
	const Asura_Direct3D_Carpet_Game_Render::CullState eCullStateFrustrum = GetCullStateOfSphereInFrustum( xCamera, xBB.CalculateCentre(), xBB.CalculateRadius() );
	if( CULL_OUTSIDE == eCullStateFrustrum)
	{
		return( CULL_OUTSIDE);
	}
	
	const bool bBoxContains = xCamera.m_xBoundingBox.Contains( xBB);
	
	if( CULL_INSIDE == eCullStateFrustrum)
	{
		if( true == bBoxContains)
		{
			return( CULL_INSIDE);
		}
	}
	
	return( CULL_TOUCHING);
}

//-------------------------------------------------------------------
// Function:    Asura_Direct3D_Carpet_Game_Render::CalcVisibleBatches
//
// Description: Fill in a list of uint32 values that are with in the view frustum
//
// Parameters:  
//				xDataFrame      - Not really used
//              pxRTree         - Root of RTree to search
//              puiBatchIndices - Pointer of array to place values into
//              uiBatchCount    - in: Size of Pointer array
//								- out: Count of uint32 values in the array
//
// Returns:     sucees == true
//				fail == false
//-------------------------------------------------------------------
template< class TvisibleBounds, bool bCheckVisability> bool Asura_Direct3D_Carpet_Game_Render::CalcVisibleBatches( const TvisibleBounds& xCamera, const Asura_Direct3D_CGR_RTreeNode* pxRTree, uint32* puiBatchIndices, uint32& uiBatchCount)
{
	const uint32 uiBatchMaxCount = uiBatchCount;
	
	uiBatchCount = 0;
	
	if( NULL == pxRTree)
	{
		return( true);
	}
	
	// DLH - 01 February 2006 16:27:49 - A stack of nodes that have to be processed
	const uint32 uiRTreeNodeToProcessMaxCount = 50;
	const Asura_Direct3D_CGR_RTreeNode* apxRTreeNodesToProcess[ 1 + uiRTreeNodeToProcessMaxCount];

#if defined ASURA_ASSERT || defined ASURA_TOOLS
	const Asura_Direct3D_CGR_RTreeNode** ppxRTreeNodeToProcessMax = &apxRTreeNodesToProcess[ uiRTreeNodeToProcessMaxCount];
#endif

	const Asura_Direct3D_CGR_RTreeNode** ppxRTreeNodeToProcess = apxRTreeNodesToProcess;
	
	// DLH - 01 February 2006 16:28:16 - Insert the root node for processing
	(*ppxRTreeNodeToProcess) = pxRTree;
	
	while( apxRTreeNodesToProcess <= ppxRTreeNodeToProcess)
	{
		// DLH - 01 February 2006 16:28:34 - Pop top of stack
		const Asura_Direct3D_CGR_RTreeNode* pxRTreeNodeProcess = *(ppxRTreeNodeToProcess);
		ppxRTreeNodeToProcess -= 1;
		
		if( NULL == pxRTreeNodeProcess)
		{
			continue;
		}
		
		Asura_Assert( (2 + ppxRTreeNodeToProcess) <= ppxRTreeNodeToProcessMax, "CGR Batch Node Buffer Overrun");
#if defined( ASURA_TOOLS)
		if( ppxRTreeNodeToProcessMax < (2 + ppxRTreeNodeToProcess))
		{
			continue;
		}
#endif

		// DLH - 01 February 2006 16:29:48 - Add the Next sibling for processing.
		ppxRTreeNodeToProcess += 1;
		(*ppxRTreeNodeToProcess) = pxRTreeNodeProcess->m_pxSibling;
		
		// DLH - 01 February 2006 16:30:06 - If this node is not visible then stop processing
		const CullState eIsAABBInFrustum = (true == bCheckVisability) ? GetCullStateOfAABBInFrustum( xCamera, pxRTreeNodeProcess->m_xBoundingBox) : CULL_TOUCHING;
		if( true == bCheckVisability)
		{
			if( CULL_OUTSIDE == eIsAABBInFrustum)
			{
				continue;
			}
			if( CULL_INSIDE == eIsAABBInFrustum)
			{
			#if 1
				uint32 uBatchNewCount = uiBatchMaxCount - uiBatchCount;
				uint32* puiBatchNewIndices = puiBatchIndices + uiBatchCount;
				CalcVisibleBatches< TvisibleBounds, false>( xCamera, pxRTreeNodeProcess->m_pxChild, puiBatchNewIndices, uBatchNewCount);
				
				uiBatchCount += uBatchNewCount;
			#endif
			}
		}
		
		if( false == bCheckVisability || CULL_TOUCHING == eIsAABBInFrustum)
		{
			ppxRTreeNodeToProcess += 1;
			(*ppxRTreeNodeToProcess) = pxRTreeNodeProcess->m_pxChild;
		}
			
		const uint32 uiMeshBatchIndex = pxRTreeNodeProcess->m_uData;
		if( uint32(-1) == uiMeshBatchIndex)
		{
			continue;
		}
		
		// DLH - 01 February 2006 16:30:25 - This node has data to batch it
		Asura_Assert( uiBatchCount < uiBatchMaxCount, "CGR Batch List Overrun");
#if defined( ASURA_TOOLS)
		if( uiBatchMaxCount <= uiBatchCount)
		{
			return( false);
		}
#endif
		puiBatchIndices[ uiBatchCount] = uiMeshBatchIndex;
		uiBatchCount += 1;
	}
	
	return( true);	
}

bool Asura_Direct3D_Carpet_Game_Render::CopyMeshPolygonToVertex( const Asura_Direct3D_CGR_Mesh& xCGRMesh, const Asura_Direct3D_CGR_Polygon& xCGRPolygon, Asura_Direct3D_CGR_StreamVertexSimple* pxVertex0, uint32& uVertexOriginalMaxCount)
{
	uVertexOriginalMaxCount = 0;
	
	Asura_Direct3D_CGR_StreamVertexSimple* pxVerts = pxVertex0;
	uint32 uiCGRPolygonStyle = xCGRPolygon.m_uPolygonStyle;
	Asura_Assertf( uiCGRPolygonStyle < xCGRMesh.m_uPolygonStyleCount, ("Polygon Style index (%d) is too high", uiCGRPolygonStyle));
	if (uiCGRPolygonStyle >= xCGRMesh.m_uPolygonStyleCount)
	{
		// MGH 25/1/2008: Better to default the style to zero than have out-of-range array accesses.
		uiCGRPolygonStyle = 0;
	}
	const Asura_Direct3D_CGR_PolygonStyle* pxCGRPolygonStyle = &xCGRMesh.m_pxPolygonStyles[ uiCGRPolygonStyle];

	// Reseed the random number generator with any value that's constant for this triangle.
	srand( xCGRPolygon.m_uRandomSeed);

	const uint32 uiCGRInstanceCount = xCGRPolygon.m_uInstanceCount;
	
	for( uint32 uiCGRInstanceIndex = 0; uiCGRInstanceIndex < uiCGRInstanceCount; uiCGRInstanceIndex += 1)
	{
		const uint32 uiStyleLookupIndex = uiCGRInstanceIndex % ASURA_DIRECT3D_CGR_POLYGONSTYLE_LOOKUP_LENGTH;
		const uint32 uiStyleIndex = pxCGRPolygonStyle->m_aucStyleLookup[ uiStyleLookupIndex];
		Asura_Assertf( uiStyleIndex < xCGRMesh.m_uStyleCount, ("Style index (%d) is out of range", uiStyleIndex));
		const Asura_Direct3D_CGR_Style* pxStyle = &xCGRMesh.m_pxStyles[ uiStyleIndex];
		
		// Pick a random point within the triangle (barycentric coords)
		float fEdgeA = ( (rand() & 255) / 255.0f);
		float fEdgeB = ( (rand() & 255) / 255.0f);

		if( (fEdgeA + fEdgeB) > 1.0f)
		{
			fEdgeA = 1.0f - fEdgeA;
			fEdgeB = 1.0f - fEdgeB;
		}

		float fEdgeC = 1.0f - (fEdgeA + fEdgeB);

		const uint32 uiCGRVertexIndexA = xCGRPolygon.m_uVertexIndex[ 0];
		const uint32 uiCGRVertexIndexB = xCGRPolygon.m_uVertexIndex[ 1];
		const uint32 uiCGRVertexIndexC = xCGRPolygon.m_uVertexIndex[ 2];
		
		const Asura_Direct3D_CGR_Vertex* pxVertices = xCGRMesh.m_pxVertices;
		const Asura_Direct3D_CGR_Vertex& xVertexA = pxVertices[ uiCGRVertexIndexA];
		const Asura_Direct3D_CGR_Vertex& xVertexB = pxVertices[ uiCGRVertexIndexB];
		const Asura_Direct3D_CGR_Vertex& xVertexC = pxVertices[ uiCGRVertexIndexC];
		Asura_Vector_3 xPositionCenter( (xVertexA.m_xPosition * fEdgeA) + (xVertexB.m_xPosition * fEdgeB) + (xVertexC.m_xPosition * fEdgeC));		


		{				
			// Add some motion
			const uint32 uiWaveIndex = (uiCGRInstanceIndex + xCGRPolygon.m_uRandomSeed) % ASURA_CARPET_MESH_DATA_FRAME_SIZE_WAVEX;

			Asura_Vector_3 xColour = (xVertexA.m_xColourRGB * fEdgeA) + (xVertexB.m_xColourRGB * fEdgeB) + (xVertexC.m_xColourRGB * fEdgeC);

#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
			xColour *= 0xFF;
			uint32 uiColour = (uint32( xColour[2]) << 16) | (uint32( xColour[1]) << 8) | (uint32( xColour[0]) << 0);
			uiColour |= 0xFF000000;
#endif

			// Calc the scale for this 
			float fScale = (xVertexA.m_fDensity * fEdgeA) + (xVertexB.m_fDensity * fEdgeB) + (xVertexC.m_fDensity * fEdgeC);
			// We only scale then down to fMinScale, not 0
			fScale = ((1.0f - fScale) * pxStyle->m_fScaleMin) + fScale;
			const uint32 uiScale = uint32( 255.0f * fScale);

			// Position and colour
			for( uint32 uiCornerIndex = 0; uiCornerIndex < 4; uiCornerIndex += 1, pxVerts += 1, uVertexOriginalMaxCount += 1)
			{		
				Asura_Direct3D_CGR_StreamVertexSimple& xVertex = *pxVerts;
					
				xVertex.m_fX = xPositionCenter[0];
				xVertex.m_fY = xPositionCenter[1];
				xVertex.m_fZ = xPositionCenter[2];

#if defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
				xVertex.m_afDiffuse[0] = xColour.x;
				xVertex.m_afDiffuse[1] = xColour.y;
				xVertex.m_afDiffuse[2] = xColour.z;
				xVertex.m_afDiffuse[3] = 1.0f;
#else
				xVertex.m_uDiffuse = uiColour;
#endif

				xVertex.m_ucScale = static_cast<uint8>( uiScale );
				xVertex.m_ucCornerIndex = static_cast<uint8>( uiCornerIndex );
				xVertex.m_ucWaveTableIndex = static_cast<uint8>( uiWaveIndex );
				xVertex.m_ucStyleIndex = static_cast<uint8>( uiStyleIndex );
			}
		}
		
	}
		
	return( true);
}

bool Asura_Direct3D_Carpet_Game_Render::RenderMeshBatch( const Asura_Direct3D_CGR_Mesh& xCGRMesh, const Asura_Direct3D_CGR_PolygonBatch& xCGRBatch)
{
	if( 0 == xCGRBatch.m_uVertexBufferSize)
	{
		return( true);
	}
	
	{
		uint32 uVertexBufferStart = xCGRBatch.m_uVertexBufferOffset;
		uint32 uVertexBufferStop = xCGRBatch.m_uVertexBufferOffset + xCGRBatch.m_uVertexBufferSize;
		uint32 uIndexBufferStart = (6 * (xCGRBatch.m_uVertexBufferOffset / 4));
		uint32 uIndexBufferStop = uIndexBufferStart + (6 * (xCGRBatch.m_uVertexBufferSize / 4));
		
		Asura_DIPCache_DIPData xData;
		
		xData.Set( uIndexBufferStart, uIndexBufferStop, uVertexBufferStart, uVertexBufferStop);
		
		s_xDIPCache.AddDIP( xData);
	}
	
	return( true);
}

template< class Tcamera> bool Asura_Direct3D_Carpet_Game_Render::RenderMesh( const Asura_Carpet_Mesh_Data_Frame& xDataFrame, const Tcamera& tCamera, Asura_Direct3D_CGR_Mesh& xCGRMesh)
{
	
	HRESULT xResult;
	// DLH - 03 February 2006 15:25:06 - Draw a batch at a time using the tree culling and Batching
#if defined( ASURA_CARPET_MESHMANAGER_PROFILE)	
	int iOverallTimer	= Asura_Profiling_CreateTimer();

	Asura_Profiling_StartTiming( iOverallTimer );
#endif

	const uint32 uiBatchMaxCount = 0x200;
	uint32 uiBatchCount = uiBatchMaxCount;
	uint32 auiBatchIndices[ uiBatchMaxCount];
	
	{
		CalcVisibleBatches< Tcamera, true>( tCamera, xCGRMesh.m_pxRTree, auiBatchIndices, uiBatchCount);
		
		#if defined( ASURA_CARPET_MESHMANAGER_PROFILE)	
			{
				const float fOverallTime = Asura_Timers::TicksToSeconds( Asura_Profiling_StopTiming( iOverallTimer ) );

				TOOLS_CARPET_STATISTICS_SET( s_fTimeRTree, fOverallTime);
				
				Asura_Profiling_DestroyTimer(iOverallTimer);
			}
		#endif
		if( 0 == uiBatchCount)
		{
			return( true);
		}

		SetTextureID( xCGRMesh.m_iTexture);
		
		if( NULL == xCGRMesh.m_pxBatchVertexBuffer)
		{
			CreateVertexBufferMesh( xCGRMesh);
		}
		
		s_xDIPCache.Clear();
			
		xResult = Asura_Direct3D::D3DDevice->SetStreamSource( 0, xCGRMesh.m_pxBatchVertexBuffer, 0, sizeof(Asura_Direct3D_CGR_StreamVertexSimple));
		ASURA_D3D_OK( xResult);	
		
		for( uint32 uiBatchIndicesIndex = 0; uiBatchIndicesIndex < uiBatchCount; uiBatchIndicesIndex += 1)
		{
			const uint32 uiBatchIndex = auiBatchIndices[ uiBatchIndicesIndex];
			Asura_Assert( uiBatchIndex < xCGRMesh.m_uPolygonBatchCount, "CGR Polygon Batch out of range");
			
			Asura_Direct3D_CGR_PolygonBatch& xCGRBatch = xCGRMesh.m_pxPolygonBatches[ uiBatchIndex];
			
			if( false == BatchAddToBuffers( xCGRMesh, xCGRBatch))
			{
				// DLH - Batch not empty, assume that there is not enough room in the VB so clear it out and continue
				s_xDIPCache.Render();
				s_xDIPCache.Clear();
				
				xCGRMesh.m_xVBRangeManager.DeleteAllBatches();
				
				if( false == BatchAddToBuffers( xCGRMesh, xCGRBatch))
				{
					continue;
				}
			}
			
			RenderMeshBatch( xCGRMesh, xCGRBatch);
		}
		
		s_xDIPCache.Render();
	}
		
	return( true);
}

bool Asura_Direct3D_Carpet_Game_Render::Render()
{
	D3DPERFCALL(BeginEvent(0xff889944, L"Carpet Game Render"));
	
	const Asura_Carpet_Mesh_Data_Frame xDataFrame( Asura_Timers::GetLocalScaledTime());
	
	bool bResetRenderStates = false;

#if defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	{
		bool bRenderStatesInitialised = false;
					
		for( uint32 uiMeshIndex = 0; uiMeshIndex < s_uMeshesCount; uiMeshIndex += 1)
		{
			Asura_Direct3D_CGR_Mesh& xCGRMesh = s_pxMeshes[ uiMeshIndex];
			
			const bool bIsMeshVisible = Asura_Camera::IsAABBInFrustum( xCGRMesh.m_xBoundingBox);
			if( false == bIsMeshVisible)
			{
				continue;
			}
	
			if( false == bRenderStatesInitialised)
			{
				if( false == BeginRenderStatic( xDataFrame))
				{
					continue;
				}
				bRenderStatesInitialised = true;
				bResetRenderStates = true;
			}
		
			for( uint32 uiStyleIndex = 0; uiStyleIndex < xCGRMesh.m_uStyleCount; uiStyleIndex += 1)
			{
				const Asura_Direct3D_CGR_Style& xStyle = xCGRMesh.m_pxStyles[ uiStyleIndex];
				
				RenderSetShaderConstantsStyles( uiStyleIndex, xStyle);
			}
			
			Asura_Direct3D_CGR_Camera xCamera;
			
			xCamera.GetCurrentCamera( xCGRMesh.m_fLodFadeFar);
			
			RenderMesh( xDataFrame, xCamera, xCGRMesh);
		}
	}
#else
	{
		bool bRenderStatesInitialised = false;
					
		for( uint32 uiMeshIndex = 0; uiMeshIndex < s_uMeshesCount; uiMeshIndex += 1)
		{
			Asura_Direct3D_CGR_Mesh& xCGRMesh = s_pxMeshes[ uiMeshIndex];
			
			const bool bIsMeshVisible = Asura_Camera::IsAABBInFrustum( xCGRMesh.m_xBoundingBox);
			if( false == bIsMeshVisible)
			{
				continue;
			}
	
			if( false == bRenderStatesInitialised)
			{
				if( false == BeginRenderStatic( xDataFrame))
				{
					continue;
				}
				bRenderStatesInitialised = true;
				bResetRenderStates = true;
			}
		
			for( uint32 uiStyleIndex = 0; uiStyleIndex < xCGRMesh.m_uStyleCount; uiStyleIndex += 1)
			{
				const Asura_Direct3D_CGR_Style& xStyle = xCGRMesh.m_pxStyles[ uiStyleIndex];
				
				RenderSetShaderConstantsStyles( uiStyleIndex, xStyle);
			}
			
			Asura_Direct3D_CGR_Camera xCamera;
			
			xCamera.GetCurrentCamera( xCGRMesh.m_fLodFadeFar);
			
			RenderMesh( xDataFrame, xCamera, xCGRMesh);
		}
	}
	
	{
		Asura_TrackingHandlePool<Asura_Photon>* pxPhotonList = Asura_Photon_System::GetDynamicPhotonList();

		if( NULL != pxPhotonList)
		{
			// DLH - 09 June 2006 16:03:54 - Find out if there are any Dynamic Photon Emitters
			Asura_TrackingHandlePool<Asura_Photon>::Iterator xIterator = pxPhotonList->GetIterator();
				
			for (xIterator.Begin(); !xIterator.End(); xIterator.Next())
			{
				Asura_Photon* pxPhoton = xIterator.GetCurrentPointer();

				if( false == pxPhoton->IsActive()) 
				{
					continue;
				}
				
				const float fRange = pxPhoton->GetRange();
				if( fRange <= 0.0f) 
				{
					continue;
				}

// DLH - 27 June 2006 17:01:02 - Does not return a valid AABB
//				const Asura_Bounding_Box& xPhotonBoundingBox = pxPhoton->GetBoundingBox();
				const Asura_Vector_3& xPhotonPosition = pxPhoton->GetPosition();
				Asura_Bounding_Box xPhotonBoundingBox;
				xPhotonBoundingBox.SetToPoint( xPhotonPosition);
				xPhotonBoundingBox.Grow( fRange);
				
				const bool bIsPhotonVisible = Asura_Camera::IsAABBInFrustum( xPhotonBoundingBox);
				if( false == bIsPhotonVisible)
				{
					continue;
				}
				
				{
					bool bRenderStatesInitialised = false;
					
					for( uint32 uiMeshIndex = 0; uiMeshIndex < s_uMeshesCount; uiMeshIndex += 1)
					{
						Asura_Direct3D_CGR_Mesh& xCGRMesh = s_pxMeshes[ uiMeshIndex];
						
						
						const bool bIsMeshInPhotonBox = xPhotonBoundingBox.Intersects( xCGRMesh.m_xBoundingBox);
						if( false == bIsMeshInPhotonBox)
						{
							continue;
						}
						
						
						const bool bIsMeshVisible = Asura_Camera::IsAABBInFrustum( xCGRMesh.m_xBoundingBox);
						if( false == bIsMeshVisible)
						{
							continue;
						}
	
						// DLH - 27 June 2006 16:52:40 - Only Initialise the render states if we have something to draw
						if( false == bRenderStatesInitialised)
						{					
							if( false == BeginRenderDynamic( xDataFrame, *pxPhoton))
							{
								continue;
							}
							bRenderStatesInitialised = true;
							bResetRenderStates = true;
						}
						
						for( uint32 uiStyleIndex = 0; uiStyleIndex < xCGRMesh.m_uStyleCount; uiStyleIndex += 1)
						{
							const Asura_Direct3D_CGR_Style& xStyle = xCGRMesh.m_pxStyles[ uiStyleIndex];
							
							RenderSetShaderConstantsStyles( uiStyleIndex, xStyle);
						}
						
						Asura_Direct3D_CGR_Camera_And_Box xCamera;
						
						xCamera.GetCurrentCamera( xCGRMesh.m_fLodFadeFar);
						xCamera.m_xBoundingBox = xPhotonBoundingBox;
						
						RenderMesh( xDataFrame, xCamera, xCGRMesh);
					}
				}
				
			}
		}

	}
#endif //defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	
	if( true == bResetRenderStates)
	{
		EndRender();
	}

	D3DPERFCALL(EndEvent());
	
	return( true);
}


bool Asura_Direct3D_Carpet_Game_Render::BeginRenderStatic( const Asura_Carpet_Mesh_Data_Frame& xDataFrame)
{
	HRESULT xResult;
	// Temporary render function, using the rendering code from the tools, until we get the 
	// proper optimized renderer (with scene management, batching, and using VU/shaders)

	// Set some renderstates
	xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	ASURA_D3D_OK( xResult);	
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS_2X);
	Asura_Render::CurrentStates.RequestAlphaTest( true , 0x00);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);

	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );

	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
	ASURA_D3D_OK( xResult);	
	
	//xResult = Asura_Direct3D::D3DDevice->SetVertexDeclaration( s_pxVertexDeclarationSimple[ 0]);
	//ASURA_D3D_OK( xResult);	
	
#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	if( true == Asura_Fog::IsGraphFogEnabled())
	{
		Asura_Direct3D_Shaders::RequestVertexShader( ASURA_D3D_VS_CARPET_FOGG_LIGHTS);
	}
	else
	{
		Asura_Direct3D_Shaders::RequestVertexShader( ASURA_D3D_VS_CARPET_FOGL_LIGHTS);
	}
	Asura_Direct3D_Shaders::RequestPixelShader( ASURA_D3D_PS_CARPET_LIGHTS);
#else //ASURA_DIRECT3D_BGR_SUPPORT_SM30

	xResult = Asura_Direct3D::D3DDevice->SetVertexDeclaration( s_pxVertexDeclarationSimple);
	ASURA_D3D_OK( xResult);
	
	Asura_Direct3D_Effect* pxEffect = Asura_Direct3D_Effects_System::BeginEffect( "Carpet" );
	
	if ( NULL == pxEffect || false == pxEffect->IsD3DEffectValid())
	{
		return( false);
	}
	
	{
		Direct3D_Effect* pxD3DEffect = pxEffect->GetD3DEffect();
		if( NULL != pxD3DEffect)
		{
			xResult = pxD3DEffect->BeginPass( 0);
			ASURA_D3D_OK( xResult);
			
			xResult = pxD3DEffect->CommitChanges();
			ASURA_D3D_OK( xResult);
		}
	}
	
#endif //ASURA_DIRECT3D_BGR_SUPPORT_SM30

	xResult = Asura_Direct3D::D3DDevice->SetIndices( s_pxBatchIndexBuffer);
	ASURA_D3D_OK( xResult);	
	
#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	// turn on fog if we have to...
	if (Asura_Fog::IsFogRenderable())
	{
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);
		// Set the fog color.
		xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGCOLOR, Asura_Fog::GetFogColourAsUint());
		ASURA_D3D_OK( xResult);
	}
#endif
	
	RenderSetShaderConstantsStatic( xDataFrame, NULL);
	
	Asura_Render::CurrentStates.ResetTextureIndexToLast( 1);
	
	s_iDrawCacheTextureID = -1;
	s_uTimeStampLast += 1;
	
	return( true);
}

#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
bool Asura_Direct3D_Carpet_Game_Render::BeginRenderDynamic( const Asura_Carpet_Mesh_Data_Frame& xDataFrame, const Asura_Photon& xPhoton)
{
	HRESULT xResult;
	// Temporary render function, using the rendering code from the tools, until we get the 
	// proper optimized renderer (with scene management, batching, and using VU/shaders)

	if( ASURA_TEXTURE_INVALID == s_iAttenuationMap)
	{
		s_iAttenuationMap = Asura_Texture_Management::AddTextureWithResetGroup("\\graphicnovel\\expattenuation.dds", ASURA_TEXTURE_LOADED_PERMANENT);
	}
	
	// Set some renderstates
	xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	ASURA_D3D_OK( xResult);	
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_ADD);
	Asura_Render::CurrentStates.RequestAlphaTest( true , 0x00);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);

	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
	ASURA_D3D_OK( xResult);	
	
	//xResult = Asura_Direct3D::D3DDevice->SetVertexDeclaration( s_pxVertexDeclarationSimple[ 0]);
	//ASURA_D3D_OK( xResult);	
	
	if( true == Asura_Fog::IsGraphFogEnabled())
	{
		Asura_Direct3D_Shaders::RequestVertexShader( ASURA_D3D_VS_CARPET_FOGG_LIGHTD);
	}
	else
	{
		Asura_Direct3D_Shaders::RequestVertexShader( ASURA_D3D_VS_CARPET_FOGL_LIGHTD);
	}
	
	Asura_Direct3D_Shaders::RequestPixelShader( ASURA_D3D_PS_CARPET_LIGHTD);
	
	xResult = Asura_Direct3D::D3DDevice->SetIndices( s_pxBatchIndexBuffer);
	ASURA_D3D_OK( xResult);	
	
	// turn on fog if we have to...
	if (Asura_Fog::IsFogRenderable())
	{
		// Set the fog color.
		xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);
		ASURA_D3D_OK( xResult);
		xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGCOLOR, 0x00000000);
		ASURA_D3D_OK( xResult);
	}
	
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	//Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_EQUALS);

	Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
	Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );

	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
	Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
	
	Asura_Render::CurrentStates.RequestTextureIndex( 1, s_iAttenuationMap);
	Asura_Render::CurrentStates.RequestTextureIndex( 2, s_iAttenuationMap);
	
	Asura_Render::CurrentStates.ResetTextureIndexToLast( 3);
	
	RenderSetShaderConstantsStatic( xDataFrame, &xPhoton);
	
	s_iDrawCacheTextureID = -1;
	s_uTimeStampLast += 1;
	
	return( true);
}
#endif //!defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)

void Asura_Direct3D_Carpet_Game_Render::EndRender()
{
	HRESULT xResult;

#if defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	{
		Asura_Direct3D_Effect* pxEffect = Asura_Direct3D_Effects_System::BeginEffect( "Carpet" );
		
		if ( NULL != pxEffect && true == pxEffect->IsD3DEffectValid())
		{
			Direct3D_Effect* pxD3DEffect = pxEffect->GetD3DEffect();
			if( NULL != pxD3DEffect)
			{
				xResult = pxD3DEffect->EndPass();
				ASURA_D3D_OK( xResult);
			}
			Asura_Direct3D_Effects_System::EndEffect( *pxEffect);
		}
	}
#endif
	
	
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_FOGENABLE, FALSE);
	Asura_Render::CurrentStates.RequestAlphaTest( false);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	
	Asura_Render::CurrentStates.ResetTextureIndexToLast( 1);

#if !defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
	if (Asura_Fog::IsFogRenderable())
	{
		// Set the fog color.
//		xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGENABLE, TRUE);
//		ASURA_D3D_OK( xResult);
		xResult = Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_FOGCOLOR, Asura_Fog::GetFogColourAsUint());
		ASURA_D3D_OK( xResult);
	}
#endif

	// DaveR - 20/02/06
	//		   These states are unsupported by Asura, so they need restoring here, otherwise
	//         it messes up platform-independent HUD/FE rendering, where there is no
	//		   platform-independent way to restore them to Asura's assumed defaults.	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
	ASURA_D3D_OK( xResult);	
	xResult = Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_SRGBTEXTURE, FALSE );
	ASURA_D3D_OK( xResult);	
}

void Asura_Direct3D_Carpet_Game_Render::SetTextureID( const int iTextureID)
{
	if( iTextureID != s_iDrawCacheTextureID)
	{
		Asura_Render::CurrentStates.RequestTextureIndex( 0, iTextureID);					
		s_iDrawCacheTextureID = iTextureID;
	}
}

Asura_Carpet_Mesh_Data_Frame::Asura_Carpet_Mesh_Data_Frame( const float fTime) :
m_fTime( fTime)
{

	Asura_Vector_3 xRight;
	Asura_Camera::GetRightVector(xRight);	
	xRight.y = 0;
	xRight.Normalise();

	float fScaledTime = fTime * Asura_Carpet_MeshManager::GetMovementSpeed();

	const u_int uiSizeWaveX = ASURA_CARPET_MESH_DATA_FRAME_SIZE_WAVEX;
	for ( u_int i = 0; i < uiSizeWaveX; i++)
	{		
		float fS = (float)i;
		// DLH - 01 February 2006 16:21:02 - Added .0f to all constant values, requested by JGP
		// DLH - 01 February 2006 16:21:27 - Changed sin to Asura_Maths::Sine(), requested by JGP
		  float fAmount = 0.1f  * Asura_Maths::Sine( (1.0f * fS) + fScaledTime * (1.0f + (fS * 0.05f)) )
					    + 0.05f * Asura_Maths::Sine( (2.0f * fS) + fScaledTime * (1.0f + ((16.0f - fS) * 0.05f)));		

		this->m_afWaveX[i] = fAmount * xRight.x;
		this->m_afWaveY[i] = fAmount * xRight.y;
		this->m_afWaveZ[i] = fAmount * xRight.z;
	}

	Asura_Camera::GetForwardVector( this->m_xCameraForwards);
	this->m_xCameraPosition = *Asura_Camera::GetPosition();

	// Use camera pos/fwd to get plane eq, and use this to get distance to centre 
	// of triangle
	this->m_fCameraPlaneDistance = -(this->m_xCameraForwards * this->m_xCameraPosition);

	// JAT 13 Feb 2006 - you don't need to do all this maths to get a view horizontal vector
	/*
	Asura_Matrix_3x3 xMat = *Asura_Camera::GetOrientation();
	Asura_Euler_3 xOri;
	xOri.FromMatrix(xMat);
	xOri.z = xOri.x = 0.0f;
	xOri.ToMatrix(xMat);
	*/
	// JAT 13 Feb 2006 - get the view x vector from the view matrix
	// might go wrong if the camera has roll on it
	const Asura_Matrix_3x3* const xView = Asura_Camera::GetOrientation();
	this->m_xBillboardRight.x = xView->e11;
	this->m_xBillboardRight.y = xView->e21;
	this->m_xBillboardRight.z = xView->e31;

	// DLH - 01 February 2006 16:21:02 - Added .0f to all constant values, requested by JGP
	this->m_xBillboardUp = Asura_Vector_3(0.0f,-1.0f,0.0f);

}


//*******************************************************************
//
// (c) 2006 Rebellion (Derby)
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************

#endif //defined( ASURA_USE_CARPET)
