//////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_Stream_WindowsFile.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     Stream for reading / writing using windows file by HANDLE
//
//////////////////////////////////////////////////////////////////////////////


#include "Core/Asura.h"
#include "ChunkSystem/Asura_Chunk_Stream_WindowsFile.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"


Asura_Chunk_Stream_WindowsFile::Asura_Chunk_Stream_WindowsFile(HANDLE xFile)
{ 
	m_xFile = xFile; 
	m_uBufferPosition = 0;
	m_uBufferStartFilePosition = 0;
	m_bOwnsFileHandle = false;
	m_bAddAsuraHeaderAndTerminator = false;

}

Asura_Chunk_Stream_WindowsFile::Asura_Chunk_Stream_WindowsFile()
{
	m_xFile = INVALID_HANDLE_VALUE;
	m_uBufferPosition = 0;
	m_uBufferStartFilePosition = 0;
	m_bOwnsFileHandle = false;
	m_bAddAsuraHeaderAndTerminator = false;
}

Asura_Chunk_Stream_WindowsFile::~Asura_Chunk_Stream_WindowsFile()
{
	if(m_uBufferPosition)
	{
		//Flush any remaining output.
		FlushOutput();
	}
	if(m_bOwnsFileHandle)
	{
		Asura_Assert(0, "File should have either been closed or deleted by now.");
		DeleteFile();
	}
}

bool Asura_Chunk_Stream_WindowsFile::CreateFile(const char* szFilename,bool bAddAsuraHeaderAndTerminator)
{
	if(m_xFile != INVALID_HANDLE_VALUE)
	{
		Asura_Assert(0,"Already have a file handle");
		return false;
	}

	//Make sure the file isn't readonly
	DWORD uAttributes = GetFileAttributes(szFilename);
	if((uAttributes & FILE_ATTRIBUTE_READONLY) && uAttributes!=0xffffffff)
	{
		return false;
	}

	//Get a temporary filename to save under
	m_xFilename = szFilename;

	//create a temporary file to save to
	Asura_String xPath,xName;
	m_xFilename.SplitFilename(&xPath, &xName, 0);

	char acTempFilename[MAX_PATH];
	if(!GetTempFileName(xPath.GetString(), xName.GetString(), 0, acTempFilename))
	{
		Asura_Assert(0,"Error generating temporary filename");
		return false;
	}
	m_xTempFilename = acTempFilename;

	//and open this file
	m_xFile = ::CreateFile(acTempFilename, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, 0 , 0);
	if(m_xFile == INVALID_HANDLE_VALUE)
	{
		Asura_Assert(0, "Error opening temporary file for writing");
		return false;
	}

	m_bOwnsFileHandle = true;
	m_bAddAsuraHeaderAndTerminator = bAddAsuraHeaderAndTerminator;

	if(bAddAsuraHeaderAndTerminator)
	{
		Write(ASURA_UNCOMPRESSED_FILE_ID_ACTUAL_CHARS, ASURA_UNCOMPRESSED_FILE_ID);
	}
	return true;
}

bool Asura_Chunk_Stream_WindowsFile::CloseFile()
{
	if(!m_bOwnsFileHandle)
	{
		Asura_Assert(0,"Only expecting Asura_Chunk_Stream_WindowsFile::CloseFile() to be used for streams opened with CreateFile()");
		return false;
	}
	//add the terminator if needed
	if(m_bAddAsuraHeaderAndTerminator)
	{
		u_int uZero = 0;
		(*this) << uZero;
		m_bAddAsuraHeaderAndTerminator = false;
	}
	//flush the output and close the file
	FlushOutput();
	if(!CloseHandle(m_xFile))
	{
		Asura_Assert(0, "Error closing file");
		return false;
	}
	m_xFile = INVALID_HANDLE_VALUE;
	m_bOwnsFileHandle = false;

	bool bOkay = true;
	
	//see if a file already exists with the destination filename
	HANDLE hFile = ::CreateFile( m_xFilename.GetString(), GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY, NULL );
    if( INVALID_HANDLE_VALUE != hFile )
	{
        CloseHandle( hFile );

		//delete this file
		if(!::DeleteFile(m_xFilename.GetString()))
		{
			Asura_Assertf(0,("Failed to delete %s so that the temporary file could be renamed", m_xFilename.GetString()));
			::DeleteFile(m_xTempFilename.GetString());
			bOkay = false;
		}
    }

	if(bOkay)
	{
		//rename the temporary file	
		if(!MoveFile(m_xTempFilename.GetString(), m_xFilename.GetString()))
		{
			Asura_Assertf(0,("Failed to rename %s to %s",m_xTempFilename.GetString(), m_xFilename.GetString()));
			bOkay = false;
		}
	}
	return bOkay;
}

void Asura_Chunk_Stream_WindowsFile::DeleteFile()
{
	if(!m_bOwnsFileHandle)
	{
		Asura_Assert(0,"Only expecting Asura_Chunk_Stream_WindowsFile::DeleteFile() to be used for streams opened with CreateFile()");
	}

	CloseHandle(m_xFile);
	::DeleteFile(m_xTempFilename.GetString());

	m_bOwnsFileHandle = false;
	m_xFile = INVALID_HANDLE_VALUE;
}


u_int Asura_Chunk_Stream_WindowsFile::Read(u_int uNumberOfBytes, void* pData)
{ 
	Asura_Assert(false, "not implemented"); 
	return 0; 
}
u_int Asura_Chunk_Stream_WindowsFile::Skip(u_int uNumberOfBytes)
{ 
	Asura_Assert(false, "not implemented"); 
	return 0; 
}
u_int Asura_Chunk_Stream_WindowsFile::Write(u_int uNumberOfBytes, const void* pData)
{ 

	if(uNumberOfBytes + m_uBufferPosition > uBUFFER_SIZE)
	{
		//Not enough buffer space left , so write out what we have so far
		FlushOutput();
		if(uNumberOfBytes > uBUFFER_SIZE)
		{
			//Too much being written for the buffer , so just write out directly
			DWORD uBytesWritten;
			WriteFile( m_xFile, pData, uNumberOfBytes, &uBytesWritten, 0 ); 
			Asura_Assert( uBytesWritten == uNumberOfBytes, "didn't write all data" );

			m_iChunkSize += uNumberOfBytes;
	
			return uBytesWritten;
		}
	}
	
	if(!m_uBufferPosition)
	{
		//First lot of data going into the buffer now , so store the file position to make
		//sure it doesn't change
		m_uBufferStartFilePosition = GetFilePosition();
	}

	//add to the buffer
	Asura_Mem::Copy(m_acBuffer + m_uBufferPosition, pData, uNumberOfBytes);
	m_uBufferPosition += uNumberOfBytes;
	
	this->m_iChunkSize += uNumberOfBytes;
	
	return uNumberOfBytes;
}

u_int Asura_Chunk_Stream_WindowsFile::Write(u_int uNumberOfBytes, const void* pData, u_int uPosition)
{ 
	u_int uStartFilePosition = GetFilePosition();

    //For simplicity ,going to assume that the write to position is never used to extend the file for now.
	if ( uPosition+uNumberOfBytes > uStartFilePosition + m_uBufferPosition )
    {
        Asura_Assert( false, "Asura_Chunk_Stream_WindowsFile::Write - Not expecting to write beyond the current position with this Write() function" );
        return 0;
    }

	const char* pcSource = static_cast<const char*>(pData);
	u_int uBytesLeft = uNumberOfBytes;
	
	if(uPosition < uStartFilePosition)
	{
		//Some/all of the data being written has already gone to the file , so update the file directly
		SetFilePosition(uPosition);

		u_int uBytesToWrite = Asura_Maths::Min(uStartFilePosition-uPosition, uBytesLeft);

		DWORD uBytesWritten;
		WriteFile( m_xFile, pcSource, uBytesToWrite, &uBytesWritten, 0 ); 
		Asura_Assert( uBytesWritten == uBytesToWrite, "didn't write all data" );

		uPosition += uBytesToWrite;
		uBytesLeft -= uBytesToWrite;
		pcSource += uBytesToWrite;

		SetFilePosition(uStartFilePosition);
	}

	if(uBytesLeft)
	{
		//any remaining bytes to write should go to the buffer
		Asura_Assert(uPosition>=uStartFilePosition, "");

		u_int uBufferOffset = uPosition - uStartFilePosition;
		Asura_Mem::Copy(m_acBuffer + uBufferOffset, pcSource, uBytesLeft);
		uBytesLeft = 0;
	}

	return uNumberOfBytes - uBytesLeft; 
}
u_int Asura_Chunk_Stream_WindowsFile::GetPosition(void) const
{ 
	return GetFilePosition() + m_uBufferPosition; 
}

int Asura_Chunk_Stream_WindowsFile::PeekInt() 
{ 
	Asura_Assert(false, "CHUNK ERROR: Asura_Chunk_Stream_WindowsFile::PeekInt - Peeking not allowed on Windows File Streams"); 
	return 0; 
}

u_int Asura_Chunk_Stream_WindowsFile::GetFilePosition(void) const
{ 
	LONG uDistanceToMoveLow = 0;
	LONG uDistanceToMoveHigh = 0;
	
	DWORD uNewFilePointer = SetFilePointer( this->m_xFile, uDistanceToMoveLow, &uDistanceToMoveHigh, FILE_CURRENT);
	
	Asura_Assertf( INVALID_SET_FILE_POINTER != uNewFilePointer, ("Error Getting the File Pointer. Error 0x%08X", GetLastError()));
	Asura_Assert( 0 == uDistanceToMoveHigh, "This file might need more than 4 byte offsets This is not Supported.");
	
	return uNewFilePointer; 
}

void Asura_Chunk_Stream_WindowsFile::SetFilePosition( const u_int uPosition) const
{ 
	LONG uDistanceToMoveLow = uPosition;
	LONG uDistanceToMoveHigh = 0;
	
	const DWORD uNewFilePointer = SetFilePointer( this->m_xFile, uDistanceToMoveLow, &uDistanceToMoveHigh, FILE_BEGIN);
	(void)uNewFilePointer;
	
	Asura_Assertf( INVALID_SET_FILE_POINTER != uNewFilePointer, ("Error Getting the File Pointer. Error 0x%08X", GetLastError()));
	Asura_Assert( 0 == uDistanceToMoveHigh, "This file might need more than 4 byte offsets This is not Supported.");
	
	return; 
}

void Asura_Chunk_Stream_WindowsFile::FlushOutput()
{
	if(m_uBufferPosition)
	{
		//Complain if the file position isn't in the correct place.
		Asura_Assert(GetFilePosition() == m_uBufferStartFilePosition, "Asura_Chunk_Stream_WindowFile - File position changed behind the stream's back");

		
		DWORD uBytesWritten;
		WriteFile( m_xFile, m_acBuffer, m_uBufferPosition, &uBytesWritten, 0 ); 
		Asura_Assert( uBytesWritten == m_uBufferPosition, "didn't write all data" );

		m_uBufferPosition = 0;
	}		
}

