//*******************************************************************
//
// File: Asura_FSFX_DoubleVision.cpp        
//
// Created: 9/10/2007 
//
// Description: Handles the Double Vision fullscreen effect
//
//*******************************************************************

// system libraries
#include "Core/Asura.h"

#if defined(ASURA_USE_FSFX) && defined (ASURA_TARGET_SM3)

// main header
#include "Core/Asura_Timers.h"

#include "FSFX/Asura_FSFX_System.h"
#include "FSFX/Asura_FSFX_Template.h"
#include "FSFX/Asura_FSFX_Component.h"

#include "Rendering/Asura_ViewPort.h"
#include "Rendering/Asura_Render.h"
#include "Camera/Asura_Camera.h"

#include "Rendering/Asura_Direct3D_RenderTargetManager.h"
#include "Rendering/Asura_Direct3D_BackBufferCapture.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Direct3D_Effects_System.h"

// other libraries

// internal dependencies

//*******************************************************************
// Constants
//*******************************************************************


//*******************************************************************
// Local Classes
//*******************************************************************


struct DoubleVisionVertex
{
	float m_fX, m_fY, m_fZ, m_fW;
	float m_fU, m_fV;
};


const Direct3D_VertexElement axVERTEX_DECLARATION[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 16, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	D3DDECL_END()
};

static Direct3D_VertexDeclaration* g_pxVertexDecl = NULL;



//*******************************************************************
// Function Definitions
//*******************************************************************

static Asura_Direct3D_Effect* g_pxEffect;
static D3DXHANDLE g_xInitialImageHandle;
static D3DXHANDLE g_xXOffsetHandle1;
static D3DXHANDLE g_xYOffsetHandle1;
static D3DXHANDLE g_xXOffsetHandle2;
static D3DXHANDLE g_xYOffsetHandle2;

static void CaptureEffect();


#if ASURA_LOAD_EFFECT_FILES == 1
static void OnEffectLoaded(Asura_Direct3D_Effect* pxEffect)
{
	CaptureEffect();
}

static void OnEffectUnloaded(Asura_Direct3D_Effect* pxEffect)
{
	g_pxEffect = NULL;
}
#endif

static void CaptureEffect()
{
	g_pxEffect = Asura_Direct3D_Effects_System::GetEffect("FSFX_SM30_DoubleVision");

	if (g_pxEffect)
	{
#if ASURA_LOAD_EFFECT_FILES == 1
		g_pxEffect->SetOnLoadCallback(OnEffectLoaded);
		g_pxEffect->SetOnUnloadCallback(OnEffectUnloaded);
#endif


		g_xInitialImageHandle = g_pxEffect->GetD3DEffect()->GetParameterByName( NULL, "g_xTexture" );
		g_xXOffsetHandle1 = g_pxEffect->GetD3DEffect()->GetParameterByName( NULL, "g_fXOffset1" );
		g_xYOffsetHandle1 = g_pxEffect->GetD3DEffect()->GetParameterByName( NULL, "g_fYOffset1" );
		g_xXOffsetHandle2 = g_pxEffect->GetD3DEffect()->GetParameterByName( NULL, "g_fXOffset2" );
		g_xYOffsetHandle2 = g_pxEffect->GetD3DEffect()->GetParameterByName( NULL, "g_fYOffset2" );
	}
}


void Asura_FSFX_Component_SM30_DoubleVision::Platform_Initialise()
{
	// when doing batch exporting of levels sometimes we don't have a D3D device to use
	if ( NULL == Asura_Direct3D::D3DDevice )
	{
		return;
	}

	CaptureEffect();

	Asura_Direct3D::D3DDevice->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxVertexDecl );
}



void Asura_FSFX_Component_SM30_DoubleVision::Platform_Reset()
{
	Platform_Shutdown();
	Platform_Initialise();
}


void Asura_FSFX_Component_SM30_DoubleVision::Platform_Shutdown()
{
	// when doing batch exporting of levels sometimes we don't have a D3D device to use
	if ( NULL == Asura_Direct3D::D3DDevice )
	{
		return;
	}
	g_pxEffect = NULL;
	g_pxVertexDecl->Release();
	g_pxVertexDecl = NULL;

}

void Asura_FSFX_Component_SM30_DoubleVision::Platform_Render( const float fIntensity )
{
	Asura_Direct3D_Effect* const pxEffect = g_pxEffect;
	if( !pxEffect || !pxEffect->IsD3DEffectValid() ) return;
	SCOPEDPIXEVENT(( D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF ), L"FSFX: Double Vision"));

	// Interpolate from zero to set parameters
	const float fXOffset1 = m_fOrigXOffset1 * fIntensity;
	const float fYOffset1 = m_fOrigYOffset1 * fIntensity;
	const float fXScale1 = m_fOrigXScale1 * fIntensity;
	const float fYScale1 = m_fOrigYScale1 * fIntensity;
	const float fXOffset2 = m_fOrigXOffset2 * fIntensity;
	const float fYOffset2 = m_fOrigYOffset2 * fIntensity;
	const float fXScale2 = m_fOrigXScale2 * fIntensity;
	const float fYScale2 = m_fOrigYScale2 * fIntensity;

	// Setup
	const float fTime = Asura_Timers::GetLocalScaledTime();
	Asura_Direct3D_BackBufferCapture::CaptureCurrentBackBuffer();

	pxEffect->SetParameterTexture( g_xInitialImageHandle, Asura_Direct3D_BackBufferCapture::GetCaptureTexture() );

	Asura_Render::AllowFullScreenRendering();

	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );

	Asura_Direct3D::D3DDevice->SetVertexDeclaration( g_pxVertexDecl );

	const float fXMultiplier = 1.0f / 100.0f;
	const float fYMultiplier = 1.0f / 100.0f;

	// Apply phase and speed scale to time
	const float fTime1 = fTime * m_fSpeed1 + m_fPhase1;
	const float fTime2 = fTime * m_fSpeed2 + m_fPhase2;

	float fXDist1;
	float fYDist1;
	float fXDist2;
	float fYDist2;

	Asura_Maths::SinCos( fTime1, fXDist1, fYDist1 );
	Asura_Maths::SinCos( fTime2, fXDist2, fYDist2 );
	fXDist1 *= fXScale1;
	fYDist1 *= fYScale1;
	fXDist2 *= fXScale2;
	fYDist2 *= fYScale2;

	const float fXTotalOffset1 = (fXOffset1 + fXDist1) * fXMultiplier;
	const float fYTotalOffset1 = (fYOffset1 + fYDist1) * fYMultiplier;
	const float fXTotalOffset2 = (fXOffset2 + fXDist2) * fXMultiplier;
	const float fYTotalOffset2 = (fYOffset2 + fYDist2) * fYMultiplier;

	const float fXMaxOffset = Asura_Maths::Max( Asura_Maths::Abs(fXOffset1) + Asura_Maths::Abs(fXDist1) , Asura_Maths::Abs(fXOffset2) + Asura_Maths::Abs(fXDist2) ) * fXMultiplier;
	const float fYMaxOffset = Asura_Maths::Max( Asura_Maths::Abs(fYOffset1) + Asura_Maths::Abs(fYDist1) , Asura_Maths::Abs(fYOffset2) + Asura_Maths::Abs(fYDist2) ) * fYMultiplier;
	const float fXUVScale = 1.0f / (1.0f + fXMaxOffset );
	const float fYUVScale = 1.0f / (1.0f + fYMaxOffset );

	pxEffect->SetParameter( g_xXOffsetHandle1, fXTotalOffset1 );
	pxEffect->SetParameter( g_xYOffsetHandle1, fYTotalOffset1 );
	pxEffect->SetParameter( g_xXOffsetHandle2, fXTotalOffset2 );
	pxEffect->SetParameter( g_xYOffsetHandle2, fYTotalOffset2 );
	
	const float fPosLeft = -1.0f;
	const float fPosRight = 1.0f;
	const float fPosTop = -1.0f;
	const float fPosBottom = 1.0f;
	const float fUVLeft = 1.0f - fXUVScale;
	const float fUVRight = fXUVScale;
	const float fUVTop = 1.0f - fYUVScale;
	const float fUVBottom = fYUVScale;
	const DoubleVisionVertex axVerts[] =
	{
		{ fPosLeft,  fPosTop, 0.0f, 1.0f,		fUVLeft, fUVBottom },
		{ fPosRight, fPosTop, 0.0f, 1.0f,		fUVRight, fUVBottom },
		{ fPosRight, fPosBottom, 0.0f, 1.0f,	fUVRight,  fUVTop },
		{ fPosLeft, fPosBottom, 0.0f, 1.0f,		fUVLeft,  fUVTop },
	};

	Asura_Direct3D::BindDefaultTargets();
	Asura_Direct3D::D3DDevice->SetStreamSource(0, NULL, 0, sizeof( DoubleVisionVertex ));

	Asura_Direct3D_Effects_System::BeginEffect( *pxEffect );
	{
		pxEffect->BeginPass(0);
		pxEffect->GetD3DEffect()->CommitChanges();
		{
			Asura_Direct3D::D3DDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, axVerts, sizeof( DoubleVisionVertex ) );
		}
		pxEffect->EndPass();
	}
 	Asura_Direct3D_Effects_System::EndEffect(*pxEffect);

	Asura_Direct3D_Effects_System::FlushEffects();

	// Reset this parameter to NULL since the texture is in the default pool (for device resets)
	pxEffect->SetParameterTexture( g_xInitialImageHandle, NULL );

	////// Finished
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);

	Asura_Direct3D_Shaders::RequestPixelShader(ASURA_D3D_PS_NONE);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);

	Asura_Direct3D_BackBufferCapture::AdviseNextFrame();

	Asura_Render::ApplyCameraSettings();

	for( u_int u = 0; u < 16; ++u )
	{
		Asura_Direct3D::D3DDevice->SetSamplerState( 0, D3DSAMP_SRGBTEXTURE, FALSE );
	}
}


//*******************************************************************
//
// (c) 2007 Rebellion
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************

#endif //defined( ASURA_USE_FSFX)
