//////////////////////////////////////////////////////////////////////////////
//
// Asura_PC_Input_Joystick.cpp
//
// Created by AAM - 14/01/2004 14:36:17
//
// Copyright (c) 2004 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Input/Asura_PC_Input_Joystick.h"

#include "Core/Asura_Windows.h"
#include "Core/Asura_PlatformBasics.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Input/Asura_DirectInput.h"
#include "Input/Asura_PC_Input_XInput.h"

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const int iASURA_PC_JOYSTICK_RANGE = 1000;
const float fASURA_PC_JOYSTICK_RANGE = 1000.0f;
const float fASURA_PC_JOYSTICK_DEADZONE = 300.0f;
const u_int fASURA_PC_JOYSTICK_POV_UNSET = 0xffffffff;


/******************************************************************************/
/* Static Class Data                                                          */
/******************************************************************************/

int Asura_PC_Input_Joystick::s_iNumJoysticks = 0;
int Asura_PC_Input_Joystick::s_iNumXInputDevices = 0;
Asura_List<Asura_PC_Input_Joystick::JoystickDefinition>* Asura_PC_Input_Joystick::s_pxJoystickList = 0;
bool* Asura_PC_Input_Joystick::s_pbTempDevicesStatus = 0;

/******************************************************************************/
/* Class Functions                                                            */
/******************************************************************************/

Asura_PC_Input_Joystick::Asura_PC_Input_Joystick()
: m_fZAxis(0.0f)
, m_fXRot(0.0f)
, m_fYRot(0.0f)
, m_fZRot(0.0f)
, m_uiPOV(0)
, m_pxDevice(NULL)
{
	m_fHAxis[0] = m_fHAxis[1] = m_fVAxis[0] = m_fVAxis[1] = 0.0f;
	for (u_int u = 0; u < 16; ++u)
	{
		m_aucButtonPressed[u] = 0;
	}
}

// initialise a joystick
// pass the GUID that identifies which DirectInput device to use
// returns true for success, false if failure
bool Asura_PC_Input_Joystick::Initialise(REFGUID xDeviceGuid)
{
	if (!Asura_DirectInput::DIObject)
	{
		Asura_Assert(false, "Tried to set up joystick, but DirectInput has not been initialised.");
		return false;
	}

	if (m_pxDevice)
	{
		Asura_Assert(0,"device already initialised");
		return false;
	}

	HRESULT	hr;

	hr = Asura_DirectInput::DIObject->CreateDevice(xDeviceGuid, &m_pxDevice, NULL);
	if (FAILED(hr))
	{
		if (m_pxDevice)
		{
			m_pxDevice->Release();
			m_pxDevice = 0;
		}
		return false;
	}

	hr = m_pxDevice->SetDataFormat(&c_dfDIJoystick2);
	if (FAILED(hr))
	{
		if (m_pxDevice)
		{
			m_pxDevice->Release();
			m_pxDevice = 0;
		}
		return false;
	}

	hr = m_pxDevice->SetCooperativeLevel(Asura_Windows::GetWindowHandle(), DISCL_EXCLUSIVE | DISCL_FOREGROUND);
	if (FAILED(hr))
	{
		if (m_pxDevice)
		{
			m_pxDevice->Release();
			m_pxDevice = 0;
		}
		return false;
	}

	hr = m_pxDevice->EnumObjects(EnumObjectsCallback, this, DIDFT_AXIS | DIDFT_BUTTON | DIDFT_POV);
	if (FAILED(hr))
	{
		if (m_pxDevice)
		{
			m_pxDevice->Release();
			m_pxDevice = 0;
		}
		return false;
	}

	DIDEVICEINSTANCE xDeviceInstance;
	xDeviceInstance.dwSize = sizeof(DIDEVICEINSTANCE);
	hr = m_pxDevice->GetDeviceInfo(&xDeviceInstance);
	if (FAILED(hr))
	{
		if (m_pxDevice)
		{
			m_pxDevice->Release();
			m_pxDevice = 0;
		}
		return false;
	}

	m_xProductName	= xDeviceInstance.tszProductName;

	m_pxDevice->Acquire();

	Asura_OutputToDebugger(("Asura_PC_Input_Joystick::Initialise():\n"));
	Asura_OutputToDebugger(("- Instance-name = \"%s\"\n", xDeviceInstance.tszInstanceName));
	Asura_OutputToDebugger(("- Product-name  = \"%s\"\n", xDeviceInstance.tszProductName));
	Asura_OutputToDebugger(("- GUID          = %08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X\n", xDeviceGuid.Data1, xDeviceGuid.Data2, xDeviceGuid.Data3, xDeviceGuid.Data4[0], xDeviceGuid.Data4[1], xDeviceGuid.Data4[2], xDeviceGuid.Data4[3], xDeviceGuid.Data4[4], xDeviceGuid.Data4[5], xDeviceGuid.Data4[6], xDeviceGuid.Data4[7]));

	return true;
}

// shut a joystick down by releasing its DirectInput interface
void Asura_PC_Input_Joystick::Shutdown()
{
	if (m_pxDevice)
	{
		m_pxDevice->Unacquire();
		m_pxDevice->Release();
		m_pxDevice = NULL;
	}
	for (u_int u = 0; u < ASURA_PC_JS_NUM_OBJECTS; ++u)
	{
		m_axObjectNames[u].MakeNull();
	}
	m_xProductName.MakeNull();
}

// once per frame, update the joystick
// this updates its internal axis and button state variables
void Asura_PC_Input_Joystick::Update()
{
	if (!m_pxDevice)
	{
		m_fHAxis[0] = m_fHAxis[1] = m_fVAxis[0] = m_fVAxis[1] = m_fZAxis = m_fXRot = m_fYRot = m_fZRot = 0.0f;
		m_uiPOV = fASURA_PC_JOYSTICK_POV_UNSET;
		for (u_int u = 0; u < 16; ++u)
		{
			m_aucButtonPressed[u] = 0;
		}

		return;
	}

	HRESULT hr = m_pxDevice->Poll();

	if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
	{
		m_pxDevice->Acquire();

		m_fHAxis[0] = m_fHAxis[1] = m_fVAxis[0] = m_fVAxis[1] = m_fZAxis = m_fXRot = m_fYRot = m_fZRot = 0.0f;
		m_uiPOV = fASURA_PC_JOYSTICK_POV_UNSET;
		for (u_int u = 0; u < 16; ++u)
		{
			m_aucButtonPressed[u] = 0;
		}

		return;
	}

	DIJOYSTATE2 xJoyState;

	hr = m_pxDevice->GetDeviceState(sizeof(DIJOYSTATE2), &xJoyState);

	if (FAILED(hr))
	{
		m_fHAxis[0] = m_fHAxis[1] = m_fVAxis[0] = m_fVAxis[1] = m_fZAxis = m_fXRot = m_fYRot = m_fZRot = 0.0f;
		m_uiPOV = fASURA_PC_JOYSTICK_POV_UNSET;
		for (u_int u = 0; u < 16; ++u)
		{
			m_aucButtonPressed[u] = 0;
		}

		return;
	}

	// in the device control enumeration function we've chosen to use [-ASURA_PC_JOYSTICK_RANGE, ASURA_PC_JOYSTICK_RANGE] as the (integral) range for axes
	// Asura uses floats over [-1, 1]

	m_fHAxis[0] = ConvertJoystickAxisToAsuraAxis(xJoyState.lX, ASURA_PC_JS_OBJECT_X_AXIS);
	m_fVAxis[0] = ConvertJoystickAxisToAsuraAxis(xJoyState.lY, ASURA_PC_JS_OBJECT_Y_AXIS);

	// Secondary sticks should map through to slider 0/1 (known as U, V axes in previous DirectInput versions)

	m_fHAxis[1] = ConvertJoystickAxisToAsuraAxis(xJoyState.rglSlider[0], ASURA_PC_JS_OBJECT_S1_AXIS);
	m_fVAxis[1] = ConvertJoystickAxisToAsuraAxis(xJoyState.rglSlider[1], ASURA_PC_JS_OBJECT_S2_AXIS);

	m_fZAxis = ConvertJoystickAxisToAsuraAxis(xJoyState.lZ, ASURA_PC_JS_OBJECT_Z_AXIS);

	m_fXRot = ConvertJoystickAxisToAsuraAxis(xJoyState.lRx, ASURA_PC_JS_OBJECT_XROT_AXIS);
	m_fYRot = ConvertJoystickAxisToAsuraAxis(xJoyState.lRy, ASURA_PC_JS_OBJECT_YROT_AXIS);
	m_fZRot = ConvertJoystickAxisToAsuraAxis(xJoyState.lRz, ASURA_PC_JS_OBJECT_ZROT_AXIS);

	for (int i = 0; i < 16; ++i)
	{
		m_aucButtonPressed[i] = (xJoyState.rgbButtons[i] & 0x80)?255:0;
	}

	m_uiPOV = xJoyState.rgdwPOV[0];
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PC_Input_Joystick::GetObjectName().
// PARAMETERS   iObject - enum of object
// RETURNS      friendly name if found, 0 if not found
// DESCRIPTION  returns friendly name of an object on this joystick (axes, buttons)
// AUTHOR       Neil Young - 16.08.2005.
//
//////////////////////////////////////////////////////////////////////////////
const char* Asura_PC_Input_Joystick::GetObjectName(int iObject) const
{
	if (iObject >= 0 && iObject < ASURA_PC_JS_NUM_OBJECTS)
	{
		return m_axObjectNames[iObject].GetString();
	}
	else if (iObject >= ASURA_PC_JS_OBJECT_POV_HAT_UP && iObject <= ASURA_PC_JS_OBJECT_POV_HAT_RIGHT)
	{
		return m_axObjectNames[ASURA_PC_JS_OBJECT_POV_HAT].GetString();
	}
	else
	{
		Asura_Assert(0,"object enum out of range");
		return 0;
	}
}
// End function Asura_PC_Input_Joystick::GetObjectName().

/******************************************************************************/
/* Static Member Functions                                                    */
/******************************************************************************/

// NY - 16 August 2005 - this should ideally take into account the axis passed in,
// as not all axes will need a deadzone (makes no sense for throttle, for eg)
float Asura_PC_Input_Joystick::ConvertJoystickAxisToAsuraAxis(LONG lAxisValue, ASURA_PC_JS_OBJECT eAxis)
{
	// Take into acount deadzones etc.
	float fA = static_cast<float>(lAxisValue);

	float fDeadZone = fASURA_PC_JOYSTICK_DEADZONE;

	if(fA > fDeadZone)
		fA = (fA - fDeadZone) * (fASURA_PC_JOYSTICK_RANGE / (fASURA_PC_JOYSTICK_RANGE - fDeadZone));
	else if(fA < - fDeadZone)
		fA = (fA + fDeadZone) * (fASURA_PC_JOYSTICK_RANGE / (fASURA_PC_JOYSTICK_RANGE - fDeadZone));
	else
		return 0.f;

	// Normalise
	fA /= fASURA_PC_JOYSTICK_RANGE;
	return fA;		// now in range -1.f to 1.f
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PC_Input_Joystick::ConvertJoystickToAsuraType
// Author       : Andy Tate
// Date         : 6-Jun-2008
////////////////////////////////////////////////////////////////////////////////

Asura_PC_Input_Joystick::ASURA_PC_JS_TYPE Asura_PC_Input_Joystick::ConvertJoystickToAsuraType(const JoystickDefinition& xDefinition)
{
	const Asura_String&	xJoystickName	= xDefinition.m_xJoystickName;

	if( (xJoystickName == "Dual PSX-USB Adaptor ") ||
		(xJoystickName == "PS Vibration Feedback Converter ") ||
		(xJoystickName == "USB Vibration Gamepad") ||
		(xJoystickName == "USB GAME PAD") ||
		(xJoystickName == "USB Joystick     ") ||
		(xJoystickName == "PSJOY (USB) Joystick") ||
		(xJoystickName == "SmartJoy PLUS Adapter") ||
		(xJoystickName == "4 axis 16 button joystick") )
	{
		return(ASURA_PC_JS_TYPE_PS2_GAMEPAD);
	}
	else if(xJoystickName == "PLAYSTATION(R)3 Controller")
	{
		return(ASURA_PC_JS_TYPE_PS3_GAMEPAD);
	}
	else if( (xJoystickName == "GIC USB Joystick") ||
			 (xJoystickName == "GIC Psx Vibration feedback (USB) Joystick") )
	{
		return(ASURA_PC_JS_TYPE_PS2_JOYSTICK);
	}
	else if( (xJoystickName == "XBOX 360 For Windows (Controller)") ||
			 (xJoystickName == "Controller (XBOX 360 For Windows)") ||
			 (xJoystickName == "Controller (Xbox 360 Wireless Receiver for Windows)") ||
			 (xJoystickName == "XBCD XBox Gamepad") ||
			 (xJoystickName == "XBCD XBox 360 Gamepad") )
	{
		return(ASURA_PC_JS_TYPE_XBOX360_GAMEPAD);
	}
	#ifdef ASURA_PC_SUPPORT_XINPUT
	else if (xJoystickName == "XInput XBOX 360 Pad")
	{
		return(ASURA_PC_JS_TYPE_XBOX360_XINPUT_GAMEPAD);
	}
	#endif
	else
	{
		return(ASURA_PC_JS_TYPE_UNRECOGNISED);
	}
}

/// End of function 'Asura_PC_Input_Joystick::ConvertJoystickToAsuraType'. ///



////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  InitJoystickList()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  builds list of attached devices
////////////////////////////////////////////////////////////////////////////////
void Asura_PC_Input_Joystick::InitJoystickList()
{
	if (!s_pxJoystickList)
	{
		s_iNumJoysticks = 0;
		s_pxJoystickList = new Asura_List<JoystickDefinition>;
		if (s_pxJoystickList)
		{
			Asura_DirectInput::DIObject->EnumDevices(DI8DEVCLASS_GAMECTRL, EnumDevicesCallback, 0, DIEDFL_ATTACHEDONLY);
		}
		else
		{
			Asura_Assert(0, "failed to create list, so no point trying to fill it!");
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  ShutdownJoystickList()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  frees list of attached devices
////////////////////////////////////////////////////////////////////////////////
void Asura_PC_Input_Joystick::ShutdownJoystickList()
{
	if (s_pxJoystickList)
	{
		JoystickDefinition* pxDefinition;
		ShutdownTempDevices();
		for (int i = 0; i < s_iNumJoysticks; ++i)
		{
			pxDefinition = GetJoystickFromIndex(static_cast<u_int>(i));
			delete pxDefinition;
		}
		s_pxJoystickList->ClearList();
		delete s_pxJoystickList;
		s_pxJoystickList = 0;
		s_iNumJoysticks = 0;
	}
}

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  InitTempDevices()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  inits a directInput device for each entry of the list
//				of attached devices
////////////////////////////////////////////////////////////////////////////////
void Asura_PC_Input_Joystick::InitTempDevices()
{
	if (s_pxJoystickList)
	{
		Asura_Assert(!s_pbTempDevicesStatus, "device status array already allocated");
		delete [] s_pbTempDevicesStatus;
		s_pbTempDevicesStatus = new bool[s_iNumJoysticks];
		JoystickDefinition* pxDefinition;
		for (int i = 0; i < s_iNumJoysticks; ++i)
		{
			s_pbTempDevicesStatus[i] = false;
			pxDefinition = GetJoystickFromIndex(static_cast<u_int>(i));
			if (pxDefinition->m_pxJoystickTempDevice)
			{
				Asura_Assert(0,"device already initialised");
			}
			else
			{
				HRESULT	hr;

				hr = Asura_DirectInput::DIObject->CreateDevice(pxDefinition->m_xJoystickGUID, &pxDefinition->m_pxJoystickTempDevice, NULL);
				if (FAILED(hr))
				{
					if (pxDefinition->m_pxJoystickTempDevice)
					{
						pxDefinition->m_pxJoystickTempDevice->Release();
						pxDefinition->m_pxJoystickTempDevice = 0;
					}
					continue;
				}

				hr = pxDefinition->m_pxJoystickTempDevice->SetDataFormat(&c_dfDIJoystick2);
				if (FAILED(hr))
				{
					if (pxDefinition->m_pxJoystickTempDevice)
					{
						pxDefinition->m_pxJoystickTempDevice->Release();
						pxDefinition->m_pxJoystickTempDevice = 0;
					}
					continue;
				}

				hr = pxDefinition->m_pxJoystickTempDevice->SetCooperativeLevel(Asura_Windows::GetWindowHandle(), DISCL_EXCLUSIVE | DISCL_FOREGROUND);
				if (FAILED(hr))
				{
					if (pxDefinition->m_pxJoystickTempDevice)
					{
						pxDefinition->m_pxJoystickTempDevice->Release();
						pxDefinition->m_pxJoystickTempDevice = 0;
					}
					continue;
				}

				pxDefinition->m_pxJoystickTempDevice->Acquire();
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  UpdateTempDevices()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  polls each temporary device, and stores whether it has any buttons pressed
////////////////////////////////////////////////////////////////////////////////
void Asura_PC_Input_Joystick::UpdateTempDevices()
{
	if (s_pxJoystickList)
	{
		JoystickDefinition* pxDefinition;
		for (int i = 0; i < s_iNumJoysticks; ++i)
		{
			s_pbTempDevicesStatus[i] = false;
			pxDefinition = GetJoystickFromIndex(static_cast<u_int>(i));
			if (pxDefinition->m_pxJoystickTempDevice)
			{
				HRESULT hr = pxDefinition->m_pxJoystickTempDevice->Poll();

				if (hr == DIERR_INPUTLOST || hr == DIERR_NOTACQUIRED)
				{
					pxDefinition->m_pxJoystickTempDevice->Acquire();
				}
				DIJOYSTATE2 xJoyState;

				hr = pxDefinition->m_pxJoystickTempDevice->GetDeviceState(sizeof(DIJOYSTATE2), &xJoyState);

				if (!FAILED(hr))
				{
					for (int u = 0; u < 16; ++u)
					{
						if (xJoyState.rgbButtons[u] & 0x80)
						{
							s_pbTempDevicesStatus[i] = true;
						}
					}
					if (xJoyState.rgdwPOV[0] != 1 && xJoyState.rgdwPOV[0] <= 36000)
					{
						s_pbTempDevicesStatus[i] = true;
					}
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  ShutdownTempDevices()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  frees temporary devices
////////////////////////////////////////////////////////////////////////////////
void Asura_PC_Input_Joystick::ShutdownTempDevices()
{
	if (s_pxJoystickList)
	{
		JoystickDefinition* pxDefinition;
		for (int i = 0; i < s_iNumJoysticks; ++i)
		{
			pxDefinition = GetJoystickFromIndex(static_cast<u_int>(i));
			if (pxDefinition->m_pxJoystickTempDevice)
			{
				pxDefinition->m_pxJoystickTempDevice->Unacquire();
				pxDefinition->m_pxJoystickTempDevice->Release();
				pxDefinition->m_pxJoystickTempDevice = 0;
			}
		}
		delete [] s_pbTempDevicesStatus;
		s_pbTempDevicesStatus = 0;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_PC_Input_Joystick::GetTempDevicesStatus().
// PARAMETERS   None.
// RETURNS      pointer to an array of Asura_PC_Input_Joystick::GetNumJoysticks() bools,
//				or zero if temp devices not created
// DESCRIPTION  fetches the array of which temporary devices have a button pressed
// AUTHOR       Neil Young - 16.08.2005.
//
//////////////////////////////////////////////////////////////////////////////
const bool* Asura_PC_Input_Joystick::GetTempDevicesStatus()
{
	return s_pbTempDevicesStatus;
}
// End function Asura_PC_Input_Joystick::GetTempDevicesStatus().

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  GetJoystickName()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  fetches the friendly name of the numbered device in the list
////////////////////////////////////////////////////////////////////////////////
char* Asura_PC_Input_Joystick::GetJoystickName(u_int uIndex)
{
	JoystickDefinition* pxDefinition = GetJoystickFromIndex(uIndex);
	if (pxDefinition)
	{
		return pxDefinition->m_xJoystickName.GetString();
	}
	return "";
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PC_Input_Joystick::GetJoystickType
// Author       : Andy Tate
// Date         : 6-Jun-2008
////////////////////////////////////////////////////////////////////////////////

Asura_PC_Input_Joystick::ASURA_PC_JS_TYPE Asura_PC_Input_Joystick::GetJoystickType(u_int uIndex)
{
	const JoystickDefinition*	pxDefinition	= GetJoystickFromIndex(uIndex);

	if(pxDefinition == NULL)
	{
		return(ASURA_PC_JS_TYPE_UNRECOGNISED);
	}
	else
	{
		return(pxDefinition->m_eJoystickType);
	}
}

/// End of function 'Asura_PC_Input_Joystick::GetJoystickType'. ///


////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  GetJoystickGUID()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  fetches the guid of the numbered device in the list
//				returns true if succesful and sets xGUID, returns false and leaves it intact if fails
////////////////////////////////////////////////////////////////////////////////
bool Asura_PC_Input_Joystick::GetJoystickGUID(u_int uIndex, GUID& xGUID)
{
	JoystickDefinition* pxDefinition = GetJoystickFromIndex(uIndex);
	if (pxDefinition)
	{
		xGUID = pxDefinition->m_xJoystickGUID;
		return true;
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  GetJoystickGUIDFromName()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  fetches the guid of the named device in the list
//				returns true if succesful and sets xGUID, returns false and leaves it intact if fails
////////////////////////////////////////////////////////////////////////////////
bool Asura_PC_Input_Joystick::GetJoystickGUIDFromName(char* szName, GUID& xGUID)
{
	if (s_pxJoystickList && s_iNumJoysticks > 0)
	{
		for (Asura_List_It<JoystickDefinition> xIterator(s_pxJoystickList); !xIterator.Done(); xIterator.Next())
		{
			JoystickDefinition* pxDefinition = xIterator.GetCurrent();
			if (pxDefinition->m_xJoystickName == szName)
			{
				xGUID = pxDefinition->m_xJoystickGUID;
				return true;
			}
		}
	}
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//  NAME     :  JoystickDefinition* Asura_PC_Input_Joystick::GetJoystickFromIndex()
//  CREATED  :  NY (16.8.2005)
//  DESC     :  internal helper function to fetch the data for the numbered device in the list
////////////////////////////////////////////////////////////////////////////////
Asura_PC_Input_Joystick::JoystickDefinition* Asura_PC_Input_Joystick::GetJoystickFromIndex(u_int uIndex)
{
	if (s_pxJoystickList && (int)uIndex < s_iNumJoysticks)
	{
		u_int uCount = 0;
		for (Asura_List_It<JoystickDefinition> xIterator(s_pxJoystickList); !xIterator.Done(); xIterator.Next() , ++uCount)
		{
			JoystickDefinition* pxDefinition = xIterator.GetCurrent();
			if (uIndex == uCount)
			{
				return pxDefinition;
			}
		}
	}
	Asura_Assert(0, "joystick index out of range");
	return 0;
}

////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  Asura_PC_Input_Joystick
//  NAME     :  InitialiseJoystickInstanceFromGUID()
//	
//	Initialise device through the correct interface
////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_PC_SUPPORT_XINPUT
bool Asura_PC_Input_Joystick::InitialiseJoystickInstanceFromGUID( Asura_PC_Input_XInputPad* const pxJoystick, const GUID& xGUID )
{
	if (s_pxJoystickList && s_iNumJoysticks > 0)
	{
		for (Asura_List_It<JoystickDefinition> xIterator(s_pxJoystickList); !xIterator.Done(); xIterator.Next())
		{
			JoystickDefinition* pxDefinition = xIterator.GetCurrent();
			if ( xGUID == pxDefinition->m_xJoystickGUID )
			{
				if ( pxDefinition->m_bXInputDevice )
				{
					pxJoystick->InitialiseXInput( pxDefinition->m_iXInputID );
					return true;
				}
				else
				{
					return pxJoystick->Initialise( xGUID );
				}
			}
		}
	}
	return false;
}
#endif

// callback used to enumerate devices by IDirectInput8::EnumDevices()
BOOL CALLBACK Asura_PC_Input_Joystick::EnumDevicesCallback(const DIDEVICEINSTANCE* pxDeviceInstance, void* pData)
{
	++s_iNumJoysticks;
	JoystickDefinition* pxDefinition = new JoystickDefinition;

	pxDefinition->m_xJoystickGUID = pxDeviceInstance->guidInstance;

	#ifdef ASURA_PC_SUPPORT_XINPUT
	// Modify the joystick definition for XInput devices
	if ( Asura_PC_Input_XInputPad::IsXInputDevice( &pxDeviceInstance->guidProduct ) )
	{
		pxDefinition->m_xJoystickName = *Asura_PC_Input_XInputPad::GetXInputProductName();
		pxDefinition->m_bXInputDevice = true;
		pxDefinition->m_iXInputID = s_iNumXInputDevices;
		s_iNumXInputDevices++;
	}
	else
	{
		pxDefinition->m_xJoystickName = pxDeviceInstance->tszProductName;
		pxDefinition->m_bXInputDevice = false;
	}

	#else
	pxDefinition->m_xJoystickName = pxDeviceInstance->tszProductName;
	#endif	// ASURA_PC_SUPPORT_XINPUT

	pxDefinition->m_eJoystickType = ConvertJoystickToAsuraType(*pxDefinition);

	//leave initialising the temp device for a separate call, we only need it for button polling
	pxDefinition->m_pxJoystickTempDevice = 0;

	if (s_pxJoystickList)
	{
		s_pxJoystickList->AddToList(pxDefinition);
	}
	else
	{
		Asura_Assert(0, "where the **** has the list gone?");
		return DIENUM_STOP; //no point asking for more
	}

	return DIENUM_CONTINUE;
}

// callback used to enumerate device controls by IDirectInputDevice8::EnumObjects()
// for the data parameter, pass an IDirectInputDevice8 pointer to the device
BOOL CALLBACK Asura_PC_Input_Joystick::EnumObjectsCallback(const DIDEVICEOBJECTINSTANCE* pxObject, void* pData)
{
	Asura_PC_Input_Joystick* pxThis = reinterpret_cast <Asura_PC_Input_Joystick*> (pData);
	if (pxObject->dwType & DIDFT_AXIS)
	{
		DIPROPRANGE xPropRange;
		xPropRange.diph.dwSize = sizeof(DIPROPRANGE);
		xPropRange.diph.dwHeaderSize = sizeof(DIPROPHEADER);
		xPropRange.diph.dwHow = DIPH_BYID;
		xPropRange.diph.dwObj = pxObject->dwType;
		xPropRange.lMin = -iASURA_PC_JOYSTICK_RANGE;
		xPropRange.lMax = +iASURA_PC_JOYSTICK_RANGE;

		Asura_Assert(pData, "Input device pointer was not passed to axis enumeration function.");

		pxThis->m_pxDevice->SetProperty(DIPROP_RANGE, &xPropRange.diph);
	}

	int iObject = -1;

	// NY - 04 August 2005 - can't do a switch on GUIDs...  hrmph....

	if (pxObject->guidType == GUID_XAxis)
	{
		iObject = ASURA_PC_JS_OBJECT_X_AXIS;
	}
	else if (pxObject->guidType == GUID_YAxis)
	{
		iObject = ASURA_PC_JS_OBJECT_Y_AXIS;
	}
	else if (pxObject->guidType == GUID_ZAxis)
	{
		iObject = ASURA_PC_JS_OBJECT_Z_AXIS;
	}
	else if (pxObject->guidType == GUID_RxAxis)
	{
		iObject = ASURA_PC_JS_OBJECT_XROT_AXIS;
	}
	else if (pxObject->guidType == GUID_RyAxis)
	{
		iObject = ASURA_PC_JS_OBJECT_YROT_AXIS;
	}
	else if (pxObject->guidType == GUID_RzAxis)
	{
		iObject = ASURA_PC_JS_OBJECT_ZROT_AXIS;
	}
	else if (pxObject->guidType == GUID_Slider)
	{
		const int	iIndex	= ASURA_PC_JS_OBJECT_FIRST_SLIDER_AXIS + DIDFT_GETINSTANCE(pxObject->dwType);

		// We only support a limited number of slider-axes...
		if((iIndex >= ASURA_PC_JS_OBJECT_FIRST_SLIDER_AXIS) && (iIndex <= ASURA_PC_JS_OBJECT_LAST_SLIDER_AXIS))
		{
			iObject = iIndex;
		}
	}
	else if (pxObject->guidType == GUID_Button)
	{
		const int	iIndex	= ASURA_PC_JS_OBJECT_FIRST_BUTTON + DIDFT_GETINSTANCE(pxObject->dwType);

		// We only support a limited number of buttons...
		if((iIndex >= ASURA_PC_JS_OBJECT_FIRST_BUTTON) && (iIndex <= ASURA_PC_JS_OBJECT_LAST_BUTTON))
		{
			iObject = iIndex;
		}
	}
	else if (pxObject->guidType == GUID_POV && DIDFT_GETINSTANCE(pxObject->dwType) == 0)//only support 1st hat for now
	{
		iObject = ASURA_PC_JS_OBJECT_POV_HAT;
	}

	// Update this particular object...
	if(iObject >= ASURA_PC_JS_NUM_OBJECTS)
	{
		Asura_Assert(false, "Object enum callback called with an object which is out of our stored range.");
	}
	else if (iObject >= 0)
	{
		const Asura_String&	xOldName(pxThis->m_axObjectNames[iObject]);
		const Asura_String	xNewName(pxObject->tszName);

		if(xOldName.IsBlank())
		{
			pxThis->m_axObjectNames[iObject] = xNewName.GetString();
		}
		else if(xOldName == xNewName)
		{
			// Same device-name a second time - don't need to do anything...
		}
		else
		{
			Asura_Assertf(false, ("object enum callback called twice for same object (%i: \"%s\" -> \"%s\")", iObject, xOldName.GetString(), xNewName.GetString()));
		}
	}

	return DIENUM_CONTINUE;
}
