
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PC_Input_XInput.cpp
// Author : Steve Barnett
// Date   : 19-Mar-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#ifdef ASURA_PC_SUPPORT_XINPUT

#include "Input/Asura_PC_Input_XInput.h"

// System headers used to determine if XInput supports a DirectInput device
#include <wbemidl.h>
#include <oleauto.h>

// Asura platform-independent
#include "Core/Asura_FixedArray.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Deadzone for the gamepad inputs, taken from MS sample code
const short sINPUT_DEADZONE = (short)( 0.24f * float(0x7FFF) );

// POV hat unset value
const u_int fASURA_PC_JOYSTICK_POV_UNSET = 0xffffffff;

const Asura_String Asura_PC_Input_XInputPad::s_xXInputProductName ("XInput XBOX 360 Pad");

// ancillary converter function taken from MS sample code
static inline float ConvertThumbstickValue( short sThumbstickValue )
{
    if( sThumbstickValue > +sINPUT_DEADZONE )
    {
        return (sThumbstickValue-sINPUT_DEADZONE) / (32767.0f-sINPUT_DEADZONE);
    }
    if( sThumbstickValue < -sINPUT_DEADZONE )
    {
        return (sThumbstickValue+sINPUT_DEADZONE+1.0f) / (32767.0f-sINPUT_DEADZONE);
    }
    return 0.0f;
}

Asura_PC_Input_XInputPad::Asura_PC_Input_XInputPad( void )
:	Asura_PC_Input_Joystick()
{
	// Assume it's not an XInput device for now
	ResetXInputData();
}

Asura_PC_Input_XInputPad::~Asura_PC_Input_XInputPad( void )
{
}

void Asura_PC_Input_XInputPad::InitialiseXInput( const u_int uID )
{
	m_bIsXInputDevice = true;
	m_usUserIndex = uID;

	Asura_PC_Input_Joystick::Shutdown();

	// Set product name to "XInput XBOX 360 Pad" to seperate it from other 360 pad drivers but keep it identifiable
	m_xProductName = s_xXInputProductName;
}

void Asura_PC_Input_XInputPad::Shutdown()
{
	ResetXInputData();
	Asura_PC_Input_Joystick::Shutdown();
}

void Asura_PC_Input_XInputPad::Update()
{
	if ( m_bIsXInputDevice )
	{
		UpdateXInputData();
	}
	else
	{
		Asura_PC_Input_Joystick::Update();
	}
}

void Asura_PC_Input_XInputPad::ResetXInputData( void )
{
    ZeroMemory( &m_xCaps, sizeof( XINPUT_CAPABILITIES ) );
    m_fLeftX = 0.0f;
    m_fLeftY = 0.0f;
    m_fRightX = 0.0f;
    m_fRightY = 0.0f;
    m_usPressedButtons = 0x0000;
    m_usUserIndex = 0;
    m_ucLeftTrigger = 0;
    m_ucRightTrigger = 0;
    m_bConnected = false;
    m_bInserted = false;
    m_bRemoved = false;
}

void Asura_PC_Input_XInputPad::UpdateXInputData( void )
{
	// Note if the controller was connected last frame, and try to get the input state (if we can't, it's not connected)
	bool bWasConnected = this->m_bConnected;

	XINPUT_STATE xInputState;
	this->m_bConnected = ( XInputGetState( m_usUserIndex, &xInputState ) == ERROR_SUCCESS );

	// Track insertion and removals
	this->m_bRemoved  = (  bWasConnected && !this->m_bConnected );
	this->m_bInserted = ( !bWasConnected &&  this->m_bConnected );

	// if the controller isn't connected, just clear the data
	if( this->m_bConnected )
	{
		// handle a newly inserted controller
		if( this->m_bInserted )
		{
			// probably can't trust what's in memory so we'll wipe it
			// (previously pressed buttons, user may have swapped controllers etc.)
			this->ResetXInputData();
	
			this->m_bConnected = true;
			this->m_bInserted  = true;
	
			// fill in the capabilities, currently (2005/10/10) not doing much with this,
			// since we only have one type of controller to play with.
			XInputGetCapabilities( m_usUserIndex, XINPUT_FLAG_GAMEPAD, &this->m_xCaps );
		}
	
		// Let's look at our input data then. I guess we should be looking at the capabilities
		// here really to see what kind of controller it is?
	
		// Let's get those thumbsticks into -1 to +1 floats, including the deadzone.
		this->m_fLeftX = ConvertThumbstickValue( xInputState.Gamepad.sThumbLX );
		this->m_fLeftY = ConvertThumbstickValue( xInputState.Gamepad.sThumbLY );
		this->m_fRightX = ConvertThumbstickValue( xInputState.Gamepad.sThumbRX );
		this->m_fRightY = ConvertThumbstickValue( xInputState.Gamepad.sThumbRY );
	
		// Which buttons have been pressed?
		this->m_usPressedButtons = xInputState.Gamepad.wButtons;
	
		// Triggers are analogue (0..255)...
		this->m_ucLeftTrigger = xInputState.Gamepad.bLeftTrigger;
		this->m_ucRightTrigger = xInputState.Gamepad.bRightTrigger;
	}
	else
	{
		ResetXInputData();
	}
	
	// Now update the base class data from the XInput data

	// Axes
	// Vertical axes are flipped
	m_fHAxis[0] = m_fLeftX;
	m_fVAxis[0] = -m_fLeftY;
	m_fHAxis[1] = m_fRightX;
	m_fVAxis[1] = -m_fRightY;
	m_fXRot = m_fRightX;
	m_fYRot = -m_fRightY;

	// Face buttons
	m_aucButtonPressed[0] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_A ) == 0 ) ? 0 : 255;
	m_aucButtonPressed[1] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_B ) == 0 ) ? 0 : 255;
	m_aucButtonPressed[2] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_X ) == 0 ) ? 0 : 255;
	m_aucButtonPressed[3] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_Y ) == 0 ) ? 0 : 255;

	// Bumpers
	m_aucButtonPressed[4] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_LEFT_SHOULDER ) == 0 ) ? 0 : 255;
	m_aucButtonPressed[5] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_RIGHT_SHOULDER ) == 0 ) ? 0 : 255;

	// Start and back
	m_aucButtonPressed[6] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_START ) == 0 ) ? 0 : 255;
	m_aucButtonPressed[7] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_BACK ) == 0 ) ? 0 : 255;

	// Left and right thumbstick buttons
	m_aucButtonPressed[8] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_LEFT_THUMB ) == 0 ) ? 0 : 255;
	m_aucButtonPressed[9] = ( ( m_usPressedButtons & XINPUT_GAMEPAD_RIGHT_THUMB ) == 0 ) ? 0 : 255;

	// Triggers
	m_aucButtonPressed[10] = m_ucLeftTrigger;
	m_aucButtonPressed[11] = m_ucRightTrigger;

	// DPAD - need to generate an angle in hundredths of degrees
	const bool bHatUp = ( ( m_usPressedButtons & XINPUT_GAMEPAD_DPAD_UP ) != 0 );
	const bool bHatDown = ( ( m_usPressedButtons & XINPUT_GAMEPAD_DPAD_DOWN ) != 0 );
	const bool bHatLeft = ( ( m_usPressedButtons & XINPUT_GAMEPAD_DPAD_LEFT ) != 0 );
	const bool bHatRight = ( ( m_usPressedButtons & XINPUT_GAMEPAD_DPAD_RIGHT ) != 0 );
	if ( bHatUp && bHatRight )
	{
		m_uiPOV = 4500;
	}
	else if ( bHatDown && bHatRight )
	{
		m_uiPOV = 13500;
	}
	else if ( bHatDown && bHatLeft )
	{
		m_uiPOV = 22500;
	}
	else if ( bHatUp && bHatLeft )
	{
		m_uiPOV = 31500;
	}
	else if ( bHatUp )
	{
		m_uiPOV = 0;
	}
	else if ( bHatRight )
	{
		m_uiPOV = 9000;
	}
	else if ( bHatDown )
	{
		m_uiPOV = 18000;
	}
	else if ( bHatLeft )
	{
		m_uiPOV = 27000;
	}
	else
	{
		m_uiPOV = fASURA_PC_JOYSTICK_POV_UNSET;
	}

}

// This function is taken from Microsoft's XInput sample code but restructured
// to remove the goto statements that the original was plagued with D:

//-----------------------------------------------------------------------------
// Enum each PNP device using WMI and check each device ID to see if it contains 
// "IG_" (ex. "VID_045E&PID_028E&IG_00").  If it does, then it's an XInput device
// Unfortunately this information can not be found by just using DirectInput 
//-----------------------------------------------------------------------------
BOOL Asura_PC_Input_XInputPad::IsXInputDevice( const GUID* pGuidProductFromDirectInput )
{
    IWbemLocator*           pIWbemLocator  = NULL;
    IEnumWbemClassObject*   pEnumDevices   = NULL;
    IWbemClassObject*       pDevices[20]   = {0};
    IWbemServices*          pIWbemServices = NULL;
    BSTR                    bstrNamespace  = NULL;
    BSTR                    bstrDeviceID   = NULL;
    BSTR                    bstrClassName  = NULL;
    DWORD                   uReturned      = 0;
    bool                    bIsXinputDevice= false;
    UINT                    iDevice        = 0;
    VARIANT                 var;
    HRESULT                 hr;

	// Bool to record if we failed and clean up
	bool bFailed = false;

    // CoInit if needed
    hr = CoInitialize(NULL);
    bool bCleanupCOM = SUCCEEDED(hr);

    // Create WMI
    hr = CoCreateInstance( __uuidof(WbemLocator),
                           NULL,
                           CLSCTX_INPROC_SERVER,
                           __uuidof(IWbemLocator),
                           (LPVOID*) &pIWbemLocator);
    if( FAILED(hr) || pIWbemLocator == NULL )
	{
		bFailed = true;
	}

	if ( !bFailed )
	{
	    bstrNamespace = SysAllocString( L"\\\\.\\root\\cimv2" );
		if( bstrNamespace == NULL )
		{
			bFailed = true;
		}
		else
		{
			bstrClassName = SysAllocString( L"Win32_PNPEntity" );
			if( bstrClassName == NULL )
			{
				bFailed = true;
			}
			else
			{
				bstrDeviceID  = SysAllocString( L"DeviceID" );
				if( bstrDeviceID == NULL )
				{
					bFailed = true;
				}
			}
		}
	}

	if ( !bFailed )
	{
	    // Connect to WMI 
	    hr = pIWbemLocator->ConnectServer( bstrNamespace, NULL, NULL, 0L, 
	                                       0L, NULL, NULL, &pIWbemServices );
	    if( FAILED(hr) || pIWbemServices == NULL )
		{
			bFailed = true;
		}
	
		if ( !bFailed )
		{
		    // Switch security level to IMPERSONATE. 
		    CoSetProxyBlanket( pIWbemServices, RPC_C_AUTHN_WINNT, RPC_C_AUTHZ_NONE, NULL, 
		                       RPC_C_AUTHN_LEVEL_CALL, RPC_C_IMP_LEVEL_IMPERSONATE, NULL, EOAC_NONE );                    
		
		    hr = pIWbemServices->CreateInstanceEnum( bstrClassName, 0, NULL, &pEnumDevices ); 
		    if( FAILED(hr) || pEnumDevices == NULL )
			{
				bFailed = true;
			}
		
		    // Loop over all devices
		    for( ; !bFailed ; )
		    {
		        // Get 20 at a time
		        hr = pEnumDevices->Next( 10000, 20, pDevices, &uReturned );
		        if( FAILED(hr) )
				{
					break;
				}
		        if( uReturned == 0 )
				{
					// No more devices
		            break;
				}
		
		        for( iDevice=0; iDevice<uReturned; iDevice++ )
		        {
		            // For each device, get its device ID
		            hr = pDevices[iDevice]->Get( bstrDeviceID, 0L, &var, NULL, NULL );
		            if( SUCCEEDED( hr ) && var.vt == VT_BSTR && var.bstrVal != NULL )
		            {
		                // Check if the device ID contains "IG_".  If it does, then it's an XInput device
						    // This information can not be found from DirectInput 
		                if( wcsstr( var.bstrVal, L"IG_" ) )
		                {
		                    // If it does, then get the VID/PID from var.bstrVal
		                    DWORD dwPid = 0, dwVid = 0;
		                    WCHAR* strVid = wcsstr( var.bstrVal, L"VID_" );
		                    if( strVid && swscanf( strVid, L"VID_%4X", &dwVid ) != 1 )
							{
		                        dwVid = 0;
							}
		                    WCHAR* strPid = wcsstr( var.bstrVal, L"PID_" );
		                    if( strPid && swscanf( strPid, L"PID_%4X", &dwPid ) != 1 )
							{
		                        dwPid = 0;
							}
		
		                    // Compare the VID/PID to the DInput device
		                    DWORD dwVidPid = MAKELONG( dwVid, dwPid );
		                    if( dwVidPid == pGuidProductFromDirectInput->Data1 )
		                    {
		                        bIsXinputDevice = true;
								break;
		                    }
		                }
		            }
					if ( pDevices[iDevice] != NULL )
					{
						pDevices[iDevice]->Release();
						pDevices[iDevice] = NULL;
					}
		        }
		    }
		}
	}

    if( bstrNamespace != NULL )
	{
        SysFreeString(bstrNamespace);
	}
    if( bstrDeviceID != NULL )
	{
        SysFreeString(bstrDeviceID);
	}
    if( bstrClassName != NULL )
	{
        SysFreeString(bstrClassName);
	}
    for( iDevice=0; iDevice<20; iDevice++ )
	{
		if ( pDevices[iDevice] != NULL )
		{
			pDevices[iDevice]->Release();
			pDevices[iDevice] = NULL;
		}
	}
	if ( pEnumDevices != NULL )
	{
		pEnumDevices->Release();
		pEnumDevices = NULL;
	}
	if ( pIWbemLocator != NULL )
	{
		pIWbemLocator->Release();
		pIWbemLocator = NULL;
	}
	if ( pIWbemServices != NULL )
	{
		pIWbemServices->Release();
		pIWbemServices = NULL;
	}

    if( bCleanupCOM )
	{
        CoUninitialize();
	}

    return bIsXinputDevice;
}

const Asura_String* Asura_PC_Input_XInputPad::GetXInputProductName( void )
{
	return &s_xXInputProductName;
}

#endif	// ASURA_PC_SUPPORT_XINPUT
