//////////////////////////////////////////////////////////////////////////////
//
// Asura_PC_ConsoleWindowIO.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     See header.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura.h"

// This file's header.
#include "Interface/Asura_PC_ConsoleWindowIO.h"

// Stdlib platform-specific.
#include <windows.h>

// Engine independent.
#include "Core/Asura_String.h"
#include "Core/Asura_Time_System.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Version.h"
#include "Debugging/Asura_Logfile.h"

bool Asura_PC_ConsoleWindowIO::s_bQuitRequested = false;

BOOL WINAPI ConsoleHandlerRoutine(DWORD dwCtrlType)
{
    switch (dwCtrlType) 
    { 
        // WGP: Return TRUE for Ctrl-C events as these are handled in Asura_PC_ConsoleWindowIO::GetInput
        case CTRL_C_EVENT:
            return TRUE;
		// SJL: Return TRUE and record the close event by setting s_bQuitRequested so the program
		//      can query this.
        case CTRL_BREAK_EVENT:
        case CTRL_CLOSE_EVENT:
		case CTRL_LOGOFF_EVENT:
        case CTRL_SHUTDOWN_EVENT:
			Asura_PC_ConsoleWindowIO::RequestQuit();
			return TRUE;

        default:
            return FALSE;
    } 
}

void ConsolePrintf(int x, int y, u_int uColour, const char* szFormat, ...)
{
	char* str = new char[256];
	va_list xArgs;
	va_start( xArgs, szFormat );
    Asura_PrintfToBfr( str, szFormat, xArgs, 256, true );
	va_end( xArgs );

	DWORD uIgnore;
	COORD coord = { static_cast<SHORT>(x), static_cast<SHORT>(y) };
    SetConsoleCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE), coord);
//	SetConsoleTextAttribute	(GetStdHandle(STD_OUTPUT_HANDLE), uColour);
	WriteConsole			(GetStdHandle(STD_OUTPUT_HANDLE), str, strlen(str + 1), &uIgnore, NULL);
	delete[] str;
}
	

///////////////////////////////////////////////////////////////////////////////
//-----------------------------------------------------------------------------
//
//
// OR - 03/09/2004 17:31:16 - IMPORTANT!!!!!
//
// Don't forget to disable "QuickEdit" in the console window properties. 
// else, left mb will pause the console and the application. Potentially fatal. 
// 
//-----------------------------------------------------------------------------
///////////////////////////////////////////////////////////////////////////////

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_PC_ConsoleWindowIO::Asura_PC_ConsoleWindowIO( const char* szGameName )
{
	// (MP - 26.08.2003) TODO: Can support a -noconsole argument by calling
    // FreeConsole(). See MSDN.
	// (MH - 16.09.2003) need to init Asura_Windows to set threadID
	FreeConsole();
	AllocConsole();
	COORD xCoord = GetLargestConsoleWindowSize(GetStdHandle(STD_OUTPUT_HANDLE));
   	SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE), xCoord);
	SetConsoleCtrlHandler(ConsoleHandlerRoutine, true);
	const SMALL_RECT xConsoleWindowRect = {0, 0, 80, 19};
	SetConsoleWindowInfo(GetStdHandle(STD_OUTPUT_HANDLE), TRUE, &xConsoleWindowRect);

    m_hStdIn  = GetStdHandle( STD_INPUT_HANDLE  );
    m_hStdOut = GetStdHandle( STD_OUTPUT_HANDLE );

    if( INVALID_HANDLE_VALUE == m_hStdIn && INVALID_HANDLE_VALUE == m_hStdOut )
    {
        // Shouldn't be possible for this to fail...
        Asura_Assert( false, "" );
        return; 
    }

	char acTitle[64];
	Asura_PrintfToBfr(acTitle, "%s : %s", 64, true, szGameName, Asura_BuildTimeAndDate);
	SetConsoleTitle( acTitle );

    m_wszCachedInputBfr[0] = 0;
	m_xOutputCursorPos.X = 0;
    m_xOutputCursorPos.Y = 0;
    m_bCtrlC             = false;

    ClearOutput();
}

Asura_PC_ConsoleWindowIO::~Asura_PC_ConsoleWindowIO( void )
{
    if( INVALID_HANDLE_VALUE != m_hStdIn  ) { CloseHandle( m_hStdIn  ); }
    if( INVALID_HANDLE_VALUE != m_hStdOut ) { CloseHandle( m_hStdOut ); }
	FreeConsole();
}


const char* DEDICATED_LOGFILE = "logfiles\\dedicated_console.txt";

void Asura_PC_ConsoleWindowIO::AddOutput( const Asura_UnicodeChar* wszData )
{
	static bool s_bLogFileInitialised = false;

	if (!s_bLogFileInitialised)
	{
		Asura_SimpleLogfileClear(DEDICATED_LOGFILE);
		s_bLogFileInitialised = true;
	}
    ClearInputBfrDisplay();
    AddOutputInternal( wszData, true, true );
	static u_int ls_uLastFrame = 0;
	u_int uCurFrame = Asura_Timers::GetFrameNumber();
	if (ls_uLastFrame != uCurFrame)
	{
		ls_uLastFrame = uCurFrame;

		const Asura_UnicodeChar* wszDateAndTime = Asura_Time_System::GetCurrentDateAndTimeString(ASURA_DATE_MINIMAL);
		static Asura_UnicodeChar ls_awcLastDateAndTime[128] = L"";
		if (wcscmp(wszDateAndTime, ls_awcLastDateAndTime))
		{
			Asura_PrintfToBfr( ls_awcLastDateAndTime, L"%s", 128, true, wszDateAndTime );
			Asura_SimpleLogfileOutput((DEDICATED_LOGFILE, "\nframe %u: %S\n", uCurFrame, wszDateAndTime));
		}
		else
		{
			Asura_SimpleLogfileOutput((DEDICATED_LOGFILE, "\nframe %u\n", uCurFrame));
		}
	}

	Asura_SimpleLogfileOutput((DEDICATED_LOGFILE, "  %S\n", wszData));
    RefreshInputBfrDisplay( false, false );
}

void Asura_PC_ConsoleWindowIO::AddOutputInternal( const Asura_UnicodeChar* wszOutput, bool bAppendNewline, bool bUpdateCursorPos )
{
    if( INVALID_HANDLE_VALUE == m_hStdOut || !wszOutput || !*wszOutput ) { return; }

    DWORD uIgnore;
    WriteConsoleW( m_hStdOut, wszOutput, wcslen(wszOutput), &uIgnore, NULL );

    if( bAppendNewline ) { WriteConsoleW( m_hStdOut, L"\n", 1, &uIgnore, NULL ); }

    if( bUpdateCursorPos )
    {
        CONSOLE_SCREEN_BUFFER_INFO xBfrInfo;
        GetConsoleScreenBufferInfo( m_hStdOut, &xBfrInfo );
        m_xOutputCursorPos = xBfrInfo.dwCursorPosition;
    }
}

void Asura_PC_ConsoleWindowIO::ClearInputBfrDisplay( void )
{
    if( INVALID_HANDLE_VALUE == m_hStdIn || INVALID_HANDLE_VALUE == m_hStdOut ) { return; }

    if( m_wszCachedInputBfr[0] )
    {
        COORD xPos = { 0, static_cast<short>(m_xOutputCursorPos.Y) };
        SetConsoleCursorPosition( m_hStdOut, xPos );

        Asura_UnicodeChar wszInputBfrOverwrite[ASURA_COMMAND_CONSOLE_INPUT_BUFFER_SIZE];
    
        for( u_int u = 0; u < 2; ++u )
        {
            const Asura_UnicodeChar* wszIn = ( u ? m_wszCachedInputBfr : ASURA_COMMAND_CONSOLE_CURSOR_PROMPT );
            
            Asura_UnicodeChar* wszOut = wszInputBfrOverwrite;
            do
            {
                *wszOut++ = ( *wszIn && L'\n' != *wszIn ? L' ' : *wszIn );
            }
            while( *wszIn++ );

            AddOutputInternal( wszInputBfrOverwrite, false, false );
        }
    }

    SetConsoleCursorPosition( m_hStdOut, m_xOutputCursorPos );
}

void Asura_PC_ConsoleWindowIO::ClearOutput( void )
{
    ClearInputBfrDisplay();

    CONSOLE_SCREEN_BUFFER_INFO xScrBfrInfo;
    GetConsoleScreenBufferInfo( m_hStdOut, &xScrBfrInfo );

    m_xOutputCursorPos.X = 0;
    m_xOutputCursorPos.Y = 0;

    u_int uNumChars = xScrBfrInfo.dwSize.X * xScrBfrInfo.dwSize.Y;
    FillConsoleOutputCharacter( m_hStdOut, ' ', uNumChars, m_xOutputCursorPos, NULL );

    RefreshInputBfrDisplay( false, false );
}

void Asura_PC_ConsoleWindowIO::GetInput( void(*pfnInputCallback)(ASURA_KEYID) )
{
    if( INVALID_HANDLE_VALUE == m_hStdIn ) { return; }

    DWORD uNumInputEvents=0;

    if( !GetNumberOfConsoleInputEvents(m_hStdIn, &uNumInputEvents) || !uNumInputEvents ) 
	{ 
		return; 
	}

    bool bAnyInput = false;

    INPUT_RECORD xNextInputRecord;
    for( DWORD uEvent = 0; uEvent < uNumInputEvents; ++uEvent )
    {
        DWORD uNumInputEventsRead;
        if( !ReadConsoleInputW(m_hStdIn, &xNextInputRecord, 1, &uNumInputEventsRead) || 1 != uNumInputEventsRead )
        {
            break;
        }
		
		if( MOUSE_EVENT == xNextInputRecord.EventType ) 
		{ 
			continue; 
		}

        if( KEY_EVENT != xNextInputRecord.EventType ) 
		{ 
			continue; 
		}

        const KEY_EVENT_RECORD& xKeyEventRecord = xNextInputRecord.Event.KeyEvent;
        if( !xKeyEventRecord.bKeyDown )
        {
            continue;
        }

        const u_int uVK = xKeyEventRecord.wVirtualKeyCode;

        ASURA_KEYID eKeyID = ASURA_KEYID_VOID;
        switch( uVK )
        {
            case VK_BACK:   eKeyID = ASURA_KEYID_BACKSPACE; break;
            case VK_DELETE: eKeyID = ASURA_KEYID_DEL;       break;
            case VK_DOWN:   eKeyID = ASURA_KEYID_DOWN;      break;
            case VK_END:    eKeyID = ASURA_KEYID_END;       break;
            case VK_ESCAPE: eKeyID = ASURA_KEYID_ESCAPE;    break;
            case VK_HOME:   eKeyID = ASURA_KEYID_HOME;      break;
            case VK_LEFT:   eKeyID = ASURA_KEYID_LEFT;      break;
            case VK_RETURN: eKeyID = ASURA_KEYID_ENTER;     break;
            case VK_RIGHT:  eKeyID = ASURA_KEYID_RIGHT;     break;
            case VK_TAB:    eKeyID = ASURA_KEYID_TAB;       break;
            case VK_UP:     eKeyID = ASURA_KEYID_UP;        break;

            default:
            {
                const Asura_UnicodeChar wIn = xKeyEventRecord.uChar.UnicodeChar;
                if( iswgraph(wIn) || L' ' == wIn )
                {
                    // It's a printable character. However, the
                    // Asura_UserInput stuff only copes with ASCII
                    // characters at the moment...
                    if( L' ' <= wIn && wIn <= ASURA_NUMBER_OF_ASCII_KEYS )
                    {
                        eKeyID = static_cast<ASURA_KEYID>( wIn );
                    }
                }
                else if( 3 == wIn )
                {
                    m_bCtrlC = true;
                }
                break;
            }
        }

        if( ASURA_KEYID_VOID != eKeyID )
        {
            for( u_int u = xKeyEventRecord.wRepeatCount; u--; )
            {
                pfnInputCallback( eKeyID );
            }

            bAnyInput = true;
        }
    }
    // End loop through input events.

    if( bAnyInput )
    {
        RefreshInputBfrDisplay( true, true );
    }
}

void Asura_PC_ConsoleWindowIO::RefreshInputBfrDisplay( bool bCheckIfReqd, bool bNeedToClearFirst )
{
    const Asura_UnicodeChar* wszCurInput;
    u_int uCurCursorPos;
    Asura_CommandConsole::GetInputBfrInfo( &wszCurInput, &uCurCursorPos );
    
    if( !bCheckIfReqd || wcscmp(wszCurInput, m_wszCachedInputBfr) )
    {
        if( bNeedToClearFirst ) { ClearInputBfrDisplay(); }

        wcscpy( m_wszCachedInputBfr, wszCurInput );

        SetConsoleCursorPosition( m_hStdOut, m_xOutputCursorPos );
        AddOutputInternal( ASURA_COMMAND_CONSOLE_CURSOR_PROMPT, false, false );
        AddOutputInternal( m_wszCachedInputBfr, false, false );
    }

    if( !bCheckIfReqd || uCurCursorPos != m_uCachedCursorPos )
    {
        m_uCachedCursorPos = uCurCursorPos;

        COORD xPos = { static_cast<short>(m_uCachedCursorPos + wcslen(ASURA_COMMAND_CONSOLE_CURSOR_PROMPT)), static_cast<short>(m_xOutputCursorPos.Y) };
        SetConsoleCursorPosition( m_hStdOut, xPos );
    }
}

void Asura_PC_ConsoleWindowIO::Render( void )
{
    RefreshInputBfrDisplay( true, true );
}

void Asura_PC_ConsoleWindowIO::Reset( void )
{
}

void Asura_PC_ConsoleWindowIO::Update( void )
{
    if( m_bCtrlC )
    {
        // Bit ugly and relies upon the quit command being so-named but 
        // it'll do for now.
        Asura_CommandConsole::Execute( L"quit" );
    }
}

const Asura_UnicodeChar* Asura_PC_ConsoleWindowIO::GetLineSplitPoint(const Asura_UnicodeChar* wszLine)
{
    return wszLine + wcslen(wszLine);

}


int Asura_PC_ConsoleWindowIO::Printf(int x, int y, const char* szformat, ...)
{

	va_list arg;
	va_start(arg, szformat);
	int ret = PrintArgs(x, y, szformat, arg);
	va_end(arg);
	return ret;
}

int Asura_PC_ConsoleWindowIO::PrintArgs(int x, int y, const char* szformat, va_list& args)
{
	char bfr[128];
	Asura_PrintfToBfr( bfr, szformat, args, 128, true);
	return PrintStr(x, y, bfr);
}

int Asura_PC_ConsoleWindowIO::PrintStr(int x, int y, const char* szstring)
{
	Asura_Assert(false, "TO DO : Implement me.");
	return y;
}
