//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_CfgFileReader.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     PC implementations of cfg file reader functions
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "Interface/Asura_CfgFileReader.h"

      
// Stdlib platform-specific.
#include <windows.h>

class Asura_Platform_CfgFileReader: public Asura_CfgFileReader
{
public:
	virtual bool    Open( const Asura_UnicodeChar* wszFilename );
    virtual READRES ReadLine( Asura_UnicodeChar* wszBfr, u_int uBfrSize );
    virtual void    Close( void );

	Asura_Platform_CfgFileReader( void );
	virtual ~Asura_Platform_CfgFileReader( void );

    private:
        // S T A T E /////////////////////////////////////////////////////////

        HANDLE      m_hFile;      // Remaining state is invalid iff this is NULL.
        void*       m_pvFileView;
        bool        m_bUnicode;
        u_int       m_uCharsLeft;

        union
        {
            const Asura_Char*        m_szData;
            const Asura_UnicodeChar* m_wszData;
        };
};


Asura_CfgFileReader* Asura_CfgFileReader::Platform_Create( void )
{
	return new Asura_Platform_CfgFileReader();
}




Asura_Platform_CfgFileReader::Asura_Platform_CfgFileReader( void ) 
: Asura_CfgFileReader()
{
    m_hFile = NULL;
}

Asura_Platform_CfgFileReader::~Asura_Platform_CfgFileReader( void )
{
    Close();
}

bool Asura_Platform_CfgFileReader::Open( const Asura_UnicodeChar* wszFilename )
{
    Asura_Assert( !m_hFile, "" );
    m_hFile = CreateFileW( wszFilename, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_READONLY | FILE_FLAG_SEQUENTIAL_SCAN, NULL );
    if( !m_hFile )
    {
        return false;
    }

    m_uCharsLeft = GetFileSize( m_hFile, NULL );
    HANDLE hMapping = CreateFileMapping( m_hFile, NULL, PAGE_READONLY, 0, 0, NULL );
    if( INVALID_FILE_SIZE == m_uCharsLeft || !hMapping )
    {
        CloseHandle( m_hFile );
        return false;
    }

    m_pvFileView = MapViewOfFile( hMapping, FILE_MAP_READ, 0, 0, 0 );
    CloseHandle( hMapping );

    if( !m_pvFileView )
    {
        CloseHandle( m_hFile );
        return false;
    }

    // (MP - 08.09.2003) NB: Not bothering to deal with byte-reversed Unicode
    // files here...
    m_bUnicode = ( 0xFEFFu == *(reinterpret_cast<const u_short*>(m_pvFileView)) );

    if( m_bUnicode )
    {
        m_wszData = reinterpret_cast<const Asura_UnicodeChar*>( m_pvFileView );

        // Skip byte order mark.
        ++m_wszData;

        // Adjust character count.
        m_uCharsLeft >>= 1;
    }
    else
    {
        m_szData = reinterpret_cast<const Asura_Char*>( m_pvFileView );
    }

    return true;
}

Asura_CfgFileReader::READRES Asura_Platform_CfgFileReader::ReadLine( Asura_UnicodeChar* wszBfr, u_int uBfrSize )
{
    READRES eReadRes = READRES_OK;

    for( Asura_UnicodeChar* const wszEOB = wszBfr + (uBfrSize - 1); wszBfr < wszEOB; )
    {
        if( !m_uCharsLeft-- )
        {
            eReadRes = READRES_EOF;
            break;
        }

        // NB: Need to cast narrow chars to unsigned char first to ensure we
        // get correct value when default is signed chars.
        const Asura_UnicodeChar wNext = ( m_bUnicode ? *m_wszData++ : static_cast<Asura_UnicodeChar>(static_cast<u_char>(*m_szData++)) );

        if( !wNext )
        {
            eReadRes = READRES_EOF;
            break;
        }

        if( L'\r' == wNext )
        {
            continue;
        }
        else if( L'\n' == wNext )
        {
            break;
        }
        else
        {
            *wszBfr++ = wNext;
        }
    }

    *wszBfr++ = L'\0';
    return eReadRes;
}

void Asura_Platform_CfgFileReader::Close( void )
{
    if( m_hFile )
    {
        UnmapViewOfFile( m_pvFileView );
        CloseHandle( m_hFile );
        m_hFile = NULL;
    }
}
