//
// Asura_PC_DirectPlay.cpp
//
// Created by Kevin Lea - 15/06/2001 15:06:01
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Asura_PC_DirectPlay.h"

#include "Core/Asura_Timers.h"
//#include "Core/Asura_Windows.h"
#include "Core/Asura_Hash.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_MessageHeader.h"
#include "Network/Asura_Network.h"

#include "Network/Asura_PC_Message.h"
#include "Asura_PC_DirectPlay_Server.h"

Asura_PC_DirectPlay_PlayerInfo* Asura_PC_DirectPlay::PlayerInfo;

// need interface to set this via console or command-line
bool Asura_PC_DirectPlay::s_bUseNetworkSimulator = false;
u_int Asura_PC_DirectPlay::s_uPort = ASURA_PC_DIRECTPLAY_DEFAULTPORT;
u_int Asura_PC_DirectPlay::s_uPasswordHash = ASURA_HASH_ID_UNSET;

#ifdef ASURA_DIRECTPLAY_MESSAGELOGGING
Asura_PC_DirectPlay::MSG_LOG_LEVEL Asura_PC_DirectPlay::s_eMsgLogLevel;
#endif

#define ASURA_PACKETBUFFERSIZE 16384
BYTE PacketBuffer[ASURA_PACKETBUFFERSIZE];



void Asura_PC_DirectPlay::Initialise(void)
{
	if (Asura_Network::IsServer())
	{
		PlayerInfo = new Asura_PC_DirectPlay_PlayerInfo[Asura_Network::GetMaxNumberOfMachines()];
	}
	else
	{
		PlayerInfo = NULL;
	}

	s_pxClient = NULL;
	s_pxServer = NULL;
	NumberOfActiveSessions = 0;

	SessionsStorage = new Asura_PC_DirectPlay_SessionInfo*[Asura_Network::GetMaxNumberOfSessions()];

	for (u_int n=0; n<Asura_Network::GetMaxNumberOfSessions(); n++)
	{
		SessionsStorage[n] = 0;
	}

#ifdef ASURA_DIRECTPLAY_MESSAGELOGGING
    SetMsgLogLevel( MSG_LOG_LEVEL_NONE );
#endif
}

void Asura_PC_DirectPlay::Shutdown(void)
{
    DeleteSessionsInfo();
    DeleteCriticalSection(&CriticalSection_EnumSessions);

	if (PlayerInfo) delete [] PlayerInfo;

	delete [] SessionsStorage;
}



void Asura_PC_DirectPlay::PackMessageIntoBuffer(const Asura_Message* message, DPN_BUFFER_DESC& buffer)
{
	Asura_MessageHeader* header = (Asura_MessageHeader*)PacketBuffer;
	
	*header = *message;

	unsigned int headerSize = sizeof(Asura_MessageHeader);
	unsigned int totalSize = headerSize + header->m_uDataSize;

	Asura_Assert(totalSize<ASURA_PACKETBUFFERSIZE,"Packet buffer size too small");

	memcpy(&PacketBuffer[headerSize],message->GetData(),header->m_uDataSize);

	buffer.dwBufferSize = totalSize;
	buffer.pBufferData = PacketBuffer;
}

bool Asura_PC_DirectPlay::ShouldAddMessageToQueue(u_int uMsgID)
{
	//if ready for messages , then add all messages we get
	if(Asura_Network::IsReadyForMessages()) return true;

	//otherwise just add a few special messages
	switch(uMsgID)
	{
		case ASURA_MSG_VALIDCLIENTGUID :
		case ASURA_PC_MSG_GCDCHALLENGE :
		case ASURA_PC_MSG_GCDRESPONSE :
		case ASURA_PC_MSG_GCDAUTHORISATION :
			return true;

		default :
			break;
	}
	return false;
}

void Asura_PC_DirectPlay::UnpackMessageFromBuffer(const BYTE* buffer, u_int uNumBytes, bool bWasGuaranteed)
{
	const Asura_MessageHeader* header = reinterpret_cast<const Asura_MessageHeader*>(buffer);

	Asura_Assert(header->m_uDataSize + sizeof(Asura_MessageHeader) == uNumBytes, "data sizes do not match");
	// SNH - 15/05/2003 15:54:54 - only add messages to queue if application is active
	if( ShouldAddMessageToQueue(header->m_usMsgID) || bWasGuaranteed)
	{
		Asura_Network::AddMessageToQueue(header, buffer + sizeof(Asura_MessageHeader), bWasGuaranteed);
	}
}


#ifdef ASURA_DIRECTPLAY_MESSAGELOGGING
// (MH - 08.01.2004) disabled this code as it's now done in independent code.

static const char* NETWORK_MSG_LOGFILE = "logfiles\\network_msgs.log";

void Asura_PC_DirectPlay::SetMsgLogLevel( MSG_LOG_LEVEL eLevel )
{
    static bool bEverBeenOn = false;

    Asura_Assert( MSG_LOG_LEVEL_NUM_LEVELS != eLevel, "" );
    s_eMsgLogLevel = eLevel;

    if( !bEverBeenOn && MSG_LOG_LEVEL_NONE != s_eMsgLogLevel )
    {
        Asura_SimpleLogfileClear( NETWORK_MSG_LOGFILE );
        bEverBeenOn = true;
    }
}

void Asura_PC_DirectPlay::LogMsgInfo( const DPN_BUFFER_DESC* pxBfrDesc, bool bIndent )
{
    // Only supporting a single level at the moment.
    Asura_Assert( MSG_LOG_LEVEL_HEADER_ONLY == s_eMsgLogLevel, "" );

    Asura_MessageHeader* pxMsgHdr = reinterpret_cast<Asura_MessageHeader*>( pxBfrDesc->pBufferData );

    #define CASE_TO_STR( _str, _case ) \
        case (_case): (_str) = #_case; break

    char acBuffer[32];
	sprintf(acBuffer, "[misc (%d)]", pxMsgHdr->m_usMsgID);
	const char* szMsgID = acBuffer;
    switch( pxMsgHdr->m_usMsgID )
    {
	    CASE_TO_STR( szMsgID, ASURA_MSG_INITIALISE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_DESTRUCT );
	    CASE_TO_STR( szMsgID, ASURA_MSG_ACTIVATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_DEACTIVATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_TRIGGER );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SPLITTER_ACTIVATEOUTPUT );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SPLITTER_DEACTIVATEOUTPUT );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SPLITTER_TRIGGEROUTPUT );
        CASE_TO_STR( szMsgID, ASURA_MSG_CHANGE_ANIMATION );
	    CASE_TO_STR( szMsgID, ASURA_MSG_GLASS_SMASH );
	    CASE_TO_STR( szMsgID, ASURA_MSG_UNTRIGGER );
        CASE_TO_STR( szMsgID, ASURA_MSG_LOCK );
	    CASE_TO_STR( szMsgID, ASURA_MSG_UNLOCK );
	    CASE_TO_STR( szMsgID, ASURA_MSG_LIFTCALL );
	    CASE_TO_STR( szMsgID, ASURA_MSG_RESET );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SWITCHON );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SWITCHOFF );
	    CASE_TO_STR( szMsgID, ASURA_MSG_TEMPLATE_TRIGGERINPUT );
	    CASE_TO_STR( szMsgID, ASURA_MSG_TEMPLATE_TRIGGEROUTPUT );
	    CASE_TO_STR( szMsgID, ASURA_MSG_ACTIVATECUTSCENEWITHTIME );
	    CASE_TO_STR( szMsgID, ASURA_MSG_END_OF_STATIC_MESSAGES );
	    CASE_TO_STR( szMsgID, ASURA_MSG_CLIENTUPDATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PACKEDMESSAGES );
	    CASE_TO_STR( szMsgID, ASURA_MSG_NETMSGUPDATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_CLIENT_REQUESTCREATEMSG );
	    CASE_TO_STR( szMsgID, ASURA_MSG_VALIDCLIENTGUID );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PING_REQUEST );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PING_RESPONSE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PLAYERCONNECTED );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PLAYERDISCONNECTED );
        CASE_TO_STR( szMsgID, ASURA_MSG_CUTSCENECONTROLLER_REQ_PROFILE_CUTSCENE );
        CASE_TO_STR( szMsgID, ASURA_MSG_CUTSCENECONTROLLER_PROFILE_CUTSCENE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_CUTSCENECONTROLLER_ACTIVATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_CUTSCENECONTROLLER_CLIENTUPDATE );
#ifdef ASURA_SUPPORT_SERVERCUTSCENES		
        CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_STANDARD );
        CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_PRIMARYANIMATION );
        CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_SECONDARYANIMATION );
        CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_PRIMARYANIMATION_TOPOS );
        CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_SECONDARYANIMATION_TOPOS );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_ADDATTACHMENTSHAPE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_ADDATTACHMENTHIERARCHY );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_ADDATTACHMENTACTOR );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_ADDATTACHMENTENTITY );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTOR_CLIENTUPDATE_REMOVEATTACHMENT );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SERVERCUTSCENEACTORSHAPE_CLIENTUPDATE_STANDARD );
#endif // #ifdef ASURA_SUPPORT_SERVERCUTSCENES
	    CASE_TO_STR( szMsgID, ASURA_MSG_GLASS_APPLYDECAL );
	    CASE_TO_STR( szMsgID, ASURA_MSG_GLASS_SMASHWITHPOSITION );
	    CASE_TO_STR( szMsgID, ASURA_MSG_GLASS_CLIENTUPDATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PICKUP_CLIENTUPDATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PICKUP_COLLECTED );
	    CASE_TO_STR( szMsgID, ASURA_MSG_ADVANCEDLIGHT_CREATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_DOOR_CREATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_LIFT_CREATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_SWITCH_CREATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_CREATE_BEAM );
	    CASE_TO_STR( szMsgID, ASURA_MSG_MUSICSYSTEM_PLAY );
	    CASE_TO_STR( szMsgID, ASURA_MSG_EXPLOSION_CREATE );
	    CASE_TO_STR( szMsgID, ASURA_MSG_PLAYERPOSITIONVALIDATED );
        
        default:
            Project_GetMsgIDDesc( pxMsgHdr->m_usMsgID, &szMsgID );
            break;
    }

    const char* szMsgDest = NULL;
    switch( pxMsgHdr->m_uToGuid )
    {
        CASE_TO_STR( szMsgDest, ASURA_GUID_SERVER );
        CASE_TO_STR( szMsgDest, ASURA_GUID_CLIENT );
        CASE_TO_STR( szMsgDest, ASURA_GUID_ALL_PLAYERS );
        CASE_TO_STR( szMsgDest, ASURA_GUID_OTHER_PLAYERS );
        CASE_TO_STR( szMsgDest, ASURA_GUID_MY_CLIENT_ENTITIES );
        CASE_TO_STR( szMsgDest, ASURA_GUID_ALL_CLIENTMACHINES );
        CASE_TO_STR( szMsgDest, ASURA_GUID_ALL_OTHER_CLIENTMACHINES );
        CASE_TO_STR( szMsgDest, ASURA_GUID_OTHER_PLAYERS_CLIENTMACHINES );
        CASE_TO_STR( szMsgDest, ASURA_GUID_PLAYERS_CLIENTMACHINE );
        CASE_TO_STR( szMsgDest, ASURA_GUID_PLAYER );
        CASE_TO_STR( szMsgDest, ASURA_GUID_PLAYER_ONTHISMACHINE );
    }

    #undef CASE_TO_STR

    if( bIndent )
    {
        if( szMsgDest )
        {
            Asura_SimpleLogfileOutput(( NETWORK_MSG_LOGFILE, "        (%u, %s) -> %s\n", pxBfrDesc->dwBufferSize, szMsgID, szMsgDest ));
        }
        else
        {
            Asura_SimpleLogfileOutput(( NETWORK_MSG_LOGFILE, "        (%u, %s) -> %s\n", pxBfrDesc->dwBufferSize, szMsgID, pxMsgHdr->m_uToGuid ));
        }
    }
    else
    {
        const float fTime = Asura_Timers::GetGlobalGameTime();
        if( szMsgDest )
        {
            Asura_SimpleLogfileOutput(( NETWORK_MSG_LOGFILE, "%6.2f: (%u, %s) -> %s\n", fTime, pxBfrDesc->dwBufferSize, szMsgID, szMsgDest ));
        }
        else
        {
            Asura_SimpleLogfileOutput(( NETWORK_MSG_LOGFILE, "%6.2f: (%u, %s) -> %u\n", fTime, pxBfrDesc->dwBufferSize, szMsgID, pxMsgHdr->m_uToGuid ));
        }
    }

    // Special case for packed messages: describe the contents.
    if( ASURA_MSG_PACKEDMESSAGES == pxMsgHdr->m_usMsgID )
    {
        u_char* pucData = reinterpret_cast<u_char*>( pxMsgHdr + 1 );
        u_char* pucEnd  = pucData + pxMsgHdr->m_uDataSize;
        
        DPN_BUFFER_DESC xTmpBfrDesc;
        while( pucData < pucEnd )
        {
            Asura_MessageHeader* pxNextHdr = reinterpret_cast<Asura_MessageHeader*>( pucData );
            u_int uTotalSize = sizeof(Asura_MessageHeader) + pxNextHdr->m_uDataSize;

            xTmpBfrDesc.pBufferData  = pucData;
            xTmpBfrDesc.dwBufferSize = uTotalSize;
            LogMsgInfo( &xTmpBfrDesc, true );

            pucData += uTotalSize;
        }

        return;
    }
}

#endif

const Asura_PC_DirectPlay_PlayerInfo* Asura_PC_DirectPlay::GetPlayerInfo(int iIndex)
{
	Asura_Assert(PlayerInfo,"");
	Asura_Assert(iIndex>=0 && iIndex<(int)Asura_Network::GetMaxNumberOfMachines(),"");
	Asura_Assert(PlayerInfo[iIndex].IsValid,"");

	return &PlayerInfo[iIndex];
}

int Asura_PC_DirectPlay::GetPlayerIPAddress(int iIndex)
{
	const Asura_PC_DirectPlay_PlayerInfo* pxPlayer = GetPlayerInfo(iIndex);
	int iReturnAddress = 0;
	
	IDirectPlay8Address* pxDPAddress = 0;
	HRESULT hres;
	if(iIndex)
	{
		hres = s_pxServer->GetClientAddress(pxPlayer->NetworkID,&pxDPAddress,0);
	}
	else
	{
		DWORD uNumAddress = 1;
		hres = s_pxServer->GetLocalHostAddresses(&pxDPAddress,&uNumAddress,0);
	}

	Asura_Assertf(hres == S_OK,("Failed to get player address. hres = 0x%x",hres));
	if(hres == S_OK)
	{
		DWORD uNumComponents;
		hres = pxDPAddress->GetNumComponents(&uNumComponents);
		Asura_Assertf(hres == S_OK,("Failed to get number of components. hres = 0x%x",hres));

		if(hres == S_OK)
		{
			bool bFoundAddress = false;
			for(u_int i=0;i<uNumComponents;i++)
			{
				WCHAR wszNameBuffer[200];
				WCHAR wszBuffer[100];
				DWORD uNameLength = 200;
				DWORD uBufferSize = 200;
				DWORD uDataType;
				
				hres =  pxDPAddress->GetComponentByIndex(i,wszNameBuffer,&uNameLength,wszBuffer,&uBufferSize,&uDataType);
				if(!wcsicmp(wszNameBuffer,L"hostname"))
				{
					int aiIP[4];
					if(swscanf(wszBuffer,L"%d.%d.%d.%d",&aiIP[0],&aiIP[1],&aiIP[2],&aiIP[3])==4)
					{
						iReturnAddress = (aiIP[0]<<0) | (aiIP[1]<<8) | (aiIP[2]<<16) | (aiIP[3]<<24);
						bFoundAddress = true;
						break;
					}
				}
			}
			Asura_Assert(bFoundAddress,"Failed to find IP address");
		}
		pxDPAddress->Release();
	}

	return iReturnAddress;
}

void Asura_PC_DirectPlay::DisconnectPlayer(int iIndex)
{
	DPNID uID = GetPlayerInfo(iIndex)->NetworkID;
	#ifdef ASURA_ASSERT
	HRESULT hres = 
	#endif
	s_pxServer->DestroyClient(uID,0,0,0);

	Asura_Assertf(hres == S_OK,("Failed to delete client. hres = 0x%x",hres));
}


void Asura_PC_DirectPlay::SetPassword(const char* szPassword)
{
	Asura_Assert(szPassword, "no password");
	if (szPassword)
	{
		s_uPasswordHash = Asura_GetHashID(szPassword);
	}

//	if (s_pxServer)
//	{
//		UpdateServerPassword(s_uPasswordHash);
//	}
}
