//////////////////////////////////////////////////////////////////////////////
//
// Asura_NetSocketSteamworks.h
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_PC_SUPPORT_STEAMWORKS

// This file's header.
#include "Network/Steamworks/Asura_PC_NetSocketSteamworks.h"

// asura
#include "Core/Asura_Multithread.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Asura_NetConnection.h"
#include "Network/Asura_NetSocket_BSD.h"
#include "Network/Steamworks/Asura_PC_Steamworks.h"

// middleware
#include "steam/steam_api.h"
#include "steam/isteamgameserver.h"
#include "steam/steam_gameserver.h"


// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

// m_xSocket is being reused as the steam handle, so make sure SOCKET and SNetSocket_t are the same.
ASURA_STATIC_ASSERT(sizeof(SNetSocket_t) == sizeof(SOCKET));

// pimpl classes

// Server Implementation
class Asura_PC_NetSocketSteamworksServerImp : public Asura_PC_NetSocketSteamworksImp
{
	public:
		Asura_PC_NetSocketSteamworksServerImp( u_int uMaxMachines, Asura_PC_NetSocketSteamworks* pxParent );
		virtual ~Asura_PC_NetSocketSteamworksServerImp();
		virtual bool Bind(const Asura_NetAddress& xInAddress, u_short usPort);
		virtual bool SendTo(const Asura_NetSocket_MessageInfo* pxMessageInfo);
		virtual bool ReceiveMessage(char* pcBuffer, int* piSize, Asura_NetAddress* pxSourceAddress, u_short* pusSourcePort );
		virtual void Close();
		virtual int GetLastError() const;

	private:

		bool Listen();

		void Close(u_int iSocketIndex);

		struct ConnectionSocketInfo // structure that describe a connected socket
		{
			ConnectionSocketInfo()
				: m_bIsActive(false),
				m_bIsConnected(false),
				m_xAddress(),
				m_usPort(0),
				m_xSocket(0) 
			{}

			bool m_bIsActive;
			bool m_bIsConnected;
			Asura_NetAddress m_xAddress;
			u_short m_usPort;
			SNetSocket_t m_xSocket; 
		};

		Asura_NetAddress m_xListenAddress;
		u_short m_usLisenPort;
		SNetListenSocket_t m_xListenSocket;
		u_int m_iMaxConnectedSockets;
		ConnectionSocketInfo *m_xConnectionSockets;

		// client connection state
		STEAM_CALLBACK( Asura_PC_NetSocketSteamworksServerImp, OnSocketStatusCallback, SocketStatusCallback_t, m_pfnCallbackSocketStatus );
};

// Client Implementation
class Asura_PC_NetSocketSteamworksClientImp : public Asura_PC_NetSocketSteamworksImp
{
	public:
		Asura_PC_NetSocketSteamworksClientImp( Asura_PC_NetSocketSteamworks* pxParent );
		virtual ~Asura_PC_NetSocketSteamworksClientImp();
		virtual bool Bind(const Asura_NetAddress& xInAddress, u_short usPort);
		virtual bool SendTo(const Asura_NetSocket_MessageInfo* pxMessageInfo);
		virtual bool ReceiveMessage(char *pcBuffer, int *iSize,  Asura_NetAddress* pxSourceAddress, u_short* pusSourcePort);
		virtual void Close();
		virtual int GetLastError() const;

	private:

		bool Connect(const Asura_NetAddress& xRemoteAddress, u_short usRemotePort);

		Asura_NetAddress m_xAddress;
		u_short m_usPort;
		SNetSocket_t m_xSocket;
		bool m_bIsFullyConnected;

		STEAM_CALLBACK( Asura_PC_NetSocketSteamworksClientImp, OnSocketStatusCallback, SocketStatusCallback_t, m_pfnSocketStatusCallback );
};


Asura_PC_NetSocketSteamworksServerImp::Asura_PC_NetSocketSteamworksServerImp( u_int uMaxMachines, Asura_PC_NetSocketSteamworks* pxParent )
	: Asura_PC_NetSocketSteamworksImp( pxParent ),
	m_xListenAddress(), 
	m_usLisenPort(0),
	m_xListenSocket(0),
	m_iMaxConnectedSockets(uMaxMachines * 2), // reserve twice as many physical sockets as there are logical sockets - this allow us to reject sockets gracefully instead of just dropping them.
	m_xConnectionSockets(0),
	m_pfnCallbackSocketStatus(this, &Asura_PC_NetSocketSteamworksServerImp::OnSocketStatusCallback)
{
	if( !Asura_PC_Steamworks::IsInitialised() )
	{
		Asura_Assert( false, "Expected steamworks to be initialised here" );
	}
	m_xConnectionSockets = new ConnectionSocketInfo[m_iMaxConnectedSockets];
}


Asura_PC_NetSocketSteamworksServerImp::~Asura_PC_NetSocketSteamworksServerImp()
{
	Close();
	delete [] m_xConnectionSockets;
}

bool Asura_PC_NetSocketSteamworksServerImp::Bind(const Asura_NetAddress& xInAddress, u_short usPort)
{
	// This function just remember what to listen on. Actual listening is done on the first read operation.
	Asura_Assert(m_xListenAddress == Asura_NetAddress() && m_usLisenPort == 0, "Bind already called.");
	m_xListenAddress = xInAddress; 
	if( usPort == 0 )
	{
		// If we weren't supplied with a port, use the netsocket default
		m_usLisenPort = ASURA_NETSOCKET_DEFAULTPORT;
	}
	else
	{
		m_usLisenPort = usPort;
	}
	
	return true;
}

bool Asura_PC_NetSocketSteamworksServerImp::SendTo(const Asura_NetSocket_MessageInfo* pxMessageInfo)
{
	for (u_int i = 0; i < m_iMaxConnectedSockets; ++i)
	{
		if (m_xConnectionSockets[i].m_bIsActive == true && 
			m_xConnectionSockets[i].m_bIsConnected == true &&
			m_xConnectionSockets[i].m_xAddress == pxMessageInfo->m_xAddress &&
			m_xConnectionSockets[i].m_usPort == pxMessageInfo->m_usPort)
		{
			if(!SteamNetworking()->SendDataOnSocket(m_xConnectionSockets[i].m_xSocket, 
													   const_cast<char*> (pxMessageInfo->m_pcMessage), 
													   pxMessageInfo->m_iLength, 
													   m_bUseReliable))
			{
				Close(i);
			}
			break;
		}
	}

	// might be still trying to send to a connection that was closed. Let the upper layer discover it on its own. (since this layer behave like UDP)
	return true;
}

bool Asura_PC_NetSocketSteamworksServerImp::ReceiveMessage(char* pcBuffer, int* piSize, Asura_NetAddress* pxSourceAddress, u_short* pusSourcePort )
{
	if (!m_xListenSocket && !Listen())
	{
		return true;
	}

	// receive data from any socket that is associated with this listen socket
	SNetSocket_t xSocket;
	uint32 uReturnSize;
	
	while(true)
	{

		if (!SteamNetworking()->RetrieveData( m_xListenSocket, pcBuffer, *piSize, &uReturnSize, &xSocket ))
		{
			return false;
		}

		bool bFound = false;
		u_int iEmptySlot = m_iMaxConnectedSockets;

		for (u_int i = 0; i < m_iMaxConnectedSockets && !bFound ; ++i) // look for our socket and also for a free slot
		{
			if (m_xConnectionSockets[i].m_xSocket == xSocket )
			{
				bFound = true;
				if (m_xConnectionSockets[i].m_bIsActive && m_xConnectionSockets[i].m_bIsConnected)
				{
					*pxSourceAddress = m_xConnectionSockets[i].m_xAddress;
					*pusSourcePort = m_xConnectionSockets[i].m_usPort;
					*piSize = uReturnSize;
					return true;
				}
			}
			else if (iEmptySlot == m_iMaxConnectedSockets && !m_xConnectionSockets[i].m_bIsActive )
			{
				iEmptySlot = i;
			}
		}

		if (!bFound && iEmptySlot < m_iMaxConnectedSockets) // the socket was not found - it is a new socket - add the socket. 
		{
			// find the actual socket address.
			Asura_NetAddress xAddress;
			u_short usPort;
			if (!SteamNetworking()->GetSocketInfo(xSocket, &xAddress.m_xSteamID, NULL, &xAddress.m_uAddress, &usPort ))
			{
				Asura_Assert(false, "Socket info retieve failed");
				return false;
			}
			else
			{
				m_xConnectionSockets[iEmptySlot].m_bIsActive = true;
				m_xConnectionSockets[iEmptySlot].m_bIsConnected = true;
				m_xConnectionSockets[iEmptySlot].m_usPort = m_usLisenPort; // TY 24.02.09 - GetSocketInfo gets the wrong port?!
				m_xConnectionSockets[iEmptySlot].m_xAddress = xAddress;
				m_xConnectionSockets[iEmptySlot].m_xSocket = xSocket;


				*pxSourceAddress = xAddress;
				*pusSourcePort = usPort;
				*piSize = uReturnSize;
				
				return true;
			}
		}
	}
}

void Asura_PC_NetSocketSteamworksServerImp::Close()
{
	if (m_xListenSocket) // destroying listen socket destroy all the sockets that were created from it
	{
		SteamNetworking()->DestroyListenSocket( m_xListenSocket, false);

		for (u_int i = 0; i < m_iMaxConnectedSockets; ++i)
		{
			m_xConnectionSockets[i] = ConnectionSocketInfo();
		}
		m_xListenSocket = 0;
	}
}

int Asura_PC_NetSocketSteamworksServerImp::GetLastError() const
{
	return 0;
}

void Asura_PC_NetSocketSteamworksServerImp::Close(u_int uSocketIndex)
{
	Asura_Assert(uSocketIndex < m_iMaxConnectedSockets, "Socket index out of range");
	if (uSocketIndex < m_iMaxConnectedSockets)
	{
		SteamNetworking()->DestroySocket(m_xConnectionSockets[uSocketIndex].m_xSocket, false);

		m_xConnectionSockets[uSocketIndex] = ConnectionSocketInfo();
	}
}

bool Asura_PC_NetSocketSteamworksServerImp::Listen()
{
	if (m_xListenSocket == 0)
	{
		m_xListenSocket = SteamNetworking()->CreateListenSocket( ASURA_PC_STEAMWORKS_DEFAULT_VPORT , m_xListenAddress.m_uAddress, m_usLisenPort, true ); 
	}

	return m_xListenSocket != 0; 
}


void Asura_PC_NetSocketSteamworksServerImp::OnSocketStatusCallback( SocketStatusCallback_t *pxCallback)
{

	// if its not for us return
	if ( pxCallback->m_hListenSocket != m_xListenSocket)
		return;

	// The server will only get updates due to disconnection (since the server is aware only of a fully connected clients)
	// Therefore just close any the mentioned socket 
	for (u_int i = 0; i < m_iMaxConnectedSockets; ++i)
	{
		if (pxCallback->m_hSocket == m_xConnectionSockets[i].m_xSocket)
		{
			// Close the Asura_NetConnection if necessary (is this hacky?)
			Asura_NetConnection* pxConnection = m_pxParent->FindConnection( m_xConnectionSockets[i].m_xAddress, m_xConnectionSockets[i].m_usPort );
			if( pxConnection )
			{
				pxConnection->SetClosed( 0.0f );
			}

			Close(i);
			return;
		}
	}
}

Asura_PC_NetSocketSteamworksClientImp::Asura_PC_NetSocketSteamworksClientImp( Asura_PC_NetSocketSteamworks* pxParent )
	: Asura_PC_NetSocketSteamworksImp( pxParent ),
	m_xAddress(), 
	m_usPort(0),
	m_xSocket(0),
	m_bIsFullyConnected(false),
	m_pfnSocketStatusCallback(this, &Asura_PC_NetSocketSteamworksClientImp::OnSocketStatusCallback)
{
	if( !Asura_PC_Steamworks::IsInitialised() )
	{
		Asura_Assert( false, "Expected steamworks to be initialised here" );
	}
}


Asura_PC_NetSocketSteamworksClientImp::~Asura_PC_NetSocketSteamworksClientImp()
{
	Close();
}

bool Asura_PC_NetSocketSteamworksClientImp::Bind(const Asura_NetAddress& xInAddress, u_short usPort)
{
	// silently ignored - bind for client is not implemented in steam
	return true;
}


bool Asura_PC_NetSocketSteamworksClientImp::SendTo(const Asura_NetSocket_MessageInfo* pxMessageInfo)
{
	if (m_xSocket == 0 && !Connect(pxMessageInfo->m_xAddress, pxMessageInfo->m_usPort))
	{
		return true;			
	}

	if(m_bIsFullyConnected && pxMessageInfo->m_xAddress == m_xAddress && pxMessageInfo->m_usPort == m_usPort)
	{
		if(!SteamNetworking()->SendDataOnSocket(m_xSocket, const_cast<char*>(pxMessageInfo->m_pcMessage), pxMessageInfo->m_iLength, m_bUseReliable)) // Reliable is done in higher level
		{
			Close(); // If sending failed close the socket - we will retry to open it a new on the next send.
		}
	}

	return true;
}

bool Asura_PC_NetSocketSteamworksClientImp::ReceiveMessage(char *pcBuffer, int *piSize,  Asura_NetAddress* pxSourceAddress, u_short* pusSourcePort )
{
	uint32 uReturnSize;

	if (!m_bIsFullyConnected || !SteamNetworking()->RetrieveDataFromSocket(m_xSocket, pcBuffer, *piSize, &uReturnSize))
	{
		return false;
	}

	*piSize = static_cast<int>(uReturnSize);

	*pxSourceAddress = m_xAddress;
	*pusSourcePort = m_usPort;
	
	return true;
}

void Asura_PC_NetSocketSteamworksClientImp::Close()
{
	if (m_xSocket)
	{
		SteamNetworking()->DestroySocket( m_xSocket, false );

		m_bIsFullyConnected = false;
		m_xSocket = 0;
		m_xAddress = Asura_NetAddress();
		m_usPort = 0;
	}
}

int Asura_PC_NetSocketSteamworksClientImp::GetLastError() const
{
	return 0;
}

void Asura_PC_NetSocketSteamworksClientImp::OnSocketStatusCallback( SocketStatusCallback_t *pxCallback)
{
	// make sure the callback is for our socket
	if ( pxCallback->m_hSocket != m_xSocket )
		return;

	if ( pxCallback->m_eSNetSocketState == k_ESNetSocketStateConnected )
	{
		m_bIsFullyConnected = true;
	}
	else if ( pxCallback->m_eSNetSocketState == k_ESNetSocketStateRemoteEndDisconnected 
			  || pxCallback->m_eSNetSocketState == k_ESNetSocketStateConnectionBroken || 
			  pxCallback->m_eSNetSocketState >= k_ESNetSocketStateDisconnecting )
	{
		
		// Close the Asura_NetConnection if necessary (is this hacky?)
		Asura_NetConnection* pxConnection = m_pxParent->FindConnection( m_xAddress, m_usPort );
		if( pxConnection )
		{
			pxConnection->Close( true, ASURA_CLOSE_LOCAL );
		}

		Close();
	}
}


// This timeout is not really used for anything 
// If connection failed - we will try again as long as the higher level socket try to send data.
DEBUGVAR int dbg_iConnectTimeoutSeconds = 20; 

bool Asura_PC_NetSocketSteamworksClientImp::Connect(const Asura_NetAddress& xRemoteAddress, u_short usRemotePort)
{
	if( !SteamNetworking() )
	{
		Asura_Assert( false, "Steam networking API not available" );
		return false;
	}

	// assume that only clients connect - use m_xConnectionSockets[0]
	Asura_Assert(m_xSocket == 0, "Trying to connect an already connected socket");
	if (m_xSocket)
	{
		Close();
	}

	if (xRemoteAddress.m_xSteamID.IsValid())
	{
		m_xSocket = SteamNetworking()->CreateP2PConnectionSocket( xRemoteAddress.m_xSteamID, ASURA_PC_STEAMWORKS_DEFAULT_VPORT, dbg_iConnectTimeoutSeconds, true);
		Asura_OutputToDebugger( ( "Created P2P Steam socket" ) );
	}
	else
	{
		m_xSocket = SteamNetworking()->CreateConnectionSocket(xRemoteAddress.m_uAddress, usRemotePort, dbg_iConnectTimeoutSeconds);
		Asura_OutputToDebugger( ( "Created Steam socket on port %u", usRemotePort ) );
	}


	if (!m_xSocket)
	{
		return false;
	}
	else
	{
		m_xAddress = xRemoteAddress;
		m_usPort = usRemotePort;

		if( m_pxParent )
		{
			m_pxParent->UpdateAddress( m_xAddress, m_usPort );
		}

		return true;
	}
}



Asura_PC_NetSocketSteamworks::Asura_PC_NetSocketSteamworks( u_int uMaxMachines, ASURA_PC_STEAMSOCKET_TYPE eType )
	: Asura_NetSocket(uMaxMachines)
{
	if( eType == ASURA_PC_STEAMSOCKET_TYPE_AUTO )
	{
		if ( Asura_Network::IsServer() )
		{
			m_pxImp = new Asura_PC_NetSocketSteamworksServerImp(uMaxMachines, this);
		}
		else
		{
			Asura_Assert(uMaxMachines == 1, "Client can only connect to a single server");
			m_pxImp = new Asura_PC_NetSocketSteamworksClientImp( this );
		}
	}
	else if( eType == ASURA_PC_STEAMSOCKET_TYPE_SERVER )
	{
		m_pxImp = new Asura_PC_NetSocketSteamworksServerImp(uMaxMachines, this);
	}
	else
	{
		m_pxImp = new Asura_PC_NetSocketSteamworksClientImp( this );
	}
}


Asura_PC_NetSocketSteamworks::~Asura_PC_NetSocketSteamworks()
{
	delete m_pxImp;
}

// In the case of steam Asura_NetAddress can be either ip/udp pair like in BSD or a 64 bits opaque value that represent a steam ID.
// There is no standard for this 64 bits - so i decided to represent it similar to the ip dotted form - 8 bytes dotted form "xxx.xxx.xxx.xxx.xxx.xxx.xxx"
const char* Asura_PC_NetSocketSteamworks::GetStringFromAddress(const Asura_NetAddress& xAddress, unsigned short usPort)
{
	// prefer to use the steam ID as it can be used for NAT bypassing.  
	if (!xAddress.m_xSteamID.IsValid()) // ip/udp - Steamworks unlike BSD assume host ordered ip. Therefore just calling BSD conversion function will not work here.
	{
		static char acAddress[sizeof("XXX.XXX.XXX.XXX:XXXXX")];
		Asura_PrintfToBfr(acAddress, "%u.%u.%u.%u:%u", sizeof("XXX.XXX.XXX.XXX:XXXXX"), true,
			(xAddress.m_uAddress >> 24) & 0xFF,
			(xAddress.m_uAddress >> 16) & 0xFF,
			(xAddress.m_uAddress >> 8) & 0xFF,
			(xAddress.m_uAddress >> 0) & 0xFF,
			usPort);

		return acAddress;
		/*return Asura_NetSocket_BSD::GetStringFromAddress(xAddress.m_uIP, usPort);*/
	}
	else
	{

		static char acAddress[sizeof("XXX.XXX.XXX.XXX.XXX.XXX.XXX.XXX")];
		uint64 ulId = xAddress.m_xSteamID.ConvertToUint64();
		int iPos =0;
		for (int i = 8; i > 0; --i) // iterate from MSB to LSB 
		{
			uint8 ucByte = static_cast<uint8>( (ulId >> (8*(i-1))) & 0xFF ); // get the i'th byte

			acAddress[iPos++] = ucByte / 100 + '0';
			ucByte = ucByte % 100;
			acAddress[iPos++] = ucByte / 10 + '0';
			ucByte = ucByte % 10;
			acAddress[iPos++] = ucByte + '0';
			acAddress[iPos++] = '.';
		}
		acAddress[iPos] = '\0';
		return acAddress;
	}

}

// steam can be either opaque 8 bytes value or the normal IP,UDP pair
bool Asura_PC_NetSocketSteamworks::GetAddressFromString(const char* szAddress, Asura_NetAddress* pxAddress, unsigned short* pusPort)
{
	*pusPort = 0;
	*pxAddress = Asura_NetAddress();

	// Passing in a null address is valid - simply return 0
	// Check whether the string is null or empty
	if(!szAddress || !szAddress[0])
	{
		return 0;
	}

	char acAddressCopy[32];
	strncpy(acAddressCopy, szAddress, 31);
	acAddressCopy[31] = '\0';

	// count the number of dots to distinguish between "XXX.XXX.XXX.XXX:XXXX"/"XXX.XXX.XXX.XXX" and "XXX.XXX.XXX.XXX.XXX.XXX.XXX.XXX"
	int iDots = 0; 
	for (int i = 0; acAddressCopy[i] != '\0' && i < 32; ++i ) // 32 just a safe number to avoid overflow 
	{
		if (acAddressCopy[i] == '.')
		{
			++iDots;
		}
	}

	if (iDots == 3) // ip / udp format - Steamworks unlike BSD assume host ordered ip. Therefore just calling BSD conversion function will not work here.
	{

		// Find the port
		char* szPort = strchr(acAddressCopy, ':');
		if(!szPort)
		{
			*pusPort = 0;
		}
		else
		{
			int iPort = atoi(szPort + 1);
			if (iPort == 0 && szPort[1] != '0')
			{
				*pusPort = 0;
				return false;
			}

			if(iPort < 0 || iPort > 0xffff)
			{
				*pusPort = 0;
				return false;
			}
			else
			{
				*pusPort = static_cast<unsigned short>(iPort);
			}
			szPort[0] = 0;
		}

		int iPos =0;		
		for (int i = 0; i < 4; ++i) // go over the bytes
		{
			uint8 ucByte = 0;
			for (int j = 0; j < 3; ++j) // go over the digits in a byte
			{
				// either digit or dot - or if this the last number after it can be anything
				if (acAddressCopy[iPos] <= '9' && acAddressCopy[iPos] >= '0')
				{
					ucByte = ucByte * 10 + (acAddressCopy[iPos] - '0');
					++iPos;
				} 
				else if ((acAddressCopy[iPos] == '.' || i == 3) && j > 0)
				{
					j = 3;
				}
				else
				{
					*pusPort = 0;
					*pxAddress = Asura_NetAddress();
					return false;
				}

			}
			pxAddress->m_uAddress = (pxAddress->m_uAddress << 8)  + ucByte;
			if (i == 3)
			{
				// do nothing
			}
			else if (acAddressCopy[iPos] == '.') // skip the '.'
			{
				++iPos; 
			}
			else 
			{
				return false;
			}
		}
		return true;
	}
	else if (iDots == 7)
	{
		int iPos =0;
		uint64 ulId = 0;
		for (int i = 0; i < 8; ++i) // go over the bytes
		{
			uint8 ucByte = 0;
			for (int j = 0; j < 3; ++j) // go over the digits in a byte
			{
				if (szAddress[iPos] <= '9' && szAddress[iPos] >= '0')
				{
					ucByte = ucByte * 10 + (szAddress[iPos] - '0');
					++iPos;
				}
				else if ((szAddress[iPos] == '.' || i == 7 ) && j > 0) 
				{
					j = 3;
				}
				else
				{
					pxAddress->m_xSteamID.SetFromUint64(0);
					return false;
				}
			}
			ulId = (ulId << 8)  + ucByte;
			if (i == 7)
			{
				// do nothing
			}
			else if (acAddressCopy[iPos] == '.') // skip the '.'
			{
				pxAddress->m_xSteamID.SetFromUint64(0);
				++iPos; 
			}
			else 
			{
				return false;
			}
		}
		pxAddress->m_xSteamID.SetFromUint64(ulId);
		return true;
	}
	else
	{
		return false;
	}
}



bool Asura_PC_NetSocketSteamworks::Initialise()
{
	if( !Asura_PC_Steamworks::IsInitialised() )
	{
		// We need to ensure Asura_PC_Steamworks is initialised, but really it should be done before now
		Asura_PC_Steamworks::Initialise();

		if( !Asura_PC_Steamworks::IsInitialised() )
		{
			return false;
		}
	}
	
	return true;
}

void Asura_PC_NetSocketSteamworks::Update()
{
	
}

bool Asura_PC_NetSocketSteamworks::API_Bind(Asura_NetAddress xAddress, u_short &usPort)
{
	// bind is only used for server side listen operations. 
	// client under steam ignore it

	// just to bypass some checks
	m_usPort = usPort;
	
	if (m_pxImp->Bind(xAddress, usPort))
	{
		m_bConnected = true;
		return true;
	}
	else
	{
		return false;
	}
}

void Asura_PC_NetSocketSteamworks::API_Close()
{
	m_pxImp->Close();
}

bool Asura_PC_NetSocketSteamworks::API_ReceiveMessages()
{
	char pcBuffer[ASURA_RECEIVE_BUFFER_SIZE];
	Asura_NetAddress xSourceAddress;
	u_short usSourcePort;

	while (true)
	{
		int iBytesReceived = ASURA_RECEIVE_BUFFER_SIZE;
		if (m_pxImp->ReceiveMessage(pcBuffer, &iBytesReceived, &xSourceAddress, &usSourcePort)) 
		{
			if(!HandleMessage(pcBuffer, iBytesReceived, xSourceAddress, usSourcePort))
			{
				return false;
			}
		}
		else
		{
			return true;
		}
	}
}

bool Asura_PC_NetSocketSteamworks::API_SendTo(const Asura_NetSocket_MessageInfo* pxMessageInfo)
{
	return m_pxImp->SendTo(pxMessageInfo);
}

void Asura_PC_NetSocketSteamworks::API_SetBroadcastEnabled(bool bEnabled)
{
	// TY 22.04.09 - Don't assert here, its going to fire every time we use Asura_Network
	//Asura_Assert(false, "Broadcast is not supported in steamworks.");
	return;
}

int Asura_PC_NetSocketSteamworks::API_GetLastError()
{
	return m_pxImp->GetLastError();
}

void Asura_PC_NetSocketSteamworks::UpdateAddress( const Asura_NetAddress xAddress, const unsigned short usPort )
{
	m_xAddress = xAddress;
	m_usPort = usPort;
}


#endif // ASURA_PC_SUPPORT_STEAMWORKS


