///////////////////////////////////////////////////////////////////////////////
//
// File: Asura_PC_SteamworksQosSocket.cpp
//
// Created by Martin G.Harris (23/4/2009)
// 
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Network/Steamworks/Asura_PC_SteamworksQosSocket.h"


#ifdef ASURA_ENABLE_QOS_SYSTEM

// Includes
#include "Core/Asura_Multithread.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Steamworks/Asura_PC_SteamworksServer.h"


// Size of header for QoS, UDP, IP etc.
// QoS: 8 bytes for password, msg id, packet id.
const u_int uASURA_QOS_DATA_HEADER_LENGTH = sizeof(Asura_Qos_MessageHeader); // 8
const int   iASURA_QOS_DATA_HEADER_LENGTH = static_cast<int>(uASURA_QOS_DATA_HEADER_LENGTH);

// Timeout for steamworks connections.
const int iASURA_PC_STEAM_QOS_CONNECT_TIMEOUT = 15;


///////////////////////////////////////////////////////////////////////////////
// Asura_QosSocket_Steamworks
///////////////////////////////////////////////////////////////////////////////
Asura_QosSocket_Steamworks::Asura_QosSocket_Steamworks()
: m_usLocalPort(0)
{
}

Asura_QosSocket_Steamworks::~Asura_QosSocket_Steamworks()
{
}

///////////////////////////////////////////////////////////////////////////////
// Asura_QosSocket_Steamworks_Server
///////////////////////////////////////////////////////////////////////////////
Asura_QosSocket_Steamworks_Server::Asura_QosSocket_Steamworks_Server()
: m_xListenSocket(0)
//, m_pfnCallbackSocketStatus(this, &Asura_QosSocket_Steamworks_Server::OnSocketStatusCallback)
{
}

Asura_QosSocket_Steamworks_Server::~Asura_QosSocket_Steamworks_Server()
{
	Asura_Assert(!m_xListenSocket, "Socket still open!");
}

//////////////////
// CreateSocket //
//////////////////
void Asura_QosSocket_Steamworks_Server::CreateSocket(Asura_Qos_Listener* pxListener, u_short usLocalPort)
{
	Asura_Assert(!m_bCreatedSocket, "Socket has already been created!");
	if (m_bCreatedSocket) return;

	m_usLocalPort = usLocalPort;

	if (!SteamNetworking())
	{
		Asura_Assert(0, "Steam networking API is null");
		return;
	}

	const u_int uAddress = 0; // Use the default local IP address.

	// Create a socket for listening for incoming connection requests.
	// Specify a vport (different from that used by the lobby) in case we wish to use a P2P connection.
	m_xListenSocket = SteamNetworking()->CreateListenSocket( ASURA_PC_STEAMWORKS_QOSSOCKET_VPORT, uAddress, usLocalPort, true ); 

	if ( !m_xListenSocket )
	{
		return;
	}

	m_bCreatedSocket = true;

#ifdef ASURA_QOS_ENABLE_LOGGING
	// Find the socket information
	uint32 uIP		= 0;
	uint16 usPort	= 0;

	if( SteamNetworking()->GetListenSocketInfo( m_xListenSocket, &uIP, &usPort ) )
	{
		in_addr xAddr;
		xAddr.s_addr = htonl(uIP);

		Asura_OutputToDebugger_Retail(("Qos: [Steam] Created listen socket %s:%u\n", inet_ntoa(xAddr), usPort));
	}
#endif //#ifdef ASURA_QOS_ENABLE_LOGGING
}

/////////////////
// CloseSocket //
/////////////////
void Asura_QosSocket_Steamworks_Server::CloseSocket()
{
	if (!m_bCreatedSocket) return;
	
	m_bCreatedSocket = false;

	if (SteamNetworking())
	{
		SteamNetworking()->DestroyListenSocket(m_xListenSocket, false);
	}
	else
	{
		Asura_Assert(0, "Steam networking API is null");
	}

	m_xListenSocket = 0;
}

////////////////////////////
// OnSocketStatusCallback //
////////////////////////////
//void Asura_QosSocket_Steamworks_Server::OnSocketStatusCallback( SocketStatusCallback_t *pxCallback )
//{
//	// If it's not for us then return.
//	if ( pxCallback->m_hListenSocket != m_xListenSocket)
//	{
//		return;
//	}
//
//	// TODO if necessary deal with the socket status.
//}

////////////
// SendTo //
////////////
// Return true if succeeded.
bool Asura_QosSocket_Steamworks_Server::SendTo(const Asura_Qos_NetConnection* pxDestination, const char* pcData, int iDataLength)
{
	Asura_Assert(pxDestination, "Null Qos destination");
	if (!pxDestination) return false;

	Asura_Assert(pcData, "Null Qos message data");
	if (!pcData) return false;

	Asura_Assertf(iDataLength>0, ("Invalid message data length (%d)", iDataLength));
	if (iDataLength <= 0) return false;

	if (!SteamNetworking())
	{
		Asura_Assert(0, "Steam networking API is null");
		return false;
	}

	if (pxDestination->GetConnectionInfo().m_xSocket == 0)
	{
		Asura_Assert(0, "SendTo: Invalid destination steam socket");
		return false;
	}
	
	// Send data to the connection socket.
	const uint32 uDataLength = static_cast<uint32>(iDataLength);
	if ( !SteamNetworking()->SendDataOnSocket( pxDestination->GetConnectionInfo().m_xSocket, const_cast<char*>(pcData), uDataLength, false ) )
	{
		// Failed for some reason.
		Asura_Assert(0, "Error sending QoS message data to socket.");
		return false;
	}
	
	return true;
}


static Asura_Qos_Message_Padded g_xBuffer; // Static buffer for receiving messages in.

/////////////
// Receive //
/////////////
// Note: Ignore uWaitTimeSeconds, as it is currently unused on the server.
bool Asura_QosSocket_Steamworks_Server::Receive(Asura_Qos_Listener* pxListener, u_int uWaitTimeSeconds /*= 0*/)
{
	Asura_Assert(pxListener, "Null Qos listener");
	if (!pxListener) return false;

	if (!m_bCreatedSocket) return false;

	if (!SteamNetworking())
	{
		Asura_Assert(0, "Null steam networking API");
		return false;
	}

	char* const pcBuffer = reinterpret_cast<char*>(&g_xBuffer);
	
	uint32			uRecvSize = 0;
	SNetSocket_t	xSourceSocket = 0;

	while( SteamNetworking()->RetrieveData( m_xListenSocket, pcBuffer, ASURA_QOS_RECEIVE_BUFFER_SIZE, &uRecvSize, &xSourceSocket ) )
	{
		const int iBytesReceived = static_cast<int>(uRecvSize);
		if (iBytesReceived < iASURA_QOS_DATA_HEADER_LENGTH)
		{
			Asura_Assert(iBytesReceived < iASURA_QOS_DATA_HEADER_LENGTH, "Too few bytes received.");
			return false;
		}

		if (iBytesReceived > ASURA_QOS_RECEIVE_BUFFER_SIZE)
		{
			Asura_Assert(iBytesReceived <= ASURA_QOS_RECEIVE_BUFFER_SIZE , "Too many bytes received, overflowed buffer.");
			return false;
		}

		const Asura_Qos_Message* pxMessage = &g_xBuffer;
		if (pxMessage->IsPasswordValid())
		{
			Asura_NetAddress xFromAddress;
			uint16 usRemotePort = 0;

			if ( SteamNetworking()->GetSocketInfo(xSourceSocket, &xFromAddress.m_xSteamID, NULL, &xFromAddress.m_uAddress, &usRemotePort ))
			{		
				Asura_Qos_NetConnectionInfo xFromInfo(xFromAddress, usRemotePort);
				xFromInfo.m_xSocket = xSourceSocket;

				pxListener->HandleMessageFrom(xFromInfo, pxMessage);
			}
			else
			{
				Asura_Assert(0, "GetSocketInfo failed");
				return false;
			}
		}

	}

	return true; // No errors
}

///////////////////////////////////////////////////////////////////////////////
// Asura_QosSocket_Steamworks_Client
///////////////////////////////////////////////////////////////////////////////
Asura_QosSocket_Steamworks_Client::Asura_QosSocket_Steamworks_Client()
: m_xSocket(0)
, m_xRemoteAddress(0)
, m_usRemotePort(0)
//, m_pfnCallbackSocketStatus(this, &Asura_QosSocket_Steamworks_Client::OnSocketStatusCallback)
{
}

Asura_QosSocket_Steamworks_Client::~Asura_QosSocket_Steamworks_Client()
{
	Asura_Assert(!m_xSocket, "Socket still open!");
}

//////////////////
// CreateSocket //
//////////////////
void Asura_QosSocket_Steamworks_Client::CreateSocket(Asura_Qos_Listener* pxListener, u_short usLocalPort)
{
	Asura_Assert(!m_bCreatedSocket, "Socket has already been created!");
	if (m_bCreatedSocket) return;

	m_usLocalPort = usLocalPort; // Not used by Steamworks.

	if (!SteamNetworking())
	{
		Asura_Assert(0, "Steam networking API is null");
		return;
	}

	Asura_Assert(pxListener, "Null listener");
	if (!pxListener) return;
	
	// Zeroth connection is the server for clients at least.
	const Asura_Qos_NetConnection* const pxServerConnection = pxListener->GetServerConnection();
	
	Asura_Assert(pxServerConnection, "Null server connection");
	if (!pxServerConnection) return;

	m_xRemoteAddress = pxServerConnection->GetRemoteAddress();	// Remote address IPv4 (host byte order)
	m_usRemotePort	 = pxServerConnection->GetRemotePort();		// Remote port (host byte order)

	const int iVPort = ASURA_PC_STEAMWORKS_QOSSOCKET_VPORT;
	
	const bool bUseP2P = ( m_xRemoteAddress.m_xSteamID.IsValid() );	
	if ( bUseP2P )
	{
		m_xSocket = SteamNetworking()->CreateP2PConnectionSocket( m_xRemoteAddress.m_xSteamID, iVPort, iASURA_PC_STEAM_QOS_CONNECT_TIMEOUT, true );
	}
	else
	{
		m_xSocket = SteamNetworking()->CreateConnectionSocket( m_xRemoteAddress.m_uAddress, m_usRemotePort, iASURA_PC_STEAM_QOS_CONNECT_TIMEOUT );
	}

	if ( !m_xSocket )
	{
		return;
	}

	m_bCreatedSocket = true;

#ifdef ASURA_QOS_ENABLE_LOGGING
	if (bUseP2P)
	{
		Asura_OutputToDebugger_Retail( ( "Qos: [Steam] Created P2P connection socket (#%u) on vport %d\n", m_xSocket, iVPort ) );
	}
	else
	{
		Asura_OutputToDebugger_Retail( ( "Qos: [Steam] Created connection socket (#%u) to remote port %u\n", m_xSocket, m_usRemotePort ) );
	}
#endif  //ASURA_QOS_ENABLE_LOGGING
}

/////////////////
// CloseSocket //
/////////////////
void Asura_QosSocket_Steamworks_Client::CloseSocket()
{
	if (!m_bCreatedSocket) return;
	
	m_bCreatedSocket = false;

	if (SteamNetworking())
	{
		SteamNetworking()->DestroySocket(m_xSocket, false);
	}
	else
	{
		Asura_Assert(0, "Steam networking API is null");
	}

	m_xSocket = 0;
}

////////////////////////////
// OnSocketStatusCallback //
////////////////////////////
//void Asura_QosSocket_Steamworks_Client::OnSocketStatusCallback( SocketStatusCallback_t *pxCallback )
//{
//	// If it's not for us then return.
//	if ( pxCallback->m_hListenSocket != m_xSocket)
//	{
//		return;
//	}
//
//	// TODO if necessary deal with the socket status.
//}

/////////////
// IsReady //
/////////////
// Return true if any low-level handshaking is finished.
bool Asura_QosSocket_Steamworks_Client::IsReady() const
{
	if (!SteamNetworking())
	{
		Asura_Assert(0, "Steam networking API is null");
		return false;
	}

	if (!m_bCreatedSocket) return false;

	// Check the socket status.
	int iStatus = 0;
	if (SteamNetworking()->GetSocketInfo(m_xSocket, 0, &iStatus, 0, 0 ))
	{
		return (iStatus == k_ESNetSocketStateConnected);
	}

	// Failed for some reason.
	return false;
}

////////////
// SendTo //
////////////
// Return true if succeeded.
bool Asura_QosSocket_Steamworks_Client::SendTo(const Asura_Qos_NetConnection* pxDestination, const char* pcData, int iDataLength)
{
	Asura_Assert(pxDestination, "Null Qos destination");
	if (!pxDestination) return false;

	Asura_Assert(pcData, "Null Qos message data");
	if (!pcData) return false;

	Asura_Assertf(iDataLength>0, ("Invalid message data length (%d)", iDataLength));
	if (iDataLength <= 0) return false;

	if (!SteamNetworking())
	{
		Asura_Assert(0, "Steam networking API is null");
		return false;
	}

	// Destination must match that specified when we created the connection socket.
	if ( !(pxDestination->GetRemoteAddress() == m_xRemoteAddress) )
	{
		Asura_Assert(0, "SendTo: Invalid destination address!");
		return false;
	}

	if ( !(pxDestination->GetRemotePort() == m_usRemotePort) )
	{
		Asura_Assert(0, "SendTo: Invalid destination port!");
		return false;
	}

	// Send data to the connection socket.
	const uint32 uDataLength = static_cast<uint32>(iDataLength);
	if ( !SteamNetworking()->SendDataOnSocket( m_xSocket, const_cast<char*>(pcData), uDataLength, false ) )
	{
		// Failed for some reason.
		int iStatus = 0;
		if (SteamNetworking()->GetSocketInfo(m_xSocket, 0, &iStatus, 0, 0 ))
		{
			// This can happen if the client sends a server connection request before the
			// socket has finished initialising. There is a delay in the code to give it
			// time to intialise, but that may not be long enough.
			if (iStatus == k_ESNetSocketStateConnected ||
				iStatus == k_ESNetSocketStateInitiated ||
				iStatus == k_ESNetSocketStateLocalCandidatesFound ||
				iStatus == k_ESNetSocketStateReceivedRemoteCandidates ||
				iStatus == k_ESNetSocketStateChallengeHandshake)
			{
				// Just log this and hope that the socket finishes initialising soon.
				#ifdef ASURA_QOS_ENABLE_LOGGING
				Asura_OutputToDebugger(("Qos: Socket #%u has status %d (initialising). SendTo failed.\n", m_xSocket, iStatus));
				#endif //ASURA_QOS_ENABLE_LOGGING
				
				// Don't stop the client. Carry on trying to connect.
				return true;
			}
		}

		Asura_Assertf(0, ("Error sending QoS message data to socket #%u (status %d).", m_xSocket, iStatus));
		if (pxDestination->IsNullConnection())
		{
			// Not connected yet. Allow the client to retry to make the connection.
			return true;
		}

		return false;
	}
	
	return true;
}

/////////////
// Receive //
/////////////
// Check the socket for messages. If a time is specified then wait for that amount of time, or until a message arrives.
bool Asura_QosSocket_Steamworks_Client::Receive(Asura_Qos_Listener* pxListener, u_int uWaitTimeSeconds /*= 0*/)
{
	Asura_Assert(pxListener, "Null Qos listener");
	if (!pxListener) return false;

	if (!m_bCreatedSocket) return false;

	if (!SteamNetworking())
	{
		Asura_Assert(0, "Null steam networking API");
		return false;
	}

	char* const pcBuffer = reinterpret_cast<char*>(&g_xBuffer);
	
	uint32 uRecvSize = 0;

	// If necessary wait for a certain amount of time.
	Asura_TickCount xTickStart;
	if (uWaitTimeSeconds > 0)
	{
		xTickStart = Asura_Timers::GetCurTicks();
	}
	
	bool bFinishedWaiting = true;

	do
	{
		while( SteamNetworking()->RetrieveDataFromSocket( m_xSocket, pcBuffer, ASURA_QOS_RECEIVE_BUFFER_SIZE, &uRecvSize ) )
		{
			const int iBytesReceived = static_cast<int>(uRecvSize);
			if (iBytesReceived < iASURA_QOS_DATA_HEADER_LENGTH)
			{
				Asura_Assert(iBytesReceived < iASURA_QOS_DATA_HEADER_LENGTH, "Too few bytes received.");
				return false;
			}

			if (iBytesReceived > ASURA_QOS_RECEIVE_BUFFER_SIZE)
			{
				Asura_Assert(iBytesReceived <= ASURA_QOS_RECEIVE_BUFFER_SIZE , "Too many bytes received, overflowed buffer.");
				return false;
			}

			const Asura_Qos_Message* pxMessage = &g_xBuffer;
			if (pxMessage->IsPasswordValid())
			{
				Asura_Qos_NetConnectionInfo xRemoteInfo(m_xRemoteAddress, m_usRemotePort);
				xRemoteInfo.m_xSocket = m_xSocket;

				pxListener->HandleMessageFrom(xRemoteInfo, pxMessage);

				// We've received a message so we can cancel the waiting timer.
				uWaitTimeSeconds = 0;
				bFinishedWaiting = true;
			}
		}

		if (uWaitTimeSeconds > 0)
		{
			const float fElapsedSeconds = Asura_Timers::TicksToSeconds(Asura_Timers::GetElapsedTicks(xTickStart));
			bFinishedWaiting = (fElapsedSeconds > static_cast<float>(uWaitTimeSeconds));
			if (!bFinishedWaiting)
			{
				// Wait for a short time and then check for messages again.
				Asura_Thread_System::DelayCurrentThread(fASURA_QOS_THREAD_SLEEP_TIME);
			}
		}
	}
	while (!bFinishedWaiting);

	return true; // No errors
}

#endif //#ifdef ASURA_ENABLE_QOS_SYSTEM

