
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_SCEUtility_FileData.cpp
// Author : Andy Tate
// Date   : 6-Oct-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////


#include "Core/Asura_MemoryManagement_DebugOff.h"
#include <atltime.h>
#include "Core/Asura_MemoryManagement_DebugOn.h"

#include "Core/Asura.h"

// This file's header...
#include "PSPEmul/SCEUtility/Asura_PSP_SCEUtility_FileData.h"

// Asura independent header-files...

// Asura platform header-files...
#include "FileAccess/Asura_DebugFileStream.h"

// Game independent header-files...

// Game platform header-files...


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::CreateDirectoryRecursive
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Creates the directory for the current file.
//
// Return value : "true" if successfully created OR it already exists,
//				  "false" if something's gone wrong.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SCEUtility_FileData::CreateDirectoryRecursive (const Asura_String& xFullPath)
{
	Asura_String	xDirectory;
	Asura_String	xFilename;
	Asura_String	xExtension;

	// Can we split this filename?
	if (xFullPath.SplitFilename(&xDirectory, &xFilename, &xExtension))
	{
		// Have we run out of directories?
		if (xDirectory.IsNull())
		{
			// Yes - create the directory using the filename...

			CreateDirectory	(xFilename.GetString(), 0);
			return			(true);
		}

		else
		{
			// Yes - recurse to the next directory...
			if (CreateDirectoryRecursive(xDirectory))
			{
				// Success - create the directory passed-in...
				CreateDirectory	(xFullPath.GetString(), 0);
				return			(true);
			}

			else
			{
				// Failure - report this...
				return	(false);
			}
		}
	}

	else
	{
		// No - create this directory...
		Asura_Assert	(false, "Failed to split filename into it's components.");
		return			(false);
	}
}

/// End of function 'Asura_PSP_SCEUtility_FileData::CreateDirectoryRecursive'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::CreateDirectoryFromFilename
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Creates the directory for the current file.
//
// Return value : "true" if successfully created OR it already exists,
//				  "false" if something's gone wrong.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SCEUtility_FileData::CreateDirectoryFromFilename (const Asura_String& xFullPath)
{
	Asura_String	xDirectory;
	Asura_String	xFilename;
	Asura_String	xExtension;

	// Can we split this filename?
	if (xFullPath.SplitFilename(&xDirectory, &xFilename, &xExtension))
	{
		// Yes - recurse to the next directory...
		return	(CreateDirectoryRecursive(xDirectory));
	}

	else
	{
		// No - return failure...
		Asura_Assert	(false, "Failed to split filename into it's components.");
		return			(false);
	}
}

/// End of function 'Asura_PSP_SCEUtility_FileData::CreateDirectoryFromFilename'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::ReadHeaderData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Reads the header data for the current file.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_SCEUtility_FileData::ReadHeaderData (Asura_Stream& xStream)
{
	int	iTitleLength;
	int	iDescriptionLength;
	int	iDateAndTimeLength;

	// Read the string lengths...
	xStream >> iTitleLength;
	xStream >> iDescriptionLength;
	xStream >> iDateAndTimeLength;

	// Delete the allocated strings...
	delete []	(m_wszTitle);
	delete []	(m_wszDescription);
	delete []	(m_szDateAndTime);

	// Allocate the strings...
	m_wszTitle			= new Asura_UnicodeChar [iTitleLength];
	m_wszDescription	= new Asura_UnicodeChar [iDescriptionLength];
	m_szDateAndTime		= new Asura_Char [iDateAndTimeLength];

	// Read the string itself...
	xStream.ReadData	(iTitleLength * sizeof(Asura_UnicodeChar),			m_wszTitle);
	xStream.ReadData	(iDescriptionLength * sizeof(Asura_UnicodeChar),	m_wszDescription);
	xStream.ReadData	(iDateAndTimeLength * sizeof(Asura_Char),			m_szDateAndTime);

	// Finally, read the data-size...
	xStream >> m_uDataSize;
}

/// End of function 'Asura_PSP_SCEUtility_FileData::ReadHeaderData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::WriteHeaderData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Writes the header data for the current file.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_SCEUtility_FileData::WriteHeaderData (Asura_Stream& xStream,
													 const u_int uDataSize) const
{
	int	iTitleLength		= wcslen (m_wszTitle) + 1;
	int	iDescriptionLength	= wcslen (m_wszDescription) + 1;
	int	iDateAndTimeLength	= strlen (m_szDateAndTime) + 1;

	// Write the string lengths...
	xStream << iTitleLength;
	xStream << iDescriptionLength;
	xStream << iDateAndTimeLength;

	// Write the string itself...
	xStream.WriteData	(iTitleLength * sizeof(Asura_UnicodeChar),			m_wszTitle);
	xStream.WriteData	(iDescriptionLength * sizeof(Asura_UnicodeChar),	m_wszDescription);
	xStream.WriteData	(iDateAndTimeLength * sizeof(Asura_Char),			m_szDateAndTime);

	// Finally, read the data-size...
	xStream << m_uDataSize;
}

/// End of function 'Asura_PSP_SCEUtility_FileData::WriteHeaderData'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::Asura_PSP_SCEUtility_FileData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Constructor function for the SCE Utility file-data.
//
// Parameters   : The filename to be used.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_PSP_SCEUtility_FileData::Asura_PSP_SCEUtility_FileData (const char* szFilename) :
	m_xFilename(szFilename),
	m_szDateAndTime(NULL),
	m_wszTitle(NULL),
	m_wszDescription(NULL),
	m_uDataSize(0)
{
	// Nothing more to initialise...
}

/// End of function 'Asura_PSP_SCEUtility_FileData::Asura_PSP_SCEUtility_FileData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::~Asura_PSP_SCEUtility_FileData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Destructor function for the SCE Utility file-data.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_PSP_SCEUtility_FileData::~Asura_PSP_SCEUtility_FileData ()
{
	// Delete the allocated strings...
	delete []	(m_wszTitle);
	delete []	(m_wszDescription);
	delete []	(m_szDateAndTime);
}

/// End of function 'Asura_PSP_SCEUtility_FileData::~Asura_PSP_SCEUtility_FileData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::UpdateHeaderData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Update's the header-data with the current data.
//
// Parameters   : The title and description strings to use.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_SCEUtility_FileData::UpdateHeaderData (const Asura_UnicodeChar* wszTitle,
													  const Asura_UnicodeChar* wszDescription,
													  const u_int uDataSize)
{
	CTime	xTime				= CTime::GetCurrentTime ();
	CString	xDateAndTimeString	= xTime.Format ("%H:%M:%S %d-%b-%Y");

	int		iTitleLength		= wcslen (wszTitle) + 1;
	int		iDescriptionLength	= wcslen (wszDescription) + 1;
	int		iDateAndTimeLength	= xDateAndTimeString.GetLength () + 1;

	// Delete the allocated strings...
	delete []	(m_wszTitle);
	delete []	(m_wszDescription);
	delete []	(m_szDateAndTime);

	// Allocate the strings...
	m_wszTitle			= new Asura_UnicodeChar [iTitleLength];
	m_wszDescription	= new Asura_UnicodeChar [iDescriptionLength];
	m_szDateAndTime		= new Asura_Char [iDateAndTimeLength];

	// Copy the old strings...
	wcscpy	(m_wszTitle,		wszTitle);
	wcscpy	(m_wszDescription,	wszDescription);
	strcpy	(m_szDateAndTime,	xDateAndTimeString.GetString());

	// Set the data size...
	m_uDataSize	= uDataSize;
}

/// End of function 'Asura_PSP_SCEUtility_FileData::UpdateHeaderData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::ReadHeaderData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Reads the header data for the current file.
//
// Return value : "true" if read successfully, "false" if failed to read.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SCEUtility_FileData::ReadHeaderData (void)
{
	Asura_DebugFileStream	xStream;

	// Can we open this file?
	if (xStream.OpenForReading(m_xFilename.GetString(), true))
	{
		// Yes - read the header data...
		ReadHeaderData	(xStream);
		return			(true);
	}

	else
	{
		// No - return failure...
		return	(false);
	}
}

/// End of function 'Asura_PSP_SCEUtility_FileData::ReadHeaderData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::ReadData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Reads the data (including header) for the current file.
//
// Parameters   : A reference to the data-size read and a pointer to the location
//				  in memory to read the data to.
//
// Return value : "true" if read successfully, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SCEUtility_FileData::ReadData (u_int& uDataSize,
											  void* pvData)
{
	Asura_DebugFileStream	xStream;

	// Can we open this file?
	if (xStream.OpenForReading(m_xFilename.GetString(), true))
	{
		// Yes - read the header data...
		ReadHeaderData	(xStream);

		// Read the data itself...
		uDataSize	= xStream.ReadData (m_uDataSize, pvData);

		// Return success...
		return	(true);
	}

	else
	{
		// No - return failure...
		return	(false);
	}
}

/// End of function 'Asura_PSP_SCEUtility_FileData::ReadData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SCEUtility_FileData::WriteData
// Author       : Andy Tate
// Date         : 6-Oct-2005
//
// Description  : Writes the data (including header) for the current file.
//
// Parameters   : A reference to the data-size to be written and a pointer to
//				  the location in memory to write the data from.
//
// Return value : "true" if written successfully, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SCEUtility_FileData::WriteData (const u_int uDataSize,
											   const void* pvData) const
{
	Asura_DebugFileStream	xStream;

	// Ensure that the directory exists...
	CreateDirectoryFromFilename	(m_xFilename);

	// Can we open this file?
	if (xStream.OpenForWriting(m_xFilename.GetString(), true))
	{
		// Yes - read the header data...
		WriteHeaderData	(xStream, uDataSize);

		// Write the data itself...
		xStream.WriteData (uDataSize, pvData);

		// Return success...
		return	(true);
	}

	else
	{
		// No - return failure...
		return	(false);
	}
}

/// End of function 'Asura_PSP_SCEUtility_FileData::WriteData'. ///


/// End of file. ///
