//
// Asura_Direct3D_BackBufferCapture.cpp
//
// Created by RJF - 14 September 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Direct3D_BackBufferCapture.h"
#include "Rendering/Asura_Render.h"

// Static class variables

Asura_Direct3D_BackBufferCapture::BackBufferData Asura_Direct3D_BackBufferCapture::LDR = 
{
	NULL,
	NULL,
	0,
	0, 
	true,
	false
};
#if defined ASURA_USE_HDR
Asura_Direct3D_BackBufferCapture::BackBufferData Asura_Direct3D_BackBufferCapture::HDR = 
{
	NULL,
	NULL,
	0,
	0, 
	true,
	false
};
#endif

// Class implementation

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::Initialise
//
// Parameters:    None
// Return value:  boolean to indicate success
//
// Creates the texture and gets the top surface level
//////////////////////////////////////////////////////
bool Asura_Direct3D_BackBufferCapture::Initialise()
{
	HRESULT hr;

	// Make sure that any resources have been wiped
	Shutdown();

	Asura_Assert(Asura_Direct3D::D3DDevice, "Tried to create the BackBufferCapture when there's no D3DDevice!");

	LDR.m_uCaptureWidth = Asura_Direct3D::GetPresentParameters().BackBufferWidth;
	LDR.m_uCaptureHeight = Asura_Direct3D::GetPresentParameters().BackBufferHeight;
#if defined ASURA_USE_HDR
	HDR.m_uCaptureWidth = Asura_Direct3D::GetPresentParameters().BackBufferWidth;
	HDR.m_uCaptureHeight = Asura_Direct3D::GetPresentParameters().BackBufferHeight;
#endif

	D3DFORMAT uFmt = Asura_Direct3D::GetPresentParameters().BackBufferFormat;

	// DMR 29/05/07 - We need a format with an alpha channel, so we can capture that too. This fixes it for
	// my Geforce8800, but we'll probably need to handle more cases here eventually?
	if (uFmt == D3DFMT_X8R8G8B8)
	{
		uFmt = D3DFMT_A8R8G8B8;
	}

	hr = Asura_Direct3D::D3DDevice->CreateTexture(LDR.m_uCaptureWidth, LDR.m_uCaptureHeight, 1, D3DUSAGE_RENDERTARGET, uFmt, D3DPOOL_DEFAULT, &LDR.m_pxCaptureTexture, NULL);
	Asura_Assert(SUCCEEDED(hr), "Failed to create BackBufferCapture texture");
	if(FAILED(hr))return false;

	hr = LDR.m_pxCaptureTexture->GetSurfaceLevel(0, &LDR.m_pxCaptureSurface);
	Asura_Assert(SUCCEEDED(hr), "Failed to get surface level 0 of the BackBufferCapture texture");
	if(FAILED(hr))return false;

	LDR.m_bIsDirty = true;

#if defined ASURA_USE_HDR
	uFmt = D3DFMT_A16B16G16R16F;

	hr = Asura_Direct3D::D3DDevice->CreateTexture(HDR.m_uCaptureWidth, HDR.m_uCaptureHeight, 1, D3DUSAGE_RENDERTARGET, uFmt, D3DPOOL_DEFAULT, &HDR.m_pxCaptureTexture, NULL);
	Asura_Assert(SUCCEEDED(hr), "Failed to create BackBufferCapture texture");
	if(FAILED(hr))return false;

	hr = HDR.m_pxCaptureTexture->GetSurfaceLevel(0, &HDR.m_pxCaptureSurface);
	Asura_Assert(SUCCEEDED(hr), "Failed to get surface level 0 of the BackBufferCapture texture");
	if(FAILED(hr))return false;

	HDR.m_bIsDirty = true;
#endif

	return true;
}

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::Shutdown
//
// Parameters:    None
// Return value:  None
//
// Frees the surface level and texture used to capture the back buffer
//////////////////////////////////////////////////////
void Asura_Direct3D_BackBufferCapture::Shutdown()
{
	if ( LDR.m_pxCaptureSurface )
	{
		LDR.m_pxCaptureSurface->Release();
		LDR.m_pxCaptureSurface = NULL;
	}

	if ( LDR.m_pxCaptureTexture )
	{
		LDR.m_pxCaptureTexture->Release();
		LDR.m_pxCaptureTexture = NULL;
	}
#if defined ASURA_USE_HDR
	if ( HDR.m_pxCaptureSurface )
	{
		HDR.m_pxCaptureSurface->Release();
		HDR.m_pxCaptureSurface = NULL;
	}

	if ( HDR.m_pxCaptureTexture )
	{
		HDR.m_pxCaptureTexture->Release();
		HDR.m_pxCaptureTexture = NULL;
	}
#endif
}

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::CaptureCurrentBackBuffer
//
// Parameters:    None
// Return value:  None
//
// Updates the texture to match the screen res if necessary, and then takes a copy of the backbuffer
//////////////////////////////////////////////////////
void Asura_Direct3D_BackBufferCapture::CaptureCurrentBackBuffer()
{
	SCOPEDPIXEVENT(( 0xFFFFFFFF, L"Capture current back buffer" ));
#if defined ASURA_USE_HDR
	BackBufferData &xData = Asura_Render::IsCurrentlyHDR() ? HDR : LDR;
#else
	BackBufferData &xData = LDR;
#endif


	HRESULT hr;

	xData.m_bIsDirty = true;

	if ( ( Asura_Direct3D::GetPresentParameters().BackBufferWidth != xData.m_uCaptureWidth ) 
	  || ( Asura_Direct3D::GetPresentParameters().BackBufferHeight != xData.m_uCaptureHeight ) 
	  || ( !xData.m_pxCaptureTexture )
	  || ( !xData.m_pxCaptureSurface ) )
	{
		if(!Initialise())return; //don't have a working texture+surface any more
	}

	// Get an interface for the current backbuffer surface
	Direct3D_Surface* pxBackBuffer = NULL;
	hr = Asura_Direct3D::D3DDevice->GetRenderTarget(0, &pxBackBuffer);
	ASURA_D3D_OK( hr);
	if( FAILED(hr) || !pxBackBuffer )return;

#if defined( ASURA_ASSERT)
	D3DSURFACE_DESC xBackBufferDesc;
	D3DSURFACE_DESC xCaptureSurfaceDesc;
	
	hr = pxBackBuffer->GetDesc( &xBackBufferDesc);
	ASURA_D3D_OK( hr);
	hr = xData.m_pxCaptureSurface->GetDesc( &xCaptureSurfaceDesc);
	ASURA_D3D_OK( hr);
	Asura_Assertf( xData.m_uCaptureWidth <= xBackBufferDesc.Width, ("Trying to StretchRect Width %d from Width %d buffer", xData.m_uCaptureWidth, xBackBufferDesc.Width));
	Asura_Assertf( xData.m_uCaptureHeight <= xBackBufferDesc.Height, ("Trying to StretchRect Height %d from Height %d buffer", xData.m_uCaptureHeight, xBackBufferDesc.Height));
	
	Asura_Assertf( xData.m_uCaptureWidth <= xCaptureSurfaceDesc.Width, ("Trying to StretchRect Width %d from Width %d buffer", xData.m_uCaptureWidth, xCaptureSurfaceDesc.Width));
	Asura_Assertf( xData.m_uCaptureHeight <= xCaptureSurfaceDesc.Height, ("Trying to StretchRect Height %d from Height %d buffer", xData.m_uCaptureHeight, xCaptureSurfaceDesc.Height));
#endif

	RECT xRectSrc = { 0, 0, xData.m_uCaptureWidth, xData.m_uCaptureHeight};
	RECT xRectDst = { 0, 0, xData.m_uCaptureWidth, xData.m_uCaptureHeight};
	hr = Asura_Direct3D::D3DDevice->StretchRect(pxBackBuffer, &xRectSrc, xData.m_pxCaptureSurface, &xRectDst, D3DTEXF_NONE);
	ASURA_D3D_OK( hr);
	Asura_Render::ApplyCameraSettings();
	
	pxBackBuffer->Release();
	ASURA_D3D_OK( hr);
	if( FAILED(hr) )return;

	xData.m_bHasCapturedThisFrame = true;

	xData.m_bIsDirty = false;
}

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::GetCaptureTexture()
//
// Parameters:    None
// Return value:  Texture
//
// Returns the backbuffer texture
//////////////////////////////////////////////////////
Direct3D_Texture* Asura_Direct3D_BackBufferCapture::GetCaptureTexture()
{
#if defined ASURA_USE_HDR
	BackBufferData &xData = Asura_Render::IsCurrentlyHDR() ? HDR : LDR;
#else
	BackBufferData &xData = LDR;
#endif
	return xData.m_pxCaptureTexture;
}

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::IsCaptureDirty()
//
// Parameters:    None
// Return value:  Flag to indicate dirty
//
// Returns the state of the dirty flag
//////////////////////////////////////////////////////
bool Asura_Direct3D_BackBufferCapture::IsCaptureDirty()
{
#if defined ASURA_USE_HDR
	BackBufferData &xData = Asura_Render::IsCurrentlyHDR() ? HDR : LDR;
#else
	BackBufferData &xData = LDR;
#endif
	return xData.m_bIsDirty;
}

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::HasCapturedThisFrame()
//
// Parameters:    None
// Return value:  boolean
//
// Returns a value indicating whether CaptureCurrentBackBuffer() has been called this frame
//////////////////////////////////////////////////////
bool Asura_Direct3D_BackBufferCapture::HasCapturedThisFrame()
{
#if defined ASURA_USE_HDR
	BackBufferData &xData = Asura_Render::IsCurrentlyHDR() ? HDR : LDR;
#else
	BackBufferData &xData = LDR;
#endif
	return xData.m_bHasCapturedThisFrame;
}

//////////////////////////////////////////////////////
// Asura_Direct3D_BackBufferCapture::AdviseNextFrame()
//
// Parameters:    None
// Return value:  None
//
// Called to tell the BackBufferCapture system that the next frame has begun
//////////////////////////////////////////////////////
void Asura_Direct3D_BackBufferCapture::AdviseNextFrame()
{
	LDR.m_bHasCapturedThisFrame = false;
#if defined ASURA_USE_HDR
	HDR.m_bHasCapturedThisFrame = false;
#endif
}
