///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Direct3D_Effect.cpp
// Author : Andy Tate
// Date   : 11-May-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Rendering/Asura_Direct3D_Effect.h"

// Asura header-files...
#include "Animation/Asura_TextureAnimation.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Light/Asura_Photon.h"
#include "Rendering/Asura_Direct3D_Caps.h"
#include "Rendering/Asura_Render.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

// Flags used when creating/compiling Direct3D effects
static const DWORD	dwEFFECT_CREATION_FLAGS	= 0
#if defined(ASURA_DIRECT3D_DEBUG_VS) || defined(ASURA_DIRECT3D_DEBUG_PS)
											  // turn on shader debugging
											  | D3DXSHADER_DEBUG | D3DXSHADER_SKIPOPTIMIZATION
#endif
											  ;

// Direct3D macros...

static const D3DXMACRO axDirect3DMacros [] =
{

#if defined(ASURA_DIRECT3D_DEBUG_VS)
	{"DEBUG_VS",						"1"},
#endif

#if defined(ASURA_DIRECT3D_DEBUG_PS)
	{"DEBUG_PS",						"1"},
#endif

#if defined(ASURA_TOOLS)
	{"ASURA_TOOLS",						"1"},
#endif

	// Terminator
	{NULL,								NULL},
};


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// Static occurrance of the Direct3D effect-pool...

Direct3D_EffectPool*	Asura_Direct3D_Effect::s_pxD3D_EffectPool	= NULL;

int Asura_Direct3D_Effect::s_iDetailMapTextureIndex = ASURA_TEXTURE_NONE;

// JAT 17/02/07 path to load shaders from
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
Asura_String Asura_Direct3D_Effect::s_xShadersPath( "" );
#if defined(ASURA_PROJECT_SHADER_PATH)
Asura_String Asura_Direct3D_Effect::s_xProjectShadersPath( "" );
#endif // defined(ASURA_PROJECT_SHADER_PATH)
#endif


#if defined( ASURA_DIRECT3D_EFFECT_DEBUG)
#define ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( XHANDLE, ECLASS, ETYPE) \
{	\
	HRESULT xResult;	\
	D3DXPARAMETER_DESC xParameterDesc = { 0};	\
	xResult = m_pxD3DEffect->GetParameterDesc( (XHANDLE), &xParameterDesc);	\
	ASURA_D3D_OK( xResult);	\
	\
	Asura_Assert( (ECLASS) == xParameterDesc.Class, "Effects Parameter is the wrong class.");	\
	Asura_Assert( (ETYPE) == xParameterDesc.Type, "Effects Parameter is the wrong type.");	\
}
#else
#define ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( XHANDLE, ECLASS, ETYPE)
#endif

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

// JAT 22/02/07 on next-gen want this in all builds, current-gen only want this in tools builds
#if ASURA_LOAD_EFFECT_FILES == 1

////////////////////////////////////////////////////////////////////////////////
//
// IncludeHandler
//
// Implements the ID3DXInclude interface to handle #includes in text .fx files.
// We need this to correctly handle network loading, in case a header has been
// updated in version control but the main effect was loaded locally
//
////////////////////////////////////////////////////////////////////////////////

class IncludeHandler : public ID3DXInclude
{
public:
	bool LoadIncludeFileSource( const char* const szIncludeFilePath, LPCVOID* ppData, UINT* puBytes );
	bool TryFile( D3DXINCLUDE_TYPE eIncludeType, const char* const szSearchPath, LPCSTR szFileName, LPCVOID pParentData, LPCVOID* ppData, UINT* puBytes );

	STDMETHOD( Open )( THIS_ D3DXINCLUDE_TYPE eIncludeType, LPCSTR szFileName, LPCVOID pParentData, LPCVOID* ppData, UINT* puBytes );
	STDMETHOD( Close )( THIS_ LPCVOID pData );
};

////////////////////////////////////////////////////////////////////////////////
//
// IncludeHandler::Open()
//
// Called by the effect framework to read the contents of a #included file when
// loading text .fx files. Also seems to be called to load the main file itself
//
// Parameters:		eIncludeType - Local or system include type
//					szFileName - Filename to read, as an absolute path
//					pParentData - Pointer to the including dataset
//					ppData - Output pointer to address of created data buffer
//					puBytes - Number of bytes output to ppData
//
// Return value:	Success code
//
////////////////////////////////////////////////////////////////////////////////

bool IncludeHandler::LoadIncludeFileSource( const char* const szIncludeFilePath, LPCVOID* ppData, UINT* puBytes )
{
	const HANDLE xFile = CreateFile( szIncludeFilePath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL );
	if( xFile == INVALID_HANDLE_VALUE )
	{
		// failing to open the file is no longer an assert case here
		//Asura_Assertf( false, ("Unable to open shader include %s.", szIncludeFilePath) );
		return false;
	}

	const DWORD uFileSize = GetFileSize( xFile, NULL );

	char *const pcData = new char[ uFileSize ];
	if( !pcData )
	{
		Asura_Assert( false, "Unable to allocate memory for shader include handler." );
		CloseHandle( xFile );
		return false;
	}

	DWORD uNumBytesRead;
	if( !ReadFile( xFile, pcData, uFileSize, &uNumBytesRead, NULL ) || uNumBytesRead != uFileSize )
	{
		Asura_Assertf( false, ("Unable to read from file %s.", szIncludeFilePath) );
		CloseHandle( xFile );
		return false;
	}

	CloseHandle( xFile );

	*ppData = static_cast< void* >( pcData );
	*puBytes = uFileSize;
	return true;
}

bool IncludeHandler::TryFile( D3DXINCLUDE_TYPE eIncludeType, const char* const szSearchPath, LPCSTR szFileName, LPCVOID pParentData, LPCVOID* ppData, UINT* puBytes )
{
	char szPath[256];

	if ( Asura_StringUtil::StartsWith( szFileName, szSearchPath, true ) )
	{
		Asura_PrintfToBfr( szPath, "%s", 256, true, szFileName );
	}
	else
	{
		Asura_PrintfToBfr( szPath, "%s%s", 256, true, szSearchPath, szFileName );
	}

	return LoadIncludeFileSource( szPath, ppData, puBytes );
}

HRESULT IncludeHandler::Open( D3DXINCLUDE_TYPE eIncludeType, LPCSTR szFileName, LPCVOID pParentData, LPCVOID* ppData, UINT* puBytes )
{
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES )
	// DJG - 15/03/2007 : Added searching from project specific shaders path if defined
	Asura_String xIncludeFileName;
	if ( (Asura_Direct3D_Effect::GetShadersPath()) && ( Asura_StringUtil::StartsWith( szFileName, Asura_Direct3D_Effect::GetShadersPath(), true ) ) )
	{
		xIncludeFileName = &szFileName[strlen(Asura_Direct3D_Effect::GetShadersPath())];
	}
#if defined(ASURA_PROJECT_SHADER_PATH)
	else if ( (Asura_Direct3D_Effect::GetProjectShadersPath()) && ( Asura_StringUtil::StartsWith( szFileName, Asura_Direct3D_Effect::GetProjectShadersPath(), true ) ) )
	{
		xIncludeFileName = &szFileName[strlen(Asura_Direct3D_Effect::GetProjectShadersPath())];
	}
#endif
	else
	{
		xIncludeFileName = "Shaders/";
		xIncludeFileName += szFileName;
	}

#if defined(ASURA_PROJECT_SHADER_PATH)
	// try the project shaders folder first
	const char* const szProjectShadersPath = Asura_Direct3D_Effect::GetProjectShadersPath();

	if ( NULL != szProjectShadersPath )
	{
		if ( TryFile( eIncludeType, szProjectShadersPath, xIncludeFileName.GetString(), pParentData, ppData, puBytes ) )
		{
			return S_OK;
		}
	}

#endif
	const char* const szShadersPath = Asura_Direct3D_Effect::GetShadersPath();
	Asura_Assert( szShadersPath != NULL, "Invalid shader name but not loading shaders from disc" );

	if ( TryFile( eIncludeType, szShadersPath, xIncludeFileName.GetString(), pParentData, ppData, puBytes ) )
	{
		return S_OK;
	}
#elif !defined( ASURA_TARGET_SM3 )
	char szPath[256];
	// JAT 17/02/07 for non next-gen builds load the shaders from cwd path e.g. Asura_Files/Shaders folder
	if( Asura_StringUtil::StartsWith( szFileName, "Shaders\\", true ) || Asura_StringUtil::StartsWith( szFileName, "\\Shaders\\", true ) )
	{
		Asura_PrintfToBfr( szPath, "%s", 256, true, szFileName );
	}
	else
	{
		Asura_PrintfToBfr( szPath, "Shaders\\%s", 256, true, szFileName );
	}
	Asura_FileAccess::PlatformGetNetworkFilename( szPath, szNetworkPath, 256 );

	if ( LoadIncludeFileSource( szNetworkPath, ppData, puBytes ) )
	{
		return S_OK;
	}
#else
	Asura_Assert( false, "IncludeHandler::Open can't load shaders from disc in present build" );
#endif

	Asura_Assertf( false, ("Unable to open shader include %s.", szFileName) );

	return E_FAIL;
}

////////////////////////////////////////////////////////////////////////////////
//
// IncludeHandler::Close
//
// Called by the effects framework to close an include file previously opened
//
// Parameters:		pData - Pointer to the data buffer
// Return value:	Success code
//
////////////////////////////////////////////////////////////////////////////////

HRESULT IncludeHandler::Close( LPCVOID pData )
{
	delete[] pData;
	return S_OK;
}

#endif // #if ASURA_LOAD_EFFECT_FILES == 1

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::LoadEffectFromFile
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Loads the Direct3D effect using the current filename.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

// JAT 22/02/07 on non next-gen only load effect files in tools builds
// JAT 22/02/07 only want this function in tools builds on current-gen (but on all builds for next-gen)
#if ASURA_LOAD_EFFECT_FILES == 1

void Asura_Direct3D_Effect::LoadEffectFromFile (void)
{
	LPD3DXBUFFER pxErrorMessages	= NULL;
	IncludeHandler xIncludeHandler;

	// Don't need to do a network lookup here because the include handler gets called
	// even for the main .fx file itself

	// Call the appropriate Direct3D function...
#if defined(ASURA_ASSERT) || (!defined(ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)&&defined(ASURA_PROJECT_SHADER_PATH))
	HRESULT iResult = S_OK;
#endif
	
	Direct3D_Effect* pxD3DEffect = NULL;
	
	// JAT 17/02/07 - load the effect file from disc or as a resource (the resource will be a compiled file)
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
	// DJG - 15/03/2007 : Check the project shader search path first, this allows projects to override the
	// default shaders if they so require, as well as providing additional shaders
#if defined(ASURA_PROJECT_SHADER_PATH)
	const char* const szProjectShaderPath = Asura_Direct3D_Effect::GetProjectShadersPath();
	bool bLoadCommonShader=true;
	if ( szProjectShaderPath )
	{
		Asura_String xProjectShaderFullName( szProjectShaderPath );
		xProjectShaderFullName += m_xFilename;

		Asura_OutputToDebugger_Retail(( "Compiling: \'%s\'\n", xProjectShaderFullName ));
		iResult = D3DXCreateEffectFromFile( Asura_Direct3D::D3DDevice,
											xProjectShaderFullName.GetString(),
											axDirect3DMacros,
											&xIncludeHandler,
											dwEFFECT_CREATION_FLAGS,
											s_pxD3D_EffectPool,
											&pxD3DEffect,
											&pxErrorMessages );
		if ( SUCCEEDED(iResult) )
		{
			Asura_OutputToDebugger_Retail(( "Finished compiling: '%s'\n", xProjectShaderFullName ));
			bLoadCommonShader = false;
		}
		else
		{
			Asura_OutputToDebugger_Retail(( "Project shader file failed to compile - " ));
		}
	}

	if ( bLoadCommonShader )
#endif
	{
		const char* const szShaderPath = Asura_Direct3D_Effect::GetShadersPath();
		Asura_String xShaderFullName( szShaderPath );
		xShaderFullName += m_xFilename;
		
		Asura_OutputToDebugger_Retail(( "Compiling: '%s'\n", xShaderFullName));

#ifdef ASURA_ASSERT
	iResult =
#endif
		D3DXCreateEffectFromFile(	Asura_Direct3D::D3DDevice,
									xShaderFullName.GetString(),
									axDirect3DMacros,
									&xIncludeHandler,
									dwEFFECT_CREATION_FLAGS,
									s_pxD3D_EffectPool,
									&pxD3DEffect,
									&pxErrorMessages);
									
		Asura_OutputToDebugger_Retail(( "Finished Compiling: '%s'\n", xShaderFullName));

#if defined(ASURA_ASSERT)
		Asura_Assertf(SUCCEEDED(iResult), ( "Error returned from D3DXCreateEffectFromFile '%s'", xShaderFullName.GetString() ) );
#endif	// defined(ASURA_ASSERT)
	}
#else
		// only allowed to load from disc in tools builds
		Asura_Assert( false, "Can't load shaders from disc in non tools builds" );
		pxD3DEffect = NULL;
#endif // #ifdef ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES


	Asura_Assertf(SUCCEEDED(iResult), ( "Error returned from D3DXCreateEffectFromResource '%s'", m_xFilename.GetString() ) );

#if defined(ASURA_ASSERT)
	// Did we manage to load the effect?
	if (NULL == pxD3DEffect)
	{
		// No - create the error string...
		char* szErrorMessage = pxErrorMessages ? static_cast<char*>(pxErrorMessages->GetBufferPointer()) : "unknown"; 

		Asura_Assertf	((NULL != pxD3DEffect), ("Failed to load effect \"%s\" due to error: %s.", m_xFilename.GetString(), szErrorMessage));
	}
#endif	// defined(ASURA_ASSERT)

	// free up any memory used by the error messages
	if ( pxErrorMessages )
	{
		pxErrorMessages->Release();
	}
	
	if( NULL != pxD3DEffect)
	{
		this->ReleaseEffect();

		this->m_pxD3DEffect = pxD3DEffect;

		QueryTechniqueHandles();
	}
	else
	{
		Asura_Assertf( NULL == this->m_pxD3DEffect, ("Effect '%s' failed to load so the old version has been left.", m_xFilename.GetString()));
	}

	if (m_pfnOnLoadCallback)
	{
		(*m_pfnOnLoadCallback)(this);
	}
}

/// End of function 'Asura_Direct3D_Effect::LoadEffectFromFile'. ///

#endif // #if ASURA_LOAD_EFFECT_FILES == 1

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::ReleaseEffect
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Destroys the existing Direct3D effect.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::ReleaseEffect (void)
{
	// Do we have an effect?

	if (NULL != m_pxD3DEffect)
	{

// JAT 22/02/07 on next-gen want this in all builds, current-gen only want this in tools builds
#if ASURA_LOAD_EFFECT_FILES == 1
		// Inform our callback, if we have one
		if (m_pfnOnUnloadCallback)
		{
			(*m_pfnOnUnloadCallback)(this);
		}
#endif // #if ASURA_LOAD_EFFECT_FILES == 1

		// Yes - release it...

		m_pxD3DEffect->Release	();

		m_pxD3DEffect	= NULL;
	}

	delete[] m_pxTechniqueHandles;
	m_pxTechniqueHandles = NULL;

	m_uNumTechniques = 0;
}

/// End of function 'Asura_Direct3D_Effect::ReleaseEffect'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Duncan Hopkins
// Date         : 26-10-2006
//
// Description  : Sets the specified value array in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameter(const D3DXHANDLE xHandle, const void* pData, const u_int uSize)
{
	Asura_Assert	((NULL != xHandle), "Invalid name for boolean parameter.");
	if (NULL == xHandle)
	{
		return;
	}
	
	HRESULT xResult;
	
	//ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_STRUCT, D3DXPT_VOID);
		
	xResult = m_pxD3DEffect->SetValue( xHandle, pData, uSize);
	ASURA_D3D_OK( xResult);
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameter (const D3DXHANDLE xHandle,
										  const bool bValue)
{
	// Was the xHandle valid?

	if (NULL == xHandle)
	{
		// No - better assert...

		Asura_Assert	((NULL != xHandle), "Invalid name for boolean parameter.");
	}

	else
	{
		// Yes - set the value and return...
		ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_SCALAR, D3DXPT_BOOL);

		m_pxD3DEffect->SetBool	(xHandle, bValue);
	}
}

/// End of function 'Asura_Direct3D_Effect::SetParameter'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameter (const D3DXHANDLE xHandle,
										  const int iValue)
{
	// Was the xHandle valid?

	if (NULL == xHandle)
	{
		// No - better assert...

		Asura_Assert	((NULL != xHandle), "Invalid name for integer parameter.");
	}

	else
	{
		// Yes - set the value and return...
		ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_SCALAR, D3DXPT_INT);
		
		m_pxD3DEffect->SetInt	(xHandle, iValue);
	}
}

/// End of function 'Asura_Direct3D_Effect::SetParameter'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameter (const D3DXHANDLE xHandle,
										  const float fValue)
{
	// Was the xHandle valid?

	if (NULL == xHandle)
	{
		// No - better assert...

		Asura_Assert	((NULL != xHandle), "Invalid name for floating-point parameter.");
	}

	else
	{
		// Yes - set the value and return...
		ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_SCALAR, D3DXPT_FLOAT);

		m_pxD3DEffect->SetFloat	(xHandle, fValue);
	}
}

/// End of function 'Asura_Direct3D_Effect::SetParameter'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameter (const D3DXHANDLE xHandle,
										  const D3DXVECTOR4& xValue)
{
	// Was the xHandle valid?

	if (NULL == xHandle)
	{
		// No - better assert...

		Asura_Assert	((NULL != xHandle), "Invalid name for D3DXVECTOR4 parameter.");
	}

	else
	{
		// Yes - set the value and return...
		ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_VECTOR, D3DXPT_FLOAT);

		m_pxD3DEffect->SetVector	(xHandle, &xValue);
	}
}

/// End of function 'Asura_Direct3D_Effect::SetParameter'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameter (const D3DXHANDLE xHandle,
										  const D3DXMATRIX& xValue)
{
	// Was the xHandle valid?

	if (NULL == xHandle)
	{
		// No - better assert...

		Asura_Assert	((NULL != xHandle), "Invalid name for D3DXMATRIX parameter.");
	}

	else
	{
		// Yes - set the value and return...
		ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_MATRIX_ROWS, D3DXPT_FLOAT);

		m_pxD3DEffect->SetMatrix	(xHandle, &xValue);
	}
}

/// End of function 'Asura_Direct3D_Effect::SetParameter'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameter
// Author       : Andy McDonald
// Date         : 3-Oct-2006
//
// Description  : Sets the specified texture in the specified parameter.
//
// Parameters   : The xHandle to the parameter and the texture to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterTexture (const D3DXHANDLE xHandle,
										  Direct3D_BaseTexture* const pxTexture)
{
	if( !xHandle )
	{
		Asura_Assert( false, "Invalid name for texture parameter." );
	}
	else
	{
		ASURA_DIRECT3D_EFFECT_DEBUG_PARAMETER( xHandle, D3DXPC_OBJECT, D3DXPT_TEXTURE);
		
		m_pxD3DEffect->SetTexture( xHandle, pxTexture );
	}
}

void Asura_Direct3D_Effect::SetParameter (const D3DXHANDLE xHandle,
										  Direct3D_BaseTexture* const pxTexture)
{
	SetParameterTexture( xHandle, pxTexture);
}

/// End of function 'Asura_Direct3D_Effect::SetParameter'. ///


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::Asura_Direct3D_Effect
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Constructor function for the creation of a Direct3D effect
//				  given the location of the data in memory.
//
// Parameters   : The ID of the effect along with the location and size of the
//				  text from the .fx file to use for it's initialisation.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Effect::Asura_Direct3D_Effect( const void* pvData, const u_int uDataSize )
	: m_pxD3DEffect(NULL)
	, m_pxTechniqueHandles(NULL)
	, m_uNumPasses(0)
	, m_uNumTechniques(0)
// JAT 22/02/07 on next-gen want this in all builds, current-gen only want this in tools builds
#if ASURA_LOAD_EFFECT_FILES == 1
	, m_szName(NULL)
	, m_xFilename(NULL)
	, m_pfnOnLoadCallback(NULL)
	, m_pfnOnUnloadCallback(NULL)
#endif	// #if ASURA_LOAD_EFFECT_FILES == 1
{
	// Use Direct3D to create the effect...
	D3DXCreateEffect	(Asura_Direct3D::D3DDevice,
						 pvData,
						 uDataSize,
						 axDirect3DMacros,
						 NULL,
						 dwEFFECT_CREATION_FLAGS,
						 s_pxD3D_EffectPool,
						 &m_pxD3DEffect,
						 0);

	QueryTechniqueHandles();
}

/// End of function 'Asura_Direct3D_Effect::Asura_Direct3D_Effect'. ///

// JAT 22/02/07 on next-gen want this in all builds, current-gen only want this in tools builds
#if ASURA_LOAD_EFFECT_FILES == 1

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::Asura_Direct3D_Effect
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Constructor function for the creation of a Direct3D effect
//				  by loading the data from the specified file.
//
// Parameters   : The name of the effect along with the filename to use when
//				  loading the data.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Effect::Asura_Direct3D_Effect (const char* szEffectName,
											  const char* szFilename) :
	m_pxD3DEffect(NULL),
	m_pxTechniqueHandles(NULL),
	m_uNumPasses(0),
	m_uNumTechniques(0),
	m_szName(szEffectName),
	m_xFilename(szFilename),
	m_pfnOnLoadCallback(NULL),
	m_pfnOnUnloadCallback(NULL)
{
	// Load the effect...

	LoadEffectFromFile	();
}

/// End of function 'Asura_Direct3D_Effect::Asura_Direct3D_Effect'. ///

#endif	// #if ASURA_LOAD_EFFECT_FILES == 1

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::~Asura_Direct3D_Effect
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Destructor function for the Direct3D effect.
//
// Return value : N/A.
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Effect::~Asura_Direct3D_Effect ()
{
	// Simply release the effect...

	ReleaseEffect	();
}

/// End of function 'Asura_Direct3D_Effect::~Asura_Direct3D_Effect'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::Tools_Reload
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Destroys the old copy of the effect, then loads the data from
//				  the .fx file again.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

#if ASURA_LOAD_EFFECT_FILES == 1 && !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
void Asura_Direct3D_Effect::Tools_Reload (void)
{
#if !defined(ASURA_PC_DIRECT3D_EFFECT_DEBUG_SOURCE_FILE)
	// Release the existing effect...

//	ReleaseEffect	();


	// Load the data...
	if (!m_xFilename.IsNull())
	{
		LoadEffectFromFile	();
	}	


#endif	// !defined(ASURA_PC_DIRECT3D_EFFECT_DEBUG_SOURCE_FILE)

	m_xParameterHandleTree.Reset();
}
#endif

/// End of function 'Asura_Direct3D_Effect::Tools_Reload'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::CreateEffectPool
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Creates the effect-pool for the Direct3D effects to use.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::CreateEffectPool (void)
{
	// Do we already have an effect-pool?

	if (NULL != s_pxD3D_EffectPool)
	{
		// Yes - destroy it...

		DestroyEffectPool	();
	}


	// Create the effect-pool...

	D3DXCreateEffectPool	(&s_pxD3D_EffectPool);
}

/// End of function 'Asura_Direct3D_Effect::CreateEffectPool'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::DestroyEffectPool
// Author       : Andy Tate
// Date         : 11-May-2005
//
// Description  : Destroys the previously allocated effect pool.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::DestroyEffectPool (void)
{
	// Do we already have an effect-pool?

	if (NULL != s_pxD3D_EffectPool)
	{
		// Yes - release it...

		s_pxD3D_EffectPool->Release	();

		s_pxD3D_EffectPool	= NULL;
	}
}

/// End of function 'Asura_Direct3D_Effect::DestroyEffectPool'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Duncan Hopkins
// Date         : 26-10-2006
//
// Description  : Sets the specified value array in the specified parameter.
//
// Parameters   : The name of the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterByName(const char* szName, const void* pData, const u_int uSize)
{
	Asura_Assert(m_pxD3DEffect, "D3D effect pointer is null");
	if(!m_pxD3DEffect) return;

	D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);

	Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for void parameter.", szName));
	
	// Set this parameter...

	SetParameter( xHandle, pData, uSize);
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The name of the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterByName (const char* szName,
												const bool bValue)
{
	Asura_Assert(m_pxD3DEffect, "D3D effect pointer is null");
	if(!m_pxD3DEffect) return;

	D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);

	Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for bool parameter.", szName));
	
	// Set this parameter...

 	SetParameter	(xHandle, bValue);
}

/// End of function 'Asura_Direct3D_Effect::SetParameterByName'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The name of the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterByName (const char* szName,
												const int iValue)
{
	Asura_Assert(m_pxD3DEffect, "D3D effect pointer is null");
	if(!m_pxD3DEffect) return;

	D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);

	Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for int parameter.", szName));
	
	// Set this parameter...

	SetParameter	(xHandle, iValue);
}

/// End of function 'Asura_Direct3D_Effect::SetParameterByName'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The name of the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterByName (const char* szName,
												const float fValue)
{
	Asura_Assert(m_pxD3DEffect, "D3D effect pointer is null");
	if(!m_pxD3DEffect) return;

	D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);

	Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for float parameter.", szName));
	
	// Set this parameter...

	SetParameter	(xHandle, fValue);
}

/// End of function 'Asura_Direct3D_Effect::SetParameterByName'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The name of the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterByName (const char* szName,
												const D3DXVECTOR4& xValue)
{
	Asura_Assert(m_pxD3DEffect, "D3D effect pointer is null");
	if(!m_pxD3DEffect) return;

	D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);

	Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for D3DXVECTOR4 parameter.", szName));
	
	// Set this parameter...

	SetParameter	(xHandle, xValue);
}

/// End of function 'Asura_Direct3D_Effect::SetParameterByName'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Andy Tate
// Date         : 18-May-2005
//
// Description  : Sets the specified value in the specified parameter.
//
// Parameters   : The name of the parameter and the value to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterByName (const char* szName,
												const D3DXMATRIX& xValue)
{
	Asura_Assert(m_pxD3DEffect, "D3D effect pointer is null");
	if(!m_pxD3DEffect) return;

	D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);

	Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for D3DXMATRIX parameter.", szName));
	
	// Set this parameter...

	SetParameter	(xHandle, xValue);
}

/// End of function 'Asura_Direct3D_Effect::SetParameterByName'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::SetParameterByName
// Author       : Andy McDonald
// Date         : 3-Oct-2006
//
// Description  : Sets the specified texture in the specified parameter.
//
// Parameters   : The name of the parameter and the texture to be set.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::SetParameterTextureByName (const char* szName,
												Direct3D_BaseTexture* const pxTexture)
{
	if( !m_pxD3DEffect )
	{
		Asura_Assert( false, "D3D effect pointer is null" );
	}
	else
	{
		const D3DXHANDLE xHandle = m_pxD3DEffect->GetParameterByName( NULL, szName );
		
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for Direct3D_BaseTexture parameter.", szName));
	
		SetParameterTexture( xHandle, pxTexture );
	}
}

void Asura_Direct3D_Effect::SetParameterByName (const char* szName,
												Direct3D_BaseTexture* const pxTexture)
{
	SetParameterTextureByName( szName, pxTexture);
}

/// End of function 'Asura_Direct3D_Effect::SetParameterByName'. ///

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, const bool bValue)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif
	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, bValue);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for bool parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, bValue);
	}
}

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, const int iValue)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif
	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, iValue);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for int parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, iValue);
	}
}

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, const float fValue)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif
	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, fValue);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for float parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, fValue);
	}
}

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, const D3DXMATRIX& xValue)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif

	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, xValue);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for D3DXMATRIX parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, xValue);
	}
}

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, const D3DXVECTOR4& xValue)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif

	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, xValue);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for D3DXVECTOR4 parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, xValue);
	}
}

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, const void* pData, const u_int uSize)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif
	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, pData, uSize);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for void parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, pData, uSize);
	}
}

void Asura_Direct3D_Effect::SetParameterByHash(const char* szName, Asura_Hash_ID uHashID, Direct3D_BaseTexture* const pxTexture)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif
	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, pxTexture);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for Direct3D_BaseTexture parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, pxTexture);
	}
}

void Asura_Direct3D_Effect::SetParameterTextureByHash(const char* szName, Asura_Hash_ID uHashID, Direct3D_BaseTexture* const pxTexture)
{
#if _DEBUG
	Asura_Assert(Asura_GetHashID(szName) == uHashID, "hash and name don't match");
#endif
	const D3DXHANDLE* pxHandle = m_xParameterHandleTree.GetDataByHash(uHashID);
	if (pxHandle)
	{
		SetParameter(*pxHandle, pxTexture);
	}
	else
	{
		D3DXHANDLE	xHandle	= m_pxD3DEffect->GetParameterByName (NULL, szName);
		Asura_Assertf((NULL != xHandle), ("Invalid name '%s' for Direct3D_BaseTexture parameter.", szName));
		m_xParameterHandleTree.AddNode( uHashID, xHandle );
		SetParameter(xHandle, pxTexture);
	}
}


// KJFL - 04 April 2006 - the presence of this function needs a bit of a rethink. Why should the effect class
// know about specific material / texture / effect parameters? (Going to change to parameter blocks too)
void Asura_Direct3D_Effect::SetTextureRenderStates (const int iMaterialID)
{
	// Do we wish to set the texture(s)?

	//if (m_uTextureTypeFlags & ASURA_D3D_EFFECT_TTF_TEXTURED)
	{
		const int	iTexture			= Asura_Material_System::GetTextureID (iMaterialID, 0);
		const bool	bDualPassTexture	= Asura_Material_System::IsMaterialDualTexture (iMaterialID);


		// Set the appropriate texture render-states...

		Asura_Render::CurrentStates.RequestTextureIndex	(0, iTexture);
		SetParameterByName								("g_bDualPassTexture", bDualPassTexture);

		if (bDualPassTexture)
		{
			const int	iDualPassTexture	= Asura_Material_System::GetTextureID (iMaterialID, 1);


			// Set the dual-pass texture...

			Asura_Render::CurrentStates.RequestTextureIndex	(1, iDualPassTexture);
		}

		else
		{
			// Set the original texture (just to be safe)...

			Asura_Render::CurrentStates.RequestTextureIndex	(1, iTexture);
		}
	}


	// Do we wish to set the normal-map(s)?

	//if (m_uTextureTypeFlags & ASURA_D3D_EFFECT_TTF_NORMAL_MAP)
	{
		const int	iNormalMap			= Asura_Material_System::GetNormalMapID (iMaterialID, 0);
		const bool	bDualPassNormalMap	= Asura_Material_System::IsMaterialDualNormalMap (iMaterialID);


		// Yes - set the appropriate texture render-states...

		Asura_Render::CurrentStates.RequestTextureIndex	(2, iNormalMap);
		SetParameterByName								("g_bDualPassNormalMap", bDualPassNormalMap);

		if (bDualPassNormalMap)
		{
			const int	iDualPassNormalMap	= Asura_Material_System::GetNormalMapID (iMaterialID, 1);


			// Set the dual-pass texture...

			Asura_Render::CurrentStates.RequestTextureIndex	(3, iDualPassNormalMap);
		}

		else
		{
			// Set the original nornal-map (just to be safe)...

			Asura_Render::CurrentStates.RequestTextureIndex	(3, iNormalMap);
		}
	}

	// Do we want to set reflectivity?

	//if (m_uTextureTypeFlags & ASURA_D3D_EFFECT_TTF_REFLECTIVE)
	//{
	//	bool bReflective = Asura_Material_System::IsMaterialReflective(iMaterialID);
	//	SetParameterByName("g_bReflectiveTexture", bReflective);
	//	if (bReflective)
	//	{
	//		// TODO - see if there's a material-specific one?
	//		Asura_Render::CurrentStates.RequestTextureIndex(4, Asura_Render::GetReflectanceMap());
	//		// TODO - set reflectivity ("g_fMaterialReflectivity"). Currently, only texture alpha is used to scale effect
	//	}
	//}

	//if (m_uTextureTypeFlags & ASURA_D3D_EFFECT_TTF_DETAIL_MAP)
	//{
	//	bool bDetailMap = Asura_Material_System::IsMaterialUsingDetailMap(iMaterialID);
	//	SetParameterByName("g_bUseDetailMap", bDetailMap);
	//	if (bDetailMap)
	//	{
	//		if (s_iDetailMapTextureIndex == ASURA_TEXTURE_NONE)
	//		{
	//			s_iDetailMapTextureIndex = Asura_Texture_Management::AddTextureWithResetGroup("\\graphicnovel\\detail.dds", ASURA_TEXTURE_LOADED_PERMANENT );
	//			Asura_Texture_Management::LoadTextures();
	//		}
	//		// TODO - see if there's a material-specific one?
	//		Asura_Render::CurrentStates.RequestTextureIndex(5, s_iDetailMapTextureIndex);
	//		// TODO - set reflectivity ("g_fMaterialReflectivity"). Currently, only texture alpha is used to scale effect
	//	}
	//}

	Asura_TextureAnimationOutput xTextureAnim;
	if (Asura_Material_System::GetTextureAnimationState(iMaterialID,0,&xTextureAnim))
	{
		D3DXVECTOR4 xOffset(xTextureAnim.m_fUOffset, xTextureAnim.m_fVOffset, 1.0f, 1.0f); // No specified scale yet
		SetParameterByName("g_xTextureCoordOffsetAndScale", xOffset);
	}
	// The waterfall currently doesn't work, as the values are updated in the old env renderer
	//	else if (Asura_Material_System::IsMaterialWaterfall(iMaterialID))
	//	{
	//		D3DXVECTOR4 xOffset(s_fScrollTextureOffsetU, s_fScrollTextureOffsetV, 1.0f, 1.0f); // No specified scale yet
	//		SetParameterByName("g_xTextureCoordOffsetAndScale", xOffset);
	//	}
	else
	{
		D3DXVECTOR4 xOffset(0.0f, 0.0f, 1.0f, 1.0f);
		SetParameterByName("g_xTextureCoordOffsetAndScale", xOffset);
	}
}

/// End of function 'Asura_Direct3D_Effect::SetTextureRenderStates'. ///




////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::Begin
// Author       : Andy Tate
// Date         : 25-May-2005
//
// Description  : Begins using this Direct3D effect.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::Begin (const bool bSaveState /*= false*/, const u_int uTechnique /*= 0*/)
{
	const u_int	uFlags	= (bSaveState)
						? 0
						: D3DXFX_DONOTSAVESTATE;

	Asura_Render::CurrentStates.InvalidateTextureOpMode();

	if( m_pxTechniqueHandles && uTechnique < m_uNumTechniques )
	{
		m_pxD3DEffect->SetTechnique( m_pxTechniqueHandles[ uTechnique ] );
	}
	else
	{
		Asura_Assertf( false, ("Invalid effect technique (%u). Number of techniques: %u",
			uTechnique, m_uNumTechniques) );
	}

	// Begin using this effect...
#ifdef ASURA_ASSERT
	HRESULT iResult = 
#endif
	m_pxD3DEffect->Begin	(&m_uNumPasses, uFlags);
	Asura_Assert(iResult == S_OK, "Failed to begin effect");
}

/// End of function 'Asura_Direct3D_Effect::Begin'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::BeginPass
// Author       : Duncan Hopkins
// Date         : 26-Oct-2006
//
// Description  : Allows direct access to the BeginPass() method
//
// Parameters   : The pass number to draw
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::BeginPass( const u_int uPass)
{
	HRESULT xResult;
	
	Asura_Assert( uPass < this->m_uNumPasses, "This effect does not support that many passed");
	xResult = m_pxD3DEffect->BeginPass( uPass);
	ASURA_D3D_OK( xResult);
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::EndPass
// Author       : Duncan Hopkins
// Date         : 26-Oct-2006
//
// Description  : Allows direct access to the EndPass() method
//
// Parameters   : The pass number to draw
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::EndPass()
{
	HRESULT xResult;
	
	xResult = m_pxD3DEffect->EndPass();
	ASURA_D3D_OK( xResult);
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::RenderAllIndexedPrimitivePasses
// Author       : Andy Tate
// Date         : 25-May-2005
//
// Description  : Renders ALL of the passes for the current vertex buffer for
//				  this Direct3D effect.
//
// Parameters   : A pointer to the function to use when committing changes on
//				  a per-pass basis.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::RenderAllIndexedPrimitivePasses (const D3DPRIMITIVETYPE ePrimitiveType,
															 const u_int uStartVertex,
															 const u_int uNumVertices,
															 const u_int uStartIndex,
															 const u_int uNumPrimitives,
															 CommitChangesPerPassFn* pxCommitChangesPerPassFn,
															 const u_int uBaseVertexIndex /* = 0 */)
{
	// Render each pass in-turn...
	for (u_int uPass = 0; uPass < m_uNumPasses; uPass++)
	{
		// Begin this pass...
		BeginPass(uPass);

		// Do we need to commit changes per pass?
		if (NULL != pxCommitChangesPerPassFn)
		{
			// Yes - call the appropriate function...
			if (pxCommitChangesPerPassFn(uPass))
			{
				// Only call "CommitChanges" if any state changes have happened after "BeginPass" is called...
				m_pxD3DEffect->CommitChanges	();
			}
		}

		// Render the mesh with the applied technique...
		Asura_Direct3D::D3DDevice->DrawIndexedPrimitive	(ePrimitiveType,
														 uBaseVertexIndex,
														 uStartVertex,
														 uNumVertices,
														 uStartIndex,
														 uNumPrimitives);

		// End this pass...
		EndPass();
	}
}

/// End of function 'Asura_Direct3D_Effect::RenderAllIndexedPrimitivePasses'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::RenderAllPrimitivePasses
// Author       : Andy Tate
// Date         : 25-May-2005
//
// Description  : Renders ALL of the passes for the current vertex buffer for
//				  this Direct3D effect.
//
// Parameters   : A pointer to the function to use when committing changes on
//				  a per-pass basis.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::RenderAllPrimitivePasses (const D3DPRIMITIVETYPE ePrimitiveType,
															 const u_int uStartVertex,
															 const u_int uNumPrimitives,
															 CommitChangesPerPassFn* pxCommitChangesPerPassFn)
{
	// Render each pass in-turn...
	for (u_int uPass = 0; uPass < m_uNumPasses; uPass++)
	{
		// Begin this pass...
		BeginPass(uPass);

		// Do we need to commit changes per pass?
		if (NULL != pxCommitChangesPerPassFn)
		{
			// Yes - call the appropriate function...
			if (pxCommitChangesPerPassFn(uPass))
			{
				// Only call "CommitChanges" if any state changes have happened after "BeginPass" is called...
				m_pxD3DEffect->CommitChanges	();
			}
		}

		// Render the mesh with the applied technique...
		Asura_Direct3D::D3DDevice->DrawPrimitive(ePrimitiveType,
			uStartVertex,
			uNumPrimitives);

		// End this pass...
		EndPass();
	}
}

/// End of function 'Asura_Direct3D_Effect::RenderAllIndexedPrimitivePasses'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::End
// Author       : Andy Tate
// Date         : 25-May-2005
//
// Description  : Ends the use of this Direct3D effect.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::End (void)
{
#if defined(ASURA_ASSERT)
	const HRESULT	iResult	= m_pxD3DEffect->End();

	// We're done with this effect...
	Asura_Assert(iResult == S_OK, "Failed to end effect");
#else
	// We're done with this effect...
	m_pxD3DEffect->End();
#endif

	// Reset the pass count...
	m_uNumPasses	= 0;
}

/// End of function 'Asura_Direct3D_Effect::End'. ///

#if ASURA_LOAD_EFFECT_FILES == 1 && !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
// JAT 17/02/07 set a path for loading shaders from (in tools)
void Asura_Direct3D_Effect::SetShadersPath( const char* const szShadersPath )
{
	s_xShadersPath = szShadersPath;	
}

const char* const Asura_Direct3D_Effect::GetShadersPath( void )
{
	return s_xShadersPath.GetString();
}

#if defined(ASURA_PROJECT_SHADER_PATH)
void Asura_Direct3D_Effect::SetProjectShadersPath( const char* const szProjectShadersPath )
{
	s_xProjectShadersPath = szProjectShadersPath;
	// DJG : Should this also set s_bLoadShadersFromDisc, or should we require
	// that SetShadersPath is called for that to be true.
}

const char* const Asura_Direct3D_Effect::GetProjectShadersPath( void )
{
	return s_xProjectShadersPath.GetString();
}
#endif

#endif

void Asura_Direct3D_Effect::ResetParameterTree()
{
	m_xParameterHandleTree.Reset();
}


bool Asura_Direct3D_Effect::PreReset(void)
{
	HRESULT xResult;
	
	Direct3D_Effect* pxD3DEffect = this->GetD3DEffect();
	
	if( NULL == pxD3DEffect)
	{
		return( false);
	}
	
	D3DXEFFECT_DESC xEffectDesc = { 0};
	xResult = pxD3DEffect->GetDesc( &xEffectDesc);
	ASURA_D3D_OK( xResult);
	
	for( u_int uParameterIndex = 0; uParameterIndex < xEffectDesc.Parameters; uParameterIndex += 1)
	{
		D3DXHANDLE xParameterHandle = pxD3DEffect->GetParameter( NULL, uParameterIndex);
		if( NULL == xParameterHandle)
		{
			continue;
		}
		
		D3DXPARAMETER_DESC xParameterDesc = { 0};
		xResult = pxD3DEffect->GetParameterDesc( xParameterHandle, &xParameterDesc);
		ASURA_D3D_OK( xResult);
		
		if( D3DXPT_TEXTURE == xParameterDesc.Type)
		{
			LPDIRECT3DBASETEXTURE9 pxTextureBase = NULL;
			xResult = pxD3DEffect->GetTexture( xParameterHandle, &pxTextureBase);
			ASURA_D3D_OK( xResult);

			if( pxTextureBase )
			{
				#if defined ASURA_ASSERT
					#if defined ASURA_LOAD_EFFECT_FILES && ASURA_LOAD_EFFECT_FILES == 1
						const char* const szName = m_szName;
					#else
						const char* const szName = "UNKNOWN";
					#endif
				#endif

				// Can't just ask the base resource type what pool it's in. Have to interpret it as a
				// specific texture type and grab it from there

				bool bUnmanagedResource = false;

				LPDIRECT3DTEXTURE9 pxTexture2D = NULL;
				LPDIRECT3DCUBETEXTURE9 pxTextureCube = NULL;
				LPDIRECT3DVOLUMETEXTURE9 pxTextureVolume = NULL;

				if( SUCCEEDED( pxTextureBase->QueryInterface( IID_IDirect3DTexture9,
															  reinterpret_cast< void** >( &pxTexture2D ) ) ) )
				{
					D3DSURFACE_DESC xSurfaceDesc;
					xResult = pxTexture2D->GetLevelDesc( 0, &xSurfaceDesc );

					if( SUCCEEDED( xResult ) )
					{
						bUnmanagedResource = (xSurfaceDesc.Pool != D3DPOOL_MANAGED);
					}

					SAFE_RELEASE( pxTexture2D );
				}
				else if( SUCCEEDED( pxTextureBase->QueryInterface( IID_IDirect3DCubeTexture9,
																   reinterpret_cast< void** >( &pxTextureCube ) ) ) )
				{
					D3DSURFACE_DESC xSurfaceDesc;
					xResult = pxTextureCube->GetLevelDesc( 0, &xSurfaceDesc );

					if( SUCCEEDED( xResult ) )
					{
						bUnmanagedResource = (xSurfaceDesc.Pool != D3DPOOL_MANAGED);
					}

					SAFE_RELEASE( pxTextureCube );
				}
				else if( SUCCEEDED( pxTextureBase->QueryInterface( IID_IDirect3DVolumeTexture9,
																   reinterpret_cast< void** >( &pxTextureVolume ) ) ) )
				{
					D3DVOLUME_DESC xVolumeDesc;
					xResult = pxTextureVolume->GetLevelDesc( 0, &xVolumeDesc );

					if( SUCCEEDED( xResult ) )
					{
						bUnmanagedResource = (xVolumeDesc.Pool != D3DPOOL_MANAGED);
					}

					SAFE_RELEASE( pxTextureVolume );
				}
				else
				{
					Asura_Assertf( false, ("Effect '%s' Parameter '%s' has texture type, but could not be queried for a texture interface.",
										   szName,
										   xParameterDesc.Name) );
				}

				if( bUnmanagedResource )
				{
					Asura_Assertf( false, ("Effect '%s' Parameter '%s' Unmanaged texture is still set.", szName, xParameterDesc.Name));
					xResult = pxD3DEffect->SetTexture( xParameterHandle, NULL);
					ASURA_D3D_OK( xResult);
				}
			}

			SAFE_RELEASE( pxTextureBase);
		}
	}
	
	return( true);
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_Effect::QueryTechniqueHandles
// Author       : Andy McDonald
// Date         : 03-Sept-2007
//
// Description  : Initialises the technique handle array, so we don't need to
//                keep querying each time we begin.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Effect::QueryTechniqueHandles( void )
{
	if( !IsD3DEffectValid() ) return;

	// Find effect name

	#ifdef ASURA_ASSERT
	#if ASURA_LOAD_EFFECT_FILES == 1
			const char* const szEffectName = m_szName;
	#else
			const char* const szEffectName = "<Unknown>";
	#endif
	#endif

	// Query effect description

	D3DXEFFECT_DESC xEffectDesc;

	if( FAILED( m_pxD3DEffect->GetDesc( &xEffectDesc ) ) )
	{
		Asura_Assertf( false, ("Failed to retrieve description for effect '%s'.", szEffectName) );
		return;
	}

	// Allocate technique handle array

	m_uNumTechniques = xEffectDesc.Techniques;
	m_pxTechniqueHandles = new D3DXHANDLE[ m_uNumTechniques ];

	if( !m_pxTechniqueHandles )
	{
		Asura_Assert( false, "Ran out of memory trying to allocate technique handles." );
		return;
	}

	// Query technique handles

	for( u_int u = 0; u < m_uNumTechniques; ++u )
	{
		m_pxTechniqueHandles[u] = m_pxD3DEffect->GetTechnique(u);

		Asura_Assertf( m_pxTechniqueHandles[u] != NULL, ("Failed to get technique %u in effect '%s'.",
			u, szEffectName) );
	}
}


/// End of file. ///
