//
// Asura_Direct3D_EffectParameterBlock.cpp
//
// Created by RJF - 11 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Asura_Direct3D_EffectParameterBlock.h"

#include "Rendering/Asura_Direct3D_Effect.h"

//////////////////////////////////////////////////////////////////////////
// Static members
//////////////////////////////////////////////////////////////////////////

#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)

Asura_List<Asura_Direct3D_EffectParameterBlock> Asura_Direct3D_EffectParameterBlock::s_xBlockList;

#endif

//////////////////////////////////////////////////////////////////////////
// Parameter structures
//////////////////////////////////////////////////////////////////////////

Asura_Direct3D_EffectParameterBlock::Parameter::Parameter(const Asura_Char* szParameterName, bool bIsStatic)
: m_szParameterName(szParameterName),
  m_xParameterHandle(NULL),
  m_bIsStatic(bIsStatic)
{
	
}

D3DXHANDLE Asura_Direct3D_EffectParameterBlock::Parameter::GetHandle(Direct3D_Effect* pxEffect)
{
	// If we've got a cached value, use it.
	if(m_xParameterHandle != NULL) return m_xParameterHandle;

	// If we don't have a cached value, calculate the new one, but DON'T STORE IT - we don't do that until we bind.
	return pxEffect->GetParameterByName(NULL, m_szParameterName);
}

//////////////////////////////////////////////////////////////////////////
// Asura_Direct3D_EffectParameterBlock
//////////////////////////////////////////////////////////////////////////

Asura_Direct3D_EffectParameterBlock::Asura_Direct3D_EffectParameterBlock() : m_pxEffect(NULL), m_xStaticParameterBlock(NULL)
{
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
	s_xBlockList.AddToList(this);
#endif
}

Asura_Direct3D_EffectParameterBlock::~Asura_Direct3D_EffectParameterBlock()
{
	// Release the parameter block
	if(m_xStaticParameterBlock && m_pxEffect && m_pxEffect->IsD3DEffectValid())
	{
		m_pxEffect->GetD3DEffect()->DeleteParameterBlock(m_xStaticParameterBlock);
	}

	// Delete any parameters left over
	Asura_List_It<Parameter> xIt(&m_xParameters);
	while(!xIt.Done())
	{
		delete xIt.GetCurrent();
		xIt.DeleteCurrent();
	}

#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
	s_xBlockList.RemoveFromList(this);
#endif
}

void Asura_Direct3D_EffectParameterBlock::Bind(Asura_Direct3D_Effect* pxEffectToBindTo)
{
	Asura_Assert(!IsBound(), "Block is already bound!");
	if(IsBound()) return;

	Asura_Assert(pxEffectToBindTo, "Trying to bind to a null effect!");
	if(!pxEffectToBindTo) return;

	Asura_Assert(pxEffectToBindTo->IsD3DEffectValid(), "Binding to an invalid effect!");
	if(!pxEffectToBindTo->IsD3DEffectValid()) return;

	// We also need a technique set before we can build a static parameter block
	Asura_Assert(pxEffectToBindTo->GetD3DEffect()->GetCurrentTechnique(), "Trying to bind to an effect that has not had its technique set");
	if(!pxEffectToBindTo->GetD3DEffect()->GetCurrentTechnique()) return;

	// OK, we're going to bind to this effect
	m_pxEffect = pxEffectToBindTo;

	// Set up handles for all parameters
	Asura_List_It<Parameter> xIt(&m_xParameters);
	while(!xIt.Done())
	{
		Parameter* pxParam = xIt.GetCurrent();

		pxParam->m_xParameterHandle = pxParam->GetHandle(m_pxEffect->GetD3DEffect());

		xIt.Next();
	}

	// Build a static parameter block for any static parameters we have
	m_xStaticParameterBlock = NULL;
	bool bHasStartedParamBlock = false;
	xIt.Restart();
	while(!xIt.Done())
	{
		Parameter* pxParam = xIt.GetCurrent();

		if(pxParam->m_bIsStatic)
		{
			if(!bHasStartedParamBlock)
			{
				m_pxEffect->GetD3DEffect()->BeginParameterBlock();
				bHasStartedParamBlock = true;
			}
			pxParam->Apply(m_pxEffect->GetD3DEffect());
			
			xIt.Next();
		}
		else
		{
			xIt.Next();
		}
	}
	if(bHasStartedParamBlock)
	{
		m_xStaticParameterBlock = m_pxEffect->GetD3DEffect()->EndParameterBlock();
	}
}

void Asura_Direct3D_EffectParameterBlock::Apply(Asura_Direct3D_Effect* pxEffectToApplyTo /* = NULL */)
{
	if(IsBound())
	{
		if(pxEffectToApplyTo != NULL)
		{
			Asura_Assert(false, "Trying to apply a locked effect block to a specific effect!");
			return;
		}
		pxEffectToApplyTo = m_pxEffect;
	}

	Asura_Assert(pxEffectToApplyTo, "Don't have an effect to apply this block to");
	if(!pxEffectToApplyTo) return;

	// If we have a static block, apply it
	if(m_xStaticParameterBlock != NULL)
	{
		pxEffectToApplyTo->GetD3DEffect()->ApplyParameterBlock(m_xStaticParameterBlock);
	}

	// Apply all remaining (dynamic) parameters
	Asura_List_It<Parameter> xIt(&m_xParameters);
	while(!xIt.Done())
	{
		Parameter* pxParam = xIt.GetCurrent();
		if(!pxParam->m_bIsStatic)
		{
			pxParam->Apply(pxEffectToApplyTo->GetD3DEffect());
		}
		xIt.Next();
	}
}

#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
void Asura_Direct3D_EffectParameterBlock::Tools_RegenerateAll()
{
	Asura_List_It<Asura_Direct3D_EffectParameterBlock> xIt(&s_xBlockList);
	while(!xIt.Done())
	{
		Asura_Direct3D_EffectParameterBlock* pxBlock = xIt.GetCurrent();
		if(pxBlock->IsBound())
		{
			// Store the effect it's currently bound to
			Asura_Direct3D_Effect* pxBoundEffect = pxBlock->m_pxEffect;
			// Forcibly unbind it
			pxBlock->m_pxEffect = NULL;
			Asura_List_It<Parameter> xItParams(&pxBlock->m_xParameters);
			while(!xItParams.Done()) { xItParams.GetCurrent()->m_xParameterHandle = NULL; xItParams.Next();}
			// Rebind
			pxBlock->Bind(pxBoundEffect);
		}
		// No need to rebind blocks that weren't previously bound
		xIt.Next();
	}
}
#endif