//////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_HDR.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_HDR

#include "Rendering/Asura_Direct3D_HDR.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_HalfFloat.h"
#include "Debugging/Asura_DebugText.h"
#include "Interface/Asura_CommandConsole.h"

#if defined(ASURA_TOOLS)
	#include "Interface/Asura_Dialog.h"
	#include "Interface/Asura_Overlay.h"
#endif	// defined(ASURA_TOOLS)

#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// Asura PC
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Effect.h"
#include "Rendering/Asura_Direct3D_Effects_System.h"
#include "Rendering/Asura_Direct3D_RenderTargetManager.h"

#if defined ASURA_TOOLS

	// Tools
	#include "Load&Save/Tools_ModeDataFile.h"

#endif // defined ASURA_TOOLS

// Misc
#include <d3d9.h>

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/
//#define DEBUG_CHECK_HDR_LUMINOSTY

// Working render targets for downsampling
static Asura_Direct3D_RenderTarget* g_pxWorkingRT_RGB = NULL;
static Asura_Direct3D_RenderTarget* g_pxWorkingRT_R_64 = NULL;
static Asura_Direct3D_RenderTarget* g_pxWorkingRT_R_16 = NULL;

// Holds the current luminance value. TODO: splitscreen - use 4 texels here?
// On PC, need to ping-pong between two
static Asura_Direct3D_RenderTarget* g_apxCurrentLuminanceTextures[2] = { NULL, NULL };
static u_int g_uCurrentLuminanceTextureIndex = 0;

#if defined ASURA_TOOLS
	static float* g_pfFixedLuminanceResult = NULL;

	static Direct3D_Texture *g_pxHistogramTempTexture;
	static Direct3D_Surface *g_pxHistogramTempSurface;
	static Direct3D_Surface *g_pxPreToneMapHistogramSurface;
	static Direct3D_Surface *g_pxPostToneMapHistogramSurface;
#endif

// Bits for dealing with the measurement region
const Direct3D_VertexElement axREGION_VERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 8, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	D3DDECL_END()
};

static Direct3D_VertexDeclaration* g_pxRegionVD = NULL;

static bool g_bShowRegion = false;
DVSVAR bool g_bShowHDRValues = false;
DVSVAR bool g_bGrabBuffer = false;

static bool s_bFirstTimeSinceReset = true;

/******************************************************************************/
/* Asura_Direct3D_HDR Static Class Functions                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_HDR::Initialise()
//
// Initialises the HDR rendering system on PC
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_HDR::Initialise()
{
	if(!Asura_Direct3D::D3DDevice)
	{
		return;
	}
	g_pxWorkingRT_RGB = Asura_Direct3D_RenderTargetManager::CreateRenderTarget( 0.25f, 0.25f, D3DFMT_A16B16G16R16F, D3DFMT_UNKNOWN, true );

	g_pxWorkingRT_R_64 = Asura_Direct3D_RenderTargetManager::CreateRenderTarget( 64.0f, 64.0f, D3DFMT_R32F, D3DFMT_UNKNOWN, false );
	g_pxWorkingRT_R_16 = Asura_Direct3D_RenderTargetManager::CreateRenderTarget( 16.0f, 16.0f, D3DFMT_R32F, D3DFMT_UNKNOWN, false );
	for( u_int u = 0; u < 2; ++u )
	{
		g_apxCurrentLuminanceTextures[u] = Asura_Direct3D_RenderTargetManager::CreateRenderTarget( 1.0f, 1.0f, D3DFMT_R32F, D3DFMT_UNKNOWN, false );
		if( !g_apxCurrentLuminanceTextures[u] ) continue;

		Asura_Direct3D::D3DDevice->ColorFill( g_apxCurrentLuminanceTextures[u]->GetSurface(), NULL, D3DCOLOR_XRGB( 0, 0, 0 ) );
	}

	#if defined ASURA_ASSERT
		const HRESULT xResult =
	#endif
		Asura_Direct3D::D3DDevice->CreateVertexDeclaration( axREGION_VERTEX_DECLARATION,
			&g_pxRegionVD );

	Asura_Assert( SUCCEEDED( xResult ), "Failed to create HDR vertex declaration." );

	Asura_Direct3D::AddPreResetCallback(PreReset, NULL);
	Asura_Direct3D::AddPostResetCallback(PostReset, NULL);

	PostReset();

#if defined ASURA_DEBUG_VARIABLE_SYSTEM && !defined ASURA_COMMAND_CONSOLE_READONLY
	Asura_CommandConsole::AddVar( L"Render.HDR.ShowRegion", &g_bShowRegion,
		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.HDR.ShowValues", &g_bShowHDRValues,
		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.HDR.GrabBuffer", &g_bGrabBuffer,
		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
}

bool Asura_Direct3D_HDR::PreReset(void*)
{
#if defined ASURA_TOOLS
	SAFE_RELEASE(g_pxHistogramTempSurface);
	SAFE_RELEASE(g_pxHistogramTempTexture);
	SAFE_RELEASE(g_pxPreToneMapHistogramSurface);
	SAFE_RELEASE(g_pxPostToneMapHistogramSurface);
#endif
	return true;
}

bool Asura_Direct3D_HDR::PostReset(void*)
{
#if defined ASURA_TOOLS

	#if defined ASURA_ASSERT
		const HRESULT xResult =
	#endif
		Asura_Direct3D::D3DDevice->CreateTexture(Asura_Viewport::GetWidth()/2, Asura_Viewport::GetHeight()/2, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &g_pxHistogramTempTexture, NULL);

	Asura_Assertf(g_pxHistogramTempTexture, ("Failed to create HDR histogram temp texture. Error code 0x%X.", xResult));

	if (g_pxHistogramTempTexture)
	{
		g_pxHistogramTempTexture->GetSurfaceLevel(0, &g_pxHistogramTempSurface);
	}

	Asura_Direct3D::D3DDevice->CreateOffscreenPlainSurface(Asura_Viewport::GetWidth()/2, Asura_Viewport::GetHeight()/2, D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, &g_pxPreToneMapHistogramSurface, NULL); 
	Asura_Direct3D::D3DDevice->CreateOffscreenPlainSurface(Asura_Viewport::GetWidth()/2, Asura_Viewport::GetHeight()/2, D3DFMT_A8R8G8B8, D3DPOOL_SYSTEMMEM, &g_pxPostToneMapHistogramSurface, NULL); 

#endif
	s_bFirstTimeSinceReset = true;
	return true;
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_HDR::Shutdown()
//
// Shuts down the HDR rendering system on PC
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_HDR::Shutdown()
{
	SAFE_RELEASE( g_pxRegionVD );

	Asura_Direct3D_RenderTargetManager::DestroyRenderTarget( g_apxCurrentLuminanceTextures[0] );
	Asura_Direct3D_RenderTargetManager::DestroyRenderTarget( g_apxCurrentLuminanceTextures[1] );

	Asura_Direct3D_RenderTargetManager::DestroyRenderTarget( g_pxWorkingRT_R_16 );
	Asura_Direct3D_RenderTargetManager::DestroyRenderTarget( g_pxWorkingRT_R_64 );
	Asura_Direct3D_RenderTargetManager::DestroyRenderTarget( g_pxWorkingRT_RGB );

	Asura_Direct3D::RemovePostResetCallback( PostReset, NULL);
	Asura_Direct3D::RemovePreResetCallback( PreReset, NULL);

	PreReset();

}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_HDR::ApplyToneMapping()
//
// Applies tone mapping to the current scene
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

static void SetTexCoordScale(Asura_Direct3D_Effect* const pxEffect, float fX, float fY)
{
	const D3DXVECTOR4 xScale( fX, fY, 0.0f, 0.0f );
	pxEffect->SetParameterByName( "g_xTexCoordScale", xScale );
}

static void SetViewportScale(Asura_Direct3D_Effect* const pxEffect, const Direct3D_Viewport &xViewport)
{
	const D3DXVECTOR4 xScale( 1.0f / static_cast<float>(xViewport.Width), 1.0f / static_cast<float>(xViewport.Height), 0.0f, 0.0f );
	pxEffect->SetParameterByName( "g_xViewportScale", xScale );
	Asura_Direct3D::D3DDevice->SetViewport( &xViewport );
}

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
static void GrabHalfFloatTexture(float *fDest, Direct3D_Texture*pxTexture, u_int uWidth, u_int uHeight)
{
	D3DLOCKED_RECT xLockedRect;

	D3DSURFACE_DESC xDesc;
	pxTexture->GetLevelDesc(0, &xDesc);
	static u_int uSurfaceWidth = xDesc.Width, uSurfaceHeight = xDesc.Height;
#pragma pack(push)
#pragma pack(1)
	struct SourceStruct
	{
		 Asura_HalfFloat r,g,b,a;
	};
	struct DestStruct
	{
		float r,g,b;
	};
#pragma pack(pop)


	if( xDesc.Format == D3DFMT_A16B16G16R16F && SUCCEEDED( pxTexture->LockRect( 0, &xLockedRect, NULL, D3DLOCK_READONLY  ) ) )
	{
		for (u_int y = 0; y < uHeight; ++y)
		{
			for (u_int x = 0; x < uWidth; ++x)
			{
				DestStruct *pxDest = ((DestStruct *)fDest) + (uHeight - y - 1) * uWidth + x;
				SourceStruct *pxSource = (SourceStruct*)(((char*)( xLockedRect.pBits )) + y * xLockedRect.Pitch) + x;
				pxDest->r = pxSource->r;
				pxDest->g = pxSource->g;
				pxDest->b = pxSource->b;
			}
		}

		pxTexture->UnlockRect(0);
	} else {
		Direct3D_Surface* pxTempSurface = NULL;
		Direct3D_Surface* pxSourceSurface = NULL;
		HRESULT xResult = Asura_Direct3D::D3DDevice->CreateOffscreenPlainSurface(uSurfaceWidth, uSurfaceHeight, D3DFMT_A16B16G16R16F, D3DPOOL_SYSTEMMEM, &pxTempSurface, NULL );
		if( SUCCEEDED( xResult ) )
		{
			if( SUCCEEDED( pxTexture->GetSurfaceLevel(0, &pxSourceSurface) ) )
			{
				if( SUCCEEDED( Asura_Direct3D::D3DDevice->GetRenderTargetData( pxSourceSurface, pxTempSurface ) ) )
				{
					if( SUCCEEDED( pxTempSurface->LockRect( &xLockedRect, NULL, D3DLOCK_READONLY ) ) )
					{
						for (u_int y = 0; y < uHeight; ++y)
						{
							for (u_int x = 0; x < uWidth; ++x)
							{
								DestStruct *pxDest = ((DestStruct *)fDest) + (uHeight - y - 1) * uWidth + x;
								SourceStruct *pxSource = (SourceStruct*)(((char*)( xLockedRect.pBits )) + y * xLockedRect.Pitch) + x;
								pxDest->r = pxSource->r;
								pxDest->g = pxSource->g;
								pxDest->b = pxSource->b;
							}
						}
						pxTempSurface->UnlockRect();
					}
				}
				pxSourceSurface->Release();
			}
			pxTempSurface->Release();
		}
	}

}
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM


static void GrabFloatSurface(float *fDest, Asura_Direct3D_RenderTarget*pRenderTarget, u_int uWidth, u_int uHeight)
{
	Direct3D_Surface* pxTempSurface = NULL;
	static u_int fSurfaceWidth, fSurfaceHeight;
	fSurfaceWidth = pRenderTarget->GetWidth();
	fSurfaceHeight = pRenderTarget->GetHeight();


	Asura_Direct3D::D3DDevice->CreateOffscreenPlainSurface(fSurfaceWidth, fSurfaceHeight, D3DFMT_R32F, D3DPOOL_SYSTEMMEM, &pxTempSurface, NULL );

	if( pxTempSurface )
	{
		HRESULT xResult = Asura_Direct3D::D3DDevice->GetRenderTargetData( pRenderTarget->GetSurface(), pxTempSurface );
		if( SUCCEEDED( xResult ) )
		{
			D3DLOCKED_RECT xLockedRect;
			xResult = pxTempSurface->LockRect( &xLockedRect, NULL, D3DLOCK_READONLY );
			if( SUCCEEDED( xResult ) )
			{
				for (u_int y = 0; y < uHeight; ++y)
					for (u_int x = 0; x < uWidth; ++x)
						fDest[y * uWidth + x] = static_cast< const float* >( xLockedRect.pBits )[y * xLockedRect.Pitch / 4 + x];

				pxTempSurface->UnlockRect();
			}
			else
			{
				Asura_Assert( false, "Failed to lock temp surface for luminance capture." );
			}
		}
		else
		{
			Asura_Assert( false, "Failed to get render target data for luminance capture." );
		}
	}
	else
	{
		Asura_Assert( false, "Failed to create temporary texture for luminance capture." );
	}
	SAFE_RELEASE( pxTempSurface );


}


#if defined ASURA_TOOLS
class Asura_Direct3D_Histogram
{
	u_int m_auR[256];
	u_int m_auG[256];
	u_int m_auB[256];
	u_int m_auL[256];

	u_int m_uRHeight;
	u_int m_uGHeight;
	u_int m_uBHeight;
	u_int m_uLHeight;

	u_int m_uNumPasses;
	u_int m_uPass;

	float m_fRange;

	Asura_Direct3D_HDR::HDR_Histogram_Data m_uHDRData;
	void (*m_pCb)(const Asura_Direct3D_HDR::HDR_Histogram_Data &xData);
public:
	void SetCallback(void (*pCb)(const Asura_Direct3D_HDR::HDR_Histogram_Data &xData), u_int uAppDefined)
	{
		m_pCb = pCb;
		m_uHDRData.m_uAppDefined = uAppDefined;
	}

	Asura_Direct3D_Histogram (u_int uNumPasses) : m_uNumPasses(uNumPasses)
	{
		Reset();
		m_uPass = m_uNumPasses;
		m_fRange = 32.0f;
	}
	void Update(void *pvData, u_int uWidth, u_int uHeight, u_int uPitch)
	{
		if (Done() || !m_pCb)
		{
			return;
		}
		u_int uNumLines = uHeight / m_uNumPasses;
		u_int uStartLine = uNumLines * (m_uPass);
		u_int uEndLine = uNumLines * (m_uPass + 1);
		if (uEndLine > uHeight)
			uEndLine = uHeight;

		for (u_int y = uStartLine ; y < uEndLine; ++y)
		{
			u_char* pxData = (u_char*)(((u_char*)pvData) + y * uPitch);
			for (u_int x = 0; x < uWidth; ++x, (pxData += 4))
			{
				m_uRHeight = Asura_Maths::Max(m_uRHeight, ++m_auR[pxData[2]]);
				m_uGHeight = Asura_Maths::Max(m_uGHeight, ++m_auG[pxData[1]]);
				m_uBHeight = Asura_Maths::Max(m_uBHeight, ++m_auB[pxData[0]]);
				m_uLHeight = Asura_Maths::Max(m_uLHeight, ++m_auL[pxData[3]]);
			}
		}
		m_uPass++;
		if (Done())
		{

			m_uHDRData.m_uRHeight = m_uRHeight;
			m_uHDRData.m_uGHeight = m_uGHeight;
			m_uHDRData.m_uBHeight = m_uBHeight;
			m_uHDRData.m_uLHeight = m_uLHeight;

			memcpy(m_uHDRData.m_auR, m_auR, sizeof(m_auR));
			memcpy(m_uHDRData.m_auG, m_auG, sizeof(m_auG));
			memcpy(m_uHDRData.m_auB, m_auB, sizeof(m_auB));
			memcpy(m_uHDRData.m_auL, m_auL, sizeof(m_auL));

			if (m_pCb)
			{
				m_pCb(m_uHDRData);
			}
		}
	}
	bool Done() {return m_uPass == m_uNumPasses;}
	void Reset()
	{
		memset(m_auR, 0, sizeof(m_auR));
		memset(m_auG, 0, sizeof(m_auR));
		memset(m_auB, 0, sizeof(m_auR));
		memset(m_auL, 0, sizeof(m_auR));
		m_uRHeight = m_uGHeight = m_uBHeight = m_uLHeight = 0;
		m_uPass = 0;
	}
	void SetRange(float fRange)
	{
		m_fRange = fRange > 0.0f ? 1.0f / fRange : 1.0f / 4.0f  ;
		m_uPass = m_uNumPasses;
	}
	float GetRange()
	{
		return m_fRange;

	}



};

#ifdef _DEBUG
static Asura_Direct3D_Histogram g_xPreToneMapHisto(2);
static Asura_Direct3D_Histogram g_xPostToneMapHisto(2);
#else
static Asura_Direct3D_Histogram g_xPreToneMapHisto(1);
static Asura_Direct3D_Histogram g_xPostToneMapHisto(1);
#endif
void Asura_Direct3D_HDR::SetPreToneMapHDRHistogramCallback( void (*pCb)(const HDR_Histogram_Data &xData), u_int uAppDefined)
{
	g_xPreToneMapHisto.SetCallback(pCb, uAppDefined);
}

void Asura_Direct3D_HDR::SetPreToneMapHDRHistogramRange( float fRange)
{
	g_xPreToneMapHisto.SetRange(fRange);
}

void Asura_Direct3D_HDR::SetPostToneMapHDRHistogramCallback( void (*pCb)(const HDR_Histogram_Data &xData), u_int uAppDefined)
{
	g_xPostToneMapHisto.SetCallback(pCb, uAppDefined);
}

void Asura_Direct3D_HDR::SetPostToneMapHDRHistogramRange( float fRange)
{
	g_xPostToneMapHisto.SetRange(fRange);
}


#endif

void SaveFloatImage(const Asura_Char*pFileName, float *pfBuf, u_int uWidth, u_int uHeight)
{
	Asura_Char acHdrBuf[1024];
	u_int uHdrsize = Asura_PrintfToBfr(acHdrBuf, "PF\n%d %d\n-1.000000\n", sizeof(acHdrBuf), false, uWidth, uHeight);
	HANDLE hFile = CreateFile(pFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile != INVALID_HANDLE_VALUE)
	{
		DWORD uWritten = 0;
		WriteFile(hFile, acHdrBuf, uHdrsize, &uWritten, NULL );
		WriteFile(hFile, pfBuf, uWidth * uHeight * 12,&uWritten, NULL );
		CloseHandle(hFile);
	}
}

void Asura_Direct3D_HDR::ApplyToneMapping()
{
	SCOPEDPIXEVENT(( 0, L"Apply tone mapping" ));

	Direct3D_Texture* pxInitialSceneTexture = NULL;
	Direct3D_Viewport xViewport;

	DWORD auMinFilter[2], auMagFilter[2], auMipFilter[2];

	// Initial setup
	D3DPERF_BeginEvent( 0, L"Initial setup" );

	Asura_Direct3D_RenderTarget* const pxDeferredRT = Asura_Direct3D::GetDeferredRenderingTarget();
	pxInitialSceneTexture = pxDeferredRT ? pxDeferredRT->GetTexture() : NULL;

	Asura_Direct3D::D3DDevice->SetVertexDeclaration( Asura_Direct3D::GetFullScreenVD() );
	Asura_Direct3D::D3DDevice->SetStreamSource( 0, Asura_Direct3D::GetFullScreenVB(), 0, sizeof( short ) * 2 );

	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
	Asura_Render::CurrentStates.RequestDitheringMode( true );
	Asura_Render::CurrentStates.RequestDynamicLights( false );
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );

	xViewport.X = 0;
	xViewport.Y = 0;
	xViewport.MinZ = 0.0f;
	xViewport.MaxZ = 1.0f;

	// Preserve initial filtering state
	for( u_int u = 0; u < 2; ++u )
	{
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_MINFILTER, &auMinFilter[u] );
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_MAGFILTER, &auMagFilter[u] );
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_MIPFILTER, &auMipFilter[u] );
	}

	D3DPERF_EndEvent();

	Asura_Direct3D_Effect* const pxEffect = Asura_Direct3D_Effects_System::BeginEffect( "HDR" );
	if( !pxEffect || !pxEffect->IsD3DEffectValid() ) return;

	Asura_CameraData::HDRSettings xSettings;
	Asura_Camera::GetHDRSettings( xSettings );

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (g_bGrabBuffer)
	{
		int iWidth = Asura_Viewport::GetWidth(), iHeight = Asura_Viewport::GetHeight();
		float *pBuf = new float[iWidth * iHeight * 3];
		GrabHalfFloatTexture(pBuf, pxInitialSceneTexture, iWidth, iHeight);

		char cBuf[512];
		static int iCallCount = 1;
		Asura_PrintfToBfr(cBuf, "PreToneMapPC_%dx%d_%d.pfm", sizeof(cBuf), true, iWidth, iHeight, iCallCount++);
		SaveFloatImage(cBuf, pBuf, iWidth, iHeight);

		float fAdaptedLum;
		GrabFloatSurface(&fAdaptedLum, g_apxCurrentLuminanceTextures[ g_uCurrentLuminanceTextureIndex ], 1, 1);

		for (int y = 0; y < iHeight; ++y)
		{
			for (int x = 0; x < iWidth; ++x)
			{
				float r = pBuf[(y * iWidth + x) * 3 + 0];
				float g = pBuf[(y * iWidth + x) * 3 + 1];
				float b = pBuf[(y * iWidth + x) * 3 + 2];

				// Scale to account for "black point"

				const float fGradient =
					xSettings.m_fWhitePoint / (xSettings.m_fWhitePoint - xSettings.m_fBlackPoint );

				const float fOffset = -xSettings.m_fBlackPoint * fGradient;

				r = r * fGradient + fOffset;
				g = g * fGradient + fOffset;
				b = b * fGradient + fOffset;

				float fHDRLuminance = 0.27f * r + 0.67f * g + 0.06f * b;
				const float fDELTA = 0.0001f;
				const float fScaledLuminance = xSettings.m_fKeyValue / (fAdaptedLum + fDELTA) * fHDRLuminance;
				const float fNumerator = fScaledLuminance * (1.0f + fScaledLuminance / (xSettings.m_fWhitePoint * xSettings.m_fWhitePoint));
				const float fDenominator = 1.0f + fScaledLuminance;

				const float fLDRLuminance = fNumerator / fDenominator;

				// Rescale the colours by the ratio of new to old luminance
				r = r * fLDRLuminance / fHDRLuminance;
				g = g * fLDRLuminance / fHDRLuminance;
				b = b * fLDRLuminance / fHDRLuminance;

				pBuf[(y * iWidth + x) * 3 + 0] = r;
				pBuf[(y * iWidth + x) * 3 + 1] = g;
				pBuf[(y * iWidth + x) * 3 + 2] = b;
			}
		}
		Asura_PrintfToBfr(cBuf, "PostToneMapPC_%dx%d_%d.pfm", sizeof(cBuf), true, iWidth, iHeight, iCallCount++);
		SaveFloatImage(cBuf, pBuf, iWidth, iHeight);


		delete[] pBuf;
		g_bGrabBuffer = false;
	}
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

#if defined ASURA_TOOLS
#if _DEBUG
	// Try to make sure that both dont get done in the same frame
	if (!g_xPostToneMapHisto.Done())
#endif
	{
		SCOPEDPIXEVENT(( 0, L"Grab HDR Histogram"));

		if (g_xPreToneMapHisto.Done())
		{
			{
				// This viewport is 1/4 dimensions of the screen
				xViewport.Width = Asura_Viewport::GetWidth() / 2;
				xViewport.Height = Asura_Viewport::GetHeight() / 2;
				SetViewportScale(pxEffect, xViewport);
				SetTexCoordScale(pxEffect, 1.0f / Asura_Viewport::GetWidth(),  1.0f / Asura_Viewport::GetHeight());
			}

			pxEffect->SetParameterByName( "g_xInputSceneTexture", pxInitialSceneTexture );
			pxEffect->SetParameterByName( "g_fHistogramCopyScale", g_xPreToneMapHisto.GetRange());
			Asura_Direct3D::D3DDevice->SetRenderTarget(0, g_pxHistogramTempSurface);
			Asura_Direct3D::D3DDevice->SetRenderTarget(1, NULL);
			Asura_Direct3D::D3DDevice->SetRenderTarget(2, NULL);
			Asura_Direct3D::D3DDevice->SetRenderTarget(3, NULL);

			pxEffect->BeginPass(8);
			Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
			pxEffect->EndPass();

			Asura_Assert(g_pxHistogramTempSurface, "g_pxHistogramTempSurface is NULL.");
			Asura_Assert(g_pxPreToneMapHistogramSurface, "g_pxPreToneMapHistogramSurface is NULL.");

			#if defined ASURA_ASSERT
				const HRESULT xRet =
			#endif
				Asura_Direct3D::D3DDevice->GetRenderTargetData(g_pxHistogramTempSurface, g_pxPreToneMapHistogramSurface);

			Asura_Assertf(SUCCEEDED(xRet) || xRet == D3DERR_DEVICELOST, ("Failed to get render target data (error 0x%X).", xRet));

			g_xPreToneMapHisto.Reset();
		}

		D3DLOCKED_RECT xLockRect;
		if (g_pxPreToneMapHistogramSurface->LockRect(&xLockRect, NULL, D3DLOCK_READONLY) == D3D_OK)
		{
			g_xPreToneMapHisto.Update(xLockRect.pBits, pxDeferredRT->GetWidth()/2, pxDeferredRT->GetHeight()/2, xLockRect.Pitch);
			g_pxPreToneMapHistogramSurface->UnlockRect();
		}

	}
#endif


	// The region is defined in [-1,1] terms relative to the screen height, so it always
	// keeps the same aspect ratio when we change to widescreen

	const float fLeft = xSettings.m_xRegion.MinX * Asura_Camera::GetYOverXAspectRatio();
	const float fRight = xSettings.m_xRegion.MaxX * Asura_Camera::GetYOverXAspectRatio();
	const float fTop = xSettings.m_xRegion.MaxY;
	const float fBottom = xSettings.m_xRegion.MinY;

	Asura_FixedArray< D3DXVECTOR4, 4 > axPoints;
	const UINT uStride = sizeof( axPoints[0] );

	// First downsample to quarter size
	{
		SCOPEDPIXEVENT(( 0, L"Downsample to quarter size" ));

		pxEffect->SetParameterByName( "g_xDownsizeTexture", pxInitialSceneTexture );

		g_pxWorkingRT_RGB->Bind();

		{
			// This viewport is 1/4 dimensions of the screen
			xViewport.Width = Asura_Viewport::GetWidth() / 4;
			xViewport.Height = Asura_Viewport::GetHeight() / 4;
			SetViewportScale(pxEffect, xViewport);
			SetTexCoordScale(pxEffect, 1.0f / Asura_Viewport::GetWidth(),  1.0f / Asura_Viewport::GetHeight());
		}

		// We'll not use a quad covering the whole RT, just the portion that corresponds to the
		// selected measurement region

		{
			Asura_Direct3D::D3DDevice->SetVertexDeclaration( g_pxRegionVD );

			axPoints[0].x = -1;  axPoints[0].y = 1;
			axPoints[1].x = 1; axPoints[1].y = 1;
			axPoints[2].x = 1; axPoints[2].y = -1;
			axPoints[3].x = -1;  axPoints[3].y = -1;

			for( u_int u = 0; u < axPoints.GetSize(); ++u )
			{
				// Don't need texcoords at this stage
				axPoints[u].z = axPoints[u].w = 0.0f;
			}
		}

		pxEffect->BeginPass(0);
		Asura_Direct3D::D3DDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, axPoints, uStride );
		pxEffect->EndPass();
	}

	// Then downsample to 64x64 and calculate average luminance for each 3x3
	{
		SCOPEDPIXEVENT(( 0, L"Downsample to 64x64 and calculate log luminance" ));

		pxEffect->SetParameterByName( "g_xDownsizeTexture", g_pxWorkingRT_RGB->GetTexture() );
		g_pxWorkingRT_R_64->Bind();

		{
			SetTexCoordScale(pxEffect, 1.0f / xViewport.Width, 1.0f / xViewport.Height);
			xViewport.Width = xViewport.Height = 64;
			SetViewportScale(pxEffect, xViewport);
		}

		// Now render to the whole RT, but only sample from a subrect of the texture where we
		// rendered before
		{
			axPoints[0].z = fLeft * 0.5f + 0.5f + (0.5f / (Asura_Viewport::GetWidth() / 4));  axPoints[0].w = 0.5f - fTop * 0.5f  + (0.5f / (Asura_Viewport::GetHeight() / 4));
			axPoints[1].z = fRight * 0.5f + 0.5f + (0.5f / (Asura_Viewport::GetWidth() / 4)); axPoints[1].w = 0.5f - fTop * 0.5f  + (0.5f / (Asura_Viewport::GetHeight() / 4));
			axPoints[2].z = fRight * 0.5f + 0.5f + (0.5f / (Asura_Viewport::GetWidth() / 4)); axPoints[2].w = 0.5f - fBottom * 0.5f  + (0.5f / (Asura_Viewport::GetHeight() / 4));
			axPoints[3].z = fLeft * 0.5f + 0.5f + (0.5f / (Asura_Viewport::GetWidth() / 4));  axPoints[3].w = 0.5f - fBottom * 0.5f  + (0.5f / (Asura_Viewport::GetHeight() / 4));

			// Vertex positions covering the whole of the RT
			axPoints[0].x = -1.0f; axPoints[0].y =  1.0f;
			axPoints[1].x =  1.0f; axPoints[1].y =  1.0f;
			axPoints[2].x =  1.0f; axPoints[2].y = -1.0f;
			axPoints[3].x = -1.0f; axPoints[3].y = -1.0f;
		}

		pxEffect->BeginPass(1);
		Asura_Direct3D::D3DDevice->DrawPrimitiveUP( D3DPT_TRIANGLEFAN, 2, axPoints, uStride );
		pxEffect->EndPass();
	}
#ifdef DEBUG_CHECK_HDR_LUMINOSTY
	float f64Lum = 0;
	{
		float f64x64Surface[64*64];
		memset(f64x64Surface, 0, sizeof(f64x64Surface));
		GrabFloatSurface(f64x64Surface, g_pxWorkingRT_R_64, 64, 64);
		for (int i = 0; i < 64* 64; ++i)
		{
			f64Lum += f64x64Surface[i];
		}
		f64Lum /= 64.0f*64.0f;
	}
	Asura_Assert(!_isnan(f64Lum), (""));
#endif

	// Down to 16x16
	{
		SCOPEDPIXEVENT(( 0, L"Downsample to 16x16" ));

		Asura_Direct3D::D3DDevice->SetVertexDeclaration( Asura_Direct3D::GetFullScreenVD() );
		Asura_Direct3D::D3DDevice->SetStreamSource( 0, Asura_Direct3D::GetFullScreenVB(), 0,
			sizeof( short ) * 2 );

		pxEffect->SetParameterByName( "g_xDownsizeTexture", g_pxWorkingRT_R_64->GetTexture() );
		g_pxWorkingRT_R_16->Bind();

		{
			SetTexCoordScale(pxEffect, 1.0f / xViewport.Width, 1.0f / xViewport.Height);
			xViewport.Width = xViewport.Height = 16;
			SetViewportScale(pxEffect, xViewport);
		}

		// Switch to unfiltered downsampling pass. Can't use filtering on a floating-point texture
		// on some hardware
		pxEffect->BeginPass(2);
		Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
	}

	// Then to 4x4
	{
		SCOPEDPIXEVENT(( 0, L"Downsample to 4x4" ));

		pxEffect->SetParameterByName( "g_xDownsizeTexture", g_pxWorkingRT_R_16->GetTexture() );
		g_pxWorkingRT_R_64->Bind();

		{
			// Now we need to use a 4x4 viewport on the 64x64 RT
			SetTexCoordScale(pxEffect, 1.0f / xViewport.Width, 1.0f / xViewport.Height);
			xViewport.Width = xViewport.Height = 4;
			SetViewportScale(pxEffect, xViewport);
		}

		pxEffect->GetD3DEffect()->CommitChanges();
		Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );

		// End the straight downsampling pass
		pxEffect->EndPass();
	}

	// In the final downsampling step, switch PS to apply the exp()
	{
		SCOPEDPIXEVENT(( 0, L"Downsample to 1x1 and calculate final scene luminance" ));

		pxEffect->SetParameterByName( "g_xDownsizeTexture", g_pxWorkingRT_R_64->GetTexture() );
		g_pxWorkingRT_R_16->Bind();

		{
			// Set viewport to 1x1 on 16x16 RT
			SetTexCoordScale(pxEffect, 1.0f / 64.0f, 1.0f / 64.0f);
			xViewport.Width = xViewport.Height = 1;
			SetViewportScale(pxEffect, xViewport);
		}

		pxEffect->BeginPass(3);
		Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
		pxEffect->EndPass();
	}


	float fFinalLum;
	if (g_bShowHDRValues
#if defined(ASURA_TOOLS)
		|| g_pfFixedLuminanceResult
#endif
		)
	{
		GrabFloatSurface(&fFinalLum, g_pxWorkingRT_R_16, 1, 1);
#if defined(ASURA_TOOLS)
		if (g_pfFixedLuminanceResult)
		{
			*g_pfFixedLuminanceResult = fFinalLum;
			g_pfFixedLuminanceResult = NULL;
		}
#endif
	}


	// Now we've found the scene's luminance value (in the top-left texel of the
	// 16x16 texture), lerp the current value towards it by a small amount
	{
		SCOPEDPIXEVENT(( 0, L"Adapt current luminance" ));

		pxEffect->SetParameterByName( "g_fMinLuminance", xSettings.m_fMinLuminance );
		pxEffect->SetParameterByName( "g_fMaxLuminance", xSettings.m_fMaxLuminance );

		const float fAdaptationFactor = Asura_Maths::Min( xSettings.m_fAdaptationRate * Asura_Timers::GetGameTimeSlice(), 1.0f );
		pxEffect->SetParameterByName( "g_fAdaptationFactor", fAdaptationFactor );

		// Scene luminance will always end up in the 16x16 RT
		pxEffect->SetParameterByName( "g_xSceneLuminanceTexture", g_pxWorkingRT_R_16->GetTexture() );

		// Ping-pong between the two 'current luminance' textures
		Direct3D_Texture* const pxCurrentLuminanceTexture = g_apxCurrentLuminanceTextures[ 1 - g_uCurrentLuminanceTextureIndex ]->GetTexture();
		pxEffect->SetParameterByName( "g_xCurrentLuminanceTexture", pxCurrentLuminanceTexture );

		g_apxCurrentLuminanceTextures[ g_uCurrentLuminanceTextureIndex ]->Bind();
		// Leave viewport at 1x1
		SetTexCoordScale(pxEffect, 1.0f / xViewport.Width, 1.0f / xViewport.Height);
		SetViewportScale(pxEffect, xViewport);

		pxEffect->BeginPass(4 + s_bFirstTimeSinceReset);
		s_bFirstTimeSinceReset = false;
		Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
		pxEffect->EndPass();

		if (g_bShowHDRValues)
		{
			float fAdapted;
			GrabFloatSurface(&fAdapted, g_apxCurrentLuminanceTextures[ g_uCurrentLuminanceTextureIndex ], 1, 1);


#ifdef DEBUG_CHECK_HDR_LUMINOSTY
			Asura_DebugTextPrint(( 10, 10, 0xFFFFFFFF, 0, "HDR %f", f64Lum));
			Asura_DebugTextPrint(( 10, 11, 0xFFFFFFFF, 0, "Final %f %f", exp(f64Lum), fFinalLum ));
			Asura_DebugTextPrint(( 10, 12, 0xFFFFFFFF, 0, "Diff %f", abs(exp(f64Lum) - fFinalLum) ));
			Asura_DebugTextPrint(( 10, 13, 0xFFFFFFFF, 0, "Adapted %f", fAdapted ));

			Asura_Assert(!_isnan(fAdapted), (""));

#else
			Asura_DebugTextPrint(( 10, 10, 0xFFFFFFFF, 0, "HDR %f", log(fFinalLum)));
			Asura_DebugTextPrint(( 10, 11, 0xFFFFFFFF, 0, "Final %f", fFinalLum ));
			Asura_DebugTextPrint(( 10, 12, 0xFFFFFFFF, 0, "Adapted %f", fAdapted ));
#endif

		}

	}


	// Final pass: apply the tone mapping
	{
		SCOPEDPIXEVENT(( 0, L"Apply tone mapping to scene" ));

		// Take adapted luminance value we just calculated
		Direct3D_Texture* const pxCurrentLuminanceTexture = g_apxCurrentLuminanceTextures[ g_uCurrentLuminanceTextureIndex ]->GetTexture();
		pxEffect->SetParameterByName( "g_xCurrentLuminanceTexture", pxCurrentLuminanceTexture );

		pxEffect->SetParameterByName( "g_xInputSceneTexture", pxInitialSceneTexture );
		pxEffect->SetParameterByName( "g_fKeyValue", xSettings.m_fKeyValue );

		const float fWhitePointSq = xSettings.m_fWhitePoint * xSettings.m_fWhitePoint;
		pxEffect->SetParameterByName( "g_fWhitePointSquared", fWhitePointSq );

		// Simple linear rescale on initial sample, to give the concept of "black point"
		{
			Asura_Assert( xSettings.m_fBlackPoint < xSettings.m_fWhitePoint,
				"HDR black point must be < white point." );

			const float fGradient =
				xSettings.m_fWhitePoint / (xSettings.m_fWhitePoint - xSettings.m_fBlackPoint);

			const float fOffset = -xSettings.m_fBlackPoint * fGradient;

			const D3DXVECTOR4 xScaleParams( fGradient, fOffset, 0.0f, 0.0f );

			pxEffect->SetParameterByHash( "g_xScaleParams", 0xFDB170B0, xScaleParams );
		}

		pxEffect->SetParameterByName( "g_bDynamicExposure", xSettings.m_bDynamicExposure );
		pxEffect->SetParameterByName( "g_fFixedLuminance", xSettings.m_fFixedLuminance );


#if defined ASURA_TOOLS
		if (1)
		{
			SCOPEDPIXEVENT(( 0, L"Grab HDR Histogram"));

			if (g_xPostToneMapHisto.Done())
			{
				pxEffect->SetParameterByName( "g_fHistogramCopyScale", g_xPostToneMapHisto.GetRange());
				Asura_Direct3D::D3DDevice->SetRenderTarget(0, g_pxHistogramTempSurface);
				// Reset the viewport
				{
					xViewport.Width = Asura_Viewport::GetWidth()/2;
					xViewport.Height = Asura_Viewport::GetHeight()/2;
					SetViewportScale(pxEffect, xViewport);
					SetTexCoordScale(pxEffect, 1.0f / xViewport.Width, 1.0f / xViewport.Height);
				}

				pxEffect->BeginPass(9);
				Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
				pxEffect->EndPass();

				HRESULT xRet;
				xRet = Asura_Direct3D::D3DDevice->GetRenderTargetData(g_pxHistogramTempSurface, g_pxPostToneMapHistogramSurface);

				Asura_Assertf(SUCCEEDED(xRet) || xRet == D3DERR_DEVICELOST, ("Failed to get render target data (error 0x%X).", xRet));


				g_xPostToneMapHisto.Reset();
			}

			D3DLOCKED_RECT xLockRect;
			if (g_pxPostToneMapHistogramSurface->LockRect(&xLockRect, NULL, D3DLOCK_READONLY) == D3D_OK)
			{
				g_xPostToneMapHisto.Update(xLockRect.pBits, pxDeferredRT->GetWidth()/2, pxDeferredRT->GetHeight()/2, xLockRect.Pitch);
				g_pxPostToneMapHistogramSurface->UnlockRect();
			}

		}
#endif



		// Set the render target to the final LDR setup
		Asura_Direct3D::SetUsingDeferredRenderingTarget( false );
		Asura_Direct3D::BindDefaultTargets();

		// Reset the viewport
		{
			xViewport.Width = Asura_Viewport::GetWidth();
			xViewport.Height = Asura_Viewport::GetHeight();
			SetViewportScale(pxEffect, xViewport);
			SetTexCoordScale(pxEffect, 1.0f / xViewport.Width, 1.0f / xViewport.Height);
		}

		if( Asura_Direct3D::UseGammaCorrection() )
		{
			// Now we're using an 8888 RT, enable sRGB writes
			Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_SRGBWRITEENABLE, TRUE );
		}

		pxEffect->BeginPass(6);
		Asura_Direct3D::D3DDevice->DrawPrimitive( D3DPT_TRIANGLEFAN, 0, 2 );
		pxEffect->EndPass();

		// Next frame, render into the other current luminance texture
		g_uCurrentLuminanceTextureIndex = 1 - g_uCurrentLuminanceTextureIndex;
	}

	// Cleanup
	{
		SCOPEDPIXEVENT(( 0, L"Cleanup" ));


		// Unbind unmanaged textures from the effect, in case the device is reset after this
		pxEffect->SetParameterTextureByName( "g_xDownsizeTexture", NULL );
		pxEffect->SetParameterTextureByName( "g_xCurrentLuminanceTexture", NULL );
		pxEffect->SetParameterTextureByName( "g_xSceneLuminanceTexture", NULL );
		pxEffect->SetParameterTextureByName( "g_xInputSceneTexture", NULL );

		pxEffect->GetD3DEffect()->CommitChanges();
		Asura_Direct3D_Effects_System::EndEffect( *pxEffect );

		// Preserve initial filtering state
		for( u_int u = 0; u < 2; ++u )
		{
			Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_MINFILTER, auMinFilter[u] );
			Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_MAGFILTER, auMagFilter[u] );
			Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_MIPFILTER, auMipFilter[u] );
		}
	}

	// Draw region if enabled
	if( g_bShowRegion )
	{
		SCOPEDPIXEVENT(( 0, L"Draw measurement region" ));

		const float fWidth = static_cast< float >( Asura_Viewport::GetWidth() );
		const float fHeight = static_cast< float >( Asura_Viewport::GetHeight() );
		const float fYOverX = fHeight / fWidth;

		#define FROMCLIP_X(x) ((x) * 0.5f + 0.5f)
		#define FROMCLIP_Y(y) (0.5f - (y) * 0.5f)

		// NB both dimensions specified relative to screen height. Convert from clip space
		const float fLeft = FROMCLIP_X( xSettings.m_xRegion.MinX * fYOverX ) * fWidth;
		const float fRight = FROMCLIP_X( xSettings.m_xRegion.MaxX * fYOverX  ) * fWidth;
		const float fTop = FROMCLIP_Y( xSettings.m_xRegion.MaxY ) * fHeight;
		const float fBottom = FROMCLIP_Y( xSettings.m_xRegion.MinY ) * fHeight;

		#undef FROMCLIP_X
		#undef FROMCLIP_Y

		Asura_Render::Render2DLine( fLeft, fTop, fRight, fTop, 0xFFFFFFFF );
		Asura_Render::Render2DLine( fRight, fTop, fRight, fBottom, 0xFFFFFFFF );
		Asura_Render::Render2DLine( fRight, fBottom, fLeft, fBottom, 0xFFFFFFFF );
		Asura_Render::Render2DLine( fLeft, fBottom, fLeft, fTop, 0xFFFFFFFF );
	}
}

#if defined ASURA_TOOLS

void Asura_Direct3D_HDR::CaptureFixedLuminance( float& fResult )
{
	g_pfFixedLuminanceResult = &fResult;
}

bool* Asura_Direct3D_HDR::GetShowRegionVar()
{
	return &g_bShowRegion;
}

#endif // defined ASURA_TOOLS

#endif // defined ASURA_USE_HDR
