//
// Asura_Direct3D_ShaderCompile.cpp
//
// Created by AS  - 8/10/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#if defined(ASURA_USE_MATERIAL_RESPONSE_SYSTEM)

#include "Core/Asura_FixedArray.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_TextureFlags.h"
#include "Rendering/Asura_Direct3D_Effect.h"
#include "Rendering/Asura_Direct3D_MaterialResponse.h"
#ifdef ASURA_TOOLS
#include "Materials/Tools_MaterialResponse_System.h"
#endif

#include "Asura_Direct3D_MRPShaderCompile.h"

#if ((ASURA_LOAD_EFFECT_FILES == 1) && !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)) || defined(ASURA_TOOLS)

Asura_TemplateHashTree< ID3DXBuffer* > Asura_Direct3D_MRPShaderCompile::s_xShaderTree;
bool Asura_Direct3D_MRPShaderCompile::s_bLoadShadersFromFile = false;;

////////////////////////////////////////////////////////////////////////////////
//
// ResourceIncludeHandler
//
// Implements the ID3DXInclude interface to handle #includes in the shaders at
// export time. We'll load them from the resource package
//
////////////////////////////////////////////////////////////////////////////////

class ResourceIncludeHandler : public ID3DXInclude
{
public:

	STDMETHOD( Open )( THIS_ D3DXINCLUDE_TYPE eIncludeType, LPCSTR szFileName, LPCVOID pParentData,
		LPCVOID* ppData, UINT* puBytes );

	STDMETHOD( Close )( THIS_ LPCVOID pData );
};


////////////////////////////////////////////////////////////////////////////////
//
// ResourceIncludeHandler::Open()
//
// Called by the shader compiler to read the contents of a #included file. We'll
// point the compiler at the data from the resource package
//
////////////////////////////////////////////////////////////////////////////////

HRESULT ResourceIncludeHandler::Open( D3DXINCLUDE_TYPE eIncludeType, LPCSTR szFileName,
									 LPCVOID pParentData, LPCVOID* ppData, UINT* puBytes )
{
	// Must use RT_RCDATA instead of "RCDATA". And we can't use a custom name because
	// D3DXCompileShaderFromResource() only looks for RCDATA
	const HRSRC xResInfoHandle = FindResource( NULL, szFileName, RT_RCDATA );
	Asura_Assertf( xResInfoHandle, ("Failed to find resource '%s'.", szFileName) );
	if( !xResInfoHandle ) return E_FAIL;

	const HGLOBAL xResHandle = LoadResource( NULL, xResInfoHandle );
	Asura_Assertf( xResHandle, ("Failed to load resource '%s'.", szFileName) );
	if( !xResHandle ) return E_FAIL;

	const LPVOID pResData = LockResource( xResHandle );
	Asura_Assertf( pResData, ("Failed to lock resource '%s'.", szFileName) );
	if( !pResData ) return E_FAIL;

	const DWORD uResSize = SizeofResource( NULL, xResInfoHandle );

	*ppData = pResData;
	*puBytes = uResSize;

	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////
//
// ResourceIncludeHandler::Close
//
// Called by the effects framework to close an include file previously opened
//
////////////////////////////////////////////////////////////////////////////////

HRESULT ResourceIncludeHandler::Close( LPCVOID pData )
{
	// Nothing to do. The compiler was pointed directly at the resource data
	return S_OK;
}


////////////////////////////////////////////////////////////////////////////////
//
// SetPercentageBits()
//
// Helper function to set the bit patterns that represent a 'percentage'-type
// define (e.g. colourise diffuse)
//
////////////////////////////////////////////////////////////////////////////////
#ifdef ASURA_TOOLS
void Asura_Direct3D_MRPShaderCompile::SetPercentageBits( const float fValue, Asura_Hash_ID& uHash, u_int& uBit )
{
	// Tools materials store values as percentages. Use two bits in the hash to represent which
	// define we used

	if( fValue <= 1.0f )
	{
		// Don't set the two bits
		uBit <<= 2;
	}
	else if( fValue >= 99.0f )
	{
		// Set the bits to 01
		uHash |= uBit;
		uBit <<= 2;
	}
	else
	{
		// Set the bits to 11
		uHash |= uBit;
		uBit <<= 1;
		uHash |= uBit;
		uBit <<= 1;
	}
}
#endif

#ifdef ASURA_TOOLS
void Asura_Direct3D_MRPShaderCompile::AddMRPShader( const Asura_Hash_ID uHash,
	const Asura_Hash_ID uMaterial1, const Asura_Hash_ID uMaterial2 )
{
	// Can be at most 32 macros from the bitfield, plus NOT_EFFECT and NULL terminator
	D3DXMACRO axMacros[34];

	bool bTmpSkin;
	bool bDeferred;
	if( !GetMRPSMacrosFromHash( uHash, axMacros, bTmpSkin, bDeferred ) )
	{
		const Tools_MaterialResponse* const pxMaterial1 =
			Tools_MaterialResponse_System::GetMaterialByHashID( uMaterial1 );
		const Tools_MaterialResponse* const pxMaterial2 =
			Tools_MaterialResponse_System::GetMaterialByHashID( uMaterial2 );

		const char* const szName1 = (pxMaterial1 ? pxMaterial1->GetName().GetString() : NULL);
		const char* const szName2 = (pxMaterial2 ? pxMaterial2->GetName().GetString() : NULL);

		Tools_Level_Export::AddMessage( Tools_Level_Export::MT_ERROR,
			"Materials '%s' and '%s' have been dual-textured together, and consume too many texture slots. Max allowed is 14.",
			szName1, szName2 );
	}

	const bool bSkin = ((uHash & uASURA_MRPS_HASH_BITMASK_SKIN) != 0);
	Asura_Assert( bSkin == bTmpSkin, "Incorrect bit mapping between GetMRPShaderHash and GetMRPSMacrosFromHash" );

	bool bDidExist = false;
	ID3DXBuffer** const ppxCompiledShader = s_xShaderTree.AddOrFindNode( uHash, &bDidExist );

	if( !bDidExist )
	{
		Asura_Assert( ppxCompiledShader, "Failed to add shader to export tree." );

		if( ppxCompiledShader )
		{
			*ppxCompiledShader = CompileMRPShader( axMacros, bSkin, bDeferred );
		}
	}
}
#endif

void Asura_Direct3D_MRPShaderCompile::RecompileShadersTraverseFunction(Asura_Hash_ID uHash, ID3DXBuffer** pxData)
{
	// Can be at most 32 macros from the bitfield, plus NOT_EFFECT and NULL terminator
	D3DXMACRO axMacros[34];

	bool bSkin;
	bool bDeferred;
	if( !GetMRPSMacrosFromHash( uHash, axMacros, bSkin, bDeferred ) )
	{
	}

	if (*pxData)
	{
		(*pxData)->Release();

	}

	*pxData = CompileMRPShader( axMacros, bSkin, bDeferred );

	if (*pxData)
	{
		Asura_Direct3D_MaterialResponse::ReplacePixelShader(uHash, (*pxData)->GetBufferPointer());
	}

}


bool Asura_Direct3D_MRPShaderCompile::RecompileShaders(void* pvContext)
{
	s_xShaderTree.Traverse(RecompileShadersTraverseFunction);
	return true;
}

void Asura_Direct3D_MRPShaderCompile::SetLoadShadersFromFile( bool bLoadShadersFromFile ) 
{ 
	if (s_bLoadShadersFromFile != bLoadShadersFromFile)
	{
		if ((s_bLoadShadersFromFile = bLoadShadersFromFile) == true)
		{
			Asura_Direct3D::AddPostResetCallback(RecompileShaders, NULL);
		}
		else
		{
			Asura_Direct3D::RemovePostResetCallback(RecompileShaders, NULL);
		}
	}
}


void Asura_Direct3D_MRPShaderCompile::AddMRPShader( Asura_Hash_ID uHash, void **ppData, u_int *puShaderSize )
{
	// Can be at most 32 macros from the bitfield, plus NOT_EFFECT and NULL terminator
	D3DXMACRO axMacros[34];

	bool bSkin;
	bool bDeferred;
	if( !GetMRPSMacrosFromHash( uHash, axMacros, bSkin, bDeferred ) )
	{
	}

	bool bDidExist = false;
	ID3DXBuffer** const ppxCompiledShader = s_xShaderTree.AddOrFindNode( uHash, &bDidExist );

	if( !bDidExist )
	{
		Asura_Assert( ppxCompiledShader, "Failed to add shader to export tree." );

		if( ppxCompiledShader )
		{
			*ppxCompiledShader = CompileMRPShader( axMacros, bSkin, bDeferred );
		}
	}
	if (*ppxCompiledShader)
	{
		if (ppData) *ppData = (*ppxCompiledShader)->GetBufferPointer();
		if (puShaderSize) *puShaderSize = (*ppxCompiledShader)->GetBufferSize();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_MRPShaderCompile::GetMRPSMacrosFromHash()
//
// Determines the macros needed to compile the material response pixel shader
// for a given hash ID. Reports back whether there were too many texture slots
// required to compile them all in
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Direct3D_MRPShaderCompile::GetMRPSMacrosFromHash( const Asura_Hash_ID uHash, D3DXMACRO* const pxMacros, bool& bSkin, bool& bDeferred )
{
	// Keep the following name array up to date if number of slots changes
	ASURA_STATIC_ASSERT( Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS * 2 == 20 );

	pxMacros[ 0].Name = "USE_ALBEDO_1";
	pxMacros[ 1].Name = "USE_ALBEDO_2";
	pxMacros[ 2].Name = "USE_NORMAL_1";
	pxMacros[ 3].Name = "USE_NORMAL_2";
	pxMacros[ 4].Name = "USE_SPECULAR_1";
	pxMacros[ 5].Name = "USE_SPECULAR_2";
	pxMacros[ 6].Name = "USE_TRANSPARENCY_1";
	pxMacros[ 7].Name = "USE_TRANSPARENCY_2";
	pxMacros[ 8].Name = "USE_SELFILLUMINATION_1";
	pxMacros[ 9].Name = "USE_SELFILLUMINATION_2";
	pxMacros[10].Name = "USE_ENVIRONMENT_1";
	pxMacros[11].Name = "USE_ENVIRONMENT_2";
	pxMacros[12].Name = "USE_REFLECTIVITY_1";
	pxMacros[13].Name = "USE_REFLECTIVITY_2";
	pxMacros[14].Name = "USE_PARALLAX_1";
	pxMacros[15].Name = "USE_PARALLAX_2";
	pxMacros[16].Name = "USE_ALBEDO_DETAIL_MAP_1";
	pxMacros[17].Name = "USE_ALBEDO_DETAIL_MAP_2";
	pxMacros[18].Name = "USE_NORMAL_DETAIL_MAP_1";
	pxMacros[19].Name = "USE_NORMAL_DETAIL_MAP_2";

	u_int uBitMask = 1, uNumTextures = 0;

	// Texture slots. Need one for the materials data textures, and another for the fog in case of
	// transparent strips. That means we can only support 14 user-defined texture slots across both
	// materials with shader model 3.0
	const u_int uMAX_NUM_TEXTURES = 14;

	for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS * 2; ++u )
	{
		if( uHash & uBitMask )
		{
			pxMacros[u].Definition = (uNumTextures <= uMAX_NUM_TEXTURES ? "true" : "false");
			++uNumTextures;
		}
		else
		{
			pxMacros[u].Definition = "false";
		}

		uBitMask <<= 1;
	}

	// Colourise diffuse
	{
		const bool bBit1 = (uHash & uBitMask) != 0;
		uBitMask <<= 1;
		const bool bBit2 = (uHash & uBitMask) != 0;
		uBitMask <<= 1;

		pxMacros[20].Name = "COLOURISE_DIFFUSE";
		pxMacros[20].Definition = (bBit1 && bBit2) ? "2" : (bBit1 && !bBit2) ? "1" : "0";
	}

	// Colourise specular
	{
		const bool bBit1 = (uHash & uBitMask) != 0;
		uBitMask <<= 1;
		const bool bBit2 = (uHash & uBitMask) != 0;
		uBitMask <<= 1;

		pxMacros[21].Name = "COLOURISE_SPECULAR";
		pxMacros[21].Definition = (bBit1 && bBit2) ? "2" : (bBit1 && !bBit2) ? "1" : "0";
	}

	// Metalicity
	{
		const bool bBit1 = (uHash & uBitMask) != 0;
		uBitMask <<= 1;
		const bool bBit2 = (uHash & uBitMask) != 0;
		uBitMask <<= 1;

		pxMacros[22].Name = "METALICITY";
		pxMacros[22].Definition = (bBit1 && bBit2) ? "2" : (bBit1 && !bBit2) ? "1" : "0";
	}

	pxMacros[23].Name = "BLEND_INTO_ALBEDO";
	pxMacros[23].Definition = ((uHash & uBitMask) != 0 ? "true" : "false");
	uBitMask <<= 1;

	pxMacros[24].Name = "IGNORE_SPECULAR";
	pxMacros[24].Definition = ((uHash & uBitMask) != 0 ? "true" : "false");
	uBitMask <<= 1;

	// Switch off effect gubbins like sampler states and static branching
	pxMacros[25].Name = "NOT_EFFECT";
	pxMacros[25].Definition = NULL;

	// Macro array must be NULL-terminated
	pxMacros[26].Name = NULL;
	pxMacros[26].Definition = NULL;

	bSkin = (uBitMask & uHash) != 0;
	uBitMask <<= 1;

	bDeferred = (uBitMask & uHash) == 0;
	uBitMask <<= 1;

	return uNumTextures <= uMAX_NUM_TEXTURES;
}

////////////////////////////////////////////////////////////////////////////////
//
// Tools_PC_Export::CompileMRPShader()
//
// Compiles the material response pixel shader with a given set of macros
//
////////////////////////////////////////////////////////////////////////////////

ID3DXBuffer* Asura_Direct3D_MRPShaderCompile::CompileMRPShader( const D3DXMACRO* const pxMacros, const bool bSkin, const bool bDeferred )
{
	ID3DXBuffer* pxCompiledShader = NULL;
	ID3DXBuffer* pxErrors = NULL;

	const char* const szFilename = (bSkin ? "Skin_PS.ps" : "Env_MaterialResponse.fx");

	// Not exporting transparent env shaders yet
	const char* const szEntrypoint =
		(bDeferred || !bSkin ? "PShader_Deferred" : "PShader_Forward");


	HRESULT xResult;
#if ASURA_LOAD_EFFECT_FILES == 1 && !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
#ifdef ASURA_TOOLS
	if (s_bLoadShadersFromFile)
#else
	if (1)
#endif
	{
		const char* const szShaderPath = Asura_Direct3D_Effect::GetShadersPath();
		Asura_String xShaderFullName( szShaderPath ? szShaderPath : "" );
#if defined(ASURA_PROJECT_SHADER_PATH)
// 		const char* const szProjectShaderPath = Asura_Direct3D_Effect::GetProjectShadersPath();
 
// 		Asura_String xProjectShaderFullName( szProjectShaderPath ? szProjectShaderPath : "" );
// 		Asura_String xOtherShaderFullName( "Shaders/" );
#endif
		xShaderFullName += "Shaders/" + Asura_String(szFilename);

		xResult = D3DXCompileShaderFromFile( xShaderFullName.GetString(), pxMacros, NULL, szEntrypoint, "ps_3_0", 0, &pxCompiledShader, &pxErrors, NULL);

		//Todo: Maybe should try alternative locations?
	}
	else
#endif
	{
		ResourceIncludeHandler xIncludeHandler;
		xResult = D3DXCompileShaderFromResource( NULL, szFilename,
			pxMacros, &xIncludeHandler, szEntrypoint, "ps_3_0", 0, &pxCompiledShader, &pxErrors,
			NULL );

	}

	if( FAILED( xResult ) )
	{
		if(pxErrors == NULL)
		{
			Asura_Assert( false, "Failed to compile material response pixel shader. No errors reported!");
		}
		else
		{
			Asura_Assertf( false, ("Failed to compile material response pixel shader. Errors:\n%s",
				static_cast< const char* >( pxErrors->GetBufferPointer() )) );
		}
	}

	SAFE_RELEASE( pxErrors );

	return pxCompiledShader;
}

#endif
#endif