/**
*
* \file Asura_Direct3D_Shaders.cpp
*
* Created by Kevin Lea - 12/07/2002 14:37:57
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

// This file's header
#include "Rendering/Asura_Direct3D_Shaders.h"

// Asura headers
#include "Core/Asura_PlatformBasics.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Direct3D_Caps.h"
#include "Rendering/Asura_Render.h"

// Static members
bool								Asura_Direct3D_Shaders::s_bBeenInitialised = false;
bool								Asura_Direct3D_Shaders::s_bShadersReady = false;
// Vertex shaders
Asura_Direct3D_VS_Definition		Asura_Direct3D_Shaders::s_axVertexShaderDefs[ASURA_D3D_NUMBER_OF_VERTEXSHADERS];
Asura_Direct3D_VS_Storage			Asura_Direct3D_Shaders::s_axVertexShaders[ASURA_D3D_NUMBER_OF_VERTEXSHADERS];
ASURA_D3D_VERTEXSHADER_ID			Asura_Direct3D_Shaders::s_eCurrentVertexShader = ASURA_D3D_NUMBER_OF_VERTEXSHADERS;

// Vertex formats
Direct3D_VertexElement*				Asura_Direct3D_Shaders::s_apxVertexDeclElements[ASURA_D3D_NUMBER_OF_VERTEX_FORMATS];
Direct3D_VertexDeclaration*			Asura_Direct3D_Shaders::s_apxVertexDeclarations[ASURA_D3D_NUMBER_OF_VERTEX_FORMATS];
unsigned int						Asura_Direct3D_Shaders::s_uVertexDeclCount = 0;
unsigned int						Asura_Direct3D_Shaders::s_uCurrentVertexDecl = ASURA_D3D_NUMBER_OF_VERTEX_FORMATS;

// Pixel shaders
//const char*							Asura_Direct3D_Shaders::s_aszPixelShaderCode[ASURA_D3D_NUMBER_OF_PIXELSHADERS];
//float								Asura_Direct3D_Shaders::s_afPixelShaderVersion[ASURA_D3D_NUMBER_OF_PIXELSHADERS];
Asura_Direct3D_PS_Definition		Asura_Direct3D_Shaders::s_axPixelShaderDef[ASURA_D3D_NUMBER_OF_PIXELSHADERS];
Direct3D_PixelShader*				Asura_Direct3D_Shaders::s_apxPixelShaders[ASURA_D3D_NUMBER_OF_PIXELSHADERS];
ASURA_D3D_PIXELSHADER_ID			Asura_Direct3D_Shaders::s_eCurrentPixelShader = ASURA_D3D_NUMBER_OF_PIXELSHADERS;

// Local Static Data
static const char* SHADERLOGFILE = "Logfiles\\ShaderLog.txt";

void Asura_Direct3D_Shaders::Initialise(void)
{
	Asura_Assert(!s_bBeenInitialised,"Already initialised!");
	
	s_bBeenInitialised = true;	
	
	CreateShaders();
}

void Asura_Direct3D_Shaders::Shutdown(void)
{
	Asura_Assert( NULL == Asura_Direct3D::D3DDevice || s_bBeenInitialised,"Shutdown called without Initialise");

	ReleaseShaders();

	s_bBeenInitialised = false;
}

int Asura_Direct3D_Shaders::RegisterVertexFormat(Direct3D_VertexElement* pxElements)
{
	Asura_Assert(s_uVertexDeclCount<ASURA_D3D_NUMBER_OF_VERTEX_FORMATS, "Maximum number of vertex formats reached!");
	Asura_Assert(pxElements, "Null pointer passed in!");

	// Scan existing declarations to see if this is a duplicate
	for (int i=0; i!=static_cast<int>(s_uVertexDeclCount); ++i)
	{
		// In theory it should be enough to compare pointers, but just to be safe
		// let's do a deep compare.
		Direct3D_VertexElement* pLeft = s_apxVertexDeclElements[i];
		Direct3D_VertexElement* pRight = pxElements;
		bool bIdentical = true;

		while (pLeft && pLeft->Stream != 0xFF && pRight && pRight->Stream != 0xFF)
		{
			if (	pLeft->Stream		!=	pRight->Stream		||
					pLeft->Offset		!=	pRight->Offset		||
					pLeft->Type			!=	pRight->Type		||
					pLeft->Method		!=	pRight->Method		||
					pLeft->Usage		!=	pRight->Usage		||
					pLeft->UsageIndex	!=	pRight->UsageIndex)
			{
				bIdentical = false;
				break;
			}

			++pLeft;
			++pRight;
		}

		if (!pLeft || !pRight || pLeft->Stream != pRight->Stream)
		{
			bIdentical = false;
		}

		if (bIdentical)
		{
			return i;
		}
	}

	// If we're here, no duplicate was found. Register it.
	s_apxVertexDeclElements[s_uVertexDeclCount++] = pxElements;

	return s_uVertexDeclCount-1;
}

void Asura_Direct3D_Shaders::CreateShaders(void)
{
	Asura_Assert(s_bBeenInitialised,"Not Initialised!");

	// Reset 'last used' values
	ClearLastShader();

	s_bShadersReady = true;

	// Assemble all pixel shaders
	for (int i = 0; i < ASURA_D3D_NUMBER_OF_PIXELSHADERS; ++i)
	{
		LPD3DXBUFFER pCode;
		LPD3DXBUFFER pErrors;
		
		const Asura_Direct3D_PS_Definition& xPSDefinition = s_axPixelShaderDef[ i];
		const char* pszPixelShaderCode = xPSDefinition.m_szCode;
		
		Asura_Assertf( NULL != pszPixelShaderCode,("Undefined pixel shader ID %d", i));
		if ( NULL == pszPixelShaderCode)
		{
			s_apxPixelShaders[i] = 0;
			continue;
		}

		// check this pixel shader version - if it's more than the hardware can handle, don't bother trying it
		if( xPSDefinition.m_fVersion > Asura_Direct3D_Caps::GetHardwarePixelShaderVersion())
		{
			s_apxPixelShaders[i] = 0;
			Asura_OutputToDebugger(("Skipping pixel shader ID %i which requires version %f (hardware is %f)\n", i, xPSDefinition.m_fVersion, Asura_Direct3D_Caps::GetHardwarePixelShaderVersion()));
			continue;
		}

		HRESULT iRes = E_FAIL;
		switch( xPSDefinition.m_eCodeType)
		{
			case ASURA_D3D_PS_CODE_TYPE_ASM:
			{
				iRes = D3DXAssembleShader
				(
					pszPixelShaderCode,
					strlen(pszPixelShaderCode),
					0,
					0,
		#ifdef ASURA_DIRECT3D_DEBUG_PS
									D3DXSHADER_DEBUG,
		#else
									0,
		#endif
					&pCode,
					&pErrors
				);
				break;
			}
			case ASURA_D3D_PS_CODE_TYPE_HLSL:
			{
				char* pszProfile = NULL;
				if( 2.0f == xPSDefinition.m_fVersion)
				{
					pszProfile = "ps_2_0";
				}
				if( 3.0f == xPSDefinition.m_fVersion)
				{
					pszProfile = "ps_3_0";
				}
				Asura_Assertf( NULL != pszProfile, ("Shader Version %f is not know for HLSL", xPSDefinition.m_fVersion));
				const char* const pszEntryFuncation = xPSDefinition.m_pszEntryFunction;
				
				iRes =	D3DXCompileShader
				(
					pszPixelShaderCode,
					strlen( pszPixelShaderCode),
					NULL,
					NULL,
					pszEntryFuncation,
					pszProfile,
#ifdef ASURA_DIRECT3D_DEBUG_PS
					D3DXSHADER_DEBUG,
					//D3DXSHADER_DEBUG | D3DXSHADER_SKIPOPTIMIZATION,
#else
					0,
#endif
					&pCode, 
					&pErrors,
					NULL
				);
				break;
			}
			default:
			{
				Asura_Assert( false, "Unknown Pixel Shader Code Type");
				break;
			}
		}

		if(FAILED(iRes))
		{
			if (pErrors)
			{
				#if defined ASURA_ASSERT
					char* szErrorMessage = static_cast<char*>(pErrors->GetBufferPointer()); 
						
					Asura_Logfile(( SHADERLOGFILE, "%s\n%s\n", szErrorMessage, pszPixelShaderCode));
				#endif

				Asura_Assertf( false, ("Errors in pixel shader %d:\n%s", i, szErrorMessage) );
			}
			else
			{
				if ( D3DXERR_INVALIDDATA == iRes )
				{
					Asura_Assert( false, "Bad pixel shader code." );
				}
				else
				{
					Asura_Assert(false, "Unknown error while assembling pixel shader");
				}
			}
			s_apxPixelShaders[i] = 0;
		}
		else
		{
			Asura_Assert(pCode, "Shader failed to assemble, but the return code SUCCEEDED()");

			if(pCode && pErrors)
			{
				#if defined ASURA_ASSERT
					char* szErrorMessage = static_cast<char*>(pErrors->GetBufferPointer());
				#endif

				Asura_Assertf( false, ("Warnings from pixel shader %d:\n%s", i, szErrorMessage) );
			}

			Asura_Direct3D::D3DDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(), &s_apxPixelShaders[i]);
		}

		if(pCode)pCode->Release();
		if(pErrors)pErrors->Release();
	}

	// VGP 15Feb2006 Create all vertex declarations
	for (int i=0; i!=static_cast<int>(s_uVertexDeclCount); ++i)
	{
		#if defined ASURA_ASSERT
			HRESULT iRes =
		#endif
			Asura_Direct3D::D3DDevice->CreateVertexDeclaration( s_apxVertexDeclElements[i], &(s_apxVertexDeclarations[i]) );

		Asura_Assert( D3D_OK == iRes, "Error Creating Vertex Declaration" );
	}

	// Assemble all vertex shaders
	for (int i=0; i<ASURA_D3D_NUMBER_OF_VERTEXSHADERS; i++)
	{
		LPD3DXBUFFER pCode = 0;

		Asura_Assertf(s_axVertexShaderDefs[i].m_szCode,("Undefined Vertex eShader ID %d", i));
		if (!s_axVertexShaderDefs[i].m_szCode)
		{
			s_axVertexShaders[i].m_pxVertexShader = 0;
//DeadCode VGP 15/02/2006 			s_axVertexShaders[i].m_pxVertexShaderDecl = 0;
			s_axVertexShaders[i].m_uVertexDeclIndex = ASURA_D3D_NUMBER_OF_VERTEX_FORMATS;
			continue;
		}
/*		HRESULT iRes =	D3DXCompileShader
						(
							s_axVertexShaderDefs[i].m_szCode,
							strlen(s_axVertexShaderDefs[i].m_szCode),
							0, // Defines
							0, // Include
							"Function name",
							"vs_2_0",
							0, // Flags
							&pCode,
							&pxBuffer,
							0
						);
*/
		LPD3DXBUFFER pxBuffer;
		
		HRESULT iRes = E_FAIL;
		Asura_Direct3D_VS_Definition& xVSDefinition = s_axVertexShaderDefs[i];
		switch( xVSDefinition.m_eCodeType)
		{
			case ASURA_D3D_VS_CODE_TYPE_ASM:
			{
				iRes =	D3DXAssembleShader
				(
					xVSDefinition.m_szCode,
					strlen( xVSDefinition.m_szCode),
					0,
					0,
#ifdef ASURA_DIRECT3D_DEBUG_VS
					D3DXSHADER_DEBUG,
#else
					0,
#endif
					&pCode, 
					&pxBuffer
				);
				break;
			}
			case ASURA_D3D_VS_CODE_TYPE_HLSL:
			{
				char* pszProfile = NULL;
				if( 1.1f == xVSDefinition.m_fVersion)
				{
					pszProfile = "vs_1_1";
				}
				if( 2.0f == xVSDefinition.m_fVersion)
				{
					pszProfile = "vs_2_0";
				}
				if( 3.0f == xVSDefinition.m_fVersion)
				{
					pszProfile = "vs_3_0";
				}
				Asura_Assertf( NULL != pszProfile, ("Shader Version %f is not know for HLSL", xVSDefinition.m_fVersion));
				const char* const pszEntryFuncation = xVSDefinition.m_pszEntryFunction;
				
				iRes =	D3DXCompileShader
				(
					xVSDefinition.m_szCode,
					strlen( xVSDefinition.m_szCode),
					NULL,
					NULL,
					pszEntryFuncation,
					pszProfile,
#ifdef ASURA_DIRECT3D_DEBUG_VS
					D3DXSHADER_DEBUG,
					//D3DXSHADER_DEBUG | D3DXSHADER_SKIPOPTIMIZATION,
#else
					0,
#endif
					&pCode, 
					&pxBuffer,
					NULL
				);
				break;
			}
			default:
			{
				Asura_Assert( false, "Unknown Vertex Shader Code Type");
				break;
			}
		}

		if( FAILED(iRes))
        {
			// DLH - 27 March 2006 11:01:21 - Commented out IF statement as ASM and HLSL compiler will returned valid Buffer with other values as well.
            //if( D3DXERR_INVALIDDATA == iRes )
            {
				if (pxBuffer)
				{
					#if defined ASURA_ASSERT
						char* szErrorMessage = static_cast<char*>(pxBuffer->GetBufferPointer()); 
						
						Asura_Logfile(( SHADERLOGFILE, "%s\n%s\n", szErrorMessage, xVSDefinition.m_szCode));
					#endif

					Asura_Assertf( false, ("Errors in vertex shader %d:\n%s", i, szErrorMessage) );
				}
				else
				{
					Asura_Assert( false, "Bad vertex shader code." );
				}
            }
            //else
            {
                //Asura_Assert( false, "Failed to create vertex shader." );
            }

			s_axVertexShaders[i].m_pxVertexShader = 0;
//DeadCode VGP 15/02/2006 			s_axVertexShaders[i].m_pxVertexShaderDecl = 0;
			s_axVertexShaders[i].m_uVertexDeclIndex = ASURA_D3D_NUMBER_OF_VERTEX_FORMATS;
        }
		else
		{
			Asura_Assert(pCode, "Shader assembly succeeded, but we don't have any output code!");
			
			if(pCode && pxBuffer)
			{
				#if defined ASURA_ASSERT
					char* szErrorMessage = static_cast<char*>(pxBuffer->GetBufferPointer());
						
					Asura_Logfile(( SHADERLOGFILE, "%s\n%s\n", szErrorMessage, xVSDefinition.m_szCode));
				#endif

				Asura_Assertf( false, ("Warnings from vertex shader %d:\n%s", i, szErrorMessage) );
			}
			
			// Create declaration
//DeadCode VGP 15/02/2006 			iRes = Asura_Direct3D::D3DDevice->CreateVertexDeclaration( s_axVertexShaderDefs[i].m_pxVertexDeclElements, &(s_axVertexShaders[i].m_pxVertexShaderDecl) );
//DeadCode VGP 15/02/2006 			Asura_Assert( D3D_OK == iRes, "Error Creating Vertex Declaration" );
			s_axVertexShaders[i].m_uVertexDeclIndex = s_axVertexShaderDefs[i].m_uVertexDeclIndex;

			// Create shader
			iRes = Asura_Direct3D::D3DDevice->CreateVertexShader( (DWORD*)pCode->GetBufferPointer(), &(s_axVertexShaders[i].m_pxVertexShader) );
			Asura_Assert( D3D_OK == iRes, "Error Creating Vertex Shader" );
		}
		if (pxBuffer) pxBuffer->Release();
		if (pCode) pCode->Release();
	}
}

void Asura_Direct3D_Shaders::ReleaseShaders(void)
{
	Asura_Assert( NULL == Asura_Direct3D::D3DDevice || s_bBeenInitialised,"Not Initialised!");
	if(s_bShadersReady)
	{
		// Release pixel shaders
		for (int i=0; i<ASURA_D3D_NUMBER_OF_PIXELSHADERS; i++)
		{
			if (s_apxPixelShaders[i])
			{
				s_apxPixelShaders[i]->Release();
				s_apxPixelShaders[i] = 0;
			}
		}

		// Release vertex shaders
		for (int i=0; i<ASURA_D3D_NUMBER_OF_VERTEXSHADERS; i++)
		{
			if (s_axVertexShaders[i].m_pxVertexShader)
			{
				s_axVertexShaders[i].m_pxVertexShader->Release();
				s_axVertexShaders[i].m_pxVertexShader = 0;
			}

//DeadCode VGP 15/02/2006 			if (s_axVertexShaders[i].m_pxVertexShaderDecl)
//DeadCode VGP 15/02/2006 			{
//DeadCode VGP 15/02/2006 				s_axVertexShaders[i].m_pxVertexShaderDecl->Release();
//DeadCode VGP 15/02/2006 				s_axVertexShaders[i].m_pxVertexShaderDecl = 0;
//DeadCode VGP 15/02/2006 			}
			s_axVertexShaders[i].m_uVertexDeclIndex = ASURA_D3D_NUMBER_OF_VERTEX_FORMATS;
		}

		// VGP 15Feb2006 Release vertex format declarations
		for (int i=0; i!=static_cast<int>(s_uVertexDeclCount); ++i)
		{
			if (s_apxVertexDeclarations[i])
			{
				s_apxVertexDeclarations[i]->Release();
				s_apxVertexDeclarations[i] = 0;
			}
		}

		// Clear 'last used' values
		s_eCurrentVertexShader = ASURA_D3D_NUMBER_OF_VERTEXSHADERS;
		s_eCurrentPixelShader = ASURA_D3D_NUMBER_OF_PIXELSHADERS;
		s_uCurrentVertexDecl = ASURA_D3D_NUMBER_OF_VERTEX_FORMATS;
	}

	s_bShadersReady = false;
}

void Asura_Direct3D_Shaders::RequestPixelShader(enum ASURA_D3D_PIXELSHADER_ID eShader)
{
	if (!s_bShadersReady) CreateShaders();

	// VGP 15Feb2006 If we're switching to the same shader, don't do anything
	if (eShader == s_eCurrentPixelShader)
	{
		return;
	}

	SCOPEDPIXEVENT(( 0xFFFFFFFF, L"Set pixel shader to %i", eShader ))

	Asura_Render::CurrentStates.InvalidateTextureOpMode();
	
	if (eShader == ASURA_D3D_PS_NONE)
	{
		Asura_Direct3D::D3DDevice->SetPixelShader(0);

		// VGP 15Feb2006
		s_eCurrentPixelShader = ASURA_D3D_NUMBER_OF_PIXELSHADERS;
	}
	else
	{
		if(!s_apxPixelShaders[eShader])
		{
			const Asura_Direct3D_PS_Definition& xPixelShaderDef = Asura_Direct3D_Shaders::s_axPixelShaderDef[ eShader];
			const float fPixelShaderVersion = xPixelShaderDef.m_fVersion;
			
			if( fPixelShaderVersion > Asura_Direct3D_Caps::GetHardwarePixelShaderVersion())
			{
				Asura_Assertf(false, ("Trying to use a pixel shader (%d), but your hardware doesn't support it. You need hardware that supports pixel shader version %f.", static_cast< u_int>( eShader), fPixelShaderVersion));
			}
			else
			{
				Asura_Assertf(false, ("Requesting a pixel shader (%d) that did not assemble", static_cast< u_int>( eShader)));
			}
		}
		Asura_Direct3D::D3DDevice->SetPixelShader(s_apxPixelShaders[eShader]);
		s_eCurrentPixelShader = eShader;
	}
}

// VGP 15Feb2006
void Asura_Direct3D_Shaders::SetVertexDeclaration(unsigned int uDeclIndex)
{
	if (uDeclIndex != s_uCurrentVertexDecl)
	{
		Asura_Direct3D::D3DDevice->SetVertexDeclaration(s_apxVertexDeclarations[uDeclIndex]);
		s_uCurrentVertexDecl = uDeclIndex;
	}
}

void Asura_Direct3D_Shaders::RequestVertexShader(enum ASURA_D3D_VERTEXSHADER_ID eShader, bool bSetDeclaration)
{
	HRESULT xResult;
	
	if (!s_bShadersReady) CreateShaders();

	/// Kevin Lea - 17/07/2002 12:18:23 - If the least significant bit of the handle is set (1) then the handle is vertex shader handler.
	/// If the least significant bit is clear (0) the handle is a legacy FVF code. At least, that's what happens in DirectX 8.1 - if
	/// necessary we could have an extra method called RequestFixedFunctionVertexShader; I just liked having one interface for all vertex shaders

#ifdef ASURA_DIRECT3D_DEBUG_VS
	xResult = Asura_Direct3D::D3DDevice->SetSoftwareVertexProcessing(TRUE);
	ASURA_D3D_OK( xResult);
#else//ASURA_DIRECT3D_DEBUG_VS
	if (Asura_Direct3D_Caps::HasHardwareTnL())
	{
		if (Asura_Direct3D_Caps::GetHardwareVertexShaderVersion() < s_axVertexShaderDefs[eShader].m_fVersion)
		{
			xResult = Asura_Direct3D::D3DDevice->SetSoftwareVertexProcessing(TRUE);
			ASURA_D3D_OK( xResult);
		}
		else
		{
			xResult = Asura_Direct3D::D3DDevice->SetSoftwareVertexProcessing(FALSE);
			ASURA_D3D_OK( xResult);
		}	
	}
#endif//!ASURA_DIRECT3D_DEBUG_VS

	// VGP 15Feb2006 Only set shader if a different shader was requested
	// KOR - 02/05/2006 - this was breaking with the new renderers, as they weren't resetting the old system.
	// Hacky fix, as we'll soon be ditching this system.
#ifndef ASURA_TARGET_SM3
	if (eShader != s_eCurrentVertexShader)
#endif
	{
		xResult = Asura_Direct3D::D3DDevice->SetVertexShader(s_axVertexShaders[eShader].m_pxVertexShader);
		ASURA_D3D_OK( xResult);
		s_eCurrentVertexShader = eShader;
	}

	// VGP 15Feb2006 Same for vertex declaration
#ifndef ASURA_TARGET_SM3
	if (bSetDeclaration && s_axVertexShaders[eShader].m_uVertexDeclIndex != s_uCurrentVertexDecl)
#endif
	{
		xResult = Asura_Direct3D::D3DDevice->SetVertexDeclaration(s_apxVertexDeclarations[s_axVertexShaders[eShader].m_uVertexDeclIndex]);
		ASURA_D3D_OK( xResult);
		s_uCurrentVertexDecl = s_axVertexShaders[eShader].m_uVertexDeclIndex;
	}
}

void Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(DWORD FVFCode)
{
	SCOPEDPIXEVENT((0xff8800, L"Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader( %d)", FVFCode));
	
#ifdef ASURA_DIRECT3D_DEBUG_VS
	Asura_Direct3D::D3DDevice->SetSoftwareVertexProcessing(TRUE);
#else//ASURA_DIRECT3D_DEBUG_VS
	if (Asura_Direct3D_Caps::HasHardwareTnL())
	{
		Asura_Direct3D::D3DDevice->SetSoftwareVertexProcessing(FALSE);
	}
#endif//!ASURA_DIRECT3D_DEBUG_VS

	Asura_Direct3D::D3DDevice->SetVertexShader(0);
	Asura_Direct3D::D3DDevice->SetFVF(FVFCode);

	// VGP 15Feb2006
	ClearLastShader();
}

HRESULT Asura_Direct3D_Shaders::SetVertexShaderConstantF( const u_int uStartRegister, const float* pfConstantData, const u_int uVector4fCount)
{
	HRESULT xResult;
	
	#if defined( ASURA_ASSERT)
	{
		const u_int uMaxConstants = Asura_Direct3D_Caps::GetNumberOfVertexShaderConstants();
		Asura_Assert( uStartRegister < uMaxConstants, "Vertex Constants Upload start after the end of the consts");
		Asura_Assert( (uStartRegister + uVector4fCount) <= uMaxConstants, "Vertex Constants Upload start after the end of the consts");
		
		#if defined( _DEBUG)
		const u_int uFloatCount = 4 * uVector4fCount;
		for( u_int uFloatIndex = 0; uFloatIndex < uFloatCount; uFloatIndex += 1)
		{
			const float fConst = pfConstantData[ uFloatIndex];
			Asura_Assertf( -1e37 < fConst && fConst < 1e37, ("Uploaded Vertex Constant[%d][%d] = %f", uFloatIndex / 4, uFloatIndex % 4, fConst));
		}
		#endif
	}
	#endif
	
	xResult = Asura_Direct3D::D3DDevice->SetVertexShaderConstantF( uStartRegister, pfConstantData, uVector4fCount);
	ASURA_D3D_OK( xResult);	
	
	return( xResult);
}

