//
// Asura_Direct3D_SmoothSkinRenderer.cpp
//
// Created by RJF - 04 August 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Asura_Direct3D_SmoothSkinRenderer.h"

// Asura Independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Memory/Asura_Mem.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#include "Shadows/Asura_Shadow_System.h"

// Asura PC
#include "Rendering/Asura_Direct3D_Effects_System.h"
#include "Rendering/Asura_Direct3D_MaterialResponse.h"

// Support for converting meshes to skins in Tools
#ifdef ASURA_TOOLS

	#include "ChunkSystem/Asura_Chunk_Loaders.h"
	#include "ChunkSystem/Asura_Chunk_ResourceFile_PC_SmoothSkin.h"
	#include "ChunkSystem/Asura_Chunk_Stream.h"
	#include "Mesh/Asura_Mesh.h"

	#include "Materials/Tools_MaterialResponse_System.h"
	#include "MeshProcessing/Tools_SmoothSkin_Processor.h"
	#include "MeshProcessing/Tools_SmoothSkin_Processor_PC.h"
	#include "Tools_System.h"

	#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		#include "Rendering/Asura_Project_Skin_Elements.h"
	#endif
#endif // defined ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
// Constant declarations
////////////////////////////////////////////////////////////////////////////////
#define SMOOTH_SKIN_EFFECT_NAME "SmoothSkin_SM3"
#define SMOOTH_SKIN_SHADOW_EFFECT_NAME "SmoothSkin_SM3_Shadow"
#define SMOOTH_SKIN_DAMAGEMAPPED_EFFECT_NAME "SmoothSkin_SM3_DamageMapped"

DEBUGVAR float	dbg_fDirect3DSmoothSkinRendererWireframeDepthBias	= -0.0001f;


////////////////////////////////////////////////////////////////////////////////
// Static member variables
////////////////////////////////////////////////////////////////////////////////
// Storage for all the smooth skins
Asura_TemplateHashTree< Asura_Direct3D_SmoothSkin > Asura_Direct3D_SmoothSkinRenderer::s_xSkinTree;

// Smooth skin vertex format declaration
// Irritatingly, DEC3N is not supported on the GeForce 6 or 7 series, only on ATI (according to the caps sheet in the DXSDK)
// However, the Float16 types are supported on all of them, so we'll use them for texcoords and blendweights like the XB360.
const Direct3D_VertexElement Asura_Direct3D_SmoothSkinRenderer::s_axSmoothSkinVertexDeclaration[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,     0 },
	{ 0, 12, D3DDECLTYPE_FLOAT16_2,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     0 },
	{ 0, 16, D3DDECLTYPE_FLOAT16_2,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     1 },
	{ 0, 20, D3DDECLTYPE_FLOAT1,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     2 },
	{ 0, 24, D3DDECLTYPE_UBYTE4,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDINDICES, 0 },
	{ 0, 28, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT,  0 },
	{ 0, 36, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,		  0 },
	{ 0, 48, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,	  0 },
	{ 0, 60, D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL,	  0 },
	D3DDECL_END()
};
// Aaand the declaration object itself...
Direct3D_VertexDeclaration* Asura_Direct3D_SmoothSkinRenderer::s_pxSmoothSkinVertexDeclarationObject = NULL;

// Pointer to the Shader Model 3 All-Singing All-Dancing Smooth Skin Rendering Ubershader (tm)
Asura_Direct3D_Effect* Asura_Direct3D_SmoothSkinRenderer::s_pxSmoothSkinUbershader = NULL;

// And a damage mapped version (multitextured)
Asura_Direct3D_Effect* Asura_Direct3D_SmoothSkinRenderer::s_pxDamageMappedSmoothSkinUbershader = NULL;

// Pointer to the slightly more sedate and less showy Smooth Skin Shadowing Ubershader
Asura_Direct3D_Effect* Asura_Direct3D_SmoothSkinRenderer::s_pxSmoothSkinShadowShader = NULL;

// Structure that holds all our effect parameter handles, for nice conceptual grouping purposes
Asura_Direct3D_SmoothSkinRenderer::ShaderParameterHandles Asura_Direct3D_SmoothSkinRenderer::s_xShaderParameterHandles;
Asura_Direct3D_SmoothSkinRenderer::ShaderParameterHandles Asura_Direct3D_SmoothSkinRenderer::s_xDamageMappedShaderParameterHandles;
Asura_Direct3D_SmoothSkinRenderer::ShadowParameterHandles Asura_Direct3D_SmoothSkinRenderer::s_xShadowParameterHandles;


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Class implementations													   
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Asura_Direct3D_SmoothSkin::Vertex
////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_TOOLS
Asura_Direct3D_SmoothSkin::Vertex& Asura_Direct3D_SmoothSkin::Vertex::operator = (const Tools_SmoothSkin_Vertex& xVertex)
{
	m_xPosition = xVertex.GetPos();
	m_xNormal = xVertex.GetNormal();
	m_fUV[0] = xVertex.GetU();
	m_fUV[1] = xVertex.GetV();
	m_fUV2[0] = xVertex.GetU2();
	m_fUV2[1] = xVertex.GetV2();
	m_fAlpha = xVertex.GetAlpha();
	u_int iBoneIndex;
	for(iBoneIndex = 0; iBoneIndex < xVertex.GetNumWeightings(); ++iBoneIndex)
	{
		m_afIndex[iBoneIndex] = xVertex.GetBoneIndices()[iBoneIndex];
		m_afWeight[iBoneIndex] = xVertex.GetBoneWeights()[iBoneIndex];
	}
	for( ; iBoneIndex<4; ++iBoneIndex)
	{
		m_afIndex[iBoneIndex] = 0;
		m_afWeight[iBoneIndex] = 0;
	}
	m_xTangent = xVertex.GetTangent();
	m_xBinormal = xVertex.GetBinormal();
	return *this;
}

#endif // ifdef ASURA_TOOLS

////////////////////////////////////////////////////////////////////////////////
// Asura_Direct3D_SmoothSkin
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_SmoothSkin::Asura_Direct3D_SmoothSkin() : m_pxStrips(0), m_uNumStrips(0), m_pxVertexBuffer(NULL), m_pxIndexBuffer(NULL), m_ePrimitiveType(D3DPT_TRIANGLESTRIP)
{
	m_pxBoneMaps = NULL;
	m_uNumBoneMaps = 0;

}

Asura_Direct3D_SmoothSkin::~Asura_Direct3D_SmoothSkin()
{
	delete[] m_pxStrips;
	delete[] m_pxBoneMaps;

	if(m_pxVertexBuffer)m_pxVertexBuffer->Release();
	if(m_pxIndexBuffer)m_pxIndexBuffer->Release();
}

// Sets the number of strips, trashing the existing strip data in the process
void Asura_Direct3D_SmoothSkin::SetNumStrips(u_int uNumStrips)
{
	delete[] m_pxStrips;

	m_pxStrips = new Strip[uNumStrips];
	m_uNumStrips = uNumStrips;

	Asura_Assert(m_pxStrips, "Unable to allocate smooth skin strips");
}

void Asura_Direct3D_SmoothSkin::SetNumBoneMaps(u_int uNumBoneMaps)
{
	delete[] m_pxBoneMaps;

	m_uNumBoneMaps = uNumBoneMaps;
	if (uNumBoneMaps)
	{
		m_pxBoneMaps = new BoneMap[uNumBoneMaps];
		Asura_Assert(m_uNumBoneMaps, "Unable to allocate smooth skin Bonemaps");
	} else {
		m_pxBoneMaps = NULL;
	}
}



// Bind the vertex and index buffer to the device ready for rendering
void Asura_Direct3D_SmoothSkin::BindBuffers()
{
	Asura_Assert(m_pxVertexBuffer && m_pxIndexBuffer, "Vertex or index buffer for smooth skin is invalid");
	Asura_Direct3D::D3DDevice->SetStreamSource(0, m_pxVertexBuffer, 0, sizeof(Vertex));
	Asura_Direct3D::D3DDevice->SetIndices(m_pxIndexBuffer);
}

// Scans all strips to see if any of them are using a transparent material
bool Asura_Direct3D_SmoothSkin::IsAnyMaterialTransparent() const
{
	for( u_int u = 0; u < m_uNumStrips; ++u )
	{
		const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( m_pxStrips[u].m_auMaterialResponseHashID[0] );

		// Materials are opaque as far as the sorted render is concerned if they're flagged as decal-type
		if( pxMaterial && pxMaterial->IsTransparent() ) return true;
	}

	return false;
}

void Asura_Direct3D_SmoothSkin::SetGeometry(u_int uNumVertices, Direct3D_VertexBuffer* pxVertices, Direct3D_IndexBuffer* pxIndices, D3DPRIMITIVETYPE ePrimitiveType)
{
	if(m_pxVertexBuffer) m_pxVertexBuffer->Release();
	if(m_pxIndexBuffer) m_pxIndexBuffer->Release();

	m_uNumVertices = uNumVertices;
	m_pxVertexBuffer = pxVertices;
	m_pxIndexBuffer = pxIndices;
	m_ePrimitiveType = ePrimitiveType;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Asura_Direct3D_SmoothSkinRenderer											
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_SmoothSkinRenderer::Initialise()
{
	Direct3D_Device* pxD3DDevice = Asura_Direct3D::D3DDevice;
	if( NULL != pxD3DDevice)
	{
		pxD3DDevice->CreateVertexDeclaration(s_axSmoothSkinVertexDeclaration, &s_pxSmoothSkinVertexDeclarationObject);
		Asura_Assert(s_pxSmoothSkinVertexDeclarationObject, "Smooth skin renderer couldn't create vertex decl");
		
		CaptureEffect();
	}
}

void Asura_Direct3D_SmoothSkinRenderer::Reset()
{
	// Unload all skins
	s_xSkinTree.Reset();

	// Forget the shader pointer, it's going to change
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES) && ASURA_LOAD_EFFECT_FILES == 1
	if (s_pxSmoothSkinUbershader)
	{
		OnEffectUnloaded(s_pxSmoothSkinUbershader);
		s_pxSmoothSkinUbershader->SetOnLoadCallback(NULL);
		s_pxSmoothSkinUbershader->SetOnUnloadCallback(NULL);
	}
	if (s_pxSmoothSkinShadowShader)
	{
		OnEffectUnloaded(s_pxSmoothSkinShadowShader);
		s_pxSmoothSkinShadowShader->SetOnLoadCallback(NULL);
		s_pxSmoothSkinShadowShader->SetOnUnloadCallback(NULL);
	}
#ifdef ASURA_USE_DAMAGE_MAPS
	if (s_pxDamageMappedSmoothSkinUbershader)
	{
		OnEffectUnloaded(s_pxDamageMappedSmoothSkinUbershader);
		s_pxDamageMappedSmoothSkinUbershader->SetOnLoadCallback(NULL);
		s_pxDamageMappedSmoothSkinUbershader->SetOnUnloadCallback(NULL);
	}
#endif
#endif
	s_pxSmoothSkinUbershader = NULL;
	s_xShaderParameterHandles.Release();
	s_pxSmoothSkinShadowShader = NULL;
	s_xShadowParameterHandles.Release();
#ifdef ASURA_USE_DAMAGE_MAPS
	s_pxDamageMappedSmoothSkinUbershader = NULL;
	s_xDamageMappedShaderParameterHandles.Release();
#endif
}

void Asura_Direct3D_SmoothSkinRenderer::Shutdown()
{
	Reset();

	if(s_pxSmoothSkinVertexDeclarationObject) s_pxSmoothSkinVertexDeclarationObject->Release();
	s_pxSmoothSkinVertexDeclarationObject = NULL;
}

Asura_Direct3D_SmoothSkin* Asura_Direct3D_SmoothSkinRenderer::CreateSmoothSkin( const Asura_Hash_ID uHash, const u_int uNumStrips )
{
#if !defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA) && defined(ASURA_ASSERT)
	const Asura_Direct3D_SmoothSkin* const pxExistingSkin = s_xSkinTree.GetDataByHash( uHash );
	Asura_Assertf( !pxExistingSkin, ("A smooth skin with this hash (0x%X) has already been created.", uHash) );
#endif	//!defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA) && defined(ASURA_ASSERT)

	Asura_Direct3D_SmoothSkin* const pxNewSkin = s_xSkinTree.AddNode( uHash );
	if( !pxNewSkin )
	{
#if !defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA)
		Asura_Assert( false, "Unable to create new character." );
#endif	//!defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA)
		return NULL;
	}

	pxNewSkin->SetNumStrips(uNumStrips);

	return pxNewSkin;
}

Asura_Direct3D_SmoothSkin* Asura_Direct3D_SmoothSkinRenderer::GetSmoothSkin( const Asura_Hash_ID uHash )
{
	return s_xSkinTree.GetDataByHash( uHash );
}

void Asura_Direct3D_SmoothSkinRenderer::ResetAllLODs(const Asura_Hierarchy_Skin* pxSkin)
{
	while(pxSkin)
	{
		s_xSkinTree.DeleteNodeByHash(pxSkin->m_uNameHashID);
		pxSkin = pxSkin->m_pxNextLowerSkin;
	}
}

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Direct3D_SmoothSkinRenderer::Render(const Asura_Animation& xAnimation,
											   const Asura_Hierarchy_Skin& xSkin,
											   Asura_AMatrix4x4_InParam xTransform,
											   Asura_AMatrix4x4_InParam xInverseTransform,
											   const Asura_AMatrix4x4* const pxBoneTransforms)
#else
void Asura_Direct3D_SmoothSkinRenderer::Render( const Asura_Animation& xAnimation,
												const Asura_Hierarchy_Skin& xSkin,
												const Asura_Vector_3& xPos,
												const Asura_Matrix_3x3& xOri,
												const Asura_Vector_3* const pxBonePos,
												const Asura_Matrix_3x3* const pxBoneOri )
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
	SCOPEDPIXEVENT(( 0xFF884499, L"Smooth skin '%S' (material response system)", xSkin.Name.GetString() ));

	Asura_Direct3D_SmoothSkin* pxSmoothSkin = GetSmoothSkin( xSkin.m_uNameHashID );

	#if defined ASURA_TOOLS

		if( !pxSmoothSkin )
		{
			pxSmoothSkin = Tools_InitialiseSmoothSkinFromRawData( &xSkin );
			Asura_Assertf( pxSmoothSkin, ("Failed to find smooth skin %s.", xSkin.Name.GetString()) );
		}

	#endif // defined ASURA_TOOLS

	if( !pxSmoothSkin ) return;

	// Change technique based on transparency. Opaque strips use the deferred renderer,
	// transparent strips the forward renderer
	const bool bTransparentPass = Asura_SortedRender::IsTransparentPass();

	Asura_Direct3D_Effect* const pxEffect = Asura_Direct3D_Effects_System::BeginEffect(
		"SmoothSkin_MaterialResponse", bTransparentPass ? 1 : 0 );

#ifdef ASURA_USE_DAMAGE_MAPS	
	bool bUseDamageMaps = xAnimation.IsUsingDamageMap();
#else
	bool bUseDamageMaps = false;
#endif

	if( !pxEffect || !pxEffect->IsD3DEffectValid() ) return;

	// Set main materials data texture
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set materials data texture" ));

		pxEffect->SetParameterByName( "g_xMaterialsTexture", Asura_Direct3D_MaterialResponse::GetMaterialsTexture());

		const D3DXVECTOR4 xMaterialProperties( 1.0f / ASURA_MATERIAL_RESPONSE_TEXTURE_WIDTH,
			Asura_Direct3D_MaterialResponse::GetMaterialsTextureScale(), 0.0f, 0.0f );

		pxEffect->SetParameterByName( "g_xMaterialsTextureProperties", xMaterialProperties );
	}

	// Set world and world-view-proj matrices
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set world-view-projection matrix" ));

		D3DXMATRIX xWorldViewProj;

#if defined(ASURA_HIERARCHY_QUAD)
		xTransform.StoreColsUnaligned(&xWorldViewProj._11);
#else
		// Rotation
		Asura_Mem::Set( &xWorldViewProj, 0, sizeof( xWorldViewProj ) );
		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xWorldViewProj.m[u][v] = xOri.Element[v][u];
			}
		}

		xWorldViewProj._44 = 1.0f;

		// Translation
		D3DXMATRIX xTemp;
		D3DXMatrixTranslation( &xTemp, xPos.x, xPos.y, xPos.z );
		xWorldViewProj *= xTemp;
#endif	// defined(ASURA_HIERARCHY_QUAD)

		pxEffect->SetParameterByName( "g_xWorldMatrix", xWorldViewProj );

		// View-projection
		xWorldViewProj *= Asura_Direct3D::GetViewProjectionMatrix();

		pxEffect->SetParameterByName( "g_xWorldViewProjectionMatrix", xWorldViewProj );
	}

	// Set bone matrices
	if (pxSmoothSkin->GetNumBoneMaps() == 0)
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set bone matrices" ));

		D3DXMATRIX axBoneMatrices[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
		Asura_Mem::Set( axBoneMatrices, 0, sizeof( axBoneMatrices ) );

		for( int i = 0; i < xSkin.NumberOfBones; ++i )
		{
#if defined(ASURA_HIERARCHY_QUAD)
			pxBoneTransforms[i].StoreColsUnaligned(&axBoneMatrices[i]._11);
#else
			for( u_int u = 0; u < 3; ++u )
			{
				for( u_int v = 0; v < 3; ++v )
				{
					axBoneMatrices[i].m[v][u] = pxBoneOri[i].Element[u][v];
				}
			}

			axBoneMatrices[i]._44 = 1.0f;

			D3DXMATRIX xTranslation;
			D3DXMatrixTranslation( &xTranslation, pxBonePos[i].x, pxBonePos[i].y, pxBonePos[i].z );

			axBoneMatrices[i] *= xTranslation;
#endif	// defined(ASURA_HIERARCHY_QUAD)

#ifdef ASURA_USE_DAMAGE_MAPS	
			// Damage maps assume the matrices are orthogonal, and we can reconstruct
			// 3 elements with a cross-product in the shader. As we're only saving 3
			// elements like this, not a whole vector, we can't split the damage values
			// from the matrix palette. So we poke them into the matrices here...
			if (bUseDamageMaps)
			{
				axBoneMatrices[i]._33 = xAnimation.GetBoneDamageMapValue(i);
			}
#endif
		}

		pxEffect->GetD3DEffect()->SetMatrixArray( "g_axBoneMatrices", axBoneMatrices, ASURA_MAX_NO_OF_BONES_PER_MATERIAL);

		pxEffect->GetD3DEffect()->SetBool( "g_bUseDamageMaps", bUseDamageMaps );
	}

	// Set render states
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set render states" ));

		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestDitheringMode( true );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );

		// Leave wireframe as it was set elsewhere
	}

	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Bind vertex data" ));

		Asura_Direct3D::D3DDevice->SetVertexDeclaration( s_pxSmoothSkinVertexDeclarationObject );
		pxSmoothSkin->BindBuffers();
	}


	if (!bTransparentPass)
	{					
		Asura_Direct3D_MaterialResponse::SetFogParameters( false, NULL, *pxEffect );
	}

	for (int iZPrePassMode = !bTransparentPass ; iZPrePassMode < 2; ++iZPrePassMode)
	{
		u_int uStartIndex = 0;
		for( u_int u = 0; u < pxSmoothSkin->GetNumStrips(); ++u )
		{
			const Asura_Direct3D_SmoothSkin::Strip& xStrip = pxSmoothSkin->GetStrip(u);


			// Set bone matrices
			if (pxSmoothSkin->GetNumBoneMaps() > 0)
			{
				SCOPEDPIXEVENT(( 0xFF884499, L"Set bone matrices" ));

				D3DXMATRIX axBoneMatrices[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
				Asura_Mem::Set( axBoneMatrices, 0, sizeof( axBoneMatrices ) );
				const Asura_Direct3D_SmoothSkin::BoneMap &xBoneMap = pxSmoothSkin->GetBoneMap(xStrip.uBoneMap);

				for( u_int i = 0; i < xBoneMap.m_uSize; ++i )
				{
					u_int uBone = xBoneMap.m_puEntries[i];
	#if defined(ASURA_HIERARCHY_QUAD)
					pxBoneTransforms[uBone].StoreColsUnaligned(&axBoneMatrices[i]._11);
	#else
					for( u_int u = 0; u < 3; ++u )
					{
						for( u_int v = 0; v < 3; ++v )
						{
							axBoneMatrices[i].m[v][u] = pxBoneOri[uBone].Element[u][v];
						}
					}

					axBoneMatrices[i]._44 = 1.0f;

					D3DXMATRIX xTranslation;
					D3DXMatrixTranslation( &xTranslation, pxBonePos[uBone].x, pxBonePos[uBone].y, pxBonePos[uBone].z );

					axBoneMatrices[i] *= xTranslation;
	#endif	// defined(ASURA_HIERARCHY_QUAD)

	#ifdef ASURA_USE_DAMAGE_MAPS	
					// Damage maps assume the matrices are orthogonal, and we can reconstruct
					// 3 elements with a cross-product in the shader. As we're only saving 3
					// elements like this, not a whole vector, we can't split the damage values
					// from the matrix palette. So we poke them into the matrices here...
					if (bUseDamageMaps)
					{
						axBoneMatrices[i]._33 = xAnimation.GetBoneDamageMapValue(uBone);
					}
	#endif
				}

				pxEffect->GetD3DEffect()->SetMatrixArray( "g_axBoneMatrices", axBoneMatrices, xBoneMap.m_uSize);

				pxEffect->GetD3DEffect()->SetBool( "g_bUseDamageMaps", bUseDamageMaps );
			}


	#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
			bool bVisible = xAnimation.GetSkinElementVisibility(xStrip.iElementID);

	#ifdef ASURA_TOOLS 
			// Need to override the visibility in tools, so we can preview things in anim/hier modes
			if (Tools_System::GetCurrentMode()==TOOLS_MODE_HIERARCHY ||
				Tools_System::GetCurrentMode()==TOOLS_MODE_ANIMATION)
			{
				bVisible = Asura_Project_Skin_Elements::IsElementVisible(xStrip.iElementID,0);
			}
	#endif
			if (bVisible)
	#endif
			{
				const Asura_Hash_ID uMaterialHash1 = xStrip.m_auMaterialResponseHashID[0];
				const Asura_Hash_ID uMaterialHash2 = xStrip.m_auMaterialResponseHashID[1];

	#if defined ASURA_TOOLS && defined ASURA_D3D_ENABLE_PIX
				const Tools_MaterialResponse* const pxToolsMaterial1 = Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );
				const char* const szMaterialName1 = pxToolsMaterial1 ? pxToolsMaterial1->GetName().GetString() : "<Unknown>";
				const Tools_MaterialResponse* const pxToolsMaterial2 = Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash2 );
				const char* const szMaterialName2 = pxToolsMaterial2 ? pxToolsMaterial2->GetName().GetString() : (uMaterialHash2 ? "<Unknown>" : "<None>");

				SCOPEDPIXEVENT(( 0xFF884499, L"Strip %u (material1 %S material2 %S)", u, szMaterialName1, szMaterialName2));
	#else
				SCOPEDPIXEVENT(( 0xFF884499, L"Strip %u", u ));
	#endif

				bool bVisible = true;

				const Asura_MaterialResponse* pxMaterial1
					= Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

				if( pxMaterial1 )
				{
					if( pxMaterial1->GetDontRender() || (iZPrePassMode == 0 && !pxMaterial1->GetZPrePass()))
					{
						bVisible = false;
					}

					// Only render transparent strips in that pass, and vice versa
					const bool bTransparentStrip = pxMaterial1->IsTransparent();
					if( bTransparentStrip != bTransparentPass )
					{
						bVisible = false;
					}
				}

				if( bVisible )
				{
					// Set material data
					{
						SCOPEDPIXEVENT(( 0xFF884499, L"Set material data" ));

						const Asura_MaterialResponse* pxMaterial2
							= Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash2 );

						Asura_Direct3D_MaterialResponse::SetEffectParameters( pxMaterial1,
							pxMaterial2, &xAnimation, *pxEffect );

						Asura_Direct3D_MaterialResponse::SetSHParameters( pxMaterial1, *pxEffect );

						if (bTransparentPass)
						{					
							Asura_Direct3D_MaterialResponse::SetFogParameters( true, pxMaterial1, *pxEffect );
						}

						Asura_Direct3D_MaterialResponse::SetBlendStates( pxMaterial1 );

						if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback())
						{
							Asura_MaterialResponse_System::GetMaterialProjectContextCallback()(pxMaterial1->GetProjectContext(), 
								pxMaterial2 ? pxMaterial2->GetProjectContext() : ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
						}
					}

					// Tri lists only used in Tools
					const u_int uPrimitiveCount = pxSmoothSkin->GetPrimitiveType() == D3DPT_TRIANGLESTRIP
												? xStrip.uNumIndices - 2
												: xStrip.uNumIndices / 3;

					if (iZPrePassMode == 0)
					{
						Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
						Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, 0);
					}


					pxEffect->RenderAllIndexedPrimitivePasses( pxSmoothSkin->GetPrimitiveType(), 0,
						pxSmoothSkin->GetNumVertices(), uStartIndex, uPrimitiveCount, NULL );
					Asura_Render::AddToTrianglesRendered( uPrimitiveCount );

					if (iZPrePassMode == 0)
					{
						Asura_Direct3D_MaterialResponse::SetBlendStates( pxMaterial1 );
						Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA);
					}
				}
			}


			uStartIndex += xStrip.uNumIndices;
		}
	}
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Clean up" ));

		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );

		for( u_int u = 0; u < 16; ++u )
		{
			Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_SRGBTEXTURE, FALSE );
		}

		const float fDepthBias = 0.0f;
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DEPTHBIAS, *(DWORD*)&fDepthBias );

		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		Asura_Direct3D_Effects_System::FlushEffects();
	}

	#if defined ASURA_TOOLS
		if( xAnimation.GetOutlineColour() )
		{
#if defined(ASURA_HIERARCHY_QUAD)
			RenderWireframe( xAnimation, xSkin, xTransform, pxBoneTransforms );
#else // #if defined(ASURA_HIERARCHY_QUAD)
			RenderWireframe( xAnimation, xSkin, xPos, xOri, pxBonePos, pxBoneOri );
#endif // #if defined(ASURA_HIERARCHY_QUAD)
		}
	#endif
}

#if defined ASURA_TOOLS

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Direct3D_SmoothSkinRenderer::RenderWireframe(const Asura_Animation& xAnimation,
														const Asura_Hierarchy_Skin& xSkin,
														Asura_AMatrix4x4_InParam xTransform,
														const Asura_AMatrix4x4* const pxBoneTransforms)
#else
void Asura_Direct3D_SmoothSkinRenderer::RenderWireframe( const Asura_Animation& xAnimation,
														 const Asura_Hierarchy_Skin& xSkin,
														 const Asura_Vector_3& xPos,
														 const Asura_Matrix_3x3& xOri,
														 const Asura_Vector_3* const pxBonePos,
														 const Asura_Matrix_3x3* const pxBoneOri )
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
	SCOPEDPIXEVENT(( 0xFF884499, L"Wireframe", xSkin.Name.GetString() ));

	Asura_Direct3D_SmoothSkin* const pxSmoothSkin = GetSmoothSkin( xSkin.m_uNameHashID );
	if( !pxSmoothSkin ) return;

	Asura_Direct3D_Effect* const pxEffect = Asura_Direct3D_Effects_System::BeginEffect( "SmoothSkin_Wireframe" );
	if( !pxEffect || !pxEffect->IsD3DEffectValid() ) return;

	// Set wireframe colour
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set wireframe colour" ));

		const Asura_RGBA xWireframeColour( xAnimation.GetOutlineColour(), 0xFF );
		const D3DXVECTOR4 xParams( xWireframeColour.m_fR, xWireframeColour.m_fG, xWireframeColour.m_fB, 1.0f );
		pxEffect->SetParameterByName( "g_xWireframeColour", xParams );
	}

	// Set world and world-view-proj matrices
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set world-view-projection matrix" ));

		D3DXMATRIX xWorldViewProj;

#if defined(ASURA_HIERARCHY_QUAD)
		xTransform.StoreColsUnaligned(&xWorldViewProj._11);
#else
		// Rotation
		Asura_Mem::Set( &xWorldViewProj, 0, sizeof( xWorldViewProj ) );
		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xWorldViewProj.m[u][v] = xOri.Element[v][u];
			}
		}

		xWorldViewProj._44 = 1.0f;

		// Translation
		D3DXMATRIX xTemp;
		D3DXMatrixTranslation( &xTemp, xPos.x, xPos.y, xPos.z );
		xWorldViewProj *= xTemp;
#endif	// defined(ASURA_HIERARCHY_QUAD)

		// View-projection
		xWorldViewProj *= Asura_Direct3D::GetViewProjectionMatrix();

		pxEffect->SetParameterByName( "g_xWorldViewProjectionMatrix", xWorldViewProj );
	}

	// Set bone matrices
	if (pxSmoothSkin->GetNumBoneMaps() == 0)
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set bone matrices" ));

		D3DXMATRIX axBoneMatrices[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
		Asura_Mem::Set( axBoneMatrices, 0, sizeof( axBoneMatrices ) );

		for( int i = 0; i < xSkin.NumberOfBones; ++i )
		{
#if defined(ASURA_HIERARCHY_QUAD)
			pxBoneTransforms[i].StoreColsUnaligned(axBoneMatrices[i]);
#else
			for( u_int u = 0; u < 3; ++u )
			{
				for( u_int v = 0; v < 3; ++v )
				{
					axBoneMatrices[i].m[v][u] = pxBoneOri[i].Element[u][v];
				}
			}

			axBoneMatrices[i]._44 = 1.0f;

			D3DXMATRIX xTranslation;
			D3DXMatrixTranslation( &xTranslation, pxBonePos[i].x, pxBonePos[i].y, pxBonePos[i].z );

			axBoneMatrices[i] *= xTranslation;
#endif	// defined(ASURA_HIERARCHY_QUAD)
		}

		pxEffect->GetD3DEffect()->SetMatrixArray( "g_axBoneMatrices", axBoneMatrices, ASURA_MAX_NO_OF_BONES_PER_MATERIAL );
	}

	// Set render states
	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Set render states" ));

		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestDitheringMode( true );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		Asura_Render::CurrentStates.RequestWireFrameMode( true );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

		// Make sure we only write to the main render target
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, 0 );
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, 0 );
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE3, 0 );

		// Apply a little depth bias so the lines don't break up across the skin
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DEPTHBIAS, *(DWORD*)&dbg_fDirect3DSmoothSkinRendererWireframeDepthBias );
	}

	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Bind vertex data" ));

		Asura_Direct3D::D3DDevice->SetVertexDeclaration( s_pxSmoothSkinVertexDeclarationObject );
		pxSmoothSkin->BindBuffers();
	}

	u_int uStartIndex = 0;

	for( u_int u = 0; u < pxSmoothSkin->GetNumStrips(); ++u )
	{
		const Asura_Direct3D_SmoothSkin::Strip& xStrip = pxSmoothSkin->GetStrip(u);

		// Set bone matrices
		if (pxSmoothSkin->GetNumBoneMaps() > 0)
		{
			SCOPEDPIXEVENT(( 0xFF884499, L"Set bone matrices" ));

			Asura_Direct3D_SmoothSkin::BoneMap &xBoneMap = pxSmoothSkin->GetBoneMap(xStrip.uBoneMap);



			D3DXMATRIX axBoneMatrices[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
			Asura_Mem::Set( axBoneMatrices, 0, sizeof( axBoneMatrices ) );

			for( u_int i = 0; i < xBoneMap.m_uSize; ++i )
			{
				u_int uBone = xBoneMap.m_puEntries[i];
#if defined(ASURA_HIERARCHY_QUAD)
				pxBoneTransforms[uBone].StoreColsUnaligned(axBoneMatrices[i]);
#else
				for( u_int u = 0; u < 3; ++u )
				{
					for( u_int v = 0; v < 3; ++v )
					{
						axBoneMatrices[i].m[v][u] = pxBoneOri[uBone].Element[u][v];
					}
				}

				axBoneMatrices[i]._44 = 1.0f;

				D3DXMATRIX xTranslation;
				D3DXMatrixTranslation( &xTranslation, pxBonePos[uBone].x, pxBonePos[uBone].y, pxBonePos[uBone].z );

				axBoneMatrices[i] *= xTranslation;
#endif	// defined(ASURA_HIERARCHY_QUAD)
			}

			pxEffect->GetD3DEffect()->SetMatrixArray( "g_axBoneMatrices", axBoneMatrices, xBoneMap.m_uSize );
		}


#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		bool bVisible = xAnimation.GetSkinElementVisibility(xStrip.iElementID);

#ifdef ASURA_TOOLS 
		// Need to override the visibility in tools, so we can preview things in anim/hier modes
		if (Tools_System::GetCurrentMode()==TOOLS_MODE_HIERARCHY ||
			Tools_System::GetCurrentMode()==TOOLS_MODE_ANIMATION)
		{
			bVisible = Asura_Project_Skin_Elements::IsElementVisible(xStrip.iElementID,0);
		}
#endif
		if (bVisible)
#endif
		{
			SCOPEDPIXEVENT(( 0xFF884499, L"Strip %u", u ));

			uint32 uPrimitiveCount;
			if (pxSmoothSkin->GetPrimitiveType() == D3DPT_TRIANGLESTRIP)
			{
				uPrimitiveCount = xStrip.uNumIndices - 2;
			}
			else
			{
				// Tri lists. Only used in tools, I think
				uPrimitiveCount = xStrip.uNumIndices / 3;
			}	
			
			pxEffect->RenderAllIndexedPrimitivePasses( pxSmoothSkin->GetPrimitiveType(), 0, pxSmoothSkin->GetNumVertices(), uStartIndex, uPrimitiveCount, NULL );
			Asura_Render::AddToTrianglesRendered( uPrimitiveCount );
		}

		uStartIndex += xStrip.uNumIndices;
	}

	{
		SCOPEDPIXEVENT(( 0xFF884499, L"Clean  up" ));

		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );

		const float fDepthBias = 0.0f;
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_DEPTHBIAS, *(DWORD*)&fDepthBias );

		Asura_Render::CurrentStates.RequestWireFrameMode( Asura_Render::GetWireframe() );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		Asura_Direct3D_Effects_System::FlushEffects();
	}
}

#endif // defined ASURA_TOOLS

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Direct3D_SmoothSkinRenderer::RenderShadow(const Asura_Animation& xAnimation,
													 const Asura_Hierarchy_Skin& xSkin,
													 const Asura_Shadow& xShadow,
													 Asura_AMatrix4x4_InParam xTransform,
													 const Asura_AMatrix4x4* const pxBoneTransforms)
#else
void Asura_Direct3D_SmoothSkinRenderer::RenderShadow(	const Asura_Animation& xAnimation,
														const Asura_Hierarchy_Skin& xSkin,
														const Asura_Shadow& xShadow,
														const Asura_Vector_3& xPos,
														const Asura_Matrix_3x3& xOri,
														const Asura_Vector_3* const pxBonePos,
														const Asura_Matrix_3x3* const pxBoneOri )
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
	SCOPEDPIXEVENT((0xff884499, L"SM3 Smooth Skin \"%S\" shadow", xSkin.Name.GetString()));

	Asura_Direct3D_SmoothSkin* pxSmoothSkin = GetSmoothSkin( xSkin.m_uNameHashID );
#ifdef ASURA_TOOLS
	if (!pxSmoothSkin )
	{
		pxSmoothSkin = Tools_InitialiseSmoothSkinFromRawData(&xSkin);
		Asura_Assertf(pxSmoothSkin, ("A smooth skin that wasn't loaded as part of an export\ncouldn't be made from existing files either. Skin name: <%s>",xSkin.Name.GetString()));

	}
#endif
	if( !pxSmoothSkin ) return;

	if(!s_pxSmoothSkinShadowShader) CaptureEffect();
	if(!s_pxSmoothSkinShadowShader) return;			// if we still don't have it, bail out

	Asura_Direct3D_Effects_System::BeginEffect(*s_pxSmoothSkinShadowShader);
	Asura_Assert(s_pxSmoothSkinShadowShader->IsD3DEffectValid(), "Smooth skin shadow renderer effect invalid");
	if (s_pxSmoothSkinShadowShader->IsD3DEffectValid())
	{
		{
			SCOPEDPIXEVENT((0xff884499, L"Set input data"));
			Asura_Direct3D::D3DDevice->SetVertexDeclaration(s_pxSmoothSkinVertexDeclarationObject);
			pxSmoothSkin->BindBuffers();
		}

#if defined(ASURA_HIERARCHY_QUAD)
		ConfigureForShadow(xTransform, xShadow);
		ConfigureForBones(s_pxSmoothSkinUbershader, &s_xShaderParameterHandles, pxSmoothSkin, &xAnimation, pxBoneTransforms, false);
#else
		ConfigureForShadow(xPos, xOri, xShadow);
		ConfigureForBones(s_pxSmoothSkinUbershader, &s_xShaderParameterHandles, pxSmoothSkin, &xAnimation, pxBonePos, pxBoneOri, false);
#endif	// defined(ASURA_HIERARCHY_QUAD)

		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

		u_int uStartIndex = 0;
		for(u_int uStripIndex = 0; uStripIndex < pxSmoothSkin->GetNumStrips(); ++uStripIndex)
		{
			SCOPEDPIXEVENT((0xff884499, L"Strip %i", uStripIndex));
			const Asura_Direct3D_SmoothSkin::Strip& xStrip(pxSmoothSkin->GetStrip(uStripIndex));

			// Set bone matrices
			if (pxSmoothSkin->GetNumBoneMaps() > 0)
			{
				SCOPEDPIXEVENT(( 0xFF884499, L"Set bone matrices" ));

				D3DXMATRIX axBoneMatrices[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
				Asura_Mem::Set( axBoneMatrices, 0, sizeof( axBoneMatrices ) );
				const Asura_Direct3D_SmoothSkin::BoneMap &xBoneMap = pxSmoothSkin->GetBoneMap(xStrip.uBoneMap);

				for( u_int i = 0; i < xBoneMap.m_uSize; ++i )
				{
					u_int uBone = xBoneMap.m_puEntries[i];
#if defined(ASURA_HIERARCHY_QUAD)
					pxBoneTransforms[uBone].StoreColsUnaligned(&axBoneMatrices[i]._11);
#else
					for( u_int u = 0; u < 3; ++u )
					{
						for( u_int v = 0; v < 3; ++v )
						{
							axBoneMatrices[i].m[v][u] = pxBoneOri[uBone].Element[u][v];
						}
					}

					axBoneMatrices[i]._44 = 1.0f;

					D3DXMATRIX xTranslation;
					D3DXMatrixTranslation( &xTranslation, pxBonePos[uBone].x, pxBonePos[uBone].y, pxBonePos[uBone].z );

					axBoneMatrices[i] *= xTranslation;
#endif	// defined(ASURA_HIERARCHY_QUAD)

				}

				s_pxSmoothSkinShadowShader->GetD3DEffect()->SetMatrixArray( "g_axBoneMatrices", axBoneMatrices, xBoneMap.m_uSize);
			}


			const Asura_Hash_ID uHash = xStrip.m_auMaterialResponseHashID[0];
			const Asura_MaterialResponse* const pxMaterial
				= Asura_MaterialResponse_System::GetMaterialByHashID( uHash );

			if( !pxMaterial || (pxMaterial->GetRenderType() != Asura_MaterialResponse::RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW) )
			{
				uint32 uPrimitiveCount;
				if (pxSmoothSkin->GetPrimitiveType() == D3DPT_TRIANGLESTRIP)
				{
					uPrimitiveCount = xStrip.uNumIndices - 2;
				}
				else
				{
					// Tri lists. Only used in tools, I think
					uPrimitiveCount = xStrip.uNumIndices / 3;
				}			
				
#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
				if (xAnimation.GetSkinElementVisibility(xStrip.iElementID))
				{			
					s_pxSmoothSkinShadowShader->RenderAllIndexedPrimitivePasses(pxSmoothSkin->GetPrimitiveType(), 0, pxSmoothSkin->GetNumVertices(), uStartIndex, uPrimitiveCount, NULL);
				}
#else
				s_pxSmoothSkinShadowShader->RenderAllIndexedPrimitivePasses(pxSmoothSkin->GetPrimitiveType(), 0, pxSmoothSkin->GetNumVertices(), uStartIndex, uPrimitiveCount, NULL);
#endif
			}

			uStartIndex += xStrip.uNumIndices;
		}
	}

	// Flush any effects we may still be using...
	Asura_Direct3D_Effects_System::FlushEffects	();
}

void Asura_Direct3D_SmoothSkinRenderer::ShaderParameterHandles::Get(Direct3D_Effect* pxSrcEffect)
{
	if(!pxSrcEffect) return;

#define LOADPARAMETER(codeName, shaderName) { codeName = pxSrcEffect->GetParameterByName(NULL, shaderName); Asura_Assert(codeName, "Could not get shader parameter '" shaderName "'."); }

	LOADPARAMETER(m_xWVPMatrix,					"g_xWorldViewProj");
	LOADPARAMETER(m_xWorldMatrix,				"g_xWorldMatrix");
	LOADPARAMETER(m_xBoneMatrices,				"g_axBoneMatrices");

	LOADPARAMETER(m_xRedSH,						"g_xRedSH");
	LOADPARAMETER(m_xGreenSH,					"g_xGreenSH");
	LOADPARAMETER(m_xBlueSH,					"g_xBlueSH");

	LOADPARAMETER(m_xTexCoordOffsetAndScale,	"g_xTexCoordOffsetAndScale");
	LOADPARAMETER(m_xSkinAlpha,					"g_fCharacterAlpha");
	LOADPARAMETER(m_xObjectSpaceCameraPosition, "g_xObjectSpaceCameraPosition");

	LOADPARAMETER(m_xNormalMapped,				"g_bNormalMapped");
	LOADPARAMETER(m_xSelfIlluminating,			"g_bSelfIlluminating");
	LOADPARAMETER(m_xSpecular,					"g_bSpecular");
	LOADPARAMETER(m_xTransparent,				"g_bTransparent");

#undef LOADPARAMETER

}

void Asura_Direct3D_SmoothSkinRenderer::ShaderParameterHandles::Release()
{
	m_xWVPMatrix					= NULL;
	m_xWorldMatrix					= NULL;
	m_xBoneMatrices					= NULL;

	m_xRedSH						= NULL;
	m_xGreenSH						= NULL;
	m_xBlueSH						= NULL;

	m_xTexCoordOffsetAndScale		= NULL;
	m_xSkinAlpha					= NULL;
	m_xObjectSpaceCameraPosition	= NULL;

	m_xNormalMapped					= NULL;
	m_xSelfIlluminating				= NULL;
	m_xSpecular						= NULL;
	m_xTransparent					= NULL;
}

void Asura_Direct3D_SmoothSkinRenderer::ShadowParameterHandles::Get(Direct3D_Effect* pxSrcEffect)
{
	if(!pxSrcEffect) return;

#define LOADPARAMETER(codeName, shaderName) { codeName = pxSrcEffect->GetParameterByName(NULL, shaderName); Asura_Assert(codeName, "Could not get shader parameter '" shaderName "'."); }

	LOADPARAMETER(m_xWVPMatrix,					"g_xWorldViewProj");
	LOADPARAMETER(m_xBoneMatrices,				"g_axBoneMatrices");

#undef LOADPARAMETER

}

void Asura_Direct3D_SmoothSkinRenderer::ShadowParameterHandles::Release()
{
	m_xWVPMatrix					= NULL;
	m_xBoneMatrices					= NULL;
}

void Asura_Direct3D_SmoothSkinRenderer::CaptureEffect()
{
	// Then regular shader
	s_pxSmoothSkinUbershader = Asura_Direct3D_Effects_System::GetEffect(SMOOTH_SKIN_EFFECT_NAME);
	Asura_Assert(s_pxSmoothSkinUbershader, "Smooth skin renderer couldn't find it's shader.");
	if(!s_pxSmoothSkinUbershader) return;
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES) && ASURA_LOAD_EFFECT_FILES == 1
	s_pxSmoothSkinUbershader->SetOnLoadCallback(OnEffectLoaded);
	s_pxSmoothSkinUbershader->SetOnUnloadCallback(OnEffectUnloaded);
#endif
	s_xShaderParameterHandles.Get(s_pxSmoothSkinUbershader->GetD3DEffect());

	// The shadow shader
	s_pxSmoothSkinShadowShader = Asura_Direct3D_Effects_System::GetEffect(SMOOTH_SKIN_SHADOW_EFFECT_NAME);
	Asura_Assert(s_pxSmoothSkinShadowShader, "Smooth skin renderer couldn't find it's shadow shader.");
	if(!s_pxSmoothSkinShadowShader) return;
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES) && ASURA_LOAD_EFFECT_FILES == 1
	s_pxSmoothSkinShadowShader->SetOnLoadCallback(OnEffectLoaded);
	s_pxSmoothSkinShadowShader->SetOnUnloadCallback(OnEffectUnloaded);
#endif
	s_xShadowParameterHandles.Get(s_pxSmoothSkinShadowShader->GetD3DEffect());


#ifdef ASURA_USE_DAMAGE_MAPS
	// The damagemapped shader
	s_pxDamageMappedSmoothSkinUbershader = Asura_Direct3D_Effects_System::GetEffect(SMOOTH_SKIN_DAMAGEMAPPED_EFFECT_NAME);
	Asura_Assert(s_pxDamageMappedSmoothSkinUbershader, "Smooth skin renderer couldn't find it's damage-mapped shader.");
	if(!s_pxDamageMappedSmoothSkinUbershader) return;
#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES) && ASURA_LOAD_EFFECT_FILES == 1
	s_pxDamageMappedSmoothSkinUbershader->SetOnLoadCallback(OnEffectLoaded);
	s_pxDamageMappedSmoothSkinUbershader->SetOnUnloadCallback(OnEffectUnloaded);
#endif
	s_xDamageMappedShaderParameterHandles.Get(s_pxDamageMappedSmoothSkinUbershader->GetD3DEffect());
#endif

}

#if !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)
void Asura_Direct3D_SmoothSkinRenderer::OnEffectLoaded(Asura_Direct3D_Effect* pxEffect)
{
	if(pxEffect == s_pxSmoothSkinUbershader)
	{
		s_xShaderParameterHandles.Get(pxEffect->GetD3DEffect());
	}
	else if(pxEffect == s_pxSmoothSkinShadowShader)
	{
		s_xShadowParameterHandles.Get(pxEffect->GetD3DEffect());
	}
#ifdef ASURA_USE_DAMAGE_MAPS
	else if(pxEffect == s_pxDamageMappedSmoothSkinUbershader)
	{
		s_xDamageMappedShaderParameterHandles.Get(pxEffect->GetD3DEffect());
	}
#endif
	else
	{
		Asura_Assert(false, "We're being informed about an effect that isn't our regular or shadow shader.");
	}
	
}

void Asura_Direct3D_SmoothSkinRenderer::OnEffectUnloaded(Asura_Direct3D_Effect* pxEffect)
{
	// No assert - at shutdown the effect can be unloaded after we Shutdown
	if(pxEffect == s_pxSmoothSkinUbershader)
	{
		s_xShaderParameterHandles.Release();
	}
	else if(pxEffect == s_pxSmoothSkinShadowShader)
	{
		s_xShadowParameterHandles.Release();
	}
}
#endif // !defined( ASURA_NO_LOAD_EFFECTS_FILES_FROM_FILES)

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Direct3D_SmoothSkinRenderer::ConfigureForBones(Asura_Direct3D_Effect* pxEffect,
														  ShaderParameterHandles* pxParams,
														  const Asura_Direct3D_SmoothSkin* pxSkin,
														  const Asura_Animation *pxAnim,
														  const Asura_AMatrix4x4* pxBoneTransforms,
														  const bool bUseDamageMaps)
#else
void Asura_Direct3D_SmoothSkinRenderer::ConfigureForBones(Asura_Direct3D_Effect* pxEffect,
														  ShaderParameterHandles* pxParams,
														  const Asura_Direct3D_SmoothSkin* pxSkin,
														  const Asura_Animation *pxAnim,
														  const Asura_Vector_3* xBonePos,
														  const Asura_Matrix_3x3* xBoneOri,
														  const bool bUseDamageMaps)
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
	SCOPEDPIXEVENT((0xff884499, L"Set bone parameters"));

#if defined(ASURA_HIERARCHY_QUAD)
	Asura_Assert(pxSkin && pxBoneTransforms && pxEffect, "Invalid parameters");
	if(!(pxSkin && pxBoneTransforms && pxEffect))return;
#else
	Asura_Assert(pxSkin && xBonePos && xBoneOri && pxEffect, "Invalid parameters");
	if(!(pxSkin && xBonePos && xBoneOri && pxEffect))return;
#endif	// defined(ASURA_HIERARCHY_QUAD)
	if(pxSkin->GetNumBoneMaps() > 0)return;

	const u_int uBoneCount = Asura_Hierarchy_System::s_pxSkin->NumberOfBones;

	D3DXMATRIX axUploadMatrices[ASURA_MAX_NO_OF_BONES_PER_MATERIAL];
	memset(axUploadMatrices, 0, sizeof(D3DXMATRIX) * ASURA_MAX_NO_OF_BONES_PER_MATERIAL);

	for(u_int uBoneIndex = 0; uBoneIndex < uBoneCount; ++uBoneIndex)
	{
		D3DXMATRIX& xResult = axUploadMatrices[uBoneIndex];

#if defined(ASURA_HIERARCHY_QUAD)
		pxBoneTransforms[uBoneIndex].StoreColsUnaligned(&xResult._11);
#else
		const Asura_Vector_3&   xPosition    = xBonePos[uBoneIndex];
		const Asura_Matrix_3x3& xOrientation = xBoneOri[uBoneIndex];

		// Expand the 3x3 orientation matrices to 4x4
		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xResult.m[v][u] = xOrientation.Element[u][v];
			}
		}
		xResult._44 = 1.0f;

		// Add in the translation
		D3DXMATRIX xTranslation;
		D3DXMatrixTranslation( &xTranslation, xPosition.x, xPosition.y, xPosition.z );
		xResult *= xTranslation;
#endif	// defined(ASURA_HIERARCHY_QUAD)

#ifdef ASURA_USE_DAMAGE_MAPS	
		// Damage maps assume the matrices are orthogonal, and we can reconstruct
		// 3 elements with a cross-product in the shader. As we're only saving 3
		// elements like this, not a whole vector, we can't split the damage values
		// from the matrix palette. So we poke them into the matrices here...
		if (bUseDamageMaps)
		{
			xResult._33 = pxAnim->GetBoneDamageMapValue(uBoneIndex);
		}
#endif
	}

	// Upload them to the shader constants
	pxEffect->GetD3DEffect()->SetMatrixArray(pxParams->m_xBoneMatrices, axUploadMatrices, uBoneCount);
	
	if( s_pxSmoothSkinShadowShader)
	{
		s_pxSmoothSkinShadowShader->GetD3DEffect()->SetMatrixArray(s_xShadowParameterHandles.m_xBoneMatrices, axUploadMatrices, uBoneCount);
	}
}

void Asura_Direct3D_SmoothSkinRenderer::ConfigureForLighting(Asura_Direct3D_Effect* pxEffect, ShaderParameterHandles* pxParams, const Asura_SphericalHarmonics_Order2_RGB& xSH)
{
	SCOPEDPIXEVENT((0xff884499, L"Set lighting parameters"));

	const float afRedSH[4] = { xSH.m_afRedCoefficients[1], xSH.m_afRedCoefficients[2], xSH.m_afRedCoefficients[3], xSH.m_afRedCoefficients[0] };
	const float afGreenSH[4] = { xSH.m_afGreenCoefficients[1], xSH.m_afGreenCoefficients[2], xSH.m_afGreenCoefficients[3], xSH.m_afGreenCoefficients[0] };
	const float afBlueSH[4] = { xSH.m_afBlueCoefficients[1], xSH.m_afBlueCoefficients[2], xSH.m_afBlueCoefficients[3], xSH.m_afBlueCoefficients[0] };

	pxEffect->GetD3DEffect()->SetFloatArray( pxParams->m_xRedSH,	afRedSH,	4);
	pxEffect->GetD3DEffect()->SetFloatArray( pxParams->m_xGreenSH,	afGreenSH,	4);
	pxEffect->GetD3DEffect()->SetFloatArray( pxParams->m_xBlueSH,	afBlueSH,	4);
}

void Asura_Direct3D_SmoothSkinRenderer::ConfigureForMaterial(Asura_Direct3D_Effect* pxEffect, ShaderParameterHandles* pxParams, const int iMaterialID, const Asura_TextureAnimationController& xTexAnimController, const int iTexIDOverride, const float fAlpha)
{
	SCOPEDPIXEVENT((0xff884499, L"Set material parameters"));

	const bool bNormalMapped = Asura_Material_System::IsMaterialNormalMapped( iMaterialID );
	const bool bSelfIlluminating = Asura_Material_System::IsMaterialSelfIlluminating(iMaterialID);
	const bool bSpecular = Asura_Material_System::IsMaterialSpecularSharp( iMaterialID ) || Asura_Material_System::IsMaterialSpecularSoft( iMaterialID );
	const bool bTransparent = Asura_Material_System::IsMaterialTransparent( iMaterialID );
	

	// Set the material property flags
	pxEffect->GetD3DEffect()->SetBool(pxParams->m_xNormalMapped, bNormalMapped);
	pxEffect->GetD3DEffect()->SetBool(pxParams->m_xSelfIlluminating, bSelfIlluminating);
	pxEffect->GetD3DEffect()->SetBool(pxParams->m_xSpecular, bSpecular);
	pxEffect->GetD3DEffect()->SetBool(pxParams->m_xTransparent, bTransparent);

	// Set the base texture
	const int iTexID = (iTexIDOverride == ASURA_TEXTURE_NONE) ? Asura_Material_System::GetTextureID(iMaterialID) : iTexIDOverride;
	Asura_Render::CurrentStates.RequestTextureIndex(0, iTexID);

#ifdef ASURA_USE_DAMAGE_MAPS
	if (Asura_Material_System::IsMaterialDualTexture(iMaterialID))
	{
		int iTextureID = Asura_Material_System::GetTextureID(iMaterialID,1);
		if (iTextureID==ASURA_TEXTURE_NONE)
		{
			iTextureID = Asura_Material_System::GetTextureID(iMaterialID,0);
		}
		Asura_Render::CurrentStates.RequestTextureIndex(2, iTextureID);
	}
#endif

	if(bNormalMapped)
	{
		// Set the normal map
		const int iNormalMapID = Asura_Material_System::GetNormalMapID(iMaterialID);
		Asura_Render::CurrentStates.RequestTextureIndex(1, iNormalMapID);

#ifdef ASURA_USE_DAMAGE_MAPS
		if (Asura_Material_System::IsMaterialDualNormalMap(iMaterialID))
		{
			int iNormalMapID = Asura_Material_System::GetNormalMapID(iMaterialID,1);
			if (iNormalMapID==ASURA_TEXTURE_NONE)
			{
				iNormalMapID = Asura_Material_System::GetNormalMapID(iMaterialID,0);
			}
			Asura_Render::CurrentStates.RequestTextureIndex(3, iNormalMapID);
		}
#endif
	}

	// Set and upload the texture animation constants
	Asura_TextureAnimationOutput xTexAnim;
	Asura_Material_System::GetTextureAnimationState( iMaterialID, &xTexAnimController, &xTexAnim );
	const float afTexCoordOffsetAndScale[4] = { xTexAnim.m_fUOffset, xTexAnim.m_fVOffset, 1.0f, 1.0f };
	pxEffect->GetD3DEffect()->SetFloatArray( pxParams->m_xTexCoordOffsetAndScale, afTexCoordOffsetAndScale, 4 );

	// (AAM - 01.09.2006) Need alpha blending if either the material is tagged as transparent,
	// or the object is not globally opaque

	if( bTransparent || fAlpha < 0.99f )
	{
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );

		// NB: on PC we don't get separate blend states per RT. We'd better just disable writing
		// to the normal-depth target, otherwise we'll get garbage
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, 0 );
	}
	else
	{
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );
	}
}

// Calculate and set the world-view-projection matrix
void Asura_Direct3D_SmoothSkinRenderer::ConfigureForObjectAndCamera(Asura_Direct3D_Effect* pxEffect, ShaderParameterHandles* pxParams, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri)
{
	SCOPEDPIXEVENT((0xff884499, L"Set object and camera parameters"));

	D3DXMATRIX xWorldViewProj;
	memset( &xWorldViewProj, 0, sizeof( D3DXMATRIX ) );

	for( u_int u = 0; u < 3; ++u )
	{
		for( u_int v = 0; v < 3; ++v )
		{
			xWorldViewProj.m[v][u] = xOri.Element[u][v];
		}
	}
	xWorldViewProj._44 = 1.0f;

	D3DXMATRIX xTranslation;
	D3DXMatrixTranslation( &xTranslation, xPos.x, xPos.y, xPos.z );
	xWorldViewProj *= xTranslation;

	pxEffect->GetD3DEffect()->SetMatrix(pxParams->m_xWorldMatrix, &xWorldViewProj);

	D3DXMATRIX xInverseWorldMatrix;
	D3DXMatrixInverse(&xInverseWorldMatrix, NULL, &xWorldViewProj);

	D3DXMATRIX xMatViewProj, xMatView, xMatProj;
	Asura_Direct3D::D3DDevice->GetTransform( D3DTS_VIEW,       &xMatView );
	Asura_Direct3D::D3DDevice->GetTransform( D3DTS_PROJECTION, &xMatProj );

	D3DXMatrixMultiply( &xMatViewProj, &xMatView, &xMatProj );
	D3DXMatrixMultiply( &xWorldViewProj, &xWorldViewProj, &xMatViewProj);

	pxEffect->GetD3DEffect()->SetMatrix(pxParams->m_xWVPMatrix, &xWorldViewProj);

	D3DXVECTOR3 xObjectSpaceCameraPosition;
	D3DXVec3TransformCoord(&xObjectSpaceCameraPosition, reinterpret_cast<const D3DXVECTOR3*>(Asura_Camera::GetPosition()), &xInverseWorldMatrix);
	pxEffect->GetD3DEffect()->SetFloatArray(pxParams->m_xObjectSpaceCameraPosition, xObjectSpaceCameraPosition, 3);
}

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Direct3D_SmoothSkinRenderer::ConfigureForShadow(Asura_AMatrix4x4_InParam xTransform,
														   const Asura_Shadow& xShadow)
#else
void Asura_Direct3D_SmoothSkinRenderer::ConfigureForShadow(const Asura_Vector_3& xPos,
														   const Asura_Matrix_3x3& xOri,
														   const Asura_Shadow& xShadow)
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
	// Character orientation

	D3DXMATRIX xWorldViewProj;
	D3DXMATRIX xTemp;

#if defined(ASURA_HIERARCHY_QUAD)
	xTransform.StoreColsUnaligned(&xWorldViewProj._11);
#else
	memset( &xWorldViewProj, 0, sizeof( D3DXMATRIX ) );

	for( u_int u = 0; u < 3; ++u )
	{
		for( u_int v = 0; v < 3; ++v )
		{
			xWorldViewProj.m[u][v] = xOri.Element[v][u];
		}
	}

	xWorldViewProj._44 = 1.0f;

	// Character position

	D3DXMatrixTranslation( &xTemp, xPos.x, xPos.y, xPos.z );
	xWorldViewProj *= xTemp;
#endif	// defined(ASURA_HIERARCHY_QUAD)

	// Camera position

	const Asura_Vector_3& xLightPos = xShadow.GetLightPosition();
	D3DXMatrixTranslation( &xTemp, -xLightPos.x, -xLightPos.y, -xLightPos.z );
	xWorldViewProj *= xTemp;

	// Camera orientation

	Asura_Matrix_3x3 xLightOri;
	xShadow.GetLightOrientation(xLightOri);
	memset( &xTemp, 0, sizeof( D3DXMATRIX ) );

	for( u_int u = 0; u < 3; ++u )
	{
		for( u_int v = 0; v < 3; ++v )
		{
			xTemp.m[u][v] = xLightOri.Element[u][v];

			if( v == 1 )
			{
				xTemp.m[u][v] = -xTemp.m[u][v];
			}
		}
	}

	xTemp._44 = 1.0f;

	xWorldViewProj *= xTemp;

	// Projection

	const float fVerticalFOV = Asura_Maths::DegToRad( xShadow.GetFrustum().GetVerticalFOVInDegrees() );
	D3DXMatrixPerspectiveFovLH( &xTemp, fVerticalFOV, 1.0f, xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane() );

	xWorldViewProj *= xTemp;

	s_pxSmoothSkinShadowShader->GetD3DEffect()->SetMatrix(s_xShadowParameterHandles.m_xWVPMatrix, &xWorldViewProj);
}

#ifdef ASURA_TOOLS
Asura_Direct3D_SmoothSkin* Asura_Direct3D_SmoothSkinRenderer::Tools_InitialiseSmoothSkinFromRawData(const Asura_Hierarchy_Skin* pxConstSkin)
{
	// Get a non-const copy of the skin to operate on
	// Need to get the top-most skin for this , since the LOD's aren't accessible via GetSkinByHash
	int iLODLevel = 0;
	const Asura_Hierarchy_Skin* pxConstParentSkin = pxConstSkin;
	while(pxConstParentSkin->m_pxNextHigherSkin)
	{
		pxConstParentSkin = pxConstParentSkin->m_pxNextHigherSkin;
		iLODLevel++;
	}
	Asura_Hierarchy_Skin* pxSkin = Asura_Animation_System::GetSkinByHash(pxConstParentSkin->m_uNameHashID);
	if(!pxSkin) return 0;
	//skip back down to the correct LOD
	for(int i=0;i<iLODLevel;i++)
	{
		if(!pxSkin->m_pxNextLowerSkin) return NULL;
		pxSkin = pxSkin->m_pxNextLowerSkin;
	}

	Tools_SmoothSkin_Processor xProcessor;
	xProcessor.SetIncludeNormalMapData( true ); // This is unnecessary for a few models, but it's going to get pretty rare.
	if( !xProcessor.Initialise( pxSkin ) ) return NULL;

	xProcessor.CreateBoneRemapTables(ASURA_MAX_NO_OF_BONES_PER_MATERIAL);

	Tools_SmoothSkin_PC_Processor xPCProcessor(xProcessor);
	if( !xPCProcessor.Initialise(false) ) return NULL;

	const bool bConvertMaterialIDs = false;

	// use dummy stream to get the memory size required.
	Asura_Chunk_Dummy_Stream xDummyStream;
	if( !xPCProcessor.OutputToChunk(xDummyStream, bConvertMaterialIDs) ) return NULL;

	u_int uSize = xDummyStream.GetPosition();
	void* pBuffer = Asura_Allocate(uSize);
	if (!pBuffer) return NULL;

	// write chunk data to memory buffer
	Asura_Chunk_Memory_Stream xMemoryStream( uSize, pBuffer );
	if (!xPCProcessor.OutputToChunk_SmoothSkinSM3(xMemoryStream, bConvertMaterialIDs))
	{
		Asura_Deallocate(pBuffer);
		return NULL;
	}

	// restart stream
	xMemoryStream.SetPosition(0);

	Asura_Chunk_ResourceFile_PC_SmoothSkin xChunk;
	xMemoryStream.Read( sizeof(xChunk), &xChunk );
	const bool bWarnIfMeshExists = false;// if creating skin data for the loaded mesh in tools we want to skip this warning.
	xMemoryStream.SkipString();	// skip resource name (mesh name is after).
	Asura_Direct3D_SmoothSkin* pxResult = 
		xChunk.ProcessFromChunkStream( xMemoryStream, bWarnIfMeshExists, bConvertMaterialIDs );

	Asura_Deallocate(pBuffer);

	return pxResult;
}
#endif // defined ASURA_TOOLS
