//*******************************************************************
//
// File: Asura_Direct3D_VertexBuffer_Dynamic.h
//
// Created: 24 Oct 2006
//
// Description: 
//
//*******************************************************************

#ifndef _ASURA_DIRECT3D_VERTEXBUFFER_DYNAMIC_H_
#define _ASURA_DIRECT3D_VERTEXBUFFER_DYNAMIC_H_

// other libraries
#include "Rendering/Asura_Direct3D.h"

// internal dependencies

// forward declarations

//*******************************************************************
// Constants
//*******************************************************************


//*******************************************************************
// Class Definitions
//*******************************************************************

//-------------------------------------------------------------------
// Class:       Asura_Direct3D_VertexBuffer_Dynamic
//
// Description: Dynamic Vertex Buffer manager class.
//				You can ask it for a block of Vertex Buffer and it will give you a unique memory block to fill in.
//				When finished you tell it how much you have used
//				
//				Do not Lock1/Unlock1, Lock2/Unlock2 Render( Lock1 Data) Render( Lock2 Data) 
//				Only Lock1/Unlock1 Render() Lock2/Unlock2 Render()
//				If the Lock1 fills the VB then Lock2 will lock at the beginning of a discarded VB 
//
//				Once the VB has been Unlocked do not try to use any data pointers returned from the Lock
//
//				Do NOT Read from the memory returned by this class
//				Do NOT access memory outside of the range that was asked for
//-------------------------------------------------------------------
class Asura_Direct3D_VertexBuffer_Dynamic
{
public:
	Asura_Direct3D_VertexBuffer_Dynamic();
	~Asura_Direct3D_VertexBuffer_Dynamic();
	
	void Clear();
	
	bool ClearDirect3DData();
	bool CreateDirect3DData();
	
	bool SetSize( const u_int uSizeBytes);
	
	Direct3D_VertexBuffer* GetVertexBuffer() const;
	
	bool Lock( const u_int uSizeBytesRequired, u_int& uLockedVertexBufferOffsetBytes, void*& pLockedMemory);
	void Unlock( const u_int uSizeBytesUsed);
	
	bool IsLocked() const { return( NULL != this->m_pVertexBufferLockedStart);}
	
#if defined( ASURA_ASSERT)
	void* Assert_LockedMemoryStart() const { return( this->m_pVertexBufferLockedStart);}
	void* Assert_LockedMemoryEnd() const { return( this->m_pVertexBufferLockedEnd);}
#endif
	
private:
	void Init();
		
private:
	
	u_int m_uVertexBufferSizeBytes;
	Direct3D_VertexBuffer* m_pxD3DVertexBuffer;
	
	u_int m_uVertexBufferUsedSizeBytes;
	u_int m_uVertexBufferLockedSizeBytes;
	void* m_pVertexBufferLockedStart;
#if defined( ASURA_ASSERT)
	void* m_pVertexBufferLockedEnd;
#endif
};

//-------------------------------------------------------------------
// Class:       Asura_Direct3D_VertexBuffer_Dynamic_Type
//
// Description: This allows a typed interface to a non-types Vertex Buffer
//-------------------------------------------------------------------
template< class Tvertex>
class Asura_Direct3D_VertexBuffer_Dynamic_Type
{
public:
	Asura_Direct3D_VertexBuffer_Dynamic_Type( Asura_Direct3D_VertexBuffer_Dynamic &xVertexBuffer);
	
	bool SetSize( const u_int uVertexCount);
	
	bool Lock( const u_int uVertexCountRequired, u_int& uLockedVertexBufferOffsetBytes, Tvertex*& ptLockedMemory);
	void Unlock( const u_int uVertexCountUsed);
	
private:
	Asura_Direct3D_VertexBuffer_Dynamic& m_xVertexBuffer;
};

//*******************************************************************
// Inline Function Definitions
//*******************************************************************

inline Direct3D_VertexBuffer* Asura_Direct3D_VertexBuffer_Dynamic::GetVertexBuffer() const
{
	Asura_Assert( NULL != this->m_pxD3DVertexBuffer, "No vertex buffer has been allocated yet.");
	Asura_Assert( NULL == this->m_pVertexBufferLockedStart, "The Vertex Buffer is locked. Unlock() before calling this method.");
	return( this->m_pxD3DVertexBuffer);
}

template< class Tvertex> 
Asura_Direct3D_VertexBuffer_Dynamic_Type< Tvertex>::Asura_Direct3D_VertexBuffer_Dynamic_Type( Asura_Direct3D_VertexBuffer_Dynamic &xVertexBuffer) :
	m_xVertexBuffer( xVertexBuffer)
{
	
}

template< class Tvertex> 
inline bool Asura_Direct3D_VertexBuffer_Dynamic_Type< Tvertex>::SetSize( const u_int uVertexCount)
{
	return( this->m_xVertexBuffer.SetSize( uVertexCount * sizeof( Tvertex)));
}

template< class Tvertex> 
inline bool Asura_Direct3D_VertexBuffer_Dynamic_Type< Tvertex>::Lock( const u_int uVertexCountRequired, u_int& uLockedVertexBufferOffsetBytes, Tvertex*& ptLockedMemory)
{
	void* pLockedMemory = NULL;
	const bool bStatus = this->m_xVertexBuffer.Lock( uVertexCountRequired * sizeof( Tvertex), uLockedVertexBufferOffsetBytes, pLockedMemory);
	ptLockedMemory = static_cast< Tvertex*>( pLockedMemory);
	return( bStatus);
}

template< class Tvertex> 
inline void Asura_Direct3D_VertexBuffer_Dynamic_Type< Tvertex>::Unlock( const u_int uVertexCountUsed)
{
	this->m_xVertexBuffer.Unlock( uVertexCountUsed * sizeof( Tvertex));
}

//*******************************************************************
//
// (c) 2006 Rebellion Derby
//
// Warning: This computer program is protected by copyright law and 
// international treaties. Unauthorized reproduction or distribution
// of this program, or any portion of it, may result in severe civil and
// criminal penalties, and will be prosecuted to the maximum extent 
// possible under the law.
//
//*******************************************************************

#endif // _ASURA_DIRECT3D_VERTEXBUFFER_DYNAMIC_H_
