/**
*
* \file Asura_PlatformRenderCharacter.h
*
* Created by Kevin Lea - 28/02/2002 16:19:37
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_PLATFORMRENDERCHARACTER_H__
#define __ASURA_PLATFORMRENDERCHARACTER_H__

#include "Animation/Asura_Hierarchy.h"
#include "Core/Asura_HashTree.h"
#include "Rendering/Asura_Direct3D.h"

#if defined(ASURA_PC_DISABLE_PRESM3_SMOOTH_SKIN_RENDERER) && ASURA_PC_DISABLE_PRESM3_SMOOTH_SKIN_RENDERER
#error The Pre-SM3 smooth skin renderer is disabled in the project config; please ensure that its files are removed and not referenced.
#endif


class Asura_Mesh;
class Asura_Stream;


class Asura_Direct3D_CharacterVertex
{
	public:
		Asura_Vector_3	m_xPosition;
		Asura_Vector_3	m_xNormal;
		float			m_fU;
		float			m_fV;
		float			m_afIndex[4];
		float			m_afWeight[4];

		bool operator == (const Asura_Direct3D_CharacterVertex& xOther) const
		{
			return (	m_xPosition		== xOther.m_xPosition 
					&&	m_xNormal		== xOther.m_xNormal
					&&	m_fU			== xOther.m_fU
					&&	m_fV			== xOther.m_fV
					&&	m_afIndex[0]	== xOther.m_afIndex[0]
					&&	m_afIndex[1]	== xOther.m_afIndex[1]
					&&	m_afIndex[2]	== xOther.m_afIndex[2]
					&&	m_afIndex[3]	== xOther.m_afIndex[3]
					&&	m_afWeight[0] 	== xOther.m_afWeight[0]
					&&	m_afWeight[1] 	== xOther.m_afWeight[1]
					&&	m_afWeight[2] 	== xOther.m_afWeight[2]
					&&	m_afWeight[3] 	== xOther.m_afWeight[3]	);
		}
};


class Asura_D3D_CharacterVertex_NormalMapped : public Asura_Direct3D_CharacterVertex
{
	public:
		Asura_Vector_3	m_xTangent;
		Asura_Vector_3	m_xBinormal;
};

enum ASURA_D3D_CHARACTER_STRIP_TYPE
{
	ASURA_D3D_CHARACTER_STRIP_BASIC = 0,
};

class Asura_Direct3D_Character_Strip
{
	public:
		Asura_Direct3D_Character_Strip()
		: m_pxVertexBuffer(NULL)
		, m_pxIndexBuffer(NULL)
		, m_iMaterialIndex(0)
		, m_uNumberOfIndices(0)
		, m_uNumberOfVertices(0)
		, m_uNumberOfTriangles(0)
		, m_iElementID(0)
		, m_eStripType(ASURA_D3D_CHARACTER_STRIP_BASIC)
		, m_ePrimitiveType(D3DPT_TRIANGLESTRIP)
		, m_aucRenderBones(0)
		, m_ucRenderBoneCount(0)
		, m_pxNextStrip(NULL)
		{}

		~Asura_Direct3D_Character_Strip();
		
		Direct3D_VertexBuffer*			m_pxVertexBuffer;
		Direct3D_IndexBuffer*			m_pxIndexBuffer;
		
		int								m_iMaterialIndex;
		u_int							m_uNumberOfIndices;
		u_int							m_uNumberOfVertices;
		u_int							m_uNumberOfTriangles;
		int								m_iElementID;
		ASURA_D3D_CHARACTER_STRIP_TYPE	m_eStripType;
		D3DPRIMITIVETYPE 				m_ePrimitiveType;


		// VGP 17Mar2006
		// A 'render bone' is a bone that is referenced by a vertex in the vertex buffer.
		// During rendering we only need to upload the render bones to the graphics card.
		// The verts in the vertex buffer have been modified to index bones in the 
		// range [0..m_usRenderBoneCount).
		// m_ausRenderBones contains the mapping from render bone to actual bone.
		//
		// Note: As a fallback (and for backwards compatibility), it is assumed that
		// if m_usRenderBoneCount is zero, then no render bone information exists, and all 
		// bones will need to be uploaded to the card.
		u_char*						m_aucRenderBones;
		u_char						m_ucRenderBoneCount;

		// For characters with multiple materials (i.e. multiple strips), the strips form a linked list
		Asura_Direct3D_Character_Strip*	m_pxNextStrip;
};


class Asura_Direct3D_CharacterRender_Handler
{
	friend class Asura_Chunk_ResourceFile_PC_Character;
	friend class Asura_Direct3D_CharacterRenderer;
	friend class Tools_Animation_Files;
	friend class Tools_Mode_Hierarchy;

public:
	// Public functions...
	Asura_Direct3D_CharacterRender_Handler  (void)	{}
//	~Asura_Direct3D_CharacterRender_Handler ()		{}

	// Reset the handler...
	void									Reset									(void);

	// Gets the material used by the skin.
	// For multiple material skins, this only gives the first strip's material...
	int										GetMaterialIndex						(const Asura_Hierarchy_Skin* pxSkin) const;

	// Check a single/multi material skin for transparent materials
	bool									IsAnyMaterialTransparent				(const Asura_Hierarchy_Skin* pxSkin) const;

private:
	// Private variables...
	Asura_TemplateHashTree<Asura_Direct3D_Character_Strip>	m_xHashTree;

	// Private functions...
	// KJL - 14/04/2003 14:36:39 - allocate a new strip for this name.
	// Returns 0 if there is no more space or if a strip of that name already exists,
	// with the appropriate asserts
	Asura_Direct3D_Character_Strip*			AddCharacterStrip						(const char* szName);

	//Free up the memory for a mesh that is no longer needed. RAR 05/09/02
	void									DeleteCharacterMesh						(const Asura_Hierarchy_Skin* pxSkin)	{ m_xHashTree.DeleteNodeByHash(pxSkin->m_uNameHashID); }

	// KJL - 14/04/2003 14:38:52 - get strip for given name; asserts and returns 0 if name not found
	const Asura_Direct3D_Character_Strip*	GetCharacterStrip						(const Asura_Hierarchy_Skin* pxSkin);

	//Deletes the render data for all LODs of a model , forcing them to be recalculated the next time it is rendered
	void ResetAllLODs(const Asura_Hierarchy_Skin* pxSkin);


#ifdef ASURA_TOOLS
	// Tools initialise skin functions
	const Asura_Direct3D_Character_Strip*	InitialiseSkinFromRawData				(const Asura_Hierarchy_Skin* pxSkin);
	const Asura_Direct3D_Character_Strip*	InitialiseNormalMappedSkinFromRawData	(const Asura_Hierarchy_Skin* pxSkin);
#endif
};


class Asura_Direct3D_CharacterRenderer
{
	friend class Asura_Direct3D_CharacterRender_Handler;

public:
	// Public typedefs
	typedef void SetupPixelShader(void);
	typedef void RevertPixelShader(void);

	typedef void RenderFunction(const class Asura_Direct3D_Character_Strip* pxCharacter, const class Asura_Animation* pxAnim, const float fAlpha, const Asura_Vector_3& xObjectCentre, bool bOnlyDepth, int iTexIDOverride);

	static void Initialise(void);
	static void Shutdown(void);
	static void Reset(void);


	// Render chooses one of the following render functions based on the strip type
	static void Render(const Asura_Hierarchy_Skin* pxSkin, const class Asura_Animation* pxAnim, const Asura_Vector_3& xObjectCentre, const float fAlpha, bool bOnlyDepth = false, int iTexIDOverride = ASURA_TEXTURE_NONE);
	static void RenderBasic(const Asura_Direct3D_Character_Strip* pxCharacter, const class Asura_Animation* pxAnim, const float fAlpha, const Asura_Vector_3& xObjectCentre, bool bOnlyDepth = false, int iTexIDOverride = ASURA_TEXTURE_NONE);
#ifdef ASURA_TOOLS
	static void RenderBasic_Outline(const Asura_Direct3D_Character_Strip* pxCharacter, const Asura_Vector_3& xObjectCentre,u_int uColour);
#endif
	// KJFL - 2004-07-09 11:02:17 - deprecated
	static void RenderGraphicNovel(const Asura_Hierarchy_Skin* pxSkin, const class Asura_Animation* pxAnim, const Asura_Vector_3& xObjectCentre, const float fAlpha, bool bOnlyDepth = false) {};

	static void RenderShadow(const Asura_Hierarchy_Skin* pxSkin, const Asura_Vector_3& xPosition);

	static void SetRenderFunction(RenderFunction* const pfnRenderFunction) { s_pfnRenderFunction = pfnRenderFunction; };

	// Gets the material used by the skin.
	// For multiple material skins, this only gives the first strip's material...
	static int GetMaterialIndex(const Asura_Hierarchy_Skin* pxSkin)	{ return(GetCurrentHandler().GetMaterialIndex(pxSkin)); }

	// Check a single/multi material skin for transparent materials
	static bool IsAnyMaterialTransparent(const Asura_Hierarchy_Skin* pxSkin)	{ return(GetCurrentHandler().IsAnyMaterialTransparent(pxSkin)); }

#ifdef ASURA_TOOLS
	//RAR 08/07/04 For testing rendering using the first 1-4 bones associated with each vertex
	static void SetMaxBonesPerVertex(int iMaxBones){s_iMaxBonesPerVertex = iMaxBones;}
	static int GetMaxBonesPerVertex(){return s_iMaxBonesPerVertex;}
#endif // ASURA_TOOLS

	// For handling the space warp lookup texture
	static void PreReset();
	static void PostReset();

	// Helper functions for various renderers
	//  RJF:  (12/05/2005 15:40:56):  Made public so that project-specific renderers can reuse as much as possible of the standard renderer
	static void SetViewVertexShaderConstants();
	static void SetLightVertexShaderConstants(float fAlpha, bool bSpecular, float fFog, bool bReflective, int iMaterialID = -1);
	static void SetBoneVertexShaderConstants(const Asura_Direct3D_Character_Strip* pxStrip = 0);
	static void SetTextureAnimVertexShaderConstants(const class Asura_Animation* pxAnim,int iMaterialID);
	static float GetFogValue(const Asura_Vector_3& xObjectCentre);

	static const Asura_Direct3D_Character_Strip* GetCharacterStrip(const Asura_Hierarchy_Skin* pxSkin) { return(GetCurrentHandler().GetCharacterStrip(pxSkin)); }

	//Deletes the render data for all LODs of a model , forcing them to be recalculated the next time it is rendered
	static void ResetAllLODs(const Asura_Hierarchy_Skin* pxSkin) { GetCurrentHandler().ResetAllLODs(pxSkin);}

#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Handler functions...
	static	void									ReparentHandler		(const Asura_Hash_ID uSourceHash, const Asura_Hash_ID uNewParentHash);
	static	void									ReparentHandler		(Asura_Direct3D_CharacterRender_Handler& xSourceHandler, const Asura_Hash_ID uNewParentHash);

	static	Asura_Direct3D_CharacterRender_Handler*	CreateHandler		(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash);
	static	bool									DestroyHandler		(const Asura_Hash_ID uHash);
	static	void									WriteHandlerReport	(u_int uHandler, Asura_Stream& xLog);

	static	void									CreateAndSetHandler	(const Asura_Hash_ID uHash, const Asura_Hash_ID uParentHandlerHash)	{ s_pxCurrentHandler = CreateHandler(uHash, uParentHandlerHash); }
	static	void									SetCurrentHandler	(Asura_Direct3D_CharacterRender_Handler* pxHandler)					{ s_pxCurrentHandler = pxHandler; }
	static	void									SetCurrentHandler	(const Asura_Hash_ID uHash)											{ s_pxCurrentHandler = s_xHandlerTree.GetDataByHash(uHash); }
	static	Asura_Direct3D_CharacterRender_Handler&	GetCurrentHandler	(void)																{ return(*s_pxCurrentHandler); }
#else
	// Handler functions...
	static	Asura_Direct3D_CharacterRender_Handler&	GetCurrentHandler	(void)																{ return(s_xHandler); }
#endif

	static void SetPixelShaderOverrides(SetupPixelShader* const pfnSetupShader = NULL, RevertPixelShader* const pfnRevertShader = NULL)
	{
		s_pfnSetupPixelShader = pfnSetupShader;
		s_pfnRevertPixelShader = pfnRevertShader;
	}
protected:

	static RenderFunction *s_pfnRenderFunction;

private:
#if defined(ASURA_USE_MULTIPLE_HANDLERS)
	// Private variables...
	static	Asura_Direct3D_CharacterRender_Handler*							s_pxCurrentHandler;		// Currently active handler.
	static	Asura_TemplateHashTree<Asura_Direct3D_CharacterRender_Handler>	s_xHandlerTree;			// Tree of all of the handlers available.
#else
	// Private variables...
	static	Asura_Direct3D_CharacterRender_Handler							s_xHandler;				// Single static-instance of a zone.
#endif

#ifdef ASURA_TOOLS
	static	int																s_iMaxBonesPerVertex;

	static void SetLightVertexShaderConstants_ForOutline(u_int uColour,float fFog);
#endif

	static SetupPixelShader* s_pfnSetupPixelShader;
	static RevertPixelShader* s_pfnRevertPixelShader;
};

#endif//__ASURA_PLATFORMRENDERCHARACTER_H__
