
//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Projector_System.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_TARGET_SM3

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Colour.h"

// Asura PC
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Effects_System.h"
#include "Rendering/Asura_Direct3D_EffectParameterBlock.h"
#include "Rendering/Asura_Direct3D_RenderTargetManager.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Direct3D_UserOptions.h"
#if defined( ASURA_USE_MATERIAL_RESPONSE_SYSTEM )
#include "Rendering/Asura_Direct3D_MaterialResponse.h"
#endif

#include "Projectors/Asura_Projector_System.h"

#ifdef ASURA_TARGET_SM3	
#include "Rendering/Asura_DeferredShading.h"
#endif

const Direct3D_VertexElement axVERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	D3DDECL_END()
};


enum
{
	// Effect Pass Indices
	Projector_Pass_ClearDestAlpha = 0,
	Projector_Pass_DynamicLight	  = 1,
	Projector_Pass_FakeLight	  = 2,
	Projector_Pass_Standard		  = 3,
	Projector_Pass_Spotlight	  = 4
};


static Direct3D_VertexDeclaration* g_pxTexProjectionVD = NULL;
static Asura_Direct3D_Effect* g_pxApplyProjectedTextureEffect;



struct Asura_PC_Projector_System_Texture_Parameters
{
public:
	Asura_Direct3D_EffectParameterBlock* m_pxParamBlock;

	D3DXMATRIX 				m_xProjectionMatrix;
	D3DXMATRIX				m_xInvViewMatrix;
	D3DXVECTOR4				m_xNearPlane;
	D3DXVECTOR4				m_xNeutralColour;
	D3DXVECTOR4				m_xProjectorPosition;
	D3DXVECTOR4				m_xProjectionDirection;
	float					m_fDistanceFade;
	float 					m_fTwoOverScreenWidth;
	float 					m_fTwoOverScreenHeight;
	float 					m_fCameraTanHalfVertFOV;
	float 					m_fCameraTanHalfHorizFOV;
	D3DXVECTOR4				m_xDetailTextureScaleAndOffset;
	Direct3D_Texture*		m_pxNormalDepthBuffer;
	Direct3D_Texture*		m_pxAlbedoBuffer;
	Direct3D_BaseTexture*	m_pxProjectedTexture;
	Direct3D_BaseTexture*	m_pxDetailTexture;
	D3DXVECTOR4				m_xFalloff;
	bool					m_bUseDetailTexture;
	D3DXVECTOR4				m_afLightColour;
	
	
#if defined( ASURA_USE_MATERIAL_RESPONSE_SYSTEM )
	Asura_Direct3D_MaterialResponse_ShaderData m_xMaterialResponse_ShaderData;
#endif
} ;

static Asura_PC_Projector_System_Texture_Parameters g_xProjectedTextureParameters;



void Asura_Projector_System::Platform_Initialise(void)
{
	if( NULL == Asura_Direct3D::D3DDevice)
	{
		return;
	}
	// Initialise effect and parameter handles

#if defined( ASURA_USE_MATERIAL_RESPONSE_SYSTEM )
	g_pxApplyProjectedTextureEffect = Asura_Direct3D_Effects_System::GetEffect("ApplyProjectedTexture_MaterialResponse");
#else
	g_pxApplyProjectedTextureEffect = Asura_Direct3D_Effects_System::GetEffect("ApplyProjectedTexture");
#endif		

	g_xProjectedTextureParameters.m_pxParamBlock = new Asura_Direct3D_EffectParameterBlock();
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		    ("g_xProjectionMatrix",					&g_xProjectedTextureParameters.m_xProjectionMatrix);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		    ("g_xInvViewMatrix",					&g_xProjectedTextureParameters.m_xInvViewMatrix);	
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		    ("g_xNearPlane",					    &g_xProjectedTextureParameters.m_xNearPlane);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		    ("g_xNeutralColour",				    &g_xProjectedTextureParameters.m_xNeutralColour);	
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		    ("g_xProjectorPosition",				&g_xProjectedTextureParameters.m_xProjectorPosition);		
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		    ("g_xProjectionDirection",				&g_xProjectedTextureParameters.m_xProjectionDirection);		
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fCameraTanHalfHorizFOV",			&g_xProjectedTextureParameters.m_fCameraTanHalfHorizFOV);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fCameraTanHalfVertFOV",				&g_xProjectedTextureParameters.m_fCameraTanHalfVertFOV);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fTwoOverScreenWidth",				&g_xProjectedTextureParameters.m_fTwoOverScreenWidth);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fTwoOverScreenHeight",				&g_xProjectedTextureParameters.m_fTwoOverScreenHeight);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>			("g_xDetailTextureScaleAndOffset",		&g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fDistanceFade",						&g_xProjectedTextureParameters.m_fDistanceFade);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_Texture*>	("g_xNormalDepthBuffer",				&g_xProjectedTextureParameters.m_pxNormalDepthBuffer);	
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_Texture*>	("g_xAlbedoBuffer",						&g_xProjectedTextureParameters.m_pxAlbedoBuffer);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_BaseTexture*>("g_xProjectedTexture",					&g_xProjectedTextureParameters.m_pxProjectedTexture);	
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_BaseTexture*>("g_xDetailTexture",					&g_xProjectedTextureParameters.m_pxDetailTexture);	
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>			("g_xFalloff",							&g_xProjectedTextureParameters.m_xFalloff);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<bool>					("g_bUseDetailTexture",					&g_xProjectedTextureParameters.m_bUseDetailTexture);
	g_xProjectedTextureParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>			("g_xLightColour",						&g_xProjectedTextureParameters.m_afLightColour);
	
#if defined( ASURA_USE_MATERIAL_RESPONSE_SYSTEM )
	g_xProjectedTextureParameters.m_xMaterialResponse_ShaderData.AddDataToParameterBlock( g_xProjectedTextureParameters.m_pxParamBlock);
#endif

	g_xProjectedTextureParameters.m_pxParamBlock->Bind(g_pxApplyProjectedTextureEffect);

	Asura_Direct3D::D3DDevice->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxTexProjectionVD );
	Asura_Assert( g_pxTexProjectionVD, "Failed to create shadow frustum vertex declaration." );
}


void Asura_Projector_System::Platform_Shutdown(void)
{
	g_pxApplyProjectedTextureEffect = NULL;
	
	if (g_xProjectedTextureParameters.m_pxParamBlock)
	{	
		delete g_xProjectedTextureParameters.m_pxParamBlock;	
		g_xProjectedTextureParameters.m_pxParamBlock = NULL;
	}

	if (g_pxTexProjectionVD)
	{
		g_pxTexProjectionVD->Release();
		g_pxTexProjectionVD = 0;
	}
}


void Asura_Projector_System::Platform_Reset(void)
{	
}




static void SetZTestAndCullMode(bool bCameraInFrustum)
{
	if(bCameraInFrustum)
	{
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_GREATER);		
	}
	else
	{	
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );		
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		
	}
}


void Asura_Projector_System::Platform_BeginRender(void)
{
	D3DPERFCALL(BeginEvent(D3DCOLOR_XRGB(0xFF, 0x80, 0x40), L"Projectors"));
}

void Asura_Projector_System::Platform_RenderProjector(Asura_Projector *pxProjector)
{
	HRESULT xResult;

#ifdef ASURA_TARGET_SM3	
	if (Asura_DeferredShading::IsShowingProjectorFrustums())
	{	
		pxProjector->RenderFrustum();
	}
#endif

	Asura_Direct3D_RenderTarget* pxDeferredRendrTarget = Asura_Direct3D::GetDeferredRenderingTarget();

	if( NULL == pxDeferredRendrTarget)
	{
		return;
	}
	
	Direct3D_Surface* pxDeferredRenderTargetFrontSurface = pxDeferredRendrTarget->GetSurface( ASURA_DIRECT3D_DEFFEDRENDERTARGET_FRONT_INDEX);
	if( NULL == pxDeferredRenderTargetFrontSurface)
	{
		return;
	}
		
	Direct3D_Texture* pxDeferredRenderTargetNormalTexture = pxDeferredRendrTarget->GetTexture( ASURA_DIRECT3D_DEFFEDRENDERTARGET_NORMAL_INDEX);
	if( NULL == pxDeferredRenderTargetNormalTexture)
	{
		return;
	}
		
	Direct3D_Texture* pxDeferredRenderTargetAlbedoTexture = pxDeferredRendrTarget->GetTexture( ASURA_DIRECT3D_DEFFEDRENDERTARGET_ALBEDO_INDEX);
	if( NULL == pxDeferredRenderTargetAlbedoTexture)
	{
		return;
	}
		
	Direct3D_Effect* pxD3DEffect = g_pxApplyProjectedTextureEffect->GetD3DEffect();
	if( NULL == pxD3DEffect)
	{
		return;
	}
	
	Direct3D_Device* pxD3DDevice = Asura_Direct3D::D3DDevice;
	
	// 6 faces/12 tris that make up a frustum - const index list
	static const uint16 au16Indices[] = {0,1,3, 1,2,3, 
								  0,4,1, 1,4,5,  
								  1,5,6, 1,6,2,
								  3,2,7, 2,6,7, 
								  0,3,7, 0,7,4,
								  4,7,5, 5,7,6};


	// Is the projector in range?
	Asura_Vector_3 xVec = *Asura_Camera::GetPosition() - pxProjector->m_xPos;
	if (xVec.MagnitudeSquared() > ((pxProjector->m_fDistFadeStart+pxProjector->m_fDistFadeRange)*(pxProjector->m_fDistFadeStart+pxProjector->m_fDistFadeRange)))
	{
		// out of range - don't draw anything
		return;
	}

	float fDistanceFade = 1.0f;
	float fDist = xVec.Magnitude();

	if (fDist> pxProjector->m_fDistFadeStart)
	{
		fDistanceFade = 1.0f - ((fDist - pxProjector->m_fDistFadeStart)/pxProjector->m_fDistFadeRange);
	}
	
	SCOPEDPIXEVENT((0xff00ff00, L"Projector"));

	Asura_Direct3D_Effects_System::BeginEffect(*g_pxApplyProjectedTextureEffect);
	Asura_Assert(g_pxApplyProjectedTextureEffect->IsD3DEffectValid(), "Apply Projected Texture effect invalid");

	Asura_Direct3D::ClearRenderTargets( 1);

	xResult = pxD3DDevice->SetRenderTarget( 0, pxDeferredRenderTargetFrontSurface);
	ASURA_D3D_OK( xResult);
	
	Asura_Render::CurrentStates.RequestAlphaTest( false );	

	xResult = pxD3DDevice->SetVertexDeclaration( g_pxTexProjectionVD );
	ASURA_D3D_OK( xResult);

	// Hacky: In case cached states are wrong, this will force the state change
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );

	// If we're inside the frustum, draw the back faces, else draw the front faces. When outside the
	// frustum, we can leave z-tests on (as an optimization - less pixel shader work)
	Asura_Render::CurrentStates.RequestAlphaTest( false );	
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );	
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );		
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

	bool bCameraInFrustum = pxProjector->PointInFrustum(*Asura_Camera::GetPosition(),1.0f);
	SetZTestAndCullMode(bCameraInFrustum);
	
	// Set the textures
	g_xProjectedTextureParameters.m_pxProjectedTexture = (0 <= pxProjector->m_iTextureIndex) ? Asura_Direct3D::GetDirect3DTextureByIndex( pxProjector->m_iTextureIndex ) : NULL;		
	g_xProjectedTextureParameters.m_pxDetailTexture = (0 <= pxProjector->m_iDetailTextureIndex) ? Asura_Direct3D::GetDirect3DTextureByIndex( pxProjector->m_iDetailTextureIndex ) : NULL;
	g_xProjectedTextureParameters.m_bUseDetailTexture = (pxProjector->m_iDetailTextureIndex!=ASURA_TEXTURE_NONE);		

	// Falloff
	g_xProjectedTextureParameters.m_xFalloff.x = pxProjector->m_fFalloffScale;
	g_xProjectedTextureParameters.m_xFalloff.y = pxProjector->m_fFalloffPower;
	g_xProjectedTextureParameters.m_xFalloff.z = pxProjector->m_fAngularFalloffOfs;
	g_xProjectedTextureParameters.m_xFalloff.w = pxProjector->m_fAngularFalloffScale;


	// Set the neutral colour - this is the colour that will have no visible effect when
	// blended in the current blend mode - and set when outside the range of projection/falloff
	Asura_RGBA xNeutralColour;

	// Hacky: In case cached states are wrong, this will force the state change
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 	


	
	bool bShadows = false;
	int  iPass = 0;

	if( true == pxProjector->m_bDynamicLight || true == pxProjector->m_bDynamicSpotlight)
	{

		if (true == pxProjector->m_bDynamicLight)
			iPass = Projector_Pass_DynamicLight;
		else
			iPass = Projector_Pass_Spotlight;

		// -----------------------------------------------------------------------------------------------------------------------
		// DMR - 03/04/07 - Shadow support. Only for 'Dynamic Light' projectors.
		// If this projector is a shadow caster, we need to build the shadow mask now (in the destination alpha channel), and set the
		// appropriate blending mode.
		//
		// Before projecting the shadows, we need to clear the area of the shadow mask that's used by this light - we do that by
		// rendering the projector frustum, with colour writes only enabled to the alpha channel
		// -----------------------------------------------------------------------------------------------------------------------
		if (pxProjector->GetShadowParameters().m_uFlags & (ASURA_SHADOWFLAG_CASTS_OWN_SHADOWS|ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS))
		{	
			// Are there any active shadows for this light?
			if (Asura_Shadow_System::LightHasActiveShadows(ASURA_SHADOW_LIGHTSOURCE_PROJECTOR,pxProjector->GetHandle()))
			{	
				// Set this flag, so we know to restore the renderstates
				bShadows = true;
			}
		}
		
		if (bShadows)
		{

			// First, render the projection frustum to fill the destination alpha with 0's 
			Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA);
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 						
			
			g_xProjectedTextureParameters.m_pxParamBlock->Apply();			
			xResult = pxD3DEffect->BeginPass(Projector_Pass_ClearDestAlpha);
			ASURA_D3D_OK( xResult);
			xResult = pxD3DEffect->CommitChanges();
			ASURA_D3D_OK( xResult);
			xResult = pxD3DDevice->DrawIndexedPrimitiveUP( D3DPT_TRIANGLELIST,0,8,12, (void*)au16Indices, D3DFMT_INDEX16, (void *)pxProjector->m_axVertices ,12);
			ASURA_D3D_OK( xResult);
			xResult = pxD3DEffect->EndPass();
			ASURA_D3D_OK( xResult);			

			// Then render the shadow mask into the destination alpha
			Asura_Direct3D_Effects_System::EndEffect(*g_pxApplyProjectedTextureEffect);			
			Asura_Shadow_System::BuildShadowMask(ASURA_SHADOW_LIGHTSOURCE_PROJECTOR, pxProjector->GetHandle());
			Asura_Direct3D::ClearRenderTargets( 1);

			// Re-enable RGB writes			
			Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA|D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE);
			
			// Restore shader and vertex decl
			Asura_Direct3D_Effects_System::BeginEffect(*g_pxApplyProjectedTextureEffect);
			xResult = pxD3DDevice->SetVertexDeclaration( g_pxTexProjectionVD );

			// Restore Z-test and culling mode
			SetZTestAndCullMode(bCameraInFrustum);

			// Asura doesn't have a mode for this, so I'm changing it manually here, then restoring it when we're done			
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_ADD); 
			Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_INVDESTALPHA );
		}		
		else
		{					
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_ADD); 		
		}

		xNeutralColour = Asura_RGBA(0x000000,0);

	}
	else if( true == pxProjector->m_bFakeLight)
	{
		// Need a special case blend mode for this - Dest + (Src * Dest). As we have an HDR target, the
		// output from the shader can be >1, so we can do this without sampling from the colour buffer like
		// the 360 version has to.
		iPass = Projector_Pass_FakeLight;

		Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING); 
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_DESTCOLOR);
		xNeutralColour = Asura_RGBA(0x000000,0);
	}
	else
	{
		iPass = Projector_Pass_Standard;

		switch (pxProjector->m_eBlendMode)
		{
			case ASURA_PROJECTOR_BLEND_MULTIPLY:
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_COLOUR); 
				xNeutralColour = Asura_RGBA(0xffffff,0);
				break;

			case ASURA_PROJECTOR_BLEND_MULTIPLYX2:
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_DOUBLECOLOUR); 
				xNeutralColour = Asura_RGBA(0x808080,0);
				break;

			case ASURA_PROJECTOR_BLEND_ADD:
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING); 
				xNeutralColour = Asura_RGBA(0x000000,0);
				break;

			default:
				Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL); 
				xNeutralColour = Asura_RGBA(0x000000,0);
				break;
		}
	}
	
	// Ideally, we'd set our border colour to the neutral colour, but we can't do this on 360 - only
	// black and white border colours are supported.
	pxD3DDevice->SetSamplerState(1,D3DSAMP_BORDERCOLOR, 0x00000000);

	// Also use a border colour for the falloff, so when we've got no texture, it still clips off at the ends 
	// of the frustum
	pxD3DDevice->SetSamplerState(2,D3DSAMP_BORDERCOLOR, 0x00000000);

	// Near plane equation - used to calc distance for falloff
	// Needs to be in view space.
	Asura_Vector_3 xDir = Asura_Vector_3(0,0,1);
	xDir = xDir * pxProjector->m_xOrient;	
	Asura_Vector_3 xPoint = pxProjector->m_xPos+ (xDir * pxProjector->m_fMinRange);

	// We need to transform this point and normal into view space
	// We'll also need the inverse of the view matrix soon.
	D3DXMATRIX xViewMatrix;
	D3DXMATRIX xViewMatrixInv;
	xViewMatrix = Asura_Direct3D::GetViewMatrix();	
	D3DXMatrixInverse(&xViewMatrixInv,NULL,&xViewMatrix);

	// D3Dify it
	D3DXVECTOR3 xNormal, xNormalWS				= D3DXVECTOR3(xDir.x,   xDir.y,   xDir.z  );
	D3DXVECTOR3 xPointOnPlaneWS					= D3DXVECTOR3(xPoint.x, xPoint.y, xPoint.z);
	D3DXVECTOR4 xPointOnPlane;

	// To view space	
	D3DXVec3Transform(&xPointOnPlane, &xPointOnPlaneWS, &xViewMatrix);	
	D3DXVec3TransformNormal(&xNormal, &xNormalWS,		&xViewMatrix);

	// Get plane equation for near plane, in view space
	// ax+by+cz+d = 0,  d = - (ax+by+cz).
	float fD= - ((xNormal.x * xPointOnPlane.x)+(xNormal.y * xPointOnPlane.y)+(xNormal.z * xPointOnPlane.z));
	D3DXVECTOR4 xPlane;
	float fRange = pxProjector->m_fMaxRange - pxProjector->m_fMinRange;
	xPlane.x = xNormal.x / fRange;
	xPlane.y = xNormal.y / fRange;
	xPlane.z = xNormal.z / fRange;
	xPlane.w = fD / fRange;

	
	if (g_pxApplyProjectedTextureEffect->IsD3DEffectValid())
	{
		SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), L"Projector")); 

		D3DXMATRIX xProjMatrix, xOrient, xPos;
		D3DXMatrixIdentity(&xOrient);		

		if (pxProjector->m_eType == ASURA_PROJECTOR_TYPE_PERSPECTIVE)
		{
			// Perspective projection
			D3DXMatrixPerspectiveFovLH(&xProjMatrix, pxProjector->m_fFoV, 1.0f, pxProjector->m_fMinRange, pxProjector->m_fMaxRange);
		}
		else
		{
			// Parallel projection
			D3DXMatrixScaling(&xProjMatrix, 1.0f/pxProjector->m_fScale, 1.0f/pxProjector->m_fScale, 1.0f);		
		}
			

		// Orientation - Convert Asura matrix to D3D matrix
		for(int i=0;i<3;i++)
		{
			for(int j=0;j<3;j++)
			{
				xOrient.m[i][j] = pxProjector->m_xOrient.Element[i][j];				
			}			
		}

		// Translation (world space pos of projector)
		D3DXMatrixTranslation(&xPos,-pxProjector->m_xPos.x,-pxProjector->m_xPos.y,-pxProjector->m_xPos.z);
		D3DXMATRIX xScale,xOffset;
		
		// Scale and offset the texture from -1..+1 projection to 0..1 tex coords
		D3DXMatrixScaling(&xScale,0.5f / pxProjector->m_fUVAspect,0.5f,0);
		D3DXMatrixTranslation(&xOffset,0.5f,0.5f,0);
		xProjMatrix =  xPos * xOrient * xProjMatrix * xScale * xOffset;
		
		// Set shader params
		g_xProjectedTextureParameters.m_xProjectionMatrix		= xViewMatrixInv*xProjMatrix;
		g_xProjectedTextureParameters.m_xInvViewMatrix			= xViewMatrixInv;
		g_xProjectedTextureParameters.m_fCameraTanHalfVertFOV	= Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
		g_xProjectedTextureParameters.m_fCameraTanHalfHorizFOV  = g_xProjectedTextureParameters.m_fCameraTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();
		g_xProjectedTextureParameters.m_fTwoOverScreenWidth		= 2.0f / Asura_Viewport::GetWidth();
		g_xProjectedTextureParameters.m_fTwoOverScreenHeight	= 2.0f / Asura_Viewport::GetHeight();	
		g_xProjectedTextureParameters.m_pxNormalDepthBuffer		= pxDeferredRenderTargetNormalTexture;
		g_xProjectedTextureParameters.m_pxAlbedoBuffer			= pxDeferredRenderTargetAlbedoTexture;
		g_xProjectedTextureParameters.m_xNearPlane              = xPlane;
		g_xProjectedTextureParameters.m_xNeutralColour.x		= xNeutralColour.m_fR;
		g_xProjectedTextureParameters.m_xNeutralColour.y		= xNeutralColour.m_fG;
		g_xProjectedTextureParameters.m_xNeutralColour.z		= xNeutralColour.m_fB;
		g_xProjectedTextureParameters.m_xNeutralColour.w		= xNeutralColour.m_fA;		

		// (AAM - 07.01.2008) Roll a x2 from the pixel shader into here
		g_xProjectedTextureParameters.m_fDistanceFade			= fDistanceFade * 2.0f;

		g_xProjectedTextureParameters.m_xProjectorPosition.x    = pxProjector->m_xPos.x;
		g_xProjectedTextureParameters.m_xProjectorPosition.y    = pxProjector->m_xPos.y;
		g_xProjectedTextureParameters.m_xProjectorPosition.z    = pxProjector->m_xPos.z;
		g_xProjectedTextureParameters.m_xProjectorPosition.w    = 1;		
		g_xProjectedTextureParameters.m_xProjectionDirection.x  = pxProjector->m_xOrient.e13;
		g_xProjectedTextureParameters.m_xProjectionDirection.y  = pxProjector->m_xOrient.e23;
		g_xProjectedTextureParameters.m_xProjectionDirection.z  = pxProjector->m_xOrient.e33;
		g_xProjectedTextureParameters.m_xProjectionDirection.w  = 1;		
		g_xProjectedTextureParameters.m_afLightColour[0]		= pxProjector->m_afRGB[0]*pxProjector->m_fIntensity;
		g_xProjectedTextureParameters.m_afLightColour[1]		= pxProjector->m_afRGB[1]*pxProjector->m_fIntensity;
		g_xProjectedTextureParameters.m_afLightColour[2]		= pxProjector->m_afRGB[2]*pxProjector->m_fIntensity;
		g_xProjectedTextureParameters.m_afLightColour[3]		= 1;

		g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset.x = pxProjector->m_fUScrollOfs;
		g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset.y = pxProjector->m_fVScrollOfs;
		g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset.z = pxProjector->m_fDetailTextureScale;		
		
		ASURA_D3D_OK( xResult);
 
#if defined( ASURA_USE_MATERIAL_RESPONSE_SYSTEM )
		g_xProjectedTextureParameters.m_xMaterialResponse_ShaderData.m_bUseSpecularMap = true;
		g_xProjectedTextureParameters.m_xMaterialResponse_ShaderData.SetupCommon();
#endif
		g_xProjectedTextureParameters.m_pxParamBlock->Apply();

		// Render the frustum
		xResult = pxD3DEffect->BeginPass(iPass);
		ASURA_D3D_OK( xResult);

		xResult = pxD3DEffect->CommitChanges();
		ASURA_D3D_OK( xResult);
		
		xResult = pxD3DDevice->DrawIndexedPrimitiveUP( D3DPT_TRIANGLELIST,0,8,12, (void*)au16Indices, D3DFMT_INDEX16, (void *)pxProjector->m_axVertices ,12);
		ASURA_D3D_OK( xResult);
		
		xResult = pxD3DEffect->EndPass();
		ASURA_D3D_OK( xResult);

		// Don't leave render targets bound to the effect - can cause bad things to happen when resetting the device?		
		g_xProjectedTextureParameters.m_pxNormalDepthBuffer = NULL;
		g_xProjectedTextureParameters.m_pxAlbedoBuffer = NULL;
		g_xProjectedTextureParameters.m_pxParamBlock->Apply();
		xResult = pxD3DEffect->CommitChanges();
		ASURA_D3D_OK( xResult);

		// Flush any effects we may still be using...
		Asura_Direct3D_Effects_System::FlushEffects();
		pxDeferredRendrTarget->Bind();
	}


	// Restore states
	if (bShadows)
	{
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE );
	}

	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );		
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL); 
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);

	// Yuck. Got to do this twice, as Asura's cached state may be invalid
	Asura_Render::CurrentStates.RequestAlphaTest( true );
	Asura_Render::CurrentStates.RequestAlphaTest( false );

	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
	Asura_Direct3D::D3DDevice->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);

	// Projectors use samplers 1 & 3 for the colour textures
	Asura_Direct3D::D3DDevice->SetSamplerState( 1, D3DSAMP_SRGBTEXTURE, FALSE );
	Asura_Direct3D::D3DDevice->SetSamplerState( 3, D3DSAMP_SRGBTEXTURE, FALSE );
}

void Asura_Projector_System::Platform_EndRender(void)
{
	D3DPERFCALL(EndEvent()); // Projectors
}

#endif //#ifdef ASURA_TARGET_SM3