//////////////////////////////////////////////////////////////////////////////////////////
//
//  FILE        :  Asura_D3D_VS_Water.cpp
//
//  CREATED     :  Richard J. Fine  (17.3.2005)
//
//  Copyright (C) 2005 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////////


#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Direct3D_Shaders.h"

#ifdef ASURA_USE_WATER

// Vertex declarator for water vertices
static Direct3D_VertexElement auVSD[] =
{
	{	0,	0,	D3DDECLTYPE_FLOAT3,		D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_POSITION,	0 },
	{	0, 12,	D3DDECLTYPE_D3DCOLOR,	D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_COLOR,		0 },
	{	0, 16,	D3DDECLTYPE_FLOAT2,		D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_TEXCOORD,	0 },
	D3DDECL_END()
};


static const char* const szVS_WATER_CODE =

/*		Source data:
c  0		-\
c  1		 }- View matrix
c  2		 |
c  3		-/
c  4		-\
c  5		 }- Proj matrix
c  6		 |
c  7		-/

c  8		(XYZ) XScale	(W) Reflection
c  9		(XYZ) YScale	(W) ?
c 10		(XYZ) ZScale	(W) Alpha
c 11		(XYZ) Falloff	(W) fSunYScale
c 12		(XYZ) T values	(W) ?
c 13		0, 1, 2, 1 / 2pi
c 14		1, -1.5, 0.5, 0
c 15		3,   -3, 0.5, 0 
c 16		(X) Horizon brightness (Y) Horizon Y scale (Z) Horizon V offset (W) Horizon V scale
c 17		(XYZ) Camera position (W) camHeading/2pi
c 18		(X) sin(camera heading) (Y) cos(camera heading) (Z) -LightScale (W) LightOffset
c 19		\
c 20		---- texture transform
c 21		(XYZ) Sun direction
c 22		(X) K1 (Y) K2 (Z) UVOffset0 (W) UVOffset1


// Outputs:
oPos.xyzw	Vertically displaced vertex position
oD0.xyzw	Vertex color (lighting + alpha)
oT0.xy		Water texture coordinates
oT1.xy		Horizon texture coordinates
oT2.z		Sun falloff
oT2.w		Fresnel term
oT3.xy		Sun texture coordinates
*/

/*

"vs_1_1									\n"
"dcl_position v0						\n"
"dcl_color v1							\n"
"dcl_texcoord v2						\n"

// calculate our phase values. One value per wave.
// frac(x * positionx + z * positionz + t)	
// then add one if it's less than zero
"mul r1.xyz, v0.xxx, c[ 8].xyz			\n"
"mad r1.xyz, v0.zzz, c[10].xyz, r1.xyz	\n"
"add r1.xyz, r1.xyz, c[12].xyz			\n"
"frc r0.xy, r1							\n"
"frc r1.y, r1.z							\n"
"mov r0.z, r1.y							\n"
"slt r1.xyz, r0.xyz, c[13].xxx			\n"
"add r0.xyz, r0.xyz, r1.xyz				\n"

// Use these values to calculate heights and derivatives
// height(x)		=  x^3 - 1.5x^2 + 0.5x
// derivative(x)	= 3x^2 -   3x   + 0.5
// we'll start by building (x^3, x^2, x, ?) / (x^2, x, 1, ?) registers for each phase
// phase 1 gets r0/r1
// phase 2 gets r2/r3
// phase 3 gets r4/r5

// phase 3
"mov r4.xyz, r0.zzz						\n"				// line 10
"mul r4.xy, r4.xy, r4.z					\n"
"mul r4.x, r4.x, r4.z					\n"
"mov r5.xy, r4.yz						\n"
"mov r5.z, c[13].y						\n"

// phase 2
"mov r2.xyz, r0.yyy						\n"
"mul r2.xy, r2.xy, r2.z					\n"
"mul r2.x, r2.x, r2.z					\n"
"mov r3.xy, r2.yz						\n"
"mov r3.z, c[13].y						\n"

// phase 1
"mov r0.z, r0.x							\n"
"mul r0.xy, r0.z, r0.z					\n"
"mul r0.x, r0.x, r0.z					\n"
"mov r1.xy, r0.yz						\n"
"mov r1.z, c[13].y						\n"

// dot each pair of constants against the coefficients, storing height results into r6 and derivative results into r7
"dp3 r6.x, r0, c[14]					\n"
"dp3 r6.y, r2, c[14]					\n"
"dp3 r6.z, r4, c[14]					\n"
"dp3 r7.x, r1, c[15]					\n"
"dp3 r7.y, r3, c[15]					\n"
"dp3 r7.z, r5, c[15]					\n"

// Calculate the vertex's camera space position as if we didn't displace it
"m4x4 r10, v0, c[ 0]					\n"

// Calculate the falloff value, (2 - (cam.z)/scales)
"mul r10.xyz, r10.zzz, c[11].xyz		\n"
"sub r10.xyz, c[13].zzz, r10.xyz		\n"
"max r10.xyz, r10.xyz, c[13].xxx		\n"	// clamp to the range
"min r10.xyz, r10.xyz, c[13].yyy		\n"	//		[0..1]

// Apply this falloff
"mul r6.xyz, r6.xyz, r10.xyz			\n"
"mul r7.xyz, r7.xyz, r10.xyz			\n"

// Scale the heights/derivatives according to the per-phase scaling constants
"mul r6.xyz, r6.xyz, c[9].xyz			\n"
"mul r7.xyz, r7.xyz, c[9].xyz			\n"

// Apply heights from phases 1 and 2 to the vertex
"mov r9, v0								\n"
"add r9.y, r9.y, r6.x					\n"
"add r9.y, r9.y, r6.y					\n"

// Push the vertex through view/proj transforms
"m4x4 r10, r9, c[0]						\n"
"m4x4 oPos, r10, c[4]					\n"

// Calculate the vertex normals
// the r0 normal is 'correct', while the r1 normal is scaled by the horizon scaling value
// We have the derivatives in r7.xyz, and they have been scaled for distance falloff and height
// Now we multiply by xScale and zScale to get components of the normal
"mul r0.xyz, c[ 8].xyz, r7.xyz			\n"
"mul r1.xyz, c[10].xyz, r7.xyz			\n"
// y scaling has already been applied
// sum them up into a (xParts, -1, zParts) vector
"dp3 r0.x, r0.xyz, c[13].yyy			\n"
"dp3 r0.z, r1.xyz, c[13].yyy			\n"
"mov r0.y, -c[13].y						\n"
// copy to r1
"mov r1.xyz, r0.xyz						\n"
// normalize r0
"dp3 r2.x, r0, r0						\n"
"rsq r2.x, r2.x							\n"
"mul r0.xyz, r0.xyz, r2.x				\n"	// 'fNormalPass2' in PS2 code
// apply horizon scaling to r1
"mul r1.y, r1.y, c[16].y				\n"
// normalize it
"dp3 r2.x, r1, r1						\n"
"rsq r2.x, r2.x							\n"
"mul r1.xyz, r1.xyz, r2.x				\n" // 'fNormal' in PS2 code

// Get the direction from the camera into r2
"sub r2.xyz, r9.xyz, c[17].xyz			\n"
// Normalise it
"dp3 r3.x, r2.xyz, r2.xyz				\n"
"rsq r3.x, r3.x							\n"
"mul r2.xyz, r2.xyz, r3.x				\n"

// Calculate (D.N)
"dp3 r3.x, r2.xyz, r1.xyz				\n" // in PS2 code: "fDotP1", aka 'fChosenDot'

// Calculate the fresnel term, which is (1 - abs(D.N))^2
"sge r3.y, r3.x, c[13].x				\n" // r3.y = (r3.x > 0)?1.f:0.f;
"mad r3.y, r3.y, c[13].z, -c[13].y		\n" // r3.y = (r3.x > 0)?1.f:-1.f;
"mad r3.z, r3.x, r3.y, -c[13].y			\n" // r3.z = 1 - abs(r3.x)
"mul r3.z, r3.z, r3.z					\n" // fresnel = (1 - abs(r3.x))^2
"mul oT2.x, r3.z, c[ 8].w				\n" // multiply by reflectivity constant

// Calculate the reflection vector D - 2(D.N)N
"mul r4.xyz, r3.x, r1.xyz				\n" // (D.N)N	// line 70
"mad r4.xyz, r4.xyz, -c[13].z, r2.xyz	\n" // -2(D.N)N + D == D - 2(D.N)N

// Figure out the horizon map texture coordinates from this reflection vector.
// U = ((Rx * cos) - (Rz * sin)) / 2pi + offset
// V = (Ry * scale) + offset +1
// where the offsets are (heading/2pi, Voffset)
// U coordinate
"mul r5.xy, r4.xz, c[18].yx				\n" // (Rx * cos, Rz * sin)
"sub r5.x, r5.x, r5.y					\n" // ((Rx * cos) - (Rz * sin))
"mul r5.x, r5.x, c[13].w				\n" // (((Rx * cos) - (Rz * sin))/2pi)
"add oT1.x, r5.x, c[17].w				\n" // (((Rx * cos) - (Rz * sin))/2pi + offset
// V coordinate
"mad r5.y, r4.y, c[16].w, c[16].z		\n" // Ry * scale + offset
"add oT1.y, r5.y, c[13].y				\n" // (Ry * scale) + offset + 1

// Lighting: (offset - scale*height) * original colour, clamped to [0..1]
"add r5.x, r6.x, r6.y					\n"	// height
"mad r5.x, r5.x, c[18].z, c[18].w		\n" // height*(-scale) + offset == offset - scale*height
"mul r5.xyz, v1.xyz, r5.x				\n" // (offset - scale*height) * original color
"mov r5.w, c[10].w						\n" // copy in alpha
"min r5, r5, c[13].y					\n" // clamp to 1
"max oD0, r5, c[13].x					\n" // clamp to 0 and output

// Sun reflection stuff. Start by calculating a new reflection vector that isn't scaled for the horizon

// Calculate (D.N)
"dp3 r3.x, r2.xyz, r0.xyz				\n" // in PS2 code: "fDot"

// Calculate the reflection vector D - 2(D.N)N
"mul r4.xyz, r3.x, r0.xyz				\n" // (D.N)N
"mad r4.xyz, r4.xyz, -c[13].z, r2.xyz	\n" // -2(D.N)N + D == D - 2(D.N)N
"mov r4.w, c[13].y						\n" // force w=1

// Apply a texture transform matrix
"dp4 oT3.x, r4, c[19]					\n"
"dp4 oT3.y, r4, c[20]					\n"

// are we looking towards the sun?
"dp3 r3.x, r2.xyz, c[21].xyz			\n"
"slt r3.x, r3.x, c[13].x				\n"

// Fade based on camera distance
"mad r4.w, r10.z, c[22].x, c[22].y		\n"
"mul r4.w, r4.w, r3.x					\n" // force fade to zero if looking away
"min r4.w, r4.w, c[13].y				\n"
"max oT2.y, r4.w, c[13].x				\n"

// Finally, base pass texture coordinates
"add oT0.xy, v2.xy, c[22].zw			\n"*/

/* Attempt #2: generated from HLSL:

vertexOutput VS_Water(vertexInput IN) 
{
    vertexOutput OUT;
    
    // calculate phase values.
    float3 fPhase = IN.position.xxx * reg8_xScale_reflection.xyz
    			  + IN.position.zzz * reg10_zScale_alpha.xyz
    			  + tValues.xyz;
    fPhase = frac(fPhase);
    
    // plug these into the polynominals
    float3 fHeights = (fPhase*fPhase*fPhase) - 1.5 * (fPhase*fPhase) + 0.5 * fPhase;
    float3 fDerivatives = 3 * (fPhase*fPhase) - 3 * fPhase + 0.5;
    
    // calculate the vertex's camera space position as if we didn't displace it
    float3 camSpacePosition = mul( float4(IN.position.xyz, 1.0), view ).xyz;
    
    // calculate falloff
    float3 fFalloff = (2).xxx - (camSpacePosition.zzz / falloff.xyz);
    fFalloff = saturate(fFalloff);
    fHeights *= fFalloff;
    fDerivatives *= fFalloff;
    
    // scale according to the per-phase scaling constants
    fHeights *= yScale;
    fDerivatives *= yScale;
    
    // apply heights from phases 1 and 2 to the vertex
    float3 xDisplacedPosition = IN.position.xyz;
    xDisplacedPosition.y += fHeights.x + fHeights.y;
    
    // push the displaced position through view/proj
    OUT.hPosition = mul( mul( float4(xDisplacedPosition , 1.0) , view), proj);
    
    // calculate the vertex normals
    float3 fNormal, fNormalPass2;
    
    fNormal.x = dot(fDerivatives * reg8_xScale_reflection.xyz, (1).xxx);
    fNormal.y = -1;
    fNormal.z = dot(fDerivatives * reg10_zScale_alpha.xyz, (1).xxx);
    fNormalPass2 = normalize(fNormal);
    
    fNormal.y *= reg16_horizonParams.y;
    fNormal = normalize(fNormal);
    
    // get the direction from the camera
    float3 xCamDirection = normalize(xDisplacedPosition - reg17_cameraParams.xyz);
    //float3 xCamDirection = normalize(xDisplacedPosition - camPosition.xyz);
    
    // calculate the fresnel term using fNormal
    float fFresnel = (1 - abs(dot(xCamDirection, fNormal)));
    fFresnel *= fFresnel;
    
    OUT.parameters.x = fFresnel;
    
    // calculate the reflection vector
    float3 refl = xCamDirection - 2*dot(xCamDirection, fNormal) * fNormal;
    
    // convert that to horizon map texcoords
    OUT.horizonTexCoords.x = (((refl.x * reg18_camHeading_lightParams.y) - (refl.z * reg18_camHeading_lightParams.x)) / (2*3.141592)) + reg17_cameraParams.w;
    OUT.horizonTexCoords.y = (refl.y * reg16_horizonParams.w) + reg16_horizonParams.z + 1;
    
    // Lighting: (offset - scale*height) * original color, saturated
    OUT.lighting = saturate(IN.lightColor * (reg18_camHeading_lightParams.w - reg18_camHeading_lightParams.z * (fHeights.x + fHeights.y)));
    
    // sun reflection - get a new reflection vector
    refl = xCamDirection - 2*dot(xCamDirection, fNormalPass2) * fNormalPass2;
    OUT.sunTexCoords.x = dot(float4(refl, 1.0), texMat1);
    OUT.sunTexCoords.y = dot(float4(refl, 1.0), texMat2);
    
    float fSunFade = saturate(mul( float4(xDisplacedPosition , 1.0) , view).z * reg22_sunFalloff_texOffset.x + reg22_sunFalloff_texOffset.y);
	// if we're looking away from the sun, force it to zero
	if(dot(xCamDirection, sunDirection) < 0) fSunFade = 0;
	OUT.parameters.y = fSunFade;
    
    OUT.baseTexCoords = IN.texCoord.xy + reg22_sunFalloff_texOffset.zw;

	// fog
    float fRangePosition = OUT.hPosition.w * reg23_fogscale_fogtest.x + reg23_fogscale_fogtest.y;
    int iSectionID = (fRangePosition >= reg23_fogscale_fogtest.z) ? 1 : 0;
    float fSectionPosition = fRangePosition * reg2425_sectionscale[iSectionID].x + reg2425_sectionscale[iSectionID].y;
    float4 fPolynomial = float4(fSectionPosition * fSectionPosition * fSectionPosition, fSectionPosition * fSectionPosition, fSectionPosition, 1);
    OUT.fogValue = dot(fPolynomial, reg2627_curvecoefficients[iSectionID]);

    return OUT;
}
  */


// ORIGINAL VERSION THAT WORKS

"vs_2_0										"

"dcl_position	v0							"
"dcl_color		v1							"
"dcl_texcoord	v2							"
"mul	r0.xyz,	v0.z,	c10					"
"mad	r0.xyz, v0.x,	c8,		r0			"
"add	r0.xyz, r0,		c12					"
"frc	r3.xyz, r0							"
"mov	r1.xyz, v0							"
"mov	r1.w,	c13.x						"
"mul	r0.xy,	r3,		r3					"
"dp4	r0.w,	r1,		c2					"
"mul	r2.xy,	r0,		c14.x				"
"mov	r4.x,	c14.w						"
"mad	r0.w,	r0.w,	-c11.x,	r4.x		"
"mad	r0.xy,	r0,		r3,		-r2			"
"max	r0.w,	r0.w,	c13.y				"
"mad	r0.xy,	c14.y,	r3,		r0			"
"min	r2.w,	r0.w,	c13.x				"
"mul	r0.xy,	r0,		r2.w				"
"mul	r0.xy,	r0,		c9					"
"add	r0.w,	r0.y,	r0.x				"
"add	r1.y,	r0.w,	v0.y				"
"mad	r3.w,	c18.z,	-r0.w,	c18.w		"
"dp4	r0.x,	r1,		c0					"
"dp4	r0.y,	r1,		c1					"
"dp4	r0.w,	r1,		c3					"
"dp4	r0.z,	r1,		c2					"
"add	r1.xyz,	r1,		-c17				"

"dp4	oPos.x,	r0,		c4					"
"mul	r2.xyz,	r3,		r3					"
"mul	r3.xyz,	r3,		c14.z				"
"dp4	oPos.y,	r0,		c5					"
"mad	r2.xyz,	c14.z,	r2,		-r3			"
"dp4	oPos.z,	r0,		c6					"

"add	r2.xyz,	r2,		c14.y				"
"dp4	r4.x,	r0,		c7					"
"mul	r2.xyz,	r2.w,	r2					"
"mad	r2.w,	r0.z,	c22.x,	c22.y		"
"mul	r0.xyz,	r2,		c9					"
"nrm	r2.xyz,	r1							"
"mul	r1.xyz,	r0,		c8					"
"mul	r0.xyz,	r0,		c10					"
"dp3	r1.x,	r1,		c13.x				"
"dp3	r1.y,	r0,		c13.x				"
"mov	r0.zw,	c13							"
"mul	r1.w,	r0.z,	c16.y				"
"abs	r4.w,	r2.y						"
"nrm	r0.xyz,	r1.xwyw						"
"add	r1.w,	-r4.w,	c13.x				"
"dp3	r3.x,	r2,		r0					"
"mul	r1.w,	r1.w,	r1.w				"
"mul	r4.w,	r3.x,	c14.w				"
"max	r1.w,	r1.w,	c13.y				"
"mad	r0.xyz,	r4.w,	-r0,	r2			"

"min	oT2.x,	r1.w,	c13.x				"
"mul	r0.z,	r0.z,	c18.x				"
"mad	r4.w,	r0.x,	c18.y,	-r0.z		"
"mad	r1.w,	r0.y,	c16.w,	c16.z		"
"mad	oT1.x,	r4.w,	r0.w,	c17.w		"
"add	oT1.y,	r1.w,	c13.x				"

"mul	r0,		r3.w,	v1					"
"mov	r1.z,	c13.z						"
"max	r0,		r0,		c13.y				"

"min	r3,		r0,		c13.x				"
"mul	r3.w,	r3,		c10.w				"	// * base texture alpha scale
"mov	oD0,	r3							"

"nrm	r3.xyz,	r1.xzyw						"

"dp3	r0.x,	r2,		r3					"
"mul	r1.w,	r0.x,	c14.w				"
"mad	r0.w,	r4.x,	c23.x,	c23.y		"
"mad	r0.xyz,	r1.w,	-r3,	r2			"
"max	r0.w,	r0.w,	c13.y				"
"dp3	r1.x,	r2,		c21					"
"min	r3.w,	r0.w,	c13.x				"
"mov	r0.w,	c13.x						"
"sge	r4.w,	r3.w,	c23.z				"

"mul	r0.y,	r0.y,	c11.w				"
"dp4	oT3.x,	r0,		c19					"

"lrp	r1.w,	r4.w,	c13.x,	c13.y		"
"dp4	oT3.y,	r0,		c20					"

"frc	r4.w,	r1.w						"
"slt	r0.w,	c13.y,	r1.x				"
"add	r5.w,	r1.w,	-r4.w				"
"slt	r1.w,	r1.w,	-r1.w				"
"slt	r4.w,	-r4.w,	r4.w				"
"max	r2.w,	r2.w,	c13.y				"
"mad	r1.w,	r1.w,	r4.w,	r5.w		"
"min	r2.w,	r2.w,	c13.x				"
"mova	a0.w,	r1.w						"
"add	oT2.y,	-r2.w,	c13.x				" // r1.w = (sun falloff) invert to "1" at near plane
"mad	r0.z,	r3.w,	c24[a0.w].x, c24[a0.w].y "

"mul	r0.y,	r0.z,	r0.z				"
"mul	r0.x,	r0.z,	r0.y				"
"mov	r0.w,	c13.x						"
"add	oT0.xy,	v2,		c22.zwzw			"
"dp4	r0.w,	r0,		c26[a0.w]			"
"mov	oPos.w,	r4.x						"
"add	oFog,	-r0.w,	c13.x				"
;

// Register the shader
static const Asura_Direct3D_VertexShader xVS_WATER_SHADER(ASURA_D3D_VS_WATER, szVS_WATER_CODE, auVSD, 2.0f);


/////////////////////////////////////////////////////////////////////////////////////////////////////////
///Function:			szVS_WATER_CODE_vs11
///Creation Date:		27/03/2006
///Author:				DH 
///
///Description:			Water Shader for vs_1_1
///						This was a quick hack-around to get it to work in VS_1_1 ( will tidy it up later )
///
/////////////////////////////////////////////////////////////////////////////////////////////////////////

static const char* const szVS_WATER_CODE_vs11 =

"vs_1_1										"

"dcl_position	v0							"
"dcl_color		v1							"
"dcl_texcoord	v2							"

"mul	r0.xyz,	v0.z,	c10					"
"mad	r0.xyz, v0.x,	c8,		r0			"
"add	r0.xyz, r0,		c12					"
"frc	r3.xy,	r0							"		// frc	r3.xyz,	r0
"sge	r3.z,	r0.x,	r0.x				"		// set Z to 1

"mov	r1.xyz, v0							"
"mov	r1.w,	c13.x						"
"mul	r0.xy,	r3,		r3					"
"dp4	r0.w,	r1,		c2					"
"mul	r2.xy,	r0,		c14.x				"
"mov	r4.x,	c14.w						"
"mad	r0.w,	r0.w,	-c11.x,	r4.x		"
"mad	r0.xy,	r0,		r3,		-r2			"
"max	r0.w,	r0.w,	c13.y				"
"mad	r0.xy,	c14.y,	r3,		r0			"
"min	r2.w,	r0.w,	c13.x				"
"mul	r0.xy,	r0,		r2.w				"
"mul	r0.xy,	r0,		c9					"
"add	r0.w,	r0.y,	r0.x				"
"add	r1.y,	r0.w,	v0.y				"
"mad	r3.w,	c18.z,	-r0.w,	c18.w		"
"dp4	r0.x,	r1,		c0					"
"dp4	r0.y,	r1,		c1					"
"dp4	r0.w,	r1,		c3					"
"dp4	r0.z,	r1,		c2					"
"add	r1.xyz,	r1,		-c17				"

"dp4	oPos.x,	r0,		c4					"
"mul	r2.xyz,	r3,		r3					"
"mul	r3.xyz,	r3,		c14.z				"
"dp4	oPos.y,	r0,		c5					"
"mad	r2.xyz,	c14.z,	r2,		-r3			"
"dp4	oPos.z,	r0,		c6					"

"add	r2.xyz,	r2,		c14.y				"
"dp4	r4.x,	r0,		c7					"
"mul	r2.xyz,	r2.w,	r2					"
"mad	r2.w,	r0.z,	c22.x,	c22.y		"
"mul	r0.xyz,	r2,		c9					"

//"nrm	r2.xyz,	r1							"
"dp3	r2,		r1,		r1					"
"rsq	r2.w,	r2.x						"
"mul	r2.xyz,	r1.xyz,	r2.w				"

"mul	r1.xyz,	r0,		c8					"
"mul	r0.xyz,	r0,		c10					"
"dp3	r1.x,	r1,		c13.x				"
"dp3	r1.y,	r0,		c13.x				"
"mov	r0.zw,	c13							"
"mul	r1.w,	r0.z,	c16.y				"

//"abs	r4.w,	r2.y						"
"mov	r4.w,	r2.y						"

//"nrm	r0.xyz,	r1.xwyw						"
"dp3	r0,		r1,		r1					"
"rsq	r0.w,	r0.x						"
"mul	r0.xyz,	r1.xyz,	r0.w				"

"add	r1.w,	-r4.w,	c13.x				"
"dp3	r3.x,	r2,		r0					"
"mul	r1.w,	r1.w,	r1.w				"
"mul	r4.w,	r3.x,	c14.w				"
"max	r1.w,	r1.w,	c13.y				"
"mad	r0.xyz,	r4.w,	-r0,	r2			"

"min	oT2.x,	r1.w,	c13.x				"
"mul	r0.z,	r0.z,	c18.x				"
"mad	r4.w,	r0.x,	c18.y,	-r0.z		"
"mad	r1.w,	r0.y,	c16.w,	c16.z		"
"mad	oT1.x,	r4.w,	r0.w,	c17.w		"
"add	oT1.y,	r1.w,	c13.x				"

"mul	r0,		r3.w,	v1					"
"mov	r1.z,	c13.z						"
"max	r0,		r0,		c13.y				"

"min	r3,		r0,		c13.x				"
"mul	r3.w,	r3,		c10.w				"	// * base texture alpha scale
"mov	oD0,	r3							"

//"nrm	r3.xyz,	r1.xzyw						"
"dp3	r3,		r1,		r1					"
"rsq	r3.w,	r3.x						"
"mul	r3.xyz,	r1.xyz,	r3.w				"

"dp3	r0.x,	r2,		r3					"
"mul	r1.w,	r0.x,	c14.w				"
"mad	r0.w,	r4.x,	c23.x,	c23.y		"
"mad	r0.xyz,	r1.w,	-r3,	r2			"
"max	r0.w,	r0.w,	c13.y				"
"dp3	r1.x,	r2,		c21					"
"min	r3.w,	r0.w,	c13.x				"
"mov	r0.w,	c13.x						"
"sge	r4.w,	r3.w,	c23.z				"

"mul	r0.y,	r0.y,	c11.w				"
"dp4	oT3.x,	r0,		c19					"

//"lrp	r1.w,	r4.w,	c13.x,	c13.y		"
"add	r1.w,	c13.x,	-c13.y				"	
"mad	r1.w,	r4.w,	r1.w,	c13.y		"

"dp4	oT3.y,	r0,		c20					"

//"frc	r4.w,	r1.w						"
"frc	r4.y,	r1.w						"
"mov	r4.w,	r4.y						"


"slt	r0.w,	c13.y,	r1.x				"
"add	r5.w,	r1.w,	-r4.w				"
"slt	r1.w,	r1.w,	-r1.w				"
"slt	r4.w,	-r4.w,	r4.w				"
"max	r2.w,	r2.w,	c13.y				"
"mad	r1.w,	r1.w,	r4.w,	r5.w		"
"min	r2.w,	r2.w,	c13.x				"
"add	oT2.y,	-r2.w,	c13.x				"		// r1.w = (sun falloff) invert to "1" at near plane

//"mova	a0.w,	r1.w						"
//"mad	r0.z,	r3.w,	c24[a0.w].x, c24[a0.w].y "
"sge	r0.z,	r3.w,	r3.w				"

"mul	r0.y,	r0.z,	r0.z				"
"mul	r0.x,	r0.z,	r0.y				"
"mov	r0.w,	c13.x						"
"add	oT0.xy,	v2,		c22.zwzw			"
//"dp4	r0.w,	r0,		c26[a0.w]			"
"mov	oPos.w,	r4.x						"
"mov	oFog,	r0.x						"
;

// Register the shader
static const Asura_Direct3D_VertexShader xVS_WATER_SHADER_vs11(ASURA_D3D_VS_WATER_VS11, szVS_WATER_CODE_vs11, auVSD, 1.0f);


#endif


