/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"
#include "Deferred_Shared.fxh"


// Light clip space -> Light view space -> World space -> Camera view space -> Camera clip space
const float4x4 g_xLightClipSpaceToCameraClipSpace;

// Camera view space -> World space -> Light view space -> Light clip space
const float4x4 g_xCameraViewSpaceToLightClipSpace;

// Camera view space -> World space -> Light view space
const float4x4 g_xCameraViewSpaceToLightViewSpace;

// Camera view space -> World space
const float4x4 g_xCameraViewSpaceToWorldSpace;
// shadow caster position in world space
const float4 g_xLightPos;

const float4x4 g_xUVXForm;


// 2 / dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float g_fTwoOverScreenWidth;
const float g_fTwoOverScreenHeight;

// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
const float g_fCameraTanHalfHorizFOV;
const float g_fCameraTanHalfVertFOV;

// Is shadow filtering turned on?
const bool g_bUseNoise;
const float g_fShadowFilterTextureScale;
const float g_fShadowFilterSize;

const float4 g_xShadowDirection;
const float  g_fShadowStrength;

const bool   g_bCreatingShadowMask;

// x is multiplied by the dot product, y is just added
const float3   g_xSelfShadowBias;

const texture g_xNormalDepthBuffer;
const sampler g_xNormalDepthBufferSampler : register(s0) = sampler_state
{
	Texture = <g_xNormalDepthBuffer>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MagFilter = POINT;
	MinFilter = POINT;
	MipFilter = POINT;
};

const texture g_xShadowBuffer;
const sampler g_xShadowBufferSampler : register(s1) = sampler_state
{
	Texture = <g_xShadowBuffer>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MagFilter = LINEAR;
	MinFilter = LINEAR;
	MipFilter = POINT;
};

const texture g_xShadowBufferCube;
const sampler g_xShadowBufferCubeSampler = sampler_state
{
	Texture = <g_xShadowBufferCube>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MagFilter = LINEAR;
	MinFilter = LINEAR;
	MipFilter = POINT;
};



const texture g_xKernelRotationTexture;
const sampler g_xKernelRotationSampler : register(s2) = sampler_state
{
	Texture = <g_xKernelRotationTexture>;
	AddressU = WRAP;
	AddressV = WRAP;
	MagFilter = LINEAR;
	MinFilter = LINEAR;
	MipFilter = LINEAR;
};




/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

static const float fTEXTURE_SIZE = 512;

#define NUM_TAPS 12

#if (NUM_TAPS > 0)
static const float2 axFILTER_TAPS[ NUM_TAPS ] =
{
#if (NUM_TAPS == 8)
	{ -0.840144, -0.073580 },
	{ -0.695914,  0.457137 },
	{ -0.203345,  0.620716 },
	{  0.473434, -0.480026 },
	{  0.519456,  0.767022 },
	{  0.185461, -0.893124 },
	{  0.896420,  0.412458 },
	{ -0.321940, -0.932615 },
#elif (NUM_TAPS == 12)
	{ -0.326212, -0.405805 },
	{ -0.840144, -0.073580 },
	{ -0.695914,  0.457137 },
	{ -0.203345,  0.620716 },
	{  0.962340, -0.194983 },
	{  0.473434, -0.480026 },
	{  0.519456,  0.767022 },
	{  0.185461, -0.893124 },
	{  0.507431,  0.064425 },
	{  0.896420,  0.412458 },
	{ -0.321940, -0.932615 },
	{ -0.791559, -0.597705 },
#endif	// (NUM_TAPS == 12)
};
#endif

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// The vertex shader recreates the shadow frustum in world space, then projects
// it as an object in the standard camera view

float4 VShader( const float4 xInPos : POSITION ) : POSITION
{
	return mul( xInPos, g_xLightClipSpaceToCameraClipSpace );
}

float4 GetCameraViewSpacePosition(const float2 xScreenPos, out half3 xNormal)
{
	// Fetch the world-space normal and screen depth
	const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * float2(g_fTwoOverScreenWidth, g_fTwoOverScreenHeight) / 2;
	const float4 xBufferNormal = tex2D(g_xNormalDepthBufferSampler, xNormalizedCoords);
	
	float fScreenDepth;
	
	Deferred_GetNormalDepth( xBufferNormal, xNormal, fScreenDepth);
	
	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	return float4(	g_fCameraTanHalfHorizFOV * fScreenDepth * (xScreenPos.x * g_fTwoOverScreenWidth  - 1.0),
													g_fCameraTanHalfVertFOV  * fScreenDepth * -(xScreenPos.y * g_fTwoOverScreenHeight - 1.0),
													fScreenDepth,
													1.0);
}

float3 GetCubicMap_ShadowUV(const float4 xCameraViewSpacePosition, out float fSceneLightspaceDepth, out float4 xLightVector)
{
	float4 xWorldSpacePosition = mul(xCameraViewSpacePosition, g_xCameraViewSpaceToWorldSpace);
	xLightVector = g_xLightPos - xWorldSpacePosition;
	
	fSceneLightspaceDepth = length(xLightVector);
	
	xLightVector /= fSceneLightspaceDepth;

	fSceneLightspaceDepth = (fSceneLightspaceDepth - g_fShadowFrustumNearPlaneDistance ) * g_fOneOverShadowFrustumPlaneSeparation;
	
	float3 xShadowBufferTexcoords;
	xShadowBufferTexcoords = xLightVector;
	xShadowBufferTexcoords.xz = -xShadowBufferTexcoords.xz;

	return xShadowBufferTexcoords;
}

float GetProjected_Depth(const float4 xCameraViewSpacePosition)
{
	// We want a linear distribution of depth in the shadow buffer so we don't divide z by w - we use the near and far planes of the shadow
	// frustum instead.
	float4 xLightViewSpacePosition = mul(xCameraViewSpacePosition, g_xCameraViewSpaceToLightViewSpace);
	return (xLightViewSpacePosition.z - g_fShadowFrustumNearPlaneDistance ) * g_fOneOverShadowFrustumPlaneSeparation;

}


float4 GetProjected_ShadowUV(const float4 xCameraViewSpacePosition)
{
	float4 xShadowBufferTexcoords = mul(xCameraViewSpacePosition, g_xCameraViewSpaceToLightClipSpace);
	xShadowBufferTexcoords =  mul(g_xUVXForm, xShadowBufferTexcoords);
	return xShadowBufferTexcoords;
}

float2 GetPerPixelRotation(float2 xUV, float fScale)
{
	float2 fSinCos;
	if (g_bUseNoise) 
	{
		fSinCos = (tex2D(g_xKernelRotationSampler, xUV*16).gb *2.0f - 1.0f) * fScale * 0.66f;
		fSinCos += (tex2D(g_xKernelRotationSampler, xUV*64).gb *2.0f - 1.0f) * fScale * 0.33f;
	} 
	else 
	{
		fSinCos = float2(fScale,0);
	}
	return fSinCos;
}

float SampleShadowBuffer(float3 xShadowBufferTexcoords, const bool bUseCubic, float fSceneLightspaceDepth)
{
	float fOccluderLightspaceDepth;
	if (bUseCubic)
	{
		fOccluderLightspaceDepth = texCUBE(g_xShadowBufferCubeSampler, xShadowBufferTexcoords).r;
	} 
	else 
	{
		fOccluderLightspaceDepth = tex2D(g_xShadowBufferSampler, xShadowBufferTexcoords.xy).r;
	}
	return fSceneLightspaceDepth - fOccluderLightspaceDepth;
}

float IsShadowed_SingleTap(float4 xShadowBufferTexcoords, float fSceneLightspaceDepth, uniform bool bUseCubic, uniform bool bDepthMap)
{
	float fShadow;
	if (bDepthMap)
	{
		fShadow = 1.0f - tex2Dproj(g_xShadowBufferSampler, xShadowBufferTexcoords);
	} 
	else 
	{
		fShadow = 0 < SampleShadowBuffer(xShadowBufferTexcoords.xyz, bUseCubic, fSceneLightspaceDepth);
	}
	return fShadow;
}

float IsShadowed_MultiTap(float4 xShadowBufferTexcoords, float fSceneLightspaceDepth, uniform bool bUseCubic, uniform bool bDepthMap)
{
	float fShadow = 0;
	
	// Regardless of whether this pixel is actually shadowed, allow 'nearby' occluders to contribute to the shadow value

	// Calculate a value to scale the sampling region on the shadow buffer by
	const float fScaling = max(0.0f, g_fShadowFilterSize + saturate(2.0f * fSceneLightspaceDepth - 1.0f) * 2.0);
	const float fScale = g_fShadowFilterTextureScale * fScaling * (bDepthMap ? xShadowBufferTexcoords.w : 1.0f);

	// To reduce banding, we randomly rotate the filter kernel per-pixel. Uses a texture of pre-generated	
	// sin/cos for random rotation angles
	float2 fSinCos = GetPerPixelRotation(bDepthMap ? xShadowBufferTexcoords.xy / xShadowBufferTexcoords.w : xShadowBufferTexcoords.xy, fScale);

	// Tweak this constant to balance self-shadowing artifacts with missing shadows near character feet.		
	float fMaxDist = 0.005f * g_fOneOverShadowFrustumPlaneSeparation;
	
	// For each tap,
	for( int i = 0; i < NUM_TAPS; ++i )
	{
		// calculate the offset texture coordinates for sampling the occluder depth			
		float3 xOffset;						               
		xOffset.x = dot(axFILTER_TAPS[i].xy , fSinCos);
		xOffset.y = dot(axFILTER_TAPS[i].xy , float2(fSinCos.y, -fSinCos.x));
		xOffset.z = 0.0f;
		

		if (bDepthMap)
		{
			const float4 xTapCoord = xShadowBufferTexcoords + xOffset.xyzz;
			fShadow += IsShadowed_SingleTap(xTapCoord, fSceneLightspaceDepth, false, true);
		} 
		else 
		{
			const float4 xTapCoord = xShadowBufferTexcoords + xOffset.xyxy;
			// See how far away this pixel is from the occluder
			float fShadow2 = SampleShadowBuffer(xTapCoord, bUseCubic, fSceneLightspaceDepth).r;
			
			// If the occluder is behind the pixel, fShadow2 has a negative sign - translate this into 'no shadow'
			fShadow += clamp( fShadow2, 0, fMaxDist );
		}
	}
	// Average out the contributions across all taps
	if (bDepthMap)
	{
		fShadow /= (NUM_TAPS);	
	} 
	else 
	{
		fShadow /= (fMaxDist * NUM_TAPS);	
	}
	
	return fShadow;	
}

float ScaleShadow(float fShadow, float fDot, float fDistScale)
{
	fShadow = fShadow * g_fShadowStrength;
	
	if (!g_bCreatingShadowMask)
	{
		fShadow = saturate(fShadow * fDistScale);		
		
	//	Taking this out, as it causes more problems than it fixes in SS2... shadows of a helmet on a character's face look particularly bad...
		fShadow *= fDot;
	} 
	return fShadow;
}

float BiasDepthValue(float fSceneLightspaceDepth, float3 xLightVector, half3 xNormal, out float fDot)
{
	fDot = saturate( dot( xLightVector, xNormal) );	
	fSceneLightspaceDepth -= ((1.0f - fDot) * g_xSelfShadowBias.x + g_xSelfShadowBias.y + fSceneLightspaceDepth * g_xSelfShadowBias.z);
	fDot = saturate(2.0f * fDot);

	return fSceneLightspaceDepth;
}

float4 CalculateShadowParams(const float2 xScreenPos, uniform bool g_bCubicMapShadow, uniform bool g_bDepthMap, out float fDot, out float fSceneLightspaceDepth)
{
	half3 xNormal;
	const float4 xCameraViewSpacePosition = GetCameraViewSpacePosition(xScreenPos, xNormal);

	float4 xShadowBufferTexcoords;

	if (g_bCubicMapShadow)
	{
		float4 xLightVector;
		xShadowBufferTexcoords = float4(GetCubicMap_ShadowUV(xCameraViewSpacePosition, fSceneLightspaceDepth, xLightVector), 1);
		

		fSceneLightspaceDepth = BiasDepthValue(fSceneLightspaceDepth, xLightVector.xyz, xNormal, fDot);

	} else {
		xShadowBufferTexcoords = GetProjected_ShadowUV(xCameraViewSpacePosition);

		float3 xShadowClipSpace = xShadowBufferTexcoords.xyz / xShadowBufferTexcoords.w;

 		clip( xShadowClipSpace);
 		clip( 1 - xShadowClipSpace);

		// fSceneLightspaceDepth is the scene depth from the light's POV
		fSceneLightspaceDepth = GetProjected_Depth(xCameraViewSpacePosition);

		if (g_bDepthMap)
		{
			xShadowBufferTexcoords.z = BiasDepthValue(xShadowBufferTexcoords.z, g_xShadowDirection, xNormal, fDot);
		} else {
			// Discard if we're behind the shadow near plane
			clip( fSceneLightspaceDepth);

			xShadowBufferTexcoords = float4(xShadowClipSpace.xyz, 1.0f);

			fSceneLightspaceDepth = BiasDepthValue(fSceneLightspaceDepth, g_xShadowDirection, xNormal, fDot);
		}

	}

	return xShadowBufferTexcoords;
}

float4 PShader_MultiTap( const float2 xScreenPos : VPOS, uniform bool g_bCubicMapShadow, uniform bool g_bDepthMap) : COLOR
{
	float fDot;
	float fSceneLightspaceDepth;

	float4 xShadowBufferTexcoords = CalculateShadowParams(xScreenPos, g_bCubicMapShadow, g_bDepthMap, fDot, fSceneLightspaceDepth);

	float fShadow = IsShadowed_MultiTap(xShadowBufferTexcoords, fSceneLightspaceDepth, g_bCubicMapShadow, g_bDepthMap);

	// Scale the shadow according to how far this pixel is away from the light
	return float4( 0.0, 0.0, 0.0, saturate( ScaleShadow(fShadow, fDot, 1.0f - fSceneLightspaceDepth) ) );	
}

float4 PShader_SingleTap( const float2 xScreenPos : VPOS, uniform bool g_bCubicMapShadow, uniform bool g_bDepthMap) : COLOR
{
	float fDot;
	float fSceneLightspaceDepth;

	float4 xShadowBufferTexcoords = CalculateShadowParams(xScreenPos, g_bCubicMapShadow, g_bDepthMap, fDot, fSceneLightspaceDepth);

	float fShadow = IsShadowed_SingleTap(xShadowBufferTexcoords, fSceneLightspaceDepth, g_bCubicMapShadow, g_bDepthMap);

	// Scale the shadow according to how far this pixel is away from the light
	return float4( 0.0, 0.0, 0.0, saturate( ScaleShadow(fShadow, fDot, 1.0f - fSceneLightspaceDepth) ) );	

}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false, false);
	}
}
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false, false);
	}
}
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false, true);
	}
}
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false, true);
	}
}
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(true, false);
	}
}
technique
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(true, false);
	}
}

