/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Deferred_Shared.fxh"
#include "ParallaxMap.fxh"
#include "Fog.fxh"

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#define ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH (252.0f)
#define DIFFUSE_TEXEL                       (ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH + 1.5f)
#define SPECULAR_TEXEL                      (DIFFUSE_TEXEL + 1.0f)
#define MISC_TEXEL                          (SPECULAR_TEXEL + 1.0f)

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float4 g_xUV_Offsets : register(c1);

// These are only used by objects, hierarchies and skins. They're constant since we just use one SH
// sample
const float4 g_xRedSH   : register( ps, c128 );
const float4 g_xGreenSH : register( ps, c129 );
const float4 g_xBlueSH  : register( ps, c130 );

// These are used when rendering transparent things, after the deferred lights
const float3 g_xSHLightDirection : register( ps, c131 );

// Brightness in w, to avoid computing in shader
const float4 g_xSHLightColour    : register( ps, c132 );

// If not specified, assume colourise values and metalicity are between 0 and 1, so we have to do
// the texture reads and lerps for them
#if !defined( COLOURISE_DIFFUSE )
	#define COLOURISE_DIFFUSE 2
#endif
#if !defined( COLOURISE_SPECULAR )
	#define COLOURISE_SPECULAR 2
#endif
#if !defined( METALICITY )
	#define METALICITY 2
#endif

#if defined( NOT_EFFECT )

// Pack values to save constant registers
// x = (material index + 0.5) / materials texture height
// y = reflectivity value
// z = parallax scale (~0.04)
// w = parallax bias (~-0.02)
const float4 g_xMaterialProperties : register(c0);

// x = 1 / materials texture width
// y = materials texture value scale
const float2 g_xMaterialsTextureProperties : register(c1);

// x   = Albedo UV Scale
// y   = Albedo Value Scale;
// z   = Normal UV Scale
// w   = Normal Value Scale;
const float4 g_xDetailMapParameters : register(c2);

// x   = Albedo UV Scale
// y   = Albedo Value Scale;
// z   = Normal UV Scale
// w   = Normal Value Scale;
const float4 g_xDetailMap2Parameters : register(c3);

// xyz = self-illumination amount * picked self-illumination colour * self-illumination colourise percent
// w   = self-illumination amount * (1.0f - self illumination colourise percent)
const float4 g_xSelfIlluminationParameters : register(c4);

// This is multiplied with albedo, for character customisation etc. Set to (1,1,1) for no effect
const float3 g_xAlbedoModifier : register(c5);

// Ensure the materials data texture is always in sampler 0
sampler2D   g_xMaterialsTextureSampler : register(s0);

// On PC SM 3.0, only have 16 samplers available. Don't bind these to specific registers, we'll
// discover where they've gone at runtime. Can't use all slots at the same time though

sampler2D   g_xAlbedoSampler;
sampler2D   g_xAlbedo2Sampler;
sampler2D   g_xNormalMapSampler;
sampler2D   g_xNormalMap2Sampler;
sampler2D   g_xSpecularMapSampler;
sampler2D   g_xSpecularMap2Sampler;
sampler2D   g_xTransparencyMapSampler;
sampler2D   g_xTransparencyMap2Sampler;
sampler2D   g_xSelfIlluminationMapSampler;
sampler2D   g_xSelfIlluminationMap2Sampler;
sampler2D   g_xParallaxMapSampler;
sampler2D   g_xParallaxMap2Sampler;
samplerCUBE g_xEnvironmentMapSampler;
sampler2D   g_xReflectivityMaskSampler;
sampler2D   g_xAlbedoDetailMapSampler;
sampler2D   g_xAlbedoDetailMap2Sampler;
sampler2D   g_xNormalDetailMapSampler;
sampler2D   g_xNormalDetailMap2Sampler;

#if !defined( USE_ALBEDO_1 )
	#define USE_ALBEDO_1 false
#endif
#if !defined( USE_ALBEDO_2 )
	#define USE_ALBEDO_2 false
#endif
#if !defined( USE_NORMAL_1 )
	#define USE_NORMAL_1 false
#endif
#if !defined( USE_NORMAL_2 )
	#define USE_NORMAL_2 false
#endif
#if !defined( USE_SPECULAR_1 )
	#define USE_SPECULAR_1 false
#endif
#if !defined( USE_SPECULAR_2 )
	#define USE_SPECULAR_2 false
#endif
#if !defined( USE_TRANSPARENCY_1 )
	#define USE_TRANSPARENCY_1 false
#endif
#if !defined( USE_TRANSPARENCY_2 )
	#define USE_TRANSPARENCY_2 false
#endif
#if !defined( USE_SELFILLUMINATION_1 )
	#define USE_SELFILLUMINATION_1 false
#endif
#if !defined( USE_SELFILLUMINATION_2 )
	#define USE_SELFILLUMINATION_2 false
#endif
#if !defined( USE_PARALLAX_1 )
	#define USE_PARALLAX_1 false
#endif
#if !defined( USE_PARALLAX_2 )
	#define USE_PARALLAX_2 false
#endif
#if !defined( USE_ENVIRONMENT_1 )
	#define USE_ENVIRONMENT_1 false
#endif
#if !defined( USE_REFLECTIVITY_1 )
	#define USE_REFLECTIVITY_1 false
#endif
#if !defined( USE_ALBEDO_DETAIL_MAP_1 )
	#define USE_ALBEDO_DETAIL_MAP_1 false
#endif
#if !defined( USE_ALBEDO_DETAIL_MAP_2 )
	#define USE_ALBEDO_DETAIL_MAP_2 false
#endif
#if !defined( USE_NORMAL_DETAIL_MAP_1 )
	#define USE_NORMAL_DETAIL_MAP_1 false
#endif
#if !defined( USE_NORMAL_DETAIL_MAP_2 )
	#define USE_NORMAL_DETAIL_MAP_2 false
#endif
#if !defined( IGNORE_SPECULAR )
	#define IGNORE_SPECULAR false
#endif

static bool g_bUseAlbedo               = USE_ALBEDO_1;
static bool g_bDualAlbedo              = USE_ALBEDO_2;
static bool g_bUseNormalMap            = USE_NORMAL_1;
static bool g_bDualNormalMap           = USE_NORMAL_2;
static bool g_bUseSpecularMap          = USE_SPECULAR_1;
static bool g_bDualSpecularMap         = USE_SPECULAR_2;
static bool g_bUseTransparencyMap      = USE_TRANSPARENCY_1;
static bool g_bDualTransparencyMap     = USE_TRANSPARENCY_2;
static bool g_bUseSelfIlluminationMap  = USE_SELFILLUMINATION_1;
static bool g_bDualSelfIlluminationMap = USE_SELFILLUMINATION_2;
static bool g_bUseParallaxMap          = USE_PARALLAX_1;
static bool g_bDualParallaxMap         = USE_PARALLAX_2;
static bool g_bUseEnvironmentMap       = USE_ENVIRONMENT_1;
static bool g_bUseReflectivityMask     = USE_REFLECTIVITY_1;
static bool g_bUseAlbedoDetailMap      = USE_ALBEDO_DETAIL_MAP_1;
static bool g_bDualAlbedoDetailMap     = USE_ALBEDO_DETAIL_MAP_2;
static bool g_bUseNormalDetailMap      = USE_NORMAL_DETAIL_MAP_1;
static bool g_bDualNormalDetailMap     = USE_NORMAL_DETAIL_MAP_2;

static bool g_bIgnoreSpecular          = IGNORE_SPECULAR;

#else // !defined( NOT_EFFECT )

// Pack values to save constant registers
// x = (material index + 0.5) / materials texture height
// y = reflectivity value
// z = parallax scale (~0.04)
// w = parallax bias (~-0.02)
const float4 g_xMaterialProperties;

// x = 1 / materials texture width
// y = materials texture value scale
const float2 g_xMaterialsTextureProperties;

// xyz = self-illumination amount * picked self-illumination colour * self-illumination colourise percent
// w   = self-illumination amount * (1.0f - self illumination colourise percent)
const float4 g_xSelfIlluminationParameters;

// This is multiplied with albedo, for character customisation etc. Set to (1,1,1) for no effect
const float3 g_xAlbedoModifier = { 1.0f, 1.0f, 1.0f };

#define DEFINE_SAMPLER( Name, sRGB )					\
	const texture g_x##Name;							\
														\
	const sampler2D g_x##Name##Sampler = sampler_state	\
	{													\
		Texture = <g_x##Name>;							\
		AddressU = WRAP;								\
		AddressV = WRAP;								\
		MinFilter = ANISOTROPIC;						\
		MagFilter = LINEAR;								\
		MipFilter = LINEAR;								\
		sRGBTexture = sRGB;								\
	};

#define DEFINE_DUAL_SAMPLERS( Name, sRGB )	\
	const bool g_bUse##Name;				\
	DEFINE_SAMPLER( Name, sRGB );			\
											\
	const bool g_bDual##Name;				\
	DEFINE_SAMPLER( Name##2, sRGB );		\

DEFINE_DUAL_SAMPLERS( Albedo, <g_bUseGammaCorrection> );
DEFINE_DUAL_SAMPLERS( NormalMap, FALSE );
DEFINE_DUAL_SAMPLERS( SpecularMap, FALSE );
DEFINE_DUAL_SAMPLERS( TransparencyMap, FALSE );
DEFINE_DUAL_SAMPLERS( SelfIlluminationMap, FALSE );
DEFINE_DUAL_SAMPLERS( ParallaxMap, FALSE );

const bool g_bUseEnvironmentMap;
const texture g_xEnvironmentMap;
const samplerCUBE g_xEnvironmentMapSampler = sampler_state
{
	Texture = <g_xEnvironmentMap>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	sRGBTexture = <g_bUseGammaCorrection>;
};

const bool g_bUseReflectivityMask;
DEFINE_SAMPLER( ReflectivityMask, FALSE );

// Can't use DEFINE_SAMPLER for this one, as we want some different states
// (Note - needs to be clamped for objects using the default material - at the bottom of the texture, v=1)
const texture g_xMaterialsTexture;
const sampler2D g_xMaterialsTextureSampler = sampler_state
{
	Texture = <g_xMaterialsTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = NONE;
	sRGBTexture = FALSE;
};

#undef DEFINE_DUAL_SAMPLERS
#undef DEFINE_SAMPLER

// Can't use these in the effect versions, since we're out of samplers
static bool g_bUseNormalDetailMap = false;
static bool g_bDualNormalDetailMap = false;
sampler2D g_xNormalDetailMapSampler;
sampler2D g_xNormalDetailMap2Sampler;
static float4 g_xDetailMapParameters;
static float4 g_xDetailMap2Parameters;

static bool g_bIgnoreSpecular = false;

#endif // !defined( NOT_EFFECT )

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// Simple implementation on PC to make it easier to diff this file with 360 version
float4 SampleTexture( sampler2D xSampler, float2 xUV )
{
	return tex2D( xSampler, xUV );
}

float GetLuminance( const float3 xColour )
{
	const float3 xLUMINANCE_CONSTANTS = { 0.27f, 0.67f, 0.06f };
	return dot( xColour, xLUMINANCE_CONSTANTS );
}

float3 SHLookup( const float4 xNormal, const float4 xRedSH, const float4 xGreenSH,
				 const float4 xBlueSH )
{
	float3 xColour;

	xColour.r = dot( xNormal, xRedSH );
	xColour.g = dot( xNormal, xGreenSH );
	xColour.b = dot( xNormal, xBlueSH );

	xColour = max( xColour, 0.0 );

	return xColour;
}

half4 MaterialResponse_SampleDataTexture( const half fU, const half fV )
{
	const half2 xCoords = half2( fU * g_xMaterialsTextureProperties.x, fV );
	half4 xTex = SampleTexture( g_xMaterialsTextureSampler, xCoords );
	half4 xScaledTex = g_xMaterialsTextureProperties.y * xTex;
	
	return( xScaledTex);
}

half DiffuseGraph( const half fNdotL, const half fV )
{
	const half fU = fNdotL * 0.5h * (ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH - 1.0h)
				  + 0.5h * ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH;

	const half fSample = MaterialResponse_SampleDataTexture( fU, fV ).r;

	return fSample;
}

half SpecularGraph( const half fNdotH, const half fV )
{
	const half fU = fNdotH * 0.5h * (ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH - 1.0h)
				  + 0.5h * ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH;

	const half fSample = MaterialResponse_SampleDataTexture( fU, fV ).g;

	return fSample;
}

half TransparencyGraph( const half fNdotV, const half fV )
{
	const half fU = fNdotV * 0.5h * (ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH - 1.0h)
				  + 0.5h * ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH;

	const half fSample = MaterialResponse_SampleDataTexture( fU, fV ).b;

	return fSample;
}

half ParallaxLookup( const half4 xUVPair, const half hDualTextureLerp )
{
	// We can assume we've got a texture bound, since we'll only call this function if we're doing
	// parallax mapping

	half hHeight = SampleTexture( g_xParallaxMapSampler, xUVPair.xy ).a;

	#if defined( NOT_EFFECT )
		if( g_bDualParallaxMap )
		{
			const half hHeight2 = SampleTexture( g_xParallaxMap2Sampler, xUVPair.zw ).a;
			hHeight = lerp( hHeight2, hHeight, hDualTextureLerp );
		}
	#endif

	return hHeight * g_xMaterialProperties.z + g_xMaterialProperties.w;
}

void ApplyParallaxMapping( const float3 xCameraDirection, // Must be in tangent space
						   const half hDualTextureLerp,
						   inout float4 xUVPair )
{
	if( g_bUseParallaxMap )
	{
		half hInitialHeight = ParallaxLookup( xUVPair, hDualTextureLerp );

		half2 xUVOffset = xCameraDirection.xy * hInitialHeight;

		//for( int i = 0; i < 2; ++i )
		{
			half4 xModifiedUVPair = xUVPair;
			xModifiedUVPair += xUVOffset.xyxy;

			const half hModifiedHeight = ParallaxLookup( xModifiedUVPair, hDualTextureLerp );

			const half hParametric = CalculateLineIntersectionParametric( hInitialHeight, hModifiedHeight );

			xUVOffset *= hParametric;
		}

		xUVPair += xUVOffset.xyxy;
	}
}

half3 AlbedoLookup( const float4 xUVPair,
					const float fDualTextureLerp,
					const half3 xNormal,
					const half3 xCameraDirection,
					out half3 xOutputAlbedo )
{
	if( g_bUseAlbedo )
	{
		half3 xAlbedoTextureColour = SampleTexture( g_xAlbedoSampler, xUVPair.xy ).rgb;

#if defined( NOT_EFFECT )
		if (g_bUseAlbedoDetailMap)
		{
			half xDetailTextureColour = SampleTexture( g_xAlbedoDetailMapSampler, xUVPair.xy * g_xDetailMapParameters.x).a * 2.0f - 1.0f;
			xDetailTextureColour *= g_xDetailMapParameters.y;
			xAlbedoTextureColour = saturate(xAlbedoTextureColour * (1 + xDetailTextureColour));
		}
#endif
		if( g_bDualAlbedo )
		{
			half3 xAlbedo2 = SampleTexture( g_xAlbedo2Sampler, xUVPair.zw ).rgb;
#if defined( NOT_EFFECT )
			if (g_bDualAlbedoDetailMap)
			{
				half xDetailTextureColour = SampleTexture( g_xAlbedoDetailMap2Sampler, xUVPair.zw * g_xDetailMap2Parameters.x).a * 2.0f - 1.0f;
				xDetailTextureColour *= g_xDetailMap2Parameters.y;
				xAlbedo2 = saturate(xAlbedo2 * (1 + xDetailTextureColour));
			}
#endif
			xAlbedoTextureColour = lerp( xAlbedo2, xAlbedoTextureColour, fDualTextureLerp );
		}

		xOutputAlbedo = xAlbedoTextureColour * g_xAlbedoModifier;
	}
	else
	{
		xOutputAlbedo = g_xAlbedoModifier;
	}

	// Environment mapping affects the albedo

	if( g_bUseEnvironmentMap )
	{
		// reflect() intrinsic wants direction from camera to point, not the standard V
		half3 xReflectionVector = reflect( -xCameraDirection, xNormal );

		// Asura y-coords are upside-down w.r.t. D3D
		xReflectionVector.y = -xReflectionVector.y;

		half3 xEnvironmentColour = texCUBE( g_xEnvironmentMapSampler, xReflectionVector ).rgb;

		half hReflectivity = g_xMaterialProperties.y;
		if( g_bUseReflectivityMask )
		{
			hReflectivity *= SampleTexture( g_xReflectivityMaskSampler, xUVPair.xy ).a;
		}

		xOutputAlbedo = lerp( xOutputAlbedo, xEnvironmentColour, hReflectivity );
	}

	// Do the x2 Asura thing. TODO: get rid of this
	return xOutputAlbedo * 2.0f;
}

half3 AlbedoAndSpecularLookup_DamageMap(const float4 xUVPair,
										const float fDualTextureLerp,
										const half3 xNormal,
										const half3 xCameraDirection,
										out half3 xOutputAlbedo,
										out half  hSpecular)
{

	if( g_bUseSpecularMap )
	{
		hSpecular = SampleTexture( g_xSpecularMapSampler, xUVPair.xy ).a;
	}
	else
	{
		hSpecular = 1.0;
	}

	if( g_bUseAlbedo )
	{
		half3 xAlbedoTextureColour = SampleTexture( g_xAlbedoSampler, xUVPair.xy ).rgb;

#if defined( NOT_EFFECT )
		if( g_bUseAlbedoDetailMap )
		{
			half hDetailTextureValue = SampleTexture( g_xAlbedoDetailMapSampler,
				xUVPair.xy * g_xDetailMapParameters.x ).a * 2.0f - 1.0f;

			hDetailTextureValue *= g_xDetailMapParameters.y;
			xAlbedoTextureColour = saturate( xAlbedoTextureColour * (1 + hDetailTextureValue) );
		}
#endif

		if( g_bDualAlbedo )
		{
			half4 xAlbedo2 = SampleTexture( g_xAlbedo2Sampler, xUVPair.zw ).rgba;

#if defined( NOT_EFFECT )
			if( g_bDualAlbedoDetailMap )
			{
				half hDetailTextureValue = SampleTexture( g_xAlbedoDetailMap2Sampler,
					xUVPair.zw * g_xDetailMap2Parameters.x ).a * 2.0f - 1.0f;

				hDetailTextureValue *= g_xDetailMap2Parameters.y;
				xAlbedo2.rgb = saturate( xAlbedo2.rgb * (1 + hDetailTextureValue) );
			}
#endif

			float fLerp = 1.0f - ((1.0f - fDualTextureLerp) * xAlbedo2.a);

			xAlbedoTextureColour = lerp( xAlbedo2.rgb, xAlbedoTextureColour, fLerp );

			if( g_bDualSpecularMap )
			{
				const half hSpecular2 = SampleTexture( g_xSpecularMap2Sampler, xUVPair.zw ).a;
				hSpecular = lerp( hSpecular2, hSpecular, fLerp);
			}
		}

		xOutputAlbedo = xAlbedoTextureColour;
	}
	else
	{
		xOutputAlbedo = float3( 1.0, 1.0, 1.0);
	}
	

	// Environment mapping affects the albedo
	if( g_bUseEnvironmentMap )
	{
		// reflect() intrinsic wants direction from camera to point, not the standard V
		half3 xReflectionVector = reflect( -xCameraDirection, xNormal );

		// Asura y-coords are upside-down w.r.t. D3D
		xReflectionVector.y = -xReflectionVector.y;

		half3 xEnvironmentColour = texCUBE( g_xEnvironmentMapSampler, xReflectionVector ).rgb;

		half hReflectivity = g_xMaterialProperties.y;
		if( g_bUseReflectivityMask )
		{
			hReflectivity *= SampleTexture( g_xReflectivityMaskSampler, xUVPair.xy ).a;
		}

		xOutputAlbedo = lerp( xOutputAlbedo, xEnvironmentColour, hReflectivity );
	}

	// Do the x2 Asura thing. TODO: get rid of this
	return xOutputAlbedo * 2.0f;
}

float3 AlbedoLookupWithAlpha( const float4 xUVPair, const float fDualTextureLerp, out float3 xOutputAlbedo, out float fOutputAlbedoAlpha)
{
	if( g_bUseAlbedo )
	{
		float4 xAlbedoTextureColour = SampleTexture( g_xAlbedoSampler, xUVPair.xy ).rgba;
		if( g_bDualAlbedo )
		{
			const float3 xAlbedo2 = SampleTexture( g_xAlbedo2Sampler, xUVPair.zw ).rgb;
			xAlbedoTextureColour.rgb = lerp( xAlbedo2, xAlbedoTextureColour.rgb, fDualTextureLerp );
		}

		xOutputAlbedo = xAlbedoTextureColour.rgb;
		fOutputAlbedoAlpha = xAlbedoTextureColour.a;
		
		// Do the x2 Asura thing. TODO: get rid of this
		return xAlbedoTextureColour * 2.0;
	}
	else
	{
		xOutputAlbedo = float3( 1.0, 1.0, 1.0);
		fOutputAlbedoAlpha = 1.0f;
		return float3( 2.0, 2.0, 2.0 );
	}
}

half4 NormalLookup( const float4 xUVPair, const half hLerpValue, const half3x3 xTangentToWorldMatrix, const half3x3 xTangentToWorldMatrix2 )
{
	// On PC, we'll use DXT5 normal maps with x swizzled into the alpha channel and y in green.
	// Would prefer to use 3Dc/ATI2 compression but it's only available on GeForce 6/7 with a
	// lower-quality hack, where they automatically fudge it into DXT5 anyway

	if( g_bUseNormalMap )
	{
		half4 xNormal;
		xNormal.xy = SampleTexture( g_xNormalMapSampler, xUVPair.xy ).ag;
		xNormal.xy = xNormal.xy * 2.0h - 1.0h;

		if( g_bUseNormalDetailMap )
		{
			const half2 xDetailUV = xUVPair.xy * g_xDetailMapParameters.z;
			half2 xDetailNormal = SampleTexture( g_xNormalDetailMapSampler, xDetailUV ).ag;
			xDetailNormal = xDetailNormal * 2.0h - 1.0h;
			xNormal.xy += xDetailNormal * g_xDetailMapParameters.w;
		}

		// Our normal maps are output upside-down. TODO: do this in a preprocessing step
		xNormal.y = -xNormal.y;
		xNormal.z = sqrt( 1.0h - dot( xNormal.xy, xNormal.xy ) );
		xNormal.xyz = mul( xNormal.xyz, xTangentToWorldMatrix );

		if( g_bDualNormalMap )
		{
			half3 xNormal2 = SampleTexture( g_xNormalMap2Sampler, xUVPair.zw ).agg;
			xNormal2 = xNormal2 * 2.0h - 1.0h;

			if( g_bDualNormalDetailMap )
			{
				const half2 xDetailUV = xUVPair.zw * g_xDetailMap2Parameters.z;
				half2 xDetailNormal = SampleTexture( g_xNormalDetailMap2Sampler, xDetailUV ).ag;
				xDetailNormal = xDetailNormal * 2.0h - 1.0h;
				xNormal2.xy += xDetailNormal * g_xDetailMap2Parameters.w;
			}

			// Our normal maps are output upside-down. TODO: do this in a preprocessing step
			xNormal2.y = -xNormal2.y;
			xNormal2.z = sqrt( 1.0h - dot( xNormal2.xy, xNormal2.xy ) );
			xNormal2.xyz = mul( xNormal2.xyz, xTangentToWorldMatrix2 );

			xNormal.xyz = lerp( xNormal2.xyz, xNormal.xyz, hLerpValue );
		}

		xNormal.xyz = normalize(xNormal.xyz);
		xNormal.w = 1.0h;

		return xNormal;
	}
	else
	{
		return half4( xTangentToWorldMatrix[2], 1.0h );
	}
}

half SpecularLookup( const float4 xUVPair, const float fLerpValue )
{
	if( g_bUseSpecularMap )
	{
		half hSpecularMask = SampleTexture( g_xSpecularMapSampler, xUVPair.xy ).a;
		if( g_bDualSpecularMap )
		{
			const half hSpecularMask2 = SampleTexture( g_xSpecularMap2Sampler, xUVPair.zw ).a;
			hSpecularMask = lerp( hSpecularMask2, hSpecularMask, fLerpValue );
		}

		return hSpecularMask;
	}
	else
	{
		return 1.0;
	}
}

half3 SelfIlluminationLookup( const float4 xUVPair, const float fLerpValue, const float3 xAlbedoTextureColour )
{
	// Slightly different to the others, since this is called from the base pass. Perform the whole
	// calculation in addition to just the texture sampling

	half3 xSelfIllumination = xAlbedoTextureColour * g_xSelfIlluminationParameters.w + g_xSelfIlluminationParameters.xyz;

	if( g_bUseSelfIlluminationMap )
	{
		half hSelfIlluminationMask = SampleTexture( g_xSelfIlluminationMapSampler, xUVPair.xy ).a;
		if( g_bDualSelfIlluminationMap )
		{
			const half hSelfIlluminationMask2 = SampleTexture( g_xSelfIlluminationMap2Sampler, xUVPair.zw ).a;
			hSelfIlluminationMask = lerp( hSelfIlluminationMask2, hSelfIlluminationMask, fLerpValue );
		}

		xSelfIllumination *= hSelfIlluminationMask;
	}

	return xSelfIllumination;
}

half TransparencyLookup( const float4 xUVPair, const float fLerpValue )
{
	if( g_bUseTransparencyMap )
	{
		half hTransparencyMask = SampleTexture( g_xTransparencyMapSampler, xUVPair.xy ).a;
		if( g_bDualTransparencyMap )
		{
			const half hTransparencyMask2 = SampleTexture( g_xTransparencyMap2Sampler, xUVPair.zw ).a;
			hTransparencyMask = lerp( hTransparencyMask2, hTransparencyMask, fLerpValue );
		}

		return hTransparencyMask;
	}
	else
	{
		return 1.0;
	}
}

float4 MaterialResponse_AnimateUVs( const float4 xUVPair )
{
	return xUVPair + g_xUV_Offsets;
}

half3 MaterialResponse_Basic(
						const float fV,
						const half3 xAlbedoTextureColour,
						const half  hSpecularMask,
						const float4 xNormal,					// N in standard notation. Must have 1.0 in w component for SH lookups
						const float3 xCameraDirection,			// V in standard notation
						const float3 xLightDirection,			// L in standard notation
						const float3 xDiffuseLightColour,		// Attenuated light colour
						const float3 xSpecularLightColour, 
						const float fLightStrength				// Need light strength too, to attenuate colourise values
						)
{
	// No need to saturate here, our graphs include back-facing values. Do need to clamp though:
	// normalising isn't coming out accurate enough, sometimes going over 1. Texture address
	// clamping will take care of the lower bound though
	const float fNdotL = min( dot( xNormal.xyz, xLightDirection ), 1.0f );

	const float3 xReflectionVector = 2.0f * xNormal.xyz * fNdotL - xLightDirection.xyz;

	const float fRdotV = min( dot( xReflectionVector.xyz, xCameraDirection.xyz ), 1.0f );

	// Picked colour in rgb, colourise in alpha
	const half4 xDiffuseColour = MaterialResponse_SampleDataTexture( DIFFUSE_TEXEL, fV );
	const half4 xSpecularColour = MaterialResponse_SampleDataTexture( SPECULAR_TEXEL, fV );

	// (Metalicity, diffuse graph scale, specular graph scale)
	const half3 xMisc = MaterialResponse_SampleDataTexture( MISC_TEXEL, fV ).rgb;

	#if COLOURISE_DIFFUSE == 0
		const float3 xFinalDiffuseLightColour = xDiffuseLightColour;
	#elif COLOURISE_DIFFUSE == 1
		const float3 xFinalDiffuseLightColour = xDiffuseColour.rgb * fLightStrength;
	#else
		const float3 xFinalDiffuseLightColour =
			lerp( xDiffuseLightColour.rgb, xDiffuseColour.rgb * fLightStrength, xDiffuseColour.a );
	#endif

	float3 xDiffuse = DiffuseGraph( fNdotL, fV )
					* xFinalDiffuseLightColour
					* xAlbedoTextureColour;

	// Normally specular highlight is the light colour. 'Metalicity' lerps it towards the albedo
	#if METALICITY == 0
		float3 xFinalSpecularLightColour = xSpecularLightColour;
	#elif METALICITY == 1
		float3 xFinalSpecularLightColour = xAlbedoTextureColour * fLightStrength;
	#else
		float3 xFinalSpecularLightColour =
			lerp( xSpecularLightColour, xAlbedoTextureColour * fLightStrength, xMisc.r );
	#endif

	#if COLOURISE_SPECULAR == 1
		xFinalSpecularLightColour = xSpecularColour.rgb * fLightStrength;
	#elif COLOURISE_SPECULAR != 0
		xFinalSpecularLightColour = lerp( xFinalSpecularLightColour,
			xSpecularColour.rgb * fLightStrength, xSpecularColour.a );
	#endif

	float3 xSpecular = SpecularGraph( fRdotV, fV ) * xFinalSpecularLightColour;

	// Mask specular contribution if necessary
	xSpecular *= hSpecularMask;

	half3 xResult = xDiffuse;

	if( !g_bIgnoreSpecular )
	{
		xResult += xSpecular;
	}

	return xResult;
}

half4 MaterialResponse( const float4 xUVPair,
						const float fDualTextureLerp,	// Dual texturing factor. 1 = entirely first texture, 0 entirely second
						const float4 xNormal,			// N in standard notation. Must have 1.0 in w component for SH lookups
						const float3 xCameraDirection,	// V in standard notation
						const float3 xLightDirection,	// L in standard notation
						const float3 xDiffuseLightColour,		// Attenuated light colour
						const float3 xSpecularLightColour, 
						const float fLightStrength,	// Need light strength too, to attenuate colourise values
						const bool bDamageMaps,
						out half3 xOutputAlbedo,
						out half hOutputSpecular,
						out float fOutputMaterial )
{
	if( bDamageMaps )
	{
		AlbedoAndSpecularLookup_DamageMap( xUVPair, fDualTextureLerp, xNormal, xCameraDirection,
			xOutputAlbedo, hOutputSpecular );
	}
	else
	{
		AlbedoLookup( xUVPair, fDualTextureLerp, xNormal, xCameraDirection, xOutputAlbedo );
		hOutputSpecular = SpecularLookup( xUVPair, fDualTextureLerp );
	}

	const half hTransparencyMask = TransparencyLookup( xUVPair, fDualTextureLerp );

	fOutputMaterial = g_xMaterialProperties.x;

	half4 xColour;

	xColour.rgb =

#if defined( ASURA_TOOLS)
		g_bDisableLighting ? xOutputAlbedo : 
#endif // #if defined( ASURA_TOOLS)

		 MaterialResponse_Basic(
					fOutputMaterial,
					xOutputAlbedo,
					hOutputSpecular,
					xNormal,
					xCameraDirection,
					xLightDirection,
					xDiffuseLightColour,
					xSpecularLightColour,
					fLightStrength
							   );

	xColour.a = hTransparencyMask;

	return xColour;
}

half4 MaterialResponseSH( const half4 xUVPair,
						  const half hDualTextureLerp,
						  const half4 xNormal,				// Must have 1.0 in w component for SH lookups
						  const half3 xCameraDirection )
{
	const float3 xDiffuseLightColour = SHLookup( xNormal, g_xRedSH, g_xGreenSH, g_xBlueSH );

	// Dummy variable, we don't care about the specular output
	half hOutputSpecular;

	float fOutputMaterial;
	half3 xOutputAlbedo;
	half4 xColour;

	xColour = MaterialResponse( xUVPair, hDualTextureLerp, xNormal, xCameraDirection,
		g_xSHLightDirection, xDiffuseLightColour, g_xSHLightColour.xyz, g_xSHLightColour.w, false,
		xOutputAlbedo, hOutputSpecular, fOutputMaterial );

	xColour.rgb += SelfIlluminationLookup( xUVPair, hDualTextureLerp, xOutputAlbedo );

	const half hNdotV = dot( xNormal.xyz, xCameraDirection );
	xColour.a *= TransparencyGraph( hNdotV, fOutputMaterial );

	return xColour;
}
