/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float4x4 g_xWorldViewProjectionMatrix;
const float4x4 g_xWorldMatrix;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 m_xPosition : POSITION;

	float3 m_xNormal   : NORMAL;
	float3 m_xTangent  : TANGENT;
	float3 m_xBinormal : BINORMAL;

	float2 m_xUV1       : TEXCOORD0;
	float2 m_xUV2				: TEXCOORD1;
	float m_fDualTextureLerp      : TEXCOORD2;
};

struct OutputVertex
{
	float4 m_xPosition              : POSITION;

	float3 m_xNormal                : NORMAL;
	float3 m_xTangent               : TANGENT;
	float3 m_xBinormal              : BINORMAL;

	float2 m_xUV1                    : TEXCOORD0;
	float2 m_xUV2					: TEXCOORD1;
	float m_fDualTextureLerp		: TEXCOORD2;

	float4 m_xCameraDirDistance     : TEXCOORD3; // Camera direction, and distance from view plane in w
	float3 m_xTangentSpaceCameraDir : TEXCOORD4;
	
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	OutputVertex xOut;

	xOut.m_xPosition = mul( xIn.m_xPosition, g_xWorldViewProjectionMatrix );

	half4 xUVPair;
	xUVPair.xy = xIn.m_xUV1;
	xUVPair.zw = xIn.m_xUV2;

	xUVPair = MaterialResponse_AnimateUVs( xUVPair );

	xOut.m_xUV1 = xUVPair.xy;
	xOut.m_xUV2 = xUVPair.zw;
	xOut.m_fDualTextureLerp = xIn.m_fDualTextureLerp;

	const float4 xWorldSpacePosition = mul( xIn.m_xPosition, g_xWorldMatrix );

	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xWorldSpacePosition.xyz;
	xOut.m_xCameraDirDistance.w = xOut.m_xPosition.w;

	xOut.m_xTangent = mul( xIn.m_xTangent, g_xWorldMatrix );
	xOut.m_xBinormal = mul( xIn.m_xBinormal, g_xWorldMatrix );
	xOut.m_xNormal = mul( xIn.m_xNormal, g_xWorldMatrix );

	const float3x3 xTangentToWorldMatrix = float3x3( xIn.m_xTangent,
													 xIn.m_xBinormal,
													 xIn.m_xNormal );

	// Note implicit transpose in the mul, by swapping operand order
	xOut.m_xTangentSpaceCameraDir = mul( xTangentToWorldMatrix, xOut.m_xCameraDirDistance.xyz );

	return xOut;
}

DeferredOutputPixel PShader_Deferred( const OutputVertex xIn, uniform bool bUseResponse )
{
	const half3 xInterpolatedTangent  = normalize( xIn.m_xTangent );
	const half3 xInterpolatedBinormal = normalize( xIn.m_xBinormal );
	const half3 xInterpolatedNormal   = normalize( xIn.m_xNormal );

	const half3x3 xTangentToWorldMatrix
		= { xInterpolatedTangent, xInterpolatedBinormal, xInterpolatedNormal };

	const half3 xViewDir = normalize( xIn.m_xCameraDirDistance.xyz );
	const half3 xTangentSpaceViewDir = normalize( xIn.m_xTangentSpaceCameraDir );

	half4 xUVPair = half4( xIn.m_xUV1, xIn.m_xUV2 );
	half fDualTextureLerp = xIn.m_fDualTextureLerp;
	ApplyParallaxMapping( xTangentSpaceViewDir, fDualTextureLerp, xUVPair );

	const half4 xWorldSpaceNormal = NormalLookup( xUVPair, fDualTextureLerp, xTangentToWorldMatrix, xTangentToWorldMatrix );


	half3 xAlbedo;
	half hSpecular;
	float fMaterialsTextureY;
	half4 xColour;

	if( bUseResponse )
	{
		const half4 xSHLookupDir = half4( xInterpolatedNormal, 1.0h );
		const float3 xLightColour
			= SHLookup( xSHLookupDir, g_xRedSH, g_xGreenSH, g_xBlueSH ) * 2.0f;

		const float fLightStrength = GetLuminance( xLightColour );

		xColour = MaterialResponse( xUVPair, fDualTextureLerp, xWorldSpaceNormal, xViewDir, xInterpolatedNormal,
			xLightColour, xLightColour, fLightStrength, false, xAlbedo, hSpecular,
			fMaterialsTextureY );

		xColour.rgb += SelfIlluminationLookup( xUVPair, fDualTextureLerp, xAlbedo );
	}
	else
	{
		fMaterialsTextureY = g_xMaterialProperties.x;

		// For this base geometry pass, we'll just use the SH lighting + self illumination

		const float3 xAlbedoTimes2 = AlbedoLookup( xUVPair,
												   fDualTextureLerp,
												   xWorldSpaceNormal,
												   xViewDir,
												   xAlbedo );

		xColour.rgb =
#if defined( ASURA_TOOLS)
			g_bDisableLighting
			? xAlbedo :
#endif // #if defined( ASURA_TOOLS)
			SHLookup( xWorldSpaceNormal, g_xRedSH, g_xGreenSH, g_xBlueSH ) * xAlbedoTimes2
			+ SelfIlluminationLookup( xUVPair, fDualTextureLerp, xAlbedo );

		// Transparent env strips aren't sorted - but we need to use the transparency maps if present
		// to allow alpha decals
		xColour.a = TransparencyLookup( xUVPair, fDualTextureLerp );

		hSpecular = SpecularLookup( xUVPair, fDualTextureLerp );
	}

	// Do 1-bit alpha testing like this, since it seems to be done for deferred objects this way
	clip( xColour.a - 0.05h );

	DeferredOutputPixel xOut = Deferred_FillOutputPixel(
										xColour,
										xWorldSpaceNormal.xyz,
										xIn.m_xCameraDirDistance.w,
										xAlbedo,
										hSpecular,
										fMaterialsTextureY
										);
	return( xOut);
}

float4 PShader_Forward( const OutputVertex xIn ) : COLOR
{
	const float3x3 xTangentToWorldMatrix = float3x3( normalize( xIn.m_xTangent ),
													 normalize( xIn.m_xBinormal ),
													 normalize( xIn.m_xNormal ) );

	const half3 xViewDir = normalize( xIn.m_xCameraDirDistance.xyz );
	const half3 xTangentSpaceViewDir = normalize( xIn.m_xTangentSpaceCameraDir );

	half4 xUVPair = half4( xIn.m_xUV1, xIn.m_xUV2 );
	half fDualTextureLerp = xIn.m_fDualTextureLerp;

	ApplyParallaxMapping( xTangentSpaceViewDir, fDualTextureLerp, xUVPair );

	const float4 xWorldSpaceNormal = NormalLookup( xUVPair, fDualTextureLerp, xTangentToWorldMatrix, xTangentToWorldMatrix );


	const float4 xUnfoggedColour = MaterialResponseSH( xUVPair, fDualTextureLerp, xWorldSpaceNormal, xViewDir );
	return ApplyFog( xUnfoggedColour, xIn.m_xCameraDirDistance.w );
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

// Simple prelighting, no material response curves
//technique Deferred_SH
//{
//	pass
//	{
//		VertexShader = compile vs_3_0 VShader();
//		PixelShader = compile ps_3_0 PShader_Deferred( false );
//	}
//}

// Perform material response using the SH
technique Deferred_MR
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Deferred( true );
	}
}

// Forward renderer, for transparent strips
technique Forward
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Forward();
	}
}