/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Fog.fxh"
#include "Shared.fxh"

const texture g_xTexture;
const sampler g_xTextureSampler = sampler_state
{
	Texture = <g_xTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	sRGBTexture = <g_bUseGammaCorrection>;
};

const bool g_bCrossfadeTextureFrames;


const float4 g_xOneOverScreenDimensions;

const texture g_xHeatHazeCaptureTexture;
const sampler g_xHeatHazeCaptureTextureSampler = sampler_state
{
	Texture = <g_xHeatHazeCaptureTexture>;
	AddressU = CLAMP;
	AddressV = CLAMP;
	MinFilter = LINEAR;
	MagFilter = LINEAR;
	MipFilter = LINEAR;
	sRGBTexture = <g_bUseGammaCorrection>;
};


/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 Position : POSITION;
	float4 Diffuse  : COLOR;
	float2 UV1      : TEXCOORD0;
	float2 UV2      : TEXCOORD1;
};

struct OutputVertex
{
	float4 Position						 : POSITION;
	float4 Diffuse						 : COLOR;
	float4 UVPair						 : TEXCOORD0;
	float2 fCrossFadeBlendWeightAndDepth : TEXCOORD2;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	OutputVertex xOut;

	xOut.Position = mul( float4(xIn.Position.xyz, 1), g_xViewProjectionMatrix );
	xOut.Diffuse = xIn.Diffuse;
	xOut.UVPair.xy = xIn.UV1;
	xOut.UVPair.zw = xIn.UV2;
	xOut.fCrossFadeBlendWeightAndDepth.x = xIn.Position.w;
	xOut.fCrossFadeBlendWeightAndDepth.y = xOut.Position.w;
	return xOut;
}


float4 PShader( const OutputVertex xIn) : COLOR
{
	const float4 xTexture[2] = { tex2D( g_xTextureSampler, xIn.UVPair.xy ), tex2D( g_xTextureSampler, xIn.UVPair.zw ) };

	float4 xColour;

	if( g_bCrossfadeTextureFrames )
	{
		xColour = lerp( xTexture[1], xTexture[0], xIn.fCrossFadeBlendWeightAndDepth.x ) * xIn.Diffuse;
	}
	else
	{
		xColour = xTexture[0] * xIn.Diffuse;
	}


	if (g_bFogEnabled)
	{
		return ApplyFog(xColour, xIn.fCrossFadeBlendWeightAndDepth.y);
	}
	else
	{
		return xColour;
	}
}


float4 PShader_HeatHaze( const OutputVertex xIn, const float2 xScreenPos : VPOS ) : COLOR
{
	const float4 xTexture[2] = { tex2D( g_xTextureSampler, xIn.UVPair.xy ), tex2D( g_xTextureSampler, xIn.UVPair.zw ) };

	float4 xColour;

	if( g_bCrossfadeTextureFrames )
	{
		xColour = lerp( xTexture[1], xTexture[0], xIn.fCrossFadeBlendWeightAndDepth.x );
	}
	else
	{
		xColour = xTexture[0];
	}
	
	xColour.a *= xIn.Diffuse.a;
	
	// Sample the heathaze texture, Scale to get an offset	
	const float2 xNormalizedCoords = (xScreenPos.xy + float2(0.5, 0.5)) * g_xOneOverScreenDimensions.xy;	
	
	float2 xOffset = (xIn.Diffuse.rg * (xColour.rg-0.5f)) / (10.0f+xIn.fCrossFadeBlendWeightAndDepth.y);		
	
			
	// Do the offset lookup from the frame capture
	xColour.rgb = 0.25f *
				  (
				  tex2D(g_xHeatHazeCaptureTextureSampler, xNormalizedCoords + 0.3f*xOffset).rgb	+ 
				  tex2D(g_xHeatHazeCaptureTextureSampler, xNormalizedCoords + 0.6f*xOffset).rgb  +
				  tex2D(g_xHeatHazeCaptureTextureSampler, xNormalizedCoords + 0.8f*xOffset).rgb +
				  tex2D(g_xHeatHazeCaptureTextureSampler, xNormalizedCoords + xOffset).rgb
				  );

	
	return xColour;	
}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique PFX_Standard
{
	pass
	{
		AlphaTestEnable = FALSE;		
		CullMode = NONE;
		ZEnable = TRUE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader();
	}
}



technique PFX_HeatHaze
{
	pass
	{
		AlphaTestEnable = FALSE;		
		CullMode = NONE;
		ZEnable = TRUE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_HeatHaze();
	}
}