/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

// DLH - Enable this define to give a colour coded view of a skinned character's Normal, Tangent and Birnormal error wrt fully boned versions of this data
//#define DEBUG_SHOW_NTB_ERROR

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float4x4 g_xWorldViewProjectionMatrix;
const float4x4 g_xWorldMatrix;
const float4x3 g_axBoneMatrices[80];

const bool g_bUseDamageMaps;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 m_xPosition			: POSITION;
	float2 m_xUV1				: TEXCOORD0;
	float2 m_xUV2				: TEXCOORD1;
	float2 m_fDualTextureLerp   : TEXCOORD2;
	int4   m_aiBoneIndices : BLENDINDICES;
	float4 m_afBoneWeights : BLENDWEIGHT;

	float3 m_xNormal       : NORMAL;
	float3 m_xTangent      : TANGENT;
	float3 m_xBinormal     : BINORMAL;
};

struct OutputVertex
{
	float4 m_xPosition              : POSITION;

	float3 m_xNormal                : NORMAL;
	float3 m_xTangent               : TANGENT;
	float3 m_xBinormal              : BINORMAL;

	float2 m_xUV1					: TEXCOORD0;
	float m_fDualTextureLerp		: TEXCOORD1;
	float2 m_xUV2					: TEXCOORD2;

	float4 m_xCameraDirDistance     : TEXCOORD3; // Direction to camera, and distance from view plane in w
	float3 m_xTangentSpaceCameraDir : TEXCOORD4;

#if defined( DEBUG_SHOW_NTB_ERROR)
	float3 m_xDebugNTBError         : TEXCOORD5;
#endif	
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	// Position and UVs
	float4   xObjectSpacePosition = float4(0,0,0,1); 
	float4x3 axBoneMatrix;
	float    fAlbedoBlend = 0;
	
	OutputVertex xOut;
	
#if defined( DEBUG_SHOW_NTB_ERROR)
	float3 xDebugNFull = 0;
	float3 xDebugTFull = 0;
	float3 xDebugBFull = 0;
#endif

	if ( g_bUseDamageMaps )
	{
		// the code section below and comment from Dave Reed

		// Decrementing loop counter doesn't seem to work, and
		// so when we finish we want the most heavily weighted bone - BoneWeights[0] - in 
		// axBoneMatrix, hence the '3-i' indexing
		for(int i=0; i<4; i++)
		{
			axBoneMatrix = g_axBoneMatrices[ xIn.m_aiBoneIndices[3-i] ];		
			
			// Get the damagemap blend value
			fAlbedoBlend += (axBoneMatrix[2][2]*xIn.m_afBoneWeights[3-i]);

			// Recompute the matrix elements that we've stored per-bone-effects values in
			axBoneMatrix[2] = cross (axBoneMatrix[0], axBoneMatrix[1]);
			const float fBoneWeight = xIn.m_afBoneWeights[3-i];
			xObjectSpacePosition.xyz += mul( xIn.m_xPosition, axBoneMatrix) * fBoneWeight;
#if defined( DEBUG_SHOW_NTB_ERROR)
			xDebugNFull += mul( xIn.m_xNormal,   axBoneMatrix ) * fBoneWeight;
			xDebugTFull += mul( xIn.m_xTangent,  axBoneMatrix ) * fBoneWeight;
			xDebugBFull += mul( xIn.m_xBinormal, axBoneMatrix ) * fBoneWeight;
#endif
		}

		// Tangent, binormal & normal (just transformed to object space to match SH)
		xOut.m_xNormal	= mul( xIn.m_xNormal,	axBoneMatrix );
		xOut.m_xTangent = mul( xIn.m_xTangent,	axBoneMatrix );
		xOut.m_xBinormal= mul( xIn.m_xBinormal,	axBoneMatrix );
		
		// Damage map values are reversed with respect to normal mat resp lookups
		fAlbedoBlend = 1.0f - fAlbedoBlend;
	}
	else
	{
		fAlbedoBlend = xIn.m_fDualTextureLerp;
		
		xObjectSpacePosition.xyz = mul( xIn.m_xPosition, g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] ) * xIn.m_afBoneWeights[0]
								 + mul( xIn.m_xPosition, g_axBoneMatrices[ xIn.m_aiBoneIndices[1] ] ) * xIn.m_afBoneWeights[1]
								 + mul( xIn.m_xPosition, g_axBoneMatrices[ xIn.m_aiBoneIndices[2] ] ) * xIn.m_afBoneWeights[2]
								 + mul( xIn.m_xPosition, g_axBoneMatrices[ xIn.m_aiBoneIndices[3] ] ) * xIn.m_afBoneWeights[3];

#if defined( DEBUG_SHOW_NTB_ERROR)
		xDebugNFull = mul( xIn.m_xNormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] ) * xIn.m_afBoneWeights[0];
		xDebugNFull += mul( xIn.m_xNormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[1] ] ) * xIn.m_afBoneWeights[1];
		xDebugNFull += mul( xIn.m_xNormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[2] ] ) * xIn.m_afBoneWeights[2];
		xDebugNFull += mul( xIn.m_xNormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[3] ] ) * xIn.m_afBoneWeights[3];

		xDebugTFull  = mul( xIn.m_xTangent, g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] ) * xIn.m_afBoneWeights[0];
		xDebugTFull += mul( xIn.m_xTangent, g_axBoneMatrices[ xIn.m_aiBoneIndices[1] ] ) * xIn.m_afBoneWeights[1];
		xDebugTFull += mul( xIn.m_xTangent, g_axBoneMatrices[ xIn.m_aiBoneIndices[2] ] ) * xIn.m_afBoneWeights[2];
		xDebugTFull += mul( xIn.m_xTangent, g_axBoneMatrices[ xIn.m_aiBoneIndices[3] ] ) * xIn.m_afBoneWeights[3];

		xDebugBFull  = mul( xIn.m_xBinormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] ) * xIn.m_afBoneWeights[0];
		xDebugBFull += mul( xIn.m_xBinormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[1] ] ) * xIn.m_afBoneWeights[1];
		xDebugBFull += mul( xIn.m_xBinormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[2] ] ) * xIn.m_afBoneWeights[2];
		xDebugBFull += mul( xIn.m_xBinormal, g_axBoneMatrices[ xIn.m_aiBoneIndices[3] ] ) * xIn.m_afBoneWeights[3];
#endif

		xOut.m_xNormal	= mul( xIn.m_xNormal,	g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] );
		xOut.m_xTangent	= mul( xIn.m_xTangent,	g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] );
		xOut.m_xBinormal= mul( xIn.m_xBinormal,	g_axBoneMatrices[ xIn.m_aiBoneIndices[0] ] );
	}
 
	xOut.m_xPosition = mul( xObjectSpacePosition, g_xWorldViewProjectionMatrix );

	half4 xUVPair;
	xUVPair.xy = xIn.m_xUV1;
	xUVPair.zw = xIn.m_xUV2;

	xUVPair = MaterialResponse_AnimateUVs( xUVPair );

	xOut.m_xUV1 = xUVPair.xy;
	// damage maps use the first set of UV's twice	
	if (g_bUseDamageMaps)
	{
		xOut.m_xUV2 = xUVPair.xy;
	} else {
		xOut.m_xUV2 = xUVPair.zw;
	}

	xOut.m_fDualTextureLerp = fAlbedoBlend;

	// Direction to camera, and world-space distance to vertex from viewer

	const float4 xWorldSpacePosition = mul( xObjectSpacePosition, g_xWorldMatrix );

	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xWorldSpacePosition.xyz;
	xOut.m_xCameraDirDistance.w = xOut.m_xPosition.w;

	const float3x3 xTangentToWorldMatrix = { xOut.m_xTangent, xOut.m_xBinormal, xOut.m_xNormal };

	// Note implicit transpose in the mul, by swapping operand order
	xOut.m_xTangentSpaceCameraDir = mul( xTangentToWorldMatrix, xOut.m_xCameraDirDistance.xyz );

	xOut.m_xNormal	 = mul( xOut.m_xNormal,	  g_xWorldMatrix );
	xOut.m_xTangent	 = mul( xOut.m_xTangent,  g_xWorldMatrix );
	xOut.m_xBinormal = mul( xOut.m_xBinormal, g_xWorldMatrix );

	return xOut;
}

DeferredOutputPixel PShader_Deferred( const OutputVertex xIn, uniform bool bUseResponse )
{
	const half3 xInterpolatedTangent  = normalize( xIn.m_xTangent );
	const half3 xInterpolatedBinormal = normalize( xIn.m_xBinormal );
	const half3 xInterpolatedNormal   = normalize( xIn.m_xNormal );

	const half3x3 xTangentToWorldMatrix
		= { xInterpolatedTangent, xInterpolatedBinormal, xInterpolatedNormal };

	const half3 xViewDir = normalize( xIn.m_xCameraDirDistance.xyz );
	const half3 xTangentSpaceViewDir = normalize( xIn.m_xTangentSpaceCameraDir );

	const half hDualTextureLerp = xIn.m_fDualTextureLerp;

	half4 xUVPair = half4( xIn.m_xUV1.xy, xIn.m_xUV2.xy );
	ApplyParallaxMapping( xTangentSpaceViewDir, hDualTextureLerp, xUVPair );

	const half4 xWorldSpaceNormal = NormalLookup( xUVPair, hDualTextureLerp, xTangentToWorldMatrix, xTangentToWorldMatrix );


	half3 xAlbedo;
	half hSpecular;
	float fMaterialsTextureY;
	half4 xColour;

	if( bUseResponse )
	{
		const half4 xSHLookupDir = half4( xInterpolatedNormal, 1.0h );
		const float3 xLightColour
			= SHLookup( xSHLookupDir, g_xRedSH, g_xGreenSH, g_xBlueSH ) * 2.0f;

		const float fLightStrength = GetLuminance( xLightColour );

		xColour = MaterialResponse( xUVPair, hDualTextureLerp, xWorldSpaceNormal, xViewDir,
			xInterpolatedNormal, xLightColour, xLightColour, fLightStrength, 
			g_bUseDamageMaps, 
			
			xAlbedo, hSpecular, fMaterialsTextureY );

		xColour.rgb += SelfIlluminationLookup( xUVPair, hDualTextureLerp, xAlbedo );
	}
	else
	{
		fMaterialsTextureY = g_xMaterialProperties.x;

		// For this base geometry pass, we'll just use the SH lighting + self illumination

		const float3 xAlbedoTimes2 = AlbedoLookup( xUVPair,
												   hDualTextureLerp,
												   xWorldSpaceNormal,
												   xViewDir,
												   xAlbedo );

		xColour.rgb =
#if defined( ASURA_TOOLS)
			g_bDisableLighting
			? xAlbedo :
#endif // #if defined( ASURA_TOOLS)
			SHLookup( xWorldSpaceNormal, g_xRedSH, g_xGreenSH, g_xBlueSH ) * xAlbedoTimes2
			+ SelfIlluminationLookup( xUVPair, hDualTextureLerp, xAlbedo );

		// Transparent env strips aren't sorted - but we need to use the transparency maps if present
		// to allow alpha decals
		xColour.a = TransparencyLookup( xUVPair, hDualTextureLerp );

		hSpecular = SpecularLookup( xUVPair, hDualTextureLerp );
	}

	// Do 1-bit alpha testing like this, since it seems to be done for deferred objects this way
	clip( xColour.a - 0.05h );

#if defined( DEBUG_SHOW_NTB_ERROR)
	const float fScaleNormalError = 1.0f;
	xAlbedo.rgb = xColour.rgb = fScaleNormalError * xIn.m_xDebugNTBError;
#endif

	DeferredOutputPixel xOut = Deferred_FillOutputPixel(
										xColour,
										xWorldSpaceNormal.xyz,
										xIn.m_xCameraDirDistance.w,
										xAlbedo,
										hSpecular,
										fMaterialsTextureY
										);
	return( xOut);
}

float4 PShader_Forward( const OutputVertex xIn ) : COLOR
{
	const float3x3 xTangentToWorldMatrix = float3x3( normalize( xIn.m_xTangent ),
													 normalize( xIn.m_xBinormal ),
													 normalize( xIn.m_xNormal ) );

	const half3 xViewDir = normalize( xIn.m_xCameraDirDistance.xyz );
	const half3 xTangentSpaceViewDir = normalize( xIn.m_xTangentSpaceCameraDir );

	half4 xUVPair = half4( xIn.m_xUV1.xy, xIn.m_xUV2.xy );
	float fAlbedoBlend = xIn.m_fDualTextureLerp;

	ApplyParallaxMapping( xTangentSpaceViewDir, fAlbedoBlend, xUVPair );

	const float4 xWorldSpaceNormal = NormalLookup( xUVPair, fAlbedoBlend, xTangentToWorldMatrix, xTangentToWorldMatrix );


	return ApplyFog( MaterialResponseSH( xUVPair, fAlbedoBlend, xWorldSpaceNormal, xViewDir ), xIn.m_xCameraDirDistance.w);
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

// Simple prelighting, no material response curves
//technique Deferred_SH
//{
//	pass
//	{
//		VertexShader = compile vs_3_0 VShader();
//		PixelShader = compile ps_3_0 PShader_Deferred( false );
//	}
//}

// Perform material response using the SH
technique Deferred_MR
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Deferred( true );
	}
}

// Forward renderer, for transparent strips
technique Forward
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Forward();
	}
}