/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"
#include "Deferred_Shared.fxh"

const float4x4 g_xWorldViewProj;
const float4x4 g_xWorldMatrix;
const float4x3 g_axBoneMatrices[80];

const float4 g_xRedSH;
const float4 g_xGreenSH;
const float4 g_xBlueSH;

const float4 g_xTexCoordOffsetAndScale;

const float3 g_xObjectSpaceCameraPosition;

const float g_fCharacterAlpha;

const bool g_bNormalMapped;
const bool g_bSelfIlluminating;
const bool g_bSpecular;
const bool g_bTransparent;

const sampler g_xTextureSampler : register(s0);

const sampler g_xNormalMapSampler : register(s1); 

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

static const float fSPECULAR_POWER = 16.0;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 Position     : POSITION;
	float2 UV           : TEXCOORD;
	int4   BoneIndices  : BLENDINDICES;
	float4 BoneWeights  : BLENDWEIGHT;
	float3 Normal		: NORMAL;
	float3 Tangent		: TANGENT;
	float3 Binormal		: BINORMAL;
};

struct OutputVertex
{
	float4		Position	: POSITION;
	float2		UV			: TEXCOORD0;
	float3		PixelPos	: TEXCOORD1;
	float4		NormalDepth	: TEXCOORD2;
	float3x3	Matrix		: TEXCOORD5;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	OutputVertex xOut;

	xOut.Position.xyz = mul( xIn.Position, g_axBoneMatrices[ xIn.BoneIndices[0] ] ) * xIn.BoneWeights[0]
					  + mul( xIn.Position, g_axBoneMatrices[ xIn.BoneIndices[1] ] ) * xIn.BoneWeights[1]
					  + mul( xIn.Position, g_axBoneMatrices[ xIn.BoneIndices[2] ] ) * xIn.BoneWeights[2]
					  + mul( xIn.Position, g_axBoneMatrices[ xIn.BoneIndices[3] ] ) * xIn.BoneWeights[3];

	xOut.Position.w = 1.0;

	xOut.PixelPos = xOut.Position;
	xOut.Position = mul( xOut.Position, g_xWorldViewProj );
	
	const float3 Normal = mul( xIn.Normal, g_axBoneMatrices[ xIn.BoneIndices[0] ] );
	const float3 Tangent = mul( xIn.Tangent, g_axBoneMatrices[ xIn.BoneIndices[0] ] );
	const float3 Binormal = mul( xIn.Binormal, g_axBoneMatrices[ xIn.BoneIndices[0] ] );

	xOut.Matrix = float3x3( Tangent, Binormal, Normal );

	xOut.UV = xIn.UV * g_xTexCoordOffsetAndScale.zw + g_xTexCoordOffsetAndScale.xy;

	xOut.NormalDepth.xyz = mul(Normal, g_xWorldMatrix);
	xOut.NormalDepth.w = xOut.Position.w;

	return xOut;
}

float3 CalculateDiffuseSHLightColour( const float3 xDirection )
{
	const float4 xDirection4 = float4( xDirection, 1.0 );

	float3 xColour;

	xColour.r = dot( xDirection4, g_xRedSH );
	xColour.g = dot( xDirection4, g_xGreenSH );
	xColour.b = dot( xDirection4, g_xBlueSH );

	return max( xColour, 0.0 );
}

float3 CalculateSpecularSHLightColour( const float3 xNormal, const float3 xEyeDirection, const float3 xBrightestLightDirection )
{
	const float3 xDiffuseColour = CalculateDiffuseSHLightColour( xBrightestLightDirection );

	const float3 xHalfwayVector = normalize( xBrightestLightDirection + xEyeDirection );
	const float fDotProduct = saturate( dot( xHalfwayVector, xNormal ) );
	const float fPower = pow( fDotProduct, fSPECULAR_POWER );

	return xDiffuseColour * fPower;
}

float3 CalculateBrightestSHLightDirection()
{
	float3 xBrightestLightDirection;

	xBrightestLightDirection.x = g_xRedSH.x + g_xGreenSH.x + g_xBlueSH.x;
	xBrightestLightDirection.y = g_xRedSH.y + g_xGreenSH.y + g_xBlueSH.y;
	xBrightestLightDirection.z = g_xRedSH.z + g_xGreenSH.z + g_xBlueSH.z;

	return xBrightestLightDirection;
}

float3 CalculateNormal( const sampler xSampler, const float2 xUV )
{
	float3 xNormal = tex2D( xSampler, xUV ).xyz;

	xNormal *= 2.0;
	xNormal -= 1.0;
	xNormal.y = -xNormal.y;

	return normalize( xNormal );
}

DeferredOutputPixel PShader( const OutputVertex xIn )
{
	const float4 xTextureColour = tex2D( g_xTextureSampler, xIn.UV );

	float3 xNormal;
	if( g_bNormalMapped )
	{
		xNormal = CalculateNormal( g_xNormalMapSampler, xIn.UV );
		xNormal = mul( xNormal, xIn.Matrix );
	}
	else
	{
		xNormal = xIn.Matrix[2];
	}

	float3 xOutputColour = CalculateDiffuseSHLightColour( xNormal );
	xOutputColour *= xTextureColour * 2.0;

	if( g_bSelfIlluminating )
	{
		xOutputColour += xTextureColour.rgb * xTextureColour.a * 2.0f;
	}

	if( g_bSpecular )
	{
		const float3 xEyeDirection = normalize( g_xObjectSpaceCameraPosition.xyz - xIn.PixelPos );
		const float3 xBrightestLightDirection = normalize( CalculateBrightestSHLightDirection() );

		xOutputColour += 2.0f * xTextureColour.a * CalculateSpecularSHLightColour( xNormal, xEyeDirection, xBrightestLightDirection );
	}

	float fAlpha = g_fCharacterAlpha;
	if( g_bTransparent )
	{
		fAlpha *= xTextureColour.a;
	}

	half4 xOutputRGBA;

	xOutputRGBA.rgb = xOutputColour;
	xOutputRGBA.a = fAlpha;
	
	half3 xOutputAlbedo = xTextureColour;
	half hOutputSpecular = xTextureColour.a;

	// Return the data...
	DeferredOutputPixel xOut = Deferred_FillOutputPixel( xOutputRGBA, xIn.NormalDepth.xyz, xIn.NormalDepth.w, xOutputAlbedo, hOutputSpecular, 0);
	
	return( xOut);
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique t0
{
	pass p0
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader();
	}
}
