/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"
#include "Deferred_Shared.fxh"

const float4x4 g_xWorldViewProj;
const float4x4 g_xWorldMatrix;
const float4x3 g_axBoneMatrices[80];

const float4 g_xRedSH;
const float4 g_xGreenSH;
const float4 g_xBlueSH;

const float4 g_xTexCoordOffsetAndScale;

const float3 g_xObjectSpaceCameraPosition;

const float g_fCharacterAlpha;

const bool g_bNormalMapped;
const bool g_bSelfIlluminating;
const bool g_bSpecular;
const bool g_bTransparent;

const sampler g_xTextureSampler    : register(s0) = sampler_state
{
	AddressU	= WRAP;
	AddressV	= WRAP;
	AddressW	= CLAMP;
	MIPFILTER	= LINEAR;
	MINFILTER	= LINEAR;
	MAGFILTER	= LINEAR;
};

const sampler g_xNormalMapSampler  : register(s1)  = sampler_state
{
	AddressU	= WRAP;
	AddressV	= WRAP;
	AddressW	= CLAMP;
	MIPFILTER	= LINEAR;
	MINFILTER	= LINEAR;
	MAGFILTER	= LINEAR;
}; 


// 2nd texture - for damage map
const sampler g_xTextureSampler1   : register(s2) = sampler_state
{
	AddressU	= WRAP;
	AddressV	= WRAP;
	AddressW	= CLAMP;
	MIPFILTER	= LINEAR;
	MINFILTER	= LINEAR;
	MAGFILTER	= LINEAR;
}; 

const sampler g_xNormalMapSampler1 : register(s3) = sampler_state
{
	AddressU	= WRAP;
	AddressV	= WRAP;
	AddressW	= CLAMP;
	MIPFILTER	= LINEAR;
	MINFILTER	= LINEAR;
	MAGFILTER	= LINEAR;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

static const float fSPECULAR_POWER = 16.0;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 Position     : POSITION;
	float2 UV           : TEXCOORD;
	int4   BoneIndices  : BLENDINDICES;
	float4 BoneWeights  : BLENDWEIGHT;
	float3 Normal		: NORMAL;
	float3 Tangent		: TANGENT;
	float3 Binormal		: BINORMAL;
};

struct OutputVertex
{
	float4		Position	: POSITION;
	float2		UV			: TEXCOORD0;
	float3		PixelPos	: TEXCOORD1;
	float4		NormalDepth	: TEXCOORD2;
	float3x3	Matrix		: TEXCOORD5;
	float		fDamage     : TEXCOORD3;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	OutputVertex xOut;

	xOut.Position.w = 1.0;
	xOut.Position = float4(0,0,0,1); 
	float4x3 axBoneMatrix;
	float fDamage = 0.0;
	
	// Decrementing loop counter doesn't seem to work, and
	// so when we finish we want the most heavily weighted bone - BoneWeights[0] - in 
	// axBoneMatrix, hence the '3-i' indexing
	for(int i=0; i<4; i++)
	{
		axBoneMatrix=g_axBoneMatrices[ xIn.BoneIndices[3-i] ];		

		// Get the weighted damage value, for the damage map (we may support up to 3 in the future)
		fDamage += axBoneMatrix[2][2]*xIn.BoneWeights[3-i];		
		
		// Recompute the matrix elements that we've stored per-bone-effects values in		
		axBoneMatrix[2] = cross (axBoneMatrix[0], axBoneMatrix[1]);		
		xOut.Position.xyz += mul( xIn.Position, axBoneMatrix) * xIn.BoneWeights[3-i];		
	}	

	xOut.PixelPos = xOut.Position;
	xOut.Position = mul( xOut.Position, g_xWorldViewProj );
	
	const float3 Normal = mul( xIn.Normal,     axBoneMatrix);
	const float3 Tangent = mul( xIn.Tangent,   axBoneMatrix);
	const float3 Binormal = mul( xIn.Binormal, axBoneMatrix);

	xOut.Matrix = float3x3( Tangent, Binormal, Normal );

	xOut.UV = xIn.UV * g_xTexCoordOffsetAndScale.zw + g_xTexCoordOffsetAndScale.xy;

	xOut.NormalDepth.xyz = mul(Normal, g_xWorldMatrix);
	xOut.NormalDepth.w = xOut.Position.w;
	
	xOut.fDamage = fDamage;

	return xOut;
}

float3 CalculateDiffuseSHLightColour( const float3 xDirection )
{
	const float4 xDirection4 = float4( xDirection, 1.0 );

	float3 xColour;

	xColour.r = dot( xDirection4, g_xRedSH );
	xColour.g = dot( xDirection4, g_xGreenSH );
	xColour.b = dot( xDirection4, g_xBlueSH );

	return max( xColour, 0.0 );
}

float3 CalculateSpecularSHLightColour( const float3 xNormal, const float3 xEyeDirection, const float3 xBrightestLightDirection )
{
	const float3 xDiffuseColour = CalculateDiffuseSHLightColour( xBrightestLightDirection );

	const float3 xHalfwayVector = normalize( xBrightestLightDirection + xEyeDirection );
	const float fDotProduct = saturate( dot( xHalfwayVector, xNormal ) );
	const float fPower = pow( fDotProduct, fSPECULAR_POWER );

	return xDiffuseColour * fPower;
}

float3 CalculateBrightestSHLightDirection()
{
	float3 xBrightestLightDirection;

	xBrightestLightDirection.x = g_xRedSH.x + g_xGreenSH.x + g_xBlueSH.x;
	xBrightestLightDirection.y = g_xRedSH.y + g_xGreenSH.y + g_xBlueSH.y;
	xBrightestLightDirection.z = g_xRedSH.z + g_xGreenSH.z + g_xBlueSH.z;

	return xBrightestLightDirection;
}

float3 CalculateNormal( const sampler xSampler, const float2 xUV )
{
	float3 xNormal = tex2D( xSampler, xUV ).xyz;

	xNormal *= 2.0;
	xNormal -= 1.0;
	xNormal.y = -xNormal.y;

	return normalize( xNormal );
}

DeferredOutputPixel PShader( const OutputVertex xIn )
{
	const float4 xTextureColour = lerp (tex2D( g_xTextureSampler, xIn.UV ),
										tex2D( g_xTextureSampler1, xIn.UV ), xIn.fDamage);
	

	float3 xNormal;
	if( g_bNormalMapped )
	{
		xNormal = lerp(CalculateNormal( g_xNormalMapSampler,  xIn.UV ),
		               CalculateNormal( g_xNormalMapSampler1, xIn.UV ) , xIn.fDamage);
		               
		xNormal = normalize (mul( xNormal, xIn.Matrix ));
	}
	else
	{
		xNormal = xIn.Matrix[2];
	}

	float3 xOutputColour = CalculateDiffuseSHLightColour( xNormal );
	xOutputColour *= xTextureColour * 2.0;

	if( g_bSelfIlluminating )
	{
		xOutputColour += xTextureColour.rgb * xTextureColour.a * 2.0f;
	}

	if( g_bSpecular )
	{
		const float3 xEyeDirection = normalize( g_xObjectSpaceCameraPosition.xyz - xIn.PixelPos );
		const float3 xBrightestLightDirection = normalize( CalculateBrightestSHLightDirection() );

		xOutputColour += 2.0f * xTextureColour.a * CalculateSpecularSHLightColour( xNormal, xEyeDirection, xBrightestLightDirection );
		
		//xOutputColour = xNormal.rgb; 
	}

	float fAlpha = g_fCharacterAlpha;
	if( g_bTransparent )
	{
		fAlpha *= xTextureColour.a;
	}

	half4 xOutputRGBA;

	xOutputRGBA.rgb = xOutputColour;
	xOutputRGBA.a = fAlpha;
	
	half3 xOutputAlbedo = xTextureColour;
	half hOutputSpecular = xTextureColour.a;

	// Return the data...
	DeferredOutputPixel xOut = Deferred_FillOutputPixel( xOutputRGBA, xIn.NormalDepth.xyz, xIn.NormalDepth.w, xOutputAlbedo, hOutputSpecular, 0);
	
	return( xOut);
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique t0
{
	pass p0
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader();
	}
}