//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Shadow_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"


#include "Shadows/Asura_Shadow_System.h"

#if defined(ASURA_TARGET_SM3)


// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#include "Entity/DynamicLight/Asura_ClientEntity_DynamicLight.h"
#include "SpecialFX/Asura_Sphere.h"

// Asura PC
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Caps.h"
#include "Rendering/Asura_Direct3D_Effects_System.h"
#include "Rendering/Asura_Direct3D_EffectParameterBlock.h"
#include "Rendering/Asura_Direct3D_EnvironmentRenderer.h"
#include "Rendering/Asura_Direct3D_RenderTargetManager.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Direct3D_UserOptions.h"
#include "Rendering/Asura_DeferredShading.h"

#include "Interface/Asura_CommandConsole.h"

#include "Asura_PC_ShadowBuffer.h"
#include "Asura_PC_ShadowStorage.h"


/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const Direct3D_VertexElement axVERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
		D3DDECL_END()
};

const u_int uVERTEX_SIZE = 12;

const D3DVECTOR axFRUSTUM_VERTICES[] =
{
	{ -1.0f,  1.0f, 0.0f },
	{  1.0f,  1.0f, 0.0f },
	{  1.0f, -1.0f, 0.0f },
	{ -1.0f, -1.0f, 0.0f },
	{ -1.0f,  1.0f, 1.0f },
	{  1.0f,  1.0f, 1.0f },
	{  1.0f, -1.0f, 1.0f },
	{ -1.0f, -1.0f, 1.0f },
};

const u_short ausFRUSTUM_INDICES[] =
{
	0, 1, 2,
	2, 3, 0,		// Near
	4, 5, 1,
	1, 0, 4,		// Bottom
	5, 6, 2,
	2, 1, 5,		// Right
	6, 7, 3,
	3, 2, 6,		// Top
	7, 4, 0,
	0, 3, 7,		// Left
	5, 4, 7,
	7, 6, 5,		// Far
};


static bool s_bEnableNvidiaDepthMapExtensions = false;

/******************************************************************************/
/* Classes                                                                    */
/******************************************************************************/


enum
{
	SHADOW_EFFECT_PASS_SHADOWMAP_SINGLETAP,
	SHADOW_EFFECT_PASS_SHADOWMAP_MULTITAP,
	SHADOW_EFFECT_PASS_DEPTHMAP_SINGLETAP,
	SHADOW_EFFECT_PASS_DEPTHMAP_MULTITAP,
	SHADOW_EFFECT_PASS_ENVCUBEMAP_SINGLETAP,
	SHADOW_EFFECT_PASS_ENVCUBEMAP_MULTITAP,
};

DVSVAR float g_fXShadowOffset;
DVSVAR float g_fYShadowOffset;
DVSVAR bool g_bDisableFiltering = false;

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Shadow
//
// Represents a single Xbox 360 shadow
//
////////////////////////////////////////////////////////////////////////////////

class Asura_Direct3D_Shadow : public Asura_Shadow
{
public:

	// Constructor / destructor
	Asura_Direct3D_Shadow();
	virtual ~Asura_Direct3D_Shadow();

	// Virtual overrides

	virtual	void	PrepareForRendering	(void);
	virtual	void	FinishedRendering	(void);

	Asura_PC_ShadowBuffer* GetShadowBuffer	(void) const	{ return m_pxShadowBuffer; }

	virtual void SetInstance (const Asura_Shadow_Instance* pxInstance)
	{ 
		m_pxInstance = pxInstance; 
		m_pxShadowBuffer = pxInstance->IsCubeMapped() ? NULL : Asura_PC_ShadowStorage::Get().GetNextShadowBuffer();
	}

protected:

	Asura_PC_ShadowBuffer* m_pxShadowBuffer;
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

// Effect and parameters

static Asura_Direct3D_Effect* g_pxApplyShadowsEffect;

struct Asura_Direct3D_Shadow_System_ShaderParameters
{
	Asura_Direct3D_EffectParameterBlock* m_pxParamBlock;

	D3DXMATRIX			m_xLightClipSpaceToCameraClipSpace;
	D3DXMATRIX			m_xCameraViewSpaceToLightClipSpace;
	D3DXMATRIX			m_xCameraViewSpaceToLightViewSpace;
	D3DXMATRIX			m_xCameraViewSpaceToWorldSpace;
	D3DXMATRIX			m_xUVXForm;
	D3DXVECTOR4			m_xLightPos;
	float				m_fTwoOverScreenWidth;
	float				m_fTwoOverScreenHeight;
	float				m_fCameraTanHalfVertFOV;
	float				m_fCameraTanHalfHorizFOV;
	bool				m_bFilterShadows;
	bool				m_b;
	bool				m_bUseNoise;
	float				m_fShadowFilterTextureScale;
	Direct3D_BaseTexture*	m_pxShadowBuffer;
	Direct3D_BaseTexture*	m_pxShadowBufferCube;
	Direct3D_BaseTexture*	m_pxNormalDepthImage;
	D3DXVECTOR4			m_xShadowDirection;
	float				m_fShadowStrength;
	Direct3D_Texture*	m_pxKernelRotationTexture;
	float               m_fShadowFilterSize;
	bool				m_bShadowMask;
	D3DXVECTOR4			m_xSelfShadowBias;
};

static D3DXVECTOR4     s_xSelfShadowBiasValues[5] = 
{
	// Slope, Bias, Dist
	// non environment
	D3DXVECTOR4(0.06f, .001f, 0.001f, 0),
	// environment
	D3DXVECTOR4(0.0f, .002f, 0.0f, 0),
	// non environment
	D3DXVECTOR4(0.001f, .000f, 0, 0),
	// environment
	D3DXVECTOR4(0.001f, .001f, 0.02f, 0),

	// cubic environment
	D3DXVECTOR4(0.001f, .001f, 0.01f, 0),
};

static Asura_Direct3D_Shadow_System_ShaderParameters g_xApplyShadowParameters;

// Shadow frustum geometry

static Direct3D_VertexBuffer* g_pxShadowFrustumVB;
static Direct3D_IndexBuffer* g_pxShadowFrustumIB;

static Direct3D_VertexDeclaration* g_pxShadowFrustumVD = NULL;

static Direct3D_Texture* g_pxFilterKernelRotationTexture = NULL;




/******************************************************************************/
/* Asura_Shadow_System Static Class Functions                                 */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Initialise()
//
// Initialises the shadow system on PC
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////
#endif
void Asura_Shadow_System::Platform_Initialise()
{
#if defined(ASURA_TARGET_SM3)
	if( NULL == Asura_Direct3D::D3DDevice)
	{
		return;
	}



	// Check for nvidia's "hardware shadow maps"
	s_bEnableNvidiaDepthMapExtensions = D3D_OK == Asura_Direct3D::D3DObject->CheckDeviceFormat(
		D3DADAPTER_DEFAULT, //default adapter
		D3DDEVTYPE_HAL, //HAL device
		D3DFMT_X8R8G8B8, //display mode
		D3DUSAGE_DEPTHSTENCIL, //shadow map is a depth/s surface
		D3DRTYPE_TEXTURE, //shadow map is a texture
		D3DFMT_D24S8 //format of shadow map
		);

	// NVidia shadow maps only work correctly when NOT using the debug runtimes. It seems that for most IHV extensions 
	// this is the case. So we need to check for their existence. 
	{

		HRESULT hr = Asura_Direct3D::D3DDevice->CreateQuery(D3DQUERYTYPE_VERTEXSTATS, NULL);

		if(SUCCEEDED(hr))
		{
			// Debug Runtime
			s_bEnableNvidiaDepthMapExtensions = false;
		}
		else
		{
			// Release Runtime
		}
	}



	Asura_PC_ShadowBuffer::SetNvidiaDepthMap(s_bEnableNvidiaDepthMapExtensions);

	Asura_Shadow**	ppxShadow	= s_apxShadows;

	for(u_int u = 0; u < s_uMaxShadows; ++u, ++ppxShadow)
	{
		*ppxShadow	= new Asura_Direct3D_Shadow();
	}


	// Create the texture used for random rotation of the shadow filter kernel. Contains sin/cos of random angles	
	HRESULT xResult = Asura_Direct3D::D3DDevice->CreateTexture( 64, 64, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &g_pxFilterKernelRotationTexture, NULL );
	ASURA_D3D_OK( xResult);
	Asura_Assert( NULL != g_pxFilterKernelRotationTexture, "Failed to create shadow filter kernel rotation texture." );

	D3DLOCKED_RECT xLockedRect;
	xResult = g_pxFilterKernelRotationTexture->LockRect( 0, &xLockedRect, NULL, 0 );
	if( FAILED( xResult ) )
	{
		Asura_Assert( false, "Failed to lock shadow filter kernel rotation texture." );	
	}
	else
	{
		u_char *pxData = static_cast<u_char*>(xLockedRect.pBits);

		for(int i=0;i<64;i++)
		{
			for (int j=0;j<64;j++)
			{
				float fAngle = Asura_Random::GetFloat(0,Asura_Maths::Pi*2.0f);
				float fSin,fCos;
				Asura_Maths::SinCos(fAngle, fSin, fCos);
				fSin = (fSin * 127.0f) + 127.0f;
				fCos = (fCos * 127.0f) + 127.0f;
				pxData[(j<<2)+0] = (u_char)fSin;
				pxData[(j<<2)+1] = (u_char)fCos;
				pxData[(j<<2)+2] = 0;
				pxData[(j<<2)+3] = 0;
			}
			pxData += xLockedRect.Pitch;
		}
		
		g_pxFilterKernelRotationTexture->UnlockRect(0);
	}
	D3DXFilterTexture(g_pxFilterKernelRotationTexture, NULL, 0, D3DX_FILTER_TRIANGLE);
	

	// Initialise effect and parameter handles
	g_pxApplyShadowsEffect = Asura_Direct3D_Effects_System::GetEffect("ApplyShadows");

	g_xApplyShadowParameters.m_pxParamBlock = new Asura_Direct3D_EffectParameterBlock();
	
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		("g_xLightClipSpaceToCameraClipSpace", &g_xApplyShadowParameters.m_xLightClipSpaceToCameraClipSpace);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		("g_xCameraViewSpaceToLightClipSpace", &g_xApplyShadowParameters.m_xCameraViewSpaceToLightClipSpace);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		("g_xCameraViewSpaceToLightViewSpace", &g_xApplyShadowParameters.m_xCameraViewSpaceToLightViewSpace);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		("g_xCameraViewSpaceToWorldSpace", &g_xApplyShadowParameters.m_xCameraViewSpaceToWorldSpace);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		("g_xLightPos",						&g_xApplyShadowParameters.m_xLightPos);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fCameraTanHalfHorizFOV",		&g_xApplyShadowParameters.m_fCameraTanHalfHorizFOV);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fCameraTanHalfVertFOV",			&g_xApplyShadowParameters.m_fCameraTanHalfVertFOV);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fTwoOverScreenWidth",			&g_xApplyShadowParameters.m_fTwoOverScreenWidth);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fTwoOverScreenHeight",			&g_xApplyShadowParameters.m_fTwoOverScreenHeight);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<bool>				("g_bUseNoise",						&g_xApplyShadowParameters.m_bUseNoise);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fShadowFilterTextureScale",		&g_xApplyShadowParameters.m_fShadowFilterTextureScale);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_BaseTexture*>	("g_xNormalDepthBuffer",			&g_xApplyShadowParameters.m_pxNormalDepthImage);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_BaseTexture*>	("g_xShadowBuffer",					&g_xApplyShadowParameters.m_pxShadowBuffer);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_BaseTexture*>	("g_xShadowBufferCube",					&g_xApplyShadowParameters.m_pxShadowBufferCube);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		("g_xShadowDirection",				&g_xApplyShadowParameters.m_xShadowDirection);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float>				("g_fShadowStrength",				&g_xApplyShadowParameters.m_fShadowStrength);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<Direct3D_Texture*>	("g_xKernelRotationTexture",		&g_xApplyShadowParameters.m_pxKernelRotationTexture);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<float> 			("g_fShadowFilterSize",				&g_xApplyShadowParameters.m_fShadowFilterSize);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<bool>				("g_bCreatingShadowMask",			&g_xApplyShadowParameters.m_bShadowMask);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXVECTOR4>		("g_xSelfShadowBias",				&g_xApplyShadowParameters.m_xSelfShadowBias);
	g_xApplyShadowParameters.m_pxParamBlock->AddDynamicParameter<D3DXMATRIX>		("g_xUVXForm", &g_xApplyShadowParameters.m_xUVXForm);

	Asura_Mem::Set(g_xApplyShadowParameters.m_xUVXForm, 0, sizeof(g_xApplyShadowParameters.m_xUVXForm));
	g_xApplyShadowParameters.m_xUVXForm._11 = 0.5f;
	g_xApplyShadowParameters.m_xUVXForm._22 = 0.5f;
	g_xApplyShadowParameters.m_xUVXForm._33 = 1.0f;
	g_xApplyShadowParameters.m_xUVXForm._44 = 1.0f;


	g_xApplyShadowParameters.m_pxParamBlock->Bind(g_pxApplyShadowsEffect);

	// Set up shadow frustum geometry

	Asura_Direct3D::D3DDevice->CreateVertexBuffer(uVERTEX_SIZE * 8, D3DUSAGE_WRITEONLY, 0, D3DPOOL_MANAGED, &g_pxShadowFrustumVB, NULL );

	void* pVertices; g_pxShadowFrustumVB->Lock(0, 0, &pVertices, 0);
	Asura_Assert( pVertices, "Failed to lock shadow frustum vertex buffer." );
	memcpy( pVertices, axFRUSTUM_VERTICES, sizeof( axFRUSTUM_VERTICES ) );
	g_pxShadowFrustumVB->Unlock();


	Asura_Direct3D::D3DDevice->CreateIndexBuffer(sizeof(ausFRUSTUM_INDICES), D3DUSAGE_WRITEONLY, D3DFMT_INDEX16, D3DPOOL_MANAGED, &g_pxShadowFrustumIB, NULL);

	void* pIndices; g_pxShadowFrustumIB->Lock(0, 0, &pIndices, 0);
	Asura_Assert( pIndices, "Failed to lock shadow frustum index buffer." );
	memcpy( pIndices, ausFRUSTUM_INDICES, sizeof( ausFRUSTUM_INDICES ) );
	g_pxShadowFrustumIB->Unlock();


	Asura_Direct3D::D3DDevice->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxShadowFrustumVD );
	Asura_Assert( g_pxShadowFrustumVD, "Failed to create shadow frustum vertex declaration." );



#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
	if(!Asura_CommandConsole::VarExists(L"Render.Shadows.Bias.Skin_Slope"))
	{
		if (Asura_PC_ShadowBuffer::IsNvidiaDepthMap())
		{
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Slope", &(s_xSelfShadowBiasValues[0].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Bias", &(s_xSelfShadowBiasValues[0].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Dist", &(s_xSelfShadowBiasValues[0].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Slope", &(s_xSelfShadowBiasValues[1].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Bias", &(s_xSelfShadowBiasValues[1].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Dist", &(s_xSelfShadowBiasValues[1].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
//			Asura_CommandConsole::AddVar( L"Render.Shadows.EnableNvidiaDepthMaps", &s_bEnableNvidiaDepthMapExtensions,, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		} else {
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Slope", &(s_xSelfShadowBiasValues[2].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Bias", &(s_xSelfShadowBiasValues[2].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Dist", &(s_xSelfShadowBiasValues[2].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Slope", &(s_xSelfShadowBiasValues[3].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Bias", &(s_xSelfShadowBiasValues[3].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Dist", &(s_xSelfShadowBiasValues[3].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		}
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Slope", &(s_xSelfShadowBiasValues[4].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Bias", &(s_xSelfShadowBiasValues[4].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Dist", &(s_xSelfShadowBiasValues[4].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_CommandConsole::AddVar( L"Render.Shadows.Offset.X", &(g_fXShadowOffset), -1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Offset.Y", &(g_fYShadowOffset), -1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.DisableFiltering", &(g_bDisableFiltering), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif

		
	}
#endif

	Asura_Direct3D::AddPostResetCallback( Asura_PC_ShadowStorage::PostResetCB, NULL );
	Asura_Direct3D::AddPreResetCallback( Asura_PC_ShadowStorage::PreResetCB, NULL );

	Asura_PC_ShadowStorage::Get().Initialise();
#endif


}


#if defined(ASURA_TOOLS) && defined(ASURA_USE_SCREENSHOT_SYSTEM)
void Asura_Shadow_System::Platform_SetHiResShadowMode(bool bValue)
{
#if defined(ASURA_TARGET_SM3)
	Asura_Shadow**	ppxShadow	= s_apxShadows;
	for(u_int u = 0; u < s_uMaxShadows; ++u, ++ppxShadow)
	{
		*ppxShadow	= new Asura_Direct3D_Shadow();
	}

	Asura_PC_ShadowStorage::Get().SetHRS(bValue);
#endif
}
#endif

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Reset()
//
// Resets the shadow system on PC. Called between levels.
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_Reset()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Shutdown()
//
// Shuts down the shadow system on PC
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_Shutdown()
{
#if defined(ASURA_TARGET_SM3)
	Asura_Direct3D::RemovePreResetCallback( Asura_PC_ShadowStorage::PreResetCB, NULL );
	Asura_Direct3D::RemovePostResetCallback( Asura_PC_ShadowStorage::PostResetCB, NULL );

	Asura_PC_ShadowStorage::Get().Shutdown();

	g_pxApplyShadowsEffect = NULL;
	delete g_xApplyShadowParameters.m_pxParamBlock;

	if(g_pxShadowFrustumVD) g_pxShadowFrustumVD->Release(); g_pxShadowFrustumVD = NULL;
	if(g_pxShadowFrustumIB) g_pxShadowFrustumIB->Release(); g_pxShadowFrustumIB = NULL;
	if(g_pxShadowFrustumVB) g_pxShadowFrustumVB->Release(); g_pxShadowFrustumVB = NULL;

	SAFE_RELEASE( g_pxFilterKernelRotationTexture);
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_BeginCreateShadows()
//
// Begins the process of creating this frame's shadow maps on PC
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_BeginCreateShadows()
{
#if defined(ASURA_TARGET_SM3)
	D3DPERFCALL(BeginEvent( D3DCOLOR_XRGB( 0xFF, 0x80, 0x40 ), L"Create Shadows" ));
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 

	Asura_PC_ShadowStorage::Get().Reset();

#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_EndCreateShadows()
//
// Finishes creating this frame's shadow maps on PC
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_EndCreateShadows()
{
#if defined(ASURA_TARGET_SM3)
	Asura_Direct3D::GetDeferredRenderingTarget()->Bind();
	D3DPERFCALL(EndEvent());
#endif
}

#if defined(ASURA_TARGET_SM3)
void Asura_Shadow_System::Platform_RenderEnvironmentShadowMap( Asura_Shadow &xShadow)
{
	SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), L"Render Env Shadow"));
	
	if (((Asura_Direct3D_Shadow&)xShadow).GetShadowBuffer())
	{
		Asura_Vector_3 xCamPos = *Asura_Camera::GetPosition();
		Asura_Matrix_3x3 xCamOrient = *Asura_Camera::GetOrientation();
		Asura_Angle fCamFOV = Asura_Camera::GetHorizontalFieldOfView();
		float fNear = Asura_Camera::GetNearClip();
		float fFar = Asura_Camera::GetFarClip();
		float fCamAspect = Asura_Camera::GetXOverYAspectRatio() / Asura_Camera::GetPhysicalScreenAspectRatio();

		Asura_Matrix_3x3 xOrient;
		xShadow.GetLightOrientation(xOrient);

		Asura_Camera::SetPosition( xShadow.GetLightPosition() );
		Asura_Camera::SetHorizontalFieldOfView( Asura_Maths::DegToRad(xShadow.GetFrustum().GetHorizontalFOVInDegrees()), 1.0f / Asura_Camera::GetPhysicalScreenAspectRatio() );
		Asura_Camera::SetOrientation(xOrient);
		Asura_Camera::SetNearClip(xShadow.GetFrustum().GetNearPlane());
		Asura_Camera::SetFarClip(xShadow.GetFrustum().GetFarPlane());
		Asura_Direct3D_EnvironmentRenderer::RenderToShadow(xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane(), xShadow.GetInstance()->IsCubeMapped());

		Asura_Camera::SetNearClip(fNear);
		Asura_Camera::SetFarClip(fFar);
		Asura_Camera::SetPosition(xCamPos);
		Asura_Camera::SetOrientation(xCamOrient);
		Asura_Camera::SetHorizontalFieldOfView(fCamFOV, 1.0f / fCamAspect);
		Asura_Direct3D_Effects_System::SetSharedParameters();
		Asura_Render::CurrentStates.RequestDynamicLights( true);
	}
}
#endif	//defined(ASURA_TARGET_SM3)


////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_RenderOntoScene()
//
// Renders this frame's shadows onto the scene
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_RenderOnToScene(ASURA_SHADOW_LIGHTSOURCE eLightSource, uint32 uLightSourceHandle)
{
#if defined(ASURA_TARGET_SM3)
	const Asura_Shadow_Instance *apxCubeMapShadows[ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS];
	u_int uCubeMapShadowsFound = 0;

	// DMR 03/04/07
	//
	// If we're rendering shadows from averaged lights, they're still done as a darkening pass, as before.
	//
	// If we're rendering shadows from a real light source, we write them to the destination alpha channel, and the deferred light
	// pass will use a D3DBLEND_INVDESTALPHA blend, to mask out the shadowed areas.
	//
	// This relies on the destination alpha channel (at least the area affected by the light) being cleared to 0 alpha before this call.

	HRESULT xResult;
	
	if( 0 == s_uNumShadows || !g_pxApplyShadowsEffect)
	{
		return;
	}
	
	SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), L"Apply %d Shadows", s_uNumShadows));

	Asura_Direct3D_RenderTarget* pxDeferredRenderTarget = Asura_Direct3D::GetDeferredRenderingTarget();
	if( NULL == pxDeferredRenderTarget)
	{
		return;
	}
	
	Direct3D_Surface* pxSurfaceRenderTarget = pxDeferredRenderTarget->GetSurface( ASURA_DIRECT3D_DEFFEDRENDERTARGET_FRONT_INDEX);
	if( NULL == pxSurfaceRenderTarget)
	{
		return;
	}
	
	Asura_Direct3D::ClearRenderTargets( 1);
	
	xResult = Asura_Direct3D::D3DDevice->SetRenderTarget( 0, pxSurfaceRenderTarget);
	ASURA_D3D_OK( xResult);

	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

	Asura_Direct3D::D3DDevice->SetVertexDeclaration( g_pxShadowFrustumVD );
	Asura_Direct3D::D3DDevice->SetStreamSource(0, g_pxShadowFrustumVB, 0, uVERTEX_SIZE);
	Asura_Direct3D::D3DDevice->SetIndices(g_pxShadowFrustumIB);
	
	// (AAM - 01.09.2006) Best preserve these, they're set in the effect
	DWORD auMinFilter[2], auMagFilter[2], auMipFilter[2], auAddressU[2], auAddressV[2];
	for( u_int u = 0; u < 2; ++u )
	{
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_MINFILTER, &auMinFilter[u] );
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_MAGFILTER, &auMagFilter[u] );
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_MIPFILTER, &auMipFilter[u] );
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_ADDRESSU, &auAddressU[u] );
		Asura_Direct3D::D3DDevice->GetSamplerState( u, D3DSAMP_ADDRESSV, &auAddressV[u] );
	}

	g_xApplyShadowParameters.m_bFilterShadows = false;

	g_xApplyShadowParameters.m_bShadowMask = eLightSource != ASURA_SHADOW_LIGHTSOURCE_AVERAGED;


	// If we're creating a shadow mask for a light source, using the MAX operation is 
	// a nice way of avoiding double-shadowing
	if (g_xApplyShadowParameters.m_bShadowMask)
	{
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
		Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_BLENDOP, D3DBLENDOP_MAX);
	}
	
	// Begin rendering
	Asura_Assert(g_pxApplyShadowsEffect->IsD3DEffectValid(), "Shadow renderer effect invalid");
	if (g_pxApplyShadowsEffect->IsD3DEffectValid())
	{

		g_xApplyShadowParameters.m_fCameraTanHalfVertFOV = Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
		g_xApplyShadowParameters.m_fCameraTanHalfHorizFOV = g_xApplyShadowParameters.m_fCameraTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();
		g_xApplyShadowParameters.m_fTwoOverScreenWidth = 2.0f / Asura_Viewport::GetWidth();
		g_xApplyShadowParameters.m_fTwoOverScreenHeight = 2.0f / Asura_Viewport::GetHeight();

		for( u_int u = 0; u < s_uNumShadows; ++u )
		{
			Asura_Direct3D_Shadow*	pxShadow	= static_cast<Asura_Direct3D_Shadow*>(s_apxShadows[u]);
			const Asura_Shadow_Instance*	pxInstance	= pxShadow->GetInstance();

			bool bAccept = false;			

			if (pxInstance->GetLightSourceType() == eLightSource)
			{
				if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
				{
					// We're casting shadows from an averaged light source - there's no light source handle 
					bAccept = true;
				}
				else
				{
					// We're casting shadows from a real light source - make sure this is the right light source
					if (uLightSourceHandle == pxInstance->GetLightSourceHandle())
					{
						bAccept = true;					
					}
				}		
			}
			else
			{
				if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
				{
					// If we're rendering averaged shadows, we also want to render any darkening pass 
					// shadows from specific light sources here, too
					if (pxInstance->GetShadowFlags()&ASURA_SHADOWFLAG_IS_DARKENING_PASS)
					{
						bAccept = true;
					}
				}
			}
			if (bAccept	&& pxInstance->IsCubeMapped())
			{
				if (uCubeMapShadowsFound < ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS)
				{
					apxCubeMapShadows[uCubeMapShadowsFound++] = pxInstance;
				}
			} else if (bAccept && pxShadow->GetShadowBuffer() && pxShadow->GetShadowBuffer()->GetTexture()) {
				const Asura_Frustum&			xFrustum	= pxInstance->GetFrustum();				

				SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), L"Shadow %u", u));

				D3DXMATRIX xShadowProj, xInvShadowProj;
				D3DXMatrixPerspectiveFovLH(&xShadowProj, xFrustum.GetVerticalFOVInDegrees() * Asura_Maths::Pi / 180.f, 1.f, xFrustum.GetNearPlane(), xFrustum.GetFarPlane());
				D3DXMatrixInverse(&xInvShadowProj, NULL, &xShadowProj);

				D3DXMATRIX xShadowView, xInvShadowView;
				{
					// Light position
					const Asura_Vector_3& xLightPos = pxInstance->GetLightPosition();
					D3DXMatrixTranslation( &xShadowView, -xLightPos.x, -xLightPos.y, -xLightPos.z );
					g_xApplyShadowParameters.m_xLightPos = D3DXVECTOR4(xLightPos.x, xLightPos.y, xLightPos.z, 1);

					// Light orientation
					D3DXMATRIX xTemp;
					Asura_Matrix_3x3 xLightOri;
					pxInstance->GetLightOrientation(xLightOri);
					D3DXMatrixIdentity( &xTemp );
					for( u_int v = 0; v < 3; ++v )
					{
						for( u_int w = 0; w < 3; ++w )
						{
							xTemp.m[v][w] = xLightOri.Element[v][w];
							// No need to invert y axis here, the shader can just compute an upside-down y
						}
					}
					xShadowView *= xTemp;
				}
				D3DXMatrixInverse(&xInvShadowView, NULL, &xShadowView);

				D3DXMATRIX xView, xInvView;
				xView = Asura_Direct3D::GetViewMatrix();
				D3DXMatrixInverse(&xInvView, NULL, &xView);

				g_xApplyShadowParameters.m_xCameraViewSpaceToWorldSpace = xInvView;
				
				g_xApplyShadowParameters.m_xLightClipSpaceToCameraClipSpace = xInvShadowProj * xInvShadowView * Asura_Direct3D::GetViewProjectionMatrix();
				g_xApplyShadowParameters.m_xCameraViewSpaceToLightClipSpace = xInvView * xShadowView * xShadowProj;
				g_xApplyShadowParameters.m_xCameraViewSpaceToLightViewSpace = xInvView * xShadowView;

				g_xApplyShadowParameters.m_fCameraTanHalfVertFOV = Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
				g_xApplyShadowParameters.m_fCameraTanHalfHorizFOV = g_xApplyShadowParameters.m_fCameraTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();

				float fFilterFade = 1.0f;
				if ( Asura_Direct3D_UserOptions::GetFilterShadows() && !g_bDisableFiltering)
				{		
					g_xApplyShadowParameters.m_bFilterShadows = (pxInstance->GetDistance() < Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering);

					fFilterFade = 0.25f * (Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering - pxInstance->GetDistance());
					fFilterFade = Asura_Maths::ClampToRange(fFilterFade, 0.0f,1.0f);
					
					g_xApplyShadowParameters.m_bUseNoise = (pxInstance->GetShadowFlags() & ASURA_SHADOWFLAG_USE_NOISE) != 0; 
				}


				g_xApplyShadowParameters.m_fShadowFilterTextureScale = fFilterFade * 11.0f / ((float)(pxShadow->GetShadowBuffer()->GetTextureSize())); 
				g_xApplyShadowParameters.m_fShadowFilterSize = 0.2f*pxInstance->GetSoftness();

				Asura_Direct3D_Effects_System::GetSharedEffect()->SetParameterByName("g_fShadowFrustumNearPlaneDistance", xFrustum.GetNearPlane());
				Asura_Direct3D_Effects_System::GetSharedEffect()->SetParameterByName("g_fOneOverShadowFrustumPlaneSeparation", 1.f / (xFrustum.GetFarPlane() - xFrustum.GetNearPlane()));

				g_xApplyShadowParameters.m_pxShadowBuffer = pxShadow->GetShadowBuffer()->GetTexture();
				g_xApplyShadowParameters.m_pxNormalDepthImage = pxDeferredRenderTarget->GetTexture( ASURA_DIRECT3D_DEFFEDRENDERTARGET_NORMAL_INDEX);

				g_xApplyShadowParameters.m_pxKernelRotationTexture = g_pxFilterKernelRotationTexture;

				Asura_Vector_3 xDir = pxShadow->GetLightDirection();
				g_xApplyShadowParameters.m_xShadowDirection.x = xDir.x;
				g_xApplyShadowParameters.m_xShadowDirection.y = xDir.y;
				g_xApplyShadowParameters.m_xShadowDirection.z = xDir.z;			

				g_xApplyShadowParameters.m_fShadowStrength = pxInstance->GetStrength();

				if (Asura_PC_ShadowBuffer::IsNvidiaDepthMap())
				{
					g_xApplyShadowParameters.m_xSelfShadowBias = s_xSelfShadowBiasValues[pxInstance->IsEnvironment()];
				} 
				else 
				{
					g_xApplyShadowParameters.m_xSelfShadowBias = s_xSelfShadowBiasValues[2 + pxInstance->IsEnvironment()];
				}


 				g_xApplyShadowParameters.m_xUVXForm._24 = g_xApplyShadowParameters.m_xUVXForm._14 
 					= 0.5f + (0.5f / pxShadow->GetShadowBuffer()->GetTextureSize());

				g_xApplyShadowParameters.m_xUVXForm._14 += g_fXShadowOffset;
				g_xApplyShadowParameters.m_xUVXForm._24 += g_fYShadowOffset;


				if (Asura_PC_ShadowBuffer::IsNvidiaDepthMap())
				{
					Asura_Direct3D_Effects_System::BeginEffect(*g_pxApplyShadowsEffect, g_xApplyShadowParameters.m_bFilterShadows ? SHADOW_EFFECT_PASS_DEPTHMAP_MULTITAP : SHADOW_EFFECT_PASS_DEPTHMAP_SINGLETAP);
				}
				else
				{
					Asura_Direct3D_Effects_System::BeginEffect(*g_pxApplyShadowsEffect, g_xApplyShadowParameters.m_bFilterShadows ? SHADOW_EFFECT_PASS_SHADOWMAP_MULTITAP : SHADOW_EFFECT_PASS_SHADOWMAP_SINGLETAP);
				}

				g_xApplyShadowParameters.m_pxParamBlock->Apply();

				g_pxApplyShadowsEffect->RenderAllIndexedPrimitivePasses(D3DPT_TRIANGLELIST, 0, 8, 0, sizeof(ausFRUSTUM_INDICES) / (sizeof(u_short)*3), NULL);				
			}
		}
	}

	for (u_int uCubeMappedShadow = 0;uCubeMappedShadow < uCubeMapShadowsFound; uCubeMappedShadow++)
	{
		const Asura_Shadow_Instance *pxInstance = apxCubeMapShadows[uCubeMappedShadow];
		float fNear = pxInstance->GetFrustum().GetNearPlane();
		float fFar = pxInstance->GetFrustum().GetFarPlane();
		float fRange = fFar - fNear;
		float fInvRange = 1.0f / fRange;


		SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), L"Cubic Env Map Shadow"));


		g_xApplyShadowParameters.m_pxShadowBuffer = NULL;
		g_xApplyShadowParameters.m_pxShadowBufferCube = NULL;
		g_xApplyShadowParameters.m_pxParamBlock->Apply();

		Asura_Direct3D_Effects_System::GetSharedEffect()->SetParameterByName("g_fShadowFrustumNearPlaneDistance", fNear);
		Asura_Direct3D_Effects_System::GetSharedEffect()->SetParameterByName("g_fOneOverShadowFrustumPlaneSeparation", fInvRange);

		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED);
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 

		//	Asura_CameraData xCamData = Asura_Camera::GetCameraData(Asura_Camera::GetCameraIndex());
		Asura_Vector_3 xCamPos = *Asura_Camera::GetPosition();
		Asura_Matrix_3x3 xCamOrient = *Asura_Camera::GetOrientation();
		Asura_Angle fCamFOV = Asura_Camera::GetHorizontalFieldOfView();
		float fCamNear = Asura_Camera::GetNearClip();
		float fCamFar = Asura_Camera::GetFarClip();
		float fCamAspect = Asura_Camera::GetXOverYAspectRatio() / Asura_Camera::GetPhysicalScreenAspectRatio();


		for (int i = 0; i < 6; ++i)
		{
			static const Asura_Matrix_3x3 axORIENTATIONS[6] =
			{
				// Looking at +x face
				Asura_Matrix_3x3(  0.0f, 0.0f, 1.0f,
				0.0f, 1.0f, 0.0f,
				-1.0f, 0.0f, 0.0f ),

				// Looking at -x face
				Asura_Matrix_3x3( 0.0f, 0.0f, -1.0f,
				0.0f, 1.0f,  0.0f,
				1.0f, 0.0f,  0.0f ),

				// Looking at -y face
				Asura_Matrix_3x3( 1.0f, 0.0f,  0.0f,
				0.0f, 0.0f, -1.0f,
				0.0f, 1.0f,  0.0f ),

				// Looking at +y face
				Asura_Matrix_3x3( 1.0f,  0.0f, 0.0f,
				0.0f,  0.0f, 1.0f,
				0.0f, -1.0f, 0.0f ),

				// Looking at +z face
				Asura_Matrix_3x3( 1.0f, 0.0f, 0.0f,
				0.0f, 1.0f, 0.0f,
				0.0f, 0.0f, 1.0f ),

				// Looking at -z face
				Asura_Matrix_3x3( -1.0f, 0.0f,  0.0f,
				0.0f, 1.0f,  0.0f,
				0.0f, 0.0f, -1.0f ),
			};
			const Asura_Matrix_3x3 &xOrient = axORIENTATIONS[i];

			IDirect3DSurface9 *pxCubeSurface;
#ifdef ASURA_HIGH_RES_SHADOW_MODE
			if (Asura_Render::IsHiResScreenshotModeEnabled())
			{
				Asura_PC_ShadowStorage::Get().GetEnvShadowCubeTexture()->GetCubeMapSurface((D3DCUBEMAP_FACES)i, 0, &pxCubeSurface );
				Asura_Direct3D::D3DDevice->SetRenderTarget(0, pxCubeSurface);
				Asura_Direct3D::D3DDevice->SetDepthStencilSurface(Asura_PC_ShadowStorage::Get().GetShadowDepthStencil(0));
				Asura_Viewport::Set(ASURA_MAX_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZE, ASURA_MAX_HIRES_SCREENSHOT_SHADOW_TEXTURE_SIZE);
			}
			else
#endif
			{
				Asura_PC_ShadowStorage::Get().GetEnvShadowCubeTexture()->GetCubeMapSurface((D3DCUBEMAP_FACES)i, 0, &pxCubeSurface );
				Asura_Direct3D::D3DDevice->SetRenderTarget(0, pxCubeSurface);
				Asura_Direct3D::D3DDevice->SetDepthStencilSurface(Asura_PC_ShadowStorage::Get().GetShadowDepthStencil(0));
				Asura_Viewport::Set(Asura_Shadow_System::s_uMaxShadowTextureSize, Asura_Shadow_System::s_uMaxShadowTextureSize);
			}

			Asura_Direct3D::D3DDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF ), 1.0f, 0 );
	

			Asura_Camera::SetPosition( pxInstance->GetLightPosition() );
			Asura_Camera::SetHorizontalFieldOfView( Asura_Maths::Deg90, 1.0f / Asura_Camera::GetPhysicalScreenAspectRatio() );
			Asura_Camera::SetOrientation(xOrient);
			Asura_Camera::SetNearClip(fCamNear);
			Asura_Camera::SetFarClip(fCamFar);
			Asura_Direct3D_EnvironmentRenderer::RenderToShadow(fNear, fFar, true);

			SAFE_RELEASE(pxCubeSurface);
		}
		Asura_Viewport::Set(Asura_Direct3D::GetDeferredRenderingTarget()->GetWidth(), Asura_Direct3D::GetDeferredRenderingTarget()->GetHeight());
		Asura_Camera::SetNearClip(fCamNear);
		Asura_Camera::SetFarClip(fCamFar);
		Asura_Camera::SetPosition(xCamPos);
		Asura_Camera::SetOrientation(xCamOrient);
		Asura_Camera::SetHorizontalFieldOfView(fCamFOV, 1.0f / fCamAspect);
		Asura_Direct3D_Effects_System::SetSharedParameters();
		Asura_Render::CurrentStates.RequestDynamicLights( true);



		Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE, pxInstance->GetShadowFlags()&ASURA_SHADOWFLAG_IS_DARKENING_PASS? 
			D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE :
			D3DCOLORWRITEENABLE_ALPHA
		);

		Asura_Direct3D::D3DDevice->SetRenderTarget( 0, pxSurfaceRenderTarget);
		Asura_Direct3D::D3DDevice->SetDepthStencilSurface( NULL);
		g_xApplyShadowParameters.m_pxShadowBufferCube = Asura_PC_ShadowStorage::Get().GetEnvShadowCubeTexture();

		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
		Asura_Render::CurrentStates.RequestZBufferEnabled( false );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

		Asura_Direct3D::D3DDevice->SetVertexDeclaration( g_pxShadowFrustumVD );
		Asura_Direct3D::D3DDevice->SetStreamSource(0, g_pxShadowFrustumVB, 0, uVERTEX_SIZE);
		Asura_Direct3D::D3DDevice->SetIndices(g_pxShadowFrustumIB);


		D3DXMATRIX xShadowView, xInvShadowView;
		{
			// Light position
			const Asura_Vector_3& xLightPos = pxInstance->GetLightPosition();
			D3DXMatrixTranslation( &xShadowView, -xLightPos.x, -xLightPos.y, -xLightPos.z );
			g_xApplyShadowParameters.m_xLightPos = D3DXVECTOR4(xLightPos.x, xLightPos.y, xLightPos.z, 1);
		}
		D3DXMatrixInverse(&xInvShadowView, NULL, &xShadowView);
		xInvShadowView._11 *= fFar; xInvShadowView._12 *= fFar; xInvShadowView._13 *= fFar;
		xInvShadowView._21 *= fFar; xInvShadowView._22 *= fFar; xInvShadowView._23 *= fFar;
		xInvShadowView._31 *= fFar; xInvShadowView._32 *= fFar; xInvShadowView._33 *= fFar;

		g_xApplyShadowParameters.m_xLightClipSpaceToCameraClipSpace = xInvShadowView * Asura_Direct3D::GetViewProjectionMatrix();


		D3DXMATRIX xView, xInvView;
		xView = Asura_Direct3D::GetViewMatrix();
		D3DXMatrixInverse(&xInvView, NULL, &xView);

		g_xApplyShadowParameters.m_xCameraViewSpaceToWorldSpace = xInvView;

		g_xApplyShadowParameters.m_xLightClipSpaceToCameraClipSpace = xInvShadowView * Asura_Direct3D::GetViewProjectionMatrix();


		g_xApplyShadowParameters.m_fCameraTanHalfVertFOV = Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
		g_xApplyShadowParameters.m_fCameraTanHalfHorizFOV = g_xApplyShadowParameters.m_fCameraTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();

		float fFilterFade = 1.0f;
		if ( Asura_Direct3D_UserOptions::GetFilterShadows() && !g_bDisableFiltering)
		{		
			g_xApplyShadowParameters.m_bFilterShadows = true;
			g_xApplyShadowParameters.m_bUseNoise = (pxInstance->GetShadowFlags() & ASURA_SHADOWFLAG_USE_NOISE) != 0; 

			fFilterFade = 0.25f * (Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering - 0.0f);
			fFilterFade = Asura_Maths::ClampToRange(fFilterFade, 0.0f,1.0f);
		}

		g_xApplyShadowParameters.m_fShadowFilterTextureScale = fFilterFade * 11.0f / ((float)(Asura_Shadow_System::s_uMaxShadowTextureSize)); 
		g_xApplyShadowParameters.m_fShadowFilterSize = 0.2f*pxInstance->GetSoftness();

		g_xApplyShadowParameters.m_pxNormalDepthImage = pxDeferredRenderTarget->GetTexture( ASURA_DIRECT3D_DEFFEDRENDERTARGET_NORMAL_INDEX);

		g_xApplyShadowParameters.m_pxKernelRotationTexture = g_pxFilterKernelRotationTexture;

		g_xApplyShadowParameters.m_fShadowStrength = pxInstance->GetStrength();

		g_xApplyShadowParameters.m_xSelfShadowBias = s_xSelfShadowBiasValues[4];

		if (g_xApplyShadowParameters.m_bShadowMask)
		{
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
			Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_BLENDOP, D3DBLENDOP_MAX);
		}


		Asura_Direct3D_Effects_System::BeginEffect(*g_pxApplyShadowsEffect, g_xApplyShadowParameters.m_bFilterShadows ? SHADOW_EFFECT_PASS_ENVCUBEMAP_MULTITAP : SHADOW_EFFECT_PASS_ENVCUBEMAP_SINGLETAP);

		g_xApplyShadowParameters.m_pxParamBlock->Apply();

		Asura_Direct3D::D3DDevice->SetVertexDeclaration( g_pxShadowFrustumVD );

		// Are we inside the volume?
		Asura_Vector_3 xOfs = *Asura_Camera::GetPosition() - pxInstance->GetLightPosition();
		float fDist = xOfs.Magnitude();

		// Pick a sphere LOD. Switch down quickly to the low LOD, as we only need a more round
		// sphere when the camera is about to move inside it, for a more accurate inside-outside test
		int iSphereLOD = (int)(fDist*0.5f/pxInstance->GetFrustum().GetFarPlane());
		if (iSphereLOD>3)
		{
			iSphereLOD = 3;
		}		
		else if (iSphereLOD<0)
		{
			iSphereLOD = 0;
		}


		extern void Asura_Direct3D_GetPhotonSphereVB(u_int uLOD, Direct3D_VertexBuffer *&pxVertexBuffer, u_int &uNumVerts);
		Direct3D_VertexBuffer *pxShadowFrustumVBSphere;
		u_int uNumVerts;

		Asura_Direct3D_GetPhotonSphereVB(iSphereLOD, pxShadowFrustumVBSphere, uNumVerts);
		Asura_Direct3D::D3DDevice->SetStreamSource(0, pxShadowFrustumVBSphere,0, sizeof (Asura_Vector_3));

		g_pxApplyShadowsEffect->RenderAllPrimitivePasses(D3DPT_TRIANGLESTRIP, 0, uNumVerts - 2);				


		g_xApplyShadowParameters.m_pxShadowBufferCube = NULL;
	}


	g_xApplyShadowParameters.m_pxShadowBuffer = NULL;
	g_xApplyShadowParameters.m_pxNormalDepthImage = NULL;
	g_xApplyShadowParameters.m_pxKernelRotationTexture = NULL;
	g_xApplyShadowParameters.m_pxParamBlock->Apply();
	g_pxApplyShadowsEffect->GetD3DEffect()->CommitChanges();

	// Flush any effects we may still be using...
	Asura_Direct3D_Effects_System::FlushEffects	();
	
	Asura_Direct3D::GetDeferredRenderingTarget()->Bind();

	for( u_int u = 0; u < 2; ++u )
	{
		Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_MINFILTER, auMinFilter[u] );
		Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_MAGFILTER, auMagFilter[u] );
		Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_MIPFILTER, auMipFilter[u] );
		Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_ADDRESSU, auAddressU[u] );
		Asura_Direct3D::D3DDevice->SetSamplerState( u, D3DSAMP_ADDRESSV, auAddressV[u] );
	}

	Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_BLENDOP, D3DBLENDOP_ADD);
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );

	s_uNumShadowInstancesRendered++;

#endif
}


/******************************************************************************/
/* Asura_Direct3D_Shadow Class Functions                                       */
/******************************************************************************/
#if defined(ASURA_TARGET_SM3)
////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Shadow::Asura_Direct3D_Shadow()
//
// Constructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Shadow::Asura_Direct3D_Shadow()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Shadow::~Asura_Direct3D_Shadow()
//
// Destructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Direct3D_Shadow::~Asura_Direct3D_Shadow()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Shadow::PrepareForRendering()
//
// Begins rendering into a shadow map
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Shadow::PrepareForRendering( void )
{
	D3DPERFCALL(BeginEvent(D3DCOLOR_XRGB( 0xC0, 0x60, 0x30 ), L"Create Shadow" ));

	if (m_pxShadowBuffer)
	{
		m_pxShadowBuffer->Bind();
		Asura_Viewport::Set(m_pxShadowBuffer->GetTextureSize(), m_pxShadowBuffer->GetTextureSize());
		if (Asura_PC_ShadowBuffer::IsNvidiaDepthMap())
		{
			Asura_Direct3D::D3DDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF ), 1.0f, 0 );
		} 
		else 
		{
			Asura_Direct3D::D3DDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF ), 1.0f, 0 );
		}
	}

	Asura_Direct3D_Effects_System::GetSharedEffect()->SetParameterByName("g_fShadowFrustumNearPlaneDistance", m_pxInstance->GetFrustum().GetNearPlane());
	Asura_Direct3D_Effects_System::GetSharedEffect()->SetParameterByName("g_fOneOverShadowFrustumPlaneSeparation", 1.f / (m_pxInstance->GetFrustum().GetFarPlane() - m_pxInstance->GetFrustum().GetNearPlane()));
}

static const char* const g_szPS_Deferred_DebugShowBuffers_Shadow_Code =
"ps_3_0								\n"
"dcl_texcoord0 v0.xy				\n"
"def c0, 0, 1, 5.9604648328104515558750364705942e-7,0\n"
"dcl_2d s0							\n"
"dcl_2d s1							\n"
"texld r0, v0, s0					\n"
"mul r0, r0, c0.z                   \n"
"sub r0, r0, c0.x                   \n"
"mul r0, r0, c0.y                   \n"
"texld r0, r0, s1					\n"
"mov oC0, r0						\n"
;

const Asura_Direct3D_PixelShader xPS_Deferred_DebugShowBuffers_Shader_Shadow( ASURA_D3D_PS_DEFERRED_DEBUGSHOWBUFFERS_SHADOWBUFFER, g_szPS_Deferred_DebugShowBuffers_Shadow_Code, 3.0);

#ifdef _DEBUG
void Asura_Shadow_System::ShowDebugShadowMaps()
{
	static Direct3D_Texture *pxTexture;
	if (!pxTexture)
	{
		D3DLOCKED_RECT Rect;
		Asura_Direct3D::D3DDevice->CreateTexture(9,1,1,0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &pxTexture, NULL);
		pxTexture->LockRect(0, &Rect, NULL, 0);
		((unsigned int*)Rect.pBits)[0] = 0xff000000;
		((unsigned int*)Rect.pBits)[1] = 0xffff0000;
		((unsigned int*)Rect.pBits)[2] = 0xff00ff00;
		((unsigned int*)Rect.pBits)[3] = 0xff0000ff;
		((unsigned int*)Rect.pBits)[4] = 0xff00ffff;
		((unsigned int*)Rect.pBits)[5] = 0xffffff00;
		((unsigned int*)Rect.pBits)[6] = 0xffff00ff;
		((unsigned int*)Rect.pBits)[7] = 0xff000000;
		((unsigned int*)Rect.pBits)[8] = 0xffffffff;
		pxTexture->UnlockRect(0);
	}
	Asura_Render::CurrentStates.RequestTextureIndex( 0, ASURA_TEXTURE_INVALID);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
	Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_QUADVERTEX);
	Asura_Render::CurrentStates.RequestZBufferEnabled( false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestAlphaTest(false);

	Asura_Direct3D::ClearRenderTargets( 1);

	Asura_Direct3D_Shaders::RequestVertexShader( ASURA_D3D_VS_DEFERRED_DEBUGSHOWBUFFERS);

	float fX = 0.0f, fY = 0.0f;

	for( u_int uDrawTargetIndex = 0; uDrawTargetIndex < Asura_Shadow_System::s_uNumShadows; uDrawTargetIndex += 1)
	{
		Asura_Direct3D_Shadow* pxShadow = (Asura_Direct3D_Shadow*) Asura_Shadow_System::s_apxShadows[uDrawTargetIndex];
		if (!pxShadow || !pxShadow->GetShadowBuffer())
			continue;

		Direct3D_Texture* pxMRTTexture = pxShadow->GetShadowBuffer()->GetTexture();

		//apxDrawTextures[ eRenderTargetTexture];
		if( NULL == pxMRTTexture)
		{
			continue;
		}
		float afPSConsts[5][4] =
			{ 
				{ 1.0f, 0.0f, 0.0f, 0.0f}, 
				{ 0.0f, 1.0f, 0.0f, 0.0f}, 
				{ 0.0f, 0.0f, 1.0f, 0.0f}, 
				{ 0.0f, 0.0f, 0.0f, 1.0f}, 
				{ 0.0f, 0.0f, 0.0f, 0.0f}
			};

		Asura_Direct3D_Shaders::RequestPixelShader( ASURA_D3D_PS_DEFERRED_DEBUGSHOWBUFFERS_SHADOWBUFFER);
		Asura_Direct3D::D3DDevice->SetPixelShaderConstantF( 0, afPSConsts[0], 5);
		Asura_Direct3D::D3DDevice->SetTexture( 0, pxMRTTexture);
		Asura_Direct3D::D3DDevice->SetTexture( 1, pxTexture);

		const static float fOutputSizeTotalX = 4.0f;
		const static float fOutputSizeTotalY = 1.0f;
		const float fStartX = fX / fOutputSizeTotalX;
		const float fEndX = (fX + 1.0f) / fOutputSizeTotalX;
		const float fStartY = fY / fOutputSizeTotalY;
		const float fEndY = (fY + 1) / fOutputSizeTotalY;
		fX += 1.0f;
		if (fX >= fOutputSizeTotalX)
		{
			fX = 0.0f;
			fY += 1.0f;
		}

		struct _D3DQuadVertex { float X, Y, Z, U, V; } axVert[4] =
		{
			{ -1.0f + (2.0f * fStartX), -0.6f - (0.4f * fStartY), 1.0f, 0.0f, 0.0f},
			{ -1.0f + (2.0f * fStartX), -0.6f - (0.4f * fEndY), 1.0f, 0.0f, 1.0f},
			{ -1.0f + (2.0f * fEndX), -0.6f - (0.4f * fStartY), 1.0f, 1.0f, 0.0f},
			{ -1.0f + (2.0f * fEndX), -0.6f - (0.4f * fEndY), 1.0f, 1.0f, 1.0f}
		};

		Asura_Direct3D::D3DDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, axVert, sizeof( axVert[0]));
	}

	// Tidy-up...
	DestroyShadowInstances();

}
#endif



////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_Shadow::FinishedRendering()
//
// Finishes rendering into a shadow map
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Direct3D_Shadow::FinishedRendering()
{
	Asura_Viewport::Set(Asura_Direct3D::GetDeferredRenderingTarget()->GetWidth(), Asura_Direct3D::GetDeferredRenderingTarget()->GetHeight());
	D3DPERFCALL(EndEvent()); // Create Shadow
}
#endif


/// End of file. ///
