
//
// Asura_Platform_PFX.cpp
//
// Created by KOR - 16/07/2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//
#include "Core/Asura.h"


// This file's header.
#include "SpecialFX/PFX/Asura_PFX_Effect.h"

// Asura headers
#include "Debugging/Asura_Debug_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"
#include "Light/Asura_GI_Solution.h"
#include "SpecialFX/PFX/Asura_PFX_Classes.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Direct3D_Caps.h"
#include "Rendering/Asura_Direct3D_Shaders.h"
#include "Rendering/Asura_Direct3D_BackBufferCapture.h"
#include "Rendering/Asura_Direct3D_RenderTargetManager.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"

#ifdef ASURA_USE_FASTPATH_PFX
#include "SpecialFX/Asura_Platform_PFX_System_FastPath.h"
#endif

#ifndef ASURA_TARGET_SM3


#ifdef ASURA_TOOLS
#include "Particles/Tools_Mode_PFX.h"
#endif

static float g_fHEAT_HAZE_OFFSET_V = 0.0f;

#ifndef ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD
    // The position beyond which a particle position is considered to be suspicious.
    #define	ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD  ( 6000.0f )
#endif 

// Texture used when we need an intermediate power of two render target
// rather than using the backbuffer as a texture directly. (Only needed
// with certain hardware.)
static Direct3D_Texture* g_pxPowerOfTwoBackbufferTexture = NULL;

#ifdef ASURA_TOOLS
static Asura_RandomGenerator* g_pxHeatHazeRNG = 0;
#endif

static void CreatePowerOfTwoBackbufferTexture()
{
	Asura_Assert( !g_pxPowerOfTwoBackbufferTexture, "Create called when texture already exists" );

	u_int uRTWidth = 1 << ((int)(logf((float)Asura_Direct3D::GetPresentParameters().BackBufferWidth) / logf(2.f)));
	u_int uRTHeight = 1 << ((int)(logf((float)Asura_Direct3D::GetPresentParameters().BackBufferWidth) / logf(2.f)));

	Asura_Direct3D::D3DDevice->CreateTexture( uRTWidth, uRTHeight, 1, D3DUSAGE_RENDERTARGET, Asura_Direct3D::GetPresentParameters().BackBufferFormat, D3DPOOL_DEFAULT, &g_pxPowerOfTwoBackbufferTexture, NULL );

	Asura_Assert( g_pxPowerOfTwoBackbufferTexture, "Could not create texture" );
}

static void DestroyPowerOfTwoBackbufferTexture()
{
	if( g_pxPowerOfTwoBackbufferTexture )
	{
		g_pxPowerOfTwoBackbufferTexture->Release();
		g_pxPowerOfTwoBackbufferTexture = NULL;
	}
}


void Asura_PFX_System::Platform_Initialise( void )
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Asura_PFX_System::Platform_Initialise");
	
#ifdef ASURA_USE_FASTPATH_PFX
	Asura_Platform_PFX_System_FastPath::Initialise();
#endif

	#ifdef ASURA_TOOLS
		// load basic textures used by the system
		// (SJB) Outside tools, these are loaded from common.asr
		Asura_Texture_Management::AddTextureWithResetGroup("\\specialfx\\heathaze.bmp", ASURA_TEXTURE_LOADED_PERMANENT );
	#endif

	// Do we need an intermediate power of two texture to use the backbuffer as a texture?
	if( Asura_Direct3D_Caps::IsNonPow2Conditional() )
	{
		// create the texture, and add callbacks to cope with Direct3D being reset
		CreatePowerOfTwoBackbufferTexture();
		Asura_Direct3D_RenderTargetManager::AddResetCallbacks( DestroyPowerOfTwoBackbufferTexture, CreatePowerOfTwoBackbufferTexture );
	}
}

void Asura_PFX_System::Platform_Reset( void )
{

}

void Asura_PFX_System::Platform_Shutdown( void )
{
#ifdef ASURA_USE_FASTPATH_PFX
	Asura_Platform_PFX_System_FastPath::Shutdown();
#endif

	if( Asura_Direct3D_Caps::IsNonPow2Conditional() )
	{
		Asura_Direct3D_RenderTargetManager::RemoveResetCallbacks( DestroyPowerOfTwoBackbufferTexture, CreatePowerOfTwoBackbufferTexture );
		DestroyPowerOfTwoBackbufferTexture();
	}
}

static DWORD FloatToDW(float fVal)
{
	return *(DWORD*)&fVal;
}

DEBUGVAR u_int		dbg_uAlphaTestRefValue = 1;

void Asura_PFX_System::Platform_BeginRender()
{
	HRESULT xResult;
	
	#if ASURA_D3D_ENABLE_PIX
		const wchar_t* const wszString = s_bIsRenderingHeatHazeSources ? L"PFX Render (Heat Haze)"
									   : GetCurrentRenderType() == ASURA_PFX_RENDER_HUD ? L"PFX Render (HUD)"
									   : GetCurrentRenderType() == ASURA_PFX_RENDER_SKYBOX ? L"PFX Render (Skybox)"
									   : GetCurrentRenderType() == ASURA_PFX_RENDER_PROJECT ? L"PFX Render (Project)"
									   : GetCurrentRenderType() == ASURA_PFX_RENDER_WATER_DISTORTER ? L"PFX Render (Water Distorter)"
									   : L"PFX Render (Normal)";
	#endif
	D3DPERFCALL(BeginEvent(0xff885577, wszString));

	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, 0 );

	int iHeatHazeTexID = Asura_Texture_Management::GetTextureIndex( "\\specialfx\\heathaze.bmp" );

	Asura_Render::CurrentStates.RequestAlphaTest(true, dbg_uAlphaTestRefValue);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);

	if( !s_bIsRenderingHeatHazeSources )
	{
		// (MH - 11.01.2006) TextureOpMode is now set for each render; 2x only needed if template uses lighting.
		//Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS_2X);
		Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);
	}
	else
	{
		// Assume that the backbuffer capture is dirty - we just rendered a bunch of particles to it

		// do we need an intermediate power of two texture (or can we use backbuffer directly)?
		if( Asura_Direct3D_Caps::IsNonPow2Conditional() )
		{
			Asura_Assert( g_pxPowerOfTwoBackbufferTexture, "Backbuffer capture texture missing" );

			if ( g_pxPowerOfTwoBackbufferTexture )
			{
				Direct3D_Surface* pxBackBuffer = NULL;
				Asura_Direct3D::D3DDevice->GetRenderTarget(0, &pxBackBuffer);
				if( !pxBackBuffer ) return;

				Direct3D_Surface* pxSL0 = NULL;
				g_pxPowerOfTwoBackbufferTexture->GetSurfaceLevel(0, &pxSL0);
				Asura_Assert(pxSL0, "Could not get SL0");

				xResult = Asura_Direct3D::D3DDevice->StretchRect(pxBackBuffer, NULL, pxSL0, NULL, D3DTEXF_NONE);
				ASURA_D3D_OK( xResult);
				pxBackBuffer->Release();
				pxSL0->Release();

				Asura_Direct3D::D3DDevice->SetTexture(2, g_pxPowerOfTwoBackbufferTexture);
			}
		}
		else // use backbuffer directly
		{
			Asura_Direct3D_BackBufferCapture::CaptureCurrentBackBuffer();
			Asura_Direct3D::D3DDevice->SetTexture(2, Asura_Direct3D_BackBufferCapture::GetCaptureTexture());
		}
		Asura_Render::AllowFullScreenRendering();

		Asura_Render::CurrentStates.RequestTextureIndex( 1, iHeatHazeTexID );
		
		Asura_Direct3D_Shaders::RequestVertexShader(ASURA_D3D_VS_PFX_HEATHAZE);
		Asura_Direct3D_Shaders::RequestPixelShader(ASURA_D3D_PS_PFX_HEATHAZE);

		Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);

		//If recording an AVI there may be a different RNG to use. RAR 23/05/2006
		float fRandom;
		#ifdef ASURA_TOOLS
		if(g_pxHeatHazeRNG)
		{
			fRandom = g_pxHeatHazeRNG->GetFloat(-0.05f, 0.05f);
		}
		else
		#endif
		{
			fRandom = Asura_Random::GetFloat(-0.05f, 0.05f);
		}
		
		// VSH output position is in homogeneous clipping space (-1.0...+1.0)
		float afVSConstants[8] = { 0.5f, -1.0f / 50.0f, -0.5f, 1.0f,
									  1.0f, 1.0f, fRandom, g_fHEAT_HAZE_OFFSET_V, };
		Asura_Direct3D::D3DDevice->SetVertexShaderConstantF(4, afVSConstants, 2);

		g_fHEAT_HAZE_OFFSET_V += Asura_Timers::GetGameTimeSlice() * 0.75f;
		if(g_fHEAT_HAZE_OFFSET_V>1.0f)g_fHEAT_HAZE_OFFSET_V -= 1.0f;

		Asura_Direct3D::D3DDevice->SetTextureStageState(2, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);

		D3DXMATRIX xMatViewProj, xMatView, xMatProj;
		Asura_Direct3D::D3DDevice->GetTransform(D3DTS_VIEW, &xMatView);
		Asura_Direct3D::D3DDevice->GetTransform(D3DTS_PROJECTION, &xMatProj);
		D3DXMatrixMultiplyTranspose(&xMatViewProj, &xMatView, &xMatProj);
		Asura_Direct3D::D3DDevice->SetVertexShaderConstantF(0, xMatViewProj, 4);

		// Bump map matrix
		float fDistortionScale = 8.0f;
		Asura_Direct3D::D3DDevice->SetTextureStageState(2, D3DTSS_BUMPENVMAT00, FloatToDW(fDistortionScale / Asura_Direct3D::GetPresentParameters().BackBufferWidth));
		Asura_Direct3D::D3DDevice->SetTextureStageState(2, D3DTSS_BUMPENVMAT11, FloatToDW(fDistortionScale / Asura_Direct3D::GetPresentParameters().BackBufferHeight));

		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
		Asura_Direct3D::D3DDevice->SetSamplerState(1, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
		Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
		Asura_Direct3D::D3DDevice->SetSamplerState(2, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);
	}
}

void Asura_PFX_System::Platform_EndRender()
{
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	if(s_bIsRenderingHeatHazeSources)
	{
		Asura_Direct3D::D3DDevice->SetTexture(2, NULL);
		Asura_Direct3D_Shaders::RequestFixedFunctionVertexShader(D3DFVF_GENERICVERTEX);
		Asura_Direct3D_Shaders::RequestPixelShader(ASURA_D3D_PS_NONE);
		Asura_Render::CurrentStates.RequestTextureIndex(2, ASURA_TEXTURE_NONE);
	}
	Asura_Direct3D::D3DDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA );

	#if defined ASURA_TARGET_SM3
		if( GetCurrentRenderType() == ASURA_PFX_RENDER_SKYBOX )
		{
			// If we were rendering skybox particles, ensure we rebind all the deferred render targets.
			// Everything except the prelighting buffer will have been switched off by the fast path
			// renderer
			if( Asura_Direct3D::GetDeferredRenderingTarget() )
			{
				Asura_Direct3D::GetDeferredRenderingTarget()->Bind();
			}
		}
	#endif // defined ASURA_TARGET_SM3

	D3DPERFCALL(EndEvent());
}

bool Asura_PFX_System::Platform_SupportsHeatHaze()
{
	return true;
}




// DH ( 20:Feb:2006 )
// This change calls "DrawPrimitiveUP" for a batch of particles rather than one call per particle.
// This can improve particle performance by 20-40% if many particles are being rendered.
//
// Numerous buffer sizes were tested and it was found the a list just 16*6 ( 96) verts was the 
// most optimal, due to "DrawPrimitiveUP" internal workings.
//
static int const iParticleVertsNumOrig = 1;
static D3DGenericVertex xParticleVertsOrig[ iParticleVertsNumOrig*6 ];

void Asura_PFX_Effect::Platform_RenderParticles( const Asura_PFX_Particle* pxParticles,
                                                 u_int uNumParticles,
                                                 const Asura_PFX_Particle_Template* pxTemplate,
                                                 const Asura_RGBA& xLightColour 
#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
												 ,void* pPlatformRenderData
#endif
												 ) const
{
	// It is valid to pass in zero particles (source could just have started or ended. Simply return)
	if(!uNumParticles) return;


	Asura_Assert(pxParticles, "Null particle pointer given");
	Asura_Assert(pxTemplate, "Null particle template pointer given");
	if(!pxTemplate || !pxParticles) return;


#ifdef ASURA_USE_FASTPATH_PFX
	// Handle custom PFX effects, that have custom update functions so won't be handled
	// by the multithreaded update, but can be rendered using the fast path...

#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
	if (pPlatformRenderData == NULL)
	{			
		pPlatformRenderData = Asura_Platform_PFX_System_FastPath::GetInstanceBufferArea(uNumParticles);			

		if (pPlatformRenderData)
		{
			Asura_Platform_PFX_System_FastPath::UpdateInstanceBuffer(pxTemplate, pxParticles, (Asura_Platform_PFXInstanceData *)pPlatformRenderData, uNumParticles);				
		}
	}
#endif

	// Can we use the fast path renderer? 	
	// DMR - 02/08/07 - Skybox particles forced through the slow path, as they're rendered early, before we're finished with MRT rendering
	if (!pxTemplate->m_bUseMotionBlur &&
		!pxTemplate->m_b3DRotation &&
		!pxTemplate->m_bSpriteRotation &&
		!(m_usPFXFlags& ASURA_PFX_EFFECT_SKYBOX)
		&& pPlatformRenderData)		
	{
		Asura_Platform_PFX_System_FastPath::RenderParticles(this,
			pxParticles,
			uNumParticles,
			pxTemplate,
			xLightColour,
			m_usPFXFlags,
			m_ucCameraFlags,
			pPlatformRenderData);
		return;	
	}
#endif

	Asura_Render::CurrentStates.RequestTextureIndex( 0, pxTemplate->m_iTextureHandle );
	Asura_Render::CurrentStates.RequestTextureIndex( 1, pxTemplate->m_iTextureHandle );

	// (MH - 11.01.2006) instead of setting TextureOpMode in Platform_BeginRender, 
	// do it here based on template flags.
	if (!Asura_PFX_System::IsRenderingHeatHazeSources())
	{
		if (pxTemplate->m_bCrossfadeTextureFrames)
		{
			Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_LERP);
		}
		else
		{
			// the code used to multiply every colour value by 0.5f instead of using normal modulate.
			Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
		}
	}

	if(m_usPFXFlags & ASURA_PFX_EFFECT_ALWAYS_ON_TOP)
	{
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_ALWAYS);
	}
	else
	{
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
	}
	Asura_Render::CurrentStates.RequestTranslucencyMode(pxTemplate->m_eTranslucency);
	// Solid particles write to the z-buffer.
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(pxTemplate->m_eTranslucency == ASURA_TRANSLUCENCY_OFF);

	Asura_Vector_3 axVertices[4];

	for( int iCam=0; iCam<Asura_Camera::GetNumberOfCameras(); ++iCam )
	{
		// RWM - Array pointer being incremented inside inner loop, need to reset for each camera...
		const Asura_PFX_Particle* pxParticlesForCamera = pxParticles;

		if((m_ucCameraFlags & (1<<iCam)) != 0)
		{
			Asura_Camera::SetCameraIndex( iCam );

			if(Asura_Camera::IsCameraActive())
			{
				Asura_Render::ApplyCameraSettings();

				const	Asura_Vector_3*		pxCameraPos		= Asura_Camera::GetPosition();
				D3DGenericVertex*	pxParticleVerts;
				u_int				uParticlesLeft	= uNumParticles;
				u_int				uParticlesToRender;
				u_int				uParticleCount;

				while ( uParticlesLeft )
				{
					uParticleCount = uParticlesLeft;
					if ( uParticleCount > iParticleVertsNumOrig )
					{
						uParticleCount	= iParticleVertsNumOrig;
					}

					uParticlesToRender	= uParticleCount;
					pxParticleVerts		= &xParticleVertsOrig[0];

					for(u_int uParticle = 0 ; uParticle < uParticleCount ; ++uParticle)
					{
						const	Asura_PFX_Particle*	pxParticle = &pxParticlesForCamera[uParticle];
						Asura_Vector_3		xPosition	= pxParticle->m_xPosition;

						if ( m_usPFXFlags & ASURA_PFX_EFFECT_SKYBOX ) xPosition += *pxCameraPos;

						Asura_RGBA xColour = pxTemplate->m_xColour.GetValueAtTime_Discrete(pxParticle->m_fAge);

						// (SB - 01/05/07) Allow project to override particle colour
						if (Asura_PFX_System::GetParticleColourOverride() != NULL)
						{
							Asura_PFX_System::GetParticleColourOverride()(this, pxTemplate, uParticle, xColour);
						}

						if(pxTemplate->m_bUseLighting)
						{
							xColour.m_fR *= xLightColour.m_fR;
							xColour.m_fG *= xLightColour.m_fG;
							xColour.m_fB *= xLightColour.m_fB;
						}

						if(pxTemplate->m_bNearCameraFade || pxTemplate->m_bFarCameraFade)
						{
							float fDist = (*pxCameraPos - xPosition).Magnitude();
							//fade when near camera
							if(fDist < pxTemplate->m_fNearFadeDistance && pxTemplate->m_bNearCameraFade)
							{
								if( (pxTemplate->m_fNearFadeDistance - pxTemplate->m_fNearZeroDistance) == 0.0f )
								{
									xColour.m_fA = 0.f;
								}
								else
								{
									xColour.m_fA *= (fDist - pxTemplate->m_fNearZeroDistance) / (pxTemplate->m_fNearFadeDistance - pxTemplate->m_fNearZeroDistance);
								}						
							}
							//fade when far from camera
							if(fDist > pxTemplate->m_fFarFadeDistance && pxTemplate->m_bFarCameraFade)
							{
								if( (pxTemplate->m_fFarFadeDistance - pxTemplate->m_fFarZeroDistance) == 0.0f )
								{
									xColour.m_fA = 0.f;
								}
								else
								{
									xColour.m_fA *= (fDist - pxTemplate->m_fFarZeroDistance) / (pxTemplate->m_fFarFadeDistance - pxTemplate->m_fFarZeroDistance);
								}				
							}
						}
						// KOR - 15/11/2005 - early out on zero alpha
						if(xColour.m_fA <= 0.f)
						{
							--uParticlesToRender;
							continue;
						}

						float fSize = pxTemplate->m_xSize.GetValueAtTime_Discrete(pxParticle->m_fAge);
						if( pxTemplate->m_bUseSizeScale )
						{
							fSize *= GetSizeScale();
						}

						fSize *= (1.f - pxParticle->m_fSeed * pxTemplate->m_xSizeVariation.GetValueAtTime_Discrete(pxParticle->m_fAge));
						const float fAgeInSeconds = pxParticle->m_fAge * pxTemplate->GetLifetimeFromSeed(pxParticle->m_fSeed);
						const float fLength = Asura_Maths::Min(pxTemplate->m_fBlurLength, fAgeInSeconds);
						const Asura_Vector_3 xPrevPosition = xPosition - (pxParticle->m_xVelocity * fLength);

						Asura_Vector_3 xDeltaX;
						Asura_Vector_3 xDeltaY;

						float fTheta;

						if(pxTemplate->m_bFixedRotation)
						{
							fTheta = Asura_Maths::DegToRad(pxTemplate->m_fFixedRotation);
							if(pxTemplate->m_uNumFixedRotations > 1)
							{
								int iRotation = Asura_Maths::FloatToInt_Chop(pxParticle->m_fSeed * pxTemplate->m_uNumFixedRotations);
								fTheta += (iRotation * Asura_Maths::Pi * 2.f) / pxTemplate->m_uNumFixedRotations;
							}
						}
						else if(pxTemplate->m_b3DRotation || !pxTemplate->m_bUseMotionBlur)
						{
							// Need a rotation
							if(pxTemplate->m_bUseRandomRotation)
							{
								fTheta = ((Asura_Timers::GetGameTimeScale() == 0.0f) ? 0.0f : GetRNG()->GetFloat()) * Asura_Maths::Pi * 2.0f;
							}
							else
							{
#ifdef ASURA_TOOLS
								fTheta = (pxParticle->m_fSeed + pxTemplate->m_xRotation.GetAreaUnderGraph(pxParticle->m_fAge)) * Asura_Maths::Pi * 2.f;
#else
								fTheta = (pxParticle->m_fSeed + pxTemplate->m_xRotation.GetValueAtTime_Discrete(pxParticle->m_fAge)) * Asura_Maths::Pi * 2.f;
#endif
							}

							if(pxTemplate->m_bMirrorRotation)
							{
								const int* piRandom = reinterpret_cast<const int*>(&pxParticle->m_fSeed);
								if(*piRandom % 2)
								{
									fTheta = -fTheta;
								}
							}
						}

						const Asura_Matrix_3x3* pxOrientation = GetOrientation();
						if(pxTemplate->m_b3DRotation)
						{
							// KOR - 9/9/2004 - not optimal, and possibly not the way we want to do 3D rotation
							// (intended for effects such as glass shards or scattering debris). Need artist feedback to improve.
							const Asura_Euler_3 xEuler(fTheta, fTheta, 0.f);
							Asura_Matrix_3x3 xMatrix;
							xEuler.ToMatrix(xMatrix);
							xDeltaX = Asura_Vector_3(xMatrix.e11, xMatrix.e21, xMatrix.e31) * fSize;
							xDeltaY = Asura_Vector_3(xMatrix.e12, xMatrix.e22, xMatrix.e32) * fSize;
						}
						else if(pxTemplate->m_bAlignEffectX && pxOrientation)
						{
							xDeltaX = Asura_Vector_3(pxOrientation->e12, pxOrientation->e22, pxOrientation->e32) * fSize;
							xDeltaY = Asura_Vector_3(pxOrientation->e13, pxOrientation->e23, pxOrientation->e33) * fSize;
						}
						else if(pxTemplate->m_bAlignEffectY && pxOrientation)
						{
							xDeltaX = Asura_Vector_3(pxOrientation->e11, pxOrientation->e21, pxOrientation->e31) * fSize;
							xDeltaY = Asura_Vector_3(pxOrientation->e13, pxOrientation->e23, pxOrientation->e33) * fSize;
						}
						else if(pxTemplate->m_bAlignEffectZ && pxOrientation)
						{
							xDeltaX = Asura_Vector_3(pxOrientation->e11, pxOrientation->e21, pxOrientation->e31) * fSize;
							xDeltaY = Asura_Vector_3(pxOrientation->e12, pxOrientation->e22, pxOrientation->e32) * fSize;
						}
						else if(pxTemplate->m_bBillboardJustY && pxOrientation)
						{
							const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();
							xDeltaX = Asura_Vector_3(pxCameraOrientation->e11,pxCameraOrientation->e21,pxCameraOrientation->e31) * fSize;
							xDeltaY = Asura_Vector_3(pxOrientation->e13, pxOrientation->e23, pxOrientation->e33) * fSize;
						}
						else if(pxTemplate->m_bBillboardJustX && pxOrientation)
						{
							const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();
							xDeltaX = Asura_Vector_3(pxOrientation->e12, pxOrientation->e22, pxOrientation->e32) * fSize;
							xDeltaY = Asura_Vector_3(pxCameraOrientation->e12,pxCameraOrientation->e22,pxCameraOrientation->e32) * fSize;
						}
						else
						{
							const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();
							xDeltaX = Asura_Vector_3(pxCameraOrientation->e11,pxCameraOrientation->e21,pxCameraOrientation->e31) * fSize;
							xDeltaY = Asura_Vector_3(pxCameraOrientation->e12,pxCameraOrientation->e22,pxCameraOrientation->e32) * fSize;
						}

						// JGP - moved the getting of frames to here


						// get the texture frame(s) for this particle
						const Asura_PFX_TextureFrame* pxTextureFrame;
						const Asura_PFX_TextureFrame* pxTextureFrame2;
						if (pxTemplate->m_bCrossfadeTextureFrames)
						{
							float fWeight;
							pxTemplate->GetTextureFramesForCrossfadeAtTime(pxParticle->m_fAge, pxTextureFrame, pxTextureFrame2, fWeight);
							// weight needs to be replicated across RGBA
							u_int uWeight = Asura_Maths::FloatToInt(fWeight * 255.0f);
							uWeight = (uWeight << 24) | (uWeight << 16) | (uWeight << 8) | (uWeight << 0);
							// set the amount of blending between texture frames for this particle
							Asura_Direct3D::D3DDevice->SetRenderState(D3DRS_TEXTUREFACTOR, uWeight);
						}
						else
						{
							pxTextureFrame = pxTemplate->GetTextureFrameAtTime(pxParticle->m_fAge);
							pxTextureFrame2 = NULL;
						}

						// JGP aspect ratio from UV

						if ( pxTemplate->m_bUseUVAsAspect )
						{

							const float fDU = pxTextureFrame->m_afTextureUVs[ 2 ] - pxTextureFrame->m_afTextureUVs[ 0 ];
							const float fDV = pxTextureFrame->m_afTextureUVs[ 3 ] - pxTextureFrame->m_afTextureUVs[ 1 ];
							const float fAspect = ( fDV ) ? fDU / fDV : 1.0f;
							xDeltaX *= fAspect;
						}

						Asura_Assert( Asura_Maths::Abs( xPosition.x ) < ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD,
							"Asura_PFX_Effect::Platform_RenderParticles - possible particle position problem" );
						Asura_Assert( Asura_Maths::Abs( xPosition.y ) < ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD,
							"Asura_PFX_Effect::Platform_RenderParticles - possible particle position problem" );
						Asura_Assert( Asura_Maths::Abs( xPosition.z ) < ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD,
							"Asura_PFX_Effect::Platform_RenderParticles - possible particle position problem" );

						/// Kevin Lea - 27/03/2002 11:43:15 - code to draw particle stretched along its path
						/// ucky and needs rewriting
						if (pxTemplate->m_bUseMotionBlur)
						{
							Asura_Vector_3 xBeta;

							xBeta = xPosition - xPrevPosition;
							if( xBeta.MagnitudeSquared() > (0.1f*0.1f) )
							{
								xBeta.Normalise();
								xBeta*=0.1f;
							}

							Asura_Assert( Asura_Maths::Abs( xPrevPosition.x ) < ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD,
								"Asura_PFX_Effect::Platform_RenderParticles - possible particle previous position problem" );
							Asura_Assert( Asura_Maths::Abs( xPrevPosition.y ) < ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD,
								"Asura_PFX_Effect::Platform_RenderParticles - possible particle previous position problem\n" );
							Asura_Assert( Asura_Maths::Abs( xPrevPosition.z ) < ASURA_PFX_SYSTEM_SUSPICIOUS_POSITION_THRESHOLD,
								"Asura_PFX_Effect::Platform_RenderParticles - possible particle previous position problem\n" );

							if (Asura_Maths::Abs(xBeta * xDeltaX) > Asura_Maths::Abs(xBeta*xDeltaY))
							{
								if (xBeta * xDeltaX > 0)
								{
									axVertices[0] = xPrevPosition	- xBeta - xDeltaX - xDeltaY; 
									axVertices[1] = xPosition		+ xBeta + xDeltaX - xDeltaY;
									axVertices[2] = xPosition		+ xBeta + xDeltaX + xDeltaY;
									axVertices[3] = xPrevPosition	- xBeta - xDeltaX + xDeltaY;
								}
								else
								{
									axVertices[0] = xPosition		+ xBeta - xDeltaX - xDeltaY; 
									axVertices[1] = xPrevPosition	- xBeta + xDeltaX - xDeltaY;
									axVertices[2] = xPrevPosition	- xBeta + xDeltaX + xDeltaY;
									axVertices[3] = xPosition		+ xBeta - xDeltaX + xDeltaY;
								}
							}
							else
							{
								if (xBeta * xDeltaY > 0)
								{
									axVertices[0] = xPrevPosition	- xBeta - xDeltaX - xDeltaY; 
									axVertices[1] = xPrevPosition	- xBeta + xDeltaX - xDeltaY;
									axVertices[2] = xPosition		+ xBeta + xDeltaX + xDeltaY;
									axVertices[3] = xPosition		+ xBeta - xDeltaX + xDeltaY;
								}
								else
								{
									axVertices[0] = xPosition		+ xBeta - xDeltaX - xDeltaY; 
									axVertices[1] = xPosition		+ xBeta + xDeltaX - xDeltaY;
									axVertices[2] = xPrevPosition	- xBeta + xDeltaX + xDeltaY;
									axVertices[3] = xPrevPosition	- xBeta - xDeltaX + xDeltaY;
								}
							}
						}
						else 
						{
							if (fTheta!=0.f)
							{
								const float fSin = Asura_Maths::Sine(fTheta);
								const float fCos = Asura_Maths::Cosine(fTheta);

								Asura_Vector_3 xCosDeltaX = xDeltaX * fCos;
								Asura_Vector_3 xSinDeltaX = xDeltaX * fSin;
								Asura_Vector_3 xCosDeltaY = xDeltaY * fCos;
								Asura_Vector_3 xSinDeltaY = xDeltaY * fSin;

								axVertices[0] = xPosition + (-xCosDeltaX + xSinDeltaX) + (-xSinDeltaY - xCosDeltaY);		
								axVertices[1] = xPosition + ( xCosDeltaX + xSinDeltaX) + ( xSinDeltaY - xCosDeltaY);
								axVertices[2] = xPosition + ( xCosDeltaX - xSinDeltaX) + ( xSinDeltaY + xCosDeltaY);
								axVertices[3] = xPosition + (-xCosDeltaX - xSinDeltaX) + (-xSinDeltaY + xCosDeltaY);
							}
							else
							{
								axVertices[0] = xPosition - xDeltaX - xDeltaY; 
								axVertices[1] = xPosition + xDeltaX - xDeltaY;
								axVertices[2] = xPosition + xDeltaX + xDeltaY;
								axVertices[3] = xPosition - xDeltaX + xDeltaY;
							}
						}


						{		
							// are we changing texture orient?
							u_int uTexOrient = 0;
							if(pxTemplate->m_bRandomTextureOrient)
							{
								uTexOrient = Asura_Maths::FloatToInt_Chop(pxParticle->m_fSeed * 4.f);
							}

							float fVOffset = 0.f;
							if(pxTemplate->m_bSpriteRotation)
							{
								Asura_Vector_3 xParticleDirection = pxParticle->m_xVelocity;
								Asura_Vector_3 xCameraDirection = pxParticle->m_xPosition - *pxCameraPos;
								// Flatten it out to the X-Z plane.
								xParticleDirection.y = 0.f;
								xCameraDirection.y = 0.f;
								if(xParticleDirection.Normalise() > 0.f && xCameraDirection.Normalise() > 0.f)
								{
									const float fParticleYAngle = Asura_Maths::ArcTangent(xParticleDirection.x, xParticleDirection.z);
									const float fCameraYAngle = Asura_Maths::ArcTangent(xCameraDirection.x, xCameraDirection.z);
									const float fAngleDiff = Asura_Maths::RadiansToZeroMinimumRange(fParticleYAngle + Asura_Maths::DegToRad(22.5f) - fCameraYAngle);
									const u_int uFrameOffset = Asura_Maths::FloatToInt_Near(fAngleDiff / Asura_Maths::DegToRad(45.f));
									fVOffset = uFrameOffset * pxTextureFrame->m_afTextureUVs[3] - pxTextureFrame->m_afTextureUVs[1];
								}
							}

							const float fFrame0U0 = pxTextureFrame->m_afTextureUVs[0];
							const float fFrame0U1 = pxTextureFrame->m_afTextureUVs[2];
							const float fFrame1U0 = pxTextureFrame2 ? (pxTextureFrame2->m_afTextureUVs[0]) : 0.f;
							const float fFrame1U1 = pxTextureFrame2 ? (pxTextureFrame2->m_afTextureUVs[2]) : 0.f;
							switch(uTexOrient)
							{
							case 0:
							case 1:
								// Normal
								pxParticleVerts[0].U = pxParticleVerts[3].U	= pxParticleVerts[5].U = fFrame0U0;
								pxParticleVerts[1].U = pxParticleVerts[2].U = pxParticleVerts[4].U = fFrame0U1;
								if (pxTextureFrame2)
								{
									pxParticleVerts[0].U2 = pxParticleVerts[3].U2 = pxParticleVerts[5].U2 = fFrame1U0;
									pxParticleVerts[1].U2 = pxParticleVerts[2].U2 = pxParticleVerts[4].U2 = fFrame1U1;
								}
								break;
							case 2:
							case 3:
							default:
								// Mirrored
								pxParticleVerts[0].U = pxParticleVerts[3].U	= pxParticleVerts[5].U = fFrame0U1;
								pxParticleVerts[1].U = pxParticleVerts[2].U = pxParticleVerts[4].U = fFrame0U0;
								if (pxTextureFrame2)
								{
									pxParticleVerts[0].U2 = pxParticleVerts[3].U2 = pxParticleVerts[5].U2 = fFrame1U1;
									pxParticleVerts[1].U2 = pxParticleVerts[2].U2 = pxParticleVerts[4].U2 = fFrame1U0;
								}
								break;
							}
							const float fFrame0V0 = pxTextureFrame->m_afTextureUVs[1] + fVOffset;
							const float fFrame0V1 = pxTextureFrame->m_afTextureUVs[3] + fVOffset;
							const float fFrame1V0 = pxTextureFrame2 ? (pxTextureFrame2->m_afTextureUVs[1] + fVOffset) : 0.f;
							const float fFrame1V1 = pxTextureFrame2 ? (pxTextureFrame2->m_afTextureUVs[3] + fVOffset) : 0.f;
							switch(uTexOrient)
							{
							case 0:
							case 3:
								// Normal
								pxParticleVerts[0].V = pxParticleVerts[1].V = pxParticleVerts[3].V = fFrame0V0;
								pxParticleVerts[2].V = pxParticleVerts[4].V = pxParticleVerts[5].V = fFrame0V1;
								if (pxTextureFrame2)
								{
									pxParticleVerts[0].V2 = pxParticleVerts[1].V2 = pxParticleVerts[3].V2 = fFrame1V0;
									pxParticleVerts[2].V2 = pxParticleVerts[4].V2 = pxParticleVerts[5].V2 = fFrame1V1;
								}
								break;
							case 1:
							case 2:
							default:
								// Mirrored
								pxParticleVerts[0].V = pxParticleVerts[1].V = pxParticleVerts[3].V = fFrame0V1;
								pxParticleVerts[2].V = pxParticleVerts[4].V = pxParticleVerts[5].V = fFrame0V0;
								if (pxTextureFrame2)
								{
									pxParticleVerts[0].V2 = pxParticleVerts[1].V2 = pxParticleVerts[3].V2 = fFrame1V1;
									pxParticleVerts[2].V2 = pxParticleVerts[4].V2 = pxParticleVerts[5].V2 = fFrame1V0;
								}
								break;
							}

							const unsigned int diffuse = xColour.ToUIntARGB(true);

							// 1st Tri
							pxParticleVerts[0].X			= axVertices[0].x;
							pxParticleVerts[0].Y			= axVertices[0].y;
							pxParticleVerts[0].Z			= axVertices[0].z;
							pxParticleVerts[1].X			= axVertices[1].x;
							pxParticleVerts[1].Y			= axVertices[1].y;
							pxParticleVerts[1].Z			= axVertices[1].z;
							pxParticleVerts[2].X			= axVertices[2].x;
							pxParticleVerts[2].Y			= axVertices[2].y;
							pxParticleVerts[2].Z			= axVertices[2].z;
							// 2nd Tri
							pxParticleVerts[3].X			= axVertices[0].x;
							pxParticleVerts[3].Y			= axVertices[0].y;
							pxParticleVerts[3].Z			= axVertices[0].z;
							pxParticleVerts[4].X			= axVertices[2].x;
							pxParticleVerts[4].Y			= axVertices[2].y;
							pxParticleVerts[4].Z			= axVertices[2].z;
							pxParticleVerts[5].X			= axVertices[3].x;
							pxParticleVerts[5].Y			= axVertices[3].y;
							pxParticleVerts[5].Z			= axVertices[3].z;

							pxParticleVerts[0].Diffuse	= pxParticleVerts[1].Diffuse = pxParticleVerts[2].Diffuse	=
								pxParticleVerts[3].Diffuse	= pxParticleVerts[4].Diffuse = pxParticleVerts[5].Diffuse	= diffuse;

							pxParticleVerts += 6;
						}
					}

					if ( uParticlesToRender )
					{
#ifdef ASURA_TOOLS
						Tools_Mode_PFX::AddToParticlesRendered(uParticlesToRender);
#endif

						// (SB - 18/05/07) Setup render states for particle effect
						if (Asura_PFX_System::GetRenderStateOverride() != NULL)
						{
							Asura_PFX_System::GetRenderStateOverride()();
						}

						Asura_Direct3D::D3DDevice->DrawPrimitiveUP(D3DPT_TRIANGLELIST, uParticlesToRender*2, (LPVOID)xParticleVertsOrig, sizeof(D3DGenericVertex) );
						Asura_Render::AddToTrianglesRendered( uParticlesToRender*2 );
					}
					pxParticlesForCamera	+= uParticleCount;
					uParticlesLeft			-= uParticleCount;
				}
			}
		}
	}
}

#ifdef ASURA_TOOLS
void Asura_PFX_System::SetHeatHazeRNG(Asura_RandomGenerator* pxRNG)
{
	g_pxHeatHazeRNG = pxRNG;
	//reset the offset at the same time , since this needs to start at a known value for avi recording to be consistent.
	//RAR 23/05/2006
	g_fHEAT_HAZE_OFFSET_V = 0.0f;
}

#endif



#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_BeginUpdate
// Author       : Dave Reed
// Date         : 13-Nov-2006
//
// Description  : Prepares vertex buffers for updating during Update.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void Asura_PFX_System::Platform_BeginUpdate (void)
{
#ifdef ASURA_USE_FASTPATH_PFX
	Asura_Platform_PFX_System_FastPath::UpdateInstanceBuffersForNewFrame();
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_GetRenderData
// Author       : Dave Reed
// Date         : 13-Nov-2006
//
// Description  : returns area of vertex buffer big enough for given number of 
//                particles - for multithreaded updates.
//
//			      Returns NULL if we're not doing this via the job queue 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void* Asura_PFX_System::Platform_GetRenderData(u_int uNumParticles)
{
#ifdef ASURA_USE_FASTPATH_PFX
	// Todo - Don't call this with slow-path effects - this is wasting buffer space at the moment!
	if (Asura_Platform_PFX_System_FastPath::IsEnabled())
	{	
		return Asura_Platform_PFX_System_FastPath::GetInstanceBufferArea(uNumParticles);
	}
#endif	

	return NULL;
}

#endif // ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

#endif // !defined(ASURA_TARGET_SM3)



/// End of file. ///
