//*******************************************************************
//
// File: Asura_PS2_Botanicals_Store.h
//
// Created: 22 March 2006 by Jake Turner
//
// Description: 
// PS2 specific botanicals storage class
// Manager class for PS2 botanicals data
//
//*******************************************************************

#include "Core/Asura.h"

#if defined( ASURA_USE_BOTANICALS )

#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
#include "Interface/Asura_CommandConsole.h"
#endif
#include "Light/Asura_Photon.h"
#include "Mesh/Asura_CollisionMesh.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Render.h"

// Platform includes.
#include "Botanicals/Asura_PS2_Botanicals_Store.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS2_StrippedObject.h"
#include "Core/Asura_PS2_DMA.h"
#include "Debugging/Asura_DebugText.h"
#include "Light/Asura_GI_Solution.h"
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_PS2_VU1.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_BotanicalsDefines.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"

#ifdef _DEBUG
#include "Core/Asura_PlatformBasics.h"
#endif
// See  "PS2/Botanicals/Asura_PS2_Botanicals_Shared.h" for details of the file format
// General.
#define g_uMeshFlags         (*reinterpret_cast<u_int*       >(0x70000000))
#define g_uPrevMeshFlags     (*reinterpret_cast<u_int*       >(0x70000004)) // Mesh flags from previous SendPendingData().
#define g_uPendingFlags      (*reinterpret_cast<u_int*       >(0x70000008))

// Camera.
#define g_uCamVisFlags       (*reinterpret_cast<u_int*       >(0x70000010))
#define g_uCamFlagsForVU     (*reinterpret_cast<u_int*       >(0x70000014))
#define g_uNumCamsVisFor     (*reinterpret_cast<u_int*       >(0x70000018))
#define g_bO2WIsIdentity     (*reinterpret_cast<bool*        >(0x7000001C))
#define g_uCurVUCamBfr       (*reinterpret_cast<u_int*       >(0x70000020))

// Texture.
#define g_bTexUploaded       (*reinterpret_cast<bool*        >(0x70000024))
#define g_iCurBaseTex        (*reinterpret_cast<int*         >(0x70000028))

// Renderstate
#define g_uCurVURSBfr        (*reinterpret_cast<u_int*       >(0x70000214))
#define g_uNumPassesPerTri   (*reinterpret_cast<u_int*       >(0x70000218))
#define g_uCurMaterialFlags  (*reinterpret_cast<u_int*       >(0x7000021C))
#define g_pxRSBfr            ( reinterpret_cast<Asura_PS2_QW*>(0x70000220)) // Up to 5 QWs of renderstate.
#define g_pxSTOvParams       ( reinterpret_cast<Asura_PS2_QW*>(0x700002C0))

// Matrices.
#define g_pxO2W              ( reinterpret_cast<Asura_PS2_QW*>(0x700002F0)) // 4x4 matrix.
#define g_fObjectScale       (*reinterpret_cast<float*>(0x70000330)) // float
#define g_pfAlphaScales      ( reinterpret_cast<float*>(0x70000334)) // float[4]

// Lighting.
#define g_pxOutLightBfr      ( reinterpret_cast<Asura_PS2_QW*>(0x700003B0)) // 2 QWs per light for up to 7 lights 
#define g_pvIntSHBfr         ( reinterpret_cast<void*        >(0x70000490)) // Space for one intermediate 2nd order SHs, for intermediate calculations.
#define g_uNumLightQWs       (*reinterpret_cast<u_int*       >(0x700004D0))
#define g_uCurVULightBfr     (*reinterpret_cast<u_int*       >(0x700004D4))
// next available address 0x70000500

// Total.
#define TOTAL_SPR_QWS_REQD                                    (0x0000050)

// Special case for light count. Only required for point lighting.
#define g_uNumLights ((g_pxOutLightBfr + ASURA_PS2_BOTANICALS__OFFSET__QW_CONTAINING_QTY)->u32s[3])

#define MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT(puExtern, uNumMSCALs) (++(uNumMSCALs), ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(puExtern))

static inline void FillQWWithNOPs( Asura_PS2_QW* pxOut )
{
	pxOut->u128 = 0;
}

// T Y P E S /////////////////////////////////////////////////////////////////

enum PENDING
{
	PENDING_UPLOAD_TEX_SYNC          =  1 <<  0,
	PENDING_UPLOAD_RENDERSTATE       =  1 <<  1,
	PENDING_UPLOAD_CAM               =  1 <<  2,
	PENDING_UPLOAD_LIGHT             =  1 <<  3,
	PENDING_UPLOAD_MESH_FLAGS        =  1 <<  4,
	PENDING_UPLOAD_CAM_FLAGS         =  1 <<  5,
};


////////////////////////////////////////////////////////////////////////////
//
// Global variables
//
////////////////////////////////////////////////////////////////////////////

struct Asura_PS2_Botanicals_Component_Data
{
	Asura_PS2_Botanicals_QW xRETtag;
	Asura_PS2_Botanicals_QW xDMAdata[1];
	//Asura_PS2_Botanicals_QW xDMAdata[nQwords];
};

struct Asura_PS2_Botanicals_Instance_Runtime_Data
{
	u_short usDistCam[ASURA_MAX_NO_OF_CAMERAS];	// Distance to camera 3 for the render instance
	u_int uVisAndClipFlags;
#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
	Asura_SphericalHarmonics_Order2_RGB m_xDynamicLight;
#endif
};

#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
DEBUGVAR float dbg_fBotanicalBillboardLightingContrastFactor = 1.f;
#endif
Asura_PS2_Botanicals_Header Asura_PS2_Botanicals_Store::s_xHeader;
const Asura_PS2_Botanicals_Type_Header** Asura_PS2_Botanicals_Store::s_ppxTypeHeaders = NULL;
const Asura_PS2_Botanicals_Type_Info** Asura_PS2_Botanicals_Store::s_ppxTypeInfos = NULL;
const Asura_PS2_Botanicals_Instance** Asura_PS2_Botanicals_Store::s_ppxInstancePtrs = NULL;
Asura_PS2_Botanicals_Instance_Runtime_Data* Asura_PS2_Botanicals_Store::s_pxInstanceRuntimeData = NULL;
Asura_PS2_Botanicals_Component_Data** Asura_PS2_Botanicals_Store::s_ppxComponentDatas = NULL;
Asura_PS2_QW* Asura_PS2_Botanicals_Store::s_pxDataPool = NULL;
Asura_RenderChainNode** Asura_PS2_Botanicals_Store::s_pxTails = NULL;
u_short *Asura_PS2_Botanicals_Store::s_pusComponentTriCounts = NULL;

extern u_int VU1_BOTANICALS_INIT[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_BEGIN_STRIP[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_BEGIN_STRIP_BREAK[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_SWITCH_RENDERSTATE_BFR[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_SWITCH_CAM_DATA_BFR[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_SWITCH_LIGHT_DATA_BFR[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_LOAD_MESH_FLAGS[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_LOAD_CAM_FLAGS[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_BEGIN_STRIP_BREAK[] __attribute__(( section(".vudata") ));
extern u_int VU1_BOTANICALS_NOP[] __attribute__(( section(".vudata") ));

static Asura_Vector_3 s_axCameraPos[ASURA_MAX_NO_OF_CAMERAS];
static Asura_Vector_3 s_axCameraFwdVector[ASURA_MAX_NO_OF_CAMERAS];
static Asura_Vector_3 s_axCameraHorizVector[ASURA_MAX_NO_OF_CAMERAS];
static Asura_Vector_3 s_axCameraVertVector[ASURA_MAX_NO_OF_CAMERAS];
static float s_afCameraHorizScale[ASURA_MAX_NO_OF_CAMERAS];
static float s_afCameraVertScale[ASURA_MAX_NO_OF_CAMERAS];
static float s_afCameraTanHalfFovX[ASURA_MAX_NO_OF_CAMERAS];
static float s_afCameraYOverXRatio[ASURA_MAX_NO_OF_CAMERAS];
static float s_afHorizFoVScale[ASURA_MAX_NO_OF_CAMERAS];

// For culling 
DEBUGVAR float dbg_fBotanicalsViewHorizScale = 1.0f;
DEBUGVAR float dbg_fBotanicalsViewVertScale = 1.0f;

#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
const Asura_UnicodeChar* s_wszEnableLightingConsoleVarName = L"Render.Botanicals.EnableLighting";
DVSVAR bool cnsl_bLightingEnabled = true;
#endif

static u_int s_uMaxNumInstancesToDrawPerType = 9999;
static u_int s_uMaxNumTypesToDraw = 9999;

static u_int s_uNumInstancesDrawn = 0;
static int s_iTypeToDraw = -1;

#define MAX_NUM_RENDER_NODES 4096
static Asura_RenderChainNode axRenderNodes[MAX_NUM_RENDER_NODES];

#define DEBUG_PS2_BOTANICALS

#if defined( DEBUG_PS2_BOTANICALS )

extern bool bCarpetBotanicalsDebugTex;
extern bool bCarpetBotanicalsDebugTexReal;

DEBUGVAR int dbg_iComponentTypeToDraw = -1;
DEBUGVAR int dbg_iComponentToDraw = -1;
DEBUGVAR bool dbg_bBotanicalsFillSPRInBegin = false;
DEBUGVAR bool dbg_bBotanicalsFillVU1MemInBegin = false;
static bool bBotanicalsVU1Debug = false;

#endif // #if defined( DEBUG_PS2_BOTANICALS )

// Used to invalidate the material flags and force an update
static u_int uINVALID_MATERIAL_FLAG = AsuraMaterialFlag_SubdividePolygon;

void Asura_PS2_Botanicals_Store::Initialize( void )
{
	s_xHeader.uNumTypes = 0;
	s_xHeader.uNumTotalInstances = 0;
	s_xHeader.uNumTotalComponents = 0;
	s_ppxTypeHeaders = NULL;
	s_ppxTypeInfos = NULL;
	s_ppxComponentDatas = NULL;
	s_ppxInstancePtrs = NULL;
	s_pxDataPool = NULL;
	s_pxTails = NULL;
	s_pxInstanceRuntimeData = NULL;
	s_pusComponentTriCounts = NULL;

	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Header ) == 5*4 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Type_Header ) == 8 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Module_Header ) == 4 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Component_Info ) == 8*4 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Instance ) == 26*4 );

#if defined( ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING ) && defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	Asura_CommandConsole::AddVar(s_wszEnableLightingConsoleVarName, &cnsl_bLightingEnabled, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif
}

void Asura_PS2_Botanicals_Store::Shutdown( void )
{
	Reset();
}

void Asura_PS2_Botanicals_Store::Reset( void )
{
	// Deallocate memory
	// Need to loop over the component data array free'ing memory by hand
	for ( u_int uI = 0; uI < s_xHeader.uNumTotalComponents; uI++ )
	{
		Asura_DeallocateAlign( s_ppxComponentDatas[uI] );
		s_ppxComponentDatas[uI] = NULL;
	}

	delete[] s_ppxTypeHeaders;
	delete[] s_ppxTypeInfos;
	delete[] s_ppxComponentDatas;
	delete[] s_ppxInstancePtrs;
	delete[] s_pxDataPool;
	delete[] s_pxTails;
	delete[] s_pxInstanceRuntimeData;
	delete[] s_pusComponentTriCounts;

	s_xHeader.uNumTypes = 0;
	s_xHeader.uNumTotalInstances = 0;
	s_xHeader.uNumTotalComponents = 0;
	s_ppxTypeHeaders = NULL;
	s_ppxTypeInfos = NULL;
	s_ppxComponentDatas = NULL;
	s_ppxInstancePtrs = NULL;
	s_pxDataPool = NULL;
	s_pxTails = NULL;
	s_pxInstanceRuntimeData = NULL;
	s_pusComponentTriCounts = NULL;
}

inline void Asura_PS2_Botanicals_Store::SetObjectToWorldMat( const Asura_Matrix_3x3& xOri, const Asura_Vector_3& xPos, const float fObjectScale )
{
	Asura_PS2_VU1::SetupPS2MatFromAsuraData( &g_pxO2W->f32s[0], xOri, xPos );
	g_bO2WIsIdentity = false;
	g_fObjectScale = fObjectScale;
	g_uPendingFlags |= PENDING_UPLOAD_CAM;
}

void Asura_PS2_Botanicals_Store::SetLights( const u_int uSH0, const u_int uSH1, const u_int uSH2, const u_int uSH3 DYNAMIC_LIGHT_PARAM IGNORE_NORMAL_PARAM  )
{
    u_int uNewPendingFlags = g_uPendingFlags;
	Asura_PS2_QW* const pxLightBfr = g_pxOutLightBfr;

	// On export the SH terms are already rotated into model space
	// For reference on export this is how the int SH co-efficients are stored
	// 0 co-efficents of RGB
/*
	xInstance.uSH0 = ( ( Asura_Maths::FloatToInt_Near( xSH.m_afRedCoefficients[0] ) & 0xFF ) << 0 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afGreenCoefficients[0] ) & 0xFF ) << 8 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afBlueCoefficients[0] ) & 0xFF ) << 16 );

	// 1st co-efficents of RGB
	xInstance.uSH1 = ( ( Asura_Maths::FloatToInt_Near( xSH.m_afRedCoefficients[1] ) & 0xFF ) << 0 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afGreenCoefficients[1] ) & 0xFF ) << 8 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afBlueCoefficients[1] ) & 0xFF ) << 16 );

	// 2nd co-efficents of RGB
	xInstance.uSH2 = ( ( Asura_Maths::FloatToInt_Near( xSH.m_afRedCoefficients[2] ) & 0xFF ) << 0 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afGreenCoefficients[2] ) & 0xFF ) << 8 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afBlueCoefficients[2] ) & 0xFF ) << 16 );

	// 3rd co-efficents of RGB
	xInstance.uSH3 = ( ( Asura_Maths::FloatToInt_Near( xSH.m_afRedCoefficients[3] ) & 0xFF ) << 0 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afGreenCoefficients[3] ) & 0xFF ) << 8 ) |
		( ( Asura_Maths::FloatToInt_Near( xSH.m_afBlueCoefficients[3] ) & 0xFF ) << 16 );
*/

	// The SH terms are stored with 64 as the fixed point scale and 0 = 127
	// Convert the int representation to floating point values
	const float fPS2Scale = 1.0f / 64.0f;
	// 0.0f is 127 so need to offset the value back again
	const float fPS2Offset = 127.0f;
	Asura_SphericalHarmonics_Order2_RGB* pxSrcSH = static_cast<Asura_SphericalHarmonics_Order2_RGB*>( g_pvIntSHBfr );
	// Convert to signed int first, to avoid slow casts.
	const int iSHR0 = ( uSH0 >>  0 ) & 0xFF;
	const int iSHG0 = ( uSH0 >>  8 ) & 0xFF;
	const int iSHB0 = ( uSH0 >> 16 ) & 0xFF;
	pxSrcSH->m_afRedCoefficients[0]		= ( static_cast<float>( iSHR0 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afGreenCoefficients[0]	= ( static_cast<float>( iSHG0 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afBlueCoefficients[0]	= ( static_cast<float>( iSHB0 ) - fPS2Offset ) * fPS2Scale;

	const int iSHR1 = ( uSH1 >>  0 ) & 0xFF;
	const int iSHG1 = ( uSH1 >>  8 ) & 0xFF;
	const int iSHB1 = ( uSH1 >> 16 ) & 0xFF;
	pxSrcSH->m_afRedCoefficients[1]		= ( static_cast<float>( iSHR1 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afGreenCoefficients[1]	= ( static_cast<float>( iSHG1 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afBlueCoefficients[1]	= ( static_cast<float>( iSHB1 ) - fPS2Offset ) * fPS2Scale;

	const int iSHR2 = ( uSH2 >>  0 ) & 0xFF;
	const int iSHG2 = ( uSH2 >>  8 ) & 0xFF;
	const int iSHB2 = ( uSH2 >> 16 ) & 0xFF;
	pxSrcSH->m_afRedCoefficients[2]		= ( static_cast<float>( iSHR2 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afGreenCoefficients[2]	= ( static_cast<float>( iSHG2 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afBlueCoefficients[2]	= ( static_cast<float>( iSHB2 ) - fPS2Offset ) * fPS2Scale;

	const int iSHR3 = ( uSH3 >>  0 ) & 0xFF;
	const int iSHG3 = ( uSH3 >>  8 ) & 0xFF;
	const int iSHB3 = ( uSH3 >> 16 ) & 0xFF;
	pxSrcSH->m_afRedCoefficients[3]		= ( static_cast<float>( iSHR3 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afGreenCoefficients[3]	= ( static_cast<float>( iSHG3 ) - fPS2Offset ) * fPS2Scale;
	pxSrcSH->m_afBlueCoefficients[3]	= ( static_cast<float>( iSHB3 ) - fPS2Offset ) * fPS2Scale;

#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
	if(pxDynamicSH)
	{
		*pxSrcSH += *pxDynamicSH;
	}
#endif

#ifdef ASURA_PS2_BOTANICAL_BILLBOARDS_IGNORE_NORMAL_FOR_LIGHTING
	if(bIgnoreNormal)
	{
		Asura_RGBA xColour;
		pxSrcSH->EvaluateInDirection(Asura_Vector_3(0.f, -1.f, 0.f), xColour);
		pxSrcSH->Zero();
		pxSrcSH->m_afRedCoefficients[0] = xColour.m_fR;
		pxSrcSH->m_afGreenCoefficients[0] = xColour.m_fG;
		pxSrcSH->m_afBlueCoefficients[0] = xColour.m_fB;
	}
#endif
	// Setup VU SH matrix.
	{
		Asura_PS2_QW* pxOut = pxLightBfr;

		// We need to scale and transpose the coefficients from independent
		// code as we copy them.
		for( u_int u = 0; u < 4; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				pxOut->f32s[v] = 255.0f * pxSrcSH->m_afCoefficient[(v << 2) + u];
			}
			++pxOut;
		}
	}
    // VU (specifically the LightVtxWithSH macro) requires w component of
    // second row to be 0.
    (pxLightBfr + 1)->f32s[3] = 0.0f;

	pxLightBfr[0].f32s[3] = 128.0f;

    uNewPendingFlags |= PENDING_UPLOAD_LIGHT;

    g_uNumLightQWs  = 4;
    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_Botanicals_Store::Render( const float fDefaultHorizFoV, const float fGlobalFade )
{
	#define VIEW_DISTANCE_SCALE_FACTOR 32768.0f
	#define PS2_BOTANICALS_NOT_IN_VIEW 65535

	if( Asura_PS2_RenderingOptions::GetOption( Asura_PS2_RenderingOptions::OPTION_DRAW_BOTANICALS ) == false )
	{
		return;
	}

	if ( s_xHeader.uNumTotalInstances == 0 )
	{
		return;
	}
	const u_int uNumTypes = ( s_xHeader.uNumTypes > s_uMaxNumTypesToDraw ) ? s_uMaxNumTypesToDraw : s_xHeader.uNumTypes;
	if ( uNumTypes == 0 )
	{
		return;
	}
	ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__BOTANICALS_MARKER );

	const u_int uNumCams = Asura_Camera::GetNumberOfCameras();
	for ( u_int uI = 0; uI < uNumCams; uI++ )
	{
		Asura_Camera::SetCameraIndex( uI );
		const Asura_Matrix_3x3* const xView = Asura_Camera::GetOrientation();
		s_axCameraHorizVector[uI].x = xView->e11;
		s_axCameraHorizVector[uI].y = xView->e21;
		s_axCameraHorizVector[uI].z = xView->e31;
		s_axCameraVertVector[uI].x = xView->e12;
		s_axCameraVertVector[uI].y = xView->e22;
		s_axCameraVertVector[uI].z = xView->e32;
		s_axCameraFwdVector[uI].x = xView->e13;
		s_axCameraFwdVector[uI].y = xView->e23;
		s_axCameraFwdVector[uI].z = xView->e33;

		s_axCameraPos[uI] = *(Asura_Camera::GetPosition());

		const Asura_CameraData& xCamData = Asura_Camera::GetCameraData( uI );
		s_afCameraHorizScale[uI] = xCamData.m_fOneOverCosHalfFovX;
		s_afCameraVertScale[uI] = xCamData.m_fOneOverCosHalfFovY;
		s_afCameraTanHalfFovX[uI] = xCamData.m_fTanHalfFovX;
		s_afCameraYOverXRatio[uI] = xCamData.m_fYOverXAspectRatio;

		// Narrower FoV's mean increase the distance of when to change the LoD's
		const float fCameraHorizFoV = ( xCamData.m_xHalfFieldOfViewX * 2.0f );
		s_afHorizFoVScale[uI] =  fCameraHorizFoV / fDefaultHorizFoV;
	}

	s_uNumInstancesDrawn = 0;

#if defined( DEBUG_PS2_BOTANICALS )
#if defined( ASURA_USE_CARPET )
	bCarpetBotanicalsDebugTex = bCarpetBotanicalsDebugTexReal;
	if ( bCarpetBotanicalsDebugTex ) 
	{
		Asura_OutputToDebugger_Retail( ( "**** BOTANICALS START ****\n" ) );
	}
#endif // #if defined( ASURA_USE_CARPET )
#endif // #if defined( DEBUG_PS2_BOTANICALS )
	 // To stop overrunning the render chain node array
	u_int uNumRenderNodes = 0;

	// Take the tail sentinel out of the pool of render chain nodes
	// Set this tail sentinel for all the render chains
	Asura_RenderChainNode& xTailSentinel = axRenderNodes[0];
	xTailSentinel.usUserData = 0xFFFF;
	xTailSentinel.usNextOffset = 0;
	const u_int uMaxNumComponents = s_xHeader.uNumTotalComponents;
	for ( u_int uI = 0; uI < uMaxNumComponents; uI++ )
	{
		s_pxTails[uI] = &xTailSentinel;
	}
	uNumRenderNodes++;

	// The starting point for putting data into the render chain
	Asura_RenderChainNode* pxThis = & (axRenderNodes[uNumRenderNodes] );

	Begin();

	ASURA_PS2_TEXBFRLAYOUT eLayout = ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL;
	Asura_PS2_TextureManagement::SetBufferLayout( eLayout, true, 0, 1 );

	// Draw the instances
	u_int uMinType = 0;
	u_int uMaxType = uNumTypes;
	if ( s_iTypeToDraw != -1 )
	{
		uMinType = s_iTypeToDraw;
		uMaxType = uMinType + 1;
		uMinType = Asura_Maths::Min( uMinType, uNumTypes );
		uMaxType = Asura_Maths::Min( uMaxType, uNumTypes );
	}

	for ( u_int uI = uMinType; uI < uMaxType; uI++ )
	{
		const Asura_PS2_Botanicals_Type_Info* const pxTypeInfo = s_ppxTypeInfos[uI];
		const Asura_PS2_Botanicals_Type_Header* const pxTypeHeader = s_ppxTypeHeaders[uI];

		const u_char* pcMemory = reinterpret_cast<const u_char*>(pxTypeHeader);
		pcMemory += sizeof( Asura_PS2_Botanicals_Type_Header );

		u_int uNumInstancesPerType = 0;
		const u_int uNumModulesInType = pxTypeHeader->usNumModules;
		for ( u_int uMod = 0; uMod < uNumModulesInType; uMod++ )
		{
			const Asura_PS2_Botanicals_Module_Header* const pxModuleHeader = reinterpret_cast<const Asura_PS2_Botanicals_Module_Header* const>(pcMemory);
			pcMemory += sizeof( Asura_PS2_Botanicals_Module_Header );

			const u_int uModuleIndex = pxModuleHeader->usModuleID;
			const u_int uNumInstancesInModule = pxModuleHeader->usNumInstances;
			const u_int uLinkedFlags = Asura_Module_System::GetModuleLinkedFlags( uModuleIndex );
			if ( uLinkedFlags != 0 )
			{
#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
				const Asura_CollisionMesh* pxCMesh = Asura_Module_System::GetCollisionMeshForModule( uModuleIndex );
				Asura_Bounding_Box xBB = pxCMesh->GetBoundingBox();
				xBB.Displace( Asura_Module_System::GetPositionOfModule(uModuleIndex) );
				const Asura_Vector_3& xModuleCentre = Asura_Module_System::GetCentreOfModule(uModuleIndex);
				Asura_Photon_System::MakePhotonListForBoundingBox( xBB, xModuleCentre, false, true );
#endif
				Asura_PS2_Botanicals_Instance* const pxInstances = const_cast<Asura_PS2_Botanicals_Instance* const>( reinterpret_cast<const Asura_PS2_Botanicals_Instance* const>(pcMemory) );
				for ( u_int uInstance = 0; uInstance < uNumInstancesInModule; uInstance++ )
				{
					if ( uNumInstancesPerType >= s_uMaxNumInstancesToDrawPerType )
					{
						break;
					}
					u_short usDistCam[ASURA_MAX_NO_OF_CAMERAS];

					Asura_PS2_Botanicals_Instance* const pxInstance = pxInstances + uInstance;
					const Asura_Vector_3 xPos( pxInstance->fSphereBoundX, pxInstance->fSphereBoundY, pxInstance->fSphereBoundZ );

					bool bVisible = false;
					// It should be only between 0-1
					float fMinUnitlessViewDist = 2.0f;
					float fMaxUnitlessViewDist = 0.0f;
					u_int uClipFlags = 0;
					u_int uClipBit = 1 << 6;
					for ( u_int uI = 0; uI < uNumCams; uI++ )
					{
						Asura_Camera::SetCameraIndex( uI );
						// Sentinel value for not in view
						usDistCam[uI] = PS2_BOTANICALS_NOT_IN_VIEW;

						// Ignore testing cameras which aren't linked to this module
						if ( ( uLinkedFlags & ( 1 << uI ) ) == 0 )
						{
							continue;
						}

						// Setup camera data for this camera
						const Asura_Vector_3 xCameraPos = s_axCameraPos[uI];
						const Asura_Vector_3 xCameraFwdVector = s_axCameraFwdVector[uI];
						const Asura_Vector_3 xCameraHorizVector = s_axCameraHorizVector[uI];
						const Asura_Vector_3 xCameraVertVector = s_axCameraVertVector[uI];

						const float fCameraHorizScale = dbg_fBotanicalsViewHorizScale * s_afCameraHorizScale[uI];
						const float fCameraVertScale = dbg_fBotanicalsViewVertScale * s_afCameraVertScale[uI];
						const float fCameraTanHalfFovX = dbg_fBotanicalsViewHorizScale * s_afCameraTanHalfFovX[uI];
						const float fCameraYOverXRatio = dbg_fBotanicalsViewVertScale * s_afCameraYOverXRatio[uI];

						// Culling checks for this camera
						const Asura_Vector_3 xPosInCameraSpace = xPos - xCameraPos;

						// Do quick ignore based on forward distance
						const float fFwdDist = xPosInCameraSpace * xCameraFwdVector;
						const float fUnitlessDistance = ( fFwdDist * pxTypeInfo->fInvMaxDrawDistance * s_afHorizFoVScale[uI] );
						if( fUnitlessDistance > 1.0f )
						{
							// Outside Maximum draw distance
							continue;
						}

						// Do forward culling check
						// See http://www.lighthouse3d.com/opengl/viewfrustum/index.php?camspace3
						const float fRadius = pxInstance->fSphereBoundR;
						// Use the bounding sphere radius
						if ( fFwdDist < -fRadius )
						{
							// Ignore things behind the camera
							continue;
						}

						const float fHorizR = fRadius * fCameraHorizScale;
						const float fHorizZ = fFwdDist * fCameraTanHalfFovX;
						const float fHorizRplusZ = fHorizR + fHorizZ;

						// Do horizontal culling check
						const float fHorizDist = Asura_Maths::Abs( xPosInCameraSpace * xCameraHorizVector );
						if ( fHorizDist > fHorizRplusZ )
						{
							// Ignore things outside camera horizontal FoV
							continue;
						}

						const float fVertR = fRadius * fCameraVertScale;
						const float fVertZ = fHorizZ * fCameraYOverXRatio;
						const float fVertRplusZ = fVertR + fVertZ;
	
						// Do vertical culling check
						const float fVertDist = Asura_Maths::Abs( xPosInCameraSpace * xCameraVertVector );
						if ( fVertDist > fVertRplusZ )
						{
							// Ignore things outside camera vertical FoV
							continue;
						}

						if(!Asura_PS2_VU1::IsSphereEntirelyWithinGuardbandFrustum(xPos, pxInstance->fSphereBoundR))
						{
							uClipFlags |= uClipBit;
						}

						// Use the bounding sphere centre distance for LoD selection (this matches the editor and PC renderer)
						// Clamp it to be >=0.001f
						const float fUnitlessViewDist = Asura_Maths::Max( 0.001f, fUnitlessDistance );

						// Convert fUnitlessViewDist into 16-bit unsigned range for storing
						usDistCam[uI] = Asura_Maths::FloatToInt( fUnitlessViewDist * VIEW_DISTANCE_SCALE_FACTOR );
						bVisible = true;
						fMinUnitlessViewDist = Asura_Maths::Min( fMinUnitlessViewDist, fUnitlessViewDist );
						fMaxUnitlessViewDist = Asura_Maths::Max( fMaxUnitlessViewDist, fUnitlessViewDist );
					}

					// Ignore instances which aren't visible to any camera
					if ( bVisible == false )
					{
						continue;
					}

					const u_short usInstanceIndex = pxInstance->usGlobalInstanceIndex;
					Asura_PS2_Botanicals_Instance_Runtime_Data* const  pxInstanceRuntimeData = s_pxInstanceRuntimeData + usInstanceIndex;
					for ( u_int uI = 0; uI < uNumCams; uI++ )
					{
						pxInstanceRuntimeData->usDistCam[uI] = usDistCam[uI];
					}
					pxInstanceRuntimeData->uVisAndClipFlags = uLinkedFlags | (uClipFlags << ASURA_MAX_NO_OF_CAMERAS);

#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
					const Asura_Vector_3 xPosition(pxInstance->fX, pxInstance->fY, pxInstance->fZ);
					Asura_SphericalHarmonics_Order2_RGB& xDynamicLight = pxInstanceRuntimeData->m_xDynamicLight;
					xDynamicLight.Zero();
					const u_int uNumPhotons = Asura_Photon_System::NumberOfLocalPhotons;
					for(u_int uPhoton = 0; uPhoton < uNumPhotons; ++uPhoton)
					{
						const Asura_Photon* pxPhoton = Asura_Photon_System::LocalPhotons[uPhoton].GetPhoton();
						const float fRange = pxPhoton->GetRange();
						Asura_Vector_3 xDir = pxPhoton->GetPosition() - xPosition;
						const float fMag = xDir.Normalise();

						if (fMag < fRange)
						{
							const float fBrightness = 1.0f - fMag / fRange;

							const float fR = fBrightness * pxPhoton->GetRedComponent();
							const float fG = fBrightness * pxPhoton->GetGreenComponent();
							const float fB = fBrightness * pxPhoton->GetBlueComponent();

							xDynamicLight.AddSampleForDirection( fR, fG, fB, xDir );		
						}
					}
					xDynamicLight.ScaleCoefficientsForIrradiance();
#endif
					// Choose which LoD components to render (add them to the render list)
					const u_int uNumComponents = pxTypeInfo->uNumComponents;
					for ( u_int uComp = 0; uComp < uNumComponents; uComp++ )
					{
						const Asura_PS2_Botanicals_Component_Info& xComponentInfo = pxTypeInfo->xComponentInfo[uComp];
						const float fFarDistance = xComponentInfo.fLoDMax;
						if ( fMinUnitlessViewDist >= fFarDistance )
						{
							// Ignore if beyond the LoD far distance
							continue;
						}
						const float fNearDistance = xComponentInfo.fLoDMin;
						if ( fMaxUnitlessViewDist <= fNearDistance )
						{
							// Ignore if closer than the LoD near distance
							continue;
						}

						// Stop overrunning the array - could be an assert
						if ( uNumRenderNodes >= MAX_NUM_RENDER_NODES )
						{
							Asura_Assert( false, "Too many trees" );
							continue;
						}

						// Lower 24-bits is index
						const u_int uComponentIndex = ( xComponentInfo.uComponentIndexAndType & 0x00FFFFFF );
						Asura_Assert( uComponentIndex < s_xHeader.uNumTotalComponents, "" );
						// So rendering into chain usChain with node data instance uInst
						pxThis->usUserData = usInstanceIndex;

						// Update the tail of this list
						Asura_RenderChainNode** const ppxTail = &( s_pxTails[uComponentIndex] );
						Asura_RenderChainNode* const pxPrev = *ppxTail;

						// Set offset ptr (assumed to be -ve from this place to the previous tail)
						pxThis->usNextOffset = pxThis - pxPrev;

						// Set a new tail for this component
						*ppxTail = pxThis;

						// Another item in the list
						uNumRenderNodes++;
						pxThis++;
					}
					uNumInstancesPerType++;
					s_uNumInstancesDrawn++;
				}
			}
			pcMemory += sizeof( Asura_PS2_Botanicals_Instance ) * uNumInstancesInModule;
		}
	}

	// Loop over the types rendering the chains
	// Do it this way because the types are sorted by texture so as to reduce texture swapping
	u_int uNumTris = 0;
	for ( u_int uI = 0; uI < uNumTypes; uI++ )
	{
		const Asura_PS2_Botanicals_Type_Info* const pxTypeInfo = s_ppxTypeInfos[uI];

		// Don't like how have to loop over the components in the type then get the chain component index
		const u_int uNumComponents = pxTypeInfo->uNumComponents;
		for ( u_int uComp = 0; uComp < uNumComponents; uComp++ )
		{
			const Asura_PS2_Botanicals_Component_Info& xComponentInfo = pxTypeInfo->xComponentInfo[uComp];
			const u_int uComponentIndex = ( xComponentInfo.uComponentIndexAndType & 0x00FFFFFF );

			Asura_RenderChainNode* pxNode = s_pxTails[uComponentIndex];
			// Is there anything in this render chain?
			if ( pxNode->usNextOffset == 0 )
			{
				continue;
			}

			// Debug only draw a certain type of component
#if defined( DEBUG_PS2_BOTANICALS )
			if ( ( dbg_iComponentToDraw != -1 ) && ( uComp != static_cast<u_int>( dbg_iComponentToDraw ) ) )
			{
				continue;
			}

			const u_int uType = xComponentInfo.usType;
			if ( ( dbg_iComponentTypeToDraw != -1 ) && ( uType != static_cast<u_int>( dbg_iComponentTypeToDraw ) ) )
			{
				continue;
			}
#endif //  #if defined( DEBUG_PS2_BOTANICALS )

			const Asura_PS2_Botanicals_Component_Data* const pxComponentData = s_ppxComponentDatas[uComponentIndex];
			const u_int uCALLAddr = reinterpret_cast<u_int>( pxComponentData );

			const float fFarDistance = xComponentInfo.fLoDMax;
			const float fNearDistance = xComponentInfo.fLoDMin;
			const float fLoDConstant1 = xComponentInfo.fLoDConstant1;
			const float fLoDConstant2 = xComponentInfo.fLoDConstant2;
			const u_int uNumTrisPerComponent = s_pusComponentTriCounts[uComponentIndex];
			//const u_int uNumTrisPerComponent = xComponentInfo.usNumTris;

			// Set the texture for this type
			const int iTexID = xComponentInfo.iTextureID;
			if ( iTexID == ASURA_MATERIAL_NONE )
			{
				continue;
			}

			const int iMMAG = SCE_GS_LINEAR;
			const int iMMIN = SCE_GS_LINEAR_MIPMAP_NEAREST; // bilinear
			//const int iMMIN = SCE_GS_LINEAR_MIPMAP_LINEAR; // trilinear
			SetTexture( iTexID, iMMAG, iMMIN, Asura_PS2_TextureManagement::BFR_AUTO );

			const u_int uMaterialFlags = xComponentInfo.uMaterialFlags;
			SetRenderState( uMaterialFlags );

			const bool bBillboard = ( uType == ASURA_PS2_BOTANICALS_BILLBOARDS );
			// Set the mesh flags based on the geometry type
			// Need to get this flag in the component info and then just copy it in!
			if ( bBillboard )
			{
				g_uMeshFlags |= ASURA_PS2_BOTANICALS__FLAG__BILLBOARD;
			}
			else
			{
				g_uMeshFlags &= ~ASURA_PS2_BOTANICALS__FLAG__BILLBOARD;
			}

			// If the component has valid data in it
			while ( pxNode->usNextOffset )
			{
				const u_int uInstanceIndex = pxNode->usUserData;
				Asura_PS2_Botanicals_Instance_Runtime_Data* const pxInstanceRuntimeData = s_pxInstanceRuntimeData + uInstanceIndex;

				bool bVisible = false;
				for ( u_int uI = 0; uI < uNumCams; uI++ )
				{
					g_pfAlphaScales[uI] = 0.0f;
					const u_short usCamDist = pxInstanceRuntimeData->usDistCam[uI];
					const float fUnitlessViewDist = static_cast<float>( usCamDist ) * ( 1.0f / VIEW_DISTANCE_SCALE_FACTOR );

					// So rendering this component add it to the render chain list
					// Need to compute its alpha fade value
	
					// 4 LoD values: 0, 1, 2, 3
					//   |
					//   |       +------------+
					//   |      /|            |\
					//   |     / |            | \
					//   |    /  |            |  \
					//   +---+--------------------+------------
					//       0   1            2   3
					//   0.0 -----------------------> 1.0
					//
					// 0 = fNearDistance
					// 1 = fFadeNear
					// 2 = fFadeFar
					// 3 = fFarDistance
					//
					// first slope = ( x - fNearDistance ) / ( fFadeNear - fNearDistance ) 
					// last slope = ( fFarDistance - x ) / ( fFarDistance - fFadeFar )

					// Near fade value
					float fFade0 = ( fUnitlessViewDist - fNearDistance ) * fLoDConstant1;
					fFade0 = Asura_Maths::Min( fFade0, 1.0f );
					fFade0 = Asura_Maths::Max( fFade0, 0.0f );
					if ( fFade0 == 0.0f )
					{
						continue;
					}
	
					// Far fade value
					float fFade1 = ( fFarDistance - fUnitlessViewDist ) * fLoDConstant2;
					fFade1 = Asura_Maths::Min( fFade1, 1.0f );
					fFade1 = Asura_Maths::Max( fFade1, 0.0f );
					if ( fFade1 == 0.0f )
					{
						continue;
					}

					const float fFade = fGlobalFade * Asura_Maths::Min( fFade0, fFade1 );
					if ( fFade == 0.0f )
					{
						continue;
					}
					Asura_Assert( fFade > 0.0f, "" );

					// The input alpha image is 0 - transparent 128->255 for solid but with noise on it however
					// on the PS2 the alpha range is 64->128 (because alpha values are halved when making PS2 textures)
					// Formula is (remember on PS2 alpha scale operation is ( alpha_texture * scale_value / 128 )
					//			cut_value = 128.0f - 64.0f * fFade;
					//          scale_value = 128.0f * ( 64.0f / cut_value );
					//			scale_value = 128.0f / ( 2.0f - fFade ) );
					// But because of integer maths used it has been found that the following formula gives better results:
					//			scale_value = ( 128.0f / ( 2.0f - fFade ) ) + fFudge;
					// With fFudge = 0.914f gives the least average square error (0.51%)
					// 
					// So: 
					//   fFade = 0.0 (0% visible) scale = 64 = 0.5
					//   fFade = 1.0 (100% visible) scale = 128 = 1.0
					float fAlphaScale = ( 128.0f / ( 2.0f - fFade ) ) + 0.914f;
					g_pfAlphaScales[uI] = fAlphaScale;
					bVisible = true;
				}

				if ( bVisible == false )
				{
					// The offset index is a -ve offset by definition
					pxNode = pxNode - pxNode->usNextOffset;
					continue;
				}

				// Set the camera and clip flags for this mesh
				const u_int uVisAndClipFlags = pxInstanceRuntimeData->uVisAndClipFlags;
                const u_int uVisFlags  = uVisAndClipFlags & ((1 << ASURA_MAX_NO_OF_CAMERAS) - 1);
                const u_int uClipFlags = uVisAndClipFlags >> ASURA_MAX_NO_OF_CAMERAS;
				SetCamFlags( uVisFlags, uClipFlags );

				const Asura_PS2_Botanicals_Instance* const pxInstance = s_ppxInstancePtrs[uInstanceIndex];
				const Asura_Vector_3 xPos( pxInstance->fX, pxInstance->fY, pxInstance->fZ );
				SetObjectToWorldMat( pxInstance->xLocal2World, xPos, pxInstance->fScaleXYZ );
#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
				Asura_SphericalHarmonics_Order2_RGB xDynamicSH;
				bool bNonZeroDynamicSH = false;
				if(cnsl_bLightingEnabled)
				{
					xDynamicSH = pxInstanceRuntimeData->m_xDynamicLight;
					for ( u_int uI = 0; uI < Asura_SphericalHarmonics_Order2_RGB::NUMBER_OF_COEFFICIENTS; ++uI )
					{
						if(xDynamicSH.m_afCoefficient[uI] != 0.f)
						{
							bNonZeroDynamicSH = true;
							break;
						}
					}
				}
#endif
				// For billboards the o2w transform is done in the billboarding code
				if ( bBillboard )
				{
					g_bO2WIsIdentity = true;
					// For billboards use the world SH
#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
					if(bNonZeroDynamicSH)
					{
#ifdef ASURA_PS2_BOTANICAL_BILLBOARDS_IGNORE_NORMAL_FOR_LIGHTING
						SetLights( pxInstance->uWorldSH0, pxInstance->uWorldSH1, pxInstance->uWorldSH2, pxInstance->uWorldSH3, &xDynamicSH, true );
#else
						SetLights( pxInstance->uWorldSH0, pxInstance->uWorldSH1, pxInstance->uWorldSH2, pxInstance->uWorldSH3, &xDynamicSH );
#endif
					}
					else
#endif
					{
#ifdef ASURA_PS2_BOTANICAL_BILLBOARDS_IGNORE_NORMAL_FOR_LIGHTING
						SetLights( pxInstance->uWorldSH0, pxInstance->uWorldSH1, pxInstance->uWorldSH2, pxInstance->uWorldSH3, true );
#else
						SetLights( pxInstance->uWorldSH0, pxInstance->uWorldSH1, pxInstance->uWorldSH2, pxInstance->uWorldSH3 );
#endif
					}
				}
				else
				{
					g_bO2WIsIdentity = false;
					// For 3D geometry use the object SH which has been rotated into object space
#ifdef ASURA_PS2_SUPPORT_BOTANICAL_LIGHTING
					if(bNonZeroDynamicSH)
					{
						xDynamicSH.RotateToModelspace(pxInstance->xLocal2World);
						SetLights( pxInstance->uObjectSH0, pxInstance->uObjectSH1, pxInstance->uObjectSH2, pxInstance->uObjectSH3, &xDynamicSH );
					}
					else
#endif
					{
						SetLights( pxInstance->uObjectSH0, pxInstance->uObjectSH1, pxInstance->uObjectSH2, pxInstance->uObjectSH3 );
					}
				}

				// Do the DMA call
				SendPendingData( true );

				// Render the Mesh
				Asura_PS2_DmaTag* pxCALLTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
				pxCALLTag->SetWithNOPs( 0,
										Asura_PS2_DmaTag::PCE_NOCHANGE,
										Asura_PS2_DmaTag::ID_CALL,
										false,
										uCALLAddr,
										false );

#if defined( DEBUG_PS2_BOTANICALS )
				// If want to do VU1 debugging then change the MSCNT to a MSCAL( VU1_BOTANICALS_BEGIN_STRIP_BREAK );
				const u_int uNumQWs = pxComponentData->xRETtag.u16s[0];
				Asura_PS2_Botanicals_QW* const pxMSCNT = const_cast<Asura_PS2_Botanicals_QW* const>( pxComponentData->xDMAdata + ( uNumQWs-1 ) );
				const u_int uMSCNT = SCE_VIF1_SET_MSCNT( false );
				const u_int uMSCAL = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_BOTANICALS_BEGIN_STRIP_BREAK );
				u_int iMSCNTplace = 0;
				if ( ( pxMSCNT->u32s[iMSCNTplace] != uMSCNT ) && ( pxMSCNT->u32s[iMSCNTplace] != uMSCAL ) ) iMSCNTplace++;
				if ( ( pxMSCNT->u32s[iMSCNTplace] != uMSCNT ) && ( pxMSCNT->u32s[iMSCNTplace] != uMSCAL ) ) iMSCNTplace++;
				if ( ( pxMSCNT->u32s[iMSCNTplace] != uMSCNT ) && ( pxMSCNT->u32s[iMSCNTplace] != uMSCAL ) ) iMSCNTplace++;
				Asura_Assert( ( pxMSCNT->u32s[iMSCNTplace] == uMSCNT ) || ( pxMSCNT->u32s[iMSCNTplace] == uMSCAL ), "" );
				if ( bBotanicalsVU1Debug == true )
				{
					pxMSCNT->u32s[iMSCNTplace] = uMSCAL;
				}
				else
				{
					pxMSCNT->u32s[iMSCNTplace] = uMSCNT;
				}
#endif // #if defined( DEBUG_PS2_BOTANICALS )
				// TODO: doesn't include extra rendering because of split-screen
				uNumTris += uNumTrisPerComponent;

				// The offset index is a -ve offset by definition
				pxNode = pxNode - pxNode->usNextOffset;
			}
		}
	}

	End();
	Asura_PS2_Render::UpdateTriStats( uNumTris, 1, Asura_PS2_Render::TRICOUNT_BOTANICALS );
#if defined( DEBUG_PS2_BOTANICALS )
#if defined( ASURA_USE_CARPET )
	if ( bCarpetBotanicalsDebugTex ) 
	{
		Asura_OutputToDebugger_Retail( ( "**** BOTANICALS END ****\n" ) );
		bCarpetBotanicalsDebugTex = true;
	}
#endif // #if defined( ASURA_USE_CARPET )
#endif // #if defined( DEBUG_PS2_BOTANICALS )

	ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__BOTANICALS_MARKER );

	#undef VIEW_DISTANCE_SCALE_FACTOR 
	#undef PS2_BOTANICALS_NOT_IN_VIEW
}

bool Asura_PS2_Botanicals_Store::Load( Asura_Handle xFile )
{
	// Remove memory
	Reset();

	/////////////////////////////////////////////////////////////////////////////
	//
	// The PS2 Botanicals data format Version 6 is:
	//
	// Asura_PS2_Botanicals_Header
	// {
	//    u_int uNumTypes;
	//	  u_int uNumTotalComponents;
	//    u_int uNumTotalInstances;
	//    u_int uTotalHeadersSizeInBytes;
	//	  u_int uTotalTypeInfosSizeInBytes;
	//	  u_int uTotalTypeDmaDataSizeInBytes;
	// };
	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Header
	//    {
	//       u_short usTypeID;
	//   	 u_short usNumModules;
	//       u_int uNumInstances;
	//    };
	//    Array of usNumModules
	//    {
	//       struct Asura_PS2_Botanicals_Module_Header
	//       {
	//       	u_short usModuleID;
	//       	u_short usNumInstances;
	//       };
	//	  Array of usNumInstances;
	//    {
	//	     struct Asura_PS2_Botanicals_Instance
	//		 {
	//			u_short usGlobalInstanceIndex;
	//			u_short usPadding;
	//
	//			float fSphereBoundX;
	//			float fSphereBoundY;
	//			float fSphereBoundZ;
	//			float fSphereBoundR;
	//
	//			float fScaleXYZ;
	//
	//		    float fX;
	//			float fY;
	//			float fZ;
	//
	//			float fQuatX;
	//			float fQuatY;
	//			float fQuatZ;
	//			float fQuatW;
	//
	//		    Asura_Matrix_3x3 xLocal2World;
	//
	//			u_int uWorldSH0;
	//			u_int uWorldSH1;
	//			u_int uWorldSH2;
	//			u_int uWorldSH3;
	//
	//			u_int uSH0;
	//			u_int uSH1;
	//			u_int uSH2;
	//			u_int uSH3;
	//		 };
	//    }
	// }
	//
	// This data is separate to the Asura_PS2_Botanicals_Type_Header data because
	// one day this type info data might be passed to VU1 for LoD selection
	//
	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Info
	//    {
	//       Asura_Bounding_Box xBB;
//			 float fInvMaxDrawDistance;			
	// 	     u_int uNumComponents;
	//		 Array of uNumComponents;
	//       {
	//          struct Asura_PS2_Botanicals_Component_Info
	//          {
	//			   u_int uMaterialID;					// On export it is a material ID which is converted to a texture ID & material flags
	//			   u_int uMaterialFlags;
	//			   u_short usNumTris;					
	//			   u_short usType;					    // enum Asura_PS2_Botanicals_Mesh_Type
	//			   float fLoDMin;
	//			   float fLoDMax;
	//			   float fLoDConstant1;
	//			   float fLoDConstant2;
	//			   u_int uComponentIndexAndType; // u_int : top 8-bits is type, bottom 24-bits is component index into component DMA data
	//			};
	//       }
	//    };
	// }
	//
	// Array of uNumTotalComponents
	// {
	//    DMA data for the mesh component: referenced by a DMA CALL tag so the DMA data includes the DMA RET tag
	// }
	//
	/////////////////////////////////////////////////////////////////////////////

	Asura_FileAccess::ReadData( xFile, &s_xHeader, sizeof( Asura_PS2_Botanicals_Header ) );
	if ( s_xHeader.uNumTypes == 0 )
	{
		return true;
	}

	u_int uDataPoolSize = s_xHeader.uTotalHeadersSizeInBytes + s_xHeader.uTotalTypeInfosSizeInBytes;
	s_pxDataPool = new Asura_PS2_QW[ ( uDataPoolSize + 15 ) / 16 ];
	Asura_Assert( s_pxDataPool, "" );

	// Read the header and info data and type info data in a single read
	Asura_FileAccess::ReadData( xFile, s_pxDataPool, uDataPoolSize );

	s_ppxTypeHeaders = new (const Asura_PS2_Botanicals_Type_Header*)[s_xHeader.uNumTypes];
	Asura_Assert( s_ppxTypeHeaders, "" );

	s_ppxInstancePtrs = new (const Asura_PS2_Botanicals_Instance*)[s_xHeader.uNumTotalInstances];
	Asura_Assert( s_ppxInstancePtrs, "" );

	s_pxInstanceRuntimeData = new Asura_PS2_Botanicals_Instance_Runtime_Data[s_xHeader.uNumTotalInstances];
	Asura_Assert( s_pxInstanceRuntimeData, "" );

	// Need to patch up the style headers ptr array and the instance ptrs array
	const u_char* pcMemory = reinterpret_cast<u_char*>( s_pxDataPool );
	const u_char* pcTypeHeadersStart = pcMemory;

	s_pusComponentTriCounts = new u_short[s_xHeader.uNumTotalComponents];// < ASURA_PS2_SPR_SIZE_BYTES/4, "Too many components to fit on temp scratch-pad" );

	u_int uInstanceIndex = 0;
	for ( u_int uI = 0; uI < s_xHeader.uNumTypes; uI++ )
	{
		const Asura_PS2_Botanicals_Type_Header* pxTypeHeader = reinterpret_cast<const Asura_PS2_Botanicals_Type_Header*>( pcMemory );
		pcMemory += sizeof( Asura_PS2_Botanicals_Type_Header );

		s_ppxTypeHeaders[uI] = pxTypeHeader;

		const u_int uNumModulesInType = pxTypeHeader->usNumModules;
		for ( u_int uMod = 0; uMod < uNumModulesInType; uMod++ )
		{
			const Asura_PS2_Botanicals_Module_Header* const pxModuleHeader = reinterpret_cast<const Asura_PS2_Botanicals_Module_Header* const>( pcMemory );
			const u_int uNumInstancesInModule = pxModuleHeader->usNumInstances;
			pcMemory += sizeof( Asura_PS2_Botanicals_Module_Header );

			const Asura_PS2_Botanicals_Instance* const pxInstances = reinterpret_cast<const Asura_PS2_Botanicals_Instance* const>( pcMemory );
			for ( u_int uInstance = 0; uInstance < uNumInstancesInModule; uInstance++ )
			{
				Asura_PS2_Botanicals_Instance* const pxInstance = const_cast<Asura_PS2_Botanicals_Instance* const>( pxInstances + uInstance );
				s_ppxInstancePtrs[uInstanceIndex] = pxInstance;
				Asura_Assert( pxInstances[uInstance].usGlobalInstanceIndex == uInstanceIndex, "" );
				uInstanceIndex++;
			}
			pcMemory += sizeof( Asura_PS2_Botanicals_Instance ) * uNumInstancesInModule;
		}
	}
	Asura_Assert( ( pcMemory - pcTypeHeadersStart ) == static_cast<int>( s_xHeader.uTotalHeadersSizeInBytes ), "" );

	// The Type Infos
	s_ppxTypeInfos = new (const Asura_PS2_Botanicals_Type_Info*)[s_xHeader.uNumTypes];
	Asura_Assert( s_ppxTypeInfos, "" );

	//Use scratch-pad memory for array of pointers
//	Asura_PS2_Botanicals_Component_Info** const ppxComponentInfos = reinterpret_cast<Asura_PS2_Botanicals_Component_Info**>( 0x70000000 );
//	Asura_Assert( s_xHeader.uNumTotalComponents < ASURA_PS2_SPR_SIZE_BYTES/4, "Too many components to fit on scratch-pad" );

	
	const u_char* pcTypeInfosStart = pcMemory;
	for ( u_int uI = 0; uI < s_xHeader.uNumTypes; uI++ )
	{
		const Asura_PS2_Botanicals_Type_Info* pxTypeInfo = reinterpret_cast<const Asura_PS2_Botanicals_Type_Info*>( pcMemory );
		s_ppxTypeInfos[uI] = pxTypeInfo;
		// Skip over the memory
		// Asura_Bounding_Box xBB;
		pcMemory += sizeof( Asura_Bounding_Box );
		// float fInvMaxDrawDist;
		pcMemory += sizeof(float);
		// u_int uNumComponents;
		pcMemory += sizeof( u_int );
		// Remap the material ID's per component info
		const u_int uNumComponents = pxTypeInfo->uNumComponents;
		for ( u_int uComp = 0; uComp < uNumComponents; uComp++ )
		{
			Asura_PS2_Botanicals_Component_Info* const pxComponentInfo = const_cast<Asura_PS2_Botanicals_Component_Info*>( pxTypeInfo->xComponentInfo + uComp );

			const int iInputMaterialID = pxComponentInfo->iTextureID;
			// Remap the materialID
			const int iNewMaterialID = Asura_Material_System::ConvertIndex( iInputMaterialID );

			// Check for invalid textures
			pxComponentInfo->iTextureID = Asura_Material_System::GetTextureID( iNewMaterialID );
			pxComponentInfo->uMaterialFlags = Asura_Material_System::GetMaterialFlags( iNewMaterialID );

			// Get the texture ID and texture flags
			if ( pxComponentInfo->iTextureID < 0 )
			{
				pxComponentInfo->iTextureID = ASURA_TEXTURE_NONE;
			}
			const u_int uComponentIndex = ( pxComponentInfo->uComponentIndexAndType & 0x00FFFFFF );
			Asura_Assert( uComponentIndex < s_xHeader.uNumTotalComponents, "" );
			//Doesn't seem to be referenced anyway
			//ppxComponentInfos[uComponentIndex] = pxComponentInfo;
		}

		// Asura_PS2_Botanicals_Component_Info xComponentInfo[uNumComponents]
		pcMemory += sizeof( Asura_PS2_Botanicals_Component_Info ) * pxTypeInfo->uNumComponents;
	}
	Asura_Assert( ( pcMemory - pcTypeInfosStart ) == static_cast<int>( s_xHeader.uTotalTypeInfosSizeInBytes ), "" );

	// The Component Datas

	// Read the component datas in by hand
	s_ppxComponentDatas = new (Asura_PS2_Botanicals_Component_Data*)[s_xHeader.uNumTotalComponents];
	Asura_Assert( s_ppxComponentDatas, "" );

	for ( u_int uI = 0; uI < s_xHeader.uNumTotalComponents; uI++ )
	{
		// Read in name.
		u_int uSizeOfNameInInts;
		const Asura_Char* szName = Asura_FileAccess::ReadString( xFile, NULL, &uSizeOfNameInInts );

		Asura_PS2_Shape_ShapeHdr xShapeHdr;
		Asura_FileAccess::ReadData( xFile, &xShapeHdr, sizeof(Asura_PS2_Shape_ShapeHdr) );

		// Do some debugging to make sure data is valid
		Asura_Assert( xShapeHdr.usTotalTris < 1024*1024, "" );
		Asura_Assert( xShapeHdr.usNumStrips < 1024*1024, "" );
		Asura_Assert( xShapeHdr.usTotalTris != 0, "" );
		Asura_Assert( xShapeHdr.usNumStrips != 0, "" );

		u_int uNumBytes;
		Asura_FileAccess::ReadData( xFile, &uNumBytes, sizeof(u_int) );

		// Allocate the object
        Asura_Assert( ( ( uNumBytes & 0xF ) == 0 ), "" );
		s_ppxComponentDatas[uI] = (Asura_PS2_Botanicals_Component_Data*)Asura_AllocateAlign( uNumBytes, ASURA_ALIGN_128_BYTE );
        Asura_FileAccess::ReadData( xFile, s_ppxComponentDatas[uI], uNumBytes );

		// Get the number of qwords from the return tag: lower 16-bits
		const Asura_PS2_Botanicals_Component_Data* const pxComponentData = s_ppxComponentDatas[uI];
		const Asura_PS2_DmaTag* const pxRETtag =  reinterpret_cast<const Asura_PS2_DmaTag* const>( &( pxComponentData->xRETtag ) );
		Asura_Assert( ( reinterpret_cast<u_int>( pxRETtag ) & 0xF ) == 0, "" );

		// Check this is a RET tag
		Asura_Assert( pxRETtag->m_eID == Asura_PS2_DmaTag::ID_RET, "" );

		// Want to copy the usTotalTris into the correct component info place
		//ppxComponentInfos[uI]->usNumTris = xShapeHdr.usTotalTris;
		s_pusComponentTriCounts[uI] = xShapeHdr.usTotalTris;
	}

	s_pxTails = new (Asura_RenderChainNode*)[s_xHeader.uNumTotalComponents];

	Asura_Assert( s_xHeader.uNumTotalComponents < 65535, "" );
	return true;
}

// Movement is height * wave value
//	pxVertices[0].x += fHeight * fWave;
//	pxVertices[1].x += fHeight * fWave;

void Asura_PS2_Botanicals_Store::Begin( void )
{
	const bool bNeedFullSetup = Asura_PS2_VU1::LoadRenderer( Asura_PS2_VU1::RENDERER_BOTANICALS );

	// Reserve some of the beginning of SPR.
	const u_int uNumModules = Asura_Module_System::GetNumberOfModules();
	const u_int uQWsForVisAndClipFlags = ( uNumModules + 3 ) >> 2;
	const u_int uTotalSPRQWs = TOTAL_SPR_QWS_REQD + uQWsForVisAndClipFlags;

	(void) Asura_PS2_SPR_Alloc( 0, uTotalSPRQWs ); // The cast fixes warning in release.


	// Couple of bits for debugging purposes, for checking everything's
	// correctly initialised, etc.
#if defined( DEBUG_PS2_BOTANICALS )
	if( dbg_bBotanicalsFillSPRInBegin )
	{
		const u_long128 uFillValue = ASURA_PS2_INIT_ULONG128( 0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDuLL );
		u_long128* puOut = reinterpret_cast<u_long128*>( 0x70000000 );
		for( u_int u = uTotalSPRQWs; u; --u )
		{
			*puOut++ = uFillValue;
		}
	}
	if( dbg_bBotanicalsFillVU1MemInBegin )
	{
		Asura_PS2_VU1::FillMem( ASURA_PS2_INIT_ULONG128(0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDuLL) );
	}
#endif // #if defined( DEBUG_PS2_BOTANICALS )

	const u_int uNumCams = Asura_Camera::GetNumberOfCameras();

	const u_int uNumQWsForCamUNPACK    = uNumCams * 5;
	const u_int uNumQWsForGIFTagUNPACK = 2; // ASURA_PS2_BOTANICALS__SIZE__GIF_TAGS;

	const u_int uNumQWsForDMA = uNumQWsForCamUNPACK
		+ 1 + uNumQWsForGIFTagUNPACK
		+ 1
		+ 1 + Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

	const u_int uTotalQWs = 1 + uNumQWsForDMA;

	Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
	Asura_PS2_QW* pxOut = pxBase;

	(reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumQWsForDMA,
		Asura_PS2_DmaTag::PCE_NOCHANGE,
		Asura_PS2_DmaTag::ID_CNT,
		false,
		0,
		false,
		SCE_VIF1_SET_STCYCL( 1, 1, false ),
		SCE_VIF1_SET_UNPACK( ASURA_PS2_BOTANICALS__ADDR__CONST_CAM_DATA, uNumQWsForCamUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false ) );

	// Const cam data.
	for( u_int uCam = 0; uCam < uNumCams; ++uCam )
	{
		pxOut = Asura_PS2_GS::GetViewportSetupGifCmds( uCam, Asura_PS2_GS::CONTEXT_1, pxOut );

		(pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrScaleVec(uCam) );
		(pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrTransVec(uCam) );
		(pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrClipScaleVec(uCam) );
	}

	// Setup BASE and OFFSET and start UNPACK of GIF tags.
	pxOut->u32s[0] = SCE_VIF1_SET_BASE( ASURA_PS2_BOTANICALS__ADDR__INPUT_BFR_0, false );
	pxOut->u32s[1] = SCE_VIF1_SET_OFFSET( ASURA_PS2_BOTANICALS__SIZE__INPUT_BFR, false );
	pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
	pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_BOTANICALS__ADDR__GIF_TAGS, uNumQWsForGIFTagUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false );
	++pxOut;

	// Geometry setup GIF tag. The default NLOOP should be for the base
	// pass; we'll fix it up on VU1 for anything else.
	{
		Asura_PS2_GifTag* pxGeomSetupGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
		pxGeomSetupGifTag->SetExceptRegs( ASURA_PS2_BOTANICALS__RS_OUT_0_NLOOP, true, false, 0, SCE_GIF_PACKED, 1 );
		pxGeomSetupGifTag->uAllRegs = SCE_GIF_PACKED_AD;
	}

	// Geometry draw GIF tag. Note that NLOOP is setup in code.
	{
		Asura_PS2_GifTag* pxGeomDrawGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );

		u_int uPrimType = SCE_GS_PRIM_TRISTRIP;
		bool  bTME      = true;

#ifndef ASURA_PS2_SONY_SUBMISSION
		if( Asura_Render::GetWireframe() )
		{
			uPrimType = SCE_GS_PRIM_LINESTRIP;
			bTME      = false;
		}
#endif

		const u_int uPRIM = SCE_GS_SET_PRIM( uPrimType, 1, bTME, false, true, false, 0, 0, 0 );
		pxGeomDrawGifTag->SetExceptRegs( 0, true, true, uPRIM, SCE_GIF_PACKED, 3 );

		pxGeomDrawGifTag->uAllRegs = (SCE_GS_ST << 0) | (SCE_GS_RGBAQ << 4) | (SCE_GS_XYZ2 << 8);
	}

	// Run INIT program
	pxOut->u32s[0] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_BOTANICALS_INIT );
	pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
	pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
	pxOut->u32s[3] = SCE_VIF1_SET_NOP( false );
	++pxOut;

	// DIRECT for default renderstate setup.
	pxOut->u32s[0] = SCE_VIF1_SET_NOP( false );
	pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
	pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
	pxOut->u32s[3] = SCE_VIF1_SET_DIRECT( Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS, false );
	++pxOut;

	Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1, pxOut, true, false );
	pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

	Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );

	// Reset buffer indexes.
	g_uCurVURSBfr     = 0;
	g_uCurVUCamBfr    = 0;
	g_uCurVULightBfr  = 0;

	// Fill in constant elements of renderstate bfr.
	(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_ALPHA      )->u64s[1] = SCE_GS_ALPHA_1;
	(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_TEX1       )->u64s[1] = SCE_GS_TEX1_1;
	(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_TEX0       )->u64s[1] = SCE_GS_TEX0_1;
	(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_MIPTBP1    )->u64s[1] = SCE_GS_MIPTBP1_1;
	(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_TEST       )->u64s[1] = SCE_GS_TEST_1;

	// Setup some defaults.
	g_uMeshFlags       = 0;
	g_uPrevMeshFlags   = 0;
	g_uPendingFlags    = PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM | PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS | PENDING_UPLOAD_LIGHT;
	g_uCamFlagsForVU   = 0;
	g_bO2WIsIdentity   = true;

	g_uNumLightQWs = ASURA_PS2_BOTANICALS__OFFSET__QW_CONTAINING_QTY + 1;
    g_uNumLightQWs  = 4;
    g_uNumLights    = 0;

	// Setup default renderstate.
	g_iCurBaseTex = ASURA_TEXTURE_NONE;
	g_uCurMaterialFlags = uINVALID_MATERIAL_FLAG;

	SetRenderState( 0 );
}

void Asura_PS2_Botanicals_Store::End( void )
{
	// Send any pending things e.g. set various states but not actually rendered a mesh
	// because the mesh was out of view
	if( g_uPendingFlags )
	{
		SendPendingData( false );
	}

	// FLUSH to prevent subsequent rendering trashing stuff in VU memory that
	// we're still in the middle of using.
	Asura_PS2_QW* pxQWForTag = Asura_PS2_DeferredRendering::GetVifBuffer( 1 );
	(reinterpret_cast<Asura_PS2_DmaTag*>(pxQWForTag))->Set( 0,
		Asura_PS2_DmaTag::PCE_NOCHANGE,
		Asura_PS2_DmaTag::ID_CNT,
		false,
		0,
		false,
		SCE_VIF1_SET_NOP(false),
		SCE_VIF1_SET_FLUSH(false) );

	// Return region of SPR that we reserved in Begin().
	Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(0x70000000) );
}

void Asura_PS2_Botanicals_Store::SendPendingData( const bool bWillMSCNTForNextStrip )
{
	const u_int uMeshFlags = g_uMeshFlags;
	u_int uPendingFlags = g_uPendingFlags;

	// Turn on need-to-send-mesh-flags flag if required.
	if( uMeshFlags != g_uPrevMeshFlags )
	{
		uPendingFlags |= PENDING_UPLOAD_MESH_FLAGS;
		g_uPrevMeshFlags = uMeshFlags;
	}

	// Early out if nothing's changed since the previous strip.
	if( !uPendingFlags ) 
	{ 
		return; 
	}

	// Anything to UNPACK? Do this first, so we can leave it as long as
	// possible before FLUSHing.
	if ( uPendingFlags & ( PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM | PENDING_UPLOAD_LIGHT ) )
	{
		u_int uEstimatedQWC = 0;
		if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE		 ) { uEstimatedQWC += 1 + 8; } // ASURA_PS2_BOTANICALS__SIZE__RENDERSTATE_BFR;							}
		if( uPendingFlags & PENDING_UPLOAD_CAM				 ) { uEstimatedQWC += 1 + ASURA_PS2_BOTANICALS__SIZE__PER_OBJ_CAM_DATA_BFR;					}
		if( uPendingFlags & PENDING_UPLOAD_LIGHT			 ) { uEstimatedQWC += 1 + ASURA_PS2_BOTANICALS__SIZE__LIGHT_DATA_BFR;							}

		const u_int uTotalQWs = 1 + uEstimatedQWC;
		Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
		Asura_PS2_QW* pxOut = pxBase;

		Asura_PS2_DmaTag* const pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut );
		pxCNTTag->Set( 0, // Will be fixed up below as necessary.
			Asura_PS2_DmaTag::PCE_NOCHANGE,
			Asura_PS2_DmaTag::ID_CNT,
			false,
			0,
			false );

		pxCNTTag->m_uWord2 = SCE_VIF1_SET_STCYCL( 1, 1, false );

		bool bNeedNOPs = false;

		if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE )
		{
			const u_int uBfr  = g_uCurVURSBfr ^ 1;
			const u_int uAddr = ( uBfr ? ASURA_PS2_BOTANICALS__ADDR__RENDERSTATE_BFR_1 : ASURA_PS2_BOTANICALS__ADDR__RENDERSTATE_BFR_0 );

			if( bNeedNOPs )
			{
				FillQWWithNOPs( pxOut );
			}
			//pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, ASURA_PS2_BOTANICALS__SIZE__RENDERSTATE_BFR, Asura_PS2_VU1::VIFPK_V4_32, false );
			pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, 5, Asura_PS2_VU1::VIFPK_V4_32, false );
			++pxOut;

			const u_long128* puSrc = &g_pxRSBfr->u128;
			//for( u_int u = ASURA_PS2_BOTANICALS__SIZE__RENDERSTATE_BFR; u; --u )
			for( u_int u = 5; u; --u )
			{
				(pxOut++)->u128 = *puSrc++;
			}

			bNeedNOPs     = true;
			g_uCurVURSBfr = uBfr;
		}

		// Send the camera data
		if( uPendingFlags & PENDING_UPLOAD_CAM )
		{
			const u_int uCamVisFlags = g_uCamVisFlags;
			const u_int uBfr         = g_uCurVUCamBfr ^ 1;
			const u_int uSize        = 4 + 4 + 3;
			u_int uAddr = ( uBfr ? ASURA_PS2_BOTANICALS__ADDR__PER_OBJ_CAM_DATA_BFR_1 : ASURA_PS2_BOTANICALS__ADDR__PER_OBJ_CAM_DATA_BFR_0 );
			const u_int uNumCams = Asura_Camera::GetNumberOfCameras();

			// Send the identity transforms down
			for( u_int uCam = 0; uCam < uNumCams; uAddr += ASURA_PS2_BOTANICALS__OFFSET__CAM_PER_OBJ_NEXT_CAM, ++uCam )
			{
				if( bNeedNOPs )
				{
					FillQWWithNOPs( pxOut );
				}
				bNeedNOPs = true;
				(pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, uSize, Asura_PS2_VU1::VIFPK_V4_32, false );

				// O->GBC matrix.
				{
					sceVu0FMATRIX& xW2GBC = Asura_PS2_VU1::GetWorldToClipMatGuardband( uCam );
					if( g_bO2WIsIdentity )
					{
						Asura_PS2_VU1::CopyPS2Mat( pxOut, xW2GBC );
					}
					else
					{
						sceVu0MulMatrix( &pxOut->Vu0FVec, xW2GBC, &g_pxO2W->Vu0FVec );
					}
					pxOut += 4;
					Asura_PS2_VU1::CopyPS2Mat( pxOut, g_pxO2W );
					pxOut += 4;
				}
				const Asura_Vector_3 xHorizViewVec = s_axCameraHorizVector[uCam];
				const Asura_Vector_3 xVertViewVec = s_axCameraVertVector[uCam];
				const Asura_Vector_3 xFwdViewVec = s_axCameraFwdVector[uCam];

				// Send the horizontal vector
				pxOut->f32s[0] = xHorizViewVec.x;
				pxOut->f32s[1] = xHorizViewVec.y;
				pxOut->f32s[2] = xHorizViewVec.z;
				pxOut->f32s[3] = g_fObjectScale;	// the instance object scale
				pxOut++;

				// Send the height vector
				pxOut->f32s[0] = xVertViewVec.x;
				pxOut->f32s[1] = xVertViewVec.y;
				pxOut->f32s[2] = xVertViewVec.z;
				pxOut->f32s[3] = g_pfAlphaScales[uCam]; // the alpha value for this component per camera
				pxOut++;

				// Send the fwd vector
				pxOut->f32s[0] = -xFwdViewVec.x;
				pxOut->f32s[1] = -xFwdViewVec.y;
				pxOut->f32s[2] = -xFwdViewVec.z;
				pxOut->f32s[3] = 0.0f;
				pxOut++;
			}
			g_uCurVUCamBfr = uBfr;
		}
		if( uPendingFlags & PENDING_UPLOAD_LIGHT )
		{
			u_int uQWs = g_uNumLightQWs;
			Asura_Assert( uQWs, "" );

			const u_int uBfr  = g_uCurVULightBfr ^ 1;
			const u_int uAddr = ( uBfr ? ASURA_PS2_BOTANICALS__ADDR__LIGHT_DATA_BFR_1 : ASURA_PS2_BOTANICALS__ADDR__LIGHT_DATA_BFR_0 );

			if( bNeedNOPs )
			{
				FillQWWithNOPs( pxOut );
			}
			(pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, uQWs, Asura_PS2_VU1::VIFPK_V4_32, false );

			const u_long128* puSrc = &g_pxOutLightBfr->u128;
			for( ; uQWs; --uQWs )
			{
				(pxOut++)->u128 = *puSrc++;
			}

			bNeedNOPs        = true;
			g_uCurVULightBfr = uBfr;
		}

		if( !bNeedNOPs ) { (pxOut++)->u32s[3] = SCE_VIF1_SET_NOP( false ); }

		// Verify that we didn't underestimate.
		Asura_Assert( pxOut <= pxBase + uTotalQWs, "" );

		// Fix up tag.
		u_int uActualQWC = pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCNTTag + 1);
		pxCNTTag->m_uQWC = uActualQWC;

		// Return remainder of buffer if necessary.
		u_int uDiff = (pxBase + uTotalQWs) - pxOut;
		if( uDiff ) { Asura_PS2_DeferredRendering::ReturnVifBuffer( pxOut, uDiff ); }
	}
	if( uPendingFlags & PENDING_UPLOAD_TEX_SYNC )
	{
		Asura_PS2_TextureManagement::SetupTextureEnvironment( g_bTexUploaded, ASURA_PS2_TEXREGFLAG_TEXFLUSH );
	}

	// Add any necessary MSCALs.
	{
		u_int uEstimatedQWC = 0;
		// MSCALs to update ptrs after UNPACKs above.
		if( uPendingFlags & (PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM | PENDING_UPLOAD_LIGHT ) ) { ++uEstimatedQWC; }
		// ITOPs and MSCALs for mesh/cam flags.
		if( uPendingFlags & (PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS) ) { ++uEstimatedQWC; }

		const u_int uTotalQWs = 1 + uEstimatedQWC;
		Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
		Asura_PS2_QW* pxOut = pxBase;

		Asura_PS2_DmaTag* const pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut );
		pxCNTTag->Set( 0, // Will be fixed up below as necessary.
			Asura_PS2_DmaTag::PCE_NOCHANGE,
			Asura_PS2_DmaTag::ID_CNT,
			false,
			0,
			false );

		u_int uNumMSCALs = 0;
		u_int uWordOut   = 2;

		if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE )
		{
			pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_BOTANICALS_SWITCH_RENDERSTATE_BFR, uNumMSCALs );
		}   
		if( uPendingFlags & PENDING_UPLOAD_CAM )
		{
			pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_BOTANICALS_SWITCH_CAM_DATA_BFR, uNumMSCALs );
		}
		if( uPendingFlags & PENDING_UPLOAD_LIGHT )
		{
			pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_BOTANICALS_SWITCH_LIGHT_DATA_BFR, uNumMSCALs );
		}
		if( uPendingFlags & PENDING_UPLOAD_MESH_FLAGS )
		{
			pxOut->u32s[uWordOut++] = SCE_VIF1_SET_ITOP( uMeshFlags, false );
			pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_BOTANICALS_LOAD_MESH_FLAGS, uNumMSCALs );

		}
		if( uPendingFlags & PENDING_UPLOAD_CAM_FLAGS )
		{
			pxOut->u32s[uWordOut++] = SCE_VIF1_SET_ITOP( g_uCamFlagsForVU, false );
			pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_BOTANICALS_LOAD_CAM_FLAGS, uNumMSCALs );
		}

		// We need to run the NOP program up to two times:
		//     (i) If we're transferring a whole chain of strips and so will
		//         be MSCNTing to render the next strip, we need a NOP to get
		//         the VU1 PC to point to the right place.
		//    (ii) If we've written an odd number of MSCALs above, including
		//         (i), we need a dummy one to get the TOP register back to
		//         the correct value, ie. pointing to the strip data that the
		//         caller UNPACKed.
		if( bWillMSCNTForNextStrip ) { pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_BOTANICALS_NOP, uNumMSCALs ); }
		if( uNumMSCALs & 1         ) { pxOut->u32s[uWordOut++] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_BOTANICALS_NOP ); }

		// Pad to end of QW.
		if( uWordOut & 3 )
		{
			for( u_int u = 4 - (uWordOut & 3); u--; )
			{
				pxOut->u32s[uWordOut++] = SCE_VIF1_SET_NOP( false );
			}
		}

		// Fix up output ptr.
		pxOut += (uWordOut >> 2);

		// Verify that we didn't underestimate.
		Asura_Assert( pxOut <= pxBase + uTotalQWs, "" );

		// Fix up tag.
		u_int uActualQWC = pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCNTTag + 1);
		pxCNTTag->m_uQWC = uActualQWC;

		// Return remainder of buffer if necessary.
		u_int uDiff = (pxBase + uTotalQWs) - pxOut;
		if( uDiff ) { Asura_PS2_DeferredRendering::ReturnVifBuffer( pxOut, uDiff ); }
	}

	// Clear pending flags.
	g_uPendingFlags = 0;
}

void Asura_PS2_Botanicals_Store::SetTexture( const int iTexID, const int iMMAG, const int iMMIN, const u_int uBfr )
{
	const u_int uCurPendingFlags = g_uPendingFlags;
	u_int uNewPendingFlags = uCurPendingFlags;

	bool bUploaded = false;

	{
		bool bUpload = true;

		// If there's another texture sync pending, say from another pass's
		// texture upload, then sync with that first.
		if( uCurPendingFlags & PENDING_UPLOAD_TEX_SYNC )
		{
			// We want to avoid doing this sync if possible since it means a
			// FLUSH much earlier, probably forcing a stall while the
			// previous strip finishes rendering. A safe case for doing this
			// is if this next texture we're going to upload is going to some
			// fixed buffer and is already resident there. Note that if we
			// skip the sync, we have to skip the upload below as well (since
			// otherwise internal texture management settings may not be
			// right when we come to do the sync later.) This means that it's
			// not safe to skip the sync if auto buffer switching since we
			// can't skip the logic in UploadTexture() that remembers which
			// buffer we used last (and therefore doesn't go overwriting the
			// texture in it while we're still using it.)
			// (MP - 09.01.2004) TODO: This restriction may become a
			// significant problem and bottleneck if/when we support more
			// general multipassing.
			if( Asura_PS2_TextureManagement::BFR_AUTO != uBfr )
			{
				if( Asura_PS2_TextureManagement::BFR_NONE != Asura_PS2_TextureManagement::HasTextureBeenUploaded( iTexID, uBfr ) )
				{
					bUpload = false;
				}
			}

			if( bUpload )
			{
				Asura_PS2_TextureManagement::SetupTextureEnvironment( g_bTexUploaded, ASURA_PS2_TEXREGFLAG_TEXFLUSH );
			}
		}

		if( bUpload )
		{
			bUploaded = Asura_PS2_TextureManagement::UploadTexture( iTexID, uBfr );

			g_bTexUploaded = bUploaded;
			uNewPendingFlags |= PENDING_UPLOAD_TEX_SYNC;
		}
	}

	// Refresh tex setup stuff if necessary. Note that a check on whether
	// or not the setup has already been done for the specified texture ID
	// breaks in two cases: (i) if they're requesting the same texture but
	// with different filtering settings; (ii) if texture caching isn't
	// working, so that we've uploaded the texture again despite it being
	// resident. I don't think (i) is likely to happen, mainly because we
	// use bilinear everywhere on meshes, and we can workaround (ii) by
	// forcing a refresh if we found the texture needed to be uploaded above.
	int& iCurTex = g_iCurBaseTex;
	if( bUploaded || iTexID != iCurTex )
	{
		const u_int uMIPTBP1Offset = ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_MIPTBP1;
		Asura_PS2_TextureManagement::GetMIPTBP1Setting( &(g_pxRSBfr + uMIPTBP1Offset)->u64s[0], iTexID, uBfr );

		const u_int uTEX1Offset = ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_TEX1;
		Asura_PS2_TextureManagement::GetTEX1Setting( &(g_pxRSBfr + uTEX1Offset)->u64s[0], iTexID, iMMAG, iMMIN );

		const u_int uTEX0Offset = ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_TEX0;
		Asura_PS2_TextureManagement::GetTEX0Setting( &(g_pxRSBfr + uTEX0Offset)->u64s[0], iTexID, uBfr );

		iCurTex = iTexID;
		uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
#if defined( DEBUG_PS2_BOTANICALS )
#if defined( ASURA_USE_CARPET )
		if ( bCarpetBotanicalsDebugTex ) 
		{
			Asura_OutputToDebugger_Retail( ( "iTexID %d TBP 0x%X DW %d DH %d\n", 
										   iTexID,
										   g_pxRSBfr[uTEX0Offset].u32s[0] & 0x3FFF,
										 ( g_pxRSBfr[uTEX0Offset].u32s[0] >> 26 ) & 0xF,
										 ( g_pxRSBfr[uTEX0Offset].u64s[0] >> 30 ) & 0xF ) );
		}
		if ( bCarpetBotanicalsDebugTex ) 
		{
			Asura_OutputToDebugger_Retail( ( "TEX1 0x%X\n",  g_pxRSBfr[uTEX1Offset].u32s[0] ) );
			Asura_OutputToDebugger_Retail( ( "MIPTBP1 0x%X\n",  g_pxRSBfr[uMIPTBP1Offset].u32s[0] ) );
			Asura_OutputToDebugger_Retail( ( "TEX0 0x%X\n",  g_pxRSBfr[uTEX0Offset].u32s[0] ) );
		}
#endif // #if defined( ASURA_USE_CARPET )
#endif // #if defined( DEBUG_PS2_BOTANICALS )
	}
	g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_Botanicals_Store::SetRenderState( const u_int uInputMaterialFlags )
{
	const u_int uOldMeshFlags = g_uMeshFlags;

	u_int uNewMeshFlags    = uOldMeshFlags;
	u_int uNewPendingFlags = g_uPendingFlags;

	// Mask on only the flags that we are interested in
	u_int uMaterialFlags = uInputMaterialFlags;
	uMaterialFlags &= (   AsuraMaterialFlag_SelfIlluminate
						| AsuraMaterialFlag_PS2AdditiveAlpha
						| AsuraMaterialFlag_3dDecal
						| AsuraMaterialFlag_Transparent
						| AsuraMaterialFlag_Reflective );

	// Mask off mesh flags for RS set up externally and after this method.
	uNewMeshFlags &= ~ASURA_PS2_BOTANICALS__MASK__ANY_VTX_MANIPULATIONS;

	if( uMaterialFlags != g_uCurMaterialFlags )
	{
		// Mask off the mesh flags that we may or may not turn on below.
		uNewMeshFlags &= ~(ASURA_PS2_BOTANICALS__FLAG__SELFILLUMINATED | ASURA_PS2_BOTANICALS__FLAG__REFLECTIVE);

		u_long uTEST = SCE_GS_SET_TEST( false, 0, 0, 0, false, 0, true, SCE_GS_ZGEQUAL );

		if( uMaterialFlags & AsuraMaterialFlag_PS2AdditiveAlpha )
		{
			// Pass 0: Cs * As + Cd
			(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
				SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
			// Disable z writes.
			uTEST = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_RGB_ONLY, false, 0, true, SCE_GS_ZGEQUAL );
		}
		else if( uMaterialFlags & AsuraMaterialFlag_Transparent )
		{
			// Pass 0: Cs
			(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
				SCE_GS_SET_ALPHA( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CS, 0 );
			// Only pass with alpha > 63 pass alpha test, so ignore pixels with alpha <= 63
			uTEST = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_GREATER, 63, SCE_GS_AFAIL_KEEP, false, 0, true, SCE_GS_ZGEQUAL );
		}
		else
		{
			// Pass 0: Cs
			(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
				SCE_GS_SET_ALPHA( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CS, 0 );
		}

		(g_pxRSBfr + ASURA_PS2_BOTANICALS__OFFSET__RS_IN_BASE_TEST)->u64s[0] = uTEST;

		uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
		g_uCurMaterialFlags = uMaterialFlags;
	}

	g_uMeshFlags    = uNewMeshFlags;
	g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_Botanicals_Store::SetCamFlags( u_int uVisFlags, u_int uClipFlags )
{
	// Callers should deal with meshes that aren't visible to any cameras.
	Asura_Assert( uVisFlags, "" );

	u_int uNumCamsThatCanSee = 0;
	u_int uVisFlagsForVU     = 0;

	// See notes in Asura_PS2_BotanicalsDefines.h for what's going on here.
	{
		u_int uEEVisBit = 1 << 0;
		u_int uVUVisBit = 1 << 7;

		for( u_int u = Asura_Camera::GetNumberOfCameras(); u; --u )
		{
			if( uVisFlags & uEEVisBit )
			{
				uVisFlagsForVU |= uVUVisBit;
				++uNumCamsThatCanSee;
			}

			uEEVisBit <<= 1;
			uVUVisBit >>= 2;
		}
	}

	// Ensure that no clip flags are set where the corresponding vis flag
	// isn't. This breaks some code that assumes the flags will be zero once
	// there are no more cameras to render for.
	Asura_Assert( !(uClipFlags & ~(uVisFlagsForVU >> 1)), "" );

	u_int uNewFlagsForVU = uVisFlagsForVU | uClipFlags;

	// If more than one camera can see the mesh, let the program know.
	if( uNumCamsThatCanSee > 1 )
	{
		uNewFlagsForVU |= ASURA_PS2_BOTANICALS__FLAG__VIS_FOR_MORE_THAN_ONE_CAM;
	}

	if( uNewFlagsForVU != g_uCamFlagsForVU )
	{
		g_uCamVisFlags = uVisFlags;
		g_uCamFlagsForVU = uNewFlagsForVU;
		g_uNumCamsVisFor = uNumCamsThatCanSee;
		g_uPendingFlags |= PENDING_UPLOAD_CAM_FLAGS;
	}
}

#endif // #if defined( ASURA_USE_BOTANICALS ))