//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Multithread.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_Multithread.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"

//PS2
#include <eekernel.h>

// N O T E ///////////////////////////////////////////////////////////////////
// 
// Limitations of PS2 Multithreading (in this implementation)
//
// (1) Between the thread function releasing the stack mutex and the thread dying 
//     it is possible (although pretty unlikely unless you have a loop creating a 
//	   thread whenever and old one dies) for a thread to take the dying threads stack 
//     space - The results of this would be undefined.
//
// (2) Only one thread can concurrently wait on a single sync event
//
// (3) The waitevent and clearevent functionality of a sync event is indivisible,
//     as a result clearevent itself is not implemented.
//

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////
class Asura_PS2_Thread_Stack
{
public:
	static const int NO_OWNER = -1;

	Asura_PS2_Thread_Stack()
	: m_iOwner(NO_OWNER)
	{
		Asura_Mem::Set(&m_acStackSpace, 0, ASURA_PS2_THREAD_STACK_SIZE);
	}

	char m_acStackSpace[ASURA_PS2_THREAD_STACK_SIZE] __attribute__ ((aligned(16)));
	int m_iOwner;
};

class Asura_PS2_ThreadParams
{
public:

	Asura_PS2_ThreadParams( Asura_ThreadFunc pfnThreadFunc, const void* pUserData )
		: m_pfnThreadFunc (pfnThreadFunc)
		, m_pUserData (pUserData)
	{
	}

	// data
	Asura_ThreadFunc m_pfnThreadFunc;
	const void* m_pUserData;
};

// G L O B A L S ////////////////////////////////////////////////////////////

static Asura_Mutex g_xAsura_PS2_ThreadStackMutex;
static Asura_PS2_Thread_Stack g_axAsura_PS2_ThreadStackData[ASURA_PS2_MAX_THREADS];

//TODO: Initialise this array and use it for the stack size
u_int Asura_Thread_System::s_auStackSizes[]  =	{ 0		//ASURA_MULTITHREAD_STACK_SIZE_SMALL
												, 0		//ASURA_MULTITHREAD_STACK_SIZE_MEDIUM
												, 0 };	//ASURA_MULTITHREAD_STACK_SIZE_LARGE

// S T A T I C   F U N C T I O N S ///////////////////////////////////////////
static void Asura_PS2_ThreadStartup( void* pParam  )
{
	//TODO: Disable FP Exceptions?

	const Asura_PS2_ThreadParams* pxParams = reinterpret_cast<const Asura_PS2_ThreadParams*>(pParam);
	Asura_ThreadFunc pfnThreadFunc = pxParams->m_pfnThreadFunc;
	const void* pUserData = pxParams->m_pUserData;
	pfnThreadFunc(pUserData);

	g_xAsura_PS2_ThreadStackMutex.Lock();
	
	int iThreadID = GetThreadId();
	int iStackElement;
	for (iStackElement = 0; iStackElement < ASURA_PS2_MAX_THREADS; iStackElement++)
	{
		if (g_axAsura_PS2_ThreadStackData[iStackElement].m_iOwner == iThreadID)
		{
			g_axAsura_PS2_ThreadStackData[iStackElement].m_iOwner = Asura_PS2_Thread_Stack::NO_OWNER;
			break;
		}
	}
	Asura_Assert(iStackElement < ASURA_PS2_MAX_THREADS,"Couldn't find threads stack space to free");

	g_xAsura_PS2_ThreadStackMutex.Unlock();

	ExitDeleteThread();
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Thread_System
//

u_int Asura_Thread_System::Platform_GetNumHardwareThreads()
{
    return 1;
}

bool Asura_Thread_System::Platform_CreateThread( const char* szThreadName, Asura_ThreadFunc pfnThreadFunc, const void* pUserData, const ASURA_MULTITHREAD_STACK_SIZE eStackSize )
{
	Asura_PS2_ThreadParams xParams(pfnThreadFunc, pUserData);

	g_xAsura_PS2_ThreadStackMutex.Lock();

	bool bResult = false;

	int iStackElement;
	for (iStackElement = 0; iStackElement < ASURA_PS2_MAX_THREADS; iStackElement++)
	{
		if (g_axAsura_PS2_ThreadStackData[iStackElement].m_iOwner == Asura_PS2_Thread_Stack::NO_OWNER) break;
	}

	if (g_axAsura_PS2_ThreadStackData[iStackElement].m_iOwner != Asura_PS2_Thread_Stack::NO_OWNER)
	{
		Asura_Assert(false,"Thread allocation used, increase ASURA_PS2_MAX_THREADS");
	}
	else
	{
		struct ThreadParam xCurrentThreadParam;
		ReferThreadStatus(GetThreadId(), &xCurrentThreadParam);

		struct ThreadParam xNewThreadParam;
		xNewThreadParam.entry = Asura_PS2_ThreadStartup;
		xNewThreadParam.stack = static_cast<void*>(&g_axAsura_PS2_ThreadStackData[iStackElement].m_acStackSpace);
		xNewThreadParam.stackSize = ASURA_PS2_THREAD_STACK_SIZE;
		xNewThreadParam.gpReg = &_gp;
		xNewThreadParam.initPriority = xCurrentThreadParam.currentPriority;
		
		int iRes =  ::CreateThread(&xNewThreadParam);
		Asura_Assert(iRes >=0, "Failed to create thread");

		if (iRes >= 0)
		{
			iRes = StartThread(iRes, &xParams);
			Asura_Assert(iRes >= 0,"Failed to start thread");
			if (iRes >= 0)
			{
				g_axAsura_PS2_ThreadStackData[iStackElement].m_iOwner = iRes;
				bResult = true;
			}
		}
	}
	
	g_xAsura_PS2_ThreadStackMutex.Unlock();

	YieldCurrentThread();

	return bResult;
}


void Asura_Thread_System::Platform_YieldCurrentThread( void )
{
	struct ThreadParam xThreadParam;
	ReferThreadStatus(GetThreadId(), &xThreadParam);

#ifdef ASURA_ASSERT
	int iRes = 
#endif
	RotateThreadReadyQueue(xThreadParam.currentPriority);
	
	Asura_Assert(iRes == xThreadParam.currentPriority,"Error yielding current thread");
}


Asura_ThreadID Asura_Thread_System::Platform_GetCurrentThreadID( void )
{ 
    return GetThreadId();
}


Asura_ThreadID Asura_Thread_System::Platform_GetInvalidThreadID( void )
{ 
    return -1;
}

void Asura_Thread_System::Platform_DelayCurrentThread( float fTimeInSeconds, bool ballowCallbacks )
{ 
	Asura_Assert( !ballowCallbacks, "Callbacks not supported" );	
	
#ifdef ASURA_ASSERT
	int iRes =
#endif
	//Surprisingly Sonys delay function takes microseconds as an argument
	DelayThread(Asura_Maths::FloatToInt(fTimeInSeconds * 1000000.0f));

	Asura_Assert(iRes >= 0,"Error delaying current thread");
}

//
// Asura_Mutex
//

void Asura_Mutex::Platform_Lock( void )
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
		WaitSema(m_iSemaphoreId);
	Asura_Assert(m_iSemaphoreId == iRet, "SignalSema for mutex failed");
}


void Asura_Mutex::Platform_Unlock()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
		SignalSema(m_iSemaphoreId);
	Asura_Assert(m_iSemaphoreId == iRet, "SignalSema for mutex failed");
}

//
// Asura_SyncEvent
//

void Asura_SyncEvent::Platform_WaitForSignal( void )
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
		WaitSema(m_iSemaphoreId);
	Asura_Assert(m_iSemaphoreId == iRet, "SignalSema for sync event failed");
	
	// Unfortunately WaitSema removes the signal
	m_bSignalled = false;
}

void Asura_SyncEvent::Platform_SetSignal( void )
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
		SignalSema(m_iSemaphoreId);
	Asura_Assert(m_iSemaphoreId == iRet, "SignalSema for sync event failed");
}

void Asura_SyncEvent::Platform_ClearSignal( void )
{
    Asura_Assert(false,"Not supported on PS2");
}


//////////////////////////////////////////////////////////////////////////
// Semaphore
//////////////////////////////////////////////////////////////////////////
void Asura_Semaphore::Platform_Wait()
{
}

bool Asura_Semaphore::Platform_TryWait()
{
	return false;
}

bool Asura_Semaphore::Platform_Signal()
{
	return false;
}
