//////////////////////////////////////////////////////////////////////////////
//
// Asura_Assert.cpp
//
// By: Richard M Smith, Morgan Parry.
//
// Copyright (c) 2001 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////



//////////////
// INCLUDES //
//////////////

#include "Core/Asura.h"

// This file's header.
#include "Debugging/Asura_Assert.h"

#ifdef ASURA_ASSERT

// Stdlib independent.
#include <stdarg.h>

// Stdlib platform-specific.
#include <devfont.h>
#include <libgraph.h>
#include <libpad.h>
#include <libsn.h>

// Engine independent.
#include "Core/Asura_String.h"
#include "Core/Asura_UserInput.h"
#include "Debugging/Asura_Logfile.h"
#include "Input/Asura_PlatformInput.h"
#include "Network/Asura_Network.h"

// Engine platform-specific.
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_PS2_DMA.h"
#include "Core/Asura_PS2_OS.h"
#include "Debugging/Asura_PS2_DebugInfo.h"
#include "Input/Asura_PS2_ControlPortDevice.h"
#include "Input/Asura_PS2_Input.h"
#include "ProjectSettings/Asura_PS2_ProjectControllerDefines.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#ifdef ASURA_PS2_DEFERRED_RENDERING
#include "Rendering/Asura_PS2_DeferredRendering.h"
#endif
#include "Rendering/Asura_PS2_GS.h"

// (MP - 24.10.2001) Warn if using asserts in CD builds. We will occasionally
// want this but, on the other hand, it's all too easy to leave on by accident.
#if !_DEBUG && defined(ASURA_ASSERT) && defined(ASURA_PS2_CDEMUL) && !defined(_MSC_VER)
    #warning ARE YOU SURE YOU WANT ASSERTS IN CD RELEASE BUILDS?
	static bool g_bIgnoreAllAsserts = true;
#else
	static bool g_bIgnoreAllAsserts = false;
#endif

	u_char		g_ucAssertIgnoreCounter	= 1;

// T Y P E S /////////////////////////////////////////////////////////////////

enum
{
    ASURA_PS2_ASSERT_INVALID    = -1,
    ASURA_PS2_ASSERT_BREAK,
    ASURA_PS2_ASSERT_IGNORE,
    ASURA_PS2_ASSERT_IGNORE_ALL
};

// G L O B A L S /////////////////////////////////////////////////////////////

static volatile bool g_bIgnoreThisAssertOnce = false;
static volatile bool g_bIgnoreThisAssertAll  = false;
static volatile bool g_bCurrentlyInAssert    = false;

// F U N C T I O N S /////////////////////////////////////////////////////////

void AsuraAssert_SetIgnoreAll( bool bIgnore )
{
	g_bIgnoreAllAsserts = bIgnore;
}

bool AsuraAssert_GetIgnoreAll()
{
	return g_bIgnoreAllAsserts;
}

int AssertConsoleCharsToLines(int iChars)
{
    return (iChars + 55)/56;
}

static void DrawAssertConsole( int iConsoleID, int iCurrentSelection, const char* szFile, int iLine, const char* szCondition, const char* szDescription )
{
	Asura_PS2_GS::DebugClearFrame();

    sceDevConsClearBox( iConsoleID, 0, 0, 56, 24 );
	sceDevConsLocate( iConsoleID, 0, 0 );// (MH - 17.04.2002) strangely, setting y to non-0 is *really* slow.

    sceDevConsAttribute( iConsoleID, 2 );
    sceDevConsPrintf( iConsoleID, "________________________________________________________\n" );
	sceDevConsPrintf( iConsoleID, "ASSERTION FAILURE\n" );
    sceDevConsPrintf( iConsoleID, "________________________________________________________\n" );

	sceDevConsAttribute( iConsoleID, 7 );
	sceDevConsPrintf( iConsoleID, "File:        %s\n", szFile );
	sceDevConsPrintf( iConsoleID, "\nLine:        %i\n", iLine );

    if( szCondition )
    {
	    sceDevConsPrintf( iConsoleID, "\nCondition:   %s\n", szCondition );
    }

    if( szDescription )
    {
	    sceDevConsPrintf( iConsoleID, "\nDescription: %s\n", szDescription );
    }

    sceDevConsPrintf( iConsoleID, "________________________________________________________\n" );

    sceDevConsAttribute( iConsoleID, (ASURA_PS2_ASSERT_BREAK == iCurrentSelection ? 4 : 7) );
	sceDevConsPrintf( iConsoleID, "    BREAK" );

    sceDevConsAttribute( iConsoleID, (ASURA_PS2_ASSERT_IGNORE == iCurrentSelection ? 4 : 7) );
	sceDevConsPrintf( iConsoleID, "                IGNORE" );

    sceDevConsAttribute( iConsoleID, (ASURA_PS2_ASSERT_IGNORE_ALL == iCurrentSelection ? 4 : 7) );
	sceDevConsPrintf( iConsoleID, "           IGNORE ALL\n" );

    sceDevConsAttribute( iConsoleID, 7 );
    sceDevConsPrintf( iConsoleID, "________________________________________________________\n" );

    // We were previously using sceDevConsDraw() here, rather than
    // sceDevConsDrawS(), in order to avoid problems when asserts fire during
    // any code using SPR, such as much of the rendering. However, that
    // function uses a vast amount of stack, which then causes problems when
    // very low on memory, in that the stack will suddenly descend into the
    // heap in release builds where we set a rather tighter stack size
    // (typically 32Kb.) So instead, we use sceDevConsDrawS() but backup the
    // contents of SPR to the stack first, which at least only uses 16Kb,
    // rather than the ~280Kb consumed by sceDevConsDraw().
    {
        void* const pvSPR = reinterpret_cast<void*>( ASURA_PS2_SPR_MEM_BASE );
        const u_int uSPRSize = 16 * 1024;

        u_char aucSPRBackup[uSPRSize];
        memcpy( aucSPRBackup, pvSPR, uSPRSize );
    
	    sceDevConsDrawS( iConsoleID );

        memcpy( pvSPR, aucSPRBackup, uSPRSize );
    }

    sceGsSyncV( 0 );
}


static void ResetRenderstateForAssertRendering( void )
{
    const u_int uNLOOP      = 7;
    const u_int uNumQWs     = 1 + uNLOOP;
    const u_int uRoundedQWs = (uNumQWs + 3) & ~3;

    static Asura_PS2_QW axBfr[uRoundedQWs] __attribute__ ((aligned(64)));

    Asura_PS2_QW* const pxStart = Asura_PS2_UncachedMem( &axBfr[0] );
    Asura_PS2_QW* pxOut = pxStart;

    Asura_PS2_GifTag* pxGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut );
    pxGifTag->SetExceptRegs( uNLOOP, true, false, 0, SCE_GIF_PACKED, 1 );
    pxGifTag->uAllRegs = SCE_GIF_PACKED_AD;
    ++pxOut;

    // (MP - 20.08.2007) Unfortunately we can't do this with this normal
    // engine functionality - eg.
    // Asura_PlatformRenderStates::CreateDefaultRenderStatePacket() or
    // similar - since an assert in rendering code will be requiring some of
    // this state.

    // From Asura_PlatformRenderStates::CreateDefaultRenderStatePacket():
    pxOut->u64s[0] = SCE_GS_SET_ZBUF( Asura_PS2_GS::GetZBfrBasePtr(), Asura_PS2_GS::GetZBfrPSM(), false );
    pxOut->u64s[1] = SCE_GS_ZBUF_1;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_TEXA( 0, false, 128 );
    pxOut->u64s[1] = SCE_GS_TEXA;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
    pxOut->u64s[1] = SCE_GS_ALPHA_1;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_TEST_1( false, 0, 0, 0, false, 0, true, SCE_GS_ZALWAYS );
    pxOut->u64s[1] = SCE_GS_TEST_1;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_CLAMP( false, false, 0, 0, 0, 0 );
    pxOut->u64s[1] = SCE_GS_CLAMP_1;
    ++pxOut;

    // From Asura_PS2_GS::GetViewportSetupGifCmds():
    const Asura_PS2_GS::ViewportParams& xViewportParams = Asura_PS2_GS::GetViewportParamsForFullScreenRendering();
    pxOut->u64s[0] = SCE_GS_SET_SCISSOR( xViewportParams.sOriginX, xViewportParams.sOriginX + xViewportParams.sWidth - 1, xViewportParams.sOriginY, xViewportParams.sOriginY + xViewportParams.sHeight - 1 );
    pxOut->u64s[1] = SCE_GS_SCISSOR_1;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_XYOFFSET( xViewportParams.sPrimTransX, xViewportParams.sPrimTransY );
    pxOut->u64s[1] = SCE_GS_XYOFFSET_1;
    ++pxOut;

    Asura_Assert( static_cast<u_int>(pxOut - pxStart) == uNumQWs, "" );

    Asura_PS2_DMA::NormalTransferFromMem( Asura_PS2_DMA::CHANNEL_GIF, Asura_PS2_DMAMem(pxStart), uNumQWs, false, true, true );
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION     : Asura_AssertFired
// PARAMETERS   : Name of file containing the ASSERT
//              : Line number of the ASSERT
//              : Condition that has failed to come here
//              : Description provided by the programmer
//              : Value to set if the user selects "Ignore All"
// RETURNS      : TRUE if the program should break; FALSE otherwise
// DESCRIPTION  : On screen assert message providing the above information with the option to
//                break into the program, ignore this one assert or ignore this assert for the
//                remainder of this program
// AUTHOR       : Richard M. Smith (19/07/01 15:12:27)
///////////////////////////////////////////////////////////////////////////////////////////////////////////
bool Asura_AssertFired( const char* szFile, int iLine, const char* szCondition, const char* szDescription, u_char* pucIgnore )
{
	// PJC (15.01.2008) Asserts in an interrupt will be hidden if they occur during an 
	// assert in the main thread, but as an interrupt function should simply return in 
	// this situation it should not matter

	// Prevent recursive asserts, eg. a problem in the input code.
    if( g_bCurrentlyInAssert ) { return false; }
    g_bCurrentlyInAssert = true;
    
    const char* ASSERTION_LOGFILE = "logfiles\\asserts.txt";

    const char* ASSERTION_OUTPUT_FORMAT_STRING_MAIN = "ASSERTION FAILURE:\n   File:        %s (line %i)\n";
    const char* ASSERTION_OUTPUT_FORMAT_STRING_COND = "   Condition:   %s\n";
    const char* ASSERTION_OUTPUT_FORMAT_STRING_DESC = "   Description: %s\n";

    // Output to debugger.
    {
        Asura_OutputToDebugger_Retail(( ASSERTION_OUTPUT_FORMAT_STRING_MAIN, szFile, iLine ));
        
        if( szCondition ) { Asura_OutputToDebugger_Retail(( ASSERTION_OUTPUT_FORMAT_STRING_COND, szCondition )); }

        if( szDescription ) { Asura_OutputToDebugger_Retail(( ASSERTION_OUTPUT_FORMAT_STRING_DESC, szDescription )); }
    }
    
	// (MP - 22.08.2007) Asserts within interrupts are going to cause
	// problems. We may ultimately want a better solution than this -
	// probably making everything that might be called as a result of the
	// assert work in an interrupt - but this'll do for now (where I'm
	// implementing a vsync callback renderer for the Simpsons loading
	// screens).
	// (PJC 15.01.2008) Moving here to allow debug output during interrupt 
	// and to try and catch asserts in BF3 vsync loading
	if( QueryIntrContext() ) 
	{ 
		if (snIsDebuggerRunning())
		{
			// *** Assert within interrupt ***
			// An assert has occurred during an interrupt. A break has been called to allow
			// the assert to be viewed in the debugger output as it cannot be rendered from
			// inside an interrupt. Pressing F9 (Start the target) will perform the same 
			// action as clicking "Ignore All" would normally. PJC 08/02/08
			ASURA_PS2_DEBUG_BREAK();
		}

		Asura_OutputToDebugger_Retail(( "\n[assertion in interrupt ignored and disabled]\n" ))
		
		g_bCurrentlyInAssert = false;
		*pucIgnore = g_ucAssertIgnoreCounter;
		
		return false; 
	}

    // Log as well.
    {
        static bool bFirstTimeAssertFired = true;
        if (bFirstTimeAssertFired)
        {
            bFirstTimeAssertFired = false;
            Asura_SimpleLogfileClear(ASSERTION_LOGFILE);
        }

        Asura_SimpleLogfileOutput(( ASSERTION_LOGFILE, ASSERTION_OUTPUT_FORMAT_STRING_MAIN, szFile, iLine ));
        
        if( szCondition ) { Asura_SimpleLogfileOutput(( ASSERTION_LOGFILE, ASSERTION_OUTPUT_FORMAT_STRING_COND, szCondition )); }

        if( szDescription ) { Asura_SimpleLogfileOutput(( ASSERTION_LOGFILE, ASSERTION_OUTPUT_FORMAT_STRING_DESC, szDescription )); }
    }

    // If there's no debugger running then breaking is a really bad idea.
    const bool bCanBreak = ( 0 != snIsDebuggerRunning() );

    const int iConsoleID = Asura_PS2_DebugInfo::GetConsoleID();

    bool bShouldBreak;

    if( g_bIgnoreAllAsserts )
    {
        *pucIgnore = g_ucAssertIgnoreCounter;
        bShouldBreak = false;
    }
    else if( iConsoleID == 0 || !Asura_UserInput::IsInitialised() )
    {
        bShouldBreak = bCanBreak;
    }
    else
    {
        u_int	uNumFrames			= 0;
		bool    boFinished          = false;
        int     iCurrentSelection   = ASURA_PS2_ASSERT_IGNORE;
        int     iPreviousSelection  = ASURA_PS2_ASSERT_INVALID;
//      u_int   paddata             = 0;
//      u_int   maskedpaddata       = 0;
//      u_int   lastpaddata         = 0;

#ifdef ASURA_PS2_DEFERRED_RENDERING
	   Asura_PS2_DeferredRendering::WaitForDeferredRendering();
#endif

		// (MH - 02.09.2004) stop vibrations
		for (int iCurrentPort=0; iCurrentPort<2; ++iCurrentPort)
		{
	        int iNumberOfSlots = Asura_PS2_Input::MultitapAvailable(iCurrentPort) ? 4 : 1;
			for (int iCurrentSlot = 0; iCurrentSlot < iNumberOfSlots; iCurrentSlot++)
			{
				Asura_PS2_ControlPortDevice* pxControlPortDevice = Asura_PS2_Input::GetControlPortDevice(iCurrentPort * 4 + iCurrentSlot);
				if(pxControlPortDevice)
				{
					// Only set small motor if magnitude is greater than a half
					pxControlPortDevice->SetRumble(0, 0);
					pxControlPortDevice->Update();
				}
			}
		}
		
        Asura_PS2_GS::EnableDrawToFrontBfr( true );
        ResetRenderstateForAssertRendering();

		while(Asura_UserInput::HasAnyKeyBeenPressed())
		{
			Asura_UserInput::Update();
#if ASURA_PS2_SUPPORT_GUNCON2
            // (NRM - 25.11.2003) update done around Vsync below
            Asura_PS2_Input::UpdateGuncon2(0);
            Asura_PS2_Input::UpdateGuncon2(1);
#endif
            
			DrawAssertConsole( iConsoleID, iCurrentSelection, szFile, iLine, szCondition, szDescription );
		}

        do
        {
			Asura_UserInput::Update();
#if ASURA_PS2_SUPPORT_GUNCON2
            Asura_PS2_Input::TranslateGuncon2(0);
            Asura_PS2_Input::TranslateGuncon2(1);
#endif
            
			bool bLeftPressed = Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD1_LEFT) |
								Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD2_LEFT) |
								Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD1_L1)	|
								Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD2_L1);

			bool bRightPressed = Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD1_RIGHT) |
								 Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD2_RIGHT) |
								 Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD1_R1)	  |
								 Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD2_R1);
			if(bLeftPressed)
			{
                if( iCurrentSelection != ASURA_PS2_ASSERT_BREAK )
                {
                    iCurrentSelection--;
                }
			}

			if(bRightPressed)		   
			{
                if( iCurrentSelection != ASURA_PS2_ASSERT_IGNORE_ALL)
                {
                    iCurrentSelection++;
                }
			}

			boFinished = Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD1_CROSS) |
						 Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_PAD2_CROSS) |
						 Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_GUNCON1_TRIGGER) |
						 Asura_UserInput::DebouncedKeyPressed(ASURA_PS2_KEYID_GUNCON2_TRIGGER);

            if (g_bIgnoreThisAssertOnce)
            {
                g_bIgnoreThisAssertOnce = false;
                iCurrentSelection = ASURA_PS2_ASSERT_IGNORE;
                boFinished = true;
            }
            else if (g_bIgnoreThisAssertAll)
            {
                g_bIgnoreThisAssertAll = false;
                iCurrentSelection = ASURA_PS2_ASSERT_IGNORE_ALL;
                boFinished = true;
            }
			else if (uNumFrames > 200 && Asura_Network::IsExternalNetworkActive() && 
				((!Asura_Network::IsServer()) || Asura_Network::GetNumberOfMachines() > 1))
			{
				// only display for a couple of seconds if external network AND 
				// if we're the server, if people are connected, or if we're just a 
				// client then always.
                iCurrentSelection = ASURA_PS2_ASSERT_IGNORE;
                boFinished = true;
			}

#if ASURA_PS2_SUPPORT_KEYBOARD
            if(Asura_UserInput::DebouncedKeyPressed('i'))
            {
                iCurrentSelection = ASURA_PS2_ASSERT_IGNORE;
                boFinished = true;
            }

            if(Asura_UserInput::DebouncedKeyPressed('a'))
            {
                iCurrentSelection = ASURA_PS2_ASSERT_IGNORE_ALL;
                boFinished = true;                
            }

            if(Asura_UserInput::DebouncedKeyPressed('b'))
            {   
                iCurrentSelection = ASURA_PS2_ASSERT_BREAK;
                boFinished = true;                
            }
#endif

            if (!boFinished && iCurrentSelection != iPreviousSelection)
            {
                // (MH - 21.08.2002) only redrawing if selection changed
                // (MP - 07.05.2002) This will solve the problem of new
                // asserts getting muddled up with old asserts on-screen if
                // we haven't drawn a frame inbetween.
#if ASURA_PS2_SUPPORT_GUNCON2
                Asura_PS2_Input::UpdateGuncon2(0);
#endif
				DrawAssertConsole( iConsoleID, iCurrentSelection, szFile, iLine, szCondition, szDescription );

#if ASURA_PS2_SUPPORT_GUNCON2
                Asura_PS2_Input::UpdateGuncon2(1);
#endif
            }
            else
            {
#if ASURA_PS2_SUPPORT_GUNCON2
                Asura_PS2_Input::UpdateGuncon2(0);
#endif
    			sceGsSyncV( 0 );

#if ASURA_PS2_SUPPORT_GUNCON2
                Asura_PS2_Input::UpdateGuncon2(1);
#endif
            }

			uNumFrames++;

        } while( !boFinished );

        // Wait for the X button to be released
		while(Asura_UserInput::HasAnyKeyBeenPressed())
		{
			Asura_UserInput::Update();
#if ASURA_PS2_SUPPORT_GUNCON2
            Asura_PS2_Input::UpdateGuncon2(0);
#endif
            
			DrawAssertConsole( iConsoleID, iCurrentSelection, szFile, iLine, szCondition, szDescription );
		}

        Asura_PS2_GS::DebugClearFrame();
        sceGsSyncV( 0 );

        Asura_PS2_GS::EnableDrawToFrontBfr( false );

        // Ensure white is the currently selected font colour
        sceDevConsAttribute( iConsoleID, 7 );

        switch( iCurrentSelection )
        {
            case ASURA_PS2_ASSERT_BREAK:
                bShouldBreak = bCanBreak;
                break;

            default:
            case ASURA_PS2_ASSERT_IGNORE:
                Asura_OutputToDebugger_Retail(( "[assertion ignored]\n" ));
                bShouldBreak = false;
                break;
        
            case ASURA_PS2_ASSERT_IGNORE_ALL:
                Asura_OutputToDebugger_Retail(( "[assertion ignored and disabled]\n" ));
				*pucIgnore = g_ucAssertIgnoreCounter;
                bShouldBreak = false;
                break;
        }

    }

    g_bCurrentlyInAssert = false;
    return bShouldBreak;
}


static const char*  g_szAsuraAssertfFile;
static int          g_iAsuraAssertfLine;
static const char*  g_szAsuraAssertfCondition;
static u_char*      g_pucAsuraAssertfIgnore;
static const int    ASURAASSERTF_OUTPUT_BUFFER_SIZE = 2048;
static char         g_szOutputBuffer[ASURAASSERTF_OUTPUT_BUFFER_SIZE];


void Asura_AssertfSetup( const char* szFile, int iLine, const char* szCondition, u_char* pucIgnore)
{
    g_szAsuraAssertfFile = szFile;
    g_iAsuraAssertfLine = iLine;
    g_szAsuraAssertfCondition = szCondition;
    g_pucAsuraAssertfIgnore = pucIgnore;
}

bool Asura_AssertfFired( const Asura_Char* szFormat, ... )
{
	va_list xArgs;
	va_start( xArgs, szFormat );
    Asura_PrintfToBfr( g_szOutputBuffer, szFormat, xArgs, ASURAASSERTF_OUTPUT_BUFFER_SIZE, true );
	va_end( xArgs );

    bool bRetVal = Asura_AssertFired(g_szAsuraAssertfFile, g_iAsuraAssertfLine, 
        g_szAsuraAssertfCondition, g_szOutputBuffer, g_pucAsuraAssertfIgnore);

    g_szAsuraAssertfFile = 0;
    g_iAsuraAssertfLine = 0;
    g_szAsuraAssertfCondition = 0;
    g_pucAsuraAssertfIgnore = 0;

    return bRetVal;
}


// (MH - 17.02.2004) implementation of library assert which should be used in
// preference to the library version, which seems to just call exit() on PS2.
// ignore all is not implemented however.
extern "C"
{
	void __assert(const char* szFile, int iLine, const char* szCondition)
	{
		bool bIgnoreAllDummy;
		if (Asura_AssertFired(szFile, iLine, szCondition, "", &bIgnoreAllDummy))
		{
			ASURA_PS2_DEBUG_BREAK();
		}
		// hopefully you can get to the calling point by stepping over.
	}
}


bool Asura_PS2_IsInAssert( void )
{
    return g_bCurrentlyInAssert;
}

void AsuraAssert_ReEnableAll()
{
	do 
	{
		++g_ucAssertIgnoreCounter;
	} while ( g_ucAssertIgnoreCounter == 0 ); // If we wrap the u_char, don't want to accidentally ignore all asserts
}

bool AsuraAssert_ConsoleCmd_EnableAsserts( bool bHelp )
{
	if ( bHelp )
	{
		return true;
	}
	AsuraAssert_SetIgnoreAll(false);
	AsuraAssert_ReEnableAll();
	return true;
}

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifdef ASURA_ASSERT

