/* Copyright (C) 2002 Sony Computer Entertainment America */
/*        All Rights Reserved.  SCEA Confidential         */

#ifndef SCEIMAGE_FILTER_H
#define SCEIMAGE_FILTER_H


// *************SMOOTHING FILTERS*******************
// The shift argument is how much the resulting filtered image
// should be right-shifted after smoothing.  The default amount
// will right-shift the image to keep the same average as the original.
// By adjusting the shift amount, you can maintain a higher precision
// when repeatedly smoothing.

// 5x5 gaussian conv, result is offset up and to the left 2 pixels
void smooth( const sceImage& _src, sceImage& _dest, int shift=8);

// 5x5 gaussian conv, result is offset down and to the right 2 pixels
void smooth2( const sceImage& _src, sceImage& _dest, int shift=8);

// 3x3 gaussian conv, result is offset up and to the left 1 pixel
void lsmooth( const sceImage& _src, sceImage& _dest, int shift=4);

// 3x3 gaussian conv, result is offset down and to the right 1 pixel
void lsmooth2( const sceImage& _src, sceImage& _dest, int shift=4);

// *************THRESHOLDING FILTERS*******************

// binarizes pixels to 0 or 0xff depending if input pixel > threshold value
void threshold(const sceImage& _src, sceImage& _dest, unsigned int threshold=0);

// binarizes pixels to 0 or 0xff depending if input pixel < threshold value
void invthreshold(const sceImage& _src, sceImage& _dest, unsigned int threshold=255);

// binarizes pixels to 0 or 0xff depending if threshold1<input pixel<threshold2
void image_band(const sceImage& _src, sceImage& _dest, unsigned int threshold1, unsigned int threshold2);

// binarizes pixels to 0 or 0xff depending if low1<_src1 pixel<high1 AND
// low2<_src2 pixel<high2.  Really just an optimized version of:
// image_band(_src1,_dest1,low1,high1);
// image_band(_src2,_dest2,low2,high2);
// image_and(_dest1,_dest2, _dest);
void image_band2(const sceImage& _src1, const sceImage& _src2,sceImage& _dest,
 unsigned int low1, unsigned int high1, unsigned int low2, unsigned int high2);

// same as above, but also returns count of pixels
int image_band2_count(const sceImage& _src1, const sceImage& _src2,sceImage& _dest,
 unsigned int low1, unsigned int high1, unsigned int low2, unsigned int high2);

// *************EDGE FILTERS*******************

// computes gradients of an image using a very simple [-1 1] operator
// image had better be pretty smooth!

// result is offset 0.5 pixels to left
void gradientX(const sceImage&_src, sceImage& _dest);

// result is offset 0.5 pixels up
void gradientY(const sceImage&_src, sceImage& _dest);


// computes absolute value of laplacian [ 0 -1 0; -1 4 -1; 0 -1 0] by left 
// shifting data, result offset up and to left 1 pixel
void edge(const sceImage& _src, sceImage& _dest);

// computes laplacian [0 0 -1 0 0;0 0 0 0 0;-1 0 4 0 -1;0 0 0 0 0;0 0 -1 0 0]
// by right shifting data (so 16-bit data pixels are okay), result is offset
// up and left 2 pixels
void laplacian(const sceImage& _src, sceImage& _dest);


// computes signum of laplacian (same laplacian as laplacian filter above)
void slaplacian(const sceImage& _src, sceImage& _dest);

// tracking-specific filter
void edgeY(const sceImage& _src, sceImage& _dest);

// various logical operations
void image_or(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);
void image_and(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);
void image_xor(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);
void image_nor(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);

// *************MORPHOLOGICAL FILTERS*******************

// dilate and erode are morphological operators, really only work
// for binary (0,0xff) data.

// dilate: Result 0xff if at least COUNT neighbors are 0xff
// dilate: Result 0xff if at least COUNT neighbors are 0xff


// Result offset up and left 1 pixel
void dilate(const sceImage& _src, sceImage& _dest, unsigned short count=3);
// Result offset down and right 1 pixel
void dilate2(const sceImage& _src, sceImage& _dest,unsigned short count=3);

// Result offset up and left 1 pixel
void erode(const sceImage& _src, sceImage& _dest, unsigned short count=3);
// Result offset down and right 1 pixel
void erode2(const sceImage& _src, sceImage& _dest,unsigned short count=3);


// *************MISCELLANEOUS FILTERS*******************

// counts the number of pixels between two threshold values
unsigned int bandcount(const sceImage& _src, int lo, int hi, const RECT&w);

// blends two images according to the formula:
// result = (src1 + (src2<<C) - src2) >> C
void blend(const sceImage& _src1, const sceImage& _src2, sceImage& _dest, int blendcoeff=4);

// achieves embossing effect by subracting image from itself (with offset)
void emboss(const sceImage& _src, sceImage& _dest);

// absolute value of difference of two images
void motion(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);
void motion3(const sceImage& _src1, const sceImage& _osrc1, 
	     const sceImage& _src2, const sceImage& _osrc2, 
	     const sceImage& _src3, const sceImage& _osrc3, 
	     sceImage& _dest);

// value of difference of two images
void smotion(const sceImage16& _src1, const sceImage16& _src2, sceImage16& _dest);

// add pixels from two images
void image_add(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);

// subtract pixels from two images
void image_sub(const sceImage& _src1, const sceImage& _src2, sceImage& _dest);

// multiply pixels by power of 2, like << (pos amount) or >> (neg amount)
void image_binshift(const sceImage& _src, sceImage& _dest, int amount);

// totals the values of all the pixels in an image (note the return value is
// only 32-bit)
unsigned int image_sum(const sceImage& _src);

unsigned int image_sum(const sceImage& _src,const RECT&rect);

// counts the number of nonzero pixels in an image
unsigned int image_count(const sceImage& _src);

void deblock(sceImage& _i);


#if 0
// counts the number of nonzero pixels in an image mask and also totals the value of all the pixels in an image whose corresponding mask pixel is nonzero (note the return values are
// only 32-bit)
void image_maskedsum(const sceImage16& _src, const sceImage16& _mask, int& return_sum, int& return_masksum);



void composite(const sceImage16& _src1, const sceImage16& _src2, const sceImage16& alpha, sceImage16& _dest);
#endif
#endif


