/* Copyright (C) 2002 Sony Computer Entertainment America */
/*        All Rights Reserved.  SCEA Confidential         */

#ifndef LIBVIDEO_H
#define LIBVIDEO_H

// NOTE:  Only a limited number of combinations of sources, formats, sizes, 
// framerates are actually implemented.  sceVideo::init() will point out
// unimplemented combinations.

#include "libimage.h"
class sceVideoStream;

class sceVideo {
public:
  // constructors/destructors
  sceVideo();
  ~sceVideo();

  typedef enum {
    UNKNOWN,
    SVIDEO,     // T-10000 s-video input
    FILE,       // extensions .ipu and .mpg/.mpeg supported
    EYETOY,     // PS2 EyeToy (IPU bitstream)
  } SceVideoSource;

  // image format for attach/acquire
  typedef enum {
    YCbrCr422, Y, CbCr411, YCbCr444, YCbCr411, RGB, YCbCr111,
    RGB16,
    RGBA32,
    YCbCrA32, // for svideo only
  } SceVideoFormat;

  typedef enum {
    STANDARD_RES, // 320x240
    LOW_RES,      // 160x112
    HIGH_RES,     // 640x480
    HIGH_RES_X,   // 640x240
    HIGH_RES_Y,   // 320x480
  } SceVideoSize;

  typedef enum {
    NO_CAMERA,      // camera is not currently plugged in (EyeToy only)
    UNINITIALIZED,  // video stream has not been initialized
    STOPPED,        // video stream is initialized but not streaming
    STARTED,        // video is streaming
  } SceVideoStatus;

  // accessors
  SceVideoSource source() const { return _source;}
  SceVideoFormat format() const { return _format;}
  sceImage::SceImageType type() const;
  SceVideoSize size() const { return _size;}
  int framerate() const { return _framerate;} // frames per second
  int width() const;
  int height() const;
  int widthColor() const;
  int heightColor() const;
  int frame() const; // current frame number
  int maxsize_bitstream() const; // only nonzero for EyeToy source.  Maximum size the bitstream can ever be.

  const char* name() const { return _name;}  // for FILE video source

  // Methods
  // Check the status of the video stream.
  SceVideoStatus status() const;

  // Initialize the video stream.  
  // A negative number indicates a problem (need to check status())
  int init(SceVideoSource src=EYETOY, SceVideoFormat fmt=RGB16, 
	   SceVideoSize size=STANDARD_RES, int framerate=60);

  // For EYETOY sources, this begins the streaming process on the IOP.
  // Compressed video frames will be obtained and copied to EE memory.
  // For SVIDEO sources, this begins the GS acquisition process.
  // Video frames will be in GS memory.
  void start();

  // Copy bitstream for a frame of video.  For EyeToy, use this to
  // have more control over decompression, etc. than attach/acquire.
  // Returns frame number if  successful, negative otherwise (check status()).
  int get_bitstream(const void* bitstream);

  // Attach acquisition images to video stream
  // If necessary, the attached images are resized to match video size.
  void attach(sceImage* y,sceImage* cb,sceImage* cr);
  void attach(sceImage* image); // for rgba, rgb16

  // Fill attached images with a frame of video.  
  // Returns frame number if  successful, negative otherwise (check status()).
  // Decompression and/or color conversion, if necessary, are performed here.
  int acquire();

  // Stop video streaming.  Only applicable to EYETOY and SVIDEO sources.
  void stop();

  // For EYETOY sources, this begins the mic streaming process 
  // on the IOP.  Mic audio will be obtained and copied to EE memory.
  void start_mic();

  // For EYETOY sources, this copies up to maxbytes of mic data to buf.
  // The # of bytes actually copied is the return value.
  // If maxbytes is passed as 0, this acts as a query to see how much
  // data is available.
  int read_mic(void*buf, int maxbytes);

  // Stop mic streaming.  Only applicable to EYETOY.
  void stop_mic();

  // necessary for FILE video source
  void setName(const char*name); 

  // setAttribute() only supported for EYETOY.  Note if waitflag is 
  // true (default), this blocks until the IOP successfully returns,
  // which may be a substantial amount of time.  But this is necessary
  // to avoid "missing" the setting of an attribute.
  typedef enum {
    GAIN=0,           // 0..255 default auto
    REDBLUE_GAIN,     // 0..255 default auto
    REDBLUE_BIAS,     // 0..255 default auto
    SATURATION,       // 0..255, default 0xa4
    EXPOSURE,         // 0..255 default auto
    BRIGHTNESS,       // 0..255 default 0x60
    AEC,              // automatic exposure control, 0 or 1, default 1
    AGC,              // automatic gain control, 0 or 1, default 1
    AWB,              // automatic white balance, 0 or 1, default 1
    ABC,              // automatic brightness, 0 or 1, default 0
    LED,              // param1=frames on, param2=frames off, default 0,0
    AUDIO_GAIN,       // 0..7, default 5
    QS,               // minimum compression level 0..30, default 0, compression always auto-adjusts if insufficient bandwidth
    NONZERO_COEFFS,   // lum 0..64,chrom 0..64 default 40,40. See Note below!
    YUV_FLAG,         // 0 for YCbCr, 1 for YUV.  See Note below!
    JPEG_BITSTREAM_FLAG, // 0 for IPU stream, 1 for JPEG stream
    BACKLIGHT_COMP,   // 0 for no backlight comp, 1 for backlight comp
  } SceVideoAttribute;
  void setAttribute(SceVideoAttribute a, int val1, int val2, bool wait_flag=true);

  // Notes to Experts

  // NONZERO_COEFFS sets how many of the block coeffs are allowed to be
  // nonzero.  0 means all coeffs zeroed (no data) 1 means keep DC value only.
  // 64 means do not zero any of the coeffs.  Can be used to balance the 
  // bandwidth dedicated to lum vs. chrom.  This parameter also has other
  // ramifications in IPU mode because the worst-case macroblock for
  // compression can dictate the required overall compression ratio, so by 
  // forcing some of the high-order coefficients to zero, we limit how bad 
  // the worst-case block can ever be.

  // YUV mode does not scale down chrominance.  This means it is better
  // for color tracking.  Also, packYUV() is slightly faster than packYCbCr()

protected:
  SceVideoSource _source;
  SceVideoFormat _format;
  SceVideoSize _size;
  int _framerate;
  char *_name; // only used for FILE source

  sceVideoStream* _vidstream;
};


#endif // LIBVIDEO_H
