//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Input.cpp
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Implementation of input controllers manager
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Input/Asura_PS2_Input.h"

// Stdlib platform-specific.
#include <eetypes.h>
#include <libmtap.h>
#include <libpad.h>
#include <libsn.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"
#include "Input/Asura_PlatformInput.h"

// Engine platform-specific.
#include "ProjectSettings/Asura_PS2_ProjectControllerDefines.h"
#include "Input/Asura_PS2_ControlPortDevice.h"
#include "Input/Asura_PS2_USBDevice.h"

#if ASURA_PS2_SUPPORT_KEYBOARD
    #include "Input/Asura_PS2_USBKeyboard.h"
    Asura_PS2_USBKeyboard* Asura_PS2_Input::s_pxUSBKeyboard = NULL;
#endif

#if ASURA_PS2_SUPPORT_MOUSE
    #include "Input/Asura_PS2_USBMouse.h"
    Asura_PS2_USBMouse* Asura_PS2_Input::s_pxUSBMouse = NULL;
#endif

#if ASURA_PS2_SUPPORT_GUNCON2
    #include "Input/Asura_PS2_USBGuncon2.h"
    Asura_PS2_USBGuncon2* Asura_PS2_Input::s_pxGuncon2[ASURA_PS2_MAX_GUNCON2S] = {NULL, NULL};
#endif

Asura_PS2_ControlPortDevice* Asura_PS2_Input::s_apxControlPortDevices[ASURA_PS2_MAX_CONTROLLERS];

float*	Asura_PS2_Input::s_pfAsuraAxis;
u_char* Asura_PS2_Input::s_pucAsuraCurrentInput;
bool	Asura_PS2_Input::s_bPressureEnabled = false;

#if ASURA_PS2_SUPPORT_MULTITAP
bool Asura_PS2_Input::s_abMultitap[ASURA_PS2_MAX_MULTITAPS];
#endif

void Asura_PS2_Input::Update(void)
{
	UpdateControlPortDevices();
	UpdateUSBDevices();
}

/********************************************************************************************************/

#if ASURA_PS2_SUPPORT_MULTITAP
void Asura_PS2_Input::CheckForMultitap(int iPort)
{
	Asura_Assert(iPort < ASURA_PS2_MAX_MULTITAPS, "Port out of range");
	s_abMultitap[iPort] = sceMtapGetConnection(iPort);
}

bool Asura_PS2_Input::MultitapAvailable(int iPort)
{
	if (iPort < ASURA_PS2_MAX_MULTITAPS)
	{
		return s_abMultitap[iPort];
	}
	return false;
}

#endif

/********************************************************************************************************/

void Asura_PS2_Input::UpdateControlPortDevices(void)
{
	int iCurrentCtrl = 0;
	int iNCtrlDevsAttached = 0;

	// Check all devices in all multitap ports and slots
	for (int iCurrentPort = 0; iCurrentPort < 2; iCurrentPort++)	// two ports
	{
		// Asura only supports 4 devices
		if (iNCtrlDevsAttached >= 4)
			return;

#if ASURA_PS2_SUPPORT_MULTITAP
		// Look for multitap
		if (MultitapAvailable(iCurrentPort))
		{
			// Update any controllers plugged into the multitap
			for (int iCurrentSlot = 0; iCurrentSlot < 4; iCurrentSlot++)
			{
				if (iCurrentSlot == 0 ||
					iCurrentPort < ASURA_PS2_MAX_MULTITAPS)
				{
					// Work out controller number
					iCurrentCtrl = iCurrentPort*4 + iCurrentSlot;
					Asura_Assert(iCurrentCtrl < ASURA_PS2_MAX_CONTROLLERS, "Device out of range");
					// Update will set availability of controller
					s_apxControlPortDevices[iCurrentCtrl]->Update();
					// If available then translate into the buffer
					if(s_apxControlPortDevices[iCurrentCtrl]->Available())
					{
						// Translate into the Asura_Axis buffer, part of Asura_UserInput
						s_apxControlPortDevices[iCurrentCtrl]->Translate(s_pucAsuraCurrentInput + iCurrentCtrl * ASURA_PS2_NUM_BUTTONS_IN_PAD,
																		 s_pfAsuraAxis + iCurrentCtrl * ASURA_PS2_NUM_AXES_IN_PAD);
						iNCtrlDevsAttached++;
					}
				}
			}
		}
		else
#endif
		{
			// Same but for single controller
			iCurrentCtrl = iCurrentPort*4;
			s_apxControlPortDevices[iCurrentCtrl]->Update();
			if(s_apxControlPortDevices[iCurrentCtrl]->Available())
			{
				int iPlayerNumber = GetPlayerNumber(iCurrentCtrl);
				switch(s_apxControlPortDevices[iCurrentCtrl]->GetId())
				{
					case DT_DIGITAL:
					case DT_ANALOG_JSTICK:
					case DT_DUALSHOCK:
					{
						s_apxControlPortDevices[iCurrentCtrl]->Translate(s_pucAsuraCurrentInput + iPlayerNumber * ASURA_PS2_NUM_BUTTONS_IN_PAD,
																		 s_pfAsuraAxis + iPlayerNumber * ASURA_PS2_NUM_AXES_IN_PAD);
					}
					break;
					case DT_GUNCON:
					{
						u_char* pucBuffer = s_pucAsuraCurrentInput;
						pucBuffer += iCurrentPort ? ASURA_PS2_KEYID_GUNCON2_TRIGGER : ASURA_PS2_KEYID_GUNCON1_TRIGGER;
						s_apxControlPortDevices[iCurrentCtrl]->Translate(pucBuffer,
																		 s_pfAsuraAxis + ASURA_AXIS_LIGHTGUN1_H + iCurrentPort * 2);
					}
					break;
					default:
					{
						//NRM - don't translate - for now
					}
				}
				iNCtrlDevsAttached++;
			}
		}
	}
}

int  Asura_PS2_Input::GetNumberOfActiveControllers(bool bInOrder /*=false*/)
{
    int iNumberOfControllers = 0;
    // (NRM - 20.05.2003) this is the "we don't care what order we're in" case
    if(!bInOrder)
    {
        for(int iCurrentPort = 0; iCurrentPort < 2; ++iCurrentPort)
        {
#if ASURA_PS2_SUPPORT_MULTITAP
            for(int iCurrentSlot = 0; iCurrentSlot < 4; ++iCurrentSlot)
            {
				if (iCurrentSlot != 0 &&
					iCurrentPort >= ASURA_PS2_MAX_MULTITAPS)
				{
					continue;
				}
                int iCurrentCtrl = iCurrentPort * 4 + iCurrentSlot;
#else
            {
                int iCurrentCtrl = iCurrentPort * 4;
#endif
				Asura_Assert(iCurrentCtrl < ASURA_PS2_MAX_CONTROLLERS, "Index out of range");
                if(s_apxControlPortDevices[iCurrentCtrl] &&
                    s_apxControlPortDevices[iCurrentCtrl]->Available())
                {
                    iNumberOfControllers++;
                }
            }
        }
        return iNumberOfControllers;
    }

    // (NRM - 20.05.2003) Make sure controllers are plugged in in order
    for(int iCurrentPort = 0; iCurrentPort < 2; ++iCurrentPort)
    {
#if ASURA_PS2_SUPPORT_MULTITAP
        if (iCurrentPort < ASURA_PS2_MAX_MULTITAPS &&
			s_abMultitap[iCurrentPort])
        {
            for(int iCurrentSlot = 0; iCurrentSlot < 4; ++iCurrentSlot)
            {
                int iCurrentCtrl = iCurrentPort * 4 + iCurrentSlot;
                if(s_apxControlPortDevices[iCurrentCtrl] &&
                    s_apxControlPortDevices[iCurrentCtrl]->Available())
                {
                    iNumberOfControllers++;
                }
                // Check that enough controllers have been added and in order
                if(iNumberOfControllers <= iCurrentCtrl)
                {
                    return iNumberOfControllers;
                }
            }
        }
        else
        {
            if(s_apxControlPortDevices[iCurrentPort * 4] &&
                s_apxControlPortDevices[iCurrentPort * 4]->Available())
            {
                iNumberOfControllers++;
            }

            if(iNumberOfControllers <= iCurrentPort)
            {
                return iNumberOfControllers;
            }
        }
#else
        if(s_apxControlPortDevices[iCurrentPort * 4] &&
            s_apxControlPortDevices[iCurrentPort * 4]->Available())
        {
            iNumberOfControllers++;
        }

        if(iNumberOfControllers <= iCurrentPort)
        {
            return iNumberOfControllers;
        }

#endif
    }
    return iNumberOfControllers;
}

bool Asura_PS2_Input::IsControllerActive(int iControllerNum)
{
#if ASURA_PS2_SUPPORT_MULTITAP
    if(s_abMultitap[0])
    {
        Asura_PS2_ControlPortDevice* pxControlPortDevice = GetControlPortDevice(iControllerNum);
        return (pxControlPortDevice && pxControlPortDevice->Available());
    }
    else
#endif
    {
        Asura_PS2_ControlPortDevice* pxControlPortDevice = GetControlPortDevice(iControllerNum * 4);
        return (pxControlPortDevice && pxControlPortDevice->Available());
    }
}

int Asura_PS2_Input::GetPlayerNumber(int iControllerNum)
{
	if (iControllerNum < 4 ||
		MultitapAvailable(0))
	{
		return iControllerNum;
	}
	else
	{
		return iControllerNum - 3;
	}
}

#if ASURA_PS2_SUPPORT_GUNCON2
void Asura_PS2_Input::UpdateGuncon2(int iDeviceNumber)
{
    s_pxGuncon2[iDeviceNumber]->Update();
}

void Asura_PS2_Input::TranslateGuncon2(int iDeviceNumber)
{
    // call twice for two player games
    if(s_pxGuncon2[iDeviceNumber]->Available())
    {
        s_pxGuncon2[iDeviceNumber]->Translate(s_pucAsuraCurrentInput, s_pfAsuraAxis);
    }
}

#endif //ASURA_PS2_SUPPORT_GUNCON2

void Asura_PS2_Input::UpdateUSBDevices(void)
{
#if ASURA_PS2_SUPPORT_MOUSE
    s_pxUSBMouse->Update();
    if(s_pxUSBMouse->Available())
    {
        s_pxUSBMouse->Translate(s_pucAsuraCurrentInput, s_pfAsuraAxis);
    }
#endif

#if ASURA_PS2_SUPPORT_KEYBOARD
    s_pxUSBKeyboard->Update();
    if(s_pxUSBKeyboard->Available())
    {
        s_pxUSBKeyboard->Translate(s_pucAsuraCurrentInput, s_pfAsuraAxis);
    }
#endif    
}

/********************************************************************************************************/

void Asura_PS2_Input::PreInitialise(void)
{
	// Initialize and open multi tap ports
#if ASURA_PS2_SUPPORT_MULTITAP
	sceMtapInit();
	for (int i=0;i<ASURA_PS2_MAX_MULTITAPS;i++)
	{
		s_abMultitap[i] = false;
		sceMtapPortOpen(i);
	}
#endif

	// Initialize and open libpad
	scePadInit(0);

	// Try and open all the ports
	int iRetVal = 0;
	int iSlot = 0;
	for(int iPort=0; iPort<2; ++iPort)
	{
#if ASURA_PS2_SUPPORT_MULTITAP
		for(iSlot = 0; iSlot < 4; ++iSlot)
		{
			if (iSlot != 0 &&
				iPort >= ASURA_PS2_MAX_MULTITAPS)
			{
				continue;
			}
#else
		{
#endif
			iRetVal = scePadPortOpen(iPort, iSlot, Asura_PS2_ControlPortDevice::GetMemoryOfBuffer(iPort, iSlot));
			Asura_Assert(iRetVal == 1, "Could not open controller ports");
		}
	}

#if _DEBUG
	// (MH - 31.07.2003) in debug, sitting on a breakpoint causes these messages.
	if (snIsDebuggerRunning())
	{
		scePadSetWarningLevel(0);
	}
#endif
}

/********************************************************************************************************/

void Asura_PS2_Input::Initialise(u_char *pucAsuraCurrentInput, float *pfAsuraAxis)
{
	// Assign asura input and axis buffers
	s_pucAsuraCurrentInput = pucAsuraCurrentInput;
	s_pfAsuraAxis = pfAsuraAxis;

	// Create up to eight new ControlPortDevices (one for each multitap position
	for (int iCurrentPort = 0; iCurrentPort < 2; iCurrentPort++)
	{
#if ASURA_PS2_SUPPORT_MULTITAP
		for(int iCurrentSlot = 0; iCurrentSlot < 4; iCurrentSlot++)
		{
			if (iCurrentSlot != 0 &&
				iCurrentPort >= ASURA_PS2_MAX_MULTITAPS)
			{
				continue;
			}
#else
		int iCurrentSlot = 0;
		{
#endif
		
			s_apxControlPortDevices[iCurrentPort * 4 + iCurrentSlot] = new Asura_PS2_ControlPortDevice(iCurrentPort, iCurrentSlot, s_bPressureEnabled);
		}
	}

    // Initialise the USB devices
    Asura_PS2_USBDevice::Initialise();

#if ASURA_PS2_SUPPORT_GUNCON2
    s_pxGuncon2[0] = new Asura_PS2_USBGuncon2(0);
    s_pxGuncon2[0]->Initialise();
    s_pxGuncon2[1] = new Asura_PS2_USBGuncon2(1);
    s_pxGuncon2[1]->Initialise();
#endif

#if ASURA_PS2_SUPPORT_MOUSE
    s_pxUSBMouse = new Asura_PS2_USBMouse(0);
    s_pxUSBMouse->Initialise();
#endif

#if ASURA_PS2_SUPPORT_KEYBOARD
    s_pxUSBKeyboard = new Asura_PS2_USBKeyboard(0);
    s_pxUSBKeyboard->Initialise();
#endif
}

/********************************************************************************************************/

void Asura_PS2_Input::Shutdown(void)
{
	// Delete all controllers created
	for (int iCurrentPort = 0; iCurrentPort < 2; iCurrentPort++)
	{
#if ASURA_PS2_SUPPORT_MULTITAP
		for(int iCurrentSlot = 0; iCurrentSlot < 4; iCurrentSlot++)
		{
			if (iCurrentSlot != 0 &&
				iCurrentPort >= ASURA_PS2_MAX_MULTITAPS)
			{
				continue;
			}
#else
		int iCurrentSlot = 0;
		{
#endif
			delete s_apxControlPortDevices[iCurrentPort * 4 + iCurrentSlot];
		}
	}

	// Close multi-tap ports
#if ASURA_PS2_SUPPORT_MULTITAP
	for (int i=0;i<ASURA_PS2_MAX_MULTITAPS;i++)
	{
		sceMtapPortClose(i);
	}
#endif

	// Delete all USB devices created

#if ASURA_PS2_SUPPORT_GUNCON2
    delete s_pxGuncon2[0];
    s_pxGuncon2[0] = NULL;
    delete s_pxGuncon2[1];
    s_pxGuncon2[1] = NULL;
#endif

#if ASURA_PS2_SUPPORT_MOUSE
    delete s_pxUSBMouse;
    s_pxUSBMouse = NULL;
#endif

#if ASURA_PS2_SUPPORT_KEYBOARD
    delete s_pxUSBKeyboard;
    s_pxUSBKeyboard = NULL;
#endif

	// Close libpad
	scePadEnd();
}

/********************************************************************************************************/

void Asura_PS2_Input::SetPressureSensitivity(bool bEnable)
{
	s_bPressureEnabled = bEnable;
}

/********************************************************************************************************/

Asura_PS2_ControlPortDevice* Asura_PS2_Input::GetControlPortDevice(int iControlPortNumber)
{
	Asura_Assert(iControlPortNumber >= 0 && iControlPortNumber < ASURA_PS2_MAX_CONTROLLERS, 
				 "iControlPortNumber out of range in Asura_PS2_Input::GetControlPortDevice");
	return s_apxControlPortDevices[iControlPortNumber];
}


#if ASURA_PS2_SUPPORT_GUNCON2
Asura_PS2_USBGuncon2* Asura_PS2_Input::GetUSBGuncon2(int iDeviceNumber)
{
    Asura_Assert(iDeviceNumber >= 0 && iDeviceNumber < ASURA_PS2_MAX_GUNCON2S,
                 "invalid Guncon2 port");
    return(s_pxGuncon2[iDeviceNumber]);
}

bool Asura_PS2_Input::Guncon2Available(int iDeviceNumber)
{
    Asura_Assert(iDeviceNumber >= 0 && iDeviceNumber < ASURA_PS2_MAX_GUNCON2S,
                 "invalid Guncon2 port");
    return(s_pxGuncon2[iDeviceNumber]->Available());
}
#endif

#if ASURA_PS2_SUPPORT_MOUSE
bool Asura_PS2_Input::MouseAvailable(void)
{
    return(s_pxUSBMouse &&
            s_pxUSBMouse->Available());
}
#endif

#if ASURA_PS2_SUPPORT_KEYBOARD
bool Asura_PS2_Input::KeyboardAvailable(void)
{
    return(s_pxUSBKeyboard &&
            s_pxUSBKeyboard->Available());
}
#endif

bool Asura_PS2_Input::IsKeyToggled(int iKey)
{
#if ASURA_PS2_SUPPORT_KEYBOARD
    if (s_pxUSBKeyboard &&
            s_pxUSBKeyboard->Available())
    {
        return s_pxUSBKeyboard->IsKeyToggled(iKey);
    }
    else
#endif
    {
        return false;
    }
}
