//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_USBGuncon2.cpp
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Derived class to handle USB Namco Guncon2 controller
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Input/Asura_PS2_USBGuncon2.h"

#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"
#include "Input/Asura_PlatformInput.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Input/SCE_USB_Guncon2.h"

const int GUNCON2_CENTER_X = 0x1a6;			// Gun hardware centre
const int GUNCON2_CENTER_Y_PAL = 0x0a8;
const int GUNCON2_CENTER_Y_NTSC = 0x086;

const int GUNCON2_ADJ_X2 = 0x55;
const int GUNCON2_ADJ_Y2_PAL = 0x78;
const int GUNCON2_ADJ_Y2_NTSC = 0x69;

const int GUNCON2_DEFAULT_ADJ_H	= 0xb3;
const int GUNCON2_DEFAULT_ADJ_V_PAL	= 0x3c;
const int GUNCON2_DEFAULT_ADJ_V_NTSC = 0x3e;

enum Guncon2BufferLocations
{
	GC2_DATA_SIZE           = 0,	//[0]:Gun data size n (0 when blank)
	GC2_DIGITAL_BUTTON_HI   = 1,	//[1]:SW0
	GC2_DIGITAL_BUTTON_LO   = 2,	//[2]:SW1
    GC2_X_COORD_LO		    = 3,	//[3]:X(LOWER8bit)
	GC2_X_COORD_HI		    = 4,	//[4]:X(UPPER8bit)
	GC2_Y_COORD_LO		    = 5,	//[5]:Y(LOWER8bit)
	GC2_Y_COORD_HI		    = 6,	//[6]:Y(UPPER8bit)
    GC2_NUM_GCON2S_CONNECTED= 9,    // Extra byte added by MH to return the number of Guncon2s currently connected
};

enum Guncon2PadMasksHi
{
	GC2_LEFT_MASK		  = 1<<7,
	GC2_DOWN_MASK		  = 1<<6,
	GC2_RIGHT_MASK		  = 1<<5,
	GC2_UP_MASK			  = 1<<4,
	GC2_A_MASK			  = 1<<3,
	GC2_B_MASK			  = 1<<2,
	GC2_C_MASK			  = 1<<1,
};

enum Guncon2PadMasksLo
{
	GC2_START_MASK		  = 1<<7,
	GC2_SELECT_MASK		  = 1<<6,
	GC2_TRIGGER_MASK	  = 1<<5,
	GC2_HDD_MASK		  = 1
};

/****************************** S T A T I C   M E M B E R S ******************************/
ASURA_PS2_USB_PHASE Asura_PS2_USBGuncon2::s_eCurrentPhase   = USB_PHASE_NO_DEVICES;
ASURA_PS2_USB_PHASE Asura_PS2_USBGuncon2::s_eNextPhase      = USB_PHASE_NO_DEVICES;

u_char Asura_PS2_USBGuncon2::s_aucGunconDeviceBuffer[GUNCON_DEVICE_BUFFER_SIZE] __attribute__((aligned(64)));
u_char Asura_PS2_USBGuncon2::s_aucGunconLocationBuffer[GUNCON_LOCATION_BUFFER_SIZE] __attribute__((aligned(64)));
u_char Asura_PS2_USBGuncon2::s_aucGunconTranslateData[GUNCON_DEVICE_BUFFER_SIZE] __attribute__((aligned(64)));

int Asura_PS2_USBGuncon2::s_iNumberOfGuncon2s = 0;
int Asura_PS2_USBGuncon2::s_iCurrentGuncon = 0;
bool g_bGunconAvailable[2] = {false, false};
int Asura_PS2_USBGuncon2::s_iGunconOffset[2] = {0, 1}; // (NRM - 12.12.2003) these can be set by the user during calibration

/*****************************************************************************************/

Asura_PS2_USBGuncon2::Asura_PS2_USBGuncon2(int iPortNumber)
					 :Asura_PS2_USBDevice(iPortNumber)
{
	m_xDeviceType               = USB_DT_GUNCON2;
    m_pucGunconReadData		    = reinterpret_cast<char*>(Asura_AllocateAlign(GUNCON_READ_DATA_SIZE, ASURA_ALIGN_64_BYTE));
	m_xCalibrateData.m_iMonitorType = 0;
	m_xCalibrateData.m_iHdsw        = 0x01;
	m_xCalibrateData.m_fXOffset = 0.0f;
	m_xCalibrateData.m_fYOffset = 0.0f;
	m_xCalibrateData.m_iH       = GUNCON2_DEFAULT_ADJ_H;
	m_xCalibrateData.m_iV       = Asura_PS2_GS::VideoModeIsPal() ? GUNCON2_DEFAULT_ADJ_V_PAL : GUNCON2_DEFAULT_ADJ_V_NTSC;
}

Asura_PS2_USBGuncon2::~Asura_PS2_USBGuncon2(void)
{
    Asura_DeallocateAlign(m_pucGunconReadData);
}

void Asura_PS2_USBGuncon2::Update(void)
{
    // Check to see if static state machine needs updating
    if(m_iDeviceNumber == 0)
    {
        UpdateAllGuncon2s();
    }
    if(s_eCurrentPhase == USB_PHASE_READY)
    {
        if(m_iDeviceNumber >= 0 && Available())
        {
            if(nmUsbGunRead(m_iDeviceNumber, m_pucGunconReadData) != 0)
	        {
                s_eCurrentPhase = USB_PHASE_BUSY;
		        s_eNextPhase    = USB_PHASE_INITIALISING;
            }
        }
    }
}

void Asura_PS2_USBGuncon2::Translate(u_char *piAsuraInputBufferOffset, float *piAsuraAxis_BufferOffset)
{
    // (NRM - 18.09.2003) move sync to just before translate to try and use some natural latency...
    nmUsbGunSync(0);
    if (s_iNumberOfGuncon2s &&                          // Check at least one plugged in
         s_iNumberOfGuncon2s != m_pucGunconReadData[GC2_NUM_GCON2S_CONNECTED])  // Check whether the number has changed
    {
        s_iNumberOfGuncon2s = 0;
		s_eCurrentPhase = USB_PHASE_BUSY;
		s_eNextPhase    = USB_PHASE_ENUMERATING;
        return;
    }
    s_iNumberOfGuncon2s = m_pucGunconReadData[GC2_NUM_GCON2S_CONNECTED];


	int	GUNCON2_ADJ_X1 = Asura_Viewport::GetWidth()  / 2;
	int	GUNCON2_ADJ_Y1 = Asura_Viewport::GetHeight() / 2;

	// Translate UDLR
	int iStartLocation = s_iGunconOffset[m_iDeviceNumber] * ASURA_PS2_NUM_BUTTONS_IN_PAD;
	*(piAsuraInputBufferOffset + iStartLocation + ASURA_PS2_KEYID_PAD1_UP)	 |= 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_UP_MASK) == 0);
	*(piAsuraInputBufferOffset + iStartLocation + ASURA_PS2_KEYID_PAD1_DOWN) |= 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_DOWN_MASK) == 0);
	*(piAsuraInputBufferOffset + iStartLocation + ASURA_PS2_KEYID_PAD1_LEFT) |= 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_LEFT_MASK) == 0);
	*(piAsuraInputBufferOffset + iStartLocation + ASURA_PS2_KEYID_PAD1_RIGHT)|= 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_RIGHT_MASK)== 0);

	// START and SELECT
	*(piAsuraInputBufferOffset + iStartLocation + ASURA_PS2_KEYID_PAD1_START)  |= 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_LO] & GC2_START_MASK) == 0);
	*(piAsuraInputBufferOffset + iStartLocation + ASURA_PS2_KEYID_PAD1_SELECT) |= 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_LO] & GC2_SELECT_MASK) == 0);

	// Translate A, B, C and trigger
	iStartLocation = s_iGunconOffset[m_iDeviceNumber] == 1 ? ASURA_PS2_KEYID_GUNCON2_TRIGGER 
                                                           : ASURA_PS2_KEYID_GUNCON1_TRIGGER;
	int iTrigger, iAButton, iBButton, iCButton;
	iTrigger = 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_LO] & GC2_TRIGGER_MASK) == 0);
	iAButton = 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_A_MASK) == 0);
	iBButton = 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_B_MASK) == 0);
	iCButton = 255 * ((m_pucGunconReadData[GC2_DIGITAL_BUTTON_HI] & GC2_C_MASK) == 0);

	*(piAsuraInputBufferOffset + iStartLocation) |= iTrigger;
	*(piAsuraInputBufferOffset + iStartLocation + 1) |= iAButton;
	*(piAsuraInputBufferOffset + iStartLocation + 2) |= iBButton;
	*(piAsuraInputBufferOffset + iStartLocation + 3) |= iCButton;

	// Translate the axis
	iStartLocation = s_iGunconOffset[m_iDeviceNumber] == 1 ? ASURA_AXIS_LIGHTGUN2_H 
                                                           : ASURA_AXIS_LIGHTGUN1_H;

	int iScreenW = Asura_Viewport::GetWidth();
	int iScreenH = Asura_Viewport::GetHeight();
	int iGunX, iGunY;
	float fGunX, fGunY;

	iGunX = (m_pucGunconReadData[GC2_X_COORD_LO] &0xFF)
		  + (m_pucGunconReadData[GC2_X_COORD_HI] << 8);

	iGunY = (m_pucGunconReadData[GC2_Y_COORD_LO] &0xFF)
		  + (m_pucGunconReadData[GC2_Y_COORD_HI] << 8);

	// Check for invalid co-ordinate
	if(iGunX == 0 || iGunX == 1)
	{
		iGunX = 
		iGunY = -1;
	}

	int   iOffsetX = (iGunX - GUNCON2_CENTER_X) * (GUNCON2_ADJ_X1 - GUNCON2_ADJ_X2) / m_xCalibrateData.m_iH
					  + GUNCON2_ADJ_X1;

	int GUNCON2_CENTER_Y = Asura_PS2_GS::VideoModeIsPal() ? GUNCON2_CENTER_Y_PAL : GUNCON2_CENTER_Y_NTSC;
	int GUNCON2_ADJ_Y2 = Asura_PS2_GS::VideoModeIsPal() ? GUNCON2_ADJ_Y2_PAL : GUNCON2_ADJ_Y2_NTSC;
  	int   iOffsetY = (iGunY - GUNCON2_CENTER_Y) * (GUNCON2_ADJ_Y1 - GUNCON2_ADJ_Y2) / m_xCalibrateData.m_iV
					  + GUNCON2_ADJ_Y1;

	float fScreenX = static_cast<float>(iOffsetX) / Asura_Viewport::GetWidth();
	float fScreenY = static_cast<float>(iOffsetY) / Asura_Viewport::GetHeight();

    // Calibrate
	fScreenX -= m_xCalibrateData.m_fXOffset;
	fScreenY -= m_xCalibrateData.m_fYOffset;
	
	*(piAsuraAxis_BufferOffset + iStartLocation) = fScreenX;
	*(piAsuraAxis_BufferOffset + iStartLocation + 1) = fScreenY;
}

bool Asura_PS2_USBGuncon2::Available(void)
{
	return (g_bGunconAvailable[m_iDeviceNumber]) && 
            (s_eCurrentPhase == USB_PHASE_READY);
}

void Asura_PS2_USBGuncon2::Initialise(void)
{
    // Set initial phase only - everything else handled by state machine below
    s_eCurrentPhase     = USB_PHASE_BUSY;
    s_eNextPhase        = USB_PHASE_NO_DEVICES;
    s_iNumberOfGuncon2s = 0;
}

/******************************** S T A T I C   M E M B E R S ********************************/

void Asura_PS2_USBGuncon2::UpdateAllGuncon2s(void)
{
    int iGunconReturnVal = -1;
	switch(s_eCurrentPhase)
	{
        case USB_PHASE_NONE:
			s_eCurrentPhase = USB_PHASE_BUSY;
			s_eNextPhase    = USB_PHASE_INITIALISING;
        break;

        case USB_PHASE_INITIALISING:
			s_eCurrentPhase = USB_PHASE_BUSY;
			s_eNextPhase    = USB_PHASE_ENUMERATING;
        break;

        case USB_PHASE_ENUMERATING:
			iGunconReturnVal = nmUsbGunGetInfo(s_aucGunconDeviceBuffer);
			s_eCurrentPhase = USB_PHASE_BUSY;
			s_eNextPhase    = USB_PHASE_ENUMERATED;
        break;

        case USB_PHASE_ENUMERATED:
			s_iNumberOfGuncon2s = s_aucGunconDeviceBuffer[1];
            g_bGunconAvailable[0] = false;
            g_bGunconAvailable[1] = false;
            s_iCurrentGuncon = 0;
			if(s_iNumberOfGuncon2s)
			{
				s_eCurrentPhase = USB_PHASE_BUSY;
				s_eNextPhase    = USB_PHASE_GET_LOCATIONS;
			}
			else
			{
				s_eCurrentPhase = USB_PHASE_NO_DEVICES;
			}
        break;

        case USB_PHASE_GET_LOCATIONS:
			// See if all locations have been got
			if(s_iCurrentGuncon >= s_iNumberOfGuncon2s)
			{
				s_iCurrentGuncon = 0;
				s_eCurrentPhase = USB_PHASE_BUSY;
				s_eNextPhase    = USB_PHASE_DEVICE_SPECIFIC;
			}
			else
			{
                // Otherwise get the location and go into busy
                // Make sure the current guncon is a valid one
                while(s_aucGunconDeviceBuffer[2+s_iCurrentGuncon] == 0)
                {
                    s_iCurrentGuncon++;
                }
                g_bGunconAvailable[s_iCurrentGuncon] = true;
				nmUsbGunGetLocation(s_iCurrentGuncon, s_aucGunconLocationBuffer);
				s_iCurrentGuncon++;
				s_eCurrentPhase =  USB_PHASE_BUSY;
				s_eNextPhase    =  USB_PHASE_GET_LOCATIONS;
			}
        break;

		case USB_PHASE_BUSY:			    // Wait for the device to be ready but asychronously
			if(nmUsbGunSync(0) == 0)		// 1= non_blocking. Return val : 0= finished, 1= not-finished
			{
				s_eCurrentPhase = s_eNextPhase;
			}
		break;

		case USB_PHASE_ERROR:
			s_eCurrentPhase = USB_PHASE_ENUMERATING;
		break;

		case USB_PHASE_NO_DEVICES:
			s_eCurrentPhase = USB_PHASE_ENUMERATING;
		break;

        case USB_PHASE_CALIBRATE:
        break;

        case USB_PHASE_READY:
        break;

        case USB_PHASE_DEVICE_SPECIFIC:
			if(s_iCurrentGuncon == s_iNumberOfGuncon2s)
			{
				s_eCurrentPhase  = USB_PHASE_BUSY;
				s_eNextPhase     = USB_PHASE_READY;
				s_iCurrentGuncon = 0;
			}
			else
			{
				// See if this is a valid device or not
                for(int i=0; i< ASURA_PS2_MAX_GUNCON2S; ++i)
                {
                    if(i == s_iCurrentGuncon)
                    {
        				TranslateControlData(s_iCurrentGuncon);
                        break;
                    }
                }
				s_iCurrentGuncon++;
				s_eCurrentPhase = USB_PHASE_BUSY;
				s_eNextPhase    = USB_PHASE_DEVICE_SPECIFIC;
			}
        break;
    }
}

void Asura_PS2_USBGuncon2::TranslateControlData(const int iDeviceNumber)
{
	int	iRetVal;

// x offset
	s_aucGunconTranslateData[0] = 0;
	s_aucGunconTranslateData[1] = 0;	// xoff

// y offset
	s_aucGunconTranslateData[2] = 0;		// yoff
	s_aucGunconTranslateData[3] = 0;		// yoff

	s_aucGunconTranslateData[4] = 0;
	s_aucGunconTranslateData[5] = 1;

	iRetVal = nmUsbTransCtrl(iDeviceNumber, s_aucGunconTranslateData);
}

void Asura_PS2_USBGuncon2::SetCalibration(float fData1, float fData2)
{
    m_xCalibrateData.m_fXOffset = fData1;
    m_xCalibrateData.m_fYOffset = fData2;
}

void Asura_PS2_USBGuncon2::SetDeviceNumber(int iPlayer, int iDevice)
{
    Asura_Assert(iPlayer >= 0 && iPlayer < 2, "Illegal player number\n");
    if(!(iPlayer >= 0 && iPlayer < 2))
    {
        return;
    }

    Asura_Assert(iDevice >= 0 && iDevice < 2, "Illegal player number\n");
    if(!(iDevice >= 0 && iDevice < 2))
    {
        return;
    }

    s_iGunconOffset[iDevice] = iPlayer;

    if(s_iGunconOffset[!iDevice] == iPlayer)
    {
        s_iGunconOffset[!iDevice] = !iPlayer;
    }
}
