//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_MemoryCard_SubSystem.cpp
//
// Created by Neil Massey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Sub system for handling state of memory card and command processing
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "MemoryCard/Asura_PS2_MemoryCard.h"
#include "MemoryCard/Asura_PS2_MemoryCard_SubSystem.h"
#include "MemoryCard/Asura_PS2_MemoryCard_Error.h"

#include "Debugging/Asura_Assert.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if _DEBUG
// MRD 07/11/08: Integrating COD changes into Asura ( CoD Change 150608 )
//#define ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
#endif

sceMcTblGetDir Asura_PS2_MemoryCard_SubSystem::s_xTableDir[MAX_SAVE_GAMES] __attribute__((aligned (64)));
Asura_PS2_MemoryCard_State Asura_PS2_MemoryCard_SubSystem::s_xState[2];
int	  Asura_PS2_MemoryCard_SubSystem::s_iNSlots[2];
int	  Asura_PS2_MemoryCard_SubSystem::s_iAsyncFunctionNumber;
int	  Asura_PS2_MemoryCard_SubSystem::s_iAsyncFunctionResult;
Asura_PS2_MemoryCard_Info Asura_PS2_MemoryCard_SubSystem::s_xMemoryCardInfo[2][4];
Asura_PS2_MemoryCard_Command  Asura_PS2_MemoryCard_SubSystem::s_xCommandList[NUMBER_OF_COMMANDS];
int	  Asura_PS2_MemoryCard_SubSystem::s_iCurrentNumberOfCommands;
char* Asura_PS2_MemoryCard_SubSystem::s_pcOutputBuffer;
int	  Asura_PS2_MemoryCard_SubSystem::s_iCurrentFileID;
int   Asura_PS2_MemoryCard_SubSystem::s_iRetryCardCommand = -1;
int   Asura_PS2_MemoryCard_SubSystem::s_iPortToRetry = 0;
int   Asura_PS2_MemoryCard_SubSystem::s_iSlotToRetry = 0;
int	  Asura_PS2_MemoryCard_SubSystem::s_iNumberOfEntries = 0;
const int NUMBER_OF_RETRIES = 3;		// incase of signal noise

void Asura_PS2_MemoryCard_SubSystem::Initialise()
{
	// Initialise the library
	switch(sceMcInit())
	{
		case sceMcIniSucceed:
		break;
		case sceMcIniErrKernel:
			Asura_Assert(0, "Could not initialise memory card system");
		break;
		case sceMcIniOldMcserv:
			Asura_Assert(0, "Old version of mcserv.irx");
		break;
		case sceMcIniOldMcman:
			Asura_Assert(0, "Old version of mcman.irx");
		break;
	}
	s_iCurrentNumberOfCommands = 0;
}

void Asura_PS2_MemoryCard_SubSystem::Update()
{
	switch(s_xState[0])
	{
		case ASURA_PS2_MEMORY_CARD_INIT:
			s_iNSlots[0] = sceMcGetSlotMax(0);
			s_iNSlots[1] = sceMcGetSlotMax(1);	// These are synchronous functions - no need to sync
			s_xState[0]  = ASURA_PS2_MEMORY_CARD_BUSY;	
			s_xState[1]  = ASURA_PS2_MEMORY_CARD_READY;	
			// Add a bunch of GetInfo commands to get the info for each slot
			for(int j=0; j<2; ++j)
			{
				for(int i=0; i<s_iNSlots[j]; ++i)
				{
					AddGetInfoCommand(j, i);
				}
			}
		break;

		case ASURA_PS2_MEMORY_CARD_READY:
		{
			ProcessCommand(GetNextCommand());
		}
		break;

		case ASURA_PS2_MEMORY_CARD_BUSY:
			if(sceMcSync(1, &s_iAsyncFunctionNumber, &s_iAsyncFunctionResult))
			{
				ProcessEndMessage();
				s_xState[0] = s_xState[1];
			}
		break;
	}
}

void Asura_PS2_MemoryCard_SubSystem::Shutdown(void)
{
	sceMcEnd();
}

void Asura_PS2_MemoryCard_SubSystem::AddGetInfoCommand(const int iPortNumber, const int iSlotNumber)
{
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoCardInfo;
	xCommand.iParameters[0] = iPortNumber;
	xCommand.iParameters[1] = iSlotNumber;
	s_iPortToRetry = iPortNumber;
	s_iSlotToRetry = iSlotNumber;

	AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_SubSystem::AddCommand(const Asura_PS2_MemoryCard_Command& xCommand)
{
	// Add a command to the queue
	Asura_Assert(s_iCurrentNumberOfCommands < NUMBER_OF_COMMANDS, "Out of space for memory card message");
	s_xCommandList[s_iCurrentNumberOfCommands++] = xCommand;
}

const Asura_PS2_MemoryCard_Command Asura_PS2_MemoryCard_SubSystem::GetNextCommand(void)
{
	// Remove an item from the queue, shuffle the rest down
	Asura_PS2_MemoryCard_Command xCommand;
	if(s_iCurrentNumberOfCommands == 0)
	{
		xCommand.iCommand = 0;
		return xCommand;
	}

	xCommand = s_xCommandList[0];

	for(int i=1; i<s_iCurrentNumberOfCommands; ++i)
	{
		s_xCommandList[i-1] = s_xCommandList[i];
	}
	s_iCurrentNumberOfCommands--;

	return xCommand;
}

/******************************** PROCESS END MESSAGES *****************************************/

void Asura_PS2_MemoryCard_SubSystem::ProcessEndMessage(void)
{
	switch(s_iAsyncFunctionNumber)
	{
		case sceMcFuncNoCardInfo:
		{
			ProcessCardInfoEnd();
		}
		break;
		case sceMcFuncNoOpen:
		{
			ProcessOpenEnd();
		}
		break;
		case sceMcFuncNoClose:
		{
			ProcessCloseEnd();
		}
		break;
		case sceMcFuncNoSeek:
		{
			ProcessSeekEnd();
		}
		break;
		case sceMcFuncNoRead:
		{
			ProcessReadEnd();
		}
		break;
		case sceMcFuncNoWrite:
		{
			ProcessWriteEnd();
		}
		break;
		case sceMcFuncNoFlush:
		{
			ProcessFlushEnd();
		}
		break;
		case sceMcFuncNoMkdir:
		{
			ProcessMkDirEnd();
		}
		break;
		case sceMcFuncNoChDir:
		{
			ProcessChDirEnd();
		}
		break;
		case sceMcFuncNoGetDir:
		{
			ProcessGetDirEnd();
		}
		break;
		case sceMcFuncNoFileInfo:
		{
			ProcessFileInfoEnd();
		}
		break;
		case sceMcFuncNoDelete:
		{
			ProcessDeleteEnd();
		}
		break;
		case sceMcFuncNoFormat:
		{
			ProcessFormatEnd();
		}
		break;
		case sceMcFuncNoUnformat:
		{
			ProcessUnformatEnd();
		}
		break;
		case sceMcFuncNoEntSpace:
		{
			ProcessEntSpaceEnd();
		}
		break;
		case sceMcFuncNoRename:
		{
			ProcessRenameEnd();
		}
		break;
		case sceMcFuncChgPrior:
		{
			ProcessChgPriorEnd();
		}
		break;
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessCardInfoEnd(void)
{
	switch(s_iAsyncFunctionResult)
	{
		case sceMcResSucceed:		// The same Memory Card has been connected continuously since the previous time sceMcGetInfo() was called
		case sceMcResChangedCard:	// Switched to formatted Memory Card
		{
			// both of these cases are a success
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
			printf("Card info got succesfully\n");
#endif
			s_iRetryCardCommand = -1;
            // check that card is actually supported
            if(s_xMemoryCardInfo[s_iPortToRetry][s_iSlotToRetry].iMemoryCardType != sceMcTypePS2)
            {
                s_iAsyncFunctionResult = sceMcResDeniedPermit;
            }
			Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoCardInfo, s_iAsyncFunctionResult);
			return;
		}

		default:
		{
		}
	}

	// Retry in case of signal noise etc.
	if(s_iRetryCardCommand == -1)
	{
		s_iRetryCardCommand = NUMBER_OF_RETRIES + 1;
	}
	if(s_iRetryCardCommand == 0)
	{
		s_iRetryCardCommand = -1;
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoCardInfo, s_iAsyncFunctionResult);
	}
	else
	{
		s_iRetryCardCommand--;
		AddGetInfoCommand(s_iPortToRetry, s_iSlotToRetry);
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessOpenEnd(void)
{
	if(s_iAsyncFunctionResult >= 0)		// Success! Equal to file handle ID
	{
		s_iCurrentFileID = s_iAsyncFunctionResult;
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("File opened succesfully\n");
#endif
		return;
	}
	Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoOpen, s_iAsyncFunctionResult);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessCloseEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoClose, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("File closed succesfully\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessSeekEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoSeek, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("File seeked succesfully\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessReadEnd(void)
{
	if(s_iAsyncFunctionResult >= 0)
	{
		// Success (number of bytes read)
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Data Read\n");
#endif
		return;
	}
	Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoRead, s_iAsyncFunctionResult);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessWriteEnd(void)
{
	if(s_iAsyncFunctionResult >= 0)
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Data written\n");
#endif
		return;
	}
	Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoWrite, s_iAsyncFunctionResult);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessFlushEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoFlush, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Flushed\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessMkDirEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoMkdir, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Mkdir finished\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessChDirEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoChDir, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Directory changed\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessGetDirEnd(void)
{
	if(s_iAsyncFunctionResult >= 0)		// Success! Equal to number of directories (or save games)
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Directory contents got successfully \n");
#endif
		s_pcOutputBuffer[0] = '\0';
		int iStringPos = 1;
		s_pcOutputBuffer[0] = s_iAsyncFunctionResult;	// first byte is number of saves
		s_iNumberOfEntries  = s_iAsyncFunctionResult;
		for(int i = 0; i < s_iAsyncFunctionResult; ++i)
		{
			strcpy(&(s_pcOutputBuffer[iStringPos]), reinterpret_cast<char*>(s_xTableDir[i].EntryName));
			iStringPos += strlen(reinterpret_cast<char*>(s_xTableDir[i].EntryName));
			s_pcOutputBuffer[iStringPos] = '\0';
			iStringPos++;
		}
		return;
	}

	// Not successfull - process the error messages
	Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoGetDir, s_iAsyncFunctionResult);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessFileInfoEnd(void)
{
}

void Asura_PS2_MemoryCard_SubSystem::ProcessDeleteEnd(void)
{
	if(s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoDelete, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("File deleted\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessFormatEnd(void)
{
	if(s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoFormat, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Format complete\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessUnformatEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoFormat, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
		printf("Unformat complete\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessEntSpaceEnd(void)
{
	if(s_iAsyncFunctionResult >= 0)
	{
		int* piOutput = reinterpret_cast<int*>(s_pcOutputBuffer);
		*piOutput = s_iAsyncFunctionResult;
	}
	else
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoFormat, s_iAsyncFunctionResult);
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessRenameEnd(void)
{
	if (s_iAsyncFunctionResult < 0)
	{
		Asura_PS2_MemoryCard_ErrorSystem::ProcessError(sceMcFuncNoFormat, s_iAsyncFunctionResult);
	}
	else
	{
#ifdef ASURA_PS2_SHOW_MEMORY_CARD_COMPLETION_MSGS
			printf("Rename complete\n");
#endif
	}
}

void Asura_PS2_MemoryCard_SubSystem::ProcessChgPriorEnd(void)
{
}

/************************************** PROCESS COMMAND MESSAGES ***************************************/

void Asura_PS2_MemoryCard_SubSystem::ProcessCommand(const Asura_PS2_MemoryCard_Command xCommand)
{
	switch(xCommand.iCommand)
	{
		case 0:
		{
			return;
		}
		case sceMcFuncNoCardInfo:
		{
			ProcessCardInfoCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoOpen:
		{
			ProcessOpenCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoClose:
		{
			ProcessCloseCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoSeek:
		{
			ProcessSeekCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoRead:
		{
			ProcessReadCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoWrite:
		{
			ProcessWriteCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoFlush:
		{
			ProcessFlushCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoMkdir:
		{
			ProcessMkDirCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoChDir:
		{
			ProcessChDirCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoGetDir:
		{
			ProcessGetDirCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoFileInfo:
		{
			ProcessFileInfoCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoDelete:
		{
			ProcessDeleteCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoFormat:
		{
			ProcessFormatCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoUnformat:
		{
			ProcessUnformatCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoEntSpace:
		{
			ProcessEntSpaceCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncNoRename:
		{
			ProcessRenameCommand(xCommand.iParameters);
		}
		break;
		case sceMcFuncChgPrior:
		{
			ProcessChgPriorCommand(xCommand.iParameters);
		}
		break;
	}
	s_xState[0] = ASURA_PS2_MEMORY_CARD_BUSY;
	s_xState[1] = ASURA_PS2_MEMORY_CARD_READY;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessCardInfoCommand(const int* piParameters)
{
	int iCurrentPort = piParameters[0];
	int iCurrentSlot = piParameters[1];
	sceMcGetInfo(iCurrentPort, iCurrentSlot, 
			     &(s_xMemoryCardInfo[iCurrentPort][iCurrentSlot].iMemoryCardType),
				 &(s_xMemoryCardInfo[iCurrentPort][iCurrentSlot].iFreeKiloBytes),
				 &(s_xMemoryCardInfo[iCurrentPort][iCurrentSlot].iFormatted));
	s_iAsyncFunctionNumber = sceMcFuncNoCardInfo;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessOpenCommand(const int* piParameters)
{
	int iCurrentPort = piParameters[0];
	int iCurrentSlot = piParameters[1];
	int iMode = piParameters[2];
	const char* pcFileName = reinterpret_cast<const char*>(&(piParameters[3]));
	sceMcOpen(iCurrentPort, iCurrentSlot, pcFileName, iMode);
	s_iAsyncFunctionNumber = sceMcFuncNoOpen;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessCloseCommand(const int* piParameters)
{
	sceMcClose(s_iCurrentFileID);
	s_iAsyncFunctionNumber = sceMcFuncNoClose;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessSeekCommand(const int* piParameters)
{
    sceMcSeek(s_iCurrentFileID, static_cast<u_int>(piParameters[0]), piParameters[1]);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessReadCommand(const int* piParameters)
{
	int iDataSize		= piParameters[0];
	void* pcDataBuffer  = reinterpret_cast<void*>(piParameters[1]);
	sceMcRead(s_iCurrentFileID, pcDataBuffer, iDataSize);
	s_iAsyncFunctionNumber = sceMcFuncNoRead;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessWriteCommand(const int* piParameters)
{
	int iDataSize		= piParameters[0];
	void* pcDataBuffer  = reinterpret_cast<void*>(piParameters[1]);
	sceMcWrite(s_iCurrentFileID, pcDataBuffer, iDataSize);
	s_iAsyncFunctionNumber = sceMcFuncNoWrite;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessFlushCommand(const int* piParameters)
{
	sceMcFlush(s_iCurrentFileID);
	s_iAsyncFunctionNumber = sceMcFuncNoFlush;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessMkDirCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	const char* pcDirectory = reinterpret_cast<const char*>(&(piParameters[2]));
	sceMcMkdir(iMemoryCardPort, iMemoryCardSlot, pcDirectory);
	s_iAsyncFunctionNumber = sceMcFuncNoMkdir;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessChDirCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	const char* pcDirectory = reinterpret_cast<const char*>(&(piParameters[2]));
	sceMcChdir(iMemoryCardPort, iMemoryCardSlot, pcDirectory, NULL);
	s_iAsyncFunctionNumber = sceMcFuncNoChDir;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessGetDirCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	const char* pcFileFilter = reinterpret_cast<const char*>(&piParameters[2]);
	s_pcOutputBuffer    = reinterpret_cast<char*>(piParameters[31]);
	int iReturn  = sceMcGetDir(iMemoryCardPort, iMemoryCardSlot, pcFileFilter, 0, MAX_SAVE_GAMES, s_xTableDir);
	s_iAsyncFunctionNumber = sceMcFuncNoGetDir;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessFileInfoCommand(const int* piParameters)
{
}

void Asura_PS2_MemoryCard_SubSystem::ProcessDeleteCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	const char* pcFileName = reinterpret_cast<const char*>(&piParameters[2]);
	sceMcDelete(iMemoryCardPort, iMemoryCardSlot, pcFileName);
	s_iAsyncFunctionNumber = sceMcFuncNoDelete;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessFormatCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	if(s_xMemoryCardInfo[iMemoryCardPort][iMemoryCardSlot].iFormatted == 1)
	{
		Asura_Assert(0, "Memory card already formatted - data would be lost!");
		return;
	}
	sceMcFormat(iMemoryCardPort, iMemoryCardSlot);
	s_iAsyncFunctionNumber = sceMcFuncNoFormat;

	// Add a command to get info after formatting
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoCardInfo;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessUnformatCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	sceMcUnformat(iMemoryCardPort, iMemoryCardSlot);
	s_iAsyncFunctionNumber = sceMcFuncNoUnformat;

	// Add a command to get info after unformatting
	Asura_PS2_MemoryCard_Command xCommand;
	xCommand.iCommand = sceMcFuncNoCardInfo;
	xCommand.iParameters[0] = iMemoryCardPort;
	xCommand.iParameters[1] = iMemoryCardSlot;
	AddCommand(xCommand);
}

void Asura_PS2_MemoryCard_SubSystem::ProcessEntSpaceCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	const char* pcPathName = reinterpret_cast<const char*>(&piParameters[2]);
	s_pcOutputBuffer = reinterpret_cast<char*>(piParameters[31]);
	sceMcGetEntSpace(iMemoryCardPort, iMemoryCardSlot, pcPathName);
	s_iAsyncFunctionNumber = sceMcFuncNoEntSpace;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessRenameCommand(const int* piParameters)
{
	int iMemoryCardPort = piParameters[0];
	int iMemoryCardSlot = piParameters[1];
	const char* pcOriginalName = reinterpret_cast<const char*>(&piParameters[2]);
	const char* pcNewName	   = reinterpret_cast<const char*>(&piParameters[16]);
	sceMcRename(iMemoryCardPort, iMemoryCardSlot, pcOriginalName, pcNewName);
	s_iAsyncFunctionNumber = sceMcFuncNoRename;
}

void Asura_PS2_MemoryCard_SubSystem::ProcessChgPriorCommand(const int* piParameters)
{
}
