//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Env_Shared.h
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     Types, constants, etc., that are shared between the PS2 engine and
//     tools. Don't put any PS2-specific code or includes in here because
//     that won't work in tools.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_ENV_SHARED_H___
#define ___ASURA_PS2_ENV_SHARED_H___ 1

// C O N S T A N T S /////////////////////////////////////////////////////////

//
// Version history:
//
//     (MH - 17.05.2002) -  2 - Separated out ADC data. 
//
//     (MP - 12.06.2001) -  3 - Add tri count to strip headers.
//
//     (MP - 22.08.2002) -  4 - Reduced max strip length from 56 to 53.
//
//     (MP - 01.10.2002) -  5 - Textures are now sorted according to
//                              render pass order, as dictated by the enum
//                              below.
//
//     (MP - 03.10.2002) -  6 - Reflective textures drawn before opaque ones
//                              along with self-illuminating ones. Water
//                              drawn before translucent textures so that it
//                              be seen through windows, etc.
//
//     (MP - 10.01.2003) -  7 - Normals and ADC bits now packed into a single
//                              V4_5 per vertex.
//
//     (MP - 10.01.2003) -  8 - Regions.
//
//     (MP - 08.12.2003) -  9 - Changed UNPACKs, etc., for use with new GSM
//                              renderer.
//
//     (MP - 10.05.2004) - 10 - Removed padding after last piece of chains.
//
//     (MP - 18.08.2004) - 11 - Chains no longer split. Previously this was
//                              quite simple since the first part would
//                              always fit in less than a single piece but
//                              that's no longer true if the environment has
//                              dual texture materials.
//
static const u_int ASURA_PS2_ENV_FILE_FORMAT_VERSION_NUMBER = 11;

static const u_int ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE = (1024 >> 4); // 1Kb.

// T Y P E S /////////////////////////////////////////////////////////////////

//
// Format structures for the environment data exported from tools. Many of 
// these are used at runtime as well.
//
// Note that the region headers, material headers, the region offset list and
// module DMA chain headers are all assumed to lie on QW boundaries. This
// mostly arises out of the fact that the DMA chains end at QW boundaries
// anyway though a small amount of additional padding is required.
//

struct Asura_PS2_Env_ExportData_EnvHdr
{
    u_int uNumUniqueMaterials;
    u_int uVersionID;
    u_int uNumFmvMaterials;
    u_int uIdxOfFirstFmvMaterial; // Value is only valid iff uNumFmvTextures > 0.
    u_int uNumDmaChains;          // AKA number of distinct (material, module) pairs.
    u_int uTotalDmaChainPieces;
    u_int uMaxDmaChainPiecesReqd; // Used for the dynamic loading; this is the maximum number of DMA chain pieces that will be required for any module in the environment.
    u_int uInitReqdRgns;          // Bit i is on iff region i is required in the first frame.
};

// <now at QW boundary>

////// Followed by uNumUniqueMaterials of:
    struct Asura_PS2_Env_ExportData_MaterialHdr
    {
        short   sMaterialID;  // NB: Polys without a material are not permitted; they should be given a special name in tools to force them to use the Hideous Pink Texture (TM) at runtime.
        u_short usNumModules; // Num modules using this material.
    };

    ////// Followed by uNumModules of:
        struct Asura_PS2_Env_ExportData_ModuleHdr
        {
            short   sModuleIdx;
            u_short usNumTris;
        };

    // <padding to QW boundary>
    
////// Followed by:
    // u_int auRegionOffsets[ASURA_ENV_MAX_REGIONS]; // QW offsets from beginning of environment header to beginning of
                                                     // corresponding region's data. Set to 0xFFFFFFFF iff the region is
                                                     // empty.

    // <padding to QW boundary - but not necessary assuming ASURA_ENV_MAX_REGIONS is 32>

// Then at each corresponding QW offset we have:
struct Asura_PS2_Env_ExportData_RgnHdr
{
    u_int uNumUniqueMaterials;
    u_int uNumDmaChainPieces;
};
// <padding to QW boundary>

////// Followed by uNumUniqueMaterials of:
    struct Asura_PS2_Env_ExportData_RgnMaterialHdr
    {
        u_int uMaterialOffset; // This is an index into the unique materials for the environment, ie. an index into the global material hdr list above.
        u_int uNumModules;     // Num modules in this region using this material.
    };
    // <padding to QW boundary>

    ////// Followed by uNumModules of:
        struct Asura_PS2_Env_ExportData_DmaChainHdr
        {
            u_int uModuleOffset;       // This is an idx into the modules using this material, ie. into the global idx list for each material above.
            u_int uNumPieces;          // Num pieces that the DMA chain is split into.
            u_int uNumQWsInFirstPiece; // To save reading unused data.
            u_int uNumQWsInLastPiece;  //             "
        };
        // <padding to QW boundary - but not necessary assuming the hdr is still a whole number of QWs in size>

        ////// Followed by:
            // u_long128 aauPieces[usNumPieces - 1][ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE];
            // u_long128 auLastPiece[usNumQWsInLastPiece];
 
// Details of the DMA chain pieces:
//
//     The VIF data is written out in DMA chain pieces that look like this:
//
//         RET      (First part of first piece.)
//         [data]
//
//         NEXT|RET
//         [data]   (Second part of first piece and all other pieces.
//                   Tag will be RET in last piece, NEXT otherwise.)
//
//     Each NEXT tag references the subsequent piece. These are fixed up at
//     runtime.
//
//     As implied by the RET tags, these chains are intended to be CALLed, ie.
//     to fit into our deferred rendering scheme.
//
//     We assume that all data can be safely placed into the spare words in
//     the tags. This is true provided TTE is on for the DMA transfers and
//     also that neither DIRECT nor DIRECTHL VIF codes are used (they're the
//     only ones that require alignment stricter than a word boundary.) We
//     can therefore fit
//
//         0.5 + (ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE - 1)
//
//     QWs of data into each piece.
//
//     Note that there's no upper limit on chain length with this scheme.
//

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_ENV_SHARED_H___
