//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Env_Store.cpp
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     Functionality for dynamically loading and managing the environment
//     data.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Mesh/Asura_PS2_Env_Store.h"

// Stdlib independent.

// Stdlib platform-specific.
#include <sifdev.h>
#include <sifrpc.h>

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Env_System.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Core/Asura_PS2_OS.h"
#include "Core/Asura_PS2_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_Platform_FileAccess.h"
#include "Sound/Asura_PS2_Multistream.h"

// C L A S S E S /////////////////////////////////////////////////////////////

class DmaPiecePool
{

    public:

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        DmaPiecePool( int iNumPieces );
        ~DmaPiecePool( void );

        // M E T H O D S /////////////////////////////////////////////////////

        u_long128* Allocate( void );
        void DeallocateChain( u_long128* puChain );

        u_int GetNumFreePieces( void ) const { return m_uNumFreePieces; }

#if _DEBUG
        u_int GetTotalPieces( void ) const { return m_uNumPieces; }
        
        void Validate( void ) const;
        void ValidateChainPiece( const Asura_PS2_DmaTag& xTag, const u_long128* puPiece ) const;
        void ValidateDmaChain( const u_long128* puChain ) const;
#endif

    private:
        
        // M E T H O D S /////////////////////////////////////////////////////

        void DeallocatePiece( u_long128* puPiece );

        // S T A T E /////////////////////////////////////////////////////////

        u_long128* m_puMem;
        u_long128* m_puNextFreePiece;
        u_int      m_uNumFreePieces;
#if _DEBUG
        u_int      m_uNumPieces;
#endif

};
// End class DmaPiecePool.

class RgnManager
{
    
    public:

        // C L A S S   M E T H O D S /////////////////////////////////////////
        
        // Top level control.
        static void Initialise( void );
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        static void Update( u_int uVisRgnsForLastFrame );
#endif
        static void Reset( void );
        static void Shutdown( void );

        // Methods for setting initial properties of the env data file.
        static void SetDataOffset( u_int uOffset );
        static void SetDataSize( u_int uSize );
        
        // For loading initial data.
        static u_int* GetRgnOffsetPtr( void ) { return s_auOffsets; }
        static void SetMaxDmaChainPiecesReqd( u_int uMax );

        // For loading from env file when it's opened for the first time.
        static void BeginLoadOfInitiallyReqdRegions( Asura_Handle iHandle );
        static u_int GetCurrentFilePosition( void );
        static void EndLoadOfInitiallyReqdRegions( void );

        // And for opening it the second time...
        static void OpenFile( const char* szFilename );
		static void CloseFile( void );

        // Methods for querying which regions are currently fully loaded.
        static u_int GetLoadedRegionBits( void ) { return s_uLoadedRegions; }
        static bool IsRegionLoaded( u_int uRgn ) { return( 0 != (s_uLoadedRegions & (1 << uRgn)) ); }

        // Methods for the region loading loop.
        static void StartRegionLoad( u_int uRgn );
        static Asura_PS2_Env_Store::LOADREQD ContinueRegionLoad( void );
        static void SetRegionLoadAbortFlag( void );
        static bool IsAnyRegionLoading( void ) { return( NO_REGION != s_iCurLoadRgn ); }
        static u_int GetRegionBeingLoaded( void ) { return s_iCurLoadRgn; }

        static void UnloadRegion( u_int uRgn );

#if _DEBUG
        static void Validate( void );
#endif

    private:

        // C O N S T A N T S /////////////////////////////////////////////////

        enum { NO_REGION = -1 };

        // T Y P E S /////////////////////////////////////////////////////////

        enum LOADSTATE
        {
            LOADSTATE_READING_RGN_HDR,
            LOADSTATE_UNLOADING_REGIONS,
            LOADSTATE_READING_RGN_MATERIAL_HDR,
            LOADSTATE_READING_DMA_CHAIN_HDR,
            LOADSTATE_READING_DMA_CHAIN_PIECES,
            LOADSTATE_FINISHED_READING_LAST_PIECE,
        };

        // C L A S S   M E T H O D S /////////////////////////////////////////
        
        static bool FindRegionToUnload( void );
        
        // C L A S S   S T A T E /////////////////////////////////////////////

        static u_int         s_auOffsets[ASURA_ENV_MAX_REGIONS];
#if _DEBUG
        static u_int         s_auPiecesPerRgn[ASURA_ENV_MAX_REGIONS]; // Each entry is 0xFFFFFFFFu iff region has never been loaded.
#endif

        static u_int         s_uLoadedRegions;                        // Bit i is high iff region i is loaded.
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        static u_int         s_uPartiallyLoadedRegionsToUnload;
#endif

        static DmaPiecePool* s_pxDmaPiecePool;

        // File state.
        static Asura_Handle  s_iFileHandle;
        static bool          s_bFileWasOpened;                        // True iff we have an open file and we opened it, as opposed to us being given the handle via UseExistingFileHandle().
        static u_int         s_uBaseOffset;                           // In bytes. Should be an integral number of QWs.
        static u_int         s_uDataSize;                             //             "                 "

        // Loading state.
        static int           s_iCurLoadRgn;                           // Remaining loading state is only valid if a region is loading.
        static u_int         s_uCycleCountAtLoadStart;
        static LOADSTATE     s_eLoadState;
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        static bool          s_bAbortLoad;                            // Flag is set to true if cancel is requested; state machine then tidies up.
#endif
        static u_int         s_uNumUniqueMaterialsInRegion;
        static u_int         s_uNumChainPiecesInThisRgn;
        static u_int         s_uMaterialBeingProcessed;
        static u_int         s_uOffsetForCurMaterial;                 // The uMaterialOffset field from the corresponding Asura_PS2_Env_ExportData_RgnMaterialHdr.
        static u_int         s_uNumModulesForCurRgnMaterial;          //  "  uNumModules                                   "
        static u_int         s_uModuleBeingProcessed;
        static u_int         s_uOffsetForCurModule;
        static u_int         s_uNumPiecesForCurModule;
        static u_int         s_uNumQWsInFirstPieceOfCurModule;
        static u_int         s_uNumQWsInLastPieceOfCurModule;
        static u_long128*    s_puFirstPieceInChain;
        static u_int         s_uPieceBeingRead;
        static u_long128*    s_puCurPiece;
        static u_int         s_uNumQWsReadForCurPiece;

        // Buffer state. This is only valid if there's currently a region
        // loading.
        static const Asura_PS2_QW* s_pxNextQW;
        static u_int               s_uNumQWsLeft;

        // Output state.
        static Asura_PS2_Env_Store::NonConstIt      s_xEnvIt;
        static Asura_PS2_Env_RuntimeData_ModuleHdr* s_pxCurModuleHdr;

};
// End class RgnManager.

// G L O B A L S /////////////////////////////////////////////////////////////

#if _DEBUG
static bool g_bEnvStoreInitialised = false;
#endif

//DEBUGVAR bool dbg_bUnloadAllUnrequiredRegions = false;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// DmaPiecePool.
//
// Basically a slightly stripped down Asura_Mempool but with the important
// property that it does not pull the memory into the cache.
//

DmaPiecePool::DmaPiecePool( int iNumPieces )
{
    Asura_Assert( iNumPieces > 0, "Invalid piece count." );

    Asura_Assert( !(ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE & 3), "Each piece needs to be cache aligned." );
    m_puMem = reinterpret_cast<u_long128*>( Asura_AllocateAlign(iNumPieces * (ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE << 4), ASURA_ALIGN_64_BYTE) );
    if( !m_puMem )
    {
        // Not much we can really do to recover here but we want the problem
        // to be noticed so let's printf() as well as assert.
        Asura_OutputToDebugger_Retail(( "ERROR: Out of memory when trying to allocate memory pool for environment geometry.\n" ));
        Asura_Assert( false, "" );
    }

    // We'll store the memory with an uncached ptr. This will help to ensure
    // it never gets pulled into the cache by accident. We hardly ever want
    // to do any reads on the memory ourselves. The only situations where we
    // do are on allocation and deallocation (to follow the chain); in both
    // of these cases, we'd get a cache miss anyway so it's no worse.
    m_puMem = Asura_PS2_UncachedMem( m_puMem );

    m_puNextFreePiece = m_puMem;

    // Store ptr to next free piece in first word of each preceding piece.
    {
        u_long128* puThis = m_puMem;
        for( int i = 0; i < iNumPieces - 1; ++i )
        {
            u_long128* const puNext = puThis + ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE;
            *(reinterpret_cast<u_long128**>(puThis)) = puNext;
            puThis = puNext;
        }

        *(reinterpret_cast<u_long128**>(puThis)) = NULL;
    }

    m_uNumFreePieces = iNumPieces;
#if _DEBUG
    m_uNumPieces     = iNumPieces;
#endif
}

DmaPiecePool::~DmaPiecePool( void )
{
    if( m_puMem )
    {
        // We store the ptr as uncached so we need to mask those bits off
        // again before deallocating.
        Asura_DeallocateAlign( Asura_PS2_DMAMem(m_puMem) );
    }
}

u_long128* DmaPiecePool::Allocate( void )
{
    // We should never fail to allocate a new piece because the environment
    // data provides a peak figure that we ensure we can always accomodate.
    // Note that we ensure externally that we never try to use more pieces
    // than we've budgeted for - see the LOADSTATE_UNLOADING_REGIONS state
    // in RgnManager - so we should never return NULL here.
    Asura_Assert( m_puNextFreePiece, "" );
    u_long128* puNewPiece = m_puNextFreePiece;

    // Update next-free-piece ptr. The new ptr is in the first word of the
    // free piece we're about to give to the caller.
    m_puNextFreePiece = *(reinterpret_cast<u_long128**>(puNewPiece));

    Asura_Assert( m_uNumFreePieces, "" );
    --m_uNumFreePieces;
    
    return puNewPiece;
}

void DmaPiecePool::DeallocateChain( u_long128* puChain )
{
    // We only want to be manipulating this memory using uncached ptrs.
    Asura_Assert( ((reinterpret_cast<u_int>(puChain)) & ASURA_PS2_UNCACHED_MEM_FLAG), "Ptr should be uncached." );

    Asura_PS2_DmaTag xNextTag = *(reinterpret_cast<const Asura_PS2_DmaTag*>(puChain));

    bool bReachedEnd = false;
    while( !bReachedEnd )
    {
#if _DEBUG
        ValidateChainPiece( xNextTag, puChain );
#endif

        u_long128* const puToDealloc = puChain;

        // Any more pieces?
        if( Asura_PS2_DmaTag::ID_RET == xNextTag.m_eID )
        {
            bReachedEnd = true;
        }
        else
        {
            puChain  = Asura_PS2_UncachedMem( reinterpret_cast<u_long128*>(xNextTag.m_uAddr) );
            xNextTag = *(reinterpret_cast<const Asura_PS2_DmaTag*>(puChain));
        }

        DeallocatePiece( puToDealloc );
    }
}

void DmaPiecePool::DeallocatePiece( u_long128* puPiece )
{
    // We only want to be writing to this memory using uncached ptrs.
    Asura_Assert( ((reinterpret_cast<u_int>(puPiece)) & ASURA_PS2_UNCACHED_MEM_FLAG), "Ptr should be uncached." );

    // Push onto the front of the free list.
    *(reinterpret_cast<u_long128**>(puPiece)) = m_puNextFreePiece;
    m_puNextFreePiece = puPiece;

    ++m_uNumFreePieces;
}

#if _DEBUG

void DmaPiecePool::Validate( void ) const
{
    u_int uNumFree = 0;
    for( const u_long128* puCur = m_puNextFreePiece; puCur; ++uNumFree )
    {
        puCur = *(reinterpret_cast<const u_long128* const *>(puCur));
    }

    Asura_Assert( uNumFree == m_uNumFreePieces, "" );
}

void DmaPiecePool::ValidateChainPiece( const Asura_PS2_DmaTag& xTag, const u_long128* puPiece ) const
{
    // Verify the address is valid.
    const u_long128* const puMemEnd = m_puMem + m_uNumPieces * ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE;
    Asura_Assert( m_puMem <= puPiece && puPiece < puMemEnd, "Invalid piece address." );
    Asura_Assert( !((puPiece - m_puMem) % ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE), "Invalid piece address." );

    // Verify that the tag is valid.
    Asura_Assert( Asura_PS2_DmaTag::ID_NEXT == xTag.m_eID || Asura_PS2_DmaTag::ID_RET == xTag.m_eID, "Region data chains should consist entirely of NEXT and RET tags." );
    Asura_Assert( !(Asura_PS2_DmaTag::ID_NEXT == xTag.m_eID && !xTag.m_uAddr), "NEXT tags should not have a NULL ptr." );
}

void DmaPiecePool::ValidateDmaChain( const u_long128* puChain ) const
{
    // We only want to be manipulating this memory using uncached ptrs.
    Asura_Assert( ((reinterpret_cast<u_int>(puChain)) & ASURA_PS2_UNCACHED_MEM_FLAG), "Ptr should be uncached." );

    const u_long128* puPiece = puChain;

    Asura_PS2_DmaTag xNextTag = *(reinterpret_cast<const Asura_PS2_DmaTag*>(puPiece));

    while( true )
    {
        ValidateChainPiece( xNextTag, puPiece );

        // Any more pieces?
        if( Asura_PS2_DmaTag::ID_RET == xNextTag.m_eID )
        {
            break;
        }
        else
        {
            puPiece  = Asura_PS2_UncachedMem( reinterpret_cast<u_long128*>(xNextTag.m_uAddr) );
            xNextTag = *(reinterpret_cast<const Asura_PS2_DmaTag*>(puPiece));
        }
    }
}

#endif // #if _DEBUG

//
// RgnManager.
//

u_int                 RgnManager::s_auOffsets[ASURA_ENV_MAX_REGIONS];
#if _DEBUG
u_int                 RgnManager::s_auPiecesPerRgn[ASURA_ENV_MAX_REGIONS];
#endif
u_int                 RgnManager::s_uLoadedRegions;
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
u_int                 RgnManager::s_uPartiallyLoadedRegionsToUnload;
#endif
DmaPiecePool*         RgnManager::s_pxDmaPiecePool;
Asura_Handle          RgnManager::s_iFileHandle;
bool                  RgnManager::s_bFileWasOpened;
u_int                 RgnManager::s_uBaseOffset;
u_int                 RgnManager::s_uDataSize;
const Asura_PS2_QW*   RgnManager::s_pxNextQW;
u_int                 RgnManager::s_uNumQWsLeft;
int                   RgnManager::s_iCurLoadRgn;
u_int                 RgnManager::s_uCycleCountAtLoadStart;
RgnManager::LOADSTATE RgnManager::s_eLoadState;
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
bool                  RgnManager::s_bAbortLoad;
#endif
u_int                 RgnManager::s_uNumUniqueMaterialsInRegion;
u_int                 RgnManager::s_uNumChainPiecesInThisRgn;
u_int                 RgnManager::s_uMaterialBeingProcessed;
u_int                 RgnManager::s_uOffsetForCurMaterial;
u_int                 RgnManager::s_uNumModulesForCurRgnMaterial;
u_int                 RgnManager::s_uModuleBeingProcessed;
u_int                 RgnManager::s_uOffsetForCurModule;
u_int                 RgnManager::s_uNumPiecesForCurModule;
u_int                 RgnManager::s_uNumQWsInFirstPieceOfCurModule;
u_int                 RgnManager::s_uNumQWsInLastPieceOfCurModule;
u_long128*            RgnManager::s_puFirstPieceInChain;
u_int                 RgnManager::s_uPieceBeingRead;
u_long128*            RgnManager::s_puCurPiece;
u_int                 RgnManager::s_uNumQWsReadForCurPiece;

Asura_PS2_Env_Store::NonConstIt      RgnManager::s_xEnvIt;
Asura_PS2_Env_RuntimeData_ModuleHdr* RgnManager::s_pxCurModuleHdr;

void Asura_PS2_Env_Store::AbortCurrentLoad()
{
    while( RgnManager::IsAnyRegionLoading() )
    {
        RgnManager::SetRegionLoadAbortFlag();
        RgnManager::ContinueRegionLoad();

        // Since we're just going to sit and wait for this, we'd better give
        // the MS thread(s) some time do their stuff.
        Asura_PS2_OS::DelayThread( 1 );
    }
}

void RgnManager::BeginLoadOfInitiallyReqdRegions( Asura_Handle iHandle )
{
    Asura_Assert( ASURA_INVALID_HANDLE == s_iFileHandle, "File already opened?" );
    Asura_Assert( ASURA_INVALID_HANDLE != iHandle, "Invalid existing file handle specified." );

    s_iFileHandle    = iHandle;
    s_bFileWasOpened = false;

    Asura_Platform_FileAccess::BeginDirectMultistreamAccess( iHandle );
}

void RgnManager::CloseFile( void )
{
    if( ASURA_INVALID_HANDLE != s_iFileHandle )
    {
        if( s_bFileWasOpened ) { Asura_FileAccess::Close( s_iFileHandle ); }

        s_iFileHandle = ASURA_INVALID_HANDLE;
    }
}

Asura_PS2_Env_Store::LOADREQD RgnManager::ContinueRegionLoad( void )
{
    if( NO_REGION == s_iCurLoadRgn )
    {
        Asura_Assert( false, "Not currently loading a region." );
        return Asura_PS2_Env_Store::LOADREQD_NO;
    }
    Asura_Assert( 0 <= s_iCurLoadRgn && static_cast<u_int>(s_iCurLoadRgn) < ASURA_ENV_MAX_REGIONS, "" );

    if( !s_uNumQWsLeft )
    {
        u_int uBytes = Asura_PS2_Multistream_FastLoad::GetData( s_iFileHandle, reinterpret_cast<const void**>(&s_pxNextQW) );
        if( !uBytes )
        {
            return Asura_PS2_Env_Store::LOADREQD_YES;
        }

// (MP - 13.04.2007) This assert can fire in the case where there's very
// little data following the PS2Env chunk, so that we don't get another full
// buffer's worth. In that case, we won't be using all of the data anyway so
// we can just round down. Other asserts should pick up genuine problems of
// stuff being un-aligned or there not being the correct amount of data so I
// think this one can go.
//      Asura_Assert( !(uBytes & 15), "" );

        s_uNumQWsLeft = (uBytes >> 4);
    }

    bool bFinishedLoading             = false;
    bool bSuspendSinceCannotUnloadYet = false;

    while( s_uNumQWsLeft && !bFinishedLoading && !bSuspendSinceCannotUnloadYet )
    {
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        // Cancel load?
        if( s_bAbortLoad )
        {
            // It's not safe to cancel while in the middle of loading a chain
            // because we won't be able to deallocate it properly.
            if( LOADSTATE_READING_DMA_CHAIN_PIECES != s_eLoadState && LOADSTATE_FINISHED_READING_LAST_PIECE != s_eLoadState )
            {
                bFinishedLoading = true;
                break;
            }
        }
#endif

        switch( s_eLoadState )
        {
            case LOADSTATE_READING_RGN_HDR:
            {
                // Read.
                {
                    const Asura_PS2_Env_ExportData_RgnHdr* pxHdr = reinterpret_cast<const Asura_PS2_Env_ExportData_RgnHdr*>( s_pxNextQW );

                    s_uNumUniqueMaterialsInRegion = pxHdr->uNumUniqueMaterials;
                    s_uNumChainPiecesInThisRgn    = pxHdr->uNumDmaChainPieces;

                    // Verify these values are sensible.
                    Asura_Assert( s_uNumUniqueMaterialsInRegion <= Asura_PS2_Env_Store::GetNumUniqueMaterials(), "" );
#if _DEBUG
                    Asura_Assert( s_uNumChainPiecesInThisRgn <= s_pxDmaPiecePool->GetTotalPieces(), "" );

                    if( 0xFFFFFFFFu == s_auPiecesPerRgn[s_iCurLoadRgn] )
                    {
                        s_auPiecesPerRgn[s_iCurLoadRgn] = s_uNumChainPiecesInThisRgn;
                    }
                    else
                    {
                        Asura_Assert( s_auPiecesPerRgn[s_iCurLoadRgn] == s_uNumChainPiecesInThisRgn, "Num pieces in rgn has changed?!?" );
                    }
#endif
                }

                s_eLoadState = LOADSTATE_UNLOADING_REGIONS;

                // Skip over hdr and subsequent padding.
                ++s_pxNextQW;
                --s_uNumQWsLeft;

                break;
            }
            // End case LOADSTATE_READING_RGN_HDR.

            case LOADSTATE_UNLOADING_REGIONS:
            {
                bool bUnloadedEnough = true;
                while( s_uNumChainPiecesInThisRgn > s_pxDmaPiecePool->GetNumFreePieces() )
                {
                    if( !FindRegionToUnload() )
                    {
                        bUnloadedEnough              = false;
                        bSuspendSinceCannotUnloadYet = true;
                        break;
                    }
                }

                if( bUnloadedEnough )
                {
                    s_eLoadState              = LOADSTATE_READING_RGN_MATERIAL_HDR;
                    s_uMaterialBeingProcessed = 0;
                }

                break;
            }
            // End case LOADSTATE_UNLOADING_REGIONS.

            case LOADSTATE_READING_RGN_MATERIAL_HDR:
            {
                // Keep a record of the previous material offset.
                u_int uPrevMatOffset = ( s_uMaterialBeingProcessed ? s_uOffsetForCurMaterial : 0 );

                // Read.
                {
                    const Asura_PS2_Env_ExportData_RgnMaterialHdr* pxHdr = reinterpret_cast<const Asura_PS2_Env_ExportData_RgnMaterialHdr*>( s_pxNextQW );

                    s_uOffsetForCurMaterial        = pxHdr->uMaterialOffset;
                    s_uNumModulesForCurRgnMaterial = pxHdr->uNumModules;
                }

                // Skip over hdr and subsequent padding.
                ++s_pxNextQW;
                --s_uNumQWsLeft;

                s_eLoadState            = LOADSTATE_READING_DMA_CHAIN_HDR;
                s_uModuleBeingProcessed = 0;

                // Calculate material offset delta.
                Asura_Assert( uPrevMatOffset <= s_uOffsetForCurMaterial && s_uOffsetForCurMaterial < Asura_PS2_Env_Store::GetNumUniqueMaterials(), "" );
                u_int uMatOffsetDelta = s_uOffsetForCurMaterial - uPrevMatOffset;

#ifndef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
                // If dynamic loading is off then we expect all modules to be
                // in region 0. If all the data's in a single region then
                // this delta should always be 1, apart from the first one.
                if( s_uMaterialBeingProcessed )
                {
                    Asura_Assert( 1 == uMatOffsetDelta, "" );
                }
#endif

                // Move on to relevant material header.
                while( uMatOffsetDelta-- )
                {
                    s_xEnvIt.Next();
                }

                // Get module hdr array for this new material.
                s_pxCurModuleHdr = s_xEnvIt.GetModuleHdrsForCurMaterial();

                break;
            }
            // End case LOADSTATE_READING_RGN_MATERIAL_HDR.

            case LOADSTATE_READING_DMA_CHAIN_HDR:
            {
                // Keep a record of the previous module offset.
                u_int uPrevModuleOffset = ( s_uModuleBeingProcessed ? s_uOffsetForCurModule : 0 );

                // Copy data from input hdr.
                {
                    const Asura_PS2_Env_ExportData_DmaChainHdr* pxChainHdr = reinterpret_cast<const Asura_PS2_Env_ExportData_DmaChainHdr*>( s_pxNextQW );

                    s_uOffsetForCurModule            = pxChainHdr->uModuleOffset;
                    s_uNumPiecesForCurModule         = pxChainHdr->uNumPieces;
                    s_uNumQWsInFirstPieceOfCurModule = pxChainHdr->uNumQWsInFirstPiece;
                    s_uNumQWsInLastPieceOfCurModule  = pxChainHdr->uNumQWsInLastPiece;

                    // These asserts are just here because the data used to
                    // be slightly different and there could be padding after
                    // the first piece, even if there were more in the module.
                    // The s_uNumQWsInFirstPieceOfCurModule field is actually
                    // redundant now.
                    if( 1 == s_uNumPiecesForCurModule )
                    {
                        Asura_Assert( s_uNumQWsInFirstPieceOfCurModule == s_uNumQWsInLastPieceOfCurModule, "" );
                    }
                    else
                    {
                        Asura_Assert( s_uNumQWsInFirstPieceOfCurModule == ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE, "" );
                    }
                }

                // Skip over hdr.
                ++s_pxNextQW;
                --s_uNumQWsLeft;

                // Calculate module offset delta.
                Asura_Assert( uPrevModuleOffset <= s_uOffsetForCurModule && s_uOffsetForCurModule < s_xEnvIt.GetNumModulesForCurMaterial(), "" );
                u_int uModuleOffsetDelta = s_uOffsetForCurModule - uPrevModuleOffset;

#ifndef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
                // If dynamic loading is off then we expect all modules to be
                // in region 0. If all the data's in a single region then
                // this delta should always be 1, apart from the first one.
                if( s_uModuleBeingProcessed )
                {
                    Asura_Assert( 1 == uModuleOffsetDelta, "" );
                }
#endif

                // Move on to correct module hdr.
                while( uModuleOffsetDelta-- )
                {
                    ++s_pxCurModuleHdr;
                }
                Asura_Assert( s_pxCurModuleHdr < s_xEnvIt.GetModuleHdrsForCurMaterial() + s_xEnvIt.GetNumModulesForCurMaterial(), "" );

                // Verify this module hasn't already had its data loaded.
                Asura_Assert( !s_pxCurModuleHdr->uChainCALLAddr, "" );
               
                // (MP - 13.05.2003) We should always have enough free pieces
                // at this point now since we continually try to unload
                // data in the LOADSTATE_UNLOADING_REGIONS state until
                // there's enough space for the entire region.
                Asura_Assert( s_uNumPiecesForCurModule <= s_pxDmaPiecePool->GetNumFreePieces(), "" );

                s_eLoadState             = LOADSTATE_READING_DMA_CHAIN_PIECES;
                s_uPieceBeingRead        = 0;
                s_uNumQWsReadForCurPiece = 0;

                break;
            }
            // End case LOADSTATE_READING_DMA_CHAIN_HDR.

            case LOADSTATE_READING_DMA_CHAIN_PIECES:
            {
                // If we haven't read any QWs for this piece yet then we need
                // to grab some memory for it from the DMA piece pool.
                if( !s_uNumQWsReadForCurPiece )
                {
                    u_long128* const puNewPiece = s_pxDmaPiecePool->Allocate();

                    if( s_uPieceBeingRead )
                    {
                        // If this is not the first piece then fix up the 
                        // address in the NEXT tag of the previous piece. We
                        // need to mask off the upper nybble of the new
                        // piece's address because we'll have an uncached ptr
                        // to it.
                        (reinterpret_cast<Asura_PS2_DmaTag*>(s_puCurPiece))->m_uAddr = reinterpret_cast<u_int>( Asura_PS2_DMAMem(puNewPiece) );
                    }
                    else
                    {
                        // Otherwise if this is our first piece for this
                        // chain then record the start address for when we
                        // complete the chain. We can't store the start
                        // address in the module hdr until we've finished
                        // loading the whole chain.
                        s_puFirstPieceInChain = puNewPiece;
                    }

                    s_puCurPiece = puNewPiece;
                }

                Asura_Assert( s_uPieceBeingRead < s_uNumPiecesForCurModule, "" );
                const bool bIsLastPiece = ( s_uNumPiecesForCurModule == s_uPieceBeingRead + 1 );

                u_int uNumQWsInThisPiece;
                if( bIsLastPiece )
                {
                    uNumQWsInThisPiece = s_uNumQWsInLastPieceOfCurModule;
                }
                else
                {
                    uNumQWsInThisPiece = ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE;
                }
                   
                // How many more QWs should we read for the current piece?
                u_int uNumQWsToReadForCurPiece = Asura_Maths::Min( s_uNumQWsLeft, uNumQWsInThisPiece - s_uNumQWsReadForCurPiece );
                Asura_Assert( uNumQWsToReadForCurPiece, "" );

                // Setup output ptr before updating global QW ctr.
                u_long128* puOut = s_puCurPiece + s_uNumQWsReadForCurPiece;

                // Update loop state.
                s_uNumQWsReadForCurPiece += uNumQWsToReadForCurPiece;
                s_uNumQWsLeft            -= uNumQWsToReadForCurPiece;

                // Copy.
                while( uNumQWsToReadForCurPiece-- )
                {
                    *puOut++ = (s_pxNextQW++)->u128;
                }

                // Read all (non-padding) QWs for this piece?
                Asura_Assert( s_uNumQWsReadForCurPiece <= uNumQWsInThisPiece, "" );
                if( s_uNumQWsReadForCurPiece == uNumQWsInThisPiece )
                {
                    if( bIsLastPiece )
                    {
                        // TODO: Don't need to go to this separate state
                        // because it doesn't read any data anymore. Just
                        // call a FinishedReadingLastPiece() method or
                        // something.

                        s_eLoadState = LOADSTATE_FINISHED_READING_LAST_PIECE;
                    }
                    else
                    {
                        // Just move on to next piece then.
                        ++s_uPieceBeingRead;
                        s_uNumQWsReadForCurPiece = 0;
                    }
                }

                break;
            }
            // End case LOADSTATE_READING_DMA_CHAIN_PIECES.

            case LOADSTATE_FINISHED_READING_LAST_PIECE:
            {
                ++s_uPieceBeingRead;
                Asura_Assert( s_uPieceBeingRead == s_uNumPiecesForCurModule, "Wasn't last piece." );

                // Store the ptr in the module hdr array now that we've
                // finished loading it.
                s_pxCurModuleHdr->uChainCALLAddr  = reinterpret_cast<u_int>( Asura_PS2_DMAMem(s_puFirstPieceInChain) );

#if _DEBUG
                s_pxDmaPiecePool->ValidateDmaChain( s_puFirstPieceInChain );
#endif

                // Was this the last module for this material?
                if( ++s_uModuleBeingProcessed == s_uNumModulesForCurRgnMaterial )
                {
                    // Any more materials for this region?
                    if( ++s_uMaterialBeingProcessed == s_uNumUniqueMaterialsInRegion )
                    {
                        // Finished all the materials too so we're done.
                        bFinishedLoading = true;
                    }
                    else
                    {
                        s_eLoadState = LOADSTATE_READING_RGN_MATERIAL_HDR;
                    }
                }
                else
                {
                    // No so move on to the next module. Note that we'll
                    // move on to the correct output module hdr once
                    // we've read the next DMA chain hdr and know how far
                    // to skip.
                    s_eLoadState = LOADSTATE_READING_DMA_CHAIN_HDR;
                }

                break;
            }
            // End case LOADSTATE_SKIPPING_PADDING_AFTER_LAST_PIECE.
        }
        // End switch( s_eLoadState ).
    }
    // End while( s_uNumQWsLeft && !bFinishedLoading ).

    // Ensure we're not going to leave a non-empty buffer post-load...
    Asura_Assert( !(bFinishedLoading && bSuspendSinceCannotUnloadYet), "" );

    if( !bSuspendSinceCannotUnloadYet )
    {
        s_pxNextQW    = NULL;
        s_uNumQWsLeft = 0;
    }

    // Finished loading this region yet?
    if( bFinishedLoading )
    {
        // Flag this region as loaded.
        s_uLoadedRegions |= (1 << s_iCurLoadRgn);

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        // Was the load actually cancelled?
        if( s_bAbortLoad )
        {
            // Flag that this region needs to be removed ASAP. We may not
            // be able to unload it immediately since some of its modules may
            // have been rendered last frame and therefore still be queued
            // if we're using deferred rendering.
            s_uPartiallyLoadedRegionsToUnload |= (1 << s_iCurLoadRgn);
            Asura_OutputToDebugger(( "Queueing cancelled region %i for unloading.\n", s_iCurLoadRgn ));
        }
        else
        {
            float fElapsed = static_cast<float>(Asura_PS2_Timers::GetElapsedCycles(s_uCycleCountAtLoadStart)) / ASURA_PS2_EE_CYCLES_PER_SEC;
            Asura_OutputToDebugger(( "Loaded region %i (%.2fs).\n", s_iCurLoadRgn, Asura_Maths::FtoD(fElapsed) ));
        }
#endif // #ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING

        // Reset load state.
        s_iCurLoadRgn = NO_REGION;

        return Asura_PS2_Env_Store::LOADREQD_NO;
    }

    // Okay, more loading to do then.
    if( bSuspendSinceCannotUnloadYet )
    {
        return Asura_PS2_Env_Store::LOADREQD_YES_BUT_NO_REGIONS_AVAILABLE_FOR_UNLOAD;
    }
    else
    {
        return Asura_PS2_Env_Store::LOADREQD_YES;
    }
}

void RgnManager::EndLoadOfInitiallyReqdRegions( void )
{
    u_int uEndOfData = s_uBaseOffset + s_uDataSize;

    Asura_PS2_Multistream_FastLoad::Seek( s_iFileHandle, uEndOfData );
    Asura_Platform_FileAccess::EndDirectMultistreamAccess( s_iFileHandle, uEndOfData );

    CloseFile(); // Won't actually close the file in this case, rather just forget the handle.
}

bool RgnManager::FindRegionToUnload( void )
{
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING

    // (MP - 13.05.2003) NB: Since the environment stuff should be getting
    // updated before the module visibilities, the vis regions for the last
    // render, ie. the queued frame, are the current ones.

    // (MP - 18.07.2003) We should be getting updated after module linking is
    // but before rendering. Therefore we're interested in our record of
    // required and visible regions for *last* frame.
    u_int uReqd         = Asura_PS2_Env_Store::GetReqdRgnsForThisFrame();
    u_int uVisLastFrame = Asura_PS2_Env_Store::GetVisRgnsForLastFrame();

    // We can't unload anything that is still required or that we rendered
    // last frame. The latter point is only relevant if we're using deferred
    // rendering but (a) everyone is, and (b) there's no harm in doing this
    // anyway.
    u_int uUnloadCandidates = ( GetLoadedRegionBits() & ~(uReqd | uVisLastFrame) );
    if( !uUnloadCandidates )
    {
        return false;
    }

    u_int uToUnload;
    for( uToUnload = 0; uToUnload < ASURA_ENV_MAX_REGIONS; ++uToUnload )
    {
        if( uUnloadCandidates & (1 << uToUnload) )
        {
            break;
        }
    }
    Asura_Assert( uToUnload < ASURA_ENV_MAX_REGIONS, "" );

    UnloadRegion( uToUnload );
    return true;

#else // not #ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING

	return true;

#endif // else not #ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
}

u_int RgnManager::GetCurrentFilePosition( void )
{
    return Asura_PS2_Multistream_FastLoad::GetFilePos( s_iFileHandle );
}

void RgnManager::Initialise( void )
{
    s_pxDmaPiecePool = NULL;
    s_iFileHandle    = ASURA_INVALID_HANDLE;

    Reset();
}

void RgnManager::OpenFile( const char* szFilename )
{
    Asura_Assert( ASURA_INVALID_HANDLE == s_iFileHandle, "File already opened?" );

    s_iFileHandle = Asura_FileAccess::Open( szFilename );
    Asura_Assert( ASURA_INVALID_HANDLE != s_iFileHandle, "Couldn't re-open main .ps2 file." );

    s_bFileWasOpened = true;
}

void RgnManager::Reset( void )
{
    // (MP - 14.01.2003) NB: No need to abort current load, unload regions,
    // etc., because that's all taken care of by simply deallocating the
    // chain piece pool below.

	CloseFile();

    if( s_pxDmaPiecePool )
    {
        delete s_pxDmaPiecePool;
        s_pxDmaPiecePool = NULL;
    }

    s_pxNextQW    = NULL;
    s_uNumQWsLeft = 0;

    s_uLoadedRegions                  = 0;
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    s_uPartiallyLoadedRegionsToUnload = 0;
#endif
    s_iCurLoadRgn                     = NO_REGION;

#if _DEBUG
    for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
    {
        s_auPiecesPerRgn[u] = 0xFFFFFFFFu;
    }
#endif
}

void RgnManager::SetDataOffset( u_int uOffset )
{
    s_uBaseOffset = uOffset;
}

void RgnManager::SetDataSize( u_int uSize )
{
    Asura_Assert( !(uSize & 15), "Data should be an integral number of QWs." );
    s_uDataSize = uSize;
}

void RgnManager::SetMaxDmaChainPiecesReqd( u_int uMax )
{
    // Setup DMA piece pool.
    Asura_Assert( !s_pxDmaPiecePool, "Dma piece pool is already allocated. Did you forget to Reset() or call this method multiple times?" );
    s_pxDmaPiecePool = new DmaPiecePool( uMax );
}

void RgnManager::SetRegionLoadAbortFlag( void )
{
    if( NO_REGION == s_iCurLoadRgn )
    {
        Asura_Assert( false, "Not currently loading a region." );
        return;
    }
    Asura_Assert( 0 <= s_iCurLoadRgn && static_cast<u_int>(s_iCurLoadRgn) < ASURA_ENV_MAX_REGIONS, "" );

    Asura_OutputToDebugger(( "Cancelling load of region %i.\n", s_iCurLoadRgn ));

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    s_bAbortLoad = true;
#endif
}

void RgnManager::Shutdown( void )
{
    Reset();
}

void RgnManager::StartRegionLoad( u_int uRgn )
{
    Asura_Assert( uRgn < ASURA_ENV_MAX_REGIONS, "Invalid region." );

    if( NO_REGION != s_iCurLoadRgn )
    {
        Asura_Assert( false, "Already loading a region. Did you check IsLoadingRegion()?" );
        return;
    }

    if( IsRegionLoaded(uRgn) )
    {
        Asura_Assert( false, "Specified region is already loaded. Did you check IsRegionLoaded()?" );
        return;
    }

    // Extract offset for this region.
    const u_int uQWOffset = s_auOffsets[uRgn];
    if( 0xFFFFFFFFu == uQWOffset )
    {
        Asura_Assert( false, "Specified region is empty." );
        return;
    }

    // Seek to correct position in file.
    Asura_PS2_Multistream_FastLoad::Seek( s_iFileHandle, s_uBaseOffset + (uQWOffset << 4) );

    // Setup load state.
    s_iCurLoadRgn            = uRgn;
    s_uCycleCountAtLoadStart = Asura_PS2_Timers::GetCycleCount();
    s_eLoadState             = LOADSTATE_READING_RGN_HDR;
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    s_bAbortLoad             = false;
#endif
    // Reset environment iterator.
    s_xEnvIt.Restart();

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    Asura_OutputToDebugger(( "Started region %u loading.\n", uRgn ));
#endif
}

void RgnManager::UnloadRegion( u_int uRgn )
{
    Asura_Assert( uRgn < ASURA_ENV_MAX_REGIONS, "Invalid region." );

    if( !IsRegionLoaded(uRgn) )
    {
        Asura_Assert( false, "Specified region isn't loaded. Did you check IsRegionLoaded()?" );
        return;
    }

    const u_int uNumFreePiecesBefore = s_pxDmaPiecePool->GetNumFreePieces();   
    u_int uNumChainsDeallocd = 0;

    const Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
    for( Asura_PS2_Env_Store::NonConstIt xIt; !xIt.IsFinished(); xIt.Next() )
    {
        Asura_PS2_Env_RuntimeData_ModuleHdr* pxNextHdr = xIt.GetModuleHdrsForCurMaterial();
        for( u_int u = xIt.GetNumModulesForCurMaterial(); u--; ++pxNextHdr )
        {
            if( pxEnv->GetModuleRegion(pxNextHdr->iModuleIdx) == uRgn )
            {
                // If we cancelled the region load then not all chains will
                // have made it into memory so deal with that case.
                const u_int uChainCALLAddr = pxNextHdr->uChainCALLAddr;
                if( uChainCALLAddr )
                {
                    s_pxDmaPiecePool->DeallocateChain( Asura_PS2_UncachedMem(reinterpret_cast<u_long128*>(uChainCALLAddr)) );
                    pxNextHdr->uChainCALLAddr = 0;
                    ++uNumChainsDeallocd;
                }
            }
        }
    }

    Asura_Assert( s_pxDmaPiecePool->GetNumFreePieces() >= uNumFreePiecesBefore, "" );

    Asura_OutputToDebugger(( "Unloaded region %u. (Deallocated %u pieces in %u chains.)\n",
                             uRgn,
                             s_pxDmaPiecePool->GetNumFreePieces() - uNumFreePiecesBefore,
                             uNumChainsDeallocd ));

    s_uLoadedRegions &= ~(1 << uRgn);
}

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
void RgnManager::Update( u_int uVisRgnsForLastFrame )
{
    // (MP - 13.05.2003) NB: Even if the region in question has become
    // required again, currently we still need to unload it, because it's 
    // only been partially loaded. We'll probably then reload it next frame
    // but this should be very rare.
    if( s_uPartiallyLoadedRegionsToUnload )
    {
        u_int uRegionsWeCanUnloadNow = s_uPartiallyLoadedRegionsToUnload & ~uVisRgnsForLastFrame;
        if( uRegionsWeCanUnloadNow )
        {
            for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
            {
                if( uRegionsWeCanUnloadNow & (1 << u) ) { UnloadRegion( u ); }
            }

            s_uPartiallyLoadedRegionsToUnload &= ~uRegionsWeCanUnloadNow;
        }
    }
}
#endif

#if _DEBUG
void RgnManager::Validate( void )
{
    s_pxDmaPiecePool->Validate();

    // We can't do the check on free pieces below if we're in the middle of
    // loading a region because it will have consumed some chain pieces but
    // won't be flagged as loaded yet, throwing off the expected used chain
    // pieces.
    if( IsAnyRegionLoading() ) { return; }

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    // Similarly, we can't do it if we have some partially loaded regions to
    // unload.
    if( s_uPartiallyLoadedRegionsToUnload ) { return; }
#endif

    u_int uUsedChainPieces = 0;
    for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
    {
        if( IsRegionLoaded(u) ) { uUsedChainPieces += s_auPiecesPerRgn[u]; }
    }

    Asura_Assert( s_pxDmaPiecePool->GetTotalPieces() - s_pxDmaPiecePool->GetNumFreePieces() == uUsedChainPieces, "" );
}
#endif // #if _DEBUG

//
// Asura_PS2_Env_Store.
//

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
u_int                                  Asura_PS2_Env_Store::s_uReqdRegionsLastFrame;
u_int                                  Asura_PS2_Env_Store::s_uVisRegionsLastFrame;
bool                                   Asura_PS2_Env_Store::s_bForceAllLoadsToCompleteOnNextUpdate;
#endif
Asura_PS2_Env_ExportData_EnvHdr        Asura_PS2_Env_Store::s_xEnvHdr;  
Asura_PS2_Env_RuntimeData_MaterialHdr* Asura_PS2_Env_Store::s_pxMaterialHdrs;
Asura_PS2_Env_RuntimeData_ModuleHdr*   Asura_PS2_Env_Store::s_pxModuleHdrs;

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
static bool g_bFirstEnvStoreUpdate = true;
static bool g_bDynamicLoadingIsReqd;
#endif

void Asura_PS2_Env_Store::CompleteOutstandingLoads( void )
{
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    const u_int uReqdRegions = GetReqdRgnsForThisFrame();
    while( LOADREQD_YES == LoadIfNecessary(uReqdRegions) ) { }
#endif
}

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
u_int Asura_PS2_Env_Store::GetReqdRgnsForThisFrame( void )
{
    u_int uReqdRegions = (Asura_Environment_System::GetEnvironment())->GetReqdRgnsForThisFrame();

    // Just use last frame's if problems elsewhere - eg. screwed up player
    // orientations for a frame or two - mean we don't have a proper required
    // regions record for this frame.
    if( !uReqdRegions )
    {
        uReqdRegions = s_uReqdRegionsLastFrame;
    }

    return uReqdRegions;
}

u_int Asura_PS2_Env_Store::GetVisRgnsForThisFrame( void )
{
    const Asura_Environment* pxEnv = Asura_Environment_System::GetEnvironment();
	u_int uVisRegions = pxEnv->GetVisRgnsForThisFrame();

    // Just use last frame's if problems elsewhere - eg. screwed up player
    // orientations for a frame or two - mean we don't have a proper vis
    // regions record for this frame.
    if( !uVisRegions )
    {
        uVisRegions = s_uVisRegionsLastFrame;
    }

    return uVisRegions;
}
#endif

void Asura_PS2_Env_Store::Initialise( void )
{
#if _DEBUG
    Asura_Assert( !g_bEnvStoreInitialised, "Already initialised." );
#endif

    s_xEnvHdr.uNumUniqueMaterials = 0;

    s_pxMaterialHdrs = NULL;
    s_pxModuleHdrs  = NULL;

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    s_bForceAllLoadsToCompleteOnNextUpdate = false;

    g_bFirstEnvStoreUpdate = true;
#endif

#if _DEBUG
    g_bEnvStoreInitialised = true;
#endif

    RgnManager::Initialise();
}

// Returns true iff any load was necessary.
Asura_PS2_Env_Store::LOADREQD Asura_PS2_Env_Store::LoadIfNecessary( u_int uReqdRegions )
{
    if( RgnManager::IsAnyRegionLoading() )
    {
        // Still required?
        if( !(uReqdRegions & (1 << RgnManager::GetRegionBeingLoaded())) )
        {
            RgnManager::SetRegionLoadAbortFlag();
        }

        LOADREQD eLoadReqd = RgnManager::ContinueRegionLoad();

        // Don't report we're done loading if there's actually other regions
        // required too. Leave it up to higher level code as to whether or
        // not we immediately start on the next region as well.
        if( LOADREQD_NO == eLoadReqd && RgnManager::GetLoadedRegionBits() != uReqdRegions )
        {
            eLoadReqd = LOADREQD_YES;
        }

        return eLoadReqd;
    }
    else
    {
        // Check for any newly-required regions.
        const u_int uLoadedRegions = RgnManager::GetLoadedRegionBits();
        if( uLoadedRegions != uReqdRegions )
        {
            for( u_int uRgn = 0; uRgn < ASURA_ENV_MAX_REGIONS; ++uRgn )
            {
                if( (uReqdRegions & (1 << uRgn)) && !(uLoadedRegions & (1 << uRgn)) )
                {
                    RgnManager::StartRegionLoad( uRgn );
                    return LOADREQD_YES;
                }
            }
        }
    }

    return LOADREQD_NO;
}

void Asura_PS2_Env_Store::LoadInitialData( Asura_Handle iHandle )
{
#if _DEBUG
    Asura_Assert( g_bEnvStoreInitialised, "Not initialised." );
#endif

    LoadRegionIndependentData( iHandle );

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING

    RgnManager::SetMaxDmaChainPiecesReqd( s_xEnvHdr.uMaxDmaChainPiecesReqd );

	Asura_OutputToDebugger(("region loading is saving %d bytes of a total %d at runtime.\n", 
		(s_xEnvHdr.uTotalDmaChainPieces - s_xEnvHdr.uMaxDmaChainPiecesReqd) * (ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE << 4),
		(s_xEnvHdr.uTotalDmaChainPieces) * (ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE << 4) ));
#else

    // We assume that regions have not actually been used if dynamic loading
    // is off - ie. everything's in region 0 - and so the max number of chain
    // pieces required simultaneously should be the same as the total number
    // of chain pieces.
    Asura_Assert( s_xEnvHdr.uMaxDmaChainPiecesReqd == s_xEnvHdr.uTotalDmaChainPieces, "" );

    // Just to be sure, we'd better use the total, in case they've stuck
    // modules in other regions.
    RgnManager::SetMaxDmaChainPiecesReqd( s_xEnvHdr.uTotalDmaChainPieces );

#endif

    RgnManager::BeginLoadOfInitiallyReqdRegions( iHandle );
    LoadInitiallyReqdRegions();
    RgnManager::EndLoadOfInitiallyReqdRegions();
}

void Asura_PS2_Env_Store::LoadInitiallyReqdRegions( void )
{
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    Asura_OutputToDebugger(( "Loading initially required regions.\n" ));
#endif

    u_int uInitiallyReqdRgns;

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    uInitiallyReqdRgns      = s_xEnvHdr.uInitReqdRgns;   
    s_uReqdRegionsLastFrame = s_xEnvHdr.uInitReqdRgns;
#else
    // All regions are always required.
    {
        const u_int* puRgnOffsets = RgnManager::GetRgnOffsetPtr();

        uInitiallyReqdRgns = 0;
        for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
        {
            const u_int uNextOffset = *puRgnOffsets++;
            if( 0xFFFFFFFFu != uNextOffset )
            {
                uInitiallyReqdRgns |= (1 << u);
            }
        }
    }

    // If dynamic loading is off then all modules should really be in region
    // 0. It'll all still work if they're not but artists shouldn't even be
    // using region mode for projects not using dynamic loading.
    Asura_Assert( (1 << 0) == uInitiallyReqdRgns, "Dynamic loading is off for this project yet you've assigned modules to regions other than 0." );
#endif

    for( u_int uUpdateCounter = 0; ; ++uUpdateCounter )
    {
        LOADREQD eLoadReqd = LoadIfNecessary( uInitiallyReqdRgns );
        Asura_Assert( LOADREQD_YES_BUT_NO_REGIONS_AVAILABLE_FOR_UNLOAD != eLoadReqd, "Initially reqd regions calculation incorrect?" );
        
        if( LOADREQD_NO == eLoadReqd )
        {
            break;
        }
        else if( !(uUpdateCounter & 15) )
        {
            // Update independent chunk system with progress, so that the
            // visual display of loading progress still moves (relatively)
            // smoothly even when we have a mass of region data to read in
            // the initial load.
            Asura_ChunkLoading_System::UpdateLoadingBarWithForcedFilePosition( RgnManager::GetCurrentFilePosition() );
        }
    }
}

void Asura_PS2_Env_Store::LoadRegionIndependentData( int iHandle )
{
#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    Asura_OutputToDebugger(( "Loading region independent data.\n" ));
#endif

    enum STATE
    {
        STATE_READING_ENV_HDR,
        STATE_READING_MATERIAL_HDR,
        STATE_READING_MODULE_HDRS,
        STATE_READING_REGION_OFFSETS,
        STATE_DONE,
    };

    u_int* const puRegionOffsets = RgnManager::GetRgnOffsetPtr();

    // (MP - 03.02.2003) TODO: It'd be quicker to read the data directly but
    // the independent interface doesn't currently expose that functionality.
    // Doing it this way should still be quicker than reading it word-by-word
    // though.
    const u_int BFRSIZE = 64;
    Asura_PS2_QW axData[BFRSIZE];

    // Loop state.
    const Asura_PS2_QW* pxData = NULL;
    u_int uNumQWsLeft          = 0;
    STATE eState               = STATE_READING_ENV_HDR;
    u_int uNumMaterialsRead    = 0;
    u_int uCumModuleHdrCnt     = 0;
    u_int uEndOfThisMatModules = 0;
    u_int uNumOffsetsRead      = 0;
#if _DEBUG
    u_int uQWsReadAfterCurData = 0; // So num QWs read is (uCumQWsAfterCurData - uNumQWsLeft).
#endif

    while( STATE_DONE != eState )
    {
        // Need more data?
        while( !uNumQWsLeft )
        {
            Asura_FileAccess::ReadData( iHandle, axData, (BFRSIZE << 4) );
            pxData = axData;
            uNumQWsLeft = BFRSIZE;

#if _DEBUG
            uQWsReadAfterCurData += uNumQWsLeft;
#endif
        }

        // If we're not done reading then there should always be at least
        // another QW left.
        Asura_Assert( uNumQWsLeft, "" );

        switch( eState )
        {
            case STATE_READING_ENV_HDR:
            {
                // We assume the env hdr is a whole multiple of QWs.
                Asura_Assert( !(sizeof(Asura_PS2_Env_ExportData_EnvHdr) & 15), "" );
                const u_int uSizeInQWs = (sizeof(Asura_PS2_Env_ExportData_EnvHdr) >> 4);

                Asura_Assert( uSizeInQWs <= uNumQWsLeft, "" );
                s_xEnvHdr = *(reinterpret_cast<const Asura_PS2_Env_ExportData_EnvHdr*>(pxData));

                // Check the env version.
                if( ASURA_PS2_ENV_FILE_FORMAT_VERSION_NUMBER > s_xEnvHdr.uVersionID )
                {
                    Asura_Assert( false, "This environment export is out of date. Re-export it from tools." );
                }
                else if( ASURA_PS2_ENV_FILE_FORMAT_VERSION_NUMBER < s_xEnvHdr.uVersionID )
                {
                    Asura_Assert( false, "This environment export is newer than this code supports. Speak to your project's lead programmer." );
                }

                // Now that we know how many unique materials this
                // environment contains, we can allocate the material
                // header buffer.
                Asura_Assert( !s_pxMaterialHdrs, "" );
                s_pxMaterialHdrs = new Asura_PS2_Env_RuntimeData_MaterialHdr[s_xEnvHdr.uNumUniqueMaterials];

                // And the module header buffer...
                Asura_Assert( !s_pxModuleHdrs, "" );
                s_pxModuleHdrs = new Asura_PS2_Env_RuntimeData_ModuleHdr[s_xEnvHdr.uNumDmaChains];

                // Warn if we've run out of memory. (As with elsewhere in
                // this code, there's not really any way we can cope with
                // such an eventuality but we at least make sure it's made
                // obvious.)
                if( !s_pxMaterialHdrs || !s_pxModuleHdrs )
                {
                    Asura_OutputToDebugger_Retail(( "ERROR: Out of memory when trying to allocate material and module headers.\n" ));
                    Asura_Assert( false, "" );
                }

                pxData      += uSizeInQWs;
                uNumQWsLeft -= uSizeInQWs;

                eState = STATE_READING_MATERIAL_HDR;

                // We only need to break if we need more data.
                if( !uNumQWsLeft ) { break; }
            }

            case STATE_READING_MATERIAL_HDR:
            {
                // We assume that the material header is smaller than a QW
                // in size and so that there's at least one module index 
                // following it.
                Asura_Assert( sizeof(Asura_PS2_Env_ExportData_MaterialHdr) <= 12, "Material hdr is too large." );

                // We also assume it's a whole multiple of words in size to
                // make this easier.
                Asura_Assert( !(sizeof(Asura_PS2_Env_ExportData_MaterialHdr) & 3), "Material hdr isn't an integral number of words in size." );

                const Asura_PS2_Env_ExportData_MaterialHdr* pxMatHdr = reinterpret_cast<const Asura_PS2_Env_ExportData_MaterialHdr*>( pxData );

                s_pxMaterialHdrs[uNumMaterialsRead].uNumModules = pxMatHdr->usNumModules;
                s_pxMaterialHdrs[uNumMaterialsRead].uCumModules = uCumModuleHdrCnt;

                // Read and convert material ID.
                {
                    // Shouldn't have truly material-less polys in the
                    // environment anymore because tools should be
                    // substituting a non-existent one, meaning it'll end
                    // up using our Hideous Pink Texture (TM). This won't be
                    // true for old exports but then they won't be v7 format
                    // either and so will be unsupported.
                    Asura_Assert( ASURA_MATERIAL_NONE != pxMatHdr->sMaterialID, "Speak to Morgan." );

                    s_pxMaterialHdrs[uNumMaterialsRead].iMaterialID = Asura_Material_System::ConvertIndex( pxMatHdr->sMaterialID );
                }

                uEndOfThisMatModules += s_pxMaterialHdrs[uNumMaterialsRead].uNumModules;

                eState = STATE_READING_MODULE_HDRS;

                // No need to break because there's at least one module index
                // left in the current QW.
            }
            // End case STATE_READING_MATERIAL_HDR.

            case STATE_READING_MODULE_HDRS:
            {
                // We assume each module hdr occupies a single word here...
                Asura_Assert( 4 == sizeof(Asura_PS2_Env_ExportData_ModuleHdr), "" );

                // How many hdrs left in the cur QW?
                u_int uNumHdrsInCurQW;
                if( s_pxMaterialHdrs[uNumMaterialsRead].uCumModules == uCumModuleHdrCnt )
                {
                    uNumHdrsInCurQW = 4 - (sizeof(Asura_PS2_Env_ExportData_MaterialHdr) >> 2);
                }
                else
                {
                    uNumHdrsInCurQW = 4;
                }

                // How many should we read?
                u_int uMaxHdrsToRead = uNumHdrsInCurQW + ((uNumQWsLeft - 1) << 2);
                u_int uNumHdrsToRead = Asura_Maths::Min( uEndOfThisMatModules - uCumModuleHdrCnt, uMaxHdrsToRead );

                // Setup local input ptr before moving global one on.
                const Asura_PS2_Env_ExportData_ModuleHdr* pxHdrs = reinterpret_cast<const Asura_PS2_Env_ExportData_ModuleHdr*>( &pxData->u32s[4 - uNumHdrsInCurQW] );

                // Update loop state.
                {
                    u_int uNumQWsRead = 1 + ( uNumHdrsToRead > uNumHdrsInCurQW ? (((uNumHdrsToRead - uNumHdrsInCurQW) + 3) >> 2) : 0 );
                    Asura_Assert( uNumQWsRead <= uNumQWsLeft, "" );

                    pxData      += uNumQWsRead;
                    uNumQWsLeft -= uNumQWsRead;
                }

                // We should either be reading all the hdrs or running out of
                // data.
                Asura_Assert( !uNumQWsLeft || uCumModuleHdrCnt + uNumHdrsToRead == uEndOfThisMatModules, "" );

                // Read.
                while( uNumHdrsToRead-- )
                {
                    Asura_Assert( uCumModuleHdrCnt < s_xEnvHdr.uNumDmaChains, "DMA chain count in env header was wrong." );

                    // Copy index and tri count.
                    s_pxModuleHdrs[uCumModuleHdrCnt].iModuleIdx = pxHdrs->sModuleIdx;
                    s_pxModuleHdrs[uCumModuleHdrCnt].usNumTris  = pxHdrs->usNumTris;

                    // Set other fields to default values.
                    s_pxModuleHdrs[uCumModuleHdrCnt].uChainCALLAddr = 0;

                    ++pxHdrs;
                    ++uCumModuleHdrCnt;
                }

                // Have we read all the indices for this material?
                if( uCumModuleHdrCnt == uEndOfThisMatModules )
                {
                    // Any more materials?
                    if( s_xEnvHdr.uNumUniqueMaterials == ++uNumMaterialsRead )
                    {
                        eState = STATE_READING_REGION_OFFSETS;
                    }
                    else
                    {
                        eState = STATE_READING_MATERIAL_HDR;
                    }
                }

                break;
            }
            // End case STATE_READING_MODULE_HDRS.

            case STATE_READING_REGION_OFFSETS:
            {
                Asura_Assert( uNumOffsetsRead < ASURA_ENV_MAX_REGIONS, "Should have stopped if we've read all offsets." );

                u_int uNumOffsetsToRead = Asura_Maths::Min( ASURA_ENV_MAX_REGIONS - uNumOffsetsRead, (uNumQWsLeft << 2) );

                // This should always be a multiple of 4 since we're working
                // in QWs and max regions will be 32.
                Asura_Assert( !(uNumOffsetsToRead & 3), "" );

                // Setup local input ptr before updating global data ptr.
                const u_int* puInputOffsets = &pxData->u32s[0];

                // Update loop state.
                pxData      += (uNumOffsetsToRead >> 2);
                uNumQWsLeft -= (uNumOffsetsToRead >> 2);

                // Read.
                while( uNumOffsetsToRead-- )
                {
                    puRegionOffsets[uNumOffsetsRead++] = *puInputOffsets++;
                }

                // Read offsets for all regions?
                if( ASURA_ENV_MAX_REGIONS == uNumOffsetsRead )
                {
                    eState = STATE_DONE;
                }
                else
                {
                    // Otherwise the only reason for not having read them all
                    // is running out of data.
                    Asura_Assert( !uNumQWsLeft, "" );
                }

                break;
            }
            // End case STATE_READING_REGION_OFFSETS.

            case STATE_DONE:
            default:
                Asura_Assert( false, "Invalid state." );
                break;
        }
        // End switch( eState ).
    }
    // End while( STATE_DONE != eState ).

    // Check that we encountered the expected number of DMA chains.
    Asura_Assert( uCumModuleHdrCnt == s_xEnvHdr.uNumDmaChains, "DMA chain count in env header was wrong." );

    // Check that we've reached the beginning of the region data as expected,
    // ie. check we've reached the lowest of the region offsets.
#if _DEBUG
    {
        u_int uFirstNonEmpty;

        // Find the first non-empty region. Obviously this should be the
        // first one output and so the one with the lowest offset.
        const u_int* puOffsets = RgnManager::GetRgnOffsetPtr();
        for( uFirstNonEmpty = 0; uFirstNonEmpty < ASURA_ENV_MAX_REGIONS; ++uFirstNonEmpty )
        {
            if( 0xFFFFFFFFu != puOffsets[uFirstNonEmpty] )
            {
                break;
            }
        }

        // Check that the number of QWs read since the beginning of the env
        // header is equal to this lowest region offset.
        Asura_Assert( puOffsets[uFirstNonEmpty] == uQWsReadAfterCurData - uNumQWsLeft, "First non-empty region's offset is incorrect." );
    }
#endif

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    Asura_OutputToDebugger(( "Loaded.\n" ));
#endif
}

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
void Asura_PS2_Env_Store::ReopenFile( void )
{
    if( g_bDynamicLoadingIsReqd )
    {
        RgnManager::OpenFile( Project_GetFilename() );
    }
}
#endif

void Asura_PS2_Env_Store::Reset( void )
{
    RgnManager::Reset();

    // Deallocate header arrays.
    if( s_pxMaterialHdrs )
    {
        delete[] s_pxMaterialHdrs;
        s_pxMaterialHdrs = NULL;
    }
    if( s_pxModuleHdrs )
    {
        delete[] s_pxModuleHdrs;
        s_pxModuleHdrs = NULL;
    }

    s_xEnvHdr.uNumUniqueMaterials = 0;

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    s_bForceAllLoadsToCompleteOnNextUpdate = false;

    g_bFirstEnvStoreUpdate = true;
#endif
}

void Asura_PS2_Env_Store::SetDataOffset( u_int uOffset )
{
    RgnManager::SetDataOffset( uOffset );
}

void Asura_PS2_Env_Store::SetDataSize( u_int uSize )
{
    RgnManager::SetDataSize( uSize );
}

void Asura_PS2_Env_Store::Shutdown( void )
{
#if _DEBUG
    Asura_Assert( g_bEnvStoreInitialised, "Not initialised." );
#endif

    RgnManager::Shutdown();

    if( s_pxMaterialHdrs )
    {
        delete[] s_pxMaterialHdrs;
        s_pxMaterialHdrs = NULL;
    }
    if( s_pxModuleHdrs )
    {
        delete[] s_pxModuleHdrs;
        s_pxModuleHdrs = NULL;
    }

#if _DEBUG
    g_bEnvStoreInitialised = false;
#endif
}

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
void Asura_PS2_Env_Store::TemporarilyCloseFile( void )
{
    Asura_Assert( !g_bFirstEnvStoreUpdate, "Things won't have been setup properly if you call this so early." );
    if( g_bDynamicLoadingIsReqd )
    {
        CompleteOutstandingLoads();
        RgnManager::CloseFile();
    }
}
#endif

void Asura_PS2_Env_Store::Update( void )
{
#if _DEBUG
    Asura_Assert( g_bEnvStoreInitialised, "Not initialised." );
#endif

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING

    if( g_bFirstEnvStoreUpdate )
    {
        s_uVisRegionsLastFrame = 0;

        // Check if we actually need to do any dynamic loading. If all
        // regions are initially required then we won't bother. If we're
        // budgeting enough memory to accomodate all of them at some point
        // then we might as well just keep them loaded all the time. Note
        // that tools will modify the initially required regions to be all
        // regions if they're ever all simultaneously required.
        {
            g_bDynamicLoadingIsReqd = false;

            const u_int* puRgnOffsets = RgnManager::GetRgnOffsetPtr();
            for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
            {
                // Dynamic loading is necessary if this region is non-empty
                // and is not initially required.
                if( 0xFFFFFFFFu != puRgnOffsets[u] && !(s_xEnvHdr.uInitReqdRgns & (1 << u)) )
                {
                    g_bDynamicLoadingIsReqd = true;
                    break;
                }
            }
        }

        if( g_bDynamicLoadingIsReqd )
        {
            RgnManager::OpenFile( Project_GetFilename() );
        }

        g_bFirstEnvStoreUpdate = false;
    }

    if( !g_bDynamicLoadingIsReqd ) { return; }

    const u_int uLoadedRegionsBeforeUpdate = RgnManager::GetLoadedRegionBits();

    // (MP - 13.05.2003) NB: Since the environment stuff should be
    // getting updated before the module visibilities, the vis regions
    // for the last render, ie. the queued frame, are the current ones.

    const u_int uReqdRegions = GetReqdRgnsForThisFrame();
    const u_int uVisRegions  = GetVisRgnsForThisFrame();

    // Reqd regions should always be a superset of the vis regions.
    // (MP - 13.05.2003) It's possible some level exports will exist
    // where this isn't the case, due to a problem in tools where region
    // requirements aren't automatically recalculated when module linking
    // is modified. This should be fixed soon.
    Asura_Assert( !(uVisRegions & ~uReqdRegions), "" );

//    if( uReqdRegions != s_uReqdRegionsLastFrame )
//    {
//        Asura_OutputToDebugger(( "Reqd regions changed from 0x%08X to 0x%08X.\n", s_uReqdRegionsLastFrame, uReqdRegions ));
//    }

    RgnManager::Update( s_uVisRegionsLastFrame );
    
//    if (dbg_bUnloadAllUnrequiredRegions)
//    {
//        for( u_int uRgn = 0; uRgn < ASURA_ENV_MAX_REGIONS; ++uRgn )
//        {
//            if (RgnManager::IsRegionLoaded(uRgn) && !(uReqdRegions & (1 << uRgn)))
//            {
//                RgnManager::UnloadRegion(uRgn);
//            }
//        }
//        dbg_bUnloadAllUnrequiredRegions = false;
//    }

    if( s_bForceAllLoadsToCompleteOnNextUpdate )
    {
        CompleteOutstandingLoads();
        s_bForceAllLoadsToCompleteOnNextUpdate = false;
    }
    else
    {
        LoadIfNecessary( uReqdRegions );
    }

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
    {
        const u_int uLoadedRegionsAfterUpdate = RgnManager::GetLoadedRegionBits();
        if( uLoadedRegionsAfterUpdate != uLoadedRegionsBeforeUpdate )
        {
            Asura_OutputToDebugger(( "Loaded regions changed from 0x%08X to 0x%08X.\n", uLoadedRegionsBeforeUpdate, uLoadedRegionsAfterUpdate ));
        }
    }
#endif

#if _DEBUG
    RgnManager::Validate();
#endif

    s_uReqdRegionsLastFrame = uReqdRegions;
    s_uVisRegionsLastFrame  = uVisRegions;

#endif // #ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
}
