//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Env_Store.h
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     Functionality for dynamically loading and managing the environment
//     data. Also includes the data structures used at runtime, as opposed to
//     those used in data exported from tools (see Asura_PS2_Env_Shared.h.)
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_ENV_STORE_H___
#define ___ASURA_PS2_ENV_STORE_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Stdlib platform-specific.
#include <eetypes.h>

// Engine independent.
#include "Core/Asura_Handle.h"
#include "Debugging/Asura_Assert.h"

// Engine platform-specific.
#include "Mesh/Asura_PS2_Env_Shared.h"

// T Y P E S /////////////////////////////////////////////////////////////////

struct Asura_PS2_Env_RuntimeData_ModuleHdr
{
    int     iModuleIdx;
    u_int   uChainCALLAddr; // 0 => this module's region isn't completely loaded yet.
    u_short usNumTris;
};

struct Asura_PS2_Env_RuntimeData_MaterialHdr
{
    int   iMaterialID;
    u_int uNumModules;
    u_int uCumModules;      // Offset into module header array for the start of this material's modules.
};

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_PS2_Env_Store
{

    public:

        // T Y P E S /////////////////////////////////////////////////////////

        enum LOADREQD
        {
            LOADREQD_NO,
            LOADREQD_YES,
            LOADREQD_YES_BUT_NO_REGIONS_AVAILABLE_FOR_UNLOAD,
        };

        // C L A S S E S /////////////////////////////////////////////////////

        class It
        {

            public:

                // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////

                inline It( void );

                // M E T H O D S /////////////////////////////////////////////

                inline void Next( void );
                inline bool IsFinished( void ) const;
                inline void Restart( void );

                inline u_int GetNumModulesForCurMaterial( void ) const;

            protected:
                
                // S T A T E /////////////////////////////////////////////////

                u_int m_uItCurMaterialHdr;

        };

        class ConstIt : public It
        {

            public:

                // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////

                inline ConstIt( void );

                // M E T H O D S /////////////////////////////////////////////

                inline const Asura_PS2_Env_RuntimeData_MaterialHdr& GetCurMaterialHdr( void ) const;
                inline const Asura_PS2_Env_RuntimeData_ModuleHdr*  GetModuleHdrsForCurMaterial( void ) const;
        };

        class NonConstIt : public It
        {

            public:

                // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////

                inline NonConstIt( void );

                // M E T H O D S /////////////////////////////////////////////

                inline Asura_PS2_Env_RuntimeData_MaterialHdr& GetCurMaterialHdr( void ) const;
                inline Asura_PS2_Env_RuntimeData_ModuleHdr*  GetModuleHdrsForCurMaterial( void ) const;

        };

        // C L A S S   M E T H O D S /////////////////////////////////////////

        static void Initialise( void );
        static void Reset( void );
        static void Shutdown( void );

        static void SetDataOffset( u_int uOffset );
        static void SetDataSize( u_int uSize );
        static void LoadInitialData( Asura_Handle iHandle );
        static void Update( void );

        static void AbortCurrentLoad( void );
        static void CompleteOutstandingLoads( void );

        // Access methods.
        static inline const Asura_PS2_Env_ExportData_EnvHdr& GetEnvHdr( void );
        static inline const Asura_PS2_Env_RuntimeData_ModuleHdr* GetModuleHdrsForMaterial( const Asura_PS2_Env_RuntimeData_MaterialHdr& xMatHdr );
        static u_int GetNumUniqueMaterials( void );

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        static u_int GetReqdRgnsForThisFrame( void );
        static u_int GetVisRgnsForThisFrame( void );
        static u_int GetVisRgnsForLastFrame( void ) { return s_uVisRegionsLastFrame; }
        static void ClearVisRgnsForLastFrame( void ) { s_uVisRegionsLastFrame = 0; }

        static void ForceAllLoadsToCompleteOnNextUpdate( void ) { s_bForceAllLoadsToCompleteOnNextUpdate = true; }

        static void TemporarilyCloseFile( void );
        static void ReopenFile( void );
#endif

    private:

        // F R I E N D S /////////////////////////////////////////////////////

        friend It;
        friend ConstIt;
        friend NonConstIt;

        // C L A S S   M E T H O D S /////////////////////////////////////////

        static LOADREQD LoadIfNecessary( u_int uReqdRegions );
        static void LoadInitiallyReqdRegions( void );
        static void LoadRegionIndependentData( Asura_Handle iHandle );

		static const char* Project_GetFilename( void );

        // C L A S S   S T A T E /////////////////////////////////////////////

        static Asura_PS2_Env_ExportData_EnvHdr        s_xEnvHdr;  
        static Asura_PS2_Env_RuntimeData_MaterialHdr* s_pxMaterialHdrs;
        static Asura_PS2_Env_RuntimeData_ModuleHdr*   s_pxModuleHdrs;

#ifdef ASURA_PS2_SUPPORT_DYNAMIC_LOADING
        static u_int s_uReqdRegionsLastFrame;
        static u_int s_uVisRegionsLastFrame;

        static bool  s_bForceAllLoadsToCompleteOnNextUpdate;
#endif

};
// End class Asura_PS2_Env_Store.

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_PS2_Env_Store.
//

inline const Asura_PS2_Env_ExportData_EnvHdr& Asura_PS2_Env_Store::GetEnvHdr( void )
{
    return s_xEnvHdr;
}

inline const Asura_PS2_Env_RuntimeData_ModuleHdr* Asura_PS2_Env_Store::GetModuleHdrsForMaterial( const Asura_PS2_Env_RuntimeData_MaterialHdr& xMaterialHdr )
{
    Asura_Assert( xMaterialHdr.uCumModules < s_xEnvHdr.uNumDmaChains, "Invalid chain idx." );
    return &s_pxModuleHdrs[xMaterialHdr.uCumModules];
}

inline u_int Asura_PS2_Env_Store::GetNumUniqueMaterials( void )
{
    return s_xEnvHdr.uNumUniqueMaterials;
}

//
// Asura_PS2_Env_Store::It.
//

inline Asura_PS2_Env_Store::It::It( void )
{
    m_uItCurMaterialHdr = 0;
}

inline u_int Asura_PS2_Env_Store::It::GetNumModulesForCurMaterial( void ) const
{
    Asura_Assert( m_uItCurMaterialHdr < s_xEnvHdr.uNumUniqueMaterials, "Overrun material hdr array. Did you check IsFinished() before calling this?" );
    return s_pxMaterialHdrs[m_uItCurMaterialHdr].uNumModules;
}

inline bool Asura_PS2_Env_Store::It::IsFinished( void ) const
{
    return( m_uItCurMaterialHdr >= s_xEnvHdr.uNumUniqueMaterials );
}

inline void Asura_PS2_Env_Store::It::Next( void )
{
    ++m_uItCurMaterialHdr;
}

inline void Asura_PS2_Env_Store::It::Restart( void )
{
    m_uItCurMaterialHdr = 0;
}

//
// Asura_PS2_Env_Store::ConstIt.
//

inline Asura_PS2_Env_Store::ConstIt::ConstIt( void ) : It()
{
}

inline const Asura_PS2_Env_RuntimeData_MaterialHdr& Asura_PS2_Env_Store::ConstIt::GetCurMaterialHdr( void ) const
{
    Asura_Assert( m_uItCurMaterialHdr < s_xEnvHdr.uNumUniqueMaterials, "Overrun material hdr array. Did you check IsFinished() before calling this?" );
    return s_pxMaterialHdrs[m_uItCurMaterialHdr];
}

inline const Asura_PS2_Env_RuntimeData_ModuleHdr* Asura_PS2_Env_Store::ConstIt::GetModuleHdrsForCurMaterial( void ) const
{
    Asura_Assert( m_uItCurMaterialHdr < s_xEnvHdr.uNumUniqueMaterials, "Overrun material hdr array. Did you check IsFinished() before calling this?" );
    const u_int uIdx = s_pxMaterialHdrs[m_uItCurMaterialHdr].uCumModules;

    Asura_Assert( uIdx < s_xEnvHdr.uNumDmaChains, "Invalid chain idx." );
    return &s_pxModuleHdrs[uIdx];
}

//
// Asura_PS2_Env_Store::NonConstIt.
//

inline Asura_PS2_Env_Store::NonConstIt::NonConstIt( void ) : It()
{
}

inline Asura_PS2_Env_RuntimeData_MaterialHdr& Asura_PS2_Env_Store::NonConstIt::GetCurMaterialHdr( void ) const
{
    Asura_Assert( m_uItCurMaterialHdr < s_xEnvHdr.uNumUniqueMaterials, "Overrun material hdr array. Did you check IsFinished() before calling this?" );
    return s_pxMaterialHdrs[m_uItCurMaterialHdr];
}

inline Asura_PS2_Env_RuntimeData_ModuleHdr* Asura_PS2_Env_Store::NonConstIt::GetModuleHdrsForCurMaterial( void ) const
{
    Asura_Assert( m_uItCurMaterialHdr < s_xEnvHdr.uNumUniqueMaterials, "Overrun material hdr array. Did you check IsFinished() before calling this?" );
    const u_int uIdx = s_pxMaterialHdrs[m_uItCurMaterialHdr].uCumModules;

    Asura_Assert( uIdx < s_xEnvHdr.uNumDmaChains, "Invalid chain idx." );
    return &s_pxModuleHdrs[uIdx];
}

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_ENV_STORE_H___
