//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_DNAS.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     interface for the DNAS authentication process
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// this file's header
#include "Network/Asura_PS2_DNAS.h"

// Asura independent includes
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_LocalisedText.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"

// Asura PS2 includes
#include "Core/Asura_PS2_OS.h"	// for DLL loading

// project includes
#include "ASURA_PS2_DNAS_Text_Enum.h"	// need to have path to this file set in project settings.

// middleware includes
#include "Sony/dnas/include/dnas2.h"

// library includes
#include <string.h>
#include <libcdvd.h>

// C O N S T A N T S /////////////////////////////////////////////////////////
static const u_int uAUTHDATASIZE = 64*1024;
DEBUGVAR int dbg_iDnasTimeout = 30;

// G L O B A L S /////////////////////////////////////////////////////////////
static sceDNAS2Status_t g_xDNASStatus;

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
char* Asura_PS2_DNAS::s_pcAuthData;
Asura_PS2_DNAS::PublicStatus Asura_PS2_DNAS::s_ePublicStatus = Asura_PS2_DNAS::DnasStatus_None;

#ifdef ASURA_PS2_SONY_SUBMISSION
bool Asura_PS2_DNAS::s_bUseDebugServer = false;
#else
bool Asura_PS2_DNAS::s_bUseDebugServer = true;
#endif

// S T A T I C   C L A S S   F U N C T I O N S ///////////////////////////////
void Asura_PS2_DNAS::ForceUseDebugDNASServer(bool bDebug)
{
	// SNH - 03/11/2004 15:44:25 - see comment in Asura_PS2_DNAS.h
	s_bUseDebugServer = bDebug;
}

void Asura_PS2_DNAS::Initialise( void )
{
#ifdef ASURA_PS2_ENABLE_DNAS
	// network should already be intialised with eznetcnf
	Asura_LocalisedText::PAGERESETGROUP eBackupResetGroup = Asura_LocalisedText::GetCurPageResetGroup();
	Asura_LocalisedText::SetPageResetGroup(Asura_LocalisedText::PAGERESETGROUP_PS2_DNAS);
	bool bOK = Asura_ChunkLoading_System::LoadAsuraFile(Project_GetDNASFile());
	Asura_Assert(bOK, "failed to load dnas file");
	Asura_LocalisedText::SetPageResetGroup(eBackupResetGroup);

	if (!bOK)
	{
		// TODO - flag error!
		s_ePublicStatus = DnasStatus_Failed;
		return;
	}
#endif // #ifdef ASURA_PS2_ENABLE_DNAS

	BeginAuth();
}


void Asura_PS2_DNAS::BeginAuth( void )
{
#ifdef ASURA_PS2_ENABLE_DNAS
	Asura_Assert(s_ePublicStatus == DnasStatus_None, "BeginAuth called while authorisation in progress");

	// init DNAS library
	sceDNAS2TitleAuthInfo_t xAuthInfo;

	memset(&xAuthInfo, 0, sizeof(sceDNAS2TitleAuthInfo_t));
	
	xAuthInfo.line_type = sceDNAS2_T_INET;
	xAuthInfo.auth_data.ptr = s_pcAuthData;
	xAuthInfo.auth_data.size = uAUTHDATASIZE;
	xAuthInfo.pass_phrase.ptr = Project_GetPassPhrase();
	xAuthInfo.pass_phrase.size = Project_GetSizeOfPassPhrase();

	// sample code sets this to be the same as eenet EENET_DEV_PRIO (48).
	// it sets main thread priority to 127; ours is around 60.
	// newsgroups suggest eenet should have higher priority than DNAS, and 
	// the timeout thread should have higher priority than the DNAS thread.
	int iDNASThreadPriority = 50;
	
	sceDNAS2TimeoutInfo_t timeout_info;
	memset(&timeout_info, 0, sizeof(sceDNAS2TimeoutInfo_t));
	timeout_info.timeout = dbg_iDnasTimeout;
	timeout_info.priority = iDNASThreadPriority - 1;

//#ifdef ASURA_PS2_SONY_SUBMISSION
//	int iDebug = 0;
//	int iForceErrorCode = 0;
//#else
	int iDebug = s_bUseDebugServer ? 1 : 0;
	int iForceErrorCode = 0;
//#endif

	// don't set a breakpoint on this function call (or a call to any function 
	// in a DLL) as it interferes with the debugger's ability to patch the jal address.
	int iRet = sceDNAS2InitNoHDD(&xAuthInfo, iDNASThreadPriority, iDebug, iForceErrorCode, &timeout_info);

	/* NOTE */
	/* sceDNAS2InitNoHDD() with Debug flag should never be used in the final release, but for test program only */
	Asura_OutputToDebugger(("return value of sceDNAS2Init(): %d\n", iRet));
	if(iRet < 0)
	{
		// TODO - flag error!
		s_ePublicStatus = DnasStatus_Failed;
		return;
	}

	// don't set a breakpoint on this function call (or a call to any function 
	// in a DLL) as it interferes with the debugger's ability to patch the jal address.
	iRet = sceDNAS2AuthNetStart();
	Asura_OutputToDebugger(("ret of AuthNetStart() = %d\n", iRet));
	if(iRet < 0)
	{
		// TODO - flag error!
		s_ePublicStatus = DnasStatus_Failed;
		return;
	}

	s_ePublicStatus = DnasStatus_Authorising;
#else // not #ifdef ASURA_PS2_ENABLE_DNAS
	s_ePublicStatus = DnasStatus_Authorised;
#endif // #else not #ifdef ASURA_PS2_ENABLE_DNAS
}


void Asura_PS2_DNAS::Update( void )
{
#ifdef ASURA_PS2_ENABLE_DNAS
	// poll for auth complete
	if (s_ePublicStatus == DnasStatus_Authorising)
	{
		// don't set a breakpoint on this function call (or a call to any function 
		// in a DLL) as it interferes with the debugger's ability to patch the jal address.
		int iRet = sceDNAS2GetStatus(&g_xDNASStatus);
		if (iRet < 0)
		{
			Asura_OutputToDebugger(("DNAS failed: ret of sceDNAS2GetStatus() = %d\n", iRet));
			// TODO - flag error!
			s_ePublicStatus = DnasStatus_Failed;
			return;
		}

		if (g_xDNASStatus.code == sceDNAS2_S_NG || g_xDNASStatus.code == sceDNAS2_S_COM_ERROR)
		{
			Asura_OutputToDebugger(("DNAS failed: g_xDNASStatus.code = %d\n", g_xDNASStatus.code));
			// TODO - flag error!
			s_ePublicStatus = DnasStatus_Failed;
			return;
		}

		if (g_xDNASStatus.code == sceDNAS2_S_END)
		{
			Asura_OutputToDebugger(("DNAS success. subcode:type %d, inst_result %d\n",
									sceDNAS2Status_SubCode_TYPE(g_xDNASStatus),
									sceDNAS2Status_SubCode_INST_RESULT(g_xDNASStatus)));
			// finished! - change state
			s_ePublicStatus = DnasStatus_Authorised;
			return;
		}
	}
#endif // #ifdef ASURA_PS2_ENABLE_DNAS
}
	
//	while(1){
//		if((ret = sceDNAS2GetStatus(&status)) < 0){
//			scePrintf("ret of sceDNAS2GetStatus() = %d\n", ret);
//			error = 1;
//			goto error;
//		}
//		if((ret = memcmp(&prev_status, &status, sizeof(sceDNAS2Status_t))) != 0){
//			/* status changed */
//			scePrintf("code: %d, sub_code: %d, progress: %d\n",
//				status.code, status.sub_code, status.progress);
//			if(status.code == sceDNAS2_S_END){
//				scePrintf("subcode:type %d, inst_result %d\n",
//					sceDNAS2Status_SubCode_TYPE(status),
//					sceDNAS2Status_SubCode_INST_RESULT(status));
//				memcpy(&prev_status, &status, sizeof(sceDNAS2Status_t));
//				break;
//			}
//			else if(status.code == sceDNAS2_S_NG || status.code == sceDNAS2_S_COM_ERROR){
//				error = 1;
//				break;
//			}
//		}

//
//	int timeout = 0;
//	int proxy_state = 0;
//	const char *proxy_host = 0;
//	u_short proxy_port = 0;
//	
//	int  ret, error = 0;
//	sceDNAS2TitleAuthInfo_t auth_info;
//	sceDNAS2Status_t status, prev_status;
//	sceDNAS2TimeoutInfo_t timeout_info;
//	int fd, proxy_state_check = 0;
//	char auth_data[1024*64];
//	char proxy_host_check[512];
//	u_short proxy_port_check;
//
//	memset(&status, 0, sizeof(sceDNAS2Status_t));
//	memset(&prev_status, 0, sizeof(sceDNAS2Status_t));
//	memset(&auth_info, 0, sizeof(sceDNAS2TitleAuthInfo_t));
//	if(timeout > 0){
//		memset(&timeout_info, 0, sizeof(sceDNAS2TimeoutInfo_t));
//		timeout_info.timeout = timeout;
//		timeout_info.priority = 96;
//	}
//	auth_info.line_type = sceDNAS2_T_INET;
//	auth_info.auth_data.ptr = auth_data;
//	auth_info.auth_data.size = sizeof(auth_data);
//	auth_info.pass_phrase.ptr = pass_phrase;
//	auth_info.pass_phrase.size = sizeof(pass_phrase);
//	scePrintf("before sceDNAS2Init()\n");
//	if(timeout > 0){
//		ret = sceDNAS2InitNoHDD(&auth_info, 48, 1, 0, &timeout_info);
//	}else{
//		ret = sceDNAS2InitNoHDD(&auth_info, 48, 1, 0, NULL);
//	}
//	/* NOTE */
//	/* sceDNAS2InitNoHDD() with Debug flag should never be used in the final release, but for test program only */
//	scePrintf("return value of sceDNAS2Init(): %d\n", ret);
//	if(ret < 0){
//		error = 1;
//		goto error;
//	}
//
//	/*set proxy parameter*/
//	if(proxy_state){
//		ret = sceDNAS2SetProxy(1, proxy_host, proxy_port);
//		scePrintf("return value of sceDNAS2SetProxy = %d\n", ret);
//		if(ret < 0){
//			error = 1;
//			goto error;
//		}
//	}
//
//	/*get proxy parameter*/
//	if(proxy_state){
//		ret = sceDNAS2GetProxy(&proxy_state_check, proxy_host_check, sizeof(proxy_host_check), &proxy_port_check);
//		scePrintf("return value of sceDNASGetProxy = %d\n", ret);
//		if(ret < 0){
//			error = 1;
//			goto error;
//		}
//		if(proxy_state_check){
//			scePrintf("Proxy Info : Server = %s Port = %d\n", proxy_host_check, proxy_port_check);
//		}else{
//			scePrintf("Proxy Info : no use \n");
//		}
//	}
//
//	ret = sceDNAS2AuthNetStart();
//	scePrintf("ret of AuthNetStart() = %d\n", ret);
//	if(ret < 0){
//		error = 1;
//		goto error;
//	}
//
//	while(1){
//		if((ret = sceDNAS2GetStatus(&status)) < 0){
//			scePrintf("ret of sceDNAS2GetStatus() = %d\n", ret);
//			error = 1;
//			goto error;
//		}
//		if((ret = memcmp(&prev_status, &status, sizeof(sceDNAS2Status_t))) != 0){
//			/* status changed */
//			scePrintf("code: %d, sub_code: %d, progress: %d\n",
//				status.code, status.sub_code, status.progress);
//			if(status.code == sceDNAS2_S_END){
//				scePrintf("subcode:type %d, inst_result %d\n",
//					sceDNAS2Status_SubCode_TYPE(status),
//					sceDNAS2Status_SubCode_INST_RESULT(status));
//				memcpy(&prev_status, &status, sizeof(sceDNAS2Status_t));
//				break;
//			}
//			else if(status.code == sceDNAS2_S_NG || status.code == sceDNAS2_S_COM_ERROR){
//				error = 1;
//				break;
//			}
//		}
//		memcpy(&prev_status, &status, sizeof(sceDNAS2Status_t));
//
//		DelayThread(1000 * 1000);
//	}
//
//error:
//	if(0 < error)
//		scePrintf("*********** DNAS failed. **********\n");
//	else
//		scePrintf("*********** DNAS succeeded.!!!! *************\n");
//
//	ret = sceDNAS2Shutdown();
//	scePrintf("ret of sceDNAS2Shutdown() is %d\n", ret);
//
//	if(0 < error)
//		return -1;
//	else
//		return 0;
//}



void Asura_PS2_DNAS::Shutdown( void )
{
#ifdef ASURA_PS2_ENABLE_DNAS
	bool bFailed = (s_ePublicStatus == DnasStatus_Failed);

// (MP - 25.04.2005) This code doesn't appear to work and BHD failed a
// submission partly due to it. Seeing as we should generally cope elsewhere
// with the tray being ejected, in that future loads should just sit and
// wait until it returns, this code is redundant anyway.
//	if(bFailed)
//	{
//		// (MH - 16.01.2004) if we failed on a disk info error reinit the CD in case
//		// it was ejected and reinserted.
//		if (GetErrorNumber() == sceDNAS2_SS_INVALID_MEDIA)
//		{
//			sceCdInit(SCECdINIT);
//			switch( sceCdGetDiskType() )
//			{
//				case SCECdPS2DVD:
//					Asura_OutputToDebugger(( "Disk type: DVD.\n" ));
//					sceCdMmode( SCECdDVD );
//					break;
//
//				default:
//					Asura_OutputToDebugger(( "Disk type: CD.\n" ));
//					sceCdMmode( SCECdCD );
//					break;  
//			}
//		}
//	}

	// shutdown DNAS library
	CancelCurrentQuery();

	// unload DLL
	Asura_PS2_OS::ShutdownAndUnloadDll( Project_GetDNASDLLName() );
#endif // #ifdef ASURA_PS2_ENABLE_DNAS

	Asura_LocalisedText::ResetPages(Asura_LocalisedText::PAGERESETGROUP_PS2_DNAS);

	delete[] s_pcAuthData;
	s_pcAuthData = NULL;
	s_ePublicStatus = DnasStatus_None;
}


void Asura_PS2_DNAS::Retry( void )
{
	bool bFailed = (s_ePublicStatus == DnasStatus_Failed);

	CancelCurrentQuery();

// (MP - 25.04.2005) This code doesn't appear to work and BHD failed a
// submission partly due to it. Seeing as we should generally cope elsewhere
// with the tray being ejected, in that future loads should just sit and
// wait until it returns, this code is redundant anyway.
//	if(bFailed)
//	{
//		// (MH - 16.01.2004) if we failed on a disk info error reinit the CD in case
//		// it was ejected and reinserted.
//		if (GetErrorNumber() == sceDNAS2_SS_INVALID_MEDIA)
//		{
//			sceCdInit(SCECdINIT);
//			switch( sceCdGetDiskType() )
//			{
//				case SCECdPS2DVD:
//					Asura_OutputToDebugger(( "Disk type: DVD.\n" ));
//					sceCdMmode( SCECdDVD );
//					break;
//
//				default:
//					Asura_OutputToDebugger(( "Disk type: CD.\n" ));
//					sceCdMmode( SCECdCD );
//					break;  
//			}
//		}
//	}
	
	BeginAuth();
	s_ePublicStatus = DnasStatus_Authorising;
}


void Asura_PS2_DNAS::CancelCurrentQuery( void )
{
#ifdef ASURA_PS2_ENABLE_DNAS
	// shutdown DNAS library

	// don't set a breakpoint on this function call (or a call to any function 
	// in a DLL) as it interferes with the debugger's ability to patch the jal address.
	int iRet = sceDNAS2GetStatus(&g_xDNASStatus);
	if (iRet < 0)
	{
		Asura_OutputToDebugger(("DNAS Restart: ret of sceDNAS2GetStatus() = %d\n", iRet));
	}
	else
	{
		Asura_OutputToDebugger(("DNAS Restart: g_xDNASStatus.code = %d\n", g_xDNASStatus.code));
	}
	
	// do we need to abort? Yes, if the DNAS process didn't end.
	if ( (g_xDNASStatus.code != sceDNAS2_S_NG) &&
		 (g_xDNASStatus.code != sceDNAS2_S_COM_ERROR) && 
		 (g_xDNASStatus.code != sceDNAS2_S_END) )
	{
		if (g_xDNASStatus.code != sceDNAS2_S_INIT)
		{
			// don't set a breakpoint on this function call (or a call to any function 
			// in a DLL) as it interferes with the debugger's ability to patch the jal address.
			iRet = sceDNAS2Abort();
			Asura_OutputToDebugger(("ret of sceDNAS2Abort() is %d\n", iRet));
			// -6 sceDNAS2_NOT_IN_COMM   DNAS thread is not communicating  
			// -8 sceDNAS2_NOT_INITIALIZED  
		}
		
		// need to wait for state to be sceDMAS2_S_INIT before calling sceDNAS2Shutdown()
		while (g_xDNASStatus.code != sceDNAS2_S_INIT)
		{
			// don't set a breakpoint on this function call (or a call to any function 
			// in a DLL) as it interferes with the debugger's ability to patch the jal address.
			sceDNAS2GetStatus(&g_xDNASStatus);
			Asura_PS2_OS::DelayThread(20);
		}
	}

	// don't set a breakpoint on this function call (or a call to any function 
	// in a DLL) as it interferes with the debugger's ability to patch the jal address.
	iRet = sceDNAS2Shutdown();
	Asura_OutputToDebugger(("ret of sceDNAS2Shutdown() is %d\n", iRet));
	Asura_PS2_OS::DelayThread(20);
#endif // #ifdef ASURA_PS2_ENABLE_DNAS

	s_ePublicStatus = DnasStatus_None;
}


void* Asura_PS2_DNAS::GetAuthDataPtr( u_int uBytesRequired )
{
	Asura_Assert(uBytesRequired == uAUTHDATASIZE, "incorrect size for authdata");
	if (uBytesRequired != uAUTHDATASIZE)
	{
		return NULL;
	}
	Asura_Assert(!s_pcAuthData, "auth data already exists");
	if (!s_pcAuthData)
	{
		s_pcAuthData = new char[uAUTHDATASIZE];
	}
	return s_pcAuthData;
}



// internal types
struct Asura_PS2_DNAS_ErrorInfo
{
	int m_iErrorCode;
	int m_iTextID;
};

// internal data
Asura_PS2_DNAS_ErrorInfo g_auBodyInfo[] = 
{
/*    { -101, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_101 }, 
    { -102, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_102 }, 
    { -103, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_103 }, 
    { -104, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_104 }, 
    { -105, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_105 }, 
    { -106, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_106 }, 
    { -107, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_107 }, 
    { -108, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_108 }, 
    { -401, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_401 }, 
    { -402, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_402 }, 
    { -403, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_403 }, 
    { -404, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_404 }, 
    { -601, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_601 }, 
    { -602, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_602 }, 
    { -603, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_603 }, 
    { -610, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_610 }, 
    { -611, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_611 }, 
    { -612, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_612 }, 
    { -613, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_613 }, 
    { -614, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_614 }, 
    { -615, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_615 }, 
    { -616, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_616 }, 
    { -617, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_617 }, 
    { -618, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_618 }, 
    { -619, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_619 }, 
    { -620, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_620 }, 
    { -621, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_621 }, 
    { -622, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_622 }, 
    { -623, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_623 }, 
    { -624, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_624 }, 
    { -625, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_625 }, 
    { -626, ASURA_PS2_DNAS_TEXT_ID_NET_ERROR_626 }, 
	*/
};

Asura_PS2_DNAS_ErrorInfo g_auFooterInfo[] = 
{
	/*
    { -102, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_102B }, 
    { -103, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_103B }, 
    { -104, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_104B }, 
    { -105, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_105B }, 
    { -106, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_106B }, 
    { -107, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_107B }, 
    { -108, ASURA_PS2_DNAS_TEXT_ID_AUTH_ERROR_108B }, 
    { -401, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_401B }, 
    { -402, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_402B }, 
    { -403, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_403B }, 
    { -404, ASURA_PS2_DNAS_TEXT_ID_MACHINE_ERROR_404B }, 
	*/
};


// static functions
static const Asura_UnicodeChar* GetDnasText(int iTextID)
{
	const Asura_LocalisedText_Page* pxPage = Asura_LocalisedText::GetPage(Asura_GetHashID("ASURA_PS2_DNAS"));
	Asura_Assert(pxPage, "Dnas text page not loaded");
	if (pxPage)
	{
		return pxPage->GetString(iTextID);
	}
	return L"";	
}

static u_int GetHeaderTextID()
{
	//if (g_xDNASStatus.code == sceDNAS2_S_COM_ERROR)
	//{
	//	return ASURA_PS2_DNAS_TEXT_ID_NETWORK_ERROR;
	//}
	return 0;
}

static u_int GetBodyTextID(int iErrorCode)
{
	// see if the error is one we know about. If not use catch all.
	for (Asura_PS2_DNAS_ErrorInfo* pxErrorInfo = g_auBodyInfo; 
		pxErrorInfo < g_auBodyInfo + sizeof(g_auBodyInfo);
		++pxErrorInfo)
	{
		if (pxErrorInfo->m_iErrorCode == iErrorCode)
		{
			return pxErrorInfo->m_iTextID;
		}
	}
	return 0;//ASURA_PS2_DNAS_TEXT_ID_OTHER_ERROR;
}

static u_int GetFooterTextID(int iErrorCode)
{
	// catch-all for error in range (-1099, -800)
	if ( (iErrorCode >= (-1099)) && (iErrorCode <= (-800)) )
	{
		return 0;// ASURA_PS2_DNAS_TEXT_ID_OTHER_ERROR_B;
	}

	// see if this is listed in the footer info
	for (Asura_PS2_DNAS_ErrorInfo* pxErrorInfo = g_auFooterInfo; 
		pxErrorInfo < g_auFooterInfo + sizeof(g_auFooterInfo);
		++pxErrorInfo)
	{
		if (pxErrorInfo->m_iErrorCode == iErrorCode)
		{
			return pxErrorInfo->m_iTextID;
		}
	}
		
	return 0;//ASURA_PS2_DNAS_TEXT_ID_NUM_TAGS;
}

const Asura_UnicodeChar* Asura_PS2_DNAS::GetErrorType( void )
{
//	if (g_xDNASStatus.code == sceDNAS2_S_NG)
//	{
//		return GetDnasText(ASURA_PS2_DNAS_TEXT_ID_DNAS_ERROR);
//	}
//	else if (g_xDNASStatus.code == sceDNAS2_S_COM_ERROR)
//	{
//		return GetDnasText(ASURA_PS2_DNAS_TEXT_ID_NETWORK_ERROR);
//	}
//	Asura_Assert(false, "invalid error type");
	return L"";
}

int Asura_PS2_DNAS::GetErrorNumber( void )
{
	Asura_Assert((g_xDNASStatus.code == sceDNAS2_S_NG) || (g_xDNASStatus.code == sceDNAS2_S_COM_ERROR), "invalid error type");
	Asura_Assert(s_ePublicStatus == DnasStatus_Failed, "not in failed state");
	return g_xDNASStatus.sub_code;
}

const Asura_UnicodeChar* Asura_PS2_DNAS::GetErrorBody( void )
{
	int iTextID = GetBodyTextID(g_xDNASStatus.sub_code);
	return GetDnasText(iTextID);
}

const Asura_UnicodeChar* Asura_PS2_DNAS::GetErrorFooter( void )
{
//	int iTextID = GetFooterTextID(g_xDNASStatus.sub_code);
//	if (iTextID == ASURA_PS2_DNAS_TEXT_ID_NUM_TAGS)
//	{
		// no footer for this error
		return L"";
//	}
//	else 
//	{
//		return GetDnasText(iTextID);
//	}
}

