//
// Asura_Platform_NetTransport_System.cpp
//
// PC/PS2 version. Note that since nearly all code is shared between PC and PS2 (apart 
// from calls to WSAStartup/WSACleanup which are only needed on PC) I have just enclosed 
// those calls in _WIN32 defines. Also we now have Platform_GetNetAdaptorIP().
//
// Created by MH - 19.07.2004
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Network/Asura_NetTransport_System.h"

#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK

#ifdef _WIN32 
// library includes
#include <winsock.h>
#endif

// Asura includes
#include "Network/Asura_Network_Session.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_NetSocket.h"
#include "Network/Asura_NetSocket_BSD.h"
#include "Network/Asura_NetConnection.h"

#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Logfile.h"


DEBUGVAR int dbg_iNumPortsToScan = 8;

struct Asura_Network_LANBroadcastServerRequest
{
	u_short     m_usSequenceNumber;
	u_short		m_usChecksum;
	u_short		m_usMagicID;		// = ASURA_NETTRANSPORT_MSG_ENUMHOSTS
	u_short		m_usGameVersion;
	u_short		m_usPadding;
	u_int		m_uApplicationGuid;
};

class Asura_Network_LANBroadcastServerReplyHeader
{
	public:

		u_short m_usSequenceNumber;
		u_short	m_usChecksum;
		u_short m_usMagicID;		// ASURA_NETTRANSPORT_MSG_HOSTREPLY
		u_short m_usGameVersion;
		u_int m_uApplicationGuid;
		u_int m_uServerSessionID;

		u_short m_usDataSize;		// this data is followed by this many bytes of user data
		Asura_UnicodeChar m_wszSessionName[uASURA_NETWORK_MAX_SESSION_NAME + 1];
		u_char m_ucNumPlayers;
		u_char m_ucMaxPlayers;
};

void Asura_NetTransport_System::Platform_Initialise()
{
#ifdef _WIN32 
	WSADATA data;
	WSAStartup(0x0101, &data);
#endif
}

void Asura_NetTransport_System::Platform_Update()
{
}

void Asura_NetTransport_System::Platform_Shutdown()
{
#ifdef _WIN32 
	WSACleanup();
#endif
}

void Asura_NetTransport_System::Platform_StartServer(bool bInternetGame)
{
}

void Asura_NetTransport_System::Platform_StopServer()
{
}

bool Asura_NetTransport_System::Platform_SendEnumHostsBroadcast()
{
	bool iResult = false;
		
	for(int i = -1; i < dbg_iNumPortsToScan; ++i)
	{
		u_int uBroadcastAddr = INADDR_BROADCAST;	// 255.255.255.255
		u_short usPort = (i == -1) ? ASURA_NETSOCKET_DEFAULTPORT : static_cast<u_short>(ASURA_NETSOCKET_FIRST_DYNAMIC_PORT + i);
		Asura_Network_LANBroadcastServerRequest xBroadcastMsg;
		xBroadcastMsg.m_usSequenceNumber = 0;
		// checksum filled out later
		xBroadcastMsg.m_usGameVersion = Project_GetVersionID();
		xBroadcastMsg.m_usMagicID = ASURA_NETTRANSPORT_MSG_ENUMHOSTS;
		xBroadcastMsg.m_uApplicationGuid = Project_GetApplicationGuid();

		// checksum on length of data after SeqNum and checksum
		xBroadcastMsg.m_usChecksum = ComputeChecksum( reinterpret_cast<const char*>(&xBroadcastMsg.m_usMagicID), sizeof(xBroadcastMsg) - 4 );

		Asura_Assert(s_pxSocket, "No socket");
		if(s_pxSocket)
		{
			iResult |= s_pxSocket->Send(uBroadcastAddr, usPort, reinterpret_cast<const char*>(&xBroadcastMsg), sizeof(xBroadcastMsg));
			Asura_Assert(iResult, "failed to send broadcast message");
			
			AddDirectSendBytes(sizeof(xBroadcastMsg));
		}
	}
	s_fLastSessionRequestBroadcastTime = Asura_Timers::GetLocalUnscaledTime();
	return iResult;
}

void Asura_NetTransport_System::Platform_ProcessEnumHostsRequest( Asura_NetAddress xAddress, u_short usPort, const char* pcData, u_int uDataLength )
{
	const Asura_Network_LANBroadcastServerRequest* pxBroadcastRequestMsg = reinterpret_cast
		<const Asura_Network_LANBroadcastServerRequest*>(pcData);

	// Ignore requests from other games or versions
	if (pxBroadcastRequestMsg->m_uApplicationGuid != Project_GetApplicationGuid())
	{
		return;
	}

	if (pxBroadcastRequestMsg->m_usGameVersion != Project_GetVersionID())
	{
		return;
	}

	if (uDataLength != sizeof(Asura_Network_LANBroadcastServerRequest))
	{
		Asura_Assert(false, "bad data length in enum hosts");
		return;
	}

	// check the checksum
	if (pxBroadcastRequestMsg->m_usChecksum != 
		ComputeChecksum( reinterpret_cast<const char*>(&pxBroadcastRequestMsg->m_usMagicID), uDataLength - 4 ))
	{
		Asura_Assert(false, "bad checksum in enum hosts");
		return;
	}
	
	Asura_OutputToDebugger(("Received enum hosts request from %0X:%i with %i bytes of data.\n", xAddress, usPort, uDataLength));
	Platform_SendEnumHostsResponse( xAddress, usPort );
}

void Asura_NetTransport_System::Platform_SendEnumHostsResponse( Asura_NetAddress xAddress, u_short usPort)
{
	// (MH - 01.07.2004) ensure the session data is fresh
	Project_RefreshSessionData();

	u_int uSizeOfReplyData = sizeof(Asura_Network_LANBroadcastServerReplyHeader) + s_uSessionDataSize;
	char* const pcReplyData = new char[uSizeOfReplyData];

	Asura_Network_LANBroadcastServerReplyHeader* pxHeader = reinterpret_cast
		<Asura_Network_LANBroadcastServerReplyHeader*>(pcReplyData);
	pxHeader->m_usSequenceNumber = 0;
	// checksum filled out later
	pxHeader->m_usMagicID = ASURA_NETTRANSPORT_MSG_HOSTREPLY;
	pxHeader->m_usGameVersion = Project_GetVersionID();
	pxHeader->m_uApplicationGuid = Project_GetApplicationGuid();
	pxHeader->m_uServerSessionID = s_uServerSessionID;
	pxHeader->m_usDataSize = static_cast<u_short>(s_uSessionDataSize);
	pxHeader->m_ucNumPlayers = static_cast<u_char>(Asura_Network::GetNumberOfPlayers());
	pxHeader->m_ucMaxPlayers = static_cast<u_char>(Asura_Network::GetMaxNumberOfPlayers());
	Asura_PrintfToBfr(pxHeader->m_wszSessionName, L"%s", uASURA_NETWORK_MAX_SESSION_NAME + 1, true, Asura_Network::GetServerName());

	u_char* pucSessionData = reinterpret_cast<u_char*>(pxHeader + 1);
	memcpy(pucSessionData, s_pucSessionData, s_uSessionDataSize);

	// checksum on length of data after SeqNum and checksum
	pxHeader->m_usChecksum = ComputeChecksum( reinterpret_cast<const char*>(&pxHeader->m_usMagicID), uSizeOfReplyData - 4 );

	bool bResult = s_pxSocket->Send(xAddress, usPort, pcReplyData, uSizeOfReplyData);
	Asura_Assert(bResult, "failed to send broadcast message");

	Asura_NetTransport_System::AddDirectSendBytes(uSizeOfReplyData);

	delete[] pcReplyData;
}

void Asura_NetTransport_System::Platform_ProcessEnumHostsResponse( Asura_NetAddress xAddress, u_short usPort, const char* pcData, int iDataSize )
{
	const Asura_Network_LANBroadcastServerReplyHeader* pxHeader = reinterpret_cast
		<const Asura_Network_LANBroadcastServerReplyHeader*>(pcData);
	const u_char* pucSessionData = reinterpret_cast<const u_char*>(pxHeader + 1);

	// check project and version
	if (pxHeader->m_uApplicationGuid != Project_GetApplicationGuid()) return;
	if (pxHeader->m_usGameVersion != Project_GetVersionID()) return;

	// check data length
	if (static_cast<u_int>(iDataSize) != sizeof(Asura_Network_LANBroadcastServerReplyHeader) + pxHeader->m_usDataSize)
	{
		Asura_Assert(false, "wrong data length in enum hosts response");
		return;
	}

	// check the checksum
	if (pxHeader->m_usChecksum != ComputeChecksum( reinterpret_cast<const char*>(&pxHeader->m_usMagicID), iDataSize - 4 ))
	{
		Asura_Assert(false, "bad checksum in enum hosts response");
		return;
	}

	if (!s_ppxSessionsStorage) return;
	
	// work out ping based on time since message was sent. Possible problem of
	// overlapping response and request...
	u_int uPing = Asura_Maths::FloatToInt(1000.0f * (Asura_Timers::GetLocalUnscaledTime() - s_fLastSessionRequestBroadcastTime));

	Asura_Network_SessionInfo* pxNewSession =  new Asura_Network_SessionInfo(xAddress, 
			usPort, uPing, pxHeader->m_uServerSessionID, NULL, 0, pucSessionData, pxHeader->m_usDataSize, pxHeader->m_wszSessionName, 
			pxHeader->m_ucNumPlayers, pxHeader->m_ucMaxPlayers);

	for (u_int u = 0; u < s_uNumberOfActiveSessions; ++u)
	{
		if (s_ppxSessionsStorage[u]->Platform_IsSameHost(pxNewSession))
		{
			// SNH: copied from xbox version. Replace the session details.
			delete s_ppxSessionsStorage[u];
			s_ppxSessionsStorage[u] = pxNewSession;
			return;
		}
	}

	// this is a new server
	if (s_uNumberOfActiveSessions < uASURA_NETWORK_MAX_NUMBER_OF_SESSIONS)
	{
		s_ppxSessionsStorage[s_uNumberOfActiveSessions++] = pxNewSession;
	}
	else
	{
		delete pxNewSession;
	}
}

void Asura_NetTransport_System::Platform_Client_ClientConnected(Asura_NetConnection* pxConnection)
{
	// KOR - This empty function was originally used for initiating the peer-to-peer voice message
	// exchange, which was moved to the Platform_StartGame (see below) to stop it overflowing during load.
	// It may come in useful later or on other platforms though.
}

void Asura_NetTransport_System::Platform_Client_ClientDisconnected(Asura_NetConnection* pxConnection)
{
	// Do nothing
}

void Asura_NetTransport_System::Platform_StartGame()
{
}

bool Asura_NetTransport_System::Platform_HandleMessage( Asura_NetAddress xAddress, u_short usPort, const char* pcMessage, int iLength)
{
	return false;
}

u_int Asura_NetTransport_System::Platform_GetNetAdaptorIP( int iAdapter ) 
{ 	
#ifdef _WIN32 
	char ac[80];
	if (gethostname(ac, sizeof(ac)) == SOCKET_ERROR)
	{
		Asura_Assertf(false, ("WINSOCK : Error (%d) retriving the Local Host Name",  WSAGetLastError()));
		return 0;
	}
	Asura_OutputToDebugger(("WINSOCK : Local Host Name : \"%s\"\n", ac));

	struct hostent *phe = gethostbyname(ac);

    if (phe == 0) 
	{
		Asura_Assertf(false, ("WINSOCK : Error (%d) in Host Name Lookup",  WSAGetLastError()));
		return 0;
	}
	
	Asura_OutputToDebugger(("WINSOCK : Known IP Addresses :\n"));
		
	in_addr addr;
        
    for (int i = 0; phe->h_addr_list[i] != 0; ++i) 
	{
        in_addr temp;

        memcpy(&temp, phe->h_addr_list[i], sizeof(temp));

        Asura_OutputToDebugger(("- Adapter [%d] : IP %s\n", i, inet_ntoa(temp)));

		if (i == iAdapter)
			addr = temp;
    }
	Asura_OutputToDebugger(("WINSOCK : Using Adapter [%d] : IP %s\n", iAdapter, inet_ntoa(addr)));

	return addr.s_addr;
#else
	return 0;
#endif
}

const char* Asura_NetTransport_System::Platform_GetStringFromAddress(Asura_NetAddress xAddress, unsigned short usPort)
{
	return Asura_NetSocket_BSD::GetStringFromAddress( xAddress, usPort );
}

void Asura_NetTransport_System::Platform_GetAddressFromString(const char* szAddress, Asura_NetAddress* pxAddress, unsigned short* pusPort)
{
	Asura_NetSocket_BSD::GetAddressFromString( szAddress, pxAddress, pusPort );
}

Asura_NetSocket* Asura_NetTransport_System::Platform_AllocateNetSocket( u_int uMaxMachines )
{
	return new Asura_NetSocket_BSD( uMaxMachines );
}

bool Asura_Network_SessionInfo::Platform_IsSameHost(Asura_Network_SessionInfo* pxSession)
{
	return ((m_xHostAddr == pxSession->m_xHostAddr) && (m_usHostPort == pxSession->m_usHostPort));
}

#endif //#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
