//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_GSMRenderer.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS2_GSMRenderer.h"

// Stdlib platform-specific.
#include <eestruct.h>
#include <libvu0.h>

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_Render.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_GSMDefines.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"
#include "Rendering/Asura_PS2_Shadows.h"
#include "Rendering/Asura_PS2_ShadowBuffer.h"
#include "Rendering/Asura_PS2_VU1.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// The following is rather ugly but we do have a lot of state and it's quite
// a lot faster to keep it in SPR. Note that some relatively strong
// assumptions are made in the memory layout about buffer sizes, etc. This
// also needs to be kept in sync with the constants in Asura_PS2_GSMDefines.h.

// General.
#define g_uMeshFlags         (*reinterpret_cast<u_int*       >(0x70000000))
#define g_uPrevMeshFlags     (*reinterpret_cast<u_int*       >(0x70000004)) // Mesh flags from previous GSMSendPendingData().
#define g_uPendingFlags      (*reinterpret_cast<u_int*       >(0x70000008))
#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
#define g_uPendingVtxCount   (*reinterpret_cast<u_int*       >(0x7000000C))
#endif
// Camera.
#define g_uCamVisFlags       (*reinterpret_cast<u_int*       >(0x70000010))
#define g_uCamFlagsForVU     (*reinterpret_cast<u_int*       >(0x70000014))
#define g_uNumCamsVisFor     (*reinterpret_cast<u_int*       >(0x70000018))
#define g_bO2WIsIdentity     (*reinterpret_cast<bool*        >(0x7000001C))
#define g_uCurVUCamBfr       (*reinterpret_cast<u_int*       >(0x70000020))
// Texture.
#define g_bTexUploaded       (*reinterpret_cast<bool*        >(0x70000024))
#define g_iCurBaseTex        (*reinterpret_cast<int*         >(0x70000028))
#define g_iCurSTOvTex        (*reinterpret_cast<int*         >(0x7000002C))
#define g_pxDualTexRegs      ( reinterpret_cast<Asura_PS2_QW*>(0x70000030)) // Up to 10 textures with 3 reg QWs each.
#define g_uDualTexCount      (*reinterpret_cast<u_int*       >(0x70000210)) 
// Renderstate
#define g_uCurVURSBfr        (*reinterpret_cast<u_int*       >(0x70000214))
#define g_uNumPassesPerTri   (*reinterpret_cast<u_int*       >(0x70000218))
#define g_uCurMaterialFlags  (*reinterpret_cast<u_int*       >(0x7000021C))
#define g_pxRSBfr            ( reinterpret_cast<Asura_PS2_QW*>(0x70000220)) // Up to 10 QWs of renderstate.
#define g_pxSTOvParams       ( reinterpret_cast<Asura_PS2_QW*>(0x700002C0))
#define g_pxYParams          ( reinterpret_cast<Asura_PS2_QW*>(0x700002D0))
// Matrices.
#define g_pxO2W              ( reinterpret_cast<Asura_PS2_QW*>(0x700002F0)) // 4x4 matrix.
#define g_pxMatTempBfr       ( reinterpret_cast<Asura_PS2_QW*>(0x70000330)) //     "
#define g_pxTexMatBfr        ( reinterpret_cast<Asura_PS2_QW*>(0x70000370)) //     "
// Lighting.
#define g_pxOutLightBfr      ( reinterpret_cast<Asura_PS2_QW*>(0x700003B0)) // 2 QWs per light for up to 7 lights or else one SH matrix.
#define g_pvIntSHBfr         ( reinterpret_cast<void*        >(0x70000490)) // Space for one intermediate 2nd order SHs, for intermediate calculations.
#define g_uNumLightQWs       (*reinterpret_cast<u_int*       >(0x700004D0))
#define g_uCurVULightBfr     (*reinterpret_cast<u_int*       >(0x700004D4))
// Shadows.
#define g_uCurVUShadowBfr    (*reinterpret_cast<u_int*       >(0x700004D8))
#define g_pxShadowMatBfr     ( reinterpret_cast<Asura_PS2_QW*>(0x700004E0)) // 4x4 matrix.
#define g_pxShadowPosAndSize ( reinterpret_cast<Asura_PS2_QW*>(0x70000520))
#define g_fShadowAlpha       (*reinterpret_cast<float*       >(0x70000530))
#define g_fShadowLightLevel  (*reinterpret_cast<float*       >(0x70000534))
#define g_uShadowBuffers     (*reinterpret_cast<u_int*       >(0x70000538)) // count of shadow buffers in use for each pass
#define g_piShadowBufIDs     ( reinterpret_cast<u_int*       >(0x7000053c)) // space for 5 entries 
#define g_pxShadowMatList    ( reinterpret_cast<Asura_PS2_QW*>(0x70000550)) // 16 lots of 4x4 matrix
// next available address 0x70000950

// Total.
#define TOTAL_SPR_QWS_REQD                                    (0x0000095 )

// Compile-time checks against stuff in the header.
#if TOTAL_SPR_QWS_REQD > ASURA_PS2_GSM_SPR_QWS_RESERVED
    #error Out of sync with header.
#endif

// Special case for light count. Only required for point lighting.
#define g_uNumLights ((g_pxOutLightBfr + ASURA_PS2_GSM__OFFSET__QW_CONTAINING_QTY)->u32s[3])

#define MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT(puExtern, uNumMSCALs) (++(uNumMSCALs), ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(puExtern))

#if _DEBUG
    #define GET_STRIP_RENDER_MSCAL() \
        ( g_bSkipNextStrip \
            \
            ? (g_bSkipNextStrip = false, ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(VU1_GSM_NOP)) \
            \
            : ( g_bBreakOnNextStrip \
                ? (g_bBreakOnNextStrip = false, ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(VU1_GSM_BEGIN_STRIP_BREAK)) \
                : ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(VU1_GSM_BEGIN_STRIP) ) )
#else
    #define GET_STRIP_RENDER_MSCAL() ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR(VU1_GSM_BEGIN_STRIP)
#endif

// T Y P E S /////////////////////////////////////////////////////////////////

enum PENDING
{
    PENDING_UPLOAD_TEX_SYNC          =  1 <<  0,
    PENDING_UPLOAD_RENDERSTATE       =  1 <<  1,
    PENDING_UPLOAD_CAM               =  1 <<  2,
    PENDING_UPLOAD_LIGHT             =  1 <<  3,
    PENDING_UPLOAD_MESH_FLAGS        =  1 <<  4,
    PENDING_UPLOAD_CAM_FLAGS         =  1 <<  5,
    PENDING_UPLOAD_SHADOW_PROJECTION =  1 <<  6,	// new one
    PENDING_UPLOAD_SHADOW_DATA       =  1 <<  7,
    PENDING_UPLOAD_ST_OV_PARAMS      =  1 <<  8,
    PENDING_UPLOAD_Y_PARAMS          =  1 <<  9,

};

// C O N S T A N T S /////////////////////////////////////////////////////////

// This is used to invalidate the cached material flags to ensure that we
// end up resending renderstate next time if something has changed the
// cached renderstate buffer but won't be picked up by the check inside
// SetRenderstate(). See, for example, DisableZWrites(). The value is chosen
// to be one we should never use here.
static u_int uINVALID_MATERIAL_FLAG = AsuraMaterialFlag_SubdividePolygon;

// G L O B A L S /////////////////////////////////////////////////////////////

// Subroutines on the VU.
extern u_int VU1_GSM_INIT[]                   __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_SWITCH_CAM_DATA_BFR[]    __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_SWITCH_LIGHT_DATA_BFR[]  __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_SWITCH_SHADOW_DATA_BFR[] __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_SWITCH_RENDERSTATE_BFR[] __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_LOAD_MESH_FLAGS[]        __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_LOAD_CAM_FLAGS[]         __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_BEGIN_STRIP[]            __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_BEGIN_STRIP_BREAK[]      __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_NOP[]                    __attribute__(( section(".vudata") ));
extern u_int VU1_GSM_RENDER_SHADOW_QUAD[]     __attribute__(( section(".vudata") ));

static Asura_PS2_QW* g_pxVifPtrFromLastEnd;
static u_int         g_uQWCFromLastEnd;
static u_int         g_uFrameOfLastEnd;

// Debugging controls.
DEBUGVAR bool dbg_bGSMFillSPRInBegin     = false;
DEBUGVAR bool dbg_bGSMFillVU1MemInBegin  = false;
DEBUGVAR bool dbg_bGSMForcePointLighting = false;
#if _DEBUG
static bool g_bBreakOnNextStrip = false;
static bool g_bSkipNextStrip    = false;
#endif

// Optimisation controls.
DEBUGVAR bool dbg_bGSMOptimiseOutRedundantFLUSHes       = true;
DEBUGVAR bool dbg_bGSMOptimiseOutRedundantRSUploads     = true;
DEBUGVAR bool dbg_bGSMOptimiseOutRedundantTexRegUploads = true;

//DEBUGVAR bool dbg_bDisplayGSMShadowBufferRGB    = false;
//DEBUGVAR bool dbg_bDisplayGSMShadowBufferAlpha  = false;

// controls for variables which you may want to see the effect of changes immediately
DEBUGVAR float dbg_fShadowYOffset = ASURA_PS2_SHADOW_OFFSET_Y;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static Asura_PS2_QW* CalculateAndWriteCamData( Asura_PS2_QW* pxOut, u_int uCamIdx, bool bO2WIsIdentity, bool bReflective );

static inline void FillQWWithNOPs( Asura_PS2_QW* pxOut )
{
    pxOut->u128 = 0;
}

static inline void FillSTScaleAndTranslationQW( float fScaleS, float fScaleT, float fTransS, float fTransT )
{
    g_pxSTOvParams->f32s[0] = fTransS;
    g_pxSTOvParams->f32s[1] = fTransT;
    g_pxSTOvParams->f32s[2] = fScaleS;
    g_pxSTOvParams->f32s[3] = fScaleT;
}

#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
static inline void RenderPendingStrip( Asura_PS2_DmaTag* pxTagWithSpareWords )
{
    pxTagWithSpareWords->SetSpareWords( SCE_VIF1_SET_ITOP(g_uPendingVtxCount, false), GET_STRIP_RENDER_MSCAL() );
    g_uPendingVtxCount = 0;
}
#endif

//static void RenderShadowQuad( void );

static void GSMSendPendingData( bool bWillMSCNTForNextStrip );

static inline void UpdatePassesPerTri( u_int uMeshFlags )
{
    g_uNumPassesPerTri = 1 + (uMeshFlags & ASURA_PS2_GSM__MASK__PASS_1 ? 1 : 0) + (uMeshFlags & ASURA_PS2_GSM__MASK__PASS_2 ? 1 : 0);
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_PS2_GSMRenderer::SetPointLightingParametersForPrelitMesh( const Light* pxLights, u_int uNumLights, float fAlpha )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewPendingFlags = g_uPendingFlags;
    u_int uNewMeshFlags    = uOldMeshFlags;

    Asura_Assert( uNumLights <= ASURA_PS2_GSM__MAX_POINT_LIGHTS, "" );
    if( uNumLights )
    {
        const bool          bO2WIsIdentity = g_bO2WIsIdentity;
        const Asura_PS2_QW* pxO2W          = g_pxO2W;

        Asura_PS2_QW* pxOut = g_pxOutLightBfr;
        for( u_int u = 0; u < uNumLights; ++u )
        {
            const Light& xLight = pxLights[u];

            // Light position.
            float fX = xLight.afLightPos[0];
            float fY = xLight.afLightPos[1];
            float fZ = xLight.afLightPos[2];
            if( bO2WIsIdentity )
            {
                // Just copy to output then.
                pxOut->f32s[0] = fX;
                pxOut->f32s[1] = fY;
                pxOut->f32s[2] = fZ;
            }
            else
            {
                // Rotate position into objectspace (by applying transpose of
                // O->W matrix.)

                fX -= (pxO2W + 3)->f32s[0];
                fY -= (pxO2W + 3)->f32s[1];
                fZ -= (pxO2W + 3)->f32s[2];

                pxOut->f32s[0] = (pxO2W + 0)->f32s[0] * fX + (pxO2W + 0)->f32s[1] * fY + (pxO2W + 0)->f32s[2] * fZ;
                pxOut->f32s[1] = (pxO2W + 1)->f32s[0] * fX + (pxO2W + 1)->f32s[1] * fY + (pxO2W + 1)->f32s[2] * fZ;
                pxOut->f32s[2] = (pxO2W + 2)->f32s[0] * fX + (pxO2W + 2)->f32s[1] * fY + (pxO2W + 2)->f32s[2] * fZ;
            }

            // Jam range into unused w component of position QW.
            const float fRange = xLight.fLightRange;
            pxOut->f32s[3] = fRange;
            ++pxOut;
                
            // RGB.
            const float fScaleOverRange = 255.0f / fRange;
            pxOut->f32s[0] = xLight.afLightRGB[0] * fScaleOverRange;
            pxOut->f32s[1] = xLight.afLightRGB[1] * fScaleOverRange;
            pxOut->f32s[2] = xLight.afLightRGB[2] * fScaleOverRange;
            ++pxOut;
        }

        g_uNumLightQWs = pxOut - g_pxOutLightBfr;

        uNewMeshFlags    |= ASURA_PS2_GSM__FLAG__POINT_LIT;
        uNewPendingFlags |= PENDING_UPLOAD_LIGHT;
    }
    else
    {
        uNewMeshFlags    &= ~ASURA_PS2_GSM__FLAG__POINT_LIT;
        uNewPendingFlags &= ~PENDING_UPLOAD_LIGHT;
    }

    uNewMeshFlags |= ASURA_PS2_GSM__FLAG__PRELIT;

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
    g_uNumLights    = uNumLights;
}

//void Asura_PS2_GSMRenderer::CalculateSHForMesh( const Asura_Vector_3& xMeshCentre, float fAlpha )
//{
//  Asura_SphericalHarmonics_Order2_RGB* pxWorldspaceSH = static_cast<Asura_SphericalHarmonics_Order2_RGB*>( g_pvIntSHBfr );
//  Asura_GI_Solution::GetTotalIrradiance( *pxWorldspaceSH, xMeshCentre );
//
//  g_pxOutLightBfr->f32s[3] = 128.0f * fAlpha;
//
//  RecalculateSHForChangedObjectToWorldMat();
//}

// special one off to allow shadow system to prevent lights from being required - otherwise it confuses the renderer
void Asura_PS2_GSMRenderer::CancelLightingUpload( void )
{
	g_uPendingFlags &= ~PENDING_UPLOAD_LIGHT;
}

void Asura_PS2_GSMRenderer::UseExistingSHForMesh( const Asura_SphericalHarmonics_Order2_RGB& xSH, float fAlpha )
{
    // Copy SH.
    *(static_cast<Asura_SphericalHarmonics_Order2_RGB*>(g_pvIntSHBfr)) = xSH;

    // Record alpha.
    g_pxOutLightBfr->f32s[3] = 128.0f * fAlpha;

    // Scale values and apply rotation as required.
    RecalculateSHForChangedObjectToWorldMat();
}

void Asura_PS2_GSMRenderer::RecalculateSHForChangedObjectToWorldMat( void )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewMeshFlags    = ( uOldMeshFlags & ~(ASURA_PS2_GSM__FLAG__POINT_LIT | ASURA_PS2_GSM__FLAG__PRELIT) );
    u_int uNewPendingFlags = g_uPendingFlags;

    const Asura_SphericalHarmonics_Order2_RGB* pxSrcSH = static_cast<Asura_SphericalHarmonics_Order2_RGB*>( g_pvIntSHBfr );

    Asura_PS2_QW* const pxLightBfr = g_pxOutLightBfr;

    // Setup VU SH matrix.
    {
        Asura_PS2_QW* pxOut = pxLightBfr;

        // We need to scale and transpose the coefficients from independent
        // code as we copy them.
        for( u_int u = 0; u < 4; ++u )
        {
            for( u_int v = 0; v < 3; ++v )
            {
                pxOut->f32s[v] = 255.0f * pxSrcSH->m_afCoefficient[(v << 2) + u];
            }
            ++pxOut;
        }
    }

    // Rotate if required.
    if( !g_bO2WIsIdentity )
    {
        const Asura_PS2_QW* pxO2W = g_pxO2W;

        // Load rotation part of O->W matrix.
        asm( "LQC2 $vf01,  0(%0)" : : "r"(pxO2W) );
        asm( "LQC2 $vf02, 16(%0)" : : "r"(pxO2W) );
        asm( "LQC2 $vf03, 32(%0)" : : "r"(pxO2W) );

        // Load relevant bits of SH data.
        asm( "LQC2 $vf04, 16(%0)" : : "r"(pxLightBfr) );
        asm( "LQC2 $vf05, 32(%0)" : : "r"(pxLightBfr) );
        asm( "LQC2 $vf06, 48(%0)" : : "r"(pxLightBfr) );

        // Rotate.
        // r1,g1,b1
        asm( "VMULAx.xyz  ACC,   $vf04, $vf01x" );
        asm( "VMADDAy.xyz ACC,   $vf05, $vf01y" );
        asm( "VMADDz.xyz  $vf07, $vf06, $vf01z" );
        // r2,g2,b2
        asm( "VMULAx.xyz  ACC,   $vf04, $vf02x" );
        asm( "VMADDAy.xyz ACC,   $vf05, $vf02y" );
        asm( "VMADDz.xyz  $vf08, $vf06, $vf02z" );
        // r3,g3,b3
        asm( "VMULAx.xyz  ACC,   $vf04, $vf03x" );
        asm( "VMADDAy.xyz ACC,   $vf05, $vf03y" );
        asm( "VMADDz.xyz  $vf09, $vf06, $vf03z" );

        // Store rotated SH data back to light bfr.
        asm( "SQC2 $vf07, 16(%1)" : "=m"((pxLightBfr + 1)->u128) : "r"(pxLightBfr) );
        asm( "SQC2 $vf08, 32(%1)" : "=m"((pxLightBfr + 2)->u128) : "r"(pxLightBfr) );
        asm( "SQC2 $vf09, 48(%1)" : "=m"((pxLightBfr + 3)->u128) : "r"(pxLightBfr) );
    }

    // VU (specifically the LightVtxWithSH macro) requires w component of
    // second row to be 0.
    (pxLightBfr + 1)->f32s[3] = 0.0f;

    uNewPendingFlags |= PENDING_UPLOAD_LIGHT;

    // Need to refresh shadow data, if shadow is enabled.
    if( uOldMeshFlags & ASURA_PS2_GSM__FLAG__SHADOW )
    {
        g_fShadowLightLevel = Asura_PS2_Shadows::CalculateShadowLightLevelFromSH( *pxSrcSH, pxLightBfr->f32s[3] / 128.0f );
        uNewPendingFlags |= PENDING_UPLOAD_SHADOW_DATA;
        // Note that the recalculation of the shadow projection matrix will
        // be done at upload time.

		// place light level into shadow buffer structure since it will be needed again later
		if (Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer)
		{
			Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer->SetLightLevel( g_fShadowLightLevel );
		}
    }

    g_uNumLightQWs  = 4;
    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_GSMRenderer::Begin( void )
{
#if _DEBUG
    // Runtime checks against stuff in the header.
    if( ASURA_PS2_GSM_NUM_CAMS_VIS_FOR_ADDR != &g_uNumCamsVisFor )
    {
        ASURA_PS2_DEBUG_BREAK();
    }
    if( ASURA_PS2_GSM_NUM_PASSES_PER_TRI_ADDR != &g_uNumPassesPerTri )
    {
        ASURA_PS2_DEBUG_BREAK();
    }
#endif

    // May end doing End() then another Begin() immediately afterwards in
    // some circumstances. This will mean we end up doing an unnecessary
    // FLUSH. Deal with that...
    if( g_pxVifPtrFromLastEnd == Asura_PS2_DeferredRendering::GetCurVifBfrPtr() && Asura_Timers::GetFrameNumber() == g_uFrameOfLastEnd )
    {
        Asura_PS2_DeferredRendering::ReturnVifBuffer( g_pxVifPtrFromLastEnd - 1, 1 );
    }
    g_pxVifPtrFromLastEnd = NULL;

    // If renderer was already there, we assume some setup stuff has
    // already been done.
    const bool bNeedFullSetup = Asura_PS2_VU1::LoadRenderer( Asura_PS2_VU1::RENDERER_GSM );

    // Reserve some of the beginning of SPR.
    (void) Asura_PS2_SPR_Alloc( 0, TOTAL_SPR_QWS_REQD ); // The cast fixes warning in release.

    if( bNeedFullSetup )
    {
        // Couple of bits for debugging purposes, for checking everything's
        // correctly initialised, etc.
        if( dbg_bGSMFillSPRInBegin )
        {
            const u_long128 uFillValue = ASURA_PS2_INIT_ULONG128( 0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDuLL );
            u_long128* puOut = reinterpret_cast<u_long128*>( 0x70000000 );
            for( u_int u = ASURA_PS2_GSM_SPR_QWS_RESERVED; u; --u )
            {
                *puOut++ = uFillValue;
            }
        }
        if( dbg_bGSMFillVU1MemInBegin )
        {
            Asura_PS2_VU1::FillMem( ASURA_PS2_INIT_ULONG128(0xCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDCDuLL) );
        }
        
        const u_int uNumCams = Asura_Camera::GetNumberOfCameras();

        const u_int uNumQWsForCamUNPACK    = uNumCams * 5;
        const u_int uNumQWsForGIFTagUNPACK = ASURA_PS2_GSM__SIZE__GIF_TAGS;
        const u_int uNumQWsForShadowUNPACK = 1;

        const u_int uNumQWsForDMA = uNumQWsForCamUNPACK
                                  + 1 + uNumQWsForGIFTagUNPACK
                                  + 1 + uNumQWsForShadowUNPACK
                                  + 1 + Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

        const u_int uTotalQWs = 1 + uNumQWsForDMA;

        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumQWsForDMA,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_CNT,
                                                             false,
                                                             0,
                                                             false,
                                                             SCE_VIF1_SET_STCYCL(1, 1, false),
                                                             SCE_VIF1_SET_UNPACK(ASURA_PS2_GSM__ADDR__CONST_CAM_DATA, uNumQWsForCamUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false) );

        // Const cam data.
        for( u_int uCam = 0; uCam < uNumCams; ++uCam )
        {
            pxOut = Asura_PS2_GS::GetViewportSetupGifCmds( uCam, Asura_PS2_GS::CONTEXT_1, pxOut );

            (pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrScaleVec(uCam) );
            (pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrTransVec(uCam) );
            (pxOut++)->u128 = reinterpret_cast<const u_long128&>( Asura_PS2_VU1::GetGBClipToScrClipScaleVec(uCam) );
        }

        // Setup BASE and OFFSET and start UNPACK of GIF tags.
        pxOut->u32s[0] = SCE_VIF1_SET_BASE( ASURA_PS2_GSM__ADDR__INPUT_BFR_0, false );
        pxOut->u32s[1] = SCE_VIF1_SET_OFFSET( ASURA_PS2_GSM__SIZE__INPUT_BFR, false );
        pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_GSM__ADDR__GIF_TAGS, uNumQWsForGIFTagUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false );
        ++pxOut;

        // Geometry setup GIF tag. The default NLOOP should be for the base
        // pass; we'll fix it up on VU1 for anything else.
        {
            Asura_PS2_GifTag* pxGeomSetupGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
            pxGeomSetupGifTag->SetExceptRegs( ASURA_PS2_GSM__RS_OUT_0_NLOOP, true, false, 0, SCE_GIF_PACKED, 1 );
            pxGeomSetupGifTag->uAllRegs = SCE_GIF_PACKED_AD;
        }
        
        // Geometry draw GIF tag. Note that NLOOP is setup in code.
        {
            Asura_PS2_GifTag* pxGeomDrawGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );

            u_int uPrimType = SCE_GS_PRIM_TRISTRIP;
            bool  bTME      = true;

#ifndef ASURA_PS2_SONY_SUBMISSION
            if( Asura_Render::GetWireframe() )
            {
                uPrimType = SCE_GS_PRIM_LINESTRIP;
                bTME      = false;
            }
#endif

            const u_int uPRIM = SCE_GS_SET_PRIM( uPrimType, 1, bTME, false, true, false, 0, 0, 0 );
            pxGeomDrawGifTag->SetExceptRegs( 0, true, true, uPRIM, SCE_GIF_PACKED, 3 );

            pxGeomDrawGifTag->uAllRegs = (SCE_GS_ST << 0) | (SCE_GS_RGBAQ << 4) | (SCE_GS_XYZ2 << 8);
        }

        // Shadow setup GIF tag. EOP isn't required since shadow geometry
        // will immediately follow.
        {
            Asura_PS2_GifTag* pxShadowSetupGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
            pxShadowSetupGifTag->SetExceptRegs( 1, false, false, 0, SCE_GIF_PACKED, 1 );
            pxShadowSetupGifTag->uAllRegs = SCE_GIF_PACKED_AD;
        }

        // Shadow draw GIF tag. Again, NLOOP is setup in code.
        {
            Asura_PS2_GifTag* pxShadowDrawGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
            pxShadowDrawGifTag->SetExceptRegs( 0, true, true, Asura_PS2_Shadows::Proj_GetRenderToBfrPRIMSetting(), SCE_GIF_PACKED, 1 );
            pxShadowDrawGifTag->uAllRegs = SCE_GS_XYZ2;
        }

        // Run INIT program and start UNPACK of shadow constant data.
        pxOut->u32s[0] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_INIT );
        pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_GSM__ADDR__CONST_SHADOW_DATA, uNumQWsForShadowUNPACK, Asura_PS2_VU1::VIFPK_V4_32, false );
        ++pxOut;

        // Shadow RGBAQ.
        pxOut->u64s[0] = Asura_PS2_Shadows::Proj_GetRenderToBfrRGBAQ();
        pxOut->u64s[1] = SCE_GS_RGBAQ;
        ++pxOut;

        // DIRECT for default renderstate setup.
        pxOut->u32s[0] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
        pxOut->u32s[3] = SCE_VIF1_SET_DIRECT( Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS, false );
        ++pxOut;

        Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1, pxOut, true, false );
        pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

        Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );

        // Reset buffer indices. We assume these are still in sync if
        // renderer was already present on VU; the only time they'll change
        // is when we run one of the VU1_GSM_SWITCH... subroutines.
        g_uCurVURSBfr     = 0;
        g_uCurVUCamBfr    = 0;
        g_uCurVULightBfr  = 0;
        g_uCurVUShadowBfr = 0;

        // Fill in constant elements of renderstate bfr.
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA      )->u64s[1] = SCE_GS_ALPHA_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX1       )->u64s[1] = SCE_GS_TEX1_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX0       )->u64s[1] = SCE_GS_TEX0_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_MIPTBP1    )->u64s[1] = SCE_GS_MIPTBP1_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEST       )->u64s[1] = SCE_GS_TEST_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_SELF_ILLUM_ALPHA)->u64s[1] = SCE_GS_ALPHA_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_ALPHA     )->u64s[1] = SCE_GS_ALPHA_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX1      )->u64s[1] = SCE_GS_TEX1_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX0      )->u64s[1] = SCE_GS_TEX0_1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_MIPTBP1   )->u64s[1] = SCE_GS_MIPTBP1_1;

        // Fill in register identifiers for dual texture bits.
        {
            Asura_PS2_QW* pxOut = g_pxDualTexRegs;
            for( u_int u = 0; u < ASURA_PS2_GSM__MAX_DUAL_TEXTURE_PALETTE_SIZE; ++u )
            {
                (pxOut + 0)->u64s[1] = SCE_GS_TEX1_1;
                (pxOut + 1)->u64s[1] = SCE_GS_MIPTBP1_1;
                (pxOut + 2)->u64s[1] = SCE_GS_TEX0_1;
                pxOut += 3;
            }
        }

        // Setup some defaults.
        g_uMeshFlags       = 0;
        g_uPrevMeshFlags   = 0;
        g_uPendingFlags    = PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM | PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS | PENDING_UPLOAD_LIGHT;
#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
        g_uPendingVtxCount = 0;
#endif
        g_uCamFlagsForVU   = 0;
        g_bO2WIsIdentity   = true;

        // (MP - 05.05.2005) We used to default this to 0 but that breaks in
        // the (incredibly perverse) case where the first mesh being rendered
        // has no polys or whatever, eg. a hierarchy with 0 sections, so
        // that we end up falling into the case in End() where there's some
        // pending data left, just from the defaults here. Sending 0 light
        // QWs is invalid; setting it to this value is value and also has the
        // added bonus of ensuring the light qty, if relevant, gets (re)set.
        g_uNumLightQWs = ASURA_PS2_GSM__OFFSET__QW_CONTAINING_QTY + 1; 

        // Setup default renderstate.
        g_iCurBaseTex = ASURA_TEXTURE_NONE;
        g_iCurSTOvTex = ASURA_TEXTURE_NONE;
        g_uCurMaterialFlags = uINVALID_MATERIAL_FLAG;
        SetRenderstate( 0 );
    }
    else // not bNeedFullSetup
    {
#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
        Asura_Assert( !g_uPendingVtxCount, "Forgot to call End()?" );
#endif

        // Reset transform.
        g_bO2WIsIdentity = true;

        // Ensure all mesh flags other than those setup in SetRenderstate()
        // are reset. Leave those on since otherwise we'll interfere with the
        // caching based on material flags.
        g_uMeshFlags &= ~(   ASURA_PS2_GSM__FLAG__SHADOW
                           | ASURA_PS2_GSM__FLAG__POINT_LIT
                           | ASURA_PS2_GSM__FLAG__PRELIT
                           | ASURA_PS2_GSM__MASK__ANY_VTX_MANIPULATIONS );

        g_uPendingFlags |= PENDING_UPLOAD_CAM | PENDING_UPLOAD_MESH_FLAGS;
    }
    
    g_uNumPassesPerTri = 1;
    g_uDualTexCount    = 0;
}

void Asura_PS2_GSMRenderer::SetDualTexturePalette( const int* piTextureIDs, u_int uNumTextureIDs, u_int uTexBfr, u_int uMMAG, u_int uMMIN )
{
    // Assumptions:
    //
    //     - Second textures are all 128x128 8-bit palettised images with
    //       up to three lower mipmap levels.
    //
    //     - The dual texture palette need contain no more than
    //       ASURA_PS2_ENV_MAX_DUAL_TEXTURE_PALETTE_SIZE entries. Note that
    //       here that constraint means the *scene*, not the environment,
    //       must contain no more than this.
    //
    //     - The calling renderer is using a texture buffer layout such that
    //       the buffer at the specified index is at least 256Kb in size.

    // NB: Currently, setting up multiple palettes in a row using the same
    // texture buffer layout and chosen buffer will likely lead to
    // synchronisation issues, assuming the second or subsequent palettes
    // go down PATH3.

    Asura_Assert( uNumTextureIDs <= ASURA_PS2_GSM__MAX_DUAL_TEXTURE_PALETTE_SIZE, "" );
    g_uDualTexCount = uNumTextureIDs;

    // Need to manually invalidate the texture buffer we're going to be
    // filling.
    Asura_PS2_TextureManagement::InvalidateTextureBuffer( uTexBfr );

    const u_int uBfrTBP           = Asura_PS2_TextureManagement::GetBfrTBP( uTexBfr );
    const u_int uMaxSpaceForCLUTs = ASURA_PS2_NUM_64W_UNITS_FOR_CLUT * ASURA_PS2_GSM__MAX_DUAL_TEXTURE_PALETTE_SIZE;
    const u_int uFirstCLUTTBP     = uBfrTBP + Asura_PS2_TextureManagement::GetBfrSizeIn64wUnits(uTexBfr) - uMaxSpaceForCLUTs;

    u_int uTBP     = uBfrTBP;
    u_int uCLUTTBP = uFirstCLUTTBP;

    Asura_PS2_QW* pxRegOut = g_pxDualTexRegs;
    while( uNumTextureIDs )
    {
        const int iTexID = *piTextureIDs++;

        // Hide the texture ID away in the upper word of the first QW of
        // the registers. This is ignored when we send it to the GS; it just
        // uses the lower 8 bits of the upper doubleword for the register
        // address and doesn't process the rest.
        (pxRegOut + 0)->i32s[3] = iTexID;

        // Cache the GS register settings for the texture. We may end up
        // switching between them quite a lot so this'll save some work later
        // on.
        {
            // Note that the upper 64 bits, ie. the register identifiers, 
            // will already have been filled out in Begin().

            Asura_PS2_TextureManagement::GetTEX1Setting( &(pxRegOut + 0)->u64s[0], iTexID, uMMAG, uMMIN );
            Asura_PS2_TextureManagement::GetMIPTBP1SettingForTBP( &(pxRegOut + 1)->u64s[0], iTexID, uTBP );
            Asura_PS2_TextureManagement::GetTEX0SettingForTBP( &(pxRegOut + 2)->u64s[0], iTexID, uTBP, uCLUTTBP );

            pxRegOut += 3;
        }

        bool bUploaded = Asura_PS2_TextureManagement::UploadTextureToAddress( iTexID, &uTBP, uFirstCLUTTBP, &uCLUTTBP );

        // Round new TBP to next page address. There are 32 blocks (256
        // bytes) to a page (8Kb).
        // TODO: Is this actually necessary?
        uTBP = (uTBP + 31) & ~31;

        --uNumTextureIDs;

        // Let texture management deal with opening PATH3, etc. For the last
        // texture in the list, TEXFLUSH as well.
        Asura_PS2_TextureManagement::SetupTextureEnvironment( bUploaded, (uNumTextureIDs ? 0 : ASURA_PS2_TEXREGFLAG_TEXFLUSH) );
    }
}

void Asura_PS2_GSMRenderer::SetObjectToWorldMat( const Asura_Matrix_3x3& xOri, const Asura_Vector_3& xPos )
{
    Asura_PS2_VU1::SetupPS2MatFromAsuraData( &g_pxO2W->f32s[0], xOri, xPos );
    g_bO2WIsIdentity = false;
    g_uPendingFlags |= PENDING_UPLOAD_CAM;
}

void Asura_PS2_GSMRenderer::SetObjectToWorldMat( const float* pfO2W )
{
    Asura_PS2_VU1::CopyPS2Mat( g_pxO2W, pfO2W );
    g_bO2WIsIdentity = false;
    g_uPendingFlags |= PENDING_UPLOAD_CAM;
}

void Asura_PS2_GSMRenderer::SetObjectToWorldMatToIdentity( void )
{
    if( !g_bO2WIsIdentity )
    {
        g_bO2WIsIdentity = true;
        g_uPendingFlags |= PENDING_UPLOAD_CAM;
    }
}

void Asura_PS2_GSMRenderer::EnableShadow( const Asura_Vector_3& xBSCentre, float fBSRadius/***RCD:not needed now**, const Asura_Vector_3& xShadowPos, float fMeshHeightAboveShadow */)
{
    const u_int uOldMeshFlags = g_uMeshFlags;
    if( !(uOldMeshFlags & ASURA_PS2_GSM__FLAG__SHADOW) || Asura_PS2_ShadowBufferSystem::s_bBufferHasChanged)
    {
        float fMinDistFromCam;
        
        float fShadowAlpha = Asura_PS2_Shadows::CalcAlphaForMinDistFromCam( xBSCentre, fBSRadius, g_uCamVisFlags, &fMinDistFromCam );
        if( fShadowAlpha <= 0.0f ) { return; }

//RCD: removed because NEW shadow system will project down and reduce alpha as necessary
//        Asura_PS2_Shadows::Proj_ModifyShadowAlphaForHeight( fMeshHeightAboveShadow, &fShadowAlpha );
//        if( fShadowAlpha <= 0.0f ) { return; }

#if 0 // not needed here for new shadow system
        g_pxShadowPosAndSize->f32s[0] = xShadowPos.x;
        g_pxShadowPosAndSize->f32s[1] = xShadowPos.y + dbg_fShadowYOffset;
        g_pxShadowPosAndSize->f32s[2] = xShadowPos.z;
        g_pxShadowPosAndSize->f32s[3] = fBSRadius;
#endif

//        Asura_PS2_Shadows::Proj_SetBfrParamsForBS( fBSRadius, fMinDistFromCam );

        {
            // Fix position of shadow-casting light a small distance
            // above the top of the BS.
            Asura_Vector_3 xLightPos = xBSCentre;
            xLightPos.y -= fBSRadius;
            xLightPos.y += ASURA_PS2_SHADOW_CASTING_LIGHT_OFFSET_Y;

            Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer->Proj_GetShadowProjMat( NULL, &g_pxShadowMatBfr->f32s[0] );
        }

        // Store alpha as a float temporarily. We need to scale it by
        // light level before sending to VU but we won't know that value
        // until we calculate the lighting later.
//rcd not needed        g_fShadowAlpha = fShadowAlpha;
        g_fShadowLightLevel = 1.0f;

        g_uMeshFlags = uOldMeshFlags | ASURA_PS2_GSM__FLAG__SHADOW;
        g_uPendingFlags |= PENDING_UPLOAD_SHADOW_DATA;
    }
}

bool Asura_PS2_GSMRenderer::IsMaterialReflective( u_int uMaterialFlags )
{
    return( (uMaterialFlags & AsuraMaterialFlag_Reflective) && Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_REFLECTION) );
}

// setting up render states for projecting shadows onto environment
void Asura_PS2_GSMRenderer::SetShadowStates( u_int uNumShadows, const u_long* puTEX0Settings, const int* piShadowBufferIDs, const u_short* pusVisClipData )
{
	// Pass 0: Cd * (1 - As)
	(g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
		SCE_GS_SET_ALPHA(SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0);

	Asura_Assert( uNumShadows < 6, "" );

	for( u_int u = 0; u < uNumShadows; ++u )
	{
		(g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_TEX0_SHADOW0 + u)->u64s[0] = puTEX0Settings[u];
		(g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_TEX0_SHADOW0 + u)->u32s[2] = pusVisClipData[u];
		g_piShadowBufIDs[u] = piShadowBufferIDs[u];
	}

	g_uPendingFlags |= PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_SHADOW_PROJECTION;	// ensure changes to renderstate get sent !

	// must also find somewhere to put the number of shadows to render, so the VU1 knows !
	// I've used this location for now but if we need an extra item per shadow buffer then I'll have to find somewhere better
	(g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_TEX0_SHADOW0)->u32s[3] = uNumShadows;
	g_uShadowBuffers = uNumShadows; // keep a copy for easy use elsewhere !

	// set projected texture mesh mode
	g_uMeshFlags |= ASURA_PS2_GSM__FLAG__PROJECTED_TEXTURE | ASURA_PS2_GSM__FLAG__PRELIT;
	g_uMeshFlags &= ~( ASURA_PS2_GSM__FLAG__REFLECTIVE | 
		               ASURA_PS2_GSM__FLAG__SELFILLUMINATED | 
					   ASURA_PS2_GSM__FLAG__DUAL_TEXTURE |
					   ASURA_PS2_GSM__FLAG__PASS_0_ST_SCALE_AND_TRANS |
					   ASURA_PS2_GSM__FLAG__PASS_2_ST_SCALE_AND_TRANS |
					   ASURA_PS2_GSM__FLAG__Y_SCALE_AND_TRANS |
					   ASURA_PS2_GSM__FLAG__POINT_LIT |
					   ASURA_PS2_GSM__FLAG__SHADOW  );

	// RCD : this was an attempt to fix a bug which was then fixed by the line above (removing reflective flag)
	// as a result, I can only assume it isn't needed, but just in case I'll leave it here commented out for now
	// force no alpha write : could save problems with reflection layer
//	(g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEST)->u64s[0] = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_RGB_ONLY, false, 0, true, SCE_GS_ZGEQUAL );
}

void Asura_PS2_GSMRenderer::GetShadowBufferPosition(u_int u_ShadowIndex)
{
	Asura_PS2_ShadowBufferSystem::GetShadowBufferPosition(g_pxShadowPosAndSize, u_ShadowIndex);
    g_pxShadowPosAndSize->f32s[1] += dbg_fShadowYOffset;
}

void Asura_PS2_GSMRenderer::UploadExtraShadowData( void )
{
    Asura_PS2_QW* pxOut = Asura_PS2_DeferredRendering::GetVifBuffer( 2 );

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 1,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         SCE_VIF1_SET_STCYCL(1, 1, false),
                                                         SCE_VIF1_SET_UNPACK(ASURA_PS2_GSM__ADDR__CONST_SHADOW_DATA, 1, Asura_PS2_VU1::VIFPK_V4_32, false) );
	
	// note : we are over-writing the area which previously was used for shadow RGBA, but is now used
	// for fall-off data
	Asura_PS2_Shadows::Proj_GetShadowFalloffControls( pxOut );
}

void Asura_PS2_GSMRenderer::SetProjectionMatrix( u_int uID, const sceVu0FMATRIX *pxData )
{
    Asura_PS2_QW* pxOut = g_pxShadowMatList + (4 * uID);
	Asura_PS2_VU1::CopyPS2Mat( pxOut->f32s, pxData );
}

void Asura_PS2_GSMRenderer::SetTextureClamp( bool bClamp )
{
    // We don't currently 'properly' support texture clamping in the GSM
    // renderer in the sense of doing it particuarly efficiently. This
    // implementation synchronises with VU1 before updating the GS CLAMP
    // register. Therefore it's a pretty bad idea to use this between strips
    // or something, since all parallelism would be destroyed. See the skybox
    // code for an example of where it's suitable for use, where the FLUSHing
    // can be tolerated.
    //
    // The clamping state is not generally reset most of the time and this
    // should be dealt with externally.

    Asura_PS2_QW* pxBfr = Asura_PS2_DeferredRendering::GetVifBuffer( 3 );

    Asura_PS2_DmaTag* pxDMATag = reinterpret_cast<Asura_PS2_DmaTag*>( pxBfr + 0 );
    pxDMATag->Set( 2,
                   Asura_PS2_DmaTag::PCE_NOCHANGE,
                   Asura_PS2_DmaTag::ID_CNT,
                   false,
                   0,
                   false,
                   SCE_VIF1_SET_FLUSH(false),
                   SCE_VIF1_SET_DIRECT(2, false) );

    Asura_PS2_GifTag* pxGIFTag = reinterpret_cast<Asura_PS2_GifTag*>( pxBfr + 1 );
    pxGIFTag->SetExceptRegs( 1, true, false, 0, SCE_GIF_PACKED, 1 );
    pxGIFTag->uAllRegs = SCE_GIF_PACKED_AD;

    Asura_PS2_QW* pxCLAMP = pxBfr + 2;
    pxCLAMP->u64s[0] = SCE_GS_SET_CLAMP( bClamp, bClamp, 0, 0, 0, 0 );;
    pxCLAMP->u64s[1] = SCE_GS_CLAMP_1;
}

void Asura_PS2_GSMRenderer::SetRenderstate( u_int uMaterialFlags )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewMeshFlags    = uOldMeshFlags;
    u_int uNewPendingFlags = g_uPendingFlags;

    // Mask off material flags that we don't care about.
    uMaterialFlags &= (   AsuraMaterialFlag_SelfIlluminate
                        | AsuraMaterialFlag_PS2AdditiveAlpha
                        | AsuraMaterialFlag_3dDecal
                        | AsuraMaterialFlag_Transparent
                        | AsuraMaterialFlag_Reflective );

    // Mask off mesh flags for RS set up externally and after this method.
    uNewMeshFlags &= ~ASURA_PS2_GSM__MASK__ANY_VTX_MANIPULATIONS;

    if( !dbg_bGSMOptimiseOutRedundantRSUploads || uMaterialFlags != g_uCurMaterialFlags )
    {
        // Mask off the mesh flags that we may or may not turn on below.
        uNewMeshFlags &= ~(ASURA_PS2_GSM__FLAG__SELFILLUMINATED | ASURA_PS2_GSM__FLAG__REFLECTIVE);

        u_long uTEST = SCE_GS_SET_TEST( false, 0, 0, 0, false, 0, true, SCE_GS_ZGEQUAL );

        if( uMaterialFlags & AsuraMaterialFlag_SelfIlluminate )
        {
            uNewMeshFlags |= ASURA_PS2_GSM__FLAG__SELFILLUMINATED;

            // Pass 0: Cs
            (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
                SCE_GS_SET_ALPHA(SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CS, 0);
            // Pass 1: Cs * As + Cd
            (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_SELF_ILLUM_ALPHA)->u64s[0] =
                SCE_GS_SET_ALPHA(SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0);
        }
        else if( uMaterialFlags & AsuraMaterialFlag_3dDecal )
        {
            // Pass 0: Cs * As + Cd * (1 - As)
            (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
                SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
            // (AW - 26.08.2004) - Change in behaviour, don't write to the frame buffer at all. Just the Z buffer.
            uTEST = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_ZB_ONLY, false, 0, true, SCE_GS_ZGEQUAL );
        }
        else
        {
            if( uMaterialFlags & AsuraMaterialFlag_PS2AdditiveAlpha )
            {
                // Pass 0: Cs * As + Cd
                (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
                    SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
                // Disable z writes.
                uTEST = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_RGB_ONLY, false, 0, true, SCE_GS_ZGEQUAL );
            }
            else if( uMaterialFlags & AsuraMaterialFlag_Transparent )
            {
                // Pass 0: Cs * As + Cd * (1 - As)
                (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
                    SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
                // For stuff that's more or less opaque, write to both the
                // frame and z buffers. For everything else, only update the
                // frame buffer.
                uTEST = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_GEQUAL, 124, SCE_GS_AFAIL_FB_ONLY, false, 0, true, SCE_GS_ZGEQUAL );
            }
            else
            {
                // Pass 0: Cs
                (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0] =
                    SCE_GS_SET_ALPHA( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CS, 0 );
            }
        }

        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEST)->u64s[0] = uTEST;

        if( IsMaterialReflective(uMaterialFlags) )
        {
            uNewMeshFlags |= ASURA_PS2_GSM__FLAG__REFLECTIVE;

            // If reflective flag wasn't previously on, setup tex matrix now.
            if( !(uOldMeshFlags & ASURA_PS2_GSM__FLAG__REFLECTIVE) )
            {
                sceVu0UnitMatrix( &g_pxTexMatBfr->Vu0FVec );
                (g_pxTexMatBfr + 0)->f32s[0] = 0.5f;
                (g_pxTexMatBfr + 1)->f32s[1] = 0.5f;
                (g_pxTexMatBfr + 3)->f32s[0] = 0.5f;
                (g_pxTexMatBfr + 3)->f32s[1] = 0.5f;

                uNewPendingFlags |= PENDING_UPLOAD_CAM;
            }

            // Pass 2: Cd += Cs * Ad
            (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_ALPHA)->u64s[0] =
                SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AD, SCE_GS_ALPHA_CD, 0 );
        }

        uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
        g_uCurMaterialFlags = uMaterialFlags;
    }

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_GSMRenderer::SetupPass0STManipulations( bool bEnable, float fScaleS, float fScaleT, float fTransS, float fTransT )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewPendingFlags = g_uPendingFlags;
    u_int uNewMeshFlags    = uOldMeshFlags;

    if( bEnable )
    {
        // Check for incompatible / mutually exclusive settings.
        Asura_Assert( !(uOldMeshFlags & ASURA_PS2_GSM__FLAG__PASS_2_ST_SCALE_AND_TRANS), "" );
        
        FillSTScaleAndTranslationQW( fScaleS, fScaleT, fTransS, fTransT );
        
        uNewPendingFlags |= PENDING_UPLOAD_ST_OV_PARAMS;
        uNewMeshFlags |= ASURA_PS2_GSM__FLAG__PASS_0_ST_SCALE_AND_TRANS;
    }
    else
    {
        uNewMeshFlags &= ~ASURA_PS2_GSM__FLAG__PASS_0_ST_SCALE_AND_TRANS;
    }

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_GSMRenderer::SetupPass2STManipulations( bool bEnable,
                                                       float fScaleS, float fScaleT, float fTransS, float fTransT,
                                                       u_int uAlphaA, u_int uAlphaB, u_int uAlphaC, u_int uAlphaD, u_int uAlphaFIX )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewPendingFlags = g_uPendingFlags;
    u_int uNewMeshFlags    = uOldMeshFlags;

    if( bEnable )
    {
        // Check for incompatible / mutually exclusive settings.
        Asura_Assert( !(uOldMeshFlags & (ASURA_PS2_GSM__FLAG__PASS_0_ST_SCALE_AND_TRANS | ASURA_PS2_GSM__FLAG__REFLECTIVE)), "" );

        FillSTScaleAndTranslationQW( fScaleS, fScaleT, fTransS, fTransT );

        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_ALPHA)->u64s[0] =
            SCE_GS_SET_ALPHA( uAlphaA, uAlphaB, uAlphaC, uAlphaD, uAlphaFIX );

        uNewPendingFlags |= (PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_ST_OV_PARAMS);
        uNewMeshFlags |= ASURA_PS2_GSM__FLAG__PASS_2_ST_SCALE_AND_TRANS;
    }
    else
    {
        uNewMeshFlags &= ~ASURA_PS2_GSM__FLAG__PASS_2_ST_SCALE_AND_TRANS;
    }

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_GSMRenderer::SetupYManipulations( bool bEnable,
                                                 float fA0, float fB0, float fA1, float fB1,
                                                 float fC0, float fC1, float fC2, float fC3 )
{
    const u_int uOldMeshFlags = g_uMeshFlags;

    u_int uNewPendingFlags = g_uPendingFlags;
    u_int uNewMeshFlags    = uOldMeshFlags;

    if( bEnable )
    {
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_ANGLE)->f32s[0] = fA0;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_ANGLE)->f32s[1] = fB0;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_ANGLE)->f32s[2] = fA1;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_ANGLE)->f32s[3] = fB1;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_Y    )->f32s[0] = fC2;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_Y    )->f32s[1] = fC0;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_Y    )->f32s[2] = fC1;
        (g_pxYParams + ASURA_PS2_GSM__OFFSET__Y_PARAMS_Y    )->f32s[3] = fC3;

        uNewPendingFlags |= PENDING_UPLOAD_Y_PARAMS;
        uNewMeshFlags |= ASURA_PS2_GSM__FLAG__Y_SCALE_AND_TRANS;
    }
    else
    {
        uNewMeshFlags &= ~ASURA_PS2_GSM__FLAG__Y_SCALE_AND_TRANS;
    }

    g_uMeshFlags    = uNewMeshFlags;
    g_uPendingFlags = uNewPendingFlags;
}

DEBUGVAR bool dbg_bGSMDisableDualTexturingWithBlend   = false;
DEBUGVAR bool dbg_bGSMDisableMipmappingOnSecondaryTex = false;

void Asura_PS2_GSMRenderer::SelectFromDualTexturePalette( int iTexID )
{
    const u_int uDualTexturePaletteSize = g_uDualTexCount;
    
    // Don't regard an empty palette as an error; we use this to disable
    // dual texturing when the relevant rendering option is off.
    if( !uDualTexturePaletteSize ) { return; }

    // Find relevant set of cached GS register settings.
    const Asura_PS2_QW* pxRegs = g_pxDualTexRegs;
    for( u_int u = 0; u < uDualTexturePaletteSize; ++u )
    {
        // Texture ID is tucked away in upper word of first QW of each
        // register triplet.
        if( (pxRegs + 0)->i32s[3] == iTexID ) { break; }

        pxRegs += 3;
    }

    // Assume the caller will only request textures that are actually in the
    // palette.
    Asura_Assert( pxRegs < g_pxDualTexRegs + 3 * uDualTexturePaletteSize, "" );

    u_int uNewMeshFlags = g_uMeshFlags;

// (MP - 05.11.2004) Tools has changed to disallow exports containing
// materials with such combinations.
//  // Verify that no other pass 2 renderstate is on.
//  if( uNewMeshFlags & ASURA_PS2_GSM__MASK__PASS_2 )
//  {
//      Asura_Assert( false, "" );
//      
//      uNewMeshFlags &= ~ASURA_PS2_GSM__MASK__PASS_2;
//
//      // Better mask off related material flags from our local record as
//      // well, since otherwise SetRenderstate() will get confused.
//      g_uCurMaterialFlags &= ~AsuraMaterialFlag_Reflective;
//  }

    // Copy GS reg QWs over.
    {
        const u_long128 uTEX1    = (pxRegs + 0)->u128;
        const u_long128 uMIPTBP1 = (pxRegs + 1)->u128;
        const u_long128 uTEX0    = (pxRegs + 2)->u128;

        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX1   )->u128 = uTEX1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_MIPTBP1)->u128 = uMIPTBP1;
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX0   )->u128 = uTEX0;

        if( dbg_bGSMDisableMipmappingOnSecondaryTex )
        {
            (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX1)->u64s[0] &= ~GS_TEX1_MXL_M;
        }
    }

    // Setup ALPHA.
    // Pass 2: Cd * As + Cs * (1 - As)
    (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_ALPHA)->u64s[0] =
        SCE_GS_SET_ALPHA( SCE_GS_ALPHA_CD, SCE_GS_ALPHA_CS, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CS, 0 );

    if( dbg_bGSMDisableDualTexturingWithBlend )
    {
        // Pass 2: Cd
        (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_ALPHA)->u64s[0] =
            SCE_GS_SET_ALPHA( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CD, 0 );
    }

    uNewMeshFlags |= ASURA_PS2_GSM__FLAG__DUAL_TEXTURE;

    g_uMeshFlags = uNewMeshFlags;
    g_uPendingFlags |= PENDING_UPLOAD_RENDERSTATE;

    // Since we're sharing the RS buffer entries used by other pass 2 stuff
    // we need to trash the corresponding texture ID record.
    g_iCurSTOvTex = ASURA_TEXTURE_NONE;
}

void Asura_PS2_GSMRenderer::ManuallyModifyRenderstate( RSCallback pfnCB )
{
    pfnCB( g_pxRSBfr );

    // Ensure that we resend renderstate.
    g_uPendingFlags |= PENDING_UPLOAD_RENDERSTATE;

    // Flag that next time we change renderstate we need to update the
    // buffer, etc. We have to assume that the callback has trashed all of
    // our cached state, including texture settings.
    g_uCurMaterialFlags |= uINVALID_MATERIAL_FLAG;
    g_iCurBaseTex = ASURA_TEXTURE_NONE;
    g_iCurSTOvTex = ASURA_TEXTURE_NONE;
}

void Asura_PS2_GSMRenderer::CB_DisableZReadsAndWrites( Asura_PS2_QW* pxRSBfr )
{
    (g_pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEST)->u64s[0] =
        SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_FB_ONLY,
                         false, 0,
                         true, SCE_GS_ZALWAYS );
}

void Asura_PS2_GSMRenderer::CB_JustWriteZ( Asura_PS2_QW* pxRSBfr )
{
    (pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEST)->u64s[0] =
        SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_ZB_ONLY,
                         false, 0,
                         true, SCE_GS_ZGEQUAL );
}

void Asura_PS2_GSMRenderer::CB_IgnoreBaseTextureAlpha( Asura_PS2_QW* pxRSBfr )
{
    (pxRSBfr + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX0)->u64s[0] &= ~GS_TEX0_TCC_M;
}

void Asura_PS2_GSMRenderer::SetTexture( int iTexID, int iMMAG, int iMMIN, u_int uBfr, bool bIsBaseTexture )
{
    const u_int uCurPendingFlags = g_uPendingFlags;
    u_int uNewPendingFlags = uCurPendingFlags;

    bool bUploaded = false;

    {
        bool bUpload = true;

        // If there's another texture sync pending, say from another pass's
        // texture upload, then sync with that first.
        if( uCurPendingFlags & PENDING_UPLOAD_TEX_SYNC )
        {
            // We want to avoid doing this sync if possible since it means a
            // FLUSH much earlier, probably forcing a stall while the
            // previous strip finishes rendering. A safe case for doing this
            // is if this next texture we're going to upload is going to some
            // fixed buffer and is already resident there. Note that if we
            // skip the sync, we have to skip the upload below as well (since
            // otherwise internal texture management settings may not be
            // right when we come to do the sync later.) This means that it's
            // not safe to skip the sync if auto buffer switching since we
            // can't skip the logic in UploadTexture() that remembers which
            // buffer we used last (and therefore doesn't go overwriting the
            // texture in it while we're still using it.)
            // (MP - 09.01.2004) TODO: This restriction may become a
            // significant problem and bottleneck if/when we support more
            // general multipassing.
            if( Asura_PS2_TextureManagement::BFR_AUTO != uBfr )
            {
                if( Asura_PS2_TextureManagement::BFR_NONE != Asura_PS2_TextureManagement::HasTextureBeenUploaded(iTexID, uBfr) )
                {
                    bUpload = false;
                }
            }

            if( bUpload )
            {
                Asura_PS2_TextureManagement::SetupTextureEnvironment( g_bTexUploaded, ASURA_PS2_TEXREGFLAG_TEXFLUSH );
            }
        }

        if( bUpload )
        {
            bUploaded = Asura_PS2_TextureManagement::UploadTexture( iTexID, uBfr, 0 );

            g_bTexUploaded = bUploaded;
            uNewPendingFlags |= PENDING_UPLOAD_TEX_SYNC;
        }
    }

    // Refresh tex setup stuff if necessary. Note that a check on whether
    // or not the setup has already been done for the specified texture ID
    // breaks in two cases: (i) if they're requesting the same texture but
    // with different filtering settings; (ii) if texture caching isn't
    // working, so that we've uploaded the texture again despite it being
    // resident. I don't think (i) is likely to happen, mainly because we
    // use bilinear everywhere on meshes, and we can workaround (ii) by
    // forcing a refresh if we found the texture needed to be uploaded above.
    int& iCurTex = ( bIsBaseTexture ? g_iCurBaseTex : g_iCurSTOvTex );
    if( !dbg_bGSMOptimiseOutRedundantTexRegUploads || bUploaded || iTexID != iCurTex )
    {
        const u_int uMIPTBP1Offset = ( bIsBaseTexture ? ASURA_PS2_GSM__OFFSET__RS_IN_BASE_MIPTBP1 : ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_MIPTBP1 );
        Asura_PS2_TextureManagement::GetMIPTBP1Setting( &(g_pxRSBfr + uMIPTBP1Offset)->u64s[0], iTexID, uBfr );

        const u_int uTEX1Offset = ( bIsBaseTexture ? ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX1 : ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX1 );
        Asura_PS2_TextureManagement::GetTEX1Setting( &(g_pxRSBfr + uTEX1Offset)->u64s[0], iTexID, iMMAG, iMMIN );

        const u_int uTEX0Offset = ( bIsBaseTexture ? ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX0 : ASURA_PS2_GSM__OFFSET__RS_IN_ST_OV_TEX0 );
        Asura_PS2_TextureManagement::GetTEX0Setting( &(g_pxRSBfr + uTEX0Offset)->u64s[0], iTexID, uBfr );

        iCurTex = iTexID;
        uNewPendingFlags |= PENDING_UPLOAD_RENDERSTATE;
    }

    g_uPendingFlags = uNewPendingFlags;
}

void Asura_PS2_GSMRenderer::SetCamFlags( u_int uVisFlags, const Asura_Bounding_Box& xBB )
{
    SetCamFlags( uVisFlags, xBB.CalculateCentre(), 0.5f * xBB.CalculateExtents().Magnitude() );
}

void Asura_PS2_GSMRenderer::SetCamFlags( u_int uVisFlags, const Asura_Vector_3& xCentre, const float fRadius )
{
    u_int uClipFlags = 0;
    
    // See notes in Asura_PS2_GSMDefines.h for what's going on here.
    u_int uClipBit = 1 << 6;

    const u_int uNumCams = Asura_Camera::GetNumberOfCameras();
    for( u_int uCam = 0; uCam < uNumCams; ++uCam )
    {
        // Don't bother checking if not visible to this camera.
        if( uVisFlags & (1 << uCam) )
        {
            Asura_Camera::SetCameraIndex( uCam );

            // Turn on relevant clip flag if not entirely within this cam's
            // guardband frustum.
            if( !Asura_PS2_VU1::IsSphereEntirelyWithinGuardbandFrustum(xCentre, fRadius) )
            {
                uClipFlags |= uClipBit;
            }
        }

        uClipBit >>= 2;
    }

    SetCamFlags( uVisFlags, uClipFlags );
}

void Asura_PS2_GSMRenderer::SetCamFlags( u_int uVisFlags, u_int uClipFlags )
{
    // Callers should deal with meshes that aren't visible to any cameras.
    Asura_Assert( uVisFlags, "" );
    
    u_int uNumCamsThatCanSee = 0;
    u_int uVisFlagsForVU     = 0;

    // See notes in Asura_PS2_GSMDefines.h for what's going on here.
    {
        u_int uEEVisBit = 1 << 0;
        u_int uVUVisBit = 1 << 7;

        for( u_int u = Asura_Camera::GetNumberOfCameras(); u; --u )
        {
            if( uVisFlags & uEEVisBit )
            {
                uVisFlagsForVU |= uVUVisBit;
                ++uNumCamsThatCanSee;
            }

            uEEVisBit <<= 1;
            uVUVisBit >>= 2;
        }
    }

    // Ensure that no clip flags are set where the corresponding vis flag
    // isn't. This breaks some code that assumes the flags will be zero once
    // there are no more cameras to render for.
    Asura_Assert( !(uClipFlags & ~(uVisFlagsForVU >> 1)), "" );

    u_int uNewFlagsForVU = uVisFlagsForVU | uClipFlags;

    // If more than one camera can see the mesh, let the program know.
    if( uNumCamsThatCanSee > 1 )
    {
        uNewFlagsForVU |= ASURA_PS2_GSM__FLAG__VIS_FOR_MORE_THAN_ONE_CAM;
    }

    if( uNewFlagsForVU != g_uCamFlagsForVU )
    {
        g_uCamVisFlags = uVisFlags;
        g_uCamFlagsForVU = uNewFlagsForVU;
        g_uNumCamsVisFor = uNumCamsThatCanSee;
        g_uPendingFlags |= PENDING_UPLOAD_CAM_FLAGS;
    }
}

void Asura_PS2_GSMRenderer::End( void )
{
#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
    // Render pending strip, if there is one.
    if( g_uPendingVtxCount )
    {
        Asura_PS2_DmaTag* pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
        pxCNTTag->Set( 0,
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        RenderPendingStrip( pxCNTTag );
    }
#endif

    // Deal with case of having to render mesh with no tris. Unfortunately
    // there are cases where this occurs, eg. with invisible objects being
    // used for collisions only.
    if( g_uPendingFlags )
    {
        GSMSendPendingData( false );
    }

#if 0	// no longer required because now the shadow system renders all the outstanding shadows when rendering is all complete
    // Render shadow if necessary.
    if( g_uMeshFlags & ASURA_PS2_GSM__FLAG__SHADOW )
    {
        RenderShadowQuad();
    }
#endif

    // FLUSH to prevent subsequent rendering trashing stuff in VU memory that
    // we're still in the middle of using. This gets optimised out wherever
    // possible; see comment below.
    Asura_PS2_QW* pxQWForTag = Asura_PS2_DeferredRendering::GetVifBuffer( 1 );
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxQWForTag))->Set( 0,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_CNT,
                                                            false,
                                                            0,
                                                            false,
                                                            SCE_VIF1_SET_NOP(false),
                                                            SCE_VIF1_SET_FLUSH(false) );

    // Record ptr and timestamp so we can remove the FLUSH again if it turns
    // out we're doing more GSM stuff straight after. Note that the code to
    // do this in Begin() makes some pretty strong assumptions about exactly
    // what we do with the tag and the FLUSH above.
    g_pxVifPtrFromLastEnd = ( dbg_bGSMOptimiseOutRedundantFLUSHes ? pxQWForTag + 1 : NULL );
    g_uFrameOfLastEnd     = Asura_Timers::GetFrameNumber();

    // Return region of SPR that we reserved in Begin().
    Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(0x70000000) );
}

void Asura_PS2_GSMRenderer::Initialise( void )
{
    Reset();
}

#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
void Asura_PS2_GSMRenderer::RenderStrip( u_int uREFAddr, u_int uNumQWsForREF, u_int uNumVertices )
{
    Asura_PS2_DmaTag* pxTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
    pxTag->Set( uNumQWsForREF,
                Asura_PS2_DmaTag::PCE_NOCHANGE,
                Asura_PS2_DmaTag::ID_REF,
                false,
                uREFAddr,
                false );

    if( g_uPendingVtxCount )
    {
        RenderPendingStrip( pxTag );
    }
    else
    {
        pxTag->SetSpareWordsToNOPs();
    }

    GSMSendPendingData( false );

    g_uPendingVtxCount = uNumVertices;
}
#endif

void Asura_PS2_GSMRenderer::RenderStripChain( u_int uCALLAddr )
{
#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
    // Render pending strip, if necessary.
    if( g_uPendingVtxCount )
    {
        Asura_PS2_DmaTag* pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
        pxCNTTag->Set( 0,
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        RenderPendingStrip( pxCNTTag );
    }
#endif

    GSMSendPendingData( true );

    // CALL strip chain.
    {
        Asura_PS2_DmaTag* pxCALLTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
        pxCALLTag->SetWithNOPs( 0,
                                Asura_PS2_DmaTag::PCE_NOCHANGE,
                                Asura_PS2_DmaTag::ID_CALL,
                                false,
                                uCALLAddr,
                                false );
    }
}

void Asura_PS2_GSMRenderer::RenderStripsWithManualSend( void )
{
#if ASURA_PS2_GSM_SUPPORT_RENDERING_SINGLE_STRIPS
    // Render pending strip, if necessary.
    if( g_uPendingVtxCount )
    {
        Asura_PS2_DmaTag* pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );
        pxCNTTag->Set( 0,
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        RenderPendingStrip( pxCNTTag );
    }
#endif

    GSMSendPendingData( true );
}

void Asura_PS2_GSMRenderer::Reset( void )
{
    g_pxVifPtrFromLastEnd = NULL;
}

void Asura_PS2_GSMRenderer::Shutdown( void )
{
    Reset();
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static Asura_PS2_QW* CalculateAndWriteCamData( Asura_PS2_QW* pxOut, u_int uCamIdx, bool bO2WIsIdentity, bool bReflective )
{
    // O->GBC matrix.
    {
        sceVu0FMATRIX& xW2GBC = Asura_PS2_VU1::GetWorldToClipMatGuardband( uCamIdx );
        if( bO2WIsIdentity )
        {
            Asura_PS2_VU1::CopyPS2Mat( pxOut, xW2GBC );
        }
        else
        {
            sceVu0MulMatrix( &pxOut->Vu0FVec, xW2GBC, &g_pxO2W->Vu0FVec );
        }
        pxOut += 4;
    }

    // O->V and texture matrices.
    if( bReflective )
    {
        // Setup O->V matrix.
        sceVu0FMATRIX& xW2V = Asura_PS2_VU1::GetWorldToViewMat( uCamIdx );
        if( bO2WIsIdentity )
        {
            Asura_PS2_VU1::CopyPS2Mat( g_pxMatTempBfr, xW2V );
        }
        else
        {
            sceVu0MulMatrix( &g_pxMatTempBfr->Vu0FVec, xW2V, &g_pxO2W->Vu0FVec );
        }
        
        // Copy to output.
        Asura_PS2_VU1::CopyPS2Mat( pxOut, g_pxMatTempBfr );
        pxOut += 4;
    
        // (MP - 06.11.2003) TODO: Is the following code correct?
        // Previously we were concatenating tex mat and v->w here
        // but I'm pretty sure that it should be tex mat and v->o.
        
        // Remove translation component from O->V matrix and transpose.
        Asura_PS2_VU1::CalcPS2MatInverseRotOnly( g_pxMatTempBfr, g_pxMatTempBfr );

        // Concatenate with texture matrix and write to output.
        sceVu0MulMatrix( &pxOut->Vu0FVec, &g_pxTexMatBfr->Vu0FVec, &g_pxMatTempBfr->Vu0FVec );
        pxOut += 4;
    }

    return pxOut;
}


void Asura_PS2_GSMRenderer::RenderShadowQuad( u_int uFlags  )
{
    const bool bNeedNewCamData = true; //!g_bO2WIsIdentity;

    // First UNPACK the renderstate and possibly also camera data, ready for
    // rendering the quad.
    {
        // We'll need to send new renderstate since we're rendering with a
        // different texture, alpha blend, etc. Note that we can jam the 
        // UNPACK in the DMA tag so we don't need an extra QW for that.
        const u_int uRenderstateQWs = ASURA_PS2_GSM__SIZE__RENDERSTATE_BFR;

        u_int uCamSetupQWs = 0;
        if( bNeedNewCamData )
        {
            // If we're rendering a shadow with a non-identity O->W transform, we
            // need to replace the existing O->GBC camera matrices with plain old
            // W->GBC ones.

            const u_int uCamVisFlags = g_uCamVisFlags;
            const u_int uNumCams     = Asura_Camera::GetNumberOfCameras();

            u_int uNumCamsThatCanSee = 0;
            for( u_int uCam = 0; uCam < uNumCams; ++uCam )
            {
                if( uCamVisFlags & (1 << uCam) ) { ++uNumCamsThatCanSee; }
            }
            Asura_Assert( uNumCamsThatCanSee, "" );

            uCamSetupQWs = 5 * uNumCamsThatCanSee;
        }

		const u_int uShadowDataQWC = 2;
        const u_int uQWC      = uRenderstateQWs + uCamSetupQWs + uShadowDataQWC + 1;	// 2 more for sending quad shadow position and colour
        const u_int uTotalQWs = 1 + uQWC;

        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        // DMA tag and renderstate UNPACK.
        {
            const u_int uBfr  = g_uCurVURSBfr ^ 1;
            const u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__RENDERSTATE_BFR_1 : ASURA_PS2_GSM__ADDR__RENDERSTATE_BFR_0 );

            (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWC,
                                                                 Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                 Asura_PS2_DmaTag::ID_CNT,
                                                                 false,
                                                                 0,
                                                                 false,
                                                                 SCE_VIF1_SET_STCYCL(1, 1, false),
                                                                 SCE_VIF1_SET_UNPACK(uAddr, uRenderstateQWs, Asura_PS2_VU1::VIFPK_V4_32, false) );

            // Copy previous state.
            const u_long128* puSrc = &g_pxRSBfr->u128;
            u_long128* puDst = &pxOut->u128;
            for( u_int u = ASURA_PS2_GSM__SIZE__RENDERSTATE_BFR; u--; )
            {
                *puDst++ = *puSrc++;
            }
        
            // Fix up the bits we care about.
            {
                // Alpha blend: Cd *= (1 - As).
                (pxOut + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_ALPHA)->u64s[0]
                    = SCE_GS_SET_ALPHA( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );

                // Texture settings.
				Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer->Proj_GetTEX1andTEX0( &(pxOut + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX1)->u64s[0],
                                                        &(pxOut + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEX0)->u64s[0] );

                // TEST - disable z write.
                (pxOut + ASURA_PS2_GSM__OFFSET__RS_IN_BASE_TEST)->u64s[0]
                    = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_RGB_ONLY, false, 0, true, SCE_GS_ZGEQUAL );
            }

            pxOut += ASURA_PS2_GSM__SIZE__RENDERSTATE_BFR;

	        g_uCurVURSBfr = uBfr;
        }

		{
			const u_int uBfr  = g_uCurVUShadowBfr^1;	// switch shadow buffer
			const u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__SHADOW_DATA_BFR_1 : ASURA_PS2_GSM__ADDR__SHADOW_DATA_BFR_0 );

			// Shadow pos and size.
			FillQWWithNOPs( pxOut);
            pxOut->u32s[1] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_SWITCH_SHADOW_DATA_BFR);
			pxOut->u32s[2] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_NOP );	// to balance out the uneven number of calls

			(pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK(uAddr + ASURA_PS2_GSM__OFFSET__SHADOW_QUAD_PARAMS_0 , uShadowDataQWC, Asura_PS2_VU1::VIFPK_V4_32, false );
			(pxOut++)->u128 = g_pxShadowPosAndSize->u128;
			g_fShadowLightLevel = Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer->GetLightLevel();
			pxOut->u32s[3] = Asura_Maths::FloatToInt( Asura_Maths::Min(255.0f, g_fShadowLightLevel * 128.0f ) );
			++pxOut;

			g_uCurVUShadowBfr = uBfr;
		}
        // Camera data UNPACK, if necessary.
        if( bNeedNewCamData )
        {
            const u_int uCamVisFlags = g_uCamVisFlags;
            const u_int uBfr         = g_uCurVUCamBfr ^ 1;
            const u_int uNumCams     = Asura_Camera::GetNumberOfCameras();

            u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__PER_OBJ_CAM_DATA_BFR_1 : ASURA_PS2_GSM__ADDR__PER_OBJ_CAM_DATA_BFR_0 );
            for( u_int uCam = 0; uCam < uNumCams; uAddr += ASURA_PS2_GSM__OFFSET__CAM_PER_OBJ_NEXT_CAM, ++uCam )
            {
                if( !(uCamVisFlags & (1 << uCam)) ) { continue; }

                FillQWWithNOPs( pxOut );
                (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, 4, Asura_PS2_VU1::VIFPK_V4_32, false );

                const void* pvW2GBC = Asura_PS2_VU1::GetWorldToClipMatGuardband( uCam );
                Asura_PS2_VU1::CopyPS2Mat( pxOut, pvW2GBC );
                pxOut += 4;
            }

            g_uCurVUCamBfr = uBfr;
        }
    }

    // MSCALs, etc.
    {
        // Only need a QWC of 1 since we can some stuff in the DMA tag.
        const u_int uQWC      = 1;
        const u_int uTotalQWs = 1 + uQWC;

        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut))->Set( uQWC,
                                                           Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                           Asura_PS2_DmaTag::ID_CNT,
                                                           false,
                                                           0,
                                                           false );
    
        {
            pxOut->u32s[2] = SCE_VIF1_SET_ITOP( uFlags, false );
            pxOut->u32s[3] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_LOAD_CAM_FLAGS );
            ++pxOut;
        }

        // Switch renderstate buffer and maybe also for the camera. Note that
        // we want to do an even total number of MSCALs here so that the TOP
        // ptr is back at the value used for the last strip rendered. (See
        // where this fn is called from, ie. Asura_PS2_GSMRenderer::End().)
        // Therefore if we don't need to switch camera buffers, we NOP on the
        // VU instead.
        pxOut->u32s[0] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_SWITCH_RENDERSTATE_BFR );
        pxOut->u32s[1] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( bNeedNewCamData ? VU1_GSM_SWITCH_CAM_DATA_BFR : VU1_GSM_NOP );

        // Setup vertex count and render.
        pxOut->u32s[2] = SCE_VIF1_SET_ITOP( 4, false );
        pxOut->u32s[3] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_RENDER_SHADOW_QUAD );
        ++pxOut;
    }

#if 0	// now handled by shadow buffer system itself
    // Maybe display some element of the shadow buffer.
    if( dbg_bDisplayGSMShadowBufferRGB || dbg_bDisplayGSMShadowBufferAlpha )
    {
        Asura_PS2_Shadows::Proj_ShowShadowBfr( 128, 128, 256, 256, dbg_bDisplayGSMShadowBufferAlpha, true );
    }
#endif

    Asura_PS2_Render::UpdateTriStatsFromVisFlags( 2, 1, g_uCamVisFlags );

    // Some bits of state will now need fixing up on the VU so record that in
    // the pending flags.
    g_uPendingFlags |= PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS;
    if( bNeedNewCamData )
    {
        g_uPendingFlags |= PENDING_UPLOAD_CAM;
    }
}


static void GSMSendPendingData( bool bWillMSCNTForNextStrip )
{
    const u_int uMeshFlags = g_uMeshFlags;

    u_int uPendingFlags = g_uPendingFlags;

    // Turn on need-to-send-mesh-flags flag if required.
    if( uMeshFlags != g_uPrevMeshFlags )
    {
        uPendingFlags |= PENDING_UPLOAD_MESH_FLAGS;
        g_uPrevMeshFlags = uMeshFlags;
    }

    // Early out if nothing's changed since the previous strip.
    if( !uPendingFlags ) { return; }

    // Anything to UNPACK? Do this first, so we can leave it as long as
    // possible before FLUSHing.
    if( uPendingFlags & ( PENDING_UPLOAD_RENDERSTATE
                        | PENDING_UPLOAD_CAM
                        | PENDING_UPLOAD_LIGHT
                        | PENDING_UPLOAD_SHADOW_DATA
                        | PENDING_UPLOAD_SHADOW_PROJECTION
                        | PENDING_UPLOAD_ST_OV_PARAMS
                        | PENDING_UPLOAD_Y_PARAMS ) )
    {
        // Estimate how many QWs we'll take up with it all. We'll err on the
        // side of overestimation and then just hand back any QWs we don't
        // use.
        u_int uEstimatedQWC = 0;
        if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE		 ) { uEstimatedQWC += 1 + ASURA_PS2_GSM__SIZE__RENDERSTATE_BFR;							}
        if( uPendingFlags & PENDING_UPLOAD_CAM				 ) { uEstimatedQWC += 1 + ASURA_PS2_GSM__SIZE__PER_OBJ_CAM_DATA_BFR;					}
        if( uPendingFlags & PENDING_UPLOAD_LIGHT			 ) { uEstimatedQWC += 1 + ASURA_PS2_GSM__SIZE__LIGHT_DATA_BFR;							}
        if( uPendingFlags & PENDING_UPLOAD_SHADOW_DATA		 ) { uEstimatedQWC += 1 + ASURA_PS2_GSM__SIZE__SHADOW_DATA_BFR;							}
        if( uPendingFlags & PENDING_UPLOAD_SHADOW_PROJECTION ) { uEstimatedQWC += 1 + ASURA_PS2_GSM__SIZE__SHADOW_PALETTE_ENTRY * g_uShadowBuffers; }
        if( uPendingFlags & PENDING_UPLOAD_ST_OV_PARAMS		 ) { uEstimatedQWC += 2 + ASURA_PS2_GSM__SIZE__ST_SCALE_AND_TRANS_PARAMS;				}
        if( uPendingFlags & PENDING_UPLOAD_Y_PARAMS			 ) { uEstimatedQWC += 2 + ASURA_PS2_GSM__SIZE__Y_SCALE_AND_TRANS_PARAMS;				}

        const u_int uTotalQWs = 1 + uEstimatedQWC;
        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        Asura_PS2_DmaTag* const pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut );
        pxCNTTag->Set( 0, // Will be fixed up below as necessary.
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        pxCNTTag->m_uWord2 = SCE_VIF1_SET_STCYCL( 1, 1, false );

        bool bNeedNOPs = false;

        if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE )
        {
            const u_int uBfr  = g_uCurVURSBfr ^ 1;
            const u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__RENDERSTATE_BFR_1 : ASURA_PS2_GSM__ADDR__RENDERSTATE_BFR_0 );

            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
            }
            pxOut->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, ASURA_PS2_GSM__SIZE__RENDERSTATE_BFR, Asura_PS2_VU1::VIFPK_V4_32, false );
            ++pxOut;

            const u_long128* puSrc = &g_pxRSBfr->u128;
            for( u_int u = ASURA_PS2_GSM__SIZE__RENDERSTATE_BFR; u; --u )
            {
                (pxOut++)->u128 = *puSrc++;
            }

            bNeedNOPs     = true;
            g_uCurVURSBfr = uBfr;
        }

        if( uPendingFlags & PENDING_UPLOAD_CAM )
        {
            const u_int uCamVisFlags = g_uCamVisFlags;
            const u_int uBfr         = g_uCurVUCamBfr ^ 1;
            const bool  bReflective  = ( uMeshFlags & ASURA_PS2_GSM__FLAG__REFLECTIVE );
            const u_int uSize        = ( bReflective ? 12 : 4 );
            const u_int uNumCams     = Asura_Camera::GetNumberOfCameras();

            u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__PER_OBJ_CAM_DATA_BFR_1 : ASURA_PS2_GSM__ADDR__PER_OBJ_CAM_DATA_BFR_0 );

            const bool bO2WIsIdentity = g_bO2WIsIdentity;
            for( u_int uCam = 0; uCam < uNumCams; uAddr += ASURA_PS2_GSM__OFFSET__CAM_PER_OBJ_NEXT_CAM, ++uCam )
            {
//                if( !(uCamVisFlags & (1 << uCam)) ) { continue; }

                if( bNeedNOPs )
                {
                    FillQWWithNOPs( pxOut );
                }
                bNeedNOPs = true;

                (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, uSize, Asura_PS2_VU1::VIFPK_V4_32, false );

                pxOut = CalculateAndWriteCamData( pxOut, uCam, bO2WIsIdentity, bReflective );
            }

            g_uCurVUCamBfr = uBfr;
        }

        if( uPendingFlags & PENDING_UPLOAD_LIGHT )
        {
            u_int uQWs = g_uNumLightQWs;
            Asura_Assert( uQWs, "" );

            const u_int uBfr  = g_uCurVULightBfr ^ 1;
            const u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__LIGHT_DATA_BFR_1 : ASURA_PS2_GSM__ADDR__LIGHT_DATA_BFR_0 );

            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
            }
            (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, uQWs, Asura_PS2_VU1::VIFPK_V4_32, false );

            const u_long128* puSrc = &g_pxOutLightBfr->u128;
            for( ; uQWs; --uQWs )
            {
                (pxOut++)->u128 = *puSrc++;
            }

            bNeedNOPs        = true;
            g_uCurVULightBfr = uBfr;
        }

        if( uPendingFlags & PENDING_UPLOAD_SHADOW_DATA )
        {
            const u_int uBfr  = g_uCurVUShadowBfr ^ 1;
            const u_int uAddr = ( uBfr ? ASURA_PS2_GSM__ADDR__SHADOW_DATA_BFR_1 : ASURA_PS2_GSM__ADDR__SHADOW_DATA_BFR_0 );
        
            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
            }
            (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( uAddr, ASURA_PS2_GSM__SIZE__SHADOW_DATA_BFR, Asura_PS2_VU1::VIFPK_V4_32, false );

            // Shadow proj mat.
            if( g_bO2WIsIdentity )
            {
                Asura_PS2_VU1::CopyPS2Mat( pxOut, g_pxShadowMatBfr );
            }
            else
            {
                sceVu0MulMatrix( &pxOut->Vu0FVec, &g_pxShadowMatBfr->Vu0FVec, &g_pxO2W->Vu0FVec );
            }
            pxOut += 4;

            // Shadow pos and size.
            (pxOut++)->u128 = g_pxShadowPosAndSize->u128;

            // Shadow ST scales and quad alpha.
//RCD:not needed            Asura_PS2_Shadows::Proj_GetShadowQuadSTScales( pxOut->i32s );
            pxOut->u32s[3] = Asura_Maths::FloatToInt( Asura_Maths::Min(2.0f, g_fShadowLightLevel /*rcd removed as handled in VU1 * g_fShadowAlpha */ ) );
            ++pxOut;
        
			// place light level into shadow buffer structure since it will be needed again later
			Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer->SetLightLevel( g_fShadowLightLevel );

            bNeedNOPs         = true;
            g_uCurVUShadowBfr = uBfr;
        }

        if( uPendingFlags & PENDING_UPLOAD_SHADOW_PROJECTION )
        {
			// have to flush when changing these, because they may still be being used by previous block
            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
                pxOut->u32s[2] = SCE_VIF1_SET_FLUSHE( false );
            }
            else
            {
                (pxOut++)->u32s[3] = SCE_VIF1_SET_FLUSHE( false );
                FillQWWithNOPs( pxOut );
            }

            (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_GSM__ADDR__SHADOW_PALETTE, ASURA_PS2_GSM__SIZE__SHADOW_PALETTE_ENTRY * g_uShadowBuffers, Asura_PS2_VU1::VIFPK_V4_32, false );


			// go through list of (up to 5) ids which specify which shadow buffers we are going to use
			// and copy over their matrices.
			// note : I am assuming that modules don't have a O2W matrix.

			for (u_int i=0; i < g_uShadowBuffers; i++)
			{
				Asura_PS2_VU1::CopyPS2Mat( pxOut, reinterpret_cast<void*>(g_pxShadowMatList + 4 * g_piShadowBufIDs [i] ) );
				pxOut += 4;

				float fD = Asura_PS2_ShadowBufferSystem::GetShadowBufferDistanceToLight(g_piShadowBufIDs[i]);
				float fR = Asura_PS2_ShadowBufferSystem::GetShadowBufferRadius(g_piShadowBufIDs[i]);
				float fEnd = Asura_PS2_ShadowBuffer::s_fFalloffEnd;
				float fLen = Asura_PS2_ShadowBuffer::s_fFalloffLength;

				fEnd = fD + fEnd;	// adjust so that distance is from light rather than from object

				pxOut->f32s[0] = fD-( fR / SHADOW_FALLOFF_SLOPE_SCALEF );		// start allowing shadows just before centre - fixed scale for how far before
				pxOut->f32s[1] = fEnd;		// stop shadows beyond preset endpoint
				pxOut->f32s[2] = 1.0f/fLen;	// start fading them preset distance before this
				// spare for lighting
				pxOut->f32s[3] = Asura_PS2_ShadowBufferSystem::GetShadowBufferLightLevel(g_piShadowBufIDs[i]);
				pxOut++;
			}

            bNeedNOPs         = true;
        }

        if( uPendingFlags & (PENDING_UPLOAD_ST_OV_PARAMS | PENDING_UPLOAD_Y_PARAMS) )
        {
            // Note that we don't bother double buffering these parameters
            // because we don't expect them to change very often. We've
            // UNPACKed all the strip data and other stuff above already
            // anyway so a FLUSHE here isn't much of an issue.

            // (MP - 17.06.2005) Might need to do something about this now
            // that we've got proper animated texture support in the engine...

            if( bNeedNOPs )
            {
                FillQWWithNOPs( pxOut );
                pxOut->u32s[2] = SCE_VIF1_SET_FLUSHE( false );
            }
            else
            {
                (pxOut++)->u32s[3] = SCE_VIF1_SET_FLUSHE( false );
                FillQWWithNOPs( pxOut );
            }

            if( uPendingFlags & PENDING_UPLOAD_ST_OV_PARAMS )
            {
                (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_GSM__ADDR__ST_SCALE_AND_TRANS_PARAMS,
                                                          ASURA_PS2_GSM__SIZE__ST_SCALE_AND_TRANS_PARAMS,
                                                          Asura_PS2_VU1::VIFPK_V4_32,
                                                          false );

                (pxOut++)->u128 = g_pxSTOvParams->u128;
            }

            if( uPendingFlags & PENDING_UPLOAD_Y_PARAMS )
            {
                if( uPendingFlags & PENDING_UPLOAD_ST_OV_PARAMS )
                {
                    FillQWWithNOPs( pxOut );
                }
                (pxOut++)->u32s[3] = SCE_VIF1_SET_UNPACK( ASURA_PS2_GSM__ADDR__Y_SCALE_AND_TRANS_PARAMS,
                                                          ASURA_PS2_GSM__SIZE__Y_SCALE_AND_TRANS_PARAMS,
                                                          Asura_PS2_VU1::VIFPK_V4_32,
                                                          false );

                (pxOut++)->u128 = (g_pxYParams + 0)->u128;
                (pxOut++)->u128 = (g_pxYParams + 1)->u128;
            }

            bNeedNOPs = true;
        }

        if( !bNeedNOPs ) { (pxOut++)->u32s[3] = SCE_VIF1_SET_NOP( false ); }

        // Verify that we didn't underestimate.
        Asura_Assert( pxOut <= pxBase + uTotalQWs, "" );

        // Fix up tag.
        u_int uActualQWC = pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCNTTag + 1);
        pxCNTTag->m_uQWC = uActualQWC;

        // Return remainder of buffer if necessary.
        u_int uDiff = (pxBase + uTotalQWs) - pxOut;
        if( uDiff ) { Asura_PS2_DeferredRendering::ReturnVifBuffer( pxOut, uDiff ); }
    }

    if( uPendingFlags & PENDING_UPLOAD_TEX_SYNC )
    {
        Asura_PS2_TextureManagement::SetupTextureEnvironment( g_bTexUploaded, ASURA_PS2_TEXREGFLAG_TEXFLUSH );
    }

    // Add any necessary MSCALs.
    {
        u_int uEstimatedQWC = 0;
        // MSCALs to update ptrs after UNPACKs above.
        if( uPendingFlags & (PENDING_UPLOAD_RENDERSTATE | PENDING_UPLOAD_CAM | PENDING_UPLOAD_LIGHT | PENDING_UPLOAD_SHADOW_DATA) ) { ++uEstimatedQWC; }
        // ITOPs and MSCALs for mesh/cam flags.
        if( uPendingFlags & (PENDING_UPLOAD_MESH_FLAGS | PENDING_UPLOAD_CAM_FLAGS) ) { ++uEstimatedQWC; }

        const u_int uTotalQWs = 1 + uEstimatedQWC;
        Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
        Asura_PS2_QW* pxOut = pxBase;

        Asura_PS2_DmaTag* const pxCNTTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut );
        pxCNTTag->Set( 0, // Will be fixed up below as necessary.
                       Asura_PS2_DmaTag::PCE_NOCHANGE,
                       Asura_PS2_DmaTag::ID_CNT,
                       false,
                       0,
                       false );

        u_int uNumMSCALs = 0;
        u_int uWordOut   = 2;

        if( uPendingFlags & PENDING_UPLOAD_RENDERSTATE )
        {
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_SWITCH_RENDERSTATE_BFR, uNumMSCALs );
        }   
        if( uPendingFlags & PENDING_UPLOAD_CAM )
        {
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_SWITCH_CAM_DATA_BFR, uNumMSCALs );
        }
        if( uPendingFlags & PENDING_UPLOAD_LIGHT )
        {
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_SWITCH_LIGHT_DATA_BFR, uNumMSCALs );
        }
        if( uPendingFlags & PENDING_UPLOAD_SHADOW_DATA )
        {
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_SWITCH_SHADOW_DATA_BFR, uNumMSCALs );
        }
        if( uPendingFlags & PENDING_UPLOAD_MESH_FLAGS )
        {
            pxOut->u32s[uWordOut++] = SCE_VIF1_SET_ITOP( uMeshFlags, false );
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_LOAD_MESH_FLAGS, uNumMSCALs );

            UpdatePassesPerTri( uMeshFlags );
        }
        if( uPendingFlags & PENDING_UPLOAD_CAM_FLAGS )
        {
            pxOut->u32s[uWordOut++] = SCE_VIF1_SET_ITOP( g_uCamFlagsForVU, false );
            pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_LOAD_CAM_FLAGS, uNumMSCALs );
        }

        // We need to run the NOP program up to two times:
        //     (i) If we're transferring a whole chain of strips and so will
        //         be MSCNTing to render the next strip, we need a NOP to get
        //         the VU1 PC to point to the right place.
        //    (ii) If we've written an odd number of MSCALs above, including
        //         (i), we need a dummy one to get the TOP register back to
        //         the correct value, ie. pointing to the strip data that the
        //         caller UNPACKed.
        if( bWillMSCNTForNextStrip ) { pxOut->u32s[uWordOut++] = MAKE_MSCAL_FROM_VU_PTR_WITH_COUNT( VU1_GSM_NOP, uNumMSCALs ); }
        if( uNumMSCALs & 1         ) { pxOut->u32s[uWordOut++] = ASURA_PS2_MAKE_MSCAL_FROM_VU_PTR( VU1_GSM_NOP ); }

        // Pad to end of QW.
        if( uWordOut & 3 )
        {
            for( u_int u = 4 - (uWordOut & 3); u--; )
            {
                pxOut->u32s[uWordOut++] = SCE_VIF1_SET_NOP( false );
            }
        }

        // Fix up output ptr.
        pxOut += (uWordOut >> 2);

        // Verify that we didn't underestimate.
        Asura_Assert( pxOut <= pxBase + uTotalQWs, "" );

        // Fix up tag.
        u_int uActualQWC = pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCNTTag + 1);
        pxCNTTag->m_uQWC = uActualQWC;

        // Return remainder of buffer if necessary.
        u_int uDiff = (pxBase + uTotalQWs) - pxOut;
        if( uDiff ) { Asura_PS2_DeferredRendering::ReturnVifBuffer( pxOut, uDiff ); }
    }

    // Clear pending flags.
    g_uPendingFlags = 0;
}
