//
// Asura_PS2_ShadowBuffer.cpp
//
// Created by RCD - 23 December 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS2_ShadowBuffer.h"

// Engine independent.
#include "Animation/Asura_Hierarchy.h"
#include "Core/Asura_Material_System.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Env_System.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_ShadowBuffer.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PS2_2D_Graphics.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_GSMRenderer.h"
#include "Rendering/Asura_PS2_GSMDefines.h"
#include "Rendering/Asura_PS2_Shadows.h"
#include "Rendering/Asura_PS2_VU1.h"

#ifdef _DEBUG
bool dbg_bShadowUseAAForRenderToBfr = true;
#endif

DEBUGVAR bool dbg_bShadowUseDATE = true;
DEBUGVAR bool dbg_bDisplayShadowBuffersAlpha = false;
DEBUGVAR bool dbg_bDisableNonHierarchyShadows = false;
DEBUGVAR bool dbg_bAlwaysFlushNewShadowBuffer = false;
DEBUGVAR bool dbg_bReportShadowBufferActivity = false;	// note : gets reset each frame, to avoid spewing millions of lines
DEBUGVAR bool dbg_bLimitShadowsChunks = false;
DEBUGVAR int  dbg_aiChunksToShadow[6] = {-1, -1, -1, -1, -1, -1};
DEBUGVAR int  dbg_iOnlyShadowThisModule = -1;
DEBUGVAR int  dbg_iOnlyUseThisShadow = -1;

DEBUGVAR float dbg_fLightHeight = -ASURA_PS2_SHADOW_CASTING_LIGHT_OFFSET_Y;
DEBUGVAR bool  dbg_bDisableSceneryModules = true;

DEBUGVAR bool dbg_bFallbackAllShadows = false;

bool bReportingActivityThisFrame = false;	// 

/*******************************************************************************/
/* ASURA_PS2_SHADOWBUFFER                                                     */
/*******************************************************************************/

/* ---------------------------------- */
/* static members                     */
/* ---------------------------------- */

Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = NULL;
u_int Asura_PS2_ShadowBuffer::s_uShadowBufferNextPointer = 0;
u_int Asura_PS2_ShadowBuffer::s_uShadowBufferLimit;

//!	shadow fall off is computed at vertices of polygons, and as such it is liable to show innaccuracies
//! whenever a vertex is beyond the distance where fall-off hits zero, the shadow strength will be
//! smeared out over the whole polygon rather than hitting zero part way across
float Asura_PS2_ShadowBuffer::s_fFalloffLength = 1.0f;	// NOTE : do NOT use small values - see above
float Asura_PS2_ShadowBuffer::s_fFalloffEnd = 5.0f;	

bool Asura_PS2_ShadowBufferSystem::s_bBufferHasChanged = true;
Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBufferSystem::s_pxLastUsedShadowBuffer = NULL;

/* ---------------------------------- */
/* Functions                          */
/* ---------------------------------- */

Asura_PS2_ShadowBuffer::Asura_PS2_ShadowBuffer(u_int uBufferWidth /* = DEFAULT_WIDTH */, u_int uBufferHeight /* = DEFAULT_HEIGHT */)
	: m_uBufferWidth(uBufferWidth), m_uBufferHeight(uBufferHeight), 
	  m_bIsCurrentlyRenderingToBuffer(false), m_bShadowTransformIsDirty(true),
	  m_pxAssociatedAnim(NULL), m_eBufferStatus(BUFFER_UNUSED)
{
	// Enforce power of two sizes
	Asura_Assert((m_uBufferWidth & (m_uBufferWidth-1))==0, "Width must be a power of two");
	Asura_Assert((m_uBufferHeight & (m_uBufferHeight-1))==0, "Height must be a power of two");

	m_fLightLevel = 1.0f;	// default to something sensible
	m_fBufferScale = 1.0f;	// default scale
}

Asura_PS2_ShadowBuffer::~Asura_PS2_ShadowBuffer()
{
}

void Asura_PS2_ShadowBuffer::AllocateVRAM(void)
{
	// Allocate VRAM for the buffer - if enough is available
	u_int uPageWidth  = (m_uBufferWidth  + 63) / 64;	// each 32 bit z buffer page is 64 pixels wide
	u_int uPageHeight = (m_uBufferHeight + 31) / 32;	// and 32 pixels high

	u_int uPagesRequired = uPageWidth * uPageHeight;

	Asura_Assert(s_uShadowBufferNextPointer + uPagesRequired <= s_uShadowBufferLimit, "Out of VRAM for shadow buffer");
	m_uBufferBasePointer = s_uShadowBufferNextPointer;

	s_uShadowBufferNextPointer += uPagesRequired;	
}

void Asura_PS2_ShadowBuffer::MakeClean()
{
	// Clear the shadow buffer
	ClearAndSetupRenderToBfr(SCE_VIF1_SET_FLUSH(false));
	m_eBufferStatus = BUFFER_CLEAN;
}

void Asura_PS2_ShadowBuffer::BeginRenderingIntoBuffer()
{
	m_bIsCurrentlyRenderingToBuffer = true;
	JustSetupRenderToBfr(SCE_VIF1_SET_FLUSH(false));	// sets up the necessary renderstates ETC
	m_eBufferStatus = BUFFER_DIRTY;
}

void Asura_PS2_ShadowBuffer::EndRenderingIntoBuffer()
{
	m_bIsCurrentlyRenderingToBuffer = false;
}

void Asura_PS2_ShadowBuffer::ReleaseShadowBuffer()
{
	m_eBufferStatus = BUFFER_UNUSED;
}

void Asura_PS2_ShadowBuffer::Proj_GetTEX1andTEX0( u_long* puTEX1, u_long* puTEX0 ) const
{
	if (puTEX1)
	{
	    *puTEX1 = SCE_GS_SET_TEX1( 0,                                           // LCM - irrelevant since MXL is 0.
		                           0,                                           // MXL.
			                       SCE_GS_LINEAR, SCE_GS_LINEAR_MIPMAP_NEAREST, // MMAG, MMIN.
				                   0, 0, 0 );                                   // MTBA, L, K - irrelevant since MXL is 0.
	}

    const u_int uW  = m_uBufferWidth; // Also H.
    const u_int uTW = Asura_Maths::GetCeilLog2(m_uBufferWidth); // Also TH.

    *puTEX0 = SCE_GS_SET_TEX0( m_uBufferBasePointer << 5,    // TBP.	// convert from page to block address
                               (uW + 63) >> 6,			// TBW.
                               SCE_GS_PSMCT32,			// PSM.
                               uTW, uTW,				// TW, TH.
                               1,						// TCC.
                               SCE_GS_MODULATE,			// TFX.
                               0, 0, 0, 0, 0 );			// CLUT - irrelevant.
}

void Asura_PS2_ShadowBuffer::SetPositionAndSize(const Asura_Vector_3& xBSCentre, float fBSRadius)
{
	// Set up the light position as being five metres above the top of the bounding sphere
	Asura_Vector_3 xLightPosition = xBSCentre;
	xLightPosition.y -= fBSRadius + dbg_fLightHeight;
	SetLightPosition(xLightPosition);

	// Set up a perspective light projection
	SetPerspectiveEnclosingSphere(xBSCentre, fBSRadius, fBSRadius * 2.0f + dbg_fLightHeight + 8.0f);	// 3rd parameter defines how deep the shadow is cast

	// set shadow object radius (used in falloff calcs)
	m_fObjectRadius = fBSRadius;

	// Use the bottom of the bounding sphere as the base height - only necessary for drawing bounding box and for fallback quad shadows
	m_fBaseLevel = xBSCentre.y + fBSRadius;	// note this is expected top be over-ridden in many cases
}

void Asura_PS2_ShadowBuffer::SetLightPosition(const Asura_Vector_3& xLightPosition)
{
	m_xLightPosition = xLightPosition;
	m_xLightDirection = Asura_Vector_3(0.0f, 3.0f, 0.0f);

	m_bShadowTransformIsDirty = true;
}

void Asura_PS2_ShadowBuffer::SetShadowBufferQuadHeight(float fQuadHeight)
{
	m_fBaseLevel = fQuadHeight;
}

static inline float SphereProjectionPlane( float fDistance, float fBSRadius, float fMaxProjSize )
{
/*
	Angle from light to tangent on sphere is Theta
	We want to know the distance at which fMaxProjSize fits inside this tangent
	The ratio of this distance to (fMaxProjSize * 0.5) is Tan Theta
	Tan Theta is also (by definition) equal to fBSRadius / tangentLength
	tangentLength (using Pythagoras) is just sqrt(distance^2 - radius^2)
	so our answer is :
	[ 0.5 * fMaxProjSize * sqrt(distance^2 - radius^2) ] / fBSRadius
*/

	const float fTangentLength = Asura_Maths::Sqrt( fDistance * fDistance - fBSRadius * fBSRadius );

	// Calculate projection distance
	return ( ( 0.5f * fMaxProjSize * fTangentLength ) / fBSRadius );
}

static void GetWorldToLightMatrix( sceVu0FMATRIX mOut, const Asura_Vector_3& xLightPos)
{
	// NOTE : we are assuming that all lights are straight down (just like the XBox version)
	// this is of course sheer laziness, but certainly cuts down on the calculations required !

    // Calculate all components.
    mOut[0][0] = 1.0f;	// X remains unchange
    mOut[0][1] = 0.0f;
    mOut[0][2] = 0.0f;
    mOut[0][3] = 0.0f;

	mOut[1][0] = 0.0f;	// Y swaps with Z
	mOut[1][1] = 0.0f;
	mOut[1][2] = 1.0f;
	mOut[1][3] = 0.0f;

	mOut[2][0] = 0.0f;	// and vice versa - strictly speaking this causes an inversion of the space
	mOut[2][1] = 1.0f;	// however it causes no problems for us because the same effect applies
	mOut[2][2] = 0.0f;	// to both the projection to render the texture and to the output UV's
	mOut[2][3] = 0.0f;	// when rendering the shadow onto the environment

	mOut[3][0] = -xLightPos.x;	// this removes the light position after it has been rotated
	mOut[3][1] = -xLightPos.z;	// hence the swapped z and y
	mOut[3][2] = -xLightPos.y;
	mOut[3][3] = 0.0f;

}



// NOTE : still editing this bit - needs to generate two matrices like the XBox one
void Asura_PS2_ShadowBuffer::CalcShadowProjMat( float fLightScreenZ, const Asura_Vector_3& xLightPos, float fNearLimit, float fFarLimit)
{
    // Setup view->screen.
    sceVu0FMATRIX xLightViewScreen;
    sceVu0ViewScreenMatrix( xLightViewScreen, fLightScreenZ,
                            1.0f, 1.0f, // Aspect Ratios (1:1 for render to texture)
                            2048.0f, 2048.0f,
                            0.0f, 1.0f,  // arbitray z values since z-buffer not used.
                            1.0f, 16777216.0f  );

	sceVu0FMATRIX xLightViewTexture; //same as above but into UV space (0-1) instead of texel space (0-width)
    sceVu0ViewScreenMatrix( xLightViewTexture, fLightScreenZ / m_uBufferWidth,
                            1.0f, 1.0f, // Aspect Ratios (1:1 for render to texture)
                            0.0f, 0.0f, // offset centre to middle of texture
                            -1.0f, 0.0f,  // set with the intention of using the matrix Z result as a means for fading the shadow
                            fNearLimit, fFarLimit  );

    // View->world and then world->view. 
    sceVu0FMATRIX xLightWorldToX;

	GetWorldToLightMatrix(xLightWorldToX, xLightPos);

 	// generate two matrices - one which goes world to shadow texture, for rendering into texture
	// and the other which does world to shadow projection, for rendering shadow texture onto the ground

    sceVu0MulMatrix( m_xWorldToShadowMatrix,    xLightViewScreen,  xLightWorldToX);
    sceVu0MulMatrix( m_xShadowProjectionMatrix, xLightViewTexture, xLightWorldToX);
}


// Note : the reason this code is broken up as it is to allow for more flexible variations of this approach.
// so for example, we have one function to compute the necessary projection distance, which is designed to work 
// for the sphere case, then we can easily put in a different one for other cases.
// But the result from whichever function decides on the projection distance can be used by the same set of
// functions, which go and build the necessary matrices
void Asura_PS2_ShadowBuffer::SetPerspectiveEnclosingSphere(const Asura_Vector_3& xSphereCenter, float fSphereRadius, float fDistance)
{
	// Assume that we already know the light position

	// We want to build a frustum that encloses the sphere described by xSphereCenter and fSphereRadius
	// i.e. the sphere touches the edges of the frustum

	m_fDistanceToLight = (xSphereCenter - m_xLightPosition).Magnitude();

    float fLightScreenZ;
    // Calculate distance to projection plane
    fLightScreenZ = SphereProjectionPlane( m_fDistanceToLight, fSphereRadius, m_uBufferWidth - 2 );

	// allow shadow to be enlarged by a fixed scale per project
	fLightScreenZ *= m_fBufferScale;

	// record the distance that the shadow will cast.  Note that on the XBox this is the depth buffer far Z 
	// and is used when computing the shadow box which in turn determines the selection of modules on which
	// to cast the shadow.  This is intended to produce approximately the same behaviour
	m_fDistanceLimit = fDistance;
	m_fBaseWidth = 0.5 * m_fDistanceLimit * m_uBufferWidth / fLightScreenZ;

	// generate the matrices required
	CalcShadowProjMat(fLightScreenZ, m_xLightPosition, m_fDistanceToLight - fSphereRadius, fDistance);

	if(m_bIsCurrentlyRenderingToBuffer)
	{
		// Upload the new matrix
		// PS2 :: Can this happen ?  How ?  What would need doing if it did ?
	}

	m_bShadowTransformIsDirty = true;
}

void Asura_PS2_ShadowBuffer::RecalcShadowTransform()
{
	if(!m_bShadowTransformIsDirty)return;

	m_xWorldSpaceShadowBox.MinY = m_xLightPosition.y;
	m_xWorldSpaceShadowBox.MaxY = m_xLightPosition.y + m_fDistanceLimit;

	m_xWorldSpaceShadowBox.MinX = m_xLightPosition.x - m_fBaseWidth;
	m_xWorldSpaceShadowBox.MaxX = m_xLightPosition.x + m_fBaseWidth;

	m_xWorldSpaceShadowBox.MinZ = m_xLightPosition.z - m_fBaseWidth;
	m_xWorldSpaceShadowBox.MaxZ = m_xLightPosition.z + m_fBaseWidth;

	m_bShadowTransformIsDirty = false;
}

const sceVu0FMATRIX *Asura_PS2_ShadowBuffer::GetShadowSpaceTransform()
{
	RecalcShadowTransform();
	return reinterpret_cast<const sceVu0FMATRIX *>(&m_xShadowProjectionMatrix);
}

const Asura_Bounding_Box& Asura_PS2_ShadowBuffer::GetWorldSpaceShadowBox()
{
	RecalcShadowTransform();
	return m_xWorldSpaceShadowBox;
}

#ifdef _DEBUG
void Asura_PS2_ShadowBuffer::DrawShadowBox()
{
	// Shadow box
	RecalcShadowTransform();
	Asura_Vector_3 xPos; xPos.Zero();
	Asura_Matrix_3x3 xOri; xOri.Identity();
	Asura_Render::RenderBB(m_xWorldSpaceShadowBox, xPos, xOri);

#if 0 /// needs a totally different approach for PS2 (if any approach at all !)
	// Light + lightLook
	sceVu0FVECTOR xVec[2] = { { 0.0f, 0.0f, 0.0f, 1.0f},
							  { 0.0f, 0.0f, 3.0f, 1.0f}, };
	sceVu0FMATRIX xInverse;
	sceVu0InversMatrix(xInverse, m_xNewViewMatrix);
	sceVu0ApplyMatrix(xVec[0], xInverse, xVec[0]);
	sceVu0ApplyMatrix(xVec[1], xInverse, xVec[1]);
	Asura_Render::RenderLine(Asura_Vector_3(xVec[0]), Asura_Vector_3(xVec[1]), 0xFFFFFFFF);

	// base plane
	Asura_Render::RenderLine(Asura_Vector_3(m_xWorldSpaceShadowBox.MinX, m_fBaseLevel, m_xWorldSpaceShadowBox.MinZ), Asura_Vector_3(m_xWorldSpaceShadowBox.MaxX, m_fBaseLevel, m_xWorldSpaceShadowBox.MinZ), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(m_xWorldSpaceShadowBox.MaxX, m_fBaseLevel, m_xWorldSpaceShadowBox.MinZ), Asura_Vector_3(m_xWorldSpaceShadowBox.MaxX, m_fBaseLevel, m_xWorldSpaceShadowBox.MaxZ), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(m_xWorldSpaceShadowBox.MaxX, m_fBaseLevel, m_xWorldSpaceShadowBox.MaxZ), Asura_Vector_3(m_xWorldSpaceShadowBox.MinX, m_fBaseLevel, m_xWorldSpaceShadowBox.MaxZ), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(m_xWorldSpaceShadowBox.MinX, m_fBaseLevel, m_xWorldSpaceShadowBox.MaxZ), Asura_Vector_3(m_xWorldSpaceShadowBox.MinX, m_fBaseLevel, m_xWorldSpaceShadowBox.MinZ), 0xFFFFFFFF);

	//frustum
	sceVu0MulMatrix(xInverse, m_xNewViewMatrix, m_xNewProjMatrix);
	sceVu0InversMatrix(xInverse, xInverse);

	sceVu0FVECTOR xF[8] = {	{-1.0f, -1.0f, 0.0f, 1.0f},
							{-1.0f, -1.0f, 1.0f, 1.0f},
							{-1.0f,  1.0f, 0.0f, 1.0f},
							{-1.0f,  1.0f, 1.0f, 1.0f},
							{ 1.0f, -1.0f, 0.0f, 1.0f},
							{ 1.0f, -1.0f, 1.0f, 1.0f},
							{ 1.0f,  1.0f, 0.0f, 1.0f},
							{ 1.0f,  1.0f, 1.0f, 1.0f} };
	for(int i = 0; i < 8; ++i)
	{
		sceVu0ApplyMatrix(xF[i], xInverse, xF[i]);
	}

	Asura_Render::RenderLine(Asura_Vector_3(xF[0]), Asura_Vector_3(xF[1]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[1]), Asura_Vector_3(xF[2]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[2]), Asura_Vector_3(xF[3]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[3]), Asura_Vector_3(xF[0]), 0xFFFFFFFF);

	Asura_Render::RenderLine(Asura_Vector_3(xF[4]), Asura_Vector_3(xF[5]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[5]), Asura_Vector_3(xF[6]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[6]), Asura_Vector_3(xF[7]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[7]), Asura_Vector_3(xF[4]), 0xFFFFFFFF);

	Asura_Render::RenderLine(Asura_Vector_3(xF[0]), Asura_Vector_3(xF[4]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[1]), Asura_Vector_3(xF[5]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[2]), Asura_Vector_3(xF[6]), 0xFFFFFFFF);
	Asura_Render::RenderLine(Asura_Vector_3(xF[3]), Asura_Vector_3(xF[7]), 0xFFFFFFFF);

#endif
}
#endif

// shadow buffer manipulation stuff
DEBUGVAR int dbg_iShadowCol = 0;

void Asura_PS2_ShadowBuffer::ClearAndSetupRenderToBfr( u_int uFLUSH )
{
    const u_int uDim = m_uBufferWidth;
    const u_int uFBP = m_uBufferBasePointer;	// this is a page address, not a block one
    const u_int uFBW = (uDim + 63) >> 6;
    const u_int uTL  = 2048 - (uDim >> 1); // Used for both x and y since width and height are assumed to be equal.

    const u_int uNLOOP    = 4 + Asura_PS2_2D_Graphics::GetNumQWsForSpriteStrips(uDim, false) + (dbg_bShadowUseAAForRenderToBfr ? 1 : 0);
    const u_int uQWC      = Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS + 1 + uNLOOP;
    const u_int uTotalQWs = 1 + uQWC;

    Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
    Asura_PS2_QW* pxOut = pxBase;

    // DMA tag.
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWC,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         dbg_bAlwaysFlushNewShadowBuffer ? SCE_VIF1_SET_FLUSH(false) : uFLUSH ,
                                                         SCE_VIF1_SET_DIRECT(uQWC, false) );

    // Setup context 2 renderstate for the clear below.
    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_2, pxOut, false, true );
    // Disable depth testing and prevent writes to the z buffer.
    Asura_PlatformRenderStates::SetPixelTest( true, false, 0, SCE_GS_ALPHA_NEVER, SCE_GS_AFAIL_FB_ONLY );
    // Update ptr.
    Asura_PlatformRenderStates::ForgetMyBuffer( pxOut );
    pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

    // GIF tag for the clear.
    {
        Asura_PS2_GifTag* pxGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
        pxGifTag->SetExceptRegs( uNLOOP,
                                 true,
                                 true,
                                 SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                                  0,                  // IIP
                                                  false,              // TME
                                                  false,              // FGE
                                                  false,              // ABE
                                                  false,              // AA1
                                                  0,                  // FST
                                                  1,                  // CTXT
                                                  0 ),                // FIX 
                                 SCE_GIF_PACKED,
                                 1 );
        pxGifTag->uAllRegs = SCE_GIF_PACKED_AD;
    }

    // Viewport setup.
    pxOut->u64s[0] = SCE_GS_SET_FRAME( uFBP, uFBW, SCE_GS_PSMCT32, 0x00FFFFFF ); // Mask RGB.
    pxOut->u64s[1] = SCE_GS_FRAME_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_XYOFFSET( uTL << 4, uTL << 4 );
    pxOut->u64s[1] = SCE_GS_XYOFFSET_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_SCISSOR( 0, uDim - 1, 0, uDim - 1 );
    pxOut->u64s[1] = SCE_GS_SCISSOR_2;
    ++pxOut;

    // Sprites to clear bfr.
    // TODO: Should be able to clear faster here using the fake z buffer
    //       trick. May be significant for larger shadows.
    pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 0, 0, 0, dbg_iShadowCol, 0x3F800000 );
    pxOut->u64s[1] = SCE_GS_RGBAQ;
    ++pxOut;
    pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut, uTL, uTL, uDim, uDim, false );

    // Viewport setup is unchanged but we need to modify the pixel test to
    // get a correct cumulative effect when used in conjunction with the GS's
    // built-in edge anti-aliasing.
    if( dbg_bShadowUseAAForRenderToBfr )
    {
        pxOut->u64s[0] = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_FB_ONLY, // ATE, ATST, AREF, AFAIL
                                          dbg_bShadowUseDATE, 0,                             // DATE, DATM
                                          true, SCE_GS_ZALWAYS );                            // ZTE, ZTST
        pxOut->u64s[1] = SCE_GS_TEST_2;
        ++pxOut;
    }

    // Verify that we wrote the predicted number of QWs.
    Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );
}

// basically identical to ClearAndSetup function above, but without the clearing sprites - needed because of multiple shadow buffers
void Asura_PS2_ShadowBuffer::JustSetupRenderToBfr( u_int uFLUSH )
{
    const u_int uDim = m_uBufferWidth;
    const u_int uFBP = m_uBufferBasePointer;	// this is a page address, not a block one
    const u_int uFBW = (uDim + 63) >> 6;
    const u_int uTL  = 2048 - (uDim >> 1); // Used for both x and y since width and height are assumed to be equal.

    const u_int uNLOOP    = 3 + (dbg_bShadowUseAAForRenderToBfr ? 1 : 0);
    const u_int uQWC      = Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS + 1 + uNLOOP;
    const u_int uTotalQWs = 1 + uQWC;

    Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
    Asura_PS2_QW* pxOut = pxBase;

    // DMA tag.
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWC,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         dbg_bAlwaysFlushNewShadowBuffer ? SCE_VIF1_SET_FLUSH(false) : uFLUSH ,
                                                         SCE_VIF1_SET_DIRECT(uQWC, false) );

    // Setup context 2 renderstate for rendering
    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_2, pxOut, false, true );
    // Disable depth testing and prevent writes to the z buffer.
    Asura_PlatformRenderStates::SetPixelTest( true, false, 0, SCE_GS_ALPHA_NEVER, SCE_GS_AFAIL_FB_ONLY );
    // Update ptr.
    Asura_PlatformRenderStates::ForgetMyBuffer( pxOut );
    pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS;

    // GIF tag for the clear.
    {
        Asura_PS2_GifTag* pxGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
        pxGifTag->SetExceptRegs( uNLOOP,
                                 true,
                                 true,
                                 SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                                  0,                  // IIP
                                                  false,              // TME
                                                  false,              // FGE
                                                  false,              // ABE
                                                  false,              // AA1
                                                  0,                  // FST
                                                  1,                  // CTXT
                                                  0 ),                // FIX 
                                 SCE_GIF_PACKED,
                                 1 );
        pxGifTag->uAllRegs = SCE_GIF_PACKED_AD;
    }

    // Viewport setup.
    pxOut->u64s[0] = SCE_GS_SET_FRAME( uFBP, uFBW, SCE_GS_PSMCT32, 0x00FFFFFF ); // Mask RGB.
    pxOut->u64s[1] = SCE_GS_FRAME_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_XYOFFSET( uTL << 4, uTL << 4 );
    pxOut->u64s[1] = SCE_GS_XYOFFSET_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_SCISSOR( 1, uDim - 2, 1, uDim - 2 );	// note : in 1 from both edges
    pxOut->u64s[1] = SCE_GS_SCISSOR_2;
    ++pxOut;

    // Viewport setup is unchanged but we need to modify the pixel test to
    // get a correct cumulative effect when used in conjunction with the GS's
    // built-in edge anti-aliasing.
    if( dbg_bShadowUseAAForRenderToBfr )
    {
        pxOut->u64s[0] = SCE_GS_SET_TEST( true, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_FB_ONLY, // ATE, ATST, AREF, AFAIL
                                          dbg_bShadowUseDATE, 0,                             // DATE, DATM
                                          true, SCE_GS_ZALWAYS );                            // ZTE, ZTST
        pxOut->u64s[1] = SCE_GS_TEST_2;
        ++pxOut;
    }

    // Verify that we wrote the predicted number of QWs.
    Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );
}

void Asura_PS2_ShadowBuffer::Proj_GetShadowProjMat( float* pfO2W, float* pfMatOut )
{
    // Just calculate {object|world}->screen.
    if( pfO2W )
    {
        sceVu0MulMatrix( reinterpret_cast<sceVu0FVECTOR*>(pfMatOut), m_xWorldToShadowMatrix, reinterpret_cast<sceVu0FVECTOR*>(pfO2W) );
    }
    else
    {
        sceVu0CopyMatrix( reinterpret_cast<sceVu0FVECTOR*>(pfMatOut), m_xWorldToShadowMatrix);
    }
}

void Asura_PS2_ShadowBuffer::FindAffectedModules()
{
	// Do a simple BB test against the module tree
	RecalcShadowTransform();

	Asura_Assert(s_pxCurrentShadowBuffer == NULL, "Something else is in the middle of testing still!");
	s_pxCurrentShadowBuffer = this;
	m_uModulesFound = 0;

	Asura_Module_System::TestModuleTree(m_xWorldSpaceShadowBox, FindAffectedModulesCallback);

	s_pxCurrentShadowBuffer = NULL;
}

/* ---------------------------------- */
/* static Functions                   */
/* ---------------------------------- */

void Asura_PS2_ShadowBuffer::FindAffectedModulesCallback(const u_int uModuleIndex)
{
	Asura_Assert(s_pxCurrentShadowBuffer, "Testing with no current shadow buffer set!");

	// if the module isn't modlinked in, ignore it
	if(!Asura_Module_System::IsModuleLinkedInForCurrentCamera(uModuleIndex))return;
	// how about a frustum cull, then?
	if(!Asura_Module_System::IsModuleInFrustum(uModuleIndex))return;

	if (s_pxCurrentShadowBuffer->m_uModulesFound != AFFECTED_MODULES_LIST_SIZE)
	{
		// I guess it qualifies, then.
		// Add it to the list.
		s_pxCurrentShadowBuffer->m_ausModuleIDs[s_pxCurrentShadowBuffer->m_uModulesFound] = uModuleIndex;

		Asura_PS2_ShadowBufferSystem::AddModuleReference(uModuleIndex, s_pxCurrentShadowBuffer->m_uIndex);
		if (bReportingActivityThisFrame)
		{
			printf("Module %d associated with shadowbuffer %d\n", uModuleIndex, s_pxCurrentShadowBuffer->m_uIndex);
		}

		++s_pxCurrentShadowBuffer->m_uModulesFound;
	}
}

/*******************************************************************************/
/* ASURA_PS2_SHADOWBUFFERSYSTEM                                               */
/*******************************************************************************/

/* ---------------------------------- */
/* static members                     */
/* ---------------------------------- */

Asura_PS2_ShadowBufferSystem::ModuleShadowReference Asura_PS2_ShadowBufferSystem::s_axModuleControls[Asura_PS2_ShadowBufferSystem::MAXIMUM_NUMBER_OF_MODULES];
Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBufferSystem::s_apxShadowBuffers[ASURA_PS2_SHADOW_MAX_BUFFERS];
bool Asura_PS2_ShadowBufferSystem::s_bStillCollectingShadowsThisFrame = false;

/* ---------------------------------- */
/* Static Functions                   */
/* ---------------------------------- */

void Asura_PS2_ShadowBufferSystem::Initialise()
{
    // This code only works with a 24-bit z buffer, since it needs to hide
    // its data in the spare 8 bits.
    Asura_Assert( Asura_PS2_GS::GetZBfrBitDepth() == 24, "" );

    Asura_PS2_ShadowBuffer::s_uShadowBufferNextPointer = Asura_PS2_GS::GetZBfrBasePtr();	// reset base pointer - note counting pages, not blocks, hence >>5

	u_int uScreenWidthPages  = Asura_PS2_GS::GetScreenWidth()  / 64;
	u_int uScreenHeightPages = Asura_PS2_GS::GetScreenHeight() / 32;

	Asura_PS2_ShadowBuffer::s_uShadowBufferLimit = Asura_PS2_ShadowBuffer::s_uShadowBufferNextPointer + uScreenWidthPages * uScreenHeightPages;

	// Allocate all shadow buffers, marking them as unused
	for(u_int u = 0; u < ASURA_PS2_SHADOW_MAX_BUFFERS; ++u)
	{
		s_apxShadowBuffers[u] = new Asura_PS2_ShadowBuffer();
		s_apxShadowBuffers[u]->m_uIndex = u;
		if ( (u < ASURA_PS2_SHADOW_MAX_PROJECTED_BUFFERS) && !dbg_bFallbackAllShadows)
		{
			s_apxShadowBuffers[u]->m_bForceFallbackRender = false;
		}
		else
		{
			s_apxShadowBuffers[u]->m_bForceFallbackRender = true;
		}
	}
}

void Asura_PS2_ShadowBufferSystem::Shutdown()
{
	for(u_int u = 0; u < ASURA_PS2_SHADOW_MAX_BUFFERS; ++u)
	{
		delete s_apxShadowBuffers[u];
		s_apxShadowBuffers[u] = NULL;
	}
}

Asura_PS2_ShadowBufferSystem::ModuleShadowReference *Asura_PS2_ShadowBufferSystem::GetModuleReference( u_short usModuleIndex)
{
	for(u_int u = 0; u < MAXIMUM_NUMBER_OF_MODULES; ++u)
	{
		if ( (s_axModuleControls[u].iModuleIndex == usModuleIndex)
		   ||(s_axModuleControls[u].iModuleIndex == -1) )
		{
			return(&s_axModuleControls[u]);
		}
	}
	return(NULL);
}

void Asura_PS2_ShadowBufferSystem::AddModuleReference(u_short usModuleIndex, u_int uShadowBufferIndex)
{
	ModuleShadowReference *p_xModuleReference = GetModuleReference(usModuleIndex);

	if (p_xModuleReference)
	{
		p_xModuleReference->iModuleIndex = usModuleIndex;	// NOTE : overwrite in case it was empty
		p_xModuleReference->usShadowBufferFlags |= (1 << uShadowBufferIndex);
	}
}

Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBufferSystem::FindUnusedShadowBuffer(void)
{
	// Check through buffers for one not currently in use
	for(int i = 0; i < ASURA_PS2_SHADOW_MAX_BUFFERS; ++i)
	{
		Asura_PS2_ShadowBuffer* pxShadowBuffer = s_apxShadowBuffers[i];
		if(pxShadowBuffer->IsUnused())
		{
			return pxShadowBuffer;
		}
	}

	return NULL;
}


void Asura_PS2_ShadowBufferSystem::PreAllocate(const Asura_Animation* pxAnim, u_int uBufferWidth, float scale)
{
	// can we find an empty one
	Asura_PS2_ShadowBuffer* pxShadowBuffer = FindUnusedShadowBuffer();

	if (pxShadowBuffer)
	{
		if (bReportingActivityThisFrame)
		{
			printf("Pre-Alloc bfr %d\n", pxShadowBuffer->m_uIndex);
		}

		pxShadowBuffer->m_uBufferWidth = uBufferWidth;
		pxShadowBuffer->m_uBufferHeight = uBufferWidth;
		pxShadowBuffer->m_fBufferScale = scale;
		pxShadowBuffer->m_pxAssociatedAnim = pxAnim;

		pxShadowBuffer->m_eBufferStatus = Asura_PS2_ShadowBuffer::BUFFER_ALLOCATED;

		s_bBufferHasChanged = (s_pxLastUsedShadowBuffer != pxShadowBuffer);
		s_pxLastUsedShadowBuffer = pxShadowBuffer;

		pxShadowBuffer->AllocateVRAM();
	}
}

Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBufferSystem::GetUnusedShadowBuffer( const Asura_Vector_3& xBSCentre,
                                                                             float fBSRadius,
                                                                             const Asura_Animation* pxAnim,
                                                                             u_int uCamVisFlagsForCaster )
{
	// can we find an empty one
	Asura_PS2_ShadowBuffer* pxShadowBuffer = FindUnusedShadowBuffer();

	if (pxShadowBuffer)
	{
		if (bReportingActivityThisFrame)
		{
			printf("Alloc bfr %d\n", pxShadowBuffer->m_uIndex);
		}

		// yup, we can
		// since its unused, it hasn't been given a size yet, so we just allocate the default size
		pxShadowBuffer->m_uBufferWidth  = Asura_PS2_ShadowBuffer::DEFAULT_WIDTH;
		pxShadowBuffer->m_uBufferHeight = Asura_PS2_ShadowBuffer::DEFAULT_WIDTH;

		// decide where its going
		pxShadowBuffer->AllocateVRAM();

		// Clean it
		pxShadowBuffer->MakeClean();

		// set up the bits which effect projection etc
		pxShadowBuffer->SetPositionAndSize(xBSCentre, fBSRadius);

		// Build the affected modules list
		pxShadowBuffer->FindAffectedModules();

		// Mark the buffer as being set for this no anim, because its for a shape
		pxShadowBuffer->m_pxAssociatedAnim = pxAnim;
 
        // Store cam vis flags.
        pxShadowBuffer->m_uCamVisFlagsForCaster = uCamVisFlagsForCaster;

		s_bBufferHasChanged = (s_pxLastUsedShadowBuffer != pxShadowBuffer);
		s_pxLastUsedShadowBuffer = pxShadowBuffer;

		return pxShadowBuffer;
	}

	if (bReportingActivityThisFrame)
	{
		printf("No empty buffer found\n");
	}
	s_bBufferHasChanged = true;
	s_pxLastUsedShadowBuffer = NULL;
	// We can't get one
	return NULL;
}

Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBufferSystem::GetTemporaryShadowBuffer( const Asura_Vector_3& xBSCentre,
                                                                                float fBSRadius,
                                                                                u_int uCamVisFlagsForCaster )
{
	if(!s_bStillCollectingShadowsThisFrame) return NULL; // no more shadows this frame

	if (dbg_bDisableNonHierarchyShadows) return NULL;

	if (bReportingActivityThisFrame)
	{
		printf("Temp Bfr Req : ");
	}

	return GetUnusedShadowBuffer( xBSCentre, fBSRadius, NULL, uCamVisFlagsForCaster );
}


Asura_PS2_ShadowBuffer* Asura_PS2_ShadowBufferSystem::GetShadowBufferForAnim( const Asura_Animation* pxAnim,
                                                                              const Asura_Vector_3& xBSCentre,
                                                                              float fBSRadius,
                                                                              u_int uCamVisFlagsForCaster )
{
	Asura_Assert(pxAnim, "Tried to get shadow buffer for a null anim");

	if(!s_bStillCollectingShadowsThisFrame) return 0; // no more shadows this frame

	if (bReportingActivityThisFrame)
	{
		printf("Anim Bfr Req : %-24s : ", pxAnim->GetSkin()->Name.GetString());
	}

	// Track back till we get the uberparent of this anim
	while(pxAnim->GetParent())
	{
		pxAnim = pxAnim->GetParent();
	}

	if (bReportingActivityThisFrame)
	{
		printf("Parent skin name : %-24s : ", pxAnim->GetSkin()->Name.GetString());
	}

	// Check through our buffers. Is there already one dedicated to this uberparent?
	for(int i = 0; i < ASURA_PS2_SHADOW_MAX_BUFFERS; ++i)
	{
		Asura_PS2_ShadowBuffer* pxShadowBuffer = s_apxShadowBuffers[i];
		if((pxShadowBuffer->m_eBufferStatus != Asura_PS2_ShadowBuffer::BUFFER_UNUSED) && (pxShadowBuffer->m_pxAssociatedAnim == pxAnim))
		{
			if (pxShadowBuffer->m_eBufferStatus == Asura_PS2_ShadowBuffer::BUFFER_ALLOCATED)
			{
				if (bReportingActivityThisFrame)
				{
					printf("Using bfr %d\n", i);
				}

				// decide where its going
				pxShadowBuffer->AllocateVRAM();

				// Clean it
				pxShadowBuffer->MakeClean();

				// set up projection etc
				pxShadowBuffer->SetPositionAndSize(pxAnim->GetBoundingSphereCentre(), pxAnim->GetTightBoundingSphereRadius());

				// Build the affected modules list
				pxShadowBuffer->FindAffectedModules();
			}
			else
			{
				if (bReportingActivityThisFrame)
				{
					printf("Alloc bfr %d\n", i);
				}
			}

			s_bBufferHasChanged = (s_pxLastUsedShadowBuffer != s_apxShadowBuffers[i]);
			s_pxLastUsedShadowBuffer = s_apxShadowBuffers[i];
			// it's this one
			return s_apxShadowBuffers[i];
		}
	}

	// Apparently not. Get an unused one (or NULL if there are none)
	return GetUnusedShadowBuffer( pxAnim->GetBoundingSphereCentre(), pxAnim->GetTightBoundingSphereRadius(), pxAnim, uCamVisFlagsForCaster );
}

void Asura_PS2_ShadowBufferSystem::EndFrame()
{
	// scratch all buffers
	for(u_int u = 0; u < ASURA_PS2_SHADOW_MAX_BUFFERS; ++u)
	{
		s_apxShadowBuffers[u]->m_eBufferStatus = Asura_PS2_ShadowBuffer::BUFFER_UNUSED;
		s_apxShadowBuffers[u]->m_pxAssociatedAnim = NULL;
		s_apxShadowBuffers[u]->m_fBufferScale = 1.0f;
		if ( (u < ASURA_PS2_SHADOW_MAX_PROJECTED_BUFFERS) && !dbg_bFallbackAllShadows)
		{
			s_apxShadowBuffers[u]->m_bForceFallbackRender = false;
		}
		else
		{
			s_apxShadowBuffers[u]->m_bForceFallbackRender = true;
		}
	}

	// scratch list of modules and associated flags
	for(u_int u=0; u < MAXIMUM_NUMBER_OF_MODULES; ++u)
	{
		s_axModuleControls[u].iModuleIndex = -1;		// invalid index I hope
		s_axModuleControls[u].usShadowBufferFlags = 0;	// clear shadow flags
	}

}

void Asura_PS2_ShadowBufferSystem::BeginNewFrame()
{
    // This code only works with a 24-bit z buffer, since it needs to hide
    // its data in the spare 8 bits.
    Asura_Assert( Asura_PS2_GS::GetZBfrBitDepth() == 24, "" );

    Asura_PS2_ShadowBuffer::s_uShadowBufferNextPointer = Asura_PS2_GS::GetZBfrBasePtr();	// reset base pointer - note counting pages, not blocks, hence >>5

	s_pxLastUsedShadowBuffer = NULL;
	Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = NULL;
	s_bBufferHasChanged = true;

	s_bStillCollectingShadowsThisFrame = true;

	if (bReportingActivityThisFrame)
	{
		bReportingActivityThisFrame = false;	// if reporting last frame, stop now
	}
#ifdef _DEBUG
	if (dbg_bReportShadowBufferActivity)
	{
		bReportingActivityThisFrame  = true;	// if asked for, start reporting now
		dbg_bReportShadowBufferActivity = false;
	}
#endif
}

void Asura_PS2_ShadowBufferSystem::FinishedCollectingShadows()
{
    ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__PROJ_SHADOW_MARKER );

    // decide whether we need to use fallback case for everything - either if asked to or if there is no module data
	bool bFallbackAllShadows = dbg_bFallbackAllShadows || (s_paiModuleChunkCounts == NULL);

	// prepare all the basic data - cameras, renderstates, etc
	Asura_PS2_GSMRenderer::Begin();

	// prevent lighting data being uploaded - shadows don't use lights (trust me on this one)
	Asura_PS2_GSMRenderer::CancelLightingUpload();

	//NOTE : must render any quad shadows here, before we do anything else, not least 
	// because we will be trashing tha values they use!
	for (int i = 0; i < ASURA_PS2_SHADOW_MAX_BUFFERS; i++)
	{
		const Asura_PS2_ShadowBuffer* pBuffer = s_apxShadowBuffers[i];
		if (!pBuffer->IsUnused())
		{
			if (bFallbackAllShadows || pBuffer->m_bForceFallbackRender)
			{
				u_int uClipVisFlags = 0;
				{
					Asura_Vector_3 xShadowCentre = pBuffer->m_xLightPosition;
					xShadowCentre.y = pBuffer->m_fBaseLevel;
					const float fShadowRadius = pBuffer->m_fObjectRadius;
					const int iCameras = Asura_Camera::GetNumberOfCameras();
                    const u_int uCamVisFlagsForCaster = pBuffer->m_uCamVisFlagsForCaster;

					// determine if this intersection is visible from ANY camera, and if so which ones
					for (int iCam=0; iCam < iCameras; iCam++)
					{
						Asura_Camera::SetCameraIndex(iCam);

                        if( !(uCamVisFlagsForCaster & (1 << iCam)) ) { continue; }

						if ( Asura_Camera::IsSphereInFrustum(xShadowCentre, fShadowRadius) )
						{
							if (!pBuffer->m_pxAssociatedAnim ||
								pBuffer->m_pxAssociatedAnim->GetParentEntityGUID() != Asura_Camera::GetEntityToIgnore())
							{
								uClipVisFlags |= 2;
//								if( !Asura_PS2_VU1::IsSphereEntirelyWithinGuardbandFrustum(xShadowCentre, fIntersectionRadius) )
								{
									uClipVisFlags |= 1;
								}
							}
						}

						uClipVisFlags <<= 2;
					}
					
					// move bits up so that camera zero is at bits 8 & 7
					uClipVisFlags <<= (2 * ( ASURA_MAX_NO_OF_CAMERAS - iCameras - 1) );

					if (uClipVisFlags == 0) 
					{
						continue;			// skip this shadow if can't be seen
					}
				}
				Asura_PS2_GSMRenderer::GetShadowBufferPosition(i);
				Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = s_apxShadowBuffers[i];
				Asura_PS2_GSMRenderer::RenderShadowQuad(uClipVisFlags);
				//s_apxShadowBuffers[i]->ReleaseShadowBuffer();	// don't use it again !
			}
		}
	}
	Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = NULL;

	// if we aren't forced to use fallback case, then use normal case
	if ( !bFallbackAllShadows )
	{
		// upload any additional data required for shadows
		Asura_PS2_GSMRenderer::UploadExtraShadowData();

		u_long	aiTex0Data[MAX_SHADOWS_IN_ONE_PASS];		// storage for up to 5 shadow buffers at a time
		int		aiShadowBuffers[MAX_SHADOWS_IN_ONE_PASS];	// storage for which 5 are chosen too
		u_short ausShadowVisFlags[MAX_SHADOWS_IN_ONE_PASS];	// storage for visibility flags for chosen 5
		const Asura_Bounding_Box* apxShadowBoxes[ASURA_PS2_SHADOW_MAX_BUFFERS];

		// get all the shadow buffer boxes sorted at once to save repeatedly checking them
		// also send matrices for all the shadow buffers, to be held on SPR until needed
		for (int i=0; i < ASURA_PS2_SHADOW_MAX_BUFFERS; i++)
		{
			if (!s_apxShadowBuffers[i]->IsUnused())
			{
				apxShadowBoxes[i] = &s_apxShadowBuffers[i]->GetWorldSpaceShadowBox();
				Asura_PS2_GSMRenderer::SetProjectionMatrix(i, s_apxShadowBuffers[i]->GetShadowSpaceTransform());
			}
		}

		Asura_Assert(s_paiModuleChunkCounts, "Module index table not built - must add a call to Traverse_ModuleData");

		// run through all modules which we've listed as being intersected by shadows and render them
		for(u_int u=0; u < MAXIMUM_NUMBER_OF_MODULES; ++u)
		{
			int iModule = s_axModuleControls[u].iModuleIndex ;

			if ((dbg_iOnlyShadowThisModule != -1) && (iModule != dbg_iOnlyShadowThisModule)) continue;

			if (iModule != -1)
			{
				// use bounding box of module to limit area which is tested for visibility
				Asura_Bounding_Box xModuleBox = Asura_Module_System::GetBoundingBoxOfModule(iModule);
				xModuleBox.Displace(Asura_Module_System::GetPositionOfModule(iModule));
				
				// get flags to show which shadow buffers are associated with this module
				u_short uFlags = s_axModuleControls[u].usShadowBufferFlags;

#ifdef _DEBUG
				if (dbg_iOnlyUseThisShadow != -1)
				{
					uFlags &= (1<<dbg_iOnlyUseThisShadow);
				}
#endif
				while ( uFlags != 0) // while still shadows to render
				{
					int iShadows = 0;

					// collect up as many shadow buffers as we can for a single pass
					for (int i=0; (i < ASURA_PS2_SHADOW_MAX_BUFFERS) && (iShadows < MAX_SHADOWS_IN_ONE_PASS) && (uFlags != 0); i++)
					{
						if ( (s_apxShadowBuffers[i]->IsUnused()) || (s_apxShadowBuffers[i]->m_bForceFallbackRender) )
						{
							uFlags &= ~(1<<i);	// clear flag - helps speed up process
							continue;	// skip over any which have already been through the fallback process or were never used anyway
						}
						if (uFlags & (1<<i))	// if this shadow is supposed to be attached to this module
						{
							uFlags &= ~(1<<i);	// remove current flag - we only want to check it once !

							Asura_Bounding_Box xIntersection = xModuleBox.Intersection(*apxShadowBoxes[i]);

							if (xIntersection.MaxX == xIntersection.MinX) 
							{
								continue;			// abandon ship if intersection is nowt - although this shouldn't be possible !
							}

							Asura_Vector_3 xIntersectionCentre = xIntersection.CalculateCentre();
							float fIntersectionRadius = xIntersection.CalculateRadius();
							int iCameras = Asura_Camera::GetNumberOfCameras();

							// determine if this intersection is visible from ANY camera, and if so which ones
							u_int uClipVisFlags = 0;
							u_int uNumActiveCams = 0;
							for (int iCam=0; iCam < iCameras; iCam++)
							{
								Asura_Camera::SetCameraIndex(iCam);

								if ( Asura_Camera::IsSphereInFrustum(xIntersectionCentre, fIntersectionRadius) )
								{
									if (!s_apxShadowBuffers[i]->m_pxAssociatedAnim ||
										s_apxShadowBuffers[i]->m_pxAssociatedAnim->GetParentEntityGUID() != Asura_Camera::GetEntityToIgnore())
									{
										uClipVisFlags |= 2;
										uNumActiveCams++;

		//								if( !Asura_PS2_VU1::IsSphereEntirelyWithinGuardbandFrustum(xIntersectionCentre, fIntersectionRadius) )
										{
											uClipVisFlags |= 1;
										}
									}
								}

								uClipVisFlags <<= 2;
							}
							
							// move bits up so that camera zero is at bits 8 & 7
							uClipVisFlags <<= (2 * ( ASURA_MAX_NO_OF_CAMERAS - iCameras - 1) );

							if (uClipVisFlags == 0) 
							{
								continue;			// skip this shadow if can't be seen
							}
							
							// include flag for multiple camera to ensure renderstates are updated properly
							if (uNumActiveCams > 1)
							{
								uClipVisFlags |= ASURA_PS2_GSM__FLAG__VIS_FOR_MORE_THAN_ONE_CAM;
							}

							// need to send these vis flags to the VU, so they have to be stored somewhere
							ausShadowVisFlags[iShadows] = uClipVisFlags;

							// record which shadow buffer is next (for when we want to upload the matrices)
							aiShadowBuffers[iShadows] = i;

							// record the texture settings too
							s_apxShadowBuffers[i]->Proj_GetTEX1andTEX0(NULL, &aiTex0Data[iShadows]);

							// move on to next shadow
							iShadows++;
						}
					}

					// We now have up to five shadow buffers, their visibility and clip flags for all the cameras
					// and their texture settings, all of which need to go to VU1

					if (iShadows)
					{
						if (bReportingActivityThisFrame)
						{
							printf("Rendering module %d, with %d shadows (", iModule, iShadows);
							printf("%d", aiShadowBuffers[0]);
							for (int i=1; i<iShadows; i++)
							{
								printf(", %d", aiShadowBuffers[i]);
							}
							printf(")\n");
						}

						// set renderstates including tex0 for multiple shadow buffers
						Asura_PS2_GSMRenderer::SetShadowStates( iShadows, aiTex0Data, aiShadowBuffers, ausShadowVisFlags );

						// finally loop over list of chunks for this module:
						int i_chunkCount = s_paiModuleChunkCounts[iModule];

						for (int i=0; i < i_chunkCount; i++)
						{
							u_int callAddr = s_paxModuleIndices[iModule][i]->uChainCALLAddr;
							if (dbg_bLimitShadowsChunks)
							{
								int j;
								for (j=0; j<6; j++)
								{
									if (dbg_aiChunksToShadow[j] == i) break;
								}
								if (j==6) 
								{
									if (bReportingActivityThisFrame)
									{
										printf("Skipping chunk  %d (%x)\n", i, callAddr);
									}
									continue;
								}
							}
							if (bReportingActivityThisFrame)
							{
								printf("Rendering chunk %d (%x)\n", i, callAddr);
							}
							Asura_PS2_GSMRenderer::RenderStripChain( callAddr );

							// stats collection - need to see which shadow buffers are visible for which camera
							{
								u_int uNumTris = 0;
								u_int uNumTrisPerCam = s_paxModuleIndices[iModule][i]->usNumTris;

								for (int j=0; (j < iShadows); j++)
								{
									u_int uCamVisFlags = ausShadowVisFlags[j];

									for( u_int u = ASURA_MAX_NO_OF_CAMERAS; u; u--)
									{
										if( uCamVisFlags & (1 << ( 2 * u) - 1 ) ) { uNumTris += uNumTrisPerCam; }
									}
								}
								Asura_PS2_Render::UpdateTriStats( uNumTris, 1, Asura_PS2_Render::TRICOUNT_SHADOW);
							}
						}
					}
				}
			}
		}
	}

	Asura_PS2_GSMRenderer::End();

    ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__PROJ_SHADOW_MARKER );

	s_bStillCollectingShadowsThisFrame = false;

	if (dbg_bDisplayShadowBuffersAlpha)
	{
		Asura_PS2_ShadowBufferSystem::ShowUsedShadowBuffers(true);
	}

	// NEED TO FLUSH to prevent anything later from trashing the texture while we used it - eg screen clear etc.
    Asura_PS2_QW* pxQWForTag = Asura_PS2_DeferredRendering::GetVifBuffer( 1 );
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxQWForTag))->Set( 0,
                                                            Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                            Asura_PS2_DmaTag::ID_CNT,
                                                            false,
                                                            0,
                                                            false,
                                                            SCE_VIF1_SET_NOP(false),
                                                            SCE_VIF1_SET_FLUSH(false) );

}


void Asura_PS2_ShadowBufferSystem::ShowUsedShadowBuffers(bool bAlpha)
{
	// single quad word to flush anything just in case
    (reinterpret_cast<Asura_PS2_DmaTag*>(Asura_PS2_DeferredRendering::GetVifBuffer(1)))->Set( 0,
                                                                                              Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                                              Asura_PS2_DmaTag::ID_CNT,
                                                                                              false,
                                                                                              0,
                                                                                              false, 
                                                                                              SCE_VIF1_SET_NOP(false),
                                                                                              SCE_VIF1_SET_FLUSH(false) );

	    // Renderstate.
    {
        Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1 );

        Asura_PlatformRenderStates::SetPixelTest( false, false );
        Asura_PlatformRenderStates::SetZBufferWriteEnabled( true );
        Asura_PlatformRenderStates::SetTextureClamp( true );

        if( bAlpha )
        {
            // See comment by TEX0 setup below.
            Asura_PlatformRenderStates::SetAlphaBlend( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_ZERO, 0 );
        }

        Asura_PlatformRenderStates::ApplyRenderStates();
    }

	u_int uGap = 20;
	u_int uRealX=uGap;
	u_int uRealY=uGap;
	u_int uW=128;
	u_int uH=128;

	for (int i=0; i < ASURA_PS2_SHADOW_MAX_BUFFERS; i++)
	{
		if (s_apxShadowBuffers[i]->IsUnused()) continue;	// skip unused ones

		u_int uX = uRealX;
		u_int uY = uRealY;

		const u_int uNumRegs  = 7;
		const u_int uQWC      = 1 + uNumRegs;
		const u_int uTotalQWs = 1 + uQWC;

		Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
		Asura_PS2_QW* pxOut = pxBase;

		// DMA tag.
		(reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWC,
															 Asura_PS2_DmaTag::PCE_NOCHANGE,
															 Asura_PS2_DmaTag::ID_CNT,
															 false,
															 0,
															 false,
															 SCE_VIF1_SET_NOP(false),
															 SCE_VIF1_SET_DIRECT(uQWC, false) );

		// GIF tag.
		{
			const u_int uPRIM = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
												 0,                  // IIP
												 true,               // TME
												 false,              // FGE
												 bAlpha,             // ABE
												 false,              // AA1
												 1,                  // FST
												 0,                  // CTXT
												 0 );                // FIX

			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( 1,                 // NLOOP
																 true,              // EOP
																 true,              // PRE
																 uPRIM,             // PRIM
																 SCE_GIF_PACKED,    // FLG
																 uNumRegs,          // NREG
																 SCE_GIF_PACKED_AD, // REG0
																 SCE_GIF_PACKED_AD, // REG1
																 SCE_GIF_PACKED_AD, // REG2
																 SCE_GS_UV,         // REG3
																 SCE_GS_XYZ2,       // REG4
																 SCE_GS_UV,         // REG5
																 SCE_GS_XYZ2 );     // REG6
		}

		// TEX1, TEX0.
		{
			s_apxShadowBuffers[i]->Proj_GetTEX1andTEX0( &((pxOut + 0)->u64s[0]), &((pxOut + 1)->u64s[0]) );

			if( bAlpha )
			{
				// Change TFX to HIGHLIGHT2. That way, Rs=Rt*Rf+Af (similarly for
				// G and B) and As=At. If we then set Rf=Gf=Bf=0 and Af=128, we
				// get Rs=Gs=Bs=128 and As=At, meaning we get Rd=Gd=Bd=At when
				// using the alpha blend setup up above.
				(reinterpret_cast<sceGsTex0&>((pxOut + 1)->u64s[0])).TFX = SCE_GS_HIGHLIGHT2;
			}

			(pxOut++)->u64s[1] = SCE_GS_TEX1_1;
			(pxOut++)->u64s[1] = SCE_GS_TEX0_1;
		}

		// RGBA. See comment by TEX0 setup above.
		pxOut->u64s[0] = ( bAlpha ? SCE_GS_SET_RGBAQ(0, 0, 0, 128, 0x3F800000) : SCE_GS_SET_RGBAQ(128, 128, 128, 128, 0x3F800000) );
		pxOut->u64s[1] = SCE_GS_RGBAQ;
		++pxOut;

		// Transform into primitive coordinate system.
		uX += ASURA_PS2_SCREEN_CENTRE_X - (Asura_PS2_GS::GetScreenWidth()  >> 1);
		uY += ASURA_PS2_SCREEN_CENTRE_Y - (Asura_PS2_GS::GetScreenHeight() >> 1);

		// TL.
		pxOut->u32s[0] = 0 << 4;
		pxOut->u32s[1] = 0 << 4;
		++pxOut;
		pxOut->u32s[0] = uX << 4;
		pxOut->u32s[1] = uY << 4;
		pxOut->u64s[1] = (1ul << Asura_PS2_GS::GetZBfrBitDepth()) - 1;
		++pxOut;

		// BR.
		pxOut->u32s[0] = s_apxShadowBuffers[i]->m_uBufferWidth << 4;
		pxOut->u32s[1] = s_apxShadowBuffers[i]->m_uBufferHeight << 4;
		++pxOut;
		pxOut->u32s[0] = (uX + uW) << 4;
		pxOut->u32s[1] = (uY + uH) << 4;
		pxOut->u64s[1] = (1ul << Asura_PS2_GS::GetZBfrBitDepth()) - 1;
		++pxOut;

		Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );

		uRealX += uW + uGap;
		if (uRealX + uW > 512)
		{
			uRealX = uGap;
			uRealY += uH + uGap;
		}
	}

}

//#############################################################################
// Code for handling curious PS2 module system
//#############################################################################
	const Asura_PS2_Env_RuntimeData_ModuleHdr*** Asura_PS2_ShadowBufferSystem::s_paxModuleIndices = NULL;
	int *Asura_PS2_ShadowBufferSystem::s_paiModuleChunkCounts = NULL;
	int Asura_PS2_ShadowBufferSystem::s_moduleCount = 0;

void Asura_PS2_ShadowBufferSystem::Traverse_ModuleData( void )
{
    Asura_Assert( !s_moduleCount, "Module data already pre-processed. This will result in a memory leak." );

	// find out how many modules we have
	s_moduleCount = Asura_Module_System::GetNumberOfModules();

	int *paiModuleChunkFillCount; // for counting entries into each buffer - just in case we overflow

	// allocate space for module index pointer list, and a temporary counter list
	s_paxModuleIndices = new const Asura_PS2_Env_RuntimeData_ModuleHdr** [s_moduleCount];
	s_paiModuleChunkCounts = new int[s_moduleCount];
	paiModuleChunkFillCount = new int[s_moduleCount];

    // (MP - 07.02.2005) Disregard hideously high poly modules.
    // Hopefully will be able to remove this again later...
    u_int* puTrisPerModule = new u_int[s_moduleCount];

	// run through counting arrays and initialise to zero !
	for (int i=0; i < s_moduleCount; i++)
	{
		s_paxModuleIndices[i] = NULL;
		s_paiModuleChunkCounts[i] = 0;
		paiModuleChunkFillCount[i] = 0;
        puTrisPerModule[i] = 0;
	}

	// run through all materials to count number of chunks for each module
    for( Asura_PS2_Env_Store::ConstIt xIt; !xIt.IsFinished(); xIt.Next() )
    {
        const Asura_PS2_Env_RuntimeData_MaterialHdr& xCurMaterialHdr = xIt.GetCurMaterialHdr();


		if (Asura_Material_System::IsMaterialTransparent( xCurMaterialHdr.iMaterialID ) )
		{
			continue; // skip transparent materials entirely, since casting onto them causes problems
		}

        const Asura_PS2_Env_RuntimeData_ModuleHdr*   pxCurModuleHdr  = xIt.GetModuleHdrsForCurMaterial();

        // Count each module associated with this material against the relevant module
        int iNumLeft = xCurMaterialHdr.uNumModules;
        for( ; iNumLeft--; ++pxCurModuleHdr )
        {
            const int iModuleIdx = pxCurModuleHdr->iModuleIdx;

			if (dbg_bDisableSceneryModules)
			{
				if(Asura_Environment_System::GetEnvironment()->IsSceneryModule(iModuleIdx))
				{
					continue;
				}
			}

            s_paiModuleChunkCounts[iModuleIdx]++;

            puTrisPerModule[iModuleIdx] += pxCurModuleHdr->usNumTris;
        }
    }

	// now allocate space for every modules little chunk of data
	for (int i=0; i < s_moduleCount; i++)
	{
#if 0
        // (MP - 07.02.2005) Disregard hideously high poly modules.
        // Hopefully will be able to remove this again later...
        if( puTrisPerModule[i] > 3000 )
        {
#if _DEBUG
            Asura_Assertf( false, ("Module %i has %u tris - ignoring for shadows.", i, puTrisPerModule[i]) );
#endif
            s_paiModuleChunkCounts[i] = 0;
        }
#endif

//		Asura_Assert( s_paiModuleChunkCounts[i] , "Found a module with no data - not expected result!");
		if (s_paiModuleChunkCounts[i])
		{
			s_paxModuleIndices[i] = new const Asura_PS2_Env_RuntimeData_ModuleHdr* [s_paiModuleChunkCounts[i]];

		}
		else
		{
			s_paxModuleIndices[i] = NULL;
		}
	}

	// now run through again, populating arrays
    for( Asura_PS2_Env_Store::ConstIt xIt; !xIt.IsFinished(); xIt.Next() )
    {
        const Asura_PS2_Env_RuntimeData_MaterialHdr& xCurMaterialHdr = xIt.GetCurMaterialHdr();

		if (Asura_Material_System::IsMaterialTransparent( xCurMaterialHdr.iMaterialID ) )
		{
			continue; // skip the same ones we skipped before, otherwise all hell will break loose
		}

        const Asura_PS2_Env_RuntimeData_ModuleHdr*   pxCurModuleHdr  = xIt.GetModuleHdrsForCurMaterial();

        // Count each module associated with this material against the relevant module
        int iNumLeft = xCurMaterialHdr.uNumModules;
        for( ; iNumLeft--; ++pxCurModuleHdr )
        {
            const int iModuleIdx = pxCurModuleHdr->iModuleIdx;

            if( !s_paiModuleChunkCounts[iModuleIdx] ) { continue; }

			Asura_Assert( paiModuleChunkFillCount[iModuleIdx] <		// number entered so far
				          s_paiModuleChunkCounts[iModuleIdx],         // total expected for this module
						  "shadow module data overflowed array");
            s_paxModuleIndices[iModuleIdx][paiModuleChunkFillCount[iModuleIdx]++] = pxCurModuleHdr;
        }
    }

    delete[] puTrisPerModule;
	delete[] paiModuleChunkFillCount;	
}

void Asura_PS2_ShadowBufferSystem::Destroy_ModuleData( void )
{
	// free allocated space for each module
	for (int i=0; i < s_moduleCount; i++)
	{
		//Asura_Assert( s_paxModuleIndices[i] , "Missing module data");
		if (s_paxModuleIndices[i])
		{
			delete[] s_paxModuleIndices[i];
			s_paxModuleIndices[i] = NULL;
		}
	}

	// finally free the list of items
	if( s_paxModuleIndices )
	{
		delete[] s_paxModuleIndices;
		s_paxModuleIndices = NULL;
	}
	if ( s_paiModuleChunkCounts )
	{
		delete[] s_paiModuleChunkCounts;	// and delete the counters
		s_paiModuleChunkCounts = NULL;
	}
	s_moduleCount = 0;
}

//****************************************************************************
// Asura_ShadowBuffer_System implementation
//****************************************************************************

void Asura_ShadowBuffer_System::PreAllocateBuffer(const Asura_Animation* pxAnim, BUFFERSIZE eBufferSize, float fScale)
{
	// implement your own platform size conversion here :
	u_int u_bufferSize;

	switch (eBufferSize)
	{
	case SIZE_LARGE:
		u_bufferSize = 256;
		break;
	case SIZE_DEFAULT:
	default:
		u_bufferSize = Asura_PS2_ShadowBuffer::DEFAULT_WIDTH;
		break;
	}

	// implement youw own platform pre-allocation function call here (if you have one - doing nothing is safe too)
	Asura_PS2_ShadowBufferSystem::PreAllocate(pxAnim, u_bufferSize, fScale);
}

void Asura_ShadowBuffer_System::PreProcessLevelData(void)
{
	Asura_PS2_ShadowBufferSystem::Traverse_ModuleData();
}
