//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformRenderObject.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PlatformRenderObject.h"

// Stdlib platform-specific.
#include <libgraph.h>

// Engine independent.
#include "Core/Asura_Material_System.h"
#include "Core/Asura_String.h"
#include "Core/Asura_TextureManagement.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Light/Asura_GI_Solution.h"
#include "Shapes/Asura_Shapes.h"

// Engine platform-specific.
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS2_StrippedObject.h"
#include "Debugging/Asura_PS2_DebugInfo.h"
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_GSMDefines.h"
#include "Rendering/Asura_PS2_GSMRenderer.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"
#include "Rendering/Asura_PS2_Shadows.h"
#include "Rendering/Asura_PS2_ShadowBuffer.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#define OBJ_MMAG (dbg_bObjectsBilinear ? SCE_GS_LINEAR : SCE_GS_NEAREST)
#define OBJ_MMIN (dbg_bObjectsTrilinear ? SCE_GS_LINEAR_MIPMAP_LINEAR : (dbg_bObjectsBilinear ? SCE_GS_LINEAR_MIPMAP_NEAREST : SCE_GS_NEAREST_MIPMAP_NEAREST))

// G L O B A L S /////////////////////////////////////////////////////////////

DEBUGVAR bool dbg_bObjectsBilinear  = true;
DEBUGVAR bool dbg_bObjectsTrilinear = false;
DEBUGVAR bool dbg_bDrawObjectBBs    = false;

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_PS2_StrippedObject::Asura_PS2_StrippedObject( const char* szName )
{   
#if _DEBUG
    m_szName     = Asura_StringUtil::Duplicate( szName );
#endif   
    m_pvDMAChain = NULL;
}

Asura_PS2_StrippedObject::~Asura_PS2_StrippedObject()
{
    Asura_DeallocateAlign( m_pvDMAChain );
#if _DEBUG
    delete[] m_szName;
#endif
}

bool Asura_PS2_StrippedObject::LoadStrippedObject( Asura_Handle iFileHandle, u_int uNumBytes, const Asura_PS2_Shape_ShapeHdr& xShapeHdr )
{
    // Currently only supporting a single material on objects so the one
    // we're interested in will just be the first in the conversion list.
    m_iMaterialID = Asura_Material_System::ConvertIndex( 0 );

    m_uNumTris = xShapeHdr.usTotalTris;

    if( uNumBytes )
    {
        Asura_Assert( !(uNumBytes & 15), "" );
        m_pvDMAChain = Asura_AllocateAlign( uNumBytes, ASURA_ALIGN_128_BYTE );

        return( m_pvDMAChain && Asura_FileAccess::ReadData(iFileHandle, m_pvDMAChain, uNumBytes) );
    }
    else
    {
        Asura_Assert( !m_uNumTris, "" );
        return true;
    }
}

#if _DEBUG
// For all once-per-frame stuff.
#include "Core/Asura_Timers.h"

DEBUGVAR u_int dbg_uNumObjectsToSkipPerFrame = 0;
DEBUGVAR u_int dbg_uMaxObjectsToDrawPerFrame = 0xFFFFFFFFu;

#include "Core/Asura_UserInput.h"
#include "Input/Asura_PlatformInput.h"
DEBUGVAR bool dbg_bEnableObjOriAndPosManipulation = false;
static Asura_Euler_3 g_xObjManipEuler;
static Asura_Vector_3 g_xObjManipPos;
#endif

void Asura_PS2_StrippedObject::Render( const Asura_Shape_Instance* pxShapeInst, u_int uObjectARGB, u_int uCamVisFlags ) const
{
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DRAW_OBJS) || !m_pvDMAChain )
    {
        return;
    }

#if  _DEBUG
    {
        static u_int uTimestamp = 0xFFFFFFFFu;
        static u_int uNumObjsThisFrame;
        static u_int uNumObjsDrawnThisFrame;

        const u_int uCurFrame = Asura_Timers::GetFrameNumber();
        bool bFirstObjThisFrame = ( uCurFrame != uTimestamp );

        if( bFirstObjThisFrame )
        {
            uNumObjsThisFrame      = 0;
            uNumObjsDrawnThisFrame = 0;
        }

        uTimestamp = uCurFrame;

        if( dbg_bEnableObjOriAndPosManipulation && bFirstObjThisFrame )
        {
            const float fFrameTime = Asura_Timers::GetFrameTime();

            g_xObjManipEuler.x += Asura_UserInput::GetAxisValue(ASURA_AXIS_PAD2_RSTICK_H) * fFrameTime;
            g_xObjManipEuler.z += Asura_UserInput::GetAxisValue(ASURA_AXIS_PAD2_RSTICK_V) * fFrameTime;
            if( g_xObjManipEuler.x < Asura_Maths::Pi ) { g_xObjManipEuler.x += 2.0f * Asura_Maths::Pi; }
            if( g_xObjManipEuler.x > Asura_Maths::Pi ) { g_xObjManipEuler.x -= 2.0f * Asura_Maths::Pi; }
            if( g_xObjManipEuler.z < Asura_Maths::Pi ) { g_xObjManipEuler.z += 2.0f * Asura_Maths::Pi; }
            if( g_xObjManipEuler.z > Asura_Maths::Pi ) { g_xObjManipEuler.z -= 2.0f * Asura_Maths::Pi; }

            g_xObjManipPos.y -= Asura_UserInput::GetAxisValue(ASURA_AXIS_PAD2_LSTICK_V) * fFrameTime;
            g_xObjManipPos.y = Asura_Maths::Sign(g_xObjManipPos.y) * Asura_Maths::Min(3.0f, Asura_Maths::Abs(g_xObjManipPos.y));
        }

        if( uNumObjsThisFrame++ < dbg_uNumObjectsToSkipPerFrame ) { return; }
        
        if( uNumObjsDrawnThisFrame++ >= dbg_uMaxObjectsToDrawPerFrame ) { return; }
    }
#endif

    ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__OBJECT_MARKER );

    u_int uMaterialFlags = Asura_Material_System::GetMaterialFlags( m_iMaterialID );

    // NB: Cast to int before casting to float to prevent software operation.
    const float fAlpha = static_cast<int>((uObjectARGB & 0xFF000000u) >> 24u) / 255.0f;
    if( fAlpha < 1.0f )
    {
        uMaterialFlags |= AsuraMaterialFlag_Transparent;
    }

    const bool  bReflective = Asura_PS2_GSMRenderer::IsMaterialReflective( uMaterialFlags );
    const u_int uSIFlags    = pxShapeInst->GetFlags();
    bool  bShadow     = !(uSIFlags & ASURA_SHAPE_INSTANCE_NO_SHADOW) && Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_OBJ_SHADOWS);

    Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL, true, 0, 1 );

    Asura_PS2_GSMRenderer::Begin();

    Asura_PS2_GSMRenderer::SetTexture( Asura_Material_System::GetTextureID(m_iMaterialID), OBJ_MMAG, OBJ_MMIN, Asura_PS2_TextureManagement::BFR_AUTO );

    if( bReflective )
    {
        Asura_PS2_GSMRenderer::SetTexture( Asura_Render::GetReflectanceMap(), OBJ_MMAG, OBJ_MMIN, 2, false );
    }

    Asura_PS2_GSMRenderer::SetRenderstate( uMaterialFlags );

    const Asura_Bounding_Box& xBB = pxShapeInst->GetBB();
    const Asura_Vector_3 xBSCentre = xBB.CalculateCentre();
    const float fBSRadius = 0.5f * xBB.CalculateExtents().Magnitude();

    Asura_PS2_GSMRenderer::SetCamFlags( uCamVisFlags, xBSCentre, fBSRadius );

	float fShadowAlpha = 0.0f;

    if( bShadow )
	{
		float fMinDistFromCam;	// not sure if needed ?
		fShadowAlpha = Asura_PS2_Shadows::CalcAlphaForMinDistFromCam( xBSCentre, fBSRadius, uCamVisFlags, &fMinDistFromCam );
		if( fShadowAlpha <= 0.0f )
		{
			bShadow = false;
		}
	}

    if( bShadow )
    {
		Asura_PS2_ShadowBuffer* pxShadowBuffer = Asura_PS2_ShadowBufferSystem::GetTemporaryShadowBuffer( xBSCentre, fBSRadius, uCamVisFlags );
		Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = pxShadowBuffer;

#ifdef _DEBUG
		extern bool bReportingActivityThisFrame;
		if (bReportingActivityThisFrame )
		{
			printf("%s\n", m_szName);
		}
#endif

		if (pxShadowBuffer)
		{
			Asura_Vector_3 xShadowPos( xBSCentre.x, xBB.MaxY, xBSCentre.z );
	#if _DEBUG
			xShadowPos += g_xObjManipPos;
	#endif

//			if( uSIFlags & ASURA_SHAPE_INSTANCE_MAY_NOT_BE_ON_GROUND )
			{
				if (pxShadowBuffer->IsForcedFallback())
				{
					float fHeight = 0.0f;
					Asura_PS2_Shadows::Proj_GetShadowYAndHeightForPos( xShadowPos, &xShadowPos.y, &fHeight );
					pxShadowBuffer->SetShadowBufferQuadHeight( xShadowPos.y );
				}
			}

			Asura_PS2_GSMRenderer::EnableShadow( xBSCentre, fBSRadius/*, xShadowPos, fHeight */);
			pxShadowBuffer->JustSetupRenderToBfr(SCE_VIF1_SET_FLUSH(false));
		}
    }

#if _DEBUG
    if( dbg_bEnableObjOriAndPosManipulation ) //TEMP
    {
        Asura_Matrix_3x3 xOri;
        g_xObjManipEuler.ToMatrix( xOri );
        xOri = pxShapeInst->GetOrientation() * xOri;

        Asura_Vector_3 xPos = pxShapeInst->GetPosition() + g_xObjManipPos;

        Asura_PS2_GSMRenderer::SetObjectToWorldMat( xOri, xPos );
    }
    else
#endif
    {
        Asura_PS2_GSMRenderer::SetObjectToWorldMat( pxShapeInst->GetOrientation(), pxShapeInst->GetPosition() );
    }

	pxShapeInst->UpdateLighting();
    Asura_PS2_GSMRenderer::UseExistingSHForMesh( Asura_GI_Solution::s_xSHForCurrentRender, fAlpha );

    Asura_PS2_GSMRenderer::RenderStripChain( reinterpret_cast<const u_int&>(m_pvDMAChain) );

    Asura_PS2_Render::UpdateTriStats( m_uNumTris * Asura_PS2_GSMRenderer::GetNumCamsVisFor(),
                                      Asura_PS2_GSMRenderer::GetNumPassesPerTri(),
                                      Asura_PS2_Render::TRICOUNT_OBJ );

    Asura_PS2_GSMRenderer::End();
    
	if (Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer)
	{
//		Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer->ReleaseShadowBuffer();	don't release it, keep it so it can be rendered later !
		Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = NULL;
	}

    ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__OBJECT_MARKER );

#ifndef ASURA_PS2_SONY_SUBMISSION
    if( dbg_bDrawObjectBBs ) { Asura_Render::RenderBB( xBB, Asura_Maths::ZeroVector3, Asura_Maths::IdentityMatrix3x3 ); }
#endif
}

