/////////////////////////////////////////////////////////////////////////////////////////
//
//	FILE		ASURA_PlatformRenderStates.CPP
//
//	CREATOR		Kevin Lea (27 March 2001)
//
//	MODIFIED	Richard M. Smith (5 September 2001)
//					Added Z Buffer Test Enable/Disable (Using ScratchPad)
//
/////////////////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PlatformRenderStates.h"

// Engine independent.
#include "Debugging/Asura_Assert.h"                   // For Asura_Assert().
#include "Debugging/Asura_Logfile.h"                  // Asura_OutputToDebugger()
#include "Rendering/Asura_Render.h"                   // For Asura_Render class.
#include "Rendering/Asura_RenderStates.h"

// Engine platform-specific.
#ifdef ASURA_PS2_DEFERRED_RENDERING
#include "Rendering/Asura_PS2_DeferredRendering.h"
#endif
#include "Core/Asura_PS2_DMA.h"                  // For Asura_PS2_DMAFromSPR().
#include "Rendering/Asura_PS2_GS.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static Asura_PS2_QW* g_pxRenderstatePacket = NULL;
static Asura_PS2_QW* g_pxRealRSPacket      = NULL;

static bool g_bZBufferWriteHasBeenSetExplicitly;

// C L A S S E S /////////////////////////////////////////////////////////////

//
// Asura_Render_States.
//

void Asura_Render_States::SetAlphaBlending( bool ) { }
void Asura_Render_States::SetAlphaTest( bool ) { }
void Asura_Render_States::SetCullMode( ASURA_CULLMODE_ID ) { }
void Asura_Render_States::SetDitheringMode( bool ) { }
void Asura_Render_States::SetDynamicLights( bool ) { }
void Asura_Render_States::SetTextureIndex( int, int ) { }
void Asura_Render_States::SetTextureOpMode( ASURA_TEXTUREOP_ID ) { }
void Asura_Render_States::SetTranslucencyMode( ASURA_TRANSLUCENCY_MODE ) { }
void Asura_Render_States::SetWireFrameMode( bool ) { }
void Asura_Render_States::SetZBufferCompareMode( ASURA_ZBUFFERCMP_ID ) { }
void Asura_Render_States::SetZBufferEnabled( bool ) { }
void Asura_Render_States::SetZBufferWriteEnabled( bool ) { }
void Asura_Render_States::SetStencilEnabled( bool ) { }

//
// Asura_PlatformRenderStates.
//

void Asura_PlatformRenderStates::SetAlphaBlend( int iA, int iB, int iC, int iD, int iFIX )
{
    Asura_Assert( g_pxRenderstatePacket, "" );

    const u_long uNewValue = SCE_GS_SET_ALPHA( iA, iB, iC, iD, iFIX );
    Asura_Assert( uNewValue != g_pxRenderstatePacket[QWID_ALPHA].u64s[0], "" );
    g_pxRenderstatePacket[QWID_ALPHA].u64s[0] = uNewValue;
}

#if _DEBUG
bool Asura_PlatformRenderStates::IsTranslucencyModeSupported( ASURA_TRANSLUCENCY_MODE eMode )
{
    switch( eMode )
	{
	 	case ASURA_TRANSLUCENCY_NORMAL:
  		case ASURA_TRANSLUCENCY_GLOWING:
  		case ASURA_TRANSLUCENCY_INVCOLOUR:
        case ASURA_TRANSLUCENCY_OFF:
            return true;

        default:
            break;
    }

    return false;
}
#endif

void Asura_PlatformRenderStates::SetTranslucencyMode( ASURA_TRANSLUCENCY_MODE eMode )
{
    Asura_Assert( g_pxRenderstatePacket, "" );

    switch( eMode )
	{
		default:
            Asura_Assert( false, "Not supported on PS2.");
            // Fall through to NORMAL case, I guess.

	 	case ASURA_TRANSLUCENCY_NORMAL:
            SetAlphaBlend( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
			break;

        case ASURA_TRANSLUCENCY_GLOWING:
            SetAlphaBlend( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
			break;

        case ASURA_TRANSLUCENCY_INVCOLOUR:
            SetAlphaBlend( SCE_GS_ALPHA_CD, SCE_GS_ALPHA_CS, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_ZERO, 0x80 );
			break;

        case ASURA_TRANSLUCENCY_OFF:
            SetAlphaBlend( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_FIX, SCE_GS_ALPHA_CS, 0 );
            break;
	}

    g_pxRenderstatePacket[QWID_TEST].u64s[0] = SCE_GS_SET_TEST_1( true,                  // ATE 
                                                                  SCE_GS_ALPHA_GREATER,  // ATST
                                                                  0x70,                  // AREF
                                                                  SCE_GS_AFAIL_FB_ONLY,  // AFAIL
                                                                  false,                 // DATE
                                                                  0,	                   // DATM
                                                                  true,                  // ZTE
                                                                  SCE_GS_DEPTH_GEQUAL ); // ZTST
}

void Asura_PlatformRenderStates::SetDefaultAlphaValue( bool bTreatBlackAsTransparent, uint32 uTA0, uint32 uTA1 )
{
    Asura_Assert( g_pxRenderstatePacket, "" );

    const u_long uNewValue = SCE_GS_SET_TEXA( uTA0, bTreatBlackAsTransparent, uTA1 );
    Asura_Assert( uNewValue != g_pxRenderstatePacket[QWID_TEXA].u64s[0], "" );
    g_pxRenderstatePacket[QWID_TEXA].u64s[0] = uNewValue;
}

void Asura_PlatformRenderStates::SetPixelTest( bool bEnableAlphaTest, 
                                               bool bEnableDepthTest, 
                                               int	iAlphaLimit,
                                               int	iAlphaMethod,
                                               int	iAlphaFail )
{
    Asura_Assert( g_pxRenderstatePacket, "" );

    const u_int uDepthTestMethod = ( bEnableDepthTest ?  SCE_GS_ZGEQUAL : SCE_GS_ZALWAYS );
    const u_long uNewValue = SCE_GS_SET_TEST_1( bEnableAlphaTest,	// ATE 
                                                iAlphaMethod,		// ATST
                                                iAlphaLimit,        // AREF
                                                iAlphaFail,			// AFAIL
                                                false,              // DATE
                                                0,					// DATM
                                                true,				// ZTE
                                                uDepthTestMethod );	// ZTST
    
    Asura_Assert( uNewValue != g_pxRenderstatePacket[QWID_TEST].u64s[0], "" );
    g_pxRenderstatePacket[QWID_TEST].u64s[0] = uNewValue;

    if( !g_bZBufferWriteHasBeenSetExplicitly )
    {
        // If they haven't made an explicit request, set ZMSK according to
        // depth test, ie. don't update the z buffer is testing it has been
        // disabled.
        g_pxRenderstatePacket[QWID_ZBUF].u32s[1] = ( bEnableDepthTest ? 0 : 1 );
    }
}

void Asura_PlatformRenderStates::SetTextureClamp( bool bClamp )
{
    Asura_Assert( g_pxRenderstatePacket, "" );

    const u_long uNewValue = SCE_GS_SET_CLAMP( bClamp, bClamp, 0, 0, 0, 0 );
    Asura_Assert( uNewValue != g_pxRenderstatePacket[QWID_CLAMP].u64s[0], "" );
    g_pxRenderstatePacket[QWID_CLAMP].u64s[0] = uNewValue;
}


void Asura_PlatformRenderStates::SetZBufferWriteEnabled( bool bEnable )
{
    Asura_Assert( g_pxRenderstatePacket, "" );

    const u_long uNewValue = SCE_GS_SET_ZBUF( Asura_PS2_GS::GetZBfrBasePtr(), Asura_PS2_GS::GetZBfrPSM(), !bEnable );
    Asura_Assert( uNewValue != g_pxRenderstatePacket[QWID_ZBUF].u64s[0], "" );
    g_pxRenderstatePacket[QWID_ZBUF].u64s[0] = uNewValue;

    g_bZBufferWriteHasBeenSetExplicitly = true;
}

void Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( CONTEXT eContextID,
                                                                 Asura_PS2_QW* pxBuffer,
                                                                 bool bSetEOP, bool bRememberMyBuffer, bool bWriteGifTag )
{
    const u_int uQWs = ( bWriteGifTag ? NUM_RENDERSTATE_PACKET_QWS : NUM_RENDERSTATE_PACKET_QWS_EXCL_GIF_TAG );

    if( pxBuffer )
    {
		g_pxRealRSPacket = pxBuffer;
    }
    else
    {
		Asura_PS2_QW* pxOut = Asura_PS2_DeferredRendering::GetVifBuffer( 1 + uQWs );

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWs,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_CNT,
                                                             false,
                                                             0,
                                                             false,
                                                             SCE_VIF1_SET_NOP(false),
                                                             SCE_VIF1_SET_DIRECT(uQWs, false) );

		g_pxRealRSPacket = pxOut;
    }

#ifdef ASURA_ASSERT
    // Fill buffer with 0xFF to ensure checks on required settings are
    // accurate.
    memset( g_pxRealRSPacket, 0xFF, uQWs << 4 );
#endif

    g_pxRenderstatePacket = g_pxRealRSPacket;
    if( bWriteGifTag )
    {
        Asura_PS2_GifTag& xGIFTag = reinterpret_cast<Asura_PS2_GifTag&>( g_pxRenderstatePacket[QWID_TAG] );
        xGIFTag.SetExceptRegs( NUM_RENDERSTATE_PACKET_QWS_EXCL_GIF_TAG, bSetEOP, false, 0, SCE_GIF_PACKED, 1 );
        xGIFTag.uAllRegs = SCE_GIF_PACKED_AD;
    }
    else
    {
        --g_pxRenderstatePacket;
    }

    // Set register IDs.
    g_pxRenderstatePacket[QWID_ZBUF].u64s[1]  = SCE_GS_ZBUF_1 + eContextID;
    g_pxRenderstatePacket[QWID_TEXA].u64s[1]  = SCE_GS_TEXA;
    g_pxRenderstatePacket[QWID_ALPHA].u64s[1] = SCE_GS_ALPHA_1 + eContextID;
    g_pxRenderstatePacket[QWID_TEST].u64s[1]  = SCE_GS_TEST_1 + eContextID;
    g_pxRenderstatePacket[QWID_CLAMP].u64s[1] = SCE_GS_CLAMP_1 + eContextID;

    // Defaults.
    SetZBufferWriteEnabled( true );
    SetDefaultAlphaValue( false, 0, 128 );
    SetAlphaBlend( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_CD, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 );
    SetPixelTest( false, true );
    SetTextureClamp( false );

    g_bZBufferWriteHasBeenSetExplicitly = false;

    if( pxBuffer && !bRememberMyBuffer )
    {
        g_pxRealRSPacket = g_pxRenderstatePacket = NULL;
    }
}

void Asura_PlatformRenderStates::ApplyRenderStates()
{
    Asura_Assert( g_pxRenderstatePacket, "" );
    g_pxRealRSPacket = g_pxRenderstatePacket = NULL;
}

void Asura_PlatformRenderStates::ForgetMyBuffer( Asura_PS2_QW* pxBfr )
{
    Asura_Assert( g_pxRealRSPacket == pxBfr, "" );
    g_pxRealRSPacket = g_pxRenderstatePacket = NULL;
}
