//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformTextureManagement.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PlatformTextureManagement.h"

// Engine independent.
#include "Core/Asura_NameTree.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_ScreenShot_System.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_CommandConsole.h"
#ifdef ASURA_SUPPORT_FMV_TEXTURES
#include "Rendering/Asura_MovieToTexture_System.h"
#endif

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Core/Asura_PS2_OS.h"
#ifdef ASURA_SUPPORT_FMV_TEXTURES
#include "Rendering/Asura_Platform_MovieToTextureStream.h"
#endif
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_TIM2.h"

// JAT 29/06/06 - the texture management code relies on using allocated memory via uncached acces
// Need this define to maek sure any memory allocated is flushed to the cachce
#if !defined( ASURA_MEMORYMANAGEMENT_PROCESS_ALLOCATED_MEM )

#error ASURA_MEMORYMANAGEMENT_PROCESS_ALLOCATED_MEM needs to be defined for PS2 texture management to work properly

#endif // #if !defined( ASURA_MEMORYMANAGEMENT_PROCESS_ALLOCATED_MEM )


// T Y P E S /////////////////////////////////////////////////////////////////

struct Asura_PS2_16BitClutEntry
{
    u_short usR : 5;
    u_short usG : 5;
    u_short usB : 5;
    u_short usA : 1;
};

struct Asura_PS2_Tim2UserData
{
	int   iID;              // ('A')+('S'<<8)+('U'<<16)+('R'<<24)
	u_int uSwizzled :  1;
	u_int __pad     : 31;
};

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

class Asura_PS2_Texture;

// G L O B A L S /////////////////////////////////////////////////////////////

static Asura_PS2_Texture* g_apxTextures[ASURA_MAX_NO_OF_TEXTURES];

DEBUGVAR bool dbg_bReportFailedTextureLoads     = true;
DEBUGVAR bool dbg_bReportSuccessfulTextureLoads = false;
#if _DEBUG
DEBUGVAR bool dbg_bPrintLoadedTextureInfo       = false;
#endif

static ASURA_PS2_TEXBFRLAYOUT g_eCurTexBfrLayout = ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS;

static int g_aiTextureInVRAM[ASURA_PS2_MAX_TEXTURE_BUFFERS] = { -1, -1, -1, -1 };
DEBUGVAR bool dbg_bTexManagementCacheTexturesInternally = true;

// These are all reset/setup when the layout is selected, and of course we
// assume that will done before any textures are uploaded or used.
static u_int g_uLastUsedBuffer;
static bool  g_bAutoBfrSwitchingEnabled; // Generally should not be used directly; call ShouldUseAutoBfrSwitching() instead.
static u_int g_uAutoBfrSwitchingMin;     // Inclusive.
static u_int g_uAutoBfrSwitchingMax;     //     "
static u_int g_uAutoBfrLastUsed;

#if ASURA_PS2_TEXTUREMANAGER__PROFILING
u_int Asura_PS2_TextureManagement::s_uProfile_UploadedBytes    = 0;
u_int Asura_PS2_TextureManagement::s_uProfile_UploadedQuantity = 0;
u_int Asura_PS2_TextureManagement::s_uProfile_CacheHits        = 0;
#endif

extern u_short g_uMipMapK;

static bool g_bAlwaysUsePATH2 = false;
static bool g_bForcePATH2UploadNext = false;
static bool g_bLastTexWentAlongPATH2;

DEBUGVAR bool dbg_bOptimiseOutTEXFLUSHWherePossible = true;
DEBUGVAR bool dbg_bGloballyDisableMipmaps           = false;
DEBUGVAR bool dbg_bInvalidateTexBfrsAtEndOfFrame    = false;

DVSVAR bool cnsl_bUseMipMapKMoireFix = true;
DVSVAR u_int cnsl_uMipMapKMoireFixOffset = 16;

// C O N S T A N T S /////////////////////////////////////////////////////////

static const int ASURA_PS2_TIM2_USER_DATA_ID = ('A' << 0) + ('S' << 8) + ('U' << 16) + ('R' << 24);

static const u_int NUM_TEXTURE_BUFFERS[ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS] =
{
    2, // ASURA_PS2_TEXBFRLAYOUT_2_LARGE
    3, // ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL
    4, // ASURA_PS2_TEXBFRLAYOUT_4_SMALL
};

static const u_int g_aauTexBfrSizesIn64wUnits[ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS][ASURA_PS2_MAX_TEXTURE_BUFFERS] =
{
    // ASURA_PS2_TEXBFRLAYOUT_2_LARGE (2 x 512Kb)
    {
        (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 1),
        (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 1),
        0,
        0,
    },

    // ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL (2 x 384Kb + 1 x 256Kb)
    {
        (384 << 2),
        (384 << 2),
        (256 << 2),
        0,
    },

    // ASURA_PS2_TEXBFRLAYOUT_4_SMALL (4 x 256Kb)
    {
        (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2),
        (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2),
        (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2),
        (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2),
    },
};

static const u_int g_aauTexBfrBasePtrs[ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS][ASURA_PS2_MAX_TEXTURE_BUFFERS] =
{
    // ASURA_PS2_TEXBFRLAYOUT_2_LARGE (2 x 512Kb)
    {
        ASURA_PS2_TEXBFR_BASE_PTR,
        ASURA_PS2_TEXBFR_BASE_PTR + (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 1),
        0,
        0,
    },

    // ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL (2 x 384Kb + 1 x 256Kb)
    {
        ASURA_PS2_TEXBFR_BASE_PTR,
        ASURA_PS2_TEXBFR_BASE_PTR + ((384 << 2) << 0),
        ASURA_PS2_TEXBFR_BASE_PTR + ((384 << 2) << 1),
        0,
    },

    // ASURA_PS2_TEXBFRLAYOUT_4_SMALL (4 x 256Kb)
    {
        ASURA_PS2_TEXBFR_BASE_PTR,
        ASURA_PS2_TEXBFR_BASE_PTR + (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2) * 1,
        ASURA_PS2_TEXBFR_BASE_PTR + (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2) * 2,
        ASURA_PS2_TEXBFR_BASE_PTR + (ASURA_PS2_TEXBFR_NUM_64W_UNITS >> 2) * 3,
    },
};

// C L A S S E S /////////////////////////////////////////////////////////////

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static inline bool CanGSCalculateMipmapAddrs( const Asura_PS2_Texture* pxTex )
{
    return( 0 != pxTex->GetPicHdr()->xTex1.MTBA );
}

static inline bool CanGSCalculateMipmapAddrs( int iTexID )
{
	Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );
    return CanGSCalculateMipmapAddrs( g_apxTextures[iTexID] );
}

static u_char* CreateDummyTexture( void );

static inline u_int GetNextAutoSwitchBfr( void )
{
    Asura_Assert( g_bAutoBfrSwitchingEnabled, "" );

    u_int uBfr = g_uAutoBfrLastUsed + 1;
    if( uBfr < g_uAutoBfrSwitchingMin || uBfr > g_uAutoBfrSwitchingMax )
    {
        uBfr = g_uAutoBfrSwitchingMin;
    }

    return uBfr;
}

static u_int GetNumQWsPerNLines( const u_int uNumLines, const u_int uPSM, const u_int uWidth );

static bool PlatformLoadTexture( const Asura_String& xName, int iTexID );

#if _DEBUG
static void PrintLoadedTextureInfo( int iTexID );
#endif

u_int SearchInBfrsForTexture( int iTexID, u_int uSearchStart, u_int uSearchMin, u_int uSearchMax )
{
    Asura_Assert( ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS != g_eCurTexBfrLayout, "Texture buffer layout hasn't been set." );
    Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );

    const u_int uBfrCount = NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout];
    Asura_Assert( uSearchMin <= uSearchMax && uSearchMax < uBfrCount, "" );

    u_int u = uSearchStart;
    do
    {
        if( iTexID == g_aiTextureInVRAM[u] ) { return u; }

        if( ++u > uSearchMax ) { u = uSearchMin; }
    }
    while( u != uSearchStart );

    return Asura_PS2_TextureManagement::BFR_NONE;
}

u_int SearchInBfrsForTexture( int iTexID, u_int uSearchStart, bool bAutoBfrSwitching )
{
    u_int uSearchMin = ( bAutoBfrSwitching ? g_uAutoBfrSwitchingMin : uSearchStart );
    u_int uSearchMax = ( bAutoBfrSwitching ? g_uAutoBfrSwitchingMax : uSearchStart );

    return SearchInBfrsForTexture( iTexID, uSearchStart, uSearchMin, uSearchMax );
}

static inline bool ShouldUseAutoBfrSwitching( u_int uInBfr )
{
    if( Asura_PS2_TextureManagement::BFR_AUTO == uInBfr )
    {
        // Shouldn't be requesting auto bfr index selection if auto bfr
        // switching has been disabled.
        Asura_Assert( g_bAutoBfrSwitchingEnabled, "" );

        return true;
    }
    else
    {
        return false;
    }
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////


//
// Asura_PS2_Texture.
//

Asura_PS2_Texture::Asura_PS2_Texture( void )
{
    m_pucRawData = NULL;
    m_pxPicHdr   = NULL;

    for( u_int u = 0; u < ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS; ++u )
    {
        for( u_int v = 0; v < ASURA_PS2_MAX_TEXTURE_BUFFERS; ++v )
        {
            m_aapxDmaChains[u][v] = NULL;
        }
    }

    m_bIsSwizzled         = false;
    m_ucNumQWsForDMAChain = 0;
#if _DEBUG
    m_szName              = NULL;
    m_bIsPinkTexture      = false;
#endif
}

Asura_PS2_Texture::~Asura_PS2_Texture( void )
{
    Reset();

    if( m_pucRawData ) { Asura_DeallocateAlign( m_pucRawData ); }

#if _DEBUG
    if( m_szName ) { delete[] m_szName; }
#endif
}

u_long128* Asura_PS2_Texture::GetClutDataOpen( void ) const
{
    return( reinterpret_cast<u_long128*>(reinterpret_cast<u_char*>(m_pxPicHdr) + m_pxPicHdr->HeaderSize + m_pxPicHdr->ImageSize) );
}

const u_long128* Asura_PS2_Texture::GetClutData( void ) const
{
    return( reinterpret_cast<const u_long128*>(reinterpret_cast<const u_char*>(m_pxPicHdr) + m_pxPicHdr->HeaderSize + m_pxPicHdr->ImageSize) );
}

u_int Asura_PS2_Texture::GetClutPSM( void ) const
{
	Asura_Assert( m_pxPicHdr->ClutType != TIM2_NONE, "Can't get PSM - this texture hasn't got a CLUT." );

    switch( m_pxPicHdr->ClutType & ASURA_PS2_TIM2_CLUT_FORMAT_MASK )
    {
        case TIM2_RGB16:
    		return SCE_GS_PSMCT16;

        case TIM2_RGB24:
    		return SCE_GS_PSMCT24;

        case TIM2_RGB32:
            return SCE_GS_PSMCT32;
	}

    Asura_Assert( false, "Invalid or unsupported CLUT type." );
    return SCE_GS_PSMCT32;
}

const u_long128* Asura_PS2_Texture::GetData( void ) const
{
    return( reinterpret_cast<const u_long128*>(reinterpret_cast<const char*>(m_pxPicHdr) + m_pxPicHdr->HeaderSize) );
}

bool Asura_PS2_Texture::HasClut( void ) const
{
    if( m_bIsSwizzled )
    {
        return true;
    }
    else
    {
        switch( m_pxPicHdr->xTex0.PSM )
        {
            case SCE_GS_PSMT8:
            case SCE_GS_PSMT4:
            case SCE_GS_PSMT8H:
            case SCE_GS_PSMT4HL:
            case SCE_GS_PSMT4HH:
                return true;
        }

        return false;
    }
}

const TIM2_MIPMAPHEADER* Asura_PS2_Texture::GetMipmapHdr( void ) const
{
    return( m_pxPicHdr->MipMapTextures > 1 ? reinterpret_cast<const TIM2_MIPMAPHEADER*>(m_pxPicHdr + 1) : NULL );
}

u_int Asura_PS2_Texture::GetWidthForUpload( void ) const
{
    return m_pxPicHdr->ImageWidth;
}

u_int Asura_PS2_Texture::GetHeightForUpload( void ) const
{
    return m_pxPicHdr->ImageHeight;
}

u_int Asura_PS2_Texture::GetWidth( void ) const
{
    u_int uW = GetWidthForUpload();
    return( m_bIsSwizzled ? uW << (256 == m_pxPicHdr->ClutColors ? 1 : 2) : uW );
}

u_int Asura_PS2_Texture::GetHeight( void ) const
{
    u_int uH = GetHeightForUpload();
    return( m_bIsSwizzled ? uH << 1 : uH );
}

u_int Asura_PS2_Texture::GetPSM( void ) const
{
    return m_pxPicHdr->xTex0.PSM;
}

u_int Asura_PS2_Texture::GetPSMForUpload( void ) const
{
    return( m_bIsSwizzled ? SCE_GS_PSMCT32 : GetPSM() );
}

u_int Asura_PS2_Texture::GetTBPOffset( u_int uLevel ) const
{
    return m_ausTBPOffsets[uLevel];
}

bool Asura_PS2_Texture::ExtractAndParsePicHdr( void )
{
    if( !m_pxPicHdr )
    {
        Asura_Assert( false, "No TIM2 header. TIM2 may have been broken." );
        return false;
    }

	Asura_Assert( (m_pxPicHdr->ClutType == TIM2_NONE)  || 
				  (m_pxPicHdr->ClutType == TIM2_RGB16) || 
				  (m_pxPicHdr->ClutType == TIM2_RGB32), "Texture has invalid clut type" );
	

    // MH added some asserts here - there should be some defines somewhere for the max size, but 
    // I'm can't remember where... the max width and height apply to 32-bit images, but they are
    // sensible limits anyway.
    Asura_Assert( m_pxPicHdr->ImageSize + m_pxPicHdr->ClutSize <= 1024 * 1024, "Max texture size is 1MB." );
    Asura_Assert( m_pxPicHdr->ImageWidth <= 512, "Max texture width is 512." );
    Asura_Assert( m_pxPicHdr->ImageHeight <= 512, "Max texture height is 512." );

    // Examine user data to determine whether this is a swizzled texture or
    // not.
    {
        // Get the user data.
        u_int uSizeOfUserData;
        const void* pvUserData = Tim2GetUserData( m_pxPicHdr, &uSizeOfUserData );

        // Swizzled?
        m_bIsSwizzled = false;
        if( pvUserData && uSizeOfUserData >= sizeof(Asura_PS2_Tim2UserData) )
        {
            const Asura_PS2_Tim2UserData* pxUserData = reinterpret_cast<const Asura_PS2_Tim2UserData*>( pvUserData );
            if( pxUserData->iID == ASURA_PS2_TIM2_USER_DATA_ID )
            {
                if( pxUserData->uSwizzled )
                {
                    m_bIsSwizzled = true;
                }
            }
        }
    }

    // Fix up the header.
    Tim2SetupTim2Header( m_pxPicHdr, m_bIsSwizzled );

    CalculateTBPOffsets();

    m_usImageDataSize = m_pxPicHdr->ImageSize + m_pxPicHdr->ClutSize;

    return true;
}

void Asura_PS2_Texture::CalculateTBPOffsets( void )
{
    // (MP - 21.10.2003)
    //
    // Rather than the method that Tim2CalcBufSize() uses - ie. just looking
    // at how many bytes does the area uW*uH occupy - we need to calculate
    // the number of texture blocks used and reserve all of those.
    //
    // ... Unfortunately, it seems to get a little more complicated than that.
    // There are various threads in sce.dev.prog.gs about this.
    //
    // I was seeing problems with, for example, a 384x384x8 texture where
    // some of the lower mip levels were overwriting higher ones. In
    // particular, it looked like level 2 was overwriting level 1 and level 3
    // was overwriting level 2.
    //
    // I think the issue was that level 1 was occupying 1.5 pages across.
    // Level 2 was then starting halfway across this second page but was
    // continuing onto the next one as well (level 2 being a 96x96.)
    //
    // I think rounding to a page boundary will always solve this. The
    // easiest way to sort this out, I think, is to just round to a page
    // boundary if we're calculating mip addresses (ie. MIPTBP1) manually.
    
    u_int uLog2PageWInBlocks;
    u_int uLog2PageHInBlocks;
    u_int uLog2BlockW;
    u_int uLog2BlockH;

    u_int uPSMForUpload = GetPSMForUpload();
    switch( uPSMForUpload )
    {
        default:
            Asura_Assert( false, "Invalid pixel format." );
            // Just fall through to a valid case.

        case SCE_GS_PSMCT32:
        case SCE_GS_PSMCT24:
        case SCE_GS_PSMT8H:
        case SCE_GS_PSMT4HH:
        case SCE_GS_PSMT4HL:
            uLog2PageWInBlocks = 3;
            uLog2PageHInBlocks = 2;
            uLog2BlockW        = 3;
            uLog2BlockH        = 3;
            break;

        case SCE_GS_PSMCT16:
        case SCE_GS_PSMCT16S:
            uLog2PageWInBlocks = 2;
            uLog2PageHInBlocks = 3;
            uLog2BlockW        = 4;
            uLog2BlockH        = 3;
            break;

        case SCE_GS_PSMT8:
            uLog2PageWInBlocks = 3;
            uLog2PageHInBlocks = 2;
            uLog2BlockW        = 4;
            uLog2BlockH        = 4;
            break;

        case SCE_GS_PSMT4:
            uLog2PageWInBlocks = 2;
            uLog2PageHInBlocks = 3;
            uLog2BlockW        = 5;
            uLog2BlockH        = 4;
            break;
    }

    // Assume we have no more than 3 mipmaps. I guess we might as well fix
    // it up if we have though.
    if( m_pxPicHdr->MipMapTextures > 4 )
    {
        Asura_Assert( false, "Texture has too many mipmaps." );
        m_pxPicHdr->MipMapTextures = 4;
    }

    u_int uUploadW = GetWidthForUpload();
    u_int uUploadH = GetHeightForUpload();

    bool bNeedPageAlignment = !CanGSCalculateMipmapAddrs( this );

    for( u_int u = 0; u < m_pxPicHdr->MipMapTextures; ++u )
    {
        u_int uBlocksUsedX = (uUploadW + ((1 << uLog2BlockW) - 1)) >> uLog2BlockW;
        u_int uBlocksUsedY = (uUploadH + ((1 << uLog2BlockH) - 1)) >> uLog2BlockH;

        if( bNeedPageAlignment )
        {
            uBlocksUsedX = ((uBlocksUsedX + ((1 << uLog2PageWInBlocks) - 1)) >> uLog2PageWInBlocks) << uLog2PageWInBlocks;
            uBlocksUsedY = ((uBlocksUsedY + ((1 << uLog2PageHInBlocks) - 1)) >> uLog2PageHInBlocks) << uLog2PageHInBlocks;
        }

        m_ausTBPOffsets[u] = uBlocksUsedX * uBlocksUsedY;

        uUploadW >>= 1;
        uUploadH >>= 1;
    }
}

bool Asura_PS2_Texture::HasNonOpaqueTexels( void ) const
{
    if( m_bIsSwizzled || m_pxPicHdr->ImageType == TIM2_IDTEX8 || m_pxPicHdr->ImageType == TIM2_IDTEX4 )
    {
        switch( m_pxPicHdr->ClutType & ASURA_PS2_TIM2_CLUT_FORMAT_MASK )
        {
            case TIM2_RGB16:
            {
                // Need to scan through the CLUT to find out...
                const Asura_PS2_16BitClutEntry* pxClutEntry = reinterpret_cast<const Asura_PS2_16BitClutEntry*>( GetClutData() );
                for( u_int uClutIdx = 0; uClutIdx < 256; ++uClutIdx )
                {
                    if( (pxClutEntry++)->usA == 0 )
                    {
                        return true;
                    }
                }
                break;
            }
    
            case TIM2_RGB24:
                break;

            case TIM2_RGB32:
                // NB: Assuming here it wouldn't be a 32-bit CLUT unless
                // there were some non-opaque texels.
                return true;

            default:
                Asura_Assert( false, "Invalid CLUT format." );
                break;
        }
    }
    else if( m_pxPicHdr->ImageType == TIM2_RGB32 )
    {
        return true;
    }

    return false;
}

bool Asura_PS2_Texture::Has8BitAlpha( void ) const
{
    if( m_bIsSwizzled || m_pxPicHdr->ImageType == TIM2_IDTEX8 || m_pxPicHdr->ImageType == TIM2_IDTEX4 )
    {
        switch( m_pxPicHdr->ClutType & ASURA_PS2_TIM2_CLUT_FORMAT_MASK )
        {
            case TIM2_RGB16:
            case TIM2_RGB24:
                return false;

            case TIM2_RGB32:
                return true;

            default:
                Asura_Assert( false, "Invalid CLUT format." );
                break;
        }
    }
    else if( m_pxPicHdr->ImageType == TIM2_RGB32 )
    {
        return true;
    }

    return false;
}

u_int Asura_PS2_Texture::GetDmaChainCallAddr( u_int uBfr ) const
{
    if( !m_aapxDmaChains[g_eCurTexBfrLayout][uBfr] )
    {
        CalculateDmaChain( uBfr );
    }
    
    return( reinterpret_cast<u_int>(m_aapxDmaChains[g_eCurTexBfrLayout][uBfr]) );
}

void Asura_PS2_Texture::CalculateDmaChain( u_int uBfr ) const
{
    Asura_Assert( ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS != g_eCurTexBfrLayout, "Texture buffer layout hasn't been set." );
    Asura_Assert( !m_aapxDmaChains[g_eCurTexBfrLayout][uBfr], "This DMA chain has already been calculated." );

    // Check there's some texture data to upload.
    if( !m_pxPicHdr->MipMapTextures )
    {
        return;
    }

	// Determine how much space we'll require per DMA chain for this texture.
	u_int uChainQWs = GetNumQWsForDmaChain();

    // Total size is that plus an extra QW for a RET tag.
    u_int uNumQWs = uChainQWs + 1;

    // Align to a 64 byte boundary to prevent cache issues.
    uNumQWs = ((uNumQWs + 3) >> 2) << 2;

	// Allocate.
	{
		// Allocate on a 8 QW (ie. 128 byte) boundary to speed up transfers
		// and to avoid cache issues.
		void* pvBfrSpace = Asura_AllocateAlign( uNumQWs << 4, ASURA_ALIGN_128_BYTE );
		if( !pvBfrSpace )
		{
			Asura_Assert( false, "Out of memory.");
			return;
		}

		m_aapxDmaChains[g_eCurTexBfrLayout][uBfr] = reinterpret_cast<Asura_PS2_QW*>( pvBfrSpace );
	}

    Asura_PS2_QW* pxOut = Asura_PS2_UncachedAcceleratedMem( m_aapxDmaChains[g_eCurTexBfrLayout][uBfr] );

	// Fill with standard chain.
    {
        u_int uBfrTBP  = Asura_PS2_TextureManagement::GetBfrTBP( uBfr );
        u_int uCLUTTBP = Asura_PS2_TextureManagement::GetBfrTBPForCLUT( uBfr );
	    WriteTextureUploadDmaData( pxOut, &uBfrTBP, uCLUTTBP, &uCLUTTBP );
    }

    // Final RET tag since this precalculated stuff will be CALLed.
	(reinterpret_cast<Asura_PS2_DmaTag*>(pxOut + uChainQWs))->SetWithNOPs( 0,
														                   Asura_PS2_DmaTag::PCE_NOCHANGE,
														                   Asura_PS2_DmaTag::ID_RET,
														                   false,
														                   0,
														                   false );
}

void Asura_PS2_Texture::Reset()
{
    for( u_int uTextureBufferLayout = 0; uTextureBufferLayout < ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS; ++uTextureBufferLayout )
    {
        for( u_int uTextureBuffer = 0; uTextureBuffer < ASURA_PS2_MAX_TEXTURE_BUFFERS; ++uTextureBuffer )
        {
            if( m_aapxDmaChains[uTextureBufferLayout][uTextureBuffer] )
            {
                Asura_DeallocateAlign( m_aapxDmaChains[uTextureBufferLayout][uTextureBuffer] );
                m_aapxDmaChains[uTextureBufferLayout][uTextureBuffer] = NULL;
            }
        }
    }
}


static const u_int NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD = 4;
static const u_int NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION  = 1 + NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD + 1;
static const u_int NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION  = 1 + NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION + 1;

u_int Asura_PS2_Texture::GetNumQWsForDmaChain( void ) const
{
    u_int uNumQWs = m_ucNumQWsForDMAChain;

    // Calculate if this is the first time.
    if( !uNumQWs )
    {
        const u_int uUploadPSM = GetPSMForUpload();
        const u_int uNumLevels = m_pxPicHdr->MipMapTextures;
        const u_int uUploadW   = GetWidthForUpload();
        const u_int uUploadH   = GetHeightForUpload();

        // If the texture has a CLUT, we'll need to transfer that as well.
        if( HasClut() )
        {
            uNumQWs += NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION;
        }
    
        // Iterate through mipmaps.
        for( u_int uLevel = 0; uLevel < uNumLevels; ++uLevel )
        {
            u_int uLevelWidth  = uUploadW >> uLevel;
            u_int uLevelHeight = uUploadH >> uLevel;
            u_int uQWsPerLine  = GetNumQWsPerNLines( 1, uUploadPSM, uLevelWidth );
            u_int uLinesPerSec = (uQWsPerLine ? Asura_PS2_GifTag::MAX_NLOOP / uQWsPerLine : uLevelHeight);
            u_int uNumSections = (uLevelHeight + (uLinesPerSec - 1)) / uLinesPerSec;
            if( 1 == uNumSections )
            {
                // This mipmap can be uploaded in a single section and
                // therefore all smaller mipmaps can be as well.
                uNumQWs += (uNumLevels - uLevel) * NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION;
                break;
            }
            else
            {
                uNumQWs += NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION * uNumSections;
            }
        }

        Asura_Assert( uNumQWs <= 255, "" );
        m_ucNumQWsForDMAChain = static_cast<u_char>( uNumQWs );
    }

    return uNumQWs;
}

void Asura_PS2_Texture::WriteUploadDmaDataForOneLevel( Asura_PS2_QW** ppxOut,
                                                       u_int uMipLevel,
										               u_int uTBP,
										               const u_long128* puImgData,
                                                       bool bNeedEOP ) const
{
    Asura_PS2_QW* pxOut = *ppxOut;

    u_int uUploadPSM = GetPSMForUpload();
    u_int uUploadW   = GetWidthForUpload()  >> uMipLevel;
    u_int uUploadH   = GetHeightForUpload() >> uMipLevel;

    if( m_bIsSwizzled )
    {
        u_int uRealPSM = GetPSM();
        u_int uRealW   = GetWidth()  >> uMipLevel;
        u_int uRealH   = GetHeight() >> uMipLevel;

        u_int uBlockW = ( SCE_GS_PSMT8 == GetPSM() ? 16 : 32 );
        u_int uBlockH = 16;

        // Any mip levels that are not multiples of the block dimensions will
        // not be swizzled.
        if( (uRealW & (uBlockW - 1)) || (uRealH & (uBlockH - 1)) )
        {
            uUploadPSM = uRealPSM;
            uUploadW   = uRealW;
            uUploadH   = uRealH;
        }
    }

	u_int uTBW = Tim2CalcBufWidth( uUploadPSM, uUploadW );

    // How many QWs per line of the texture?
    const u_int uQWsPerLine = GetNumQWsPerNLines( 1, uUploadPSM, uUploadW );

    // How many lines of the texture will we do in each packet?
    const u_int uLinesPerPacket = ( uQWsPerLine ? Asura_PS2_GifTag::MAX_NLOOP / uQWsPerLine : uUploadH );

	u_int uLinesRemaining = uUploadH;
    u_int uDestOffset     = 0;

	do
	{
        const u_int uNumLinesInThisPacket = Asura_Maths::Min( uLinesPerPacket, uLinesRemaining );
        const u_int uNumImageDataQWs = GetNumQWsPerNLines( uNumLinesInThisPacket, uUploadPSM, uUploadW );

        Asura_PS2_QW* const pxSectionStart = pxOut;

		// Write DMA tag.
		// PSJE - Added a VIF_DIRECT to the end of the dma tag.. this allows the texture to go:
		//		via path2 - The DIRECT code is used by vif1 
		//		via path3 - When dmaing to GIF we don't send the dma tag (TTE=0)
        Asura_PS2_DmaTag* const pxCntDmaTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut++ );
        pxCntDmaTag->Set( NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION,
                          Asura_PS2_DmaTag::PCE_NOCHANGE,
                          Asura_PS2_DmaTag::ID_CNT,
                          false,
                          0,
                          false,
                          0,
                          SCE_VIF1_SET_DIRECT( NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION, false ) );

        // GIF tag for register settings.
        Asura_PS2_GifTag* const pxRegGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
        pxRegGifTag->Set( NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD,
                          false,
                          false,
                          0,
                          SCE_GIF_PACKED,
                          1,
                          SCE_GIF_PACKED_AD );

	    // Set transmission between buffers.
		pxOut->u64s[0] = SCE_GS_SET_BITBLTBUF( 0, 0, 0, uTBP, uTBW, uUploadPSM );
		pxOut->u64s[1] = SCE_GS_BITBLTBUF;
        ++pxOut;

	    // Set transmission area between buffers (source (x, y), dest (x, y)
        // and direction.)
		pxOut->u64s[0] = SCE_GS_SET_TRXPOS( 0, 0, 0, uDestOffset, 0 );
		pxOut->u64s[1] = SCE_GS_TRXPOS;
        ++pxOut;

	    // Set size of transmission area.
		pxOut->u64s[0] = SCE_GS_SET_TRXREG( uUploadW, uNumLinesInThisPacket );
		pxOut->u64s[1] = SCE_GS_TRXREG;
        ++pxOut;

	    // Set transmission direction to HOST->LOCAL.
		pxOut->u64s[0] = SCE_GS_SET_TRXDIR( 0 );
		pxOut->u64s[1] = SCE_GS_TRXDIR;
        ++pxOut;

        Asura_Assert( static_cast<u_int>(pxOut - reinterpret_cast<Asura_PS2_QW*>(pxRegGifTag)) - 1 == NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD, "" );

        // GIF tag for IMAGE data.
		// Check if we need an EOP yet... (only put EOPs when bNeedEOP and it's the last chunk of a big upload)
		const bool bEOPThisTime = ( bNeedEOP && uLinesRemaining == uNumLinesInThisPacket );
#ifdef ASURA_ASSERT
		if  (uNumImageDataQWs == 0)
		{
			Asura_Assert( 0, "This texture is too small, and will crash the engine in a non Assert build" );
			Asura_Assert( 0, "Absaloute Minimum texture widths are : 32bit - 4, 16bit - 8, 8bit - 32, 4bit - 64" );

			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( 1,
				                                                 bEOPThisTime,
					                                             false,
						                                         0,
							                                     SCE_GIF_IMAGE,
								                                 0 );
		}
		else
#endif
		{
			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNumImageDataQWs,
				                                                 bEOPThisTime,
					                                             false,
						                                         0,
							                                     SCE_GIF_IMAGE,
								                                 0 );
		}

        Asura_Assert( static_cast<u_int>(pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCntDmaTag)) - 1 == NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION, "" );

    	// Transfer IMAGE data.
		// PSJE - Added a VIF_DIRECT to the end of the dma tag.. this allows the texture to go:
		//		via path2 - The DIRECT code is used by vif1 
		//		via path3 - When dmaing to GIF we don't send the dma tag (TTE=0)
        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumImageDataQWs,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_REF,
                                                             false,
                                                             reinterpret_cast<u_int>(puImgData),
                                                             false,
                                                             0,
                                                             SCE_VIF1_SET_DIRECT( uNumImageDataQWs, false ) );

        Asura_Assert( static_cast<u_int>(pxOut - pxSectionStart) == NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION, "" );

		uLinesRemaining -= uNumLinesInThisPacket;
		uDestOffset     += uNumLinesInThisPacket;
		puImgData       += uNumImageDataQWs;
	}
    while( uLinesRemaining );

    // Update output ptr.
    *ppxOut = pxOut;
}

bool Asura_PS2_Texture::WriteTextureUploadDmaData( Asura_PS2_QW* pxOut, u_int* puTBP, u_int uEndTBP, u_int* puCLUTTBP ) const
{
    Asura_PS2_QW* const pxOutputStart  = pxOut;

    // Get a pointer to the texture image data.
	const u_long128* puImgData = GetData();

    u_int       uTBP        = *puTBP;
    u_int       uCLUTTBP    = *puCLUTTBP;
    const u_int uUploadPSM  = GetPSMForUpload();
    const bool  bHasClut    = HasClut();
    const u_int uNumMipmaps = m_pxPicHdr->MipMapTextures;
    u_int       uUploadW    = GetWidthForUpload();
    u_int       uUploadH    = GetHeightForUpload();
	
    if( bHasClut )
    {
		const u_long128* puClutData		  = GetClutData();
        const u_int		 uClutPSM		  = GetClutPSM();
			  u_int		 uNumImageDataQWs = m_pxPicHdr->ClutColors;

		switch( uClutPSM )
		{
		    case SCE_GS_PSMCT16:
			    uNumImageDataQWs >>= 3; break;

		    case SCE_GS_PSMCT24:
		    case SCE_GS_PSMCT32:
			    uNumImageDataQWs >>= 2; break;

		    default:
			    Asura_Assert( false, "Invalid CLUT pixel format." );
			    break;
		}

        Asura_PS2_QW* const pxClutStart = pxOut;

	    // Use a CNT tag to transfer the GIF tags and register settings for
        // this block of texture.

		// PSJE - Added a VIF_DIRECT to the end of the dma tag.. this allows the texture to go:
		//		via path2 - The DIRECT code is used by vif1 
		//		via path3 - When dmaing to GIF we don't send the dma tag (TTE=0)
        Asura_PS2_DmaTag* const pxCntDmaTag = reinterpret_cast<Asura_PS2_DmaTag*>( pxOut++ );
        pxCntDmaTag->Set( NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION,
                          Asura_PS2_DmaTag::PCE_NOCHANGE,
                          Asura_PS2_DmaTag::ID_CNT,
                          false,
                          0,
                          false,
                          0,
						  SCE_VIF1_SET_DIRECT( NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION, false ) );

	    // GIF tag for texture settings
        Asura_PS2_GifTag* const pxRegGifTag = reinterpret_cast<Asura_PS2_GifTag*>( pxOut++ );
        pxRegGifTag->Set( NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD,
                          false,
                          false,
                          0,
                          SCE_GIF_PACKED,
                          1,
                          SCE_GIF_PACKED_AD );

    	// set transmission between buffers
		pxOut->u64s[0] = SCE_GS_SET_BITBLTBUF( 0, 0, 0,	uCLUTTBP, 1, uClutPSM );
		pxOut->u64s[1] = SCE_GS_BITBLTBUF;
        ++pxOut;

    	// set transmission area between buffers ( source x,y  dest x,y  and direction )
        pxOut->u64s[0] = SCE_GS_SET_TRXPOS( 0, 0, 0, 0, 0 );
		pxOut->u64s[1] = SCE_GS_TRXPOS;
        ++pxOut;

    	// set size of transmission area 
		// (MH - 10.01.2005) 4-bit (16 colour) palettes use a different arrangement
		if (SCE_GS_PSMT4 == GetPSM())
		{
			pxOut->u64s[0] = SCE_GS_SET_TRXREG( 8, 2 );
		}
		else
		{
			pxOut->u64s[0] = SCE_GS_SET_TRXREG( 16, (m_pxPicHdr->ClutColors >> 4) );
		}
		pxOut->u64s[1] = SCE_GS_TRXREG;
        ++pxOut;

    	// set transmission direction  ( HOST -> LOCAL Transmission 
		pxOut->u64s[0] = SCE_GS_SET_TRXDIR( 0 );
		pxOut->u64s[1] = SCE_GS_TRXDIR;
        ++pxOut;

        Asura_Assert( static_cast<u_int>(pxOut - reinterpret_cast<Asura_PS2_QW*>(pxRegGifTag)) - 1 == NUM_REGISTERS_TO_SET_FOR_TEX_UPLOAD, "" );

    	// GIF tag for texture.
#ifdef ASURA_ASSERT
		if  (uNumImageDataQWs == 0)
		{
			Asura_Assert( 0, "This texture is too small, and will crash the engine in a non Assert build" );
			Asura_Assert( 0, "Absaloute Minimum texture widths are : 32bit - 4, 16bit - 8, 8bit - 32, 4bit - 64" );

			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( 1,
				                                                 false,
					                                             false,
						                                         0,
							                                     SCE_GIF_IMAGE,
								                                 0 );
		}
		else
#endif
		{
			(reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNumImageDataQWs,
				                                                 false,
					                                             false,
						                                         0,
							                                     SCE_GIF_IMAGE,
								                                 0 );
		}

        Asura_Assert( static_cast<u_int>(pxOut - reinterpret_cast<Asura_PS2_QW*>(pxCntDmaTag)) - 1 == NUM_GIF_QWS_PER_TEX_UPLOAD_SECTION, "" );

    	// Transfer IMAGE data.
		// PSJE - Added a VIF_DIRECT to the end of the dma tag.. this allows the texture to go:
		//		via path2 - The DIRECT code is used by vif1 
		//		via path3 - When dmaing to GIF we don't send the dma tag (TTE=0)
        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumImageDataQWs,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_REF,
                                                             false,
                                                             reinterpret_cast<u_int>(puClutData),
                                                             false,
                                                             0,
															 SCE_VIF1_SET_DIRECT( uNumImageDataQWs, false ) );

        Asura_Assert( static_cast<u_int>(pxOut - pxClutStart) == NUM_DMA_QWS_PER_TEX_UPLOAD_SECTION, "" );

        uCLUTTBP += ASURA_PS2_NUM_64W_UNITS_FOR_CLUT;
	}

    const TIM2_MIPMAPHEADER* pxMipmapHdr = GetMipmapHdr();

    // Mipmap header must be present if we have anything more than the top
    // level image.
    Asura_Assert( 1 == uNumMipmaps || pxMipmapHdr, "" );

    // Transfer each mipmap (including top level image.)
    {
        Asura_Assert( uNumMipmaps > 0, "" );

        const u_int uEndLevel = uNumMipmaps - 1;
        for( u_int uLevel = 0; ; ++uLevel )
        {
		    // Check this texture isn't too big for the buffer it's going to
		    Asura_Assert( uTBP + m_ausTBPOffsets[uLevel] <= uEndTBP, "" );

            WriteUploadDmaDataForOneLevel( &pxOut, uLevel, uTBP, puImgData, uLevel == uEndLevel );

            uTBP += m_ausTBPOffsets[uLevel];

            if( uLevel == uEndLevel ) { break; }

            puImgData = reinterpret_cast<const u_long128*>( reinterpret_cast<const char*>(puImgData) + pxMipmapHdr->MMImageSize[uLevel] );
        
            uUploadW >>= 1;
            uUploadH >>= 1;
        }
    }

    Asura_Assert( GetNumQWsForDmaChain() == static_cast<u_int>(pxOut - pxOutputStart), "" );

    // Store updated pointers.
    *puTBP     = uTBP;
    *puCLUTTBP = uCLUTTBP;

    return true;
}

#if _DEBUG
void Asura_PS2_Texture::SetName( const char* szName )
{
    Asura_Assert( !m_szName, "" );
    m_szName = Asura_StringUtil::Duplicate( szName );
}
#endif

bool Asura_PS2_Texture::Setup( u_char* pucRawData )
{
    // Ensure we have some valid data to play with.
	if( !pucRawData || !Tim2CheckFileHeader(pucRawData) )
	{
        if( pucRawData )
        {
            Asura_DeallocateAlign( pucRawData );
        }

        pucRawData = CreateDummyTexture();
        if( !pucRawData ) { return false; }

#if _DEBUG
        m_bIsPinkTexture = true;
#endif
	}

    m_pucRawData = pucRawData;
    m_pxPicHdr   = Tim2GetPictureHeader( pucRawData, 0 );

    if( !ExtractAndParsePicHdr() )
    {
        return false;
    }

    return true;
}

//
// Asura_Texture_Management.
//

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Texture_Management::Platform_LoadTexture
// PARAMETERS	Name of texture
//				Index to load to
// RETURNS		None
// DESCRIPTION	Loads a texture to memory and keeps a record of its location
// AUTHOR		KJL
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Texture_Management::Platform_LoadTexture( const Asura_String& name, int index)
{
	if( IsTextureLoaded(index) )
	{
		return;
	}

	g_apxTextures[index] = Asura_PS2_TextureManagement::Project_CreateTexture(); // DJP
    if( !g_apxTextures[index] )
    {
        Asura_OutputToDebugger_Retail(( "ERROR: Out of memory.\n" ));
        return;
    }
	
	bool bLoadedOK = PlatformLoadTexture( name, index );
    if( bLoadedOK )
    {
    	Loaded[index] = s_uCurResetGroup;
    }
    else
    {
        delete g_apxTextures[index];
        g_apxTextures[index] = NULL;
    }
}
// End Of Function 'Asura_Texture_Management::Platform_LoadTexture'

void Asura_Texture_Management::Platform_LoadTextureFromResource(Asura_Handle fileHandle, int fileSize, int index, u_int uChunkFlags)
{
	u_char* pucTim2Buffer = reinterpret_cast<u_char*>( Asura_AllocateAlign(fileSize, ASURA_ALIGN_128_BYTE) );
    Asura_Assert( pucTim2Buffer, "failed to create temp buffer for texture load from resource file");
    if (pucTim2Buffer)
    {
	    Asura_FileAccess::ReadData(fileHandle,pucTim2Buffer,fileSize);

		g_apxTextures[index] = Asura_PS2_TextureManagement::Project_CreateTexture(); // DJP
        if( !g_apxTextures[index] )
        {
            Asura_OutputToDebugger_Retail(( "ERROR: Out of memory.\n" ));
            Asura_DeallocateAlign( pucTim2Buffer );
            return;
        }

        bool bOK = g_apxTextures[index]->Setup( pucTim2Buffer );

// (MP - 23.07.2002) Redundant - done in independent code.
//        Loaded[index] = true;
        
#if _DEBUG
        const Asura_String* pxTextureName = Asura_Texture_Management::GetTextureName( index );

        g_apxTextures[index]->SetName( pxTextureName->GetString() );

        if( bOK )
        {
            if( dbg_bPrintLoadedTextureInfo )
            {
                PrintLoadedTextureInfo( index );
            }
            else if( dbg_bReportSuccessfulTextureLoads )
            {
                Asura_OutputToDebugger(( "PlatformLoadTextureFromResource() loaded \"%s\" (index %d.)\n", pxTextureName->GetString(), index ));
            }
        }
        else
        {
            if( dbg_bReportFailedTextureLoads )
            {
                Asura_OutputToDebugger(( "PlatformLoadTextureFromResource() failed to load \"%s\" (index %d.)\n", pxTextureName->GetString(), index ));
            }
        }
#endif

        if( !bOK )
        {
            delete g_apxTextures[index];
            g_apxTextures[index] = NULL;
            Loaded[index] = ASURA_TEXTURE_NOT_LOADED;
        }
    }
}

void Asura_Texture_Management::ReloadTexture(int /*index*/)
{
	//Not supported on PS2. (RAR 8/8/02)
}

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Texture_Management::Platform_DestroyTexture
// PARAMETERS	Index of texture to destroy
// RETURNS		None
// DESCRIPTION	Deletes the record of the loaded texture
// AUTHOR		KJL
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Texture_Management::Platform_DestroyTexture( int iTexID )
{
    // (MH - 17.12.2002) in case the texture data is still referenced in a DMA chain
    // (MP - 18.10.2005) I don't see why we need to do anything more than
    // wait here. Resetting is a bit heavy handed.
    Asura_PS2_DeferredRendering::WaitForDeferredRendering();

	if( g_apxTextures[iTexID] )
	{
		delete g_apxTextures[iTexID];
    	g_apxTextures[iTexID] = NULL;
    }
}
// End Of Function 'Asura_Texture_Management::Platform_DestroyTexture'

void Asura_Texture_Management::Platform_Initialise( void )
{
    Asura_PS2_TextureManagement::Initialise();
}

void Asura_Texture_Management::Platform_Shutdown( void )
{
    Asura_PS2_TextureManagement::Shutdown();
}

void Asura_Texture_Management::Platform_Reset( void )
{
    Asura_PS2_TextureManagement::Reset();
}

//
// Asura_PS2_TextureManagement.
//

void Asura_PS2_TextureManagement::Initialise( void )
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar(L"Render.PS2.MoireFixToggle", &cnsl_bUseMipMapKMoireFix, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"Render.PS2.MoireFixValue", &cnsl_uMipMapKMoireFixOffset, 0, 255, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif
}

void Asura_PS2_TextureManagement::Shutdown( void )
{
}

void Asura_PS2_TextureManagement::Reset( void )
{
	// (MH - 17.12.2002) in case the texture data is still referenced in a DMA chain
    Asura_PS2_DeferredRendering::Reset();

    // (MH - 15.12.2002) to avoid fragmentation, delete the DMA chains - they will be recalculated as required.
    for (int iTexID = 0; iTexID < ASURA_MAX_NO_OF_TEXTURES; ++iTexID)
    {
        if( g_apxTextures[iTexID] )
	    {
            g_apxTextures[iTexID]->Reset();
        }
    }
}

bool Asura_PS2_TextureManagement::UploadTexture( int iTexID, u_int uBfr, void(*pfnAddrCallback)(u_int, u_int*, u_int*) )
{
    const bool bAutoBfrSwitching = ShouldUseAutoBfrSwitching( uBfr );
	if( bAutoBfrSwitching )
	{
        uBfr = GetNextAutoSwitchBfr();
	}

    Asura_Assert( uBfr < NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout], "Invalid texture buffer." );

    u_int uDmaChainCallAddr = 0;
    u_int uDataSize         = 0;
	

	if( pfnAddrCallback )
    {
        Asura_Assert( ASURA_TEXTURE_NONE == iTexID, "Can't specify texture ID if you're providing the DMA chain." );
        pfnAddrCallback( GetBfrTBP(uBfr), &uDmaChainCallAddr, &uDataSize );
    }
    else
    {
        Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );

        // Check if this texture is already in VRAM.
        if( dbg_bTexManagementCacheTexturesInternally )
        {
            u_int uExistingBfr = SearchInBfrsForTexture( iTexID, uBfr, bAutoBfrSwitching );
            if( BFR_NONE != uExistingBfr )
            {
	            g_uLastUsedBuffer = uExistingBfr;   
                if( bAutoBfrSwitching ) { g_uAutoBfrLastUsed = uExistingBfr; }
#if ASURA_PS2_TEXTUREMANAGER__PROFILING
	            ++s_uProfile_CacheHits;
#endif
                return false;
            }
        }

        const Asura_PS2_Texture* pxTexture = g_apxTextures[iTexID];
        if( !pxTexture ) { return false; }

#ifdef ASURA_SUPPORT_FMV_TEXTURES
	    if( Asura_Texture_Management::IsTextureFMV(iTexID) )
	    {
		    Asura_Platform_MovieToTextureStream* pxStream = Asura_MovieToTexture_System::GetStreamByTexID( iTexID );
		    if( pxStream )
		    {
				uDmaChainCallAddr = pxStream->GetDmaChainCallAddr( uBfr );
				uDataSize         = pxStream->GetDataSize();
		    }
	    }
        else
#endif
        {		    
			uDmaChainCallAddr = pxTexture->GetDmaChainCallAddr( uBfr );
            uDataSize         = pxTexture->GetImageDataSize();
        }
        
        if( !uDmaChainCallAddr ) { return false; }
    }


    const bool bNeedPATH2 = ( g_bAlwaysUsePATH2 || g_bForcePATH2UploadNext || uDataSize < ASURA_PS2_DEFERRED_RENDERING__MIN_PATH3_UPLOAD );

    if( bNeedPATH2 )	
	{
		Asura_PS2_DmaTag* pxTag = reinterpret_cast<Asura_PS2_DmaTag*>( Asura_PS2_DeferredRendering::GetVifBuffer(1) );

        // Need a FLUSH if we're forcing the texture along PATH2 due to a
        // reason in SetBufferLayout() - eg. because they're not using auto
        // layout stuff.
		pxTag->Set( 0,
					Asura_PS2_DmaTag::PCE_NOCHANGE,
					Asura_PS2_DmaTag::ID_CALL,
					false,
					uDmaChainCallAddr,
					false,
					SCE_VIF1_SET_NOP(false),
                    (g_bForcePATH2UploadNext ? SCE_VIF1_SET_FLUSH(false) : SCE_VIF1_SET_NOP(false)) );
	}
	else
	{
		// DMA tag to CALL the relevant DMA chain.
		// CALL TexUpload + GIF TAG (EOP=0) + GS_NOP  -  See sce.dev.prog.dma-Using CALL/RET with IMT mode for a full explanation
		Asura_PS2_QW *pxGifBuffer = Asura_PS2_DeferredRendering::GetGifBuffer(3);
        
        Asura_PS2_DmaTag* pxDMATag = reinterpret_cast<Asura_PS2_DmaTag*>( pxGifBuffer++ );
        pxDMATag->Set( 2, Asura_PS2_DmaTag::PCE_NOCHANGE, Asura_PS2_DmaTag::ID_CALL, false, uDmaChainCallAddr, false );
        // NB: Don't need to worry about upper 64 bits of tag since TTE is
        // off for the GIF channel.

        Asura_PS2_GifTag* pxGIFTag = reinterpret_cast<Asura_PS2_GifTag*>( pxGifBuffer++ );
        pxGIFTag->SetExceptRegs( 1, false, false, 0, SCE_GIF_PACKED, 1 );
        pxGIFTag->uAllRegs = SCE_GIF_PACKED_AD;

        pxGifBuffer->u64s[1] = SCE_GS_NOP;
	}

	g_aiTextureInVRAM[uBfr]  = iTexID;
    g_bForcePATH2UploadNext  = false;
    g_bLastTexWentAlongPATH2 = bNeedPATH2;

#if ASURA_PS2_TEXTUREMANAGER__PROFILING
	s_uProfile_UploadedBytes += uDataSize;
	++s_uProfile_UploadedQuantity;
#endif

	g_uLastUsedBuffer = uBfr;
    if( bAutoBfrSwitching ) { g_uAutoBfrLastUsed = uBfr; }
    
    return true;
}
// End Asura_PS2_TextureManagement::UploadTexture().

bool Asura_PS2_TextureManagement::UploadTextureToAddress( int iTexID, u_int* puTBP, u_int uEndTBP, u_int* puCLUTTBP )
{
    // Check the TBP is valid.
    Asura_Assert( ASURA_PS2_TEXBFR_BASE_PTR <= *puTBP && *puTBP < ASURA_PS2_TEXBFR_BASE_PTR + ASURA_PS2_TEXBFR_NUM_64W_UNITS, "" );

#ifdef ASURA_SUPPORT_FMV_TEXTURES
    // Not currently supporting FMV textures in this method.
	Asura_Assert( !Asura_Texture_Management::IsTextureFMV(iTexID), "" );
#endif

    const Asura_PS2_Texture* pxTexture = g_apxTextures[iTexID];
    if( !pxTexture ) { return false; }

    const u_int uChainQWs = pxTexture->GetNumQWsForDmaChain();
	const u_int uDataSize = pxTexture->GetImageDataSize();

    const bool bNeedPATH2 = ( g_bAlwaysUsePATH2 || g_bForcePATH2UploadNext || uDataSize < ASURA_PS2_DEFERRED_RENDERING__MIN_PATH3_UPLOAD );

    Asura_PS2_QW* pxOut;
    if( bNeedPATH2 )
    {
        if( g_bForcePATH2UploadNext )
        {
            pxOut = Asura_PS2_DeferredRendering::GetVifBuffer( 1 + uChainQWs );

		    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 0,
					                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
					                                             Asura_PS2_DmaTag::ID_CNT,
					                                             false,
					                                             0,
					                                             false,
					                                             SCE_VIF1_SET_NOP(false),
                                                                 SCE_VIF1_SET_FLUSH(false) );
        }
        else
        {
            pxOut = Asura_PS2_DeferredRendering::GetVifBuffer( uChainQWs );
        }
    }
    else
    {
        // NB: Don't need to worry about GIF tag + NOP stuff since we're not
        // using CALL/RET here.
        pxOut = Asura_PS2_DeferredRendering::GetGifBuffer( uChainQWs );
    }

    pxTexture->WriteTextureUploadDmaData( pxOut, puTBP, uEndTBP, puCLUTTBP );

    g_bForcePATH2UploadNext  = false;
    g_bLastTexWentAlongPATH2 = bNeedPATH2;

    // Hmm... maybe invalidate g_uLastUsedBuffer?

#if ASURA_PS2_TEXTUREMANAGER__PROFILING
	s_uProfile_UploadedBytes += uDataSize;
	++s_uProfile_UploadedQuantity;
#endif

    return true;
}

u_int Asura_PS2_TextureManagement::GetBfrUsedForLastTexture( void )
{
    return g_uLastUsedBuffer;
}

// Returns TBP value for specified buffer, ready for use with TEX0, etc.
u_int Asura_PS2_TextureManagement::GetBfrTBP( u_int uBfr )
{
    return g_aauTexBfrBasePtrs[g_eCurTexBfrLayout][TranslateBfrIdx(uBfr)];
}

u_int Asura_PS2_TextureManagement::GetBfrSizeIn64wUnits( u_int uBfr )
{
    return g_aauTexBfrSizesIn64wUnits[g_eCurTexBfrLayout][uBfr];
}

u_int Asura_PS2_TextureManagement::GetNumTexBfrsForCurLayout( void )
{
    return NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout];
}

u_int Asura_PS2_TextureManagement::GetBfrTBPForCLUT( u_int uBuffer )
{
	Asura_Assert( static_cast<u_int>(uBuffer) < NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout], "" );
	return( g_aauTexBfrBasePtrs[g_eCurTexBfrLayout][uBuffer] + g_aauTexBfrSizesIn64wUnits[g_eCurTexBfrLayout][uBuffer] - ASURA_PS2_NUM_64W_UNITS_FOR_CLUT - Project_GetCLUTBPOffset() );
}

u_int Asura_PS2_TextureManagement::TranslateBfrIdx( u_int uInBfr )
{
	if( ShouldUseAutoBfrSwitching(uInBfr) )
	{
		return g_uAutoBfrLastUsed;
	}
	else
	{
		// If they're asking for a particular buffer instead, make sure it's a
		// legal index.
		Asura_Assert( uInBfr < NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout], "" );

		return uInBfr;
	}
}

void Asura_PS2_TextureManagement::GetTEX0SettingForTBP( u_long* puTEX0, int iTexID, u_int uTBP, u_int uCLUTTBP )
{
	Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "" );  

#ifdef ASURA_SUPPORT_FMV_TEXTURES
	bool bIsFMV = Asura_Texture_Management::IsTextureFMV(iTexID);
	if( bIsFMV )
	{
		const Asura_Platform_MovieToTextureStream* pxStream = Asura_MovieToTexture_System::GetStreamByTexID( iTexID );
		if( pxStream )
		{
			pxStream->GetTex0( puTEX0, uTBP );
		}
		else
		{
			bIsFMV = false;
		}
	}

	if( !bIsFMV )
#endif
	{
        u_long uTEX0 = g_apxTextures[iTexID]->GetPicHdr()->uTex0;

        // TBP and CBP should be 0 by default so that we don't need to bother
        // masking
        Asura_Assert( !(uTEX0 & (GS_TEX0_TBP0_M | GS_TEX0_CBP_M)), "" );

        uTEX0 |= static_cast<u_long>(uTBP    ) << GS_TEX0_TBP0_O;
        uTEX0 |= static_cast<u_long>(uCLUTTBP) << GS_TEX0_CBP_O;

        *puTEX0 = uTEX0;
	}
}

void Asura_PS2_TextureManagement::GetTEX0Setting( u_long* puTEX0, int iTexID, u_int uBfr )
{
    uBfr = TranslateBfrIdx( uBfr );
    GetTEX0SettingForTBP( puTEX0, iTexID, GetBfrTBP(uBfr), GetBfrTBPForCLUT(uBfr) );
}

// Return value in range [1, 7].
u_int Asura_PS2_TextureManagement::GetNumMipmapLevels( int iTexID )
{
	Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "" );
    return g_apxTextures[iTexID]->GetPicHdr()->MipMapTextures;
}

void Asura_PS2_TextureManagement::GetTEX1Setting( u_long* puTEX1, int iTexID, u_int uMMAG, u_int uMMIN, bool bForceNoMipmaps /* = false */, int iMipMapKBias /* = 0 */ )
{
#ifdef ASURA_SUPPORT_FMV_TEXTURES
	// no mipmaps if we're uploading an FMV texture.
	if( Asura_Texture_Management::IsTextureFMV(iTexID) )
	{
		Asura_Platform_MovieToTextureStream* pxStream = Asura_MovieToTexture_System::GetStreamByTexID( iTexID );
		if( pxStream )
		{
			bForceNoMipmaps = true;
		}
	}
#endif
	const u_int uFlags = Asura_Texture_Management::GetTextureFlags(iTexID);
	const bool bMoireFix = (uFlags & AsuraTextureFlag_PS2MoireFix) != 0;
	const u_int uMipMapKOffset = (bMoireFix && cnsl_bUseMipMapKMoireFix) ? cnsl_uMipMapKMoireFixOffset : 0;

    u_long uTEX1 = g_apxTextures[iTexID]->GetPicHdr()->uTex1;

	if( dbg_bGloballyDisableMipmaps || bForceNoMipmaps || Asura_ScreenShot_System::IsTakingHiResScreenshot() )
    {
        uTEX1 &= ~GS_TEX1_MXL_M;
    }

    // LCM and L should be 0 by default.
    Asura_Assert( !(uTEX1 & (GS_TEX1_LCM_M | GS_TEX1_L_M)), "" );

    uTEX1 |= static_cast<u_long>(uMMAG     ) << GS_TEX1_MMAG_O;
    uTEX1 |= static_cast<u_long>(uMMIN     ) << GS_TEX1_MMIN_O;
    uTEX1 |= static_cast<u_long>(g_uMipMapK + uMipMapKOffset + iMipMapKBias) << GS_TEX1_K_O;

    *puTEX1 = uTEX1;
}

void Asura_PS2_TextureManagement::GetMIPTBP1SettingForTBP( u_long* puMIPTBP1, int iTexID, u_int uTBP )
{
    const Asura_PS2_Texture* pxTex = g_apxTextures[iTexID];

    u_int uPSM   = pxTex->GetPSM();
    u_int uWidth = pxTex->GetWidth();

    u_int aauValues[3][2];
    for( u_int u = 0; u < 3; ++u )
    {
        uTBP   += pxTex->GetTBPOffset( u );
        uWidth >>= 1;

        aauValues[u][0] = uTBP;
        aauValues[u][1] = Tim2CalcBufWidth( uPSM, uWidth );
    }

    *puMIPTBP1 = SCE_GS_SET_MIPTBP1( aauValues[0][0], aauValues[0][1], aauValues[1][0], aauValues[1][1], aauValues[2][0], aauValues[2][1] );
}

void Asura_PS2_TextureManagement::GetMIPTBP1Setting( u_long* puMIPTBP1, int iTexID, u_int uBfr )
{
    GetMIPTBP1SettingForTBP( puMIPTBP1, iTexID, GetBfrTBP(TranslateBfrIdx(uBfr)) );
}

void Asura_PS2_TextureManagement::BeginFrame( void )
{
#if ASURA_PS2_TEXTUREMANAGER__PROFILING
	s_uProfile_UploadedBytes	= 0;
	s_uProfile_UploadedQuantity = 0;
	s_uProfile_CacheHits		= 0;
#endif
}

void Asura_PS2_TextureManagement::EndFrame( void )
{
    // (MP - 08.08.2005) This shouldn't actually be required and taking it
    // out should allow some neat tricks with more advanced loading screens,
    // with textures resident in VRAM rather than in main RAM.
    if( dbg_bInvalidateTexBfrsAtEndOfFrame )
    {
    	InvalidateAllTextureBuffers();
    	g_eCurTexBfrLayout = ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS;
    }
}

void Asura_PS2_TextureManagement::InvalidateTextureBuffer( u_int uBfr )
{
    Asura_Assert( uBfr < NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout], "" );
    g_aiTextureInVRAM[uBfr] = ASURA_TEXTURE_NONE;
}

void Asura_PS2_TextureManagement::InvalidateAllTextureBuffers( void )
{
    for( u_int u = 0; u < ASURA_PS2_MAX_TEXTURE_BUFFERS; ++u )
    {
        g_aiTextureInVRAM[u] = ASURA_TEXTURE_NONE;
    }
}

void Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT eLayout, bool bEnableAutoSwitch, int iMinForAutoSwitch, int iMaxForAutoSwitch )
{
    bool bNeedToUsePATH2 = false;

    if( eLayout != g_eCurTexBfrLayout )
    {
		// If the buffer layout has changed then we should send the next
        // texture along PATH2 to avoid overwriting textures in use in the
        // previous layout.
        bNeedToUsePATH2 = true;

        g_eCurTexBfrLayout = eLayout;
        InvalidateAllTextureBuffers();

        // We'd better also reset all of our "last buffer" records since
        // these have no meaning for a different layout.
        g_uLastUsedBuffer = g_uAutoBfrLastUsed = NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout];
    }

    // We also need to use PATH2 if we're not using auto buffer switching or
    // if it has just been turned on. We can't really rely on the renderer in
    // question to not go uploading another texture into the last used buffer,
    // especially since we're not generally expecting external code to pay
    // any attention to which buffer is being used.
    if( !bEnableAutoSwitch || !g_bAutoBfrSwitchingEnabled )
    {
        bNeedToUsePATH2 = true;
    }

    g_bAutoBfrSwitchingEnabled = bEnableAutoSwitch;

    if( g_bAutoBfrSwitchingEnabled )
    {
        // Fill in defaults as necessary.
        if( -1 == iMinForAutoSwitch ) { iMinForAutoSwitch = 0; }
        if( -1 == iMaxForAutoSwitch ) { iMaxForAutoSwitch = NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout] - 1; }

        // Check limits are valid.
        Asura_Assert( 0 <= iMinForAutoSwitch, "" );
        Asura_Assert( iMinForAutoSwitch < iMaxForAutoSwitch, "" );
        Asura_Assert( static_cast<u_int>(iMaxForAutoSwitch) < NUM_TEXTURE_BUFFERS[g_eCurTexBfrLayout], "" );

        if( static_cast<u_int>(iMinForAutoSwitch) != g_uAutoBfrSwitchingMin || static_cast<u_int>(iMaxForAutoSwitch) != g_uAutoBfrSwitchingMax )
        {
            // If auto buffer switching limits have changed, a PATH2 upload
            // is again required, for the same reason as above.
            bNeedToUsePATH2 = true;

            g_uAutoBfrSwitchingMin = iMinForAutoSwitch;
            g_uAutoBfrSwitchingMax = iMaxForAutoSwitch;
        }
    }

    if( bNeedToUsePATH2 )
    {
        g_bForcePATH2UploadNext = true;
    }
}

void Asura_PS2_TextureManagement::SetupTextureEnvironment( bool bTextureUploaded,
                                                           u_int uRegFlags,
														   int iTextureID, u_int uBuffer,
														   int iMMAG, int iMMIN,
                                                           bool bForceNoMipmaps,
                                                           u_int uCTXT )
{
    // Check context is valid.
    Asura_Assert( 0 == uCTXT || 1 == uCTXT, "" );

    // If we're doing any registers other than TEXFLUSH, the texture ID
    // better have been set...
    if( uRegFlags & ~ASURA_PS2_TEXREGFLAG_TEXFLUSH ) { Asura_Assert( ASURA_TEXTURE_NONE != iTextureID, "" ); }

    // (MP - 06.08.2003) In the case of a PATH3 upload, according to the
    // "VIFcode FLUSHA Command Problem" Sony technote, the sequence has to
    // look like this:
    //
    //     FLUSH - DIRECT - FLUSHA
    //
    // However, we cannot put our TEX0 in that DIRECT since we need to do it
    // after the FLUSHA. So our sequence ends up being:
    //
    //     FLUSH - DIRECT [TEX1, MIPTBP1] - FLUSHA - DIRECT [TEXFLUSH, TEX0]
    //
    // Note that some or all of those register settings may be omitted,
    // depending on uRegFlags.
    //
    // In the case of a PATH2 upload, we end up with:
    //
    //     FLUSH - DIRECT [TEX1, MIPTBP1] - DIRECT [TEXFLUSH, TEX0]

    bool bWasPath3Upload = false;
    
	if( bTextureUploaded )
	{
		if( g_bLastTexWentAlongPATH2 )
		{
			g_bLastTexWentAlongPATH2 = false;
		}
		else
		{
            bWasPath3Upload = true;
        }
	}
    else
    {
        // If this fires, you've probably missed out or misordered a previous
        // call to this method.
        Asura_Assert( !g_bLastTexWentAlongPATH2, "" );
    }

    if( !bWasPath3Upload )
    {
        Asura_PS2_DeferredRendering::CancelLastPath3Gate();
    }

    bool bNeedMIPTBP1  = ( !bForceNoMipmaps && (uRegFlags & ASURA_PS2_TEXREGFLAG_MIPTBP1) && !CanGSCalculateMipmapAddrs(iTextureID) );
    bool bNeedTEXFLUSH = ( 0 != (uRegFlags & ASURA_PS2_TEXREGFLAG_TEXFLUSH) );

    if( dbg_bOptimiseOutTEXFLUSHWherePossible ) { bNeedTEXFLUSH &= bTextureUploaded; }

    u_int uNumRegsSet0 = 0;
    u_int uNumRegsSet1 = 0;
    
    if( uRegFlags & ASURA_PS2_TEXREGFLAG_TEX1 ) { ++uNumRegsSet0; }
    if( bNeedMIPTBP1                          ) { ++uNumRegsSet0; }
    if( bNeedTEXFLUSH                         ) { ++uNumRegsSet1; }
    if( uRegFlags & ASURA_PS2_TEXREGFLAG_TEX0 ) { ++uNumRegsSet1; }

    const u_int uQWsForDIRECT0 = 1 + uNumRegsSet0;
    const u_int uQWsForDIRECT1 = 1 + uNumRegsSet1;
    const u_int uQWsForDMA     = uQWsForDIRECT0 + 1 + uQWsForDIRECT1;
    const u_int uTotalQWs      = 1 + uQWsForDMA;

    Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
    Asura_PS2_QW* pxOut = pxBase;

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uQWsForDMA,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         SCE_VIF1_SET_FLUSH(false),
                                                         SCE_VIF1_SET_DIRECT(uQWsForDIRECT0, false) );

    (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNumRegsSet0,
                                                         true,
                                                         false,
                                                         0,
                                                         SCE_GIF_PACKED,
                                                         1,
                                                         SCE_GIF_PACKED_AD );

    if( uRegFlags & ASURA_PS2_TEXREGFLAG_TEX1 )
    {
        // Caller must provide the MMAG and MMIN parameters if they want this
        // register set.
        Asura_Assert( -1 != iMMAG && -1 != iMMIN, "" );

        GetTEX1Setting( &pxOut->u64s[0], iTextureID, iMMAG, iMMIN, bForceNoMipmaps );
        pxOut->u64s[1] = SCE_GS_TEX1_1 + uCTXT;
        ++pxOut;
    }

    if( bNeedMIPTBP1 )
    {
        GetMIPTBP1Setting( &pxOut->u64s[0], iTextureID, uBuffer );
        pxOut->u64s[1] = SCE_GS_MIPTBP1_1 + uCTXT;
        ++pxOut;
    }

    pxOut->u32s[0] = ( bWasPath3Upload ? SCE_VIF1_SET_FLUSHA(false) : SCE_VIF1_SET_NOP(false) );
    pxOut->u32s[1] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[2] = SCE_VIF1_SET_NOP( false );
    pxOut->u32s[3] = SCE_VIF1_SET_DIRECT( uQWsForDIRECT1, false );
    ++pxOut;

    (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNumRegsSet1,
                                                         true,
                                                         false,
                                                         0,
                                                         SCE_GIF_PACKED,
                                                         1,
                                                         SCE_GIF_PACKED_AD );

    if( bNeedTEXFLUSH )
    {
        (pxOut++)->u64s[1] = SCE_GS_TEXFLUSH;
    }

    if( uRegFlags & ASURA_PS2_TEXREGFLAG_TEX0 )
    {
        GetTEX0Setting( &pxOut->u64s[0], iTextureID, uBuffer );
        pxOut->u64s[1] = SCE_GS_TEX0_1 + uCTXT;
        ++pxOut;
    }

    Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );

    // Open/close PATH3 for next texture.
	Asura_PS2_DeferredRendering::AddPath3Gate();
}

u_int Asura_PS2_TextureManagement::HasTextureBeenUploaded( int iTexID, u_int uBfr )
{
    if( !dbg_bTexManagementCacheTexturesInternally ) { return Asura_PS2_TextureManagement::BFR_NONE; }

    const bool bAutoBfrSwitching = ShouldUseAutoBfrSwitching( uBfr );
	if( bAutoBfrSwitching )
	{
        uBfr = GetNextAutoSwitchBfr();
	}

    return SearchInBfrsForTexture( iTexID, uBfr, bAutoBfrSwitching );
}

bool Asura_PS2_TextureManagement::HasTexGot8BitAlpha( int iTexID )
{
    Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );

    if( g_apxTextures[iTexID] )
    {
        return g_apxTextures[iTexID]->Has8BitAlpha();
    }

    return false;
}

u_int Asura_PS2_TextureManagement::GetTextureWidth( int iTexID )
{
    Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );
    return( g_apxTextures[iTexID] ? g_apxTextures[iTexID]->GetWidth() : 0 );
}

u_int Asura_PS2_TextureManagement::GetTextureHeight( int iTexID )
{
    Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );
    return( g_apxTextures[iTexID] ? g_apxTextures[iTexID]->GetHeight() : 0 );
}

Asura_PS2_Texture* Asura_PS2_TextureManagement::GetTexture( int iTexID )
{
	if( iTexID < 0 || ASURA_MAX_NO_OF_TEXTURES <= iTexID )
	{
		Asura_Assert( 0 <= iTexID && iTexID < ASURA_MAX_NO_OF_TEXTURES, "Invalid texture ID." );
		return NULL;
	}
    return g_apxTextures[iTexID];
}


void Asura_PS2_TextureManagement::ForcePATH2Uploads( bool bForce )
{
    g_bAlwaysUsePATH2 = bForce;
}

void Asura_PS2_TextureManagement::ForcePATH2UploadNext()
{
	g_bForcePATH2UploadNext = true;
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

// following is to cope with nWidth going to zero unless first multiplied by nlines
static u_int GetNumQWsPerNLines( const u_int uNumLines, const u_int uPSM, const u_int uWidth )
{
    u_int uNumQWs = 0;
    switch( uPSM )
	{
	    case SCE_GS_PSMCT32:
		    uNumQWs = ((uNumLines * uWidth) >> 2); break;

	    case SCE_GS_PSMCT16:
		    uNumQWs = ((uNumLines * uWidth) >> 3); break;

	    case SCE_GS_PSMT8:
		    uNumQWs = ((uNumLines * uWidth) >> 4); break;

	    case SCE_GS_PSMT4:
		    uNumQWs = ((uNumLines * uWidth) >> 5); break;

        default:
            Asura_Assert( false, "Invalid or unhandled texel format." );
    }

    return uNumQWs;
}


#if _DEBUG

static const char* GetClutOrImageType(int type)
{
    switch (type)
    {
    case TIM2_NONE:
        return "TIM2_NONE";

    case TIM2_RGB16:
        return "TIM2_RGB16";

    case TIM2_RGB24:
        return "TIM2_RGB24";

    case TIM2_RGB32:
        return "TIM2_RGB32";

    case TIM2_IDTEX4:
        return "TIM2_IDTEX4";

    case TIM2_IDTEX8:
        return "TIM2_IDTEX8";

    default:
        return "<invalid>";
    }

}

static const char* GetDescriptionOfPSM( u_int uPSM )
{
    switch( uPSM )
    {
        case SCE_GS_PSMCT32:  return "PSMCT32";
        case SCE_GS_PSMCT24:  return "PSMCT24";
        case SCE_GS_PSMCT16:  return "PSMCT16";
        case SCE_GS_PSMCT16S: return "PSMCT16S";
        case SCE_GS_PSMT8:    return "PSMT8";
        case SCE_GS_PSMT4:    return "PSMT4";
        case SCE_GS_PSMT8H:   return "PSMT8H";
        case SCE_GS_PSMT4HL:  return "PSMT4HL";
        case SCE_GS_PSMT4HH:  return "PSMT4HH";
        case SCE_GS_PSMZ32:   return "PSMZ32";
        case SCE_GS_PSMZ24:   return "PSMZ24";
        case SCE_GS_PSMZ16:   return "PSMZ16";
        case SCE_GS_PSMZ16S:  return "PSMZ16S";
    }

    return "<invalid>";
}

static void PrintLoadedTextureInfo( int iTexID )
{
    const Asura_String* pxTextureName       = Asura_Texture_Management::GetTextureName( iTexID );
    const Asura_PS2_Texture* pxTexture      = g_apxTextures[iTexID];
    const TIM2_PICTUREHEADER* pxPicHdr      = pxTexture->GetPicHdr();

    Asura_OutputToDebugger_Retail(( "Loaded %s.\n"
//            "DMA chain addrs: 0 = %u; 1 = %u\n"
            "Img type=%s%s; clut cols=%d; clut type=%s; PSM=%s; TCC=%lu\n", 
            pxTextureName->GetString(), 
//            reinterpret_cast<u_int>(pxTexture->apxDmaChains[0]),
//            reinterpret_cast<u_int>(pxTexture->apxDmaChains[1]),
            GetClutOrImageType(pxPicHdr->ImageType),
            (pxTexture->IsSwizzled() ? " (swizzled)" : ""),
            pxPicHdr->ClutColors, 
            GetClutOrImageType(pxPicHdr->ClutType),
            GetDescriptionOfPSM(pxPicHdr->xTex0.PSM),
            pxPicHdr->xTex0.TCC ));
}

#endif // #if _DEBUG

static u_char* CreateDummyTexture( void )
{
	const int image_width = 4;
	const int image_height = 4;
	const int image_data_size = image_width * image_height * 2;// 16 bit = 2 bytes per pixel
	const int total_size = sizeof(TIM2_FILEHEADER) + sizeof(TIM2_PICTUREHEADER) + image_data_size;

	u_char* texture_data = reinterpret_cast<u_char*>( Asura_AllocateAlign(total_size, ASURA_ALIGN_16_BYTE) );
	if (!texture_data)
	{
        Asura_OutputToDebugger_Retail(( "ERROR: Out of memory.\n" ));
		return NULL;
	}

	TIM2_FILEHEADER * fhdr = (TIM2_FILEHEADER *) texture_data;
	fhdr->FileId[0] = 'T';
	fhdr->FileId[1] = 'I';
	fhdr->FileId[2] = 'M';
	fhdr->FileId[3] = '2';
	fhdr->FormatVersion = 3;
	fhdr->FormatId  = 0;					// 16-byte alignment mode
	fhdr->Pictures  = 1;					// Number of pictures is 1

	for( int i = 0; i < 8; i++ )
	{
		fhdr->pad[i] = 0x00;				// Clear padding member with 0x00
	}

	TIM2_PICTUREHEADER * ph = (TIM2_PICTUREHEADER *) (fhdr + 1);
	ph->TotalSize = image_data_size + sizeof(TIM2_PICTUREHEADER);				// Total size in bytes of all picture data
	ph->ClutSize = 0;				// Size in bytes of CLUT data part
	ph->ImageSize = image_data_size;				// Size in bytes of image data part
	ph->HeaderSize = sizeof(TIM2_PICTUREHEADER);				// Image header size
	ph->ClutColors = 0;				// CLUT size (number of colors in CLUT part)
	ph->PictFormat = 0;				// Image ID
	ph->MipMapTextures = 1;			// Number of MIPMAPs
	ph->ClutType = 0;				// CLUT part type
	ph->ImageType = TIM2_RGB16;		// Image part type
	ph->ImageWidth = image_width;	// Image width (not bit count)
	ph->ImageHeight = image_height;	// Image height (not bit count)
	(*reinterpret_cast<u_long*>(&ph->xTex0)) = 0;					// TEX0
	(*reinterpret_cast<u_long*>(&ph->xTex1)) = 0;					// TEX1
	ph->GsTexaFbaPabe = 0;			// Combination of TEXA, FBA, and PABE bits
	ph->GsTexClut = 0;				// TEXCLUT (low-order 32 bits directly)

	Asura_PS2_16BitClutEntry* pixel_16bit = reinterpret_cast<Asura_PS2_16BitClutEntry*>( ph + 1 );
	for (int h = 0; h < image_height; h++)
	{
		for (int w = 0; w < image_width; w++)
		{
			pixel_16bit->usR = (1 << 5) - 1;
			pixel_16bit->usG = (1 << 4);
			pixel_16bit->usB = (1 << 4);
			pixel_16bit->usA = 1;

			pixel_16bit++;
		}
	}

//	//MH - 02.05.2002) code for writing out texture
//	int filehandle = sceOpen("host0:\\sshots\\test.tm2", SCE_WRONLY | SCE_CREAT);
//	
//	if( filehandle < 0 )
//	{
//	// Processing for opening file failed
//		printf("file open failure: is it read-only?\n");
//		return texture_data;
//	}
//
//	sceWrite( filehandle, texture_data, total_size );
//	sceClose( filehandle );

	return texture_data;
}


/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		PlatformLoadTexture
// PARAMETERS	Name of texture to load
//				Index of where to load to
// RETURNS		None
// DESCRIPTION	Loads the texture file into memory and then uploads it to VRAM
// AUTHOR		KJL
/////////////////////////////////////////////////////////////////////////////////////////
static bool PlatformLoadTexture( const Asura_String& xName, int iTexID )
{
    static char szBuffer[256];
    int iNumChars;

    const char* const szInputName = xName.GetString();

    // Prepend "graphics\" to the path.
    {
        // (MP - 17.09.2002) Could use Asura_PrintfToBfr() here but
        // truncating will be just as bad as crashing here and it'll be
        // easier to spot in the latter case; failing to load a texture is a
        // pretty fatal error in the final game, I think.
        iNumChars = sprintf( szBuffer, "graphics%s%s", ('\\' == *szInputName ? "" : "\\"), szInputName );
        Asura_Assert( iNumChars < 256, "Path to texture is too long." );
    }

    // Replace extension.
    {
        char* szExt = strrchr( szBuffer, '.' );
        if( szExt )
        {
            // Check there's space for the "tm2" extension and the nul
            // terminator.
            Asura_Assert( (szExt + 1) - szBuffer <= 256 - 4, "Path to texture is too long." );
        }
        else
        {
            // Check there's space for the string ".tm2" and the nul
            // terminator.
            Asura_Assert( iNumChars <= 256 - 5, "Path to texture is too long." );

            szExt = &szBuffer[iNumChars];
            *szExt = '.';
        }

        *(szExt + 1) = 't';
        *(szExt + 2) = 'm';
        *(szExt + 3) = '2';
        *(szExt + 4) = '\0';
    }

    u_char* pucTim2Buffer = Asura_PS2_OS::LoadFileAligned( szBuffer, ASURA_ALIGN_128_BYTE );
    bool bOK = g_apxTextures[iTexID]->Setup( pucTim2Buffer );

#if _DEBUG
    g_apxTextures[iTexID]->SetName( szInputName );

    if( bOK )
    {
        if( dbg_bPrintLoadedTextureInfo )
        {
            PrintLoadedTextureInfo( iTexID );
        }
        else if( dbg_bReportSuccessfulTextureLoads )
        {
            Asura_OutputToDebugger(( "Loaded %d (%s).\n", iTexID, szBuffer ));
        }
    }
    else
    {
        if( dbg_bReportFailedTextureLoads )
        {
            Asura_OutputToDebugger(( "ERROR: Failed to load %d (%s).\n", iTexID, szBuffer ));
        }
    }
#endif

    return bOK;
}

