//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformTextureManagement.h
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PLATFORMTEXTUREMANAGEMENT_H___
#define ___ASURA_PLATFORMTEXTUREMANAGEMENT_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Engine platform-specific.
#include "Animation/Asura_Animation.h"
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_Handle.h"
#include "Rendering/Asura_TIM2.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#ifndef ASURA_PS2_SONY_SUBMISSION
	#define ASURA_PS2_TEXTUREMANAGER__PROFILING	(1)
#else
	#define ASURA_PS2_TEXTUREMANAGER__PROFILING	(0)
#endif

// T Y P E S /////////////////////////////////////////////////////////////////

// Texture buffer layout in VRAM.
enum ASURA_PS2_TEXBFRLAYOUT
{
    ASURA_PS2_TEXBFRLAYOUT_2_LARGE,          // 2 x 512Kb.
    ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL, // 2 x 384Kb + 1 x 256Kb.
    ASURA_PS2_TEXBFRLAYOUT_4_SMALL,          // 4 x 256Kb.
    ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS,
};

enum ASURA_PS2_TEXREGFLAG
{
    ASURA_PS2_TEXREGFLAG_NONE     = (1 << 0),
    ASURA_PS2_TEXREGFLAG_TEX1     = (1 << 1),
    ASURA_PS2_TEXREGFLAG_TEXFLUSH = (1 << 2),
    ASURA_PS2_TEXREGFLAG_TEX0     = (1 << 3),
    ASURA_PS2_TEXREGFLAG_MIPTBP1  = (1 << 4),     // Only sent if required, ie. if the texture is non-square or its dimensions are not powers of two.
    ASURA_PS2_TEXREGFLAG_ALL      = (1 << 5) - 1,
};

// C O N S T A N T S /////////////////////////////////////////////////////////

const u_int ASURA_PS2_MAX_TEXTURE_BUFFERS = 4;

// Values assume an 32-bit CLUT. Obviously halve them for 16-bit.
const u_int ASURA_PS2_NUM_BYTES_FOR_CLUT     = 1024;
const u_int ASURA_PS2_NUM_64W_UNITS_FOR_CLUT = (ASURA_PS2_NUM_BYTES_FOR_CLUT >> 8);

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_PS2_Texture
{

public:

	// C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

	Asura_PS2_Texture( void );

	virtual ~Asura_PS2_Texture( void );

	// M E T H O D S /////////////////////////////////////////////////////

	virtual bool Setup( u_char* pucRawData );
	bool ExtractAndParsePicHdr( void );

	// Will return 0 if out of memory.
	u_int GetDmaChainCallAddr( u_int uBfr ) const;

	// delete any extra memory that gets allocated when the texture is first used.
	// This helps avoid fragmentation.
	void Reset();

	inline const TIM2_PICTUREHEADER* GetPicHdr( void ) const { return m_pxPicHdr; }

	bool HasNonOpaqueTexels( void ) const;
	bool Has8BitAlpha( void ) const;

	inline bool IsSwizzled( void ) const { return m_bIsSwizzled; }

	u_int GetWidth( void ) const;
	u_int GetHeight( void ) const;
	u_int GetPSM( void ) const;

	u_int GetWidthForUpload( void ) const;
	u_int GetHeightForUpload( void ) const;
	u_int GetPSMForUpload( void ) const;

	bool HasClut( void ) const;
	u_int GetClutPSM( void ) const;

	u_int GetTBPOffset( u_int uLevel ) const;

#if _DEBUG
	void SetName( const char* szName );
#endif

	virtual u_int GetNumQWsForDmaChain( void ) const;
	u_int GetImageDataSize( void ) const { return m_usImageDataSize; }
	virtual bool WriteTextureUploadDmaData( Asura_PS2_QW* pxOut, u_int* puTBP, u_int uEndTBP, u_int* puCLUTTBP ) const;
	void WriteUploadDmaDataForOneLevel( Asura_PS2_QW** ppxOut,
		u_int uMipLevel, 
		u_int uTBP,
		const u_long128* puImgData,
		bool bNeedEOP ) const;

	u_long128* GetClutDataOpen( void ) const;

protected:

	// M E T H O D S /////////////////////////////////////////////////////

	const u_long128* GetData( void ) const;
	const u_long128* GetClutData( void ) const;

	const TIM2_MIPMAPHEADER* GetMipmapHdr( void ) const;

	void CalculateDmaChain( u_int uBfr ) const;

	void CalculateTBPOffsets( void );


	// S T A T E /////////////////////////////////////////////////////////

	// Image data.
	u_char* m_pucRawData;
	TIM2_PICTUREHEADER* m_pxPicHdr;

	// DMA chains, ready to be CALLed. These are lazily evaluated. This
	// class currently assumes a given texture is only used in
	// conjunction with a single memory layout; this is verified in debug
	// builds.
	mutable Asura_PS2_QW* m_aapxDmaChains[ASURA_PS2_TEXBFRLAYOUT_NUM_LAYOUTS][ASURA_PS2_MAX_TEXTURE_BUFFERS];

	// Num blocks used by a particular level and therefore TBP offset to
	// the next one.
	u_short m_ausTBPOffsets[4];

	// True iff this is a 4- or 8-bit palettised texture pretending to be
	// 32-bit to speed up texture uploads.
	bool m_bIsSwizzled; 

	// Cached size for DMA chain required to upload this texture. Lazily
	// evaluated. 0 is an invalid value (and implies it hasn't been
	// evaluated yet.)
	mutable u_char m_ucNumQWsForDMAChain;

	// Cached data size, extracted from the TIM2 picture header, in bytes.
	u_short m_usImageDataSize;

#if _DEBUG
	// Extra debug bits.
	char* m_szName;
	bool m_bIsPinkTexture;
#endif

};
// End class Asura_PS2_Texture.


class Asura_PS2_TextureManagement
{

    public:

        // C O N S T A N T S /////////////////////////////////////////////////

        // These are just abitrary constants much higher than the maximum
        // number of buffers we'll ever allow. They're powers of two to make
        // them easy to spot in the debugger.
        enum
        {
            BFR_AUTO = 64,
            BFR_NONE = 128,
        };

        // C L A S S   M E T H O D S /////////////////////////////////////////

        static void Initialise( void );
        static void Shutdown( void );
        static void Reset( void );

        static void BeginFrame( void );
        static void EndFrame( void );

        // All renderers should call this before drawing anything to ensure
        // the texture buffer layout is as intended. It is very cheap to call,
        // especially if the layout is already as requested.
        static void SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT eLayout, bool bEnableAutoSwitch = true, int iMinForAutoSwitch = -1, int iMaxForAutoSwitch = -1 );

        static void ForcePATH2Uploads( bool bForce );
		static void ForcePATH2UploadNext();
        
        // Functionality for simple texture uploads.
        static bool UploadTexture( int iTexID, u_int uBuffer = BFR_AUTO, void(*pfnAddrCallback)(u_int, u_int*, u_int*) = NULL );
        static bool UploadTextureToAddress( int iTexID, u_int* puTBP, u_int uEndTBP, u_int* puCLUTTBP );
		static void SetupTextureEnvironment( bool bTextureUploaded,
                                             u_int uRegFlags,                        // At least ASURA_PS2_TEXREGFLAG_TEXFLUSH must be set.
                                             int iTexID = -1, u_int uBfr = BFR_AUTO, // Ignored if only TEXFLUSHing.
                                             int iMMAG = -1, int iMMIN = -1,         // Ignored if not setting TEX1.
                                             bool bForceNoMipmaps = false,           // Ignored if not setting TEX1 or MIPTBP1.
                                             u_int uCTXT = 0 );

        // For corresponding renderstate setup.
        static void GetTEX0SettingForTBP( u_long* puTEX0, int iTexID, u_int uTBP, u_int uCLUTTBP );
        static void GetTEX0Setting( u_long* puTEX0, int iTexID, u_int uBfr = BFR_AUTO );
        static void GetTEX1Setting( u_long* puTEX1, int iTexID, u_int uMMAG, u_int uMMIN, bool bForceNoMipmaps = false, int iMipMapKBias = 0 );
        static void GetMIPTBP1SettingForTBP( u_long* puMIPTBP1, int iTexID, u_int uTBP );
        static void GetMIPTBP1Setting( u_long* puMIPTBP1, int iTexID, u_int uBfr = BFR_AUTO );
        static u_int GetNumMipmapLevels( int iTexID );

        // For querying results of previous uploads.
        static u_int HasTextureBeenUploaded( int iTexID, u_int uBuffer = BFR_AUTO );
        static u_int GetBfrUsedForLastTexture( void );

        // For use when texture buffer contents have been modified outside of
        // this class.
        static void InvalidateTextureBuffer( u_int uBuffer );
        static void InvalidateAllTextureBuffers( void );

        static u_int GetBfrTBP( u_int uBuffer = BFR_AUTO );
        static u_int GetBfrSizeIn64wUnits( u_int uBfr );
        static u_int GetNumTexBfrsForCurLayout( void );
		static u_int GetBfrTBPForCLUT( u_int uBuffer );
		static u_int TranslateBfrIdx( u_int uInBfr );

#if ASURA_PS2_TEXTUREMANAGER__PROFILING
        static u_int Profile_GetNumUploadedBytes( void ) { return s_uProfile_UploadedBytes; }
        static u_int Profile_GetNumUploads( void ) { return s_uProfile_UploadedQuantity; }
        static u_int Profile_GetNumCacheHits( void ) { return s_uProfile_CacheHits; }
#endif

        static bool HasTexGot8BitAlpha( int iTexID );
        
        static u_int GetTextureWidth( int iTexID );
        static u_int GetTextureHeight( int iTexID );

		static Asura_PS2_Texture* GetTexture( int iTexID );

		// Project specific functions
		static Asura_PS2_Texture *Project_CreateTexture(); // normally { return new Asura_PS2_Texture(); }
		static int Project_GetCLUTBPOffset(); // normally { return 0; }

    private:

        // C L A S S   S T A T E /////////////////////////////////////////////

#if ASURA_PS2_TEXTUREMANAGER__PROFILING
		static u_int s_uProfile_UploadedBytes;
		static u_int s_uProfile_UploadedQuantity;
		static u_int s_uProfile_CacheHits;
#endif

};
// End class Asura_PS2_TextureManagement.

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PLATFORMTEXTUREMANAGEMENT_H___
