//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformRenderHierarchy.cpp
//
// Created by Morgan Parry.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Platform_RenderHierarchy.h"

// Stdlib platform-specific.
#include <libgraph.h>

// Engine independent.
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#if _DEBUG
#include "Camera/Asura_Camera.h"
#endif
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Maths.h"
#if _DEBUG
#include "Core/Asura_Timers.h"
#endif
#include "Light/Asura_GI_Solution.h"
#if _DEBUG
#include "Rendering/Asura_Render.h"
#endif

// Engine platform-specific.
#include "Animation/Asura_PS2_Hierarchy.h"
#include "Animation/Asura_PS2_SkinTypes.h"
#include "Debugging/Asura_PS2_DebugInfo.h"
#include "Profiling/Asura_PS2_Profiling.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_GSMRenderer.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"
#include "Rendering/Asura_PS2_Shadows.h"
#include "Rendering/Asura_PS2_VU1.h"
#include "Rendering/Asura_PS2_ShadowBuffer.h"

// T Y P E S /////////////////////////////////////////////////////////////////

// Just to tidy up the list of memory layout defines below.
typedef const Asura_Animation*                     HS_AnimPtr;
typedef const Asura_SphericalHarmonics_Order2_RGB* HS_SHPtr;
typedef const Asura_PS2_HierSkin_SkinHdr*          HS_SkinHdrPtr;

// G L O B A L S /////////////////////////////////////////////////////////////

DEBUGVAR bool dbg_bDrawHierBBs = false;

#if _DEBUG
DEBUGVAR u_int dbg_uNumHiersToSkipPerFrame   = 0;
DEBUGVAR u_int dbg_uMaxHiersToDrawPerFrame   = 0xFFFFFFFFu;
DEBUGVAR u_int dbg_uHierOnlyRenderOneSection = 0xFFFFFFFFu;
#endif

#define HS_SPR_QW_OFFSET  (ASURA_PS2_GSM_SPR_QWS_RESERVED)
#define HS_SPR_START_ADDR (0x70000000 + (HS_SPR_QW_OFFSET << 4))

// General.
#define g_pxAnim                 (*reinterpret_cast<HS_AnimPtr*    >(HS_SPR_START_ADDR + 0x00))
#define g_pxSkinHdr              (*reinterpret_cast<HS_SkinHdrPtr* >(HS_SPR_START_ADDR + 0x04))
// Camera.
#define g_uEECamVisFlags         (*reinterpret_cast<u_int*         >(HS_SPR_START_ADDR + 0x08))
// Matrices.
#define g_bWorldspaceBones       (*reinterpret_cast<bool*          >(HS_SPR_START_ADDR + 0x0C))
#define g_pxO2W                  ( reinterpret_cast<Asura_PS2_QW*  >(HS_SPR_START_ADDR + 0x10))
#define g_pxSec2O                ( reinterpret_cast<Asura_PS2_QW*  >(HS_SPR_START_ADDR + 0x50))
#define g_pxSec2W                ( reinterpret_cast<Asura_PS2_QW*  >(HS_SPR_START_ADDR + 0x90))
// Bounding sphere.
#define g_xBSCentre              (*reinterpret_cast<Asura_Vector_3*>(HS_SPR_START_ADDR + 0xD0))
#define g_fBSRadius              (*reinterpret_cast<float*         >(HS_SPR_START_ADDR + 0xDC))
// Lighting.
#define g_pxSH                   (*reinterpret_cast<HS_SHPtr*      >(HS_SPR_START_ADDR + 0xE0))
#define g_fAlpha                 (*reinterpret_cast<float*         >(HS_SPR_START_ADDR + 0xE4))
// Total.
#define TOTAL_SPR_QWS_REQD                                                              (0xF)

#if HS_SPR_QW_OFFSET + TOTAL_SPR_QWS_REQD > 1024
    #error Oops, overran SPR.
#endif
#if HS_SPR_QW_OFFSET + TOTAL_SPR_QWS_REQD > ASURA_PS2_HIERSKIN_SPR_QWS_RESERVED
    #error Out of sync with header.
#endif

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Hierarchy_System.
//

void Asura_Hierarchy_System::Platform_RenderHierarchy( const Asura_Animation* pxAnim,
                                                       u_int uCamVisFlags,
                                                       u_int uRenderFlags,
                                                       const Asura_Vector_3& xBSCentre,
                                                       float fBSRadius, 
													   const int iTexIDOverride /*=ASURA_TEXTURE_NONE*/ )
{
    if( !Asura_PS2_HierarchySkin_Renderer::Begin(pxAnim, uCamVisFlags, xBSCentre, fBSRadius) )
    {
        return;
    }
    
	const Asura_PS2_HierSkin_SkinHdr* pxSkinHdr;
    const u_short* pusSecMatCounts;
    Asura_PS2_HierarchySkin_Renderer::GetHdrData( &pxSkinHdr, &pusSecMatCounts );
			
	Asura_PS2_HierarchySkin_Renderer::SetTexture( (iTexIDOverride != ASURA_TEXTURE_NONE) ? iTexIDOverride : pxSkinHdr->sTextureID );

    if( !(uRenderFlags & ASURA_HIERARCHY_RENDER_NO_SHADOW) )
    {
        Asura_PS2_HierarchySkin_Renderer::SetupShadow();
    }

    Asura_PS2_HierarchySkin_Renderer::SetupMatrices();

    Asura_PS2_HierarchySkin_Renderer::SetLighting( Asura_GI_Solution::s_xSHForCurrentRender, pxAnim->GetAlpha() );

    const u_int uNumSections = pxSkinHdr->usSectionCount;
	const Asura_PS2_HierSkin_SectionMatHdr* pxSecMatHdr = pxSkinHdr->pxSectionMatHdrStart;
    u_int uStripCALLAddr = pxSkinHdr->uStripDataStart;

    bool bFirstNonEmptySection = true;
    int  iCurMaterialID        = -999;
    
    for( u_int uSection = 0; uSection < uNumSections; ++pusSecMatCounts, ++uSection )
    {
        u_int uSecMatsLeft = *pusSecMatCounts;
        if( !uSecMatsLeft ) { continue; }

#if _DEBUG
        const bool bSkipThisSection = ( 0xFFFFFFFFu != dbg_uHierOnlyRenderOneSection && dbg_uHierOnlyRenderOneSection != uSection );
#endif

        // NB: Check the number of triangles rendered so far, not whether or
        // not this is the first section, since the first section (or more)
        // may have 0 materials. Bloody artists.
        Asura_PS2_HierarchySkin_Renderer::SetupSection( uSection, bFirstNonEmptySection );
        bFirstNonEmptySection = false;

        for( ; uSecMatsLeft; ++pxSecMatHdr, --uSecMatsLeft )
        {
#if _DEBUG
            if( !bSkipThisSection )
#endif
            {
                const int iNextMaterialID = pxSecMatHdr->sMaterialID;
                if( iNextMaterialID != iCurMaterialID )
                {
                    u_int uMaterialFlags = Asura_Material_System::GetMaterialFlags( iNextMaterialID );

					if(pxAnim->GetAlpha() != 1.0f)
					{
						uMaterialFlags |= AsuraMaterialFlag_Transparent;
					}

                    Asura_PS2_GSMRenderer::SetRenderstate( uMaterialFlags );
                
                    if( uMaterialFlags & AsuraMaterialFlag_TextureAnim )
                    {
                        Asura_TextureAnimationOutput xTexAnimState;
                        Asura_Material_System::GetTextureAnimationState( iNextMaterialID, pxAnim->GetTextureAnimController(), &xTexAnimState );
                        Asura_PS2_GSMRenderer::SetupPass0STManipulations( true, 1.0f, 1.0f, xTexAnimState.m_fUOffset, xTexAnimState.m_fVOffset );
                    }
                
                    iCurMaterialID = iNextMaterialID;
                }

                Asura_PS2_GSMRenderer::RenderStripChain( uStripCALLAddr );

                Asura_PS2_Render::UpdateTriStats( pxSecMatHdr->usTriCount * Asura_PS2_GSMRenderer::GetNumCamsVisFor(),
                                                  Asura_PS2_GSMRenderer::GetNumPassesPerTri(),
                                                  Asura_PS2_Render::TRICOUNT_HIER );
            }

            uStripCALLAddr += pxSecMatHdr->uByteOffsetToNextChain;
        }
    }

    Asura_PS2_HierarchySkin_Renderer::End();
}

void Asura_Hierarchy_System::Platform_RenderHierarchyShadow( const Asura_Animation& xAnim, const Asura_Shadow& xShadow, const u_int uCamVisFlags, const u_int uRenderFlags )
{
}

//
// Asura_PS2_HierarchySkin_Renderer.
//

bool Asura_PS2_HierarchySkin_Renderer::Begin( const Asura_Animation* pxAnim,
                                              u_int uCamVisFlags,
                                              const Asura_Vector_3& xBSCentre,
                                              float fBSRadius )
{
    Asura_Assert( uCamVisFlags, "" );

    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DRAW_HIERS) ) { return false; }

#if _DEBUG
    if( 0xFFFFFFFFu != dbg_uMaxHiersToDrawPerFrame )
    {
        static u_int uTimestamp = 0xFFFFFFFFu;
        static u_int uNumHiersDrawnThisFrame;

        const u_int uCurFrame = Asura_Timers::GetFrameNumber();
        if( uCurFrame != uTimestamp )
        {
            uNumHiersDrawnThisFrame = 0;
            uTimestamp = uCurFrame;
        }

        ++uNumHiersDrawnThisFrame;

        if( uNumHiersDrawnThisFrame > dbg_uMaxHiersToDrawPerFrame ) { return false; }

        if( uNumHiersDrawnThisFrame <= dbg_uNumHiersToSkipPerFrame ) { return false; }
    }
#endif

    const Asura_PS2_HierSkin_SkinHdr* pxSkinHdr;

    // Get appropriate skin hdr.
    {
        const Asura_PS2_SkinEntry* pxSkinEntry = Asura_PS2_SkinSystem::GetSkinEntry( Asura_Hierarchy_System::s_pxSkin->m_uNameHashID );
        if( !pxSkinEntry ) { return false; }

        pxSkinHdr = Asura_PS2_SkinSystem::GetHierSkinHdr( pxSkinEntry, Asura_Hierarchy_System::GetWorldspaceBonePos(0), uCamVisFlags, NULL );
        if( !pxSkinHdr ) { return false; }
    }

    // Special handling for empty skins, which sometimes get abused by
    // artists for various reasons. Note that in this case, the render data's
    // section count won't match up with the skin's bone count - specifically,
    // it will be 0. (Of course, this check also handles the perverse case
    // where a skin genuinely has no sections, though I'm not sure that can
    // ever happen.)
    if( !pxSkinHdr->usSectionCount ) { return false; }

	ASURA_PS2_PROFILE__ADD_VIF_START( ASURA_PS2_PROFILE__HIERARCHY_MARKER );

    // Verify it matches the independent skin data.
    Asura_Assert( static_cast<int>(pxSkinHdr->usSectionCount) == Asura_Hierarchy_System::s_pxSkin->NumberOfBones, "" );

    Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_MEDIUM_1_SMALL, true, 0, 1 );

    Asura_PS2_GSMRenderer::Begin();

    (void) Asura_PS2_SPR_Alloc( HS_SPR_QW_OFFSET, TOTAL_SPR_QWS_REQD );
    
    g_pxAnim           = pxAnim;
    g_pxSkinHdr        = pxSkinHdr;
    g_uEECamVisFlags   = uCamVisFlags;
    g_bWorldspaceBones = Asura_Hierarchy_System::s_bBonesInWorldspace;
    g_xBSCentre        = xBSCentre;
    g_fBSRadius        = fBSRadius;

    Asura_PS2_GSMRenderer::SetCamFlags( uCamVisFlags, g_xBSCentre, g_fBSRadius );

    return true;
}

void Asura_PS2_HierarchySkin_Renderer::GetHdrData( const Asura_PS2_HierSkin_SkinHdr** ppxSkinHdr, const u_short** ppusSecMatCounts )
{
    *ppxSkinHdr       = g_pxSkinHdr;
    *ppusSecMatCounts = reinterpret_cast<const u_short*>( g_pxSkinHdr + 1 );
}

void Asura_PS2_HierarchySkin_Renderer::SetTexture( int iTexID )
{
    const Asura_PS2_HierSkin_SkinHdr* pxSkinHdr = g_pxSkinHdr;

	Asura_PS2_GSMRenderer::SetTexture( iTexID, SCE_GS_LINEAR, SCE_GS_LINEAR_MIPMAP_NEAREST, Asura_PS2_TextureManagement::BFR_AUTO, true );

    // Also upload reflective texture, if it's required.
    if( Asura_PS2_GSMRenderer::IsMaterialReflective(pxSkinHdr->uAllMaterialFlags) )
    {
        Asura_PS2_GSMRenderer::SetTexture( Asura_Render::GetReflectanceMap(), SCE_GS_LINEAR, SCE_GS_LINEAR_MIPMAP_NEAREST, 2, false );
    }
}

void Asura_PS2_HierarchySkin_Renderer::SetupShadow( void )
{
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_HIER_SHADOWS) )
    {
        return;
    }

    const Asura_Vector_3& xBSCentre    = g_xBSCentre;
    const float           fBSRadius    = g_fBSRadius;
    const u_int           uCamVisFlags = g_uEECamVisFlags;

    if( Asura_PS2_Shadows::CalcAlphaForMinDistFromCam(xBSCentre, fBSRadius, uCamVisFlags) <= 0.00f )
    {
        return;
    }

    const Asura_Animation* pxAnim = g_pxAnim;

    Asura_PS2_ShadowBuffer* pxShadowBfr = Asura_PS2_ShadowBufferSystem::GetShadowBufferForAnim( pxAnim, xBSCentre, fBSRadius, uCamVisFlags );
    if( !pxShadowBfr ) { return; }

    Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = pxShadowBfr;

    // save time on doing ground height test if we aren't using fallback shadow quad
    if( pxShadowBfr->IsForcedFallback() )
    {
        float fQuadHeight;
        Asura_PS2_Shadows::Proj_GetShadowYAndHeightForAnim( pxAnim, *Asura_Hierarchy_System::s_pxPos, &fQuadHeight, NULL );
        pxShadowBfr->SetShadowBufferQuadHeight( fQuadHeight );
    }

    Asura_PS2_GSMRenderer::EnableShadow( xBSCentre, fBSRadius );
    pxShadowBfr->JustSetupRenderToBfr( SCE_VIF1_SET_FLUSH(false) );
}

void Asura_PS2_HierarchySkin_Renderer::SetupMatrices( void )
{
    if( g_bWorldspaceBones )
    {
        sceVu0UnitMatrix( &g_pxO2W->f32s );
    }
    else
    {
        Asura_PS2_VU1::SetupPS2MatFromAsuraData( &g_pxO2W->f32s[0], *Asura_Hierarchy_System::s_pxOri, *Asura_Hierarchy_System::s_pxPos );
    }
}

void Asura_PS2_HierarchySkin_Renderer::InsertWorldspaceTransform( const Asura_PS2_QW* pxTransform )
{
    sceVu0MulMatrix( &g_pxO2W->f32s, &(const_cast<Asura_PS2_QW*>(pxTransform))->f32s, &g_pxO2W->f32s );
}

void Asura_PS2_HierarchySkin_Renderer::SetupSection( u_int uSectionIdx, bool bResetLighting )
{
    const Asura_Matrix_3x3& xBoneOri = Asura_Hierarchy_System::s_axBoneOri[uSectionIdx];
    const Asura_Vector_3&   xBonePos = Asura_Hierarchy_System::s_axBonePos[uSectionIdx];

    Asura_PS2_VU1::SetupPS2MatFromAsuraData( &g_pxSec2O->f32s[0], xBoneOri, xBonePos );
    sceVu0MulMatrix( &g_pxSec2W->Vu0FVec, &g_pxO2W->Vu0FVec, &g_pxSec2O->Vu0FVec );
    Asura_PS2_GSMRenderer::SetObjectToWorldMat( &g_pxSec2W->f32s[0] );
    
    if( bResetLighting )
    {
        // First time through we need to setup the lighting data from
        // scratch. Subsequent sections just need to fix up the
        // directions for the updated transform.
        Asura_PS2_GSMRenderer::UseExistingSHForMesh( *g_pxSH, g_fAlpha );
    }
    else
    {
        Asura_PS2_GSMRenderer::RecalculateSHForChangedObjectToWorldMat();
    }
}

void Asura_PS2_HierarchySkin_Renderer::SetLighting( const Asura_SphericalHarmonics_Order2_RGB& xSH, float fAlpha )
{
    g_pxSH   = &xSH;
    g_fAlpha = fAlpha;
}

void Asura_PS2_HierarchySkin_Renderer::End( void )
{
    Asura_PS2_GSMRenderer::End();

	// ensure that later code doesn't think it should use this same shadow buffer !
	Asura_PS2_ShadowBuffer::s_pxCurrentShadowBuffer = NULL;

	ASURA_PS2_PROFILE__ADD_VIF_END( ASURA_PS2_PROFILE__HIERARCHY_MARKER );

#ifndef ASURA_PS2_SONY_SUBMISSION
    if( dbg_bDrawHierBBs ) { Asura_PS2_DebugInfo::RenderBBForSphere( g_xBSCentre, g_fBSRadius ); }
#endif

    Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(HS_SPR_START_ADDR) );
}
