
//////////////////////////////////////////////////////////////////////////////
//
// VU1_FractalSky.vcl
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     Basic tri renderer with the addition of some texture coordinate
//     manipulation to create the desired effect. Effect mimics KJL's PC
//     version.
//
//     Does no work with RGBs, instead assuming that all tris need to be
//     drawn with the same constant colour and that the GS's RGBAQ register
//     has already been set appropriately. Also doesn't bother perspective
//     correcting the UVs; didn't seem to make any difference and it makes
//     things quite a lot easier not to bother.
//
//     The renderer assumes that there are three render passes.
//
//     See Asura_PS2_FractalSkyDefines.h for memory layout, etc.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Asura_PS2_FractalSkyDefines.h"
#include "vcl_sml.i"

// V C L   S E T U P /////////////////////////////////////////////////////////

.init_vf_all
.init_vi_all

.syntax new

.global VU1_FS_SETUP

// M A C R O S ///////////////////////////////////////////////////////////////

// These values are assumed in VU1_FS_INTERPOLATE.
#define INTERPOLATE_X (2)
#define INTERPOLATE_Y (1)
#define INTERPOLATE_Z (0)

// Switches pCur between buffers. Assumes pCur is already set to first_bfr or
// second_bfr and then switches it to the other. Also requires that first_bfr
// and second_bfr are integer constants and that second_bfr >= first_bfr.
// Uses a temp register called _pDelta.
.macro SwitchBfr pCur, first_bfr, second_bfr
    ISUBIU _pDelta, \pCur, \second_bfr
    // _pDelta is now either 0 if pCur was pSecond or -firstsize if pCur was
    // pFirst. Hence (pFirst - _pDelta) will give other buffer.
    IADDIU \pCur, VI00, \first_bfr
    ISUB   \pCur, \pCur, _pDelta
.endm

// Swaps the two specified integer registers. Uses a temporary register
// called _iTmp.
.macro Swap iA, iB
    IADDIU _iTmp, \iA, 0
    IADDIU \iA, \iB, 0
    IADDIU \iB, _iTmp, 0
.endm

// Linear interpolation along the vector A->B according to the value R, in
// range [0,1]. Note that this macro trashes ACC.
.macro Lerp fOut, fA, fB, fR, componentMask=xyzw
    ADDAx.\componentMask  ACC,   \fA, VF00 // = A
    MADDAx.\componentMask ACC,   \fB, \fR  // = A + (B * R)
    MSUBx.\componentMask  \fOut, \fA, \fR  // = A + (B * R) - (A * R) = (1 - R) * A + R * B
.endm

// Sets up mask for last three vertices' clip flags for a particular plane.
// Deals with case where immediate value requires more than 15 bits.
#define MAKE_PLANE_MASK_FOR_LAST_3_VERTS(plane_flag) (((plane_flag) << 12) | ((plane_flag) << 6) | (plane_flag))

// C O D E ///////////////////////////////////////////////////////////////////

    --enter
    --endenter

VU1_FS_SETUP:

    // Get data ptr.
    XTOP pIn
    // Setup initial output bfr ptr.
    IADDIU pCurOutputBfr, VI00, ASURA_PS2_FS__ADDR__OUTPUT_BFR_0
    // We'll also need a temp ptr to the other one for copying below.
    IADDIU pOtherOutputBfr, VI00, ASURA_PS2_FS__ADDR__OUTPUT_BFR_1
    // Load GIF tags and ST generation parameter QW.
    LQ     fSetupGIFTag, 0(pIn)
    LQ     fDrawGIFTag,  1(pIn)
    LQ.xyz fSTGenParams, 2(pIn)
    // Copy GIF tags to both output buffers. We'll leave these untouched
    // apart from to fix up the NLOOP.
    SQ fSetupGIFTag, ASURA_PS2_FS__OFFSET__OUTPUT_SETUP_GIF_TAG(pCurOutputBfr)
    SQ fDrawGIFTag,  ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    SQ fSetupGIFTag, ASURA_PS2_FS__OFFSET__OUTPUT_SETUP_GIF_TAG(pOtherOutputBfr)
    SQ fDrawGIFTag,  ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pOtherOutputBfr)
    // Read camera count.
    XITOP iNumCams
    // To keep VCL happy...
    FCSET 0

    --cont

VU1_FS_BEGIN_STRIP:

    // Reset cam data ptr.
    IADDIU pCamData, VI00, ASURA_PS2_FS__ADDR__CAM_DATA
    // Reset cam remaining count.
    IADDIU iCamsRemaining, iNumCams, 0

VU1_FS_NEXT_CAM:

    // Load transform parameters.
    MatrixLoad mV2GBC, ASURA_PS2_FS__OFFSET__CAM_V2GBC_MAT, pCamData
    LQ fGBC2SScale, ASURA_PS2_FS__OFFSET__CAM_GBC2S_SCALE(pCamData)
    LQ fGBC2STrans, ASURA_PS2_FS__OFFSET__CAM_GBC2S_TRANS(pCamData)
    // Get data ptr and vertex count. Obviously these are constant across all
    // cameras but it helps with register pressure to reload them.
    XTOP  pIn
    XITOP iVertexQty
    // Setup NLOOP+EOP in cur output bfr.
    IADDIU i0x8000, VI00, 0x7FFF
    IADDIU i0x8000, i0x8000, 1
    IOR iNLOOPandEOP, iVertexQty, i0x8000
    ISW.x iNLOOPandEOP, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    // Setup end-of-input ptr.
    IADD pInEnd, pIn,    iVertexQty
    IADD pInEnd, pInEnd, iVertexQty
    // Setup output ptrs.
    IADDIU pOut, pCurOutputBfr, ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX
    IADDIU pClipOut, VI00, ASURA_PS2_FS__ADDR__CLIP_INPUT_BFR
    // Reset any/all flags.
    IADDIU iAnyClipped, VI00, 0
    IADDIU iAllClipped, VI00, 0x3F

    VU1_FS_TRANSFORM_LOOP:

        --LoopCS  3, 0
        --LoopABS 21

        // Load worldspace vertex position and ST+ADC.
        LQ.xyz fWPos,     0(pIn)
        LQ.xy  fInST,     1(pIn)
        ILW.z  iInADC,    1(pIn)
        // Convert input data.
        ITOF12.xy fInST, fInST
        // Calculate pass 0 STs from the input ones.
        ADD.xy fPass0ST, fInST, fSTGenParams
        // Calculate GBC vtx pos and perform clip test.
        MatrixMultiplyVertexW1 fGBCPos, mV2GBC, fWPos
        CLIPw.xyz fGBCPos, fGBCPos
        // Scale and translate to screenspace.
        MULA  ACC,   fGBCPos,     fGBC2SScale
        MADDw fSPos, fGBC2STrans, fGBCPos[w]
        // Perspective divide.
        DIV Q, VF00[w], fSPos[w]
        MULq.xyz fSPos, fSPos, Q
        // Get clip result and OR on degeneracy flag.
        FCAND VI01, 0x3FFFF
        IADDIU iClipADC, VI01, 0x7FFF
        IOR iOutADC, iClipADC, iInADC
        // Update any/all flags.
        FCGET iClipFlags
        IAND iAllClipped, iAllClipped, iClipFlags
        IOR  iAnyClipped, iAnyClipped, iClipFlags
        // Store GBC pos and pass 0 ST ready for use by clipper, if we end
        // up having to run it over this strip.
        SQ    fGBCPos,  0(pClipOut)
        SQ.xy fPass0ST, 1(pClipOut)
        // In spare w component of clipper input STQ, store sum of iInADC and
        // iClipADC. This will be 0x8000 iff the vertex was rejected here
        // but is not degenerate.
        IADD  iSum, iInADC, iClipADC
        ISW.w iSum, 1(pClipOut)
        // Convert output data.
        FTOI4.xyz fSPos, fSPos        
        // Output.
        SQ.xy  fPass0ST, 0(pOut)
        SQ.xyz fSPos,    1(pOut)
        ISW.w  iOutADC,  1(pOut)
        // Update loop state.
        IADDIU pIn,      pIn,      2
        IADDIU pClipOut, pClipOut, 2
        IADDIU pOut,     pOut,     2
        // Done yet?
        IBNE pIn, pInEnd, VU1_FS_TRANSFORM_LOOP

    // If the entire strip was rejected then we're done.
    IBNE iAllClipped, VI00, VU1_FS_FINISHED_CUR_CAM
    // Render stuff that didn't get clipped.
    BAL pReturnAddress, VU1_FS_RENDER_ALL_PASSES
    // Any clipping required?
    IBEQ iAnyClipped, VI00, VU1_FS_FINISHED_CUR_CAM
    // Ugh, yes.
    B VU1_FS_CLIP_TRIANGLES

VU1_FS_FINISHED_CUR_CAM:
       
    // Reduce cams-remaining count and check there are more to.
    ISUBIU iCamsRemaining, iCamsRemaining, 1
    IBEQ iCamsRemaining, VI00, VU1_FS_FINISHED_STRIP
    // Update cam data ptr.
    IADDIU pCamData, pCamData, ASURA_PS2_FS__OFFSET__CAM_NEXT_CAM
    // Ok, next...
    B VU1_FS_NEXT_CAM

VU1_FS_FINISHED_STRIP:

    --cont
    B VU1_FS_BEGIN_STRIP

VU1_FS_RENDER_ALL_PASSES:

    // Load SCISSOR and XYOFFSET settings for cur cam.
    LQ fSCISSOR,  ASURA_PS2_FS__OFFSET__CAM_SCISSOR(pCamData)
    LQ fXYOFFSET, ASURA_PS2_FS__OFFSET__CAM_XYOFFSET(pCamData)

VU1_FS_RENDER_PASS_0:

    // Copy SCISSOR and XYOFFSET to output buffer.
    SQ fSCISSOR,  ASURA_PS2_FS__OFFSET__OUTPUT_SCISSOR(pCurOutputBfr)
    SQ fXYOFFSET, ASURA_PS2_FS__OFFSET__OUTPUT_XYOFFSET(pCurOutputBfr)
    // Load pass 0 ALPHA and copy to output buffer.
    LQ fPass0ALPHA, (ASURA_PS2_FS__ADDR__RENDERSTATE + ASURA_PS2_FS__OFFSET__RS_ALPHA_PASS_0)(VI00)
    SQ fPass0ALPHA, ASURA_PS2_FS__OFFSET__OUTPUT_ALPHA(pCurOutputBfr)
    // Send to GS.
    XGKICK pCurOutputBfr
    
VU1_FS_RENDER_PASS_1:

    // Switch output buffers. Keep a ptr to the old one since we'll need to
    // read from it.
    IADDIU pOtherOutputBfr, pCurOutputBfr, 0
    SwitchBfr pCurOutputBfr, ASURA_PS2_FS__ADDR__OUTPUT_BFR_0, ASURA_PS2_FS__ADDR__OUTPUT_BFR_1
    // Copy SCISSOR and XYOFFSET to our new buffer.
    SQ fSCISSOR,  ASURA_PS2_FS__OFFSET__OUTPUT_SCISSOR(pCurOutputBfr)
    SQ fXYOFFSET, ASURA_PS2_FS__OFFSET__OUTPUT_XYOFFSET(pCurOutputBfr)
    // Load pass 1 ALPHA and copy.
    LQ fPass1ALPHA, (ASURA_PS2_FS__ADDR__RENDERSTATE + ASURA_PS2_FS__OFFSET__RS_ALPHA_PASS_1)(VI00)
    SQ fPass1ALPHA, ASURA_PS2_FS__OFFSET__OUTPUT_ALPHA(pCurOutputBfr)
    // Grab NLOOP+EOP from other buffer's draw GIF tag and copy that as well.
    ILW.x iNLOOPandEOP, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pOtherOutputBfr)
    ISW.x iNLOOPandEOP, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    // Setup input and output ptrs.
    IADDIU pIn,  pOtherOutputBfr, ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX
    IADDIU pOut, pCurOutputBfr,   ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX
    // Extract NLOOP from NLOOP+EOP.
    IADDIU iMask, VI00, 0x7FFF
    IAND iNLOOP, iNLOOPandEOP, iMask
    // Setup end-of-input ptr.
    IADD pInEnd, pIn,    iNLOOP
    IADD pInEnd, pInEnd, iNLOOP
    
    VU1_FS_RENDER_PASS_1_ST_OVERWRITE_LOOP:

        --LoopCS  3, 0
        --LoopABS 7

        // Read pass 0 vertex data.
        LQ.xy fPass0ST,   0(pIn)
        LQ    fPosAndADC, 1(pIn)
        // Calculate pass 1 ST.
        SUB.xy fInST, fPass0ST, fSTGenParams
        ADDAy.xy ACC, VF00, fSTGenParams       // ( -p1, -p1 )
        MADDz.xy fPass1ST, fInST, fSTGenParams // ( 0.9s - p1, 0.9t - p1 )
        // Output.
        SQ.xy fPass1ST,   0(pOut)
        SQ    fPosAndADC, 1(pOut)
        // Update loop state.
        IADDIU pIn,  pIn,  2
        IADDIU pOut, pOut, 2
        // Done yet?
        IBNE pIn, pInEnd, VU1_FS_RENDER_PASS_1_ST_OVERWRITE_LOOP

    // Send to GS.
    XGKICK pCurOutputBfr

VU1_FS_RENDER_PASS_2:

    // There's little point in switching output buffer because we'll just be
    // going back again in a minute. We've also already got NLOOP, SCISSOR
    // and XYOFFSET settings and all the XYZs in this buffer. We just need to
    // use the right ALPHA setting and fix up the STs. First, copy the ALPHA.
    LQ fPass2ALPHA, (ASURA_PS2_FS__ADDR__RENDERSTATE + ASURA_PS2_FS__OFFSET__RS_ALPHA_PASS_2)(VI00)
    SQ fPass2ALPHA, ASURA_PS2_FS__OFFSET__OUTPUT_ALPHA(pOtherOutputBfr)
    // Input and output ptrs are going to be equal. Set one up.
    IADDIU pInOut, pOtherOutputBfr, ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX
    // Setup end-of-input ptr.
    IADD pInOutEnd, pInOut,    iNLOOP
    IADD pInOutEnd, pInOutEnd, iNLOOP

    VU1_FS_RENDER_PASS_2_ST_OVERWRITE_LOOP:

        --LoopCS  3, 0
        --LoopABS 5

        // Read pass 0 ST.
        LQ.xy fPass0ST, 0(pInOut)
        // Calculate pass 2 ST.
        SUB.xy fInST, fPass0ST, fSTGenParams
        SUBy.xy fPass2ST, fInST, fSTGenParams
        // Overwrite.
        SQ.xy fPass2ST, 0(pInOut)
        // Update loop state.
        IADDIU pInOut, pInOut, 2
        // Done yet?
        IBNE pInOut, pInOutEnd, VU1_FS_RENDER_PASS_2_ST_OVERWRITE_LOOP

    // Send to GS.
    XGKICK pOtherOutputBfr

VU1_FS_RENDERED_ALL_PASSES:

    // Ok, pCurOutputBfr is still correct from when we did pass 1 so there's
    // no need to switch that back. And so we're done.
    JR pReturnAddress

VU1_FS_CLIP_FINISHED:

    // Check if there's anything in the output buffer that needs rendering.
    // We could just kick the packet anyway but we don't want to do all the
    // multipass stuff if there's nothing in the buffer.
    ILW.x iNLOOPandEOP, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    IADDIU iMask, VI00, 0x7FFF
    IAND iNLOOP, iNLOOPandEOP, iMask
    IBEQ iNLOOP, VI00, VU1_FS_FINISHED_CUR_CAM
    // Ok, we have something to render.
    BAL pReturnAddress, VU1_FS_RENDER_ALL_PASSES
    // And start on the next camera.
    B VU1_FS_FINISHED_CUR_CAM

VU1_FS_CLIP_TRIANGLES:

    // Clipper input data format:
    //
    //        x    y    z    w                  -- in GBC space
    //        s    t    X    should clip flag
    //
    // where X is "don't care".

    // Load GBC->SC scale vector.
    LQ fGBC2SCScale, ASURA_PS2_FS__OFFSET__CAM_GBC2SC_SCALE(pCamData)
    // Setup output pointer.
    IADDIU pOut, pCurOutputBfr, ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX
    // Push onto stack.
    ISW.x pOut, ASURA_PS2_FS__ADDR__VI_STACK_1(VI00)
    // Setup some stuff for loop.
    IADDIU pIn,        VI00,      ASURA_PS2_FS__ADDR__CLIP_INPUT_BFR
    IADD   pInEnd,     pIn,       iVertexQty
    IADD   pInEnd,     pInEnd,    iVertexQty
    IADDIU iClipMask,  VI00,      0x3F
    IADDIU i0x8000,    iClipMask, (0x8000 - 0x3F)
    // Write a GIF tag to new output buffer (with EOP=1 and NLOOP=0.)
    ISW.x i0x8000, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    // Get clipping results for vertex (n - 2) - but against our smaller
    // screen frustum, not the guardband one.
    LQ        fClipPos, 0(pIn)
    MUL       fClipPos, fClipPos, fGBC2SCScale
    CLIPw.xyz fClipPos, fClipPos
    FCGET     iClipFlag2
    IAND      iClipFlag2, iClipFlag2, iClipMask
    // Same for vertex (n - 1)...
    LQ        fClipPos, 2(pIn)
    MUL       fClipPos, fClipPos, fGBC2SCScale
    CLIPw.xyz fClipPos, fClipPos
    FCGET     iClipFlag1
    IAND      iClipFlag1, iClipFlag1, iClipMask
    // Skip ahead to last vertex of first tri.
    IADDIU pIn,  pIn, 4
    // Calculate half of first clipping test.
    IAND iClipAll21, iClipFlag2, iClipFlag1

    VU1_FS_FIND_CLIP_TRIANGLE_LOOP:

        --LoopCS 1, 0

        // Have we finished yet?
        IBEQ pIn, pInEnd, VU1_FS_CLIP_FINISHED
        // Get clipping results for current vertex - but against our smaller
        // screen frustum, not the guardband one.
        LQ        fClipPos, 0(pIn)
        MUL       fClipPos, fClipPos, fGBC2SCScale
        CLIPw.xyz fClipPos, fClipPos
        FCGET     iClipFlag0
        IAND      iClipFlag0, iClipFlag0, iClipMask
        // Calculate second half of clipping test.
        IAND iClipAll10, iClipFlag1, iClipFlag0
        IAND iClipAll,   iClipAll21, iClipAll10
        // Move clipping flags/result along 1.
        IADD iClipFlag1, VI00, iClipFlag0
        IADD iClipAll21, VI00, iClipAll10
        // Update pointers.
        IADDIU pIn, pIn, 2
        // Don't bother clipping if vertex was degenerate or if we accepted it
        // in the main render loop. We have to check the latter since we're
        // using a smaller frustum here and may end up scissoring tris that
        // actually fit inside the larger one.
        ILW.w iShouldClipFlag, -1(pIn)
        IBNE  iShouldClipFlag, i0x8000, VU1_FS_FIND_CLIP_TRIANGLE_LOOP
        // If the triangle is completely outside at least one plane then
        // just reject.
        IBNE iClipAll, VI00, VU1_FS_FIND_CLIP_TRIANGLE_LOOP
        // At this point, we know the triangle is not completely inside all
        // planes since otherwise it would have been accepted in the main
        // render loop. We now also know it's not entirely outside one plane;
        // hence, it must cross one and require scissoring.

--barrier
    // Ok, this triangle needs clipping. Push some int registers onto stack.
    ISW.x pIn,        ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    ISW.y pInEnd,     ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    ISW.z iClipFlag1, ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    ISW.w iClipAll21, ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    // Setup buffers and initial vtx count.
    IADDIU pClipSrc,     pCurOutputBfr, ASURA_PS2_FS__OFFSET__CLIP_WORKSPACE
    IADDIU pClipDst,     pClipSrc,      ASURA_PS2_FS__SIZE__CLIP_WORKSPACE
    IADDIU iSrcPointQty, VI00,          3
    // Copy the source vertices into the first clip workspace. We have to do
    // this because we have split input data
    LQ    fClipPos, -6(pIn)
    LQ.xy fSTQ,     -5(pIn)
    SQ    fClipPos,  0(pClipSrc)
    SQ.xy fSTQ,      1(pClipSrc)
    LQ    fClipPos, -4(pIn)
    LQ.xy fSTQ,     -3(pIn)
    SQ    fClipPos,  2(pClipSrc)
    SQ.xy fSTQ,      3(pClipSrc)
    LQ    fClipPos, -2(pIn)
    LQ.xy fSTQ,     -1(pIn)
    SQ    fClipPos,  4(pClipSrc)
    SQ.xy fSTQ,      5(pClipSrc)
#if NORMALISE_CLIPPED_STS
    // Initialise min and max STs.
    LOI 0x49742400 // Arbitrary large float.
    ADDi.xy fMinSTs, VF00, I
    SUBi.xy fMaxSTs, VF00, I
#endif

    // See which planes we need to clip against. We don't want to bother
    // clipping against a plane unless at least one vertex from the source
    // tri lies on or outside it.
    // -z
    FCAND VI01, MAKE_PLANE_MASK_FOR_LAST_3_VERTS(ASURA_PS2_FS__CLIP_NEAR)
    IADD iPlaneClipFlags, VI00, VI01
    // +z
    IADD iPlaneClipFlags, iPlaneClipFlags, iPlaneClipFlags                   // Shift up.
    FCAND VI01, MAKE_PLANE_MASK_FOR_LAST_3_VERTS(ASURA_PS2_FS__CLIP_FAR)    // Get flag for next plane.
    IADD iPlaneClipFlags, iPlaneClipFlags, VI01                           // OR onto cumulative flags.
    // -y
    IADD iPlaneClipFlags, iPlaneClipFlags, iPlaneClipFlags
    FCAND VI01, MAKE_PLANE_MASK_FOR_LAST_3_VERTS(ASURA_PS2_FS__CLIP_TOP)    //       "
    IADD iPlaneClipFlags, iPlaneClipFlags, VI01
    // +y
    IADD iPlaneClipFlags, iPlaneClipFlags, iPlaneClipFlags
    FCAND VI01, MAKE_PLANE_MASK_FOR_LAST_3_VERTS(ASURA_PS2_FS__CLIP_BOTTOM) //       "
    IADD iPlaneClipFlags, iPlaneClipFlags, VI01
    // -x
    IADD iPlaneClipFlags, iPlaneClipFlags, iPlaneClipFlags
    FCAND VI01, MAKE_PLANE_MASK_FOR_LAST_3_VERTS(ASURA_PS2_FS__CLIP_LEFT)   //       "
    IADD iPlaneClipFlags, iPlaneClipFlags, VI01
    // +x
    IADD iPlaneClipFlags, iPlaneClipFlags, iPlaneClipFlags
    FCAND VI01, MAKE_PLANE_MASK_FOR_LAST_3_VERTS(ASURA_PS2_FS__CLIP_RIGHT)  //       "
    IADD iPlaneClipFlags, iPlaneClipFlags, VI01

VU1_FS_CLIP_NEAR_PLANE: // -z

    // Setup clip mask.
    IADDIU iCurrentClipMask, VI00, ASURA_PS2_FS__CLIP_NEAR
    // Skip if no vertices are on or outside this plane.
    IAND iAnyOnOrOutside, iPlaneClipFlags, iCurrentClipMask
    IBEQ iAnyOnOrOutside, VI00, VU1_FS_CLIP_FAR_PLANE
    // Clear output vertex count and set plane sign to negative.
    SUBw.xw fPlaneSign, VF00, VF00
    // Clip against this plane.
    IADDIU iInterpolate, VI00, INTERPOLATE_Z
    BAL pRtA, VU1_FS_CLIP_POINTS
    // Switch buffers.
    Swap pClipSrc, pClipDst
    // Extract output vertex count.
    FTOI0.w fPlaneSign, fPlaneSign
    MTIR iSrcPointQty, fPlaneSign[w]
    // Check tri hasn't been entirely clipped out. While we try to reject
    // such cases in the find-clip-tri loop above, I think numerical
    // inaccuracies can lead to it still happening.
    IBEQ iSrcPointQty, VI00, VU1_FS_CLIP_FINISHED_THIS_TRIANGLE
    // If we've clipped to the near plane, force x and y clips. I think the
    // vertex being behind the near plane somehow invalidates our results
    // above. Ought to double-check the maths behind this to make sure that's
    // actually the case and that this isn't just hiding a bug somewhere.
    IADDIU iXandYClips, VI00, (ASURA_PS2_FS__CLIP_LEFT | ASURA_PS2_FS__CLIP_RIGHT | ASURA_PS2_FS__CLIP_TOP | ASURA_PS2_FS__CLIP_BOTTOM)
    IOR iPlaneClipFlags, iPlaneClipFlags, iXandYClips

VU1_FS_CLIP_FAR_PLANE: // +z

    // Setup clip mask.
    IADDIU iCurrentClipMask, VI00, ASURA_PS2_FS__CLIP_FAR
    // Skip if no vertices are on or outside this plane.
    IAND iAnyOnOrOutside, iPlaneClipFlags, iCurrentClipMask
    IBEQ iAnyOnOrOutside, VI00, VU1_FS_CLIP_TOP_PLANE
    // Clear output vertex count and set plane sign to positive.
    SUBw.w fPlaneSign, VF00, VF00
    ADDw.x fPlaneSign, VF00, VF00
    // Clip against this plane.
    IADDIU iInterpolate, VI00, INTERPOLATE_Z
    BAL pRtA, VU1_FS_CLIP_POINTS
    // Switch buffers.
    Swap pClipSrc, pClipDst
    // Extract output vertex count.
    FTOI0.w fPlaneSign, fPlaneSign
    MTIR iSrcPointQty, fPlaneSign[w]
    // Check tri hasn't been entirely clipped out. (See also comment in
    // VU1_FS_CLIP_NEAR_PLANE.)
    IBEQ iSrcPointQty, VI00, VU1_FS_CLIP_FINISHED_THIS_TRIANGLE

VU1_FS_CLIP_TOP_PLANE: // -y

    // Setup clip mask.
    IADDIU iCurrentClipMask, VI00, ASURA_PS2_FS__CLIP_TOP
    // Skip if no vertices are on or outside this plane.
    IAND iAnyOnOrOutside, iPlaneClipFlags, iCurrentClipMask
    IBEQ iAnyOnOrOutside, VI00, VU1_FS_CLIP_BOTTOM_PLANE
    // Clear output vertex count and set plane sign to negative.
    SUBw.xw fPlaneSign, VF00, VF00
    // Clip against this plane.
    IADDIU iInterpolate, VI00, INTERPOLATE_Y
    BAL pRtA, VU1_FS_CLIP_POINTS
    // Switch buffers.
    Swap pClipSrc, pClipDst
    // Extract output vertex count.
    FTOI0.w fPlaneSign, fPlaneSign
    MTIR iSrcPointQty, fPlaneSign[w]
    // Check tri hasn't been entirely clipped out. (See also comment in
    // VU1_FS_CLIP_NEAR_PLANE.)
    IBEQ iSrcPointQty, VI00, VU1_FS_CLIP_FINISHED_THIS_TRIANGLE

VU1_FS_CLIP_BOTTOM_PLANE: // +y

    // Setup clip mask.
    IADDIU iCurrentClipMask, VI00, ASURA_PS2_FS__CLIP_BOTTOM
    // Skip if no vertices are on or outside this plane.
    IAND iAnyOnOrOutside, iPlaneClipFlags, iCurrentClipMask
    IBEQ iAnyOnOrOutside, VI00, VU1_FS_CLIP_LEFT_PLANE
    // Clear output vertex count and set plane sign to positive.
    SUBw.w fPlaneSign, VF00, VF00
    ADDw.x fPlaneSign, VF00, VF00
    // Clip against this plane.
    IADDIU iInterpolate, VI00, INTERPOLATE_Y
    BAL pRtA, VU1_FS_CLIP_POINTS
    // Switch buffers.
    Swap pClipSrc, pClipDst
    // Extract output vertex count.
    FTOI0.w fPlaneSign, fPlaneSign
    MTIR iSrcPointQty, fPlaneSign[w]
    // Check tri hasn't been entirely clipped out. (See also comment in
    // VU1_FS_CLIP_NEAR_PLANE.)
    IBEQ iSrcPointQty, VI00, VU1_FS_CLIP_FINISHED_THIS_TRIANGLE

VU1_FS_CLIP_LEFT_PLANE: // -x

    // Setup clip mask.
    IADDIU iCurrentClipMask, VI00, ASURA_PS2_FS__CLIP_LEFT
    // Skip if no vertices are on or outside this plane.
    IAND iAnyOnOrOutside, iPlaneClipFlags, iCurrentClipMask
    IBEQ iAnyOnOrOutside, VI00, VU1_FS_CLIP_RIGHT_PLANE
    // Clear output vertex count and set plane sign to negative.
    SUBw.xw fPlaneSign, VF00, VF00
    // Clip against this plane.
    IADDIU iInterpolate, VI00, INTERPOLATE_X
    BAL pRtA, VU1_FS_CLIP_POINTS
    // Switch buffers.
    Swap pClipSrc, pClipDst
    // Extract output vertex count.
    FTOI0.w fPlaneSign, fPlaneSign
    MTIR iSrcPointQty, fPlaneSign[w]
    // Check tri hasn't been entirely clipped out. (See also comment in
    // VU1_FS_CLIP_NEAR_PLANE.)
    IBEQ iSrcPointQty, VI00, VU1_FS_CLIP_FINISHED_THIS_TRIANGLE

VU1_FS_CLIP_RIGHT_PLANE: // +x

    // Setup clip mask.
    IADDIU iCurrentClipMask, VI00, ASURA_PS2_FS__CLIP_RIGHT
    // Skip if no vertices are on or outside this plane.
    IAND iAnyOnOrOutside, iPlaneClipFlags, iCurrentClipMask
    IBEQ iAnyOnOrOutside, VI00, VU1_FS_CLIP_FINISHED_PLANES
    // Clear output vertex count and set plane sign to positive.
    SUBw.w fPlaneSign, VF00, VF00
    ADDw.x fPlaneSign, VF00, VF00
    // Clip against this plane.
    IADDIU iInterpolate, VI00, INTERPOLATE_X
    BAL pRtA, VU1_FS_CLIP_POINTS
    // Switch buffers.
    Swap pClipSrc, pClipDst
    // Extract output vertex count.
    FTOI0.w fPlaneSign, fPlaneSign
    MTIR iSrcPointQty, fPlaneSign[w]
    // Check tri hasn't been entirely clipped out. (See also comment in
    // VU1_FS_CLIP_NEAR_PLANE.)
    IBEQ iSrcPointQty, VI00, VU1_FS_CLIP_FINISHED_THIS_TRIANGLE

VU1_FS_CLIP_FINISHED_PLANES:

    // Retrieve int registers from stack.
    ILW.x pOut, ASURA_PS2_FS__ADDR__VI_STACK_1(VI00)
    // Check there is enough space in the output buffer for this many
    // vertices.
    ISUB   iTemp, pOut,  pCurOutputBfr                        // space already used
    IADD   iTemp, iTemp, iSrcPointQty                         // + half space required for this strip
    IADD   iTemp, iTemp, iSrcPointQty                         // + half space required for this strip
    IADDIU iTemp, iTemp, 2                                    // + space for overrun (see comment by VU1_FS_CLIP_OUTPUT_STRIP_LOOP below)
    ISUBIU iTemp, iTemp, ASURA_PS2_FS__SIZE__CLIP_OUTPUT_BFR  // => iTemp < 0 iff there is space for these vertices
    IBLTZ  iTemp, VU1_FS_CLIP_ENOUGH_SPACE_IN_OUTPUT
    // Not enough space so kick what we have and swap buffers.
    BAL pReturnAddress, VU1_FS_RENDER_ALL_PASSES
    // Reset output pointers.
    IADDIU pOut, pCurOutputBfr, ASURA_PS2_FS__OFFSET__OUTPUT_FIRST_VERTEX
    // Write new GIF tag to output buffer with EOP=1 and NLOOP=0.
    IADDIU iNLOOPandEOP, VI00,         0x7FFF
    IADDIU iNLOOPandEOP, iNLOOPandEOP, 1
    ISW.x  iNLOOPandEOP, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    
VU1_FS_CLIP_ENOUGH_SPACE_IN_OUTPUT:

    // Increment NLOOP count in GIF tag.
    ILW.x iTemp, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    IADD  iTemp, iTemp, iSrcPointQty
    ISW.x iTemp, ASURA_PS2_FS__OFFSET__OUTPUT_DRAW_GIF_TAG(pCurOutputBfr)
    // Setup end-of-input pointer.
    IADD   pClipSrcE, pClipSrc,  iSrcPointQty
    IADD   pClipSrcE, pClipSrcE, iSrcPointQty
    ISUBIU pClipSrcE, pClipSrcE, 2            // Point to last vertex in fan.
    // Set degeneracy flag for first two output verts. We'll add this value
    // to itself each time; that'll give us 0xC000, 0x8000, 0x0000 (and
    // obviously 0s thereafter), which means the ADC bit is on for precisely
    // the first two vertices.
    IADDIU iADC, VI00, 0x6000
#if NORMALISE_CLIPPED_STS
    // Calc ST normalising offset.
    ADD.xy   fNormST, fMinSTs, fMaxSTs
    MULi.xy  fNormST, fNormST, I
    FTOI0.xy fNormST, fNormST
    ITOF0.xy fNormST, fNormST
#endif

VU1_FS_CLIP_OUTPUT_STRIP:

    VU1_FS_CLIP_OUTPUT_STRIP_LOOP:

        --LoopCS  2, 0
        --LoopAbs 20

        // (MP - 23.10.2003) We'll always write a multiple of two vertices
        // here to save branching in the loop. This then allows VCL to
        // optimise it for us. Of course, this means we'll overrun by up to
        // one vertex so we have to account for this when checking for free
        // space above. The GIF tag's NLOOP won't count this extra vertex but
        // we'll have to fix up the output ptr after the loop.

        // Load vertex from the start.
        LQ     fGBCPos, 0(pClipSrc)
        LQ.xy  fST,     1(pClipSrc)
        // Scale and translate to screenspace.// (MP - 19.11.2003) 
        MULA  ACC,   fGBCPos,     fGBC2SScale
        MADDw fSPos, fGBC2STrans, fGBCPos[w]
        // Calculate 1/w.
        DIV Q, VF00[w], fSPos[w]
#if NORMALISE_CLIPPED_STS
        // Normalise ST.
        SUB.xy fST, fST, fNormST
#endif
        // Perspective divide.
        MULq.xyz fSPos, fSPos, Q
        // Convert output data.
        FTOI4.xyz fSPos, fSPos
        // Store.
        SQ.xy  fST,   0(pOut)
        SQ.xyz fSPos, 1(pOut)
        // Set the ADC bit.
        IADD  iADC, iADC, iADC
        ISW.w iADC, 1(pOut)
        // Update loop state.
        IADDIU pClipSrc, pClipSrc, 2
        IADDIU pOut,     pOut,     2
        // Load vertex from the end.
        LQ     fGBCPos, 0(pClipSrcE)
        LQ.xy  fST,     1(pClipSrcE)
        // Scale and translate to screenspace.
        MULA  ACC,   fGBCPos,     fGBC2SScale
        MADDw fSPos, fGBC2STrans, fGBCPos[w]
        // Calculate 1/w.
        DIV Q, VF00[w], fSPos[w]
#if NORMALISE_CLIPPED_STS
        // Normalise ST.
        SUB.xy fST, fST, fNormST
#endif
        // Perspective divide.
        MULq.xyz fSPos, fSPos, Q
        // Convert output data.
        FTOI4.xyz fSPos, fSPos
        // Store.
        SQ.xy  fST,   0(pOut)
        SQ.xyz fSPos, 1(pOut)
        // Set the ADC bit.
        IADD  iADC, iADC, iADC
        ISW.w iADC, 1(pOut)
        // Update loop state.
        ISUBIU pClipSrcE, pClipSrcE, 2
        IADDIU pOut,      pOut,      2
        // Done yet?
        ISUB  iRemaining, pClipSrcE, pClipSrc
        IBGEZ iRemaining, VU1_FS_CLIP_OUTPUT_STRIP_LOOP

VU1_FS_CLIP_FINISHED_OUTPUT_THIS_TRIANGLE:

    // Did we have an odd number of vertices?
    IADDIU iMask, VI00,         0x1
    IAND   iOdd,  iSrcPointQty, iMask
    IBEQ   iOdd,  VI00,         VU1_FS_CLIP_SKIP_OUTPUT_PTR_FIX_UP
    // Yes. We need to correct the output ptr.
    ISUBIU pOut, pOut, 2

VU1_FS_CLIP_SKIP_OUTPUT_PTR_FIX_UP:

    // Push int register back onto stack.
    ISW.x pOut, ASURA_PS2_FS__ADDR__VI_STACK_1(VI00)

VU1_FS_CLIP_FINISHED_THIS_TRIANGLE:

    // Retrieve int registers from stack.
    ILW.x pIn,        ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    ILW.y pInEnd,     ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    ILW.z iClipFlag1, ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    ILW.w iClipAll21, ASURA_PS2_FS__ADDR__VI_STACK_0(VI00)
    // Recreate some constant values so VCL doesn't have to preserve them for
    // us.
    IADDIU iClipMask, VI00,      0x3F
    IADDIU i0x8000,   iClipMask, (0x8000 - 0x3F)
    // Carry out previous 2 clips again so the find triangle loop has the
    // correct clip history.
    LQ        fClipPos, -4(pIn)
    MUL       fClipPos, fClipPos, fGBC2SCScale
    CLIPw.xyz fClipPos, fClipPos
    LQ        fClipPos, -2(pIn)
    MUL       fClipPos, fClipPos, fGBC2SCScale
    CLIPw.xyz fClipPos, fClipPos
    // And look for further tris that require scissoring.
    B VU1_FS_FIND_CLIP_TRIANGLE_LOOP

--barrier

VU1_FS_CLIP_POINTS:

    // (MP - 23.10.2003) We work in guardband clip space in general. However,
    // our clipping logic works in the screen clip space, so we don't end up
    // sending large bits of polys that are outside the screen bounds. Hence
    // the following code keeps track of the current and next positions in
    // both spaces.

    // Keep track of the start of the input buffer so we can add the first
    // vertex to the end of the list when we're done.
    IADDIU pClipSrcStart, pClipSrc, 0
    // Setup an output ptr.
    IADDIU pClipOut, pClipDst, 0
    // Load the first vertex from the buffer.
    LQ    fNGBCVert, 0(pClipSrc)
    LQ.xy fNST,      1(pClipSrc)
    IADDIU pClipSrc, pClipSrc, 2
#if NORMALISE_CLIPPED_STS
    // Track min and max STs so we can normalise later.
    MINI.xy fMinSTs, fMinSTs, fNST
    MAX.xy  fMaxSTs, fMaxSTs, fNST
#endif
    // Check if this first one is in or out.
    MUL       fNSCVert, fNGBCVert, fGBC2SCScale
    CLIPw.xyz fNSCVert, fNSCVert[w]
    FCGET     iTemp
    IAND      iTemp, iTemp, iCurrentClipMask
    IBEQ      iTemp, VI00, VU1_FS_CLIP_CURRENT_IN

VU1_FS_CLIP_CURRENT_OUT:

    // Done yet?
    ISUBIU iSrcPointQty, iSrcPointQty, 1
    IBEQ   iSrcPointQty, VI00, VU1_FS_CLIP_POINTS_END
    // Move next point to current point.
    MOVE    fCGBCVert, fNGBCVert
    MOVE    fCSCVert,  fNSCVert
    MOVE.xy fCST,      fNST
    // And load a new point.
    LQ    fNGBCVert, 0(pClipSrc)
    LQ.xy fNST,      1(pClipSrc)
    IADDIU pClipSrc, pClipSrc, 2
#if NORMALISE_CLIPPED_STS
    // Track min and max STs so we can normalise later.
    MINI.xy fMinSTs, fMinSTs, fNST
    MAX.xy  fMaxSTs, fMaxSTs, fNST
#endif
    // Check if this new vertex is in or out.
    MUL       fNSCVert, fNGBCVert, fGBC2SCScale
    CLIPw.xyz fNSCVert, fNSCVert[w]
    FCGET     iTemp
    IAND      iTemp, iTemp, iCurrentClipMask
    IBEQ      iTemp, VI00, VU1_FS_CLIP_OUT_IN

VU1_FS_CLIP_OUT_OUT:

    // Just carry on to next point.
    B VU1_FS_CLIP_CURRENT_OUT

VU1_FS_CLIP_OUT_IN:

    // Add 1 to output point count.
    ADD.w fPlaneSign, fPlaneSign, VF00
    // Interpolate new point on clip plane.
    BAL pRtA2, VU1_FS_INTERPOLATE
    // Ok, on to next point.
    B VU1_FS_CLIP_CURRENT_IN

VU1_FS_CLIP_CURRENT_IN:
    
    // Done yet?
    ISUBIU iSrcPointQty, iSrcPointQty, 1
    IBEQ   iSrcPointQty, VI00, VU1_FS_CLIP_POINTS_END
    // Move next point to current point.
    MOVE    fCGBCVert, fNGBCVert
    MOVE    fCSCVert,  fNSCVert
    MOVE.xy fCST,      fNST
    // And load a new point.
    LQ    fNGBCVert, 0(pClipSrc)
    LQ.xy fNST,      1(pClipSrc)
    IADDIU pClipSrc, pClipSrc, 2
#if NORMALISE_CLIPPED_STS
    // Track min and max STs so we can normalise later.
    MINI.xy fMinSTs, fMinSTs, fNST
    MAX.xy  fMaxSTs, fMaxSTs, fNST
#endif
    // Check if this new vertex is in or out.
    MUL       fNSCVert, fNGBCVert, fGBC2SCScale
    CLIPw.xyz fNSCVert, fNSCVert[w]
    FCGET     iTemp
    IAND      iTemp, iTemp, iCurrentClipMask
    IBNE      iTemp, VI00, VU1_FS_CLIP_IN_OUT

VU1_FS_CLIP_IN_IN:

    // Add 1 to output point count.
    ADD.w fPlaneSign, fPlaneSign, VF00
    // Copy current to output.
    SQ    fCGBCVert, 0(pClipOut)
    SQ.xy fCST,      1(pClipOut)
    IADDIU pClipOut, pClipOut, 2
    // And on to next vertex.
    B VU1_FS_CLIP_CURRENT_IN

VU1_FS_CLIP_IN_OUT:

    // Add 2 to output point count.
    ADD.w fPlaneSign, fPlaneSign, VF00
    ADD.w fPlaneSign, fPlaneSign, VF00
    // Copy current to output.
    SQ    fCGBCVert, 0(pClipOut)
    SQ.xy fCST,      1(pClipOut)
    IADDIU pClipOut, pClipOut, 2
    // Interpolate a new point on the clip plane.
    BAL pRtA2, VU1_FS_INTERPOLATE
    // And on to next vertex.
    B VU1_FS_CLIP_CURRENT_OUT

VU1_FS_CLIP_POINTS_END:

    // Check if we've still got to do the final line, ie. last -> first.
    IBNE pClipSrcStart, VI00, VU1_FS_CLIP_POINT_LAST_SEGMENT
    // Ok, we're finished.
    JR pRtA

VU1_FS_CLIP_POINT_LAST_SEGMENT:

    // Get the start of the input buffer and clear the input buffer pointer
    // so we know we've finished when we get to VU1_FS_CLIP_POINTS_END again.
    IADDIU pClipSrc,      pClipSrcStart, 0
    IADDIU pClipSrcStart, VI00,          0
    // One final line segment to do.
    IADDIU iSrcPointQty, iSrcPointQty, 2
    // Check if the last clip was in or out.
    FCGET iTemp
    IAND  iTemp, iTemp, iCurrentClipMask
    IBEQ  iTemp, VI00, VU1_FS_CLIP_CURRENT_IN
    B VU1_FS_CLIP_CURRENT_OUT

--barrier

    //       |-------L-------|
    //      C ----------|---- N
    //                  P
    //       |----A-----l--B-|
    //                  a
    //                  n
    //                  e
    //
    // We need to find the ratio R = |A/L|

    // (MP - 27.10.2003) NB: We calculate R from the screenclipspace
    // positions but actually interpolate the guardbandclipspace ones. We
    // need to calculate R in that space since the clip logic is done there.

VU1_FS_INTERPOLATE:

    IBEQ   iInterpolate, VI00, VU1_FS_INTERPOLATE_Z
    ISUBIU iTemp, iInterpolate, 1
    IBEQ   iTemp, VI00, VU1_FS_INTERPOLATE_Y

VU1_FS_INTERPOLATE_X:

    // Get distance from clip plane to current, taking into account +/- clip
    // plane. This is -A in the above diagram.
    ADDAx.x ACC, VF00, fCSCVert
    MSUBw.x fC2Plane, fPlaneSign, fCSCVert  // -A = current.x - plane sign * current.w
    // Get distance from next to clip plane, again taking into account plane
    // sign. This is B above.
    ADDAx.x ACC, VF00, fNSCVert
    MSUBw.x fN2Plane, fPlaneSign, fNSCVert  // B = next.x - plane sign * next.w
    // And use these values to interpolate a new point.
    B VU1_FS_INTERPOLATE_COMMON

VU1_FS_INTERPOLATE_Y:

    // Get distance from clip plane to current, taking into account +/- clip
    // plane. This is -A in the above diagram.
    ADDAy.x ACC, VF00, fCSCVert
    MSUBw.x fC2Plane, fPlaneSign, fCSCVert  // -A = current.y - plane sign * current.w
    // Get distance from next to clip plane, again taking into account plane
    // sign. This is B above.
    ADDAy.x ACC, VF00, fNSCVert
    MSUBw.x fN2Plane, fPlaneSign, fNSCVert  // B = next.y - plane sign * next.w
    // And use these values to interpolate a new point.
    B VU1_FS_INTERPOLATE_COMMON

VU1_FS_INTERPOLATE_Z:

    // Get distance from clip plane to current, taking into account +/- clip
    // plane. This is -A in the above diagram.
    ADDAz.x ACC, VF00, fCSCVert
    MSUBw.x fC2Plane, fPlaneSign, fCSCVert  // -A = current.z - plane sign * current.w
    // Get distance from next to clip plane, again taking into account plane
    // sign. This is B above.
    ADDAz.x ACC, VF00, fNSCVert
    MSUBw.x fN2Plane, fPlaneSign, fNSCVert  // B = next.z - plane sign * next.w
    // And use these values to interpolate a new point.
//  B VU1_FS_INTERPOLATE_COMMON

VU1_FS_INTERPOLATE_COMMON:

    // Calculate L = B + A = B - -A
    SUB.x fL, fN2Plane, fC2Plane
    // Calculate R = | A / L |
    DIV    Q, fC2Plane[x], fL[x]
    ADDq.x fR, vf00, Q
    ABS.x  fR, fR
    // R is then our lerp coefficient along the current->next vector.
    Lerp fNewGBCPos, fCGBCVert, fNGBCVert, fR
    Lerp fNewST,     fCST,      fNST,     fR, xy
    // Output.
    SQ    fNewGBCPos, 0(pClipOut)
    SQ.xy fNewST,     1(pClipOut)
    // Update output ptr.
    IADDIU pClipOut, pClipOut, 2
    // Done.
    JR pRtA2

.end
