;/////////////////////////////////////////////////////////////////////////////
;//
;// VU1_HierSkin_Dredd.vcl
;//
;// Created by Morgan Parry.
;//
;// Changes:
;//
;//     (MP - 27.05.2002) Initial version.
;//
;// Memory layout:
;//
;//     We want to double-buffer the input and the output but independently,
;//     so that we can switch output buffers during a single execution (ie.
;//     between the two render passes required.) Each input buffer needs to
;//     contain
;//
;//         2 GIF tags + vertex data == 2 + 3N
;//
;//     while the output buffers just need to hold
;//
;//         2 GIF tag + 2 GIF cmd QWs + vertex data == 4 + 3N
;//
;//     giving 12 + 12N QWs of data, where N is the maximum number of
;//     vertices allowed per strip. Hence
;//
;//               1024 - 12
;//          N == --------- == 84
;//                   12
;//
;//     and we have an buffer sizes of
;//
;//          input:  2 + 3N == 254
;//          output: 4 + 3N == 256
;//
;//     We therefore use the following memory layout:
;//
;//         /-----------------\
;//         | input buffer 0  |  254 QWs
;//         |-----------------|
;//         | input buffer 1  |  254 QWs
;//         |-----------------|
;//         | output buffer 0 |  256 QWs
;//         |-----------------|
;//         | output buffer 1 |  256 QWs
;//         |-----------------|
;//         |     unused      |    4 QWs
;//         \-----------------/
;//
;//     The input buffer setup is done by the hardware while we deal with the
;//     output buffer selection ourselves. This is, however, particularly
;//     simple because we don't need to determine at runtime which output
;//     buffer to use; we just write the first pass to output buffer 0 and 
;//     the second to output buffer 1 on every execution.
;//
;/////////////////////////////////////////////////////////////////////////////

;/////////////////////////////////////////////////////////////////////////////
;//
;// VCL setup.
;//
;/////////////////////////////////////////////////////////////////////////////

.init_vf_all
.init_vi_all

.syntax new

--enter
--endenter

;/////////////////////////////////////////////////////////////////////////////
;//
;// Includes.
;//
;/////////////////////////////////////////////////////////////////////////////

#include "vcl_sml.i"

;/////////////////////////////////////////////////////////////////////////////
;//
;// Macros.
;//
;/////////////////////////////////////////////////////////////////////////////

;// The value of these defines should be kept in the sync with the double
;// buffer settings in Asura_PlatformRenderHierarchy.cpp.
#define INPUT_BUFFER_SIZE  (254)
#define OUTPUT_BUFFER_SIZE (256)

#define NUM_INPUT_GIF_TAG_QWS          (2)
#define NUM_OUTPUT_GIF_TAG_AND_CMD_QWS (4)

;/////////////////////////////////////////////////////////////////////////////
;//
;// A short microprogram to setup stuff that is invariant across the entire
;// skin rendering.
;//
;/////////////////////////////////////////////////////////////////////////////

.global HSR_D_GLOBAL_SETUP
HSR_D_GLOBAL_SETUP:
    ;// Determine start of cur input buffer.
    XTOP pIn
    ;// Load GIF tag for texture register setup.
    LQI vTexSetupGifTag, (pIn++)
    ;// Load tex register settings for base pass.
    LQI vBasePassTEX0QW, (pIn++)
    LQI vBasePassTEX1QW, (pIn++)
    ;// And for specular pass.
    LQI vSpecPassTEX0QW, (pIn++)
    LQI vSpecPassTEX1QW, (pIn++)
    ;// Setup output buffer ptrs.
    IADDIU pOutNormal, VI00,     (INPUT_BUFFER_SIZE * 2)
    IADDIU pOutSpecular, pOutNormal, OUTPUT_BUFFER_SIZE
	;// As we are alwas using output buffer 0 for the plain render, we can write out the setup gif packet now
    SQ	vTexSetupGifTag, 0(pOutNormal)
    SQ	vBasePassTEX0QW, 1(pOutNormal)
    SQ	vBasePassTEX1QW, 2(pOutNormal)
	;// And output buffer 1 for the specular render, we can write out the setup gif packet now
    SQ	vTexSetupGifTag, 0(pOutSpecular)
    SQ	vSpecPassTEX0QW, 1(pOutSpecular)
    SQ	vSpecPassTEX1QW, 2(pOutSpecular)
    ;// Prepare value for masking usage later.
    IADDIU i0x7fff, VI00, 0x7fff
    ;// Keep VCL quiet.
    FCSET 0

    --cont
    
;/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
;//
;// This microprogram loads in the data for the next hierarchy section. The following input data is expected:
;//
;//				Non Environment Mapped										Non Environment Mapped
;//     /-----------------------------------\						/-----------------------------------\
;//     |         section->screen mat       | 4 QWs					|         section->screen mat       | 4 QWs
;//     |-----------------------------------|						|-----------------------------------|
;//     |          section->clip mat        | 4 QWs					|          section->clip mat        | 4 QWs
;//     |-----------------------------------|						|-----------------------------------|
;//     |    sectionspace light normal X    | 1 QW					|    sectionspace light normal X    | 1 QW
;//     |    sectionspace light normal Y    | 1 QW					|	 sectionspace light normal Y    | 1 QW
;//     |    sectionspace light normal Z    | 1 QW					|    sectionspace light normal Z    | 1 QW
;//     |-----------------------------------|						|-----------------------------------|
;//     |           light colour 0          | 1 QW					|           light colour 0          | 1 QW
;//     |           light colour 1          | 1 QW					|           light colour 1          | 1 QW
;//     |           light colour 2          | 1 QW					|           light colour 2          | 1 QW
;//     |-----------------------------------|						|-----------------------------------|
;//     |   sectionspace model to cam dir   | 1 QW					| sectionspace specular halfway vec | 1 QW
;//     |-----------------------------------|						|-----------------------------------|
;//     | sectionspace specular halfway vec | 1 QW					|		  section->view mat			| 4 QW
;//     |-----------------------------------|						|-----------------------------------|
;//     |          specular colour          | 1 QW					|             texture mat			| 4 QW
;//     \-----------------------------------/						\-----------------------------------/
;//
;/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

.global HSR_D_LOAD_SECTION_DATA
HSR_D_LOAD_SECTION_DATA:

    ;// Get address of current double buffer.
    XTOP pInputBuffer
    ;// Load matrices.
    MatrixLoad	mS2S,	0,	pInputBuffer
    MatrixLoad	mS2C,	4,	pInputBuffer
    ;// Load lighting parameters.
    LQ.xyz	fLightNormalXs,	 8(pInputBuffer)
	LQ.xyz	fLightNormalYs,	 9(pInputBuffer)
	LQ.xyz	fLightNormalZs,	10(pInputBuffer)
	LQ.xyzw	fLightColour1,	11(pInputBuffer)
	LQ.xyzw	fLightColour2,	12(pInputBuffer)
	LQ.xyzw	fLightColour3,	13(pInputBuffer)

    LQ.xyz	vMdlToCamDir,	14(pInputBuffer)
    LQ.xyz	vSpecHalfway,	15(pInputBuffer)
    LQ.xyzw	vSpecColour,	16(pInputBuffer)
    ;// Since the specular colour is constant, we might as well convert it to GS format now.
    FTOI0	vSpecColour,	vSpecColour

	;// Set alpha component to 128 and convert to gs format (replace this with a real alpha value at some point ?)
	LOI 128
	MULi.w	fRgbaOut,	VF00,	I
	FTOI0.w	fRgbaOut,	fRgbaOut

	;// Setup constants in a register for colour scaling and clamping
	LOI 128
	ADDi.x vColourScaleAndClampConstants, VF00, I
	LOI	255
	ADDi.y vColourScaleAndClampConstants, VF00, I

    --cont

;/////////////////////////////////////////////////////////////////////////////
;//
;// And the actual renderer...
;//
;/////////////////////////////////////////////////////////////////////////////

HSR_D_BASE_PASS_BEGIN_RENDER:

    ;// Get address of current input buffer.
    XTOP	pBaseIn
    ;// Load GIF tags.
    LQ		vBasePassGifTag, 0(pBaseIn)
	IADDIU	pIn, pBaseIn, NUM_INPUT_GIF_TAG_QWS
    ;// Extract NLOOP (remembering to mask off the EOP bit.)
    MTIR	iNumVertices, vBasePassGifTag[x]
    IAND	iNumVertices, iNumVertices, i0x7fff
    ;// Copy for loop ctr.
    IADDIU	iLoopCtr, iNumVertices, 0
    ;// Setup output ptr. (Step over the setup gif packet)
    IADDIU	pOut, pOutNormal, 4
    ;// Output TEXx setup stuff and GIF tag for this strip.
    SQ	vBasePassGifTag, -1(pOut)

HSR_D_BASE_PASS_VERTEX_LOOP:
    --LoopCS 3, 0
		;// Load (sectionspace) position, normal and texture data. Recall that
		;// the ADC bit is stored in the normal's W so we need to load that as
		;// well.
		LQI.xyz	vPos, (pIn++)
		LQI     vNrm, (pIn++)
		LQI.xyz vTex, (pIn++)
		;// Convert texture coordinates from 4:12 fixed to floating point.
		ITOF12.xy vTex, vTex
		;// Transform input position to clipspace and to screenspace. 
		MatrixMultiplyVertexW1 vClipPos, mS2C, vPos
		MatrixMultiplyVertexW1 vScrnPos, mS2S, vPos
		;// Calculate 1/w.
		DIV Q, VF00[w], vScrnPos[w]
		;// Perspective divide.
		MULq.xyz	vScrnPos, vScrnPos, Q
		MULq.xyz	vTex,     vTex,     Q
		;// Calculate light.normal
		MULAx.xyz	ACC,				fLightNormalXs,	vNrm
		MADDAy.xyz  ACC,				fLightNormalYs,	vNrm
		MADDz.xyz   fLightDotNormal,	fLightNormalZs,	vNrm
		;// Clamp to >= 0
		MAXx.xyz	fLightDotNormal,	fLightDotNormal,	VF00
		;// Light the vertex
		MULAx.xyz	ACC,		fLightColour1,	fLightDotNormal
		MADDAy.xyz  ACC,		fLightColour2,	fLightDotNormal
		MADDz.xyz	fRgbaOut,	fLightColour3,	fLightDotNormal
		;// Scale the colour and then clamp to [0, 255].
		MULx.xyz    fRgbaOut,   fRgbaOut,   vColourScaleAndClampConstants
		MINIy.xyz   fRgbaOut,   fRgbaOut,   vColourScaleAndClampConstants
		;// Convert XYZ and RGBA to GS format.
		FTOI0.xyz	fRgbaOut,	fRgbaOut
		FTOI4.xyz	vScrnPos, vScrnPos
		;// Output.
		SQ.xyz		vTex,		0(pOut)
		SQ.xyzw     fRgbaOut,	1(pOut)
		SQ.xyz		vScrnPos,	2(pOut)
		;// Increment pOut (switched from SQI's so VCL can better optimise instructions)
		IADDIU	pOut,	pOut,	3
		;// Perform clipping judgement.
		CLIPw.xyz vClipPos, vClipPos
		;// Check result.
		FCAND  VI01, 0x3ffff
		IADDIU iADC, VI01, 0x7fff ;// ADC bit is now set iff any of the last three vertices intersected with or were outside the guardband frustrum planes.
		;// OR on the degenerate ADC bit setting that was uploaded in the normal's w component.
		MTIR iDegenerateADC, vNrm[w]
		IOR  iADC, iADC, iDegenerateADC
		;// Output to the position's w component.
		ISW.w iADC, -1(pOut)
		;// Decrement vertex count and check if we're done yet.
		IADDI iLoopCtr, iLoopCtr, -1
		IBNE  iLoopCtr, VI00, HSR_D_BASE_PASS_VERTEX_LOOP
    ;// Okay, we've processed all of this strip's vertices. Send to the GS.
    XGKICK pOutNormal

	--barrier

HSR_D_SPECULAR_PASS_BEGIN_RENDER:
    ;// Setup input ptrs. This time we'll need to read in data from the first
    ;// pass's output buffer as well as from the relevant input buffer.
    LQ	vSpecPassGifTag, 1(pBaseIn)
	IADDIU	pIn, pBaseIn, NUM_INPUT_GIF_TAG_QWS
    IADDIU	pBasePassOutputData,	pOutNormal,	NUM_OUTPUT_GIF_TAG_AND_CMD_QWS  ;// Skip over base pass's output GIF tag.
    ;// Reset loop ctr.
    IADDIU iLoopCtr, iNumVertices, 0
    ;// Setup output ptr.
    IADDIU pOut, pOutSpecular, 4
    ;// Output TEXx setup stuff and GIF tag for this strip.
    SQ	vSpecPassGifTag, -1(pOut)

HSR_D_SPECULAR_PASS_VERTEX_LOOP:
    --LoopCS 3, 0
		;// Load Q (ie. 1/w which is stored in the texture QW's z component) and
		;// XYZ and the ADC setting from base pass's output buffer.
		LQ.xyzw		vTex,     0(pBasePassOutputData)
		LQ.xyzw		vScrnPos, 2(pBasePassOutputData)
		IADDIU		pBasePassOutputData, pBasePassOutputData, 3
		;// Load position and normal from current input buffer and update ptr.
		LQ.xyz		vPos,	0(pIn)
		LQ.xyz		vNrm,	1(pIn)
		IADDIU		pIn,	pIn,	3
		;// Calculate S texture coordinate. This determines the outlining effect
		;// and we calculate it as max(0, to_cam . normal).
		;// (PSJE) - Outlining switched off in heirachy renderer
		;//			 Because the F.P. gun is alway outlined
;//		VectorDotProduct vToCamDotNrm, vMdlToCamDir, vNrm
;//		MAX.x vTex, vToCamDotNrm, VF00
		MAXw.x vTex, VF00, VF00
		;// Calculate T texture coordinate. This determines the specular
		;// highlight effect. We calculate it as max(0, h.n ^ 5), where h is the
		;// specular halfway vector.
		VectorDotProduct vHalfwayDotNrm, vSpecHalfway, vNrm
		ADDx.y vTex, VF00, vHalfwayDotNrm ;// vTex.y == h.n
		MUL.y  vTex, vTex, vTex           ;// vTex.y == h.n ^ 2
		MUL.y  vTex, vTex, vTex           ;// vTex.y == h.n ^ 4
		MULx.y vTex, vTex, vHalfwayDotNrm ;// vTex.y == h.n ^ 5
		MAX.y  vTex, vTex, VF00           ;// vTex.y == max(0, h.n ^ 5)
		;// Perspective divide the texture coordinates.
		MULz.xy vTex, vTex, vTex
		;// Output.
		SQ.xyz		vTex,			0(pOut)
		SQ.xyzw     vSpecColour,	1(pOut)
		SQ.xyzw		vScrnPos,		2(pOut)
		;// Increment pOut (switched from SQI's so VCL can better optimise instructions)
		IADDIU	pOut,	pOut,	3
		;// Decrement vertex count and check if we're done yet.
		IADDI iLoopCtr, iLoopCtr, -1
		IBNE  iLoopCtr, VI00, HSR_D_SPECULAR_PASS_VERTEX_LOOP
    ;// Send to the GS.
    XGKICK pOutSpecular
	;// If this barrier is removed VCL creates a nasty NOP[E] XGKICK...
	--barrier
    --cont

    B HSR_D_BASE_PASS_BEGIN_RENDER

.end
