;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;    
;//;    FILE            VU1_Skinning.vcl
;//;    
;//;    DESCRIPTION     VU1 microcode for skinning and rendering characters
;//;
;//;                    Also contains a second quad renderer for rendering the shadow
;//;                    textures to screen
;//;    
;//;    AUTHOR          Pete Ellacott 17.09.02
;//;    
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//; 
;//;    The quad renderer can only be used directly after the character renderer because
;//;    it use data in memory and registers setup by the character renderer
;//;
;//;    (PSJE: 20-11-02) Extended to use variable number of bones upto a maximum of : 63 (goverened by bone index/address packing. ie 64*4qws = 0x100)
;//;
;//;    The input buffer is used to hold camera data and shadow data once skinning has been performed.
;//;    This gives a minimum size for the input buffer and so a maximum number of bones.
;//;    Per camera data which is moved = 10 Qws (stored 2 Qws per vertex ) = 6 Verts per camera
;//;    Per shadow data which is moved = 5 Qws (stored 2 Qws per vertex ) = 3 Verts per camera
;//;    So in the maximum case of 4 cameras and 4 shadows = 6*4 + 3*4 = minimum of 36 verts per input buffer
;//;    This gives a maximum bone count of : 108 Bones
;//;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;
;//;    Calculation for buffer sizes :
;//;    ------------------------------
;//;    Setup : 13QWs (Camera + multipass setup used in dredd renderer, not all used here, but kept in calculation to keep both renderers data compatible)
;//;    OutTags : 2QWs ( 1 gif tag per output buffer )
;//;    Bone Buffer : nBones * 4QWs
;//;    Input Vert Size = 5Qws ( Position, Normal, UV, Weight, Bone Index )
;//;    Output Vert Size = 3Qws ( Position, RGBA, UV )
;//;    VertsPerBuffer = (1024Qws - (BoneBuffer + Setup + OutTags)) / (8*2) = (1009 - BoneBuffer)/16
;//;    Input Buffer Size = 5Qws * VertsPerBuffer
;//;    Output Buffer Size = 1Qw + 3Qws * VertsPerBuffer
;//;
;//;    Constants uploaded at startup:
;//;    ------------------------------
;//;    (BoneBuffer - 14)   Drawing Gif Tag
;//;    (BoneBuffer - 13)   Light Colour Scale
;//;    (BoneBuffer - 12)   Light Directions
;//;    (BoneBuffer -  8)   Light Colours
;//;    (BoneBuffer -  4)   Camera Setup - Giftag
;//;    (BoneBuffer -  3)   Shadow Drawing Gif Tag
;//;    (BoneBuffer -  2)   Shadow Setup Gif Tag
;//;    (BoneBuffer -  1)   Shadow RGBAQ register
;//;    (BoneBuffer +  0)   Camera 1 Screeen Matrix
;//;    (BoneBuffer +  4)   Camera 1 Clipping Matrix
;//;    (BoneBuffer +  8)   Camera 1 Scissor Value
;//;    (BoneBuffer +  9)   Camera 1 XYOffset Value
;//;    
;//;    Optional data depending on camera count
;//;    ---------------------------------------
;//;    4Qwds   Camera 2 Screeen Matrix
;//;    4Qwds   Camera 2 Clipping Matrix
;//;    1Qwds   Camera 2 Scissor Value
;//;    1Qwds   Camera 2 XYOffset Value
;//;    4Qwds   Camera 3 Screeen Matrix
;//;    4Qwds   Camera 3 Clipping Matrix
;//;    1Qwds   Camera 3 Scissor Value
;//;    1Qwds   Camera 3 XYOffset Value
;//;    4Qwds   Camera 4 Screeen Matrix
;//;    4Qwds   Camera 4 Clipping Matrix
;//;    1Qwds   Camera 4 Scissor Value
;//;    1Qwds   Camera 4 XYOffset Value
;//;
;//;    Optional data depending on shadow count
;//;    ---------------------------------------
;//;    1Qwds   Shadow 1 Frame Setting
;//;    4Qwds   Shadow 1 Screeen Matrix
;//;    1Qwds   Shadow 2 Frame Setting
;//;    4Qwds   Shadow 2 Screeen Matrix
;//;    1Qwds   Shadow 3 Frame Setting
;//;    4Qwds   Shadow 3 Screeen Matrix
;//;    1Qwds   Shadow 5 Frame Setting
;//;    4Qwds   Shadow 4 Screeen Matrix

;//;    0x000   Bone Array      : BoneMatrix[MaxBones] - Loaded at startup, after the constants have been set

;//;
;//;    Buffer layout during rendering :
;//;    --------------------------------
;//;    0x000
;//;            Bone Array      : BoneMatrix[nBones]
;//;    0xXXX
;//;            Output Buffer 0 : GifTag,  (STQ,RGB,XYZ)[VertQty]
;//;    0xXXX
;//;            Output Buffer 1 : GifTag,  (STQ,RGB,XYZ)[VertQty]
;//;    0xXXX
;//;            Input Buffer 0  : Initially - (Normal,BoneIndex,BoneWeight,Position, UV)[VertQty]
;//;            Input Buffer 0  : After Skinning -  (Normal,    Camera1 Matrix 1, Camera1 Matrix 2, Position, UV),
;//;                                                (Normal,    Camera1 Matrix 3, Camera1 Matrix 4, Position, UV),
;//;                                                (Normal,    Camera1 Clip 1,   Camera1 Clip 2,   Position, UV),
;//;                                                (Normal,    Camera1 Clip 3,   Camera1 Clip 4,   Position, UV),
;//;                                                (Normal,    Camera2 Matrix 1, Camera2 Matrix 2, Position, UV),
;//;                                                (Normal,    Camera2 Matrix 3, Camera2 Matrix 4, Position, UV),
;//;                                                (Normal,    Camera2 Clip 1,   Camera2 Clip 2,   Position, UV),
;//;                                                (Normal,    Camera2 Clip 3,   Camera2 Clip 4,   Position, UV),
;//;                                                (Normal,    Camera3 Matrix 1, Camera3 Matrix 2, Position, UV),
;//;                                                (Normal,    Camera3 Matrix 3, Camera3 Matrix 4, Position, UV),
;//;                                                (Normal,    Camera3 Clip 1,   Camera3 Clip 2,   Position, UV),
;//;                                                (Normal,    Camera3 Clip 3,   Camera3 Clip 4,   Position, UV),
;//;                                                (Normal,    Camera4 Matrix 1, Camera4 Matrix 2, Position, UV),
;//;                                                (Normal,    Camera4 Matrix 3, Camera4 Matrix 4, Position, UV),
;//;                                                (Normal,    Camera4 Clip 1,   Camera4 Clip 2,   Position, UV),
;//;                                                (Normal,    Camera4 Clip 3,   Camera4 Clip 4,   Position, UV),
;//;                                                (Normal,    unused,           unused,           Position, UV)[VertQty-16]
;//;    0xXXX
;//;            Input Buffer 1  : See input buffer 0 for layout
;//;    0x3f3
;//;            Camera Setup    : GifTag, Scissor, XYOffset ( Dredd Renderer also has - First Pass setup : Tex1, Tex0, Test )
;//;    0x3f9
;//;            ( Dredd Renderer also has - Second Pass setup   : GifTag, Tex1, Tex0, Test )
;//;    0x3fd
;//;            Shadow Setup    : Gif Tag, RGBAQ, Frame_2
;//;    0x400
;//;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;// I N C L U D E S   &   D E F I N E S ////////////////////////////////////////
#include "Asura_PS2_CharDefines.h"
#define PASS1_SETUP_BUFFER          (1024 - ASURA_PS2_SKINNING__QWS_TOTAL_SETUP)
#define PASS2_SETUP_BUFFER          (PASS1_SETUP_BUFFER + ASURA_PS2_SKINNING__QWS_PASS1_SETUP)
#define SHADOW_SETUP_BUFFER         (PASS2_SETUP_BUFFER + ASURA_PS2_SKINNING__QWS_PASS2_SETUP)

;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

.syntax new

.init_vf_all
.init_vi VI01, VI02, VI03, VI04, VI05, VI06, VI07, VI08, VI09, VI10, VI11, VI13

#define REG_THIS_OUTPUT_BUFFER VI15
#define REG_NEXT_OUTPUT_BUFFER VI14
#define REG_NUM_CAMERAS        VI12

.macro ConvertInputNormal nrm
    ITOF4.xyz \nrm, \nrm
    LOI -1.0
    ADDAi.xyz ACC, VF00, I
    LOI 0.125
    MADDi.xyz \nrm, \nrm, I  ;// N = (1/8)N - 1
    ;// Convert normal ADC degeneracy flag
    ITOF0.w \nrm, \nrm
    LOI 0.5
    ADDi.w  \nrm, \nrm, I
.endm


    --enter
    --endenter

START:
    ;// Get the address of the end of the bone matrix (and the start of the constants)
    XITOP   REG_THIS_OUTPUT_BUFFER
    ISUBIU  pConstants,     REG_THIS_OUTPUT_BUFFER, ASURA_PS2_SKINNING__QWS_CONSTANTS
    ;// We are switching output buffers manually so we can find our camera data 
    ILW.w   iOutputBufferSize, 1(pConstants)w
    IADD    REG_NEXT_OUTPUT_BUFFER, REG_THIS_OUTPUT_BUFFER,     iOutputBufferSize
;// Preload registers with common data
    ;// Load the Drawing GIF tag
    LQ      fGifTag,             0(pConstants)
;// ;// Light Scale Colour (ie light values are returned in the range 0.0 to 1.0)
;// LQ      fLightScale,         1(pConstants)
    ;// Load camera count from Pass 1 Drawing Gif Tag - NLOOP member
    ILW.x   REG_NUM_CAMERAS,             0(pConstants)x
    ;// Load the Light data
    LQ      fLightNormalXs,      2(pConstants)
    LQ      fLightNormalYs,      3(pConstants)
    LQ      fLightNormalZs,      4(pConstants)
    ;// Load the rendering alpha value
    LQ.w    fRgbaOut,            5(pConstants)
    LQ      fLightColour1,       6(pConstants)
    LQ      fLightColour2,       7(pConstants)
    LQ      fLightColour3,       8(pConstants)
    LQ      fAmbientColour,      9(pConstants)
    ;// Load the Camera Setup GIF tag
    LQ      fCamera_SetupGIF,   10(pConstants)
    ;// Load Shadow Drawing GifTag and create Shadow Setup drawlist
    LQ      fShadow_DrawGIF,    11(pConstants)
    LQ      fShadow_SetupGIF,   12(pConstants)
    LQ      fShadow_RGBAQ,      13(pConstants)
    SQ      fShadow_SetupGIF,   (SHADOW_SETUP_BUFFER+0)(VI00)
    SQ      fShadow_RGBAQ,      (SHADOW_SETUP_BUFFER+1)(VI00)
    ;// Load shadow count from shadow drawing GIF Tag - NLOOP member
    ILW.x   iShadowQty,         11(pConstants)x
    ;// Load the World->Screen matrix
    LQ      fWorldToScreen1,    14(pConstants)
    LQ      fWorldToScreen2,    15(pConstants)
    LQ      fWorldToScreen3,    16(pConstants)
    LQ      fWorldToScreen4,    17(pConstants)
    ;// Load the World->Clip matrix
    LQ      fWorldToClip1,      18(pConstants)
    LQ      fWorldToClip2,      19(pConstants)
    LQ      fWorldToClip3,      20(pConstants)
    LQ      fWorldToClip4,      21(pConstants)
    ;// Load first cameras Scissor and xyoffset values
    LQ      fSCISSOR,           22(pConstants)
    LQ      fXYOFFSET,          23(pConstants)
;// Register preloading complete - Set Register constants
    ;// Create the first cameras setup drawlist in the right place
    SQ      fCamera_SetupGIF,   (PASS1_SETUP_BUFFER+0)(VI00)
    SQ      fSCISSOR,           (PASS1_SETUP_BUFFER+1)(VI00)
    SQ      fXYOFFSET,          (PASS1_SETUP_BUFFER+2)(VI00)
    --cont                    

;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

STRIP_SETUP:

;// Get which double buffer we are in
    XTOP pInputBuffer
;// Read the vertex count from ITOP
    XITOP   iVertexQty
;// Clear the clipping flags
    FCSET   0
;// First pass - Skinning
    IADD    iVertexCount,   iVertexQty,     VI00
    IADD    pIn,    pInputBuffer,           VI00

//////////////////////////////////////////////////////////////////////////////
// (MP - 24.06.2004) Quick hack to fix up weighting data to get the old
// renderer to cope with the new data format.
//
    MINIx fAll0s, VF00, VF00
    // Load max weightings per vertex.
    ILW.w iMaxWeightingsPerVtx, 4(pIn)
    // Setup 
    // Branch to correct fix-up routine.
    ISUBIU iMaxWeightingsPerVtx, iMaxWeightingsPerVtx, 1
    IBEQ iMaxWeightingsPerVtx, VI00, FIX_UP_YZW
    ISUBIU iMaxWeightingsPerVtx, iMaxWeightingsPerVtx, 1
    IBEQ iMaxWeightingsPerVtx, VI00, FIX_UP_ZW
    ISUBIU iMaxWeightingsPerVtx, iMaxWeightingsPerVtx, 1
    IBEQ iMaxWeightingsPerVtx, VI00, FIX_UP_W
    B DONE_FIX_UP
FIX_UP_YZW:
    SQ.yzw fAll0s, 2(pIn)
    IADDIU pIn, pIn, 5
    ISUBIU iVertexCount, iVertexCount, 1
    IBNE iVertexCount, VI00, FIX_UP_YZW
    B DONE_FIX_UP
FIX_UP_ZW:
    SQ.zw fAll0s, 2(pIn)
    IADDIU pIn, pIn, 5
    ISUBIU iVertexCount, iVertexCount, 1
    IBNE iVertexCount, VI00, FIX_UP_ZW
    B DONE_FIX_UP
FIX_UP_W:
    SQ.w fAll0s, 2(pIn)
    IADDIU pIn, pIn, 5
    ISUBIU iVertexCount, iVertexCount, 1
    IBNE iVertexCount, VI00, FIX_UP_W
DONE_FIX_UP:
    // Reset input ptr and vertex count for loop below.
    XTOP pIn
    IADDIU iVertexCount, iVertexQty, 0
//
//////////////////////////////////////////////////////////////////////////////

    IADDIU iXIsZeroMACFlag, VI00, 0x8

    SKINNING_VERTEX_LOOP:

        --LoopCS    3,0
        ;// load in our data
        LQ.xyzw fNormal,        0(pIn)
        LQ.yzw  fBoneIds,       1(pIn)
        LQ.xyzw fBoneWeights,   2(pIn)
        LQ.xyz  fVertexIn,      3(pIn)
        ;// Unpack Normal & weight
        ConvertInputNormal  fNormal
        ITOF12.xyz          fVertexIn,    fVertexIn
        ITOF15.xyzw         fBoneWeights, fBoneWeights
        ;// set output point to (0, 0, 0, X)
        MOVE.xyz    fVertexOut, VF00
        // Load the matrix address for this bone. We can load it directly
        // here and reduce latency (as opposed to having to use MTIR in the
        // loop.)
        ILW.x iBoneIndex, 1(pIn)x
        ;// Load first bone matrix.
        LQ.xyz fBoneMatrix1, 4(iBoneIndex)
        LQ.xyz fBoneMatrix2, 5(iBoneIndex)
        LQ.xyz fBoneMatrix3, 6(iBoneIndex)
        LQ.xyz fBoneMatrix4, 7(iBoneIndex)
        ;// transform normal to output reg
        MULAx.xyz  ACC,     fBoneMatrix1,   fNormal
        MADDAy.xyz ACC,     fBoneMatrix2,   fNormal
        MADDz.xyz  fNormal, fBoneMatrix3,   fNormal

        SKINNING_BONE_LOOP:

            --LoopCS 1, 0

            ;// transform point to temp reg 
            MULAw.xyz  ACC,   fBoneMatrix4, VF00[w]
            MADDAx.xyz ACC,   fBoneMatrix1, fVertexIn[x]
            MADDAy.xyz ACC,   fBoneMatrix2, fVertexIn[y]
            MADDz.xyz  fTemp, fBoneMatrix3, fVertexIn[z]
            ;// scale transformed point by weight and add to output point
            MULx.xyz    fTemp,      fTemp,      fBoneWeights
            ADD.xyz     fVertexOut, fVertexOut, fTemp
            ;// Rotate bone indicies and weights ready for next bone.
            MR32.xyz fBoneIds,     fBoneIds
            MR32.xyz fBoneWeights, fBoneWeights
            // Force 0 into w weight component to deal with termination in
            // the case of 4 weightings.
            SUB.w fBoneWeights, VF00, VF00
            // Get the matrix index for the next bone.
            MTIR iBoneIndex, fBoneIds[x]
            // Load next bone matrix.
            LQ.xyz fBoneMatrix1, 4(iBoneIndex)
            LQ.xyz fBoneMatrix2, 5(iBoneIndex)
            LQ.xyz fBoneMatrix3, 6(iBoneIndex)
            LQ.xyz fBoneMatrix4, 7(iBoneIndex)
            // Stop when we encounter a 0 weight.
            SUB.x VF00, fBoneWeights, VF00
            FMAND iWeightIsZero, iXIsZeroMACFlag
            IBEQ iWeightIsZero, VI00, SKINNING_BONE_LOOP

    FINISHED_SKINNING_BONE_LOOP:

        ;// Store Weighted vert and normal back in the input buffer ready for rendering
        SQ.xyz  fVertexOut, 3(pIn)
        SQ.xyzw fNormal,    0(pIn)
        IADDIU  pIn,    pIn,    5
        ;// Decrement counter
        IADDI   iVertexCount,   iVertexCount,   -1
        ;// Are we done yet?
        IBNE    iVertexCount,   VI00,   SKINNING_VERTEX_LOOP

    ;// Everythings Skinned .. now we need to move the extra camera data and shadow data, out of the output buffers and stash it in the the unused input buffer
    IADD    pOut,   pInputBuffer,       VI00
    IADD    pIn,    REG_THIS_OUTPUT_BUFFER, VI00
    IADDIU  iCameraCount,   REG_NUM_CAMERAS,    0

MOVE_CAMERA_DATA_LOOP:
        --LoopCS    1,0
    
            LQ  fTemp1,  0(pIn)
            LQ  fTemp2,  1(pIn)
            LQ  fTemp3,  2(pIn)
            LQ  fTemp4,  3(pIn)
            LQ  fTemp5,  4(pIn)
            LQ  fTemp6,  5(pIn)
            LQ  fTemp7,  6(pIn)
            LQ  fTemp8,  7(pIn)
            LQ  fTemp9,  8(pIn)
            LQ  fTemp10, 9(pIn)

            SQ  fTemp1,  1(pOut)
            SQ  fTemp2,  2(pOut)
            SQ  fTemp3,  6(pOut)
            SQ  fTemp4,  7(pOut)
            SQ  fTemp5, 11(pOut)
            SQ  fTemp6, 12(pOut)
            SQ  fTemp7, 16(pOut)
            SQ  fTemp8, 17(pOut)
            SQ  fTemp9, 21(pOut)
            SQ  fTemp10,22(pOut)

            IADDIU  pIn, pIn,   10
            IADDIU  pOut, pOut, 25

            IADDI   iCameraCount,   iCameraCount,   -1
            IBNE    iCameraCount,   VI00,   MOVE_CAMERA_DATA_LOOP

CAMERA_DATA_MOVED:

    ;// Now move the shadow data - skip if no shadows
    IBEQ    iShadowQty, VI00,   SHADOW_DATA_MOVED
    IADD    iShadowCount,   iShadowQty,     VI00
MOVE_SHADOW_DATA_LOOP:
        --LoopCS    1,0
    
            LQ  temp1,  0(pIn)
            LQ  temp2,  1(pIn)
            LQ  temp3,  2(pIn)
            LQ  temp4,  3(pIn)
            LQ  temp5,  4(pIn)

            SQ  temp1,  1(pOut)
            SQ  temp2,  2(pOut)
            SQ  temp3,  6(pOut)
            SQ  temp4,  7(pOut)
            SQ  temp5, 11(pOut)

            IADDIU  pIn, pIn,   5
            IADDIU  pOut, pOut, 15

            IADDI   iShadowCount,   iShadowCount,   -1
            IBNE    iShadowCount,   VI00,   MOVE_SHADOW_DATA_LOOP
SHADOW_DATA_MOVED:

    ;// Lets render each of the cameras
    IADD    iCameraCount,   REG_NUM_CAMERAS,        VI00
    IADDIU  pCameraData,    pInputBuffer,   25          ;// The first cameras data is already loaded so we'll start reloading data fromt he second camera
    ;// Setup constants in a register for colour scaling and clamping
    LOI 128
    ADDi.x vColourScaleAndClampConstants, VF00, I
    LOI 255
    ADDi.y vColourScaleAndClampConstants, VF00, I

PER_CAMERA_RENDERING_LOOP:
    --LoopCS    1,0

    ;// RENDERING
    IADDIU  iVertexCount,   iVertexQty,         0
    IADD    pIn,            pInputBuffer,       VI00
    IADDIU  pOut,           REG_THIS_OUTPUT_BUFFER, 1

    ;// Set the nloop count for this output buffers gif tag
    IADDIU  iTemp,  VI00,   0x7fff
    IADDIU  iTemp,  iTemp,  1
    IOR     iEOP_and_NLoop, iTemp,  iVertexQty
    ISWR.x  iEOP_and_NLoop, (REG_THIS_OUTPUT_BUFFER)x
    ;// Write out the gif tags for this buffer
    SQ.yzw  fGifTag,        0(REG_THIS_OUTPUT_BUFFER)

    RENDERING_VERTEX_LOOP:
            --LoopCS    3,0
            LQ.xyzw     fNormalIn,      0(pIn)      ;// Load Normal        
            LQ.xyz      fVertexIn,      3(pIn)      ;// Load XYZW
            LQ.xy       fSTQ,           4(pIn)      ;// Load STQ
            IADDIU      pIn,    pIn,    5
        ;// Transform the vertex into Clip Space
            MULAx.xyzw  ACC,            fWorldToClip1,  fVertexIn
            MADDAy.xyzw ACC,            fWorldToClip2,  fVertexIn
            MADDAz.xyzw ACC,            fWorldToClip3,  fVertexIn
            MADDw.xyzw  fClipVertex,    fWorldToClip4,  VF00
            CLIPw.xyz   fClipVertex,    fClipVertex
        ;// Transform the vertex to screen space
            MULAx.xyzw  ACC,        fWorldToScreen1,    fVertexIn
            MADDAy.xyzw ACC,        fWorldToScreen2,    fVertexIn
            MADDAz.xyzw ACC,        fWorldToScreen3,    fVertexIn
            MADDw.xyzw  fVertexOut, fWorldToScreen4,    VF00
            ;// while waiting, convert the STQs from 4.12 to floats, 
            ITOF12.xy   fSTQ,   fSTQ
            ADDw.z      fSTQ,   VF00,   VF00   ;// set Q to 1.0f
        ;// Calculate 1/W
            DIV     Q,  VF00[w],    fVertexOut[w]
        ;// Perspective divide both the XYZ and ST
            MULq.xyz    fVertexOut, fVertexOut, Q                
            MULq.xyz    fSTQ,       fSTQ,       Q            
        ;// Calculate light.normal
            MULAx.xyzw  ACC,                fLightNormalXs,     fNormalIn
            MADDAy.xyzw ACC,                fLightNormalYs,     fNormalIn
            MADDz.xyzw  fLightDotNormal,    fLightNormalZs,     fNormalIn
        ;// Clamp to >= 0
            MAXx.xyzw   fLightDotNormal,    fLightDotNormal,    VF00
        ;// Create light colour
            ADDA.xyz    ACC,        fAmbientColour, VF00
            MADDAx.xyz  ACC,        fLightColour1,  fLightDotNormal
            MADDAy.xyz  ACC,        fLightColour2,  fLightDotNormal
            MADDz.xyz   fRgbaOut,   fLightColour3,  fLightDotNormal
        ;// Scale the colour and then clamp to [0, 255].
            MULx.xyz    fRgbaOut,   fRgbaOut,   vColourScaleAndClampConstants
            MINIy.xyz   fRgbaOut,   fRgbaOut,   vColourScaleAndClampConstants
        ;// Decrement the vertex loop counter
            IADDI   iVertexCount,   iVertexCount,   -1
        ;// Convert RGBA and XYZ for the GIF packet
            FTOI4.xyz   fVertexOut, fVertexOut
            FTOI0.xyz   fRgbaOut,   fRgbaOut
        ;// Output STQ, RGBA, XYZ
            SQ.xyz      fSTQ,           0(pOut) 
            SQ.xyzw     fRgbaOut,       1(pOut)
            SQ.xyz      fVertexOut,     2(pOut)
            IADDIU      pOut,   pOut,   3
            ;// Copy export ADC bit from normal's w to integer register.
            MTIR    iDegeneracyADCFlag, fNormalIn[w]
            ;// Test if any of the clip flags are set for last check
            FCAND   VI01,       0x3ffff             ;// previous 3 . any set, then vi01=1, else 0
            IADDIU  iADCFlag,   VI01,   0x7fff      ;// add 0x7fff so that iADCFlag will be 0x8000 if clip flags set or 0x7fff otherwise
            ;// OR in the ADC flag from the input data.
            IOR     iADCFlag,   iADCFlag,   iDegeneracyADCFlag
            ISW.w   iADCFlag,    -1(pOut)w
            IBNE    iVertexCount,   VI00,   RENDERING_VERTEX_LOOP
        ;// Kick the camera setup list
        IADDIU  pKickMe,    VI00,   PASS1_SETUP_BUFFER
        XGKICK  pKickMe
        ;// And Send this strip to the GS
        XGKICK  REG_THIS_OUTPUT_BUFFER

        ;// Swap output buffers
        IADDIU  iTemp,              REG_THIS_OUTPUT_BUFFER, 0
        IADDIU  REG_THIS_OUTPUT_BUFFER, REG_NEXT_OUTPUT_BUFFER, 0
        IADDIU  REG_NEXT_OUTPUT_BUFFER, iTemp,              0

        ;// Are there anymore cameras to come ?
        IADDI   iCameraCount,   iCameraCount,   -1
        IBEQ    iCameraCount,   VI00,   FINISHED_RENDERING_CAMERAS

        ;// Load the next cameras settings
        LQ  fWorldToScreen1, 1(pCameraData)
        LQ  fWorldToScreen2, 2(pCameraData)
        LQ  fWorldToScreen3, 6(pCameraData)
        LQ  fWorldToScreen4, 7(pCameraData)
        LQ  fWorldToClip1,  11(pCameraData)
        LQ  fWorldToClip2,  12(pCameraData)
        LQ  fWorldToClip3,  16(pCameraData)
        LQ  fWorldToClip4,  17(pCameraData)
        LQ  fSCISSOR,       21(pCameraData)
        LQ  fXYOFFSET,      22(pCameraData)
        ;// Update the camera setup draw list
        SQ  fSCISSOR,       (PASS1_SETUP_BUFFER+1)(VI00)
        SQ  fXYOFFSET,      (PASS1_SETUP_BUFFER+2)(VI00)
        
        IADDIU  pCameraData, pCameraData, 25

        B   PER_CAMERA_RENDERING_LOOP

FINISHED_RENDERING_CAMERAS:

    ;// Now render any shadows
    IBEQ    iShadowQty, VI00,   FINISHED_RENDERING_SHADOWS
    IADD    iShadowCount,   iShadowQty,     VI00

PER_SHADOW_RENDERING_LOOP:
    --LoopCS    1,0
    ;// Load the shadow matrix
        LQ  fFRAME,           1(pCameraData)
        LQ  fWorldToScreen1,  2(pCameraData)
        LQ  fWorldToScreen2,  6(pCameraData)
        LQ  fWorldToScreen3,  7(pCameraData)
        LQ  fWorldToScreen4, 11(pCameraData)
        ;// Update the shadow setup draw list
        SQ  fFRAME, (SHADOW_SETUP_BUFFER+2)(VI00)
        ;// Move the shadow data pointer on, ready for the next shadow
        IADDIU  pCameraData, pCameraData, 15
    ;// Counter for vertex quantity
        IADDIU  iVertexCount,   iVertexQty, 0
    ;// Input and Output pointers for skinned verts and normals and UVs
        IADD    pIn,    pInputBuffer,   VI00
        IADDIU  pOut,   REG_THIS_OUTPUT_BUFFER, 1

        ;// Write shadow drawing GifTag
        SQ.yzw  fShadow_DrawGIF, 0(REG_THIS_OUTPUT_BUFFER)
        ;// Set the nloop count for this output buffers gof tag
        IADDIU  iTemp,  VI00,   0x7fff
        IADDIU  iTemp,  iTemp,  1
        IOR     iEOP_and_NLoop, iTemp,  iVertexQty
        ISW.x   iEOP_and_NLoop, 0(REG_THIS_OUTPUT_BUFFER)x

SHADOW_RENDERING_VERTEX_LOOP:
            --LoopCS    3,0

            LQ.w    fScrnPos,       0(pIn)      ;// Load degeneracy flag
            LQ.xyz  fVertexIn,      3(pIn)      ;// Load XYZ
            IADDIU  pIn,    pIn,    5
            ;// Transform the vertex to screen space - no need to clip the matrixs ensures it's all visible
            MULAx.xyzw  ACC,        fWorldToScreen1,    fVertexIn
            MADDAy.xyzw ACC,        fWorldToScreen2,    fVertexIn
            MADDAz.xyzw ACC,        fWorldToScreen3,    fVertexIn
            MADDw.xyzw  fVertexOut, fWorldToScreen4,    VF00
            ;// Calculate 1/W
            DIV Q,  VF00[w],    fVertexOut[w]
            ;// Perspective divide both the XYZ
            MULq.xyz    fScrnPos,   fVertexOut, Q                
            ;// Convert XYZ for the GIF packet
            FTOI4.xyz   fScrnPos,   fScrnPos
            ;// Write out XYZ
            SQ.xyzw     fScrnPos,       0(pOut)  ;// Output XYZ
            IADDIU      pOut,   pOut,   1
            ;// Decrement the vertex loop counter
            IADDI   iVertexCount,   iVertexCount,   -1
            IBNE    iVertexCount,   VI00,   SHADOW_RENDERING_VERTEX_LOOP
        ;// Kick the shadow setup drawlist
        IADDIU  pKickMe,    VI00,   SHADOW_SETUP_BUFFER
        XGKICK  pKickMe
        ;// Kick the Output Buffer
        XGKICK  REG_THIS_OUTPUT_BUFFER

        ;// Swap output buffers
        IADDIU  iTemp,              REG_THIS_OUTPUT_BUFFER, 0
        IADDIU  REG_THIS_OUTPUT_BUFFER, REG_NEXT_OUTPUT_BUFFER, 0
        IADDIU  REG_NEXT_OUTPUT_BUFFER, iTemp,              0

        ;// Are we finished with the shadows ?
        IADDI   iShadowCount,   iShadowCount,   -1
        IBNE    iShadowCount,   VI00,   PER_SHADOW_RENDERING_LOOP

FINISHED_RENDERING_SHADOWS:

    ;// This strip is all rendered in each of the cameras. Put all the camera data back in the right places.. If we have shadows put that data back in the right place
    ;// Set up these pointers now in case we have shadows but only one camera
    IADD    pIn,            pInputBuffer,       VI00
    IADD    pOut,           REG_THIS_OUTPUT_BUFFER, VI00
    IADD    pCameraData,    pInputBuffer,       VI00

    ;// Reload first camera settings
    LQ  fWorldToScreen1, 1(pCameraData)
    LQ  fWorldToScreen2, 2(pCameraData)
    LQ  fWorldToScreen3, 6(pCameraData)
    LQ  fWorldToScreen4, 7(pCameraData)
    LQ  fWorldToClip1,  11(pCameraData)
    LQ  fWorldToClip2,  12(pCameraData)
    LQ  fWorldToClip3,  16(pCameraData)
    LQ  fWorldToClip4,  17(pCameraData)
    LQ  fSCISSOR,       21(pCameraData)
    LQ  fXYOFFSET,      22(pCameraData)
    ;// Update the camera setup draw list
    SQ  fSCISSOR,       (PASS1_SETUP_BUFFER+1)(VI00)
    SQ  fXYOFFSET,      (PASS1_SETUP_BUFFER+2)(VI00)

    ;// Move the camera data back into the output buffers ready for then next strip
    IADDIU  iCameraCount,   REG_NUM_CAMERAS,    0
MOVE_CAMERA_DATA_BACK_LOOP:
    --LoopCS    1,0
        LQ  fTemp1,  1(pIn)
        LQ  fTemp2,  2(pIn)
        LQ  fTemp3,  6(pIn)
        LQ  fTemp4,  7(pIn)
        LQ  fTemp5, 11(pIn)
        LQ  fTemp6, 12(pIn)
        LQ  fTemp7, 16(pIn)
        LQ  fTemp8, 17(pIn)
        LQ  fTemp9, 21(pIn)
        LQ  fTemp10,22(pIn)

        SQ  fTemp1,  0(pOut)
        SQ  fTemp2,  1(pOut)
        SQ  fTemp3,  2(pOut)
        SQ  fTemp4,  3(pOut)
        SQ  fTemp5,  4(pOut)
        SQ  fTemp6,  5(pOut)
        SQ  fTemp7,  6(pOut)
        SQ  fTemp8,  7(pOut)
        SQ  fTemp9,  8(pOut)
        SQ  fTemp10, 9(pOut)

        IADDIU  pIn, pIn, 25
        IADDIU  pOut, pOut, 10

        IADDI   iCameraCount,   iCameraCount,   -1
        IBNE    iCameraCount,   VI00,   MOVE_CAMERA_DATA_BACK_LOOP

CAMERA_DATA_MOVED_BACK:

    ;// Now move the shadow data - skip if no shadows
    IBEQ    iShadowQty,     VI00,   SHADOW_DATA_MOVED_BACK
    IADD    iShadowCount,   iShadowQty,     VI00

MOVE_SHADOW_DATA_BACK_LOOP:
        --LoopCS    1,0
            LQ  temp1,  1(pIn)
            LQ  temp2,  2(pIn)
            LQ  temp3,  6(pIn)
            LQ  temp4,  7(pIn)
            LQ  temp5, 11(pIn)

            SQ  temp1,  0(pOut)
            SQ  temp2,  1(pOut)
            SQ  temp3,  2(pOut)
            SQ  temp4,  3(pOut)
            SQ  temp5,  4(pOut)

            IADDIU  pIn, pIn,   15
            IADDIU  pOut, pOut, 5

            IADDI   iShadowCount,   iShadowCount,   -1
            IBNE    iShadowCount,   VI00,   MOVE_SHADOW_DATA_BACK_LOOP
SHADOW_DATA_MOVED_BACK:

    --cont
    
    B    STRIP_SETUP

;// Quad renderer to draw the shadows back into the the scene once the character has finished rendering
;// It's here so we can keep the same renderer in memory, and luckily all the camera data is still present

    --enter
        .init_vi_all
        in_vi pThisOutputBuffer(VI15)
        in_vi pNextOutputBuffer(VI14)
        in_vi iCameraQty(VI12)
    --endenter

.global VU1_Micro_Skinning_QuadRender
VU1_Micro_Skinning_QuadRender:

    ;// At the end of the character rendering, all the camera data we need is sitting in one of the two output buffers.
    ;// No we are going to be setting up scissor and offset values and drawing quads, giving us very little ouput data
    ;// So we'll leave the output buffer with the camera data alone, and split the other output buffer into 2 and use that for output

    ;// Load the gif tags for the quads from 0 - the bone buffer is no longer needed
    LQ      fQuad_GIF,  0(VI00)

    ;// And the output buffer size
    ILW.w   iHalfOutputBufferSize, 1(VI00)w

    // Setup the output pointer and camera data pointer
    IADD    pSkinCamData, pThisOutputBuffer, VI00
    IADD    pThisOutputBuffer, pNextOutputBuffer, iHalfOutputBufferSize

    --cont

QUAD_RENDER_SETUP:

    ;// Get which double buffer we are in (where our memory starts)
    XTOP    pInputBuffer

    ;// Per camera quad rendering loop
    IADDIU pCamIn,       pInputBuffer, 0
    IADDIU pGSSetupIn,   pSkinCamData, 0
    IADDIU iCameraCount, iCameraQty,   0

PER_CAMERA_QUAD_RENDERING_LOOP:
    --LoopCS    1,0

        ;// Load the settings for this camera
        LQ  fWorldToScreen1,  0(pCamIn)
        LQ  fWorldToScreen2,  1(pCamIn)
        LQ  fWorldToScreen3,  2(pCamIn)
        LQ  fWorldToScreen4,  3(pCamIn)
        LQ  fWorldToClip1,    4(pCamIn)
        LQ  fWorldToClip2,    5(pCamIn)
        LQ  fWorldToClip3,    6(pCamIn)
        LQ  fWorldToClip4,    7(pCamIn)
        LQ  fRGBA,            8(pCamIn)
        LQ  fSCISSOR,         8(pGSSetupIn)
        LQ  fXYOFFSET,        9(pGSSetupIn)
        IADDIU pCamIn,     pCamIn,     9
        IADDIU pGSSetupIn, pGSSetupIn, 10

        ;// Build the start of the draw list
        SQ  fQuad_GIF,  0(pThisOutputBuffer)
        SQ  fSCISSOR,   1(pThisOutputBuffer)
        SQ  fXYOFFSET,  2(pThisOutputBuffer)

        ;// Clear clipping flags
        FCSET   0

        ;// It's a quad, so it has 4 verts
        IADDIU  iVertexCount,   VI00,   4

        IADDIU  pIn,    pInputBuffer,       (4 * 9)
        IADDIU  pOut,   pThisOutputBuffer,  3

QUAD_RENDERING_VERTEX_LOOP:
            --LoopCS    4,0

            LQ.xyz      fSTQ,           0(pIn)      ;// Load STQ        
            LQ.xyz      fVertexIn,      1(pIn)      ;// Load XYZW

            IADDIU  pIn,    pIn,    2

            ;// Transform the vertex into Clip Space and Clip
            MULAx.xyzw  ACC,            fWorldToClip1,  fVertexIn
            MADDAy.xyzw ACC,            fWorldToClip2,  fVertexIn
            MADDAz.xyzw ACC,            fWorldToClip3,  fVertexIn
            MADDw.xyzw  fClipVertex,    fWorldToClip4,  VF00
            CLIPw.xyz   fClipVertex,    fClipVertex
    
            ;// Transform the vertex to screen space
            MULAx.xyzw  ACC,        fWorldToScreen1,    fVertexIn
            MADDAy.xyzw ACC,        fWorldToScreen2,    fVertexIn
            MADDAz.xyzw ACC,        fWorldToScreen3,    fVertexIn
            MADDw.xyzw  fVertexOut, fWorldToScreen4,    VF00

            ;// Test if any of the clip flags are set for last check
            FCAND   VI01,       0x3ffff             ;// previous 3 . any set, then vi01=1, else 0
            IADDIU  iADCFlag,   VI01,   0x7fff      ;// add 0x7fff so that iADCFlag will be 0x8000 if clip flags set or 0x7fff otherwise

            ;// Calculate 1/W
            DIV Q,  VF00[w],    fVertexOut[w]

            ;// Copy clipping KICK flag into output vertex
            MFIR.w  fVertexOut, iADCFlag

            ;// Perspective divide both the XYZ and ST
            MULq.xyz    fVertexOut, fVertexOut, Q
            MULq.xyz    fSTQ,       fSTQ,       Q            

            ;// Convert XYZ for the GIF packet
            FTOI4.xyz   fVertexOut, fVertexOut

            ;// Write out STQ, RGBA, and XYZ 
            SQ.xyz  fSTQ,           0(pOut) ;// Output STQ
            SQ      fRGBA,          1(pOut) ;// Output RGBA
            SQ      fVertexOut,     2(pOut)
            IADDIU  pOut,   pOut,   3

            ;// Decrement the vertex loop counter
            IADDI   iVertexCount,   iVertexCount,   -1
            IBNE    iVertexCount,   VI00,   QUAD_RENDERING_VERTEX_LOOP
    
        ;// Kick Output Buffer 0 (First Pass Drawing)
        XGKICK  pThisOutputBuffer

        ;// Swap output buffers
        IADDIU  iTemp,              pThisOutputBuffer,  0
        IADDIU  pThisOutputBuffer,  pNextOutputBuffer,  0
        IADDIU  pNextOutputBuffer,  iTemp,              0

        ;// Have we done all the cameras ?
        IADDI   iCameraCount,   iCameraCount,   -1
        IBNE    iCameraCount,   VI00,   PER_CAMERA_QUAD_RENDERING_LOOP

    ;// Ok, we are done with this quad

END:
    --cont
    
    B    QUAD_RENDER_SETUP

.end

