//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Shape.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Shapes/Asura_Platform_Shape.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Macros.h"
#include "Core/Asura_Material_System.h"
#include "Debugging/Asura_Assert.h"
#include "Light/Asura_Light.h"
#include "Shapes/Asura_Shapes.h"

// Engine platform-specific.
#include "Rendering/Asura_PlatformRenderObject.h"
#include "Rendering/Asura_PS2_Object_System.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Platform_Shape_RenderData.
//

Asura_Platform_Shape_RenderData::Asura_Platform_Shape_RenderData( const char*& pcChunkBuffer )
{
    const char* szName = pcChunkBuffer;
    pcChunkBuffer += 260;

#if _DEBUG
    m_szName = new char[strlen(szName) + 1];
    if( m_szName ) { strcpy( m_szName, szName ); }
#endif
}

#if _DEBUG
Asura_Platform_Shape_RenderData::~Asura_Platform_Shape_RenderData( void )
{
    delete[] m_szName;
}
#endif

//
// Asura_Shape.
//

int Asura_Shape::Platform_GetMaterialIdx( void ) const
{
    return Asura_PS2_Object_System::GetMaterialIdx( m_uHashID );
}

bool Asura_Shape::Platform_IsOpaque( void ) const
{
    const u_int uMaterialFlags = Asura_Material_System::GetMaterialFlags( Platform_GetMaterialIdx() );
    return( 0 == (uMaterialFlags & (AsuraMaterialFlag_Transparent | AsuraMaterialFlag_PS2AdditiveAlpha)) );
}

void Asura_Shape::Platform_Render( const Asura_Shape_Instance* pxShapeInst, u_int uAmbientARGB, u_int uCamVisFlags ) const
{
	Asura_Assert( uCamVisFlags, "Shouldn't get to this render call without visibility." );

    const Asura_Vector_3&   xPos = pxShapeInst->GetPosition();
    const Asura_Matrix_3x3& xOri = pxShapeInst->GetOrientation();

#ifdef ASURA_PS2_RENDER_OBJECT_LOD_PER_CAMERA

	// Warning - assumption of a maximum of 4 cameras here.
	const Asura_PS2_StrippedObject* apxObjectToRender[4];

	// First, find out which LOD is needed for each camera.
	for (int i=0;i<Asura_Camera::GetNumberOfCameras();i++)
	{
		Asura_Assert(i<4, "Assumption about max-cameras in-correct");
		if ((1<<i) & uCamVisFlags)
		{
			Asura_Camera::SetCameraIndex(i);
			Asura_Vector_3 xDiff = (xPos - (*Asura_Camera::GetPosition()));
			float fDistanceSqd = xDiff * xDiff * Asura_Camera::GetLODMetric(i) * Asura_Camera::GetLODMetric(i);

			apxObjectToRender[i] = Asura_PS2_Object_System::GetPS2ObjectToRender( m_uHashID, fDistanceSqd );
            if( !apxObjectToRender[i] )
            {
                uCurVisFlags &= ~(1 << i);
            }
		}
	}

	// Next, render each camera, but also render for any other camera that
	// shares the same LOD model.
	unsigned uCurVisFlags = uCamVisFlags;
	for (int i=0;i<Asura_Camera::GetNumberOfCameras();i++)
	{
		if ((1<<i) & uCurVisFlags)
		{
			// Now, we loop through the other cameras to see if any others 
			//share the same LOD model.
			unsigned uNewCamVisFlags = (1<<i);
			for (int j=i+1;j<Asura_Camera::GetNumberOfCameras();j++)
			{
				if ((1<<j) & uCurVisFlags)
				{
					// Shape is visible in this camera, too.
					if (apxObjectToRender[j] == apxObjectToRender[i])
					{
						// We will be drawing the same LOD in this camera.
						uNewCamVisFlags |= (1<<j);
						// Make sure we don't render this one again later in the loop
						uCurVisFlags &= ~(1<<j);
					}
				}
			}
			apxObjectToRender[i]->Render( pxShapeInst, uAmbientARGB, uNewCamVisFlags );
		}
	}

#else // not def ASURA_PS2_RENDER_OBJECT_LOD_PER_CAMERA

	float fMinDistanceSqd = Asura_Maths::LargeFloat;
	for (int i=0;i<Asura_Camera::GetNumberOfCameras();i++)
	{
		if ((1<<i) & uCamVisFlags)
		{
			// Shape is visible in this camera.
			Asura_Camera::SetCameraIndex(i);
			Asura_Vector_3 xDiff = (xPos - (*Asura_Camera::GetPosition()));
			float fDistanceSqd = xDiff * xDiff * Asura_Camera::GetLODMetric(i) * Asura_Camera::GetLODMetric(i);
			if (fDistanceSqd < fMinDistanceSqd)
			{
				fMinDistanceSqd = fDistanceSqd;
			}
		}
	}
	const Asura_PS2_StrippedObject* pxObj = Asura_PS2_Object_System::GetPS2ObjectToRender( m_uHashID, fMinDistanceSqd );
    if( pxObj )
    {
        pxObj->Render( pxShapeInst, uAmbientARGB, uCamVisFlags );
    }

#endif // end not def ASURA_PS2_RENDER_OBJECT_LOD_PER_CAMERA
}

void Asura_Shape::Platform_RenderShadow( const Asura_Shape_Instance* pxShapeInst, const Asura_Shadow& xShadow, u_int uCamVisFlags ) const
{
}

//
// Asura_Shape_Instance.
//

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::Platform_SetAmbientARGB
// PARAMETERS	components of colour in one 32 bit value (ARGB)
// RETURNS		None
// DESCRIPTION	
// AUTHOR		Neil Massey (13th Sept. 2002)
// MODIFIED     Mike Healey (26.9.02) -- changed halving so it rounds 255 to 128
//                                       also got order of ARGB right
//              Andrew Weinkove (27.02.04) - No longer need to half because Asura_PS2_GSMRenderer::SetMeshColour
//											will do this.
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::Platform_SetAmbientARGB( const unsigned int uARGB )
{
	m_uAmbientARGB = uARGB;
	m_uFlags &= ~ASURA_SHAPE_INSTANCE_USES_DEFAULT_AMBIENT;
}

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::Platform_SetAmbientRGB
// PARAMETERS	components of colour in four 8 bit value (RGBA)
// RETURNS		None
// DESCRIPTION	
// AUTHOR		Neil Massey (13th Sept. 2002)
// MODIFIED     Mike Healey (26.9.02) -- changed halving so it rounds 255 to 128
//                                       also got order of ARGB right
//              Andrew Weinkove (27.02.04) - No longer need to half because Asura_PS2_GSMRenderer::SetMeshColour
//											will do this.
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::Platform_SetAmbientRGB( const unsigned char ucR, const unsigned char ucG, const unsigned char ucB, const unsigned char ucA)
{
	// (AW - 27.02.2004) - No longer half for PS2 as this is done in Asura_PS2_GSMRenderer::SetMeshColour.
	m_uAmbientARGB = (ucA << 24) | (ucR << 16) | (ucG << 8) | (ucB);
	m_uFlags &= ~ASURA_SHAPE_INSTANCE_USES_DEFAULT_AMBIENT;
}

/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Asura_Shape_Instance::Platform_SetDefaultAmbientRGB
// PARAMETERS	None
// RETURNS		None
// DESCRIPTION	
// AUTHOR		Richard M. Smith (9 August 2002)
// MODIFIED     Mike Healey (26.9.02) now calling the appropriate function to set up
//                          ambient to correct values for each platform.
// MODIFIED		Andrew Weinkove (01.03.04) Now just set a flag to say it uses default ambient.
//							If level has no entity ambient set, revert back to old system.
/////////////////////////////////////////////////////////////////////////////////////////
void Asura_Shape_Instance::Platform_SetDefaultAmbientRGB( void )
{
	m_uFlags |= ASURA_SHAPE_INSTANCE_USES_DEFAULT_AMBIENT;
} 
