//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_Multistream.h
//
// Created by Mike Healey.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Class to provide an Asura interface to PS2 multistream library
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_MULTISTREAM_H___
#define ___ASURA_PS2_MULTISTREAM_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Engine independent.
#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "FileAccess/Asura_FileAccess.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#define ASURA_PS2_DEBUGGING_DISABLE_MULTISTREAM 0

#define ASURA_PS2_MAX_CONCURRENT_SOUND_STREAMS (2)

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    // You can override this limit in your project's / platform's
    // Asura_EngineConfiguration.h.
    #ifndef ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS
        #define ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS (2)
    #endif
#endif

#define ASURA_PS2_VOICES_PER_CORE (24)
#define ASURA_PS2_VOICES_FOR_STREAMING (4)
#ifdef ASURA_PS2_SUPPORT_DTS	// DTS takes a voice on each core per sound
#define ASURA_PS2_MS_AVAILABLE_VOICES (ASURA_PS2_VOICES_PER_CORE - ASURA_PS2_VOICES_FOR_STREAMING)	// 20
#define ASURA_PS2_MS_DTS_FRONT_VOICE(iVoice) (iVoice)
#define ASURA_PS2_MS_DTS_REAR_VOICE(iVoice) ((iVoice) + ASURA_PS2_VOICES_PER_CORE)
#else
#define ASURA_PS2_MS_AVAILABLE_VOICES ((2*ASURA_PS2_VOICES_PER_CORE) - ASURA_PS2_VOICES_FOR_STREAMING)	// 44
#endif

// T Y P E S /////////////////////////////////////////////////////////////////

enum ASURA_PS2_STREAM_ID
{
    ASURA_PS2_STREAM_ID_SOUND_STREAM_0_L,
    ASURA_PS2_STREAM_ID_SOUND_STREAM_1_L,
    ASURA_PS2_STREAM_ID_SOUND_STREAM_0_R,   // child stream
    ASURA_PS2_STREAM_ID_SOUND_STREAM_1_R,   // child stream
    ASURA_PS2_STREAM_ID_SFX_LOADING,
    ASURA_PS2_STREAM_ID_FILELOAD_FIRST,
    ASURA_PS2_STREAM_ID_FILELOAD_LAST = ASURA_PS2_STREAM_ID_FILELOAD_FIRST + (ASURA_MAX_NO_OF_OPEN_FILES - 1),
#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST,
    ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_LAST = ASURA_PS2_STREAM_ID_PARTIALLY_STREAMED_SOUND_FIRST + (ASURA_PS2_MAX_CONCURRENT_PARTIALLY_STREAMED_SOUNDS - 1),
#endif
    ASURA_PS2_STREAM_ID_COUNT, // Must be last.
};

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_PS2_Multistream
{
    class ReplaceableSound;

public:

    enum ReverbType
    {
        REVERB_OFF,
        REVERB_ROOM,
        REVERB_STUDIO_A,
        REVERB_STUDIO_B,
        REVERB_STUDIO_C,
        REVERB_HALL,
        REVERB_SPACE,
        REVERB_ECHO,
        REVERB_DELAY,
        REVERB_PIPE,
    };

	enum EQ_FILTER_TYPE
	{
		// correspond to SOUND_FILTER_LOWPASS, SOUND_FILTER_HIPASS, SOUND_FILTER_BANDPASS
		FILTER_LOWPASS,
		FILTER_HIPASS,
		FILTER_BANDPASS,
	};

    static bool Initialise(int iMaxLoadedSpotFX);
    static void Shutdown();
    static void SetResetPoint();    // set pointer to SPU RAM at this point
    static void Reset();            // reset pointer to that marker.

 //   static void SetSoundGroupVolume(u_int uSoundGroup, u_int uVolume);  // 0-4096
 //   static u_int GetSoundGroupVolume(u_int uSoundGroup);
    
    // reverb settings SOUND_FX_OFF for no reverb, 
    // depth from -0x8000 to 0x7fff 
    // SOUND_FX_ECHO, SOUND_FX_DELAY can have delay 0-127
    // SOUND_FX_ECHO can have feedback 0-127
    static void SetReverb(u_int uType, u_int uDepth, u_int uDelay = 0, u_int uFeedback = 0);   

	// type from FILTER_LOWPASS, FILTER_HIPASS, FILTER_BANDPASS
	// freq from 0 to 512 ( not exactly sure about the scale, roughly 0Hz to 20kHz? )
	static void SetEQ(int iEffectSlot, EQ_FILTER_TYPE eFilterType, u_int uFilterFreq);
	static void SetEQMute(int iEffectSlot, bool bMute);
	static void SetVoiceUseEQ(int iVoiceNum, bool bUseEQ);
	// DisableDSPProcessing / EnableDSPProcessing - use to resolve conflicts for PCM channel usage.
	// Shutting it down can take a short while, hence the blocking parameter and query function.
	static void DisableDSPProcessing(bool bBlocking);
	static void EnableDSPProcessing();
	static bool IsDSPProcessingEnabled();

    static void Update();
    static void UpdateIOPBlocking();
    static int  GetNumFramesSinceIOPUpdate() { return s_iNumFramesSinceLastIOPUpdate; }


    static bool ReplaceSound(const int iExistingLoadedSoundID, const int iNewLoadedSoundID, u_short& usNewBasePitch);
    static bool LoadReplaceableSound(const int iLoadedSoundID, u_short& usNewBasePitch);
    static bool LoadSoundFromMemory( const char* vag_in_memory, const int num_bytes, const int loadedSoundIndex );
    static bool UnloadSound(const int loadedSoundIndex);
    static bool IsVoiceFree(int voice_num); // for checking independent resource allocation
    static bool IsSoundPlaying(int voice_num);

    static void PlaySound(int loadedSoundIndex, int voice_num, int volumeL, int volumeR, int pitch);
    static void PlayLoopingSound(int loadedSoundIndex, int voice_num, int volumeL, int volumeR, int pitch, u_int uLoopCount);// 0 = forever
    static void StopSound(int voice_num);
    static void SetVolume(int voice_num, int volumeL, int volumeR);
    static void SetPitch(int voice_num, int pitch);

    static bool OpenStreamingSoundFile(int iStreamID, const char* szFilename);
    static void PlayStreamingSound(int iStreamID, bool bLoop, bool bPreload, bool bMono);
    static void StopStreamingSound(int iStreamID);
    static void PauseStreamingSound(int iStreamID);
    static void ResumeStreamingSound(int iStreamID);
    static void CloseStreamingSoundFile(int iStreamID);
    static bool StreamingSoundIsPreloaded(int iStreamID);
    static void StartPreloadedSound(int iStreamID);
    static void ChangeMonoOrStereo(int iStreamID, bool bMono);
    static void SetStreamVolume(int iStreamID, int iVol);
	static bool IsStreamPlaying(int iStreamID);
	static float GetStreamPlayedSeconds(int iStreamID);
	// WARNING: this play offset only gets updated every time a buffer is filled
	// so it's not at all accurate. You may want to use GetStreamPlayedSeconds()
	// instead.
	static u_int GetStreamPlayOffset(int iStreamID);
	static u_int GetStreamFilesize(int iStreamID);
	static u_int GetStreamFrequency(int iStreamID);
	
    static int  GetReplaceableSoundSize(const int iLoadedSoundIndex);

    // (MP - 20.09.2002) Slightly hacky functionality for MPEG system to
    // determine what file and stream IDs to use.
    static int GetFirstUnusedFileIdAfterSounds( void );
    static int GetFirstUnusedStreamIdAfterSounds( void );
    static u_int AllocateTempSPURAM(u_int uBytes);
    static void FreeTempSPURAM(u_int uAddress);

	static void SetStreamingSoundsAsGlobal(bool bGlobalSounds) { s_bStreamingSoundsAreGlobal = bGlobalSounds; }
	static bool AreStreamingSoundsGlobal() { return s_bStreamingSoundsAreGlobal; }

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    struct PartiallyStreamedSound;
    struct ActivePartiallyStreamedSound;

    static bool IsLoadedIndexReplaceableSound( int iLoadedIndex );
	static bool IsLoadedIndexPartiallyStreamedSound(int iLoadedIndex);
	static bool IsLoadedIndexFullyStreamedSound(int iLoadedIndex);
    static bool PlayPartiallyStreamingSound( int iIdx, int iVoiceNum, int iVolumeL, int iVolumeR );
	static void StopPartiallyStreamingSound(int iVoiceNum);
	static bool IsPartiallyStreamedSoundPlaying(int iVoiceNum);
#endif

#ifdef ASURA_PS2_SUPPORT_DTS
	static bool IsDTSOn() { return s_bDTSOn; }
	static void SetDTS(bool bEnable) { s_bDTSOn = bEnable; }
#endif
	
    static u_int GetTotalFreeSPURAM( void );
	
private:

    // wrapper for SOUND_MemAllocate()
    static int AllocateSPURAM(u_int uBytes);

#ifdef ASURA_PS2_SUPPORT_PARTIALLY_STREAMED_SOUNDS
    friend class Asura_Chunk_ResourceFile_PS2_PartiallyStreamedSounds;

    // TEMP HACK FOR TESTING
    friend void SetupPartiallyStreamingSoundTest( void );

    // For use by chunk class.
    static void InitPartiallyStreamingSounds( u_int uNumPSSs, const char* szStreamFilename );
    static bool AddPartiallyStreamingSoundData( int iSoundID, int iFrequency, int iOffset1, int iSize1, int iOffset2, int iSize2, Asura_Hash_ID uFilenameHash );
    //static PartiallyStreamedSound* GetPartiallyStreamingSoundsArray( void );

    // Internal use.
    static void ResetPartiallyStreamedSoundState( void );
#endif

    static int  s_iNumFramesSinceLastIOPUpdate;
    static u_int* s_auSPUAddressOfLoadedSound;
    
    static int s_iFastLoadStackAddress;
#ifdef ASURA_PS2_SUPPORT_DTS
	static bool s_bDTSOn;
#endif
	static bool s_bDSPOn;
	static bool s_bStreamingSoundsAreGlobal;

    static ReplaceableSound* s_axReplaceableSounds;
	static u_int s_uEQOnBitfield;	// bitfield for EQ on voices 0-23
};
// End class Asura_PS2_Multistream.

class Asura_PS2_Multistream_FastLoad
{
    public:

        // C O N S T A N T S /////////////////////////////////////////////////

        enum
        {
            REACHED_EOF = -1,
        };

        // C L A S S   M E T H O D S /////////////////////////////////////////

        // Standard functionality.
        static bool OpenFile( Asura_Handle iHandle, const char* szFilename );
        static void Seek( Asura_Handle iHandle, u_int uByteOffset );
        static int  GetData( Asura_Handle iHandle, const void** ppvBfr, bool bCancelLoad = false );
        static void CloseFile( Asura_Handle iHandle );

		// function for asynchronous loading - preload an entire file to a 
		// dynamically-allocated buffer and allow seeking, GetData etc. after.
		static bool OpenFileForPreload( Asura_Handle iHandle, const char* szFilename );
		static bool IsPreloadFinished( Asura_Handle iHandle );
		static u_int GetNumBytesPreloaded( Asura_Handle iHandle );

		// functions to allow sceOpen to be used without conflict
		static void PauseFastload();
		static void ResumeFastload();

        // Additional query methods for open files.
        static u_int GetFilePos( Asura_Handle iHandle );
        static u_int GetFileSize( Asura_Handle iHandle );

		static bool AreAnyFilesOpen( void ) { return s_uNumOpenFiles != 0; }

    private:

        // F R I E N D S /////////////////////////////////////////////////////

        friend Asura_PS2_Multistream;

        // C L A S S   M E T H O D S /////////////////////////////////////////

        // Called by Asura_PS2_Multistream.
        static bool Initialise( void );
        static void Shutdown( void );

		// D A T A ///////////////////////////////////////////////////////////
		static u_int s_uNumOpenFiles;
		static u_int s_uLastFastloadMode;

};
// End class Asura_PS2_Multistream_FastLoad.

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_MULTISTREAM_H___
