//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS2_SkyboxDefines.h
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     Defines shared between VCL program and EE-side code.
//
// Memory layout:
//
//     Given s subdivisions per edge, we have 2^2s quads, 2^(2s+1) tris and
//     ((2^s)+1)^2 vertices per face.
//
//     However, since we're stripping this stuff, each row of quads will have
//     to send both the top and bottom vertices, meaning we transfer
//     (2((2^s)+1))(2^s) vertices.
//
//     To allow arbitrary levels of subdivision when the geometry is
//     generated on the EE, we'll double buffer the input and upload it in
//     batches as necessary.
//
//     Indexing the input vertices would a possibility and would save
//     bandwidth but it would also make the batching rather more complicated.
//
//     Given v verts per batch, our memory usage in QWs is:
//
//         rotation matrices: 6*3           [one per face but don't need translation row]
//         camera data:       4*(4+2+1+1+2) [W->CG, GC->S scale&trans, GC->SC scale, pos, SCISSOR, XYOFFSET]
//         GIF tags:          2             [GS setup,geometry]
//         input buffers:     2*2v          [STQ,XYZ]
//         output buffers:    2*(4+3v)      [GIFtag,SCISSOR,XYOFFSET,GIFtag,STQ,RGBA,XYZ]
//         clip workspace:    2*9*2
//
//     Hence 10v + 104 <= 1024 => v <= 92.
//
//     Our layout therefore looks like this:
//
//        0    /--------------------------\ 
//             |     rotation matrices    |   18 QWs -
//       18    |--------------------------|
//             |        camera data       |   40 QWs - [W->CG, GC->S scale&trans, GC->SC scale, pos, SCISSOR, XYOFFSET] * 4
//       58    |--------------------------|
//             |         GIF tags         |    2 QWs - GS setup [NLOOP 2, EOP 1, NREG 1:A+D], geometry [NLOOP code, EOP 1, NREG 3:ST,RGBAQ,XYZ]
//       60    |--------------------------|
//             |      input buffer 0      |  184 QWs - [STQ,XYZ] * v
//      244    |--------------------------|
//             |      input buffer 1      |  184 QWs - [STQ,XYZ] * v
//      428    |--------------------------|
//             |      output buffer 0     |  280 QWs - [GIFtag,SCISSOR,XYOFFSET,GIFtag,[STQ,RGBA,XYZ]*v]
//      708    |--------------------------|
//             |      output buffer 1     |  280 QWs - [GIFtag,SCISSOR,XYOFFSET,GIFtag,[STQ,RGBA,XYZ]*v]
//      988    |--------------------------|
//             |     clip workspace 0     |   18 QWs -
//     1006    |--------------------------|
//             |     clip workspace 1     |   18 QWs -
//     1024    \--------------------------/
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS2_SKYBOXDEFINES_H___
#define ___ASURA_PS2_SKYBOXDEFINES_H___ 1

// M A C R O S ///////////////////////////////////////////////////////////////

#define ASURA_PS2_SKYBOX_NUM_FACES            (    6)
#define ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_IN   (    2)
#define ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_OUT  (    3)
#define ASURA_PS2_SKYBOX_NUM_QWS_PER_VTX_CLIP (    2)

#define ASURA_PS2_SKYBOX_BATCH_MAX_VERTS      (   92)

#define ASURA_PS2_SKYBOX_VIS_FLAG_SHIFT       (    3)
#define ASURA_PS2_SKYBOX_FIRST_VIS_FLAG       (0x1 << ASURA_PS2_SKYBOX_VIS_FLAG_SHIFT)
#define ASURA_PS2_SKYBOX_MASK_VIS_FLAGS       (0x3F << ASURA_PS2_SKYBOX_VIS_FLAG_SHIFT)
#define ASURA_PS2_SKYBOX_MASK_MAT_PTR_UPDATES ((1 << ASURA_PS2_SKYBOX_VIS_FLAG_SHIFT) - 1)

#define ASURA_PS2_SKYBOX_SIZE_ROT_MATS        (   18)
#define ASURA_PS2_SKYBOX_SIZE_CAM_DATA        (   40)
#define ASURA_PS2_SKYBOX_SIZE_GIF_TAGS        (    2)
#define ASURA_PS2_SKYBOX_SIZE_INPUT_BFR       (  184)
#define ASURA_PS2_SKYBOX_SIZE_OUTPUT_BFR      (  280)
#define ASURA_PS2_SKYBOX_SIZE_CLIP_WORKSP     (   27)

#define ASURA_PS2_SKYBOX_ADDR_ROT_MATS        (    0)
#define ASURA_PS2_SKYBOX_ADDR_CAM_DATA        (   18)
#define ASURA_PS2_SKYBOX_ADDR_GIF_TAGS        (   58)
#define ASURA_PS2_SKYBOX_ADDR_INPUT_BFR_0     (   60)
#define ASURA_PS2_SKYBOX_ADDR_INPUT_BFR_1     (  244)
#define ASURA_PS2_SKYBOX_ADDR_OUTPUT_BFR_0    (  428)
#define ASURA_PS2_SKYBOX_ADDR_OUTPUT_BFR_1    (  708)
#define ASURA_PS2_SKYBOX_ADDR_CLIP_WORKSP_0   (  988)
#define ASURA_PS2_SKYBOX_ADDR_CLIP_WORKSP_1   ( 1006)

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS2_SKYBOXDEFINES_H___
