//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Corona.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     PS2 implementation of corona system rendering.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "SpecialFX/Asura_Corona_System.h"

// Engine platform-specific.
#include "Core/Asura_PS2_DMA.h"
#include "Rendering/Asura_PlatformRendering.h"
#include "Rendering/Asura_PlatformRenderStates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"
#include "Rendering/Asura_PS2_VU1.h"
#include "SpecialFX/Asura_PS2_CoronaDefines.h"

// M A C R O S ///////////////////////////////////////////////////////////////

#ifndef ASURA_PS2_DEFERRED_RENDERING
#error Not implemented.
#endif

// C O N S T A N T S /////////////////////////////////////////////////////////

static const u_int MAX_CORONAS_PER_FRAME                  = 300;
static const u_int MAX_DISTINCT_CORONA_TEXTURES_PER_FRAME = 16;

// T Y P E S /////////////////////////////////////////////////////////////////

struct CoronaData
{
    Asura_PS2_QW xPosAndSize;
    Asura_PS2_QW xRGBAndVariousFlags; // w component is (((~camvisflags & 0xF) << 1) | fadeout), where fadeout is 1 iff the corona should fade out with distance.
    Asura_PS2_QW xDirAndUseDirFlag;   // w component is 0 if should use dir effect and 128 otherwise.
};

struct CoronaTex
{
    int   iTexID;
    u_int uFirstQueueEntry;
    u_int uLastQueueEntry;
};

// G L O B A L S /////////////////////////////////////////////////////////////

static CoronaData*  g_pxCoronaData = NULL;
static CoronaData** g_ppxCoronaQueue;      // Sorted version of g_pxCoronaData.
static u_int*       g_puNumQueuedCoronas;
static CoronaTex*   g_pxCoronaTextures;
static u_int*       g_puNumCoronaTextures;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void CopyCoronaDataToVifBuffer( const CoronaData* pxData, Asura_PS2_QW** ppxVifBfrOut )
{
    u_long128 uPosAndSize         = pxData->xPosAndSize.u128;
    u_long128 uRGBAndVariousFlags = pxData->xRGBAndVariousFlags.u128;
    u_long128 uDirAndUseDirFlag   = pxData->xDirAndUseDirFlag.u128;

    ((*ppxVifBfrOut)++)->u128 = uPosAndSize;
    ((*ppxVifBfrOut)++)->u128 = uRGBAndVariousFlags;
    ((*ppxVifBfrOut)++)->u128 = uDirAndUseDirFlag;
}

static void FinishBatch( Asura_PS2_QW* pxVifBfrOut, bool bIsEndOfTex )
{
    pxVifBfrOut->u128 = 0;

    pxVifBfrOut->u32s[0] = SCE_VIF1_SET_MSCNT( false );

    if( bIsEndOfTex )
    {
        pxVifBfrOut->u32s[1] = SCE_VIF1_SET_FLUSH( false );
    }
}

static void InitVU1( void );

static void StartNewBatch( u_int uCount, Asura_PS2_QW** pxOut );

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_Corona_System::Platform_Initialise()
{
}

void Asura_Corona_System::Platform_Shutdown()
{
}

void Asura_Corona_System::Platform_RenderStart( void )
{
    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_DRAW_CORONAS) ) { return; }

    // Ensure our copy of the fade parameters in Asura_PS2_CoronaDefines.h
    // are kept in sync with independent code.
    Asura_Assert( ASURA_CORONA_FADE_DIST_BEGIN == ASURA_PS2_CORONA_FADE_DIST_BEGIN && ASURA_CORONA_FADE_DIST_END == ASURA_PS2_CORONA_FADE_DIST_END, "Need to update PS2 corona fade parameters." );

    Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_LARGE );

    Asura_PS2_VU1::LoadRenderer( Asura_PS2_VU1::RENDERER_CORONAS );

    InitVU1();

    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1 );
    Asura_PlatformRenderStates::SetZBufferWriteEnabled( false );
    Asura_PlatformRenderStates::SetTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
    Asura_PlatformRenderStates::ApplyRenderStates();

    {
        u_int uSPRBytesReqd = MAX_CORONAS_PER_FRAME * sizeof(CoronaData)
                            + MAX_CORONAS_PER_FRAME * sizeof(CoronaData*)
                            + sizeof(u_int)
                            + MAX_DISTINCT_CORONA_TEXTURES_PER_FRAME * sizeof(CoronaTex)
                            + sizeof(u_int);

        g_pxCoronaData        = reinterpret_cast<CoronaData* >( Asura_PS2_SPR_Alloc(0, ((uSPRBytesReqd + 15) >> 4)) );
        g_ppxCoronaQueue      = reinterpret_cast<CoronaData**>( g_pxCoronaData + MAX_CORONAS_PER_FRAME );
        g_puNumQueuedCoronas  = reinterpret_cast<u_int*      >( g_ppxCoronaQueue + MAX_CORONAS_PER_FRAME );
        g_pxCoronaTextures    = reinterpret_cast<CoronaTex*  >( g_puNumQueuedCoronas + 1 );
        g_puNumCoronaTextures = reinterpret_cast<u_int*      >( g_pxCoronaTextures + MAX_DISTINCT_CORONA_TEXTURES_PER_FRAME );
    }

    *g_puNumQueuedCoronas  = 0;
    *g_puNumCoronaTextures = 0;
}

void Asura_Corona_System::Platform_RenderCorona( const Asura_Corona* pxCorona, u_int uCamVisFlags )
{   
    if( !g_pxCoronaData ) { return; }

    const int iTexID = pxCorona->m_iTextureIndex;
    Asura_Assert( ASURA_TEXTURE_NONE != iTexID, "Untextured coronas are not supported." );

    if( *g_puNumQueuedCoronas >= MAX_CORONAS_PER_FRAME || *g_puNumCoronaTextures >= MAX_DISTINCT_CORONA_TEXTURES_PER_FRAME )
    {
        Asura_Assert( false, "Too many coronas or corona textures this frame." );
        return;
    }

    u_int uDataIdx = (*g_puNumQueuedCoronas)++;

    u_int uTexEntry = 0;
    for( ; uTexEntry < *g_puNumCoronaTextures; ++uTexEntry )
    {
        if( g_pxCoronaTextures[uTexEntry].iTexID == iTexID ) { break; }
    }

    u_int uQueueIdx;
    if( uTexEntry == *g_puNumCoronaTextures )
    {
        uQueueIdx = uDataIdx;

        g_pxCoronaTextures[uTexEntry].iTexID           = iTexID;
        g_pxCoronaTextures[uTexEntry].uFirstQueueEntry = uQueueIdx;
        g_pxCoronaTextures[uTexEntry].uLastQueueEntry  = uQueueIdx;

        ++(*g_puNumCoronaTextures);
    }
    else
    {
        // We want to insert the corona after the last existing queue entry.
        // Unfortunately, this means shuffling later entries around.
        for( u_int u = *g_puNumCoronaTextures - 1; u > uTexEntry; --u )
        {
            // Move first entry for this texture to the position after its
            // last entry. This will leave space for the next texture down to
            // move or insert an entry at the end of its section.
            g_ppxCoronaQueue[++g_pxCoronaTextures[u].uLastQueueEntry] = g_ppxCoronaQueue[g_pxCoronaTextures[u].uFirstQueueEntry++];
        }

        uQueueIdx = ++g_pxCoronaTextures[uTexEntry].uLastQueueEntry;
    }

    // Store corona data.
    {
        g_pxCoronaData[uDataIdx].xPosAndSize.f32s[0] = pxCorona->m_xPosition.x;
        g_pxCoronaData[uDataIdx].xPosAndSize.f32s[1] = pxCorona->m_xPosition.y;
        g_pxCoronaData[uDataIdx].xPosAndSize.f32s[2] = pxCorona->m_xPosition.z;
        g_pxCoronaData[uDataIdx].xPosAndSize.f32s[3] = pxCorona->m_fRadius;

        g_pxCoronaData[uDataIdx].xRGBAndVariousFlags.f32s[0] = pxCorona->m_fRed;
        g_pxCoronaData[uDataIdx].xRGBAndVariousFlags.f32s[1] = pxCorona->m_fGreen;
        g_pxCoronaData[uDataIdx].xRGBAndVariousFlags.f32s[2] = pxCorona->m_fBlue;
        g_pxCoronaData[uDataIdx].xRGBAndVariousFlags.u32s[3] = ((~uCamVisFlags & 0xF) << 1) | (pxCorona->m_bDoNotFadeOutWithDistance ? 0 : 1);

        g_pxCoronaData[uDataIdx].xDirAndUseDirFlag.u128 = 0;
        if( pxCorona->m_bUseDirectionEffect )
        {
            g_pxCoronaData[uDataIdx].xDirAndUseDirFlag.f32s[0] = pxCorona->m_xDirection.x;
            g_pxCoronaData[uDataIdx].xDirAndUseDirFlag.f32s[1] = pxCorona->m_xDirection.y;
            g_pxCoronaData[uDataIdx].xDirAndUseDirFlag.f32s[2] = pxCorona->m_xDirection.z;
        }
        else
        {
            g_pxCoronaData[uDataIdx].xDirAndUseDirFlag.f32s[3] = 128.0f;
        }
    }

    // And store ptr to that in our sorted queue.
    g_ppxCoronaQueue[uQueueIdx] = &g_pxCoronaData[uDataIdx];

    Asura_PS2_Render::UpdateTriStatsFromVisFlags( 2, 1, uCamVisFlags );
}

void Asura_Corona_System::Platform_RenderEnd( void )
{
    if( !g_pxCoronaData ) { return; }

    u_int uCorona= 0;

    for( u_int uTex = 0; uTex < *g_puNumCoronaTextures; ++uTex )
    {
        const CoronaTex& xTex = g_pxCoronaTextures[uTex];

        Asura_PS2_TextureManagement::SetupTextureEnvironment( Asura_PS2_TextureManagement::UploadTexture(xTex.iTexID),
                                                              ASURA_PS2_TEXREGFLAG_ALL,
                                                              xTex.iTexID,
                                                              Asura_PS2_TextureManagement::BFR_AUTO,
                                                              SCE_GS_LINEAR,
                                                              SCE_GS_LINEAR_MIPMAP_NEAREST );

        u_int uCoronasLeftForThisTex = (xTex.uLastQueueEntry + 1) - xTex.uFirstQueueEntry;
        while( uCoronasLeftForThisTex )
        {
            u_int uBatchSize = Asura_Maths::Min( uCoronasLeftForThisTex, static_cast<u_int>(ASURA_PS2_CORONA_MAX_CORONAS_PER_BATCH) );

            uCoronasLeftForThisTex -= uBatchSize;

            Asura_PS2_QW* pxOut;
            StartNewBatch( uBatchSize, &pxOut );

            for( u_int u = uBatchSize; u--; )
            {
                CopyCoronaDataToVifBuffer( g_ppxCoronaQueue[uCorona++], &pxOut );
            }

            FinishBatch( pxOut, !uCoronasLeftForThisTex );
        }
    }

    Asura_Assert( uCorona == *g_puNumQueuedCoronas, "" );

    Asura_PS2_SPR_Dealloc( reinterpret_cast<Asura_PS2_QW*>(g_pxCoronaData) );
    g_pxCoronaData = NULL;
}

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void InitVU1( void )
{
    u_int uNumActiveCameras = 0;
    for( int i = 0; i < Asura_Camera::GetNumberOfCameras(); ++i )
    {
        Asura_Camera::SetCameraIndex( i );

        if( Asura_Camera::IsCameraActive() ) { ++uNumActiveCameras; }
    }

    const u_int uNumQWsOfCamData = uNumActiveCameras * ASURA_PS2_CORONA_NUM_QWS_PER_CAM;
    const u_int uNumQWsOfGIFTags = 2;
    const u_int uTotalQWs        = 1 + uNumQWsOfCamData + 1 + uNumQWsOfGIFTags + 1; // The 1s are for the 2 DMA tags and extra VIF codes at the end.

    Asura_PS2_QW* const pxBase = Asura_PS2_DeferredRendering::GetVifBuffer( uTotalQWs );
    Asura_PS2_QW* pxOut = pxBase;
    
    // DMA tag to transfer cam matrices, positions and respective corona
    // axes.
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumQWsOfCamData,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         SCE_VIF1_SET_STCYCL(1, 1, false),
                                                         SCE_VIF1_SET_UNPACK(ASURA_PS2_CORONA_CAM_DATA_BASE, uNumQWsOfCamData, CMD_V4_32, false) );

    for( u_int uCam = 0; uCam < uNumActiveCameras; ++uCam )
    {
        Asura_PS2_QW* const pxOutBeforeThisCam = pxOut;

        Asura_Camera::SetCameraIndex( uCam );
        if( !Asura_Camera::IsCameraActive() ) { continue; }

        // Viewport setup.
        pxOut = Asura_PS2_GS::GetViewportSetupGifCmds( uCam, Asura_PS2_GS::CONTEXT_1, pxOut );

        // Matrices.
        {
            const u_long128* puW2S = reinterpret_cast<const u_long128*>( Asura_PS2_VU1::GetWorldToScreenMat() );
            for( u_int u = 4; u--; )
            {
                (pxOut++)->u128 = *puW2S++;
            }

            const u_long128* puW2C = reinterpret_cast<const u_long128*>( Asura_PS2_VU1::GetWorldToClipMatGuardband() );
            for( u_int u = 4; u--; )
            {
                (pxOut++)->u128 = *puW2C++;
            }
        }

        // Cam pos.
        {
            const Asura_Vector_3* pxCamPos = Asura_Camera::GetPosition();
            pxOut->f32s[0] = pxCamPos->x;
            pxOut->f32s[1] = pxCamPos->y;
            pxOut->f32s[2] = pxCamPos->z;
//            pxOut->f32s[3] = 1.0f;
            ++pxOut;
        }

        // X and Y axes for particle.
        {
            const Asura_Matrix_3x3* pxCamOri = Asura_Camera::GetOrientation();

            pxOut->f32s[0] = pxCamOri->e11;
            pxOut->f32s[1] = pxCamOri->e21;
            pxOut->f32s[2] = pxCamOri->e31;
//            pxOut->u32s[3] = uNumActiveCameras;
            ++pxOut;

            pxOut->f32s[0] = pxCamOri->e12;
            pxOut->f32s[1] = pxCamOri->e22;
            pxOut->f32s[2] = pxCamOri->e32;
//            pxOut->f32s[3] = 0.0f;
            ++pxOut;
        }

        Asura_Assert( pxOut - pxOutBeforeThisCam == ASURA_PS2_CORONA_NUM_QWS_PER_CAM, "" );
    }

    // DMA tag to transfer GIF tags and extra VIF codes. Start setting up the
    // double buffer in the spare word.
    (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( uNumQWsOfGIFTags + 1,
                                                         Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                         Asura_PS2_DmaTag::ID_CNT,
                                                         false,
                                                         0,
                                                         false,
                                                         SCE_VIF1_SET_BASE(ASURA_PS2_CORONA_INPUT_BUFFERS_BASE, false),
                                                         SCE_VIF1_SET_UNPACK(ASURA_PS2_CORONA_INPUT_BUFFERS_BASE, uNumQWsOfGIFTags, CMD_V4_32, false) );

    // Cam setting GIF tag.
    (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( Asura_PS2_GS::NUM_QWS_FOR_VIEWPORT_SETUP_PER_CONTEXT, // NLOOP
                                                         false,                                                // EOP
                                                         false,                                                // PRE
                                                         0,                                                    // PRIM
                                                         SCE_GIF_PACKED,                                       // FLG
                                                         1,                                                    // NREG
                                                         SCE_GIF_PACKED_AD );                                  // REG0

    // GIF tag for the actual quads.
    {
        const u_int uPRIM = SCE_GS_SET_PRIM( SCE_GS_PRIM_TRISTRIP, // PRIM
                                             0,                    // IIP
                                             1,                    // TME
                                             0,                    // FGE
                                             1,                    // ABE
                                             0,                    // AA1
                                             0,                    // FST
                                             0,                    // CTXT
                                             0 );                  // FIX

        (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( 0,                   // NLOOP (will be fixed up in microcode per strip)
                                                             true,                // EOP
                                                             true,                // PRE
                                                             uPRIM,               // PRIM
                                                             SCE_GIF_PACKED,      // FLG
                                                             9,                   // NREG
                                                             SCE_GS_RGBAQ,        // REG0
                                                             SCE_GS_ST,           //
                                                             SCE_GS_XYZ2,         //
                                                             SCE_GS_ST,           //  .
                                                             SCE_GS_XYZ2,         //  .
                                                             SCE_GS_ST,           //  .
                                                             SCE_GS_XYZ2,         //
                                                             SCE_GS_ST,           //
                                                             SCE_GS_XYZ2 );       // REG8
    }

    // Finish setting up double buffer, set camera count and kick off the
    // microprogram initialisation routine.
    pxOut->u32s[0] = SCE_VIF1_SET_OFFSET( ASURA_PS2_CORONA_INPUT_BUFFERS_SIZE, 0 );
    pxOut->u32s[1] = SCE_VIF1_SET_ITOP( uNumActiveCameras, false );
    pxOut->u32s[2] = SCE_VIF1_SET_MSCALF( 0, false );
    pxOut->u32s[3] = SCE_VIF1_SET_NOP( false );
    ++pxOut;

    Asura_Assert( static_cast<u_int>(pxOut - pxBase) == uTotalQWs, "" );
}

static void StartNewBatch( u_int uCount, Asura_PS2_QW** pxOut )
{
    u_int uCoronaDataSize = uCount * ASURA_PS2_CORONA_NUM_QWS_PER_INPUT_CORONA;
    u_int uDmaQWs         = uCoronaDataSize + 1; // Extra QW for terminal VIF cmds - MSCNT, etc.

    Asura_PS2_QW* pxBatchStart = Asura_PS2_DeferredRendering::GetVifBuffer( 1 + uDmaQWs );

    (reinterpret_cast<Asura_PS2_DmaTag*>(pxBatchStart))->Set( uDmaQWs,
                                                              Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                              Asura_PS2_DmaTag::ID_CNT,
                                                              false,
                                                              0,
                                                              false,
                                                              SCE_VIF1_SET_ITOP(uCount, false),
                                                              SCE_VIF1_SET_UNPACKR(0, uCoronaDataSize, Asura_PS2_VU1::VIFPK_V4_32, false) );

    *pxOut = pxBatchStart + 1;
}
