//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Weather_SunGlare.cpp
//
// Copyright (c) 2003 Rebellion. All rights reserved.
//
// Description:
//
//     PS2 implementation of sun glare special effect.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib platform-specific.
#include <eestruct.h>
#include <libgraph.h>

// Engine independent.
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureManagement.h"
#include "SpecialFX/Asura_Weather_System.h"

// Engine platform-specific.
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_PS2_DMA.h"
#include "Rendering/Asura_PlatformRenderstates.h"
#include "Rendering/Asura_PlatformTextureManagement.h"
#include "Rendering/Asura_PS2_2D_Graphics.h"
#include "Rendering/Asura_PS2_DeferredRendering.h"
#include "Rendering/Asura_PS2_GS.h"
#include "Rendering/Asura_PS2_RenderingOptions.h"

// G L O B A L S /////////////////////////////////////////////////////////////

static void* g_pvSunChainPartOne = NULL;
static u_int g_uSizeOfSunChainPartOne;
static void* g_pvSunChainPartTwo = NULL;
static u_int g_uSizeOfSunChainPartTwo;

#if _DEBUG
static bool g_bSunForceChainReset = false;
#endif

static int g_iSunCachedNumCams = 0;

static bool g_bAlphaSetForLensFlare = false;

//
// Debugging stuff.
//

// Setting this to an invalid value, ie. anything apart from powers of 2
// between 1 and 256 inclusive, will break everything horribly.
DEBUGVAR u_int dbg_uSunOSBWidthToDisplay = 0; // 0 => none

DEBUGVAR bool dbg_bSunOSBDisplayShouldBeColoured = false;

DEBUGVAR bool dbg_bSunEnableFinalGlareSprite = true;

DEBUGVAR bool dbg_bSunFan = true;

//
// Render parameters.
//

DEBUGVAR bool  dbg_bSunBlendDownsampledBuffersIntoDrawBuffer = true;

DEBUGVAR u_int dbg_uSunGlareFilter      = SCE_GS_LINEAR;
DEBUGVAR u_int dbg_uSunDownsampleFilter = SCE_GS_LINEAR;

DEBUGVAR float dbg_fSunViewLimitScale = 0.845f;

DEBUGVAR float dbg_fSunMaxRadius = 350.0f;
DEBUGVAR u_int dbg_uSunZ         = 0x00000010;

DEBUGVAR u_int dbg_uSunCentreAlpha = 255;
DEBUGVAR u_int dbg_uSunEdgeAlpha   = 0;

DEBUGVAR u_int dbg_uBleedBlendUVInset = 4;

DEBUGVAR u_int dbg_uSunFinalGlareSpriteRGBMul = 3;

DEBUGVAR float dbg_fHorizonDist = 1.0f;
DEBUGVAR float dbg_fHorizonIncrement = 0.3f;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void GenerateDmaChains( const Asura_ViewingArea* pxCamViewingArea, float fR, float fG, float fB );
static void GenerateDmaChainsIfNecessary( void );


// S T A T I C   F U N C T I O N S ///////////////////////////////////////////
// function: LineIsOutsideScreen()
// description: returns true iff all points on line between 2 points are outside
// screen rectangle [0,0]-[1,1]
static bool LineIsOutsideScreen( const Asura_Vector_2& xPoint1, const Asura_Vector_2& xPoint2 )
{
	// first check if either point is on screen
	if ( xPoint1.x >= 0.0f && xPoint1.x <= 1.0f && xPoint1.y >= 0.0f && xPoint1.y <= 1.0f ) 
	{
		return false;
	}
	if ( xPoint2.x >= 0.0f && xPoint2.x <= 1.0f && xPoint2.y >= 0.0f && xPoint2.y <= 1.0f )
	{
		return false;
	}

	// both points are offscreen.

	// check if both points are outside any screen edge
	if ( xPoint1.x < 0.0f && xPoint2.x < 0.0f ) { return true; }
	if ( xPoint1.y < 0.0f && xPoint2.y < 0.0f ) { return true; }
	if ( xPoint1.x > 1.0f && xPoint2.x > 1.0f ) { return true; }
	if ( xPoint1.y > 1.0f && xPoint2.y > 1.0f ) { return true; }

	// not so simple: find intersection of line with each screen edge and see if that point is offscreen
	// use formula xIntersection = xPoint1 + fLamba * xDelta
	// fLamba = (xIntersection - xPoint1) / xDelta
	const Asura_Vector_2 xDelta = xPoint2 - xPoint1;
	float fLambda;
	Asura_Vector_2 xIntersection;

	// if we have an almost horizontal or vertical line (near-zero in delta component) then
	// the line must be grazing the screen edge they are stradling. Call this offscreen for efficiency.

	// MC (20/06/08): Doesn't cover the case where one point is above the screen, one is below; catch those cases.
	if ( (Asura_Maths::Abs(xDelta.x) < 0.001f) )
	{
		// Vertical line
		// Verify the line comes entirely between the left and right edge of the screen
		if (xPoint1.x > 0.0f && xPoint1.x < 1.0f && xPoint2.x > 0.0f && xPoint2.x < 1.0f)
		{
			return false;
		}
		return true;
	}

	if ( (Asura_Maths::Abs(xDelta.y) < 0.001f) )
	{
		// Horizontal line
		// Verify the line comes entirely between the top and bottom edge of the screen
		if (xPoint1.y > 0.0f && xPoint1.y < 1.0f && xPoint2.y > 0.0f && xPoint2.y < 1.0f)
		{
			return false;
		}
		return true;
	}

	// x = 0
	xIntersection.x = 0.0f;
	fLambda = (xIntersection.x - xPoint1.x) / xDelta.x;
	xIntersection.y = xPoint1.y + (fLambda * xDelta.y);
	if (xIntersection.y < 0.0f || xIntersection.y > 1.0f) { return true; }

	// x = 1
	xIntersection.x = 1.0f;
	fLambda = (xIntersection.x - xPoint1.x) / xDelta.x;
	xIntersection.y = xPoint1.y + (fLambda * xDelta.y);
	if (xIntersection.y < 0.0f || xIntersection.y > 1.0f) { return true; }

	// y = 0
	xIntersection.y = 0.0f;
	fLambda = (xIntersection.y - xPoint1.y) / xDelta.y;
	xIntersection.x = xPoint1.x + (fLambda * xDelta.x);
	if (xIntersection.x < 0.0f || xIntersection.x > 1.0f) { return true; }

	// y = 1
	xIntersection.y = 1.0f;
	fLambda = (xIntersection.y - xPoint1.y) / xDelta.y;
	xIntersection.x = xPoint1.x + (fLambda * xDelta.x);
	if (xIntersection.x < 0.0f || xIntersection.x > 1.0f) { return true; }

	return false;
}

//
// function: RenderSunGlare()
// description: renders either a rough circle or a horizon rectangle into the 
// alpha channel of the Z-buffer, downsamples this to temp buffers, adding a 
// bit each time to the frame buffer.
static void RenderSunGlare( bool bHorizon )
{
	g_bAlphaSetForLensFlare = false;

    if( !Asura_PS2_RenderingOptions::GetOption(Asura_PS2_RenderingOptions::OPTION_ENABLE_SUNGLARE) ) { return; }

    GenerateDmaChainsIfNecessary();

    u_int uNumCamsRendered = 0;
    
    // (MP - 01.10.2003) I've changed this so that we use the z-buffer for
    // generating the alpha channel. We can't use the display buffer since
    // that may only be 16-bit. Of course, this new approach only works if
    // we've got a 24-bit z buffer.
    Asura_Assert( 24 == Asura_PS2_GS::GetZBfrBitDepth(), "" );

    const u_int uScreenWidth   = Asura_PS2_GS::GetScreenWidth();
    const u_int uScreenHeight  = Asura_PS2_GS::GetScreenHeight();
    const float fScreenWidth   = static_cast<float>( static_cast<int>(uScreenWidth ) );
    const float fScreenHeight  = static_cast<float>( static_cast<int>(uScreenHeight) );
    const u_int uTempBufferFBP = Asura_PS2_GS::GetZBfrBasePtr();
    const u_int uDrawBufferFBP = Asura_PS2_GS::GetDrawBfrFBP();
    const u_int uTempBufferTW  = Asura_Maths::GetCeilLog2( uScreenWidth  );
    const u_int uTempBufferTH  = Asura_Maths::GetCeilLog2( uScreenHeight );
    const u_int uDrawBfrPSM    = Asura_PS2_GS::GetDrawBfrPSM();
    const bool  bModifyAlpha   = Asura_PS2_RenderingOptions::GetOption( Asura_PS2_RenderingOptions::OPTION_SUNGLARE_LEAVE_BRIGHTNESS_VALUE_IN_ALPHA );

    // Iterate through cameras, running through the entire process for each
    // one. We could possibly speed this up by performing some of the steps
    // only once for all cameras but we're likely to save more time in general
    // by not burning any fillrate for the cameras that can't see the sun.
    for( int iCam = 0; iCam < Asura_Camera::GetNumberOfCameras(); ++iCam )
    {
        Asura_Camera::SetCameraIndex( iCam );
        if( !Asura_Camera::IsCameraActive() ) { continue; }

        // Calculate view to sun angle.
        Asura_Vector_3 xCameraVec;
        Asura_Vector_3 xSunVector = Asura_Weather_System::GetSunPos() - *Asura_Camera::GetPosition();
        xSunVector.Normalise();
        Asura_Camera::GetForwardVector( xCameraVec );

        // Early reject for offscreen sun?
		const bool bDoHorizonGlare = bHorizon;
		const float fSunViewLimit = Asura_Maths::Cosine(Asura_Camera::GetHorizontalFieldOfView() * dbg_fSunViewLimitScale);
        const float fSunAngle = xSunVector * xCameraVec;
        const bool bDoSun = (fSunAngle >= fSunViewLimit) && !bHorizon;
		if ( (!bDoSun) && (!bDoHorizonGlare) )
		{
			continue;
		}
    
        if( !uNumCamsRendered )
        {
            Asura_PS2_DeferredRendering::CancelLastPath3Gate();

            Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_LARGE );
            Asura_PS2_TextureManagement::InvalidateAllTextureBuffers();
        }

        const Asura_ViewingArea* pxCamViewingArea = Asura_Camera::GetViewingArea();
        const u_int uVAWidth       = Asura_Maths::FloatToInt( fScreenWidth  * pxCamViewingArea->m_fWidth   );
        const u_int uVAHeight      = Asura_Maths::FloatToInt( fScreenHeight * pxCamViewingArea->m_fHeight  );
        const float fVAWidth       = uVAWidth;
        const float fVAHeight      = uVAHeight;
        const u_int uVAPrimOriginX = ASURA_PS2_SCREEN_CENTRE_X - (uVAWidth  >> 1);
        const u_int uVAPrimOriginY = ASURA_PS2_SCREEN_CENTRE_Y - (uVAHeight >> 1);

        // Calculate sun radius based on angle to it.
        const float fSunRadius = dbg_fSunMaxRadius * ((fSunAngle - fSunViewLimit) / (1.0f - fSunViewLimit));

        const u_int uNumSunSlices = 16;

        const u_int uNLOOPForPrePartOne = 4;                    // FRAME, XYOFFSET and SCISSOR registers for context 1 and TEX0 for context 2

		const u_int uQWsForSun =	( 1 + 1 + 1                 // PRIM then RGBA and XYZ for sun centre.
									+ 1 + uNumSunSlices + 1);	// RGBA and XYZs for sun edge vertices, including first one being repeated.
		const u_int uQWsForHorizGlare = 7;						// PRIM, then 2 RGBA values and 4 XYZs for sun quad
        const u_int uNLOOPForPrePartTwo = (bDoSun ? uQWsForSun : 0) 
										+ (bDoHorizonGlare ? uQWsForHorizGlare : 0) 
                                        + 1;                    // FRAME register setup for context 1.

        const u_int uNumQWs = 1                                 // CALL DMA tag for pre part 1 stuff and then its chain.
                            + 1                                 // GIF tag.
                            + uNLOOPForPrePartOne
                            + 1                                 // CALL DMA tag for pre part 2 stuff and then its chain.
                            + 1                                 // GIF tag.
                            + uNLOOPForPrePartTwo;

        Asura_PS2_QW* const pxOutStart = Asura_PS2_DeferredRendering::GetVifBuffer( uNumQWs );
        Asura_PS2_QW* pxOut = pxOutStart;

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 1 + uNLOOPForPrePartOne,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_CALL,
                                                             false,
                                                             reinterpret_cast<u_int>(g_pvSunChainPartOne),
                                                             false,
                                                             SCE_VIF1_SET_NOP(false),
                                                             SCE_VIF1_SET_DIRECT(1 + uNLOOPForPrePartOne, false) );

        (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNLOOPForPrePartOne,
                                                             true,
                                                             false,
                                                             0,
                                                             SCE_GIF_PACKED,
                                                             1,
                                                             SCE_GIF_PACKED_AD );

        pxOut->u64s[0] = SCE_GS_SET_FRAME_1( uTempBufferFBP, uScreenWidth >> 6, SCE_GS_PSMCT32, 0x00FFFFFF ); // Mask RGB.
        pxOut->u64s[1] = SCE_GS_FRAME_1;
        ++pxOut;

        // XYOFFSET and SCISSOR for context 1.
        pxOut = Asura_PS2_GS::GetViewportSetupGifCmds( iCam, Asura_PS2_GS::CONTEXT_1, pxOut );

        // TEX0 for context 2.
        {
            Asura_Assert( 0.0f == pxCamViewingArea->m_fXOffset || 0.5f == pxCamViewingArea->m_fXOffset, "" );
            Asura_Assert( 0.0f == pxCamViewingArea->m_fYOffset || 0.5f == pxCamViewingArea->m_fYOffset, "" );
            Asura_Assert( 0.5f == pxCamViewingArea->m_fHeight  || 1.0f == pxCamViewingArea->m_fHeight,  "" );

            u_int uTBP = uTempBufferFBP << 5;
            u_int uTBW = uScreenWidth >> 6;
            u_int uTW  = uTempBufferTW;
            u_int uTH  = uTempBufferTH;

            // Assuming 32-bit colour here. In 2 player split screen, we just
            // offset by half the screen buffer to get to the bottom half. In
            // 4 player split screen, it's a little more complicated to get to
            // the righthand quadrants. Due to the fiendish GS memory layout,
            // the additional offset is basically 4 (ie. half width / 64)
            // texture pages, ie. 4 * 32 texels.
            if( pxCamViewingArea->m_fXOffset ) { uTBP += ((uScreenWidth >> 1) << 5) >> 6; }
            if( pxCamViewingArea->m_fYOffset ) { uTBP += (uScreenWidth * (uScreenHeight >> 1)) >> 6; }

            pxOut->u64s[0] = SCE_GS_SET_TEX0_2( uTBP,                // TBP0
                                                uScreenWidth >> 6,   // TBW
                                                SCE_GS_PSMCT32,      // PSM
                                                uTW,                 // TW
                                                uTH,                 // TH
                                                1,                   // TCC
                                                SCE_GS_DECAL,        // TFX
                                                0,                   // CBP
                                                0,                   // CPSM
                                                0,                   // CSM
                                                0,                   // CSA
                                                0 );                 // CLD
            pxOut->u64s[1] = SCE_GS_TEX0_2;
            ++pxOut;
        }

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 1 + uNLOOPForPrePartTwo,
                                                             Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                             Asura_PS2_DmaTag::ID_CALL,
                                                             false,
                                                             reinterpret_cast<u_int>(g_pvSunChainPartTwo),
                                                             false,
                                                             SCE_VIF1_SET_NOP(false),
                                                             SCE_VIF1_SET_DIRECT(1 + uNLOOPForPrePartTwo, false) );

        (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( uNLOOPForPrePartTwo,
                                                             true,
                                                             false,
                                                             0,
                                                             SCE_GIF_PACKED,
                                                             1,
                                                             SCE_GIF_PACKED_AD );

		// Horizon glare RGBAs and XYZs
		if (bDoHorizonGlare)
		{
			// Draw sun in temp buffer's alpha channel.
			pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_TRISTRIP, // PRIM
											  1,					// IIP
											  0,					// TME
											  0,					// FGE
											  0,					// ABE
											  0,					// AA1
											  0,					// FST
											  0,					// CTXT
											  0 );					// FIX
			pxOut->u64s[1] = SCE_GS_PRIM;
			++pxOut;


            // horizon point directly infront of the camera at same Y
			Asura_Vector_3 xHorizonPoint = *Asura_Camera::GetPosition() + Asura_Vector_3(xCameraVec.x * dbg_fHorizonDist, 0.0f, xCameraVec.z * dbg_fHorizonDist);

			Asura_Camera::ProjectWorldCoordsToScreenSpace( xHorizonPoint );

			// getting the horizon glow area rendered is trivial if camera eulerZ is 0.
			// If the camera is rotated things get a bit hairy. To solve this problem I decided
			// to find a quad which would cover the whole screen area affected by horizon glow
			// rather than try to subdivide the area for an exact fit. The closeness of the fit
			// can be adjusted by dbg_fHorizonIncrement.
			// The algorithm is as follows:
			// - get the point directly above the horizon point using the artist-set height.
			// - keep moving right in increments of dbg_fHorizonIncrement (normalised screen 
			// coords) along the rotated horizon line until the line from the current point 
			// directly down to bottom of the glow area is completely offscreen. This gives 
			// the top-right and bottom-right corners. 
			// - repeat the process moving left from the middle to get the top-left and bottom-
			// left corners.
			// Note that since we are working in screen coordinates 0-1 in both directions 
			// one of the directions must be scaled by the aspect ratio (4:3) or (16:9) to 
			// maintain the correct dimensions. Since y is specified by the artists I change
			// x by 1/GetPhysicalScreenAspectRatio() rather than rescale y.
			
			const float fAngle = Asura_Camera::GetOriEuler()->z;
			float fSinA;
			float fCosA;
			Asura_Maths::SinCos( fAngle, fSinA, fCosA );
			
			const float fHeightUp = Asura_Weather_System::GetHorizonGlareSizeUp();
			const float fTotalHeight = fHeightUp + Asura_Weather_System::GetHorizonGlareSizeDown();

			const float fXScaleForScreenAspect = 1.0f / Asura_Camera::GetPhysicalScreenAspectRatio();
			const Asura_Vector_2 xTopMiddlePoint( xHorizonPoint.x - (fXScaleForScreenAspect * fHeightUp * fSinA), xHorizonPoint.y - (fHeightUp * fCosA));
			const Asura_Vector_2 xTopToBottom( fXScaleForScreenAspect * fTotalHeight * fSinA, fTotalHeight * fCosA );
			const Asura_Vector_2 xLeftToRightVector( fXScaleForScreenAspect * dbg_fHorizonIncrement * fCosA, -dbg_fHorizonIncrement * fSinA );

			Asura_Vector_2 xTopRight;
			Asura_Vector_2 xBottomRight;

			for ( xTopRight = xTopMiddlePoint; ; xTopRight += xLeftToRightVector )
			{
				xBottomRight = xTopRight + xTopToBottom;
				if (LineIsOutsideScreen( xTopRight, xBottomRight ) )
				{
					break;
				}
			}

			Asura_Vector_2 xTopLeft;
			Asura_Vector_2 xBottomLeft;

			for ( xTopLeft = xTopMiddlePoint; ; xTopLeft -= xLeftToRightVector )
			{
				xBottomLeft = xTopLeft + xTopToBottom;
				if (LineIsOutsideScreen( xTopLeft, xBottomLeft ) )
				{
					break;
				}
			}

			const float fVAPrimOriginX = static_cast<float>(static_cast<int>(uVAPrimOriginX));
			const float fVAPrimOriginY = static_cast<float>(static_cast<int>(uVAPrimOriginY));
			u_int aauCorners[4][2];
			aauCorners[0][0] = Asura_Maths::FloatToInt( (xTopLeft.x) * fVAWidth + fVAPrimOriginX );
			aauCorners[0][1] = Asura_Maths::FloatToInt( (xTopLeft.y) * fVAHeight + fVAPrimOriginY );
			aauCorners[1][0] = Asura_Maths::FloatToInt( (xTopRight.x) * fVAWidth + fVAPrimOriginX );
			aauCorners[1][1] = Asura_Maths::FloatToInt( (xTopRight.y) * fVAHeight + fVAPrimOriginY );
			aauCorners[2][0] = Asura_Maths::FloatToInt( (xBottomLeft.x) * fVAWidth + fVAPrimOriginX );
			aauCorners[2][1] = Asura_Maths::FloatToInt( (xBottomLeft.y) * fVAHeight + fVAPrimOriginY );
			aauCorners[3][0] = Asura_Maths::FloatToInt( (xBottomRight.x) * fVAWidth + fVAPrimOriginX );
			aauCorners[3][1] = Asura_Maths::FloatToInt( (xBottomRight.y) * fVAHeight + fVAPrimOriginY );

			// finally, some paranoia clipping - better to keep coords in valid range and have 
			// the glare shape a bit skewed than wrap around GS range.
			for (u_int u = 0; u < 4; ++u)
			{
				aauCorners[u][0] = Asura_Maths::Max(0u, aauCorners[u][0]);
				aauCorners[u][1] = Asura_Maths::Max(0u, aauCorners[u][1]);
				aauCorners[u][0] = Asura_Maths::Min(static_cast<u_int>(ASURA_PS2_FRAMEBFR_WIDTH - 1), aauCorners[u][0]);
				aauCorners[u][1] = Asura_Maths::Min(static_cast<u_int>(ASURA_PS2_FRAMEBFR_HEIGHT - 1), aauCorners[u][1]);
			}
			
			// top-left, top-right: edge alpha
            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 0, 0, 0, dbg_uSunEdgeAlpha, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;
			
            // TL
			pxOut->u64s[0] = SCE_GS_SET_XYZ( aauCorners[0][0] << 4,
                                             aauCorners[0][1] << 4,
                                             dbg_uSunZ );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;

			// TR
			pxOut->u64s[0] = SCE_GS_SET_XYZ( aauCorners[1][0] << 4,
                                             aauCorners[1][1] << 4,
                                             dbg_uSunZ );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;

			// bottom-left, bottom-right: centrealpha
            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 0, 0, 0, dbg_uSunCentreAlpha, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;

            // BL
			pxOut->u64s[0] = SCE_GS_SET_XYZ( aauCorners[2][0] << 4,
                                             aauCorners[2][1] << 4,
                                             dbg_uSunZ );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;

			// BR
			pxOut->u64s[0] = SCE_GS_SET_XYZ( aauCorners[3][0] << 4,
                                             aauCorners[3][1] << 4,
                                             dbg_uSunZ );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;
		}

        // Sun RGBAs and XYZs.
        if (bDoSun)
		{
			// Draw sun in temp buffer's alpha channel.
			pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_TRIFAN, // PRIM
											  1,                  // IIP
											  0,                  // TME
											  0,                  // FGE
											  0,                  // ABE
											  0,                  // AA1
											  0,                  // FST
											  0,                  // CTXT
											  0 );                // FIX
			pxOut->u64s[1] = SCE_GS_PRIM;
			++pxOut;

            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 0, 0, 0, dbg_uSunCentreAlpha, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;

            Asura_Vector_3 xSunCentre = Asura_Weather_System::GetSunPos();
            Asura_Camera::ProjectWorldCoordsToScreenSpace( xSunCentre, false );

            xSunCentre.x *= fScreenWidth  * pxCamViewingArea->m_fWidth;
            xSunCentre.y *= fScreenHeight * pxCamViewingArea->m_fHeight;

			u_int uXCoord = (uVAPrimOriginX << 4) + Asura_Maths::FloatToInt(xSunCentre.x * 16.0f);
			u_int uYCoord = (uVAPrimOriginY << 4) + Asura_Maths::FloatToInt(xSunCentre.y * 16.0f);
			Asura_Assert(uXCoord < (ASURA_PS2_FRAMEBFR_WIDTH << 4) && uYCoord < (ASURA_PS2_FRAMEBFR_HEIGHT << 4), "invalid sun coordinate");
            pxOut->u64s[0] = SCE_GS_SET_XYZ( uXCoord,
                                             uYCoord,
                                             dbg_uSunZ );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;

            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 0, 0, 0, dbg_uSunEdgeAlpha, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;

            const float fSlice = (2.0f * Asura_Maths::Pi) / uNumSunSlices;

            float fTheta = 0.0f;
            Asura_Vector_3 xVtxTmp;
            
            for( u_int u = 0; u < uNumSunSlices; fTheta += fSlice, ++u )
            {
                xVtxTmp = xSunCentre + Asura_Vector_3(fSunRadius * Asura_Maths::Sine(fTheta), fSunRadius * Asura_Maths::Cosine(fTheta), 0.0f);

				uXCoord = (uVAPrimOriginX << 4) + Asura_Maths::FloatToInt(xVtxTmp.x * 16.0f);
				uYCoord = (uVAPrimOriginY << 4) + Asura_Maths::FloatToInt(xVtxTmp.y * 16.0f);
				Asura_Assert(uXCoord < (ASURA_PS2_FRAMEBFR_WIDTH << 4) && uYCoord < (ASURA_PS2_FRAMEBFR_HEIGHT << 4), "invalid sun coordinate");
                pxOut->u64s[0] = SCE_GS_SET_XYZ( uXCoord,
                                                 uYCoord,
                                                 dbg_uSunZ );
                pxOut->u64s[1] = SCE_GS_XYZ2;

                if( !u ) { (pxOut + uNumSunSlices)->u128 = pxOut->u128; }

                ++pxOut;
            }
            ++pxOut;
        }
		
        // FRAME 1 for blends back to draw buffer.
        pxOut->u64s[0] = SCE_GS_SET_FRAME_1( uDrawBufferFBP, uScreenWidth >> 6, uDrawBfrPSM, (bModifyAlpha ? 0 : 0xFF000000) );
        pxOut->u64s[1] = SCE_GS_FRAME_1;
        ++pxOut;

        Asura_Assert( static_cast<u_int>(pxOut - pxOutStart) == uNumQWs, "" );

        ++uNumCamsRendered;

		// We've set up the alpha value now; it's safe to render lens flare.
		g_bAlphaSetForLensFlare = true;
    }

    if( uNumCamsRendered )
    {
        // Reset FRAME register for context 1. Note that other registers will
        // be taken care of by setting default renderstate (ZBUF, ALPHA, TEST,
        // CLAMP) or by setting stuff up for the next camera (XYOFFSET,
        // SCISSOR.)
        {
            Asura_PS2_QW* pxOut = Asura_PS2_DeferredRendering::GetVifBuffer( 3 );

            (reinterpret_cast<Asura_PS2_DmaTag*>(pxOut++))->Set( 2,
                                                                 Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                 Asura_PS2_DmaTag::ID_CNT,
                                                                 false,
                                                                 0,
                                                                 false,
                                                                 SCE_VIF1_SET_NOP(false),
                                                                 SCE_VIF1_SET_DIRECT(2, false) );

            (reinterpret_cast<Asura_PS2_GifTag*>(pxOut++))->Set( 1,
                                                                 true,
                                                                 false,
                                                                 0,
                                                                 SCE_GIF_PACKED,
                                                                 1,
                                                                 SCE_GIF_PACKED_AD );

            pxOut->u64s[0] = SCE_GS_SET_FRAME_1( uDrawBufferFBP, uScreenWidth >> 6, uDrawBfrPSM, 0 );
            pxOut->u64s[1] = SCE_GS_FRAME_1;
            ++pxOut;
        }

        Asura_PS2_DeferredRendering::AddPath3Gate();
    }
}// end RenderSunGlare()


// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////


void Asura_Weather_System::Platform_RenderSunGlare( void )
{
	RenderSunGlare(false);
}

void Asura_Weather_System::Platform_RenderHorizonGlare( void )
{
	RenderSunGlare(true);
}

static int g_iLensflareTextureId = ASURA_TEXTURE_NONE;
static int g_iUVBitShift = 0;
void Asura_Weather_System::Platform_Reset( void )
{
	Platform_Shutdown();

	// (AW - 20.10.2004) - Don't assume lensflare is in the gfx.asr anymore.
	g_iLensflareTextureId = ASURA_TEXTURE_NONE;
}

void Asura_Weather_System::Platform_Shutdown( void )
{
    Asura_DeallocateAlign( g_pvSunChainPartTwo );
    g_pvSunChainPartTwo = NULL;

    Asura_DeallocateAlign( g_pvSunChainPartOne );
    g_pvSunChainPartOne = NULL;
}


static u_int s_auFlareUStart[4] = {    8, 4104,    8, 4104 };
static u_int s_auFlareUEnd[4]   = { 4104, 8200, 4104, 8200 };
static u_int s_auFlareVStart[4] = {    8,    8, 4104, 4104 };
static u_int s_auFlareVEnd[4]   = { 4104, 4104, 8200, 8200 };

void Asura_Weather_System::Platform_RenderLensFlare( void )
{
	Asura_Assert(Asura_PS2_RenderingOptions::GetOption( Asura_PS2_RenderingOptions::OPTION_SUNGLARE_LEAVE_BRIGHTNESS_VALUE_IN_ALPHA ), "This rendering option should be enabled for lens flare to work properly");
	Asura_Assert(IsSunGlareEnabled(), "Sun glare should be switched on for lens flare to work properly");

	// We need to have rendered sun glare this frame for the display buffer to be set up with the correct alpha settings for this.
	if(!g_bAlphaSetForLensFlare)
	{
		return;
	}
	g_bAlphaSetForLensFlare = false;

	// Calculate view to sun angle
	Asura_Vector_3 xCameraVec;
	Asura_Vector_3 xSunVector = s_xSunPos - *Asura_Camera::GetPosition();
	xSunVector.Normalise();
	Asura_Camera::GetForwardVector( xCameraVec );

	// Early reject for offscreen sun
	const float fSunAngle = (xSunVector * xCameraVec);
	const float fSunViewLimit = Asura_Camera::GetHorizontalFieldOfView() * dbg_fSunViewLimitScale;
	if (fSunAngle < fSunViewLimit)	return;

	// Project the suns position to the screen
	Asura_Vector_3 xProjectedSunPosition = s_xSunPos;
	Asura_Camera::ProjectWorldCoordsToScreenSpace( xProjectedSunPosition, true );


	// Calculate brightness based on view/sun angle
	const float fSunViewScale = 1.0f / (1.0f - fSunViewLimit);
	const float fBrightness = ( fSunAngle - fSunViewLimit ) * fSunViewScale;

	// Load the textures if not already loaded
	if ( g_iLensflareTextureId == ASURA_TEXTURE_NONE )
	{
#ifdef ASURA_ASSERT
		bool bWasNew;
		g_iLensflareTextureId = Asura_Texture_Management::AddTexture("\\specialfx\\lensflr.tm2", &bWasNew );

		if ( bWasNew )
		{
			Asura_Texture_Management::LoadTextures();
			Asura_Assert( 0, "Lensflare texture should already be loaded.  This will fail on a cd" );
		}
#else
		g_iLensflareTextureId = Asura_Texture_Management::AddTexture( "\\specialfx\\lensflr.tm2" );
#endif
		g_iUVBitShift = (512 / Asura_PS2_TextureManagement::GetTextureWidth(g_iLensflareTextureId)) >> 1;
	}

    Asura_PS2_TextureManagement::SetBufferLayout( ASURA_PS2_TEXBFRLAYOUT_2_LARGE );
    
	Asura_PS2_TextureManagement::SetupTextureEnvironment( Asura_PS2_TextureManagement::UploadTexture(g_iLensflareTextureId),
                                                          ASURA_PS2_TEXREGFLAG_ALL,
                                                          g_iLensflareTextureId,
                                                          Asura_PS2_TextureManagement::BFR_AUTO,
                                                          SCE_GS_LINEAR,
                                                          SCE_GS_LINEAR_MIPMAP_NEAREST );

	// Calculate dma size
	static u_int uFlareQty = 4;
	const u_int uDMAQty = 4 + (uFlareQty*4);
	// Add a dmatag
	Asura_PS2_QW* pxOut = Asura_PS2_DeferredRendering::GetVifBuffer(uDMAQty+2);
	pxOut->u32s[0] = PY_DMA_CNT | (uDMAQty+1);
	pxOut->u32s[1] = 0;
	pxOut->u32s[2] = SCE_VIF1_SET_NOP(false);
	pxOut->u32s[3] = SCE_VIF1_SET_DIRECT( uDMAQty+1, false );
	pxOut++;
	// Add a gif tag
	pxOut->u64s[0] = SCE_GIF_SET_TAG( uDMAQty,			// NLOOP
									  1,				// EOP
									  1,				// Prim Enable : no
									  SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE,		// PRIM Type
													   0,						// IIP (Flat/Gouraud Shading)
													   1,						// TME (Texture Mapping Enabled)
													   0,						// FGE (Fogging Enabled)
													   1,						// ABE (Alpha Blending Enabled)
													   0,						// AA1 (1 Pass Anti-Aliasing Enabled)
													   1,						// FST (Method of specifying Texture Coords)
													   0,						// CTXT (Context)
													   0 ),						// Register quantity
									  SCE_GIF_PACKED,	// PACKET type
									  1 );				// Register quantity
    pxOut->u64s[1] = SCE_GIF_PACKED_AD;
	pxOut++;
	// Set blend mode additive
	pxOut->u64s[0] = SCE_GS_SET_ALPHA_1( 0, 2, 1, 1, 0 );
	pxOut->u64s[1] = SCE_GS_ALPHA_1;
	pxOut++;
	// Clear clamping
	pxOut->u64s[0] = SCE_GS_SET_CLAMP_1( SCE_GS_CLAMP, SCE_GS_CLAMP, 0, 512, 0, 512 );
	pxOut->u64s[1] = SCE_GS_CLAMP_1;
	pxOut++;
	// Set colour to white (don't care about alpha or Q )
	static int uFlareColour = 0xff;
	pxOut->u64s[0] = SCE_GS_SET_RGBAQ( uFlareColour, uFlareColour, uFlareColour, 0, 0 );
	pxOut->u64s[1] = SCE_GS_RGBAQ;
	pxOut++;
	// Set test to only alow RGB writes
	pxOut->u64s[0] = SCE_GS_SET_TEST_1( 1, SCE_GS_ALPHA_NEVER, 0, SCE_GS_AFAIL_RGB_ONLY, 0, 0, 1, SCE_GS_ZALWAYS );
	pxOut->u64s[1] = SCE_GS_TEST_1;
	pxOut++;
	// Calculate viewport data
	Asura_Assert(Asura_Camera::GetNumberOfCameras() == 1, "Lens flare code currently only copes with one camera");
	const Asura_ViewingArea *pxViewArea =Asura_Camera::GetViewingArea();
	const Asura_Vector_2 xScreenCenter( pxViewArea->m_fWidth * (Asura_PS2_GS::GetScreenWidth()>>1), pxViewArea->m_fHeight * (Asura_PS2_GS::GetScreenHeight()>>1) );
	const u_int uXOffset = (ASURA_PS2_SCREEN_CENTRE_X<<4) - (Asura_PS2_GS::GetScreenWidth()<<3) + (u_int)(int)(pxViewArea->m_fXOffset*(float)Asura_PS2_GS::GetScreenWidth());
	const u_int uYOffset = (ASURA_PS2_SCREEN_CENTRE_Y<<4) - (Asura_PS2_GS::GetScreenHeight()<<3) + (u_int)(int)(pxViewArea->m_fYOffset*(float)Asura_PS2_GS::GetScreenHeight());

	// Calculate the flare data
	const float fFlareQty = (float)uFlareQty;
	const Asura_Vector_2 xFlareStart( xProjectedSunPosition.x, xProjectedSunPosition.y );
		  Asura_Vector_2 xFlareVector = (xScreenCenter - xFlareStart) * 2.0f;
	const float fFlareLength = xFlareVector.Normalise();

	static float fFlareSizes[6] = { 200.0f, 300.0f, 50.0f, 25.0f, 90.0f, 100.0f };
	static float fFlareOffset[6] = { 1.0f, 0.75f, 0.5f, 0.25f, 0.0f, 0.0f };
	
	// Render the flares
	for ( u_int u = 0; u<uFlareQty; u++ )
	{
		// Calculate position and size of this flare
		const Asura_Vector_2 xFlareCenter = xFlareStart + (xFlareVector * (fFlareLength*fFlareOffset[u]));
		const float fFlareSize = fFlareSizes[u] * fBrightness;
		
		// Calculate coordinate for the flare
		const u_int uXStart = uXOffset + (u_int)(int)((xFlareCenter.x - fFlareSize) * 16.0f);
		const u_int uYStart = uYOffset + (u_int)(int)((xFlareCenter.y - fFlareSize) * 16.0f);
		const u_int uXEnd   = uXOffset + (u_int)(int)((xFlareCenter.x + fFlareSize) * 16.0f);
		const u_int uYEnd   = uYOffset + (u_int)(int)((xFlareCenter.y + fFlareSize) * 16.0f);
		// Draw the flare
		pxOut->u64s[0] = SCE_GS_SET_UV( s_auFlareUStart[ u&3 ] >> g_iUVBitShift, s_auFlareVStart[ u&3] >> g_iUVBitShift);
		pxOut->u64s[1] = SCE_GS_UV;
		pxOut++;
		pxOut->u64s[0] = SCE_GS_SET_XYZ2( uXStart, uYStart, 0 );
		pxOut->u64s[1] = SCE_GS_XYZ2;
		pxOut++;
		pxOut->u64s[0] = SCE_GS_SET_UV( s_auFlareUEnd[ u&3 ] >> g_iUVBitShift, s_auFlareVEnd[ u&3] >> g_iUVBitShift);
		pxOut->u64s[1] = SCE_GS_UV;
		pxOut++;
		pxOut->u64s[0] = SCE_GS_SET_XYZ2( uXEnd, uYEnd, 0 );
		pxOut->u64s[1] = SCE_GS_XYZ2;
		pxOut++;
	}
}


// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void GenerateDmaChains( const Asura_ViewingArea* pxCamViewingArea )
{
    //
    // Algorithm
    //
    //     For each camera...
    //
    //         Camera's viewing area is AX x AY.
    //
    //         Use two offscreen buffers, each AX/2 x AY/2 in size.
    //         To simplify matters, actually use one offscreen buffer that's
    //         AX/2 x AY in size and draw at appropriate Y offsets.
    //
    //         Use the alpha channel of the temp buffer to avoid disturbing
    //         the draw buffer's alpha. Clear this to 255. Draw the sun into
    //         it, testing against the z buffer to get occlusion information.
    //
    //         Downsample this into the alpha channel of offscreen buffer 1.
    //         (Now AX/2 x AY/2 in size.)
    //
    //         Downsample into alternate offscreen buffers (1->2->1...) until
    //         we reach 2 x 2 in size. At each step, blend the result over
    //         the draw buffer to get the effect of the sun bleeding round
    //         the edges of occluders.
    //
    //         Sample all four texels using bilinear filtering. Use the
    //         resultant value to blend a sprite over the whole screen to
    //         simulate sun glare.
    //
    // Notes
    //
    //     We assume all camera viewing areas are the same size and constant
    //     between resets.
    //
    //     We assume the locations of the z buffer and texture buffer 0 are
    //     constant.
    //
    //     We assume the screen width and height are constant.
    //

    const u_int uNumSPRQWsToReserve = 768;
    Asura_PS2_QW* const pxSPRBfr = Asura_PS2_SPR_Alloc( 0, uNumSPRQWsToReserve );

    const u_int uScreenWidth        = Asura_PS2_GS::GetScreenWidth();
    const u_int uScreenHeight       = Asura_PS2_GS::GetScreenHeight();
    const u_int uZBfrPSM            = Asura_PS2_GS::GetZBfrPSM();
    const u_int uOffscreenBufferTBP = ASURA_PS2_TEXBFR_BASE_PTR;

    const u_int uSunGlareRed    = Asura_Maths::FloatToInt( Asura_Maths::Min(255.0f, Asura_Weather_System::GetSunColourR() * 128.0f) );
    const u_int uSunGlareGreen  = Asura_Maths::FloatToInt( Asura_Maths::Min(255.0f, Asura_Weather_System::GetSunColourG() * 128.0f) );
    const u_int uSunGlareBlue   = Asura_Maths::FloatToInt( Asura_Maths::Min(255.0f, Asura_Weather_System::GetSunColourB() * 128.0f) );
    const u_int uSunGlareAlpha  = 128;

    const u_int uVAWidth       = Asura_Maths::FloatToInt( uScreenWidth  * pxCamViewingArea->m_fWidth   );
    const u_int uVAHeight      = Asura_Maths::FloatToInt( uScreenHeight * pxCamViewingArea->m_fHeight  );
    const u_int uVAPrimOriginX = ASURA_PS2_SCREEN_CENTRE_X - (uVAWidth  >> 1);
    const u_int uVAPrimOriginY = ASURA_PS2_SCREEN_CENTRE_Y - (uVAHeight >> 1);
        
    const u_int uOSBWidth       = uVAWidth >> 1;
    const u_int uOSBHeight      = uVAHeight;
    const u_int uOSBPrimOriginX = ASURA_PS2_SCREEN_CENTRE_X - (uOSBWidth  >> 1);
    const u_int uOSBPrimOriginY = ASURA_PS2_SCREEN_CENTRE_Y - (uOSBHeight >> 1);
    const u_int uOSBTW          = Asura_Maths::GetCeilLog2( uOSBWidth  );
    const u_int uOSBTH          = Asura_Maths::GetCeilLog2( uOSBHeight );

    Asura_PS2_QW* pxOut = pxSPRBfr;
    Asura_PS2_QW* const pxPartOneStart = pxOut;

    // Leave gap for DMA tag and GIF tag. We'll fill them in once we know
    // how many QWs are required.
    pxOut += 2;

    // Setup constant render settings.
    // Context 1.
    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_1, pxOut, false, true, false );
    Asura_PlatformRenderStates::SetZBufferWriteEnabled( false );
    Asura_PlatformRenderStates::SetTextureClamp( true );
    Asura_PlatformRenderStates::SetPixelTest( false, false );
    Asura_PlatformRenderStates::ForgetMyBuffer( pxOut );
    pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS_EXCL_GIF_TAG;
    pxOut->u64s[0] = SCE_GS_SET_TEX1_1( 1, 0, dbg_uSunGlareFilter, dbg_uSunGlareFilter, 1, 0, 0); // LOD, MXL, MMAG, MMIN, MTBA, L, K
    pxOut->u64s[1] = SCE_GS_TEX1_1;
    ++pxOut;
    // Context 2 
    Asura_PlatformRenderStates::CreateDefaultRenderStatePacket( Asura_PlatformRenderStates::CONTEXT_2, pxOut, false, true, false );
    Asura_PlatformRenderStates::SetZBufferWriteEnabled( false );
    Asura_PlatformRenderStates::SetTextureClamp( true );
    Asura_PlatformRenderStates::SetPixelTest( false, false );
    Asura_PlatformRenderStates::ForgetMyBuffer( pxOut );
    pxOut += Asura_PlatformRenderStates::NUM_RENDERSTATE_PACKET_QWS_EXCL_GIF_TAG;
    pxOut->u64s[0] = SCE_GS_SET_FRAME_2( uOffscreenBufferTBP >> 5, (uOSBWidth + 63) >> 6, SCE_GS_PSMCT32, 0 );
    pxOut->u64s[1] = SCE_GS_FRAME_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_XYOFFSET_2( uOSBPrimOriginX << 4, uOSBPrimOriginY << 4 );
    pxOut->u64s[1] = SCE_GS_XYOFFSET_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_SCISSOR_2( 0, uOSBWidth - 1, 0, uOSBHeight - 1 );
    pxOut->u64s[1] = SCE_GS_SCISSOR_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_TEX1_2( 1, 0, dbg_uSunDownsampleFilter, dbg_uSunDownsampleFilter, 1, 0, 0 ); // LOD, MXL, MMAG, MMIN, MTBA, L, K
    pxOut->u64s[1] = SCE_GS_TEX1_2;
    ++pxOut;

    // Clear offscreen buffer to white.
    pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE,   // PRIM Type
                                      0,                    // IIP (Flat/Gouraud Shading)
                                      0,                    // TME (Texture Mapping Enabled)
                                      0,                    // FGE (Fogging Enabled)
                                      0,                    // ABE (Alpha Blending Enabled)
                                      0,                    // AA1 (1 Pass Anti-Aliasing Enabled)
                                      0,                    // FST (Method of specifying Texture Coords)
                                      1,                    // CTXT (Context)
                                      0 );                  // FIX (Fragment Value Control)
    pxOut->u64s[1] = SCE_GS_PRIM;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 128, 128, 128, 128, 0 );
    pxOut->u64s[1] = SCE_GS_RGBAQ;
    ++pxOut;
    pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut, uOSBPrimOriginX, uOSBPrimOriginY, uOSBWidth, uOSBHeight, false );
        
    // Clear alpha channel of temp buffer.
    pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                      0,                  // IIP
                                      0,                  // TME
                                      0,                  // FGE
                                      0,                  // ABE
                                      0,                  // AA1
                                      0,                  // FST
                                      0,                  // CTXT
                                      0 );                // FIX
    pxOut->u64s[1] = SCE_GS_PRIM;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 0, 0, 0, 0, 0 );
    pxOut->u64s[1] = SCE_GS_RGBAQ;
    ++pxOut;
    pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut, uVAPrimOriginX, uVAPrimOriginY, uVAWidth, uVAHeight, false );
        
// (MH - 22.10.2004) - moved PRIM setting to just before setting XYZ and RGB so I can add horizon glare.
//  // Draw sun in temp buffer's alpha channel.
//  pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_TRIFAN, // PRIM
//                                    1,                  // IIP
//                                    0,                  // TME
//                                    0,                  // FGE
//                                    0,                  // ABE
//                                    0,                  // AA1
//                                    0,                  // FST
//                                    0,                  // CTXT
//                                    0 );                // FIX
//  pxOut->u64s[1] = SCE_GS_PRIM;
//  ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_TEST_1( 0, 0, 0, 0, 0, 0, 1, SCE_GS_ZGEQUAL ); // ATE, ATST, AREF, AFAIL, DATE, DATM, ZTE, ZTST
    pxOut->u64s[1] = SCE_GS_TEST_1;
    ++pxOut;

    const u_int uNumQWsInPartOne = pxOut - pxPartOneStart;

    // Fill in part one's DMA and GIF tags now.
    {
        const u_int uQWC   = uNumQWsInPartOne - 1;
        const u_int uNLOOP = uQWC - 1;

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxPartOneStart))->Set( uQWC,
                                                                    Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                    Asura_PS2_DmaTag::ID_RET,
                                                                    false,
                                                                    0,
                                                                    false,
                                                                    SCE_VIF1_SET_NOP(false),
                                                                    SCE_VIF1_SET_DIRECT(uQWC, false) );

        // And the GIF tag.
        (reinterpret_cast<Asura_PS2_GifTag*>(pxPartOneStart + 1))->Set( uNLOOP,
                                                                        true,
                                                                        false,
                                                                        0,
                                                                        SCE_GIF_PACKED,
                                                                        1,
                                                                        SCE_GIF_PACKED_AD );
    }

    // Allocate buffer.
    {
        u_int uQWsToAlloc = ((uNumQWsInPartOne + 3) >> 2) << 2;

        if( g_pvSunChainPartOne )
        {
            Asura_Assert( uQWsToAlloc <= g_uSizeOfSunChainPartOne, "" );
        }
        else
        {
            g_pvSunChainPartOne = Asura_AllocateAlign( uQWsToAlloc << 4, ASURA_ALIGN_64_BYTE );
            g_uSizeOfSunChainPartOne = uQWsToAlloc;
        }
    }
    
    // Start part one DMAing into that buffer, while we prepare part two.
    Asura_PS2_DMA::TransferFromSPR( pxPartOneStart, Asura_PS2_DMAMem(g_pvSunChainPartOne), uNumQWsInPartOne, true, false );

    Asura_PS2_QW* const pxPartTwoStart = pxOut;

    // Leave gap for DMA tag and GIF tag. We'll fill them in once we know
    // how many QWs are required.
    pxOut += 2;

    // Setup context 2 ready for the downsampling process.
    pxOut->u64s[0] = SCE_GS_SET_ALPHA_2( SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 ); // Leave colour; update alpha.
    pxOut->u64s[1] = SCE_GS_ALPHA_2;
    ++pxOut;

    // Downsample from the temp buffer's alpha to first offscreen buffer's
    // alpha.
    pxOut->u64s[1] = SCE_GS_TEXFLUSH;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_CLAMP_2( SCE_GS_REGION_CLAMP, SCE_GS_REGION_CLAMP, 0, uVAWidth - 1, 0, uVAHeight - 1 );
    pxOut->u64s[1] = SCE_GS_CLAMP_2;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                      0,                  // IIP
                                      1,                  // TME
                                      0,                  // FGE
                                      1,                  // ABE
                                      0,                  // AA1
                                      1,                  // FST
                                      1,                  // CTXT
                                      0 );                // FIX
    pxOut->u64s[1] = SCE_GS_PRIM;
    ++pxOut;
    pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut,
                                                      uOSBPrimOriginX, uOSBPrimOriginY, (uVAWidth >> 1), (uVAHeight >> 1),
                                                      true,
                                                      1, 1, uVAWidth, uVAHeight );

    // Now setup context 1 for our blends back to the draw buffer.
    pxOut->u64s[0] = SCE_GS_SET_TEST_1( 0, 0, 0, 0, 0, 0, 1, SCE_GS_ZALWAYS ); // ATE, ATST, AREF, AFAIL, DATE, DATM, ZTE, ZTST
    pxOut->u64s[1] = SCE_GS_TEST_1;
    ++pxOut;
    // NB: This RGBA is only used for final glare sprite pass.
    pxOut->u64s[0] = SCE_GS_SET_RGBAQ( uSunGlareRed, uSunGlareGreen, uSunGlareBlue, uSunGlareAlpha, 0 );
    pxOut->u64s[1] = SCE_GS_RGBAQ;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_ALPHA_1( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 ); // Cd += Cs * As
    pxOut->u64s[1] = SCE_GS_ALPHA_1;
    ++pxOut;

    // Downsample repeatedly to reach 2x2.
    u_int uSrcBfr    = 0;
    u_int uDestBfr   = 1;
    u_int uSrcWidth  = uVAWidth  >> 1;
    u_int uSrcHeight = uVAHeight >> 1;
    // Throughout downsampling, source texture buffer is just the entire
    // OSB and we use offsets to get the correct part.
    pxOut->u64s[0] = SCE_GS_SET_TEX0_1( uOffscreenBufferTBP,
                                        (uSrcWidth + 63) >> 6,
                                        SCE_GS_PSMCT32,
                                        uOSBTW,
                                        uOSBTH,
                                        1,
                                        SCE_GS_MODULATE,
                                        0,
                                        0,
                                        0,
                                        0,
                                        0 );
    pxOut->u64s[1] = SCE_GS_TEX0_1;
    ++pxOut;
    pxOut->u64s[0] = SCE_GS_SET_TEX0_2( uOffscreenBufferTBP,
                                        (uSrcWidth + 63) >> 6,
                                        SCE_GS_PSMCT32,
                                        uOSBTW,
                                        uOSBTH,
                                        1,
                                        SCE_GS_DECAL,
                                        0,
                                        0,
                                        0,
                                        0,
                                        0 );
    pxOut->u64s[1] = SCE_GS_TEX0_2;
    ++pxOut;
    //          PAL                        NTSC
    // 0 256x256 -> 1 128x128     0 256x224 -> 1 128x112
    // 1 128x128 -> 0 64x64       1 128x112 -> 0 64x56
    // 0 64x64   -> 1 32x32       0 64x56   -> 1 32x28
    // 1 32x32   -> 0 16x16       1 32x28   -> 0 16x14
    // 0 16x16   -> 1 8x8         0 16x14   -> 1 8x7
    // 1 8x8     -> 0 4x4         1 8x7     -> 0 4x3
    // 0 4x4     -> 1 2x2         0 4x3     -> 1 2x2
    while( true )
    {
        const u_int uSrcV  = (uSrcBfr ? (uVAHeight >> 1) : 0);
        const u_int uDestY = uOSBPrimOriginY + (uDestBfr ? (uVAHeight >> 1) : 0);

        pxOut->u64s[1] = SCE_GS_TEXFLUSH;
        ++pxOut;

        // Maybe display on screen, for debugging purposes. Note that we
        // deal with displaying the final 1x1 sample below.
        if( uSrcWidth == dbg_uSunOSBWidthToDisplay )
        {
            pxOut->u64s[0] = SCE_GS_SET_CLAMP_1( SCE_GS_REGION_CLAMP, SCE_GS_REGION_CLAMP, 0, uSrcWidth - 1, uSrcV, uSrcV + uSrcHeight - 1 );
            pxOut->u64s[1] = SCE_GS_CLAMP_1;
            ++pxOut;
            // No filtering; we want to see the texture as-is.
            pxOut->u64s[0] = SCE_GS_SET_TEX1_1( 1, 0, SCE_GS_NEAREST, SCE_GS_NEAREST, 1, 0, 0 ); // LOD, MXL, MMAG, MMIN, MTBA, L, K
            pxOut->u64s[1] = SCE_GS_TEX1_1;
            ++pxOut;
            // Cd = Cs * As = As, since we cleared the OSBs to white
            // earlier. 
            pxOut->u64s[0] = SCE_GS_SET_ALPHA_1( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_ZERO, 0 );
            pxOut->u64s[1] = SCE_GS_ALPHA_1;
            ++pxOut;
            if( !dbg_bSunOSBDisplayShouldBeColoured )
            {
                pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 128, 128, 128, 128, 0 );
                pxOut->u64s[1] = SCE_GS_RGBAQ;
                ++pxOut;
            }
            pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                              0,                  // IIP
                                              1,                  // TME
                                              0,                  // FGE
                                              1,                  // ABE
                                              0,                  // AA1
                                              1,                  // FST
                                              0,                  // CTXT
                                              0 );                // FIX
            pxOut->u64s[1] = SCE_GS_PRIM;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_UV( 0 << 4, uSrcV << 4 );
            pxOut->u64s[1] = SCE_GS_UV;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_XYZ2( (ASURA_PS2_SCREEN_CENTRE_X - 128) << 4, (ASURA_PS2_SCREEN_CENTRE_Y - 128) << 4, 0 );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_UV( uSrcWidth << 4, (uSrcV + uSrcHeight) << 4 );
            pxOut->u64s[1] = SCE_GS_UV;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_XYZ2( (ASURA_PS2_SCREEN_CENTRE_X + 128) << 4, (ASURA_PS2_SCREEN_CENTRE_Y + 128) << 4, 0 );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;
            // Reset stuff...
            pxOut->u64s[0] = SCE_GS_SET_TEX1_1( 1, 0, dbg_uSunGlareFilter, dbg_uSunGlareFilter, 1, 0, 0 ); // LOD, MXL, MMAG, MMIN, MTBA, L, K
            pxOut->u64s[1] = SCE_GS_TEX1_1;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_ALPHA_1( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_CD, 0 ); // Cd += Cs * As
            pxOut->u64s[1] = SCE_GS_ALPHA_1;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( uSunGlareRed, uSunGlareGreen, uSunGlareBlue, uSunGlareAlpha, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;
        }

        // Stop if we've reached 2x2, now that we've dealt with the
        // possibility of having to display it on screen.
        if( 2 == uSrcWidth && 2 == uSrcHeight ) { break; }

        // Update CLAMP registers.
        pxOut->u64s[0] = SCE_GS_SET_CLAMP_1( SCE_GS_REGION_CLAMP, SCE_GS_REGION_CLAMP, 0, uSrcWidth - 1, uSrcV, uSrcV + uSrcHeight - 1 );
        pxOut->u64s[1] = SCE_GS_CLAMP_1;
        ++pxOut;
        pxOut->u64s[0] = SCE_GS_SET_CLAMP_2( SCE_GS_REGION_CLAMP, SCE_GS_REGION_CLAMP, 0, uSrcWidth - 1, uSrcV, uSrcV + uSrcHeight - 1 );
        pxOut->u64s[1] = SCE_GS_CLAMP_2;
        ++pxOut;

        const u_int uDestWidth  = Asura_Maths::Max( 2u, uSrcWidth  >> 1 );
        const u_int uDestHeight = Asura_Maths::Max( 2u, uSrcHeight >> 1 );

        // Downsample to target.
        pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                          0,                  // IIP
                                          1,                  // TME
                                          0,                  // FGE
                                          1,                  // ABE
                                          0,                  // AA1
                                          1,                  // FST
                                          1,                  // CTXT
                                          0 );                // FIX
        pxOut->u64s[1] = SCE_GS_PRIM;
        ++pxOut;
        pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut,
                                                          uOSBPrimOriginX, uDestY, uDestWidth, uDestHeight,
                                                          true,
                                                          1, uSrcV + 1, uSrcWidth, uSrcHeight );

        // Blend into draw buffer. This gives the effect of light bleeding
        // around the edges of occluders. However, don't bother with tiny
        // textures. They waste fillrate - this is a full-screen pass per
        // downsample level, remember - and don't add much, apart from an
        // overall brightening that's achieved via the final glare sprite
        // below anyway.
        if( dbg_bSunBlendDownsampledBuffersIntoDrawBuffer && uSrcWidth >= (uVAWidth >> 6) )
        {
            pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                              0,                  // IIP
                                              1,                  // TME
                                              0,                  // FGE
                                              1,                  // ABE
                                              0,                  // AA1
                                              1,                  // FST
                                              0,                  // CTXT
                                              0 );                // FIX
            pxOut->u64s[1] = SCE_GS_PRIM;
            ++pxOut;

            // Use an inset for larger textures to increase bleeding
            // effect.
            const u_int uInset = ( uSrcWidth >= 64 ? dbg_uBleedBlendUVInset : 0 );

            pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut,
                                                              uVAPrimOriginX, uVAPrimOriginY, uVAWidth, uVAHeight,
                                                              true,
                                                              0 + uInset, uSrcV + uInset, uSrcWidth - (uInset << 1), uSrcHeight - (uInset << 1) );
        }

        uSrcBfr    ^= 1;
        uDestBfr   ^= 1;
        uSrcWidth   = uDestWidth;
        uSrcHeight  = uDestHeight;
    }

    // Final fullscreen (well, full viewing area) glare sprite using the 2x2
    // downsampled texture to give us a brightness value. Note that we'll
    // have done the TEXFLUSH in the last iteration of the loop.
    if( dbg_bSunEnableFinalGlareSprite ) //|| 1 == dbg_uSunOSBWidthToDisplay )
    {
        const u_int uSrcV = (uSrcBfr ? (uVAHeight >> 1) : 0);

        pxOut->u64s[0] = SCE_GS_SET_CLAMP_1( SCE_GS_REGION_CLAMP, SCE_GS_REGION_CLAMP, 0, 0 + 2 - 1, uSrcV, uSrcV + 2 - 1 );
        pxOut->u64s[1] = SCE_GS_CLAMP_1;
        ++pxOut;
        pxOut->u64s[0] = SCE_GS_SET_PRIM( SCE_GS_PRIM_SPRITE, // PRIM
                                          0,                  // IIP
                                          1,                  // TME
                                          0,                  // FGE
                                          1,                  // ABE
                                          0,                  // AA1
                                          1,                  // FST
                                          0,                  // CTXT
                                          0 );                // FIX
        pxOut->u64s[1] = SCE_GS_PRIM;
        ++pxOut;
        pxOut->u64s[0] = SCE_GS_SET_UV( 1 << 4, (uSrcV + 1) << 4 );
        pxOut->u64s[1] = SCE_GS_UV;
        ++pxOut;
        if( 1 == dbg_uSunOSBWidthToDisplay )
        {
            pxOut->u64s[0] = SCE_GS_SET_ALPHA_1( SCE_GS_ALPHA_CS, SCE_GS_ALPHA_ZERO, SCE_GS_ALPHA_AS, SCE_GS_ALPHA_ZERO, 0 );
            pxOut->u64s[1] = SCE_GS_ALPHA_1;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( 128, 128, 128, 128, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_XYZ2( (ASURA_PS2_SCREEN_CENTRE_X - 128) << 4, (ASURA_PS2_SCREEN_CENTRE_Y - 128) << 4, 0 );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;
            pxOut->u64s[0] = SCE_GS_SET_XYZ2( (ASURA_PS2_SCREEN_CENTRE_X + 128) << 4, (ASURA_PS2_SCREEN_CENTRE_Y + 128) << 4, 0 );
            pxOut->u64s[1] = SCE_GS_XYZ2;
            ++pxOut;
        }
        else
        {
            const u_int uFinalGlareSpriteR = Asura_Maths::Min( 255u, uSunGlareRed   * dbg_uSunFinalGlareSpriteRGBMul );
            const u_int uFinalGlareSpriteG = Asura_Maths::Min( 255u, uSunGlareGreen * dbg_uSunFinalGlareSpriteRGBMul );
            const u_int uFinalGlareSpriteB = Asura_Maths::Min( 255u, uSunGlareBlue  * dbg_uSunFinalGlareSpriteRGBMul );

            pxOut->u64s[0] = SCE_GS_SET_RGBAQ( uFinalGlareSpriteR, uFinalGlareSpriteG, uFinalGlareSpriteB, uSunGlareAlpha, 0 );
            pxOut->u64s[1] = SCE_GS_RGBAQ;
            ++pxOut;
            pxOut = Asura_PS2_2D_Graphics::WriteSpriteStrips( pxOut, uVAPrimOriginX, uVAPrimOriginY, uVAWidth, uVAHeight, false );
        }
    }

    const u_int uNumQWsInPartTwo = pxOut - pxPartTwoStart;

    // Fill in part two's DMA and GIF tags now.
    {
        const u_int uQWC   = uNumQWsInPartTwo - 1;
        const u_int uNLOOP = uQWC - 1;

        (reinterpret_cast<Asura_PS2_DmaTag*>(pxPartTwoStart))->Set( uQWC,
                                                                    Asura_PS2_DmaTag::PCE_NOCHANGE,
                                                                    Asura_PS2_DmaTag::ID_RET,
                                                                    false,
                                                                    0,
                                                                    false,
                                                                    SCE_VIF1_SET_NOP(false),
                                                                    SCE_VIF1_SET_DIRECT(uQWC, false) );

        // And the GIF tag.
        (reinterpret_cast<Asura_PS2_GifTag*>(pxPartTwoStart + 1))->Set( uNLOOP,
                                                                        true,
                                                                        false,
                                                                        0,
                                                                        SCE_GIF_PACKED,
                                                                        1,
                                                                        SCE_GIF_PACKED_AD );
    }

    // Allocate buffer.
    {
        u_int uQWsToAlloc = ((uNumQWsInPartTwo + 3) >> 2) << 2;

        if( g_pvSunChainPartTwo )
        {
            Asura_Assert( uQWsToAlloc <= g_uSizeOfSunChainPartTwo, "" );
        }
        else
        {
            g_pvSunChainPartTwo = Asura_AllocateAlign( uQWsToAlloc << 4, ASURA_ALIGN_64_BYTE );
            g_uSizeOfSunChainPartTwo = uQWsToAlloc;
        }
    }

    // Wait for part 1 transfer to complete.
    ASURA_PS2_WAIT_FOR_DMA_COMPLETION( 1 << Asura_PS2_DMA::CHANNEL_FROMSPR, false );

    // And transfer second part (blocking.)
    Asura_PS2_DMA::TransferFromSPR( pxPartTwoStart,
                                    Asura_PS2_DMAMem(g_pvSunChainPartTwo),
                                    uNumQWsInPartTwo,
                                    true,
                                    true );

    Asura_PS2_SPR_Dealloc( pxSPRBfr );
}

static void GenerateDmaChainsIfNecessary( void )
{
#if _DEBUG
    if( g_bSunForceChainReset )
    {
        if( g_pvSunChainPartTwo ) { Asura_DeallocateAlign( g_pvSunChainPartTwo ); g_pvSunChainPartTwo = NULL; }
        if( g_pvSunChainPartOne ) { Asura_DeallocateAlign( g_pvSunChainPartOne ); g_pvSunChainPartOne = NULL; }

        g_bSunForceChainReset = false;
    }
#endif

    if( !g_pvSunChainPartOne || Asura_Camera::GetNumberOfCameras() != g_iSunCachedNumCams )
    {
        // Since we assume all cameras have the same viewing area, just
        // deal with camera 0.
        Asura_Camera::SetCameraIndex( 0 );

        if( !g_pvSunChainPartOne )
        {
            // Post-reset or post-initialisation, first make sure the
            // buffer is large enough for worst case. This is full screen
            // since then the draw area will be largest and possibly we
            // may need to split stuff up to deal with UV limits.
            const Asura_ViewingArea xFullScreenViewingArea( 0.0f, 0.0f, 1.0f, 1.0f );
            GenerateDmaChains( &xFullScreenViewingArea );
        }

        GenerateDmaChains( Asura_Camera::GetViewingArea() );

        g_iSunCachedNumCams = Asura_Camera::GetNumberOfCameras();
    }
}
