//
// Asura_Platform_Animation_RenderCache.cpp
//
// Ben Weston - 13/06/2008
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

//
// Includes
//

#include "Core/Asura.h"

#include "Animation/Asura_Platform_Animation_RenderCache.h"

// Asura
#include "Animation/Asura_Animation.h"
#include "Core/Memory/Asura_Mem.h"
#include "Rendering/Asura_Render.h"
#include "Interface/Asura_CommandConsole.h"
#include "Debugging/Asura_DebugText.h"

// Asura Platform
#include "Debugging/Asura_PS3_DebugInfo.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_GCM_IndexBuffer.h"
#include "Rendering/Asura_PS3_SPU_SkinningTypes.h"
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"


//
// File-scope Variables
//

//Asura_Platform_Animation_RenderCache*		Asura_Platform_Animation_RenderCache::s_pxActive = NULL;

static Asura_PS3_RenderCache_Buffer*		g_apxCleanUp[2] = { NULL, NULL };
static u_int								g_uCleanupBuffer = 0;

const u_int										g_uAgeToDelete = 100; // if it has been this many frames since this cache was last used, delete it.
static Asura_Platform_Animation_RenderCache*	g_pxSortedListStart = NULL;
static Asura_Platform_Animation_RenderCache*	g_pxSortedListEnd = NULL;

DVSVAR bool dvs_bShowRenderCacheStats = false;

//
// Static functions, for globally managing clean-up of render caches
//

Asura_Platform_Animation_RenderCache* Asura_Platform_Animation_RenderCache::Create()
{
	Asura_Platform_Animation_RenderCache* pxNew = new Asura_Platform_Animation_RenderCache;
	
	Asura_Assert(pxNew,"Asura_Platform_Animation_RenderCache: failed to allocate new rendercache object");
	
	return pxNew;
}


void Asura_Platform_Animation_RenderCache::Cleanup()
{
	// aggressively find & destroy buffers that have not been used for a while
	Asura_Platform_Animation_RenderCache* pxCheck = g_pxSortedListStart;
	while ( pxCheck && pxCheck->m_uBoneFrameStamp + g_uAgeToDelete < Asura_Timers::GetFrameNumber() )
	{
		Asura_Platform_Animation_RenderCache* pxNext = pxCheck->m_pxNext;
		if ( pxCheck->m_uBoneFrameStamp != 0 ) // tend to have some of these hanging around in the list
		{
// can't delete, parent still refers, reset to init
// 			pxCheck->Destroy();
 			pxCheck->Reset();

			Asura_OutputToDebugger_Retail(("Deleting cache for unseen skin\n"));
		}
		pxCheck = pxNext;
	}
	
	// clean up all buffers added to the cleanup list 2-frames ago
	g_uCleanupBuffer = 1-g_uCleanupBuffer;
	
	int iDebugCounter = 0;
	while ( g_apxCleanUp[g_uCleanupBuffer] )
	{
		Asura_PS3_RenderCache_Buffer* pxThis = g_apxCleanUp[g_uCleanupBuffer];
		g_apxCleanUp[g_uCleanupBuffer] = pxThis->m_pxNext;

		delete pxThis; //call the destructor & free the memory
		iDebugCounter++;
	}
	
	if ( iDebugCounter > 0 )
	{
		Asura_OutputToDebugger_Retail(( "Cleaned up %d SPU skinning buffers\n", iDebugCounter ));
	}
	

#if defined ASURA_DEBUG_VARIABLE_SYSTEM
	const bool bDVSOpen = Asura_DebugVariableSystem::IsOpen();
	if ( dvs_bShowRenderCacheStats && !bDVSOpen )
	{
		RenderDebugInfo();
	}
#endif

}

void Asura_Platform_Animation_RenderCache::Initialise()
{
	g_apxCleanUp[0] = NULL;
	g_apxCleanUp[1] = NULL;
	g_uCleanupBuffer = 0;
	
#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.RenderCacheStats", &dvs_bShowRenderCacheStats, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
}

void Asura_Platform_Animation_RenderCache::Shutdown()
{
	Cleanup();
	Cleanup();

	g_apxCleanUp[0] = NULL;
	g_apxCleanUp[1] = NULL;
	g_uCleanupBuffer = 0;
}

// output statistics about how many caches exist, how much memory is used, etc.
void Asura_Platform_Animation_RenderCache::RenderDebugInfo()
{
#if defined(ASURA_USE_DEBUG_TEXT)

	int iTotalPos  = 0;
	int iTotalIdx  = 0;
	int iTotalAA   = 0;
	int iTotalBone = 0;
	int iCount     = 0;

	Asura_Platform_Animation_RenderCache*	pxIter = g_pxSortedListStart;
	while ( pxIter )
	{
		for ( int i=0; i < ASURA_ANIMATION_SKIN_COUNT; i++ )
		{
			Asura_Assert( ASURA_PS3_RENDERCACHE_NUM_INDEXBUFFERS == 2, "Asura_Platform_Animation_RenderCache::RenderDebugInfo: update this code" );
			if ( pxIter->m_apxBuffers[i]->m_apxIndexBuffer[0] ) iTotalIdx += pxIter->m_apxBuffers[i]->m_apxIndexBuffer[0]->GetIndexBufferSize();
			if ( pxIter->m_apxBuffers[i]->m_apxIndexBuffer[1] ) iTotalAA  += pxIter->m_apxBuffers[i]->m_apxIndexBuffer[1]->GetIndexBufferSize();
			
			Asura_Assert( ASURA_PS3_RENDERCACHE_NUM_VERTEXBUFFERS == 1, "Asura_Platform_Animation_RenderCache::RenderDebugInfo: update this code" );
			if ( pxIter->m_apxBuffers[i]->m_apxVertexBuffer[0] ) iTotalPos += pxIter->m_apxBuffers[i]->m_apxVertexBuffer[0]->GetVertexBufferSize();
		}
		iTotalBone += pxIter->m_uNumBonesAlloced * sizeof(Asura_AMatrix4x4);
		if ( pxIter->m_uNumBonesAlloced > 0 ) iCount++;
	
		pxIter = pxIter->m_pxNext;
	}

	// peaks
	static int iMaxPos = 0;
	static int iMaxIdx = 0;
	static int iMaxAA = 0;
	static int iMaxBone = 0;
	static int iMaxCount = 0;

	iMaxPos   = Asura_Maths::Max( iTotalPos , iMaxPos   );
	iMaxIdx   =	Asura_Maths::Max( iTotalIdx , iMaxIdx   );
	iMaxAA    = Asura_Maths::Max( iTotalAA  , iMaxAA    );
	iMaxBone  =	Asura_Maths::Max( iTotalBone, iMaxBone  );
	iMaxCount =	Asura_Maths::Max( iCount    , iMaxCount );

	
	Asura_DebugText::AddTextBlock( 15,  8, 0xff44ff88, 0, "SPU Skinning Caches: current       " );
	Asura_DebugText::AddTextBlock( 15,  8, 0xffff0000, 0, "                               max " );
	Asura_DebugText::AddTextBlock( 15,  9, 0xff44ff88, 0, "         num caches:  %4d          ", iCount );
	Asura_DebugText::AddTextBlock( 15,  9, 0xffff0000, 0, "                              %4d  ", iMaxCount );
	Asura_DebugText::AddTextBlock( 15, 10, 0xff44ff88, 0, "    position buffer:%6.2fMB        ", (float)(iTotalPos)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 10, 0xffff0000, 0, "                            %6.2fMB", (float)(iMaxPos)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 12, 0xff44ff88, 0, "     culled indices:%6.2fMB        ", (float)(iTotalIdx)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 12, 0xffff0000, 0, "                            %6.2fMB", (float)(iMaxIdx)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 13, 0xff44ff88, 0, " anti-alias indices:%6.2fMB        ", (float)(iTotalAA)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 13, 0xffff0000, 0, "                            %6.2fMB", (float)(iMaxAA)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 14, 0xff44ff88, 0, "        bone caches: %6.2fk         ", (float)(iTotalBone)/(1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 14, 0xffff0000, 0, "                            %6.2fk ", (float)(iMaxBone)/(1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 15, 0xff44ff88, 0, "              TOTAL:%6.2fMB        ", (float)(iTotalPos+iTotalIdx+iTotalAA+iTotalBone)/(1024.0f*1024.0f) );
	Asura_DebugText::AddTextBlock( 15, 15, 0xffff0000, 0, "                            %6.2fMB", (float)(iMaxPos+iMaxIdx+iMaxAA+iMaxBone)/(1024.0f*1024.0f) );

	Asura_Render::FlushDisplay();
	
#endif	//defined(ASURA_USE_DEBUG_TEXT)
}


//
// Methods
//

Asura_Platform_Animation_RenderCache::Asura_Platform_Animation_RenderCache()
{
	for ( int j=0; j < ASURA_ANIMATION_SKIN_COUNT; j++ )
	{
		m_apxBuffers[j] = new Asura_PS3_RenderCache_Buffer;
	}
	m_pxBoneCache = NULL;
	m_uNumBonesAlloced = 0;
	m_uBoneFrameStamp = 0;
	
	m_pxPrev = g_pxSortedListEnd;
	g_pxSortedListEnd = this;
	if ( m_pxPrev ) m_pxPrev->m_pxNext = this; else g_pxSortedListStart = this;
	m_pxNext = NULL;
}

Asura_Platform_Animation_RenderCache::~Asura_Platform_Animation_RenderCache()
{
	for ( int j=0; j < ASURA_ANIMATION_SKIN_COUNT; j++ )
	{
		if ( m_apxBuffers[j] != NULL )
		{
			delete m_apxBuffers[j];
			m_apxBuffers[j] = NULL;
		}
	}

	if ( m_pxBoneCache != NULL )
	{
		Asura_DeallocateAlign(m_pxBoneCache);
		m_pxBoneCache = NULL;
	}

	m_uNumBonesAlloced = 0;
	m_uBoneFrameStamp = 0;

	if ( m_pxPrev ) m_pxPrev->m_pxNext = m_pxNext; else g_pxSortedListStart = m_pxNext;
	if ( m_pxNext ) m_pxNext->m_pxPrev = m_pxPrev; else g_pxSortedListEnd = m_pxPrev;
}


// Vertex data will be different next time, handle our vertex buffers in a way that won't break the current render but that prepares them to be replaced next frame
void Asura_Platform_Animation_RenderCache::Flush( u_int uSkin )
{
	if ( m_apxBuffers[uSkin] != NULL )
	{
		//add this to the list of ones to destroy in a couple of frames time (so we don't piss off SPUs or RSX by deleting something they're still using)
		if ( m_apxBuffers[uSkin]->Destroy() )
		{
			// it has been destroyed (if not it was never initialised)
			m_apxBuffers[uSkin] = new Asura_PS3_RenderCache_Buffer;
		}
	}
}


// Queue this cache for destruction after the RSX has finished reading from it.
void Asura_Platform_Animation_RenderCache::Destroy()
{
	// add buffers to the cleanup list (so they can still be accessed by RSX/SPU)
	for (u_int uSkin=0; uSkin < ASURA_ANIMATION_SKIN_COUNT; uSkin++)
	{
		//add this to the list of ones to destroy in a couple of frames time (so we don't piss off SPUs or RSX by deleting something they're still using)
		if ( !m_apxBuffers[uSkin]->Destroy() )
		{
			//it doesn't need to be slow-destroyed, so just delete it - BW this is ugly I think I need something like "IsUsed()" and a Destroy that always deletes it.
			delete m_apxBuffers[uSkin];
		}
		m_apxBuffers[uSkin] = NULL;
	}
	
	delete this; //call my destructor
}

void Asura_Platform_Animation_RenderCache::Reset()
{
	// Destroy

	// add buffers to the cleanup list (so they can still be accessed by RSX/SPU)
	for (u_int uSkin=0; uSkin < ASURA_ANIMATION_SKIN_COUNT; uSkin++)
	{
		//add this to the list of ones to destroy in a couple of frames time (so we don't piss off SPUs or RSX by deleting something they're still using)
		if ( !m_apxBuffers[uSkin]->Destroy() )
		{
			//it doesn't need to be slow-destroyed, so just delete it - BW this is ugly I think I need something like "IsUsed()" and a Destroy that always deletes it.
			delete m_apxBuffers[uSkin];
		}
		m_apxBuffers[uSkin] = NULL;
	}

	// destructor
	for ( int j=0; j < ASURA_ANIMATION_SKIN_COUNT; j++ )
	{
		if ( m_apxBuffers[j] != NULL )
		{
			delete m_apxBuffers[j];
			m_apxBuffers[j] = NULL;
		}
	}

	if ( m_pxBoneCache != NULL )
	{
		Asura_DeallocateAlign(m_pxBoneCache);
		m_pxBoneCache = NULL;
	}

	m_uNumBonesAlloced = 0;
	m_uBoneFrameStamp = 0;

	if ( m_pxPrev ) m_pxPrev->m_pxNext = m_pxNext; else g_pxSortedListStart = m_pxNext;
	if ( m_pxNext ) m_pxNext->m_pxPrev = m_pxPrev; else g_pxSortedListEnd = m_pxPrev;

	// constructor
	for ( int j=0; j < ASURA_ANIMATION_SKIN_COUNT; j++ )
	{
		m_apxBuffers[j] = new Asura_PS3_RenderCache_Buffer;
	}
	m_pxBoneCache = NULL;
	m_uNumBonesAlloced = 0;
	m_uBoneFrameStamp = 0;

	m_pxPrev = g_pxSortedListEnd;
	g_pxSortedListEnd = this;
	if ( m_pxPrev ) m_pxPrev->m_pxNext = this; else g_pxSortedListStart = this;
	m_pxNext = NULL;
}

Asura_AMatrix4x4* Asura_Platform_Animation_RenderCache::UpdateCachedBones( const Asura_AMatrix4x4* pxBones, u_int uNumBones )
{
	Asura_Assert( !IsBoneCacheCurrent(), "UpdateCachedBones - updating bone cache twice in the same frame for the same animation, this is wasteful" );

	// first check our cache exists & is big enough
	if ( uNumBones != m_uNumBonesAlloced )
	{
		if ( m_pxBoneCache != NULL )
		{
			Asura_Assert( 0, "Asura_Platform_Animation_RenderCache::UpdateCachedBones - Number of bones in animated skin has changed, this will cause a realloc & is probably a bad thing." );
			
			Asura_DeallocateAlign( m_pxBoneCache );
		}
		
		// alloc the space needed
		m_pxBoneCache = reinterpret_cast<Asura_AMatrix4x4*>(Asura_AllocateAlign( uNumBones * sizeof(Asura_AMatrix4x4), ASURA_ALIGN_16_BYTE ));

		m_uNumBonesAlloced = uNumBones;
	}
	
	// cache the bones
	Asura_Mem::Copy( m_pxBoneCache, pxBones, uNumBones * sizeof(Asura_AMatrix4x4) );

	SetBoneCacheCurrent();

	// bump it to the top of the age-sorted list
	// delete
	if ( m_pxPrev ) m_pxPrev->m_pxNext = m_pxNext; else g_pxSortedListStart = m_pxNext;
	if ( m_pxNext ) m_pxNext->m_pxPrev = m_pxPrev; else g_pxSortedListEnd = m_pxPrev;
	
	// reinsert
	if ( g_pxSortedListEnd ) g_pxSortedListEnd->m_pxNext = this; else g_pxSortedListStart = this;
	m_pxPrev = g_pxSortedListEnd;
	g_pxSortedListEnd = this;
	m_pxNext = NULL;
	
	return m_pxBoneCache;
}


Asura_PS3_GCM_VertexBuffer* Asura_Platform_Animation_RenderCache::ObtainVertexBuffer( ASURA_PS3_RENDERCACHE_VERTEXBUFFER_ID eBuffer, u_int uSkin, u_int uNumVerts, size_t uVertexSize )
{
	Asura_PS3_GCM_VertexBuffer* pVertexBuffer = m_apxBuffers[uSkin]->m_apxVertexBuffer[eBuffer];

	// alloc if necessary
	if ( pVertexBuffer == NULL || uNumVerts > pVertexBuffer->GetNumVerts() || uVertexSize != pVertexBuffer->GetSize() )
	{
		// Assert if the data changed. This is handled but it's better if the code that changed vertex data call Flush() explicitly.
		Asura_Assert(pVertexBuffer==NULL,"Rellocing vertex buffer, vertex data for this skin must have changed.\n");

		if ( pVertexBuffer == NULL )
		{
			pVertexBuffer = new Asura_PS3_GCM_VertexBuffer;
		}
		else if ( pVertexBuffer->IsInitialized() )
		{
			Asura_Assertf( uVertexSize == pVertexBuffer->GetSize(), ("Asura_Platform_Animation_RenderCache::ObtainVertexBuffer - Cached vertex size in animated skin has changed (buffer %d, skin %d), this will cause a realloc & is probably a bad thing.", eBuffer, uSkin));
			//BW if the above assert fires, try always calling this with the same num verts
			// i.e. if you're changing the LOD of the skin on this animation: always call this with the numverts of the highest LOD skin, should make life simpler.

			// have got to throw out the old buffer, but it might still be used by SPU & RSX, so destroy it nicely
			bool bRet = m_apxBuffers[uSkin]->Destroy();
			Asura_Assert(bRet,"Sanity check failed, has behaviour of Asura_PS3_RenderCache_Buffer::Destroy() changed?");
			m_apxBuffers[uSkin] = new Asura_PS3_RenderCache_Buffer;

			pVertexBuffer = new Asura_PS3_GCM_VertexBuffer;
		}

		//create a buffer sized for this number of verts
		pVertexBuffer->Initialise( uVertexSize, uNumVerts, CELL_GCM_LOCATION_MAIN, true );
		
		m_apxBuffers[uSkin]->m_apxVertexBuffer[eBuffer] = pVertexBuffer;
	}

	return pVertexBuffer;
}


Asura_PS3_GCM_IndexBuffer* Asura_Platform_Animation_RenderCache::ObtainIndexBuffer( ASURA_PS3_RENDERCACHE_INDEXBUFFER_ID eBuffer, u_int uSkin, u_int uNumIndices )
{
	Asura_PS3_GCM_IndexBuffer* pIndexBuffer = m_apxBuffers[uSkin]->m_apxIndexBuffer[eBuffer];
	
	// alloc if necessary
	if ( pIndexBuffer == NULL || uNumIndices > pIndexBuffer->GetIndexBufferSize()/2 )
	{
		// Similar to the assert for vertex buffer (above), but much worse.
		// At the time of writing this function is called after ObtainVertexBuffer, so if this fires it means index data got bigger without vertex data changing.
		// I believe this is impossible, so this is not handled well, in fact it will destroy the old index AND VERTEX buffers!
		// This destruction has a 2-frame delay, so despite this case not being well handled it's entirely possible it's non-fatal.
		// In any case (whether it crashes or survives after this assert) you should call Flush() after changing the vertex/index data of the skin.
		Asura_Assert(pIndexBuffer==NULL,"Reallocing index buffer - THIS IS BAD!\n");

		if ( pIndexBuffer == NULL )
		{
			pIndexBuffer = new Asura_PS3_GCM_IndexBuffer;
		}
 		else if ( pIndexBuffer->IsInitialized() )
		{
			Asura_Assert(0,"It probably shouldn't ever get here, I think it might break - BenW"); // because it will throw out the vertex buffer too.
			//BW if the above assert fires, try always calling this with the same num verts
			// i.e. if you're changing the LOD of the skin on this animation: always call this with the numverts of the highest LOD skin, should make life simpler.

			// have got to throw out the old buffer, but it might still be used by SPU & RSX, so destroy it nicely
			bool bRet = m_apxBuffers[uSkin]->Destroy();
			Asura_Assert(bRet,"Sanity check failed, has behaviour of Asura_PS3_RenderCache_Buffer::Destroy() changed?");
			m_apxBuffers[uSkin] = new Asura_PS3_RenderCache_Buffer;

			pIndexBuffer = new Asura_PS3_GCM_IndexBuffer;
		}

		//create a buffer sized for this number of verts
		pIndexBuffer->Initialise( uNumIndices, CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16, CELL_GCM_LOCATION_MAIN, true );

		m_apxBuffers[uSkin]->m_apxIndexBuffer[eBuffer] = pIndexBuffer;
	}

	return pIndexBuffer;
}


Asura_PS3_SPU_Skinning_Batch* Asura_Platform_Animation_RenderCache::GetBatches( u_int uSkin, u_int uNumStrips )
{
	return m_apxBuffers[uSkin]->GetBatches( uNumStrips );
}

Asura_PS3_SPU_Skinning_Batch* Asura_PS3_RenderCache_Buffer::GetBatches( u_int uNumStrips )
{
	if ( m_auNumBatches < uNumStrips+1 )
	{
		if ( m_apxBatches ) delete [] m_apxBatches; //BW this shouldn't cause a crash, since SPUs should have finished by now, but ideally this should destroy & new the same as reallocs in vertex & index buffers
		
		m_apxBatches = new Asura_PS3_SPU_Skinning_Batch[uNumStrips+1];
		m_auNumBatches = uNumStrips+1;
	}
	
	return m_apxBatches;
}

Asura_PS3_RenderCache_Buffer::Asura_PS3_RenderCache_Buffer()
{
	for ( int i=0; i < ASURA_PS3_RENDERCACHE_NUM_VERTEXBUFFERS; i++ )
	{
		m_apxVertexBuffer[i] = NULL;
		m_auVertexFrameStamp[i] = 0;
	}

	for ( int i=0; i < ASURA_PS3_RENDERCACHE_NUM_INDEXBUFFERS; i++ )
	{
		m_apxIndexBuffer[i] = NULL;
		m_auIndexFrameStamp[i] = 0;
	}

	m_apxBatches = NULL;
	m_auNumBatches = 0;
}

Asura_PS3_RenderCache_Buffer::~Asura_PS3_RenderCache_Buffer()
{
	for ( int i=0; i < ASURA_PS3_RENDERCACHE_NUM_VERTEXBUFFERS; i++ )
	{
		if ( m_apxVertexBuffer[i] != NULL )
		{
			m_apxVertexBuffer[i]->Destroy();
			delete m_apxVertexBuffer[i];
			m_apxVertexBuffer[i] = NULL;
		}

		m_auVertexFrameStamp[i] = 0;
	}

	for ( int i=0; i < ASURA_PS3_RENDERCACHE_NUM_INDEXBUFFERS; i++ )
	{
		if ( m_apxIndexBuffer[i] )
		{
			m_apxIndexBuffer[i]->Shutdown();
			delete m_apxIndexBuffer[i];
			m_apxIndexBuffer[i] = NULL;
		}
		m_auIndexFrameStamp[i] = 0;
	}

	if ( m_apxBatches != NULL )
	{
		delete[] m_apxBatches;
		m_apxBatches = NULL;
	}
	m_auNumBatches = 0;
}

bool Asura_PS3_RenderCache_Buffer::Destroy()
{
	if ( m_apxVertexBuffer[0] == NULL ) // everything that uses render caches sets this.
		return false; // this one was never used, so no need to clean up.

	// note: we could free it here if we know it wasn't drawn from last frame, which might be helpful if this is fragging the heap a lot when increasing buffer size.

	// add to the cleanup list
	m_pxNext = g_apxCleanUp[g_uCleanupBuffer];
	g_apxCleanUp[g_uCleanupBuffer] = this;
	
	return true;
}
