//
// Asura_Platform_Animation_RenderCache.h
//
// Ben Weston - 13/06/2008
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __ASURA_PLATFORM_ANIMATION_RENDERCACHE_H__
#define __ASURA_PLATFORM_ANIMATION_RENDERCACHE_H__

#if !defined(ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE)
#error PS3 must have ASURA_USE_ANIMATION_PLATFORM_RENDER_CACHE defined
#endif

#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"

class Asura_PS3_GCM_VertexBuffer;
class Asura_PS3_GCM_IndexBuffer;
class Asura_PS3_SPU_Skinning_Batch;

// support multiple vertex/index buffers, useful for sharing some data. e.g. between shadow & main render
//BW these buffer names are rather specific to skinning, if another module starts using this render cache we should pick appropriately generic names
//BW i.e. there's no reason for the rendercache system to know what the buffer contains, it just manages alloc/free
enum ASURA_PS3_RENDERCACHE_VERTEXBUFFER_ID {
	ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS, // SPU processed world position
	ASURA_PS3_RENDERCACHE_NUM_VERTEXBUFFERS
};

enum ASURA_PS3_RENDERCACHE_INDEXBUFFER_ID {
	ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED, // Indices of faces that passed the SPU culling for the main camera view
	ASURA_PS3_RENDERCACHE_BUFFER_EDGEAA, // Indices of edges to be anti-aliased
	ASURA_PS3_RENDERCACHE_NUM_INDEXBUFFERS
};



class Asura_PS3_RenderCache_Buffer
{
	friend class Asura_Platform_Animation_RenderCache;

protected:
	Asura_PS3_RenderCache_Buffer*	m_pxNext; // maintain linked lists for any globally applicable actions

	// vertex buffers
	Asura_PS3_GCM_VertexBuffer*		m_apxVertexBuffer[ASURA_PS3_RENDERCACHE_NUM_VERTEXBUFFERS];		// Buffers for each type of vertex data
	u_int							m_auVertexFrameStamp[ASURA_PS3_RENDERCACHE_NUM_VERTEXBUFFERS];	// so we can tell if this data has already been update for this frame

	// index buffer - for culled triangle lists.
	// we only cull geometry for the main camera view (since we want multiple shadows without having to cull for each shadow view), so we only have 1 buffer.
	Asura_PS3_GCM_IndexBuffer*		m_apxIndexBuffer[ASURA_PS3_RENDERCACHE_NUM_INDEXBUFFERS];
	u_int							m_auIndexFrameStamp[ASURA_PS3_RENDERCACHE_NUM_INDEXBUFFERS];	// so we can tell if this data has already been update for this frame

	// SPU batch info
	Asura_PS3_SPU_Skinning_Batch*	m_apxBatches;
	u_int							m_auNumBatches;

private:
	Asura_PS3_RenderCache_Buffer();
	~Asura_PS3_RenderCache_Buffer();
	bool Destroy();
	Asura_PS3_SPU_Skinning_Batch* GetBatches( u_int uNumStrips );
};


class Asura_Platform_Animation_RenderCache
{
friend class Asura_Animation;

public:
	Asura_AMatrix4x4* UpdateCachedBones( const Asura_AMatrix4x4* pxBones, u_int uNumBones );
	Asura_AMatrix4x4* GetCachedBones() { return m_pxBoneCache; }
	u_int GetNumCachedBones() { return m_uNumBonesAlloced; }
	bool IsBoneCacheCurrent() { return (Asura_Timers::GetFrameNumber() == m_uBoneFrameStamp); }
	void SetBoneCacheCurrent() { m_uBoneFrameStamp = Asura_Timers::GetFrameNumber(); }

	Asura_PS3_SPU_Skinning_Batch* GetBatches( u_int uSkin, u_int uNumStrips );

	Asura_PS3_GCM_VertexBuffer* ObtainVertexBuffer( ASURA_PS3_RENDERCACHE_VERTEXBUFFER_ID eBuffer, u_int uSkin, u_int uNumVerts, size_t uVertexSize );
	Asura_PS3_GCM_VertexBuffer* ObtainVertexBuffer( ASURA_PS3_RENDERCACHE_VERTEXBUFFER_ID eBuffer, u_int uSkin ) { return m_apxBuffers[uSkin]->m_apxVertexBuffer[eBuffer]; }
	Asura_PS3_GCM_IndexBuffer* ObtainIndexBuffer( ASURA_PS3_RENDERCACHE_INDEXBUFFER_ID eBuffer, u_int uSkin, u_int uNumIndices );
	Asura_PS3_GCM_IndexBuffer* ObtainIndexBuffer( ASURA_PS3_RENDERCACHE_INDEXBUFFER_ID eBuffer, u_int uSkin ) { return m_apxBuffers[uSkin]->m_apxIndexBuffer[eBuffer]; }

	bool IsVertexCacheCurrent( ASURA_PS3_RENDERCACHE_VERTEXBUFFER_ID eBuffer, u_int uSkin )
	{
		return (Asura_Timers::GetFrameNumber() == m_apxBuffers[uSkin]->m_auVertexFrameStamp[eBuffer]);
	}
	void SetVertexCacheCurrent( ASURA_PS3_RENDERCACHE_VERTEXBUFFER_ID eBuffer, u_int uSkin )
	{
		m_apxBuffers[uSkin]->m_auVertexFrameStamp[eBuffer] = Asura_Timers::GetFrameNumber();
	}
	
	bool IsIndexCacheCurrent( ASURA_PS3_RENDERCACHE_INDEXBUFFER_ID eBuffer, u_int uSkin )
	{
		return (Asura_Timers::GetFrameNumber() == m_apxBuffers[uSkin]->m_auIndexFrameStamp[eBuffer]);
	}
	void SetIndexCacheCurrent( ASURA_PS3_RENDERCACHE_INDEXBUFFER_ID eBuffer, u_int uSkin )
	{
		m_apxBuffers[uSkin]->m_auIndexFrameStamp[eBuffer] = Asura_Timers::GetFrameNumber();
	}

	void Flush( u_int uSkin );

	void Destroy();

	// functions to globally manage all render caches
	static void Initialise();
	static void Shutdown();
	static void Cleanup();
	static Asura_Platform_Animation_RenderCache* Create();

	static void RenderDebugInfo();

private:
	Asura_Platform_Animation_RenderCache();
	~Asura_Platform_Animation_RenderCache();
	void Reset();

	// linked list, sorted by age
	class Asura_Platform_Animation_RenderCache* m_pxNext;
	class Asura_Platform_Animation_RenderCache* m_pxPrev;

	// cached data
	// these buffers are allocated when needed (hence keep them private so we can create them when accessed)
	
	// bone cache
	Asura_AMatrix4x4*				m_pxBoneCache;
	u_int							m_uNumBonesAlloced;
	u_int							m_uBoneFrameStamp;

	// vertex/index buffers and SPU batch info
	Asura_PS3_RenderCache_Buffer*	m_apxBuffers[ASURA_ANIMATION_SKIN_COUNT];

};


#endif//__ASURA_PLATFORM_ANIMATION_RENDERCACHE_H__
