//////////////////////////////////////////////////////////////////////////////
//
// Asura_Chunk_ResourceFile_PS3_ObjectHierarchy.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS3_ObjectHierarchy.h"

// Engine platform-specific.
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Material_System.h"
#include "Rendering/Asura_PS3_ObjectHierarchyRenderer.h"
#include "Rendering/Shaders/Asura_PS3_GCM_ShadersLibrary.h"
#include "Rendering/Asura_PS3_MaterialResponse_System.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS3.h"
#include "Rendering/ShadersDefinition/VD/Object_VD.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_Chunk_ResourceFile_PS3_ObjectHierarchy
//

void Asura_Chunk_ResourceFile_PS3_ObjectHierarchy::Process( const char* szResourceFilename, Asura_Chunk_Stream& xStream )
{
	Asura_Assertf( Version == ASURA_PS3_OBJECTHIERARCHY_CHUNK_VERSION, ("Invalid object hierarchy chunk version (%d) encountered. Should be %d. You may need new code, or to re-export the level.", Version, ASURA_PS3_OBJECTHIERARCHY_CHUNK_VERSION) );

	Asura_Hash_ID	xNameHashID;
	u_int			uTotalNumberOfStrips;
	u_int			uTotalNumberOfVertices;
	u_int			uTotalNumberOfIndices;

	// first get the unique hash id for this hierarchy
	xStream >> xNameHashID;

	// stream in the hierarchy description
	xStream >> uTotalNumberOfStrips;
	xStream >> uTotalNumberOfVertices;
	xStream >> uTotalNumberOfIndices;

	// attempt to create a new hierarchy based on the hash id
	Asura_PS3_ObjectHierarchyRenderer::Hierarchy* pxHierarchy = Asura_PS3_ObjectHierarchyRenderer::CreateHierarchy( xNameHashID ); 
	
	// is this invalid?
	if ( !pxHierarchy )
	{
		// Out of memory, or hierarchy already exists. Either way there should have been an assert already.
		xStream.Skip( sizeof( Asura_PS3_ObjectHierarchyRenderer::Strip ) * uTotalNumberOfStrips );
		const u_int uVertexBufferSize = uTotalNumberOfVertices * Asura_PS3_ObjectHierarchyRenderer::GetVertexSize();
		xStream.Skip( uVertexBufferSize );
		const u_int uIndexBufferSize = uTotalNumberOfIndices * sizeof(short);
		xStream.Skip( uIndexBufferSize );
		
		if ( Version > 5 )
		{
			u_int	uNumBones = 0;
			u_int	uSkipNumEdgeIndices = 0;
			xStream >> uNumBones;
			for ( u_int uBone = 0 ; uBone < uNumBones; uBone++ )
			{
				int iNumEdgeIndices;
				xStream >> iNumEdgeIndices;
				uSkipNumEdgeIndices = iNumEdgeIndices * sizeof(u_short);
				xStream.Skip( uSkipNumEdgeIndices );
			}
		}
		
		return;
	}

	pxHierarchy->m_uNumberOfStrips = uTotalNumberOfStrips;

    // Some hierarchies, or sections of them, won't have any vertices or
    // indices, since artists sometimes use empty ones for faking up other
    // things. In this case, it shouldn't have any strips either.
    Asura_Assert( (uTotalNumberOfStrips == 0) == (uTotalNumberOfVertices == 0), "" );
    Asura_Assert( (uTotalNumberOfStrips == 0) == (uTotalNumberOfIndices == 0), "" );

    // And so we can just stop at this point.
    if( !uTotalNumberOfStrips ) { return; }

	// we now need to allocate enough strip descriptors, and then read in the data for them
	pxHierarchy->m_pxStrips	= new Asura_PS3_ObjectHierarchyRenderer::Strip[ uTotalNumberOfStrips ];

	xStream.Read( sizeof( Asura_PS3_ObjectHierarchyRenderer::Strip ) * uTotalNumberOfStrips, pxHierarchy->m_pxStrips );

	// setup vertex buffer and stream in the data (in one big chunk)
	if( uTotalNumberOfVertices > 0)
	{
		Asura_PS3_GCM_VertexBuffer& xVertexBuffer = pxHierarchy->m_xVertexBuffer;

		// generate the vertex buffer
        const u_int uVtxSize = Asura_PS3_ObjectHierarchyRenderer::GetVertexSize();
		const u_int uVertexBufferSize = uTotalNumberOfVertices * uVtxSize;
		xVertexBuffer.Initialise( uVtxSize, uTotalNumberOfVertices, CELL_GCM_LOCATION_LOCAL, true );

		// lock the buffer
		void* pBuffer = xVertexBuffer.LockForWriting();
		
		// read in data
		xStream.Read( uVertexBufferSize, pBuffer );

		// unlock the buffer
		xVertexBuffer.Unlock();
	}

	// set up index buffer and stream in the data (in one big chunk)
	if( uTotalNumberOfIndices > 0 )
	{
		Asura_PS3_GCM_IndexBuffer& xIndexBuffer	= pxHierarchy->m_xIndexBuffer;

		// generate the index buffer
		const u_int uIndexBufferSize = uTotalNumberOfIndices * sizeof(short);
		xIndexBuffer.Initialise( uTotalNumberOfIndices, CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16, CELL_GCM_LOCATION_LOCAL, true );
		
		// lock the buffer
		void* pBuffer = xIndexBuffer.LockForWriting();
		
		// read in data
		xStream.Read( uIndexBufferSize, pBuffer );

		// unlock the buffer
		xIndexBuffer.Unlock();
	}
	

	// read in the Anti-Alias edges.
	// Note: this is compatible with objects that don't have any, and the old code will ignore this extra data (though you get an assert)

	if ( Version > 4 )
	{
		u_int uNumBones = 0;

		if ( Version > 5 )
		{
			xStream >> uNumBones;
		}
		else
		{
			// in older versions bone number was deduced as follows, this is inconsistent with tools so was replaced.
			for ( u_int i=0; i < uTotalNumberOfStrips; i++ )
			{
				if ( uNumBones < pxHierarchy->m_pxStrips[i].m_uBoneIndex ) uNumBones = pxHierarchy->m_pxStrips[i].m_uBoneIndex;
			}
			uNumBones++; //obviously if bone 1 is indexed there's 2 bones!
		}
	
		pxHierarchy->m_uNumAABuffers = uNumBones;
		pxHierarchy->m_pxAntiAliasEdges = new Asura_PS3_GCM_IndexBuffer[uNumBones];
	
		for ( u_int uBone=0; uBone < uNumBones; uBone++ )
		{
			int iNumEdgeIndices;
			xStream >> iNumEdgeIndices;
			
			if ( iNumEdgeIndices > 0 )
			{
				pxHierarchy->m_pxAntiAliasEdges[uBone].Initialise( iNumEdgeIndices, CELL_GCM_DRAW_INDEX_ARRAY_TYPE_16, CELL_GCM_LOCATION_LOCAL, true );
				
				u_short* puIdxOut = static_cast<u_short*>(pxHierarchy->m_pxAntiAliasEdges[uBone].LockForWriting());
				xStream.Read( iNumEdgeIndices*2, puIdxOut );
			}
		}
	}
}
