///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_FMV_Streamer.cpp
// Author    : Philip Hall
// Created   : 19/10/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "FMV/Asura_PS3_FMV_Streamer.h"

// Asura headers
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_LogFile.h"

#include <cell/sysmodule.h>
#include <sys/timer.h>
#include <sys/ppu_thread.h>


////PGH/////////////////////////////////////////
// General constants (from pamf_dmux sample)
////////////////////////////////////////////////
const int       STMIN_PPU_THREAD_PRIO = 600;
const size_t    STMIN_PPU_STACK_SIZE  = 8192;
const int       STMIN_BUFFER_COUNT    = 3;          // Number of stream read buffers 
const usecond_t STREAM_RETRY_WAIT     = 1000 * 50;  // usec to wait between read attempts
const usecond_t STREAM_READ_WAIT      = 1000 * 1;   // usec to wait between reads


Asura_PS3_FMV_Streamer::Asura_PS3_FMV_Streamer()
{
	m_StreamHandle = ASURA_INVALID_HANDLE;
}

Asura_PS3_FMV_Streamer::~Asura_PS3_FMV_Streamer()
{
	Close();
}


bool Asura_PS3_FMV_Streamer::Initialise(Asura_PS3_FMV_Control* pControl, const char* szFilename, off64_t fileOffset, off64_t streamSize)
{
	strncpy(m_szFilename, szFilename, ASURA_PS3_MAX_PATH);
	m_pControl   = pControl;
	m_fileOffset = fileOffset;
	m_streamSize = streamSize;
	m_bFinished  = false;

	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::Initialise] - Setup streaming for file: %s\n", m_szFilename));
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_Streamer::Open
// Description : Open this streamer on the configured file and initialise the
//               read data banks (taken from PAMF sample).
//
// Parameters  : none
//
// Returns     : bool - Returns true on success 
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV_Streamer::Open()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::Open] - Opening Streamer\n"));
	
	if ( !Asura_PS3_FileSystem::OpenForReading(m_szFilename, &m_StreamHandle) )
	{
		Asura_Assertf(false, ("[Asura_PS3_FMV_Streamer::Open] - Unable to open file: %s\n", m_szFilename));
		return false;
	}
	
	if ( !m_xReadBuffer.Create( STMIN_BUFFER_COUNT, ASURA_ALIGN_128_BYTE ) )
	{
		Asura_Assert(false, "[Asura_PS3_FMV_Streamer::Open] - Unable to allocate read buffer memory\n");
		return false;
	}

	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_Streamer::Close
// Description : Close & clean up after stream reading
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV_Streamer::Close()
{
	Asura_PS3_FileSystem::Close( &m_StreamHandle );
	m_StreamHandle = ASURA_INVALID_HANDLE;

	m_xBufferLock.Destroy();
	m_xReadySignal.Destroy();
	m_xReadBuffer.Destroy();
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_Streamer::Begin
// Description : Start streaming data from the requried source file
//
// Parameters  : none
//
// Returns     : bool - Returns true on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_FMV_Streamer::Begin()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::Begin] - Beginning...\n"));

	m_bFinished	= false;
	m_uTotalBytes = 0;
	
	int iResult = sys_ppu_thread_create(&m_threadId, &Asura_PS3_FMV_Streamer::ThreadProxy, (uintptr_t)this, 
	                                     STMIN_PPU_THREAD_PRIO, STMIN_PPU_STACK_SIZE, SYS_PPU_THREAD_CREATE_JOINABLE, "Asura_PS3_FMV_Streamer");
	if (iResult < CELL_OK)
	{
		Asura_Assert(false, ("[Asura_PS3_FMV_Streamer::Start] - Unable to create PPU thread\n"));
		return false;
	}
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_Streamer::End
// Description : Request end of streaming and join thread
//
// Parameters  : none
//
// Returns     : int - Returns CELL result code
///////////////////////////////////////////////////////////////////////////////
int Asura_PS3_FMV_Streamer::End()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::End] - Ending...\n"));

	m_bFinished = true;

	uint64_t thrStat;
	int	iResult = sys_ppu_thread_join(m_threadId, &thrStat);
	if (iResult < CELL_OK)
	{
		Asura_Assert(false, "[Asura_PS3_FMV_Streamer::WaitComplete] - Unable to join thread\n");
		return iResult;
	}

	return CELL_OK;
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_Streamer::Abort
// Description : Abort file streaming
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV_Streamer::Abort()
{
	m_bFinished = true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_FMV_Streamer::Thread
// Description : Streaming threads
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_FMV_Streamer::Thread()
{
	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::Thread] - Started...\n"));

	m_bFinished = false;
	m_uTotalBytes = 0;
	
	// Initial seek
	if ( !Asura_PS3_FileSystem::SetCurrentPosition(m_StreamHandle, m_fileOffset) )
	{
		m_bFinished = true;
	}

	// Stream file contents
	while ( !m_bFinished )
	{
		if ( m_pControl->IsComplete() ) { break; }

		// Begin block allocation
		Asura_PS3_StreamBlock* pBlock = NULL;
		if ( m_xBufferLock.Lock(STREAM_READ_WAIT) )
		{
			pBlock = m_xReadBuffer.Alloc();
			m_xBufferLock.Unlock();	
		}			

		// If no free blocks yet then sleep & continue
		if (pBlock == NULL)
		{
			sys_timer_usleep(STREAM_RETRY_WAIT);
			continue;
		}

		pBlock->bStart = (m_uTotalBytes == 0);
		
		// Read data from the source file
		u_int uReadBytes = Asura_PS3_FileSystem::ReadData(m_StreamHandle, STMIN_BANK_SIZE, pBlock->acData);
		pBlock->uSize = uReadBytes;
		if (uReadBytes == 0)
		{
			m_pControl->Abort(CELL_OK); // We've already finished
			break;
		}
		else if ( uReadBytes < STMIN_BANK_SIZE )
		{
			m_bFinished = true;	        // This is the last block
		}
		
		m_uTotalBytes += uReadBytes;    // Track number of bytes streamed
		
		m_xReadySignal.Signal();
	}
	
	m_pControl->EndStream();

	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::Thread] - Streamer read %d bytes total\n", m_uTotalBytes));
	Asura_OutputToDebugger(("[Asura_PS3_FMV_Streamer::Thread] - ...Finished\n"));
	sys_ppu_thread_exit(0);
}
