
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PS3_FileSystem.cpp
// Author : Andy Tate
// Date   : 23-Sep-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "FileAccess/Asura_PS3_FileSystem.h"

// Stdlib independent.
#include <string.h>

// Stdlib platform-specific.
#include <cell/cell_fs.h>
#include <np/drm_inline.h>
#include <sys/ppu_thread.h>
#include <sysutil/sysutil_syscache.h>
#include <unistd.h>

// Engine independent.
#include "Core/Asura_Core.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
#include "Network/Asura_DLC_System.h"
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT

// Engine PS3.
#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
#include "Network/Asura_PS3_DLC_System.h"
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
#include "Rendering/Asura_PS3_Graphics.h"

// D E F I N E S /////////////////////////////////////////////////////////////
#define ASURA_PS3_LOG_FILE_OPERATION(x) // printf x

// C O N S T A N T S /////////////////////////////////////////////////////////

static const char* const szMOUNT_PREFIX_HOSTFS = "/app_home";
static const char* const szMOUNT_PREFIX_BDVD   = "/dev_bdvd/PS3_GAME/USRDIR";

static const u_int uMAX_TOTAL_PATH_LEN = 256; // Including nul terminator.

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////
const char* Asura_PS3_FileSystem::s_szMountPrefix = szMOUNT_PREFIX_BDVD;
bool Asura_PS3_FileSystem::s_bInitialised;

CellSysCacheParam	s_xCacheParam = { ASURA_PS3_CACHE_ID, "",  NULL };
////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::Initialise
// Author       : Andy Tate
// Date         : 31-Aug-2005
//
// Description  : Handles the initialisation of the PS3 file-system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PS3_FileSystem::Initialise(const char* szBootfile, bool bForceDiscAccess)
{
	if (bForceDiscAccess)
	{
		printf("forcing disk access\n");
	}
	else if( szBootfile && ( !strnicmp( szBootfile, "/app_home", 9 ) || !strnicmp( szBootfile, "/host_root", 10 ) ) )
	{
		s_szMountPrefix = szMOUNT_PREFIX_HOSTFS;
	}
	printf("file root = %s\n", s_szMountPrefix);
	s_bInitialised = true;

	int iRet = cellSysCacheMount( &s_xCacheParam );
	if( iRet < 0 )
	{
		Asura_OutputToDebugger( ( "cellSysCacheMount FAILED!" ) );
	}
}

/// End of function 'Asura_PS3_FileSystem::Initialise'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::Shutdown
// Author       : Andy Tate
// Date         : 31-Aug-2005
//
// Description  : Handles the shutdown of the PS3 file-system.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PS3_FileSystem::Shutdown (void)
{
	s_bInitialised = false;
}

/// End of function 'Asura_PS3_FileSystem::Shutdown'. ///



bool Asura_PS3_FileSystem::OpenForAppending( const char* szFilename, Handle* pxHandle, bool bForceHOSTFS, bool bOpenInCache )
{
	if (!s_bInitialised)
	{
		return false;
	}

    char szFullPath[uMAX_TOTAL_PATH_LEN];
	if( bOpenInCache )
	{
		GenerateFilenameInCache( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN );
	}
	else
	{
		GenerateFullFilename( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN, bForceHOSTFS );
	}

    // Open existing file, if possible.
    CellFsErrno iError = cellFsOpen( szFullPath, CELL_FS_O_WRONLY, pxHandle, NULL, 0 );
    if( CELL_FS_SUCCEEDED == iError )
    {
        uint64_t uCurPos;
        iError = cellFsLseek( *pxHandle, 0, CELL_FS_SEEK_END, &uCurPos );

        return( CELL_FS_SUCCEEDED == iError );
    }

    // Ok, try opening a new one.
    if( OpenForWriting(szFilename, pxHandle, bForceHOSTFS) )
    {
        return true;
    }

    return false;
}


bool Asura_PS3_FileSystem::OpenForReading( const char* szFilename, Handle* pxHandle, bool bForceHOSTFS, bool bOpenInCache )
{
	if (!s_bInitialised)
	{
		return false;
	}

	ASURA_PS3_LOG_FILE_OPERATION(("File %s, OpenForReading() - %u ms\n", szFilename, Asura_Timers::GetMSecsSinceStart_Continuous()));
	
	char szFullPath[uMAX_TOTAL_PATH_LEN];

#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	// DLC can be opened for reading, so check that here.
	const bool bDLCFile = Asura_DLC_System::IsDownloadedFile(szFilename);
	const bool bEncriptedFile = Asura_PS3_DLC_System::IsEncryptedFile(szFilename);
	Asura_Assert(!bEncriptedFile || bDLCFile, "Attempted to open an encrypted file which hasn't been registered with the DLC system. Is this a file we don't have access rights to?");
	if (bEncriptedFile && !bDLCFile) { return false; }

	if (bDLCFile)
	{
		GenerateDLCFilename(szFilename, szFullPath, uMAX_TOTAL_PATH_LEN);
	}
	else
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	if( bOpenInCache )
	{
		GenerateFilenameInCache( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN );
	}
	else
	{
		GenerateFullFilename( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN, bForceHOSTFS );
	}

    CellFsErrno xError;
#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	if (bEncriptedFile)
	{
		xError = sceNpDrmOpen(Asura_PS3_DLC_System::GetDRMKey(), szFullPath, CELL_FS_O_RDONLY, pxHandle, NULL, 0);
	}
	else
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	{
		xError = cellFsOpen( szFullPath, CELL_FS_O_RDONLY, pxHandle, NULL, 0 );
	}

	if (xError != CELL_OK)
	{
		if (xError == CELL_FS_ERROR_ENOTMOUNTED)
		{
			printf("open %s failed with CELL_FS_ERROR_ENOTMOUNTED: aborting\n", szFilename);
			Asura_PS3_Graphics::SetIsInShutdown();
			if (!Asura_Core::IsMainThread())
			{
				printf("exiting thread...\n");
				sys_ppu_thread_exit(-1);
			}
		}
		else
		{
			printf("open %s failed (0x%x)\n", szFilename, xError);
		}
	}

	ASURA_PS3_LOG_FILE_OPERATION(("File %s, OpenForReading() return %x, handle %u - %u ms\n", szFilename, xError, *pxHandle, Asura_Timers::GetMSecsSinceStart_Continuous()));

	// CELL_FS_ENOENT (0x80010006) means file doesn't exist
    return( CELL_FS_SUCCEEDED == xError );
}

bool Asura_PS3_FileSystem::OpenForWriting( const char* szFilename, Handle* pxHandle, bool bForceHOSTFS, bool bOpenInCache )
{
	if (!s_bInitialised)
	{
		return false;
	}

	char szFullPath[uMAX_TOTAL_PATH_LEN];
	if( bOpenInCache )
	{
		GenerateFilenameInCache( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN );
	}
	else
	{
		GenerateFullFilename( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN, bForceHOSTFS );
	}

    CellFsErrno iError = cellFsOpen( szFullPath,
                                     CELL_FS_O_WRONLY | CELL_FS_O_CREAT | CELL_FS_O_TRUNC,
                                     pxHandle, NULL, 0 );

    if( CELL_FS_SUCCEEDED == iError )
    {
        return true;
    }
    else
    {
    	// CELL_FS_ENOENT (0x80010006) means file doesn't exist
    	Asura_OutputToDebugger_Retail(( "Failed to open '%s' for writing (0x%08X)\n", szFullPath, iError ));
        return false;
    }
}

bool Asura_PS3_FileSystem::OpenForReadWrite( const char* szFilename, Handle* pxHandle, bool bForceHOSTFS /* = false */, bool bOpenInCache /*= false */ )
{
	if (!s_bInitialised)
	{
		return false;
	}

	char szFullPath[uMAX_TOTAL_PATH_LEN];
	if( bOpenInCache )
	{
		GenerateFilenameInCache( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN );
	}
	else
	{
		GenerateFullFilename( szFilename, szFullPath, uMAX_TOTAL_PATH_LEN );
	}

	CellFsErrno xError = cellFsOpen( szFullPath, CELL_FS_O_RDWR | CELL_FS_O_CREAT | CELL_FS_O_TRUNC, pxHandle, NULL, 0 );

	// CELL_FS_ENOENT (0x80010006) means file doesn't exist
	return( CELL_FS_SUCCEEDED == xError );
}

bool Asura_PS3_FileSystem::CreateDir( const char* const szDir )
{
	if (!s_bInitialised)
	{
		return false;
	}

	char szFullPath[uMAX_TOTAL_PATH_LEN];
	GenerateFullFilename( szDir, szFullPath, uMAX_TOTAL_PATH_LEN );
	const int iError = mkdir( szFullPath, S_IROTH | S_IWOTH | S_IXOTH );

	return ( iError == 0 );
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::ReadData
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Reads the specified amount of data from the stream into the
//                buffer given.
//
// Parameters   : The handle of the file to read from, the number of bytes to be
//                read and a pointer to the address to read the data into.
//
// Return value : The number of bytes successfully read.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_PS3_FileSystem::ReadData (const Handle xHandle,
                                      const u_int uNumberOfBytes,
                                      void* pData)
{
	ASURA_PS3_LOG_FILE_OPERATION(("File %u, ReadData() %u bytes to 0x%x - %u ms\n", xHandle, uNumberOfBytes, pData, Asura_Timers::GetMSecsSinceStart_Continuous()));

    uint64_t uNumBytesRead;
    CellFsErrno xError = cellFsRead( xHandle, pData, uNumberOfBytes, &uNumBytesRead );
	if( Asura_PS3_Graphics::IsInShutdown() )
	{
		printf("read data: aborting\n");
		return 0;
	}

	#ifdef ASURA_ASSERT
    // Validation...
	switch(xError)
	{
		case CELL_FS_SUCCEEDED:		break;
		case CELL_FS_EBADF:			Asura_Assert(false, "cellFsRead() Failed: File descriptor is invalid or not open for read"); break;
		case CELL_FS_EIO:			Asura_Assert(false, "cellFsRead() Failed: I/O error has occurred"); break;
		case CELL_FS_ENOMEM:		Asura_Assert(false, "cellFsRead() Failed: Out of memory"); break;
		case CELL_FS_EFSSPECIFIC:	Asura_Assert(false, "File system specific internal error has occurred"); break;
		case CELL_FS_EPERM:			Asura_Assert(false, "Read permission does not exist in file"); break;
		case CELL_FS_EFAULT:		Asura_Assert(false, "cellFsRead() Failed: Buffer is NULL"); break;

		default:					Asura_Assertf(false, ("cellFsRead() Failed with an invalid error code! (%d)", xError));
	}
	#endif

	// EIO is a fatal error as of SDK 210.
	if (CELL_FS_EIO == xError)
	{
		printf("read data failed with EIO: aborting\n");
		Asura_PS3_Graphics::SetIsInShutdown();
		if (!Asura_Core::IsMainThread())
		{
			printf("exiting thread...\n");
			sys_ppu_thread_exit(-1);
		}
	}

	ASURA_PS3_LOG_FILE_OPERATION(("File %u, ReadData() complete (%u bytes) - %u ms\n", xHandle, uNumberOfBytes, Asura_Timers::GetMSecsSinceStart_Continuous()));

    // Return the number of bytes read...
    return uNumBytesRead;
}

/// End of function 'Asura_PS3_FileSystem::ReadData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::SkipData
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Reads the specified amount of data from the stream into the
//                buffer given.
//
// Parameters   : The handle of the file to skip and the number of bytes to be
//                skipped.
//
// Return value : The number of bytes successfully skipped.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_PS3_FileSystem::SkipData (const Handle xHandle,
                                      const u_int uNumberOfBytes)
{
    uint64_t uCurPos;
    CellFsErrno xError  = cellFsLseek( xHandle, uNumberOfBytes, CELL_FS_SEEK_CUR, &uCurPos );

    // Was the skip successful?
    if (CELL_FS_SUCCEEDED == xError)
    {
        // Yes - return the number of bytes skipped...
        return  (uNumberOfBytes);
    }

    else
    {
        // No - return no bytes skipped...
        Asura_Assertf   ((CELL_FS_SUCCEEDED == xError), ("cellFsLseek(): Failed %d\n", xError));
		// EIO is a fatal error as of SDK 210.
		if (CELL_FS_EIO == xError)
		{
			printf("SkipData failed with EIO: aborting\n");
			Asura_PS3_Graphics::SetIsInShutdown();
			if (!Asura_Core::IsMainThread())
			{
				printf("exiting thread...\n");
				sys_ppu_thread_exit(-1);
			}
		}
        return (0);
    }
}

/// End of function 'Asura_PS3_FileSystem::SkipData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::WriteData
// Author       : Andy Tate
// Date         : 9-Jun-2005
//
// Description  : Writes the specified amount of data from the stream into the
//                buffer given.
//
// Parameters   : The handle of the file to write to, the number of bytes to be
//                written and a pointer to the address to write the data from.
//
// Return value : The number of bytes successfully written.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_PS3_FileSystem::WriteData (const Handle xHandle,
                                       const u_int uNumberOfBytes,
                                       const void* pData)
{
    uint64_t uNumBytesWritten;
#ifdef ASURA_ASSERT
    CellFsErrno xError = 
#endif // #ifdef ASURA_ASSERT
		cellFsWrite( xHandle, pData, uNumberOfBytes, &uNumBytesWritten );

    // Return the number of bytes written...
    Asura_Assertf   ((CELL_FS_SUCCEEDED == xError), ("cellFsWrite(): Failed %d\n", xError));
    return uNumBytesWritten;
}

/// End of function 'Asura_PS3_FileSystem::WriteData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::SetCurrentPosition
// Author       : Andy Tate
// Date         : 28-Jun-2005
//
// Description  : Sets the current byte position within the currently open file.
//
// Parameters   : The handle of the file to set the current position for, and
//                the offset position within the current file.
//
// Return Value : "true" if successful, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PS3_FileSystem::SetCurrentPosition (const Handle xHandle,
                                               const u_int uPosition)
{
    uint64_t uCurPos;
    CellFsErrno xError = cellFsLseek( xHandle, uPosition, CELL_FS_SEEK_SET, &uCurPos );

	#ifdef ASURA_ASSERT
	// Validation...
	switch(xError)
	{
	case CELL_FS_SUCCEEDED:		break;
	case CELL_FS_EBADF:			Asura_Assert(false, "cellFsRead() Failed: File descriptor is invalid or not open for read"); break;
	case CELL_FS_EIO:			Asura_Assert(false, "cellFsRead() Failed: I/O error has occurred"); break;
	case CELL_FS_ENOMEM:		Asura_Assert(false, "cellFsRead() Failed: Out of memory"); break;
	case CELL_FS_EFSSPECIFIC:	Asura_Assert(false, "File system specific internal error has occurred"); break;
	case CELL_FS_EPERM:			Asura_Assert(false, "Read permission does not exist in file"); break;
	case CELL_FS_EFAULT:		Asura_Assert(false, "cellFsRead() Failed: Buffer is NULL"); break;

	default:					Asura_Assertf(false, ("cellFsRead() Failed with an invalid error code! (%d)", xError));
	}
	#endif

    // Return the current position...
    Asura_Assertf   ((CELL_FS_SUCCEEDED == xError), ("cellFsLseek(): Failed %d\n", xError));
    return          (CELL_FS_SUCCEEDED == xError);
}

/// End of function 'Asura_PS3_FileSystem::SetCurrentPosition'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::GetCurrentPosition
// Author       : Andy Tate
// Date         : 28-Jun-2005
//
// Description  : Gets the current byte position within the currently open file.
//
// Parameters   : The handle of the file to get the current position for.
//
// Return value : The offset position within the current file.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_PS3_FileSystem::GetCurrentPosition (const Handle xHandle)
{
    uint64_t uCurPos;
#ifdef ASURA_ASSERT
    CellFsErrno xError = 
#endif // #ifdef ASURA_ASSERT
		cellFsLseek( xHandle, 0, CELL_FS_SEEK_CUR, &uCurPos );

    // Return the current position...
    Asura_Assertf   ((CELL_FS_SUCCEEDED == xError), ("cellFsLseek(): Failed %d\n", xError));
    return uCurPos;
}

/// End of function 'Asura_PS3_FileSystem::GetCurrentPosition'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::GetFileSize
// Author       : Andy Tate
// Date         : 28-Jun-2005
//
// Description  : Gets the size of the currently open file.
//
// Parameters   : The handle of the file to get the current size of.
//
// Return value : The size of the file.
////////////////////////////////////////////////////////////////////////////////

u_int Asura_PS3_FileSystem::GetFileSize (const Handle xHandle)
{
    CellFsStat  xFileStats;
	xFileStats.st_size = 0;

#ifdef ASURA_ASSERT
    CellFsErrno xError = 
#endif // #ifdef ASURA_ASSERT
		cellFsFstat (xHandle, &xFileStats);

    // Return the size of the file...
    Asura_Assertf   ((CELL_FS_SUCCEEDED == xError), ("cellFsFstat(): Failed %d\n", xError));
    return          (xFileStats.st_size);
}

/// End of function 'Asura_PS3_FileSystem::GetFileSize'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_FileSystem::Close
// Author       : Andy Tate
// Date         : 10-Jun-2005
//
// Description  : Closes the currently open file.
//
// Parameters   : A pointer to the handle of the file to be closed.
//
// Return value : "true" if closed successfully, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PS3_FileSystem::Close (Handle* pxHandle)
{
	ASURA_PS3_LOG_FILE_OPERATION(("File %u, Close() - %u ms\n", *pxHandle, Asura_Timers::GetMSecsSinceStart_Continuous()));

	CellFsErrno xError  = cellFsClose (*pxHandle);

	ASURA_PS3_LOG_FILE_OPERATION(("File %u, Close() complete - %u ms\n", *pxHandle, Asura_Timers::GetMSecsSinceStart_Continuous()));

    // Reset the handle...
    *pxHandle   = INVALID_HANDLE;

    // Return whether we were successful or not...
    return  (CELL_FS_SUCCEEDED == xError);
}

/// End of function 'Asura_PS3_FileSystem::Close'. ///

// Functions for generating complete filenames.
const char* Asura_PS3_FileSystem::GetMountPrefix(bool bForceHOSTFS/* = false*/)
{
	return (bForceHOSTFS ? szMOUNT_PREFIX_HOSTFS : s_szMountPrefix);
}

void Asura_PS3_FileSystem::GenerateFullFilename( const char* szInput, char* szOutputBfr, 
												u_int uOutputBfrSize, bool bForceHOSTFS )
{
	GenerateFilenameFromPrefix(GetMountPrefix(bForceHOSTFS), szInput, szOutputBfr, uOutputBfrSize);
}

void Asura_PS3_FileSystem::GenerateFilenameInCache( const char* szInput, char* szOutputBfr, u_int uOutputBfrSize )
{
	GenerateFilenameFromPrefix(s_xCacheParam.getCachePath, szInput, szOutputBfr, uOutputBfrSize);
}

#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
void Asura_PS3_FileSystem::GenerateDLCFilename(const char* szInput, char* szOutputBfr, u_int uOutputBfrSize)
{
	GenerateFilenameFromPrefix(Asura_PS3_DLC_System::GetRootDirectory(), szInput, szOutputBfr, uOutputBfrSize);
}
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT

void Asura_PS3_FileSystem::GenerateFilenameFromPrefix(const char* szPrefix, const char* szInput, char* szOutputBfr, u_int uOutputBfrSize)
{
	char acUppercase[uMAX_TOTAL_PATH_LEN];
	Asura_PrintfToBfr(acUppercase, "%s", uMAX_TOTAL_PATH_LEN, true, szInput);
	Asura_StringUtil::MakeUppercase(acUppercase);

    Asura_PrintfToBfr(szOutputBfr,
                       "%s%s%s",
                       uOutputBfrSize, true,
					   szPrefix, ('/' == *szInput || '\\' == *szInput ? "" : "/"), acUppercase );

    // Replace all Windows back slashes with Unix forward slashes.
    Asura_StringUtil::ReplaceAllChars(szOutputBfr, '\\', '/');
}


void Asura_PS3_FileSystem::ClearCache( void )
{
	cellSysCacheClear();
}

void Asura_PS3_FileSystem::FindFiles(Asura_List<Asura_String>& xFiles, const char* szMountPrefix, const char* szDirectory, bool bRecursive/* = true*/)
{
	char acFullPath[uMAX_TOTAL_PATH_LEN];
	char acUppercaseDirectory[uMAX_TOTAL_PATH_LEN];
	int iRet;
	const bool bLeadingSlash = ('/' == *szDirectory) || ('\\' == *szDirectory);
	const bool bTrailingSlash = Asura_StringUtil::EndsWith(szDirectory, "/", true) || Asura_StringUtil::EndsWith(szDirectory, "\\", true);

	Asura_PrintfToBfr(acUppercaseDirectory, "%s", uMAX_TOTAL_PATH_LEN, true, szDirectory);
	Asura_StringUtil::MakeUppercase(acUppercaseDirectory);

	Asura_PrintfToBfr(acFullPath, "%s%s%s", uMAX_TOTAL_PATH_LEN, true, szMountPrefix, bLeadingSlash ? "" : "/", acUppercaseDirectory);
	Asura_StringUtil::ReplaceAllChars(acFullPath, '\\', '/');

	bool bWhileOk = true;
	PS3_DirectoryEntry xContents;
	int xFileDescriptor = OpenDir(acFullPath);

	while (bWhileOk)
	{
		iRet = ReadDir(xFileDescriptor, &xContents);
		if (iRet < 0)
		{
			Asura_Assert(false , "Error during reading on PS3 (ReadDir)");

			CloseDir(xFileDescriptor);
			bWhileOk = false;
		}		
		else if (iRet == 0) //finished reading context of directory
		{
			CloseDir(xFileDescriptor);
			bWhileOk = false;
		}
		else
		{
			if (xContents.szName && (xContents.szName[0] != '.')) // ignore . and ..
			{
				Asura_PrintfToBfr(acFullPath, "%s%s%s", uMAX_TOTAL_PATH_LEN, true, acUppercaseDirectory, bTrailingSlash ? "" : "/", xContents.szName );		

				switch( xContents.ucType )
				{
				case PS3_TOF_File:
					{
						xFiles.AddToList(new Asura_String(acFullPath));
						break;
					}
				case PS3_TOF_Directory:
					{
						if( bRecursive )
						{
							FindFiles(xFiles, szMountPrefix, acFullPath, bRecursive);
						}
						break;
					}
				default:
					{
						break;
					}
				}
			}
		}
	}//while
}

int Asura_PS3_FileSystem::OpenDir( const char* szFullPath )
{
	int iFileDesc = -1;

#ifdef ASURA_ASSERT
	CellFsErrno xError = 
#endif
	cellFsOpendir( szFullPath, &iFileDesc );
#ifdef ASURA_ASSERT
	switch( xError )
	{
	case CELL_FS_SUCCEEDED:
		break;
	case CELL_FS_ENOTMOUNTED:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::OpenDir -> File system corresponding to %s is not mounted", szFullPath) );
		break;
	case CELL_FS_ENOENT:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::OpenDir -> The specified directory does not exist: %s", szFullPath) );
		break;
	case CELL_FS_ENOTDIR:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::OpenDir -> The specified path is not a directory: %s", szFullPath) );
		break;
	case CELL_FS_EMFILE:
		Asura_Assert( false, ("Asura_PS3_FileSystem::OpenDir -> The number of file descriptors that can be opened at the same time has reached the maximum") );
		break;
	case CELL_FS_EIO:
		Asura_Assert( false, ("Asura_PS3_FileSystem::OpenDir -> I/O error has occurred") );
		break;
	case CELL_FS_ENOMEM:
		Asura_Assert( false, ("Asura_PS3_FileSystem::OpenDir -> Memory is insufficient") );
		break;
	case CELL_FS_ENAMETOOLONG:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::OpenDir -> Path or components in path exceed the maximum length:%f", szFullPath) );
		break;
	case CELL_FS_EFSSPECIFIC:
		Asura_Assert( false, ("Asura_PS3_FileSystem::OpenDir -> File system specific internal error has occurred") );
		break;
	case CELL_FS_EPERM:
		Asura_Assert( false, ("Asura_PS3_FileSystem::OpenDir -> Permission is invalid") );
		break;
	case CELL_FS_EFAULT:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::OpenDir -> %s or fileDescriptor is NULL", szFullPath) );
		break;
	default:
		Asura_Assert( false, ("Asura_PS3_FileSystem::OpenDir -> Unknown error") );
		break;
	}
#endif
	return iFileDesc;
}

void Asura_PS3_FileSystem::CloseDir( int iFileDescriptor )
{
#ifdef ASURA_ASSERT
	CellFsErrno xError = 
#endif
	cellFsClosedir( iFileDescriptor );

#ifdef ASURA_ASSERT
	switch( xError )
	{
	case CELL_FS_SUCCEEDED:
		break;
	case CELL_FS_EBADF:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::CloseDir -> File descriptor( %d ) is invalid", iFileDescriptor) );
		break;
	case CELL_FS_EIO:
		Asura_Assert( false, ("Asura_PS3_FileSystem::CloseDir -> I/O error has occurred") );
		break;
	case CELL_FS_EFSSPECIFIC:
		Asura_Assert( false, ("Asura_PS3_FileSystem::CloseDir -> File system specific internal error has occurred") );
		break;
	default:
		Asura_Assert( false, ("Asura_PS3_FileSystem::CloseDir -> Unknown error") );
		break;
	}
#endif
}

int Asura_PS3_FileSystem::ReadDir( int iFileDescriptor, PS3_DirectoryEntry* pxDirectoryEntry )
{
#ifdef ASURA_ASSERT
	if( !pxDirectoryEntry )
	{
		Asura_Assert( false, ("Asura_PS3_FileSystem::ReadDir -> Inavlid input: PS3_DirectoryEntry is NULL") );
		return -1;
	}
#endif

	uint64_t		xRead;
	CellFsDirent	xDirEntry;
	int				iErr = -1;

	switch( cellFsReaddir( iFileDescriptor, &xDirEntry, &xRead ) )
	{
	case CELL_FS_SUCCEEDED:
		iErr = 1;
		break;
#ifdef ASURA_ASSERT
	case CELL_FS_EBADF:
		Asura_Assertf( false, ("Asura_PS3_FileSystem::ReadDir -> File descriptor( %d ) is invalid", iFileDescriptor) );
		break;
	case CELL_FS_EIO:
		Asura_Assert( false, ("Asura_PS3_FileSystem::ReadDir -> I/O error has occurred") );
		break;
	case CELL_FS_ENOMEM:
		Asura_Assert( false, ("Asura_PS3_FileSystem::ReadDir -> Memory is insufficient") );
		break;
	case CELL_FS_EFSSPECIFIC:
		Asura_Assert( false, ("Asura_PS3_FileSystem::ReadDir -> File system specific internal error has occurred") );
		break;
	default:
		Asura_Assert( false, ("Asura_PS3_FileSystem::ReadDir -> Unknown error") );
		break;
#endif
	}

	if( iErr == 1 )
	{
		if( xRead == 0 )
		{
			return 0;
		}
	}

	Asura_Mem::Copy( pxDirectoryEntry->szName, xDirEntry.d_name, xDirEntry.d_namlen + 1 );
	switch( xDirEntry.d_type )
	{
	case CELL_FS_TYPE_DIRECTORY:
		{
			pxDirectoryEntry->ucType = PS3_TOF_Directory;
			break;
		}
	case CELL_FS_TYPE_REGULAR:
		{
			pxDirectoryEntry->ucType = PS3_TOF_File;
			break;
		}
	case CELL_FS_TYPE_SYMLINK:
		{
			pxDirectoryEntry->ucType = PS3_TOF_SymbolicLink;
			break;
		}
	case CELL_FS_TYPE_UNKNOWN:
		{
			pxDirectoryEntry->ucType = PS3_TOF_Unknown;
			break;
		}
	}

	return iErr;
}
