//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_USBPad.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Input/Asura_PS3_USBPad.h"

// Asura Headers
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"

// Stdlib platform-specific.
#include <cell/padfilter.h>

//TEMP
#include <sdk_version.h>
#if CELL_SDK_VERSION < 0x090002
#define CELL_PAD_ERROR_UNSUPPORTED_GAMEPAD CELL_PAD_ERROR_UNSUPPRTED_GAMEPAD
#endif

DEBUGVAR float dbg_fAsura_PS3_Rumble_SmallMotorTheshold = 0.25f;

// Debug vars
// shaking params
float dbg_fPS3PadShakeThresholds	= 0.15f;
u_int dbg_uMinimumPS3PadShakes		= 4;	// just remote shakes
float dbg_fPS3PadShakeTimer			= 0.5f;

//rocking params
float dbg_fPS3PadRockThresholds		= 0.05f;
u_int dbg_uMinimumPS3Rocks			= 2;
float dbg_fPS3PadRockTimer			= 0.5f;


// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Asura_PS3_USBPad
//
Asura_PS3_USBPad::Asura_PS3_USBPad( u_int uControllerIndex )
#if ASURA_USE_NEW_INPUT
	: Asura_Controller( uControllerIndex, ASURA_PS3_OUTPUTID_MAX_NUMBER_OF_KEYS )
#endif // #if ASURA_USE_NEW_INPUT
{
	m_uID =  uControllerIndex;

#if ASURA_USE_NEW_INPUT
	m_pfCurrentKeys[ASURA_PS3_OUTPUTID_VOID] = 0.0f; // must set this to zero
#endif // #if ASURA_USE_NEW_INPUT
}


void Asura_PS3_USBPad::Initialise( u_int uID )
{
	m_uID = uID;

	// Set our status to DISCONNECTED initially. We'll pick up the pad in the
	// first update, if it's already there.
	m_eStatus = STATUS_DISCONNECTED;

	// SixAxis movement sensors
	m_bSensorSupported = false; 
	m_pxCellPadFilter = NULL; 

	// Rumble support
	m_bRumbleSupported = false;

	m_fShakeTimer = 0.0f;
	m_uNumberOfSensorSpikes = 0;
	m_uCurrentSensorSample = 0;

	for(u_int uSample = 0; uSample < uNUM_ACCELERATION_SAMPLES; ++uSample)
	{
		m_axSensorAccSamples[uSample].Zero();
	}

	m_fRockXTimer = 0.0f;
	m_eRockState = ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE;
	m_uNumberOfRocks = 0;
	m_bIntercepted = false;
}


void Asura_PS3_USBPad::Reset( void )
{
	// Ensure motors are off
	SetMotorSpeeds(0, 0);

    m_eStatus = STATUS_DISCONNECTED;

	// SixAxis movement sensors/Rumble
	m_bSensorSupported = false; 
	m_bRumbleSupported = false;
	m_bIntercepted = false;

	m_fShakeTimer = 0.0f;
	m_uNumberOfSensorSpikes = 0;
	m_uCurrentSensorSample = 0;

	m_fRockXTimer = 0.0f;
	m_eRockState = ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE;
	m_uNumberOfRocks = 0;

	if ( m_pxCellPadFilter )
	{
		delete m_pxCellPadFilter; 
		m_pxCellPadFilter = NULL; 	
	}

	for(u_int uSample = 0; uSample < uNUM_ACCELERATION_SAMPLES; ++uSample)
	{
		m_axSensorAccSamples[uSample].Zero();
	}
}


void Asura_PS3_USBPad::RefreshStatusWhenConnected( const Asura_PS3_PadInfo& xCurPadInfo )
{
    CellPadData xPadData;

    int iResult = cellPadGetData( m_uID, &xPadData );
    if( CELL_PAD_OK == iResult )
    {
        m_eStatus = STATUS_CONNECTED;
    }
    else if( CELL_PAD_ERROR_UNSUPPORTED_GAMEPAD == iResult )
    {
#if ASURA_PS3_USBPAD_SUPPORT_MANUAL_TRANSLATION
        if( CanManuallyTranslate(xCurPadInfo) )
        {
            m_eStatus = STATUS_CONNECTED;
        }
        else
#endif
        {
            m_eStatus = STATUS_UNSUPPORTED;
        }
    }
    else
    {
        // We don't really know so leave the status unchanged.
    }
    
    // Only query pad if we're sure it's connected
    if ( m_eStatus == STATUS_CONNECTED )
    {
		#if CELL_SDK_VERSION < 0x250000

			CellPadCapabilityInfo xPadCaps;
			cellPadGetCapabilityInfo(m_uID, &xPadCaps);

			// Dump pad capabilities in debug mode
			#if defined _DEBUG
			
				if ( xPadCaps.info[0] & CELL_PAD_CAPABILITY_PS3_CONFORMITY ) { Asura_OutputToDebugger(("USB pad: Pad %d is PS3 conformant\n", m_uID)); }
				if ( xPadCaps.info[0] & CELL_PAD_CAPABILITY_PRESS_MODE ) { Asura_OutputToDebugger(("USB pad: Pad %d supports Press mode (buttons)\n", m_uID)); }
				if ( xPadCaps.info[0] & CELL_PAD_CAPABILITY_SENSOR_MODE ) { Asura_OutputToDebugger(("USB pad: Pad %d supports Sensor mode (tilt)\n", m_uID)); }
				if ( xPadCaps.info[0] & CELL_PAD_CAPABILITY_HP_ANALOG_STICK ) { Asura_OutputToDebugger(("USB pad: Pad %d is High Precision\n", m_uID)); }
				if ( xPadCaps.info[0] & CELL_PAD_CAPABILITY_ACTUATOR ) { Asura_OutputToDebugger(("USB pad: Pad %d has Actuators\n", m_uID)); }
				
			#endif

			m_bRumbleSupported = (xPadCaps.info[0] & CELL_PAD_CAPABILITY_ACTUATOR) != 0;
			m_bSensorSupported = (xPadCaps.info[0] & CELL_PAD_CAPABILITY_SENSOR_MODE) != 0 ; 

		#else

			// Dump pad capabilities in debug mode
			#if defined _DEBUG

				if ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_PS3_CONFORMITY	) { Asura_OutputToDebugger(("USB pad: Pad %d is PS3 conformant\n"				, m_uID ) ); }
				if ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_PRESS_MODE		) { Asura_OutputToDebugger(("USB pad: Pad %d supports Press mode (buttons)\n"	, m_uID ) ); }
				if ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_SENSOR_MODE		) { Asura_OutputToDebugger(("USB pad: Pad %d supports Sensor mode (tilt)\n"		, m_uID ) ); }
				if ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_HP_ANALOG_STICK	) { Asura_OutputToDebugger(("USB pad: Pad %d is High Precision\n"				, m_uID ) ); }
				if ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_ACTUATOR		) { Asura_OutputToDebugger(("USB pad: Pad %d has Actuators\n"					, m_uID ) ); }

			#endif

			m_bRumbleSupported = ( ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_ACTUATOR    ) != 0 );
			m_bSensorSupported = ( ( xCurPadInfo.device_capability[m_uID] & CELL_PAD_CAPABILITY_SENSOR_MODE ) != 0 );

		#endif

		// Check for sensor capabilities
		if( m_bSensorSupported )
		{
			// Activate sensor output once we know is supported 
			if( !SetSensorModeOn() )
			{
				// Something went wrong
				m_bSensorSupported = false; 
			}
		}
	}
}

void Asura_PS3_USBPad::Update( const Asura_PS3_PadInfo& xCurPadInfo, u_char* pucButtonOutput, float* pfAxisOutput, float* pfSensorOutput )
{
    bool bCurrentlyConnected = IsPadConnected( xCurPadInfo );
    bool bWasConnected = ( STATUS_DISCONNECTED != m_eStatus );
	m_bIntercepted = false;

    // Deal with the pad having been connected / disconnected.
    if( bCurrentlyConnected != bWasConnected )
    {
        if( bCurrentlyConnected )
        {
            Asura_OutputToDebugger( ( "USB pad: Pad %u connected.\n", m_uID ) );

            // Reset status to disconnected first so that if the status
            // refresh fails then it stays that way.
            m_eStatus = STATUS_DISCONNECTED;

            RefreshStatusWhenConnected( xCurPadInfo );

            if( STATUS_CONNECTED == m_eStatus )
            {
                ResetButtonAndAxisStates();
				ResetSensorStates();
            }
            else if( STATUS_UNSUPPORTED == m_eStatus )
            {
                Asura_OutputToDebugger(( "USB pad: ERROR! Pad is not supported.\n" ));
            }
        }
        else
        {
            Asura_OutputToDebugger(( "USB pad: Pad %u disconnected.\n", m_uID ));
            m_eStatus = STATUS_DISCONNECTED;
        }
    }

    // Check if we have anything to do.
    if( STATUS_CONNECTED != m_eStatus )
    {
		ResetButtonAndAxisStates();
		ResetSensorStates();
        return;
    }

	m_bIntercepted = IsIntercepted( xCurPadInfo );
	

	// Otherwise we just need the latest input, if there is any.
	{
		CellPadData xPadData;
		bool bGotData = false;

		int iReturnValue = cellPadGetData( m_uID, &xPadData );
		if( CELL_PAD_OK == iReturnValue )
		{
			bGotData = ( xPadData.len > 0 );
//			Asura_DebugText::AddTextBlock(6, 24, 0xFFFFFFFF, 0, "CELL PAD OK");
		}
#if ASURA_PS3_USBPAD_SUPPORT_MANUAL_TRANSLATION
		else if( CELL_PAD_ERROR_UNSUPPORTED_GAMEPAD == iReturnValue )
		{
			bGotData = ManuallyTranslate( xCurPadInfo, &xPadData );
		}
#endif

		if( bGotData )
		{
//			Asura_DebugText::AddTextBlock(6, 23, 0xFFFFFFFF, 0, "GOT DATA");
			UpdateButtonAndAxisStates( xPadData );
			if( m_bSensorSupported )
			{
				UpdateSensorStates( xPadData ); 

				UpdateSensorActions( xPadData);
			}
		}
		else if( CELL_PAD_OK != iReturnValue )
		{
			ResetButtonAndAxisStates();
			ResetSensorStates();
		}
	}

//	Asura_DebugText::AddTextBlock(6, 22, 0xFFFFFFFF, 0, "ROCKS %d", m_uNumberOfRocks);

	// Copy button states to output buffers	(if requested)
	if ( pucButtonOutput != NULL )
	{
		ASURA_STATIC_ASSERT( (sizeof(m_aucButtonStates)/sizeof(m_aucButtonStates[0])) <= (ASURA_KEYID_PAD2_BUTTON1 - ASURA_KEYID_PAD1_BUTTON1) );
		Asura_Mem::Copy(pucButtonOutput, m_aucButtonStates, sizeof(m_aucButtonStates));
	}
	// Copy axis states to output buffers (if requested)
	if ( pfAxisOutput != NULL )
	{
		ASURA_STATIC_ASSERT( (sizeof(m_afAxisStates)/sizeof(m_afAxisStates[0])) <= (ASURA_AXIS_PAD2_LSTICK_H - ASURA_AXIS_PAD1_LSTICK_H) );
		Asura_Mem::Copy(pfAxisOutput, m_afAxisStates, sizeof(m_afAxisStates));
	}
	// Copy sensor states to output buffers (if requested)
	if ( pfSensorOutput != NULL )
	{
		Asura_Assert(false, "Unbounded and untested sensor copy - needs implementing\n");
		Asura_Mem::Copy(pfSensorOutput, m_afSensorStates, sizeof(m_afSensorStates));
	}
	
}



void Asura_PS3_USBPad::UpdateButtonAndAxisStates( const CellPadData& xPadData )
{
    // The headers don't currently define any useful values for this stuff
    // so it's all hardcoded based on the reference documentation for
    // cellPadGetData().

    #define GET_BUTTON_VALUE( byte_idx, bit_idx ) \
        ((xPadData.button[byte_idx] & (1 << bit_idx)) ? 0xFF : 0)
    
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_CROSS]    = GET_BUTTON_VALUE( 3, 6 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_CIRCLE]   = GET_BUTTON_VALUE( 3, 5 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_SQUARE]   = GET_BUTTON_VALUE( 3, 7 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_TRIANGLE] = GET_BUTTON_VALUE( 3, 4 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_R2]       = GET_BUTTON_VALUE( 3, 1 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_L2]       = GET_BUTTON_VALUE( 3, 0 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_START]    = GET_BUTTON_VALUE( 2, 3 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_SELECT]   = GET_BUTTON_VALUE( 2, 0 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_L1]       = GET_BUTTON_VALUE( 3, 2 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_R1]       = GET_BUTTON_VALUE( 3, 3 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_L3]       = GET_BUTTON_VALUE( 2, 1 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_R3]       = GET_BUTTON_VALUE( 2, 2 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_UP]       = GET_BUTTON_VALUE( 2, 4 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_DOWN]     = GET_BUTTON_VALUE( 2, 6 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_LEFT]     = GET_BUTTON_VALUE( 2, 7 );
    m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_RIGHT]    = GET_BUTTON_VALUE( 2, 5 );

    #undef GET_BUTTON_VALUE

    m_afAxisStates[ASURA_PS3_PADAXISOFFSET_LSTICK_H] = ConvertToAsuraAxis( xPadData.button[CELL_PAD_BTN_OFFSET_ANALOG_LEFT_X], false );
    m_afAxisStates[ASURA_PS3_PADAXISOFFSET_LSTICK_V] = ConvertToAsuraAxis( xPadData.button[CELL_PAD_BTN_OFFSET_ANALOG_LEFT_Y], true  );
    m_afAxisStates[ASURA_PS3_PADAXISOFFSET_RSTICK_H] = ConvertToAsuraAxis( xPadData.button[CELL_PAD_BTN_OFFSET_ANALOG_RIGHT_X], false );
    m_afAxisStates[ASURA_PS3_PADAXISOFFSET_RSTICK_V] = ConvertToAsuraAxis( xPadData.button[CELL_PAD_BTN_OFFSET_ANALOG_RIGHT_Y], false  );

#if ASURA_USE_NEW_INPUT
	m_pfCurrentKeys[ASURA_PS3_KEYID_CROSS]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_CROSS]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_CIRCLE]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_CIRCLE]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_SQUARE]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_SQUARE]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_TRIANGLE]	= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_TRIANGLE]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_R2]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_R2]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_L2]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_L2]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_START]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_START]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_SELECT]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_SELECT]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_L1]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_L1]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_R1]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_R1]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_L3]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_L3]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_R3]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_R3]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_UP]			= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_UP]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_DOWN]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_DOWN]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_LEFT]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_LEFT]) / 255.0f; 
	m_pfCurrentKeys[ASURA_PS3_KEYID_RIGHT]		= static_cast<float>(m_aucButtonStates[ASURA_PS3_PADBUTTONOFFSET_RIGHT]) / 255.0f; 

	m_pfCurrentKeys[ASURA_PS3_AXISID_LSTICK_H]	= m_afAxisStates[ASURA_PS3_PADAXISOFFSET_LSTICK_H]; 
	m_pfCurrentKeys[ASURA_PS3_AXISID_LSTICK_V]	= m_afAxisStates[ASURA_PS3_PADAXISOFFSET_LSTICK_V]; 
	m_pfCurrentKeys[ASURA_PS3_AXISID_RSTICK_H]	= m_afAxisStates[ASURA_PS3_PADAXISOFFSET_RSTICK_H]; 
	m_pfCurrentKeys[ASURA_PS3_AXISID_RSTICK_V]	= m_afAxisStates[ASURA_PS3_PADAXISOFFSET_RSTICK_V]; 
#endif // #if ASURA_USE_NEW_INPUT

}

void Asura_PS3_USBPad::ResetButtonAndAxisStates( void )
{
    for( u_int u = 0; u < ASURA_PS3_PADBUTTONOFFSET_NEXT_PAD; ++u )
    {
        m_aucButtonStates[u] = 0;
    }

    for( u_int u = 0; u < ASURA_PS3_PADAXISOFFSET_NEXT_PAD; ++u )
    {
        m_afAxisStates[u] = 0.0f;
    }
}
void Asura_PS3_USBPad::UpdateSensorStates( const CellPadData& xPadData )
{
	u_int uSensorOutputAxisX = xPadData.button[CELL_PAD_BTN_OFFSET_SENSOR_X]; 
	u_int uSensorOutputAxisY = xPadData.button[CELL_PAD_BTN_OFFSET_SENSOR_Y]; 
	u_int uSensorOutputAxisZ = xPadData.button[CELL_PAD_BTN_OFFSET_SENSOR_Z]; 
	u_int uSensorOutputGyro	= xPadData.button[CELL_PAD_BTN_OFFSET_SENSOR_G];

	// Filter the data if requested
	if( m_pxCellPadFilter )
	{
		uSensorOutputAxisX = cellPadFilterIIRFilter( m_pxCellPadFilter, uSensorOutputAxisX ); 
		uSensorOutputAxisY = cellPadFilterIIRFilter( m_pxCellPadFilter, uSensorOutputAxisY ); 
		uSensorOutputAxisZ = cellPadFilterIIRFilter( m_pxCellPadFilter, uSensorOutputAxisZ ); 
		uSensorOutputGyro = cellPadFilterIIRFilter( m_pxCellPadFilter, uSensorOutputGyro ); 
	}

	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_X]		= ConvertToAsuraAccelerometer( uSensorOutputAxisX ); 
	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Y]		= ConvertToAsuraAccelerometer( uSensorOutputAxisY ); 
	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Z]		= ConvertToAsuraAccelerometer( uSensorOutputAxisZ ); 
	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_GYRO_YAW]	= ConvertToAsuraGyroscope( uSensorOutputGyro ); 

#if ASURA_USE_NEW_INPUT
	m_pfCurrentKeys[ASURA_PS3_SENSORID_AXIS_X]				= m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_X]; 
	m_pfCurrentKeys[ASURA_PS3_SENSORID_AXIS_Y]				= m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Y]; 
	m_pfCurrentKeys[ASURA_PS3_SENSORID_AXIS_Z]				= m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Z]; 
	m_pfCurrentKeys[ASURA_PS3_SENSORID_GYRO_YAW]			= m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_GYRO_YAW]; 
#endif // #if ASURA_USE_NEW_INPUT

//#ifdef ASURA_USE_DEBUG_TEXT
//	Asura_DebugText::AddTextBlock(6, 17, 0xFFFFFFFF, 0, "Raw X:  %4.4f\t%05d",	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_X], uSensorOutputAxisX);
//	Asura_DebugText::AddTextBlock(6, 18, 0xFFFFFFFF, 0, "Raw Y:  %4.4f\t%05d",	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Y], uSensorOutputAxisY);
//	Asura_DebugText::AddTextBlock(6, 19, 0xFFFFFFFF, 0, "Raw Z:  %4.4f\t%05d",	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Z], uSensorOutputAxisZ);
//	Asura_DebugText::AddTextBlock(6, 20, 0xFFFFFFFF, 0, "Raw Yaw: %4.4f\t%05d",	m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_GYRO_YAW], uSensorOutputGyro);
//#endif // #ifdef ASURA_USE_DEBUG_TEXT
}

void Asura_PS3_USBPad::UpdateSensorActions( const CellPadData& xPadData)
{

	m_axSensorAccSamples[m_uCurrentSensorSample].x = m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_X]; 
	m_axSensorAccSamples[m_uCurrentSensorSample].y = m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Y]; 
	m_axSensorAccSamples[m_uCurrentSensorSample].z = m_afSensorStates[ASURA_PS3_PADSENSOROFFSET_AXIS_Z]; 

	const u_int uPrevSample = (m_uCurrentSensorSample - 1) % uNUM_ACCELERATION_SAMPLES;
	const Asura_Vector_3 xCurrGradient	= m_axSensorAccSamples[m_uCurrentSensorSample] - m_axSensorAccSamples[uPrevSample];

	for(u_int uAxis = 0; uAxis < 3; ++uAxis)
	{
		// Shaking
		if( (m_uNumberOfSensorSpikes) && (m_fShakeTimer > 0.f)) // at least one spike was detected and in time
		{
			const int iRemoteSpike = DetectSensorSpike(m_xShakeSpikeDetection, xCurrGradient, uAxis, dbg_fPS3PadShakeThresholds, dbg_fPS3PadShakeThresholds);
			if ( iRemoteSpike ) // spike detected, increase spike counter and reset timer
			{ 
				m_uNumberOfSensorSpikes++;
				m_fShakeTimer = dbg_fPS3PadShakeTimer;
			}
			else // decrease timer
			{
				m_fShakeTimer -= Asura_Timers::GetFrameTime();
				if (m_fShakeTimer < 0.f) { m_fShakeTimer = 0.f; }
			}
		}
		else if ((m_fShakeTimer == 0.f) && (m_uNumberOfSensorSpikes)) // timer = 0 and at least one spike had been detected, meaning this has timed out, reset spike counter
		{
			m_uNumberOfSensorSpikes = 0;
//			if (dbg_bShowShakeDebugInfo) { Asura_OutputToDebugger(("Remote Timeout\n")); }
		}
		else // detecting first spike
		{
			const int iRemoteSpike = DetectSensorSpike(m_xShakeSpikeDetection, xCurrGradient, uAxis, dbg_fPS3PadShakeThresholds, dbg_fPS3PadShakeThresholds);
			if ( iRemoteSpike ) // spike detected, increase spike counter and reset timer
			{ 
				m_uNumberOfSensorSpikes++;				
				m_fShakeTimer = dbg_fPS3PadShakeTimer;
			}
		}

		if( uAxis == 0)	// currently only checking rocking on x-axis
		{
			if( (m_uNumberOfRocks) && (m_fRockXTimer > 0.0f)) // at least one rock was detected and in time
			{
				const u_int uRock = DetectRocking( m_axSensorAccSamples[m_uCurrentSensorSample].x, m_eRockState, dbg_fPS3PadRockThresholds, -dbg_fPS3PadRockThresholds);
				if ( uRock) // rock detected
				{ 
					m_uNumberOfRocks++;
					m_fRockXTimer = dbg_fPS3PadRockTimer;
				}
				else // decrease timer
				{
					m_fRockXTimer -= Asura_Timers::GetFrameTime();
					if (m_fRockXTimer < 0.f) { m_fRockXTimer = 0.f; }
				}
			}
			else if ((m_fRockXTimer == 0.0f) && (m_uNumberOfRocks)) // timer = 0 and at least one spike had been detected, meaning this has timed out, reset spike counter
			{
				m_uNumberOfRocks = 0;
				m_eRockState = ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE;
			}
			else // detecting first spike
			{
				const u_int uRock = DetectRocking( m_axSensorAccSamples[m_uCurrentSensorSample].x, m_eRockState, dbg_fPS3PadRockThresholds, -dbg_fPS3PadRockThresholds);
				if ( uRock)
				{ 
					m_uNumberOfRocks++;				
					m_fRockXTimer = dbg_fPS3PadRockTimer;
				}
			}

		}	
		
	}

	if( m_uNumberOfRocks >= dbg_uMinimumPS3Rocks)
	{
		#if ASURA_USE_NEW_INPUT
		m_pfCurrentKeys[ASURA_PS3_ACTION_ROCKING_X] = fASURA_CONTROLLER_INPUT_MAX;
		#endif

		m_uNumberOfRocks = 0;
		m_eRockState = ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE;
		m_fRockXTimer = 0.0f;

//		Asura_DebugText::AddTextBlock(6, 22, 0xFFFFFFFF, 0, "ROCKS %d", m_uNumberOfRocks);
	}
	else
	{
		#if ASURA_USE_NEW_INPUT
		m_pfCurrentKeys[ASURA_PS3_ACTION_ROCKING_X] = 0.0f;
		#endif
	}

	if( m_uNumberOfSensorSpikes >= dbg_uMinimumPS3PadShakes)
	{
		#if ASURA_USE_NEW_INPUT
		m_pfCurrentKeys[ASURA_PS3_ACTION_SHAKE] = fASURA_CONTROLLER_INPUT_MAX;
		#endif

		m_uNumberOfSensorSpikes = 0;
	}
	else
	{
		#if ASURA_USE_NEW_INPUT
		m_pfCurrentKeys[ASURA_PS3_ACTION_SHAKE] = 0.0f;
		#endif
	}

//	Asura_DebugText::AddTextBlock(6, 22, 0xFFFFFFFF, 0, "ROCKS %d", m_uNumberOfRocks);

	if(++m_uCurrentSensorSample >= uNUM_ACCELERATION_SAMPLES)
	{
		m_uCurrentSensorSample = 0;
	}

//	if( m_pfCurrentKeys[ASURA_PS3_SENSORID_AXIS_X] < -0.1f || m_pfCurrentKeys[ASURA_PS3_SENSORID_AXIS_X] > 0.1f)
//	{
//		if( m_pfCurrentKeys[ASURA_PS3_ACTION_SHAKE] >= fASURA_CONTROLLER_INPUT_MAX)
//		{
//			m_pfCurrentKeys[ASURA_PS3_ACTION_SHAKE] = 0.0f;
//		}
//		else
//		{
//			m_pfCurrentKeys[ASURA_PS3_ACTION_SHAKE] = fASURA_CONTROLLER_INPUT_MAX;
//		}
//
//#ifdef ASURA_USE_DEBUG_TEXT
//		Asura_DebugText::AddTextBlock(6, 22, 0xFFFFFFFF, 0, "FAKE SHAKE");
//#endif
}

int Asura_PS3_USBPad::DetectSensorSpike( Asura_Vector_3& xSpikeDetection, const Asura_Vector_3& xGradient, u_int uAxis, float fPosSpikeThreshold, float fNegSpikeThreshold)
{
//	Asura_OutputToDebugger( ("Axis %d : Gradiant Val: %4.4f\n", uAxis, xGradient[uAxis]));

	if( xSpikeDetection[uAxis] != 0.f)
	{
		if(xSpikeDetection[uAxis] * xGradient[uAxis] < 0.f)
		{
			// Hit a peak.
			if (Asura_Maths::Sign(xSpikeDetection[uAxis]) > 0.f)
			{
				xSpikeDetection[uAxis] = 0.f;				
				return 1;
			}
			else
			{
				xSpikeDetection[uAxis] = 0.f;				
				return -1;
			}
		}
	}
	else if (((xGradient[uAxis] > 0.f) && (Asura_Maths::Abs(xGradient[uAxis]) > fPosSpikeThreshold))  // positive gradient greater than positive threshold
		|| ((xGradient[uAxis] < 0.f) && (Asura_Maths::Abs(xGradient[uAxis]) > fNegSpikeThreshold))) // negative gradient greater than negative threshold
	{
		xSpikeDetection[uAxis] = Asura_Maths::Sign(xGradient[uAxis]);
	}

	return 0;
}

u_int Asura_PS3_USBPad::DetectRocking( float fCurrentTilt, ASURA_PS3_USB_SENSOR_ROCK_STATE& eRockState, float fPosThreshold, float fNegThreshold)
{
	int iRet = 0;
	ASURA_PS3_USB_SENSOR_ROCK_STATE	ePrevRockState = eRockState;

	switch( eRockState)
	{
		case ASURA_PS3_USB_SENSOR_ROCKSTATE_NEG:
			eRockState = (fCurrentTilt > fPosThreshold) ? ASURA_PS3_USB_SENSOR_ROCKSTATE_POS : eRockState;	
			break;

		case ASURA_PS3_USB_SENSOR_ROCKSTATE_POS:
			eRockState = (fCurrentTilt < fNegThreshold) ? ASURA_PS3_USB_SENSOR_ROCKSTATE_NEG : eRockState;	
			break;

		case ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE:
		default:
			eRockState = (fCurrentTilt < fNegThreshold) ? ASURA_PS3_USB_SENSOR_ROCKSTATE_NEG : ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE;
			eRockState = (fCurrentTilt > fPosThreshold) ? ASURA_PS3_USB_SENSOR_ROCKSTATE_POS : ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE;	
			break;
	}

	if( (eRockState != ePrevRockState) && (eRockState != ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE))// && (ePrevRockState != ASURA_PS3_USB_SENSOR_ROCKSTATE_NONE))
	{
		iRet = 1;
	}


	return iRet;
}


void Asura_PS3_USBPad::ResetSensorStates( void )
{
	for( u_int u = 0; u < ASURA_PS3_PADSENSOROFFSET_NEXT_PAD; ++u )
    {
        m_afSensorStates[u] = 0.0f;
    }

	for(u_int uSample = 0; uSample < uNUM_ACCELERATION_SAMPLES; ++uSample)
	{
		m_axSensorAccSamples[uSample].Zero();
	}
}

#if ASURA_PS3_USBPAD_SUPPORT_MANUAL_TRANSLATION

bool Asura_PS3_USBPad::CanManuallyTranslate( const Asura_PS3_PadInfo& xCurPadInfo )
{
    // This needs to be kept in sync with ManuallyTranslate() below.


    if( xCurPadInfo.vendor_id[m_uID] == 3090 && xCurPadInfo.product_id[m_uID] == 5 )
    {
        return true;
    }

    return false;
}

bool Asura_PS3_USBPad::ManuallyTranslate( const Asura_PS3_PadInfo& xCurPadInfo, CellPadData* pxDataOut )
{
    // This needs to be kept in sync with CanManuallyTranslate() above.

    if( xCurPadInfo.vendor_id[m_uID] == 3090 && xCurPadInfo.product_id[m_uID] == 5 )
    {
        // Sony Dualshock 2.
        // (MP - 06.02.2006) This may be specific to it being connected via
        // a specific USB adapter. If that's the case, we should just need
        // to add the other IDs to this if() case.
        if( CELL_PAD_OK == cellPadGetRawData(m_uID, pxDataOut) && pxDataOut->len > 0 )
        {
            u_char aucUntranslated[8];
            memcpy( aucUntranslated, pxDataOut->button, 8 * sizeof(u_char) );

            // Hardcoded values that the cellPadGetData() function claims
            // to return.
            pxDataOut->button[0] = 0;
            pxDataOut->button[1] = 0x74;

            // Swap some bytes and nybbles around.
            pxDataOut->button[2] = aucUntranslated[5] >> 4; // ST, R3, L3, SEL
            pxDataOut->button[3] = (aucUntranslated[4] & 0xF0) | (aucUntranslated[5] & 0x0F); // SQ, X, O, TRI, R1, L1, R2, L2
            pxDataOut->button[4] = aucUntranslated[3]; // RSTICK_H
            pxDataOut->button[5] = aucUntranslated[2]; // RSTICK_V
            pxDataOut->button[6] = aucUntranslated[0]; // LSTICK_H
            pxDataOut->button[7] = aucUntranslated[1]; // LSTICK_V

            // Constructing the d-pad nybble is a bit more hassle...
            switch( aucUntranslated[4] & 0x0F )
            {
                case 0: pxDataOut->button[2] |= 0x00; break; // None (centred).
                case 1: pxDataOut->button[2] |= 0x10; break; // Up.
                case 2: pxDataOut->button[2] |= 0x30; break; // Up and right.
                case 3: pxDataOut->button[2] |= 0x20; break; // Right.
                case 4: pxDataOut->button[2] |= 0x60; break; // Down and right.
                case 5: pxDataOut->button[2] |= 0x40; break; // Down.
                case 6: pxDataOut->button[2] |= 0xC0; break; // Down and left.
                case 7: pxDataOut->button[2] |= 0x80; break; // Left.
                case 8: pxDataOut->button[2] |= 0x90; break; // Up and left.
            }

            return true;
        }
    }
    else
    {
        Asura_Assert( false, "Pad should be set as unsupported in this case and we shouldn't be in here." );
    }

    return false;
}

#endif // #if ASURA_PS3_USBPAD_SUPPORT_MANUAL_TRANSLATION

float Asura_PS3_USBPad::ConvertToAsuraAxis( int iPS3Value, bool bIsVerticalAxis )
{
    // Using (an equivalent to) the PS2 deadzones for now...
    const float fDEADZONE = 0.25f;

    // Convert to [-1, 1] range.
    float fNormVal = static_cast<float>(iPS3Value);
    fNormVal /= 127.5f;
    fNormVal -= 1.0f;

    // Apply deadzone and clamp.
    float fAsuraValue = (Asura_Maths::Abs(fNormVal) - fDEADZONE) * (1.0f / (1.0f - fDEADZONE));
    fAsuraValue = Asura_Maths::Max( fAsuraValue, 0.0f );
    fAsuraValue = Asura_Maths::Min( fAsuraValue, 1.0f );
    fAsuraValue *= Asura_Maths::Sign( fNormVal );

    // For historical reasons, the vertical axes are the opposite way round
    // from that reported by the pad.
    if( bIsVerticalAxis )
    {
        fAsuraValue = -fAsuraValue;
    }

    return fAsuraValue;
}


float Asura_PS3_USBPad::ConvertToAsuraAccelerometer( int iPS3Value )
{
	const float fDEADZONE = 0.015f;

	// Scale into range -1.0f .. 1.0f
	float fNormVal( iPS3Value - 512 );
	fNormVal /= 512.0f;
	
    // Apply deadzone and clamp.
    float fAsuraValue = (Asura_Maths::Abs(fNormVal) - fDEADZONE) / (1.0f - fDEADZONE);
	fAsuraValue = Asura_Maths::Max( fAsuraValue, 0.0f );
	fAsuraValue = Asura_Maths::Min( fAsuraValue, 1.0f );
	fAsuraValue *= Asura_Maths::Sign( fNormVal );

	return fAsuraValue;
}

float Asura_PS3_USBPad::ConvertToAsuraGyroscope( int iPS3Value )
{
	// TODO: no deadzone for the gyroscope?
	return ConvertToAsuraAccelerometer( iPS3Value );
}

void Asura_PS3_USBPad::SetSensorFilterMode( const ASURA_PS3_USBPAD_SENSOR_FILTER_MODE uSensorFilterMode )
{
	// Are we already filtering the data
	if( m_pxCellPadFilter )
	{
		if( ASURA_PS3_USBPAD_SENSORFILTER_NO_FILTER == uSensorFilterMode ) 
		{
			delete m_pxCellPadFilter; 
			m_pxCellPadFilter = NULL; 
		}
	}
	else
	{
		if ( ASURA_PS3_USBPAD_SENSORFILTER_NO_FILTER != uSensorFilterMode )
		{
			m_pxCellPadFilter = new CellPadFilterIIRSos; 
			if ( ASURA_PS3_USBPAD_SENSORFILTER_50_PERCENT == uSensorFilterMode )
			{
				cellPadFilterIIRInit( m_pxCellPadFilter, CELL_PADFILTER_IIR_CUTOFF_2ND_LPF_BT_050 ); 
			}
			else if ( ASURA_PS3_USBPAD_SENSORFILTER_20_PERCENT == uSensorFilterMode )
			{
				cellPadFilterIIRInit( m_pxCellPadFilter, CELL_PADFILTER_IIR_CUTOFF_2ND_LPF_BT_020 ); 
			}
			else if ( ASURA_PS3_USBPAD_SENSORFILTER_10_PERCENT == uSensorFilterMode )
			{
				cellPadFilterIIRInit( m_pxCellPadFilter, CELL_PADFILTER_IIR_CUTOFF_2ND_LPF_BT_010 ); 
			}
			else
			{
				Asura_Assert( false, "Undefined sensor filter mode. Setting filter mode on, cutoff frequency to 20%" ); 
				cellPadFilterIIRInit( m_pxCellPadFilter, CELL_PADFILTER_IIR_CUTOFF_2ND_LPF_BT_020 ); 
			}
		}
	}
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_USBPad::SetMotorSpeeds
// Description : Set the actuator motor speeds (if supported)
//
// Parameters  : 
//   float fSmall  - Speed for small motor (0 .. 1.0f)
//   float fLarge  - Speed for large motor (0 .. 1.0f)
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_USBPad::SetMotorSpeeds(float fSmall, float fLarge)
{
	// Only set if supported by controller and controller is connected
	if ( m_bRumbleSupported && (m_eStatus == STATUS_CONNECTED) )
	{
		Asura_Assert( (fSmall >= 0.0f) && (fSmall <= 1.0f) && (fLarge >= 0.0f) && (fLarge <= 1.0f), "Given rumble motor speeds are out of range");

		CellPadActParam xParams;
		Asura_Mem::Set( &xParams, 0, sizeof(xParams) );				 // SDK PDF suggests zeroing out the padded area

		xParams.motor[0] = (fSmall > dbg_fAsura_PS3_Rumble_SmallMotorTheshold)?1:0;	// turn on small motor if over theshold level (we can't set speed)
		xParams.motor[1] = static_cast<uint8_t>(255.0f * fLarge);    // Set large motor speed
		
		cellPadSetActDirect(m_uID, &xParams);
	}
}


bool Asura_PS3_USBPad::GetAllPadInfo ( Asura_PS3_PadInfo& xPadInfo )
{
	#if CELL_SDK_VERSION < 0x250000
	
		return ( cellPadGetInfo( &xPadInfo ) == CELL_PAD_OK );
	
	#else

		return ( cellPadGetInfo2( &xPadInfo ) == CELL_PAD_OK );

	#endif
}

bool Asura_PS3_USBPad::IsPadConnected ( const Asura_PS3_PadInfo& xInfo ) const
{
	#if CELL_SDK_VERSION < 0x250000

		return ( CELL_PAD_STATUS_CONNECTED == xInfo.status[m_uID] );

	#else

		return ( CELL_PAD_STATUS_CONNECTED == xInfo.port_status[m_uID] );

	#endif
}

bool Asura_PS3_USBPad::IsIntercepted ( const Asura_PS3_PadInfo& xInfo ) const
{
	#if CELL_SDK_VERSION < 0x250000

		return ( xInfo.info & CELL_PAD_INFO_INTERCEPTED );

	#else

		return ( xInfo.system_info & CELL_PAD_INFO_INTERCEPTED );

	#endif
}

bool Asura_PS3_USBPad::SetSensorModeOn () const
{
	#if CELL_SDK_VERSION < 0x250000

		return ( cellPadSetSensorMode( m_uID, CELL_PAD_SENSOR_MODE_ON ) == CELL_PAD_OK ); 
		
	#else

		return ( cellPadSetPortSetting( m_uID, CELL_PAD_SETTING_SENSOR_ON ) == CELL_PAD_OK );
		
	#endif
}

//eof
