//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformInput.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Input/Asura_PlatformInput.h"

// Asura headers
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#if ASURA_USE_NEW_INPUT
	#include "Input/Asura_Input.h"
#endif // #if ASURA_USE_NEW_INPUT

// Asura PS3 headers
#include "Input/Asura_PS3_USBKeyboard.h"
#include "Input/Asura_PS3_USBPad.h"

// Stdlib platform-specific.
#include <cell/keyboard.h>
#include <cell/sysmodule.h>


// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_PS3_Input::s_bInputIntercepted;


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::Asura_PS3_Input
// Description : Default constructor - just sets default values
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_Input::Asura_PS3_Input()
{
	m_bUSBPadInitialised = false;
	m_bUSBKBInitialised = false;
	m_bLeaveKeyboardsForPSGL = false;
	
	// Clear arrays
	Asura_Mem::Set(m_apxPads, 0, sizeof(m_apxPads));
	Asura_Mem::Set(m_apxKeyboards, 0, sizeof(m_apxKeyboards));
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::Initialise
// Description : Initialise the requested parts of PS3 input
//
// Parameters  : 
//   bool bPads  - Set true to initialise pads
//   bool bKeyboards - Set true to initialise keyboards
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Input::Initialise(bool bPads, bool bKeyboards)
{
	// Load the libio PRX.
#ifdef ASURA_ASSERT
	int iResult = 
#endif // #ifdef ASURA_ASSERT
		cellSysmoduleLoadModule( CELL_SYSMODULE_IO );
	Asura_Assert( iResult == CELL_OK, "[Asura_PS3_Input::Initialise] - PRX for libio potentially already loaded" );

	if ( bPads )
	{
		m_bUSBPadInitialised = InitialiseUSBPads();
	}
	
	if ( bKeyboards )
	{
#if 0 //TODO
//    const char* szGLRenderer = reinterpret_cast<const char*>( glGetString(GL_RENDERER) );
//    g_bLeaveKeyboardsForPSGL = ( NULL != strstr(szGLRenderer, "HUD") );
#endif
		m_bLeaveKeyboardsForPSGL = false;
		m_bUSBKBInitialised = InitialiseUSBKeyboards();
	}
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::Shutdown
// Description : Shutdown PS3 input - uninitialises whatever we started up
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Input::Shutdown()
{
	// Shutdown controller pads
	if( m_bUSBPadInitialised )
	{
		cellPadEnd();
		m_bUSBPadInitialised = false;

		for( int iPadID = 0; iPadID < GetMaxControllers(); iPadID++ )
		{
			if ( m_apxPads[iPadID] != NULL )
			{
				delete m_apxPads[iPadID];
				m_apxPads[iPadID] = NULL;
			}
		}
    }

	// Shutdown keyboards
	if( m_bUSBKBInitialised )
	{
		cellKbEnd();
		m_bUSBKBInitialised = false;

		for( int iKBID = 0; iKBID < GetMaxKeyboards(); ++iKBID )
		{
			if ( m_apxKeyboards[iKBID] != NULL )
			{
				delete m_apxKeyboards[iKBID];
				m_apxKeyboards[iKBID] = NULL;
			}
		}
	}

	Asura_Mem::Set(m_apxKeyboards, 0, sizeof(m_apxKeyboards));

	// Unload the libio PRX.
#ifdef ASURA_ASSERT
	int iResult = 
#endif // #ifdef ASURA_ASSERT
		cellSysmoduleUnloadModule( CELL_SYSMODULE_IO );
	Asura_Assert( iResult == CELL_OK, "[Asura_PS3_Input::Shutdown] - PRX for libio potentially already unloaded" );
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::Reset
// Description : Reset all initialised input system components
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Input::Reset()
{
	// Reset pads if initialised
	if( m_bUSBPadInitialised )
	{
		for ( int iLoop=0; iLoop < GetMaxControllers(); iLoop++)
		{
			m_apxPads[iLoop]->Reset();
		}
	}

	// Reset keyboards if initialised
	if( m_bUSBKBInitialised )
	{
		for ( int iLoop=0; iLoop < GetMaxKeyboards(); iLoop++)
		{
			m_apxKeyboards[iLoop]->Reset();
		}
	}
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::UpdatePads
// Description : Update the currently connected active control pads
//
// Parameters  : 
//   u_char* pucButtonOutput  - Output button array
//   float* pfAxisOutput      - Output axis value array
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Input::UpdatePads(u_char* pucButtonOutput, float* pfAxisOutput)
{
	// Abort if not initialised
	if ( !m_bUSBPadInitialised ) { return; }

	// Number of allocated keys (beware overlap!)
	const u_int uKeysPerController = ASURA_KEYID_PAD2_BUTTON1 - ASURA_KEYID_PAD1_BUTTON1;
	const u_int uAxesPerController = ASURA_AXIS_PAD2_LSTICK_H - ASURA_AXIS_PAD1_LSTICK_H;
	
	// Update all known pads
	bool bInputIntercepted = false;
	Asura_PS3_PadInfo xPadInfo;
	if ( Asura_PS3_USBPad::GetAllPadInfo( xPadInfo ) )
	{
		for( u_int uPadID = 0; uPadID < ASURA_MAX_NUMBER_OF_CONTROLLER_PADS; uPadID++ )
		{
			if ( m_apxPads[uPadID] != NULL )
			{
				// Setup targets for button information copy
				u_char* pucButtons = (pucButtonOutput != NULL) ? (pucButtonOutput + ASURA_KEYID_PAD1_BUTTON1  + (uPadID * uKeysPerController)) : NULL;
				float*  pfAxes     = (pfAxisOutput != NULL) ? (pfAxisOutput + ASURA_AXIS_PAD1_LSTICK_H + (uPadID * uAxesPerController)) : NULL;

				// Update the individual pad
				m_apxPads[uPadID]->Update( xPadInfo, pucButtons, pfAxes, NULL );
				bInputIntercepted = bInputIntercepted || m_apxPads[uPadID]->IsInputIntercepted();
			}
		}
	}

	s_bInputIntercepted = bInputIntercepted;
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::UpdateKeyboards
// Description : Update the currently connected active keyboards
//
// Parameters  : 
//   u_char* pucKeyOutput                    - Output key array
//   const Asura_UnicodeChar** pwszInputBfr  - String input buffer return (pointer!)
//   u_int* puInputBfrLen                    - Input buffer length return
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Input::UpdateKeyboards(u_char* pucKeyOutput, const Asura_UnicodeChar** pwszInputBfr, u_int* puInputBfrLen)
{
	// Abort if not initialised
	if ( !m_bUSBKBInitialised ) { return; }

	// Update keyboards
	Asura_PS3_USBKeyboard::ResetInputBfr();

	if( !m_bLeaveKeyboardsForPSGL )
	{
		CellKbInfo xKBInfo;
		if ( cellKbGetInfo(&xKBInfo) == CELL_KB_OK )
		{
			for( int iKBID = 0; iKBID < GetMaxKeyboards(); ++iKBID )
			{
				m_apxKeyboards[iKBID]->Update( xKBInfo, pucKeyOutput );
			}
		}
	}

	// Retrieve the keyboard's input buffer
	Asura_PS3_USBKeyboard::GetInputBfr( pwszInputBfr, puInputBfrLen );    
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::InitialiseUSBPads
// Description : Initialise the USB pads
//
// Parameters  : none
//
// Returns     : bool - Returns true on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Input::InitialiseUSBPads()
{
	Asura_Assert( !m_bUSBPadInitialised, "[Asura_PS3_Input::InitialiseUSBPads] - Pads already initialised!\n" );
	Asura_OutputToDebugger(( "USB pad: Initialising with support for up to %i pad(s)...\n", ASURA_MAX_NUMBER_OF_CONTROLLER_PADS ));

	int iPadError = cellPadInit( GetMaxControllers() );
	if( (CELL_PAD_OK != iPadError) && (CELL_PAD_ERROR_ALREADY_INITIALIZED != iPadError) )
	{
		Asura_OutputToDebugger(( "USB pad: ERROR! cellPadInit() failed (code: 0x%08X).\n", iPadError ));
		cellPadEnd();
		return false;
	}

	for( int iPadID = 0; iPadID < GetMaxControllers(); ++iPadID )
	{
		m_apxPads[iPadID] = new Asura_PS3_USBPad( iPadID );
		m_apxPads[iPadID]->Initialise( iPadID );
	}

	Asura_OutputToDebugger(( "USB pad: Initialisation completed.\n" ));
	return true;    
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::InitialiseUSBKeyboards
// Description : Initialise the USB keyboards
//
// Parameters  : none
//
// Returns     : bool - Returns true on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Input::InitialiseUSBKeyboards()
{
	Asura_Assert( !m_bUSBKBInitialised, "[Asura_PS3_Input::InitialiseUSBKeyboards] - Keyboards already initialised!\n" );
	Asura_OutputToDebugger(( "USB keyboard: Initialising with support for %i keyboard(s)...\n", ASURA_MAX_NUMBER_OF_KEYBOARDS ));

	if( !m_bLeaveKeyboardsForPSGL )
	{
		int iKBError = cellKbInit( GetMaxKeyboards() );
		if( CELL_KB_OK != iKBError )
		{
			Asura_OutputToDebugger(( "USB keyboard: ERROR! cellKbInit() failed (code: 0x%08X).\n", iKBError ));
			cellKbEnd();
			return false;
		}
	}

	for( int iKBID = 0; iKBID < GetMaxKeyboards(); ++iKBID )
	{
		m_apxKeyboards[iKBID] = new Asura_PS3_USBKeyboard( );
		m_apxKeyboards[iKBID]->Initialise( iKBID );
	}

	Asura_OutputToDebugger(( "USB keyboard: Initialisation completed.\n" ));
	return true;    
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::IsKeyToggled
// Description : Determine the state of an individual key
//
// Parameters  : 
//   int iKeyID - ID of key to query
//
// Returns     : bool - Key state
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_Input::IsKeyToggled(int iKeyID)
{
	return false;
}


///////////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::CountActiveControllers
// Description : Count the number of connected active controllers
//
// Parameters  :
//  bool bConsecutive - Set true to count only consecutive controllers
//
// Returns     : int - Number of connected active controllers
///////////////////////////////////////////////////////////////////////////////
int Asura_PS3_Input::CountActiveControllers(bool bConsecutive) const
{
    int iCount = 0;
    for( int iPadID = 0; iPadID < GetMaxControllers(); iPadID++ )
    {
		if ( (m_apxPads[iPadID] != NULL) && m_apxPads[iPadID]->IsConnectedAndSupported() )
        {
            iCount++;
        }
        else if ( bConsecutive )
        {
            break;
        }
    }

    return iCount;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_Input::GetController
// Description : Get the controller at a specific index
//
// Parameters  : 
//   int iController - Controller index to locate
//
// Returns     : Asura_PS3_USBPad* - Pointer to valid PAD, or NULL on failure
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_USBPad* Asura_PS3_Input::GetController(int iController)
{
	if ( (iController < 0) || (iController >= GetMaxControllers()) )
	{
		Asura_Assertf(false, ("Controller index %d is out of range", iController));
		return NULL;
	}		
	
	return m_apxPads[iController];
}



// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void Asura_UserInput::Platform_Initialise( void )
{
	Asura_PS3_Input::GetAPI().Initialise(true, true);
}

void Asura_UserInput::Platform_Reset( void )
{
	Asura_PS3_Input::GetAPI().Reset();
}

void Asura_UserInput::Platform_Shutdown (void)
{
	Asura_PS3_Input::GetAPI().Shutdown();
}

void Asura_UserInput::Platform_Update( void )
{
	Asura_PS3_Input::GetAPI().UpdatePads( CurrentInput, AxisValue );
	Asura_PS3_Input::GetAPI().UpdateKeyboards( CurrentInput, &s_wszInputBuffer, &s_uInputBufferSize );
}

int Asura_UserInput::Platform_GetNumberOfActiveControllers( bool bConsecutive )
{
	return Asura_PS3_Input::GetAPI().CountActiveControllers(bConsecutive);
}

bool Asura_UserInput::Platform_IsControllerActive( int iPlayerNum )
{
	Asura_PS3_USBPad* pController = Asura_PS3_Input::GetAPI().GetController(iPlayerNum-1);
	return (pController != NULL) && pController->IsConnectedAndSupported();
}

void Asura_UserInput::Platform_UpdateInputBuffer( Asura_UnicodeChar* wszBuffer,
                                                  u_int uMaxChars,
                                                  u_int& uInputCursorPos,
                                                  bool (*pfnHandleInputOverride)(ASURA_KEYID eKey), 
                                                  const Asura_UnicodeChar* wszInputBuffer,
                                                  u_int uInputBufferSize )
{
}

bool Asura_UserInput::Platform_IsKeyToggled( int iKeyID )
{
    return Asura_PS3_Input::GetAPI().IsKeyToggled( iKeyID );
}

// NEW INPUT /////////////////////////////////////////////////////////////////

#if ASURA_USE_NEW_INPUT

void Asura_Input::Platform_Initialise( void )
{
	Asura_PS3_Input::GetAPI().Initialise(true, true);
}

void Asura_Input::Platform_Reset( void )
{
	Asura_PS3_Input::GetAPI().Reset();
}

void Asura_Input::Platform_Update( void )
{
	Asura_PS3_Input::GetAPI().UpdatePads( NULL, NULL );
	// Asura_Input doesn't support storage of keyboard keypresses (it's controller-centric)
	//Asura_PS3_Input::GetAPI().UpdateKeyboards( NULL, &s_wszInputBuffer, &s_uInputBufferSize );
}

void Asura_Input::Platform_Shutdown( void )
{
	Asura_PS3_Input::GetAPI().Shutdown();
}

#endif // #if ASURA_USE_NEW_INPUT
