//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_Main.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Stdlib independent.
#include <stdlib.h>

// Stdlib platform-specific.
#include <sys/spu_initialize.h>
#include <sys/process.h>
#include <sys/memory.h>

// Engine independent.
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_Core.h"
#ifdef ASURA_USE_DEBUG_TEXT
#include "Core/Asura_TextureManagement.h"
#endif // #ifdef ASURA_USE_DEBUG_TEXT
#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#include "Core/Asura_MaterialResponse_System.h"
#endif //#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#ifdef ASURA_SUPPORT_LOBBY
#include "Network/Online/Asura_FriendsList.h"
#endif
#ifdef ASURA_USE_AUDIO_SYSTEM
#include "Sound/Asura_Audio_Systems_Manager.h"
#endif

// Engine platform-specific.
#include "FileAccess/Asura_PS3_FileSystem.h"
#include "Network/Asura_PS3_Network.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_SwapChain.h"
#include "Sound/Asura_PS3_Sound_System.h"
#include "Sound/Asura_PS3_MultiStream.h"

// Project independent.
#include "Main/Asura_Project_Main.h"

#include <netex/libnetctl.h>
#include <sys/process.h>
#include <sysutil/sysutil_sysparam.h>

// G L O B A L S /////////////////////////////////////////////////////////////
static u_int g_uPresentIntervalBackup;	// only used in main thread.
static u_int g_uSysUtilDrawingCount;	// only used in main thread.

SYS_PROCESS_PARAM (1000, 0x10000)

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

#ifdef ASURA_PS3_SUPPORT_SNTUNER
void SNTunerSyncFunction( void );
#endif

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

static void Asura_PS3_SysUtilCallback(uint64_t status, uint64_t param, void* userdata)
{
	Asura_Assert(Asura_Core::IsMainThread(), "only expecting this callback on main thread");

	(void) param;
	(void) userdata;

	switch(status)
	{
	case CELL_SYSUTIL_REQUEST_EXITGAME:
		Asura_PS3_Graphics::SetIsInShutdown();
		break;
	case CELL_SYSUTIL_DRAWING_BEGIN:
		{
			// force max game frame rate of 60/50 fps while rendering system stuff
			Asura_Assert(g_uSysUtilDrawingCount == 0, "received more than one CELL_SYSUTIL_DRAWING_BEGIN messages");
			if (g_uSysUtilDrawingCount == 0)
			{
				g_uPresentIntervalBackup = Asura_PS3_SwapChain::GetPresentInterval();
				Asura_PS3_SwapChain::SetPresentInterval(1);
			}
			++g_uSysUtilDrawingCount;
			break;
		}
	case CELL_SYSUTIL_DRAWING_END:
		{
			// restore default game frame rate
			// it seems if the sysutil started rendering something before our app gets control (or maybe
			// we're just a bit late registering the callback) we can get one of these without getting 
			// a corresponding CELL_SYSUTIL_DRAWING_BEGIN, so test for that case here.
			if (g_uSysUtilDrawingCount > 0)
			{
				--g_uSysUtilDrawingCount;
				if (g_uSysUtilDrawingCount == 0)
				{
					Asura_PS3_SwapChain::SetPresentInterval(g_uPresentIntervalBackup);
				}
			}
		}
		break;
	case 0x0131://CELL_SYSUTIL_SYSTEM_MENU_OPEN:
		break;
	case 0x0132://CELL_SYSUTIL_SYSTEM_MENU_CLOSE
#ifdef ASURA_SUPPORT_LOBBY
		Asura_FriendsList::Refresh();
#endif
		break;
#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
	case CELL_SYSUTIL_NET_CTL_NETSTART_LOADED:
	case CELL_SYSUTIL_NET_CTL_NETSTART_FINISHED:
	case CELL_SYSUTIL_NET_CTL_NETSTART_UNLOADED:
		Asura_PS3_Network::SysUtilNetCtlEvent(status);
		break;
#endif
	default:
		Asura_OutputToDebugger( ( "Asura_PS3_SysUtilCallback: Unknown status received: 0x%X\n", status ) );
		// this may happen with codes not defined in the SDK if you just update the firmware.
	}
}


int main( int iNumArgs, const char** aszArgs )
{
	int iRet;

	// register sysutil exit callback
	iRet = cellSysutilRegisterCallback(0, Asura_PS3_SysUtilCallback, NULL);
	if( iRet != CELL_OK )
	{
		Asura_OutputToDebugger( ( "Registering sysutil callback failed...: error=0x%x\n", iRet ) );
		return iRet;
	}

	// (MH - 11.01.2008) available memory print - helps projects find a good limit for debug builds.
	sys_memory_info_t xMemInfo;
	sys_memory_get_user_memory_size( &xMemInfo );
	printf("mem free: %.02f MB (%d bytes) [total user memory %.02f MB]\n", 
		Asura_Maths::FtoD(static_cast< float >( xMemInfo.available_user_memory ) / 1048576.0f), xMemInfo.available_user_memory, 
		Asura_Maths::FtoD(static_cast< float >( xMemInfo.total_user_memory ) / 1048576.0f) );

    // first arg is num physical SPUs, second is num allowed to be RAW SPUs.
    // We need a RAW SPU for PSGL 
	sys_spu_initialize( 6, 0 );

    // (MP - 07.04.2006) Not much point in handling that failing. It should
    // only fail if we're trying to reinitialise incorrectly or if the
    // arguments are otherwise invalid (eg. requesting more SPUs than exist).
    Asura_Assert( CELL_OK == iRet, "sys_spu_initialize() failed!" );

    const char* szBootfile = ( iNumArgs > 0 ? aszArgs[0] : NULL );

    // Initialise file access
	// first see if the command line says to force BDVD access
	bool bForceDiscAccess = false;
	for (int i = 1; i < iNumArgs; ++i)
	{
		if (!strnicmp(aszArgs[i], "-forcebdvd", 10))
		{
			bForceDiscAccess = true;
			break;
		}
	}
    Asura_PS3_FileSystem::Initialise( szBootfile, bForceDiscAccess );
    
    Asura_Core::EarlyInitialise();
    Asura_Project::EarlyInitialise();

#ifdef ASURA_USE_DEBUG_TEXT
    // Do this before Asura_Core::LateInitialise() since that will call
    // Asura_DebugText::Initialise() which looks for the texture from
    // this archive.
	Asura_Texture_Management::LoadAsuraFileUsingResetGroup("graphics\\dbgfont." ASURA_STANDARD_FILE_EXTENSION, ASURA_TEXTURE_LOADED_PERMANENT );
#endif

    Asura_Core::LateInitialise();
    Asura_Project::LateInitialise();

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	Asura_MaterialResponse_System::SetGlobalResetPointHit();
#endif //#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	Asura_PS3_Graphics::SetIsInSecondaryDrawThread( false );

	while( Asura_Project::MainLoop() && !Asura_PS3_Graphics::IsInShutdown() )
	{
		Asura_PS3_Graphics::CheckExitCallBack();

#ifdef ASURA_PS3_SUPPORT_SNTUNER
        SNTunerSyncFunction();
#endif
	}

	Asura_PS3_Graphics::SetIsInShutdown();
	#ifndef ASURA_USE_AUDIO_SYSTEM
	Asura_PS3_Sound_System::GetAPI().Shutdown();
	Asura_PS3_MultiStream::GetStreamingAPI().Shutdown();
	#else	
	Asura_Audio_Systems_Manager::Shutdown();
	#endif
	Asura_PS3_Graphics::Shutdown();

	printf("Asura: exit main\n");

	sys_process_exit( EXIT_SUCCESS );

	// Return success...
	return EXIT_SUCCESS;
}


#ifdef ASURA_PS3_SUPPORT_SNTUNER
// Intended for use as SN Tuner's sync function across all projects.
void SNTunerSyncFunction( void )
{
    asm volatile ( "nop" );
}
#endif