//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_NP_Room.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     wrapper for creating/joining PSN rooms
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Online/Asura_PS3_NP_Room.h"

// Asura independent includes
#include "Core/Asura_HeapArray.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Memory/Asura_Mem.h"

// Asura PS3 includes
#include "Network/Online/Asura_PS3_AVChat2.h"
#include "Network/Online/Asura_PS3_NP.h"
#include "Network/Online/Asura_PS3_NP_Basic.h"

// library includes
#include <wchar.h>

// I N T E R N A L   T Y P E S ///////////////////////////////////////////////
class Asura_PS3_JoinRoom_CallbackData 
{
public:
	Asura_PS3_JoinRoom_CallbackData()
	{
		Reset();
	}

	void Reset()
	{ 
		m_bDone = false;
		m_bError = false;
		m_iErrorCode = 0;
	}

	bool m_bDone;
	bool m_bError;
	int m_iErrorCode;
};



// G L O B A L S /////////////////////////////////////////////////////////////
static Asura_PS3_JoinRoom_CallbackData g_xJoinRoomCallbackData;
static u_int g_uReqSeqNum;
static SceNpMatching2RequestId g_xAssignedReqId;

//Signaling callback is currently only producing debug output
static bool g_bUseSignallingCallback = false;

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
Asura_PS3_NP_Room* Asura_PS3_NP_Room::s_pxThis;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
Asura_PS3_NP_Room::Asura_PS3_NP_Room(u_int uMaxPlayers)
{
	Asura_Assert(uMaxPlayers > 0, "invalid max players");
	m_eState = ASURA_PS3_NP_ROOM_STATE_NONE;
	m_iErrorCode = 0;
	m_bIsHost = false;
	m_uMaxPlayers = uMaxPlayers;
	m_uNumPlayers = 0;
	m_pxPlayers = new SceNpId[uMaxPlayers];
	m_usHostMemberId = 0;
	m_usLocalMemberId = 0;
	m_xRoomId = 0;

	// initialise member data
	for (u_int u = 0; u < SCE_NP_MATCHING2_SESSION_PASSWORD_SIZE; ++u)
	{
		m_xPassword.data[u] = Asura_Random::GetIntU('0', 'z');
	}

	Asura_Assert(!s_pxThis, "only one instance currently supported");
	s_pxThis = this;

#ifdef ASURA_ASSERT
	int iRet = 
#endif
		sceNpMatching2RegisterRoomEventCallback(Asura_PS3_NP::GetContextID(), RoomEventCallback, NULL);
	Asura_Assertf(iRet == CELL_OK, ("sceNpMatching2RegisterRoomEventCallback failed (0x%x)", iRet));

	// Register for signaling
	if (g_bUseSignallingCallback)
	{
#ifdef ASURA_ASSERT
		iRet = 
#endif
			sceNpMatching2RegisterSignalingCallback(Asura_PS3_NP::GetContextID(), RoomSignalingCallback, NULL);
		Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2RegisterSignalingCallback failed (0x%x)", iRet) );
	}
}

Asura_PS3_NP_Room::~Asura_PS3_NP_Room()
{
	Asura_Assert(s_pxThis == this, "only one instance currently supported");
	s_pxThis = NULL;

	// leave the room - assuming we don't need a callback
	const bool bForcedDisconnect = (m_eState == ASURA_PS3_NP_ROOM_STATE_DESTROYED);
	if (m_eState == ASURA_PS3_NP_ROOM_STATE_IDLE)
	{
		Asura_PS3_AVChat2::LeaveChat(bForcedDisconnect);

		// don't call np functions if there is an error.
		if (!Asura_PS3_NP::IsInErrorState())
		{
			SceNpMatching2LeaveRoomRequest xReqParam;
			SceNpMatching2RequestOptParam xOptParam;
			SceNpMatching2RequestId xAssignedReqId;

			Asura_Mem::Set(&xReqParam, 0, sizeof(xReqParam));
			xReqParam.roomId = m_xRoomId;

			Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
			xOptParam.cbFunc = NULL;
			xOptParam.cbFuncArg = NULL;
			xOptParam.timeout = 0; // default
			xOptParam.appReqId = ++g_uReqSeqNum;

#ifdef ASURA_ASSERT
			const int iRet = 
#endif
				sceNpMatching2LeaveRoom(Asura_PS3_NP::GetContextID(), &xReqParam, &xOptParam, &xAssignedReqId);
			Asura_Assertf(iRet == CELL_OK, ("sceNpMatching2LeaveRoom failed (0x%x)", iRet));
		}
	}
	else if (m_eState == ASURA_PS3_NP_ROOM_STATE_DESTROYED)
	{
		Asura_PS3_AVChat2::LeaveChat(bForcedDisconnect);
	}
	else
	{
		Asura_Assert(m_xRoomId == 0, "should probably have left the room");
	}

	delete[] m_pxPlayers;
}

void Asura_PS3_NP_Room::HostRoom(u_int uPublicSlots, u_int uPrivateSlots, 
								 const SceNpMatching2IntAttr* pxSearchableAttributes, 
								 u_int uNumSearchableAttributes)
{
	const u_int uTotalSlots = uPublicSlots + uPrivateSlots;

	SceNpMatching2RoomPasswordSlotMask xRoomPasswordSlotMask;
	Asura_Mem::Set(&xRoomPasswordSlotMask, 0, sizeof(xRoomPasswordSlotMask));
	for (u_int u = 0; u < uPrivateSlots; ++u) 
	{
		SCE_NP_MATCHING2_ADD_SLOTNUM_TO_ROOM_PASSWORD_SLOT_MASK(
			xRoomPasswordSlotMask, (uTotalSlots - u));
	}

	SceNpMatching2SignalingOptParam xSigOptParam;
	Asura_Mem::Set(&xSigOptParam, 0, sizeof(xSigOptParam));
	xSigOptParam.type = SCE_NP_MATCHING2_SIGNALING_TYPE_STAR;
	xSigOptParam.hubMemberId = 0; //Indicates room owner is the hub
	
	SceNpMatching2CreateJoinRoomRequest xReqParam;
	SceNpMatching2RequestOptParam xOptParam;

	Asura_Mem::Set(&xReqParam, 0, sizeof(xReqParam));
	//E Request parameters
	xReqParam.worldId = Asura_PS3_NP::GetWorldID();
	xReqParam.lobbyId = 0;  //E Specify 0 if the room does not belong to a lobby
	//E Parameters for creating a room
	xReqParam.maxSlot = uTotalSlots;
	xReqParam.groupConfig = NULL;
	xReqParam.groupConfigNum = 0;
	xReqParam.roomPassword = &m_xPassword;
	xReqParam.passwordSlotMask = &xRoomPasswordSlotMask;
	xReqParam.flagAttr = 0;
	xReqParam.roomSearchableIntAttrExternal = const_cast<SceNpMatching2IntAttr*>(pxSearchableAttributes);// library is not const-correct
	xReqParam.roomSearchableIntAttrExternalNum = uNumSearchableAttributes;
	xReqParam.roomBinAttrInternal = NULL;
	xReqParam.roomBinAttrInternalNum = 0;
	xReqParam.roomSearchableBinAttrExternal = NULL;
	xReqParam.roomSearchableBinAttrExternalNum = 0;
	xReqParam.roomBinAttrExternal = NULL;
	xReqParam.roomBinAttrExternalNum = 0;
	xReqParam.allowedUser = NULL;
	xReqParam.allowedUserNum = 0;
	xReqParam.blockedUser = const_cast<SceNpId*>(Asura_PS3_NP_Basic::GetBlockList()); // SDK is not const correct
	xReqParam.blockedUserNum = Asura_PS3_NP_Basic::GetBlockListSize();
	//E Parameters for joining a room
	xReqParam.roomMemberBinAttrInternal = NULL;
	xReqParam.roomMemberBinAttrInternalNum = 0;
	xReqParam.teamId = 0;
	xReqParam.joinRoomGroupLabel = NULL;
	xReqParam.sigOptParam = &xSigOptParam;

	Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
	xOptParam.cbFunc = RoomRequestCallback;
	xOptParam.cbFuncArg = NULL;
	xOptParam.timeout = 0; // default
	xOptParam.appReqId = ++g_uReqSeqNum;

	g_xJoinRoomCallbackData.Reset();

	const int iRet = sceNpMatching2CreateJoinRoom(Asura_PS3_NP::GetContextID(), 
		&xReqParam, &xOptParam, &g_xAssignedReqId);

	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2CreateJoinRoom failed (0x%x)", iRet) );

	if (iRet < 0) 
	{
		m_eState = ASURA_PS3_NP_ROOM_STATE_ERROR;
	}
	else
	{
		m_eState = ASURA_PS3_NP_ROOM_STATE_CREATING;
		m_bIsHost = true;
	}
}

void Asura_PS3_NP_Room::JoinRoom(SceNpMatching2RoomId xRoomID)
{
	SceNpMatching2JoinRoomRequest xReqParam;
	SceNpMatching2RequestOptParam xOptParam;

	Asura_Mem::Set(&xReqParam, 0, sizeof(xReqParam));
	xReqParam.roomId = xRoomID;
	xReqParam.roomPassword = NULL; //E Room password of reserved slot
	xReqParam.joinRoomGroupLabel = NULL;
	xReqParam.teamId = 0; //E Initial team ID
	xReqParam.roomMemberBinAttrInternal = NULL;
	xReqParam.roomMemberBinAttrInternalNum = 0;

	Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
	xOptParam.cbFunc = RoomRequestCallback;
	xOptParam.cbFuncArg = NULL;
	xOptParam.timeout = 0; // default
	xOptParam.appReqId = ++g_uReqSeqNum;

	g_xJoinRoomCallbackData.Reset();

	const int iRet = sceNpMatching2JoinRoom(Asura_PS3_NP::GetContextID(), 
		&xReqParam, &xOptParam, &g_xAssignedReqId);

	Asura_Assertf( iRet == CELL_OK, ("sceNpMatching2JoinRoom failed (0x%x)", iRet) );

	if (iRet < 0) 
	{
		m_eState = ASURA_PS3_NP_ROOM_STATE_ERROR;
	}
	else
	{
		m_eState = ASURA_PS3_NP_ROOM_STATE_CREATING;
		m_bIsHost = false;
	}

}

bool Asura_PS3_NP_Room::GrantOwnership( SceNpMatching2RoomMemberId usNewOwner )
{
	Asura_Assert(m_bIsHost,"Only the host can grant room ownership");
	bool bResult = false;
	if (m_bIsHost)
	{
		SceNpMatching2GrantRoomOwnerRequest xReqParam;
		SceNpMatching2RequestId xAssignedReqId;

		xReqParam.roomId = m_xRoomId;
		xReqParam.newOwner = usNewOwner;

		int iRet = sceNpMatching2GrantRoomOwner( Asura_PS3_NP::GetContextID(), &xReqParam, NULL, &xAssignedReqId);
		bResult = (iRet == CELL_OK);
		Asura_Assertf(bResult, ("sceNpMatching2GrantRoomOwner failed (0x%x)", iRet));
	}
	
	return bResult;
}

bool Asura_PS3_NP_Room::SetOwnerAutoGrant( const SceNpMatching2RoomMemberId* pusOwners, u_int uNumOwners )
{
	Asura_Assert(m_bIsHost,"Only the host can set the owner privilege ranking")
	Asura_Assert((pusOwners != NULL) && (uNumOwners > 0),"Invalid parameters supplied");
	bool bResult = false;
	if (m_bIsHost)
	{
		//Treat an empty owner list, or 1 or less members as a request to disable auto ownership transfer
		//HACK change > 0 back to > 1
		bool bEnable = ((uNumOwners > 0) && (pusOwners));

		SceNpMatching2RequestOptParam xOptParam;
		SceNpMatching2SetRoomDataInternalRequest xReqParam;

		Asura_Mem::Set(&xOptParam, 0, sizeof(xOptParam));
		xOptParam.cbFunc = RoomRequestCallback;
		xOptParam.cbFuncArg = NULL;
		xOptParam.timeout = 0; // default
		xOptParam.appReqId = ++g_uReqSeqNum;

		memset(&xReqParam, 0, sizeof(xReqParam));
		xReqParam.roomId = m_xRoomId; //E Specify ID of current room
		xReqParam.flagFilter = SCE_NP_MATCHING2_ROOM_FLAG_ATTR_OWNER_AUTO_GRANT;
		xReqParam.flagAttr = (bEnable) ? SCE_NP_MATCHING2_ROOM_FLAG_ATTR_OWNER_AUTO_GRANT : 0;
		xReqParam.roomBinAttrInternal = NULL;
		xReqParam.roomBinAttrInternalNum = 0;
		xReqParam.passwordConfig = NULL;
		xReqParam.passwordConfigNum = 0;
		xReqParam.passwordSlotMask = NULL;
		xReqParam.ownerPrivilegeRank = (bEnable) ? const_cast<SceNpMatching2RoomMemberId*>(pusOwners) : NULL; // SDK is not const correct
		xReqParam.ownerPrivilegeRankNum = (bEnable) ? uNumOwners : 0;

		int iRet = sceNpMatching2SetRoomDataInternal( Asura_PS3_NP::GetContextID(), &xReqParam, &xOptParam, &g_xAssignedReqId);
		bResult = (iRet == CELL_OK);
		Asura_Assertf(bResult, ("sceNpMatching2SetRoomDataInternal failed (0x%x)", iRet));
	}

	return bResult;
}

void Asura_PS3_NP_Room::StartGame(bool bAllowJoinInProgress)
{
	// set the room as closed and hidden if join in progress, always disable migration
	int iRet;
	u_int uJoinInProgressFlags = (bAllowJoinInProgress) ? 0 : (SCE_NP_MATCHING2_ROOM_FLAG_ATTR_HIDDEN | SCE_NP_MATCHING2_ROOM_FLAG_ATTR_CLOSED);

	SceNpMatching2SetRoomDataInternalRequest xReqParam;
	SceNpMatching2RequestId xAssignedReqId;

	//E Request parameters
	memset(&xReqParam, 0, sizeof(xReqParam));
	xReqParam.roomId = m_xRoomId; //E Specify ID of current room
	xReqParam.flagFilter = (uJoinInProgressFlags | SCE_NP_MATCHING2_ROOM_FLAG_ATTR_OWNER_AUTO_GRANT);
	xReqParam.flagAttr = (uJoinInProgressFlags);
	xReqParam.roomBinAttrInternal = NULL;
	xReqParam.roomBinAttrInternalNum = 0;
	xReqParam.passwordConfig = NULL;
	xReqParam.passwordConfigNum = 0;
	xReqParam.passwordSlotMask = NULL;
	xReqParam.ownerPrivilegeRank = NULL;
	xReqParam.ownerPrivilegeRankNum = 0;

	//E Assuming that appropriate values are stored to ctxId, optParam
	iRet = sceNpMatching2SetRoomDataInternal( Asura_PS3_NP::GetContextID(), 
		&xReqParam, NULL, &xAssignedReqId);
	Asura_Assertf(iRet == CELL_OK, ("sceNpMatching2SetRoomDataInternal failed (0x%x)", iRet));
}

void Asura_PS3_NP_Room::InviteFriendToRoom(const Asura_UnicodeChar* wszSubject, const Asura_UnicodeChar* wszBody) const
{
	char* pcUtf8Subject = NULL;
	char* pcUtf8Body = NULL;

	// allow buffers to be big enough for every character to be a triple-byte character.
	char acUtf8Subject[SCE_NP_BASIC_SUBJECT_CHARACTER_MAX*3+1];	// 18*3+1 = 55 bytes
	char acUtf8Body[SCE_NP_BASIC_BODY_CHARACTER_MAX*3+1];		// 512*3+1 = 1537 bytes

	Asura_Assert(wszSubject && wszBody, "must specify subject and body");
	if (!wszSubject || !wszBody)
	{
		return;
	}

	Asura_Assertf(wcslen(wszSubject) <= SCE_NP_BASIC_SUBJECT_CHARACTER_MAX, 
		("invitation subject %S too long (max length %d)", wszSubject, SCE_NP_BASIC_SUBJECT_CHARACTER_MAX));
	Asura_Assertf(wcslen(wszBody) <= SCE_NP_BASIC_BODY_CHARACTER_MAX, 
		("invitation body %S too long (max length %d)", wszBody, SCE_NP_BASIC_BODY_CHARACTER_MAX));
	Asura_UnicodeToUTF8(wszSubject, acUtf8Subject, sizeof(acUtf8Subject));
	Asura_UnicodeToUTF8(wszBody, acUtf8Body, sizeof(acUtf8Body));

	pcUtf8Subject = acUtf8Subject;
	pcUtf8Body = acUtf8Body;

	SceNpBasicMessageDetails xMsg;
	Asura_PS3_InvitationInfo xInvitationInfo;

	//E Set invitation information
	Asura_Mem::Set(&xInvitationInfo, 0, sizeof(xInvitationInfo));
	xInvitationInfo.m_xRoomId = m_xRoomId;
	// here's how to do passwords:
	//xInvitationInfo.m_bWithPassword = true;
	//memcpy(xInvitationInfo.m_xRoomPassword.data, "PASSWORD", 8);

	Asura_Mem::Set(&xMsg, 0, sizeof(xMsg));
	xMsg.msgId = 0;
	xMsg.mainType = SCE_NP_BASIC_MESSAGE_MAIN_TYPE_INVITE;
	xMsg.subType = SCE_NP_BASIC_MESSAGE_INVITE_SUBTYPE_ACTION_ACCEPT_DENY;
	xMsg.msgFeatures = SCE_NP_BASIC_MESSAGE_FEATURES_MULTI_RECEIPIENTS;
	xMsg.npids = NULL;
	xMsg.count = 0;
	xMsg.subject = pcUtf8Subject;
	xMsg.body = pcUtf8Body;
	xMsg.data = &xInvitationInfo;
	xMsg.size = sizeof(xInvitationInfo);

#ifdef ASURA_ASSERT
	int iRet = 
#endif
		sceNpBasicSendMessageGui(&xMsg, SYS_MEMORY_CONTAINER_ID_INVALID);
	Asura_Assertf(iRet == CELL_OK, ("sceNpBasicSendMessageGui failed (%x)", iRet));
}

void Asura_PS3_NP_Room::AddPlayer( const SceNpId& xNpId )
{
	Asura_Assert(!IsPlayerInRoom(xNpId), "shouldn't add player already in the room");
	Asura_Assert(m_uNumPlayers < m_uMaxPlayers, "Can't add player - room is full");
	if (m_uNumPlayers < m_uMaxPlayers)
	{
		m_pxPlayers[m_uNumPlayers] = xNpId;
		++m_uNumPlayers;
	}
}

void Asura_PS3_NP_Room::RemovePlayer( const SceNpId& xNpId )
{
#ifdef ASURA_ASSERT
	bool bFound = false;
#endif
	for (u_int u = 0; u < m_uNumPlayers; ++u)
	{
		if (sceNpUtilCmpNpId(&m_pxPlayers[u], &xNpId) == 0)
		{
			// swap with last
			m_pxPlayers[u] = m_pxPlayers[m_uNumPlayers - 1];
			--m_uNumPlayers;
#ifdef ASURA_ASSERT
			bFound = true;
#endif
			break;
		}
	}

	Asura_Assert(bFound, "RemovePlayer(): player not found");
}

bool Asura_PS3_NP_Room::IsPlayerInRoom( const SceNpId& xNpId ) const
{
	for (u_int u = 0; u < m_uNumPlayers; ++u)
	{
		if (sceNpUtilCmpNpId(&m_pxPlayers[u], &xNpId) == 0)
		{
			return true;
		}
	}
	return false;
}

const SceNpId& Asura_PS3_NP_Room::GetPlayer( u_int uIndex ) const
{ 
	Asura_Assert(uIndex < m_uNumPlayers, "invalid index");
	return m_pxPlayers[uIndex];
}

void Asura_PS3_NP_Room::Update()
{
	switch (m_eState)
	{
		case ASURA_PS3_NP_ROOM_STATE_CREATING:
		{
			if (g_xJoinRoomCallbackData.m_bDone)
			{
				if (g_xJoinRoomCallbackData.m_bError)
				{
					m_eState = ASURA_PS3_NP_ROOM_STATE_ERROR;
					m_iErrorCode = g_xJoinRoomCallbackData.m_iErrorCode;
				}
				else
				{
					Asura_PS3_AVChat2::JoinChat(&m_xRoomId, m_usLocalMemberId);
					m_eState = ASURA_PS3_NP_ROOM_STATE_IDLE;
				}
			}
			break;
		}

		default:
		{
			break;
		}
	}
}

// Callback function will be executed on a different thread to the main one.
void Asura_PS3_NP_Room::RoomRequestCallback(SceNpMatching2ContextId xCtxId, 
										 SceNpMatching2RequestId xReqId, 
										 SceNpMatching2Event usEvent,
										 SceNpMatching2EventKey xEventKey, 
										 int iErrorCode, 
										 size_t uDataSize, 
										 void* pArg)
{
	Asura_Assert(s_pxThis, "no room object");
	if (!s_pxThis)
	{
		return;
	}

	if (xReqId != g_xAssignedReqId)
	{
		Asura_Assert(false, "unexpected room request callback");
		g_xJoinRoomCallbackData.m_bError = true;
		g_xJoinRoomCallbackData.m_bDone = true;
		g_xJoinRoomCallbackData.m_iErrorCode = SCE_NP_MATCHING2_ERROR_REQUEST_NOT_FOUND;
		return;
	}

	if (iErrorCode < 0) 
	{
		Asura_Assertf(false, ("room request callback (Request %d, Event: %d) error 0x%x", xReqId, xEventKey, iErrorCode));
		g_xJoinRoomCallbackData.m_bError = true;
		g_xJoinRoomCallbackData.m_bDone = true;
		g_xJoinRoomCallbackData.m_iErrorCode = iErrorCode;
		return;
	}

	// using HeapArray to create a temporary buffer that is automatically 
	// deleted when it goes out of scope.
	Asura_HeapArray<u_char> xBuffer;
	void* pResponseData = NULL;
	if (uDataSize)
	{
		xBuffer.SetSize(uDataSize);
		pResponseData = &xBuffer[0];
		if (!pResponseData) 
		{
			g_xJoinRoomCallbackData.m_bError = true;
			g_xJoinRoomCallbackData.m_bDone = true;
			g_xJoinRoomCallbackData.m_iErrorCode = SCE_NP_MATCHING2_ERROR_OUT_OF_MEMORY;
			return;
		}

		Asura_Mem::Set(pResponseData, 0, uDataSize);

		int iRet = sceNpMatching2GetEventData(xCtxId, xEventKey, pResponseData, uDataSize);
		Asura_Assertf( iRet >= 0, ("sceNpMatching2GetEventData failed (0x%x)", iRet) );
		if (iRet < 0) 
		{
			Asura_Assert(false, "create / join room failed");
			g_xJoinRoomCallbackData.m_bError = true;
			g_xJoinRoomCallbackData.m_bDone = true;
			g_xJoinRoomCallbackData.m_iErrorCode = iRet;
			return;
		}	
	}

	switch (usEvent)
	{
		case SCE_NP_MATCHING2_REQUEST_EVENT_JoinRoom:
		{
			if (!pResponseData) 
			{
				Asura_Assert(pResponseData,"Missing SCE_NP_MATCHING2_REQUEST_EVENT_JoinRoom data"); 
				return;
			}
			const SceNpMatching2JoinRoomResponse* pxJoinRoomResponse = reinterpret_cast
				<const SceNpMatching2JoinRoomResponse*>(pResponseData);
			s_pxThis->m_usHostMemberId = pxJoinRoomResponse->roomDataInternal->memberList.owner->memberId;
			s_pxThis->m_usLocalMemberId = pxJoinRoomResponse->roomDataInternal->memberList.me->memberId;
			s_pxThis->m_xRoomId = pxJoinRoomResponse->roomDataInternal->roomId;
			break;
		}
		case SCE_NP_MATCHING2_REQUEST_EVENT_CreateJoinRoom :
		{
			if (!pResponseData) 
			{
				Asura_Assert(pResponseData,"Missing SCE_NP_MATCHING2_REQUEST_EVENT_CreateJoinRoom data"); 
				return;
			}
			const SceNpMatching2CreateJoinRoomResponse* pxJoinRoomResponse = reinterpret_cast
				<const SceNpMatching2CreateJoinRoomResponse*>(pResponseData);
			s_pxThis->m_usHostMemberId = pxJoinRoomResponse->roomDataInternal->memberList.owner->memberId;
			s_pxThis->m_usLocalMemberId = pxJoinRoomResponse->roomDataInternal->memberList.me->memberId;
			s_pxThis->m_xRoomId = pxJoinRoomResponse->roomDataInternal->roomId;
			break;
		}
		case SCE_NP_MATCHING2_REQUEST_EVENT_SetRoomDataInternal:
		{
			Asura_Assertf(iErrorCode == 0,("A request to SetRoomDataInternal failed. Error %d", iErrorCode));
			break;
		}
		default:
		{
			break;
		}
	}

	g_xJoinRoomCallbackData.m_bDone = true;
}


#ifdef ASURA_NO_DEBUG_OUTPUT
#define OUTPUT_CASE_TO_DEBUGGER( x )
#else
#define OUTPUT_CASE_TO_DEBUGGER( x ) \
		case (x): Asura_OutputToDebugger(("Received " #x "\n")); break
#endif

void Asura_PS3_NP_Room::RoomEventCallback(SceNpMatching2ContextId xCtxId, 
											SceNpMatching2RoomId xRoomId,
											SceNpMatching2Event xEvent,
											SceNpMatching2EventKey xEventKey,
											int iErrorCode, 
											size_t uDataSize, 
											void* pArg)
{
	Asura_Assert(s_pxThis, "no room object");
	if (!s_pxThis)
	{
		return;
	}

	// using HeapArray to create a temporary buffer that is automatically 
	// deleted when it goes out of scope.
	Asura_HeapArray<u_char> xBuffer;
	void* pEventData = NULL;
	if (uDataSize > 0)
	{
		xBuffer.SetSize(uDataSize);
		pEventData = &xBuffer[0];
		Asura_Assert(pEventData,"Unable to create event data");
		if (pEventData) 
		{
			Asura_Mem::Set(pEventData, 0, uDataSize);
			int iRet = sceNpMatching2GetEventData( xCtxId, xEventKey, pEventData, uDataSize );
			Asura_Assertf(iRet == static_cast<int>(uDataSize),("sceNpMatching2GetEventData failed (%x)",iRet));
			if (iRet != static_cast<int>(uDataSize))
			{
				pEventData = NULL;
			}
		}
		else
		{
			int iRet = sceNpMatching2ClearEventData( xCtxId, xEventKey );
			Asura_Assertf(iRet == static_cast<int>(uDataSize),("sceNpMatching2ClearEventData failed (%x)",iRet));
		}
	}

	switch (xEvent)
	{
		case SCE_NP_MATCHING2_ROOM_EVENT_RoomDestroyed:
		{
			Asura_OutputToDebugger(("Received SCE_NP_MATCHING2_ROOM_EVENT_RoomDestroyed\n"));
			if (s_pxThis)
			{
				s_pxThis->m_eState = ASURA_PS3_NP_ROOM_STATE_DESTROYED;
			}
			break;
		}
		case SCE_NP_MATCHING2_ROOM_EVENT_RoomOwnerChanged:
		{
			Asura_OutputToDebugger(("Received SCE_NP_MATCHING2_ROOM_EVENT_RoomOwnerChanged\n"));
			Asura_Assert(pEventData,"This event should have data");
			if (pEventData)
			{
				const SceNpMatching2RoomOwnerUpdateInfo* pxOwnerUpdateInfo = static_cast<SceNpMatching2RoomOwnerUpdateInfo*>(pEventData);
				Asura_Assert(!s_pxThis->m_bIsHost || (s_pxThis->m_usLocalMemberId == pxOwnerUpdateInfo->prevOwner),"Something went wrong changing hosts");
				s_pxThis->m_bIsHost = (pxOwnerUpdateInfo->newOwner == s_pxThis->m_usLocalMemberId);
				s_pxThis->m_usHostMemberId = pxOwnerUpdateInfo->newOwner;
				Asura_OutputToDebugger(("Cause: %d, Old Owner: %d, New owner: %d\n", pxOwnerUpdateInfo->eventCause, pxOwnerUpdateInfo->prevOwner, pxOwnerUpdateInfo->newOwner));
			}
			break;
		}
		OUTPUT_CASE_TO_DEBUGGER(SCE_NP_MATCHING2_ROOM_EVENT_MemberJoined);
		OUTPUT_CASE_TO_DEBUGGER(SCE_NP_MATCHING2_ROOM_EVENT_MemberLeft);
		OUTPUT_CASE_TO_DEBUGGER(SCE_NP_MATCHING2_ROOM_EVENT_Kickedout);
		OUTPUT_CASE_TO_DEBUGGER(SCE_NP_MATCHING2_ROOM_EVENT_UpdatedRoomDataInternal);
		OUTPUT_CASE_TO_DEBUGGER(SCE_NP_MATCHING2_ROOM_EVENT_UpdatedRoomMemberDataInternal);
		OUTPUT_CASE_TO_DEBUGGER(SCE_NP_MATCHING2_ROOM_EVENT_UpdatedSignalingOptParam);
		default:
		{
			Asura_OutputToDebugger(("received room message callback for event %d\n", xEvent));
			break;
		}
	}
}

void Asura_PS3_NP_Room::RoomMessageCallback(SceNpMatching2ContextId xCtxId, 
											SceNpMatching2RoomId xRoomId,
											SceNpMatching2RoomMemberId xSrcMemberId,
											SceNpMatching2Event xEvent,
											SceNpMatching2EventKey xEventKey,
											int iErrorCode, 
											size_t uDataSize, 
											void* pArg)
{
	// not sure if we need this or not yet...
	Asura_OutputToDebugger(("received room message callback for event %d\n", xEvent));

	// need to clear data to prevent buffer overflows, replace this with sceNpMatching2GetEventData if we become
	// interested in it.
	sceNpMatching2ClearEventData( xCtxId, xEventKey );
}

void Asura_PS3_NP_Room::RoomSignalingCallback(SceNpMatching2ContextId xCtxId,
											  SceNpMatching2RoomId xRoomId,
											  SceNpMatching2RoomMemberId usPeerRoomMemberId,
											  SceNpMatching2Event usEvent,
											  int iErrorCode,
											  void *arg)
{
	Asura_OutputToDebugger_Retail(("[ %s ]\n", __FUNCTION__));

	Asura_OutputToDebugger_Retail(("Asura_PS3_NP::SignalCallback, ctxId = %d, roomId = %d, peerRoomId = %d, error = %d, event = ", 
		xCtxId, xRoomId, usPeerRoomMemberId, iErrorCode));

	switch (usEvent)
	{
		case SCE_NP_MATCHING2_SIGNALING_EVENT_Established:	Asura_OutputToDebugger_Retail(("SCE_NP_MATCHING2_SIGNALING_EVENT_Established\n"));	break;
		case SCE_NP_MATCHING2_SIGNALING_EVENT_Dead:			Asura_OutputToDebugger_Retail(("SCE_NP_MATCHING2_SIGNALING_EVENT_Dead\n"));			break;
		default:											Asura_OutputToDebugger_Retail(("unknown (%d)\n", usEvent));							break;
	}
}

void Asura_PS3_NP_Room::RegisterRoomCallback()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
		sceNpMatching2RegisterRoomMessageCallback( Asura_PS3_NP::GetContextID(), RoomMessageCallback, NULL );
	Asura_Assertf(iRet == CELL_OK, ("sceNpMatching2RegisterRoomMessageCallback failed (%x)", iRet));
}

