//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_NP_TUS.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     wrappers for PS3 Title User Storage SDK
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Online/Asura_PS3_NP_TUS.h"

// Asura independent
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Asura PS3
#include "Network/Online/Asura_PS3_NP.h"

// PS3 library
#include <cell/sysmodule.h>
#include <np.h>

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
int Asura_PS3_NP_TUS::s_iTitleContext;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_PS3_NP_TUS::Initialise()
{
	// load PRX for the sysutil.
	// docs indicate this loads NP and NP2 automatically. In theory we 
	// already have those loaded so hopefully this will still work.
	int iRet = cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_NP_TUS);
	Asura_Assertf(iRet == CELL_OK,
		("cellSysmoduleLoadModule(CELL_SYSMODULE_SYSUTIL_NP_TUS) failed 0x%x", iRet));

	// init score utility and create title context
	iRet = sceNpTusInit(0); // 0 => use default thread priority
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusInit() failed 0x%x", iRet));

	iRet = sceNpTusCreateTitleCtx( &Asura_PS3_NP::GetCommunicationID(), 
		&Asura_PS3_NP::GetCommunicationPassphrase(), &Asura_PS3_NP::GetNpID());
	Asura_Assertf(iRet > 0, ("sceNpTusCreateTitleCtx() failed 0x%x", iRet));
	if (iRet > 0)
	{
		s_iTitleContext = iRet;
	}
}

void Asura_PS3_NP_TUS::Shutdown()
{
	// shutdown and destroy title context
	int iRet;

	iRet = sceNpTusDestroyTitleCtx(s_iTitleContext);
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusDestroyTitleCtx() failed 0x%x", iRet));

	iRet = sceNpTusTerm();
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusTerm() failed 0x%x", iRet));

	// unload PRX module
	// docs indicate this unloads NP and NP2 automatically. We don't want that!
	// I'm hoping that since NP/NP2 were already loaded when TUS was loaded, it
	// will know not to unload them.
	cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_NP_TUS);
	Asura_Assertf(iRet == CELL_OK, 
		("cellSysmoduleUnloadModule(CELL_SYSMODULE_SYSUTIL_NP_TUS) failed 0x%x", iRet));
}

Asura_PS3_TUSTransactionManagerBase::Asura_PS3_TUSTransactionManagerBase()
: m_bError(false)
, m_iTransactionID(0)
, m_iResult(0)
{
}

Asura_PS3_TUSTransactionManagerBase::~Asura_PS3_TUSTransactionManagerBase()
{
	int iRet;
	// see if we must abort a pending transaction
	if (m_iTransactionID > 0)
	{
		iRet = sceNpTusAbortTransaction(m_iTransactionID);
		Asura_Assertf(iRet == CELL_OK, ("sceNpTusAbortTransaction() failed 0x%x", iRet));

		if (iRet == CELL_OK)
		{
			int iResult;
			iRet = sceNpTusWaitAsync(m_iTransactionID, &iResult);
			Asura_Assertf(iRet >= 0, ("sceNpTusWaitAsync() failed 0x%x", iRet));
		}
	}
}

bool Asura_PS3_TUSTransactionManagerBase::Update()
{
	// if we've already finished, the transaction ID will be set to 0 so there's nothing to do.
	if (m_iTransactionID == 0)
	{
		return true;
	}

	int iResult = 0;
	int iRet = sceNpTusPollAsync(m_iTransactionID, &iResult);
	Asura_Assertf(iRet >= 0, ("sceNpTusPollAsync() failed 0x%x", iRet));
	if (iRet == 0)
	{
		// finished. Record the result/error code. Calling code must check it.
		m_iResult = iResult;
		m_bError = (iResult < 0);
		iRet = sceNpTusDestroyTransactionCtx(m_iTransactionID);
		Asura_Assertf(iRet >= 0, ("sceNpTusDestroyTransactionCtx() failed 0x%x", iRet));
		Asura_OutputToDebugger(("transaction ID 0x%08x finished (result = 0x%x)\n", m_iTransactionID, iRet));
		m_iTransactionID = 0;
		return true;
	}
	else if (iRet == 1)
	{
		// still active
		return false;
	}

	m_bError = true;
	return true;
}

int Asura_PS3_TUSTransactionManagerBase::CreateTransactionID()
{
	Asura_Assert(m_iTransactionID == 0, "already have transaction context");
	int iRet = sceNpTusCreateTransactionCtx(Asura_PS3_NP_TUS::GetTitleContext());
	Asura_Assertf(iRet > 0, ("sceNpTusCreateTransactionCtx() failed 0x%x", iRet));
	if (iRet > 0)
	{
		m_iTransactionID = iRet;
		m_bError = false;
		m_iResult = EINPROGRESS;

		Asura_OutputToDebugger(("created transaction ID 0x%08x\n", iRet));
	}
	return iRet;
}


//////////////////////////////////////////////////////////////////////////
// TUS Data functions
//////////////////////////////////////////////////////////////////////////
int Asura_PS3_TUSTransactionManagerData::GetDataAsync(const SceNpId& xTargetNpId, 
													  ASURA_PS3_TUS_DATA_SLOT eSlotID, 
													  void* pData, 
													  u_int uDataSize)
{
	const int iTransactionID = CreateTransactionID();
	// do a get data call for the entire amount of data.
	const int iRet = sceNpTusGetDataAsync( iTransactionID, &xTargetNpId, 
		eSlotID, &m_xDataStatus, sizeof(m_xDataStatus), pData, uDataSize, NULL );
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusGetDataAsync() failed 0x%x", iRet));
	return iRet;
}

int Asura_PS3_TUSTransactionManagerData::SetDataAsync(const SceNpId& xTargetNpId, ASURA_PS3_TUS_DATA_SLOT eSlotID, 
													  const void* pData, u_int uDataSize)
{
	const int iTransactionID = CreateTransactionID();
	// do a set data call for the entire amount of data.
	const int iRet = sceNpTusSetDataAsync( iTransactionID, &xTargetNpId, 
		eSlotID, uDataSize, uDataSize, pData, NULL, 0, NULL);
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusSetDataAsync() failed 0x%x", iRet));
	return iRet;
}

bool Asura_PS3_TUSTransactionManagerData::IsDataValid(const char* m_pcData, u_int uExpectedDataSize) const
{
	Asura_Assert(m_xDataStatus.hasData && m_xDataStatus.data == m_pcData && 
		m_xDataStatus.dataSize == uExpectedDataSize, "TUS data error");
	return (m_xDataStatus.hasData && m_xDataStatus.data == m_pcData && 
		m_xDataStatus.dataSize == uExpectedDataSize);
}


//////////////////////////////////////////////////////////////////////////
// TUS Var functions
//////////////////////////////////////////////////////////////////////////
Asura_PS3_TUSTransactionManagerGetVars::Asura_PS3_TUSTransactionManagerGetVars(u_int uMaxVarsToGet)
{
	Asura_Assert(uMaxVarsToGet <= 101, "too many vars");// limit specified in SDK docs
	m_pxVarArray = new SceNpTusVariable[uMaxVarsToGet];
	m_uMaxVarsToGet = uMaxVarsToGet;
}

Asura_PS3_TUSTransactionManagerGetVars::~Asura_PS3_TUSTransactionManagerGetVars()
{
	delete[] m_pxVarArray;
}

// wrapper function for sceNpTusGetMultiSlotVariableAsync
int Asura_PS3_TUSTransactionManagerGetVars::GetVarsAsync(const SceNpId& xTargetNpId, u_int uNumVars, 
													  const ASURA_PS3_TUS_VAR_SLOT* aeSlotIDs)
{
	Asura_Assert(uNumVars <= m_uMaxVarsToGet, "too many vars");
	const int iTransactionID = CreateTransactionID();
	// need a const_cast due to SDK header file not matching documented declaration :(
	const int iRet = sceNpTusGetMultiSlotVariableAsync(iTransactionID, &xTargetNpId, reinterpret_cast<int*>
		(const_cast<ASURA_PS3_TUS_VAR_SLOT*>(aeSlotIDs)), m_pxVarArray, sizeof(SceNpTusVariable) * m_uMaxVarsToGet, uNumVars, NULL);
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusGetMultiSlotVariableAsync() failed 0x%x", iRet));
	return iRet;
}

int64 Asura_PS3_TUSTransactionManagerGetVars::GetVarValue(u_int uIndex) const
{
	Asura_Assert(uIndex < m_uMaxVarsToGet, "invalid index");
	if (!m_pxVarArray[uIndex].hasData)
	{
		Asura_OutputToDebugger_Retail(("TUS var has never been set, returning 0.\n"));
		return 0;
	}
	return m_pxVarArray[uIndex].variable;
}

int Asura_PS3_TUSTransactionManagerAddToVar::AddToVarAsync(const SceNpId& xTargetNpId, 
													   ASURA_PS3_TUS_VAR_SLOT eSlotID, 
													   int64 lValueToAdd)
{
	const int iTransactionID = CreateTransactionID();
	const int iRet = sceNpTusAddAndGetVariableAsync(iTransactionID, &xTargetNpId, eSlotID, 
		lValueToAdd, &m_xVarStorage, sizeof(m_xVarStorage), NULL);
	Asura_Assertf(iRet == CELL_OK, ("sceNpTusAddAndGetVariableAsync() failed 0x%x", iRet));
	return iRet;
}

