//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_SPUGroupController.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "PPU_SPU_Utils/Asura_PS3_SPUGroupController.h"

// Stdlib platform-specific.
#include <cell/sync/lfqueue.h>
#ifndef ASURA_NO_DEBUG_OUTPUT
#include <spu_printf.h>
#endif

// Engine independent.
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"

// Engine platform-specific.
#include "PPU_SPU_Utils/spu_thread_group_utils.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_PS3_SPUGroupController::Asura_PS3_SPUGroupController( const char* szGroupName,
                                                            const void* pvELFAddr,
                                                            u_int uThreadCount,
                                                            u_int uGroupPriority,
                                                            u_int uQueueEntrySize, u_int uMaxQueueEntries,
                                                            const char* const * pszThreadNames,
                                                            bool bSuspendGroup )
{
	UtilSpuThreadGroupArg xGroupArgs;

	xGroupArgs.group_name = szGroupName;
	xGroupArgs.priority   = uGroupPriority;
	xGroupArgs.nThreads   = uThreadCount;

    for( u_int u = 0; u < uThreadCount; ++u )
	{
        xGroupArgs.thread[u].name   = ( pszThreadNames ? pszThreadNames[u] : szGroupName );
		xGroupArgs.thread[u].option = SYS_SPU_THREAD_OPTION_NONE;
		xGroupArgs.thread[u].elf    = reinterpret_cast<sys_addr_t>( pvELFAddr );

        // Note: args are setup properly later, before we actually start the
        // threads.
        Asura_Mem::Set( xGroupArgs.thread[u].arg, 0, sizeof(xGroupArgs.thread[u].arg) );
	}

    m_pxThreadGroupInfo = new UtilSpuThreadGroupInfo;
	int iRetCode = utilInitializeSpuThreadGroupAll( m_pxThreadGroupInfo, &xGroupArgs );
	Asura_Assert( CELL_OK == iRetCode, "Thread group initialisation failed" );

#ifndef ASURA_NO_DEBUG_OUTPUT
	iRetCode = spu_printf_attach_group( m_pxThreadGroupInfo->group_id );
    Asura_Assert( CELL_OK == iRetCode, "spu_printf_attach_group() failed" );
#endif

    uQueueEntrySize = (uQueueEntrySize + 15) & ~15;

	// allocate queue, including space for terminators (one terminator per thread)
    m_pvQueueBfr = Asura_AllocateAlign( uQueueEntrySize * ( uMaxQueueEntries + uThreadCount ), ASURA_ALIGN_16_BYTE );

    ASURA_STATIC_ASSERT( CELL_SYNC_LFQUEUE_ALIGN == ASURA_ALIGN_128_BYTE );
    m_pxQueue = static_cast<CellSyncLFQueue*>( Asura_AllocateAlign(CELL_SYNC_LFQUEUE_SIZE, ASURA_ALIGN_128_BYTE) );
    Asura_Mem::Set( m_pxQueue, 0, CELL_SYNC_LFQUEUE_SIZE );

    iRetCode = cellSyncLFQueueInitialize( m_pxQueue, m_pvQueueBfr, uQueueEntrySize, uMaxQueueEntries, CELL_SYNC_QUEUE_PPU2SPU, NULL );
    Asura_Assert( CELL_OK == iRetCode, "Failed to initialise queue" );

    m_pvTerminatorJob     = NULL;
    m_uTempThreadLimit    = 0;
	m_bQueueTerminated    = false;
    m_bHasEverBeenStarted = false;
    m_bSuspendGroup       = bSuspendGroup;

    (void)iRetCode; // Keep compiler quiet in non-assert builds.
}


Asura_PS3_SPUGroupController::~Asura_PS3_SPUGroupController( void )
{
    // Threads should already be back in INITIALIZED state by now, having
    // been terminated at the end of the final frame, so just some final
    // clean up here.
#ifndef ASURA_NO_DEBUG_OUTPUT
    spu_printf_detach_group( m_pxThreadGroupInfo->group_id );
#endif
    utilFinalizeSpuThreadGroupAll( m_pxThreadGroupInfo );

    Asura_DeallocateAlign( m_pxQueue );
    m_pxQueue = NULL;

    Asura_DeallocateAlign( m_pvQueueBfr );
    m_pvQueueBfr = NULL;

    delete m_pxThreadGroupInfo;
    m_pxThreadGroupInfo = NULL;
}


void Asura_PS3_SPUGroupController::Begin( uint64_t* puArg1, uint64_t* puArg2, uint64_t* puArg3, uint64_t* puArg4 )
{
    int iRetCode;

//BW I add stuff to the SPU skinning queue before starting it, so check this in End()
//     // The queue should be empty at this point, assuming whatever work was
//     // done previously has been waited for via End().
//     {
//         u_int uQueueSize;
//         iRetCode = cellSyncLFQueueSize( m_pxQueue, &uQueueSize );
// 
//         Asura_Assert( CELL_OK == iRetCode && !uQueueSize, "Asura_PS3_SPUGroupController::Begin - Queue not empty?" );
//     }

    if( !m_bHasEverBeenStarted || !m_bSuspendGroup )
    {
        // Pass new info to SPU threads.
        {
            sys_spu_thread_argument_t xArgs;

            // Pass specified override values or else defaults, ie. queue address.
            xArgs.arg2 = ( puArg2 ? *puArg2 : reinterpret_cast<uint64_t>(m_pxQueue) );
            xArgs.arg3 = ( puArg3 ? *puArg3 : 0 );
            xArgs.arg4 = ( puArg4 ? *puArg4 : 0 );

            for( int i = 0; i < m_pxThreadGroupInfo->nThreads; ++i )
            {
                // Pass specified override value or else default of thread number.
                xArgs.arg1 = ( puArg1 ? *puArg1 : SYS_SPU_THREAD_ARGUMENT_LET_32(i) );

                iRetCode = sys_spu_thread_set_argument( m_pxThreadGroupInfo->thread_id[i], &xArgs );
                Asura_Assert( CELL_OK == iRetCode, "Failed to set SPU thread args" );
            }
        }

        // And start them running again.
        iRetCode = sys_spu_thread_group_start( m_pxThreadGroupInfo->group_id );    
        Asura_Assert( CELL_OK == iRetCode, "Failed to start SPU thread group running" );

        m_bHasEverBeenStarted = true;
    }
    else
    {
        // NB: arguments cannot change in this case.
        iRetCode = sys_spu_thread_group_resume( m_pxThreadGroupInfo->group_id );    
        Asura_Assert( CELL_OK == iRetCode, "Failed to resume SPU thread group" );
    }

    // If there's a thread limit then immediately push terminator jobs to the
    // redundant threads.
    if( m_uTempThreadLimit )
    {
        for( int i = m_uTempThreadLimit; i < m_pxThreadGroupInfo->nThreads; ++i )
        {
            iRetCode = cellSyncLFQueuePush( m_pxQueue, m_pvTerminatorJob );
            Asura_Assert( CELL_OK == iRetCode, "Queue push failed" );
        }
    }
    
    (void)iRetCode; // Keep compiler quiet in non-assert builds.
}


void Asura_PS3_SPUGroupController::PushTerminators( void )
{
	if( !m_pvTerminatorJob ) return;

	Asura_Assert( !(reinterpret_cast<uintptr_t>(m_pvTerminatorJob) & 15), "" );

	u_int uNumTerminatorJobsReqd;
	if( m_uTempThreadLimit )
	{
		uNumTerminatorJobsReqd = Asura_Maths::Min( m_uTempThreadLimit, static_cast<u_int>(m_pxThreadGroupInfo->nThreads) );
	}
	else
	{
		uNumTerminatorJobsReqd = m_pxThreadGroupInfo->nThreads;
	}

	for( u_int u = 0; u < uNumTerminatorJobsReqd; ++u )
	{
		int iRetCode = cellSyncLFQueuePush( m_pxQueue, m_pvTerminatorJob );
		Asura_Assert( CELL_OK == iRetCode, "Queue push failed" );
	}
	
	m_bQueueTerminated = true;
}


void Asura_PS3_SPUGroupController::End( void )
{
    int iRetCode;

    // Push terminator job to each thread, if one has been supplied.
    if ( !m_bQueueTerminated )
    {
		PushTerminators();
    }

    if( m_bSuspendGroup )
    {
        // Caller deals with termination synchronisation in this case.
        iRetCode = sys_spu_thread_group_suspend( m_pxThreadGroupInfo->group_id );
        Asura_Assert( CELL_OK == iRetCode, "Failed to suspend thread group" );
    }
    else
    {
        // Once all work has been processed, via terminator jobs or some
        // other means, all threads in the group are expected to exit and
        // end up back in initialised state.
        int iCause, iStatus;
        iRetCode = sys_spu_thread_group_join( m_pxThreadGroupInfo->group_id, &iCause, &iStatus );
        Asura_Assert( CELL_OK == iRetCode && SYS_SPU_THREAD_GROUP_JOIN_ALL_THREADS_EXIT == iCause, "Unexpected thread group state" );
    }

    // The queue should be empty at this point.
    {
        u_int uQueueSize;
        iRetCode = cellSyncLFQueueSize( m_pxQueue, &uQueueSize );

        Asura_Assert( CELL_OK == iRetCode && !uQueueSize, "Asura_PS3_SPUGroupController::End - Queue not empty after SPUs finished?" );
    }

	m_bQueueTerminated = false; //empty queue contains no terminator

    (void)iRetCode; // Keep compiler quiet in non-assert builds.
}
