//
// Asura_PS3_EdgeAntiAliasing.cpp
//
// Ben Weston (18/9/2008)
//

// INCLUDES

#include "Core/Asura.h"

#ifdef ASURA_EDGE_ANTIALIASING

// This file's header...
#include "Rendering/Asura_PS3_EdgeAntiAliasing.h"

// PS3
#include "Rendering/Asura_PS3_QuadOverlayRenderer.h"
#include "Rendering/Asura_PS3_DeferredRendering.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"
#include "Rendering/Asura_PS3_SPU_Skinning.h"
#include "Rendering/Asura_PS3_SPU_SkinningTypes.h"
#include "Rendering/Shaders/Asura_PS3_GCM_ShadersLibrary.h"
#include "Rendering/ShadersDefinition/FP/EdgeAntiAlias_FP.h"

// independent
#include "Core/Asura_Timers.h"
#include "Rendering/Asura_Viewport.h"


// VARIABLES

Asura_PS3_EdgeAntiAliasing::AntiAliasEntity		Asura_PS3_EdgeAntiAliasing::s_axAntiAliasEntities[MAX_ANTIALIAS_ENTITIES];
u_int											Asura_PS3_EdgeAntiAliasing::s_uNumAntiAliasEntities = 0;
int												Asura_PS3_EdgeAntiAliasing::s_iDebugMode = Asura_PS3_EdgeAntiAliasing::AA_ENABLED;

static Asura_PS3_GCM_VertexShader*	g_pxVertexShader;
static Asura_PS3_FragmentShader*	g_pxFragmentShader;

//vertex size is already set in the buffer, so just specify that position is be 3 floats at the start of the vertex (we don't care about the rest).
static const Asura_PS3_VertexBufferDeclaration g_xVertex[] =
{
	{ "xIn",				0,	CELL_GCM_VERTEX_F,		3, 0 },
};


static const char* g_aszDebugModeName[] =   //corresponding to AA_DEBUG_MODE
{
	"Disabled",
	"Enabled",
	"Debug",
};

static u_int g_uNumEntitiesDrawn = 0;
static u_int g_uNumEdgesDrawn = 0;

static float g_fDepthOffset = -3000.0f; // good compromise between jaggies on very shallow edges (fixed by bigger numbers) & zipper artefact on overlapping edges on bevelled corners (often can be avoided using smaller numbers).

static float g_fLineThickness = 0.9f; 	// set this to slightly less than 1-pixel => only touch the worst pixels. This improves render performance (less pixels to shade) and reduces artefacts.


// FUNCTIONS

void Asura_PS3_EdgeAntiAliasing::Initialise( void )
{
	g_pxVertexShader = new Asura_PS3_EdgeAntiAlias_VP();
	g_pxFragmentShader =  new Asura_PS3_EdgeAntiAlias_FP();

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	Asura_CommandConsole::AddCmd(L"Render.AntiAlias.EdgeAntiAliasing", SwitchAAModes, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar(L"Render.AntiAlias.DepthOffset", &g_fDepthOffset, -FLT_MAX, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar(L"Render.AntiAlias.LineThickness", &g_fLineThickness, 0, 63.875f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
}

void Asura_PS3_EdgeAntiAliasing::Shutdown( void )
{
	delete g_pxVertexShader;
	delete g_pxFragmentShader;
	g_pxVertexShader = NULL;
	g_pxFragmentShader = NULL;
}

void Asura_PS3_EdgeAntiAliasing::BeginFrame( void )
{
	s_uNumAntiAliasEntities = 0;
}

void Asura_PS3_EdgeAntiAliasing::EndFrame( void )
{
}

//
// Draw over the edges with an anti-aliasing blur.
//
// The blur works by mimicking the rounding that originally created the jaggies, then sampling adjacent pixels to reconstruct the colour according to the
// true coverage of that pixel. e.g. if we're antialiasing the edge of an object in front of the sky, when we draw a line over the edge of that object half
// the pixels of the line will be on pixels of the object, and half will be on sky pixels. All of these pixels need to be anti-aliased to some extent, and
// we achieve this by sampling pixels either side of the line. So the object pixels gain a bit of sky colour, and the sky ones gain a bit of object colour.
// Since this is technically a directional blur of 1-pixel, some artefacts are to be expected on high-frequency patterns - they will appear to stretch by
// 1-pixel under the anti-aliasing.
//
void Asura_PS3_EdgeAntiAliasing::Render( void )
{
	g_uNumEntitiesDrawn = s_uNumAntiAliasEntities;
	g_uNumEdgesDrawn = 0;

	if ( s_uNumAntiAliasEntities == 0 || s_iDebugMode == AA_DISABLED )
		return;

	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__MISC, "Render Edge Anti-Aliasing" );


	// Clear the dest alpha to 0 first (code pinched from water, could probably do this faster in both cases)
	float fWidth =  (float)Asura_Viewport::GetWidth();		
	float fHeight = (float)Asura_Viewport::GetHeight();			
	Asura_Overlay_Quad xQuad;
	xQuad.SetRectXY(0,0, fWidth, fHeight);
	xQuad.Colour = 0x00FFFFFF;
	xQuad.TextureIndex = ASURA_TEXTURE_NONE;
	xQuad.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;	
	Asura_PS3_RenderingStates::RequestAlphaTest( false , 0x00);
	Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALPHA );
	Asura_PS3_QuadOverlayRenderer::RenderCurrentRenderTarget( xQuad );
	Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALL );
	Asura_PS3_Graphics::Flush();


	Asura_PS3_DeferredRendering::ResolveBackBuffer();
	g_pxFragmentShader->SetTexture(0, Asura_PS3_DeferredRendering::GetResolvedTexture(), false );
//	g_pxFragmentShader->SetTexture(0, Asura_PS3_DeferredRendering::GetRTAsTexture( ASURA_RENDER_TARGET_CURRENT_FRONT_BUFFER ), false );

	Asura_PS3_DeferredRendering::BindAsSingleRT();
	Asura_PS3_Graphics::SetViewportToDefaults(true);


	g_pxVertexShader->Bind();
	g_pxFragmentShader->Bind();

	Asura_PS3_RenderingStates::SetColorMaskMrt0( 0xf );
	Asura_PS3_RenderingStates::RequestAlphaTest( false );
	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
	Asura_PS3_RenderingStates::RequestZBufferEnabled( true ); //test against depth
    Asura_PS3_RenderingStates::RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		

	// setup blending - this can help resolve some issues when lines overlap
	// frame buffer alpha is 0 at the moment, our shader outputs 0.5, combine these in such a way as to approximately average the values written to the buffer.
	Asura_PS3_RenderingStates::InvalidateTranslucencyMode();
  	Asura_PS3_RenderingStates::SetAlphaBlending( true );
  	cellGcmSetBlendFuncUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_ONE_MINUS_DST_ALPHA, CELL_GCM_DST_ALPHA, CELL_GCM_ONE, CELL_GCM_SRC_ALPHA ); //alpha values will go 0, 1/2, 3/4, 7/8, etc which loosely approximates a true average (0, 1/2, 2/3, 3/4)


	Asura_PS3_RenderingStates::SetLineWidth( g_fLineThickness );

	if ( s_iDebugMode == AA_DEBUG )
	{
		//flip to a different shader & render additively (so can see overdraw)
		Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_GLOBAL, ASURA_GLOBAL_SHADER_BLANK )->Bind();
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
	}

	// offset depth
// 	Asura_PS3_RenderingStates::SetUseDepthBias( true ); - these don't work, either because I'm drawing lines or because they're already in screen space.
// 	Asura_PS3_RenderingStates::SetDepthBias( -3.0f, 0.0f ); - so instead we'll just tweak the projection matrix


	Asura_AVector4 xVecs( g_fDepthOffset*2.0f/16777216.0f, 0, 0, 0 ); //z range is -1 to 1 in 24-bit, so a difference of 1 value = a 16777216th of [-1,1]
	g_pxVertexShader->SetUniformParameter( VS_DepthOffset, xVecs.GetElements() );
	
	Asura_AVector4 xConsts( Asura_Viewport::GetWidth(), Asura_Viewport::GetHeight(), 0, 0 );
	g_pxVertexShader->SetUniformParameter( VS_Constants, xConsts.GetElements() );

	for ( u_int uEntity=0; uEntity < s_uNumAntiAliasEntities; uEntity++ )
	{
		AntiAliasEntity& xEnt = s_axAntiAliasEntities[uEntity];
		g_pxVertexShader->SetUniformParameter( VS_ViewProjectionMatrix, xEnt.m_xProjectionMatrix.GetElements() );

		xEnt.m_pxVertexBuffer->Bind( g_xVertex, 1 );

		xEnt.m_pxIndexBuffer->RenderLines( xEnt.m_iNumVerts, xEnt.m_iStartOffset );		
		
		g_uNumEdgesDrawn += xEnt.m_iNumVerts/2;
	}

	// tell SPU skinning not to overwrite the vertices till it passes this point.
	Asura_PS3_SPU_Skinning::NotifyBufferUsed();

	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__MISC );

}

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
bool Asura_PS3_EdgeAntiAliasing::SwitchAAModes ( bool bInc, bool bInterrogate, bool bHelp )
{
	if ( bHelp || bInterrogate )
	{
		Asura_CommandConsole::Printf( "%s\nentities: %d\nedges: %d", g_aszDebugModeName[s_iDebugMode], g_uNumEntitiesDrawn, g_uNumEdgesDrawn );
		return true;
	}
	
	s_iDebugMode += bInc?1:-1;
	if ( s_iDebugMode < 0 ) s_iDebugMode = NUM_AA_DEBUG_MODES-1;
	if ( s_iDebugMode >= NUM_AA_DEBUG_MODES ) s_iDebugMode = 0;

	return true;
}
#endif

#endif // #ifdef ASURA_EDGE_ANTIALIASING
