///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PS3_GCM_VertexBuffer.cpp
// Author : Andy Tate
// Date   : 22-Jul-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"

// Asura headers
#include "Debugging/Asura_Assert.h"

// Asura PS3 headers
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Rendering/Shaders/Asura_PS3_VertexShader.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_RenderingStates.h"

#include "cell/gcm/gcm_method_data.h"

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_GCM_VertexBuffer::Initialise
// Author		: Andy Tate (22-Jul-2005)
// Modified		: Catalin Arsenescu (28-March-2007)
//
// Description  : Handles the initialisation of the vertex-buffer.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GCM_VertexBuffer::Initialise( u_char uVertSize, u_int uNumVerts, const u_char ucLocation, bool bForceLocation )
{
	Asura_Assert(m_uNumVerts == 0 && m_ucVertSize == 0, "buffer already initialised");

	m_ucLocation	= ucLocation;
	m_uNumVerts		= uNumVerts;
	m_ucVertSize	= uVertSize;
	m_uBufferSize	= uNumVerts * uVertSize;

	if( !bForceLocation )
	{
		if( Asura_PS3_VideoMemoryManager::OverrideLocation( m_ucLocation ) )
		{
		}
#ifdef ASURA_PS3_USE_BALANCED_MEMORY_USAGE
		else
		{
			if( Asura_PS3_VideoMemoryManager::GetFreeMain() > Asura_PS3_VideoMemoryManager::GetFreeLocal() )
			{
				m_ucLocation	= CELL_GCM_LOCATION_MAIN;
			}
			else
			{
				m_ucLocation	= CELL_GCM_LOCATION_LOCAL;
			}
		}
#endif // ASURA_PS3_USE_BALANCED_MEMORY_USAGE
	}

	if( m_ucLocation == CELL_GCM_LOCATION_LOCAL )
	{
		m_pxVertexData	= ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( m_uBufferSize, 128 );
	}
	else
	{
		m_pxVertexData	= ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_uBufferSize, 128 );
	}

	if( !m_pxVertexData )
	{
		if( m_ucLocation == CELL_GCM_LOCATION_LOCAL )
		{
			m_ucLocation = CELL_GCM_LOCATION_MAIN;
			m_pxVertexData	= ASURA_PS3_VIDEOMEMORY_ALLOCMAIN( m_uBufferSize, 128 );
		}
		else
		{
			m_ucLocation = CELL_GCM_LOCATION_LOCAL;
			m_pxVertexData	= ASURA_PS3_VIDEOMEMORY_ALLOCLOCAL( m_uBufferSize, 128 );
		}
	}

	if(m_pxVertexData == NULL)
	{
		Asura_Assertf( false, ("Asura_PS3_GCM_VertexBuffer::Initialise() - Failed to allocate memory for vertex buffer (%d vertices)\n", uNumVerts));
	}
	else
	{
		cellGcmAddressToOffset( m_pxVertexData, &m_uBaseOffest );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_GCM_VertexBuffer::Shutdown
// Author		: Andy Tate (22-Jul-2005)
// Modified		: Catalin Arsenescu (28-March-2007)
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GCM_VertexBuffer::Shutdown (void)
{
	Destroy();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_GCM_VertexBuffer::Destroy
// Author		: Andy Tate (22-Jul-2005)
// Modified		: Catalin Arsenescu (28-March-2007)
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_GCM_VertexBuffer::Destroy (void)
{
	// Delete the GL vertex-buffers...
	if( m_pxVertexData )
	{
		if( m_ucLocation == CELL_GCM_LOCATION_LOCAL )
		{
			Asura_PS3_VideoMemoryManager::FreeLocal( m_pxVertexData );
		}
		else
		{
			Asura_PS3_VideoMemoryManager::FreeMain( m_pxVertexData );
		}
		m_ucVertSize	= 0;
		m_uNumVerts		= 0;
		m_uBufferSize	= 0;
		m_ucLocation	= 0;
		m_uBaseOffest	= 0;
		m_ucVertSize	= 0;
		m_pxVertexData	= NULL;
	}
}

void Asura_PS3_GCM_VertexBuffer::Bind( const Asura_PS3_VertexBufferDeclaration* pxDeclaration, unsigned char ucElements, u_short usFrequency, bool bClearUnusedVertsDataArray, bool *pbKeep ) const
{
	u_int										uiBaseOffset;
	uiBaseOffset	= GetBaseOffset();

	BindInternal( pxDeclaration, ucElements, uiBaseOffset, usFrequency, bClearUnusedVertsDataArray, pbKeep );
}

void Asura_PS3_GCM_VertexBuffer::BindFromVertex( const Asura_PS3_VertexBufferDeclaration* pxDeclaration, unsigned char ucElements, unsigned int uiVertex, u_short usFrequency, bool bClearUnusedVertsDataArray, bool *pbKeep ) const
{
	u_int										uiBaseOffset;
	u_char										ucVertexSize;
	
	ucVertexSize	= GetSize();
	uiBaseOffset	= GetBaseOffset() + uiVertex * ucVertexSize;

	BindInternal( pxDeclaration, ucElements, uiBaseOffset, usFrequency, bClearUnusedVertsDataArray, pbKeep );
}

void Asura_PS3_GCM_VertexBuffer::BindInternal( const Asura_PS3_VertexBufferDeclaration* pxDeclaration, unsigned char ucElements, u_int uiBaseOffset, u_short usFrequency, bool bClearUnusedVertsDataArray, bool* pbKeep ) const
{
	u_int										uElement;
	u_char										ucLoc, ucIndex, ucVertexSize;
	const Asura_PS3_VertexBufferDeclaration*	pxVBElement;
	bool										bBinded[ 16 ] = { false };   //BW awesome bug if this gets initialised to true.

	ucVertexSize	= GetSize();
	ucLoc			= GetLocation();

	struct
	{
		u_int	uiOffset;
		u_char	ucElement;
		bool	bAttributes;
		bool	bOffset;
	} xaToSet[ 17 ] = { { 0,0,false,false} };

	if( Asura_PS3_Graphics::UseGCmMacros() )
	{
		for( uElement = 0; uElement < ucElements; ++uElement)
		{
			pxVBElement	= &pxDeclaration[ uElement ];
			ucIndex		= pxVBElement->m_ucAttributeIndex;
			if (ucIndex != 0xFF)
			{
				xaToSet[ ucIndex ].uiOffset = uiBaseOffset + pxVBElement->m_uOffset;
				xaToSet[ ucIndex ].ucElement = uElement;
				bBinded[ ucIndex ] = true;

				if( Asura_PS3_RenderingStates::SetVertexAttributeNeeded( ucIndex, usFrequency, ucVertexSize, pxVBElement->m_uNumberOfCoords, pxVBElement->m_uType) )
				{
					xaToSet[ ucIndex ].bAttributes = true;
				}
				if( Asura_PS3_RenderingStates::SetVertexOffsetNeeded( ucIndex, ucLoc, xaToSet[ ucIndex ].uiOffset ) )
				{
					xaToSet[ ucIndex ].bOffset = true;
				}
			}
		}
	}
	else
	{
		for( uElement = 0; uElement < ucElements; ++uElement)
		{
			pxVBElement	= &pxDeclaration[ uElement ];
			ucIndex		= pxVBElement->m_ucAttributeIndex;
			if (ucIndex != 0xFF)
			{
				xaToSet[ ucIndex ].uiOffset = uiBaseOffset + pxVBElement->m_uOffset;
				xaToSet[ ucIndex ].ucElement = uElement;
				bBinded[ ucIndex ] = true;

				Asura_PS3_RenderingStates::SetVertexDataArray( ucIndex, usFrequency, ucVertexSize, pxVBElement->m_uNumberOfCoords, pxVBElement->m_uType, ucLoc, xaToSet[ ucIndex ].uiOffset );
			}
		}
	}

	if( Asura_PS3_Graphics::UseGCmMacros() )
	{
		CellGcmContextData* pxCommandBuffer = Asura_PS3_Graphics::GetGcmCurrentContext();
		uint32_t* puiCmd = pxCommandBuffer->current;
		//Set Vertex Attribute
		int iStart	= -1;
		int iStop	= -1;
		for( int i = 0; i < 17; ++i )
		{
			if( xaToSet[ i ].bAttributes )
			{
				if( iStart == -1 )
				{
					iStart	= i;
					iStop	= i;
				}
				else
				{
					iStop	= i;
				}
			}
			else
			{
				if( iStart != -1 )
				{
					//Set the data
					int iElements = iStop - iStart + 1;
					puiCmd[ 0 ] = CELL_GCM_METHOD( CELL_GCM_NV4097_SET_VERTEX_DATA_ARRAY_FORMAT + iStart * 0x04, iElements );
					for( int j = 0; j < iElements; ++j )
					{
						pxVBElement	= &pxDeclaration[ xaToSet[ iStart + j ].ucElement ];
						puiCmd[ j + 1 ] = CELL_GCM_METHOD_DATA_VERTEX_DATA_ARRAY_FORMAT( usFrequency, ucVertexSize, pxVBElement->m_uNumberOfCoords, pxVBElement->m_uType );
					}
					pxCommandBuffer->current += iElements + 1;
					puiCmd = pxCommandBuffer->current;
					//Reset
					iStart	= -1;
					iStop	= -1;
				}
			}
		}

		//Set Vertex Offset
		iStart	= -1;
		iStop	= -1;
		for( int i = 0; i < 17; ++i )
		{
			if( xaToSet[ i ].bOffset )
			{
				if( iStart == -1 )
				{
					iStart	= i;
					iStop	= i;
				}
				else
				{
					iStop	= i;
				}
			}
			else
			{
				if( iStart != -1 )
				{
					//Set the data
					int iElements = iStop - iStart + 1;
					puiCmd[ 0 ] = CELL_GCM_METHOD( CELL_GCM_NV4097_SET_VERTEX_DATA_ARRAY_OFFSET + iStart * 0x04, iElements );
					for( int j = 0; j < iElements; ++j )
					{
						pxVBElement	= &pxDeclaration[ xaToSet[ iStart + j ].ucElement ];
						puiCmd[ j + 1 ] = CELL_GCM_METHOD_DATA_VERTEX_DATA_ARRAY_OFFSET( ucLoc, xaToSet[ iStart + j ].uiOffset );
					}
					pxCommandBuffer->current += iElements + 1;
					puiCmd = pxCommandBuffer->current;
					//Reset
					iStart	= -1;
					iStop	= -1;
				}
			}
		}

		if( bClearUnusedVertsDataArray )
		{
			for( int i = 0; i < 17; ++i )
			{
				xaToSet[ i ].bAttributes	= false;
				xaToSet[ i ].bOffset		= false;
			}
			
			if ( pbKeep )
			{
				for( uElement = 0; uElement < 16; ++uElement )
				{
					if( !bBinded[ uElement ] && !pbKeep[ uElement ] )
					{
						if( Asura_PS3_RenderingStates::SetVertexAttributeNeeded( uElement, 0, 0, 0, CELL_GCM_VERTEX_F ) )
						{
							xaToSet[ uElement ].bAttributes	= true;
						}
					}
				}
			}
			else
			{
				for( uElement = 0; uElement < 16; ++uElement )
				{
					if( !bBinded[ uElement ] )
					{
						if( Asura_PS3_RenderingStates::SetVertexAttributeNeeded( uElement, 0, 0, 0, CELL_GCM_VERTEX_F ) )
						{
							xaToSet[ uElement ].bAttributes	= true;
						}
					}
				}
			}

			iStart	= -1;
			iStop	= -1;
			for( int i = 0; i < 17; ++i )
			{
				if( xaToSet[ i ].bAttributes )
				{
					if( iStart == -1 )
					{
						iStart	= i;
						iStop	= i;
					}
					else
					{
						iStop	= i;
					}
				}
				else
				{
					if( iStart != -1 )
					{
						//Set the data
						int iElements = iStop - iStart + 1;
						puiCmd[ 0 ] = CELL_GCM_METHOD( CELL_GCM_NV4097_SET_VERTEX_DATA_ARRAY_FORMAT + iStart * 0x04, iElements );
						for( int j = 0; j < iElements; ++j )
						{
							puiCmd[ j + 1 ] = CELL_GCM_METHOD_DATA_VERTEX_DATA_ARRAY_FORMAT( 0, 0, 0, CELL_GCM_VERTEX_F );
						}
						pxCommandBuffer->current += iElements + 1;
						puiCmd = pxCommandBuffer->current;
						//Reset
						iStart	= -1;
						iStop	= -1;
					}
				}
			}
		}
	}
	else
	{
		if( bClearUnusedVertsDataArray )
		{
			if ( pbKeep )
			{
				for( uElement = 0; uElement < 16; ++uElement )
				{
					if( !bBinded[ uElement ] && !pbKeep[uElement] )
					{
						Asura_PS3_RenderingStates::SetVertexDataArray( uElement, 0, 0, 0, CELL_GCM_VERTEX_F, CELL_GCM_LOCATION_LOCAL, 0 );
					}
				}
			}
			else
			{
				for( uElement = 0; uElement < 16; ++uElement )
				{
					if( !bBinded[ uElement ] )
					{
						Asura_PS3_RenderingStates::SetVertexDataArray( uElement, 0, 0, 0, CELL_GCM_VERTEX_F, CELL_GCM_LOCATION_LOCAL, 0 );
					}
				}
			}
		}
	}
}

/// End of file. ///
