//
// Asura_PS3_SPU_Skinning.cpp
//
// Author: Ben Weston 28/5/2008
//
// TEMPORARY! Really straightforward interface for running skinning code on SPUs
// Todo: use this and whatever Kevin does & Morgan's GeomCulling stuff as a basis for designing some kind of SPU task manager.
//


/* The model being followed here is:

spu DMA								|			spu processor
write processed vertices			|	process vertices
read next vertices					|

Since we know how many verts an SPU can take, we create a list of tasks and stick them in a queue.
But that's a pretty laborious way to do something we can do in the SPUs.
So ideally it should work like this:

data: static block of character data
task: process this ^ (pointer to that data)
spu: request next N vertices - update the read-from pointer shared between all processes.
	setup output address - update write-to pointer to where the end of the outputted data will be.

*/

//
// Includes
//

#include "Core/Asura.h"

// This file's header
#include "Asura_PS3_SPU_Skinning.h"

// Asura
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_CommandConsole.h"

// Asura Platform
#include "Animation/Asura_Platform_Animation_RenderCache.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile_PS3.h"
#include "PS3/PPU_SPU_UTILS/Asura_PS3_SPU_HelperStructs.h"
#include "PPU_SPU_Utils/Asura_PS3_SPUGroupController.h"
#include "Rendering/Asura_PS3_EdgeAntiAliasing.h"

// Platform
#include <cell/gcm.h>
#include <cell/sync/lfqueue.h>
#include <spu_printf.h>


#ifdef SPURS_SKINNING	//	See Asura_Engine_Configuration for appropriate project
	#include "Core/Memory/Asura_Mem.h"
	#include "PS3/PPU_SPU_UTILS/Asura_PS3_SPURS.h"
	#include "PS3/PPU_SPU_UTILS/Asura_PS3_SPURS_TaskController.h"
#endif

//
// Macros
//

#ifdef SPURS_SKINNING
	#define NUM_SPU_SKINNING_THREADS 4
#else
	#define NUM_SPU_SKINNING_THREADS 5

	#if (!defined(_DEBUG) && NUM_SPU_SKINNING_THREADS < 4)
	#error NUM_SPU_SKINNING_THREADS is low in non-debug build, did you mean to do that?
	#endif

#endif

#ifndef ASURA_PS3_MAX_SPU_SKINNING_TASKS
	#define ASURA_PS3_MAX_SPU_SKINNING_TASKS 64
#endif //#ifndef ASURA_PS3_MAX_SPU_SKINNING_TASKS

#if (0)
	#define DebugSpuSync(a) Asura_OutputToDebugger_Retail(a)
#else
	#define DebugSpuSync(a)
#endif


//
// Variables
//

u_int							Asura_PS3_SPU_Skinning::s_uWriteBuffer = 0;
int								Asura_PS3_SPU_Skinning::s_iRSXBufferReads = 0;
int								Asura_PS3_SPU_Skinning::s_iRSXReadyValue = 0;
bool							Asura_PS3_SPU_Skinning::s_bFirstCharacter = true;
bool							Asura_PS3_SPU_Skinning::s_bRanLastFrame = false;
bool							Asura_PS3_SPU_Skinning::s_bRanThisFrame = false;
volatile bool					Asura_PS3_SPU_Skinning::s_bStartedRunning = false;


#ifdef	SPURS_SKINNING
	extern const char _binary_task_Skinning_Taskset_elf_start[];	

	static Asura_PS3_SPURS_TaskController*	g_pxSPURSTaskController_Skinning = NULL;

	static	Skinning_TaskParameters	g_xSkinParams;

	static CellSpursBarrier g_xSkinningSPUBarrier;

#else
	extern u_int _binary_Skinning_SPU_elf_start;

	static Asura_PS3_SPUGroupController*	g_pxSPUThreadGroupInfo = NULL;

	static CellSyncBarrier g_xBarrier;

#endif

static const Asura_PS3_SPU_SkinningJob	g_xTerminatorJob(0);

static Asura_PS3_SPU_Skinning_StartupInfo g_xStartupInfo __attribute__((aligned(128)));

static bool g_bDoSpuSkinning = true;
static bool g_bHighlightSpuSkins = false;

//
// Functions
//


void Asura_PS3_SPU_Skinning::Shutdown( void )
{
#ifdef	SPURS_SKINNING
	g_pxSPURSTaskController_Skinning->End();

	delete g_pxSPURSTaskController_Skinning;
#else
	g_pxSPUThreadGroupInfo->End(); // just in case
	
	delete g_pxSPUThreadGroupInfo;
#endif

}

void Asura_PS3_SPU_Skinning::Initialise( void )
{
#ifdef	SPURS_SKINNING
	g_pxSPURSTaskController_Skinning = new Asura_PS3_SPURS_TaskController( NUM_SPU_SKINNING_THREADS, "Skin_Taskset", _binary_task_Skinning_Taskset_elf_start, sizeof(Asura_PS3_SPU_SkinningJob), ASURA_PS3_MAX_SPU_SKINNING_TASKS ); 
	
	g_pxSPURSTaskController_Skinning->SetTerminator( &g_xTerminatorJob );
#else
	g_pxSPUThreadGroupInfo = new Asura_PS3_SPUGroupController( "Skinning_SPU",
																&_binary_Skinning_SPU_elf_start,
																NUM_SPU_SKINNING_THREADS,
																Asura_PS3_SPUGroupController::PRIORITY_LOW,
																sizeof(Asura_PS3_SPU_SkinningJob), ASURA_PS3_MAX_SPU_SKINNING_TASKS );

	g_pxSPUThreadGroupInfo->SetTerminatorJob( &g_xTerminatorJob );
#endif
	s_uWriteBuffer = 0;
	s_iRSXBufferReads = 0;
	s_iRSXReadyValue = 0;

	s_bStartedRunning = false;
	s_bRanLastFrame = false;
	s_bRanThisFrame = false;
	
	
 #if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
 	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.Enable", &g_bDoSpuSkinning, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.SpuSkinning.Highlight", &g_bHighlightSpuSkins, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
 #endif

	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_0) = 0;
	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_1) = 0;
	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_0) = 0;
	*cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_1) = 0;
	
#ifdef	SPURS_SKINNING
	cellSpursBarrierInitialize( g_pxSPURSTaskController_Skinning->GetTaskset()->GetTaskset(), &g_xSkinningSPUBarrier, NUM_SPU_SKINNING_THREADS /* Change to reflect what is set in SPURS */ );
#else
	cellSyncBarrierInitialize( &g_xBarrier, NUM_SPU_SKINNING_THREADS );
#endif
}


void Asura_PS3_SPU_Skinning::BeginFrame( void )
{
	s_uWriteBuffer = 1-s_uWriteBuffer;
	s_bFirstCharacter = true;
	
	//RSX read flag, set by RSX once it's finished reading each vertex buffer
	//Wait for the last read
	s_iRSXReadyValue = s_iRSXBufferReads;
	s_iRSXBufferReads = 0;

	u_int uLabel = s_uWriteBuffer?ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_1:ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_0;
	*cellGcmGetLabelAddress(uLabel) = 0xfeedbac2;

	s_bRanLastFrame = s_bRanThisFrame;
	s_bRanThisFrame = false;
}
	
	
// check that ProcessSkin will run correctly, but don't actually run it
//BW this function used to check lots of things because it locally handled caching of bones & output vertices, now it does very little.
bool Asura_PS3_SPU_Skinning::IsSpuable(const Asura_PS3_CharacterRenderer_Skin* pxCharacter)
{
	if ( !g_bDoSpuSkinning )
	{
		return false;
	}

	if ( pxCharacter->m_xVertexBuffer.GetSize() != sizeof(Asura_PS3_SPU_Skinning_InputVertex) )
	{
		Asura_Assert( 0, "Vertex size in buffer is wrong! SPU skinning can't handle this model, deactivating" );
		return false;
	}

	u_int uNumQueuedJobs = 0;
#ifdef	SPURS_SKINNING
	cellSpursLFQueueSize( g_pxSPURSTaskController_Skinning->GetQueue(), &uNumQueuedJobs );
#else
	cellSyncLFQueueSize( g_pxSPUThreadGroupInfo->GetQueue(), &uNumQueuedJobs );
#endif
	if ( uNumQueuedJobs >= ASURA_PS3_MAX_SPU_SKINNING_TASKS )
	{
		Asura_Assert( 0, "Ran out of SPU skinning tasks! Increase ASURA_PS3_MAX_SPU_SKINNING_TASKS." );
		return false;
	}

	return true;
}



//
// ProcessSkin
//
// Ben Weston June '08
//
// Add a character's skin to the SPU task list, and return a pointer to where the vertex data will be.
// This assumes you're about to add the draw strip commands, so this puts a wait label in the RSX command buffer, where it will wait for SPU skinning to complete.
//
// pOutputVertexBuffer - vertex buffer for the output vertices - for SPU to write to, and RSX to read from //BW todo: cull vertices so we can have less [+implement a cap so don't need to alloc the full space]
//						 at present this should be the same size as the input vertices.
//

static 	Asura_PS3_SPU_SkinningJob xJob  __attribute__(( aligned(16) ));

bool Asura_PS3_SPU_Skinning::ProcessSkin(	const Asura_PS3_CharacterRenderer_Skin* pxCharacter,
											Asura_PS3_SPU_SkinningConstants* pxShaderConstants,
											Asura_Platform_Animation_RenderCache* pxRenderCache,
											u_int uSkin, u_int uNumVerts, SPU_SKINNING_MODE eMode )
{
	
	if ( !IsSpuable(pxCharacter) )
	{
		Asura_Assert(0,"ProcessSkin: IsSpuable() is false, this function should not be called!");
		return false;
	}

	if ( s_bFirstCharacter && s_bRanLastFrame )
	{
		WaitForSPUs();
		//BW do this as late as possible, but before any of the input data changes for this frame.
		//BW actually this might be too late, since the first character's bone cache has already been updated. But provided SPUs have started by now that won't cause a problem
	}


	//
	// Sync with RSX
	//

	// ensure RSX doesn't draw until SPUs have finished (this is unlikely to happen)
	//BW I was trying to make it sync per task but that's not practical, plus the SPUs should be finished well before the RSX starts, so just wait for the last task to finish.
	u_int uLabel = s_uWriteBuffer?ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_1:ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_0; //double buffer since our SPUs start running (and possibly finish) 1 frame before RSX reads it
	u_int* pLabel = cellGcmGetLabelAddress(uLabel);

	if ( s_bFirstCharacter )
	{
 		*pLabel = 0; // init to something other than uSPU_SKINNING_RSX_FLAG

		//make RSX wait for SPU to complete before it draws the character (don't do this in BeginFrame, in case we don't draw any chars!)
		cellGcmSetWaitLabelInline(Asura_PS3_Graphics::GetGcmCurrentContext(),uLabel,uSPU_SKINNING_RSX_FLAG);
	}

	// Get the vertex output buffer
	Asura_PS3_GCM_VertexBuffer* pOutputVertexPosBuffer = pxRenderCache->ObtainVertexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS,uSkin,uNumVerts,sizeof(Asura_PS3_SPU_Skinning_OutputVertex_PosOnly));

	//Sanity checks: make sure vertex buffer is in RSX-mapped memory (otherwise we can't render from it)
	//Note we map all of main memory to the RSX, but on devkits it's possible to exceed the 256MB limit of mappable memory.
#ifdef ASURA_ASSERT
	{
		u_int uTestOffset;
		char* pTestBuffer = reinterpret_cast<char*>(pOutputVertexPosBuffer->LockForWriting());
		Asura_Assert( cellGcmAddressToOffset(pTestBuffer, &uTestOffset) == CELL_OK, "Asura_PS3_SPU_Skinning::ProcessSkin - CANNOT DRAW SKIN! Vertex buffer is not in memory mapped to RSX." );
	}
#endif	
	
	// Create SPU Job
	xJob.m_uVertPosOutAddr = reinterpret_cast<uint64_t>(pOutputVertexPosBuffer->LockForWriting()); //pointer to output vertex positions
	xJob.m_uAAIndicesOutAddr = 0;

	if ( eMode == SPU_SKINNING_POS_ONLY_NO_CULL )
	{
		xJob.m_uInputAddr = reinterpret_cast<uint64_t>(pxCharacter->m_xVertexBuffer.GetReadPtr());
	}
	else
 	{
		int iNumStrips = pxCharacter->GetNumStrips();
		Asura_PS3_SPU_Skinning_Batch* pxBatches = pxRenderCache->GetBatches( uSkin, iNumStrips );
		
		const u_short* pusIndexBuffer = static_cast<const u_short*>(pxCharacter->m_xIndexBuffer.GetReadPtr());
		const Asura_PS3_SPU_Skinning_InputVertex* pxVertexBuffer = static_cast<const Asura_PS3_SPU_Skinning_InputVertex*>(pxCharacter->m_xVertexBuffer.GetReadPtr());
		
		for ( int iStrip = 0; iStrip < iNumStrips; iStrip++ )
		{
			Asura_PS3_SPU_Skinning_Batch& xBatch = pxBatches[iStrip];
			
			const Asura_PS3_CharacterRenderer_Strip& xStrip = pxCharacter->GetStrip( iStrip );

			xBatch.m_uStartIndexAddr = reinterpret_cast<uint64_t>(&pusIndexBuffer[xStrip.m_uStartIndex]);
			xBatch.m_uNumIndices = xStrip.m_uNumIndices;
			
			// assume the first vertex referenced by the strip is the 1st one we need (this is how the tools setup the data).
			u_int uFirstVert = pusIndexBuffer[xStrip.m_uStartIndex];
			xBatch.m_uVertAddr = reinterpret_cast<uint64_t>(&pxVertexBuffer[uFirstVert]);
			
			if ( iStrip < iNumStrips-1 )
			{
				u_int uNextFirstVert = pusIndexBuffer[ xStrip.m_uStartIndex + xStrip.m_uNumIndices ]; //the first vertex of the next strip
				xBatch.m_uNumVerts = uNextFirstVert - uFirstVert;
			}
			else
			{
				xBatch.m_uNumVerts = uNumVerts - uFirstVert;
			}
//			xBatch.m_uCommand = um;

			Asura_Assertf( xStrip.m_uNumIndices <= ASURA_PS3_SPUSKINNING_NUM_INDICES,("Too many indices in batch for SPU. %d indices - max is %d\n", xStrip.m_uNumIndices, ASURA_PS3_SPUSKINNING_NUM_INDICES));
			Asura_Assertf( xBatch.m_uNumVerts <= ASURA_PS3_SPUSKINNING_NUM_VERTICES,("Too many verts in batch for SPU. %d verts - max is %d\n", xBatch.m_uNumVerts, ASURA_PS3_SPUSKINNING_NUM_VERTICES ));

// 			#define NUM_TRIS_IN_BATCH	((ASURA_PS3_SPUSKINNING_BUFFER_SIZE-0xf)/((64+2)*3)) //duplicated from tools, sanity check, should definitely be impossible if chunk had correct character version (version 6 or later)
// 			Asura_Assertf( xStrip.m_uNumIndices <= NUM_TRIS_IN_BATCH*3, ("Too many triangles in data, somehow the tools didn't export this skin correctly! %.1f triangles (%d max)\n", ((float)xStrip.m_uNumIndices)/3, NUM_TRIS_IN_BATCH ) );
// 
// 			//BW Test code - this is surprisingly slow. But good for spotting if tools have broken a rule I'm assuming here.
// 			// sanity check the assumption about how strips are arranged (that the first vert indexed will be the first vert in the vert buffer)
// 			// n.b. if these fire someone's probably messed with the tools code (which sorts vertex data by first use, hence this assumption will always be true)
// 			for ( u_int i=xStrip.m_uStartIndex; i < xStrip.m_uStartIndex+xStrip.m_uNumIndices; i++ )
// 			{
// 				Asura_Assertf(pusIndexBuffer[i] >= uFirstVert,("Vertex needed by strip is not in batch sent to SPU! Culling will go mad\n%d >= %d\nbatch %d index %d localindex %d",
// 																pusIndexBuffer[i], uFirstVert, iStrip, i, i - xStrip.m_uStartIndex ));
// 				Asura_Assertf(pusIndexBuffer[i] - uFirstVert < xBatch.m_uNumVerts,("Vertex needed by strip is not in batch sent to SPU! Culling will go mad\n%d-%d < %d\nbatch %d index %d localindex %d",
// 																						pusIndexBuffer[i], uFirstVert, xBatch.m_uNumVerts, iStrip, i, i - xStrip.m_uStartIndex ));
// 			}
		}
		
		pxBatches[iNumStrips].m_uNumVerts = 0;
		pxBatches[iNumStrips].m_uNumIndices = 0;

		Asura_PS3_GCM_IndexBuffer* pOutputIndexBuffer = pxRenderCache->ObtainIndexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED, uSkin, pxCharacter->GetIndexBuffer()->GetIndexBufferSize()/2);

#ifdef ASURA_ASSERT
		{
			u_int uTestOffset;
			char* pTestBuffer = reinterpret_cast<char*>(pOutputIndexBuffer->LockForWriting());
			Asura_Assert( cellGcmAddressToOffset(pTestBuffer, &uTestOffset) == CELL_OK, "Asura_PS3_SPU_Skinning::ProcessSkin - CANNOT DRAW SKIN! Index buffer is not in memory mapped to RSX." );
		}
#endif
		
#ifdef ASURA_EDGE_ANTIALIASING
		if ( Asura_PS3_EdgeAntiAliasing::Enabled() ) //don't waste time finding the edges if AA's turned off.
		{
			Asura_PS3_GCM_IndexBuffer* pOutputAAIndexBuffer = pxRenderCache->ObtainIndexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_EDGEAA, uSkin, SPU_SKINNING_MAX_AA_EDGES*2);

			xJob.m_uAAIndicesOutAddr = reinterpret_cast<uint64_t>(pOutputAAIndexBuffer->LockForWriting());

 			Asura_PS3_EdgeAntiAliasing::AddEdgeList( pOutputAAIndexBuffer, pOutputVertexPosBuffer, &pxShaderConstants->m_xWorldViewProjectionMatrix );
		}
#endif

		xJob.m_uInputAddr = reinterpret_cast<uint64_t>(pxBatches);
		xJob.m_uIndicesOutAddr = reinterpret_cast<uint64_t>(pOutputIndexBuffer->LockForWriting());
	} // if ( eMode == SPU_SKINNING_POS_ONLY_NO_CULL ) ... else ...
	
	xJob.m_eMode = eMode;


	// Shader constants
	Asura_Mem::Copy( &xJob.m_xShaderConstants, pxShaderConstants, sizeof(Asura_PS3_SPU_SkinningConstants) );


	if ( g_bHighlightSpuSkins )
	{
		// mess with the matrix to make SPU skins look different
		xJob.m_xShaderConstants.m_uBoneArrayAddr += 64;
	}

	//write this last to avoid SPUs ever getting nonsense data (though if they're reading at this point some of them might think they're finished when some others keep running)
	xJob.m_uNumVerts = uNumVerts;


	u_int uSize = 0;
#ifdef	SPURS_SKINNING
	cellSpursLFQueueSize( g_pxSPURSTaskController_Skinning->GetQueue(), &uSize );
#else
	cellSyncLFQueueSize( g_pxSPUThreadGroupInfo->GetQueue(), &uSize );
#endif
	if ( uSize < ASURA_PS3_MAX_SPU_SKINNING_TASKS )
	{
#ifdef	SPURS_SKINNING
		g_pxSPURSTaskController_Skinning->PushTask_Blocking( &xJob );
#else
		cellSyncLFQueuePush( g_pxSPUThreadGroupInfo->GetQueue(), &xJob );
#endif
	}
	else
	{
		Asura_Assert( 0, "Ran out of SPU skinning tasks! Increase ASURA_PS3_MAX_SPU_SKINNING_TASKS." );
		return false;
	}


	//
	// Start the SPUS
	//

	if ( s_bFirstCharacter )
	{
		s_bFirstCharacter = false;
		TryKickSPUs();
	}


	// update flags
	if ( eMode == SPU_SKINNING_POS_AND_CULL )
	{
		pxRenderCache->SetIndexCacheCurrent( ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED, uSkin );
#ifdef ASURA_EDGE_ANTIALIASING
		if ( Asura_PS3_EdgeAntiAliasing::Enabled() )
		{
			pxRenderCache->SetIndexCacheCurrent( ASURA_PS3_RENDERCACHE_BUFFER_EDGEAA, uSkin );
		}
#endif
	}
	pxRenderCache->SetVertexCacheCurrent( ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS, uSkin );

	return true;
}

void Asura_PS3_SPU_Skinning::EndFrame( void )
{
	ForceKickSPUs();

	if ( !s_bRanThisFrame )
	{
		if ( s_bRanLastFrame )
		{
			WaitForSPUs(); //but we did draw on the previous frame, make sure SPUs finish that (previous) frame's work
		}

		return;
	}

#ifdef	SPURS_SKINNING
	g_pxSPURSTaskController_Skinning->PushTerminators();
#else
	g_pxSPUThreadGroupInfo->PushTerminators(); // spus will take up time until this is done. - so do it as soon as we know there's no more characters
#endif

//BW debug - force spus to finish by end of frame	
// WaitForSPUs(); //but we did draw on the previous frame, make sure SPUs finish that (previous) frame's work
// s_bRanThisFrame=false;
}

//
// Asura_PS3_SPU_Skinning::NotifyBufferUsed
//
// Call this whenever you've added render calls that read from a buffer filled by Asura_PS3_SPU_Skinning::ProcessSkin
// the SPU skinning uses this for synchronisation
//
void Asura_PS3_SPU_Skinning::NotifyBufferUsed( void )
{
	s_iRSXBufferReads++;

	u_int uLabel = s_uWriteBuffer?ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_1:ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_0;

	cellGcmSetWriteTextureLabelInline(
		Asura_PS3_Graphics::GetGcmCurrentContext(),
		uLabel, //note buffer ID's flipped, since we're making next-frame's SPUs wait for this
		s_iRSXBufferReads); //this label increments each time any of our vertex buffers is used for rendering (since we don't know which one's used last until the end of the frame), then we can just wait for the last usage

	DebugSpuSync(( "Told RSX to write %d to 0x%08x\n", s_iRSXBufferReads, cellGcmGetLabelAddress(uLabel) ));
}


// kick the SPUs, after the RSX has finished rendering the skinned data from the previous frame
void Asura_PS3_SPU_Skinning::KickSPUs( bool bForce )
{
	if ( s_bFirstCharacter ) return;	//don't kick if there's no data.
	

	// check if the SPUs are already running
	u_int uSpuToRsxLabel = s_uWriteBuffer?ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_1:ASURA_GCM_LABEL_SPUSKINNING_WAITFORSPU_0; //double buffer since our SPUs start running (and possibly finish) 1 frame before RSX reads it
	u_int* pSpuToRsxLabel = cellGcmGetLabelAddress(uSpuToRsxLabel);

	if ( s_bStartedRunning	// we have started them
		&& *pSpuToRsxLabel != uSPU_SKINNING_RSX_FLAG )	// they haven't finished (they set this flag when they finish)
	{
		// no need to kick them, they're already running and will pick up any new data.
		return;

		//BW this can go wrong, if some of the SPUs have finished and are waiting for the others then only the ones still running will process successive data.
		//BW I doubt if this will happen enough to cause visible slow-down
	}
	else
	{
		s_bStartedRunning = false;
	}

	
	s_bRanThisFrame = true; // set this as soon as we try to kick the SPUs, even if we don't succeed this time (they will be kicked at the end of the frame)


	// wait for the RSX to finish reading last frame's data
	u_int uRsxToPpuLabel = s_uWriteBuffer?ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_0:ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_1; //INVERTED! we're checking last frame's flag.
	u_int* pRsxToPpuLabel = cellGcmGetLabelAddress(uRsxToPpuLabel);
	if ( s_iRSXReadyValue > 0 ) // if we drew anything from our vertex buffers last time
	{
		Asura_TickCount xStartTicks = Asura_Timers::GetCurTicks();

		if ( bForce )
		{
			DebugSpuSync(( "Waiting for RSX to write %d to 0x%08x... ", s_iRSXReadyValue, pRsxToPpuLabel ));
		}
		else
		{
			DebugSpuSync(( "Checking RSX to write %d to 0x%08x... ", s_iRSXReadyValue, pRsxToPpuLabel ));
		}
	
		while ( bForce && *pRsxToPpuLabel != (u_int)s_iRSXReadyValue
				&& Asura_Timers::GetElapsedTicks(xStartTicks) < Asura_Timers::SecondsToTicks(3.0f) )
		{
			// wait for the RSX to finish with SPU buffers
			sys_timer_usleep( 1000 );

#ifdef ASURA_USE_GCM_REPLAY
			if ( cell::Gcm::Replay::Capture::IsCurrentlyCapturing() )
			{
				// prevent timeout during GPAD captures
				xStartTicks = Asura_Timers::GetCurTicks();
			}		
#endif
		}

		if ( *pRsxToPpuLabel != (u_int)s_iRSXReadyValue )
		{
			DebugSpuSync(( "failure\n" ));
		
			Asura_Assert( !bForce, "Asura_PS3_SPU_Skinning::KickSPUs - wait for RSX timed out.\n" );
			return;
		}
		else
		{
			DebugSpuSync(( "SUCCESS!\n" ));
		}
	}
	
	// start the spus
	Asura_Assert(!s_bStartedRunning,"SPUs already running"); // sanity check

	g_xStartupInfo.m_auScissorArea[0] = 0;
	g_xStartupInfo.m_auScissorArea[1] = 0;
	g_xStartupInfo.m_auScissorArea[2] = Asura_PS3_Graphics::GetScreenWidth();
	g_xStartupInfo.m_auScissorArea[3] = Asura_PS3_Graphics::GetScreenHeight();

#ifdef	SPURS_SKINNING
	g_xStartupInfo.m_uBarrierAddr		= reinterpret_cast<uint64_t>(&g_xSkinningSPUBarrier);
#else
	g_xStartupInfo.m_uBarrierAddr		= reinterpret_cast<uint64_t>(&g_xBarrier);
#endif

	uint64_t uLabelAddr = reinterpret_cast<uint64_t>(pSpuToRsxLabel);
	uint64_t uStartupInfoAddr = reinterpret_cast<uint64_t>(&g_xStartupInfo);

	DebugSpuSync(( "Telling SPUs to flag at 0x%08x\n", pSpuToRsxLabel ));
	
#ifdef	SPURS_SKINNING
	g_xSkinParams.uLabelAddr	= uLabelAddr;
	g_xSkinParams.xStartupInfo	= uStartupInfoAddr;
	g_pxSPURSTaskController_Skinning->Begin( &g_xSkinParams );
#else
	g_pxSPUThreadGroupInfo->Begin(
		NULL, // will be replaced with SPU num
		NULL, // will be replaced with pointer to queue
		&uLabelAddr,
		&uStartupInfoAddr
		);
#endif

	s_bStartedRunning = true;
}


void Asura_PS3_SPU_Skinning::WaitForSPUs()
{
	if ( !s_bRanLastFrame ) return;

	s_bRanLastFrame = false; // so we can detect if we wait twice

	// make sure SPUs have actually started!
	Asura_TickCount xStartTicks = Asura_Timers::GetCurTicks();

	if ( !s_bStartedRunning )
	{

		//This assert is a sanity check
		//Previously this indicated something got stuck, but with changes to SPU kicking code it is now logically impossible to get in here!
		Asura_Assert(0, "Asura_PS3_SPU_Skinning::WaitForSPUs - this assert should not fire, was some logic changed? This is probably fatal" );

		//Previous code, for posterity:
		//Asura_OutputToDebugger_Retail(("WaitForSPUs - SPU kicker thread seems to be stuck. RSX wait label is 0x%08x or \n", cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_0), cellGcmGetLabelAddress(ASURA_GCM_LABEL_SPUSKINNING_WAITFORRSX_1)));
		//ASURA_PS3_DEBUGBREAK();
		// this would happen if the kicker thread is stuck waiting for the RSX, which is waiting for the SPU
		// but the only time I've seen this has been when I've erroneously called this function when SPUs weren't meant to be running (i.e. wait for something that's not happening!)
 		//return;
	}


	// wait for SPUs to finish
#ifdef	SPURS_SKINNING
	g_pxSPURSTaskController_Skinning->End();
#else
	g_pxSPUThreadGroupInfo->End();
#endif

	s_bStartedRunning = false;
}


void Asura_PS3_SPU_Skinning::SetSPUSkinningEnabled(bool bEnable)
{
	g_bDoSpuSkinning = bEnable;
}

bool Asura_PS3_SPU_Skinning::GetSPUSkinningEnabled()
{
	return g_bDoSpuSkinning;
}
