//
// Asura_PS3_SwapChain.cpp
//
// Created by Catalin Arsenescu - 29 April 2008
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PS3_SwapChain.h"

// Stdlib platform-specific.
#include <stdlib.h>
#include <sdk_version.h>
#include <sysutil/sysutil_sysparam.h>
#include <cell/sysmodule.h>
#if VIDEO_OUT_CONFIGURE_NON_BLOCK
	#include <sys/timer.h>
#endif //#if VIDEO_OUT_CONFIGURE_NON_BLOCK
#ifdef ASURA_USE_GCM_REPLAY
	#include <netex/net.h>
	#include <netex/libnetctl.h>
#endif

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_CPUProfile.h"
#if defined(ASURA_USE_PROFILING_SYSTEM)
#include "Debugging/Asura_CPUProfile_Renderer.h"
#endif
#include "Debugging/Asura_Logfile.h"
#if defined(ASURA_USE_SCREENSHOT_SYSTEM)
#include "Debugging/Asura_ScreenShot_System.h"
#endif
#include "Debugging/Asura_PS3_CommandBufferInspector.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// Engine platform-specific.
#include "Debugging/Asura_PS3_DebugInfo.h"
#include "Rendering/Asura_PS3_DeferredRendering.h"
#include "Rendering/Asura_PS3_DeferredShading.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_Helper.h"
#include "Rendering/Asura_PS3_Projector_System.h"
#include "Rendering/Asura_PS3_QuadOverlayRenderer.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"
#include "Rendering/Asura_PS3_ScratchRenderMemory.h"
#include "Rendering/Asura_PS3_TextureManagement.h"
#include "Rendering/Asura_PS3_LineRenderer.h"

volatile u_int		Asura_PS3_SwapChain::s_uFlipCounter				= 0;
volatile u_int		Asura_PS3_SwapChain::s_uVSyncCounter			= 0;
volatile u_int		Asura_PS3_SwapChain::s_uVSyncNext				= 0;
volatile u_int		Asura_PS3_SwapChain::s_uLastFlipId				= 0xF;
volatile u_int		Asura_PS3_SwapChain::s_uFinishRC				= 0;

u_int*				Asura_PS3_SwapChain::s_pxDefaultCommandBuffer	= NULL;				// small ring buffer
u_int*				Asura_PS3_SwapChain::s_paxCommandBuffer[2]		= { NULL, NULL };	// large per-frame command buffers
u_int				Asura_PS3_SwapChain::s_uWhich					= 0;
u_int				Asura_PS3_SwapChain::s_uTotalSize				= 0;
u_int				Asura_PS3_SwapChain::s_uRingBufferSize			= 0;
u_int				Asura_PS3_SwapChain::s_uCommandBufferSize		= 0;
u_int				Asura_PS3_SwapChain::s_uLastFlipCounter			= 0;
u_int				Asura_PS3_SwapChain::s_uPresentationInterval	= 1;
CellGcmContextData	Asura_PS3_SwapChain::s_xCurrentCmdCtxt;			// current command buffer context
sys_lwmutex_t		Asura_PS3_SwapChain::s_xMutex;					// prevent the render thread and the interrupt thread from colliding!
bool				Asura_PS3_SwapChain::s_bIsInFMV					= false;


static const u_char		g_ucFlipLabel	= 69;
static const usecond_t	g_xSleepTime	= 30;

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_SwapChain::Initialise
// Author       : Catalin Arsenescu (29-April-2008)
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_SwapChain::Initialise( void )
{
	#ifdef ASURA_PS3_PRESENT_INTERVAL
		s_uPresentationInterval = ASURA_PS3_PRESENT_INTERVAL;
	#else
		s_uPresentationInterval = 2;
	#endif

	#ifdef ASURA_PS3_RING_BUFFFER_SIZE
		s_uRingBufferSize = ASURA_PS3_RING_BUFFFER_SIZE;
	#else
		s_uRingBufferSize = 192 * 1024;
	#endif

	#ifdef ASURA_PS3_COMMAND_BUFFFER_TOTAL_MB_SIZE
		s_uTotalSize = ASURA_PS3_COMMAND_BUFFFER_TOTAL_MB_SIZE * 1024 * 1024; 
	#else
		s_uTotalSize = 4 * 1024 * 1024; // 2MB per command buffer, double buffered command buffer
	#endif
	s_uCommandBufferSize		= ( s_uTotalSize - s_uRingBufferSize ) / 2;	//double buffered command buffer
	s_uWhich					= 0;
	// create a mutex to control access to the flips
	sys_lwmutex_attribute_t xAttr;
	sys_lwmutex_attribute_initialize( xAttr );
	xAttr.attr_recursive = SYS_SYNC_RECURSIVE;
	sys_lwmutex_create( &s_xMutex, &xAttr );
}

void Asura_PS3_SwapChain::Create( void )
{
	s_pxDefaultCommandBuffer = static_cast< u_int* >( memalign( 1048576, s_uTotalSize ) );
	if( s_pxDefaultCommandBuffer == NULL )
	{
		Asura_OutputToDebugger( ( "Failed to allocate memory for the command buffer!\n" ) );
	}
	s_paxCommandBuffer[ 0 ]		= ( u_int* )( ( char* )s_pxDefaultCommandBuffer + s_uRingBufferSize );
	s_paxCommandBuffer[ 1 ]		= ( u_int* )( ( char* )s_paxCommandBuffer[ 0 ] + s_uCommandBufferSize );

#if defined(ASURA_USE_GCM_REPLAY) && defined(ASURA_SUPPORT_EXTERNAL_NETWORK)
	// GCM HUD is implicitly initialised in cellGcmInit, and if we init our network after that it all goes wrong.
	cellSysmoduleLoadModule(CELL_SYSMODULE_NET);
	sys_net_initialize_network();
	cellNetCtlInit();
#endif

	if( cellGcmInit( s_uRingBufferSize, s_uTotalSize, s_pxDefaultCommandBuffer ) != CELL_OK )
	{
		Asura_OutputToDebugger( ("cellGcmInit -> FAILED\n") );
		return;
	}

	// clear the label we'll use for synchronizing the RSX and vsync
	cellGcmSetWriteBackEndLabelInline( gCellGcmCurrentContext, g_ucFlipLabel, 0xFFFFFFFF );

	// also make sure the RSX updates it
	cellGcmFlush( gCellGcmCurrentContext );
	++s_uFinishRC;
	cellGcmSetReferenceCommand( gCellGcmCurrentContext, s_uFinishRC );
	cellGcmFinish( gCellGcmCurrentContext, s_uFinishRC );

	// set our flip and vsync handlers
	cellGcmSetFlipHandler( &Asura_PS3_SwapChain::FlipHandler );
	cellGcmSetVBlankHandler( &Asura_PS3_SwapChain::VSyncHandler );

	// flip immediate still respects this
	cellGcmSetFlipMode(CELL_GCM_DISPLAY_HSYNC);

	// setup our current context, we are ready to render!
	cellGcmSetupContextData( &s_xCurrentCmdCtxt, s_paxCommandBuffer[ s_uWhich ], s_uCommandBufferSize, &CommandBufferCallback );
	gCellGcmCurrentContext = &s_xCurrentCmdCtxt;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PS3_SwapChain::Shutdown
// Author       : Catalin Arsenescu (29-April-2008)
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_SwapChain::Shutdown( void )
{
	free( s_pxDefaultCommandBuffer );
}

Asura_PS3_SwapChain::Asura_PS3_SwapChain( void )
{
}

Asura_PS3_SwapChain::~Asura_PS3_SwapChain( void )
{
}

#ifdef ASURA_ASSERT
void Asura_PS3_SwapChain::SetDefaultCommandBuffer( void )
{
	cellGcmSetDefaultCommandBuffer();
}

void Asura_PS3_SwapChain::RestoreCommandBuffer( void )
{
	gCellGcmCurrentContext = &s_xCurrentCmdCtxt;
}

void Asura_PS3_SwapChain::PresentAssert( u_int uDisplayID )
{
	// don't get ahead more than 1 frame!
	BlockOnFlip( 1 );
	++s_uLastFlipCounter;

	// insert a prepare flip (get as much done as possible before vsync)
	int32_t uFlipId = cellGcmSetPrepareFlip( gCellGcmCurrentContext, uDisplayID );

	// mark it so we can wait on the GPU to finish up to here
	cellGcmSetWriteBackEndLabelInline( gCellGcmCurrentContext, g_ucFlipLabel, uFlipId );

	// get it out there, this actually restarts the RSX again!
	cellGcmFlush( Asura_PS3_Graphics::GetGcmCurrentContext() );
	// No GCM Replay support if we are in an assert
}

#endif

//Before calling this function make sure you lock the graphics mutex
void Asura_PS3_SwapChain::BlockUntilIdle( void )
{
	// we need this for proper resource unloading...
	// do we want to present this? we could have some ugly half frame?!
	// no that is not true! we will always be in the system request section

	// all flips should be done! we need to wait a bit
	// we use > 1 because the last flip has not been pushed yet!
	BlockOnFlip(1);

	// switch back to default command buffer
	cellGcmSetDefaultCommandBuffer();
	cellGcmFlush( gCellGcmCurrentContext );
	gCellGcmCurrentContext = &s_xCurrentCmdCtxt;
}

void Asura_PS3_SwapChain::Present( u_int uDisplayID )
{
	// don't get ahead more than 1 frame!
	BlockOnFlip( 1 );
	++s_uLastFlipCounter;

	// switch back to default command buffer
	cellGcmSetDefaultCommandBuffer();

	// make sure we have room to write our flip commands (always 4 bytes each?)
	// we are writing 3 (jump, prepare, label)
	static const u_int suFlipSize = 4 * 5;
	if( ( gCellGcmCurrentContext->current + suFlipSize ) > gCellGcmCurrentContext->end )
	{
		gCellGcmCurrentContext->callback( gCellGcmCurrentContext, suFlipSize );
	}

	// get the *next* location in the default command buffer
	// so the deferred can jump back to it
	u_int uOffset;
	cellGcmAddressToOffset( cellGcmGetCurrentBuffer() + 1, &uOffset );
	cellGcmSetJumpCommandInline( &s_xCurrentCmdCtxt, uOffset);

	// insert a jump to the deferred command buffer in the default
	cellGcmAddressToOffset( s_paxCommandBuffer[ s_uWhich ], &uOffset );
	cellGcmSetJumpCommandInline( gCellGcmCurrentContext, uOffset );

	// insert a prepare flip (get as much done as possible before vsync)
	int32_t uFlipId = cellGcmSetPrepareFlip( gCellGcmCurrentContext, uDisplayID );

	// mark it so we can wait on the GPU to finish up to here
	cellGcmSetWriteBackEndLabelInline( gCellGcmCurrentContext, g_ucFlipLabel, uFlipId );

	// get it out there, this actually restarts the RSX again!
	cellGcmFlush( Asura_PS3_Graphics::GetGcmCurrentContext() );

	// next frame please!
	s_uWhich = s_uWhich ^ 1;

	// start a new command buffer
	cellGcmSetupContextData( &s_xCurrentCmdCtxt, s_paxCommandBuffer[ s_uWhich ], s_uCommandBufferSize, &CommandBufferCallback );
	gCellGcmCurrentContext = &s_xCurrentCmdCtxt;
}

void Asura_PS3_SwapChain::FlipHandler( const u_int Value )
{
	++s_uFlipCounter;
}

void Asura_PS3_SwapChain::VSyncHandler( const u_int Value )
{
	// bump the counter
	++s_uVSyncCounter;

	// try and flip if we can
	TryFlipImmediate();
}

void Asura_PS3_SwapChain::TryFlipImmediate( void )
{
	int Ret = sys_lwmutex_trylock( &s_xMutex );
	if( Ret == CELL_OK )
	{
		// unpack the RSX label
		u_int Data		= *( volatile u_int* )cellGcmGetLabelAddress( g_ucFlipLabel );
		u_int FlipId	= Data & 0xF;	// 0-7 valid

		// handle first frame and re-entry before RSX updates the label
		if( ( FlipId < 8 )  && ( FlipId != s_uLastFlipId ) )
		{
			// check our presentation interval
			if( s_uVSyncCounter >= s_uVSyncNext )
			{
				// save the current flip so we don't try again
				s_uLastFlipId = FlipId;

				// flip it right now (this does not touch the command buffer)
				cellGcmSetFlipImmediate( FlipId );

				// we can flip again in X vsyncs
				s_uVSyncNext = s_uVSyncCounter + s_uPresentationInterval;
			}
		}
		// done with the mutex
		sys_lwmutex_unlock( &s_xMutex );
	}
	else
	{
		// better be busy! else we are messed
		if( Ret != EBUSY )
		{
			Asura_OutputToDebugger( ( "Warning: Flip immediate not busy" ) );
		}
	}
}

void Asura_PS3_SwapChain::BlockOnFlip( unsigned int FlipDelta )
{
#if defined(ASURA_PS3_USE_COMMAND_BUFFER_INSPECTOR)
	Asura_TickCount xStartTicks = Asura_Timers::GetCurTicks();
	static bool bInspected = false;
#endif

	while( ( s_uLastFlipCounter - s_uFlipCounter ) > FlipDelta )
	{
		sys_timer_usleep( g_xSleepTime );
		if( !s_bIsInFMV )
		{
			// the RSX might have missed its VSYNC by a little bit
			TryFlipImmediate();
		}

#if defined(ASURA_PS3_USE_COMMAND_BUFFER_INSPECTOR)

#if defined(ASURA_USE_GCM_REPLAY)
		if ( cell::Gcm::Replay::Capture::IsCurrentlyCapturing() )
		{
			// prevent timeout during GPAD captures
			xStartTicks = Asura_Timers::GetCurTicks();
		}
#endif

		if ( !bInspected && Asura_Timers::TicksToSeconds(Asura_Timers::GetElapsedTicks(xStartTicks)) > 5.0f )
		{
			Asura_OutputToDebugger_Retail(("\n*************************************************************************************\n"));
			Asura_OutputToDebugger_Retail((  "Asura_PS3_SwapChain::BlockOnFlip - RSX seems to be stuck, analysing command buffer...\n"));
			Asura_OutputToDebugger_Retail((  "*************************************************************************************\n"));
			
			Asura_PS3_CommandBufferInspector::Run( s_paxCommandBuffer[ 1^s_uWhich ], s_paxCommandBuffer[ 1^s_uWhich ] + s_uCommandBufferSize );
			
			bInspected = true;
		}
#endif

	}
}

int Asura_PS3_SwapChain::CommandBufferCallback( struct CellGcmContextData* pContextData, unsigned int Amount )
{
	Asura_OutputToDebugger( ( "Asura_PS3_SwapChain - out of command buffer space\n" ) );
	return 0;
}

// SkipFlip
// by Ben Weston 14/7/08
// Fix a bug when we free memory when quitting the game but RSX still wants to draw the next frame
// Quite simply, when graphics is reset, it calls this to discard the command buffer we've just built
void Asura_PS3_SwapChain::SkipFlip( void )
{
	// adapted from Present (with the bits that actually present removed!)
	BlockOnFlip(1);

	// next frame please!
	s_uWhich = s_uWhich ^ 1;

	// start a new command buffer
	cellGcmSetupContextData( &s_xCurrentCmdCtxt, s_paxCommandBuffer[ s_uWhich ], s_uCommandBufferSize, &CommandBufferCallback );
 	gCellGcmCurrentContext = &s_xCurrentCmdCtxt;
}