//////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_TextureManagement.h
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___ASURA_PS3_TEXTUREMANAGEMENT_H___
#define ___ASURA_PS3_TEXTUREMANAGEMENT_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Engine independent.
#include "Core/Asura_Handle.h"

// Engine platform-specific.
#include "Rendering/Asura_PS3_TextureHeader.h"
#include "FileAccess/Asura_PS3_FileSystem.h"
#include "Rendering/Asura_PS3_Graphics.h"

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

class Asura_PS3_Texture;
class Asura_MemAllocator;

// C L A S S E S /////////////////////////////////////////////////////////////

class Asura_PS3_TextureManagement
{

public:

    enum
    {
        SPECIALTEXTURE_FLAG = 0x10000000,

        SPECIALTEXTURE_ID_PLACEHOLDER   = 0 | SPECIALTEXTURE_FLAG, // Placeholder pink texture, used by default for missing textures.
        SPECIALTEXTURE_ID_NOTEXTURE     = 1 | SPECIALTEXTURE_FLAG, // Blank white texture, used by default if ASURA_TEXTURE_NONE is specified.
        SPECIALTEXTURE_ID_FLATNORMALMAP = 2 | SPECIALTEXTURE_FLAG, // Can be used when normal mapping to give flat surface (ie. equivalent of not normal mapping).
        SPECIALTEXTURE_ID_BLACKTEXTURE	= 3 | SPECIALTEXTURE_FLAG, // Black, useful for removing specular maps.
        SPECIALTEXTURE_NUM_IDS          = 4,
    };

    static void Initialise( void );
    static void Reset( void );
    static void Shutdown( void );

    static Asura_PS3_Texture* GetTexture( int iTexID );

	static bool TextureBudgetEnabled()						{return s_bTextureBudgetEnabled;}
	static void SetTextureBudget(bool bEnable, int iBytes)  {s_bTextureBudgetEnabled = bEnable; s_iTextureBudget = iBytes;}
	static int  GetTextureBudget()							{return s_iTextureBudget;}
	static void DecrementTextureBudget(int iBytes)			{s_iTextureBudget-=iBytes;}
	static void SetTextureBudgetExceededAmount(int iBytes)  {s_iTextureBudgetExceededBy = iBytes;}
	static int  GetTextureBudgetExceededAmount(void)		{return s_iTextureBudgetExceededBy;}
protected:
	static bool s_bTextureBudgetEnabled;
	static int	s_iTextureBudget;
	static int  s_iTextureBudgetExceededBy;
};
// End class Asura_PS3_TextureManagement.

class Asura_PS3_Texture
{

public:
	typedef void fnDeallocateTexture(void* pData);
	typedef void* fnAllocateTexture(const u_int uSize, const u_int uAlignment, CellGcmTexture& xGCMTexture);

	Asura_PS3_Texture( CellGcmTexture* pxTexture );
    Asura_PS3_Texture( u_char m_ucLocation = CELL_GCM_LOCATION_LOCAL );
    ~Asura_PS3_Texture( void );

    CellGcmTexture* GetTexture( void ) { return &m_xTexture; }
	void* GetTextureData( void ) { return m_pxTextureData; }
	void  SetTextureData( void* pxTextureData );

	void CreateTexture( u_short usWidth, u_short usHeight, ASURA_PS3_TEXTUREFORMAT eFormat = ASURA_PS3_TEXTUREFORMAT_8888_RGBA, bool bAllocateMemory = true, u_short usFalgs = 0 );

	void SetAsTextureIndex( int iIndex, bool bNormalizedTextureCoordinates = true );
	void SetUseNormalizedCoordinates( bool bNormalizedTextureCoordinates );
	int GetTextureFaceSize( void ) { return m_iTextureFaceSize; }

    void Destroy( fnDeallocateTexture* pfnDeallocator = NULL );

	u_int GetTextureOffset( void )		{ return m_uiTextureOffset; }
	u_int GetTextureFormat( void )		{ return m_uiTextureFormat; }
	u_int GetTextureControl1( void )	{ return m_uiTextureControl1; }
	u_int GetTextureControl3( void )	{ return m_uiTextureControl3; }
	u_int GetTextureImageRect( void )	{ return m_uiTextureImageRect; }

	bool Is8BitTexture( void );

private:

    friend class Asura_Texture_Management;
    friend class Asura_PS3_TextureManagement;
	friend class Asura_Texture_Streaming_System;

	bool InitFromFile( Asura_Handle iFileHandle, Asura_PS3_FileSystem::Handle iCacheHandle = Asura_PS3_FileSystem::INVALID_HANDLE, fnAllocateTexture* pfnAllocator = NULL );
    void SetupDummyTexture( u_int uRGBA, fnAllocateTexture* pfnAllocator = NULL );

    int CreateTextureObject( fnAllocateTexture* pfnAllocator = NULL );
	void InitGCMMacroData( void );

	int	CalculatePitch( const int iWidth, const int iBlockSize, const int iBytesPerPixel );

    Asura_PS3_TextureHeader	m_xTexHdr;
    CellGcmTexture			m_xTexture;

	u_int					m_uiTextureOffset; // offset
	u_int					m_uiTextureFormat; // format (location, cubemap, dimension, format, mipmap)
	u_int					m_uiTextureControl1; //Remap
	u_int					m_uiTextureControl3; // pitch, depth
	u_int					m_uiTextureImageRect; // width, height

	void*					m_pxTextureData;
	int						m_iTextureFaceSize;
	bool					m_bAllocatedOwnTextureData;

#if _DEBUG
    const char*             m_szName;
#endif

};
// End class Asura_PS3_Texture.



////////////////////////////////////////////////////////////////////////////////
// Class        : Asura_PS3_Texture_Streaming_MemoryAllocator
// Author       : Andy Tate
// Date         : 18-Dec-2008
//
// Description  : Singleton class used to 
////////////////////////////////////////////////////////////////////////////////

class Asura_PS3_Texture_Streaming_MemoryAllocator
{
public:
	// Public functions...
	static void Initialise(const u_int uSizeInMB, const u_int uMaxTextures);
	static void Shutdown(void);
	static void Reset(void);

	// Allocation functions...
	static void* Allocate(const u_int uSize, const u_int uAlignment, CellGcmTexture& xGCMTexture);
	static void Free(void* pData);

	// Accessor functions...
	static bool	IsInitialised(void)	{ return(s_pxMemAllocator != NULL); }


protected:
	// Protected variables...

	// Protected functions...


private:
	// Private variables...
	static bool					s_bMemoryIsLocal;
	static void*				s_pMemory;
	static Asura_MemAllocator*	s_pxMemAllocator;

	// Private functions...
	Asura_PS3_Texture_Streaming_MemoryAllocator(void);
	Asura_PS3_Texture_Streaming_MemoryAllocator(const Asura_PS3_Texture_Streaming_MemoryAllocator& xOther);
	~Asura_PS3_Texture_Streaming_MemoryAllocator();

	Asura_PS3_Texture_Streaming_MemoryAllocator&	operator=(const Asura_PS3_Texture_Streaming_MemoryAllocator& xOther);

};   // End of class 'Asura_PS3_Texture_Streaming_MemoryAllocator'.

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___ASURA_PS3_TEXTUREMANAGEMENT_H___
