#include "Shared.FXH"
#include "MaterialResponse_Shared.ps3FXH"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

// Camera view space -> World space -> Light view space -> Light clip space
const float4x4 g_xCameraViewSpaceToLightClipSpace;

// Camera view space -> World space -> Light view space
const float4x4 g_xCameraViewSpaceToLightViewSpace;

// Camera view space -> World space
const float4x4 g_xCameraViewSpaceToWorldSpace;
// shadow caster position in world space
const float4 g_xLightPos;

// 2 / dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float2 g_xTwoOverScreen;
const float2 g_fScreenOffset;

// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
const float2  g_fCameraTanHalfFOV;

const float g_fShadowFilterTextureScale;
const float g_fShadowFilterSize;

const float4 g_xShadowDirection;
const float  g_fShadowStrength;

// Shadow frustum near/far values, used to map shadow depth textures into the appropriate range
const float	g_fShadowFrustumNearPlaneDistance;
const float	g_fOneOverShadowFrustumPlaneSeparation;
const float3 g_xSelfShadowBias;

const sampler g_xNormalMtlBuffer		: register( s0 );
const sampler g_xDepthBuffer			: register( s2 );
const sampler g_xShadowBuffer			: register( s5 );
const samplerCUBE g_xShadowCubeBuffer	: register( s6 );
const sampler g_xKernelRotationSampler	: register( s7 );

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#define NUM_TAPS 4

#if (NUM_TAPS > 0)
static const float2 axFILTER_TAPS[ NUM_TAPS ] =
{
#if (NUM_TAPS == 4)

	#ifdef SHADER_USE_NOISE
		{ -0.695914f,  0.457137f },
		{  0.519456f,  0.767022f },
		{  0.473434f, -0.480026f },
		{ -0.321940f, -0.932615f },
	#else
	//	{ -0.177f,  0.152f },
	//	{  0.152f,  0.177f },
	//	{  0.177f, -0.152f },
	//	{ -0.152f, -0.177f },
		{ -0.177f,  0.306f },
		{  0.306f,  0.177f },
		{  0.177f, -0.306f },
		{ -0.306f, -0.177f },
	#endif

#elif (NUM_TAPS == 8)
	{ -0.840144h, -0.073580h },
	{ -0.695914h,  0.457137h },
	{ -0.203345h,  0.620716h },
	{  0.473434h, -0.480026h },
	{  0.519456h,  0.767022h },
	{  0.185461h, -0.893124h },
	{  0.896420h,  0.412458h },
	{ -0.321940h, -0.932615h },
#elif (NUM_TAPS == 12)
	{ -0.326212h, -0.405805h },
	{ -0.840144h, -0.073580h },
	{ -0.695914h,  0.457137h },
	{ -0.203345h,  0.620716h },
	{  0.962340h, -0.194983h },
	{  0.473434h, -0.480026h },
	{  0.519456h,  0.767022h },
	{  0.185461h, -0.893124h },
	{  0.507431h,  0.064425h },
	{  0.896420h,  0.412458h },
	{ -0.321940h, -0.932615h },
	{ -0.791559h, -0.597705h },
#endif	// (NUM_TAPS == 12)
};
#endif


/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// The vertex shader recreates the shadow frustum in world space, then projects
// it as an object in the standard camera view
float GetShadowDepth3( float3 xUVSceneLightspaceDepth )
{
	#ifdef SHADER_CUBIC_MAP_SHADOW
		return f1texCUBE( g_xShadowCubeBuffer, xUVSceneLightspaceDepth );
	#else
		return f1texcompare2D( g_xShadowBuffer, xUVSceneLightspaceDepth );
	#endif
}

float AdjustShadow(	float fShadow
	#ifndef SHADER_CREATE_SHADOW_MASK
	, const float fDot
	, const float fDistScale
	#endif
	)
{
	fShadow = fShadow * g_fShadowStrength;
	#ifndef SHADER_CREATE_SHADOW_MASK
		fShadow = saturate( fShadow * fDistScale ) * fDot;
	#endif

	return saturate( fShadow );
}

float BiasDepthValue( float fSceneLightspaceDepth, inout float fDot )
{
	float3 xSelfShadowBias = g_xSelfShadowBias;
	xSelfShadowBias.x = 0.000001f;
	fSceneLightspaceDepth -= ( ( 1.0f - fDot ) * xSelfShadowBias.x + g_xSelfShadowBias.y + fSceneLightspaceDepth * xSelfShadowBias.z );
	#ifndef SHADER_CREATE_SHADOW_MASK
		fDot = saturate( 2.0f * fDot );
	#endif

	return fSceneLightspaceDepth;
}

float2 GetSinCos( const float fScale , const float2 xUV )
{
	float2 fSinCos;
	#ifdef SHADER_USE_NOISE
		fSinCos  = ( h2tex2D( g_xKernelRotationSampler, xUV.xy * 16.0f ) ) * fScale * 0.66f;
		fSinCos += ( h2tex2D( g_xKernelRotationSampler, xUV.xy * 64.0f ) ) * fScale * 0.33f;
	#else
		fSinCos = float2( fScale, 0.0f );
	#endif
	return fSinCos;
}

#if SHADER_CUBIC_MAP_SHADOW

float GetFilteredShadow( float4 xUVDepth, float fDistScale, float fSceneLightspaceDepth )
{
	// Regardless of whether this pixel is actually shadowed, allow 'nearby' occluders to contribute to the shadow value
	float fShadow = 0.0f;

	// Calculate a value to scale the sampling region on the shadow buffer by
	const float fScaling = g_fShadowFilterSize + saturate( 1.0f - fDistScale * 2.0f ) * 2.0f;
	const float fScale = g_fShadowFilterTextureScale * fScaling;

	// To reduce banding, we randomly rotate the filter kernel per-pixel. Uses a texture of pre-generated
	// sin/cos for random rotation angles
	const float2 fSinCos = GetSinCos( fScale, xUVDepth.xy / xUVDepth.w );
	float fMaxDist = 0.05 * g_fOneOverShadowFrustumPlaneSeparation;

	for( int i = 0; i < NUM_TAPS; ++i )
	{
		// calculate the offset texture coordinates for sampling the occluder depth			
		const float3 xOffset = float3( dot( axFILTER_TAPS[i].xy , fSinCos ), dot( axFILTER_TAPS[i].xy , float2( fSinCos.y, -fSinCos.x ) ), 0.0f );
		const float3 xTapCoord = xUVDepth.xyz + xOffset;
		const float  fShadow2 = ( fSceneLightspaceDepth - GetShadowDepth3( xTapCoord.xyz ) );
		// Accumulate it
		fShadow += min( max( fShadow2, 0.0f ),  fMaxDist );
	}
	// Average out the contributions across all taps (plus the original shadow sample)
	fShadow /= ( NUM_TAPS * fMaxDist );
	
	return fShadow;
}

float GetUnFilteredShadow( float3 xUVDepth, float fSceneLightspaceDepth )
{
	return ( fSceneLightspaceDepth > GetShadowDepth3( xUVDepth ) ) ? 1.0h : 0.0h;
}

float3 CalculateLightClipViewSpacePosition(
	  const float4 xCameraViewSpacePosition
	, const half3 xNormal
	, out float4 xUVDepth
	, out float fSceneLightspaceDepth
	, out float fDistScale
	, out float fDot
	)
{
	float4 xWorldSpacePosition = mul( xCameraViewSpacePosition, g_xCameraViewSpaceToWorldSpace );
	float4 xLightVector = g_xLightPos - xWorldSpacePosition;
	fSceneLightspaceDepth = length( xLightVector );
	xUVDepth.xyz = xLightVector.xyz / fSceneLightspaceDepth;
	fSceneLightspaceDepth = ( fSceneLightspaceDepth - g_fShadowFrustumNearPlaneDistance ) * g_fOneOverShadowFrustumPlaneSeparation;
	fDistScale = 1.0f - fSceneLightspaceDepth;
	fDot = saturate( dot( xLightVector.xyz, xNormal ) );

	clip( min( fSceneLightspaceDepth, fDistScale ) );

	fSceneLightspaceDepth = BiasDepthValue( fSceneLightspaceDepth, fDot );
	xUVDepth.xz = -xUVDepth.xz;
	xUVDepth.w = fSceneLightspaceDepth;

	return xLightVector.xyz;
}

float CalculateCubicMapShadow( float4 xCameraViewSpacePosition, half3 xNormal )
{
	float4 xUVDepth;
	float fShadow;
	float fSceneLightspaceDepth;
	float fDistScale;
	float fDot;
	float3 xLightVector = CalculateLightClipViewSpacePosition( xCameraViewSpacePosition, xNormal, xUVDepth, fSceneLightspaceDepth, fDistScale, fDot );
	
	#ifdef SHADER_FILTER_SHADOWS
		fShadow = GetFilteredShadow( xUVDepth, fDistScale, fSceneLightspaceDepth );
	#else
		fShadow = GetUnFilteredShadow( xUVDepth.xyz, fSceneLightspaceDepth );
	#endif

 	#ifndef SHADER_CREATE_SHADOW_MASK
 		return AdjustShadow( fShadow, fDot, fDistScale );
 	#else
 		return AdjustShadow( fShadow );
 	#endif
}

#else

float GetFilteredShadow( float3 xUVDepth, float fDistScale )
{
	// Regardless of whether this pixel is actually shadowed, allow 'nearby' occluders to contribute to the shadow value
	float fShadow = 0.0f;

	// Calculate a value to scale the sampling region on the shadow buffer by
	const float fScaling = g_fShadowFilterSize + saturate( 1.0f - fDistScale * 2.0f ) * 2.0f;
	const float fScale = g_fShadowFilterTextureScale * fScaling;

	// To reduce banding, we randomly rotate the filter kernel per-pixel. Uses a texture of pre-generated
	// sin/cos for random rotation angles
	const float2 fSinCos = GetSinCos( fScale, xUVDepth.xy );

	for( int i = 0; i < NUM_TAPS; ++i )
	{
		// calculate the offset texture coordinates for sampling the occluder depth			
		const float3 xOffset = float3( dot( axFILTER_TAPS[i].xy , fSinCos ), dot( axFILTER_TAPS[i].xy , float2( fSinCos.y, -fSinCos.x ) ), 0.0f );
		const float3 xTapCoord = xUVDepth.xyz + xOffset;
		// Accumulate it
		fShadow += GetShadowDepth3( xTapCoord.xyz );
	}
	// Average out the contributions across all taps (plus the original shadow sample)
	fShadow /= NUM_TAPS;
	
	return fShadow;
}

float GetUnFilteredShadow( float3 xUVDepth )
{
	return GetShadowDepth3( xUVDepth );
}

void CalculateLightClipViewSpacePosition(
	  const float4 xCameraViewSpacePosition
	, const half3 xNormal
	, out float4 xLightClipSpacePosition
	, out float fDistScale
	, out float fDot
	)
{
	xLightClipSpacePosition = mul( xCameraViewSpacePosition, g_xCameraViewSpaceToLightClipSpace );

	const float4 xLightViewSpacePosition = mul( xCameraViewSpacePosition, g_xCameraViewSpaceToLightViewSpace );
	float fLightDist = ( ( xLightViewSpacePosition.z - g_fShadowFrustumNearPlaneDistance ) * g_fOneOverShadowFrustumPlaneSeparation );
	fDistScale = 1.0f - fLightDist;
	fDot = saturate( dot( g_xShadowDirection.xyz, xNormal ) );

	clip( min( fLightDist, fDistScale ) );

	xLightClipSpacePosition.xyz /= xLightClipSpacePosition.w;

	xLightClipSpacePosition.z = BiasDepthValue( xLightClipSpacePosition.z, fDot );

	xLightClipSpacePosition.xyz += 1.0f;
	xLightClipSpacePosition.xyz *= 0.5f;

}

float CalculateShadow( float4 xCameraViewSpacePosition, half3 xNormal )
{
	float4 xLightClipSpacePosition;
	float fShadow;
	float fDistScale;
	float fDot;
	CalculateLightClipViewSpacePosition( xCameraViewSpacePosition, xNormal, xLightClipSpacePosition, fDistScale, fDot );
	
	#ifdef SHADER_FILTER_SHADOWS
		fShadow = GetFilteredShadow( xLightClipSpacePosition.xyz, fDistScale );
	#else
		fShadow = GetUnFilteredShadow( xLightClipSpacePosition.xyz );
	#endif
	
	#ifndef SHADER_CREATE_SHADOW_MASK
		return AdjustShadow( fShadow, fDot, fDistScale );
	#else
		return AdjustShadow( fShadow );
	#endif
}
#endif

float4 main( const float2 xScreenPos : WPOS ) : COLOR
{
	const half3	xNormal				= h3tex2D( g_xNormalMtlBuffer, xScreenPos );
	// Fetch the world-space normal and screen depth
	const float fScreenDepth	= Deferred_GetDepth(h4tex2D( g_xDepthBuffer, xScreenPos ));
	
	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	const float4 xCameraViewSpacePosition = float4(	
				g_fCameraTanHalfFOV * fScreenDepth * ( ( xScreenPos - g_fScreenOffset ) * g_xTwoOverScreen - 1.0h ),
				fScreenDepth,
				1.0h );
#ifdef SHADER_CUBIC_MAP_SHADOW
	return float4( 0.0f, 0.0f, 0.0f, CalculateCubicMapShadow( xCameraViewSpacePosition, xNormal ) );
#else
	return float4( 0.0f, 0.0f, 0.0f, CalculateShadow( xCameraViewSpacePosition, xNormal ) );
#endif
}
