//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Shadow_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Engine independent.
#include "Interface/Asura_CommandConsole.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_HalfFloat.h"
#include "Core/Memory/Asura_Mem.h"
#include "Rendering/Asura_Viewport.h"
#include "Shadows/Asura_Shadow_System.h"

// Engine platform-specific.
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Rendering/Asura_PS3_GCM_IndexBuffer.h"
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_QuadOverlayRenderer.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"
#include "Rendering/Asura_PS3_EnvironmentRenderer.h"
#include "Rendering/Asura_PS3_DeferredShading.h"

//Shaders definition
#include "Rendering/ShadersDefinition/VD/Apply_Shadows_VD.h"
#include "Rendering/ShadersDefinition/VP/Apply_Shadows_VP.h"
#include "Rendering/ShadersDefinition/FP/Apply_Shadows_FP.h"

#include "Rendering/ShadersDefinition/FP/Shadows_FP.h"

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Asura_PS3_Texture*				g_pxFilterKernelRotationTexture = NULL;
static Asura_PS3_GCM_VertexBuffer		g_xShadowFrustumVB;
static Asura_PS3_GCM_VertexBuffer		g_xShadowFrustumGeneratedVB;
static Asura_PS3_GCM_IndexBuffer		g_xShadowFrustumIB;

static Asura_PS3_GCM_VertexShader*		g_pxVertexShader				= NULL;
static Asura_PS3_FragmentShaderCopy*	g_pxFragmentShader				= NULL;
static	Asura_PS3_FragmentShader*		g_pxFragmentShaderShadow		= NULL;
static bool								g_bDisableFiltering				= false;

static bool								g_bUseFrustumVertices			= true;

#define USE_QUADS						1

#if !defined( ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS )
#define ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS	( ASURA_NUM_SHADOW_TEXTURE_SIZES * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE )
#endif	//!defined(ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS)

#if !defined( ASURA_MAX_NUMBER_OF_SHADOWS )
#define ASURA_MAX_NUMBER_OF_SHADOWS				( ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS + ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS )
#endif	//!defined(ASURA_MAX_NUMBER_OF_SHADOWS)

static float s_xSelfShadowBiasValues[ 3 ][ 3 ] = 
{
	// Slope, Bias,   Dist
	{ 0.001f, 0.001f,  0.01f   }, // non environment
	{ 0.000f, 0.0002f, 0.0001f }, // environment
	{ 0.001f, 0.001f,  0.01f   }  // cubic environment
};

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Platofrm_ShadowBuffer
//
////////////////////////////////////////////////////////////////////////////////
class Asura_Platform_ShadowBuffer
{
public:
	Asura_Platform_ShadowBuffer( u_int uTextureSize, Asura_PS3_FrameBuffer* pxFrameBuffer )
		: m_uTextureSize(uTextureSize),
		m_pxFrameBuffer( pxFrameBuffer )
	{
	}

	~Asura_Platform_ShadowBuffer()
	{
		delete m_pxFrameBuffer;
		m_pxFrameBuffer = NULL;
	}

	void Bind( void ) const
	{
		m_pxFrameBuffer->BindAsZPassOnly();
		cellGcmSetZcullStatsEnableUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_FALSE );
		cellGcmSetClearSurfaceUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_CLEAR_Z | CELL_GCM_CLEAR_S );
		g_pxFragmentShaderShadow->Bind();
	}

	void UnBind( void )
	{
		Asura_PS3_DeferredRendering::BindAsSingleRT();
		cellGcmSetZcullStatsEnableUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRUE );
	}

	Asura_PS3_Texture*		GetTexture			(void)			{ return m_pxFrameBuffer->GetZBufferAsTexture(); }
	u_int					GetTextureSize		(void) const	{ return( m_uTextureSize ); }

private:
	// Data
	u_int						m_uTextureSize;
	Asura_PS3_FrameBuffer*		m_pxFrameBuffer;
};

static Asura_PS3_FrameBuffer*			s_pxCubeMapFrameBuffer;
static Asura_PS3_Texture				s_pxEnvShadowCubeTexture;
static Asura_Platform_ShadowBuffer*		s_axShadowTextures[ ASURA_NUM_SHADOW_TEXTURE_SIZES * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE ];
static u_int							s_uShadowTexturesAllocated;

////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_Shadow
//
// Represents a single PS3 shadow
//
////////////////////////////////////////////////////////////////////////////////

class Asura_PS3_Shadow : public Asura_Shadow
{
public:

	// Constructor / destructor
	Asura_PS3_Shadow() {}
	virtual ~Asura_PS3_Shadow() {}

	// Virtual overrides

	virtual void PrepareForRendering( void );
	virtual void FinishedRendering( void );

	Asura_Platform_ShadowBuffer* GetShadowBuffer ( void ) const	{ return m_pxShadowBuffer; }

	virtual void SetInstance( const Asura_Shadow_Instance* pxInstance )
	{ 
		m_pxInstance = pxInstance; 
		if( ( !pxInstance->IsCubeMapped() ) && ( s_uShadowTexturesAllocated < ( ASURA_NUM_SHADOW_TEXTURE_SIZES * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE ) ) )
		{
			m_pxShadowBuffer = s_axShadowTextures[ s_uShadowTexturesAllocated++ ];
		}
		else
		{
			m_pxShadowBuffer = NULL;
		}
	}

	void SetIsEnvironmentShadow( bool bEnvironmentShadow ) { m_bIsEnvironmentShadow = bEnvironmentShadow; }
	bool GetIsEnvironmentShadow( void ) { return m_bIsEnvironmentShadow; }

protected:
	// Data
	Asura_Platform_ShadowBuffer*	m_pxShadowBuffer;
	bool							m_bIsEnvironmentShadow;
};

static bool CreateEnvShadowTextures( void* = NULL )
{
	s_pxCubeMapFrameBuffer = new Asura_PS3_FrameBuffer( true, ASURA_MAX_SHADOW_TEXTURE_SIZE, ASURA_MAX_SHADOW_TEXTURE_SIZE, 1, ASURA_PS3_FRAMEBUFFER_32F, true, false, "shadow cubemap", -1, true );
	// Create cube texture
	s_pxEnvShadowCubeTexture.CreateTexture( ASURA_MAX_SHADOW_TEXTURE_SIZE, ASURA_MAX_SHADOW_TEXTURE_SIZE, ASURA_PS3_TEXTUREFORMAT_X32F, true, ASURA_PS3_TEXTURE_CUBEMAP );

	for( u_int i = 0, uTexSize = ASURA_MAX_SHADOW_TEXTURE_SIZE; i < ASURA_NUM_SHADOW_TEXTURE_SIZES; ++i, ( uTexSize >>= 1 ) )
	{
		for (u_int j = 0; j < ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE; ++j)
		{
			Asura_PS3_FrameBuffer* pxFrameBuffer;
			pxFrameBuffer = new Asura_PS3_FrameBuffer( true, uTexSize, uTexSize, 0, ASURA_PS3_FRAMEBUFFER_RGBA, true, false, "shadow", -1, false ); //BW todo: place these in 1 shared tiled region (we can't use tiled memory for each one because there's too many of them)
			s_axShadowTextures[ j + i * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE ] = new Asura_Platform_ShadowBuffer( uTexSize, pxFrameBuffer );
		}
	}
	return true;
}

static bool DestroyEnvShadowTextures( void* = NULL )
{
	delete s_pxCubeMapFrameBuffer;
	s_pxEnvShadowCubeTexture.Destroy();

	for( u_int i = 0, uTexSize = ASURA_MAX_SHADOW_TEXTURE_SIZE; i < ASURA_NUM_SHADOW_TEXTURE_SIZES; ++i, ( uTexSize >>= 1 ) )
	{
		for( u_int j = 0; j < ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE; ++j )
		{
			delete s_axShadowTextures[ j + i * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE ];
			s_axShadowTextures[ j + i * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE ] = NULL;
		}
	}

	return true;
}

#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
static bool ReloadCubmapShaders( bool bHelp )
{
	if( bHelp )
	{
		Asura_CommandConsole::Printf("Reload ApplyShadows_FP1000 to FP1111");
		return true;
	}

	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0000.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 0, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0001.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0010.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 10, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0011.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 11, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0100.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 100, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0101.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 101, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0110.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 110, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP0111.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 111, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );

	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1000.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1000, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1001.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1001, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1010.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1010, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1011.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1011, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1100.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1100, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1101.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1101, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1110.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1110, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader("~/Asura/PS3/Shaders/FragmentPrograms/ApplyShadows/ApplyShadows_FP1111.fpx",
		static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1111, NULL, NULL, new Asura_PS3_ShadowSystem_Handles() );
	return true;
}
#endif //#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Initialise()
//
// Initialises the shadow system on PS 3
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_Initialise()
{
	typedef struct _D3DVECTOR {
		float x;
		float y;
		float z;
	} D3DVECTOR;

	const D3DVECTOR axFRUSTUM_VERTICES[] =
	{
		// this order is the same as Asura_Frustum::GenerateVertices()
		{  1.0f,  1.0f, 0.0f },
		{ -1.0f,  1.0f, 0.0f },
		{  1.0f, -1.0f, 0.0f },
		{ -1.0f, -1.0f, 0.0f },
		{  1.0f,  1.0f, 1.0f },
		{ -1.0f,  1.0f, 1.0f },
		{  1.0f, -1.0f, 1.0f },
		{ -1.0f, -1.0f, 1.0f },
	};
#if USE_QUADS
	const u_short ausFRUSTUM_INDICES[] =
	{
		1, 0, 2, 3,		// Near
		6, 7, 3, 2,		// Bottom
		2, 0, 4, 6,		// Right
		4, 0, 1, 5,		// Top
		1, 3, 7, 5,		// Left
		6, 4, 5, 7,		// Far
	};
#else
	// this order is wrong, I updated the vertex order to match Asura_Frustum but didn't touch this because it wasn't used. - BenW
	#error triangles are no longer supported
	const u_short ausFRUSTUM_INDICES[] =
	{
		0, 1, 2,
		2, 3, 0,		// Near
		4, 5, 1,
		1, 0, 4,		// Bottom
		5, 6, 2,
		2, 1, 5,		// Right
		6, 7, 3,
		3, 2, 6,		// Top
		7, 4, 0,
		0, 3, 7,		// Left
		5, 4, 7,
		7, 6, 5,		// Far
	};
#endif

#if defined( ASURA_SUPPORT_COMMAND_CONSOLE )
	Asura_CommandConsole::AddVar( L"Render.Shadows.DisableFiltering",	&g_bDisableFiltering,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"Render.Shadows.UseInlineVertices",	&g_bUseFrustumVertices,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif
	g_pxVertexShader			= new Asura_PS3_VP_ApplyShadows();
	g_pxFragmentShaderShadow	= new Asura_PS3_FragmentShader( ( unsigned char *)&_binary_PS3_Shadows_FP_bin_start, (unsigned char *)&_binary_PS3_Shadows_FP_bin_end, NULL );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 0, (unsigned char *)&_binary_PS3_ApplyShadows_FP0000_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0000_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0000_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1, (unsigned char *)&_binary_PS3_ApplyShadows_FP0001_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0001_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0001_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 10, (unsigned char *)&_binary_PS3_ApplyShadows_FP0010_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0010_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0010_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 11, (unsigned char *)&_binary_PS3_ApplyShadows_FP0011_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0011_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0011_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 100, (unsigned char *)&_binary_PS3_ApplyShadows_FP0100_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0100_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0100_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 101, (unsigned char *)&_binary_PS3_ApplyShadows_FP0101_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0101_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0101_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 110, (unsigned char *)&_binary_PS3_ApplyShadows_FP0110_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0110_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0110_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 111, (unsigned char *)&_binary_PS3_ApplyShadows_FP0111_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP0111_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP0111_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1000, (unsigned char *)&_binary_PS3_ApplyShadows_FP1000_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1000_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1000_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1001, (unsigned char *)&_binary_PS3_ApplyShadows_FP1001_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1001_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1001_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1010, (unsigned char *)&_binary_PS3_ApplyShadows_FP1010_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1010_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1010_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1011, (unsigned char *)&_binary_PS3_ApplyShadows_FP1011_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1011_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1011_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1100, (unsigned char *)&_binary_PS3_ApplyShadows_FP1100_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1100_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1100_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1101, (unsigned char *)&_binary_PS3_ApplyShadows_FP1101_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1101_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1101_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1110, (unsigned char *)&_binary_PS3_ApplyShadows_FP1110_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1110_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1110_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_SHADOW ), 1111, (unsigned char *)&_binary_PS3_ApplyShadows_FP1111_bin_start, ( (unsigned char *)&_binary_PS3_ApplyShadows_FP1111_bin_end - (unsigned char *)&_binary_PS3_ApplyShadows_FP1111_bin_start), new Asura_PS3_ShadowSystem_Handles(), 1, true );

	g_pxVertexShader->InitVertexBufferDeclaration( g_axVertexDeclaration, VD_ELEMENTS );

	Asura_Shadow**	ppxShadow		= s_apxShadows;

	for(u_int u = 0; u < ASURA_MAX_NUMBER_OF_SHADOWS; ++u, ++ppxShadow)
	{
		*ppxShadow	= new Asura_PS3_Shadow();
	}

	g_pxFilterKernelRotationTexture = new Asura_PS3_Texture();
	g_pxFilterKernelRotationTexture->CreateTexture( 64, 64, ASURA_PS3_TEXTUREFORMAT_YX16F );
	Asura_HalfFloat* pxData = static_cast< Asura_HalfFloat* >( g_pxFilterKernelRotationTexture->GetTextureData() );
	int iPitch = ( g_pxFilterKernelRotationTexture->GetTexture()->pitch / sizeof( Asura_HalfFloat ) );

	for( int i = 0; i < 64; ++i )
	{
		for( int j = 0; j < 127; )
		{
			float fAngle = Asura_Random::GetFloat( 0, Asura_Maths::Pi * 2.0f );
			float fSin,fCos;

			Asura_Maths::SinCos( fAngle, fSin, fCos );
			pxData[ j++ ] = fSin;	//->It can be read from Green or Blue 
			pxData[ j++ ] = fCos;	//->It can be read from Red or Alpha
		}
		pxData += iPitch;
	}

	// TODO: Need to add filtered mipmap levels for g_pxFilterKernelRotationTexture

	g_xShadowFrustumVB.Initialise( 12, 8, CELL_GCM_LOCATION_LOCAL );
	g_xShadowFrustumGeneratedVB.Initialise( 12, 8, CELL_GCM_LOCATION_MAIN );

	void* pVertices = g_xShadowFrustumVB.LockForWriting();
	Asura_Assert( pVertices, "Failed to lock shadow frustum vertex buffer." );
	Asura_Mem::Copy( pVertices, axFRUSTUM_VERTICES, sizeof( axFRUSTUM_VERTICES ) );

	g_xShadowFrustumIB.Initialise( 24 );
	void* pIndices = g_xShadowFrustumIB.LockForWriting();
	Asura_Assert( pIndices, "Failed to lock shadow frustum index buffer." );
	Asura_Mem::Copy( pIndices, ausFRUSTUM_INDICES, sizeof( ausFRUSTUM_INDICES ) );

	CreateEnvShadowTextures();
#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
	Asura_CommandConsole::AddCmd(L"Render.ReloadShaders.Shadows", ReloadCubmapShaders, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
	if( !Asura_CommandConsole::VarExists( L"Render.Shadows.Bias.Skin_Normal" ) )
	{
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Slope", &( s_xSelfShadowBiasValues[ 0 ][ 0 ]),	0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Bias", &(s_xSelfShadowBiasValues[ 0 ][ 1 ] ),	0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Dist", &(s_xSelfShadowBiasValues[ 0 ][ 2 ] ),	0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Slope", &(s_xSelfShadowBiasValues[ 1 ][ 0 ] ),	0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Bias", &(s_xSelfShadowBiasValues[ 1 ][ 1 ] ),	0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Dist", &(s_xSelfShadowBiasValues[ 1 ][ 2 ] ),	0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Slope", &(s_xSelfShadowBiasValues[ 2 ][ 0 ] ),	-1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Bias", &(s_xSelfShadowBiasValues[ 2 ][ 1 ] ),	-1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Dist", &(s_xSelfShadowBiasValues[ 2 ][ 2 ] ),	-1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	}
#endif
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Reset()
//
// Resets the shadow system on PS 3. Called between levels
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////
void Asura_Shadow_System::Platform_Reset()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Shutdown()
//
// Shuts down the shadow system on PS3
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////
void Asura_Shadow_System::Platform_Shutdown()
{
	DestroyEnvShadowTextures();

	delete g_pxVertexShader;
	delete g_pxFilterKernelRotationTexture;
	delete g_pxFragmentShaderShadow;

	g_pxFilterKernelRotationTexture	= NULL;
	g_pxVertexShader				= NULL;
	g_pxFragmentShaderShadow		= NULL;

	g_xShadowFrustumVB.Shutdown();
	g_xShadowFrustumGeneratedVB.Shutdown();
	g_xShadowFrustumIB.Shutdown();
}


void Asura_Shadow_System::Platform_BeginCreateShadows()
{
    ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__CREATESHADOWS, "Create Shadows" );
	s_uShadowTexturesAllocated = 0;
}


void Asura_Shadow_System::Platform_EndCreateShadows()
{
    ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__CREATESHADOWS );
}

static void RenderShadowOnToScene(const Asura_Shadow_Instance *pxInstance, bool bEnvMap)
{
	g_pxVertexShader->Bind();

	Asura_Matrix_3x3 xLightOri;
	Asura_AMatrix4x4 xShadowProj, xInvShadowProj, xInvView, xInvViewProj;
	const Asura_Frustum&	xFrustum			= pxInstance->GetFrustum();				
	const Asura_AMatrix4x4&	xViewProj			= Asura_PS3_Graphics::GetViewProjectionMatrix();
	const Asura_AMatrix4x4&	xView				= Asura_PS3_Graphics::GetViewMatrix();
	const float				fTanHalfVertFOV		= Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
	const float				fTanHalfHorizFOV	= fTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();
	const Asura_Vector_3&	xLightPos			= pxInstance->GetLightPosition();
	const float				fNear				= xFrustum.GetNearPlane();
	const float				fFar				= xFrustum.GetFarPlane();
	const float				fRange				= fFar - fNear;
	const float				fInvRange			= 1.0f / fRange;
	const float				fShadowStrength		= pxInstance->GetStrength();
	const float				afLightPos[ 4 ]		= { xLightPos.x, xLightPos.y, xLightPos.z, 1.0f };
	const float				afShadowDir[ 4 ]	= { pxInstance->GetLightDirection().x, pxInstance->GetLightDirection().y, pxInstance->GetLightDirection().z, 1.0f };
	const float				afCamTanHalfFOV[ 2 ]= { fTanHalfHorizFOV, fTanHalfVertFOV };
	const float				afScrOffset[ 2 ]	= { Asura_Camera::GetViewingArea()->m_fXOffset * Asura_Viewport::GetWidth(), Asura_Camera::GetViewingArea()->m_fYOffset * Asura_Viewport::GetHeight() };
	const float				afTwoOverScr[ 2 ]	= { 2.0f / ( Asura_Viewport::GetWidth() * Asura_Camera::GetViewingArea()->m_fWidth ), 2.0f / ( Asura_Viewport::GetHeight() * Asura_Camera::GetViewingArea()->m_fHeight ) };

	pxInstance->GetLightOrientation( xLightOri );
	xShadowProj.SetPerspectiveForLH( Asura_Maths::DegToRad( xFrustum.GetVerticalFOVInDegrees() ), 1.0f, fNear, fFar );
	xShadowProj.Invert( xInvShadowProj );
	xView.Invert( xInvView );
	xViewProj.Invert( xInvViewProj );

	float fFilterFade = 1.0f;
	if( 1 )
	{
		fFilterFade = 0.25f * ( Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering - Asura_Maths::Max(pxInstance->GetDistance(), 0.0f ) );
		fFilterFade = Asura_Maths::ClampToRange( fFilterFade, 0.0f, 1.0f );
	}

	float fFilterTextureScale = fFilterFade * 11.0f / ( ( float )( ASURA_MAX_SHADOW_TEXTURE_SIZE ) );
	float fFilterSize = 0.2f * pxInstance->GetSoftness();

	if( bEnvMap )
	{
		Asura_AMatrix4x4 xShadowView, xInvScale;
		Asura_AMatrix4x4 xInvShadowView(
			1.0f, 0.0f, 0.0f, xLightPos.x,
			0.0f, 1.0f, 0.0f, xLightPos.y,
			0.0f, 0.0f, 1.0f, xLightPos.z,
			0.0f, 0.0f, 0.0f, 1.0f );
		Asura_AMatrix4x4 xScale( fFar, 0.0f, 0.0f, 0.0f,
								 0.0f, fFar, 0.0f, 0.0f,
								 0.0f, 0.0f, fFar, 0.0f,
								 0.0f, 0.0f, 0.0f, 1.0f );

		Asura_AMatrix4x4 xLightClipSpaceToCameraClipSpace = xScale * xInvShadowView * xViewProj;

		g_pxVertexShader->SetUniformParameter( VP_LightClipSpaceToCameraClipSpace, xLightClipSpaceToCameraClipSpace.GetElements() );
		g_pxFragmentShader->SetArray( FP_CameraViewSpaceToWorldSpace, 4, 4, xInvView.GetElements() );
		g_pxFragmentShader->SetFloats( FP_SelfShadowBias, 3, &s_xSelfShadowBiasValues[ 2 ][ 0 ] );
	}
	else
	{
		Asura_AMatrix4x4 xShadowView;
		Asura_AMatrix4x4 xInvShadowView( xLightOri, xLightPos );
		xInvShadowView.InvertOrthogonal( xShadowView );

		Asura_AMatrix4x4 xCameraViewSpaceToLightClipSpace = xInvView * xShadowView * xShadowProj;
		Asura_AMatrix4x4 xCameraViewSpaceToLightViewSpace = xInvView * xShadowView;

		if ( g_bUseFrustumVertices )
		{
			g_pxVertexShader->SetUniformParameter( VP_LightClipSpaceToCameraClipSpace, xViewProj.GetElements() );
			
			Asura_Vector_3* pVertices = reinterpret_cast<Asura_Vector_3*>(g_xShadowFrustumGeneratedVB.LockForWriting());
			xFrustum.GenerateVertices( pVertices, xLightPos, xLightOri );
		}
		else
		{
			// use static vertex data, a box, which will be transformed by this concatenated matrix
			// this is prone to errors on environment shadows cast over a long distance, causing the shadows to clip to a shape that doesn't exactly match the light source.
			Asura_AMatrix4x4 xLightClipSpaceToCameraClipSpace = xInvShadowProj * xInvShadowView * xViewProj;
			g_pxVertexShader->SetUniformParameter( VP_LightClipSpaceToCameraClipSpace, xLightClipSpaceToCameraClipSpace.GetElements() );
		}


		g_pxFragmentShader->SetArray( FP_CameraViewSpaceToLightClipSpace, 4, 4, xCameraViewSpaceToLightClipSpace.GetElements() );
		g_pxFragmentShader->SetArray( FP_CameraViewSpaceToLightViewSpace, 4, 4, xCameraViewSpaceToLightViewSpace.GetElements() );
		g_pxFragmentShader->SetFloats( FP_SelfShadowBias, 3, &s_xSelfShadowBiasValues[ pxInstance->IsEnvironment() ][ 0 ] );
	}

	g_pxFragmentShader->SetArray( FP_CameraViewSpaceToWorldSpace, 4, 4, xInvView.GetElements() );
	g_pxFragmentShader->SetFloats( FP_LightPos, 4, afLightPos );
	g_pxFragmentShader->SetFloats( FP_TwoOverScreen, 2, afTwoOverScr );
	g_pxFragmentShader->SetFloats( FP_ScreenOffset, 2, afScrOffset );
	g_pxFragmentShader->SetFloats( FP_CameraTanHalfFOV, 2, afCamTanHalfFOV );
	g_pxFragmentShader->SetFloats( FP_ShadowFilterTextureScale, 1, &fFilterTextureScale );
	g_pxFragmentShader->SetFloats( FP_ShadowFilterSize, 1, &fFilterSize );
	g_pxFragmentShader->SetFloats( FP_ShadowDirection, 4, afShadowDir );
	g_pxFragmentShader->SetFloats( FP_ShadowStrength, 1, &fShadowStrength );
	g_pxFragmentShader->SetFloats( FP_ShadowFrustumNearPlaneDistance, 1, &fNear );
	g_pxFragmentShader->SetFloats( FP_OneOverShadowFrustumPlaneSeparation, 1, &fInvRange );
	
	Asura_PS3_MaterialResponse_System::SetDepthParameters( g_pxFragmentShader, FP_DepthAdjustment );

	g_pxFragmentShader->SetTexture( FP_NormalMtlBuffer, Asura_PS3_DeferredRendering::GetDeferredFrameBuffer()->GetRTAsTexture( ASURA_DEFERRED_RENDER_TARGET_NORMAL_MATERIAL ), false );
	g_pxFragmentShader->SetTexture( FP_DepthBuffer, Asura_PS3_DeferredRendering::GetDeferredFrameBuffer()->GetZBufferAsTexture(), false );
	g_pxFragmentShader->SetTexture( FP_KernelRotationSampler, g_pxFilterKernelRotationTexture, true );
	g_pxFragmentShader->Bind();

#if 0
	// override the shader with a flat coloured one
	Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_GLOBAL, ASURA_GLOBAL_SHADER_BLANK )->Bind();
#endif

	if( bEnvMap )
	{
		// Are we inside the volume?
		Asura_Vector_3 xOfs = *Asura_Camera::GetPosition() - pxInstance->GetLightPosition();
		float fDist = xOfs.Magnitude();

		// Pick a sphere LOD. Switch down quickly to the low LOD, as we only need a more round
		// sphere when the camera is about to move inside it, for a more accurate inside-outside test
		int iSphereLOD = ( int )( fDist * 0.5f / pxInstance->GetFrustum().GetFarPlane() );
		if( iSphereLOD >= 3 )
		{
			iSphereLOD = 3;
		}
		else if( iSphereLOD < 0 )
		{
			iSphereLOD = 0;
		}

		Asura_PS3_DeferredShading::RenderSphereLOD( iSphereLOD );
	}
	else
	{
#if (!USE_QUADS)
		#error Triangles are no longer supported
#endif

		if ( g_bUseFrustumVertices )
		{
			//BW do I need to bind it before the inline draw? to setup values, perhaps.
			g_xShadowFrustumGeneratedVB.Bind( g_axVertexDeclaration, VD_ELEMENTS );

			g_xShadowFrustumIB.RenderQuadsInline( &g_xShadowFrustumGeneratedVB, 24 );
		}
		else
		{
			g_xShadowFrustumVB.Bind( g_axVertexDeclaration, VD_ELEMENTS );

			g_xShadowFrustumIB.RenderQuads( 24 );
		}
	}
}

bool g_bUseZTestForShadows = true;

void Asura_Shadow_System::Platform_RenderOnToScene( ASURA_SHADOW_LIGHTSOURCE eLightSource, uint32 uLightSourceHandle )
{
	const Asura_Shadow_Instance *apxCubeMapShadows[ ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS ];
	u_int uCubeMapShadowsFound = 0;

	if( 0 == s_uNumShadows)
	{
		return;
	}

	// For specific lights we handle setting up the profiling externally.
	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__APPLYSHADOWS, "Apply shadow" );

	Asura_PS3_DeferredRendering::BindAsSingleRT();
	Asura_PS3_RenderingStates::RequestAlphaTest( false );

    if( g_bUseZTestForShadows )
    {
	    Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
        Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_GREATER );
    }
    else
    {
	    Asura_PS3_RenderingStates::RequestZBufferEnabled( false );
    }

	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
	Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CCW );
	Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
	Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALL );

	if (eLightSource != ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
	{
		cellGcmSetBlendFuncUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_ONE, CELL_GCM_ONE, CELL_GCM_ONE, CELL_GCM_ONE );
		cellGcmSetBlendEquationUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_MAX, CELL_GCM_MAX );
		Asura_PS3_RenderingStates::InvalidateTranslucencyMode();
	}
	
#if 0
	Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
	Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
    Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
    Asura_PS3_RenderingStates::SetStencilTestEnabled(false);
    Asura_PS3_RenderingStates::SetAlphaBlending(false);

	//check depth vals of shadow by writing in z
	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled(true);
	Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_NONE );
#endif

	Asura_PS3_Graphics::ApplyCameraSettings();
	for( u_int u = 0; u < s_uNumShadows; ++u )
	{
			  Asura_PS3_Shadow&			xShadow		= *static_cast< Asura_PS3_Shadow* >( s_apxShadows[u] );
		const Asura_Shadow_Instance*	pxInstance	= xShadow.GetInstance();

		bool bAccept = false;

		if (pxInstance->GetLightSourceType() == eLightSource)
		{
			if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
			{
				// We're casting shadows from an averaged light source - there's no light source handle 
				bAccept = true;
			}
			else
			{
				// We're casting shadows from a real light source - make sure this is the right light source
				if (uLightSourceHandle == pxInstance->GetLightSourceHandle())
				{
					bAccept = true;					
				}
			}
		}
		else
		{
			if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
			{
				// If we're rendering averaged shadows, we also want to render any darkening pass 
				// shadows from specific light sources here, too
				if (pxInstance->GetShadowFlags()&ASURA_SHADOWFLAG_IS_DARKENING_PASS)
				{
					bAccept = true;
				}
			}
		}

		if (bAccept)
		{
			if( pxInstance->IsCubeMapped() )
			{
				if( uCubeMapShadowsFound < ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS )
				{
					apxCubeMapShadows[uCubeMapShadowsFound++] = pxInstance;
				}
			}
			else
			{
				Asura_Hash_ID	uHash = 0;

				if( pxInstance->GetDistance() < Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering )
				{
					uHash += 1;
				}
				if( pxInstance->GetShadowFlags() & ASURA_SHADOWFLAG_USE_NOISE )
				{
					uHash += 10;
				}
				if( eLightSource != ASURA_SHADOW_LIGHTSOURCE_AVERAGED )
				{
					uHash += 100;
				}
				g_pxFragmentShader = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_SHADOW, uHash );
				g_pxFragmentShader->SetTexture( FP_ShadowBuffer, xShadow.GetShadowBuffer()->GetTexture(), true );
				
				Asura_PS3_RenderingStates::SetTextureBorderColour( g_pxFragmentShader->GetTextureIndex(FP_ShadowBuffer), 0xffffffff );

				RenderShadowOnToScene( pxInstance, false );
				s_uNumShadowInstancesRendered++;
			}
		}
	}
	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__APPLYSHADOWS );

	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__APPLYSHADOWS, "Cubic Env Map Shadow" );
	
	//TODO: environment shadows
	for( u_int uCubeMappedShadow = 0; uCubeMappedShadow < uCubeMapShadowsFound; ++uCubeMappedShadow )
	{
		s_pxCubeMapFrameBuffer->BindAsSingleRT();
		g_pxFragmentShaderShadow->Bind();

		const Asura_Shadow_Instance *pxInstance = apxCubeMapShadows[uCubeMappedShadow];
		float fNear = pxInstance->GetFrustum().GetNearPlane();
		float fFar = pxInstance->GetFrustum().GetFarPlane();

		Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
		Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( true );
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF ); 

		//	Asura_CameraData xCamData = Asura_Camera::GetCameraData(Asura_Camera::GetCameraIndex());
		Asura_Vector_3 xCamPos		= *Asura_Camera::GetPosition();
		Asura_Matrix_3x3 xCamOrient	= *Asura_Camera::GetOrientation();
		Asura_Angle fCamFOV			=  Asura_Camera::GetHorizontalFieldOfView();
		float fCamNear				=  Asura_Camera::GetNearClip();
		float fCamFar				=  Asura_Camera::GetFarClip();
		float fPhysicalAspect		=  Asura_Camera::GetPhysicalScreenAspectRatio();
		float fCamAspect			=  Asura_Camera::GetXOverYAspectRatio() / fPhysicalAspect;
		Asura_ViewingArea xCamArea	= *Asura_Camera::GetViewingArea();
		Asura_Viewport::Set( ASURA_MAX_SHADOW_TEXTURE_SIZE, ASURA_MAX_SHADOW_TEXTURE_SIZE );
		Asura_PS3_RenderingStates::SetViewport( ASURA_MAX_SHADOW_TEXTURE_SIZE, ASURA_MAX_SHADOW_TEXTURE_SIZE );

		for( int i = 0; i < 6; ++i )
		{
			static const Asura_Matrix_3x3 axORIENTATIONS[ 6 ] =
			{
				// Looking at +x face
				Asura_Matrix_3x3(	 0.0f, 0.0f, 1.0f,
									 0.0f, 1.0f, 0.0f,
									-1.0f, 0.0f, 0.0f ),

				// Looking at -x face
				Asura_Matrix_3x3(	0.0f, 0.0f, -1.0f,
									0.0f, 1.0f,  0.0f,
									1.0f, 0.0f,  0.0f ),

				// Looking at -y face
				Asura_Matrix_3x3(	1.0f, 0.0f,  0.0f,
									0.0f, 0.0f, -1.0f,
									0.0f, 1.0f,  0.0f ),

				// Looking at +y face
				Asura_Matrix_3x3(	1.0f,  0.0f, 0.0f,
									0.0f,  0.0f, 1.0f,
									0.0f, -1.0f, 0.0f ),

				// Looking at +z face
				Asura_Matrix_3x3(	1.0f, 0.0f, 0.0f,
									0.0f, 1.0f, 0.0f,
									0.0f, 0.0f, 1.0f ),

				// Looking at -z face
				Asura_Matrix_3x3(	-1.0f, 0.0f,  0.0f,
									 0.0f, 1.0f,  0.0f,
									 0.0f, 0.0f, -1.0f ),
			};
			const Asura_Matrix_3x3 &xOrient = axORIENTATIONS[i];

			cellGcmSetClearSurfaceUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_CLEAR_Z | CELL_GCM_CLEAR_S );

			Asura_Camera::SetPosition( pxInstance->GetLightPosition() );
			Asura_Camera::SetPhysicalScreenAspectRatio(1.0f);
			Asura_Camera::SetHorizontalFieldOfView( Asura_Maths::Deg90, 1.0f  );
			Asura_Camera::SetOrientation( xOrient );
			Asura_Camera::SetNearClip( fCamNear );
			Asura_Camera::SetFarClip( fCamFar );
			Asura_Camera::SetViewingArea();
			Asura_PS3_EnvironmentRenderer::RenderToShadow( fNear, fFar, true );

			//Resolve to the corespondent Texture Or Send the Texture to the Environment Renderer
			s_pxCubeMapFrameBuffer->ResolveToCubemapTexture( &s_pxEnvShadowCubeTexture, i );
		}

		Asura_Hash_ID	uHash = 1000;

		if( pxInstance->GetDistance() < Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering )
		{
			uHash += 1;
		}
		if( pxInstance->GetShadowFlags() & ASURA_SHADOWFLAG_USE_NOISE )
		{
			uHash += 10;
		}
		if( eLightSource != ASURA_SHADOW_LIGHTSOURCE_AVERAGED )
		{
			uHash += 100;
		}

		g_pxFragmentShader = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_SHADOW, uHash );

		Asura_PS3_DeferredRendering::BindAsSingleRT();

		Asura_Viewport::Set( Asura_PS3_Graphics::GetScreenWidth(), Asura_PS3_Graphics::GetScreenHeight() );
		Asura_Camera::SetNearClip( fCamNear );
		Asura_Camera::SetFarClip( fCamFar );
		Asura_Camera::SetPosition( xCamPos );
		Asura_Camera::SetOrientation( xCamOrient );
		Asura_Camera::SetPhysicalScreenAspectRatio( fPhysicalAspect );
		Asura_Camera::SetHorizontalFieldOfView( fCamFOV, 1.0f / fCamAspect );
		Asura_Camera::SetViewingArea( xCamArea );

		Asura_PS3_Graphics::ApplyCameraSettings();
		Asura_PS3_Graphics::SetViewportToDefaults();

		Asura_PS3_DeferredRendering::BindAsSingleRT();
		Asura_PS3_RenderingStates::RequestAlphaTest( false );
		Asura_PS3_RenderingStates::RequestZBufferEnabled( false );
		Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
		Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CW );
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
		Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALL );
#if 0
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
#endif
		cellGcmSetBlendOptimizationUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_FALSE );
		if( pxInstance->GetShadowFlags() & ASURA_SHADOWFLAG_IS_DARKENING_PASS )
		{
			Asura_PS3_RenderingStates::SetColorMaskMrt0( ASURA_PS3_COLOR_MASK_MRT0_RGB );
		}
		else
		{
			Asura_PS3_RenderingStates::SetColorMaskMrt0( ASURA_PS3_COLOR_MASK_MRT0_ALPHA );
		}

		g_pxFragmentShader->SetTexture( FP_ShadowCubeBuffer, &s_pxEnvShadowCubeTexture, true );
		RenderShadowOnToScene( pxInstance, true );

		Asura_PS3_RenderingStates::SetColorMaskMrt0( ASURA_PS3_COLOR_MASK_MRT0_ALL );
	}
	
	cellGcmSetBlendEquationUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_FUNC_ADD, CELL_GCM_FUNC_ADD );
	Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_NOT_SET );
	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__APPLYSHADOWS );

	Asura_PS3_Graphics::Flush();
}

/******************************************************************************/
/* Asura_PS3_Shadow Class Functions                                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_Shadow::PrepareForRendering()
//
// Begins rendering into a shadow map
//
// Parameters:		xLightDirection - Direction of most important light source
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Shadow::PrepareForRendering( void )
{
	//TODO: Clear the surface by drawing a Quad
	if( m_pxShadowBuffer )
	{
		SetIsEnvironmentShadow( false );
		ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__MISC, "Render Shadow" );

		Asura_Viewport::Set( m_pxShadowBuffer->GetTextureSize(), m_pxShadowBuffer->GetTextureSize() );
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALL );
		Asura_PS3_RenderingStates::SetBlendEnableMrt( false, false, false );
		Asura_PS3_RenderingStates::SetAlphaBlending( false );
		Asura_PS3_RenderingStates::RequestAlphaTest( false );
		m_pxShadowBuffer->Bind();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_PS3_Shadow::FinishedRendering()
//
// Finishes rendering into a shadow map
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////
void Asura_PS3_Shadow::FinishedRendering( void )
{
	if( m_pxShadowBuffer )
	{
		m_pxShadowBuffer->UnBind();
		Asura_Viewport::Set(Asura_PS3_Graphics::GetScreenWidth(), Asura_PS3_Graphics::GetScreenHeight());
		Asura_PS3_Graphics::SetViewportToDefaults();
		Asura_PS3_Graphics::ApplyCameraSettings();
		ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__MISC );

	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Shadow_System::Platform_RenderEnvironmentShadowMap
// Author       : Andy Tate
// Date         : 13-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_RenderEnvironmentShadowMap(Asura_Shadow &xShadow)
{
	Asura_PS3_Shadow& xPS3Shadow = static_cast<Asura_PS3_Shadow&>(xShadow);

	if( xPS3Shadow.GetShadowBuffer() )
	{
		Asura_Vector_3 xCamPos = *Asura_Camera::GetPosition();
		Asura_Matrix_3x3 xCamOrient = *Asura_Camera::GetOrientation();
		Asura_Angle fCamFOV = Asura_Camera::GetHorizontalFieldOfView();
		float fNear = Asura_Camera::GetNearClip();
		float fFar = Asura_Camera::GetFarClip();
		float fPhysicalAspect = Asura_Camera::GetPhysicalScreenAspectRatio();
		float fCamAspect = Asura_Camera::GetXOverYAspectRatio() / fPhysicalAspect;
		Asura_ViewingArea xCamArea = *Asura_Camera::GetViewingArea();

		Asura_Matrix_3x3 xOrient;
		xShadow.GetLightOrientation(xOrient);

		Asura_Camera::SetPosition( xShadow.GetLightPosition() );
		Asura_Camera::SetPhysicalScreenAspectRatio(1.0f);
		Asura_Camera::SetHorizontalFieldOfView( Asura_Maths::DegToRad(xShadow.GetFrustum().GetHorizontalFOVInDegrees()), 1.0f);
		Asura_Camera::SetOrientation(xOrient);
		Asura_Camera::SetNearClip(xShadow.GetFrustum().GetNearPlane());
		Asura_Camera::SetFarClip(xShadow.GetFrustum().GetFarPlane());
		Asura_Camera::SetViewingArea();
		Asura_PS3_EnvironmentRenderer::RenderToShadow( xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane(), xShadow.GetInstance()->IsCubeMapped() );

		Asura_Camera::SetNearClip(fNear);
		Asura_Camera::SetFarClip(fFar);
		Asura_Camera::SetPosition(xCamPos);
		Asura_Camera::SetOrientation(xCamOrient);
		Asura_Camera::SetPhysicalScreenAspectRatio(fPhysicalAspect);
		Asura_Camera::SetHorizontalFieldOfView(fCamFOV, 1.0f / fCamAspect);
		Asura_Camera::SetViewingArea(xCamArea);
		Asura_PS3_Graphics::ApplyCameraSettings();

		xPS3Shadow.SetIsEnvironmentShadow( true );
	}
}

/// End of function 'Asura_Shadow_System::Platform_RenderEnvironmentShadowMap'. ///


/// End of file. ///
