///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_MultiStream.cpp
// Author    : Philip Hall
// Created   : 13/09/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifndef ASURA_USE_AUDIO_SYSTEM

// Asura headers
#include "Core/Asura_Maths.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_CommandConsole.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Sound/Asura_Sound.h"

// Asura PS3 headers
#include "Core/Asura_PS3_VectorMaths.h"
#include "Sound/Asura_PS3_Sound_System.h"
#include "FileAccess/Asura_PS3_FileSystem.h" 
#include "Core/Asura_PlatformBasics.h"

// This file's header
#include "Sound/Asura_PS3_MultiStream.h"

#if defined(ASURA_PS3_USE_MULTISTREAM) && (ASURA_PS3_USE_MULTISTREAM != 0)

// SDK headers
#include <errno.h>
#include <sys/timer.h>
#include <sys/ppu_thread.h>
#include <sysutil/sysutil_sysparam.h>
#include <cell/audio.h>
#include <cell/mstream.h>
#include <cell/sysmodule.h>
#include <cell/spurs/task.h>
#include <cell/spurs/control.h>
#include <spu_printf.h>

#include "Core/Asura_Memheap.h"

#if (CELL_SDK_VERSION < 0x200000)
#error ****************************************************
#error *    PS3 multistream requires SDK 200 or later     *
#error ****************************************************
#endif

//-----------------------------------------------------

// E X T E R N A L S /////////////////////////////////////////////////////////

extern u_int _binary_mstream_dsp_aux_send_pic_start;
extern u_int _binary_mstream_dsp_aux_return_pic_start;
extern u_int _binary_mstream_dsp_filter_pic_start;

// G L O B A L S /////////////////////////////////////////////////////////////

static volatile bool	g_bMSThreadExitFlag = false;
static int				g_iMSNumChannelsUsed = 0;

static Asura_MemHeap*	g_pxDSPHeap		= 0;
static void*			g_pDSPMemory	= 0;

// Constants
static const int		MS_SPU_THREAD_PRIORITY = 100;
static const int		MS_PPU_THREAD_PRIORITY = 0;
static const size_t		MS_THREAD_STACK_SIZE =  16 * 1024;				// 16Kb thread stack
DEBUGVAR usecond_t		MS_UPDATE_INTERVAL_SLEEP = 520;// 520 is from sample code. as (1000000 / 100) / 4; // Multistream reference PDF recommends minimum of 94updates/second (change third number to scale)
static const int		MS_REVERB_SLOTS = 8;

// SubBus mappings
static const int MS_REVERB_BUS_NUMBER		= 3;			// Reverb is on sub-bus 3
static const int MS_REVERB_BUS				= CELL_MS_BUS_FLAG | MS_REVERB_BUS_NUMBER;

static const int MS_DEAD_BUS_NUMBER			= 5;			// The dead bus just allows us to route from streams to master with zero gain https://ps3.scedev.net/forums/thread/47205
static const int MS_DEAD_BUS				= CELL_MS_BUS_FLAG | MS_DEAD_BUS_NUMBER;

// Streaming system
Asura_PS3_MultiStream_Streaming Asura_PS3_MultiStream::s_xStreamingAPI;

// Master output volumes
static const float afOUTPUT_VOLUMES[64] = {	1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f, 0.0f,
											0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f, 0.0f,
											0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 1.0f};

// Predefined EQ points	for underwater effect
static const CellEQPoint xEQ_UNDERWATER[]	= { {"1",     0.0f, 110.0f},
												{"2",   540.0f,  80.0f},
												{"3",  1260.0f,  22.0f},
												{"4",  1896.0f,  10.0f},
												{"5",  2592.0f,  10.0f},
												{"6",  3456.0f,   0.0f},
												{"7", 24000.0f,   0.0f} };

// Predefined EQ points	for deafened effect
static const CellEQPoint xEQ_DEAFENED[]		= { {"1",     0.0f, 120.0f},
												{"2",   576.0f,  40.0f},
												{"3",  1272.0f,  10.0f},
												{"4",  2136.0f,   5.0f},
												{"5", 24000.0f,   0.0f} };

// Global SPURS information for this system
static CellSpurs	g_xSpursInfo  __attribute__((aligned (128)));
static const int	SPURS_SPU_COUNT = 1;
static const int	SPURS_THREAD_PRIORITY = 250;

static const u_int ASURA_PS3_SOUND_MAX_DSP_PAGES = 2;
static const u_int uASURA_PS3_DSP_MEMORY_SIZE = 1572864;	//1.5Mb (Reverb is huge for some reason and takes about 1Meg alone)

#if defined _DEBUG
static int   dbg_iReverbType = Asura_Sound_System::ASURA_REVERB_ID_NONE;
static float dbg_fDryVolume = 1.0f;
static float dbg_fWetVolume = 1.0f;
#endif

static const float g_fFilterCutoff = 5000.0f;
static const float g_fFilterMaxGain = 42.0f;

DEBUGVAR int dbg_iMSDecay = 70;

#ifdef ASURA_PS3_SUPPORT_SULPHA
static const int iASURA_MS_SULPHA_NAMED_OBJECTS = 32;
static const int iASURA_MS_SULPHA_MEMORY_SIZE = 1024 * 1024 * 2;
#endif



// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

// Useful start->end interpolator function
inline float IPol(float fStart, float fEnd, float fRatio)
{
	return fStart + ((fEnd-fStart)*fRatio);
}

static int LoadDSPFromMemory(void* pDSPMemory)
{
	//TY 27.02.08 - make this a static function since multiple classes require the functionality
	CellMSDSP xDSPInfo;

	// Load DSP from the temporary image we have in memory
	int iDSPHandle = cellMSDSPLoadDSPFromMemory(pDSPMemory, &xDSPInfo, CELL_MS_ALLOC_DSPHANDLE, CELL_MS_ALLOC_PAGE);
	if ( iDSPHandle < 0 )
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultiStream::LoadDSPFromMemory] - DSP load failed\n"));
	}
	else
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultiStream::LoadDSPFromMemory] - DSP Handle %d used %d bytes\n", iDSPHandle, xDSPInfo.memoryUsed));
	}

	return iDSPHandle;
}


#if defined ASURA_PS3_USE_I3DL2_REVERB
////PGH/////////////////////////////////////////
// I3DL2 reverb data
////////////////////////////////////////////////
extern char _binary_mstream_dsp_i3dl2_pic_start[];
extern char _binary_mstream_dsp_i3dl2_pic_size[];

struct Asura_PS3_Sound_I3DL2ReverbPreset
{
	Asura_UnicodeChar*  m_szName;
	CellMSFXI3DL2Params	m_xReverb;
};

static const Asura_PS3_Sound_I3DL2ReverbPreset s_axI3DL2ReverbPresets[] =
{
	#include "Asura_PS3_Reverb_I3DL2Presets.h"
};

static const CellMSFXI3DL2Params &GetReverbParams(Asura_Sound_System::ReverbType eType)
{
	return s_axI3DL2ReverbPresets[eType].m_xReverb;
}


static void InterpolateReverbParams(CellMSFXI3DL2Params &xResult, const CellMSFXI3DL2Params &xStart, const CellMSFXI3DL2Params &xEnd, float fRatio)
{
	xResult.Room              =  IPol(xStart.Room,              xEnd.Room,              fRatio);
	xResult.Room_HF           =  IPol(xStart.Room_HF,           xEnd.Room_HF,           fRatio);
	xResult.Decay_time        =  IPol(xStart.Decay_time,        xEnd.Decay_time,        fRatio);
	xResult.Decay_HF_ratio    =  IPol(xStart.Decay_HF_ratio,    xEnd.Decay_HF_ratio,    fRatio);
	xResult.Reflections       =  IPol(xStart.Reflections,       xEnd.Reflections,       fRatio);
	xResult.Reflections_delay =  IPol(xStart.Reflections_delay, xEnd.Reflections_delay, fRatio);
	xResult.Reverb            =  IPol(xStart.Reverb,            xEnd.Reverb,            fRatio);
	xResult.Reverb_delay      =  IPol(xStart.Reverb_delay,      xEnd.Reverb_delay,      fRatio);
	xResult.Diffusion         =  IPol(xStart.Diffusion,         xEnd.Diffusion,         fRatio);
	xResult.Density           =  IPol(xStart.Density,           xEnd.Density,           fRatio);
	xResult.HF_reference      =  IPol(xStart.HF_reference,      xEnd.HF_reference,      fRatio);
	xResult.MixChannel        =  0;
	xResult.MixAmount         =  IPol(xStart.MixAmount,         xEnd.MixAmount,         fRatio);
	xResult.EarlyReflectionPattern = (fRatio > 0.5f) ? xEnd.EarlyReflectionPattern  :  xStart.EarlyReflectionPattern;
	xResult.LateReverbPattern      = (fRatio > 0.5f) ? xEnd.LateReverbPattern  :  xStart.LateReverbPattern;
	xResult.EarlyReflectionScaler  = (fRatio > 0.5f) ? xEnd.EarlyReflectionScaler  :  xStart.EarlyReflectionScaler;
	xResult.LF_reference           = IPol(xStart.LF_reference, xEnd.LF_reference, fRatio);
	xResult.Room_LF                = IPol(xStart.Room_LF,      xEnd.Room_LF,      fRatio);
}


#elif defined ASURA_PS3_USE_IR_REVERB
////PGH/////////////////////////////////////////
// Impulse response reverb data
////////////////////////////////////////////////
extern u_int _binary_mstream_dsp_IR_pic_start;
extern u_int _binary_mstream_dsp_IR_pic_size;


#elif defined ASURA_PS3_USE_PARAMETRIC_REVERB
////PGH/////////////////////////////////////////
// Parametric reverb data
////////////////////////////////////////////////

extern u_int _binary_mstream_dsp_reverb_pic_start;
extern u_int _binary_mstream_dsp_reverb_pic_size;

struct Asura_PS3_Sound_ParametricReverbPreset
{
	Asura_UnicodeChar*   m_szName;
	CellMSFXReverbParams m_xReverb;
};

static const Asura_PS3_Sound_ParametricReverbPreset s_axParametricReverbPresets[] = 
{
	#include "Asura_PS3_Reverb_ParametricPresets.h"
};

static const CellMSFXReverbParams &GetReverbParams(Asura_Sound_System::ReverbType eType)
{
	return s_axParametricReverbPresets[eType].m_xReverb;
}


static void InterpolateReverbParams(CellMSFXReverbParams &xResult, const CellMSFXReverbParams &xStart, const CellMSFXReverbParams &xEnd, float fRatio)
{
	xResult.fWetVolume  = IPol(xStart.fWetVolume , xEnd.fWetVolume , fRatio);
	xResult.fDryVolume  = IPol(xStart.fDryVolume , xEnd.fDryVolume , fRatio);
	xResult.fPreDelayMs = IPol(xStart.fPreDelayMs, xEnd.fPreDelayMs, fRatio);
	xResult.fLowPassHz  = IPol(xStart.fLowPassHz , xEnd.fLowPassHz , fRatio);
	xResult.fHighPassHz = IPol(xStart.fHighPassHz, xEnd.fHighPassHz, fRatio);
	xResult.fRoomSize   = IPol(xStart.fRoomSize  , xEnd.fRoomSize  , fRatio);
	xResult.fDamping    = IPol(xStart.fDamping   , xEnd.fDamping   , fRatio);
	xResult.fDiffusion  = IPol(xStart.fDiffusion , xEnd.fDiffusion , fRatio);
	xResult.fEarlyLate  = IPol(xStart.fEarlyLate , xEnd.fEarlyLate , fRatio);
	xResult.fLeftRight  = IPol(xStart.fLeftRight , xEnd.fLeftRight , fRatio);
	xResult.fFrontBack  = IPol(xStart.fFrontBack , xEnd.fFrontBack , fRatio);
}

#endif



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ReverbInfo::Asura_PS3_ReverbInfo
// Description : Init the reverb structure to default values.
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_ReverbInfo::Asura_PS3_ReverbInfo()
{
	SetReverb( Asura_Sound_System::ASURA_REVERB_ID_NONE, 1.0f, 1.0f );
}

void Asura_PS3_ReverbInfo::InterpolateTo(Asura_PS3_ReverbInfo &xResult, const Asura_PS3_ReverbInfo &xTarget, float fRatio)
{
	if (fRatio > 1.0f ) { fRatio = 1.0f; }
	InterpolateReverbParams(xResult.m_xParams, m_xParams, xTarget.m_xParams, fRatio);
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ReverbInfo::SetReverb
// Description : Load reverb information into this class
//
// Parameters  : 
//   Asura_Sound_System::ReverbType eType  - Reverb type
//   float fWetVolumeScale                 - Custom wet volume scalar
//   float fDryVolumeScale                 - Custom dry volume scalar
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_ReverbInfo::SetReverb(Asura_Sound_System::ReverbType eType, float fWetVolumeScale, float fDryVolumeScale)
{
	m_eReverbType = eType;
	m_xParams = GetReverbParams(eType);

#if defined ASURA_PS3_USE_I3DL2_REVERB
	// Wet volume, tweak these to modify scaling
	if ( fWetVolumeScale < 0.01f )
	{
		m_xParams.Room -= 10000.0f;		// Silent reverb
	}
	else
	{
		m_xParams.Room += log10f(fWetVolumeScale) * 1000.0f;
	}
	m_xParams.Room = Asura_Maths::ClampToRange(m_xParams.Room, -10000.0f, 0.0f);

	// Dry volume, tweak these to modify scaling
	if ( fDryVolumeScale < 0.01f )
	{
		m_xParams.Reverb -= 10000.0f;   // Silent reverb
	}
	else
	{
		m_xParams.Reverb += log10f(fDryVolumeScale) * 1000.0f;
	}
	m_xParams.Reverb = Asura_Maths::ClampToRange(m_xParams.Reverb, -10000.0f, 2000.0f);

#elif defined ASURA_PS3_USE_PARAMETRIC_REVERB
	m_xParams.fWetVolume *= fWetVolumeScale;
	m_xParams.fDryVolume *= fDryVolumeScale;
#endif	
}




////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::Asura_PS3_MultiStream
// Description : Default constructor - Init members to known defaults 
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_MultiStream::Asura_PS3_MultiStream()
{
	m_uMSPortNum = ~0;
	m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB] = NULL;
	m_pMSSystemMemory = NULL;
	m_fReverbIPolRatio = 0.0f;
	#ifdef ASURA_PS3_SUPPORT_SULPHA
	m_pSulphaMemory = NULL;
	#endif	
	
	for(u_int u = 0; u < ASURA_PS3_SOUND_DSP_EFFECT_NUM_HANDLES; ++u)
	{
		m_iDSPHandle [u] = -1;
		m_pDSPBuffer [u] = 0;
	}

	m_xStreamsToDestroy.Create(ASURA_PS3_SOUND_MAX_STREAMS_TO_DESTROY);
}

Asura_PS3_MultiStream::~Asura_PS3_MultiStream()
{
	m_xStreamsToDestroy.Destroy();
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::Initialise
// Description : Initialise the MultiStream subsystem - this is slightly 
//               involved as it requires libAudio and SPURS initialising too.
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultiStream::Initialise()
{
	int iResult;
	uint8_t auPriorities[8] = {1, 0, 0, 0, 0, 0, 0, 0};

	#ifdef ASURA_PS3_SUPPORT_SULPHA
	m_pSulphaMemory = Asura_AllocateAlign(iASURA_MS_SULPHA_MEMORY_SIZE, ASURA_ALIGN_16_BIT);
	if(m_pSulphaMemory)
	{
		cellMSSulphaInit(m_pSulphaMemory,iASURA_MS_SULPHA_MEMORY_SIZE,iASURA_MS_SULPHA_NAMED_OBJECTS);
	}
	BeginSulphaCapture();
	#endif

	// Initialise the SPURS system
	if ( !InitialiseSPURS() )
	{
		Asura_Assert(false, "Unable to initialise SPURS\n" );
		return false;
	}

	// Initialise the audio system (libAudio)
	if ( !InitialiseAudio() )
	{
		Asura_Assert(false, "Unable to initialise libAudio\n" );
		return false;
	}

	// Set the required MS configuration - NB we would be using iASURA_PS3_SOUND_MAX_VOICES for the max streams 
	// but https://ps3.scedev.net/forums/nodejump/216885 says when using routable streams each voice takes up a stream.
	CellMSSystemConfig xMSConfig = {iASURA_PS3_SOUND_MAX_VOICE_CHANNELS, 
									ASURA_PS3_SOUND_MAX_SUB_BUSES,
									ASURA_PS3_SOUND_MAX_DSP_PAGES,
									CELL_MS_ROUTABLE_STREAMS_FLAG | CELL_MS_DISABLE_SPU_PRINTF_SERVER | CELL_MS_TD_ONLY_512 | CELL_MS_SEPERATE_VOLUMES_PER_ROUTE }; 

	// Determine system memory requirements and allocate
	int iMemSize = cellMSSystemGetNeededMemorySize( &xMSConfig );
	Asura_OutputToDebugger(("[Asura_PS3_MultiStream::Initialise] - MultiStream using %d bytes of system memory\n", iMemSize));
	m_pMSSystemMemory = Asura_AllocateAlign(iMemSize, ASURA_ALIGN_128_BYTE);

	// Initialise a MultiStream SPURS thread
	iResult = cellMSSystemInitSPURS(m_pMSSystemMemory, &xMSConfig, &g_xSpursInfo, auPriorities);
	if ( iResult != 0 )
	{
		cellAudioQuit();
		Asura_Assertf(false, ("[Asura_PS3_MultiStream::Initialise] - cellMSSystemInitSPURS() error : %d\n", cellMSSystemGetLastError() ) );
		return false;
	}

	#ifndef ASURA_NO_DEBUG_OUTPUT
	// attach MS SPURS thread to our printf server
	sys_spu_thread_t xThreadID;
	u_int uNumThreads = 1;
	cellSpursGetSpuThreadId( &g_xSpursInfo, &xThreadID, &uNumThreads );
	iResult = spu_printf_attach_thread( xThreadID );
	Asura_Assertf(  iResult == CELL_OK, ("spu_printf_attach_thread() failed for multistream thread with code 0x%08X", iResult) );
	#endif

	// Initialise the streaming API early
	if ( !GetStreamingAPI().Initialise() )
	{
		cellAudioQuit();
		return false;
	}

	// Setup the MASTER output channel volumes to full (initially)
	cellMSCoreSetVolume64(CELL_MS_MASTER_BUS, CELL_MS_WET_AND_DRY, afOUTPUT_VOLUMES);

	// Set the volumes for the reverb bus
	cellMSCoreSetVolume64(MS_REVERB_BUS, CELL_MS_WET_AND_DRY, afOUTPUT_VOLUMES);

	// Set the global stream callback (called for _all_ stream events unless overridden, like for streamed sound playback)
	cellMSSystemSetGlobalCallbackFunc( &Asura_PS3_MultiStream::StreamGlobalCallback );

	// Start a thread to update the MultiStream/libaudio
	iResult = sys_ppu_thread_create(	&m_xMultiStreamPuThread
									,	&Asura_PS3_MultiStream::MultiStream_Update_Thread
									,	reinterpret_cast<uint64_t>(this)
									,	MS_PPU_THREAD_PRIORITY
									,	MS_THREAD_STACK_SIZE
									,	SYS_PPU_THREAD_CREATE_JOINABLE
									,	"MultiStream PU Thread");

	if ( iResult != 0 )
	{
		cellAudioQuit();
		Asura_Assertf( false, ("[Asura_PS3_MultiStream::Initialise] - sys_ppu_thread_create() error : %d\n", iResult) );
		return false;
	}

	Asura_OutputToDebugger(("[Asura_PS3_MultiStream::Initialise] - MultiStream update thread ID = %u\n", (int)m_xMultiStreamPuThread));

	#if defined _DEBUG
	Asura_CommandConsole::AddVar(L"Reverb.Type", &dbg_iReverbType, 0, 114, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"Reverb.DryVolume", &dbg_fDryVolume, 0, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"Reverb.WetVolume", &dbg_fWetVolume, 0, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	#endif

	#ifdef ASURA_PS3_SUPPORT_SULPHA
	cellMSSulphaNameObject( CELL_MS_SULPHA_NAMED_TYPE_SUB, MS_REVERB_BUS_NUMBER	, "Reverb Bus" );
	cellMSSulphaNameObject( CELL_MS_SULPHA_NAMED_TYPE_SUB, MS_DEAD_BUS_NUMBER	, "Dead Bus" );
	#endif

	//Create a memory area for DSP effects and a thread safe memory heap to manage it
	g_pDSPMemory = Asura_Allocate(uASURA_PS3_DSP_MEMORY_SIZE);

	g_pxDSPHeap = new Asura_MemHeap(g_pDSPMemory, uASURA_PS3_DSP_MEMORY_SIZE);
	g_pxDSPHeap->EnableThreadSafety();

	return InitDSPEffects();
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::InitialiseAudio
// Description : Initiailise the cell audio system - loads libAudio and
//				 configures the audio output port
// Parameters  : none
//
// Returns     : bool - returns tru on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultiStream::InitialiseAudio()
{
	int iResult;
	unsigned int uRetSysUtil; 

	#if CELL_SDK_VERSION >= 0x090002
	// Load the libaudio PRX.
	iResult = cellSysmoduleLoadModule( CELL_SYSMODULE_AUDIO );
	Asura_Assert( iResult == CELL_OK, "PRX for libaudio already loaded?" );
	#endif

	// Init the audio core
	iResult = cellAudioInit();
	if ( iResult != 0 )
	{
		Asura_Assertf( iResult == CELL_OK, ("[Asura_PS3_MultiStream::Initialise] - cellAudioInit() error : %d\n", iResult) );
		return false;
	}

	// Open an audio port
	CellAudioPortParam   xAudioParam;
	xAudioParam.nChannel = CELL_AUDIO_PORT_8CH;
	xAudioParam.nBlock   = CELL_AUDIO_BLOCK_16;
	xAudioParam.attr     = CELL_AUDIO_PORTATTR_INITLEVEL;
	xAudioParam.level	 = 1.0f;
	iResult = cellAudioPortOpen(&xAudioParam, &m_uMSPortNum);
	if ( iResult != CELL_OK )
	{
		Asura_Assertf( false, ("[Asura_PS3_MultiStream::Initialise] - cellAudioPortOpen() error : %d\n", iResult) );
		return false;
	}

	// Manual audio channel query
	CellAudioOutState xCurrentState;
	cellAudioOutGetState(CELL_AUDIO_OUT_PRIMARY, 0, &xCurrentState);  // There's only 1 primary output device.

	// This resets the chosen audio output channels and downmixing mechanism to the default

	#if CELL_SDK_VERSION < 0x220000
	// fix for DD not working under SDK 210, from forum post https://ps3.scedev.net/forums/nodejump/196338
	uRetSysUtil = cellMSSystemConfigureSysUtilEx_220(CELL_MS_AUDIOMODESELECT_SUPPORTSLPCM | CELL_MS_AUDIOMODESELECT_SUPPORTSDOLBY);
	#else
	uRetSysUtil = cellMSSystemConfigureSysUtilEx(CELL_MS_AUDIOMODESELECT_SUPPORTSLPCM | CELL_MS_AUDIOMODESELECT_SUPPORTSDOLBY);
	#endif
	Asura_OutputToDebugger(("[Asura_PS3_MultiStream::Initialise] - Channels = %d, Dolby = %d\n", (uRetSysUtil & 0x0000000F), (uRetSysUtil & 0x00000010) >> 4 ) );

	/*
	// This should allow you to test 5.1 surround with the PS3 devkits analogue outs.
	CellAudioOutConfiguration xOutConfig;
	xOutConfig.encoder = xCurrentState.soundMode.type;
	xOutConfig.channel = 6;
	xOutConfig.downMixer = CELL_AUDIO_OUT_DOWNMIXER_TYPE_B;
	iResult = cellAudioOutConfigure(CELL_AUDIO_OUT_PRIMARY, &xOutConfig, NULL, 0);
	*/

	// Configure the output port
	CellAudioPortConfig  xPortConfig;
	iResult = cellAudioGetPortConfig(m_uMSPortNum, &xPortConfig);
	Asura_Assertf( iResult == CELL_OK, ("[Asura_PS3_MultiStream::Initialise] - cellAudioGetPortConfig() error : %d\n", iResult) );

	// Allow output from MultiStream to libAudio - MultiStream REQUIRES 'libmstreamSPURSAT3.a' and 'libspurs_stub.a'
	iResult = cellMSSystemConfigureLibAudio(&xAudioParam, &xPortConfig);  
	Asura_Assertf( iResult == CELL_OK, ("[Asura_PS3_MultiStream::Initialise] - cellMSSystemConfigureLibAudio() error : %d\n", cellMSSystemGetLastError() ) );

	return true;	
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : InitialiseSPURS
// Description : Initialise the SPU Runtime System (SPURS)
//
// Parameters  : none
//
// Returns     : bool - True on success, false otherwise
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultiStream::InitialiseSPURS()
{
	sys_ppu_thread_t  thread_id;
	int               ppu_thr_prio __attribute__((aligned (128)));

	sys_ppu_thread_get_id(&thread_id);
	int iResult = sys_ppu_thread_get_priority(thread_id, &ppu_thr_prio);
	if ( iResult != CELL_OK )
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultiStream::Initialise] - sys_ppu_thread_get_priority() error %d\n", iResult));
		return false;
	}

	iResult = cellSpursInitialize(&g_xSpursInfo, SPURS_SPU_COUNT, SPURS_THREAD_PRIORITY, ppu_thr_prio - 1, 1);
	if ( iResult != 0 )
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultiStream::Initialise] - cellSpursInitialize() error %d\n", iResult));
		return false;
	}

	return true;
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::InitDSPEffects
// Description : Initialise/Load the required DSP effects linked into this ELF
//
// Parameters  : none
//
// Returns     : bool - returns true on success 
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultiStream::InitDSPEffects()
{
	int &iEffect = m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB];
	#if defined ASURA_PS3_USE_I3DL2_REVERB
	// Load the I3DL2 Reverb DSP
	iEffect = LoadDSPFromMemory( reinterpret_cast<void*>(&_binary_mstream_dsp_i3dl2_pic_start) );
	Asura_Assert(iEffect >= 0, "[Asura_PS3_MultiStream::InitDSPEffects] - I3DL2 DSP failed to load\n");

	// If the I3DL2 DSP program is available then initialise it
	if (iEffect >= 0)
	{
		if ( !InitI3DL2ReverbDSP() )
		{
			iEffect = -1;
		}
	}
	#elif defined ASURA_PS3_USE_IR_REVERB
	// Load the Impulse response reverb DSP
	iEffect = LoadDSPFromMemory( reinterpret_cast<void*>(&_binary_mstream_dsp_IR_pic_start) );
	// If the Impulse Response Reverb DSP program is available then initialise it
	if (iEffect >= 0)
	{
		if ( !InitImpulseResponseDSP() )
		{
			iEffect = -1;
		}
	}
	#elif defined ASURA_PS3_USE_PARAMETRIC_REVERB
	// Load the Parametric Reverb DSP
	iEffect = LoadDSPFromMemory( reinterpret_cast<void*>(&_binary_mstream_dsp_reverb_pic_start) );
	Asura_Assert(iEffect >= 0, "[Asura_PS3_MultiStream::InitDSPEffects] - Reverb DSP failed to load\n");

	// If the Reverb DSP program is available then initialise it
	if (iEffect >= 0)
	{
		if ( !InitParametricReverbDSP() )
		{
			iEffect = -1;
		}
	}
	#else
	#error "Reverb processor not selected";
	#endif

	// Set the initial reverb effect
	m_fReverbIPolRatio = 0.0f;
	m_xCurrentReverb.SetReverb( Asura_Sound_System::ASURA_REVERB_ID_NONE, 1.0f, 1.0f );
	m_xNextReverb.SetReverb( Asura_Sound_System::ASURA_REVERB_ID_NONE, 1.0f, 1.0f );
	UpdateDSPReverb(m_xNextReverb);

	//Lowpass filter
	m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_LPF] = LoadDSPFromMemory( reinterpret_cast<void*>(&_binary_mstream_dsp_filter_pic_start) );
	Asura_Assert(m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_LPF] >= 0, "[Asura_PS3_MultiStream::InitDSPEffects] - LPF DSP failed to load\n");

	//Aux Send
	m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_SEND] = LoadDSPFromMemory( reinterpret_cast<void*>(&_binary_mstream_dsp_aux_send_pic_start) );
	Asura_Assert(m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_SEND] >= 0, "[Asura_PS3_MultiStream::InitDSPEffects] - Aux send DSP failed to load\n");

	//Aux return
	m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN] = LoadDSPFromMemory( reinterpret_cast<void*>(&_binary_mstream_dsp_aux_return_pic_start) );
	Asura_Assert(m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN] >= 0, "[Asura_PS3_MultiStream::InitDSPEffects] - Aux return DSP failed to load\n");

	//Setup the aux return on the first DSP slot
	cellMSCoreSetDSP( MS_REVERB_BUS, CELL_MS_DSP_SLOT_0, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN]);
	void* pAuxReturnDSPParamAddr = cellMSCoreGetDSPParamAddr( MS_REVERB_BUS, CELL_MS_DSP_SLOT_0);
	if(pAuxReturnDSPParamAddr)
	{
		for (int iIndex = 0; iIndex < 8; iIndex++ )
		{
			cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_INMASK, CELL_MS_DSP_SLOT_0, iIndex, 0);	// Allow input from none
			cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_OUTMASK, CELL_MS_DSP_SLOT_0, iIndex, 0);	// Allow output to none
		}
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_INMASK,	CELL_MS_DSP_SLOT_0, 0, 1);
		
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_OUTMASK,	CELL_MS_DSP_SLOT_0, 0,		( 1 << CELL_MS_SPEAKER_FL ) 
																					+	( 1 << CELL_MS_SPEAKER_FR )
																					+	( 1 << CELL_MS_SPEAKER_RL ) 
																					+	( 1 << CELL_MS_SPEAKER_RR ) );
				
		int iMemSize = cellMSFXAuxReturnGetNeededMemorySize(1);
		m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN] = g_pxDSPHeap->Alloc( iMemSize, 16 );

		#ifdef ASURA_ASSERT
		int iRet = cellMSFXAuxReturnInit(pAuxReturnDSPParamAddr, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN], 1, MS_REVERB_BUS );
		Asura_Assertf(iRet >= 0, ("cellMSFXAuxReturnInit failed, error %d", cellMSSystemGetLastError()) );
		#else
		cellMSFXAuxReturnInit(pAuxReturnDSPParamAddr, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN], 1, MS_REVERB_BUS );
		#endif

		#ifdef ASURA_ASSERT
		iRet = cellMSFXAuxReturnSetThruMix( pAuxReturnDSPParamAddr, 0, 0.0f );
		Asura_Assertf(iRet >= 0, ("cellMSFXAuxReturnSetThruMix failed, error %d", cellMSSystemGetLastError()) );
		#else
		cellMSFXAuxReturnSetThruMix( pAuxReturnDSPParamAddr, 0, 0.0f );
		#endif

		cellMSCoreBypassDSP( MS_REVERB_BUS, CELL_MS_DSP_SLOT_0, CELL_MS_NOTBYPASSED );
	}

	#ifdef ASURA_PS3_SUPPORT_SULPHA
	//Name our DSPs
	cellMSSulphaNameObject(CELL_MS_SULPHA_NAMED_TYPE_DSP, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB]	, "Reverb");
	cellMSSulphaNameObject(CELL_MS_SULPHA_NAMED_TYPE_DSP, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_LPF]		, "Lowpass Filter");
	cellMSSulphaNameObject(CELL_MS_SULPHA_NAMED_TYPE_DSP, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_SEND]	, "Aux Send");
	cellMSSulphaNameObject(CELL_MS_SULPHA_NAMED_TYPE_DSP, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN], "Aux Return");
	#endif

	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::InitI3DL2ReverbDSP
// Description : Initialise the I3DL2 reverb DSP
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
#if defined ASURA_PS3_USE_I3DL2_REVERB 
bool Asura_PS3_MultiStream::InitI3DL2ReverbDSP()
{
	// Check if reverb DSP is actually available
	if ( m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB] < 0 ) { return false; }

	// Set DSP effect to slot 0 of stream buses
	cellMSCoreSetDSP( MS_REVERB_BUS, CELL_MS_DSP_SLOT_1, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB]);
	m_pI3DL2ParamAddr = cellMSCoreGetDSPParamAddr( MS_REVERB_BUS, CELL_MS_DSP_SLOT_1 );	// Get size required for active reverbs for a required slot.

	int iMemSize = cellMSFXI3DL2GetNeededMemorySize(MS_REVERB_SLOTS);
	m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB] = g_pxDSPHeap->Alloc(iMemSize,16);

	Asura_OutputToDebugger(("[Asura_PS3_MultiStream::InitI3DL2ReverbDSP] - Allocated Reverb buffer %dKb @ 0x%x\n", iMemSize/1024, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB]));

	// Initialise buffer to zero (required for SDK200 - oddly)
	Asura_Mem::Set(m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB], 0, (u_int)iMemSize);

	// Select the 'none' reverb	initially
	CellMSFXI3DL2Params xTempReverb = GetReverbParams(Asura_Sound_System::ASURA_REVERB_ID_NONE);

	int iResult = cellMSFXI3DL2Init(m_pI3DL2ParamAddr, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB], &xTempReverb, MS_REVERB_SLOTS );
	if ( iResult < 0)
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultiStream::InitI3DL2ReverbDSP] - Init reverb return failed\n"));
		return false;
	}

	// Set reverb DSP on input/output stage of reverb DSP (in slot 0)
	for (int iIndex = 0; iIndex < 8; iIndex++ )
	{
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_INMASK, CELL_MS_DSP_SLOT_1, iIndex, 0);	// Allow input from none
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_OUTMASK, CELL_MS_DSP_SLOT_1, iIndex, 0);	// Allow input from none
	}
	for (int iIndex = 0; iIndex < MS_REVERB_SLOTS; iIndex++ )
	{
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_INMASK, CELL_MS_DSP_SLOT_1, iIndex, 1<<iIndex);	// Allow input from speaker n
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_OUTMASK, CELL_MS_DSP_SLOT_1, iIndex, 1<<iIndex);	// Allow output to speaker n
	}

	// Set routing and turn effect on for the DSP assigned to slot 0
	cellMSCoreSetRouting( MS_REVERB_BUS, CELL_MS_MASTER_BUS, CELL_MS_WET_AND_DRY, CELL_MS_ROUTE_ON, 1.0f, 1.0f );
	cellMSCoreBypassDSP( MS_REVERB_BUS, CELL_MS_DSP_SLOT_1, CELL_MS_NOTBYPASSED );

	return true;
}
#endif



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::InitParametricReverbDSP
// Description : Initialise the reverb effect (if possible)
//
// Parameters  : none
//
// Returns     : bool - Returns trus on success, false to disable reverb
///////////////////////////////////////////////////////////////////////////////
#if defined ASURA_PS3_USE_PARAMETRIC_REVERB
bool Asura_PS3_MultiStream::InitParametricReverbDSP()
{
	// Check if reverb DSP is actually available
	if ( m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB] < 0 ) { return false; }

	// Get size required for 1 active reverb for a required slot.
	int iMemSize = cellMSFXReverbGetNeededMemorySize(MS_REVERB_SLOTS);
	m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB] = g_pxDSPHeap->Alloc(iMemSize,16);

	Asura_OutputToDebugger(("Allocated Reverb buffer size: 0x%x to address 0x%x\n", iMemSize, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB]));

	// Initialise reverb parameters (take from simple_reverb_td sample in the SDK)
	CellMSFXReverbParams xReverbParams;
	xReverbParams.fWetVolume  = 1.0f;
	xReverbParams.fDryVolume  = 1.0f;
	xReverbParams.fPreDelayMs = 120.0f;
	xReverbParams.fLowPassHz  = 20000.0f;
	xReverbParams.fHighPassHz = 1000.0f;
	xReverbParams.fRoomSize   = 1.0f;
	xReverbParams.fDamping    = 0.0f;
	xReverbParams.fDiffusion  = 0.2f;
	xReverbParams.fEarlyLate  = 0.5f;
	xReverbParams.fLeftRight  = 0.5f;
	xReverbParams.fFrontBack  = 0.5f;

	// Set DSP effect to slot 0 of stream
	cellMSCoreSetDSP( MS_REVERB_BUS, CELL_MS_DSP_SLOT_1, m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB]);
	void* pParamAddr = cellMSCoreGetDSPParamAddr( MS_REVERB_BUS, CELL_MS_DSP_SLOT_1 );

	int iResult = cellMSFXReverbInit(pParamAddr, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB], &xReverbParams, MS_REVERB_SLOTS );
	if ( iResult < 0)
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultiStream::InitDSPReverb] - Init reverb return failed\n"));
		return false;
	}

	// Set reverb DSP on input/output stage of reverb DSP (in slot 0)
	for (int iIndex = 0; iIndex < 8; iIndex++ )
	{
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_INMASK, CELL_MS_DSP_SLOT_1, iIndex, 0);	// Allow input from none
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_OUTMASK, CELL_MS_DSP_SLOT_1, iIndex, 0);	// Allow input from none
	}
	for (int iIndex = 0; iIndex < MS_REVERB_SLOTS; iIndex++ )
	{
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_INMASK, CELL_MS_DSP_SLOT_1, iIndex, 1<<iIndex);	// Allow input from speaker n
		cellMSCoreSetMask(MS_REVERB_BUS, CELL_MS_OUTMASK, CELL_MS_DSP_SLOT_1, iIndex, 1<<iIndex);	// Allow output to speaker n
	}

	// Turn effect on for the DSP assigned to slot 0
	cellMSCoreBypassDSP( MS_REVERB_BUS, CELL_MS_DSP_SLOT_1, CELL_MS_NOTBYPASSED );

	return true;
}
#endif




////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::InitImpulseResponseDSP
// Description : Initialise the impulse-response based DSP effect
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
#if defined ASURA_PS3_USE_IR_REVERB
bool Asura_PS3_MultiStream::InitImpulseResponseDSP()
{
	Asura_Assert(false, "Unimplemented\n");
	return true;
}
#endif

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::DSPReverb
// Description : Call this when the reverb parameters change
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultiStream::UpdateDSPReverb(Asura_PS3_ReverbInfo &xReverbInfo)
{
	#if defined ASURA_PS3_USE_I3DL2_REVERB
	// Channels go: L, R, C, LFE, RL, RR, EL, ER
	static int s_aiMix[MS_REVERB_SLOTS]	= {  1, 0, 
											-1,-1, 
											 5, 4, 
											 7, 6 };
	static CELL_MSFX_I3DL2_REFLECTIONS s_aiEarlyReverb[MS_REVERB_SLOTS] = {	CELL_MSFX_ROOM1_LEFT, CELL_MSFX_ROOM1_RIGHT, 
																			CELL_MSFX_ROOM1_LEFT, CELL_MSFX_ROOM1_RIGHT, 
																			CELL_MSFX_ROOM1_LEFT, CELL_MSFX_ROOM1_RIGHT,
																			CELL_MSFX_ROOM2_LEFT, CELL_MSFX_ROOM2_RIGHT	};
	static CELL_MSFX_I3DL2_LATE_MODEL s_aiLateReverb[MS_REVERB_SLOTS] = {	CELL_MSFX_ROOM1_LATE_LEFT, CELL_MSFX_ROOM1_LATE_RIGHT, 
																			CELL_MSFX_ROOM1_LATE_LEFT, CELL_MSFX_ROOM1_LATE_RIGHT, 
																			CELL_MSFX_ROOM1_LATE_LEFT, CELL_MSFX_ROOM1_LATE_RIGHT,
																			CELL_MSFX_ROOM1_LATE_LEFT, CELL_MSFX_ROOM1_LATE_RIGHT};

	for (int iIndex = 0; iIndex < MS_REVERB_SLOTS; iIndex++)
	{
		xReverbInfo.m_xParams.MixChannel             = s_aiMix[iIndex];
		xReverbInfo.m_xParams.EarlyReflectionPattern = s_aiEarlyReverb[iIndex];
		xReverbInfo.m_xParams.LateReverbPattern      = s_aiLateReverb[iIndex];
		cellMSFXI3DL2SetParams(m_pI3DL2ParamAddr, m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB], &xReverbInfo.m_xParams, iIndex);
	}
	#elif defined ASURA_PS3_USE_PARAMETRIC_REVERB
	for (int iIndex = 0; iIndex < MS_REVERB_SLOTS; iIndex++)
	{
		xReverbInfo.m_xParams.fLeftRight = (iIndex==0)||(iIndex==4) ? 0 : 1;
		xReverbInfo.m_xParams.fFrontBack = (iIndex<3) ? 0 : 1;
		cellMSFXReverbSetParams(m_pDSPBuffer[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB], &xReverbInfo.m_xParams, iIndex);
	}
	#endif

}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::MultiStream_Update_Thread
// Description : Update thread from multistream SCE samples - 
//               - Updates MultiStream to generate more PCM data for libAudio
//               - MultiStream callbacks are then called
// Parameters  : 
//   uint64_t param - 
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultiStream::MultiStream_Update_Thread(uint64_t param)
{
	Asura_PS3_MultiStream* pThis = reinterpret_cast<Asura_PS3_MultiStream*>(param);
	if ( pThis != NULL )
	{
		float* pOutputBuffer = NULL;

		cellAudioPortStart(pThis->m_uMSPortNum);

		// Loop until exit is requested by Asura_PS3_MultiStream::Shutdown
		while ( !g_bMSThreadExitFlag )
		{
			pOutputBuffer = cellMSSystemSignalSPU();		// Note - return buffer position is unusable

			//Destroy any sounds that we need to
			Asura_PS3_SoundToDestroy xNextToDestroy;
			while(pThis->GetNextSoundToDestroy(&xNextToDestroy))
			{
				//cellMSCoreStop(xNextToDestroy.m_iStreamHandle, 0 );
				cellMSStreamClose(xNextToDestroy.m_iStreamHandle);

				if ( xNextToDestroy.m_bStreamed )
				{
					if ( xNextToDestroy.m_xStreamingFile != Asura_PS3_FileSystem::INVALID_HANDLE )
					{
						Asura_PS3_MultiStream_Streaming::QueuedFileClose(xNextToDestroy.m_xStreamingFile);
					}
					Asura_DeallocateAlign(xNextToDestroy.m_pStreamBuffers[0]);
					Asura_DeallocateAlign(xNextToDestroy.m_pStreamBuffers[1]);
				}

				//delete the DSP effects
				if(xNextToDestroy.m_pOcclusionDSPBuffer)	{	g_pxDSPHeap->Free(xNextToDestroy.m_pOcclusionDSPBuffer);	}
				if(xNextToDestroy.m_pAuxSendDSPBuffer)		{	g_pxDSPHeap->Free(xNextToDestroy.m_pAuxSendDSPBuffer);		}
				if(xNextToDestroy.m_pDistanceDSPBuffer)		{	g_pxDSPHeap->Free(xNextToDestroy.m_pDistanceDSPBuffer);		}
			}

			sys_timer_usleep( MS_UPDATE_INTERVAL_SLEEP );
			cellMSSystemGenerateCallbacks();
		}

		cellAudioPortStop(pThis->m_uMSPortNum);
	}

	sys_ppu_thread_exit(0);
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::Update
// Description : Update the currently playing sounds
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultiStream::Update()
{
	#if defined _DEBUG
	if ( static_cast<Asura_Sound_System::ReverbType>(dbg_iReverbType) != m_xNextReverb.m_eReverbType )
	{
		SetReverb(static_cast<Asura_Sound_System::ReverbType>(dbg_iReverbType), dbg_fDryVolume, dbg_fWetVolume);
	}
	#endif
	// Blend global reverbs together if they've changed
	if ( m_xCurrentReverb.m_eReverbType != m_xNextReverb.m_eReverbType )
	{
		m_fReverbIPolRatio += 0.05f;  // This is more reliable than frame timing but of course varies with update rate.

		static Asura_PS3_ReverbInfo xReverbTemp;
		m_xCurrentReverb.InterpolateTo( xReverbTemp, m_xNextReverb, m_fReverbIPolRatio);	
		UpdateDSPReverb( xReverbTemp );

		if ( m_fReverbIPolRatio >= 1.0f )
		{
			m_xCurrentReverb = m_xNextReverb;
			m_fReverbIPolRatio = 0.0f;
		}
	}

	// Update individual sounds
	for (int iLoop=0; iLoop < GetMaxConcurrentSounds(); iLoop++ )
	{	
		Asura_PS3_ActiveSound &xActiveSound = Asura_PS3_ActiveSound::s_axActiveSounds[iLoop];
		if ( xActiveSound.IsValid() && xActiveSound.IsPlaying() )
		{
			xActiveSound.Update();
		}
	}
}


void Asura_PS3_MultiStream::Reset()
{
	// Sounds stopped by sound system	
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::Shutdown
// Description : Shutdown and clean up the libAudio and MultiStream systems
//
// Parameters  : none
//
// Returns     : bool - Returns true on success (Always)
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultiStream::Shutdown()
{
	GetStreamingAPI().Shutdown();

	// Ask update thread to exit and wait for it to complete (blocks)
	g_bMSThreadExitFlag = true;
	uint64_t uThreadResult = 0;
	sys_ppu_thread_join( m_xMultiStreamPuThread, &uThreadResult );

	ShutdownDSPEffects();

	// Shutdown MultiStream system
	#ifdef ASURA_ASSERT
	void* pMemory = 
	#endif // #ifdef ASURA_ASSERT
		cellMSSystemClose();
	Asura_Assert(pMemory == m_pMSSystemMemory, "[Asura_PS3_MultiStream::Shutdown] - Memory pointers should be identical\n");

	// Deallocate the memory we allocated in Initialise()
	if ( m_pMSSystemMemory != NULL )
	{
		Asura_DeallocateAlign(m_pMSSystemMemory);
	}
	m_pMSSystemMemory = NULL;

	int iResult = cellAudioQuit();
	Asura_Assertf( iResult == CELL_OK, ("[Asura_PS3_MultiStream::Shutdown] - cellAudioQuit() error\n") );

	#if CELL_SDK_VERSION >= 0x090002
	// Unload the libaudio PRX.
	iResult = cellSysmoduleUnloadModule( CELL_SYSMODULE_AUDIO );
	Asura_Assert( iResult == CELL_OK, "PRX for libaudio already unloaded?" );
	#endif

	Asura_Deallocate(g_pDSPMemory);
	delete g_pxDSPHeap;

	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::SetReverb
// Description : Set the reverb type - These need setting properly!!!
//
// Parameters  : 
//   Asura_Sound_System::ReverbType eType - Reverb type
//   float fDryVolumeScale - New dry volume scale factor
//   float fWetVolumeScale - New wet volume scale factor
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultiStream::SetReverb( Asura_Sound_System::ReverbType eType, float fDryVolumeScale, float fWetVolumeScale  )
{
	// Check if reverb DSP is actually available
	if ( m_iDSPHandle[ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_REVERB] < 0 ) { return; }

	#if defined _DEBUG
	dbg_iReverbType = eType;
	#endif

	#if defined ASURA_PS3_USE_I3DL2_REVERB
	Asura_OutputToDebugger_Retail(("[Asura_PS3_MultiStream::SetReverb] - Reverb set to %S\n", s_axI3DL2ReverbPresets[eType].m_szName));
	#elif defined ASURA_PS3_USE_PARAMETRIC_REVERB
	Asura_OutputToDebugger_Retail(("[Asura_PS3_MultiStream::SetReverb] - Reverb set to %S\n", s_axParametricReverbPresets[eType].m_szName));
	#else
	#error "Reverb type not implemented"
	#endif

	m_xNextReverb.SetReverb(eType, fWetVolumeScale, fDryVolumeScale);
	m_fReverbIPolRatio = 0.0f;
}

// MH quick function to help profile voice channel usage
int Asura_PS3_MultiStream::GetUsedVoiceChannels()
{
	return g_iMSNumChannelsUsed;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::ShutdownDSPEffects
// Description : Shutdown and clean up after the DSP effects 
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultiStream::ShutdownDSPEffects()
{
	// Clean up DSP memory
	g_pxDSPHeap->Wipe();
	for(u_int uDSP = 0; uDSP < ASURA_PS3_SOUND_DSP_EFFECT_NUM_HANDLES; ++uDSP)
	{
		m_pDSPBuffer[uDSP] = 0;
	}
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultiStream::StreamGlobalCallback
// Description : Global stream callback
//
// Parameters  : 
//   int iStreamHandle   - Handle to stream making callback
//   void* pUserData     - Pointer to custom user data
//   int iCallbackType   - Type of callback request
//   void* pWriteBuffer  - Next write buffer to use
//   int nBufferSize     - Size of write buffer
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultiStream::StreamGlobalCallback(int iStreamHandle, void* pUserData, int iCallbackType, void* pWriteBuffer, int nBufferSize)
{

	switch( iCallbackType )
	{
		case CELL_MS_CALLBACK_MOREDATA:
		{
		}
		break;

		case CELL_MS_CALLBACK_CLOSESTREAM:
		case CELL_MS_CALLBACK_FINISHSTREAM:
		{
		}
		break;

		default: break;
	}
}

bool Asura_PS3_MultiStream::GetNextSoundToDestroy (Asura_PS3_SoundToDestroy* pxOut)
{
	if(!pxOut)
	{
		return false;
	}
	bool bRet = false;
	m_xDestroyListMutex.Lock();
	Asura_PS3_SoundToDestroy* pxToDestroy = m_xStreamsToDestroy.Peek();
	if(pxToDestroy && Asura_Timers::GetMSecsSinceStart_Continuous() >= pxToDestroy->m_uDestroyTime)
	{
		Asura_Mem::Copy(pxOut, pxToDestroy, sizeof(Asura_PS3_SoundToDestroy));
		bRet = true;
		m_xStreamsToDestroy.Drop();
	}
	m_xDestroyListMutex.Unlock();
	return bRet;
}

void Asura_PS3_MultiStream::QueueSoundToDestroy (Asura_PS3_SoundToDestroy pxIn)
{
	m_xDestroyListMutex.Lock();

	Asura_PS3_SoundToDestroy* pxToDestroy = m_xStreamsToDestroy.Alloc();
	Asura_Assert(pxToDestroy, "Ran out of sound to destroy ring buffer space, maybe increase ASURA_PS3_SOUND_MAX_STREAMS_TO_DESTROY");
	if(pxToDestroy)
	{
		Asura_Mem::Copy(pxToDestroy, &pxIn, sizeof(Asura_PS3_SoundToDestroy));
	}
	m_xDestroyListMutex.Unlock();
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Asura_PS3_ActiveMSSound
// Description : MultiStream 'active' sound - initialise to known defaults
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
Asura_PS3_ActiveMSSound::Asura_PS3_ActiveMSSound()
{
	m_iMSStreamHandle		= -1;
	m_xStreamingFile		= Asura_PS3_FileSystem::INVALID_HANDLE;
	m_bUseHardware3d		= false;
	m_bStreamed				= false;
	m_fDistanceFilterGain	= 0.0f;
	m_fOcclusionFilterGain	= 0.0f;
	m_uStreamingBytesRead	= 0;
	m_pStreamBuffers[0]		= m_pStreamBuffers[1] = NULL;

	m_pOcclusionDSPBuffer	= 0;
	m_pOcclusionDSPParamAddr= 0;
	m_bOcclusionDSPActive	= false;

	m_pAuxSendDSPBuffer		= 0;
	m_pAuxSendDSPParamAddr	= 0;
	m_bAuxSendDSPActive		= false;

	m_pDistanceDSPBuffer	= 0;
	m_pDistanceDSPParamAddr = 0;
	m_bDistanceDSPActive	= false;;

	m_bUseReverb			= false;
							//a dynamic_cast would be safer here, but runtime type checking is expensive
	m_pxParent				= reinterpret_cast<Asura_PS3_MultiStream*>(&Asura_PS3_Sound_System::GetAPI());
	m_iNumChannels			= 0;
	m_bPausedByPitch		= false;
	m_bErrorOnStream		= false;
	m_bStreamFinished		= false;

	m_bSoundCanBeDeleted	= false;
}

Asura_PS3_ActiveMSSound::~Asura_PS3_ActiveMSSound()
{
	//Deallocate our local effects
	if(m_pOcclusionDSPBuffer)
	{
		g_pxDSPHeap->Free(m_pOcclusionDSPBuffer);
		m_pOcclusionDSPBuffer = 0;
	}
	if(m_pAuxSendDSPBuffer)
	{
		g_pxDSPHeap->Free(m_pAuxSendDSPBuffer);
		m_pAuxSendDSPBuffer = 0;
	}
	if(m_pDistanceDSPBuffer)
	{
		g_pxDSPHeap->Free(m_pDistanceDSPBuffer);
		m_pDistanceDSPBuffer = 0;
	}
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Create
// Description : Create a new active sound using multisound's Streams mechanism
//
// Parameters  : 
//   int iLoadedSoundIndex             - Index of loaded sound object
//   Asura_Handle iPlayingSoundHandle  - Playback handle from Asura_Sound_System
//   bool bUseEQ                       - Set true to use EQ for this sound
//   bool bUse3d                       - Set true to use 'hardware' 3d 
//
// Returns     : bool - Returns true if created sound is valid
///////////////////////////////////////////////////////////////////////////////

bool Asura_PS3_ActiveMSSound::Create(int iLoadedSoundIndex, Asura_Handle iSoundHandle, bool bUseEQ, bool bUse3d)
{
	if ( IsStreamValid() )
	{
		Asura_Assert(false, "[Asura_PS3_ActiveMSSound::Create] - Stream sound not destroyed before use\n");
		Destroy();
	}

	Asura_Assert(iLoadedSoundIndex >= 0 && iLoadedSoundIndex < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, "invalid loaded sound index - shouldn't have got this far");

	const Asura_PS3_LoadedSound& xLoadedSound = Asura_PS3_LoadedSound::s_axLoadedSounds[iLoadedSoundIndex];
	// assuming streamed sounds are mono - we don't actually load the number of channels in till later.
	m_bStreamed = xLoadedSound.IsStreamed();
	const int iNumChannels = m_bStreamed ? 1 : xLoadedSound.GetNumChannels();
	if (g_iMSNumChannelsUsed + iNumChannels > iASURA_PS3_SOUND_MAX_VOICE_CHANNELS)
	{
		Asura_Assert(false, "ran out of voice channels, probably too many stereo/multi-channel sounds");
		return false;
	}

	m_iMSStreamHandle = cellMSStreamOpen();
	if ( m_iMSStreamHandle < 0 )
	{
		Asura_Assertf(false, ("[Asura_PS3_ActiveMSSound::Create] - Unable to open new stream (error %d)\n", cellMSSystemGetLastError()));
		m_bErrorOnStream = true;
		return false;
	}

	SetSoundCanBeDeleted(true);

	m_bStreamFinished = false;

	g_iMSNumChannelsUsed += iNumChannels;

	// Set multistream callback data
	cellMSStreamSetCallbackData(m_iMSStreamHandle, this);
	cellMSStreamSetCallbackFunc(m_iMSStreamHandle, &Asura_PS3_MultiStream::StreamGlobalCallback );

	m_iLoadedSoundIndex = iLoadedSoundIndex;
	m_bDecaying = false;
	m_bUseEQ = bUseEQ;
	m_bUseHardware3d = bUse3d;
	m_bPausedByPitch = false;
	m_bErrorOnStream = false;

	// Special processing for streaming sources

	if ( m_bStreamed )
	{
		m_uStreamingBytesRead = 0;

		// Allocate streaming buffers
		m_pStreamBuffers[0] = Asura_AllocateAlign(ASURA_PS3_STREAMING_BUFFER_SIZE, ASURA_ALIGN_16_BYTE);
		m_pStreamBuffers[1] = Asura_AllocateAlign(ASURA_PS3_STREAMING_BUFFER_SIZE, ASURA_ALIGN_16_BYTE);

		// Adjust the callback if we're streaming
		cellMSStreamSetCallbackFunc(m_iMSStreamHandle, &Asura_PS3_ActiveMSSound::StreamLocalCallback );
	}

	Asura_Assert(!m_pOcclusionDSPBuffer, "Expected no occlusion buffer");
	Asura_Assert(!m_pAuxSendDSPBuffer, "Expected no aux send/return buffer");
	Asura_Assert(!m_pDistanceDSPBuffer, "Expected no distance buffer");

	return true;
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Destroy
// Description : Destroy this active stream sound (always succeeds)
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_ActiveMSSound::Destroy()
{
	while (!SoundCanBeDeleted()) {}

	//Queue this sound to be destroyed in the multistream callback
	if ( IsStreamValid() )
	{
		Asura_PS3_SoundToDestroy xToDestroy;

		xToDestroy.m_iStreamHandle = m_iMSStreamHandle;
		// give sounds which are fading out a little while before they get destroyed.
		xToDestroy.m_uDestroyTime = m_bDecaying ? (Asura_Timers::GetMSecsSinceStart_Continuous() + dbg_iMSDecay) : 0;

		if(m_bStreamed)
		{
			xToDestroy.m_bStreamed			= true;
			xToDestroy.m_xStreamingFile		= m_xStreamingFile;
			xToDestroy.m_pStreamBuffers[0]	= m_pStreamBuffers[0];
			xToDestroy.m_pStreamBuffers[1]	= m_pStreamBuffers[1];
		}
		else
		{
			xToDestroy.m_bStreamed			= false;
			xToDestroy.m_xStreamingFile		= 0;
			xToDestroy.m_pStreamBuffers[0]	= 0;
			xToDestroy.m_pStreamBuffers[1]	= 0;
		}

		xToDestroy.m_pOcclusionDSPBuffer	= m_pOcclusionDSPBuffer;
		xToDestroy.m_pAuxSendDSPBuffer		= m_pAuxSendDSPBuffer;
		xToDestroy.m_pDistanceDSPBuffer		= m_pDistanceDSPBuffer;

		if(m_pxParent)
		{
			m_pxParent->QueueSoundToDestroy(xToDestroy);
		}
	}

	Asura_Assert( m_iLoadedSoundIndex >= 0 && m_iLoadedSoundIndex < ASURA_MAX_NUMBER_OF_LOADED_SOUNDS, "invalid loaded sound index");
	const int iNumChannels = Asura_PS3_LoadedSound::s_axLoadedSounds[m_iLoadedSoundIndex].GetNumChannels();
	g_iMSNumChannelsUsed -= iNumChannels;
	Asura_Assert(g_iMSNumChannelsUsed >= 0, "bug in sound channel management");

	//Clear this active sound, everything is still allocated and active but will be cleaned up in
	//the multistream callback

	m_iMSStreamHandle		= -1;
	m_bPlaying				= false;

	m_bStreamed				= false;
	m_xStreamingFile		= Asura_PS3_FileSystem::INVALID_HANDLE;
	m_pStreamBuffers[0]		= 0;
	m_pStreamBuffers[1]		= 0;
	m_uStreamingBytesRead	= 0;

	//delete the occlusion filter
	m_pOcclusionDSPBuffer	= 0;
	m_bOcclusionDSPActive	= false;
	m_fOcclusionFilterGain	= 0.0f;

	m_pAuxSendDSPBuffer		= 0;
	m_bAuxSendDSPActive		= false;

	m_pDistanceDSPBuffer	= 0;
	m_bDistanceDSPActive	= false;
	m_fDistanceFilterGain	= 0.0f;

	m_iLoadedSoundIndex		= -1;

	m_bUseReverb			= false;
	m_iNumChannels			= 0;
	m_bChanged				= false;
	m_bErrorOnStream		= false;
	m_bStreamFinished		= true;
	m_bPausedByPitch		= false;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Update
// Description : Update this sound - recalculate 3d position values
//				 if necessary
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_ActiveMSSound::Update()
{
	if( !IsStreamValid() )
	{
		return;
	}

	if(IsStreamed() && m_bPending)
	{
		//This sound is streaming and cannot be updated yet
		return;
	}

	if(IsStreamFinished())
	{
		SetStreamFinished(false);	//dont come back into this logic

		m_bPending = false;
		m_bPlaying = false;

		if ( m_xStreamingFile != Asura_PS3_FileSystem::INVALID_HANDLE )
		{
			Asura_PS3_MultiStream_Streaming::QueuedFileClose(m_xStreamingFile);
		}
		//dont bother doing any further processing
		return;
	}

	// (MH - 18.01.2008) we no longer use hardware pan so this is redundant code.
	#if 0
	// Update for spatial sounds
	if ( m_bSpatial && m_bUseHardware3d )
	{
		Asura_AVector3  xDelta( m_xPosition - Asura_PS3_Sound_System::GetAPI().GetListener() );
		float           fDistance = xDelta.NormaliseReturningMagnitude();

		// PGH: Calculate 3D surround volumes (Asura_Sound_System::Calculate3dVolumePanAndFade 
		// sounds WRONG - and makes a mess of an extremely simple calculation)
		Asura_Matrix_3x3 xOrientation( Asura_PS3_Sound_System::GetAPI().GetOrientation() );
		xOrientation.Transpose();   // This will invert if the matrix contains _only_ rotations.

		float fVolume = 1.0f;
		float fPan = 0.5f;
		float fFade = 0.5f;

		// Only adjust volume for nonzero distances
		if ( fDistance > 0.001f )
		{
			if ( fDistance < m_fInnerRadius )
			{
				fVolume = 1.0f;
			}
			else if ( fDistance < m_fOuterRadius )
			{
				fVolume = (m_fOuterRadius - fDistance) / (m_fOuterRadius - m_fInnerRadius);
			}
			else
			{
				fVolume = 0.0f;
			}

			// Unrotate the delta back into camera space 
			xDelta = xDelta * xOrientation;

			fPan  = 0.5f + (xDelta.GetX() / 2.0f); 
			fFade = 0.5f + (xDelta.GetZ() / 2.0f);
		}

		m_bChanged = m_bChanged || (!Asura_Maths::Equal(m_fPan, fPan)		||
									!Asura_Maths::Equal(m_fVolume, fVolume) ||
									!Asura_Maths::Equal(m_fFade, fFade)		);
		m_fPan = fPan;
		m_fFade = fFade;
		m_fVolume = fVolume;
	}
	#endif

	// If settings have changed then apply the settings
	if ( m_bChanged )
	{
		m_bChanged = false;
		ApplySettings();
	}
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::UpdateDistanceFilter
// Description : Update this sound's distance filter (if there's been motion)
//
// Parameters  : 
//   float fDistance - Distance to listener
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////

void Asura_PS3_ActiveMSSound::UpdateDistanceFilter(float fGain)
{
	if ( IsStreamValid() )
	{
		// Only adjust the filter if the change in significant
		if ( Asura_Maths::Equal(m_fDistanceFilterGain, fGain, 0.01f) )
		{
			return;
		}

		//If the DSP is active, update it; otherwise dont, just cache the value
		if(m_bDistanceDSPActive)
		{
			CellMSFXFilter xFilterInfo;

			xFilterInfo.fFrequency		= g_fFilterCutoff;
			xFilterInfo.FilterMode		= CELL_MSFX_FILTERMODE_HIGHSHELF;
			xFilterInfo.fGain			= m_fDistanceFilterGain * -g_fFilterMaxGain;

			#ifdef ASURA_ASSERT
			int iRet = cellMSFXFilterSet( m_pDistanceDSPParamAddr, &xFilterInfo, -1);
			Asura_Assertf(iRet >= 0, ("cellMSFXFilterSet failed, error %d", cellMSSystemGetLastError() ) );
			#else
			cellMSFXFilterSet( m_pDistanceDSPParamAddr, &xFilterInfo, -1);
			#endif
		}
		else if ( ( fGain > 0.0f ) && m_bPlaying )
		{
			//The distance filter isn't active yet, lets create it
			m_fDistanceFilterGain = fGain;
			CreateDistanceFilter();
		}
	}
	// Record last gain 
	m_fDistanceFilterGain = fGain;
}


void Asura_PS3_ActiveMSSound::UpdateOcclusionFilter(float fGain)
{
	if ( IsStreamValid() )
	{
		// Only adjust the filter if the change in significant
		if ( Asura_Maths::Equal(m_fOcclusionFilterGain, fGain, 0.01f) )
		{
			return;
		}

		//If the DSP is active, update it; otherwise dont, just cache the value
		if(m_bOcclusionDSPActive)
		{
			CellMSFXFilter xFilterInfo;

			xFilterInfo.fFrequency		= g_fFilterCutoff;
			xFilterInfo.FilterMode		= CELL_MSFX_FILTERMODE_HIGHSHELF;
			xFilterInfo.fGain			= m_fOcclusionFilterGain * -g_fFilterMaxGain;

			#ifdef ASURA_ASSERT
			int iRet = cellMSFXFilterSet( m_pOcclusionDSPParamAddr, &xFilterInfo, -1);
			Asura_Assertf(iRet >= 0, ("cellMSFXFilterSet failed, error %d", cellMSSystemGetLastError() ) );
			#else
			cellMSFXFilterSet( m_pOcclusionDSPParamAddr, &xFilterInfo, -1);
			#endif
		}
		else if ( ( fGain > 0.0f ) && m_bPlaying )
		{
			//The occlusion filter isnt active yet, lets create it
			m_fOcclusionFilterGain = fGain;
			CreateOcclusionFilter();
		}
	}
	// Record last gain 
	m_fOcclusionFilterGain = fGain;
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::IsPlaying
// Description : Read stream status to determing if stream is playing
//
// Parameters  : none
//
// Returns     : bool - Returns true if stream is playing (including paused)
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::IsPlaying() const
{
	if ( m_iMSStreamHandle < 0 )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::IsPlaying] - invalid stream handle\n"));
		return false;
	}

	int iStatus = cellMSStreamGetStatus(m_iMSStreamHandle);
	m_bPlaying = (iStatus & CELL_MS_STREAM_ON) != 0;   // Cache playing state
	return m_bPlaying || (IsStreamed() && m_bPending);   // If streamed it could be pending playback
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::IsPaused
// Description : Read stream status to determing if stream is paused 
//
// Parameters  : none
//
// Returns     : bool - Returns true if stream is paused 
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::IsPaused() const
{
	if ( m_iMSStreamHandle < 0 )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::IsPaused] - invalid stream handle\n"));
		return false;
	}

	int iStatus = cellMSStreamGetStatus(m_iMSStreamHandle);
	return (iStatus & (CELL_MS_STREAM_PAUSE | CELL_MS_STREAM_PAUSE_SUB) ) != 0;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Play
// Description : Start this active sound playing (see the function
//               Asura_PS3_Sound_System::PlaySound for the entry point)
//
// Parameters  : 
//   bool bLooping - Set true to loop during playback
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::Play( bool bLooping )
{
	if ( !IsStreamValid() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::Play] - invalid stream handle\n"));
		return false;
	}

	if ( m_iLoadedSoundIndex < 0 )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::Play] - Loaded sound index is invalid"));
		return false;
	}

	// Check that the file isn't already being streamed by this mechanism
	if ( m_xStreamingFile != Asura_PS3_FileSystem::INVALID_HANDLE )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::Play] - Streaming sound is already playing"));
		return false;
	}

	m_bLooping = bLooping;

	// Retrieve information for the loaded sound
	Asura_PS3_LoadedSound&	xLoadedSound = Asura_PS3_LoadedSound::s_axLoadedSounds[m_iLoadedSoundIndex];
	if ( !xLoadedSound.IsLoaded() )
	{
		Asura_Assert(xLoadedSound.IsLoaded(), "[Asura_PS3_ActiveMSSound::Play] - Sound not loaded for playback\n");
		return false;
	}

	// File-streamed sound startup
	if ( m_bStreamed )
	{
		const char *szFilename = Asura_PS3_Sound_System::GetAPI().GetStreamingFilename( xLoadedSound.GetSourceFileIndex() );

		if ( !Asura_PS3_FileSystem::OpenForReading(szFilename, &m_xStreamingFile) )
		{
			Asura_Assert(false, "[Asura_PS3_ActiveMSSound::Play] - Unable to open streaming file\n");
			return false;
		}

		// Cache the sound info if required
		if ( xLoadedSound.NeedsSoundInfo() )
		{
			// Seek to file start
			if(!Asura_PS3_FileSystem::SetCurrentPosition(m_xStreamingFile, xLoadedSound.GetFileOffset() ) )
			{
				Asura_Assertf(false, ("[Multistream] Failed to seek in audio file %s", szFilename) );
				m_bErrorOnStream = true;
				Asura_PS3_FileSystem::Close(&m_xStreamingFile);
				return false;
			}

			// Load sound information
			Asura_PS3_Sound_Info xSoundInfo;
			Asura_PS3_Sound_Loader::LoadFileInfoAuto(m_xStreamingFile, xSoundInfo);
			Asura_Assert(xSoundInfo.GetChannelCount() == 1, "voice channel count is assuming streamed sounds e.g. dialogue are mono");
			xLoadedSound.SetSoundInfo(xSoundInfo);
		}

		// Seek to audio start
		if( !Asura_PS3_FileSystem::SetCurrentPosition(m_xStreamingFile, xLoadedSound.GetFileOffset() + xLoadedSound.GetSoundInfo().GetDataOffset() ) )
		{
			Asura_Assertf(false, ("[Multistream] Failed to seek in audio file %s", szFilename) );
			m_bErrorOnStream = true;
			Asura_PS3_FileSystem::Close(&m_xStreamingFile);
			return false;
		}
	}

	CellMSInfo xWaveInfo;
	xWaveInfo.SubBusGroup		= CELL_MS_MASTER_BUS;
	xWaveInfo.flags				= CELL_MS_STREAM_NOFLAGS; //CELL_MS_STREAM_AUTOCLOSE;
	xWaveInfo.initialOffset		= 0;                          // Initial delay (in samples) before playback starts
	xWaveInfo.inputType			= xLoadedSound.GetSoundInfo().GetSampleFormat();
	xWaveInfo.Pitch				= xLoadedSound.GetSampleRate();
	xWaveInfo.numChannels		= xLoadedSound.GetNumChannels();

	if ( m_bStreamed )
	{
		// Setup has to cope with over-read
		u_int uReadSize = Asura_Maths::Min( ASURA_PS3_STREAMING_BUFFER_SIZE, xLoadedSound.GetSampleDataSize() );
		xWaveInfo.FirstBufferSize	= Asura_PS3_FileSystem::ReadData( m_xStreamingFile, uReadSize, m_pStreamBuffers[0] );
		xWaveInfo.FirstBuffer		= (xWaveInfo.FirstBufferSize > 0) ? m_pStreamBuffers[0] : NULL;

		if ( uReadSize < xLoadedSound.GetSampleDataSize() )
		{
			uReadSize = Asura_Maths::Min( ASURA_PS3_STREAMING_BUFFER_SIZE, xLoadedSound.GetSampleDataSize() - uReadSize );
			xWaveInfo.SecondBufferSize = Asura_PS3_FileSystem::ReadData( m_xStreamingFile, uReadSize, m_pStreamBuffers[1] );
		}
		else
		{
			xWaveInfo.SecondBufferSize = 0;
		}
		xWaveInfo.SecondBuffer = (xWaveInfo.SecondBufferSize > 0) ? m_pStreamBuffers[1] : NULL;

		m_uStreamingBytesRead  = xWaveInfo.FirstBufferSize + xWaveInfo.SecondBufferSize;
	}
	else
	{
		xWaveInfo.FirstBuffer		= xLoadedSound.GetSampleData();
		xWaveInfo.FirstBufferSize	= xLoadedSound.GetSampleDataSize();
		xWaveInfo.SecondBuffer		= m_bLooping ? xLoadedSound.GetSampleData() : NULL;
		xWaveInfo.SecondBufferSize	= m_bLooping ? xLoadedSound.GetSampleDataSize() : 0;
	}

	if ( IsStreamValid() )
	{
		m_iNumChannels = xLoadedSound.GetNumChannels();

		// Bind audio information to the current stream
		int iResult = cellMSStreamSetInfo(m_iMSStreamHandle, &xWaveInfo);
		if ( iResult != CELL_OK )
		{
			m_bErrorOnStream = true;
			Asura_Assertf(false, ("[Asura_PS3_ActiveMSSound::Play] - cellMSStreamSetInfo() error: %d\n", cellMSSystemGetLastError()) );
			return false;
		}

		if(!m_bStreamed && m_bLooping)
		{
			cellMSStreamSetFirstReadSkip	(m_iMSStreamHandle, xLoadedSound.GetSampleData(),xLoadedSound.GetSampleDataSize(),0,0);
			cellMSStreamSetSecondReadSkip	(m_iMSStreamHandle, xLoadedSound.GetSampleData(),xLoadedSound.GetSampleDataSize(),0,0);
		}

		// Initial settings application (including volume & distance filtering)
		m_bChanged = true;
		m_bPausedByPitch = false;

		Update();

		// Set this stream playing
		if(!m_bPausedByPitch)
		{
			cellMSStreamPlay(m_iMSStreamHandle);
		}
		else
		{
			cellMSStreamPlay(m_iMSStreamHandle);
			cellMSCoreSetPause(m_iMSStreamHandle, CELL_MS_PAUSED);
		}

		m_bPlaying = true;
		m_bPending = false;

		//If this is a mono sound we just want the filtered path
		if( IsMono() )
		{
			cellMSStreamSetRoutingDry( m_iMSStreamHandle, CELL_MS_MASTER_BUS, 0.0f);
		} 

		//If we previously cached a non-zero distance filter gain, create and apply it now
		if( ( m_fDistanceFilterGain > 0.0f ) && !m_bDistanceDSPActive )
		{
			CreateDistanceFilter ();
		}

		//If we previously cached a non-zero occlusion filter gain, create and apply it now
		if( ( m_fOcclusionFilterGain > 0.0f ) && !m_bOcclusionDSPActive)
		{
			CreateOcclusionFilter ();
		}

		//If we are using reverb, create and aux send to the reverb bus
		if( ( m_bUseReverb ) && !m_bAuxSendDSPActive)
		{
			CreateAuxSend ();
		}

		//Route to the 'dead' bus so that processing will always happen on a stream even if volume is zero
		cellMSStreamSetRoutingDry( m_iMSStreamHandle, MS_DEAD_BUS, 1.0f);

		ApplySettings();
	}

	return true;
}


void Asura_PS3_ActiveMSSound::CreateOcclusionFilter ()
{
	Asura_Assert(!m_bOcclusionDSPActive,"occlusion filter is active, but we are about to create it");

	if( !IsStreamValid() )
	{
		Asura_Assert(false, "Stream is not valid, cannot create distance filter");
		return;
	}

	if(!m_pxParent)
	{
		return;
	}

	int iDSPHandle = m_pxParent->GetDSPHandle(ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_LPF);

	// Set DSP effect to slot 2 of stream
	cellMSCoreSetDSP( m_iMSStreamHandle, CELL_MS_DSP_SLOT_2, iDSPHandle);
	m_pOcclusionDSPParamAddr = cellMSCoreGetDSPParamAddr( m_iMSStreamHandle, CELL_MS_DSP_SLOT_2 );

	Asura_Assert( m_pOcclusionDSPParamAddr, "Failed to get DSP Param addresss" );
	if(m_pOcclusionDSPParamAddr)
	{
		int iMemSize = cellMSFXFilterGetNeededMemorySize(1);
		m_pOcclusionDSPBuffer = g_pxDSPHeap->Alloc(iMemSize,16);

		Asura_Assert(m_pOcclusionDSPBuffer, "Failed to allocate space for DSP effect");
		if(m_pOcclusionDSPBuffer)
		{
			// Initialise buffer to zero (required for SDK200 - oddly)
			Asura_Mem::Set(m_pOcclusionDSPBuffer, 0, (u_int)iMemSize);

			CellMSFXFilter xFilterInfo;

			xFilterInfo.fFrequency		= g_fFilterCutoff;
			xFilterInfo.FilterMode		= CELL_MSFX_FILTERMODE_HIGHSHELF;
			xFilterInfo.fGain			= m_fOcclusionFilterGain * -g_fFilterMaxGain;

			int iRet = cellMSFXFilterInit( m_pOcclusionDSPParamAddr, m_pOcclusionDSPBuffer, &xFilterInfo );
			Asura_Assert(iRet >= 0, "Failed to set filter params");
			if(iRet >= 0)
			{
				cellMSCoreBypassDSP( m_iMSStreamHandle, CELL_MS_DSP_SLOT_2, CELL_MS_NOTBYPASSED );
				m_bOcclusionDSPActive = true;
			}
		}
	}
}

void Asura_PS3_ActiveMSSound::CreateAuxSend ()
{
	Asura_Assert(!m_bAuxSendDSPActive,"aux send is active, but we are about to create it");
	Asura_Assert(m_bUseReverb,"why are we sending data to the reverb bus when we are not using reverb?");
	if(!m_pxParent)
	{
		Asura_Assert(m_pxParent, "Active sound has no pointer to parent");
		return;
	}

	if(!IsStreamValid())
	{
		Asura_Assert(false, "trying to create an aux send dsp on an invalid stream");
		return;
	}

	int iDSPHandle = m_pxParent->GetDSPHandle(ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_SEND);

	// Set DSP effect to slot 1 of stream bus
	cellMSCoreSetDSP( m_iMSStreamHandle, CELL_MS_DSP_SLOT_1, iDSPHandle);
	m_pAuxSendDSPParamAddr = cellMSCoreGetDSPParamAddr( m_iMSStreamHandle, CELL_MS_DSP_SLOT_1);

	if(m_pAuxSendDSPParamAddr)
	{
		CellMSFXAuxSend xSend;

		xSend.nSendChannels = 1;		//Stream DSP is mono
		xSend.nSendsPerChannel = 1;	

		int iMemSize = cellMSFXAuxSendGetNeededMemorySize(&xSend);
		m_pAuxSendDSPBuffer = g_pxDSPHeap->Alloc(iMemSize,16);

		int iRet = cellMSFXAuxSendInit(m_pAuxSendDSPParamAddr, m_pAuxSendDSPBuffer, &xSend );
		Asura_Assert(iRet >= 0, "cellMSFXAuxSendInit failed");
		if(iRet >= 0)
		{
			//Set the send to the reverb bus
			iRet = cellMSFXAuxSendSet(	m_pAuxSendDSPBuffer
									,	0
									,	0
									,	m_pxParent->GetDSPMemory(ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_AUX_RETURN)
									,	0
									,	0.0f );
			if(iRet < 0)
			{
				Asura_Assertf( false, ( "cellMSFXAuxSendSet failed, error : %d\n", cellMSSystemGetLastError() ) )
				
				return;
			}
			
			cellMSCoreBypassDSP( m_iMSStreamHandle, CELL_MS_DSP_SLOT_1, CELL_MS_NOTBYPASSED );
			m_bAuxSendDSPActive = true;
		}
	}
}


void Asura_PS3_ActiveMSSound::CreateDistanceFilter	()
{
	Asura_Assert(!m_bDistanceDSPActive,"distance filter is active, but we are about to create it");

	if( !IsStreamValid() )
	{
		Asura_Assert(false, "Stream is not valid, cannot create distance filter");
		return;
	}

	if(!m_bUseDistanceFilter || !m_pxParent)
	{
		return;
	}

	int iDSPHandle = m_pxParent->GetDSPHandle(ASURA_PS3_SOUND_DSP_EFFECT_HANDLE_LPF);

	if(iDSPHandle >= 0)
	{
		// Set DSP effect to slot 0 of stream buses
		cellMSCoreSetDSP( m_iMSStreamHandle, CELL_MS_DSP_SLOT_0, iDSPHandle);
		m_pDistanceDSPParamAddr = cellMSCoreGetDSPParamAddr( m_iMSStreamHandle, CELL_MS_DSP_SLOT_0 );

		Asura_Assert( m_pDistanceDSPParamAddr, "Failed to get DSP Param addresss" );
		if(m_pDistanceDSPParamAddr)
		{
			int iMemSize = cellMSFXFilterGetNeededMemorySize(1);
			m_pDistanceDSPBuffer = g_pxDSPHeap->Alloc(iMemSize,16);

			Asura_Assert(m_pDistanceDSPBuffer, "Failed to allocate space for DSP effect");
			if(m_pDistanceDSPBuffer)
			{
				// Initialise buffer to zero (required for SDK200 - oddly)
				Asura_Mem::Set(m_pDistanceDSPBuffer, 0, (u_int)iMemSize);

				CellMSFXFilter xFilterInfo;

				xFilterInfo.fFrequency		= g_fFilterCutoff;
				xFilterInfo.FilterMode		= CELL_MSFX_FILTERMODE_HIGHSHELF;
				xFilterInfo.fGain			= m_fDistanceFilterGain * -g_fFilterMaxGain;

				int iRet = cellMSFXFilterInit( m_pDistanceDSPParamAddr, m_pDistanceDSPBuffer, &xFilterInfo );
				Asura_Assert(iRet >= 0, "Failed to set filter params");
				if(iRet >= 0)
				{
					cellMSCoreBypassDSP( m_iMSStreamHandle, CELL_MS_DSP_SLOT_0, CELL_MS_NOTBYPASSED );
					m_bDistanceDSPActive = true;
				}
			}
		}
	}
}

////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Pause
// Description : Pause this stream
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::Pause()
{
	if ( !IsStreamValid() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::Pause] - invalid stream handle\n"));
		return false;
	}

	if ( m_bPlaying )
	{
		cellMSCoreSetPause(m_iMSStreamHandle, CELL_MS_PAUSED);
	}
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Unpause
// Description : Resume this stream
//
// Parameters  : none
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::Unpause()
{
	if ( !IsStreamValid() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::Unpause] - invalid stream handle\n"));
		return false;
	}

	if ( m_bPlaying )
	{
		cellMSCoreSetPause(m_iMSStreamHandle, CELL_MS_NOTPAUSED);
	}
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::Stop
// Description : Stop this sound from playing (immediate)
//
// Parameters  : none
//	int iDecayMS - Decay time in milliseconds
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::Stop(int iDecayMS)
{
	if ( !IsStreamValid() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::Stop] - invalid stream handle\n"));
		return true; // Stream doesn't exist
	}

	if ( m_bPlaying )
	{
		m_bDecaying = true;
		cellMSCoreStop(m_iMSStreamHandle, dbg_iMSDecay);
	}

	m_bPlaying = false;
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::SetUseReverb
// Description : Turn on reverb for this sample (i.e. switch it between sub buses)
//
// Parameters  : 
//   bool bReverb - Set true to enable reverb, false otherwise
//
// Returns     : bool - Returns true on success (always)
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::SetUseReverb( bool bReverb )
{
	m_bUseReverb = bReverb;
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::SetEQType
// Description : Set the EQ type for the playback of this sound
//
// Parameters  : 
//   Asura_Sound_System::EQType eType - EQ type
//
// Returns     : bool - Returns true on success
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_ActiveMSSound::SetEQType(Asura_Sound_System::EQType eType)
{
	if ( !IsStreamValid() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::SetEQType] - invalid stream handle"));
		return false;
	}
	/*
	if ( eType != Asura_Sound_System::EQ_OFF )
	{
		// Calculate EQ curve based on selected EQ type
		switch ( eType )
		{
			case Asura_Sound_System::EQ_DEAFENED:
			{
				CellEQSection xEQSection;
				xEQSection.curveSwitch = CELL_MS_EQ_LINE;
				xEQSection.nPoints = sizeof(xEQ_DEAFENED)/sizeof(xEQ_DEAFENED[0]);
				xEQSection.xy = const_cast<CellEQPoint*>(xEQ_DEAFENED);

				cellMSEQGet512Points( &xEQSection, m_fEQTable ); 
			}
			break;		

			case Asura_Sound_System::EQ_UNDERWATER:
			{
				CellEQSection xEQSection;
				xEQSection.curveSwitch = CELL_MS_EQ_LINE;
				xEQSection.nPoints = sizeof(xEQ_UNDERWATER)/sizeof(xEQ_UNDERWATER[0]);
				xEQSection.xy = const_cast<CellEQPoint*>(xEQ_UNDERWATER);

				cellMSEQGet512Points( &xEQSection, m_fEQTable ); 
			}	
			break;		

			default: break;
		}

		// Enable and use the calculated EQ curve
		cellMSCoreSetEQ(m_iMSStreamHandle, m_fEQTable);
		cellMSCoreSetEQEnable(m_iMSStreamHandle, CELL_MS_DRY, CELL_MS_EQENABLED);
	}
	else
	{
		cellMSCoreSetEQEnable(m_iMSStreamHandle, CELL_MS_DRY, CELL_MS_EQDISABLED);
	}*/

	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_ActiveMSSound::ApplySettings
// Description : Apply volume, pan, fade and pitch settings to this stream's 
//               speaker volumes to hopefully position it correctly.
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////

DEBUGVAR bool bUseNewSpacialisation = true;

void Asura_PS3_ActiveMSSound::ApplySettings()
{
	if ( !IsStreamValid() )
	{
		Asura_OutputToDebugger(("[Asura_PS3_ActiveMSSound::ApplySettings] - invalid stream handle"));
		return;
	}

	#ifdef ASURA_ASSERT
	int iRet = 0;
	#endif

	//TY - volume modifier, unfortunately completely empirical, mainly to prevent clipping on bassy sounds
	const float fVolumeMod = 0.4f;

	float fVolume = m_fVolume  * fVolumeMod;

	// Resulting volume calculations
	float afVolumes[8];

	if( IsMono() )
	{
		if(bUseNewSpacialisation)
		{
			const float fCartesianX = (m_fPan * 2.0f - 1.0f);
			const float fCartesianY = (m_fFade * 2.0f - 1.0f);

			Asura_Vector_2 xOriginToSound(fCartesianX,fCartesianY);
			xOriginToSound.Normalise();

			Asura_Vector_2 xOriginToSpeaker;

			//take the angle to each speaker (origin->sound x origin->speaker)
			//we precalc normalised vectors to the speakers, assuming they are positioned on a square '
			
			const float fOneOverRootTwo = 0.7071f;			//This is used for defining speaker positions
			const float fBleed = 0.5f;						//by default each speaker renders sounds up to 180 degrees away in the sound field, this value increases it (1.0 = 360 degrees)
			const float fBleedMul = 1.0f / (1.0f + fBleed);	//see above
			
			afVolumes[CELL_MS_SPEAKER_FL]	= Asura_Maths::Max(((Asura_Vector_2(-fOneOverRootTwo, fOneOverRootTwo	) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);	
			afVolumes[CELL_MS_SPEAKER_FR]	= Asura_Maths::Max(((Asura_Vector_2(fOneOverRootTwo	, fOneOverRootTwo	) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);
			afVolumes[CELL_MS_SPEAKER_FC]	= Asura_Maths::Max(((Asura_Vector_2(0.0f			, 1.0f				) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);
			afVolumes[CELL_MS_SPEAKER_RL]	= Asura_Maths::Max(((Asura_Vector_2(-fOneOverRootTwo, -fOneOverRootTwo	) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);
			afVolumes[CELL_MS_SPEAKER_RR]	= Asura_Maths::Max(((Asura_Vector_2(fOneOverRootTwo	, -fOneOverRootTwo	) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);
			afVolumes[CELL_MS_SPEAKER_EL]	= Asura_Maths::Max(((Asura_Vector_2(-1.0f			, 0.0f				) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);
			afVolumes[CELL_MS_SPEAKER_ER]	= Asura_Maths::Max(((Asura_Vector_2(1.0f			, 0.0f				) * xOriginToSound) + fBleed) * fBleedMul * fVolume, 0.0f);
			afVolumes[CELL_MS_SPEAKER_LFE]	= 0.0f;
		}
		else
		{
			const float fRight = Asura_Maths::Sqrt( m_fPan );
			const float fFront = Asura_Maths::Sqrt( m_fFade );
			const float fLeft  = Asura_Maths::Sqrt( 1.0f - m_fPan );
			const float fRear  = Asura_Maths::Sqrt( 1.0f - m_fFade );

			afVolumes[CELL_MS_SPEAKER_FL] = fVolume * fLeft * fFront;	// Front left
			afVolumes[CELL_MS_SPEAKER_FR] = fVolume * fRight * fFront;	// Front right
			afVolumes[CELL_MS_SPEAKER_FC] = fVolume * fFront;			// Front Centre
			afVolumes[CELL_MS_SPEAKER_LFE] = 0.0f;						// Low Frequency
			afVolumes[CELL_MS_SPEAKER_RL] = fVolume * fLeft * fRear;	// Rear left
			afVolumes[CELL_MS_SPEAKER_RR] = fVolume * fRight * fRear;	// Rear right
			afVolumes[CELL_MS_SPEAKER_EL] = 0.0f;						// Extra left
			afVolumes[CELL_MS_SPEAKER_ER] = 0.0f;						// Extra right
		}
	}
	else
	{
		afVolumes[CELL_MS_SPEAKER_FL] = fVolume;		// Front left
		afVolumes[CELL_MS_SPEAKER_FR] = fVolume;		// Front right
		afVolumes[CELL_MS_SPEAKER_RL] = 0.0f;			// Rear left
		afVolumes[CELL_MS_SPEAKER_RR] = 0.0f;			// Rear right
		afVolumes[CELL_MS_SPEAKER_FC] = 0.0f;			// Front Centre
		afVolumes[CELL_MS_SPEAKER_LFE] = 0.0f;			// Low Frequency
		afVolumes[CELL_MS_SPEAKER_EL] = 0.0f;			// Extra left
		afVolumes[CELL_MS_SPEAKER_ER] = 0.0f;			// Extra right

		Asura_Assert(m_iNumChannels <= 2, "This sound has more channels than the volume code is designed to use");
	}


	// Retrieve the loaded sound 
	Asura_PS3_LoadedSound&	xLoadedSound = Asura_PS3_LoadedSound::s_axLoadedSounds[m_iLoadedSoundIndex];

	// Set pitch by adjusting the loaded sample's rate	
	if ( m_fPitch > 0.0f )
	{
		if(m_bPausedByPitch)
		{
			Unpause();
			m_bPausedByPitch = false;
		}
		cellMSStreamSetPitch( m_iMSStreamHandle, static_cast<int>( m_fPitch * xLoadedSound.GetSampleRate() ) );
	}
	else if ( ( m_fPitch == 0.0f ) && !m_bPausedByPitch )
	{
		Pause();
		m_bPausedByPitch = true;
	}

	// Set the output volumes for each speaker
	if( IsMono() )
	{
		for (int i=0; i<8; i++)
		{
			#ifdef ASURA_ASSERT
			iRet = cellMSCoreSetVolume1(m_iMSStreamHandle, CELL_MS_WET, static_cast<CELL_MS_SPEAKERS>(i), CELL_MS_CHANNEL_0, afVolumes[i]);
			Asura_Assertf(iRet >= 0, ("cellMSCoreSetVolume1 failed, error %d", cellMSSystemGetLastError()) );		
			#else
			cellMSCoreSetVolume1(m_iMSStreamHandle, CELL_MS_WET, static_cast<CELL_MS_SPEAKERS>(i), CELL_MS_CHANNEL_0, afVolumes[i]);
			#endif
		}

		//Modify the route to the master with our dry volume
		#ifdef ASURA_ASSERT
		iRet = cellMSStreamSetRoutingWet( m_iMSStreamHandle, CELL_MS_MASTER_BUS, m_fDryVolumeScale);
		Asura_Assertf(iRet >= 0, ("cellMSStreamSetRoutingWet failed, error %d", cellMSSystemGetLastError()) );
		#else
		cellMSStreamSetRoutingWet( m_iMSStreamHandle, CELL_MS_MASTER_BUS, m_fDryVolumeScale);
		#endif
	}
	else
	{
		//Wet is just one channel
		#ifdef ASURA_ASSERT
		iRet = cellMSCoreSetVolume8(m_iMSStreamHandle, CELL_MS_WET, CELL_MS_CHANNEL_0, afVolumes);
		Asura_Assertf(iRet >= 0, ("cellMSCoreSetVolume8 failed, error %d", cellMSSystemGetLastError()) );		
		#else
		cellMSCoreSetVolume8(m_iMSStreamHandle, CELL_MS_WET, CELL_MS_CHANNEL_0, afVolumes);
		#endif

		//Dry is 8
		Asura_Mem::Set( afVolumes, 0, sizeof(float) * 8 );
		for (int i = 0; i < 8; i++)
		{
			afVolumes[i] = fVolume;
			if(i)
			{
				afVolumes[i - 1] = 0;
			}
			#ifdef ASURA_ASSERT
			iRet = cellMSCoreSetVolume8(m_iMSStreamHandle, CELL_MS_DRY, static_cast<CELL_MS_CHANNELS>(i), afVolumes);
			Asura_Assertf(iRet >= 0, ("cellMSCoreSetVolume8 failed, error %d", cellMSSystemGetLastError()) );		
			#else
			cellMSCoreSetVolume8(m_iMSStreamHandle, CELL_MS_DRY, static_cast<CELL_MS_CHANNELS>(i), afVolumes);
			#endif
		}

		//Modify the route to the master with our dry volume
		#ifdef ASURA_ASSERT
		iRet = cellMSStreamSetRoutingWet( m_iMSStreamHandle, CELL_MS_MASTER_BUS, m_fWetVolumeScale);
		Asura_Assertf(iRet >= 0, ("cellMSStreamSetRoutingWet failed, error %d", cellMSSystemGetLastError()) );
		#else
		cellMSStreamSetRoutingWet( m_iMSStreamHandle, CELL_MS_MASTER_BUS, m_fWetVolumeScale);
		#endif

		#ifdef ASURA_ASSERT
		iRet = cellMSStreamSetRoutingDry( m_iMSStreamHandle, CELL_MS_MASTER_BUS, m_fDryVolumeScale);
		Asura_Assertf(iRet >= 0, ("cellMSStreamSetRoutingDry failed, error %d", cellMSSystemGetLastError()) );
		#else
		cellMSStreamSetRoutingDry( m_iMSStreamHandle, CELL_MS_MASTER_BUS, m_fDryVolumeScale);
		#endif
	}

	if( m_bAuxSendDSPActive )
	{
		//If we have the aux send (i.e. reverb is enabled and setup) , modify it to refelect our wet volume (and stream volume, sends are done before volume changes)
		//the 0.15 is a bit empirical unfortunately. It needs to be scaled down because we are sending to multiple speakers (4 to be exact) but we also want reverb quieter than the original sound
		const float fSendVolume = m_fWetVolumeScale * fVolume * 0.15f;	
		#ifdef ASURA_ASSERT
		int iRet = cellMSFXAuxSendSetVolumes(m_pAuxSendDSPBuffer, 0, &fSendVolume, 1);
		Asura_Assertf(iRet >= 0, ("cellMSFXAuxSendSetVolumes failed, error %d", cellMSSystemGetLastError()) );		
		#else
		cellMSFXAuxSendSetVolumes(m_pAuxSendDSPBuffer, 0, &fSendVolume, 1);
		#endif
	}
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MSStreamingSound::StreamLocalCallback
// Description : Local stream callback
//
// Parameters  :
//   int iStreamHandle   - Handle to stream making callback
//   void* pUserData     - Pointer to custom user data
//   int iCallbackType   - Type of callback request
//   void* pWriteBuffer  - Next write buffer to use
//   int nBufferSize     - Size of write buffer
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_ActiveMSSound::StreamLocalCallback(int iStreamHandle, void* pUserData, int iCallbackType, void* pWriteBuffer, int nBufferSize)
{
	switch( iCallbackType )
	{
		case CELL_MS_CALLBACK_MOREDATA:
		{
			Asura_PS3_ActiveMSSound *pxThis = reinterpret_cast<Asura_PS3_ActiveMSSound*>(pUserData);

			pxThis->m_xLoadDataRequest.m_iStreamHandle = iStreamHandle;
			pxThis->m_xLoadDataRequest.m_pWriteBuffer = pWriteBuffer;
			Asura_PS3_MultiStream_Streaming::AddToLoadQueue( LoadDataRequest, pxThis, sizeof(Asura_PS3_MSStreamingSound) );
		}
		break;

		case CELL_MS_CALLBACK_CLOSESTREAM:   //	CELL_MS_CALLBACK_FINISHSTREAM
		case CELL_MS_CALLBACK_FINISHSTREAM:
		{
			Asura_OutputToDebugger(("Playback of Streamed Sound completed (handle=%d)\n", iStreamHandle));

			Asura_PS3_ActiveMSSound *pThis = reinterpret_cast<Asura_PS3_ActiveMSSound*>(pUserData);
			if ( pThis != NULL )
			{
				pThis->SetStreamFinished(true);
			}
		}
		break;

	default: break;
	}
}

void Asura_PS3_ActiveMSSound::LoadDataRequest( void* pThis, u_int uSize )
{
	Asura_PS3_ActiveMSSound *pxThis = reinterpret_cast<Asura_PS3_ActiveMSSound*>(pThis);

	if ( pxThis != NULL )
	{
		int iStreamHandle = pxThis->m_xLoadDataRequest.m_iStreamHandle;
		Asura_Assert(uSize = sizeof(Asura_PS3_ActiveMSSound), "unexpected size");
		void* pWriteBuffer = pxThis->m_xLoadDataRequest.m_pWriteBuffer;

		Asura_PS3_LoadedSound&	xLoadedSound = Asura_PS3_LoadedSound::s_axLoadedSounds[pxThis->m_iLoadedSoundIndex];

		if ( pxThis->m_uStreamingBytesRead < xLoadedSound.GetSampleDataSize() )
		{
			u_int uReadSize = Asura_Maths::Min( ASURA_PS3_STREAMING_BUFFER_SIZE, xLoadedSound.GetSampleDataSize() - pxThis->m_uStreamingBytesRead );
			u_int uBytesRead = Asura_PS3_FileSystem::ReadData( pxThis->m_xStreamingFile, uReadSize, pWriteBuffer );

			cellMSStreamSetSecondRead(iStreamHandle, (uBytesRead > 0) ? pWriteBuffer : NULL, uBytesRead);
			pxThis->m_uStreamingBytesRead += uBytesRead;
		}
		else
		{
			// Nothing more to play so abort
			cellMSStreamSetSecondRead(iStreamHandle, NULL, 0);
		}

		pxThis->m_bPending = false;      // Definitely not pending any more
	}
}

#endif // #if defined(ASURA_PS3_USE_MULTISTREAM) && (ASURA_PS3_USE_MULTISTREAM != 0)

#endif // #ifndef ASURA_USE_AUDIO_SYSTEM

//eof
