///////////////////////////////////////////////////////////////////////////////
//
// File      : Asura_PS3_MultipleStorage.cpp
// Author    : Philip Hall
// Created   : 11/12/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "Utility/Asura_PS3_MultipleStorage.h"

// Asura headers
#include "Core/Asura_Maths.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Asura PS3 headers
#include "FileAccess/Asura_PS3_FileSystem.h" 

////PGH/////////////////////////////////////////
// Useful constants
////////////////////////////////////////////////
const int    STORAGE_PPU_PRIORITY   = 750;
const size_t STORAGE_PPU_STACK_SIZE = (1024 * 16);
const size_t STORAGE_MEMORY_CONTAINER_SIZE = (1024 * 1024 * 5);

extern "C++" bool Asura_PS3_AllocateFromFile(const char* szFilename, void** pResult, int* piSize);

Asura_PS3_MultipleStorage::Asura_PS3_MultipleStorage(const char* szProductCode, Asura_PS3_StorageCallback pfnCompletion, void* pUserData):
                                                     Asura_PS3_SingleStorage(szProductCode, pfnCompletion, pUserData)
{
	m_bDeleteAll = 0;
	m_szMultiDirectory[0] = 0;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::Load
// Description : Proxy to the single-storage Load function just validates 
//               the slot index is valid in autosave mode.
//
// Parameters  : 
//   const char* szPrefix  - Directory prefix for load
//   bool bAutoLoad        - Set true to auto-load the data
//   int iIndex            - Save slot index if required, negative to ignore
//
// Returns     : bool - Returns true if initiated successfully
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultipleStorage::Load(const char* szPrefix, bool bAutoLoad, int iIndex)
{
	if ( bAutoLoad )
	{
		if ( (iIndex < 0) || (iIndex > ASURA_PS3_MAX_MULTIPLE_SAVES) )
		{
			Asura_Assert(false, "[Asura_PS3_MultipleStorage::Load] - Autoload requires a valid slot index\n");
			return false;
		}
	}
	else if ( iIndex >= 0 )
	{
		Asura_Assert(false, "[Asura_PS3_MultipleStorage::Load] - Can't accept an index if not auto loading\n");
		iIndex = -1;
	}
	
	return Asura_PS3_SingleStorage::Load(szPrefix, bAutoLoad, iIndex);
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::Save
// Description : Proxy to the single-storage Save function just validates 
//               the slot index is valid in autosave mode.
//
// Parameters  : 
//   const char* szPrefix  - Directory prefix for save
//   bool bAutoSave        - Set true to auto-save the data (doesn't prompt user)
//   int iIndex            - Save slot index if required, negative to ignore
//
// Returns     : bool - Returns true if initiated successfully
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultipleStorage::Save(const char* szPrefix, bool bAutoSave, int iIndex)
{
	if ( bAutoSave )
	{
		if ( (iIndex < 0) || (iIndex > ASURA_PS3_MAX_MULTIPLE_SAVES) )
		{
			Asura_Assert(false, "[Asura_PS3_MultipleStorage::Save] - Autosave requires a valid slot index\n");
			return false;
		}
	}
	else if ( iIndex >= 0 )
	{
		Asura_Assert(false, "[Asura_PS3_MultipleStorage::Save] - Can't accept an index if not auto saving\n");
		iIndex = -1;
	}
	
	return Asura_PS3_SingleStorage::Save(szPrefix, bAutoSave, iIndex);
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::Delete
// Description : Show the list of items for deletion 
//
// Parameters  : 
//	bool bAll - Set true to allow the deletion of all saves
//
// Returns     : bool - Returns true if initiated successfully
///////////////////////////////////////////////////////////////////////////////
bool Asura_PS3_MultipleStorage::Delete(bool bAll)
{
	m_bDeleteAll = bAll;
	
	// Launch load thread
	int iResult = sys_ppu_thread_create(&m_xThreadID, &ThreadDelete, 
	                                     reinterpret_cast<uint64_t>(this), STORAGE_PPU_PRIORITY, 
	                                     STORAGE_PPU_STACK_SIZE, 0, "Storage_Delete");

	Asura_Assert(iResult >= 0, "[Asura_PS3_MultipleStorage::Delete] - Unable to create storage deletion thread\n");
	return (iResult == CELL_OK);
}



////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::ThreadStorage
// Description : Custom storage thread for list mode
//
// Parameters  : none
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultipleStorage::ThreadStorage()
{
	CellSaveDataSetList xSetList;
	xSetList.sortType      = CELL_SAVEDATA_SORTTYPE_SUBTITLE;
	xSetList.sortOrder     = CELL_SAVEDATA_SORTORDER_ASCENT;
	xSetList.dirNamePrefix = m_szDirectory;
	xSetList.reserved      = NULL;

	CellSaveDataSetBuf  xSetBuf;
	Asura_Mem::Set(&xSetBuf, 0, sizeof(xSetBuf));
	xSetBuf.dirListMax     = ASURA_PS3_MAX_MULTIPLE_SAVES;
	xSetBuf.fileListMax    = ASURA_PS3_MAX_DATA_ENTRIES;
	xSetBuf.bufSize        = Asura_Maths::Max( ASURA_PS3_MAX_MULTIPLE_SAVES * sizeof(CellSaveDataDirList), ASURA_PS3_MAX_DATA_ENTRIES * sizeof(CellSaveDataFileStat) );
	xSetBuf.buf            = Asura_Allocate( xSetBuf.bufSize );
	
	if( xSetBuf.buf == NULL )
	{
		Asura_Assert( false, "[Asura_PS3_MultipleStorage::ThreadStorage] - Unable to allocate buffer memory\n" );
		return;
	}

	// Allocate a system memory container for use by the save utility
	sys_memory_container_t xMemoryContainer;
	if ( sys_memory_container_create( &xMemoryContainer, STORAGE_MEMORY_CONTAINER_SIZE) != 0 )
	{
		Asura_Assert( false, "[Asura_PS3_MultipleStorage::ThreadStorage] - Unable to allocate system memory container\n" );
		Asura_Deallocate(xSetBuf.buf);
		return;
	}

	int iResult = 0;

	// LOADING
	if ( m_bLoading )
	{
		if ( m_bAuto )
		{
			Asura_OutputToDebugger(("[Asura_PS3_MultipleStorage::ThreadStorage] - Autoloading (directory filter '%s')\n", m_szDirectory));
			iResult = cellSaveDataListAutoLoad(CELL_SAVEDATA_VERSION_CURRENT, CELL_SAVEDATA_ERRDIALOG_NOREPEAT, 
			                                   &xSetList, &xSetBuf, 
			                                   &LoadFixedCallback, &LoadStatusCallback, &LoadFileCallback,
			                                   xMemoryContainer, this);
			if ( iResult < CELL_SAVEDATA_RET_OK )
			{
				Asura_Assertf(false, ("[Asura_PS3_MultipleStorage::ThreadStorage] - cellSaveDataListAutoLoad 0x%08X\n", iResult));
			}
		}
		else
		{
			Asura_OutputToDebugger(("[Asura_PS3_MultipleStorage::ThreadStorage] - Loading (directory filter '%s')\n", m_szDirectory));
			iResult = cellSaveDataListLoad2(CELL_SAVEDATA_VERSION_CURRENT, &xSetList, &xSetBuf,
			                                 &LoadListCallback, &LoadStatusCallback, &LoadFileCallback,
			                                 xMemoryContainer, this);
			if ( iResult < CELL_SAVEDATA_RET_OK )
			{
				Asura_Assertf(false, ("[Asura_PS3_MultipleStorage::ThreadStorage] - cellSaveDataListLoad2 0x%08X\n", iResult));
			}
		}
	}
	// SAVING
	else
	{
		if ( m_bAuto )
		{
			Asura_OutputToDebugger(("[Asura_PS3_MultipleStorage::ThreadStorage] - Autosaving (directory filter '%s')\n", m_szDirectory));
			iResult = cellSaveDataListAutoSave(CELL_SAVEDATA_VERSION_CURRENT, CELL_SAVEDATA_ERRDIALOG_NOREPEAT, 
			                                   &xSetList, &xSetBuf, 
			                                   &SaveFixedCallback, &SaveStatusCallback, &SaveFileCallback,
			                                   xMemoryContainer, this);
			if ( iResult < CELL_SAVEDATA_RET_OK )
			{
				Asura_Assertf(false, ("[Asura_PS3_MultipleStorage::ThreadStorage] - cellSaveDataListAutoSave 0x%08X\n", iResult));
			}
		}
		else
		{
			Asura_OutputToDebugger(("[Asura_PS3_MultipleStorage::ThreadStorage] - Saving (directory filter '%s')\n", m_szDirectory));
			iResult = cellSaveDataListSave2(CELL_SAVEDATA_VERSION_CURRENT, &xSetList, &xSetBuf,
			                                &SaveListCallback, &SaveStatusCallback, &SaveFileCallback,
			                                xMemoryContainer, this);
			if ( iResult < CELL_SAVEDATA_RET_OK )
			{
				Asura_Assertf(false, ("[Asura_PS3_MultipleStorage::ThreadStorage] - cellSaveDataListSave2 0x%08X\n", iResult));
			}
		}
	}

	iResult = sys_memory_container_destroy( xMemoryContainer );
	if ( iResult < 0 )
	{
		Asura_Assert(false, "[Asura_PS3_MultipleStorage::ThreadStorage] - Potential leak in memory container\n");
	}
	
	// Memory cleanup
	Asura_Deallocate(xSetBuf.buf);
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::ThreadDelete
// Description : Deletion thread
//
// Parameters  : 
//   uint64_t uParam - Pointer to parent object
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultipleStorage::ThreadDelete(uint64_t uParam)
{
	Asura_PS3_MultipleStorage* pThis = reinterpret_cast<Asura_PS3_MultipleStorage*>(uParam);

	// Allocate a system memory container for use by the save utility
	sys_memory_container_t xMemoryContainer;
	if ( sys_memory_container_create( &xMemoryContainer, STORAGE_MEMORY_CONTAINER_SIZE) != 0 )
	{
		Asura_Assert( false, "[Asura_PS3_MultipleStorage::ThreadDelete] - Unable to allocate system memory container\n" );
		return;
	}
	
	// call the deletion function
	int iResult = 0;
	if ( pThis->m_bDeleteAll )
	{
		iResult = cellSaveDataDelete2( xMemoryContainer );
	}
	else
	{
		Asura_Assert( false, "[Asura_PS3_MultipleStorage::ThreadDelete] - Filtered delete unimplemented\n" );
	}
	

	iResult = sys_memory_container_destroy( xMemoryContainer );
	if ( iResult < 0 )
	{
		Asura_Assert(false, "[Asura_PS3_MultipleStorage::ThreadDelete] - Potential leak in memory container\n");
	}


	bool bDeleteThis = false;

	// Completion callback
	if ( pThis->m_pfnCompletionCallback != NULL )
	{
		pThis->m_pfnCompletionCallback(pThis, bDeleteThis);
	}

	// Signal the sync semaphore (in case anybody was waiting for it)
	pThis->m_xCompleted.Signal();

	if ( bDeleteThis )
	{
		delete pThis;
	}

	sys_ppu_thread_exit(0);
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::SaveListCallback
// Description : List oriented saving callback handler
//
// Parameters  : 
//   CellSaveDataCBResult* pResult  - Result of processing this step
//   CellSaveDataListGet* pGet      - Input from the save system
//   CellSaveDataListSet* pSet      - Output from this step
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultipleStorage::SaveListCallback(CellSaveDataCBResult* pResult, CellSaveDataListGet* pGet, CellSaveDataListSet* pSet )
{
	Asura_PS3_MultipleStorage* pThis = reinterpret_cast<Asura_PS3_MultipleStorage*>( pResult->userdata );

	Asura_Mem::Set(&pThis->m_xIconData, 0, sizeof(pThis->m_xIconData));
	Asura_Mem::Set(&pThis->m_xNewData, 0, sizeof(pThis->m_xNewData));

	if ( pGet->dirNum > pGet->dirListNum )
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultipleStorage::SaveListCallback] - Too many data directories: expected %d, found %d\n", pGet->dirListNum, pGet->dirNum));
	}

	pSet->fixedList     = pGet->dirList;
	pSet->fixedListNum  = (pGet->dirListNum < CELL_SAVEDATA_LISTITEM_MAX) ? pGet->dirListNum : CELL_SAVEDATA_LISTITEM_MAX;
	pSet->focusPosition = CELL_SAVEDATA_FOCUSPOS_LISTHEAD;
	pSet->focusDirName  = NULL;
	pSet->reserved      = NULL;
	pSet->newData       = NULL;

	// If there's enough room create a new save
	if ( pGet->dirListNum < ASURA_PS3_MAX_MULTIPLE_SAVES )
	{
		Asura_Assert(pThis->m_bAuto == false, "[Asura_PS3_MultipleStorage::SaveListCallback] - Should't be using auto in here\n");
	
		// Create an compare all possible directory names
		u_int uInnerLoop = 0;
		for( int iLoop = 0; iLoop < ASURA_PS3_MAX_MULTIPLE_SAVES; iLoop++ )
		{
			Asura_PrintfToBfr(pThis->m_szMultiDirectory, "-%c", CELL_SAVEDATA_DIRNAME_SIZE, true, pThis->m_szDirectory, 'A' + iLoop );

			// Determine if this directory is already in use
			for ( uInnerLoop = 0; uInnerLoop < pGet->dirListNum; uInnerLoop++ )
			{
				if ( strncmp(pThis->m_szMultiDirectory, pGet->dirList[uInnerLoop].dirName, CELL_SAVEDATA_DIRNAME_SIZE ) == 0 ) { break; }
			}

			// if not found, break the outer loop  (this name is usable)
			if( uInnerLoop == pGet->dirListNum ) { break; }
		}

		pSet->newData = &pThis->m_xNewData;
		pThis->m_xNewData.dirName = pThis->m_szMultiDirectory;
		pThis->m_xNewData.iconPosition = CELL_SAVEDATA_ICONPOS_HEAD;
		pThis->m_xNewData.reserved = NULL;
		pThis->m_xNewData.icon = NULL;

		int iSize;
		if ( Asura_PS3_AllocateFromFile("/MISC/ICON0.PNG", &(pThis->m_pDataBuffer), &iSize) )
		{
			pThis->m_xIconData.title = "Create New Save";
			pThis->m_xIconData.iconBuf = pThis->m_pDataBuffer;
			pThis->m_xIconData.iconBufSize = static_cast<u_int>(iSize);
			pThis->m_xIconData.reserved = NULL;
			pThis->m_xNewData.icon = &pThis->m_xIconData;
		}
		else
		{
			pResult->result = CELL_SAVEDATA_CBRESULT_ERR_FAILURE;
			return;
		}
	}

	pResult->result = CELL_SAVEDATA_CBRESULT_OK_NEXT;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Asura_PS3_MultipleStorage::LoadListCallback
// Description : List oriented loading callback handler
//
// Parameters  : 
//   CellSaveDataCBResult* pResult  - Result of processing this step
//   CellSaveDataListGet* pGet      - Input from the load system
//   CellSaveDataListSet* pSet      - Output from this step
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
void Asura_PS3_MultipleStorage::LoadListCallback(CellSaveDataCBResult* pResult, CellSaveDataListGet* pGet, CellSaveDataListSet* pSet )
{
	Asura_PS3_MultipleStorage* pThis = reinterpret_cast<Asura_PS3_MultipleStorage*>( pResult->userdata );

	Asura_Mem::Set(&pThis->m_xIconData, 0, sizeof(pThis->m_xIconData));
	Asura_Mem::Set(&pThis->m_xNewData, 0, sizeof(pThis->m_xNewData));

	if ( pGet->dirNum > pGet->dirListNum )
	{
		Asura_OutputToDebugger(("[Asura_PS3_MultipleStorage::LoadListCallback] - Too many data directories: expected %d, found %d\n", pGet->dirListNum, pGet->dirNum));
	}

	pSet->fixedList     = pGet->dirList;
	pSet->fixedListNum  = (pGet->dirListNum < CELL_SAVEDATA_LISTITEM_MAX) ? pGet->dirListNum : CELL_SAVEDATA_LISTITEM_MAX;
	pSet->focusPosition = CELL_SAVEDATA_FOCUSPOS_LISTHEAD;
	pSet->newData = NULL;
	pResult->result = CELL_SAVEDATA_CBRESULT_OK_NEXT;
	return;
}





