///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_PFX_System.cpp
// Author : Andy Tate
// Date   : 29-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Engine independent.
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Viewport.h"
#include "SpecialFX/PFX/Asura_PFX_Effect.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

// Engine platform-specific.
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_Fog.h"
#include "Rendering/Asura_PS3_GCM_VertexBuffer.h"
#include "Rendering/Asura_PS3_MaterialResponse_System.h"
#include "Rendering/Asura_PS3_RenderingOptions.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"
#include "SpecialFX/Asura_PS3_PFXRender.h"

#ifdef ASURA_USE_GCM_REPLAY
#include <cell/gcm.h>
#endif

//Shaders definition
#include "Rendering/ShadersDefinition/VD/SFX_PFX_VD.h"
#include "Rendering/ShadersDefinition/VP/SFX_PFX_VP.h"
#include "Rendering/ShadersDefinition/FP/SFX_PFX_FP.h"

// D E B U G V A R S /////////////////////////////////////////////////////////
// dbg_bAllowPFXRenderingOnSPU controls whether vertex buffer is sent to SPU program
/*DEBUGVAR */bool dbg_bAllowPFXRenderingOnSPU = true;
// dbg_bPrintPFXError controls printing of error messages
#if defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
DEBUGVAR bool dbg_bPrintPFXError = true;
// dbg_bTestPFXSPURender enables code to do the same computations again on PPU 
// and see if they match. see also dbg_bMultithreadPFX which controls whether 
// SPU is used for updating positions etc.
DEBUGVAR bool dbg_bTestPFXSPURender = false;
#endif // #if defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

//#define MAX_PARTICLES_PER_FRAME       4096
#define MAX_PARTICLES_PER_FRAME     4096*4
#define MAX_VERTICES_PER_FRAME      (MAX_PARTICLES_PER_FRAME * NUM_VERTICES_PER_PARTICLE)

#define MAX_PARTICLE_RENDER_DATA    2048
static const u_int uNUM_VERTEX_BUFFERS = 4;

typedef void (*ParticleColourFunc)(const Asura_PFX_Particle& xParticle,
                                   const Asura_PFX_Particle_Template& xTemplate,
                                   const Asura_RGBA& xLightColour,
                                   Asura_RGBA& xColour);
typedef float (*ParticleAngleFunc)(const Asura_PFX_Particle& xParticle, 
                                   const Asura_PFX_Particle_Template& xTemplate);

////////////////////////////////////////////////////////////////////////////////
// Structure declarations...
////////////////////////////////////////////////////////////////////////////////

struct Asura_PS3_ParticleRenderData
{
    int                     m_iTexID;
    ASURA_TRANSLUCENCY_MODE m_eTranslucency;
    u_int                   m_uCameraIndex;
    u_int                   m_uStartVertex;
    u_int                   m_uNumVertices;
	u_int					m_usPFXFlags;
	float					m_fDepth;
	union
	{
		struct
		{
			unsigned int	m_bUseFogging					:1;
			unsigned int	m_bCrossfadeTextureFrames		:1;
		};
		unsigned int		m_uFlags;
	};
};


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// The axis data for the camera...
static Asura_Vector_3				g_xCameraDeltaX;
static Asura_Vector_3				g_xCameraDeltaY;


// Vertex-buffer data...
static Asura_PS3_GCM_VertexBuffer	g_xVertexBuffer[uNUM_VERTEX_BUFFERS];
static u_int						g_uCurVertexBuffer = 0;
static void*						g_apVertexBufferPtr[uNUM_VERTEX_BUFFERS];

// Vertex/fragment shader data...

static Asura_PS3_GCM_VertexShader*		g_pxVertexShader;

// Particle tracking data...
static Asura_PS3_ParticleRenderData g_axParticleRenderData [MAX_PARTICLE_RENDER_DATA];

static u_int                        g_uNumParticleRenderData    = 0;
static u_int                        g_uCurrentVertex            = 0;


// Vertex tracking data...

static Asura_PS3_ParticleRender_Vertex* g_pxCurrentVertices         = NULL;
static Asura_PS3_ParticleRender_Vertex* g_pxCurrentVerticesStart    = NULL;

DEBUGVAR bool dbg_bDontRenderParticles = false;

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

static void UpdateVertexBuffersForNewFrame (void)
{
    g_uCurVertexBuffer = (g_uCurVertexBuffer + 1) % uNUM_VERTEX_BUFFERS;

    // Reset the particle-data...
    g_uCurrentVertex = 0;

    g_pxCurrentVertices = ( Asura_PS3_ParticleRender_Vertex* )( g_apVertexBufferPtr[ g_uCurVertexBuffer ] );
    g_pxCurrentVerticesStart = g_pxCurrentVertices;
}

static void* GetVertexBufferArea(u_int uNumParticles)
{
    // alignment must be in full vertices not just 128 bytes. If the vertex struct 
    // is 24 bytes, we can align to 384 bytes to get a vertex boundary with 128 byte-alignment.
    // This can be achieved by always aligning number of vertices to 16, or number of particles to 4.

    // KJFL - 03 February 2006 - currently changed to 36 bytes. So... lets say aligning to 32 vertices
	// CA - 20 June 2007 - currently changed to 40 bytes. We now align to 16 vertices
    Asura_Assert(sizeof(Asura_PS3_ParticleRender_Vertex) == 40, "size of struct changed");

    u_int uNumVertices = uNumParticles * NUM_VERTICES_PER_PARTICLE;
    uNumVertices = (uNumVertices + 31) & ~31;
    // TODO: could optimise buffer usage by taking account of which cameras can see the effect.
    uNumVertices *= Asura_Camera::GetNumberOfCameras();

    void* pRet;
    if ( g_uCurrentVertex + uNumVertices <= MAX_VERTICES_PER_FRAME )
    {
        pRet = ( char* )( g_pxCurrentVertices );
		if( (reinterpret_cast<uint64_t>(pRet) & 127) )
		{
			Asura_Assert(!(reinterpret_cast<uint64_t>(pRet) & 127), "vertex buffer area is not aligned to 128 bytes");
			pRet = NULL;
		}
		else
		{
	        g_pxCurrentVertices += uNumVertices;
			g_uCurrentVertex += uNumVertices;
	    }
	}
    else
    {
        Asura_Assert(false, "too many particles this frame");
        pRet = NULL;
    }
    return pRet;
}

#if defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
static bool PFXRenderVertsAreEqual(const Asura_PS3_ParticleRender_Vertex* pxVerts1, 
                                   const Asura_PS3_ParticleRender_Vertex* pxVerts2, 
                                   u_int uNumVerts)
{
    // for each pair of verts compare the first 4 u_chars and the next 5 floats within a slight error.
    for (u_int u = 0; u < uNumVerts; ++u)
    {
        const u_char* puc1 = reinterpret_cast<const u_char*>(pxVerts1++);
        const u_char* puc2 = reinterpret_cast<const u_char*>(pxVerts2++);
        for (u_int uByte = 0; uByte < 4; ++uByte)
        {
            int i1 = static_cast<int>(*puc1++);
            int i2 = static_cast<int>(*puc2++);
            if (Asura_Maths::Abs(i1-i2) > 2)
            {
                return false;
            }
        }

        const float* pfFloat1 = reinterpret_cast<const float*>(puc1);
        const float* pfFloat2 = reinterpret_cast<const float*>(puc2);
        for (u_int uFloat = 0; uFloat < 5; ++uFloat)
        {
            float f1 = *pfFloat1++;
            float f2 = *pfFloat2++;
            if (fabsf(f1 - f2) > 0.0001f)
            {
                return false;
            }
        }
    }
    return true;
}
#endif // #if defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)

////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////
void Asura_PFX_System::Platform_Initialise (void)
{
	g_pxVertexShader				= new Asura_PS3_VP_PFX();

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 0, (unsigned char *)&_binary_PS3_SFX_PFX_FP0000_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0000_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0000_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1, (unsigned char *)&_binary_PS3_SFX_PFX_FP0001_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0001_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0001_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 10, (unsigned char *)&_binary_PS3_SFX_PFX_FP0010_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0010_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0010_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 11, (unsigned char *)&_binary_PS3_SFX_PFX_FP0011_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0011_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0011_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 100, (unsigned char *)&_binary_PS3_SFX_PFX_FP0100_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0100_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0100_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 101, (unsigned char *)&_binary_PS3_SFX_PFX_FP0101_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0101_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0101_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 110, (unsigned char *)&_binary_PS3_SFX_PFX_FP0110_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0110_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0110_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 111, (unsigned char *)&_binary_PS3_SFX_PFX_FP0111_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP0111_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP0111_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1000, (unsigned char *)&_binary_PS3_SFX_PFX_FP1000_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1000_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1000_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1001, (unsigned char *)&_binary_PS3_SFX_PFX_FP1001_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1001_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1001_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1010, (unsigned char *)&_binary_PS3_SFX_PFX_FP1010_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1010_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1010_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1011, (unsigned char *)&_binary_PS3_SFX_PFX_FP1011_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1011_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1011_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1100, (unsigned char *)&_binary_PS3_SFX_PFX_FP1100_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1100_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1100_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1101, (unsigned char *)&_binary_PS3_SFX_PFX_FP1101_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1101_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1101_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1110, (unsigned char *)&_binary_PS3_SFX_PFX_FP1110_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1110_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1110_bin_start), new Asura_PS3_PFX_Handles(), 1, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( static_cast<Asura_Hash_ID> ( ASURA_SHADER_HASH_PFX ), 1111, (unsigned char *)&_binary_PS3_SFX_PFX_FP1111_bin_start, ( (unsigned char *)&_binary_PS3_SFX_PFX_FP1111_bin_end - (unsigned char *)&_binary_PS3_SFX_PFX_FP1111_bin_start), new Asura_PS3_PFX_Handles(), 1, true );

	g_pxVertexShader->InitVertexBufferDeclaration	( g_axVertexDeclaration, VD_ELEMENTS );

    // Create the vertex-buffer...
    for (u_int u = 0; u < uNUM_VERTEX_BUFFERS; ++u)
    {
        g_xVertexBuffer[u].Initialise( PFX_VERTEX_SIZE, MAX_VERTICES_PER_FRAME, CELL_GCM_LOCATION_LOCAL, true );
		g_apVertexBufferPtr[u] = g_xVertexBuffer[u].LockForWriting();
    }

    UpdateVertexBuffersForNewFrame();
}

void Asura_PFX_System::Platform_BeginRender (void)
{
    const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation ();

    // Store the X and Y axes from the camera...
    g_xCameraDeltaX.x   = pxCameraOrientation->e11;
    g_xCameraDeltaX.y   = pxCameraOrientation->e21;
    g_xCameraDeltaX.z   = pxCameraOrientation->e31;

    g_xCameraDeltaY.x   = pxCameraOrientation->e12;
    g_xCameraDeltaY.y   = pxCameraOrientation->e22;
    g_xCameraDeltaY.z   = pxCameraOrientation->e32;

    // Reset the particle-data...
    g_uNumParticleRenderData    = 0;

    ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__PFX, "Asura PFX" );


	if( s_bIsRenderingHeatHazeSources )
	{
		Asura_PS3_DeferredRendering::ResolveBackBuffer();
	}
}

void Asura_PFX_System::Platform_EndRender (void)
{
	if( dbg_bDontRenderParticles || !g_uNumParticleRenderData )
	{
		ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__PFX );
		return;
	}

    // Render everything...
	Asura_PS3_FragmentShaderCopy* pxFP	= NULL;
	bool bRestoreRenderStates			= false;

	Asura_PS3_RenderingStates::RequestZBufferWriteEnabled	( false );
	Asura_PS3_RenderingStates::RequestZBufferCompareMode	( ASURA_ZBUFFERCMP_LESSEQUAL );
	Asura_PS3_RenderingStates::RequestCullMode				( ASURA_CULLMODE_NONE );
	Asura_PS3_RenderingStates::SetStencilTestEnabled		( false );
	Asura_PS3_RenderingStates::SetStencilFunction( ASURA_PS3_STENCIL_NEVER, 0, 0 );
	Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP );
	cellGcmSetZcullStatsEnableUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_FALSE );

	if(Asura_PFX_System::s_eCurrentRenderType == ASURA_PFX_RENDER_SKYBOX)
	{
		Asura_PS3_Graphics::BeginRenderToSkybox();
	}

	g_pxVertexShader->Bind();

	Asura_PS3_DeferredRendering::BindAsSingleRT();
	Asura_PS3_Graphics::SetViewportToDefaults();
	Asura_PS3_RenderingStates::RequestAlphaTest( false );
	Asura_PS3_RenderingStates::SetAlphaBlending( true );

	const Asura_AMatrix4x4& xViewProjectionMatrix = Asura_PS3_Graphics::GetViewProjectionMatrix();
	g_pxVertexShader->SetUniformParameter( VIEW_PROJECTION_MATRIX, xViewProjectionMatrix.GetElements() );

	for (u_int u = 0; u < g_uNumParticleRenderData; u++)
    {
        const Asura_PS3_ParticleRenderData& xRenderData = g_axParticleRenderData[u];

        // Set the appropriate render-states...
        Asura_Camera::SetCameraIndex( xRenderData.m_uCameraIndex );

		float faPosOffsetAndSizeScaleVolDepthScale[4] = { 0.0f, 0.0f, 0.0f, xRenderData.m_fDepth };
		if ( xRenderData.m_usPFXFlags & ASURA_PFX_EFFECT_SKYBOX )
		{
			const	Asura_Vector_3 *pxCameraPos	= Asura_Camera::GetPosition();
			faPosOffsetAndSizeScaleVolDepthScale[0] = pxCameraPos->x;
			faPosOffsetAndSizeScaleVolDepthScale[1] = pxCameraPos->y;
			faPosOffsetAndSizeScaleVolDepthScale[2] = pxCameraPos->z;
			float fDistance = Asura_Camera::GetFarClip() + 500.0f;
			Asura_PS3_RenderingStates::RequestZBufferEnabled( false );
//			Asura_PS3_RenderingStates::SetDepthBias( 1.0f, fDistance );
//			Asura_PS3_RenderingStates::SetUseDepthBias( true );
		}
		else
		{
			Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
//			Asura_PS3_Graphics::SetViewportToDefaults();
//			Asura_PS3_RenderingStates::SetUseDepthBias( false );
		}
		g_pxVertexShader->SetUniformParameter( VP_PosOffsetAndSizeScaleVolDepthScale, faPosOffsetAndSizeScaleVolDepthScale );

		if( xRenderData.m_usPFXFlags & ASURA_PFX_EFFECT_WATER_DISTORTER )
		{
			Asura_PS3_RenderingStates::RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
			Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALPHA );
			bRestoreRenderStates = true;
		}
		else
		{
			if( xRenderData.m_eTranslucency == ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE )
			{
				Asura_PS3_RenderingStates::RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
			}
			else
			{
				Asura_PS3_RenderingStates::RequestTranslucencyMode( xRenderData.m_eTranslucency );
			}
		}
		Asura_PS3_FogParameters &xFogParams = Asura_PS3_Fog::GetFogParameters();
		u_int uShaderID = 0;
		if( xRenderData.m_eTranslucency == ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE )
		{
			uShaderID += 1000;
		}
		else
		{
			if( xRenderData.m_bUseFogging && xFogParams.m_bFogEnabled )
			{
				uShaderID += 1;
			}
			if( xRenderData.m_fDepth > 0.0f )
			{
				uShaderID += 10;
			}
		}
		if( xRenderData.m_bCrossfadeTextureFrames )
		{
			uShaderID += 100;
		}

		pxFP = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_PFX, uShaderID );

		static float afTemp[ 2 ] = { 1.0f / static_cast<float>(Asura_PS3_Graphics::GetScreenWidth()), 1.0f / static_cast<float>(Asura_PS3_Graphics::GetScreenHeight()) };
		pxFP->SetFloats( FP_OneOverScreenDimensions, 2, afTemp );

		if( xRenderData.m_eTranslucency == ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE )
		{
			pxFP->SetTexture( FP_HeatHazeCaptureTexture, Asura_PS3_DeferredRendering::GetResolvedTexture(), true );
		}
		else
		{
			Asura_PS3_MaterialResponse_System::SetDepthParameters( pxFP, FP_DepthAdjustment );
			
			pxFP->SetTexture( FP_DepthTexture, Asura_PS3_DeferredRendering::GetDeferredFrameBuffer()->GetZBufferAsTexture(), false );
			if( xRenderData.m_bUseFogging && xFogParams.m_bFogEnabled )
			{
				pxFP->SetTexture( FP_FogTexture,	xFogParams.m_pxFogTexture, true );
				float afColour[3];
				switch( xRenderData.m_eTranslucency )
				{
				case ASURA_TRANSLUCENCY_GLOWING:
				case ASURA_TRANSLUCENCY_INVCOLOUR:
					afColour[ 0 ] = 0.0f; afColour[ 1 ] = 0.0f; afColour[ 2 ] = 0.0f;
					break;
				default:
					afColour[ 0 ] = xFogParams.m_afFogColour[ 0 ];
					afColour[ 1 ] = xFogParams.m_afFogColour[ 1 ];
					afColour[ 2 ] = xFogParams.m_afFogColour[ 2 ];
					break;
				}
				const float afFogProperties [ 2 ] = { xFogParams.m_fFogNearPlane, xFogParams.m_fOneOverFogDepth };
				pxFP->SetFloats( FP_FogColour,				3, afColour );
				pxFP->SetFloats( FP_FogProperties,			2, afFogProperties );
			}
		}
		pxFP->SetTexture( FP_DiffuseMap, xRenderData.m_iTexID );
		pxFP->Bind();

		g_xVertexBuffer[g_uCurVertexBuffer].BindFromVertex( g_axVertexDeclaration, VD_ELEMENTS, xRenderData.m_uStartVertex );
		g_xVertexBuffer[g_uCurVertexBuffer].RenderQuads( xRenderData.m_uNumVertices );
		Asura_PS3_Graphics::Flush();
	}


	if(Asura_PFX_System::s_eCurrentRenderType == ASURA_PFX_RENDER_SKYBOX)
	{
		Asura_PS3_Graphics::EndRenderToSkybox();
	}


#ifndef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
    // this function is called by Platform_BeginUpdate() if the define is on.
    UpdateVertexBuffersForNewFrame();
#endif // #ifndef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

	if( bRestoreRenderStates )
	{
		Asura_PS3_RenderingStates::SetColorMaskMrtState( ASURA_PS3_MRT_MRT0_ALL );
	}
	Asura_PS3_RenderingStates::RequestAlphaTest( false );
	Asura_PS3_RenderingStates::SetUseDepthBias( false );
	cellGcmSetZcullStatsEnableUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), CELL_GCM_TRUE );

	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__PFX );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_Reset
// Author       : Andy Tate
// Date         : 30-Jun-2005
//
// Description  : 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PFX_System::Platform_Reset (void)
{
}

/// End of function 'Asura_PFX_System::Platform_Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_Shutdown
// Author       : Andy Tate
// Date         : 30-Jun-2005
//
// Description  : 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PFX_System::Platform_Shutdown (void)
{
    // Destroy the buffers...
    for( u_int u = 0; u < uNUM_VERTEX_BUFFERS; ++u )
    {
        g_xVertexBuffer[u].Destroy();
    }
	delete g_pxVertexShader;
	g_pxVertexShader = NULL;
}

/// End of function 'Asura_PFX_System::Platform_Shutdown'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_SupportsHeatHaze
// Author       : Andy Tate
// Date         : 30-Jun-2005
//
// Description  : Returns whether the PS3 supports heat-haze particles.
//
// Return value : "true" if heat-haze is supported, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PFX_System::Platform_SupportsHeatHaze (void)
{
    return  (true);
}

/// End of function 'Asura_PFX_System::Platform_SupportsHeatHaze'. ///





////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_Effect::Platform_RenderParticles
// Author       : Andy Tate
// Date         : 30-Jun-2005
//
// Description  : Renders the specified list of particles all of which are of 
//                the same type of particle-template.
//
// Parameters   : An array of pointers to PFX particles to render, the number of
//                particles to render, a pointer to the particle-template to
//                render along with the colour of the light.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_PFX_Effect::Platform_RenderParticles (const Asura_PFX_Particle* const pxParticles,
                                                 u_int uNumParticles,
                                                 const Asura_PFX_Particle_Template* pxTemplate,
                                                 const Asura_RGBA& xLightColour
#if defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
                                                , void* pPlatformRenderData
#endif  // defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
                                                ) const
{
    if (!Asura_PS3_RenderingOptions::GetOption(Asura_PS3_RenderingOptions::OPTION_DRAW_PARTICLES))
    {
        return;
    }

    if( !uNumParticles ) { return; }

    Asura_Assert    ((NULL != pxParticles), "Null particle pointer given");
    Asura_Assert    ((NULL != pxTemplate), "Null particle template pointer given");

	// Loop through ALL of the camera...
    for (int i = 0; i < Asura_Camera::GetNumberOfCameras(); ++i)
    {
        // Can this PFX been seen by this camera?
        if( !(m_ucCameraFlags & (1 << i)) ) { continue; }

        Asura_PS3_ParticleRenderData& xRenderData = g_axParticleRenderData[g_uNumParticleRenderData];

        xRenderData.m_iTexID					= pxTemplate->m_iTextureHandle;
        xRenderData.m_eTranslucency				= pxTemplate->m_eTranslucency;
        xRenderData.m_uCameraIndex				= i;
		xRenderData.m_usPFXFlags				= m_usPFXFlags;
		xRenderData.m_fDepth					= pxTemplate->m_fVolumetricDepthScale;
		xRenderData.m_bUseFogging				= pxTemplate->m_bUseFogging;
		xRenderData.m_bCrossfadeTextureFrames	= pxTemplate->m_bCrossfadeTextureFrames;

        u_int uStartVertex;
        bool bOK;
        u_int uNumVertices = uNumParticles * NUM_VERTICES_PER_PARTICLE;

#if defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
		// we will need to at least take this flag into account if there will be any lag introduced by rendering PFX on the SPUs
		//if ( !( m_usPFXFlags & ASURA_PFX_EFFECT_FORCE_RENDER_ON_MAIN_THREAD ) &&
        if (pPlatformRenderData)
        {
            // vertex buffer already created, just set the index correctly
            const Asura_PS3_ParticleRender_Vertex* pxStartVertex = reinterpret_cast
                <const Asura_PS3_ParticleRender_Vertex*>(pPlatformRenderData);
            bOK = (pxStartVertex >= g_pxCurrentVerticesStart) && 
                    (pxStartVertex < g_pxCurrentVerticesStart + MAX_VERTICES_PER_FRAME);
            Asura_Assert(bOK, "start vertex out of range");
            if (!bOK && dbg_bPrintPFXError)
            {
                Asura_OutputToDebugger(("start vertex out of range\n"));
            }
            uStartVertex = pxStartVertex - g_pxCurrentVerticesStart;

            if (dbg_bTestPFXSPURender)
            {
                static u_int ls_uNumEqual = 0;
                static u_int ls_uNumNotEqual = 0;
                Asura_PS3_ParticleRender_Vertex* pxVertices = new Asura_PS3_ParticleRender_Vertex[uNumParticles*4];
                bool bPaused = Asura_Timers::GetGameTimeScale() == 0.0f;
                Asura_PS3_RenderPFXForCurrentCamera(pxVertices, GetSizeScale(), 
                    GetOrientation(), pxParticles, uNumParticles, *pxTemplate, xLightColour,
                    g_xCameraDeltaX, g_xCameraDeltaY, *Asura_Camera::GetPosition(), bPaused);
                if (PFXRenderVertsAreEqual(pxVertices, pxStartVertex, uNumParticles*4))
                {
                    ++ls_uNumEqual;
                }
                else
                {
                    ++ls_uNumNotEqual;
                }
                delete[] pxVertices;
            }
        }
        else
#endif  // defined(ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE)
        {
            Asura_PS3_ParticleRender_Vertex* pxVertices = reinterpret_cast
                <Asura_PS3_ParticleRender_Vertex*>(GetVertexBufferArea(uNumParticles));
            uStartVertex = pxVertices - g_pxCurrentVerticesStart;
            bOK = pxVertices != NULL;
            Asura_Assert(bOK, "too many particles this frame");
            if (bOK)
            {
                // write the particles to the vertex buffer
                bool bPaused = Asura_Timers::GetGameTimeScale() == 0.0f;
                Asura_PS3_RenderPFXForCurrentCamera(pxVertices, GetSizeScale(), 
                    GetOrientation(), pxParticles, uNumParticles, *pxTemplate, xLightColour,
                    g_xCameraDeltaX, g_xCameraDeltaY, *Asura_Camera::GetPosition(), bPaused);
            }
        }

        if (bOK)
        {
            // Update the render-data...
            xRenderData.m_uStartVertex = uStartVertex;
            xRenderData.m_uNumVertices = uNumVertices;
            g_uNumParticleRenderData++;
			if( g_uNumParticleRenderData >= MAX_PARTICLE_RENDER_DATA )
			{
				Asura_Assert( false, "Increase MAX_PARTICLE_RENDER_DATA" );
				g_uNumParticleRenderData = MAX_PARTICLE_RENDER_DATA - 1;
			}
        }
    }
}
/// End of function 'Asura_PFX_Effect::Platform_RenderParticles'. ///


#ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE
////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_BeginUpdate
// Author       : Mike Healey
// Date         : 08-Sep-2005
//
// Description  : Prepares vertex buffers for updating during Update.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void Asura_PFX_System::Platform_BeginUpdate (void)
{
    UpdateVertexBuffersForNewFrame();
}
/// End of function 'Asura_PFX_System::Platform_BeginUpdate'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PFX_System::Platform_GetRenderData
// Author       : Mike Healey
// Date         : 09-Sep-2005
//
// Description  : returns area of vertex buffer big enough for given number of 
//                particles. Must be 128-byte-aligned for psglSPUWriteMappedBuffer
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////
void* Asura_PFX_System::Platform_GetRenderData(u_int uNumParticles)
{
    if (!dbg_bAllowPFXRenderingOnSPU)
    {
        return NULL;
    }
    return GetVertexBufferArea(uNumParticles);
}
/// End of function 'Asura_PFX_System::Platform_GetRenderData'. ///
#endif // #ifdef ASURA_SUPPORT_MULTITHREADED_JOB_QUEUE

/// End of file. ///
