///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_BotanicalsSystem.cpp
// Author : Catalin Arsenescu
// Date   : 09-March-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////
#include "Core\Asura.h"

#ifdef ASURA_USE_BOTANICALS

// This file's header
#include "Botanicals/Asura_PSP_BotanicalsSystem.h"

// Asura headers
#include "Camera/Asura_Camera.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Light/Asura_Lighting_Solution.h"
#include "Light/Asura_Photon.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_PSP_Graphics.h"
#include "Debugging\Asura_Assert.h"
#include "Botanicals/Asura_Botanical_Game_Render_Manager.h"

// Sony libraries

// PSP headers
#include "Botanicals/Asura_PSP_BotanicalsSystem.h"
#include "Botanicals/Asura_PSP_BotanicalsTemplate.h"
#include "Botanicals/Asura_PSP_BotanicalsShape.h"
#include "Rendering/Asura_PSP_BotanicalRenderer_Internal.h"
#include "Botanicals/Asura_PSP_BotanicalsElement.h"
#include "Botanicals/Asura_Botanical_Game_Render_Instance.h"
#include "Rendering/Asura_PSP_TextureManagement.h"

//*******************************************************************
// Static 
//*******************************************************************
const unsigned int						g_uASURA_PSP_BOTANICALRENDER_BILLBOARD_BUFFER_SIZE		= 4 * 256;

Asura_PSP_BotanicalsShape*				Asura_PSP_BotanicalsSystem::s_pxShapes					= NULL;
Asura_PSP_BotanicalsTemplate*			Asura_PSP_BotanicalsSystem::s_pxTemplates				= NULL;

u_int									Asura_PSP_BotanicalsSystem::s_uNumShapes				= 0;
u_int									Asura_PSP_BotanicalsSystem::s_uNumTemplates				= 0;

Asura_PSP_BotanicalBillboard_Vertex*	Asura_PSP_BotanicalsSystem::s_pxBillboardBuffer			= NULL; 
unsigned int							Asura_PSP_BotanicalsSystem::s_uBillboardBufferOffset	= 0; 

static Asura_PSP_BotanicalsSystem		g_xPSP_BotanicalsSystem;

////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_BGR_Manager::GetBGRManager
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
Asura_BGR_Manager& Asura_BGR_Manager::GetBGRManager( void)
{
	return( g_xPSP_BotanicalsSystem );
}
/// End of function 'Asura_BGR_Manager::GetBGRManager'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::Initialise
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_BotanicalsSystem::Initialise()
{
	if (!s_pxBillboardBuffer)
	{
		// Allocate memory for our billboard buffer
		Asura_ResourcePushType(ASURA_RESOURCE_BOTANICAL_MESH);
		{
			Asura_ResourceAllocateArray( s_pxBillboardBuffer, g_uASURA_PSP_BOTANICALRENDER_BILLBOARD_BUFFER_SIZE ); 
		}
		Asura_ResourcePopType(); 
		s_uBillboardBufferOffset = 0; 
	}
	else
	{
		Asura_Assert( false, "Asura_PSP_BotanicalRenderer has already been initialised" );
	}
}
/// End of function 'Asura_PSP_BotanicalsSystem::Initialise'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::Reset
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_BotanicalsSystem::Reset()
{
	delete[] s_pxShapes;
	s_pxShapes = NULL;
	delete[] s_pxTemplates;
	s_pxTemplates = NULL;

	s_uNumTemplates	= 0;
	s_uNumShapes	= 0;
}
/// End of function 'Asura_PSP_BotanicalsSystem::Reset'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::Shutdown
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_BotanicalsSystem::Shutdown()
{
	Reset();
	Asura_ResourceDeallocateArray( s_pxBillboardBuffer ); 
}
/// End of function 'Asura_PSP_BotanicalsSystem::Shutdown'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::Render
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_BotanicalsSystem::Render(const float fFade)
{
	// Setup some basic rendering states
	Asura_PSP_Graphics::SetUseLighting(true); 
	Asura_PSP_Graphics::SetUseColourDoubling(true);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);

	// Reset the billboard buffer
	s_uBillboardBufferOffset = 0; 
	for(unsigned int uTemplates = 0; uTemplates < s_uNumTemplates; ++uTemplates)
	{
		s_pxTemplates[ uTemplates ].Render();
	}
}
/// End of function 'Asura_PSP_BotanicalsSystem::Render'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::Update
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
bool Asura_PSP_BotanicalsSystem::Update(const float fTime)
{
	return true;
}
/// End of function 'Asura_PSP_BotanicalsSystem::Update'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::GetInstanceHeaderCount
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
u_int Asura_PSP_BotanicalsSystem::GetInstanceHeaderCount( void) const
{
	return( s_uNumTemplates);
}
/// End of function 'Asura_PSP_BotanicalsSystem::GetInstanceHeaderCount'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::GetInstanceHeader
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
const Asura_BGR_Instances_Header* const Asura_PSP_BotanicalsSystem::GetInstanceHeader( const u_int uInstanceHeaderIndex) const
{
#ifdef ASURA_ASSERT
	Asura_Assertf( uInstanceHeaderIndex < static_cast< u_int>( this->s_uNumTemplates), ("Instance Header %d is out of %d range.", uInstanceHeaderIndex, this->s_uNumTemplates));
#endif
	const Asura_BGR_Instances_Header* const pxInstanceHeader = s_pxTemplates + uInstanceHeaderIndex;

	return( pxInstanceHeader);
}
/// End of function 'Asura_PSP_BotanicalsSystem::GetInstanceHeader'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::GetInstanceHeader
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
Asura_BGR_Instances_Header* Asura_PSP_BotanicalsSystem::GetInstanceHeader( const u_int uInstanceHeaderIndex)
{
#ifdef ASURA_ASSERT
	Asura_Assertf( uInstanceHeaderIndex < s_uNumTemplates, ("Instance Header %d is out of %d range.", uInstanceHeaderIndex, s_uNumTemplates));
#endif
	Asura_BGR_Instances_Header* pxInstanceHeader = s_pxTemplates + uInstanceHeaderIndex;
	return( pxInstanceHeader);
}
/// End of function 'Asura_PSP_BotanicalsSystem::GetInstanceHeader'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_BotanicalsSystem::GetType
// Author       : Catalin Arsenescu
// Date         : 9-March-2007
////////////////////////////////////////////////////////////////////////////////
Asura_BGR_Type* Asura_PSP_BotanicalsSystem::GetType( const Asura_BGR_TypeID xTypeID)
{
#ifdef ASURA_ASSERT
	Asura_Assertf( xTypeID.GetID() < s_uNumShapes, ("xTypeID %d is out of %d range.", xTypeID.GetID(), s_uNumShapes));
#endif
	Asura_BGR_Type* pxType = s_pxShapes + xTypeID.GetID();

	return( pxType);
}
/// End of function 'Asura_PSP_BotanicalsSystem::GetType'. ///

#endif	//#ifdef ASURA_USE_BOTANICALS
