
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_KernelManagement.cpp
// Author : Andy Tate
// Date   : 18-May-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Core/Asura_PSP_KernelManagement.h"

// Asura independent header-files...
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Asura platform header-files...
#include "Core/Asura_PSP_Thread_Priorities.h"
#include "Core/Asura_PSP_VolatileMemoryManagement.h"
#include "FileAccess/Asura_PSP_FileIO.h"
#include "MoviePlayback/Asura_PSP_MoviePlayback.h"
#include "SCEUtility/Asura_PSP_SCEUtilityWrapper.h"
#include "Sound/Asura_PSP_Atrac.h"
#if defined(ASURA_USE_STREAMING)
#include "Streaming/Asura_Streaming_System.h"
#endif	// defined(ASURA_USE_STREAMING)

#include <fatms.h>
#include <kernel.h>
#include <mediaman.h>
#include <powersvc.h>
#include <stdio.h>

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// Callback IDs...
SceUID 	Asura_PSP_KernelManagement::s_iUmdCallBackId			= -1;
SceUID 	Asura_PSP_KernelManagement::s_iPowerCallBackId			= -1;
SceUID 	Asura_PSP_KernelManagement::s_iExitGameCallBackId		= -1;
SceUID 	Asura_PSP_KernelManagement::s_iMemoryStickCallBackId	= -1;

volatile u_int	Asura_PSP_KernelManagement::s_uSuspensionReasonBitfield	= 0;
volatile bool Asura_PSP_KernelManagement::s_bUMDReady = true;//assume it's good unless we hear otherwise
int Asura_PSP_KernelManagement::s_iMainThreadID = 0;
Asura_Mutex Asura_PSP_KernelManagement::s_xCPUSpeedMutex;

static Asura_SyncEvent	g_xPowerReadyEvent ("Power Ready Event", false);


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::UMDCallback
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

int	Asura_PSP_KernelManagement::UMDCallback(int iCount, int iArg, void *pParam)
{
	int err_stat;

	Asura_OutputToDebugger( ("** %s(): iCount = %d, iArg = 0x%08x\n", __FUNCTION__, iCount, iArg) );

	s_bUMDReady = ((iArg & 0x3F) == (SCE_UMD_MEDIA_IN | SCE_UMD_READABLE | SCE_UMD_READY));

	if (iArg & SCE_UMD_MEDIA_IN)
	{
		Asura_OutputToDebugger( ("* MEDIA IN\n") );
	}
	if (iArg & SCE_UMD_MEDIA_OUT)
	{
		Asura_OutputToDebugger( ("* NO MEDIA\n") );
	}

	if (iArg & SCE_UMD_READABLE)
	{
		Asura_OutputToDebugger( ("* READABLE\n") );
	}
	if (iArg & SCE_UMD_READY)
	{
		Asura_OutputToDebugger( ("* DEVICE READY\n") );
	}
	if (iArg & SCE_UMD_NOT_READY)
	{
		Asura_OutputToDebugger( ("* NOT DEVICE READY\n") );
		if (sceUmdCheckMedium())
		{
			err_stat = sceUmdGetErrorStat();
			sceKernelPrintf("*** ERROR: %s(0x%08x)\n", __FUNCTION__, err_stat);			
		}
	}

	return 0;
}

/// End of function 'Asura_PSP_KernelManagement::UMDCallback'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::ExitgameCallback
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////
#include "Sound/Asura_PSP_SimpleAudio.h"

int	Asura_PSP_KernelManagement::ExitgameCallback(int iCount, int iArg, void *pParam)
{
	Asura_OutputToDebugger( ("** %s(): iCount = %d, iArg = %x\n", __FUNCTION__, iCount, iArg) );
	
	// MGH 17/5/2007: When pressing Home and selecting Yes, the game would hang if playing a movie
	// and the memory stick was absent and the WLAN switch off! Shutting down the simple audio fixes the hang!
	Asura_PSP_SimpleAudio::Shutdown();

	// NY - 19 July 2006 - not calling suspend code for now, as it's been causing issues with exiting with drive open.
	// it shouldn't be needed anyway, as sceKernelExitGame triggers a reboot!
	// NY - 14 August 2006 - wrapping a couple of delays round this seem to stop occasional hanging for unknown reasons...
	// RDM - 21/08/07 - sceKernelExitGame call wrapped in mutex to prevent deadlock 
	// when sceKernelExitGame is called during scePowerSetClockFrequency.
	// This is an issue with the current libraries that Sony claim will be fixed
	// in 3.70, so we may want to remove this when the updated libraries are released.
	sceKernelDelayThread(10000);
	Asura_PSP_SCEUtilityWrapper::LockMutex();
	s_xCPUSpeedMutex.Lock();
	sceKernelExitGame();
	s_xCPUSpeedMutex.Unlock();
	Asura_PSP_SCEUtilityWrapper::UnlockMutex();
	sceKernelDelayThread(10000);
	return 0;
}

/// End of function 'Asura_PSP_KernelManagement::ExitgameCallback'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::PowerCallback
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

int	Asura_PSP_KernelManagement::PowerCallback(int iCount, int iArg, void *pParam)
{
	Asura_OutputToDebugger( ("** %s(): iCount = %d, iArg = %x\n", __FUNCTION__, iCount, iArg) );

	if (iArg & (SCE_POWER_CALLBACKARG_SUSPENDING | SCE_POWER_CALLBACKARG_RESUMING))
	{
		if (iArg & SCE_POWER_CALLBACKARG_SUSPENDING)
		{
			Asura_OutputToDebugger( ("** Asura_PSP_KernelManagement::PowerCallback() suspending...\n") );
			Suspend(POWER_OFF);
			Asura_OutputToDebugger( ("** Asura_PSP_KernelManagement::PowerCallback() suspend complete.\n") );
		}
	}
	else if (iArg & SCE_POWER_CALLBACKARG_RESUME_COMP)
	{
		Asura_OutputToDebugger( ("** Asura_PSP_KernelManagement::PowerCallback(): resuming...\n") );
		Resume(POWER_OFF);
		Asura_OutputToDebugger( ("** Asura_PSP_KernelManagement::PowerCallback(): resume complete.\n") );
	}
	else if (iArg & SCE_POWER_CALLBACKARG_STANDINGBY)
	{
		Asura_OutputToDebugger( ("** Asura_PSP_KernelManagement::PowerCallback() suspending...\n") );
		Suspend(POWER_OFF);
		Asura_OutputToDebugger( ("** Asura_PSP_KernelManagement::PowerCallback() suspend complete.\n") );
	}

	return 0;
}

/// End of function 'Asura_PSP_KernelManagement::PowerCallback'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::MemoryStickCallback
// Author       : Mike Healey
// Date         : 02-May-2007
//
// from https://psp.scedev.net/forums/thread/9768
//1) save your data
//2) check how much free space is on the Memory Stick
//3) remember that value
//4) if you receive a Memory Stick inserted callback, check the free space
//5) if it's the same value as before it's fairly safe to assume the 
//Memory Stick has not been swapped
// 
////////////////////////////////////////////////////////////////////////////////
int Asura_PSP_KernelManagement::MemoryStickCallback(int iCount, int iArg, void *pParam)
{
	Asura_OutputToDebugger(("%s();\n", __FUNCTION__));
	Asura_OutputToDebugger(("iArg = 0x%08x\n", iArg));

	if( iArg & SCE_FATMS_NOTIFY_ASSIGNED )
	{
		Asura_PSP_SCEUtilityWrapper::NotifyMemoryStickInserted();
	}

	return 0;
}
/// End of function 'Asura_PSP_KernelManagement::MemoryStickCallback'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::CallbackThread
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

int Asura_PSP_KernelManagement::CallbackThread(SceSize args, ScePVoid argp)
{
// RDM - 17/08/06 - don't want fp exceptions at all, in any threads
#if !defined ASURA_ENABLE_FLOATING_POINT_EXCEPTIONS \
	|| defined ASURA_PSP_ARTIST_BUILD \
	|| defined ASURA_PSP_SONY_SUBMISSION

	// Disable fp exceptions if not requested
	__asm__ volatile("ctc1 %0, $31" : : "r" (0));
#endif

	// Initialise callbacks
	int iRet;

	// 1. power callback
	iRet = sceKernelCreateCallback("ASURA_POWER_CB", PowerCallback, NULL);
	if (iRet < 0)
	{
		return iRet;
	}
	s_iPowerCallBackId = iRet;

	// 2. UMD callback
	iRet = sceKernelCreateCallback("ASURA_UMD_CB", UMDCallback, NULL);
	if (iRet < 0)
	{
		sceKernelDeleteCallback(s_iPowerCallBackId);
		return iRet;
	}
	s_iUmdCallBackId = iRet;

	// 3. Memory Stick callback
	iRet = sceKernelCreateCallback("ASURA_MEMORYSTICK_CB", MemoryStickCallback, NULL);
	if (iRet < 0)
	{
		sceKernelDeleteCallback(s_iUmdCallBackId);
		sceKernelDeleteCallback(s_iPowerCallBackId);
		return iRet;
	}
	s_iMemoryStickCallBackId = iRet;

	// 4. Exit game callback
	iRet = sceKernelCreateCallback("ASURA_EXIT_CB", ExitgameCallback, NULL);
	if (iRet < 0)
	{
		sceKernelDeleteCallback(s_iMemoryStickCallBackId);
		sceKernelDeleteCallback(s_iUmdCallBackId);
		sceKernelDeleteCallback(s_iPowerCallBackId);
		return iRet;
	}
	s_iExitGameCallBackId = iRet;

	// Register the callbacks...

	// 1. register UMD callback
	iRet = sceUmdRegisterUMDCallBack(s_iUmdCallBackId);
	if (iRet < 0)
	{
		scePowerUnregisterCallback(0);

		sceKernelDeleteCallback(s_iMemoryStickCallBackId);
		sceKernelDeleteCallback(s_iUmdCallBackId);
		sceKernelDeleteCallback(s_iPowerCallBackId);
		sceKernelDeleteCallback(s_iExitGameCallBackId);
		return iRet;
	}

	// 2. register exit game callback
	iRet = sceKernelRegisterExitCallback(s_iExitGameCallBackId);
	if (iRet < 0)
	{
		sceUmdUnRegisterUMDCallBack(s_iUmdCallBackId);
		scePowerUnregisterCallback(0);

		sceKernelDeleteCallback(s_iMemoryStickCallBackId);
		sceKernelDeleteCallback(s_iUmdCallBackId);
		sceKernelDeleteCallback(s_iPowerCallBackId);
		sceKernelDeleteCallback(s_iExitGameCallBackId);
		return iRet;
	}

	// 3. register power callback
	g_xPowerReadyEvent.WaitForSignal();
	iRet = scePowerRegisterCallback(0, s_iPowerCallBackId);
	if (iRet < 0)
	{
		sceKernelDeleteCallback(s_iMemoryStickCallBackId);
		sceKernelDeleteCallback(s_iUmdCallBackId);
		sceKernelDeleteCallback(s_iPowerCallBackId);
		sceKernelDeleteCallback(s_iExitGameCallBackId);
		return iRet;
	}

	// 4. register Memory Stick callback
	iRet = sceFatmsRegisterNotifyCallback(s_iMemoryStickCallBackId);
	if (iRet < 0)
	{
		sceUmdUnRegisterUMDCallBack(s_iUmdCallBackId);
		scePowerUnregisterCallback(0);
		sceKernelDeleteCallback(s_iMemoryStickCallBackId);
		sceKernelDeleteCallback(s_iUmdCallBackId);
		sceKernelDeleteCallback(s_iPowerCallBackId);
		sceKernelDeleteCallback(s_iExitGameCallBackId);
		return iRet;
	}

	while(1)
	{
		// (MH - 06.06.2006) was 1000us i.e. once every millisecond. 
		// I think once every 20ms is Ok.
		sceKernelDelayThreadCB(20000);	
	}
	sceKernelExitThread(0);
	return 0;
}

/// End of function 'Asura_PSP_KernelManagement::CallbackThread'. ///


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::EarlyInitialise
// Author       : Andy Tate
// Date         : 6-Sep-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_KernelManagement::EarlyInitialise (void)
{
	// KOR - register callbacks as early as we can
	const int iCallbackThreadID = sceKernelCreateThread("Callbacks", CallbackThread, iASURA_THREAD_PRIORITY_CALLBACKS, 4096, SCE_KERNEL_TH_USE_VFPU, NULL);
	Asura_Assertf(iCallbackThreadID >= 0, ("sceKernelCreateThread error:%08X\n", iCallbackThreadID));

	const int iThreadStatus = sceKernelStartThread(iCallbackThreadID, 0, 0);
	Asura_Assertf(iThreadStatus == SCE_KERNEL_ERROR_OK, ("sceKernelStartThread error:%08X\n", iThreadStatus));
}

/// End of function 'Asura_PSP_KernelManagement::EarlyInitialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::Initialise
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_KernelManagement::Initialise (void)
{
	// Initialise the variables...
	Asura_Assert((s_uSuspensionReasonBitfield == 0), "Cannot initialise if we're already suspended.");

	// Allow the power-callback...
	g_xPowerReadyEvent.SetSignal();

	// Initialise the various kernel interfaces...
	Asura_PSP_VolatileMemoryManagement::Initialise();
}

/// End of function 'Asura_PSP_KernelManagement::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::Shutdown
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_KernelManagement::Shutdown (void)
{
	// Shutdown the various kernel interfaces...
	Asura_PSP_VolatileMemoryManagement::Shutdown();

	// Callback shutdown
	sceFatmsUnRegisterNotifyCallback(s_iMemoryStickCallBackId);
	sceUmdUnRegisterUMDCallBack(s_iUmdCallBackId);
	scePowerUnregisterCallback(0);

	sceKernelDeleteCallback(s_iMemoryStickCallBackId);
	sceKernelDeleteCallback(s_iUmdCallBackId);
	sceKernelDeleteCallback(s_iPowerCallBackId);
	sceKernelDeleteCallback(s_iExitGameCallBackId);

	// Reset the suspension reason...
	s_uSuspensionReasonBitfield	= 0;
}

/// End of function 'Asura_PSP_KernelManagement::Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::Suspend
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_KernelManagement::Suspend (const SuspendReason eReason)
{
	const u_int	uReasonBit	= (1 << eReason);

	// Are we already suspended?
	if(s_uSuspensionReasonBitfield == 0)
	{
		// RDM - 21/08/06 - Moved set of the suspension reason here to prevent multiple suspensions 
		// being called from different threads
		
		// Set the appropriate bit...
		s_uSuspensionReasonBitfield	|= uReasonBit;

		// pause...
		Asura_Timers::SetSystemPause( true );

		// Shutdown the appropriate systems...
#if defined(ASURA_USE_STREAMING)
		Asura_Streaming_System::Suspend();
		Asura_OutputToDebugger(("Asura_Streaming_System::Suspend() done\n"));
#endif	// defined(ASURA_USE_STREAMING)
#if defined(ASURA_SUPPORT_PSP_MPEG)
		Asura_PSP_MoviePlayback::Suspend();
		Asura_OutputToDebugger(("Asura_PSP_MoviePlayback::Suspend() done\n"));
#endif	// defined(ASURA_SUPPORT_PSP_MPEG)
		Asura_PSP_Atrac::Suspend();
		Asura_OutputToDebugger(("Asura_PSP_Atrac::Suspend() done\n"));
		
		Asura_PSP_FileIO::Suspend();
		Asura_OutputToDebugger(("Asura_PSP_FileIO::Suspend() done\n"));
		
		Asura_PSP_VolatileMemoryManagement::Suspend();
		Asura_OutputToDebugger(("Asura_PSP_VolatileMemoryManagement::Suspend() done\n"));

		// Project-specific suspension...
		Project_Suspend();
		Asura_OutputToDebugger(("Asura_PSP_KernelManagement::Project_Suspend() done\n"));
	}
	else
	{
		// RDM - 21/08/06 - Added this case to prevent suspension occurring twice. Still need to 
		// set the suspension reason, however.

        // (MH - 09.03.2007) check we're not suspending twice for the same reason
		// AJG - 11/07/07 - Allow POWER_OFF, since it's easy for someone to hit the sleep button twice before the PSP actually manages to enter sleep mode, but that is handled without error.
        // (MH - 21.06.2007) replaced assert with debugger output as assert causes stack overflow.
		if ((eReason != POWER_OFF) && (s_uSuspensionReasonBitfield & uReasonBit))
		{
			Asura_OutputToDebugger_Retail(("already suspended for reason %d - probable lockup imminent\n", eReason));
		}

		// Set the appropriate bit...
		s_uSuspensionReasonBitfield	|= uReasonBit;
	}
}

/// End of function 'Asura_PSP_KernelManagement::Suspend'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_KernelManagement::Resume
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_KernelManagement::Resume (const SuspendReason eReason)
{
	const u_int	uReasonBit	= (1 << eReason);

	// Reset the appropriate bit...
	s_uSuspensionReasonBitfield	&= ~uReasonBit;

	// Are we already suspended?
	if(s_uSuspensionReasonBitfield == 0)
	{
		// Shutdown the appropriate systems...
		Asura_PSP_VolatileMemoryManagement::Resume();
		Asura_PSP_FileIO::Resume();
		Asura_PSP_Atrac::Resume();
#if defined(ASURA_SUPPORT_PSP_MPEG)
		Asura_PSP_MoviePlayback::Resume();
#endif	// defined(ASURA_SUPPORT_PSP_MPEG)
#if defined(ASURA_USE_STREAMING)
		Asura_Streaming_System::Resume();
#endif	// defined(ASURA_USE_STREAMING)

		// Restore the time-scale...
		Asura_Timers::SetSystemPause(false);

		// Project-specific resumption...
		Project_Resume();
	}
}

/// End of function 'Asura_PSP_KernelManagement::Resume'. ///

bool Asura_PSP_KernelManagement::IsUMDReady()
{
#if defined(ASURA_PSP_UMD) || defined(ASURA_PSP_ARTIST_BUILD)
	return s_bUMDReady;
#else // #if defined(ASURA_PSP_UMD) || defined(ASURA_PSP_ARTIST_BUILD)
	return true;
#endif // #if defined(ASURA_PSP_UMD) || defined(ASURA_PSP_ARTIST_BUILD)
}

void Asura_PSP_KernelManagement::StoreMainThreadID()
{
    s_iMainThreadID = sceKernelGetThreadId();
}

// Set the PSP CPU (and bus) clock speed
// RJT TODO: Better place?
//
bool Asura_PSP_KernelManagement::SetCPUSpeed( ESpeed eSpeed )
{
	bool bSuccess = false;

	// RDM - 21/08/07 - CPU speed manipulation wrapped in mutex to prevent deadlock 
	// when sceKernelExitGame is called during scePowerSetClockFrequency.
	// This is an issue with the current libraries that Sony claim will be fixed
	// in 3.70, so we may want to remove this when the updated libraries are released.
	s_xCPUSpeedMutex.Lock();

#if ( SCE_DEVKIT_VERSION >= 0x03050000 )
	switch( eSpeed )
	{
		case SPEED_SLOW:
		{
			bSuccess = scePowerSetClockFrequency( 190, 95, 95 ) ? false : true;
			break;
		}

		case SPEED_FULL:
		{
			bSuccess = scePowerSetClockFrequency( 333, 333, 166 ) ? false : true;
			break;
		}

		case SPEED_STANDARD: // falls through
		default:
		{		
			bSuccess = scePowerSetClockFrequency( 222, 222, 111 ) ? false : true;
			break;
		}
	}

#if !defined ASURA_PSP_SONY_SUBMISSION
	if( bSuccess )
	{
		sceKernelPrintf( "    CPU clocked at %iMHz\n", scePowerGetCpuClockFrequencyInt() );    
	}
#endif	
#endif

	s_xCPUSpeedMutex.Unlock();
	return bSuccess;
}

/// End of file. ///
