
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_VolatileMemoryManagement.cpp
// Author : Andy Tate
// Date   : 9-May-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include <kernel.h>
#include "Core/Asura.h"

// This file's header...
#include "Core/Asura_PSP_VolatileMemoryManagement.h"

// Asura independent header-files...
#include "Core/Memory/Asura_MemoryBank.h"
#include "Debugging/Asura_Logfile.h"



////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// Current data with in memory allocated...
bool	Asura_PSP_VolatileMemoryManagement::s_bInitialised;
bool	Asura_PSP_VolatileMemoryManagement::s_bSuspended;
bool	Asura_PSP_VolatileMemoryManagement::s_bPendingInitialise;
u_char*	Asura_PSP_VolatileMemoryManagement::s_pucDataStart		= NULL;
u_int	Asura_PSP_VolatileMemoryManagement::s_uSizeOfData		= 0;
u_int	Asura_PSP_VolatileMemoryManagement::s_uCurrentOffset	= 0;

// The memory-block data...
u_int							Asura_PSP_VolatileMemoryManagement::s_uNumMemoryBlocksInUse	= 0;
Asura_PSP_VolatileMemoryBlock	Asura_PSP_VolatileMemoryManagement::s_axMemoryBlocks[ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS];



Asura_PSP_VolatileMemoryBlock::Asura_PSP_VolatileMemoryBlock()
{
	m_uSize = 0;
	m_pucDataStart = NULL;
	m_pfnAllocationCallback = NULL;
	m_pfnDeallocationCallback = NULL;
}

Asura_PSP_VolatileMemoryBlock::~Asura_PSP_VolatileMemoryBlock()
{
	if (m_pucDataStart)
	{
		Shutdown();
	}
}

void Asura_PSP_VolatileMemoryBlock::Initialise(	const u_int uNumBytes,
												u_char* pucDataStart,
												AllocationCallbackFunction* pfnAllocationCallback,
												DeallocationCallbackFunction* pfnDeallocationCallback )
{
	Asura_Assert( !m_uSize && !m_pucDataStart && !m_pfnAllocationCallback && !m_pfnDeallocationCallback, "not shutdown");
	m_uSize = uNumBytes;
	m_pfnAllocationCallback = pfnAllocationCallback;
	m_pfnDeallocationCallback = pfnDeallocationCallback;

	Reallocate(pucDataStart);
}

void Asura_PSP_VolatileMemoryBlock::Shutdown()
{
	Deallocate();

	m_uSize = 0;
	m_pucDataStart = NULL;
	m_pfnAllocationCallback = NULL;
	m_pfnDeallocationCallback = NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryBlock::Deallocate
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_VolatileMemoryBlock::Deallocate (void)
{
	// Validation...
	Asura_Assert((m_pucDataStart != NULL), "We have no data to deallocate.");

	// Do we have an allocation function?
	if(m_pfnDeallocationCallback != NULL)
	{
		m_pfnDeallocationCallback(m_pucDataStart);
	}

	// Reset the data pointer..,
	m_pucDataStart	= NULL;
}

/// End of function 'Asura_PSP_VolatileMemoryBlock::Deallocate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryBlock::Reallocate
// Author       : Andy Tate
// Date         : 18-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_VolatileMemoryBlock::Reallocate (u_char* pucDataStart)
{
	// Validation...
	Asura_Assert((m_pucDataStart == NULL), "We already have data - reallocation shouldn't be occurring.");

	// Do we have an allocation function?
	if(m_pfnAllocationCallback != NULL)
	{
		m_pfnAllocationCallback(pucDataStart);
	}

	// Set the data pointer..,
	m_pucDataStart	= pucDataStart;
}

/// End of function 'Asura_PSP_VolatileMemoryBlock::Reallocate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryManagement::Initialise
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_VolatileMemoryManagement::Initialise (void)
{
	if (s_bSuspended)
	{
		// flag to initialise when we resume
		s_bPendingInitialise = true;
		return;
	}

	// Validation...
	if(s_pucDataStart == NULL)
	{
		// Allocate the memory-blocks...
		void* pData;

		const int iStatus = sceKernelVolatileMemLock(SCE_KERNEL_VOLATILE_MEM_DEFAULT, &pData, &s_uSizeOfData);

		// Did our allocation succeed?
		if(iStatus == 0)
		{
			// Initialise the variables...
			s_pucDataStart	= static_cast<u_char*>(pData);
			Asura_OutputToDebugger(("*** Successfully initialised volatile memory-bank (%u bytes). ***\n", s_uSizeOfData));
		}
		else
		{
			Asura_Assert(false, "Failed to allocate PSP volatile memory");
		}
	}
	else
	{
		Asura_Assert(false, "Volatile memory-manager has already been initialised.");
	}

	s_bInitialised = true;
}

/// End of function 'Asura_PSP_VolatileMemoryManagement::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryManagement::Shutdown
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_VolatileMemoryManagement::Shutdown (void)
{
	// Validation...
	if(s_pucDataStart == NULL)
	{
		Asura_Assert(false, "Volatile memory-manager is already shutdown.");
	}
	else
	{
		// Deallocate the memory-blocks...
		for (u_int u = 0; u < s_uNumMemoryBlocksInUse; u++)
		{
			s_axMemoryBlocks[u].Shutdown();
		}

		// Reset the variables...
		s_pucDataStart			= NULL;
		s_uNumMemoryBlocksInUse	= 0;
		s_uCurrentOffset		= 0;
		s_uSizeOfData			= 0;

		// Finally, release the memory...
		sceKernelVolatileMemUnlock(0);
		Asura_OutputToDebugger(("*** Successfully shutdown volatile memory. ***\n"));
	}

	s_bInitialised = false;
}

/// End of function 'Asura_PSP_VolatileMemoryManagement::Shutdown'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryManagement::Allocate
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

Asura_Handle Asura_PSP_VolatileMemoryManagement::Allocate (const u_int uNumBytes,
														   AllocationCallbackFunction* pfnAllocationCallback,
														   DeallocationCallbackFunction* pfnDeallocationCallback)
{
	// Do we a memory to use?
	if(s_pucDataStart == NULL)
	{
		Asura_Assert(false, "Cannot allocate volatile memory because the manager has not been initialised.");
		return(ASURA_INVALID_HANDLE);
	}
	else if(s_uNumMemoryBlocksInUse < ASURA_PSP_NUM_VOLATILE_MEMORY_BLOCKS)
	{
		const u_int	uTotalBytes	= s_uCurrentOffset + uNumBytes;

		// Is there enough memory?
		if(uTotalBytes <= s_uSizeOfData)
		{
			const Asura_Handle iHandle = s_uNumMemoryBlocksInUse++;

			u_char* pucDataStart = &s_pucDataStart[s_uCurrentOffset];
			Asura_PSP_VolatileMemoryBlock* pxBlock = &s_axMemoryBlocks[iHandle];
			pxBlock->Initialise(uNumBytes, pucDataStart, pfnAllocationCallback, pfnDeallocationCallback);

			s_uCurrentOffset = uTotalBytes;
			Asura_OutputToDebugger(("+++ Successfully allocated %u bytes of volatile memory. +++\n", uNumBytes));
			return(iHandle);
		}
		else
		{
			Asura_Assert(false, "Not enough volatile memory for allocation.");
			return(ASURA_INVALID_HANDLE);
		}
	}
	else
	{
		Asura_Assert(false, "Not enough volatile memory-blocks for another allocation.");
		return(ASURA_INVALID_HANDLE);
	}
}

/// End of function 'Asura_PSP_VolatileMemoryManagement::Allocate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryManagement::Deallocate
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_VolatileMemoryManagement::Deallocate (Asura_Handle& iHandle)
{
	// Do we a memory to use?
	if(s_pucDataStart == NULL)
	{
		Asura_Assert(false, "Cannot deallocate volatile memory because the manager has not been initialised.");
		return(false);
	}
	else if(iHandle == ASURA_INVALID_HANDLE)
	{
		Asura_Assert(false, "Unexpected handle for volatile memory-block deallocation.");
		return(false);
	}
	else
	{
		Asura_PSP_VolatileMemoryBlock* pxBlock = &s_axMemoryBlocks[iHandle];

		const u_int uSize = pxBlock->GetSize();

		// Remove this memory-block...
		pxBlock->Shutdown();
		s_uNumMemoryBlocksInUse--;

		// Is this the last memory-block?
		if(iHandle == s_uNumMemoryBlocksInUse)
		{
			s_uCurrentOffset = s_uCurrentOffset - uSize;
		}

		// Blank the handle...
		iHandle	= ASURA_INVALID_HANDLE;

		// We've got our memory-block - initialise the data...
		Asura_OutputToDebugger(("--- Successfully deallocated %u bytes from volatile memory. ---\n", uSize));
		return(true);
	}
}

/// End of function 'Asura_PSP_VolatileMemoryManagement::Deallocate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryManagement::Suspend
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_VolatileMemoryManagement::Suspend (void)
{
	s_bSuspended = true;
	if (!s_bInitialised)
	{
		return;
	}

	// Deallocate each memory-block in-turn...
	for(u_int u = 1; u <= s_uNumMemoryBlocksInUse; u++)
	{
		const u_int	uIndex = s_uNumMemoryBlocksInUse - u;
		Asura_PSP_VolatileMemoryBlock* pxBlock = &s_axMemoryBlocks[uIndex];

		// Do we have a block to suspend?
		if(pxBlock != NULL)
		{
			pxBlock->Deallocate();
		}
	}

	// Reset the variables...
	s_uCurrentOffset	= 0;
	s_uSizeOfData		= 0;
	s_pucDataStart		= NULL;

	// Finally, release the memory...
	sceKernelVolatileMemUnlock(0);
	Asura_OutputToDebugger(("*** Successfully suspended volatile memory-management. ***\n"));
}

/// End of function 'Asura_PSP_VolatileMemoryManagement::Suspend'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_VolatileMemoryManagement::Resume
// Author       : Andy Tate
// Date         : 9-May-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_VolatileMemoryManagement::Resume (void)
{
	s_bSuspended = false;
	if (s_bPendingInitialise)
	{
		// we got an Initialise() call after the Suspend was called.
		s_bPendingInitialise = false;
		Initialise();
		return;
	}
	if (!s_bInitialised)
	{
		return;
	}

	void*		pData;

	const int	iStatus	= sceKernelVolatileMemLock(SCE_KERNEL_VOLATILE_MEM_DEFAULT, &pData, &s_uSizeOfData);

	// Did our allocation succeed?
	if(iStatus == 0)
	{
		// Initialise the variables...
		s_pucDataStart	= static_cast<u_char*>(pData);

		// Reallocate each memory-block in-turn...
		for(u_int u = 0; u < s_uNumMemoryBlocksInUse; u++)
		{
			Asura_PSP_VolatileMemoryBlock* pxBlock = &s_axMemoryBlocks[u];

			// Do we have a block to suspend?
			if(pxBlock != NULL)
			{
				pxBlock->Reallocate(&s_pucDataStart[s_uCurrentOffset]);
				s_uCurrentOffset += pxBlock->GetSize();
			}
		}

		Asura_OutputToDebugger(("*** Successfully resumed volatile memory-management. ***\n"));
	}
	else
	{
		Asura_Assert(false, "Failed to resume volatile memory-management.");
	}
}

/// End of function 'Asura_PSP_VolatileMemoryManagement::Resume'. ///


/// End of file. ///
