/**
*
* \file Asura_Platform_MemoryManagement.cpp
*
* Created by Kevin Lea - 08/12/2001 16:05:53
*
* Copyright (C) 2001 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"

#include "ProjectSettings/Asura_PSPConfiguration.h"

// Asura_MemoryManagement_System defined in Asura.h, so for once no
// other header required
#include "Debugging/Asura_Assert.h"

// Engine independent.
#include "Core/Asura_Core.h"
#include "Core/Asura_PlatformBasics.h"
#include "Core/Asura_Multithread.h"

// Engine PSP.
#include "Core/Asura_PSP_KernelManagement.h"
#if !defined ASURA_PSP_SONY_SUBMISSION
	#include <stdio.h>
	#include "Debugging/Asura_PSP_ScreenLog.h"
#endif
#include "Debugging/Asura_PSP_GPIO.h"

#include <kernel.h>
#include <libvfpu.h>

// C O N S T A N T S /////////////////////////////////////////////////////////
// KibiByte and MibiByte
static const u_int uASURA_PSP_MM_KIB = 1<<10;
static const u_int uASURA_PSP_MM_MIB = 1<<20;

// D E B U G V A R S /////////////////////////////////////////////////////////
DEBUGVAR u_int dbg_uLeaveFreeForInfrastructure = 448;	// yikes, this is big. need to find ways of reducing it.
DEBUGVAR u_int dbg_uLeaveFreeForAdhoc = 64;
DEBUGVAR u_int dbg_uMaxHeapSize = 0; // 19221504

// G L O B A L S /////////////////////////////////////////////////////////////
static unsigned int g_uHeapSize;
static unsigned int g_uEndOfMainMem = 0;

// used to determine whether we need to unlock the mutex (which may or may not 
// have been used depending on the value of s_bMultithreadingEnabled when 
// Platform_Lock() was called.
static u_int g_uNumLocks;

static bool g_bOutOfMemoryScreenEnabled = true;

// C + +   F U N C T I O N S /////////////////////////////////////////////////

// Access function to enable/disable the red screen of death when out of memory.
void Asura_PSP_EnableOutOfMemoryScreen(bool bEnable)
{
	g_bOutOfMemoryScreenEnabled = bEnable;
}


// C   F U N C T I O N S /////////////////////////////////////////////////////
// KOR - 30/03/2005 - Functions for dlmalloc. Don't ask me how they work
// - I just copied the PS2 version.

extern "C"
{
/*
	We need to give an estimate of the amount of memory that needs to
	be left free for the system's use.
*/
static u_int Asura_PSP_LeaveFree(const u_int uMaxFree)
{
	register u_int uLeaveFree;

	/*
		JG - A little bit more memory for new library, please. Seems to
		be somewhere in between 450 and 475 kb.

		KOR - 12/01/2006 - Going up to 2.6.0 seems to have changed
		memory requirements. Increased by 21Kb accordingly.
		KOR - 26/01/2006 - Seems alright now. Back down again.

		Initialisation order appears to matter - perhaps there's fragmentation
		in the Sony kernel memory management.
	*/
	uLeaveFree = 475;

#ifdef ASURA_SUPPORT_PSP_MPEG
	/*
		(AW - 17/11/05) - Allow a little bit more if we need the mpeg
		libraries. (between 53 and 69, makes 565kb in total)
	*/
	uLeaveFree += 69 + 64;
#endif

#if defined(ASURA_USE_STREAMING)
	// ART 15/05/06 - Extra stack for streaming thread...
	uLeaveFree	+= 64;
#endif	// defined(ASURA_USE_STREAMING)

#ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK
	uLeaveFree += dbg_uLeaveFreeForInfrastructure;
#elif defined(ASURA_SUPPORT_EXTERNAL_NETWORK)
	uLeaveFree += dbg_uLeaveFreeForAdhoc;
#endif

#ifdef ASURA_PSP_GAMESHARING
	uLeaveFree += 400;//MKR 23/11/06: game sharing
#endif

	return uLeaveFree * uASURA_PSP_MM_KIB;
}

#define MORECORE_FAILURE    ((void*)(-1))
void* Asura_PSP_GetHeapMem(int iSize)
{
	static bool bFirstTime = true;
    if (bFirstTime && (iSize > 0))
	{
		const u_int memTotalFree = sceKernelTotalFreeMemSize();
		const u_int memMaxFree = sceKernelMaxFreeMemSize();

		bFirstTime = false;

		sceKernelPrintf("sceKernelTotalFreeMemSize: %u bytes\n",memTotalFree);		
		sceKernelPrintf("sceKernelMaxFreeMemSize: %u bytes\n",memMaxFree);		
		
		g_uHeapSize = sceKernelMaxFreeMemSize() - Asura_PSP_LeaveFree(memMaxFree);

#if defined ASURA_MEMORYMANAGEMENT_ENABLED
		/*
			Leave something for the utility memory,
		*/
		g_uHeapSize -= ((g_uHeapSize > (32*uASURA_PSP_MM_MIB)) ? 4 : 2) * uASURA_PSP_MM_MIB;
#endif
		if (dbg_uMaxHeapSize && (g_uHeapSize > dbg_uMaxHeapSize))
		{
			g_uHeapSize = dbg_uMaxHeapSize;
		}

		sceKernelPrintf("Heap size: %i bytes\n", g_uHeapSize);		

		// Create FPL and allocate from it
		int iRet;
		void* pHeapStart;
		iRet = sceKernelCreateFpl("Asura", SCE_KERNEL_PRIMARY_USER_PARTITION, SCE_KERNEL_FA_THPRI, g_uHeapSize, 1, NULL);
		if(iRet < 0)
		{
			sceKernelPrintf("Failed to create FPL\n");			
			return MORECORE_FAILURE;
		}
		iRet = sceKernelTryAllocateFpl(iRet, &pHeapStart);
		if(iRet < 0)
		{
			sceKernelPrintf("Failed to allocate from FPL\n");			
			return MORECORE_FAILURE;
		}
		u_int uHeapStart = (u_int)pHeapStart;

		g_uEndOfMainMem = uHeapStart + g_uHeapSize;
		// align prog end to page size (4096)
		if(uHeapStart)
		{
			uHeapStart = (uHeapStart + 4095) & ~4095;
			return (void*)uHeapStart;
		}
	}

    if (iSize == 0)
    {
        // return end of mem
        return (void*)g_uEndOfMainMem;
    }
    else
    {
        return MORECORE_FAILURE;
    }
}

void Asura_PSP_OutOfMem(u_int uBytes)
{
#if !defined ASURA_PSP_SONY_SUBMISSION
	if (g_bOutOfMemoryScreenEnabled)
	{
		// Handy for debugging purposes to have mallinfo available.
		const struct mallinfo xMallInfo = mallinfo();

		char szDebuggerBuf[2048];
		int len = 0;

#if 0
		// Full information
		len += sprintf( szDebuggerBuf + len, "Out of memory:\n\
	requested bytes : %u\n\
	space alloced from sys : %i\n\
	number of free chunks: %i\n\
	number of fastbin blocks: %i\n\
	maximum total allocated space: %i\n\
	space available in freed fastbin blocks: %i\n\
	total allocated space: %i\n\
	total free space: %i\n\
	top-most, releasable space: %i\n",
			uBytes,
			xMallInfo.arena,
			xMallInfo.ordblks,
			xMallInfo.smblks,
			xMallInfo.usmblks,
			xMallInfo.fsmblks,
			xMallInfo.uordblks,
			xMallInfo.fordblks,
			xMallInfo.keepcost);
#else
		// Short information
		len += sprintf( szDebuggerBuf + len, "Out of memory:\n\
	requested bytes : %u\n\
	maximum total allocated space: %i\n\
	total allocated space: %i\n\
	total free space: %i\n\
	top-most, releasable space: %i\n",
			uBytes,
			xMallInfo.usmblks,
			xMallInfo.uordblks,
			xMallInfo.fordblks,
			xMallInfo.keepcost);
#endif

		fputs(szDebuggerBuf,stdout);

		Asura_PSP_ScreenLog::FatalError(szDebuggerBuf);
	}
#endif // #if !defined ASURA_PSP_SONY_SUBMISSION

	// light up GPO light 6
	Asura_PSP_GPIO::SetGPO((1 << 6));
	// Now wait a little while, we're going to crash soon.
	u_int uCount = 0;
	while(uCount++ < 5)
	{
		// pause 1 second.
		sceKernelDelayThreadCB(1000);
	}

	// force bluescreen, which should have the printfs above.
	{ asm("break 1"); asm("nop"); }
}


int Asura_PSP_DLMallocPreAction( void )
{
    Asura_MemoryManagement_System::Platform_Lock();
    return 0;
}


int Asura_PSP_DLMallocPostAction( void )
{
    Asura_MemoryManagement_System::Platform_Unlock();
    return 0;
}


} // extern "C"


static Asura_Mutex& GetMemManagementMutex( void )
{
    static Asura_Mutex xMutex;
    return xMutex;
}


void Asura_MemoryManagement_System::Platform_Lock(void)
{
    if (s_bMultithreadingEnabled)
	{
		GetMemManagementMutex().Lock();
		++g_uNumLocks;
	}
	else
	{
		Asura_Assert( Asura_Core::IsMainThread(), "should only allocate from main thread unless using mutex" );
	}
}


void Asura_MemoryManagement_System::Platform_Unlock(void)
{
	// if there is an outstanding lock but we stopped requiring use of the 
	// mutex, unlock it anyway
    if (g_uNumLocks)
	{
		--g_uNumLocks;
		GetMemManagementMutex().Unlock();
	}
}

#ifdef ASURA_MEMORYMANAGEMENT_ENABLED

void Asura_MemoryManagement_System::Platform_Initialise(void)
{
}

void Asura_MemoryManagement_System::Platform_Shutdown(void)
{
}

#ifdef ASURA_MEMORYMANAGEMENT_FILL

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Asura_MemoryManagement_System::Platform_FillMem().
// PARAMETERS   pAddr          - Address, or actual address if aligned.
//              uSize          - Size, or actual size if aligned.
//              eEventID       - Type of memory management event.
//              uAlignment     - Alignment type, or 0 if no alignment.
//              pReportedAddr  - Reported address if memory is aligned.
//              uReportedSize  - Reported size if memory is aligned.
// RETURNS      None.
// DESCRIPTION  Fills specified region of memory. If the caller requested
//              memory aligned on a 64-byte boundary and a multiple of 64
//              bytes in size then we ensure the returned memory is not
//              present in the cache.
// AUTHOR       Morgan Parry - 29.07.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Asura_MemoryManagement_System::Platform_FillMem( void* pAddr,
                                                      u_int uSize,
                                                      ASURA_MM_EVENT_ID eEventID,
                                                      u_int uAlignment,
                                                      void* pReportedAddr /*= NULL*/,
                                                      u_int uReportedSize /*= 0*/ )
{
    if( uAlignment )
    {
        Asura_Assert( pReportedAddr != NULL, "" );
        Asura_Assert( uReportedSize > 0, "" );
    }

    u_char ucFill;
	switch( eEventID )
    {
        case ASURA_NEW:
        case ASURA_NEWALIGN:
        case ASURA_NEWARRAY:
        case ASURA_ALLOC:
            ucFill = ucFillValueAllocate;           
            break;

        case ASURA_DELETE:
        case ASURA_DELETEALIGN:
        case ASURA_DELETEARRAY:
        case ASURA_DEALLOC:
            ucFill = ucFillValueFree;           
            break;

        default:
            Asura_Assert( false, "Invalid memory manager event ID." );
            return;
    }
    
    // Fill.
	if (Asura_PSP_KernelManagement::IsSuspended())
	{
        // (MH - 12.07.2007) get coprocessor unusable when suspended
		memset( pAddr, ucFill, uSize );
	}
	else
	{
		sceVfpuMemset( pAddr, ucFill, uSize );
	}
}
// End function Asura_MemoryManagement_System::Platform_FillMem().

#endif // #ifdef ASURA_MEMORYMANAGEMENT_FILL

void Asura_MemoryManagement_System::Platform_PostDeallocate( void )
{
}

#else // #ifdef ASURA_MEMORYMANAGEMENT_ENABLED

void* Asura_PSP_AllocateAlign(unsigned int bytes, enum ASURA_ALIGN_ID alignment)
{
    int iTotalBytes = bytes + alignment + 4;    // 4 bytes before reported address is ptr to start of block

    char* const pcMem = (char*)Asura_Allocate(iTotalBytes);
    
    if (!pcMem) 
    {
        return NULL;
    }

    unsigned int offset = (alignment - ((unsigned int)(pcMem + 4)) & (alignment-1));
	char* const pcReportedMem = (char*)((unsigned char*)(pcMem + 4) + offset);
    u_int* const pcActualAddressStore = (u_int*)(pcReportedMem - 4);
    *pcActualAddressStore = (u_int)pcMem;
	//sceKernelPrintf("Asura_PS2_AllocateAlign: %d bytes, %d align, addr = 0x%8p\n", bytes, alignment, pcReportedMem);
    return pcReportedMem;
}

void Asura_PSP_DeallocateAlign(void* address)
{
    if (!address) return;

    u_int* puActualAddress = ((u_int*)address) - 1;
    u_int uActualAddress = *puActualAddress;
	//sceKernelPrintf("Asura_PS2_DeallocateAlign: addr = 0x%8p, actual = 0x%8x\n", address, uActualAddress);
    //printf("Asura_PS2_DeallocateAlign: addr = 0x%8p, actual = 0x%8x\n", address, uActualAddress);
    Asura_Deallocate((void*)uActualAddress);
}

#endif // else def ASURA_MEMORYMANAGEMENT_ENABLED

#if defined ASURA_MEMORYMANAGEMENT_ENABLED

static SceUID g_iAsura_MemoryUtility_Arena  = 0;

void Asura_MemoryUtility::Platform_Initialise(void)
{
	register const unsigned uMaxFree = sceKernelMaxFreeMemSize();
	const unsigned uLeaveFree = Asura_PSP_LeaveFree(uMaxFree);

	Asura_Assert(uMaxFree >= (uLeaveFree + 1*uASURA_PSP_MM_MIB),
		"less than 1MiB for utility memory");

	sceKernelPrintf("Utility memory size: %u (left free %u)\n",
		uMaxFree-uLeaveFree,uLeaveFree);

	g_iAsura_MemoryUtility_Arena = sceKernelCreateVpl("Utility",
		SCE_KERNEL_PRIMARY_USER_PARTITION,SCE_KERNEL_FA_THPRI,
		uMaxFree-uLeaveFree,NULL);
}

void Asura_MemoryUtility::Platform_Shutdown(void)
{
	(void) sceKernelDeleteVpl(g_iAsura_MemoryUtility_Arena);
	g_iAsura_MemoryUtility_Arena = 0;
}

void* Asura_MemoryUtility::Platform_PreAllocate(/*const*/ size_t size)
{
	void* pAddr = 0;

#ifdef ASURA_ASSERT
	u_int uReturnVal = 
#endif
		sceKernelTryAllocateVpl(g_iAsura_MemoryUtility_Arena,size,&pAddr);

	Asura_Assertf( uReturnVal == SCE_KERNEL_ERROR_OK, ( "Error 0x%08x allocating memory inside Asura_MemoryUtility_Arena", uReturnVal ));

	return pAddr;
}

void* Asura_MemoryUtility::Platform_PostAllocate(void* /*const*/ pAddr)
{
	return pAddr;
}

bool Asura_MemoryUtility::Platform_PreDeallocate(void* /*const*/ pAddr)
{
	(void) sceKernelFreeVpl(g_iAsura_MemoryUtility_Arena,pAddr);
	return true;
}

void Asura_MemoryUtility::Platform_PostDeallocate(void* /*const*/ pAddr)
{
}

#endif
