//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Multithread.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     PSP-specific code for multithreading code
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Core/Asura_Multithread.h"

// Asura independent
#include "Core/Asura_Maths.h"
#include "Debugging/Asura_Assert.h"


//TODO: Initialise this array and use it for the stack size
u_int Asura_Thread_System::s_auStackSizes[]  =	{ 0		//ASURA_MULTITHREAD_STACK_SIZE_SMALL
												, 0		//ASURA_MULTITHREAD_STACK_SIZE_MEDIUM
												, 0 };	//ASURA_MULTITHREAD_STACK_SIZE_LARGE

// C L A S S   D E C L A R A T I O N S ///////////////////////////////////////
class Asura_PSP_ThreadParams
{
public:

	Asura_PSP_ThreadParams( Asura_ThreadFunc pfnThreadFunc, const void* pUserData )
		: m_pfnThreadFunc (pfnThreadFunc)
		, m_pUserData (pUserData)
	{
	}

	// data
	Asura_ThreadFunc m_pfnThreadFunc;
	const void* m_pUserData;
};

// S T A T I C   F U N C T I O N S ///////////////////////////////////////////
static int Asura_PSP_ThreadStartup( SceSize uSize, void* pParam  )
{
// RDM - 17/08/06 - don't want fp exceptions at all, in any threads
#if !defined ASURA_ENABLE_FLOATING_POINT_EXCEPTIONS \
	|| defined ASURA_PSP_ARTIST_BUILD \
	|| defined ASURA_PSP_SONY_SUBMISSION

	// Disable fp exceptions if not requested
	__asm__ volatile("ctc1 %0, $31" : : "r" (0));
#endif

	const Asura_PSP_ThreadParams* pxParams = reinterpret_cast<const Asura_PSP_ThreadParams*>(pParam);
	Asura_ThreadFunc pfnThreadFunc = pxParams->m_pfnThreadFunc;
	const void* pUserData = pxParams->m_pUserData;
	pfnThreadFunc(pUserData);

	sceKernelExitDeleteThread( 0 );

	return 0;
}

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
u_int Asura_Thread_System::Platform_GetNumHardwareThreads()
{
    return 1;
}

bool Asura_Thread_System::Platform_CreateThread( const char* szThreadName, Asura_ThreadFunc pfnThreadFunc, const void* pUserData, const ASURA_MULTITHREAD_STACK_SIZE eStackSize )
{
	Asura_PSP_ThreadParams xParams(pfnThreadFunc, pUserData);

	int iCurrentThreadPriority = sceKernelGetThreadCurrentPriority();
	int iStacksize = 40*1024;	// default stacksize

	int iThreadID = sceKernelCreateThread(szThreadName, Asura_PSP_ThreadStartup, iCurrentThreadPriority, iStacksize, SCE_KERNEL_TH_USE_VFPU, NULL );
	Asura_Assert(iThreadID > 0, "sceKernelCreateThread failed");
	if( iThreadID <= 0 ) { asm("break 1"); asm("nop"); } // can't recover from this, just die

	int iThreadStatus = sceKernelStartThread(iThreadID, sizeof(xParams), &xParams);
	Asura_Assertf(iThreadStatus == SCE_KERNEL_ERROR_OK, ("sceKernelStartThread error:%08X\n", iThreadStatus));
	if( iThreadStatus != SCE_KERNEL_ERROR_OK ) { asm("break 1"); asm("nop"); } // can't recover from this, just die

	YieldCurrentThread();

	return (iThreadID > 0) && (iThreadStatus == SCE_KERNEL_ERROR_OK);
}

void Asura_Thread_System::Platform_YieldCurrentThread()
{
	sceKernelRotateThreadReadyQueue(SCE_KERNEL_TPRI_RUN);
}

void Asura_Thread_System::Platform_DelayCurrentThread(float fTimeInSeconds, bool bAllowCallbacks)
{
	if (bAllowCallbacks)
	{
		sceKernelDelayThreadCB( Asura_Maths::FloatToInt_Chop(fTimeInSeconds * 1000000) /* In microseconds */);
	}
	else
	{
		sceKernelDelayThread( Asura_Maths::FloatToInt_Chop(fTimeInSeconds * 1000000) /* In microseconds */);
	}
}

Asura_ThreadID Asura_Thread_System::Platform_GetCurrentThreadID()
{ 
	return sceKernelGetThreadId();
}

Asura_ThreadID Asura_Thread_System::Platform_GetInvalidThreadID()
{ 
	return -1;
}


void Asura_Mutex::Platform_Lock()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sceKernelWaitSema(m_iSemaphoreId, 1, NULL);
	Asura_Assert(SCE_KERNEL_ERROR_OK == iRet, "sceKernelWaitSema failed");
}

void Asura_Mutex::Platform_Unlock()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sceKernelSignalSema(m_iSemaphoreId, 1);
	Asura_Assert(SCE_KERNEL_ERROR_OK == iRet, "sceKernelSignalSema failed");
}


void Asura_SyncEvent::Platform_WaitForSignal()
{
	{
#ifdef ASURA_ASSERT
		int iRet = 
#endif
		sceKernelWaitEventFlag(m_iEventId, 1, SCE_KERNEL_EW_OR, NULL, NULL);
		Asura_Assert(SCE_KERNEL_ERROR_OK == iRet, "sceKernelWaitEventFlag failed");
	}
}

void Asura_SyncEvent::Platform_SetSignal()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sceKernelSetEventFlag(m_iEventId, 1);
	Asura_Assert(SCE_KERNEL_ERROR_OK == iRet, "sceKernelSetEventFlag failed");
}

void Asura_SyncEvent::Platform_ClearSignal()
{
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sceKernelClearEventFlag(m_iEventId, 0);
	Asura_Assert(SCE_KERNEL_ERROR_OK == iRet, "sceKernelClearEventFlag failed");
}


//////////////////////////////////////////////////////////////////////////
// Semaphore
//////////////////////////////////////////////////////////////////////////
void Asura_Semaphore::Platform_Wait()
{
}

bool Asura_Semaphore::Platform_TryWait()
{
	return false;
}

bool Asura_Semaphore::Platform_Signal()
{
	return false;
}

