//////////////////////////////////////////////////////////////////////////////
//
// Asura_PSP_InfrastructureMode.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
// Description:
//
//     wrapper for infrastructure mode setup and shutdown.
//     some code is based on sample\utility\inet\game_net\hotspot_unsupport
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK

#ifndef ASURA_SUPPORT_EXTERNAL_NETWORK
#error "inconsistent defines"
#endif

// This file's header.
#include "Network/Asura_PSP_InfrastructureMode.h"

// Asura PSP
#include "Core/Asura_PSP_Thread_Priorities.h"
#include "Debugging/Asura_Assert.h"
#include "SceUtility/Asura_PSP_SCEUtilityWrapper.h"

#include <pspnet.h>
#include <pspnet_error.h>
#include <pspnet_inet.h>
#include <pspnet_apctl.h>
#include <pspnet_resolver.h>
#if (SCE_DEVKIT_VERSION >= 0x03010110)
#include <utility/utility_module.h>
#else //#if (SCE_DEVKIT_VERSION >= 0x03010110)
#include <utility/utility_netmodule.h>
#endif //#else //#if (SCE_DEVKIT_VERSION >= 0x03010110)


// E N U M S /////////////////////////////////////////////////////////////////
#define ASURA_PSP_INFRASTRUCTURE_PSPNET_POOLSIZE		(128 * 1024)
#define ASURA_PSP_APCTL_HANDLER_STACKSIZE (1024 * 3)
#define ASURA_PSP_APCTL_STACKSIZE (SCE_NET_APCTL_LEAST_STACK_SIZE + ASURA_PSP_APCTL_HANDLER_STACKSIZE)
#define ASURA_PSP_APCTL_FLAG_CONNECTED     0x01
#define ASURA_PSP_APCTL_FLAG_DISCONNECTED  0x02

// G L O B A L S /////////////////////////////////////////////////////////////
static bool g_bIsInitialised = false;
static bool g_bIsConnected = false;
static bool g_bIsConnecting = false;

static int g_iApctlHandlerID = -1;
static int g_iApctlFlag;


// I N T E R N A L   F U N C T I O N S ///////////////////////////////////////
static void apctl_handler( int iPrevState, int iNewState, int iEvent, int iError, void* pArg )
{
	(void)iPrevState;
	(void)iError;
	(void)pArg;

	if(iNewState == SCE_NET_APCTL_STATE_IPObtained)
	{
		g_iApctlFlag |= ASURA_PSP_APCTL_FLAG_CONNECTED;
		g_bIsConnected = true;
	}
	else if(iNewState == SCE_NET_APCTL_STATE_Disconnected)
	{
	    if(iEvent == SCE_NET_APCTL_EVENT_DISCONNECT_REQ)
		{
			g_iApctlFlag |= ASURA_PSP_APCTL_FLAG_DISCONNECTED;
		}
		if((g_iApctlFlag & ASURA_PSP_APCTL_FLAG_CONNECTED) &&
		    iEvent == SCE_NET_APCTL_EVENT_ERROR)
		{
			Asura_Assertf(false, ("apctl error happened, error = 0x%x\n", iError) );
		}
		g_iApctlFlag &= ~ASURA_PSP_APCTL_FLAG_CONNECTED;
		g_bIsConnected = false;
	}
}

static void term_apnet()
{
	int iRet;
	
	if (g_iApctlHandlerID >= 0) 
	{
		iRet = sceNetApctlDelHandler(g_iApctlHandlerID);
		Asura_Assertf(iRet >= 0, ("sceNetApctlDelHandler() failed. iRet = 0x%x\n", iRet) );
		g_iApctlHandlerID = -1;
	}

	iRet = sceNetApctlTerm();
	Asura_Assertf(iRet >= 0, ("sceNetApctlTerm() failed. iRet = 0x%x\n", iRet) );
}

static int init_apnet()
{
	int iRet;

	iRet = sceNetApctlInit(ASURA_PSP_APCTL_STACKSIZE, 
		iASURA_THREAD_PRIORITY_INFRASTRUCTUREMODE_APCTL);
	if (iRet < 0) 
	{
		Asura_Assertf(false, ("sceNetApctlInit() failed. iRet = 0x%x\n", iRet) );
		return iRet;
	}
	g_iApctlFlag = 0;

	iRet = sceNetApctlAddHandler(apctl_handler, NULL);
	if (iRet < 0) 
	{
		Asura_Assertf(false, ("sceNetApctlAddHandler() failed. iRet = 0x%x\n", iRet) );
		term_apnet();
		return iRet;
	}
	g_iApctlHandlerID = iRet;

	return iRet;
}

#if 0 // TODO: dynamic load/unload of network modules
static int UnloadNetModules()
{
	int iRet;
	iRet = sceUtilityUnloadNetModule(SCE_UTILITY_NET_MODULE_INET);
	Asura_Assertf(iRet >= 0, ("sceUtilityUnloadNetModule(SCE_UTILITY_NET_MODULE_INET) failed. iRet = 0x%x\n", iRet) );

	iRet = sceUtilityUnloadNetModule(SCE_UTILITY_NET_MODULE_COMMON);
	Asura_Assertf(iRet >= 0, ("sceUtilityUnloadNetModule(SCE_UTILITY_NET_MODULE_COMMON) failed. iRet = 0x%x\n", iRet) );

	return iRet;
}

static int LoadNetModules()
{
	int iRet = 0;

	iRet = sceUtilityLoadNetModule(SCE_UTILITY_NET_MODULE_COMMON);
	if(iRet < 0)
	{
		Asura_Assertf(false, ("sceUtilityLoadNetModule(SCE_UTILITY_NET_MODULE_COMMON) failed. iRet = 0x%x\n", iRet) );
		UnloadNetModules();
		return iRet;
	}

	iRet = sceUtilityLoadNetModule(SCE_UTILITY_NET_MODULE_INET);
	if(iRet < 0)
	{
		Asura_Assertf(false, ("sceUtilityLoadNetModule(SCE_UTILITY_NET_MODULE_INET) failed. iRet = 0x%x\n", iRet) );
		UnloadNetModules();
		return iRet;
	}

	return iRet;
}
#endif // 0


// C L A S S   M E T H O D S /////////////////////////////////////////////////
void Asura_PSP_InfrastructureMode::BeginConnecting()
{
	Asura_Assert(!g_bIsInitialised, "already initialised");
	Asura_Assert(!g_bIsConnected, "already connected");
	Asura_Assert(!g_bIsConnecting, "already connecting");

	// TODO: LoadNetModules();

	if (InitialiseNetworkLibraries() == 0)
	{
		g_bIsInitialised = true;
		g_bIsConnecting = true;
		Asura_PSP_SCEUtilityWrapper::InitialiseInfrastructureMode();
	}
}

void Asura_PSP_InfrastructureMode::UpdateConnecting()
{
	Asura_Assert(g_bIsInitialised, "not initialised");
	Asura_PSP_SCEUtilityWrapper::Update();
}

void Asura_PSP_InfrastructureMode::RenderConnecting()
{
	Asura_Assert(g_bIsInitialised, "not initialised");
	Asura_PSP_SCEUtilityWrapper::BeginFrame();
	Asura_PSP_SCEUtilityWrapper::Render();
	Asura_PSP_SCEUtilityWrapper::CommandState eState = Asura_PSP_SCEUtilityWrapper::EndFrame();
	switch (eState)
	{
		case Asura_PSP_SCEUtilityWrapper::COMMAND_COMPLETED:
		{
			g_bIsConnecting = false;
			break;
		}
		case Asura_PSP_SCEUtilityWrapper::COMMAND_IN_PROGRESS:
		{
			break;
		}
		default:
		{
			g_bIsConnecting = false;
			break;
		}
	}
}

void Asura_PSP_InfrastructureMode::CloseConnection()
{
	Asura_Assert(g_bIsInitialised, "not initialised");

	ShutdownNetworkLibraries();
	// TODO: UnloadNetModules();

	g_bIsInitialised = false;
	g_bIsConnected = false;
	g_bIsConnecting = false;
}

bool Asura_PSP_InfrastructureMode::IsConnected()
{
	return g_bIsConnected;
}

bool Asura_PSP_InfrastructureMode::IsConnecting()
{
	return g_bIsConnecting;
}

int Asura_PSP_InfrastructureMode::InitialiseNetworkLibraries()
{
	sceKernelChangeThreadPriority(SCE_KERNEL_TH_SELF, iASURA_THREAD_PRIORITY_MAIN_THREAD_DURING_ADHOC);

	int iRet = sceNetInit(ASURA_PSP_INFRASTRUCTURE_PSPNET_POOLSIZE, 
		iASURA_THREAD_PRIORITY_ADHOC_CALLOUT, 0, iASURA_THREAD_PRIORITY_NETINTR, 0);
	if (iRet < 0) 
	{
		Asura_Assertf(false, ("sceNetInit() failed. iRet = 0x%x\n", iRet));
		return iRet;
	}

	iRet = sceNetInetInit();
	if (iRet < 0) 
	{
		Asura_Assertf(false, ("sceNetInetInit() failed. iRet = 0x%x\n", iRet) );
		ShutdownNetworkLibraries();
		return iRet;
	}

	iRet = sceNetResolverInit();
	if (iRet < 0) 
	{
		Asura_Assertf(false, ("sceNetResolverInit() failed. iRet = 0x%x\n", iRet) );
		ShutdownNetworkLibraries();
		return iRet;
	}

	iRet = init_apnet();
	if (iRet < 0) 
	{
		ShutdownNetworkLibraries();
		return iRet;
	}

	return 0;
}

void Asura_PSP_InfrastructureMode::ShutdownNetworkLibraries()
{
	int iRet;

	term_apnet();

	iRet = sceNetResolverTerm();
	Asura_Assertf(iRet >= 0, ("sceNetResolverTerm() failed. iRet = 0x%x\n", iRet) );

	iRet = sceNetInetTerm();
	Asura_Assertf(iRet >= 0, ("sceNetInetTerm() failed. iRet = 0x%x\n", iRet) );

	iRet = sceNetTerm();
	Asura_Assertf(iRet >= 0, ("sceNetTerm() failed. iRet = 0x%x\n", iRet) );

	sceKernelChangeThreadPriority(SCE_KERNEL_TH_SELF, iASURA_THREAD_PRIORITY_MAIN_THREAD_NORMAL);
}

#endif // #ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK
