//
// Asura_PSP_Filter.cpp
//
// Created by JG - 01/11/2005 14:31:19
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core\Asura.h"

// This file's header
#include "Rendering\Asura_PSP_Filter.h"

// Asura independent renderer
#include "Rendering\Asura_Render.h"

// PSP headers
#include "Debugging/Asura_PSP_GPIO.h"
#include "Debugging/Asura_PSP_Profiling.h"
#include "Rendering/Asura_PSP_Graphics.h"
#include "Rendering/Asura_PSP_RenderStates.h"
#include "Rendering/Asura_PSP_Stencil_Enums.h"
#include "Rendering/Asura_PSP_VRAMManagement.h"
#include "Rendering/Asura_PSP_StencilBufferManager.h"

// Various parameters for different effects
DEBUGVAR bool dbg_bBlendFilter = true; 

const unsigned int uWidthFilter = 120;
const unsigned int uHeightFilter = 120;
const unsigned int uSizeFilter = uWidthFilter*uHeightFilter*4;
unsigned int uColourPaletteR[256] __attribute__((aligned(16))); 
unsigned int uColourPaletteG[256] __attribute__((aligned(16))); 
unsigned int uColourPaletteB[256] __attribute__((aligned(16))); 

unsigned int* uOutBufferReadFilter = NULL; 
unsigned int* uOutBufferWriteFilter = NULL; 
unsigned int* uFilterBufferReadFilter = NULL; 
unsigned int* uFilterBufferWriteFilter = NULL;

const unsigned int cuOptimalWidth = 64; 

int Asura_PSP_Filter::s_iRed = 85;
int Asura_PSP_Filter::s_iGreen = 92;
int Asura_PSP_Filter::s_iBlue = 159;
// KOR - initialise to -1 to force the palettes to be generated on the first SetFilter
int Asura_PSP_Filter::s_iAlpha = -1;
bool Asura_PSP_Filter::s_bEnabled = false;

void Asura_PSP_Filter::Initialise( void )
{
	SetFilter(false, 0, 0, 0, 0);
}

void Asura_PSP_Filter::SetFilter(bool bEnable, int iRed, int iGreen, int iBlue, int iAlpha)
{
	s_bEnabled = bEnable;
	Asura_Assert(iRed <= 255 && iGreen <= 255 && iBlue <= 255 && iAlpha <= 255, "Colour out of range");
	Asura_Assert(iRed >= 0 && iGreen >= 0 && iBlue >= 0 && iAlpha >= 0, "Colour out of range");
	if (s_bEnabled &&
		(s_iRed != iRed ||
		 s_iGreen != iGreen ||
		 s_iBlue != iBlue ||
		 s_iAlpha != iAlpha))
	{
		for (int i = 0; i < 256; i++)
		{
			// Color remap renderer
			int iScaledRed = Asura_Maths::ClampToRange(Asura_Maths::FloatToInt((i / 255.0f) * iRed), 0, 255); 
			int iScaledGreen = Asura_Maths::ClampToRange(Asura_Maths::FloatToInt((i / 255.0f) * iGreen), 0, 255); 
			int iScaledBlue = Asura_Maths::ClampToRange(Asura_Maths::FloatToInt((i / 255.0f) * iBlue), 0, 255); 
			int iScaledAlpha = Asura_Maths::ClampToRange(Asura_Maths::FloatToInt((i / 255.0f) * iAlpha), 0, 255); 

			// Right now every palette component is identical, but can prime glows/blur per component 
			// if we need to
			uColourPaletteR[i] = iScaledRed | iScaledGreen << 8 | iScaledBlue << 16 | static_cast<u_int>(iScaledAlpha) << 24; 
			uColourPaletteG[i] = iScaledRed | iScaledGreen << 8 | iScaledBlue << 16 | static_cast<u_int>(iScaledAlpha) << 24; 
			uColourPaletteB[i] = iScaledRed | iScaledGreen << 8 | iScaledBlue << 16 | static_cast<u_int>(iScaledAlpha) << 24; 
		}
		s_iRed = iRed;
		s_iGreen = iGreen;
		s_iBlue = iBlue;
		s_iAlpha = iAlpha;
	}
}


void Asura_PSP_Filter::SetSourceScreen( const void* pSource )
{
	SetSourceBuffer( pSource, 512, 512);
}

void Asura_PSP_Filter::SetSourceBuffer( const void* pSource, const unsigned int uSourceWidth, const unsigned int uSourceHeight, bool bRemap, bool bShortIndex )
{
	const unsigned int uSourceWidthPowerOfTwo = Asura_Maths::GetCeilPow2(uSourceWidth); 
	const unsigned int uSourceHeightPowerOfTwo = Asura_Maths::GetCeilPow2(uSourceHeight); 
	if (uSourceWidthPowerOfTwo == uSourceWidth && uSourceHeightPowerOfTwo == uSourceHeight)
	{
		GU_TexImage(0, uSourceWidth, uSourceHeight, uSourceWidth,((char*)pSource));
	}
	else
	{
		GU_TexImage(0, uSourceWidthPowerOfTwo, uSourceHeightPowerOfTwo, uSourceWidth, ((char*)pSource)); 
	}
	const unsigned int uIndexMode = bShortIndex ? SCEGU_PFIDX16 : SCEGU_PFIDX32; 
	GU_TexMode( bRemap ? uIndexMode : SCEGU_PF8888, 0, SCEGU_SINGLE_CLUT, SCEGU_TEXBUF_NORMAL);
	GU_TexWrap(SCEGU_CLAMP, SCEGU_CLAMP);
	GU_TexOffset(0.0f, 0.0f); 
	GU_TexScale(1.0f, 1.0f); 
	GU_TexFlush();
}

void Asura_PSP_Filter::SetDestinationBuffer( void* pDest, const unsigned int uDestWidth, const unsigned int uDestHeight )
{
#ifdef ASURA_PSP_DEFERREDRENDER
	GU_DrawBufferList(SCEGU_PF8888, pDest, uDestWidth);
#else
	GU_DrawBuffer(SCEGU_PF8888, pDest, uDestWidth);
#endif
	GU_Offset(0, 0);
	GU_Viewport(0, 0, uDestWidth, uDestHeight);
	GU_Scissor(0, 0, uDestWidth, uDestHeight);
}

void Asura_PSP_Filter::SetDestinationScreen( void* pDest )
{
#ifdef ASURA_PSP_DEFERREDRENDER
	GU_DrawBufferList(SCEGU_PF8888, pDest, SCEGU_VRAM_WIDTH);
#else
	GU_DrawBuffer(SCEGU_PF8888, pDest, SCEGU_VRAM_WIDTH);
#endif
	GU_Offset(SCEGU_SCR_OFFSETX, SCEGU_SCR_OFFSETY);
	GU_Viewport(2048, 2048, SCEGU_SCR_WIDTH, SCEGU_SCR_HEIGHT);
	GU_Scissor(0, 0, SCEGU_SCR_WIDTH, SCEGU_SCR_HEIGHT);
}

void Asura_PSP_Filter::CopyFromScreen(const void* pSource, void* pDest, const unsigned int uDestWidth, const unsigned int uDestHeight )
{

	SetSourceScreen(pSource);
	SetDestinationBuffer(pDest, uDestWidth, uDestHeight); 
	
	Copy(480, 272, uDestWidth, uDestHeight); 
}

void Asura_PSP_Filter::Copy(const unsigned int uSourceWidth, const unsigned int uSourceHeight, const unsigned int uDestWidth, const unsigned int uDestHeight, const bool bLinearInterpolation, const unsigned int uHorizontalOffset, const unsigned int uVerticalOffset )
{
	// Use several buffer height sprites to speed up copies
	float fBandsHorizontal = uSourceWidth / (float) cuOptimalWidth;

	unsigned int uColumns = static_cast<unsigned int>(fBandsHorizontal); 
	unsigned int uRemainingHorizontal = uSourceWidth - uColumns * cuOptimalWidth; 

	unsigned int uBandWidth = uDestWidth / fBandsHorizontal; 
	unsigned int uLastBandWidth = uDestWidth - uColumns * uBandWidth; 

	if (bLinearInterpolation)
	{
		GU_TexFilter(SCEGU_LINEAR, SCEGU_LINEAR);
	}
	else
	{
		GU_TexFilter(SCEGU_NEAREST, SCEGU_NEAREST); 
	}

	GU_SpriteMode(cuOptimalWidth, uSourceHeight, uBandWidth, uDestHeight);

	for (unsigned int i = 0; i < uColumns; i++)
	{
		GU_DrawSprite( i * uBandWidth - uHorizontalOffset, -uVerticalOffset, 0, i * cuOptimalWidth, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	}

	if (uRemainingHorizontal != 0)
	{
		GU_SpriteMode(uRemainingHorizontal, uSourceHeight, uLastBandWidth, uDestHeight); 
		GU_DrawSprite( uColumns * uBandWidth  - uHorizontalOffset, -uVerticalOffset, 0, uColumns * cuOptimalWidth, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	}
}

void Asura_PSP_Filter::Render()
{
	Render( NULL, Asura_PSP_Graphics::GetCurrentRenderingBuffer() );
}

void Asura_PSP_Filter::Render( const char* acDisplayList, const unsigned int uCurrentRenderingBuffer )
{
	// Are we rendering this effect?
	if (!s_bEnabled) return; 

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_FILTER_START ); 
#endif // #ifdef ASURA_PSP_PROFILING

	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENTITY);

	// Get a couple of pointers to the back buffer, one to read from it, another to write on it
	unsigned int* pReadBackBufferFilter = (unsigned int*)((unsigned int)sceGeEdramGetAddr() + uCurrentRenderingBuffer);
	unsigned int* pWriteBackBufferFilter = (unsigned int*)(uCurrentRenderingBuffer);
	
	// Pointer to intermediate buffers

	// Read and write access to down copy of the back buffer
	uOutBufferWriteFilter	= (unsigned int*)((unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 
	uOutBufferReadFilter	= (unsigned int*)((unsigned int)sceGeEdramGetAddr() + (unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 	

	// Read and write access to the filtered buffer
	uFilterBufferWriteFilter	= (unsigned int*)((unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE + uSizeFilter); 
	uFilterBufferReadFilter	= (unsigned int*)((unsigned int)sceGeEdramGetAddr() + (unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE + uSizeFilter); 	

	// Draw the current frame buffer into our intermediate buffer
	unsigned int uCurrentWidthFilter = uWidthFilter; 
	unsigned int uCurrentHeightFilter = uHeightFilter; 

	// Texture, no depth test and no blending
	GU_Enable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_BLEND);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);
	GU_Disable(SCEGU_ALPHA_TEST);

	// Copy our back buffer down into the intermediate buffer
	CopyFromScreen(pReadBackBufferFilter, (void*)uOutBufferWriteFilter, uCurrentWidthFilter, uCurrentHeightFilter); 




	// Filter the three component using a gradient function
	SetDestinationBuffer( (void*)uFilterBufferWriteFilter, uCurrentWidthFilter, uCurrentHeightFilter); 
	SetSourceBuffer(uOutBufferReadFilter, uCurrentWidthFilter, uCurrentHeightFilter, true); 

	// Combine RGB channels back
	GU_Enable(SCEGU_BLEND);
	GU_BlendFunc(SCEGU_ADD, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, 0x00FFffFF, 0x00000000); 

	// Remap colours through red palette
	GU_ClutMode(SCEGU_PF8888, 0, 0xff, 0); 
	GU_ClutLoad(32, uColourPaletteR);

	Copy(uCurrentWidthFilter, uCurrentHeightFilter, uCurrentWidthFilter, uCurrentHeightFilter);

	GU_BlendFunc(SCEGU_ADD, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, 0x00FFffFF, 0x00FFffFF); 

	// Remap colours through green palette
	GU_ClutMode(SCEGU_PF8888, 8, 0xff, 0); 
	GU_ClutLoad(32, uColourPaletteG);

	Copy(uCurrentWidthFilter, uCurrentHeightFilter, uCurrentWidthFilter, uCurrentHeightFilter);

	// Remap colours through blue palette
	GU_ClutMode(SCEGU_PF8888, 16, 0xff, 0); 
	GU_ClutLoad(32, uColourPaletteB);

	Copy(uCurrentWidthFilter, uCurrentHeightFilter, uCurrentWidthFilter, uCurrentHeightFilter);

	// Draw the full screen pass onto the back buffer
	SetDestinationScreen(pWriteBackBufferFilter); 
	SetSourceBuffer(uFilterBufferReadFilter, uCurrentWidthFilter, uCurrentHeightFilter);  
	
	if (dbg_bBlendFilter)
	{
		GU_Enable(SCEGU_BLEND);
		unsigned int uExtendedAlpha = 0xff000000 | s_iAlpha << 16 | s_iAlpha << 8 | s_iAlpha; 
		unsigned int uOneMinusAlpha = 0xff - s_iAlpha;
		unsigned int uExtendedOneMinusAlpha = 0xff000000 | uOneMinusAlpha << 16 | uOneMinusAlpha << 8 | uOneMinusAlpha; 
		GU_BlendFunc(SCEGU_ADD, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, uExtendedAlpha, uExtendedOneMinusAlpha); 
	}
	else
	{
		GU_Disable(SCEGU_BLEND);
	}


	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Enable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_ALPHA_TEST); 

	Copy( uCurrentWidthFilter, uCurrentHeightFilter, 480, 272 );

	// Go back to previous texture filter 
	GU_TexFilter(SCEGU_LINEAR_MIPMAP_LINEAR, SCEGU_LINEAR);

	// Set textures to repeat by default
	GU_TexWrap(SCEGU_REPEAT, SCEGU_REPEAT); 

	Asura_Render::CurrentStates.ResetAll(); 

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_FILTER_END ); 
#endif // #ifdef ASURA_PSP_PROFILING

}
void Asura_PSP_Filter::StartStencilOutline()
{
	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_OUTLINE, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);
	GU_PixelMask(0x0);
}

void Asura_PSP_Filter::StopStencilOutline()
{
	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_NONE, 0xFF);
}
	
void Asura_PSP_Filter::FillStencilValueWithColor( const unsigned int uStencilValue, const unsigned int uColor )
{
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Disable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_ALPHA_TEST); 
	GU_Disable(SCEGU_BLEND); 

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_EQUAL, uStencilValue, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);
	GU_PixelMask(0x0); 

	Asura_PSP_RenderStates::RequestColourMaterial(0);
	GU_Material(SCEGU_AMBIENT, uColor);
	GU_Material(SCEGU_DIFFUSE, 0xffffffff);
	GU_Material(SCEGU_SPECULAR, 0xffffffff);
	// (MH - 19.12.2006) tell Asura_PSP_Graphics that its cached material type is invalid due to GU_Material calls.
	Asura_PSP_Graphics::InvalidateMaterialType();
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	
	Asura_PSP_Filter::Copy(480, 272, 480, 272, false);

	Asura_Render::CurrentStates.ResetAll(); 
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);
}

void Asura_PSP_Filter::TintStencilValueWithColor( const unsigned int uStencilValue, const unsigned int uColor, const unsigned int uStrength )
{
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Disable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_ALPHA_TEST); 

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_EQUAL, uStencilValue, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);
	GU_PixelMask(0x0); 

	Asura_PSP_RenderStates::RequestColourMaterial(0);
	GU_Material(SCEGU_AMBIENT, uColor);
	GU_Material(SCEGU_DIFFUSE, 0xffffffff);
	GU_Material(SCEGU_SPECULAR, 0xffffffff);

	// Setup blend for tinting
	GU_Enable(SCEGU_BLEND);
	GU_BlendFunc(SCEGU_ADD, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, uStrength, ~uStrength);

	// (MH - 19.12.2006) tell Asura_PSP_Graphics that its cached material type is invalid due to GU_Material calls.
	Asura_PSP_Graphics::InvalidateMaterialType();
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	
	Asura_PSP_Filter::Copy(480, 272, 480, 272, false);

	GU_Disable(SCEGU_BLEND);
	Asura_Render::CurrentStates.ResetAll(); 
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);
}

// (SB - 15/05/07) Render depth buffer through a CLUT with alpha blending masked by given stencil value
void Asura_PSP_Filter::FillStencilValueUsingDepthBuffer(const unsigned int uStencilValue, const unsigned int uShift, const unsigned char uMask, const void* const puClut, const bool bIgnoreStencil /*=false*/)
{
	// Setup rendering states
	GU_Disable(SCEGU_COLOR_TEST);
	GU_Enable(SCEGU_TEXTURE);
	GU_Disable(SCEGU_DEPTH_TEST);
	GU_Disable(SCEGU_ALPHA_TEST);
	GU_PixelMask(0x0);
	GU_Material(SCEGU_AMBIENT, 0xffffffff);
	GU_Material(SCEGU_DIFFUSE, 0xffffffff);
	GU_Material(SCEGU_SPECULAR, 0xffffffff);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(!bIgnoreStencil);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_EQUAL, uStencilValue, 0xff);
	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP);
	Asura_PSP_RenderStates::RequestColourMaterial(0);
	Asura_Render::CurrentStates.RequestZBufferEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);

	// Use depth buffer as a texture
	GU_TexLevelMode(SCEGU_LOD_CONSTANT, 0.0f);
	GU_TexImage(0, 512, 512, SCEGU_VRAM_WIDTH, (void*)((unsigned int)sceGeEdramGetAddr() + (unsigned int)(SCEGU_VRAM_BP32_2) + 0x00600000));

	// Remap colours through depth palette - high order bits used
	GU_ClutMode(SCEGU_PF8888, uShift, uMask, 0); 
	GU_ClutLoad(32, puClut);

	// Setup blend mode
	GU_Enable(SCEGU_BLEND);
	GU_BlendFunc(SCEGU_ADD, SCEGU_ONE_MINUS_SRC_ALPHA, SCEGU_SRC_ALPHA, 0x00, 0x00);

	// Copy with multiple sprites to stay in texture cache
	GU_TexMode(SCEGU_PFIDX16, 0, SCEGU_SINGLE_CLUT, SCEGU_TEXBUF_NORMAL);
	GU_TexFilter(SCEGU_LINEAR_MIPMAP_LINEAR, SCEGU_LINEAR);
	GU_SpriteMode(64, 272, 64, 272); //texture cache friendly
	GU_DrawSprite(  0, 0, 0x0000,   0, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite( 64, 0, 0x0000,  64, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(128, 0, 0x0000, 128, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(192, 0, 0x0000, 192, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(256, 0, 0x0000, 256, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(320, 0, 0x0000, 320, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(384, 0, 0x0000, 384, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(448, 0, 0x0000, 448, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );

	// Reset states
	Asura_PSP_Graphics::InvalidateMaterialType();
	GU_Disable(SCEGU_BLEND);
	GU_TexMode(SCEGU_PFIDX8, 2, SCEGU_SINGLE_CLUT, SCEGU_TEXBUF_FAST);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);
	Asura_Render::CurrentStates.ResetAll(); 
}

void Asura_PSP_Filter::ClipDepthBufferAtDistance(const unsigned int uDistance)
{
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Disable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_ALPHA_TEST);
	GU_PixelMask(0x00ffffff);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);

	Asura_PSP_RenderStates::RequestColourMaterial(0);
	GU_Material(SCEGU_AMBIENT, 0xffffffff);
	GU_Material(SCEGU_DIFFUSE, 0xffffffff);
	GU_Material(SCEGU_SPECULAR, 0xffffffff);
	GU_SpriteMode(64, 272, 64, 272);
	GU_DrawSprite(  0, 0, uDistance,   0, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite( 64, 0, uDistance,  64, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite(128, 0, uDistance, 128, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite(192, 0, uDistance, 192, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite(256, 0, uDistance, 256, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite(320, 0, uDistance, 320, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite(384, 0, uDistance, 384, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);
	GU_DrawSprite(448, 0, uDistance, 448, 0, SCEGU_NOFLIP, SCEGU_NOROTATE);

	GU_PixelMask(0x00000000);
	Asura_Render::CurrentStates.ResetAll();
}

DEBUGVAR unsigned int	dbg_uHorizontalShift = 1; 
DEBUGVAR unsigned int	dbg_uVerticalShift = 1; 
DEBUGVAR bool			dbg_bDiagonalShift = false; 

void Asura_PSP_Filter::RenderOutlineUsingColorBuffer( const char* cDisplayList, const unsigned int uCurrentRenderingBuffer, bool bColourBased /* = false */, unsigned int uOutlineColour /* = 0x00 */ )
{
#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_OUTLINE_START); 
#endif // #ifdef ASURA_PSP_PROFILING

	// color based outline
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Enable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_BLEND); 
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);

	GU_Disable(SCEGU_ALPHA_TEST); 
	GU_AlphaFunc(SCEGU_ALWAYS, 0, 0); 

	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENTITY);
	
	unsigned int* pReadBackBuffer = (unsigned int*)((unsigned int)sceGeEdramGetAddr() + uCurrentRenderingBuffer);
	unsigned int* pWriteBackBuffer = (unsigned int*)(uCurrentRenderingBuffer);

	unsigned int* pReadOutlineBuffer	= (unsigned int*)((unsigned int)sceGeEdramGetAddr() + (unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 
	unsigned int* pWriteOutlineBuffer	= (unsigned int*)((unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 

	Asura_PSP_Filter::SetSourceScreen((void*)pReadBackBuffer); 
	Asura_PSP_Filter::SetDestinationBuffer((void*)pWriteOutlineBuffer, 480, 272); 

	Asura_PSP_RenderStates::RequestTexSlope(0.3f);

	///////////////////////////////////////
	// First pass, horizontal shift
	///////////////////////////////////////

	GU_TexFilter(SCEGU_NEAREST, SCEGU_NEAREST);

	GU_ClearColor(0);
	Asura_PSP_StencilBufferManager::SetStencilBufferClearBits(ASURA_PSP_STENCIL_NONE);
	GU_Clear(SCEGU_CLEAR_COLOR);// | SCEGU_CLEAR_STENCIL);									//

	if (!bColourBased)
	{
		GU_PixelMask(0x000000ff);				// Lose R
	}

	Asura_PSP_Filter::Copy(480, 272, 480, 272, false);	
		
	GU_Enable(SCEGU_BLEND); 
	GU_BlendFunc(SCEGU_ABS, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, 0xffffffff, 0xffffffff);
	Asura_PSP_Filter::Copy(480, 272, 480, 272, false, dbg_uHorizontalShift, 0);
	GU_Disable(SCEGU_BLEND); 

	Asura_PSP_Filter::SetSourceBuffer((void*)pReadOutlineBuffer, 480, 272);
	Asura_PSP_Filter::SetDestinationScreen((void*)pWriteBackBuffer); 

	GU_Enable(SCEGU_COLOR_TEST); 
	GU_Disable(SCEGU_ALPHA_TEST); 
	GU_AlphaFunc(SCEGU_ALWAYS, 0, 0); 
	GU_ColorFunc(SCEGU_NOTEQUAL, 0x0, 0x00ffffff);

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_GREATER, ASURA_PSP_STENCIL_EDGE, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);

	Asura_PSP_Filter::Copy(480, 272, 480, 272, false);

	///////////////////////////////////////
	// Second pass, vertical shift
	///////////////////////////////////////

	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Enable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_BLEND); 
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);

	GU_Disable(SCEGU_ALPHA_TEST); 
	GU_AlphaFunc(SCEGU_ALWAYS, 0, 0); 

	Asura_PSP_Filter::SetSourceScreen((void*)pReadBackBuffer); 
	Asura_PSP_Filter::SetDestinationBuffer((void*)pWriteOutlineBuffer, 480, 272); 

	GU_ClearColor(0);
	Asura_PSP_StencilBufferManager::SetStencilBufferClearBits(ASURA_PSP_STENCIL_NONE);
	GU_Clear(SCEGU_CLEAR_COLOR);// | SCEGU_CLEAR_STENCIL);										//

	if (!bColourBased)
	{
		GU_PixelMask(0x000000ff);				// Lose R
	}

	Asura_PSP_Filter::Copy(480, 272, 480, 272, false);
	
	GU_Enable(SCEGU_BLEND); 
	GU_BlendFunc(SCEGU_ABS, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, 0xffffffff, 0xffffffff);
	Asura_PSP_Filter::Copy(480, 272, 480, 272, false, 0, dbg_uVerticalShift);
	GU_Disable(SCEGU_BLEND); 

	Asura_PSP_Filter::SetSourceBuffer((void*)pReadOutlineBuffer, 480, 272);
	Asura_PSP_Filter::SetDestinationScreen((void*)pWriteBackBuffer); 

	GU_Enable(SCEGU_COLOR_TEST); 
	GU_Disable(SCEGU_ALPHA_TEST); 
	GU_ColorFunc(SCEGU_NOTEQUAL, 0x0, 0x00ffffff);

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_GREATER, ASURA_PSP_STENCIL_EDGE, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);

	Asura_PSP_Filter::Copy(480, 272, 480, 272, false);

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_NONE, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);

	GU_AlphaFunc(SCEGU_GREATER, 0x00, 0xff);
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_TexFilter(SCEGU_LINEAR_MIPMAP_LINEAR, SCEGU_LINEAR); 
	GU_TexFunc(SCEGU_TEX_MODULATE, SCEGU_RGBA);

	GU_ClearDepth(0xffff);
	GU_ClearColor(0x00333333);

	Asura_PSP_Graphics::InvalidateMaterialType();
	Asura_Render::CurrentStates.InvalidateTextureOpMode();
	Asura_Render::CurrentStates.ResetAll();

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_OUTLINE_END); 
#endif // #ifdef ASURA_PSP_PROFILING
}

void Asura_PSP_Filter::RenderOutlineUsingStencilBuffer( const char* cDisplayList, const unsigned int uCurrentRenderingBuffer )
{
#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_OUTLINE_START); 
#endif // #ifdef ASURA_PSP_PROFILING

	// Stencil outline
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Enable(SCEGU_TEXTURE); 
	GU_Disable(SCEGU_DEPTH_TEST); 
	GU_Disable(SCEGU_BLEND); 

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_INCREMENT);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_OUTLINE, 0xFF);
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);
	GU_PixelMask(0x0); 

	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);

	GU_Enable(SCEGU_ALPHA_TEST); 
	GU_AlphaFunc(SCEGU_EQUAL, 0x01, 0xff);

	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENTITY);
	
	unsigned int* pReadBackBuffer = (unsigned int*)((unsigned int)sceGeEdramGetAddr() + uCurrentRenderingBuffer);
	unsigned int* pWriteBackBuffer = (unsigned int*)(uCurrentRenderingBuffer);

	Asura_PSP_Filter::SetSourceScreen((void*)pReadBackBuffer); 
	Asura_PSP_Filter::SetDestinationScreen((void*)pWriteBackBuffer); 

	Asura_PSP_Filter::Copy(480, 272, 480, 272, false, dbg_uHorizontalShift, dbg_uVerticalShift);

	Asura_PSP_Filter::FillStencilValueWithColor(ASURA_PSP_STENCIL_OUTLINE, 0x00000000);

	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_NONE, 0xFF);

	GU_AlphaFunc(SCEGU_GREATER, 0x00, 0xff);
	GU_TexFilter(SCEGU_LINEAR_MIPMAP_LINEAR, SCEGU_LINEAR); 

	Asura_Render::CurrentStates.ResetAll();

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_OUTLINE_END); 
#endif // #ifdef ASURA_PSP_PROFILING
}

//+================================================================================-//
// Remap buffer image to palette, using R channel as clut index						//
//+================================================================================-//

void Asura_PSP_Filter::RemapImageToPalette (const unsigned int uCurrentRenderingBuffer, const void* const puClut)
{
	unsigned int* pReadBackBuffer  = (unsigned int*)((unsigned int)sceGeEdramGetAddr() + uCurrentRenderingBuffer);
	unsigned int* pWriteBackBuffer = (unsigned int*)(uCurrentRenderingBuffer);

	GU_TexImage(0, SCEGU_VRAM_WIDTH, 512, SCEGU_VRAM_WIDTH, pReadBackBuffer);
	Asura_PSP_Filter::SetDestinationScreen((void*)pWriteBackBuffer);

	// Don't remap areas marked as using bilinear filtering
	Asura_PSP_StencilBufferManager::SetStencilEnabled(true);
	Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_NOT_EQUAL, ASURA_PSP_STENCIL_BILINEAR, 0xff);
	Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP);

	// Setup for palette remapping
	GU_TexMode(SCEGU_PFIDX32, 0, SCEGU_SINGLE_CLUT, SCEGU_TEXBUF_NORMAL);

	GU_TexFunc(SCEGU_TEX_REPLACE, SCEGU_RGBA);
	GU_TexScale(1.0f, 1.0f);
	GU_TexWrap(SCEGU_CLAMP, SCEGU_CLAMP);

	GU_TexOffset(0.0f, 0.0f);
	GU_TexLevelMode(SCEGU_LOD_CONSTANT, 0.0f);
	GU_TexFilter(SCEGU_NEAREST, SCEGU_NEAREST);
	GU_Enable(SCEGU_TEXTURE);

	GU_Disable(SCEGU_ALPHA_TEST); 
	GU_Disable(SCEGU_COLOR_TEST); 
	GU_Disable(SCEGU_DEPTH_TEST); 

	GU_PixelMask(0xff000000);

	GU_ClutMode(SCEGU_PF8888, 0, 0xff, 0);				// Use red channel to generate index
	GU_ClutLoad(32, puClut);

	GU_SpriteMode(64, 272, 64, 272); //texture cache friendly
	GU_DrawSprite(  0, 0, -0xffff,   0, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite( 64, 0, -0xffff,  64, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(128, 0, -0xffff, 128, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(192, 0, -0xffff, 192, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(256, 0, -0xffff, 256, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(320, 0, -0xffff, 320, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(384, 0, -0xffff, 384, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );
	GU_DrawSprite(448, 0, -0xffff, 448, 0, SCEGU_NOFLIP, SCEGU_NOROTATE );

	// Restore state
	GU_TexMode(SCEGU_PF8888, 0, SCEGU_SINGLE_CLUT, SCEGU_TEXBUF_NORMAL);
	GU_TexFunc(SCEGU_TEX_MODULATE, SCEGU_RGBA);

	GU_ClearDepth(0xffff);

	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);

	Asura_Render::CurrentStates.ResetAll();
}

void *Asura_PSP_Filter::TakeScreenShotCurrentFrame(int iX, int iY, int iWidth, int iHeight)
{
	unsigned int* uOutBufferWriteBlur = NULL; 
	unsigned int* uFilterBufferWriteBlur = NULL; 
	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENTITY);

	unsigned int* pReadBackBuffer = (unsigned int*)((unsigned int)sceGeEdramGetAddr() + Asura_PSP_Graphics::GetCurrentRenderingBuffer());
	unsigned int* pWriteBackBuffer = (unsigned int*)(Asura_PSP_Graphics::GetCurrentRenderingBuffer());

	uOutBufferWriteBlur	= (unsigned int*)((unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 
	pReadBackBuffer += iY * SCEGU_VRAM_WIDTH + iX;

	return CopyFromAddressToAddress(pReadBackBuffer, uOutBufferWriteBlur, pWriteBackBuffer, iWidth, iHeight);
}

void *Asura_PSP_Filter::TakeScreenShotLastFrame(int iX, int iY, int iWidth, int iHeight)
{
	unsigned int* uOutBufferWriteBlur = NULL; 
	unsigned int* uFilterBufferWriteBlur = NULL; 
	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENTITY);

	unsigned int* pReadBackBuffer = (unsigned int*)((unsigned int)sceGeEdramGetAddr() + Asura_PSP_Graphics::GetCurrentDisplayedBuffer());
	unsigned int* pWriteBackBuffer = (unsigned int*)(Asura_PSP_Graphics::GetCurrentRenderingBuffer());

	uOutBufferWriteBlur	= (unsigned int*)((unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 
	pReadBackBuffer += iY * SCEGU_VRAM_WIDTH + iX;

	return CopyFromAddressToAddress(pReadBackBuffer, uOutBufferWriteBlur, pWriteBackBuffer, iWidth, iHeight);
}

void*Asura_PSP_Filter::CopyFromAddressToAddress(void* pCopyFromAddress, void* pCopyToAddress, void* pRestoreBackBufferAddress, int iWidth, int iHeight)
{
	// Texture, no depth test and no blending
	GU_Enable(SCEGU_TEXTURE); 

	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestAlphaTest(false);

	Asura_PSP_StencilBufferManager::SetStencilEnabled(false);

	Asura_PSP_Graphics::SetUseColourDoubling(false);


	Asura_PSP_Filter::SetDestinationBuffer(pCopyToAddress, 128, 128); 
	Asura_PSP_Filter::SetSourceScreen(pCopyFromAddress); 

	Asura_PSP_Filter::Copy(iWidth, iHeight, 128, 128);
	Asura_PSP_Filter::SetDestinationScreen(pRestoreBackBufferAddress); 

//	Asura_Render::CurrentStates.ResetAll(); 

	static void* puOutBufferReadBlur = (void*)((unsigned int)sceGeEdramGetAddr() + (unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE); 
	return puOutBufferReadBlur;
}