//
// Asura_PSP_MotionBlur.cpp
//
// Created by KJFL - 24 February 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PSP_MotionBlur.h"

// Asura headers
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Rendering/Asura_Render.h"

// PSP headers
#include "Debugging/Asura_PSP_Profiling.h"
#include "Rendering/Asura_PSP_Graphics.h"
#include "Rendering/Asura_PSP_RenderStates.h"

// KJFL - 24 February 2005 - 
// Simple THROUGH mode vertex for full-screen overlay
//  - Order of members determined by PSP GPU
//  - Data sizes determined by THROUGH mode 
struct Asura_ScreenSpace_Overlay_Vertex
{
	u_short			m_usU;
	u_short			m_usV;
	u_int			m_uDiffuse;
	short			m_asPosition[3];
};

// One vertex in the centre of the screen (first one)
// Eight vertices in the corners and edge-midpoints
static Asura_ScreenSpace_Overlay_Vertex axVertices[] = 
{
	// U,	V,		Colour,			X,		Y,		Z
	{ 240,	136,	0x7fffffff,		240,	136,	32767 },
	{ 0,	0,		0xefffffff,		0,		0,		32767 },
	{ 240,	0,		0xefffffff,		240,	0,		32767 },
	{ 480,	0,		0xefffffff,		480,	0,		32767 },
	{ 480,	136,	0xefffffff,		480,	136,	32767 },
	{ 480,	272,	0xefffffff,		480,	272,	32767 },
	{ 240,	272,	0xefffffff,		240,	272,	32767 },
	{ 0,	272,	0xefffffff,		0,		272,	32767 },
	{ 0,	136,	0xefffffff,		0,		136,	32767 },
	{ 0,	0,		0xefffffff,		0,		0,		32767 },
};

bool Asura_PSP_MotionBlur::s_bIsEnabled = false;
float Asura_PSP_MotionBlur::s_fTransparency	= 1.000f;

void Asura_PSP_MotionBlur::Render( void )
{
	if(!s_bIsEnabled || (0.000f == s_fTransparency)) return;

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_MOTIONBLUR_START ); 
#endif // #ifdef ASURA_PSP_PROFILING

	// Magnitude of blur effect comes from game time scale
	// 1.0f			-> Off
	// 1.0f to 0.7f -> Fade in effect round edges of screen, don't affect centre
	// 0.7f to 0.1f -> Effect at max along edges, fade in in centre
	const float fBlurScale = Asura_Timers::GetGameTimeScale();
	if (fBlurScale >= 0.99f) return;

	const float fBlurScaleLowLevel = 0.1f;
	const float fBlurScaleMidLevel = 0.7f;

	u_int uInnerAlpha = 0xffffff;
	u_int uOuterAlpha = 0xffffff;

	const u_int uInnerMaxBlur = Asura_Maths::FloatToInt(0x40 * s_fTransparency);
	const u_int uOuterMaxBlur = Asura_Maths::FloatToInt(0xF0 * s_fTransparency);

	if ( fBlurScale <= fBlurScaleLowLevel )
	{
		uInnerAlpha |= (uInnerMaxBlur << 24);
		uOuterAlpha |= (uOuterMaxBlur << 24);
	}
	else if ( fBlurScale < fBlurScaleMidLevel )
	{
		// find a scale between 0 to 1
		const float fScale = 1.0f - (fBlurScale-fBlurScaleLowLevel) / ( fBlurScaleMidLevel - fBlurScaleLowLevel );
		
		uInnerAlpha |= (Asura_Maths::FloatToInt(fScale*uInnerMaxBlur) << 24);
		uOuterAlpha |= (uOuterMaxBlur << 24);
	}
	else // BlurScale between 1.0f and fBlurScaleMidLevel
	{
		// find a scale between 0 to 1
		const float fScale = 1.0f - (fBlurScale-fBlurScaleMidLevel) / ( 1.0f - fBlurScaleMidLevel);
		
		uInnerAlpha |= (uInnerMaxBlur << 24);
		uOuterAlpha |= (Asura_Maths::FloatToInt(fScale*240.0f) << 24);
	}

	// touch up the vertices with required alpha values
	// centre vert
	axVertices[0].m_uDiffuse = uInnerAlpha;
	// edge verts
	for (u_int u=1; u<10; ++u)
	{
		axVertices[u].m_uDiffuse = uOuterAlpha;
	}

	const unsigned int vram_top = (unsigned int)sceGeEdramGetAddr();
	
	// KJFL - 24 February 2005 - PSPTODO - Lots of state changes to optimise here
	GU_Disable( SCEGU_CULL_FACE );
	Asura_PSP_Graphics::SetUseLighting(false);
	GU_Disable( SCEGU_DEPTH_TEST );
	GU_Enable(SCEGU_TEXTURE);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(1); // No z-write.

	GU_TexLevelMode( SCEGU_LOD_CONSTANT, 0 );
	GU_TexFilter( SCEGU_LINEAR_MIPMAP_LINEAR, SCEGU_LINEAR );

	GU_TexMode( SCEGU_PF8888, 0, SCEGU_SINGLE_CLUT, SCEGU_TEXBUF_NORMAL );
	GU_TexFunc( SCEGU_TEX_MODULATE, SCEGU_RGB );
	GU_TexWrap( SCEGU_CLAMP, SCEGU_CLAMP );
	Asura_PSP_RenderStates::RequestTexScale(  1.f, 1.f );
	Asura_PSP_RenderStates::RequestTexOffset( 0.f, 0.f );

	// use the back buffer (i.e. last frame's render target) as a texture
	GU_TexImage( 0, SCEGU_VRAM_WIDTH, 512, SCEGU_VRAM_WIDTH, (void*)( vram_top + Asura_PSP_Graphics::GetCurrentDisplayedBuffer() ) );
	
	GU_Offset(SCEGU_SCR_OFFSETX, SCEGU_SCR_OFFSETY);
	GU_Viewport(2048, 2048, SCEGU_SCR_WIDTH, SCEGU_SCR_HEIGHT);

	GU_Scissor(0, 0, SCEGU_SCR_WIDTH, SCEGU_SCR_HEIGHT);

	GU_Enable( SCEGU_BLEND );

	GU_BlendFunc( SCEGU_ADD , SCEGU_SRC_ALPHA,SCEGU_ONE_MINUS_SRC_ALPHA,0,0);
	GU_AlphaFunc( SCEGU_ALWAYS,0,0xff );
	
	GU_TexLevelMode( SCEGU_LOD_CONSTANT, 0.0f );

	// draw the overlay as a fan of tris from the centre
	GUM_DrawArray
	(
		SCEGU_PRIM_TRIANGLE_FAN,
		SCEGU_TEXTURE_USHORT|SCEGU_COLOR_PF8888|SCEGU_VERTEX_SHORT|SCEGU_THROUGH,
		10,
		0,
		&axVertices[0]
	);

	GU_Enable( SCEGU_CULL_FACE );
	GU_Enable( SCEGU_DEPTH_TEST );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( 0 );
	GU_TexFunc( SCEGU_TEX_MODULATE, SCEGU_RGBA );

	GU_TexLevelMode(SCEGU_LOD_SLOPE, 0.0f);

	// JG - 01/06/06: Restore the default alpha function, otherwise objects rendered after 
	// this effect don't get alpha stenciled
	GU_AlphaFunc(SCEGU_GREATER, 0x04, 0xff);

	Asura_Render::CurrentStates.ResetAll();

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_MOTIONBLUR_END ); 
#endif // #ifdef ASURA_PSP_PROFILING
}
