//
// Asura_PSP_RenderEnvironment.cpp
//
// Created by KJFL - Thursday, February 05, 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_PSP_RenderEnvironment.h"
#include "Rendering/Asura_PSP_Project_RenderEnvironment.h"

// Asura headers
#include "Animation/Asura_TextureAnimation.h"
#include "Core/Asura_Material_System.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"
#include "Camera/Asura_Camera.h"
#include "Light/Asura_Photon.h"
#include "Mesh/Asura_CollisionMesh.h" // for bounding box
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Module.h"

// PSP headers
#include "Debugging/Asura_PSP_GPIO.h"
#include "Rendering/Asura_PSP_Fog.h"
#include "Rendering/Asura_PSP_Graphics.h"
#include "Rendering/Asura_PSP_RenderStates.h"
#include "Rendering/Asura_PSP_StencilBufferManager.h"
#include "Rendering/Asura_PSP_Stencil_Enums.h"
#include "Rendering/Asura_PSP_TextureManagement.h"
#include "Rendering/Asura_PSP_Water.h"
#include "SpecialFX/Asura_PSP_Skybox.h"
#include "SpecialFX/Asura_Skybox.h"


DEBUGVAR float	dbg_fWaterfallSpeed = 0.03f;
DEBUGVAR bool	dbg_bFastClip = true; 
DEBUGVAR bool	dbg_bMovingWater = false; 
DEBUGVAR u_int	dbg_uBlendWater = 0x00404040; 
DEBUGVAR bool	dbg_bShowClippedPolys = false; 
DEBUGVAR bool	dbg_bExpandClippedPolys = false;	// This was true in GUN to try and prevent bad tears at the expense of some less-obvious artefacts.
DEBUGVAR float	dbg_fExpansionRatio = 1.1f; 
DEBUGVAR bool	dbg_bShowExpandedClippedPolys = false; 
DEBUGVAR bool	dbg_bAllowDualPassPolys = true;
DEBUGVAR bool	dbg_bAllowFirstPassPolys = true;

// Depth buffer offsets...
DEBUGVAR int	dbg_iAlphaDecalDepthBufferOffset	= -10;

#ifdef ASURA_PSP_RENDER_CLIP_OFFSET
DEBUGVAR int	dbg_iClippingDepthBufferOffset		= ASURA_PSP_RENDER_CLIP_OFFSET;
#else
DEBUGVAR int	dbg_iClippingDepthBufferOffset		= +10;
#endif

//#ifndef PSP_NEW_ENV_RENDER
int										Asura_PSP_RenderEnvironment_Zone::s_iDepthBufferOffset	= 0;

/* Value based, not definition based */
#define PG_RENDER_DEBUG 0

#if PG_RENDER_DEBUG
	static bool dbg_bPrintStrips = false;
#endif

#if !defined ASURA_PSP_SONY_SUBMISSION && !defined(ASURA_PSP_UMD) && !defined(ASURA_PSP_TUNER)
	#define ASURA_PSP_RENDER_SELECT 1
#endif

#if defined ASURA_PSP_RENDER_SELECT
	/*
		Initialize to -1 which is nonzero to avoid ending
		up in the BSS, not the DATA segment.
	*/
	static int dbg_iRenderSelect = -1;
	/* Select strips in modules in >= min && < max */
	static int dbg_iModMin = -1;
	static int dbg_iModMax = 999999;
	/* Select strips with index count >= min && < max */
	static int dbg_iIdxMin = -1;
	static int dbg_iIdxMax = 999999;
	/* Select strips with substrisp >= min && < max */
	static int dbg_iSubMin = -1;
	static int dbg_iSubMax = 999999;

	static int dbg_iRenderColour = -1;
	/* Clip indices and substrips above this to 1.0 */
	static int dbg_iIdxClip = 1000;
	static int dbg_iSubClip = 1000;
	/* Minimum percentage saturation and value */
	static int dbg_iSatMin = 30;
	static int dbg_iValMin = 30;

	static u_int RGB_From_HSV(
		const float fH,const float fS,const float fV
	)
	{
		float fR,fG,fB;

		if (fS == 0.0f)
		{
			/* Grey */
			fR = fG = fB = fV;
		}
		else
		{
			/* Scale hue into one of the 6 60deg sectors */
			const float fHS = fH/60.0f;

			const unsigned uSector = Asura_Maths::FloatToInt(fHS);
			const float fOffset = fHS-uSector;

			const float fX = fV * (1.0f - fS);
			const float fY = fV * (1.0f - fS*fOffset);
			const float fZ = fV * (1.0f - fS*(1.0f-fOffset));

			switch (uSector)
			{
				case 0: fR = fV; fG = fZ; fB = fX; break;
				case 1:	fR = fY; fG = fV; fB = fX; break;
				case 2:	fR = fX; fG = fV; fB = fZ; break;
				case 3:	fR = fX; fG = fY; fB = fV; break;
				case 4:	fR = fZ; fG = fX; fB = fV; break;
				case 5:	fR = fV; fG = fX; fB = fY; break;
				default:
					Asura_Assert(uSector >= 6,"Sector must be 0-5");
					fR = fG = fB = 0.0f;
					break;
			}
		}

		{
			const unsigned uR = Asura_Maths::FloatToInt(fR*255.0f);
			const unsigned uG = Asura_Maths::FloatToInt(fG*255.0f);
			const unsigned uB = Asura_Maths::FloatToInt(fB*255.0f);

			/* bits 0-7: R, 8-15: G, 16-23: B, 24-31: alpha */
			return (uR&0xffU) | (uG&0xffU)<<8 | (uB&0xffU)<<16 | (0xffU<<24);
		}
	}

	static unsigned RGB_From_Strip(const Asura_PSP_RenderEnvironment_Strip& xS)
	{
		const unsigned uColourIndex
			= ((unsigned) xS.m_uModule<<5) | (unsigned) xS.m_iMaterialIndex;

		const float fH = ((float) (uColourIndex%99)/99.5f)*360.0f;

		const float fS
			= Asura_Maths::Max(dbg_iSatMin/100.0f,
				Asura_Maths::Min(1.0f,
					(float) xS.m_uNumberOfStrips/(float) dbg_iSubClip));

		const float fV
			= Asura_Maths::Max(dbg_iValMin/100.0f,
				Asura_Maths::Min(1.0f,
					(float) xS.m_uNumberOfIndices/(float) dbg_iIdxClip));

		return RGB_From_HSV(fH,fS,fV);
	}
#endif

#ifndef ASURA_PSP_MAX_CLIPPED_VERTS
#define ASURA_PSP_MAX_CLIPPED_VERTS 1024
#endif

#ifdef ASURA_PSP_DEFERREDRENDER
	static Asura_PSP_RenderEnvironment_VertexClip		g_xClippedVerts[ 2 ][ASURA_PSP_MAX_CLIPPED_VERTS] __attribute__( ( aligned( 64 ) ) );
#else
	static Asura_PSP_RenderEnvironment_VertexClip		g_xClippedVerts[ 1 ][ASURA_PSP_MAX_CLIPPED_VERTS] __attribute__( ( aligned( 64 ) ) );
#endif

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
	#define ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS		256
	static Asura_PSP_RenderEnvironment_VertexClipQuad					g_aaxClippedVertices[ 2 ][ ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ] __attribute__( ( aligned( 64 ) ) );
#endif

static u_int g_uNumClippedVerts = 0;
static u_int g_uMaxClippedVerts = 0;

DEBUGVAR bool dbg_bDontRenderCompleteStripsBeforeClipping = true; 

//#define ASURA_PSP_USE_WATER_EFFECT

// JG - 14/09/06: This method used to work on Gun, but was relying on a particular way of setting matrices. We had a problem with 
// a new project and it has been removed, but I will leave it here for now, investigate further
//#define ASURA_PSP_USE_FAST_DRAW_ARRAY

// KJFL 2004/09/30 - temporary method to only upload textures once per material
//static bool g_bTextureListValid = false;
//static u_int g_aiMaterialsUsed[128];
//static u_int g_uNumberOfMaterialsUsed;


////////////////////////////////////////////////////////////////////////////////
//
// Asura_PSP_RenderEnvironment_Zone
//
////////////////////////////////////////////////////////////////////////////////

Asura_PSP_RenderEnvironment_Zone::RenderStateOverride* Asura_PSP_RenderEnvironment_Zone::s_pfnRenderStateOverride = NULL;
Asura_PSP_RenderEnvironment_Zone::EarlyOutCallback* Asura_PSP_RenderEnvironment_Zone::s_pfnEarlyOutCallback = NULL;

Asura_PSP_RenderEnvironment_Zone::Asura_PSP_RenderEnvironment_Zone (void)
: m_uNumberOfModules(0)
, m_uTotalNumberOfStrips(0)
, m_uTotalNumberOfDualPassStrips(0)
, m_pxModules(NULL)
, m_pxStrips(NULL)
, m_pxModuleLightingCache(NULL)
, m_pucVertexBuffer(NULL)
, m_pusIndexBuffer(NULL)
, m_fCurrentWaterfallPosition(0.0f)
, m_puModuleClipFlags(NULL)
, m_puReflectedModuleClipFlags(NULL)
{
	// Nothing more to initialise...
}


Asura_PSP_RenderEnvironment_Zone::~Asura_PSP_RenderEnvironment_Zone ()
{
	Asura_ResourcePushType(ASURA_RESOURCE_ENVIRONMENT_MESH);
	{
		// KJFL - 2004/02/05 16:20:11 - free up all the used memory

		Asura_ResourceDeallocateArray(m_pxModules);
		Asura_ResourceDeallocateArray(m_pxStrips);
		Asura_ResourceDeallocateArray(m_pxModuleLightingCache);

		Asura_ResourceDeallocateArray(m_pucVertexBuffer);
		Asura_ResourceDeallocateArray(m_pusIndexBuffer);

		Asura_ResourceDeallocateArray(m_puModuleClipFlags);
		Asura_ResourceDeallocateArray(m_puReflectedModuleClipFlags);
	}
	Asura_ResourcePopType();
}

void Asura_PSP_RenderEnvironment_Zone::Reset()
{
	// KOR - 24/03/2006 - ART says this is what it should do:
	Asura_ResourcePushType(ASURA_RESOURCE_ENVIRONMENT_MESH);
	{
		// KJFL - 2004/02/05 16:20:11 - free up all the used memory

		Asura_ResourceDeallocateArray(m_pxModules);
		Asura_ResourceDeallocateArray(m_pxStrips);
		Asura_ResourceDeallocateArray(m_pxModuleLightingCache);

		Asura_ResourceDeallocateArray(m_pucVertexBuffer);
		Asura_ResourceDeallocateArray(m_pusIndexBuffer);

		Asura_ResourceDeallocateArray(m_puModuleClipFlags);
		Asura_ResourceDeallocateArray(m_puReflectedModuleClipFlags);
	}
	Asura_ResourcePopType();
	m_uNumberOfModules = 0;
	m_uTotalNumberOfStrips = 0;
	m_uTotalNumberOfDualPassStrips = 0;
	m_fCurrentWaterfallPosition = 0.f;
}

// KOR - 16/03/2006 - actual size is 4096. Was set to 4095, but still had rare drop-outs, probably due to floating point errors
// Hopefully this is conservative enough to catch everything without too much performance hit.
// JG - 28/04/2006: it seems as is not conservative enough. Some further testing in Gun + some reference code from Core suggest that 
// this value should be around 3800 
DEBUGVAR float dbg_fFudgedGuardbandSize = 3800.f;
DEBUGVAR float dbg_fFudgedScreenRatio = 1.05f;

// Helper functions for guardband clipping
void Asura_PSP_RenderEnvironment_Zone::SetupFrusta()
{
	float fHorizontalFOV				= Asura_Camera::GetHorizontalFieldOfView() * 0.5f;
	float fVerticalFOV					= Asura_Camera::GetVerticalFieldOfView() * 0.5f;
	
	//	const float fHorizontalGuardbandAngle	= Asura_Maths::ArcTangent(4096.f * Asura_Maths::Tangent(fHorizontalFOV)	, 480.f);
	//	const float fVerticalGuardbandAngle		= Asura_Maths::ArcTangent(4096.f * Asura_Maths::Tangent(fVerticalFOV)	, 272.f);
	// KOR - try bringing the guardband value in a bit to be safe
	const float fHorizontalGuardbandAngle	= Asura_Maths::ArcTangent(dbg_fFudgedGuardbandSize * Asura_Maths::Tangent(fHorizontalFOV)	, 480.f);
	const float fVerticalGuardbandAngle		= Asura_Maths::ArcTangent(dbg_fFudgedGuardbandSize * Asura_Maths::Tangent(fVerticalFOV)	, 272.f);

	// JG - 10/07/06: push the screen planes out a little bit to avoid fudging the lambda later on (causing dropouts on large polys)
	fHorizontalFOV *= dbg_fFudgedScreenRatio; 
	fVerticalFOV *= dbg_fFudgedScreenRatio; 

	const Asura_Vector_3& xCameraPosition	= *Asura_Camera::GetPosition();
	Asura_Vector_3 xCameraUp;		Asura_Camera::GetUpVector		(xCameraUp);
	Asura_Vector_3 xCameraRight;	Asura_Camera::GetRightVector	(xCameraRight);
	Asura_Vector_3 xCameraForward;	Asura_Camera::GetForwardVector	(xCameraForward);

	Asura_Quat xRotation;
	Asura_PSP_Plane* pxPlane;
	// Left frustum
	pxPlane = &m_xFrustum.m_xLeft;
	xRotation.FromRotation(-fHorizontalFOV, xCameraUp);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ xCameraUp;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Left guardband
	pxPlane = &m_xGuardband.m_xLeft;
	xRotation.FromRotation(-fHorizontalGuardbandAngle, xCameraUp);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ xCameraUp;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Down frustum
	pxPlane = &m_xFrustum.m_xDown;
	xRotation.FromRotation(-fVerticalFOV, xCameraRight);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ xCameraRight;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Down guardband
	pxPlane = &m_xGuardband.m_xDown;
	xRotation.FromRotation(-fVerticalGuardbandAngle, xCameraRight);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ xCameraRight;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Right frustum
	pxPlane = &m_xFrustum.m_xRight;
	xRotation.FromRotation(fHorizontalFOV, xCameraUp);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ -xCameraUp;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Right guardband
	pxPlane = &m_xGuardband.m_xRight;
	xRotation.FromRotation(fHorizontalGuardbandAngle, xCameraUp);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ -xCameraUp;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Up frustum
	pxPlane = &m_xFrustum.m_xUp;
	xRotation.FromRotation(fVerticalFOV, xCameraRight);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ -xCameraRight;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;
	// Up guardband
	pxPlane = &m_xGuardband.m_xUp;
	xRotation.FromRotation(fVerticalGuardbandAngle, xCameraRight);
	pxPlane->m_xNormal		= xRotation.RotateVector(xCameraForward);
	pxPlane->m_xNormal		= pxPlane->m_xNormal ^ -xCameraRight;
	pxPlane->m_fPosition	= pxPlane->m_xNormal * xCameraPosition;

	// Get the transforms from GUM_
	Asura_PSP_Matrix4 xViewMat;
	Asura_PSP_Matrix4 xProjMat;
	GUM_MatrixMode(SCEGU_MATRIX_VIEW);
	GUM_StoreMatrix(&xViewMat); 
#ifdef ASURA_PSP_USE_FAST_DRAW_ARRAY
	GUM_SetCurMatrix(); 
#endif // ASURA_PSP_USE_FAST_DRAW_ARRAY
	GUM_MatrixMode(SCEGU_MATRIX_PROJECTION);
	GUM_StoreMatrix(&xProjMat); 
#ifdef ASURA_PSP_USE_FAST_DRAW_ARRAY
	GUM_SetCurMatrix(); 
#endif // ASURA_PSP_USE_FAST_DRAW_ARRAY
	GUM_MatrixMode(SCEGU_MATRIX_WORLD);
#ifdef ASURA_PSP_USE_FAST_DRAW_ARRAY
	GUM_SetCurMatrix(); 
#endif // ASURA_PSP_USE_FAST_DRAW_ARRAY

	// Calculate the world to clip space transform
	m_xProjMat.x.x = xViewMat.x.x * xProjMat.x.x + xViewMat.x.y * xProjMat.y.x + xViewMat.x.z * xProjMat.z.x + xViewMat.x.w * xProjMat.w.x;
	m_xProjMat.x.y = xViewMat.x.x * xProjMat.x.y + xViewMat.x.y * xProjMat.y.y + xViewMat.x.z * xProjMat.z.y + xViewMat.x.w * xProjMat.w.y;
	m_xProjMat.x.z = xViewMat.x.x * xProjMat.x.z + xViewMat.x.y * xProjMat.y.z + xViewMat.x.z * xProjMat.z.z + xViewMat.x.w * xProjMat.w.z;
	m_xProjMat.x.w = xViewMat.x.x * xProjMat.x.w + xViewMat.x.y * xProjMat.y.w + xViewMat.x.z * xProjMat.z.w + xViewMat.x.w * xProjMat.w.w;

	m_xProjMat.y.x = xViewMat.y.x * xProjMat.x.x + xViewMat.y.y * xProjMat.y.x + xViewMat.y.z * xProjMat.z.x + xViewMat.y.w * xProjMat.w.x;
	m_xProjMat.y.y = xViewMat.y.x * xProjMat.x.y + xViewMat.y.y * xProjMat.y.y + xViewMat.y.z * xProjMat.z.y + xViewMat.y.w * xProjMat.w.y;
	m_xProjMat.y.z = xViewMat.y.x * xProjMat.x.z + xViewMat.y.y * xProjMat.y.z + xViewMat.y.z * xProjMat.z.z + xViewMat.y.w * xProjMat.w.z;
	m_xProjMat.y.w = xViewMat.y.x * xProjMat.x.w + xViewMat.y.y * xProjMat.y.w + xViewMat.y.z * xProjMat.z.w + xViewMat.y.w * xProjMat.w.w;

	m_xProjMat.z.x = xViewMat.z.x * xProjMat.x.x + xViewMat.z.y * xProjMat.y.x + xViewMat.z.z * xProjMat.z.x + xViewMat.z.w * xProjMat.w.x;
	m_xProjMat.z.y = xViewMat.z.x * xProjMat.x.y + xViewMat.z.y * xProjMat.y.y + xViewMat.z.z * xProjMat.z.y + xViewMat.z.w * xProjMat.w.y;
	m_xProjMat.z.z = xViewMat.z.x * xProjMat.x.z + xViewMat.z.y * xProjMat.y.z + xViewMat.z.z * xProjMat.z.z + xViewMat.z.w * xProjMat.w.z;
	m_xProjMat.z.w = xViewMat.z.x * xProjMat.x.w + xViewMat.z.y * xProjMat.y.w + xViewMat.z.z * xProjMat.z.w + xViewMat.z.w * xProjMat.w.w;

	m_xProjMat.w.x = xViewMat.w.x * xProjMat.x.x + xViewMat.w.y * xProjMat.y.x + xViewMat.w.z * xProjMat.z.x + xViewMat.w.w * xProjMat.w.x;
	m_xProjMat.w.y = xViewMat.w.x * xProjMat.x.y + xViewMat.w.y * xProjMat.y.y + xViewMat.w.z * xProjMat.z.y + xViewMat.w.w * xProjMat.w.y;
	m_xProjMat.w.z = xViewMat.w.x * xProjMat.x.z + xViewMat.w.y * xProjMat.y.z + xViewMat.w.z * xProjMat.z.z + xViewMat.w.w * xProjMat.w.z;
	m_xProjMat.w.w = xViewMat.w.x * xProjMat.x.w + xViewMat.w.y * xProjMat.y.w + xViewMat.w.z * xProjMat.z.w + xViewMat.w.w * xProjMat.w.w;

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
	m_xClipFrustum.m_xLeft.m_xNormal = -m_xFrustum.m_xLeft.m_xNormal;
	m_xClipFrustum.m_xLeft.m_fPosition = m_xFrustum.m_xLeft.m_fPosition;
	m_xClipFrustum.m_xRight.m_xNormal = -m_xFrustum.m_xRight.m_xNormal;
	m_xClipFrustum.m_xRight.m_fPosition = m_xFrustum.m_xRight.m_fPosition;
	m_xClipFrustum.m_xUp.m_xNormal = -m_xFrustum.m_xUp.m_xNormal;
	m_xClipFrustum.m_xUp.m_fPosition = m_xFrustum.m_xUp.m_fPosition;
	m_xClipFrustum.m_xDown.m_xNormal = -m_xFrustum.m_xDown.m_xNormal;
	m_xClipFrustum.m_xDown.m_fPosition = m_xFrustum.m_xDown.m_fPosition;

	m_xAProjMat = Asura_AMatrix4x4(
		m_xProjMat.x.x, m_xProjMat.y.x, m_xProjMat.z.x, m_xProjMat.w.x,
		m_xProjMat.x.y, m_xProjMat.y.y, m_xProjMat.z.y, m_xProjMat.w.y,
		m_xProjMat.x.z, m_xProjMat.y.z, m_xProjMat.z.z, m_xProjMat.w.z,
		m_xProjMat.x.w, m_xProjMat.y.w, m_xProjMat.z.w, m_xProjMat.w.w
	);
#endif
}

static bool BBIsPartiallyAbovePlane(const Asura_Bounding_Box& xBB, const Asura_PSP_Plane& xPlane);
// BBIsCompletelyAboveAnyFrustumPlane returns value from ASURA_PSP_MODULE_CLIP_FLAGS, or 0 if false
static u_int BBIsCompletelyAboveAnyFrustumPlane(const Asura_Bounding_Box& xBB, const Asura_PSP_FrustumLimits& xFrustum);
// GetGuardbandClipFlags returns bitwise OR of relevant ASURA_PSP_MODULE_CLIP_FLAGS.
static u_int GetGuardbandClipFlags(const Asura_Bounding_Box& xBB, const Asura_PSP_FrustumLimits& xGuardband);
// GetGuardbandClipFlags2: given clip flags for containing BB (module) get the clip flags for this sub-BB (strip)
static u_int GetGuardbandClipFlags2(const Asura_Bounding_Box& xBB, const Asura_PSP_FrustumLimits& xGuardband, const u_int uCurFlags);

void Asura_PSP_RenderEnvironment_Zone::SetModuleClipFlags(bool bReflected)
{
	for ( u_int u=0; u<m_uNumberOfModules; ++u )
	{
		u_int* puModuleClipFlags = bReflected ? &m_puReflectedModuleClipFlags[u]: &m_puModuleClipFlags[u];
		const Asura_CollisionMesh* pxCollisionMesh = Asura_Module_System::GetCollisionMeshForModule(u);

		Asura_Bounding_Box xBB = pxCollisionMesh->GetBoundingBox();

		xBB.Displace(Asura_Module_System::GetPositionOfModule(u));

		if(bReflected)
		{
			// Now reflect the bounding box on Y (for water reflection culling)
			const float fTemp = -xBB.MinY; 
			xBB.MinY = -xBB.MaxY; 
			xBB.MaxY = fTemp; 
		}

		// KOR - now use the BB to find out if we need to do guardband clipping on this module
		++m_uNumModuleTests;
		// Reset flags
		*puModuleClipFlags = 0;

		// Check against frustum
		// We only care if it's outside any one of the frustum planes. If any one is true, we can ignore it.
		u_int uClipFlag = BBIsCompletelyAboveAnyFrustumPlane(xBB, m_xFrustum);
		if (uClipFlag)
		{
			*puModuleClipFlags = uClipFlag;
		}
		else
		{
			// Definitely inside frustum - see if we're intersecting guardband
			// We need to know which guardband planes are being intersected, so we can clip against the appropriate ones.
			*puModuleClipFlags |= GetGuardbandClipFlags(xBB, m_xGuardband);
			if(*puModuleClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_ANY)
			{
				++m_uNumModuleAccepts;
			}
		}
	}
}

//#define ASURA_PSP_SHOW_COLLISION_MESH

#ifdef ASURA_PSP_SHOW_COLLISION_MESH
#include <kernel.h>
#include "Mesh/Asura_Mesh.h"
#define ASURA_PSP_NUM_COLLISIONMESH_VERTS 65535
static Asura_PSP_RenderEnvironment_Vertex g_xCollisionMeshVerts[ASURA_PSP_NUM_COLLISIONMESH_VERTS];
static u_int g_uCollisionMeshVertNum = 0;

void FlushEnv()
{
	g_uCollisionMeshVertNum = 0;
}

static inline void SetupNextVert(const Asura_Vector_3& xPos, const Asura_Vector_3& xOffset, u_short usFlags)
{
	Asura_Assert(g_uCollisionMeshVertNum < ASURA_PSP_NUM_COLLISIONMESH_VERTS, "Too many collision mesh verts to render");
	if(g_uCollisionMeshVertNum >= ASURA_PSP_NUM_COLLISIONMESH_VERTS) return;

	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_afPosition[0] = xPos.x + xOffset.x;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_afPosition[1] = xPos.y + xOffset.y;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_afPosition[2] = xPos.z + xOffset.z;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_asNormal[0] = 0;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_asNormal[1] = 0;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_asNormal[2] = 0;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 = 0;
	//g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 = 0xffff;
	//if(usFlags & AsuraPolygonFlag_IgnoreLOS) g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 |= (0x03 << 11); // Blue
	//if(usFlags & AsuraPolygonFlag_DontRender) g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 |= (0x03 << 5); // Green
	//if(usFlags & AsuraPolygonFlag_NoCollisions) g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 |= (0x03); // Red
	//g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 |= (0x03 << 11); // Blue
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 |= (0x03 << 5); // Green
	//g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_uRGB565 |= (0x03); // Red
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_usU = 0;
	g_xCollisionMeshVerts[g_uCollisionMeshVertNum].m_usV = 0;
	++g_uCollisionMeshVertNum;
}

static void RenderCollisionMesh(void)
{
	Asura_PSP_Graphics::SetUseLighting(false);
	Asura_PSP_Graphics::SetNumberOfLights(0);
	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENV);
	Asura_PSP_Graphics::SetUseColourDoubling(true);
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	//Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
	Asura_PSP_TextureManagement::RequestTextureIndex( ASURA_TEXTURE_NONE );

	Asura_PSP_RenderStates::RequestColourMaterial(0);// Use all values below
	GU_Material(SCEGU_AMBIENT, 0x3010ff10);// Multiplied by ambient lighting colour.
	Asura_PSP_RenderStates::RequestAmbient(0xffffffff);
	Asura_Camera::SetCameraIndex(0);

	if (Asura_Camera::IsCameraActive())
	{
		Asura_Render::ApplyCameraSettings();
		const u_int uFirstVertex = g_uCollisionMeshVertNum;

		const u_int uNumModules = Asura_Module_System::GetNumberOfModules();
		for(u_int uModule = 0; uModule < uNumModules; ++uModule)
		{
			if(!Asura_Module_System::IsModuleVisible(uModule)) continue;

			const Asura_CollisionMesh* pxMesh = Asura_Module_System::GetCollisionMeshForModule(uModule);
			const Asura_Vector_3& xOffset = Asura_Module_System::GetPositionOfModule(uModule);

			for(u_int uPoly = 0; uPoly < pxMesh->GetNumberOfPolys(); ++uPoly)
			{
				//const bool bIgnoreLOS = pxMesh->GetFlags(uPoly) & (AsuraPolygonFlag_IgnoreLOS | AsuraPolygonFlag_DontRender);
				const u_short usFlags = pxMesh->GetFlags(uPoly);
				if(usFlags & AsuraPolygonFlag_NoCollisions) continue;
				SetupNextVert(pxMesh->GetVertex(pxMesh->GetVertexIndex(uPoly, 0)), xOffset, usFlags);
				SetupNextVert(pxMesh->GetVertex(pxMesh->GetVertexIndex(uPoly, 1)), xOffset, usFlags);
				SetupNextVert(pxMesh->GetVertex(pxMesh->GetVertexIndex(uPoly, 2)), xOffset, usFlags);
				if(pxMesh->IsQuad(uPoly))
				{
					SetupNextVert(pxMesh->GetVertex(pxMesh->GetVertexIndex(uPoly, 0)), xOffset, usFlags);
					SetupNextVert(pxMesh->GetVertex(pxMesh->GetVertexIndex(uPoly, 2)), xOffset, usFlags);
					SetupNextVert(pxMesh->GetVertex(pxMesh->GetVertexIndex(uPoly, 3)), xOffset, usFlags);
				}
			}
		}
		const u_int uVertexFormat = SCEGU_TEXTURE_USHORT|SCEGU_COLOR_PF5650|SCEGU_NORMAL_SHORT|SCEGU_VERTEX_FLOAT;

		//sceKernelDcacheWritebackRange(&g_xCollisionMeshVerts[uFirstVertex]);
		sceKernelDcacheWritebackAll();
		DrawArray
			(
			SCEGU_PRIM_TRIANGLES,
			uVertexFormat,
			g_uCollisionMeshVertNum - uFirstVertex,
			0,
			&g_xCollisionMeshVerts[uFirstVertex]
			);
	}
	Asura_PSP_RenderStates::RequestColourMaterial(SCEGU_AMBIENT);// Use ambient from vertex colour
	GU_Material(SCEGU_AMBIENT, 0xffffffff);// Multiplied by ambient lighting colour.

	// (MH - 19.12.2006) tell Asura_PSP_Graphics that its cached material type is invalid due to GU_Material calls.
	Asura_PSP_Graphics::InvalidateMaterialType();
}
#else// def ASURA_PSP_SHOW_COLLISION_MESH
void FlushEnv() 
{
	g_uMaxClippedVerts = Asura_Maths::Max(g_uMaxClippedVerts, g_uNumClippedVerts);
	g_uNumClippedVerts = 0; 
}
#endif// else ASURA_PSP_SHOW_COLLISION_MESH

DEBUGVAR bool dbg_bDynamicLights = true;
DEBUGVAR bool dbg_bTexturedEnv = true;
DEBUGVAR u_int dbg_uAmbientColor = 0xffffffff;
DEBUGVAR float dbg_fPSPLightScale = 9.0f;
DEBUGVAR bool dbg_bUseWaterVertexColour = false;

const Asura_Vector_3 *pxCamPos;
Asura_PSP_Plane xCameraFrontPlane;

DEBUGVAR float dbg_fNearTexSlopeEnvironment = 0.3f; 
DEBUGVAR float dbg_fDistantTexSlopeEnvironment = 0.15f; 


void Asura_PSP_RenderEnvironment_Zone::Render(void)
{
	m_bAnyWaterThisFrame = false;

	if (!m_uTotalNumberOfStrips) return;

	Asura_PSP_Fog::TurnOnFogIfNecessary();

	Asura_PSP_Graphics::SetUseLighting(dbg_bDynamicLights);
	Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENV);
	Asura_PSP_Graphics::SetUseColourDoubling(true);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);

	// Multiplied by prelit colour
	Asura_PSP_RenderStates::RequestAmbient(dbg_uAmbientColor);

	if(Asura_Render::GetWireframe() || Asura_PSP_GPIO::GetGPI(ASURA_PSP_GPI_WIREFRAME_ENV))
	{
		m_eRenderMode = SCEGU_PRIM_LINE_STRIP;
	}
	else
	{
		m_eRenderMode = SCEGU_PRIM_TRIANGLE_STRIP;
	}

	m_fCurrentWaterfallPosition += dbg_fWaterfallSpeed * Asura_Timers::GetGameTimeSlice();
	if(m_fCurrentWaterfallPosition > 1.f) m_fCurrentWaterfallPosition -= 1.f;
	Asura_PSP_RenderStates::RequestTexScale(128.f, 128.f);

	// iterate through modules and calculate lighting values for them
	if(dbg_bDynamicLights)
		for ( u_int u=0; u<m_uNumberOfModules; ++u )
		{
			// skip module if it's not visible to any camera
			if(!Asura_Module_System::IsModuleVisible(u, 0x0f)) continue;

			// Calculate bounding box for current module
			const Asura_CollisionMesh* pxCollisionMesh = Asura_Module_System::GetCollisionMeshForModule(u);
			Asura_Bounding_Box xBB = pxCollisionMesh->GetBoundingBox();
			xBB.Displace(Asura_Module_System::GetPositionOfModule(u));
			// Get photons for the bounding box
			Asura_Photon_System::MakePhotonListForBoundingBox( xBB, Asura_Module_System::GetCentreOfModule(u), false, true, 4 );

			// Put information into lighting cache
			u_int uNumberOfPhotons = Asura_Photon_System::NumberOfLocalPhotons;
			for(u_int uPhoton = 0; uPhoton < uNumberOfPhotons; ++uPhoton)
			{
				const Asura_Local_Photon* pxPhoton = &Asura_Photon_System::LocalPhotons[uPhoton];
				const float fR = Asura_Maths::ClampToRange(pxPhoton->GetRedComponent(),		0.f, 1.f);
				const float fG = Asura_Maths::ClampToRange(pxPhoton->GetGreenComponent(),	0.f, 1.f);
				const float fB = Asura_Maths::ClampToRange(pxPhoton->GetBlueComponent(),	0.f, 1.f);
				u_int uPhotonABGR = 0xff000000;
				uPhotonABGR += static_cast<int>(fR * 0xff);
				uPhotonABGR += static_cast<int>(fG * 0xff) << 8;
				uPhotonABGR += static_cast<int>(fB * 0xff) << 16;

				Asura_Vector_3 xV = pxPhoton->GetPosition();
				Asura_PSP_Vector3	vec;
				vec.x = xV.x;
				vec.y = xV.y;
				vec.z = xV.z;

				m_pxModuleLightingCache[u].m_axPhoton[uPhoton].m_fOneOverRange = 1.0f / Asura_Photon_System::LocalPhotons[uPhoton].GetRange();
				m_pxModuleLightingCache[u].m_axPhoton[uPhoton].m_xPhotonPosition = vec;
				m_pxModuleLightingCache[u].m_axPhoton[uPhoton].m_uPhotonColour = uPhotonABGR;
			}
			// KOR - TODO - check if this is really necessary?
			// Fill out the remaining photons with blank info
			for(u_int uPhoton = uNumberOfPhotons; uPhoton < 4; ++uPhoton)
			{
				m_pxModuleLightingCache[u].m_axPhoton[uPhoton].m_uPhotonColour = 0;
			}
		}

		// Set the initial Z-buffer offset...
		s_iDepthBufferOffset	= 0;
		Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);

		// Iterate through cameras
		int i=0;
		// Only one camera on PSP for now
		//for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
		{
			Asura_Camera::SetCameraIndex(i);

			if (Asura_Camera::IsCameraActive())
			{
				Asura_Render::ApplyCameraSettings();

				// Calculate frustum limits
				SetupFrusta();
				SetModuleClipFlags(false);

				// Calculate camera front plane (should be done outside loop)
				pxCamPos = Asura_Camera::GetPosition();
				Asura_Camera::GetForwardVector(xCameraFrontPlane.m_xNormal);
				xCameraFrontPlane.m_fPosition = xCameraFrontPlane.m_xNormal * *pxCamPos;

				// iterate through strips
				if(dbg_bAllowFirstPassPolys)
				{
					for (u_int s=0; s<m_uTotalNumberOfStrips; ++s)
					{
						RenderStrip(s, ASURA_PSP_ENV_STRIP_FIRST_PASS);
					}// end strip loop
				}

				if(dbg_bAllowDualPassPolys)
				{
					for (u_int s=0; s<m_uTotalNumberOfDualPassStrips; ++s)
					{
						RenderStrip(m_uTotalNumberOfStrips + s, ASURA_PSP_ENV_STRIP_SECOND_PASS);
					}// end strip loop
				}

#ifdef ASURA_PSP_USE_WATER_EFFECT
				// Now do water strips
				if(m_bAnyWaterThisFrame)
				{
					// Should pull this out to another function

					// KOR - hack - fiddle with materials
					if(dbg_bUseWaterVertexColour)
					{
						Asura_PSP_RenderStates::RequestColourMaterial(SCEGU_AMBIENT);// Use ambient from vertex colour
						Asura_PSP_RenderStates::RequestAmbient(0x00ffffff | (Asura_PSP_Water::GetWaterOpacity() << 24));
					}
					else
					{
						Asura_PSP_RenderStates::RequestColourMaterial(0);// Use all values below
						GU_Material(SCEGU_AMBIENT, 0x00ffffff | (Asura_PSP_Water::GetWaterOpacity() << 24));// Multiplied by ambient lighting colour.
					}
					GU_Material(SCEGU_DIFFUSE, 0xffffffff);// Multiplied by diffuse lighting colour, then added to previous result
					GU_Material(SCEGU_SPECULAR, 0xffffffff);// Not currently used

					// (MH - 19.12.2006) tell Asura_PSP_Graphics that its cached material type is invalid due to GU_Material calls.
					Asura_PSP_Graphics::InvalidateMaterialType();

					GU_Enable(SCEGU_BLEND);
					// Fix Alpha value
					GU_BlendFunc(SCEGU_ADD, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, dbg_uBlendWater, 0xffffffff - dbg_uBlendWater);

					//GU_PixelMask(0x0000001f);
					GU_PixelMask(0x00000000);
					// Only write if stencil and z-test succeed, otherwise do nothing.
					Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
					// Write to stencil buffer (always, stencil value, full mask)
					Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_WATER, 0xFF);
					Asura_PSP_StencilBufferManager::SetStencilEnabled(true);

					// Don't update the depth buffer, to allow reflection to be seen through
					Asura_Render::CurrentStates.RequestZBufferWriteEnabled( 1 );

					for (u_int s=0; s<m_uTotalNumberOfStrips; ++s)
					{
						RenderStrip(s, ASURA_PSP_ENV_STRIP_WATER);
					}// end water strip loop

					// Update the depth buffer again
					Asura_Render::CurrentStates.RequestZBufferWriteEnabled( 0 );

					// Flip ori
					{
						Asura_Matrix_3x3 xOri = *Asura_Camera::GetOrientation();
						xOri.e21 = -xOri.e21;
						xOri.e22 = -xOri.e22;
						xOri.e23 = -xOri.e23;
						//Asura_Camera::SetOrientation(xOri);
						Asura_Vector_3 xInvertedPos(*Asura_Camera::GetPosition()); 
						xInvertedPos[1] *= -1.0f; 
						Asura_PSP_Graphics::SetupPSPMatFromAsuraData( xOri, xInvertedPos);
					}

					// Mask this pass to draw only on top of water
					Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP);
					Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_EQUAL, ASURA_PSP_STENCIL_WATER, 0xFF);
					Asura_PSP_StencilBufferManager::SetStencilEnabled(true);

					GU_Enable(SCEGU_BLEND);
					GU_AlphaFunc(SCEGU_GREATER, 0x04, 0xff);
					GU_Enable(SCEGU_ALPHA_TEST);
					const char uReflectAlpha = Asura_PSP_Water::GetWaterReflectivity();
					const u_int uReflectBlend = (uReflectAlpha << 16) | (uReflectAlpha << 8) | (uReflectAlpha);
					GU_BlendFunc(SCEGU_ADD, SCEGU_FIX_VALUE, SCEGU_FIX_VALUE, uReflectBlend, 0xffffffff - uReflectBlend);

					Asura_PSP_RenderStates::RequestTexScale(1.f, 1.f);
					Asura_PSP_RenderStates::RequestTexOffset(0.f, 0.f);
					Asura_Render::CurrentStates.RequestZBufferWriteEnabled( 1 ); 
					Asura_PSP_Skybox::Render();
					Asura_Render::CurrentStates.RequestZBufferWriteEnabled( 0 ); 
					Asura_PSP_RenderStates::RequestTexScale(128.f, 128.f);

					GU_Disable(SCEGU_BLEND); 


					// KOR - 17/01/2006 - Reflect mesh pass disabled as it's no longer used by any project
					/*
					SetModuleClipFlags(true);

					Asura_PSP_Graphics::SetUseColourDoubling(true);

					// iterate through strips
					m_uVertexFormat	= uFirstPassVertexFormat;
					for (u_int s=0; s<m_uTotalNumberOfStrips; ++s)
					{
					RenderStrip(s, ASURA_PSP_ENV_STRIP_REFLECTED);
					}// end strip loop

					Asura_PSP_Graphics::SetUseColourDoubling(false);
					*/
					// Flip ori back
					Asura_PSP_Graphics::SetupPSPMatFromAsuraData( *Asura_Camera::GetOrientation(), *Asura_Camera::GetPosition());

					// JG - keep stencil test active to update all other non glowing surfaces
					//GU_Disable(SCEGU_STENCIL_TEST);		
					Asura_PSP_StencilBufferManager::SetStencilOperation(ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_KEEP, ASURA_PSP_STENCIL_OPERATION_REPLACE);
					Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_NONE, 0xFF);

				}// end if any water
#endif // ASURA_PSP_USE_WATER_EFFECT
			}// end if camera is active
		}// end camera loop
		Asura_PSP_RenderStates::RequestTexScale(1.f, 1.f);
		Asura_PSP_RenderStates::RequestTexOffset(0.f, 0.f);
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
		Asura_Render::CurrentStates.RequestAlphaTest(true);
		// We've been screwing with the translucency in the water code, so reset it.
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		Asura_PSP_Fog::TurnOffFogIfNecessary();
		Asura_PSP_RenderStates::RequestTexSlope(dbg_fNearTexSlopeEnvironment); 

		if (m_bAnyWaterThisFrame)
		{
			if (dbg_bMovingWater)
			{
				Asura_PSP_Water::Render(NULL, Asura_PSP_Graphics::GetCurrentRenderingBuffer());
			}
		}

#ifdef ASURA_PSP_SHOW_COLLISION_MESH
		RenderCollisionMesh();
#endif

		Asura_Assert((s_iDepthBufferOffset == 0), "Z-buffer offset has not been correctly reset.");
}

DEBUGVAR bool dbg_bSkipAlphaDecals = false; 

void Asura_PSP_RenderEnvironment_Zone::RenderStrip(u_int uStripIndex, ASURA_PSP_ENV_STRIP_TYPE eType)
{
	Asura_PSP_RenderEnvironment_Strip& xStrip = m_pxStrips[uStripIndex];

	const bool	bSecondPass = (eType == ASURA_PSP_ENV_STRIP_SECOND_PASS);
	const size_t xVertexSize = bSecondPass ? m_xSecondPassVertexSize : m_xFirstPassVertexSize;
	const size_t xVertexPosOffset = bSecondPass ? m_xSecondPassVertexPosOffset : m_xFirstPassVertexPosOffset;

	// skip module if it's not visible
	if(!Asura_Module_System::IsModuleVisible(xStrip.m_uModule)) return;

	const u_int	uMaterialFlags	= Asura_Material_System::GetMaterialFlags(xStrip.m_iMaterialIndex);
	// skip strip if it is flagged as don't render
	if(uMaterialFlags & AsuraMaterialFlag_DontRender) { return; }

	const u_int uProjectMaterialFlags = Asura_Material_System::GetProjectFlags(xStrip.m_iMaterialIndex);

	// skip module if it's outside the frustum (more accurate check than usual modlinking)
	if(m_puModuleClipFlags[xStrip.m_uModule] & ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_ANY) return;
	//if(dbg_bTexturedEnv && !(m_puModuleClipFlags[xStrip.m_uModule] & ASURA_PSP_MODULE_CLIPS_GUARDBAND_ANY))
	//if(dbg_bTexturedEnv && !(m_puModuleClipFlags[xStrip.m_uModule] & ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_ANY))

	// Ignore strips with no polys
	if (xStrip.m_uNumberOfIndices < 3)
	{
		return;
	}

	// (SB - 27/04/07) Early out callback used to abort rendering the current strip
	if (s_pfnEarlyOutCallback)
	{
		if (s_pfnEarlyOutCallback(*this, xStrip) == true) { return; }
	}

	// Check the bigger bounding box to discard the whole lot of internal strips
	if(!BBIsPartiallyAbovePlane(xStrip.m_xBB, xCameraFrontPlane))
	{
#ifdef PSP_TEST_GUARDBAND_CLIPPING
		++m_uNumStripBehinds;
#endif
		return;
	}

	// Check to see if strip is outside frustum
	if (BBIsCompletelyAboveAnyFrustumPlane(xStrip.m_xBB, m_xFrustum)) return;
	//	if(BBIsCompletelyAbovePlane(xBB, m_xFrustum.m_xLeft))		return;
	//	if(BBIsCompletelyAbovePlane(xBB, m_xFrustum.m_xRight))		return;
	//	if(BBIsCompletelyAbovePlane(xBB, m_xFrustum.m_xDown))		return;
	//	if(BBIsCompletelyAbovePlane(xBB, m_xFrustum.m_xUp))		return;

	const u_int	uTextureIndex	= Asura_Material_System::GetTextureID(xStrip.m_iMaterialIndex, (bSecondPass) ? 1 : 0);

#ifdef ASURA_PSP_USE_WATER_EFFECT
	const u_int	uProjectFlags	= Asura_Material_System::GetProjectFlags(xStrip.m_iMaterialIndex);
	switch(eType)
	{
	case ASURA_PSP_ENV_STRIP_FIRST_PASS:
	case ASURA_PSP_ENV_STRIP_SECOND_PASS:
		{
			// skip strip if it's water - these are all done at the end
			if( uMaterialFlags & AsuraMaterialFlag_FlowingWater)
			{
				m_bAnyWaterThisFrame = true;
				return;
			}
			// skip strip if is flag as only reflected - these are all done after main geometry, before water
			if( Asura_PSP_RenderEnvironment::Project_IsStripReflectedOnly(uProjectFlags) ) { return; }
		}
		break;
	case ASURA_PSP_ENV_STRIP_REFLECTED:
		{
			// skip strip if it's water - these are all done at the end
			if( uMaterialFlags & AsuraMaterialFlag_FlowingWater) { return; }
			// only render strips if they are flag as reflected in any sense
			if ( !Asura_PSP_RenderEnvironment::Project_IsStripReflected(uProjectFlags) ) { return; }
		}
		break;
	case ASURA_PSP_ENV_STRIP_WATER:
		{
			// only rendering water
			if( !(uMaterialFlags & AsuraMaterialFlag_FlowingWater)) { return; }
		}
		break;
	}
#endif // ASURA_PSP_USE_WATER_EFFECT

	if(dbg_bTexturedEnv)
	{
		Asura_PSP_TextureManagement::RequestTextureIndex( uTextureIndex );
	}
	else
	{
		Asura_PSP_TextureManagement::RequestTextureIndex( ASURA_TEXTURE_NONE );
	}

	// Get offset to this strip's vertex buffer
#if defined ASURA_PSP_RENDER_SELECT
	const unsigned char* pucVertices;
	const u_short* pxIndices;
	pxIndices = 0;
	pucVertices = GetVertexPtr(xStrip, 0);
#endif // #if defined ASURA_PSP_RENDER_SELECT

	// Fill out lighting from cache
	u_int uPhoton;
	for(uPhoton = 0; uPhoton < 4; ++uPhoton)
	{
		const Asura_PSP_Render_ModuleLightingCache_Photon& xPhoton = m_pxModuleLightingCache[xStrip.m_uModule].m_axPhoton[uPhoton];

		if (!xPhoton.m_uPhotonColour) break;

		GU_LightColor(uPhoton, SCEGU_DIFFUSE, xPhoton.m_uPhotonColour);
		GU_Light(uPhoton, SCEGU_LIGHT_POINT, SCEGU_DIFFUSE, &xPhoton.m_xPhotonPosition);
		GU_LightAtt(uPhoton, 0.0f, 0.0f, dbg_fPSPLightScale * xPhoton.m_fOneOverRange * xPhoton.m_fOneOverRange);
	}		
	Asura_PSP_Graphics::SetNumberOfLights(uPhoton);

#ifdef ASURA_PSP_USE_WATER_EFFECT
	if(eType != ASURA_PSP_ENV_STRIP_WATER)
#endif // ASURA_PSP_USE_WATER_EFFECT
	{
		Asura_TextureAnimationOutput xTextureAnim;
		if(Asura_Material_System::GetTextureAnimationState(xStrip.m_iMaterialIndex,0,&xTextureAnim))
		{
			Asura_PSP_RenderStates::RequestTexOffset(xTextureAnim.m_fUOffset, xTextureAnim.m_fVOffset);
		}
		else if ( uMaterialFlags & AsuraMaterialFlag_Waterfall )
		{
			Asura_PSP_RenderStates::RequestTexOffset(m_fCurrentWaterfallPosition, 0.f);
		}
		else
		{
			Asura_PSP_RenderStates::RequestTexOffset(0.f, 0.f);
		}

		// JG - 15/06/06: Hack for mipmapping problems, reusing AsuraTextureFlag_PS2DistantTexture to flag problematic textures
		Asura_PSP_RenderStates::RequestTexSlope(
			( Asura_Texture_Management::GetTextureFlags( uTextureIndex ) & AsuraTextureFlag_PS2DistantTexture ) ? dbg_fDistantTexSlopeEnvironment : dbg_fNearTexSlopeEnvironment );

		// KOR - this should be the correct behaviour but isn't quite working with the clouds in the MV dynamic map
		// Investigate further
		//	if(( uMaterialFlags & AsuraMaterialFlag_AlphaDecal ) || (Asura_Texture_Management::GetTextureFlags(uTextureIndex) & AsuraTextureFlag_8BitAlpha) )
		if( uMaterialFlags & AsuraMaterialFlag_AlphaDecal )
		{
			if (dbg_bSkipAlphaDecals)
				return; 

			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
			Asura_Render::CurrentStates.RequestAlphaTest(true);
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

			s_iDepthBufferOffset += dbg_iAlphaDecalDepthBufferOffset;
			Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);
		}
		else if ( uMaterialFlags & AsuraMaterialFlag_Transparent )
		{
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
			Asura_Render::CurrentStates.RequestAlphaTest(true);
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		}
		else if( uMaterialFlags & AsuraMaterialFlag_PS2AdditiveAlpha)
		{
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
			Asura_Render::CurrentStates.RequestAlphaTest(true);
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
		}
		else if ( uMaterialFlags & AsuraMaterialFlag_SelfIlluminate)
		{
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
			Asura_Render::CurrentStates.RequestAlphaTest( false );
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		}
		else if (bSecondPass)
		{
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
			Asura_Render::CurrentStates.RequestAlphaTest( true );
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
		}
		else
		{
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
			Asura_Render::CurrentStates.RequestAlphaTest( false );
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		}
	}

	// (SB - 27/02/07) Allow projects to override render state via this callback
	if (s_pfnRenderStateOverride != NULL)
	{
		s_pfnRenderStateOverride(*this, xStrip);
	}

	// RJT NOTE: These can be changed by the override so need to be set per-strip
	Asura_PSP_Graphics::SetMaterialType( ASURA_PSP_GRAPHICS_MATERIALTYPE_ENV );
	Asura_PSP_RenderStates::RequestAmbient( dbg_uAmbientColor );

	// Centralised state override
	Asura_PSP_Graphics::DoStateOverride( uProjectMaterialFlags );

#ifdef ASURA_PSP_USE_WATER_EFFECT
	if(eType == ASURA_PSP_ENV_STRIP_REFLECTED)
	{
		Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW);
	}
#endif // ASURA_PSP_USE_WATER_EFFECT

#if PG_RENDER_DEBUG
	if (dbg_bPrintStrips)
	{
		sceKernelPrintf(" %3u strip: 0x%08x %d %d\n",
			xStrip.m_uModule,(long unsigned) &xStrip,
			xStrip.m_uNumberOfIndices,xStrip.m_uNumberOfStrips
			);		
	}
#endif

#if defined ASURA_PSP_RENDER_SELECT
	if (dbg_iRenderSelect > 0
		&& ((int) xStrip.m_uModule >= dbg_iModMin
		&& xStrip.m_uModule < dbg_iModMax)
		&& ((int) xStrip.m_uNumberOfIndices >= dbg_iIdxMin
		&& xStrip.m_uNumberOfIndices < dbg_iIdxMax)
		&& ((int) xStrip.m_uNumberOfStrips >= dbg_iSubMin
		&& xStrip.m_uNumberOfStrips < dbg_iSubMax))
	{
		if (dbg_iRenderColour > 0)
		{
			const unsigned uColour = RGB_From_Strip(xStrip);

			Asura_PSP_RenderStates::RequestColourMaterial(0);
			GU_Material(SCEGU_AMBIENT,uColour);
			Asura_PSP_RenderStates::RequestAmbient(0xffffffff);
		}

		DrawArray
			(
			(m_eRenderMode == SCEGU_PRIM_TRIANGLE_STRIP) ? SCEGU_PRIM_LINE_STRIP : SCEGU_PRIM_TRIANGLE_STRIP,
			xStrip.m_usVertexFormat,
			xStrip.m_uNumberOfIndices,
			(void*)pxIndices,
			(void*)pucVertices
			);

		if (dbg_iRenderColour > 0)
		{
			Asura_PSP_RenderStates::RequestColourMaterial(SCEGU_AMBIENT);
			GU_Material(SCEGU_AMBIENT,0xffffffff);
			// (MH - 19.12.2006) tell Asura_PSP_Graphics that its cached material type is invalid due to GU_Material calls.
			Asura_PSP_Graphics::InvalidateMaterialType();
		}
	}
	else
#endif
	{
		if (!dbg_bDontRenderCompleteStripsBeforeClipping)
		{
			DrawStripComplete(xStrip, xVertexSize);
		}
	}

	// Perform clipping on strip if necessary
	if( m_puModuleClipFlags[xStrip.m_uModule] & ASURA_PSP_MODULE_CLIPS_GUARDBAND_ANY)
	{
		if (dbg_bFastClip)
		{
			ClipStripSplitThroughDegenerates(xStrip,
				(eType == ASURA_PSP_ENV_STRIP_REFLECTED),
				(eType == ASURA_PSP_ENV_STRIP_SECOND_PASS),
				xVertexSize,
				xVertexPosOffset);
		}
		else
		{
			ClipStrip(xStrip, (eType == ASURA_PSP_ENV_STRIP_SECOND_PASS));
		}
	}
	else
	{
		// Render the whole strip without clipping
		if (dbg_bDontRenderCompleteStripsBeforeClipping)
		{
			DrawStripComplete(xStrip, xVertexSize); 
		}
	}

	if ( uMaterialFlags & AsuraMaterialFlag_SelfIlluminate )
	{
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
		Asura_PSP_Graphics::SetNumberOfLights(0);
		Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENTITY);

		// Centralised state override
		Asura_PSP_Graphics::DoStateOverride( uProjectMaterialFlags );

#if PG_RENDER_DEBUG
		if (dbg_bPrintStrips)
		{
			sceKernelPrintf(" %3u strip: 0x%08x %d %d\n",
				xStrip.m_uModule,(long unsigned) &xStrip,
				xStrip.m_uNumberOfIndices,xStrip.m_uNumberOfStrips
				);			
		}
#endif

#if defined ASURA_PSP_RENDER_SELECT
		if (dbg_iRenderSelect > 1
			&& ((int) xStrip.m_uModule >= dbg_iModMin
			&& xStrip.m_uModule < dbg_iModMax)
			&& ((int) xStrip.m_uNumberOfIndices >= dbg_iIdxMin
			&& xStrip.m_uNumberOfIndices < dbg_iIdxMax)
			&& ((int) xStrip.m_uNumberOfStrips >= dbg_iSubMin
			&& xStrip.m_uNumberOfStrips < dbg_iSubMax))
		{
			DrawArray
				(
				(m_eRenderMode == SCEGU_PRIM_TRIANGLE_STRIP)
				? SCEGU_PRIM_LINE_STRIP  : SCEGU_PRIM_TRIANGLE_STRIP,
				xStrip.m_usVertexFormat,
				xStrip.m_uNumberOfIndices,
				(void*)pxIndices,
				(void*)pucVertices
				);
		}
		else
#endif
		{
			if (!dbg_bDontRenderCompleteStripsBeforeClipping)
			{
				DrawStripComplete(xStrip, xVertexSize); 
			}
		}

		// Perform clipping on strip if necessary
		if(m_puModuleClipFlags[xStrip.m_uModule] & ASURA_PSP_MODULE_CLIPS_GUARDBAND_ANY)
		{
			if (dbg_bFastClip)
			{
				ClipStripSplitThroughDegenerates(xStrip,
					(eType == ASURA_PSP_ENV_STRIP_REFLECTED),
					(eType == ASURA_PSP_ENV_STRIP_SECOND_PASS),
					xVertexSize,
					xVertexPosOffset);
			}
			else
			{
				ClipStrip(xStrip, (eType == ASURA_PSP_ENV_STRIP_SECOND_PASS));
			}
		}
		else
		{
			// Render the whole strip without clipping
			if (dbg_bDontRenderCompleteStripsBeforeClipping)
			{
				DrawStripComplete(xStrip, xVertexSize); 
			}
		}

		Asura_PSP_Graphics::SetMaterialType(ASURA_PSP_GRAPHICS_MATERIALTYPE_ENV);

#if defined ASURA_RENDER_TTYPES
		Asura_Render::AddToTrianglesRendered(xStrip.m_uNumberOfIndices-2,1,
			ASURA_RENDER_TTYPE_SELF);
#else
		Asura_Render::AddToTrianglesRendered(xStrip.m_uNumberOfIndices-2);
#endif
		Asura_Render::AddToPatchesRendered(1);			
	}

#ifdef ASURA_PSP_USE_WATER_EFFECT
	if(eType == ASURA_PSP_ENV_STRIP_REFLECTED)
	{
		Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);
	}
#endif // ASURA_PSP_USE_WATER_EFFECT

#if defined ASURA_RENDER_TTYPES
	{
		const int iNumTotal = xStrip.m_uNumberOfIndices-2;
		const int iNumGlue = (xStrip.m_uNumberOfStrips-1)*4;
		Asura_Render::AddToTrianglesRendered(iNumTotal-iNumGlue,1,ASURA_RENDER_TTYPE_ENV);
		Asura_Render::AddToTrianglesRendered(iNumGlue,1,ASURA_RENDER_TTYPE_ENVGLUE);
	}
#else
	Asura_Render::AddToTrianglesRendered(xStrip.m_uNumberOfIndices-2);
#endif
	Asura_Render::AddToPatchesRendered(1);

	if((eType != ASURA_PSP_ENV_STRIP_WATER) && (uMaterialFlags & AsuraMaterialFlag_AlphaDecal))
	{
		s_iDepthBufferOffset -= dbg_iAlphaDecalDepthBufferOffset;
		Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);
	}
}

DEBUGVAR float dbg_fLambdaFudge = 1.0f;
DEBUGVAR bool dbg_bNewColourInterpolation = true; 

// KOR - taken and modified from decal system
u_int Asura_PSP_RenderEnvironment_Zone::ClipPolyToPlane(const Asura_PSP_RenderEnvironment_VertexClip* pxInVertices, u_int& uNumberOfVertices, Asura_PSP_RenderEnvironment_VertexClip* pxOutVertices, const Asura_PSP_Plane& xPlane)
{
	const u_int		uRedMask		= 0x000000FF;
	const u_int		uGreenMask		= 0x0000FF00;
	const u_int		uBlueMask		= 0x00FF0000;
	const u_int		uAlphaMask		= 0xFF000000;
	const u_int		uRedShift		= 0;
	const u_int		uGreenShift		= 8;
	const u_int		uBlueShift		= 16;
	const u_int		uAlphaShift		= 24;

	Asura_Assert(uNumberOfVertices >= 3, "Not enough input verts");
	u_int uNewVertexCount = 0;

	float fDot[9];

	bool bAllNegative = true;
	bool bAllPositive = true;
	for (u_int u=0; u<uNumberOfVertices; u++)
	{
		Asura_Vector_3 xPos(pxInVertices[u].m_afPosition[0], pxInVertices[u].m_afPosition[1], pxInVertices[u].m_afPosition[2]);
		fDot[u] = xPos * xPlane.m_xNormal + xPlane.m_fPosition;
		if (fDot[u]>=0.0f)
		{
			bAllNegative = false;
		}
		else
		{
			bAllPositive = false;
		}
	}

	/// Kevin Lea - 14/11/2002 10:49:43 - early exit if all vertices would be clipped
	if (bAllNegative)
	{
		uNumberOfVertices = 0;
		return false;
	}

	for (u_int u=0; u<uNumberOfVertices; u++)
	{
		u_int v = (u!=0) ? u-1 : uNumberOfVertices-1; // previous vertex

		Asura_Vector_3 xPosU(pxInVertices[u].m_afPosition[0], pxInVertices[u].m_afPosition[1], pxInVertices[u].m_afPosition[2]);
		Asura_Vector_3 xPosV(pxInVertices[v].m_afPosition[0], pxInVertices[v].m_afPosition[1], pxInVertices[v].m_afPosition[2]);

		// Transform this vertices to screen space
		float fProjUW = xPosU.GetX() * m_xProjMat.x.w + xPosU.GetY() * m_xProjMat.y.w + xPosU.GetZ() * m_xProjMat.z.w + m_xProjMat.w.w + 0.001f;;
		Asura_Vector_3 xProjU;
		xProjU.SetX(240.f * (xPosU.GetX() * m_xProjMat.x.x + xPosU.GetY() * m_xProjMat.y.x + xPosU.GetZ() * m_xProjMat.z.x + m_xProjMat.w.x) / fProjUW); 
		xProjU.SetY(136.f * (xPosU.GetX() * m_xProjMat.x.y + xPosU.GetY() * m_xProjMat.y.y + xPosU.GetZ() * m_xProjMat.z.y + m_xProjMat.w.y) / fProjUW); 
		xProjU.SetZ((xPosU.GetX() * m_xProjMat.x.z + xPosU.GetY() * m_xProjMat.y.z + xPosU.GetZ() * m_xProjMat.z.z + m_xProjMat.w.z) / fProjUW); 
		
		float fProjVW = xPosV.GetX() * m_xProjMat.x.w + xPosV.GetY() * m_xProjMat.y.w + xPosV.GetZ() * m_xProjMat.z.w + m_xProjMat.w.w + 0.001f;
		Asura_Vector_3 xProjV;
		xProjV.SetX(240.f * (xPosV.GetX() * m_xProjMat.x.x + xPosV.GetY() * m_xProjMat.y.x + xPosV.GetZ() * m_xProjMat.z.x + m_xProjMat.w.x) / fProjVW); 
		xProjV.SetY(136.f * (xPosV.GetX() * m_xProjMat.x.y + xPosV.GetY() * m_xProjMat.y.y + xPosV.GetZ() * m_xProjMat.z.y + m_xProjMat.w.y) / fProjVW); 
		xProjV.SetZ((xPosV.GetX() * m_xProjMat.x.z + xPosV.GetY() * m_xProjMat.y.z + xPosV.GetZ() * m_xProjMat.z.z + m_xProjMat.w.z) / fProjVW); 

		if (fDot[u]<0.0f)
		{
			if (fDot[v]>=0.0f)
			{
				// Current vertex is on negative side of plane,
				// but previous vertex is on positive side.
				Asura_Vector_3 xEdge = xPosV - xPosU;
				float f = xEdge * xPlane.m_xNormal;

				Asura_Assert( f!=0.0f, "edge parallel to plane");

				float fLambda = Asura_Maths::ClampToRange(fDot[v] / f  * dbg_fLambdaFudge, 0.f, 1.f);
				float fOneMinusLambda = 1.f - fLambda;

				Asura_Vector_3 xOutPos = xPosV - xEdge * fLambda;

				float fOutProjW = xOutPos.GetX() * m_xProjMat.x.w + xOutPos.GetY() * m_xProjMat.y.w + xOutPos.GetZ() * m_xProjMat.z.w + m_xProjMat.w.w + 0.001f;
				Asura_Vector_3 xOutProj;
				xOutProj.SetX(240.f * (xOutPos.GetX() * m_xProjMat.x.x + xOutPos.GetY() * m_xProjMat.y.x + xOutPos.GetZ() * m_xProjMat.z.x + m_xProjMat.w.x) / fOutProjW); 
				xOutProj.SetY(136.f * (xOutPos.GetX() * m_xProjMat.x.y + xOutPos.GetY() * m_xProjMat.y.y + xOutPos.GetZ() * m_xProjMat.z.y + m_xProjMat.w.y) / fOutProjW); 
				xOutProj.SetZ((xOutPos.GetX() * m_xProjMat.x.z + xOutPos.GetY() * m_xProjMat.y.z + xOutPos.GetZ() * m_xProjMat.z.z + m_xProjMat.w.z) / fOutProjW); 

#ifndef ASURA_PSP_NO_CLIPPER_COLOUR_INTERPOLATION
				float fHorRange = xProjV.GetX() - xProjU.GetX(); 
				float fVerRange = xProjV.GetY() - xProjU.GetY(); 
				float fLambdaXY = 0.0f;

				if (Asura_Maths::Abs(fHorRange) > Asura_Maths::Abs(fVerRange))
				{
					fLambdaXY = Asura_Maths::ClampToRange((xOutProj.GetX() - xProjU.GetX()) / fHorRange, 0.f, 1.f); 
				}
				else
				{
					fLambdaXY = Asura_Maths::ClampToRange((xOutProj.GetY() - xProjU.GetY()) / fVerRange, 0.f, 1.f); 
				}
#endif

				//Asura_Assert(xOutPos.x > -100000.f && xOutPos.y > -100000.f && xOutPos.z > -100000.f, "Position very negative");
				pxOutVertices[uNewVertexCount].m_afPosition[0] = xOutPos.x;
				pxOutVertices[uNewVertexCount].m_afPosition[1] = xOutPos.y;
				pxOutVertices[uNewVertexCount].m_afPosition[2] = xOutPos.z;

				pxOutVertices[uNewVertexCount].m_usU = static_cast<u_short>(pxInVertices[v].m_usU * fOneMinusLambda + pxInVertices[u].m_usU * fLambda);
				pxOutVertices[uNewVertexCount].m_usV = static_cast<u_short>(pxInVertices[v].m_usV * fOneMinusLambda + pxInVertices[u].m_usV * fLambda);

				pxOutVertices[uNewVertexCount].m_asNormal[0] = static_cast<u_short>(pxInVertices[v].m_asNormal[0] * fOneMinusLambda + pxInVertices[u].m_asNormal[0] * fLambda);
				pxOutVertices[uNewVertexCount].m_asNormal[1] = static_cast<u_short>(pxInVertices[v].m_asNormal[1] * fOneMinusLambda + pxInVertices[u].m_asNormal[1] * fLambda);
				pxOutVertices[uNewVertexCount].m_asNormal[2] = static_cast<u_short>(pxInVertices[v].m_asNormal[2] * fOneMinusLambda + pxInVertices[u].m_asNormal[2] * fLambda);

#ifndef ASURA_PSP_NO_CLIPPER_COLOUR_INTERPOLATION
				const int		iAlphaU	= (pxInVertices[u].m_iRGBA8888 & uAlphaMask) >> uAlphaShift;
				const int		iBlueU	= (pxInVertices[u].m_iRGBA8888 & uBlueMask)  >> uBlueShift;
				const int		iGreenU	= (pxInVertices[u].m_iRGBA8888 & uGreenMask) >> uGreenShift;
				const int		iRedU	= (pxInVertices[u].m_iRGBA8888 & uRedMask)   >> uRedShift;

				const int		iAlphaV	= (pxInVertices[v].m_iRGBA8888 & uAlphaMask) >> uAlphaShift;
				const int		iBlueV	= (pxInVertices[v].m_iRGBA8888 & uBlueMask)  >> uBlueShift;
				const int		iGreenV	= (pxInVertices[v].m_iRGBA8888 & uGreenMask) >> uGreenShift;
				const int		iRedV	= (pxInVertices[v].m_iRGBA8888 & uRedMask)   >> uRedShift;

				if (dbg_bNewColourInterpolation)
				{
					fLambda = 1.0f - fLambdaXY; 
					fOneMinusLambda = fLambdaXY; 
				}

				const int		iAlpha	=	static_cast<int>(Asura_Maths::ClampToRange((iAlphaU	* fLambda + iAlphaV	* fOneMinusLambda), 0.f, 255.f));
				const int		iBlue	=	static_cast<int>(Asura_Maths::ClampToRange((iBlueU	* fLambda + iBlueV	* fOneMinusLambda), 0.f, 255.f));
				const int		iGreen	=	static_cast<int>(Asura_Maths::ClampToRange((iGreenU	* fLambda + iGreenV	* fOneMinusLambda), 0.f, 255.f));
				const int		iRed	=	static_cast<int>(Asura_Maths::ClampToRange((iRedU	* fLambda + iRedV	* fOneMinusLambda), 0.f, 255.f));

				pxOutVertices[uNewVertexCount].m_iRGBA8888 = (iAlpha << uAlphaShift) + (iBlue << uBlueShift) + (iGreen << uGreenShift) + (iRed << uRedShift);
#else
				pxOutVertices[uNewVertexCount].m_iRGBA8888 = pxInVertices[u].m_iRGBA8888;
#endif

				uNewVertexCount++;
			}
		}
		else
		{
			if (fDot[v]<0.0f)
			{
				// Current vertex is on positive side of plane,
				// but previous vertex is on negative side.
				const Asura_Vector_3 xEdge = xPosU - xPosV;
				float f = xEdge * xPlane.m_xNormal;

				Asura_Assert( f!=0.0f, "edge parallel to plane");

				float fLambda = Asura_Maths::ClampToRange(fDot[u] / f * dbg_fLambdaFudge, 0.f, 1.f);
				float fOneMinusLambda = 1.f - fLambda;

				Asura_Vector_3 xOutPos = xPosU - xEdge * fLambda;

				float fOutProjW = xOutPos.GetX() * m_xProjMat.x.w + xOutPos.GetY() * m_xProjMat.y.w + xOutPos.GetZ() * m_xProjMat.z.w + m_xProjMat.w.w + 0.001f;
				Asura_Vector_3 xOutProj;
				xOutProj.SetX(240.f * (xOutPos.GetX() * m_xProjMat.x.x + xOutPos.GetY() * m_xProjMat.y.x + xOutPos.GetZ() * m_xProjMat.z.x + m_xProjMat.w.x) / fOutProjW); 
				xOutProj.SetY(136.f * (xOutPos.GetX() * m_xProjMat.x.y + xOutPos.GetY() * m_xProjMat.y.y + xOutPos.GetZ() * m_xProjMat.z.y + m_xProjMat.w.y) / fOutProjW); 
				xOutProj.SetZ((xOutPos.GetX() * m_xProjMat.x.z + xOutPos.GetY() * m_xProjMat.y.z + xOutPos.GetZ() * m_xProjMat.z.z + m_xProjMat.w.z) / fOutProjW); 

#ifndef ASURA_PSP_NO_CLIPPER_COLOUR_INTERPOLATION
				float fHorRange = xProjU.GetX() - xProjV.GetX(); 
				float fVerRange = xProjU.GetY() - xProjV.GetY();
				float fLambdaXY = 0.0f;

				if (Asura_Maths::Abs(fHorRange) > Asura_Maths::Abs(fVerRange))
				{
					fLambdaXY = Asura_Maths::ClampToRange((xOutProj.GetX() - xProjV.GetX()) / fHorRange, 0.f, 1.f); 
				}
				else
				{
					fLambdaXY = Asura_Maths::ClampToRange((xOutProj.GetY() - xProjV.GetY()) / fVerRange, 0.f, 1.f); 
				}
#endif

				//Asura_Assert(xOutPos.x > -100000.f && xOutPos.y > -100000.f && xOutPos.z > -100000.f, "Position very negative");
				pxOutVertices[uNewVertexCount].m_afPosition[0] = xOutPos.x;
				pxOutVertices[uNewVertexCount].m_afPosition[1] = xOutPos.y;
				pxOutVertices[uNewVertexCount].m_afPosition[2] = xOutPos.z;

				pxOutVertices[uNewVertexCount].m_usU = static_cast<u_short>(pxInVertices[u].m_usU * fOneMinusLambda + pxInVertices[v].m_usU * fLambda);
				pxOutVertices[uNewVertexCount].m_usV = static_cast<u_short>(pxInVertices[u].m_usV * fOneMinusLambda + pxInVertices[v].m_usV * fLambda);

				pxOutVertices[uNewVertexCount].m_asNormal[0] = static_cast<u_short>(pxInVertices[u].m_asNormal[0] * fOneMinusLambda + pxInVertices[v].m_asNormal[0] * fLambda);
				pxOutVertices[uNewVertexCount].m_asNormal[1] = static_cast<u_short>(pxInVertices[u].m_asNormal[1] * fOneMinusLambda + pxInVertices[v].m_asNormal[1] * fLambda);
				pxOutVertices[uNewVertexCount].m_asNormal[2] = static_cast<u_short>(pxInVertices[u].m_asNormal[2] * fOneMinusLambda + pxInVertices[v].m_asNormal[2] * fLambda);

#ifndef ASURA_PSP_NO_CLIPPER_COLOUR_INTERPOLATION
				const int		iAlphaU	= (pxInVertices[u].m_iRGBA8888 & uAlphaMask) >> uAlphaShift;
				const int		iBlueU	= (pxInVertices[u].m_iRGBA8888 & uBlueMask)  >> uBlueShift;
				const int		iGreenU	= (pxInVertices[u].m_iRGBA8888 & uGreenMask) >> uGreenShift;
				const int		iRedU	= (pxInVertices[u].m_iRGBA8888 & uRedMask)   >> uRedShift;

				const int		iAlphaV	= (pxInVertices[v].m_iRGBA8888 & uAlphaMask) >> uAlphaShift;
				const int		iBlueV	= (pxInVertices[v].m_iRGBA8888 & uBlueMask)  >> uBlueShift;
				const int		iGreenV	= (pxInVertices[v].m_iRGBA8888 & uGreenMask) >> uGreenShift;
				const int		iRedV	= (pxInVertices[v].m_iRGBA8888 & uRedMask)   >> uRedShift;

				if (dbg_bNewColourInterpolation)
				{
					fLambda = 1.0f - fLambdaXY; 
					fOneMinusLambda = fLambdaXY; 
				}

				const int		iAlpha	=	static_cast<int>(Asura_Maths::ClampToRange((iAlphaV	* fLambda	+ iAlphaU	* fOneMinusLambda), 0.f, 255.f));
				const int		iBlue	=	static_cast<int>(Asura_Maths::ClampToRange((iBlueV	* fLambda	+ iBlueU	* fOneMinusLambda), 0.f, 255.f));
				const int		iGreen	=	static_cast<int>(Asura_Maths::ClampToRange((iGreenV	* fLambda	+ iGreenU	* fOneMinusLambda), 0.f, 255.f));
				const int		iRed	=	static_cast<int>(Asura_Maths::ClampToRange((iRedV	* fLambda	+ iRedU		* fOneMinusLambda), 0.f, 255.f));

				pxOutVertices[uNewVertexCount].m_iRGBA8888 = (iAlpha << uAlphaShift) + (iBlue << uBlueShift) + (iGreen << uGreenShift) + (iRed << uRedShift);
#else
				pxOutVertices[uNewVertexCount].m_iRGBA8888 = pxInVertices[u].m_iRGBA8888;
#endif


				uNewVertexCount++;
			}

			/// Kevin Lea - 14/11/2002 11:22:04 - include current vertex
			pxOutVertices[uNewVertexCount] = pxInVertices[u];
			uNewVertexCount++;
		}
	}
	uNumberOfVertices = uNewVertexCount;
	Asura_Assert(uNumberOfVertices >= 3, "Not enough output verts");

	return !bAllPositive;
}

#include <kernelutils.h>


union Asura_Frustum_Test_Result_Bitfield
{

	struct
	{
		unsigned int	m_bFirstLeft:	1,
m_bSecondLeft:	1,
m_bThirdLeft:	1,
				:29;
		unsigned int	m_bFirstRight:	1,
m_bSecondRight:	1,
m_bThirdRight:	1,
				:29;
		unsigned int	m_bFirstUp:		1,
m_bSecondUp:	1,
m_bThirdUp:		1,
				:29;
		unsigned int	m_bFirstDown:	1,
m_bSecondDown:	1,
m_bThirdDown:	1,
				:29;
	} m_xFlags;

	struct
	{
		unsigned int	m_uiPlaneLeft;
		unsigned int	m_uiPlaneRight;
		unsigned int	m_uiPlaneUp;
		unsigned int	m_uiPlaneDown;

	} m_xPlane;

};

inline void ShiftPlaneBitfields( Asura_Frustum_Test_Result_Bitfield& xResult )
{
	xResult.m_xPlane.m_uiPlaneLeft = xResult.m_xPlane.m_uiPlaneLeft >> 1;
	xResult.m_xPlane.m_uiPlaneRight = xResult.m_xPlane.m_uiPlaneRight >> 1;
	xResult.m_xPlane.m_uiPlaneUp = xResult.m_xPlane.m_uiPlaneUp >> 1;
	xResult.m_xPlane.m_uiPlaneDown = xResult.m_xPlane.m_uiPlaneDown >> 1;
}

inline void ResetPlaneBitfields( Asura_Frustum_Test_Result_Bitfield& xResult )
{
	xResult.m_xPlane.m_uiPlaneLeft = 0;
	xResult.m_xPlane.m_uiPlaneRight = 0;
	xResult.m_xPlane.m_uiPlaneUp = 0;
	xResult.m_xPlane.m_uiPlaneDown = 0;
}

static void TestPointAgainstFrustum(const Asura_Vector_3& xPos, const Asura_PSP_FrustumLimits& xFrustum, Asura_Frustum_Test_Result_Bitfield& xResult);

void Asura_PSP_RenderEnvironment_Zone::ClipStrip(Asura_PSP_RenderEnvironment_Strip& xStrip, const bool bSecondPass)
{
}

void Asura_PSP_RenderEnvironment_Zone::CalculateNumberOfStrips(Asura_PSP_RenderEnvironment_Strip &xStrip)
{
	Asura_ResourcePushType(ASURA_RESOURCE_ENVIRONMENT_MESH);
	{
		const bool bSecondPass = (&xStrip - m_pxStrips) >= m_uTotalNumberOfStrips;
		const size_t xVertexPosOffset = bSecondPass ? m_xSecondPassVertexPosOffset : m_xFirstPassVertexPosOffset;
		const size_t xVertexPosComponentSize = Asura_PSP_VertexFormat::GetSizeOfVertexComponent(xStrip.m_usVertexFormat);

		if(xStrip.m_uNumberOfIndices >= 3)
		{
			u_int uNumberOfStrips = 1;

			for (unsigned int uCurrentVertex = 0; uCurrentVertex < xStrip.m_uNumberOfIndices - 3; uCurrentVertex++)
			{
				// update the pointers
				const unsigned char* pucFirstVertex = GetVertexPtr(xStrip, uCurrentVertex);
				const unsigned char* pucSecondVertex = GetVertexPtr(xStrip, uCurrentVertex + 1);
				Asura_Assert(pucFirstVertex, "Invalid vertex address");
				Asura_Assert(pucSecondVertex, "Invalid vertex address");

				// compare the first and second vertices
				bool bFirstDegeneratedTriangle = Asura_PSP_VertexFormat::PositionsAreEqual(pucFirstVertex, pucSecondVertex, xVertexPosOffset, xVertexPosComponentSize);

				if (!bFirstDegeneratedTriangle) continue; 

				const unsigned char* pucThirdVertex = GetVertexPtr(xStrip, uCurrentVertex + 2);
				const unsigned char* pucForthVertex = GetVertexPtr(xStrip, uCurrentVertex + 3);

				// compare the third and forth vertices
				bool bSecondDegeneratedTriangle = Asura_PSP_VertexFormat::PositionsAreEqual(pucThirdVertex, pucForthVertex, xVertexPosOffset, xVertexPosComponentSize);

				if (!bSecondDegeneratedTriangle) continue; 

				bool bJump = !Asura_PSP_VertexFormat::PositionsAreEqual(pucSecondVertex, pucThirdVertex, xVertexPosOffset, xVertexPosComponentSize);

				if (bFirstDegeneratedTriangle && bSecondDegeneratedTriangle)
				{
					if (bJump)
					{
						// we have a new strip
						uNumberOfStrips++; 

						// skip the next degenerate triangle 
						uCurrentVertex++; 
					}
					else
					{
						// TODO: check if this assert is valid
						//Asura_Assert(false, "Four consecutive vertices collapsed to the same one, check what the strip process is doing!"); 
					}
				}

			}

			// Allocate BB, ...
			xStrip.m_uNumberOfStrips = uNumberOfStrips;
			Asura_ResourceAllocateArray(xStrip.m_axBB, uNumberOfStrips);
			Asura_ResourceAllocateArray(xStrip.m_auStartIndex, uNumberOfStrips);

			// Loop again to calculate the starting index for each individual strip...
			// could save this with a dynamic list structure

			u_int uCurrentStrip = 0;
			// (SB - 23/01/07) With variable format strips the start index is relative to the strip, not the vertex buffer
			xStrip.m_auStartIndex[uCurrentStrip] = 0;//xStrip.m_uStartIndex;
			xStrip.m_axBB[uCurrentStrip].Reset();
#ifndef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
			const
#endif
				unsigned char* pucVertex = GetVertexPtr(xStrip, 0);
			Asura_Vector_3 xPos = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
			xStrip.m_axBB[uCurrentStrip].Set(xPos);


			// KOR - 24/03/2006 - define ASURA_PSP_ENV_RENDER_COLOUR_STRIPS to show each substrip as a different colour.
			// This should help track down bad mesh creating lots of degenerates.
			// Ideally this should become a realtime tools feature instead of on load.
#ifdef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
			u_int uCurrentColour = 0xffffffff;
#endif
			for (unsigned int uCurrentVertex = 0; uCurrentVertex < xStrip.m_uNumberOfIndices - 3; uCurrentVertex++)
			{
				// update the pointers
#ifndef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
				const
#endif
					unsigned char* pucFirstVertex = GetVertexPtr(xStrip, uCurrentVertex);
				const unsigned char* pucSecondVertex = GetVertexPtr(xStrip, uCurrentVertex + 1);
				const unsigned char* pucThirdVertex = GetVertexPtr(xStrip, uCurrentVertex + 2);
				const unsigned char* pucForthVertex = GetVertexPtr(xStrip, uCurrentVertex + 3);

				// compare the first and second vertices
				bool bFirstDegeneratedTriangle = Asura_PSP_VertexFormat::PositionsAreEqual(pucFirstVertex, pucSecondVertex, xVertexPosOffset, xVertexPosComponentSize);

				// compare the third and forth vertices
				bool bSecondDegeneratedTriangle = Asura_PSP_VertexFormat::PositionsAreEqual(pucThirdVertex, pucForthVertex, xVertexPosOffset, xVertexPosComponentSize);

				// check that there is a jump between them 
				bool bJump = !Asura_PSP_VertexFormat::PositionsAreEqual(pucSecondVertex, pucThirdVertex, xVertexPosOffset, xVertexPosComponentSize);

				if (bFirstDegeneratedTriangle && bSecondDegeneratedTriangle && bJump)
				{
					// add the last vertex to the previous strip bounding box 
					Asura_Vector_3 xCurrentPos = Asura_PSP_VertexFormat::GetPositionFromVertex(pucFirstVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
					xStrip.m_axBB[uCurrentStrip].Set(xCurrentPos);

					// we have a new strip
					uCurrentStrip++;
					xStrip.m_axBB[uCurrentStrip].Reset();
					// (SB - 23/01/07) With variable format strips the start index is relative to the strip, not the vertex buffer
					u_int uCurrentOffset = uCurrentVertex + 2;
					xStrip.m_auStartIndex[uCurrentStrip] = (uCurrentVertex % 2) ? uCurrentOffset + 1: uCurrentOffset;
					Asura_Vector_3 xNextPos = Asura_PSP_VertexFormat::GetPositionFromVertex(pucThirdVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
					xStrip.m_axBB[uCurrentStrip].Set(xNextPos);

					// skip the degenerate triangle
					uCurrentVertex++; 

#ifdef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
					uCurrentColour = Asura_Random::GetIntU();
#endif
				}
				else
				{
					// we don't include degenerated triangles in BB calculation
					//Asura_Vector_3 xPos(pxFirstVertex->m_afPosition[0], pxFirstVertex->m_afPosition[1], pxFirstVertex->m_afPosition[2]);
					Asura_Vector_3 xPos = Asura_PSP_VertexFormat::GetPositionFromVertex(pucFirstVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
					xStrip.m_axBB[uCurrentStrip].Set(xPos);
				}

#ifdef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
				Asura_PSP_VertexFormat::SetColourInVertex(pucFirstVertex, uCurrentColour, xStrip.m_usVertexFormat);
#endif
			}

			// last three vertices have to be added to the bounding box 
			pucVertex = GetVertexPtr(xStrip, xStrip.m_uNumberOfIndices - 3);
			Asura_Vector_3 xPosPrePreLast = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
			xStrip.m_axBB[uCurrentStrip].Set(xPosPrePreLast);

#ifdef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
			Asura_PSP_VertexFormat::SetColourInVertex(pucVertex, uCurrentColour, xStrip.m_usVertexFormat);
#endif

			pucVertex = GetVertexPtr(xStrip, xStrip.m_uNumberOfIndices - 2);
			Asura_Vector_3 xPosPreLast = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
			xStrip.m_axBB[uCurrentStrip].Set(xPosPreLast);

#ifdef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
			Asura_PSP_VertexFormat::SetColourInVertex(pucVertex, uCurrentColour, xStrip.m_usVertexFormat);
#endif

			pucVertex = GetVertexPtr(xStrip, xStrip.m_uNumberOfIndices - 1);
			Asura_Vector_3 xPosLast = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertex, xStrip.m_usVertexFormat, xVertexPosOffset);
			xStrip.m_axBB[uCurrentStrip].Set(xPosLast);

#ifdef ASURA_PSP_ENV_RENDER_COLOUR_STRIPS
			Asura_PSP_VertexFormat::SetColourInVertex(pucVertex, uCurrentColour, xStrip.m_usVertexFormat);
#endif

			Asura_Assert(uCurrentStrip == (uNumberOfStrips - 1), "Wrong number of strips!!"); 
		}
		else
		{
			Asura_Assert(false, "Strip does not have at least 3 vertices!"); 

			// Reset the data to contain no strips...
			xStrip.m_uNumberOfStrips	= 0;
			xStrip.m_axBB				= NULL;
			xStrip.m_auStartIndex		= NULL;
		}
	}
	Asura_ResourcePopType();
}

void Asura_PSP_RenderEnvironment_Zone::DrawStripComplete(Asura_PSP_RenderEnvironment_Strip &xStrip, const size_t xVertexSize)
{
	const unsigned char* pucVertices;
	const u_short* pxIndices;
	if(m_pusIndexBuffer)
	{
		pucVertices = GetVertexPtr(xStrip, 0);
		pxIndices = m_pusIndexBuffer + xStrip.m_uStartIndex;
	}
	else
	{
		pucVertices = GetVertexPtr(xStrip, 0);
		pxIndices = 0;
	}

	DrawArray
		(
		m_eRenderMode,
		xStrip.m_usVertexFormat,
		xStrip.m_uNumberOfIndices,
		(void*)pxIndices,
		(void*)pucVertices
		);
}

void Asura_PSP_RenderEnvironment_Zone::DrawInternalStripComplete(Asura_PSP_RenderEnvironment_Strip& xStrip, u_int uInternalStripIndex, const size_t xVertexSize)
{
	const unsigned char* pucVertexBuffer = GetVertexPtr(xStrip, xStrip.m_auStartIndex[uInternalStripIndex]);
	int uVertices = 0;

	// How many vertices in this substrip?
	if (uInternalStripIndex != xStrip.m_uNumberOfStrips - 1)
	{
		// Rest of the substrip
		uVertices = xStrip.m_auStartIndex[uInternalStripIndex + 1] - xStrip.m_auStartIndex[uInternalStripIndex];
	}
	else
	{
		// Rest of the triangles
		uVertices = xStrip.m_uNumberOfIndices - xStrip.m_auStartIndex[uInternalStripIndex];
	}

	DrawArray
		(
		m_eRenderMode,
		xStrip.m_usVertexFormat,
		uVertices,
		NULL,
		(void*)pucVertexBuffer
		);
}

void Asura_PSP_RenderEnvironment_Zone::DrawInternalStripPartial(Asura_PSP_RenderEnvironment_Strip& xStrip, u_int uInternalStripIndex, u_int uFirst, u_int uLast, const size_t xVertexSize)
{
		const unsigned char* pucVertexBuffer = GetVertexPtr(xStrip, uFirst + xStrip.m_auStartIndex[uInternalStripIndex]);

		Asura_Assert( uLast - uFirst <= xStrip.m_uNumberOfIndices, "Wrong first or last vertex" );

		// Ignore the wingding by now
		GU_Disable(SCEGU_CULL_FACE); 
		DrawArray
			(
			m_eRenderMode,
			xStrip.m_usVertexFormat,
			uLast - uFirst,
			NULL,
			(void*)pucVertexBuffer
			);
		GU_Enable(SCEGU_CULL_FACE); 
}

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
// Clip the specified vertices to the specified plane using degenerates
//
bool Asura_PSP_RenderEnvironment_Zone::DegeneratePlaneClip( const Asura_PSP_RenderEnvironment_VertexClipQuad *axSrcVertices, Asura_PSP_RenderEnvironment_VertexClipQuad *axDstVertices, uint32& ruNumberOfVertices, const Asura_PSP_Plane& rxPlane ) const
{
	// RJT TODO: Bounds checking each vertex is the most accurate but could just check maximum possible addition, i.e:
//	Asura_Assert( ( ruNumberOfVertices * 5 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices for quad scratch" );
//	if( ( ruNumberOfVertices * 5 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }

	// Clip data
	struct SClip
	{
		float	fT;
		uint32	uClip;
	};

	// Pre-calculate clip data (so early out's can be performed)
	SClip *axClip = reinterpret_cast< SClip * >( alloca( sizeof( SClip ) * ruNumberOfVertices ) );
	uint32 uTotal = 0;
	for( uint32 u = 0; u < ruNumberOfVertices; ++u )
	{
		axClip[ u ].fT = rxPlane.DistanceTo( axSrcVertices[ u ].xXYZV );
		// RJT TODO: Avoid branch! (conditional move)
		axClip[ u ].uClip = ( axClip[ u ].fT < 0.0f );

		uTotal += axClip[ u ].uClip;
	}

	// Early out's
	if( uTotal == ruNumberOfVertices )
	{
		// All clipped (don't bother swapping buffers)
		ruNumberOfVertices = 0;
		return false;
	}
	else if( !uTotal )
	{
		// None clipped so don't bother swapping buffers
		return false;
	}

	// Test Q
	SClip *pxClip = &axClip[ -1 ];
	uint32 uClipCode = ( pxClip[ 1 ].uClip << 1 );

	// Test R
	uClipCode |= ( pxClip[ 2 ].uClip << 2 );

	// 4 cases for first triangle
#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_VFPU
	// RJT NOTE: This make's sure that subsequent reads of this new data come directly from memory
	sceKernelDcacheInvalidateRange( axDstVertices, sizeof( Asura_PSP_RenderEnvironment_VertexClipQuad ) * ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS );
	axDstVertices = reinterpret_cast< Asura_PSP_RenderEnvironment_VertexClipQuad * >( reinterpret_cast< uint32 >( axDstVertices ) | 0x40000000 );
#endif
	const Asura_PSP_RenderEnvironment_VertexClipQuad *pxSrcVertex = &axSrcVertices[ -1 ];
	Asura_PSP_RenderEnvironment_VertexClipQuad *pxDstVertex = axDstVertices;
	switch( uClipCode >> 1 )
	{
	case 0: // Q and R inside
		Asura_Assert( ( pxDstVertex - axDstVertices + 2 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
		if( ( pxDstVertex - axDstVertices + 2 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
		pxDstVertex[ 0 ] = pxSrcVertex[ 1 ];
		pxDstVertex[ 1 ] = pxSrcVertex[ 2 ];
		pxDstVertex = &pxDstVertex[ 2 ];
		break;
	case 1:	// Q outside, R inside
		Asura_Assert( ( pxDstVertex - axDstVertices + 2 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
		if( ( pxDstVertex - axDstVertices + 2 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
		pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
		pxDstVertex[ 1 ] = pxSrcVertex[ 2 ];
		pxDstVertex = &pxDstVertex[ 2 ];
		break;
	case 2:	// Q inside, R outside
		Asura_Assert( ( pxDstVertex - axDstVertices + 2 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
		if( ( pxDstVertex - axDstVertices + 2 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
		pxDstVertex[ 0 ] = pxSrcVertex[ 1 ];
		pxDstVertex[ 1 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
		pxDstVertex = &pxDstVertex[ 2 ];
		break;
//	case 3:	// Q outside, R outside
//	default:
//		break;
	}

	// Process remaining vertices in strip
	for( uint32 u = 2; u < ruNumberOfVertices; ++u )
	{
		// Next vertex...
		uClipCode >>= 1;
		++pxClip;
		++pxSrcVertex;

		// Test P
		uClipCode |= ( pxClip[ 2 ].uClip << 2 );

		// 8 possible cases
		switch( uClipCode )
		{
		case 0:	// All inside
			Asura_Assert( ( pxDstVertex - axDstVertices + 1 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
			if( ( pxDstVertex - axDstVertices + 1 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
			pxDstVertex[ 0 ] = pxSrcVertex[ 2 ];
			pxDstVertex = &pxDstVertex[ 1 ];
			break;
		case 1:	// P outside, Q and R inside
			Asura_Assert( ( pxDstVertex - axDstVertices + 3 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
			if( ( pxDstVertex - axDstVertices + 3 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
			pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );
			pxDstVertex[ 1 ] = pxSrcVertex[ 1 ];
			pxDstVertex[ 2 ] = pxSrcVertex[ 2 ];
			pxDstVertex = &pxDstVertex[ 3 ];
			break;
		case 2:	// P inside, Q outside and R inside
			Asura_Assert( ( pxDstVertex - axDstVertices + 3 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
			if( ( pxDstVertex - axDstVertices + 3 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
			pxDstVertex[ 0 ] = pxSrcVertex[ 2 ];
			pxDstVertex[ 1 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
			pxDstVertex[ 2 ] = pxSrcVertex[ 2 ];
			pxDstVertex = &pxDstVertex[ 3 ];
			break;
		case 3:	// P and Q outside, R inside
			// RJT NOTE: This check is to prevent winding change
			if( ( ( pxDstVertex - axDstVertices ) & 1 ) == ( u & 1 ) )
			{
				Asura_Assert( ( pxDstVertex - axDstVertices + 5 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
				if( ( pxDstVertex - axDstVertices + 5 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }

				pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );

				// RJT FIXME: Can't do this as we're un-cached! (when using VFPU write-back cache)
#if 0
				pxDstVertex[ 1 ] = pxDstVertex[ 0 ];
				pxDstVertex[ 2 ] = pxDstVertex[ 0 ];
#else
				pxDstVertex[ 1 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );
				pxDstVertex[ 2 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );
#endif
				pxDstVertex[ 3 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
				pxDstVertex[ 4 ] = pxSrcVertex[ 2 ];
				pxDstVertex = &pxDstVertex[ 5 ];
			}
			else
			{
				Asura_Assert( ( pxDstVertex - axDstVertices + 4 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
				if( ( pxDstVertex - axDstVertices + 4 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }

				pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );

				// RJT FIXME: Can't do this as we're un-cached! (when using VFPU write-back cache)
#if 0
				pxDstVertex[ 1 ] = pxDstVertex[ 0 ];
#else
				pxDstVertex[ 1 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );
#endif

				pxDstVertex[ 2 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
				pxDstVertex[ 3 ] = pxSrcVertex[ 2 ];
				pxDstVertex = &pxDstVertex[ 4 ];
			}
			break;
		case 4:	// P and Q inside, R outside
			Asura_Assert( ( pxDstVertex - axDstVertices + 3 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
			if( ( pxDstVertex - axDstVertices + 3 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
			pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );
			pxDstVertex[ 1 ] = pxSrcVertex[ 1 ];
			pxDstVertex[ 2 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
			pxDstVertex = &pxDstVertex[ 3 ];
			break;
		case 5:	// P outside, Q inside and R outside
			Asura_Assert( ( pxDstVertex - axDstVertices + 1 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
			if( ( pxDstVertex - axDstVertices + 1 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }
			pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 1 ], pxSrcVertex[ 2 ], pxClip[ 1 ].fT, pxClip[ 2 ].fT, m_xAProjMat );
			pxDstVertex = &pxDstVertex[ 1 ];
			break;
		case 6:	// P inside, Q and R outside
			Asura_Assert( ( pxDstVertex - axDstVertices + 2 ) <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices in quad scratch" );
			if( ( pxDstVertex - axDstVertices + 2 ) > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { ruNumberOfVertices = 0; return false; }

			pxDstVertex[ 0 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );

			// RJT FIXME: Can't do this as we're un-cached! (when using VFPU write-back cache)
#if 0
			pxDstVertex[ 1 ] = pxDstVertex[ 0 ];
#else
			pxDstVertex[ 1 ].Clipped( pxSrcVertex[ 2 ], pxSrcVertex[ 0 ], pxClip[ 2 ].fT, pxClip[ 0 ].fT, m_xAProjMat );
#endif

			pxDstVertex = &pxDstVertex[ 2 ];
			break;
//		case 7:	// All outside
//		default:
//			break;
		}
	}

	ruNumberOfVertices = pxDstVertex - axDstVertices;

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_VFPU
		// Flush VFPU write-back cache now we're done
	#ifdef ASURA_PSP_ASM_USE_SN_SYNTAX_NEW
		__asm {
			vflush
			vnop
		}
	#elif ASURA_PSP_USE_VFPU_ASSEMBLER
		__asm__ (
			".set		push\n"							// Save assembler option
			".set		noreorder\n"					// Suppress reordering
			"vflush\n"
			"vnop\n"
			".set		pop\n"							// Restore assembler option
			);
	#endif
#endif

	return true;
}
#endif	// ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING

DEBUGVAR bool dbg_bDrawRemainingClippedStrips = true; 
DEBUGVAR float dbg_fForcedClippedDistance = 1.0f; 

void Asura_PSP_RenderEnvironment_Zone::ClipStripSplitThroughDegenerates(Asura_PSP_RenderEnvironment_Strip& xStrip, const bool bReflected, const bool bSecondPass, const size_t& xVertexSize, const size_t& xVertexPosOffset)
{
	// KOR - this function will only work with non-indexed data

	Asura_Bounding_Box xBB = xStrip.m_xBB;

	// Are we rendering reflected geometry?
	if (bReflected)
	{
		xBB.MinY = -xStrip.m_xBB.MaxY;
		xBB.MaxY = -xStrip.m_xBB.MinY;
	}

	// Initial error checking
	Asura_Assert(xStrip.m_uNumberOfIndices >= 3, "Less than one triangle?");

	// Check to see if strip is completely inside guardband
	u_int uModuleClipFlags = 0;

	if (!bReflected)
	{
		uModuleClipFlags = m_puModuleClipFlags[xStrip.m_uModule];
	}
	else
	{
		uModuleClipFlags = m_puReflectedModuleClipFlags[xStrip.m_uModule];
	}

	//	bool bTestL = (iModuleClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_L) && (BBIsPartiallyAbovePlane(xBB, m_xGuardband.m_xLeft));
	//	bool bTestR = (iModuleClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_R) && (BBIsPartiallyAbovePlane(xBB, m_xGuardband.m_xRight));
	//	bool bTestD = (iModuleClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_D) && (BBIsPartiallyAbovePlane(xBB, m_xGuardband.m_xDown));
	//	bool bTestU = (iModuleClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_U) && (BBIsPartiallyAbovePlane(xBB, m_xGuardband.m_xUp));
	//	if (!bTestL && !bTestR && !bTestD && !bTestU) return;
	// check module strip BB against guardband, using what we know about the entire module
	u_int uModuleStripClipFlags = GetGuardbandClipFlags2( xBB, m_xGuardband, uModuleClipFlags );
	if ( !(uModuleStripClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_ANY) )
	{
		if (dbg_bDontRenderCompleteStripsBeforeClipping)
		{
			DrawStripComplete(xStrip, xVertexSize); 
		}
		return;
	}

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
	uint32 uTotalClippedVertices = 0;
#endif

    for (unsigned int uInternalStrips = 0; uInternalStrips < xStrip.m_uNumberOfStrips; uInternalStrips++)
	{
#ifdef PSP_TEST_GUARDBAND_CLIPPING
		++m_uNumStripTests;
#endif

		Asura_Bounding_Box xInternalBB = xStrip.m_axBB[uInternalStrips];

		if (bReflected)
		{
			xInternalBB.MinY = -xStrip.m_axBB[uInternalStrips].MaxY;
			xInternalBB.MaxY = -xStrip.m_axBB[uInternalStrips].MinY;
		}

		if(!BBIsPartiallyAbovePlane(xInternalBB, xCameraFrontPlane))
		{
#ifdef PSP_TEST_GUARDBAND_CLIPPING
			++m_uNumStripBehinds;
#endif
			continue;
		}

		// Check to see if strip is outside frustum
		if (BBIsCompletelyAboveAnyFrustumPlane(xInternalBB, m_xFrustum)) continue;
		//		if(BBIsCompletelyAbovePlane(xInternalBB, m_xFrustum.m_xLeft))		continue;
		//		if(BBIsCompletelyAbovePlane(xInternalBB, m_xFrustum.m_xRight))		continue;
		//		if(BBIsCompletelyAbovePlane(xInternalBB, m_xFrustum.m_xDown))		continue;
		//		if(BBIsCompletelyAbovePlane(xInternalBB, m_xFrustum.m_xUp))			continue;

		// Check to see if strip is completely inside guardband
		//		bool bTestL = (uModuleClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_L) && (BBIsPartiallyAbovePlane(xInternalBB, m_xGuardband.m_xLeft));
		//		bool bTestR = (uModuleClipFlags	 & ASURA_PSP_MODULE_CLIPS_GUARDBAND_R) && (BBIsPartiallyAbovePlane(xInternalBB, m_xGuardband.m_xRight));
		//		bool bTestD = (uModuleClipFlags	 & ASURA_PSP_MODULE_CLIPS_GUARDBAND_D) && (BBIsPartiallyAbovePlane(xInternalBB, m_xGuardband.m_xDown));
		//		bool bTestU = (uModuleClipFlags	 & ASURA_PSP_MODULE_CLIPS_GUARDBAND_U) && (BBIsPartiallyAbovePlane(xInternalBB, m_xGuardband.m_xUp));
		//		if (!bTestL && !bTestR && !bTestD && !bTestU) continue;

		u_int uClipFlags = GetGuardbandClipFlags2( xInternalBB, m_xGuardband, uModuleStripClipFlags );
		if ( !(uClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_ANY) )
		{
			if (dbg_bDontRenderCompleteStripsBeforeClipping)
			{
				DrawInternalStripComplete(xStrip, uInternalStrips, xVertexSize); 
			}
			continue;
		}
		bool bTestL = uClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_L;
		bool bTestR = uClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_R;
		bool bTestD = uClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_D;
		bool bTestU = uClipFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_U;

#ifdef PSP_TEST_GUARDBAND_CLIPPING
		++m_uNumStripAccepts;
#endif

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
		uint32 uNumberOfVertices = ( uInternalStrips == ( xStrip.m_uNumberOfStrips - 1 ) ) ?
			( xStrip.m_uNumberOfIndices - xStrip.m_auStartIndex[ uInternalStrips ] ) :
			( xStrip.m_auStartIndex[ uInternalStrips + 1 ] - xStrip.m_auStartIndex[ uInternalStrips ] );

		// Bounds check
		Asura_Assert( uNumberOfVertices <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many vertices for quad scratch" );
		if( uNumberOfVertices > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { continue; }

		// RJT TODO: Specialise for certain types?
		// Uncompress vertices
		uint32 uBuffer = 0, uOffset = 0;
		Asura_PSP_RenderEnvironment_VertexClipQuad *axClippedVertices = g_aaxClippedVertices[ uBuffer ];
		const uint8 *auCompressedVertices = GetVertexPtr( xStrip, xStrip.m_auStartIndex[ uInternalStrips ] );
		for( uint32 u = 0; u < uNumberOfVertices; ++u )
		{
			axClippedVertices[ u ].Uncompress( &auCompressedVertices[ uOffset ], xStrip.m_usVertexFormat );
			uOffset += xVertexSize;
		}

		// Reflect if necessary
		if( bReflected ) { for( uint32 u = 0; u < uNumberOfVertices; ++u ) { axClippedVertices[ u ].xXYZV.SetY( axClippedVertices[ u ].xXYZV.GetY() * -1.0f ); } }

		// Down
		uint32 uClippedVertices = uNumberOfVertices;
		if( bTestD && uClippedVertices )
		{
			uint32 uOneMinusBuffer = 1 - uBuffer;
			if( DegeneratePlaneClip( g_aaxClippedVertices[ uBuffer ], g_aaxClippedVertices[ uOneMinusBuffer ], uClippedVertices, m_xClipFrustum.m_xDown ) )
			{
				uBuffer = uOneMinusBuffer;
			}
		}

		// Left
		if( bTestL && uClippedVertices )
		{
			uint32 uOneMinusBuffer = 1 - uBuffer;
			if( DegeneratePlaneClip( g_aaxClippedVertices[ uBuffer ], g_aaxClippedVertices[ uOneMinusBuffer ], uClippedVertices, m_xClipFrustum.m_xLeft ) )
			{
				uBuffer = uOneMinusBuffer;
			}
		}

		// Right
		if( bTestR && uClippedVertices )
		{
			uint32 uOneMinusBuffer = 1 - uBuffer;
			if( DegeneratePlaneClip( g_aaxClippedVertices[ uBuffer ], g_aaxClippedVertices[ uOneMinusBuffer ], uClippedVertices, m_xClipFrustum.m_xRight ) )
			{
				uBuffer = uOneMinusBuffer;
			}
		}

		// Up
		if( bTestU && uClippedVertices )
		{
			uint32 uOneMinusBuffer = 1 - uBuffer;
			if( DegeneratePlaneClip( g_aaxClippedVertices[ uBuffer ], g_aaxClippedVertices[ uOneMinusBuffer ], uClippedVertices, m_xClipFrustum.m_xUp ) )
			{
				uBuffer = uOneMinusBuffer;
			}
		}

		if( uClippedVertices )
		{
			// Bounds check
//			Asura_Assert( uClippedVertices <= ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS, "Too many clipped vertices in quad scratch" );
//			if( uClippedVertices > ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING_SCRATCHVERTS ) { continue; }

			Asura_Assert( ( g_uNumClippedVerts + uClippedVertices + 3 ) <= ASURA_PSP_MAX_CLIPPED_VERTS, "Too many clipped verts this frame" );
			if( ( g_uNumClippedVerts + uClippedVertices + 3 ) > ASURA_PSP_MAX_CLIPPED_VERTS ) { continue; }

			// Compress vertices
			Asura_PSP_RenderEnvironment_VertexClip *axClippedPackedVertices = &g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][ g_uNumClippedVerts + uTotalClippedVertices ];
			Asura_PSP_RenderEnvironment_VertexClip *pxClippedPackedVertex = reinterpret_cast< Asura_PSP_RenderEnvironment_VertexClip * >( reinterpret_cast< uint32 >( axClippedPackedVertices ) | 0x40000000 );
			Asura_PSP_RenderEnvironment_VertexClipQuad *pxClippedVertices = g_aaxClippedVertices[ uBuffer ];

			// RJT TODO: We're potentially compressing the same vertex more than once (be careful as we're using un-cached writes though!)
			// Initial degenerate
			if( uTotalClippedVertices )
			{
				pxClippedVertices[ 0 ].Compress( pxClippedPackedVertex++ );
				++uTotalClippedVertices;

				// Reverse winding if odd
				if( uTotalClippedVertices & 1 )
				{
					pxClippedVertices[ 0 ].Compress( pxClippedPackedVertex++ );
					++uTotalClippedVertices;
				}
			}

			// Sub-strip
			for( uint32 u = 0; u < uClippedVertices; ++u )
			{
				pxClippedVertices[ u ].Compress( &pxClippedPackedVertex[ u ] );
			}

			// Final degenerate
			if( uInternalStrips < ( xStrip.m_uNumberOfStrips - 1 ) )
			{
				pxClippedVertices[ uClippedVertices - 1 ].Compress( &pxClippedPackedVertex[ uClippedVertices ] );
				++uTotalClippedVertices;
			}

			uTotalClippedVertices += uClippedVertices;
		}
#else	// ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
		u_int uNumClippedVerts = 0;
		Asura_PSP_RenderEnvironment_VertexClip xVertexBuffer[2][9];
		const unsigned char* pucVertices = GetVertexPtr(xStrip, xStrip.m_auStartIndex[uInternalStrips]);

		Asura_Frustum_Test_Result_Bitfield xVertexFrustumTest;
		Asura_Frustum_Test_Result_Bitfield xVertexGuardbandTest;

		Asura_Vector_3 xPos0 = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertices, xStrip.m_usVertexFormat, xVertexPosOffset);
		Asura_Vector_3 xPos1 = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertices + xVertexSize, xStrip.m_usVertexFormat, xVertexPosOffset);

		if (bReflected)
		{
			xPos0.y *= -1.f;
			xPos1.y *= -1.f;
		}

		ResetPlaneBitfields(xVertexFrustumTest);
		ResetPlaneBitfields(xVertexGuardbandTest);

		// Check position 0 against frustum
		TestPointAgainstFrustum(xPos0, m_xFrustum, xVertexFrustumTest);
		TestPointAgainstFrustum(xPos0, m_xGuardband, xVertexGuardbandTest);

		ShiftPlaneBitfields(xVertexFrustumTest);
		ShiftPlaneBitfields(xVertexGuardbandTest);

		// Check position 1 against frustum
		TestPointAgainstFrustum(xPos1, m_xFrustum, xVertexFrustumTest);
		TestPointAgainstFrustum(xPos1, m_xGuardband, xVertexGuardbandTest);

		bool bForceClipFirstVertex = Asura_Maths::Abs(((xPos0 - *pxCamPos)) * xCameraFrontPlane.m_xNormal) < dbg_fForcedClippedDistance; 

		//u_int uNextVertexIndex = 2;

		//u_int uFirst = xStrip.m_auStartIndex[uInternalStrips];
		u_int uFirst = 0;
		u_int uLast = 0;
		// Last iteration?
		if (uInternalStrips != xStrip.m_uNumberOfStrips - 1)
		{
			// loop for the rest of the strip
			uLast = xStrip.m_auStartIndex[uInternalStrips + 1] - xStrip.m_auStartIndex[uInternalStrips] - 2;
		}
		else
		{
			// loop for the rest of triangles
			uLast = xStrip.m_uNumberOfIndices - xStrip.m_auStartIndex[uInternalStrips] - 2;
		}

		bool bForceClipSecondVertex = Asura_Maths::Abs(((xPos1 - *pxCamPos)) * xCameraFrontPlane.m_xNormal) < dbg_fForcedClippedDistance;
		bool bForceClipThirdVertex = false; 
		bool bForceClipTriangles = false; 

		// Calculate which part of the strip needs to be draw
		u_int uFirstTrianglePending = uFirst; 
		u_int uLastTrianglePending = uLast+2; 

		for(u_int uTriangle = uFirst; uTriangle < uLast; ++uTriangle, pucVertices += xVertexSize)
		{
			// Check for degeneracy
			/* if(pxVertices[1] == pxVertices[2])
			{
			// There are always 4 degenerates.
			uTriangle += 3;
			pxVertices += 3;
			// Update frustum info buffer
			Asura_Vector_3 xPos1(pxVertices[1].m_afPosition[0], pxVertices[1].m_afPosition[1], pxVertices[1].m_afPosition[2]);
			Asura_Vector_3 xPos2(pxVertices[2].m_afPosition[0], pxVertices[2].m_afPosition[1], pxVertices[2].m_afPosition[2]);
			TestPointAgainstFrustum(xPos1, m_xFrustum, xVertexFrustumTest[uNextVertexIndex]);
			TestPointAgainstFrustum(xPos1, m_xGuardband, xVertexGuardbandTest[uNextVertexIndex]);
			uNextVertexIndex = (uNextVertexIndex + 1) % 3;
			TestPointAgainstFrustum(xPos2, m_xFrustum, xVertexFrustumTest[uNextVertexIndex]);
			TestPointAgainstFrustum(xPos2, m_xGuardband, xVertexGuardbandTest[uNextVertexIndex]);
			uNextVertexIndex = (uNextVertexIndex + 1) % 3;
			continue;
			} */

			m_uNumTriangleTests ++;
			m_uMaxStripLength = Asura_Maths::Max(m_uMaxStripLength, uLast - uFirst);

			Asura_Vector_3 xPos2 = Asura_PSP_VertexFormat::GetPositionFromVertex(pucVertices + (xVertexSize * 2), xStrip.m_usVertexFormat, xVertexPosOffset);

			bForceClipThirdVertex = Asura_Maths::Abs(((xPos2 - *pxCamPos)) * xCameraFrontPlane.m_xNormal) < dbg_fForcedClippedDistance;
			bForceClipTriangles = bForceClipFirstVertex || bForceClipSecondVertex || bForceClipThirdVertex; 
			bForceClipFirstVertex = bForceClipSecondVertex; 
			bForceClipSecondVertex = bForceClipThirdVertex; 

			if (bReflected)
			{
				xPos2.y *= -1.f;
			}

			ShiftPlaneBitfields(xVertexFrustumTest);
			ShiftPlaneBitfields(xVertexGuardbandTest);

			TestPointAgainstFrustum(xPos2, m_xFrustum, xVertexFrustumTest);
			TestPointAgainstFrustum(xPos2, m_xGuardband, xVertexGuardbandTest);

			// Rotate next vertex
			//uNextVertexIndex = (uNextVertexIndex + 1) % 3;

			// All outside frustum
			bool bContinue = false;
			if (xVertexFrustumTest.m_xPlane.m_uiPlaneLeft == 0x7)
			{
				bContinue = true;
			}
			else if (xVertexFrustumTest.m_xPlane.m_uiPlaneRight == 0x7)
			{
				bContinue = true;
			}
			else if (xVertexFrustumTest.m_xPlane.m_uiPlaneDown == 0x7)
			{
				bContinue = true;
			}
			else if (xVertexFrustumTest.m_xPlane.m_uiPlaneUp == 0x7)
			{
				bContinue = true;
			}

			if (bContinue) continue;

			/*
			// All inside frustum
			bool bAllInside = true;

			//if((xVertexFrustumTest[0].m_bL || xVertexFrustumTest[1].m_bL || xVertexFrustumTest[2].m_bL))
			if (xVertexFrustumTest.m_xPlane.m_uiPlaneLeft)
			{
			bAllInside = false;

			}

			//if((xVertexFrustumTest[0].m_bR || xVertexFrustumTest[1].m_bR || xVertexFrustumTest[2].m_bR))
			if (xVertexFrustumTest.m_xPlane.m_uiPlaneRight)
			{
			bAllInside = false;
			}

			//if((xVertexFrustumTest[0].m_bD || xVertexFrustumTest[1].m_bD || xVertexFrustumTest[2].m_bD))
			if (xVertexFrustumTest.m_xPlane.m_uiPlaneDown)
			{
			bAllInside = false;
			}

			//if((xVertexFrustumTest[0].m_bU || xVertexFrustumTest[1].m_bU || xVertexFrustumTest[2].m_bU))
			if (xVertexFrustumTest.m_xPlane.m_uiPlaneUp)
			{
			bAllInside = false;
			}

			if(bAllInside) continue;
			*/

			bool bAllInside =	!(xVertexFrustumTest.m_xPlane.m_uiPlaneLeft		||
				xVertexFrustumTest.m_xPlane.m_uiPlaneRight	||
				xVertexFrustumTest.m_xPlane.m_uiPlaneUp		||
				xVertexFrustumTest.m_xPlane.m_uiPlaneDown);

			if (bAllInside) continue;


			// Now we have a triangle that's definitely crossing the frustum
			// Test whether it goes outside the guardband
			//bool bTestTriL = bTestL && (xVertexGuardbandTest[0].m_bL || xVertexGuardbandTest[1].m_bL || xVertexGuardbandTest[2].m_bL);
			//bool bTestTriR = bTestR && (xVertexGuardbandTest[0].m_bR || xVertexGuardbandTest[1].m_bR || xVertexGuardbandTest[2].m_bR);
			//bool bTestTriD = bTestD && (xVertexGuardbandTest[0].m_bD || xVertexGuardbandTest[1].m_bD || xVertexGuardbandTest[2].m_bD);
			//bool bTestTriU = bTestU && (xVertexGuardbandTest[0].m_bU || xVertexGuardbandTest[1].m_bU || xVertexGuardbandTest[2].m_bU);

			bool bTestTriL = bTestL && (xVertexGuardbandTest.m_xPlane.m_uiPlaneLeft);
			bool bTestTriR = bTestR && (xVertexGuardbandTest.m_xPlane.m_uiPlaneRight);
			bool bTestTriD = bTestD && (xVertexGuardbandTest.m_xPlane.m_uiPlaneDown);
			bool bTestTriU = bTestU && (xVertexGuardbandTest.m_xPlane.m_uiPlaneUp);


			if (bForceClipTriangles)
			{
				bTestTriL = xVertexFrustumTest.m_xPlane.m_uiPlaneLeft; 
				bTestTriR = xVertexFrustumTest.m_xPlane.m_uiPlaneRight; 
				bTestTriD = xVertexFrustumTest.m_xPlane.m_uiPlaneDown; 
				bTestTriU = xVertexFrustumTest.m_xPlane.m_uiPlaneUp; 
			}

			if(!(bTestTriL || bTestTriR || bTestTriD || bTestTriU)) continue;

			u_int uCurrentVertexBuffer = 0;

			xVertexBuffer[0][0].Uncompress(pucVertices, xStrip.m_usVertexFormat);
			xVertexBuffer[0][1].Uncompress(pucVertices + xVertexSize, xStrip.m_usVertexFormat);
			xVertexBuffer[0][2].Uncompress(pucVertices + (xVertexSize * 2), xStrip.m_usVertexFormat);

			if (bReflected)
			{
				xVertexBuffer[0][0].m_afPosition[1] *= -1.f;
				xVertexBuffer[0][1].m_afPosition[1] *= -1.f;
				xVertexBuffer[0][2].m_afPosition[1] *= -1.f;
			}

			u_int uNumVerts = 3;

			bool bIntersected = false;

			// Down most likely to discard, so do first. (TODO - statistical analysis of most discards)
			if(bTestTriD)
			{
				Asura_PSP_Plane xPlane;
				//			xPlane.m_xNormal	= -m_xGuardband.m_xDown.m_xNormal;
				//			xPlane.m_fPosition	=  m_xGuardband.m_xDown.m_fPosition - .1f;
				xPlane.m_xNormal	= -m_xFrustum.m_xDown.m_xNormal;
				xPlane.m_fPosition	=  m_xFrustum.m_xDown.m_fPosition;
				bIntersected |= ClipPolyToPlane(xVertexBuffer[uCurrentVertexBuffer], uNumVerts, xVertexBuffer[!uCurrentVertexBuffer], xPlane);
				uCurrentVertexBuffer = !uCurrentVertexBuffer;

				if(!uNumVerts) continue;
			}
			if(bTestTriL)
			{
				Asura_PSP_Plane xPlane;
				//			xPlane.m_xNormal	= -m_xGuardband.m_xLeft.m_xNormal;
				//			xPlane.m_fPosition	=  m_xGuardband.m_xLeft.m_fPosition - .1f;
				xPlane.m_xNormal	= -m_xFrustum.m_xLeft.m_xNormal;
				xPlane.m_fPosition	=  m_xFrustum.m_xLeft.m_fPosition;
				bIntersected |= ClipPolyToPlane(xVertexBuffer[uCurrentVertexBuffer], uNumVerts, xVertexBuffer[!uCurrentVertexBuffer], xPlane);
				uCurrentVertexBuffer = !uCurrentVertexBuffer;

				if(!uNumVerts) continue;
			}
			if(bTestTriR)
			{
				Asura_PSP_Plane xPlane;
				//			xPlane.m_xNormal	= -m_xGuardband.m_xRight.m_xNormal;
				//			xPlane.m_fPosition	=  m_xGuardband.m_xRight.m_fPosition - .1f;
				xPlane.m_xNormal	= -m_xFrustum.m_xRight.m_xNormal;
				xPlane.m_fPosition	=  m_xFrustum.m_xRight.m_fPosition;
				bIntersected |= ClipPolyToPlane(xVertexBuffer[uCurrentVertexBuffer], uNumVerts, xVertexBuffer[!uCurrentVertexBuffer], xPlane);
				uCurrentVertexBuffer = !uCurrentVertexBuffer;

				if(!uNumVerts) continue;
			}
			// Up is least likely to discard, so do last
			if(bTestTriU)
			{
				Asura_PSP_Plane xPlane;
				//			xPlane.m_xNormal	= -m_xGuardband.m_xUp.m_xNormal;
				//			xPlane.m_fPosition	=  m_xGuardband.m_xUp.m_fPosition - .1f;
				xPlane.m_xNormal	= -m_xFrustum.m_xUp.m_xNormal;
				xPlane.m_fPosition	=  m_xFrustum.m_xUp.m_fPosition;
				bIntersected |= ClipPolyToPlane(xVertexBuffer[uCurrentVertexBuffer], uNumVerts, xVertexBuffer[!uCurrentVertexBuffer], xPlane);
				uCurrentVertexBuffer = !uCurrentVertexBuffer;

				if(!uNumVerts) continue;
			}
			if(!bIntersected) continue;// nothing clipped - don't bother

			++m_uNumTriangleAccepts;

			Asura_Assert(g_uNumClippedVerts + (uNumClippedVerts + uNumVerts + 2) <= ASURA_PSP_MAX_CLIPPED_VERTS, "Too many clipped verts this frame");
			if(g_uNumClippedVerts + (uNumClippedVerts + uNumVerts + 2) > ASURA_PSP_MAX_CLIPPED_VERTS) continue;

			Asura_PSP_RenderEnvironment_VertexClip* pxVerticesOut = &g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][g_uNumClippedVerts + uNumClippedVerts + 1];
			// Double first for degenerate
			pxVerticesOut[-1] = xVertexBuffer[uCurrentVertexBuffer][0];
			// Copy all verts, reordering from fan to strip
			{
				u_int u = 0;
				u_int v = 0;
				u_int w = uNumVerts-1;

				pxVerticesOut[u++] = xVertexBuffer[uCurrentVertexBuffer][v++];

				while(true)
				{
					pxVerticesOut[u++] = xVertexBuffer[uCurrentVertexBuffer][v++];
					if (u == uNumVerts) break;

					pxVerticesOut[u++] = xVertexBuffer[uCurrentVertexBuffer][w--];
					if (u == uNumVerts) break;
				}
				// Double last for degenerate
				pxVerticesOut[uNumVerts] = pxVerticesOut[uNumVerts - 1];

				if (bReflected)
				{
					// Reflect back the clipped strip
					for(int iVert = -1; iVert < int(uNumVerts + 1); ++iVert)
					{
						pxVerticesOut[iVert].m_afPosition[1] *= -1.f;
					}
				}

			}

			uNumClippedVerts += uNumVerts + 2;

			// Draw the pending list up to this triangle and update the first pending triangle
			if (dbg_bDrawRemainingClippedStrips && dbg_bDontRenderCompleteStripsBeforeClipping)
			{
				DrawInternalStripPartial(xStrip, uInternalStrips, uFirstTrianglePending, uTriangle+2, xVertexSize); 
				uFirstTrianglePending = uTriangle+1;
			}
		}

		if (dbg_bDrawRemainingClippedStrips && dbg_bDontRenderCompleteStripsBeforeClipping)
		{
			DrawInternalStripPartial(xStrip, uInternalStrips, uFirstTrianglePending, uLastTrianglePending, xVertexSize); 
		}

		const u_int uVertexFormatUncompressedColour = SCEGU_TEXTURE_USHORT|SCEGU_COLOR_PF8888|SCEGU_NORMAL_SHORT|SCEGU_VERTEX_FLOAT;

		Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
		sceKernelDcacheWritebackRange((char*)&g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][g_uNumClippedVerts], sizeof(Asura_PSP_RenderEnvironment_VertexClip) * uNumClippedVerts);
		s_iDepthBufferOffset	+= dbg_iClippingDepthBufferOffset;
		Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);
		DrawArray
			(
			SCEGU_PRIM_TRIANGLE_STRIP,
			uVertexFormatUncompressedColour,
			uNumClippedVerts,
			NULL,
			g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ] + g_uNumClippedVerts
			);

		// if debugvar is set, expand clipped polys slightly to cover tears
		if (dbg_bExpandClippedPolys && uNumClippedVerts)
		{
			Asura_Assert(g_uNumClippedVerts + (2 * uNumClippedVerts) <= ASURA_PSP_MAX_CLIPPED_VERTS, "Too many clipped verts this frame to do tear-fixing");
		}
		if (dbg_bExpandClippedPolys && uNumClippedVerts && (g_uNumClippedVerts + (2 * uNumClippedVerts) <= ASURA_PSP_MAX_CLIPPED_VERTS))
		{
			const Asura_PSP_RenderEnvironment_VertexClip* pxVerticesIn = &g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][g_uNumClippedVerts];
			Asura_PSP_RenderEnvironment_VertexClip* pxVerticesOut = &g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][g_uNumClippedVerts + uNumClippedVerts];
			for (u_int uCopy = 0; uCopy < uNumClippedVerts; ++uCopy)
			{
				pxVerticesOut[uCopy] = pxVerticesIn[uCopy]; 

				// tear covering, purple
				if (dbg_bShowExpandedClippedPolys)
				{
					pxVerticesOut[uCopy].m_iRGBA8888 = 0xffff00ff;
				}
			}

			bool bRestart = true; 

			for (u_int uVert = 2; uVert < uNumClippedVerts; ++uVert)
			{
				// Expand it along the triangle radius
				Asura_Vector_3 xVertexA(pxVerticesIn[uVert  ].m_afPosition[0], pxVerticesIn[uVert  ].m_afPosition[1], pxVerticesIn[uVert  ].m_afPosition[2]); 
				Asura_Vector_3 xVertexB(pxVerticesIn[uVert-1].m_afPosition[0], pxVerticesIn[uVert-1].m_afPosition[1], pxVerticesIn[uVert-1].m_afPosition[2]); 
				Asura_Vector_3 xVertexC(pxVerticesIn[uVert-2].m_afPosition[0], pxVerticesIn[uVert-2].m_afPosition[1], pxVerticesIn[uVert-2].m_afPosition[2]); 

				Asura_Vector_3 xCenter = (xVertexA + xVertexB) / 2.0f; 
				xCenter = (xCenter + xVertexC) / 2.0f;

				if ((xVertexA != xVertexB) && (xVertexB != xVertexC) && (xVertexA != xVertexC))
				{
					if (bRestart)
					{		
						Asura_Vector_3 xExpansionB = xVertexB - xCenter; 
						xExpansionB *= dbg_fExpansionRatio; 
						Asura_Vector_3 xVertexBExpanded = xCenter + xExpansionB; 

						pxVerticesOut[uVert-1].m_afPosition[0] = xVertexBExpanded.GetX(); 
						pxVerticesOut[uVert-1].m_afPosition[1] = xVertexBExpanded.GetY(); 
						pxVerticesOut[uVert-1].m_afPosition[2] = xVertexBExpanded.GetZ(); 

						Asura_Vector_3 xExpansionC = xVertexC - xCenter; 
						xExpansionC *= dbg_fExpansionRatio; 
						Asura_Vector_3 xVertexCExpanded = xCenter + xExpansionC; 

						pxVerticesOut[uVert-2].m_afPosition[0] = xVertexCExpanded.GetX(); 
						pxVerticesOut[uVert-2].m_afPosition[1] = xVertexCExpanded.GetY(); 
						pxVerticesOut[uVert-2].m_afPosition[2] = xVertexCExpanded.GetZ(); 

						// Check if the previous one was a degenerate jump
						if (uVert > 2)
						{
							Asura_Vector_3 xVertexX(pxVerticesIn[uVert-3].m_afPosition[0], pxVerticesIn[uVert-3].m_afPosition[1], pxVerticesIn[uVert-3].m_afPosition[2]); 
							if (xVertexC == xVertexX)
							{
								pxVerticesOut[uVert-3].m_afPosition[0] = xVertexCExpanded.GetX(); 
								pxVerticesOut[uVert-3].m_afPosition[1] = xVertexCExpanded.GetY(); 
								pxVerticesOut[uVert-3].m_afPosition[2] = xVertexCExpanded.GetZ(); 
							}
						}

						bRestart = false;

					}

					Asura_Vector_3 xExpansionA = xVertexA - xCenter; 
					xExpansionA *= dbg_fExpansionRatio; 
					Asura_Vector_3 xVertexExpanded = xCenter + xExpansionA; 
					
					// Set the new output with the expanded clipped triangles
					pxVerticesOut[uVert].m_afPosition[0] = xVertexExpanded.GetX(); 
					pxVerticesOut[uVert].m_afPosition[1] = xVertexExpanded.GetY(); 
					pxVerticesOut[uVert].m_afPosition[2] = xVertexExpanded.GetZ(); 

					// Check if the next one is a degenerate jump
					if (uVert < uNumClippedVerts-1)
					{
						Asura_Vector_3 xVertexX(pxVerticesIn[uVert+1].m_afPosition[0], pxVerticesIn[uVert+1].m_afPosition[1], pxVerticesIn[uVert+1].m_afPosition[2]); 
						if (xVertexA == xVertexX)
						{
							pxVerticesOut[uVert+1].m_afPosition[0] = xVertexExpanded.GetX(); 
							pxVerticesOut[uVert+1].m_afPosition[1] = xVertexExpanded.GetY(); 
							pxVerticesOut[uVert+1].m_afPosition[2] = xVertexExpanded.GetZ(); 
						}
					}
				}
				else
				{
					bRestart = true; 
				}

			}

			sceKernelDcacheWritebackRange((char*)pxVerticesOut, sizeof(Asura_PSP_RenderEnvironment_VertexClip) * uNumClippedVerts);
			Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_NOT_EQUAL, ASURA_PSP_STENCIL_NONE, 0xFF);
			bool bIsZWriteEnabled = Asura_Render::CurrentStates.GetZBufferWriteEnabled();
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false); 
			DrawArray
				(
				SCEGU_PRIM_TRIANGLE_STRIP,
				uVertexFormatUncompressedColour,
				uNumClippedVerts,
				0,
				pxVerticesOut
				);
			Asura_PSP_StencilBufferManager::SetStencilFunction(ASURA_PSP_STENCIL_FUNCTION_ALWAYS, ASURA_PSP_STENCIL_NONE, 0xFF);
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled(bIsZWriteEnabled); 
		}

		// if debugvar is set, turn clipped polys a different colour
		if (dbg_bShowClippedPolys && uNumClippedVerts)
		{
			Asura_Assert(g_uNumClippedVerts + (3 * uNumClippedVerts) <= ASURA_PSP_MAX_CLIPPED_VERTS, "Too many clipped verts this frame to show clipped polys");
		}
		if (dbg_bShowClippedPolys && (g_uNumClippedVerts + (3 * uNumClippedVerts) <= ASURA_PSP_MAX_CLIPPED_VERTS))
		{
			Asura_PSP_RenderEnvironment_VertexClip* pxVerticesOut = &g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][g_uNumClippedVerts + 2 * uNumClippedVerts];
			for (u_int uVert = 0; uVert < uNumClippedVerts; ++uVert)
			{
				pxVerticesOut[uVert] = g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][g_uNumClippedVerts+uVert]; 

				if (bSecondPass)
				{
					// second pass, blue
					pxVerticesOut[uVert].m_iRGBA8888 = 0xffff0000;
				}
				else
				{
					// normal pass, red
					pxVerticesOut[uVert].m_iRGBA8888 = 0xff0000ff;
				}

				// Degenerate jumps, green
				if (uVert >= 3)
				{
					if (pxVerticesOut[uVert].m_afPosition[0] == pxVerticesOut[uVert-1].m_afPosition[0] && pxVerticesOut[uVert].m_afPosition[1] == pxVerticesOut[uVert-1].m_afPosition[1] && pxVerticesOut[uVert].m_afPosition[2] == pxVerticesOut[uVert-1].m_afPosition[2] && 
						pxVerticesOut[uVert-2].m_afPosition[0] == pxVerticesOut[uVert-3].m_afPosition[0] && pxVerticesOut[uVert-2].m_afPosition[1] == pxVerticesOut[uVert-3].m_afPosition[1] && pxVerticesOut[uVert-2].m_afPosition[2] == pxVerticesOut[uVert-3].m_afPosition[2] &&
						!(pxVerticesOut[uVert-1].m_afPosition[0] == pxVerticesOut[uVert-2].m_afPosition[0] && pxVerticesOut[uVert-1].m_afPosition[1] == pxVerticesOut[uVert-2].m_afPosition[1] && pxVerticesOut[uVert-1].m_afPosition[2] == pxVerticesOut[uVert-2].m_afPosition[2]))
					{
						
						pxVerticesOut[uVert-1].m_iRGBA8888 = 0xff00ff00;
						pxVerticesOut[uVert-2].m_iRGBA8888 = 0xff00ff00;
					}
				}
			}

			sceKernelDcacheWritebackRange((char*)pxVerticesOut, sizeof(Asura_PSP_RenderEnvironment_VertexClip) * uNumClippedVerts);
			s_iDepthBufferOffset	+= 10;
			Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);
			DrawArray
				(
				SCEGU_PRIM_LINE_STRIP,
				uVertexFormatUncompressedColour,
				uNumClippedVerts,
				0,
				pxVerticesOut
				);
			s_iDepthBufferOffset	-= 10;
			Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);

		}

		s_iDepthBufferOffset	-= dbg_iClippingDepthBufferOffset;
		Asura_PSP_RenderStates::RequestDepthOffset(s_iDepthBufferOffset);
		g_uNumClippedVerts += uNumClippedVerts;
		if (dbg_bExpandClippedPolys)
		{
			g_uNumClippedVerts += uNumClippedVerts;
		}
		if (dbg_bShowClippedPolys)
		{
			g_uNumClippedVerts += uNumClippedVerts;
		}
		

		Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);
#endif	// ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
	}

#ifdef ASURA_PSP_RENDERENVIRONMENT_QUADCLIPPING
	if( uTotalClippedVertices )
	{
		// Depth offset
		s_iDepthBufferOffset += dbg_iClippingDepthBufferOffset;
		Asura_PSP_RenderStates::RequestDepthOffset( s_iDepthBufferOffset );

		// Draw
		DrawArray(
			SCEGU_PRIM_TRIANGLE_STRIP,
			SCEGU_TEXTURE_USHORT | SCEGU_COLOR_PF8888 | SCEGU_NORMAL_SHORT | SCEGU_VERTEX_FLOAT,
			uTotalClippedVertices,
			NULL,
			&g_xClippedVerts[ Asura_PSP_Graphics::GetCurrentDisplayList() ][ g_uNumClippedVerts ]
		);

		// Restore depth offset
		s_iDepthBufferOffset -= dbg_iClippingDepthBufferOffset;
		Asura_PSP_RenderStates::RequestDepthOffset( s_iDepthBufferOffset );

		g_uNumClippedVerts += uTotalClippedVertices;
	}
#endif
}

// JG - 20/07/06: If we can guarantee the the matrix stack is not going to change while we are rendering the environment (all is in the same space)
// we can speed up GUM_DrawArray with this fastest version (no state checking, or whatever the library is doing inside)
// Could be optimised even more, but right now I'm not going to bother
// VTYPE is quite unlikely to change, could set this before first pass / dual pass, but the clipped triangles are using a different format 
// (uncompressed) so they will need fixing
// BASE / VADR doesn't need to be changed if this function is called for consecutive strips on the same vertex buffer pool (GU register gets incremented automatically)
// BASE doesn't have to be set again if we are on the same 'region' than before
void Asura_PSP_RenderEnvironment_Zone::DrawArray(unsigned int uPrimType, unsigned int uVertexFormat, unsigned int uVertices, void* pIndices, void* pBuffer) 
{
#ifdef ASURA_PSP_USE_FAST_DRAW_ARRAY
	// Get a pointer to the display list buffer
	u_int	puDisplayListAddress =	(u_int)	GU_GetMemory(2 * sizeof(u_int));
	
	// Use uncached memory access
	u_int*	puDisplayList =			(u_int*)(puDisplayListAddress | 0x40000000); 

	// Write ONLY the commands needed
	puDisplayList[-2]	= (0x12000000) | ((0x00FFFFFF) & uVertexFormat);					// CMD_VTYPE: Set the vertex format (could be set less often, i.e. first pass, dual pass)
	puDisplayList[-1]	= (0x10000000) | ((0x00FF0000) & ((unsigned int)pBuffer >> 8));		// CMD_BASE: Set the vertex buffer base address higher bits
	puDisplayList[ 0]	= (0x01000000) | ((0x00FFFFFF) & (unsigned int)pBuffer);			// CMD_VADR: Set the vertex buffer base address lower bits 
	puDisplayList[ 1]	= (0x04000000) | (uPrimType << 16) | ((0x0000FFFF) & uVertices);	// CMD_PRIM: Set the primitive to be render (actual draw function)
#else
	GUM_DrawArray(uPrimType, uVertexFormat, uVertices, pIndices, pBuffer); 
#endif // ASURA_PSP_USE_FAST_DRAW_ARRAY
}

static inline bool PointIsAbovePlane_C(
									   register const Asura_Vector_3& xPo,
									   register const Asura_PSP_Plane& xPl
									   )
{
	register const float fDotProduct = xPo * xPl.m_xNormal;
	return xPl.m_fPosition < fDotProduct;
}

#if defined __GNUC__ && !defined __SNC__
#include <allegrex.h>
#include <vfpu_intrinsics.h>
#endif

static inline bool PointIsAbovePlane(
									 register const Asura_Vector_3& xPo,
									 register const Asura_PSP_Plane& xPl
									 )
{
#if defined __GNUC__ && !defined __SNC__
	register const Asura_Vector_3* const xNv = &xPl.m_xNormal;
	register const vfpu_v4 afNv
		= vfpu_insert_float_into_v4xyz(afNv,xNv->x,xNv->y,xNv->z);
	register const vfpu_v4 afPo
		= vfpu_insert_float_into_v4xyz(afPo,xPo.x,xPo.y,xPo.z);

	return xPl.m_fPosition <
		vfpu_extract_float_from_v4x(vfpu_vdot_t(afPo,afNv,0));
#else
	return PointIsAbovePlane_C(xPo,xPl);
#endif
}

static inline bool BBIsCompletelyAbovePlane_C(
	register const Asura_Bounding_Box& xBB,
	register const Asura_PSP_Plane& xPl
	)
{
#if 1
	register const Asura_Vector_3 xFarPoint(
		(xPl.m_xNormal.GetX() < 0.f) ? xBB.MaxX : xBB.MinX,
		(xPl.m_xNormal.GetY() < 0.f) ? xBB.MaxY : xBB.MinY,
		(xPl.m_xNormal.GetZ() < 0.f) ? xBB.MaxZ : xBB.MinZ
		);

	return PointIsAbovePlane_C(xFarPoint,xPl);
#else
	/* Test all 8 vertices of the AABB */
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MaxX, xBB.MaxY, xBB.MaxZ), xPl)) return false;
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MaxX, xBB.MaxY, xBB.MinZ), xPl)) return false;
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MaxX, xBB.MinY, xBB.MaxZ), xPl)) return false;
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MaxX, xBB.MinY, xBB.MinZ), xPl)) return false;

	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MinX, xBB.MaxY, xBB.MaxZ), xPl)) return false;
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MinX, xBB.MaxY, xBB.MinZ), xPl)) return false;
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MinX, xBB.MinY, xBB.MaxZ), xPl)) return false;
	if(!PointIsAbovePlane_C(Asura_Vector_3(xBB.MinX, xBB.MinY, xBB.MinZ), xPl)) return false;
	return true;
#endif
}

static u_int BBIsCompletelyAboveAnyFrustumPlane_C(
	register const Asura_Bounding_Box& xBB,
	register const Asura_PSP_FrustumLimits& xFrustum
	)
{
	return
		(BBIsCompletelyAbovePlane_C(xBB, xFrustum.m_xLeft))
		? ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_L

		: (BBIsCompletelyAbovePlane_C(xBB, xFrustum.m_xRight))
		? ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_R

		: (BBIsCompletelyAbovePlane_C(xBB, xFrustum.m_xUp))
		? ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_U

		: (BBIsCompletelyAbovePlane_C(xBB, xFrustum.m_xDown))
		? ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_D

		: 0;
}

#if defined __GNUC__ && !defined __SNC__
/*
Return a vector whose elements are taken from 'xThen' if the
corresponding element of 'xA' is less than 'xB', else from 'xElse'.

Therefore 'vfpu_choose_lt_t(a,b,a,b)' is then elementwise minimum,
'vfpu_choose_lt_t(a,b,b,a)' is elementwise maximum.
*/
static inline vfpu_v4 vfpu_vchooselt_t(
									   register const vfpu_v4 xA,register const vfpu_v4 xB,
									   register const vfpu_v4 xThen,register const vfpu_v4 xElse
									   )
{
	register vfpu_v4 xChosen;

	__asm__
		(
		ASURA_PSP_ASM_NRBEGIN

		/*
		Conditions specifier 6 means elementwise move
		*/
		"vcmp.t		LT, %1, %2 \n\t"	// xA < xB?
		"vmov.t		%0, %3 \n\t"		// xChosen = xThen
		"vcmovf.t	%0, %4, 6 \n\t"		// xChosen = xElse for each !LT

		ASURA_PSP_ASM_NREND

		: "=&j" (xChosen)
		: "j" (xA), "j" (xB), "j" (xThen), "j" (xElse)
		);

	return xChosen;
}

static inline unsigned vfpu_vlt_t(
								  register const vfpu_v4 xA,register const vfpu_v4 xB
								  )
{
	unsigned uResult;

	__asm__
		(
		"vcmp.q		LT, %1, %2\n\t"
		"mfvc		%0, $131"

		: "=r" (uResult)
		: "j" (xA), "j" (xB)
		);

	return uResult & 0x0f;
}
#endif

// 5-8% profiler

/*
We have a bounding box and a frustum with four planes.

From the four planes we compute four ''far'' points; the nth element
of each point is the nth element of the max of the BB if the nth
element of the normal is negative, else the nth element of the min
of the BB.

Then we test whether each point is above the corresponding plane,
and if it is we return a plane-dependent code, if none, we return 0.

We can test for each element serially or in parallel. If serially we
need only fetch the plane data as needed, else we need to fetch the
data for all four planes. Since fetching the normal and position of
each plane is 4 bytes for each, and has a latency of 7 cycles, we
might as well load all of them.

The definition of a frustum limits object at the beginning of its file
implies its layout is:

00 m_xLeft.m_xNormal.x
04 m_xLeft.m_xNormal.y
08 m_xLeft.m_xNormal.z
12 m_xLeft.m_xfPosition
16 m_xRight.m_xNormal.x
20 m_xRight.m_xNormal.y
24 m_xRight.m_xNormal.z
28 m_xRight.m_xfPosition
32 m_xUp.m_xNormal.x
36 m_xUp.m_xNormal.y
40 m_xUp.m_xNormal.z
44 m_xUp.m_xfPosition
48 m_xDown.m_xNormal.x
52 m_xDown.m_xNormal.y
56 m_xDown.m_xNormal.z
60 m_xDown.m_xfPosition

In the definition of a bounding box the layout looks like:

00 MinX
04 MaxX
08 MinY
12 MaxY
16 MinZ
20 MaxZ

In the assembler inline we actually use the numeric offsets,
in the VFPU intrinsics we take care to perform operations in
ascending address order if possible.

Also, as to the results, we have:

ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_U		0x01
ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_D		0x02
ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_L		0x04
ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_R		0x08
*/

static u_int BBIsCompletelyAboveAnyFrustumPlane(
	const Asura_Bounding_Box& xBB,
	const Asura_PSP_FrustumLimits& xFr
	)
{
#if 0 && defined __GNUC__ && !defined __SNC__
#error GCC intrinsics version not debugged yet
	u_int uResult;

	register const vfpu_v4 afZero = vfpu_vzero_q(0);

	/*
	Gather the minimum and maximum points of the BB into two
	registers.

	The BB coordinates are in increasing address by dimension
	(min x, max x, min y, max y, min z, max z), but we don't
	care much here, hoping a cache line encompasses them.
	*/
	register const vfpu_v4 afBoxMin = vfpu_assign_float_to_v4xyzw
		(xBB.MinX,xBB.MinY,xBB.MinZ,0.0f);
	register const vfpu_v4 afBoxMax = vfpu_assign_float_to_v4xyzw
		(xBB.MaxX,xBB.MaxY,xBB.MaxZ,0.0f);

	/*
	Gather the elements of the four plane normals into four
	registers.

	We need to test planes in [up,down,left,right] order because
	that's the order of bits we need to return, but we load their
	normals in [left,right,up,down] order because that's how they
	are laid out in memory.
	*/
	register const vfpu_v4 afPlaneL
		= vfpu_assign_float_to_v4xyzw(xFr.m_xLeft.m_xNormal.x,
		xFr.m_xLeft.m_xNormal.y,xFr.m_xLeft.m_xNormal.z,0.0f);
	register const vfpu_v4 afPlaneR
		= vfpu_assign_float_to_v4xyzw(xFr.m_xRight.m_xNormal.x,
		xFr.m_xRight.m_xNormal.y,xFr.m_xRight.m_xNormal.z,0.0f);
	register const vfpu_v4 afPlaneU
		= vfpu_assign_float_to_v4xyzw(xFr.m_xUp.m_xNormal.x,
		xFr.m_xUp.m_xNormal.y,xFr.m_xUp.m_xNormal.z,0.0f);
	register const vfpu_v4 afPlaneD
		= vfpu_assign_float_to_v4xyzw(xFr.m_xDown.m_xNormal.x,
		xFr.m_xDown.m_xNormal.y,xFr.m_xDown.m_xNormal.z,0.0f);

	/*
	Gather the positions of the four planes into a single register.
	*/
	register const vfpu_v4 afPositions = vfpu_assign_float_to_v4xyzw
		(xFr.m_xUp.m_fPosition,xFr.m_xDown.m_fPosition,
		xFr.m_xLeft.m_fPosition,xFr.m_xRight.m_fPosition);

	/*
	Select into registers the ''farthest'' points of the BB wrt
	each plane.
	*/
	register const vfpu_v4
		afFarU = vfpu_vchooselt_t(afPlaneU,afZero,afBoxMax,afBoxMin);
	register const vfpu_v4
		afFarD = vfpu_vchooselt_t(afPlaneD,afZero,afBoxMax,afBoxMin);
	register const vfpu_v4
		afFarL = vfpu_vchooselt_t(afPlaneL,afZero,afBoxMax,afBoxMin);
	register const vfpu_v4
		afFarR = vfpu_vchooselt_t(afPlaneR,afZero,afBoxMax,afBoxMin);

	/*
	Compute the distances between farthest point of the BB and
	each plane into a register.
	*/
	register const vfpu_v4 afDistances
		= vfpu_assign_float_to_v4xyzw(
		vfpu_extract_float_from_v4x(vfpu_vdot_t(afFarU,afPlaneU,0)),
		vfpu_extract_float_from_v4x(vfpu_vdot_t(afFarD,afPlaneD,0)),
		vfpu_extract_float_from_v4x(vfpu_vdot_t(afFarL,afPlaneL,0)),
		vfpu_extract_float_from_v4x(vfpu_vdot_t(afFarR,afPlaneR,0))
		);

	uResult = vfpu_vlt_t(afPositions,afDistances);

#if PG_DEBUG
	{
		const unsigned uGood = BBIsCompletelyAboveAnyFrustumPlane_C(xBB,xFr);
		if (uResult != uGood)
		{
			sceKernelPrintf("our result: 0x%x, right result: 0x%x\n",uResult,uGood);			
			__asm__("break 1");
		}
	}
#endif

	return uResult;
#elif defined __SNC__
	u_int uResult;


	__asm__ volatile
		(
		".set push\n"				// save assember option
		".set noreorder\n"			// suppress reordering (we fill our own delay slots)

		/*
		In the loading we have these latencies:

		vzero latency:	3
		lv latency:		3
		*/

		"vzero.t c600 \n"			// set s600, s601, s602 to zero

		/*
		We test planes in order of increasing address for efficiency.
		it just happens that in a frustum the planes are listed in
		left, right, up and down physical order.

		(but then the codes are bit 1 for up, bit 2 for down, bit 3
		for left and bit 4 for tight)
		*/


		/* Check left plane */
		"lv.q c700, 0+%1 \n"		// Load frustum.leftplane

		/*
		This creates two 3-vectors, 'c710' with (MinX,MinY,MinZ) and
		'c720' with (MaxX,MaxY,MaxZ). The loading is interleaved for
		efficiency, and it is here to be in the shadow of the 'lv.q'
		above.
		*/
		"lv.s s710, 0+%2 \n"		// Load xBB.MinX
		"lv.s s720, 4+%2 \n"		// Load xBB.MaxX
		"lv.s s711, 8+%2 \n"		// Load xBB.MinY
		"lv.s s721, 12+%2 \n"		// Load xBB.MaxY
		"lv.s s712, 16+%2 \n"		// Load xBB.MinZ
		"lv.s s722, 20+%2 \n"		// Load xBB.MaxZ

		/*
		In the four code fragments below:
		vcmp latency:	3
		vmov latency:	3
		vcmov latency:	5
		vdot latency:	7
		*/

		/*
		Also, with 'vcmovf.t', CC 6 has a special meaning: do the
		move elementwise.
		*/

		"vcmp.t lt, c700, c600 \n"	// test left plane normal < zero
		"vmov.t c610, c720 \n"		// c610 = xBB.Max
		"vcmovf.t c610, c710, 6 \n"	// c610 = xBB.Min if normal >= zero

		"vdot.t s713, c610, c700 \n"// Dot product of point and normal
		"vcmp.s GT, s713, s703 \n"	// test dot product > plane distance
		"bvt 0, outside_left \n"	// jump if scalar comparison is true
		"vnop \n"

		/* Check right plane */
		"lv.q c700, 16+%1 \n"		// Load frustum.rightplane
		"vcmp.t lt, c700, c600 \n"	// test right plane normal < zero
		"vmov.t c610, c720 \n"		// c610 = xBB.Max
		"vcmovf.t c610, c710, 6 \n"	// c610 = xBB.Min iff normal >= zero

		"vdot.t s713, c610, c700 \n"// Dot product of point and plane normal
		"vcmp.s GT, s713, s703 \n"	// test dot product > plane distance
		"bvt 0, outside_right\n"	// jump if scalar comparison is true
		"vnop \n"

		/* Check up plane */
		"lv.q c700, 32+%1 \n"		// Load frustum.upplane
		"vcmp.t lt, c700, c600 \n"	// test up plane normal < zero
		"vmov.t c610, c720 \n"		// c610 = xBB.Max
		"vcmovf.t c610, c710, 6 \n"	// c610 = xBB.Min iff normal >= zero

		"vdot.t s713, c610, c700 \n"// Dot product of point and plane normal
		"vcmp.s GT, s713, s703 \n"	// test dot product > plane distance
		"bvt 0, outside_up\n"		// jump if scalar comparison is true
		"vnop \n"

		/* Check down plane */
		"lv.q c700, 48+%1 \n"		// Load frustum.downplane
		"vcmp.t lt, c700, c600 \n"	// test down plane normal < zero
		"vmov.t c610, c720 \n"		// c610 = xBB.Max
		"vcmovf.t c610, c710, 6 \n"	// c610 = xBB.Min if normal => zero

		"vdot.t s713, c610, c700 \n"// Dot product of point and plane normal
		"vcmp.s GT, s713, s703 \n"	// test dot product > plane distance
		"bvt 0, outside_down\n"		// jump if scalar comparison is true
		"vnop \n"

		/* Set the result codes */


		"li $8, 0 \n"				// uResult = 0
		"b done \n"
		"nop \n"

		"outside_up:\n"
		"li $8, 0x01 \n"			// uResult = ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_U		= 0x01,
		"b done \n"
		"nop \n"

		"outside_down:\n"
		"li $8, 0x02 \n"			// uResult = ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_D		= 0x02,
		"b done \n"
		"nop \n"

		"outside_left:\n"
		"li $8, 0x04 \n"			// uResult = ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_L		= 0x04,
		"b done \n"
		"nop \n"

		"outside_right:\n"
		"li $8, 0x08 \n"			// uResult = ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_R		= 0x08,
		"b done \n"
		"nop \n"

		"done:\n"

		"move %0, $8 \n"			// workaround SNC not coping with %0 in branching code

		".set pop\n"				// restore assember option

		: "=r" (uResult)
		: "m" (xFr), "m" (xBB)
		: "$8"
		);

#if PG_DEBUG
	// Test we get the same results
	Asura_Assert(uResult == BBIsCompletelyAboveAnyFrustumPlane_C(xBB,xFr),
		"BBIsCompletelyAboveAnyFrustumPlane changed");
#endif

	return uResult;
#else
	return BBIsCompletelyAboveAnyFrustumPlane_C(xBB,xFr);
#endif
}

static inline bool BBIsPartiallyAbovePlane(const Asura_Bounding_Box& xBB, const Asura_PSP_Plane& xPl)
{
#if 1
	Asura_Vector_3 xPoint;

	xPoint.x = (xPl.m_xNormal.GetX() > 0.f) ? xBB.MaxX : xBB.MinX;
	xPoint.y = (xPl.m_xNormal.GetY() > 0.f) ? xBB.MaxY : xBB.MinY;
	xPoint.z = (xPl.m_xNormal.GetZ() > 0.f) ? xBB.MaxZ : xBB.MinZ;

	return PointIsAbovePlane(xPoint, xPl);
#else
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MinX, xBB.MinY, xBB.MinZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MaxX, xBB.MinY, xBB.MinZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MinX, xBB.MinY, xBB.MaxZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MaxX, xBB.MinY, xBB.MaxZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MinX, xBB.MaxY, xBB.MinZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MaxX, xBB.MaxY, xBB.MinZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MinX, xBB.MaxY, xBB.MaxZ), xPl)) return true;
	if(PointIsAbovePlane(Asura_Vector_3(xBB.MaxX, xBB.MaxY, xBB.MaxZ), xPl)) return true;
	return false;
#endif
}

// less than 1% in profiler
static u_int GetGuardbandClipFlags(const Asura_Bounding_Box& xBB, const Asura_PSP_FrustumLimits& xGuardband)
{
	u_int uClipFlags = 0;
	if(BBIsPartiallyAbovePlane(xBB, xGuardband.m_xLeft))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_L;
	}
	if(BBIsPartiallyAbovePlane(xBB, xGuardband.m_xRight))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_R;
	}
	if(BBIsPartiallyAbovePlane(xBB, xGuardband.m_xUp))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_U;
	}
	if(BBIsPartiallyAbovePlane(xBB, xGuardband.m_xDown))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_D;
	}
	return uClipFlags;
}

// given clip flags for containing BB (module) get the clip flags for
// this sub-BB (strip) 3-5% in profiler. 2nd priority after
// BBIsCompletelyAboveAnyFrustumPlane
static u_int GetGuardbandClipFlags2(const Asura_Bounding_Box& xBB,
									const Asura_PSP_FrustumLimits& xGuardband, const u_int uCurFlags)
{
	u_int uClipFlags = 0;
#if 1
#if 0
	Asura_Vector_3 xPoint;
	xPoint.x = (xPl.m_xNormal.GetX() > 0.f) ? xBB.MaxX : xBB.MinX;
	xPoint.y = (xPl.m_xNormal.GetY() > 0.f) ? xBB.MaxY : xBB.MinY;
	xPoint.z = (xPl.m_xNormal.GetZ() > 0.f) ? xBB.MaxZ : xBB.MinZ;

	// return dotproduct(pos, plane normal) > plane distance
	return (xPoint * xPl.m_xNormal) > xPl.m_fPosition;
#endif
	if( (uCurFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_L) && BBIsPartiallyAbovePlane(xBB, xGuardband.m_xLeft))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_L;
	}
	if( (uCurFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_R) && BBIsPartiallyAbovePlane(xBB, xGuardband.m_xRight))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_R;
	}
	if( (uCurFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_U) && BBIsPartiallyAbovePlane(xBB, xGuardband.m_xUp))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_U;
	}
	if( (uCurFlags & ASURA_PSP_MODULE_CLIPS_GUARDBAND_D) && BBIsPartiallyAbovePlane(xBB, xGuardband.m_xDown))
	{
		uClipFlags |= ASURA_PSP_MODULE_CLIPS_GUARDBAND_D;
	}
#else
	__asm__
		(
		"lv.q c700, 0+%1 \n"		// Load xGuardband.leftplane						(lv latency 3)

		"lv.s s710, 0+%2 \n"		// Load xBB.MinX
		"lv.s s720, 4+%2 \n"		// Load xBB.MaxX
		"lv.s s711, 8+%2 \n"		// Load xBB.MinY
		"lv.s s721, 12+%2 \n"		// Load xBB.MaxY
		"lv.s s712, 16+%2 \n"		// Load xBB.MinZ
		"lv.s s722, 20+%2 \n"		// Load xBB.MaxZ

		"vzero.t c600 \n"			// set s600, s601, s602 to zero						(vzero latency 3)

		"vcmp.t gt, c700, c600 \n"	// test plane normal > zero							(vcmp latency 3)
		"vmov.t c610, c720 \n"		// c610 = xBB.Max									(vmov latency 3)
		"vcmovf.t c610, c710, 6 \n"	// for those tests that failed put c610 = xBB.Min	(vcmov latency 5)

		"vdot.t s713, c610, c700 \n" // Dot product of point and plane normal			(vdot latency 7)
		"vcmp.s GT, s713, s703 \n"	// test dot product > plane distance
		"bvf 0, 0f \n"	// jump if scalar comparison is false
		"vnop \n"

		"ori %3, %3, 0x04 \n"		// uClipFlags |= ASURA_PSP_MODULE_OUTSIDE_FRUSTUM_L		= 0x04,

		"0: \n"

		"done:\n"

		: "+r" (uClipFlags)
		: "m" (xGuardband), "m" (xBB)
		);

	// TODO check clobbers and Dylan's rules for uClipFlags
	// TODO: finish writing this function.

#endif
	return uClipFlags;
}

// 3-5% in profiler
static void TestPointAgainstFrustum(const Asura_Vector_3& xPos, const Asura_PSP_FrustumLimits& xFrustum, Asura_Frustum_Test_Result_Bitfield& xResult)
{
	xResult.m_xFlags.m_bThirdLeft = PointIsAbovePlane(xPos, xFrustum.m_xLeft);
	xResult.m_xFlags.m_bThirdRight = PointIsAbovePlane(xPos, xFrustum.m_xRight);
	xResult.m_xFlags.m_bThirdUp = PointIsAbovePlane(xPos, xFrustum.m_xUp);
	xResult.m_xFlags.m_bThirdDown = PointIsAbovePlane(xPos, xFrustum.m_xDown);
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_PSP_RenderEnvironment
//
////////////////////////////////////////////////////////////////////////////////

#if defined(ASURA_USE_STREAMING)
Asura_PSP_RenderEnvironment_Zone*							Asura_PSP_RenderEnvironment::s_pxCurrentZone	= NULL;		// Currently active handler.
Asura_TemplateHashTree<Asura_PSP_RenderEnvironment_Zone>	Asura_PSP_RenderEnvironment::s_xZoneTree;					// Tree of all of the handlers available.
#else
Asura_PSP_RenderEnvironment_Zone							Asura_PSP_RenderEnvironment::s_xZone;						// Single static-instance of a zone.
#endif	// defined(ASURA_USE_STREAMING)


void Asura_PSP_RenderEnvironment::Initialise(void)
{
#if defined(ASURA_USE_STREAMING)
	const Asura_Hash_ID	uHash	= 0x5C13D641; //Asura_GetHashID("Default");

	// Create our instance - for now...
	s_pxCurrentZone	= CreateZone (uHash);
#endif	// defined(ASURA_USE_STREAMING)
}


void Asura_PSP_RenderEnvironment::Reset(void)
{
	Shutdown();
	Initialise();
}


void Asura_PSP_RenderEnvironment::Shutdown(void)
{
#if defined(ASURA_USE_STREAMING)
	// Reset the entire zone-tree...
	s_xZoneTree.Reset();

	// Remove the current zone...
	s_pxCurrentZone	= NULL;
#else
	Asura_PSP_RenderEnvironment_Zone&	xZone	= GetCurrentZone();

	// Reset the current instance...
	xZone.Reset();
#endif	// defined(ASURA_USE_STREAMING)
}

void Asura_PSP_RenderEnvironment::Render (void)
{
	// ARTTODO - loop through all zones...
	GetCurrentZone().Render();
}


#if defined(ASURA_USE_STREAMING)

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_RenderEnvironment::CreateZone
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Attempts to create a render-environment-zone with the
//				  specified hash-ID.
//
// Parameters   : The hash-ID of the zone to create.
//
// Return value : A pointer to the zone created.
////////////////////////////////////////////////////////////////////////////////

Asura_PSP_RenderEnvironment_Zone* Asura_PSP_RenderEnvironment::CreateZone (const Asura_Hash_ID uHash)
{
	Asura_PSP_RenderEnvironment_Zone*	pxZone	= s_xZoneTree.GetDataByHash(uHash);

	// Do we already have this zone?
	if(pxZone == NULL)
	{
		// No - create one...
		pxZone	= s_xZoneTree.AddNode(uHash);
	}
	else
	{
		// Yes - this shouldn't happen...
		Asura_Assert(false, "Already have an physics-zone with this hash-ID.");
	}

	// Return the zone...
	return(pxZone);
}

/// End of function 'Asura_PSP_RenderEnvironment::CreateZone'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_RenderEnvironment::DestroyZone
// Author       : Andy Tate
// Date         : 14-Mar-2006
//
// Description  : Attempts to destroy the render-environment-zone with the
//				  specified hash-ID.
//
// Parameters   : The hash-ID of the zone to destroy.
//
// Return value : "true" if destroyed, "false" if failed.
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_RenderEnvironment::DestroyZone (const Asura_Hash_ID uHash)
{
	// Call the delete function...
	return(s_xZoneTree.DeleteNodeByHash(uHash));
}

/// End of function 'Asura_PSP_RenderEnvironment::DestroyZone'. ///

#endif	// defined(ASURA_USE_STREAMING)

/// End of file. ///
