///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_StencilBufferManager.cpp
// Author : Catalin Arsenescu
// Date   : 1-March-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////
#include "Core\Asura.h"

// This file's header
#include "Rendering\Asura_PSP_StencilBufferManager.h"

// Sony libraries

// PSP headers
#include "Rendering/Asura_PSP_GU.h"
#include "Rendering\Asura_PSP_Stencil_Enums.h"

// Asura independent headers
#include "Debugging\Asura_Assert.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////
static const int s_iStencilOperationMap[ASURA_PSP_NUM_STENCIL_OPERATIONS] =
{
	SCEGU_KEEP,
	SCEGU_ZERO,
	SCEGU_REPLACE,
	SCEGU_INVERT,
	SCEGU_INCR,
	SCEGU_DECR
};

static const int s_iStencilFunctionMap[ASURA_PSP_NUM_STENCIL_FUNCTIONS] =
{
	SCEGU_NEVER,
	SCEGU_ALWAYS,
	SCEGU_EQUAL,
	SCEGU_NOTEQUAL,
	SCEGU_LESS,
	SCEGU_LEQUAL,
	SCEGU_GREATER,
	SCEGU_GEQUAL
};

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////
ASURA_PSP_STENCIL_OPERATION		Asura_PSP_StencilBufferManager::s_eStencilOperationStencilTestFail	= ASURA_PSP_STENCIL_OPERATION_KEEP;
ASURA_PSP_STENCIL_OPERATION		Asura_PSP_StencilBufferManager::s_eStencilOperationZTestFail		= ASURA_PSP_STENCIL_OPERATION_KEEP;
ASURA_PSP_STENCIL_OPERATION		Asura_PSP_StencilBufferManager::s_eStencilOperationZTestPass		= ASURA_PSP_STENCIL_OPERATION_REPLACE;

ASURA_PSP_STENCIL_FUNCTION		Asura_PSP_StencilBufferManager::s_eStencilFunction					= ASURA_PSP_STENCIL_FUNCTION_ALWAYS;
u_char							Asura_PSP_StencilBufferManager::s_ucStencilReferenceValue			= ASURA_PSP_STENCIL_OUTLINE;
u_char							Asura_PSP_StencilBufferManager::s_ucStencilMaskValue				= 0x00;

u_char							Asura_PSP_StencilBufferManager::s_ucDefaultStencilClearBits			= 0x00;

bool							Asura_PSP_StencilBufferManager::s_bStencilEnabled					= false;
bool							Asura_PSP_StencilBufferManager::s_bUpdateStencilFunction			= true;
bool							Asura_PSP_StencilBufferManager::s_bUpdateStencilOperation			= true;


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::SetStencilOperation
// Author       : Catalin Arsenescu
// Date         : 1-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::SetStencilOperation ( ASURA_PSP_STENCIL_OPERATION eStencilFailOperation, ASURA_PSP_STENCIL_OPERATION eZTestFailOperation, ASURA_PSP_STENCIL_OPERATION eZTestPassOperation )
{
#ifdef ASURA_ASSERT
	Asura_Assert(eStencilFailOperation >= 0 && eStencilFailOperation < ASURA_PSP_NUM_STENCIL_OPERATIONS, "Invalid stencil fail operation");
	Asura_Assert(eZTestFailOperation >= 0 && eZTestFailOperation < ASURA_PSP_NUM_STENCIL_OPERATIONS, "Invalid Z fail operation");
	Asura_Assert(eZTestPassOperation >= 0 && eZTestPassOperation < ASURA_PSP_NUM_STENCIL_OPERATIONS, "Invalid Z pass operation");
#endif

	if( eStencilFailOperation	!=	s_eStencilOperationStencilTestFail	||
		eZTestFailOperation		!=	s_eStencilOperationZTestFail		||
		eZTestPassOperation		!=	s_eStencilOperationZTestPass		)
	{
		s_bUpdateStencilOperation = true;
		s_eStencilOperationStencilTestFail	= eStencilFailOperation;
		s_eStencilOperationZTestFail		= eZTestFailOperation;
		s_eStencilOperationZTestPass		= eZTestPassOperation;
	}

	if( s_bStencilEnabled )
	{
		UpdateStencilOperation();
	}
}

/// End of function 'Asura_PSP_StencilBufferManager::SetStencilOperation'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::SetStencilFunction
// Author       : Catalin Arsenescu
// Date         : 1-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::SetStencilFunction ( ASURA_PSP_STENCIL_FUNCTION eFunction, u_char ucReferenceValue, u_char ucMaskValue )
{
#ifdef ASURA_ASSERT
	Asura_Assert(eFunction >= 0 && eFunction < ASURA_PSP_NUM_STENCIL_FUNCTIONS, "Invalid stencil function");
#endif

	if( eFunction			!=	s_eStencilFunction			||
		ucReferenceValue	!=	s_ucStencilReferenceValue	||
		ucMaskValue			!=	s_ucStencilMaskValue		)
	{
		s_eStencilFunction			= eFunction;
		s_ucStencilReferenceValue	= ucReferenceValue;
		s_ucStencilMaskValue		= ucMaskValue;
		s_bUpdateStencilFunction	= true;
	}

	if( s_bStencilEnabled )
	{
		UpdateStencilFunction();
	}
}

/// End of function 'Asura_PSP_StencilBufferManager::SetStencilFunction'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::SetStencilEnabled
// Author       : Catalin Arsenescu
// Date         : 1-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::SetStencilEnabled ( bool bEnabled )
{
	if( bEnabled)
	{
		if( !s_bStencilEnabled )
		{
			UpdateStencilFunction();
			UpdateStencilOperation();

			s_bStencilEnabled = true;
			GU_Enable ( SCEGU_STENCIL_TEST );
		}
	}
	else
	{
		if( s_bStencilEnabled )
		{
			s_bStencilEnabled = false;
			GU_Disable ( SCEGU_STENCIL_TEST );
		}
	}
}

/// End of function 'Asura_PSP_StencilBufferManager::SetStencilEnabled'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::ClearStencilBuffer
// Author       : Catalin Arsenescu
// Date         : 2-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::ClearStencilBuffer(void)
{
	GU_Clear ( SCEGU_CLEAR_FAST | SCEGU_CLEAR_STENCIL );
}

/// End of function 'Asura_PSP_StencilBufferManager::ClearStencilBuffer'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::SetStencilBufferClearBits
// Author       : Catalin Arsenescu
// Date         : 2-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::SetStencilBufferClearBits(u_char ucDefaultClearBits)
{
	if(s_ucDefaultStencilClearBits != ucDefaultClearBits)
	{
		s_ucDefaultStencilClearBits = ucDefaultClearBits;
		GU_ClearStencil ( s_ucDefaultStencilClearBits );
	}
}

/// End of function 'Asura_PSP_StencilBufferManager::SetStencilBufferClearBits'. ///


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::UpdateStencilFunction
// Author       : Catalin Arsenescu
// Date         : 1-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::UpdateStencilFunction(void)
{
	if( s_bUpdateStencilFunction )
	{
		s_bUpdateStencilFunction = false;
		GU_StencilFunc ( s_iStencilFunctionMap [ s_eStencilFunction ], s_ucStencilReferenceValue, s_ucStencilMaskValue );
	}
}

/// End of function 'Asura_PSP_StencilBufferManager::UpdateStencilFunction'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_StencilBufferManager::UpdateStencilOperation
// Author       : Catalin Arsenescu
// Date         : 1-March-2007
////////////////////////////////////////////////////////////////////////////////
void Asura_PSP_StencilBufferManager::UpdateStencilOperation(void)
{
	if( s_bUpdateStencilOperation )
	{
		s_bUpdateStencilOperation = false;

		GU_StencilOp ( s_iStencilOperationMap[ s_eStencilOperationStencilTestFail ],
						s_iStencilOperationMap[ s_eStencilOperationZTestFail ],
						s_iStencilOperationMap[ s_eStencilOperationZTestPass ] );
	}
}

/// End of function 'Asura_PSP_StencilBufferManager::UpdateStencilOperation'. ///

/// End of the file
