//
// Asura_PSP_VRAMManagement.cpp
//
// Created by JG - 02/11/2005 14:31:19
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core\Asura.h"

// This file's 
#include "Rendering/Asura_PSP_VRAMManagement.h"

#include "Debugging/Asura_Assert.h"

#include "Rendering/Asura_PSP_TextureManagement.h"

#include "Rendering/Asura_PSP_GU.h"

// We need to access the VRAM, so we need the start address
#include <geman.h>

// Size of frame buffers, etc.

// Total VRAM size
const unsigned int uVRAM_SIZE							= 2 * 1024 * 1024; 

// Buffer reserved for fullscreen effects, after frame / depth buffers
const unsigned int uRESERVED_VRAM_FULLSCREEN_EFFECTS	= /* Screen grab*/ 128 * 128 * 4 + /* Effect buffer */ 128 * 128 * 4; 

// Free VRAM start address (after reserved buffers)
const unsigned int uFREE_VRAM_START_ADDRESS				= ((unsigned int)SCEGU_VRAM_BP32_2 + SCEGU_VRAM_BUFSIZE + uRESERVED_VRAM_FULLSCREEN_EFFECTS); 

// Free VRAM 
const unsigned int uFREE_VRAM_SIZE						= uVRAM_SIZE - uFREE_VRAM_START_ADDRESS; 

bool			Asura_PSP_VRAMManagement::s_bInitialised = false;		// Manager initialised?
char*			Asura_PSP_VRAMManagement::s_pcVRAM = NULL;				// Pointer to actual VRAM address (including mapping offset)
unsigned int	Asura_PSP_VRAMManagement::s_uAllocatedSize = 0;			// Allocated VRAM 

int				Asura_PSP_VRAMManagement::s_iCurrentHandler;						// Current handler (index to unused block)
int				Asura_PSP_VRAMManagement::s_iOffsetHandler;							// Timestamp
char*			Asura_PSP_VRAMManagement::s_apHandlers[uMAX_NUMBER_OF_TEXTURES-1];	// Array of handlers (pointers to VRAM blocks)

unsigned int	Asura_PSP_VRAMManagement::s_uHitsThisFrame;							// Number of VRAM accesses this frame
unsigned int	Asura_PSP_VRAMManagement::s_uFramesWithoutResetting;				// Number of frames running without a reset

DEBUGVAR unsigned int dbg_uMinimumNumberOfVRAMTextureHitsPerFrame = 40;				// Minimum number of accesses before resetting the VRAM manager
DEBUGVAR unsigned int dbg_uMinimumNumberOfFramesBeforeResetting = 150;				// Minimum number of frames running before resetting the manager again

void Asura_PSP_VRAMManagement::Initialise()
{
	// Is it already initialised?
	if (!s_bInitialised)
	{
		// Fix the pointer to the actual VRAM address (beginning of free space)
		s_pcVRAM = (char*)sceGeEdramGetAddr() + uFREE_VRAM_START_ADDRESS; 

		// Reset various counters
		s_uAllocatedSize = 0;
		s_iCurrentHandler = 0; 
		s_iOffsetHandler = 0; 
		s_uHitsThisFrame = 0; 
		s_uFramesWithoutResetting = 0; 

		// Make sure we don't initialise this again
		s_bInitialised = true; 
	}
	else
	{
		Asura_Assert(false, "Asura_PSP_VRAMManagement already initialised!"); 
	}
}

void Asura_PSP_VRAMManagement::Reset()
{
	// Reset the VRAM manager (old content is ignored from now)
	s_uAllocatedSize = 0; 
	s_iCurrentHandler = 0; 
	s_uHitsThisFrame = 0; 
	s_uFramesWithoutResetting = 0; 

	// Update the timestamp to invalidate access to previous content
	s_iOffsetHandler += uMAX_NUMBER_OF_TEXTURES;

	// Textures cached in VRAM are no longer valid
	Asura_PSP_TextureManagement::InvalidateTexturesInVRAM();
}

int Asura_PSP_VRAMManagement::AllocateVRAM(unsigned int uSize)
{
	// If we can't allocate, return and invalid handler
	int iHandler = ASURA_PSP_VRAM_INVALID; 

	// Have we allocated too many textures?
	if (s_iCurrentHandler < uMAX_NUMBER_OF_TEXTURES-1)
	{
		// Check that we have enough space to allocate this size
		if (uSize < uFREE_VRAM_SIZE - s_uAllocatedSize)
		{
			// Use the last handler, combined with the timestamp
			iHandler = s_iOffsetHandler + s_iCurrentHandler; 

			// Update the pointer to actual VRAM data
			s_apHandlers[s_iCurrentHandler] = s_pcVRAM + s_uAllocatedSize; 

			// Increase the allocated size
			s_uAllocatedSize += uSize; 

			// Next handler
			++s_iCurrentHandler; 
		}
		else
		{
			// Not enough space, return handle indicating not enough space to fit this data
			iHandler = s_iOffsetHandler + ASURA_PSP_VRAM_FULL;
		}
	}
	else
	{
		// We've run out of handlers
		Asura_Assert(false, "Reached maximum number of textures in VRAM, please increase it"); 
	}

	return iHandler; 
}

bool Asura_PSP_VRAMManagement::IsValidHandler(int iHandler)
{
	// Remove the timestamp
	int iIndex = iHandler - s_iOffsetHandler; 
	
	// Is this handler valid?
	if ((iIndex >= 0 && iIndex < s_iCurrentHandler && iIndex < uMAX_NUMBER_OF_TEXTURES - 1) || iIndex == ASURA_PSP_VRAM_FULL)
	{
		// Yes
		return true;
	}

	// The handler isn't valid
	return false; 
}

char* Asura_PSP_VRAMManagement::GetVRAM(int iHandler)
{
	// Remove timestamp
	int iIndex = iHandler - s_iOffsetHandler; 
	
	// Is it a handler allocated with the current timestamp?
	if (iIndex >= 0 && iIndex < s_iCurrentHandler)
	{
		// Good, we have a hit
		++s_uHitsThisFrame; 

		// Give back a pointer to VRAM
		return s_apHandlers[iIndex]; 
	}

	// Nope, invalid handler
	return NULL; 
}

void Asura_PSP_VRAMManagement::OnBeginFrame()
{
	// Reset the number of hits every frame
	s_uHitsThisFrame = 0; 
}

void Asura_PSP_VRAMManagement::OnEndFrame()
{
	// Do we have to reload stuff on VRAM?
	if (s_uHitsThisFrame < dbg_uMinimumNumberOfVRAMTextureHitsPerFrame && s_uFramesWithoutResetting >= dbg_uMinimumNumberOfFramesBeforeResetting)
	{
		// Yes, not that much hits for a while, reset the manager
		// clearing the way for new data
		Reset(); 
	}
	else
	{
		// No, count another frame without resetting
		++s_uFramesWithoutResetting; 
	}
}