
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_SCEUtilityWrapper.h
// Author : Andy Tate
// Date   : 16-Mar-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __ASURA_PSP_SCEUTILITYWRAPPER_H__
#define __ASURA_PSP_SCEUTILITYWRAPPER_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include <thread.h>
#include <utility/utility_osk.h>
#include <utility/utility_savedata.h>
#include <utility/utility_msgdialog.h>
#ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK
#include <utility/utility_netconf.h>
#endif
#ifdef ASURA_PSP_GAMESHARING 
#include <pspnet_adhocctl.h>
#include <utility/utility_gamesharing.h>
#endif

#include "Core/Asura_Multithread.h"
#include "Core/Asura_PlatformBasics.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const u_int	uASURA_PSP_NICKNAME_MAX_CHARS	= 128;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

class Asura_PSP_SCEUtilityWrapper
{
public:
	// Public typedefs...

	typedef bool LoadDataValidationFunction (const void* pData, const u_int uDataSize, const u_int uBufferSize);

	// Public enumerated types...

	enum CommandState
	{
		COMMAND_INACTIVE,
		COMMAND_IN_PROGRESS,
		COMMAND_COMPLETED,
		COMMAND_CANCELLED,
		COMMAND_ERROR,
		COMMAND_RETRY,

		MAX_COMMAND_STATES
	};

	enum OSKCharacterSet
	{
		LATIN_ALPHABET,
		LATIN_ALPHABET_UPPERCASE,
		LATIN_ALPHABET_LOWERCASE,
		LATIN_NUMERALS,
		LATIN_ALPHANUMERIC,
		LATIN_ALPHANUMERIC_UPPERCASE,
		LATIN_ALPHANUMERIC_LOWERCASE,

		MAX_OSK_CHARACTER_SET
	};

	// Public functions...

	// Static initialisation/shutdown functions...
	static	void						Initialise						(const Asura_UnicodeChar* wszTitleName);
	static	void						Shutdown						(void);

	// Update function - only EndFrame() does anything, and it assumes you're not currently rendering anything in Asura.
	static	CommandState				BeginFrame						(void)	{ return s_eCommandState; }
	static	CommandState				Update							(void)	{ return s_eCommandState; }
	static	CommandState				Render							(void)	{ return s_eCommandState; }
	static	CommandState				EndFrame						(void);

	// Initialisation functions (start up the utility library for relevant task)
	static	CommandState				InitialiseOnScreenKeyboard		(const Asura_UnicodeChar* wszMessage, const Asura_UnicodeChar* wszTextBeforeEdit, Asura_UnicodeChar* wszTextAfterEdit, const int iMaxCharsInclNul, const OSKCharacterSet eCharacterSet);
#ifdef ASURA_PSP_GAMESHARING
	//MKR 23/11/06: starts up the network libs and then the game sharing utility
	static  CommandState				InitialiseGameSharing			(const Asura_Char* const szFileName);
#endif

	static	CommandState				InitialiseSaveDataAutoload		(void* pvData, const u_int uDataSize, const u_int uBufferSize, LoadDataValidationFunction* pfnValidation = NULL);
	static	CommandState				InitialiseSaveDataAutosave		(const Asura_UnicodeChar* wszSaveTitle, const Asura_UnicodeChar* wszSaveDescription, const Asura_UnicodeChar* wszNewData, void const* pvData, const u_int uDataSize, const u_int uBufferSize);
	static	CommandState				InitialiseSaveDataDelete		(void);
	static	CommandState				InitialiseSaveDataLoadList		(const int iNumSaveSlots, void* pvData, const u_int uDataSize, const u_int uBufferSize, LoadDataValidationFunction* pfnValidation = NULL);
	static	CommandState				InitialiseSaveDataSaveList		(const int iNumSaveSlots, const Asura_UnicodeChar* wszSaveTitle, const Asura_UnicodeChar* wszSaveDescription, const Asura_UnicodeChar* wszNewData, const void* pData, const u_int uDataSize, const u_int BufferSize);

	//"single file" functions - use for storing a single file in parallel to the list functionality above
	// user id equates to a file name, allowing multiple single files
	// the list calls use user ID's of the form 0000, 0001 etc, so avoid these
	// autoload fails silently if no file or mem card found, so can be used on boot
	static	CommandState				InitialiseSaveDataLoadSingle	(const char* szUserID, void* pvData, const u_int uDataSize, const u_int uBufferSize, LoadDataValidationFunction* pfnValidation = NULL);
	static	CommandState				InitialiseSaveDataSaveSingle	(const char* szUserID, const Asura_UnicodeChar* wszSaveTitle, const Asura_UnicodeChar* wszSaveDescription, const Asura_UnicodeChar* wszNewData, const void* pData, const u_int uDataSize, const u_int BufferSize);
	static	CommandState				InitialiseSaveDataAutoloadSingle	(const char* szUserID, void* pvData, const u_int uDataSize, const u_int uBufferSize, LoadDataValidationFunction* pfnValidation = NULL);
	static	CommandState				InitialiseSaveDataAutosaveSingle	(const char* szUserID, const Asura_UnicodeChar* wszSaveTitle, const Asura_UnicodeChar* wszSaveDescription, const Asura_UnicodeChar* wszNewData, const void* pData, const u_int uDataSize, const u_int BufferSize);
	// Check save size (AN 12/09/07)
	static	CommandState				InitialiseSaveDataCheckSize		(const u_int uDataSize, const u_int BufferSize);

	static	CommandState				InitialiseMessageDialog			(const u_int uErrorNumber);

#ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK
	static	CommandState				InitialiseInfrastructureMode	(void);
#endif

	// Accessor functions...
	static	bool						IsMemoryStickPresent			(void);
	static	void						NotifyMemoryStickInserted		(void);
	static	bool						HasMemoryStickChangedSinceLastLoadOrSave(void) { return s_bMemoryStickChanged; }
	static	bool						IsAbleToPerformSavedataAutosave	(void)	{ return (s_szSavedataAutosaveUserID [0] != '\0'); }
	static	bool						IsCircleButtonEnter				(void)	{ return (SCE_UTILITY_CTRL_ASSIGN_CIRCLE_IS_ENTER == s_iSceUtilitySystemParamCtrlAssign); }
	static	const Asura_UnicodeChar*	GetNickname						(void)	{ return (s_wszNickname); }
	static	bool 						GetNeedDelete					(void)	{ return s_bNeedDelete; }
	static	void 						ResetNeedDelete					(void)	{ s_bNeedDelete = false; }
	static	CommandState				GetState						(void)	{ return s_eCommandState; }
	static	bool						ShouldHaveFocus					(void);
	static	void						SetGoToDeleteIfNoSpace			(bool bGoToDeleteIfNoSpace) { s_bGoToDeleteIfNoSpace = bGoToDeleteIfNoSpace; }

	// Return save data size in KB when last checked, in different formats (AN 13/09/07)
	static	u_int						GetLastSaveDataSizeUInt(void)	{ return ( ( s_xSceUtilitySavedataUtilityDataSize.sizeKB > 0 ) ? s_xSceUtilitySavedataUtilityDataSize.sizeKB : s_xSceUtilitySavedataUtilityDataSize.size32KB ); }
	static	const char*					GetLastSaveDataSizeChar(void)	{ return ( ( s_xSceUtilitySavedataUtilityDataSize.sizeKB > 0 ) ? s_xSceUtilitySavedataUtilityDataSize.sizeStr : s_xSceUtilitySavedataUtilityDataSize.size32Str ); }

	// Return save data size in KB when last checked, calculated with a 32K cluster size, in different formats (AN 13/09/07)
	static	u_int						GetLastSaveDataMaxSizeUInt(void)	{ return s_xSceUtilitySavedataUtilityDataSize.size32KB; }
	static	const char*					GetLastSaveDataMaxSizeChar(void)	{ return s_xSceUtilitySavedataUtilityDataSize.size32Str; }

	// Return save data cluster size when last checked (AN 13/09/07)
	static	u_int						GetLastSaveDataClusterSize(void)	{ return s_xSceUtilitySavedataUtilityDataSize.cluster; }

	// Menu page to be rendered behind utility menus
	static const Asura_Hash_ID&			GetBackgroundMenuPage(void)								{ return s_uBackgroundMenuPage; }
	static void							SetBackgroundMenuPage( Asura_Hash_ID uPage )			{ s_uBackgroundMenuPage = uPage; }

	static inline void LockMutex() { s_xUtilityMutex.Lock(); }
	static inline void UnlockMutex() { s_xUtilityMutex.Unlock(); }

private:
	// Private enumerated types...

	enum Command
	{
		// Default state...
		NONE,

		// On-screen keyboard...
		ON_SCREEN_KEYBOARD,

		// Save-data...
		SAVEDATA_AUTOLOAD,
		SAVEDATA_AUTOSAVE,
		SAVEDATA_DELETE,
		SAVEDATA_LOAD_LIST,
		SAVEDATA_SAVE_LIST,
		SAVEDATA_LOAD_SINGLE,
		SAVEDATA_SAVE_SINGLE,
		SAVEDATA_AUTOLOAD_SINGLE,
		SAVEDATA_AUTOSAVE_SINGLE,
		SAVEDATA_CHECKSAVESIZE,

		// Message dialog...
		MESSAGE_DIALOG,

#ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK
		// Infrastructure mode connection
		INFRASTRUCTURE_CONNNECTION,
#endif

#ifdef ASURA_PSP_GAMESHARING
		GAME_SHARING,
#endif

		// Count...
		MAX_COMMANDS
	};


	// Private functions
	
	// - class is not instantiable (or copiable)
	Asura_PSP_SCEUtilityWrapper();
	~Asura_PSP_SCEUtilityWrapper();
	Asura_PSP_SCEUtilityWrapper(const Asura_PSP_SCEUtilityWrapper& xOther);
	Asura_PSP_SCEUtilityWrapper& operator= (const Asura_PSP_SCEUtilityWrapper& xOther);

#ifdef ASURA_PSP_GAMESHARING
	//MKR 23/11/06: shuts down the network libraries that the game sharing utility relies on 
	static  void						GameSharingNetTerm				(void);
#endif

	// Load function...
	static	bool						LoadSceUtilitySDExtFile			(char const* szFilename, SceUtilitySDExtFile& rSceUtilitySDExtFile);

	// Initialisation functions...
	static	void						InitialiseStructure				(const int iSize, void* pvStructure, SceUtilityParamBase& rBase);
	static	void						InitialiseUserIDs				(const int iNumUserIDs);

	// Commencing functions...
	static	void						HandleLoadCommencing			(void* pData, const u_int uDataSize, const u_int uBufferSize);
	static	void						HandleSaveCommencing			(const Asura_UnicodeChar* wszSaveTitle, const Asura_UnicodeChar* wszSaveDescription, const void* pData, const u_int uDataSize, const u_int uBufferSize);
	// Check save size (AN 12/09/07)
	static	void						HandleCheckSizeCommencing		(const u_int uDataSize, const u_int uBufferSize);


	// Completion functions...
	static	void						HandleOSKCompletion				(void);

	static	void						HandleDeleteCompletion			(void);
	static	void						HandleLoadCompletion			(void);
	static	void						HandleSaveCompletion			(void);
	static	void						HandleCheckSizeCompletion		(void); // Check save size (AN 12/09/07)

	// Sony Utility library wrapper functions...
	static	CommandState				InitialiseUtilityLibrary		(const Command eCommand, void* pvStructure);
	static	int							GetUtilityLibraryStatus			(void);
	static	CommandState 				UpdateUtilityLibrary			(const int iNumFrames);
	static	CommandState				ShutdownUtilityLibrary			();

	// helper for memory card free space
	static  u_int						GetMemoryStickFreeBlocks		(void);

	// Private variables...

	// What we're currently doing...
	static	Command								s_eCommand;
	static	CommandState						s_eCommandState;

	// The command status at shutdown...
	static	CommandState						s_eShutdownCommandState;
	static	u_int								s_uShutdownErrorNumber;

	// Sony's on-screen keyboard structures...
	static	SceUtilityOskInputFieldInfo			s_xSceUtilityOskInputFieldInfo;
	static	SceUtilityOskParam					s_xSceUtilityOskParam;

	// Sony's save-data structures...
	static	void*								s_pvDestinationForSavedataLoad;
	static	u_int								s_uExpectedSizeOfSavedataLoad;

	static	char								s_szSavedataNewDataTitle [SCE_UTILITY_SD_SYSF_TITLE_SIZE];

	static	char								s_szSavedataAutosaveUserID [SCE_UTILITY_SD_USERID_SIZE];
	static	char								s_aszSavedataUserIDs [SCE_UTILITY_SD_LISTITEM_MAX][SCE_UTILITY_SD_USERID_SIZE];

	static	SceUtilitySDExtFile					s_xSceUtilitySDExtFileBackground;
	static	SceUtilitySDExtFile					s_xSceUtilitySDExtFileIcon;
	static	SceUtilitySavedataListSaveNewData	s_xSceUtilitySavedataListSaveNewData;

	static	SceUtilitySavedataParam				s_xSceUtilitySavedataParam;

	// Sony's message-dialog strutures...
	static	SceUtilityMsgDialogParam			s_xSceUtilityMsgDialogParam;

	// Utility data that holds the calculated save sizes (AN 13/09/07)
	static	SceUtilitySavedataUtilityDataSize	s_xSceUtilitySavedataUtilityDataSize;

#ifdef ASURA_PSP_GAMESHARING
	static SceUtilityGameSharingParam			s_xGameSharingParam;
	static SceNetAdhocctlAdhocId				s_xGameSharingID;	
	static SceNetAdhocctlGroupName				s_xGamesharingGroupName;
#endif

#ifdef ASURA_PSP_SUPPORT_INFRASTRUCTURE_NETWORK
	// Sony's infrastructure mode structures
	static	SceUtilityNetconfParam				s_xSceUtilityNetconfParam;
#endif

	// Circle button variables...
	static	int									s_iSceUtilitySystemParamCtrlAssign;

	// Language to use for system utilities
	static  int									s_iSceUtilitySystemLanguage;

	// Nickname...
	static	Asura_UnicodeChar					s_wszNickname [uASURA_PSP_NICKNAME_MAX_CHARS];

	// Project-specific data...
	static	const Asura_UnicodeChar*			s_wszTitleName;

	static	LoadDataValidationFunction			*s_pfnValidation;

	static bool	s_bNeedDelete;
	
	// Go straight to the delete menu if a save failed due to insufficient space.
	static bool	s_bGoToDeleteIfNoSpace;

	static Asura_Mutex s_xUtilityMutex;
	static bool s_bMemoryStickChanged;
	static u_int s_uMemoryStickFreeBlocks;

	static Asura_Hash_ID s_uBackgroundMenuPage;

};   // End of class 'Asura_PSP_SCEUtilityWrapper'.


#endif   // ifndef __ASURA_PSP_SCEUTILITYWRAPPER_H__


/// End of file. ///
