// Asura_PSP_Atrac.cpp
//
// Created by Kim Randell - 07/09/2004
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

// core
#include "Core/Asura.h"

// This file's header
#include "Sound/Asura_PSP_Atrac.h"

// Asura independent headers
#include "Core/Asura_Maths.h"
#include "Core/Asura_Multithread.h"
#include "Core/Asura_String.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileCache.h"
#include "Sound/Asura_Sound.h"

// Asura PSP headers
#include "Core/Asura_PSP_Thread_Priorities.h"
#include "FileAccess/Asura_PSP_FileIO.h"
#include "Sound/Asura_PSP_SAS.h"
#include "Sound/Asura_PSP_SimpleAudio.h"


// Project headers...
#include "ProjectSettings/Asura_PSPConfiguration.h"

// PSP headers
#include <kernel.h> // for threading
#include <libatrac3plus.h>
#include <libwave.h>
#include <audiooutput.h>
#include <libsas.h>


// Useful debugging defines...
#if 1
#define LOG_OUTPUT(s)
#else
#define LOG_OUTPUT(s)	Asura_OutputToDebugger_Retail( s )
#endif

// Buffer defines...
#if !defined(ASURA_PSP_ATRAC_INPUT_BUFFER_SIZE)
#define ASURA_PSP_ATRAC_INPUT_BUFFER_SIZE	(16 * 1024)
#endif	// !defined(ASURA_PSP_ATRAC_INPUT_BUFFER_SIZE)

#if !defined(ASURA_PSP_ATRAC_FILE_CACHE_SIZE)
#define ASURA_PSP_ATRAC_FILE_CACHE_SIZE		(128 * 1024)
#endif	// !defined(ASURA_PSP_ATRAC_FILE_CACHE_SIZE)

// Thread defines...
#define PLAYTHREAD_STACKSIZE   (32 * 1024)

// Constants
static const u_int uASURA_PSP_ATRAC_MAX_FILENAME_LENGTH = 260;

// Structures...
struct Asura_PSP_Atrac_Desc
{
	// NOTE: These booleans must not be put into bitfields - they are referenced in
	// ~~~~~ 2 separate threads and if you use bitfields the access operations are
	// no longer atomic and can cause nasty problems when threads swap between reading
	// data from the bitfield and writing to it. ART - 24/07/06.
	bool		m_bDecoding;
	bool		m_bPaused;
	bool		m_bPlaying;
	bool		m_bLooping;
	bool		m_bPreloading;
	bool		m_bFading;
	bool		m_bFinished;

	bool		m_bSuspended;
	bool		m_bPlayingBeforeSuspend;
	bool		m_bPausedBeforeSuspend;

	float		m_fVolume;
	u_int		m_uDecodeOffset;

	SceInt32	m_iAtracId;
	SceInt32	m_iSampleLength;
	u_int		m_uDataEnd;

	Asura_Handle m_xHandle;
};

static void Asura_PSP_Reset_Atrac_Desc( volatile Asura_PSP_Atrac_Desc* const pxAtracDesc )
{
	pxAtracDesc->m_bDecoding			= false;
	pxAtracDesc->m_bPaused				= false;
	pxAtracDesc->m_bPlaying				= false;
	pxAtracDesc->m_bLooping				= false;
	pxAtracDesc->m_bPreloading			= false;
	pxAtracDesc->m_bFading				= false;
	pxAtracDesc->m_bFinished			= false;

	pxAtracDesc->m_bSuspended			= false;
	pxAtracDesc->m_bPlayingBeforeSuspend = false;
	pxAtracDesc->m_bPausedBeforeSuspend	= false;

	pxAtracDesc->m_fVolume				= 1.0f;
	pxAtracDesc->m_uDecodeOffset		= 0;

	pxAtracDesc->m_iAtracId				= -1;
	pxAtracDesc->m_iSampleLength		= 0;
	pxAtracDesc->m_uDataEnd				= 0;

	pxAtracDesc->m_xHandle				= ASURA_INVALID_HANDLE;
}

class Asura_PSP_Atrac_File_Info
{
public:

	void SetFilename( const char* szFilename )
	{
		Asura_Assertf(strlen(szFilename) < uASURA_PSP_ATRAC_MAX_FILENAME_LENGTH, ("filename %s too long for atrac", szFilename));
		Asura_PrintfToBfr( m_acFilename, "%s", uASURA_PSP_ATRAC_MAX_FILENAME_LENGTH, true, szFilename );
	}

	// public data
	char			m_acFilename[uASURA_PSP_ATRAC_MAX_FILENAME_LENGTH];
	u_int			m_uStartOffset;
	int				m_iDataSize;
	int				m_iStreamID;
};

// Common buffer for streaming data or reading in the whole track, need to be aligned to 64 when allocated
static SceUChar8*						g_pucCommonBuf[ASURA_PSP_MAX_NUM_ATRAC_STREAMS]; 

static volatile Asura_PSP_Atrac_Desc	g_xPlayArg[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];

static Asura_Mutex						g_xAtracMutex;
static Asura_SyncEvent*					g_pxFinishPlayingEvent[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];

#if ASURA_PSP_MAX_NUM_ATRAC_STREAMS > 0
static Asura_SyncEvent					g_xFinishPlayingEvent0("Atrac0 finish playing", false);
#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS > 0

#if ASURA_PSP_MAX_NUM_ATRAC_STREAMS > 1
static Asura_SyncEvent					g_xFinishPlayingEvent1("Atrac1 finish playing", false);
#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS > 1

static Asura_PSP_Atrac_File_Info		g_xFileInfo[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];

static Asura_FileCache*					g_pxFileCache[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];
static volatile SceUID					g_iThreadID[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];

bool Asura_PSP_Atrac::s_bInitialised = false;
float Asura_PSP_Atrac::s_fMasterVolume = 1.0f;

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_Atrac::OpenFile
// Author       : Andy Tate
// Date         : 4-Jul-2006
//
// uDataSize = 0 means use the file size
//
////////////////////////////////////////////////////////////////////////////////

// iDataSize = -1 means the whole file
bool Asura_PSP_Atrac::OpenFile( const int iStreamID, const char* const szFilename, const u_int uStartOffset, const int iDataSize )
{
	bool	bSuccess	= false;

	// This loop is necessary because "Asura_FileCache::OpenFile()" can fail if the
	// disc is removed, in this instance we need to retry but outside of the mutex
	// in order to prevent a lock-up on the PSP when engaging sleep mode whilst in
	// the middle of the open - ART 04/09/06.
	do 
	{
		// Ensure that nothing else is using the Atrac data...
		g_xAtracMutex.Lock();
		{
			// Do we already have an open-file?
			if(g_xPlayArg[iStreamID].m_iAtracId >= 0)
			{
				Asura_Assert(false, "Atrac file is already open.");
			}
			else if(g_xPlayArg[iStreamID].m_bSuspended)
			{
				// Playback is suspended - don't do anything until we've been resumed...
			}
			else
			{
				g_xPlayArg[iStreamID].m_bFinished		= false;

				// Open the file...
				LOG_OUTPUT(("Opening Atrac file[%d] '%s'...\n", iStreamID, szFilename));

				const u_int uOffset = uStartOffset;

				while (!(g_pxFileCache[iStreamID]->OpenFile(szFilename, false, uOffset )))
				{
					g_xAtracMutex.Unlock();
					Asura_Thread_System::DelayCurrentThread(0.05f, false);
					g_xAtracMutex.Lock();
				}

				{
					SceInt32  iStatus;
					u_int	  uReadSize;

					// Initialise the flags...
					g_xPlayArg[iStreamID].m_uDecodeOffset	= 0;
					g_xPlayArg[iStreamID].m_iSampleLength	= -1;

					// AJG - 18/04/07 - Moved to before the file size get, as it seemed to be breaking debug builds in library version 3.1.1
					// Read the first chunk of data from the file...
					LOG_OUTPUT( ("Reading Atrac file[%d] '%s'...\n", iStreamID, szFilename) );
					g_xAtracMutex.Unlock();
					uReadSize = g_pxFileCache[iStreamID]->ReadData((4 * 1024), g_pucCommonBuf[iStreamID]);
					g_xAtracMutex.Lock();

					u_int uDataEnd;
					if ( iDataSize == -1 )
					{
						// JAT 03/04/07 get the filesize directly to allow for async operations on the open
						uDataEnd = g_pxFileCache[iStreamID]->GetFileSize();
						LOG_OUTPUT( ("Atrac file[%d] '%s' Size %d...\n", iStreamID, szFilename, uDataEnd ) );
					}
					else
					{
						uDataEnd = uStartOffset + iDataSize;
#ifdef ASURA_ASSERT
						Asura_Assert( iDataSize > 0, "" );

						// JAT 03/04/07 get the filesize directly to allow for async operations on the open
						const u_int uFileSize = g_pxFileCache[iStreamID]->GetFileSize();
						Asura_Assert( uFileSize > uDataEnd, "" );

#endif // #ifdef ASURA_ASSERT
					}
					g_xPlayArg[iStreamID].m_uDataEnd = uDataEnd;

					iStatus		= sceAtracSetDataAndGetID(g_pucCommonBuf[iStreamID], uReadSize);

					if (iStatus < 0)
					{
						Asura_Assertf(false, ("sceAtracSetDataAndGetID(): Failed %08X", iStatus));
						g_pxFileCache[iStreamID]->CloseFile();
						g_xPlayArg[iStreamID].m_bFinished		= true;
					}
					else
					{
						SceInt32	iDummy;
						SceInt32	iSampleLength;

						// Get the length in samples for this track...
						g_xPlayArg[iStreamID].m_iAtracId	= iStatus;
						iStatus					= sceAtracGetSoundSample(g_xPlayArg[iStreamID].m_iAtracId, &iSampleLength, &iDummy, &iDummy);

						if (iStatus < 0)
						{
							Asura_Assertf(false, ("sceAtracGetSoundSample(): Failed %08X", iStatus));
							CloseFile( iStreamID );
							g_xPlayArg[iStreamID].m_bFinished		= true;
							g_xPlayArg[iStreamID].m_bDecoding		= false;							
						}
						else
						{
							bSuccess					= true;
							g_xPlayArg[iStreamID].m_bFinished		= false;
							g_xPlayArg[iStreamID].m_bDecoding		= true;							
							g_xPlayArg[iStreamID].m_bPreloading	= false;
							g_xPlayArg[iStreamID].m_iSampleLength	= iSampleLength;
						}
					}
				}
			}
		}
		g_xAtracMutex.Unlock();

		// Ensure that we do something else...
		if ( !bSuccess && Asura_FileAccess::FileExists( szFilename ) )
		{
			Asura_Thread_System::DelayCurrentThread(0.02f);
		}

	// m_bFinished = true is used to mark an error condition
	// Carry on if not successful && the file exists && the file isn't flaged to be finished
	} while( !bSuccess && Asura_FileAccess::FileExists( szFilename ) && ( g_xPlayArg[iStreamID].m_bFinished	== false ) );

	return(bSuccess);
}

/// End of function 'Asura_PSP_Atrac::OpenFile'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_Atrac::CloseFile
// Author       : Andy Tate
// Date         : 4-Jul-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_Atrac::CloseFile( const int iStreamID )
{
	bool	bSuccess	= false;

	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
		// Do we already have an Atrac file?
		if(g_xPlayArg[iStreamID].m_iAtracId >= 0)
		{
			LOG_OUTPUT(("Closing Atrac file[%d].\n", iStreamID ) );
			g_pxFileCache[iStreamID]->CloseFile(true);
			sceAtracReleaseAtracID( g_xPlayArg[iStreamID].m_iAtracId);

			g_xPlayArg[iStreamID].m_bPreloading	= false;
			g_xPlayArg[iStreamID].m_bDecoding	= false;
			g_xPlayArg[iStreamID].m_iAtracId	= -1;
			bSuccess					= true;
		}
		else
		{
			Asura_Assert(false, "Attempting to close an Atrac file when we don't have one open.");
		}
	}
	g_xAtracMutex.Unlock();
	return(bSuccess);
}

/// End of function 'Asura_PSP_Atrac::CloseFile'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_Atrac::Decode
// Author       : Andy Tate
// Date         : 4-Jul-2006
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_Atrac::Decode( const int iStreamID, 
							  short* psCurrentBuffer,
							  short* psNextBuffer,
							  float& fVolume )
{
	if(!g_xPlayArg[iStreamID].m_bDecoding)
	{
		fVolume = 0.0f;
		return(false);
	}
	else if(g_xPlayArg[iStreamID].m_bPaused)
	{
		fVolume = 0.0f;
		// START - JAT 02/05/07 Fix for problem that if a stream is paused then trying to stop it causes a deadlock
		g_xAtracMutex.Lock();
		// Check for end of playing... (signalled or error)
		if ( ( g_xPlayArg[iStreamID].m_bFinished ) || ( g_xPlayArg[iStreamID].m_iAtracId == -1 ) )
		{
			if ( g_xPlayArg[iStreamID].m_bPreloading )
			{
				// Pre-loading - we're not decoding but we also don't need to signal
				// that we've finished with the file...
			}
			else
			{
				if(g_pxFinishPlayingEvent[iStreamID]->IsSignalled())
				{
					LOG_OUTPUT(("Asura_PSP_Atrac::Decode(): Already signalled\n"));
				}
				else
				{
					LOG_OUTPUT(("Asura_PSP_Atrac::Decode(): Not signalled\n"));
					g_pxFinishPlayingEvent[iStreamID]->SetSignal();
				}
			}
		}

		g_xAtracMutex.Unlock();
		// END - JAT 02/05/07 Fix for problem that if a stream is paused then trying to stop it causes a deadlock
		return(false);
	}
	else if(Asura_FileAccess::DiskExists() && !Asura_PSP_FileIO::IsSuspended())
	{
		bool	bDecoding	= false;

		// Ensure that nothing else is using the Atrac data...
		g_xAtracMutex.Lock();
		{
			if(g_xPlayArg[iStreamID].m_bDecoding && !g_xPlayArg[iStreamID].m_bPaused)
			{
				short*		psPCMData			= &psCurrentBuffer[g_xPlayArg[iStreamID].m_uDecodeOffset * ASURA_PSP_SA_NUM_STEREO_CHANNELS];
				SceUInt32	uNumSamplesToDecode	= ASURA_PSP_SA_NUM_SAMPLES - g_xPlayArg[iStreamID].m_uDecodeOffset;

				// Loop until we've decoded all the samples we need...
				while(uNumSamplesToDecode > 0)
				{
					SceUChar8*	pucWriteAddress;
					SceInt32	iRemainingFrames;
					SceInt32	iStatus;
					SceUInt32	uNumBytesToWrite;
					SceUInt32	uReadPosition;
					SceUInt32	uNumSamplesDecoded;
					SceUInt32	uFinishFlag;

					// Test for the stream has been stopped outside of our thread control
					if ( g_xPlayArg[iStreamID].m_iAtracId == -1 )
					{
						g_xPlayArg[iStreamID].m_bDecoding = false;
					}
					if ( g_xPlayArg[iStreamID].m_bDecoding == false )
					{
						uNumSamplesToDecode = 0;
						break;
					}
					// Decode the data we've just read...
					iStatus					= sceAtracDecodeData(g_xPlayArg[iStreamID].m_iAtracId, psPCMData, &uNumSamplesDecoded, &uFinishFlag, &iRemainingFrames);
					g_xPlayArg[iStreamID].m_bDecoding	= (iStatus >= 0);

					// Update the variables...
					psPCMData					+= (uNumSamplesDecoded * ASURA_PSP_SA_NUM_STEREO_CHANNELS);
					g_xPlayArg[iStreamID].m_uDecodeOffset	+= uNumSamplesDecoded;

					// Do we have any data left to decode?
					if(iStatus == SCE_ATRAC_ERROR_ALLDATA_WAS_DECODED)
					{
						// No - better finish by blanking the remaining data...
						if(g_xPlayArg[iStreamID].m_uDecodeOffset > ASURA_PSP_SA_NUM_SAMPLES)
						{
							const u_int	uOffset	= g_xPlayArg[iStreamID].m_uDecodeOffset - ASURA_PSP_SA_NUM_SAMPLES;
							const u_int	uSize	= (ASURA_PSP_SA_NUM_SAMPLES - uOffset) * sizeof(short) * ASURA_PSP_SA_NUM_STEREO_CHANNELS;

							// Blank the remainder of the next buffer...
							memset(&psNextBuffer[uOffset * ASURA_PSP_SA_NUM_STEREO_CHANNELS], 0, uSize);
						}
						else
						{
							const u_int	uOffset	= g_xPlayArg[iStreamID].m_uDecodeOffset;
							const u_int	uSize	= (ASURA_PSP_SA_NUM_SAMPLES - uOffset) * sizeof(short) * ASURA_PSP_SA_NUM_STEREO_CHANNELS;

							// Blank the remainder of the current buffer...
							memset(&psCurrentBuffer[uOffset * ASURA_PSP_SA_NUM_STEREO_CHANNELS], 0, uSize);
						}

						uNumSamplesToDecode	= 0;
					}
					else
					{
						// Decode validation...
						Asura_Assertf((iStatus >= 0), ("sceAtracDecodeData(): Error = %08X", iStatus));

						// Have we reached the the end of the file?
						const u_int uFilePos = g_pxFileCache[iStreamID]->GetFilePosition();
						const u_int uDataEnd = g_xPlayArg[iStreamID].m_uDataEnd;
						if ( uFilePos < uDataEnd )
						{
							// Set-up the Atrac decoder...
							iStatus	= sceAtracGetStreamDataInfo(g_xPlayArg[iStreamID].m_iAtracId, &pucWriteAddress, &uNumBytesToWrite, &uReadPosition);
							Asura_Assertf((iStatus >= 0), ("sceAtracGetStreamDataInfo(): Error = %08X", iStatus));

							if((uNumBytesToWrite > 0) && g_pxFileCache[iStreamID]->CanRead(uNumBytesToWrite))
							{
								SceUInt32	uNumBytesRead;

								// Read the data into the input-buffer...
								g_xAtracMutex.Unlock();
								uNumBytesRead	= g_pxFileCache[iStreamID]->ReadData(uNumBytesToWrite, pucWriteAddress);
								g_xAtracMutex.Lock();
								
								// Should also test to see if the thread was cancelled
								if ( g_xPlayArg[iStreamID].m_bFinished == true )
								{
									uNumSamplesToDecode = 0;
									break;
								}
								// Because the mutex is unlocked another thread can cancel the atrac loading
								if ( g_xPlayArg[iStreamID].m_iAtracId != -1 )
								{
									iStatus	= sceAtracAddStreamData(g_xPlayArg[iStreamID].m_iAtracId, uNumBytesRead);
									Asura_Assertf((iStatus >= 0), ("sceAtracAddStreamData(): Error = %08X", iStatus));
								}
								else
								{
									uNumSamplesToDecode = 0;
									break;
								}
							}
							else
							{
								// We don't have anything to read here - perhaps we should fail...
								Asura_Assertf(false, ("Failed to read %u bytes - aborting decode.", uNumBytesToWrite));
								uNumSamplesToDecode	= 0;
								break;
							}
						}
						// The atrac playing was cancelled
						if ( g_xPlayArg[iStreamID].m_iAtracId == -1 )
						{
							// Leave the buffer because it was decoded fine 
							// It can only be cancelled during a file read operation so there should be valid decoded data 
							// But clear out the data after the decode - just as if the decoding had finished

							if ( g_xPlayArg[iStreamID].m_uDecodeOffset > ASURA_PSP_SA_NUM_SAMPLES )
							{
								const u_int	uOffset	= g_xPlayArg[iStreamID].m_uDecodeOffset - ASURA_PSP_SA_NUM_SAMPLES;
								const u_int	uSize	= (ASURA_PSP_SA_NUM_SAMPLES - uOffset) * sizeof(short) * ASURA_PSP_SA_NUM_STEREO_CHANNELS;

								// Blank the remainder of the next buffer...
								memset( &psNextBuffer[uOffset * ASURA_PSP_SA_NUM_STEREO_CHANNELS], 0, uSize );
							}
							else
							{
								const u_int	uOffset	= g_xPlayArg[iStreamID].m_uDecodeOffset;
								const u_int	uSize	= (ASURA_PSP_SA_NUM_SAMPLES - uOffset) * sizeof(short) * ASURA_PSP_SA_NUM_STEREO_CHANNELS;

								// Blank the remainder of the current buffer...
								memset( &psCurrentBuffer[uOffset * ASURA_PSP_SA_NUM_STEREO_CHANNELS], 0, uSize );
							}

							uNumSamplesToDecode	= 0;
							uNumSamplesDecoded	= 0;
							bDecoding = false;
						}

						// Have we reached the end of our decoding?
						if(uNumSamplesToDecode > uNumSamplesDecoded)
						{
							// No - keep going...
							uNumSamplesToDecode	-= uNumSamplesDecoded;
						}
						else if(uNumSamplesToDecode == uNumSamplesDecoded)
						{
							// Yes - exactly...
							uNumSamplesToDecode			= 0;
							g_xPlayArg[iStreamID].m_uDecodeOffset	= 0;
						}
						else
						{
							const short*	psStartOfOverspill	= &psCurrentBuffer[ASURA_PSP_SA_NUM_SAMPLES * ASURA_PSP_SA_NUM_STEREO_CHANNELS];

							// Yes - and we've spilled over the end of the buffer...
							uNumSamplesToDecode			 = 0;
							g_xPlayArg[iStreamID].m_uDecodeOffset	-= ASURA_PSP_SA_NUM_SAMPLES;

							// Do we need to copy our data to the desired destination?
							if(psNextBuffer != psStartOfOverspill)
							{
								const u_int	uSize	= g_xPlayArg[iStreamID].m_uDecodeOffset * sizeof(short) * ASURA_PSP_SA_NUM_STEREO_CHANNELS;

								// Copy from the overspill into the desired destination...
								Asura_Mem::Copy(psNextBuffer, psStartOfOverspill, uSize);
							}
						}
					}
				}

				// Are we still decoding?
				if ( g_xPlayArg[iStreamID].m_bDecoding ) 
				{
					// Check for end of playing... (signalled or error)
					if ( ( g_xPlayArg[iStreamID].m_bFinished ) || ( g_xPlayArg[iStreamID].m_iAtracId == -1 ) )
					{
						bDecoding	= false;
					}
					else if ( g_xPlayArg[iStreamID].m_bFading )
					{
						g_xPlayArg[iStreamID].m_fVolume	-= 0.01f;
						bDecoding				 = (g_xPlayArg[iStreamID].m_fVolume > 0.0f);
					}
					else
					{
						bDecoding	= true;
					}
				}
				else
				{
					LOG_OUTPUT(("Decoding[%d] complete.\n", iStreamID));
					bDecoding	= false;
				}

				// Error handling - set volume to 0
				if ( g_xPlayArg[iStreamID].m_iAtracId == -1 )
				{
					fVolume = 0.0f;
				}

				// Time to return...
				if ( bDecoding )
				{
					const float fTrackVolume = g_xPlayArg[iStreamID].m_fVolume;
					const float fGroupVolume = 1.0f;
					fVolume	= s_fMasterVolume * fTrackVolume * fGroupVolume;
					fVolume	= Asura_Maths::ClampToRange(fVolume, 0.0f, 1.0f);
				}
				else if(g_xPlayArg[iStreamID].m_bPreloading)
				{
					// Pre-loading - we're not decoding but we also don't need to signal
					// that we've finished with the file...
				}
				else
				{
					if(g_pxFinishPlayingEvent[iStreamID]->IsSignalled())
					{
						LOG_OUTPUT(("Asura_PSP_Atrac[%d]::Decode(): Already signalled\n", iStreamID));
					}
					else
					{
						LOG_OUTPUT(("Asura_PSP_Atrac[%d]::Decode(): Not signalled\n", iStreamID));
						g_pxFinishPlayingEvent[iStreamID]->SetSignal();
					}
				}
			}
		}
		g_xAtracMutex.Unlock();
		return(bDecoding);
	}
	else
	{
		return(false);
	}
}

/// End of function 'Asura_PSP_Atrac::Decode'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_Atrac::UpdateThread
// Author       : Andy Tate
// Date         : 4-Jul-2006
////////////////////////////////////////////////////////////////////////////////

int Asura_PSP_Atrac::UpdateThread (SceSize args, void *argp )
{
// RDM - 17/08/06 - don't want fp exceptions at all, in any threads
#if !defined ASURA_ENABLE_FLOATING_POINT_EXCEPTIONS \
	|| defined ASURA_PSP_ARTIST_BUILD \
	|| defined ASURA_PSP_SONY_SUBMISSION

	// Disable fp exceptions if not requested
	__asm__ volatile("ctc1 %0, $31" : : "r" (0));
#endif

	const Asura_PSP_Atrac_File_Info* const pxFileInfo = (const Asura_PSP_Atrac_File_Info* const)( argp );

	// Loop until we're told to finish...
	const int iStreamID = pxFileInfo->m_iStreamID;
	do
	{
		// Clear the finish event...
		if(g_pxFinishPlayingEvent[iStreamID]->IsSignalled())
		{
			g_pxFinishPlayingEvent[iStreamID]->ClearSignal();
		}

		// Open the file...
		if ( OpenFile( iStreamID, pxFileInfo->m_acFilename, pxFileInfo->m_uStartOffset, pxFileInfo->m_iDataSize ) )
		{
			// Wait for decoding to finish...
			LOG_OUTPUT(("Asura_PSP_Atrac[%d]::UpdateThread(): Wait for finish signal\n", iStreamID));
			g_pxFinishPlayingEvent[iStreamID]->WaitForSignal();
			LOG_OUTPUT(("Asura_PSP_Atrac[%d]::UpdateThread(): Got the finish signal\n", iStreamID));

			// Close the file...
			CloseFile( iStreamID );
		}
		else
		{
			// Something went wrong - time to leave this thread and delete it
			LOG_OUTPUT( ( "StreamID[%d] 0x%X OpenFile ERRRO\n", iStreamID, g_xPlayArg[iStreamID].m_xHandle ));
			g_xPlayArg[iStreamID].m_bLooping = false;
			g_xPlayArg[iStreamID].m_bFinished = true;
		}

	} while(g_xPlayArg[iStreamID].m_bLooping && !g_xPlayArg[iStreamID].m_bFinished);

	g_xPlayArg[iStreamID].m_bLooping = false;
	g_xPlayArg[iStreamID].m_bFinished = true;

	// Only now are we no longer playing...
	LOG_OUTPUT( ( "UpdateThread() StreamID[%d] 0x%X free\n", iStreamID, g_xPlayArg[iStreamID].m_xHandle ));
	g_xPlayArg[iStreamID].m_bPlaying	= false;
	g_xPlayArg[iStreamID].m_xHandle		= ASURA_INVALID_HANDLE;

	// We're done with this thread...
	LOG_OUTPUT(("Asura_PSP_Atrac[%d]::UpdateThread(): 0x%X SCE_KERNEL_EXIT_SUCCESS\n", iStreamID, g_iThreadID[iStreamID] ));
	g_iThreadID[iStreamID] = -1;
	sceKernelExitDeleteThread( SCE_KERNEL_EXIT_SUCCESS );
	return SCE_KERNEL_EXIT_SUCCESS;
}

/// End of function 'Asura_PSP_Atrac::UpdateThread'. ///


void Asura_PSP_Atrac::Initialise( const int iStreamID )
{
	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
		g_iThreadID[iStreamID] = -1;
		g_pxFileCache[iStreamID] = NULL;

		Asura_PSP_Reset_Atrac_Desc( g_xPlayArg + iStreamID );

		g_xFileInfo[iStreamID].SetFilename("invalid");
		g_xFileInfo[iStreamID].m_uStartOffset	= 0;
		g_xFileInfo[iStreamID].m_iDataSize		= 0;
		g_xFileInfo[iStreamID].m_iStreamID		= -1;

		Asura_Assert(!g_pucCommonBuf[iStreamID], "Already allocated buffer?");

		// Allocate a smaller temporary buffer to decode track's bits
		char szTitle[32];
		Asura_PrintfToBfr( szTitle, "Atrac File-Cache[%d]", 32, false, iStreamID );
		g_pxFileCache[iStreamID]	= new Asura_FileCache( szTitle, ASURA_PSP_ATRAC_FILE_CACHE_SIZE, 2, false);
		g_pucCommonBuf[iStreamID]	= (SceUChar8*) Asura_AllocateAlign(ASURA_PSP_ATRAC_INPUT_BUFFER_SIZE + 2, ASURA_ALIGN_16_BYTE); 
	}
	g_xAtracMutex.Unlock();
}

void Asura_PSP_Atrac::Shutdown( void )
{
	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
		LOG_OUTPUT(("Asura_PSP_Atrac::Shutdown()\n"));
		Asura_Assert(s_bInitialised, "Not initialised");

		s_bInitialised = false;	

		// stop any tracks, if any
		Reset();

		// Free buffer 
		for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
		{
			delete( g_pxFileCache[i] );
			Asura_DeallocateAlign( g_pucCommonBuf[i] ); 
			g_pxFileCache[i]	= NULL;
			g_pucCommonBuf[i]	= 0;
		}
	}
	g_xAtracMutex.Unlock();
}

void Asura_PSP_Atrac::Suspend( const int iStreamID )
{
	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
		LOG_OUTPUT(("Asura_PSP_Atrac[%d]::Suspend()\n", iStreamID));
		Asura_Handle xHandle = g_xPlayArg[iStreamID].m_xHandle;

		g_xPlayArg[iStreamID].m_bSuspended				= true;
		g_xPlayArg[iStreamID].m_bPlayingBeforeSuspend	= g_xPlayArg[iStreamID].m_bPlaying;
		g_xPlayArg[iStreamID].m_bPausedBeforeSuspend	= g_xPlayArg[iStreamID].m_bPaused;

		if(!g_xPlayArg[iStreamID].m_bPausedBeforeSuspend)
		{
			PauseTrack( xHandle, true);
		}
	}
	g_xAtracMutex.Unlock();
}

void Asura_PSP_Atrac::Resume( const int iStreamID )
{
	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
		LOG_OUTPUT(("Asura_PSP_Atrac[%d]::Resume()\n", iStreamID));
		Asura_Handle xHandle = g_xPlayArg[iStreamID].m_xHandle;

		if(g_xPlayArg[iStreamID].m_bPlayingBeforeSuspend && !g_xPlayArg[iStreamID].m_bPausedBeforeSuspend)
		{
			PauseTrack( xHandle, false);
		}

		g_xPlayArg[iStreamID].m_bSuspended				= false;
		g_xPlayArg[iStreamID].m_bPlayingBeforeSuspend	= false;
		g_xPlayArg[iStreamID].m_bPausedBeforeSuspend	= false;
	}
	g_xAtracMutex.Unlock();
}

void Asura_PSP_Atrac::Reset( const int iStreamID )
{
	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
		
		LOG_OUTPUT(("Asura_PSP_Atrac[%d]::Reset()\n", iStreamID));

		// stop any tracks, if any
		if( g_iThreadID[iStreamID] != -1 )
		{
			StopStream( iStreamID, false );
		}		

		g_xPlayArg[iStreamID].m_iAtracId		= -1;

		g_xPlayArg[iStreamID].m_bDecoding		= false;
		g_xPlayArg[iStreamID].m_bPaused		= false;
		g_xPlayArg[iStreamID].m_bPlaying		= false;
		g_xPlayArg[iStreamID].m_bLooping		= false;
		g_xPlayArg[iStreamID].m_bPreloading	= false;
		g_xPlayArg[iStreamID].m_bFading		= false;
		g_xPlayArg[iStreamID].m_bFinished		= false;
	}
	// Allow access to the Atrac code...
	g_xAtracMutex.Unlock();
}


float Asura_PSP_Atrac::GetLength( const Asura_Handle xHandle )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return 0.0f;
	}
	// In samples...
	const int iStreamID = FindStreamID( xHandle );
	Asura_Assert( ( iStreamID >= 0 && iStreamID < ASURA_PSP_MAX_NUM_ATRAC_STREAMS ), "" );
	if ( iStreamID < 0 )
	{
		LOG_OUTPUT(("Asura_PSP_Atrac::GetLength() 0x%X bad stream ID", xHandle ));
        return 0.0f;
	}


	return static_cast<float> (g_xPlayArg[iStreamID].m_iSampleLength / (44100.0f)); 
}

float Asura_PSP_Atrac::GetPosition( const Asura_Handle xHandle )
{
	// Decoded data, so it is in samples...
	// KOR - because this is the decode position, not the play position, we have to offset it.
	// Yes, it's a magic number, but it works reasonably well for our purposes. GetPosition is
	// most useful for the meta-music system - the worst that can happen is that we report a
	// low time and it starts playing the next beat before changing track, resulting in a nasty
	// sound glitch. Therefore it is better that we use an overestimate than under.
//	return static_cast<float> (g_xPlayArg[iStreamID].m_iPlayPosition / (44100.0f)) + 0.1f; 
	return(0.0f);
}

// iDataSize = -1 means use the file size
bool Asura_PSP_Atrac::Create( const Asura_Handle xHandle, 
							  const char* const szFilename, const u_int uStartOffset, const int iDataSize)
{
	bool bSuccess = false;

	// Ensure that nothing else is using the Atrac data...
	g_xAtracMutex.Lock();
	{
#if 1
		// Find a streamID for this xHandle
		const int iStreamID = FindFreeStreamID( xHandle );
		LOG_OUTPUT( ( "Create[%d] 0x%X -> %d\n", iStreamID, xHandle, iStreamID ) );
		if ( iStreamID < 0 )
		{
			g_xAtracMutex.Unlock();
			return false;
		}

		// Do we already have a file open?
		if ( ( g_xPlayArg[iStreamID].m_iAtracId >= 0 ) || ( g_iThreadID[iStreamID] >= 0 ) )
		{
			// Have we stopped the thread?
			Asura_Assert(false, "Already playing an Atrac file - will stop this before starting the new one.");
			LOG_OUTPUT( ( "Already playing an Atrac file - will stop this before starting the new one.") );
			if(!g_xPlayArg[iStreamID].m_bFinished)
			{
				g_xAtracMutex.Unlock();
				StopStream( iStreamID, false );
				g_xAtracMutex.Lock();
			}

			// Wait for the track to finish...
			while(g_xPlayArg[iStreamID].m_iAtracId >= 0)
			{
				Asura_Thread_System::YieldCurrentThread();
			}
		}

		// Does the file exist?
		if(Asura_FileAccess::FileExists(szFilename))
		{
			SceUID	iThreadID;

			// Create a new thread...
			g_xPlayArg[iStreamID].m_bDecoding	= false;
			g_xPlayArg[iStreamID].m_bPaused		= false;
			g_xPlayArg[iStreamID].m_bPlaying	= false;
			g_xPlayArg[iStreamID].m_bLooping	= false;
			g_xPlayArg[iStreamID].m_bPreloading	= false;
			g_xPlayArg[iStreamID].m_bFading		= false;
			g_xPlayArg[iStreamID].m_bFinished	= false;
			g_xPlayArg[iStreamID].m_uDataEnd	= 0;
			g_xPlayArg[iStreamID].m_xHandle		= xHandle;

			LOG_OUTPUT( ( "StreamID[%d] 0x%X alloc\n", iStreamID, xHandle ) );

			g_xFileInfo[iStreamID].SetFilename(szFilename);
			g_xFileInfo[iStreamID].m_uStartOffset = uStartOffset;
			g_xFileInfo[iStreamID].m_iDataSize	  = iDataSize;
			g_xFileInfo[iStreamID].m_iStreamID	  = iStreamID;

			char szTitle[32];
			Asura_PrintfToBfr( szTitle, "Atrac Playing[%d]", 32, false, iStreamID );
			iThreadID = sceKernelCreateThread( szTitle, UpdateThread, iASURA_THREAD_PRIORITY_ATRAC_UPDATE_THREAD, PLAYTHREAD_STACKSIZE, SCE_KERNEL_TH_USE_VFPU, NULL);
			if (iThreadID < 0)
			{
				Asura_Assertf(false, ("sceKernelCreateThread(): Failed %08X\n", iThreadID));
			}
			else
			{
				SceInt32	iStatus	= sceKernelStartThread( iThreadID, sizeof( g_xFileInfo ), g_xFileInfo+iStreamID );

				if ( iStatus == SCE_KERNEL_ERROR_OK )
				{
					g_xPlayArg[iStreamID].m_bPaused		= true;
					g_xPlayArg[iStreamID].m_bPlaying	= true;
					g_xPlayArg[iStreamID].m_bPreloading	= true;
					g_iThreadID[iStreamID]				= iThreadID;
					bSuccess					= true;
					LOG_OUTPUT( ( "StreamID[%d] 0x%X thread start\n", iStreamID, iThreadID ) );
				}
				else
				{
					Asura_Assertf(false, ("sceKernelStartThread:0x%08X\n",iStatus));
				}
			}
		}
		else
		{
			Asura_Assertf(false, ("Cannot create Atrac for \"%s\" - the file doesn't exist.", szFilename));
			LOG_OUTPUT( ( "Cannot create Atrac for \"%s\" - the file doesn't exist.", szFilename) );
		}
#endif
	}
	g_xAtracMutex.Unlock();
	return(bSuccess);
}

void Asura_PSP_Atrac::Release( const Asura_Handle xHandle )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return;
	}
	const int iStreamID = FindStreamID( xHandle );
	ReleaseStream( iStreamID );
}

void Asura_PSP_Atrac::ReleaseStream( const int iStreamID )
{
	// Ensure that nothing else is using the Atrac data...

	int	iThreadID = -1;
	g_xAtracMutex.Lock();

	if ( iStreamID >= 0 )
	{
		// Save thie variable here to stop 2 threads changing it
		g_xAtracMutex.Lock();
		iThreadID	= g_iThreadID[iStreamID];
		g_xAtracMutex.Unlock();

		// Ignore this if the atrac has finished playing
		if ( g_xPlayArg[iStreamID].m_bFinished == false )
		{

			// Check if thread has been created, then release it
			Asura_Assert((g_pxFileCache[iStreamID] != NULL) && g_pxFileCache[iStreamID]->IsOpen(), "No file was open"); 
			Asura_Assert(g_iThreadID[iStreamID] != -1, "No thread was created"); 

			// Set flag for music to stop
			g_xPlayArg[iStreamID].m_bFinished	= true;
			g_xPlayArg[iStreamID].m_bDecoding	= false;
			g_xPlayArg[iStreamID].m_bLooping	= false;
			g_xPlayArg[iStreamID].m_bPaused		= true;

			// Ensure that the thread knows we're finished...
			if(g_pxFinishPlayingEvent[iStreamID]->IsSignalled())
			{
				LOG_OUTPUT(("Asura_PSP_Atrac[%d]::ReleaseStream(): Already signalled\n", iStreamID));
			}
			else
			{
				LOG_OUTPUT(("Asura_PSP_Atrac[%d]::ReleaseStream(): Not signalled\n", iStreamID));
				g_pxFinishPlayingEvent[iStreamID]->SetSignal();
			}
		}
	}
	g_xAtracMutex.Unlock();

	// Kill thread once it's faded
	if ( iThreadID >= 0 )
	{
		// If the stream is playing sleep for a bit to give the thread a chance to notice the finish signal and handle things
		LOG_OUTPUT( ( "ReleaseStream StreamID[%d] 0x%X sleep 0x%X\n", iStreamID, g_xPlayArg[iStreamID].m_xHandle, iThreadID ));
		sceKernelDelayThread( 20000 ); 

		sceKernelWaitThreadEndCB( iThreadID, NULL );

		sceKernelDeleteThread( iThreadID );
		if ( iStreamID >= 0 )
		{
			LOG_OUTPUT( ( "ReleaseStream StreamID[%d] 0x%X free\n", iStreamID, g_xPlayArg[iStreamID].m_xHandle ));
			g_xPlayArg[iStreamID].m_xHandle		= ASURA_INVALID_HANDLE;
		}
	}
	if ( iStreamID >= 0 )
	{
		g_xAtracMutex.Lock();
		g_iThreadID[iStreamID] = -1;
		g_xAtracMutex.Unlock();
	}
}

//////////////////////////////////////////////////////////////////////////
// NOTE: These functions used to use mutexs but since the operation of
// ~~~~~ streamed sounds is "Create" followed immediately by "Play".
// "Create" starts the thread which immediately grabs the mutex whilst
// opening the file, so the "Play" instruction was then blocking until
// the file was opened (sometimes in excess of a second) thus causing big
// halts when playing the game. I've changed the code to use atomic
// operations on the appropriate flags. ART - 24/07/06.
//////////////////////////////////////////////////////////////////////////

void Asura_PSP_Atrac::Play( const Asura_Handle xHandle, const bool bLooping )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return;
	}
	const int iStreamID = FindStreamID( xHandle );
	Asura_Assert( ( iStreamID >= 0 && iStreamID < ASURA_PSP_MAX_NUM_ATRAC_STREAMS ), "" );
	if ( iStreamID < 0 )
	{
		LOG_OUTPUT(("Asura_PSP_Atrac::Play() 0x%X bad stream ID", xHandle ));
		return;
	}
	LOG_OUTPUT(("Asura_PSP_Atrac[%d]::Play(): bLooping = %d\n", iStreamID, bLooping));
	g_xPlayArg[iStreamID].m_bPaused	= false;
	g_xPlayArg[iStreamID].m_bLooping	= bLooping;
}

void Asura_PSP_Atrac::PauseTrack( const Asura_Handle xHandle, const bool bPauseOn )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return;
	}

	const int iStreamID = FindStreamID( xHandle );
	Asura_Assert( ( iStreamID >= 0 && iStreamID < ASURA_PSP_MAX_NUM_ATRAC_STREAMS ), "" );

    if( iStreamID >= 0 )
    {
	    LOG_OUTPUT(("Asura_PSP_Atrac[%d]::PauseTrack(): bPauseOn = %d\n", iStreamID, bPauseOn));
	    g_xPlayArg[iStreamID].m_bPaused = bPauseOn;
    }
}

void Asura_PSP_Atrac::StopTrack( const Asura_Handle xHandle, const bool bFadeOut )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return;
	}

	const int iStreamID = FindStreamID( xHandle );
	Asura_Assert( ( iStreamID >= 0 && iStreamID < ASURA_PSP_MAX_NUM_ATRAC_STREAMS ), "" );

    if( iStreamID >= 0 )
    {
	    StopStream( iStreamID, bFadeOut );
    }
}

// Internal function - stops a streamID
void Asura_PSP_Atrac::StopStream( const int iStreamID, const bool bFadeOut )
{
	LOG_OUTPUT(("Asura_PSP_Atrac[%d]::StopStream(): bFadeOut = %d\n", iStreamID, bFadeOut));
	if ( bFadeOut )
	{
		g_xPlayArg[iStreamID].m_bFading	= true;
	}
	else
	{
		ReleaseStream( iStreamID );
	}
}

bool Asura_PSP_Atrac::IsPaused( const Asura_Handle xHandle )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return false;
	}
	const int iStreamID = FindStreamID( xHandle );
	if ( iStreamID < 0 )
	{
		// Not paused if the handle doesn't exist
		return false;
	}
	return(g_xPlayArg[iStreamID].m_bPaused);
}

bool Asura_PSP_Atrac::IsPlaying( const Asura_Handle xHandle )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return false;
	}
	const int iStreamID = FindStreamID( xHandle );
	if ( iStreamID < 0 )
	{
		// Not playing if the handle doesn't exist
		return false;
	}
	return(g_xPlayArg[iStreamID].m_bPlaying);
}

bool Asura_PSP_Atrac::IsPreloading( const Asura_Handle xHandle )
{
	if ( xHandle == ASURA_INVALID_HANDLE )
	{
		return false;
	}
	const int iStreamID = FindStreamID( xHandle );
	if ( iStreamID < 0 )
	{
		// Not preloading if the handle doesn't exist
		return false;
	}
	return(g_xPlayArg[iStreamID].m_bPreloading);
}

// Global functions which work on all the atrac streams
void Asura_PSP_Atrac::Initialise( void )
{
	LOG_OUTPUT(("Asura_PSP_Atrac::Initialise\n"));
	Asura_Assert(!s_bInitialised, "Already initiliased");
	s_bInitialised = true;

	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		g_pxFinishPlayingEvent[0] = NULL;
	}

	g_pxFinishPlayingEvent[0] = &g_xFinishPlayingEvent0;
#if ASURA_PSP_MAX_NUM_ATRAC_STREAMS > 1
	g_pxFinishPlayingEvent[1] = &g_xFinishPlayingEvent1;
#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2

	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		Initialise( i );
	}
}

void Asura_PSP_Atrac::Suspend( void )
{
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		Suspend( i );
	}
}

void Asura_PSP_Atrac::Resume( void )
{
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		Resume( i );
	}
}

void Asura_PSP_Atrac::Reset( void )
{
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		Reset( i );
	}
}

// JAT 29/03/07 - stop all active streams
void Asura_PSP_Atrac::StopAll( const bool bFadeOut )
{
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		const Asura_Handle xHandle = g_xPlayArg[i].m_xHandle;
		if ( xHandle != ASURA_INVALID_HANDLE )
		{
			StopTrack( xHandle, bFadeOut );
		}
	}
}

// set to true to pause all tracks, call again with false to resume all tracks
void Asura_PSP_Atrac::PauseAll( const bool bPauseOn )
{
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		const Asura_Handle xHandle = g_xPlayArg[i].m_xHandle;
		if ( xHandle != ASURA_INVALID_HANDLE )
		{
			PauseTrack( xHandle, bPauseOn );
		}
	}
}
// Find the stream ID for this handle
int Asura_PSP_Atrac::FindStreamID( const Asura_Handle xHandle )
{
	int iStreamID = -1;
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		if ( g_xPlayArg[i].m_xHandle == xHandle )
		{
			return i;
		}
	}
	return iStreamID;
}

// Like FindStreamID but also returns the first free streamID
int Asura_PSP_Atrac::FindFreeStreamID( const Asura_Handle xHandle )
{
	int iStreamID = -1;
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		if ( g_xPlayArg[i].m_xHandle == xHandle )
		{
			return i;
		}
		if ( iStreamID == -1 )
		{
			if ( g_xPlayArg[i].m_xHandle == ASURA_INVALID_HANDLE )
			{
				iStreamID = i;
			}
		}
	}

	// Special case - if only one stream then return 0 if none free - so streams can interupt each other
#if ( ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 1 )
	iStreamID = 0;
#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 1

	return iStreamID;
}

void Asura_PSP_Atrac::SetVolume( const Asura_Handle xHandle, const float fVolume )
{
	const int iStreamID = FindStreamID( xHandle );
	Asura_Assert( ( iStreamID >= 0 && iStreamID < ASURA_PSP_MAX_NUM_ATRAC_STREAMS ), "" );

    if( iStreamID >= 0 )
    {
	    g_xPlayArg[iStreamID].m_fVolume = fVolume;
    }
}

/// End of file. ///
