
///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PSP_SimpleAudio.cpp
// Author : Andy Tate
// Date   : 21-Jun-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Sound/Asura_PSP_SimpleAudio.h"

// Asura independent header-files...
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_Profiling_System.h"

// Asura platform header-files...
#include "Core/Asura_PSP_Thread_Priorities.h"
#include "MoviePlayback/Asura_PSP_MoviePlayback.h"
#include "Sound/Asura_PSP_Atrac.h"
#include "Sound/Asura_PSP_SAS.h"

// Sony header-files...
#include <audiooutput.h>
#include <libsas.h>
#include <psperror.h>	// for SCE_OK
#include <string.h>

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#define	ASURA_PSP_SIMPLE_AUDIO_THREAD_STACKSIZE	(32 * 1024)

DEBUGVAR float dbg_fVolScaleChangeRate = 0.02f;
#if 0
float s_fAtracMixSpeed = 0.0f;
#endif // #if 0

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

// The output buffers to use (including one overspill buffer for Atrac decoding)...
static short	g_aaasOutputBuffer[ASURA_PSP_MAX_NUM_ATRAC_STREAMS][ASURA_PSP_SA_NUM_OUTPUT_BUFFERS + 1][ASURA_PSP_SA_NUM_SAMPLES * ASURA_PSP_SA_NUM_STEREO_CHANNELS] __attribute__((aligned(64)));

// flag for thread running
static volatile bool g_bSimpleAudioRunning = true;

// Statics used by this class...
int		Asura_PSP_SimpleAudio::s_iThreadID		= -1;
u_int	Asura_PSP_SimpleAudio::s_uCurrentBuffer[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

bool Asura_PSP_SimpleAudio::DoAtracMixing( short*& psOutputBufferOut, float& fAtracVolumeOut )
{
	bool	bHasAtracData = false;

	float	fAtracVolume = 1.0f;

	// Use the output buffer from stream 0 as the output buffer
	short* const psOutputBuffer = &g_aaasOutputBuffer[0][ s_uCurrentBuffer[0] ][0];
	psOutputBufferOut = psOutputBuffer;

	// Initialise the buffers...
#if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2
	short* apsAtracBuffer[ASURA_PSP_MAX_NUM_ATRAC_STREAMS];
#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2
	float fAtracVolume0 = 0.0f;
	float fAtracVolume1 = 0.0f;
	u_int uStreamsWithData = 0;
	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		short* const psCurrentBuffer = &g_aaasOutputBuffer[i][ s_uCurrentBuffer[i] ][0];
#if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2
		apsAtracBuffer[i] = psCurrentBuffer;
#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2
		s_uCurrentBuffer[i]	= ( s_uCurrentBuffer[i] + 1 ) % ASURA_PSP_SA_NUM_OUTPUT_BUFFERS;
		short* const psNextBuffer = &g_aaasOutputBuffer[i][ s_uCurrentBuffer[i] ][0];

		// Mix any audio into it...
		float fVolume;
		const bool bStreamHasData = Asura_PSP_Atrac::Decode( i, psCurrentBuffer, psNextBuffer, fVolume) && ( fVolume > 0.0f );
		// Zero out the stream data
		if ( bStreamHasData )
		{
			if (i==0)
			{
				fAtracVolume0 = fVolume;
			}
			else
			{
				fAtracVolume1 = fVolume;
			}
			++uStreamsWithData;
		}
		bHasAtracData |= bStreamHasData;
	}

#if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 1
	fAtracVolume = fAtracVolume0;
#elif ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2

	// Mix the atrac streams together into one output buffer
	// Use the output buffer from stream 0 as the output buffer
	const float fASURA_PSP_ATRAC_VOL_SCALE1 = 0.75f;
	const float fASURA_PSP_ATRAC_VOL_SCALE2 = 0.5f;
	static float ls_fVolumeScale = fASURA_PSP_ATRAC_VOL_SCALE1;
	float fDesiredVolumeScale = (uStreamsWithData == 2) ? fASURA_PSP_ATRAC_VOL_SCALE2 : fASURA_PSP_ATRAC_VOL_SCALE1;// was ( 1.0f / 4.0f );
	// try to smoothly scale between the 2 volumes
	if (ls_fVolumeScale < fDesiredVolumeScale)
	{
		ls_fVolumeScale = Asura_Maths::Min( fDesiredVolumeScale, ls_fVolumeScale + dbg_fVolScaleChangeRate );
	}
	else if (ls_fVolumeScale > fDesiredVolumeScale)
	{
		ls_fVolumeScale = Asura_Maths::Max( fDesiredVolumeScale, ls_fVolumeScale - dbg_fVolScaleChangeRate );
	}

	if ( bHasAtracData )
	{
		const float fVolume0 = fAtracVolume0 * ls_fVolumeScale;
		const float fVolume1 = fAtracVolume1 * ls_fVolumeScale;

		// Mix the atrac streams together into one output buffer

		// Do this 8 samples at a time using VFPU operations
		// lv.q, - load quad word (8 samples)
		// vs2i.xy, - short to int
		// vi2f.q,  - int to float
		// vscl.q,  - vector multiply
		// vadd.q, - add together
		// vf2id.q, - float to int
		// vi2s.q,  - int to short
		// svl.q wb  - store quad word (unaligned)
#if 1 && defined ASURA_PSP_USE_VFPU_ASSEMBLER
		short* psNewOutputBuffer = psOutputBuffer;
		short* psInputBuffer0 = apsAtracBuffer[0];
		short* psInputBuffer1 = apsAtracBuffer[1];

		// Put these into the correct VFPU registers
		__asm__ volatile
		(	

			"mtv %0, s200\n"					// volume1 scale
			"mtv %1, s201\n"					// volume2 scale
				:
				: "r" (fVolume0), "r" (fVolume1)
		);

		// Tried interleaving the loads with the mixing but it just causes other stalls :(
		// Now doing 16 samples at once e.g. 32-bytes at once
		for( u_int u = 0; u < ASURA_PSP_SA_NUM_SAMPLES * ASURA_PSP_SA_NUM_STEREO_CHANNELS; u += 16 )
		{
			__asm__ volatile
			(	
				// Load all 16 samples into memory
				"lv.q c600, 0(%1)\n"				// apsAtracBuffer[0][sample+0:sample+7]
				"lv.q c700, 0(%2)\n"				// apsAtracBuffer[1][sample+0:sample+7]
				"lv.q c610, 16(%1)\n"				// apsAtracBuffer[0][sample+8:sample+15]
				"lv.q c710, 16(%2)\n"				// apsAtracBuffer[1][sample+8:sample+15]

				// Interleave the first 4, second 4, third 4 and fourth 4 samples to try and remove stalls

				// First 4 samples
				"vs2i.p c400, c600\n"				// short to int (first 4 samples)
				"vs2i.p c500, c700\n"				// short to int (first 4 samples)

				// Second 4 samples
				"vs2i.p c410, c602\n"				// short to int (second 4 samples - c602)
				"vs2i.p c510, c702\n"				// short to int (second 4 samples - c702)

				// Third 4 samples
				"vs2i.p c420, c610\n"				// short to int (third 4 samples)
				"vs2i.p c520, c710\n"				// short to int (third 4 samples)

				// Fourth 4 samples
				"vs2i.p c430, c612\n"				// short to int (fourth 4 samples - c612)
				"vs2i.p c530, c712\n"				// short to int (fourth 4 samples - c712)

				// First 4 samples
				"vi2f.q c400, c400, 14\n"			// int to float (first 4 samples)
				"vi2f.q c500, c500, 14\n"			// int to float (first 4 samples)

				// Second 4 samples
				"vi2f.q c410, c410, 14\n"			// int to float (second 4 samples)
				"vi2f.q c510, c510, 14\n"			// int to float (second 4 samples)

				// Third 4 samples
				"vi2f.q c420, c420, 14\n"			// int to float (third 4 samples)
				"vi2f.q c520, c520, 14\n"			// int to float (third 4 samples)

				// Fourth 4 samples
				"vi2f.q c430, c430, 14\n"			// int to float (fourth 4 samples)
				"vi2f.q c530, c530, 14\n"			// int to float (fourth 4 samples)

				// First 4 samples
				"vscl.q c400, c400, s200\n"			// volume1 scale (first 4 samples)
				"vscl.q c500, c500, s201\n"			// volume2 scale (first 4 samples)

				// Second 4 samples
				"vscl.q c410, c410, s200\n"			// volume1 scale (second 4 samples)
				"vscl.q c510, c510, s201\n"			// volume2 scale (second 4 samples)

				// Third 4 samples
				"vscl.q c420, c420, s200\n"			// volume1 scale (third 4 samples)
				"vscl.q c520, c520, s201\n"			// volume2 scale (third 4 samples)

				// Fourth 4 samples
				"vscl.q c430, c430, s200\n"			// volume1 scale (fourth 4 samples)
				"vscl.q c530, c530, s201\n"			// volume2 scale (fourth 4 samples)

				// First 4 samples
				"vadd.q c400, c400, c500\n"			// mix the streams (first 4 samples)

				// Second 4 samples
				"vadd.q c410, c410, c510\n"			// mix the streams (second 4 samples)

				// Third 4 samples
				"vadd.q c420, c420, c520\n"			// mix the streams (third 4 samples)

				// Fourth 4 samples
				"vadd.q c430, c430, c530\n"			// mix the streams (fourth 4 samples)

				// First 4 samples
				"vf2in.q c400, c400, 14\n"			// float to nearest int (first 4 samples)

				// Second 4 samples
				"vf2in.q c410, c410, 14\n"			// float to nearest int (second 4 samples)

				// Third 4 samples
				"vf2in.q c420, c420, 14\n"			// float to nearest int (third 4 samples)

				// Fourth 4 samples
				"vf2in.q c430, c430, 14\n"			// float to nearest int (fourth 4 samples)

				// First 4 samples
				"vi2s.q c300, c400\n"				// int to short (first 4 samples)

				// Second 4 samples
				"vi2s.q c302, c410\n"				// int to short (second 4 samples)

				// Third 4 samples
				"vi2s.q c310, c420\n"				// int to short (third 4 samples)

				// Fourth 4 samples
				"vi2s.q c312, c430\n"				// int to short (fourth 4 samples)

				"addiu %1, %1, 0x20\n"				// psInputBuffer0 += 16
				"sv.q c300, 0(%0)\n"				// psOuputBuffer[sample:sample+7]
				"addiu %2, %2, 0x20\n"				// psInputBuffer1 += 16
				"sv.q c310, 16(%0)\n"				// psOuputBuffer[sample+8:sample+15]
	
				:: "r" (psNewOutputBuffer), "r" (psInputBuffer0), "r" (psInputBuffer1)
			);
			psNewOutputBuffer += 16;
		}

#else // #if defined ASURA_PSP_USE_VFPU_ASSEMBLER

		for( u_int u = 0; u < ASURA_PSP_SA_NUM_SAMPLES * ASURA_PSP_SA_NUM_STEREO_CHANNELS; u += 1 )
		{
			const int iSample0 = apsAtracBuffer[0][u];
			const float fSample0 = iSample0;

			const int iSample1 = apsAtracBuffer[1][u];
			const float fSample1 = iSample1;

			const float fSum = fSample0 * fVolume0 + fSample1 * fVolume1;
			int iSum = static_cast<int>( fSum );

			iSum = __builtin_allegrex_min(32767, iSum);
			iSum = __builtin_allegrex_max(-32767, iSum);
			psOutputBuffer[u] = iSum;
		}
#endif // #if defined ASURA_PSP_ASM_USE_SN_SYNTAX
	}

#else // #elif ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 2

#error ASURA_PSP_MAX_NUM_ATRAC_STREAMS can only be 1 or 2

#endif // #if ASURA_PSP_MAX_NUM_ATRAC_STREAMS == 1

	fAtracVolumeOut = fAtracVolume;
	return bHasAtracData;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SimpleAudio::UpdateThread
// Author       : Andy Tate
// Date         : 21-Jun-2006
////////////////////////////////////////////////////////////////////////////////


int Asura_PSP_SimpleAudio::UpdateThread (SceSize args, ScePVoid argp)
{
// RDM - 17/08/06 - don't want fp exceptions at all, in any threads
#if !defined ASURA_ENABLE_FLOATING_POINT_EXCEPTIONS \
	|| defined ASURA_PSP_ARTIST_BUILD \
	|| defined ASURA_PSP_SONY_SUBMISSION

	// Disable fp exceptions if not requested
	__asm__ volatile("ctc1 %0, $31" : : "r" (0));
#endif

	// Blank the current buffer...
	Asura_Mem::Set(g_aaasOutputBuffer, 0, sizeof(g_aaasOutputBuffer));

	// Loop forever...
	while (g_bSimpleAudioRunning)
	{
		bool	bHasAtracData;
		bool	bHasMovieData;
		bool	bHasSASData;
		float	fMovieVolume;

		// Atrac mixing now done in another function to help profiling and maintenance.
		short* psOutputBuffer;
		float fAtracVolume;
#if 0
		Asura_Handle xTimingHandle = Asura_Profiling_CreateTimer();
		Asura_Profiling_StartTiming( xTimingHandle );
		bHasAtracData	= DoAtracMixing(psOutputBuffer, fAtracVolume);
		Asura_TickCount xTicks = Asura_Profiling_StopTiming( xTimingHandle );
		Asura_Profiling_DestroyTimer( xTimingHandle );
		const float fTime = Asura_Timers::TicksToSeconds( xTicks );
		s_fAtracMixSpeed = s_fAtracMixSpeed * 0.95f + fTime * 0.05f;
#else
		bHasAtracData	= DoAtracMixing(psOutputBuffer, fAtracVolume);
#endif // #if 0


#if defined(ASURA_SUPPORT_PSP_MPEG)
		bHasMovieData	= Asura_PSP_MoviePlayback::DecodeAudio( psOutputBuffer, fMovieVolume ) && ( fMovieVolume > 0.0f );
#else
		bHasMovieData	= false;
		fMovieVolume = 0.0f;
#endif	// defined(ASURA_SUPPORT_PSP_MPEG)

		Asura_PSP_SAS::BeginUsage();
		{
			bHasSASData		= Asura_PSP_SAS::Decode();

			// Mix the SAS and Atrac data (if necessary)...
			if(bHasSASData)
			{
				if(bHasMovieData)
				{
					const int	iVolume	= Asura_Maths::FloatToInt(fMovieVolume * SCE_SAS_VOLUME_MAX);

					// Mix the SAS data with the Atrac audio...
					for(u_int u = 0; u < ASURA_PSP_SA_NUM_SAMPLES; u += ASURA_PSP_SA_AUDIO_OUTPUT_SAMPLE)
					{
#if defined(ASURA_ASSERT)
						int	iStatus	=
#endif	// defined(ASURA_ASSERT)
						sceSasCoreWithMix(&psOutputBuffer[u * ASURA_PSP_SA_NUM_STEREO_CHANNELS], iVolume, iVolume);
						Asura_Assertf((iStatus == SCE_OK), ("sceSasCoreWithMix(): Failed %08X", iStatus));
					}
				}
				else if(bHasAtracData)
				{
					const int	iVolume	= Asura_Maths::FloatToInt(fAtracVolume * SCE_SAS_VOLUME_MAX);

					// Mix the SAS data with the Atrac audio...
					for(u_int u = 0; u < ASURA_PSP_SA_NUM_SAMPLES; u += ASURA_PSP_SA_AUDIO_OUTPUT_SAMPLE)
					{
#if defined(ASURA_ASSERT)
						int	iStatus	=
#endif	// defined(ASURA_ASSERT)
						sceSasCoreWithMix(&psOutputBuffer[u * ASURA_PSP_SA_NUM_STEREO_CHANNELS], iVolume, iVolume);
						Asura_Assertf((iStatus == SCE_OK), ("sceSasCoreWithMix(): Failed %08X", iStatus));
					}
				}
				else
				{
					// Use the SAS audio...
					for(u_int u = 0; u < ASURA_PSP_SA_NUM_SAMPLES; u += ASURA_PSP_SA_AUDIO_OUTPUT_SAMPLE)
					{
#if defined(ASURA_ASSERT)
						int	iStatus	=
#endif	// defined(ASURA_ASSERT)
						sceSasCore(&psOutputBuffer[u * ASURA_PSP_SA_NUM_STEREO_CHANNELS]);
						Asura_Assertf((iStatus == SCE_OK), ("sceSasCore(): Failed %08X", iStatus));
					}
				}
			}
			else if(bHasMovieData)
			{
				Asura_Assert(!bHasAtracData, "Cannot have both Atrac and Movie audio data.");
			}
			else if(bHasAtracData)
			{
				Asura_Assert(!bHasMovieData, "Cannot have both Atrac and Movie audio data.");
			}
			else
			{
				// We don't have SAS or Atrac data so we need to ensure that we
				// have silence as opposed to the last sounds we played being
				// constantly looped.
				Asura_Mem::Set( psOutputBuffer, 0, sizeof( g_aaasOutputBuffer[0][0] ) );
			}
		}
		Asura_PSP_SAS::EndUsage();

		// Send the necessary data to the hardware...
		if(bHasAtracData || bHasSASData)
		{
			sceAudioOutput2OutputBlocking(SCE_AUDIO_VOLUME_0dB, psOutputBuffer);
		}
	}

	// All done...
	sceKernelExitThread(SCE_KERNEL_EXIT_SUCCESS);
	return(SCE_KERNEL_EXIT_SUCCESS);
}

/// End of function 'Asura_PSP_SimpleAudio::UpdateThread'. ///



////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SimpleAudio::Initialise
// Author       : Andy Tate
// Date         : 21-Jun-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_SimpleAudio::Initialise (void)
{
	SceUID	iStatus;

	for ( int i = 0; i < ASURA_PSP_MAX_NUM_ATRAC_STREAMS; i++ )
	{
		s_uCurrentBuffer[i] = 0;
	}

	// Reserve memory for simple-audio to use...
	sceAudioOutput2Reserve(ASURA_PSP_SA_NUM_SAMPLES);

	// Create and start the thread...
	iStatus	= sceKernelCreateThread("Simple-audio thread", UpdateThread, iASURA_THREAD_PRIORITY_SIMPLE_AUDIO_UPDATE_THREAD, ASURA_PSP_SIMPLE_AUDIO_THREAD_STACKSIZE, SCE_KERNEL_TH_USE_VFPU, NULL);
	if (iStatus < 0)
	{
		Asura_Assertf(false, ("sceKernelCreateThread(): Failed %08X", iStatus));
	}
	else
	{
		s_iThreadID	= iStatus;
		iStatus		= sceKernelStartThread(s_iThreadID, 0, 0);

		if(iStatus < 0)
		{
			Asura_Assertf(false, ("sceKernelStartThread(): Failed %08X", iStatus));
			sceKernelDeleteThread(s_iThreadID);
			s_iThreadID	= -1;
		}
	}
}

/// End of function 'Asura_PSP_SimpleAudio::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_PSP_SimpleAudio::Shutdown
// Author       : Andy Tate
// Date         : 21-Jun-2006
////////////////////////////////////////////////////////////////////////////////

void Asura_PSP_SimpleAudio::Shutdown (void)
{
	// flag the thread to stop
	Asura_PSP_SAS::BeginUsage();
	g_bSimpleAudioRunning = false;
	Asura_PSP_SAS::EndUsage();
	
	SceUInt uTimeoutusecs = 1000000;
#ifdef ASURA_ASSERT
	int iRet = 
#endif
	sceKernelWaitThreadEnd(s_iThreadID, &uTimeoutusecs);
	Asura_Assertf( iRet == 0, ("sceKernelWaitThreadEnd failed (%x)", iRet ));

	Asura_PSP_SAS::BeginUsage();

	// Destroy the thread...
	sceKernelDeleteThread(s_iThreadID);
	s_iThreadID	= -1;

	// Release memory used by simple audio...
	sceAudioOutput2Release();

	Asura_PSP_SAS::EndUsage();
}

/// End of function 'Asura_PSP_SimpleAudio::Shutdown'. ///


/// End of file. ///
