//
// Asura_Platform_Corona.cpp
//
// Created by KJFL - Tuesday, March 30, 2004
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "SpecialFX/Asura_Corona_System.h"

// Asura headers
#include "Camera/Asura_Camera.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Rendering/Asura_Render.h"

// PSP headers
#include "Debugging/Asura_PSP_Profiling.h"
#include "Rendering/Asura_PSP_Fog.h"
#include "Rendering/Asura_PSP_Graphics.h"
#include "Rendering/Asura_PSP_TextureManagement.h"


#include <kernel.h> // cache writebacks

#ifndef ASURA_PSP_MAX_NUM_CORONA_VERTS
	static const u_int uMAXIMUM_NUMBER_OF_CORONA_VERTS = 256;
#else
	static const u_int uMAXIMUM_NUMBER_OF_CORONA_VERTS = ASURA_PSP_MAX_NUM_CORONA_VERTS;
#endif

struct Asura_Corona_Vertex
{
	u_char			m_ucU;
	u_char			m_ucV;
	u_int			m_uDiffuse;
	Asura_Vector_3	m_afPosition;
};

static Asura_Corona_Vertex g_axCorona [uMAXIMUM_NUMBER_OF_CORONA_VERTS];

static u_int g_uNumberOfCoronasVertsThisFrame;
static u_int g_uStartPointOfCoronaVerts;
static int g_iCurrentTextureID;

void Asura_Corona_System::Platform_Initialise()
{
}

void Asura_Corona_System::Platform_Shutdown()
{
}

void Asura_Corona_System::Platform_RenderStart(void)
{
#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_CORONAS_START ); 
#endif // #ifdef ASURA_PSP_PROFILING

	// KJFL - 28 February 2005 - disable vertex fog if it's on.
	Asura_PSP_Fog::TurnOffFogIfNecessary();

	Asura_PSP_Graphics::SetUseColourDoubling(false);

	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);

	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
	Asura_Render::CurrentStates.RequestAlphaTest(true);
 
	Asura_PSP_Graphics::SetNumberOfLights( 0 );
	Asura_PSP_Graphics::SetUseLighting( false );

	Asura_PSP_Graphics::SetMaterialType( ASURA_PSP_GRAPHICS_MATERIALTYPE_ENV );

	// Centralised state override
//	Asura_PSP_Graphics::DoStateOverride();

	// JG - 17/07/06: Same fix as applied to PFX render, this buffer reset was being called several times per frame. 
	// To fix it properly, we need a pair BeginFrame/EndFrame called only one per frame. Meanwhile a timestamp to
	// avoid resetting more than one will do it
	static unsigned int s_uFrameStamp = 0; 
	if (s_uFrameStamp != Asura_Timers::GetFrameNumber())
	{	
		g_uNumberOfCoronasVertsThisFrame = 0;
		g_uStartPointOfCoronaVerts = 0;
		s_uFrameStamp = Asura_Timers::GetFrameNumber(); 
	}

	g_iCurrentTextureID = ASURA_TEXTURE_NONE;
	
}

static void FlushCoronas( void )
{
	if (!g_uNumberOfCoronasVertsThisFrame) return;
	if (g_uNumberOfCoronasVertsThisFrame - g_uStartPointOfCoronaVerts <= 0) return;

    sceKernelDcacheWritebackRange( &g_axCorona[g_uStartPointOfCoronaVerts], (g_uNumberOfCoronasVertsThisFrame - g_uStartPointOfCoronaVerts) * sizeof(Asura_Corona_Vertex) );

	GUM_DrawArray
	(
		SCEGU_PRIM_RECTANGLES,
		SCEGU_TEXTURE_UBYTE|SCEGU_COLOR_PF8888|SCEGU_VERTEX_FLOAT,
		g_uNumberOfCoronasVertsThisFrame-g_uStartPointOfCoronaVerts,
		0,
		&g_axCorona[g_uStartPointOfCoronaVerts]
	);

	// JG - 17/07/06: This was only set after changing textures middle frame, but 
	// is part of the flushing process. Otherwise the EndRender flush was broken
	// if happened more than once per frame. 
	g_uStartPointOfCoronaVerts = g_uNumberOfCoronasVertsThisFrame; 
}


void Asura_Corona_System::Platform_RenderCorona(const Asura_Corona* pxCorona, u_int uCamVisFlags)
{
	Asura_Assert(g_uNumberOfCoronasVertsThisFrame < uMAXIMUM_NUMBER_OF_CORONA_VERTS, "Too many coronas to render this frame. Reduce the number of active coronas or change the value of ASURA_PSP_MAX_NUM_CORONA_VERTS (default is 256).");
	if (uMAXIMUM_NUMBER_OF_CORONA_VERTS == g_uNumberOfCoronasVertsThisFrame) return;

	//if(!uCamVisFlags) return;

	if (g_iCurrentTextureID != pxCorona->m_iTextureIndex)
	{
		// Uh-oh. Texture changed. Render what we have so far
		FlushCoronas();

		// set up texture in advance, ready for the next Flush call
		Asura_PSP_TextureManagement::RequestTextureIndex( pxCorona->m_iTextureIndex );
		g_iCurrentTextureID = pxCorona->m_iTextureIndex;
	}

	float fSize = pxCorona->m_fRadius;
	

	//for(int i=0; i<Asura_Camera::GetNumberOfCameras(); i++)
	//{
	//  if(!(uCamVisFlags & (1<<i))) continue;
	//			
	//	Asura_Camera::SetCameraIndex(i);
	//	Asura_Render::ApplyCameraSettings();

	Asura_Camera::SetCameraIndex(0);
	Asura_Render::ApplyCameraSettings();


	{
		const Asura_Vector_3*	pxCameraPosition = Asura_Camera::GetPosition();
		const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();

		Asura_Vector_3 xDeltaX = Asura_Vector_3(pxCameraOrientation->e11,pxCameraOrientation->e21,pxCameraOrientation->e31)*fSize;
		Asura_Vector_3 xDeltaY = Asura_Vector_3(pxCameraOrientation->e12,pxCameraOrientation->e22,pxCameraOrientation->e32)*fSize;

		g_axCorona[0+g_uNumberOfCoronasVertsThisFrame].m_ucU = 0;
		g_axCorona[0+g_uNumberOfCoronasVertsThisFrame].m_ucV = 0;
		g_axCorona[1+g_uNumberOfCoronasVertsThisFrame].m_ucU = 128;
		g_axCorona[1+g_uNumberOfCoronasVertsThisFrame].m_ucV = 128;

		g_axCorona[0+g_uNumberOfCoronasVertsThisFrame].m_afPosition = pxCorona->m_xPosition - xDeltaX - xDeltaY;
		g_axCorona[1+g_uNumberOfCoronasVertsThisFrame].m_afPosition = pxCorona->m_xPosition + xDeltaX + xDeltaY;

	
		float fLightScale = 255.0f; // KJL - 21/07/2003 15:03:24 - this is set to default for (m_bUseDirectionEffect==false)

		Asura_Vector_3 xDelta =  *pxCameraPosition - pxCorona->m_xPosition ;
		float fDistance = xDelta.Magnitude();
		
		if(pxCorona->m_bUseDirectionEffect)
		{
			if (fDistance>0.0f)
			{
				fLightScale = (pxCorona->m_xDirection * xDelta) / fDistance;
			
				if (fLightScale<0)
				{
					fLightScale = 64.0f;
				}
				else
				{
					fLightScale = 191.0f*fLightScale+64.0f;
				}
			}
			else
			{
				fLightScale = 0.0f;
			}
		}

		if( !pxCorona->m_bDoNotFadeOutWithDistance )
		{
			float fDistanceScale = fDistance - ASURA_CORONA_FADE_DIST_BEGIN;
			
			if (fDistanceScale > 0.0f)
			{
				fDistanceScale *= ( 1.0f / (ASURA_CORONA_FADE_DIST_END - ASURA_CORONA_FADE_DIST_BEGIN) );

				fLightScale *= (1.0f - fDistanceScale);
			}
		}

		unsigned int uColour = 0xff000000 	
						 + (Asura_Maths::FloatToInt(fLightScale*pxCorona->m_fRed))
						 + (Asura_Maths::FloatToInt(fLightScale*pxCorona->m_fGreen)<<8)
						 + (Asura_Maths::FloatToInt(fLightScale*pxCorona->m_fBlue)<<16);

		
		g_axCorona[0 + g_uNumberOfCoronasVertsThisFrame].m_uDiffuse = uColour;
		g_axCorona[1 + g_uNumberOfCoronasVertsThisFrame].m_uDiffuse = uColour;

		g_uNumberOfCoronasVertsThisFrame += 2;


	
	}

}

void Asura_Corona_System::Platform_RenderEnd(void)
{
	FlushCoronas();

#ifdef ASURA_PSP_PROFILING
	Asura_PSP_Profiling::Signal( Asura_PSP_Profiling::ASURA_PSP_SIGNAL_CORONAS_END ); 
#endif // #ifdef ASURA_PSP_PROFILING
}
