//
// Asura_GX_Shape.cpp
//
// Created by KOR - 10 October 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_GX_Shape.h"

// Asura headers
#include "Camera/Asura_Camera.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Material_System.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Rendering/Asura_Render.h"

// Platform headers
#include "Core/Asura_Wii_MEM2.h"
#include "Rendering/Asura_GX.h"

// Library headers
#include <revolution/gx.h>

// Static members
Asura_TemplateHashTree<Asura_GX_Shape> Asura_GX_ShapeRenderer::s_xHashTree;

void Asura_GX_ShapeRenderer::Initialise()
{
}

void Asura_GX_ShapeRenderer::Update()
{
	// Nothing to update, I think
}

void Asura_GX_ShapeRenderer::Reset()
{
	s_xHashTree.Reset();
}

void Asura_GX_ShapeRenderer::Shutdown()
{
	s_xHashTree.Reset();
}

bool Asura_GX_ShapeRenderer::AddShape(Asura_Handle iFileHandle)
{
	Asura_Hash_ID uHashID;
	Asura_FileAccess::ReadData(iFileHandle, &uHashID, sizeof(Asura_Hash_ID));
	// Add to hash tree
	Asura_GX_Shape* pxShape = s_xHashTree.AddNode(uHashID);
	Asura_Assert(pxShape, "Failed to allocated shape");
	if(pxShape)
	{
		// Add to storage
		pxShape->LoadFromChunk(iFileHandle);
	}

	return (pxShape != NULL);
}

void Asura_GX_ShapeRenderer::RenderShape(Asura_Hash_ID uHashID, u_int uCamVisFlags, const Asura_Vector_3& xPos, const Asura_Matrix_3x3& xOri, float fAlpha)
{
	const Asura_GX_Shape* pxShape = GetShape(uHashID);
	Asura_Assertf(pxShape, ("Could not find shape with hash %i.", uHashID));
	if(!pxShape)
	{
		return;
	}
	pxShape->RenderSetup(fAlpha);

	// Set up position and orientation
	Mtx xPosTransform;
	Asura_GX::OriPosToMTX(xOri, xPos, xPosTransform);

	for( int iCam = Asura_Camera::GetNumberOfCameras(); iCam--; )
	{
		if (Asura_Camera::IsCameraActive(iCam) && ((1 << iCam) & uCamVisFlags) )
		{
			Asura_Camera::SetCameraIndex( iCam );
			Asura_Render::ApplyCameraSettings();
			Asura_GX::SetUpEntityLights(xPos);

			Mtx mv, mvi;
			const Mtx& xViewMatrix = Asura_GX::GetViewMatrix();
			MTXConcat(xViewMatrix, xPosTransform, mv);
			GXLoadPosMtxImm(mv, GX_PNMTX0);
			MTXInverse(mv, mvi); 
			MTXTranspose(mvi, mv); 
			GXLoadNrmMtxImm(mv, GX_PNMTX0);

			Asura_GX::SetUpEntitySpecular(xPos, pxShape->GetMaterialID());

			pxShape->Render();
		}
	}
}

const Asura_GX_Shape* Asura_GX_ShapeRenderer::GetShape(Asura_Hash_ID uHashID)
{
	return s_xHashTree.GetDataByHash(uHashID);
}

int Asura_GX_ShapeRenderer::GetShapeMaterial(Asura_Hash_ID uHashID)
{
	const Asura_GX_Shape* pxShape = GetShape(uHashID);
	Asura_Assertf(pxShape, ("Could not find shape with hash %i.", uHashID));
	if(pxShape)
	{
		return pxShape->GetMaterialID();
	}
	return ASURA_MATERIAL_NONE;
}

Asura_GX_Shape::Asura_GX_Shape()
: m_iMaterialID(ASURA_MATERIAL_NONE)
, m_uNumberOfIndices(0)
, m_uNumberOfVertices(0)
, m_pxVertices(NULL)
, m_pusIndices(NULL)
{
}


void Asura_GX_Shape::LoadFromChunk(Asura_Handle iFileHandle)
{
	Asura_FileAccess::ReadData(iFileHandle, &m_iMaterialID, sizeof(unsigned int));
	m_iMaterialID = Asura_Material_System::ConvertIndex(m_iMaterialID);
	Asura_FileAccess::ReadData(iFileHandle, &m_uNumberOfIndices, sizeof(unsigned int));
	Asura_FileAccess::ReadData(iFileHandle, &m_uNumberOfVertices, sizeof(unsigned int));
	Asura_Bounding_Box xBB;
	Asura_FileAccess::ReadData(iFileHandle, &xBB, sizeof(Asura_Bounding_Box));
	if(m_uNumberOfVertices)
	{
		m_pxVertices = static_cast<Asura_GX_Shape_Vertex*>(ASURA_WII_MEM2_ALLOC(sizeof(Asura_GX_Shape_Vertex) * m_uNumberOfVertices));
		Asura_FileAccess::ReadData(iFileHandle, m_pxVertices, sizeof(Asura_GX_Shape_Vertex) * m_uNumberOfVertices);
	}
	if(m_uNumberOfIndices)
	{
		m_pusIndices = new unsigned short[m_uNumberOfIndices];
		Asura_FileAccess::ReadData(iFileHandle, m_pusIndices, sizeof(unsigned short) * m_uNumberOfIndices);	
	}
}

Asura_GX_Shape::~Asura_GX_Shape()
{
	delete [] m_pusIndices;
	ASURA_WII_MEM2_FREE(m_pxVertices);
}

void Asura_GX_Shape::RenderSetup(float fAlpha) const
{
	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS, GX_INDEX16);
	GXSetVtxDesc(GX_VA_NRM, GX_INDEX16);
	GXSetVtxDesc(GX_VA_TEX0, GX_INDEX16);

	GXSetArray(GX_VA_POS, m_pxVertices, sizeof(Asura_GX_Shape_Vertex));
	GXSetArray(GX_VA_NRM, (char*)m_pxVertices + 12, sizeof(Asura_GX_Shape_Vertex));
	// Offset to texture co-ords in Asura_GX_Shape_Vertex is 6 * 4 bytes
	GXSetArray(GX_VA_TEX0, (char*)m_pxVertices + 24, sizeof(Asura_GX_Shape_Vertex));

	GXInvalidateVtxCache();

	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CCW);
	Asura_GX::SetupMaterial(m_iMaterialID, 0, 0, ASURA_TEXTURE_NONE, fAlpha);
	Asura_GX::Set3DProjection(true);
}

void Asura_GX_Shape::Render() const
{	
	GXBegin(GX_TRIANGLESTRIP, GX_VTXFMT3, m_uNumberOfIndices);
	for(unsigned int uIndex = 0; uIndex < m_uNumberOfIndices; uIndex++ )
	{
		const u_short usIndex = m_pusIndices[uIndex];
		Asura_Assert(usIndex < m_uNumberOfVertices, "Vertex index out of range.");
		GXPosition1x16(usIndex);
		GXNormal1x16(usIndex);
		GXTexCoord1x16(usIndex);
	}
	GXEnd();
	Asura_Render::AddToTrianglesRendered(m_uNumberOfIndices - 2, 1, ASURA_RENDER_TTYPE_SHA);
}

