//
// Asura_Platform_Render.cpp
//
// Created by KOR - 10 October 2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Rendering/Asura_Render.h"

// Asura headers
#include "Core/Asura_TextureManagement.h"
#include "Entity/Glass/Asura_WorldSpace_Rectangle.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_DisplayLimits.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Viewport.h"

// Platform headers
#include "Rendering/Asura_GX.h"
#include "Rendering/Asura_GX_Hierarchy.h"
#include "Rendering/Asura_GX_Shape.h"
#include "Rendering/Asura_GX_SmoothSkin.h"
#include "Rendering/Asura_GX_TextureManagement.h"

// For screen profiling output
#ifndef ASURA_SUBMISSION_BUILD
// Asura
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_Entity_System.h"
#include "Sound/Asura_Sound.h"

// Wii
#include "Core/Asura_Wii_MEM2.h"
#include "Rendering/Asura_Wii_ROMFont.h"
#include "Sound/Asura_Wii_StreamSound.h"

#endif

// Library headers
#include <revolution/gx.h>

DVSVAR bool dvs_bShowFPS = false;
DVSVAR bool dvs_bShowDebugInfo = false;
DVSVAR bool dvs_bShowQAInfo = false;
DVSVAR bool dvs_bShowStreamSoundInfo = false;

DEBUGVAR bool dbg_bScaleXForWidescreen = true;

const float fSCREENWIDTH = 640.f;

static inline float ScaleXForWidescreen(float fX)
{
	if (dbg_bScaleXForWidescreen && Asura_Render::ShouldRenderForWidescreen())
	{
		fX -= fSCREENWIDTH * .5f;	// in range -320,320
		fX *= 0.75f;	// up to edge of of 4:3 part of 16:9 widescreen
		fX += fSCREENWIDTH * .5f;	// in range 0,640
	}
	return fX;
}

void Asura_Render::Platform_Initialise(void)
{
	Asura_Render::SetCutsceneHasCameraControl(false);

	Asura_GX::Initialise();
	Asura_GX_SmoothSkinRenderer::Initialise();
	Asura_GX_HierarchyRenderer::Initialise();
	Asura_GX_ShapeRenderer::Initialise();

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	Asura_CommandConsole::AddVar(L"Wii.ShowFPS", &dvs_bShowFPS, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"Wii.ShowDebugInfo", &dvs_bShowDebugInfo, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"Wii.ShowQAInfo", &dvs_bShowQAInfo, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	Asura_CommandConsole::AddVar(L"Wii.ShowStreamingSoundInfo", &dvs_bShowStreamSoundInfo, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif
	const float fHorizontalSafeBorder = 0.07f;
	const float fVerticalSafeBorder = 0.1f;
	s_xSafeArea.Set(fHorizontalSafeBorder, fVerticalSafeBorder, 1.f - fHorizontalSafeBorder, 1.f - fVerticalSafeBorder);
}

void Asura_Render::Platform_EarlyShutdown(void)
{
}

void Asura_Render::Platform_LateShutdown(void)
{
	Asura_GX_SmoothSkinRenderer::Shutdown();
	Asura_GX_HierarchyRenderer::Shutdown();
	Asura_GX_ShapeRenderer::Shutdown();
	Asura_GX::Shutdown();
}

void Asura_Render::Platform_Reset(void)
{
	OSReport("Calling Asura_Render::Reset\n");
	Asura_GX_ShapeRenderer::Reset();
	Asura_GX_HierarchyRenderer::Reset();
	Asura_GX_SmoothSkinRenderer::Reset();
}

void Asura_Render::Platform_ApplyCameraSettings(void)
{
	Asura_GX::SetViewingArea();
}

void Asura_Render::Platform_AllowFullScreenRendering(void)
{
	if (!Asura_GX::ReadyToDraw())
		return;
	GXSetScissor(0, 0, (unsigned int)Asura_Viewport::GetWidth(), (unsigned int)Asura_Viewport::GetHeight());
}


void Asura_Render::Platform_ClearZBuffer(void)
{
	Asura_GX::ClearZBuffer();
}


void Asura_Render::Platform_BeginFrame(void)
{
	s_bCurrentlyRendering = true;
	Asura_GX::BeginFrame();

	if (!Asura_GX::ReadyToDraw())
		return;

	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
	GXSetCullMode(GX_CULL_NONE);
}

DEBUGVAR short dbg_sFontSize = 16;
DEBUGVAR short dbg_sDebugInfoX = 32;
DEBUGVAR short dbg_sDebugInfoY = 6;

// KOR - yuck - this should be in a header somewhere.
extern "C"
{

struct mallinfo {
	int arena;    /* non-mmapped space allocated from system */
	int ordblks;  /* number of free chunks */
	int smblks;   /* number of fastbin blocks */
	int hblks;    /* number of mmapped regions */
	int hblkhd;   /* space in mmapped regions */
	int usmblks;  /* maximum total allocated space */
	int fsmblks;  /* space available in freed fastbin blocks */
	int uordblks; /* total allocated space */
	int fordblks; /* total free space */
	int keepcost; /* top-most, releasable (via malloc_trim) space */
};

struct mallinfo mallinfo();

};


void Asura_Render::Platform_EndFrame(void)
{
	if (!Asura_GX::ReadyToDraw())
		return;

#ifndef ASURA_SUBMISSION_BUILD

	if(dvs_bShowStreamSoundInfo)
	{
		Asura_Wii_Streaming_Sound_System::DebugRender();
	}

	Asura_Wii_ROMFont::SetSize(dbg_sFontSize, 0);
	const short sColumn = dbg_sDebugInfoX;
	short sRow = dbg_sDebugInfoY;
	// Display FPS
	const u_int uBUFFER_LENGTH = 200;
	char szBuffer[uBUFFER_LENGTH];
	if (dvs_bShowFPS)
	{
		const float fFPS = Asura_Timers::GetSmoothedFrameRate() + 0.05f;
		Asura_PrintfToBfr(szBuffer, "FPS: %3.1f", uBUFFER_LENGTH, true, fFPS);
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);
	}

	if (dvs_bShowDebugInfo)
	{
#if defined ASURA_RENDER_TTYPES
		Asura_PrintfToBfr(szBuffer, "Tris: %5u Env %5u Hier %4u Skin %4u", uBUFFER_LENGTH, true, 
			Asura_Render::GetNumberOfTrianglesRendered(),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_ENV),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_HIER),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_SKIN)
			);
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);
		Asura_PrintfToBfr(szBuffer, "          Objects %2u Glass %2u Water %2u Shapes %2u PFX %2u Other %3u", uBUFFER_LENGTH, true, 
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_OBJ),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_GLS),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_WAT),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_SHA),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_PFX),
			Asura_Render::GetNumberOfTrianglesRendered(ASURA_RENDER_TTYPE_OTHER)
			);
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);

		Asura_PrintfToBfr(szBuffer, "Hiers %2u Skins %2u Shapes %2u PFX %3u", uBUFFER_LENGTH, true, 
			Asura_Render::GetHierarchiesRenderedThisFrame(),
			Asura_Render::GetSmoothSkinsRenderedThisFrame(),
			Asura_Render::GetShapesRenderedThisFrame(),
			Asura_Render::GetParticlesRenderedThisFrame()
			);
#else
		Asura_PrintfToBfr(szBuffer, "Tris: %5u; Hiers %2u Skins %2u Shapes %2u PFX %3u", uBUFFER_LENGTH, true, 
			Asura_Render::GetNumberOfTrianglesRendered(),
			Asura_Render::GetHierarchiesRenderedThisFrame(),
			Asura_Render::GetSmoothSkinsRenderedThisFrame(),
			Asura_Render::GetShapesRenderedThisFrame(),
			Asura_Render::GetParticlesRenderedThisFrame()
			);
#endif
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);

		const struct mallinfo xMallInfo = mallinfo();
		Asura_PrintfToBfr(szBuffer, "MEM1: %iKB FREE, %i blks. MEM2: %iKB FREE, %i allocs\n", uBUFFER_LENGTH, true,
			xMallInfo.fordblks / 1024,
			xMallInfo.ordblks,
			Asura_Wii_MEM2::GetFreeBytes() / 1024,
			Asura_Wii_MEM2::GetNumAllocations());
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);
		//RenderMallInfo(sColumn, sRow++);

		const u_int uPlayingSounds = Asura_Sound_System::GetNumberOfPlayingSounds();
		Asura_PrintfToBfr(szBuffer, "Playing sounds: %2i", uBUFFER_LENGTH, true, uPlayingSounds);
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);
	}

	if (dvs_bShowQAInfo)
	{
		const Asura_Vector_3*	pxCamPos	= Asura_Camera::GetPosition();

		Asura_PrintfToBfr(szBuffer, "Cam Pos: (%3.1f, %3.1f, %3.1f)", uBUFFER_LENGTH, true, pxCamPos->GetX(), pxCamPos->GetY(), pxCamPos->GetZ()); 
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);

		// ARC - 04 January 2007 17:02:22 - Can only do this if ASURA_PSP_PROFILING is defined (Simpsons doesn't have it defined)
#if defined(ASURA_USE_PROFILING_SYSTEM)
		Asura_PrintfToBfr(szBuffer, "Updating Entities: %d", uBUFFER_LENGTH, true, Asura_Entity_System::GetNumServerEntitiesUpdating()); 
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);

		Asura_PrintfToBfr(szBuffer, "Renderable Entities: %d", uBUFFER_LENGTH, true, Asura_Entity_System::GetNumRenderableEntities()); 
		Asura_Wii_ROMFont::RenderString(sColumn, (dbg_sFontSize * sRow++), 0, szBuffer);
#endif // ASURA_PSP_PROFILING
	}
#endif//ndef ASURA_SUBMISSION_BUILD

	FlushDisplay();
	Asura_GX::EndFrame();
	s_bCurrentlyRendering = false;
}

void Asura_Render::Platform_FlushDisplay(void)
{
	Asura_GX::FlushFrame();
}

bool Asura_Render::Platform_SetVideoMode(int width, int height, bool fsaa)
{
	return false;
}

void Asura_Overlay_Quad::Platform_Render()
{	
	if (!Asura_GX::ReadyToDraw())
		return;

	/// Kim Randell - 03/02/2003 16:29:07 - TODO: find out if quad batching will be efficient on the Gamecube
	/// For now, just render it straight out.
	Asura_GX::SetZBufferMode(false, ASURA_ZBUFFERCMP_ALWAYS, false);

	Asura_GX::Set3DProjection(false);

	// No lighting
	GXSetChanCtrl(GX_COLOR0A0, GX_DISABLE, GX_SRC_REG, GX_SRC_VTX, 0, GX_DF_NONE, GX_AF_NONE);

	// Pull desired texture out of texture management
	Asura_Render::CurrentStates.RequestTextureIndex(0, TextureIndex);
	if(!Asura_GX_Texture_Management::IsColourTextureLoaded())
	{
		Asura_Render::CurrentStates.RequestTextureIndex(0, Asura_GX::GetWhiteTextureIndex());
	}

	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS, GX_DIRECT);
	GXSetVtxDesc(GX_VA_CLR0, GX_DIRECT);
	GXSetVtxDesc(GX_VA_TEX0, GX_DIRECT);

	// Channel Color Used
	GXSetNumChans(1);
	GXSetTevSwapModeTable(GX_TEV_SWAP0, GX_CH_RED, GX_CH_GREEN, GX_CH_BLUE, GX_CH_ALPHA);
	GXSetTevSwapMode(GX_TEVSTAGE0, GX_TEV_SWAP0, GX_TEV_SWAP0);

	GXTexCoordID eAlphaTexCoord = GX_TEXCOORD0;
	if(Asura_GX_Texture_Management::IsAlphaTextureLoaded() && Asura_GX_Texture_Management::IsAlphaTextureDifferentRes())
	{
		GXSetNumTexGens(2);
		GXSetTexCoordGen(GX_TEXCOORD1, GX_TG_MTX2x4, GX_TG_TEX0, GX_IDENTITY);
		eAlphaTexCoord = GX_TEXCOORD1;
	}
	else
	{
		GXSetNumTexGens(1);
	}
	GXSetTexCoordGen(GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_IDENTITY);

	GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
	if(Asura_GX_Texture_Management::IsColourTextureLoaded())
	{
		if(Asura_GX_Texture_Management::IsAlphaTextureLoaded())
		{
			GXSetNumTevStages(2);

			GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
			GXSetTevOrder(GX_TEVSTAGE1, eAlphaTexCoord, Asura_GX_Texture_Management::GetAlphaTexMapID(), GX_COLOR0A0);

			GXSetTevColorIn(GX_TEVSTAGE0, GX_CC_ZERO, GX_CC_TEXC, GX_CC_RASC, GX_CC_ZERO);
			GXSetTevColorOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
			GXSetTevAlphaIn(GX_TEVSTAGE0, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO, GX_CA_ZERO);
			GXSetTevAlphaOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);

			GXSetTevSwapMode(GX_TEVSTAGE1, GX_TEV_SWAP0, GX_TEV_SWAP1);
			GXSetTevColorIn(GX_TEVSTAGE1, GX_CC_ZERO, GX_CC_ZERO, GX_CC_ZERO, GX_CC_CPREV);
			GXSetTevColorOp(GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
			GXSetTevAlphaIn(GX_TEVSTAGE1, GX_CA_ZERO, GX_CA_RASA, GX_CA_TEXA, GX_CA_ZERO);
			GXSetTevAlphaOp(GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
		}
		else
		{
			GXSetNumTevStages(1);

			GXSetTevColorIn(GX_TEVSTAGE0, GX_CC_ZERO, GX_CC_TEXC, GX_CC_RASC, GX_CC_ZERO);
			GXSetTevColorOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
			GXSetTevAlphaIn(GX_TEVSTAGE0, GX_CA_ZERO, GX_CA_TEXA, GX_CA_RASA, GX_CA_ZERO);
			GXSetTevAlphaOp(GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO, GX_CS_SCALE_1, GX_ENABLE, GX_TEVPREV);
			GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
		}
	}
	else
	{
		return;
	}


	Asura_Render::CurrentStates.RequestTranslucencyMode(TranslucencyMode);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
	Asura_Render::CurrentStates.RequestAlphaTest(TranslucencyMode == ASURA_TRANSLUCENCY_NORMAL);

	const u_int uColour = Asura_GX::ARGBToRGBA(Colour);
	GXBegin(GX_QUADS, GX_VTXFMT2, 4);
	for(int i = 0; i < 4; ++i)
	{
		GXPosition2f32(ScaleXForWidescreen(X[i]), Y[i]);
		GXColor1u32(uColour);
		GXTexCoord2f32(U[i], V[i]);
	}
	GXEnd();

//	GXFlush();
//	Asura_GX::Set3DProjection(true);
	Asura_GX::SetZBufferMode(true, ASURA_ZBUFFERCMP_LESSEQUAL, true);
}

void Asura_Render::Platform_InitialiseSplashScreen(const char* szName)
{
}

void Asura_Render::Platform_ShutdownSplashScreen(void)
{
}

void Asura_Render::Platform_SplashScreenRender(void)
{
}

void Asura_Render::Platform_RenderDynamicShadowsOntoEnvironment()
{
}

// Primitive rendering
void Render_Rectangle( const Asura_WorldSpace_Rectangle& xRect, u_int uColour, int iTexIdx, const float* pfUs, const float* pfVs, bool bWireframe )
{
	GXSetTevSwapMode(GX_TEVSTAGE0, GX_TEV_SWAP0, GX_TEV_SWAP0);
	Asura_GX::Set3DProjection(true);
	Asura_Vector_3 axCorners[4];
	xRect.CalculateCorners(axCorners);

	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);
	Asura_Render::CurrentStates.RequestTextureIndex(0,iTexIdx);

	// No lighting
	GXSetChanCtrl(GX_COLOR0A0, GX_DISABLE, GX_SRC_REG, GX_SRC_VTX, 0, GX_DF_NONE, GX_AF_NONE);

	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS, GX_DIRECT);
	GXSetVtxDesc(GX_VA_CLR0, GX_DIRECT);
	GXSetVtxDesc(GX_VA_TEX0, GX_DIRECT);
	// Initialize lighting, texgen, and tev parameters
	GXSetNumChans(1);

	GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);
	GXSetTexCoordGen(GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_IDENTITY);
	GXSetNumTexGens(1);
	GXSetNumChans(1);
	GXSetNumTevStages(1);
	if(Asura_GX_Texture_Management::IsColourTextureLoaded())
	{
		GXSetTevOp(GX_TEVSTAGE0, GX_MODULATE);
	}
	else
	{
		GXSetTevOp(GX_TEVSTAGE0, GX_PASSCLR);
	}

	uColour = Asura_GX::ARGBToRGBA(uColour);
	GXBegin(GX_TRIANGLEFAN, GX_VTXFMT3, 4);
	GXPosition3f32(axCorners[0].x, axCorners[0].y, axCorners[0].z);
	GXColor1u32(uColour);
	GXTexCoord2f32(pfUs[0], pfVs[0]);

	GXPosition3f32(axCorners[1].x, axCorners[1].y, axCorners[1].z);
	GXColor1u32(uColour);
	GXTexCoord2f32(pfUs[1], pfVs[1]);

	GXPosition3f32(axCorners[3].x, axCorners[3].y, axCorners[3].z);
	GXColor1u32(uColour);
	GXTexCoord2f32(pfUs[3], pfVs[3]);

	GXPosition3f32(axCorners[2].x, axCorners[2].y, axCorners[2].z);
	GXColor1u32(uColour);
	GXTexCoord2f32(pfUs[2], pfVs[2]);

	GXEnd();
}

void Asura_Render::Platform_Render2DLine(const Asura_Vector_2& xV1, const Asura_Vector_2& xV2, u_int uColour)
{
	GXSetNumChans(1);
	GXSetNumTexGens(0);
	GXSetNumTevStages(1);

	GXSetTevSwapMode(GX_TEVSTAGE0, GX_TEV_SWAP0, GX_TEV_SWAP0);

	GXSetChanCtrl(GX_COLOR0A0, GX_FALSE, GX_SRC_REG, GX_SRC_VTX, 0, GX_DF_NONE, GX_AF_NONE );

	Asura_GX::Set3DProjection(false);
	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS, GX_DIRECT);
	GXSetVtxDesc(GX_VA_CLR0, GX_DIRECT);
	GXSetTevOp(GX_TEVSTAGE0, GX_PASSCLR);

	uColour = Asura_GX::ARGBToRGBA(uColour);
	GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP0, GX_COLOR0A0);

	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_ALWAYS);
	Asura_Render::CurrentStates.RequestZBufferEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestTextureIndex(0, ASURA_TEXTURE_NONE);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);

	GXBegin(GX_LINES, GX_VTXFMT2, 2);
	GXPosition2f32(ScaleXForWidescreen(xV1.x), xV1.y);
	GXColor1u32(uColour);
	GXPosition2f32(ScaleXForWidescreen(xV2.x), xV2.y);
	GXColor1u32(uColour);
	GXEnd();
}

void Asura_Render::Platform_RenderLine(const Asura_Vector_3& p1, const Asura_Vector_3& p2, u_int uColour)
{
	GXSetNumChans(1);
	GXSetNumTexGens(0);
	GXSetNumTevStages(1);

	GXSetTevSwapMode(GX_TEVSTAGE0, GX_TEV_SWAP0, GX_TEV_SWAP0);

	GXSetChanCtrl(GX_COLOR0, GX_FALSE, GX_SRC_REG, GX_SRC_VTX, 0, GX_DF_NONE, GX_AF_NONE );

	GXSetTevOrder(GX_TEVSTAGE0, GX_TEXCOORD_NULL, GX_TEXMAP_NULL, GX_COLOR0A0);

	Asura_GX::Set3DProjection(true);

	GXLoadPosMtxImm(Asura_GX::GetViewMatrix(), GX_PNMTX0);

	GXClearVtxDesc();
	GXSetVtxDesc(GX_VA_POS, GX_DIRECT);
	GXSetVtxDesc(GX_VA_CLR0, GX_DIRECT);
	GXSetTevOp(GX_TEVSTAGE0, GX_PASSCLR);

	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestTextureIndex(0, ASURA_TEXTURE_NONE);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL);
	Asura_Render::CurrentStates.RequestAlphaTest(false);
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);

	uColour = Asura_GX::ARGBToRGBA(uColour);

	GXBegin(GX_LINES, GX_VTXFMT3, 2);
	GXPosition3f32(p1.x, p1.y, p1.z);
	GXColor1u32(uColour);
	GXPosition3f32(p2.x, p2.y, p2.z);
	GXColor1u32(uColour);
	GXEnd();
}

// Miscellaneous bits that don't really deserve their own file right now.

Asura_Bounding_Rectangle Asura_DisplayLimits::Platform_GetPrimaryDisplayBounds()
{
	return Asura_Bounding_Rectangle(0, 0, static_cast<float>(Asura_Viewport::GetWidth()), static_cast<float>(Asura_Viewport::GetHeight()));
}

Asura_Bounding_Rectangle Asura_DisplayLimits::Platform_GetBoundingRectangle()
{
		return Platform_GetPrimaryDisplayBounds();
}

float Asura_Font_System::GetPlatformYScaleFactor(void)
{
	return 1.0f;
}

void Asura_Render::Platform_PrepareToRenderFirstPersonMesh ()
{
	Asura_Assert(false,"Special rendering settings for the first person mesh are not supported on this platform");
}

void Asura_Render::Platform_FinishRenderFirstPersonMesh ()
{
	//Do nothing
}

//eof
