///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_PlatformHierarchy.cpp
// Author : Kevin Floyer-Lea
// Date   : 28-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Animation/Asura_Hierarchy.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_TextureFlags.h"
#include "Rendering/Asura_Render.h"
#include "Shadows/Asura_Shadow_System.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_CharacterRenderer.h"
#include "Rendering/Asura_Xbox360_ObjectHierarchyRenderer.h"
#include "Rendering/Asura_Xbox360_SkinRenderer.h"

/******************************************************************************/
/* Asura_Hierarchy_Skin Class Functions                                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Hierarchy_Skin::Platform_GetTextureIndex()
//
// Returns the texture index associated with the hierarchy skin. Now that we
// allow multiple textures, this doesn't quite make sense, but we return the
// index of the first texture used
//
// Parameters:		None
// Return value:	Texture index
//
////////////////////////////////////////////////////////////////////////////////

int Asura_Hierarchy_Skin::Platform_GetTextureIndex() const
{
	Asura_Hash_ID uMaterialHash	= ASURA_HASH_ID_UNSET;

	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		const Asura_Xbox360_Skin* const pxSkin = Asura_Xbox360_SkinRenderer::GetSkin( m_uNameHashID );

		if( pxSkin && pxSkin->GetNumStrips() > 0 )
		{
			uMaterialHash = pxSkin->GetStrip(0).m_auMaterials[0];
		}
	}
	else if( IsSkinnedModel() )
	{
		const Asura_Xbox360_CharacterRenderer_Skin* const pxSkin = Asura_Xbox360_CharacterRenderer::GetCharacter( m_uNameHashID );

		if((pxSkin != NULL) && (pxSkin->GetNumStrips() > 0))
		{
			uMaterialHash = pxSkin->GetStrip(0).m_auMaterialResponseHashIDs[0];
		}
	}
	else
	{
		const Asura_Xbox360_ObjectHierarchyRenderer::Hierarchy* const pxHierarchy = Asura_Xbox360_ObjectHierarchyRenderer::GetHierarchyTree().GetDataByHash( m_uNameHashID );

		if((pxHierarchy != NULL) && (pxHierarchy->GetNumberOfStrips() > 0))
		{
			uMaterialHash = pxHierarchy->GetStrip(0).m_auMaterialResponseHashIDs[0];
		}
	}

	const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash );
	return pxMaterial ? pxMaterial->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) : ASURA_TEXTURE_NONE;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Hierarchy_Skin::Platform_IsOpaque()
//
// Queries whether a hierarchy skin is opaque or not. Used by the sorted render
// etc.
//
// Parameters:		None
// Return value:	true iff the entire model is opaque
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Hierarchy_Skin::Platform_IsOpaque() const
{
	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		const Asura_Hierarchy_Skin* pxHierarchy = this;

		while( pxHierarchy )
		{
			const Asura_Xbox360_Skin* const pxSkin =
				Asura_Xbox360_SkinRenderer::GetSkin( pxHierarchy->GetHashID() );

			if( pxSkin && pxSkin->IsAnyMaterialTransparent() ) return false;

			pxHierarchy = pxHierarchy->m_pxNextLowerSkin;
		}

		return true;

	}
	else
	{
		if( IsSkinnedModel() )
		{
			const Asura_Hierarchy_Skin*	pxHierarchySkin	= this;

			while(pxHierarchySkin != NULL)
			{
				if(!Asura_Xbox360_CharacterRenderer::IsCharacterOpaque( m_uNameHashID ))
				{
					// This LOD is transparent - assume that the skin is too...
					return(false);
				}
				pxHierarchySkin	= pxHierarchySkin->m_pxNextLowerSkin;
			}

			// No transparent LODs means the skin is opaque...
			return(true);
		}
		else
		{
			const Asura_Hierarchy_Skin*	pxHierarchySkin	= this;

			while(pxHierarchySkin != NULL)
			{
				if(!Asura_Xbox360_ObjectHierarchyRenderer::IsHierarchyOpaque( m_uNameHashID ))
				{
					// This LOD is transparent - assume that the skin is too...
					return(false);
				}
				pxHierarchySkin	= pxHierarchySkin->m_pxNextLowerSkin;
			}

			// No transparent LODs means the skin is opaque...
			return(true);
		}
	}
}

/******************************************************************************/
/* Asura_Hierarchy_System Static Class Functions                              */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Hierarchy_System::Platform_Initialise()
//
// Initialises the hierarchy system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Hierarchy_System::Platform_Initialise()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Hierarchy_System::Platform_Reset()
//
// Resets the hierarchy system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Hierarchy_System::Platform_Reset()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Hierarchy_System::Platform_Shutdown()
//
// Shuts down the hierarchy system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Hierarchy_System::Platform_Shutdown()
{
}

static void MoveBonesToModelSpace(const Asura_Hierarchy_Skin* pxSkin)
{
	// If the bones are in world space, have to switch them back to model space to avoid skinning inaccuracies
#if !defined(ASURA_HIERARCHY_QUAD)
	Asura_Matrix_3x3 xOriTranspose(*Asura_Hierarchy_System::s_pxOri);
	xOriTranspose.Transpose();
#endif

	for(int i = 0; i < pxSkin->NumberOfBones; ++i)
	{
#if defined(ASURA_HIERARCHY_QUAD)
		Asura_Hierarchy_System::s_xInverseBaseTransform.Apply(Asura_Hierarchy_System::s_axBoneTransforms[i], Asura_Hierarchy_System::s_axBoneTransforms[i]);
#else
		Asura_Hierarchy_System::s_axBoneOri[i] = xOriTranspose * Asura_Hierarchy_System::s_axBoneOri[i];
		Asura_Hierarchy_System::s_axBonePos[i] -= *Asura_Hierarchy_System::s_pxPos;
		Asura_Hierarchy_System::s_axBonePos[i] = Asura_Hierarchy_System::s_axBonePos[i] * xOriTranspose;
#endif
	}
	Asura_Hierarchy_System::s_bBonesInWorldspace = false;

}

static void RemoveBindPose(const Asura_Hierarchy_Skin* pxSkin)
{
	// Remove the bind-pose from the bone-data...
	for( int i = 0; i < pxSkin->NumberOfBones; ++i )
	{
#ifdef ASURA_HIERARCHY_QUAD
		Asura_Hierarchy_System::s_axBoneTransforms[i].Apply(pxSkin->m_pxBoneRestPoseInverseTransforms[i], Asura_Hierarchy_System::s_axBoneTransforms[i]);
#else
		Asura_Hierarchy_System::s_axBoneOri[i] = Asura_Hierarchy_System::s_axBoneOri[i] * pxSkin->BoneRestOrientation[i];
		Asura_Hierarchy_System::s_axBonePos[i] -= pxSkin->BoneRestPosition[i] * Asura_Hierarchy_System::s_axBoneOri[i];
#endif
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Hierarchy_System::Platform_RenderSkin
// Author       : KJFL
// Date         : ??-Jul-2005
//
// Description  : Handles the platform-specified rendering of a skinned-object.
//
// Parameters   : 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Hierarchy_System::Platform_RenderSkin( const Asura_Animation* pxAnim, float fAlpha,
	int iTexIDOverride, u_int uCamVisFlags, u_int uRenderFlags, const Asura_Vector_3& xBSCentre,
	float fBSRadius)
{
	// Select LOD mesh based on distance to first bone, since object may be animating far away from its location


	if( s_bBonesInWorldspace )
	{
		MoveBonesToModelSpace(s_pxSkin);
	}

	RemoveBindPose(s_pxSkin);

	if (s_pxSkin != pxAnim->GetSkin())
	{
		const float fDistance = (GetWorldspaceBonePos(0) - *Asura_Camera::GetPosition()).Magnitude();
		const Asura_Hierarchy_Skin* const pxSkinLOD = s_pxSkin->GetDetailSkinFromDistance( fDistance );

		// Time to render the character...
		if( ShouldJustRenderForCurCam() )
		{
#ifdef ASURA_HIERARCHY_QUAD
			Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, s_xBaseTransform, s_xInverseBaseTransform, s_axBoneTransforms );
#else
			Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, *s_pxPos, *s_pxOri, s_axBonePos, s_axBoneOri );
#endif
		}
		else
		{
			for( int u = 0; u < Asura_Camera::GetNumberOfCameras(); ++u )
			{
				if((uCamVisFlags & (1 << u)) != 0)
				{
					Asura_Camera::SetCameraIndex(u);
					Asura_Render::ApplyCameraSettings();

#ifdef ASURA_HIERARCHY_QUAD
					Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, s_xBaseTransform, s_xInverseBaseTransform, s_axBoneTransforms );
#else
					Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, *s_pxPos, *s_pxOri, s_axBonePos, s_axBoneOri );
#endif
				}
			}
		}
	} 
	else 
	{
		for( int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex )
		{
			const Asura_Hierarchy_Skin *pxSkin = pxAnim->GetSkin(iSkinIndex);

			if (pxSkin)
			{
				const float fDistance = (GetWorldspaceBonePos(0) - *Asura_Camera::GetPosition()).Magnitude();
				const Asura_Hierarchy_Skin* const pxSkinLOD = pxSkin->GetDetailSkinFromDistance( fDistance );

				// Time to render the character...
				if( ShouldJustRenderForCurCam() )
				{
#ifdef ASURA_HIERARCHY_QUAD
					Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, s_xBaseTransform, s_xInverseBaseTransform, s_axBoneTransforms );
#else
					Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, *s_pxPos, *s_pxOri, s_axBonePos, s_axBoneOri );
#endif
				}
				else
				{
					for( int u = 0; u < Asura_Camera::GetNumberOfCameras(); ++u )
					{
						if((uCamVisFlags & (1 << u)) != 0)
						{
							Asura_Camera::SetCameraIndex(u);
							Asura_Render::ApplyCameraSettings();

#ifdef ASURA_HIERARCHY_QUAD
							if( Asura_Xbox360_SkinRenderer::IsEnabled() )
							{
								Asura_Xbox360_SkinRenderer::RenderSkin( *pxAnim, *pxSkinLOD,
									s_xBaseTransform, s_axBoneTransforms );
							}
							else
							{
								Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim,
									*pxSkinLOD, s_xBaseTransform, s_xInverseBaseTransform,
									s_axBoneTransforms );
							}
#else
							Asura_Xbox360_CharacterRenderer::RenderCharacter( *pxAnim, *pxSkinLOD, *s_pxPos, *s_pxOri, s_axBonePos, s_axBoneOri );
#endif
						}
					}
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Hierarchy_System::Platform_RenderHierarchy
// Author       : Kevin Floyer-Lea
// Date         : 27-Jul-2005
//
// Description  : Handles the platform specific render of an object-hierarchy.
//
// Parameters   : 
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Hierarchy_System::Platform_RenderHierarchy( const Asura_Animation* pxAnim,
	u_int uCamVisFlags, u_int uRenderFlags, const Asura_Vector_3& xBSCentre, float fBSRadius,
	const int iTexIDOverride /*=ASURA_TEXTURE_NONE*/)
{
	Asura_Assert( pxAnim, "NULL parameter passed to Asura_Hierarchy_System::Platform_RenderHierarchy()." );

	// Select LOD mesh based on distance to first bone, since object may be animating far away from its location

	const float fDistance = (GetWorldspaceBonePos(0) - *Asura_Camera::GetPosition()).Magnitude();
	const Asura_Hierarchy_Skin* const pxSkinLOD = s_pxSkin->GetDetailSkinFromDistance( fDistance );

	if( ShouldJustRenderForCurCam() )
	{
		Asura_Xbox360_ObjectHierarchyRenderer::RenderHierarchy( *pxAnim, *pxSkinLOD );
	}
	else
	{
		for( int u = 0; u < Asura_Camera::GetNumberOfCameras(); ++u )
		{
			if( !(uCamVisFlags & (1 << u)) ) continue;

			Asura_Camera::SetCameraIndex(u);
			Asura_Render::ApplyCameraSettings();

			if( Asura_Xbox360_SkinRenderer::IsEnabled() )
			{
				Asura_Xbox360_SkinRenderer::RenderSkin( *pxAnim, *pxSkinLOD, s_xBaseTransform,
					s_axBoneTransforms );
			}
			else
			{
				Asura_Xbox360_ObjectHierarchyRenderer::RenderHierarchy( *pxAnim, *pxSkinLOD );
			}
		}
	}
}

void Asura_Hierarchy_System::Platform_RenderSkinShadow( const Asura_Animation& xAnim, const Asura_Shadow& xShadow, const u_int uCamVisFlags, const u_int uRenderFlags )
{
	// Select LOD mesh based on distance to first bone, since object may be animating far away from its location

	const float fDistance = (GetWorldspaceBonePos(0) - *Asura_Camera::GetPosition()).Magnitude();
#ifdef ASURA_USE_LOWER_SKIN_LOD_FOR_SHADOWS
	// Drop LOD an extra level for shadow
	const Asura_Hierarchy_Skin* const pxSkinLOD = s_pxSkin->GetDetailSkinFromDistance( fDistance )->GetRelativeDetailSkin(1);
#else
	const Asura_Hierarchy_Skin* const pxSkinLOD = s_pxSkin->GetDetailSkinFromDistance( fDistance );
#endif

#if defined(ASURA_HIERARCHY_QUAD)
	if(xAnim.GetSkin() == s_pxSkin)
	{
		if( s_bBonesInWorldspace )
		{
			// If the bones are in world space, have to switch them back to model space to avoid skinning inaccuracies
			for(int i = 0; i < pxSkinLOD->NumberOfBones; ++i)
			{
				//			s_axBoneTransforms[i]	= s_xInverseBaseTransform * s_axBoneTransforms[i];
				s_xInverseBaseTransform.Apply(s_axBoneTransforms[i], s_axBoneTransforms[i]);
			}
			s_bBonesInWorldspace = false;

		}

		for( int i = 0; i < pxSkinLOD->NumberOfBones; ++i )
		{
			//		s_axBoneTransforms[i]	= s_axBoneTransforms[i] * pxSkinLOD->m_pxBoneRestPoseInverseTransforms[i];
			s_axBoneTransforms[i].Apply(pxSkinLOD->m_pxBoneRestPoseInverseTransforms[i], s_axBoneTransforms[i]);
		}
	}

	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		Asura_Xbox360_SkinRenderer::RenderSkin( xAnim, *pxSkinLOD, s_xBaseTransform,
			s_axBoneTransforms, &xShadow );
	}
	else
	{
		Asura_Xbox360_CharacterRenderer::RenderShadow(xAnim, *pxSkinLOD, xShadow, s_xBaseTransform,
			s_axBoneTransforms);
	}

#else
	if(xAnim.GetSkin() == s_pxSkin)
	{
		if( s_bBonesInWorldspace )
		{
			Asura_Matrix_3x3 xOriTranspose = *s_pxOri;

			// If the bones are in world space, have to switch them back to model space to avoid skinning inaccuracies
			xOriTranspose.Transpose();

			for( int i = 0; i < pxSkinLOD->NumberOfBones; ++i )
			{
				s_axBoneOri[i] = xOriTranspose * s_axBoneOri[i];

				s_axBonePos[i] -= *s_pxPos;
				s_axBonePos[i] = s_axBonePos[i] * xOriTranspose;
			}
			s_bBonesInWorldspace = false;

		}

		for( int i = 0; i < xAnim.GetSkin()->NumberOfBones; ++i )
		{
			s_axBoneOri[i] = s_axBoneOri[i] * pxSkinLOD->BoneRestOrientation[i];
			s_axBonePos[i] -= pxSkinLOD->BoneRestPosition[i] * s_axBoneOri[i];
		}
	}

	Asura_Xbox360_CharacterRenderer::RenderShadow( xAnim, *pxSkinLOD, xShadow, *s_pxPos, *s_pxOri, s_axBonePos, s_axBoneOri );
#endif	// defined(ASURA_HIERARCHY_QUAD)
}

void Asura_Hierarchy_System::Platform_RenderHierarchyShadow( const Asura_Animation& xAnim, const Asura_Shadow& xShadow, const u_int uCamVisFlags, const u_int uRenderFlags )
{
	// Select LOD mesh based on distance to first bone, since object may be animating far away from its location

	const float fDistance = (GetWorldspaceBonePos(0) - *Asura_Camera::GetPosition()).Magnitude();
#ifdef ASURA_USE_LOWER_SKIN_LOD_FOR_SHADOWS
	// Drop LOD an extra level for shadow
	const Asura_Hierarchy_Skin* const pxSkinLOD = s_pxSkin->GetDetailSkinFromDistance( fDistance )->GetRelativeDetailSkin(1);
#else
	const Asura_Hierarchy_Skin* const pxSkinLOD = s_pxSkin->GetDetailSkinFromDistance( fDistance );
#endif

	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		Asura_Xbox360_SkinRenderer::RenderSkin( xAnim, *pxSkinLOD, s_xBaseTransform,
			s_axBoneTransforms, &xShadow );
	}
	else
	{
		Asura_Xbox360_ObjectHierarchyRenderer::RenderShadow( xAnim, *pxSkinLOD, xShadow );
	}
}


/// End of file. ///
