// Asura_Xbox360_Profile_System.cpp
// Thomas Young - 09.10.08
// Copyright (C) 2008 Rebellion, All Rights Reserved.

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#include "Core/Asura_Timers.h"						// Asura
#include "Network/Asura_NetLinkManager.h"

#include "Core/Asura_Xbox360_OS.h"					// Asura Xbox360
#include "Core/Asura_Xbox360_Profile_System.h"

#include <xbox.h>									// Xbox360
#include <qnet.h>
#include <Xonline.h>

// D E F I N E S /////////////////////////////////////////////////////////////

#define ASURA_XUID_ONLINE		0
#define ASURA_XUID_OFFLINE		1

// C O N S T A N T S /////////////////////////////////////////////////////////

static const char* g_szSaveRoot			= "save";			// root for all save files
static const char* g_szSavePrefix		= "save:\\";		// prefix this to all filenames

// S T A T I C   M E M B E R   V A R I A B L E S /////////////////////////////

Asura_Mutex											Asura_Xbox360_Profile_System::s_xMutex;
Asura_Xbox360_Profile_Error_State					Asura_Xbox360_Profile_System::s_eErrorState				= ASURA_XBOX360_PROFILE_ERROR_NONE;
Asura_Xbox360_Profile_State							Asura_Xbox360_Profile_System::s_eSystemState			= ASURA_XBOX360_PROFILE_STATE_IDLE;
Asura_Xbox360_Profile_Query_State					Asura_Xbox360_Profile_System::s_eQueryState				= ASURA_XBOX360_PROFILE_QUERY_NONE;
char*												Asura_Xbox360_Profile_System::s_szBaseFilename			= 0;
Asura_UnicodeChar*									Asura_Xbox360_Profile_System::s_szBaseDisplayName		= 0;
Asura_ProducerConsumerQueue							Asura_Xbox360_Profile_System::s_xThreadedOps			( 1, 1, ASURA_MULTITHREAD_STACK_SIZE_MEDIUM );
Asura_Xbox360_Profile_System::LoadSaveThreadData	Asura_Xbox360_Profile_System::s_xThreadData;
volatile bool										Asura_Xbox360_Profile_System::s_bUIOpen					= false;	
volatile bool										Asura_Xbox360_Profile_System::s_bIgnoreUINotification	= false;
volatile int										Asura_Xbox360_Profile_System::s_iActivePlayer			= XUSER_INDEX_ANY;
XUID												Asura_Xbox360_Profile_System::s_xActivePlayerXUID		[2];
volatile int										Asura_Xbox360_Profile_System::s_iActiveDevice			= XCONTENTDEVICE_ANY;
Asura_Xbox360_Profile_System::SaveDataSlot			Asura_Xbox360_Profile_System::s_xSlotEntries			[ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS];

// F U N C T I O N   D E F I N I T I O N S ///////////////////////////////////

void Asura_Xbox360_Profile_System::Initialise ()
{
	s_eSystemState	= ASURA_XBOX360_PROFILE_STATE_IDLE;
	s_eErrorState	= ASURA_XBOX360_PROFILE_ERROR_NONE;

	for( int iSlot = 0; iSlot < ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS; ++iSlot )
	{
		s_xSlotEntries[iSlot].szDisplayName	[0] = 0;
		s_xSlotEntries[iSlot].szFilename	[0] = 0;
	}

	s_xActivePlayerXUID[ASURA_XUID_ONLINE ] = INVALID_XUID;
	s_xActivePlayerXUID[ASURA_XUID_OFFLINE] = INVALID_XUID;
}


void Asura_Xbox360_Profile_System::Update ()
{

}


void Asura_Xbox360_Profile_System::Shutdown ()
{

}

void Asura_Xbox360_Profile_System::SetBaseFilename ( const char* const szFilename, const Asura_UnicodeChar* const szDisplayName )
{
	Asura_Assert( !s_szBaseFilename, "You have already specified a base filename in the profile system" );
	Asura_Assert( strlen(szFilename) < XCONTENT_MAX_FILENAME_LENGTH, "You have specified a filename which is too long" );

	s_szBaseFilename	= Asura_StringUtil::Duplicate( szFilename );
	s_szBaseDisplayName	= Asura_StringUtil::Duplicate( szDisplayName );
}

void Asura_Xbox360_Profile_System::BeginProfileFlow ( void* const pData, const u_int uDataSize )
{
	if( !s_szBaseFilename )
	{
		Asura_Assert( false, "You must specify a base filename for the profile sytem, aborting" );
		return;
	}

	Asura_Xbox360_Profile_State xCurrentState = GetState();

	Asura_Assert( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	if( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		SetState( ASURA_XBOX360_PROFILE_STATE_PROCESSING );

		// Since we are starting our profile flow, reset any saved information about the user and device, they are no longer relevant
		s_iActivePlayer = XUSER_INDEX_ANY;
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		s_xActivePlayerXUID[ASURA_XUID_ONLINE ] = INVALID_XUID;
		s_xActivePlayerXUID[ASURA_XUID_OFFLINE] = INVALID_XUID;

		// We only have one static data structure for thread data, but since we can only have one thread
		// for the profile system running at once, it safe to do so

		//TODO: Check here if the producer consumer queue is using this data, otherwise we will break stuff here

		Asura_Assert( !s_xThreadData.bInUse, "The thread shared storage is in use" );

		s_xThreadData.bInUse		= true;
		s_xThreadData.pData			= pData;
		s_xThreadData.uDataSize		= uDataSize;
		s_xThreadData.iSlot			= -1;

		#ifdef ASURA_ASSERT
		Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_BeginProfileFlow, 0, 0 ), "Profile system is busy, cannot perform operation" );
		#else
		s_xThreadedOps.TryToAddToQueue( Threaded_BeginProfileFlow, 0, 0 );
		#endif
	}
}

void Asura_Xbox360_Profile_System::SelectStorageDevice ( void* const pData, const u_int uDataSize )
{
	if( !s_szBaseFilename )
	{
		Asura_Assert( false, "You must specify a base filename for the profile sytem, aborting" );
		return;
	}

	Asura_Xbox360_Profile_State xCurrentState = GetState();

	Asura_Assert( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	if( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		SetState( ASURA_XBOX360_PROFILE_STATE_PROCESSING );

		//Reset the active device
		s_iActiveDevice = XCONTENTDEVICE_ANY;

		//TODO: Check here if the producer consumer queue is using this data, otherwise we will break stuff here

		Asura_Assert( !s_xThreadData.bInUse, "The thread shared storage is in use" );

		s_xThreadData.bInUse		= true;
		s_xThreadData.pData			= pData;
		s_xThreadData.uDataSize		= uDataSize;
		s_xThreadData.iSlot			= -1;

		#ifdef ASURA_ASSERT
		Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_SelectDevice, 0, 0 ), "Profile system is busy, cannot perform operation" );
		#else
		s_xThreadedOps.TryToAddToQueue( Threaded_SelectDevice, 0, 0 );
		#endif
	}
}

void Asura_Xbox360_Profile_System::Threaded_BeginProfileFlow ( void* pData, u_int uDataSize )
{
	LoadSaveThreadData* pxThreadData = &s_xThreadData;

	// Check if we need to prompt for the user to sign in
	int iActivePlayer = static_cast<int>( Asura_Xbox360_OS::GetActivePlayerIndex() );
	
	XUSER_SIGNIN_STATE xSigninState = XUserGetSigninState( iActivePlayer );
	if( xSigninState == eXUserSigninState_NotSignedIn )
	{
		if( !PromptForUser() )
		{
			// The user elected not to pick a gamer profile, we should notify the user
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_NO_PROFILE_SELECTED );
			return;
		}
	}

	// Update our active player index
	s_iActivePlayer = iActivePlayer;

	// Update our XUIDS
	if( !GetUserXUIDs ( s_xActivePlayerXUID[ ASURA_XUID_OFFLINE ], s_xActivePlayerXUID[ ASURA_XUID_ONLINE ] ) )
	{
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_NO_PROFILE_SELECTED );
		return;
	}
		
	// Prompt the user for a device on which to save, we always do this
	if( !PromptForDevice() )
	{
		// The user elected not to pick a device, we should notify the user
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_NO_DEVICE_SELECTED );
		return;
	}

	// Save the server time so that we can wait to clear the state later
	const float fServerTime = Asura_Timers::GetServerAbsoluteTime();

	// Now we have a profile and a device, check if we should load or save to it

	XCONTENT_DATA xContentData	= {0};
	xContentData.DeviceID		= s_iActiveDevice;
	xContentData.dwContentType	= XCONTENTTYPE_SAVEDGAME;
	Asura_PrintfToBfr( xContentData.szFileName,	"%s", XCONTENT_MAX_FILENAME_LENGTH,	true, s_szBaseFilename );
	swprintf( xContentData.szDisplayName, XCONTENT_MAX_DISPLAYNAME_LENGTH, L"%s", s_szBaseDisplayName );

	DWORD uDisposition = 0;
	DWORD uRet = XContentCreate(s_iActivePlayer,
								g_szSaveRoot, 
								&xContentData,
								XCONTENTFLAG_OPENALWAYS, 
								&uDisposition,
								NULL, 
								NULL );

	if( uRet != ERROR_SUCCESS )
	{
		// Failure, additional handling below
		XContentClose( g_szSaveRoot, NULL );

		if (uRet == ASURA_XBOX360_PROFILE_ERROR_CORRUPT)
		{
			// The data on this device has become corrupted, we should prompt the user to overwrite it
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_CORRUPT );
			return;
		}
		else if (	( uRet == ERROR_NOT_READY				) ||
					( uRet == ERROR_DEVICE_REMOVED			) ||
					( uRet == ERROR_MEDIA_CHANGED			) ||
					( uRet == ERROR_DEVICE_NOT_CONNECTED	) ||
					( uRet == ERROR_FILE_NOT_FOUND			) ||
					( uRet == ERROR_PATH_NOT_FOUND			) )
		{
			// Failed to create the content, report this failure and reset our active device
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
			return;
		}

		Asura_Assertf(false,("unknown Failure from XContentCreate %d %d", uRet, GetLastError() ) );
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Build the filename we want to save or load
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr( szFilename, "%s%s%s", MAX_PATH, true, g_szSavePrefix, s_szBaseFilename, ".sav" );

	if( uDisposition == XCONTENT_CREATED_NEW )
	{
		#if ( ( defined ASURA_ASSERT ) && ( defined _DEBUG ) )
		JumpToSave:
		#endif
		// We should save our data to this device
		
		SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

		HANDLE xFile = CreateFile(	szFilename,
									GENERIC_WRITE,
									0,
									0,
									CREATE_ALWAYS,
									FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
									NULL );

		bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

		if ( !bFailure )
		{
			DWORD uDataBytesWritten = 0;
			bFailure = ( WriteFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uDataBytesWritten, NULL) == FALSE );
			if( !bFailure )
			{
				bFailure = ( uDataBytesWritten != pxThreadData->uDataSize );
			}
			CloseHandle( xFile );
		}
		
		XContentClose( g_szSaveRoot, NULL );
		
		if ( bFailure )
		{
			Asura_Assertf(false,("Failure writing data to device %d", GetLastError() ) );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
			return;
		}
	}
	else if( uDisposition == XCONTENT_OPENED_EXISTING )
	{
		// We should load our data from this device

		SetState( ASURA_XBOX360_PROFILE_STATE_LOAD_PROFILE );

		// Check that the creator of this content is the logged in player

		BOOL bUserIsCreator = FALSE;
		uRet = XContentGetCreator( s_iActivePlayer, &xContentData, &bUserIsCreator, NULL, NULL);

		if( uRet == ERROR_SUCCESS)
		{
			if (bUserIsCreator == FALSE) 
			{
				XContentClose( g_szSaveRoot, NULL );
				s_iActiveDevice = XCONTENTDEVICE_ANY;
				SetErrorState( ASURA_XBOX360_PROFILE_ERROR_WRONG_USER );
				return;
			}
		}
		else
		{
			XContentClose( g_szSaveRoot, NULL );
			Asura_Assert( false, "XContentGetCreator failed" );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
			return;
		}

		// Load our data

		HANDLE xFile = CreateFile(	szFilename,
									GENERIC_READ,
									0,
									0,
									OPEN_EXISTING,
									FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
									NULL );

		bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

		if( !bFailure )
		{
			DWORD uBytesRead;
			bFailure = ( ReadFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uBytesRead, NULL) == FALSE);

			if( !bFailure )
			{
				bFailure = ( uBytesRead != pxThreadData->uDataSize );

				#if ( ( defined ASURA_ASSERT ) && ( defined _DEBUG ) )
				if( bFailure )
				{
					CloseHandle(xFile);
					Asura_Assert( false, "Did not read the number of bytes I expected from the profile save.  \
										 Because this is a debug build I will assume the profile data is just \
										 out of date and overwrite it with a fresh save when you hit ignore"  );
					DeleteFile( szFilename );
					goto JumpToSave;
				}
				#endif
			}

			CloseHandle(xFile);
			
			if ( bFailure )
			{
				XContentClose( g_szSaveRoot, NULL );
				Asura_Assertf(false,("Failure reading data from device %d", GetLastError() ) );
				s_iActiveDevice = XCONTENTDEVICE_ANY;
				SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
				return;
			}
		}
	}

	XContentClose( g_szSaveRoot, NULL );

	// check if there are also save files present
	EnumerateSaveSlots ();

	// Wait for 1 second to have passed before reverting to the idle state
	WaitUntilTime( fServerTime + 1.0f );

	// Do one last check to make sure the user didn't yank out the storage device
	if( !IsDeviceStillValid() )
	{
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED );
		return;
	}

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::Threaded_SelectDevice ( void* pData, u_int uDataSize )
{
	LoadSaveThreadData* pxThreadData = &s_xThreadData;

	// Check if we need to prompt for the user to sign in
	int iActivePlayer = static_cast<int>( Asura_Xbox360_OS::GetActivePlayerIndex() );

	XUSER_SIGNIN_STATE xSigninState = XUserGetSigninState( iActivePlayer );
	if( xSigninState == eXUserSigninState_NotSignedIn )
	{
		Asura_Assert( false, "User is not signed in, cannot select device" );
		return;
	}

	// Prompt the user for a device on which to save
	if( !PromptForDevice() )
	{
		// The user elected not to pick a device, we should notify the user
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_NO_DEVICE_SELECTED );
		return;
	}

	// Now we have a profile and a device, check if we should load or save to it

	XCONTENT_DATA xContentData	= {0};
	xContentData.DeviceID		= s_iActiveDevice;
	xContentData.dwContentType	= XCONTENTTYPE_SAVEDGAME;
	Asura_PrintfToBfr( xContentData.szFileName,	"%s", XCONTENT_MAX_FILENAME_LENGTH,	true, s_szBaseFilename );
	swprintf( xContentData.szDisplayName, XCONTENT_MAX_DISPLAYNAME_LENGTH, L"%s", s_szBaseDisplayName );

	DWORD uDisposition = 0;
	DWORD uRet = XContentCreate(s_iActivePlayer,
								g_szSaveRoot, 
								&xContentData,
								XCONTENTFLAG_OPENALWAYS, 
								&uDisposition,
								NULL, 
								NULL );

	if( uRet != ERROR_SUCCESS )
	{
		// Failure, additional handling below
		XContentClose( g_szSaveRoot, NULL );

		if (uRet == ASURA_XBOX360_PROFILE_ERROR_CORRUPT)
		{
			// The data on this device has become corrupted, we should prompt the user to overwrite it
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_CORRUPT );
			return;
		}
		else if (	( uRet == ERROR_NOT_READY				) ||
					( uRet == ERROR_DEVICE_REMOVED			) ||
					( uRet == ERROR_MEDIA_CHANGED			) ||
					( uRet == ERROR_DEVICE_NOT_CONNECTED	) ||
					( uRet == ERROR_FILE_NOT_FOUND			) ||
					( uRet == ERROR_PATH_NOT_FOUND			) )
		{
			// Failed to create the content, report this failure and reset our active device
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
			return;
		}

		Asura_Assertf(false,("unknown Failure from XContentCreate %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Save the server time so that we can wait to clear the state later
	const float fServerTime = Asura_Timers::GetServerAbsoluteTime();

	// Build the filename we want to save or load
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr( szFilename, "%s%s%s", MAX_PATH, true, g_szSavePrefix, s_szBaseFilename, ".sav" );

	if( uDisposition == XCONTENT_CREATED_NEW )
	{
		// We should save our data to this device

		SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

		HANDLE xFile = CreateFile(	szFilename,
									GENERIC_WRITE,
									0,
									0,
									CREATE_ALWAYS,
									FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
									NULL );

		bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

		if ( !bFailure )
		{
			DWORD uDataBytesWritten = 0;
			bFailure = ( WriteFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uDataBytesWritten, NULL) == FALSE );
			if( !bFailure )
			{
				bFailure = ( uDataBytesWritten != pxThreadData->uDataSize );
			}
			CloseHandle( xFile );
		}

		XContentClose( g_szSaveRoot, NULL );

		if ( bFailure )
		{
			Asura_Assertf(false,("Failure writing data to device %d", GetLastError() ) );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
			return;
		}
	}
	else if( uDisposition == XCONTENT_OPENED_EXISTING )
	{
		// We should load our data from this device

		SetState( ASURA_XBOX360_PROFILE_STATE_LOAD_PROFILE );

		// Check that the creator of this content is the logged in player

		BOOL bUserIsCreator = FALSE;
		uRet = XContentGetCreator( s_iActivePlayer, &xContentData, &bUserIsCreator, NULL, NULL);

		if( uRet == ERROR_SUCCESS)
		{
			if (bUserIsCreator == FALSE) 
			{
				XContentClose( g_szSaveRoot, NULL );
				s_iActiveDevice = XCONTENTDEVICE_ANY;
				SetErrorState( ASURA_XBOX360_PROFILE_ERROR_WRONG_USER );
				return;
			}
		}
		else
		{
			XContentClose( g_szSaveRoot, NULL );
			Asura_Assert( false, "XContentGetCreator failed" );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
			return;
		}

		// Load our data

		HANDLE xFile = CreateFile(	szFilename,
			GENERIC_READ,
			0,
			0,
			OPEN_EXISTING,
			FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
			NULL );

		bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

		if( !bFailure )
		{
			DWORD uBytesRead;
			bFailure = ( ReadFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uBytesRead, NULL) == FALSE);

			if( !bFailure )
			{
				bFailure = ( uBytesRead != pxThreadData->uDataSize );
			}

			CloseHandle(xFile);

			if ( bFailure )
			{
				XContentClose( g_szSaveRoot, NULL );
				Asura_Assertf(false,("Failure reading data from device %d", GetLastError() ) );
				s_iActiveDevice = XCONTENTDEVICE_ANY;
				SetQueryState( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT );
				return;
			}
		}
	}

	XContentClose( g_szSaveRoot, NULL );

	// check if there are also save files present
	EnumerateSaveSlots ();

	// Wait for 1 second to have passed before reverting to the idle state
	WaitUntilTime( fServerTime + 1.0f );

	// Do one last check to make sure the user didn't yank out the storage device
	if( !IsDeviceStillValid() )
	{
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED );
		return;
	}

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::Threaded_LoadProfileData ( void* pData, u_int uDataSize )
{
	LoadSaveThreadData* pxThreadData = &s_xThreadData;

	SetState( ASURA_XBOX360_PROFILE_STATE_LOAD_PROFILE );

	XCONTENT_DATA xContentData	= {0};
	xContentData.DeviceID		= s_iActiveDevice;
	xContentData.dwContentType	= XCONTENTTYPE_SAVEDGAME;
	Asura_PrintfToBfr( xContentData.szFileName,	"%s", XCONTENT_MAX_FILENAME_LENGTH,	true, s_szBaseFilename );
	swprintf( xContentData.szDisplayName, XCONTENT_MAX_DISPLAYNAME_LENGTH, L"%s", s_szBaseDisplayName );

	DWORD uDisposition = 0;
	DWORD uRet = XContentCreate(s_iActivePlayer,
								g_szSaveRoot, 
								&xContentData,
								XCONTENTFLAG_OPENEXISTING, 
								&uDisposition,
								NULL, 
								NULL );

	// Because this is a sort of re-load, we are less forgiving of errors, most of them are unrecoverable to us
	if( uRet != ERROR_SUCCESS )
	{
		// Failure, additional handling below
		XContentClose( g_szSaveRoot, NULL );

		Asura_Assertf(false,("Failure from XContentCreate %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Save the server time so that we can wait to clear the state later
	const float fServerTime = Asura_Timers::GetServerAbsoluteTime();

	// Build the filename we want to save or load
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr( szFilename, "%s%s%s", MAX_PATH, true, g_szSavePrefix, s_szBaseFilename, ".sav" );

	// Check that the creator of this content is the logged in player

	BOOL bUserIsCreator = FALSE;
	uRet = XContentGetCreator( s_iActivePlayer, &xContentData, &bUserIsCreator, NULL, NULL);

	if( uRet == ERROR_SUCCESS)
	{
		if (bUserIsCreator == FALSE) 
		{
			XContentClose( g_szSaveRoot, NULL );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_WRONG_USER );
			return;
		}
	}
	else
	{
		Asura_Assertf(false,("Failure from XContentGetCreator %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Load our data

	HANDLE xFile = CreateFile(	szFilename,
								GENERIC_READ,
								0,
								0,
								OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );

	bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

	if( !bFailure )
	{
		DWORD uBytesRead;
		bFailure = ( ReadFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uBytesRead, NULL) == FALSE);

		if( !bFailure )
		{
			bFailure = ( uBytesRead != pxThreadData->uDataSize );
		}

		CloseHandle(xFile);

		if ( bFailure )
		{
			XContentClose( g_szSaveRoot, NULL );
			Asura_Assertf(false,("Failure reading data from device %d", GetLastError() ) );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
			return;
		}
	}

	XContentClose( g_szSaveRoot, NULL );

	// Wait for 3 seconds to have passed before reverting to the idle state
	WaitUntilTime( fServerTime + 1.0f );

	// Do one last check to make sure the user didn't yank out the storage device
	if( !IsDeviceStillValid() )
	{
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED );
		return;
	}

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::Threaded_SaveProfileData ( void* pData, u_int uDataSize )
{
	LoadSaveThreadData* pxThreadData = &s_xThreadData;

	SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

	// Save the server time so that we can wait to clear the state later
	const float fServerTime = Asura_Timers::GetServerAbsoluteTime();

	XCONTENT_DATA xContentData	= {0};
	xContentData.DeviceID		= s_iActiveDevice;
	xContentData.dwContentType	= XCONTENTTYPE_SAVEDGAME;
	Asura_PrintfToBfr( xContentData.szFileName,	"%s", XCONTENT_MAX_FILENAME_LENGTH,	true, s_szBaseFilename );
	swprintf( xContentData.szDisplayName, XCONTENT_MAX_DISPLAYNAME_LENGTH, L"%s", s_szBaseDisplayName );

	DWORD uDisposition = 0;
	DWORD uRet = XContentCreate(s_iActivePlayer,
								g_szSaveRoot, 
								&xContentData,
								XCONTENTFLAG_OPENALWAYS, 
								&uDisposition,
								NULL, 
								NULL );

	// we are fairly intolerant of errors here, plus we dont give them the oppertunity to select
	// another device because it isn't really appropriate
	if( uRet != ERROR_SUCCESS )
	{
		// Failure, additional handling below
		XContentClose( g_szSaveRoot, NULL );

		Asura_Assertf(false,("Failure from XContentCreate %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Build the filename we want to save
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr( szFilename, "%s%s%s", MAX_PATH, true, g_szSavePrefix, s_szBaseFilename, ".sav" );

	HANDLE xFile = CreateFile(	szFilename,
								GENERIC_WRITE,
								0,
								0,
								CREATE_ALWAYS,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );

	bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

	if ( !bFailure )
	{
		DWORD uDataBytesWritten = 0;
		bFailure = ( WriteFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uDataBytesWritten, NULL) == FALSE );
		if( !bFailure )
		{
			bFailure = ( uDataBytesWritten != pxThreadData->uDataSize );
		}
		CloseHandle( xFile );
	}

	XContentClose( g_szSaveRoot, NULL );

	if ( bFailure )
	{
		Asura_Assertf(false,("Failure writing data to device %d", GetLastError() ) );
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}
	

	XContentClose( g_szSaveRoot, NULL );

	// Wait for 1 seconds to have passed before reverting to the idle state
	WaitUntilTime( fServerTime + 1.0f );

	// Do one last check to make sure the user didn't yank out the storage device
	if( !IsDeviceStillValid() )
	{
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED );
		return;
	}

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::Threaded_LoadCheckpoint ( void* pData, u_int uDataSize )
{
	LoadSaveThreadData* pxThreadData = &s_xThreadData;

	SetState( ASURA_XBOX360_PROFILE_STATE_LOAD_CHECKPOINT );

	XCONTENT_DATA xContentData	= {0};
	xContentData.DeviceID		= s_iActiveDevice;
	xContentData.dwContentType	= XCONTENTTYPE_SAVEDGAME;
	Asura_PrintfToBfr( xContentData.szFileName,	"%s", XCONTENT_MAX_FILENAME_LENGTH,	true, s_xSlotEntries[ pxThreadData->iSlot].szFilename );
	swprintf( xContentData.szDisplayName, XCONTENT_MAX_DISPLAYNAME_LENGTH, L"%s", s_xSlotEntries[ pxThreadData->iSlot].szDisplayName );

	DWORD uDisposition = 0;
	DWORD uRet = XContentCreate(s_iActivePlayer,
								g_szSaveRoot, 
								&xContentData,
								XCONTENTFLAG_OPENEXISTING, 
								&uDisposition,
								NULL, 
								NULL );

	// We are unforgiving of errors in this case
	if( uRet != ERROR_SUCCESS )
	{
		// Failure, additional handling below
		XContentClose( g_szSaveRoot, NULL );

		Asura_Assertf(false,("Failure from XContentCreate %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Build the filename we want to save or load
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr( szFilename, "%s%s%s", MAX_PATH, true, g_szSavePrefix, s_szBaseFilename, ".sav" );

	// Check that the creator of this content is the logged in player

	BOOL bUserIsCreator = FALSE;
	uRet = XContentGetCreator( s_iActivePlayer, &xContentData, &bUserIsCreator, NULL, NULL);

	if( uRet == ERROR_SUCCESS)
	{
		if (bUserIsCreator == FALSE) 
		{
			XContentClose( g_szSaveRoot, NULL );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_WRONG_USER );
			return;
		}
	}
	else
	{
		Asura_Assertf(false,("Failure from XContentGetCreator %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Load our data

	HANDLE xFile = CreateFile(	szFilename,
								GENERIC_READ,
								0,
								0,
								OPEN_EXISTING,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );

	bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

	if( !bFailure )
	{
		DWORD uBytesRead;
		bFailure = ( ReadFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uBytesRead, NULL) == FALSE);

		if( !bFailure )
		{
			bFailure = ( uBytesRead != pxThreadData->uDataSize );
		}

		CloseHandle(xFile);

		if ( bFailure )
		{
			XContentClose( g_szSaveRoot, NULL );
			Asura_Assertf(false,("Failure reading data from device %d", GetLastError() ) );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
			return;
		}
	}

	XContentClose( g_szSaveRoot, NULL );

	// We dont bother waiting to set the state when loading a checkpoint

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::Threaded_SaveCheckpoint ( void* pData, u_int uDataSize )
{
	LoadSaveThreadData* pxThreadData = &s_xThreadData;

	SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_CHECKPOINT );

	// Save the server time so that we can wait to clear the state later
	const float fServerTime = Asura_Timers::GetServerAbsoluteTime();

	XCONTENT_DATA xContentData	= {0};
	xContentData.DeviceID		= s_iActiveDevice;
	xContentData.dwContentType	= XCONTENTTYPE_SAVEDGAME;
	Asura_PrintfToBfr( xContentData.szFileName,	"%s", XCONTENT_MAX_FILENAME_LENGTH,	true, s_xSlotEntries[ pxThreadData->iSlot].szFilename );
	swprintf( xContentData.szDisplayName, XCONTENT_MAX_DISPLAYNAME_LENGTH, L"%s", s_xSlotEntries[ pxThreadData->iSlot].szDisplayName );

	DWORD uDisposition = 0;
	DWORD uRet = XContentCreate(s_iActivePlayer,
								g_szSaveRoot, 
								&xContentData,
								XCONTENTFLAG_OPENALWAYS, 
								&uDisposition,
								NULL, 
								NULL );

	// We are unforgiving of errors in this case
	if( uRet != ERROR_SUCCESS )
	{
		// Failure, additional handling below
		XContentClose( g_szSaveRoot, NULL );

		Asura_Assertf(false,("Failure from XContentCreate %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Check that the creator of this content is the logged in player

	BOOL bUserIsCreator = FALSE;
	uRet = XContentGetCreator( s_iActivePlayer, &xContentData, &bUserIsCreator, NULL, NULL);

	if( uRet == ERROR_SUCCESS)
	{
		if (bUserIsCreator == FALSE) 
		{
			XContentClose( g_szSaveRoot, NULL );
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_WRONG_USER );
			return;
		}
	}
	else
	{
		Asura_Assertf(false,("Failure from XContentGetCreator %d", uRet));
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	// Save our data

	// Build the filename we want to save
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr( szFilename, "%s%s%s", MAX_PATH, true, g_szSavePrefix, s_szBaseFilename, ".sav" );

	HANDLE xFile = CreateFile(	szFilename,
								GENERIC_WRITE,
								0,
								0,
								CREATE_ALWAYS,
								FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN,
								NULL );

	bool bFailure = ( xFile == INVALID_HANDLE_VALUE );

	if ( !bFailure )
	{
		DWORD uDataBytesWritten = 0;
		bFailure = ( WriteFile( xFile, pxThreadData->pData, pxThreadData->uDataSize, &uDataBytesWritten, NULL ) == FALSE );
		if( !bFailure )
		{
			bFailure = ( uDataBytesWritten != pxThreadData->uDataSize );
		}
		CloseHandle( xFile );
	}

	if ( bFailure )
	{
		XContentClose( g_szSaveRoot, NULL );
		Asura_Assertf(false,("Failure reading data from device %d", GetLastError() ) );
		s_iActiveDevice = XCONTENTDEVICE_ANY;
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_GENERIC_DEVICE_FAILURE );
		return;
	}

	XContentClose( g_szSaveRoot, NULL );

	// Re-enumerate the save slots since we just changed one
	EnumerateSaveSlots ();

	// Wait for 1 second to have passed before reverting to the idle state
	WaitUntilTime( fServerTime + 1.0f );

	// Do one last check to make sure the user didn't yank out the storage device
	if( !IsDeviceStillValid() )
	{
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED );
		return;
	}

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::EnumerateSaveSlots ()
{
	// Clear down our current slot entries

	for( int iSlot = 0; iSlot < ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS; ++iSlot )
	{
		s_xSlotEntries[iSlot].szDisplayName	[0] = 0;
		s_xSlotEntries[iSlot].szFilename	[0] = 0;
	}

	// Create a content enumerator

	DWORD	uBufferSize;
	HANDLE	xEnumHandle;

	DWORD iRet = XContentCreateEnumerator	( s_iActivePlayer
											, s_iActiveDevice
											, XCONTENTTYPE_SAVEDGAME
											, XCONTENTFLAG_ENUM_EXCLUDECOMMON
											, ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS + 1		//Save slots + profile data
											, &uBufferSize
											, &xEnumHandle );

	void* pBuffer = Asura_Allocate( uBufferSize );

	if( iRet != ERROR_SUCCESS )
	{
		Asura_Assert( false, "XContentCreateEnumerator failed" );
		return;
	}

	DWORD uItems;

	iRet = XEnumerate( xEnumHandle, pBuffer, uBufferSize, &uItems, NULL );
	
	if( iRet != ERROR_SUCCESS )
	{
		Asura_Deallocate( pBuffer );
		Asura_Assert( false, "XEnumerate failed" );
		return;
	}

	// Copy the results of our enumeration into our slot list

	int iSlot = 0;

	const int iSlotDigit = strlen( s_szBaseFilename );

	for( u_int uEnumResult = 0; uEnumResult < uItems; ++uEnumResult )
	{
		XCONTENT_DATA* pxData = ( reinterpret_cast<XCONTENT_DATA*>( pBuffer ) + uEnumResult );
		// Is this the profile data?
		if( strncmp( pxData->szFileName, s_szBaseFilename, XCONTENT_MAX_FILENAME_LENGTH ) != 0 )
		{
			// no, copy this into our slot list

			//first we need to identify which slot this file represents, this will be the digit after the filename (NOTE: expects 1 digit slot number!)
			iSlot = static_cast<int>( pxData->szFileName[iSlotDigit] ) - 48;

			if( ( iSlot >= 0 ) && ( iSlot < ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS ) )
			{
				Asura_Assert( !s_xSlotEntries[iSlot].szFilename[0], "Slot collision" );

				wcsncpy( s_xSlotEntries[iSlot].szDisplayName	, pxData->szDisplayName,	XCONTENT_MAX_DISPLAYNAME_LENGTH	);
				strncpy( s_xSlotEntries[iSlot].szFilename		, pxData->szFileName,		XCONTENT_MAX_FILENAME_LENGTH	);

				//need to null terminate the filename as well
				s_xSlotEntries[iSlot].szFilename[ XCONTENT_MAX_FILENAME_LENGTH ] = 0;
			}
		}
	}

	Asura_Deallocate( pBuffer );
}

bool Asura_Xbox360_Profile_System::PromptForUser ()
{
	IgnoreUINotification( true );

	XShowSigninUI( 1, XSSUI_FLAGS_LOCALSIGNINONLY );
	s_iActiveDevice = XCONTENTDEVICE_ANY;

	do
	{ 
		Asura_Thread_System::DelayCurrentThread( 0.1f );
	}
	while( s_bUIOpen );
	
	IgnoreUINotification( false );

	XUSER_SIGNIN_STATE xSigninState = XUserGetSigninState( Asura_Xbox360_OS::GetActivePlayerIndex() );
	if(xSigninState == eXUserSigninState_NotSignedIn)
	{
		//The user elected to not sign in
		return false;
	}
	return true;
}




bool Asura_Xbox360_Profile_System::PromptForDevice ( int iFreeSpaceRequired )
{
	// Invalidate our device since we are about to pick a new one		
	s_iActiveDevice = XCONTENTDEVICE_ANY;

	DWORD dwDeviceUIFlags = XCONTENTFLAG_FORCE_SHOW_UI;

	XOVERLAPPED xOverlapped = {0};

	ULARGE_INTEGER xBytesRequested = {0};
	xBytesRequested.QuadPart = XContentCalculateSize( iFreeSpaceRequired, 0 );
	
	XCONTENTDEVICEID xDevice;

	IgnoreUINotification( true );

	#ifdef ASURA_ASSERT
	DWORD iRet = 
	#endif
	XShowDeviceSelectorUI(	s_iActivePlayer, 
							XCONTENTTYPE_SAVEDGAME, 
							dwDeviceUIFlags,
							xBytesRequested, 
							&xDevice, 
							&xOverlapped			);

	

	Asura_Assert( iRet == ERROR_IO_PENDING ,"Error showing device selector UI" );

	WaitForOverlapped( &xOverlapped );

	IgnoreUINotification( false );

	s_iActiveDevice = static_cast<int>( xDevice );
	
	// if the user picked a valid device, it wont be XCONTENTDEVICE_ANY
	return ( s_iActiveDevice != XCONTENTDEVICE_ANY );
}


Asura_Xbox360_Profile_State Asura_Xbox360_Profile_System::GetState ()
{
	Asura_Xbox360_Profile_State xSystemState;
	
	s_xMutex.Lock	();
	
		xSystemState = s_eSystemState;

	s_xMutex.Unlock	();
	
	return xSystemState;
}


bool Asura_Xbox360_Profile_System::HasValidGamerProfile ()
{
	#ifdef ASURA_ASSERT
	Asura_Xbox360_Profile_State xState = GetState();
	Asura_Assert(	( xState != ASURA_XBOX360_PROFILE_STATE_PROCESSING		) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_LOAD_PROFILE	) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE	) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_LOAD_CHECKPOINT	) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_SAVE_CHECKPOINT	) ,
					"Probably shouldnt be trying to access this information at this time" );
	#endif
	return ( s_iActivePlayer != XUSER_INDEX_ANY );
}


bool Asura_Xbox360_Profile_System::HasValidStorageDevice ()
{
	#ifdef ASURA_ASSERT
	Asura_Xbox360_Profile_State xState = GetState();
	Asura_Assert(	( xState != ASURA_XBOX360_PROFILE_STATE_PROCESSING		) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_LOAD_PROFILE	) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE	) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_LOAD_CHECKPOINT	) &&
					( xState != ASURA_XBOX360_PROFILE_STATE_SAVE_CHECKPOINT	) ,
					"Probably shouldnt be trying to access this information at this time" );
	#endif
	return ( s_iActiveDevice != XCONTENTDEVICE_ANY );
}


int Asura_Xbox360_Profile_System::GetActiveStorageDevice()
{
	Asura_Assert(HasValidStorageDevice(), "Should check if the storage device is valid before getting it.");

	return s_iActiveDevice;
}


void Asura_Xbox360_Profile_System::SaveProfileData ( const void* const pData, const u_int uDataSize )
{
	if( !s_szBaseFilename )
	{
		Asura_Assert( false, "You must specify a base filename for the profile sytem, aborting" );
		return;
	}

	if( !HasValidGamerProfile() || !HasValidStorageDevice() )
	{
		// NOTE: This might seem like it could just return silently, but if this function has been called with an invalid user/device
		// it probably means the logic is flawed higher up, e.g. a 'load profile' button being enabled when we cannot load
		Asura_Assert( HasValidGamerProfile(),	"There is no valid gamer profile, you should not be allowing this function to be called" );
		Asura_Assert( HasValidStorageDevice(),	"There is no valid storage device, you should not be allowing this function to be called" );
		return;
	}

	Asura_Xbox360_Profile_State xCurrentState = GetState();

	Asura_Assert( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	if( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

		//TODO: Check here if the producer consumer queue is using this data, otherwise we will break stuff here

		Asura_Assert( !s_xThreadData.bInUse, "The thread shared storage is in use" );

		s_xThreadData.bInUse		= true;
		s_xThreadData.pData			= const_cast<void*>( pData );
		s_xThreadData.uDataSize		= uDataSize;
		s_xThreadData.iSlot			= -1;

		#ifdef ASURA_ASSERT
		Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_SaveProfileData, 0, 0 ), "Profile system is busy, cannot perform operation" );
		#else
		s_xThreadedOps.TryToAddToQueue( Threaded_SaveProfileData, 0, 0 );
		#endif
	}
}


void Asura_Xbox360_Profile_System::LoadProfileData ( const void* const pData, const u_int uDataSize )
{
	// This is more of a re-load since we should already have loaded as part of the regular flow
	if( !s_szBaseFilename )
	{
		Asura_Assert( false, "You must specify a base filename for the profile sytem, aborting" );
		return;
	}

	if( !HasValidGamerProfile() || !HasValidStorageDevice() )
	{
		// NOTE: This might seem like it could just return silently, but if this function has been called with an invalid user/device
		// it probably means the logic is flawed higher up, e.g. a 'load profile' button being enabled when we cannot load
		Asura_Assert( HasValidGamerProfile(),	"There is no valid gamer profile, you should not be allowing this function to be called" );
		Asura_Assert( HasValidStorageDevice(),	"There is no valid storage device, you should not be allowing this function to be called" );
		return;
	}

	Asura_Xbox360_Profile_State xCurrentState = GetState();

	Asura_Assert( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	if( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

		//TODO: Check here if the producer consumer queue is using this data, otherwise we will break stuff here

		Asura_Assert( !s_xThreadData.bInUse, "The thread shared storage is in use" );

		s_xThreadData.bInUse	= true;
		s_xThreadData.pData		= const_cast<void*>(pData);
		s_xThreadData.uDataSize	= uDataSize;
		s_xThreadData.iSlot		= -1;

		#ifdef ASURA_ASSERT
		Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_LoadProfileData, 0, 0 ), "Profile system is busy, cannot perform operation" );
		#else
		s_xThreadedOps.TryToAddToQueue( Threaded_LoadProfileData, 0, 0 );
		#endif
	}
}


void Asura_Xbox360_Profile_System::SaveCheckpoint ( const int iSlot, const void* const pData, const u_int uDataSize, const Asura_UnicodeChar* const szDisplayName )
{
	if( !s_szBaseFilename )
	{
		Asura_Assert( false, "You must specify a base filename for the profile sytem, aborting" );
		return;
	}

	if( !HasValidGamerProfile() || !HasValidStorageDevice() )
	{
		// NOTE: This might seem like it could just return silently, but if this function has been called with an invalid user/device
		// it probably means the logic is flawed higher up, e.g. a 'load profile' button being enabled when we cannot load
		Asura_Assert( HasValidGamerProfile(),	"There is no valid gamer profile, you should not be allowing this function to be called" );
		Asura_Assert( HasValidStorageDevice(),	"There is no valid storage device, you should not be allowing this function to be called" );
		return;
	}

	Asura_Xbox360_Profile_State xCurrentState = GetState();

	Asura_Assert( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	if( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

		//TODO: Check here if the producer consumer queue is using this data, otherwise we will break stuff here

		Asura_Assert( !s_xThreadData.bInUse, "The thread shared storage is in use" );

		s_xThreadData.bInUse		= true;
		s_xThreadData.pData			= const_cast<void*>(pData);
		s_xThreadData.uDataSize		= uDataSize;
		s_xThreadData.iSlot			= iSlot;

		wcsncpy( s_xSlotEntries[iSlot].szDisplayName	, szDisplayName,		XCONTENT_MAX_DISPLAYNAME_LENGTH	);
		Asura_PrintfToBfr( s_xSlotEntries[iSlot].szFilename, "%s%d", XCONTENT_MAX_FILENAME_LENGTH, true, s_szBaseFilename, iSlot );

		#ifdef ASURA_ASSERT
		Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_SaveCheckpoint, 0, 0 ), "Profile system is busy, cannot perform operation" );
		#else
		s_xThreadedOps.TryToAddToQueue( Threaded_SaveCheckpoint, 0, 0 );
		#endif
	}
}


void Asura_Xbox360_Profile_System::LoadCheckpoint ( int iSlot, void* const pData, const u_int uDataSize )
{
	if( !s_szBaseFilename )
	{
		Asura_Assert( false, "You must specify a base filename for the profile sytem, aborting" );
		return;
	}

	if( !SlotContainsData( iSlot ) )
	{
		Asura_Assert( false, "Attempting to load from a slot which does not contain any data" );
		return;
	}

	if( !HasValidGamerProfile() || !HasValidStorageDevice() )
	{
		// NOTE: This might seem like it could just return silently, but if this function has been called with an invalid user/device
		// it probably means the logic is flawed higher up, e.g. a 'load profile' button being enabled when we cannot load
		Asura_Assert( HasValidGamerProfile(),	"There is no valid gamer profile, you should not be allowing this function to be called" );
		Asura_Assert( HasValidStorageDevice(),	"There is no valid storage device, you should not be allowing this function to be called" );
		return;
	}

	Asura_Xbox360_Profile_State xCurrentState = GetState();

	Asura_Assert( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	if( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE )
	{
		SetState( ASURA_XBOX360_PROFILE_STATE_SAVE_PROFILE );

		//TODO: Check here if the producer consumer queue is using this data, otherwise we will break stuff here

		Asura_Assert( !s_xThreadData.bInUse, "The thread shared storage is in use" );

		s_xThreadData.bInUse		= true;
		s_xThreadData.pData			= pData;
		s_xThreadData.uDataSize		= uDataSize;
		s_xThreadData.iSlot			= iSlot;

		#ifdef ASURA_ASSERT
		Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_LoadCheckpoint, 0, 0 ), "Profile system is busy, cannot perform operation" );
		#else
		s_xThreadedOps.TryToAddToQueue( Threaded_LoadCheckpoint, 0, 0 );
		#endif
	}
}


const Asura_UnicodeChar* Asura_Xbox360_Profile_System::GetSlotDisplayText ( const int iSlot )
{
	Asura_Assert( GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	Asura_Assert( ( iSlot >= 0 ) && ( iSlot < ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS ), "Invalid slot specified" );

	if( s_xSlotEntries[iSlot].szFilename[0] == 0 )
	{
		return 0;
	}
	return &s_xSlotEntries[iSlot].szDisplayName[0];
}


bool Asura_Xbox360_Profile_System::SlotContainsData ( const int iSlot )
{
	Asura_Assert( GetState() == ASURA_XBOX360_PROFILE_STATE_IDLE, "Profile system should be idle when calling this function" );
	Asura_Assert( ( iSlot >= 0 ) && ( iSlot < ASURA_XBOX360_PROFILE_MAX_SAVE_SLOTS ), "Invalid slot specified" );
	
	return ( s_xSlotEntries[iSlot].szFilename[0] != 0 );
}


Asura_Xbox360_Profile_Query_State Asura_Xbox360_Profile_System::GetPendingQuery ()
{
	Asura_Xbox360_Profile_Query_State xQueryState;

	s_xMutex.Lock	();

		xQueryState	= s_eQueryState;
		
	s_xMutex.Unlock	();

	return xQueryState;
}


void Asura_Xbox360_Profile_System::SetQueryResponse ( const Asura_Xbox360_Profile_QueryResponse eResponse )
{
	Asura_Assert( GetState() == ASURA_XBOX360_PROFILE_STATE_QUERY, "Profile system should be expecting a response when calling this function" );

	Asura_Xbox360_Profile_Query_State eQueryState = GetPendingQuery();

	switch( eQueryState )
	{
		case ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT:
		{
			if( eResponse == ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT_YES )
			{
				// The user wants to select a new device
				Asura_Assert( s_xThreadData.bInUse, "The thread data does not contain the information we need" );
				
				if( s_xThreadData.bInUse )
				{
					#ifdef ASURA_ASSERT
					Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_SelectDevice, 0, 0 ), "Profile system is busy, cannot perform operation" );
					#else
					s_xThreadedOps.TryToAddToQueue( Threaded_SelectDevice, 0, 0 );
					#endif
				}
				return;
			}
			else if( ASURA_XBOX360_PROFILE_QUERY_DEVICE_SELECT_NO )
			{
				// The user does not want to select a different device
				SetErrorState( ASURA_XBOX360_PROFILE_ERROR_NO_DEVICE_SELECTED );
				return;
			}
		}
		case ASURA_XBOX360_PROFILE_QUERY_CORRUPT:
		{
			// The user wants to select a new device
			Asura_Assert( s_xThreadData.bInUse, "The thread data does not contain the information we need" );

			switch ( eResponse )
			{
				case ASURA_XBOX360_PROFILE_QUERY_CORRUPT_OVERWRITE:
				{
					if( s_xThreadData.bInUse )
					{
						#ifdef ASURA_ASSERT
						Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_SaveProfileData, 0, 0 ), "Profile system is busy, cannot perform operation" );
						#else
						s_xThreadedOps.TryToAddToQueue( Threaded_SaveProfileData, 0, 0 );
						#endif
					}
					return;
				}
				break;

				case ASURA_XBOX360_PROFILE_QUERY_CORRUPT_SELECT_DEVICE:
				{
					if( s_xThreadData.bInUse )
					{
						#ifdef ASURA_ASSERT
						Asura_Assert( s_xThreadedOps.TryToAddToQueue( Threaded_SelectDevice, 0, 0 ), "Profile system is busy, cannot perform operation" );
						#else
						s_xThreadedOps.TryToAddToQueue( Threaded_SelectDevice, 0, 0 );
						#endif
					}
					return;
				}
				break;

				case ASURA_XBOX360_PROFILE_QUERY_CORRUPT_IGNORE:
				{
					SetErrorState(  ASURA_XBOX360_PROFILE_ERROR_NO_DEVICE_SELECTED );
					return;
				}
				break;
			}
		}
	}

	Asura_Assert( false, "Query response was invalid" );
}


Asura_Xbox360_Profile_Error_State Asura_Xbox360_Profile_System::GetErrorState ()
{
	Asura_Xbox360_Profile_Error_State eErrorState;

	s_xMutex.Lock	();

		eErrorState = s_eErrorState;

	s_xMutex.Unlock	();

	return eErrorState;
}


void Asura_Xbox360_Profile_System::ClearErrorState ()
{
	Asura_Assert( GetState() == ASURA_XBOX360_PROFILE_STATE_ERROR, "The profile system is not in error should not be clearing error state" );

	SetState( ASURA_XBOX360_PROFILE_STATE_IDLE );
}


void Asura_Xbox360_Profile_System::SetErrorState ( const Asura_Xbox360_Profile_Error_State eErrorState )
{
	s_xMutex.Lock	();

		// Set our state variables

		s_eErrorState	= eErrorState;
		s_eSystemState	= ASURA_XBOX360_PROFILE_STATE_ERROR; 

		// Invalidate our thread data

		s_xThreadData.bInUse = false;

	s_xMutex.Unlock	();
}

void Asura_Xbox360_Profile_System::SetQueryState ( const Asura_Xbox360_Profile_Query_State eQueryState )
{
	s_xMutex.Lock	();

		s_eQueryState	= eQueryState;
		s_eSystemState	= ASURA_XBOX360_PROFILE_STATE_QUERY; 

	s_xMutex.Unlock	();
}

void Asura_Xbox360_Profile_System::SetState ( const Asura_Xbox360_Profile_State eState )
{
	Asura_Assert( eState != ASURA_XBOX360_PROFILE_STATE_ERROR, "Call SetErrorState to safely set the error state"  );
	Asura_Assert( eState != ASURA_XBOX360_PROFILE_STATE_QUERY, "Call SetQueryState to safely set the query state"  );

	s_xMutex.Lock	();

		s_eSystemState = eState;

		// Invalidate our thread data

		if( eState == ASURA_XBOX360_PROFILE_STATE_IDLE )
		{
			s_xThreadData.bInUse = false;
		}

	s_xMutex.Unlock	();
}

void Asura_Xbox360_Profile_System::GetFullState ( Asura_Xbox360_Profile_State& eSystemState, Asura_Xbox360_Profile_Query_State& eQueryState, Asura_Xbox360_Profile_Error_State& eErrorState )
{
	s_xMutex.Lock	();

		eSystemState	= s_eSystemState;
		eQueryState		= s_eQueryState;
		eErrorState		= s_eErrorState;

	s_xMutex.Unlock	();
}

void Asura_Xbox360_Profile_System::SetUIOpen ( const bool bOpen )
{
	s_bUIOpen = bOpen;

	if(	ShouldIgnoreUINotification() )
	{
		return;
	}

	// Check if the user has changed signin

	if( s_iActivePlayer == XUSER_INDEX_ANY )
	{
		// there is no active user
		return;
	}

	XUID xOffline, xOnline;
	GetUserXUIDs ( xOffline, xOnline );

	if(		( xOffline	!= s_xActivePlayerXUID[ ASURA_XUID_OFFLINE ] ) 
		||	( xOnline	!= s_xActivePlayerXUID[ ASURA_XUID_ONLINE  ] )  )
	{
		// User has changed
		SetErrorState( ASURA_XBOX360_PROFILE_ERROR_LOGOUT );

		// The user will have to go through the profile flow again so we might as well zero these
		s_xActivePlayerXUID[ ASURA_XUID_OFFLINE ] = 0;
		s_xActivePlayerXUID[ ASURA_XUID_ONLINE  ] = 0;

		#ifdef ASURA_SUPPORT_EXTERNAL_NETWORK
		// Set ourselves as disconnected from the network
		Asura_NetLinkManager::DropConnection();
		#endif
	}
}

void Asura_Xbox360_Profile_System::WaitForOverlapped ( const XOVERLAPPED* const pxOverlap)
{
	do 
	{
		Asura_Thread_System::DelayCurrentThread( 0.1f );
	} 
	while ( *( ( volatile ULONG_PTR* )( &( pxOverlap )->InternalLow ) ) == ERROR_IO_PENDING );
}

void Asura_Xbox360_Profile_System::WaitUntilTime ( const float fServerTime )
{
	do 
	{
		Asura_Thread_System::DelayCurrentThread( 0.1f );
	} 
	while( Asura_Timers::GetServerAbsoluteTime() < fServerTime );
}

bool Asura_Xbox360_Profile_System::GetUserXUIDs ( XUID& xOffline, XUID& xOnline )
{
	// Get the offline XUID

	XUSER_SIGNIN_INFO xInfo;
	DWORD uSuccess = XUserGetSigninInfo(static_cast<DWORD>( s_iActivePlayer ), XUSER_GET_SIGNIN_INFO_OFFLINE_XUID_ONLY, &xInfo );
	Asura_Assert((uSuccess == ERROR_SUCCESS) || (uSuccess == ERROR_NO_SUCH_USER) , "Failed to get offline signin info for current user");
	
	if( uSuccess == ERROR_SUCCESS )
	{
		xOffline = xInfo.xuid;
	}
	else if(uSuccess == ERROR_NO_SUCH_USER)
	{
		xOffline = 0;
	}
	else
	{
		xOffline	= 0;
		xOnline		= 0;
		return false;
	}

	// Get the online XUID

	uSuccess = XUserGetSigninInfo(static_cast<DWORD>( s_iActivePlayer ), XUSER_GET_SIGNIN_INFO_ONLINE_XUID_ONLY, &xInfo );
	Asura_Assert((uSuccess == ERROR_SUCCESS) || (uSuccess == ERROR_NO_SUCH_USER) , "Failed to get offline signin info for current user");

	if( uSuccess == ERROR_SUCCESS )
	{
		xOnline = xInfo.xuid;
	}
	else if(uSuccess == ERROR_NO_SUCH_USER)
	{
		xOnline = 0;
	}
	else
	{
		xOffline	= 0;
		xOnline		= 0;
		return false;
	}

	return true;
}


void Asura_Xbox360_Profile_System::StorageDevicesChanged ( )
{
	Asura_Xbox360_Profile_State xCurrentState = GetState();

	// The other states *should* deal with device removal as an error
	if( ( xCurrentState == ASURA_XBOX360_PROFILE_STATE_IDLE		) ||
		( xCurrentState == ASURA_XBOX360_PROFILE_STATE_ERROR	) ||
		( xCurrentState == ASURA_XBOX360_PROFILE_STATE_QUERY	) )
	{
		if( s_iActiveDevice != XCONTENTDEVICE_ANY )
		{
			DWORD dwRet = XContentGetDeviceState( s_iActiveDevice, NULL );

			if (dwRet == ERROR_SUCCESS)
			{
				return;
			}
			// The user removed the active device
			s_iActiveDevice = XCONTENTDEVICE_ANY;
			SetErrorState( ASURA_XBOX360_PROFILE_ERROR_DEVICE_REMOVED );
		}
	}
}

bool Asura_Xbox360_Profile_System::IsDeviceStillValid ()
{
	DWORD dwRet = XContentGetDeviceState( s_iActiveDevice, NULL );

	if (dwRet == ERROR_SUCCESS)
	{
		return true;
	}
	return false;
}



//eof
