//////////////////////////////////////////////////////////////////////////////
//
// Asura_PlatformLogfile.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Debugging/Asura_Logfile.h"

// Asura Platform-independent
#include "Core/Asura_NameTree.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

// Misc
#include <xbdm.h>
#include <xtl.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#if defined ASURA_LOGFILES_ENABLED || !defined ASURA_NO_DEBUG_OUTPUT
	const int iOUTPUT_BUFFER_SIZE = 1024;
#endif

#if defined ASURA_LOGFILES_ENABLED

static const u_int uMAX_OPENED_LOGFILES = 16;

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/
static Asura_Hash_ID g_auOpenedLogfileHashes[uMAX_OPENED_LOGFILES];
static u_int         g_uNumOpenedLogfiles;

static const char* g_szFile = NULL;
static u_int g_uLine = 0;

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

static bool CreateDirectoryRecursive( const char* szFullPath )
{
	if( CreateDirectory( szFullPath, NULL ) )
	{
		return true;
	}

	if( GetLastError() == ERROR_PATH_NOT_FOUND )
	{
		char szParentDir[ MAX_PATH + 1 ];
		Asura_PrintfToBfr( szParentDir, "%s", MAX_PATH, true, szFullPath );

		char* const szSlash = strrchr( szParentDir, '\\' );
		if( szSlash )
		{
			*szSlash = '\0';

			if( !CreateDirectoryRecursive( szParentDir ) ) return false;
			if( CreateDirectory( szFullPath, NULL ) ) return true;
		}
	}

	return false;
}

static bool TryToCreateDirForLogfile( const Asura_Char* szFilename )
{
	static char szDir[ MAX_PATH + 1 ];
	Asura_PrintfToBfr( szDir, "%s", MAX_PATH, true, szFilename );

	char* const szSlash = strrchr( szDir, '\\' );
	if( szSlash )
	{
		*szSlash = '\0';

		return CreateDirectoryRecursive( szDir );
	}

	return false;
}

static bool ClearLogfile( const char* szLogFilename )
{
	// Construct absolute path

	DM_XBE xExeInfo;
	DmGetXbeInfo( NULL, &xExeInfo );

	{
		char* const szLastSlash = strrchr( xExeInfo.LaunchPath, '\\' );
		if( szLastSlash )
		{
			*szLastSlash = '\0';
		}
	}

	static char szPath[ MAX_PATH + 1 ];
	Asura_PrintfToBfr( szPath, "%s\\%s", MAX_PATH, true, xExeInfo.LaunchPath, szLogFilename );

	// Open the file

	HANDLE hFileHandle = CreateFile( szPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL );

	if( hFileHandle == INVALID_HANDLE_VALUE && GetLastError() == ERROR_PATH_NOT_FOUND )
	{
		// Attempt to create directory

		Asura_OutputToDebugger_Retail(( "Could not find existing directory for log file %s. Creating it.\n", szLogFilename ));
		if( TryToCreateDirForLogfile( szPath ) )
		{
			hFileHandle = CreateFile( szPath, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL );
		}
	}

	if( hFileHandle != INVALID_HANDLE_VALUE )
	{
		// Write a simple header

		char szHeader[MAX_PATH];
		const u_int uNumChars = Asura_PrintfToBfr( szHeader, "Asura Logfile \"%s\"\n\n", MAX_PATH, true, szLogFilename );

		DWORD ulNumBytesWritten;
		WriteFile( hFileHandle, szHeader, uNumChars, &ulNumBytesWritten, NULL );

		CloseHandle( hFileHandle );

		return true;
	}

	return false;
}

bool __Asura_Logfile_CanLogYet()
{
	return true;
}

void __Asura_LogfileDebugData( const char* szFile, u_int uLine )
{
	g_szFile = szFile;
	g_uLine = uLine;
}

void __Asura_LogfileOutput( const char* szLogFilename, const char* szFormat, ... )
{
	// Automatically clear this logfile if it hasn't previously been opened.
	{
		const Asura_Hash_ID uHashID = Asura_GetHashID( szLogFilename );

		bool bFound = false;
		for( u_int u = 0; u < g_uNumOpenedLogfiles; ++u )
		{
			if( uHashID == g_auOpenedLogfileHashes[u] )
			{
				bFound = true;
				break;
			}
		}

		if( !bFound )
		{
			if (!ClearLogfile( szLogFilename ))
			{
				Asura_Assertf( false, ("Failed to create log file %s.", szLogFilename) );
				return;
			}

			Asura_Assert( g_uNumOpenedLogfiles < uMAX_OPENED_LOGFILES, "You need to increase uMAX_OPENED_LOGFILES." );
			if (g_uNumOpenedLogfiles < uMAX_OPENED_LOGFILES)
			{
				g_auOpenedLogfileHashes[g_uNumOpenedLogfiles++] = uHashID;
			}
		}
	}

	// Format the text into a buffer

	static char szBuffer[ iOUTPUT_BUFFER_SIZE + 1 ];

	va_list xArgList;
	va_start( xArgList, szFormat );
	const u_int uNumChars = Asura_PrintfToBfr( szBuffer, szFormat, xArgList, iOUTPUT_BUFFER_SIZE, true );
	va_end( xArgList );

	// Output the text

	__Asura_SimpleLogfileOutput_PreFormatted( szLogFilename, szBuffer, uNumChars );
}

void __Asura_SimpleLogfileClear( const char* szLogFilename )
{
	ClearLogfile( szLogFilename );
}

void __Asura_SimpleLogfileOutput( const char* szLogFilename, const char* szFormat, ... )
{
	// Format the text into a buffer

	static char szBuffer[ iOUTPUT_BUFFER_SIZE + 1 ];

	va_list xArgList;
	va_start( xArgList, szFormat );
	const u_int uNumChars = Asura_PrintfToBfr( szBuffer, szFormat, xArgList, iOUTPUT_BUFFER_SIZE, true );
	va_end( xArgList );

	// Output the text

	__Asura_SimpleLogfileOutput_PreFormatted( szLogFilename, szBuffer, uNumChars );
}

void __Asura_SimpleLogfileOutput_PreFormatted( const char* szLogFilename, const char* szStr, u_int uLen )
{
	// Construct absolute path

	DM_XBE xExeInfo;
	DmGetXbeInfo( NULL, &xExeInfo );

	{
		char* const szLastSlash = strrchr( xExeInfo.LaunchPath, '\\' );
		if( szLastSlash )
		{
			*szLastSlash = '\0';
		}
	}

	static char szPath[ MAX_PATH + 1 ];
	Asura_PrintfToBfr( szPath, "%s\\%s", MAX_PATH, true, xExeInfo.LaunchPath, szLogFilename );

	// Open the file
	const HANDLE hFileHandle = CreateFile( szPath, GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, 0, NULL );

	if( hFileHandle != INVALID_HANDLE_VALUE )
	{
		// Append the text

		SetFilePointer( hFileHandle, NULL, NULL, FILE_END );

		DWORD ulNumBytesWritten;
		WriteFile( hFileHandle, szStr, strlen( szStr ), &ulNumBytesWritten, NULL );

		CloseHandle( hFileHandle );
	}
}

bool __Asura_OverrideLogfilesDir( const char* szDir, bool bArchiveExisting )
{
	return false;
}

#endif // defined ASURA_LOGFILES_ENABLED

#if !defined ASURA_NO_DEBUG_OUTPUT

void __Asura_OutputToDebugger( const char* szFormat, ... )
{
	static char szBuffer[ iOUTPUT_BUFFER_SIZE + 1 ];

	va_list xArgList;
	va_start( xArgList, szFormat );
	Asura_PrintfToBfr( szBuffer, szFormat, xArgList, iOUTPUT_BUFFER_SIZE, true );
	va_end( xArgList );

	OutputDebugString( szBuffer );
}

#endif // !defined ASURA_NO_DEBUG_OUTPUT
