
//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_GPUProfile.cpp
//
// Created by: Dave Reed 13/05/07
//
// Copyright (c) 2004-2007 Rebellion. All rights reserved.
//
// Xbox360 GPU profile bars - basically a copy of the CPU profile renderer
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_USE_PROFILING_SYSTEM

#include "Debugging/Asura_Platform_GPUProfile.h"

#include "Rendering/Asura_Xbox360_Graphics.h"

#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_DebugText.h"
#include "Debugging/Asura_Logfile.h"
#include "Debugging/Asura_Profiling_BarRenderer.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

#ifdef ASURA_ASSERT
#include <string.h>
#endif

// G L O B A L S /////////////////////////////////////////////////////////////

// Timers are triple-buffered - we're using InsertCallback for timing, and it's not synchronized with the rendering of the bars
static Asura_TickCount g_auTimerStartTicks[3][ASURA_GPUPROFILE_MAX_TIMERS] = {0};
static Asura_TickCount g_auTimerTicks[3][ASURA_GPUPROFILE_MAX_TIMERS]	   = {0};

// This is the timer buffer we're currently filling in. The previous one should always be ready to render
static uint32  g_uCurrentBuffer = 0;

static const u_int g_auProfileColours[ASURA_GPUPROFILE_MAX_TIMERS] =
{
	0xff604020, 0xff2222ff, 0xff22ff22, 0xffff2222, 0xff22ffff, 0xffffff22, 0xffff22ff, 0xff222222,
	0xffffffff, 0xff5555dd, 0xff55dd55, 0xffdd5555, 0xff55dddd, 0xffdddd55, 0xffdd55dd, 0xff555555,
	0xffdddddd, 0xff2222ff, 0xff22ff22, 0xffff2222, 0xff22ffff, 0xffffff22, 0xffff22ff, 0xff222222,
	0xffffffff, 0xff5555dd, 0xff55dd55, 0xffdd5555, 0xff55dddd, 0xffdddd55, 0xffdd55dd, 0xff555555,
};


static const char* g_aszProfileNames[ASURA_GPUPROFILE_MAX_TIMERS] =
{
	"Total Frame",
	"Skybox",
	"Environment",
	"Skins",
	"Shapes",
	"Hierarchies",
	"Botanicals",
	"Carpet",
	"Decals",
	"Resolve",
	"Copy to EDRAM",
	"Create Shadows",
	"Apply Shadows",
	"Deferred Photons",
	"Projectors",
	"Water",
	"Fog",
	"PFX",
	"Coronas",
	"Lightshafts",
	"FSFX",
	"HUD"
};


#ifdef ASURA_ASSERT
static u_int g_uInProfileDepth = 0;
static u_int g_uInProfile = 0;
#endif // #ifdef ASURA_ASSERT

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool Asura_GPUProfile_Renderer::s_bBarsEnabled = false;
#if ASURA_USE_DEBUG_TEXT
bool Asura_GPUProfile_Renderer::s_bTextEnabled = false;
#endif

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void Asura_GPUProfile_Renderer::Initialise()
{
// JAT 27/02/07 add the variable to command console & DVS for easy toggle of the bars & text on/off
#if defined( ASURA_SUPPORT_COMMAND_CONSOLE )
	Asura_CommandConsole::AddVar( L"Debug.Profile.GPUBars", &s_bBarsEnabled, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#if ASURA_USE_DEBUG_TEXT
	Asura_CommandConsole::AddVar( L"Debug.Profile.GPUText", &s_bTextEnabled, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif // #if ASURA_USE_DEBUG_TEXT
#endif // #if defined( ASURA_USE_COMMANDCONSOLE )

	ClearValues();
}

void Asura_GPUProfile_Renderer::Shutdown()
{
}



void StartCallback(DWORD iProfileType)
{
	g_auTimerStartTicks[g_uCurrentBuffer][iProfileType] = Asura_Timers::GetCurTicks();
}


void StopCallback(DWORD iProfileType)
{	
	Asura_TickCount uEndTicks = Asura_Timers::GetCurTicks();
	Asura_TickCount uElapsedTicks = uEndTicks - g_auTimerStartTicks[g_uCurrentBuffer][iProfileType];
	g_auTimerTicks[g_uCurrentBuffer][iProfileType] += uElapsedTicks;

	if (iProfileType==0)
	{
		// Select next buffer
		g_uCurrentBuffer++;
		if (g_uCurrentBuffer>2)
		{
			g_uCurrentBuffer= 0;
		}

		// Clear next buffer for the next frame
		for (int i=0; i<ASURA_GPUPROFILE_MAX_TIMERS; i++)
		{
			g_auTimerTicks[g_uCurrentBuffer][i] = 0;		
		}	
	}	
}



void Asura_GPUProfile_Renderer::StartProfile(int iProfileType)
{
	if (!(s_bBarsEnabled  
#ifdef ASURA_USE_DEBUG_TEXT
		|| s_bTextEnabled
#endif
		))
	{
		return;
	}

#ifdef ASURA_ASSERT
	Asura_Assert(iProfileType >= 0 && iProfileType < ASURA_GPUPROFILE_MAX_TIMERS, "bad timer index");
	if (iProfileType)
	{
		Asura_Assertf( 0 == g_uInProfileDepth, ("profile depth is already %d. already profiling type %d. Trying to profile %d.", g_uInProfileDepth, g_uInProfile, iProfileType));
		if(0 == g_uInProfileDepth)
		{
			g_uInProfile = iProfileType;
		}
		g_uInProfileDepth += 1;
	}
#endif

	Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_ALL_RENDER | D3DPRED_ALL_Z);
	Asura_Xbox360_Graphics::GetDevice()->InsertCallback(D3DCALLBACK_IDLE, StartCallback, iProfileType);
	Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
}


void Asura_GPUProfile_Renderer::StopProfile(int iProfileType)
{
	if (!(s_bBarsEnabled
#ifdef ASURA_USE_DEBUG_TEXT
		|| s_bTextEnabled
#endif
		))
	{
		return;
	}

#ifdef ASURA_ASSERT
	Asura_Assert(iProfileType >= 0 && iProfileType < ASURA_GPUPROFILE_MAX_TIMERS, "bad timer index");
	if (iProfileType)
	{
		Asura_Assertf( 0 < g_uInProfileDepth, ("profiling of type %d not started so cannot leave.", iProfileType));
		g_uInProfileDepth = (0 < g_uInProfileDepth) ? (g_uInProfileDepth - 1) : 0;
		if(0 == g_uInProfileDepth)
		{
			g_uInProfile = 0;
		}
	}
#endif

	Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_ALL_RENDER | D3DPRED_ALL_Z);
	Asura_Xbox360_Graphics::GetDevice()->InsertCallback(D3DCALLBACK_IDLE, StopCallback, iProfileType);
	Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
}




DEBUGVAR u_int dbg_uProfTextX = 85;
// JAT 27/02/07 try to put the profile text below the DVS render overlay
#if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
DEBUGVAR u_int dbg_uProfTextY = 24;
#else // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )
DEBUGVAR u_int dbg_uProfTextY = 2;
#endif // #if defined( ASURA_DEBUG_VARIABLE_SYSTEM )

void Asura_GPUProfile_Renderer::Render(bool bIncludeText)
{
	if (false == s_bBarsEnabled
#if defined( ASURA_USE_DEBUG_TEXT)
	&&  false == s_bTextEnabled
#endif
	)
	{
		return;
	}

	int iBuffer = g_uCurrentBuffer;
	if (iBuffer==0)	
		iBuffer = 2;
	else
		iBuffer--;

	Asura_Render::FlushDisplay();
    
	if (!g_auTimerTicks[iBuffer][ASURA_GPUPROFILE_TOTALFRAME])
	{
		// We don't have any valid timings yet
		return;
	}

	float fTotalTime = Asura_Timers::TicksToSeconds(g_auTimerTicks[iBuffer][ASURA_GPUPROFILE_TOTALFRAME]);

	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;

	float fRunningTotalTime;

	// Pre-sum the total time (for bars and text _both_ require it)
	if (s_bBarsEnabled
#if ASURA_USE_DEBUG_TEXT
		|| s_bTextEnabled)
#else
		)
#endif
	{
		Asura_TickCount	xRunningTickCount	= 0;

		for (int i = 1; i<ASURA_GPUPROFILE_MAX_TIMERS; i++)
		{
			xRunningTickCount	+= g_auTimerTicks[iBuffer][i];
		}

		fRunningTotalTime	= Asura_Timers::TicksToSeconds(xRunningTickCount);
	}
	else
	{
		fRunningTotalTime	= 0.0f;
	}

	if (s_bBarsEnabled)
	{
		static Asura_Profiling_BarRenderer_Data xData;

		xData.SetUseDefaultBackgroundColours();
		xData.SetBarColours( &g_auProfileColours[1], ASURA_GPUPROFILE_MAX_TIMERS - 1, g_auProfileColours[0] );
		xData.SetYOffset( 0.026f );

		float afTimers[ASURA_GPUPROFILE_MAX_TIMERS];
		for( u_int u = 0; u < ASURA_GPUPROFILE_MAX_TIMERS; ++u )
		{
			afTimers[u]	= Asura_Timers::TicksToSeconds( g_auTimerTicks[iBuffer][u] );
		}

		xData.SetNewFrameData( afTimers[0], &afTimers[1], ASURA_GPUPROFILE_MAX_TIMERS - 1 );
		Asura_Profiling_BarRenderer::Render( xData );
	}


#ifdef ASURA_USE_DEBUG_TEXT
	if (s_bTextEnabled)
	{
		u_int uNumRendered = 0;
		for (int i = 1; i<ASURA_GPUPROFILE_MAX_TIMERS; i++)
		{
			if (g_auTimerTicks[iBuffer][i])
			{
				float fTime = Asura_Timers::TicksToSeconds(g_auTimerTicks[iBuffer][i]);
				if (fTime > 0.0f)
				{
					Asura_Assert(fTime < fTotalTime, "timed event took longer than entire frame?");
					const char* szName = g_aszProfileNames[i] ? g_aszProfileNames[i] : "???";
					Asura_DebugText::AddTextBlock( dbg_uProfTextX, dbg_uProfTextY + uNumRendered, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
						"%2d %5.02f%% %s: (%d us)", i, 100.0f * fTime / fTotalTime, szName, Asura_Maths::FloatToInt(fTime * 1000000.0f));

					// render a colour coded square
					float fX = (static_cast<float>(dbg_uProfTextX - 1) * ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
					float fY = 1.0f + (static_cast<float>(dbg_uProfTextY + uNumRendered) * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
					float fW = ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH - 2.0f;
					float fH = ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT - 2.0f;
					Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
					Asura_Render::Overlay.Colour = g_auProfileColours[i];
					Asura_Render::Overlay.Render();
					
					++uNumRendered;
				}
			}
		}

		{
			// render a colour coded square for the untimed section
			float fX = (static_cast<float>(dbg_uProfTextX - 1) * ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH);
			float fY = 1.0f + (static_cast<float>(dbg_uProfTextY + uNumRendered) * ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT);
			float fW = ASURA_DEBUGTEXT_OUTPUT_FONT_WIDTH - 2.0f;
			float fH = ASURA_DEBUGTEXT_OUTPUT_FONT_HEIGHT - 2.0f;
			Asura_Render::Overlay.SetRectXY(fX, fY, fW, fH);
			Asura_Render::Overlay.Colour = g_auProfileColours[0];
			Asura_Render::Overlay.Render();
			// and the text...
			const float fTime = fTotalTime - fRunningTotalTime;
			Asura_DebugText::AddTextBlock( dbg_uProfTextX, dbg_uProfTextY + uNumRendered, 0xffffffff, ASURA_DEBUGTEXT_FORMAT_LEFTJUSTIFIED, 
				"   %5.02f%% unprofiled: (%d us)", 100.0f * fTime / fTotalTime, Asura_Maths::FloatToInt(fTime * 1000000.0f));
		}
	}
#endif // #ifdef ASURA_USE_DEBUG_TEXT

	Asura_Render::FlushDisplay();
}

void Asura_GPUProfile_Renderer::ClearValues()
{
	for (int i=0; i<ASURA_GPUPROFILE_MAX_TIMERS; i++)
	{
		for(int j=0; j<3; j++)
		{				
			g_auTimerTicks[j][i] = 0;
		}
	}
}

#endif //#ifdef ASURA_USE_PROFILING_SYSTEM
