//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_FileAccess.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "FileAccess/Asura_FileAccess.h"

// Asura Platform-independent
#include "Core/Asura_List.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

// Asura 360
#include "Core/Asura_Xbox360_OS.h"
#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
#include "Network/Asura_Xbox360_DLC_System.h"
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT

// Misc
#include <xtl.h>
#include <xfilecache.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Uncomment this to verify all reads from the file cache with another uncached
// read from the file, opened at the same time
//#define VERIFY_CACHE

#if !defined ASURA_XBOX360_FILE_CACHE_SIZE
	// Doesn't seem to be any benefit in going above 128 KB. For reads larger
	// that the kernel splits up requests anyway. Must be a multiple of 2048
	// (360 DVD sector size) if we use unbuffered I/O
	#define ASURA_XBOX360_FILE_CACHE_SIZE (128 * 1024)
#endif

ASURA_STATIC_ASSERT( (ASURA_XBOX360_FILE_CACHE_SIZE & 2047) == 0 );

/******************************************************************************/
/* Static Functions                                                           */
/******************************************************************************/

static void CheckShowDirtyDiscDialog (const char*pFileName)
{
	DWORD uError = GetLastError();
	
	switch(uError)
	{
		//TY 03.04.08 - These are the error codes we ignore when deciding if we
		//should show the dirty disc dialog. An alternative would be to check
		//explicitly for ERROR_DISK_CORRUPT
		case ERROR_SUCCESS:	
		case ERROR_INVALID_USER_BUFFER:
		case ERROR_NOT_ENOUGH_MEMORY:
		case ERROR_FILE_NOT_FOUND:
		case ERROR_PATH_NOT_FOUND:
		case ERROR_INVALID_DRIVE:
		case ERROR_WRITE_PROTECT:
		case ERROR_HANDLE_EOF:

			return;

		case ERROR_INVALID_NAME:

			Asura_Assertf( false, ("Badly-constructed path (check for double slashes etc.): '%s'",
				pFileName) );

			return;
	}
#ifdef ASURA_ASSERT
	Asura_Assertf(0, ("File '%s' failed to open or read.\n in a release build this will trigger the dirty disc UI and the game will exit.", pFileName));
#else
	XShowDirtyDiscErrorUI(Asura_Xbox360_OS::GetActivePlayerIndex());
#endif

}

static void CreateDriveList(Asura_List<Asura_String>& xOutputList)
{
	xOutputList.AddToList(new Asura_String("GAME:\\"));

#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	Asura_Xbox360_DLC_System::FillDriveList(xOutputList);
#endif //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
}

static void FindDirectories( const char* szSearchString, Asura_List<Asura_String>& xOutputList )
{
	Asura_Assert( szSearchString && *szSearchString, "" );

	Asura_List_ReverseIt<Asura_String> xIt( &xOutputList );
	//xIt.end();

	char szBuffer[MAX_PATH];
	WIN32_FIND_DATA xFindData;

	const int iInitialNumEntries = xOutputList.GetNumEntries();
	for( int iDir = 0; iDir < 1 + (xOutputList.GetNumEntries() - iInitialNumEntries); ++iDir )
	{
		char* szFilenameStart;

		//copy the search string into a buffer for buidling the filenames , 
		//and work out where the filename part should be copied
		{
			if( iDir )
			{
				if(iDir==1 && iInitialNumEntries==0)
				{
					//If there were no entries in the list to start with , then the iterator will currently be
					//on the sentinel , so we need to restart to get to the first of the newly added entries.
					//If not , then the iterator will be on the last of the original entries , so using next()
					//will get us to the first newly added entry.RAR 17/12/03
					xIt.Restart();
				}
				else
				{
					xIt.Next();
				}

				const Asura_String& xNext = *xIt.GetCurrent();
				strcpy( szBuffer, xNext.GetString() );
				strcpy( &szBuffer[xNext.GetLength()], "\\*.*" );

			}
			else
			{
				strcpy( szBuffer, szSearchString );
			}

			char* szSlash = strrchr( szBuffer, '\\' );
			if( !szSlash )
			{
				szSlash = strrchr( szBuffer, '/' );
			}
			szFilenameStart = ( szSlash ? szSlash + 1 : szBuffer );
		}

		HANDLE hFindFile = ::FindFirstFile( szBuffer, &xFindData );
		if( INVALID_HANDLE_VALUE != hFindFile )
		{
			do
			{	
				if( xFindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY )
				{

					// Ignore the "." and ".." directories.
					if( '.' == xFindData.cFileName[0] )
					{
						if( !xFindData.cFileName[1] || ('.' == xFindData.cFileName[1] && !xFindData.cFileName[2]) )
						{
							continue;
						}
					}

					strcpy( szFilenameStart, xFindData.cFileName );
					xOutputList.AddToList( new Asura_String(szBuffer) );
				}
			}
			while( ::FindNextFile(hFindFile, &xFindData) );

			::FindClose( hFindFile );
		}
	}
}

/******************************************************************************/
/* Classes                                                                    */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// OpenFileInfo
//
// Holds information about an open file on 360
//
////////////////////////////////////////////////////////////////////////////////

struct OpenFileInfo
{
	HANDLE m_xWin32Handle;
	u_int m_uSize;
	u_int m_uFilePosition;
	u_int m_uCachePosition;
	char m_acCache[ ASURA_XBOX360_FILE_CACHE_SIZE ];

#if defined VERIFY_CACHE
	HANDLE m_xWin32Handle2;
#endif
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static OpenFileInfo g_axOpenFiles[ ASURA_MAX_NO_OF_OPEN_FILES ];

#if defined VERIFY_CACHE
	static char g_acVerificationBuffer[ ASURA_XBOX360_FILE_CACHE_SIZE ];
#endif

/******************************************************************************/
/* Asura_FileAccess Static Class Functions                                    */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_Initialise()
//
// Initialises file access on 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_FileAccess::Platform_Initialise()
{
//	XFileCacheInit(XFILECACHE_CLEAR_ALL, 0, XFILECACHE_DEFAULT_THREAD, 0, 0x058CF3F2);	// Asura_GetHashID("Asura");
	XFileEnableTransparentDecompression(NULL);

	for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
	{
		g_axOpenFiles[u].m_xWin32Handle = INVALID_HANDLE_VALUE;

		#if defined VERIFY_CACHE
			g_axOpenFiles[u].m_xWin32Handle2 = INVALID_HANDLE_VALUE;
		#endif
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_Shutdown()
//
// Shuts down file access on 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_FileAccess::Platform_Shutdown()
{
	for( u_int u = 0; u < ASURA_MAX_NO_OF_OPEN_FILES; ++u )
	{
		Asura_Assertf( INVALID_HANDLE_VALUE == g_axOpenFiles[u].m_xWin32Handle, ("File with handle %u wasn't closed.", u) );

		#if defined VERIFY_CACHE
			Asura_Assertf( INVALID_HANDLE_VALUE == g_axOpenFiles[u].m_xWin32Handle2, ("File with handle %u wasn't closed.", u) );
		#endif
	}

	XFileDisableTransparentDecompression();
//	XFileCacheShutdown();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_FileExists()
//
// Determines whether a file exists
//
// Parameters:		szFilename - Filename to check, without "drive:\"
// Return value:	true iff the file exists
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_FileExists( const char* szFilename )
{
	Asura_Assertf(!strstr(szFilename, "/"), ("Filename %s contains forward slash ('/') and will not be found; use backslash instead", szFilename));
	char szFullPath[ MAX_PATH ];
	Asura_PrintfToBfr( szFullPath, "GAME:\\%s", MAX_PATH, true, szFilename );

	return GetFileAttributes( szFullPath ) != -1;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_GetFileSize()
//
// Queries a file's size by filename
//
// Parameters:		szFilename - Filename to check, without "drive:\"
// Return value:	File size in bytes, or 0 for failure
//
////////////////////////////////////////////////////////////////////////////////

u_int Asura_FileAccess::Platform_GetFileSize( const char* szFilename )
{
	Asura_Assertf(!strstr(szFilename, "/"), ("Filename %s contains forward slash ('/') and will not be found; use backslash instead", szFilename));

	char szFullPath[ MAX_PATH ];
	Asura_PrintfToBfr( szFullPath, "GAME:\\%s", MAX_PATH, true, szFilename );

	const HANDLE xWin32Handle = CreateFile( szFullPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL );
	if( xWin32Handle != INVALID_HANDLE_VALUE )
	{
		const DWORD uFileSize = ::GetFileSize( xWin32Handle, NULL );
		CloseHandle( xWin32Handle );

		return uFileSize;
	}
	else
	{
		return 0;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_GetFileSize()
//
// Queries a file's size by Asura handle
//
// Parameters:		iFileHandle - Asura handle to open file
// Return value:	File size in bytes, or 0 for failure
//
////////////////////////////////////////////////////////////////////////////////

u_int Asura_FileAccess::Platform_GetFileSize( const Asura_Handle iFileHandle )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid file handle (%d) passed to Asura_FileAccess::Platform_Open().", iFileHandle) );
		return false;
	}

	return g_axOpenFiles[ iFileHandle ].m_uSize;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_Open()
//
// Opens a file for reading
//
// Parameters:		iFileHandle - Asura handle to use for this file
//					szFilename - Filename without "game:\\"
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_Open( Asura_Handle iFileHandle, const char* szFilename )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid file handle (%d) passed to Asura_FileAccess::Platform_Open().", iFileHandle) );
		return false;
	}

	Asura_Assertf(!strstr(szFilename, "/"), ("Filename %s contains forward slash ('/') and will not be found; use backslash instead", szFilename));

	HANDLE& xWin32Handle = g_axOpenFiles[ iFileHandle ].m_xWin32Handle;
	Asura_Assertf( xWin32Handle == INVALID_HANDLE_VALUE, ("File handle %d is already open.", iFileHandle) );

	char szFullPath[ MAX_PATH ];
#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	Asura_Xbox360_DLC_System::FillFullPath(szFilename, szFullPath); // Downloaded files can be opened in the same way as normal files, they just need a different absolute path.
#else //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT
	Asura_PrintfToBfr( szFullPath, "GAME:\\%s", MAX_PATH, true, szFilename );
#endif //#else //#ifdef ASURA_SUPPORT_DOWNLOADABLE_CONTENT

	// We'll almost always use sequential reads
	xWin32Handle = CreateFile( szFullPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN /*| FILE_FLAG_NO_BUFFERING*/, NULL );
	if( xWin32Handle == INVALID_HANDLE_VALUE ) 
	{
		CheckShowDirtyDiscDialog(szFullPath);
		return false;
	}

	g_axOpenFiles[ iFileHandle ].m_uSize = ::GetFileSize( xWin32Handle, NULL );

	// Prime the cache
	{
		void* const pCache = g_axOpenFiles[ iFileHandle ].m_acCache;

		DWORD uNumBytesRead;
		if( !ReadFile( xWin32Handle, pCache, ASURA_XBOX360_FILE_CACHE_SIZE, &uNumBytesRead, NULL ) )
		{
			CheckShowDirtyDiscDialog(szFullPath);
			Asura_Assertf( false, ("Failed to read from file '%s'.", szFilename) );
		}

		g_axOpenFiles[ iFileHandle ].m_uCachePosition = 0;
		g_axOpenFiles[ iFileHandle ].m_uFilePosition = 0;
	}

	#if defined VERIFY_CACHE
		g_axOpenFiles[ iFileHandle ].m_xWin32Handle2 = CreateFile( szFullPath, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL );
		if( g_axOpenFiles[ iFileHandle ].m_xWin32Handle2 == INVALID_HANDLE_VALUE )
		{
			Asura_Assertf( false, ("Failed to open file '%s'.", szFilename) );
		}
	#endif // defined VERIFY_CACHE

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_OpenFromMemory()
//
// Opens a file in memory for reading. Not implemented on 360
//
// Parameters:		iFileHandle - Asura handle to use for this file
//					pMem - Pointer to file contents in memory
//					uSize - File size
//
// Return value:	Always false
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_OpenFromMemory( Asura_Handle iFileHandle, const void* pMem, u_int uSize )
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_LoadAsynchronous()
//
// Opens a file for asynchronous loading. Not implemented on 360, since we're
// probably better off using multithreading instead
//
// Parameters:		iFileHandle - Asura handle to use for this file
//					szFilename - Filename to load
//
// Return value:	Always false
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_LoadAsynchronous( Asura_Handle iFileHandle, const char* szFilename )
{
	return false;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_Close()
//
// Closes an open file handle
//
// Parameters:		iFileHandle - Asura handle of the file to close
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_Close( Asura_Handle iFileHandle )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid file handle (%d) passed to Asura_FileAccess::Platform_Close().", iFileHandle) );
		return false;
	}

	HANDLE& xWin32Handle = g_axOpenFiles[ iFileHandle ].m_xWin32Handle;

	if( CloseHandle( xWin32Handle ) )
	{
		xWin32Handle = INVALID_HANDLE_VALUE;

		#if defined VERIFY_CACHE
			if( !CloseHandle( g_axOpenFiles[ iFileHandle ].m_xWin32Handle2 ) )
			{
				Asura_Assertf( false, ("Failed to close verification handle for file %d.", iFileHandle) );
			}
		#endif

		return true;
	}
	else
	{
		return false;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_ReadData()
//
// Reads data from an open file at the current read position, and increments it
//
// Parameters:		iFileHandle - Asura handle of file from which to read
//					pDest - Destination memory buffer
//					uNumBytes - Number of bytes to read
//
// Return value:	true iff successfully read the correct number of bytes
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_ReadData( Asura_Handle iFileHandle, void* pDest, u_int uNumBytes )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid file handle (%d) passed to Asura_FileAccess::Platform_ReadData().", iFileHandle) );
		return false;
	}

	const HANDLE xWin32Handle = g_axOpenFiles[ iFileHandle ].m_xWin32Handle;
	char* const pcCache = g_axOpenFiles[ iFileHandle ].m_acCache;
	u_int& uCachePos = g_axOpenFiles[ iFileHandle ].m_uCachePosition;
	u_int& uFilePos = g_axOpenFiles[ iFileHandle ].m_uFilePosition;

	u_int uCacheOffset = uFilePos - uCachePos;
	int iBytesLeftInFile = g_axOpenFiles[ iFileHandle ].m_uSize - uFilePos;

	if( uFilePos < uCachePos )
	{
		// No point supporting reads that start before the cached portion
		Asura_Assert( false, "Read pointer is before the cached file portion." );
		return false;
	}

	char* pcDest = static_cast< char* >( pDest );

	int iCachedBytesToRead = Asura_Maths::Min( uNumBytes, ASURA_XBOX360_FILE_CACHE_SIZE - uCacheOffset );
	iCachedBytesToRead = Asura_Maths::Min( iCachedBytesToRead, iBytesLeftInFile );
	if( iCachedBytesToRead > 0 )
	{
		memcpy( pDest, pcCache + uCacheOffset, iCachedBytesToRead );

		#if defined VERIFY_CACHE

			DWORD uNumBytesReadTemp;

			#if defined ASURA_ASSERT
				const BOOL bVerifyRead =
			#endif
				ReadFile( g_axOpenFiles[ iFileHandle ].m_xWin32Handle2, g_acVerificationBuffer, iCachedBytesToRead, &uNumBytesReadTemp, NULL );

			Asura_Assert( bVerifyRead, "Failed to read for verification." );

			for( int i = 0; i < iCachedBytesToRead; ++i )
			{
				Asura_Assertf( pcDest[i] == g_acVerificationBuffer[i], ("File cache verification failed at entry %d.", i) );
			}

		#endif

		uNumBytes -= iCachedBytesToRead;
		uFilePos += iCachedBytesToRead;
		uCacheOffset += iCachedBytesToRead;
		pcDest += iCachedBytesToRead;
		iBytesLeftInFile -= iCachedBytesToRead;
	}
	else
	{
		// We should always be able to read at least one byte from the cache. Either
		// something's gone wrong, or we're past the end of the file. Can't assert
		// in case of the latter
		return false;
	}

	// Exhausted the cache. Grab some more from the file
	DWORD uNumBytesRead;

	while( uNumBytes >= ASURA_XBOX360_FILE_CACHE_SIZE )
	{
		// More than one cache's worth of data to read. We'll still need to go through the cache
		// though, because we need to write to DWORD-aligned addresses, and we can't guarantee
		// that from the user's arbitrary location

		Asura_Assert( (reinterpret_cast< u_int >( pcCache ) & 3) == 0, "Read address isn't DWORD-aligned." );

		const BOOL bSuccess = ReadFile( xWin32Handle, pcCache, ASURA_XBOX360_FILE_CACHE_SIZE, &uNumBytesRead, NULL );

		if( bSuccess )
		{
			uNumBytes -= ASURA_XBOX360_FILE_CACHE_SIZE;
			uCachePos += ASURA_XBOX360_FILE_CACHE_SIZE;
			uFilePos += ASURA_XBOX360_FILE_CACHE_SIZE;

			iBytesLeftInFile -= ASURA_XBOX360_FILE_CACHE_SIZE;
			Asura_Assertf( iBytesLeftInFile >= 0, ("Read past the end of file %d.", iFileHandle) );

			memcpy( pcDest, pcCache, ASURA_XBOX360_FILE_CACHE_SIZE );

			#if defined VERIFY_CACHE

				DWORD uNumBytesReadTemp;

				#if defined ASURA_ASSERT
					const BOOL bVerifyRead =
				#endif
					ReadFile( g_axOpenFiles[ iFileHandle ].m_xWin32Handle2, g_acVerificationBuffer, ASURA_XBOX360_FILE_CACHE_SIZE, &uNumBytesReadTemp, NULL );

				Asura_Assert( bVerifyRead, "Failed to read for verification." );

				for( u_int u = 0; u < ASURA_XBOX360_FILE_CACHE_SIZE; ++u )
				{
					Asura_Assertf( pcCache[u] == g_acVerificationBuffer[u], ("File cache verification failed at entry %u.", u) );
				}

			#endif // defined VERIFY_CACHE

			pcDest += ASURA_XBOX360_FILE_CACHE_SIZE;
		}
		else
		{
			CheckShowDirtyDiscDialog("Unknown");
			Asura_Assertf( false, ("Failed to read from file with handle %d.\n", iFileHandle) );
			return false;
		}
	}

	if( uFilePos == uCachePos + ASURA_XBOX360_FILE_CACHE_SIZE )
	{
		// Time to grab more from the file into the cache

		const BOOL bSuccess = ReadFile( xWin32Handle, pcCache, ASURA_XBOX360_FILE_CACHE_SIZE, &uNumBytesRead, NULL );
		uCachePos += ASURA_XBOX360_FILE_CACHE_SIZE;

		if( !bSuccess )
		{
			Asura_Assertf( false, ("Failed to read from file with handle %d.", iFileHandle) );
			return false;
		}
	}

	if( uNumBytes > 0 )
	{
		memcpy( pcDest, pcCache, uNumBytes );
		uFilePos += uNumBytes;

		#if defined VERIFY_CACHE

			DWORD uNumBytesTemp;

			#if defined ASURA_ASSERT
				const BOOL bVerifyRead =
			#endif
				ReadFile( g_axOpenFiles[ iFileHandle ].m_xWin32Handle2, g_acVerificationBuffer, uNumBytes, &uNumBytesTemp, NULL );

			Asura_Assert( bVerifyRead, "Failed to read file for verification." );

			for( u_int u = 0; u < uNumBytes; ++u )
			{
				Asura_Assertf( pcDest[u] == g_acVerificationBuffer[u], ("File cache verification failed at entry %u.", u) );
			}

		#endif // defined VERIFY_CACHE
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_SkipUncompressedData()
//
// Skips on a number of bytes from the current position
//
// Parameters:		iFileHandle - Asura handle of file in which to seek
//					uNumBytes - Number of byte to move forward
//
// Return value:	true iff successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_SkipUncompressedData( Asura_Handle iFileHandle, u_int uNumBytes )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid file handle (%d) passed to Seek().", iFileHandle) );
		return false;
	}
	return Platform_SeekFile(iFileHandle, g_axOpenFiles[ iFileHandle ].m_uFilePosition + uNumBytes);
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_SeekFile()
//
// Sets the file pointer to the given position. Not implemented on 360
//
// Parameters:		iFileHandle - Asura handle of file in which to seek
//					uPosition - Absolute byte position for new file pointer
//
// Return value:	true if successful
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_SeekFile( Asura_Handle iFileHandle, u_int uPosition )
{
	if( iFileHandle < 0 || iFileHandle >= ASURA_MAX_NO_OF_OPEN_FILES )
	{
		Asura_Assertf( false, ("Invalid file handle (%d) passed to Seek().", iFileHandle) );
		return false;
	}

	#if defined VERIFY_CACHE
		#if defined ASURA_ASSERT
			const DWORD uResult =
		#endif
			SetFilePointer( g_axOpenFiles[ iFileHandle ].m_xWin32Handle2, uPosition, NULL, FILE_BEGIN );

		Asura_Assert( uResult != INVALID_SET_FILE_POINTER, "Failed to seek file for verification.");
	#endif

	const HANDLE xWin32Handle = g_axOpenFiles[ iFileHandle ].m_xWin32Handle;
	void* const pCache = g_axOpenFiles[ iFileHandle ].m_acCache;
	u_int& uCachePos = g_axOpenFiles[ iFileHandle ].m_uCachePosition;

	g_axOpenFiles[ iFileHandle ].m_uFilePosition = uPosition;

	// Don't need to do anything at all if we're within the same cache block
	if( uPosition < uCachePos || (uPosition - uCachePos) >= ASURA_XBOX360_FILE_CACHE_SIZE ) 
	{
		const u_int uNewCachePos = (uPosition / ASURA_XBOX360_FILE_CACHE_SIZE) * ASURA_XBOX360_FILE_CACHE_SIZE;

		// Need to move the file pointer on if we're more than one cache block apart (after previous
		// reads, the file pointer is left at the start of the next cache block)
		if( uNewCachePos < uCachePos || uNewCachePos > uCachePos + ASURA_XBOX360_FILE_CACHE_SIZE )
		{
			const DWORD uSFPResult = SetFilePointer( xWin32Handle, uNewCachePos, NULL, FILE_BEGIN );
			if( uSFPResult == INVALID_SET_FILE_POINTER )
			{
				Asura_Assertf( false, ("Failed to set file pointer for file %d.", iFileHandle) );
				return false;
			}
		}
		uCachePos = uNewCachePos;

		DWORD uNumBytesRead;
		const BOOL bSuccess = ReadFile( xWin32Handle, pCache, ASURA_XBOX360_FILE_CACHE_SIZE, &uNumBytesRead, NULL );
		if( !bSuccess )
		{
			Asura_Assertf( false, ("Failed to read from file %d.", iFileHandle) );
			return false;
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_GetAsynchronousLoadingStatus()
//
// Retrieves async load status for a file. Not implemented on 360
//
// Parameters:		iFileHandle - Asura handle of file to query
// Return value:	Always ASURA_ASYNCLOADING_STATUS_ERROR
//
////////////////////////////////////////////////////////////////////////////////

ASURA_ASYNCLOADING_STATUS Asura_FileAccess::Platform_GetAsynchronousLoadingStatus( Asura_Handle iFileHandle )
{
	return ASURA_ASYNCLOADING_STATUS_ERROR;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_DiscExists()
//
// Queries whether the disc is currently in the tray. Does nothing on 360, since
// as soon as the disc is removed the game is terminated anyway
//
// Parameters:		None
// Return value:	Always true
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_FileAccess::Platform_DiscExists()
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir()
//
// On PC, this strips out everything up to and including the executable
// directory. On consoles we just return the original path unmodified
//
////////////////////////////////////////////////////////////////////////////////

const char* Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir( const char* szPath )
{
	return szPath;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_FileAccess::Platform_FindFiles()
//
// Iterates over files in a directory, using a callback for each one
//
////////////////////////////////////////////////////////////////////////////////

void Asura_FileAccess::Platform_FindFiles(const char* szDirectory, const char* szExtension, void (*pfnCallback)(const char* szFilename), bool bRecursive /* = true */)
{
	// Check the strings have the right formatting. Could insert the slash or dot manually here, but
	// that means more Asura_String work which we might as well avoid
	{
		#if defined ASURA_ASSERT
			bool bSlash = Asura_StringUtil::EndsWith(szDirectory, "\\", false);
			bSlash = bSlash || Asura_StringUtil::EndsWith(szDirectory, "/", false);

			const bool bDot = (szExtension[0] == '.');
		#endif

		Asura_Assertf(bSlash, ("\"%s\" must end with a slash.", szDirectory));
		Asura_Assertf(bDot, ("\"%s\" must start with a dot.", szExtension));
	}

	char acWildcard[MAX_PATH];
	Asura_PrintfToBfr(acWildcard, "*%s", MAX_PATH, true, szExtension);

	Asura_List<Asura_String> xDriveList;
	CreateDriveList(xDriveList);

	Asura_List<Asura_String> xDirList;

	for (Asura_List_It<Asura_String> xIt(xDriveList); !xIt.Done();)
	{
		Asura_String& xCurrent = *(xIt.GetCurrent());
		xCurrent += szDirectory;
		xCurrent += acWildcard;
		xCurrent.ReplaceAllChars('/', '\\');	// Backslash only
		const char* szFullWildCard = xCurrent.GetString();

		const char* szSlash = strrchr( szFullWildCard, '\\' );

		{
			char szTempBuffer[MAX_PATH];

			if( szSlash && szSlash > szFullWildCard )
			{
				const int iLen = szSlash - szFullWildCard;
				strncpy( szTempBuffer, szFullWildCard, iLen );
				szTempBuffer[iLen] = '\0';
				xDirList.AddToList( new Asura_String(szTempBuffer) );
			}
			else
			{
				xDirList.AddToList( new Asura_String(szFullWildCard) );
			}

			if( bRecursive )
			{
				const Asura_String& xRoot = *xDirList.GetHeadObject();
				strcpy( szTempBuffer, xRoot.GetString() );
				strcpy( &szTempBuffer[xRoot.GetLength()], "\\*.*" );
				FindDirectories( szTempBuffer, xDirList );
			}
		}

		delete xIt.GetCurrent();
		xIt.DeleteCurrent();
	}

	char szBuffer[MAX_PATH];

	for( Asura_List_It<Asura_String> xIt(&xDirList); !xIt.Done(); xIt.Next() )
	{
		char* szFilenameStart;

		//copy the search string into a buffer for buidling the filenames , 
		//and work out where the filename part should be copied
		{
			const Asura_String& xNext = *xIt.GetCurrent();
			const int iLen = xNext.GetLength();

			strcpy( szBuffer, xNext.GetString() );
			szBuffer[iLen] = '\\';
			strcpy( &szBuffer[iLen + 1], acWildcard );

			szFilenameStart = &szBuffer[iLen + 1];
		}

		WIN32_FIND_DATA xFindData;
		HANDLE hFindFile = ::FindFirstFile( szBuffer, &xFindData );
		if( INVALID_HANDLE_VALUE != hFindFile )
		{
			do
			{	
				if( !(xFindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) )
				{
					// (MP - 17.03.2006) Make sure this really does match
					// the wildcard. The file search stuff in Windows does
					// some slightly odd stuff in specific cases. For example,
					// *.asr will match *.asrBE, though bizarrely
					// *.a will not match *.asr. I think it's related to
					// 3 letter extensions somewhow.
					if( Asura_StringUtil::CompareAgainstWildcardString(xFindData.cFileName, acWildcard, true) )
					{
						strcpy( szFilenameStart, xFindData.cFileName );

						// Skip past the drive prefix
						pfnCallback(strstr(szBuffer, ":\\") + 2);
					}
				}
			}
			while( ::FindNextFile(hFindFile, &xFindData) );

			::FindClose( hFindFile );
		}
	}
	//delete strings from xDirList
	for( Asura_List_It<Asura_String> xIt(&xDirList); !xIt.Done(); xIt.Next() )
	{
		delete (xIt.GetCurrent());
	}
}
