//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Leaderboards.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     Xbox 360 support for leaderboards interface.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_LEADERBOARDS

#include "Network/Online/Asura_Leaderboards.h"

// Asura 360
#include "Network/Online/Asura_Xbox360_Friends.h"
#include "Network/Online/Asura_Xbox360_Stats.h"


void Asura_Leaderboard_System::Platform_Initialise()
{
	// nothing to do yet
}

void Asura_Leaderboard_System::Platform_Shutdown()
{
	// nothing to do yet
}

void Asura_Leaderboard::Platform_Initialise()
{
	m_pPlatformData = new Asura_Xbox360_StatsReader();
}

void Asura_Leaderboard::Platform_Shutdown()
{
	Asura_Xbox360_StatsReader* pxStatsReader = reinterpret_cast
		<Asura_Xbox360_StatsReader*>(m_pPlatformData);
	delete pxStatsReader;
}

bool Asura_Leaderboard::Platform_FetchData()
{
	Asura_Xbox360_StatsReader* pxStatsReader = reinterpret_cast
		<Asura_Xbox360_StatsReader*>(m_pPlatformData);

	Asura_Assert(pxStatsReader, "no stats reader!");

	const u_int uLeaderboard = m_uLeaderboardIndex;

	// since the "rated by" stat is automatically fetched, we don't need to request it.
	// Therefore column 0 should always be 0 and we skip it from the stats spec.
	// xStatsSpec must not be on the stack when it's used for FetchStatsForXuids().
	static XUSER_STATS_SPEC xStatsSpec;
	xStatsSpec.dwViewId = Asura_Xbox360_Stats::Project_GetLeaderboardID(uLeaderboard);
	const u_int uNumColumns = Asura_Leaderboard_System::Project_GetNumLeaderboardColumns(uLeaderboard);
	Asura_Assert(uNumColumns && uNumColumns < XUSER_STATS_ATTRS_IN_SPEC, "too many columns");
	xStatsSpec.dwNumColumnIds = uNumColumns - 1;
	Asura_Assert(Asura_Xbox360_Stats::Project_GetColumnID(uLeaderboard, 0) == 0, "expecting 0 for column 0 since that's filled in by rated stat.");
	for (u_int uCol = 1; uCol < uNumColumns; ++uCol)
	{
		xStatsSpec.rgwColumnIds[uCol-1] = Asura_Xbox360_Stats::Project_GetColumnID(uLeaderboard, uCol);
	}

	bool bRet;
	if (m_bFriendsOnly)
	{
		// buffers must remain on the stack
		bRet = pxStatsReader->FetchStatsForXuids(Asura_Xbox360_Friends::GetFriendsListIncludingSelf(), 
			Asura_Xbox360_Friends::GetNumFriendsIncludingSelf(), xStatsSpec);
	}
	else
	{
		bRet = pxStatsReader->FetchStats(m_uFirstRow, uASURA_LEADERBOARD_MAX_ROWS_IN_MEMORY, xStatsSpec);
	}
	return bRet;
}

bool Asura_Leaderboard::Platform_Update()
{
	Asura_Xbox360_StatsReader* pxStatsReader = reinterpret_cast
		<Asura_Xbox360_StatsReader*>(m_pPlatformData);

	Asura_Assert(pxStatsReader, "no stats reader!");
	
	if (pxStatsReader->Update())
	{
		if (pxStatsReader->HasErrorOccurred())
		{
			m_bError = true;
			return true;
		}

		// copy the stats into the independent arrays
		const XUSER_STATS_VIEW* pxView = pxStatsReader->GetData();

		u_int uNumEntriesProcessed = 0;
		// fill these in now - for friends LBs they get changed later
		m_uTotalEntries = pxView->dwTotalViewRows;
		m_uNumRows = pxView->dwNumRows;
		const XUSER_STATS_ROW* pxRow = pxView->pRows;
		for (u_int uRow = 0; uRow < pxView->dwNumRows; ++uRow)
		{
			// check if this row contains valid data. if not move on to the next one without recording this row.
			// This is needed for friends leaderboards where you're asking for stats on gamers who might not
			// have any valid stats.
			if (!pxRow->dwNumColumns || (pxRow->pColumns[0].Value.type == XUSER_DATA_TYPE_NULL))
			{
				Asura_Assert(m_bFriendsOnly, "only expect this to happen with friends-only leaderboards");
				++pxRow;
				continue;
			}

			SetASCIINameAndRank(uNumEntriesProcessed, pxRow->szGamertag, pxRow->dwRank);

			// column 0 is the rated stat
			float fData = static_cast<float>(pxRow->i64Rating);
			SetData(uNumEntriesProcessed, 0, fData);

			const XUSER_STATS_COLUMN* pxCol = pxRow->pColumns;
			for (u_int uCol = 0; uCol < pxRow->dwNumColumns; ++uCol)
			{
				Asura_Assert(pxCol->wColumnId == Asura_Xbox360_Stats::Project_GetColumnID(
					m_uLeaderboardIndex, uCol + 1), "unexpected column ID");

				float fData;
				// convert the value to float.
				switch (pxCol->Value.type)
				{
				case XUSER_DATA_TYPE_UNICODE:
				case XUSER_DATA_TYPE_BINARY:
				case XUSER_DATA_TYPE_DATETIME:
				default:
					Asura_Assert(false, "unknown/unhandled data type");
					// fall through
				case XUSER_DATA_TYPE_CONTEXT:
				case XUSER_DATA_TYPE_INT32:
					fData = static_cast<float>(pxCol->Value.nData);
					break;
				case XUSER_DATA_TYPE_INT64:
					fData = static_cast<float>(pxCol->Value.i64Data);
					break;
				case XUSER_DATA_TYPE_DOUBLE:
					fData = static_cast<float>(pxCol->Value.dblData);
					break;
				case XUSER_DATA_TYPE_FLOAT:
					fData = pxCol->Value.fData;
					break;
				}

				// colulmns 1 to n are the unrated columns
				SetData(uNumEntriesProcessed, uCol + 1, fData);
				++pxCol;
			}

			++uNumEntriesProcessed;
			++pxRow;
		}

		if (m_bFriendsOnly)
		{
			m_uTotalEntries = uNumEntriesProcessed;
			m_uNumRows = uNumEntriesProcessed;
			SortData();
		}
		else
		{
			Asura_Assert(uNumEntriesProcessed == m_uNumRows, "didn't process some rows");
		}

		// free the stats reader data ready for next time
		pxStatsReader->FreeData();
		return true;
	}
	return false;
}

#endif // #ifdef ASURA_SUPPORT_LEADERBOARDS
