//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MatchFinder.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     class to manage searching for Xbox Live Sessions
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Network/Online/Asura_Xbox360_MatchFinder.h"

// Asura 360 includes
#include "Core/Asura_Xbox360_OS.h"
// do not include Asura_Lobby.h etc.

static const u_int uASURA_XBOX360_NUM_REQUIRED_CONTEXTS = 2; // X_CONTEXT_GAME_TYPE and X_CONTEXT_GAME_MODE

Asura_Xbox360_MatchFinder::Asura_Xbox360_MatchFinder( u_int uMaxResults, u_int uMaxContexts, bool bRankedGames )
{
	m_uMaxResults = uMaxResults;
	m_uMaxContexts = uMaxContexts + uASURA_XBOX360_NUM_REQUIRED_CONTEXTS;
	m_pxContexts = new XUSER_CONTEXT[m_uMaxContexts];
	m_uNumContexts = 0;
	m_uMaxProperties = uMaxContexts;
	m_pxProperties = new XUSER_PROPERTY[m_uMaxProperties];
	m_uNumProperties = 0;
	ZeroMemory(&m_xSearchOverlapped, sizeof(m_xSearchOverlapped));
	m_pxSearchResults = NULL;
	m_bRankedGames = bRankedGames;
	m_bQoSTesting = false;
	ZeroMemory(&m_xQoSResult, sizeof(m_xQoSResult));
	m_pxQoSResult = &m_xQoSResult;
	m_ppxQoSxnaddr = new const XNADDR*[uMaxResults];
	m_ppxQoSxnkid = new const XNKID*[uMaxResults];
	m_ppxQoSxnkey = new const XNKEY*[uMaxResults];

	ResetPropertiesAndContexts();
}

Asura_Xbox360_MatchFinder::~Asura_Xbox360_MatchFinder()
{
	delete[] m_pxContexts;
	delete[] m_pxProperties;
	delete[] m_pxSearchResults;

	XNetQosRelease(m_pxQoSResult);
	delete[] m_ppxQoSxnkey;
	delete[] m_ppxQoSxnkid;
	delete[] m_ppxQoSxnaddr;
}

bool Asura_Xbox360_MatchFinder::BeginSearch()
{
	// Figure out the size of the max search results
	DWORD uResultsSize = 0;
	u_int uRet;
	u_int uSearchType = 0;	// e.g. SESSION_MATCH_QUERY_DEFAULT, defined in the project.spa.h file
	u_int uUserIndex = Asura_Xbox360_OS::GetActivePlayerIndex();

	ZeroMemory( &m_xSearchOverlapped, sizeof( m_xSearchOverlapped ) );

	// Pass in 0 for uResultsSize to have the buffer size calculated for us by XTL
	uRet = XSessionSearch(
		uSearchType,      // Procedure index
		uUserIndex,       // User index
		m_uMaxResults,    // Maximum results
		0,                // Number of properties   (ignored)
		0,                // Number of contexts     (ignored)
		NULL,             // Properties             (ignored)
		NULL,             // Contexts               (ignored)
		&uResultsSize,    // Size of result buffer
		NULL,             // Pointer to results     (ignored)
		NULL              // This call is always synchronous
		);

	if(( uRet != ERROR_INSUFFICIENT_BUFFER ) || ( uResultsSize == 0 ))
	{
		Asura_Assert(false, "failed to get buffer for XSessionSearch");
		return false;
	}

	m_pxSearchResults = (XSESSION_SEARCHRESULT_HEADER *)new BYTE[ uResultsSize ];

	if( !m_pxSearchResults )
	{
		Asura_Assert(false, "Failed to allocate memory for search results." );
		return false;
	}

	u_short usNumContexts = static_cast<u_short>(m_uNumContexts);
	u_short usNumProperties = static_cast<u_short>(m_uNumProperties);

	// Fire off the query
	uRet = XSessionSearch(
		uSearchType,            // Procedure index
		uUserIndex,             // User index
		m_uMaxResults,			// Maximum results
		usNumProperties,        // Number of properties
		usNumContexts,          // Number of contexts
		m_pxProperties,         // Properties
		m_pxContexts,           // Contexts
		&uResultsSize,          // Size of result buffer
		m_pxSearchResults,      // Pointer to results
		&m_xSearchOverlapped    // Overlapped data structure
		);

	if( uRet != ERROR_IO_PENDING && uRet != ERROR_SUCCESS )
	{
		Asura_Assertf( false, ("XSessionSearch failed with error %d\n", uRet ));
		delete[] m_pxSearchResults;
		m_pxSearchResults = NULL;
		return false;
	}

	return true;
}

bool Asura_Xbox360_MatchFinder::IsSearchComplete()
{
	if (!m_pxSearchResults)
	{
		// error, so search is finished
		return true;
	}

	if( !XHasOverlappedIoCompleted( &m_xSearchOverlapped ) )
	{
		return false;
	}

	// We have results if the overlapped operation succeeded and if there are
	// more than zero results.
	bool bHasResults =
		SUCCEEDED( XGetOverlappedExtendedError( &m_xSearchOverlapped ) ) &&
		m_pxSearchResults->dwSearchResults > 0;

	if( bHasResults )
	{
		// start the QOS
		if( !m_bQoSTesting )
		{
			for( u_int uSessions = 0; uSessions < m_pxSearchResults->dwSearchResults; ++uSessions )
			{
				m_ppxQoSxnaddr[uSessions] = &(m_pxSearchResults->pResults[uSessions].info.hostAddress);
				m_ppxQoSxnkid[uSessions] = &(m_pxSearchResults->pResults[uSessions].info.sessionID);
				m_ppxQoSxnkey[uSessions] = &(m_pxSearchResults->pResults[uSessions].info.keyExchangeKey);
			}
			int iRet = XNetQosLookup( m_pxSearchResults->dwSearchResults, m_ppxQoSxnaddr,
				m_ppxQoSxnkid, m_ppxQoSxnkey, 0, NULL, NULL, 0, 0, 0,
				NULL, &m_pxQoSResult );
			if( 0 != iRet )
			{
				Asura_Assertf(false, ( "XNetQosLookup failed with error 0x%08x", iRet));
			}

			m_bQoSTesting = true;
		}
	}

	// TODO: should set a time limit to receive the QOS data and report unfinished until 
	// it's all in or the time limit has expired. For now ignore the QOS data, but leave 
	// in the calls to make it easier later...
	return true;
}

u_int Asura_Xbox360_MatchFinder::GetNumSearchResults() const
{
	if (m_pxSearchResults)
	{
		return m_pxSearchResults->dwSearchResults;
	}
	return 0;
}

const XSESSION_SEARCHRESULT& Asura_Xbox360_MatchFinder::GetResult( u_int uIndex ) const
{
	Asura_Assert(m_pxSearchResults && uIndex < m_pxSearchResults->dwSearchResults, "invalid index");
	return m_pxSearchResults->pResults[uIndex];
}

void Asura_Xbox360_MatchFinder::ResetPropertiesAndContexts( u_int uGameMode /* = 0 */ )
{
	Asura_Assert(m_uMaxContexts >= uASURA_XBOX360_NUM_REQUIRED_CONTEXTS && m_pxContexts, "array too small");
	m_pxContexts[0].dwContextId = X_CONTEXT_GAME_TYPE;
	m_pxContexts[0].dwValue     = m_bRankedGames ? X_CONTEXT_GAME_TYPE_RANKED : X_CONTEXT_GAME_TYPE_STANDARD;
	m_pxContexts[1].dwContextId = X_CONTEXT_GAME_MODE;
	m_pxContexts[1].dwValue     = uGameMode;

	ASURA_STATIC_ASSERT( uASURA_XBOX360_NUM_REQUIRED_CONTEXTS == 2 );

	m_uNumContexts = uASURA_XBOX360_NUM_REQUIRED_CONTEXTS;
	m_uNumProperties = 0;
}

void Asura_Xbox360_MatchFinder::AddSearchContext( u_int uContextID, u_int uContextValue )
{
	Asura_Assert(m_uNumContexts < m_uMaxContexts, "no room to add search criterion");
	if (m_uNumContexts < m_uMaxContexts)
	{
		u_int uIndex = m_uNumContexts++;
		m_pxContexts[uIndex].dwContextId = uContextID;
		m_pxContexts[uIndex].dwValue = uContextValue;
	}
}

void Asura_Xbox360_MatchFinder::AddSearchProperty( u_int uPropertyID, int iPropertyValue )
{
	Asura_Assert(m_uNumProperties < m_uMaxProperties, "no room to add search criterion");
	if (m_uNumProperties < m_uMaxProperties)
	{
		u_int uIndex = m_uNumProperties++;
		m_pxProperties[uIndex].dwPropertyId = uPropertyID;
		m_pxProperties[uIndex].value.type = XUSER_DATA_TYPE_INT32;
		m_pxProperties[uIndex].value.nData = iPropertyValue;
	}

}
