//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Stats.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     code for reading/writing stats to 360 Live! leaderboards
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_SUPPORT_LEADERBOARDS

// This file's header.
#include "Network/Online/Asura_Xbox360_Stats.h"

// Asura includes
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"
#include "Network/Online/Asura_Leaderboards.h"

// static class data
Asura_Xbox360_Stats_PreGameReader* Asura_Xbox360_Stats_PreGameReader::s_pxInstance;

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
void Asura_Xbox360_Stats::ClearAllStats()
{
	for (u_int u = 0; u < Asura_Leaderboard_System::Project_GetNumLeaderboards(); ++u)
	{
#ifdef ASURA_ASSERT
		HRESULT uResult = 
#endif
			XUserResetStatsViewAllUsers(Project_GetLeaderboardID(u), NULL);
		Asura_Assertf(uResult == ERROR_SUCCESS, ("XUserResetStatsViewAllUsers failed (0x%x)", uResult));
	}

	// also reset TrueSkill
#ifdef ASURA_ASSERT
	HRESULT uResult = 
#endif
		XUserResetStatsViewAllUsers(Project_GetTrueSkillLeaderboardID(true), NULL);
	Asura_Assertf(uResult == ERROR_SUCCESS, ("XUserResetStatsViewAllUsers failed (0x%x)", uResult));
#ifdef ASURA_ASSERT
	uResult = 
#endif
		XUserResetStatsViewAllUsers(Project_GetTrueSkillLeaderboardID(false), NULL);
	Asura_Assertf(uResult == ERROR_SUCCESS, ("XUserResetStatsViewAllUsers failed (0x%x)", uResult));
}
#endif // #ifdef ASURA_DEBUG_VARIABLE_SYSTEM


Asura_Xbox360_StatsWriter::Asura_Xbox360_StatsWriter(HANDLE xSession, bool bRanked)
: m_xSession(xSession)
, m_bRanked(bRanked)
{
	// unranked sessions only write TrueSkill
	const u_int uNumProjectLeaderboards = bRanked ? Asura_Leaderboard_System::Project_GetNumLeaderboards() : 0;
	Asura_Assert(uNumProjectLeaderboards <= 5, "can only write stats for up to 5 leaderboards, plus one trueskill leaderboard");
	m_uNumLeaderboards = uNumProjectLeaderboards + 1;
	m_pxViewProperties = new XSESSION_VIEW_PROPERTIES[m_uNumLeaderboards];

	u_int uNumProperties = 2;	// one property for TrueSkill
	for (u_int u = 0; u < uNumProjectLeaderboards; ++u)
	{
		uNumProperties += Asura_Leaderboard_System::Project_GetNumLeaderboardColumns(u);
	}
	m_uTotalUserProperties = uNumProperties;
	m_pxUserProperties = new XUSER_PROPERTY[uNumProperties];
	Asura_Mem::Set(m_pxUserProperties, 0, sizeof(XUSER_PROPERTY) * uNumProperties);
}

Asura_Xbox360_StatsWriter::~Asura_Xbox360_StatsWriter()
{
	delete[] m_pxUserProperties;
	delete[] m_pxViewProperties;
}

void Asura_Xbox360_StatsWriter::WriteStatsForPlayer( bool bDisconnectedEarly, u_int uGuid, XUID xXuid )
{
	u_int uNumUserProperties = 0;
	u_int uNumLeaderboards = 0;
	Asura_Assert(m_uNumLeaderboards, "should at least have a trueskill leaderboard");
	for (u_int uLeaderboard = 0; uLeaderboard < m_uNumLeaderboards - 1; ++uLeaderboard)
	{
		// first set the view properties for this leaderboard
		const u_int uNumStatsForLeaderboard = Asura_Leaderboard_System::Project_GetNumLeaderboardColumns(uLeaderboard);

		const u_int uInitialNumProperties = uNumUserProperties;
		// now fill out all the stats (user properties) for this leaderboard
		for (u_int uStat = 0; uStat < uNumStatsForLeaderboard; ++uStat)
		{
			Asura_Assert(uNumUserProperties < m_uTotalUserProperties, "project lied about total user properties");
			bool bGotStat = Asura_Xbox360_Stats::Project_GetStat( xXuid, bDisconnectedEarly, uGuid, uLeaderboard, uStat, m_pxUserProperties[uNumUserProperties] );
			Asura_Assert(bGotStat, "failed to get stat from project code");
			if (bGotStat)
			{
				++uNumUserProperties;
			}
		}

		const u_int uNumPropertiesWritten = uNumUserProperties - uInitialNumProperties;
		if (uNumPropertiesWritten)
		{
			m_pxViewProperties[uNumLeaderboards].dwViewId = Asura_Xbox360_Stats::Project_GetLeaderboardID(uLeaderboard);
			m_pxViewProperties[uNumLeaderboards].dwNumProperties = uNumPropertiesWritten;
			m_pxViewProperties[uNumLeaderboards].pProperties = &m_pxUserProperties[uInitialNumProperties];
			++uNumLeaderboards;
		}
	}
	// now the trueskill leaderboard
	const u_int uInitialNumProperties = uNumUserProperties;
	Asura_Assert((uNumUserProperties  + 1) < m_uTotalUserProperties, "project lied about total user properties");
	m_pxUserProperties[uNumUserProperties].dwPropertyId = X_PROPERTY_RELATIVE_SCORE;
	m_pxUserProperties[uNumUserProperties].value.type = XUSER_DATA_TYPE_CONTEXT;
	m_pxUserProperties[uNumUserProperties].value.nData = 
		Asura_Xbox360_Stats::Project_GetTrueSkillRelativeScore( bDisconnectedEarly, uGuid );
	++uNumUserProperties;
	m_pxUserProperties[uNumUserProperties].dwPropertyId = X_PROPERTY_SESSION_TEAM;
	m_pxUserProperties[uNumUserProperties].value.type = XUSER_DATA_TYPE_CONTEXT;
	m_pxUserProperties[uNumUserProperties].value.nData = Asura_Xbox360_Stats::Project_GetTrueSkillTeam(uGuid);
	++uNumUserProperties;

	m_pxViewProperties[uNumLeaderboards].dwViewId = Asura_Xbox360_Stats::Project_GetTrueSkillLeaderboardID(m_bRanked);
	m_pxViewProperties[uNumLeaderboards].dwNumProperties = 2;
	m_pxViewProperties[uNumLeaderboards].pProperties = &m_pxUserProperties[uInitialNumProperties];
	++uNumLeaderboards;
	Asura_Assert(uNumUserProperties == m_uTotalUserProperties, "didn't write expected number of user properties");

	// all properties now written, so write the stats.
#ifdef ASURA_ASSERT
	HRESULT uResult = 
#endif
		XSessionWriteStats(m_xSession, xXuid, uNumLeaderboards, m_pxViewProperties, NULL); // TODO: use overlapped?
	Asura_Assertf(uResult == ERROR_SUCCESS, ("XSessionWriteStats failed (0x%x)", uResult));

#if 0 // _DEBUG // turn this on to dump the written stats to a logfile. This can help compare what server and client are writing.
	char szLogfile[128];
	char acPlayerName[32];
	XUserGetName(0, acPlayerName, 32);
	Asura_PrintfToBfr(szLogfile, "logfiles\\XSessionWriteStats_%s.txt", 128, true, acPlayerName);
	Asura_Logfile((szLogfile, "XSessionWriteStats for XUID %llx\n", xXuid));
	for (u_int u = 0; u < uNumLeaderboards; ++u)
	{
		Asura_Logfile((szLogfile, "Leaderboard %u, viewId = 0x%08x, num properties = %u\n", u, m_pxViewProperties[u].dwViewId, m_pxViewProperties[u].dwNumProperties ));
		for (u_int v = 0; v < m_pxViewProperties[u].dwNumProperties; ++v)
		{
			const XUSER_PROPERTY& xUserProperty = m_pxViewProperties[u].pProperties[v];
			Asura_Logfile((szLogfile, "Property %u, propertyId = 0x%08x, value.type = %u, value.nData = %d, value.i64Data = %lld\n", 
				u, xUserProperty.dwPropertyId, xUserProperty.value.type, xUserProperty.value.nData, xUserProperty.value.i64Data ));
		}
	}
#endif
}


Asura_Xbox360_StatsReader::Asura_Xbox360_StatsReader()
{
	m_xEnumeratorHandle = INVALID_HANDLE_VALUE;
	ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );
	m_pBuffer = NULL;
	m_bDataIsReady = false;
	m_bError = false;
	m_bFetchInProgress = false;
}

Asura_Xbox360_StatsReader::~Asura_Xbox360_StatsReader()
{
	Asura_Assert(!m_bFetchInProgress, "stats fetch still active - must block until operation is complete.");
	if (m_bFetchInProgress)
	{
		while (!XHasOverlappedIoCompleted(&m_xOverlapped))
		{
			SleepEx(100, TRUE);
		}
#ifdef ASURA_ASSERT
		BOOL bOK = 
#endif
			CloseHandle(m_xEnumeratorHandle);
		Asura_Assertf(bOK, ("CloseHandle failed 0x%x", GetLastError()));
	}
	Asura_Deallocate(m_pBuffer);
}

bool Asura_Xbox360_StatsReader::FetchStats( u_int uStart, u_int uNumRows, const XUSER_STATS_SPEC& xStatsSpec )
{
	Asura_Assert(m_xEnumeratorHandle == INVALID_HANDLE_VALUE, "enumerator still in use");
	Asura_Assert(!m_bFetchInProgress, "stats fetch still active");
	if (m_bFetchInProgress)
	{
		return false;
	}

	// first create the enumerator object which will tell us the buffer size we need
	DWORD uBufferSize = 0;
	
	HRESULT uResult = 
		XUserCreateStatsEnumeratorByRank(
		0,
		uStart + 1,
		uNumRows,
		1,
		&xStatsSpec,
		&uBufferSize,
		&m_xEnumeratorHandle
		);
	Asura_Assertf(uResult == ERROR_SUCCESS, ("XUserCreateStatsEnumeratorByRank failed (0x%x)", uResult));
	if (uResult != ERROR_SUCCESS)
	{
		return false;
	}

	// allocate the buffer
	Asura_Assert(!m_pBuffer, "memory leak on stats buffer");
	m_pBuffer = Asura_Allocate(uBufferSize);

	// start an overlapped operation - we will check for completion in the Update() function
	uResult = XEnumerate( m_xEnumeratorHandle, m_pBuffer, uBufferSize, NULL, &m_xOverlapped );
	Asura_Assertf(uResult == ERROR_IO_PENDING, ("XEnumerate failed for stats read (0x%x)", uResult));
	if (uResult != ERROR_IO_PENDING)
	{
		// possibly out of memory. clean up and return false.
#ifdef ASURA_ASSERT
		BOOL bOK = 
#endif
			CloseHandle(m_xEnumeratorHandle);
		Asura_Assertf(bOK, ("CloseHandle failed 0x%x", GetLastError()));
		m_xEnumeratorHandle = INVALID_HANDLE_VALUE;
		Asura_Deallocate(m_pBuffer);
		m_pBuffer = NULL;
		return false;
	}

	m_bDataIsReady = false;
	m_bError = false;
	m_bFetchInProgress = true;

	return true;
}

bool Asura_Xbox360_StatsReader::FetchStatsForXuids( const XUID* pxXuids, u_int uNumXuids, const XUSER_STATS_SPEC& xStatsSpec )
{
	Asura_Assert(!m_bFetchInProgress, "stats fetch still active");
	if (m_bFetchInProgress)
	{
		return false;
	}

	// first create the enumerator object which will tell us the buffer size we need
	DWORD uBufferSize = 0;

	HRESULT uResult = XUserReadStats( 0, uNumXuids, pxXuids, 1, &xStatsSpec, &uBufferSize, NULL, NULL );
	Asura_Assertf(uResult == ERROR_INSUFFICIENT_BUFFER, ("XUserReadStats failed (0x%x)", uResult));
	if (uResult != ERROR_INSUFFICIENT_BUFFER)
	{
		m_bError = true;
		return false;
	}

	// allocate the buffer
	Asura_Assert(!m_pBuffer, "memory leak on stats buffer");
	m_pBuffer = Asura_Allocate(uBufferSize);

	// start an overlapped operation - we will check for completion in the Update() function
	uResult = XUserReadStats( 0, uNumXuids, pxXuids, 1, &xStatsSpec, &uBufferSize, 
		static_cast<XUSER_STATS_READ_RESULTS*>(m_pBuffer), &m_xOverlapped );

	Asura_Assertf(uResult == ERROR_IO_PENDING, ("XUserReadStats failed (0x%x)", uResult));
	if (uResult != ERROR_IO_PENDING)
	{
		// possibly out of memory. clean up and return false.
		Asura_Deallocate(m_pBuffer);
		m_pBuffer = NULL;
		m_bError = true;
		return false;
	}

	m_bDataIsReady = false;
	m_bError = false;
	m_bFetchInProgress = true;

	return true;
}

bool Asura_Xbox360_StatsReader::Update()
{
	// check for pending results being complete.
	if (m_bFetchInProgress)
	{
		if( XHasOverlappedIoCompleted( &m_xOverlapped ) )
		{
			HRESULT uResult = XGetOverlappedExtendedError( &m_xOverlapped );

			Asura_Assertf(uResult == ERROR_SUCCESS, ("XEnumerate for Stats failed (0x%x)", uResult));
			if( SUCCEEDED( uResult ) )
			{
				m_bDataIsReady = true;
			}
			else
			{
				m_bError = true;
				FreeData();
			}
		}
	}
	return m_bDataIsReady || m_bError;
}


const XUSER_STATS_VIEW* Asura_Xbox360_StatsReader::GetData()
{
	Asura_Assert(m_bDataIsReady, "data not ready yet");

	// copy the data to wherever we want it.
	const XUSER_STATS_READ_RESULTS* pxResults = static_cast<const XUSER_STATS_READ_RESULTS*>(m_pBuffer);
	Asura_Assert(pxResults->dwNumViews == 1, "unexpected number of leaderboards returned");
	return pxResults->pViews;
}


void Asura_Xbox360_StatsReader::FreeData()
{
	// clean up
	if (m_xEnumeratorHandle != INVALID_HANDLE_VALUE)
	{
#ifdef ASURA_ASSERT
		BOOL bOK = 
#endif
			CloseHandle(m_xEnumeratorHandle);
		Asura_Assertf(bOK, ("CloseHandle failed 0x%x", GetLastError()));
		m_xEnumeratorHandle = INVALID_HANDLE_VALUE;
	}
	ZeroMemory( &m_xOverlapped, sizeof( m_xOverlapped ) );
	Asura_Deallocate(m_pBuffer);
	m_pBuffer = NULL;
	m_bDataIsReady = false;
	m_bFetchInProgress = false;
}


//////////////////////////////////////////////////////////////////////////
// Asura_Xbox360_Stats_PreGameReader
//////////////////////////////////////////////////////////////////////////
Asura_Xbox360_Stats_PreGameReader::Asura_Xbox360_Stats_PreGameReader( const XUID* pxXuids, u_int uNumXuids )
{
	// copy XUID array since it must persist for duration of call.
	m_pxXuids = new XUID[uNumXuids];
	for (u_int u = 0; u < uNumXuids; ++u)
	{
		m_pxXuids[u] = pxXuids[u];
	}
	m_uNumXuids = uNumXuids;

	m_pxReader = new Asura_Xbox360_StatsReader;
	m_pxStatsSpec = new XUSER_STATS_SPEC;

	m_pxStatsSpec->dwViewId = Project_GetPreGameLeaderboardID();
	const u_int uNumColumns = Project_GetNumColumns();
	Asura_Assert(uNumColumns && uNumColumns < XUSER_STATS_ATTRS_IN_SPEC, "too many columns");
	m_pxStatsSpec->dwNumColumnIds = uNumColumns;
	for (u_int uCol = 0; uCol < uNumColumns; ++uCol)
	{
		Asura_Assert(Project_GetColumnID(uCol) != 0, "unexpected column ID");
		m_pxStatsSpec->rgwColumnIds[uCol] = Project_GetColumnID(uCol);
	}

	m_pxReader->FetchStatsForXuids( m_pxXuids, uNumXuids, *m_pxStatsSpec );
}

Asura_Xbox360_Stats_PreGameReader::~Asura_Xbox360_Stats_PreGameReader()
{
	if (m_pxReader)
	{
		m_pxReader->FreeData();
	}
	delete m_pxStatsSpec;
	delete m_pxReader;
	delete[] m_pxXuids;
}

void Asura_Xbox360_Stats_PreGameReader::Initialise( const XUID* pxXuids, u_int uNumXuids )
{
	Asura_Assert(!s_pxInstance, "already initialised");
	s_pxInstance = new Asura_Xbox360_Stats_PreGameReader(pxXuids, uNumXuids);
}

bool Asura_Xbox360_Stats_PreGameReader::Update()
{
	// returns true on completion
	if (s_pxInstance)
	{
		return s_pxInstance->m_pxReader->Update();
	}
	return true;
}

void Asura_Xbox360_Stats_PreGameReader::Shutdown()
{
	delete s_pxInstance;
	s_pxInstance = NULL;
}

const XUSER_STATS_ROW* Asura_Xbox360_Stats_PreGameReader::GetDataForXuid( const XUID& xXuid )
{
	// may return NULL

	if (!s_pxInstance)
	{
		return NULL;
	}
	return s_pxInstance->GetDataForXuid_Internal( xXuid );
}

const XUSER_STATS_ROW* Asura_Xbox360_Stats_PreGameReader::GetDataForXuid_Internal( const XUID& xXuid ) const
{
	const XUSER_STATS_VIEW* pxStatsView = m_pxReader->GetData();

	Asura_Assert(pxStatsView, "no stats data");
	if (!pxStatsView)
	{
		return NULL;
	}

	for (u_int uRow = 0; uRow < pxStatsView->dwNumRows; ++uRow)
	{
		if (pxStatsView->pRows[uRow].xuid == xXuid)
		{
			return &pxStatsView->pRows[uRow];
		}
	}
	Asura_Assert(false, "failed to find stats data for this XUID");
	return NULL;
}

#endif //ASURA_SUPPORT_LEADERBOARDS

//eof
