
//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Projector_System.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifdef ASURA_TARGET_SM3

#include "Projectors/Asura_Projector_System.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_RenderStates.h"
#include "Rendering/Asura_Viewport.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Debugging/Asura_Platform_GPUProfile.h"

// Misc
#include <fxl.h>
#include <xtl.h>

// Asura Shader headers
#include "ApplyProjectedTexture_MR.h"

const Direct3D_VertexElement axVERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	D3DDECL_END()
};

enum
{
	// Effect Pass Indices
	Projector_Pass_ClearDestAlpha = 0,
	Projector_Pass_DynamicLight	  = 1,
	Projector_Pass_FakeLight	  = 2,
	Projector_Pass_Standard		  = 3,
	Projector_Pass_Spotlight	  = 4
};

static Direct3D_VertexDeclaration* g_pxTexProjectionVD = NULL;
static Asura_Xbox360_Effect*	   g_pxApplyProjectedTextureEffect = NULL;


static class
{
public:

	D3DXMATRIX 				m_xProjectionMatrix;
	D3DXMATRIX				m_xInvViewMatrix;
	D3DXVECTOR4				m_xNearPlane;
	D3DXVECTOR4				m_xNeutralColour;
	D3DXVECTOR4				m_xProjectorPosition;
	D3DXVECTOR4				m_xProjectionDirection;
	float					m_fDistanceFade;
	float 					m_fTwoOverScreenWidth;
	float 					m_fTwoOverScreenHeight;
	float 					m_fCameraTanHalfVertFOV;
	float 					m_fCameraTanHalfHorizFOV;
	D3DXVECTOR4				m_xDetailTextureScaleAndOffset;
	Direct3D_BaseTexture*	m_pxNormalBuffer;
	Direct3D_BaseTexture*	m_pxDepthMtlBuffer;
	Direct3D_BaseTexture*	m_pxAlbedoBuffer;
	Direct3D_BaseTexture*	m_pxProjectedTexture;
	Direct3D_BaseTexture*	m_pxDetailTexture;
	D3DXVECTOR4				m_xFalloff;
	bool					m_bUseDetailTexture;
	float					m_afLightColour[3];

private:

	Asura_Xbox360_Effect_System::Handle m_xProjectionMatrixHandle;
	Asura_Xbox360_Effect_System::Handle m_xInvViewMatrixHandle;
	Asura_Xbox360_Effect_System::Handle m_xNearPlaneHandle;
	Asura_Xbox360_Effect_System::Handle m_xNeutralColourHandle;
	Asura_Xbox360_Effect_System::Handle m_xProjectorPositionHandle;
	Asura_Xbox360_Effect_System::Handle m_xProjectionDirectionHandle;
	Asura_Xbox360_Effect_System::Handle m_xDistanceFadeHandle;
	Asura_Xbox360_Effect_System::Handle m_xTwoOverScreenWidthHandle;
	Asura_Xbox360_Effect_System::Handle m_xTwoOverScreenHeightHandle;
	Asura_Xbox360_Effect_System::Handle m_xCameraTanHalfVertFOVHandle;
	Asura_Xbox360_Effect_System::Handle m_xCameraTanHalfHorizFOVHandle;
	Asura_Xbox360_Effect_System::Handle m_xDetailTextureScaleAndOffsetHandle;
	Asura_Xbox360_Effect_System::Handle m_xNormalBufferHandle;
	Asura_Xbox360_Effect_System::Handle m_xDepthMtlBufferHandle;
	Asura_Xbox360_Effect_System::Handle m_xAlbedoBufferHandle;
	Asura_Xbox360_Effect_System::Handle m_xProjectedTextureHandle;
	Asura_Xbox360_Effect_System::Handle m_xFalloffHandle;
	Asura_Xbox360_Effect_System::Handle m_xDetailTextureHandle;
	Asura_Xbox360_Effect_System::Handle m_xUseDetailTextureHandle;
	Asura_Xbox360_Effect_System::Handle m_xLightColourHandle;	
	Asura_Xbox360_Effect_System::Handle m_xMaterialsTextureHandle;
	Asura_Xbox360_Effect_System::Handle m_xUseSpecularMapHandle;

public:
	// Get the parameter handles
	void Initialise(void)
	{
		Asura_Xbox360_Effect *pxEffect = g_pxApplyProjectedTextureEffect;

		m_xProjectionMatrixHandle		= pxEffect->GetParameterHandle("g_xProjectionMatrix");
		m_xInvViewMatrixHandle			= pxEffect->GetParameterHandle("g_xInvViewMatrix");
		m_xNearPlaneHandle				= pxEffect->GetParameterHandle("g_xNearPlane");
		m_xNeutralColourHandle			= pxEffect->GetParameterHandle("g_xNeutralColour");
		m_xProjectorPositionHandle		= pxEffect->GetParameterHandle("g_xProjectorPosition");
		m_xProjectionDirectionHandle    = pxEffect->GetParameterHandle("g_xProjectionDirection");
		m_xDistanceFadeHandle			= pxEffect->GetParameterHandle("g_fDistanceFade");
		m_xTwoOverScreenWidthHandle		= pxEffect->GetParameterHandle("g_fTwoOverScreenWidth");
		m_xTwoOverScreenHeightHandle	= pxEffect->GetParameterHandle("g_fTwoOverScreenHeight");
		m_xCameraTanHalfVertFOVHandle	= pxEffect->GetParameterHandle("g_fCameraTanHalfVertFOV");
		m_xCameraTanHalfHorizFOVHandle	= pxEffect->GetParameterHandle("g_fCameraTanHalfHorizFOV");
		m_xDetailTextureScaleAndOffsetHandle = pxEffect->GetParameterHandle("g_xDetailTextureScaleAndOffset");
		m_xDepthMtlBufferHandle			= pxEffect->GetParameterHandle("g_xDepthMtlBufferSampler");
		m_xNormalBufferHandle			= pxEffect->GetParameterHandle("g_xNormalBufferSampler");
		m_xAlbedoBufferHandle			= pxEffect->GetParameterHandle("g_xAlbedoBufferSampler");
		m_xProjectedTextureHandle		= pxEffect->GetParameterHandle("g_xProjectedTextureSampler");		
		m_xDetailTextureHandle			= pxEffect->GetParameterHandle("g_xDetailTextureSampler");
		m_xFalloffHandle				= pxEffect->GetParameterHandle("g_xFalloff");
		m_xUseDetailTextureHandle		= pxEffect->GetParameterHandle("g_bUseDetailTexture");		
		m_xLightColourHandle			= pxEffect->GetParameterHandle("g_xLightColour");		
		m_xMaterialsTextureHandle		= pxEffect->GetParameterHandle("g_xMaterialsTextureSampler" );
		m_xUseSpecularMapHandle			= pxEffect->GetParameterHandle("g_bUseSpecularMap" );

	};	

	// Apply the parameters
	void Apply()
	{
		Asura_Xbox360_Effect *pxEffect = g_pxApplyProjectedTextureEffect;
		FXLEffect *pxFXLEffect = pxEffect->GetInterface();

		pxFXLEffect->SetMatrixF(m_xProjectionMatrixHandle,		m_xProjectionMatrix);
		pxFXLEffect->SetMatrixF(m_xInvViewMatrixHandle,			m_xInvViewMatrix);
		pxFXLEffect->SetVectorF(m_xNearPlaneHandle,				m_xNearPlane);
		pxFXLEffect->SetVectorF(m_xNeutralColourHandle,			m_xNeutralColour);
		pxFXLEffect->SetVectorF(m_xProjectorPositionHandle,		m_xProjectorPosition);
		pxFXLEffect->SetVectorF(m_xProjectionDirectionHandle,	m_xProjectionDirection);
		pxFXLEffect->SetFloat(m_xDistanceFadeHandle,			m_fDistanceFade);
		pxFXLEffect->SetFloat(m_xTwoOverScreenWidthHandle,		m_fTwoOverScreenWidth);
		pxFXLEffect->SetFloat(m_xTwoOverScreenHeightHandle,		m_fTwoOverScreenHeight);
		pxFXLEffect->SetFloat(m_xCameraTanHalfHorizFOVHandle,	m_fCameraTanHalfHorizFOV);
		pxFXLEffect->SetFloat(m_xCameraTanHalfVertFOVHandle,	m_fCameraTanHalfVertFOV);
		pxFXLEffect->SetVectorF(m_xDetailTextureScaleAndOffsetHandle,	m_xDetailTextureScaleAndOffset);
		pxFXLEffect->SetSampler(m_xNormalBufferHandle,			m_pxNormalBuffer);
		pxFXLEffect->SetSampler(m_xDepthMtlBufferHandle,		m_pxDepthMtlBuffer);
		pxFXLEffect->SetSampler(m_xAlbedoBufferHandle,		    m_pxAlbedoBuffer);
		pxFXLEffect->SetSampler(m_xProjectedTextureHandle,      m_pxProjectedTexture);
		pxFXLEffect->SetVectorF(m_xFalloffHandle,				m_xFalloff);
		pxFXLEffect->SetSampler(m_xDetailTextureHandle,		    m_pxDetailTexture);
		pxFXLEffect->SetBool(m_xUseDetailTextureHandle,			m_bUseDetailTexture);	
		pxFXLEffect->SetVectorF(m_xLightColourHandle,			m_afLightColour);			
		pxFXLEffect->SetSampler( m_xMaterialsTextureHandle, Asura_Xbox360_MaterialResponse_System::GetDataTexture() );
		pxFXLEffect->SetBool( m_xUseSpecularMapHandle, true );
	};


} g_xProjectedTextureParameters;



void Asura_Projector_System::Platform_Initialise(void)
{
	// Initialise effect and parameter handles
	g_pxApplyProjectedTextureEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory( g_auApplyProjectedTexture_MR );

	g_xProjectedTextureParameters.Initialise();
	
	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxTexProjectionVD );
	Asura_Assert( g_pxTexProjectionVD, "Failed to create shadow frustum vertex declaration." );
}


void Asura_Projector_System::Platform_Shutdown(void)
{

	g_pxApplyProjectedTextureEffect = NULL;

	if (g_pxTexProjectionVD)
	{
		g_pxTexProjectionVD->Release();
		g_pxTexProjectionVD = 0;
	}
}


void Asura_Projector_System::Platform_Reset(void)
{	
}




static void SetZTestAndCullMode(bool bInsideFrustum)
{	

	// Force this change, as Asura's cached state is likely to be invalid, causing it to ignore the state change request...
	// (I think it's the shadow system messing up Asura's cached states by setting them directly?)
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_ALWAYS);

	if(bInsideFrustum)
	{
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_GREATER);		
	}
	else
	{	
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );		
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		
	}
}


void Asura_Projector_System::Platform_BeginRender(void)
{
#if defined(ASURA_XBOX360_ENABLE_PIX)
	PIXBeginNamedEvent(D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Projectors");
#endif	//defined(ASURA_XBOX360_ENABLE_PIX)
}


void Asura_Projector_System::Platform_RenderProjector(Asura_Projector *pxProjector)
{
	// 6 faces/12 tris that make up a frustum - const index list
	const uint16 au16Indices[] = {0,1,3, 1,2,3, 
								  0,4,1, 1,4,5,  
								  1,5,6, 1,6,2,
								  3,2,7, 2,6,7, 
								  0,3,7, 0,7,4,
								  4,7,5, 5,7,6};

	if (Asura_DeferredShading::IsShowingProjectorFrustums())
	{	
		pxProjector->RenderFrustum();
	}
	

	// Is the projector in range?
	const Asura_Vector_3&	xProjectorPos	= pxProjector->m_xPos;
	Asura_Vector_3			xVec			= *Asura_Camera::GetPosition() - xProjectorPos;

	if (xVec.MagnitudeSquared() > ((pxProjector->m_fDistFadeStart+pxProjector->m_fDistFadeRange)*(pxProjector->m_fDistFadeStart+pxProjector->m_fDistFadeRange)))
	{
		// out of range - don't draw anything
		return;
	}

	float fDistanceFade = 1.0f;
	float fDist = xVec.Magnitude();

	if (fDist> pxProjector->m_fDistFadeStart)
	{ 
		fDistanceFade = 1.0f - ((fDist - pxProjector->m_fDistFadeStart)/pxProjector->m_fDistFadeRange);
	}

	if (fDistanceFade<=0.0f)
	{
		return;
	}

	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_DEFERRED_PROJECTORS);
	SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Projector @ (%.1f, %.1f, %.1f)", xProjectorPos.GetX(), xProjectorPos.GetY(), xProjectorPos.GetZ())); 

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxTexProjectionVD );	
 
	// If we're inside the frustum, draw the back faces, else draw the front faces. When outside the
	// frustum, we can leave z-tests on (as an optimization - less pixel shader work)
	Asura_Render::CurrentStates.RequestAlphaTest( false );	

	// I've had some problems with Asura's cached renderstates being out-of-sync with D3D, so we change the
	// state twice to force the change...
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );		
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );		
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );


	bool bCameraInFrustum = pxProjector->PointInFrustum(*Asura_Camera::GetPosition(),1.0f);
	SetZTestAndCullMode(bCameraInFrustum);
	
	
	// Set the textures
	Asura_Xbox360_Texture* pxTexture = Asura_Xbox360_TextureManagement::GetTexture( pxProjector->m_iTextureIndex );
	g_xProjectedTextureParameters.m_pxProjectedTexture = pxTexture->GetD3DTexture();	
	
	pxTexture = Asura_Xbox360_TextureManagement::GetTexture( pxProjector->m_iDetailTextureIndex );
	g_xProjectedTextureParameters.m_pxDetailTexture = pxTexture->GetD3DTexture();	
	g_xProjectedTextureParameters.m_bUseDetailTexture = (pxProjector->m_iDetailTextureIndex!=ASURA_TEXTURE_NONE);		

	// Falloff
	g_xProjectedTextureParameters.m_xFalloff.x = pxProjector->m_fFalloffScale;	
	g_xProjectedTextureParameters.m_xFalloff.y = pxProjector->m_fFalloffPower;	
	g_xProjectedTextureParameters.m_xFalloff.z = pxProjector->m_fAngularFalloffOfs;
	g_xProjectedTextureParameters.m_xFalloff.w = pxProjector->m_fAngularFalloffScale;

	
	// Set the neutral colour - this is the colour that will have no visible effect when
	// blended in the current blend mode - and set when outside the range of projection/falloff
	Asura_RGBA xNeutralColour;

	// Hacky: In case cached states are wrong, this will force the state change
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 	

	bool bShadows = false;
	int  iEffectPass = 0;


	bool bUseStencil = Asura_DeferredShading::GetUseStencilBuffer();


	// -----------------------------------------------------------------------------------------------------------------------
	// DMR - 03/04/07 - Stencil buffer optimization
	//
	// The stencil mask is created before rendering shadows or the light itself, as it should help with both.
	// We only care about the Hi-Stencil, as the per-pixel stencil test is executed after the shader - so not an optimization
	//
	// It takes 2 passes to fill the Hi-Stencil - the first one draws back faces, setting the stencil to 1 on z-pass and 0 on
	// z-fail. The 2nd pass draws front faces, setting the hi-stencil if the stencil was previously set to 1 (i.e. if both
	// the front and back are visible => nothing to light)
	//
	// (The camera-inside-the-volume case needs a special-case in the 2nd pass - draws the back faces again, with z-testing off)
	//
	// -----------------------------------------------------------------------------------------------------------------------	
	if (bUseStencil)
	{

		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
		Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_ALWAYS);		
		// need to make sure the last bound RT was 32bpp as the hardware wont do fast z writes with a 64bit RT!
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetPostToneMapRT());
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, NULL);

		for (int iPass=0; iPass<2; iPass++)
		{			
			g_pxApplyProjectedTextureEffect->Begin(0, Projector_Pass_ClearDestAlpha);
			g_xProjectedTextureParameters.Apply();
			g_pxApplyProjectedTextureEffect->Commit();

			switch(iPass)
			{		
			case 0:		
				// Back faces
				Asura_Xbox360_Graphics::GetDevice()->SetPixelShader(NULL);
				Asura_Render::CurrentStates.RequestStencilEnabled(true);
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILFUNC,			D3DCMP_ALWAYS );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILREF,			ASURA_STENCIL_MASK_PHOTON );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILWRITEMASK,		ASURA_STENCIL_MASK_PHOTON );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILMASK,			ASURA_STENCIL_MASK_PHOTON );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILPASS,			D3DSTENCILOP_REPLACE );				
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILZFAIL,			D3DSTENCILOP_ZERO );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE, TRUE );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILFUNC,		D3DHSCMP_NOTEQUAL );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILREF,			0x3 );

				Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
				Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		

				break;

			case 1:
				// Front faces
				Asura_Xbox360_Graphics::GetDevice()->SetPixelShader(NULL);				
				Asura_Render::CurrentStates.RequestStencilEnabled(true);
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILFUNC,			D3DCMP_ALWAYS );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILREF,			ASURA_STENCIL_MASK_PHOTON );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILWRITEMASK,		ASURA_STENCIL_MASK_PHOTON );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILMASK,			ASURA_STENCIL_MASK_PHOTON );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILPASS,			D3DSTENCILOP_KEEP );				
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILZFAIL,			D3DSTENCILOP_KEEP );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE, TRUE );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILFUNC,		D3DHSCMP_EQUAL );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILREF,			1 );

				if (bCameraInFrustum)
				{
					Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
					Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_ALWAYS);		
				}
				else
				{
					Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );		
					Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		
				}

				Asura_Xbox360_Graphics::GetDevice()->FlushHiZStencil( D3DFHZS_ASYNCHRONOUS );

				break;
			}


			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedPrimitiveUP( D3DPT_TRIANGLELIST,0,8,12, (void*)au16Indices, D3DFMT_INDEX16, (void *)pxProjector->m_axVertices ,12);
			
			g_pxApplyProjectedTextureEffect->End();				
		}

		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, FALSE );
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE,	FALSE );
		Asura_Render::CurrentStates.RequestStencilEnabled(false);
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetFinalRT());
	}









	if (true == pxProjector->m_bDynamicLight || true == pxProjector->m_bDynamicSpotlight)
	{

		if (true == pxProjector->m_bDynamicLight)
			iEffectPass = Projector_Pass_DynamicLight;
		else
			iEffectPass = Projector_Pass_Spotlight;

		// -----------------------------------------------------------------------------------------------------------------------
		// DMR - 03/04/07 - Shadow support. Only for 'Dynamic Light' projectors.
		// If this projector is a shadow caster, we need to build the shadow mask now (in the destination alpha channel), and set the
		// appropriate blending mode.
		//
		// Before projecting the shadows, we need to clear the area of the shadow mask that's used by this light - we do that by
		// rendering the projector frustum, with colour writes only enabled to the alpha channel
		// -----------------------------------------------------------------------------------------------------------------------
		if (pxProjector->GetShadowParameters().m_uFlags & (ASURA_SHADOWFLAG_CASTS_OWN_SHADOWS|ASURA_SHADOWFLAG_CASTS_ENVIRONMENT_SHADOWS))
		{	
			// Are there any active shadows for this light?
			if (Asura_Shadow_System::LightHasActiveShadows(ASURA_SHADOW_LIGHTSOURCE_PROJECTOR,pxProjector->GetHandle()))
			{	
				// Set this flag, so we know to restore the renderstates
				bShadows = true;
			}
		}

		if (bShadows)
		{
			ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_DEFERRED_PROJECTORS);
			ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_APPLYSHADOWS);
			
			// First, render the projection frustum to fill the destination alpha with 0's 
			Asura_Xbox360_Graphics::GetDevice()->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA);
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 			
			
			g_pxApplyProjectedTextureEffect->Begin(0, Projector_Pass_ClearDestAlpha);
			g_xProjectedTextureParameters.Apply();
			g_pxApplyProjectedTextureEffect->Commit();

			if (bUseStencil)
			{
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, TRUE );				
				Asura_Xbox360_Graphics::GetDevice()->FlushHiZStencil( D3DFHZS_ASYNCHRONOUS );
			}

			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedPrimitiveUP( D3DPT_TRIANGLELIST,0,8,12, (void*)au16Indices, D3DFMT_INDEX16, (void *)pxProjector->m_axVertices ,12);

			g_pxApplyProjectedTextureEffect->End();

			// Then render the shadow mask into the destination alpha			
			Asura_Shadow_System::BuildShadowMask(ASURA_SHADOW_LIGHTSOURCE_PROJECTOR, pxProjector->GetHandle());

			if (bUseStencil)
			{
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, FALSE );			
			}

			// Re-enable RGB writes			
			Asura_Xbox360_Graphics::GetDevice()->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA|D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE);

			// Restore  vertex decl			
			Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxTexProjectionVD );

			// Restore Z-test and culling mode
			SetZTestAndCullMode(bCameraInFrustum);

			// Asura doesn't have a mode for this, so I'm changing it manually here, then restoring it when we're done			
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING); 
			D3DBLENDSTATE xBlendState;
			xBlendState.BlendOp = xBlendState.BlendOpAlpha = D3DBLENDOP_ADD;
			xBlendState.SrcBlend = xBlendState.SrcBlendAlpha = D3DBLEND_INVDESTALPHA;
			xBlendState.DestBlend = xBlendState.DestBlendAlpha = D3DBLEND_ONE;
			Asura_Xbox360_Graphics::GetDevice()->SetBlendState( 0, xBlendState );
			xNeutralColour = Asura_RGBA(0x000000,0);

			ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_APPLYSHADOWS);
			ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_DEFERRED_PROJECTORS);
		}
		else
		{					
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING); 		
			xNeutralColour = Asura_RGBA(0x000000,0);
		}		
	}
	else
	{	
		if (pxProjector->m_bFakeLight)
		{
			iEffectPass = Projector_Pass_FakeLight;
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING); 	
			xNeutralColour = Asura_RGBA(0x000000,0);
		}
		else
		{	
			iEffectPass = Projector_Pass_Standard;
			switch (pxProjector->m_eBlendMode)
			{				
				case ASURA_PROJECTOR_BLEND_MULTIPLY:
					Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_COLOUR); 
					xNeutralColour = Asura_RGBA(0xffffff,0);
					break;

				case ASURA_PROJECTOR_BLEND_MULTIPLYX2:
					Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_DOUBLECOLOUR); 
					xNeutralColour = Asura_RGBA(0x808080,0);
					break;

				case ASURA_PROJECTOR_BLEND_ADD:
					Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING); 
					xNeutralColour = Asura_RGBA(0x000000,0);
					break;

				default:
					Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_NORMAL); 
					xNeutralColour = Asura_RGBA(0x000000,0);
					break;
			}
		}
	}
	

	
	// Ideally, we'd set our border colour to the neutral colour, but we can't do this on 360 - only
	// black and white border colours are supported.
	Asura_Xbox360_Graphics::GetDevice()->SetSamplerState(1,D3DSAMP_BORDERCOLOR, 0x00000000);

	// Also use a border colour for the falloff, so when we've got no texture, it still clips off at the ends 
	// of the frustum
	Asura_Xbox360_Graphics::GetDevice()->SetSamplerState(2,D3DSAMP_BORDERCOLOR, 0x00000000);

	// Near plane equation - used to calc distance for falloff
	// Needs to be in view space.
	Asura_Vector_3 xDir = Asura_Vector_3(0,0,1);
	xDir = xDir * pxProjector->m_xOrient;	
	Asura_Vector_3 xPoint = pxProjector->m_xPos+ (xDir * pxProjector->m_fMinRange);

	// We need to transform this point and normal into view space
	// We'll also need the inverse of the view matrix soon.
	D3DXMATRIX xViewMatrix;
	D3DXMATRIX xViewMatrixInv;
	xViewMatrix = Asura_Xbox360_Graphics::GetViewMatrix();	
	D3DXMatrixInverse(&xViewMatrixInv,NULL,&xViewMatrix);

	// D3Dify it
	D3DXVECTOR3 xNormal, xNormalWS				= D3DXVECTOR3(xDir.x,   xDir.y,   xDir.z  );
	D3DXVECTOR3 xPointOnPlaneWS					= D3DXVECTOR3(xPoint.x, xPoint.y, xPoint.z);
	D3DXVECTOR4 xPointOnPlane;

	// To view space	
	D3DXVec3Transform(&xPointOnPlane, &xPointOnPlaneWS, &xViewMatrix);	
	D3DXVec3TransformNormal(&xNormal, &xNormalWS,		&xViewMatrix);

	// Get plane equation for near plane, in view space
	// ax+by+cz+d = 0,  d = - (ax+by+cz).
	float fD= - ((xNormal.x * xPointOnPlane.x)+(xNormal.y * xPointOnPlane.y)+(xNormal.z * xPointOnPlane.z));
	D3DXVECTOR4 xPlane;
	float fRange = pxProjector->m_fMaxRange - pxProjector->m_fMinRange;
	xPlane.x = xNormal.x / fRange;
	xPlane.y = xNormal.y / fRange;
	xPlane.z = xNormal.z / fRange;
	xPlane.w = fD / fRange;

	if (bUseStencil)
	{
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, TRUE );
		Asura_Xbox360_Graphics::GetDevice()->FlushHiZStencil( D3DFHZS_ASYNCHRONOUS );
	}

	D3DXMATRIX xProjMatrix, xOrient, xPos;
	D3DXMatrixIdentity(&xOrient);		

	if (pxProjector->m_eType == ASURA_PROJECTOR_TYPE_PERSPECTIVE)
	{
		// Perspective projection
		D3DXMatrixPerspectiveFovLH(&xProjMatrix, pxProjector->m_fFoV, 1.0f, pxProjector->m_fMinRange, pxProjector->m_fMaxRange);
	}
	else
	{
		// Parallel projection
		D3DXMatrixScaling(&xProjMatrix, 1.0f/pxProjector->m_fScale, 1.0f/pxProjector->m_fScale, 1.0f);		
	}
		

	// Orientation - Convert Asura matrix to D3D matrix
	for(int i=0;i<3;i++)
	{
		for(int j=0;j<3;j++)
		{
			xOrient.m[i][j] = pxProjector->m_xOrient.Element[i][j];				
		}			
	}

	// Translation (world space pos of projector)
	D3DXMatrixTranslation(&xPos,-pxProjector->m_xPos.x,-pxProjector->m_xPos.y,-pxProjector->m_xPos.z);
	D3DXMATRIX xScale,xOffset;
	
	// Scale and offset the texture from -1..+1 projection to 0..1 tex coords
	D3DXMatrixScaling(&xScale,0.5f / pxProjector->m_fUVAspect,0.5f,0);
	D3DXMatrixTranslation(&xOffset,0.5f,0.5f,0);
	xProjMatrix =  xPos * xOrient * xProjMatrix * xScale * xOffset;


	// Set shader params
	g_xProjectedTextureParameters.m_xProjectionMatrix		= xViewMatrixInv*xProjMatrix;
	g_xProjectedTextureParameters.m_xInvViewMatrix			= xViewMatrixInv;
	g_xProjectedTextureParameters.m_fCameraTanHalfVertFOV	= Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
	g_xProjectedTextureParameters.m_fCameraTanHalfHorizFOV  = g_xProjectedTextureParameters.m_fCameraTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();
	g_xProjectedTextureParameters.m_fTwoOverScreenWidth		= 2.0f / Asura_Viewport::GetWidth();
	g_xProjectedTextureParameters.m_fTwoOverScreenHeight	= 2.0f / Asura_Viewport::GetHeight();	
	g_xProjectedTextureParameters.m_pxNormalBuffer			= Asura_Xbox360_Graphics::GetNormalTexture();
	g_xProjectedTextureParameters.m_pxDepthMtlBuffer		= Asura_Xbox360_Graphics::GetDepthMtlTexture();

	if (pxProjector->m_bFakeLight)
	{					
		g_xProjectedTextureParameters.m_pxAlbedoBuffer		= Asura_Xbox360_Graphics::GetSceneResolveTexture();
	}
	else
	{
		g_xProjectedTextureParameters.m_pxAlbedoBuffer		= Asura_Xbox360_Graphics::GetAlbedoTexture();
	}

	g_xProjectedTextureParameters.m_xNearPlane              = xPlane;
	g_xProjectedTextureParameters.m_xNeutralColour.x		= xNeutralColour.m_fR;
	g_xProjectedTextureParameters.m_xNeutralColour.y		= xNeutralColour.m_fG;
	g_xProjectedTextureParameters.m_xNeutralColour.z		= xNeutralColour.m_fB;
	g_xProjectedTextureParameters.m_xNeutralColour.w		= xNeutralColour.m_fA;

	// (AAM - 07.01.2008) Roll a x2 from the pixel shader into here
	g_xProjectedTextureParameters.m_fDistanceFade			= fDistanceFade * 2.0f;

	g_xProjectedTextureParameters.m_xProjectorPosition.x    = pxProjector->m_xPos.x;
	g_xProjectedTextureParameters.m_xProjectorPosition.y    = pxProjector->m_xPos.y;
	g_xProjectedTextureParameters.m_xProjectorPosition.z    = pxProjector->m_xPos.z;
	g_xProjectedTextureParameters.m_xProjectorPosition.w    = 1;		
	g_xProjectedTextureParameters.m_xProjectionDirection.x  = pxProjector->m_xOrient.e13;
	g_xProjectedTextureParameters.m_xProjectionDirection.y  = pxProjector->m_xOrient.e23;
	g_xProjectedTextureParameters.m_xProjectionDirection.z  = pxProjector->m_xOrient.e33;
	g_xProjectedTextureParameters.m_xProjectionDirection.w  = 1;
	g_xProjectedTextureParameters.m_afLightColour[0]		= pxProjector->m_fIntensity * pxProjector->m_afRGB[0];
	g_xProjectedTextureParameters.m_afLightColour[1]		= pxProjector->m_fIntensity * pxProjector->m_afRGB[1];
	g_xProjectedTextureParameters.m_afLightColour[2]		= pxProjector->m_fIntensity * pxProjector->m_afRGB[2];

	
	g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset.x = pxProjector->m_fUScrollOfs;	
	g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset.y = pxProjector->m_fVScrollOfs;	
	g_xProjectedTextureParameters.m_xDetailTextureScaleAndOffset.z = pxProjector->m_fDetailTextureScale;		
	g_pxApplyProjectedTextureEffect->Begin(0,iEffectPass);

	g_xProjectedTextureParameters.Apply();



	// Render the frustum
	for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
	{
		if (Asura_Xbox360_Graphics::IsTiling2())
		{
			Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
			Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
		}
		g_pxApplyProjectedTextureEffect->Commit();
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );

		Asura_Xbox360_Graphics::GetDevice()->DrawIndexedPrimitiveUP(D3DPT_TRIANGLELIST,0,8,12, (void*)au16Indices, D3DFMT_INDEX16, (void *)pxProjector->m_axVertices ,12);

		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );
	}
	if (Asura_Xbox360_Graphics::IsTiling2())
	{
		Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
	}


	g_pxApplyProjectedTextureEffect->End();
	
	// Restore states
	if (bUseStencil)
	{
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, FALSE );		
	}

	if (bShadows)
	{
		// If we've been rendering with shadows, restore states, so we aren't doing the multiply by the dest alpha	
		D3DBLENDSTATE xBlendState;
		xBlendState.BlendOp = xBlendState.BlendOpAlpha = D3DBLENDOP_ADD;
		xBlendState.SrcBlend = xBlendState.SrcBlendAlpha = D3DBLEND_ONE;
		xBlendState.DestBlend = xBlendState.DestBlendAlpha = D3DBLEND_ONE;
		Asura_Xbox360_Graphics::GetDevice()->SetBlendState( 0, xBlendState );
	}
	
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
	Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );		


	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_DEFERRED_PROJECTORS);
}


void Asura_Projector_System::Platform_EndRender(void)
{
#if defined(ASURA_XBOX360_ENABLE_PIX)
	PIXEndNamedEvent();
#endif	//defined(ASURA_XBOX360_ENABLE_PIX)
}


#endif //#ifdef ASURA_TARGET_SM3
