///////////////////////////////////////////////////////////////////////////////
//
// File   : Asura_Platform_Render.cpp
// Author : Kevin Floyer-Lea
// Date   : 24-Jun-2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Rendering/Asura_Render.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_CharacterRenderer.h"
#include "Rendering/Asura_Xbox360_EdgeAntiAliasing.h"
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_EnvironmentRenderer.h"
#include "Rendering/Asura_Xbox360_Fog.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_HDR.h"
#include "Rendering/Asura_Xbox360_LineRenderer.h"
#include "Rendering/Asura_Xbox360_ObjectHierarchyRenderer.h"
#include "Rendering/Asura_Xbox360_ObjectRenderer.h"
#include "Rendering/Asura_Xbox360_QuadOverlayRenderer.h"
#include "Rendering/Asura_Xbox360_SkinRenderer.h"

//How close should the viewport max z be to prevent the player model clipping into stuff
const float fASURA_XBOX360_PLAYER_RENDER_FAR_Z = 0.001f;

/******************************************************************************/
/* Asura_Render Static Class Functions                                        */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_Initialise
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles Xbox360 specific renderer initilisation.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_Initialise()
{
	// Intialise the graphics...
	Asura_Xbox360_Graphics::Initialise();

	// Initialise the effects system before all the renderers (which will need effects)
	Asura_Xbox360_Effect_System::Initialise();

	// Initialise the specific renderers...
	Asura_Xbox360_EnvironmentRenderer::Initialise();
	Asura_Xbox360_CharacterRenderer::Initialise			();
	Asura_Xbox360_ObjectHierarchyRenderer::Initialise();
	Asura_Xbox360_ObjectRenderer::Initialise();
	Asura_Xbox360_SkinRenderer::Initialise();

	Asura_Xbox360_QuadOverlayRenderer::Initialise();
	Asura_Xbox360_LineRenderer::Initialise();

	Asura_Xbox360_Fog::Initialise();

	#if defined ASURA_USE_HDR
		Asura_Xbox360_HDR::Initialise();
	#endif

#ifdef ASURA_EDGE_ANTIALIASING
	Asura_Xbox360_EdgeAntiAliasing::Initialise();
#endif

}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_EarlyShutdown
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles Xbox360 specific early-shutdown code.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_EarlyShutdown()
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_LateShutdown
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles Xbox360 specific late-shutdown code.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_LateShutdown()
{
	// Shutdown the specific renderers...
	Asura_Xbox360_SkinRenderer::Shutdown();
	Asura_Xbox360_ObjectRenderer::Shutdown();
	Asura_Xbox360_ObjectHierarchyRenderer::Shutdown();
	Asura_Xbox360_CharacterRenderer::Shutdown();
	Asura_Xbox360_EnvironmentRenderer::Shutdown();

	Asura_Xbox360_QuadOverlayRenderer::Shutdown();
	Asura_Xbox360_LineRenderer::Shutdown();

	#if defined ASURA_USE_HDR
		Asura_Xbox360_HDR::Shutdown();
	#endif

	Asura_Xbox360_Fog::Shutdown();

	// Shutdown the effects system after all the renderers
	Asura_Xbox360_Effect_System::Shutdown();

	// Shutdown the graphics...
	Asura_Xbox360_Graphics::Shutdown();

#ifdef ASURA_EDGE_ANTIALIASING
	Asura_Xbox360_EdgeAntiAliasing::Shutdown();
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_Reset
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles the Xbox360 specific renderer reset.
//
// Return value : 
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_Reset()
{
	// MH test - I got an error "ERR[D3D]: The current thread doesn't own the primary D3D device"
	// while trying to destroy a resource Hopefully taking ownership of the device will fix this.
	SCOPED_DEVICE_OWNERSHIP();

	// Unbind all resources first
	Asura_Xbox360_Graphics::Reset();

	// Reset the renderers...
	Asura_Xbox360_SkinRenderer::Reset();
	Asura_Xbox360_ObjectHierarchyRenderer::Reset();
	Asura_Xbox360_ObjectRenderer::Reset();
	Asura_Xbox360_CharacterRenderer::Reset();
	Asura_Xbox360_EnvironmentRenderer::Reset();

	Asura_Xbox360_LineRenderer::Reset();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_ApplyCameraSettings
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Copies Asura_Camera settings through to the 360 APIs
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_ApplyCameraSettings()
{
	Asura_Xbox360_Graphics::ApplyCameraSettings();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_AllowFullScreenRendering
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Not relevant on the Xbox360.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_AllowFullScreenRendering()
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_ClearZBuffer
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Clears the current Xbox360 Z-buffer.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_ClearZBuffer()
{
	Asura_Xbox360_Graphics::ClearZBuffer();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_BeginFrame
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles the Xbox360 specific start-of-frame code.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_BeginFrame()
{
	Asura_Xbox360_Graphics::BeginFrame();
	Asura_Xbox360_SkinRenderer::ResetBoneBuffer();
#ifdef ASURA_EDGE_ANTIALIASING
	Asura_Xbox360_EdgeAntiAliasing::BeginFrame();
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_EndFrame
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles the Xbox360 specific end-of-frame code.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_EndFrame()
{
#ifdef ASURA_EDGE_ANTIALIASING
	Asura_Xbox360_EdgeAntiAliasing::BeginFrame();
#endif
	Asura_Xbox360_Graphics::EndFrame();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_FlushDisplay
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Handles the flushing of the Xbox360 display.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_FlushDisplay()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Render::Platform_BeginMRT()
//
// Begins MRT usage within the frame. Needed to pull this out of BeginFrame()
// so the CSM creation could come first
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_BeginMRT()
{
	Asura_Xbox360_Graphics::BeginTiling();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_EndMRT
// Author       : Andrew McDonald
// Date         : 26-Oct-2006
//
// Description  : Ends MRT usage during the render frame.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_EndMRT()
{
	Asura_Xbox360_Graphics::EndTiling();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_SetVideoMode
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Sets the specified video mode.
//
// Parameters   : The width, height and whether to enable antialiasing.
//
// Return value : "true" if successfully set, "false" if not.
////////////////////////////////////////////////////////////////////////////////

bool Asura_Render::Platform_SetVideoMode( int iWidth, int iHeight, bool bFsaa )
{
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_InitialiseSplashScreen
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Splash-screen initialisation code.
//
// Parameters   : The name of the splash-screen file to be loaded.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_InitialiseSplashScreen( const char* szName )
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_ShutdownSplashScreen
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Shuts down the previously opened splash-screen.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_ShutdownSplashScreen()
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_SplashScreenRender
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Renders the previously loaded splash-screen.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_SplashScreenRender()
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_RenderLine
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Renders a line in 3D space.
//
// Parameters   : The start and end positions of the line and it's colour.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_RenderLine( const Asura_Vector_3& xP1, const Asura_Vector_3& xP2, u_int uColour )
{
	Asura_Xbox360_LineRenderer::Render3DLine( xP1, xP2, uColour );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_Render2DLine
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Renders a line in 2D space.
//
// Parameters   : The start and end positions of the line and it's colour.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_Render2DLine( const Asura_Vector_2& xP1, const Asura_Vector_2& xP2, u_int uColour )
{
	Asura_Xbox360_LineRenderer::Render2DLine( xP1, xP2, uColour );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_RenderDynamicShadowsOntoEnvironment
// Author       : Kevin Floyer-Lea
// Date         : 24-Jun-2005
//
// Description  : Xbox360 specific shadow rendering.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Render::Platform_RenderDynamicShadowsOntoEnvironment()
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Render::Platform_ApplyToneMapping
// Author       : Andrew McDonald
// Date         : 13-Jul-2007
//
// Description  : Xbox360 specific HDR rendering.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

#if defined ASURA_USE_HDR

void Asura_Render::Platform_ApplyToneMapping()
{
	Asura_Xbox360_HDR::ApplyToneMapping();
}

#endif // defined ASURA_USE_HDR

#ifdef ASURA_TARGET_SM3
void Asura_Render::Platform_BeginSkyboxRendering(void)
{
	Asura_Xbox360_Graphics::s_bIsCurrentlyRenderingInSkybox = true;

	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		Asura_Xbox360_SkinRenderer::OwnRegisters();
	}
}
void Asura_Render::Platform_EndSkyboxRendering(void)
{
	Asura_Xbox360_Graphics::s_bIsCurrentlyRenderingInSkybox = false;

	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		Asura_Xbox360_SkinRenderer::DisownRegisters();
	}
}
#endif

#ifdef ASURA_EDGE_ANTIALIASING
void Asura_Render::Platform_SetCollectAAEdges( bool bCollect )
{
	Asura_Xbox360_EdgeAntiAliasing::SetCollectEdges( bCollect );
}
void Asura_Render::Platform_RenderEdgeAntiAliasing( void )
{
	Asura_Xbox360_EdgeAntiAliasing::Render();
}
#endif


// Temp. This is not named as a platform function...
#include "Rendering/Asura_ShadowBuffer.h"
void Asura_ShadowBuffer_System::PreAllocateBuffer( const Asura_Animation* pxAnim, BUFFERSIZE eBufferSize, float fScale )
{
}

void Asura_Render::Platform_PrepareToRenderFirstPersonMesh ()
{
	Asura_Xbox360_Graphics::SetViewportMinZ(0.0f);
	Asura_Xbox360_Graphics::SetViewportMaxZ(fASURA_XBOX360_PLAYER_RENDER_FAR_Z);
	Asura_Xbox360_Graphics::ApplyCameraSettings();
}

void Asura_Render::Platform_FinishRenderFirstPersonMesh ()
{
	Asura_Xbox360_Graphics::SetViewportMinZ(0.0f);
	Asura_Xbox360_Graphics::SetViewportMaxZ(1.0f);
	Asura_Xbox360_Graphics::ApplyCameraSettings();
}

//eof
