//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Rendering/Asura_Xbox360_CharacterRenderer.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Animation/Asura_TextureAnimation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Material_System.h"
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureFlags.h"
#include "Debugging/Asura_Assert.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#include "Shadows/Asura_Shadow_System.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Debugging/Asura_Platform_GPUProfile.h"

// Shader Include Files
#include "Smoothskin_MaterialResponse.h"
#include "Skin_Shadow.h"

// Misc
#include <fxl.h>
#include <xtl.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Remember to update the vertex size in Asura_Chunk_ResourceFile_Xbox360_Character.cpp
// if this changes

const D3DVERTEXELEMENT9 axVERTEX_DECLARATION[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION,     0 },
	{ 0, 12, D3DDECLTYPE_FLOAT16_2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     0 },
	{ 0, 16, D3DDECLTYPE_FLOAT16_2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,     1 },
	{ 0, 20, D3DDECLTYPE_FLOAT1,	D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD,	  2 },
	{ 0, 24, D3DDECLTYPE_UBYTE4,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDINDICES, 0 },
	{ 0, 28, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BLENDWEIGHT,  0 },
	{ 0, 36, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,       0 },
	{ 0, 40, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,      0 },
	{ 0, 44, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL,     0 },
	D3DDECL_END()
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Asura_TemplateHashTree< Asura_Xbox360_CharacterRenderer_Skin > g_xCharacterTree;

// Standard skinned shader
u_int	Asura_Xbox360_CharacterRenderer::s_uNumRendered	= 0;
Asura_Xbox360_Effect* Asura_Xbox360_CharacterRenderer::s_pxSkinEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_xRedSHHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xGreenSHHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xBlueSHHandle = 0;

static Asura_Xbox360_Effect_System::Handle g_xWorldViewProjHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWorldMatrixHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xBoneMatricesHandle = 0;

static Asura_Xbox360_Effect_System::Handle g_xUseDamageMapsHandle = 0;

Asura_Xbox360_MaterialResponse_Handles g_xMaterialResponseHandles;

// Skin shadow effect
static Asura_Xbox360_Effect* g_pxSkinShadowEffect = NULL;
static Asura_Xbox360_Effect_System::Handle g_xShadowWorldViewProjHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowBoneMatricesHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowFrustumNearPlaneHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xOneOverShadowFrustumDepthHandle = 0;

// Vertex declaration
static Direct3D_VertexDeclaration* g_pxCharacterVertexDeclaration = NULL;

/******************************************************************************/
/* Asura_Xbox360_CharacterRenderer_Skin Class Functions                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer_Skin::Asura_Xbox360_CharacterRenderer_Skin()
//
// Constructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_CharacterRenderer_Skin::Asura_Xbox360_CharacterRenderer_Skin()
	: m_pxStrips( NULL )
	, m_uNumStrips(0)
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer_Skin::~Asura_Xbox360_CharacterRenderer_Skin()
//
// Destructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_CharacterRenderer_Skin::~Asura_Xbox360_CharacterRenderer_Skin()
{
	m_xIndexBuffer.Destroy();
	m_xVertexBuffer.Destroy();

	delete[] m_pxStrips;

	delete [] m_pxBoneMaps;
}

/******************************************************************************/
/* Asura_Xbox360_CharacterRenderer Static Class Functions                     */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer::Initialise()
//
// Initialises the Xbox 360 character renderer
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_CharacterRenderer::Initialise()
{
	// Standard skin effect

	s_pxSkinEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory( g_auSmoothSkin_MaterialResponse);
	g_xMaterialResponseHandles.InitialiseFromEffect( *s_pxSkinEffect );

	g_xRedSHHandle = s_pxSkinEffect->GetParameterHandle( "g_xRedSH" );
	g_xGreenSHHandle = s_pxSkinEffect->GetParameterHandle( "g_xGreenSH" );
	g_xBlueSHHandle = s_pxSkinEffect->GetParameterHandle( "g_xBlueSH" );

	g_xWorldViewProjHandle = s_pxSkinEffect->GetParameterHandle( "g_xWorldViewProjectionMatrix" );
	g_xWorldMatrixHandle = s_pxSkinEffect->GetParameterHandle( "g_xWorldMatrix" );
	g_xBoneMatricesHandle = s_pxSkinEffect->GetParameterHandle( "g_axBoneMatrices" );

	g_xUseDamageMapsHandle = s_pxSkinEffect->GetParameterHandle( "g_bUseDamageMaps" );

	// Skin shadow effect
	g_pxSkinShadowEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auSkin_Shadow );
	g_xShadowWorldViewProjHandle = g_pxSkinShadowEffect->GetParameterHandle( "g_xWorldViewProj" );
	g_xShadowBoneMatricesHandle = g_pxSkinShadowEffect->GetParameterHandle( "g_axBoneMatrices" );
	g_xShadowFrustumNearPlaneHandle = g_pxSkinShadowEffect->GetParameterHandle( "g_fShadowFrustumNearPlane" );
	g_xOneOverShadowFrustumDepthHandle = g_pxSkinShadowEffect->GetParameterHandle( "g_fOneOverShadowFrustumDepth" );

	// Create vertex declaration
	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxCharacterVertexDeclaration );
	Asura_Assert( g_pxCharacterVertexDeclaration, "Failed to create character vertex declaration." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer::Reset()
//
// Resets the Xbox 360 character renderer. Called between level loads
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_CharacterRenderer::Reset()
{
	g_xCharacterTree.Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer::Shutdown()
//
// Shuts the Xbox 360 character renderer down
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_CharacterRenderer::Shutdown()
{
	Reset();

	Asura_Xbox360_Effect_System::Destroy( s_pxSkinEffect );
	s_pxSkinEffect = NULL;

	Asura_Xbox360_Effect_System::Destroy( g_pxSkinShadowEffect );
	g_pxSkinShadowEffect = NULL;

	g_pxCharacterVertexDeclaration->Release();
	g_pxCharacterVertexDeclaration = NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer::CreateCharacter()
//
// Creates a new Xbox 360 character. Called by the chunk loader
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_CharacterRenderer_Skin* Asura_Xbox360_CharacterRenderer::CreateCharacter( const Asura_Hash_ID uHash, const u_int uNumStrips )
{
#if !defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA) && defined(ASURA_ASSERT)
		const Asura_Xbox360_CharacterRenderer_Skin* const pxExistingCharacter = g_xCharacterTree.GetDataByHash( uHash );
		Asura_Assertf( !pxExistingCharacter, ("A character with this hash (0x%X) has already been created.", uHash) );
#endif	//!defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA) && defined(ASURA_ASSERT)

	Asura_Xbox360_CharacterRenderer_Skin* const pxNewCharacter = g_xCharacterTree.AddNode( uHash );
	if( !pxNewCharacter )
	{
#if !defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA)
		Asura_Assert( false, "Unable to create new character." );
#endif	//!defined(ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA)
		return NULL;
	}

	pxNewCharacter->m_pxStrips = new Asura_Xbox360_CharacterRenderer_Strip[ uNumStrips ];
	Asura_Assert( pxNewCharacter->m_pxStrips, "Unable to allocate character materials." );

	pxNewCharacter->m_uNumStrips = uNumStrips;

	return pxNewCharacter;
}

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Xbox360_CharacterRenderer::RenderCharacter(const Asura_Animation& xAnimation,
													  const Asura_Hierarchy_Skin& xSkin,
													  Asura_AMatrix4x4_InParam xTransform,
													  Asura_AMatrix4x4_InParam xInverseTransform,
													  const Asura_AMatrix4x4* const pxBones)
#else
void Asura_Xbox360_CharacterRenderer::RenderCharacter( const Asura_Animation& xAnimation,
													   const Asura_Hierarchy_Skin& xSkin,
													   const Asura_Vector_3& xPos,
													   const Asura_Matrix_3x3& xOri,
													   const Asura_Vector_3* const pxBonePos,
													   const Asura_Matrix_3x3* const pxBoneOri )
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
//	const float fAlpha = xAnimation.GetAlpha();
//	const int iTexIDOverride = xAnimation.GetAnimTexID();

	Asura_Xbox360_CharacterRenderer_Skin* const pxCharacter = GetCharacter( xSkin.m_uNameHashID );
	if( !pxCharacter ) return;

	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xFF, 0, 0xFF ), "Character %u - '%s'", ++s_uNumRendered, xSkin.Name.GetString() );
	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_SKINS);

#ifdef ASURA_USE_DAMAGE_MAPS	
	const bool bUseDamageMaps = xAnimation.IsUsingDamageMap();
#else
	const bool bUseDamageMaps = false;
#endif

	Asura_Xbox360_Effect *pxEffect = s_pxSkinEffect;

	const bool bTransparentPass = Asura_SortedRender::IsTransparentPass();


	// Set world and world-view-proj matrices
	{
		D3DXMATRIX xWorldViewProj;

#if defined(ASURA_HIERARCHY_QUAD)
		xTransform.StoreColsUnaligned(&xWorldViewProj._11);
#else
		// Rotation
		Asura_Mem::Set( &xWorldViewProj, 0, sizeof( xWorldViewProj ) );
		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xWorldViewProj.m[u][v] = xOri.Element[v][u];

			}
		}

		xWorldViewProj._44 = 1.0f;

		// Translation
		D3DXMATRIX xTemp;
		D3DXMatrixTranslation( &xTemp, xPos.x, xPos.y, xPos.z );
		xWorldViewProj *= xTemp;
#endif	// defined(ASURA_HIERARCHY_QUAD)

		pxEffect->SetMatrix( g_xWorldMatrixHandle, xWorldViewProj );

		// View-projection
		xWorldViewProj *= Asura_Xbox360_Graphics::GetViewProjectionMatrix();

		pxEffect->SetMatrix( g_xWorldViewProjHandle, xWorldViewProj );
	}

	// Set bone matrices
	if (pxCharacter->GetNumBoneMaps() == 0)
	{
		D3DXMATRIX axBoneMatrix[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
		Asura_Mem::Set( axBoneMatrix, 0, sizeof( axBoneMatrix ) );

		for( int i = 0; i < xSkin.NumberOfBones; ++i )
		{
#if defined(ASURA_HIERARCHY_QUAD)
			pxBones[i].StoreColsUnaligned(&axBoneMatrix[i]._11);
#else
			for( u_int u = 0; u < 3; ++u )
			{
				for( u_int v = 0; v < 3; ++v )
				{
					axBoneMatrix[i].m[u][v] = pxBoneOri[i].Element[v][u];
				}
			}

			axBoneMatrix[i]._44 = 1.0f;

			D3DXMATRIX xTranslation;
			D3DXMatrixTranslation( &xTranslation, pxBonePos[i].x, pxBonePos[i].y, pxBonePos[i].z );

			axBoneMatrix[i] *= xTranslation;
#endif	// defined(ASURA_HIERARCHY_QUAD)

#ifdef ASURA_USE_DAMAGE_MAPS
			if (bUseDamageMaps)
			{
				axBoneMatrix[i]._33 = xAnimation.GetBoneDamageMapValue(i);
			}
#endif
		}

		pxEffect->GetInterface()->SetMatrixArrayF4x4( g_xBoneMatricesHandle, axBoneMatrix[0], xSkin.NumberOfBones );

		pxEffect->SetBool( g_xUseDamageMapsHandle, bUseDamageMaps );
	}

	pxCharacter->m_xIndexBuffer.Bind();
	pxCharacter->m_xVertexBuffer.Bind();
	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxCharacterVertexDeclaration );

	// Render one strip for each material



	if (bTransparentPass==false)
	{	
		// No fogging if we're doing the opaque pass
		Asura_Xbox360_MaterialResponse_System::SetFogParameters( false, NULL, g_xMaterialResponseHandles, *s_pxSkinEffect );
	} 

	const u_int uTechnique = bTransparentPass ? 1 : 0;
	for (int iZPrePassMode = !bTransparentPass ; iZPrePassMode < 2; ++iZPrePassMode)
	{
		u_int uStartIndex = 0;
		pxEffect->Begin( uTechnique );
		for( u_int u = 0; u < pxCharacter->m_uNumStrips; ++u )
		{

	#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
			if (xAnimation.GetSkinElementVisibility(pxCharacter->m_pxStrips[u].m_iElementID))
	#endif
			{			

				if (pxCharacter->GetNumBoneMaps() > 0)
				{
					Asura_Xbox360_CharacterRenderer_Strip* pxStrip = pxCharacter->m_pxStrips + u;

					D3DXMATRIX axBoneMatrix[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
					Asura_Mem::Set( axBoneMatrix, 0, sizeof( axBoneMatrix ) );
					const Asura_Xbox360_CharacterRenderer_Skin::BoneMap &xBoneMap = pxCharacter->GetBoneMap(pxStrip->m_uBoneMap);

					for( u_int i = 0; i < xBoneMap.m_uSize; ++i )
					{
						u_int uBone = xBoneMap.m_puEntries[i];
	#if defined(ASURA_HIERARCHY_QUAD)
						pxBones[uBone].StoreColsUnaligned(&axBoneMatrix[i]._11);
	#else
						for( u_int u = 0; u < 3; ++u )
						{
							for( u_int v = 0; v < 3; ++v )
							{
								axBoneMatrix[i].m[u][v] = pxBoneOri[uBone].Element[v][u];
							}
						}

						axBoneMatrix[i]._44 = 1.0f;

						D3DXMATRIX xTranslation;
						D3DXMatrixTranslation( &xTranslation, pxBonePos[uBone].x, pxBonePos[uBone].y, pxBonePos[uBone].z );

						axBoneMatrix[i] *= xTranslation;
	#endif	// defined(ASURA_HIERARCHY_QUAD)

	#ifdef ASURA_USE_DAMAGE_MAPS
						if (bUseDamageMaps)
						{
							axBoneMatrix[i]._33 = xAnimation.GetBoneDamageMapValue(uBone);
						}
	#endif
					}

					pxEffect->GetInterface()->SetMatrixArrayF4x4( g_xBoneMatricesHandle, axBoneMatrix[0], xBoneMap.m_uSize );

					pxEffect->SetBool( g_xUseDamageMapsHandle, bUseDamageMaps );
				}


				const Asura_Hash_ID uMaterialHash1 = pxCharacter->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
				const Asura_Hash_ID uMaterialHash2 = pxCharacter->m_pxStrips[u].m_auMaterialResponseHashIDs[1];

				const Asura_MaterialResponse* const pxMaterial1
					= Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

				bool bVisible = true;

				if( pxMaterial1 )
				{
					if( pxMaterial1->GetDontRender() || (iZPrePassMode == 0 && !pxMaterial1->GetZPrePass()) )
					{
						bVisible = false;
					}

					const bool bTransparentStrip = pxMaterial1->IsTransparent();
					if( bTransparentStrip != bTransparentPass )
					{
						bVisible = false;
					}
				}


				if( bVisible )
				{
					PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xC0, 0, 0xC0 ), "Strip %u (primary material 0x%X)", u , uMaterialHash1 );

					// Set material data
					{
						const Asura_MaterialResponse* const pxMaterial2 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash2 );

						Asura_Xbox360_MaterialResponse_System::SetEffectParameters( pxMaterial1, pxMaterial2, &xAnimation, g_xMaterialResponseHandles, *s_pxSkinEffect );
						Asura_Xbox360_MaterialResponse_System::SetBlendStates( pxMaterial1 );
						Asura_Xbox360_MaterialResponse_System::SetSHParameters( pxMaterial1, g_xMaterialResponseHandles, *s_pxSkinEffect );

						if (bTransparentPass)
						{
							Asura_Xbox360_MaterialResponse_System::SetFogParameters( true, pxMaterial1, g_xMaterialResponseHandles, *s_pxSkinEffect );
						}

						if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback())
						{
							Asura_MaterialResponse_System::GetMaterialProjectContextCallback() (pxMaterial1->GetProjectContext(),
								pxMaterial2 ? pxMaterial2->GetProjectContext() : ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
						}

					}

					pxEffect->Commit();

					if (iZPrePassMode == 0)
					{
						Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
						Asura_Xbox360_Graphics::GetDevice()->SetPixelShader( NULL);
						Asura_Xbox360_Graphics::GetDevice()->SetRenderState(D3DRS_COLORWRITEENABLE, 0);
					}


					Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0, uStartIndex, pxCharacter->m_pxStrips[u].m_uNumIndices );
					PIXEndNamedEvent(); // Material

					if (iZPrePassMode == 0)
					{
						Asura_Xbox360_MaterialResponse_System::SetBlendStates( pxMaterial1 );
						Asura_Xbox360_Graphics::GetDevice()->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE | D3DCOLORWRITEENABLE_ALPHA);
					}

				}
			}

			uStartIndex += pxCharacter->m_pxStrips[u].m_uNumIndices;
		}
		pxEffect->End();
	}


	Asura_Render::AddToTrianglesRendered( pxCharacter->m_uNumTriangles );
	
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

	Asura_Xbox360_Graphics::ResetDepthBias();
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );

	PIXEndNamedEvent(); // Character
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_SKINS);
}

#if defined(ASURA_HIERARCHY_QUAD)
void Asura_Xbox360_CharacterRenderer::RenderShadow(const Asura_Animation& xAnimation,
												   const Asura_Hierarchy_Skin& xSkin,
												   const Asura_Shadow& xShadow,
												   Asura_AMatrix4x4_InParam xTransform,
												   const Asura_AMatrix4x4* const pxBones )
#else
void Asura_Xbox360_CharacterRenderer::RenderShadow( const Asura_Animation& xAnimation,
													const Asura_Hierarchy_Skin& xSkin,
													const Asura_Shadow& xShadow,
													const Asura_Vector_3& xPos,
													const Asura_Matrix_3x3& xOri,
													const Asura_Vector_3* const pxBonePos,
													const Asura_Matrix_3x3* const pxBoneOri )
#endif	// defined(ASURA_HIERARCHY_QUAD)
{
	Asura_Xbox360_CharacterRenderer_Skin* const pxCharacter = GetCharacter( xSkin.m_uNameHashID );
	if( !pxCharacter ) return;

	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xFF, 0, 0xFF ), "Character Shadow %s", xSkin.Name.GetString() );

	// Set common parameters

	{
		// Character orientation

		D3DXMATRIX xWorldViewProj;

#if defined(ASURA_HIERARCHY_QUAD)
		xTransform.StoreColsUnaligned(&xWorldViewProj._11);
#else
		Asura_Mem::Set( &xWorldViewProj, 0, sizeof( D3DXMATRIX ) );

		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xWorldViewProj.m[u][v] = xOri.Element[v][u];
			}
		}

		xWorldViewProj._44 = 1.0f;

		// Character position

		D3DXMATRIX xTemp;
		D3DXMatrixTranslation( &xTemp, xPos.x, xPos.y, xPos.z );
		xWorldViewProj *= xTemp;
#endif	// defined(ASURA_HIERARCHY_QUAD)

		// Light position

		const Asura_Vector_3& xLightPos = xShadow.GetLightPosition();
		D3DXMATRIX xLightTransform;
		D3DXMatrixTranslation( &xLightTransform, -xLightPos.x, -xLightPos.y, -xLightPos.z );
		xWorldViewProj *= xLightTransform;

		// Light orientation

		Asura_Matrix_3x3 xLightOri;
		xShadow.GetLightOrientation(xLightOri);
		Asura_Mem::Set( &xLightTransform, 0, sizeof( D3DXMATRIX ) );

		for( u_int u = 0; u < 3; ++u )
		{
			for( u_int v = 0; v < 3; ++v )
			{
				xLightTransform.m[u][v] = xLightOri.Element[u][v];

				if( v == 1 )
				{
					xLightTransform.m[u][v] = -xLightTransform.m[u][v];
				}
			}
		}

		xLightTransform._44 = 1.0f;

		xWorldViewProj *= xLightTransform;

		// Projection

		const float fVerticalFOV = Asura_Maths::DegToRad( xShadow.GetFrustum().GetVerticalFOVInDegrees() );
		D3DXMATRIX	xProjectionTransform;
		D3DXMatrixPerspectiveFovLH( &xProjectionTransform, fVerticalFOV, 1.0f, xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane() );

		xWorldViewProj *= xProjectionTransform;

		g_pxSkinShadowEffect->GetInterface()->SetMatrixF( g_xShadowWorldViewProjHandle, xWorldViewProj );
	}

	if (pxCharacter->GetNumBoneMaps() == 0)
	{
#if defined(ASURA_HIERARCHY_QUAD)
		const float*	pfBoneData	= reinterpret_cast<const float*>(&pxBones[0]);

		// Bones are already stored in a 4x4 matrix - no need to transform anything...
		g_pxSkinShadowEffect->GetInterface()->SetMatrixArrayF4x4( g_xShadowBoneMatricesHandle, pfBoneData, xSkin.NumberOfBones );
#else
		D3DXMATRIX axBoneMatrix[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
		Asura_Mem::Set( axBoneMatrix, 0, sizeof( axBoneMatrix ) );

		for( int i = 0; i < xSkin.NumberOfBones; ++i )
		{
			for( u_int u = 0; u < 3; ++u )
			{
				for( u_int v = 0; v < 3; ++v )
				{
					axBoneMatrix[i].m[u][v] = pxBoneOri[i].Element[v][u];
				}
			}

			axBoneMatrix[i]._44 = 1.0f;

			D3DXMATRIX xTranslation;
			D3DXMatrixTranslation( &xTranslation, pxBonePos[i].x, pxBonePos[i].y, pxBonePos[i].z );

			axBoneMatrix[i] *= xTranslation;
		}

		g_pxSkinShadowEffect->GetInterface()->SetMatrixArrayF4x4( g_xShadowBoneMatricesHandle, axBoneMatrix[0], xSkin.NumberOfBones );
#endif	// defined(ASURA_HIERARCHY_QUAD)
	}

	// Frustum parameters

	const float fShadowFrustumNearPlane = xShadow.GetFrustum().GetNearPlane();
	g_pxSkinShadowEffect->GetInterface()->SetFloat( g_xShadowFrustumNearPlaneHandle, fShadowFrustumNearPlane );

	const float fOneOverShadowFrustumDepth = 1.0f / (xShadow.GetFrustum().GetFarPlane() - fShadowFrustumNearPlane);
	g_pxSkinShadowEffect->GetInterface()->SetFloat( g_xOneOverShadowFrustumDepthHandle, fOneOverShadowFrustumDepth );

	pxCharacter->m_xIndexBuffer.Bind();
	pxCharacter->m_xVertexBuffer.Bind();
	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxCharacterVertexDeclaration );

	// Render one strip for each material

	u_int uStartIndex = 0;

	g_pxSkinShadowEffect->Begin();
	g_pxSkinShadowEffect->Commit();

	for( u_int u = 0; u < pxCharacter->m_uNumStrips; ++u )
	{
		if (pxCharacter->GetNumBoneMaps() > 0)
		{
			Asura_Xbox360_CharacterRenderer_Strip* pxStrip = pxCharacter->m_pxStrips + u;

			D3DXMATRIX axBoneMatrix[ ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];
			Asura_Mem::Set( axBoneMatrix, 0, sizeof( axBoneMatrix ) );
			const Asura_Xbox360_CharacterRenderer_Skin::BoneMap &xBoneMap = pxCharacter->GetBoneMap(pxStrip->m_uBoneMap);


			for( u_int i = 0; i < xBoneMap.m_uSize; ++i )
			{
				u_int uBone = xBoneMap.m_puEntries[i];
#if defined(ASURA_HIERARCHY_QUAD)
				Asura_Mem::Copy(axBoneMatrix[i], pxBones + uBone, sizeof(float) * 16);
#else
				for( u_int u = 0; u < 3; ++u )
				{
					for( u_int v = 0; v < 3; ++v )
					{
						axBoneMatrix[i].m[u][v] = pxBoneOri[uBone].Element[v][u];
					}
				}

				axBoneMatrix[i]._44 = 1.0f;

				D3DXMATRIX xTranslation;
				D3DXMatrixTranslation( &xTranslation, pxBonePos[uBone].x, pxBonePos[uBone].y, pxBonePos[uBone].z );

				axBoneMatrix[i] *= xTranslation;

#endif
			}

			g_pxSkinShadowEffect->GetInterface()->SetMatrixArrayF4x4( g_xShadowBoneMatricesHandle, axBoneMatrix[0], xBoneMap.m_uSize );
			g_pxSkinShadowEffect->Commit();
		}

#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		if (xAnimation.GetSkinElementVisibility(pxCharacter->m_pxStrips[u].m_iElementID))
#endif
		{
			const Asura_Hash_ID uHash = pxCharacter->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
			const Asura_MaterialResponse* const pxMaterial =
				Asura_MaterialResponse_System::GetMaterialByHashID( uHash );

			if( !pxMaterial || (pxMaterial->GetRenderType() != Asura_MaterialResponse::RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW) )
			{
				SCOPEDPIXEVENT(( 0xFFC000C0, "Strip %u", u ));

				Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0,
					uStartIndex, pxCharacter->m_pxStrips[u].m_uNumIndices );
			}
		}
		
		uStartIndex += pxCharacter->m_pxStrips[u].m_uNumIndices;
	}

	g_pxSkinShadowEffect->End();

	Asura_Render::AddToTriPassesRendered( pxCharacter->m_uNumTriangles );

	PIXEndNamedEvent(); // Character Shadow
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer::GetCharacter()
//
// Retrieves the data associated with a given character
//
// Parameters:		uHash - Hash ID of character's name
// Return value:	Pointer to the data
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_CharacterRenderer_Skin* Asura_Xbox360_CharacterRenderer::GetCharacter( const Asura_Hash_ID uHash )
{
	Asura_Xbox360_CharacterRenderer_Skin* const pxSkin = g_xCharacterTree.GetDataByHash( uHash );
	Asura_Assert( pxSkin, "Failed to get character data." );

	return pxSkin;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_CharacterRenderer::IsCharacterOpaque()
//
// Queries whether a character is entirely opaque
//
// Parameters:		uHash - Hash ID of character name
// Return value:	true iff all strips are opaque
//
////////////////////////////////////////////////////////////////////////////////

bool Asura_Xbox360_CharacterRenderer::IsCharacterOpaque( const Asura_Hash_ID uHash )
{
	Asura_Xbox360_CharacterRenderer_Skin* const pxSkin = GetCharacter( uHash );
	if( !pxSkin ) return false;

	for( u_int u = 0; u < pxSkin->m_uNumStrips; ++u )
	{
		const Asura_Hash_ID uMaterialResponseHashID = pxSkin->m_pxStrips[u].m_auMaterialResponseHashIDs[0];
		const Asura_MaterialResponse* const pxMaterial
			= Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialResponseHashID );

		// Decal-type materials count as opaque

		if( pxMaterial && pxMaterial->IsTransparent() ) return false;
	}

	return true;
}

Direct3D_VertexDeclaration* Asura_Xbox360_CharacterRenderer::GetVertexDeclaration()
{
	return g_pxCharacterVertexDeclaration;
}