//
// Asura_Xbox360_EnvironmentRenderer.cpp
//
// Created by KJFL - 22 June 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Rendering/Asura_Xbox360_EnvironmentRenderer.h"

// Asura Platform-independent
#include "Animation/Asura_TextureAnimation.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_Timers.h"
#include "Light/Asura_GI_Solution.h"
#include "Light/Asura_Photon.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Environment_Renderer.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "Visibility/Asura_Visibility_System.h"

// Asura Xbox 360
#ifdef ASURA_EDGE_ANTIALIASING
#include "Rendering/Asura_Xbox360_EdgeAntiAliasing.h"
#endif
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_IndexBuffer.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Rendering/Asura_Xbox360_VertexBuffer.h"
#include "Rendering/Asura_Xbox360_Fog.h"
#include "Debugging/Asura_Platform_GPUProfile.h"

// Shader Include Files
#include "Env_MaterialResponse.h"
#include "Env_Water.h"
#include "Env_Shadow.h"
#include "Env_ZPass.h"
#include "Env_ZPassAABB.h"

// Misc
#include <fxl.h>
#include <xtl.h>

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct AABBVertex
{
	float x, y, z;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const AABBVertex axCUBE_VERTICES[] =
{
	{ 0.0f, 0.0f, 0.0f },
	{ 0.0f, 0.0f, 1.0f },
	{ 0.0f, 1.0f, 0.0f },
	{ 0.0f, 1.0f, 1.0f },
	{ 1.0f, 0.0f, 0.0f },
	{ 1.0f, 0.0f, 1.0f },
	{ 1.0f, 1.0f, 0.0f },
	{ 1.0f, 1.0f, 1.0f },
};

const u_short axCUBE_INDICES[] =
{
	0, 1, 2,
	1, 3, 2,

	1, 5, 3,
	5, 7, 3,

	4, 6, 7,
	7, 5, 4,

	0, 2, 6,
	6, 4, 0,

	2, 3, 7,
	7, 6, 2,

	0, 5, 1,
	0, 4, 5,
};

// Remember to update the vertex size in Asura_Chunk_ResourceFile_Xbox360_Environment.cpp
// if this changes

const Direct3D_VertexElement axVERTEX_DECLARATION[] =
{
	{ 0,  0, D3DDECLTYPE_FLOAT3,    D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
	{ 0, 12, D3DDECLTYPE_D3DCOLOR,  D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR,    0},
	{ 0, 16, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_NORMAL,   0},
	{ 0, 20, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,  0},
	{ 0, 24, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 0},
	{ 0, 28, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TANGENT,  1},
	{ 0, 32, D3DDECLTYPE_DEC3N,     D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_BINORMAL, 1},
	{ 0, 36, D3DDECLTYPE_FLOAT16_2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
	{ 0, 40, D3DDECLTYPE_FLOAT16_2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 1},
	{ 0, 44, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 2},
	{ 0, 52, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 3},
	{ 0, 60, D3DDECLTYPE_FLOAT16_4, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 4},
	D3DDECL_END()
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

// Strip rendering data

static Asura_Xbox360_Effect* g_pxEnvPrelightingEffect = NULL;
static Asura_Xbox360_Effect* g_pxEnvWaterEffect = NULL;
static Asura_Xbox360_Effect* g_pxEnvShadowEffect = NULL;

static Asura_Xbox360_MaterialResponse_Handles g_xMaterialResponseHandles;

static Asura_Xbox360_Effect_System::Handle g_xPrelightingColourFilterHandle = 0;

// For the water materials...
static Asura_Xbox360_Effect_System::Handle g_xWaterUseTransparencyMapHandle =0;
static Asura_Xbox360_Effect_System::Handle g_xWaterDualNormalMapHandle = 0;

static Asura_Xbox360_Effect_System::Handle g_xWaterFogScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterVolumetricDepthScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterRippleScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterDistortionScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterTwoOverScreenWidthHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterTwoOverScreenHeightHandle = 0;

static Asura_Xbox360_Effect_System::Handle g_xWaterReflectionColourHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterRefractionColourHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterFogColourHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterUVScrollHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterUVScroll2Handle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterWavesXHandle	= 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterWavesZHandle	= 0;

static Asura_Xbox360_Effect_System::Handle g_xWaterDepthTextureHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterBackBufferTextureHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterTransparencyMapHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterNormalMapHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterNormalMap2Handle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterReflectionTextureHandle = 0;

static Asura_Xbox360_Effect_System::Handle g_xWaterForwardFogEnabledHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterForwardFogColourHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterForwardFogParamsHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xWaterForwardFogTextureHandle = 0;


static Asura_Xbox360_Effect_System::Handle g_xShadowFrustumNearPlaneHandle;
static Asura_Xbox360_Effect_System::Handle g_xShadowOneOverShadowFrustumDepthHandle;



static Direct3D_VertexDeclaration* g_pxEnvVertexDeclaration = NULL;

 // Strip depth data

static Asura_Xbox360_Effect* g_pxEnvDepthEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_xZWriteHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xTextureSamplerHandle = 0;

// AABB rendering data

static Asura_Xbox360_Effect* g_pxAABBEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_xAABBScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xAABBOffsetHandle = 0;

static Asura_Xbox360_IndexBuffer g_xAABBIndexBuffer;
static Asura_Xbox360_VertexBuffer g_xAABBVertexBuffer;

static Direct3D_VertexDeclaration* g_pxAABBVertexDeclaration;

/******************************************************************************/
/* Asura_Xbox360_EnvironmentRenderer Static Class Variables                   */
/******************************************************************************/

u_int										Asura_Xbox360_EnvironmentRenderer::s_uNumberOfStrips;
Asura_Xbox360_EnvironmentRenderer_Strip* 	Asura_Xbox360_EnvironmentRenderer::s_pxStrips;

void*										Asura_Xbox360_EnvironmentRenderer::s_pVertexData;
Asura_Xbox360_VertexBuffer*					Asura_Xbox360_EnvironmentRenderer::s_pxVBs;
Asura_Xbox360_IndexBuffer					Asura_Xbox360_EnvironmentRenderer::s_xIB;

Asura_Xbox360_IndexBuffer					Asura_Xbox360_EnvironmentRenderer::s_xAntiAliasEdges;

int											Asura_Xbox360_EnvironmentRenderer::s_aiQueuedWaterStrips[ASURA_MAX_VISIBLE_WATER_STRIPS];		
int											Asura_Xbox360_EnvironmentRenderer::s_iNumQueuedWaterStrips = 0;		


/******************************************************************************/
/* Asura_Xbox360_EnvironmentRenderer Static Class Functions                   */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_EnvironmentRenderer::Initialise()
//
// Initialises the Xbox 360 environment renderer
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_EnvironmentRenderer::Initialise()
{
	s_uNumberOfStrips = 0;
	s_pxStrips = 0;
	s_pVertexData = 0;
	s_pxVBs = 0;

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxEnvVertexDeclaration );
	Asura_Assert( g_pxEnvVertexDeclaration, "Failed to create environment vertex declaration." );

	g_pxEnvPrelightingEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auEnv_MaterialResponse);	

	g_xMaterialResponseHandles.InitialiseFromEffect( *g_pxEnvPrelightingEffect );

	g_xPrelightingColourFilterHandle = g_pxEnvPrelightingEffect->GetParameterHandle( "g_xPrelightingColourFilter" );

	// Now get the shader params for water materials
	g_pxEnvWaterEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auEnv_Water);	

	g_xWaterUseTransparencyMapHandle			= g_pxEnvWaterEffect->GetParameterHandle( "g_bUseTransparencyMap" );
	g_xWaterDualNormalMapHandle					= g_pxEnvWaterEffect->GetParameterHandle( "g_bDualNormalMap" );

	g_xWaterFogScaleHandle						= g_pxEnvWaterEffect->GetParameterHandle( "g_fWaterFogScale" );
	g_xWaterVolumetricDepthScaleHandle			= g_pxEnvWaterEffect->GetParameterHandle( "g_fVolumetricDepthScale" );
	g_xWaterRippleScaleHandle					= g_pxEnvWaterEffect->GetParameterHandle( "g_fRippleScale" );
	g_xWaterDistortionScaleHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_fDistortionScale" );
	g_xWaterTwoOverScreenWidthHandle			= g_pxEnvWaterEffect->GetParameterHandle( "g_fTwoOverScreenWidth" );
	g_xWaterTwoOverScreenHeightHandle			= g_pxEnvWaterEffect->GetParameterHandle( "g_fTwoOverScreenHeight" );

	g_xWaterReflectionColourHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xWaterReflectionColour" );
	g_xWaterRefractionColourHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xWaterRefractionColour" );
	g_xWaterFogColourHandle						= g_pxEnvWaterEffect->GetParameterHandle( "g_xWaterFogColour" );
	g_xWaterUVScrollHandle						= g_pxEnvWaterEffect->GetParameterHandle( "g_xUVScroll" );
	g_xWaterUVScroll2Handle						= g_pxEnvWaterEffect->GetParameterHandle( "g_xUVScroll2" );
	g_xWaterWavesXHandle						= g_pxEnvWaterEffect->GetParameterHandle( "g_xWavesX" );
	g_xWaterWavesZHandle						= g_pxEnvWaterEffect->GetParameterHandle( "g_xWavesZ" );

	g_xWaterDepthTextureHandle					= g_pxEnvWaterEffect->GetParameterHandle( "g_xDepthTextureSampler" );
	g_xWaterBackBufferTextureHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xBackBufferTextureSampler" );
	g_xWaterTransparencyMapHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xTransparencyMapSampler" );
	g_xWaterNormalMapHandle						= g_pxEnvWaterEffect->GetParameterHandle( "g_xNormalMapSampler" );
	g_xWaterNormalMap2Handle					= g_pxEnvWaterEffect->GetParameterHandle( "g_xNormalMap2Sampler" );
	g_xWaterReflectionTextureHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xReflectionTextureSampler" );

	g_xWaterForwardFogEnabledHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_bFogEnabled" );
	g_xWaterForwardFogColourHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xFogColour" );
	g_xWaterForwardFogParamsHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xFogParams" );
	g_xWaterForwardFogTextureHandle				= g_pxEnvWaterEffect->GetParameterHandle( "g_xFogTexture" );


	g_pxEnvShadowEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auEnv_Shadow);

	g_xShadowFrustumNearPlaneHandle	= g_pxEnvShadowEffect->GetParameterHandle( "g_fShadowFrustumNearPlaneDistance" );
	g_xShadowOneOverShadowFrustumDepthHandle = g_pxEnvShadowEffect->GetParameterHandle( "g_fOneOverShadowFrustumPlaneSeparation" );


	// Get parameter handles specific to the dynamic lighting effect


	// Load visibility effects
	{
		g_pxEnvDepthEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auEnv_ZPass);

		g_xZWriteHandle = g_pxEnvDepthEffect->GetParameterHandle( "g_bZBufferWrite" );
		g_xTextureSamplerHandle = g_pxEnvDepthEffect->GetParameterHandle( "TextureSampler" );
	}

	CreateAABBBuffers();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_EnvironmentRenderer::CreateAABBBuffers()
//
// Initialises the effect and buffers for rendering AABB z-passes
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_EnvironmentRenderer::CreateAABBBuffers()
{
	g_pxAABBEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auEnv_ZPassAABB);

	g_xAABBScaleHandle = g_pxAABBEffect->GetParameterHandle( "g_xScale" );
	g_xAABBOffsetHandle = g_pxAABBEffect->GetParameterHandle( "g_xOffset" );

	void* pData;

	// Create vertex declaration

	const Direct3D_VertexElement axVERTEX_DECLARATION[] =
	{
		{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
		D3DDECL_END()
	};

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxAABBVertexDeclaration );
	Asura_Assert( g_pxAABBVertexDeclaration, "Failed to create AABB vertex declaration." );

	// Create vertex buffer

	g_xAABBVertexBuffer.CreateStatic( sizeof( AABBVertex ), 8 );

	pData = g_xAABBVertexBuffer.Lock();
	Asura_Assert( pData, "Failed to lock AABB vertex buffer." );

	memcpy( pData, axCUBE_VERTICES, sizeof( axCUBE_VERTICES ) );
	g_xAABBVertexBuffer.Unlock();

	// Create index buffer

	g_xAABBIndexBuffer.CreateStatic( sizeof( u_short ) * 36 );

	pData = g_xAABBIndexBuffer.Lock();
	Asura_Assert( pData, "Failed to lock AABB index buffer." );

	memcpy( pData, axCUBE_INDICES, sizeof( axCUBE_INDICES ) );
	g_xAABBIndexBuffer.Unlock();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_EnvironmentRenderer::Reset()
//
// Resets the Xbox 360 environment renderer; called between levels
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_EnvironmentRenderer::Reset()
{
	delete[] s_pxVBs;
	s_pxVBs = NULL;

	Asura_Xbox360_MemoryManagement::FreePhysical(Asura_Xbox360_MemoryManagement::MT_READ_WRITE_COMBINE, s_pVertexData, XALLOC_PHYSICAL_ALIGNMENT_DEFAULT);
	s_pVertexData = NULL;

	s_xIB.Destroy();

	s_xAntiAliasEdges.Destroy();

	delete[] s_pxStrips;
	s_pxStrips = NULL;

	s_uNumberOfStrips = 0;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_EnvironmentRenderer::Shutdown()
//
// Shuts down the Xbox 360 environment renderer
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_EnvironmentRenderer::Shutdown()
{
	Reset();

	Asura_Xbox360_Effect_System::Destroy( g_pxEnvPrelightingEffect );
	Asura_Xbox360_Effect_System::Destroy( g_pxEnvDepthEffect );
	Asura_Xbox360_Effect_System::Destroy( g_pxAABBEffect );

	g_pxEnvPrelightingEffect = NULL;
	g_pxEnvDepthEffect = NULL;
	g_pxAABBEffect = NULL;

	g_xAABBVertexBuffer.Destroy();
	g_xAABBIndexBuffer.Destroy();

	g_pxEnvVertexDeclaration->Release();
	g_pxAABBVertexDeclaration->Release();

	g_pxEnvVertexDeclaration = NULL;
	g_pxAABBVertexDeclaration = NULL;
}

void Asura_Xbox360_EnvironmentRenderer::Render( const bool bTransparentStrips )
{
	if( s_uNumberOfStrips == 0 ) return;

	SCOPEDPIXEVENT(( 0xFFFF0000, "Environment (%s)",
		bTransparentStrips ? "Transparent Strips" : "Opaque Strips" ));

	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_ENVIRONMENT);

	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	Asura_Render::ApplyCameraSettings();

#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM
	if( Asura_Visibility_System::ShouldRenderInXRayMode() || bTransparentStrips==false )
#else
	if (bTransparentStrips==false)
#endif
	{	
		// No fogging if we're doing the opaque pass
		Asura_Xbox360_MaterialResponse_System::SetFogParameters( false, NULL, g_xMaterialResponseHandles, *g_pxEnvPrelightingEffect );
	}

	pxDevice->SetVertexDeclaration( g_pxEnvVertexDeclaration );

	s_xIB.Bind();

	// Set render states
	{
		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestDitheringMode( true );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );

		// Leave wireframe as it was set elsewhere
	}

	const u_int uTechnique = bTransparentStrips ? 1 : 0;
	g_pxEnvPrelightingEffect->Begin( uTechnique );

	if( !bTransparentStrips )
	{
		// [3/25/2008 AS] Temp fix for mantis#50905 until AM can have a look at this.
		// Seems what is happening is g_pxEnvPrelightingEffect is getting an old texture hanging around
		// which upon commit causes it to crash, clearing all textures seems to fix the issue but
		// parhaps there is a better solution (clear only on level close/load?
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xAlbedo, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xNormalMap, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xSpecularMap, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xTransparencyMap, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xSelfIlluminationMap, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xEnvironmentMap, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xReflectivityMask, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xParallaxMap, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xAlbedo2, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xNormalMap2, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xSpecularMap2, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xTransparencyMap2, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xSelfIlluminationMap2, ASURA_TEXTURE_NONE );
		g_pxEnvPrelightingEffect->SetTexture( g_xMaterialResponseHandles.m_xParallaxMap2, ASURA_TEXTURE_NONE );

		// Commit the effect to get the vertex shader in place
		g_pxEnvPrelightingEffect->Commit();

		// Clamp the materials and fog textures, but wrap the artist-selected ones

		for( u_int u = 0; u < 2; ++u )
		{
			pxDevice->SetSamplerState_Inline( u, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
			pxDevice->SetSamplerState_Inline( u, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
		}

		for( u_int u = 2; u <= 19; ++u )
		{
			pxDevice->SetSamplerState_Inline( u, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP );
			pxDevice->SetSamplerState_Inline( u, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP );
		}
	}

#if defined(USE_PIX)
	u_int											uNumStripsRendered	= 0;
#endif	//defined(USE_PIX)
	const Asura_Xbox360_EnvironmentRenderer_Strip*	pxStrip	= s_pxStrips;

	// Ensure that our strip visibility data is up-to-date...
	Asura_Environment_Renderer::UpdateVisibilityOfStrips();

	// Pre-fetch the first strip to attempt to prevent cache misses...
	Asura_Mem::PreFetchData(pxStrip, sizeof(*pxStrip));

	for( u_int u = 0; u < s_uNumberOfStrips; ++u, pxStrip++ )
	{
		// Pre-fetch the next strip to attempt to prevent cache misses...
		Asura_Mem::PreFetchData(pxStrip + 1, sizeof(*pxStrip));

		if(!Asura_Environment_Renderer::IsStripVisibleFromCurrentCamera(u)) continue;

		const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( pxStrip->m_auMaterialResponseHashIDs[0] );
		const Asura_MaterialResponse* const pxMaterial2 = Asura_MaterialResponse_System::GetMaterialByHashID( pxStrip->m_auMaterialResponseHashIDs[1] );

		if( pxMaterial1 )
		{
			if( pxMaterial1->GetDontRender() ) continue;

			if( bTransparentStrips )
			{
				if( !pxMaterial1->IsTransparent() ) continue;
			}
			else
			{
				if( pxMaterial1->IsTransparent() ) continue;
			}

			if (pxMaterial1->GetSpecialFXMode()==Asura_MaterialResponse::SPECIALFX_MODE_WATER)
			{
				QueueWaterStrip(u);
				continue;
			}

			Asura_RGBA xPrelightingColour = pxMaterial1->GetAmbientColour();
			xPrelightingColour *= pxMaterial1->GetAmbientContribution();

			PIXBeginNamedEvent( 0xFF800000, "Strip #%u (%u)", u, ++uNumStripsRendered );

			// Set prelighting colour parameters
			if( bTransparentStrips )
			{
				const D3DXVECTOR4 xPrelightingColourVector( xPrelightingColour.m_fR, xPrelightingColour.m_fG, xPrelightingColour.m_fB, 0.0f );
				g_pxEnvPrelightingEffect->SetVector( g_xPrelightingColourFilterHandle, xPrelightingColourVector );
			}
			else
			{
				pxDevice->SetVertexShaderConstantF( 0, &xPrelightingColour.m_fR, 1 );
			}
			Asura_Xbox360_MaterialResponse_System::SetBlendStates( pxMaterial1 );

			// For transparent env strips, grab an SH at the centre of the bounding box and
			// use that for lighting

			if( bTransparentStrips )
			{
				Asura_Xbox360_MaterialResponse_System::SetEffectParameters( pxMaterial1,
					pxMaterial2, NULL, g_xMaterialResponseHandles, *g_pxEnvPrelightingEffect );

				Asura_SphericalHarmonics_Order2_RGB& xSH =
					Asura_GI_Solution::s_xDynamicSHForCurrentRender;

				const Asura_Vector_3 xStripCentre = pxStrip->m_xBoundingBox.CalculateCentre();

				Asura_GI_Solution::GetDynamicIrradiance( xSH, xStripCentre );

				// Static SH will come from the per-vertex values
				Asura_GI_Solution::s_xStaticSHForCurrentRender.Zero();

				Asura_Xbox360_MaterialResponse_System::SetSHParameters( pxMaterial1, g_xMaterialResponseHandles, *g_pxEnvPrelightingEffect );

#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM
				if( !Asura_Visibility_System::ShouldRenderInXRayMode() )
				{
#endif
				// Also set up the fog
				Asura_Xbox360_MaterialResponse_System::SetFogParameters( true, pxMaterial1, g_xMaterialResponseHandles, *g_pxEnvPrelightingEffect );
#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM
				}
#endif

				g_pxEnvPrelightingEffect->Commit();
			}

			if (Asura_MaterialResponse_System::GetMaterialProjectContextCallback())
			{
				Asura_MaterialResponse_System::GetMaterialProjectContextCallback() (pxMaterial1->GetProjectContext(),
					pxMaterial2 ? pxMaterial2->GetProjectContext() : ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);
			}
		}

		// Just one IB, but a separate VB per strip, since they max out at 64 MB on 360
		s_pxVBs[u].Bind();

		if( !bTransparentStrips )
		{
			// Replace the pixel shader from the effect with the faster version. TODO: remove the
			// effect entirely

			const Asura_Hash_ID uCombinedHash =
				Asura_MaterialResponse_System::CombineShaderHashes( pxMaterial1, pxMaterial2,
				false );

			Direct3D_PixelShader* const pxPS =
				Asura_Xbox360_MaterialResponse_System::GetPixelShader( uCombinedHash );

			if( pxPS )
			{
				pxDevice->SetPixelShader( pxPS );
			}

			pxDevice->GpuOwnPixelShaderConstantF( 0, 8 );

			Asura_Xbox360_MaterialResponse_System::SetMaterialConstants( pxMaterial1, pxMaterial2,
				pxStrip->m_auMaterialResponseHashIDs[0], NULL );
		}

		pxDevice->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0, pxStrip->m_uStartIndex,
			pxStrip->m_uNumberOfIndices );

		Asura_Render::AddToTrianglesRendered( pxStrip->m_uNumberOfTriangles );

#ifdef ASURA_EDGE_ANTIALIASING
		// Add Anti-Alias edges to list to render later
		if ( !Asura_Xbox360_EnvironmentRenderer::s_xAntiAliasEdges.IsEmpty() && pxStrip->m_uAntiAliasEdgeIdxNum > 0 )
		{
			Asura_Xbox360_EdgeAntiAliasing::AddEdgeList( &Asura_Xbox360_EnvironmentRenderer::s_xAntiAliasEdges, 
				&Asura_Xbox360_EnvironmentRenderer::s_pxVBs[u], 
				Asura_Xbox360_EnvironmentRenderer::GetVD(),
				Asura_Xbox360_Graphics::GetViewProjectionMatrix(), pxStrip->m_uAntiAliasEdgeIdxStart, pxStrip->m_uAntiAliasEdgeIdxNum );
		}
#endif

		if( !bTransparentStrips )
		{
			pxDevice->GpuDisownPixelShaderConstantF( 0, 8 );
		}

		if(pxMaterial1)
		{
			PIXEndNamedEvent();
		}
	}

	g_pxEnvPrelightingEffect->End();

	Asura_Xbox360_Graphics::ResetDepthBias();

	Asura_Xbox360_Graphics::SetTranslucencyModeMRT( 2, ASURA_TRANSLUCENCY_OFF );
	pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );		
	pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL );
	pxDevice->SetRenderState( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL );


	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_ENVIRONMENT);
}


void Asura_Xbox360_EnvironmentRenderer::RenderToShadow(float fNear, float fFar, bool bEnvMap)
{

	SCOPEDPIXEVENT(( 0, "Render Environment to Shadow" ));

	if( s_uNumberOfStrips == 0 ) return;

	if( s_uNumberOfStrips == 0 || !g_pxEnvShadowEffect)
	{
		return;
	}
	g_pxEnvShadowEffect->SetFloat( g_xShadowFrustumNearPlaneHandle, fNear);
	g_pxEnvShadowEffect->SetFloat( g_xShadowOneOverShadowFrustumDepthHandle , 1.0f / (fFar - fNear));


	// Set up
	{
		SCOPEDPIXEVENT(( 0, "Set Up" ));

		Asura_Xbox360_Graphics::ApplyCameraSettings();

		g_pxEnvShadowEffect->Begin(bEnvMap,0);
		g_pxEnvShadowEffect->Commit();

		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxEnvVertexDeclaration );

		s_xIB.Bind();

		Asura_Render::CurrentStates.RequestAlphaTest( true );
		Asura_Render::CurrentStates.RequestAlphaTest( false );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF);
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
		Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
		Asura_Render::CurrentStates.RequestDitheringMode( true );
		Asura_Render::CurrentStates.RequestDynamicLights( false );
		Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
		Asura_Render::CurrentStates.RequestZBufferEnabled( false );
		Asura_Render::CurrentStates.RequestZBufferEnabled( true );

		// Leave wireframe as it was set elsewhere
	}

	// make a rough bounding box around the light, for trivial rejection of env strips
	Asura_Bounding_Box xLightAABB;
	xLightAABB.SetToPoint(*Asura_Camera::GetPosition());
	xLightAABB.Grow( fFar );
	const Asura_Xbox360_EnvironmentRenderer_Strip*	pxStrip	= s_pxStrips;

	// Pre-fetch the first strip to attempt to prevent cache misses...
	Asura_Mem::PreFetchData(pxStrip, sizeof(*pxStrip));

	for( u_int u = 0; u < s_uNumberOfStrips; ++u, pxStrip++ )
	{
		// Pre-fetch the next strip to attempt to prevent cache misses...
		Asura_Mem::PreFetchData(pxStrip + 1, sizeof(*pxStrip));

		if (!xLightAABB.Intersects(pxStrip->m_xBoundingBox)) continue;

		if(Asura_Camera::IsAABBInFrustum( pxStrip->m_xBoundingBox ) )
		{
			const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( pxStrip->m_auMaterialResponseHashIDs[0] );

			// Strip out unwanted materials
			if( pxMaterial1 )
			{
				if( pxMaterial1->GetDontCastShadow() || pxMaterial1->IsTransparent() ) continue;

				const bool bSpecialRender =	pxMaterial1->GetSpecialFXMode() != Asura_MaterialResponse::SPECIALFX_MODE_NONE;

				if( bSpecialRender ) continue;
			}

			SCOPEDPIXEVENT(( 0xFF800000, "Strip %u", u ));

			s_pxVBs[u].Bind();

			// Render
			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0,
				pxStrip->m_uStartIndex, pxStrip->m_uNumberOfIndices );

			Asura_Render::AddToTrianglesRendered( pxStrip->m_uNumberOfTriangles );
		}
	}

	// Clean up
	{
		SCOPEDPIXEVENT(( 0, "Clean Up" ));

		g_pxEnvShadowEffect->End();
	}
}

void Asura_Xbox360_EnvironmentRenderer::RenderToShadowCubemap_Initialise( float fNear, float fFar )
{
	g_pxEnvShadowEffect->SetFloat( g_xShadowFrustumNearPlaneHandle, fNear);
	g_pxEnvShadowEffect->SetFloat( g_xShadowOneOverShadowFrustumDepthHandle , 1.0f / (fFar - fNear));

	
	Asura_Render::CurrentStates.RequestAlphaTest( true );
	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
	Asura_Render::CurrentStates.RequestDitheringMode( true );
	Asura_Render::CurrentStates.RequestDynamicLights( false );
	Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );

	
	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxEnvVertexDeclaration );

	s_xIB.Bind();
}

void Asura_Xbox360_EnvironmentRenderer::RenderToShadowCubemap( u_char* pxStripVisibilities )
{

	SCOPEDPIXEVENT(( 0, "Render Environment to Shadow" ));

	if( s_uNumberOfStrips == 0 ) return;

	if( s_uNumberOfStrips == 0 || !g_pxEnvShadowEffect)
	{
		return;
	}
	

	// Set up
	{
		SCOPEDPIXEVENT(( 0, "Set Up" ));

		Asura_Xbox360_Graphics::ApplyCameraSettings();

		g_pxEnvShadowEffect->Begin(true,0);
		g_pxEnvShadowEffect->Commit();
	}

	const Asura_Xbox360_EnvironmentRenderer_Strip*	pxStrip	= s_pxStrips;

	for( u_int u = 0; u < s_uNumberOfStrips; ++u, pxStrip++ )
	{
		if ( !pxStripVisibilities[u] ) continue;

		if(Asura_Camera::IsAABBInFrustum( pxStrip->m_xBoundingBox ) )
		{
			SCOPEDPIXEVENT(( 0xFF800000, "Strip %u", u ));

			s_pxVBs[u].Bind();

			// Render
			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0,
				pxStrip->m_uStartIndex, pxStrip->m_uNumberOfIndices );

			Asura_Render::AddToTrianglesRendered( pxStrip->m_uNumberOfTriangles );
		}
	}

	// Clean up
	{
		SCOPEDPIXEVENT(( 0, "Clean Up" ));

		g_pxEnvShadowEffect->End();
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Xbox360_EnvironmentRenderer::QueueWaterStrip
// Author       : Dave Reed
// Date         : 25-May-2007
//
// Description  : Queues a water strip for rendering after the deferred lighting passes
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_EnvironmentRenderer::QueueWaterStrip(int iStripIdx)
{
	if (s_iNumQueuedWaterStrips<ASURA_MAX_VISIBLE_WATER_STRIPS)
	{
		s_aiQueuedWaterStrips[s_iNumQueuedWaterStrips] = iStripIdx;
		s_iNumQueuedWaterStrips++;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : Asura_Direct3D_EnvironmentRenderer::RenderWater
// Author       : Dave Reed
// Date         : 17-May-2007
//
// Description  : Called after the deferred lighting passes to render any visible water
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_EnvironmentRenderer::RenderWater(void)
{
	if (s_iNumQueuedWaterStrips==0)
	{
		return;
	}

	if (!Asura_Render::IsSystemEnabled(ASURA_RENDER_SYSTEM_WATER))
	{
		s_iNumQueuedWaterStrips = 0;
		return;
	}

	SCOPEDPIXEVENT(( 0xFF0000FF, "Environment (Water Strips)" ));

	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_WATER);

	// Clear the dest alpha to 0 first
	float fWidth =  (float)Asura_Viewport::GetWidth();		
	float fHeight = (float)Asura_Viewport::GetHeight();			
	Asura_Overlay_Quad xQuad;
	xQuad.SetRectXY(0,0, fWidth, fHeight);
	xQuad.Colour = 0x00ffffff;
	xQuad.TextureIndex = ASURA_TEXTURE_NONE;
	xQuad.TranslucencyMode = ASURA_TRANSLUCENCY_OFF;	
	Asura_Render::CurrentStates.RequestAlphaTest( false , 0x00);
	Asura_Render::CurrentStates.RequestStencilEnabled( false);
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALPHA);
	xQuad.Render(false);
	Asura_Render::FlushDisplay();
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState(D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE|D3DCOLORWRITEENABLE_ALPHA);

	// PFX and resolve have their own profile sections
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_WATER);

	// Water ripples - esentially a heightmap - gets rendered into the the dest. alpha channel
	Asura_PFX_System::Render(ASURA_PFX_RENDER_WATER_DISTORTER);

	// Resolve back buffer here! 
	Asura_Xbox360_Graphics::Resolve( Asura_Xbox360_Graphics::GetSceneResolveTexture() );

	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_WATER);

	Asura_Render::ApplyCameraSettings();

	g_pxEnvWaterEffect->Begin();

	// Set render states
	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );		
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
	Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );

	// (AAM - 13.09.2007) Force z-buffer state to be changed
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );
	Asura_Render::CurrentStates.RequestZBufferEnabled( true );

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxEnvVertexDeclaration ); 

	s_xIB.Bind();

	g_pxEnvWaterEffect->SetFloat( g_xWaterTwoOverScreenWidthHandle  , 2.0f / Asura_Viewport::GetWidth());
	g_pxEnvWaterEffect->SetFloat( g_xWaterTwoOverScreenHeightHandle , 2.0f / Asura_Viewport::GetHeight());

	g_pxEnvWaterEffect->SetTexture( g_xWaterDepthTextureHandle, Asura_Xbox360_Graphics::GetDepthMtlTexture());
	g_pxEnvWaterEffect->SetTexture( g_xWaterBackBufferTextureHandle, Asura_Xbox360_Graphics::GetSceneResolveTexture());	


	for(int i=0;i<s_iNumQueuedWaterStrips;i++)
	{
		int iStripIdx = s_aiQueuedWaterStrips[i];

		Asura_Xbox360_EnvironmentRenderer_Strip *pxStrip = &s_pxStrips[iStripIdx];

		const Asura_MaterialResponse* pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( pxStrip->m_auMaterialResponseHashIDs[0] );
		const Asura_MaterialResponse* pxMaterial2 = Asura_MaterialResponse_System::GetMaterialByHashID( pxStrip->m_auMaterialResponseHashIDs[1] );

		if( pxMaterial1 )
		{

			g_pxEnvWaterEffect->SetTexture( g_xWaterTransparencyMapHandle,		Asura_Xbox360_TextureManagement::GetTexture(pxMaterial1->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ))->GetD3DTexture());
			g_pxEnvWaterEffect->SetTexture( g_xWaterNormalMapHandle,			Asura_Xbox360_TextureManagement::GetTexture(pxMaterial1->GetTexture( Asura_MaterialResponse::NORMAL_MAP ))->GetD3DTexture());
			g_pxEnvWaterEffect->SetTexture( g_xWaterNormalMap2Handle,			Asura_Xbox360_TextureManagement::GetTexture(pxMaterial2 ? pxMaterial2->GetTexture( Asura_MaterialResponse::NORMAL_MAP ) : ASURA_TEXTURE_NONE)->GetD3DTexture());
			g_pxEnvWaterEffect->SetTexture( g_xWaterReflectionTextureHandle,	Asura_Xbox360_TextureManagement::GetTexture(pxMaterial1->GetTexture( Asura_MaterialResponse::ENVIRONMENT_MAP ))->GetD3DTexture());

			g_pxEnvWaterEffect->SetBool(g_xWaterDualNormalMapHandle,     (pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::NORMAL_MAP ) != ASURA_TEXTURE_NONE));
			g_pxEnvWaterEffect->SetBool(g_xWaterUseTransparencyMapHandle, (pxMaterial1->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ) != ASURA_TEXTURE_NONE));

			g_pxEnvWaterEffect->SetFloat(g_xWaterFogScaleHandle,			  pxMaterial1->GetWaterFogStrength());
			g_pxEnvWaterEffect->SetFloat(g_xWaterVolumetricDepthScaleHandle , 5.0f/pxMaterial1->GetWaterEdgeSoftness());
			g_pxEnvWaterEffect->SetFloat(g_xWaterRippleScaleHandle,			  pxMaterial1->GetWaterRippleScale());
			g_pxEnvWaterEffect->SetFloat(g_xWaterDistortionScaleHandle,		  pxMaterial1->GetWaterDistortionScale());


			float afTemp[4];
			afTemp[0] = pxMaterial1->GetWaterReflectionColour().m_fR;
			afTemp[1] = pxMaterial1->GetWaterReflectionColour().m_fG;
			afTemp[2] = pxMaterial1->GetWaterReflectionColour().m_fB;
			afTemp[3] = pxMaterial1->GetWaterReflectionColour().m_fA;			
			g_pxEnvWaterEffect->SetVector(g_xWaterReflectionColourHandle, afTemp);

			afTemp[0] = pxMaterial1->GetWaterRefractionColour().m_fR;
			afTemp[1] = pxMaterial1->GetWaterRefractionColour().m_fG;
			afTemp[2] = pxMaterial1->GetWaterRefractionColour().m_fB;
			afTemp[3] = pxMaterial1->GetWaterRefractionColour().m_fA;			
			g_pxEnvWaterEffect->SetVector(g_xWaterRefractionColourHandle, afTemp);

			afTemp[0] = pxMaterial1->GetWaterFogColour().m_fR;
			afTemp[1] = pxMaterial1->GetWaterFogColour().m_fG;
			afTemp[2] = pxMaterial1->GetWaterFogColour().m_fB;
			afTemp[3] = pxMaterial1->GetWaterFogColour().m_fA;			
			g_pxEnvWaterEffect->SetVector(g_xWaterFogColourHandle, afTemp);

			afTemp[0] = pxMaterial1->GetWaterUVControl(0) * Asura_Timers::GetLocalScaledTime();
			afTemp[1] = pxMaterial1->GetWaterUVControl(1) * Asura_Timers::GetLocalScaledTime();
			afTemp[2] = pxMaterial1->GetWaterUVControl(2);			
			g_pxEnvWaterEffect->SetVector(g_xWaterUVScrollHandle, afTemp);

			afTemp[0] = pxMaterial1->GetWaterUVControl(3) * Asura_Timers::GetLocalScaledTime();
			afTemp[1] = pxMaterial1->GetWaterUVControl(4) * Asura_Timers::GetLocalScaledTime();
			afTemp[2] = pxMaterial1->GetWaterUVControl(5);			
			g_pxEnvWaterEffect->SetVector(g_xWaterUVScroll2Handle, afTemp);

			afTemp[0] = pxMaterial1->GetWaterWaveControl(0);
			afTemp[1] = pxMaterial1->GetWaterWaveControl(1);
			afTemp[2] = pxMaterial1->GetWaterWaveControl(2) * Asura_Timers::GetLocalScaledTime();
			g_pxEnvWaterEffect->SetVector(g_xWaterWavesXHandle, afTemp);

			afTemp[0] = pxMaterial1->GetWaterWaveControl(3);
			afTemp[1] = pxMaterial1->GetWaterWaveControl(4);
			afTemp[2] = pxMaterial1->GetWaterWaveControl(5) * Asura_Timers::GetLocalScaledTime();
			g_pxEnvWaterEffect->SetVector(g_xWaterWavesZHandle, afTemp);

			s_pxVBs[iStripIdx].Bind();

			g_pxEnvWaterEffect->SetBool(g_xWaterForwardFogEnabledHandle, Asura_Fog::IsFogEnabled());

			if (Asura_Fog::IsFogEnabled())
			{
				const FXLVECTOR xFogData = {Asura_Fog::GetFogNearPlane(),
					Asura_Fog::GetOneOverFogPlaneSeparation(), 0.0f, 0.0f};

				g_pxEnvWaterEffect->SetVector(g_xWaterForwardFogParamsHandle, xFogData);

				const Asura_RGBA& xFogColour = Asura_Fog::GetFogColour();
				const FXLVECTOR xFogColourVec = {xFogColour.m_fR, xFogColour.m_fG, xFogColour.m_fB,
					xFogColour.m_fA};

				g_pxEnvWaterEffect->SetVector(g_xWaterFogColourHandle, xFogColourVec);

				g_pxEnvWaterEffect->SetTexture(g_xWaterForwardFogTextureHandle,
					Asura_Xbox360_Fog::GetRenderTexture());
			}			

			for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
			{
				if (Asura_Xbox360_Graphics::IsTiling2())
				{
					Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
					Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
				}

				g_pxEnvWaterEffect->Commit();

				Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0,
					pxStrip->m_uStartIndex, pxStrip->m_uNumberOfIndices );
			}

			if (Asura_Fog::IsFogEnabled())
			{
				g_pxEnvWaterEffect->SetTexture(g_xWaterForwardFogTextureHandle, NULL );
			}
		}
	}

	if (Asura_Xbox360_Graphics::IsTiling2())
	{
		Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
	}

	g_pxEnvWaterEffect->End();
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );		
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );

	s_iNumQueuedWaterStrips=0;

	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_WATER);
}

Direct3D_VertexDeclaration* Asura_Xbox360_EnvironmentRenderer::GetVD()
{
	return g_pxEnvVertexDeclaration;
}

void Asura_Xbox360_EnvironmentRenderer::RenderGeometry()
{
	// Renders all strip geometry, assuming render state has been set up by the caller. No
	// visibility culling at the moment, not sure how best to implement that yet

	SCOPEDPIXEVENT(( 0, "Environment Geometry" ));

	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();
	if( !pxDevice ) return;

	pxDevice->SetVertexDeclaration( g_pxEnvVertexDeclaration );

	s_xIB.Bind();

	for( u_int u = 0; u < s_uNumberOfStrips; ++u )
	{
		const Asura_Xbox360_EnvironmentRenderer_Strip& xStrip = s_pxStrips[u];

		s_pxVBs[u].Bind();

		pxDevice->DrawIndexedVertices( D3DPT_TRIANGLESTRIP, 0, xStrip.m_uStartIndex,
			xStrip.m_uNumberOfIndices );
	}
}

u_int Asura_Environment_Renderer::Platform_GetNumberOfStrips( void )
{
	return Asura_Xbox360_EnvironmentRenderer::s_uNumberOfStrips;
}

void Asura_Environment_Renderer::Platform_PrefetchStrip(u_int uStrip)
{
	const Asura_Xbox360_EnvironmentRenderer_Strip&	xStrip	= Asura_Xbox360_EnvironmentRenderer::s_pxStrips[uStrip];

	Asura_Mem::PreFetchData(&xStrip, sizeof(xStrip));
}

const Asura_Bounding_Box& Asura_Environment_Renderer::Platform_GetBoundingBox( u_int uStrip )
{
	return Asura_Xbox360_EnvironmentRenderer::s_pxStrips[uStrip].m_xBoundingBox;
}

const Asura_MaterialResponse* Asura_Environment_Renderer::Platform_GetMaterial( u_int uStrip, u_int uPass )
{
	const Asura_Hash_ID uMaterialHash = Asura_Xbox360_EnvironmentRenderer::s_pxStrips[uStrip].m_auMaterialResponseHashIDs[ uPass ];
	const Asura_MaterialResponse* pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash );

	return pxMaterial;
}

u_int Asura_Environment_Renderer::Platform_GetModuleIndex( u_int uStrip )
{
	return Asura_Xbox360_EnvironmentRenderer::s_pxStrips[uStrip].m_iModuleIndex;
}


/// End of file. ///
