//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_HDR.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_HDR

#include "Rendering/Asura_Xbox360_HDR.h"
#include "XGraphics.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_HalfFloat.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Graphics.h"

// Shader Include Files
#include "HDR.h"

// Misc
#include <xtl.h>

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

//#define DEBUG_CHECK_HDR_LUMINOSTY

#ifdef _DEBUG
//#define DEBUG_ALLOW_FRAMEBUFFER_GRAB
#endif

// Working render targets for downsampling
static Direct3D_Texture* g_pxWorkingTexture_RGB = NULL;
static Direct3D_Texture* g_pxWorkingTexture_R   = NULL;
static Direct3D_Surface* g_pxWorkingRT_RGB      = NULL;
static Direct3D_Surface* g_pxWorkingRT_R        = NULL;

// Holds the current luminance value. TODO: splitscreen - use 4 texels here?
static Direct3D_Texture* g_pxCurrentLuminanceTexture = NULL;

static Asura_Xbox360_Effect* g_pxHDREffect                            = NULL;
static Asura_Xbox360_Effect_System::Handle g_xDownsizeSamplerHandle   = 0;
static Asura_Xbox360_Effect_System::Handle g_xInputSceneSamplerHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xExtractLumUVScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xAdaptionFactorHandle    = 0;
static Asura_Xbox360_Effect_System::Handle g_xMinLuminanceHandle      = 0;
static Asura_Xbox360_Effect_System::Handle g_xMaxLuminanceHandle      = 0;
static Asura_Xbox360_Effect_System::Handle g_xKeyValueHandle          = 0;
static Asura_Xbox360_Effect_System::Handle g_xWhitePointSquaredHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xScaleParamsHandle       = 0;
static Asura_Xbox360_Effect_System::Handle g_xDynamicExposureHandle   = 0;
static Asura_Xbox360_Effect_System::Handle g_xFixedLuminanceHandle    = 0;
static Asura_Xbox360_Effect_System::Handle g_xXboxGammaFudgeHandle    = 0;

// Bits for dealing with the measurement region
const Direct3D_VertexElement axREGION_VERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	{ 0, 8, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
	D3DDECL_END()
};

static Direct3D_VertexDeclaration* g_pxRegionVD = NULL;

DVSVAR bool g_bShowRegion = false;
DVSVAR bool g_bShowHDRValues = false;
#ifdef DEBUG_ALLOW_FRAMEBUFFER_GRAB
DVSVAR bool g_bGrabBuffer = false;
#endif

DVSVAR float g_fXboxGammaFudge = 2.0f;
/******************************************************************************/
/* Asura_Xbox360_HDR Static Class Functions                                   */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_HDR::Initialise()
//
// Initialises the HDR rendering system on 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_HDR::Initialise()
{
	// Create all render targets in the same place

	const D3DSURFACE_PARAMETERS& xHDRSurfaceParams =
		Asura_Xbox360_Graphics::GetSurfaceParams( Asura_Xbox360_Graphics::SP_HDR );

	Asura_Xbox360_Graphics::GetDevice()->CreateRenderTarget
	(
		Asura_Xbox360_Graphics::GetScreenWidth() / 4,
		Asura_Xbox360_Graphics::GetScreenHeight() / 4,
		D3DFMT_A16B16G16R16_EDRAM,
		D3DMULTISAMPLE_NONE,
		0,
		FALSE,
		&g_pxWorkingRT_RGB,
		&xHDRSurfaceParams
	);

	Asura_Xbox360_Graphics::GetDevice()->CreateRenderTarget
	(
		64,
		64,
		D3DFMT_R32F,
		D3DMULTISAMPLE_NONE,
		0,
		FALSE,
		&g_pxWorkingRT_R,
		&xHDRSurfaceParams
	);

	// Create textures into which we'll resolve those RTs

	Asura_Xbox360_Graphics::GetDevice()->CreateTexture(
		Asura_Xbox360_Graphics::GetScreenWidth() / 4,
		Asura_Xbox360_Graphics::GetScreenHeight() / 4,
		1,
		0,
		D3DFMT_A16B16G16R16F_EXPAND,
		D3DPOOL_DEFAULT,
		&g_pxWorkingTexture_RGB,
		NULL
	);

	Asura_Xbox360_Graphics::GetDevice()->CreateTexture(
		64,
		64,
		1,
		0,
		D3DFMT_R32F,
		D3DPOOL_DEFAULT,
		&g_pxWorkingTexture_R,
		NULL
	);

	// Create 1x1 texture to hold the current adaptation value
	Asura_Xbox360_Graphics::GetDevice()->CreateTexture(
		1,
		1,
		1,
		0,
		D3DFMT_R32F,
		D3DPOOL_DEFAULT,
		&g_pxCurrentLuminanceTexture,
		NULL
	);

	Asura_Assert( g_pxWorkingRT_R && g_pxWorkingRT_RGB
			   && g_pxWorkingTexture_R && g_pxWorkingTexture_RGB
			   && g_pxCurrentLuminanceTexture,
			   "Failed to create HDR buffers." );

	if( g_pxCurrentLuminanceTexture )
	{
		// Initialise the current luminance texture

		D3DLOCKED_RECT xLockedRect;
		g_pxCurrentLuminanceTexture->LockRect( 0, &xLockedRect, NULL, 0 );

		*static_cast< float* >( xLockedRect.pBits ) = 0.18f;
		g_pxCurrentLuminanceTexture->UnlockRect(0);
	}

	// Load effect

	g_pxHDREffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auHDR);

	if( g_pxHDREffect )
	{
		g_xDownsizeSamplerHandle   = g_pxHDREffect->GetParameterHandle( "g_xDownsizeSampler" );
		g_xInputSceneSamplerHandle = g_pxHDREffect->GetParameterHandle( "g_xInputSceneSampler" );
		g_xExtractLumUVScaleHandle = g_pxHDREffect->GetParameterHandle( "g_xExtractLumUVScale" );
		g_xAdaptionFactorHandle    = g_pxHDREffect->GetParameterHandle( "g_fAdaptationFactor" );
		g_xMinLuminanceHandle      = g_pxHDREffect->GetParameterHandle( "g_fMinLuminance" );
		g_xMaxLuminanceHandle      = g_pxHDREffect->GetParameterHandle( "g_fMaxLuminance" );
		g_xKeyValueHandle          = g_pxHDREffect->GetParameterHandle( "g_fKeyValue" );
		g_xWhitePointSquaredHandle = g_pxHDREffect->GetParameterHandle( "g_fWhitePointSquared" );
		g_xScaleParamsHandle       = g_pxHDREffect->GetParameterHandle( "g_xScaleParams" );
		g_xDynamicExposureHandle   = g_pxHDREffect->GetParameterHandle( "g_bDynamicExposure" );
		g_xFixedLuminanceHandle    = g_pxHDREffect->GetParameterHandle( "g_fFixedLuminance" );
		g_xXboxGammaFudgeHandle    = g_pxHDREffect->GetParameterHandle( "g_fXboxGammaFudge" );

		// These parameters won't change
		const Asura_Xbox360_Effect_System::Handle xCurrentLuminanceHandle = g_pxHDREffect->GetParameterHandle( "g_xCurrentLuminanceSampler" );
		const Asura_Xbox360_Effect_System::Handle xSceneLuminanceHandle   = g_pxHDREffect->GetParameterHandle( "g_xSceneLuminanceSampler" );
		g_pxHDREffect->SetTexture( xCurrentLuminanceHandle, g_pxCurrentLuminanceTexture );
		g_pxHDREffect->SetTexture( xSceneLuminanceHandle,   g_pxWorkingTexture_R );
	}

	// Measurement region stuff

	#if defined ASURA_ASSERT
		const HRESULT xResult =
	#endif
		Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axREGION_VERTEX_DECLARATION,
			&g_pxRegionVD );

	Asura_Assert( SUCCEEDED( xResult ), "Failed to create HDR vertex declaration." );

	#if defined ASURA_DEBUG_VARIABLE_SYSTEM
		Asura_CommandConsole::AddVar( L"Render.HDR.ShowRegion", &g_bShowRegion,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.HDR.ShowValues", &g_bShowHDRValues,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.HDR.XboxGammaFudge", &g_fXboxGammaFudge, 1.0f, 3.0f,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#ifdef DEBUG_ALLOW_FRAMEBUFFER_GRAB
		Asura_CommandConsole::AddVar( L"Render.HDR.GrabBuffer", &g_bGrabBuffer,
			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif // DEBUG_ALLOW_FRAMEBUFFER_GRAB

	#endif

}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_HDR::Shutdown()
//
// Shuts down the HDR rendering system on 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_HDR::Shutdown()
{
	SAFE_RELEASE( g_pxRegionVD );

	SAFE_RELEASE( g_pxWorkingRT_R );
	SAFE_RELEASE( g_pxWorkingRT_RGB );
	SAFE_RELEASE( g_pxWorkingTexture_R );
	SAFE_RELEASE( g_pxWorkingTexture_RGB );
	SAFE_RELEASE( g_pxCurrentLuminanceTexture );
}


static void GrabFloatSurface(float *fDest, u_int uWidth, u_int uHeight, Direct3D_Texture*pTexture)
{
	D3DLOCKED_RECT xLockedRect;
	Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, pTexture, NULL, 0, 0, NULL, 1.0f, 0, NULL );

	if (SUCCEEDED( pTexture->LockRect(0, &xLockedRect, NULL, D3DLOCK_READONLY)))
	{
		for (u_int y = 0; y < uHeight; ++y)
		{
			static const float *pfSourceRow;
			static float *pfDestRow;
			pfSourceRow =  static_cast< const float* >( xLockedRect.pBits ) + y * xLockedRect.Pitch / 4;
			pfDestRow =  fDest + y * uWidth;

			for (u_int x = 0; x < uWidth; ++x)
			{
				pfDestRow[x] = pfSourceRow[x];
			}
		}
	} else {
		Asura_Assert( false, "Failed to get render target data for luminance capture." );
	}
	pTexture->UnlockRect(0);
}


#ifdef DEBUG_ALLOW_FRAMEBUFFER_GRAB

static void GrabHalfFloatSurface(float *pfDest, u_int uWidth, u_int uHeight, Direct3D_Texture*pTexture)
{
	D3DLOCKED_RECT xLockedRect;

	if (SUCCEEDED( pTexture->LockRect(0, &xLockedRect, NULL, D3DLOCK_READONLY)))
	{
		Asura_HalfFloat *pxTempBuf = new Asura_HalfFloat[uWidth * uHeight * 4];

		XGUntileSurface(pxTempBuf, uWidth * sizeof(Asura_HalfFloat) * 4, NULL, xLockedRect.pBits, uWidth, uHeight,  NULL, sizeof(Asura_HalfFloat) * 4);
		for (u_int y = 0; y < uHeight; ++y)
		{
			Asura_HalfFloat *pxSource = pxTempBuf + (uHeight - y - 1) * uWidth * 4;
			for (u_int x = 0; x < uWidth; ++x)
			{
				*(pfDest++) = (float)*(pxSource++);
				*(pfDest++) = (float)*(pxSource++);
				*(pfDest++) = (float)*(pxSource++);
				++pxSource;
			}
		}
		delete[] pxTempBuf;
	} else {
		Asura_Assert( false, "Failed to get render target data for capture." );
	}
	pTexture->UnlockRect(0);
}



// [16/10/2008 AS] Kept in case it could be useful again...
// static void Grab8888Surface(float *pfDest, u_int uWidth, u_int uHeight, Direct3D_Texture*pTexture)
// {
// 	D3DLOCKED_RECT xLockedRect;
// 	Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, pTexture, NULL, 0, 0, NULL, 1.0f, 0, NULL );
// 
// 	if (SUCCEEDED( pTexture->LockRect(0, &xLockedRect, NULL, D3DLOCK_READONLY)))
// 	{
// 		unsigned char *pxTempBuf = new unsigned char[uWidth * uHeight * 4];
// 
// 		XGUntileSurface(pxTempBuf, uWidth * sizeof(unsigned char) * 4, NULL, xLockedRect.pBits, uWidth, uHeight,  NULL, sizeof(unsigned char) * 4);
// 		for (u_int y = 0; y < uHeight; ++y)
// 		{
// 			unsigned char *pxSource = pxTempBuf + (uHeight - y - 1) * uWidth * 4;
// 			for (u_int x = 0; x < uWidth; ++x)
// 			{
// 				++pxSource;
// 				*(pfDest+0) = ((float)*(pxSource++)) / 255.0f;
// 				*(pfDest+1) = ((float)*(pxSource++)) / 255.0f;
// 				*(pfDest+2) = ((float)*(pxSource++)) / 255.0f;
// 				pfDest += 3;
// 			}
// 		}
// 		delete [] pxTempBuf;
// 	} else {
// 		Asura_Assert( false, "Failed to get render target data for luminance capture." );
// 	}
// 	pTexture->UnlockRect(0);
// }

template <typename T> static T endian_swap(T t)
{
	unsigned char d[sizeof(t)];
	for (int i = 0; i < sizeof(t); ++i)
	{
		d[i] = ((unsigned char*)&t)[sizeof(t) - i - 1];
	}
	return *(T*)d;
}
void SaveFloatImage(const Asura_Char*pFileName, float *pfBuf, u_int uWidth, u_int uHeight)
{
	Asura_Char acHdrBuf[1024];
	u_int uHdrsize = Asura_PrintfToBfr(acHdrBuf, "PF\n%d %d\n-1.000000\n", sizeof(acHdrBuf), false, uWidth, uHeight);
	HANDLE hFile = CreateFile(pFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
	if (hFile != INVALID_HANDLE_VALUE)
	{
		DWORD uWritten = 0;
		float *pRow = new float[uWidth * 3];
		WriteFile(hFile, acHdrBuf, uHdrsize, &uWritten, NULL );
		for (u_int y = 0; y < uHeight; ++y)
		{
			for (u_int x = 0; x < uWidth; ++x)
			{
				pRow[x * 3 + 0] = endian_swap(pfBuf[(y * uWidth + x) * 3 + 0]);
				pRow[x * 3 + 1] = endian_swap(pfBuf[(y * uWidth + x) * 3 + 1]);
				pRow[x * 3 + 2] = endian_swap(pfBuf[(y * uWidth + x) * 3 + 2]);
			}
			WriteFile(hFile, pRow, sizeof(float) * uWidth * 3,&uWritten, NULL );
		}
		delete [] pRow;
		CloseHandle(hFile);

	}
}
#endif //DEBUG_ALLOW_FRAMEBUFFER_GRAB

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_HDR::ApplyToneMapping()
//
// Applies tone mapping to the current scene
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////
void Asura_Xbox360_HDR::ApplyToneMapping()
{
	if( !g_pxHDREffect ) return;

	if( Asura_Xbox360_Graphics::IsTiling2() )
	{
		// This is horrible but if we have a pix event that is started within a tiling bracket
		// and ended outside of it it screwes with pix
		// so we need to end whatever the current event is and then start another one afterwards
#if defined(ASURA_XBOX360_ENABLE_PIX)
		PIXEndNamedEvent();
#endif	//defined(ASURA_XBOX360_ENABLE_PIX)
		Asura_Xbox360_Graphics::EndTiling2();
#if defined(ASURA_XBOX360_ENABLE_PIX)
		PIXBeginNamedEvent(D3DCOLOR_XRGB(0xFF, 0xFF, 0xFF), "Post Tiling2 PixBracket");
#endif	//defined(ASURA_XBOX360_ENABLE_PIX)

	} else {
		Asura_Xbox360_Graphics::Resolve( Asura_Xbox360_Graphics::GetSceneResolveTexture() );
	}

	SCOPEDPIXEVENT(( 0, "Apply tone mapping" ));

	Direct3D_Viewport xViewport;
	Asura_CameraData::HDRSettings xSettings;
	Asura_Camera::GetHDRSettings( xSettings );

	g_pxHDREffect->SetFloat( g_xXboxGammaFudgeHandle, 1.0f / g_fXboxGammaFudge );


	// Initial setup
	{
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 1, NULL );
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 2, NULL );
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 3, NULL );

		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );

		xViewport.X = 0;
		xViewport.Y = 0;
		xViewport.MinZ = 0.0f;
		xViewport.MaxZ = 1.0f;
	}


	// The region is defined in [0,1] terms relative to the screen height, so it always
	// keeps the same aspect ratio when we change to widescreen

	const float fLeft = xSettings.m_xRegion.MinX * Asura_Camera::GetYOverXAspectRatio();
	const float fRight = xSettings.m_xRegion.MaxX * Asura_Camera::GetYOverXAspectRatio();
	const float fTop = xSettings.m_xRegion.MaxY;
	const float fBottom = xSettings.m_xRegion.MinY;

	const UINT uStride = sizeof( D3DXVECTOR4 );

	// First downsample to quarter size
	{
		SCOPEDPIXEVENT(( 0, "Downsample to quarter size" ));

		g_pxHDREffect->SetTexture( g_xDownsizeSamplerHandle, Asura_Xbox360_Graphics::GetSceneResolveTexture() );

		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, g_pxWorkingRT_RGB );

		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxRegionVD );

		g_pxHDREffect->Begin();
		g_pxHDREffect->Commit();

		// We'll not use a quad covering the whole RT, just the portion that corresponds to the
		// selected measurement region

		void* pVerts = NULL;
		Asura_Xbox360_Graphics::GetDevice()->BeginVertices( D3DPT_RECTLIST, 4, uStride, &pVerts );

		D3DXVECTOR4* const pxPoints = static_cast< D3DXVECTOR4* >( pVerts );

		if( pxPoints )
		{
			// As we're writing into write-combined memory, must do it in order with no reads

			pxPoints[0].x = fLeft;
			pxPoints[0].y = fTop;
			pxPoints[0].z = pxPoints[0].w = 0.0f; // Don't need texcoords at this stage

			pxPoints[1].x = fRight;
			pxPoints[1].y = fTop;
			pxPoints[1].z = pxPoints[1].w = 0.0f;

			pxPoints[2].x = fRight;
			pxPoints[2].y = fBottom;
			pxPoints[2].z = pxPoints[2].w = 0.0f;

			pxPoints[3].x = fLeft;
			pxPoints[3].y = fBottom;
			pxPoints[3].z = pxPoints[3].w = 0.0f;

			Asura_Xbox360_Graphics::GetDevice()->EndVertices();
		}
	}

	// Then downsample to 64x64 and calculate average luminance for each 3x3
	{
		SCOPEDPIXEVENT(( 0, "Downsample to 64x64 and calculate log luminance" ));

		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, g_pxWorkingTexture_RGB, NULL, 0, 0, NULL, 1.0f, 0, NULL );
		g_pxHDREffect->SetTexture( g_xDownsizeSamplerHandle, g_pxWorkingTexture_RGB );

		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, g_pxWorkingRT_R );

		{
			xViewport.Width = xViewport.Height = 64;
			Asura_Xbox360_Graphics::GetDevice()->SetViewport( &xViewport );

			const float afScale[] = { 320.0f / 64.0f, 180.0f / 64.0f };
			g_pxHDREffect->SetVector( g_xExtractLumUVScaleHandle, afScale );
		}

		g_pxHDREffect->BeginPass(1);
		g_pxHDREffect->Commit();

		// Now render to the whole RT, but only sample from a subrect of the texture where we
		// rendered before
		void* pVerts = NULL;
		Asura_Xbox360_Graphics::GetDevice()->BeginVertices( D3DPT_RECTLIST, 4, uStride, &pVerts );

		D3DXVECTOR4* const pxPoints = static_cast< D3DXVECTOR4* >( pVerts );

		if( pxPoints )
		{
			// Vertex positions covering the whole of the RT
			pxPoints[0].x = -1.0f; pxPoints[0].y = 1.0f;
			// The new texcoords are the previous positions, converted back from clip space
			pxPoints[0].z = fLeft * 0.5f + 0.5f;
			pxPoints[0].w = 0.5f - fTop * 0.5f; // y increases down the texture

			pxPoints[1].x = 1.0f; pxPoints[1].y = 1.0f;
			pxPoints[1].z = fRight * 0.5f + 0.5f;
			pxPoints[1].w = 0.5f - fTop * 0.5f;

			pxPoints[2].x = 1.0f; pxPoints[2].y = -1.0f;
			pxPoints[2].z = fRight * 0.5f + 0.5f;
			pxPoints[2].w = 0.5f - fBottom * 0.5f;

			pxPoints[3].x = -1.0f; pxPoints[3].y = -1.0f;
			pxPoints[3].z = fLeft * 0.5f + 0.5f;
			pxPoints[3].w = 0.5f - fBottom * 0.5f;

			Asura_Xbox360_Graphics::GetDevice()->EndVertices();
		}
	}


#ifdef DEBUG_CHECK_HDR_LUMINOSTY
	g_pxHDREffect->SetTexture( g_xDownsizeSamplerHandle, NULL );
	g_pxHDREffect->Commit();
	float f64x64Surface[64*64];
	memset(f64x64Surface, 0, sizeof(f64x64Surface));
	GrabFloatSurface(f64x64Surface, 64, 64, g_pxWorkingTexture_R);
#endif
	// Switch to sourcing from the single-channel working texture. Switch to unfiltered
	// downsampling pass
	g_pxHDREffect->SetTexture( g_xDownsizeSamplerHandle, g_pxWorkingTexture_R );
	g_pxHDREffect->BeginPass(2);
	g_pxHDREffect->Commit();

	Asura_Xbox360_Graphics::GetFullScreenVB().Bind();
	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( Asura_Xbox360_Graphics::GetFullScreenVD() );

	// Down to 16x16
	{
		SCOPEDPIXEVENT(( 0, "Downsample to 16x16" ));

		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, g_pxWorkingTexture_R, NULL, 0, 0, NULL, 1.0f, 0, NULL );

		xViewport.Width = xViewport.Height = 16;
		Asura_Xbox360_Graphics::GetDevice()->SetViewport( &xViewport );
		Asura_Xbox360_Graphics::GetDevice()->DrawPrimitive( D3DPT_RECTLIST, 0, 1 );
	}

	// Then to 4x4
	{
		SCOPEDPIXEVENT(( 0, "Downsample to 4x4" ));

		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, g_pxWorkingTexture_R, NULL, 0, 0, NULL, 1.0f, 0, NULL );

		xViewport.Width = xViewport.Height = 4;
		Asura_Xbox360_Graphics::GetDevice()->SetViewport( &xViewport );
		Asura_Xbox360_Graphics::GetDevice()->DrawPrimitive( D3DPT_RECTLIST, 0, 1 );
	}

	// In the final downsampling step, switch PS to apply the exp()
	{
		SCOPEDPIXEVENT(( 0, "Downsample to 1x1 and calculate final scene luminance" ));

		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, g_pxWorkingTexture_R, NULL, 0, 0, NULL, 1.0f, 0, NULL );

		xViewport.Width = xViewport.Height = 1;
		Asura_Xbox360_Graphics::GetDevice()->SetViewport( &xViewport );

		g_pxHDREffect->BeginPass(3);
		g_pxHDREffect->Commit();
		Asura_Xbox360_Graphics::GetDevice()->DrawPrimitive( D3DPT_RECTLIST, 0, 1 );
	}
	
	float fFinalLum = 0.0f;
#ifdef DEBUG_CHECK_HDR_LUMINOSTY
	float f64Lum = 0;
#endif
	if (g_bShowHDRValues)
	{
		g_pxHDREffect->SetTexture( g_xDownsizeSamplerHandle, NULL );
		g_pxHDREffect->Commit();
		GrabFloatSurface(&fFinalLum, 1, 1,g_pxWorkingTexture_R);

#ifdef DEBUG_CHECK_HDR_LUMINOSTY
		for (int i = 0; i < 64* 64; ++i)
			f64Lum += f64x64Surface[i];

		f64Lum /= 64.0f*64.0f;
#endif
		g_pxHDREffect->SetTexture( g_xDownsizeSamplerHandle, g_pxWorkingTexture_R );
		g_pxHDREffect->Commit();
	}


	// Now we've found the scene's luminance value (in the top-left texel of the
	// viewport), retrieve it and lerp the current value towards it by a small amount
	{
		SCOPEDPIXEVENT(( 0, "Adapt current luminance" ));
		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, g_pxWorkingTexture_R, NULL, 0, 0, NULL, 1.0f, 0, NULL );

		g_pxHDREffect->SetFloat( g_xMinLuminanceHandle, xSettings.m_fMinLuminance );
		g_pxHDREffect->SetFloat( g_xMaxLuminanceHandle, xSettings.m_fMaxLuminance );

		const float fAdaptationFactor = Asura_Maths::Min( xSettings.m_fAdaptationRate * Asura_Timers::GetGameTimeSlice(), 1.0f );
		g_pxHDREffect->SetFloat( g_xAdaptionFactorHandle, fAdaptationFactor );

		// No need to reset the viewport, should be left at 1x1 from last time

		g_pxHDREffect->BeginPass(4);
		g_pxHDREffect->Commit();
		Asura_Xbox360_Graphics::GetDevice()->DrawPrimitive( D3DPT_RECTLIST, 0, 1 );
	}

#ifdef DEBUG_ALLOW_FRAMEBUFFER_GRAB
	if (g_bGrabBuffer)
	{
		int iWidth = Asura_Viewport::GetWidth(), iHeight = Asura_Viewport::GetHeight();
 		float *pBuf = new float[iWidth * iHeight * 3];
 		GrabHalfFloatSurface(pBuf, iWidth, iHeight, Asura_Xbox360_Graphics::GetSceneResolveTexture());

		char cBuf[512];
 		static int iCallCount = 1;
		Asura_PrintfToBfr(cBuf, "devkit:\\PreToneMapXbox360_%dx%d_%d.pfm", sizeof(cBuf), true, iWidth, iHeight, iCallCount);
		SaveFloatImage(cBuf, pBuf, iWidth, iHeight);

		const Asura_Xbox360_Effect_System::Handle xCurrentLuminanceHandle = g_pxHDREffect->GetParameterHandle( "g_xCurrentLuminanceSampler" );
		g_pxHDREffect->SetTexture( xCurrentLuminanceHandle, NULL );
		g_pxHDREffect->Commit();
		float fAdaptedLum;
		GrabFloatSurface(&fAdaptedLum, 1, 1,g_pxCurrentLuminanceTexture);
		g_pxHDREffect->SetTexture( xCurrentLuminanceHandle, g_pxCurrentLuminanceTexture );
		g_pxHDREffect->Commit();

		for (int y = 0; y < iHeight; ++y)
		{
			for (int x = 0; x < iWidth; ++x)
			{
				float r = pBuf[(y * iWidth + x) * 3 + 0];
				float g = pBuf[(y * iWidth + x) * 3 + 1];
				float b = pBuf[(y * iWidth + x) * 3 + 2];
				float fHDRLuminance = 0.27f * r + 0.67f * g + 0.06f * b;
				const float fDELTA = 0.0001f;
				const float fScaledLuminance = xSettings.m_fKeyValue / (fAdaptedLum + fDELTA) * fHDRLuminance;
				const float fNumerator = fScaledLuminance * (1.0f + fScaledLuminance / (xSettings.m_fWhitePoint * xSettings.m_fWhitePoint));
				const float fDenominator = 1.0f + fScaledLuminance;

				const float fLDRLuminance = fNumerator / fDenominator;

				// Rescale the colours by the ratio of new to old luminance
				r = r * fLDRLuminance / fHDRLuminance;
				g = g * fLDRLuminance / fHDRLuminance;
				b = b * fLDRLuminance / fHDRLuminance;
				pBuf[(y * iWidth + x) * 3 + 0] = r;
				pBuf[(y * iWidth + x) * 3 + 1] = g;
				pBuf[(y * iWidth + x) * 3 + 2] = b;
			}
		}
		Asura_PrintfToBfr(cBuf, "devkit:\\PostToneMapXbox360_%dx%d_%d.pfm", sizeof(cBuf), true, iWidth, iHeight, iCallCount++);
		SaveFloatImage(cBuf, pBuf, iWidth, iHeight);



		delete[] pBuf;
		g_bGrabBuffer = false;
	}
#endif

	float fAdaptedLum = 0.0f;
	if (g_bShowHDRValues)
	{
		const Asura_Xbox360_Effect_System::Handle xCurrentLuminanceHandle = g_pxHDREffect->GetParameterHandle( "g_xCurrentLuminanceSampler" );
		g_pxHDREffect->SetTexture( xCurrentLuminanceHandle, NULL );
		g_pxHDREffect->Commit();
		GrabFloatSurface(&fAdaptedLum, 1, 1,g_pxCurrentLuminanceTexture);
		g_pxHDREffect->SetTexture( xCurrentLuminanceHandle, g_pxCurrentLuminanceTexture );
		g_pxHDREffect->Commit();
	}
	// Final pass: apply the tone mapping
	{
		SCOPEDPIXEVENT(( 0, "Apply tone mapping to scene" ));
		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, g_pxCurrentLuminanceTexture, NULL, 0, 0, NULL, 1.0f, 0, NULL );

		// SetRenderTarget() will automatically reset the viewport

		Direct3D_Surface* const pxRT = Asura_Xbox360_Graphics::GetPostToneMapLinearRT();
		Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, pxRT );

		g_pxHDREffect->SetTexture( g_xInputSceneSamplerHandle, Asura_Xbox360_Graphics::GetSceneResolveTexture() );
		g_pxHDREffect->SetFloat( g_xKeyValueHandle, xSettings.m_fKeyValue );

		const float fWhitePointSq = xSettings.m_fWhitePoint * xSettings.m_fWhitePoint;
		g_pxHDREffect->SetFloat( g_xWhitePointSquaredHandle, fWhitePointSq );

		// Simple linear rescale on initial sample, to give the concept of "black point"
		{
			Asura_Assert( xSettings.m_fBlackPoint < xSettings.m_fWhitePoint,
				"HDR black point must be < white point." );

			const float fGradient =
				xSettings.m_fWhitePoint / (xSettings.m_fWhitePoint - xSettings.m_fBlackPoint);

			const float fOffset = -xSettings.m_fBlackPoint * fGradient;

			const float afScaleParams[] = { fGradient, fOffset };
			g_pxHDREffect->SetVector( g_xScaleParamsHandle, afScaleParams );
		}

		g_pxHDREffect->SetBool( g_xDynamicExposureHandle, xSettings.m_bDynamicExposure );
		g_pxHDREffect->SetFloat( g_xFixedLuminanceHandle, xSettings.m_fFixedLuminance );
		

		g_pxHDREffect->BeginPass(5);
		g_pxHDREffect->Commit();
		Asura_Xbox360_Graphics::GetDevice()->DrawPrimitive( D3DPT_RECTLIST, 0, 1 );
	}
	Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetPostToneMapRT() );


	// Cleanup
	g_pxHDREffect->End();



	// Draw region if enabled
#if defined ASURA_DEBUG_VARIABLE_SYSTEM
	if( g_bShowRegion )
	{
		SCOPEDPIXEVENT(( 0, "Draw measurement region" ));

		const float fWidth = static_cast< float >( Asura_Viewport::GetWidth() );
		const float fHeight = static_cast< float >( Asura_Viewport::GetHeight() );
		const float fYOverX = fHeight / fWidth;

		#define FROMCLIP_X(x) ((x) * 0.5f + 0.5f)
		#define FROMCLIP_Y(y) (0.5f - (y) * 0.5f)

		// NB both dimensions specified relative to screen height. Convert from clip space
		const float fLeft = FROMCLIP_X( xSettings.m_xRegion.MinX * fYOverX ) * fWidth;
		const float fRight = FROMCLIP_X( xSettings.m_xRegion.MaxX * fYOverX  ) * fWidth;
		const float fTop = FROMCLIP_Y( xSettings.m_xRegion.MaxY ) * fHeight;
		const float fBottom = FROMCLIP_Y( xSettings.m_xRegion.MinY ) * fHeight;

		#undef FROMCLIP_X
		#undef FROMCLIP_Y

		Asura_Render::Render2DLine( fLeft, fTop, fRight, fTop, 0xFFFFFFFF );
		Asura_Render::Render2DLine( fRight, fTop, fRight, fBottom, 0xFFFFFFFF );
		Asura_Render::Render2DLine( fRight, fBottom, fLeft, fBottom, 0xFFFFFFFF );
		Asura_Render::Render2DLine( fLeft, fBottom, fLeft, fTop, 0xFFFFFFFF );
	}


	if (g_bShowHDRValues)
	{
		const u_int uFormat = ASURA_TEXTFORMAT_RELATIVEPOSITION | ASURA_TEXTFORMAT_FORMATSTRING | ASURA_TEXTFORMAT_DROPSHADOWS;
		Asura_Font_System::SetFont(0);
#ifdef DEBUG_CHECK_HDR_LUMINOSTY
		Asura_Font_System::Print( 0.1, 0.3, 0.8f, 0xFFFFFFFF, uFormat, "HDR %f", f64Lum);
		Asura_Font_System::Print( 0.1, 0.32, 0.8f, 0xFFFFFFFF, uFormat, "Final %f, %f", exp(f64Lum), fFinalLum );
		Asura_Font_System::Print( 0.1, 0.34, 0.8f, 0xFFFFFFFF, uFormat, "Diff %f", abs(exp(f64Lum) - fFinalLum));
		Asura_Font_System::Print( 0.1, 0.36, 0.8f, 0xFFFFFFFF, uFormat, "Adapted %f", fAdaptedLum);
#else
		Asura_Font_System::Print( 0.1, 0.3, 0.8f, 0xFFFFFFFF, uFormat, "HDR %f", log(fFinalLum));
		Asura_Font_System::Print( 0.1, 0.32, 0.8f, 0xFFFFFFFF, uFormat, "Final %f", fFinalLum );
		Asura_Font_System::Print( 0.1, 0.34, 0.8f, 0xFFFFFFFF, uFormat, "Adapted %f", fAdaptedLum );
#endif
	}
#endif //#if defined ASURA_DEBUG_VARIABLE_SYSTEM
}

#endif // defined ASURA_USE_HDR
