//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_LineRenderer.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Rendering/Asura_Xbox360_LineRenderer.h"

// Asura Platform-independent
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Render.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Viewport.h"

// Shader Include Files
#include "3DLine.h"
#include "2DLine.h"


// Misc
#include <xtl.h>

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const Direct3D_VertexElement axLINE_VERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	D3DDECL_END()
};

const u_int uLINE_VERTEX_SIZE = 12;

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/

static Direct3D_VertexDeclaration* g_pxLineVD = NULL;

static Asura_Xbox360_Effect* g_px2DLineEffect = NULL;
static Asura_Xbox360_Effect* g_px3DLineEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_x3DLineColourHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_x2DLineColourHandle = 0;

/******************************************************************************/
/* Asura_Xbox360_LineRenderer Static Class Functions                          */
/******************************************************************************/

void Asura_Xbox360_LineRenderer::Initialise()
{
	g_px3DLineEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_au3DLine);
	g_px2DLineEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_au2DLine);

	g_x3DLineColourHandle = g_px3DLineEffect->GetParameterHandle( "g_xColour" );
	g_x2DLineColourHandle = g_px2DLineEffect->GetParameterHandle( "g_xColour" );

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axLINE_VERTEX_DECLARATION, &g_pxLineVD );
	Asura_Assert( g_pxLineVD, "Failed to create debug line vertex declaration." );
}

void Asura_Xbox360_LineRenderer::Reset()
{
}

void Asura_Xbox360_LineRenderer::Shutdown()
{
	Asura_Xbox360_Effect_System::Destroy( g_px2DLineEffect );
	g_px2DLineEffect = NULL;

	Asura_Xbox360_Effect_System::Destroy( g_px3DLineEffect );
	g_px3DLineEffect = NULL;

	g_pxLineVD->Release();
	g_pxLineVD = NULL;
}

void Asura_Xbox360_LineRenderer::Render2DLine( const Asura_Vector_2& xP1, const Asura_Vector_2& xP2, const u_int uColour )
{
	
	if (!Asura_Xbox360_Graphics::HasThreadOwnership())
	{
		Asura_Assert(false, "Can't render lines now! - Not within BeginFrame()/EndFrame()" );
		return;
	}
	
	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	pxDevice->SetVertexDeclaration( g_pxLineVD );

	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );

	const float fAlpha = static_cast< float >( (uColour & 0xFF000000) >> 24 ) / 255.0f;
	const float fRed   = static_cast< float >( (uColour & 0x00FF0000) >> 16 ) / 255.0f;
	const float fGreen = static_cast< float >( (uColour & 0x0000FF00) >>  8 ) / 255.0f;
	const float fBlue  = static_cast< float >( (uColour & 0x000000FF)       ) / 255.0f;

	const float afColour[4] = { fRed, fGreen, fBlue, fAlpha };
	g_px2DLineEffect->GetInterface()->SetVectorF( g_x2DLineColourHandle, afColour );

	g_px2DLineEffect->Begin();
	g_px2DLineEffect->Commit();

	D3DXVECTOR3* pxVertices = NULL;

	const HRESULT hResult = pxDevice->BeginVertices( D3DPT_LINELIST, 2, uLINE_VERTEX_SIZE, reinterpret_cast< void** >( &pxVertices ) );
	if( SUCCEEDED( hResult ) )
	{
		pxVertices[0].x = ((xP1.x*2.0f)/Asura_Viewport::GetWidth())-1.0f;
		pxVertices[0].y = -(((xP1.y*2.0f)/Asura_Viewport::GetHeight())-1.0f);
		pxVertices[0].z = 0;

		pxVertices[1].x = ((xP2.x*2.0f)/Asura_Viewport::GetWidth())-1.0f;
		pxVertices[1].y = -(((xP2.y*2.0f)/Asura_Viewport::GetHeight())-1.0f);
		pxVertices[1].z = 0;

		pxDevice->EndVertices();
	}

	g_px2DLineEffect->End();
}


void Asura_Xbox360_LineRenderer::Render3DLine( const Asura_Vector_3& xP1, const Asura_Vector_3& xP2, const u_int uColour )
{
	if (!Asura_Xbox360_Graphics::HasThreadOwnership())
	{
		Asura_Assert(false, "Can't render lines now! - Not within BeginFrame()/EndFrame()" );
		return;
	}

	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	pxDevice->SetVertexDeclaration( g_pxLineVD );

	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );

	const float fAlpha = static_cast< float >( (uColour & 0xFF000000) >> 24 ) / 255.0f;
	const float fRed   = static_cast< float >( (uColour & 0x00FF0000) >> 16 ) / 255.0f;
	const float fGreen = static_cast< float >( (uColour & 0x0000FF00) >>  8 ) / 255.0f;
	const float fBlue  = static_cast< float >( (uColour & 0x000000FF)       ) / 255.0f;

	const float afColour[4] = { fRed, fGreen, fBlue, fAlpha };
	g_px3DLineEffect->GetInterface()->SetVectorF( g_x3DLineColourHandle, afColour );

	g_px3DLineEffect->Begin();
	g_px3DLineEffect->Commit();

	D3DXVECTOR3* pxVertices = NULL;

	const HRESULT hResult = pxDevice->BeginVertices( D3DPT_LINELIST, 2, uLINE_VERTEX_SIZE, reinterpret_cast< void** >( &pxVertices ) );
	if( SUCCEEDED( hResult ) )
	{
		pxVertices[0].x = xP1.x;
		pxVertices[0].y = xP1.y;
		pxVertices[0].z = xP1.z;

		pxVertices[1].x = xP2.x;
		pxVertices[1].y = xP2.y;
		pxVertices[1].z = xP2.z;

		pxDevice->EndVertices();
	}

	g_px3DLineEffect->End();
}
