//////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Core/Asura_MaterialResponse_System.h"
#include "Rendering/Asura_Xbox360_MaterialResponse_System.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_TextureAnimation.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_HalfFloat.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Light/Asura_GI_Solution.h"
#include "Rendering/Asura_DeferredShading.h"
#include "Rendering/Asura_Fog.h"
#include "Rendering/Asura_Render.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Fog.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Rendering/Asura_Xbox360_XPR_System.h"

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// DLH - 14 June 2007 11:40:09 - Turn this on to fill all the graphs with a 0.25 apart from the first and last item
// DLH - 14 June 2007 11:40:09 - This makes checking the graph texture reads in the shader easier.
//#define ASURA_XBOX360_MATERIALRESPONSE_UPDATETEXTURE_FILL_WITHDEBUGLIMIT

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Convert()
//
// Reads a float from a chunk stream and stores it as a half
//
// Parameters:		xStream - Chunk stream from which to read
//					xStore - Unsigned short in which to store the value
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

static void Convert( Asura_Chunk_Stream& xStream, u_short& xStore )
{
	float f;
	xStream >> f;
	f *= 65535.0f;
	f = Asura_Maths::ClampToRange( f, 0.0f, 65535.0f );
	xStore = static_cast< u_short >( Asura_Maths::FloatToInt_Near(f) );
}

/******************************************************************************/
/* Asura_Xbox360_MRPShader Class Functions                                    */
/******************************************************************************/

Asura_Xbox360_MRPShader::Asura_Xbox360_MRPShader( Direct3D_PixelShader* const pxShader,
												  Asura_Xbox360_XPR& xXPR )
	: m_pxShader( pxShader )
	, m_xXPR( xXPR )
{
	xXPR.AddRef();
}

Asura_Xbox360_MRPShader::~Asura_Xbox360_MRPShader()
{
	SAFE_RELEASE( m_pxShader );
	m_xXPR.Release();
}

/******************************************************************************/
/* Asura_Xbox360_MaterialResponse_System Static Class Data                    */
/******************************************************************************/

Direct3D_Texture* Asura_Xbox360_MaterialResponse_System::s_pxMaterialsTexture = NULL;

Asura_TemplateHashTree< Asura_Xbox360_MRPShader* > Asura_Xbox360_MaterialResponse_System::s_xPSTree;
Asura_TemplateHashTree< Direct3D_ConstantBuffer* > Asura_Xbox360_MaterialResponse_System::s_xCBTree;

/******************************************************************************/
/* Asura_Xbox360_MaterialResponse_System Static Class Functions               */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::Initialise()
//
// Initialises the material response system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::Initialise()
{
	//(AAM - 30.01.2007) Used to use D3DXCreateTexture() here, but it simply can't create most
	// 360 extension formats. No need to use it on a fixed platform anyway
	Asura_Xbox360_Graphics::GetDevice()->CreateTexture( ASURA_MATERIAL_RESPONSE_TEXTURE_WIDTH,
														ASURA_XBOX360_MATERIALS_TEXTURE_HEIGHT,
														1, 0, D3DFMT_LIN_A16B16G16R16,
														D3DPOOL_DEFAULT, &s_pxMaterialsTexture,
														NULL );

	Asura_Assert( s_pxMaterialsTexture, "Failed to create material response data texture." );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::Reset()
//
// Resets the material response system on Xbox 360. Called between level loads
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::Reset()
{
	// Remove pixel shaders. TODO: we're going to have problems with global materials, but not an
	// issue for the moment since we're only doing this for the environment
	{
		Asura_TemplateHashTree_Iterator< Asura_Xbox360_MRPShader* > xIter( s_xPSTree );
		while( !xIter.Done() )
		{
			const Asura_Hash_ID uNameHash = xIter.GetCurrentHash();
			Asura_Xbox360_MRPShader** const ppxShader = xIter.GetCurrent();

			xIter.Next();

			delete *ppxShader;
			s_xPSTree.DeleteNodeByHash( uNameHash );
		}
	}

	// Remove CBs from the tree for non-permanent materials
	{
		Asura_TemplateHashTree_Iterator< Direct3D_ConstantBuffer* > xIter( s_xCBTree );
		while( !xIter.Done() )
		{
			const Asura_Hash_ID uNameHash = xIter.GetCurrentHash();
			Direct3D_ConstantBuffer** const ppxCB = xIter.GetCurrent();

			xIter.Next();

			const Asura_MaterialResponse* const pxMaterial
				= Asura_MaterialResponse_System::GetMaterialByHashID( uNameHash );

			if( pxMaterial && pxMaterial->GetRemoveOnReset() )
			{
				SAFE_RELEASE( *ppxCB );
				s_xCBTree.DeleteNodeByHash( uNameHash );
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::Shutdown()
//
// Shuts the material response system down on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::Shutdown()
{
	Reset();

	if( s_pxMaterialsTexture )
	{
		s_pxMaterialsTexture->Release();
		s_pxMaterialsTexture = NULL;
	}

	// Remove all CBs from the tree

	Asura_TemplateHashTree_Iterator< Direct3D_ConstantBuffer* > xIter( s_xCBTree );
	while( !xIter.Done() )
	{
		Direct3D_ConstantBuffer** const ppxCB = xIter.GetCurrent();
		SAFE_RELEASE( *ppxCB );

		xIter.Next();
	}

	s_xCBTree.Reset();
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::CreateConstantBuffer()
//
// Creates a constant buffer to hold the material properties which are uploaded
// as shader constants. Called when a material is loaded from a chunk
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::CreateConstantBuffer( const Asura_Hash_ID uNameHash,
	const Asura_MaterialResponse& xMaterial )
{
	Direct3D_ConstantBuffer* pxCB = NULL;
	Asura_Xbox360_Graphics::GetDevice()->CreateConstantBuffer( 4, 0, &pxCB );

	Asura_Assert( pxCB, "Failed to create material constant buffer." );
	if( !pxCB ) return;

	D3DVECTOR4* pxBufferContents = NULL;
	pxCB->Lock( &pxBufferContents, 0 );

	Asura_Assert( pxBufferContents, "Failed to lock material constant buffer." );
	if( pxBufferContents )
	{
		const float fParallaxScale = xMaterial.GetParallaxScale();

		pxBufferContents[0].x = static_cast< float >( xMaterial.GetMaterialIndex() );
		pxBufferContents[0].y = xMaterial.GetReflectivity();
		pxBufferContents[0].z = fParallaxScale;
		pxBufferContents[0].w = -fParallaxScale * 0.5f;

		pxBufferContents[1].x = xMaterial.GetAlbedoDetailMapUVScale();
		pxBufferContents[1].y = xMaterial.GetAlbedoDetailMapValueScale();
		pxBufferContents[1].z = xMaterial.GetNormalDetailMapUVScale();
		pxBufferContents[1].w = xMaterial.GetNormalDetailMapValueScale();

		pxCB->Unlock();
	}

	s_xCBTree.AddNode( uNameHash, pxCB );
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::GetConstantBuffer()
//
// Retrieves the constant buffer associated with the given material response
//
////////////////////////////////////////////////////////////////////////////////

Direct3D_ConstantBuffer* Asura_Xbox360_MaterialResponse_System::GetConstantBuffer( const Asura_Hash_ID uMaterialName )
{
	Asura_Hash_ID uCBHash =
		(uMaterialName != ASURA_HASH_ID_UNSET ? uMaterialName : ASURA_DEFAULT_MATERIAL_RESPONSE);

	// Need to check the override system
	uCBHash = Asura_MaterialResponse_System::GetOverriddenHash( uCBHash );

	Direct3D_ConstantBuffer** const ppxCB = s_xCBTree.GetDataByHash( uCBHash );
	Asura_Assertf( ppxCB, ("Failed to find constant buffer for material 0x%X.", uCBHash) );

	return ppxCB ? *ppxCB : NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::UpdateDataTexture()
//
// Adds material response curves from a chunk stream to the data texture
//
// Parameters:		xStream - Chunk stream from which to read
//					iVersion - Version number of the chunk
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::UpdateDataTexture( Asura_Chunk_Stream &xStream, const int iVersion )
{
	// JAT 16/05/07 this version number is common to PC, 360, PS3 exports
	const u_int uCURRENT_VERSION = Asura_MaterialResponse_System::EXPORTED_TEXTURE_VERSION;
	if ( iVersion < uCURRENT_VERSION )
	{
		Asura_Assert( false, "Bad version re-export the level to fix this" );
		return;
	}

	if( !s_pxMaterialsTexture )
	{
		Asura_Assert( false, "Trying to update material response texture, but it doesn't exist." );
		return;
	}

	D3DLOCKED_RECT xLockedRect;
	const HRESULT xResult = s_pxMaterialsTexture->LockRect( 0, &xLockedRect, NULL, 0 );
	if( FAILED( xResult ) )
	{
		Asura_Assert( false, "Failed to lock material response texture." );
		return;
	}

	// Pitch is specified in bytes for non-DXT formats
	const u_int uPitchInComponents = xLockedRect.Pitch / sizeof( u_short );
	
	u_int uNumberOfMaterials;
	xStream >> uNumberOfMaterials;

	u_short* pxData = static_cast< u_short* >( xLockedRect.pBits );
	u_short* const pxDataLast =
		pxData + uPitchInComponents * (ASURA_XBOX360_MATERIALS_TEXTURE_HEIGHT - 1);

	const u_int uNumLines = Asura_MaterialResponse_System::GetNumMaterialsTextureLines();
	pxData += uNumLines * uPitchInComponents;
	Asura_MaterialResponse_System::SetNumMaterialsTextureLines( uNumLines + uNumberOfMaterials );

	Asura_Assert( uNumLines + uNumberOfMaterials <= ASURA_XBOX360_MATERIALS_TEXTURE_HEIGHT,
				  "Too many material responses. Data texture needs to be enlarged." );

	for ( u_int uMaterial=0; uMaterial<uNumberOfMaterials; ++uMaterial )
	{
		// DLH - Stop the pointer going outside of the texture space
		pxData = (pxData < pxDataLast) ? pxData : pxDataLast;

		// First we'll have the graphs, one in each colour channel

		u_int uElement=0;
 
		for ( ; uElement< (ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH*4);  )
		{
#if defined( ASURA_XBOX360_MATERIALRESPONSE_UPDATETEXTURE_FILL_WITHDEBUGLIMIT)
			Asura_HalfFloat xDummy;
			// m_xDiffuseGraph.GetValueAtTime( fTime ) / 100.0f;
			Convert(xStream, xDummy);
			// m_xSpecularGraph.GetValueAtTime( fTime ) / 100.0f;
			Convert(xStream, xDummy);
			// m_xTransparencyGraph.GetValueAtTime( fTime ) / 100.0f;
			Convert(xStream, xDummy);
			// 0.0f;
			Convert(xStream, xDummy);
			
			if( 0 == uElement || ((uGRAPH_PIXEL_WIDTH-1)*4) == uElement)
			{
				pxData[uElement++] = 0;
				pxData[uElement++] = 0;
				pxData[uElement++] = 0;
				pxData[uElement++] = 0;
			}
			else
			{
				pxData[uElement++] = 0.25f;
				pxData[uElement++] = 0.25f;
				pxData[uElement++] = 0.25f;
				pxData[uElement++] = 0.25f;
			}
#else
			// m_xDiffuseGraph.GetValueAtTime( fTime ) / 100.0f;
			Convert(xStream, pxData[uElement++]);
			// m_xSpecularGraph.GetValueAtTime( fTime ) / 100.0f;
			Convert(xStream, pxData[uElement++]);
			// m_xTransparencyGraph.GetValueAtTime( fTime ) / 100.0f;
			Convert(xStream, pxData[uElement++]);
			// 0.0f;
			Convert(xStream, pxData[uElement++]);
#endif
		}

		// Then the rest of the material's properties

		// m_xAmbientColour.m_fR;
		// m_xAmbientColour.m_fG;
		// m_xAmbientColour.m_fB;
		// m_fColouriseAmbient / 100.0f;
#if defined( ASURA_XBOX360_MATERIALRESPONSE_UPDATETEXTURE_FILL_WITHDEBUGLIMIT)
		Asura_HalfFloat xDummy;

		Convert(xStream, xDummy);
		Convert(xStream, xDummy);
		Convert(xStream, xDummy);
		Convert(xStream, xDummy);
		
		pxData[uElement++] = 1;
		pxData[uElement++] = 1;
		pxData[uElement++] = 1;
		pxData[uElement++] = 1;
#else
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
#endif

		// m_xDiffuseColour.m_fR;
		// m_xDiffuseColour.m_fG;
		// m_xDiffuseColour.m_fB;
		// m_fColouriseDiffuse / 100.0f;
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);

		// m_xSpecularColour.m_fR;
		// m_xSpecularColour.m_fG;
		// m_xSpecularColour.m_fB;
		// m_fColouriseSpecular / 100.0f;
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);

		// m_fAmbientContribution / 100.0f;
		// m_fMetalicity / 100.0f;
		// 0.0f;
		// 0.0f;
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);
		Convert(xStream, pxData[uElement++]);

		pxData += uPitchInComponents;
	}

	s_pxMaterialsTexture->UnlockRect(0);
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::SetBlendStates()
//
// Sets alpha blend state on the GPU appropriate for the given material response
//
// Parameters:		pxMaterial - Material to set. May be NULL for default values
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::SetBlendStates( const Asura_MaterialResponse* const pxMaterial )
{
#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM
	if ( Asura_Visibility_System::ShouldRenderInXRayMode() )
	{
		SCOPEDPIXEVENT(( 0xFF884499, "Set blend mode (additive)" ));

		Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
		return;
	}
#endif

	const Asura_MaterialResponse::BLEND_MODE eBlendMode = pxMaterial
														? pxMaterial->GetBlendMode()
														: Asura_MaterialResponse::BLEND_MODE_OPAQUE;

	switch( eBlendMode )
	{
		default:
		case Asura_MaterialResponse::BLEND_MODE_OPAQUE:
		{
			SCOPEDPIXEVENT(( 0xFF884499, "Set blend mode (opaque)" ))

			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
			Asura_Xbox360_Graphics::SetTranslucencyModeMRT( 2, ASURA_TRANSLUCENCY_OFF );

			Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );

			break;
		}

		case Asura_MaterialResponse::BLEND_MODE_ADDITIVE:
		{
			SCOPEDPIXEVENT(( 0xFF884499, "Set blend mode (additive)" ));

			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
			Asura_Xbox360_Graphics::SetTranslucencyModeMRT( 2, ASURA_TRANSLUCENCY_GLOWING );

			Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );

			break;
		}

		case Asura_MaterialResponse::BLEND_MODE_BACKGROUNDBLEND:
		{
			SCOPEDPIXEVENT(( 0xFF884499, "Set blend mode (background)" ));

			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
			Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
			Asura_Xbox360_Graphics::SetTranslucencyModeMRT( 2, ASURA_TRANSLUCENCY_NORMAL );

			Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );

			break;
		}
	}

	if( pxMaterial && pxMaterial->GetIsDecal() )
	{
		// Apply depth bias so these can display above coplanar geometry
		Asura_Xbox360_Graphics::SetDecalDepthBias();

		if( pxMaterial->GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_OPAQUE )
		{
			// Need to disable writes to the normal-depth buffer and specular channel
			Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, 0 );
			Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2,
				D3DCOLORWRITEENABLE_RED | D3DCOLORWRITEENABLE_GREEN | D3DCOLORWRITEENABLE_BLUE );
			Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, 0 );
		}
	}
	else
	{
		Asura_Xbox360_Graphics::ResetDepthBias();

		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL);
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE2, D3DCOLORWRITEENABLE_ALL);
		Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE3, D3DCOLORWRITEENABLE_ALL);
	}
	if (Asura_Xbox360_Graphics::s_bIsCurrentlyRenderingInSkybox)
	{
		Asura_Xbox360_Graphics::SetSkyboxDepthBias();
	}

}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_System::SetEffectParameters()
//
// Sets parameters for a material response effect as appropriate for the given
// material
//
// Parameters:		pxMaterial1 - Material to use. May be NULL for default values
//					pxMaterial2 - Dual-texturing material to use. May be NULL
//					pxTextureAnimationController - Tex anim controller, if any
//					xHandles - Bundle of handles to the params in this effect
//					xEffect - Effect in which to set the parameters
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::SetEffectParameters(
	const Asura_MaterialResponse* const pxMaterial1,
	const Asura_MaterialResponse* const pxMaterial2, const Asura_Animation* const pxAnim,
	const Asura_Xbox360_MaterialResponse_Handles& xHandles, Asura_Xbox360_Effect& xEffect )
{
	xEffect.SetTexture( xHandles.m_xMaterialsTexture, GetDataTexture() );

	const Asura_TextureAnimationController* const pxTextureAnimationController =
		(pxAnim ? pxAnim->GetTextureAnimController() : NULL);

	if( pxMaterial1 )
	{
		// Set texture usage flags
		{
			xEffect.SetBool( xHandles.m_xUseAlbedo, pxMaterial1->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseNormalMap, pxMaterial1->GetTexture( Asura_MaterialResponse::NORMAL_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseSpecularMap, pxMaterial1->GetTexture( Asura_MaterialResponse::SPECULAR_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseTransparencyMap, pxMaterial1->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseSelfIlluminationMap, pxMaterial1->GetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseEnvironmentMap, pxMaterial1->GetTexture( Asura_MaterialResponse::ENVIRONMENT_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseReflectivityMask, pxMaterial1->GetTexture( Asura_MaterialResponse::REFLECTIVITY_MASK ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xUseParallaxMap, pxMaterial1->GetTexture( Asura_MaterialResponse::PARALLAX_MAP ) != ASURA_TEXTURE_NONE );

			xEffect.SetBool( xHandles.m_xDualAlbedo, pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xDualNormalMap, pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::NORMAL_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xDualSpecularMap, pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::SPECULAR_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xDualTransparencyMap, pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xDualSelfIlluminationMap, pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP ) != ASURA_TEXTURE_NONE );
			xEffect.SetBool( xHandles.m_xDualParallaxMap, pxMaterial2 && pxMaterial2->GetTexture( Asura_MaterialResponse::PARALLAX_MAP ) != ASURA_TEXTURE_NONE );
		}

		// Set textures
		{
			xEffect.SetTexture( xHandles.m_xAlbedo, pxMaterial1->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) );
			xEffect.SetTexture( xHandles.m_xNormalMap, pxMaterial1->GetTexture( Asura_MaterialResponse::NORMAL_MAP ) );
			xEffect.SetTexture( xHandles.m_xSpecularMap, pxMaterial1->GetTexture( Asura_MaterialResponse::SPECULAR_MAP ) );
			xEffect.SetTexture( xHandles.m_xTransparencyMap, pxMaterial1->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ) );
			xEffect.SetTexture( xHandles.m_xSelfIlluminationMap, pxMaterial1->GetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP ) );
			xEffect.SetTexture( xHandles.m_xEnvironmentMap, pxMaterial1->GetTexture( Asura_MaterialResponse::ENVIRONMENT_MAP ) );
			xEffect.SetTexture( xHandles.m_xReflectivityMask, pxMaterial1->GetTexture( Asura_MaterialResponse::REFLECTIVITY_MASK ) );
			xEffect.SetTexture( xHandles.m_xParallaxMap, pxMaterial1->GetTexture( Asura_MaterialResponse::PARALLAX_MAP ) );

			if( pxMaterial2 )
			{
				xEffect.SetTexture( xHandles.m_xAlbedo2, pxMaterial2->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ) );
				xEffect.SetTexture( xHandles.m_xNormalMap2, pxMaterial2->GetTexture( Asura_MaterialResponse::NORMAL_MAP ) );
				xEffect.SetTexture( xHandles.m_xSpecularMap2, pxMaterial2->GetTexture( Asura_MaterialResponse::SPECULAR_MAP ) );
				xEffect.SetTexture( xHandles.m_xTransparencyMap2, pxMaterial2->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ) );
				xEffect.SetTexture( xHandles.m_xSelfIlluminationMap2, pxMaterial2->GetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP ) );
				xEffect.SetTexture( xHandles.m_xParallaxMap2, pxMaterial2->GetTexture( Asura_MaterialResponse::PARALLAX_MAP ) );
			}
			else
			{
				xEffect.SetTexture( xHandles.m_xAlbedo2, ASURA_TEXTURE_NONE );
				xEffect.SetTexture( xHandles.m_xNormalMap2, ASURA_TEXTURE_NONE );
				xEffect.SetTexture( xHandles.m_xSpecularMap2, ASURA_TEXTURE_NONE );
				xEffect.SetTexture( xHandles.m_xTransparencyMap2, ASURA_TEXTURE_NONE );
				xEffect.SetTexture( xHandles.m_xSelfIlluminationMap2, ASURA_TEXTURE_NONE );
				xEffect.SetTexture( xHandles.m_xParallaxMap2, ASURA_TEXTURE_NONE );
			}
		}

		// Set misc packed properties
		{
			FXLVECTOR xProperties;

			xProperties.x = static_cast< float >( pxMaterial1->GetMaterialIndex() );
			xProperties.y = pxMaterial1->GetReflectivity();

			// Set parallax bias to half the scale for now, centres effect on the polygon position
			xProperties.z = pxMaterial1->GetParallaxScale();
			xProperties.w = -xProperties.z * 0.5f;

			xEffect.SetVector( xHandles.m_xMaterialProperties, xProperties );
		}

		// Set self-illumination parameters
		{
			float fSelfIlluminationAmount;
			float fSelfIlluminationColourise;
			Asura_RGBA xSelfIlluminationColour;
			pxMaterial1->GetSelfIllumination(pxTextureAnimationController, fSelfIlluminationAmount, fSelfIlluminationColourise, xSelfIlluminationColour);


			D3DXVECTOR4 xSelfIlluminationParameters( xSelfIlluminationColour.m_fR, xSelfIlluminationColour.m_fG, xSelfIlluminationColour.m_fB, 0.0f );
			xSelfIlluminationParameters *= fSelfIlluminationAmount * fSelfIlluminationColourise;
			xSelfIlluminationParameters.w = fSelfIlluminationAmount * (1.0f - fSelfIlluminationColourise);

			xEffect.SetVector( xHandles.m_xSelfIlluminationParameters, xSelfIlluminationParameters );
		}

		// Set texture animation
		{
			// Grab the texture anim UV offsets for both channels
			Asura_TextureAnimationOutput xTexAnim1, xTexAnim2;

			Asura_TextureAnimation_System::GetAnimationState( pxMaterial1->GetTextureAnim(), pxTextureAnimationController, &xTexAnim1 );
			Asura_TextureAnimation_System::GetAnimationState( pxMaterial2 ? pxMaterial2->GetTextureAnim() : ASURA_HASH_ID_UNSET, pxTextureAnimationController, &xTexAnim2 );

			const float afTexCoordOffset[4] = { xTexAnim1.m_fUOffset, xTexAnim1.m_fVOffset, xTexAnim2.m_fUOffset, xTexAnim2.m_fVOffset };
			xEffect.SetVector( xHandles.m_xUV_Offsets, afTexCoordOffset );
		}

		// Albedo modifier
		{
			Asura_RGBA xAlbedoModifier( 1.0f, 1.0f, 1.0f, 1.0f );

			if( Asura_MaterialResponse_System::GetAlbedoModifierCallback() )
			{
				Asura_MaterialResponse_System::GetAlbedoModifierCallback()( pxAnim,
					pxMaterial1->GetProjectContext(), xAlbedoModifier );
			}

			xEffect.SetVector( xHandles.m_xAlbedoModifier, &xAlbedoModifier.m_fR );
		}

		// Misc
		{
			const bool bBlendIntoAlbedo = (pxMaterial1->GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_OPAQUE)
										&& pxMaterial1->GetIsDecal();

			xEffect.SetBool( xHandles.m_xBlendIntoAlbedo, bBlendIntoAlbedo );
		}
	}
	else
	{
		// Set texture usage flags
		{
			xEffect.SetBool( xHandles.m_xUseAlbedo, false );
			xEffect.SetBool( xHandles.m_xUseNormalMap, false );
			xEffect.SetBool( xHandles.m_xUseSpecularMap, false );
			xEffect.SetBool( xHandles.m_xUseTransparencyMap, false );
			xEffect.SetBool( xHandles.m_xUseSelfIlluminationMap, false );
			xEffect.SetBool( xHandles.m_xUseEnvironmentMap, false );
			xEffect.SetBool( xHandles.m_xUseReflectivityMask, false );
		}

		// Set self-illumination parameters
		{
			const D3DXVECTOR4 xSelfIlluminationParameters( 0.0f, 0.0f, 0.0f, 0.0f );
			xEffect.SetVector( xHandles.m_xSelfIlluminationParameters, xSelfIlluminationParameters );
		}

		// Zero the uv offsets
		{
			const float afTexCoordOffset[4] = { 0.0f, 0.0f, 0.0f, 0.0f };
			xEffect.SetVector( xHandles.m_xUV_Offsets, afTexCoordOffset );
		}

		// Misc
		{
			// Just use the first material's responses
			const FXLVECTOR xProperties = { 0.0f, 0.0f, 0.0f, 0.0f };
			xEffect.SetVector( xHandles.m_xMaterialProperties, xProperties );
			xEffect.SetBool( xHandles.m_xBlendIntoAlbedo, false );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_MaterialResponse::SetSHParameters()
//
// Sets effect parameters relating to an SH, including brightest light stuff
// for transparent strip rendering post-deferred
//
// Parameters:		pxMaterial - Material to use. May be NULL for default values
//					xHandles - Bundle of handles to the params in this effect
//					xEffect - Effect in which to set the parameters
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::SetSHParameters( const Asura_MaterialResponse* const pxMaterial,
															 const Asura_Xbox360_MaterialResponse_Handles& xHandles,
															 Asura_Xbox360_Effect& xEffect )
{
	Asura_SphericalHarmonics_Order2_RGB xSH;

	if( pxMaterial )
	{
		// Set SH lighting by considering static and dynamic components separately

		xSH = Asura_GI_Solution::s_xStaticSHForCurrentRender;

		Asura_RGBA xColour = pxMaterial->GetAmbientColour();

		const float fPrelightingColourise = pxMaterial->GetPrelightingColourise();
		xSH.Colourise( xColour.m_fR, xColour.m_fG, xColour.m_fB, fPrelightingColourise );

		const float fAmbientContribution = pxMaterial->GetAmbientContribution();
		xSH.Scale( fAmbientContribution );

		if( pxMaterial->IsTransparent() )
		{	
			xSH += Asura_GI_Solution::s_xDynamicSHForCurrentRender;
		}
	}
	else
	{
		// send through basic SH
		xSH = Asura_GI_Solution::s_xSHForCurrentRender;
	}

	// Set effect parameters

	const FXLVECTOR xRedSH = { xSH.m_afRedCoefficients[1], xSH.m_afRedCoefficients[2],
							   xSH.m_afRedCoefficients[3], xSH.m_afRedCoefficients[0] };

	const FXLVECTOR xGreenSH = { xSH.m_afGreenCoefficients[1], xSH.m_afGreenCoefficients[2],
								 xSH.m_afGreenCoefficients[3], xSH.m_afGreenCoefficients[0] };

	const FXLVECTOR xBlueSH = { xSH.m_afBlueCoefficients[1], xSH.m_afBlueCoefficients[2],
								xSH.m_afBlueCoefficients[3], xSH.m_afBlueCoefficients[0] };

	xEffect.SetVector( xHandles.m_xRedSH, xRedSH );
	xEffect.SetVector( xHandles.m_xGreenSH, xGreenSH );
	xEffect.SetVector( xHandles.m_xBlueSH, xBlueSH );

	// Now we've scaled the SH, calculate the related light properties for transparent render
	if( pxMaterial && pxMaterial->IsTransparent() )
	{
		// Find the brightest direction

		Asura_Vector_3 xLightDirection_Asura;
		xSH.GetBrightestDirection( xLightDirection_Asura );

		if( xLightDirection_Asura.ApproximatelyEquals( Asura_Maths::ZeroVector3 ) )
		{
			// If there are no lights, we'll get a zero direction. Not much we can do here,
			// might as well consider the light as coming from above
			xLightDirection_Asura.y = -1.0f;
		}

		const FXLVECTOR xLightDirection_FXL = { xLightDirection_Asura.x,
												xLightDirection_Asura.y,
												xLightDirection_Asura.z,
												1.0f };

		// Evaluate the colour in that direction

		FXLVECTOR xLightColour;
		{
			xLightColour.x = XMVector4Dot( xLightDirection_FXL, xRedSH ).x;
			xLightColour.y = XMVector4Dot( xLightDirection_FXL, xGreenSH ).x;
			xLightColour.z = XMVector4Dot( xLightDirection_FXL, xBlueSH ).x;
			xLightColour.w = 0.0f;

			xLightColour = XMVectorMax( xLightColour, XMVectorZero() );
		}

		// Use the same method to calculate light brightness as the photon shaders. Should
		// probably use standard luminance calculation instead
		const XMVECTOR xAveragingVector = { 0.333f, 0.333f, 0.333f, 0.0f };

		// Store the result in the w component to minimise number of constant registers used
		xLightColour.w = XMVector3Dot( xLightColour, xAveragingVector ).x;

		xEffect.SetVector( xHandles.m_xSHLightDirection, xLightDirection_FXL );
		xEffect.SetVector( xHandles.m_xSHLightColour, xLightColour );
	}
}



////////////////////////////////////////////////////////////////////////////////
//
// Asura_Direct3D_MaterialResponse::SetFogParameters()
//
// Sets effect parameters relating to fog, for for transparent strip rendering post-deferred
//
// Parameters:		pxMaterial - Material to use. May be NULL for default values
//					xHandles - Bundle of handles to the params in this effect
//					xEffect - Effect in which to set the parameters
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_System::SetFogParameters(bool bIsForwardPass, 
															 const Asura_MaterialResponse* const pxMaterial, 
															 const Asura_Xbox360_MaterialResponse_Handles& xHandles,
															 Asura_Xbox360_Effect& xEffect )
{
	if (!bIsForwardPass)
	{
		xEffect.SetBool(xHandles.m_xFogEnabled, false);
	}
	else if (pxMaterial)
	{
		xEffect.SetBool(xHandles.m_xFogEnabled, Asura_Fog::IsFogEnabled());

		if (Asura_Fog::IsFogEnabled())
		{
			xEffect.SetTexture(xHandles.m_xFogTexture, Asura_Xbox360_Fog::GetRenderTexture());

			const FXLVECTOR xFogData = {Asura_Fog::GetFogNearPlane(),
				Asura_Fog::GetOneOverFogPlaneSeparation(), 0.0f, 0.0f};

			xEffect.SetVector(xHandles.m_xFogParams, xFogData);

			FXLVECTOR xCol;

			if (pxMaterial->GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_ADDITIVE)
			{
				// Additive-blended materials fog out to black, not the fog colour
				xCol.x = 0;
				xCol.y = 0;
				xCol.z = 0;
			}
			else
			{
				xCol.x = Asura_Fog::GetFogColour().m_fR;
				xCol.y = Asura_Fog::GetFogColour().m_fG;
				xCol.z = Asura_Fog::GetFogColour().m_fB;
			}	

			xEffect.SetVector(xHandles.m_xFogColour,  xCol);
		}
	}
}

void Asura_Xbox360_MaterialResponse_System::AddPixelShader( const char* const szName,
	void* const pData, Asura_Xbox360_XPR& xXPR )
{
	IDirect3DPixelShader9* pxShader = NULL;

	const HRESULT xResult =
		Asura_Xbox360_Graphics::GetDevice()->CreatePixelShader( (const DWORD*)pData, &pxShader );

	if( SUCCEEDED( xResult ) && pxShader )
	{
		int iHash = ASURA_HASH_ID_UNSET;
		if( Asura_StringUtil::GetIntValue( szName, &iHash ) )
		{
			bool						bDidExist;
			Asura_Xbox360_MRPShader**	ppxMPRShader	= s_xPSTree.AddOrFindNode(iHash, &bDidExist);

			if(bDidExist)
			{
#if !defined ASURA_DONT_ASSERT_ON_DUPLICATE_RESOURCE_DATA
				// Might actually be out of memory, or during a tree traversal etc, but those
				// are covered by separate asserts
				Asura_Assertf( false, ("Pixel shader %s already loaded.", szName) );
#endif
			}
			else
			{
				*ppxMPRShader = new Asura_Xbox360_MRPShader( pxShader, xXPR );
			}
		}
		else
		{
			Asura_Assertf( false, ("Failed to get hash from '%s'.", szName) );
		}
	}
	else
	{
		Asura_Assert( false, "Couldn't create pixel shader." );
	}
}

Direct3D_PixelShader* Asura_Xbox360_MaterialResponse_System::GetPixelShader( const Asura_Hash_ID uHash )
{
	Asura_Xbox360_MRPShader** const ppxResult = s_xPSTree.GetDataByHash( uHash );
	Asura_Assertf( ppxResult, ("Failed to get pixel shader with hash ID 0x%X.", uHash) );

	return ppxResult ? (*ppxResult)->GetShader() : NULL;
}

// For some reason the compiler won't inline this on its own. Definitely a good idea though,
// otherwise you get a big load-hit-store penalty from the prologue/epilogue
static __forceinline void BindTexture( const int iTexID, const u_int uFetchReg )
{
	if( iTexID == ASURA_TEXTURE_NONE ) return;

	Asura_Xbox360_Texture* const pxTexture = Asura_Xbox360_TextureManagement::GetTexture( iTexID );
	if( pxTexture )
	{
		Asura_Xbox360_Graphics::GetDevice()->SetTexture( uFetchReg, pxTexture->GetD3DTexture() );
	}
}

void Asura_Xbox360_MaterialResponse_System::SetMaterialConstants(
	const Asura_MaterialResponse* const pxMaterial1,
	const Asura_MaterialResponse* const pxMaterial2, const Asura_Hash_ID uCBHash,
	const Asura_Animation* const pxAnim )
{
	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();

	const Asura_TextureAnimationController* const pxTextureAnimController
		= (pxAnim ? pxAnim->GetTextureAnimController() : NULL);

	// Use a constant buffer for the fixed material properties
	{
		Direct3D_ConstantBuffer* const pxCB = GetConstantBuffer( uCBHash );
		if( pxCB )
		{
			pxDevice->GpuLoadPixelShaderConstantF4( 0, 4, pxCB, 0 );
		}
	}

	// Need to recompute the self-illumination values because they can change each frame

	// Constants must be set in multiples of 4

	D3DVECTOR4* pxConstants = NULL;
	pxDevice->GpuBeginPixelShaderConstantF4( 4, &pxConstants, 4 );
	Asura_Assert( pxConstants, "Failed to begin pixel shader constants." );

	// Set self-illumination values
	{
		float fAmount = 0.0f;
		float fColourise = 0.0f;
		Asura_RGBA xColour(0);

		if( pxMaterial1 )
		{
			pxMaterial1->GetSelfIllumination( pxTextureAnimController, fAmount, fColourise, xColour );
		}

		pxConstants[0].x = fAmount * xColour.m_fR * fColourise;
		pxConstants[0].y = fAmount * xColour.m_fG * fColourise;
		pxConstants[0].z = fAmount * xColour.m_fB * fColourise;
		pxConstants[0].w = fAmount * (1.0f - fColourise);
	}

	// Detail map 2 parameters. Those for material 1 are in the constant buffer
	if( pxMaterial2 )
	{
		pxConstants[1].x = pxMaterial2->GetAlbedoDetailMapUVScale();
		pxConstants[1].y = pxMaterial2->GetAlbedoDetailMapValueScale();
		pxConstants[1].z = pxMaterial2->GetNormalDetailMapUVScale();
		pxConstants[1].w = pxMaterial2->GetNormalDetailMapValueScale();
	}
	else
	{
		pxConstants[1].x = pxConstants[1].y = pxConstants[1].z = pxConstants[1].w = 0.0f;
	}

	// Albedo modifier
	{
		Asura_RGBA xAlbedoModifier( 1.0f, 1.0f, 1.0f, 1.0f );

		Asura_MaterialResponse_System::AlbedoModifierCallback pfnCallback =
			Asura_MaterialResponse_System::GetAlbedoModifierCallback();

		if( pfnCallback )
		{
			const u_int uContext = (pxMaterial1 ? pxMaterial1->GetProjectContext() :
				ASURA_MATERIAL_RESPONSE_PROJECT_CONTEXT_INVALID);

			pfnCallback( pxAnim, uContext, xAlbedoModifier );
		}

		pxConstants[2].x = xAlbedoModifier.m_fR;
		pxConstants[2].y = xAlbedoModifier.m_fG;
		pxConstants[2].z = xAlbedoModifier.m_fB;
		pxConstants[2].w = 0.0f;
	}

	pxConstants[3].x = pxConstants[3].y = pxConstants[3].z = pxConstants[3].w = 0.0f;

	pxDevice->GpuEndPixelShaderConstantF4();

	// Also need to set texture animation constants for the vertex shader
	{
		// Grab the texture anim UV offsets for both channels
		Asura_TextureAnimationOutput xTexAnim1, xTexAnim2;

		const Asura_Hash_ID uAnim1 =
			(pxMaterial1 ? pxMaterial1->GetTextureAnim() : ASURA_HASH_ID_UNSET);
		const Asura_Hash_ID uAnim2 =
			(pxMaterial2 ? pxMaterial2->GetTextureAnim() : ASURA_HASH_ID_UNSET);

		Asura_TextureAnimation_System::GetAnimationState( uAnim1, pxTextureAnimController,
			&xTexAnim1 );
		Asura_TextureAnimation_System::GetAnimationState( uAnim2, pxTextureAnimController,
			&xTexAnim2 );

		const float afTexCoordOffset[4] = { xTexAnim1.m_fUOffset, xTexAnim1.m_fVOffset,
			xTexAnim2.m_fUOffset, xTexAnim2.m_fVOffset };

		pxDevice->SetVertexShaderConstantF( 1, afTexCoordOffset, 1 );
	}

	// Bind textures
	pxDevice->SetTexture( 0, s_pxMaterialsTexture );

	if( pxMaterial1 )
	{
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ),        2 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::NORMAL_MAP ),            4 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::SPECULAR_MAP ),          6 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ),      8 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP ), 10 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::PARALLAX_MAP ),         12 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::ENVIRONMENT_MAP ),      14 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::REFLECTIVITY_MASK ),    15 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::DETAIL_MAP ),           16 );
		BindTexture( pxMaterial1->GetTexture( Asura_MaterialResponse::DETAIL_NORMAL_MAP ),    18 );
	}

	if( pxMaterial2 )
	{
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE ),        3 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::NORMAL_MAP ),            5 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::SPECULAR_MAP ),          7 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP ),      9 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP ), 11 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::PARALLAX_MAP ),         13 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::DETAIL_MAP ),           17 );
		BindTexture( pxMaterial2->GetTexture( Asura_MaterialResponse::DETAIL_NORMAL_MAP ),    19 );
	}
}

/******************************************************************************/
/* Asura_Xbox360_MaterialResponse_Handles Class Functions                     */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_Handles::Asura_Xbox360_MaterialResponse_Handles()
//
// Constructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_MaterialResponse_Handles::Asura_Xbox360_MaterialResponse_Handles()
	: m_xMaterialProperties(0)
	, m_xMaterialsTexture(0)

	, m_xUseAlbedo(0)
	, m_xUseNormalMap(0)
	, m_xUseSpecularMap(0)
	, m_xUseTransparencyMap(0)
	, m_xUseSelfIlluminationMap(0)
	, m_xUseEnvironmentMap(0)
	, m_xUseReflectivityMask(0)
	, m_xUseParallaxMap(0)

	, m_xDualAlbedo(0)
	, m_xDualNormalMap(0)
	, m_xDualSpecularMap(0)
	, m_xDualTransparencyMap(0)
	, m_xDualSelfIlluminationMap(0)
	, m_xDualParallaxMap(0)

	, m_xAlbedo(0)
	, m_xNormalMap(0)
	, m_xSpecularMap(0)
	, m_xTransparencyMap(0)
	, m_xSelfIlluminationMap(0)
	, m_xEnvironmentMap(0)
	, m_xReflectivityMask(0)
	, m_xParallaxMap(0)

	, m_xAlbedo2(0)
	, m_xNormalMap2(0)
	, m_xSpecularMap2(0)
	, m_xTransparencyMap2(0)
	, m_xSelfIlluminationMap2(0)
	, m_xParallaxMap2(0)

	, m_xUV_Offsets(0)
	, m_xSelfIlluminationParameters(0)
	, m_xAlbedoModifier(0)
	, m_xBlendIntoAlbedo(0)

	, m_xRedSH(0)
	, m_xGreenSH(0)
	, m_xBlueSH(0)

	, m_xSHLightDirection(0)
	, m_xSHLightColour(0)

	, m_xFogEnabled(0)
	, m_xFogTexture(0)
	, m_xFogColour(0)
	, m_xFogParams(0)
{}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_MaterialResponse_Handles::InitialiseFromEffect()
//
// Initialises a bundle of MR parameter handles from a given effect
//
// Parameters:		xEffect - Effect from which to grab the handles
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_MaterialResponse_Handles::InitialiseFromEffect( const Asura_Xbox360_Effect& xEffect )
{
	#define GET_HANDLE( Prefix, Name ) m_x##Name = xEffect.GetParameterHandle( "g_" ## #Prefix ## #Name )

	GET_HANDLE( x, MaterialProperties );

	GET_HANDLE( b, UseAlbedo );
	GET_HANDLE( b, UseNormalMap );
	GET_HANDLE( b, UseSpecularMap );
	GET_HANDLE( b, UseTransparencyMap );
	GET_HANDLE( b, UseSelfIlluminationMap );
	GET_HANDLE( b, UseEnvironmentMap );
	GET_HANDLE( b, UseReflectivityMask );
	GET_HANDLE( b, UseParallaxMap );

	GET_HANDLE( b, DualAlbedo );
	GET_HANDLE( b, DualNormalMap );
	GET_HANDLE( b, DualSpecularMap );
	GET_HANDLE( b, DualTransparencyMap );
	GET_HANDLE( b, DualSelfIlluminationMap );
	GET_HANDLE( b, DualParallaxMap );

	GET_HANDLE( x, UV_Offsets );
	GET_HANDLE( x, SelfIlluminationParameters );
	GET_HANDLE( x, AlbedoModifier );
	GET_HANDLE( b, BlendIntoAlbedo );

	GET_HANDLE( x, RedSH );
	GET_HANDLE( x, GreenSH );
	GET_HANDLE( x, BlueSH );
	GET_HANDLE( x, SHLightDirection );
	GET_HANDLE( x, SHLightColour );

	GET_HANDLE( b, FogEnabled );
	GET_HANDLE( x, FogColour );
	GET_HANDLE( x, FogParams );
	GET_HANDLE( x, FogTexture );

	#undef GET_HANDLE

	// On 360, need to access sampler parameters, not texture
	#define GET_HANDLE( Prefix, Name ) m_x##Name = xEffect.GetParameterHandle( "g_" ## #Prefix ## #Name ## "Sampler" )

	GET_HANDLE( x, MaterialsTexture );

	GET_HANDLE( x, Albedo );
	GET_HANDLE( x, NormalMap );
	GET_HANDLE( x, SpecularMap );
	GET_HANDLE( x, TransparencyMap );
	GET_HANDLE( x, SelfIlluminationMap );
	GET_HANDLE( x, EnvironmentMap );
	GET_HANDLE( x, ReflectivityMask );
	GET_HANDLE( x, ParallaxMap );

	GET_HANDLE( x, Albedo2 );
	GET_HANDLE( x, NormalMap2 );
	GET_HANDLE( x, SpecularMap2 );
	GET_HANDLE( x, TransparencyMap2 );
	GET_HANDLE( x, SelfIlluminationMap2 );
	GET_HANDLE( x, ParallaxMap2 );

	#undef GET_HANDLE
}

/******************************************************************************/
/* Asura_MaterialResponse_System Static Class Functions                       */
/******************************************************************************/

void Asura_MaterialResponse_System::Platform_Initialise()
{
	Asura_Xbox360_MaterialResponse_System::Initialise();
}

void Asura_MaterialResponse_System::Platform_Reset( const bool bFullReset )
{
	Asura_Xbox360_MaterialResponse_System::Reset();
}

void Asura_MaterialResponse_System::Platform_Shutdown()
{
	Asura_Xbox360_MaterialResponse_System::Shutdown();
}

void Asura_MaterialResponse_System::Platform_InitialiseMaterial( const Asura_Hash_ID uNameHash,
	const Asura_MaterialResponse& xMaterial )
{
	Asura_Xbox360_MaterialResponse_System::CreateConstantBuffer( uNameHash, xMaterial );
}
