/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#include "Shared.fxh"

// Light clip space -> Light view space -> World space -> Camera view space -> Camera clip space
const float4x4 g_xLightClipSpaceToCameraClipSpace;

// Camera view space -> World space -> Light view space -> Light clip space
const float4x4 g_xCameraViewSpaceToLightClipSpace;

// Camera view space -> World space -> Light view space
const float4x4 g_xCameraViewSpaceToLightViewSpace;

// Camera view space -> World space
const float4x4 g_xCameraViewSpaceToWorldSpace;
// shadow caster position in world space
const float4 g_xLightPos;

const float4x4 g_xUVXForm;

const bool   g_bCubicMapShadow;

// 2 / dimensions of screen, in pixels. Used to convert VPOS pixel shader input into normalised coordinates.
const float g_fTwoOverScreenWidth;
const float g_fTwoOverScreenHeight;
const float2 g_fScreenOffset;

// Slopes of camera frustum. Used to reconstruct viewspace positions from screen coordinates and depth.
const float  g_fCameraTanHalfHorizFOV;
const float  g_fCameraTanHalfVertFOV;

// Is shadow filtering turned on?
const float g_fShadowFilterTextureScale;
const float g_fShadowFilterSize;

const float4 g_xShadowDirection;
const float  g_fShadowStrength;

const float g_fShadowSelfShadowTweak = 0.05f;

const bool   g_bCreatingShadowMask; 

const float g_fShadowTexSize;
// x is multiplied by the dot product, y is just added
const float3   g_xSelfShadowBias;

// Shadow frustum near/far values, used to map shadow depth textures into the appropriate range
const float	g_fShadowFrustumNearPlaneDistance;
const float	g_fOneOverShadowFrustumPlaneSeparation;


const sampler g_xNormalBuffer = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};

const sampler g_xDepthMtlBuffer = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};


const sampler g_xShadowBuffer = sampler_state
{
	AddressU = CLAMP;
	AddressV = CLAMP;
};
const sampler g_xShadowCubeBuffer = sampler_state
{

};


const sampler g_xKernelRotationSampler = sampler_state
{	
	AddressU = WRAP;
	AddressV = WRAP;
};



/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#define NUM_TAPS 4

#if (NUM_TAPS > 0)
static const float2 axFILTER_TAPS[ NUM_TAPS ] =
{
#if (NUM_TAPS == 8)
	{ -0.840144h, -0.073580h },
	{ -0.695914h,  0.457137h },
	{ -0.203345h,  0.620716h },
	{  0.473434h, -0.480026h },
	{  0.519456h,  0.767022h },
	{  0.185461h, -0.893124h },
	{  0.896420h,  0.412458h },
	{ -0.321940h, -0.932615h },
#elif (NUM_TAPS == 4)
	{ -1.0h,  0.0h },
	{  0.0h, -1.0h },
	{  1.0h,  0.0h },
	{  0.0h,  1.0h },
#elif (NUM_TAPS == 12)
	{ -0.326212, -0.405805 },
	{ -0.840144, -0.073580 },
	{ -0.695914,  0.457137 },
	{ -0.203345,  0.620716 },
	{  0.962340, -0.194983 },
	{  0.473434, -0.480026 },
	{  0.519456,  0.767022 },
	{  0.185461, -0.893124 },
	{  0.507431,  0.064425 },
	{  0.896420,  0.412458 },
	{ -0.321940, -0.932615 },
	{ -0.791559, -0.597705 },
#endif	// (NUM_TAPS == 12)
};
#endif


/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

// The vertex shader recreates the shadow frustum in world space, then projects
// it as an object in the standard camera view

float4 VShader( const float4 xInPos : POSITION ) : POSITION
{
	return mul( xInPos, g_xLightClipSpaceToCameraClipSpace );
}

float4 GetCameraViewSpacePosition(const float2 xScreenPos, out half3 xNormal)
{
	float4 xNormalMtl;
	float4 xDepthMtl;
	
	asm
	{
		tfetch2D xNormalMtl,   xScreenPos, g_xNormalBuffer,   UnnormalizedTextureCoords=true, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xDepthMtl, xScreenPos, g_xDepthMtlBuffer, UnnormalizedTextureCoords=true, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled
	};
	
	xNormal = (xNormalMtl * 2.0f) - 1.0f;

	// Reconstruct the camera view space position. I'm not 100% sure I understand why we can't construct the clip space position
	// and then push it through the inverse projection matrix, but I think it's because we don't have a value for w and so can't
	// reverse the transform correctly. Note that the y coordinate gets flipped; this is because Asura's Y coordinates are upside down
	return float4( g_fCameraTanHalfHorizFOV * xDepthMtl.x * ((xScreenPos.x - g_fScreenOffset.x)* g_fTwoOverScreenWidth  - 1.0),
					 g_fCameraTanHalfVertFOV  * xDepthMtl.x * -((xScreenPos.y - g_fScreenOffset.y) * g_fTwoOverScreenHeight - 1.0),
					 xDepthMtl.x,
					 1.0);
}


float3 GetCubicMap_ShadowUV(const float4 xCameraViewSpacePosition, out float fSceneLightspaceDepth, out float4 xLightVector)
{
	float4 xWorldSpacePosition = mul(xCameraViewSpacePosition, g_xCameraViewSpaceToWorldSpace);
	xLightVector = g_xLightPos - xWorldSpacePosition;
	
	fSceneLightspaceDepth = length(xLightVector);
	
	xLightVector /= fSceneLightspaceDepth;

	fSceneLightspaceDepth = (fSceneLightspaceDepth - g_fShadowFrustumNearPlaneDistance ) * g_fOneOverShadowFrustumPlaneSeparation;
	
	float3 xShadowBufferTexcoords;
	xShadowBufferTexcoords = xLightVector;
	xShadowBufferTexcoords.xz = -xShadowBufferTexcoords.xz;

	return xShadowBufferTexcoords;
}

float GetProjected_Depth(const float4 xCameraViewSpacePosition)
{
	// We want a linear distribution of depth in the shadow buffer so we don't divide z by w - we use the near and far planes of the shadow
	// frustum instead.
	float4 xLightViewSpacePosition = mul(xCameraViewSpacePosition, g_xCameraViewSpaceToLightViewSpace);
	return (xLightViewSpacePosition.z - g_fShadowFrustumNearPlaneDistance ) * g_fOneOverShadowFrustumPlaneSeparation;

}

float4 GetProjected_ShadowUV(const float4 xCameraViewSpacePosition)
{
	float4 xShadowBufferTexcoords = mul(xCameraViewSpacePosition, g_xCameraViewSpaceToLightClipSpace);
	xShadowBufferTexcoords =  mul(g_xUVXForm, xShadowBufferTexcoords);
	return xShadowBufferTexcoords;
}

float2 GetPerPixelRotation(float2 xUV, float fScale, uniform bool bScreenSpace)
{
	float2 fSinCos;
	if (bScreenSpace)
	{
		fSinCos = (TexFetch2D_LinearLinearLinear(g_xKernelRotationSampler, xUV*1.0f/64.0f).gb *2.0f - 1.0f) * fScale;
	} 
	else 
	{
		fSinCos = (TexFetch2D_LinearLinearLinear(g_xKernelRotationSampler, xUV*16).gb *2.0f - 1.0f) * fScale * 0.66f;
		fSinCos += (TexFetch2D_LinearLinearLinear(g_xKernelRotationSampler, xUV*64).gb *2.0f - 1.0f) * fScale * 0.33f;
	}
	return fSinCos;
}

float4 Fetch4(sampler g_xSampler, float2 xUV)
{
	float4 xTexture1, xTexture2, xTexture3, xTexture4; 
	asm
	{
		tfetch2D xTexture1, xUV, g_xSampler,  UnnormalizedTextureCoords=true, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 0.5
		tfetch2D xTexture2, xUV, g_xSampler,  UnnormalizedTextureCoords=true, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX = 1.5, OffsetY = 0.5
		tfetch2D xTexture3, xUV, g_xSampler,  UnnormalizedTextureCoords=true, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX = 0.5, OffsetY = 1.5
		tfetch2D xTexture4, xUV, g_xSampler,  UnnormalizedTextureCoords=true, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX = 1.5, OffsetY = 1.5
	};
	return float4(xTexture1.r, xTexture2.r, xTexture3.r, xTexture4.r);
}

float f2wayLerp(float4 xVals, float2 fT)
{
	return lerp(lerp(xVals.x, xVals.y, fT.x),
				lerp(xVals.z, xVals.w, fT.x), fT.y);
}


float SampleShadowBuffer(float3 xShadowBufferTexcoords, const bool bUseCubic, float fSceneLightspaceDepth, uniform bool bEnablePCF)
{
	float fOccluderLightspaceDepth;
	if (bUseCubic)
	{
		float4 xShadowTexture; 
		const float f1_5 = 1.5f;
		asm
		{
			cube xShadowTexture, xShadowBufferTexcoords.zzxy, xShadowBufferTexcoords.yxz
			rcp xShadowTexture.z, xShadowTexture_abs.z
			mad xShadowTexture.xy, xShadowTexture, xShadowTexture.zzzw, f1_5

			tfetchCube xShadowTexture, xShadowTexture.yxw, g_xShadowCubeBuffer, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled
		};
		fOccluderLightspaceDepth = saturate(sign(fSceneLightspaceDepth.x - xShadowTexture.x));
	} 
	else 
	{
		if (bEnablePCF)
		{
			xShadowBufferTexcoords.xy *= g_fShadowTexSize;
			float2 xLerpFactors = frac(xShadowBufferTexcoords.xy);
			xShadowBufferTexcoords.xy -= xLerpFactors;

			fOccluderLightspaceDepth = f2wayLerp(saturate(sign(fSceneLightspaceDepth.xxxx - Fetch4(g_xShadowBuffer, xShadowBufferTexcoords.xy))), xLerpFactors);
		} else {
			float4 xTexture;
			asm
			{
				tfetch2D xTexture, xShadowBufferTexcoords, g_xShadowBuffer, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled
			};
			fOccluderLightspaceDepth = saturate(sign(fSceneLightspaceDepth.x - xTexture.x));
		}
	}
	return fOccluderLightspaceDepth;
}

float2 GetNoiseOffset(float2 xUV, float fSceneLightspaceDepth, float2 xScreenPos, uniform bool bUseNoise, uniform bool bScreenSpace)
{
	// Calculate a value to scale the sampling region on the shadow buffer by
	const float fScaling = max(0.0f, g_fShadowFilterSize + saturate(2.0f * fSceneLightspaceDepth - 1.0f) * 2.0);
	const float fScale = g_fShadowFilterTextureScale * fScaling;

	// To reduce banding, we randomly rotate the filter kernel per-pixel. Uses a texture of pre-generated	
	// sin/cos for random rotation angles
	return bUseNoise ? GetPerPixelRotation(bScreenSpace ? xScreenPos : xUV.xy, fScale, bScreenSpace) : float2(fScale, 0.0f);
}


float IsShadowed_SingleTap(float4 xShadowBufferTexcoords, float fSceneLightspaceDepth, float2 xScreenPos, uniform bool bUseCubic, uniform bool bUseNoise, uniform bool bScreenBasedNoise, uniform bool bEnablePCF)
{
	float fShadow;


	float3 xOffset = float3(GetNoiseOffset(xShadowBufferTexcoords.xy, fSceneLightspaceDepth, xScreenPos, bUseNoise, bScreenBasedNoise), 0.0f);						               


	fShadow = SampleShadowBuffer(xShadowBufferTexcoords.xyz + xOffset.xyz, bUseCubic, fSceneLightspaceDepth, bEnablePCF) ;

	return fShadow;
}


float IsShadowed_MultiTap(float4 xShadowBufferTexcoords, float fSceneLightspaceDepth, float2 xScreenPos, uniform bool bUseCubic, uniform bool bUseNoise, uniform bool bScreenBasedNoise, uniform bool bEnablePCF)
{
	float fShadow = 0;
	
	// Regardless of whether this pixel is actually shadowed, allow 'nearby' occluders to contribute to the shadow value
	float2 fSinCos = GetNoiseOffset(xShadowBufferTexcoords.xy, fSceneLightspaceDepth, xScreenPos, bUseNoise, bScreenBasedNoise);

	
	// For each tap,
	for( int i = 0; i < NUM_TAPS; ++i )
	{
		// calculate the offset texture coordinates for sampling the occluder depth			
		float3 xOffset;						               
		xOffset.x = dot(axFILTER_TAPS[i].xy , fSinCos);
		xOffset.y = dot(axFILTER_TAPS[i].xy , float2(fSinCos.y, -fSinCos.x));
		xOffset.z = 0.0f;

		const float4 xTapCoord = xShadowBufferTexcoords + xOffset.xyxy;
		// See how far away this pixel is from the occluder
		fShadow += SampleShadowBuffer(xTapCoord, bUseCubic, fSceneLightspaceDepth, bEnablePCF);
		
		// If the occluder is behind the pixel, fShadow2 has a negative sign - translate this into 'no shadow'
	}
	// Average out the contributions across all taps
	fShadow /= NUM_TAPS;	
	
	return fShadow;	
}

float ScaleShadow(float fShadow, float fDot, float fDistScale)
{
	fShadow = fShadow * g_fShadowStrength;
	
	if (!g_bCreatingShadowMask)
	{
		fShadow = saturate(fShadow * fDistScale);		
		
	//	Taking this out, as it causes more problems than it fixes in SS2... shadows of a helmet on a character's face look particularly bad...
		fShadow *= fDot;
	} 
	return fShadow;
}


float BiasDepthValue(float fSceneLightspaceDepth, float3 xLightVector, half3 xNormal, out float fDot)
{
	fDot = saturate( dot( xLightVector, xNormal) );	
	fSceneLightspaceDepth -= ((1.0f - fDot) * g_xSelfShadowBias.x + g_xSelfShadowBias.y + fSceneLightspaceDepth * g_xSelfShadowBias.z);
	fDot = saturate(2.0f * fDot);

	return fSceneLightspaceDepth;
}

float4 CalculateShadowParams(const float2 xScreenPos, uniform bool g_bCubicMapShadow, out float fDot, out float fSceneLightspaceDepth)
{
	half3 xNormal;
	const float4 xCameraViewSpacePosition = GetCameraViewSpacePosition(xScreenPos, xNormal);

	float4 xShadowBufferTexcoords;

	if (g_bCubicMapShadow)
	{
		float4 xLightVector;
		xShadowBufferTexcoords = float4(GetCubicMap_ShadowUV(xCameraViewSpacePosition, fSceneLightspaceDepth, xLightVector), 1);
		
		fSceneLightspaceDepth = BiasDepthValue(fSceneLightspaceDepth, xLightVector.xyz, xNormal, fDot);

	} else {
		xShadowBufferTexcoords = GetProjected_ShadowUV(xCameraViewSpacePosition);

		float3 xShadowClipSpace = xShadowBufferTexcoords.xyz / xShadowBufferTexcoords.w;

 		clip( xShadowClipSpace);
 		clip( 1 - xShadowClipSpace);

		// fSceneLightspaceDepth is the scene depth from the light's POV
		fSceneLightspaceDepth = GetProjected_Depth(xCameraViewSpacePosition);

		// Discard if we're behind the shadow near plane
		clip( fSceneLightspaceDepth);

		xShadowBufferTexcoords = float4(xShadowClipSpace.xyz, 1.0f);

		fSceneLightspaceDepth = BiasDepthValue(fSceneLightspaceDepth, g_xShadowDirection, xNormal, fDot);
	}

	return xShadowBufferTexcoords;
}


float4 PShader_MultiTap( const float2 xScreenPos : VPOS, uniform bool g_bCubicMapShadow, uniform bool bUseNoise, uniform bool bScreenBasedNoise, uniform bool bEnablePCF) : COLOR
{
	float fDot;
	float fSceneLightspaceDepth;

	float4 xShadowBufferTexcoords = CalculateShadowParams(xScreenPos + g_xVPOSOffsetVector, g_bCubicMapShadow, fDot, fSceneLightspaceDepth);

	float fShadow = IsShadowed_MultiTap(xShadowBufferTexcoords, fSceneLightspaceDepth, xScreenPos, g_bCubicMapShadow, bUseNoise, bScreenBasedNoise, bEnablePCF);

	// Scale the shadow according to how far this pixel is away from the light
	return float4( 0.0, 0.0, 0.0, saturate( ScaleShadow(fShadow, fDot, 1.0f - fSceneLightspaceDepth) ) );	
}

float4 PShader_SingleTap( const float2 xScreenPos : VPOS, uniform bool g_bCubicMapShadow, uniform bool bUseNoise, uniform bool bScreenBasedNoise, uniform bool bEnablePCF) : COLOR
{
	float fDot;	
	float fSceneLightspaceDepth;

	float4 xShadowBufferTexcoords = CalculateShadowParams(xScreenPos + g_xVPOSOffsetVector, g_bCubicMapShadow, fDot, fSceneLightspaceDepth);

	float fShadow = IsShadowed_SingleTap(xShadowBufferTexcoords, fSceneLightspaceDepth, xScreenPos, g_bCubicMapShadow, bUseNoise, bScreenBasedNoise, bEnablePCF);

	// Scale the shadow according to how far this pixel is away from the light
	return float4( 0.0, 0.0, 0.0, saturate( ScaleShadow(fShadow, fDot, 1.0f - fSceneLightspaceDepth) ) );	

}

float Chebychev( const float t, const float m, const float s2 )
{
	const float d = t - m;
	return s2 / (s2 + d*d);
}

#define STANDARD_NUM_TAPS  8
static const float2 axSTANDARD_FILTER_TAPS[ STANDARD_NUM_TAPS ] =
{
	{ -0.840144h, -0.073580h },
	{ -0.695914h,  0.457137h },
	{ -0.203345h,  0.620716h },
	{  0.473434h, -0.480026h },
	{  0.519456h,  0.767022h },
	{  0.185461h, -0.893124h },
	{  0.896420h,  0.412458h },
	{ -0.321940h, -0.932615h }
};

float4 PShader_Standardised_NonCubic( const float2 xScreenPos : VPOS ) : COLOR
{
	float fDot;	
	
	half3 xNormal;

	const float4 xCameraViewSpacePosition = GetCameraViewSpacePosition(xScreenPos + g_xVPOSOffsetVector, xNormal);
	
	float4 xShadowBufferTexcoords = GetProjected_ShadowUV(xCameraViewSpacePosition);

	float3 xShadowClipSpace = xShadowBufferTexcoords.xyz / xShadowBufferTexcoords.w;

 	clip( xShadowClipSpace);
 	clip( 1 - xShadowClipSpace);

	// fSceneLightspaceDepth is the scene depth from the light's POV
	float fSceneLightspaceDepth = GetProjected_Depth(xCameraViewSpacePosition);

	// Discard if we're behind the shadow near plane
	clip( fSceneLightspaceDepth);

	xShadowBufferTexcoords = float4(xShadowClipSpace.xyz, 1.0f);

	fSceneLightspaceDepth = BiasDepthValue(fSceneLightspaceDepth, g_xShadowDirection, xNormal, fDot);
			
	// Calculate a value to scale the sampling region on the shadow buffer by
	const float fScaling = max(0.0f, g_fShadowFilterSize + saturate(2.0f * fSceneLightspaceDepth - 1.0f) * 2.0);
	const float fScale = g_fShadowFilterTextureScale * fScaling;

	float fShadow = 0;
	
	// Regardless of whether this pixel is actually shadowed, allow 'nearby' occluders to contribute to the shadow value
	float2 fSinCos = (TexFetch2D_LinearLinearLinear(g_xKernelRotationSampler, xShadowBufferTexcoords.xy*16).gb *2.0f - 1.0f) * fScale * 0.66f;
	//fSinCos += (TexFetch2D_LinearLinearLinear(g_xKernelRotationSampler, xShadowBufferTexcoords.xy*64).gb *2.0f - 1.0f) * fScale * 0.33f;
	float2 fSinCos2 = float2(fSinCos.y, -fSinCos.x);

	for( int i = 0; i < 4; ++i )
	{
		// calculate the offset texture coordinates for sampling the occluder depth			
		float2 xOffset;						               
		xOffset.x = dot(axSTANDARD_FILTER_TAPS[i].xy , fSinCos);
		xOffset.y = dot(axSTANDARD_FILTER_TAPS[i].xy , fSinCos2);

		float4 xTapCoord = xShadowBufferTexcoords + xOffset.xyxy;

		// See how far away this pixel is from the occluder
		float4 xTexture;
		asm
		{
			tfetch2D xTexture, xTapCoord, g_xShadowBuffer, MagFilter=point, MinFilter=point, MipFilter=point, AnisoFilter=disabled
		};

		fShadow += (fSceneLightspaceDepth.x > xTexture.x) ? 0.25f : 0.0f;
	}
	
	//if ( fShadow < 1.0f ) fShadow= saturate(fSinCos.x)*fShadow*2);
		
	fShadow *= g_fShadowStrength;

	if (!g_bCreatingShadowMask)
	{
		fShadow = saturate(fShadow * (1.0f-fSceneLightspaceDepth));		
		fShadow *= fDot;
	} 

	// Scale the shadow according to how far this pixel is away from the light
	return float4( 0.0, 0.0, 0.0, saturate(fShadow) );	
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

// PCF Off
//******************************************************************************************************
// No Noise


technique NonCubic_SingleTap_NoPCF_NoNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false /*cubic*/, false /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}
technique NonCubic_MultiTap_NoPCF_NoNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false /*cubic*/, false /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}

technique Cubic_SingleTap_NoPCF_NoNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(true /*cubic*/, false /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}
technique Cubic_MultiTap_NoPCF_NoNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(true /*cubic*/, false /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}
//******************************************************************************************************
// Noise On

technique NonCubic_SingleTap_NoPCF_Noise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false /*cubic*/, true /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}
technique NonCubic_MultiTap_NoPCF_Noise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false /*cubic*/, true /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}

technique Cubic_SingleTap_NoPCF_Noise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(true /*cubic*/, true /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}
technique Cubic_MultiTap_NoPCF_Noise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(true /*cubic*/, true /*Noise*/, false /*Screen*/, false /* PCF */);
	}
}

//******************************************************************************************************
// Screen based Noise On

technique NonCubic_SingleTap_NoPCF_ScreenNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false /*cubic*/, true /*Noise*/, true /*Screen*/, false /* PCF */);
	}
}
technique NonCubic_MultiTap_NoPCF_ScreenNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false /*cubic*/, true /*Noise*/, true /*Screen*/, false /* PCF */);
	}
}

technique Cubic_SingleTap_NoPCF_ScreenNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(true /*cubic*/, true /*Noise*/, true /*Screen*/, false /* PCF */);
	}
}
technique Cubic_MultiTap_NoPCF_ScreenNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(true /*cubic*/, true /*Noise*/, true /*Screen*/, false /* PCF */);
	}
}


//**************************************************************************************************************
//* PCF On

// No Noise


technique NonCubic_SingleTap_PCF_NoNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false /*cubic*/, false /*Noise*/, false /*Screen*/, true /* PCF */);
	}
}
technique NonCubic_MultiTap_PCF_NoNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false /*cubic*/, false /*Noise*/, false /*Screen*/, true /* PCF */);
	}
}
// Noise On

technique NonCubic_SingleTap_PCF_Noise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false /*cubic*/, true /*Noise*/, false /*Screen*/, true /* PCF */);
	}
}
technique NonCubic_MultiTap_PCF_Noise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false /*cubic*/, true /*Noise*/, false /*Screen*/, true /* PCF */);
	}
}

//******************************************************************************************************
// Screen based Noise On

technique NonCubic_SingleTap_PCF_ScreenNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_SingleTap(false /*cubic*/, true /*Noise*/, true /*Screen*/, true /* PCF */);
	}
}
technique NonCubic_MultiTap_PCF_ScreenNoise
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_MultiTap(false /*cubic*/, true /*Noise*/, true /*Screen*/, true /* PCF */);
	}
}

technique Standardised_NonCubic
{
	pass
	{
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Standardised_NonCubic();
	}
}



