/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#include "Botanicals_Common.fxh"

shared const float4x4 g_xViewMatrix;
shared const float4x4 g_xProjectionMatrix;

const float  g_fInstanceSize; // In vertices. Ints are cast to floats anyway outside of loop and vertex indices
const float  g_fFirstInstance;

const float3 g_xCameraRightVector;
const float3 g_xCameraUpVector;
const float3 g_xCameraForwardsVector;

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex_Botanicals_Prelighting VShader( const int iIndex : INDEX )
{
	// Integer arithmetic has some rounding errors; see help topic "Vertex Fetching in HLSL"
	// and thread "Hardware instancing oddness" in xds360.graphics
	
	// TODO: this still produces errors in AVP_Demo level, need better kludge

	int iInstanceIndex = (iIndex + 0.5) / g_fInstanceSize;
	const int iVertexIndex = iIndex - iInstanceIndex * g_fInstanceSize; // We don't use indexing for billboards
	const int iBillboardIndex = iVertexIndex / 4;

	iInstanceIndex += g_fFirstInstance;

	float2 xVertexPositionOffset; // xy offset from billboard centre, in view space, of this vertex
	float2 xUV;

	float4 xBillboardPositionOffset; // xyz offset from instance position, in object space, of this billboard
	float3 xNormal;
	float3 xTangent;
	float3 xBinormal;

	float3 xInstanceOrientation0;
	float3 xInstanceOrientation1;
	float3 xInstanceOrientation2;
	float4 xInstancePositionAndAlpha;
	float4 xRedSH;
	float4 xGreenSH;
	float4 xBlueSH;	
	float  fBoneWeight;
	float4 xBoneOffsetAndScale;
	
	asm
	{
		// Get the billboard vertex attributes
		vfetch xVertexPositionOffset.xy, iVertexIndex, position0
		vfetch xUV.xy, iVertexIndex, texcoord0

		// Retrieve the billboard properties
		vfetch xBillboardPositionOffset, iBillboardIndex, position1
		vfetch xNormal.xyz,   iBillboardIndex, normal0
		vfetch xTangent.xyz,  iBillboardIndex, tangent0
		vfetch xBinormal.xyz, iBillboardIndex, binormal0
		vfetch fBoneWeight.x, iBillboardIndex, blendweight

		// Grab the instance's world matrix
		vfetch xInstanceOrientation0.xyz, iInstanceIndex, texcoord1
		vfetch xInstanceOrientation1.xyz, iInstanceIndex, texcoord2
		vfetch xInstanceOrientation2.xyz, iInstanceIndex, texcoord3
		vfetch xInstancePositionAndAlpha, iInstanceIndex, texcoord4

		// Grab the spherical harmonic
		vfetch xRedSH,   iInstanceIndex, texcoord5
		vfetch xGreenSH, iInstanceIndex, texcoord6
		vfetch xBlueSH,  iInstanceIndex, texcoord7

		// Bone Position Offset
		vfetch xBoneOffsetAndScale, iInstanceIndex, texcoord8
	};

	const float3x4	xWorldMatrix	= float3x4(float4(xInstanceOrientation0.xyz * xBoneOffsetAndScale.w, xInstancePositionAndAlpha.x),
											   float4(xInstanceOrientation1.xyz * xBoneOffsetAndScale.w, xInstancePositionAndAlpha.y),
											   float4(xInstanceOrientation2.xyz * xBoneOffsetAndScale.w, xInstancePositionAndAlpha.z));

	xNormal = mul( xWorldMatrix, xNormal ); // Only first 3 columns of world matrix will be used, since 3-vector

	OutputVertex_Botanicals_Prelighting xOut;

	float3 xWorldPosition = mul( xWorldMatrix, xBillboardPositionOffset ); // Matrix is transposed to get the (0 0 0 1) in the last register (as it were)
	xOut.m_xPosition.xyz  = xWorldPosition;
	xOut.m_xPosition.w	  = 1.0;

	xOut.m_xPosition	 = mul( xOut.m_xPosition, g_xViewMatrix ); // TODO: try premultiplying this with instance world matrix CPU-side when LODs are calculated
	xOut.m_xPosition.xy += xVertexPositionOffset * xBoneOffsetAndScale.w;
	
	//xOut.m_xPosition.xyz+= (fInstanceScale * fBoneWeight * mul( xWorldMatrix, xBoneOffset ));
	xOut.m_xPosition.xyz = Bot_CalcNewBonedPosition( xOut.m_xPosition.xyz, fBoneWeight, xBoneOffsetAndScale.w * mul( xWorldMatrix, xBoneOffsetAndScale.xyz ));
	
	xOut.m_xPosition	 = mul( xOut.m_xPosition, g_xProjectionMatrix );

	xOut.m_xUVPair.xyzw = xUV.xyxy;	
	
	
	

		
	// Lighting calculations all done in world space now
	// --------------------------------------------------------------------------------------	
	xOut.m_afAlphaCut   = xInstancePositionAndAlpha.w;
		
	const float3x3 xMatrix = float3x3( -g_xCameraRightVector,
		 							    g_xCameraUpVector,
									    g_xCameraForwardsVector );
	
	xOut.m_xTangent     = -mul(  xMatrix, -xTangent  );
	xOut.m_xBinormal    =  mul(  xMatrix,  xBinormal );
	xOut.m_xNormal      = -g_xCameraForwardsVector; 
	
	
	// Light it - from front and back, then lerp
	half3 xColour, xBackColour;		
	half4 xNorm = half4(xOut.m_xNormal,1);						
	xOut.m_xColour.rgb = SHLookup( xNorm, xRedSH, xGreenSH, xBlueSH );
		
	xNorm.rgb=-xNorm.rgb;	
	xOut.m_xBackColour.rgb =  SHLookup( xNorm, xRedSH, xGreenSH, xBlueSH );
	 
		
	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xWorldPosition.xyz;
	xOut.m_xCameraDirDistance.w   = xOut.m_xPosition.w;

	return xOut;
}


/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass
	{
		AlphaTestEnable = TRUE;
		AlphaFunc = GREATER;
		AlphaRef = 0;
		CullMode = CW;
		ZEnable = TRUE;
		ZWriteEnable = TRUE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Botanicals_Prelighting();
	}
}