/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

// Includes must be this way round here
#include "Shared.fxh"
#include "MaterialResponse_Shared.fxh"

/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

const float3 g_xPrelightingColourFilter : register(c0);

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

struct InputVertex
{
	float4 m_xPosition : POSITION;
	float4 m_xColour   : COLOR;

	float3 m_xNormal   : NORMAL;
	float3 m_xTangent  : TANGENT0;
	float3 m_xBinormal : BINORMAL0;
	float3 m_xTangent2  : TANGENT1;
	float3 m_xBinormal2 : BINORMAL1;

	float2 m_axUV[2]    : TEXCOORD0;

	float4 m_xRedSH     : TEXCOORD2;
	float4 m_xGreenSH   : TEXCOORD3;
	float4 m_xBlueSH    : TEXCOORD4;
};

struct OutputVertex
{
	float4 m_xPosition              : POSITION;

	float4 m_xUVPair                : TEXCOORD0;

	float4 m_xRedSH                 : TEXCOORD1;
	float4 m_xGreenSH               : TEXCOORD2;
	float4 m_xBlueSH                : TEXCOORD3;

	float4 m_xCameraDirDistance     : TEXCOORD4; // Camera direction, and distance from view plane in w
	float4 m_xTangentSpaceCameraDir : TEXCOORD5;

	float3 m_xTangent               : TANGENT0;
	float3 m_xBinormal              : BINORMAL0;
	float3 m_xTangent2              : TANGENT1;
	float3 m_xBinormal2             : BINORMAL1;
	float3 m_xNormal                : NORMAL;
};

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

OutputVertex VShader( const InputVertex xIn )
{
	OutputVertex xOut;

	xOut.m_xPosition = mul( xIn.m_xPosition, g_xViewProjectionMatrix );
	xOut.m_xTangentSpaceCameraDir.w = xIn.m_xColour.w;

	// Pack UV co-ords to save an interpolator
	xOut.m_xUVPair.xy = xIn.m_axUV[0];
	xOut.m_xUVPair.zw = xIn.m_axUV[1];
	xOut.m_xUVPair = MaterialResponse_AnimateUVs( xOut.m_xUVPair );

	// (AAM - 07.01.2008) Prelighting scale now includes the x4 from the tone mapping
	xOut.m_xRedSH = xIn.m_xRedSH * g_xPrelightingColourFilter.r;
	xOut.m_xGreenSH = xIn.m_xGreenSH * g_xPrelightingColourFilter.g;
	xOut.m_xBlueSH = xIn.m_xBlueSH * g_xPrelightingColourFilter.b;

	// (AAM 13.02.2007) - Used to output data in tangent space, but need world space for
	// reflectivity transforms

	xOut.m_xTangent = xIn.m_xTangent;
	xOut.m_xBinormal = xIn.m_xBinormal;
	xOut.m_xTangent2 = xIn.m_xTangent2;
	xOut.m_xBinormal2 = xIn.m_xBinormal2;
	xOut.m_xNormal = xIn.m_xNormal;

	xOut.m_xCameraDirDistance.xyz = g_xCameraPosition.xyz - xIn.m_xPosition.xyz;
	xOut.m_xCameraDirDistance.w = xOut.m_xPosition.w;

	const float3x3 xTangentToWorldMatrix = float3x3( xIn.m_xTangent,
													 xIn.m_xBinormal,
													 xIn.m_xNormal );

	// Note implicit transpose in the mul, by swapping operand order
	xOut.m_xTangentSpaceCameraDir.xyz = mul( xTangentToWorldMatrix, xOut.m_xCameraDirDistance.xyz );

	return xOut;
}

[sampreg( 0, 19 )]
DeferredOutputPixel PShader_Deferred( const OutputVertex xIn, uniform bool bUseResponse )
{
	const half3 xInterpolatedTangent  = normalize( xIn.m_xTangent );
	const half3 xInterpolatedBinormal = normalize( xIn.m_xBinormal );
	const half3 xInterpolatedTangent2  = normalize( xIn.m_xTangent2 );
	const half3 xInterpolatedBinormal2 = normalize( xIn.m_xBinormal2 );
	const half3 xInterpolatedNormal   = normalize( xIn.m_xNormal );

	const half3x3 xTangentToWorldMatrix
		= { xInterpolatedTangent, xInterpolatedBinormal, xInterpolatedNormal };
	const half3x3 xTangentToWorldMatrix2
		= { xInterpolatedTangent2, xInterpolatedBinormal2, xInterpolatedNormal };

	const half3 xViewDir = normalize( xIn.m_xCameraDirDistance.xyz );
	const half3 xTangentSpaceViewDir = normalize( xIn.m_xTangentSpaceCameraDir );

	const half hDualTextureLerp = xIn.m_xTangentSpaceCameraDir.w;

	half4 xUVPair = xIn.m_xUVPair;
	ApplyParallaxMapping( xTangentSpaceViewDir, hDualTextureLerp, xUVPair );

	const half4 xWorldSpaceNormal = NormalLookup( xUVPair, hDualTextureLerp, xTangentToWorldMatrix, xTangentToWorldMatrix2 );

	half3 xAlbedo;
	half hSpecular;
	float fMaterialsTextureY;
	half4 xColour;

#if defined( NOT_EFFECT)
	if( true )
#else
	if( bUseResponse )
#endif
	{
		const half4 xSHLookupDir = half4( xInterpolatedNormal, 1.0h );
		const float3 xLightColour
			= SHLookup( xSHLookupDir, xIn.m_xRedSH, xIn.m_xGreenSH, xIn.m_xBlueSH ) * 2.0f;

		const float fLightStrength = GetLuminance( xLightColour );

		xColour = MaterialResponse( xUVPair, hDualTextureLerp, xWorldSpaceNormal, xViewDir,
			xInterpolatedNormal, xLightColour, xLightColour, fLightStrength, false, xAlbedo,
			hSpecular, fMaterialsTextureY );

		xColour.rgb += SelfIlluminationLookup( xUVPair, hDualTextureLerp, xAlbedo );
	}
	else
	{
		fMaterialsTextureY = g_xMaterialProperties.x;

		// For this base geometry pass, we'll just use the SH lighting + self illumination

		const float3 xAlbedoTimes2 = AlbedoLookup( xUVPair,
												   hDualTextureLerp,
												   xWorldSpaceNormal,
												   xViewDir,
												   xAlbedo );

		xColour.rgb =
			SHLookup( xWorldSpaceNormal, xIn.m_xRedSH, xIn.m_xGreenSH, xIn.m_xBlueSH ) * xAlbedoTimes2
			+ SelfIlluminationLookup( xUVPair, hDualTextureLerp, xAlbedo );

		// Transparent env strips aren't sorted - but we need to use the transparency maps if present
		// to allow alpha decals
		xColour.a = TransparencyLookup( xUVPair, hDualTextureLerp );

		hSpecular = SpecularLookup( xUVPair, hDualTextureLerp );
	}

	// Do 1-bit alpha testing like this, since it seems to be done for deferred objects this way
	clip( xColour.a - 0.05h );

	DeferredOutputPixel xOut = Deferred_FillOutputPixel(
										xColour,
										xWorldSpaceNormal.xyz,
										xIn.m_xCameraDirDistance.w,
										xAlbedo,
										hSpecular,
										fMaterialsTextureY
										);
	return( xOut);
}

float4 PShader_Forward( const OutputVertex xIn ) : COLOR
{
	const half3 xInterpolatedTangent  = normalize( xIn.m_xTangent );
	const half3 xInterpolatedBinormal = normalize( xIn.m_xBinormal );
	const half3 xInterpolatedTangent2  = normalize( xIn.m_xTangent2 );
	const half3 xInterpolatedBinormal2 = normalize( xIn.m_xBinormal2 );
	const half3 xInterpolatedNormal   = normalize( xIn.m_xNormal );

	const half3x3 xTangentToWorldMatrix
		= { xInterpolatedTangent, xInterpolatedBinormal, xInterpolatedNormal };
	const half3x3 xTangentToWorldMatrix2
		= { xInterpolatedTangent2, xInterpolatedBinormal2, xInterpolatedNormal };


	const half3 xViewDir = normalize( xIn.m_xCameraDirDistance.xyz );
	const half3 xTangentSpaceViewDir = normalize( xIn.m_xTangentSpaceCameraDir );

	const float fDualTextureLerp = xIn.m_xTangentSpaceCameraDir.w;

	half4 xUVPair = xIn.m_xUVPair;
	ApplyParallaxMapping( xTangentSpaceViewDir, fDualTextureLerp, xUVPair );

	const float4 xWorldSpaceNormal = NormalLookup( xUVPair, fDualTextureLerp, xTangentToWorldMatrix, xTangentToWorldMatrix2 );

	half3 xAlbedo;
	const half3 xAlbedoTimes2 = AlbedoLookup( xUVPair, fDualTextureLerp, xWorldSpaceNormal,
		xViewDir, xAlbedo );

	half4 xColour = MaterialResponseSH( xUVPair, fDualTextureLerp, xWorldSpaceNormal, xViewDir );

	xColour.rgb += SHLookup( xWorldSpaceNormal, xIn.m_xRedSH, xIn.m_xGreenSH, xIn.m_xBlueSH )
				 * xAlbedoTimes2;

	return ApplyFog(xColour, xIn.m_xCameraDirDistance.w);
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

// Simple prelighting, no material response curves
//technique Deferred_SH
//{
//	pass
//	{
//		AlphaTestEnable = FALSE;
//		CullMode = CW;
//		ZEnable = TRUE;
//		ZWriteEnable = TRUE;
//
//		VertexShader = compile vs_3_0 VShader();
//		PixelShader = compile ps_3_0 PShader_Deferred( false );
//	}
//}

// Perform material response using the SH
technique Deferred_MR
{
	pass
	{
		AlphaTestEnable = FALSE;
		CullMode = CW;
		ZEnable = TRUE;
		ZWriteEnable = TRUE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Deferred( true );
	}
}

// Forward renderer, for transparent strips
technique Forward
{
	pass
	{
		AlphaTestEnable = FALSE;
		CullMode = CW;
		ZEnable = TRUE;
		ZWriteEnable = FALSE;

		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Forward();
	}
}