/******************************************************************************/
/* Parameters                                                                 */
/******************************************************************************/

#define DEFINE_SAMPLER( Name )                   \
	texture g_x##Name##Texture;                  \
	sampler2D g_x##Name##Sampler = sampler_state \
	{                                            \
		Texture = <g_x##Name##Texture>;          \
		AddressU = CLAMP;                        \
		AddressV = CLAMP;                        \
	};

// Copy of the scene to be downsized
DEFINE_SAMPLER( Downsize );

// Single texel holding the current adapted luminance
DEFINE_SAMPLER( CurrentLuminance );

// Single texel holding the scene's luminance, our target value over time
DEFINE_SAMPLER( SceneLuminance );

// Current HDR scene to apply tone mapping on
DEFINE_SAMPLER( InputScene );

#undef DEFINE_SAMPLER

float2 g_xExtractLumUVScale;

float g_fAdaptationFactor;  // Adaption rate * time elapsed this frame
float g_fMaxLuminance;      // Maximum luminance to adapt to
float g_fMinLuminance;      // Minimum luminance to adapt to
float g_fKeyValue;          // Exposure setting for the scene
float g_fWhitePointSquared; // World luminances beyond this point are whited out
float g_fXboxGammaFudge = 1.0f / 2.2f;
// (white pt / (white pt - black pt), -black pt * white pt / (white pt - black pt)
float2 g_xScaleParams;

// Allow the user to specify a fixed luminance
bool g_bDynamicExposure = true;
float g_fFixedLuminance;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

// Vertex structure used in second pass
struct PosUVVert
{
	float4 m_xPos : POSITION;
	float4 m_xUV  : TEXCOORD;
};

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

// Luminance = 0.27r + 0.67g + 0.06b
static float3 xLUMINANCE_CONSTANTS = float3( 0.27f, 0.67f, 0.06f );

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

float4 VShader( const float4 xIn : POSITION ) : POSITION
{
	return xIn;
}

float4 Average16_Filtering( const float2 xUV )
{
	// Use bilinear filtering on each 2x2 sub-block, then average the results
	float4 xSample1, xSample2, xSample3, xSample4;
	asm
	{
		tfetch2D xSample1, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=point, AnisoFilter=disabled, OffsetX=+1.0, OffsetY=+1.0
		tfetch2D xSample2, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=point, AnisoFilter=disabled, OffsetX=+2.0, OffsetY=+1.0
		tfetch2D xSample3, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=point, AnisoFilter=disabled, OffsetX=+2.0, OffsetY=+3.0
		tfetch2D xSample4, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=linear, MagFilter=linear, MipFilter=point, AnisoFilter=disabled, OffsetX=+1.0, OffsetY=+3.0

	};

	return (xSample1 + xSample2 + xSample3 + xSample4) / 4.0f;
}

float4 Average16_Point( const float2 xUV )
{
	// Can't use filtering on float textures. Manually average each block
	float4 xSample01, xSample02, xSample03, xSample04, xSample05, xSample06, xSample07, xSample08,
		   xSample09, xSample10, xSample11, xSample12, xSample13, xSample14, xSample15, xSample16;

	asm
	{
		tfetch2D xSample01, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=0.5, OffsetY=0.5
		tfetch2D xSample02, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=1.5, OffsetY=0.5
		tfetch2D xSample03, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=2.5, OffsetY=0.5
		tfetch2D xSample04, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=3.5, OffsetY=0.5

		tfetch2D xSample05, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=0.5, OffsetY=1.5
		tfetch2D xSample06, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=1.5, OffsetY=1.5
		tfetch2D xSample07, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=2.5, OffsetY=1.5
		tfetch2D xSample08, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=3.5, OffsetY=1.5

		tfetch2D xSample09, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=0.5, OffsetY=2.5
		tfetch2D xSample10, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=1.5, OffsetY=2.5
		tfetch2D xSample11, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=2.5, OffsetY=2.5
		tfetch2D xSample12, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=3.5, OffsetY=2.5

		tfetch2D xSample13, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=0.5, OffsetY=3.5
		tfetch2D xSample14, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=1.5, OffsetY=3.5
		tfetch2D xSample15, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=2.5, OffsetY=3.5
		tfetch2D xSample16, xUV, g_xDownsizeSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=3.5, OffsetY=3.5
	};

	return (xSample01 + xSample02 + xSample03 + xSample04 + xSample05 + xSample06 + xSample07 + xSample08
		  + xSample09 + xSample10 + xSample11 + xSample12 + xSample13 + xSample14 + xSample15 + xSample16) / 16.0f;
}

float4 PShader_Downsample( float2 xScreenPos : VPOS, uniform const bool bUseFiltering ) : COLOR
{
	// Convert coords to the 4x4-times-larger source texture
	xScreenPos *= 4.0f;

	return bUseFiltering
		 ? Average16_Filtering( xScreenPos )
		 : Average16_Point( xScreenPos );
}

PosUVVert VShader_CalcLogLuminance( const PosUVVert xIn )
{
	return xIn;
}

float4 PShader_CalcLogLuminance( const float2 xUV : TEXCOORD ) : COLOR
{
	// We should be squishing the scene into a smaller RT at this point (for 720p, 320x180 -> 64x64)

	// Wish we could use arrays in asm blocks...
	float4 xSample1, xSample2, xSample3, xSample4, xSample5, xSample6, xSample7, xSample8, xSample9;
	asm
	{
		tfetch2D xSample1, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=-1.0, OffsetY=-1.0
		tfetch2D xSample2, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=+0.0, OffsetY=-1.0
		tfetch2D xSample3, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=+1.0, OffsetY=-1.0
		tfetch2D xSample4, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=-1.0, OffsetY=+0.0
		tfetch2D xSample5, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=+0.0, OffsetY=+0.0
		tfetch2D xSample6, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=+1.0, OffsetY=+0.0
		tfetch2D xSample7, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=-1.0, OffsetY=+1.0
		tfetch2D xSample8, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=+0.0, OffsetY=+1.0
		tfetch2D xSample9, xUV, g_xDownsizeSampler, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled, OffsetX=+1.0, OffsetY=+1.0
	};

	// Find the luminance of each sample
	const float afLuminance[9] =
	{
		dot( xSample1.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample2.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample3.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample4.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample5.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample6.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample7.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample8.rgb, xLUMINANCE_CONSTANTS ),
		dot( xSample9.rgb, xLUMINANCE_CONSTANTS ),
	};

	// Add a tiny amount to avoid log(0) on pure black pixels
	const float fDELTA = 0.0001f;

	// Find the average natural log
	float fAvg = log( afLuminance[0] + fDELTA );
	for( int i = 1; i < 9; ++i )
	{
		fAvg += log( afLuminance[i] + fDELTA );
	}

	return fAvg / 9.0f;
}

float4 PShader_CalcFinalLuminance() : COLOR
{
	// We should now be rendering into a 1x1 RT, from a 4x4 source. Take one final average and then
	// undo the log

	const float2 xUV = float2( 0.0f, 0.0f );

	// NB: exp() HLSL intrinsic operates on every component, but assembly instruction is scalar.
	// Leaving off the .r here dramatically increases the register usage and number of ALU ops
	const float fAvg = Average16_Point( xUV ).r;

	return exp( fAvg );
}

float4 PShader_AdaptLuminance() : COLOR
{
	const float2 xUV = float2( 0.5f, 0.5f );
	float4 xCurrentLuminance, xSceneLuminance;
	asm
	{
		tfetch2D xCurrentLuminance, xUV, g_xCurrentLuminanceSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
		tfetch2D xSceneLuminance, xUV, g_xSceneLuminanceSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point, AnisoFilter=disabled
	};

	const float fAdaptedLuminance = lerp( xCurrentLuminance.r, xSceneLuminance.r, g_fAdaptationFactor );
	return clamp( fAdaptedLuminance, g_fMinLuminance, g_fMaxLuminance );
}

float4 PShader_ApplyToneMapping( const float2 xScreenPos : VPOS ) : COLOR
{
	float4 xInputSample;
	asm
	{
		tfetch2D xInputSample, xScreenPos, g_xInputSceneSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point
	};

	// Scale to account for "black point"
	xInputSample.rgb = max( xInputSample.rgb * g_xScaleParams.x + g_xScaleParams.y, 0.0f );

	float4 xAdaptedLuminance;

	if( g_bDynamicExposure )
	{
		const float2 xUV = float2( 0.5f, 0.5f );
		asm
		{
			tfetch2D xAdaptedLuminance, xUV, g_xCurrentLuminanceSampler, UnnormalizedTextureCoords=true, MinFilter=point, MagFilter=point, MipFilter=point
		};
	}
	else
	{
		xAdaptedLuminance.r = g_fFixedLuminance;
	}

	const float fHDRLuminance = dot( xInputSample.rgb, xLUMINANCE_CONSTANTS );

	// Apply a slight delta to prevent divide-by-zero
	const float fDELTA = 0.0001f;
	const float fScaledLuminance = g_fKeyValue / (xAdaptedLuminance.r + fDELTA) * fHDRLuminance;

	const float fNumerator = fScaledLuminance * (1.0f + fScaledLuminance / g_fWhitePointSquared);
	const float fDenominator = 1.0f + fScaledLuminance;

	const float fLDRLuminance = fNumerator / fDenominator;

	// Rescale the colours by the ratio of new to old luminance. Slight delta to avoid divide by 0
	xInputSample.rgb = xInputSample.rgb * fLDRLuminance / max( fHDRLuminance, fDELTA );

	// Simulate a power 2.2 curve by taking a sqrt, technically this is pow(x,1/2.0) rather than pow(x,1/2.2)
	// but its a lot quicker to calculate and should be "close enough"
	// AS 30/10/08 - Nope changed back to 2.2 and just going to have to eat the slight performance cost
	// 2.0 was just too dark on some of the dark points on AVP. 2.4 actually looks
	// closer to PC on the dell monitors but will leave it at 2.2 for now.
	return saturate( pow(xInputSample, g_fXboxGammaFudge) ); 
}

/******************************************************************************/
/* Techniques                                                                 */
/******************************************************************************/

technique
{
	pass
	{
		AlphaTestEnable = FALSE;
		CullMode = NONE;
		ZEnable = FALSE;
		ZWriteEnable = FALSE;

		// First pass downsamples from 1280x720 to 320x180. We can use filtering
		// because we source from a 16F_EXPAND texture
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Downsample( true );
	}

	pass
	{
		// Second pass calculates log luminance averages, writing into a 64x64
		// R32F target
		VertexShader = compile vs_3_0 VShader_CalcLogLuminance();
		PixelShader = compile ps_3_0 PShader_CalcLogLuminance();
	}

	pass
	{
		// Then need to downsample that R32F source down to 4x4. Can't use
		// filtering here because of the texture format. This pass is executed
		// twice (64x64 -> 16x16 -> 4x4)
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_Downsample( false );
	}

	pass
	{
		// Final pass downsamples from 4x4 to 1x1 and undoes the log
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_CalcFinalLuminance();
	}

	pass
	{
		// This pass adapts the eye luminance towards the target
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_AdaptLuminance();
	}

	pass
	{
		// Finally we apply tone mapping using the current light adaptation
		VertexShader = compile vs_3_0 VShader();
		PixelShader = compile ps_3_0 PShader_ApplyToneMapping();
	}
}