//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Shadow_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Shadows/Asura_Shadow_System.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Memory/Asura_Mem.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_DeferredShading.h"

// Asura Xbox 360
#include "Debugging/Asura_Platform_GPUProfile.h"
#include "Rendering/Asura_Xbox360_DeferredShading.h"
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_EnvironmentRenderer.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_IndexBuffer.h"
#include "Rendering/Asura_Xbox360_QuadOverlayRenderer.h"
#include "Rendering/Asura_Xbox360_SkinRenderer.h"
#include "Rendering/Asura_Xbox360_VertexBuffer.h"

// Misc
#include <xgraphics.h>
#include <xtl.h>

// Shader Include Files
#include "ApplyShadows.h"
#ifdef _DEBUG
#include "ShadowBufferOverlayRender.h"
#endif

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

#if !defined(ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS)
#define ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS	(ASURA_NUM_SHADOW_TEXTURE_SIZES * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE) 
#endif	//!defined(ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS)

#if !defined(ASURA_MAX_NUMBER_OF_SHADOWS)
#define ASURA_MAX_NUMBER_OF_SHADOWS		(ASURA_MAX_NUMBER_OF_NON_CUBEMAP_SHADOWS + ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS)
#endif	//!defined(ASURA_MAX_NUMBER_OF_SHADOWS)

#define ASURA_CUBEMAP_SHADOW_SIZE 512

const Direct3D_VertexElement axVERTEX_DECLARATION[] =
{
	{ 0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
	D3DDECL_END()
};

const u_int uVERTEX_SIZE = 12;

const D3DVECTOR axFRUSTUM_VERTICES[] =
{
	{ -1.0f,  1.0f, 0.0f },
	{  1.0f,  1.0f, 0.0f },
	{  1.0f, -1.0f, 0.0f },
	{ -1.0f, -1.0f, 0.0f },
	{ -1.0f,  1.0f, 1.0f },
	{  1.0f,  1.0f, 1.0f },
	{  1.0f, -1.0f, 1.0f },
	{ -1.0f, -1.0f, 1.0f },
};





const u_short ausFRUSTUM_INDICES[] =
{
	0, 1, 2, 3,		// Near
	4, 5, 1, 0,		// Bottom
	5, 6, 2, 1,		// Right
	6, 7, 3, 2,		// Top
	7, 4, 0, 3,		// Left
	5, 4, 7, 6,		// Far
};

/******************************************************************************/
/* Statics                                                                    */
/******************************************************************************/

static u_int				g_uCreateShadowTextureSize;
static Direct3D_Viewport	g_xCreateShadowViewport;

static Asura_Vector_3		g_xLastLightPosition[2];
static float				g_fLastLightRange[2];

/******************************************************************************/
/* Classes                                                                    */
/******************************************************************************/

class Asura_Platform_ShadowBuffer
{
public:
	Asura_Platform_ShadowBuffer(u_int uTextureSize);
	~Asura_Platform_ShadowBuffer();

	void Bind() const;

	u_int				GetTextureSize		(void) const	{ return(m_uTextureSize); }
	Direct3D_Texture*	GetTexture			(void) const    { return(m_pShadowTexture);}

private:
	u_int m_uTextureSize;
	Direct3D_Texture* m_pShadowTexture;
	Direct3D_Surface* m_pShadowTextureSurface;
	Direct3D_Surface* m_pDepthSurface;
};

Asura_Platform_ShadowBuffer::Asura_Platform_ShadowBuffer(u_int uTextureSize) 
: m_uTextureSize(uTextureSize)
{
	Asura_Xbox360_Graphics::GetDevice()->CreateTexture(uTextureSize,uTextureSize, 1, D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &m_pShadowTexture, NULL);
	Asura_Assert(m_pShadowTexture, "Unable to create shadow texture");
	
	m_pShadowTexture->GetSurfaceLevel(0, &m_pShadowTextureSurface);
	Asura_Assert(m_pShadowTextureSurface, "Unable to get surface of shadow texture");

}

Asura_Platform_ShadowBuffer::~Asura_Platform_ShadowBuffer()
{
	SAFE_RELEASE(m_pShadowTextureSurface);
	SAFE_RELEASE(m_pShadowTexture);
}

void Asura_Platform_ShadowBuffer::Bind() const
{
}

static Direct3D_CubeTexture*			s_pxEnvShadowCubeTexture[2];
static Direct3D_Surface*				s_pxShadowDepthStencil[ASURA_NUM_SHADOW_TEXTURE_SIZES];
static Asura_Platform_ShadowBuffer*		s_axShadowTextures[ASURA_NUM_SHADOW_TEXTURE_SIZES*ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE];
static u_int s_uShadowTexturesAllocated;
static Direct3D_Surface* g_pxRT = NULL;
static Direct3D_Surface* g_pxDepthRT = NULL;

enum ShadowTechnique
{
	Technique_NonCubic_SingleTap_NoPCF_NoNoise,
	Technique_NonCubic_MultiTap_NoPCF_NoNoise,
	Technique_Cubic_SingleTap_NoPCF_NoNoise,
	Technique_Cubic_MultiTap_NoPCF_NoNoise,
	Technique_NonCubic_SingleTap_NoPCF_Noise,
	Technique_NonCubic_MultiTap_NoPCF_Noise,
	Technique_Cubic_SingleTap_NoPCF_Noise,
	Technique_Cubic_MultiTap_NoPCF_Noise,
	Technique_NonCubic_SingleTap_NoPCF_ScreenNoise,
	Technique_NonCubic_MultiTap_NoPCF_ScreenNoise,
	Technique_Cubic_SingleTap_NoPCF_ScreenNoise,
	Technique_Cubic_MultiTap_NoPCF_ScreenNoise,
	Technique_NonCubic_SingleTap_PCF_NoNoise,
	Technique_NonCubic_MultiTap_PCF_NoNoise,
	Technique_NonCubic_SingleTap_PCF_Noise,
	Technique_NonCubic_MultiTap_PCF_Noise,
	Technique_NonCubic_SingleTap_PCF_ScreenNoise,
	Technique_NonCubic_MultiTap_PCF_ScreenNoise,
	Technique_Standardised_NonCubic, // no static branches!
};

static ShadowTechnique GetShadowTechnique(bool bFilter, bool bCubic, bool bUsePCF, bool bUseNoise, bool bUseScreenBasedNoise)
{
	static ShadowTechnique xTechniques[2*2*2*2*2] = 
	{
		//	Mode											Filter		Cubic		PCF			Noise		ScreenSpaceNoise
		Technique_NonCubic_SingleTap_NoPCF_NoNoise,		//	false		false		false		false		false
		Technique_NonCubic_MultiTap_NoPCF_NoNoise,		//	true		false		false		false		false
		Technique_Cubic_SingleTap_NoPCF_NoNoise,		//	false		true		false		false		false
		Technique_Cubic_MultiTap_NoPCF_NoNoise,			//	true		true		false		false		false
		Technique_NonCubic_SingleTap_PCF_NoNoise,		//	false		false		true		false		false
		Technique_NonCubic_MultiTap_PCF_NoNoise,		//	true		false		true		false		false
		Technique_Cubic_SingleTap_NoPCF_NoNoise,		//	false		true		true		false		false
		Technique_Cubic_MultiTap_NoPCF_NoNoise,			//	true		true		true		false		false
		Technique_NonCubic_SingleTap_NoPCF_Noise,		//	false		false		false		true		false
		Technique_NonCubic_MultiTap_NoPCF_Noise,		//	true		false		false		true		false
		Technique_Cubic_SingleTap_NoPCF_Noise,			//	false		true		false		true		false
		Technique_Cubic_MultiTap_NoPCF_Noise,			//	true		true		false		true		false
		Technique_NonCubic_SingleTap_PCF_Noise,			//	false		false		true		true		false
		Technique_NonCubic_MultiTap_PCF_Noise,			//	true		false		true		true		false
		Technique_Cubic_SingleTap_NoPCF_Noise,			//	false		true		true		true		false
		Technique_Cubic_MultiTap_NoPCF_Noise,			//	true		true		true		true		false

		Technique_NonCubic_SingleTap_NoPCF_NoNoise,		//	false		false		false		false		true
		Technique_NonCubic_MultiTap_NoPCF_NoNoise,		//	true		false		false		false		true
		Technique_Cubic_SingleTap_NoPCF_NoNoise,		//	false		true		false		false		true
		Technique_Cubic_MultiTap_NoPCF_NoNoise,			//	true		true		false		false		true
		Technique_NonCubic_SingleTap_PCF_NoNoise,		//	false		false		true		false		true
		Technique_NonCubic_MultiTap_PCF_NoNoise,		//	true		false		true		false		true
		Technique_Cubic_SingleTap_NoPCF_NoNoise,		//	false		true		true		false		true
		Technique_Cubic_MultiTap_NoPCF_NoNoise,			//	true		true		true		false		true
		Technique_NonCubic_SingleTap_NoPCF_ScreenNoise,	//	false		false		false		true		true
		Technique_NonCubic_MultiTap_NoPCF_ScreenNoise,	//	true		false		false		true		true
		Technique_Cubic_SingleTap_NoPCF_ScreenNoise,	//	false		true		false		true		true
		Technique_Cubic_MultiTap_NoPCF_ScreenNoise,		//	true		true		false		true		true
		Technique_NonCubic_SingleTap_PCF_ScreenNoise,	//	false		false		true		true		true
		Technique_NonCubic_MultiTap_PCF_ScreenNoise,	//	true		false		true		true		true
		Technique_Cubic_SingleTap_NoPCF_ScreenNoise,	//	false		true		true		true		true
		Technique_Cubic_MultiTap_NoPCF_ScreenNoise,		//	true		true		true		true		true
	};

	return xTechniques[bFilter*1 + bCubic*2 + bUsePCF*4 + bUseNoise*8 + bUseScreenBasedNoise*16];

}

DVSVAR float g_fXShadowOffset;
DVSVAR float g_fYShadowOffset;
DVSVAR bool g_bDisableFiltering = false;
DVSVAR bool g_bEnablePCF = true;
DVSVAR u_int g_uNoiseMode = 1; // 0 = Force Noise Off, 1 = Use Shadow Flag, 2 = Force Noise On
DVSVAR bool g_bEnableScreenBasedNoise = false;
DVSVAR bool dvs_bEnableStandardisedShadows = false;
DVSVAR bool g_bEnableStencil = true;
DVSVAR bool g_bEnableCubeMap = true;

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Shadow
//
// Represents a single Xbox 360 shadow
//
////////////////////////////////////////////////////////////////////////////////

class Asura_Xbox360_Shadow : public Asura_Shadow
{
public:

	// Constructor / destructor
	Asura_Xbox360_Shadow();
	virtual ~Asura_Xbox360_Shadow();

	// Virtual overrides

	virtual	void	PrepareForRendering	(void);
	virtual	void	FinishedRendering	(void);

	Asura_Platform_ShadowBuffer* GetShadowBuffer	(void) const	{ return m_pxShadowBuffer; }

	virtual void SetInstance (const Asura_Shadow_Instance* pxInstance)
	{ 
		m_pxInstance = pxInstance; 
		if (!pxInstance->IsCubeMapped() && s_uShadowTexturesAllocated < ASURA_NUM_SHADOW_TEXTURE_SIZES*ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE)
		{
			m_pxShadowBuffer = s_axShadowTextures[s_uShadowTexturesAllocated++];
		} else {
			m_pxShadowBuffer = NULL;
		}
	}

protected:

	Asura_Platform_ShadowBuffer* m_pxShadowBuffer;
};

/******************************************************************************/
/* Global Variables                                                           */
/******************************************************************************/


// Effect and parameters

static Asura_Xbox360_Effect* g_pxApplyShadowsEffect = NULL;

static Asura_Xbox360_Effect_System::Handle g_xLightClipSpaceToCameraClipSpaceHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xCameraViewSpaceToLightClipSpaceHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xCameraViewSpaceToLightViewSpaceHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xCameraViewSpaceToWorldSpaceHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xUVXFormHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xSelfShadowBiasHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xLightPosHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xTwoOverScreenWidthHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xTwoOverScreenHeightHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xScreenOffsetHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xCameraTanHalfVertFOVHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xCameraTanHalfHorizFOVHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowTexSizeHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowFilterTextureScaleHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowBufferHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowBufferCubeHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xNormalBufferHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xDepthMtlBufferHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowDirectionHandle;
static Asura_Xbox360_Effect_System::Handle g_xShadowStrengthHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xKernelRotationTextureHandle = 0;			
static Asura_Xbox360_Effect_System::Handle g_xShadowFilterSizeHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xShadowMaskHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xCubicMapShadowHandle;
static Asura_Xbox360_Effect_System::Handle g_xShadowFrustumNearPlaneDistanceHandle = 0;
static Asura_Xbox360_Effect_System::Handle g_xOneOverShadowFrustumPlaneSeparationHandle;

static D3DXMATRIX			s_xUVXForm;
static D3DXVECTOR4			s_xSelfShadowBias;


static D3DXVECTOR4     s_xSelfShadowBiasValues[3] = 
{
	// Slope, Bias, Dist
	// non environment
	D3DXVECTOR4(0.001f, .000f, 0.03, 0),
	// environment
	D3DXVECTOR4(0.001f, .001f, 0.02f, 0),

	// cubic environment
	D3DXVECTOR4(0.013f, .000f, 0.002f, 0),
};




// Shadow frustum geometry

static Asura_Xbox360_VertexBuffer   g_xShadowFrustumVB;
static Asura_Xbox360_IndexBuffer    g_xShadowFrustumIB;  

static Direct3D_VertexDeclaration*  g_pxShadowFrustumVD = NULL;
static Direct3D_VertexDeclaration* g_pxShadowFrustumVDSphere = NULL;

static Direct3D_Texture*			g_pxFilterKernelRotationTexture = NULL;

static bool CreateEnvShadowTextures( void* = NULL )
{
	// Create cube texture
	{
		/*const HRESULT xResult =*/ Asura_Xbox360_Graphics::GetDevice()->CreateCubeTexture( ASURA_CUBEMAP_SHADOW_SIZE, 1,
			D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &s_pxEnvShadowCubeTexture[0], NULL );
		Asura_Xbox360_Graphics::GetDevice()->CreateCubeTexture( ASURA_CUBEMAP_SHADOW_SIZE, 1,
			D3DUSAGE_RENDERTARGET, D3DFMT_R32F, D3DPOOL_DEFAULT, &s_pxEnvShadowCubeTexture[1], NULL );
		//ASURA_D3D_OK( xResult );
	}
	for (u_int i = 0, uTexSize = ASURA_MAX_SHADOW_TEXTURE_SIZE; i < ASURA_NUM_SHADOW_TEXTURE_SIZES; ++i, (uTexSize>>=1))
	{
		for (u_int j = 0; j < ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE; ++j)
		{
			s_axShadowTextures[j + i * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE] = new Asura_Platform_ShadowBuffer(uTexSize);
		}
	}
	return true;
}

static bool DestroyEnvShadowTextures( void* = NULL )
{
	SAFE_RELEASE( s_pxEnvShadowCubeTexture[0] );
	s_pxEnvShadowCubeTexture[0] = NULL;
	SAFE_RELEASE( s_pxEnvShadowCubeTexture[1] );
	s_pxEnvShadowCubeTexture[1] = NULL;

	for (u_int i = 0, uTexSize = ASURA_MAX_SHADOW_TEXTURE_SIZE; i < ASURA_NUM_SHADOW_TEXTURE_SIZES; ++i, (uTexSize>>=1))
	{
		for (u_int j = 0; j < ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE; ++j)
		{
			delete s_axShadowTextures[j + i * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE];
			s_axShadowTextures[j + i * ASURA_NUM_SHADOWS_PER_TEXTURE_SIZE] = NULL;
		}
	}

	return true;
}





/******************************************************************************/
/* Asura_Shadow_System Static Class Functions                                 */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Initialise()
//
// Initialises the shadow system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_Initialise()
{
	Asura_Shadow**	ppxShadow		= s_apxShadows;

	for(u_int u = 0; u < ASURA_MAX_NUMBER_OF_SHADOWS; ++u, ++ppxShadow)
	{
		*ppxShadow	= new Asura_Xbox360_Shadow();
	}

	// Create the texture used for random rotation of the shadow filter kernel. Contains sin/cos of random angles	
	HRESULT xResult = Asura_Xbox360_Graphics::GetDevice()->CreateTexture( 64, 64, 0, 0, D3DFMT_A8R8G8B8, D3DPOOL_MANAGED, &g_pxFilterKernelRotationTexture, NULL );	
	Asura_Assert( NULL != g_pxFilterKernelRotationTexture, "Failed to create shadow filter kernel rotation texture." );

	D3DLOCKED_RECT xLockedRect;
	xResult = g_pxFilterKernelRotationTexture->LockRect( 0, &xLockedRect, NULL, 0 );
	if( FAILED( xResult ) )
	{
		Asura_Assert( false, "Failed to lock shadow filter kernel rotation texture." );	
	}
	else
	{
		u_char *pxData = static_cast<u_char*>(xLockedRect.pBits);

		for(int i=0;i<64;i++)
		{
			for (int j=0;j<64;j++)
			{
				float fAngle = Asura_Random::GetFloat(0,Asura_Maths::Pi*2.0f);
				float fSin,fCos;
				Asura_Maths::SinCos(fAngle, fSin, fCos);
				fSin = (fSin * 127.0f) + 127.0f;
				fCos = (fCos * 127.0f) + 127.0f;
				pxData[(j<<2)+3] = (u_char)fSin;
				pxData[(j<<2)+2] = (u_char)fCos;
				pxData[(j<<2)+1] = 0;
				pxData[(j<<2)+0] = 0;
			}
			pxData += xLockedRect.Pitch;
		}

		g_pxFilterKernelRotationTexture->UnlockRect(0);
	}
	D3DXFilterTexture(g_pxFilterKernelRotationTexture, NULL, 0, D3DX_FILTER_TRIANGLE);


	// These surfaces can overlap others, since we won't be using them at the same time

	const D3DSURFACE_PARAMETERS& xShadowRTParams =
		Asura_Xbox360_Graphics::GetSurfaceParams( Asura_Xbox360_Graphics::SP_SHADOW_RT );

	Asura_Xbox360_Graphics::GetDevice()->CreateRenderTarget( ASURA_MAX_SHADOW_TEXTURE_SIZE,
		ASURA_MAX_SHADOW_TEXTURE_SIZE, D3DFMT_R32F, D3DMULTISAMPLE_NONE, 0, FALSE, &g_pxRT,
		&xShadowRTParams );

	const D3DSURFACE_PARAMETERS& xShadowDepthParams =
		Asura_Xbox360_Graphics::GetSurfaceParams( Asura_Xbox360_Graphics::SP_SHADOW_DEPTH );

	Asura_Xbox360_Graphics::GetDevice()->CreateDepthStencilSurface( ASURA_MAX_SHADOW_TEXTURE_SIZE,
		ASURA_MAX_SHADOW_TEXTURE_SIZE, D3DFMT_D24S8, D3DMULTISAMPLE_NONE, 0, FALSE, &g_pxDepthRT,
		&xShadowDepthParams );

	Asura_Assert( g_pxRT && g_pxDepthRT, "Unable to create shadow depth surface." );

	// Initialise effect and parameter handles
	g_pxApplyShadowsEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auApplyShadows);

	g_xLightClipSpaceToCameraClipSpaceHandle	= g_pxApplyShadowsEffect->GetParameterHandle( "g_xLightClipSpaceToCameraClipSpace" );
	g_xCameraViewSpaceToLightClipSpaceHandle	= g_pxApplyShadowsEffect->GetParameterHandle( "g_xCameraViewSpaceToLightClipSpace" );
	g_xCameraViewSpaceToLightViewSpaceHandle	= g_pxApplyShadowsEffect->GetParameterHandle( "g_xCameraViewSpaceToLightViewSpace" );
	g_xCameraViewSpaceToWorldSpaceHandle		= g_pxApplyShadowsEffect->GetParameterHandle( "g_xCameraViewSpaceToWorldSpace" );
	g_xLightPosHandle							= g_pxApplyShadowsEffect->GetParameterHandle( "g_xLightPos" );
	g_xCameraTanHalfHorizFOVHandle				= g_pxApplyShadowsEffect->GetParameterHandle( "g_fCameraTanHalfHorizFOV" );
	g_xCameraTanHalfVertFOVHandle				= g_pxApplyShadowsEffect->GetParameterHandle( "g_fCameraTanHalfVertFOV" );
	g_xTwoOverScreenWidthHandle					= g_pxApplyShadowsEffect->GetParameterHandle( "g_fTwoOverScreenWidth" );
	g_xTwoOverScreenHeightHandle				= g_pxApplyShadowsEffect->GetParameterHandle( "g_fTwoOverScreenHeight" );
	g_xScreenOffsetHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_fScreenOffset" );
	g_xShadowTexSizeHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_fShadowTexSize");	
	g_xShadowFilterTextureScaleHandle			= g_pxApplyShadowsEffect->GetParameterHandle( "g_fShadowFilterTextureScale");
	g_xShadowBufferHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_xShadowBuffer" );
	g_xShadowBufferCubeHandle					= g_pxApplyShadowsEffect->GetParameterHandle( "g_xShadowCubeBuffer" );
	g_xNormalBufferHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_xNormalBuffer" );
	g_xDepthMtlBufferHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_xDepthMtlBuffer" );
	g_xShadowDirectionHandle					= g_pxApplyShadowsEffect->GetParameterHandle( "g_xShadowDirection" );
	g_xShadowStrengthHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_fShadowStrength" );			
	g_xKernelRotationTextureHandle				= g_pxApplyShadowsEffect->GetParameterHandle( "g_xKernelRotationSampler" );			
	g_xShadowFilterSizeHandle					= g_pxApplyShadowsEffect->GetParameterHandle( "g_fShadowFilterSize" );			
	g_xShadowMaskHandle							= g_pxApplyShadowsEffect->GetParameterHandle( "g_bCreatingShadowMask" );			
	g_xCubicMapShadowHandle						= g_pxApplyShadowsEffect->GetParameterHandle( "g_bCubicMapShadow" );
	g_xShadowFrustumNearPlaneDistanceHandle 	= g_pxApplyShadowsEffect->GetParameterHandle( "g_fShadowFrustumNearPlaneDistance" );
	g_xOneOverShadowFrustumPlaneSeparationHandle= g_pxApplyShadowsEffect->GetParameterHandle( "g_fOneOverShadowFrustumPlaneSeparation" );
	g_xUVXFormHandle							= g_pxApplyShadowsEffect->GetParameterHandle( "g_xUVXForm" );
	g_xSelfShadowBiasHandle	                    = g_pxApplyShadowsEffect->GetParameterHandle( "g_xSelfShadowBias" );

	Asura_Mem::Set(s_xUVXForm, 0, sizeof(s_xUVXForm));
	s_xUVXForm._11 = 0.5f;
	s_xUVXForm._22 = 0.5f;
	s_xUVXForm._33 = 1.0f;
	s_xUVXForm._44 = 1.0f;

	// Set up shadow frustum geometry

	g_xShadowFrustumVB.CreateStatic( uVERTEX_SIZE, 8 );

	void* pVertices = g_xShadowFrustumVB.Lock();
	Asura_Assert( pVertices, "Failed to lock shadow frustum vertex buffer." );
	XMemCpyStreaming_WriteCombined( pVertices, axFRUSTUM_VERTICES, sizeof( axFRUSTUM_VERTICES ) );
	g_xShadowFrustumVB.Unlock();


	g_xShadowFrustumIB.CreateStatic( sizeof( u_short ) * 24 );

	void* pIndices = g_xShadowFrustumIB.Lock();
	Asura_Assert( pIndices, "Failed to lock shadow frustum index buffer." );
	XMemCpyStreaming_WriteCombined( pIndices, ausFRUSTUM_INDICES, sizeof( ausFRUSTUM_INDICES ) );
	g_xShadowFrustumIB.Unlock();

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( axVERTEX_DECLARATION, &g_pxShadowFrustumVD );
	Asura_Assert( g_pxShadowFrustumVD, "Failed to create shadow frustum vertex declaration." );

	CreateEnvShadowTextures();

#if defined(ASURA_DEBUG_VARIABLE_SYSTEM) 
	if(!Asura_CommandConsole::VarExists(L"Render.Shadows.Bias.Skin_Slope"))
	{
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Slope", &(s_xSelfShadowBiasValues[0].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Bias", &(s_xSelfShadowBiasValues[0].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Skin_Dist", &(s_xSelfShadowBiasValues[0].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Slope", &(s_xSelfShadowBiasValues[1].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Bias", &(s_xSelfShadowBiasValues[1].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Env_Dist", &(s_xSelfShadowBiasValues[1].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Slope", &(s_xSelfShadowBiasValues[2].x), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Bias", &(s_xSelfShadowBiasValues[2].y), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Bias.Cube_Dist", &(s_xSelfShadowBiasValues[2].z), 0.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Offset.X", &(g_fXShadowOffset), -1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.Offset.Y", &(g_fYShadowOffset), -1.0f, 1.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.DisableFiltering", &(g_bDisableFiltering), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.EnablePCF", &(g_bEnablePCF), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.NoiseMode", &(g_uNoiseMode), 0, 2, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.EnableScreenBasedNoise", &(g_bEnableScreenBasedNoise), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.EnableStencil", &(g_bEnableStencil), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar( L"Render.Shadows.EnableCubeMap", &(g_bEnableCubeMap), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );


		Asura_CommandConsole::AddVar( L"Render.Shadows.EnableStandardisedShadows", &(dvs_bEnableStandardisedShadows), Asura_CommandConsole::REMOVE_AT_SHUTDOWN );



	}
#endif

}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Reset()
//
// Resets the shadow system on Xbox 360. Called between levels
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_Reset()
{
	// reset the last light info just in case we have matching values between levels
	for (u_int u=0; u<2; ++u)
	{
		g_xLastLightPosition[u] = Asura_Maths::ZeroVector3;
		g_fLastLightRange[u] = 0.0f;
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_Shutdown()
//
// Shuts down the shadow system on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_Shutdown()
{
	if (g_pxFilterKernelRotationTexture)
	{
		g_pxFilterKernelRotationTexture->Release();
		g_pxFilterKernelRotationTexture = NULL;
	}

	DestroyEnvShadowTextures();

	Asura_Xbox360_Effect_System::Destroy( g_pxApplyShadowsEffect );
	g_pxApplyShadowsEffect = NULL;

	if( g_pxDepthRT )
	{
		g_pxDepthRT->Release();
		g_pxDepthRT = NULL;
	}
	if( g_pxRT )
	{
		g_pxRT->Release();
		g_pxRT = NULL;
	}

}



////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_BeginCreateShadows()
//
// Begins the process of creating this frame's shadow maps on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_BeginCreateShadows()
{
	// (AAM - 19.01.2009) Now that we deal with shadows during a second predicated tiling bracket,
	// ensure we don't duplicate the work in the second tile!
	Asura_Xbox360_Graphics::GetDevice()->SetPredication( D3DPRED_TILE(0) );

	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_CREATESHADOWS);

	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xFF, 0x80, 0x40 ), "Create Shadows" );

	Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, g_pxRT );
	Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 1, NULL);
	Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 2, NULL);
	Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 3, NULL);
	Asura_Xbox360_Graphics::GetDevice()->SetDepthStencilSurface( g_pxDepthRT );

	// Invalidate the current shadow texture...
	g_uCreateShadowTextureSize	= 0;	
	
	s_uShadowTexturesAllocated = 0;

	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		Asura_Xbox360_SkinRenderer::OwnRegisters();
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_EndCreateShadows()
//
// Finishes creating this frame's shadow maps on Xbox 360
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Shadow_System::Platform_EndCreateShadows()
{
	if( Asura_Xbox360_SkinRenderer::IsEnabled() )
	{
		// Disown the registers we acquired for skin rendering
		Asura_Xbox360_SkinRenderer::DisownRegisters();
	}

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_ALL );
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_COLORWRITEENABLE1, D3DCOLORWRITEENABLE_ALL );	

	Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetFinalRT() );
	Asura_Xbox360_Graphics::GetDevice()->SetDepthStencilSurface( Asura_Xbox360_Graphics::GetFinalDepth() );

	PIXEndNamedEvent(); // Create Shadows

	Asura_Viewport::Set(Asura_Xbox360_Graphics::GetScreenWidth(), Asura_Xbox360_Graphics::GetScreenHeight());
	Asura_Xbox360_Graphics::ApplyCameraSettings();

	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_CREATESHADOWS);

	// Restore predication to normal
	Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
}

void Asura_Shadow_System::Platform_RenderEnvironmentShadowMap( Asura_Shadow &xShadow)
{
	SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Render Env Shadow @ (%.1f, %.1f, %.1f)", xShadow.GetLightPosition().GetX(), xShadow.GetLightPosition().GetY(), xShadow.GetLightPosition().GetZ()));
	
	if (((Asura_Xbox360_Shadow&)xShadow).GetShadowBuffer())
	{

	//	Asura_CameraData xCamData = Asura_Camera::GetCameraData(Asura_Camera::GetCameraIndex());
		Asura_Vector_3 xCamPos = *Asura_Camera::GetPosition();
		Asura_Matrix_3x3 xCamOrient = *Asura_Camera::GetOrientation();
		Asura_Angle fCamFOV = Asura_Camera::GetHorizontalFieldOfView();
		float fNear = Asura_Camera::GetNearClip();
		float fFar = Asura_Camera::GetFarClip();
		float fPhysicalAspect = Asura_Camera::GetPhysicalScreenAspectRatio();
		float fCamAspect = Asura_Camera::GetXOverYAspectRatio() / fPhysicalAspect;
		Asura_ViewingArea xCamArea = *Asura_Camera::GetViewingArea();

		Asura_Matrix_3x3 xOrient;
		xShadow.GetLightOrientation(xOrient);

		Asura_Camera::SetPosition( xShadow.GetLightPosition() );
		Asura_Camera::SetPhysicalScreenAspectRatio(1.0f);
		Asura_Camera::SetHorizontalFieldOfView( Asura_Maths::DegToRad(xShadow.GetFrustum().GetHorizontalFOVInDegrees()), 1.0f);
		Asura_Camera::SetOrientation(xOrient);
		Asura_Camera::SetNearClip(xShadow.GetFrustum().GetNearPlane());
		Asura_Camera::SetFarClip(xShadow.GetFrustum().GetFarPlane());
		Asura_Camera::SetViewingArea();
		Asura_Xbox360_EnvironmentRenderer::RenderToShadow(xShadow.GetFrustum().GetNearPlane(), xShadow.GetFrustum().GetFarPlane(), xShadow.GetInstance()->IsCubeMapped());

		Asura_Camera::SetNearClip(fNear);
		Asura_Camera::SetFarClip(fFar);
		Asura_Camera::SetPosition(xCamPos);
		Asura_Camera::SetOrientation(xCamOrient);
		Asura_Camera::SetPhysicalScreenAspectRatio(fPhysicalAspect);
		Asura_Camera::SetHorizontalFieldOfView(fCamFOV, 1.0f / fCamAspect);
		Asura_Camera::SetViewingArea(xCamArea);
		Asura_Xbox360_Graphics::ApplyCameraSettings();

		Asura_Render::CurrentStates.RequestDynamicLights( true);

		// Invalidate the current shadow texture...
		g_uCreateShadowTextureSize	= 0;	

	}
}

static void RenderShadowOnToScene(const Asura_Shadow_Instance *pxInstance, bool bEnvMap)
{
	const Asura_Frustum&			xFrustum	= pxInstance->GetFrustum();				
	float fNear = pxInstance->GetFrustum().GetNearPlane();
	float fFar = pxInstance->GetFrustum().GetFarPlane();
	float fRange = fFar - fNear;
	float fInvRange = 1.0f / fRange;


	D3DXMATRIX xShadowProj, xInvShadowProj;
	D3DXMatrixPerspectiveFovLH(&xShadowProj, xFrustum.GetVerticalFOVInDegrees() * Asura_Maths::Pi / 180.f, 1.f, xFrustum.GetNearPlane(), xFrustum.GetFarPlane());
	D3DXMatrixInverse(&xInvShadowProj, NULL, &xShadowProj);

	D3DXMATRIX xShadowView, xInvShadowView;
	{
		// Light position
		const Asura_Vector_3 &xLightPos = pxInstance->GetLightPosition();
		D3DXMatrixTranslation( &xShadowView, -xLightPos.x, -xLightPos.y, -xLightPos.z );
		g_pxApplyShadowsEffect->GetInterface()->SetVectorF( g_xLightPosHandle, &xLightPos.x );
	}

	if (!bEnvMap)
	{
		// Light orientation
		D3DXMATRIX xTemp;
		Asura_Matrix_3x3 xLightOri;
		pxInstance->GetLightOrientation(xLightOri);
		D3DXMatrixIdentity( &xTemp );
		for( u_int v = 0; v < 3; ++v )
		{
			for( u_int w = 0; w < 3; ++w )
			{
				xTemp.m[v][w] = xLightOri.Element[v][w];
				// No need to invert y axis here, the shader can just compute an upside-down y
			}
		}
		xShadowView *= xTemp;
	}
	D3DXMatrixInverse(&xInvShadowView, NULL, &xShadowView);
	if (bEnvMap)
	{
		xInvShadowView._11 *= fFar; xInvShadowView._12 *= fFar; xInvShadowView._13 *= fFar;
		xInvShadowView._21 *= fFar; xInvShadowView._22 *= fFar; xInvShadowView._23 *= fFar;
		xInvShadowView._31 *= fFar; xInvShadowView._32 *= fFar; xInvShadowView._33 *= fFar;
	}

	D3DXMATRIX xView, xInvView;
	xView = Asura_Xbox360_Graphics::GetViewMatrix();
	D3DXMatrixInverse(&xInvView, NULL, &xView);

	g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xCameraViewSpaceToWorldSpaceHandle, xInvView );

	if (bEnvMap)
	{
		D3DXMATRIX xLightClipSpaceToCameraClipSpace = xInvShadowView * Asura_Xbox360_Graphics::GetViewProjectionMatrix();
		g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xLightClipSpaceToCameraClipSpaceHandle, xLightClipSpaceToCameraClipSpace );

	} else {
		D3DXMATRIX xLightClipSpaceToCameraClipSpace = xInvShadowProj * xInvShadowView * Asura_Xbox360_Graphics::GetViewProjectionMatrix();
		g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xLightClipSpaceToCameraClipSpaceHandle, xLightClipSpaceToCameraClipSpace );
		D3DXMATRIX xCameraViewSpaceToLightClipSpace = xInvView * xShadowView * xShadowProj;
		g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xCameraViewSpaceToLightClipSpaceHandle, xCameraViewSpaceToLightClipSpace );
		D3DXMATRIX xCameraViewSpaceToLightViewSpace = xInvView * xShadowView;
		g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xCameraViewSpaceToLightViewSpaceHandle, xCameraViewSpaceToLightViewSpace );
	}

	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xShadowFrustumNearPlaneDistanceHandle, fNear );
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xOneOverShadowFrustumPlaneSeparationHandle, fInvRange );


	float fFilterFade = 1.0f;
	bool bFilterShadows = false;
	bool bUseNoise = false;
	if ( !g_bDisableFiltering)
	{
		bFilterShadows = (pxInstance->GetDistance() < Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering);

		fFilterFade = 0.25f * (Asura_Shadow_System::GetConfig().m_fMaxDistanceForShadowFiltering - Asura_Maths::Max(pxInstance->GetDistance(), 0.0f));
		fFilterFade = Asura_Maths::ClampToRange(fFilterFade, 0.0f,1.0f);
	} 

	bUseNoise = g_uNoiseMode == 1 ? ((pxInstance->GetShadowFlags() & ASURA_SHADOWFLAG_USE_NOISE) != 0) : (g_uNoiseMode == 2);

	g_pxApplyShadowsEffect->GetInterface()->SetFloat(g_xShadowFilterTextureScaleHandle, fFilterFade * 11.0f / ((float)(ASURA_MAX_SHADOW_TEXTURE_SIZE))); 
	g_pxApplyShadowsEffect->GetInterface()->SetFloat(g_xShadowFilterSizeHandle , 0.2f*pxInstance->GetSoftness());

	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xShadowFrustumNearPlaneDistanceHandle, xFrustum.GetNearPlane() );
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xOneOverShadowFrustumPlaneSeparationHandle, 1.f / (xFrustum.GetFarPlane() - xFrustum.GetNearPlane()) );
	
	const float fTanHalfVertFOV = Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xCameraTanHalfVertFOVHandle, fTanHalfVertFOV );

	const float fTanHalfHorizFOV = fTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xCameraTanHalfHorizFOVHandle, fTanHalfHorizFOV );

	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xTwoOverScreenWidthHandle, 2.0f / (Asura_Viewport::GetWidth() * Asura_Camera::GetViewingArea()->m_fWidth) );
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xTwoOverScreenHeightHandle, 2.0f / (Asura_Viewport::GetHeight() * Asura_Camera::GetViewingArea()->m_fHeight) );
	FXLVECTOR xScreenOffset;
	xScreenOffset.x = Asura_Camera::GetViewingArea()->m_fXOffset * Asura_Viewport::GetWidth();
	xScreenOffset.y = Asura_Camera::GetViewingArea()->m_fYOffset * Asura_Viewport::GetHeight();
	g_pxApplyShadowsEffect->GetInterface()->SetVector( g_xScreenOffsetHandle,  xScreenOffset);
	


	g_pxApplyShadowsEffect->GetInterface()->SetSampler( g_xNormalBufferHandle,   Asura_Xbox360_Graphics::GetNormalTexture() );
	g_pxApplyShadowsEffect->GetInterface()->SetSampler( g_xDepthMtlBufferHandle, Asura_Xbox360_Graphics::GetDepthMtlTexture() );

	g_pxApplyShadowsEffect->GetInterface()->SetSampler(g_xKernelRotationTextureHandle, g_pxFilterKernelRotationTexture);


	g_pxApplyShadowsEffect->GetInterface()->SetVectorF( g_xShadowDirectionHandle, &pxInstance->GetLightDirection().x );

	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xShadowStrengthHandle, pxInstance->GetStrength());


	if (bEnvMap)
	{
		g_pxApplyShadowsEffect->GetInterface()->SetVectorF( g_xSelfShadowBiasHandle, s_xSelfShadowBiasValues[2]);
		// Are we inside the volume?
		Asura_Vector_3 xOfs = *Asura_Camera::GetPosition() - pxInstance->GetLightPosition();
		float fDist = xOfs.Magnitude();

		// Pick a sphere LOD. Switch down quickly to the low LOD, as we only need a more round
		// sphere when the camera is about to move inside it, for a more accurate inside-outside test
		int iSphereLOD = (int)(fDist*0.5f/pxInstance->GetFrustum().GetFarPlane());
		if (iSphereLOD>3)
		{
			iSphereLOD = 3;
		}		
		else if (iSphereLOD<0)
		{
			iSphereLOD = 0;
		}

		Asura_Xbox360_DeferredShading::s_apxSphereVB[iSphereLOD]->Bind();
		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( g_pxShadowFrustumVD );
		g_pxApplyShadowsEffect->Begin(GetShadowTechnique(bFilterShadows, true, false, bUseNoise, g_bEnableScreenBasedNoise),0);

		for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
		{
			if (Asura_Xbox360_Graphics::IsTiling2())
			{
				Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
				Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
			}
			g_pxApplyShadowsEffect->Commit();
			Asura_Xbox360_Graphics::GetDevice()->DrawVertices( D3DPT_TRIANGLESTRIP, 0, Asura_Xbox360_DeferredShading::s_aiSphereNumVerts[iSphereLOD]-2);		
		}

		g_pxApplyShadowsEffect->End();
	}
	else
	{
		g_pxApplyShadowsEffect->GetInterface()->SetVectorF( g_xSelfShadowBiasHandle, s_xSelfShadowBiasValues[pxInstance->IsEnvironment()]);
		
		if (dvs_bEnableStandardisedShadows)
		{
			g_pxApplyShadowsEffect->Begin(Technique_Standardised_NonCubic,0);
		}
		else
		{
			g_pxApplyShadowsEffect->Begin(GetShadowTechnique(bFilterShadows, false, g_bEnablePCF, bUseNoise, g_bEnableScreenBasedNoise),0);
		}
		
		for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
		{
			if (Asura_Xbox360_Graphics::IsTiling2())
			{
				Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
				Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
			}
			g_pxApplyShadowsEffect->Commit();
			Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_QUADLIST, 0, 0, 24 );		
		}
		g_pxApplyShadowsEffect->End();
	}
	if (Asura_Xbox360_Graphics::IsTiling2())
	{
		Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
		Asura_Xbox360_Graphics::CorrectVPOSForTiling(0);
	}
}


////////////////////////////////////////////////////////////////////////////////
//
// Asura_Shadow_System::Platform_RenderOntoScene()
//
// Renders this frame's shadows onto the scene
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////


void Asura_Shadow_System::Platform_RenderOnToScene(ASURA_SHADOW_LIGHTSOURCE eLightSource, uint32 uLightSourceHandle)
{
	Direct3D_Device* const pxDevice = Asura_Xbox360_Graphics::GetDevice();
	const Asura_Shadow_Instance *apxCubeMapShadows[ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS];
	u_int uCubeMapShadowsFound = 0;
	u_int uTotalCubeMapShadowsFound = 0;;



	// DMR 03/04/07
	//
	// If we're rendering shadows from averaged lights, they're still done as a darkening pass, as before.
	//
	// If we're rendering shadows from a real light source, we write them to the destination alpha channel, and the deferred light
	// pass will use a D3DBLEND_INVDESTALPHA blend, to mask out the shadowed areas.
	//
	// This relies on the destination alpha channel (at least the area affected by the light) being cleared to 0 alpha before this call.
	
	if( 0 == s_uNumShadows)
	{
		return;
	}
	
	SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Apply %d Shadows", s_uNumShadows));

	if(!g_pxApplyShadowsEffect) 
	{
		return;
	}

	pxDevice->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetFinalRT() );
	pxDevice->SetDepthStencilSurface( Asura_Xbox360_Graphics::GetFinalDepth() );

	Asura_Render::CurrentStates.RequestAlphaTest( false );
	Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
	Asura_Render::CurrentStates.RequestZBufferEnabled( false );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );

	pxDevice->SetVertexDeclaration( g_pxShadowFrustumVD );
	g_xShadowFrustumVB.Bind();
	g_xShadowFrustumIB.Bind();

	g_pxApplyShadowsEffect->GetInterface()->SetBool(g_xShadowMaskHandle, eLightSource != ASURA_SHADOW_LIGHTSOURCE_AVERAGED);
	// If we're creating a shadow mask for a light source, using the MAX operation is 
	// a nice way of avoiding double-shadowing	
	if (eLightSource != ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
	{
		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
		D3DBLENDSTATE xBlendState;
		xBlendState.BlendOp = xBlendState.BlendOpAlpha = D3DBLENDOP_MAX;
		xBlendState.SrcBlend = xBlendState.SrcBlendAlpha = D3DBLEND_ONE;
		xBlendState.DestBlend = xBlendState.DestBlendAlpha = D3DBLEND_ONE;
		Asura_Xbox360_Graphics::GetDevice()->SetBlendState( 0, xBlendState );
	}

	// Set parameters that don't depend on the shadow

	const float fTanHalfVertFOV = Asura_Maths::Tangent( Asura_Camera::GetVerticalFieldOfView() * 0.5f );
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xCameraTanHalfVertFOVHandle, fTanHalfVertFOV );

	const float fTanHalfHorizFOV = fTanHalfVertFOV * Asura_Camera::GetXOverYAspectRatio();
	g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xCameraTanHalfHorizFOVHandle, fTanHalfHorizFOV );

	g_pxApplyShadowsEffect->GetInterface()->SetBool(g_xCubicMapShadowHandle, false);

	Asura_Xbox360_Graphics::ApplyCameraSettings();

	// Begin rendering
	{
		SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "%d Shadows", s_uNumShadows));
		for( u_int u = 0; u < s_uNumShadows; ++u )
		{
			const Asura_Xbox360_Shadow&		xShadow				= *static_cast< const Asura_Xbox360_Shadow* >( s_apxShadows[u] );
			const Asura_Shadow_Instance*	pxInstance	= xShadow.GetInstance();

			bool bAccept = false;			

			if (pxInstance->GetLightSourceType() == eLightSource)
			{
				if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
				{
					// We're casting shadows from an averaged light source - there's no light source handle 
					bAccept = true;
				}
				else
				{
					// We're casting shadows from a real light source - make sure this is the right light source
					if (uLightSourceHandle == pxInstance->GetLightSourceHandle())
					{
						bAccept = true;					
					}
				}		
			}
			else
			{
				if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED)
				{
					// If we're rendering averaged shadows, we also want to render any darkening pass 
					// shadows from specific light sources here, too
					if (pxInstance->GetShadowFlags()&ASURA_SHADOWFLAG_IS_DARKENING_PASS)
					{
						bAccept = true;
					}
				}
			}
			if (pxInstance->IsCubeMapped())
			{
				uTotalCubeMapShadowsFound++;
			}

			if (bAccept	&& pxInstance->IsCubeMapped())
			{
				if (uCubeMapShadowsFound < ASURA_MAX_NUMBER_OF_CUBEMAP_SHADOWS)
				{
					apxCubeMapShadows[uCubeMapShadowsFound++] = pxInstance;
				}
			} 
			else if (bAccept) 
			{
				SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Shadow %u @ (%.1f, %.1f, %.1f)", u, xShadow.GetLightPosition().GetX(), xShadow.GetLightPosition().GetY(), xShadow.GetLightPosition().GetZ()));
				g_pxApplyShadowsEffect->GetInterface()->SetSampler( g_xShadowBufferHandle, xShadow.GetShadowBuffer()->GetTexture() );
				s_xUVXForm._24 = s_xUVXForm._14 = 0.5f + (0.5f / xShadow.GetShadowBuffer()->GetTextureSize());
				s_xUVXForm._14 += g_fXShadowOffset;
				s_xUVXForm._24 += g_fYShadowOffset;
				g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xUVXFormHandle, s_xUVXForm );

				g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xShadowTexSizeHandle, (float)xShadow.GetShadowBuffer()->GetTextureSize() );


				if (eLightSource == ASURA_SHADOW_LIGHTSOURCE_AVERAGED && g_bEnableStencil && Asura_DeferredShading::GetUseStencilBuffer())
				{
					Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_NONE );
					Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_ALWAYS);		
					Asura_Render::CurrentStates.RequestZBufferEnabled( true );

					const Asura_Frustum&			xFrustum	= pxInstance->GetFrustum();				


					g_pxApplyShadowsEffect->Begin(Technique_Standardised_NonCubic, 0);	

					D3DXMATRIX xShadowProj, xInvShadowProj;
					D3DXMatrixPerspectiveFovLH(&xShadowProj, xFrustum.GetVerticalFOVInDegrees() * Asura_Maths::Pi / 180.f, 1.f, xFrustum.GetNearPlane(), xFrustum.GetFarPlane());
					D3DXMatrixInverse(&xInvShadowProj, NULL, &xShadowProj);

					D3DXMATRIX xShadowView, xInvShadowView;
					const Asura_Vector_3 &xLightPos = pxInstance->GetLightPosition();
					D3DXMatrixTranslation( &xShadowView, -xLightPos.x, -xLightPos.y, -xLightPos.z );
					// Light orientation
					D3DXMATRIX xTemp;
					Asura_Matrix_3x3 xLightOri;
					pxInstance->GetLightOrientation(xLightOri);
					D3DXMatrixIdentity( &xTemp );
					for( u_int v = 0; v < 3; ++v )
					{
						for( u_int w = 0; w < 3; ++w )
						{
							xTemp.m[v][w] = xLightOri.Element[v][w];
							// No need to invert y axis here, the shader can just compute an upside-down y
						}
					}
					xShadowView *= xTemp;

					D3DXVECTOR4 xCamPos;
					D3DXVECTOR4 xCamWorldPos(Asura_Camera::GetPosition()->GetX(),Asura_Camera::GetPosition()->GetY(),Asura_Camera::GetPosition()->GetZ(), 1.0f);
					D3DXVec4Transform(&xCamPos, &xCamWorldPos, &xShadowView);
					bool bInside = xFrustum.Intersects(*(Asura_Vector_3*)&xCamPos, 0.1f);

					D3DXMatrixInverse(&xInvShadowView, NULL, &xShadowView);

					D3DXMATRIX xLightClipSpaceToCameraClipSpace = xInvShadowProj * xInvShadowView * Asura_Xbox360_Graphics::GetViewProjectionMatrix();
					g_pxApplyShadowsEffect->GetInterface()->SetMatrixF( g_xLightClipSpaceToCameraClipSpaceHandle, xLightClipSpaceToCameraClipSpace );

					g_pxApplyShadowsEffect->Commit();
					// need to make sure the last bound RT was 32bpp as the hardware wont do fast z writes with a 64bit RT!
					Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetPostToneMapRT());
					Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, NULL);

					for (int iPass=0; iPass<!bInside + 1; iPass++)
					{

						switch(iPass)
						{		
						case 0:		
							// Back faces
							Asura_Xbox360_Graphics::GetDevice()->SetPixelShader(NULL);
							Asura_Render::CurrentStates.RequestStencilEnabled(true);
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILFUNC,			D3DCMP_ALWAYS );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILREF,			ASURA_STENCIL_MASK_PHOTON );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILWRITEMASK,		ASURA_STENCIL_MASK_PHOTON );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILMASK,			ASURA_STENCIL_MASK_PHOTON );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILPASS,			D3DSTENCILOP_ZERO );				
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILZFAIL,			D3DSTENCILOP_REPLACE);
							if (bInside)
							{
								Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE, TRUE );
								Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILFUNC,		D3DHSCMP_EQUAL );
								Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILREF,			0x0 );
							} else {
								Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE, FALSE );
							}

							Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
							Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		

							break;

						case 1:
							// Front faces
							Asura_Xbox360_Graphics::GetDevice()->SetPixelShader(NULL);				
							Asura_Render::CurrentStates.RequestStencilEnabled(true);
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILFUNC,			D3DCMP_ALWAYS );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILREF,			ASURA_STENCIL_MASK_PHOTON );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILWRITEMASK,		ASURA_STENCIL_MASK_PHOTON );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILMASK,			ASURA_STENCIL_MASK_PHOTON );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILPASS,			D3DSTENCILOP_DECR );				
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILZFAIL,			D3DSTENCILOP_REPLACE );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE, TRUE );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILFUNC,		D3DHSCMP_NOTEQUAL );
							Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILREF,			0 );


							Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );		
							Asura_Render::CurrentStates.RequestZBufferCompareMode(ASURA_ZBUFFERCMP_LESSEQUAL);		

							break;
						}			
						Asura_Xbox360_Graphics::GetDevice()->SetPixelShader(NULL);
						Asura_Xbox360_Graphics::GetDevice()->DrawIndexedVertices( D3DPT_QUADLIST, 0, 0, 24 );		
					}
					g_pxApplyShadowsEffect->End();				
					Asura_Xbox360_Graphics::GetDevice()->FlushHiZStencil( D3DFHZS_ASYNCHRONOUS );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetFinalRT());


					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, TRUE );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE,	FALSE );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILFUNC,				bInside ? D3DCMP_NOTEQUAL : D3DCMP_EQUAL );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILREF,				0x00 );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILWRITEMASK,		ASURA_STENCIL_MASK_PHOTON );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_STENCILMASK,			ASURA_STENCIL_MASK_PHOTON );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILWRITEENABLE,		FALSE );
					Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE,			TRUE );		
					Asura_Render::CurrentStates.RequestZBufferCompareMode( ASURA_ZBUFFERCMP_GREATER );
					Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CCW );
					Asura_Render::CurrentStates.RequestZBufferEnabled( true );

				}


				RenderShadowOnToScene(pxInstance, false);
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HISTENCILENABLE, FALSE );
				s_uNumShadowInstancesRendered++;
			}
		}
		g_pxApplyShadowsEffect->GetInterface()->SetBool(g_xCubicMapShadowHandle, true);
	}

	if (uCubeMapShadowsFound && g_bEnableCubeMap)
	{
		SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "%d Cubic Env Map Shadows", uCubeMapShadowsFound));
	
		// KJFL - 7 May 2009 -
		// Make a list of the strip visibilities for the cubemap. May be better to make a list
		// of used strips rather than using a bitfield - toying with making this multithreaded
		// if we're still CPU bound. (Not happy with the allocate, but it's less costly than
		// repeating the tests for each cube face, and I'd rather not allocate a buffer globally
		// at this stage.)
		u_int uNumberOfStrips = Asura_Xbox360_EnvironmentRenderer::GetNumStrips();
		u_char* pxStripVisibilities = new u_char[ uNumberOfStrips ];

		for (u_int uCubeMappedShadow = 0;uCubeMappedShadow < uCubeMapShadowsFound; uCubeMappedShadow++)
		{
			const Asura_Shadow_Instance *pxInstance = apxCubeMapShadows[uCubeMappedShadow];
			float fNear = pxInstance->GetFrustum().GetNearPlane();
			float fFar = pxInstance->GetFrustum().GetFarPlane();
			float fRange = fFar - fNear;
			float fInvRange = 1.0f / fRange;

			SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Shadow %d @ (%.1f, %.1f, %.1f)", uCubeMappedShadow, pxInstance->GetLightPosition().GetX(), pxInstance->GetLightPosition().GetY(), pxInstance->GetLightPosition().GetZ()));

			g_pxApplyShadowsEffect->GetInterface()->SetSampler( g_xShadowBufferHandle, NULL );
			g_pxApplyShadowsEffect->GetInterface()->SetSampler( g_xShadowBufferCubeHandle, NULL );

			g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xShadowFrustumNearPlaneDistanceHandle, fNear );
			g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xOneOverShadowFrustumPlaneSeparationHandle, fInvRange );

			Asura_Xbox360_Graphics::GetDevice()->SetRenderState( D3DRS_COLORWRITEENABLE, D3DCOLORWRITEENABLE_RED);
			Asura_Render::CurrentStates.RequestZBufferEnabled( true );
			Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );
			Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_OFF); 

			//	Asura_CameraData xCamData = Asura_Camera::GetCameraData(Asura_Camera::GetCameraIndex());
			Asura_Vector_3 xCamPos = *Asura_Camera::GetPosition();
			Asura_Matrix_3x3 xCamOrient = *Asura_Camera::GetOrientation();
			Asura_Angle fCamFOV = Asura_Camera::GetHorizontalFieldOfView();
			float fCamNear = Asura_Camera::GetNearClip();
			float fCamFar = Asura_Camera::GetFarClip();
			float fPhysicalAspect = Asura_Camera::GetPhysicalScreenAspectRatio();
			float fCamAspect = Asura_Camera::GetXOverYAspectRatio() / fPhysicalAspect;
			Asura_ViewingArea xCamArea = *Asura_Camera::GetViewingArea();

			Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, g_pxRT );
			Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 1, NULL);
			Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 2, NULL);
			Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 3, NULL);
			Asura_Xbox360_Graphics::GetDevice()->SetDepthStencilSurface( g_pxDepthRT );

			static u_int uNextCubeMap = 0;
			u_int uCubeMapToUse = 0;

			// KJFL - 7 May 2009 - quick test optimisation: if the light position and range 
			// is the same as the last one rendered, re-use the cube texture. Saves a lot of
			// time for when there is just one environment shadow light in the scene! (But
			// currently goes back to normal behaviour if there's two env shadow lights or 
			// more - we'd need to change the light ordering in the independent code.)
			// KJFL - 14 May 2009 - doubled up the buffers so we can have two of them.
			if ( uTotalCubeMapShadowsFound <= 2 && pxInstance->GetLightPosition() == g_xLastLightPosition[0] && pxInstance->GetFrustum().GetFarPlane() == g_fLastLightRange[0] )
			{
				// reuse cube map 0
				// make sure we overwrite this one last
				uNextCubeMap = 1;
			}
 			else if ( uTotalCubeMapShadowsFound <= 2 && pxInstance->GetLightPosition() == g_xLastLightPosition[1] && pxInstance->GetFrustum().GetFarPlane() == g_fLastLightRange[1] )
 			{
 				// reuse cube map 1
 				uCubeMapToUse = 1;
				// make sure we overwrite this one last
				uNextCubeMap = 0;
 			}
 			else
			{
				uCubeMapToUse = uNextCubeMap;
				g_xLastLightPosition[uCubeMapToUse] = pxInstance->GetLightPosition();
				g_fLastLightRange[uCubeMapToUse] = pxInstance->GetFrustum().GetFarPlane();

				uNextCubeMap = (uNextCubeMap+1)&1;

				{
					SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Shadow %d EnvCull", uCubeMappedShadow ));
					const float fSX = pxInstance->GetLightPosition().GetX();
					const float fSY = pxInstance->GetLightPosition().GetY();
					const float fSZ = pxInstance->GetLightPosition().GetZ();
				
					float fLightRangeSqd =  pxInstance->GetFrustum().GetFarPlane();

					fLightRangeSqd *= fLightRangeSqd;
					for( u_int u = 0; u < uNumberOfStrips; ++u )
					{
						const Asura_Xbox360_EnvironmentRenderer_Strip* pxStrip = &Asura_Xbox360_EnvironmentRenderer::GetStrip(u);

						pxStripVisibilities[u] = 0;

						const Asura_Bounding_Box& xBB = pxStrip->m_xBoundingBox;
						const float fPX = Asura_Maths::ClampToRange( fSX, xBB.MinX, xBB.MaxX );
						const float fPY = Asura_Maths::ClampToRange( fSY, xBB.MinY, xBB.MaxY );
						const float fPZ = Asura_Maths::ClampToRange( fSZ, xBB.MinZ, xBB.MaxZ );
					
						float fX = fPX - fSX;
						float fY = fPY - fSY;
						float fZ = fPZ - fSZ;
						
						float fDistanceSquared = fX*fX+fY*fY+fZ*fZ;

						if (fDistanceSquared>fLightRangeSqd) continue;

						{
							const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( pxStrip->m_auMaterialResponseHashIDs[0] );

							// Strip out unwanted materials
							if( pxMaterial1 )
							{
								if( pxMaterial1->GetDontCastShadow() || pxMaterial1->IsTransparent() ) continue;

								const bool bSpecialRender =	pxMaterial1->GetSpecialFXMode() != Asura_MaterialResponse::SPECIALFX_MODE_NONE;

								if( bSpecialRender ) continue;
							}
						}

						pxStripVisibilities[u] = 0xff;
					}

					Asura_Xbox360_EnvironmentRenderer::RenderToShadowCubemap_Initialise(fNear, fFar);
				}
			
				Asura_Camera::SetPosition( pxInstance->GetLightPosition() );
				Asura_Camera::SetPhysicalScreenAspectRatio(1.0f);
				Asura_Camera::SetHorizontalFieldOfView( Asura_Maths::Deg90, 1.0f  );

				Asura_Camera::SetNearClip(fCamNear);
				Asura_Camera::SetFarClip(fCamFar);
				Asura_Camera::SetViewingArea();

				for (int i = 0; i < 6; ++i)
				{
					SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Shadow %d RenderToCubemap Face:%d", uCubeMappedShadow, i ));

					static const Asura_Matrix_3x3 axORIENTATIONS[6] =
					{
						// Looking at +x face
						Asura_Matrix_3x3
						(
							 0.0f, 0.0f, 1.0f,
							 0.0f, 1.0f, 0.0f,
							-1.0f, 0.0f, 0.0f
						),

						// Looking at -x face
						Asura_Matrix_3x3
						(
							 0.0f, 0.0f, -1.0f,
							 0.0f, 1.0f,  0.0f,
							 1.0f, 0.0f,  0.0f
						),

						// Looking at -y face
						Asura_Matrix_3x3
						(
							 1.0f, 0.0f,  0.0f,
							 0.0f, 0.0f, -1.0f,
							 0.0f, 1.0f,  0.0f
						),

						// Looking at +y face
						Asura_Matrix_3x3
						(
							 1.0f,  0.0f, 0.0f,
							 0.0f,  0.0f, 1.0f,
							 0.0f, -1.0f, 0.0f
						),

						// Looking at +z face
						Asura_Matrix_3x3
						(
							 1.0f, 0.0f, 0.0f,
							 0.0f, 1.0f, 0.0f,
							 0.0f, 0.0f, 1.0f
						),

						// Looking at -z face
						Asura_Matrix_3x3
						(
							-1.0f, 0.0f,  0.0f,
							 0.0f, 1.0f,  0.0f,
							 0.0f, 0.0f, -1.0f
						),
					};
					const Asura_Matrix_3x3 &xOrient = axORIENTATIONS[i];

					Asura_Viewport::Set(ASURA_CUBEMAP_SHADOW_SIZE, ASURA_CUBEMAP_SHADOW_SIZE);
					Asura_Xbox360_Graphics::GetDevice()->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_XRGB( 0xFF, 0xFF, 0xFF ), 1.0f, 0 );

					Asura_Camera::SetOrientation(xOrient);
					Asura_Xbox360_EnvironmentRenderer::RenderToShadowCubemap( pxStripVisibilities );

					Asura_Xbox360_Graphics::GetDevice()->Resolve(D3DRESOLVE_RENDERTARGET0, NULL, s_pxEnvShadowCubeTexture[uCubeMapToUse], NULL, 0, (D3DCUBEMAP_FACES)i, NULL, 1.0f, 0, NULL);
				}
			}

			{
				SCOPEDPIXEVENT((D3DCOLOR_XRGB(0xFF, 0x80, 0x40), "Shadow RenderToScene %d", uCubeMappedShadow ));

				Asura_Xbox360_Graphics::GetDevice()->SetRenderTarget( 0, Asura_Xbox360_Graphics::GetFinalRT() );
				Asura_Xbox360_Graphics::GetDevice()->SetDepthStencilSurface( Asura_Xbox360_Graphics::GetFinalDepth() );


				Asura_Viewport::Set(Asura_Xbox360_Graphics::GetScreenWidth(), Asura_Xbox360_Graphics::GetScreenHeight());
				Asura_Camera::SetNearClip(fCamNear);
				Asura_Camera::SetFarClip(fCamFar);
				Asura_Camera::SetPosition(xCamPos);
				Asura_Camera::SetOrientation(xCamOrient);
				Asura_Camera::SetPhysicalScreenAspectRatio(fPhysicalAspect);
				Asura_Camera::SetHorizontalFieldOfView(fCamFOV, 1.0f / fCamAspect);
				Asura_Camera::SetViewingArea(xCamArea);

				Asura_Xbox360_Graphics::ApplyCameraSettings();
				Asura_Render::CurrentStates.RequestDynamicLights( true);

				Asura_Xbox360_Graphics::GetDevice()->SetRenderState
				(
					D3DRS_COLORWRITEENABLE,
					(pxInstance->GetShadowFlags()&ASURA_SHADOWFLAG_IS_DARKENING_PASS) ? (D3DCOLORWRITEENABLE_RED|D3DCOLORWRITEENABLE_GREEN|D3DCOLORWRITEENABLE_BLUE) : D3DCOLORWRITEENABLE_ALPHA
				);

				g_pxApplyShadowsEffect->GetInterface()->SetSampler( g_xShadowBufferCubeHandle, s_pxEnvShadowCubeTexture[uCubeMapToUse] );
				pxDevice->SetVertexDeclaration( g_pxShadowFrustumVDSphere );


				Asura_Render::CurrentStates.RequestAlphaTest( false );
				Asura_Render::CurrentStates.RequestCullMode( ASURA_CULLMODE_CW );
				Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
				Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );
				Asura_Render::CurrentStates.RequestZBufferEnabled( false );

				Asura_Render::CurrentStates.RequestZBufferWriteEnabled( false );

				if (!(pxInstance->GetShadowFlags()&ASURA_SHADOWFLAG_IS_DARKENING_PASS))
				{
					Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
					D3DBLENDSTATE xBlendState;
					xBlendState.BlendOp = xBlendState.BlendOpAlpha = D3DBLENDOP_MAX;
					xBlendState.SrcBlend = xBlendState.SrcBlendAlpha = D3DBLEND_ONE;
					xBlendState.DestBlend = xBlendState.DestBlendAlpha = D3DBLEND_ONE;
					Asura_Xbox360_Graphics::GetDevice()->SetBlendState( 0, xBlendState );
				}
				g_pxApplyShadowsEffect->GetInterface()->SetFloat( g_xShadowTexSizeHandle, (float)ASURA_CUBEMAP_SHADOW_SIZE );

				RenderShadowOnToScene(pxInstance, true);
			}
		}
		delete [] pxStripVisibilities;
		
	}

	// Restore the blend state to what Asura thinks it's still set to
	D3DBLENDSTATE xBlendState;
	xBlendState.BlendOp   = xBlendState.BlendOpAlpha   = D3DBLENDOP_ADD;
	xBlendState.SrcBlend  = xBlendState.SrcBlendAlpha  = D3DBLEND_SRCALPHA;
	xBlendState.DestBlend = xBlendState.DestBlendAlpha = D3DBLEND_INVSRCALPHA;
	Asura_Xbox360_Graphics::GetDevice()->SetBlendState( 0, xBlendState );	

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState(D3DRS_BLENDOP, D3DBLENDOP_ADD);
	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled( true );


}




/******************************************************************************/
/* Asura_Xbox360_Shadow Class Functions                                       */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Shadow::Asura_Xbox360_Shadow()
//
// Constructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_Shadow::Asura_Xbox360_Shadow ()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Shadow::~Asura_Xbox360_Shadow()
//
// Destructor
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

Asura_Xbox360_Shadow::~Asura_Xbox360_Shadow()
{
}

////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Shadow::PrepareForRendering()
//
// Begins rendering into a shadow map
//
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_Shadow::PrepareForRendering( void )
{
	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xC0, 0x60, 0x30 ), "Create Shadow" );

	if (m_pxShadowBuffer)
	{
		m_pxShadowBuffer->Bind();
		Asura_Viewport::Set(m_pxShadowBuffer->GetTextureSize(), m_pxShadowBuffer->GetTextureSize());
		const D3DVECTOR4 xClearColour = { 1.0f, 0.0f, 0.0f, 0.0f };
		Asura_Xbox360_Graphics::GetDevice()->ClearF( D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, NULL, &xClearColour, 1.0f, 0 );
		if(g_uCreateShadowTextureSize != m_pxShadowBuffer->GetTextureSize())
		{
			// Fix up the viewport
			g_xCreateShadowViewport.X = g_xCreateShadowViewport.Y = 0;
			g_xCreateShadowViewport.Width = g_xCreateShadowViewport.Height = m_pxShadowBuffer->GetTextureSize();
			g_xCreateShadowViewport.MinZ = 0.0f;
			g_xCreateShadowViewport.MaxZ = 1.0f;
	
			Asura_Xbox360_Graphics::GetDevice()->SetViewport( &g_xCreateShadowViewport );
	
			g_uCreateShadowTextureSize = m_pxShadowBuffer->GetTextureSize();
		}
	}

	Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
}

#ifdef _DEBUG
void Asura_Shadow_System::ShowDebugShadowMaps()
{
	static Asura_Xbox360_Effect *pxEffect;

	if (!pxEffect)	
		pxEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auShadowBufferOverlayRender);

	Asura_Assert( pxEffect, "Failed to create normal map debug shader" );

	Asura_Overlay_Quad xQuad;

	const static float fOutputSizeTotalX = 8.0f;
	const static float fOutputSizeTotalY = 8.0f;
	float fW = 128;//(float)(Asura_Viewport::GetWidth()/fOutputSizeTotalX);
	float fH = 128;//(float)(Asura_Viewport::GetHeight()/fOutputSizeTotalY);

	float fX = 0.0f, fY = 0.0f;

	for( u_int uDrawTargetIndex = 0; uDrawTargetIndex < Asura_Shadow_System::s_uNumShadows; uDrawTargetIndex += 1)
	{
		Asura_Xbox360_Shadow* pxShadow = (Asura_Xbox360_Shadow*) Asura_Shadow_System::s_apxShadows[uDrawTargetIndex];
		if (!pxShadow || !pxShadow->GetShadowBuffer())
			continue;
		
		Direct3D_Texture* pxMRTTexture = pxShadow->GetShadowBuffer()->GetTexture();

		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
		xQuad.Colour = 0xff0000ff;
		xQuad.SetRectUV(0,0,1,1);
		xQuad.SetRectXY(fX,fY,fW,fH);	
		fX += fW + 2;
		if (fX + fW + 2 > Asura_Viewport::GetWidth())
		{
			fX = 0.0f;
			fY += fH;
		}

		Asura_Xbox360_QuadOverlayRenderer::RenderEx(xQuad, pxMRTTexture, pxEffect);
		Asura_Xbox360_QuadOverlayRenderer::Flush();
	}

	// Tidy-up...
	DestroyShadowInstances();
}
#endif


////////////////////////////////////////////////////////////////////////////////
//
// Asura_Xbox360_Shadow::FinishedRendering()
//
// Finishes rendering into a shadow map
//
// Parameters:		None
// Return value:	None
//
////////////////////////////////////////////////////////////////////////////////

void Asura_Xbox360_Shadow::FinishedRendering()
{
	if( m_pxShadowBuffer )
	{
		Asura_Xbox360_Graphics::GetDevice()->Resolve( D3DRESOLVE_RENDERTARGET0, NULL, m_pxShadowBuffer->GetTexture(), NULL, 0, 0, NULL, 1.0f, 0, NULL );
	}

	PIXEndNamedEvent(); // Create Shadow
}
