//
// Asura_Platform_OpenAL.cpp
//
// Created by Andy Patrick - 20-09-2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Sound/Asura_OpenAL.h"
#if defined( ASURA_USE_OPENAL ) && ( ASURA_USE_OPENAL == 1 )
#include "Core/Asura_Timers.h"
#include "Debugging//Asura_Logfile.h"
#include "OpenAL/OpenAL1-1SDK/include/Xbox360/alconsole.h"
#include "OpenAL/OpenAL1-1SDK/include/Xbox360/alDebugXbox360.h"
#include "Sound/Asura_Xbox360_Sound.h"
#include <xauddefs.h>

/* ------------------------------------------------------------------------- */

// We currently leak about 4KB per sound with this turned off, plus an occasional
// 300KB if a sound plays when no other sounds are currently playing. OpenAL leak.
#define LIMIT_TO_ONE_AUXILIARY_SEND_PER_VOICE	(1)

// Only define this when you are specifically looking for OpenAL problems
// and have linked to openal32d.lib. (Not needed the rest of the time)
//#define LINKING_TO_OPENAL_DEBUG_LIB (1)

/* ------------------------------------------------------------------------- */

// Forward definitions
static void Xbox360_OpenAL_XMAError( ALuint uStatus, ALuint uSource, ALuint uBuffer );
static void* Xbox360_OpenAL_Alloc( u_int size );
static void Xbox360_OpenAL_Dealloc( void* ptr );

/* ------------------------------------------------------------------------- */

Asura_OpenAL_Sound_API& Asura_OpenAL_Sound_API::Platform_GetAPI()
{
	return Asura_Xbox360_Sound::GetAPI();
}

/* ------------------------------------------------------------------------- */

void Asura_OpenAL_Sound_API::Platform_Initialise()
{
	#if defined( _DEBUG ) && defined( LINKING_TO_OPENAL_DEBUG_LIB )
		g_pDebugLevels[OPENAL_DEBUG_FILTER] = OPENAL_DEBUG_NORMAL;
		g_pDebugLevels[OPENAL_DEBUG_ERRORS] = OPENAL_DEBUG_NORMAL;
		g_pDebugLevels[OPENAL_DEBUG_FXREVERB] = OPENAL_DEBUG_NORMAL;
		g_pDebugLevels[OPENAL_DEBUG_MEMORY] = OPENAL_DEBUG_NORMAL;
		g_pDebugLevels[OPENAL_DEBUG_SOURCE_AUXSENDS] = OPENAL_DEBUG_NORMAL;
		g_pDebugLevels[OPENAL_DEBUG_SOURCE_SUBMIX] = OPENAL_DEBUG_NORMAL;
		//g_pDebugLevels[OPENAL_DEBUG_SOURCE_PARAMS] = OPENAL_DEBUG_VERBOSE;
		g_pDebugLevels[OPENAL_DEBUG_SOURCE] = OPENAL_DEBUG_NORMAL;
	#endif

	//alMemoryAllocFns( Asura_Allocate, Asura_Deallocate );
	alMemoryAllocFns( Xbox360_OpenAL_Alloc, Xbox360_OpenAL_Dealloc );
	alSetXMAErrorCallback( Xbox360_OpenAL_XMAError );
}

/* ------------------------------------------------------------------------- */

// JAT 18/10/07 - let the h/w platform specify this value
int Asura_OpenAL_Sound_API::Platform_GetNumAuxSendsPerSource( void ) 
{
	int iAuxSendsPerSource = 0;
	// Get the actual number of auxiliary sends available
	alcGetIntegerv( m_pxDevice, ALC_MAX_AUXILIARY_SENDS, 1, &iAuxSendsPerSource );
#ifdef ASURA_ASSERT
	ALenum xError = alGetError();
	Asura_Assertf( xError == AL_NO_ERROR, ( "Couldn't get max auxiliary sends, xError 0x%X", xError ) );
#endif // #ifdef ASURA_ASSERT

	#ifdef LIMIT_TO_ONE_AUXILIARY_SEND_PER_VOICE
		// JAT 18/10/07 - cross-fading reverb is causing hassles on 360 so for now disable it
		iAuxSendsPerSource = 1;
	#endif

	return iAuxSendsPerSource;
}

/* ------------------------------------------------------------------------- */

bool Asura_OpenAL_ActiveSound_API::Platform_FillBuffer( ALuint m_uBufferID, const void* pvFormat, const u_char* pucData, u_int uSize )
{
	ALenum xError = alGetError();
	ALenum xFormat = alGetEnumValue( "AL_FORMAT_XMA2" );
	// NOTE could support XMA1 and WAV as well - but everything should be XMA2

	if ( xFormat == AL_INVALID )
	{
		Asura_Assert( xFormat != AL_INVALID, "XMA2 not supported?!?" );
		return false;
	}

	bool bUsedStatic = false;

	// This check is used by alBufferDataDesc internally. It sometimes fails, and we therefore
	// need to call alBufferDataDesc in that case. (Easiest to leave the check in our code now
	// that we know about it). However - the check appears to be wrong. The first comparison
	// should be against 0x80000000 (an extra zero) and the other two should be against 2047
	// not 2043. Now awaiting confirmation (and a fix) from Creative about this, but seems
	// like a pretty reasonable and convincing explanation.
	if ( ( ( 0x8000000 & (ALuint)pucData ) &&
		( ( 2043 & (ALuint)pucData ) == 0 ) &&
		( ( 2043 & (ALuint)uSize ) == 0 ) ) ) 
	{
		bUsedStatic = true;
		alBufferDataDescStatic( m_uBufferID, xFormat, pvFormat, pucData, uSize );
	}
	else
	{
		alBufferDataDesc( m_uBufferID, xFormat, pvFormat, pucData, uSize );
	}

	if ( ( xError = alGetError() ) != AL_NO_ERROR )
	{
		Asura_Assertf( xError == AL_NO_ERROR, ( "Failed to buffer data using %s desc, error %x", 
			xError, bUsedStatic ? "static" : "non-static" ) );
	}

	return true;
}

/* ------------------------------------------------------------------------- */

static void Xbox360_OpenAL_XMAError( ALuint uStatus, ALuint uSource, ALuint uBuffer )
{
	Asura_Assertf( false, ( "XMA Error %x detected in Source %x, Buffer %x", uStatus, uSource, uBuffer ) );
}

/* ------------------------------------------------------------------------- */

static void* Xbox360_OpenAL_Alloc( u_int size )
{
	return Asura_Allocate( size );
}

/* ------------------------------------------------------------------------- */

static void Xbox360_OpenAL_Dealloc( void* ptr )
{
	Asura_Deallocate( ptr );
}

/* ------------------------------------------------------------------------- */
// Memory management functions, for use investigating OpenAL memory leaks.
// It seems that OpenAL is asking XAudio to create submix voices when the
// auxiliary send filters are set, and the memory for them is never freed.
// Memory sizes that leak: 312376 (occasional), 1464, 2488.
/* ------------------------------------------------------------------------- */

#if !defined( _DEBUG )

struct AudioAllocation
{
	LPVOID address;
	SIZE_T dwSize;
	u_int frameNumber;

	AudioAllocation()
	{
		Reset();
	}

	void Reset()
	{
		address = NULL;
		dwSize = 0;
		frameNumber = 0;
	}
};

static const int MaxAudioAllocations = 8192;
static AudioAllocation s_audioAllocations[MaxAudioAllocations];
static int NumAudioAllocs = 0;

/* ------------------------------------------------------------------------- */

LPVOID XMemAlloc( SIZE_T dwSize, DWORD dwAllocAttributes )
{
	LPVOID lpAddress = XMemAllocDefault( dwSize, dwAllocAttributes );

	DWORD dwAllocatorID = ( dwAllocAttributes >> 16 ) & 0xFF;

	if ( dwAllocatorID == eXALLOCAllocatorId_XAUDIO )
	{
		Asura_Assert( NumAudioAllocs < MaxAudioAllocations, "run out of room!" );

		for ( int i = 0; i < MaxAudioAllocations; ++i )
		{
			if ( !s_audioAllocations[i].address )
			{
				++NumAudioAllocs;
				s_audioAllocations[i].address = lpAddress;
				s_audioAllocations[i].dwSize = dwSize;
				s_audioAllocations[i].frameNumber = Asura_Timers::GetFrameNumber();
				break;
			}
		}
	}

	return lpAddress;
}

/* ------------------------------------------------------------------------- */

VOID XMemFree( PVOID pAddress, DWORD dwAllocAttributes )
{
	DWORD dwAllocatorID = ( dwAllocAttributes >> 16 ) & 0xFF;

	if ( dwAllocatorID == eXALLOCAllocatorId_XAUDIO )
	{
		for ( int i = 0; i < MaxAudioAllocations; ++i )
		{
			if ( s_audioAllocations[i].address == pAddress )
			{
				--NumAudioAllocs;
				s_audioAllocations[i].Reset();
				break;
			}
		}
	}

	XMemFreeDefault( pAddress, dwAllocAttributes );
}

/* ------------------------------------------------------------------------- */

SIZE_T XMemSize( PVOID pAddress, DWORD dwAllocAttributes )
{
	DWORD size = XMemSizeDefault( pAddress, dwAllocAttributes );

	DWORD dwAllocatorID = ( dwAllocAttributes >> 16 ) & 0xFF;

	if ( dwAllocatorID == eXALLOCAllocatorId_XAUDIO )
	{
		for ( int i = 0; i < MaxAudioAllocations; ++i )
		{
			if ( s_audioAllocations[i].address == pAddress )
			{
				Asura_Assert( size == s_audioAllocations[i].dwSize, "Size wrong!" );
			}
		}
	}

	return size;
}

#endif // !defined _DEBUG

/* ------------------------------------------------------------------------- */

#endif // defined( ASURA_USE_OPENAL ) && ( ASURA_USE_OPENAL == 1 )
