//////////////////////////////////////////////////////////////////////////////
//
// Asura_Platform_Corona_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "SpecialFX/Asura_Corona_System.h"

// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Rendering/Asura_Render.h"

// Asura Xbox 360
#include "Rendering/Asura_Xbox360_Effect_System.h"
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "Rendering/Asura_Xbox360_TextureManagement.h"
#include "Debugging/Asura_Platform_GPUProfile.h"

// Shader Include Files
#include "Corona.h"
// Misc
#include <fxl.h>
#include <xtl.h>

// DMR 11/07/07 - New SM3 corona renderer - with soft particle support and some batching

#include "Rendering/Asura_Viewport.h"

const int iASURA_MAX_BATCHABLE_CORONA_TEXTURES	   = 8;
const int iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE = 32;


struct CoronaVertex
{
	float X,Y,Z;
	float Diffuse[3];
	float U;
	float V;
	float U2;
	float V2;
};


class Asura_Xbox360_Corona_System
{
public:
	static void		Initialise();
	static void		Shutdown();
	static void		RenderStart();
	static void		RenderEnd();
	static void		RenderCorona(const Asura_Corona* pxCorona, u_int uCamVisFlags);	

private:
	static void		FlushBatch(int iTextureSlot);
	static float	CalcLightScale(const Asura_Corona *pxCorona, const Asura_Vector_3 *pxCameraPosition, const Asura_Vector_3 &xShiftForward, const Asura_Matrix_3x3 *pxCameraOrientation);

	static int				   s_aiBatchedTextures  [iASURA_MAX_BATCHABLE_CORONA_TEXTURES];
	static int				   s_aiNumBatchedCoronas[iASURA_MAX_BATCHABLE_CORONA_TEXTURES];
	static const Asura_Corona* s_apxBatchedCoronas  [iASURA_MAX_BATCHABLE_CORONA_TEXTURES][iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE];
	static uint32			   s_auBatchedVisFlags	[iASURA_MAX_BATCHABLE_CORONA_TEXTURES][iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE];
	static CoronaVertex		   s_axVertices			[iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE*4];

	static Asura_Xbox360_Effect*			   s_pxCoronaEffect;
	static Direct3D_VertexDeclaration*		   s_pxCoronaVertexDecl;

	static Asura_Xbox360_Effect_System::Handle s_xTextureHandle;
	static Asura_Xbox360_Effect_System::Handle s_xDepthTextureHandle;
};


int									Asura_Xbox360_Corona_System::s_aiBatchedTextures  [iASURA_MAX_BATCHABLE_CORONA_TEXTURES];
int									Asura_Xbox360_Corona_System::s_aiNumBatchedCoronas[iASURA_MAX_BATCHABLE_CORONA_TEXTURES];
const Asura_Corona*					Asura_Xbox360_Corona_System::s_apxBatchedCoronas  [iASURA_MAX_BATCHABLE_CORONA_TEXTURES][iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE];
uint32								Asura_Xbox360_Corona_System::s_auBatchedVisFlags  [iASURA_MAX_BATCHABLE_CORONA_TEXTURES][iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE];
CoronaVertex						Asura_Xbox360_Corona_System::s_axVertices		  [iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE*4];

Direct3D_VertexDeclaration*			Asura_Xbox360_Corona_System::s_pxCoronaVertexDecl				= 0;
Asura_Xbox360_Effect*				Asura_Xbox360_Corona_System::s_pxCoronaEffect					= 0;

Asura_Xbox360_Effect_System::Handle Asura_Xbox360_Corona_System::s_xTextureHandle					= 0;
Asura_Xbox360_Effect_System::Handle Asura_Xbox360_Corona_System::s_xDepthTextureHandle				= 0;


// All platform calls passed straight to Asura_Direct3D_CoronaSystem
void Asura_Corona_System::Platform_Initialise()														{ Asura_Xbox360_Corona_System::Initialise();							}
void Asura_Corona_System::Platform_Shutdown()														{ Asura_Xbox360_Corona_System::Shutdown();								}
void Asura_Corona_System::Platform_RenderStart(void)												{ Asura_Xbox360_Corona_System::RenderStart();							}
void Asura_Corona_System::Platform_RenderEnd(void)													{ Asura_Xbox360_Corona_System::RenderEnd();								}
void Asura_Corona_System::Platform_RenderCorona(const Asura_Corona* pxCorona, u_int uCamVisFlags)	{ Asura_Xbox360_Corona_System::RenderCorona(pxCorona, uCamVisFlags);    }



// ---------------------------------------------------------------------------------
// Initialise() - Clear the batches, init the shader
// ---------------------------------------------------------------------------------

void Asura_Xbox360_Corona_System::Initialise()
{
	for(int i=0; i<iASURA_MAX_BATCHABLE_CORONA_TEXTURES; i++)
	{
		s_aiNumBatchedCoronas[i] = 0;
		s_aiBatchedTextures[i] = -1;

		for(int j=0; j<iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE; j++)
		{
			s_apxBatchedCoronas[i][j]=0;
		}
	}

	for(int i=0; i<iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE; i++)
	{
		// UVs are constant
		s_axVertices[i*4+0].U = 0.0f;
		s_axVertices[i*4+0].V = 0.0f;
		s_axVertices[i*4+1].U = 1.0f;
		s_axVertices[i*4+1].V = 0.0f;
		s_axVertices[i*4+2].U = 1.0f;
		s_axVertices[i*4+2].V = 1.0f;
		s_axVertices[i*4+3].U = 0.0f;
		s_axVertices[i*4+3].V = 1.0f;
	}	

	Direct3D_VertexElement auVSD[] =
	{
		{	0,	0,	D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_POSITION,	0 },
		{	0, 12,	D3DDECLTYPE_FLOAT3,	D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_COLOR,		0 },
		{	0, 24,	D3DDECLTYPE_FLOAT4,	D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_TEXCOORD,	0 },
		D3DDECL_END()
	};

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( auVSD, &(s_pxCoronaVertexDecl) );

	s_pxCoronaEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auCorona);
	Asura_Assert( s_pxCoronaEffect, "Failed to load corona effect." );

	if (s_pxCoronaEffect)
	{	
		s_xTextureHandle				 = s_pxCoronaEffect->GetParameterHandle( "g_xTextureSampler" );
		s_xDepthTextureHandle			 = s_pxCoronaEffect->GetParameterHandle( "g_xDepthTextureSampler" );	
	}
}


// ---------------------------------------------------------------------------------
// Shutdown()
// ---------------------------------------------------------------------------------

void Asura_Xbox360_Corona_System::Shutdown()
{
	s_pxCoronaVertexDecl->Release();
	s_pxCoronaVertexDecl = NULL;

	Asura_Xbox360_Effect_System::Destroy( s_pxCoronaEffect);
	s_pxCoronaEffect = NULL;
}



// ---------------------------------------------------------------------------------
// RenderStart() - set up all constant renderstates/shader params
// ---------------------------------------------------------------------------------

void Asura_Xbox360_Corona_System::RenderStart()
{
	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xA0, 0xFF, 0x20 ), "Coronas" );
	ASURA_GPU_PROFILE_START(ASURA_GPUPROFILE_CORONAS);

	// Setup initial renderstates
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);	

	// This double call is intentional. Cached states are shafted again, this is the workaround
	Asura_Render::CurrentStates.RequestZBufferEnabled(false);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);

	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
	Asura_Render::CurrentStates.RequestAlphaTest(false);

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline(D3DRS_HIGHPRECISIONBLENDENABLE,
															   TRUE);

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration(s_pxCoronaVertexDecl);	
}



// ---------------------------------------------------------------------------------
// RenderEnd() - Flush all batches, then restore all renderstates
// ---------------------------------------------------------------------------------

void Asura_Xbox360_Corona_System::RenderEnd()
{
	for(int i=0; i<iASURA_MAX_BATCHABLE_CORONA_TEXTURES; i++)
	{	
		FlushBatch(i);
	}

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline(D3DRS_HIGHPRECISIONBLENDENABLE,
															   FALSE);

	s_pxCoronaEffect->SetTexture(s_xDepthTextureHandle, ASURA_TEXTURE_NONE);

	PIXEndNamedEvent();
	ASURA_GPU_PROFILE_END(ASURA_GPUPROFILE_CORONAS);
}


// ---------------------------------------------------------------------------------
// RenderCorona() - This queues the corona for batched rendering - but will
//					flush a batch if there isn't space to add the new corona
// ---------------------------------------------------------------------------------

void Asura_Xbox360_Corona_System::RenderCorona(const Asura_Corona* pxCorona, u_int uCamVisFlags)
{	

	if (pxCorona->m_iTextureIndex<0)
	{
		// Invalid texture
		return;
	}

	// Do we have a batch for this texture?
	int iTextureSlot = -1;
	for(int i=0; i<iASURA_MAX_BATCHABLE_CORONA_TEXTURES; i++)
	{							
		if (s_aiBatchedTextures[i] == pxCorona->m_iTextureIndex)
		{
			iTextureSlot = i;
		}
	}

	if (iTextureSlot>=0)
	{
		// If we've got a batch, but it's full, we need to flush it
		if (s_aiNumBatchedCoronas[iTextureSlot]>=iASURA_MAX_BATCHABLE_CORONAS_PER_TEXTURE)
		{
			FlushBatch(iTextureSlot);
		}		
	}
	else
	{
		// No - Do we have an empty slot for a new texture batch?		
		int iBestToFlush  = 0;		
		int iBiggestBatch = 0;		

		for(int i=0; i<iASURA_MAX_BATCHABLE_CORONA_TEXTURES; i++)
		{					
			// Find the first empty slot, if we have one
			if (s_aiBatchedTextures[i]<0)
			{
				if (iTextureSlot<0)
				{
					iTextureSlot = i;
				}				
			}

			// Update best candidate to flush (the one with the biggest batch already)
			if (s_aiNumBatchedCoronas[i]>iBiggestBatch)
			{
				iBiggestBatch = s_aiNumBatchedCoronas[i];
				iBestToFlush = i;
			}
		}

		if (iTextureSlot<0)
		{
			FlushBatch(iBestToFlush);
			iTextureSlot = iBestToFlush;
		}
	}

	s_aiBatchedTextures  [iTextureSlot] = pxCorona->m_iTextureIndex;	
	s_apxBatchedCoronas  [iTextureSlot][s_aiNumBatchedCoronas[iTextureSlot]] = pxCorona;	
	s_auBatchedVisFlags  [iTextureSlot][s_aiNumBatchedCoronas[iTextureSlot]] = uCamVisFlags;
	s_aiNumBatchedCoronas[iTextureSlot]++;
}




// ---------------------------------------------------------------------------------
// CalcLightScale() - Calculate the intensity of the corona - handle the distance fade
//                    and direction effect
// ---------------------------------------------------------------------------------

float Asura_Xbox360_Corona_System::CalcLightScale(const Asura_Corona *pxCorona, const Asura_Vector_3 *pxCameraPosition,
												   const Asura_Vector_3 &xShiftForward, const Asura_Matrix_3x3 *pxCameraOrientation)
{
	float fLightScale = 255.0f; 

	Asura_Vector_3 xDelta =  *pxCameraPosition - (pxCorona->m_xPosition + xShiftForward);
	float fDistance = xDelta.Magnitude();

	if(pxCorona->m_bUseDirectionEffect)
	{
		if (fDistance>0.0f)
		{
			fLightScale = (pxCorona->m_xDirection * xDelta) / fDistance;

			if (fLightScale<0)
			{
				fLightScale = 64.0f;
			}
			else
			{
				fLightScale = 191.0f*fLightScale+64.0f;
			}
		}
		else
		{
			fLightScale = 0.0f;
		}
	}

	if( !pxCorona->m_bDoNotFadeOutWithDistance )
	{
		float fDistanceScale = fDistance - ASURA_CORONA_FADE_DIST_BEGIN;

		if (fDistanceScale > 0.0f)
		{
			fDistanceScale *= ( 1.0f / (ASURA_CORONA_FADE_DIST_END - ASURA_CORONA_FADE_DIST_BEGIN) );

			fLightScale *= (1.0f - fDistanceScale);
		}
	}

	const float fNearFadeDistance = pxCorona->GetNearFadeDistance();
	if(fNearFadeDistance)
	{
		Asura_Vector_3 xZDir = Asura_Vector_3(pxCameraOrientation->e13,pxCameraOrientation->e23,pxCameraOrientation->e33);
		float fZDistance = -(xDelta*xZDir);

		if(fZDistance<fNearFadeDistance)
		{
			fLightScale *= fZDistance/fNearFadeDistance;
			if(fLightScale<0) fLightScale = 0;
		}
	}

	return fLightScale;
}


// ---------------------------------------------------------------------------------
// FlushBatch() - Render a batch of coronas that share the same texture
// ---------------------------------------------------------------------------------

void Asura_Xbox360_Corona_System::FlushBatch(int iTextureSlot)
{
	// Have we got something to render?
	if (s_aiBatchedTextures[iTextureSlot]>=0 && s_aiNumBatchedCoronas[iTextureSlot]>0)
	{
		for(int iCamera=0; iCamera<Asura_Camera::GetNumberOfCameras(); iCamera++)
		{			
			Asura_Camera::SetCameraIndex(iCamera);
			Asura_Render::ApplyCameraSettings();

			const Asura_Vector_3*	pxCameraPosition = Asura_Camera::GetPosition();
			const Asura_Matrix_3x3* pxCameraOrientation = Asura_Camera::GetOrientation();
			Asura_Vector_3 xDeltaX = Asura_Vector_3(pxCameraOrientation->e11,pxCameraOrientation->e21,pxCameraOrientation->e31);
			Asura_Vector_3 xDeltaY = Asura_Vector_3(pxCameraOrientation->e12,pxCameraOrientation->e22,pxCameraOrientation->e32);

			// 2 passes through the coronas - first the soft ones, then the non-soft ones?
			for(int iPass=0; iPass<2; iPass++)
			{
				int iNumCoronas = 0;
				int iNumBatched = s_aiNumBatchedCoronas[iTextureSlot];

				CoronaVertex *pxVtx = &s_axVertices[0];

				for(int i=0; i<iNumBatched; i++)
				{
					const Asura_Corona *pxCorona = s_apxBatchedCoronas[iTextureSlot][i];
					
					if (iPass==0 && pxCorona->m_fVolumetricDepthScale<=0.0f || 
						iPass==1 && pxCorona->m_fVolumetricDepthScale>0.0f)
					{
						if((s_auBatchedVisFlags[iTextureSlot][i] & (1<<iCamera))) 
						{
							Asura_Vector_3 xForward = pxCorona->m_xPosition - *pxCameraPosition;
							xForward.Normalise();
							Asura_Vector_3 xShiftForward =- ( xForward * pxCorona->m_fRadius * pxCorona->m_fVolumetricDepthScale );

							// Calculate the colour
							float fLightScale = CalcLightScale(pxCorona, pxCameraPosition, xShiftForward, pxCameraOrientation);

							if (fLightScale>0.0f)
							{							
								// Calculate the billboarded vert positions
								float fSize = pxCorona->m_fRadius;
								float afDeltaXScale[4] = {-fSize, fSize, fSize,-fSize};
								float afDeltaYScale[4] = {-fSize,-fSize, fSize, fSize};

								for(int i=0; i<4; i++)
								{
									pxVtx->X = pxCorona->m_xPosition.x + (afDeltaXScale[i] * xDeltaX.x) + (afDeltaYScale[i] * xDeltaY.x) + xShiftForward.x;
									pxVtx->Y = pxCorona->m_xPosition.y + (afDeltaXScale[i] * xDeltaX.y) + (afDeltaYScale[i] * xDeltaY.y) + xShiftForward.y;
									pxVtx->Z = pxCorona->m_xPosition.z + (afDeltaXScale[i] * xDeltaX.z) + (afDeltaYScale[i] * xDeltaY.z) + xShiftForward.z;
									pxVtx->Diffuse[0] = fLightScale * pxCorona->m_fRed / 255.0f;
									pxVtx->Diffuse[1] = fLightScale * pxCorona->m_fGreen / 255.0f;
									pxVtx->Diffuse[2] = fLightScale * pxCorona->m_fBlue / 255.0f;
									pxVtx->U2 = fSize * pxCorona->m_fVolumetricDepthScale * 2.0f;
									pxVtx++;
								}

								// Add the corona to the VB
								iNumCoronas++;
							}
						}						
					}				
				}

				// Render 'em now
				if (iNumCoronas>0)
				{
					s_pxCoronaEffect->Begin(0, iPass);

					// Set the texture
					s_pxCoronaEffect->SetTexture( s_xTextureHandle,s_aiBatchedTextures[iTextureSlot]);					
					
					s_pxCoronaEffect->GetInterface()->SetSampler(s_xDepthTextureHandle, Asura_Xbox360_Graphics::GetDepthMtlTexture());	

					for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
					{
						if (Asura_Xbox360_Graphics::IsTiling2())
						{
							Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
							Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
						}

						s_pxCoronaEffect->Commit();
						Asura_Xbox360_Graphics::GetDevice()->DrawPrimitiveUP(D3DPT_QUADLIST, iNumCoronas, (LPVOID)s_axVertices, sizeof(CoronaVertex));
					}
					if (Asura_Xbox360_Graphics::IsTiling2())
					{
						Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
						Asura_Xbox360_Graphics::CorrectVPOSForTiling(0);
					}

					s_pxCoronaEffect->End();
				}
			}
		}
	}

	s_aiBatchedTextures[iTextureSlot]   = -1;
	s_aiNumBatchedCoronas[iTextureSlot] = 0;
}




