#include <revolution.h>

#include <stdlib.h>
#include <math.h>

#include "binktextures.h"


RADDEFFUNC void * radmalloc( U32 );
RADDEFFUNC void radfree( void * );

//############################################################################
//##                                                                        ##
//## Free the textures that we allocated.                                   ##
//##                                                                        ##
//############################################################################

void Free_Bink_textures( BINKTEXTURESET * set_textures )
{
  if ( set_textures->base_ptr )
  {
    radfree( set_textures->base_ptr );
   
    set_textures->base_ptr = 0;
  }
}


static U32 nextpow2( U32 x )
{
  return( 1 << ( 32 - __cntlzw( x ) ) );
}


static void  fill_offset_buffer( U8 * buffer, U32 sourcewidth, U32 offwidth )
{
   U32 finalwidth = sourcewidth / 8;

   S32 srcx, srcy, x, y;

   srcx = 0;
   srcy = 0;

   for ( y = 0 ; y < 4 ; ++y )
   {
     for ( x = 0 ; x < finalwidth ; ++x )
     {
       U32 texoffset = ( y * offwidth + x ) * 2;
       S32 xoffset = srcx - x;
       S32 yoffset = srcy - y;

       buffer[ texoffset ] = (U8) ( xoffset + 128 );
       buffer[ texoffset + 1 ] = (U8) ( yoffset + 128 );
      
       if ( ++srcy == 4 )
       {
         srcy = 0;
         ++srcx;
       }
     }
   }
}


static void swizzle_offsets( U8 * swizz, U8 const * offsets, U32 sourcewidth, U32 sourceheight, U32 bytesperpixel )
{
   // Only works (tested, anyways) for 8 and 16 bit formats for the moment.
   // Texture tiling format is specified in the GX programmer's
   // guide, Appendix D, as of this writing.

   // Not super efficient, but it doesn't really have to be.

   U32 tilex, tiley;
   U32 pitch = sourcewidth * bytesperpixel;
   U32 tilesw = pitch / 8;
   U32 tilesh   = sourceheight / 4;
   U32 swizzoffset = 0;
  
   for ( tiley = 0 ; tiley < tilesh ; ++tiley )
   {
     for ( tilex = 0 ; tilex < tilesw; ++tilex )
     {
       U32 row;
       U32 x = tilex * 8;
       U32 y = tiley * 4;
       U32 sourceoffset = y * pitch + x;

       for ( row = 0; row < 4; ++row )
       {
         U32 * dest = (U32*) ( swizz + swizzoffset + ( row * 8 ) );
         U32 * src  = (U32*) ( offsets + sourceoffset + ( row * pitch ) );

         dest[ 0 ] = src[ 0 ];
         dest[ 1 ] = src[ 1 ];
       }
       swizzoffset += 32;
     }
   }
}


//############################################################################
//##                                                                        ##
//## Create a texture while allocating the memory ourselves.                ##
//##                                                                        ##
//############################################################################

static S32 make_texture( U32 pitch, U32 height,
                         GXTexFmt format,
                         GXTexWrapMode type,
                         GXTexObj * texture, void * ptr )
{
  GXInitTexObj( texture,
                ptr,
                (U16) pitch, (U16) height,
                format,
                type, type,
                GX_FALSE);
  
  GXInitTexObjLOD( texture,
                   GX_NEAR, GX_NEAR,
                   0, 0, 0.0f,
                   GX_FALSE, 0, GX_ANISO_1);
  return( 1 );
}


static U32 divisible128( U32 val )
{
  return( ( val + 127 ) & ~127 );
}


static U32 swizz_div( U32 sourcewidth )
{
  U32 offwidth;
  
  // The width of the offset texture is the source/8, but it must be
  // padded out to both the next power of 2.  (Also the next
  // multiple of 4, for very small textures.)

  offwidth = ( ( sourcewidth / 8 ) + 3 ) & ~3;
  
  if ( offwidth & ( offwidth - 1 ) )
  {
    // must be a power of 2
    offwidth = nextpow2( offwidth );
  }

  return( offwidth );
}

// create a texture object for the inverse swizzle pattern
static U32 make_swizzler( void * data, GXTexObj* obj,
                          U32 * out_swidth, U32 * out_sheight,
                          U32 sourcewidth )
{
  U32 offwidth, offsize;
  U8 buffer[ 256 * 4 * 2 ];

  offwidth = swizz_div( sourcewidth );
  
  *out_swidth = offwidth;
  *out_sheight = 4;

  offsize = divisible128( offwidth * 2 * 4 );

  // Fill the buffer, and swizzle the result
  fill_offset_buffer( buffer, sourcewidth, offwidth );
  swizzle_offsets( (U8*) data, buffer, offwidth, 4, 2 );

  DCFlushRange( data, offsize );

  make_texture( offwidth, 4, GX_TF_IA8, GX_REPEAT, obj, data );

  return( offsize );
}


// make the two deswizzler textures (which are only around 512 bytes long)
static void do_deswizzlers( BINKTEXTURESET * set_textures, U8 * data )
{
  data += 
    make_swizzler( data, &set_textures->YAdeswizzle, 
                         &set_textures->YAdeswizzle_width, &set_textures->YAdeswizzle_height,
                         set_textures->bink_buffers.YABufferWidth );
  data += 
    make_swizzler( data, &set_textures->cRcBdeswizzle, 
                         &set_textures->cRcBdeswizzle_width, &set_textures->cRcBdeswizzle_height,
                         set_textures->bink_buffers.cRcBBufferWidth );
}


static void calc_layout( BINKFRAMEBUFFERS * buffers, U32 * cRoffset, U32 * cBoffset, U32 * Aoffset, 
                                                     U32 * framesize, U32 * swizzlesize, U32 * YApitch, U32 * cRcBpitch )
{
  U32 ywidth  = ( buffers->Frames[ 0 ].YPlane.Allocate ) ? buffers->YABufferWidth : 0;
  U32 awidth = ( buffers->Frames[ 0 ].APlane.Allocate ) ? buffers->YABufferWidth : 0;
  U32 crcbwidth = ( buffers->Frames[ 0 ].cRPlane.Allocate ) ? buffers->cRcBBufferWidth : 0;
  U32 yaheight = buffers->YABufferHeight;
  U32 crcbheight = buffers->cRcBBufferHeight;

  U32 ysize, asize, crcbsize;
               
  ysize = divisible128( ywidth * yaheight );
  asize = divisible128( awidth * yaheight );
  crcbsize = divisible128( crcbwidth * crcbheight );

  *swizzlesize = divisible128( swizz_div( buffers->YABufferWidth ) * 2 * 4 ) +
                 divisible128( swizz_div( buffers->cRcBBufferWidth ) * 2 * 4 );
  
  *Aoffset = ysize;
  *cRoffset = ysize + asize;
  *cBoffset = ysize + asize + crcbsize;
  *framesize = ysize + asize + 2 * crcbsize;
  *YApitch = ywidth | awidth; // they will be the same, or one will be zero
  *cRcBpitch = crcbwidth;
}


//############################################################################
//##                                                                        ##
//## Create 2 sets of textures for Bink to decompress into...               ##
//##                                                                        ##
//############################################################################

S32 Create_Bink_textures( BINKTEXTURESET * set_textures )
{
  int i;

  U32 cRoffset;
  U32 cBoffset;
  U32 Aoffset;
  U32 framesize;
  U32 swizzlesize;
  U32 YApitch;
  U32 cRcBpitch;
  U8* data;

  calc_layout( &set_textures->bink_buffers,
               &cRoffset,
               &cBoffset,
               &Aoffset,
               &framesize,
               &swizzlesize,
               &YApitch,
               &cRcBpitch) ;

  // videos have to be smaller than 1024 for the dependent texture read to work
  if ( YApitch > 1024 )
    return( 0 );

  //
  // Allocate all texture memory at once
  //

  data = (U8*) radmalloc( framesize * set_textures->bink_buffers.TotalFrames + swizzlesize + 127 );
  if ( data == 0 )
    goto fail;

  set_textures->base_ptr = data;

  // align the start of the block
  data = (U8*) ( ( ( (U32) data ) + 127 ) & ~127 );

  //
  // Create our textures
  //

  for( i = 0 ; i < set_textures->bink_buffers.TotalFrames ; i++ )
  {
    if ( set_textures->bink_buffers.Frames[ i ].YPlane.Allocate )
    {
      set_textures->bink_buffers.Frames[ i ].YPlane.Buffer = data;
      set_textures->bink_buffers.Frames[ i ].YPlane.BufferPitch = YApitch;
      
      // create Y plane
      if ( ! make_texture( YApitch, 
                           set_textures->bink_buffers.YABufferHeight,
                           GX_TF_I8,
                           GX_CLAMP,                           
                           &set_textures->textures[ i ].Ytexture, 
                           data ) )
        goto fail;
    }
                                              
    if ( set_textures->bink_buffers.Frames[ i ].cRPlane.Allocate )
    {
      set_textures->bink_buffers.Frames[ i ].cRPlane.Buffer = data + cRoffset;
      set_textures->bink_buffers.Frames[ i ].cRPlane.BufferPitch = cRcBpitch;
      
      // create cR plane
      if ( ! make_texture( cRcBpitch, 
                           set_textures->bink_buffers.cRcBBufferHeight,
                           GX_TF_I8,
                           GX_CLAMP,                           
                           &set_textures->textures[ i ].cRtexture, 
                           data + cRoffset ) )
        goto fail;
    }

    if ( set_textures->bink_buffers.Frames[ i ].cBPlane.Allocate )
    {
      set_textures->bink_buffers.Frames[ i ].cBPlane.Buffer = data + cBoffset;
      set_textures->bink_buffers.Frames[ i ].cBPlane.BufferPitch = cRcBpitch;

      // create cB plane
      if ( ! make_texture( cRcBpitch, 
                           set_textures->bink_buffers.cRcBBufferHeight,
                           GX_TF_I8,
                           GX_CLAMP,                           
                           &set_textures->textures[ i ].cBtexture, 
                           data + cBoffset ) )
        goto fail;
    }
    
    if ( set_textures->bink_buffers.Frames[ i ].APlane.Allocate )
    {
      set_textures->bink_buffers.Frames[ i ].APlane.Buffer = data + Aoffset;
      set_textures->bink_buffers.Frames[ i ].APlane.BufferPitch = YApitch;
      
      // create alpha plane
      if ( ! make_texture( YApitch, 
                           set_textures->bink_buffers.YABufferHeight,
                           GX_TF_I8,
                           GX_CLAMP,                           
                           &set_textures->textures[ i ].Atexture, 
                           data + Aoffset ) )
        goto fail;
    }  

    data += framesize;
  }

  set_textures->framesize = framesize;
  do_deswizzlers( set_textures, data );
 
  return( 1 );
  
 fail:

  Free_Bink_textures( set_textures );
  return( 0 );
}


// Constants
static const GXColorS10 reg0color = { -102, 0, -128, 154 };
static const GXColor color0 = {   0,   0, 255, 100 };
static const GXColor color1 = { 204,   0,   0, 208 };
static const GXColor color2 = { 255,   0, 255, 128 };

static const F32 indMtx[ 2 ][ 3 ] = {
  8.0f / 16.0f, 0.0f, 0.0f,
  0.0f, 1.0f / 16.0f, 0.0f
};

static const F32 ATTRIBUTE_ALIGN(32) id_texmtx[ 2 ][ 4 ] = 
{
  1.0f, 0.0f, 0.0f, 0.0f,
  0.0f, 1.0f, 0.0f, 0.0f
};

static F32 ATTRIBUTE_ALIGN(32) yswiz_texmtx[ 2 ][ 4 ] = 
{
  0.0f, 0.0f, 0.0f, 0.0f,
  0.0f, 0.0f, 0.0f, 0.0f
};

static F32 ATTRIBUTE_ALIGN(32) crcbswiz_texmtx[ 2 ][ 4 ] = {
  0.0f, 0.0f, 0.0f, 0.0f,
  0.0f, 0.0f, 0.0f, 0.0f
};


//############################################################################
//##                                                                        ##
//## Draw our textures onto the screen with our vertex and pixel shaders.   ##
//##                                                                        ##
//############################################################################

void Draw_Bink_textures( BINKTEXTURESET * set_textures,
                         U32 width,
                         U32 height,
                         F32 x_offset,
                         F32 y_offset,
                         F32 x_scale,
                         F32 y_scale,
                         F32 alpha_level,
                         S32 is_premultiplied_alpha )
{
  int doalpha;
  U8 alv;
  U8 alvc;
  
  if ( alpha_level >= 0.99f )
    alv = 255;
  else if ( alpha_level <= 0.01f )
    alv = 0;
  else
    alv = (alpha_level * 255.0f + 0.5f);
  alvc = ( is_premultiplied_alpha ) ? alv : 255;
    
  doalpha = set_textures->bink_buffers.Frames[ 0 ].APlane.Buffer ? 1 : 0;

  // mark these textures as being in use and drop a sentinal in the GPU buffer
  //   that we can see to know when we are in use
    
  set_textures->drawing[ set_textures->bink_buffers.FrameNum ] = 1;
  GXSetDrawSync( 0x4444 + set_textures->bink_buffers.FrameNum );

  // Position has 3 elements (x,y,z), each of type f32
  GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS, GX_POS_XYZ, GX_S16, 0);
  GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGBA8, 0);
  GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX0, GX_TEX_ST, GX_F32, 0);
  GXSetVtxAttrFmt(GX_VTXFMT0, GX_VA_TEX1, GX_TEX_ST, GX_F32, 0);

  GXClearVtxDesc();
  GXLoadTexObj( &set_textures->textures[ set_textures->bink_buffers.FrameNum ].Ytexture,  GX_TEXMAP0 );
  GXLoadTexObj( &set_textures->YAdeswizzle,  GX_TEXMAP1 );
  GXLoadTexObj( &set_textures->cRcBdeswizzle, GX_TEXMAP2 );
  GXLoadTexObj( &set_textures->textures[ set_textures->bink_buffers.FrameNum ].cRtexture,  GX_TEXMAP3 );
  GXLoadTexObj( &set_textures->textures[ set_textures->bink_buffers.FrameNum ].cBtexture,  GX_TEXMAP4 );

  GXSetVtxDesc( GX_VA_POS,  GX_DIRECT );
  GXSetVtxDesc( GX_VA_CLR0, GX_DIRECT );  // pass in alpha val */
  GXSetVtxDesc( GX_VA_TEX0, GX_DIRECT );  // Note that we pass the same texcoord in twice, to
  GXSetVtxDesc( GX_VA_TEX1, GX_DIRECT );  // get around the automatic scaling that takes place

  // If there is alpha present
  if ( doalpha )
    GXLoadTexObj( &set_textures->textures[ set_textures->bink_buffers.FrameNum ].Atexture,  GX_TEXMAP5);

  GXSetNumChans(1);
  GXSetNumTexGens( 4 );
  GXSetNumTevStages( 4 );
  GXSetNumIndStages( 3 );

  GXSetIndTexOrder( GX_INDTEXSTAGE0, GX_TEXCOORD1, GX_TEXMAP1 );  // y swiz
  GXSetIndTexOrder( GX_INDTEXSTAGE1, GX_TEXCOORD2, GX_TEXMAP2 );  // uv swiz

  // No alpha blending
  if ( ( doalpha ) || ( alv != 255) )
  {
    if ( is_premultiplied_alpha )
      GXSetBlendMode( GX_BM_BLEND,
                      GX_BL_ONE,
                      GX_BL_INVSRCALPHA,
                      GX_LO_CLEAR );
    else
      GXSetBlendMode( GX_BM_BLEND,
                      GX_BL_SRCALPHA,
                      GX_BL_INVSRCALPHA,
                      GX_LO_CLEAR );
  }
  else
  {
    GXSetBlendMode( GX_BM_BLEND,
                    GX_BL_ONE,
                    GX_BL_ZERO,
                    GX_LO_CLEAR );
  }
  
  
  // First stage: cB
  {
    GXSetTevIndirect( GX_TEVSTAGE0,
                      GX_INDTEXSTAGE1,
                      GX_ITF_8,
                      GX_ITB_STU,
                      GX_ITM_0,
                      GX_ITW_OFF, GX_ITW_OFF,
                      GX_FALSE, GX_FALSE,
                      GX_ITBA_OFF );
    GXSetTevOrder( GX_TEVSTAGE0, GX_TEXCOORD0, GX_TEXMAP4, GX_COLOR_NULL );

    GXSetTevColorIn( GX_TEVSTAGE0, GX_CC_ZERO, GX_CC_TEXC,
                     GX_CC_KONST, GX_CC_C0 );
    GXSetTevColorOp( GX_TEVSTAGE0, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_FALSE, GX_TEVPREV );
    GXSetTevAlphaIn( GX_TEVSTAGE0, GX_CA_ZERO, GX_CA_TEXA,
                     GX_CA_KONST, GX_CA_A0 );
    GXSetTevAlphaOp( GX_TEVSTAGE0, GX_TEV_SUB, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_FALSE, GX_TEVPREV );
    GXSetTevKColorSel( GX_TEVSTAGE0, GX_TEV_KCSEL_K0 );
    GXSetTevKAlphaSel( GX_TEVSTAGE0, GX_TEV_KASEL_K0_A );
    GXSetTevSwapMode( GX_TEVSTAGE0, GX_TEV_SWAP0, GX_TEV_SWAP0 );
  }

  // Stage 2: cR
  {
    // cR
    GXSetTevIndirect( GX_TEVSTAGE1,
                      GX_INDTEXSTAGE1,
                      GX_ITF_8,
                      GX_ITB_STU,
                      GX_ITM_0,
                      GX_ITW_OFF, GX_ITW_OFF,
                      GX_FALSE, GX_FALSE,
                      GX_ITBA_OFF );
    GXSetTevOrder( GX_TEVSTAGE1, GX_TEXCOORD0, GX_TEXMAP3, GX_COLOR_NULL);

    GXSetTevColorIn( GX_TEVSTAGE1, GX_CC_ZERO, GX_CC_TEXC,
                     GX_CC_KONST, GX_CC_CPREV );
    GXSetTevColorOp( GX_TEVSTAGE1, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_2, GX_FALSE, GX_TEVPREV );
    GXSetTevAlphaIn( GX_TEVSTAGE1, GX_CA_ZERO, GX_CA_TEXA,
                     GX_CA_KONST, GX_CA_APREV );
    GXSetTevAlphaOp( GX_TEVSTAGE1, GX_TEV_SUB, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_FALSE, GX_TEVPREV );
    GXSetTevKColorSel( GX_TEVSTAGE1, GX_TEV_KCSEL_K1 );
    GXSetTevKAlphaSel( GX_TEVSTAGE1, GX_TEV_KASEL_K1_A );
    GXSetTevSwapMode( GX_TEVSTAGE1, GX_TEV_SWAP0, GX_TEV_SWAP0 );
  }

  // Stage 3: Y
  {
    // y
    GXSetTevIndirect( GX_TEVSTAGE2,
                      GX_INDTEXSTAGE0,
                      GX_ITF_8,
                      GX_ITB_STU,
                      GX_ITM_0,
                      GX_ITW_OFF, GX_ITW_OFF,
                      GX_FALSE, GX_FALSE,
                      GX_ITBA_OFF );
    GXSetTevOrder( GX_TEVSTAGE2, GX_TEXCOORD3, GX_TEXMAP0, GX_COLOR_NULL );

    GXSetTevColorIn( GX_TEVSTAGE2, GX_CC_ZERO, GX_CC_TEXC,
                     GX_CC_ONE, GX_CC_CPREV );
    GXSetTevColorOp( GX_TEVSTAGE2, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV );
    GXSetTevAlphaIn( GX_TEVSTAGE2, GX_CA_TEXA, GX_CA_ZERO,
                     GX_CA_ZERO, GX_CA_APREV );
    GXSetTevAlphaOp( GX_TEVSTAGE2, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV );
    GXSetTevSwapMode( GX_TEVSTAGE2, GX_TEV_SWAP0, GX_TEV_SWAP0 );
  }

  if ( doalpha )
  {
    GXSetTevIndirect( GX_TEVSTAGE3,
                      GX_INDTEXSTAGE0,
                      GX_ITF_8,
                      GX_ITB_STU,
                      GX_ITM_0,
                      GX_ITW_OFF, GX_ITW_OFF,
                      GX_FALSE, GX_FALSE,
                      GX_ITBA_OFF );
    GXSetTevOrder( GX_TEVSTAGE3, GX_TEXCOORD3, GX_TEXMAP5, GX_COLOR0A0 );

    GXSetTevColorIn( GX_TEVSTAGE3, GX_CC_APREV, GX_CC_CPREV,
                     GX_CC_KONST, GX_CC_ZERO );
    GXSetTevColorOp( GX_TEVSTAGE3, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV );
    GXSetTevAlphaIn( GX_TEVSTAGE3, GX_CA_ZERO, GX_CA_TEXA,
                     GX_CA_RASA, GX_CA_ZERO );
    GXSetTevAlphaOp( GX_TEVSTAGE3, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV );
    GXSetTevSwapMode( GX_TEVSTAGE3, GX_TEV_SWAP0, GX_TEV_SWAP0 );
    GXSetTevKColorSel( GX_TEVSTAGE3, GX_TEV_KCSEL_K2 );
  }
  else
  {
    GXSetTevOrder( GX_TEVSTAGE3, GX_TEXCOORD_NULL, GX_TEXMAP_NULL, GX_COLOR0A0 );
    GXSetTevColorIn( GX_TEVSTAGE3, GX_CC_APREV, GX_CC_CPREV,
                     GX_CC_KONST, GX_CC_ZERO );
    GXSetTevColorOp( GX_TEVSTAGE3, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV );
    GXSetTevAlphaIn( GX_TEVSTAGE3, GX_CA_ZERO, GX_CA_ONE,
                     GX_CA_RASA, GX_CA_ZERO );
    GXSetTevAlphaOp( GX_TEVSTAGE3, GX_TEV_ADD, GX_TB_ZERO,
                     GX_CS_SCALE_1, GX_TRUE, GX_TEVPREV );
    GXSetTevSwapMode( GX_TEVSTAGE3, GX_TEV_SWAP0, GX_TEV_SWAP0 );
    GXSetTevKColorSel( GX_TEVSTAGE3, GX_TEV_KCSEL_K2 );
  }

  GXSetTevColorS10( GX_TEVREG0, reg0color );
  GXSetTevKColor( GX_KCOLOR0, color0 );
  GXSetTevKColor( GX_KCOLOR1, color1 );
  GXSetTevKColor( GX_KCOLOR2, color2 );

  // Swap mode tables
  GXSetTevSwapModeTable( GX_TEV_SWAP0, GX_CH_RED,   GX_CH_GREEN,
                         GX_CH_BLUE,  GX_CH_ALPHA ); // RGBA

  // Setup the indirect matrix
  // You can't actually store 1.0 in this matrix, so we use the
  // bias exponent to get 8 and 1 in the locations where we need
  // them
    
  GXSetIndTexMtx( GX_ITM_0, indMtx, 4 );
  

  // Setup the texgens...
  {
    // adjust for the pow2 expansion
    F32 yaswizzletexU = ( (F32) ( set_textures->bink_buffers.YABufferWidth / 8 ) ) / ( (F32) set_textures->YAdeswizzle_width );
    F32 yaSwizzletexV = ( (F32) set_textures->bink_buffers.YABufferHeight ) / ( (F32) set_textures->YAdeswizzle_height );

    F32 crcbswizzletexU = ( (F32) ( set_textures->bink_buffers.cRcBBufferWidth / 8 ) ) / ( (F32) set_textures->cRcBdeswizzle_width );
    F32 crcbswizzletexV = ( (F32) set_textures->bink_buffers.cRcBBufferHeight ) / ( (F32) set_textures->cRcBdeswizzle_height );

    yswiz_texmtx[ 0 ][ 0 ] = yaswizzletexU;
    yswiz_texmtx[ 1 ][ 1 ] = yaSwizzletexV;
    crcbswiz_texmtx[ 0 ][ 0 ] = crcbswizzletexU;
    crcbswiz_texmtx[ 1 ][ 1 ] = crcbswizzletexV;

    GXSetTexCoordGen( GX_TEXCOORD0, GX_TG_MTX2x4, GX_TG_TEX0, GX_TEXMTX0 );
    GXSetTexCoordGen( GX_TEXCOORD1, GX_TG_MTX2x4, GX_TG_TEX0, GX_TEXMTX1 );
    GXSetTexCoordGen( GX_TEXCOORD2, GX_TG_MTX2x4, GX_TG_TEX0, GX_TEXMTX2 );
    GXSetTexCoordGen( GX_TEXCOORD3, GX_TG_MTX2x4, GX_TG_TEX1, GX_TEXMTX3 );

    GXLoadTexMtxImm( id_texmtx, GX_TEXMTX0, GX_MTX2x4 );
    GXLoadTexMtxImm( yswiz_texmtx, GX_TEXMTX1, GX_MTX2x4 );
    GXLoadTexMtxImm( crcbswiz_texmtx, GX_TEXMTX2, GX_MTX2x4 );
    GXLoadTexMtxImm( id_texmtx, GX_TEXMTX3, GX_MTX2x4 );
  }

  #define NUM_BITS 5
  #define SCALE_FACTOR (1 << NUM_BITS)

  {
    F32 w_factor, h_factor;
    S16 nw = (S16) ( ( (F32) width ) * x_scale );
    S16 nh = (S16) ( ( (F32) height ) * y_scale );

    S16 x1 = (S16) x_offset;
    S16 y1 = (S16) y_offset;
    S16 x2 = x1 + nw;
    S16 y2 = y1 + nh;

    w_factor = (F32) ( set_textures->bink_buffers.YABufferWidth * SCALE_FACTOR ) / (F32) nw;
    h_factor = (F32) ( set_textures->bink_buffers.YABufferHeight * SCALE_FACTOR ) / (F32) nh;

    if (nw != 0 && nh != 0)
    {

      GXBegin( GX_QUADS, GX_VTXFMT0, 4 );
      {
        F32 FactorU, FactorV;

        FactorU = floor( w_factor );
        FactorU /= w_factor; 

        FactorV = floor( h_factor );
        FactorV /= h_factor; 

        GXPosition3s16( x1, y1, 0); GXColor4u8(alvc, alvc, alvc, alv); GXTexCoord2f32( 0,       0 );       GXTexCoord2f32( 0,       0 );
        GXPosition3s16( x2, y1, 0); GXColor4u8(alvc, alvc, alvc, alv); GXTexCoord2f32( FactorU, 0 );       GXTexCoord2f32( FactorU, 0 );
        GXPosition3s16( x2, y2, 0); GXColor4u8(alvc, alvc, alvc, alv); GXTexCoord2f32( FactorU, FactorV ); GXTexCoord2f32( FactorU, FactorV ); 
        GXPosition3s16( x1, y2, 0); GXColor4u8(alvc, alvc, alvc, alv); GXTexCoord2f32( 0,       FactorV ); GXTexCoord2f32( 0,       FactorV );
      }
      GXEnd();
    }
  }

  // KOR - note - always comment this line out after upgrading the Bink SDK,
  // in order to keep the change made by Andrew Southgate.
//  GXSetDrawSync( 0 );
}


// sync the textures out to main memory, so that the GPU can see them (this is fast)
RADDEFFUNC void Sync_Bink_textures( BINKTEXTURESET * set_textures )
{
  DCStoreRange( set_textures->bink_buffers.Frames[ set_textures->bink_buffers.FrameNum ].YPlane.Buffer, set_textures->framesize );
}


// make sure the GPU is done with the textures that we are about to write into
RADDEFFUNC void Wait_for_Bink_textures( BINKTEXTURESET * set_textures )
{
  S32 next;
  
  next = set_textures->bink_buffers.FrameNum + 1;
  if ( next >= set_textures->bink_buffers.TotalFrames )
    next = 0;

  if ( set_textures->drawing[ next ] )
  {
    for(;;)
    {
      U16 v;
    
      v = GXReadDrawSync();
      if ( v != ( 0x4444 + next ) )
        break;
    }
    set_textures->drawing[ next ] = 0;
  }
}

