// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdConnection/bdConnectionStore.h>

#include <bdConnection/bdConnection.h>
#include <bdSocket/bdAddressMap.h>
#include <bdConnection/bdPacket/bdDefaultPacket.h>
#include <bdConnection/bdPacket/bdCookieEchoChunk.h>
#include <bdConnection/bdPacket/bdInitChunk.h>
#include <bdConnection/bdPacket/bdInitAckChunk.h>
#include <bdConnection/bdLoopbackConnection.h>
#include <bdConnection/bdUnicastConnection.h>
#include <bdSocket/bdSocketRouter.h>

#if defined BD_SOCKET_DO_BROADCAST && !defined BD_PLATFORM_PSP_ADHOC
#error "BD_SOCKET_DO_BROADCAST can only be defined on psp adhoc."
#endif

#define BD_LOG_LEVEL	"bdConnection/connectionstore"


#define BD_MAX_FLUSH_ITERATIONS (5u)

bdConnectionStore::bdConnectionStore(bdSocketRouter *socket,
									 bdCommonAddrRef me,
									 bdSecurityKeyMap * keyMap,
									 bdECCKey *const key)
: m_socket(socket),
  m_keyMap(keyMap),
  m_ECCKey(key),
  m_me(me)
{
	bdAddressMap::getInstance()->setTitleCommonAddr(me);
}

bdConnectionStore::~bdConnectionStore()
{
	closeAll();
	flushAll();

	m_ECCKey = BD_NULL;
	m_keyMap = BD_NULL;
}

void bdConnectionStore::registerListener(bdConnectionListener *listener)
{
#ifdef BD_DEBUG
	bdUInt32 index;
	if(m_listeners.findFirst(listener, index))
	{
		BD_WARN(BD_LOG_LEVEL, "same listener registered multiple times.");
	}
#endif // BD_DEBUG

	m_listeners.pushBack(listener);
}

void bdConnectionStore::unregisterListener(bdConnectionListener *listener)
{
	m_listeners.removeAllKeepOrder(listener);
}

bdConnectionRef bdConnectionStore::create(bdCommonAddrRef addr, const bdSecurityID &secID)
{
	BD_ASSERT(addr.notNull(), "Common address is null.");

	bdConnectionRef connection;

	if (addr->isLoopback())
	{
		BD_INFO(BD_LOG_LEVEL, "Creating new loopback connection.");
		connection = new bdLoopbackConnection(addr);
	}
	else
	{
		BD_INFO(BD_LOG_LEVEL, "Creating new unicast connection.");
		connection = new bdUnicastConnection(addr);
	}
	
	for (bdUInt i = 0 ; i < m_listeners.getSize(); i++)
	{
		connection->registerListener(m_listeners[i]);
	}

	if(connection)
	{
		bdAddrHandleRef addrHandle;
		bdAddressMap::getInstance()->commonAddrToAddr(addr, secID, addrHandle);
		connection->setAddressHandle(addrHandle);

		// first check to make sure the connection isn't already in the map.
		if(!m_connectionMap.get(addrHandle, connection))
		{
			if(!addr->isLoopback())
			{
				m_socket->connect(addrHandle);
			}

			m_connectionMap.put(addrHandle, connection);
		}
		else
		{
			BD_WARN(BD_LOG_LEVEL, "a connection already exists to host.");
		}
	}
	

	return connection;
}

void bdConnectionStore::remove(bdConnectionRef connection)
{	
	BD_INFO(BD_LOG_LEVEL, "Removing connection.");

	bdBool removed = m_connectionMap.remove(connection->getAddressHandle());
	
	if(!removed)
	{
		BD_WARN(BD_LOG_LEVEL, "failed to remove connection from store, connection not present.");
	}
}

void bdConnectionStore::closeAll()
{
	BD_INFO(BD_LOG_LEVEL, "Closing all connections.");

	bdHashMap<bdAddrHandleHashmapWrapper, bdConnectionRef, bdAddrHandleHashmapWrapper>::Iterator iter = m_connectionMap.getIterator();

	while(iter)
	{
		bdConnectionRef connection = m_connectionMap.getValue(iter);

		connection->close();

		m_connectionMap.next(iter);
	}

	m_connectionMap.clear();
}


void bdConnectionStore::disconnectAll()
{
	BD_INFO(BD_LOG_LEVEL, "Disconnecting all connections.");

	bdHashMap<bdAddrHandleHashmapWrapper, bdConnectionRef, bdAddrHandleHashmapWrapper>::Iterator iter = m_connectionMap.getIterator();

	while(iter)
	{
		bdConnectionRef connection = m_connectionMap.getValue(iter);

		connection->disconnect();

		m_connectionMap.next(iter);
	}
}

bdBool bdConnectionStore::flushAll()
{
	bdQueue<bdConnectionRef> toDisconnect;

	// Pump socket.
	m_socket->pump();

	// Flush outgoing connections
	bdHashMap<bdAddrHandleHashmapWrapper, bdConnectionRef, bdAddrHandleHashmapWrapper>::Iterator it = m_connectionMap.getIterator();

	while(it)
	{
		bdConnectionRef connection = m_connectionMap.getValue(it);

		if(!connection->getAddress()->isLoopback())
		{
			const bdSocketStatus &status = m_socket->getStatus(connection->getAddressHandle());
			bdUInt flushIters = 0;
			switch(status)
			{
				case BD_SOCKET_IDLE:
					BD_WARN(BD_LOG_LEVEL, "bad socket status.");
					break;
				case BD_SOCKET_PENDING:
					break;
				case BD_SOCKET_CONNECTED:
					
					while((flushIters < BD_MAX_FLUSH_ITERATIONS) && flush(*connection) > 0)
					{
						flushIters++;
					}

					if(connection->getStatus() == bdConnection::BD_DISCONNECTED)
					{
						const bdAddrHandleRef addrHandle = connection->getAddressHandle();
						bdNChar8 addrStr[BD_ADDR_STRING_SIZE];
						bdAddressMap::getInstance()->addrToString(addrHandle, addrStr, sizeof(addrStr));
						
						BD_INFO(BD_LOG_LEVEL, "Connection state disconnected. Disconnecting %s.", addrStr);
						toDisconnect.enqueue(connection);
					}
					break;
				case BD_SOCKET_LOST:
					{
						const bdAddrHandleRef addrHandle = connection->getAddressHandle();
						bdNChar8 addrStr[BD_ADDR_STRING_SIZE];
						bdAddressMap::getInstance()->addrToString(addrHandle, addrStr, sizeof(addrStr));

						BD_INFO(BD_LOG_LEVEL, "Socket router reports socket lost. Disconnecting %s.", addrStr);
						toDisconnect.enqueue(connection);						
					}
					break;
			}
		}
		else
		{
			(reference_cast<bdLoopbackConnection>(connection))->updateStatus();

			if(connection->getStatus() == bdConnection::BD_DISCONNECTED)
			{		
				toDisconnect.enqueue(connection);
			}
		}
		
		m_connectionMap.next(it);
	}

	// remove disconnected connections
	while(!toDisconnect.isEmpty())
	{
		bdConnectionRef connection = toDisconnect.peek();
		bdAddrHandleRef addr = connection->getAddressHandle();
		connection->close();

		// remove from the connection map.
		m_connectionMap.remove(bdAddrHandleHashmapWrapper(addr));
		// unregister the address. (this will tear down the secure association).
		bdAddressMap::getInstance()->unregisterAddr(addr);

		toDisconnect.dequeue();
	}

#if defined BD_SOCKET_DO_BROADCAST
	const bdInt status = m_socket->flushBuffer();
	if(status < 0)
	{
		switch(status)
		{
		case BD_NET_SUBSYTEM_ERROR:
			// something very wrong.
			BD_WARN(BD_LOG_LEVEL, "net subsystem error!");
			break;
		case BD_NET_ADDRESS_INVALID:
			BD_INFO(BD_LOG_LEVEL, "Invalid address. Closing connection.");
			break;
		case BD_NET_ERROR:
			// unknown error..
			BD_WARN(BD_LOG_LEVEL, "unknown error.");
			break;
		case BD_NET_WOULD_BLOCK:
			// doesn't happen.
			BD_WARN(BD_LOG_LEVEL, "would block.");
			break;
		case BD_NET_ADDRESS_IN_USE:
			// can't happen.. not binding
			break;
		case BD_NET_CONNECTION_RESET:
		case BD_NET_MSG_SIZE:
			// can't happen.. not receiving
			break;
		default:
			BD_WARN(BD_LOG_LEVEL, "unknown error.");
			break;
		}
	}
#endif

	return true;
}

bdUInt bdConnectionStore::flush(bdConnection& connection)
{
	bdUByte8 data[BD_MAX_PAYLOAD_SIZE];
	const bdUInt size = connection.getDataToSend(data, sizeof(data));

	if(size)
	{		
		// send to socket
		const bdAddrHandleRef addrHandle = connection.getAddressHandle();
		const bdInt status = m_socket->sendTo(addrHandle, data, size);

		if(status < 0)
		{
			switch(status)
			{
			case BD_NET_SUBSYTEM_ERROR:
				// something very wrong.
				BD_WARN(BD_LOG_LEVEL, "net subsystem error!");
				break;
			case BD_NET_ADDRESS_INVALID:
				BD_INFO(BD_LOG_LEVEL, "Invalid address. Closing connection.");
				connection.close();
				break;
			case BD_NET_ERROR:
				// unknown error..
				BD_WARN(BD_LOG_LEVEL, "unknown error.");
				break;
			case BD_NET_WOULD_BLOCK:
				// doesn't happen.
				BD_WARN(BD_LOG_LEVEL, "would block.");
				break;
			case BD_NET_ADDRESS_IN_USE:
				// can't happen.. not binding
				break;
			case BD_NET_CONNECTION_RESET:
			case BD_NET_MSG_SIZE:
				// can't happen.. not receiving
				break;
			default:
				BD_WARN(BD_LOG_LEVEL, "unknown error.");
				break;
			}
		}
	}

	return size;
}

bdBool bdConnectionStore::receiveFrom(bdConnectionRef &conn)
{
	// return once we have received a packed from a connected conection or
	// there is nothing left to receive	
	bdBool socketEmpty;	
	bdUByte8 data[BD_MAX_DATAGRAM_SIZE];	
	bdConnectionRef connection;

	do
	{
		// fail safe, will cause the loop to terminate
		socketEmpty = true;
		BD_ASSERT(connection.isNull(), "Socket receive loop failed to terminate.");

		bdAddrHandleRef address;
		const bdInt received = m_socket->receiveFrom(address, data, sizeof(data));

		// lookup up the address
		bdConnectionRef receivedConnection;
		if(address.notNull())
		{
			m_connectionMap.get(address, receivedConnection);
		}

		// check the error conditions
		if(received < 0)
		{
			switch(received)
			{
				// connection has been 
				case BD_NET_CONNECTION_RESET:
					{
						if(receivedConnection)
						{
							BD_INFO(BD_LOG_LEVEL, "Received 'connection reset'. Closing connection.");
							receivedConnection->close();
						}
					}
					break;
				// packet cannot fit in the buffer
				case BD_NET_MSG_SIZE:
					{
						BD_WARN(BD_LOG_LEVEL, "Couldn't receive message. Buffer too small?");
					}
					break;
				// socket is empty
				case BD_NET_WOULD_BLOCK:
				// blocking call was cancelled
				case BD_NET_BLOCKING_CALL_CANCELLED: 
				// socket not yet bound
				case BD_NET_NOT_BOUND: 
					break;
				default:
					BD_WARN(BD_LOG_LEVEL, "Unexpected socket error");
					break;
			}
		}
		else
		{
			// we have received something (may be zero sized)
			socketEmpty = false;
			
			const bdUInt bytesRecieved = static_cast<bdUInt>(received);
			
			// is there a matching connection
			if(receivedConnection.notNull() && receivedConnection->receive(data, bytesRecieved))
			{	
				connection = receivedConnection;				
			}
			else
			{
				// this is a new connection					

				// copy the incoming packet, to use again if the connection 
				// is complete
				bdUByte8 dataCopy[BD_MAX_DATAGRAM_SIZE];
				bdMemcpy(dataCopy, data, bytesRecieved);

				bdUnicastConnection newConnection;
				newConnection.setAddressHandle(address);
				bdBool ok = newConnection.receive(data, bytesRecieved);

				if(ok)
				{
					if(newConnection.getStatus() == bdConnection::BD_CONNECTED)
					{
						// this is an incomming connection that is now setup
						// so allocate it on the heap and record it
						const bdConnectionRef establishedConnection = new bdUnicastConnection;
						establishedConnection->setAddressHandle(address);

						// regsiter any listeners
						for(bdUInt i = 0; i < m_listeners.getSize(); i++)
						{
							establishedConnection->registerListener(m_listeners[i]);
						}

						// process the packet
						ok = establishedConnection->receive(dataCopy, bytesRecieved);

						if(ok && (establishedConnection->getStatus() == bdConnection::BD_CONNECTED))
						{
							if(m_connectionMap.put(address, establishedConnection))
							{								
								BD_INFO(BD_LOG_LEVEL, "New incoming connection created.");
								connection = establishedConnection;
							}							
						}
					}
					else
					{
						// give the connection a chance to send any connection packets
						flush(newConnection);
					}
				}
			}
		}
	}
	while(!socketEmpty && connection.isNull());

	// termination cases
	// empty  && null  ok - nothing to recieve
	// !empty && !null ok - received something
	// empty  && !null error - got something from an empty socket
	// !empty && null  error - terminated with nothing and the socket in not empty
	// some error checking
	if(socketEmpty)
	{
		BD_ASSERT(connection.isNull(), "No data but valid connection.");	
	}
	else
	{
		BD_ASSERT(connection.notNull(), "Data received, but no connection.");		
	}

	// if nothing is available on the network look at the loopback connections
	//TODO: should there ever be more than one loopback connection?	
	bdHashMap<bdAddrHandleHashmapWrapper, bdConnectionRef, bdAddrHandleHashmapWrapper>::Iterator it = m_connectionMap.getIterator();

	while(it && connection.isNull())
	{
		const bdConnectionRef iConnection = m_connectionMap.getValue(it);

		if(iConnection.notNull() && iConnection->getAddress().notNull() && iConnection->getAddress()->isLoopback())
		{
			const bdLoopbackConnectionRef loopbackConnection = reference_cast<bdLoopbackConnection>(iConnection);
			if(loopbackConnection->checkNewMessages())
			{
				connection = iConnection;
			}
		}

		m_connectionMap.next(it);
	}

	m_connectionMap.releaseIterator(it);

	const bdBool haveConnection = connection.notNull();
	if(haveConnection)
	{
		conn = connection;
	}

	return haveConnection;
}

