// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 PS2 PSP-INFRA XENON XBOX PS3 UNIX
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdCore/bdCore.h>

#include <bdConnection/bdStreamConnection.h>
#include <bdConnection/bdConnectionListener.h>
#include <bdConnection/bdMessage/bdMessage.h>

#include <bdCore/bdCrypto/bdCypher3Des.h>
#include <bdCore/bdCrypto/bdCryptoUtils.h>

//#include <bdPlatform/bdPlatformTiming/bdPlatformTiming.h>

#define BD_LOG_LEVEL	"bdConnection/connections"

bdStreamConnection::bdStreamConnection( bdCommonAddrRef addr)
:bdConnection(addr),
 m_messageSize(0)
 #if (!defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
 , m_messageCount(0)
#endif
{
	m_outgoingBuffer = new bdCircularBuffer<bdUByte8>(BD_STREAM_BUFFER_SIZE, false);
	BD_ASSERT(m_outgoingBuffer != 0, "Failed to allocate buffer");
	m_incomingBuffer = new bdCircularBuffer<bdUByte8>(BD_STREAM_BUFFER_SIZE, false);
	BD_ASSERT(m_incomingBuffer != 0, "Failed to allocate buffer");
	m_socket.create();
	m_socket.setBlocking(false);

	
}

bdStreamConnection::bdStreamConnection( bdCommonAddrRef addr,
									   void *sendBuffer,
									   void *recvBuffer,
									   bdUWord bufSize) :
bdConnection(addr),
m_messageSize(0)
#if (!defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
, m_messageCount(0)
#endif
{
	m_outgoingBuffer = new bdCircularBuffer<bdUByte8>((bdUInt)bufSize, sendBuffer, false);
	m_incomingBuffer = new bdCircularBuffer<bdUByte8>((bdUInt)bufSize, recvBuffer, false);
	m_socket.create();
	m_socket.setBlocking(false);
}

#if (defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
bdStreamConnection::bdStreamConnection( bdCommonAddrRef addr, 
										const bdSecurityID &secID)
:bdConnection(addr),
 m_messageSize(0),
 m_secID(secID)
{
	m_socket.create();
	m_socket.setBlocking(false);

	m_outgoingBuffer = new bdCircularBuffer<bdUByte8>(BD_STREAM_BUFFER_SIZE, false);
	BD_ASSERT(m_outgoingBuffer != 0, "Failed to allocate buffer");
	m_incomingBuffer = new bdCircularBuffer<bdUByte8>(BD_STREAM_BUFFER_SIZE, false);
	BD_ASSERT(m_incomingBuffer != 0, "Failed to allocate buffer");
}

#elif (defined BD_PLATFORM_XENON)
bdStreamConnection::bdStreamConnection( bdAddr addr, 
										const bdSecurityID &secID)
:m_messageSize(0),
 m_secID(secID)
{
	m_serverAddr = addr;
	m_socket.create();
	m_socket.setBlocking(false);

	m_outgoingBuffer = new bdCircularBuffer<bdUByte8>(BD_STREAM_BUFFER_SIZE, false);
	BD_ASSERT(m_outgoingBuffer != 0, "Failed to allocate buffer");
	m_incomingBuffer = new bdCircularBuffer<bdUByte8>(BD_STREAM_BUFFER_SIZE, false);
	BD_ASSERT(m_incomingBuffer != 0, "Failed to allocate buffer");
}

#else // This only exists on non X* platforms

void bdStreamConnection::setSessionKey(const bdUByte8 sessionKey[BD_ENCRYPTION_KEY_SIZE])
{
	bdMemcpy(m_sessionKey, sessionKey, BD_ENCRYPTION_KEY_SIZE); 
}
#endif // #if (defined BD_PLATFORM_XBOX) || (defined BD_PLATFORM_XENON)

bdStreamConnection::~bdStreamConnection()
{
	m_socket.close();
	delete m_outgoingBuffer;
	delete m_incomingBuffer;
}

bdBool bdStreamConnection::receive(bdUByte8 *, 
								 const bdUInt32)
{
	BD_ASSERT(g_assertFalse, "Should never be called.");

	return false;
}

bdBool bdStreamConnection::send(const bdMessageRef message,
				  const bdBool encrypt)
{
	bdBool result = false;

	bdByte8 encryptionType = encrypt ? BD_SEND_ENCRYPTED : BD_SEND_PLAINTEXT;
	if(m_status == BD_CONNECTED || m_status == BD_CONNECTING)
	{
		// Message header:
		// 1: Byte8 encryption type
		// 2: UInt32 IV seed
		// 3: Byte8[] Encrypted message
		//     |_ 4 bytes = 0xDEADBEEF

		BD_ASSERT(message->getPayload()->getDataSize() < BD_STREAM_MAX_MESSAGE_SIZE,
				"Message too big.");

#if (!defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
		if (encrypt)
		{
			bdUInt32 signature = BD_SIGNATURE_CHECK;
			bdCypher3Des cypher;

			// Prepare some size constants.
			const bdUInt cypherSize = BD_NEXT_MULTIPLE_OF_M(
					static_cast<bdUInt16>(sizeof(signature) +
					sizeof(bdMessageType) +
					message->getPayload()->getDataSize()),
					8);
			const bdUInt16 sizePrefix = static_cast<bdUInt16>(cypherSize + sizeof(encryptionType)
				+ sizeof(m_messageCount));
			const bdInt headerSize = sizeof(sizePrefix) + sizeof(encryptionType) + sizeof(m_messageCount);
			
			bdUInt offset = 0;

			// Check that there's enough buffer space for this message.
			pump();
			if (m_outgoingBuffer->getTotalFreeSpace() < cypherSize + headerSize)
			{
				BD_INFO(BD_LOG_LEVEL, "Not enough buffer space for the message.");
				return false;
			}

			if (m_outgoingBuffer->getFreeBlockSize() < cypherSize + headerSize)
			{
				m_outgoingBuffer->alignUsedBlock();
				BD_ASSERT(m_outgoingBuffer->getFreeBlockSize() >= cypherSize + headerSize,
					"Outgoing buffer align failed.");
			}

			bdUByte8 initialVector[BD_TIGER_HASH_SIZE];
			bdCryptoUtils::calculateInitialVector(m_messageCount, initialVector);

			// Put unencrypted portion of the message in.
			result = true;
			result = result && m_outgoingBuffer->appendBasicType(sizePrefix);
			result = result && m_outgoingBuffer->appendBasicType(encryptionType);
			result = result && m_outgoingBuffer->appendBasicType(m_messageCount);

			// Above operations should never fail.
			BD_ASSERT(result, "Failed to write into send buffer.");

			bdUByte8 *encBuffer = m_outgoingBuffer->getFreeBlockStart();
			bdUInt encBufferSize = m_outgoingBuffer->getFreeBlockSize();

			offset = 0;
			result = result && bdBytePacker::appendBasicType(encBuffer, encBufferSize, offset, offset, signature);
			result = result && bdBytePacker::appendBasicType(encBuffer, encBufferSize, offset, offset, message->getType());
			result = result && bdBytePacker::appendBuffer(encBuffer, encBufferSize, offset, offset,
						message->getPayload()->getData(), message->getPayload()->getDataSize());

			// Above operations should never fail.
			BD_ASSERT(result, "Failed to write into send buffer.");

			bdUInt tmp = cypherSize;
			result = result && cypher.encrypt(  m_sessionKey,
												BD_ENCRYPTION_KEY_SIZE,
												initialVector,
												encBuffer,
												encBuffer,
												tmp);

			// Encryption should never fail.
			BD_ASSERT(result, "Encryption failed.");
			
			m_outgoingBuffer->protectBlock(cypherSize);

			m_messageCount++;

			// Preceding operations should never fail
			BD_ASSERT(result, "Failed to prepare the packet.");
		}
		else
#endif // XBoxes
		{
#if (defined BD_PLATFORM_XBOX) || (defined BD_PLATFORM_XENON)

			encryptionType = BD_SEND_PLAINTEXT;

#endif // XBoxes
			const bdUInt16 payloadSize = static_cast<bdUInt16>(message->getPayload()->getDataSize());
			const bdUInt16 sizePrefix = payloadSize + sizeof(encryptionType) + sizeof(bdMessageType);
			const bdUInt headerSize = sizeof(sizePrefix) + sizeof(encryptionType) + sizeof(bdMessageType);

			// Check that there's enough buffer space for this
			pump();
			if (m_outgoingBuffer->getTotalFreeSpace() < payloadSize + headerSize)
			{
				BD_INFO(BD_LOG_LEVEL, "Not enough buffer space for the message.");
				return false;
			}

			if (m_outgoingBuffer->getFreeBlockSize() < payloadSize + headerSize)
			{
				m_outgoingBuffer->alignUsedBlock();
				BD_ASSERT(m_outgoingBuffer->getFreeBlockSize() >= payloadSize + headerSize,
					"Outgoing buffer align failed.");
			}

			result = true;
			result = result && m_outgoingBuffer->appendBasicType(sizePrefix);
			result = result && m_outgoingBuffer->appendBasicType(encryptionType);
			const bdMessageType msgType = message->getType();
			result = result && m_outgoingBuffer->appendBasicType(msgType);

			result = result && m_outgoingBuffer->appendBuffer(message->getPayload()->getData(), payloadSize);

			// Preceding operations should never fail
			BD_ASSERT(result, "Failed to prepare the header.");
		}

	}

	pump();
	return result;
}


bdBool bdStreamConnection::getMessageToDispatch(bdMessageRef &message)
{
	if (m_status == BD_DISCONNECTED)
	{
		return false;
	}
	pump();
	bdBool result = false;

	if(m_status == BD_CONNECTED || m_status == BD_CONNECTING)
	{
		//bdBool parse = false;
		bdInt status = 0;

		// If we don't have a full message in the buffer, attempt to get one.
		if (!receivedFullMessage())
		{
			status = m_socket.recv(m_incomingBuffer->getFreeBlockStart(),
					m_incomingBuffer->getFreeBlockSize());
			if (status > 0)
			{
				m_incomingBuffer->protectBlock(static_cast<bdUInt>(status));
#if defined BD_SEND_KEEP_ALIVE
				m_lastReceivedTimer.reset();
				m_lastReceivedTimer.start();
#endif
			}
		}

		if (m_messageSize > BD_STREAM_MAX_MESSAGE_SIZE)
		{
			BD_WARN(BD_LOG_LEVEL, "Message received is too large to fit in the receive buffer.");
			close();
			return false;
		}

		if(status <= 0 && (!receivedFullMessage()))
		{
			// handle error!
			switch(status)
			{
				case BD_NET_CONNECTION_CLOSED:
					{
						m_status = BD_DISCONNECTING;
						callListenersDisconnect();
						m_status = BD_DISCONNECTED;
						break;
					}
				case BD_NET_WOULD_BLOCK:
					if(m_status == BD_CONNECTING)
					{
						m_status = BD_CONNECTED;
						callListenersConnect(true);
					}
					break;
				case BD_NET_CONNECTION_RESET:
					{
						BD_INFO(BD_LOG_LEVEL, "Received 'connection reset'. Closing connection.");
						close();
					}
					break;
				case BD_NET_MSG_SIZE:
					BD_WARN(BD_LOG_LEVEL, "Couldn't receive message. Buffer too small?");
					break;
				case BD_NET_BLOCKING_CALL_CANCELLED:
					//blocking call was canceled.
					break;
				case BD_NET_NOT_BOUND:
					break;
				case BD_NET_NOT_CONNECTED:
					break;
				default:
					BD_WARN(BD_LOG_LEVEL, "net error. status : %i" , status);
					break;
			}
		}
		else if(receivedFullMessage())
		{
			if(m_status == BD_CONNECTING)
			{
				m_status = BD_CONNECTED;
				callListenersConnect(true);
			}

			//BD_INFO("info","Received %u bytes\n", m_status);

			result = true;
			bdByte8 encryptionType = 0;
			result = result && m_incomingBuffer->removeBasicType(m_messageSize);
			result = result && m_incomingBuffer->removeBasicType(encryptionType);

			BD_ASSERT(result, "Preceeding operations should never fail.");

			if (encryptionType == BD_SEND_ENCRYPTED)
			{
#if (!defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
				bdUInt32 signature = 0;
				bdUInt plainTextSize = m_messageSize - (sizeof(encryptionType) + sizeof(m_messageCount));
				bdUInt plaintextHeaderSize = sizeof(bdMessageType) + sizeof(signature);
				bdUInt32 IVseed = 0;
				result = result && m_incomingBuffer->removeBasicType(IVseed);

				bdUByte8 initialVector[BD_TIGER_HASH_SIZE];
				bdCryptoUtils::calculateInitialVector(IVseed, initialVector);
				bdCypher3Des cypher;

				BD_ASSERT(plainTextSize <= m_incomingBuffer->getUsedBlockSize(), "Lost buffer alignment.");

				result = result && cypher.decrypt(  m_sessionKey,
													BD_ENCRYPTION_KEY_SIZE,
													initialVector,
													m_incomingBuffer->getUsedBlockStart(),
													m_incomingBuffer->getUsedBlockStart(),
													plainTextSize);

				BD_ASSERT(result, "Decryption failed."); // Should never happen

				
				result = result && m_incomingBuffer->removeBasicType(signature);
				if (signature != BD_SIGNATURE_CHECK)
				{
					BD_WARN(BD_LOG_LEVEL, "Decryption failed.");
					result = false;
				}
				else
				{
					bdMessageType type = 0;
					result = result && m_incomingBuffer->removeBasicType(type);
					if (result)
					{
						message = new bdMessage(type, m_incomingBuffer->getUsedBlockStart(),
												plainTextSize - plaintextHeaderSize, true);
					}
					m_incomingBuffer->freeBlock(plainTextSize - plaintextHeaderSize);
					m_messageSize = 0;
				}
#else // Not XBoxes
				BD_ASSERT(g_assertFalse, "Should never be called.");
#endif // XBoxes
			}
			else
			{
				bdMessageType type = 0;
				bdUInt payloadSize = m_messageSize - (sizeof(encryptionType) + sizeof(type));
				result = true;
				result = result && m_incomingBuffer->removeBasicType(type);
				if (result)
				{
					message = new bdMessage(type, m_incomingBuffer->getUsedBlockStart(),
											payloadSize, true);
				}
				m_incomingBuffer->freeBlock(payloadSize);
				m_messageSize = 0;
			}
		}
	}
	return result;
}

bdBool bdStreamConnection::connect()
{
	m_status = BD_CONNECTING;

#if (defined BD_PLATFORM_XBOX) && (!defined BD_PLATFORM_XENON)
	// FIXME , this only permits connections to Title Servers....
	bdBool result = false;
	const DWORD titleid = static_cast<DWORD>(m_addr->getTitleId());
	IN_ADDR server;
		
	result = XNetTsAddrToInAddr(&(m_addr->getXNAddr()), titleid, &m_secID, &server) == 0;

	bdAddr serverAddr(server.S_un.S_addr, m_addr->getPort());
	const bdSocketStatusCode status = m_socket.connect(serverAddr);
#elif defined BD_PLATFORM_XENON
	// FIXME , this only permits connections to Title Servers....
	const bdSocketStatusCode status = m_socket.connect(m_serverAddr);
#else
	const bdSocketStatusCode status = m_socket.connect(m_addr->getPublicAddr());
#endif

	switch(status)
	{
		case BD_NET_SUCCESS:
			m_status = BD_CONNECTED;
			callListenersConnect(true);
			return true;
		case BD_NET_WOULD_BLOCK:
			return true;
		default:
			close();
			return false;
	}
}

void bdStreamConnection::disconnect()
{
	close();
}

void bdStreamConnection::close()
{

	if(m_status == BD_CONNECTING)
	{
		callListenersConnect(false);
	}
	else if(m_status != BD_DISCONNECTED)
	{
		callListenersDisconnect();	
	}
	m_status = BD_DISCONNECTED;
	m_socket.close();
}

// protected:
bdUInt bdStreamConnection::getDataToSend(bdUByte8 *const,
										   const bdUInt32)
{
	BD_ASSERT(g_assertFalse, "Should never be called.");

	return 0;
}


void bdStreamConnection::callListenersConnect(const bdBool success)
{
	for(bdUInt i = 0; i < m_listeners.getSize(); i++)
	{
		if(success)
		{
#if defined BD_SEND_KEEP_ALIVE
			m_keepAliveTimer.reset();
			m_keepAliveTimer.start();
			m_lastReceivedTimer.reset();
			m_lastReceivedTimer.start();
#endif
			m_listeners[i]->onConnect(this);
		}
		else
		{
			m_listeners[i]->onConnectFailed(this);
		}
	}
}

void bdStreamConnection::callListenersDisconnect()
{
	for(bdUInt i = 0; i < m_listeners.getSize(); i++)
	{
		m_listeners[i]->onDisconnect(this);
	}
}

void bdStreamConnection::callListenersReconnect()
{
	for(bdUInt i = 0; i < m_listeners.getSize(); i++)
	{
		m_listeners[i]->onReconnect(this);
	}
}

bdBool bdStreamConnection::pump()
{
	while (m_outgoingBuffer->getUsedBlockSize() > 0 && 
		(m_status == BD_CONNECTED || m_status == BD_CONNECTING))
	{
		const bdInt sent = m_socket.send(m_outgoingBuffer->getUsedBlockStart(),
										m_outgoingBuffer->getUsedBlockSize());
		if (sent > 0)
		{
#if defined BD_SEND_KEEP_ALIVE
			m_keepAliveTimer.start();
#endif
			if (m_status == BD_CONNECTING)
			{
				m_status = BD_CONNECTED;
				callListenersConnect(true);
			}
			m_outgoingBuffer->freeBlock(static_cast<bdUInt>(sent));
		}
		else
		{
			switch(sent)
			{
				case BD_NET_SUBSYTEM_ERROR:
					// something very wrong.
					BD_WARN(BD_LOG_LEVEL, "net subsystem error!");
					close();
					break;
				case BD_NET_ADDRESS_INVALID:
					BD_INFO(BD_LOG_LEVEL, "Invalid address. Closing connection.");
					close();
					break;
				case BD_NET_ERROR:
					// unknown error..
					BD_WARN(BD_LOG_LEVEL, "unknown error.");
					close();
					break;
				case BD_NET_WOULD_BLOCK:
					// Is actually ok to happen. Means there is no buffer space at all, or something.
					//BD_WARN(BD_LOG_LEVEL, "would block.");
					break;
				case BD_NET_ADDRESS_IN_USE:
					// can't happen.. not binding
					break;
				case BD_NET_CONNECTION_RESET:
				case BD_NET_MSG_SIZE:
					close();
					break;
				case BD_NET_INVALID_HANDLE:
					BD_WARN(BD_LOG_LEVEL, "invalid handle.");
					close();
					break;
				case BD_NET_NOT_CONNECTED:
					/*
					if (m_status == BD_CONNECTING)
					{
						//BD_INFO(BD_LOG_LEVEL, "Failed sending because connection is still pending.");
					}
					else
					{
						close();
					}*/
					break;
				default:
					BD_WARN(BD_LOG_LEVEL, "unknown error.");
					close();
					break;
			}
			return m_status == BD_CONNECTED || m_status == BD_CONNECTING;
		}
	}
#if defined BD_SEND_KEEP_ALIVE
	if (m_keepAliveTimer.getElapsedTimeInSeconds() > BD_KEEP_ALIVE_PERIOD)
	{
		bdUInt16 keepAlive = 0;
		m_socket.send(&keepAlive, sizeof(keepAlive));
		m_keepAliveTimer.start();
	}
	if (m_lastReceivedTimer.getElapsedTimeInSeconds() > 2 * BD_KEEP_ALIVE_PERIOD)
	{
		BD_WARN("warn","Connection timed out\n");
		close();
	}

#endif
	return true;
}

bdBool bdStreamConnection::receivedFullMessage()
{
	// See if we need to shift
	if (m_incomingBuffer->getUsedBlockSize() < sizeof(m_messageSize) && 
		m_incomingBuffer->getFreeBlockSize() + m_incomingBuffer->getUsedBlockSize() < sizeof(m_messageSize))
	{
		m_incomingBuffer->alignUsedBlock();
	}

	
	if (m_incomingBuffer->getUsedBlockSize() >= sizeof(m_messageSize))
	{
		//m_messageSize = *(reinterpret_cast<bdUInt16*>(m_incomingBuffer->getUsedBlockStart()));
		bdUInt offset =0;
		bdBytePacker::removeBasicType(m_incomingBuffer->getUsedBlockStart(), m_incomingBuffer->getUsedBlockSize(), 0, offset, m_messageSize);
		offset++;
		// Special case: a message size of 0 indicates a reply from the server 
		// following a keep alive message sent from the client. 
#if defined BD_SEND_KEEP_ALIVE
		if (m_messageSize == 0)
		{
			bdUInt16 temp;
			m_incomingBuffer->removeBasicType(temp);
		}
#endif
	}

	if (m_messageSize > 0 && m_incomingBuffer->getUsedBlockSize() >= m_messageSize + sizeof(m_messageSize))
	{
		return true;
	}
	

	// See if we need to shift to be able to fit the whole message
	if (m_incomingBuffer->getUsedBlockSize() + m_incomingBuffer->getFreeBlockSize() < 
			sizeof(m_messageSize) + m_messageSize)
	{
		m_incomingBuffer->alignUsedBlock();
	}

	return false;
}


