// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 PS2 UNIX OSX 
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdCore/bdSocket/bdThreadedSocket.h>

#include <bdPlatform/bdPlatformThread/bdPlatformThread.h>

#ifdef BD_PS2_USE_LIBEENET
#include <libeenet.h>
#undef bind
#endif // BD_PS2_USE_LIBEENET

#include <bdCore/bdThread/bdThread.h>

#define BD_LOG_LEVEL "threadedsocket"

bdThreadedSocket::bdThreadedSocket()
: m_receiveThread(BD_NULL),
  m_receiver(BD_NULL), 
  m_sendThread(BD_NULL),
  m_sender(BD_NULL),
  m_incomingBufferSize(BD_MAX_DATAGRAM_SIZE)
{
}

bdThreadedSocket::~bdThreadedSocket()
{
	close();
}

void bdThreadedSocket::setIncomingBufferSize(const bdUInt size)
{
	m_incomingBufferSize = size;
}

bdBool bdThreadedSocket::create(const bdBool broadcast)
{
	bdInt priority = BD_DEFAULT_THREAD_PRIORITY;
#if (defined BD_PLATFORM_PS2 ||  defined BD_PLATFORM_PSP)
	// try to pick a thread proirity that is higher that 'this' thread
	priority = bdPlatformThread::getPriority();
	BD_ASSERT(priority > 1, "bdThreadedSocket::create, threaded socket threads "
		"must be of higher priority than the main thread.");
	priority -= 1;
#endif // BD_PLATFORM_PS2

	const bdBool created = bdSocket::create(broadcast);

	if(created)
	{
		bdSocket::setBlocking(true);
		if (!m_receiver)
		{
			m_receiver = new bdReceiveThread(m_handle, m_incomingBufferSize);
		}
		if (!m_receiveThread)
		{
			m_receiveThread = new bdThread(m_receiver, priority);
		}
		if (!m_sender)
		{
			m_sender = new bdSendThread(m_handle);
		}
		m_sender->m_isBound = false;
		if (!m_sendThread)
		{
			m_sendThread = new bdThread(m_sender, priority);
		}
		// Start the send-thread. The receive thread is started when either
		// sendTo() or bind() is called
		m_sendThread->start(BD_NULL,0);
	}
	
	return created;
}


bdSocketStatusCode bdThreadedSocket::bind(const bdPort port)
{	
	bdSocketStatusCode success = bdSocket::bind(port);
	if (success)
	{
		m_receiveThread->start(BD_NULL,0);
	}
	return success;
}

bdSocketStatusCode bdThreadedSocket::bind(const bdAddr &addr)
{
	bdSocketStatusCode success = bdSocket::bind(addr);
	if (success)
	{
		m_receiveThread->start(BD_NULL,0);
	}
	return success;
}

bdInt bdThreadedSocket::sendTo(const bdAddr &addr, 
							   const void *data, 
							   const bdUInt length)
{
	const bdInt status = m_sender->sendTo(addr, data, length);
	if (!m_receiveThread->isRunning())
	{
		// Wait until the socket is bound
		m_sender->waitForBind();
		m_receiveThread->start(BD_NULL,0);
	}
	return status;
}


bdInt bdThreadedSocket::receiveFrom(bdAddr &addr, 
									void *data,
									const bdUInt size)
{
	return m_receiver->receiveFrom(addr, data, size);
}

bdBool bdThreadedSocket::close()
{	
	if (m_handle == BD_INVALID_SOCKET_HANDLE)
	{
		return false;
	}
	if (m_receiveThread)
	{
		m_receiveThread->stop();

#ifdef BD_PS2_USE_LIBEENET
		sceEENetThreadAbort( m_receiveThread->getThreadHandle() );
#endif
		
	}
	const bdBool status = bdSocket::close();
	
	if (m_sendThread)
	{ 
		m_sendThread->stop();
		m_sendThread->join();
#if defined BD_PLATFORM_PS2 
		bdPlatformSocket::unregisterThread(m_sendThread->getThreadHandle());
#endif
		m_sendThread->cleanup();
		m_sendThread = BD_NULL;
	}
	if (m_receiveThread)
	{
		m_receiveThread->join();
#if defined BD_PLATFORM_PS2
		bdPlatformSocket::unregisterThread(m_receiveThread->getThreadHandle());
#endif
		m_receiveThread->cleanup();
		m_receiveThread = BD_NULL;
	}
	
	delete m_receiver;
	m_receiver = BD_NULL;
	delete m_sender;
	m_sender = BD_NULL;
	return status;
}


// class bdSendThread

bdSendThread::bdSendThread(const bdInt handle)
: m_handle(handle),
  m_isBound(false)
{
}

bdSendThread::~bdSendThread()
{
}

bdInt bdSendThread::sendTo(const bdAddr &addr,
						   const void *data,
						   const bdUInt length)
{
	bdByteBufferRef buffer = new bdByteBuffer(length);
	bdMemcpy(buffer->getData(), data, length);

	m_lock.lock();
	m_toSend.enqueue(bdSendData(addr, buffer));
	m_lock.unlock();

	// data available
	m_signal.signal();
	return length;
}

bdUInt bdSendThread::run(void *)
{

	while(!m_stop)
	{	
		m_signal.wait();
		if(!m_toSend.isEmpty())
		{
			m_lock.lock();
			const bdSendData sendData = m_toSend.peek();
			m_toSend.dequeue();
			m_lock.unlock();

			bdPlatformSocket::sendTo(m_handle,
									 sendData.m_addr.getAddress().getInAddr(),
									 sendData.m_addr.getPort(),
									 sendData.m_data->getData(), 
									 sendData.m_data->getSize());
			// After the first sendTo() the socket will be bound. We keep track of it
			// here in order to safely start the receiving thread
			if (!m_isBound)
			{
				m_bindSignal.signal();
				m_isBound = true;
			}
		}
	}

	return 0;
}

void bdSendThread::stop()
{
	if(!m_stop)
	{
		bdRunnable::stop();
		m_signal.signal();
	}
}

void bdSendThread::waitForBind()
{
	if (!m_isBound)
	{	
		m_bindSignal.wait();
	}
}

// class bdReceiveThread

bdReceiveThread::bdReceiveThread(const bdInt handle, 
								 const bdUInt maxQueueSize)
: m_recvQueueSize(0),
  m_maxRecvQueueSize(maxQueueSize),
  m_handle(handle)
{
}

bdReceiveThread::~bdReceiveThread()
{
}

bdInt bdReceiveThread::receiveFrom(bdAddr &addr, 
								   void *data,
								   const bdUInt size)
{
	bdInt status = BD_NET_WOULD_BLOCK;

	m_lock.lock();

	if(!m_received.isEmpty())
	{
		bdReceiveData rdata = m_received.peek();
		m_received.dequeue();

		if(rdata.m_status >= 0)
		{
			if(rdata.m_data->getSize() <= size)
			{
				addr = rdata.m_addr;
				status = rdata.m_status;
				bdMemcpy(data, rdata.m_data->getData(), rdata.m_data->getSize());
			}
			else
			{
				status = BD_NET_MSG_SIZE;
			}

			m_recvQueueSize -= status;
		}
		else
		{
			addr = rdata.m_addr;
			status = rdata.m_status;
		}
	}

	m_lock.unlock();

	return status;
}

bdUInt bdReceiveThread::run(void *)
{
	const bdUInt size = BD_MAX_DATAGRAM_SIZE;
	bdUByte8 data[BD_MAX_DATAGRAM_SIZE];
	bdInAddr inaddr;
	bdPort port;

	while(!g_assertFalse)
	{
		bdInt status = bdPlatformSocket::receiveFrom(m_handle, inaddr, port, data, size);
		if (m_stop)
		{
			return 0;
		}
		
		if(status < 0 || m_recvQueueSize + status < m_maxRecvQueueSize)
		{
			bdByteBufferRef buffer;

			if(status >= 0)
			{
				buffer = new bdByteBuffer(static_cast<bdUInt>(status));
				bdMemcpy(buffer->getData(), data, static_cast<bdUInt>(status));

				m_recvQueueSize += status;
			}
			bdInetAddr inetAddr(inaddr);
			bdAddr addr(inetAddr, port);
			m_lock.lock();
			m_received.enqueue(bdReceiveData(status, addr, buffer));
			m_lock.unlock();
		}
		else
		{
			//BD_WARN(BD_LOG_LEVEL, "bdReceiveThread::run, buffer too small. Packet discarded");
		}
	}

	return 0;
}

