// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: PS2
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdCore/bdStreams/bdPs2FileStreambuf.h>
#include <bdCore/bdLog/bdLog.h>
#include <bdCore/bdUtilities/bdBitOperations.h>

#include <bdPlatform/bdPlatformString/bdPlatformString.h>

#include <sifdev.h>

bdPs2FileStreambuf::bdPs2FileStreambuf(const bdChar8 *const filename,
									   const OpenMode mode)
{
	bdInt32 fileMode = 0;

	switch(mode)
	{
		case READ:
			fileMode = SCE_RDONLY;
			break;
		case WRITE:
			fileMode = SCE_WRONLY | SCE_CREAT | SCE_TRUNC;
			break;
		case APPEND:
			fileMode = SCE_WRONLY | SCE_CREAT | SCE_APPEND;
			break;
		default:
			BD_ASSERT(false, "Unknown OpenMode for bdPs2FileStreambuf");
	}

	m_handle = ::sceOpen(filename, fileMode);
	if (m_handle < 0)
	{
		m_flags |= END_OF_FILE;
		m_readWritebuffer = BD_NULL;
	}
	else
	{
		m_readWritebuffer = bdAlignedAllocate<bdChar8>(BD_PS2_FILE_BUFFER_SIZE*2, 64);
		setGetBuffer(m_readWritebuffer, BD_PS2_FILE_BUFFER_SIZE);
		setPutBuffer(m_readWritebuffer+BD_PS2_FILE_BUFFER_SIZE, BD_PS2_FILE_BUFFER_SIZE);
	}
}

bdPs2FileStreambuf::~bdPs2FileStreambuf()
{
	if (m_handle >= 0)
	{
		flush();
		sceClose(m_handle);
		m_handle = -1;
	}

	if(m_readWritebuffer)
	{
		bdDeallocate<bdChar8>(m_readWritebuffer);
		m_readWritebuffer = BD_NULL;
	}
}

bdUInt32 bdPs2FileStreambuf::bufGetFrom(bdChar8 *const dest, 
										const bdUInt32 nbytes)
{
	BD_ASSERT(m_handle >= 0, "bdPs2FileStreambuf has invalid handle.");

	BD_ASSERT(BD_IS_MULTIPLE_OF_M(reinterpret_cast<bdUInt32>(dest), 64), 
		"dest must be 64 byte aligned.");

	BD_ASSERT(BD_IS_MULTIPLE_OF_M(nbytes, 512), 
		"read size must be a multiple of 512.");

	const bdInt32 status = sceRead(m_handle, dest, nbytes);
	if(status < 0)
	{
		// something went wrong
		return 0;
	}

	return status;
}

bdUInt32 bdPs2FileStreambuf::bufPutTo(bdChar8 *const src, 
									  const bdUInt32 nbytes)
{
	BD_ASSERT(m_handle >= 0, "bdPs2FileStreambuf has invalid handle.");

	BD_ASSERT(BD_IS_MULTIPLE_OF_M(reinterpret_cast<bdUInt32>(src), 64), 
		"src must be 64 byte aligned.");
	
	if (nbytes != BD_PS2_FILE_BUFFER_SIZE)
	{
		// should only get in here on the final flush
		// as we use a 512 byte buffer by default.
		
		// clear the end of the buffer and proceed
		bdMemset(src+nbytes, BD_NULL, BD_PS2_FILE_BUFFER_SIZE-nbytes);
	}

	const bdInt32 status = sceWrite(m_handle, src, BD_PS2_FILE_BUFFER_SIZE);

	if(status < 0)
	{
		// something went wrong
		return 0;
	}

	return status;
}

bdBool bdPs2FileStreambuf::bufSeek(const bdInt32 offset, 
								   const SeekFrom from)
{
	BD_ASSERT(m_handle >= 0, "bdPs2FileStreambuf has invalid handle.");

	BD_ASSERT(BD_IS_MULTIPLE_OF_M(offset, 512), 
		"read size must be a multiple of 512.");

	bdInt32 whence = 0;
	switch(from)
	{
		case BEG:
			whence = SCE_SEEK_SET;
			break;
		case CUR:
			whence = SCE_SEEK_CUR;
			break;
		case END:
			whence = SCE_SEEK_END;
			break;
		default:
			BD_ASSERT(false, "Unknown SeekFrom for bdPs2FileStreambuf");		
	}

	if (sceLseek(m_handle, offset, whence) >= 0)
	{
		return true;
	}

	return false;
}

bdBool bdPs2FileStreambuf::bufTell(bdUInt32 &offset) const
{
	BD_ASSERT(m_handle >= 0, "bdPs2FileStreambuf has invalid handle.");

	const bdInt32 status = sceLseek(m_handle, 0, SCE_SEEK_CUR);
	if (status >= 0)
	{
		offset = status;
		return true;
	}

	return false;
}

