// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE: Clients side interface to the DemonWare friends service. 

#ifndef BD_FRIENDS_H
#define BD_FRIENDS_H

#include <bdCore/bdCore.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdLobbyErrorCode.h>
#include <bdLobby/bdLobbyCommon/bdQueryResult.h>

class bdRemoteTask;
class bdRemoteTaskManager;

BD_REFERENCE(bdConnection);

// Friends operation types.
#define BD_FRIEND_PROPOSE_FRIENDSHIP		(1u)
#define BD_FRIEND_ACCEPT_PROPOSAL			(2u)
#define BD_FRIEND_REJECT_PROPOSAL			(3u)
#define BD_FRIEND_REVOKE_FRIENDSHIP			(4u)
#define BD_FRIEND_GET_FRIENDS				(5u)
#define BD_FRIEND_SET_BLOCKED_STATUS		(6u)
#define BD_FRIEND_GET_BLOCKED				(7u)
#define BD_FRIEND_INVITE_TO_SESSION			(8u)
#define BD_FRIEND_ACCEPT_INVITE				(9u)
#define BD_FRIEND_REJECT_INVITE				(10u)
#define BD_FRIEND_GET_INVITES				(11u)
#define BD_FRIEND_SEND_FRIEND_MSG			(12u)
#define BD_FRIEND_WITHDRAW_PROPOSAL			(13u)
#define BD_FRIEND_SET_GROUP					(14u)
#define BD_FRIEND_SET_GROUP_NAME			(15u)
#define BD_FRIEND_GET_GROUPS				(16u)
#define BD_FRIEND_GET_GROUPED_FRIENDS		(17u)
#define BD_FRIEND_DELETE_GROUP_NAME			(18u) 
#define BD_FRIEND_GET_PROPOSALS				(19u) 

/// Clients side interface to the DemonWare friends service. 
/// Manages friends lists, block lists and friend messaging.
/// Use the functions contained in this class to perform operations on the server.
/// Each of these function returns a task handle, which should be pumped to 
/// completion by calling its getStatus() function. When the status of a task is 
/// bdRemoteTask::BD_DONE, you can pass it into the bdFriends::handleTask()
/// function. Once this is done, the error code for the task can be accessed via
/// bdFriends::getErrorCode(). If you expect a result set to be returned from
/// the server as a result of the task you started, you must pass an instance of 
/// a bdQueryResult into bdFriends::handleTask. The implementation must be
/// specialized to match the results you are expecting. Standard implementations
/// are provided.
class bdFriends
{
	public:
		BD_DECLARE_NEW_AND_DELETE_OPERATORS		
		
		enum bdMessageDeliveryMethod
		{
			BD_PULL_ONLY,
			BD_PUSH_ONLY,
			BD_PUSH_AND_PULL
		};

		enum bdBlockLevel
		{
			BD_NO_BLOCK,
			BD_FULL_BLOCK,
			BD_INVITES_BLOCK
		};

		~bdFriends();

		/// Invite a user to become a friend.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param friendID[in]			The friend's user ID.
		/// \param attachment[in]		An optional attachment to send with the
		///								proposal message.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned if an error occurred. The invited user will 
		///				be sent an proposal and must accept it in order to become
		///				a friend.
		/// \note		A BD_FRIENDSHIP_PROPOSAL_MSG will be delivered to the
		///				specified user ID.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success. 
		///				BD_PENDING_FRIENDSHIP_EXISTS - if an existing friendship 
		///				proposal has yet to be dealt with.
		///				BD_FRIENDSHIP_EXISTS - if the specified friendID is already
		///				present in this user's friend list.
		bdRemoteTask *proposeFriendship(bdUByte8 controllerIndex, 
										bdOnlineUserID friendID,
										bdByteBuffer * const attachment = BD_NULL);

		/// Revoke a proposal for friendship that has not yet been accepted or
		/// rejected.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param friendID[in]			The user to which the original proposal was 
		///								addressed.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned if an error occurred.
		/// \note		A BD_FRIENDSHIP_PROPOSAL_CANCEL_MSG will be delivered to 
		///				the user specified by friendID.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		bdRemoteTask *withdrawProposal(bdUByte8 controllerIndex, 
									   bdOnlineUserID friendID);

		/// Get a list of all friendship proposals sent by this user that have not
		/// yet been accepted or rejected.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \return		A pointer to the task that handles the request. BD_NULL is 
		///				returned if an error occurred. 
		/// \note		When this task completes, you can obtain the result set by 
		///				invoking bdFriends::handleTask with an instance of 
		///				bdGetProposalListResult.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success. 
		bdRemoteTask *getProposalList(bdUByte8 controllerIndex);

		/// Accept a friendship proposal.
		/// This function should be invoked when the local users wishes to accept
		/// a friendship proposal message that has been received. This completes
		/// the friendship proposal process.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param proposer[in]			The user ID of the user that made the proposal.
		/// \return		A pointer to the task handle that should be pumped to completion,
		///				or BD_NULL if the operation could not be performed (message 
		///				could not be sent, or maximum number of tasks was exceeded).
		/// \note		A BD_FRIENDSHIP_ACCEPT_MSG will be delivered to the
		///				original proposer, and a BD_FRIEND_STATUS_ONLINE
		///				message will be delivered to the local user if the user
		///				specified by proposer is currently online.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		///				BD_FRIENDSHIP_NOT_REQUSTED - if the potential friend never 
		///				proposed friendship.
		///				BD_FRIENDSHIP_EXISTS - if the potential friend is already
		///				in this user's friends list.
		bdRemoteTask* acceptProposal(bdUByte8 controllerIndex, 
									 bdOnlineUserID proposer);

		/// Reject a friendship proposal.
		/// This function should be invoked when the local users wishes to reject
		/// a friendship proposal message that has been received. This completes
		/// the friendship proposal process.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param proposer[in]			The user ID of the user that made the proposal.
		/// \return		A pointer to the task handle that should be pumped to completion,
		///				or BD_NULL if the operation could not be performed (message 
		///				could not be sent, or maximum number of tasks was exceeded).
		/// \note		A BD_FRIENDSHIP_REJECT_MSG will be delivered to the
		///				original proposer.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		///				BD_FRIENDSHIP_NOT_REQUSTED - if the potential friend never 
		///				proposed friendship.
		bdRemoteTask* rejectProposal(bdUByte8 controllerIndex, 
									 bdOnlineUserID proposer);

		/// Remove a user from the friends list. 
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param friendID[in]			The friend's user ID.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned in case an error occurred.
		/// \note		A BD_FRIENDSHIP_REVOKE_MSG will be delivered to the
		///				user specified by friendID.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		///				BD_NOT_A_FRIEND - if the user specified by friendID is
		///				not in this user's friend list.
		bdRemoteTask *removeFriend(bdUByte8 controllerIndex, 
								   bdOnlineUserID friendID);

		/// Retrieve the friend list.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned in case an error occurred.
		/// \note		When this task completes, you can obtain the result set by 
		///				invoking bdFriends::handleTask with an instance of 
		///				bdGetFriendListResult.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		bdRemoteTask *getFriends(bdUByte8 controllerIndex);	
		
		/// Invite a friend to a game session.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param friendID[in]			The User ID of the friend to invite.
		/// \param securityID[in]		The securityID of the game session, as used
		///								with the matchmaking service.
		/// \param attachment[in]		An optional attachment to send with the 
		///								invite message.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned if an error occurred.
		/// \note		A BD_FRIEND_SESSION_INVITATION_MSG will be delivered to the
		///				user specified by friendID.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		///				BD_NOT_A_FRIEND - if the user specified by friendID is
		///				not in this user's friend list.
		///				BD_USERID_BANNED - if the user specified by friendID has
		///				added this user blocked list.
		bdRemoteTask *inviteFriend(bdUByte8 controllerIndex, 
								   bdOnlineUserID friendID,
								   const bdSecurityID &securityID,
								   bdByteBuffer * const attachment = BD_NULL);
		
		/// Accept an invite from a friend to a game session.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param inviter[in]			User ID of the friend that sent the invite.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned in case an error occurred.
		/// \note		A BD_FRIEND_SESSION_ACCEPT_MSG will be delivered to the
		///				original inviter.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		bdRemoteTask* acceptInvitation(bdUByte8 controllerIndex, 
									   bdOnlineUserID inviter);

		/// Reject an invite from a friend to a game session.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param inviter[in]			User ID of the friend that sent the invite.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned in case an error occurred.
		/// \note		A BD_FRIEND_SESSION_REJECT_MSG will be delivered to the
		///				original inviter.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		bdRemoteTask* rejectInvitation(bdUByte8 controllerIndex,
									   bdOnlineUserID inviter);

		/// Adds/remove a user to/from the block list. 
		/// Blocking a user means that no further messages will be received from
		/// that user. For example, you can block a player from sending friend 
		/// proposals.
		/// When blocking a user, a mandatory  'blockLevel' is required. 
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param userID[in]			The user ID to be added to/removed from 
		///								the blocked list.
		/// \param blockLevel[in]		Value indicating the severity of block. 
		///								A blockLevel of 1 blocks the specified user 
		///								from sending any session invites or friendship 
		///								proposals. 
		///								A blockLevel of 2 blocks the user from 
		///								sending session invites only.
		/// \return		A pointer to the task that handles the request. BD_NULL 
		///				is returned if an error occurred.
		/// \note		A BD_FRIEND_BLOCK_MSG or BD_FRIEND_UNBLOCK_MSG will be 
		///				delivered to the user specified by userID.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		///				BD_SELF_BLOCK_NOT_ALLOWED - if the user specified by userID
		///				is also that of the caller.
		bdRemoteTask *setBlockedStatus(bdUByte8 controllerIndex, 
									   bdOnlineUserID userID,
                                       bdBlockLevel blockLevel);
		
		/// Associate a groupID with a particular friend.
		/// The groupIDs are created and associated with strings using 
		/// bdFriends::setGroupName, so that the friend list can be classified 
		/// into sections with readable names.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param friendID[in]			The user ID of the friend to be classified.
		/// \param groupID[in]			The ID of the group, which was previously
		///								created with bdFriends::setGroupName.
		/// \return		A pointer to the task that handles the request. BD_NULL is returned
		///				in case an error occurred.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		///				BD_NOT_A_FRIEND - if the user specified by friendID
		///				is not in the caller's friends list.
		bdRemoteTask *setFriendGroup(bdUByte8 controllerIndex, 
									 bdOnlineUserID friendID,
									 bdUByte8 groupID);

		/// Creates or updates a group's name.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param groupID[in]			The group ID to set.
		/// \param groupName[in]		The name to associate with groupID
		/// \return		A pointer to the task handle that should be pumped to completion,
		///				or BD_NULL if the operation could not be performed (message 
		///				could not be sent, or maximum number of tasks was exceeded).
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		bdRemoteTask *setGroupName(bdUByte8 controllerIndex, 
								   bdUByte8 groupID,
								   const bdNChar8 *groupName);

		/// Get a list of group names and IDs that have been created by this user.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note		When this task completes, you can obtain the result set by 
		///				invoking bdFriends::handleTask with an instance of 
		///				bdGetGroupNamesResult.
		/// \note		When complete, the returned bdRemoteTask will specify:
		///				BD_NO_ERROR - on success.
		bdRemoteTask *getGroupNames(bdUByte8 controllerIndex);

		/// Similar to bdFriends::getFriends, this method returns the friend list,
		/// but also includes each friend's group classification in the result set.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdFriends::handleTask with an instance of 
		///			bdGetGroupedFriendsResult.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		bdRemoteTask *getGroupedFriends(bdUByte8 controllerIndex);

		/// Remove a group name.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param groupID[in]			The ID of the group name to be removed.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		bdRemoteTask *deleteGroupName(bdUByte8 controllerIndex, 
									  bdUByte8 groupID);

		/// Get the list of blocked users and the type of block in each case.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdFriends::handleTask with an instance of 
		///			bdGetBlockListResult.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		bdRemoteTask *getBlockedList(bdUByte8 controllerIndex);
		
		/// Send a custom message to a friend.
		/// Use this method when you need to send messages that you have 
		/// designed. This is useful for message types that are specific to your
		/// title, and have not been defined by Matchmaking+.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								performs this operation.
		/// \param friendId[in]			The User ID of the friend to send the 
		///								message to.
		/// \param payload[in]			The custom message contained in a byte buffer.
		/// \param deliveryMethod[in]	An enumerated value that specifies whether 
		///								to use push or pull messaging, or both.
		///								See bdMessageDeliveryMethod.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	A BD_FRIEND_CUSTOM_MSG will be delivered to the user
		///			specified by friendId.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success.
		///			BD_NOT_A_FRIEND - if the user specified by friendID is not
		///			in the caller's friend list. 
		///			BD_USERID_BANNED - if the user specified by friendID has blocked
		///			the caller by calling bdFriends::setBlockedStatus.
		bdRemoteTask *sendFriendMessage(bdUByte8 controllerIndex, 
										bdOnlineUserID friendId,
										bdByteBuffer& payload,
										bdMessageDeliveryMethod deliveryMethod);

		/// Handle a completed task. 
		/// This retrieves the error code returned from the server, and fills out
		/// a result set of data, if appropriate. The specified task must have 
		/// been created by this class. 
		/// If the task is expected to return results, the second parameter should
		/// be specified. bdQueryResult is the base-class for classes that handle
		/// result sets returned from the server. A specialized class for each kind
		/// of result set is located in the same directory as this class.
		/// \param remoteTask[in]	A pointer to the remote task to handle.
		/// \param queryResult[in]	An optional pointer to an object that handles 
		///							the expected type of result data.
		/// \return - True if the task was successfully handled.
		///			- False otherwise. Call bdFriends::getErrorCode() to get the 
		///			  error condition.
		bdBool handleTask(const bdRemoteTask *remoteTask, 
						  bdQueryResult *queryResult = BD_NULL);

		/// Return the error code of the last operation that failed.
		/// \return		The last error code
		bdLobbyErrorCode getErrorCode() const;

	protected:

		friend class bdLobbyServiceImpl;

		/// Constructs a friend manager with the given task handler.
		/// Note: never call this constructor directly, use bdLobbyService::getFriends() instead.
		/// \param remoteTaskManager[in] A pointer to the remote task manager that handles
		/// requests and replies
		bdFriends(bdRemoteTaskManager *const remoteTaskManager);

		bdLobbyErrorCode m_errorCode;

		bdRemoteTaskManager *m_remoteTaskManager;

};

#endif // BD_FRIENDS_H
