// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// PURPOSE : Advertise and discover game session(s) on Matchmaking+.

#ifndef BD_MATCH_MAKING_H
#define BD_MATCH_MAKING_H

#include <bdCore/bdCore.h>

#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdLobbyCommon/bdCustomQueryParams.h>
#include <bdLobby/bdLobbyErrorCode.h>

class bdRemoteTask;
class bdRemoteTaskManager;
class bdMatchMakingInfo;
class bdQueryResult;
class bdQueryParams;

// MatchMaking Operations
#define BD_MM_CREATE_SESSION			(1u)
#define BD_MM_UPDATE_SESSION			(2u)
#define BD_MM_DELETE_SESSION			(3u)
#define BD_MM_FIND_SESSION_BY_ID		(4u)
#define BD_MM_FIND_SESSIONS				(5u)

/// Handles all communications to the matchmaking service.
/// Use the functions contained in this class to perform operations on the server.
/// Each of these function returns a task handle, which should be pumped to 
/// completion by calling its getStatus() function. When the status of a task is 
/// bdRemoteTask::BD_DONE, you can pass it into the bdMatchmaking::handleTask()
/// function. Once this is done, the error code for the task can be accessed via
/// bdMatchMaking::getErrorCode(). If you expect a result set to be returned from
/// the server as a result of the task you started, you must pass an instance of 
/// a bdQueryResult into bdMatchMaking::handleTask. The implementation must 
/// match the results you are expecting. Standard implementations are provided
/// for standard types of result sets. If the result set being returned contains
/// customized Matchmaking+ attributes, you will need a specialized implementation
/// of bdQueryResult.
class bdMatchMaking
{
	public:
		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		/// Destructor
		~bdMatchMaking();

		/// Advertise a game session.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								creates the session.
		/// \param params[in]			A pointer to an instance of 
		///								bdMatchMakingInfo that contains session 
		///								specific parameters. The implementation
		///								of this class must match the custom 
		///								parameters that have been configured for
		///								your title ID on Matchmaking+.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdMatchMaking::handleTask() with an instance of
		///			bdSessionCreateResult, which will then contain the security 
		///			key and security ID that has been assigned to this session. 
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_PARAM_MISMATCHED_TYPE - if the parameters specified in params
		///			do not match the schema of the onlinegames table.
		bdRemoteTask* createSession(const bdUByte8 controllerIndex,
									bdMatchMakingInfo *const params);

		/// Update an existing session.
		/// \param controllerIndex[in]	The controller index of the user who 
		///								created the session.
		/// \param params[in]			A pointer to an instance of 
		///								bdMatchMakingInfo that contains session 
		///								specific parameters. The implementation
		///								of this class must match the custom 
		///								parameters that have been configured for
		///								your title ID on Matchmaking+.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_PARAM_MISMATCHED_TYPE - if the parameters specified in params
		///			do not match the schema of the onlinegames table.
		///			BD_NO_ENTRY_TO_UPDATE - if no online game was found with a
		///			bdSecurityID matching the one specified in params.
		bdRemoteTask* updateSession(const bdUByte8 controllerIndex,
									bdMatchMakingInfo *const params);

		/// Stop advertising a session.
		/// \param controllerIndex[in]	The controller index of the user that 
		///								created the session.
		/// \param securityID[in]		The bdSecurityID of the session to be 
		///								deleted.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask* deleteSession(const bdUByte8 controllerIndex,
								    const bdSecurityID& securityID);

		/// Find the details of the online session with a given bdSecurityID.
		/// \param securityID[in]	The bdSecurityID of the session to be found.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdMatchMaking::handleTask(). An instance of a class
		///			derived from bdQueryResult, which matches the results
		///			expected from the server, should be passed into it to receive
		///			the results.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask* findSessionFromID(const bdSecurityID& securityID);

		/// Find any online sessions that match the specified parameters.
		/// \param params[in]	The search parameters for finding a session. This 
		///						should be a class derived from bdCustomQueryParams,
		///						which has been implemented to serialize the values
		///						expected by the server for this search. 
		///						Searches are identified by the the value of
		///						bdCustomQueryParams::m_queryID. You can limit the
		///						number of results that the server will return by
		///						setting the bdCustomQueryParams::m_maxNumResults member.
		///						See the MatchMaking+ Tutorials for more information.
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdMatchMaking::handleTask(). An instance of a class
		///			derived from bdQueryResult, which matches the results
		///			expected from the server, should be passed into it to receive
		///			the results.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_INVALID_QUERY_ID - if the queryID specified in params does not
		///			exist.
		///			BD_PARAM_PARSE_ERROR - if the parameters specified in the custom
		///			query are not found in the schema for the onlinegames table. 
		bdRemoteTask* findSessions(bdCustomQueryParams *const params);

		/// Handle a completed task. 
		/// This retrieves the error code returned from the server, and fills out
		/// a result set of data, if appropriate. The specified task must have 
		/// been created by this class. 
		/// If the task is expected to return results, the second parameter should
		/// be specified. bdQueryResult is the base-class for classes that handle
		/// result sets returned from the server. A class matching the standard
		/// result set for calls to bdMatchMaking::createSession is located in 
		/// the same directory as this class (bdSessionCreateResult). 
		/// The format of results for other calls is non-standard, and depends 
		/// on this title ID's Matchmaking+ configuration. Special implementations
		//// of bdQueryResult must be used in these cases.
		/// \param remoteTask[in]	A pointer to the remote task to handle.
		/// \param queryResult[in]	An optional pointer to an object that handles 
		///							the expected type of result data.
		/// \return - True if the task was successfully handled.
		///			- False otherwise. Call bdMatchMaking::getErrorCode() to get the 
		///			  error condition.
		bdBool handleTask(const bdRemoteTask *const remoteTask,
						  bdQueryResult *const queryResult = BD_NULL);

		/// Return the error code of the last operation that failed.
		/// \return The last error code
		bdLobbyErrorCode getErrorCode() const;

	protected:

		friend class bdLobbyServiceImpl;

		/// Constructor
		/// \param remoteTaskManager[in]
		bdMatchMaking(bdRemoteTaskManager *const remoteTaskManager);

		/// a utility function used internally in bdMatchMaking
		bdRemoteTask* startTask(const bdUByte8 controllerIndex,
								bdQueryParams *params,
								const bdUByte8 BD_TASK_ID);

		/// remote task manager
		bdRemoteTaskManager *m_remoteTaskManager;

		/// error code store for look up
		bdLobbyErrorCode m_errorCode;
};

#endif // BD_MATCH_MAKING_H
