// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

// Purpose: bdMatchMakingInfo contains information about online game sessions. 
// It is used to advertise game sessions on the matchmaking service, and
// to connect to them.

#ifndef BD_MATCH_MAKING_INFO_H
#define BD_MATCH_MAKING_INFO_H

#include <bdCore/bdCore.h>

#include <bdLobby/bdLobbyCommon/bdQueryParams.h>
#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdLobbyCommon/bdQueryResult.h>

/// Contains information about a single online game session.
/// It is used to create/delete/update game sessions on the Matchmaking service,
/// and to connect to hosted games.
class bdMatchMakingInfo : public bdQueryParams
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

 		/// Default Constructor
		bdMatchMakingInfo();

		/// Create an instance with the given info.
		/// \param hostAddr[in]			The address at which the game is hosted.
		/// \param publicOpen[in]		Number of public open player slots.
		/// \param privateOpen[in]		Number of private open player slots.
		/// \param publicFilled[in]		Number of filled public player slots.
		/// \param privateFilled[in]	Number of filled private player slots.
		bdMatchMakingInfo(const bdCommonAddrRef hostAddr,
						  const bdUByte8 publicOpen,
						  const bdUByte8 privateOpen,
						  const bdUByte8 publicFilled,
						  const bdUByte8 privateFilled);

		/// Virtual Destructor
		virtual ~bdMatchMakingInfo();

		/// Serialize the entire game info into a buffer.
		/// \param buffer[in|out]	The buffer into which the member variables 
		///							are serialized.
		virtual void serialize(bdBitBuffer &buffer);

		/// Deserialize the entire game info from a buffer.
		/// \param buffer[in|out]	The buffer from which the member variables 
		///							are deserialized.
		/// \return 
		///		- True, if deserialization completed successfully.
		///		- False, otherwise.
		virtual bdBool deserialize(bdBitBufferRef buffer);

		/// Get the security ID for this game session.
		/// \return		The game session's security ID.
		const bdSecurityID& getSecurityID() const;

		/// Set the security ID for this game session.
		/// \param	secID[in]	The game session's security ID.
		void setSecurityID(const bdSecurityID& secID);

		/// Get the security key for this game session.
		/// \return		The game session's security key.
		const bdSecurityKey& getSecurityKey() const;

		/// Set the security key for this game session.
		/// \param secKey[in]	The game session's security key.
		void setSecurityKey(const bdSecurityKey& secKey);

		/// Get the address at which the game session is hosted.
		/// \return		The game session host's common address.
		bdCommonAddrRef getHostAddr() const;

		/// Set the address at which the game session is hosted.
		/// \param	hostAddr[in]	The game session host's common address.
		void setHostAddr(const bdCommonAddrRef hostAddr);

		/// Get the number of public open slots in the game session.
		/// \return		The number of public open slots.
		bdUByte8 getPublicOpen() const;

		/// Set the number of public open slots in the game session.
		/// \param publicOpen[in]	The number of public open slots
		void setPublicOpen(const bdUByte8 publicOpen);

		/// Get the number of private open slots in the game session.
		/// \return The number of private open slots.
		bdUByte8 getPrivateOpen() const;

		/// Set the number of private open slots in the game session.
		/// \param privateOpen[in]	The number of private open slots
		void setPrivateOpen(const bdUByte8 privateOpen);

		/// Get the number of filled public slots in the game session.
		/// \return The number of filled public slots.
		bdUByte8 getPublicFilled() const;

		/// Set the number of filled public slots in the game session.
		/// \param publicFilled[in]		The number of filled public slots.
		void setPublicFilled(const bdUByte8 publicFilled);

		/// Get the number of filled private slots in the game session.
		/// \return The number of filled private slots.
		bdUByte8 getPrivateFilled() const;

		/// Set the number of filled private slots in the game session.
		/// \param privateFilled[in]	The number of filled private slots.
		void setPrivateFilled(const bdUByte8 privateFilled);

	protected:

		/// Security credentials
		bdSecurityID m_secID;

		bdSecurityKey m_secKey;

		/// The address at which the game is hosted
		bdCommonAddrRef m_hostAddr;

		/// number of public spaces available
		bdUByte8 m_publicOpen;

		/// number of public spaces filled
		bdUByte8 m_publicFilled;

		/// number of private spaces available
		bdUByte8 m_privateOpen;

		/// number of private spaces filled
		bdUByte8 m_privateFilled;
};

#endif // BD_MATCH_MAKING_INFO_H
