// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#ifndef BD_TITLE_UTILITIES_H
#define BD_TITLE_UTILITIES_H

#include <bdCore/bdCore.h>

#include <bdLobby/bdLobbyCommon/bdLobbyServicesCommon.h>
#include <bdLobby/bdLobbyCommon/bdQueryResult.h>
#include <bdLobby/bdLobbyErrorCode.h>
#include <bdLobby/bdTitleUtilities/bdUnitTracker.h>

class bdRemoteTask;
class bdRemoteTaskManager;

#define BD_VERIFY_STRING		(1u)
#define BD_GET_AD_UNIT			(2u)
#define BD_TRACK_IMPRESSION		(3u)
#define BD_GET_TITLE_STATS		(4u)
#define BD_RECORD_EVENT			(5u)

/// Collection of various utility functions 
class bdTitleUtilities
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS

		bdTitleUtilities(bdRemoteTaskManager *const remoteTaskManager);

		~bdTitleUtilities();

		enum bdScreenType
		{
			BD_UNKNOWN,
			BD_PAL,
			BD_NTSC,
			BD_SECAM
		};

		enum bdUISection
		{
			BD_LOBBY,
			BD_MATCHMAKING,
			BD_FRIENDS,
			BD_TEAMS,
			BD_STATS,
			BD_STORAGE,
			BD_CHANNEL_CHAT,
			BD_PROFILES
		};

		/// Specifies the format of the ad unit,
		/// see http://www.iab.net/standards/adunits.asp
		enum bdUnitLayout
		{
			BD_0x0,
			BD_300x250, 
			BD_250x250, 
			BD_240x400, 
			BD_336x280, 
			BD_180x150, 
			BD_468x60, 
			BD_234x60, 
			BD_88x31 ,
			BD_120x90, 
			BD_120x60, 
			BD_120x240, 
			BD_125x125, 
			BD_728x90,
			BD_160x600, 
			BD_120x600, 
			BD_300x600 
		};

		/// Determines whether the given string passes the vulgarity filter. 
		/// \param str[in] : specifies the string to be verified
		/// \param length[in] : specifies the length of the string
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdTitleUtilities::handleTask() with an instance of 
		///			bdVerifyStringResult.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask* verifyString(bdNChar8 *str, 
								   bdUInt length);
		
		/// Records an 'event' for a specific user, at the current time.
		/// The event is a string, and is opaque to the server, can be used for later analysis.
		/// \param eventStr[in] : specifies the event that has just occurred
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask* recordEvent(const bdNChar8 *eventStr);

		/// Retrieves general usage statistics for this title, specifically: 
		/// - current number of online sessions 
		/// - total users that have created accounts
		/// - current number of online users
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, you can obtain the result set by 
		///			invoking bdTitleUtilities::handleTask() with an instance of 
		///			bdTitleStatsResult.
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask* getTitleStats();

		/// Downloads an advertisement unit
		/// \param layout[in] : specifies the shape and size of the unit,
		///							http://www.iab.net/standards/adunits.asp
		/// \param mimeType[in] : string specifying the unit's MIME type
		/// \param language[in] : a two character language identifier, see iso639-2
		///					http://www.loc.gov/standards/iso639-2/langcodes.html
		/// \param screenType[in] : eg. PAL, NTSC, SECAM
		/// \param UISection[in] : the section of the lobby where the ad is to be
		///                           displayed, eg. friends, teams, matchmaking, etc..
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When this task completes, the result can be obtained by 
		///			invoking bdTitleUtilities::handleTask() and passing in an instance
		///			of bdAdUnit, which will contain the requested ad unit
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		///			BD_UTILITY_NO_MATCHING_UNIT - if no ad unit could be found to match
		///			the language and lobbysection parameters.
		bdRemoteTask* getAdUnit(bdUnitLayout layout, 
							    const bdNChar8 *mimeType, 
								const bdNChar8 *language,
								bdScreenType screenType,
								bdUISection UISection);

		/// Reports a collection of impressions of ad units.
		/// \param numUnits[in] : The number of bdUnitTrackers to be updated
		/// \param units[in] : A pointer to an array of bdUnitTrackers
		/// \return	A pointer to the task handle that should be pumped to completion,
		///			or BD_NULL if the operation could not be performed (message 
		///			could not be sent, or maximum number of tasks was exceeded).
		/// \note	When complete, the returned bdRemoteTask will specify:
		///			BD_NO_ERROR - on success. 
		bdRemoteTask* trackImpression(bdUWord numUnits,
									  bdUnitTracker *units);

		bdBool handleTask(const bdRemoteTask *const remoteTask,
						  bdQueryResult *const queryResult = BD_NULL);

		/// Return the error code of the last operation that failed.
		/// \return The last error code
		bdLobbyErrorCode getErrorCode() const;

	protected:

		/// remote task manager
		bdRemoteTaskManager *m_remoteTaskManager;

		/// error code store for look up
		bdLobbyErrorCode m_errorCode;


};

#endif //BD_TITLE_UTILITIES_H

