// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: WIN32 UNIX OSX PS2 PSP-INFRA PS3
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#include <bdNet/bdNet.h>
#include <bdNet/bdGetHostByName.h>

#include <bdPlatform/bdPlatformTiming/bdPlatformTiming.h>

#include <bdSocket/bdNAT/bdIPDiscoveryClient.h>
#include <bdSocket/bdNAT/bdNATTypeDiscoveryClient.h>

#include <bdConnection/bdConnection.h>
#include <bdSocket/bdSocketRouter.h>
#include <bdConnection/bdConnectionStore.h>

#define BD_LOG_LEVEL	"bdNet/net"

BD_REGISTER_SINGLETON(bdNetImpl)

bdBool bdNetImpl::start(const bdNetStartParams& params)
{
	bdBool success = true;

	if (m_status != BD_NET_STOPPED)
	{
		BD_WARN(BD_LOG_LEVEL, "bdNetImpl::start, bdNet subsystem already started");
		success = false;
	}
	else
	{
		m_UPnPFinished = false;
		m_params = params;

		bdAddr addr = getBindAddr();

		if(findFreePort(addr))
		{
			const bdBool tmpSocketCreateResult = m_tmpSocket.create();
			if(false == tmpSocketCreateResult)
			{
				BD_WARN(BD_LOG_LEVEL, "Failed to create temporary socket");
				m_status = BD_NET_BIND_FAILED;
				success = false;
			}
			else
			{
				// the tmp socket is used to keep the port free for us
				const bdSocketStatusCode bindResult = m_tmpSocket.bind(addr);
				if(BD_NET_SUCCESS != bindResult)
				{
					BD_WARN(BD_LOG_LEVEL, "Failed to bind temporary socket to addr");
					m_status = BD_NET_BIND_FAILED;
					success = false;
				}
				else
				{
					m_params.m_gamePort = addr.getPort();
					BD_INFO(BD_LOG_LEVEL, "Requested port %u, using port %u", params.m_gamePort, m_params.m_gamePort);

					bdBool socketCreateResult = true;
					bdBool keyCreateResult = true;

					if(!m_params.m_socket)
					{
						m_params.m_socket = new bdDefaultSocket;
						socketCreateResult = m_params.m_socket->create();
					}

					if (!m_ECCKey)
					{
						m_ECCKey = new bdECCKey;
						keyCreateResult = m_ECCKey->create();
					}
					
					if(false == socketCreateResult )
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to create game socket");
						m_status = BD_NET_BIND_FAILED;
						success = false;
					}
					else if (false == keyCreateResult )
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to create key ");
						m_status = BD_NET_INIT_FAILED;
						success = false;
					}
					else
					{
						// if it's an online game then start looking for our local game address (public address).
						if(m_params.m_onlineGame)
						{
							m_status = BD_NET_PENDING;
							m_currentNatTravAddrIndex = 0;
							m_currentNatTravHostIndex = 0;

							m_getHostByName = new bdGetHostByName();
							const bdBool getHostByNameStartResult = m_getHostByName->start(m_params.m_natTravHosts[m_currentNatTravHostIndex], m_params.m_timeout);
							
							if(false == getHostByNameStartResult)
							{
								BD_WARN(BD_LOG_LEVEL, "Failed to start DNS lookup task.");
								m_status = BD_NET_INIT_FAILED;
								success = false;
							}
							else
							{
								// start UPnP discovery
								m_UPnP = new bdUPnP;
								m_UPnP->setDiscoveryTimeOut(m_params.m_upnpTimeout);
								m_UPnP->startDiscovery();
								m_UPnPState = m_UPnP->getState();
								if(bdUPnP::BD_DISCOVERING == m_UPnPState)
								{
									BD_INFO(BD_LOG_LEVEL, "UPnP discovery successfully started, timeout: %u seconds", m_params.m_upnpTimeout);
								}
							}
						}
						else
						{
							m_status = BD_NET_DONE;

							const bdBool tmpSocketCloseResult = m_tmpSocket.close();
							if(false == tmpSocketCloseResult)
							{
								BD_WARN(BD_LOG_LEVEL, "Tmp socket close failed.");
							}

							// bind to the specified address and port
							if(m_params.m_socket->bind(addr) == BD_NET_SUCCESS)
							{
								// Get the local IP address
								BD_ASSERT(!m_params.m_localAddresses.isEmpty(), "No local address specified when starting bdNet. See bdNetStartParams.");

								bdArray<bdAddr> localAddresses;
								for(bdUInt i = 0; i < m_params.m_localAddresses.getSize(); i++)
								{
									const bdAddr local(m_params.m_localAddresses[i], m_params.m_gamePort);
									localAddresses.pushBack(local);
								}

								// Use an invalid public address because we are not online.
								bdAddr remAddr;

								// create the common address
								m_localCommonAddr = new bdCommonAddr(localAddresses,
																		remAddr);

								//create the socket router
								m_socketRouter = new bdSocketRouter(m_params.m_socket,
																	&m_keyStore,
																	m_ECCKey );

								// create the connection store.
								m_connectionStore = new bdConnectionStore(	m_socketRouter,
																			m_localCommonAddr,
																			&m_keyStore,
																			m_ECCKey);
							}
							else
							{
								BD_WARN(BD_LOG_LEVEL, "Failed to bind");
								m_status = BD_NET_BIND_FAILED;
							}
						}
					}
				}
			}
		}
		else
		{
			BD_WARN(BD_LOG_LEVEL, "Failed to find a free port");
			m_status = BD_NET_BIND_FAILED;
			success = false;
		}
	}

	return success;
}

void bdNetImpl::pump()
{
	if(m_params.m_onlineGame && m_status == BD_NET_PENDING)
	{
		bdBool finishedIPLookup = false;
		bdBool finishedNATLookup = false;
		const bdUInt numNATHosts = m_params.m_natTravHosts.getSize();
		bdBool finishedNATHostLookup = m_currentNatTravHostIndex == numNATHosts;

		// pump the upnp module
		if (m_UPnP && !m_UPnPFinished)
		{
			m_UPnP->pump();

			// check if the upnp module is ready for mapping
			bdUPnP::bdUPnPState upnpState = m_UPnP->getState();

			if (upnpState == bdUPnP::BD_READY)
			{
				if (m_UPnPState == bdUPnP::BD_DISCOVERING)
				{
					// discovery has finished, see if a port mapping already exists
					// for the desired port
					m_UPnP->queryMapping(m_params.m_gamePort, &m_params.m_localAddresses);
					m_UPnPState = m_UPnP->getState();
				}
				else if (m_UPnPState == bdUPnP::BD_QUERYING)
				{
					// open the game port
					m_UPnP->openPort(m_params.m_gamePort, m_params.m_localAddresses[0].getInAddr());
					m_UPnPState = m_UPnP->getState();
				}
			}
			else if (upnpState == bdUPnP::BD_COLLISION_FOUND)
			{
				// the desired port mapping already exists, pick the next one
				m_params.m_gamePort++;

				BD_INFO(BD_LOG_LEVEL, "UPnP Mapping collision found. New port Selected: %d", m_params.m_gamePort);

				// start a new query to make sure this new port has no collisions
				m_UPnP->queryMapping(m_params.m_gamePort, &m_params.m_localAddresses);
				m_UPnPState = m_UPnP->getState();
			}
			else if (upnpState == bdUPnP::BD_MAPPED)
			{
				// mapping has finished or existing usable mapping found at
				// the requested port, no further UPnP work required
				m_UPnPFinished = true;
			}
			else if (upnpState == bdUPnP::BD_FAILED)
			{
				// todo -> Why do we call getState() again here ?
				m_UPnPState = m_UPnP->getState();

				BD_WARN(BD_LOG_LEVEL, "UPnP Traversal failed");
				delete m_UPnP;
				m_UPnP = BD_NULL;
			}
			else if (upnpState == bdUPnP::BD_NONE_DISCOVERED)
			{
				// todo -> Why do we call getState() again here ?
				m_UPnPState = m_UPnP->getState();

				BD_INFO(BD_LOG_LEVEL, "no UPnP NATs found");
				delete m_UPnP;
				m_UPnP = BD_NULL;
			}
		}

		if (m_getHostByName)
		{
			bdGetHostByName::bdStatus status = m_getHostByName->getStatus();
			if (status == bdGetHostByName::BD_READY)
			{
				bdUInt numAddrs = m_getHostByName->getNumAddresses();

				for(bdUInt j = 0; j < numAddrs; j++)
				{
					const bdInetAddr inAddr( m_getHostByName->getAddressAt(j).inUn.m_iaddr);
					const bdAddr addr(inAddr, m_params.m_natTravPort);
					m_natTravAddrs.pushBack(addr);
				}
				m_currentNatTravHostIndex++;

			}
			else if (status == bdGetHostByName::BD_FAILED)
			{
				m_currentNatTravHostIndex++;
			}

			// Start a new lookup if there are still hostnames left
			if (m_currentNatTravHostIndex < numNATHosts && status != bdGetHostByName::BD_PENDING)
			{
				delete m_getHostByName;
				m_getHostByName = new bdGetHostByName();
				const bdBool getHostByNameStartResult = m_getHostByName->start(m_params.m_natTravHosts[m_currentNatTravHostIndex], m_params.m_timeout);
				if(false == getHostByNameStartResult)
				{
					BD_WARN(BD_LOG_LEVEL, "Failed to start DNS lookup task.");
					m_status = BD_NET_INIT_FAILED;
				}
			}
			if (m_currentNatTravHostIndex == numNATHosts)
			{
				delete m_getHostByName;
				m_getHostByName = BD_NULL;
				// No more addresses to lookup, now we can start the IP discovery and NAT-type
				// discovery tasks
			}
		}

		if (!m_getHostByName && (m_UPnP == BD_NULL || m_UPnPFinished) &&
			!m_ipDiscClient && !m_natTypeDiscClient)
		{
			const bdBool closedResult = m_tmpSocket.close();
			if(false == closedResult)
			{
				BD_WARN(BD_LOG_LEVEL, "Failed to close temporary socket.");
			}

			const bdAddr bindAddr = getBindAddr();

			// bind to the specified port
			const bdSocketStatusCode bindStatus = m_params.m_socket->bind(bindAddr);
			if(BD_NET_SUCCESS == bindStatus)
			{
				// start ip and nat type discovery
				if(!m_natTravAddrs.isEmpty())
				{
					bdAddr addr = m_natTravAddrs[m_currentNatTravAddrIndex];

					m_ipDiscClient = new bdIPDiscoveryClient(m_params.m_socket, addr);
					const bdBool startIPDiscoveryResult = m_ipDiscClient->startDiscovery();

					m_natTypeDiscClient = new bdNATTypeDiscoveryClient(bindAddr.getAddress(), addr);
					const bdBool startNATDiscoveryResult = m_natTypeDiscClient->startDiscovery();

					if(false == startIPDiscoveryResult )
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to start IP discovery");
						m_status = BD_NET_ONLINE_FAILED;
					}
					else if (false == startNATDiscoveryResult)
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to start NAT type discovery");
						m_status = BD_NET_ONLINE_FAILED;
					}
				}
				else
				{
					for(bdUInt i = 0; i < m_params.m_natTravHosts.getSize(); i++)
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to lookup address %s", m_params.m_natTravHosts[i].getBuffer());
					}
					m_status = BD_NET_ONLINE_FAILED;
				}
			}
			else
			{
				BD_WARN(BD_LOG_LEVEL, "Failed to bind. Status : %i " , bindStatus);
				m_status = BD_NET_BIND_FAILED;
			}
		}
		if(m_ipDiscClient)
		{
			m_ipDiscClient->pump();
			if (!m_ipDiscClient->isPending())
			{
				finishedIPLookup = true;
			}
		}

		if(m_natTypeDiscClient)
		{
			m_natTypeDiscClient->pump();
			if (!m_natTypeDiscClient->isPending())
			{
				finishedNATLookup = true;
			}
		}

		if(finishedNATHostLookup && finishedIPLookup && finishedNATLookup)
		{
			if(m_ipDiscClient->wasSuccess() && m_natTypeDiscClient->wasSuccess())
			{
				// Get the local IP address
				BD_ASSERT(!m_params.m_localAddresses.isEmpty(), "No local address specified when starting bdNet. See bdNetStartParams.");

				bdArray<bdAddr> localAddresses;
				for(bdUInt i = 0; i < m_params.m_localAddresses.getSize(); i++)
				{
					const bdAddr local(m_params.m_localAddresses[i], m_params.m_gamePort);
					localAddresses.pushBack(local);
				}

				// the public address is the address the IP discovery client found.
				bdAddr publicAddr = m_ipDiscClient->getPublicAddress();

				// check if bdUPnP has changed the effective NAT type
				bdNATType clientNAT = m_natTypeDiscClient->getNATType();
				if (m_UPnP && m_UPnP->isMapped(publicAddr.getPort()))
				{
					// UPNP external port mapped is the same that STUN reports
					// this is now effectively an open NAT
					clientNAT = BD_NAT_OPEN;
					BD_INFO(BD_LOG_LEVEL, "UPnP mapping confirmed, changing NAT type to open.");
				}

				// create the common address
				m_localCommonAddr = new bdCommonAddr(localAddresses,
													 publicAddr,
													 clientNAT);

				// clean up memory used for NAT and IP discovery.
				delete m_ipDiscClient;
				m_ipDiscClient = BD_NULL;
				delete m_natTypeDiscClient;
				m_natTypeDiscClient = BD_NULL;

				bdArray<bdAddr> introducers = m_natTravAddrs;
				// we want to put the working introducer as the first in the array. we can
				// leave the rest in the array (even if they aren't working) because
				// they might start working later and it would be a good idea to send
				// keep alives to them in that case.
				if(m_currentNatTravAddrIndex != 0)
				{
					const bdAddr tmpAddr = m_natTravAddrs[0];
					introducers[0] = m_natTravAddrs[m_currentNatTravAddrIndex];
					introducers[m_currentNatTravAddrIndex] = tmpAddr;
				}

				// create the socket router.
				m_socketRouter = new bdSocketRouter(m_params.m_socket,
													&m_keyStore,
													m_ECCKey ,
													introducers );
				// create the connection store.
				m_connectionStore = new bdConnectionStore(	m_socketRouter,
															m_localCommonAddr,
															&m_keyStore,
															m_ECCKey);

				m_status = BD_NET_DONE;
			}
			else
			{
				if(m_currentNatTravAddrIndex + 1 < m_natTravAddrs.getSize())
				{
					BD_INFO(BD_LOG_LEVEL, "Trying next IP address for IP and NAT discovery.");

					m_currentNatTravAddrIndex++;
					bdAddr addr = m_natTravAddrs[m_currentNatTravAddrIndex];

					if (m_ipDiscClient)
					{
						delete m_ipDiscClient;
					}
					m_ipDiscClient = new bdIPDiscoveryClient(m_params.m_socket, addr);
					const bdBool startIPDiscoveryResult = m_ipDiscClient->startDiscovery();

					if (m_natTypeDiscClient)
					{
						delete m_natTypeDiscClient;
					}

					const bdAddr bindAddr = getBindAddr();

					m_natTypeDiscClient = new bdNATTypeDiscoveryClient(bindAddr.getAddress(), addr);
					const bdBool startNATDiscoveryResult = m_natTypeDiscClient->startDiscovery();

					if(false == startIPDiscoveryResult )
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to start IP discovery");
						m_status = BD_NET_ONLINE_FAILED;
					}
					else if(false == startNATDiscoveryResult )
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to start IP discovery");
						m_status = BD_NET_ONLINE_FAILED;
					}
				}
				else
				{
					for(bdUInt i = 0; i < m_params.m_natTravHosts.getSize(); i++)
					{
						BD_WARN(BD_LOG_LEVEL, "Failed to do IP and NAT discovery using %s", m_params.m_natTravHosts[i].getBuffer());
					}
					m_status = BD_NET_ONLINE_FAILED;
				}
			}
		}
	}
}

void bdNetImpl::stop()
{
	if (m_UPnP)
	{
		// remove any UPnP port mappings set up by demonware
		if (m_UPnP->getNumMappings())
		{
			m_UPnP->clearPortMap();
			m_UPnP->pump();
			// to reduce total blocking time the rest of UPnP clean up
			// is placed after the 1 second delay below
		}
	}

	if(m_connectionStore)
	{
		m_connectionStore->disconnectAll();

		// final flush for messages in the connections

		bdStopwatch timer;
		timer.start();

		while(timer.getElapsedTimeInSeconds() < 1.f)
		{
			receiveAndDispatchAll();
			sendAll();
			bdPlatformTiming::sleep(20);
		}

		m_connectionStore->closeAll();

		delete m_connectionStore;
		m_connectionStore = BD_NULL;
	}

	if (m_socketRouter)
	{
		delete m_socketRouter;
		m_socketRouter = BD_NULL;
	}


	// UPnP clean up continuation
	if (m_UPnP)
	{
		while (m_UPnP->getState() == bdUPnP::BD_CLOSING)
		{
			m_UPnP->pump();
		}

		delete m_UPnP;
		m_UPnP = BD_NULL;
	}

	// Close the tmp socket, in case stop is called 
	// while the tmp socket is in use.
	const bdBool closedResult = m_tmpSocket.close();
	if(false == closedResult)
	{
		// This should not occur , as it is ok to close a closed socket again!
		BD_WARN(BD_LOG_LEVEL, "Failed to close temporary socket.");
	}

	if(BD_NULL != m_params.m_socket)
	{
		const bdBool socketCloseResult = m_params.m_socket->close();
		if(false == socketCloseResult)
		{
			BD_WARN(BD_LOG_LEVEL, "Game socket failed to close with an error");
		}
		delete m_params.m_socket;
		m_params.m_socket = BD_NULL;
	}

	if (m_ipDiscClient)
	{
		delete m_ipDiscClient;
		m_ipDiscClient = BD_NULL;
	}
	if (m_natTypeDiscClient)
	{
		delete m_natTypeDiscClient;
		m_natTypeDiscClient = BD_NULL;
	}

	m_status = BD_NET_STOPPED;
	m_currentNatTravAddrIndex = 0;
	m_currentNatTravHostIndex = 0;
	m_params.m_localAddresses.clear();
	m_params.m_natTravHosts.clear();
	m_keyStore.clear();
	m_natTravAddrs.clear();
}

bdNetImpl::bdNetStatus bdNetImpl::getStatus() const
{
	return m_status;
}

void bdNetImpl::sendAll()
{
	BD_ASSERT(m_status == BD_NET_DONE, "bdNetImpl::sendAll: not ready to send");

	const bdBool flushAllResult = m_connectionStore->flushAll();
	if(false == flushAllResult)
	{
		BD_WARN(BD_LOG_LEVEL, "Failed to flush all connections in the connection store");
	}
}

void bdNetImpl::receiveAndDispatchAll()
{
	BD_ASSERT(m_status == BD_NET_DONE, "bdNetImpl::receiveAndDispatchAll: not ready to work");

	bdConnectionRef connection;
	bdArray<bdConnectionRef> receivedConnections;

	// decouple the receiving from the dispatching
	while(m_connectionStore->receiveFrom(connection))
	{
		bdUInt index;
		// check the connection is not already there
		if(!receivedConnections.findFirst(connection, index))
		{
			receivedConnections.pushBack(connection);
		}
	}

	const bdUInt numReceivedConnections = receivedConnections.getSize();
	for(bdUInt i = 0; i < numReceivedConnections; i++)
	{
		m_dispatcher.process(receivedConnections[i]);
	}
}

void bdNetImpl::registerDispatchInterceptor(bdDispatchInterceptor *const interceptor)
{
	m_dispatcher.registerInterceptor(interceptor);
}

void bdNetImpl::unregisterDispatchInterceptor(bdDispatchInterceptor *const interceptor)
{
	m_dispatcher.unregisterInterceptor(interceptor);
}


bdSecurityKeyMap *bdNetImpl::getKeyMap()
{
	return &m_keyStore;
}

bdConnectionStore *bdNetImpl::getConnectionStore()
{
	return m_connectionStore;
}

bdCommonAddrRef bdNetImpl::getLocalCommonAddr() const
{
	// todo : should this only be available once it's valid ?
	return m_localCommonAddr;
}

const bdNetStartParams& bdNetImpl::getParams()
{
	return m_params;
}

bdNetImpl::bdNetImpl()
: m_ECCKey(BD_NULL),
  m_connectionStore(BD_NULL),
  m_socketRouter(BD_NULL),
  m_status(BD_NET_STOPPED),
  m_ipDiscClient(BD_NULL),
  m_natTypeDiscClient(BD_NULL),
  m_getHostByName(BD_NULL),
  m_currentNatTravAddrIndex(0),
  m_currentNatTravHostIndex(0),
  m_UPnP(BD_NULL),
  m_UPnPFinished(false),
  m_UPnPState(bdUPnP::BD_INITIAL)
{
}

bdNetImpl::~bdNetImpl()
{
	if (m_ECCKey)
	{
		delete m_ECCKey;
		m_ECCKey = BD_NULL;
	}
	if (m_UPnP)
	{
		delete m_UPnP;
		m_UPnP = BD_NULL;
	}
	if (m_getHostByName)
	{
		delete m_getHostByName;
		m_getHostByName = BD_NULL;
	}
	if (m_natTypeDiscClient)
	{
		delete m_natTypeDiscClient;
		m_natTypeDiscClient = BD_NULL;
	}
	if (m_ipDiscClient)
	{
		delete m_ipDiscClient;
		m_ipDiscClient = BD_NULL;
	}
	if (m_connectionStore)
	{
		delete m_connectionStore;
		m_connectionStore = BD_NULL;
	}
	if (m_socketRouter)
	{
		delete m_socketRouter;
		m_socketRouter = BD_NULL;
	}

}


bdBool bdNetImpl::findFreePort(bdAddr& addr)
{
	bdPort port = addr.getPort();
	bdBool success = false;

	for(bdUInt i = 0; i < 100; i++)
	{
		bdSocket socket;
		const bdBool socketCreateResult = socket.create();
		if(false == socketCreateResult)
		{
			BD_WARN(BD_LOG_LEVEL, "Create socket failed ");
			break;
		}
		else 
		{
			const bdBool socketSetBlockingResult = socket.setBlocking(false);
			if(false == socketSetBlockingResult)
			{
				BD_WARN(BD_LOG_LEVEL, "Socket set blocking failed ");
				break;
			}
			else 
			{

				const bdAddr tryAddr(addr.getAddress(), port);

				const bdSocketStatusCode bindResult = socket.bind(tryAddr);
				if(BD_NET_SUCCESS == bindResult)
				{
					const bdBool socketCloseResult = socket.close();
					if(false == socketCloseResult)
					{
						BD_WARN(BD_LOG_LEVEL, "Socket failed to close with an error");
					}
					success = true;
					addr = tryAddr;
					break;
				}
				else if (BD_NET_ADDRESS_IN_USE == bindResult)
				{
					// We expect the socket closing to fail if the bind failed because 
					// of the addr already being in use.
					// So we will flag success as unexpected!
					const bdBool socketCloseResult = socket.close();
					if(true == socketCloseResult)
					{
						BD_WARN(BD_LOG_LEVEL, "Socket bind failed, but subsequent close succeeded!");
					}
				}
				else
				{
					const bdBool socketCloseResult = socket.close();
					if(false == socketCloseResult)
					{
						BD_WARN(BD_LOG_LEVEL, "Socket failed to close with an error");
					}
				}
				port++;
			}
		}
	}

	return success;
}

bdAddr bdNetImpl::getBindAddr()
{
	bdAddr addr;

	if (m_params.m_localAddresses.getSize() == 1)
	{
		addr = bdAddr(m_params.m_localAddresses[0], m_params.m_gamePort);
	}
	else
	{
		addr = bdAddr("0.0.0.0", m_params.m_gamePort);
	}

	return addr;
}

bdSocketRouter* bdNetImpl::getSocketRouter()
{
	return m_socketRouter;
}
