// DWBS ------------------------------------------------------------------ DWBS
//
// PLATFORM		: ALL
// PRODUCT		: BIT_DEMON
// VISIBILITY	: PUBLIC
//
// DWBS ------------------------------------------------------------------ DWBS

#ifndef BD_SOCKET_ROUTER_H
#define BD_SOCKET_ROUTER_H

#include <bdCore/bdCore.h>
#include <bdCore/bdSocket/bdSecurityKeyMap.h>
#include <bdSocket/bdSocketConfig.h>

#if !defined BD_PLATFORM_PSP_ADHOC

#include <bdSocket/bdQoS/bdQoSProbe.h>

#endif // Not PSP-ADHOC

#if !defined BD_PLATFORM_XBOX && !defined BD_PLATFORM_PSP_ADHOC

#include <bdSocket/bdServiceBandwidthArbitrator.h>
#include <bdSocket/bdNAT/bdNATTravClient.h>
#include <bdCore/bdSocket/bdSecurityKeyMapListener.h>

#endif

class bdDTLSAssociation;
class bdPacketInterceptor;

/// Socket-like interface to NAT traversed & encrypted association.
/// This class can be used to create a NAT traversed & encrypted
/// association that can then be used send & receive data unreliably
/// between instances.
#if !defined BD_PLATFORM_XBOX && !defined BD_PLATFORM_PSP_ADHOC
class bdSocketRouter : public bdNATTravListener,
					   public bdSecurityKeyMapListener
#else
class bdSocketRouter 
#endif //#ifndef BD_PLATFORM_PSP_ADHOC
{
	public:

		BD_DECLARE_NEW_AND_DELETE_OPERATORS;

		/// Internal status of the socket router.
		enum bdSocketRouterStatus
		{
			BD_SOCKET_ROUTER_UNINITIALIZED,
			BD_SOCKET_ROUTER_INITIALIZED
		};

		/// Basic constructor.
		/// A instance created with this constructor cannot be used until 
		/// init() has been called in it.
		bdSocketRouter();

		/// Complex constructor
		/// The class can be used immediately after calling this constructor.
		/// \param socket [in] The socket instance to use.
		/// \param keyMap [in] The key map to use
		/// \param key [in] The key to use
		/// \param introducers [in] An optional list of introducer addresses.
		///							These are used for NAT traversal and are 
		///							not needed when playing a LAN game.
		bdSocketRouter(bdDefaultSocket *socket, 
						bdSecurityKeyMap *keyMap, 
						bdECCKey *key, 
						const bdArray<bdAddr> &introducers = bdArray<bdAddr>());

		/// Virtual destructor
		virtual ~bdSocketRouter();

		/// Initialize the class.
		/// If the class was created using the basic constructor or if quit()
		/// has been called then this function must be run before the class
		/// can be used.
		/// \param socket [in] The socket instance to use.
		/// \param keyMap [in] The key map to use
		/// \param key [in] The key to use
		/// \param introducers [in] An optional list of introducer addresses.
		///							These are used for NAT traversal and are 
		///							not needed when playing a LAN game.
		/// \return 
		///  - True, if the class was successfully initialized.
		///  - False otherwise.
		bdBool init(bdDefaultSocket *socket, 
					bdSecurityKeyMap *keyMap, 
					bdECCKey *key, 
					const bdArray<bdAddr> &introducers = bdArray<bdAddr>());

		/// Stops all activity in this class.
		/// This will reset the class to the uninitialized state.
		/// After this function has been called, it is safe to call init again.
		/// \return 
		///   - True if the quit was error free.
		///   - False otherwise.
		bdBool quit();

		/// Start the association setup process.
		/// This is an asymmetric process as NAT traversal can take of the 
		/// order of tens of seconds to complete or fail.
		/// \param addrHandle [in] The address to connect to.
		/// \return 
		///  - True, if the connect process was successfully started.
		///  - False otherwise.
		bdBool connect(bdAddrHandleRef &addrHandle);

		/// Get the status of an individual association.
		/// \param addrHandle [in] The address to identify the association.
		/// \return The socket status of the underlying association.
		virtual bdSocketStatus getStatus(const bdAddrHandleRef addrHandle) const;

		/// Send data to addr.
		/// \param addr[in]		The address and port to send the data to.
		/// \param data[in]		A pointer to the data. <B> See note below. </B>
		/// \param length[in]	The length of the data.
		/// \return An signed integer representing:
		///  - If positive, the amount of amount of data sent.
		///  - If negative then an error from bdSocketStatusCode.
		/// \note	The data passed to this function must be in a certain format.
		///			It is possible to send both encrypted & unencrypted data in one
		///			packet. The special formatting indicates which parts of the buffer
		///			are to be encrypted and which need not be.
		///			The <B>first 2 bytes in buffer</B> represent the size of the 
		///			encrypted portion. The size of the unencrypted data is then 
		///			determined to be the remainder of the data. See the SocketRouter 
		///			example for more information on using this class.
		virtual bdInt sendTo(const bdAddrHandleRef addrHandle, 
							 const void *data, 
							 const bdUInt length);

		/// Receive data. 
		/// Make sure size is greater than the max packet size or data will be
		/// lost.
		/// \param addr[out]	The address of the sender.
		/// \param data[out]	A pointer to a buffer into which the data read
		///						from the socket will be read. <B> See note below. </B>
		/// \return An signed integer representing:
		///  - If positive, the amount of amount of data read into buffer.
		///  - If negative then an error from bdSocketStatusCode.
		/// \note	The data received from this function is in a certain format.
		///			It is possible to transmit both encrypted & unencrypted data in one
		///			packet. The special formatting indicates how much of the buffer
		///			is encrypted and how much is not. The <B>first 2 bytes in the buffer</B>
		///			represent the size of the encrypted portion.
		///			The size of the unencrypted data is then determined to be the 
		///			remainder of the data. See the SocketRouter example for more 
		///			information on using this class.
		virtual bdInt receiveFrom(	bdAddrHandleRef &addrHandle, 
									void *data,
									const bdUInt size);

		/// Pump the assorted associations.
		/// This drives the socket router, and should be pumped every
		/// frame.
		void pump();

		/// Registers an interceptor with the router.
		/// \param interceptor[in]	The interceptor to be registered with this
		///							router.
		void registerInterceptor(bdPacketInterceptor *const interceptor);

		/// Removes an existing interceptor from the router.
		/// If the same interceptor has been registered multiple times, this 
		/// call will remove all instances.
		/// \param interceptor[in]	The interceptor to be removed from this
		///							router.
		void unregisterInterceptor(bdPacketInterceptor *const interceptor);

#if !defined BD_PLATFORM_XBOX && !defined BD_PLATFORM_PSP_ADHOC && !defined BD_PLATFORM_XENON
		// Implemented from bdNATTravListener
		// see base class description.
		virtual void onNATAddrDiscovery(bdCommonAddrRef remote, 
										const bdAddr &realAddr);

		// Implemented from bdNATTravListener
		// see base class description.
		virtual void onNATAddrDiscoveryFailed(bdCommonAddrRef remote);

		// Implemented from bdSecurityKeyMapListener
		// see base class description.
		virtual void onSecurityKeyRemove(const bdSecurityID &id);

		/// Setup the introducer array that this class should use.
		/// This sets up the introducers that are used during NAT traversal.
		/// This function will overwrite any existing introducers that are 
		/// setup.
		/// \param introducers[in]	An array of introducers addrs.
		/// \return :
		///  - True if the introducers were correctly setup.
		///  - False otherwise.
		bdBool setupIntroducers(const bdArray<bdAddr> &introducers);

#endif

#if !defined BD_PLATFORM_PSP_ADHOC
		/// Returns a pointer to the QoS prober object. The pointer is only good
		/// while this object instance is valid.
		bdQoSProbe* getQoSProber();
#endif 

#if defined BD_SOCKET_DO_BROADCAST
		/// Flush the buffer.
		/// This is only implemented on PSP in ad-hoc mode, it ensures data is sent 
		/// immediately rather than during the next frame.
		bdInt flushBuffer();
#endif // BD_SOCKET_DO_BROADCAST

	protected:

		/// Pointer to the socket used by this class.
		bdDefaultSocket *m_socket;
#if !defined BD_PLATFORM_XBOX && !defined BD_PLATFORM_PSP_ADHOC
		bdNATTravClient m_natTrav;

		bdSecurityKeyMap *m_keyMap;

		bdECCKey *m_ECCKey;

		bdServiceBandwidthArbitrator m_qosBandwidth;

		// map of real addrs to sockets.
		bdHashMap<bdAddr, bdDTLSAssociation *, bdAddrHash> m_dtls;

		bdHashMap<bdCommonAddrRef, bdAddrHandleRef> m_caToAddrMap;
#endif

#if !defined BD_PLATFORM_PSP_ADHOC
		bdQoSProbe m_qosProber;
#endif

#if defined BD_SOCKET_DO_BROADCAST
		bdUByte8 *m_buffer;
		bdUInt m_bufferOffset;
		bdPort m_destPort;
		bdInetAddr m_localAddr;
#endif // BD_SOCKET_DO_BROADCAST
		
		bdFastArray<bdPacketInterceptor*> m_interceptors;

		bdSocketRouterStatus m_status;
};
#endif // BD_SOCKET_ROUTER_H
