/*
[SCE CONFIDENTIAL]
$PSDocId$
                  Copyright (C) 2002-4 Sony Computer Entertainment America
                                                     All Rights Reserved
//  ------------------------------------------------------------------------

File: i18n.c

libeenet phase codes
-------------------------------------

    * sceEENETCTL_P_PppDead                    => template
    * sceEENETCTL_P_PppEstablish
          o sceEENETCTL_PP_ModemStarting       => template
          o sceEENETCTL_PP_ModemConnecting     => template + phone number
          o sceEENETCTL_PP_ModemConnect        => template + line speed
          o sceEENETCTL_PP_ModemDisconnecting  => template + CHAT output
          o sceEENETCTL_PP_ModemDisconnect     => template + CHAT output
    
          o sceEENETCTL_PP_Pppoe____           => template
    * sceEENETCTL_P_PppTerminate               => template
    * sceEENETCTL_P_PppAuth                    => template
    * sceEENETCTL_P_PppNetwork                 => template + line speed
    
    * sceEENETCTL_P_Dhcp____                   => template

libeenet error codes
-------------------------------------

    * sceEENETCTL_E_Dhcp____                   => template
    
    * sceEENETCTL_E_PppLinkEstab
          o sceEENETCTL_EE_ModemCannotConnect  => template + CHAT output
          o sceEENETCTL_EE_ModemHangUp         => template + CHAT output
          o sceEENETCTL_EE_ModemInit           => template + CHAT output
          o sceEENETCTL_EE_ModemBusy           => template
          o sceEENETCTL_EE_ModemUnknown        => template + CHAT output
    
          o sceEENETCTL_EE_Pppoe___            => template
    * sceEENETCTL_E_PppAuth                    => template + CHAP/PAP
    * sceEENETCTL_E_PppNetwork                 => template
    * sceEENETCTL_E_PppTerminate               => template

//  ------------------------------------------------------------------------
*/

#include <ctype.h>
#include <stdio.h>
#include <string.h>

#include <eekernel.h>

#include "../include/eznetcnf.h"
#include "../include/eznetctl_i18n.h"
#include "internal.h"


/*  --------------------------------------------------------
 *  default message localization tables and format callbacks
 *
 *  - getMsgEntry() matches an input code to the ranges [lo,hi].
 *    the first match wins. list specific codes ahead of wider ranges
 *    to achieve a "cascading" messaging system.
 *
 *  - with negative error codes, be careful to maintain lo <= hi order.
 *
 *  - supply a callback function to achieve custom message formats.
 *  --------------------------------------------------------
 */

static
ezMsgFormat_t m_errorMsgs_en[] = {
    //  DHCP errors
    //  libeenet only
    {   sceEENETCTL_E_DhcpMalformedPkt,
        sceEENETCTL_E_DhcpNoServer,
        "DHCP configuration error",
        NULL
    },

    //  PPP/PPPoE errors (cascaded)
    //  libeenet and Inet
    {   sceEENETCTL_E_PppAuth,
        sceEENETCTL_E_PppAuth,
        "User authentication error with ISP\n%s",
        cbFormatAuthMessage
    },
    {   sceEENETCTL_E_PppTerminate,
        sceEENETCTL_E_PppLinkEstab,
        "Connection error with ISP",
        NULL
    },

    //  modem errors (cascaded)
    //  libeenet and Inet
    {   sceEENETCTL_EE_ModemCannotConnect,
        sceEENETCTL_EE_ModemCannotConnect,
        "Modem cannot connect: %s",
        cbFormatPPPStatus
    },
    {   sceEENETCTL_EE_ModemHangUp,
        sceEENETCTL_EE_ModemHangUp,
        "Invalid dial-in number: %s",
        cbFormatPPPStatus
    },
    {   sceEENETCTL_EE_ModemBusy,
        sceEENETCTL_EE_ModemBusy,
        "Line busy",
        NULL
    },
    {   sceEENETCTL_EE_ModemUnknown,
        sceEENETCTL_EE_ModemInit,
        "Modem error: %s",
        cbFormatPPPStatus
    },

    //  PPPoE errors (cascaded)
    //  libeenet only
    {   sceEENETCTL_EE_PppoeGetPADT,
        sceEENETCTL_EE_PppoeGetPADT,
        "PPPoE session terminated by ISP",
        NULL
    },
    {   sceEENETCTL_EE_PppoeUnknown,
        sceEENETCTL_EE_PppoeServiceName,
        "PPPoE session error with ISP",
        NULL
    },

    //  terminator
    {   0, 0, NULL, NULL }
};

static
ezMsgFormat_t m_phaseMsgs_en[] = {
    //  DHCP phases (cascaded)
    //  libeenet only
    {   sceEENETCTL_P_DhcpBound,
        sceEENETCTL_P_DhcpBound,
        "DHCP configuration successful",
        NULL
    },
    {   sceEENETCTL_P_DhcpInit,
        sceEENETCTL_P_DhcpRebinding,
        "Requesting DHCP configuration",
        NULL
    },
    
    //  PPP/PPPoE phases (cascaded)
    //  libeenet and Inet
    {   sceEENETCTL_P_PppDead,
        sceEENETCTL_P_PppDead,
        "Connection dead",
        NULL
    },
    {   sceEENETCTL_P_PppNetwork,
        sceEENETCTL_P_PppNetwork,
        "Connection established with ISP",
        NULL
    },
    {   sceEENETCTL_P_PppEstablish,
        sceEENETCTL_P_PppAuth,
        "Establishing connection to ISP",
        NULL
    },

    //  modem phases (complete)
    //  libeenet only, except sceEENETCTL_PP_ModemConnecting
    {   sceEENETCTL_PP_ModemDisconnect,
        sceEENETCTL_PP_ModemDisconnect,
        "Line disconnected: %s",
        cbFormatPPPStatus
    },
    {   sceEENETCTL_PP_ModemConnecting,
        sceEENETCTL_PP_ModemConnecting,
        "Dialing %s",
        cbFormatPhoneNumber
    },
    {   sceEENETCTL_PP_ModemConnect,
        sceEENETCTL_PP_ModemConnect,
        "Line speed %s",
        cbFormatLineSpeed
    },
    {   sceEENETCTL_PP_ModemDisconnecting,
        sceEENETCTL_PP_ModemDisconnecting,
        "Disconnecting line: %s",
        cbFormatPPPStatus
    },
    {   sceEENETCTL_PP_ModemStarting,
        sceEENETCTL_PP_ModemStarting,
        "Initializing modem",
        NULL
    },

    //  PPPoE phases (cascaded)
    //  libeenet only
    {   sceEENETCTL_PP_PppoeSession,
        sceEENETCTL_PP_PppoeSession,
        "PPPoE session established",
        NULL
    },
    {   sceEENETCTL_PP_PppoeInit,
        sceEENETCTL_PP_PppoeClosing,
        "Establishing PPPoE session",
        NULL
    },

    //  terminator
    {   0, 0, NULL, NULL }
};

static
ezMsgFormat_t m_stateMsgs_en[] = {
    //  generic network states (complete)
    //  libeenet and Inet
    {   sceEENETCTL_S_DETACHED,
        sceEENETCTL_S_DETACHED,
        "Network hardware is detached",
        NULL
    },
    {   sceEENETCTL_S_STARTING,
        sceEENETCTL_S_STARTING,
        "Connecting to network",
        NULL
    },
    {   sceEENETCTL_S_RETRYING,
        sceEENETCTL_S_RETRYING,
        "Waiting to retry",
        NULL
    },
    {   sceEENETCTL_S_STARTED,
        sceEENETCTL_S_STARTED,
        "Connected to network",
        NULL
    },
    {   sceEENETCTL_S_STOPPING,
        sceEENETCTL_S_STOPPING,
        "Disconnecting from network",
        NULL
    },
    {   sceEENETCTL_S_STOPPED,
        sceEENETCTL_S_STOPPED,
        "Disconnected from network",
        NULL
    },
    {   sceEENETCTL_S_DETACHING,
        sceEENETCTL_S_DETACHING,
        "Network hardware is detaching",
        NULL
    },

    //  terminator
    {   0, 0, NULL, NULL }
};


/*  --------------------------------------------------------
 *  message table indirection and search function
 *  --------------------------------------------------------
 */

static ezMsgFormat_t *m_errorMsgs = m_errorMsgs_en;
static ezMsgFormat_t *m_phaseMsgs = m_phaseMsgs_en;
static ezMsgFormat_t *m_stateMsgs = m_stateMsgs_en;

void
ezNetCtlSwapMessageTables(ezMsgFormat_t *errorMsgs,
                          ezMsgFormat_t *phaseMsgs,
                          ezMsgFormat_t *stateMsgs)
{
    m_errorMsgs = errorMsgs ? errorMsgs : m_errorMsgs_en;
    m_phaseMsgs = phaseMsgs ? phaseMsgs : m_phaseMsgs_en;
    m_stateMsgs = stateMsgs ? stateMsgs : m_stateMsgs_en;
}

static ezMsgFormat_t *
getMsgEntry(int val, ezMsgFormat_t *msgList) 
{
    ezMsgFormat_t *e;

    for (e = msgList; e->msg != NULL; e++) {
        if (val >= e->lo && val <= e->hi) {
            return e;
        }
    }

    dprintf("no match for %d\n", val);
    return NULL;
}


/*  --------------------------------------------------------
 *  message localization function
 *
 *  input:  ezNetCtlStatus_t *s
 *          ezNetCtlIfstat_u *ifstat
 *  output: char *msgBuf
 *
 *  Note: make allowances for recreated ezNetCtlIfstat_u values.
 *  --------------------------------------------------------
 */

void
ezNetCtlLocalizeMessage(char *str, size_t size,
                        ezNetCtlStatus_t *s, ezNetCtlIfstat_u *ifstat)
{
    int phase = 0;
    ezMsgFormat_t *msgFormat = NULL;

#if (DUMP_IFSTAT)    
    static ezNetCtlIfstat_u lastIfstat;

    switch (s->iftype) {
    case sceEENET_IFTYPE_ETHER:
        if (lastIfstat.dhcp.phase != ifstat->dhcp.phase) {
            dprintf("dhcp phase %d\n", ifstat->dhcp.phase);
        }
        break;

    case sceEENET_IFTYPE_PPP:
        if (memcmp(&lastIfstat.ppp, &ifstat->ppp, sizeof(lastIfstat.ppp))) {
            dprintf("ppp phase %d, modem phase %d\n", 
                    ifstat->ppp.ppp_phase,
                    ifstat->ppp.modem_phase);
            dprintf("status  : %s\n", ifstat->ppp.status_string);
            dprintf("phone   : %s\n", ifstat->ppp.phone_number);
            dprintf("result  : %s\n", ifstat->ppp.last_string);
            dprintf("CHAP/PAP: %s\n", ifstat->ppp.auth_message);
        }
        break;

    case sceEENET_IFTYPE_PPPOE:
        if (memcmp(&lastIfstat.pppoe, &ifstat->pppoe, sizeof(lastIfstat.pppoe))) {
            dprintf("ppp phase %d, pppoe phase %d\n", 
                    ifstat->pppoe.ppp_phase,
                    ifstat->pppoe.pppoe_phase);
            dprintf("CHAP/PAP: %s\n", ifstat->pppoe.auth_message);
        }
        break;
    }
    memcpy(&lastIfstat, ifstat, sizeof(lastIfstat));
#endif

    if (str == NULL || size == 0) {
        return;
    }

    /*
     *  check if phase information is available for stateful connections
     */

    switch (s->iftype) {
    case sceEENET_IFTYPE_ETHER:
        phase = ifstat->dhcp.phase;
        break;

    case sceEENET_IFTYPE_PPP:
        phase = (ifstat->ppp.ppp_phase == sceEENETCTL_P_PppEstablish) ?
            ifstat->ppp.modem_phase :
            ifstat->ppp.ppp_phase;
        break;

    case sceEENET_IFTYPE_PPPOE:
        phase = (ifstat->pppoe.ppp_phase == sceEENETCTL_P_PppEstablish) ?
            ifstat->pppoe.pppoe_phase :
            ifstat->pppoe.ppp_phase;
        break;
    }

    /*
     *  select a message format entry using the following criteria:
     */

    //  phase info is only useable between S_STARTING and S_STARTED
    if (phase &&
        s->state >= sceEENETCTL_S_STARTING &&
        s->state <= sceEENETCTL_S_STARTED)
    {
        msgFormat = getMsgEntry(phase, m_phaseMsgs);
    }

    //  when there is an error, override the phase description
    if (s->error)
    {
        msgFormat = getMsgEntry(s->error, m_errorMsgs);
    }

    //  as a last resort, use the generic state description
    if (!msgFormat)
    {
        msgFormat = getMsgEntry(s->state, m_stateMsgs);
    }

    /*
     *  format the resulting status message
     */

    memset(str, 0, size);

    if (msgFormat) {
        if (msgFormat->callback)
            msgFormat->callback(str, size, msgFormat->msg, s, ifstat);
        else
            strncpy(str, msgFormat->msg, size);
    }
}


/*  --------------------------------------------------------
 *  formatter callbacks used by the default message tables
 *  --------------------------------------------------------
 */

void
cbFormatAuthMessage(char *str, size_t size, const char *tmpl, 
                    ezNetCtlStatus_t *s, ezNetCtlIfstat_u *ifstat)
{
    if (s->iftype == sceEENET_IFTYPE_PPP) {
        snprintf( str, size, tmpl, ifstat->ppp.auth_message );
    } else 
    if (s->iftype == sceEENET_IFTYPE_PPPOE) {
        snprintf( str, size, tmpl, ifstat->pppoe.auth_message );
    }
}

void
cbFormatPPPStatus(char *str, size_t size, const char *tmpl, 
                  ezNetCtlStatus_t *s, ezNetCtlIfstat_u *ifstat)
{
    char *cp = ifstat->ppp.status_string;

    // remove "ABORT" unfriendliness
    if (!strncmp(cp, "ABORT", 5))
        cp += 6;
    snprintf( str, size, tmpl, cp );
}

void
cbFormatLineSpeed(char *str, size_t size, const char *tmpl, 
                  ezNetCtlStatus_t *s, ezNetCtlIfstat_u *ifstat)
{
    char *cp = ifstat->ppp.last_string;
    int len;

    // remove trailing whitespaces
    len = strlen( cp );
    while ( len && isspace(cp[len-1]) )
        cp[--len] = '\0';
    snprintf( str, size, tmpl, cp );
}

void
cbFormatPhoneNumber(char *str, size_t size, const char *tmpl, 
                    ezNetCtlStatus_t *s, ezNetCtlIfstat_u *ifstat)
{
    snprintf( str, size, tmpl, ifstat->ppp.phone_number );
}

