//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Alien_Attack.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/Alien/AVP_Behaviour_Alien_Attack.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#endif

// avp
#include "Actor/Vertebrates/Alien/AVP_Alien_MeleeCounter_System.h"
#include "Animation/AVP_AnimFlags.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Controller/AVP_ServerEntity_Squad_Alien.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_Vertebrate_Messages.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Alien/AVP_ClientActor_Alien.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Alien/AVP_ServerActor_Alien.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesMelee.h"

// avp axon
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Behaviour/BehaviourTypes/Alien/AVP_Behaviour_Alien_Movement.h"
#include "Core/BrainTypes/AVP_Brain_Actor_Alien.h"
#include "Knowledge/AVP_Knowledge_System_AlienTeam.h"

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Alien_Attack
//
//////////////////////////////////////////////////////////////////////////////

static const float fMIN_ATTACK_RANGE_JUMP = 5.0f;
static const float fMAX_ATTACK_RANGE_JUMP = 6.0f;
static const float fMIN_ATTACK_RANGE_JUMP_BURST = 8.0f;
static const float fMAX_ATTACK_RANGE_JUMP_BURST = 10.0f;

static const float fMIN_ATTACK_RANGE_STAND = 1.75f;
static const float fMAX_ATTACK_RANGE_STAND = 4.75f;
static const u_int uMIN_CYCLES_ROUND = 3;
static const u_int uMAX_CYCLES_ROUND = 5;

AVP_Behaviour_Alien_Attack::AVP_Behaviour_Alien_Attack( AVP_Brain_Actor_Alien* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_ALIEN_ATTACK )
, m_uMeleeType( AVP_ALIEN_MELEEATTACK_TYPE_PRIMARYATTACK )
, m_uInitialMeleeMoveHash( ASURA_HASH_ID_UNSET )
, m_eMeleeHandedness( NONE )
, m_fAttackRangeMin( 0.0f )
, m_fAttackRangeMax( 0.0f )
, m_uNumRemainingAttacks( 0 )
, m_uNumRemainingCyclesRound( 0 )
{
	SetCurrentState( UNSET );
}

bool AVP_Behaviour_Alien_Attack::Activate()
{
	if ( !PARENT::Activate() ) { return false; }
	SwitchToState( RUNNING_IN );
	m_uNumRemainingCyclesRound = Asura_Random::GetIntU( uMIN_CYCLES_ROUND, uMAX_CYCLES_ROUND );
	return true;
}

void AVP_Behaviour_Alien_Attack::Deactivate()
{
	SwitchToState( UNSET );
	PARENT::Deactivate();
}

void AVP_Behaviour_Alien_Attack::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	if ( fTimeslice == 0.0f ) { return; }

	UpdateStateTimer( fTimeslice );

	if (!GetAlienBrain()->CanPlanMove())
	{
		return;
	}

	const AVP_AlienTeam_Knowledge* pxTargetKnowledge = GetAlienBrain()->GetCurrentTargetKnowledge();
	if ( !pxTargetKnowledge )
	{
		SwitchToState( FINISHED );
		return;
	}

	switch ( GetCurrentState() )
	{
		case RUNNING_IN:
		{

			const bool bAttemptingJumpSlashAttack = ( m_uMeleeType == AVP_ALIEN_MELEEATTACK_TYPE_JUMPSLASH );
			const float fTargetDistance = GetAlienBrain()->GetCurrentTargetDistance();

			AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
			if ( xMovement.IsInError() )
			{
				SetInError();
				SwitchToState( FINISHED );
			}
			else if ( xMovement.IsFinished() )
			{
				// Should have kicked off a grapple/melee by now - we'll have to cycle round.
				if ( m_uNumRemainingCyclesRound )
				{
					SwitchToState( CYCLING_ROUND );
				}
				else
				{
					SwitchToState( FINISHED );
				}
			}
			else if ( fTargetDistance < m_fAttackRangeMin )
			{
				//Asura_OutputToDebugger_Retail( ( "%u: Too near to attack now - giving up.\n", GetBrain()->GetOwnerGuid() ) );
				// We've got too near - we'll have to cycle round.
				if ( m_uNumRemainingCyclesRound )
				{
					SwitchToState( CYCLING_ROUND );
				}
				else
				{
					SwitchToState( FINISHED );
				}
			}
			else if ( m_fAttackRangeMin < fTargetDistance && fTargetDistance < m_fAttackRangeMax )
			{
				const bool bCanAttack = GetAlienBrain()->CurrentTargetWithinAttackFOV() && GetAlienBrain()->HaveLOSToCurrentTarget();
				if ( bCanAttack )
				{
					if ( bAttemptingJumpSlashAttack )
					{
						SwitchToState( REQUESTED_JUMP_ATTACK );
					}
					else if ( GetAlienBrain()->GetActor()->GetGroundType() == AVP_Actor_PhysicsObjectHandler::ON_FLOOR )
					{
						SwitchToState( REQUESTED_STAND_ATTACK );
					}
					else
					{
						//Asura_OutputToDebugger_Retail( ( "%u: Not on floor for stand attack.\n", GetBrain()->GetOwnerGuid() ) );
					}
				}
				else if ( bAttemptingJumpSlashAttack )
				{
					// If we're down for the jump attack and are just about to run out of range see if we can switch to the stand attack.
					if ( ( ( fTargetDistance - m_fAttackRangeMin ) < 0.2f ) )
					{
						if ( EndOfRouteAllowsStandAttack() )
						{
							//Asura_OutputToDebugger_Retail( ( "%u: No FOV/LOS for jump attack - changing attack to stand.\n", GetBrain()->GetOwnerGuid() ) );
							SetupStandSlashAttack();
						}
						else
						{
							//Asura_OutputToDebugger_Retail( ( "%u: No FOV/LOS for jump attack - nearly too close.\n", GetBrain()->GetOwnerGuid() ) );
						}
					}
					else
					{
						//Asura_OutputToDebugger_Retail( ( "%u: No FOV/LOS for jump attack.\n", GetBrain()->GetOwnerGuid() ) );
					}
				}
				else
				{
					//Asura_OutputToDebugger_Retail( ( "%u: No FOV/LOS for stand attack.\n", GetBrain()->GetOwnerGuid() ) );
				}
			}
			else if ( GetAlienBrain()->ShouldRefreshPath((pxTargetKnowledge->GetPosition()- GetAlienBrain()->GetNavData().m_xDestinationPosition).MagnitudeSquared() ) )
			{
				xMovement.GetToPosition( pxTargetKnowledge->GetPosition(), 0.0f, AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM );
				if ( !bAttemptingJumpSlashAttack )
				{
					// Need to check again that the end of our route allows the stand attack.
					if ( !EndOfRouteAllowsStandAttack() )
					{
						SetupJumpSlashAttack();
					}
				}
			}

			break;
		}

		case REQUESTED_JUMP_ATTACK:
		case REQUESTED_STAND_ATTACK:
		{
			// We're waiting for the move start message to notify us we've started our melee move.
			// Hardcoded timeout in case this doesn't arrive.
			if ( GetStateTimer() > 0.5f )
			{
				if ( m_uNumRemainingCyclesRound )
				{
					SwitchToState( CYCLING_ROUND );
				}
				else
				{
					SwitchToState( FINISHED );
				}
			}
			break;
		}

		case STARTED_JUMP_ATTACK:
		{
			// Sitting and waiting for the move chain complete message.
			if ( GetStateTimer() > 3.0f )
			{
				// Hardcoded timeout in case, for whatever reason, our anims break.
				if ( m_uNumRemainingCyclesRound )
				{
					SwitchToState( CYCLING_ROUND );
				}
				else
				{
					SwitchToState( FINISHED );
				}
			}

			break;
		}

		case STARTED_STAND_ATTACK:
		case STARTED_LOOP_STAND_ATTACK:
		{
			// The allow new melee flag lets us know we are at the end of the melee.
			// Use this to either trigger a looped attack or exit slightly early so we can continue moving smoothly.
			const Asura_Move* pxMove = GetAlienBrain()->GetClientActor()->GetMovementSystem()->GetCurrentMove();
			const u_int uMoveAnimFlags = pxMove ? pxMove->GetAnimFlagsForLastFrame() : 0;
			if ( ( uMoveAnimFlags & AVP_ANIMFLAG_ALLOW_NEW_MELEE ) != 0 )
			{
				const bool bCanDoMoreAttacks = ( m_uNumRemainingAttacks &&
					( GetAlienBrain()->GetCurrentTargetDistance() < m_fAttackRangeMax ) &&
					GetAlienBrain()->CurrentTargetWithinAttackFOV() );
				if ( bCanDoMoreAttacks )
				{
					SwitchToState( STARTED_LOOP_STAND_ATTACK );
				}
				else
				{
					if ( m_uNumRemainingCyclesRound )
					{
						SwitchToState( CYCLING_ROUND );
					}
					else
					{
						SwitchToState( FINISHED );
					}
				}
			}
			else
			{
				// Hardcoded timeout.
				if ( GetStateTimer() > 3.0f )
				{
					if ( m_uNumRemainingCyclesRound )
					{
						SwitchToState( CYCLING_ROUND );
					}
					else
					{
						SwitchToState( FINISHED );
					}
				}
			}

			break;
		}

		case CYCLING_ROUND:
		{
			// If we're a little way away and we can attack again then we should.
			if ( GetAlienBrain()->GetCurrentTargetDistance() > 7.5f )
			{
				SwitchToState( RUNNING_IN );
			}
			else
			{
				// If we've finished cycling we should attack again.
				const Axon_Behaviour* pxCyclingChild = GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIEN_CYCLEROUNDIMMEDIATE );
				if ( !pxCyclingChild || pxCyclingChild->IsInError() )
				{
					SetInError();
					SwitchToState( FINISHED );
				}
				else if ( pxCyclingChild->IsFinished() )
				{
					SwitchToState( RUNNING_IN );
				}
			}
			break;
		}

		case COUNTERED:
		{
			// Sitting and waiting for the move chain complete or abort message.
			if ( GetStateTimer() > 6.0f )
			{
				SwitchToState( FINISHED );
			}

			break;
		}

		case UNSET:
		case FINISHED:
		default:
		{
			break;
		}
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT
const char* AVP_Behaviour_Alien_Attack::s_aszStateStrings[ NUM_STATES ] =
{
	"UNSET",
	"RUNNING_IN",
	"REQUESTED_JUMP_ATTACK",
	"STARTED_JUMP_ATTACK",
	"REQUESTED_STAND_ATTACK",
	"STARTED_STAND_ATTACK",
	"STARTED_LOOP_STAND_ATTACK",
	"CYCLING_ROUND",
	"COUNTERED",
	"FINISHED"
};

void AVP_Behaviour_Alien_Attack::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "Alien Attack: %s", s_aszStateStrings[ GetCurrentState() ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

bool AVP_Behaviour_Alien_Attack::HandleMessage( const Asura_Message* pxMessage )
{
	switch ( GetCurrentState() )
	{
		case REQUESTED_JUMP_ATTACK:
		{
			// Waiting for the move start message to notify us that we have started our melee.
			if ( pxMessage->GetMsgID() == ASURA_MSG_MOVE_START )
			{
				const AVP_Msg_Move* pxData = static_cast< const AVP_Msg_Move* >( pxMessage->GetData() );
				m_uInitialMeleeMoveHash = pxData->GetMoveID();
				SwitchToState( STARTED_JUMP_ATTACK );
				return true;
			}
			break;
		}

		case REQUESTED_STAND_ATTACK:
		{
			// Waiting for the move start message to notify us that we have started our melee.
			if ( pxMessage->GetMsgID() == ASURA_MSG_MOVE_START )
			{
				const AVP_Msg_Move* pxData = static_cast< const AVP_Msg_Move* >( pxMessage->GetData() );
				m_uInitialMeleeMoveHash = pxData->GetMoveID();
				SwitchToState( STARTED_STAND_ATTACK );
				return true;
			}
			break;
		}

		case STARTED_JUMP_ATTACK:
		case STARTED_STAND_ATTACK:
		case STARTED_LOOP_STAND_ATTACK:
		{
			if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE )
			{
				const Asura_Msg_MoveChainComplete* pxData = static_cast< const Asura_Msg_MoveChainComplete* >( pxMessage->GetData() );
				if ( pxData->GetInitialMoveID() == m_uInitialMeleeMoveHash )
				{
					// If our chain is complete then so are we.
					if ( m_uNumRemainingCyclesRound )
					{
						SwitchToState( CYCLING_ROUND );
					}
					else
					{
						SwitchToState( FINISHED );
					}
					return true;
				}
			}
			else if ( pxMessage->GetMsgID() == AVP_MSGID_COUNTER_MELEE )
			{
				// If we've been countered then we shouldn't bother doing any more melees or cycle round.
				SwitchToState( COUNTERED );
				return true;
			}
			break;
		}

		case COUNTERED:
		{
			if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE )
			{
				const Asura_Msg_MoveChainComplete* pxData = static_cast< const Asura_Msg_MoveChainComplete* >( pxMessage->GetData() );
				if ( pxData->GetInitialMoveID() == m_uInitialMeleeMoveHash )
				{
					if ( GetAlienBrain()->GetCurrentTargetDistance() < fMAX_ATTACK_RANGE_STAND )
					{
						SetupStandSlashAttack();
						SwitchToState( REQUESTED_STAND_ATTACK );
					} 
					else
					{
						SwitchToState( FINISHED );
					}

					return true;
				}
			}
			break;
		}
	}

	return PARENT::HandleMessage( pxMessage );
}

AVP_Brain_Actor_Alien* AVP_Behaviour_Alien_Attack::GetAlienBrain() const
{
	return static_cast< AVP_Brain_Actor_Alien* >( GetBrain() );
}

void AVP_Behaviour_Alien_Attack::RequestMelee( const u_int uMeleeType )
{
	if ( !GetAlienBrain()->GetCurrentTargetKnowledge() )
	{
		Asura_Assert( false, "AVP_Behaviour_Alien_Attack::RequestMelee - we have tested this already!" );
		return;
	}

	//Asura_OutputToDebugger( ( "AVP_Behaviour_Alien_Attack::RequestMelee - requesting melee (%08x)\n", uMeleeType ) );

	// Send a request melee message.
	Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_REQUEST_MELEE, sizeof( AVP_Message_Vertebrate_RequestMelee ) );
	if ( pxMessage )
	{
		AVP_Message_Vertebrate_RequestMelee* pxData = static_cast< AVP_Message_Vertebrate_RequestMelee* >( pxMessage->GetData() );

		pxData->SetType( uMeleeType );
		pxData->SetTargetGuid( GetAlienBrain()->GetCurrentTargetKnowledge()->GetGuid() );

		pxMessage->SetToGuid( GetAlienBrain()->GetOwnerGuid() );
		pxMessage->SetFromGuid( GetAlienBrain()->GetOwnerGuid() );
		Asura_Network::Send( pxMessage );
	}
}

void AVP_Behaviour_Alien_Attack::SwitchToState( const State eState )
{
	//Asura_OutputToDebugger( ( "AVP_Behaviour_Alien_Attack::SwitchToState - switching to %s\n", s_aszStateStrings[ eState ] ) );

	switch ( GetCurrentState() )
	{
		case RUNNING_IN:
		{
			GetAlienBrain()->GetMovement().ResetIgnoreTargetForAvoidCalcuation();
			break;
		}

		case STARTED_LOOP_STAND_ATTACK:
		case STARTED_STAND_ATTACK:
		{
			if ( m_uNumRemainingAttacks == 0 )
			{
				GetAlienBrain()->GetAlienActor()->GetAlienAIAnimStateMachine()->RequestIdle();
			}
			break;
		}

		case CYCLING_ROUND:
		{
			DestroyChildWithType( AVP_BEHAVIOUR_TYPE_ALIEN_CYCLEROUNDIMMEDIATE );
			break;
		}

		case COUNTERED:
		case UNSET:
		case STARTED_JUMP_ATTACK:
		case REQUESTED_JUMP_ATTACK:
		case REQUESTED_STAND_ATTACK:
		{
			break;
		}

		case FINISHED:
		{
			SetFinished( false );
			break;
		}
	}

    SetCurrentState( eState );
	ResetStateTimer();
	switch ( GetCurrentState() )
	{
		case UNSET:
		{
			// We don't want to leave the state machine in the stand idle state by mistake.
// 			if ( GetBrain()->IsActive() )
// 			{
// 				GetAlienBrain()->GetAlienActor()->GetAlienAIAnimStateMachine()->RequestIdle();
// 			}
			break;
		}

		case RUNNING_IN:
		{
			const AVP_AlienTeam_Knowledge* pxTargetKnowledge = GetAlienBrain()->GetCurrentTargetKnowledge();

			AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
			if ( pxTargetKnowledge && xMovement.GetToPosition( pxTargetKnowledge->GetPosition(), 0.0f,  AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM ) )
			{
				GetAlienBrain()->GetMovement().SetIgnoreTargetForAvoidCalcuation(pxTargetKnowledge->GetGuid());
			}
			else
			{
				// If we can't get to our target then we should ignore it for a while to see if we can target someone else.
				GetAlienBrain()->GetMovement().ResetIgnoreTargetForAvoidCalcuation();
				GetAlienBrain()->IgnoreTarget( GetAlienBrain()->GetCurrentTargetGuid(), 5.0f );
				SetInError( true );
				SwitchToState( FINISHED );
			}

            const bool bJumpAttackValid  = GetAlienBrain()->GetMovement().IsBursting() ?  
				( GetAlienBrain()->GetCurrentTargetDistance() > fMIN_ATTACK_RANGE_JUMP_BURST ) :  ( GetAlienBrain()->GetCurrentTargetDistance() > fMIN_ATTACK_RANGE_JUMP );
			const bool bStandAttackValid = ( GetAlienBrain()->GetCurrentTargetDistance() > fMIN_ATTACK_RANGE_STAND ) && EndOfRouteAllowsStandAttack();
			if ( bJumpAttackValid && bStandAttackValid )
			{
				if ( Asura_Random::GetFloat() < GetAlienBrain()->GetAlienSquad()->GetJumpVsStandFraction() )
				{
					SetupStandSlashAttack();
				}
				else
				{
					SetupJumpSlashAttack();
				}
			}
			else if ( bJumpAttackValid )
			{
				SetupJumpSlashAttack();
			} 
			else if ( bStandAttackValid )
			{
				SetupStandSlashAttack();
			} 
			else
			{
				SetupStandSlashAttack();
			}

			break;
		}

		case REQUESTED_JUMP_ATTACK:
		{
			// If we're on the wall request the wall versions of our attack.
			//const Asura_AI_NavVertex* pxCurrentWP = Asura_AI_Navigation_System::GetWaypoint( GetAlienBrain()->GetNavData().m_usCurrentWPID );
			//if ( pxCurrentWP && pxCurrentWP->IsWallWaypoint() )
			//{
			//	m_uMeleeType = AVP_ALIEN_MELEEATTACK_TYPE_AI_JUMPSLASHWALL;
			//}
			GetAlienBrain()->GetMovement().Stop();
			RequestMelee( m_uMeleeType );
			break;
		}

		case REQUESTED_STAND_ATTACK:
		{
			GetAlienBrain()->GetMovement().Stop();
			RequestMelee( m_uMeleeType );
			GetAlienBrain()->GetAlienActor()->GetAlienAIAnimStateMachine()->RequestStandIdle();
			break;
		}

		case STARTED_LOOP_STAND_ATTACK:
		{
			--m_uNumRemainingAttacks;

			Asura_Assert( m_eMeleeHandedness != NONE, "AVP_Behaviour_Alien_Attack::SwitchToState - bad melee handedness" );
			if ( m_eMeleeHandedness == LEFT )
			{
				RequestMelee( AVP_ALIEN_MELEEATTACK_TYPE_AI_STANDSLASHLOOP_R );
				m_eMeleeHandedness = RIGHT;
			}
			else // if ( m_eMeleeHandedness == RIGHT )
			{
				RequestMelee( AVP_ALIEN_MELEEATTACK_TYPE_AI_STANDSLASHLOOP_L );
				m_eMeleeHandedness = LEFT;
			}

			break;
		}

		case CYCLING_ROUND:
		{

			if (!GetAlienBrain()->CanAttackCurrentTarget())
			{
				SwitchToState( FINISHED );
			}
			else 
			{
				--m_uNumRemainingCyclesRound;

				if ( !CreateChildWithType( AVP_BEHAVIOUR_TYPE_ALIEN_CYCLEROUNDIMMEDIATE ) )
				{
					SwitchToState( FINISHED );
				}
			}

			break;
		}

		case STARTED_JUMP_ATTACK:
		case STARTED_STAND_ATTACK:
		case COUNTERED:
		{
			break;
		}

		case FINISHED:
		{
			SetFinished( true );
			break;
		}
	}
}

bool AVP_Behaviour_Alien_Attack::EndOfRouteAllowsStandAttack() const
{
	// Work backwards along the route and determine whether we are on the floor until at least fMIN_ATTACK_RANGE_STAND away.
	const float fDistanceRequiredSq = ( fMIN_ATTACK_RANGE_STAND * fMIN_ATTACK_RANGE_STAND );

	const AVP_Navigation::NavData& xNavData = GetAlienBrain()->GetNavData();
	if ( xNavData.m_usRouteLength == 0 )
	{
		return false;
	}
	else if ( xNavData.m_usRouteLength == 1 )
	{
		// Must be in the same waypoint as the target.
		const Asura_AI_NavVertex* pxWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ 0 ] );

		if ( pxWP->IsCeilingWaypoint() || pxWP->IsWallWaypoint() )
		{
			return false;
		}

		return ( GetAlienBrain()->GetCurrentTargetDistance() > fMIN_ATTACK_RANGE_STAND );
	}
	else
	{
		float fDistanceSq = 0.0f;
		for ( u_short usWPIndex = xNavData.m_usRouteLength - 1; usWPIndex > 0; --usWPIndex )
		{
			const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usWPIndex ] );
			const Asura_AI_NavVertex* pxToWP   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usWPIndex - 1 ] );

			if ( pxFromWP->IsCeilingWaypoint() || pxFromWP->IsWallWaypoint() )
			{
				return false;
			}
			if ( pxToWP->IsCeilingWaypoint() || pxToWP->IsWallWaypoint() )
			{
				return false;
			}

			fDistanceSq += ( pxToWP->GetPosition() - pxFromWP->GetPosition() ).MagnitudeSquared();
			if ( fDistanceSq > fDistanceRequiredSq )
			{
				return true;
			}
		}

		return false;
	}
}

void AVP_Behaviour_Alien_Attack::SetupStandSlashAttack()
{
	const float fRand = Asura_Random::GetFloat();
	if ( fRand < 0.5f )
	{
		m_uMeleeType = AVP_ALIEN_MELEEATTACK_TYPE_AI_CRAWLTOSTANDSLASH_R;
		m_eMeleeHandedness = RIGHT;
		m_fAttackRangeMin = fMIN_ATTACK_RANGE_STAND;
		m_fAttackRangeMax = fMAX_ATTACK_RANGE_STAND;
		m_uNumRemainingAttacks = Asura_Random::GetIntU( 1, 2 );
	}
	else
	{
		m_uMeleeType = AVP_ALIEN_MELEEATTACK_TYPE_AI_CRAWLTOSTANDSLASH_L;
		m_eMeleeHandedness = LEFT;
		m_fAttackRangeMin = fMIN_ATTACK_RANGE_STAND;
		m_fAttackRangeMax = fMAX_ATTACK_RANGE_STAND;
		m_uNumRemainingAttacks = Asura_Random::GetIntU( 1, 2 );
	}
}

void AVP_Behaviour_Alien_Attack::SetupJumpSlashAttack()
{
	m_uMeleeType = AVP_ALIEN_MELEEATTACK_TYPE_JUMPSLASH;
	m_eMeleeHandedness = NONE;
	if (GetAlienBrain()->GetMovement().IsBursting())
	{
		m_fAttackRangeMin = fMIN_ATTACK_RANGE_JUMP_BURST;
		m_fAttackRangeMax = fMAX_ATTACK_RANGE_JUMP_BURST;
	}
	else
	{
		m_fAttackRangeMin = fMIN_ATTACK_RANGE_JUMP;
		m_fAttackRangeMax = fMAX_ATTACK_RANGE_JUMP;
	}
	m_uNumRemainingAttacks = 0;
}

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Alien_JumpAttack
//
//////////////////////////////////////////////////////////////////////////////

AVP_Behaviour_Alien_JumpAttack::AVP_Behaviour_Alien_JumpAttack( AVP_Brain_Actor_Alien* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_ALIEN_JUMPATTACK )
{
	SetCurrentState( UNSET );
}

bool AVP_Behaviour_Alien_JumpAttack::Activate()
{
	if ( !PARENT::Activate() ) { return false; }
	SwitchToState( RUNNING_IN );
	return true;
}

void AVP_Behaviour_Alien_JumpAttack::Deactivate()
{
	SwitchToState( UNSET );
	PARENT::Deactivate();
}

void AVP_Behaviour_Alien_JumpAttack::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	UpdateStateTimer( fTimeslice );

	if (!GetAlienBrain()->CanPlanMove())
	{
		return;
	}

	const AVP_AlienTeam_Knowledge* pxTargetKnowledge = GetAlienBrain()->GetCurrentTargetKnowledge();
	if ( !pxTargetKnowledge )
	{
		SwitchToState( FINISHED );
		return;
	}

	switch ( GetCurrentState() )
	{
		case RUNNING_IN:
		{
			// If we are in range and facing the target then attack.
			const float fTargetDistance = GetAlienBrain()->GetCurrentTargetDistance();
			const bool bIsBursting = GetAlienBrain()->GetMovement().IsBursting();
			float fMinAttackRange = bIsBursting ? fMIN_ATTACK_RANGE_JUMP_BURST : fMIN_ATTACK_RANGE_JUMP;
			float fMaxAttackRange = bIsBursting ? fMAX_ATTACK_RANGE_JUMP_BURST : fMAX_ATTACK_RANGE_JUMP;

			if ( fTargetDistance < fMinAttackRange )
			{
				// We've got too near - we'll have to cycle round.
				SwitchToState( FINISHED );
			}
			else if ( fMinAttackRange < fTargetDistance && fTargetDistance < fMaxAttackRange )
			{
				if ( GetAlienBrain()->CurrentTargetWithinAttackFOV() && GetAlienBrain()->HaveLOSToCurrentTarget() )
				{
					SwitchToState( STARTED_JUMP_ATTACK );
				}
			}
			else
			{
				AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
				if ( xMovement.IsInError() || xMovement.IsFinished() )
				{
					// Should have kicked off a grapple/melee by now - we'll have to cycle round.
					SwitchToState( FINISHED );
				}
				else
				{
					if ( GetAlienBrain()->ShouldRefreshPath((pxTargetKnowledge->GetPosition()- GetAlienBrain()->GetNavData().m_xDestinationPosition).MagnitudeSquared() ) )
					{			
						xMovement.GetToPosition( pxTargetKnowledge->GetPosition(), 0.0f, AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM );
					}
				}
			}

			break;
		}

		case STARTED_JUMP_ATTACK:
		{
			// Sitting and waiting for the move chain complete message.
			if ( GetStateTimer() > 3.0f )
			{
				// Hardcoded timeout in case, for whatever reason, our anims break.
				SetInError( true );
				SwitchToState( FINISHED );
			}

			break;
		}

		case COUNTERED:
		{
			// Sitting and waiting for the move chain complete or abort message.
			if ( GetStateTimer() > 6.0f )
			{
				SwitchToState( FINISHED );
			}

			break;
		}

		case UNSET:
		case FINISHED:
		default:
		{
			break;
		}
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT
const char* AVP_Behaviour_Alien_JumpAttack::s_aszStateStrings[ NUM_STATES ] =
{
	"UNSET",
	"RUNNING_IN",
	"STARTED_JUMP_ATTACK",
	"COUNTERED",
	"FINISHED"
};

void AVP_Behaviour_Alien_JumpAttack::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "Alien Jump Attack: %s", s_aszStateStrings[ GetCurrentState() ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

bool AVP_Behaviour_Alien_JumpAttack::HandleMessage( const Asura_Message* pxMessage )
{
	if ( GetCurrentState() == COUNTERED )
	{
		if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE || pxMessage->GetMsgID() == ASURA_MSG_MOVE_ABORT )
		{
			SwitchToState( FINISHED );
			return true;
		}
	}
	else if ( GetCurrentState() != UNSET )
	{
		if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE )
		{
			// If our chain is complete then so are we.
			SwitchToState( FINISHED );
			return true;
		}
		else if ( pxMessage->GetMsgID() == AVP_MSGID_COUNTER_MELEE )
		{
			// If we've been countered then we shouldn't bother doing any more melees or cycle round.
			SwitchToState( COUNTERED );
			return true;
		}
	}

	return PARENT::HandleMessage( pxMessage );
}

AVP_Brain_Actor_Alien* AVP_Behaviour_Alien_JumpAttack::GetAlienBrain() const
{
	return static_cast< AVP_Brain_Actor_Alien* >( GetBrain() );
}

void AVP_Behaviour_Alien_JumpAttack::RequestJumpAttackMelee()
{
	if ( !GetAlienBrain()->GetCurrentTargetKnowledge() )
	{
		Asura_Assert( false, "AVP_Behaviour_Alien_JumpAttack::RequestMelee - we have tested this already!" );
		return;
	}

	// Send a request melee message.
	Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_REQUEST_MELEE, sizeof( AVP_Message_Vertebrate_RequestMelee ) );
	if ( pxMessage )
	{
		AVP_Message_Vertebrate_RequestMelee* pxData = static_cast< AVP_Message_Vertebrate_RequestMelee* >( pxMessage->GetData() );

		pxData->SetType( AVP_ALIEN_MELEEATTACK_TYPE_JUMPSLASH );
		pxData->SetTargetGuid( GetAlienBrain()->GetCurrentTargetKnowledge()->GetGuid() );

		pxMessage->SetToGuid( GetAlienBrain()->GetOwnerGuid() );
		pxMessage->SetFromGuid( GetAlienBrain()->GetOwnerGuid() );
		Asura_Network::Send( pxMessage );
	}
}

void AVP_Behaviour_Alien_JumpAttack::SwitchToState( const State eState )
{
	switch ( GetCurrentState() )
	{
		case RUNNING_IN:
		{
			GetAlienBrain()->GetMovement().ResetIgnoreTargetForAvoidCalcuation();
			break;
		}

		case UNSET:
		case STARTED_JUMP_ATTACK:
		case COUNTERED:
		{
			break;
		}

		case FINISHED:
		{
			SetFinished( false );
			break;
		}
	}

    SetCurrentState( eState );
	ResetStateTimer();
	switch ( GetCurrentState() )
	{
		case UNSET:
		case COUNTERED:
		{
			break;
		}

		case RUNNING_IN:
		{
			const AVP_AlienTeam_Knowledge* pxTargetKnowledge = GetAlienBrain()->GetCurrentTargetKnowledge();
			if (pxTargetKnowledge)
			{
				AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
				if ( xMovement.GetToPosition( pxTargetKnowledge->GetPosition(), 0.0f, AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM ) )
				{
					GetAlienBrain()->GetMovement().SetIgnoreTargetForAvoidCalcuation(pxTargetKnowledge->GetGuid());					
				}
				else
				{
					SetInError( true );
					SwitchToState( FINISHED );
				}
			}
			else
			{
				SetInError( true );
				SwitchToState( FINISHED );
			}

			break;
		}

		case STARTED_JUMP_ATTACK:
		{
			GetAlienBrain()->GetMovement().Stop();
			RequestJumpAttackMelee();
			break;
		}

		case FINISHED:
		{
			SetFinished( true );
			break;
		}
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Alien_StandAttack
//
//////////////////////////////////////////////////////////////////////////////

AVP_Behaviour_Alien_StandAttack::AVP_Behaviour_Alien_StandAttack( AVP_Brain_Actor_Alien* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_ALIEN_STANDATTACK )
, m_uMeleeType( AVP_ALIEN_MELEEATTACK_TYPE_PRIMARYATTACK )
, m_eMeleeHandedness( NONE )
, m_uNumRemainingAttacks( 0 )
{
	SetCurrentState( UNSET );
}

bool AVP_Behaviour_Alien_StandAttack::Activate()
{
	if ( !PARENT::Activate() ) { return false; }
	SwitchToState( RUNNING_IN );
	return true;
}

void AVP_Behaviour_Alien_StandAttack::Deactivate()
{
	SwitchToState( UNSET );
	PARENT::Deactivate();
}

void AVP_Behaviour_Alien_StandAttack::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	UpdateStateTimer( fTimeslice );

	if (!GetAlienBrain()->CanPlanMove())
	{
		return;
	}

	const AVP_AlienTeam_Knowledge* pxTargetKnowledge = GetAlienBrain()->GetCurrentTargetKnowledge();
	if ( !pxTargetKnowledge )
	{
		SwitchToState( FINISHED );
		return;
	}

	switch ( GetCurrentState() )
	{
		case RUNNING_IN:
		{
			// If we are in range and facing the target then attack.
			const float fTargetDistance = GetAlienBrain()->GetCurrentTargetDistance();
			if ( fTargetDistance < fMIN_ATTACK_RANGE_STAND )
			{
				// We've got too near - we'll have to cycle round.
				SwitchToState( FINISHED );
			}
			else if ( fMIN_ATTACK_RANGE_STAND < fTargetDistance && fTargetDistance < fMAX_ATTACK_RANGE_STAND )
			{
				if ( GetAlienBrain()->CurrentTargetWithinAttackFOV() && GetAlienBrain()->HaveLOSToCurrentTarget() )
				{
					SwitchToState( STARTED_STAND_ATTACK );
				}
			}
			else
			{
				AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
				if ( xMovement.IsInError() || xMovement.IsFinished() )
				{
					// Should have kicked off a grapple/melee by now - we'll have to cycle round.
					SwitchToState( FINISHED );
				}
				else
				{
					if ( GetAlienBrain()->ShouldRefreshPath((pxTargetKnowledge->GetPosition()- GetAlienBrain()->GetNavData().m_xDestinationPosition).MagnitudeSquared() ) )
					{
						xMovement.GetToPosition( pxTargetKnowledge->GetPosition(), 0.0f, AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM );
					}
				}
			}

			break;
		}

		case STARTED_STAND_ATTACK:
		case STARTED_LOOP_STAND_ATTACK:
		{
			// The allow new melee flag lets us know we are at the end of the melee.
			// Use this to either trigger a looped attack or exit slightly early so we can continue moving smoothly.
			const Asura_Move* pxMove = GetAlienBrain()->GetClientActor()->GetMovementSystem()->GetCurrentMove();
			const u_int uMoveAnimFlags = pxMove ? pxMove->GetAnimFlagsForLastFrame() : 0;
			if ( ( uMoveAnimFlags & AVP_ANIMFLAG_ALLOW_NEW_MELEE ) != 0 )
			{
				const bool bCanDoMoreAttacks = ( m_uNumRemainingAttacks &&
					( GetAlienBrain()->GetCurrentTargetDistance() < fMAX_ATTACK_RANGE_STAND ) &&
					GetAlienBrain()->CurrentTargetWithinAttackFOV() );
				if ( bCanDoMoreAttacks )
				{
					SwitchToState( STARTED_LOOP_STAND_ATTACK );
				}
				else
				{
					SwitchToState( FINISHED );
				}
			}
			else
			{
				// Hardcoded timeout.
				if ( GetStateTimer() > 3.0f )
				{
					SwitchToState( FINISHED );
				}
			}

			break;
		}

		case COUNTERED:
		{
			// Sitting and waiting for the move chain complete or abort message.
			if ( GetStateTimer() > 6.0f )
			{
				SwitchToState( FINISHED );
			}

			break;
		}

		case UNSET:
		case FINISHED:
		default:
		{
			break;
		}
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT
const char* AVP_Behaviour_Alien_StandAttack::s_aszStateStrings[ NUM_STATES ] =
{
	"UNSET",
	"RUNNING_IN",
	"STARTED_STAND_ATTACK",
	"STARTED_LOOP_STAND_ATTACK",
	"COUNTERED",
	"FINISHED"
};

void AVP_Behaviour_Alien_StandAttack::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "Alien Stand Attack: %s", s_aszStateStrings[ GetCurrentState() ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

bool AVP_Behaviour_Alien_StandAttack::HandleMessage( const Asura_Message* pxMessage )
{
	if ( GetCurrentState() == COUNTERED )
	{
		if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE || pxMessage->GetMsgID() == ASURA_MSG_MOVE_ABORT )
		{
			SwitchToState( FINISHED );
			return true;
		}
	}
	else if ( GetCurrentState() != UNSET )
	{
		if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE )
		{
			// If our chain is complete then so are we.
			SwitchToState( FINISHED );
			return true;
		}
		else if ( pxMessage->GetMsgID() == AVP_MSGID_COUNTER_MELEE )
		{
			// If we've been countered then we shouldn't bother doing any more melees or cycle round.
			SwitchToState( COUNTERED );
			return true;
		}
	}

	return PARENT::HandleMessage( pxMessage );
}

AVP_Brain_Actor_Alien* AVP_Behaviour_Alien_StandAttack::GetAlienBrain() const
{
	return static_cast< AVP_Brain_Actor_Alien* >( GetBrain() );
}

void AVP_Behaviour_Alien_StandAttack::RequestMelee( const u_int uMeleeType )
{
	if ( !GetAlienBrain()->GetCurrentTargetKnowledge() )
	{
		Asura_Assert( false, "AVP_Behaviour_Alien_StandAttack::RequestMelee - we have tested this already!" );
		return;
	}

	// Send a request melee message.
	Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_REQUEST_MELEE, sizeof( AVP_Message_Vertebrate_RequestMelee ) );
	if ( pxMessage )
	{
		AVP_Message_Vertebrate_RequestMelee* pxData = static_cast< AVP_Message_Vertebrate_RequestMelee* >( pxMessage->GetData() );

		pxData->SetType( uMeleeType );
		pxData->SetTargetGuid( GetAlienBrain()->GetCurrentTargetKnowledge()->GetGuid() );

		pxMessage->SetToGuid( GetAlienBrain()->GetOwnerGuid() );
		pxMessage->SetFromGuid( GetAlienBrain()->GetOwnerGuid() );
		Asura_Network::Send( pxMessage );
	}
}

void AVP_Behaviour_Alien_StandAttack::SwitchToState( const State eState )
{
	switch ( GetCurrentState() )
	{
		case UNSET:
		case COUNTERED:
		{
			break;
		}

		case RUNNING_IN:
		{
			GetAlienBrain()->GetMovement().ResetIgnoreTargetForAvoidCalcuation();
			break;
		}

		case STARTED_LOOP_STAND_ATTACK:
		case STARTED_STAND_ATTACK:
		{
			if ( m_uNumRemainingAttacks == 0 )
			{
				GetAlienBrain()->GetAlienActor()->GetAlienAIAnimStateMachine()->RequestIdle();
			}
			break;
		}

		case FINISHED:
		{
			SetFinished( false );
			break;
		}
	}

    SetCurrentState( eState );
	ResetStateTimer();
	switch ( GetCurrentState() )
	{
		case UNSET:
		case COUNTERED:
		{
			break;
		}

		case RUNNING_IN:
		{
			const AVP_AlienTeam_Knowledge* pxTargetKnowledge = GetAlienBrain()->GetCurrentTargetKnowledge();
			if (pxTargetKnowledge)
			{
				AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
				if ( xMovement.GetToPosition( pxTargetKnowledge->GetPosition(), 0.0f, AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM ) )
				{
					GetAlienBrain()->GetMovement().SetIgnoreTargetForAvoidCalcuation(pxTargetKnowledge->GetGuid());
				}
				else
				{
					SetInError( true );
					SwitchToState( FINISHED );
				}

				const float fRand = Asura_Random::GetFloat();
				if ( fRand < 0.5f )
				{
					m_uMeleeType = AVP_ALIEN_MELEEATTACK_TYPE_AI_CRAWLTOSTANDSLASH_R;
					m_eMeleeHandedness = RIGHT;
					m_uNumRemainingAttacks = Asura_Random::GetIntU( 1, 2 );
				}
				else
				{
					m_uMeleeType = AVP_ALIEN_MELEEATTACK_TYPE_AI_CRAWLTOSTANDSLASH_L;
					m_eMeleeHandedness = LEFT;
					m_uNumRemainingAttacks = Asura_Random::GetIntU( 1, 2 );
				}

			}
			else
			{
				SetInError( true );
				SwitchToState( FINISHED );
			}
			break;
		}

		case STARTED_STAND_ATTACK:
		{
			GetAlienBrain()->GetMovement().Stop();
			RequestMelee( m_uMeleeType );
			break;
		}

		case STARTED_LOOP_STAND_ATTACK:
		{
			--m_uNumRemainingAttacks;

			Asura_Assert( m_eMeleeHandedness != NONE, "AVP_Behaviour_Alien_Attack::SwitchToState - bad melee handedness" );
			if ( m_eMeleeHandedness == LEFT )
			{
				RequestMelee( AVP_ALIEN_MELEEATTACK_TYPE_AI_STANDSLASHLOOP_R );
				m_eMeleeHandedness = RIGHT;
			}
			else // if ( m_eMeleeHandedness == RIGHT )
			{
				RequestMelee( AVP_ALIEN_MELEEATTACK_TYPE_AI_STANDSLASHLOOP_L );
				m_eMeleeHandedness = LEFT;
			}

			break;
		}

		case FINISHED:
		{
			SetFinished( true );
			break;
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
