//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Alien_Wander.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/Alien/AVP_Behaviour_Alien_Wander.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "Network/Asura_Message.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#endif

// avp
#include "Actor/Vertebrates/Alien/AVP_Alien_AIAnimStateMachine.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Moves/AVP_Movement_System.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesJump.h"

// axon
#include "Behaviour/BehaviourTypes/Alien/AVP_Behaviour_Alien_Movement.h"
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Core/AVP_Navigation.h"
#include "Core/BrainTypes/AVP_Brain_Actor_Alien.h"


AVP_Behaviour_Alien_Wander::AVP_Behaviour_Alien_Wander( AVP_Brain_Actor_Alien* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_ALIEN_WANDER )
, m_xStartPosition(pxBrain->GetActor()->GetPosition())
, m_usStartWPID(ASURA_INVALID_WAYPOINT_ID)
, m_fTimeToWait(0.0f)
{
	SetCurrentState( UNSET );
	AVP_Navigation::NavData &xNavData = GetAlienBrain()->GetNavData();
	const AVP_Navigation::TestParams xTestParams( xNavData.m_uPathfinderFlags, ASURA_NAVVERTEX_MAX_ZONE_ID );
	if ( GetAlienBrain()->IsInCurrentWaypoint() ) 
	{
		m_usStartWPID = GetAlienBrain()->GetNavData().m_usCurrentWPID;
	}
	else if ( !AVP_Navigation::GetNearestWaypointID( m_xStartPosition, m_usStartWPID, xTestParams ) )
	{
		Asura_Assert(false, "Invalid start position");
	}
}

void AVP_Behaviour_Alien_Wander::Deactivate()
{
 	SwitchToState(UNSET);
    PARENT::Deactivate();
}

DEBUGVAR float dbg_fChanceToWait = 0.05f;

void AVP_Behaviour_Alien_Wander::Update( const float fTimeslice )
{
    PARENT::Update( fTimeslice );

    UpdateStateTimer( fTimeslice );

	if (!GetAlienBrain()->CanPlanMove())
	{
		return;
	}

    switch ( GetCurrentState() )
    {
	case UNSET:
		SwitchToState(WANDERING);
		break;
    case WAITING:
		if ( GetStateTimer() > m_fTimeToWait )
		{
			SwitchToState( WANDERING );
		}
		break;
    case WANDERING:
		{
			AVP_Alien_Movement& xMovement = GetAlienBrain()->GetMovement();
			if ( xMovement.IsInError() || xMovement.IsFinished()  )
			{
				if (Asura_Random::GetFloat() < dbg_fChanceToWait)
				{
					SwitchToState( WAITING );
				}
				else
				{
					SwitchToState( WANDERING );
				}
			}
		}
		break;
	default:
		Asura_Assert(false, "Unexpected State");
		break;
    }
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
	"UNSET",
	"WAITING",
	"WANDERING"
};

void AVP_Behaviour_Alien_Wander::RenderDebugInfo()
{
    PARENT::RenderDebugInfo();

    Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "Alien Wander: %s", s_aszStateStrings[ GetCurrentState() ] );
    GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

DEBUGVAR float dbg_fMinTimeToWait = 1.0f;
DEBUGVAR float dbg_fMaxTimeToWait = 2.0f;

void AVP_Behaviour_Alien_Wander::SwitchToState( const State eNextState )
{
    switch ( GetCurrentState() )
    {
	case UNSET:
	case WANDERING:
	case WAITING:
		break;
	default:
		Asura_Assert(false, "Unexpected State");
		break;
	}

    SetCurrentState( eNextState );
    ResetStateTimer();
    switch ( GetCurrentState() )
    {
	case UNSET:
		break;
	case WANDERING:
		PlanPath();
		break;
	case WAITING:
		m_fTimeToWait = Asura_Random::GetFloat(dbg_fMinTimeToWait, dbg_fMaxTimeToWait);
		GetAlienBrain()->GetMovement().Stop();
		break;
	default:
		Asura_Assert(false, "Unexpected State");
		break;
    }
}

AVP_Brain_Actor_Alien* AVP_Behaviour_Alien_Wander::GetAlienBrain() const
{
    return static_cast< AVP_Brain_Actor_Alien* >( GetBrain() );
}

DEBUGVAR float dbg_fMinWanderRAnge = 3.0f;
DEBUGVAR float dbg_fMaxWanderRAnge = 10.0f;

void AVP_Behaviour_Alien_Wander::PlanPath()
{

	AVP_Navigation::NavData xNavData;
	GetAlienBrain()->GetNavData().CopyConfigTo(xNavData);

	xNavData.m_xStartPosition = m_xStartPosition;
	xNavData.m_usCurrentWPID = m_usStartWPID;

	const Asura_AI_NavVertex* pxWP = Asura_AI_Navigation_System::GetWaypoint(m_usStartWPID);
	if (!pxWP)
	{
		SwitchToState(WAITING);
		return;
	}
	const Asura_Vector_3 &xUp = pxWP->GetNormal();
	Asura_Vector_3 xStartDir;
	// pick up some random direction in the plane.
	if (xUp.ApproximatelyEquals(Asura_Vector_3(1.0f, 0.0f, 0.0f))) 
	{
		xStartDir = xUp ^ Asura_Vector_3(0.0f, 0.0f, 1.0f);
	}
	else
	{
		xStartDir = xUp ^ Asura_Vector_3(1.0f, 0.0f, 0.0f);
	}

	// rotate it by random amount
	Asura_Quat xTurn;
	float fTurnAngle = Asura_Maths::Pi * Asura_Random::GetFloat(-1.0f,1.0f);
	xTurn.FromRotation(fTurnAngle, xUp);
	const Asura_Vector_3 xDirection = xTurn.RotateVector(xStartDir);

	AVP_Navigation::AlienAwayFromInDirFunctor xAwayFunctor(m_xStartPosition, Asura_Random::GetFloat(dbg_fMinWanderRAnge, dbg_fMaxWanderRAnge), xDirection);
	
	if(!AVP_Navigation::FindRouteAwayFrom( xNavData, &xAwayFunctor, false ) || xNavData.m_usRouteLength == 0)
	{
		SwitchToState(WAITING);
		return;
	}

	const Asura_Vector_3 xDestination = 
		Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[xNavData.m_usRouteLength-1] )->GetPosition();

	if ( !GetAlienBrain()->GetMovement().GetToPosition( xDestination, AVP_Alien_Movement::GetMaxArrivalRange(), AVP_Navigation::s_xAlienAStarFunctor, AVP_Alien_Movement::EMM_MEDIUM, false ) )
	{
		SwitchToState(WAITING);
		return;
	}
	else
	{
		GetAlienBrain()->GetMovement().SetMovementMode(AVP_Alien_Movement::EMM_MEDIUM);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
