///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Behaviour_AlienQueen_Stalk.cpp
// Author : David Wooldridge
// Date   : 13/11/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/AlienQueen/AVP_Behaviour_AlienQueen_Stalk.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#endif

// avp
#include "Actor/Vertebrates/AVP_Vertebrate_Description.h"
#include "Actor/Vertebrates/AlienQueen/AVP_AlienQueen_AIAnimStateMachine.h"
#include "Core/AVP_Navigation.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h" 

// axon
#include "Behaviour/BehaviourTypes/AlienQueen/AVP_Behaviour_AlienQueen_Movement.h"
#include "Behaviour/BehaviourTypes/AlienQueen/AVP_Behaviour_AlienQueen_Attack.h"
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Core/AVP_Navigation.h"
#include "Core/BrainTypes/AVP_Brain_Actor_AlienQueen.h"


DEBUGVAR float dbg_fMaxToRecalcRoute = 5.0f;

AVP_Behaviour_AlienQueen_Stalk::AVP_Behaviour_AlienQueen_Stalk( AVP_Brain_Actor_AlienQueen* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_STALK )
, m_xLastUpdatePosition(Asura_Maths::ZeroVector3)
, m_eCurrentState( WAIT )
, m_fTimeInState( 0.0f )
, m_fLastChargeRouteCheckTimer( 0.0f )
, m_fTimeSinceLastAcidSpitAttack( 0.0f )
, m_bIsCenterValid(false)
{
	// default for this mode is fast movement. 
	GetAlienQueenBrain()->SetMovementMode(AVP_Navigation::MM_FAST);
}

void AVP_Behaviour_AlienQueen_Stalk::Deactivate()
{
	// Destroy any children.
	Axon_Behaviour* pxChildBehaviour = GetChild();
	while( pxChildBehaviour )
	{
		UnlinkChild( pxChildBehaviour );
		GetProjectBrain()->DestroyBehaviour( pxChildBehaviour );
		pxChildBehaviour = GetChild();
	}

	PARENT::Deactivate();
}

void AVP_Behaviour_AlienQueen_Stalk::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	m_fTimeInState += fTimeslice;
	m_fTimeSinceLastAcidSpitAttack += fTimeslice;

	m_fLastChargeRouteCheckTimer -= fTimeslice;

	if( ! GetAlienQueenBrain() || ! GetAlienQueenBrain()->HasCurrentTarget() )
	{
		// if we dont have a brain or a target then we need to get out of here fast. 
		SwitchToState( WAIT );
		SetFinished();
		return;
	}

	switch( m_eCurrentState )
	{
		////////////////////
		case WAIT:
		////////////////////
		{
			if( GetAlienQueenBrain()->GetLookAtGuid() != ASURA_GUID_UNREGISTERED )
			{
				// if we have an entity we have to look at look at it
				GetAlienQueenBrain()->LookAtEntityFromMsgData();
			}

			// If an attack option is available, launch an attack
			if( ShouldLaunchAttack() )
			{
				SwitchToState( ATTACKING );
			}
			else
			{
				if( GetAlienQueenBrain() && GetAlienQueenBrain()->GetTargetKnowledge() )
				{
					if( GetAlienQueenBrain()->GetTargetKnowledge()->m_uSourceClassification != AVP_ENTITYCLASS_ALIEN )
					{
						// if we have a knowlegd system and we know the player is NOT an alien then follow them. 
						SwitchToState( FOLLOW_TARGET );
					}
				}
			}
			break;
		}
		////////////////////
		case FOLLOW_TARGET:
		////////////////////
		{
			// How far has the target moved from his last desitination if enough recalc route
			if( GetAlienQueenBrain() && GetAlienQueenBrain()->GetTargetKnowledge() )
			{
				float fDist = ( m_xLastUpdatePosition - GetAlienQueenBrain()->GetTargetKnowledge()->m_xPosition ).MagnitudeSquared();

				// Too far away recalc route
				if( fDist > dbg_fMaxToRecalcRoute )
				{
					SwitchToState( FOLLOW_TARGET );
				}

				if( GetAlienQueenBrain()->GetTargetKnowledge()->m_uSourceClassification == AVP_ENTITYCLASS_ALIEN )
				{
					// if we are following the player and they are an alien then we shouldnt be, switch to waiting state. 
					SwitchToState( WAIT );
				}
				else
				{
					// If an attack option is available, launch an attack
					if( ShouldLaunchAttack() )
					{
						SwitchToState( ATTACKING );
					}
					else
					{
						AVP_Behaviour_AlienQueen_Movement* pxBehaviour = static_cast<AVP_Behaviour_AlienQueen_Movement*>( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT ) );
						if( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished() )
						{
							// We have reached our target but cant attack them, so finish and let the wander move us closer
							SetFinished();
							SwitchToState( WAIT );
						}
					}
				}
			}
			else
			{
				AVP_Behaviour_AlienQueen_Movement* pxBehaviour = static_cast<AVP_Behaviour_AlienQueen_Movement*>( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT ) );
				if( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished() )
				{
					// We have reached our target
					SetFinished();
					SwitchToState( WAIT );
				}
			}
			break;
		}
		////////////////////
		case TARGET_OUT_OF_REACH:
		////////////////////
		{
			break;
		}
		////////////////////
		case ATTACKING:
		////////////////////
		{
			AVP_Behaviour_AlienQueen_Attack* pxBehaviour = static_cast<AVP_Behaviour_AlienQueen_Attack*>( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_ATTACK ) );
			if( !pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished() )
			{
				// Finished the attack
				SwitchToState( WAIT );
			}
			break;
		}
	}
}

void AVP_Behaviour_AlienQueen_Stalk::SwitchToState( const State eNextState )
{
	switch( m_eCurrentState )
	{
		////////////////////
		case WAIT:
		////////////////////
		{
			break;
		}
		////////////////////
		case FOLLOW_TARGET:
		////////////////////
		{
			DestroyChildWithType( AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT );
			break;
		}
		////////////////////
		case TARGET_OUT_OF_REACH:
		////////////////////
		{
			break;
		}
		////////////////////
		case ATTACKING:
		////////////////////
		{
			DestroyChildWithType( AVP_BEHAVIOUR_TYPE_ALIENQUEEN_ATTACK );
			break;
		}
		////////////////////
		default:
		////////////////////
		{
			break;
		}
	}

	m_eCurrentState = eNextState;
	m_fTimeInState = 0.0f;
	
	switch( m_eCurrentState )
	{
		////////////////////
		case WAIT:
		////////////////////
		{
			//GetAlienQueenBrain()->GetActor()->GetAIAnimStateMachine()->RequestIdle();
			GetAlienQueenBrain()->SetMovementMode( AVP_Navigation::MM_SLOW );
			break;
		}
		////////////////////
		case FOLLOW_TARGET:
		////////////////////
		{
			GetAlienQueenBrain()->SetMovementMode( AVP_Navigation::MM_FAST );

			// pick out a way point infront of the players is currently over
			AVP_Navigation::NavData& xNavData = GetAlienQueenBrain()->GetNavData();

			//u_short usCurrentWayPoint = ASURA_INVALID_WAYPOINT_ID;
			//AVP_Navigation::GetNearestWaypointID(GetAlienQueenBrain()->GetActor()->GetPosition(), usCurrentWayPoint, xTestParams );

			if( GetAlienQueenBrain()->GetTargetKnowledge() )
			{
				Asura_Vector_3 xDestination = GetAlienQueenBrain()->GetTargetKnowledge()->m_xPosition;

				AVP_Navigation::TestParams xTestParams;
				xTestParams.m_uPathfinderFlags = xNavData.m_uPathfinderFlags;
				xTestParams.m_usZone           = xNavData.m_usCurrentZone;
				AVP_Navigation::ResolveToValidPosition( xDestination, xTestParams );

				AVP_Behaviour_AlienQueen_Movement* pxBehaviour = static_cast<AVP_Behaviour_AlienQueen_Movement*>( CreateChildWithType( AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT ) );

				// No behaviour then we are in error
				if( ! pxBehaviour )
				{
					m_bIsInError = true;
				}

				// cannot get to that position we are in error
				if( pxBehaviour && ( ! pxBehaviour->GetToPosition( xDestination ) ) )
				{
					m_bIsInError = true;
				}

				m_xLastUpdatePosition = xDestination;
			}	
			else
			{
				AVP_Behaviour_AlienQueen_Movement* pxBehaviour = static_cast<AVP_Behaviour_AlienQueen_Movement*>( GetChildWithType( this, AVP_BEHAVIOUR_TYPE_ALIENQUEEN_MOVEMENT ) );
				if( ! pxBehaviour || pxBehaviour->IsInError() || pxBehaviour->IsFinished() )
				{
					// We have reached our target
					SwitchToState( WAIT );
				}
			}
			break;
		}
		////////////////////
		case TARGET_OUT_OF_REACH:
		////////////////////
		{
			GetAlienQueenBrain()->SetMovementMode( AVP_Navigation::MM_SLOW );

			break;
		}
		////////////////////
		case ATTACKING:
		////////////////////
		{
			GetAlienQueenBrain()->SetMovementMode( AVP_Navigation::MM_FAST );

			AVP_Behaviour_AlienQueen_Attack* pxBehaviour = static_cast<AVP_Behaviour_AlienQueen_Attack*>(CreateChildWithType( AVP_BEHAVIOUR_TYPE_ALIENQUEEN_ATTACK) );

			// No behaviour then we are in error
			if( ! pxBehaviour )
			{
				m_bIsInError = true;
			}
			break;
		}
		////////////////////
		default:
		////////////////////
		{
			break;
		}
	}
}

DEBUGVAR float dbg_fTimeBetweenSpits = 6.0f;

bool AVP_Behaviour_AlienQueen_Stalk::ShouldLaunchAttack()
{
	if( ! ( GetAlienQueenBrain() && GetAlienQueenBrain()->GetActor() ) )
	{
		return false;
	}

	if( ! GetAlienQueenBrain()->GetTargetKnowledge() || ! GetAlienQueenBrain()->HasCurrentTarget() ) 
	{
		return false;
	}

	if( ! GetAlienQueenBrain()->HaveLOSToCurrentTarget() ) 
	{
		return false;
	}

	if( GetAlienQueenBrain()->GetTargetKnowledge()->m_uSourceClassification == AVP_ENTITYCLASS_ALIEN )
	{
		// dont attack if the target is an alien. 
		return false;
	}

	const Asura_Vector_3& xOurPosition = GetAlienQueenBrain()->GetActor()->GetPosition();
	const Asura_Vector_3& xTargetPosition = GetAlienQueenBrain()->GetTargetKnowledge()->m_xPosition;

	Asura_Vector_3 xToTarget( xTargetPosition - xOurPosition );
	const float fDistToTarget = xToTarget.Normalise();

	const float fMeleeChargeMinDist = GetAlienQueenBrain()->GetMeleeChargeMinDistance();
	const float fMeleeChargeMaxDist = GetAlienQueenBrain()->GetMeleeChargeMaxDistance();
	
	const float fSpitAcidMinDist	= GetAlienQueenBrain()->GetSpitAcidMinRange();
	const float fSpitAcidMaxDist	= GetAlienQueenBrain()->GetSpitAcidMaxRange();

	const bool bChargeDistOk = ( ( fDistToTarget < fMeleeChargeMaxDist ) && ( fDistToTarget > fMeleeChargeMinDist ) ); 
	const bool bSpitAcidDistOk = ( ( fDistToTarget < fSpitAcidMaxDist ) && ( fDistToTarget > fSpitAcidMinDist ) ); 
	const bool bSlashDistOk = ( fDistToTarget < GetAlienQueenBrain()->GetSlashAttackMinRange() );

	// Are any possible attacks in range?
	if( ! ( bChargeDistOk || bSpitAcidDistOk || bSlashDistOk ) )
	{
		return false;
	}

	// Are we facing the target?
	const Asura_Matrix_3x3& xOurOrientation = GetAlienQueenBrain()->GetActor()->GetOrientation();
	Asura_Vector_3 xOurForward( xOurOrientation.e13, xOurOrientation.e23, xOurOrientation.e33 );
	xOurForward.Normalise();

	const float fDot = xOurForward * xToTarget;

	if( fDot < 0.86f )
	{
		return false;
	}

	bool bCanCharge = false;
	bool bOkToSpitAcid = ( ( m_fTimeSinceLastAcidSpitAttack > dbg_fTimeBetweenSpits ) && bSpitAcidDistOk );

	if( bChargeDistOk )
	{
		// If target is standing near a ledge, dont charge and fall over the ledge
		if( IsTargetOnLedge(xOurPosition, xTargetPosition) ) return false;

		// Return whether waypoint distance to target is approx equal to LOS distance between us and target, if so good to go!!
		bCanCharge = HasRouteToTargetForCharge( xOurPosition, xTargetPosition );
	}
	
	if( GetAlienQueenBrain()->CanSpitAcid() && bOkToSpitAcid )
	{
		GetAlienQueenBrain()->SetNextAttackType( AVP_AQ_ATTACK_SPIT_ACID );
		m_fTimeSinceLastAcidSpitAttack = 0.0f;
		return true;
	}
	else if( bCanCharge )
	{
		GetAlienQueenBrain()->SetNextAttackType( AVP_AQ_ATTACK_CHARGE );
		return true;
	}
	else if( bSlashDistOk )
	{
		GetAlienQueenBrain()->SetNextAttackType( AVP_AQ_ATTACK_CLAW_SWIPE );
		return true;
	}

	// Could not find any valid attacks to perform
	return false;
}

DEBUGVAR float dbg_fRouteMaxDiffProportion = 1.3f;

bool AVP_Behaviour_AlienQueen_Stalk::HasRouteToTargetForCharge( const Asura_Vector_3& xOrigin, const Asura_Vector_3& xTarget )
{
	if( m_fLastChargeRouteCheckTimer > 0.0f ) 
	{
		return false;
	}

	// Early out - if it's far enough away anyway, we won't bother doing this test.
	Asura_Vector_3 xDir( xTarget - xOrigin );
	const float fOffsetDist = xDir.Normalise();

	int iRouteLength = 0;
	u_short* pusRouteBuffer = AVP_Navigation::GetWorkingRouteBuffer();

	AVP_Navigation::TestParams xTestParams = AVP_Navigation::s_xTPStandardParams;
	xTestParams.m_uPathfinderFlags = AVP_Navigation::AVP_AI_PATHFINDER_FLAG_HUMANOID;
	u_short usOriginWP, usTargetWP;
	if( ! AVP_Navigation::GetNearestWaypointID( xOrigin, usOriginWP, xTestParams ) ) { return false; }
	if( ! Asura_AI_Navigation_System::IsPositionInWaypoint( xOrigin, usOriginWP ) ) { return false; }
	if( ! AVP_Navigation::GetNearestWaypointID( xTarget, usTargetWP, xTestParams ) ) { return false; }
	if( ! Asura_AI_Navigation_System::IsPositionInWaypoint( xTarget, usTargetWP ) ) { return false; }

	m_fLastChargeRouteCheckTimer = 3.0f;

	Asura_AI_Navigation_System::FindRouteCustom(pusRouteBuffer,
												iRouteLength,
												AVP_Navigation::GetWorkingRouteBufferLength(),
												usOriginWP,
												usTargetWP,
												AVP_Navigation::AVP_AI_PATHFINDER_FLAG_HUMANOID,
												&AVP_Navigation::s_xAlienQueenAStarFunctor,
												AVP_Navigation::FunctorGoalCostFunction,
											    AVP_Navigation::FunctorTraversalCostFunction,
											    AVP_Navigation::FunctorGoalAchievedFunction,
											    AVP_Navigation::FunctorLinkFlagsTestFunction );

	// See if we failed to get a route at all.
	if( iRouteLength <= 0 ) { return false; }

	// Go through the route and see how far it is.
	// Just use the estimate of the waypoints only, so we aren't affected by which side of the waypoint we're on.
	float fRouteDist = 0.0f;
	const Asura_AI_NavVertex* pxCurrentWaypoint = Asura_AI_Navigation_System::GetWaypoint( pusRouteBuffer[0] );
	for( int i=1; i<iRouteLength; ++i )
	{
		const Asura_AI_NavVertex* pxNewWaypoint = Asura_AI_Navigation_System::GetWaypoint( pusRouteBuffer[i] );
		fRouteDist += ( pxNewWaypoint->GetPosition() - pxCurrentWaypoint->GetPosition() ).Magnitude();
		pxCurrentWaypoint = pxNewWaypoint;
	}

	return( ( fRouteDist / fOffsetDist ) <= dbg_fRouteMaxDiffProportion );
}

DEBUGVAR float dbg_fDownAngleForEnvLOSTest = 0.36f;

bool AVP_Behaviour_AlienQueen_Stalk::IsTargetOnLedge( const Asura_Vector_3& xOrigin, const Asura_Vector_3& xTarget ) const
{
	if ( !(GetAlienQueenBrain() && GetAlienQueenBrain()->HasCurrentTarget()) ) return false;

	const Asura_Guid uCurrentTargetGuid = GetAlienQueenBrain()->GetCurrentTargetGuid();
	const AVP_ServerEntity_Actor* pxTargetActor = AVP_ServerEntity_Actor::FindEntity( uCurrentTargetGuid );
	if ( !pxTargetActor ) return false;

	Asura_Vector_3 xHeadPos;
	if ( ! pxTargetActor->GetWorldSpaceMarkerPoint(AVP_Vertebrate_Description::GetMarkerPointHashID( AVP_Vertebrate_Description::MPID_EYES ), &xHeadPos, NULL ) )
	{
		xHeadPos = pxTargetActor->GetPosition();
		xHeadPos.y -= 1.8f;
	}

	Asura_Vector_3 xDir( xTarget - xOrigin );
	xDir.Normalise();

	xDir.y = dbg_fDownAngleForEnvLOSTest;

	Asura_LOS_InputData xInput( xHeadPos, xDir, uCurrentTargetGuid, 6.0f, false );
	const Asura_LOS_OutputData* pxOutput = Asura_LOS::TestAgainstEnvironment( &xInput );

	if( pxOutput->m_bHitEnvironment )
	{
		return false;
	}

	return true;
}



AVP_Brain_Actor_AlienQueen* AVP_Behaviour_AlienQueen_Stalk::GetAlienQueenBrain() const
{
	return static_cast<AVP_Brain_Actor_AlienQueen*>( GetProjectBrain() );
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
	"WAIT",
	"FOLLOWING TARGET",
	"LOST PLAYER",
	"ATTACKING",
};

void AVP_Behaviour_AlienQueen_Stalk::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "AlienQueen Wander: %s", s_aszStateStrings[ m_eCurrentState ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
}
#endif

