//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Predator_AtVentagePoint.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/Predator/AVP_Behaviour_Predator_AtVentagePoint.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#include "Debugging/Asura_DebugLineRender.h"
#endif
#include "Entity/PhysicalObject/Asura_ClientEntity_Actor.h"

// axon
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Behaviour/BehaviourTypes/Predator/AVP_Behaviour_Predator_Movement.h"
#include "Core/AVP_Waypoint_Manager.h"
#include "Core/Axon_Project_System.h"
#include "Core/BrainTypes/AVP_Brain_Actor_Predator.h"
#include "Knowledge/AVP_Knowledge_System_Predator.h"
#include "Knowledge/Axon_Knowledge.h"

// avp
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ServerActor_Predator.h"

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Predator_AtVentagePoint
//
//////////////////////////////////////////////////////////////////////////////

AVP_Behaviour_Predator_AtVentagePoint::AVP_Behaviour_Predator_AtVentagePoint( AVP_Brain_Actor_Predator* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_PREDATOR_AT_VENTAGE_POINT )
, m_eCurrentState(UNSET)
, m_fTimeInState(0.0f)
, m_iConsecutiveFire(0)
, m_iConsecutiveFailedFireAttempts(0)
, m_bFireOccured(true)
, m_xStartTrackingPosition(Asura_Maths::ZeroVector3)
, m_xSelectedTargetPosition(Asura_Maths::ZeroVector3)
, m_xDeltaAimDirection(Asura_Maths::ZeroVector3)
, m_xDesiredAimDirection(Asura_Maths::ZeroVector3)
, m_xDesiredTargetPosition(Asura_Maths::ZeroVector3)
, m_bIsCharging(false)
, m_bAbort(false)
{
}


bool AVP_Behaviour_Predator_AtVentagePoint::Activate()
{
#ifndef ASURA_NO_DEBUG_OUTPUT
	GetPredatorBrain()->SetRenderRoute( true );
#endif

	return PARENT::Activate();
}

void AVP_Behaviour_Predator_AtVentagePoint::Deactivate()
{
	Axon_Behaviour* pxChild = GetChild();
	while ( pxChild )
	{
		UnlinkChild( pxChild );
		GetProjectBrain()->DestroyBehaviour( pxChild );
		pxChild = GetChild();
	}

	PARENT::Deactivate();
}

DEBUGVAR int dbg_iMaxConsecutiveFailedFireAttempts = 3;
DEBUGVAR float dbg_fWaitAfterAbortedFire = 0.5f;


void AVP_Behaviour_Predator_AtVentagePoint::PerFrameUpdate( const float fTimeslice )
{
#ifndef ASURA_NO_DEBUG_OUTPUT
	if (Axon_Project_System::GetRenderingDebugInfo())
	{
		Asura_DebugLineRender::AddCross(m_xDesiredTargetPosition, 0xfff0f0f0);
		Asura_DebugLineRender::AddCross(m_xSelectedTargetPosition, 0xff00f0f0);
	}
#endif
	
	PARENT::PerFrameUpdate( fTimeslice );

	if ( IsFinished() || IsInError() )
	{
		return;
	}

	m_fTimeInState += fTimeslice; 

	// we not suppose to fall or jump from ventage point ...
	if (!GetPredatorBrain()->CanPlanMove())
	{
		SetInError(true);
		SwitchToState(FINISHED);
		return;
	}

	switch (m_eCurrentState)
	{
	case UNSET:
		SwitchToState(FIND_TARGET);
		break;
	case IDLE:
		break;
	case FIND_TARGET:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}
		
		if (GetPredatorBrain()->IsAimAtTestFinished())
		{
			if (GetPredatorBrain()->GetAimAtTestResult(m_xSelectedTargetPosition))
			{
				SwitchToState(ORIENT_TO_TARGET);
			}
			else
			{
				SetInError(true);
				SwitchToState(FINISHED);
			}
		}
		break;
	case ORIENT_TO_TARGET:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}

		OrientToTarget();
		if(IsAtDesiredAim())
		{
			SwitchToState(EQUIP_PLASMACASTER);
		}
		break;
	case EQUIP_PLASMACASTER:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}

		if (GetPredatorBrain()->IsHoldingPlasmaCaster())
		{
			SwitchToState(AIM_NEAR_TARGET);
		}
		break;
	case AIM_NEAR_TARGET:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}

		if (GetPredatorBrain()->IsAimAtTestFinished())
		{
			if (!GetPredatorBrain()->GetAimAtTestResult(m_xSelectedTargetPosition))
			{
				SwitchToState(FINISHED);
				return;
			}
			GetPredatorBrain()->StartAimAtTest(true);
		}

		{
			float fAimDelay = GetPredatorBrain()->GetConfigInfo().m_xAtVPBehaviour.m_fAimDelay;
			if (m_fTimeInState < fAimDelay)
			{
				if (IsAtDesiredAim())
				{
					AimNearTarget();
				}
			}
			else
			{
				SwitchToState(DECLOAK);
			}
		}
		break;
	case DECLOAK:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}
		
		if (GetPredatorBrain()->IsAimAtTestFinished() && m_fTimeInState < GetPredatorBrain()->GetConfigInfo().m_xWeaponConfig.m_fPlasmaCasterDelayBetweenShots )
		{
			if (!GetPredatorBrain()->GetAimAtTestResult(m_xSelectedTargetPosition))
			{
				SwitchToState(FINISHED);
				return;
			}
			GetPredatorBrain()->StartAimAtTest(true);
		}
		AimAtTarget();
		if (!GetPredatorBrain()->IsCloaked())
		{
			m_iConsecutiveFire = Asura_Random::GetInt( GetPredatorBrain()->GetConfigInfo().m_xWeaponConfig.m_iPlasmaCasterMiniumNumberOfShots, GetPredatorBrain()->GetConfigInfo().m_xWeaponConfig.m_iPlasmaCasterMaximumNumberOfShots );
			SwitchToState(AIM_AT_TARGET);
		}
		break;
	case AIM_AT_TARGET:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}

		if(m_bFireOccured && m_fTimeInState < GetPredatorBrain()->GetConfigInfo().m_xWeaponConfig.m_fPlasmaCasterDelayBetweenShots)
		{
			if (GetPredatorBrain()->IsAimAtTestFinished())
			{
				if (!GetPredatorBrain()->GetAimAtTestResult(m_xSelectedTargetPosition))
				{
					SwitchToState(FINISHED);
					return;
				}
				GetPredatorBrain()->StartAimAtTest(false);
			}
		}
		AimAtTarget();
		if((!m_bFireOccured || m_fTimeInState > GetPredatorBrain()->GetConfigInfo().m_xWeaponConfig.m_fPlasmaCasterDelayBetweenShots) && IsAtDesiredAim())  
		{
			SwitchToState(FIRE);
		}
		break;

	case REAQUIRE_TARGET:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}

		if (GetPredatorBrain()->IsAimAtTestFinished())
		{
			if (!GetPredatorBrain()->GetAimAtTestResult(m_xSelectedTargetPosition))
			{
				SwitchToState(FINISHED);
				return;
			}
			GetPredatorBrain()->StartAimAtTest(false);
			SwitchToState(AIM_AT_TARGET);
		}
		break;
	case FIRE:
	
		AimAtTarget();
		if(Fire())
		{
			if (m_iConsecutiveFire >= 0)
			{
				if (m_iConsecutiveFailedFireAttempts >= dbg_iMaxConsecutiveFailedFireAttempts)
				{
					SetInError(true);
					SwitchToState(FINISHED);
				}
				else if (m_bFireOccured) // fire happened 
				{
					SwitchToState(REAQUIRE_TARGET);			
				}
				else
				{
					SwitchToState(WAIT_AFTER_ABORTED_FIRE);			
				}
			}
			else
			{
				SwitchToState(FINISHED);
			}
		}
		break;
	case WAIT_AFTER_ABORTED_FIRE:
		if (m_bAbort)
		{
			SwitchToState(FINISHED);
			return;
		}

		if (m_fTimeInState > dbg_fWaitAfterAbortedFire)
		{
			SwitchToState(REAQUIRE_TARGET);
		}
		break;
	case FINISHED:
	default:
		Asura_Assert(false, "unexpected state");
		break;
	}
}

void AVP_Behaviour_Predator_AtVentagePoint::Update( const float fTimeslice )
{
	PARENT::Update( fTimeslice );

	if ( IsFinished() || IsInError() )
	{
		return;
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
	"UNSET",
	"IDLE",
	"FIND_TARGET",
	"ORIENT_TO_TARGET",
	"DECLOAK",
	"EQUIP_PLASMACASTER",
	"AIM_NEAR_TARGET",
	"REAQUIRE_TARGET",
	"AIM_AT_TARGET",
	"FIRE",
	"WAIT_AFTER_ABORTED_FIRE",
	"FINISHED",
	"NUM_STATES"
};

void AVP_Behaviour_Predator_AtVentagePoint::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffffffff,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "%s", s_aszStateStrings[ m_eCurrentState ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;


}
#endif

AVP_Brain_Actor_Predator* AVP_Behaviour_Predator_AtVentagePoint::GetPredatorBrain() const
{
	return static_cast< AVP_Brain_Actor_Predator* >( GetProjectBrain() );
}


void AVP_Behaviour_Predator_AtVentagePoint::SwitchToState( const State eNextState )
{

	switch ( m_eCurrentState )
	{
	case AIM_AT_TARGET:
		GetPredatorBrain()->SetNormalAimSpeed();
		break;
	case UNSET:
	case IDLE:
	case FIND_TARGET:
	case ORIENT_TO_TARGET:
	case AIM_NEAR_TARGET:
	case REAQUIRE_TARGET:	
	case DECLOAK:
	case EQUIP_PLASMACASTER:
	case WAIT_AFTER_ABORTED_FIRE:
		break;
	case FIRE:
		if (m_bIsCharging)
		{
			GetPredatorBrain()->AbortPlasmaCasterCharge();
		}
		break;
	case FINISHED:
	default:
		Asura_Assert(false, "unexpected state");
	}

	m_eCurrentState = eNextState;
	m_fTimeInState  = 0.0f;
	switch ( m_eCurrentState )
	{
	case IDLE:
		break;
	case FIND_TARGET:
		GetPredatorBrain()->StartAimAtTest(true);		
		break;
	case ORIENT_TO_TARGET:
		m_xDeltaAimDirection.RandomDirectionSeeded();
		break;
	case EQUIP_PLASMACASTER:
		GetPredatorBrain()->EquipPlasmaCaster();
		break;
	case AIM_NEAR_TARGET:
		{
			const Axon_Knowledge& xThreat = GetPredatorBrain()->GetKnowledgeSystem()->GetBestThreatKnowledge();
			if (xThreat.GetGuid() == ASURA_GUID_UNREGISTERED)
			{
				SwitchToState(FINISHED);
			}
			else
			{
				GetPredatorBrain()->StartAimAtTest(false);
			}
		}
		break;
	case AIM_AT_TARGET:
		GetPredatorBrain()->SetFastTurnSpeed();
		GetPredatorBrain()->StartAimAtTest(false);
		break;
	case REAQUIRE_TARGET:
		GetPredatorBrain()->StartAimAtTest(true);
		break;
	case DECLOAK:
		GetPredatorBrain()->Decloak();
		break;
	case FIRE:
		break;
	case WAIT_AFTER_ABORTED_FIRE:
		break;
	case FINISHED:
		GetPredatorBrain()->SetNormalAimSpeed();
		SetFinished(true);
		break;
	case UNSET:
	default:
		Asura_Assert(false, "unexpected state");
		break;
	}
}

DEBUGVAR float g_fAimRange = 8.0f;
void AVP_Behaviour_Predator_AtVentagePoint::OrientToTarget()
{
	m_xDesiredTargetPosition = m_xSelectedTargetPosition + m_xDeltaAimDirection * g_fAimRange;
		
	Asura_Vector_3 xOurPosition;
	GetPredatorBrain()->GetPredatorActor()->GetHeadPosition(xOurPosition); // here we can not get yet the muzzle since we might not be equipping the weapon yet.

	m_xDesiredAimDirection = m_xDesiredTargetPosition - xOurPosition;
	m_xDesiredAimDirection.Normalise();

	GetPredatorBrain()->SetAimDirection(m_xDesiredAimDirection);
	GetPredatorBrain()->SetLookAtPosition(m_xDesiredTargetPosition);
}

void AVP_Behaviour_Predator_AtVentagePoint::AimNearTarget()
{
	float fAimDelay = GetPredatorBrain()->GetConfigInfo().m_xAtVPBehaviour.m_fAimDelay;
	float fAimFraction = 1.0f-m_fTimeInState/fAimDelay;
	if (fAimFraction < 0.0f)
	{
		fAimFraction = 0.0f;
	}

	m_xDesiredTargetPosition = m_xSelectedTargetPosition + m_xDeltaAimDirection * g_fAimRange * fAimFraction;

	Asura_Vector_3 xOurPosition;
	GetPredatorBrain()->GetPredatorActor()->GetMuzzlePosition(xOurPosition);

	m_xDesiredAimDirection = m_xDesiredTargetPosition - xOurPosition;
	m_xDesiredAimDirection.Normalise();

	GetPredatorBrain()->SetAimDirection(m_xDesiredAimDirection);
	GetPredatorBrain()->SetLookAtPosition(m_xDesiredTargetPosition);
}


void AVP_Behaviour_Predator_AtVentagePoint::AimAtTarget()
{
	m_xDesiredTargetPosition = m_xSelectedTargetPosition;

	Asura_Vector_3 xOurPosition;
	GetPredatorBrain()->GetPredatorActor()->GetMuzzlePosition(xOurPosition);

	m_xDesiredAimDirection = m_xDesiredTargetPosition - xOurPosition;
	m_xDesiredAimDirection.Normalise();

	GetPredatorBrain()->SetAimDirection(m_xDesiredAimDirection);
	GetPredatorBrain()->SetLookAtPosition(m_xDesiredTargetPosition);
}

DEBUGVAR float dbg_fEstimateMoveDistance = 8.0f;
DEBUGVAR float dbg_fMinimumRotateAngleCosine = Asura_Maths::Cosine(Asura_Maths::DegToRad(30.0f));

bool AVP_Behaviour_Predator_AtVentagePoint::EstimateNewOrientation()
{
	const Axon_Knowledge& xThreat = GetPredatorBrain()->GetKnowledgeSystem()->GetBestThreatKnowledge();
	if (xThreat.GetGuid() == ASURA_GUID_UNREGISTERED)
	{
		return false;
	}

	Asura_Vector_3 xEstimatedPosition = xThreat.m_xPosition + xThreat.m_xDirection * dbg_fEstimateMoveDistance;
	Asura_Vector_3 xEstimateDirection = xEstimatedPosition - GetPredatorBrain()->GetActor()->GetPosition();
	xEstimatedPosition.Normalise();

	if (GetPredatorBrain()->GetAimDirection() * xEstimatedPosition < dbg_fMinimumRotateAngleCosine)
	{
		m_xDesiredTargetPosition = xEstimatedPosition;
		return true;
	}
	else
	{
		return false;
	}
}

bool AVP_Behaviour_Predator_AtVentagePoint::IsAtDesiredAim()
{
	Asura_AVector3 xCurrecntAimDirection;
	GetPredatorBrain()->GetClientActor()->GetAimDirection(xCurrecntAimDirection);
	return m_xDesiredAimDirection.ApproximatelyEquals(Asura_Vector_3(xCurrecntAimDirection));
}

DEBUGVAR float dbg_fSafeDistance = 8.0f;


bool AVP_Behaviour_Predator_AtVentagePoint::Fire()
{
	if(!m_bIsCharging)
	{
		m_bIsCharging = GetPredatorBrain()->StartPlasmaCasterCharge(GetPredatorBrain()->GetConfigInfo().m_xWeaponConfig.m_fPlasmaCasterDamage);
	}
	else
	{
		if (GetPredatorBrain()->IsPlasmaCasterCharged())
		{
			Asura_AVector3 xCurrecntAimDirection;
			GetPredatorBrain()->GetClientActor()->GetAimDirection(xCurrecntAimDirection);
			Asura_Vector_3 xSafeTarget;
			GetPredatorBrain()->GetActor()->GetMuzzlePosition(xSafeTarget);
			xSafeTarget += Asura_Vector_3(xCurrecntAimDirection) * dbg_fSafeDistance;

			if(GetPredatorBrain()->CanFireAt(Asura_Vector_3(xSafeTarget), true))
			{
				--m_iConsecutiveFire;
				m_iConsecutiveFailedFireAttempts = 0;
				GetPredatorBrain()->ReleasePlasmaCasterCharge();
				m_bFireOccured = true;
			}
			else
			{
				++m_iConsecutiveFailedFireAttempts;
				GetPredatorBrain()->AbortPlasmaCasterCharge();
				m_bFireOccured = false;
			}
			m_bIsCharging = false;
			return true;
		}
	}

	return false;
}
//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
