//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Sentrygun_PointAndShoot.cpp
//
// Copyright (C) 2008 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Behaviour/BehaviourTypes/Sentrygun/AVP_Behaviour_Sentrygun_PointAndShoot.h"

// asura
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Debugging/Asura_DebugLineRender.h"
#include "Rendering/Asura_Font_System.h"
#include "Actor/AVP_Actor_DynamicStat.h"
#endif

// axon
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Core/BrainTypes/AVP_Brain_Actor_Sentrygun.h"
#include "Knowledge/AVP_Knowledge_System_Sentrygun.h"

// avp
#include "Actor/Weapons/Sentrygun/AVP_Sentrygun_AIAnimStateMachine.h"
#include "Entity/PhysicalObject/Actor/Weapons/Sentrygun/AVP_ServerWeapon_Sentrygun.h"
#include "Entity/PhysicalObject/Actor/Weapons/Sentrygun/AVP_ClientWeapon_Sentrygun.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/Weapons/AVP_InventoryWeapon_RangedWeapon.h"

DEBUGVAR float dbg_fCloseDotProduct = Asura_Maths::Cosine( Asura_Maths::DegToRad( 5.0f ) ); // 5 degrees
static const Asura_Vector_3 g_xSentrygunTargetingOffset( 0.0f, -0.6f, 0.0f );

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Behaviour_Sentrygun_PointAndShoot
//
//////////////////////////////////////////////////////////////////////////////

AVP_Behaviour_Sentrygun_PointAndShoot::AVP_Behaviour_Sentrygun_PointAndShoot( AVP_Brain_Actor_Sentrygun* pxBrain )
: PARENT( pxBrain, AVP_BEHAVIOUR_TYPE_SENTRYGUN_POINTANDSHOOT )
, m_uTargetGuid( ASURA_GUID_UNREGISTERED )
, m_eFiringState( FIRINGSTATE_UNSET )
, m_fShootTimer( 0.0f )
, m_fInvestigateTime( 1.0f )
, m_fFireTime( 2.0f )
, m_fWaitTime( 0.25f )
, m_fLostTime( 1.0f )
, m_xTargetPosition( Asura_Maths::ZeroVector3 )
, m_bInvItemEndUsagePending( false )
{
	SetCurrentState( UNSET );
}

bool AVP_Behaviour_Sentrygun_PointAndShoot::Activate()
{
	return PARENT::Activate();
}

void AVP_Behaviour_Sentrygun_PointAndShoot::Deactivate()
{
	DestroyAllChildren();
	PARENT::Deactivate();
}

void AVP_Behaviour_Sentrygun_PointAndShoot::Update( const float fTimeslice )
{
	PARENT::Update( fTimeslice );

	if ( IsInError() || IsFinished() )
	{
		return;
	}

	if ( m_bInvItemEndUsagePending )
	{
		AVP_Inventory_System*		pxInventorySystem	= GetSentrygunBrain()->GetActor()->GetInventorySystem();
		AVP_InventoryItem_Weapon*	pxCurrentWeapon		= static_cast<AVP_InventoryItem_Weapon*>(pxInventorySystem ? pxInventorySystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL);
		if ( pxCurrentWeapon && pxCurrentWeapon->IsUsing( true ) )
		{
			// stop firing weapon
			pxCurrentWeapon->RequestEndUsage( AVP_InventoryWeapon_RangedWeapon::TT_ABORT_DESTROY_CHARGE );
			m_bInvItemEndUsagePending = false;
		}
	}

	switch ( GetCurrentState() )
	{
		case UNSET:
		{
			SwitchToState( SWEEPING );
			break;
		}

		case SWEEPING:
		{
			if ( HaveTarget() )
			{
				SwitchToState( TARGETTING );
			}
			break;
		}

		case TARGETTING:
		{
			if ( !HaveTarget() )
			{
				SwitchToState( LOST_TARGET );
			}
			else
			{
				TurnTowardsTarget();
			}
			break;
		}

		case LOST_TARGET:
		{
			if ( HaveTarget() )
			{
				SwitchToState( TARGETTING );
			}
			else if ( m_fShootTimer > m_fLostTime )
			{
				SwitchToState( SWEEPING );
			}
			break;
		}

		case DISABLED:
		{
			// do nothing
		}
		break;

		default:
		{
			Asura_Assert( false, "Unhandled state in AVP_Behaviour_Sentrygun_PointAndShoot::Update()" ); 
		}
		break;
	}

	switch ( m_eFiringState )
	{
		case FIRINGSTATE_INVESTIGATING:
		{
			if ( m_fShootTimer > m_fInvestigateTime )
			{
				if ( IsAbleToFire() )
				{
					SwitchToFiringState( FIRINGSTATE_FIRING );
				}
				else
				{
					SwitchToFiringState( FIRINGSTATE_WAITING );
				}
			}
			break;
		}
		
		case FIRINGSTATE_FIRING:
		{
			if ( m_fShootTimer > m_fFireTime || !IsAbleToFire() )
			{
				SwitchToFiringState( FIRINGSTATE_WAITING );
			}
			break;
		}

		case FIRINGSTATE_WAITING:
		{
			if ( m_fShootTimer > m_fWaitTime && IsAbleToFire() )
			{
				SwitchToFiringState( FIRINGSTATE_FIRING );
			}
			break;
		}

		default:
		case FIRINGSTATE_UNSET:
		case FIRINGSTATE_DISABLED:
		{
			break;
		}
	}
}

void AVP_Behaviour_Sentrygun_PointAndShoot::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	UpdateStateTimer( fTimeslice );
	m_fShootTimer += fTimeslice;

	if ( IsInError() || IsFinished() )
	{
		return;
	}

	UpdateTargetDetails();
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[ AVP_Behaviour_Sentrygun_PointAndShoot::NUM_STATES ] =
{
	"UNSET",	
	"SWEEPING",
	"TARGETTING",
	"LOST_TARGET",
	"DISABLED"
};

static const char* s_aszFiringStateStrings[ AVP_Behaviour_Sentrygun_PointAndShoot::NUM_FIRING_STATES ] =
{
	"FIRINGSTATE_UNSET",
	"FIRINGSTATE_INVESTIGATING",
	"FIRINGSTATE_FIRING",
	"FIRINGSTATE_WAITING",
	"FIRINGSTATE_DISABLED"
};

void AVP_Behaviour_Sentrygun_PointAndShoot::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffff8080,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "%s : %s",
		s_aszStateStrings[ GetCurrentState() ], s_aszFiringStateStrings[ GetCurrentFiringState() ] );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;

	const Asura_Guid					uActorGuid			= GetSentrygunBrain()->GetActor()->GetGuid();
	const AVP_ClientWeapon_Sentrygun*	pxClient			= AVP_ClientWeapon_Sentrygun::FindEntity(uActorGuid);
	const AVP_Inventory_System*			pxClientSystem		= (pxClient == NULL) ? NULL : pxClient->GetInventorySystem();
	const AVP_InventoryItem_Weapon*		pxClientWeapon		= static_cast<const AVP_InventoryItem_Weapon*>(pxClientSystem ? pxClientSystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL);

	const AVP_Inventory_System*			pxServerSystem		= GetSentrygunBrain()->GetActor()->GetInventorySystem();
	const AVP_InventoryItem_Weapon*		pxServerWeapon		= static_cast<const AVP_InventoryItem_Weapon*>(pxServerSystem ? pxServerSystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL);

	if(pxClientWeapon != NULL)
	{
		Asura_Char	szClientTemp[256]	= "CLNT: ";
		pxClientWeapon->CreateDebugString(250, &szClientTemp[6]);

		Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffff8080,
			ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, szClientTemp );
		GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
	}

	if(pxServerWeapon != NULL)
	{
		Asura_Char	szServerTemp[256]	= "SRVR: ";
		pxServerWeapon->CreateDebugString(250, &szServerTemp[6]);

		Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffff8080,
			ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, szServerTemp );
		GetProjectBrain()->GetDebugOutputPos().y += 10.0f;
	}

	Asura_Font_System::Print( GetProjectBrain()->GetDebugOutputPos().x, GetProjectBrain()->GetDebugOutputPos().y, 0.5f, 0xffff8080,
		ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "Target GUID: %d", m_uTargetGuid );
	GetProjectBrain()->GetDebugOutputPos().y += 10.0f;

	Asura_DebugLineRender::AddSphere( m_xTargetPosition, 0.3f, 0xffff8080, 4, 2 );
}
#endif

AVP_Brain_Actor_Sentrygun* AVP_Behaviour_Sentrygun_PointAndShoot::GetSentrygunBrain() const
{
	return static_cast< AVP_Brain_Actor_Sentrygun* >( GetProjectBrain() );
}

void AVP_Behaviour_Sentrygun_PointAndShoot::UpdateTargetDetails()
{
	m_uTargetGuid = GetSentrygunBrain()->GetKnowledgeSystem()->GetBestThreatKnowledge().GetGuid();
	if ( m_uTargetGuid != ASURA_GUID_UNREGISTERED )
	{
		const AVP_ServerEntity_Actor* pxActorTarget = AVP_ServerEntity_Actor::FindEntity( m_uTargetGuid );
		if ( pxActorTarget )
		{
			pxActorTarget->GetTargetPosition( m_xTargetPosition );
		}
		else
		{
			m_uTargetGuid = ASURA_GUID_UNREGISTERED;
		}
	}
}

bool AVP_Behaviour_Sentrygun_PointAndShoot::HaveTarget() const
{
	return ( m_uTargetGuid != ASURA_GUID_UNREGISTERED );
}

void AVP_Behaviour_Sentrygun_PointAndShoot::SwitchToState( const State eNextState )
{
	switch ( GetCurrentState() )
	{
		case TARGETTING:
		{
			SwitchToFiringState( FIRINGSTATE_DISABLED );
			break;
		}

		case LOST_TARGET:
		{
			if ( eNextState == SWEEPING )
			{
				GetSentrygunBrain()->GetSentrygunActor()->SendSoundMessage( AVP_ClientWeapon_Sentrygun::SOUND_RESUMING_SWEEP );
			}
		}
		break;

		case UNSET:
		case SWEEPING:
		case DISABLED:
		default:
			break;
	}

	SetCurrentState( eNextState );
	switch ( GetCurrentState() )
	{
		case TARGETTING:
		{
			GetSentrygunBrain()->GetActor()->GetAIAnimStateMachine()->RequestIdle();
			SwitchToFiringState( FIRINGSTATE_INVESTIGATING );
			break;
		}

		case SWEEPING:
		{
			static_cast< AVP_Sentrygun_AIAnimStateMachine* >( GetSentrygunBrain()->GetActor()->GetAIAnimStateMachine() )->RequestSweeping();
		}
		break;

		case LOST_TARGET:
		{
			GetSentrygunBrain()->GetSentrygunActor()->SendSoundMessage( AVP_ClientWeapon_Sentrygun::SOUND_LOST_TARGET );
		}

		case UNSET:
		case DISABLED:
		default:
			break;
	}
}

void AVP_Behaviour_Sentrygun_PointAndShoot::SwitchToFiringState( const FiringState eNextFiringState )
{
	AVP_Inventory_System*		pxInventorySystem	= GetSentrygunBrain()->GetActor()->GetInventorySystem();
	AVP_InventoryItem_Weapon*	pxCurrentWeapon		= static_cast<AVP_InventoryItem_Weapon*>(pxInventorySystem ? pxInventorySystem->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL);

	switch ( m_eFiringState )
	{
		case FIRINGSTATE_FIRING:
		{
			if ( pxCurrentWeapon && pxCurrentWeapon->IsUsing( true ) )
			{
				// fire weapon
				pxCurrentWeapon->RequestEndUsage( AVP_InventoryWeapon_RangedWeapon::TT_ABORT_DESTROY_CHARGE );
			}
			else
			{
				m_bInvItemEndUsagePending = true;
			}
		}
		break;

		case FIRINGSTATE_INVESTIGATING:
		{
			if ( eNextFiringState == FIRINGSTATE_FIRING || eNextFiringState == FIRINGSTATE_WAITING )
			{
				GetSentrygunBrain()->GetSentrygunActor()->SendSoundMessage( AVP_ClientWeapon_Sentrygun::SOUND_LOCKED_ON );
			}
		}
		break;

		case FIRINGSTATE_WAITING:
		case FIRINGSTATE_DISABLED:
		case FIRINGSTATE_UNSET:
		default:
			break; // no nothing
	}

	m_eFiringState = eNextFiringState;
	m_fShootTimer  = 0.0f;
	
	switch ( m_eFiringState )
	{
		case FIRINGSTATE_FIRING:
		{
			if ( pxCurrentWeapon && pxCurrentWeapon->IsIdling( true ) )
			{
				// fire weapon
				pxCurrentWeapon->RequestBeginUsage( AVP_InventoryWeapon_RangedWeapon::UT_FULL_AUTO_PRIMARY );
			}
		}
		break;

		case FIRINGSTATE_INVESTIGATING:
		{
			GetSentrygunBrain()->GetSentrygunActor()->SendSoundMessage( AVP_ClientWeapon_Sentrygun::SOUND_INVESTIGATING );
		}
		break;

		case FIRINGSTATE_WAITING:
		case FIRINGSTATE_DISABLED:
		case FIRINGSTATE_UNSET:
		default:
			break; // no nothing
	}
}

void AVP_Behaviour_Sentrygun_PointAndShoot::TurnTowardsTarget()
{
	if ( m_uTargetGuid != ASURA_GUID_UNREGISTERED )
	{
		// Ideally we'd turn eye to eye, but this'll do for now.
		Asura_Vector_3 xDirection = m_xTargetPosition - ( GetSentrygunBrain()->GetActor()->GetPosition() + g_xSentrygunTargetingOffset );
		xDirection.Normalise();
		const Asura_Matrix_3x3& xOri = GetSentrygunBrain()->GetActor()->GetOrientation();
		const Asura_Vector_3 xFwd( xOri.e13, xOri.e23, xOri.e33 );
		float fDot = xFwd * xDirection;
		if ( fDot > 0.0f ) // if target is in front of us
		{
			GetSentrygunBrain()->SetAimDirection( xDirection );
			GetSentrygunBrain()->SetLookAtPosition( m_xTargetPosition );
		}
	}
}

bool AVP_Behaviour_Sentrygun_PointAndShoot::IsAbleToFire() const
{
	// Get our aim direction
	const Asura_Vector_3 xAimDir = GetSentrygunBrain()->GetAimDirection();
	
	// Get the Pulse rifles muzzle marker point
	Asura_Matrix_3x3 xMuzOri;
	if ( !GetSentrygunBrain()->GetActor()->GetWorldSpaceMarkerPoint( 0x0001A7D2, NULL, &xMuzOri ) ) // Asura_GetHashID("muz")
	{
		Asura_Assert( false, "Couldn't get muzzle MP" );
		return false;
	}
	const Asura_Vector_3	xMuzFwd	= Asura_Vector_3( xMuzOri.e13, xMuzOri.e23, xMuzOri.e33 );
	return ( xMuzFwd * xAimDir > dbg_fCloseDotProduct );
}
