//////////////////////////////////////////////////////////////////////////////
//
// AVP_Navigation.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Core/AVP_Navigation.h"

// asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Mempool.h"
#include "Core/Asura_OrientedVolume.h"
#include "Core/Memory/Asura_Mem.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Debugging/Asura_DebugLineRender.h"
#endif
#include "Debugging/Asura_Logfile.h"
#include "Jobs/Asura_Job.h"
#include "Light/Asura_GI_Solution.h"

// axon
#include "Core/AVP_Waypoint_Manager.h"
#include "Core/BrainTypes/Axon_Project_Brain.h"

// avp
#include "Actor/Vertebrates/Alien/AVP_Alien_JumpHandler.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesJump.h"

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Navigation
//
//////////////////////////////////////////////////////////////////////////////

const AVP_Navigation::TestParams AVP_Navigation::s_xTPStandardParams;
const AVP_Navigation::APParams   AVP_Navigation::s_xAPStandardParams;

AVP_Navigation::AlienAStarFunctor		AVP_Navigation::s_xAlienAStarFunctor;
AVP_Navigation::HumanoidAStarFunctor	AVP_Navigation::s_xHumanoidAStarFunctor;
AVP_Navigation::PredatorAStarFunctor	AVP_Navigation::s_xPredatorAStarFunctor;
AVP_Navigation::AlienQueenAStarFunctor	AVP_Navigation::s_xAlienQueenAStarFunctor;

#ifndef ASURA_NO_DEBUG_OUTPUT
int AVP_Navigation::s_iDbgTotalRoutesAllocated    = 0;
int AVP_Navigation::s_iDbgNumShortRoutesAllocated = 0;
int AVP_Navigation::s_iDbgNumMedRoutesAllocated   = 0;
int AVP_Navigation::s_iDbgNumLongRoutesAllocated  = 0;
#endif

Asura_Mempool* AVP_Navigation::s_pxShortRouteMempool		= NULL;
Asura_Mempool* AVP_Navigation::s_pxMedSmallRouteMempool     = NULL;
Asura_Mempool* AVP_Navigation::s_pxMedRouteMempool			= NULL;
Asura_Mempool* AVP_Navigation::s_pxMedLongRouteMempool		= NULL;
Asura_Mempool* AVP_Navigation::s_pxLongRouteMempool			= NULL;
u_short        AVP_Navigation::s_ausWorkingRouteBuffer[] = { 0xffff };

#ifndef ASURA_NO_DEBUG_OUTPUT
u_short        AVP_Navigation::s_usPreLength = 0;
u_short        AVP_Navigation::s_ausPreRouteBuffer[]     = { 0xffff };
u_short        AVP_Navigation::s_usPostLength = 0;
u_short        AVP_Navigation::s_ausPostRouteBuffer[]    = { 0xffff };
#endif

static AVP_Navigation::TestParams s_xTestParams;
static Asura_Vector_3             s_xTestPosition		= Asura_Maths::ZeroVector3;
static u_short                    s_usBestWaypointID	= ASURA_INVALID_WAYPOINT_ID;
static float                      s_fBestDistanceSq		= Asura_Maths::LargeFloat;
static float                      s_fBestDistance		= Asura_Maths::LargeFloat;

static bool NearestWaypointTreeCallback( u_short usWaypointID )
{
	if ( !AVP_Navigation::IsValidWaypointID( usWaypointID, s_xTestParams ) )
	{
		return true;
	}

	const Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint(usWaypointID);

	if ( s_xTestParams.m_bRejectNonalignedWPs )
	{
		// Allow up to 50 difference so we handle the case where we're between surfaces correctly.
		if ( pxWaypoint->GetNormal() * s_xTestParams.m_xNormal < 0.6427876f )
		{
			return true;
		}
	}

	// left original code as active change to #if 1 to run code that takes radius into account.
#if 1

	float fDistance = (pxWaypoint->m_xPosition - s_xTestPosition).Magnitude();

	fDistance -= pxWaypoint->GetRadius();

	if (fDistance < s_fBestDistance)
	{
		s_fBestDistance  = fDistance;
		s_usBestWaypointID = usWaypointID;
	}

#else

	const float fDistanceSq = ( pxWaypoint->m_xPosition - s_xTestPosition ).MagnitudeSquared();

	if (fDistanceSq < s_fBestDistanceSq)
	{
		s_fBestDistanceSq  = fDistanceSq;
		s_usBestWaypointID = usWaypointID;
	}
#endif

	return true;
}

bool AVP_Navigation::GetNearestWaypointID( const Asura_Vector_3& xPosition, u_short& usNearestWaypointID, const TestParams& xParams /*= s_xTPStandardParams*/ )
{
    // Reset the static search parameters.
    s_xTestPosition    = xPosition;
    s_xTestParams      = xParams;
    s_usBestWaypointID = ASURA_INVALID_WAYPOINT_ID;
    s_fBestDistanceSq  = Asura_Maths::LargeFloat;
	s_fBestDistance	   = Asura_Maths::LargeFloat;

    // Find all of the valid waypoints we're inside. Pick the nearest to us.
    Asura_Bounding_Box xBB;
    xBB.Reset();
    xBB.Set( xPosition );
    xBB.Grow( 10.0f );
    Asura_AI_Navigation_System::SearchWaypoints( xBB, NearestWaypointTreeCallback );
    usNearestWaypointID = s_usBestWaypointID;

    // If we haven't got a hit we're obviously not near enough anything.
    return ( usNearestWaypointID != ASURA_INVALID_WAYPOINT_ID );
}

void AVP_Navigation::UpdateCurrentWaypointID( const Asura_Vector_3& xPosition, u_short& usCurrentWaypointID, const TestParams& xParams /*= s_xTPStandardParams*/ )
{
    if ( usCurrentWaypointID == ASURA_INVALID_WAYPOINT_ID )
    {
        // We'll have to do an exhaustive search.
        GetNearestWaypointID( xPosition, usCurrentWaypointID, xParams );
    }
    else
    {
        // We can use our last value to attempt a best guess.
        // See if we're still in our current waypoint.
        // If we're not then test that waypoint's connections to see if we're in a neighbouring waypoint.
        float   fNearestDistanceSq = Asura_Maths::LargeFloat;
        u_short usNearestWaypoint  = ASURA_INVALID_WAYPOINT_ID;

        const Asura_AI_NavVertex* pxNavVertex = Asura_AI_Navigation_System::GetWaypoint( usCurrentWaypointID );
		if ( pxNavVertex && AVP_Navigation::IsValidWaypointID( usCurrentWaypointID, xParams ) )
        {
			float fDistSq = Asura_Maths::LargeFloat;
            if ( pxNavVertex->Contains( xPosition, &fDistSq ) )
            {
                fNearestDistanceSq = fDistSq;
                usNearestWaypoint  = usCurrentWaypointID;
            }

            for ( u_char ucLinkIndex = 0; ucLinkIndex < pxNavVertex->m_ucNumConnections; ++ucLinkIndex )
            {
                if ( !IsValidLinkID( usCurrentWaypointID, ucLinkIndex, xParams ) )
                {
                    continue;
                }

                const Asura_AI_NavConnection* pxConnection    = &pxNavVertex->m_pxConnections[ ucLinkIndex ];
                const Asura_AI_NavVertex*     pxNextNavVertex = Asura_AI_Navigation_System::GetWaypoint( pxConnection->m_usIndex );
				if ( AVP_Navigation::IsValidWaypointID( pxConnection->m_usIndex, xParams ) && pxNextNavVertex->Contains( xPosition, &fDistSq ) )
				{
					if ( fDistSq < fNearestDistanceSq )
					{
						fNearestDistanceSq = fDistSq;
						usNearestWaypoint  = pxConnection->m_usIndex;
					}
				}
            }
        }

        if ( usNearestWaypoint != ASURA_INVALID_WAYPOINT_ID )
        {
            usCurrentWaypointID = usNearestWaypoint;
        }
        else
        {
            // No good, we'll have to do an exhaustive search.
            GetNearestWaypointID( xPosition, usCurrentWaypointID, xParams );
        }
    }
}

void AVP_Navigation::ResolveToValidPosition( Asura_Vector_3& xPosition, const TestParams& xParams /*= s_xTPStandardParams*/ )
{
    u_short usNearestWaypointID;
    if ( !GetNearestWaypointID( xPosition, usNearestWaypointID, xParams ) )
    {
        Asura_OutputToDebugger( ( "AVP_Navigation::ResolveToValidPosition - failed\n" ) );
        return;
    }

    const Asura_AI_NavVertex* pxNearestWaypoint = Asura_AI_Navigation_System::GetWaypoint( usNearestWaypointID );
    if ( !pxNearestWaypoint ) { return; }

    Asura_Vector_3 xWaypointToPosition = xPosition - pxNearestWaypoint->m_xPosition;
    Asura_Vector_3 xDirection          = xWaypointToPosition;
    const float	   fWaypointToPosition = xDirection.Normalise();

    // If the position is outside the nearest waypoint bring it 0.5m inside.
    if ( fWaypointToPosition > pxNearestWaypoint->GetRadius() )
    {
        xPosition           = pxNearestWaypoint->m_xPosition + ( xDirection * ( pxNearestWaypoint->GetRadius() - 0.5f ) );
        xWaypointToPosition = xPosition - pxNearestWaypoint->m_xPosition;
    }

    // Resolve the position to the surface.
    const Asura_Vector_3& xWaypointNormal = pxNearestWaypoint->GetNormal();
    const float fDotProduct = xWaypointToPosition * xWaypointNormal;
    xPosition = xPosition - ( xWaypointNormal * fDotProduct );
}


bool AVP_Navigation::AtPosition( const Asura_Vector_3& xCurrentPosition, const Asura_Vector_3& xDestinationPosition, const APParams& xParams /*= s_xAPStandardParams*/ )
{
    // Do an early rejection test based on spherical range.
    float fLargestRangeSq = Asura_Maths::Max( xParams.m_fMaxCoplanarOffset, xParams.m_fMaxPerpendicularOffset );
    fLargestRangeSq *= fLargestRangeSq;
    if ( ( xCurrentPosition - xDestinationPosition ).MagnitudeSquared() > fLargestRangeSq )
    {
        return false;
    }

    // Do the full orientated volume test.
    Asura_OrientedVolume xOrientedVolume;
    xOrientedVolume.m_xBox.Reset();
    xOrientedVolume.m_xBox.Set( Asura_Vector_3( -xParams.m_fMaxCoplanarOffset, -xParams.m_fMaxPerpendicularOffset, -xParams.m_fMaxCoplanarOffset ) );
    xOrientedVolume.m_xBox.Set( Asura_Vector_3(  xParams.m_fMaxCoplanarOffset,  xParams.m_fMaxPerpendicularOffset,  xParams.m_fMaxCoplanarOffset ) );
    xOrientedVolume.m_xBox.Displace( xDestinationPosition );

    Asura_Matrix_3x3 xOrientationMatrix;
    Asura_Maths::OrientationFromNormal( xOrientationMatrix, xParams.m_xDestinationNormal );
    xOrientedVolume.m_xOrient.FromMatrix( xOrientationMatrix );

    return ( xOrientedVolume.Contains( xCurrentPosition ) );
}

bool AVP_Navigation::IsValidWaypointID( const u_short usWaypointID, const TestParams& xParams /*= s_xTPStandardParams*/ )
{
    const Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( usWaypointID );

    // If it is not in the correct zone then reject it.
    if ( xParams.m_usZone != ASURA_NAVVERTEX_MAX_ZONE_ID && pxWaypoint->m_usZone != xParams.m_usZone )
    {
        return false;
    }

    // The waypoint must match our pathfinder.
    if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_Navigation::AVP_AI_PATHFINDER_FLAG_HUMANOID ) )
    {
        const bool bIsHumanoidWaypoint = AVP_FlagIsSet( pxWaypoint->m_usFlags, AVP_Navigation::AVP_AI_NAVVERTEX_FLAG_HUMANOID );
        if ( !bIsHumanoidWaypoint )
        {
            return false;
        }
    }
	else if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_Navigation::AVP_AI_PATHFINDER_FLAG_PREDATOR ) )
	{
		const bool bIsPredatorWaypoint = AVP_FlagIsSet( pxWaypoint->m_usFlags, AVP_Navigation::AVP_AI_NAVVERTEX_FLAG_PREDATOR );
		if ( !bIsPredatorWaypoint )
		{
			return false;
		}
	}
	else if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_Navigation::AVP_AI_PATHFINDER_FLAG_ALIEN_HIGH_MOBILITY ) )
    {
        const bool bIsAlienWaypoint = AVP_FlagIsSet( pxWaypoint->m_usFlags, AVP_Navigation::AVP_AI_NAVVERTEX_FLAG_ALIEN );
        if ( !bIsAlienWaypoint )
        {
            return false;
        }

        // High mobility Aliens can travel on any waypoint.
    }
	else if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_Navigation::AVP_AI_PATHFINDER_FLAG_ALIEN_MED_MOBILITY ) )
	{
		const bool bIsAlienWaypoint = AVP_FlagIsSet( pxWaypoint->m_usFlags, AVP_Navigation::AVP_AI_NAVVERTEX_FLAG_ALIEN );
		if ( !bIsAlienWaypoint )
		{
			return false;
		}

// 		Med mobility can only travel on floor/slopes/wall.
// 				const bool bIsCeilingWaypoint = AVP_FlagIsSet( pxWaypoint->m_usFlags, ASURA_AI_NAVVERTEX_FLAG_CEILING );
// 				if ( bIsCeilingWaypoint )
// 				{
// 					return false;
// 				}
	}
	else if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_Navigation::AVP_AI_PATHFINDER_FLAG_ALIEN_LOW_MOBILITY ) )
	{
		const bool bIsAlienWaypoint = AVP_FlagIsSet( pxWaypoint->m_usFlags, AVP_Navigation::AVP_AI_NAVVERTEX_FLAG_ALIEN );
		if ( !bIsAlienWaypoint )
		{
			return false;
		}

// 		// Low mobility can only travel on floor/slopes.
// 		const bool bIsCeilingWallWaypoint = ( AVP_FlagIsSet( pxWaypoint->m_usFlags, ASURA_AI_NAVVERTEX_FLAG_CEILING ) ||
// 											  AVP_FlagIsSet( pxWaypoint->m_usFlags, ASURA_AI_NAVVERTEX_FLAG_WALL ) );
// 		if ( bIsCeilingWallWaypoint )
// 		{
// 			return false;
// 		}
	}

    if ( xParams.m_bRejectNonalignedWPs )
    {
        // If the normal doesn't match our test normal then reject this waypoint.
        // Allow up to 50 difference so we handle the case where we're between surfaces correctly.
        if ( pxWaypoint->GetNormal() * xParams.m_xNormal < 0.6427876f )
        {
            return false;
        }
    }

    return true;
}

bool AVP_Navigation::IsValidLinkID( const u_short usWaypointID, const u_char ucLinkIndex, const TestParams& xParams /*= s_xTPStandardParams*/ )
{
    const Asura_AI_NavVertex*     const pxFromWP = Asura_AI_Navigation_System::GetWaypoint( usWaypointID );
    const Asura_AI_NavConnection* const pxLink   = &pxFromWP->m_pxConnections[ ucLinkIndex ];

    // First test the link flags using the standard functors.
	if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_AI_PATHFINDER_FLAG_HUMANOID ) )
	{
		if ( s_xHumanoidAStarFunctor.LinkFlagsTest( pxLink, xParams.m_uPathfinderFlags) )
		{
			// Remember this function returns whether to ignore the link or not.
			return false;
		}
	}
	else if ( AVP_FlagIsSet( xParams.m_uPathfinderFlags, AVP_AI_PATHFINDER_FLAG_PREDATOR ) )
	{
		if ( s_xHumanoidAStarFunctor.LinkFlagsTest( pxLink, xParams.m_uPathfinderFlags ) )
		{
			// Remember this function returns whether to ignore the link or not.
			return false;
		}
	} 
	else
	{
		if ( s_xHumanoidAStarFunctor.LinkFlagsTest( pxLink, xParams.m_uPathfinderFlags ) )
		{
			// Remember this function returns whether to ignore the link or not.
			return false;
		}
	}

    // Then test the destination waypoint normal, if necessary.
    // (We can skip a bunch of tests in the IsValidWaypointID function, so we don't call that from here.
    //  Specifically, the waypoint's always going to be in the right zone, as it's linked from our current waypoint,
    //  and it should always meet our pathfinder requirements, as the link meets them.)
    if ( xParams.m_bRejectNonalignedWPs )
    {
        const Asura_AI_NavVertex* const pxToWP = Asura_AI_Navigation_System::GetWaypoint( pxLink->m_usIndex );
        // Allow up to 50 difference so we handle the case where we're between surfaces correctly.
        if ( pxToWP->GetNormal() * xParams.m_xNormal < 0.6427876f )
        {
            return false;
        }
    }

    return true;
}

AVP_Navigation::LinkTransitionType AVP_Navigation::GetLinkTransitionType( const u_short usFromWaypointID, const u_short usToWaypointID )
{
    // Get the waypoints.
    const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( usFromWaypointID );
    const Asura_AI_NavVertex* pxToWP   = Asura_AI_Navigation_System::GetWaypoint( usToWaypointID );
    if ( !pxFromWP || !pxToWP )
    {
        Asura_Assert( false, "AVP_Navigation::GetLinkTransitionType - failed to find waypoint(s)" );
        return NUM_LINK_TRANSITION_TYPES;
    }

    // Find the correct link.
    Asura_AI_NavConnection* pxLink = NULL;
    for ( u_char ucLink = 0; ucLink < pxFromWP->m_ucNumConnections && !pxLink; ++ucLink )
    {
        if ( pxFromWP->m_pxConnections[ ucLink ].m_usIndex == usToWaypointID )
        {
            pxLink = &pxFromWP->m_pxConnections[ ucLink ];
        }
    }
    if ( !pxLink )
    {
		// There may be no link associated with this because the route could have been refined.
        return LTT_NO_TRANSITION;
    }

	if ( IsEnvHookLink( pxLink->m_usFlags ) )
	{
		if (pxFromWP->GetFlags() & AVP_AI_NAVVERTEX_FLAG_ALIEN_JUMP_OVER)
		{
			return LTT_ENVHOOK_JUMPOVER;
		}
		
		return LTT_ENVHOOK;
	}

    if ( !IsTransitionLink( pxLink->m_usFlags ) )
    {
        return LTT_NO_TRANSITION;
    }

    // Get the waypoint types.
    enum WaypointType { FLOOR, WALL, CEILING } eFromWPType, eToWPType;
    if ( pxFromWP->IsWallWaypoint() )
    {
        eFromWPType = WALL;
    }
    else if ( pxFromWP->IsCeilingWaypoint() )
    {
        eFromWPType = CEILING;
    }
    else
    {
        eFromWPType = FLOOR;
    }
    if ( pxToWP->IsWallWaypoint() )
    {
        eToWPType = WALL;
    }
    else if ( pxToWP->IsCeilingWaypoint() )
    {
        eToWPType = CEILING;
    }
    else
    {
        eToWPType = FLOOR;
    }

    // Determine if concave or convex.
    const Asura_Vector_3  xLinkVector   = pxToWP->m_xPosition - pxFromWP->m_xPosition;
    const Asura_Vector_3& xToWPNormal   = pxToWP->GetNormal();
    const Asura_Vector_3& xFromWPNormal = pxFromWP->GetNormal();
    const bool bConcave    = ( xLinkVector * xToWPNormal < 0.0f );
    const bool bRightAngle = ( Asura_Maths::Abs( xFromWPNormal * xToWPNormal ) < 0.134f );  // Approx 30 either way.

    // Calculate the transition type from the link flags and the waypoint flags/normals.
    switch ( eFromWPType )
    {
        case WALL:
        {
            switch ( eToWPType )
            {
			case WALL:    return ( bRightAngle ? ( bConcave ? LTT_WALL_TO_WALL_90_CONCAVE : LTT_WALL_TO_WALL_90_CONVEX ) : ( bConcave ? LTT_WALL_TO_WALL_180 : LTT_NO_TRANSITION ));
                case CEILING: return ( bConcave    ? LTT_WALL_TO_CEILING_CONCAVE : LTT_WALL_TO_CEILING_CONVEX );
                case FLOOR:   return ( bConcave    ? LTT_WALL_TO_FLOOR_CONCAVE   : LTT_WALL_TO_FLOOR_CONVEX   );
            }
            break;
        }

        case CEILING:
        {
            switch ( eToWPType )
            {
                case WALL:    return ( bConcave ? LTT_CEILING_TO_WALL_CONCAVE : LTT_CEILING_TO_WALL_CONVEX );
                case CEILING: return LTT_CEILING_TO_CEILING;
                case FLOOR:   return LTT_CEILING_TO_FLOOR;
            }
            break;
        }

        case FLOOR:
        {
            switch ( eToWPType )
            {
                case WALL:    return ( bConcave ? LTT_FLOOR_TO_WALL_CONCAVE : LTT_FLOOR_TO_WALL_CONVEX );
                case CEILING: return LTT_FLOOR_TO_CEILING;
                case FLOOR:   return LTT_FLOOR_TO_FLOOR;
            }
            break;
        }
    }

    return NUM_LINK_TRANSITION_TYPES;
}

AVP_Alien_JumpHandler::AlienJumpType AVP_Navigation::GetJumpTypeFromLinkTransitionType( const LinkTransitionType eLinkTransitionType )
{
    switch ( eLinkTransitionType )
    {
        case LTT_FLOOR_TO_WALL_CONCAVE:   return AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_FLOOR_TO_WALL;
        case LTT_WALL_TO_FLOOR_CONCAVE:   return AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_WALL_TO_FLOOR;
        case LTT_WALL_TO_CEILING_CONCAVE: return AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_WALL_TO_CEILING;
        case LTT_CEILING_TO_WALL_CONCAVE: return AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_CEILING_TO_WALL;
        case LTT_FLOOR_TO_CEILING:        return AVP_Alien_JumpHandler::JT_JUMP_180_FLOOR_TO_CEILING;
        case LTT_CEILING_TO_FLOOR:        return AVP_Alien_JumpHandler::JT_JUMP_180_CEILING_TO_FLOOR;
        case LTT_WALL_TO_WALL_90_CONCAVE: return AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_WALL_TO_WALL;
        case LTT_WALL_TO_WALL_180:        return AVP_Alien_JumpHandler::JT_JUMP_180_WALL_TO_WALL;

        default:
        case LTT_WALL_TO_CEILING_CONVEX:
        case LTT_CEILING_TO_WALL_CONVEX:
        case LTT_FLOOR_TO_WALL_CONVEX:
        case LTT_WALL_TO_FLOOR_CONVEX:
        case LTT_WALL_TO_WALL_90_CONVEX:
        case LTT_NO_TRANSITION:
        case NUM_LINK_TRANSITION_TYPES:   return AVP_Alien_JumpHandler::JT_NORMAL;
    }
}

bool AVP_Navigation::IsStandardLink( const u_short usLinkFlags )
{
	return ( !IsTransitionLink( usLinkFlags ) && !IsGS2Link( usLinkFlags ) && !IsEnvHookLink( usLinkFlags ) );
}

bool AVP_Navigation::IsTransitionLink( const u_short usLinkFlags )
{
    return AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_TRANSITION_LOW_MOBILITY ) ||
           AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_TRANSITION_MED_MOBILITY ) ||
           AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_TRANSITION_HIGH_MOBILITY ) ||
		   AVP_FlagIsSet( usLinkFlags, ASURA_AI_CONNECTION_FLAGS_JUMP );
}

bool AVP_Navigation::IsGS2Link( const u_short usLinkFlags )
{
	return AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_GS2 );
}

bool AVP_Navigation::IsEnvHookLink( const u_short usLinkFlags )
{
	return AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_ENV_HOOK );
}

bool AVP_Navigation::IsLowMobilityLink( const u_short usLinkFlags )
{
    return AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_TRANSITION_LOW_MOBILITY );
}

bool AVP_Navigation::IsMedMobilityLink( const u_short usLinkFlags )
{
    return AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_TRANSITION_MED_MOBILITY );
}

bool AVP_Navigation::IsHighMobilityLink( const u_short usLinkFlags )
{
    return AVP_FlagIsSet( usLinkFlags, AVP_Navigation::AVP_AI_CONNECTION_FLAG_TRANSITION_HIGH_MOBILITY ) ||
           AVP_FlagIsSet( usLinkFlags, ASURA_AI_CONNECTION_FLAGS_JUMP );
}

bool AVP_Navigation::IsConcaveLink( const Asura_Vector_3& xStartPos, const Asura_Vector_3& xFinishPos, const Asura_Vector_3& xFinishNormal )
{
	return ( ( ( xFinishPos - xStartPos ) * xFinishNormal ) < 0.0f );
}

bool AVP_Navigation::FindRoute( NavData& xNavData, AStarFunctorBase* pxFunctor /*= NULL*/, bool bRefineRoute /*= true*/ )
{
#ifdef ASURA_TOOLS
	if ( !xNavData.Verify() ) { return false; }
#endif

	// Make doubly sure that the route is reset.
	xNavData.ResetRoute();

	const Asura_AI_NavVertex* pxCurrentWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usCurrentWPID );
	if ( pxCurrentWP && pxCurrentWP->Contains( xNavData.m_xDestinationPosition ) )
	{
		// No need to create a route if the destination is within the current waypoint.
		xNavData.m_pusRoute = CreateRouteBuffer( 1 );
		xNavData.m_pusOrigRoute = CreateRouteBuffer( 1 );
		if ( xNavData.m_pusRoute )
		{
			xNavData.m_usRouteLength = 1;
			xNavData.m_pusRoute[ 0 ] = xNavData.m_usDestinationWPID;
			xNavData.m_usOrigRouteLength = 1;
			xNavData.m_pusOrigRoute[ 0 ] = xNavData.m_usDestinationWPID;
		}
	}
	else 
	{
		if ( pxFunctor == NULL )
		{
			if ( AVP_FlagIsSet( xNavData.m_uPathfinderFlags, AVP_AI_PATHFINDER_FLAG_HUMANOID ) )
			{
				pxFunctor = &s_xHumanoidAStarFunctor;
			}
			else if ( AVP_FlagIsSet( xNavData.m_uPathfinderFlags, AVP_AI_PATHFINDER_FLAG_PREDATOR ) )
			{
				pxFunctor = &s_xPredatorAStarFunctor;
			}
			else
			{
				pxFunctor = &s_xAlienAStarFunctor;
			}
		}

		int iRouteLength = 0;
		Asura_AI_Navigation_System::FindRouteCustom( GetWorkingRouteBuffer(), iRouteLength, GetWorkingRouteBufferLength(),
			xNavData.m_usCurrentWPID, xNavData.m_usDestinationWPID, xNavData.m_uPathfinderFlags, pxFunctor,
			FunctorGoalCostFunction,
			FunctorTraversalCostFunction,
			FunctorGoalAchievedFunction,
			FunctorLinkFlagsTestFunction );
		if ( iRouteLength )
		{
#ifndef ASURA_NO_DEBUG_OUTPUT
			s_usPreLength = static_cast< u_short >( iRouteLength );
			Asura_Mem::Copy( s_ausPreRouteBuffer, GetWorkingRouteBuffer(), sizeof( u_short ) * s_usPreLength );
#endif

			// Bit of trickery to do the route refining before we allocate any route memory.
			xNavData.m_pusRoute      = GetWorkingRouteBuffer();
			xNavData.m_usRouteLength = static_cast< u_short >( iRouteLength );

			xNavData.m_uConnectionGeneration = Asura_AI_Navigation_System::GetConnectionGeneration();
			xNavData.m_usOrigRouteLength = xNavData.m_usRouteLength;
			xNavData.m_pusOrigRoute = CreateRouteBuffer( xNavData.m_usOrigRouteLength );
			if ( xNavData.m_pusOrigRoute )
			{
				Asura_Mem::Copy( xNavData.m_pusOrigRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xNavData.m_usOrigRouteLength );
			}

			if ( bRefineRoute )
			{
				const TestParams xTestParams( xNavData.m_uPathfinderFlags, xNavData.m_usCurrentZone );
				RefineRoute( xNavData, xTestParams );
			}
			xNavData.m_pusRoute = CreateRouteBuffer( xNavData.m_usRouteLength );
			if ( xNavData.m_pusRoute )
			{
				Asura_Mem::Copy( xNavData.m_pusRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xNavData.m_usRouteLength );

#ifndef ASURA_NO_DEBUG_OUTPUT
				s_usPostLength = xNavData.m_usRouteLength;
				Asura_Mem::Copy( s_ausPostRouteBuffer, xNavData.m_pusRoute, sizeof( u_short ) * s_usPostLength );
#endif
			}
		}
	}

	return ( xNavData.m_pusOrigRoute && xNavData.m_pusRoute && xNavData.m_usRouteLength > 0 );
}

bool AVP_Navigation::FindRouteAwayFrom( NavData& xNavData, AStarFunctorBase* pxFunctor, bool bRefineRoute /*= true*/ )
{
	// Make doubly sure that the route is reset.
	xNavData.ResetRoute();

	int iRouteLength = 0;
	Asura_AI_Navigation_System::FindRouteCustom( GetWorkingRouteBuffer(), iRouteLength, GetWorkingRouteBufferLength(),
		xNavData.m_usCurrentWPID, xNavData.m_uPathfinderFlags, pxFunctor,
		FunctorGoalCostFunction,
		FunctorTraversalCostFunction,
		FunctorGoalAchievedFunction,
		FunctorLinkFlagsTestFunction );
	if ( iRouteLength )
	{
		// Bit of trickery to do the route refining before we allocate any route memory.
		xNavData.m_pusRoute      = GetWorkingRouteBuffer();
		xNavData.m_usRouteLength = static_cast< u_short >( iRouteLength );

		xNavData.m_uConnectionGeneration = Asura_AI_Navigation_System::GetConnectionGeneration();
		xNavData.m_usOrigRouteLength = xNavData.m_usRouteLength;
		xNavData.m_pusOrigRoute = CreateRouteBuffer( xNavData.m_usOrigRouteLength );
		if ( xNavData.m_pusOrigRoute )
		{
			Asura_Mem::Copy( xNavData.m_pusOrigRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xNavData.m_usOrigRouteLength );
		}


		const TestParams xTestParams( xNavData.m_uPathfinderFlags, xNavData.m_usCurrentZone );
		RefineRoute( xNavData, xTestParams );
		xNavData.m_pusRoute = CreateRouteBuffer( xNavData.m_usRouteLength );
		if ( xNavData.m_pusRoute )
		{
			Asura_Mem::Copy( xNavData.m_pusRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xNavData.m_usRouteLength );
			xNavData.m_usDestinationWPID    = xNavData.m_pusRoute[ xNavData.m_usRouteLength - 1 ];
			xNavData.m_xDestinationPosition = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usDestinationWPID )->m_xPosition;
		}
	}

	return ( xNavData.m_pusRoute && xNavData.m_usRouteLength > 0 );
}

bool AVP_Navigation::AdoptRoute( NavData& xToNavData, const NavData& xFromNavData, const Asura_Vector_3& xAdopterPosition, AStarFunctorBase* pxFunctor, bool bRefineRoute /*= true*/ )
{
	if ( xFromNavData.m_usRouteLength == 0 )
	{
		return false;
	}

	xToNavData.ResetRoute();
	xToNavData.m_xStartPosition = xAdopterPosition;
	xToNavData.m_xDestinationPosition = xFromNavData.m_xDestinationPosition;
	xToNavData.m_usDestinationWPID = xFromNavData.m_usDestinationWPID;
	xToNavData.m_usRouteLength = xFromNavData.m_usRouteLength;
	xToNavData.m_pusRoute = AVP_Navigation::CreateRouteBuffer( xToNavData.m_usRouteLength );
	if ( !xToNavData.m_pusRoute )
	{
		return false;
	}

	// Whilst we are copying the route, try and determine where we are on the route.
	u_short usCurrentIdx = xToNavData.m_usRouteLength;
	for ( u_short usIdx = 0; usIdx < xToNavData.m_usRouteLength; ++usIdx )
	{
		xToNavData.m_pusRoute[ usIdx ] = xFromNavData.m_pusRoute[ usIdx ];
		if ( xToNavData.m_pusRoute[ usIdx ] == xToNavData.m_usCurrentWPID )
		{
			usCurrentIdx = usIdx;
		}
	}

	if ( usCurrentIdx < xToNavData.m_usRouteLength )
	{

		xToNavData.m_usOrigRouteLength = xFromNavData.m_usOrigRouteLength;
		xToNavData.m_pusOrigRoute = CreateRouteBuffer( xToNavData.m_usOrigRouteLength );
		xToNavData.m_uConnectionGeneration = xFromNavData.m_uConnectionGeneration;
		if ( xToNavData.m_pusOrigRoute )
		{
			Asura_Mem::Copy( xToNavData.m_pusOrigRoute, xFromNavData.m_pusOrigRoute, sizeof( u_short ) * xToNavData.m_usOrigRouteLength );
		}
		else
		{
			return false;
		}


		// We are already on the route - fantastic.
		xToNavData.m_usRouteSegmentIndex = usCurrentIdx;
		return true;
	}
	else if ( xToNavData.m_usRouteLength == 1 )
	{
		// We need to route-find to the single WP in the route.
		int iRouteLength = 0;
		Asura_AI_Navigation_System::FindRouteCustom( GetWorkingRouteBuffer(), iRouteLength, GetWorkingRouteBufferLength(),
			xToNavData.m_usCurrentWPID, xFromNavData.m_pusRoute[ 0 ], xToNavData.m_uPathfinderFlags, pxFunctor,
			FunctorGoalCostFunction,
			FunctorTraversalCostFunction,
			FunctorGoalAchievedFunction,
			FunctorLinkFlagsTestFunction );
		if ( iRouteLength > 0 )
		{
			xToNavData.ResetRoute();
			xToNavData.m_usRouteLength = static_cast< u_short >( iRouteLength );
			xToNavData.m_pusRoute = AVP_Navigation::CreateRouteBuffer( xToNavData.m_usRouteLength );
			xToNavData.m_pusOrigRoute = AVP_Navigation::CreateRouteBuffer( xToNavData.m_usRouteLength );
			if ( xToNavData.m_pusRoute && xToNavData.m_pusOrigRoute)
			{
				Asura_Mem::Copy( xToNavData.m_pusRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xToNavData.m_usRouteLength );
				xToNavData.m_uConnectionGeneration = Asura_AI_Navigation_System::GetConnectionGeneration();
				xToNavData.m_usOrigRouteLength = xToNavData.m_usRouteLength;
				Asura_Mem::Copy( xToNavData.m_pusOrigRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xToNavData.m_usOrigRouteLength );

				return true;
			}
			else
			{
				return false;
			}
		}
		else
		{
			return false;
		}
	}
	else
	{
		xToNavData.m_usRouteSegmentIndex = 0;

		// We need to route-find to the nearest point on the route.
		Asura_Vector_3 xNearestPosition;
		u_short usNearestSegmentIdx;
		if ( AVP_Navigation::FindNearestPointOnRoute( xToNavData, xAdopterPosition, 0.0f, xNearestPosition, usNearestSegmentIdx ) )
		{
			// We're at the end, set us to the last segment.
			usNearestSegmentIdx = xToNavData.m_usRouteLength - 2;
		}

		// Route-find to the WP at the end of the nearest segment.
		const u_short usTargetWPIndex = usNearestSegmentIdx + 1;

		// find the target in the orig route 
		bool bFound = false;
		u_short usTargetOrigWPIndex = 0;
		for (u_short u = 0; u < xFromNavData.m_usOrigRouteLength && !bFound; ++u )
		{
			if (xFromNavData.m_pusOrigRoute[u] == xToNavData.m_pusRoute[usTargetWPIndex])
			{
				usTargetOrigWPIndex = u; 
				bFound = true;
			}
		}

		if (!bFound)
		{
			return false;
		}

		int iRouteLength = 0;
		Asura_AI_Navigation_System::FindRouteCustom( GetWorkingRouteBuffer(), iRouteLength, GetWorkingRouteBufferLength(),
			xToNavData.m_usCurrentWPID, xFromNavData.m_pusRoute[ usTargetWPIndex ], xToNavData.m_uPathfinderFlags, pxFunctor,
			FunctorGoalCostFunction,
			FunctorTraversalCostFunction,
			FunctorGoalAchievedFunction,
			FunctorLinkFlagsTestFunction );
		if ( iRouteLength > 0 )
		{

			xToNavData.ResetRoute();

			int iOrigRouteLength = iRouteLength;
			// Tack the main route (from the usTargetWPIndex + 1 onwards) onto the end of this route.
			// '+ 1' because the last WP in the new route will be the same as the target WP in the main route.
			for ( u_short usIdx = usTargetWPIndex + 1; usIdx < xFromNavData.m_usRouteLength; ++usIdx )
			{
				GetWorkingRouteBuffer()[ iRouteLength++ ] = xFromNavData.m_pusRoute[ usIdx ];
			}
			xToNavData.m_usRouteLength = static_cast< u_short >( iRouteLength );
			xToNavData.m_pusRoute = AVP_Navigation::CreateRouteBuffer( xToNavData.m_usRouteLength );
			if (xToNavData.m_pusRoute)
			{
				Asura_Mem::Copy( xToNavData.m_pusRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xToNavData.m_usRouteLength );
			}
			else
			{
				xToNavData.ResetRoute();
				return false;
			}

			// Tack the main route (from the usTargetWPIndex + 1 onwards) onto the end of this route.
			// '+ 1' because the last WP in the new route will be the same as the target WP in the main route.
			for ( u_short usIdx = usTargetOrigWPIndex + 1; usIdx < xFromNavData.m_usOrigRouteLength; ++usIdx )
			{
				GetWorkingRouteBuffer()[ iOrigRouteLength++ ] = xFromNavData.m_pusOrigRoute[ usIdx ];
			}
			xToNavData.m_usOrigRouteLength = static_cast< u_short >( iOrigRouteLength );
			xToNavData.m_pusOrigRoute = AVP_Navigation::CreateRouteBuffer( xToNavData.m_usOrigRouteLength );
			if (xToNavData.m_pusOrigRoute)
			{
				Asura_Mem::Copy( xToNavData.m_pusOrigRoute, GetWorkingRouteBuffer(), sizeof( u_short ) * xToNavData.m_usOrigRouteLength );
			}
			else
			{
				xToNavData.ResetRoute();
				return false;
			}

			return true;
		}
		else
		{
			return false;
		}
	}
}

void AVP_Navigation::SearchWaypoints( const u_int uPathfinderFlags, SearchFunctor* pxSearchFunctor )
{
	int iRouteLength = 0;
	Asura_AI_Navigation_System::FindRouteCustom( GetWorkingRouteBuffer(), iRouteLength, GetWorkingRouteBufferLength(),
		pxSearchFunctor->GetStartWPID(), uPathfinderFlags, pxSearchFunctor,
		FunctorGoalCostFunction,
		FunctorTraversalCostFunction,
		FunctorGoalAchievedFunction,
		FunctorLinkFlagsTestFunction );
}

void AVP_Navigation::CalculateRouteAgility( NavData& xNavData )
{
    xNavData.m_eRouteAgilityRating = AR_LOW;

    bool bDone = false;
    for ( u_short usRouteIndex = 0; usRouteIndex < xNavData.m_usRouteLength && !bDone; ++usRouteIndex )
    {
        const Asura_AI_NavVertex* pxWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usRouteIndex ] );
        if ( !pxWP ) { continue; }

        if ( pxWP->IsCeilingWaypoint() )
        {
            xNavData.m_eRouteAgilityRating = AR_HIGH;
            // Rating can't go any higher, so we're done.
            bDone = true;
        }
        else if ( pxWP->IsWallWaypoint() )
        {
            xNavData.m_eRouteAgilityRating = AR_MED;
        }
    }
}

//--------------------------------------------------------------------------------------------------------
// AVP_Navigation :: CalculateRouteDistance
//--------------------------------------------------------------------------------------------------------
float AVP_Navigation::CalculateRouteDistance( const NavData& xNavData, const Asura_Vector_3& xStartPosition, u_short uStartRouteIndex )
{
	u_short uStartIndexBackup = uStartRouteIndex;

	Asura_Assert( xNavData.m_pusRoute, "AVP_Navigation :: CalculateRouteDistance - Route NavData Is Not Valid , Route Distance Will Be Returned As -1.0f, This happens when a route find has failed, It is likley that the associated actor cannot find a route please check the waypointing in the relevant area" );

	if ( xNavData.m_pusRoute )
	{
		// Route length is only 1, so just return the distance from the start point to the destination
		if ( ( xNavData.m_usRouteLength == 1 ) || ( uStartRouteIndex == xNavData.m_usRouteLength ) )
		{
			return ( xStartPosition - xNavData.m_xDestinationPosition ).Magnitude();
		}

		//  Some times we have a route length of 2 and it can be basically a 1 length route.
		if ( ( xNavData.m_usRouteLength == 2 ) && ( uStartRouteIndex == xNavData.m_usRouteLength - 2 ) )
		{
			const Asura_Vector_3	xNearestWaypointPos( Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ uStartRouteIndex ] )->m_xPosition );
			const Asura_Vector_3	xNextWaypointPos( Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ uStartRouteIndex + 1 ] )->m_xPosition );
			const Asura_Vector_3	xActorToNearestDirection( xNearestWaypointPos - xStartPosition );
			const Asura_Vector_3	xActorToNextDirection( xNextWaypointPos - xStartPosition );
			const float				fActorDot = xActorToNearestDirection * xActorToNextDirection;

			if (fActorDot < 0.0f)
			{
				return ( xStartPosition - xNavData.m_xDestinationPosition ).Magnitude();
			}
		}


		if ( uStartRouteIndex < xNavData.m_usRouteLength - 1 )
		{
			const Asura_Vector_3	xFirstWaypointPos( Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ uStartRouteIndex ] )->m_xPosition );
			const Asura_Vector_3	xSecondWaypointPos( Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ uStartRouteIndex + 1 ] )->m_xPosition );

			Asura_Vector_3 xDirToFirstWp( xFirstWaypointPos - xStartPosition );
			Asura_Vector_3 xDirToSecondWp( xSecondWaypointPos - xStartPosition );

			float fDot = xDirToFirstWp * xDirToSecondWp;

			if (fDot < 0.0f)
			{
				++uStartRouteIndex;
			}

			float fDistanceToWP1 = (xStartPosition - xFirstWaypointPos).MagnitudeSquared();
			float fDistanceToWP2 = (xStartPosition - xSecondWaypointPos).MagnitudeSquared();

			if ( ( fDistanceToWP2 < fDistanceToWP1 ) && ( uStartIndexBackup == uStartRouteIndex ))
			{
				++uStartRouteIndex;
			}
		}

		// Find the first waypoint on our route
		const Asura_AI_NavVertex* pxFirstWaypoint = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ uStartRouteIndex ] );

		// Calculate the distance from our starting position to the first waypoint
		float fRouteDistance = ( xStartPosition - pxFirstWaypoint->m_xPosition ).Magnitude();

		// Loop round our route starting with our current route index [route length - 2 because we never end up traveling to the last waypoint]
		for ( u_short usWP = uStartRouteIndex; usWP < xNavData.m_usRouteLength - 2; ++usWP )
		{
			// Find the waypoint on our route
			const Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usWP ] );

			bool bFound = false;

			// Loop round this waypoints links to find the next waypoint on our route and grab the length
			for ( u_char uc = 0; uc < pxWaypoint->m_ucNumConnections; ++uc )
			{
				// Check if the connection index from the links is the same as the next waypoint on our route
				if ( pxWaypoint->m_pxConnections[ uc ].m_usIndex == xNavData.m_pusRoute[ usWP + 1 ] )
				{
					bFound = true;
					// add the distance taken from the connection length
					fRouteDistance += pxWaypoint->m_pxConnections[ uc ].GetLength();
				}
			}

			// For smoothed portions of routes - we will have to do the more expensive magnitude calculation
			if ( !bFound )
			{
				const Asura_AI_NavVertex* pxTestWaypoint = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usWP + 1 ] );
				fRouteDistance += ( pxWaypoint->GetPosition() - pxTestWaypoint->GetPosition() ).Magnitude();
			}
		}

		// Find the last waypoint on our route [route length - 2 because we never end up traveling to the last waypoint]
		const Asura_AI_NavVertex* pxLastWaypoint = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ xNavData.m_usRouteLength - 2 ] );

		// Finally add on the distance form our last waypoint to the destination position in the route
		fRouteDistance += ( xNavData.m_xDestinationPosition - pxLastWaypoint->m_xPosition ).Magnitude();

		return fRouteDistance;
	}
	else
	{
		return -1.0f;
	}
}

void AVP_Navigation::FindNearestPointOnOtherRoute( const NavData& xNavData, const Asura_Vector_3& xCurrentPosition, Asura_Vector_3& xNearestPosition, u_short& usNearestRouteSegmentIndex )
{
	if (xNavData.m_usRouteLength == 0)
	{
		Asura_Assert(0, "AVP_Navigation::FindNearestPointOnRoute - zero route length");
		xNearestPosition = xCurrentPosition;
		usNearestRouteSegmentIndex = 0;
		return;
	}

	Asura_Vector_3 xCurrentSegmentStart;
	Asura_Vector_3 xCurrentSegmentEnd;

	if ( usNearestRouteSegmentIndex < xNavData.m_usRouteLength - 1 )
	{
		// find the closest waypoint
		float fLeastDistanceSq = Asura_Maths::LargeFloat;

		for ( u_short usSegmentIndex = usNearestRouteSegmentIndex; usSegmentIndex < xNavData.m_usRouteLength - 1; ++usSegmentIndex )
		{
			xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usSegmentIndex ] )->m_xPosition;
			xCurrentSegmentEnd   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usSegmentIndex + 1 ] )->m_xPosition;

			Asura_Vector_3 xPosition;
			Asura_Maths::GetNearestPointOnLineSegment( xCurrentSegmentStart, xCurrentSegmentEnd, xCurrentPosition, xPosition );

			const float fDistanceSq = ( xPosition - xCurrentPosition ).MagnitudeSquared();
			const float fYDiffDist = Asura_Maths::Abs( xPosition.y - xCurrentPosition.y );

			if ( fYDiffDist > 4.0f )
			{
				continue;
			}

			if ( fDistanceSq < fLeastDistanceSq )
			{
				fLeastDistanceSq			= fDistanceSq;
				usNearestRouteSegmentIndex	= usSegmentIndex;
				xNearestPosition			= xPosition;
			}
		}
	}
	else
    {
		Asura_AI_NavVertex* pxTemp = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usDestinationWPID );

		Asura_Assert(pxTemp, "Invalid waypoint ID");

		if(pxTemp)
		{
			xNearestPosition = pxTemp->m_xPosition;
		}
    }
}

bool AVP_Navigation::FindNearestPointOnRoute( const NavData& xNavData, const Asura_Vector_3& xCurrentPosition, float fLookAheadDistance,
                                              Asura_Vector_3& xNearestPosition, u_short& usNearestRouteSegmentIndex )
{
	if (xNavData.m_usRouteLength == 0)
	{
		Asura_Assert(0, "AVP_Navigation::FindNearestPointOnRoute - zero route length");
		xNearestPosition = xCurrentPosition;
		return false;
	}

    if ( xNavData.m_usRouteSegmentIndex < xNavData.m_usRouteLength - 1 )
    {
        // Work out the nearest position to us on our route plus our look ahead distance.
        /*
            For ( all segments from this segment until the end of the route )
            {
                Get the nearest point on the segment.
                If ( nearest point distance to us < best nearest point distance to us )
                {
                    best nearest point = nearest point
                    best segment = current segment
                }
            }
            If ( we are at the end of the route )
            {
                we're done - return true
            }

            Do
            {
                Add on look ahead distance in direction of best segment until segment end.
                Calc distance left.
                If ( distance left )
                    Move to next segment.
            } while ( distance left )
        */

        Asura_Vector_3 xCurrentSegmentStart;
        Asura_Vector_3 xCurrentSegmentEnd;

        usNearestRouteSegmentIndex = xNavData.m_usRouteSegmentIndex;

        float fLeastDistanceSq = Asura_Maths::LargeFloat;
        for ( u_short usSegmentIndex = xNavData.m_usRouteSegmentIndex; usSegmentIndex < xNavData.m_usRouteLength - 1; ++usSegmentIndex )
        {
            xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usSegmentIndex ] )->m_xPosition;
            xCurrentSegmentEnd   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usSegmentIndex + 1 ] )->m_xPosition;

            Asura_Vector_3 xPosition;
            Asura_Maths::GetNearestPointOnLineSegment( xCurrentSegmentStart, xCurrentSegmentEnd, xCurrentPosition, xPosition );
            const float fDistanceSq = ( xPosition - xCurrentPosition ).MagnitudeSquared();
            if ( fDistanceSq < fLeastDistanceSq )
            {
                fLeastDistanceSq    = fDistanceSq;
                usNearestRouteSegmentIndex = usSegmentIndex;
                xNearestPosition    = xPosition;
            }
        }

        xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex ] )->m_xPosition;
        xCurrentSegmentEnd   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex + 1 ] )->m_xPosition;
        if ( usNearestRouteSegmentIndex == xNavData.m_usRouteLength - 2 && xNearestPosition.ApproximatelyEquals( xCurrentSegmentEnd ) )
        {
            return true;
        }

        while ( fLookAheadDistance > 0.0f )
        {
            Asura_Vector_3 xSegmentDirection = xCurrentSegmentEnd - xCurrentSegmentStart;
            xSegmentDirection.Normalise();
            const float fToSegmentEnd = ( xCurrentSegmentEnd - xNearestPosition ).Magnitude();

            xNearestPosition   += xSegmentDirection * Asura_Maths::Min( fLookAheadDistance, fToSegmentEnd );
            fLookAheadDistance -= fToSegmentEnd;
            if ( fLookAheadDistance > 0.0f )
            {
                if ( ++usNearestRouteSegmentIndex < xNavData.m_usRouteLength - 1 )
                {
					if ( xNavData.m_pusRoute[ usNearestRouteSegmentIndex ] < Asura_AI_Navigation_System::GetNumberOfWaypoints() && 
						xNavData.m_pusRoute[ usNearestRouteSegmentIndex + 1] < Asura_AI_Navigation_System::GetNumberOfWaypoints() && 
						xNavData.m_pusRoute[ usNearestRouteSegmentIndex ] != ASURA_INVALID_WAYPOINT_ID && 
						xNavData.m_pusRoute[ usNearestRouteSegmentIndex + 1 ] != ASURA_INVALID_WAYPOINT_ID )
					{
						xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex ] )->m_xPosition;
						xCurrentSegmentEnd   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex + 1 ] )->m_xPosition;
					}
                }
                else
                {
					if ( xNavData.m_usDestinationWPID < Asura_AI_Navigation_System::GetNumberOfWaypoints() && xNavData.m_usDestinationWPID != ASURA_INVALID_WAYPOINT_ID )
					{
						xNearestPosition = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usDestinationWPID )->m_xPosition;
						usNearestRouteSegmentIndex = xNavData.m_usRouteLength - 2;
						return true;
					}	
					else
					{
						Asura_Assert(false, "Invalid destination WPID"); 
						return false; 
					}
                }
            }
        }

        return false;
    }
    else
    {
		Asura_AI_NavVertex* pxTemp = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usDestinationWPID );
		Asura_Assert(pxTemp, "Invalid waypoint ID");
		if(pxTemp)
		{
			xNearestPosition = pxTemp->m_xPosition;
		}
		return true;   
    }
}

// Similar to FindNearestPointOnRoute - but use only the XZ plane
// This function help avoid turning in circle when droping a ledge.
bool AVP_Navigation::FindNearestPointOnRouteHorizontal( const NavData& xNavData, const Asura_Vector_3& xCurrentPosition3D, float fLookAheadDistance,
														Asura_Vector_3& xNearestPosition, u_short& usNearestRouteSegmentIndex )
{

	if (xNavData.m_usRouteLength == 0)
	{
		Asura_Assert(0, "AVP_Navigation::FindNearestPointOnRoute - zero route length");
		xNearestPosition = xCurrentPosition3D;
		return false;
	}

	if ( xNavData.m_usRouteSegmentIndex < xNavData.m_usRouteLength - 1 )
	{
		Asura_Vector_3 xCurrentSegmentStart;
		Asura_Vector_3 xCurrentSegmentEnd;

		Asura_Vector_3 xCurrentPosition = xCurrentPosition3D; 
		xCurrentPosition.SetY(0.0f);

		usNearestRouteSegmentIndex = xNavData.m_usRouteSegmentIndex;

		float fLeastDistanceSq = Asura_Maths::LargeFloat;
		for ( u_short usSegmentIndex = xNavData.m_usRouteSegmentIndex;
			usSegmentIndex < xNavData.m_usRouteLength - 1 && usSegmentIndex < xNavData.m_usRouteSegmentIndex + 5;
			++usSegmentIndex )
		{
			xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usSegmentIndex ] )->m_xPosition;
			xCurrentSegmentStart.SetY(0.0f);
			xCurrentSegmentEnd = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usSegmentIndex + 1 ] )->m_xPosition;
			xCurrentSegmentEnd.SetY(0.0f);

			Asura_Vector_3 xPosition;
			Asura_Maths::GetNearestPointOnLineSegment( xCurrentSegmentStart, xCurrentSegmentEnd, xCurrentPosition, xPosition );
			const float fDistanceSq = ( xPosition - xCurrentPosition ).MagnitudeSquared();
			if ( fDistanceSq < fLeastDistanceSq )
			{
				fLeastDistanceSq    = fDistanceSq;
				usNearestRouteSegmentIndex = usSegmentIndex;
				xNearestPosition    = xPosition;				
			}
		}

		xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex ] )->m_xPosition;
		xCurrentSegmentStart.SetY(0.0f);
		xCurrentSegmentEnd   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex + 1 ] )->m_xPosition;
		xCurrentSegmentEnd.SetY(0.0f);

		if ( usNearestRouteSegmentIndex == xNavData.m_usRouteLength - 1 && xNearestPosition.ApproximatelyEquals( xCurrentSegmentEnd ) )
		{
			xNearestPosition.SetY(xCurrentPosition3D.GetY());
			return true;
		}

		while ( fLookAheadDistance > 0.0f )
		{
			Asura_Vector_3 xSegmentDirection = xCurrentSegmentEnd - xCurrentSegmentStart;
			xSegmentDirection.Normalise();


			const float fToSegmentEnd = ( xCurrentSegmentEnd - xNearestPosition ).Magnitude();

			xNearestPosition   += xSegmentDirection * Asura_Maths::Min( fLookAheadDistance, fToSegmentEnd );
			fLookAheadDistance -= fToSegmentEnd;
			fLookAheadDistance -= fToSegmentEnd * Asura_Maths::Sqrt(xSegmentDirection.x * xSegmentDirection.x + xSegmentDirection.y * xSegmentDirection.y );
			if ( fLookAheadDistance > 0.0f )
			{
				if ( ++usNearestRouteSegmentIndex < xNavData.m_usRouteLength - 1 )
				{
					xCurrentSegmentStart = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex ] )->m_xPosition;
					xCurrentSegmentStart.SetY(0.0f);
					xCurrentSegmentEnd   = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usNearestRouteSegmentIndex + 1 ] )->m_xPosition;
					xCurrentSegmentEnd.SetY(0.0f);

				}
				else
				{
					xNearestPosition = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usDestinationWPID )->m_xPosition;
					xNearestPosition.SetY(xCurrentPosition3D.GetY());
					return true;
				}
			}
		}

		xNearestPosition.SetY(xCurrentPosition3D.GetY());
		return false;
	}
	else
	{
		Asura_AI_NavVertex* pxTemp = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usDestinationWPID );
		Asura_Assert(pxTemp, "Invalid waypoint ID");
		if(pxTemp)
		{
			xNearestPosition = pxTemp->m_xPosition;
			xNearestPosition.SetY(xCurrentPosition3D.GetY());
		}
		return true;   
	}
}


float AVP_Navigation::FunctorTraversalCostFunction( const int iWaypointIndex, const int iLinkIndex, u_int uPathfinderFlags, void* pUserData )
{
	AVP_Navigation::AStarFunctorBase &xFunctor = *static_cast<AVP_Navigation::AStarFunctorBase *>(pUserData);
	return xFunctor.EstimateTraversalCost(iWaypointIndex, iLinkIndex, uPathfinderFlags);
}

bool AVP_Navigation::FunctorLinkFlagsTestFunction( const Asura_AI_NavConnection* pxLink, const u_int uPathfinderFlags, void* pUserData )
{
	AVP_Navigation::AStarFunctorBase &xFunctor = *static_cast<AVP_Navigation::AStarFunctorBase *>(pUserData);
	return xFunctor.LinkFlagsTest(pxLink, uPathfinderFlags);
}

float AVP_Navigation::FunctorGoalCostFunction( const int iWaypointIndex, void* pUserData )
{
	AVP_Navigation::AStarFunctorBase &xFunctor = *static_cast<AVP_Navigation::AStarFunctorBase *>(pUserData);
	return xFunctor.EstimateCostToGoal(iWaypointIndex);
}

bool AVP_Navigation::FunctorGoalAchievedFunction( const int iWaypointIndex, void* pUserData )
{
	AVP_Navigation::AStarFunctorBase &xFunctor = *static_cast<AVP_Navigation::AStarFunctorBase *>(pUserData);
	return xFunctor.IsGoalAchieved(iWaypointIndex);
}


bool AVP_Navigation::IsRouteBlocked(const NavData& xNavData)
{
	if (xNavData.m_uConnectionGeneration == 0 || !xNavData.m_pusOrigRoute) // generation zero is invalid
	{
		return true; 
	}

	if (xNavData.m_uConnectionGeneration == Asura_AI_Navigation_System::GetConnectionGeneration() || xNavData.m_usOrigRouteLength < 1)
	{
		return false;
	}

	for (u_short u = 0; u < xNavData.m_usOrigRouteLength - 1; ++u)
	{
		const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusOrigRoute[ u ] );

		if (!pxFromWP) // something wrong
		{
			return true;
		}
		const Asura_AI_NavConnection* pxLink = pxFromWP->GetLinkTo( xNavData.m_pusOrigRoute[ u + 1 ] );
		if (!pxLink) // something wrong
		{
			return true;
		}
		if (pxLink->GetFlags() & ASURA_AI_CONNECTION_FLAGS_BLOCKED)
		{
			return true;
		}

	}

	// since this route is compatible with the new generation update the generation number
	xNavData.m_uConnectionGeneration = Asura_AI_Navigation_System::GetConnectionGeneration();

	return false;
}


#ifndef ASURA_NO_DEBUG_OUTPUT
void AVP_Navigation::RenderRoute( const NavData& xNavData )
{
    if ( xNavData.m_pusRoute && xNavData.m_usRouteLength > 0 )
    {
        for ( u_short us = 0; us < xNavData.m_usRouteLength - 1; ++us )
        {
            const Asura_AI_NavVertex* pxWP1 = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ us ] );
            const Asura_AI_NavVertex* pxWP2 = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ us + 1 ] );

			if (pxWP1 && pxWP2)
			{
				Asura_Vector_3 xLocalUp1 = pxWP1->GetNormal();
				xLocalUp1 *= 0.125f;
				Asura_Vector_3 xLocalUp2 = pxWP2->GetNormal();
				xLocalUp2 *= 0.125f;

				Asura_DebugLineRender::AddLine( pxWP1->m_xPosition + xLocalUp1, pxWP2->m_xPosition + xLocalUp2, 0xffffffff );
				Asura_DebugLineRender::AddSphere( pxWP1->m_xPosition, 0.5f, 0xffffffff, 4, 2 );

				Asura_Vector_3 xOutputPos = pxWP1->GetPosition();

				// GBT : Left commented out as it doesn't need to be in can be useful for debugging, writes the waypoint ID in the little sphere we are drawing for each route node.
// 				if (Asura_Camera::ProjectWorldCoordsToScreenSpace( xOutputPos, true ) || xOutputPos.z <= 0.0f)
// 				{
// 					Asura_Font_System::Print(xOutputPos.x, xOutputPos.y, 0.5f, 0xffffffff, ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, "%d", xNavData.m_pusRoute[ us ] );
// 				}
			}
        }
        const Asura_AI_NavVertex* pxWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ xNavData.m_usRouteLength - 1 ] );

		if (pxWP)
		{
	        Asura_DebugLineRender::AddSphere( pxWP->m_xPosition, 0.5f, 0xffffffff, 4, 2 );
			Asura_DebugLineRender::AddSphere( xNavData.m_xDestinationPosition, 1.0f, 0xffffb0ff, 4, 2 );
		}
    }


    if ( s_usPreLength )
    {
        for ( u_short us = 0; us < s_usPreLength - 1; ++us )
        {
            const Asura_AI_NavVertex* pxWP1 = Asura_AI_Navigation_System::GetWaypoint( s_ausPreRouteBuffer[ us ] );
            const Asura_AI_NavVertex* pxWP2 = Asura_AI_Navigation_System::GetWaypoint( s_ausPreRouteBuffer[ us + 1 ] );

			if (pxWP1 && pxWP2)
			{
				Asura_Vector_3 xLocalUp1 = pxWP1->GetNormal();
				xLocalUp1 *= 0.125f;
				Asura_Vector_3 xLocalUp2 = pxWP2->GetNormal();
				xLocalUp2 *= 0.125f;

				Asura_DebugLineRender::AddLine( pxWP1->m_xPosition + xLocalUp1, pxWP2->m_xPosition + xLocalUp2, 0xffff0000 );
			}
        }
    }
    if ( s_usPostLength )
    {
        for ( u_short us = 0; us < s_usPostLength - 1; ++us )
        {
            const Asura_AI_NavVertex* pxWP1 = Asura_AI_Navigation_System::GetWaypoint( s_ausPostRouteBuffer[ us ] );
            const Asura_AI_NavVertex* pxWP2 = Asura_AI_Navigation_System::GetWaypoint( s_ausPostRouteBuffer[ us + 1 ] );

			if (pxWP1 && pxWP2)
			{
				Asura_Vector_3 xLocalUp1 = pxWP1->GetNormal();
				xLocalUp1 *= 0.125f;
				Asura_Vector_3 xLocalUp2 = pxWP2->GetNormal();
				xLocalUp2 *= 0.125f;

				Asura_DebugLineRender::AddLine( pxWP1->m_xPosition + xLocalUp1, pxWP2->m_xPosition + xLocalUp2, 0xff00ff00 );
			}
        }
    }
}

int AVP_Navigation::GetDebugNumRoutesAllocated()
{
    return s_iDbgTotalRoutesAllocated;
}
#endif

void AVP_Navigation::InitialiseRouteMempools()
{
    Asura_Assert( !s_pxShortRouteMempool, "AVP_Navigation::InitialiseRouteMempools - s_pxShortRouteMempool mempool is already allocated" );
    Asura_Assert( !s_pxMedSmallRouteMempool,   "AVP_Navigation::InitialiseRouteMempools - s_pxMedSmallRouteMempool mempool is already allocated" );
	Asura_Assert( !s_pxMedRouteMempool,   "AVP_Navigation::InitialiseRouteMempools - s_pxMedRouteMempool mempool is already allocated" );
	Asura_Assert( !s_pxMedLongRouteMempool,   "AVP_Navigation::InitialiseRouteMempools - s_pxMedLongRouteMempool mempool is already allocated" );
    Asura_Assert( !s_pxLongRouteMempool,  "AVP_Navigation::InitialiseRouteMempools - s_pxLongRouteMempool mempool is already allocated" );

    s_pxShortRouteMempool = new Asura_Mempool( sizeof( u_short ) * 16, 256 );
    s_pxMedSmallRouteMempool   = new Asura_Mempool( sizeof( u_short ) * 32, 128 );
	s_pxMedRouteMempool   = new Asura_Mempool( sizeof( u_short ) * 64, 128 );
	s_pxMedLongRouteMempool   = new Asura_Mempool( sizeof( u_short ) * 128, 64 );
    s_pxLongRouteMempool  = new Asura_Mempool( sizeof( u_short ) * 256, 64 );
}

void AVP_Navigation::ShutdownRouteMempools()
{
    if ( s_pxShortRouteMempool )
    {
        Asura_Assert( s_pxShortRouteMempool->GetNumAllocated() == 0, "AVP_Navigation::ShutdownBehaviourMempool - s_pxShortRouteMempool not empty" );
        delete s_pxShortRouteMempool;
        s_pxShortRouteMempool = NULL;
    }
    if ( s_pxMedSmallRouteMempool )
    {
        Asura_Assert( s_pxMedSmallRouteMempool->GetNumAllocated() == 0, "AVP_Navigation::ShutdownBehaviourMempool - s_pxMedRouteMempool not empty" );
        delete s_pxMedSmallRouteMempool;
        s_pxMedSmallRouteMempool = NULL;
    }
	if ( s_pxMedRouteMempool )
	{
		Asura_Assert( s_pxMedRouteMempool->GetNumAllocated() == 0, "AVP_Navigation::ShutdownBehaviourMempool - s_pxMedRouteMempool not empty" );
		delete s_pxMedRouteMempool;
		s_pxMedRouteMempool = NULL;
	}
	if ( s_pxMedLongRouteMempool )
	{
		Asura_Assert( s_pxMedLongRouteMempool->GetNumAllocated() == 0, "AVP_Navigation::ShutdownBehaviourMempool - s_pxMedLongRouteMempool not empty" );
		delete s_pxMedLongRouteMempool;
		s_pxMedLongRouteMempool = NULL;
	}
    if ( s_pxLongRouteMempool )
    {
        Asura_Assert( s_pxLongRouteMempool->GetNumAllocated() == 0, "AVP_Navigation::ShutdownBehaviourMempool - s_pxLongRouteMempool not empty" );
        delete s_pxLongRouteMempool;
        s_pxLongRouteMempool = NULL;
    }
}

u_short* AVP_Navigation::CreateRouteBuffer( const u_int uRouteLength )
{
    u_short* pusBuffer = NULL;

    if ( !pusBuffer && uRouteLength <= 16 )
    {
        pusBuffer = static_cast< u_short* >( s_pxShortRouteMempool->Allocate() );
    }
    if ( !pusBuffer && uRouteLength <= 32 )
    {
        pusBuffer = static_cast< u_short* >( s_pxMedSmallRouteMempool->Allocate() );
    }
	if ( !pusBuffer && uRouteLength <= 64 )
	{
		pusBuffer = static_cast< u_short* >( s_pxMedRouteMempool->Allocate() );
	}
	if ( !pusBuffer && uRouteLength <= 128 )
	{
		pusBuffer = static_cast< u_short* >( s_pxMedLongRouteMempool->Allocate() );
	}
	if ( !pusBuffer && uRouteLength <= 256 )
    {
        pusBuffer = static_cast< u_short* >( s_pxLongRouteMempool->Allocate() );
	}
	Asura_Assert( uRouteLength <= 256 , "AVP_Navigation::CreateRouteBuffer - route length is too long" );
	Asura_Assert( pusBuffer, "AVP_Navigation::CreateRouteBuffer - no more route space available" );
#ifndef ASURA_NO_DEBUG_OUTPUT
        if ( pusBuffer )
        {
            ++s_iDbgNumLongRoutesAllocated;
            ++s_iDbgTotalRoutesAllocated;
        }
#endif
    return pusBuffer;
}

void AVP_Navigation::DestroyRouteBuffer( u_short*& pusRouteBuffer )
{
    if ( s_pxShortRouteMempool->AllocatedThisMemory( pusRouteBuffer ) )
    {
        s_pxShortRouteMempool->Deallocate( pusRouteBuffer );
        pusRouteBuffer = NULL;

#ifndef ASURA_NO_DEBUG_OUTPUT
        --s_iDbgNumShortRoutesAllocated;
#endif
    }
    else if ( s_pxMedSmallRouteMempool->AllocatedThisMemory( pusRouteBuffer ) )
    {
        s_pxMedSmallRouteMempool->Deallocate( pusRouteBuffer );
        pusRouteBuffer = NULL;

#ifndef ASURA_NO_DEBUG_OUTPUT
        --s_iDbgNumMedRoutesAllocated;
#endif
    }
	else if ( s_pxMedRouteMempool->AllocatedThisMemory( pusRouteBuffer ) )
	{
		s_pxMedRouteMempool->Deallocate( pusRouteBuffer );
		pusRouteBuffer = NULL;

#ifndef ASURA_NO_DEBUG_OUTPUT
		--s_iDbgNumMedRoutesAllocated;
#endif
	}
	else if ( s_pxMedLongRouteMempool->AllocatedThisMemory( pusRouteBuffer ) )
	{
		s_pxMedLongRouteMempool->Deallocate( pusRouteBuffer );
		pusRouteBuffer = NULL;

#ifndef ASURA_NO_DEBUG_OUTPUT
		--s_iDbgNumMedRoutesAllocated;
#endif
	}
    else if ( s_pxLongRouteMempool->AllocatedThisMemory( pusRouteBuffer ) )
    {
        s_pxLongRouteMempool->Deallocate( pusRouteBuffer );
        pusRouteBuffer = NULL;

#ifndef ASURA_NO_DEBUG_OUTPUT
        --s_iDbgNumLongRoutesAllocated;
#endif
    }
    else
    {
        Asura_Assert( false, "AVP_Navigation::DestroyRouteBuffer - could not destroy route buffer" );
    }

#ifndef ASURA_NO_DEBUG_OUTPUT
    --s_iDbgTotalRoutesAllocated;
#endif
}

int AVP_Navigation::GetWorkingRouteBufferLength()
{
    return LONGEST_ROUTE_ALLOWED;
}

u_short* AVP_Navigation::GetWorkingRouteBuffer()
{
    return s_ausWorkingRouteBuffer;
}

//////////////////////////////////////////////////////////////////////////////
//
// Asura_Job_RouteFind
//
//////////////////////////////////////////////////////////////////////////////

AVP_RouteFind::Request AVP_RouteFind::s_axRequests[ MAX_SIMULTANEOUS_JOBS ];
int					   AVP_RouteFind::s_iNumUsedRequests   = 0;
int					   AVP_RouteFind::s_iNumActiveRequests = 0;
Asura_Mutex			   AVP_RouteFind::s_xRequestMutex;
Asura_SyncEvent		   AVP_RouteFind::s_xStopThread( "AVP Axon Route Find Stop", false );

Asura_Handle AVP_RouteFind::RequestRouteFind( AVP_Navigation::NavData& xNavData )
{
	//if ( s_iNumUsedRequests < MAX_SIMULTANEOUS_JOBS )
	//{
	//	// Find the first free request slot and set it up.
	//	for ( int i = 0; i < MAX_SIMULTANEOUS_JOBS; ++i )
	//	{
	//		if ( s_axRequests[ i ].m_eState == Request::FREE )
	//		{
	//			s_xRequestMutex.Lock();
 //               s_axRequests[ i ].m_eState   = Request::PENDING;
	//			s_axRequests[ i ].m_xNavData = xNavData;
	//			++s_iNumUsedRequests;
	//			++s_iNumActiveRequests;
	//			s_xRequestMutex.Unlock();

	//			return i;
	//		}
	//	}

	//	return ASURA_INVALID_HANDLE;
	//}
	//else
	{
		Asura_Assert( false, "Asura_Job_RouteFind::RequestRouteFind - too many pending route requests" );
		return ASURA_INVALID_HANDLE;
	}
}

void AVP_RouteFind::CancelRouteFind( Asura_Handle& iHandle )
{
	if ( iHandle == ASURA_INVALID_HANDLE )
	{
		Asura_Assert( false, "AVP_RouteFind::CancelRouteFind - invalid handle" );
		return;
	}

	// If this request has already been finished then just free it up, otherwise mark it as cancelled.
	s_xRequestMutex.Lock();
	if ( s_axRequests[ iHandle ].m_eState == Request::FINISHED )
	{
		s_axRequests[ iHandle ].m_eState = Request::FREE;
		--s_iNumUsedRequests;
	}
	else
	{
		s_axRequests[ iHandle ].m_eState = Request::CANCELLING;
	}
	s_xRequestMutex.Unlock();

	iHandle = ASURA_INVALID_HANDLE;
}

bool AVP_RouteFind::RouteFindFinished( const Asura_Handle& iHandle )
{
	if ( iHandle == ASURA_INVALID_HANDLE )
	{
		Asura_Assert( false, "AVP_RouteFind::RouteFindFinished - invalid handle" );
		return true;
	}

	return ( s_axRequests[ iHandle ].m_eState == Request::FINISHED );
}

bool AVP_RouteFind::GetRouteFindResults( Asura_Handle& iHandle, AVP_Navigation::NavData& xNavData )
{
	if ( iHandle == ASURA_INVALID_HANDLE )
	{
		Asura_Assert( false, "AVP_RouteFind::GetRouteFindResults - invalid handle" );
		xNavData.m_usRouteLength = 0;
		return true;
	}

	if ( s_axRequests[ iHandle ].m_eState == Request::FINISHED )
	{
		s_xRequestMutex.Lock();
		xNavData.m_usRouteLength = s_axRequests[ iHandle ].m_xNavData.m_usRouteLength;
		xNavData.m_pusRoute		 = s_axRequests[ iHandle ].m_xNavData.m_pusRoute;

		s_axRequests[ iHandle ].m_eState = Request::FREE;
		--s_iNumUsedRequests;
		s_xRequestMutex.Unlock();

		iHandle = ASURA_INVALID_HANDLE;
		return true;
	}
	else
	{
		Asura_Assert( s_axRequests[ iHandle ].m_eState != Request::FREE, "Asura_Job_RouteFind::GetRouteFindResults - this request does not exist" );
		return false;
	}
}

void AVP_RouteFind::Initialise()
{
	Asura_Thread_System::CreateThread( "AVP Axon Route Find Thread", ThreadUpdateFunction, NULL );
}

void AVP_RouteFind::Reset()
{
	CancelAllRouteFinds();
}

void AVP_RouteFind::ThreadUpdateFunction( const void* pUserData )
{
	do
	{
		Asura_Thread_System::DelayCurrentThread( 0.033333333333333333333333333333333f ); //30 times per second

		// Work through the requests until we have none left active.
        int iRequestIndex = 0;
		while ( s_iNumActiveRequests > 0 )
		{
			// If we've been cancelled then set the request as free.
			if ( s_axRequests[ iRequestIndex ].m_eState == Request::CANCELLING )
			{
				s_xRequestMutex.Lock();
				s_axRequests[ iRequestIndex ].m_eState = Request::FREE;
				--s_iNumUsedRequests;
				--s_iNumActiveRequests;
				s_xRequestMutex.Unlock();
			}
			else if ( s_axRequests[ iRequestIndex ].m_eState == Request::PENDING )
			{
				FindRoute( s_axRequests[ iRequestIndex ].m_xNavData );

				// This request may have been cancelled whilst we were working.
				// If that is the case then just free up the request, otherwise mark it as finished.
				s_xRequestMutex.Lock();
				if ( s_axRequests[ iRequestIndex ].m_eState == Request::CANCELLING )
				{
					s_axRequests[ iRequestIndex ].m_eState = Request::FREE;
					--s_iNumUsedRequests;
					--s_iNumActiveRequests;
				}
				else
				{
		            s_axRequests[ iRequestIndex ].m_eState = Request::FINISHED;
					--s_iNumActiveRequests;
				}
				s_xRequestMutex.Unlock();
			}

			iRequestIndex = (iRequestIndex + 1) % MAX_SIMULTANEOUS_JOBS;
		}
	} while ( !s_xStopThread.IsSignalled() );
	s_xStopThread.ClearSignal();
}

void AVP_RouteFind::Shutdown()
{
	CancelAllRouteFinds();
	s_xStopThread.SetSignal();
}

void AVP_RouteFind::FindRoute( AVP_Navigation::NavData& xNavData )
{
//#ifdef ASURA_TOOLS
//    if ( !xNavData.Verify() ) { return; }
//#endif
//
//    // Make doubly sure that the route is reset.
//    xNavData.ResetRoute();
//
//    const Asura_AI_NavVertex* pxCurrentWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_usCurrentWPID );
//    if ( pxCurrentWP && pxCurrentWP->Contains( xNavData.m_xDestinationPosition ) )
//    {
//        // No need to create a route if the destination is within the current waypoint.
//		xNavData.m_pusRoute = AVP_Navigation::CreateRouteBuffer( 1 );
//        if ( xNavData.m_pusRoute )
//        {
//            xNavData.m_usRouteLength = 1;
//            xNavData.m_pusRoute[ 0 ] = xNavData.m_usDestinationWPID;
//        }
//    }
//    else
//    {
//        //## In here we will eventually check the route type and call the different A* functions based on this.
//        //switch ( xNavData.m_eRouteType ) {}
//
//        int iRouteLength = 0;
//        Asura_AI_Navigation_System::FindRouteCustom( AVP_Navigation::GetWorkingRouteBuffer(), iRouteLength, AVP_Navigation::GetWorkingRouteBufferLength(),
//            xNavData.m_usCurrentWPID, xNavData.m_usDestinationWPID, xNavData.m_uPathfinderFlags,
//            Asura_AI_Navigation_System::DefaultGoalCostEstimateFunction,
//            AVP_Navigation::TraversalCostFunction,
//            Asura_AI_Navigation_System::DefaultGoalAchievedFunction,
//            AVP_Navigation::LinkFlagsTestFunction );
//        if ( iRouteLength )
//        {
//			// Bit of trickery to do the route refining before we allocate any route memory.
//			xNavData.m_pusRoute      = AVP_Navigation::GetWorkingRouteBuffer();
//			xNavData.m_usRouteLength = static_cast< u_short >( iRouteLength );
//			const AVP_Navigation::TestParams xTestParams( xNavData.m_uPathfinderFlags, xNavData.m_ucCurrentZone );
//            AVP_Navigation::RefineRoute( xNavData, xTestParams );
//			xNavData.m_pusRoute = AVP_Navigation::CreateRouteBuffer( xNavData.m_usRouteLength );
//			if ( xNavData.m_pusRoute )
//			{
//				Asura_Mem::Copy( xNavData.m_pusRoute, AVP_Navigation::GetWorkingRouteBuffer(), sizeof( u_short ) * xNavData.m_usRouteLength );
//			}
//        }
//    }
}

void AVP_RouteFind::CancelAllRouteFinds()
{
	s_xRequestMutex.Lock();
	for ( int i = 0; i < MAX_SIMULTANEOUS_JOBS; ++i )
	{
		if ( s_axRequests[ i ].m_eState == Request::FINISHED )
		{
			s_axRequests[ i ].m_eState = Request::FREE;
			--s_iNumUsedRequests;
		}
		else
		{
			s_axRequests[ i ].m_eState = Request::CANCELLING;
		}
	}
	s_xRequestMutex.Unlock();
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
