//////////////////////////////////////////////////////////////////////////////
//
// AVP_Navigation_RouteRefine.cpp
//
// Route refining implementation for AVP_Navigation.
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Core/AVP_Navigation.h"

//////////////////////////////////////////////////////////////////////////////
//
// AVP_Navigation
//
//////////////////////////////////////////////////////////////////////////////

static AVP_Navigation::TestParams s_xTestParams;
static Asura_Vector_3			  s_xLineStart            = Asura_Maths::ZeroVector3;
static Asura_Vector_3			  s_xLineDirection		  = Asura_Maths::ZeroVector3;

enum { MAX_HITS_ALLOWED = 32 };
static struct ParametricHit
{
	float m_fMin;
	float m_fMax;
	bool  m_bUsed;
}     s_axParametricHits[ MAX_HITS_ALLOWED ];
static u_int s_uNumParametricHits = 0;

static bool WaypointIsValidForSmoothing( const u_short usWaypointID )
{
	// Don't shortcut through waypoints that may be dangerous.
	if ( Asura_AI_Navigation_System::GetWaypointWeight( usWaypointID ) > 0 )
	{
		return false;
	}

	// Don't shortcut through waypoints that have blocked links.
	// This is because we can't be sure that we won't be traversing along what was (roughly) the blocked link.
	const Asura_AI_NavVertex* pxWP = Asura_AI_Navigation_System::GetWaypoint( usWaypointID );
	for ( u_int u = 0; u < pxWP->GetNumLinks(); ++u )
	{
		if ( pxWP->GetLink( u ).IsBlocked() )
		{
			return false;
		}
	}

	return true;
}

static bool LinkIsValidForSmoothing( const u_short usFromWPIndex, const u_short usToWPIndex )
{
	const Asura_AI_NavVertex* pxFromWP = Asura_AI_Navigation_System::GetWaypoint( usFromWPIndex );
	if ( !pxFromWP ) { return false; }

	// Find the correct link.
	const Asura_AI_NavConnection* pxLink = NULL;
	for ( u_char ucLinkIndex = 0; ucLinkIndex < pxFromWP->m_ucNumConnections; ++ucLinkIndex )
	{
		if ( pxFromWP->m_pxConnections[ ucLinkIndex ].m_usIndex == usToWPIndex )
		{
			pxLink = &pxFromWP->m_pxConnections[ ucLinkIndex ];
			break;
		}
	}
	if ( !pxLink ) { return false; }

	// If this is a special link then it's not valid for smoothing through.
	if ( AVP_Navigation::IsTransitionLink( pxLink->m_usFlags ) ||
		AVP_Navigation::IsGS2Link( pxLink->m_usFlags ) ||
		AVP_Navigation::IsEnvHookLink( pxLink->m_usFlags ) )
	{
		return false;
	}

	return true;
}

//-------------------------------------------------------------------------
// Function : RayIntersectsCylinder
//-------------------------------------------------------------------------
// Adapted from RayIntersectsCylinder in Asura_Maths
// However that only check for the first intersection this has been 
// modified as we need both places where the line intersects the cylinder
//-------------------------------------------------------------------------
static bool RayIntersectsCylinder(const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, const Asura_Vector_3& xCylinderStart, const Asura_Vector_3& xCylinderNormal, const float fCylinderHeight, const float fCylinderRadius, float* fMin, float* fMax)
{
	// First test against each cap.
	const Asura_Vector_3	xRayEnd								= xRayOrigin + xRayDirection;
	const float				fRadiusSq							= fCylinderRadius * fCylinderRadius;
	const Asura_Vector_3	xCylinderEnd						= xCylinderStart + (xCylinderNormal * fCylinderHeight);
	const float				fRayDotCylinderAxis					= xRayDirection * xCylinderNormal;
	const Asura_Vector_3	xRayOriginToCylinderStart			= xCylinderStart - xRayOrigin;
	const Asura_Vector_3	xRayEndToCylinderStart				= xCylinderStart - xRayEnd;
	const float				fDistanceROToCylinderStartAlongAxis	= xRayOriginToCylinderStart * xCylinderNormal;
	const float				fDistanceREToCylinderStartAlongAxis	= xRayEndToCylinderStart * xCylinderNormal;
	Asura_Vector_3			xIntersection[2]					= {Asura_Maths::ZeroVector3, Asura_Maths::ZeroVector3};
	u_int					uNumIntersections					= 0;


	//----------------------------
	// Start Cap
	//----------------------------
	if ((((fDistanceROToCylinderStartAlongAxis < 0.0f) && (fRayDotCylinderAxis >= 0.0f)) ||
		((fDistanceROToCylinderStartAlongAxis > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
		(fRayDotCylinderAxis != 0.0f))
	{
		// Ray points towards start cap. Test for intersection point.
		const float				fIntersectionDistance	= fDistanceROToCylinderStartAlongAxis / fRayDotCylinderAxis;
		const Asura_Vector_3	xIntersectionPoint		= xRayOrigin + xRayDirection * fIntersectionDistance;
		const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderStart).MagnitudeSquared();

		if (fDistanceOnCapSquared <= fRadiusSq)
		{
			Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
			xIntersection[uNumIntersections] = xIntersectionPoint;
			++uNumIntersections;
		}
	}
	
	if ((((fDistanceREToCylinderStartAlongAxis < 0.0f) && (fRayDotCylinderAxis >= 0.0f)) ||
		((fDistanceREToCylinderStartAlongAxis > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
		(fRayDotCylinderAxis != 0.0f))
	{
		// Ray points towards start cap. Test for intersection point.
		const float				fIntersectionDistance	= fDistanceREToCylinderStartAlongAxis / -fRayDotCylinderAxis;
		const Asura_Vector_3	xIntersectionPoint		= xRayEnd - xRayDirection * fIntersectionDistance;
		const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderStart).MagnitudeSquared();

		if (fDistanceOnCapSquared <= fRadiusSq)
		{
			Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
			xIntersection[uNumIntersections] = xIntersectionPoint;
			++uNumIntersections;
		}
	}

	//----------------------------
	// End Cap
	//----------------------------
	if (uNumIntersections < 2)
	{
		if (((((fDistanceROToCylinderStartAlongAxis + fCylinderHeight) > 0.0f) && (fRayDotCylinderAxis <= 0.0f)) ||
			(((fDistanceROToCylinderStartAlongAxis + fCylinderHeight) > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
			(fRayDotCylinderAxis != 0.0f))
		{
			// Ray points towards end cap. Test for intersection point.
			const float				fIntersectionDistance	= (fDistanceROToCylinderStartAlongAxis + fCylinderHeight) / fRayDotCylinderAxis;
			const Asura_Vector_3	xIntersectionPoint		= xRayOrigin + xRayDirection * fIntersectionDistance;
			const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderEnd).MagnitudeSquared();

			if (fDistanceOnCapSquared <= fRadiusSq)
			{
				Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
				xIntersection[uNumIntersections] = xIntersectionPoint;
				++uNumIntersections;
			}
		}
	}
	
	if (uNumIntersections < 2)
	{
		if (((((fDistanceREToCylinderStartAlongAxis + fCylinderHeight) < 0.0f) && (fRayDotCylinderAxis >= 0.0f)) ||
			(((fDistanceREToCylinderStartAlongAxis + fCylinderHeight) > 0.0f) && (fRayDotCylinderAxis <= 0.0f))) && 
			(fRayDotCylinderAxis != 0.0f))
		{
			// Ray points towards end cap. Test for intersection point.
			const float				fIntersectionDistance	= (fDistanceREToCylinderStartAlongAxis + fCylinderHeight) / -fRayDotCylinderAxis;
			const Asura_Vector_3	xIntersectionPoint		= xRayEnd - xRayDirection * fIntersectionDistance;
			const float				fDistanceOnCapSquared	= (xIntersectionPoint - xCylinderEnd).MagnitudeSquared();

			if (fDistanceOnCapSquared <= fRadiusSq)
			{
				Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
				xIntersection[uNumIntersections] = xIntersectionPoint;
				++uNumIntersections;
			}
		}
	}

	// Only do the following if we haven't hit both end caps
	if (uNumIntersections < 2)
	{
		// Test the sides of the cylinder
		Asura_Vector_3	xPerpendicularDirection			= xRayDirection ^ xCylinderNormal;
		const float		fPerpendicularLength			= xPerpendicularDirection.Normalise();
		const float		fShortestDistanceToAxis			= xRayOriginToCylinderStart * xPerpendicularDirection;
		const float		fShortestDistanceToAxisSquared	= fShortestDistanceToAxis * fShortestDistanceToAxis;

		if (fShortestDistanceToAxisSquared <= fRadiusSq)
		{
			// Calculate distance to closest point on ray to axis.
			const Asura_Vector_3	xD1							= xRayOriginToCylinderStart ^ xCylinderNormal;
			const float				fDistanceToClosestPoint		= (xD1 * xPerpendicularDirection) / fPerpendicularLength;
			// Calculate distance along ray from closest point to cylinder surface.
			const Asura_Vector_3	xD2							= xPerpendicularDirection ^ xCylinderNormal;
			const float				fS							= Asura_Maths::Abs(Asura_Maths::Sqrt(fRadiusSq - fShortestDistanceToAxisSquared) / (xRayDirection * xD2));
			const float				fFirstIntersectionDistance	= fDistanceToClosestPoint - fS;
			const Asura_Vector_3	xFirstIntersectionPoint		= xRayOrigin + xRayDirection * fFirstIntersectionDistance;

			if (fFirstIntersectionDistance >= 0.0f)
			{
				Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
				xIntersection[uNumIntersections] = xFirstIntersectionPoint;
				++uNumIntersections;
			}

			// Do an extra number intersections check here to avoid the three intersections crash. 
			if (uNumIntersections < 2)
			{
				const float	fSecondIntersectionDistance = fDistanceToClosestPoint + fS;

				if (fFirstIntersectionDistance < fSecondIntersectionDistance)
				{
					const Asura_Vector_3	xSecondIntersectionPoint	= xRayOrigin + xRayDirection * fSecondIntersectionDistance;

					if (fSecondIntersectionDistance >= 0.f)
					{
						Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
						xIntersection[uNumIntersections] = xSecondIntersectionPoint;
						++uNumIntersections;
					}
				}
				else
				{
					Asura_Assert(uNumIntersections < 2, "RayIntersectsCylinder - More than two intersections!");
					xIntersection[uNumIntersections] = xFirstIntersectionPoint;
					++uNumIntersections;
				}
			}
		}
	}

	// Calculate our parametric values
	if (uNumIntersections == 1)
	{
		float fDistanceRayOriginToIntersection	= (xIntersection[0] - xRayOrigin).Magnitude();
		float fLineDistance						= xRayDirection.Magnitude();

		*fMin = 0.0f;
		*fMax = fDistanceRayOriginToIntersection / fLineDistance;

		return true;
	}
	// We have two intersections, so we can find out what portion of the ray our cylinder covers
	else if (uNumIntersections == 2)
	{
		float fDistanceToFirstIntersection	= (xIntersection[0] - xRayOrigin).Magnitude();
		float fDistanceToSecondIntersection	= (xIntersection[1] - xRayOrigin).Magnitude();
		float fLineDistance					= xRayDirection.Magnitude();

		// Smallest value to be fMin, and largest value to be fMax
		if (fDistanceToFirstIntersection > fDistanceToSecondIntersection)
		{
			*fMin = fDistanceToFirstIntersection / fLineDistance;
			*fMax = fDistanceToSecondIntersection / fLineDistance;
		}
		else
		{
			*fMin = fDistanceToFirstIntersection / fLineDistance;
			*fMax = fDistanceToSecondIntersection / fLineDistance;
		}
		return true;
	}

	// We haven't intersected with anything so set fMin and fMax to zero and retrun
	*fMin = 0.0f;
	*fMax = 0.0f;
	return false;
}

static bool WaypointSpaceTreeCallback( const u_int uWaypointID )
{
	// Check this waypoint matches our test params.
	if ( !AVP_Navigation::IsValidWaypointID( static_cast< u_short >( uWaypointID ), s_xTestParams ) )
	{
		return true;
	}

	// Make a BB up from this waypoint. Make it symmetrical for simplicity's sake.
	const Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( uWaypointID );

	// Work out the parametric portion of the line this box covers.
	if ( s_uNumParametricHits < MAX_HITS_ALLOWED )
	{
		const Asura_Vector_3&	xCylinderNormal		= pxWaypoint->GetNormal();
		const Asura_Vector_3	xCylinderPosition	= pxWaypoint->m_xPosition - (xCylinderNormal);
		const float				fCylinderRadius		= pxWaypoint->GetRadius();
		const float				fCylinderHeight		= fCylinderRadius + 1.0f;
		float fMin, fMax;

		if (RayIntersectsCylinder(s_xLineStart, s_xLineDirection, xCylinderPosition, xCylinderNormal, fCylinderHeight, fCylinderRadius, &fMin, &fMax))
		{
			// This is a hit. See if we can merge it with any existing hits. If we can't then add it on the end.
			bool bMerged = false;
			for ( u_int u = 0; u < s_uNumParametricHits && !bMerged; ++u )
			{
				if ( s_axParametricHits[ u ].m_fMin <= fMin && fMax <= s_axParametricHits[ u ].m_fMax )
				{
					// This hit is completely enclosed in an already existing hit, so it can be ignored.
					bMerged = true;
				}
				else if ( fMin <= s_axParametricHits[ u ].m_fMin && s_axParametricHits[ u ].m_fMin <= fMax )
				{
					// We intersect the lower bound. We can expand this existing hit.
					fMin = Asura_Maths::Min( fMin, s_axParametricHits[ u ].m_fMin );
					fMax = Asura_Maths::Max( fMax, s_axParametricHits[ u ].m_fMax );
					s_axParametricHits[ u ].m_fMin = fMin;
					s_axParametricHits[ u ].m_fMax = fMax;
					bMerged = true;
				}
				else if ( fMin <= s_axParametricHits[ u ].m_fMax && s_axParametricHits[ u ].m_fMax <= fMax )
				{
					// We intersect the upper bound. We can expand this existing hit.
					fMin = Asura_Maths::Min( fMin, s_axParametricHits[ u ].m_fMin );
					fMax = Asura_Maths::Max( fMax, s_axParametricHits[ u ].m_fMax );
					s_axParametricHits[ u ].m_fMin = fMin;
					s_axParametricHits[ u ].m_fMax = fMax;
					bMerged = true;
				}
			}

			if ( !bMerged )
			{
				s_axParametricHits[ s_uNumParametricHits ].m_fMin = fMin;
				s_axParametricHits[ s_uNumParametricHits ].m_fMax = fMax;
				++s_uNumParametricHits;
			}
		}

		// If we now cover the whole of the line then we can terminate the waypoint graph test.
		return ( 0.0f < fMin || fMax < 1.0f );
	}
	else
	{
		Asura_Assert( false, "WaypointSpaceTreeCallback - run out of hit space" );
		return false;
	}
}

static bool LineSegmentStaysInValidWaypointSpace( const Asura_Vector_3& xStart, const Asura_Vector_3& xFinish, const AVP_Navigation::TestParams& xParams )
{
	// Test the waypoint BB tree with the line.
	// Each waypoint returned covers another section of the line.
	// If the line is entirely covered when the test is finished then this waypoint stays in waypoint space.
	Asura_Vector_3 xLineDirection = xFinish - xStart;

	s_xTestParams	          = xParams;
	s_xLineStart              = xStart;
	s_xLineDirection		  = xLineDirection;
	s_uNumParametricHits      = 0;

	// We expand the distance to test the tree with to ensure we hit boxes that we be just past our finish position but still cover the line.
	float fLineLength = xLineDirection.Normalise() + 15.0f;
	Asura_AI_Navigation_System::SearchWaypoints( xStart, xLineDirection, &fLineLength, WaypointSpaceTreeCallback );

	if ( s_uNumParametricHits == 0 )
	{
		return false;
	}

	// Iterate over the hits, merging them until they are all used.
	u_int u;
	for ( u = 0; u < s_uNumParametricHits; ++u )
	{
		s_axParametricHits[ u ].m_bUsed = false;
	}
	float fMin = s_axParametricHits[ 0 ].m_fMin, fMax = s_axParametricHits[ 0 ].m_fMax;
	s_axParametricHits[ 0 ].m_bUsed = true;
	u_int uPrevUsedCount = 1;
	u_int uUsedCount     = 1;
	do 
	{
		uPrevUsedCount = uUsedCount;

		for ( u = 0; u < s_uNumParametricHits; ++u )
		{
			if ( !s_axParametricHits[ u ].m_bUsed )
			{
				if ( s_axParametricHits[ u ].m_fMin < fMin && fMin < s_axParametricHits[ u ].m_fMax )
				{
					fMin = s_axParametricHits[ u ].m_fMin;
					s_axParametricHits[ u ].m_bUsed = true;
				}
				if ( s_axParametricHits[ u ].m_fMin < fMax && fMax < s_axParametricHits[ u ].m_fMax )
				{
					fMax = s_axParametricHits[ u ].m_fMax;
					s_axParametricHits[ u ].m_bUsed = true;
				}
				if ( s_axParametricHits[ u ].m_bUsed )
				{
					++uUsedCount;
				}
			}
		}

		if ( fMin <= 0.0f && 1.0f <= fMax )
		{
			break;
		}

		// We stop if we've either used all of the hits up or we used none up this iteration (meaning we can't merge any further).
	} while ( ( uUsedCount < s_uNumParametricHits ) && ( uUsedCount != uPrevUsedCount ) );

	return ( fMin <= 0.0f && 1.0f <= fMax );
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void AVP_Navigation::RefineRoute( NavData& xNavData, const TestParams& xParams /*= s_xTPStandardParams*/ )
{
	// Go through the route removing redundant waypoints.
	// The base index is the index from which we are trying to smooth.
	// The target index is the index we've got to with the smoothing.
	u_short usBaseIndex = xNavData.m_usRouteSegmentIndex;
	while ( usBaseIndex < xNavData.m_usRouteLength - 2 )
	{
		const Asura_AI_NavVertex* pxBaseWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usBaseIndex ] );
		const Asura_Vector_3& xBaseNormal = pxBaseWP->GetNormal();

		// Iterate the test index away from the base index until we cannot shortcut any more.
		u_short usTargetIndex = usBaseIndex;
		u_short usTestIndex   = usBaseIndex + 1;
		while ( usTestIndex < xNavData.m_usRouteLength )
		{
			if ( !WaypointIsValidForSmoothing( xNavData.m_pusRoute[ usTestIndex ] ) )
			{
				break;
			}

			if ( !LinkIsValidForSmoothing( xNavData.m_pusRoute[ usTestIndex - 1 ], xNavData.m_pusRoute[ usTestIndex ] ) )
			{
				break;
			}

			const Asura_AI_NavVertex* pxTestWP = Asura_AI_Navigation_System::GetWaypoint( xNavData.m_pusRoute[ usTestIndex ] );
			const Asura_Vector_3& xTestNormal = pxTestWP->GetNormal();

			// Work out a sideways vector so we can incorporate the character width.
			Asura_Vector_3 xAverageNormalVector = ( xBaseNormal + xTestNormal ) * 0.5f;
			const Asura_Vector_3 xLineVector = pxTestWP->m_xPosition - pxBaseWP->m_xPosition;
			Asura_Vector_3 xSideways = xAverageNormalVector ^ xLineVector;
			xSideways.Normalise();

			// We test two lines, one either side, to ensure the refined route is wide enough for the character.
			Asura_Vector_3 xOffsetVector = xSideways * ( xNavData.m_fCharacterWidth * 0.5f );
			if ( !LineSegmentStaysInValidWaypointSpace( pxBaseWP->m_xPosition + xOffsetVector, pxTestWP->m_xPosition + xOffsetVector, xParams ) )
			{
				break;
			}
			if ( !LineSegmentStaysInValidWaypointSpace( pxBaseWP->m_xPosition - xOffsetVector, pxTestWP->m_xPosition - xOffsetVector, xParams ) )
			{
				break;
			}

			// We test another line to ensure the refined route is high enough for the character.
			xAverageNormalVector.Normalise();
			xOffsetVector = xAverageNormalVector * xNavData.m_fCharacterHeight;
			if ( !LineSegmentStaysInValidWaypointSpace( pxBaseWP->m_xPosition + xOffsetVector, pxTestWP->m_xPosition + xOffsetVector, xParams ) )
			{
				break;
			}

			usTargetIndex = usTestIndex++;
		}

		// Remove the intervening path points and adjust the route length.
		if ( usTargetIndex > usBaseIndex )
		{
			u_short usCopySource = usTargetIndex;
			u_short usCopyDest   = usBaseIndex + 1;
			while ( usCopySource < xNavData.m_usRouteLength )
			{
				xNavData.m_pusRoute[ usCopyDest++ ] = xNavData.m_pusRoute[ usCopySource++ ];
			}

			xNavData.m_usRouteLength -= ( usCopySource - usCopyDest );
		}

		// Advance the start position.
		++usBaseIndex;
	}
}

//bool AVP_Navigation::TestRayBBTest( const Asura_Vector_3& xStartPosition, const Asura_Vector_3& xEndPosition, const Asura_Bounding_Box& xBB,
//								    float& fT0, float& fT1 )
//{
//	return LineSegmentIntersectsBB( xStartPosition, xEndPosition, xBB, fT0, fT1 );
//}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
