
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Actor_AIAnimStateMachine.cpp
// Author : Will Plano
// Date   : 29/02/2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/AVP_Actor_AIAnimStateMachine.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Core/Asura_Timers.h"
#include "Entity/PhysicalObject/Moves/Asura_Movement_System.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Message.h"
#include "Rendering/Asura_Font_System.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_AIAnimStateConnection.h"
#include "Core/AVP_Core.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Moves/AVP_Moves.h"

DEBUGVAR float dbg_fNewStateTweenTime = 0.05f;
DEBUGVAR float dbg_fStartNextStateAnimTime = 0.3f;
DEBUGVAR float dbg_fMissingTransitionAnimTweenTime = 0.5f;

#ifndef ASURA_NO_DEBUG_OUTPUT
DVSVAR bool dvs_bRenderDebugInfo = false;
#endif

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
int AVP_Actor_AIAnimStateMachine::s_iConsoleVarsCount	= 0;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

static const char* g_aszStateString[ AVP_Actor_AIAnimStateMachine::ACTOR_STATE_END - AVP_Actor_AIAnimStateMachine::ACTOR_STATE_BEGIN ] = 
{
	"Dead",							// ACTOR_STATE_DEAD
	"InMove",						// ACTOR_STATE_MOVE
	"InGamescene",					// ACTOR_STATE_IN_GAMESCENE
	"Idle"							// ACTOR_STATE_IDLE
};

static const bool g_abLoopedState[ AVP_Actor_AIAnimStateMachine::ACTOR_STATE_END - AVP_Actor_AIAnimStateMachine::ACTOR_STATE_BEGIN ] = 
{
	true,							// ACTOR_STATE_DEAD
	true,							// ACTOR_STATE_MOVE
	true,							// ACTOR_STATE_IN_GAMESCENE
	true							// ACTOR_STATE_IDLE
};

////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_AIAnimStateMachine::AStarNode AVP_Actor_AIAnimStateMachine::s_axAStarNodeScratchPad[ AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES ];

AVP_Actor_AIAnimStateMachine::AVP_Actor_AIAnimStateMachine( AVP_ServerEntity_Actor& xServerActor )
: m_xServerActor( xServerActor )
, m_uStateAfterMove( ACTOR_STATE_IDLE )
, m_uStateQueueStart( 0 )
, m_uStateQueueEnd( 0 )
, m_uCurrentState( ACTOR_STATE_IDLE )
, m_uFlags( 0 )
, m_fTimeInState( 0.0f )
, m_fCurrentTweenTime( 0.0f )
, m_fOSMoveDirAngle( 0.0f )
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_iConsoleVarsCount == 0)
	{
#ifndef ASURA_NO_DEBUG_OUTPUT
		Asura_CommandConsole::AddVar(L"AVP.States.RenderAIDebugInfo", &dvs_bRenderDebugInfo, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
#endif
	}
	++s_iConsoleVarsCount;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

#ifdef _DEBUG
	for ( u_int uItr = 0; uItr != AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE; ++uItr )
	{
		// Set the queue to a default value.
		m_auStateQueue[ uItr ] = 0xFFFFFFFF;
	}
#endif // _DEBUG
}

AVP_Actor_AIAnimStateMachine::~AVP_Actor_AIAnimStateMachine()
{
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	--s_iConsoleVarsCount;
	if( s_iConsoleVarsCount == 0 )
	{
#ifndef ASURA_NO_DEBUG_OUTPUT
		Asura_CommandConsole::RemoveVar(L"AVP.States.RenderAIDebugInfo" );
#endif
	}
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}

u_int AVP_Actor_AIAnimStateMachine::GetTargetState() const
{
	if ( m_uStateQueueStart == m_uStateQueueEnd )
	{
		return m_uCurrentState;
	}
	else
	{
		return m_auStateQueue[ m_uStateQueueEnd - 1 ];
	}
}

void AVP_Actor_AIAnimStateMachine::Update()
{
	m_fTimeInState += Asura_Timers::GetGameTimeSlice();
	if ( HasStateInQueue() && IsCurrentStateFinished() )
	{
		if ( EndState( m_uCurrentState ) )
		{
			BeginState( m_auStateQueue[ m_uStateQueueStart ] );
			m_uCurrentState	= m_auStateQueue[ m_uStateQueueStart ];
#ifdef _DEBUG
			m_auStateQueue[ m_uStateQueueStart ] = 0xFFFFFFFF;
#endif
			m_fTimeInState	= 0.0f;
			++m_uStateQueueStart;
			Asura_Assert( m_uStateQueueStart < AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE, "Too many states in queue" );
		}
	}
	UpdateState( m_uCurrentState );

#ifndef ASURA_NO_DEBUG_OUTPUT
	if ( dvs_bRenderDebugInfo )
	{
		const Asura_Handle iFontHandle = AVP_Core::GetFontHandle( AVP_Core::HUD_FONT );
		Asura_Font_System::SetFont( iFontHandle );
		RenderDebugInfo();
	}
#endif
}

bool AVP_Actor_AIAnimStateMachine::HandleMessage( const Asura_Message* pxMessage )
{
	bool bOk( true );
	Asura_Assert( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE, "Unhandled message" );
	if ( pxMessage->GetMsgID() == ASURA_MSG_MOVECHAINCOMPLETE )
	{
		// only need to handle this message if we're in the ACTOR_STATE_IN_MOVE state
		if ( m_uCurrentState == ACTOR_STATE_IN_MOVE )
		{
			// if we've not got any states queued up use the parameter passed to RequestInMove
			if ( !HasStateInQueue() )
			{
				ForceState( m_uStateAfterMove );
			}
			else
			{
				Asura_Assert( false, "Shouldn't have a statequeue while in a move" );
			}
			m_bInMove = false;
		}
	}
	return bOk;
}

void AVP_Actor_AIAnimStateMachine::RequestInMove( Asura_Message* pxMoveStartMessage, const u_int uStateAfterMove /*= ACTOR_STATE_IDLE*/ )
{
	Asura_Msg_Move* pxData = static_cast< Asura_Msg_Move* >( pxMoveStartMessage->GetData() );
	if ( pxData != NULL )
	{
		m_bInMove = true;

		// Quick hack - forward the message directly to the client...
		pxData->SetIgnoredByOwner( false );
		AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity( m_xServerActor.GetGuid() );
		if ( pxClient != NULL && pxClient->HandleMessage( pxMoveStartMessage ) )
		{
			// ...before informing all other client's as well...
			pxData->SetIgnoredByOwner( true );
			Asura_Network::Send( pxMoveStartMessage );
			m_uStateAfterMove = uStateAfterMove;
			ForceState( ACTOR_STATE_IN_MOVE );
			return;
		}
	}

	Asura_Network::DestroyMessage( pxMoveStartMessage );
	Asura_Assert( false, "Failed to start move on server" );

	m_bInMove = false;
}

void AVP_Actor_AIAnimStateMachine::RequestInMove( const u_int uStateAfterMove /*= ACTOR_STATE_IDLE*/ )
{
	m_bInMove = true;
	m_uStateAfterMove = uStateAfterMove;
	ForceState( ACTOR_STATE_IN_MOVE );
}


float AVP_Actor_AIAnimStateMachine::GetDistanceRequired( const u_int uStateFrom, const u_int uStateTo )
{
	u_int auRouteQueue[ AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE ];
	u_int uRouteLength = 0;
	if ( !AStarSearch( auRouteQueue, uRouteLength, uStateFrom, uStateTo ) )
	{
		return 0.0f;
	}
	float fDist = 0.0f;
	for ( u_int u = 0; u < uRouteLength; ++u )
	{
		const AVP_Actor_AnimStateData* xStateData = GetStateAnimData( auRouteQueue[ u ] );
		Asura_Assert( xStateData, "Unable to get state data for calculating distance required" );
		if ( xStateData )
		{
			fDist += xStateData->GetMoveDistance();
		}
	}
	return fDist;
}

void AVP_Actor_AIAnimStateMachine::RequestState( const u_int uNewState )
{
	if ( uNewState != GetTargetState() )
	{
#ifdef _DEBUG
		for ( u_int uItr = 0; uItr != AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE; ++uItr )
		{
			// Set the queue to a default value.
			m_auStateQueue[ uItr ] = 0xFFFFFFFF;
		}
#endif
		m_uStateQueueStart = m_uStateQueueEnd = 0;

		if ( uNewState != m_uCurrentState )
		{
			if ( uNewState == ACTOR_STATE_IN_MOVE || uNewState == ACTOR_STATE_IN_GAMESCENE || uNewState == ACTOR_STATE_DEAD ) // just set in_move, in_gamescene or dead state as our next state
			{
				m_uStateQueueEnd						= 1;
				m_auStateQueue[ m_uStateQueueStart ]	= uNewState;
				SetNeedToTween( 0.2f );
				m_bPlayNextAnimFromStart = true;
				m_xServerActor.SetAIStateAnimsNeedToTween( 0.2f );
				m_xServerActor.SetPlayNextAIStateAnimFromStart();
			}
			else if ( m_uCurrentState == ACTOR_STATE_IN_MOVE /* || m_uCurrentState == ACTOR_STATE_IN_GAMESCENE */ ) // just set next state if coming from in_move or in_gamescene
			{
				m_uStateAfterMove = uNewState;
			}
			else if ( !AStarSearch( m_auStateQueue, m_uStateQueueEnd, m_uCurrentState, uNewState/*, fMaxCost*/ ) )
			{
				// If we can't route find to the target state then just go straight to it.
				Asura_Assertf( false, ( "AStar can't get from state %d to state %d forcing state as a fall back", m_uCurrentState, uNewState ) );
				ForceState( uNewState );
			}
		}
	}
}

void AVP_Actor_AIAnimStateMachine::ForceState( const u_int uNewState )
{
	EndState( m_uCurrentState );

#ifdef _DEBUG
	for ( u_int uItr = 0; uItr != AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE; ++uItr )
	{
		// Set the queue to a default value.
		m_auStateQueue[ uItr ] = 0xFFFFFFFF;
	}
#endif
	m_uStateQueueStart = 0;
	m_uStateQueueEnd = 0;
	m_uCurrentState = uNewState;
	m_fTimeInState	= 0.0f;

	BeginState( m_uCurrentState );
}

void AVP_Actor_AIAnimStateMachine::BeginState( const u_int uState )
{
	// get move dir in OS
	Asura_Vector_3 xMoveDir = m_xServerActor.GetMovementDirection();
	xMoveDir = m_xServerActor.GetOrientation().ApplyTranspose( xMoveDir );
	m_fOSMoveDirAngle = Asura_Maths::RadiansToZeroMinimumRange( Asura_Maths::ArcTangent( xMoveDir.GetX(), xMoveDir.GetZ() ) );

	const AVP_Actor_AnimStateData* pxNewStateData = GetStateAnimData( uState );
	if ( pxNewStateData != NULL )
	{
		m_xServerActor.SetCollisionDescriptor( pxNewStateData->GetCollisionDescriptorHash() );
	}

	const AVP_Actor_AIAnimStateConnectionData* pxConnection = GetConnectionData( GetCurrentState(), uState );
	if ( pxConnection )
	{
		float fTweenTime = pxConnection->GetTweenTime();
		const AVP_Actor_AnimStateData* pxCurrentStateData = GetStateAnimData( m_uCurrentState );
		if ( pxCurrentStateData && pxCurrentStateData->GetAnimHash() == ASURA_HASH_ID_UNSET && pxCurrentStateData->GetNumBlendRegions() == 0 )
		{
			fTweenTime = dbg_fMissingTransitionAnimTweenTime;
		}
		SetNeedToTween( fTweenTime );
		m_bPlayNextAnimFromStart = !pxConnection->GetSync();
		m_xServerActor.SetAIStateAnimsNeedToTween( fTweenTime );
		if ( !pxConnection->GetSync() )
		{
			m_xServerActor.SetPlayNextAIStateAnimFromStart();
		}
	} 
	else
	{
		SetNeedToTween( dbg_fNewStateTweenTime );
		m_bPlayNextAnimFromStart = true;
		m_xServerActor.SetAIStateAnimsNeedToTween( dbg_fNewStateTweenTime );
		m_xServerActor.SetPlayNextAIStateAnimFromStart();
	}

	if ( uState == ACTOR_STATE_IN_GAMESCENE )
	{
		m_bGamesceneFinished = false;
		m_xServerActor.SetMoving( true );
	}
}

void AVP_Actor_AIAnimStateMachine::UpdateState( const u_int uState )
{
	AVP_ClientEntity_Actor*		pxClientActor = AVP_ClientEntity_Actor::FindEntity( m_xServerActor.GetGuid() );
	AVP_Actor_AnimationHandler* pxAnimHandler = ( pxClientActor != NULL ) ? pxClientActor->GetAnimationHandler() : NULL;
	if ( !pxAnimHandler ) { return; }

	bool bPauseNextAnimState = false;
	const AVP_Actor_AnimStateData* pxCurrentStateData;
	if ( ( uState == ACTOR_STATE_IN_GAMESCENE || uState == ACTOR_STATE_IN_MOVE ) )
	{

		const AVP_Move* pxMove = pxClientActor->GetMovementSystem() ? static_cast< const AVP_Move* >(pxClientActor->GetMovementSystem()->GetCurrentMove()) : NULL;
		const AVP_MoveData_Animation* pxMoveData = pxMove ? pxMove->GetAVPAnimationData() : NULL;
		const float fTweenInTime = pxMoveData ? pxMoveData->GetTweenInDuration() : dbg_fStartNextStateAnimTime;


		if ( GetTimeInState() > fTweenInTime )
		{
			if ( uState == ACTOR_STATE_IN_MOVE && !pxMove )
			{
				ForceState( m_uStateAfterMove );
			}

			pxCurrentStateData = GetStateAnimData( m_uStateAfterMove );
			m_fCurrentTweenTime = 0.0f;
			m_bPlayNextAnimFromStart = !m_bPlayingNextAnim;
			bPauseNextAnimState = true;
			m_bPlayingNextAnim = true;
		}
		else
		{
			m_bPlayingNextAnim = false;
			pxCurrentStateData = GetStateAnimData( uState );
		}
	}
	else
	{
		m_bPlayingNextAnim = false;
		pxCurrentStateData = GetStateAnimData( uState );
	}

	if ( ShouldStateUpdateMoveDirAngle( uState ) )
	{
		// get move dir in OS
		Asura_Vector_3 xMoveDir = m_xServerActor.GetMovementDirection();
		xMoveDir = m_xServerActor.GetOrientation().ApplyTranspose( xMoveDir );
		m_fOSMoveDirAngle = Asura_Maths::RadiansToZeroMinimumRange( Asura_Maths::ArcTangent( xMoveDir.GetX(), xMoveDir.GetZ() ) );
	}
	if ( pxCurrentStateData )
	{
		pxClientActor->UpdateAnimStateData( pxCurrentStateData, NULL, pxCurrentStateData, NULL, IsLoopedState( uState ), 0.0f, 0.0f, m_fOSMoveDirAngle, m_fCurrentTweenTime, !m_bPlayNextAnimFromStart );
		pxClientActor->SetAIAnimState( uState );
		pxClientActor->SetMoving( m_xServerActor.IsMoving() );
		pxClientActor->SetMovingSlow( m_xServerActor.IsMovingSlow() );
		pxClientActor->SetMovingFast( m_xServerActor.IsMovingFast() );
		if ( bPauseNextAnimState )
		{
			pxClientActor->PauseAnimState();
		}
		else
		{
			pxClientActor->PlayAnimState();
		}
	}

	if ( HasStateInQueue() )
	{
		if ( !IsCurrentStateFinished() )
		{
			pxAnimHandler->SetCurrentAnimStateLooped( false );
		}
	}
	else if ( IsLoopedState( uState ) )
	{
		pxAnimHandler->SetCurrentAnimStateLooped( true );
		if ( !bPauseNextAnimState )
		{
			pxAnimHandler->PlayCurrentAnimState();
		}
	}

	m_fCurrentTweenTime = 0.0f;
	m_bPlayNextAnimFromStart = false;
}

bool AVP_Actor_AIAnimStateMachine::EndState( const u_int uState )
{
	if ( uState == ACTOR_STATE_IN_GAMESCENE )
	{
		m_xServerActor.SetMoving( false );
	}

	return true;
}

#ifndef ASURA_NO_DEBUG_OUTPUT
void AVP_Actor_AIAnimStateMachine::RenderDebugInfo()
{
	Asura_Vector_3 xDebugPos( m_xServerActor.GetPosition() );
	if ( Asura_Camera::ProjectWorldCoordsToScreenSpace( xDebugPos, true ) && xDebugPos.z > 0.0f)
	{
		Asura_Font_System::Print
		(
			xDebugPos.GetX(),
			xDebugPos.GetY() - 10.0f,
			0.6f,
			0xffffffff,
			ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, 
			"Current State: %d - %s",
			m_uCurrentState, GetStateDescription( m_uCurrentState )
		);
	}
}
#endif

bool AVP_Actor_AIAnimStateMachine::IsCurrentStateFinished() const
{
	bool bFinished( true );
	switch( m_uCurrentState )
	{
		case ACTOR_STATE_DEAD:
		{
			bFinished = false; // not sure if we ever finish being dead
		}
		break;

		case ACTOR_STATE_IN_MOVE:
		{
			bFinished = !m_bInMove;
		}
		break;

		case ACTOR_STATE_IN_GAMESCENE:
		{
			bFinished = m_bGamesceneFinished;
		}
		break;

		default: // This is a typical anim based state
		{
			const AVP_Actor_AnimStateData* pxCurrentStateData = GetStateAnimData( m_uCurrentState );
			const AVP_Actor_AIAnimStateConnectionData* pxConnectionToNext = HasStateInQueue() ? GetConnectionData( m_uCurrentState, m_auStateQueue[ m_uStateQueueStart ] ) : NULL;
			Asura_Assert( pxCurrentStateData, "Bad state data" );
			if ( pxCurrentStateData == NULL || ( pxCurrentStateData->GetAnimHash() == ASURA_HASH_ID_UNSET && pxCurrentStateData->GetNumBlendRegions() == 0 ) || pxConnectionToNext == NULL || !pxConnectionToNext->GetWaitForEndOfLoop() )
			{
				bFinished = true;
			}
			else
			{
				const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity( m_xServerActor.GetGuid() );
				const AVP_Actor_AnimationHandler* pxAnimationHandler = pxClient != NULL ? pxClient->GetAnimationHandler() : NULL;
				if ( pxAnimationHandler != NULL )
				{
					return pxAnimationHandler->IsCurrentAnimStateFinished();
				}
				else
				{
					bFinished = true;
				}
			}
		} // end default:
		break;
	} // end switch( m_uCurrentState )
	return bFinished;
}

bool AVP_Actor_AIAnimStateMachine::AStarSearch( u_int* puRouteArray, u_int& uRouteLength, const u_int uFromState, const u_int uToState/*, float fMaxCost*/ )
{
	float fMaxCost( Asura_Maths::LargeFloat );
	Asura_Assert( uFromState < AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES && uFromState < GetNumStates(), "From state out of range" );
	Asura_Assert( uToState < AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES && uToState < GetNumStates(), "To state out of range" );
	if ( uFromState >= AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES || uToState >= AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES )
	{
		return false;
	}

	for ( u_int u = 0; u < AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES; ++u )
	{
		s_axAStarNodeScratchPad[ u ].m_uFlags = 0;
		s_axAStarNodeScratchPad[ u ].m_pxParent = NULL;
	}

	AVP_Actor_AIAnimStateMachine::AStarNode* pxCurrentNode = &s_axAStarNodeScratchPad[ uFromState ];
	pxCurrentNode->m_fCostToHere	= 0.0f;
	pxCurrentNode->m_pxParent		= NULL;

	while ( pxCurrentNode )
	{
		for ( u_int uToStateItr = 0; uToStateItr < GetNumStates(); ++uToStateItr )
		{
			u_int uCurrentState = static_cast< u_int >( pxCurrentNode - s_axAStarNodeScratchPad );
			const AVP_Actor_AIAnimStateConnectionData* pxConnectionData = GetConnectionData( uCurrentState, uToStateItr );
			if ( pxConnectionData != NULL )
			{
				AVP_Actor_AIAnimStateMachine::AStarNode* pxNextNode = &s_axAStarNodeScratchPad[ uToStateItr ];

				float fNewCostToHere = pxCurrentNode->m_fCostToHere + pxConnectionData->GetCost();

				// If we've already been here and had a better cost last time, skip out.
				if ( pxNextNode->m_uFlags && pxNextNode->m_fCostToHere <= fNewCostToHere )
				{
					continue;
				}

				pxNextNode->m_fCostToHere = fNewCostToHere;
				pxNextNode->m_pxParent    = pxCurrentNode;
				if ( pxNextNode == &s_axAStarNodeScratchPad[ uToState ] )
				{
					pxNextNode->m_bOpen   = false;
					pxNextNode->m_bClosed = true;
					fMaxCost = pxNextNode->m_fCostToHere;
				}
				pxNextNode->m_bOpen   = true;
				pxNextNode->m_bClosed = false;
			}
		}
		
		do
		{
			pxCurrentNode->m_bOpen   = false;
			pxCurrentNode->m_bClosed = true;
			pxCurrentNode = GetOpenAStarNode();
		} while ( pxCurrentNode != NULL && pxCurrentNode->m_fCostToHere > fMaxCost );
	}

	// Copy out the route
	pxCurrentNode = s_axAStarNodeScratchPad + uToState;
	if ( !pxCurrentNode->m_pxParent )
	{
		return false;
	}

	uRouteLength = 0;
	while ( pxCurrentNode->m_pxParent )
	{
		if ( uRouteLength < AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE )
		{
			puRouteArray[ AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE - ( 1 + uRouteLength ) ] = u_int( pxCurrentNode - s_axAStarNodeScratchPad );
			++uRouteLength;
			pxCurrentNode = pxCurrentNode->m_pxParent;
		}
		else
		{
			Asura_Assertf( false, ( "Not enough space in route buffer to store best route from state %d to state %d", uFromState, uToState ) );
			return false;
		}
	}

	// Push route buffer to beginning of array
	for ( u_int uRouteItr = 0; uRouteItr < uRouteLength; ++uRouteItr )
	{
		puRouteArray[ uRouteItr ] = puRouteArray[ uRouteItr + AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES_IN_QUEUE - uRouteLength ];
	}
	return true;
}

AVP_Actor_AIAnimStateMachine::AStarNode* AVP_Actor_AIAnimStateMachine::GetOpenAStarNode()
{
	for ( u_int uStateItr = 0; uStateItr < AVP_ACTOR_AIANIMSTATEMACHINE_MAX_STATES; ++uStateItr )
	{
		if ( s_axAStarNodeScratchPad[uStateItr].m_bOpen )
		{
			return &s_axAStarNodeScratchPad[uStateItr];
		}
	}

	// if no open nodes are found return NULL
	return NULL;
}

const char* AVP_Actor_AIAnimStateMachine::GetStateDescriptionStatic( const u_int uState )
{
	Asura_Assert( uState < ACTOR_STATE_END, "Invalid state" );
	return g_aszStateString[ uState ];
}

bool AVP_Actor_AIAnimStateMachine::IsLoopedStateStatic( const u_int uState )
{
	Asura_Assert( uState < ACTOR_STATE_END, "Invalid state" );
	return g_abLoopedState[ uState ];
}

/// End of file. ///
