
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Actor_AnimationHandler.cpp
// Author : Andy Tate
// Date   : 18-Oct-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/AVP_Actor_AnimationHandler.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationAttachment.h"
#include "Animation/Asura_AnimationBlend.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Core/Asura_Timers.h"
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "FacialAnimation/Asura_FacialAnimation_System.h"
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "Interface/Asura_CommandConsole.h"
#include "Shapes/Asura_Shapes.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AttachmentHandler.h"
#include "Actor/AVP_Actor_FidgetAnimHandler.h"
#include "Actor/Vertebrates/Human/AVP_Human_Description.h"
#include "Actor/Vertebrates/Alien/AVP_Alien_Description.h"
#include "Actor/Vertebrates/FaceHugger/AVP_FaceHugger_Description.h"
#include "Actor/Vertebrates/Predator/AVP_Predator_Description.h"
#include "Animation/AVP_Animation_BodyRig.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Game_Core.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ServerActor_Human.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/Items/Weapons/AVP_InventoryWeapon_RangedWeapon.h"
#include "Moves/AVP_Moves.h"

// axon
#include "AnimTemplates/Axon_Gamescene_AnimTemplates.h"
#include "AnimTemplates/Axon_Gamescene_AnimTemplate_System.h"
#include "Core/BrainTypes/Axon_Project_Brain.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const Asura_Hash_ID	g_uCameraMarkerPointHash	= 0x002FBC62;	// Asura_GetHashID("eyes");

DVSVAR Asura_Guid	dvs_uAnimationDisplayGuid	= ASURA_GUID_UNREGISTERED;

#ifdef _DEBUG

static const char* s_szRootStatesGroupName = "StatesGroup";
static const char* s_szRootWeaponGroupName = "WeaponGroup";

#else

static const Asura_Hash_ID s_uRootStatesGroupHash = 0x7ced99bd; //"StatesGroup"
static const Asura_Hash_ID s_uRootWeaponGroupHash = 0xe92425e3; //"WeaponGroup"

#endif

#ifdef ASURA_TOOLS
const char* s_aszAVPPostureStrings[ NUM_AVP_ANIM_POSTURE ] =
{
	"Stand",
	"Crouch",
	"Front",
	"Back"
};

const char* s_aszAVPPoseStrings[ NUM_AVP_ANIM_POSE ] =
{
	"Relaxed",
	"Neutral",
	"Alert",
	"Aim"
};
#endif

AVP_GS2_AnimMappingData AVP_Actor_AnimationHandler::s_xAnimMappingData;

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

const Asura_Hash_ID AVP_Actor_AnimationHandler::s_auTargetMarkerPointHashes[MAX_MARKER_POINTS] =
{
	0xE5A85D03,		// Asura_GetHashID("target_1");
	0xE5A85D04,		// Asura_GetHashID("target_2");
};

#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
int AVP_Actor_AnimationHandler::s_iConsoleVarsCount = 0;
#endif	// defined(ASURA_DEBUG_VARIABLE_SYSTEM)

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
static bool SetDebugGuidToPlayerGuid( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf( L"AVP.Animation.SetDebugGuidToPlayerGuid" L": Set guid to player guid" );
	}
	else
	{
		dvs_uAnimationDisplayGuid = 2000000;
	}
	return true;
}
#endif // ASURA_DEBUG_VARIABLE_SYSTEM

////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_AnimationHandler::UpdateSkin
// Author       : Andy Tate
// Date         : 18-Oct-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_AnimationHandler::UpdateSkin(void)
{
	const Asura_Blueprint*	pxBlueprint	= m_xOwnerActor.GetBlueprint();

	if(pxBlueprint != NULL)
	{
		Asura_Hash_ID	uSkinHash	= ASURA_HASH_ID_UNSET;

		// Decide on which skin to use...
		if(IsFirstPerson())
		{
			Asura_Value	xValue;

			if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FIRSTPERSON_SKIN_MAIN, xValue))
			{
				uSkinHash	= xValue.GetHashIDValue();
			}
		}
		else 
		{
			Asura_Value	xValue;

			if(IsDamaged())
			{
				if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_THIRDPERSON_SKIN_DAMAGED_MAIN, xValue))
				{
					uSkinHash	= xValue.GetHashIDValue();
				}
			}
			else
			{
				if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_THIRDPERSON_SKIN_MAIN, xValue))
				{
					uSkinHash	= xValue.GetHashIDValue();
				}
			}
		}

		// Change to this skin (if appropriate)...
		if((uSkinHash != ASURA_HASH_ID_UNSET) && (uSkinHash != m_uSkinHash))
		{
			if(m_pxAnimation != NULL)
			{
				Asura_Hierarchy_Skin* pxNewSkin = Asura_Animation_System::GetSkinByHash(uSkinHash);
				Asura_Assert( pxNewSkin, "New skin is invalid" );
				Asura_Assert( m_pxAnimation->GetSkin(), "Current skin is invalid" );
				Asura_Assert( m_pxAnimation->GetSkin() == NULL || pxNewSkin == NULL || m_pxAnimation->GetSkin()->NumberOfBones == pxNewSkin->NumberOfBones, "Skins have different numbers of bones, this is bad" );
				if ( m_pxAnimation->GetSkin() == NULL || pxNewSkin == NULL || m_pxAnimation->GetSkin()->NumberOfBones == pxNewSkin->NumberOfBones )
				{
					m_pxAnimation->SetSkin(uSkinHash);
				}
				m_pxAnimation->DestroyAllAttachments();
			}
			else
			{
				// Handle this here, subclasses should be overriding this function anyway if they need special skin setup handling.
				m_pxAnimation	= new Asura_Animation(uSkinHash);
			}

			// Finally, update the skin we're using..
			m_uSkinHash	= uSkinHash;
		}
	}
}

/// End of function 'AVP_Actor_AnimationHandler::UpdateSkin'. ///


bool AVP_Actor_AnimationHandler::PlayingGamesceneAnimSequence() const
{
    return ( m_xCurrentAnimSequenceData.m_uAnimTemplateSequenceID != uAXON_ANIMATION_SEQUENCE_ID_NONE );
}

Asura_Hash_ID AVP_Actor_AnimationHandler::PlayAnimation( const u_int uAnimFlags,
														 const u_int uPriority,
														 Asura_Hash_ID uAnimHashID,
														 Asura_AnimationBlend* pxBlend,
														 Asura_AnimationController_Group* pxParentGroup /* = NULL */,
														 const float fParametricStartTime /*= 0.0f*/ ) const
{
	Asura_AnimationController_Animation* pxAnimation = NULL;
#if defined(_DEBUG)
	const Asura_Hierarchy_Anim*			pxHierarchyAnim	= Asura_Animation_System::GetAnimByHash(uAnimHashID);
	if ( pxHierarchyAnim )
	{
		const char*							szAnimName		= pxHierarchyAnim->GetName();
		pxAnimation = m_pxAnimation->CreateControllerAnimation(szAnimName, uAnimFlags, uPriority, uAnimHashID, pxBlend, pxParentGroup );
	}
	else
	{
		pxAnimation = m_pxAnimation->CreateControllerAnimation( uAnimFlags, uPriority, uAnimHashID, pxBlend, pxParentGroup );
	}
#else
	pxAnimation = m_pxAnimation->CreateControllerAnimation( uAnimFlags, uPriority, uAnimHashID, pxBlend, pxParentGroup );
#endif

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	Asura_FacialAnimation_AnimationHandler*	pxFacialAnimationHandler	= Asura_FacialAnimation_System::FindAnimationHandler(m_uOwnerGuid);

	// Play the facial-animation (if there is one)...
	if(pxFacialAnimationHandler != NULL)
	{
		Asura_FacialAnimation_FaceAnimation* pxFaceAnimation	= Asura_FacialAnimation_System::FindFaceAnimation(uAnimHashID);

		if(pxFaceAnimation != NULL)
		{
			pxFacialAnimationHandler->PlayFaceAnimation(uAnimHashID);
		}
	}
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)

	if ( pxAnimation )
	{
		pxAnimation->SetParametricTimer( fParametricStartTime );
		if (pxAnimation->GetAnimation())
		{
			pxAnimation->Play();
			return pxAnimation->GetControllerHash();
		}
		else
		{
			m_pxAnimation->DestroyController(pxAnimation->GetControllerHash());
			return ASURA_HASH_ID_UNSET;
		}
	}
	else
	{
		Asura_Assertf( false, ( "Unable to create animation with hash ID 0x%08X", uAnimHashID ) );
		return ASURA_HASH_ID_UNSET;
	}
}



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_AnimationHandler::AVP_Actor_AnimationHandler
// Author       : Andy Tate
// Date         : 18-Oct-2006
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_AnimationHandler::AVP_Actor_AnimationHandler(AVP_ClientEntity_Actor& xOwner)
: m_xGS2SequenceHandler()
, m_pxAnimation( NULL )
, m_pxBodyRig( NULL )
, m_pxFidgetAnimHandler( NULL )
, m_pxCurrentAnimationTemplateEntry( NULL )
, m_uOwnerGuid(xOwner.GetGuid())
, m_xOwnerActor(xOwner)
, m_uSkinHash(ASURA_HASH_ID_UNSET)
, m_uRootStatesGroupHash(ASURA_HASH_ID_UNSET)
, m_uRootWeaponGroupHash(ASURA_HASH_ID_UNSET)
, m_uCurrentGamesceneAnimControllerHash( ASURA_HASH_ID_UNSET )
, m_uCurrentGamesceneFidgetAnimControllerHash( ASURA_HASH_ID_UNSET )
, m_uCurrentRequestedAnimControllerHash( ASURA_HASH_ID_UNSET )
, m_uCurrentExtendWeaponAnimHashID( ASURA_HASH_ID_UNSET )
, m_uCurrentRetractWeaponAnimHashID( ASURA_HASH_ID_UNSET )
, m_uCurrentPermanentWeaponAnimHashID( ASURA_HASH_ID_UNSET )
, m_uDeathAnimController(ASURA_HASH_ID_UNSET)
, m_uNumTargetMarkerPoints( 0 )
, m_uFlags( 0 )
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_iConsoleVarsCount == 0)
	{
		if(!Asura_CommandConsole::VarExists(L"AVP.Animation.DebugDisplayGuid"))
		{
			Asura_CommandConsole::AddVar(L"AVP.Animation.DebugDisplayGuid", &dvs_uAnimationDisplayGuid, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
		}
		if(!Asura_CommandConsole::CmdExists(L"AVP.Animation.SetDebugGuidToPlayerGuid"))
		{
			Asura_CommandConsole::AddCmd(L"AVP.Animation.SetDebugGuidToPlayerGuid", SetDebugGuidToPlayerGuid, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
		}
	}
	++s_iConsoleVarsCount;
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

	m_bFirstPerson = AVP_Game_Core::IsLocalPlayer(xOwner.GetGuid());
}
/// End of function 'AVP_Actor_AnimationHandler::AVP_Actor_AnimationHandler'. ///

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_AnimationHandler::~AVP_Actor_AnimationHandler
// Author       : Andy Tate
// Date         : 9-Nov-2006
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_AnimationHandler::~AVP_Actor_AnimationHandler()
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	--s_iConsoleVarsCount;
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
}

/// End of function 'AVP_Actor_AnimationHandler::~AVP_Actor_AnimationHandler'. ///

void AVP_Actor_AnimationHandler::Initialise( void )
{
	Asura_Blueprint*	pxBlueprint	= m_xOwnerActor.GetBlueprint();

	if(pxBlueprint == NULL)
	{
		Asura_Assertf(false, ("Cannot find the blueprint for actor %d - initialisation failed.", m_uOwnerGuid));
	}
	else
	{
		Asura_Shape_Instance&		xShapeInstance	= m_xOwnerActor.GetShapeInstance();
		const Asura_Hierarchy_Skin*	pxSkin			= CreateAnimation();

		// Initialise the shape-instance...
		m_pxAnimation->TimersUpdatedBeforeRender	= true;
		m_pxAnimation->SetCalculateBSUsingBones(true);
		m_pxAnimation->SetParentEntity( m_uOwnerGuid );

		xShapeInstance.SetAnimationPointer( m_pxAnimation );

		// Set target points to 0 ready for initialisation
		m_uNumTargetMarkerPoints = 0;

		if(pxSkin)
		{
			//Do we have a marker-point with the name target+i ?
			for(int i=0;i<MAX_MARKER_POINTS;i++)
			{
				int	iIndex;

				if(	pxSkin->GetMarkerPointIndex(s_auTargetMarkerPointHashes[i], iIndex))
				{
					//Increment marker-point count
					m_uNumTargetMarkerPoints++;
				}
			}
		}

		Asura_AnimationBlend*	pxStatesGroupBlend	= new Asura_AnimationBlend(1.0f);
		Asura_AnimationBlend*	pxWeaponGroupBlend	= new Asura_AnimationBlend(1.0f);

		Asura_Assert( m_uRootStatesGroupHash == ASURA_HASH_ID_UNSET, "Already got a weapon extend root group?  Shouldn't have one.");
		Asura_Assert( m_uRootWeaponGroupHash == ASURA_HASH_ID_UNSET, "Already got a weapon root group?  Shouldn't have one.");

#ifdef _DEBUG
		Asura_AnimationController_Group* pxRootStatesGroup = m_pxAnimation->CreateControllerGroup(s_szRootStatesGroupName, AAC_FLAG_ACTIVE | AACG_FLAG_DONT_PASS_DESTROY_TO_CHILDREN, AVP_AAPC_STATES_GROUP, pxStatesGroupBlend);
		Asura_AnimationController_Group* pxRootWeaponGroup = m_pxAnimation->CreateControllerGroup(s_szRootWeaponGroupName, AAC_FLAG_ACTIVE | AACG_FLAG_DONT_PASS_DESTROY_TO_CHILDREN | AACG_FLAG_INHERIT_PARENT_TWEEN_STORE, AVP_AAPC_WEAPON_GROUP, pxWeaponGroupBlend);
#else
		Asura_AnimationController_Group* pxRootStatesGroup = m_pxAnimation->CreateControllerGroup(s_uRootStatesGroupHash, AAC_FLAG_ACTIVE | AACG_FLAG_DONT_PASS_DESTROY_TO_CHILDREN, AVP_AAPC_STATES_GROUP, pxStatesGroupBlend);
		Asura_AnimationController_Group* pxRootWeaponGroup = m_pxAnimation->CreateControllerGroup(s_uRootWeaponGroupHash, AAC_FLAG_ACTIVE | AACG_FLAG_DONT_PASS_DESTROY_TO_CHILDREN | AACG_FLAG_INHERIT_PARENT_TWEEN_STORE, AVP_AAPC_WEAPON_GROUP, pxWeaponGroupBlend);
#endif

		Asura_Assert( pxRootWeaponGroup != NULL, "Did not create root human weapons animation group" );
		if ( pxRootWeaponGroup == NULL )
		{
			delete pxWeaponGroupBlend;
		}
		else
		{
			m_uRootWeaponGroupHash = pxRootWeaponGroup->GetControllerHash();
		}

		Asura_Assert( pxRootStatesGroup != NULL, "Did not create root human states animation group" );
		if ( pxRootStatesGroup == NULL )
		{
			delete pxStatesGroupBlend;
		}
		else
		{
			m_uRootStatesGroupHash = pxRootStatesGroup->GetControllerHash();
		}
	}
}

void AVP_Actor_AnimationHandler::Shutdown( void )
{
	// Delete the attachment handler
	if (m_xOwnerActor.GetAttachmentHandler())
	{
		delete(m_xOwnerActor.GetAttachmentHandler());
		m_xOwnerActor.ClearAttachmentHandler();
	}

	// Destroy all moves, as these have animation pointers in them.
	Asura_Movement_System* pxMS = m_xOwnerActor.GetMovementSystem();
	if ( pxMS )
	{
		pxMS->Reset();
	}

	if( m_pxFidgetAnimHandler )
	{
		delete m_pxFidgetAnimHandler;
		m_pxFidgetAnimHandler = NULL;
	}

	if(m_xOwnerActor.GetParentGuid() != ASURA_GUID_UNREGISTERED)
	{
		const float	fServerTime	= Asura_Timers::GetServerAbsoluteTime();

		m_xOwnerActor.UnlinkFromParent(fServerTime);
	}

	m_pxAnimation = NULL;

	// Clean up the shape-instance...
	Asura_Shape_Instance& xShapeInstance = m_xOwnerActor.GetShapeInstance();
	xShapeInstance.SetAnimationPointer( NULL ); // deletes the animation

	// Set target points to 0 ready for initialisation
	m_uNumTargetMarkerPoints = 0;

	m_uRootStatesGroupHash = ASURA_HASH_ID_UNSET;
	m_uRootWeaponGroupHash = ASURA_HASH_ID_UNSET;
}

void AVP_Actor_AnimationHandler::HandleGamesceneAnimSequenceData( const Axon_Gamescene_AnimSequenceData& xAnimSequenceData )
{
	m_xGS2SequenceHandler.HandleData( xAnimSequenceData );
}

void AVP_Actor_AnimationHandler::UpdateGamesceneAnimSequence()
{
	m_xGS2SequenceHandler.UpdateAnimSequence( GetAnimMappingData() );
}

bool AVP_Actor_AnimationHandler::RequestedAnimationHasFinished()
{
    if ( m_uCurrentRequestedAnimControllerHash != ASURA_HASH_ID_UNSET )
    {
        Asura_AnimationController_Animation* pxAnimController = m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentRequestedAnimControllerHash );
        if ( pxAnimController )
        {
            Asura_Assert( !pxAnimController->IsLooped(), "AVP_Actor_AnimationHandler::RequestedAnimationHasFinished - animation is set to loop" );
            return pxAnimController->IsFinished();
        }
        else
        {
            return true;
        }
    }
    else
    {
        return true;
    }
}

void AVP_Actor_AnimationHandler::Reset()
{
	m_uDeathAnimController = ASURA_HASH_ID_UNSET;
}

void AVP_Actor_AnimationHandler::Update()
{
	if ( m_pxBodyRig )
	{
		m_pxBodyRig->Update();
	}

	// Update Fidget Engine.
	if ( m_pxFidgetAnimHandler )
	{
		m_pxFidgetAnimHandler->Update();
	}

    UpdateGamesceneAnimSequence();

	if ( m_pxAnimation )
	{
		Asura_AnimationController_Animation* pxGS2AACA = m_uCurrentGamesceneAnimControllerHash != ASURA_HASH_ID_UNSET ? m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentGamesceneAnimControllerHash ) : NULL;
		if ( m_xGS2SequenceHandler.AnimChangedThisUpdate() )
		{
			if ( m_xGS2SequenceHandler.GetAnimHashID() == ASURA_HASH_ID_UNSET )
			{
				const Asura_Movement_System* pxMS = m_xOwnerActor.GetMovementSystem();
				const Asura_Move* pxMove = pxMS ? pxMS->GetCurrentMove() : NULL;
				if ( pxMove && pxMove->GetHash() == AVP_Move_GS2RigidAnim::GetStaticHash() )
				{
					// We're on a rigid GS2 spline and want to finish early (normally the move will finish before the spline behaviour does), abort the move (we'll handle the tween out there)
					Asura_Message* pxMoveMessage = Asura_Network::CreateMessage( ASURA_MSG_MOVE_ABORT, sizeof( Asura_Msg_Move ) );
					if ( pxMoveMessage != NULL )
					{
						Asura_Msg_Move*		pxData		= static_cast<Asura_Msg_Move*>(pxMoveMessage->GetData());
						const Asura_Guid	uOwnerGuid	= m_xOwnerActor.GetGuid();
						if(pxData != NULL)
						{
							pxData->ClearFlags();
							pxData->SetMoveID(AVP_Move_GS2RigidAnim::GetStaticHash());
						}
						pxMoveMessage->SetToGuid(uOwnerGuid);
						pxMoveMessage->SetFromGuid(uOwnerGuid);

						// Forward the message directly to the client... we will do this on each client
						m_xOwnerActor.HandleMessage( pxMoveMessage );
						Asura_Network::DestroyMessage( pxMoveMessage );
					}
				}
				else
				{
					// Our GS2 behaviour has finished (and there are no further GS2 anims to play) so clean up the anim left behind (either by the rigid spline move, or the anim created below for normal GS2)
					// It's valid for the anim to be missing here as if a GS2RigisSpline move gets aborted from an external source then the anim's going to have been cleaned up. Hence no assert
					if ( pxGS2AACA )
					{
						// this will stop us or tween us out as necessary
						u_int uFlags = pxGS2AACA->GetControllerFlags();
						uFlags |= AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE;
						uFlags &= ~AACA_FLAG_LOOPED;
						pxGS2AACA->SetControllerFlags( uFlags );

						Asura_AnimationBlend_Tween* pxTween = static_cast<Asura_AnimationBlend_Tween*>( pxGS2AACA->GetBlendToPreviousController() );
						if ( pxTween )
						{
							pxTween->ChangeTweenDuration( 0.0f, DEFAULT_TWEEN_TIME );
						}

						if ( pxGS2AACA->GetTimeUntilFinished() > 0.2f )
						{
							pxGS2AACA->SetParametricTargetTime( pxGS2AACA->GetParametricTimer() + ( DEFAULT_TWEEN_TIME / pxGS2AACA->GetDuration() ) );
						}
					}
				}
				m_uCurrentGamesceneAnimControllerHash = ASURA_HASH_ID_UNSET;
			}
			else
			{
				// We've got a new GS2 anim to play
				if ( pxGS2AACA )
				{
					// If we've currently got a GS2 anim playing (left behind after a prior node/spline anim has finished normally) then allow it to be tweened over by the new one.
					pxGS2AACA->SetControllerFlags( pxGS2AACA->GetControllerFlags() | AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE );
				}

				u_int uAnimFlags = AAC_FLAG_ACTIVE | AAC_FLAG_ENSURE_UNIQUE_HASH | AACA_FLAG_DONT_TWEEN_VELOCITY | AACA_FLAG_DONT_TWEEN_ANGULAR_VELOCITY | AACA_FLAG_DONT_TWEEN_MOVEMENT_OFFSET;
				uAnimFlags |= m_xGS2SequenceHandler.GetAnimFlags();

				const float fTweenInTime = ( m_bNotFirstUpdate && m_xGS2SequenceHandler.AnimShouldTween() && GetCurrentAnimStateData() != NULL )
											? m_xGS2SequenceHandler.GetAnimTweenInTime() : 0.0f;

				if ( m_xGS2SequenceHandler.IsRigidAnimTemplate() )
				{
					// If it's a rigid spline anim then play it in the move.
					Asura_Message* pxMoveMessage = Asura_Network::CreateMessage( ASURA_MSG_MOVE_START, sizeof( AVP_Msg_Move_Start_GS2RigidAnim ) );
					if ( pxMoveMessage != NULL )
					{
						AVP_Msg_Move_Start_GS2RigidAnim* pxMoveData = static_cast< AVP_Msg_Move_Start_GS2RigidAnim* >( pxMoveMessage->GetData() );

						// Fill-in the message-body...
						pxMoveMessage->SetFromGuid( m_xOwnerActor.GetGuid() );
						pxMoveMessage->SetToGuid( m_xOwnerActor.GetGuid() );

						pxMoveData->ClearFlags();
						pxMoveData->SetMoveID( AVP_Move_GS2RigidAnim::GetStaticHash() );

						// play the animation associated with this new entry.
						pxMoveData->SetData( m_xGS2SequenceHandler.GetAnimHashID(), uAnimFlags, fTweenInTime, m_xGS2SequenceHandler.GetAnimParametricStartTime(), m_xGS2SequenceHandler.GetAnimStartOrientQ(), m_xGS2SequenceHandler.GetAnimStartPosition() );

						// Forward the message directly to the client... we will do this on each client
						m_xOwnerActor.HandleMessage( pxMoveMessage );
						Asura_Network::DestroyMessage( pxMoveMessage );
					}
				}
				else
				{
					// Otherwise play it here...
					const float fParametricStartTime = m_xGS2SequenceHandler.GetAnimParametricStartTime();
					Asura_AnimationBlend* pxAnimBlend = new Asura_AnimationBlend_AnimatedTween( fTweenInTime, 0.0f, false );
					m_uCurrentGamesceneAnimControllerHash = PlayAnimation( uAnimFlags, AVP_AAPC_GAMESCENE_ANIM,
						m_xGS2SequenceHandler.GetAnimHashID(), pxAnimBlend, 0, fParametricStartTime );
				}
			}
		}
		
		if ( m_xGS2SequenceHandler.FidgetAnimChangedThisUpdate() )
		{
			Asura_AnimationController_Animation* pxGS2FidgetAACA = m_uCurrentGamesceneFidgetAnimControllerHash != ASURA_HASH_ID_UNSET ? m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentGamesceneFidgetAnimControllerHash ) : NULL;
			if ( m_xGS2SequenceHandler.GetFidgetAnimHashID() == ASURA_HASH_ID_UNSET )
			{
				if ( pxGS2FidgetAACA )
				{
					// tween out of our figet anim revelaing the main anim underneath
					u_int uFlags = pxGS2FidgetAACA->GetControllerFlags();

					Asura_AnimationBlend_Tween* pxTween = static_cast<Asura_AnimationBlend_Tween*>( pxGS2FidgetAACA->GetBlendToPreviousController() );
					if ( pxTween )
					{
						pxTween->ChangeTweenDuration( m_xGS2SequenceHandler.GetFidgetTweenInTime(), m_xGS2SequenceHandler.GetFidgetTweenInTime() );
					}

					if ( ( uFlags & AACA_FLAG_LOOPED ) != 0 )
					{
						pxGS2FidgetAACA->SetControllerFlags( uFlags & ( ~AACA_FLAG_LOOPED ) );
						pxGS2FidgetAACA->SetParametricTargetTime( pxGS2FidgetAACA->GetParametricTimer() + ( m_xGS2SequenceHandler.GetFidgetTweenInTime() / pxGS2FidgetAACA->GetDuration() ) );
					}
					else if ( pxGS2FidgetAACA->GetTimeUntilFinished() > m_xGS2SequenceHandler.GetFidgetTweenInTime() )
					{
						pxGS2FidgetAACA->SetParametricTargetTime( pxGS2FidgetAACA->GetParametricTimer() + ( m_xGS2SequenceHandler.GetFidgetTweenInTime() / pxGS2FidgetAACA->GetDuration() ) );
					}
				}
			}
			else
			{
				// play the animation associated with this fidget.
				u_int uAnimFlags = AAC_FLAG_ACTIVE | AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE | AAC_FLAG_ENSURE_UNIQUE_HASH;
				uAnimFlags |= m_xGS2SequenceHandler.GetFidgetAnimFlags();
				Asura_AnimationBlend* pxAnimBlend =
					new Asura_AnimationBlend_AnimatedTween(
						m_xGS2SequenceHandler.GetFidgetTweenInTime(),
						m_xGS2SequenceHandler.GetFidgetTweenInTime(),
						false );
				m_uCurrentGamesceneFidgetAnimControllerHash = PlayAnimation( uAnimFlags, AVP_AAPC_GAMESCENE_FIDGET_ANIM,
					m_xGS2SequenceHandler.GetFidgetAnimHashID(), pxAnimBlend, 0, 0.0f );		
			}
		}
	}
	m_bNotFirstUpdate = true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_AnimationHandler::Render
// Author       : Andy Tate
// Date         : 9-Nov-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_AnimationHandler::Render(void) const
{
}

/// End of function 'AVP_Actor_AnimationHandler::Render'. ///

#ifdef ASURA_USE_DEBUG_TEXT
void AVP_Actor_AnimationHandler::RenderDebugInfo()
{
	// Do we need to do debugging of animations?
	if((dvs_uAnimationDisplayGuid != ASURA_GUID_UNREGISTERED) && (dvs_uAnimationDisplayGuid == m_uOwnerGuid))
	{
		u_int									uColumn		= 4;
		u_int									uRow		= 4;

		// Display the hierarchy...
		m_pxAnimation->PrintDebugInfo(uColumn, uRow, true);
	}
}
#endif

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_AnimationHandler::GetCameraData
// Author       : Andy Tate
// Date         : 18-Oct-2006
////////////////////////////////////////////////////////////////////////////////

bool AVP_Actor_AnimationHandler::GetCameraData(Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientation) const
{
	const AVP_ClientEntity_Actor*	pxClient			= AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);
	const Asura_Vector_3&			xClientPosition		= pxClient->GetCurrentPosition();
	const Asura_Matrix_3x3&			xClientOrientation	= pxClient->GetCurrentOrientation();
	const u_int						uFlags				= ASURA_HIERARCHY_RENDER_NO_FRUSTUM_TEST | ASURA_HIERARCHY_RENDER_NO_MODULELINK_TEST | ASURA_HIERARCHY_RENDER_DONTRENDER | ASURA_HIERARCHY_RENDER_UPDATEBONESALWAYS;

	// Ensure that the animation has been updated this frame...
	m_pxAnimation->Update(xClientPosition, xClientOrientation, ANIMTIME_USEGAMETIME, uFlags);

	// Get the camera position from the marker-point...
	if(pxClient->GetWorldSpaceMarkerPoint(g_uCameraMarkerPointHash, &xPosition, &xOrientation))
	{
		return(true);
	}
	else
	{
		Asura_Assert(false, "Cannot find marker-point to use for camera.");
		return(false);
	}
}

/// End of function 'AVP_Actor_AnimationHandler::GetCameraData'. ///


DEBUGVAR float dbg_fDeathAnimTweenDuration = DEFAULT_TWEEN_TIME;

bool AVP_Actor_AnimationHandler::PlayDeathAnimation(Asura_Hash_ID uDeathAnimID)
{
	if (uDeathAnimID == ASURA_HASH_ID_UNSET) { return true; } // This means that we didn't actually want to play a death anim in the first place. We must be in a move that's going to control our death.
	if (!m_pxAnimation) { return false; }

	Asura_AnimationBlend_AnimatedTween* pxBlend = new Asura_AnimationBlend_AnimatedTween(dbg_fDeathAnimTweenDuration, false);
	if (!pxBlend) { return false; }

	const u_int uFlags = AAC_FLAG_ACTIVE | AAC_FLAG_ENSURE_UNIQUE_HASH | AACA_FLAG_CALCULATE_VELOCITY | AACA_FLAG_CALCULATE_ANGULAR_VELOCITY | AACA_FLAG_CALCULATE_MOVEMENT_OFFSET;
	const u_int uPriority = AVP_AAPC_DEATH_ANIM;

	Asura_AnimationController_Animation* pxDeathAnim = m_pxAnimation->CreateControllerAnimation(uDeathAnimID, uFlags, uPriority, uDeathAnimID, pxBlend);
	if (pxDeathAnim)
	{
		pxDeathAnim->Play();
		m_uDeathAnimController = pxDeathAnim->GetControllerHash();
		if ( pxDeathAnim->GetAnimation() )
		{
			AVP_ClientEntity_Actor* pxOwner = AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);
			if (pxOwner)
			{
				pxOwner->FlagMovementSystemForDestruction();
			}

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
			Asura_FacialAnimation_AnimationHandler*	pxFacialAnimationHandler = Asura_FacialAnimation_System::FindAnimationHandler(m_uOwnerGuid);

			// Play the facial-animation (if there is one)...
			if (pxFacialAnimationHandler)
			{
				Asura_FacialAnimation_FaceAnimation* pxFaceAnimation = Asura_FacialAnimation_System::FindFaceAnimation(uDeathAnimID);

				if (pxFaceAnimation)
				{
					pxFacialAnimationHandler->PlayFaceAnimation(uDeathAnimID);
				}
			}
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
			return true;
		}
	}
	return false;
}

bool AVP_Actor_AnimationHandler::OnWeaponExtending( void )
{
	if ( ASURA_HASH_ID_UNSET != m_uCurrentExtendWeaponAnimHashID )
	{
		Asura_AnimationController_Animation* pxExtendAACA = m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentExtendWeaponAnimHashID );
		if ( pxExtendAACA )
		{
			return(pxExtendAACA->Play());
		}
		else
		{
			Asura_Assert(false, "Unable to find current extend anim" );
			return(false);
		}
	}
	else
	{
		return(false);
	}
}

bool AVP_Actor_AnimationHandler::OnWeaponTweening(const Asura_Hash_ID uAnimHash, const bool bInstant, const bool bCalculateMovement)
{
	if(ASURA_HASH_ID_UNSET == uAnimHash)
	{
		m_uCurrentExtendWeaponAnimHashID = ASURA_HASH_ID_UNSET;
		if ( ASURA_HASH_ID_UNSET != m_uCurrentRetractWeaponAnimHashID )
		{
			if ( m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentRetractWeaponAnimHashID ) )
			{
				m_pxAnimation->DestroyController( m_uCurrentRetractWeaponAnimHashID );
			}
			m_uCurrentRetractWeaponAnimHashID = ASURA_HASH_ID_UNSET;
		}
	}
	else
	{
		Asura_AnimationController_Group* pxRootWeaponGroup = m_pxAnimation->GetControllerGroupFromHash( m_uRootWeaponGroupHash );
		Asura_Assert( pxRootWeaponGroup, "Unable to get weapon group" );
		if ( pxRootWeaponGroup != NULL )
		{
			u_int								uAnimFlags		= AAC_FLAG_ACTIVE | AAC_FLAG_DESTROY_WHEN_FINISHED | AAC_FLAG_ENSURE_UNIQUE_HASH;
			if ( bCalculateMovement )
			{
				uAnimFlags |= ( AACA_FLAG_CALCULATE_MOVEMENT_OFFSET | AACA_FLAG_CALCULATE_ANGULAR_VELOCITY | AACA_FLAG_CALCULATE_VELOCITY );
			}
			const u_int								uAnimPriority	= AVP_AAPC_WEAPON_ANIM;
			Asura_AnimationBlend*					pxAnimBlend		= new Asura_AnimationBlend_Tween( 0.0f, false );
			pxAnimBlend->SetDeltasUseBlends( true );
			Asura_AnimationController_Animation*	pxAnim;

			m_uCurrentExtendWeaponAnimHashID	= PlayAnimation( uAnimFlags, uAnimPriority, uAnimHash, pxAnimBlend, pxRootWeaponGroup );
			pxAnim								= m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentExtendWeaponAnimHashID );

			Asura_Assert(pxAnim != NULL, "Unable to create extend weapon anim");
			if(pxAnim != NULL)
			{
				pxAnim->Pause();
			}
		}
	}
	return true;
}

bool AVP_Actor_AnimationHandler::OnWeaponRetracting(const Asura_Hash_ID uAnimHash, const bool bCalculateMovement)
{
	Asura_AnimationController_Group* pxRootWeaponGroup = m_pxAnimation->GetControllerGroupFromHash( m_uRootWeaponGroupHash );
	Asura_Assert( pxRootWeaponGroup, "Unable to get weapon group" );
	if ( pxRootWeaponGroup != NULL )
	{
		u_int	uAnimFlags	= AAC_FLAG_ACTIVE | AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE | AAC_FLAG_ENSURE_UNIQUE_HASH;
		if ( bCalculateMovement )
		{
			uAnimFlags |= ( AACA_FLAG_CALCULATE_MOVEMENT_OFFSET | AACA_FLAG_CALCULATE_ANGULAR_VELOCITY | AACA_FLAG_CALCULATE_VELOCITY );
		}
		const u_int	uPriority	= AVP_AAPC_WEAPON_ANIM;

		if ( ASURA_HASH_ID_UNSET != uAnimHash )
		{
			Asura_AnimationBlend*	pxBlend		= new Asura_AnimationBlend_Tween( 0.0f, false );
			pxBlend->SetDeltasUseBlends( true );

			m_uCurrentRetractWeaponAnimHashID = PlayAnimation( uAnimFlags, uPriority, uAnimHash, pxBlend, pxRootWeaponGroup );
		}
		else
		{
			m_uCurrentRetractWeaponAnimHashID = ASURA_HASH_ID_UNSET;
		}

		if ( ASURA_HASH_ID_UNSET != m_uCurrentPermanentWeaponAnimHashID )
		{
			Asura_AnimationController_Animation* pxPermAACA = m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentPermanentWeaponAnimHashID );
			m_uCurrentPermanentWeaponAnimHashID = ASURA_HASH_ID_UNSET;
			if( pxPermAACA )
			{
				pxPermAACA->Stop();
			}
			else
			{
				Asura_Assert(false, "Unable to look up permanent weapon anim" );
			}
		}
	}
	return m_uCurrentRetractWeaponAnimHashID != ASURA_HASH_ID_UNSET;
}

bool AVP_Actor_AnimationHandler::OnWeaponExtended(const Asura_Hash_ID uAnimHash)
{
	Asura_AnimationController_Group* pxRootWeaponGroup = m_pxAnimation->GetControllerGroupFromHash( m_uRootWeaponGroupHash );
	Asura_Assert( pxRootWeaponGroup, "Unable to get weapon group" );
	if ( pxRootWeaponGroup != NULL )
	{
		const u_int	uAnimFlags	= AAC_FLAG_ACTIVE | AAC_FLAG_DESTROY_WHEN_FINISHED | AAC_FLAG_ENSURE_UNIQUE_HASH | AACA_FLAG_LOOPED;
		const u_int	uPriority	= AVP_AAPC_WEAPON_ANIM;

		if( ASURA_HASH_ID_UNSET != uAnimHash )
		{
			Asura_AnimationBlend*	pxBlend		= new Asura_AnimationBlend_Tween( 0.0f, false );
			pxBlend->SetDeltasUseBlends( true );
			m_uCurrentPermanentWeaponAnimHashID = PlayAnimation( uAnimFlags, uPriority, uAnimHash, pxBlend, pxRootWeaponGroup );
			return ASURA_HASH_ID_UNSET != m_uCurrentPermanentWeaponAnimHashID;
		}
	}
	return false;
}

bool AVP_Actor_AnimationHandler::IsActivateAnimPlaying() const
{
	if ( m_uCurrentExtendWeaponAnimHashID == ASURA_HASH_ID_UNSET )
	{
		return false;
	}
	else
	{
		const Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentExtendWeaponAnimHashID );
		if ( pxAACA == NULL )
		{
			return false;
		}
		else
		{
			return pxAACA->IsPlaying();
		}
	}
}

bool AVP_Actor_AnimationHandler::IsDeactivateAnimPlaying() const
{
	if ( m_uCurrentRetractWeaponAnimHashID == ASURA_HASH_ID_UNSET )
	{
		return false;
	}
	else
	{
		const Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_uCurrentRetractWeaponAnimHashID );
		if ( pxAACA == NULL )
		{
			return false;
		}
		else
		{
			return pxAACA->IsPlaying();
		}
	}
}

void AVP_Actor_AnimationHandler::AttachItem( AVP_Inventory_Item* pxItem )
{
	Asura_Assert( pxItem != NULL, "No item given" );
	if ( pxItem != NULL && !IsItemAttached( pxItem ))
	{
		Asura_Animation* pxWeaponAnimation = pxItem->GetAnimation( true ); // creates if necessary
		if ( pxWeaponAnimation != NULL )
		{
			if ( m_pxAnimation->LinkAnimation( pxWeaponAnimation, pxItem->GetLinkMarkerPoint(), false ) )
			{
				pxWeaponAnimation->SetParent( m_pxAnimation );
				pxWeaponAnimation->SetParentEntity( m_uOwnerGuid );

				pxItem->OnAttachItem( pxWeaponAnimation ); // Only call if linking was successfull!
			}
		}
		pxItem->InitialiseAnimation();
	}
}

void AVP_Actor_AnimationHandler::DetachItem( AVP_Inventory_Item* pxItem )
{
	Asura_Assert( pxItem != NULL, "No item given" );
	if ( pxItem != NULL )
	{
		Asura_Animation* pxWeaponAnimation = pxItem->GetAnimation();
		if ( pxWeaponAnimation != NULL )
		{
			pxItem->OnDetachItem();
			m_pxAnimation->UnlinkAnimation( pxWeaponAnimation );
		}
	}
}

bool AVP_Actor_AnimationHandler::IsItemAttached( const AVP_Inventory_Item* pxItem ) const
{
	Asura_Assert( pxItem != NULL, "No item given" );
	if ( pxItem != NULL )
	{
		const Asura_Animation* pxWeaponAnimation = pxItem->GetAnimation();
		if ( pxWeaponAnimation != NULL )
		{
			const Asura_AnimationAttachment* pxAttItr = m_pxAnimation->GetFirstAttachment();
			while ( pxAttItr != NULL )
			{
				if ( pxAttItr->GetAttachedAnimation() == pxWeaponAnimation )
				{
					return true;
				}
				pxAttItr = pxAttItr->GetNextAttachment();
			}
		}
	}

	return false;
}

Asura_Hash_ID AVP_Actor_AnimationHandler::PlayWeaponAnimation( const Asura_Hash_ID uAnimHash, const float fTweenTime, const bool bLooped,  u_int uFlags )
{
	Asura_AnimationController_Group* pxRootWeaponGroup = m_pxAnimation->GetControllerGroupFromHash( m_uRootWeaponGroupHash );
	Asura_Assert( pxRootWeaponGroup, "Unable to get weapon group" );
	if ( pxRootWeaponGroup != NULL )
	{
		Asura_AnimationBlend*	pxBlend	= new Asura_AnimationBlend_AnimatedTween( fTweenTime, fTweenTime, false );
		pxBlend->SetDeltasUseBlends( true );
		
		if ( bLooped )
		{
			uFlags |= AACA_FLAG_LOOPED;
		}

		u_int uPriority = AVP_AAPC_WEAPON_ANIM;

		return PlayAnimation( uFlags, uPriority, uAnimHash, pxBlend, pxRootWeaponGroup );
	}
	//else
	return ASURA_HASH_ID_UNSET;
}

void AVP_Actor_AnimationHandler::StopWeaponAnimation( const Asura_Hash_ID uAnimControllerHash, const bool bDestroy )
{
	Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( uAnimControllerHash );
	if ( pxAACA != NULL )
	{
		if ( bDestroy )
		{
			pxAACA->SetControllerFlags( pxAACA->GetControllerFlags() | AAC_FLAG_DESTROY_WHEN_FINISHED );
		}
		pxAACA->Stop();
	}
}

bool AVP_Actor_AnimationHandler::IsWeaponAnimationPlaying( const Asura_Hash_ID uAnimControllerHash ) const
{
	Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( uAnimControllerHash );
	if ( pxAACA != NULL )
	{
		return !pxAACA->IsFinished();
	}
	else
	{
		return false;
	}
}

AVP_GS2_AnimMappingData* AVP_Actor_AnimationHandler::GetAnimMappingData()
{
	// Fill in the data and return it.
	s_xAnimMappingData.m_auFactorValues[ AVP_GS2_MF_SKIN ] = m_xOwnerActor.GetShapeInstance().GetShapeOrSkinHashID();
	const AVP_Inventory_Item* pxCurrentItem = m_xOwnerActor.GetInventorySystem()->GetCurrentItem( uAVP_INVENTORY_GROUPHASH_WEAPON );
	if ( pxCurrentItem )
	{
		switch ( pxCurrentItem->GetItemType() )
		{
			case AVP_INVENTORYITEMTYPE_WEAPON_FLAMETHROWER:
			{
				s_xAnimMappingData.m_auFactorValues[ AVP_GS2_MF_WEAPON ] = AVP_WMF_FLAMETHROWER;
				break;
			}

			case AVP_INVENTORYITEMTYPE_SHOTGUN:
			{
				s_xAnimMappingData.m_auFactorValues[ AVP_GS2_MF_WEAPON ] = AVP_WMF_SHOTGUN;
				break;
			}

			case AVP_INVENTORYITEMTYPE_SMARTGUN:
			{
				s_xAnimMappingData.m_auFactorValues[ AVP_GS2_MF_WEAPON ] = AVP_WMF_SMARTGUN;
				break;
			}

			default:
			{
				s_xAnimMappingData.m_auFactorValues[ AVP_GS2_MF_WEAPON ] = AVP_WMF_X;
				break;
			}
		}
	}
	else
	{
		s_xAnimMappingData.m_auFactorValues[ AVP_GS2_MF_WEAPON ] = AVP_WMF_X;
	}
	return &s_xAnimMappingData;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_AnimationHandler::CreateAnimation
// Author       : Andy Tate
// Date         : 28-Jul-2008
////////////////////////////////////////////////////////////////////////////////

const Asura_Hierarchy_Skin* AVP_Actor_AnimationHandler::CreateAnimation(void)
{
	// UpdateSkin handles creation of a new animation if we don't have one, so it should be fine to reuse that.
	UpdateSkin();
	if ( m_pxAnimation )
	{
		// Return the skin...
		return(m_pxAnimation->GetSkin());
	}
	return NULL;
}

void AVP_Actor_AnimationHandler::SetFirstPerson( void )
{
	if ( !m_pxAnimation )
	{
		m_bFirstPerson = true;
		UpdateSkin();
	}
	else if ( !m_bFirstPerson )
	{
		Shutdown();
		m_bFirstPerson = true;
		Initialise();
	}
}

void AVP_Actor_AnimationHandler::SetThirdPerson( void )
{
	if ( !m_pxAnimation )
	{
		m_bFirstPerson = false;
		UpdateSkin();
	}
	else if ( m_bFirstPerson )
	{
		Shutdown();
		m_bFirstPerson = false;
		Initialise();
	}
}
/// End of function 'AVP_Actor_AnimationHandler::CreateAnimation'. ///


/// End of file. ///
