
///////////////////////////////////////////////////////////////////////////////
//
// File		: AVP_Actor_CameraHandler.cpp
// Author	: Paul Humphries
// Date		: 18-May-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

// Core Asura
#include "Core/Asura.h"

// This file's header...
#include "Actor/AVP_Actor_CameraHandler.h"

// Asura platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Camera/Asura_Camera.h"
#include "Camera/Asura_FlyCam.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_DebugLineRender.h"
#include "Interface/Asura_CommandConsole.h"
#include "Physics/Asura_Physics_MovementQuery.h"
#include "Rendering/Asura_Render.h"
#include "Debugging/Asura_DebugText.h"
#include "Rendering/Asura_Text_Output.h"

// Game platform-independent
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_VelocityInterceptor.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Core.h"
#include "Core/AVP_Maths.h"
#include "Core/AVP_Settings.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Mesh/AVP_PolygonFlags.h"

// Constants
DEBUGVAR Asura_Euler_3 dbg_xCameraEuler90Left(0.0f, Asura_Maths::Deg90, 0.0f);
DEBUGVAR float dbg_fMinFov = Asura_Maths::DegToRad(10.0f);

bool AVP_Actor_CameraHandler::s_bFlyCam = false;
bool AVP_Actor_CameraHandler::s_bFixCam = false;

//#define DRAWLINES 1
//#define DRAWTRAILS 1

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	int AVP_Actor_CameraHandler::s_iConsoleVarsCount	= 0;
#endif	// defined(ASURA_SUPPORT_COMMAND_CONSOLE)


AVP_Actor_CameraMovement::AVP_Actor_CameraMovement(void)
: m_xRelativeOffset(Asura_Maths::ZeroVector3)
, m_xRelativeEuler(Asura_Maths::ZeroEuler3)
{
}

AVP_Actor_CameraMovement::AVP_Actor_CameraMovement(const AVP_Actor_CameraMove& xMove)
: m_xRelativeOffset(Asura_Maths::ZeroVector3)
, m_xRelativeEuler(Asura_Maths::ZeroEuler3)
{
	m_xRelativeOffset.Reset(xMove.m_xOffset);
	m_xRelativeEuler.Reset(xMove.m_xEuler);
}

// Camera handler constructor, requires the owner guid

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
AVP_Actor_CameraHandler::AVP_Actor_CameraHandler (const Asura_Guid uOwnerGuid, AVP_Actor_CameraProperties& xCameraProperties)
#else
AVP_Actor_CameraHandler::AVP_Actor_CameraHandler (const Asura_Guid uOwnerGuid, const AVP_Actor_CameraProperties& xCameraProperties)
#endif
: m_uOwnerGuid(uOwnerGuid)
, m_eCameraType(CT_FIRST_PERSON)
, m_fCameraFieldOfViewFirstPerson(85.0f)
, m_xCameraOrientation( Asura_Maths::IdentityMatrix3x3 )
, m_xCameraOrientationThirdToFirst( Asura_Maths::IdentityMatrix3x3 )
, m_xCameraPosition( Asura_Maths::ZeroVector3 )
, m_uZoomLevel(0)
, m_fWallDuration(0.2f)
, m_fWallExp(22.0f)
, m_xCameraProperties(xCameraProperties)
, m_fZoomSpeed(1.0f)
, m_uFlags(0)
, m_xFirstToThirdFactor(0.0f)
, m_xInsideOutside(1.0f, 0.0f)
, m_fCameraRotationSmoothed(0.0f)
{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	if(s_iConsoleVarsCount==0)
	{
#if defined(ASURA_SUPPORT_FLYCAM)
			Asura_CommandConsole::AddVar(L"AVP.FlyCam", &s_bFlyCam, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
			Asura_CommandConsole::AddVar(L"AVP.FixCam", &s_bFixCam, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif	// defined(ASURA_SUPPORT_FLYCAM)

		// DVS variables needed for adjusting camera properties
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Inside.Distance", &(m_xCameraProperties.m_fDistanceFollowInside), 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Inside.Follow Speed", &(m_xCameraProperties.m_fExpFollowInside), 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Outside.Distance", &(m_xCameraProperties.m_fDistanceFollowOutside), 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Outside.Follow Speed", &(m_xCameraProperties.m_fExpFollowOutside), 0.0f, FLT_MAX, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		
		// The others will be useful later!
#if 0
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Duration To", &(m_xCameraProperties.m_fDurationToThird));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Field Of View", &(m_xCameraProperties.m_fFovThirdPerson));

		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Height Smoothing.Follow Character", &(m_xCameraProperties.m_fExpCharacterHeight));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Height Smoothing.Follow Camera", &(m_xCameraProperties.m_fExpCameraHeight));

		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Offset.Amount", &(m_xCameraProperties.m_fDistanceOffset));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Offset.Reduce Below Distance", &(m_xCameraProperties.m_fDistanceReduceOffset));

		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Inside.Duration To", &(m_xCameraProperties.m_fDurationOutsideToInside));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person. Outside.Duration To", &(m_xCameraProperties.m_fDurationInsideToOutside));

		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Rotation Lock.Angular Velocity", &(m_xCameraProperties.m_fRotationMaxAngularVelocity));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Rotation Lock.Half Velocity Duration", &(m_xCameraProperties.m_fRotationHalfLife));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Rotation Lock.Lock Speed", &(m_xCameraProperties.m_fExpLock));

		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Around Walls.Half Player Width", &(m_xCameraProperties.m_fDistanceHalfPlayerWidth));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Around Walls.Extra Factor", &(m_xCameraProperties.m_fFactorClearCorner));

		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Close.Move Back Distance", &(m_xCameraProperties.m_fDistanceMoveBack));
		Asura_CommandConsole::AddVar(L"AVP.Camera.3rd Person.Close.Fade Player Distance", &(m_xCameraProperties.m_fDistanceFadeOutActor));

		Asura_CommandConsole::AddVar(L"AVP.Camera.1st Person. Duration To", &(m_xCameraProperties.m_fDurationToFirst));
		Asura_CommandConsole::AddVar(L"AVP.Camera.1st Person.Character Turn Speed", &(m_xCameraProperties.m_fExpTurnToFirst));
		Asura_CommandConsole::AddVar(L"AVP.Camera.1st Person. Field Of View 1", &(m_xCameraProperties.m_afFovZoom[0]));
		Asura_CommandConsole::AddVar(L"AVP.Camera.1st Person. Field Of View 2", &(m_xCameraProperties.m_afFovZoom[1]));
		Asura_CommandConsole::AddVar(L"AVP.Camera.1st Person. Field Of View 3", &(m_xCameraProperties.m_afFovZoom[2]));
		Asura_CommandConsole::AddVar(L"AVP.Camera.1st Person. Field Of View 4", &(m_xCameraProperties.m_afFovZoom[3]));
#endif
		
	}
	s_iConsoleVarsCount++;
#endif	// defined(ASURA_SUPPORT_COMMAND_CONSOLE)

	//Reset(); //No longer called in constructor, called afterwards while in MakePlayer()
}

AVP_Actor_CameraHandler::~AVP_Actor_CameraHandler()
{
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	s_iConsoleVarsCount--;
	if( s_iConsoleVarsCount == 0 )
	{
#if defined(ASURA_SUPPORT_FLYCAM)
			Asura_CommandConsole::RemoveVar(L"AVP.FlyCam" );
			Asura_CommandConsole::RemoveVar(L"AVP.FixCam" );
#endif	// defined(ASURA_SUPPORT_FLYCAM)

		// DVS variables needed for adjusting camera properties
		Asura_CommandConsole::RemoveVar(L"AVP.Camera.3rd Person. Inside.Distance" );
		Asura_CommandConsole::RemoveVar(L"AVP.Camera.3rd Person. Inside.Follow Speed" );
		Asura_CommandConsole::RemoveVar(L"AVP.Camera.3rd Person. Outside.Distance" );
		Asura_CommandConsole::RemoveVar(L"AVP.Camera.3rd Person. Outside.Follow Speed" );
	}
#endif
}

void AVP_Actor_CameraHandler::Reset(void)
{
	m_uFlags		= 0;
	m_uZoomLevel	= 0;
	for ( u_int uFovModItr = 0; uFovModItr < MAX_FOVMOD; ++uFovModItr )
	{
		m_afFOVModifier[ uFovModItr ] = 0.0f;
	}

	AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);

	m_eCameraType			= CT_FIRST_PERSON;
	m_xFirstToThirdFactor.Reset(0.0f);
// 	SetInsideCamera(false, true);	// set outside camera (instantly)

	// set current first person zoom
	RefreshBlueprintData();
	m_fCameraFieldOfViewFirstPerson	= m_afFovZoom[0];

	// m_xCameraMovement will probably be used for the grapple camera..
	//m_xCameraMovement.Reset(m_xCameraProperties.m_xCameraOutside);

	// set the current camera distance from the inside/outside velocity interceptor
	m_xInsideOutside.CalculatePosition(m_fCameraDistance);

	// Position the camera behind the player
	Asura_Vector_3 xPlayerPosition = pxClient->GetCurrentPosition();
	m_xCameraOrientation = pxClient->GetCurrentOrientation();
	m_xCameraPosition = xPlayerPosition + (m_xCameraProperties.m_xOffsetHead + Asura_Vector_3(m_xCameraProperties.m_fDistanceOffset, 0.0f, -m_fCameraDistance)) * m_xCameraOrientation;

	m_xFirstPersonPosition.Zero();
	m_bRemovedClientFromRenderableList = false;

	// Initialise smoothed position and rotation speed
	m_xPlayerPositionSmoothed = xPlayerPosition;
	m_fCameraRotation = 0.0f;

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
	// Initialise distances for checking when changed via DVS
	m_fDistanceFollowInsideLast = m_xCameraProperties.m_fDistanceFollowInside;
	m_fDistanceFollowOutsideLast = m_xCameraProperties.m_fDistanceFollowOutside;
#endif
}

// only set the asura camera in this method - nowhere else!!!

void AVP_Actor_CameraHandler::SetAsuraCamera(AVP_ClientEntity_Actor& xOwner)
{
#if defined(ASURA_SUPPORT_FLYCAM)
	if ( s_bFlyCam )
	{
		Asura_FlyCam::SetInvertY(!AVP_Settings::GetCurrentProfile()->GetControls().GetInvertY());
		m_fAlpha = 1.0f;
		Asura_FlyCam::Update();
	}
	else if ( !s_bFixCam )
#endif	// defined(ASURA_SUPPORT_FLYCAM)
	{
		if (m_eCameraType == CT_EXTERNAL)
		{
			m_fAlpha = 1.0f;
		}
		else
		{
			Asura_Vector_3 xActorPosition, xActorHead;
			Asura_Matrix_3x3 xActorOrientation, xActorTurn;

			Asura_Vector_3 xCameraPosition;
			Asura_Matrix_3x3 xCameraOrientation;
			float fCameraFieldOfView;

			xActorOrientation = xOwner.GetCurrentOrientation();
			xActorPosition = xOwner.GetCurrentPosition();

			// get the player's turn orientation matrix (without the tilt)
			AVP_Maths::ExtractTurnFrom(xActorOrientation, xActorTurn);

			xActorHead = xActorPosition + m_xCameraProperties.m_xOffsetHead * xActorTurn;

			float fFactor;
			m_xFirstToThirdFactor.CalculatePosition(fFactor);

			// if we are using first person camera OR we are changing from first to third,
			// get first person position and orientation from the animation handler

			if (m_eCameraType == CT_FIRST_PERSON || m_xFirstToThirdFactor.GetComplete() < 1.0f)
			{
				AVP_Actor_AnimationHandler* pxAnimationHandler = xOwner.GetAnimationHandler();
				if (!pxAnimationHandler)
				{
					Asura_Assert(false, "Unable to get animation handler");
					return;
				}

				// get the first person camera position and orientation from the animation handler
				if (!pxAnimationHandler->GetCameraData(xCameraPosition, xCameraOrientation))
				{
					// Some animation problem, break out, hopefully this isn't a lasting problem.
					return;
				}

				// if not complete, we need to lerp i.e. merge in third person values
				if (m_xFirstToThirdFactor.GetComplete() < 1.0f)
				{
					// update the time
					m_xFirstToThirdFactor.AddTime(Asura_Timers::GetGameTimeSlice());

					// lerp field of view between first and third person
					fCameraFieldOfView = Asura_Maths::Lerp(m_fCameraFieldOfViewFirstPerson, m_xCameraProperties.m_fFovThirdPerson, fFactor);

					// lerp between first person and third person camera positions
					xCameraPosition *= (1.0f - fFactor);
					xCameraPosition += m_xCameraPosition * fFactor;

					// lerp between first person and third person orientations
					Asura_Quat xQuat, xQuat1, xQuat2;
					xQuat1.FromMatrix(xCameraOrientation);
					xQuat2.FromMatrix(m_xCameraOrientation);

					xQuat.NLerp(xQuat1, xQuat2, fFactor);
					xQuat.ToMatrix(xCameraOrientation);
				}
				else
				{
					fCameraFieldOfView = m_fCameraFieldOfViewFirstPerson;
				}
			}
			else
				// just third person camera
			{
				xCameraPosition = m_xCameraPosition;
				xCameraOrientation = m_xCameraOrientation;
				fCameraFieldOfView = m_xCameraProperties.m_fFovThirdPerson;
			}

			// finally, actually set the Asura camera
			UpdateCameraOrientation( xOwner, xCameraOrientation );

			float fFOV = fCameraFieldOfView;
			for ( u_int uFovModItr = 0; uFovModItr < MAX_FOVMOD; ++uFovModItr )
			{
				fFOV -= m_afFOVModifier[ uFovModItr ];
			}
			if ( fFOV < dbg_fMinFov )
			{
				fFOV = dbg_fMinFov;
			}

			Asura_Camera::SetFieldOfView(fFOV);
			Asura_Camera::SetPosition(xCameraPosition);
			Asura_Camera::SetOrientation(xCameraOrientation);

			// store the actual camera variables so we can access them when required
			m_xCameraActualPosition = xCameraPosition;
			m_xCameraActualOrientation = xCameraOrientation;

			// and set the alpha value for the actor (fades the actor out when the camera is close)
			Asura_Vector_3 xActorToCamera = xCameraPosition - xActorHead;
			float fCameraDistance = xActorToCamera.Magnitude();
			if (fCameraDistance < m_xCameraProperties.m_fDistanceFadeOutActor)
			{
				float fAlpha = fCameraDistance / m_xCameraProperties.m_fDistanceFadeOutActor;
				m_fAlpha = fAlpha * fAlpha;
			}
			else
			{
				m_fAlpha = 1.0f;
			}
		}
	}
}

// update the third person camera position and orientation - must be called once per frame

void AVP_Actor_CameraHandler::Update(AVP_ClientEntity_Actor& xOwner)
{
#if defined(ASURA_SUPPORT_FLYCAM)
	if ( s_bFlyCam || s_bFixCam ) return;
#endif	// defined(ASURA_SUPPORT_FLYCAM)

	// don't need to update if third person camera is not required
	if (m_eCameraType != CT_THIRD_PERSON && !m_xFirstToThirdFactor.IsMoving()) return;

	// reference to camera properties
	AVP_Actor_CameraProperties& xC = m_xCameraProperties;

	float fTimeSlice = Asura_Timers::GetGameTimeSlice();

	Asura_Matrix_3x3 xPlayerOrientation = xOwner.GetCurrentOrientation();
	Asura_Vector_3 xPlayerPosition = xOwner.GetCurrentPosition();
	Asura_Vector_3 xPlayerVelocity;
	Asura_Matrix_3x3 xPlayerTurn;

	// get the player's turn orientation matrix (without the tilt)
	AVP_Maths::ExtractTurnFrom(xPlayerOrientation, xPlayerTurn);

	Asura_Vector_3 xOffset = Asura_Vector_3(1.0f, 0.0f, 0.0f) * m_xCameraOrientation * xC.m_fDistanceOffset;

	xPlayerPosition += xOffset;	// use offset position for following player

	xOwner.GetCurrentVelocity(xPlayerVelocity);

	// smooth out player height
	m_xPlayerPositionSmoothed += (xPlayerPosition - m_xPlayerPositionSmoothed) * fTimeSlice * xC.m_fExpCharacterHeight;
	xPlayerPosition[1] = m_xPlayerPositionSmoothed[1];

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) 
// 	if (m_bInside) 
// 	{
// // 		if (m_fDistanceFollowInsideLast != xC.m_fDistanceFollowInside)
// // 		{
// // 			SetInsideCamera(true);
// // 		}
// 	}
// 	else
// 	{
// // 		if (m_fDistanceFollowOutsideLast != xC.m_fDistanceFollowOutside)
// // 		{
// // 			SetInsideCamera(false);
// // 		}
// 	}
// 	m_fDistanceFollowInsideLast = xC.m_fDistanceFollowInside;
// 	m_fDistanceFollowOutsideLast = xC.m_fDistanceFollowOutside;
#endif

	Asura_Vector_3 xPlayerHead = xPlayerPosition + xC.m_xOffsetHead * xPlayerTurn;

	// get the smoothed camera distance from the interceptor
	m_xInsideOutside.CalculatePosition(m_fCameraDistance);
	m_xInsideOutside.AddTime(fTimeSlice);

	// obtain vector from player to camera
	Asura_Vector_3 xFromPlayer = m_xCameraPosition - xPlayerHead;

	// lock camera behind player
	if (m_bLocked && xFromPlayer.Magnitude() > 0.01f)
	{
		Asura_Vector_3 xTargetPosition = xPlayerHead + Asura_Vector_3(0.0f, 0.0f, -m_fCameraDistance) * xPlayerOrientation;
		xFromPlayer += (xTargetPosition - m_xCameraPosition) * fTimeSlice * xC.m_fExpLock;
	}

	// get the vector from the player in the XZ plane
	Asura_Vector_3 xFromPlayerXZ = xFromPlayer;
	xFromPlayerXZ[1] = 0.0f;

	// move camera height towards player height
	AVP_Maths::ExponentiallyTowards(xFromPlayer[1], 0.0f, xC.m_fExpCameraHeight);

	// set exponential follow factor for inside/outside
	float fExpFollow	= xC.m_fExpFollowOutside;

	// if XZ distance greater than current desired distance, move camera to catch up with player
	if (xFromPlayerXZ.Magnitude() > m_fCameraDistance)
	{
		xFromPlayer[0] += xFromPlayer[0] * (m_fCameraDistance - xFromPlayer.Magnitude()) * fTimeSlice * fExpFollow;
		xFromPlayer[2] += xFromPlayer[2] * (m_fCameraDistance - xFromPlayer.Magnitude()) * fTimeSlice * fExpFollow;
	}

	xFromPlayer += xOffset;
	xPlayerHead -= xOffset;

	xFromPlayerXZ = xFromPlayer;
	xFromPlayerXZ[1] = 0.0f;

	// otherwise, if less than current desired minimum distance and not locked, fix at this distance
	if (xFromPlayerXZ.Magnitude() > 0.01f && xFromPlayerXZ.Magnitude() < m_fCameraDistance - 0.5f && !m_bLocked)
	{
		xFromPlayer[0] *= (m_fCameraDistance - 0.5f) / xFromPlayerXZ.Magnitude();
		xFromPlayer[2] *= (m_fCameraDistance - 0.5f) / xFromPlayerXZ.Magnitude();
	}

	// rotate the camera around the player
	float fGamma = log(2.0f) / xC.m_fRotationHalfLife;
	float fAngularAcceleration = xC.m_fRotationMaxAngularVelocity * fGamma;
	Asura_Vector_3 xTangent = xFromPlayer ^ Asura_Vector_3(0.0f, -1.0f, 0.0f);
	xFromPlayer += xTangent * fTimeSlice * m_fCameraRotationSmoothed;
	m_fCameraRotationSmoothed += fAngularAcceleration * m_fCameraRotation * fTimeSlice;
	AVP_Maths::ExponentiallyTowards(m_fCameraRotationSmoothed, 0.0f, fGamma);

	// perform a movement query test from the head to the camera
	float fSize = xFromPlayer.Magnitude();
	if (fSize > 0.01f)
	{
		Asura_Vector_3 xWallCamera, xWallPlayer, xNormalCamera, xNormalPlayer, xNormalPlayer2;

		Asura_Vector_3 xFromPlayerNormalised = xFromPlayer;
		xFromPlayerNormalised.Normalise();

		Asura_Vector_3 xShift = xFromPlayer ^ Asura_Vector_3(0.0f, 1.0f, 0.0f);
		xShift.Normalise();
		xShift *= xC.m_fDistanceHalfPlayerWidth;		// how far perpendicularly left and right of xActorHead for the line of sight tests

		// fDistancePlayer is the distance from the player to a wall (in the direction of the camera)
		// fDistanceCamera is the distance from the camera to a wall (in the direction of the player)

		float fDistancePlayer = LineOfSight(xPlayerHead + xShift, xFromPlayer, fSize, xNormalPlayer);
		float fDistancePlayer2 = LineOfSight(xPlayerHead - xShift, xFromPlayer, fSize, xNormalPlayer2);

		float fDistanceCamera;
		if (fDistancePlayer < fSize - 0.01f || fDistancePlayer2 < fSize - 0.01f)
		{
			// if second test closer to player, use these values (and negate the shift to reflect this)
			if (fDistancePlayer2 < fDistancePlayer)
			{
				fDistancePlayer = fDistancePlayer2;
				xNormalPlayer = xNormalPlayer2;
				xShift = -xShift;
			}

			fDistanceCamera = LineOfSight(m_xCameraPosition + xShift, -xFromPlayer, fSize, xNormalCamera);				

			// xWallCamera is the point on a wall nearest the camera
			// xWallPlayer is the point on a wall nearest the player

			xWallCamera = m_xCameraPosition + xShift - xFromPlayerNormalised * fDistanceCamera;
			xWallPlayer = xPlayerHead + xShift + xFromPlayerNormalised * fDistancePlayer;

			//AVP_Maths::CreateArrow(m_xCameraPosition, xPlayerHead, 0xff777777);
			//AVP_Maths::CreateArrow(m_xCameraPosition + xShift, xPlayerHead + xShift, 0xffaaffaa);

			Asura_Vector_3 xPreviousOffset = xFromPlayer;

			// dot the normals together - we want to check whether they are 90 degrees to each other..
			float fDot = xNormalCamera * xNormalPlayer;

			// check if this is a corner we want to try and move around:
			// * camera greater than move-back distance from player
			// * player moving around the corner i.e. in direction xShift
			// * normals are approxmiately 90 degrees from each other

			if (m_fCameraDistance > xC.m_fDistanceMoveBack && // m_fCameraRotation == 0.0f &&
				xPlayerVelocity * xShift > 0.0f && Asura_Maths::Abs(fDot) < 0.005f
				&& xNormalCamera.Magnitude() > 0.01f && xNormalPlayer.Magnitude() > 0.01f)
			{
				//AVP_Maths::CreateArrow(m_xCameraPosition, xPlayerHead, 0xffff7777);
#ifdef DRAWLINES
				AVP_Maths::CreateArrow(xWallCamera, xWallCamera + xNormalCamera*0.3f, 0xffffffff);
				AVP_Maths::CreateArrow(xWallPlayer, xWallPlayer + xNormalPlayer*0.3f, 0xffffffff);
#endif
				// fLamba is the parameter for a line starting at xWallCamera along the wall.
				// The value at the corner is calculated by saying that the vector from xWallPlayer to a point
				// on the line is perpendicular to xNormalPlayer at the corner.

				float fLamba = ( (xWallPlayer[2] - xWallCamera[2]) * xNormalPlayer[2]
								+ (xWallPlayer[0] - xWallCamera[0]) * xNormalPlayer[0] )
								/ (xNormalPlayer[2] * xNormalPlayer[2] + xNormalPlayer[0] * xNormalPlayer[0]);

				// now we can calculate where the corner is..

				Asura_Vector_3 xCorner;
				xCorner[0] = xWallPlayer[0] + fLamba * xNormalPlayer[0];
				xCorner[2] = xWallPlayer[2] + fLamba * xNormalPlayer[2];
				xCorner[1] = -1.5f;	// any sensible value (shouldn't matter)

				//AVP_Maths::CreateArrow(xCorner, xCorner + Asura_Vector_3(0.0f, -0.2f, 0.0f), 0xff7777ff);

				// work out the displacement we need to apply to the camera

				Asura_Vector_3 xMoveCamera = xCorner - xPlayerHead - xShift;	// vector from player to corner
				if (xMoveCamera.Magnitude() > 0.01f)
				{
					xMoveCamera[1] = 0.0f;					// we are working in the XZ plane
					xMoveCamera.Normalise();
					xMoveCamera *= xFromPlayer.Magnitude();	// same size as current distance from player to camera
					xMoveCamera += xPlayerHead;				// we now have the position we want the camera
					xMoveCamera -= m_xCameraPosition;		// the displacement we want to apply

					// move slightly more to ensure we clear the corner
					xFromPlayer += xMoveCamera * xC.m_fFactorClearCorner;
#ifdef DRAWLINES
					AVP_Maths::CreateArrow(xPlayerHead + xFromPlayer, xPlayerHead, 0x88888888);
					AVP_Maths::CreateArrow(xPlayerHead + xPreviousOffset, xPlayerHead + xFromPlayer, 0xffffff77);
#endif
				}

				// perform a final test to ensure we are clear and if not, give up and pull the camera in
				fDistancePlayer = LineOfSight(xPlayerHead, xFromPlayer, fSize, xNormalPlayer);
				if (fDistancePlayer < fSize)
				{
					xFromPlayer *= fDistancePlayer / fSize;
#ifdef DRAWLINES
					AVP_Maths::CreateArrow(xPlayerHead + xPreviousOffset, xPlayerHead + xFromPlayer, 0xffff9999);
#endif
				}
			}
			else
			// not a corner we want to try and move around, so pull in the camera
			{
				xFromPlayer *= fDistancePlayer / fSize;
#ifdef DRAWLINES
				AVP_Maths::CreateArrow(xPlayerHead + xPreviousOffset, xPlayerHead + xFromPlayer, 0xff00ff00);
#endif
			}
		}
	}

	// set the camera position
	m_xCameraPosition = xPlayerHead + xFromPlayer;

	// subtract the offset from the vector from the player, but reduce when close
	float fCloseFactor = Asura_Maths::Min(xFromPlayer.Magnitude(), xC.m_fDistanceReduceOffset) / xC.m_fDistanceReduceOffset;
	xFromPlayer -= xOffset * fCloseFactor * fCloseFactor;

	// always point camera directly at the actor (well, the actor's smoothed-height position)
	Asura_Maths::OrientationFromDirection(m_xCameraOrientation, -xFromPlayer);

#ifdef DRAWTRAILS
	// for debugging - leave a trail of camera positions if trail is switched on
	if (true || m_bTrail) {
		AVP_Maths::CreateArrow(m_xCameraPositionLast, m_xCameraPosition, uTrailColour);
		AVP_Maths::CreateArrow(m_xCameraPosition, xPlayerPosition, 0x22ffff00);
	}
#endif

}

// set the camera type

void AVP_Actor_CameraHandler::SetCameraType(const CameraType eType)
{
	AVP_ClientEntity_Actor* pxClient	= AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);
	AVP_Actor_AnimationHandler* pxAnimationHandler = pxClient ? pxClient->GetAnimationHandler() : NULL;

	// Move between third and first person cameras swiftly
	if (eType == CT_FIRST_PERSON)
	{
		m_xFirstToThirdFactor.SetTarget(0.0f, m_xCameraProperties.m_fDurationToFirst);
		m_xCameraOrientationThirdToFirst = m_xCameraOrientation;
	}
	else if (eType == CT_THIRD_PERSON)
	{
		m_xFirstToThirdFactor.SetTarget(1.0f, m_xCameraProperties.m_fDurationToThird);
	}

	m_eCameraType = eType;

	if (!pxAnimationHandler)
	{
		Asura_Assert(false, "Unable to get animation handler for player");
		return;
	}
}

// Change to inside camera view
// 
// void AVP_Actor_CameraHandler::SetInsideCamera(bool bInside, bool bInstant /* = false */)
// {
// 	m_bInside = bInside;
// 	if (bInstant)
// 	{
// 		m_xInsideOutside.Reset(bInside ? m_xCameraProperties.m_fDistanceFollowInside : m_xCameraProperties.m_fDistanceFollowOutside);
// 	}
// 	else if (!m_bGrappling)
// 	{
// 		if (bInside) 
// 		{
// 			m_xInsideOutside.SetTarget(m_xCameraProperties.m_fDistanceFollowInside, m_xCameraProperties.m_fDurationOutsideToInside);
// 		}
// 		else
// 		{
// 			m_xInsideOutside.SetTarget(m_xCameraProperties.m_fDistanceFollowOutside, m_xCameraProperties.m_fDurationInsideToOutside);
// 		}
// 	}
// }

// Call when a grapple attack is about to start - move the camera to the left or right to view the grapple attack
// 
// void AVP_Actor_CameraHandler::SetGrappleCamera(bool bGrapple)
// {
// 	m_bGrappling = true;
// 
// 	const AVP_ClientEntity_Actor* pxPlayer = AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);
// 	if (!pxPlayer)
// 	{
// 		Asura_Assert(false, "Can't find owner (player) for camera handler");
// 		return;
// 	}
// 
// 	bool bLeft = IsMoreSpacePerpendicularLeft(*pxPlayer);
// 
// 	if (m_bInside)
// 	{
// 		if (bLeft)
// 		{
// 			//m_xCameraMovement.AddMove(m_xCameraProperties.m_xCameraGrapple);
// 		}
// 		else
// 		{
// 			//m_xCameraMovement.AddMoveMirror(m_xCameraProperties.m_xCameraGrapple);
// 		}
// 	}
// 	else
// 	{
// 		if (bLeft)
// 		{
// 			//m_xCameraMovement.AddMove(m_xCameraProperties.m_xCameraGrapple1);
// 			//m_xCameraMovement.AddMove(m_xCameraProperties.m_xCameraGrapple2);
// 		}
// 		else
// 		{
// 			//m_xCameraMovement.AddMoveMirror(m_xCameraProperties.m_xCameraGrapple1);
// 			//m_xCameraMovement.AddMoveMirror(m_xCameraProperties.m_xCameraGrapple2);
// 		}
// 	}
// }

// Set whether the player has lock held

// void AVP_Actor_CameraHandler::SetLockedCamera(bool bLocked)
// {
// 	m_bLocked = bLocked;
// }

// orientate the player relative to the camera orientation given the movement direction
// 
// void AVP_Actor_CameraHandler::OrientatePlayerRelativeToCamera (float fDirectionZ, float fDirectionX, Asura_Euler_3& xPlayerEuler)
// {
// 	Asura_Matrix_3x3 xPlayerOrientation;
// 	xPlayerEuler.ToMatrix(xPlayerOrientation);
// 
// 	// obtain the movement vector relative to the camera
// 	Asura_Vector_3 xMovement = Asura_Vector_3(fDirectionX, 0.0f, fDirectionZ) * m_xCameraOrientation;
// 	if (xMovement.Magnitude() > 0.01f)
// 	{
// 		// obtain the orientation from the movement direction
// 		Asura_Maths::OrientationFromDirection(xPlayerOrientation, xMovement);
// 		Asura_Euler_3 xTargetEuler;
// 		xTargetEuler.FromMatrix(xPlayerOrientation);
// 		AVP_Maths::TurnExponentiallyTowards(xPlayerEuler, xTargetEuler, 15.0f);
// 	}
// }

// place the camera the current camera distance (for inside or outside camera) directly behind the player
// 
// void AVP_Actor_CameraHandler::PlaceCameraBehindPlayer()
// {
// 	const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);
// 	if (!pxClient)
// 	{
// 		Asura_Assert(false, "Unable to find camera handler owner");
// 		return;
// 	}
// 
// 	Asura_Vector_3 xPlayerPosition = pxClient->GetCurrentPosition();
// 
// 	Asura_Matrix_3x3 xPlayerOrientation = pxClient->GetCurrentOrientation();
// 	Asura_Matrix_3x3 xTurnOrientation;
// 
// 	AVP_Maths::ExtractTurnFrom(xPlayerOrientation, xTurnOrientation);
// 
// 	const AVP_Actor_AnimationHandler* pxAnimationHandler = pxClient->GetAnimationHandler();
// 	if (!pxAnimationHandler)
// 	{
// 		Asura_Assert(false, "Unable to get animation handler");
// 		return;
// 	}
// 	Asura_Vector_3 xCameraPosition;
// 	Asura_Matrix_3x3 xCameraOrientation;
// 	pxAnimationHandler->GetCameraData(xCameraPosition, xCameraOrientation);
// 	Asura_Vector_3 xFromPlayer = Asura_Vector_3(0.0f, 0.0f, -m_fCameraDistance) * xCameraOrientation;
// 	m_xCameraPosition = xPlayerPosition + m_xCameraProperties.m_xOffsetHead * xTurnOrientation + xFromPlayer;
// 	Asura_Maths::OrientationFromDirection(m_xCameraOrientation, -xFromPlayer);		// face the actor
// }

// perform a physics movement query with a sphere of the specified radius

float AVP_Actor_CameraHandler::MovementQuery(const Asura_Vector_3& xStart, Asura_Vector_3 xDirection, float fMaximumDistance, float fRadius)
{
	Asura_OBB xSphere = Asura_OBB(fRadius);
	Asura_AVector3 xBegin(xStart);
	Asura_AVector3 xDir(xDirection);
	xDir.Normalise();
	Asura_Physics_MovementQuery xInput(xBegin, xDir * fMaximumDistance, xSphere);
	xInput.SetCollideWithActorObjects(false);
	Asura_Physics_MovementQuery_Results xResults(xInput);
	return xResults.GetAllowedDisplacement().Magnitude();
}

// perform a line of sight test, write the normal and return distance to intersection

float AVP_Actor_CameraHandler::LineOfSight(const Asura_Vector_3& xStart, Asura_Vector_3 xDirection, float fMaximumDistance, Asura_Vector_3& xNormal)
{
	Asura_Vector_3 xDir(xDirection);
	xDir.Normalise();
	Asura_LOS_InputData xInputData (xStart, xDir, m_uOwnerGuid, fMaximumDistance);
	const Asura_LOS_OutputData* pxOutputData = Asura_LOS::TestAgainstEnvironment(&xInputData);
	xNormal = pxOutputData->m_xNormal;
	return pxOutputData->m_fDistanceToIntersection;
}

// Perform two line of sight tests from the player's head in the left and right directions perpendicular to the player's
// orientation. Return true if there is more space to the left and false if there is more space to the right. If there
// is lots of space in both directions (at least dbg_fLOSRoomLimit), return a random bool.
// 
// bool AVP_Actor_CameraHandler::IsMoreSpacePerpendicularLeft(const AVP_ClientEntity_Actor& xPlayer)
// {
// 	Asura_Matrix_3x3 xPlayerOrientation, xTurnOrientation, xPerpendicular;
// 	const AVP_Actor_OrientationHandler*		pxOrientationHandler		= xPlayer.GetOrientationHandler();
// 
// 	if ( pxOrientationHandler == NULL )
// 	{
// 		Asura_Assert(false, "No look handler");
// 		return true;
// 	}
// 
// 	pxOrientationHandler->GetLookOrientation().ToMatrix(xPlayerOrientation);
// 
// 	// Get the player's turn matrix (without the tilt) and rotate it 90 degrees left
// 	AVP_Maths::ExtractTurnFrom(xPlayerOrientation, xTurnOrientation);
// 	dbg_xCameraEuler90Left.ToMatrix(xPerpendicular);
// 	xPerpendicular = xTurnOrientation * xPerpendicular;		// rotate the turn orientation 90 degrees left
// 
// 	Asura_Vector_3 xDirection = Asura_Vector_3(0.0f, 0.0f, -1.0f) * xPerpendicular;	// obtain the left perpendicular direction and normalise for LOS tests
// 	xDirection.Normalise();
// 
// 	const Asura_Vector_3 xActorHead = xPlayer.GetCurrentPosition() + m_xCameraProperties.m_xOffsetHead * xTurnOrientation;
// 
// 	// Perform the left and right perpendicular LOS tests
// 	Asura_LOS_InputData xInputData1 (xActorHead, xDirection, m_uOwnerGuid, m_xCameraProperties.m_fDistanceGrappleCheck);
// 	const Asura_LOS_OutputData* pxOutputData1 = Asura_LOS::TestAgainstEverything(&xInputData1);
// 	const float fDistanceLeft = pxOutputData1->m_fDistanceToIntersection;
// 	//AVP_Maths::CreateLine(xPlayer.GetCurrentPosition() + dbg_xCameraHead * xTurnOrientation, xDirection, dbg_fLOSRoomLimit);
// 
// 	Asura_LOS_InputData xInputData2 (xActorHead, -xDirection, m_uOwnerGuid, m_xCameraProperties.m_fDistanceGrappleCheck);
// 	const Asura_LOS_OutputData* pxOutputData2 = Asura_LOS::TestAgainstEverything(&xInputData2);
// 	const float fDistanceRight = pxOutputData2->m_fDistanceToIntersection;
// 	//AVP_Maths::CreateLine(xPlayer.GetCurrentPosition() + dbg_xCameraHead * xTurnOrientation, -xDirection, dbg_fLOSRoomLimit, 0xff00ff00);
// 
// 	if (fDistanceLeft == m_xCameraProperties.m_fDistanceGrappleCheck && fDistanceRight == m_xCameraProperties.m_fDistanceGrappleCheck)	// reached LOS distance limit
// 	{
// 		if (Asura_Random::GetIntU()%2 == 0) return true;						// choose random direction
// 		return false;
// 	}
// 
// 	if (fDistanceLeft > pxOutputData2->m_fDistanceToIntersection) return true;	// more room to the left
// 	return false;																// more room to the right
// }

void AVP_Actor_CameraHandler::RefreshBlueprintData(void)
{
	const AVP_ClientEntity_Actor*	pxOwner	= AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);

	if(pxOwner != NULL)
	{
		const Asura_Hash_ID		uBlueprintHash	= pxOwner->GetBlueprintHashID();
		const Asura_Hash_ID		uTypeHash		= uAVP_BLUEPRINT_TYPE_PLAYERS;

		const Asura_Blueprint*	pxBlueprint		= Asura_Blueprint_System::GetBlueprintByHashID(uTypeHash, uBlueprintHash);
		if(pxBlueprint)
		{
			Asura_Value	xValue;

			if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_DEFAULT_FOV, xValue))
			{
				m_afFovZoom[0] = Asura_Maths::DegToRad(xValue.GetFloatValue());
			}
			if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ZOOM_LEVEL_1_FOV, xValue))
			{
				m_afFovZoom[1] = Asura_Maths::DegToRad(xValue.GetFloatValue());
			}
			if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ZOOM_LEVEL_2_FOV, xValue))
			{
				m_afFovZoom[2] = Asura_Maths::DegToRad(xValue.GetFloatValue());
			}
			if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ZOOM_LEVEL_3_FOV, xValue))
			{
				m_afFovZoom[3] = Asura_Maths::DegToRad(xValue.GetFloatValue());
			}
			if(pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ZOOM_SPEED_SCALE, xValue))
			{
				m_fZoomSpeed = xValue.GetFloatValue();
			}
		}
		else
		{
			// Set usable values and assert
			Asura_Assert(false,"Camera blueprints could not be read.");
			m_afFovZoom[0] = Asura_Maths::DegToRad(85.0f);
			m_afFovZoom[1] = Asura_Maths::DegToRad(60.0f);
			m_afFovZoom[2] = Asura_Maths::DegToRad(45.0f);
			m_afFovZoom[3] = Asura_Maths::DegToRad(30.0f);
		}
	}
}

/// End of file. ///

