///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Actor_InputHandler.cpp
// Author : Andy Tate
// Date   : 4-Oct-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/AVP_Actor_InputHandler.h"

// Asura independent header-files...
#include "Core/Asura_Maths.h"
#include "Core/Asura_Texture_Streaming.h"
#include "Core/Asura_Timers.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Debugging/Asura_DebugText.h"
#include "Input/Asura_ControlMapping_Key.h"

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	#include "Interface/Asura_CommandConsole.h"
#endif

#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"


// Asura platform header-files...

// Game independent header-files...
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_CameraHandler.h"
#include "Actor/AVP_Actor_InteractionHandler.h"
#include "Actor/AVP_Actor_PickupHandler.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Core.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_Settings.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ClientActor_Human.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ClientActor_Predator.h"
#include "Entity/PhysicalObject/Pickup/AVP_ClientEntity_Pickup.h"
#include "Entity/Physicalobject/Pickup/AVP_Pickup.h"
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
#include "GUIMenu/AVP_GUIMenu_System.h"
#endif
#include "Input/Asura_ControlManager.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Moves/AVP_Movement_System.h"
#include "Moves/AVP_Moves.h"
#include "Multiplayer/AVP_Multiplayer_ScoreBoard.h"
#include "Network/AVP_Client.h"
#include "Network/AVP_Server.h"
#include "Network/AVP_Player.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

const float fCUTSCENE_MIN_PLAYING_TIME = 1.0f;

DEBUGVAR Asura_Angle	dbg_fMaxTilt		= Asura_Maths::DegToRad(72.0f);
DEBUGVAR Asura_Angle	dbg_fMinTilt		= Asura_Maths::DegToRad(-72.0f);

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

float AVP_Actor_InputHandler::s_fInnerAimAssistMultiplier = 0.5f;
float AVP_Actor_InputHandler::s_fOuterAimAssistMultiplier = 1.0f;
float AVP_Actor_InputHandler::s_fLookAssistAngleInDegree = 10.0f;
float AVP_Actor_InputHandler::s_fLookAssistHorizontalMaxSpeedInDegree = 20.0f;
float AVP_Actor_InputHandler::s_fLookAssistVerticalMaxSpeedInDegree = 0.0f;
float AVP_Actor_InputHandler::s_fMouseVSensitivity = 1.0f; // Bigger is quicker
float AVP_Actor_InputHandler::s_fMouseHSensitivity = 1.0f; // Bigger is quicker

float AVP_Actor_InputHandler::s_fTiltMaxSpeedInDegree = 90.0f;
float AVP_Actor_InputHandler::s_fTiltRampUpTime = 0.25f;
float AVP_Actor_InputHandler::s_fTiltRampDownTime = 0.05f;
float AVP_Actor_InputHandler::s_fTiltExponent = 2.0f;

float AVP_Actor_InputHandler::s_fTurnDefaultMaxSpeedInDegree = 180.0f;
float AVP_Actor_InputHandler::s_fTurnTurboMaxSpeedInDegree = 360.0f;
float AVP_Actor_InputHandler::s_fTurnTurboDelay = 1.0f;
float AVP_Actor_InputHandler::s_fTurnTurboRampUpTime = 0.5f;
float AVP_Actor_InputHandler::s_fTurnRampUpTime = 0.5f;
float AVP_Actor_InputHandler::s_fTurnRampDownTime = 0.10f;
float AVP_Actor_InputHandler::s_fTurnExponent = 2.0f;

float AVP_Actor_InputHandler::s_fTanLookAssistHalfAngleInRadian = 0.0f;
float AVP_Actor_InputHandler::s_fLookAssistHorizontalMaxSpeedInRadian = 0.0f;
float AVP_Actor_InputHandler::s_fLookAssistVerticalMaxSpeedInRadian = 0.0f;
float AVP_Actor_InputHandler::s_fTiltMaxSpeedInRadian = 0.0f;
float AVP_Actor_InputHandler::s_fTurnDefaultMaxSpeedInRadian = 0.0f;
float AVP_Actor_InputHandler::s_fTurnTurboMaxSpeedInRadian = 0.0f;





#if defined(AVP_ALLOW_DEBUG_MOVEMENT)
AVP_Actor_InputHandler::MovementType	AVP_Actor_InputHandler::s_eMovementType	= MT_PLAYER_FIRST_PERSON;
#endif

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(AVP_DEBUG_INPUT)
u_int AVP_Actor_InputHandler::s_uConsoleVarReferenceCount = 0;
#endif

bool AVP_Actor_InputHandler::s_bStreamingTextureDelayActive = false;

float AVP_Actor_InputHandler::s_fSavedTimeScale = 1.0f;


void AVP_Actor_InputHandler::MapMovementAxes(Asura_ControlMap& xMap, const u_int uAxisForwardsBackwards, const u_int uAxisStrafeLeftRight, const u_int uKeySlow, const u_int uKeyFast, const u_int uKeyStanceToggle)
{
	// Note this ignores several of the values, which have been included to fix hiding warnings.
	xMap.AddAxisMapping(FORWARDS_BACKWARDS, uAxisForwardsBackwards);
	xMap.AddAxisMapping(STRAFE_LEFT_RIGHT, uAxisStrafeLeftRight);
}


void AVP_Actor_InputHandler::MapMovementKeys(Asura_ControlMap& xMap, const u_int uKeyForwards, const u_int uKeyBackwards, const u_int uKeyStrafeLeft, const u_int uKeyStrafeRight, const u_int uKeySlow, const u_int uKeyFast, const u_int uKeyStanceToggle)
{
	// Note this ignores several of the values, which have been included to fix hiding warnings.
	xMap.AddKeyMapping(FORWARDS_BACKWARDS, uKeyForwards, Asura_ControlMapping::ON_PRESS, 1.0f);
	xMap.AddKeyMapping(FORWARDS_BACKWARDS, uKeyBackwards, Asura_ControlMapping::ON_PRESS, -1.0f);
	xMap.AddKeyMapping(STRAFE_LEFT_RIGHT, uKeyStrafeLeft, Asura_ControlMapping::ON_PRESS, -1.0f);
	xMap.AddKeyMapping(STRAFE_LEFT_RIGHT, uKeyStrafeRight, Asura_ControlMapping::ON_PRESS, 1.0f);
}

void AVP_Actor_InputHandler::MapWeaponSelect(Asura_ControlMap& xMap,
											 const u_int uKeyUp,
											 const u_int uKeyDown,
											 const u_int uKeyLeft,
											 const u_int uKeyRight)
{
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyUp, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyLeft, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyRight, Asura_ControlMapping::ON_DEBOUNCED_PRESS);

	xMap.AddKeyMapping(WEAPON_SELECTOR_HOLD, uKeyUp, Asura_ControlMapping::ON_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_HOLD, uKeyLeft, Asura_ControlMapping::ON_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_HOLD, uKeyRight, Asura_ControlMapping::ON_PRESS);


	Asura_ControlMapping*	pxUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_PRESS, 1.0f);
	Asura_ControlMapping*	pxNotLeft	= new Asura_ControlMapping_Key (uKeyLeft,	Asura_ControlMapping::ON_RELEASE, 1.0f);
	Asura_ControlMapping*	pxNotRight	= new Asura_ControlMapping_Key (uKeyRight,	Asura_ControlMapping::ON_RELEASE, 1.0f);
	xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_UP, new Asura_ControlMapping_Combination( pxUp, new Asura_ControlMapping_Combination( pxNotLeft, pxNotRight, 1.0f ), 1.0f ) );

							pxUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_PRESS, 1.0f);
	Asura_ControlMapping*	pxRight		= new Asura_ControlMapping_Key (uKeyRight,	Asura_ControlMapping::ON_PRESS, 1.0f);
	xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_UP_RIGHT, new Asura_ControlMapping_Combination( pxUp, pxRight, 1.0f ) );

							pxUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_PRESS, 1.0f);
	Asura_ControlMapping*	pxLeft		= new Asura_ControlMapping_Key (uKeyLeft,	Asura_ControlMapping::ON_PRESS, 1.0f);
	xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_UP_LEFT, new Asura_ControlMapping_Combination( pxUp, pxLeft, 1.0f ) );

	if ( uKeyDown == ASURA_KEYID_VOID )
	{
		pxRight		= new Asura_ControlMapping_Key (uKeyRight,	Asura_ControlMapping::ON_PRESS, 1.0f);
		Asura_ControlMapping*	pxNotUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_RELEASE, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_RIGHT, new Asura_ControlMapping_Combination( pxRight, pxNotUp, 1.0f ) );

		pxLeft		= new Asura_ControlMapping_Key (uKeyLeft,	Asura_ControlMapping::ON_PRESS, 1.0f);
		pxNotUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_RELEASE, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_LEFT, new Asura_ControlMapping_Combination( pxLeft, pxNotUp, 1.0f ) );
	}
	else
	{
		xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyDown, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
		xMap.AddKeyMapping(WEAPON_SELECTOR_HOLD, uKeyDown, Asura_ControlMapping::ON_PRESS);

		pxRight		= new Asura_ControlMapping_Key (uKeyRight,	Asura_ControlMapping::ON_PRESS, 1.0f);
		Asura_ControlMapping*	pxNotUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_RELEASE, 1.0f);
		Asura_ControlMapping*	pxNotDown	= new Asura_ControlMapping_Key (uKeyDown,	Asura_ControlMapping::ON_RELEASE, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_RIGHT, new Asura_ControlMapping_Combination( pxRight, new Asura_ControlMapping_Combination( pxNotUp, pxNotDown, 1.0f ), 1.0f ) );

		pxLeft		= new Asura_ControlMapping_Key (uKeyLeft,	Asura_ControlMapping::ON_PRESS, 1.0f);
		pxNotUp		= new Asura_ControlMapping_Key (uKeyUp,		Asura_ControlMapping::ON_RELEASE, 1.0f);
		pxNotDown	= new Asura_ControlMapping_Key (uKeyDown,	Asura_ControlMapping::ON_RELEASE, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_LEFT, new Asura_ControlMapping_Combination( pxLeft, new Asura_ControlMapping_Combination( pxNotUp, pxNotDown, 1.0f ), 1.0f ) );

		Asura_ControlMapping*	pxDown		= new Asura_ControlMapping_Key (uKeyDown,	Asura_ControlMapping::ON_PRESS, 1.0f);
		pxNotLeft	= new Asura_ControlMapping_Key (uKeyLeft,	Asura_ControlMapping::ON_RELEASE, 1.0f);
		pxNotRight	= new Asura_ControlMapping_Key (uKeyRight,	Asura_ControlMapping::ON_RELEASE, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_DOWN, new Asura_ControlMapping_Combination( pxDown, new Asura_ControlMapping_Combination( pxNotLeft, pxNotRight, 1.0f ), 1.0f ) );

		pxDown		= new Asura_ControlMapping_Key (uKeyDown,	Asura_ControlMapping::ON_PRESS, 1.0f);
		pxRight		= new Asura_ControlMapping_Key (uKeyRight,	Asura_ControlMapping::ON_PRESS, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_DOWN_RIGHT, new Asura_ControlMapping_Combination( pxDown, pxRight, 1.0f ) );

		pxDown		= new Asura_ControlMapping_Key (uKeyDown,	Asura_ControlMapping::ON_PRESS, 1.0f);
		pxLeft		= new Asura_ControlMapping_Key (uKeyLeft,	Asura_ControlMapping::ON_PRESS, 1.0f);
		xMap.AddCombinationMapping( WEAPON_SELECTOR_SLOT_DOWN_LEFT, new Asura_ControlMapping_Combination( pxDown, pxLeft, 1.0f ) );
	}
}

void AVP_Actor_InputHandler::MapWeaponSelect(Asura_ControlMap& xMap,
											 const u_int uKeyUp,
											 const u_int uKeyUpRight,
											 const u_int uKeyRight,
											 const u_int uKeyDownRight,
											 const u_int uKeyDown,
											 const u_int uKeyDownLeft,
											 const u_int uKeyLeft,
											 const u_int uKeyUpLeft)
{
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyUp, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyUpRight, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyRight, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyDownRight, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyDown, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyDownLeft, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyLeft, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(WEAPON_SELECTOR_BEGIN, uKeyUpLeft, Asura_ControlMapping::ON_DEBOUNCED_PRESS);

	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_UP,			uKeyUp, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_UP_RIGHT,		uKeyUpRight, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_RIGHT,			uKeyRight, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_DOWN_RIGHT,	uKeyDownRight, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_DOWN,			uKeyDown, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_DOWN_LEFT,		uKeyDownLeft, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_LEFT,			uKeyLeft, Asura_ControlMapping::ON_PRESS );
	xMap.AddKeyMapping( WEAPON_SELECTOR_SLOT_UP_LEFT,		uKeyUpLeft, Asura_ControlMapping::ON_PRESS );
}

void AVP_Actor_InputHandler::MapVision(Asura_ControlMap& xMap,
										const u_int uKeyHold,
										const u_int uKeyNext,
										const u_int uKeyPrev)
{
	xMap.AddKeyMapping(VISION_SELECTOR_HOLD, uKeyHold, Asura_ControlMapping::ON_PRESS);
	xMap.AddKeyMapping(VISION_SELECTOR_NEXT, uKeyNext, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(VISION_SELECTOR_PREV, uKeyPrev, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
}

void AVP_Actor_InputHandler::MapVisionSelect(Asura_ControlMap& xMap,
											const u_int uKeyUp,
											const u_int uKeyDown,
											const u_int uKeyLeft,
											const u_int uKeyRight)
{
	xMap.AddKeyMapping(VISION_SELECTOR_SLOT_UP, uKeyUp, Asura_ControlMapping::ON_PRESS);
	xMap.AddKeyMapping(VISION_SELECTOR_SLOT_DOWN, uKeyDown, Asura_ControlMapping::ON_PRESS);
	xMap.AddKeyMapping(VISION_SELECTOR_SLOT_LEFT, uKeyLeft, Asura_ControlMapping::ON_PRESS);
	xMap.AddKeyMapping(VISION_SELECTOR_SLOT_RIGHT, uKeyRight, Asura_ControlMapping::ON_PRESS);
}

void AVP_Actor_InputHandler::MapLookAxes(Asura_ControlMap& xMap,
										const u_int uAxisLeftRight,
										const u_int uAxisUpDown)
{
	xMap.AddAxisMapping(LOOK_UP_DOWN, uAxisUpDown, -1.0f, &s_fTiltRampUpTime, &s_fTiltRampDownTime, ASURA_CONTROLMAPPING_AXIS_MASK_ALL, &s_fTiltExponent );
	xMap.AddAxisMapping(LOOK_LEFT_RIGHT, uAxisLeftRight, 1.0f, &s_fTurnRampUpTime, &s_fTurnRampDownTime, ASURA_CONTROLMAPPING_AXIS_MASK_ALL, &s_fTurnExponent );
}

void AVP_Actor_InputHandler::MapLookKeys(Asura_ControlMap& xMap,
										 const u_int uKeyUp,
										 const u_int uKeyDown,
										 const u_int uKeyLeft,
										 const u_int uKeyRight)
{
	xMap.AddKeyMapping(LOOK_UP_DOWN, uKeyUp, Asura_ControlMapping::ON_PRESS, 1.0f);
	xMap.AddKeyMapping(LOOK_UP_DOWN, uKeyDown, Asura_ControlMapping::ON_PRESS, -1.0f);
	xMap.AddKeyMapping(LOOK_LEFT_RIGHT, uKeyLeft, Asura_ControlMapping::ON_PRESS, -1.0f);
	xMap.AddKeyMapping(LOOK_LEFT_RIGHT, uKeyRight, Asura_ControlMapping::ON_PRESS, 1.0f);
}

void AVP_Actor_InputHandler::MapLookMouse(Asura_ControlMap& xMap)
{
	xMap.AddAxisMapping(LOOK_MOUSE_UP_DOWN, ASURA_AXIS_MOUSE_V, -s_fMouseVSensitivity, 0.0f );
	xMap.AddAxisMapping(LOOK_MOUSE_LEFT_RIGHT, ASURA_AXIS_MOUSE_H, s_fMouseHSensitivity, 0.0f );
}

void AVP_Actor_InputHandler::MapZoom(Asura_ControlMap& xMap,
										const u_int uKeyUp,
										const u_int uKeyDown,
										const u_int uKeyChange)
{
	xMap.AddKeyMapping(ZOOM_IN, uKeyUp, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(ZOOM_OUT, uKeyDown, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(ZOOM_HOLD, uKeyChange, Asura_ControlMapping::ON_PRESS);
}

void AVP_Actor_InputHandler::MapInteract(Asura_ControlMap& xMap, const u_int uKeyInteract, float fDelayTime )
{
	if( fDelayTime==0.0f )
	{
		xMap.AddKeyMapping(INTERACT, uKeyInteract, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	}
	else
	{
		xMap.AddHoldKeyMapping(INTERACT, uKeyInteract, fDelayTime );
	}
}

void AVP_Actor_InputHandler::MapCutscene(Asura_ControlMap& xMap, const u_int uKey1, const u_int uKey2 /* = ASURA_KEYID_VOID */, const u_int uKey3 /* = ASURA_KEYID_VOID */, const u_int uKey4 /* = ASURA_KEYID_VOID */, const u_int uKey5 /* = ASURA_KEYID_VOID */)
{
	xMap.AddKeyMapping(SKIP_CUTSCENE, uKey1, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	if ( uKey2 != ASURA_KEYID_VOID ) { xMap.AddKeyMapping(SKIP_CUTSCENE, uKey2, Asura_ControlMapping::ON_DEBOUNCED_PRESS); }
	if ( uKey3 != ASURA_KEYID_VOID ) { xMap.AddKeyMapping(SKIP_CUTSCENE, uKey3, Asura_ControlMapping::ON_DEBOUNCED_PRESS); }
	if ( uKey4 != ASURA_KEYID_VOID ) { xMap.AddKeyMapping(SKIP_CUTSCENE, uKey4, Asura_ControlMapping::ON_DEBOUNCED_PRESS); }
	if ( uKey5 != ASURA_KEYID_VOID ) { xMap.AddKeyMapping(SKIP_CUTSCENE, uKey5, Asura_ControlMapping::ON_DEBOUNCED_PRESS); }
}

void AVP_Actor_InputHandler::MapMenuToggle(Asura_ControlMap& xMap, const u_int uKey)
{
	xMap.AddKeyMapping(MENU_TOGGLE, uKey, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
}

DEBUGVAR float dbg_fSelectHoldTime = 0.5f;
void AVP_Actor_InputHandler::MapObjectivesAndSelectToggle(Asura_ControlMap& xMap, const u_int uKeyObjectives, const u_int uKeySelect)
{
#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	// We allow these keys to be the same, but when they are they need a special set-up.
	if (uKeyObjectives == uKeySelect)
	{
		xMap.AddTapKeyMapping(OBJECTIVES_TOGGLE,  uKeyObjectives, dbg_fSelectHoldTime);
		xMap.AddHoldKeyMapping(DVS_TOGGLE, uKeySelect, dbg_fSelectHoldTime, 1.0f, Asura_ControlMapping_HoldKey::NO_RAMPING, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	}
	else
	{
		xMap.AddKeyMapping(OBJECTIVES_TOGGLE, uKeyObjectives, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
		xMap.AddKeyMapping(DVS_TOGGLE, uKeySelect, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	}

#else
	xMap.AddKeyMapping(OBJECTIVES_TOGGLE, uKeyObjectives, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(SCOREBOARD, uKeySelect, Asura_ControlMapping::ON_PRESS);
#endif	// defined(ASURA_DEBUG_VARIABLE_SYSTEM)
}

void AVP_Actor_InputHandler::MapFreezeTime(Asura_ControlMap& xMap, const u_int uKey)
{
	xMap.AddKeyMapping(FREEZE_TIME_TOGGLE, uKey, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_InputHandler::AVP_Actor_InputHandler
// Author       : Andy Tate
// Date         : 4-Oct-2006
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_InputHandler::AVP_Actor_InputHandler (const Asura_Guid uOwnerGuid, const u_int uMaxMappings)
: m_uOwnerGuid(uOwnerGuid)
, m_fTurboTimer( 0.0f )
, m_xMapFirstPerson(uMaxMappings)
, m_xMapVisionSelector(uMaxMappings)
, m_xMapCutscene(uMaxMappings)
{
	Asura_ControlMap*	pxDefaultControlMap	= &m_xMapFirstPerson;

	// Set the default control map...
	Asura_ControlManager::AddDefaultControls(m_uOwnerGuid, pxDefaultControlMap);
	m_pxCurrentControlMap	= pxDefaultControlMap;

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(AVP_DEBUG_INPUT)
	if ( s_uConsoleVarReferenceCount == 0 )
	{
		if ( !Asura_CommandConsole::VarExists( L"Controls.InnerAimAssistMultiplier" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.InnerAimAssistMultiplier", &s_fInnerAimAssistMultiplier );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.OuterAimAssistMultiplier" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.OuterAimAssistMultiplier", &s_fOuterAimAssistMultiplier );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.LookAssistStrength" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.LookAssistStrength", &s_fLookAssistAngleInDegree );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.LookAssistHorizontalMaxSpeed" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.LookAssistHorizontalMaxSpeed", &s_fLookAssistHorizontalMaxSpeedInDegree );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.LookAssistVerticalMaxSpeed" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.LookAssistVerticalMaxSpeed", &s_fLookAssistVerticalMaxSpeedInDegree );
		}

		if ( !Asura_CommandConsole::VarExists( L"Controls.Tilt.MaxSpeed" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Tilt.MaxSpeed", &s_fTiltMaxSpeedInDegree );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Tilt.RampUpTime" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Tilt.RampUpTime", &s_fTiltRampUpTime );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Tilt.RampDownTime" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Tilt.RampDownTime", &s_fTiltRampDownTime );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Tilt.Exponent" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Tilt.Exponent", &s_fTiltExponent );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Tilt.MouseSensitivity" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Tilt.MouseSensitivity", &s_fMouseVSensitivity );
		}

		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.DefaultMaxSpeed" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.DefaultMaxSpeed", &s_fTurnDefaultMaxSpeedInDegree );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.TurboMaxSpeed" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.TurboMaxSpeed", &s_fTurnTurboMaxSpeedInDegree );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.TurboDelay" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.TurboDelay", &s_fTurnTurboDelay );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.TurboRampUpTime" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.TurboRampUpTime", &s_fTurnTurboRampUpTime );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.RampUpTime" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.RampUpTime", &s_fTurnRampUpTime );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.RampDownTime" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.RampDownTime", &s_fTurnRampDownTime );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.Exponent" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.Exponent", &s_fTurnExponent );
		}
		if ( !Asura_CommandConsole::VarExists( L"Controls.Turn.MouseSensitivity" ) )
		{
			Asura_CommandConsole::AddVar( L"Controls.Turn.MouseSensitivity", &s_fMouseHSensitivity );
		}
	}
	++s_uConsoleVarReferenceCount;
#endif
}

/// End of function 'AVP_Actor_InputHandler::AVP_Actor_InputHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_InputHandler::~AVP_Actor_InputHandler
// Author       : Andy Tate
// Date         : 19-Apr-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_InputHandler::~AVP_Actor_InputHandler()
{
	// Remove this control-map...
	Asura_ControlManager::RemoveDefaultControls(m_uOwnerGuid);
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE) && defined(AVP_DEBUG_INPUT)
	--s_uConsoleVarReferenceCount;
#endif
}

/// End of function 'AVP_Actor_InputHandler::~AVP_Actor_InputHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_InputHandler::Reset
// Author       : Andy Tate
// Date         : 18-Oct-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_InputHandler::Reset (void)
{
	m_xLookOrientationInput.Zero();
	m_xMovementInput.Zero();
}

////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_InputHandler::Update
// Author       : Andy Tate
// Date         : 4-Oct-2006
////////////////////////////////////////////////////////////////////////////////

Asura_Message* AVP_Actor_InputHandler::Update ( AVP_ClientEntity_Actor& xClient )
{
	if ( !xClient.IsDead() )
	{
		AVP_Actor_InteractionHandler*	pxInteractionHandler	= xClient.GetInteractionHandler();
		AVP_Actor_PickupHandler*	    pxPickupHandler	        = xClient.GetPickupHandler();

		const AVP_Inventory_System*		pxInventory				= xClient.GetInventorySystem();
		const AVP_Inventory_Item*		pxStim					= pxInventory ? pxInventory->GetItem(uAVP_BLUEPRINT_PARAM_STIM) : NULL;
		const AVP_Inventory_Item*		pxCurrentWeapon			= pxInventory ? pxInventory->GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON) : NULL;
		const AVP_Inventory_Item*	    pxCloak                 = pxInventory ? pxInventory->GetItem( uAVP_INVENTORY_ITEMHASH_CLOAK ): NULL;
		
		bool                            bCloakIdling            = true;
		if ( pxCloak )
		{
			const bool bIdling      = pxCloak->IsIdling(false);
			const bool bFullyIdling = pxCloak->IsIdling(true);
			const bool bUsing       = pxCloak->IsUsing(false);
			const bool bFullyUsing  = pxCloak->IsUsing(true);

			if ( ((bIdling && !bFullyIdling) || (bUsing && !bFullyUsing)) ) bCloakIdling = false;
		}

		const bool						bStimIdling				= pxStim ? pxStim->IsIdling(true) : true;
		const bool						bWeaponIdling			= pxCurrentWeapon ? pxCurrentWeapon->IsIdling(true) : true;

		const AVP_Movement_System*		pxMovementSystem		= static_cast<const AVP_Movement_System*>(xClient.GetMovementSystem());
		const AVP_Move*					pxCurrentMove			= (pxMovementSystem == NULL) ? NULL : static_cast<const AVP_Move*>(pxMovementSystem->GetCurrentMove());

#if defined(ASURA_USE_GUI_MENU_SYSTEM)
		const bool					bInMenu				= AVP_GUIMenu_System::IsCurrentlyInNonHudOrLoadingMenu();
#endif //defined(ASURA_USE_GUI_MENU_SYSTEM)

		if(Asura_Render::GetCutsceneHasCameraControl())
		{
			if ( m_pxCurrentControlMap != &m_xMapCutscene )
			{
				PushControls( &m_xMapCutscene );
			}

			if(GetControlMapValue(SKIP_CUTSCENE, true, true))
			{
				Asura_Cutscene_System::CheckAndRequestSkipCutscenes(fCUTSCENE_MIN_PLAYING_TIME);
			}
		}
		else
		{
			if ( m_pxCurrentControlMap == &m_xMapCutscene )
			{
				PopControls();
			}

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
			s_fTanLookAssistHalfAngleInRadian = Asura_Maths::Tangent(Asura_Maths::DegToRad(s_fLookAssistAngleInDegree) / 2.0f);
			s_fLookAssistHorizontalMaxSpeedInRadian = Asura_Maths::DegToRad(s_fLookAssistHorizontalMaxSpeedInDegree);
			s_fLookAssistVerticalMaxSpeedInRadian = Asura_Maths::DegToRad(s_fLookAssistVerticalMaxSpeedInDegree);
			s_fTiltMaxSpeedInRadian = Asura_Maths::DegToRad(s_fTiltMaxSpeedInDegree);
			s_fTurnDefaultMaxSpeedInRadian = Asura_Maths::DegToRad(s_fTurnDefaultMaxSpeedInDegree);
			s_fTurnTurboMaxSpeedInRadian = Asura_Maths::DegToRad(s_fTurnTurboMaxSpeedInDegree);
#endif	// defined(ASURA_SUPPORT_COMMAND_CONSOLE)


			// Interaction handling
			// Update the interaction...
			if ( bStimIdling && bWeaponIdling && bCloakIdling )
			{
				// See if the player is already interacting
				if( pxInteractionHandler->IsInteracting() && pxInteractionHandler->IsInIdleInteraction()) 
				{
					float fControlValue = GetControlMapValue( STRAFE_LEFT_RIGHT, true, true );
					if ( fControlValue != 0.0f )
					{
						pxInteractionHandler->HandleInteraction(AVP_INTERACTION_TYPE_DEFAULT, STRAFE_LEFT_RIGHT, fControlValue);
						m_pxCurrentControlMap->ResetValue( STRAFE_LEFT_RIGHT );
					}
					fControlValue       = GetControlMapValue( INTERACT, true, true );
					if ( fControlValue != 0.0f )
					{
						pxInteractionHandler->HandleInteraction(AVP_INTERACTION_TYPE_DEFAULT, INTERACT, fControlValue);
						m_pxCurrentControlMap->ResetValue( INTERACT );
					}
				}
				else // Check for start of interaction, and pickups
				{
					bool bInteractionHandled = false;
					const bool bInteractRequested = GetControlMapValue(INTERACT) > 0.0f;
					if( pxInteractionHandler && bInteractRequested && pxInteractionHandler->IsInteractionAllowed())
					{
						bInteractionHandled = pxInteractionHandler->HandleInteraction(AVP_INTERACTION_TYPE_DEFAULT, INTERACT, 1.0f);
					}

					// Update Pickups, uses the same button as the interaction, and less privileged than interaction handling;
					if (!bInteractionHandled && (pxPickupHandler != NULL) && bInteractRequested && (pxPickupHandler->GetNearestPickupGuid() != ASURA_GUID_UNREGISTERED))
					{
						const AVP_ClientEntity_Pickup* pxPickup =  AVP_ClientEntity_Pickup::FindEntity(pxPickupHandler->GetNearestPickupGuid());
						if(pxPickup)
						{
							const AVP_ClientActor_Human* const pxHuman = AVP_ClientActor_Human::FindEntity(xClient.GetGuid());
							if ( !pxHuman || !pxHuman->IsCarryingBudddy() )
							{
								if(    pxPickupHandler->IsSpaceForNewWeapon() 
									|| pxPickupHandler->CanCurrentWeaponBeDropped()
									|| ( pxPickup->GetPickupObject()->GetPickupType() == AVP_Pickup::PICKUP_TYPE_TOOL) ) // Can't pick up a weapon if we can't drop the current one..But if it is a tool, pick it.
								{
									pxPickupHandler->SendRequestMessage();
								}
							}
						}
					}
				}
			}

			if ( Asura_Maths::Abs( GetControlMapValue( LOOK_LEFT_RIGHT ) ) < 0.9f )
			{
				m_fTurboTimer = 0.0f;
			}
			else
			{
				m_fTurboTimer += Asura_Timers::GetGameTimeSlice();
			}

			float fTurnMax = s_fTurnDefaultMaxSpeedInRadian;
			if ( m_fTurboTimer > s_fTurnTurboDelay )
			{
				fTurnMax = Asura_Maths::Lerp( s_fTurnDefaultMaxSpeedInRadian, s_fTurnTurboMaxSpeedInRadian, Asura_Maths::ClampToRange( ( m_fTurboTimer - s_fTurnTurboDelay ) / s_fTurnTurboRampUpTime, 0.0f, 1.0f ) );
			}

			// Update the looking input...
			UpdateLookInput( xClient, s_fTiltMaxSpeedInRadian, fTurnMax, s_fInnerAimAssistMultiplier, s_fOuterAimAssistMultiplier, s_fTanLookAssistHalfAngleInRadian, s_fLookAssistHorizontalMaxSpeedInRadian, s_fLookAssistVerticalMaxSpeedInRadian );

			// Update the movement input...
			UpdateMovementInput( xClient );

			//	AVP_ClientActor_Predator* pxClientPredator = AVP_ClientActor_Predator::FindEntity(m_uOwnerGuid);
			//	if (!(pxClientPredator && !pxClientPredator->GetState().IsCriticallyBusy()))
			{
				// If we are in a move that doesn't allow movement...
				if((pxCurrentMove == NULL) || !pxCurrentMove->IsDisableMovement())
				{
					if( !bInMenu )			
					{
#if defined(AVP_ALLOW_DEBUG_MOVEMENT)
						if( s_eMovementType != MT_FLYING_CAMERA)
#endif
						{
							UpdateMovement( xClient ); // This actually only updates jumping
							// AJK; AVP Functionality for tutorial; check if func allowed
							if (AVP_Settings::IsVisionModesFuncOn())
							{
								UpdateVisionModes( xClient );
							}
							UpdateWeapons( xClient );
						}
					}
				}
			}
		} // end if cutscene control
	}

#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	if(GetControlMapValue(DVS_TOGGLE, false))
	{
		Asura_DebugVariableSystem::OpenDisplayDVS();
	}
#else
	if( GetControlMapValue(SCOREBOARD, false) )
	{
		AVP_Multiplayer_ScoreBoard::SetScoreBoardVisible( true ); 
	}
	else
	{
		AVP_Multiplayer_ScoreBoard::SetScoreBoardVisible( false ); 
	}
#endif	// defined(ASURA_DEBUG_VARIABLE_SYSTEM)

	if( GetControlMapValue(MENU_TOGGLE, false) )
	{
		//GJDH Activate PauseMenu here
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
		if( !AVP_GUIMenu_System::IsMenuActive(AVP_GUIMenu_System::GUIMENU_ID_PAUSE) )
#endif //defined(ASURA_USE_GUI_MENU_SYSTEM)
		{
			// AJG - 09/11/07 - Removed any attempt to sync this over the network - the previous implementation didn't work anyway, and it's generally something that's more effort than it's worth.
			//					Instead, we'll just not change the time scale if we're playing across the network.
			AVP_Game_Core::RequestPause();
		}
	}
	
	if(AVP_Settings::IsSinglePlayer())
	{
		if( GetControlMapValue(OBJECTIVES_TOGGLE, false) )
		{
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
			if( !AVP_GUIMenu_System::IsMenuActive(AVP_GUIMenu_System::GUIMENU_ID_OBJECTIVES) )
#endif //defined(ASURA_USE_GUI_MENU_SYSTEM)
			{
				AVP_Game_Core::RequestObjectives();
			}
		}
	}

	if( GetControlMapValue(FREEZE_TIME_TOGGLE, false))
	{
		float fCurrentTimeScale = Asura_Timers::GetGameTimeScale();
		if( fCurrentTimeScale == 0.0f )
		{
			Asura_Timers::SetGameTimeScale( s_fSavedTimeScale );
		}
		else
		{
			s_fSavedTimeScale = fCurrentTimeScale;
			Asura_Timers::SetGameTimeScale(0.0f);
		}
	}

#if defined(ASURA_USE_TEXTURE_STREAMING)
	if (GetControlMapValue(STREAMINGTEXTURES_TOGGLEDELAY, false))
	{
		s_bStreamingTextureDelayActive = !s_bStreamingTextureDelayActive;
		Asura_Texture_Streaming_System::Test_SetLoadingDelay(s_bStreamingTextureDelayActive ? 0.5f : 0.0f);
	}

	if (GetControlMapValue(STREAMINGTEXTURES_UNLOADALL, false))
	{
		Asura_Texture_Streaming_System::Test_UnloadAll();
	}
#endif	//defined(ASURA_USE_TEXTURE_STREAMING)

	// Create the message to be sent to the server...
	return(CreatePlayerCommandMessageToServer( xClient ));
}

/// End of function 'AVP_Actor_InputHandler::Update'. ///



#if defined(AVP_ALLOW_DEBUG_MOVEMENT)

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_InputHandler::RenderDebugInfo
// Author       : Andy Tate
// Date         : 5-Oct-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_InputHandler::RenderDebugInfo (void) const
{
	const float					fIncrement	= Asura_Font_System::GetFontHeight();
	const float					fX			= Asura_Viewport::GetWidth() - fIncrement;
	const float					fY			= fIncrement;
	const float					fScale		= 1.0f;
	const u_int					uColour		= 0xFFFFFFFF;
	const u_int					uFormat		= ASURA_TEXTFORMAT_RIGHTJUSTIFIED | ASURA_TEXTFORMAT_VERTICALBOTTOMJUSTIFIED;
	const Asura_UnicodeChar*	wszText		= NULL;

	// Find the appropriate string...
	switch(s_eMovementType)
	{
		case MT_PLAYER_FIRST_PERSON:
		{
			wszText	= L"Player Movement (First-Person)";
			break;
		}

		case MT_PLAYER_THIRD_PERSON:
		{
			wszText	= L"Player Movement (Third-Person)";
			break;
		}

		case MT_FLYING_PLAYER_COLLISIONS:
		{
			wszText	= L"Flying Player Movement (Collisions)";
			break;
		}

		case MT_FLYING_PLAYER_NO_COLLISIONS:
		{
			wszText	= L"Flying Player Movement (No Collisions)";
			break;
		}

		case MT_FLYING_CAMERA:
		{
			wszText	= L"Camera Movement";
			break;
		}
		default:
		{
			Asura_Assertf( false, ("Invalid switch state %d", s_eMovementType) );
		}
	}

	// Display the text...
	if(wszText != NULL)
	{
		Asura_Font_System::Print(fX, fY, fScale, uColour, uFormat, wszText);
	}
}

/// End of function 'AVP_Actor_InputHandler::RenderDebugInfo'. ///

#endif	// defined(AVP_ALLOW_DEBUG_MOVEMENT)



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_InputHandler::GetControlMapValue
// Author       : Andy Tate
// Date         : 19-Apr-2007
////////////////////////////////////////////////////////////////////////////////

float AVP_Actor_InputHandler::GetControlMapValue(const u_int uMapping,
												 const bool bAllowConditions /*= true*/,
												 const bool bConsumeKeypress /*= false*/) const
{
	float fRet = 0.0f;
#if defined(ASURA_SUPPORT_FLYCAM)
	AVP_ClientEntity_Actor*		pxClient		= AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);
	AVP_Actor_CameraHandler*	pxCameraHandler	= pxClient->GetCameraHandler();

	// Don't allow controls in fly-cam mode...
	if(!bAllowConditions || !pxCameraHandler || !pxCameraHandler->GetFlyCam())
#endif	// defined(ASURA_SUPPORT_FLYCAM)
	{
		Asura_Assert(m_pxCurrentControlMap != NULL, "Cannot find control-map to use when getting value.");
		if(m_pxCurrentControlMap != NULL)
		{
			fRet = m_pxCurrentControlMap->GetValue(uMapping);
			if ( fRet > 0.0f && bConsumeKeypress )
			{
				m_pxCurrentControlMap->ResetValue( uMapping );
			}
		}
	}
	return fRet;
}

/// End of function 'AVP_Actor_InputHandler::GetControlMapValue'. ///


#if defined(AVP_ALLOW_DEBUG_MOVEMENT)

void AVP_Actor_InputHandler::SetMovementType(MovementType eType)	
{  
	if(eType != s_eMovementType)
	{
		if(AVP_Game_Core::GetPlayer())
		{
			AVP_ClientEntity_Actor*		pxClient			= AVP_ClientEntity_Actor::FindEntity(AVP_Game_Core::GetPlayer()->GetPuppetGuid());
			AVP_Actor_CameraHandler*	pxCameraHandler		= (pxClient == NULL) ? NULL : pxClient->GetCameraHandler();
		
			s_eMovementType = eType; 

			if(pxCameraHandler != NULL)
			{
				switch(s_eMovementType)
				{
				case MT_FLYING_CAMERA:
					{
						pxCameraHandler->SetCameraType(AVP_Actor_CameraHandler::CT_EXTERNAL);
						break;
					}

				case MT_PLAYER_FIRST_PERSON:
					{
						pxCameraHandler->SetCameraType(AVP_Actor_CameraHandler::CT_FIRST_PERSON);
						break;
					}

				case MT_PLAYER_THIRD_PERSON:
					{
						pxCameraHandler->SetCameraType(AVP_Actor_CameraHandler::CT_THIRD_PERSON);
						break;
					}

				case MT_FLYING_PLAYER_COLLISIONS:
				case MT_FLYING_PLAYER_NO_COLLISIONS:
					{
						break;
					}

				default:
					{
						Asura_Assertf( false, ("Invalid switch state %d", s_eMovementType) );
					}
				}		// End of switch.
			}
		}
	}
}

#endif	// defined(AVP_ALLOW_DEBUG_MOVEMENT)

void AVP_Actor_InputHandler::RefreshBlueprintData()
{
	const AVP_ClientEntity_Actor*	pxOwner	= AVP_ClientEntity_Actor::FindEntity(m_uOwnerGuid);

	if(pxOwner != NULL)
	{
		const Asura_Hash_ID		uBlueprintHash	= pxOwner->GetBlueprintHashID();
		const Asura_Hash_ID		uTypeHash		= uAVP_BLUEPRINT_TYPE_PLAYERS;

		const Asura_Blueprint*	pxBlueprint		= Asura_Blueprint_System::GetBlueprintByHashID(uTypeHash, uBlueprintHash);
		if(pxBlueprint)
		{
			Asura_Value	xValue;

			if(pxBlueprint->GetParameterValueByHashID(0x67971619, xValue))	//Asura_GetHashID("InnerAimAssistMultiplier")
			{
				s_fInnerAimAssistMultiplier = xValue.GetFloatValue();
			}
			if(pxBlueprint->GetParameterValueByHashID(0x7e1dffd4, xValue))	//Asura_GetHashID("OuterAimAssistMultiplier")
			{
				s_fOuterAimAssistMultiplier = xValue.GetFloatValue();
			}
			if(pxBlueprint->GetParameterValueByHashID(0x6673400b, xValue))	//Asura_GetHashID("LookAssistAngle")
			{
				s_fLookAssistAngleInDegree = xValue.GetFloatValue();
				s_fTanLookAssistHalfAngleInRadian = Asura_Maths::Tangent(Asura_Maths::DegToRad(s_fLookAssistAngleInDegree) / 2.0f);
			}
			if(pxBlueprint->GetParameterValueByHashID(0x55caf54f, xValue))	//Asura_GetHashID("LookAssistHorizontalMaxSpeed")
			{
				s_fLookAssistHorizontalMaxSpeedInDegree = xValue.GetFloatValue();
				s_fLookAssistHorizontalMaxSpeedInRadian = Asura_Maths::DegToRad(s_fLookAssistHorizontalMaxSpeedInDegree);
			}
			if(pxBlueprint->GetParameterValueByHashID(0x4bdcb0a1, xValue))	//Asura_GetHashID("LookAssistVerticalMaxSpeed")
			{
				s_fLookAssistVerticalMaxSpeedInDegree = xValue.GetFloatValue();
				s_fLookAssistVerticalMaxSpeedInRadian = Asura_Maths::DegToRad(s_fLookAssistVerticalMaxSpeedInDegree);
			}

			if(pxBlueprint->GetParameterValueByHashID(0x75b3c58e, xValue))	//Asura_GetHashID("MouseVSensitivity")
			{
				s_fMouseVSensitivity = xValue.GetFloatValue();
			}
			if(pxBlueprint->GetParameterValueByHashID(0x09fcba5c, xValue))	//Asura_GetHashID("MouseHSensitivity")
			{
				s_fMouseHSensitivity = xValue.GetFloatValue();
			}

			// get the platform dependant variables. 
			Platform_RefreshTurnAndTiltParams( pxBlueprint );
		}
		else
		{
			// Set useable values and assert
			Asura_Assert(false,"Input blueprints could not be read.");

			s_fInnerAimAssistMultiplier = 0.5f;
			s_fOuterAimAssistMultiplier = 1.0f;
			s_fLookAssistAngleInDegree = 10.0f;
			s_fTanLookAssistHalfAngleInRadian = Asura_Maths::Tangent(Asura_Maths::DegToRad(s_fLookAssistAngleInDegree) / 2.0f);
			s_fLookAssistHorizontalMaxSpeedInDegree = 20.0f;
			s_fLookAssistHorizontalMaxSpeedInRadian = Asura_Maths::DegToRad(s_fLookAssistHorizontalMaxSpeedInDegree);
			s_fLookAssistVerticalMaxSpeedInDegree = 20.0f;
			s_fLookAssistVerticalMaxSpeedInRadian = Asura_Maths::DegToRad(s_fLookAssistVerticalMaxSpeedInDegree);

			s_fTiltMaxSpeedInDegree = 90.0f;
			s_fTiltMaxSpeedInRadian = Asura_Maths::DegToRad(s_fTiltMaxSpeedInDegree);
			s_fTiltRampUpTime = 0.25f;
			s_fTiltRampDownTime = 0.05f;
			s_fTiltExponent = 2.0f;
			s_fMouseVSensitivity = 1.0f;

			s_fTurnDefaultMaxSpeedInDegree = 180.0f;
			s_fTurnDefaultMaxSpeedInRadian = Asura_Maths::DegToRad(s_fTurnDefaultMaxSpeedInDegree);
			s_fTurnTurboMaxSpeedInDegree = 180.0f;
			s_fTurnTurboMaxSpeedInRadian = Asura_Maths::DegToRad(s_fTurnTurboMaxSpeedInDegree);
			s_fTurnTurboDelay = 1.0f;
			s_fTurnTurboRampUpTime = 0.5f;
			s_fTurnRampUpTime = 0.5f;
			s_fTurnRampDownTime = 0.10f;
			s_fTurnExponent = 2.0f;
			s_fMouseHSensitivity = 1.0f;
		}
	}
}

// void AVP_Actor_InputHandler::SetGrappleControls(bool bSet) 
// {
// 	if(bSet)
// 	{
// 		m_pxCurrentControlMap	= Asura_ControlManager::PushDefaultControls(m_uOwnerGuid, &m_xMapGrapple);
// 	}
// 	else
// 	{
// 		m_pxCurrentControlMap	= Asura_ControlManager::PopDefaultControls(m_uOwnerGuid);
// 	}
// }

void AVP_Actor_InputHandler::PushControls( Asura_ControlMap* pxControlMap )
{
	m_pxCurrentControlMap	= Asura_ControlManager::PushDefaultControls(m_uOwnerGuid, pxControlMap);
}

void AVP_Actor_InputHandler::PopControls() 
{
	m_pxCurrentControlMap	= Asura_ControlManager::PopDefaultControls(m_uOwnerGuid);
}

bool AVP_Actor_InputHandler::IsIdle() const
{
	bool bIdle = true;
	bIdle &= GetControlMapValue( LOOK_UP_DOWN ) == 0.0f;
	bIdle &= GetControlMapValue( LOOK_LEFT_RIGHT ) == 0.0f;
	bIdle &= GetControlMapValue( LOOK_MOUSE_UP_DOWN ) == 0.0f;
	bIdle &= GetControlMapValue( LOOK_MOUSE_LEFT_RIGHT ) == 0.0f;
	bIdle &= GetControlMapValue( FORWARDS_BACKWARDS ) == 0.0f;
	bIdle &= GetControlMapValue( STRAFE_LEFT_RIGHT ) == 0.0f;
	return bIdle;
}

/// End of file. ///
