
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Actor_PhysicsObjectHandler.cpp
// Author : Andy Tate
// Date   : 10-May-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/AVP_Actor_PhysicsObjectHandler.h"

// Asura independent header-files...
#include "Core/Asura_LOS.h"
#include "Entity/Asura_PosAndOriInterpolator.h"
#include "Physics/Asura_Physics.h"
#include "Physics/Asura_Physics_Object.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_InputHandler.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ClientActor_Vertebrate.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ServerActor_Vertebrate.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float	dbg_fActorOnGroundLOSDistance	= 0.100f;


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::AVP_Actor_PhysicsObjectHandler
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_PhysicsObjectHandler::AVP_Actor_PhysicsObjectHandler(const Asura_Guid uOwnerGuid,
															   const Type eType,
															   const Asura_Collision_Object& xCollisionObject,
															   const Asura_Physics_Object::COLLISION_PROXY eCollisionProxy,
															   const u_short usFlagsToIgnore)
: PARENT(NULL)
, m_uOwnerGuid(uOwnerGuid)
, m_usFlagsToIgnore(usFlagsToIgnore)
{
	// Create the physics object...
	Reset(eType, xCollisionObject, eCollisionProxy);
	AVP_Actor_PhysicsObjectHandler::Initialise(m_pxPhysicsObject);
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::AVP_Actor_PhysicsObjectHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::~AVP_Actor_PhysicsObjectHandler
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_PhysicsObjectHandler::~AVP_Actor_PhysicsObjectHandler()
{
	// Simply remove the existing physics object...
	Asura_Physics_Object::Destroy(m_pxPhysicsObject);
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::~AVP_Actor_PhysicsObjectHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::CalculateGroundType
// Author       : Andy Tate
// Date         : 3-Oct-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_PhysicsObjectHandler::GroundType AVP_Actor_PhysicsObjectHandler::CalculateGroundType(const Asura_AVector3& xNormal)
{
	const Asura_AVector3	xUp(0.0f, -1.0f, 0.0f);
	const float				fDotProduct	= xNormal * xUp;

	// Floor - where are we jumping to?
	if(fDotProduct > 0.707f)
	{
		return(ON_FLOOR);
	}
	else if (fDotProduct < -0.707f)
	{
		return(ON_CEILING);
	}
	else
	{
		return(ON_WALL);
	}
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::CalculateGroundType'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::CalculateGroundType
// Author       : Andy Tate
// Date         : 3-Oct-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_PhysicsObjectHandler::GroundType AVP_Actor_PhysicsObjectHandler::CalculateGroundType(const Asura_Matrix_3x3& xOrientation)
{
	const Asura_AVector3	xDirection(-xOrientation.e12, -xOrientation.e22, -xOrientation.e32);

	return(CalculateGroundType(xDirection));
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::CalculateGroundType'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::PerformOnGroundCheck
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

AVP_Actor_PhysicsObjectHandler::GroundType AVP_Actor_PhysicsObjectHandler::PerformOnGroundCheck(void) const
{
	const Asura_Matrix_3x3&	xCurrentOrientation	= m_pxPhysicsObject->GetOrientation();

	if(m_pxPhysicsObject->IsSupported() || !m_pxPhysicsObject->FeelsGravity())
	{
		return(CalculateGroundType(xCurrentOrientation));
	}
	else
	{
		const Asura_Vector_3&		xCurrentPosition	= m_pxPhysicsObject->GetPosition();

		const Asura_Vector_3		xDirection(xCurrentOrientation.e12, xCurrentOrientation.e22, xCurrentOrientation.e32);
		const Asura_LOS_InputData	xLOSInput(xCurrentPosition - xDirection * dbg_fActorOnGroundLOSDistance, xDirection, m_uOwnerGuid, 2.0f * dbg_fActorOnGroundLOSDistance);

		const Asura_LOS_OutputData*	pxLOSOutput	= Asura_LOS::TestAgainstEverything(&xLOSInput);

		// Did we hit a surface?
		if(pxLOSOutput->m_fDistanceToIntersection < 2.0f * dbg_fActorOnGroundLOSDistance)
		{
			return(CalculateGroundType(-xDirection));
		}
		else
		{
			return(OFF_GROUND);
		}
	}
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::PerformOnGroundCheck'. ///


Asura_Physics_Object::COLLISION_PROXY AVP_Actor_PhysicsObjectHandler::GetActorCollisionProxy(u_short usActorClassification)
{
	switch (usActorClassification)
	{
	default:
		Asura_Assert(false, "Unexpected actor classification, will use an OBB collision proxy.");
		return Asura_Physics_Object::COLLISION_PROXY_OBB;

	case AVP_ENTITYCLASS_ALIEN:
	case AVP_ENTITYCLASS_ALIENFACEHUGGER:
	case AVP_ENTITYCLASS_ALIEN_HYBRID:
		return Asura_Physics_Object::COLLISION_PROXY_OBB;

	case AVP_ENTITYCLASS_DROPSHIP:
	case AVP_ENTITYCLASS_WEAPON:
	case AVP_ENTITYCLASS_SENTRYGUN:
		return Asura_Physics_Object::COLLISION_PROXY_OBB;

	case AVP_ENTITYCLASS_ALIENQUEEN:
	case AVP_ENTITYCLASS_PREDATOR:
	case AVP_ENTITYCLASS_ANDROID:
	case AVP_ENTITYCLASS_CIVILIAN:
	case AVP_ENTITYCLASS_MARINE:
	case AVP_ENTITYCLASS_PREDALIEN:
		return Asura_Physics_Object::COLLISION_PROXY_SPHERESTACK;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::UpdateFromInterpolator
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_PhysicsObjectHandler::UpdateFromInterpolator(const Asura_PosAndOriInterpolator& xInterpolator)
{
	Asura_Vector_3		xCurrentPosition;
	Asura_Vector_3		xVelocity;
	Asura_Matrix_3x3	xOrientation;

	xInterpolator.GetCurrentPosition(xCurrentPosition, xVelocity);
	xInterpolator.GetCurrentOrientation(xOrientation);

	m_pxPhysicsObject->SetPosition(xCurrentPosition);
	m_pxPhysicsObject->SetOrientation(xOrientation);
	m_pxPhysicsObject->SetLinVelocity(xVelocity);
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::UpdateFromInterpolator'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::UpdateInterpolator
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_PhysicsObjectHandler::UpdateInterpolator(Asura_PosAndOriInterpolator& xInterpolator) const
{
	const float	fTimeStamp		= Asura_Timers::GetServerAbsoluteTime();
	const float	fTimeDifference	= fTimeStamp - m_fTimeOfLastInterpolatorData;
	const float	fTimeSlice		= Asura_Physics_System::GetTimeSlice();

	if(fTimeDifference >= fTimeSlice)
	{
		Asura_Quat		xCurrentOrientation;
		Asura_Vector_3	xCurrentPosition;
		Asura_Vector_3	xCurrentVelocity;

		// Get the data from the physics-object and pass it to the interpolator...
		xCurrentOrientation.FromMatrix(m_pxPhysicsObject->GetOrientation());

		xCurrentPosition	= m_pxPhysicsObject->GetPosition();
		xCurrentVelocity	= m_pxPhysicsObject->GetLinVelocity();

		xInterpolator.SetProcessNewDataIncrement(fTimeSlice * 2.0f);
		xInterpolator.ProcessNewData(xCurrentPosition, xCurrentOrientation, fTimeStamp, &xCurrentVelocity);

		m_fTimeOfLastInterpolatorData	= fTimeStamp;
	}
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::UpdateInterpolator'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::ResetUsingWorldSpaceOrientation
// Author       : Andy Tate
// Date         : 13-Sep-2007
////////////////////////////////////////////////////////////////////////////////


/// End of function 'AVP_Actor_PhysicsObjectHandler::ResetUsingWorldSpaceOrientation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Actor_PhysicsObjectHandler::Reset
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_Actor_PhysicsObjectHandler::Reset(const Type eType,
										   const Asura_Collision_Object& xCollisionObject,
										   const Asura_Physics_Object::COLLISION_PROXY eCollisionProxy)
{
	Asura_Matrix_3x3	xOrientation;

	// Ensure we've got a physics object...
	if(m_pxPhysicsObject == NULL)
	{
		m_pxPhysicsObject	= (eType == REMOTE_CLIENT || eType == STATIC_AI)
							? Asura_Physics_Object::Create(ASURA_PHYSICS_OBJECT_EXTERNALCONTROL)
							: Asura_Physics_Object::Create(ASURA_PHYSICS_OBJECT_ACTOR);

		m_pxPhysicsObject->SetFlagsOfPolysToIgnore(m_pxPhysicsObject->GetFlagsOfPolysToIgnore() | m_usFlagsToIgnore);

		xOrientation.Identity();
	}
	else
	{
		switch(eType)
		{
			case GAMESCENE_AI:
			case FREE_AI:
			case PLAYER:
			{
				m_pxPhysicsObject->ChangeType(ASURA_PHYSICS_OBJECT_ACTOR);
				break;
			}

			case REMOTE_CLIENT:
			case STATIC_AI:
			{
				m_pxPhysicsObject->ChangeType(ASURA_PHYSICS_OBJECT_EXTERNALCONTROL);
				break;
			}

			default:
			{
				Asura_Assert(false, "Unexpected type for physics-object handler.");
				break;
			}
		}		// End of switch.

		xOrientation	= m_pxPhysicsObject->GetOrientation();
	}

	m_eDefaultPhysicsObjectType = m_pxPhysicsObject ? m_pxPhysicsObject->GetType() : ASURA_PHYSICS_OBJECT_EXTERNALCONTROL;

	// Reset the variables...
	m_fTimeOfLastInterpolatorData	= 0.0f;

	// Reset the physics object...
	m_pxPhysicsObject->SetIgnore(true);
	m_pxPhysicsObject->SetAxisConstraint(Asura_Physics_Object::AXIS_CONSTRAINT_NONE);
	m_pxPhysicsObject->SetCollisionObject(&xCollisionObject);
	m_pxPhysicsObject->SetCollisionProxy(eCollisionProxy);
	m_pxPhysicsObject->SetAssociatedEntityGuid(m_uOwnerGuid);
	m_pxPhysicsObject->SetActorsShouldSlideOffMe(true);
}

/// End of function 'AVP_Actor_PhysicsObjectHandler::Reset'. ///


/// End of file. ///
