
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Vertebrate_Focus.cpp
// Author : Anthony Gregson
// Date   : 31/01/08
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/Vertebrates/AVP_Vertebrate_Focus.h"

// Asura
#include "Attractors/Asura_Attractor_System.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Interface/Asura_CommandConsole.h"
#include "Rendering/Asura_Render.h"

// AVP
#include "Actor/Vertebrates/AVP_Vertebrate_HUDHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_InputHandler.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ClientActor_Vertebrate.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Moves/AVP_Moves.h"

float AVP_Vertebrate_Focus::s_fToggleAutomaticTimeout = 1.5f;
bool AVP_Vertebrate_Focus::s_bOnlyUseToggle = false;
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
int AVP_Vertebrate_Focus::s_iConsoleVarsCount	= 0;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

AVP_Vertebrate_Focus::AVP_Vertebrate_Focus(AVP_ClientActor_Vertebrate& xOwner)
: m_xOwner(xOwner)
, m_uTargetEnemyGuid( ASURA_GUID_UNREGISTERED )
, m_uBestTargetGuid( ASURA_GUID_UNREGISTERED )
, m_fTimeActive(0.0f)
, m_uFlags(0)
, m_uNumPounceAttackVariations(0)
, m_pxPounceAttacks(NULL)
{
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	if (s_iConsoleVarsCount == 0)
	{
		if (!Asura_CommandConsole::VarExists(L"AVP.Focus.AutomaticTimeoutTime"))
		{
			Asura_CommandConsole::AddVar(L"AVP.Focus.AutomaticTimeoutTime", &s_fToggleAutomaticTimeout, 0.0f, 10.0f);
		}
		if (!Asura_CommandConsole::VarExists(L"AVP.Focus.OnlyUseToggle"))
		{
			Asura_CommandConsole::AddVar(L"AVP.Focus.OnlyUseToggle", &s_bOnlyUseToggle);
		}
	}
	++s_iConsoleVarsCount;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}

AVP_Vertebrate_Focus::~AVP_Vertebrate_Focus()
{
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	--s_iConsoleVarsCount;
	if (s_iConsoleVarsCount == 0)
	{
		if (Asura_CommandConsole::VarExists(L"AVP.Focus.AutomaticTimeoutTime"))
		{
			Asura_CommandConsole::RemoveVar(L"AVP.Focus.AutomaticTimeoutTime");
		}
		if (Asura_CommandConsole::VarExists(L"AVP.Focus.OnlyUseToggle"))
		{
			Asura_CommandConsole::RemoveVar(L"AVP.Focus.OnlyUseToggle");
		}
	}
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE

	delete[] m_pxPounceAttacks;
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Vertebrate_Focus::Activate
// Author       : Andy Tate
// Date         : 1-Jul-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Vertebrate_Focus::Activate(void)
{
	if(!m_bActive)
	{
		AVP_Vertebrate_HUDHandler*	pxHUDHandler	= static_cast<AVP_Vertebrate_HUDHandler*>(m_xOwner.GetHUDHandler());

		m_bActive	= true;

		if(pxHUDHandler != NULL)
		{
			pxHUDHandler->OnBeginFocus();
		}
	}
}

/// End of function 'AVP_Vertebrate_Focus::Activate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Vertebrate_Focus::Deactivate
// Author       : Andy Tate
// Date         : 1-Jul-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Vertebrate_Focus::Deactivate(void)
{
	if(m_bActive)
	{
		AVP_Vertebrate_HUDHandler*	pxHUDHandler	= static_cast<AVP_Vertebrate_HUDHandler*>(m_xOwner.GetHUDHandler());

		m_bActive	= false;

		if(pxHUDHandler != NULL)
		{
			pxHUDHandler->OnEndFocus();
		}
	}
}

/// End of function 'AVP_Vertebrate_Focus::Deactivate'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Vertebrate_Focus::Update
// Author       : Anthony Gregson
// Date         : 31/01/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Vertebrate_Focus::Update()
{
	const AVP_Vertebrate_InputHandler* pxInput = static_cast<const AVP_Vertebrate_InputHandler*>(m_xOwner.GetInputHandler());
	if (!pxInput)
	{
		Deactivate();
		return;
	}

	if (Asura_Render::GetCutsceneHasCameraControl())
	{
		Deactivate();
		return;
	}

	// Some moves can stop use of focus.
	const Asura_Movement_System* pxMovementSystem = m_xOwner.GetMovementSystem();
	const AVP_Move* pxCurrentMove = (pxMovementSystem && pxMovementSystem->IsMoveInProgress()) ? static_cast<const AVP_Move*>(pxMovementSystem->GetCurrentMove()) : NULL;
	if (pxCurrentMove && pxCurrentMove->GetPreventsFocus())
	{
		Deactivate();
		return;
	}

	// Certain items block focus too.
	const AVP_Inventory_System* pxInventory = m_xOwner.GetInventorySystem();
	const AVP_Inventory_Item* pxStim = pxInventory ? pxInventory->GetItem(uAVP_BLUEPRINT_PARAM_STIM) : NULL;
	if (pxStim && !pxStim->IsIdling(true))
	{
		Deactivate();
		return ;
	}

	bool bWasActive = m_bActive;

	if (s_bOnlyUseToggle)
	{
		if (pxInput->GetControlMapValue(AVP_Vertebrate_InputHandler::FOCUS_TOGGLE) > 0.0f)
		{
			if ( m_bActive )
			{
				Deactivate();
			}
			else
			{
				Activate();
			}
		}
	}
	else
	{
		if (pxInput->GetControlMapValue(AVP_Vertebrate_InputHandler::FOCUS_HELD) > 0.0f)
		{
			Activate();
			// Reset timer if tapped again.
			bWasActive &= (pxInput->GetControlMapValue(AVP_Vertebrate_InputHandler::FOCUS_TOGGLE) <= 0.0f);
		}
		else if (m_fTimeActive >= s_fToggleAutomaticTimeout)
		{
			Deactivate();
		}
	}

	if (m_bActive)
	{
		if (!bWasActive)
		{
			m_fTimeActive = 0.0f;
		}
		else
		{
			m_fTimeActive += Asura_Timers::GetGameTimeSlice();
		}
	}
}

/// End of function 'AVP_Vertebrate_Focus::Update'. ///

void AVP_Vertebrate_Focus::SetPounceAttackVariationData(u_int uNumVariations, float* pfMaxDistances, float* pfMinDrops)
{
	Asura_Assert(m_uNumPounceAttackVariations == 0, "Allocating pounce attack data twice, should be unnecessary");
	if (m_uNumPounceAttackVariations > 0)
	{
		delete[] m_pxPounceAttacks;
	}

	m_uNumPounceAttackVariations = uNumVariations;

	m_pxPounceAttacks = new PounceAttackData[uNumVariations];

	for (u_int u=0; u<uNumVariations; ++u)
	{
		m_pxPounceAttacks[u].m_fMaxDistance = pfMaxDistances[u];
		m_pxPounceAttacks[u].m_fMinDrop = pfMinDrops[u];
	}
}

bool AVP_Vertebrate_Focus::GetCanPounceToTarget(const AVP_ClientEntity_Actor& xTarget) const
{
	Asura_Vector_3 xOwnerOrigin;
	m_xOwner.GetMeleeOrigin(xOwnerOrigin);
	Asura_Vector_3 xOffset;
	xTarget.GetMeleeOrigin(xOffset);
	xOffset -= xOwnerOrigin;
	const float fDistSquared = xOffset.MagnitudeSquared();
	const float fDrop = xOffset.y; // Intentionally allowing negative as most pounces have a minimum that lets you jump up.

	for (u_int u=0; u<m_uNumPounceAttackVariations; ++u)
	{
		const float fCurrentDistance = m_pxPounceAttacks[u].m_fMaxDistance;

		if ((fDrop >= m_pxPounceAttacks[u].m_fMinDrop) && (fDistSquared <= (fCurrentDistance * fCurrentDistance)))
		{
			// We just need to return that at least one pounce is valid for this to succeed.
			return true;
		}
	}

	return false;
}

/// End of file. ///
