
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Alien_InputHandler.cpp
// Author : Chris Wallace
// Date   : 5-Feb-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/Vertebrates/Alien/AVP_Alien_InputHandler.h"

// Asura independent header-files...
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Core/Asura_BitPacker.h"
#include "Core/Asura_Timers.h"
#include "Input/Asura_ControlMapping_Combination.h"
#include "Input/Asura_ControlMapping_Key.h"
#include "Interface/Asura_CommandConsole.h"
#include "Mesh/Asura_Mesh_PolygonFlags.h"
#include "Network/Asura_Network.h"
#include "ProjectSettings/Asura_PhysicsConfiguration.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Primitive_Render.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_CameraHandler.h"
#include "Actor/AVP_Actor_HUDHandler.h"
#include "Actor/AVP_Actor_InteractionHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Focus.h"
#include "Actor/Vertebrates/Alien/AVP_Alien_CameraHandler.h"
#include "Actor/Vertebrates/Alien/AVP_Alien_MeleeCounter_System.h"
#include "Attractors/AVP_Attractors_System.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_Plane.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Alien/AVP_ClientActor_Alien.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Alien/AVP_Messages_Alien.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Mesh/AVP_PolygonFlags.h"
#include "Moves/AVP_Movement_System.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesBranch.h"
#include "Moves/Vertebrates/Alien/AVP_Alien_MovesJump.h"
#include "Physics/AVP_CalculateSurfaceReorientationNormal.h"

//#define LOG(params) Asura_OutputToDebugger_Retail(params)
//#define LOG2(params) Asura_OutputToDebugger_Retail(params)
#define LOG(params)
#define LOG2(params)

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float dbg_fCosStandMoveTriggerAngle = 0.17f; // ~ cos ( 80 )

DVSVAR bool dvs_bEnableAlienBranchEnvHooks = false;
DVSVAR bool dvs_bEnableAlienDuctEnvHooks = true;
DVSVAR bool dvs_bUsePushToTransition = true;
DVSVAR bool dvs_bUseLeapEnvHooks = true;
DVSVAR bool dvs_bLeapEnvHooksNeedButtonPress = false;
DVSVAR bool dvs_bEnableLookPrediction = true;

// TODO all these angles need hard coding as cosines once we're happy with the values.
DVSVAR float dvs_fPushTransAngle = 130.0f;
DVSVAR float dvs_fNoPushTransAngle = 50.0f;
DVSVAR float dvs_fAcceptAttractorNormalAngle = 45.0f;
DVSVAR float dvs_fAcceptAttractorDirectionAngle = 45.0f;
DVSVAR float dvs_fAcceptAttractorLookAngle = 45.0f;
DVSVAR float dvs_fAcceptAttractorConeAngle = 30.0f;

DVSVAR float dvs_fMinLeapAttractorDist = 1.0f; // ignore attractors closer that 0.5m
DVSVAR float dvs_fLookAheadAttractorTime = 1.0f; // look 1 second ahead for attractors


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////

void AVP_Alien_InputHandler::Reset( void )
{
	const AVP_ClientActor_Alien* pxAlien = AVP_ClientActor_Alien::FindEntity( m_uOwnerGuid );
	if ( pxAlien )
	{
		const Asura_Matrix_3x3& xOrientation = pxAlien->GetCurrentOrientation();
		m_xLastUpVector = Asura_AVector3( -xOrientation.e12, -xOrientation.e22, -xOrientation.e32 );
	}
	m_fLookSmoothingTimer = 0.0f;

	PARENT::Reset();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_InputHandler::MapTailAttack
// Author       : Anthony Gregson
// Date         : 18/03/08
////////////////////////////////////////////////////////////////////////////////
 
void AVP_Alien_InputHandler::MapTailAttack(Asura_ControlMap& xMap, const u_int uKeyTail)
{
	xMap.AddKeyMapping(TAIL_ATTACK_TAP, uKeyTail, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
	xMap.AddKeyMapping(TAIL_ATTACK_HELD, uKeyTail, Asura_ControlMapping::ON_PRESS);
	Asura_ControlMapping* pxReleaseMapping = new Asura_ControlMapping_Key(uKeyTail, Asura_ControlMapping::ON_DEBOUNCED_RELEASE, 1.0f);
	pxReleaseMapping->SetPauseWhileInactive(true);
	xMap.StoreMapping(TAIL_ATTACK_RELEASE, pxReleaseMapping);
}

/// End of function 'AVP_Alien_InputHandler::MapTailAttack'. ///


void AVP_Alien_InputHandler::MapPetFacehuggerControl(Asura_ControlMap& xMap, const u_int uKeySpawn)
{
	xMap.AddKeyMapping(PET_FACEHUGGER_CONTROL, uKeySpawn, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
}


DEBUGVAR float dbg_fInteractOrSpawnHoldTime = 0.5f;

void AVP_Alien_InputHandler::MapInteractAndPetFacehuggerControlToSameKey(Asura_ControlMap& xMap, const u_int uKeyInteractOrSpawn)
{
	xMap.DeleteMapping(PET_FACEHUGGER_CONTROL);
	xMap.DeleteMapping(INTERACT);

	xMap.AddTapKeyMapping(PET_FACEHUGGER_CONTROL, uKeyInteractOrSpawn, dbg_fInteractOrSpawnHoldTime);
	xMap.AddHoldKeyMapping(INTERACT, uKeyInteractOrSpawn, dbg_fInteractOrSpawnHoldTime, 1.0f, Asura_ControlMapping_HoldKey::NO_RAMPING, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
}


void AVP_Alien_InputHandler::MapCreateDistraction(Asura_ControlMap& xMap, const u_int uKeyDistract)
{
	xMap.AddKeyMapping(CREATE_DISTRACTION, uKeyDistract, Asura_ControlMapping::ON_DEBOUNCED_PRESS);
}

void AVP_Alien_InputHandler::MapTransition( Asura_ControlMap& xMap, const u_int uKeyTransition )
{
	xMap.AddKeyMapping(TRANSITION, uKeyTransition, Asura_ControlMapping::ON_PRESS );
}

////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float dbg_fAlienBranchMinNormalCosAngleForTraverse = -0.5f;
DEBUGVAR float dbg_fAlienBranchMinNormalCosAngleForEasyTraverse = 0.8f;
DEBUGVAR float dbg_fAlienBranchMaxXInputForEasyTraverse = 0.86f;
DEBUGVAR float dbg_fAlienBranchMinCosAngleForJunctionTransition = 0.85f;
DEBUGVAR float dbg_fAlienBranchQuickAttachDistanceSquared = 1.0f;
DEBUGVAR float dbg_fAlienBranchQuickAttachOrientationSimilarity = 0.1f;
DEBUGVAR float dbg_fAlienBranchApproachingAttractorEndPositionRange = 0.8f;
DEBUGVAR float dbg_fAlienBranchApproachingAttractorEndLOSRange = 0.5f;
DEBUGVAR float dbg_fAlienBranchNearbyEndNodePositionRange = 3.0f;
DEBUGVAR float dbg_fAlienBranchBestNormalDPAllowedReduction = 0.2f;
DEBUGVAR float dbg_fAlienBranchTraverseMinRightVectorSimilarity = -0.5f;
DEBUGVAR float dbg_fAlienBranchLOSTestOffset = 0.5f;
DEBUGVAR float dbg_fAlienBranchLOSTestNearWallLeeway = 0.1f;

void AVP_Alien_InputHandler::UpdateMovementInput(AVP_ClientEntity_Actor &xClient)
{
	AVP_ClientActor_Alien& xAlien = static_cast<AVP_ClientActor_Alien&>(xClient);

	// do this before parent update as that's where sprinting stuff is calculated
	// if we're about to stand up reset our sprinting condition
	if ( GetControlMapValue( STANCE_TOGGLE ) > 0.1f && !xAlien.IsStanding() )
	{
		if ( m_fSprintTimer > 0.0f )
		{ 
			m_fSprintTimer = 0.0f;
			m_fSprintChargeTimer = m_fSprintChargeTime;
		}
	}

	// First, get our input values up to date.
	PARENT::UpdateMovementInput(xClient);

	Asura_Vector_3 xGravityDirection;
	if (xAlien.GetPhysicsObjectHandler())
	{
		xAlien.GetPhysicsObjectHandler()->GetCurrentGravityDirection(xGravityDirection);
	}
	else
	{
		xGravityDirection = Asura_Vector_3(0.0f, 1.0f, 0.0f);
	}

	Asura_Vector_3 xVel;
	xAlien.GetCurrentVelocity(xVel);
	const float fSpeedInGravityDirection = xVel * xGravityDirection;
	const bool bJumping = xAlien.IsJumping() || ( ( xAlien.GetGroundType() == AVP_Actor_PhysicsObjectHandler::OFF_GROUND ) && ( fSpeedInGravityDirection < 0.01f ) );

	Asura_Movement_System* pxMovementSystem = xAlien.GetMovementSystem();
	const Asura_Move* pxCurrentMove = pxMovementSystem != NULL ? pxMovementSystem->GetCurrentMove() : NULL;
	if ( pxCurrentMove != NULL && pxCurrentMove->GetHash() == AVP_Move_Alien_DismountBranch::GetStaticHash() )
	{
		m_xMovementInput = static_cast< const AVP_Move_Alien_DismountBranch* >( pxCurrentMove )->GetMovementDirection();
	}
	else if ( !bJumping && pxCurrentMove == NULL )
	{
		// Now see if we want to connect to/disconnect from a branch attractor.
		const Asura_Vector_3& xCurrentPosition = xAlien.GetCurrentPosition();
		const Asura_Matrix_3x3& xCurrentOri = xAlien.GetCurrentOrientation();
		const Asura_Vector_3 xForward(xCurrentOri.e13, xCurrentOri.e23, xCurrentOri.e33);
		Asura_Vector_3 xLosDir( m_xMovementInput );
		const bool bNotMoving = xLosDir.ApproximatelyEquals(Asura_Maths::ZeroVector3);
		if (bNotMoving)
		{
			// Fire the LOS anyway to see if we're inside something already, since we may fall onto a branch. Use gravity direction for the alien.
			xLosDir = xGravityDirection;
		}
		else
		{
			Asura_AVector3 xAim;
			xAlien.GetAimDirection(xAim);
			Asura_Vector_3 xAimDir(xAim);
			// Make perpendicular to the gravity direction.
			xAimDir -= xGravityDirection * (xGravityDirection * xAimDir);
			Asura_Matrix_3x3 xLosOri;
			if (xAimDir.ApproximatelyEquals(Asura_Maths::ZeroVector3))
			{
				xLosOri = xCurrentOri;
			}
			else
			{
				Asura_Maths::OrientationFromDirection(xLosOri, xAimDir); // This normalises the vector internally.
			}

			xLosDir = xLosDir * xLosOri;
			xLosDir.Normalise();
		}

		Asura_Matrix_3x3 xLosOri;
		Asura_Maths::OrientationFromDirection(xLosOri, xLosDir);

		Asura_Vector_3 xCurrentAttractorNormal;
		if (xAlien.GetPhysicsObjectHandler())
		{
			xAlien.GetPhysicsObjectHandler()->GetCurrentSurfaceNormal(xCurrentAttractorNormal);
		}
		else
		{
			xCurrentAttractorNormal = Asura_Vector_3(0.0f, -1.0f, 0.0f);
		}
		const Asura_Attractor_Compiled* pxCurrentAttractor = Asura_Attractor_System::GetInstance()->GetAttractorByID(xAlien.GetConstrainedMovementAttractorID());
		bool bEasyTraversalOnly = true;
		bool bApproachingAttractorEnd = false;
		bool bAllowJunctionTransition = true;
		if (pxCurrentAttractor)
		{
			pxCurrentAttractor->GetNormal(xCurrentAttractorNormal);

			if (!bNotMoving)
			{
				// Easy traversal is only relevant to branches.
				const Asura_Hash_ID uType = pxCurrentAttractor->GetType();
				if (uType == uAVP_ATTRACTOR_TYPE_BRANCH)
				{
					// Easy traversal means we're pushing roughly along the branch, and changes the test so we can't transition to attractors with a significantly different normal to the current one.
					// Now changed to the raw movement input, so you have to strafe to rotate around the branch.
					bEasyTraversalOnly = Asura_Maths::Abs(m_xMovementInput.x) < dbg_fAlienBranchMaxXInputForEasyTraverse;
				}

				// Attractor end is for both though
				Asura_Vector_3 xAxisDirection;
				pxCurrentAttractor->GetAxisDirection(xAxisDirection);

				const int iNodeIndex = ((xAxisDirection * xLosDir) < 0.0f) ? 0 : 1;
				const Asura_Vector_3 xOffset = pxCurrentAttractor->GetNode(iNodeIndex)->GetPosition() - xCurrentPosition;
				bApproachingAttractorEnd = xOffset.MagnitudeSquared() < (dbg_fAlienBranchApproachingAttractorEndPositionRange * dbg_fAlienBranchApproachingAttractorEndPositionRange);

				// Further to the above, do a quick LOS test to see if we're about to run into a wall.
				if (!bApproachingAttractorEnd && !xLosDir.ApproximatelyEquals(xGravityDirection) && (Asura_Maths::Abs(xLosDir * xForward) > 0.00001f))
				{
					Asura_Vector_3 xHeadPos;
					xAlien.GetHeadPosition(xHeadPos);
					Asura_LOS_InputData xInput(xHeadPos, xForward * Asura_Maths::Sign(xLosDir * xForward), xAlien.GetGuid(), dbg_fAlienBranchApproachingAttractorEndLOSRange, false, xAlien.GetFlagsToIgnore() | AsuraPolygonFlag_NoCollisions);
					const Asura_LOS_OutputData* pxOutput = Asura_LOS::TestAgainstEverything(&xInput);
					bApproachingAttractorEnd = pxOutput->m_fDistanceToIntersection < dbg_fAlienBranchApproachingAttractorEndLOSRange;
				}

				// As is allowing junction transitions.
				bAllowJunctionTransition = Asura_Maths::Abs(xAxisDirection * xLosDir) <= dbg_fAlienBranchMinCosAngleForJunctionTransition;
			}
		}

		BranchCallbackData xCallbackData(*this, xAlien, xLosDir, xGravityDirection, xForward, xCurrentAttractorNormal, bEasyTraversalOnly, bApproachingAttractorEnd, bAllowJunctionTransition);

		// TODO - May need to do this from multiple positions so we can do some kind of reorientation between attractors that meet
		Asura_Attractor_System::GetInstance()->CollectAttractors(xCurrentPosition, xLosOri, 0.0f, m_fAlienBranchLOSDistance, TestBranchAttractors_Callback, &xCallbackData);

		// See if we found a new attractor
		if (xCallbackData.m_uBestAttractor != xAlien.GetConstrainedMovementAttractorID())
		{
			if (xCallbackData.m_uBestAttractor == uASURA_ATTRACTOR_ID_NONE)
			{
				// Just disconnect in this case, moves seem unnecessary if we've lost our current attractor.
				if (xAlien.GetPhysicsObjectHandler())
				{
					xAlien.GetPhysicsObjectHandler()->UnsetForcedTargetSurfaceNormal();
				}
				pxCurrentAttractor = NULL;
			}
			else
			{
				// Change to current attractor, so update the pointer
				pxCurrentAttractor = Asura_Attractor_System::GetInstance()->GetAttractorByID(xCallbackData.m_uBestAttractor);
				if (pxCurrentAttractor)
				{
					const Asura_Vector_3 xOffset = xCallbackData.m_xBestPosOnAttractor - xCurrentPosition;
					Asura_Vector_3 xNewSurfaceNormal;
					pxCurrentAttractor->GetNormal(xNewSurfaceNormal);
					// The current attractor normal is always valid, as it's the combined surface normal without one.
					const float fNormalsDP = xNewSurfaceNormal * xCurrentAttractorNormal;
					const Asura_Quat xCurrentQuat(xCurrentOri);

					if ((xOffset.MagnitudeSquared() > dbg_fAlienBranchQuickAttachDistanceSquared) || (fNormalsDP < dbg_fAlienBranchMinNormalCosAngleForEasyTraverse) || !xCallbackData.m_xTargetOrientation.ApproximatelyEquals(xCurrentQuat, dbg_fAlienBranchQuickAttachOrientationSimilarity))
					{
						SendMountBranchMoveMessage(xCallbackData.m_xBestPosOnAttractor, xCallbackData.m_xTargetOrientation, xAlien);
						// The move handles setting gravity direction
					}
					else
					{
						// Just go straight to it in this case, so the gravity direction needs to be set here
						if (xAlien.GetPhysicsObjectHandler())
						{
							xAlien.GetPhysicsObjectHandler()->SetForcedTargetSurfaceNormal(xNewSurfaceNormal);
						}
					}
				}
			}
			xAlien.SetConstrainedMovementAttractorID(xCallbackData.m_uBestAttractor);
			m_fBranchDismountInputTimer = 0.0f;
		}
		else
		{
			// Are we pushing roughly perpendicular to the branch?
			// TODO - we should be linking up the jump control to the dismount move.
			if (false)//!s_bOnlyLeaveBranchByJump && (xAlien.GetConstrainedMovementAttractorID() != uASURA_ATTRACTOR_ID_NONE) && !xLosDir.ApproximatelyEquals(xGravityDirection) && (xCallbackData.m_fBestDPResult < m_fAlienBranchMaxCosAngleForDismount))
			{
				m_fBranchDismountInputTimer += Asura_Timers::GetGameTimeSlice();
				if (m_fBranchDismountInputTimer >= m_fAlienBranchInputTimeForDismount)
				{
					// Dismount!
					SendDismountBranchMoveMessage( m_xMovementInput, xAlien );
					if (xAlien.GetPhysicsObjectHandler())
					{
						xAlien.GetPhysicsObjectHandler()->UnsetForcedTargetSurfaceNormal();
					}
					xAlien.SetConstrainedMovementAttractorID(uASURA_ATTRACTOR_ID_NONE);
				}
			}
			else
			{
				m_fBranchDismountInputTimer = 0.0f;
			}
		}
	
		const Asura_Matrix_3x3& xClientOri( xAlien.GetCurrentOrientation() );
		// Time to check for the Alien's stance chance...
		if ( GetControlMapValue( STANCE_TOGGLE, true, true ) > 0.1f )
		{
			// AJG - 26/09/08 - Manual stance change removed, so just see if we're on a wall/ceiling and drop if we are.
			// However, don't drop if we're in a duct. We'll have updated the pointer if we changed, so just use the same one.
			if ( (xClientOri.e22 <= dbg_fCosStandMoveTriggerAngle) && (!pxCurrentAttractor || (pxCurrentAttractor->GetType() != uAVP_ATTRACTOR_TYPE_DUCT)) )
			{
				AVP_Alien_JumpHandler::SendJumpMoveStartMessage( m_uOwnerGuid, AVP_Alien_JumpHandler::JT_DROP, Asura_Maths::ZeroVector3, Asura_Maths::ZeroVector3 );
			}
		}
	}
	else if (bJumping && pxCurrentMove == NULL)
	{
		// AJG - 28/07/08 - This doesn't work properly at the moment, and is actively breaking some situations with jumping and falling.
		//					We need to revisit the whole jumping stuff at some point anyway.
 		if (xAlien.GetPhysicsObjectHandler())
 		{
 			xAlien.GetPhysicsObjectHandler()->UnsetForcedTargetSurfaceNormal();
 		}
		xAlien.SetConstrainedMovementAttractorID(uASURA_ATTRACTOR_ID_NONE);
	}

	if ( dvs_bUseLeapEnvHooks && xAlien.GetConstrainedMovementAttractorID() == uASURA_ATTRACTOR_ID_NONE && !pxCurrentMove )
	{
		if ( !dvs_bLeapEnvHooksNeedButtonPress || GetControlMapValue( TRANSITION ) > 0.0f )
		{
			Asura_Vector_3 xLookDisp;
			xAlien.GetCurrentVelocity( xLookDisp );
			xLookDisp *= dvs_fLookAheadAttractorTime;
			Asura_Bounding_Box xBB;
			xBB.SetToPoint( xAlien.GetCurrentPosition() + xLookDisp * 0.5f);
			xBB.Grow( xLookDisp.Magnitude() * 0.5f );

			Asura_Vector_3 xOutPos;
			Asura_Vector_3 xOutNormal;
			LeapCallbackData xLeapCallbackData(*this, xAlien, xLookDisp, dvs_fMinLeapAttractorDist, xOutPos, xOutNormal );
			Asura_Attractor_System::GetInstance()->CollectAttractors( xBB, TestLeapAttractors_Callback, &xLeapCallbackData );

			// See if we found a new attractor
			if (xLeapCallbackData.m_uBestAttractor != uASURA_ATTRACTOR_ID_NONE)
			{
				AVP_Alien_JumpHandler::SendJumpMoveStartMessage( m_uOwnerGuid, AVP_Alien_JumpHandler::JT_JUMP_CONCAVE_90_FLOOR_TO_WALL, xOutPos, xOutNormal );
			}
		}
	}

	if ( GetControlMapValue( FAST ) > 0.0f )
	{
		if ( xAlien.IsStanding() )
		{
			xAlien.SetStanding( false, true );
		}
	}
}

static const float fCOS45 = 0.707f;

void AVP_Alien_InputHandler::TestBranchAttractor(const Asura_Attractor_Compiled& xAttractor, AVP_Alien_InputHandler::BranchCallbackData& xData)
{
	const Asura_Hash_ID uType = xAttractor.GetType();

	const bool bBranch = uType == uAVP_ATTRACTOR_TYPE_BRANCH;
	const bool bDuct = uType == uAVP_ATTRACTOR_TYPE_DUCT;

	if (!bBranch && !bDuct) { return; }
	if (bBranch && !dvs_bEnableAlienBranchEnvHooks) { return; }
	if (bDuct && !dvs_bEnableAlienDuctEnvHooks) { return; }

	const Asura_Vector_3& xAlienPos = xData.m_xAlien.GetCurrentPosition();
	const Asura_Matrix_3x3& xAlienOri = xData.m_xAlien.GetCurrentOrientation();

	Asura_Vector_3 xBaseAttractorDirection;
	Asura_Vector_3 xAttractorDirection;
	xAttractor.GetAxisDirection(xBaseAttractorDirection);
	// Remove the component in the gravity direction
	xAttractorDirection = xBaseAttractorDirection - (xData.m_xGravityDirection * (xBaseAttractorDirection * xData.m_xGravityDirection));
	if (!xAttractorDirection.ApproximatelyEquals(Asura_Maths::ZeroVector3))
	{
		xAttractorDirection.Normalise();
	}

	const bool bNotMoving = xData.m_xTestDirection.ApproximatelyEquals(xData.m_xGravityDirection);

	const Asura_Vector_3& xNode0Pos = xAttractor.GetNode(0)->GetPosition();
	const Asura_Vector_3& xNode1Pos = xAttractor.GetNode(1)->GetPosition();
	Asura_Vector_3 xNode0Offset = xAlienPos - xNode0Pos;
	Asura_Vector_3 xNode1Offset = xAlienPos - xNode1Pos;
	const float fNode0PastEndDP = xNode0Offset * (bNotMoving ? xData.m_xCharacterForward : xData.m_xTestDirection);
	const float fNode1PastEndDP = xNode1Offset * (bNotMoving ? xData.m_xCharacterForward : xData.m_xTestDirection);

	// Find out if we're past the end.
	// To avoid being past the end in our current direction, at least one of the offsets should be in the opposite direction.
	if ((fNode0PastEndDP > 0.0f) && (fNode1PastEndDP > 0.0f))
	{
		return;
	}

	if (xAttractor.GetID() == xData.m_xAlien.GetConstrainedMovementAttractorID())
	{
		// This is the attractor we're already on. Just check the dot product of the input against the direction
		if (bNotMoving)
		{
			// No input, make this the best choice.
			xData.m_uBestAttractor = xAttractor.GetID();
			xData.m_fBestDirectionDPResult = 1.0f;
			xData.m_fBestNormalDPResult = 1.0f;
			xData.m_xBestPosOnAttractor = xAlienPos;
			xData.m_xTargetOrientation.FromMatrix(xAlienOri);
		}
		else
		{
			// Abs means it's irrelevant whether we're pushing along or opposite to the line, which is what we want because we must be somewhere in the middle
			const float fDP = Asura_Maths::Abs(xAttractorDirection * xData.m_xTestDirection);
			// This should take priority on the equality case (if it ever comes up).
			if (fDP >= xData.m_fBestDirectionDPResult)
			{
				// However, if we're approaching the end we want to prioritise the change.
				if (!xData.m_bApproachingAttractorEnd || (xData.m_uBestAttractor == uASURA_ATTRACTOR_ID_NONE))
				{
					xData.m_uBestAttractor = xAttractor.GetID();
					xData.m_fBestDirectionDPResult = fDP;
					xData.m_fBestNormalDPResult = 1.0f;
					xData.m_xBestPosOnAttractor = xAlienPos;
					xData.m_xTargetOrientation.FromMatrix(xAlienOri);
				}
			}
		}
		return;
	}

	// Only continue we're we're nearing the end of our current attractor or we're allowed junction transitions.
	if (!xData.m_bApproachingAttractorEnd && !xData.m_bAllowJunctionTransition) { return; }

	Asura_Vector_3 xNormal;
	xAttractor.GetNormal(xNormal);

	const float fNormalsDotProduct = xNormal * xData.m_xCurrentAttractorNormal;

	if (!xData.m_xCurrentAttractorNormal.ApproximatelyEquals(Asura_Maths::ZeroVector3))
	{
		if (fNormalsDotProduct < (xData.m_bEasyTraversalOnly ? dbg_fAlienBranchMinNormalCosAngleForEasyTraverse : dbg_fAlienBranchMinNormalCosAngleForTraverse))
		{
			// Extra case if we're approaching the end of our current attractor - compare the current attractor normal with the attractor axis direction, and the current forward direction with the attractor normal.
			// If neither of the above pairs are approaching perpendicular, allow the transition. This enables 90-degree transitions over edges or onto walls.
			if (xData.m_bApproachingAttractorEnd)
			{
				const float fCurNormVsAxisDP = Asura_Maths::Abs(xData.m_xCurrentAttractorNormal * xBaseAttractorDirection);
				const float fForwardVsNormalDP = Asura_Maths::Abs(xData.m_xCharacterForward * xNormal);
				if ((fCurNormVsAxisDP < fCOS45) || (fForwardVsNormalDP < fCOS45))
				{
					return;
				}
			}
			else
			{
				// Otherwise, that's it anyway.
				return;
			}
		}
	}

	const bool bHardTraversal = (fNormalsDotProduct < dbg_fAlienBranchMinNormalCosAngleForEasyTraverse) && !xData.m_bEasyTraversalOnly;

	const Asura_Vector_3 xAttractorVector = xNode1Pos - xNode0Pos;

	Asura_Vector_3 xPosOnAttractor, xPosOffset;
	Asura_Maths::GetLeastOffsetBetweenLineSegmentAndRay(xAlienPos, xData.m_xTestDirection, xNode0Pos, xAttractorVector, xPosOffset, &xPosOnAttractor);

	// The actual offset we care about is the offset between the point on the attractor and the alien position.
	xPosOffset = xPosOnAttractor - xAlienPos;

	const float fAboveBranchInGravityDirection = xPosOffset * xData.m_xGravityDirection;

	if (-fAboveBranchInGravityDirection > m_fMaxDistanceBelowBranchForAttach) { return; }

	if (xPosOffset.MagnitudeSquared() > m_fMaxDistanceBetweenAlienAndBranchForAttachSquared) { return; }

	const u_int uCurrentAttractorID = xData.m_xAlien.GetConstrainedMovementAttractorID();

	// Always test against the axis direction.
	// Abs means it's irrelevant whether we're pushing along or opposite to the line, which is what we want.
	const float fDP = Asura_Maths::Abs(xAttractorDirection * xData.m_xTestDirection);
	// If we're approaching the end, other attractors take priority over our current one.
	bool bSuccess = fDP > xData.m_fBestDirectionDPResult;
	if (xData.m_bApproachingAttractorEnd && (xData.m_uBestAttractor == uCurrentAttractorID))
	{
		bSuccess = true;
	}
	else
	{
		bSuccess &= !xData.m_bEasyTraversalOnly || (fNormalsDotProduct > (xData.m_fBestNormalDPResult - dbg_fAlienBranchBestNormalDPAllowedReduction));
	}

	// See where we are relative to the ends of the new attractor.
	int iNearbyNode = -1;
	xNode0Offset = xPosOnAttractor - xNode0Pos;
	xNode1Offset = xPosOnAttractor - xNode1Pos;
	const float fNode0MagSquared = xNode0Offset.MagnitudeSquared();
	const float fNode1MagSquared = xNode1Offset.MagnitudeSquared();
	if ((fNode0MagSquared < fNode1MagSquared) && (xNode0Offset.MagnitudeSquared() < (dbg_fAlienBranchNearbyEndNodePositionRange * dbg_fAlienBranchNearbyEndNodePositionRange)))
	{
		iNearbyNode = 0;
	}
	else if ((fNode1MagSquared < fNode0MagSquared) && (xNode1Offset.MagnitudeSquared() < (dbg_fAlienBranchNearbyEndNodePositionRange * dbg_fAlienBranchNearbyEndNodePositionRange)))
	{
		iNearbyNode = 1;
	}

	// If we're about to hit a wall, cancel our change if we're mid-new-attractor and don't have a definite direction to go in.
	bSuccess &= (iNearbyNode != -1) || xData.m_bAllowJunctionTransition || !xData.m_bApproachingAttractorEnd;

	if (bSuccess)
	{
		// Make sure we aren't blocked from reaching the new attractor. These days, there's no issue with using LOS within an attractor callback.
		const Asura_Vector_3 xLOSOrigin = xAlienPos + (xData.m_xCurrentAttractorNormal * dbg_fAlienBranchLOSTestOffset);
		const Asura_Vector_3 xLOSTarget = xPosOnAttractor + (xData.m_xCurrentAttractorNormal * dbg_fAlienBranchLOSTestOffset);
		if (!xLOSOrigin.ApproximatelyEquals(xLOSTarget))
		{
			Asura_Vector_3 xLOSDir = xLOSTarget - xLOSOrigin;
			const float fDist = xLOSDir.Normalise();
			Asura_LOS_InputData xInput(xLOSOrigin, xLOSDir, xData.m_xAlien.GetGuid(), fDist, false, xData.m_xAlien.GetFlagsToIgnore() | AsuraPolygonFlag_NoCollisions);
			const Asura_LOS_OutputData* pxOutput = Asura_LOS::TestAgainstEverything(&xInput);
			if (pxOutput->m_fDistanceToIntersection < (fDist - dbg_fAlienBranchLOSTestNearWallLeeway))
			{
				return;
			}
		}

		if (bHardTraversal)
		{
			// For hard traversals, retain the old orientation behaviour of trying to maintain the current orientation through the normal change.
			Asura_AMatrix3x3 xTargetOri;
			AVP_Alien_JumpHandler::CalculateTargetOrientation(AVP_Alien_JumpHandler::JT_JUMP_180_WALL_TO_WALL, xData.m_xAlien, xNormal, xTargetOri);
			xData.m_xTargetOrientation.FromMatrix(xTargetOri);
		}
		else
		{
			// To work out what orientation we want either use our input direction or, if that isn't usable, our character forward direction.
			Asura_Vector_3 xForward;
			if (bNotMoving)
			{
				xForward = ((xData.m_xCurrentAttractorNormal * xNormal) >= 0.0f) ? xData.m_xCharacterForward : -xData.m_xCharacterForward;
			}
			else
			{
				if (iNearbyNode == 0)
				{
					xForward = xBaseAttractorDirection;
				}
				else if (iNearbyNode == 1)
				{
					xForward = -xBaseAttractorDirection;
				}
				else
				{
					xForward = xData.m_xTestDirection;
				}
			}

			// Further validity check if we're not nearing the end of our current attractor - are we pushing roughly towards the direction we're going to end up facing?
			if (!bNotMoving && !xData.m_bApproachingAttractorEnd && xData.m_bEasyTraversalOnly)
			{
				if ((xForward * xData.m_xTestDirection) < 0.0f)
				{
					return;
				}
			}

			if (m_xMovementInput.z < 0.0f)
			{
				// Pushing backwards, reverse direction.
				xForward = -xForward;
			}

			Asura_Vector_3 xDown = -xNormal;
			Asura_Vector_3 xRight;
			if ((xBaseAttractorDirection * xForward) >= 0.0f)
			{
				xForward = xBaseAttractorDirection;
			}
			else
			{
				xForward = -xBaseAttractorDirection;
			}
			xRight = xDown ^ xForward;

			// Final validity check - are we maintaining approximately the same right vector?
			if ((xRight * Asura_Vector_3(xAlienOri.e11, xAlienOri.e21, xAlienOri.e31)) < dbg_fAlienBranchTraverseMinRightVectorSimilarity)
			{
				return;
			}

			Asura_Matrix_3x3 xOri(xRight.x, xDown.x, xForward.x, xRight.y, xDown.y, xForward.y, xRight.z, xDown.z, xForward.z);
			xData.m_xTargetOrientation.FromMatrix(xOri);
		}

		xData.m_uBestAttractor = xAttractor.GetID();
		xData.m_fBestDirectionDPResult = fDP;
		xData.m_fBestNormalDPResult = fNormalsDotProduct;
		xData.m_xBestPosOnAttractor = xPosOnAttractor;
	}
}

void AVP_Alien_InputHandler::TestBranchAttractors_Callback(void* pData, const Asura_Attractor_Compiled* const pxAttractor)
{
	if (pData && pxAttractor)
	{
		BranchCallbackData* pxData = static_cast<BranchCallbackData*>(pData);
		pxData->m_xInputHandler.TestBranchAttractor(*pxAttractor, *pxData);
	}
}

void AVP_Alien_InputHandler::SendMountBranchMoveMessage(const Asura_Vector_3& xTargetPos, const Asura_Quat& xTargetOrientation, AVP_ClientActor_Alien& xOwner) const
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Message_Alien_Move_MountBranch_Start));

	if (!pxMessage) { return; }

	AVP_Message_Alien_Move_MountBranch_Start* pxData = static_cast<AVP_Message_Alien_Move_MountBranch_Start*>(pxMessage->GetData());

	// Fill-in the message-body...
	pxData->ClearFlags();
	pxData->SetMoveID(AVP_Move_Alien_MountBranch::GetStaticHash());
	pxData->SetForceAbort(true);
	pxData->SetIgnoredByOwner(false);

	pxData->SetDestinationPosition(xTargetPos);
	pxData->SetTargetOrientation(xTargetOrientation);

	// Set the message source/destination...
	pxMessage->SetFromGuid(m_uOwnerGuid);
	pxMessage->SetToGuid(m_uOwnerGuid);

	// Quick hack - forward the message directly to the client...
	xOwner.HandleMessage(pxMessage);
	// ...and don't bother sending it to the other machines, as they can just update based on position changes as normal.
	Asura_Network::DestroyMessage(pxMessage);
}

void AVP_Alien_InputHandler::SendDismountBranchMoveMessage(const Asura_Vector_3& xDirection, AVP_ClientActor_Alien& xOwner) const
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Message_Alien_Move_DismountBranch_Start));

	if (!pxMessage) { return; }

	AVP_Message_Alien_Move_DismountBranch_Start* pxData = static_cast<AVP_Message_Alien_Move_DismountBranch_Start*>(pxMessage->GetData());

	// Fill-in the message-body...
	pxData->ClearFlags();
	pxData->SetMoveID(AVP_Move_Alien_DismountBranch::GetStaticHash());
	pxData->SetForceAbort(true);
	pxData->SetIgnoredByOwner(false);

	pxData->SetDirection(xDirection);

	// Set the message source/destination...
	pxMessage->SetFromGuid(m_uOwnerGuid);
	pxMessage->SetToGuid(m_uOwnerGuid);

	// Quick hack - forward the message directly to the client...
	xOwner.HandleMessage(pxMessage);
	// ...and don't bother sending it to the other machines, as they can just update based on position changes as normal.
	Asura_Network::DestroyMessage(pxMessage);
}

void AVP_Alien_InputHandler::TestLeapAttractor( const Asura_Attractor_Compiled& xAttractor, AVP_Alien_InputHandler::LeapCallbackData& xData )
{
	const Asura_Hash_ID uType = xAttractor.GetType();
	if ( uType != uAVP_ATTRACTOR_TYPE_ALIEN_LEAP )
	{
		return;
	}
	const Asura_Matrix_3x3& xAlienOri = xData.m_xAlien.GetCurrentOrientation();
	const Asura_Vector_3 xAlienUp( -xAlienOri.e12, -xAlienOri.e22, -xAlienOri.e32 );
	Asura_Vector_3 xAttractorNormal;
	xAttractor.GetNormal( xAttractorNormal );

	if ( xAttractorNormal * xAlienUp < Asura_Maths::Cosine( Asura_Maths::DegToRad( dvs_fAcceptAttractorNormalAngle ) ) )
	{
		return;
	}

	const Asura_Vector_3& xAlienPos = xData.m_xAlien.GetCurrentPosition();
	const Asura_Vector_3& xNode0Pos = xAttractor.GetNode(0)->GetPosition();
	const Asura_Vector_3& xNode1Pos = xAttractor.GetNode(1)->GetPosition();
	Asura_Vector_3 xAttractorDir = xNode1Pos - xNode0Pos;
	float fAttractorLen = xAttractorDir.Normalise();

	// calculate the point on the attractor axis which is in the test direction/up vector plane.
	Asura_Vector_3 xPlaneNormal = xAlienUp ^ xData.m_xTestDir;
	xPlaneNormal.Normalise();
	const AVP_Plane xPlane( xAlienPos, xPlaneNormal );

	float fT;
	if ( xPlane.CalculateIntersectionT( xNode0Pos, xAttractorDir, fT ) && fT >= 0.0f && fT <= fAttractorLen )
	{
		Asura_Vector_3 xPosOnPlane = xNode0Pos + xAttractorDir * fT;
		Asura_Vector_3 xToPoint = xPosOnPlane - xAlienPos;
		float fDistToPoint = xToPoint.Normalise();
		const Asura_Vector_3 xAttractorFwd = xAttractorNormal ^ xAttractorDir;
		if ( xToPoint * xAttractorFwd > Asura_Maths::Cosine( Asura_Maths::DegToRad( dvs_fAcceptAttractorDirectionAngle ) ) )
		{
			if ( xToPoint * xData.m_xTestDir > Asura_Maths::Cosine( Asura_Maths::DegToRad( dvs_fAcceptAttractorConeAngle ) ) && fDistToPoint > xData.m_fMinDist && fDistToPoint < xAttractor.GetRadius() && fDistToPoint < xData.m_fClosestTestDist && fDistToPoint < xData.m_fTestDist )
			{
				Asura_AVector3 xAimDir;
				xData.m_xAlien.GetAimDirection( xAimDir );
				if ( Asura_AVector3( xToPoint ) * xAimDir > Asura_Maths::Cosine( Asura_Maths::DegToRad( dvs_fAcceptAttractorLookAngle ) ) )
				{
					xData.m_uBestAttractor = xAttractor.GetGuid();
					xData.m_fClosestTestDist = fDistToPoint;
					xData.m_xLeapPositionOut = xPosOnPlane;
					xData.m_xLeapNormalOut = xAttractorNormal;
				}
			}
		}
	}
}

void AVP_Alien_InputHandler::TestLeapAttractors_Callback( void* pData, const Asura_Attractor_Compiled* const pxAttractor )
{
	if ( pData && pxAttractor )
	{
		LeapCallbackData* pxData = static_cast< LeapCallbackData* >( pData );
		pxData->m_xInputHandler.TestLeapAttractor( *pxAttractor, *pxData );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_InputHandler::UpdateMeleeRequests
// Author       : Anthony Gregson
// Date         : 22/02/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Alien_InputHandler::UpdateMeleeRequests(AVP_ClientActor_Vertebrate& xClient)
{
	const bool bPrimaryAttack = GetControlMapValue(PRIMARY_ATTACK_TAP) > 0.0f;
	const bool bNormalMeleeAttack = GetControlMapValue(MELEE) > 0.0f;
	const bool bTailAttackPressed = GetControlMapValue(TAIL_ATTACK_TAP) > 0.0f;
	const bool bTailAttackReleased = GetControlMapValue(TAIL_ATTACK_RELEASE) > 0.0f;
	const bool bOffGround = (xClient.GetGroundType() == AVP_Actor_PhysicsObjectHandler::OFF_GROUND) && (xClient.GetConstrainedMovementAttractorID() == uASURA_ATTRACTOR_ID_NONE);
	const bool bJump = GetControlMapValue(JUMP) > 0.0f;
	const AVP_Vertebrate_Focus* pxFocus = xClient.GetFocus();
	const Asura_Guid uMeleeTarget = xClient.GetMeleeTarget();
	if ((bPrimaryAttack || bJump || bNormalMeleeAttack) && !bOffGround && pxFocus && pxFocus->GetActive() && (uMeleeTarget != ASURA_GUID_UNREGISTERED))
	{
		// Send a message to the server
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_REQUEST_MELEE, sizeof(AVP_Message_Vertebrate_RequestMelee));

		// Tell the server we want to pounce...
		if(pxMessage != NULL)
		{
			AVP_Message_Vertebrate_RequestMelee* pxData = static_cast<AVP_Message_Vertebrate_RequestMelee*>(pxMessage->GetData());

			// Fill-in the message...
			if (bNormalMeleeAttack)
			{
				pxData->SetType(AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE);
			}
			else
			{
				pxData->SetType(bPrimaryAttack ? AVP_ALIEN_MELEEATTACK_TYPE_FOCUSATTACK : AVP_ALIEN_MELEEATTACK_TYPE_FOCUSPOUNCE);
			}

			pxData->SetTargetGuid(uMeleeTarget);

			// Send the request...
			pxMessage->SetToGuid(m_uOwnerGuid);
			pxMessage->SetFromGuid(m_uOwnerGuid);
			Asura_Network::Send(pxMessage);
		} 
	}
	else if (bPrimaryAttack || bTailAttackPressed || bTailAttackReleased)
	{
		u_int uType;
		if (bPrimaryAttack)
		{
			uType = AVP_ALIEN_MELEEATTACK_TYPE_PRIMARYATTACK;
		}
		else
		{
			uType = bTailAttackPressed ? AVP_ALIEN_MELEEATTACK_TYPE_SECONDARYATTACK : AVP_ALIEN_MELEEATTACK_TYPE_SECONDARYRELEASE;
		}

		// Send a message to the server
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_REQUEST_MELEE, sizeof(AVP_Message_Vertebrate_RequestMelee));

		// Tell the server we want to interact...
		if(pxMessage != NULL)
		{
			AVP_Message_Vertebrate_RequestMelee* pxData = static_cast<AVP_Message_Vertebrate_RequestMelee*>(pxMessage->GetData());

			// Fill-in the message...
			pxData->SetType(uType);

			pxData->SetTargetGuid(uMeleeTarget);

			// Send the request...
			pxMessage->SetToGuid(m_uOwnerGuid);
			pxMessage->SetFromGuid(m_uOwnerGuid);
			Asura_Network::Send(pxMessage);
		} 
	}
	else if ((GetControlMapValue(HEALTH_REGAIN) > 0.0f) && xClient.GetInteractionHandler() && (xClient.GetInteractionHandler()->GetGuidForInteraction(AVP_INTERACTION_TYPE_HEADBITE) == ASURA_GUID_UNREGISTERED))
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_REQUEST_MELEE, sizeof(AVP_Message_Vertebrate_RequestMelee));

		// Tell the server we want to pounce...
		if(pxMessage != NULL)
		{
			AVP_Message_Vertebrate_RequestMelee* pxData = static_cast<AVP_Message_Vertebrate_RequestMelee*>(pxMessage->GetData());

			// Fill-in the message...
			pxData->SetType(AVP_ALIEN_MELEEATTACK_TYPE_GRAPPLEHEADBITE);

			pxData->SetTargetGuid(uMeleeTarget);

			// Send the request...
			pxMessage->SetToGuid(m_uOwnerGuid);
			pxMessage->SetFromGuid(m_uOwnerGuid);
			Asura_Network::Send(pxMessage);
		} 
	}
	else
	{
		PARENT::UpdateMeleeRequests(xClient);
	}
}

/// End of function 'AVP_Alien_InputHandler::UpdateMeleeRequests'. ///


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_InputHandler::SendPlayerCommandMessageToServer
// Author       : Andy Tate
// Date         : 4-Oct-2006
////////////////////////////////////////////////////////////////////////////////

Asura_Message* AVP_Alien_InputHandler::CreatePlayerCommandMessageToServer ( const AVP_ClientEntity_Actor& xClient ) const
{
	bool bShouldCompressMessage = AVP_Game_Core::ShouldCompressPlayerCommandsMessages(); 
	
	if (bShouldCompressMessage)
	{
		// Fill message data
		const AVP_ClientActor_Alien& xAlien = static_cast< const AVP_ClientActor_Alien& >( xClient );
		AVP_Message_AlienPlayerCommands xAlienPlayerCommands( xAlien ); 
		const u_int uNoCompressionSize = sizeof( AVP_Message_AlienPlayerCommands ); 

		// Compress data
		Asura_BitPacker xPacker( uNoCompressionSize ); 
		xAlienPlayerCommands.Pack( xPacker ); 

		// Send the data
		const u_char* pucSrcData = 0;
		u_int uDataSize = 0;
		xPacker.GetResult( &pucSrcData, &uDataSize );

		Asura_Message*	pxMessage	= Asura_Network::CreateMessage( AVP_MSGID_ALIENPLAYERCOMMANDS_COMPRESSED, uDataSize );

		// Multiple sanity checks
		if (pxMessage && pucSrcData && uDataSize)
		{
			u_char* pucDstData = static_cast<u_char*>( pxMessage->GetData() ); 
			Asura_Mem::Copy( pucDstData, pucSrcData, uDataSize ); 
		}
		else
		{
			Asura_Assert(false, "Failed to create message.");
		}
		
		return pxMessage; 
	}
	else
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_ALIENPLAYERCOMMANDS, sizeof(AVP_Message_AlienPlayerCommands));

		if (pxMessage == NULL)
		{
			Asura_Assert(false, "Failed to create message.");
		}
		else
		{
			AVP_Message_AlienPlayerCommands*	pxData = static_cast<AVP_Message_AlienPlayerCommands*>(pxMessage->GetData());

			if (pxData == NULL)
			{
				Asura_Assert(false, "Cannot find data in message.");
			}
			else
			{
				const AVP_ClientActor_Alien& xAlien	= static_cast< const AVP_ClientActor_Alien& > ( xClient );
				pxData->FillAlien( xAlien );
			}
		}
		
		// Return the message created...
		return(pxMessage);
	}
}

/// End of function 'AVP_Alien_InputHandler::SendPlayerCommandMessageToServer'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Alien_InputHandler::AVP_Alien_InputHandler
// Author       : Andy Tate
// Date         : 3-Oct-2006
////////////////////////////////////////////////////////////////////////////////

AVP_Alien_InputHandler::AVP_Alien_InputHandler (const Asura_Guid uOwnerGuid)
: AVP_Vertebrate_InputHandler(uOwnerGuid)
, m_fBranchDismountInputTimer(0.0f)
, m_fLookSmoothingTimer(0.0f)
, m_fMaxDistanceBetweenAlienAndBranchForAttachSquared(2.25f)
, m_fMaxDistanceBelowBranchForAttach(0.25f)
, m_fAlienBranchLOSDistance(1.0f)
, m_fAlienBranchInputTimeForDismount(0.25f)
, m_fAlienBranchMaxCosAngleForDismount(0.4f)
{
	// Initialise the control maps..
	RefreshBlueprintData();
	Platform_InitialiseControlMap();
}

/// End of function 'AVP_Alien_InputHandler::AVP_Alien_InputHandler'. ///

void AVP_Alien_InputHandler::RefreshBlueprintData(void)
{
	const AVP_ClientActor_Alien*	pxOwner	= AVP_ClientActor_Alien::FindEntity(m_uOwnerGuid);

	PARENT::RefreshBlueprintData();

	if(pxOwner != NULL)
	{
		const Asura_Hash_ID		uBlueprintHash	= pxOwner->GetBlueprintHashID();
		const Asura_Hash_ID		uTypeHash		= uAVP_BLUEPRINT_TYPE_PLAYERS;

		const Asura_Blueprint*	pxBlueprint		= Asura_Blueprint_System::GetBlueprintByHashID(uTypeHash, uBlueprintHash);
		if(pxBlueprint)
		{
			Asura_Value	xValue;

			if(pxBlueprint->GetParameterValueByHashID(0x44C21088, xValue))	//Asura_GetHashID("BranchAttachDistanceMax"), xValue))
			{
				m_fMaxDistanceBetweenAlienAndBranchForAttachSquared = xValue.GetFloatValue();
				m_fMaxDistanceBetweenAlienAndBranchForAttachSquared *= m_fMaxDistanceBetweenAlienAndBranchForAttachSquared;
			}
			if(pxBlueprint->GetParameterValueByHashID(0xFDD8F3D1, xValue))	//Asura_GetHashID("BranchAttachBelowLimit"), xValue))
			{
				m_fMaxDistanceBelowBranchForAttach = xValue.GetFloatValue();
			}
			if(pxBlueprint->GetParameterValueByHashID(0xD835567E, xValue))	//Asura_GetHashID("BranchAttachLOSDistance"), xValue))
			{
				m_fAlienBranchLOSDistance = xValue.GetFloatValue();
			}
			if(pxBlueprint->GetParameterValueByHashID(0x9DC91582, xValue))	//Asura_GetHashID("BranchDetachInputTime"), xValue))
			{
				m_fAlienBranchInputTimeForDismount = xValue.GetFloatValue();
			}
			if(pxBlueprint->GetParameterValueByHashID(0x24504986, xValue))	//Asura_GetHashID("BranchDetachMinInputAngle"), xValue))
			{
				m_fAlienBranchMaxCosAngleForDismount = Asura_Maths::Cosine(Asura_Maths::DegToRad(xValue.GetFloatValue()));
			}
		}
	}
}

Asura_Message* AVP_Alien_InputHandler::Update(AVP_ClientEntity_Actor& xClient)
{
	if ( !xClient.IsDead() )
	{
		// Headbite request.
		AVP_Actor_InteractionHandler* pxInteractionHandler = xClient.GetInteractionHandler();
		if(((GetControlMapValue(HEALTH_REGAIN) > 0.0f) || (GetControlMapValue(PRIMARY_ATTACK_TAP) > 0.0f)) && pxInteractionHandler && pxInteractionHandler->IsInteractionAllowed())
		{
			pxInteractionHandler->HandleInteraction(AVP_INTERACTION_TYPE_HEADBITE, HEALTH_REGAIN, GetControlMapValue(HEALTH_REGAIN));
		}

		if ( xClient.GetPhysicsObjectHandler() && xClient.GetPhysicsObjectHandler()->GetCalculateSurfaceReorientationNormal() )
		{
			if ( !dvs_bUsePushToTransition || GetControlMapValue( TRANSITION ) > 0.0f )
			{
				xClient.GetPhysicsObjectHandler()->GetCalculateSurfaceReorientationNormal()->SetCosLimit( Asura_Maths::Cosine( Asura_Maths::DegToRad( dvs_fPushTransAngle ) ) );
			}
			else
			{
				xClient.GetPhysicsObjectHandler()->GetCalculateSurfaceReorientationNormal()->SetCosLimit( Asura_Maths::Cosine( Asura_Maths::DegToRad( dvs_fNoPushTransAngle ) ) ); // cos 50 degrees
			}
		}

		UpdateAimDirection( xClient );
	}

	return PARENT::Update(xClient);
}

void AVP_Alien_InputHandler::SetNewLookDirection( AVP_ClientEntity_Actor& xClient, const Asura_AVector3& xDir, const bool bImmediate )
{
	if ( bImmediate )
	{
		Asura_AVector3 xOldAim;
		xClient.GetAimDirection( xOldAim );
		Asura_Quat xRot;
		// This results in a y != up rotation for the camera direction which I think looks quite cool
		// If someone decides I'm wrong tho, this rotation should be generated from eulers instead of an AtoB.
		xRot.RotationFromAToB( xOldAim, xDir );
		AVP_Alien_CameraHandler* pxCamHan = static_cast< AVP_Alien_CameraHandler* >( xClient.GetCameraHandler() );
		if ( pxCamHan )
		{
			pxCamHan->RotateCurrentAimDirection( xRot );
		}
	}

	PARENT::SetNewLookDirection( xClient, xDir, bImmediate );
}

void AVP_Alien_InputHandler::CreateConsoleVars()
{
#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.UseBranchEnvHooks", &dvs_bEnableAlienBranchEnvHooks, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.UseDuctEnvHooks", &dvs_bEnableAlienDuctEnvHooks, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.UsePushToTransition", &dvs_bUsePushToTransition, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.UseLeapEnvHooks", &dvs_bUseLeapEnvHooks, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.UseLookPrediction", &dvs_bEnableLookPrediction, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooksNeedButtonPress", &dvs_bLeapEnvHooksNeedButtonPress, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.PushToTransition.NoPushMaxAngle", &dvs_fPushTransAngle, 0.0f, 180.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.PushToTransition.PushMaxAngle", &dvs_fNoPushTransAngle, 0.0f, 180.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooks.AcceptAttractorNormalAngle", &dvs_fAcceptAttractorNormalAngle, 0.0f, 180.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooks.AcceptAttractorDirectionAngle", &dvs_fAcceptAttractorDirectionAngle, 0.0f, 180.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooks.AcceptAttractorLookAngle", &dvs_fAcceptAttractorLookAngle, 0.0f, 180.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooks.AcceptAttractorConeAngle", &dvs_fAcceptAttractorConeAngle, 0.0f, 180.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooks.MinLeapAttractorDist", &dvs_fMinLeapAttractorDist, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.AlienPlayer.LeapEnvHooks.LookAheadAttractorTime", &dvs_fLookAheadAttractorTime, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
#endif //ASURA_DEBUG_VARIABLE_SYSTEM
}


DEBUGVAR float dbg_fLookDirectionSmoothingAngle = Asura_Maths::DegToRad( 30.0f );
DEBUGVAR float dbg_fLookDirectionSmoothingCosAngle = Asura_Maths::Cosine( dbg_fLookDirectionSmoothingAngle );
DEBUGVAR float dbg_fLookSmoothingTime = 0.2f;

void AVP_Alien_InputHandler::UpdateAimDirection( AVP_ClientEntity_Actor& xClient )
{
	const Asura_Matrix_3x3& xOri = xClient.GetCurrentOrientation();
	const Asura_Vector_3 xUp( -xOri.e12, -xOri.e22, -xOri.e32 );
	const Asura_Vector_3 xFwd( xOri.e13, xOri.e23, xOri.e33 );
	const Asura_Vector_3 xRight( xOri.e11, xOri.e21, xOri.e31 );
	Asura_Assert( Asura_Maths::Abs( xUp * xFwd ) < 0.0001f, "" );

	const Asura_Movement_System*	pxMovementSystem	= xClient.GetMovementSystem();
	const AVP_Move*					pxMove				= pxMovementSystem != NULL ? static_cast< const AVP_Move* >( pxMovementSystem->GetCurrentMove() ) : NULL;

	if ( pxMove && pxMove->GetAimDirectionUsage() != AVP_ClientEntity_Actor::AIM_DIRECTION_NOT_USED )
	{
		m_xLastUpVector = xUp;
		m_fLookSmoothingTimer = 0.0f;
	}
	else
	{
		LOG( ( "Old Up Vector: %+.4f %+.4f %+.4f ", m_xLastUpVector.GetX(), m_xLastUpVector.GetY(), m_xLastUpVector.GetZ() ) );
		LOG( ( "New Up Vector: %+.4f %+.4f %+.4f ", xUp.GetX(), xUp.GetY(), xUp.GetZ() ) );
		LOG( ( "New Fwd Vector: %+.4f %+.4f %+.4f ", xFwd.GetX(), xFwd.GetY(), xFwd.GetZ() ) );
		LOG2( ( "Ori Fwd.Up = %.4f ", xFwd.GetY() ) );
		
		Asura_AVector3 xAAim;
		xClient.GetAimDirection( xAAim );
		Asura_Vector_3 xAim( xAAim );
		LOG( ( "Old Aim Dir: %+.4f %+.4f %+.4f ", xAim.GetX(), xAim.GetY(), xAim.GetZ() ) );

		Asura_Quat xRot;
		xRot.RotationFromAToB( m_xLastUpVector, xUp );
		m_xLastUpVector = xUp;
		xAim = xRot.RotateVector( xAim );
		LOG( ( "Rotated Aim Dir: %+.4f %+.4f %+.4f ", xAim.GetX(), xAim.GetY(), xAim.GetZ() ) );

		UpdateAimDirectionPrediction( xClient, xAim );

		LOG( ( "New Aim Dir: %+.4f %+.4f %+.4f\n", xAim.GetX(), xAim.GetY(), xAim.GetZ() ) );

		SetNewLookDirection( xClient, Asura_AVector3(xAim), false );
	}
}

DEBUGVAR float dbg_fGroundOffset = 0.01f; // start ray this far above the ground
DEBUGVAR float dbg_fPredictionTime = 1.0f; // look 1s ahead max
DEBUGVAR u_int dbg_uNumSamples = 10;
DEBUGVAR float dbg_fCosFwdAngle = Asura_Maths::Cosine( Asura_Maths::DegToRad( 6.0f ) );
DEBUGVAR float dbg_fCosPredictionLimit = Asura_Maths::Cosine( Asura_Maths::DegToRad( 6.0f ) );
DEBUGVAR float dbg_fCosRejctAngle = Asura_Maths::Cosine( Asura_Maths::DegToRad( 45.0f ) );
DEBUGVAR float dbg_fMinStepHeight = 0.5f;

//#define DEBUGDRAWLINE(params)
//#define DEBUGDRAWBALL(params)
#define DEBUGDRAWLINE(s,e,c) Asura_Primitive_Render::DrawLineWorld( (s), (e), (c) )
#define DEBUGDRAWBALL(p, r, c, s1, s2) Asura_Primitive_Render::DrawSphereWorld( (p), (r), (c), (s1), (s2) )

void AVP_Alien_InputHandler::UpdateAimDirectionPrediction( AVP_ClientEntity_Actor& xClient, Asura_Vector_3& xAim )
{
	const Asura_Matrix_3x3& xOri = xClient.GetCurrentOrientation();
	const Asura_Vector_3 xDown( xOri.e12, xOri.e22, xOri.e32 );
	const Asura_Vector_3 xFwd( xOri.e13, xOri.e23, xOri.e33 );
	const Asura_Vector_3 xRight( xOri.e11, xOri.e21, xOri.e31 );
	Asura_Vector_3 xVel;
	xClient.GetCurrentVelocity( xVel );
	bool bPredicted = false;

	if ( dvs_bEnableLookPrediction && ( !dvs_bUsePushToTransition || GetControlMapValue( TRANSITION ) > 0.0f ) && m_xMovementInput.GetZ() > 0.0f && Asura_Maths::Abs( m_xMovementInput.GetX() ) < 0.5f && xVel.MagnitudeSquared() > 0.001f )
	{
		const float fSpeed = xVel.Magnitude();
		const float fPredictionDist = fSpeed * dbg_fPredictionTime;
		const Asura_Vector_3 xOriginalOrigin = xClient.GetCurrentPosition() - xClient.GetDownVector() * dbg_fGroundOffset ;

		// Do LOS ahead of alien
		Asura_LOS_InputData xFwdInputData( xOriginalOrigin, xFwd, m_uOwnerGuid, fPredictionDist, false, AsuraPolygonFlag_NoCollisions | AVP_PolygonFlag_IgnoreForAlienCollision | AVP_PolygonFlag_IgnoreForReorientation | AVP_PolygonFlag_IgnoreForPlayerCollision );
		Asura_LOS_OutputData xFwdOutputData = *Asura_LOS::TestAgainstEnvironment( &xFwdInputData );
		PredictionLOSFunctor xFwdFunctor( xFwdOutputData );
		Asura_Entity_System::TestEntitiesWithLOSFunctor( &xFwdInputData, xFwdFunctor );
		// technically should do this but no need here
		// 		if ( xOutputData.m_uHitEntity )
		// 		{
		// 			xOutputData.m_bHitEnvironment = false;
		// 		}

		Asura_Vector_3 xAimTarget = xOriginalOrigin + xFwd * xFwdOutputData.m_fDistanceToIntersection;
		if ( ( xFwdOutputData.m_bHitEnvironment || xFwdOutputData.m_uHitEntity != 0 ) && xFwdOutputData.m_fDistanceToIntersection < fPredictionDist )
		{
			Asura_Vector_3 xNewFwd = xRight ^ -xFwdOutputData.m_xNormal;
			xNewFwd.Normalise();
			DEBUGDRAWLINE( xOriginalOrigin, xAimTarget, 0xFF00FF00 );
			DEBUGDRAWLINE( xAimTarget, xAimTarget + xNewFwd * ( fPredictionDist - xFwdOutputData.m_fDistanceToIntersection ), 0xFF0000FF );
			xAimTarget += xNewFwd * ( fPredictionDist - xFwdOutputData.m_fDistanceToIntersection );
		}
		else
		{
			DEBUGDRAWLINE( xOriginalOrigin, xAimTarget, 0xFFFF0000 );
		}

		for ( u_int uCount = 1; uCount < dbg_uNumSamples; ++uCount )
		{
			const float fFwdDist = ( xFwdOutputData.m_fDistanceToIntersection * uCount ) / dbg_uNumSamples;
			Asura_Vector_3 xDownOrigin = xOriginalOrigin + xFwd * fFwdDist;
			const float fDistance = xFwdOutputData.m_fDistanceToIntersection - fFwdDist;

			Asura_LOS_InputData xDownInputData( xDownOrigin, xDown, m_uOwnerGuid, fDistance, false, AsuraPolygonFlag_NoCollisions | AVP_PolygonFlag_IgnoreForAlienCollision | AVP_PolygonFlag_IgnoreForReorientation | AVP_PolygonFlag_IgnoreForPlayerCollision );
			Asura_LOS_OutputData xDownOutputData = *Asura_LOS::TestAgainstEnvironment( &xDownInputData );
			PredictionLOSFunctor xDownFunctor( xDownOutputData );
			Asura_Entity_System::TestEntitiesWithLOSFunctor( &xDownInputData, xDownFunctor );

			if ( ( xDownOutputData.m_bHitEnvironment || xDownOutputData.m_uHitEntity != 0 ) && xDownOutputData.m_fDistanceToIntersection < fDistance )
			{
				const float fOffset = Asura_Maths::Max( 0.0f, xDownOutputData.m_fDistanceToIntersection - dbg_fGroundOffset );
				if ( fOffset > dbg_fMinStepHeight )
				{
					Asura_Vector_3 xNewFwd = xRight ^ -xDownOutputData.m_xNormal;
					xNewFwd.Normalise();
					xAimTarget = xDownOrigin + xDown * xDownOutputData.m_fDistanceToIntersection + xNewFwd * ( fDistance - xDownOutputData.m_fDistanceToIntersection );

					DEBUGDRAWLINE( xDownOrigin, xDownOrigin + xDown * xDownOutputData.m_fDistanceToIntersection, 0xFF00FF00 );
					DEBUGDRAWLINE( xDownOrigin + xDown * xDownOutputData.m_fDistanceToIntersection, xAimTarget, 0xFF0000FF );
					uCount = dbg_uNumSamples;
				}
			}
			else
			{
				xAimTarget = xDownOrigin + xDown * fDistance;
				DEBUGDRAWLINE( xDownOrigin, xAimTarget, 0xFFFF0000 );
				uCount = dbg_uNumSamples;
			}
		}

		Asura_Vector_3 xNewDir = xAimTarget - xOriginalOrigin;
		xNewDir.Normalise();

		if ( !xNewDir.ApproximatelyEquals( xFwd, 0.01f ) )
		{
			xAim = xNewDir;
			DEBUGDRAWBALL( xAimTarget, 0.2f, 0xFFFFFFFF, 4, 2 );
			m_fLookSmoothingTimer = dbg_fLookSmoothingTime;
			bPredicted = true;
		}
	}
	
	if ( !bPredicted )
	{
		if ( m_fLookSmoothingTimer > 0.0f )
		{
			m_fLookSmoothingTimer -= Asura_Timers::GetGameTimeSlice();
			xAim = xFwd;
		}

		LOG( ( "Reset Aim Dir: %+.4f %+.4f %+.4f\n", xAim.GetX(), xAim.GetY(), xAim.GetZ() ) );
	}
}

bool AVP_Alien_InputHandler::PredictionLOSFunctor::CheckEntity( Asura_Entity_Renderable* pxEntity, const Asura_LOS_InputData& xInputData )
{
	Asura_Assert( pxEntity, "Null leaf object pointer" );
	if ( pxEntity )
	{
		if ( AVP_ServerEntity_PhysicalObject::CanUseForSurfaceReorientation( pxEntity->GetGuid() ) )
		{
			return pxEntity->LOSTest( &xInputData, &m_rxOutData );
		}
	}
	return false;
}

/// End of file. ///
