///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Human_AnimationHandler.cpp
// Author : Chris Wallace
// Date   : 25-Apr-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Actor/Vertebrates/Human/AVP_Human_AnimationHandler.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationBlend.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Network/Asura_Network.h"
#include "Entity/AVP_Entity_Classification.h"
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "FacialAnimation/Asura_FacialAnimation_AnimationHandler_PhysicalObject.h"
#include "FacialAnimation/Asura_FacialAnimation_System.h"
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
#include "Physics/Asura_Physics_Ragdoll.h"
#include "ProjectSettings/Asura_PhysicsConfiguration.h"
#include "SpecialFX/PFX/Asura_PFX_Entity_Effect.h"

#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Debugging/Asura_DebugLineRender.h"
#include "Interface/Asura_CommandConsole.h"
#endif

// Game independent header-files...
#include "Actor/AVP_Actor_CameraHandler.h"
#include "Actor/Vertebrates/Human/AVP_Human_Description.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Game_Core.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ClientActor_Human.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ServerActor_Human.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR Asura_Vector_3	dbg_xPulseRifleToCrosshairOffset			(0.0f, 0.0f, 50.0f);

#define DEAFULT_TWEEN_IN_AMOUNT 0.2f
#define DEAFULT_TWEEN_OUT_AMOUNT 0.2f

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

static const Asura_Hash_ID g_uHUDLeftHandMarkerHash = 0x374C76BA;// Asura_GetHashID("left_hand_ik")
static const Asura_Hash_ID g_uTPLeftHandMarkerHash = 0x2b2dca21;// Asura_GetHashID("left_hand_ik_TP")

static const Asura_AVector3	g_xDownVector( 0.0f, 1.0f, 0.0f );

void AVP_Human_AnimationHandler::Initialise(void)
{
	PARENT::Initialise();

	if ( m_pxBodyRig )
	{
		if ( AVP_Game_Core::IsLocalPlayer( m_uOwnerGuid ) )
		{
			m_pxBodyRig->SetMaxAimRightLeft( Asura_Maths::Deg90 );
		}
		else
		{
			m_pxBodyRig->SetMaxAimRightLeft( Asura_Maths::Deg90 );
		}
	}

	InitialiseIKSolver();
}

void AVP_Human_AnimationHandler::Shutdown( void )
{
	DisableFacialAnimation();

	PARENT::Shutdown();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Human_AnimationHandler::CalculateMarkerPointToCrosshairData
// Author       : Andy Tate
// Date         : 3-Jan-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_Human_AnimationHandler::CalculateMarkerPointToCrosshairData (const Asura_Hash_ID uSourceMarkerPointHash,
																		 const float fInitialSpeed,
																		 Asura_AVector3& xPosition,
																		 Asura_AMatrix3x3& xOrientation,
																		 Asura_AVector3& xVelocity) const
{
	const AVP_ClientActor_Human*	pxClient	= AVP_ClientActor_Human::FindEntity(m_uOwnerGuid);
	Asura_Vector_3					xMarkerPosition;
	Asura_Matrix_3x3				xMarkerOrientation;

	// Find the marker-point orientation/position...
	if ( pxClient->GetMarkerLocationFromHash( uSourceMarkerPointHash, &xMarkerPosition, &xMarkerOrientation, true ) )
	{
		Asura_AVector3				xForwards;
		Asura_AVector3				xDown ( xMarkerOrientation.e12, xMarkerOrientation.e22, xMarkerOrientation.e32 );
		pxClient->GetAimDirection( xForwards );

		// make "Down" column perpendicular to forwards, and normalise
		xDown -= xForwards * ( xDown * xForwards );
		xDown.Normalise();

		// and now the "right" column must be perpendicular to both of these: x = + y^z
		xOrientation.SetColumn0( xDown ^ xForwards );
		xOrientation.SetColumn1( xDown );
		xOrientation.SetColumn2( xForwards );

		// Calculate the velocity...
		xVelocity	= xForwards * fInitialSpeed;

		return(true);
	}
	else
	{
		return(false);
	}
}

/// End of function 'AVP_Human_AnimationHandler::CalculateMarkerPointToCrosshairData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Human_AnimationHandler::UpdateSkin
// Author       : Andy Tate
// Date         : 27-Oct-2006
////////////////////////////////////////////////////////////////////////////////

void AVP_Human_AnimationHandler::UpdateSkin()
{
	Asura_Hierarchy_Skin*	pxSkin;

	// Firstly, call the parent function...
	PARENT::UpdateSkin();

	// Update the marker-point indices...
	pxSkin	= Asura_Animation_System::GetSkinByHash(m_uSkinHash);

	if(pxSkin != NULL)
	{
		pxSkin->GetMarkerPointIndex(AVP_Human_Description::GetMarkerPointHashID(AVP_Human_Description::MPT_RIGHT_HAND), m_iRightHandMarkerPointIndex);
	}

#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	// Decide what to do with the facial-animation...
	if(IsFirstPerson() || IsRagdoll())
	{
		DisableFacialAnimation();
	}
	else
	{
		EnableFacialAnimation();
	}
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
}

/// End of function 'AVP_Human_AnimationHandler::UpdateSkin'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////


AVP_Human_AnimationHandler::~AVP_Human_AnimationHandler()
{
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
	DisableFacialAnimation();
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Human_AnimationHandler::AVP_Alien_AnimationHandler
// Author       : Andy Tate
// Date         : 18-Oct-2006
////////////////////////////////////////////////////////////////////////////////

AVP_Human_AnimationHandler::AVP_Human_AnimationHandler(AVP_ClientActor_Human& xOwner)
: PARENT(xOwner)
, m_xOwnerHuman(xOwner)
#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
, m_pxFacialAnimationHandler(NULL)
#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)
{
}


//////////////////////////////////////////////////////////////////////////////
//	AVP_Human_AnimationHandler :: HandleAniamtionRequest
//////////////////////////////////////////////////////////////////////////////
void AVP_Human_AnimationHandler::HandleAnimationRequest( const AVP_Message_Actor_RequestAnimation& xAnimRequestData )
{
#if 0
	u_int			uAnimFlags		= s_DefaultAnimFlags;
	const u_int		uAnim			= xAnimRequestData.m_uAnimation;
	float			fTweenInAmt		= 0.0f;
	float			fTweenOutAmt	= 0.0f;

	// Check if the actor is currently on a gamescene performing gamescene animations, if he is return while assertion
	if (m_xCurrentAnimSequenceData.IsValid() || m_xPendingAnimSequenceData.IsValid())
	{
		Asura_Assert( false, "AVP_Human_AnimationHandler::HandleAnimationRequest - shouldn't request an anim whilst playing a gamescene anim template" );
		return;
	}	

	// Add animations loop flag if needed
	if (LoopflagIsNeeded(xAnimRequestData))
	{
		uAnimFlags |= AACA_FLAG_LOOPED | AAC_FLAG_DESTROY_WHEN_FINISHED | AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE;
	}
	
	// Add calculation flags if needed
	if (CalculationFlagsIsNeeded(uAnim))
	{
		uAnimFlags |= AACA_FLAG_CALCULATE_MOVEMENT_OFFSET | AACA_FLAG_CALCULATE_VELOCITY | AACA_FLAG_CALCULATE_ANGULAR_VELOCITY;
	}

	// Add Subset of bones flags if needed
	if (SubsetOfBonesFlagIsNeeded(uAnim))
	{
		uAnimFlags |= AACA_FLAG_USING_SUBSET_OF_BONES;
	}

	// Get the Tween In / out amounts for the Animations
	fTweenInAmt		= GetTweenInValueForAnim(uAnim);
	fTweenOutAmt	= GetTweenOutValueForAnim(uAnim);
	
	// Finally check the Anim ID is actually valid before creating the animation
	if (IsAnimIdValid(uAnim))
	{
		CreateRequestedAnimation(AVP_Human_Description::GetAnimationHashID(uAnim), uAnimFlags, fTweenInAmt, fTweenOutAmt);
	}
#endif
}


void AVP_Human_AnimationHandler::Update()
{
	PARENT::Update();

	if ( GetAnimation() )
	{
		const u_int uAnimFlags = GetAnimation()->GetFlagsForAllAnimsLastFrame();

		m_xLeftLegIKChain.SetAnimVeto( ( uAnimFlags & AVP_ANIMFLAG_NOT_IK_FEET ) != 0 );
		m_xRightLegIKChain.SetAnimVeto( ( uAnimFlags & AVP_ANIMFLAG_NOT_IK_FEET ) != 0 );
		m_xLeftArmIKChain.SetAnimVeto( ( uAnimFlags & AVP_ANIMFLAG_NOT_IK_HANDS_AND_TAIL ) != 0 );
	}
}

void AVP_Human_AnimationHandler::CreateRequestedAnimation( const Asura_Hash_ID uAnimHashID, const u_int uAnimFlags, float fTweenIn, float fTweenOut )
{
	Asura_AnimationBlend*	pxAnimBlend	= new Asura_AnimationBlend_AnimatedTween( fTweenIn, fTweenOut, false );
	const u_int				uPriority	= AVP_AAPC_REQUEST_ANIM;

	m_uCurrentRequestedAnimControllerHash = PlayAnimation( uAnimFlags, uPriority, uAnimHashID, pxAnimBlend );
}

const Asura_Animation_IK_Chain_LOS* AVP_Human_AnimationHandler::GetLimbIKChainLOS( u_int uLimbItr ) const
{
	switch ( uLimbItr )
	{
		case AVP_ClientEntity_Actor::FOOT_REAR_LEFT:
		{
			return &m_xLeftLegIKChain;
		}

		case AVP_ClientEntity_Actor::FOOT_REAR_RIGHT:
		{
			return &m_xRightLegIKChain;
		}

		default:
		{
			return NULL;
		}
	}
}

void AVP_Human_AnimationHandler::InitialiseIKSolver()
{
	Asura_Assert( m_xIKSolver.GetFirstChain() == NULL, "IK Solver already initialised, clearing it..." );
	if ( m_xIKSolver.GetFirstChain() )
	{
		m_xIKSolver.Clear();
	}

	static const int NUM_EXCEPTIONS = 7;
	static const u_short s_ausClassificationExceptionArray[NUM_EXCEPTIONS] = { 
		AVP_ENTITYCLASS_DISC, 
		AVP_ENTITYCLASS_DROPSHIP_MISSILE, 
		AVP_ENTITYCLASS_GENERICPROJECTILE, 
		AVP_ENTITYCLASS_HUMANFLARE, 
		AVP_ENTITYCLASS_PROJECTILE_FLAMETHROWER,
		AVP_ENTITYCLASS_MINE,
		AVP_ENTITYCLASS_PROJECTILE_ACIDSPIT
	};

	if ( IsFirstPerson() )
	{
		m_xLeftArmIKChain.SetMarkerPointHash( g_uHUDLeftHandMarkerHash );
		m_xLeftArmIKChain.SetAnim( m_pxAnimation );
		m_xLeftArmIKChain.SetUseOri( true );
		m_xLeftArmIKChain.SetPosSmoothing( false );
		m_xLeftArmIKChain.SetOriSmoothing( false );
		m_xLeftArmIKChain.SetBonesArrays( AVP_Human_Description::GetFirstPersonLeftArmChainBonesArray(), AVP_Human_Description::GetFirstPersonLeftArmChainTwistBonesArray() );
		if ( m_pxAnimation && m_pxAnimation->GetSkin() )
		{
			m_xLeftArmIKChain.SetAxesFromRestPose( m_pxAnimation->GetSkin()->BoneRestPose );
		}
		m_xLeftArmIKChain.SetExtendLimit( 0.01f );
		m_xIKSolver.AddChain( m_xLeftArmIKChain );

		m_xLeftLegIKChain.SetLocalLOSDirection( g_xDownVector );
		m_xLeftLegIKChain.SetTestAgainstEverything( m_uOwnerGuid );
		m_xLeftLegIKChain.SetExclusionLists(  NUM_EXCEPTIONS, s_ausClassificationExceptionArray );
		m_xLeftLegIKChain.SetBonesArrays( AVP_Human_Description::GetFirstPersonLeftLegChainBonesArray() );
		if ( m_pxAnimation && m_pxAnimation->GetSkin() )
		{
			m_xLeftLegIKChain.SetAxesFromRestPose( m_pxAnimation->GetSkin()->BoneRestPose );
		}
		m_xLeftLegIKChain.SetIsRootBoneMover();
		m_xLeftLegIKChain.SetExtendLimit( 0.01f );
		m_xIKSolver.AddChain( m_xLeftLegIKChain );

		m_xRightLegIKChain.SetLocalLOSDirection( g_xDownVector );
		m_xRightLegIKChain.SetTestAgainstEverything( m_uOwnerGuid );
		m_xRightLegIKChain.SetExclusionLists(  NUM_EXCEPTIONS, s_ausClassificationExceptionArray );
		m_xRightLegIKChain.SetBonesArrays( AVP_Human_Description::GetFirstPersonRightLegChainBonesArray() );
		if ( m_pxAnimation && m_pxAnimation->GetSkin() )
		{
			m_xRightLegIKChain.SetAxesFromRestPose( m_pxAnimation->GetSkin()->BoneRestPose );
		}
		m_xRightLegIKChain.SetIsRootBoneMover();
		m_xRightLegIKChain.SetExtendLimit( 0.01f );
		m_xIKSolver.AddChain( m_xRightLegIKChain );
	}
	else
	{
		m_xLeftArmIKChain.SetMarkerPointHash( g_uTPLeftHandMarkerHash );
		m_xLeftArmIKChain.SetAnim( m_pxAnimation );
		m_xLeftArmIKChain.SetUseOri( true );
		m_xLeftArmIKChain.SetPosSmoothing( false );
		m_xLeftArmIKChain.SetOriSmoothing( false );
		m_xLeftArmIKChain.SetBonesArrays( AVP_Human_Description::GetThirdPersonLeftArmChainBonesArray(), AVP_Human_Description::GetThirdPersonLeftArmChainTwistBonesArray() );
		if ( m_pxAnimation && m_pxAnimation->GetSkin() )
		{
			m_xLeftArmIKChain.SetAxesFromRestPose( m_pxAnimation->GetSkin()->BoneRestPose );
		}
		m_xLeftArmIKChain.SetExtendLimit( 0.01f );
		m_xIKSolver.AddChain( m_xLeftArmIKChain );

		m_xLeftLegIKChain.SetLocalLOSDirection( g_xDownVector );
		m_xLeftLegIKChain.SetTestAgainstEverything( m_uOwnerGuid );
		m_xLeftLegIKChain.SetExclusionLists(  NUM_EXCEPTIONS, s_ausClassificationExceptionArray );
		m_xLeftLegIKChain.SetBonesArrays( AVP_Human_Description::GetThirdPersonLeftLegChainBonesArray() );
		if ( m_pxAnimation && m_pxAnimation->GetSkin() )
		{
			m_xLeftLegIKChain.SetAxesFromRestPose( m_pxAnimation->GetSkin()->BoneRestPose );
		}
		m_xLeftLegIKChain.SetIsRootBoneMover();
		m_xLeftLegIKChain.SetExtendLimit( 0.01f );
		m_xLeftLegIKChain.SetDistanceFadeRange( 20.0f, 30.0f );
		m_xIKSolver.AddChain( m_xLeftLegIKChain );

		m_xRightLegIKChain.SetLocalLOSDirection( g_xDownVector );
		m_xRightLegIKChain.SetTestAgainstEverything( m_uOwnerGuid );
		m_xRightLegIKChain.SetExclusionLists(  NUM_EXCEPTIONS, s_ausClassificationExceptionArray );
		m_xRightLegIKChain.SetBonesArrays( AVP_Human_Description::GetThirdPersonRightLegChainBonesArray() );
		if ( m_pxAnimation && m_pxAnimation->GetSkin() )
		{
			m_xRightLegIKChain.SetAxesFromRestPose( m_pxAnimation->GetSkin()->BoneRestPose );
		}
		m_xRightLegIKChain.SetIsRootBoneMover();
		m_xRightLegIKChain.SetExtendLimit( 0.01f );
		m_xRightLegIKChain.SetDistanceFadeRange( 20.0f, 30.0f );
		m_xIKSolver.AddChain( m_xRightLegIKChain );
	}

	EnableIKSolver();
}


#if defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Human_AnimationHandler::EnableFacialAnimation
// Author       : Andy Tate
// Date         : 5-Sep-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Human_AnimationHandler::EnableFacialAnimation(void)
{
	if(m_pxFacialAnimationHandler == NULL)
	{
		const Asura_Blueprint*	pxBlueprint	= m_xOwnerHuman.GetBlueprint();

		if(pxBlueprint != NULL)
		{
			Asura_Value			xValue;

			const Asura_Hash_ID	uFaceHash	= (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FACE, xValue))
											? xValue.GetHashIDValue()
											: ASURA_HASH_ID_UNSET;

			if(uFaceHash != ASURA_HASH_ID_UNSET)
			{
				// Time to create (and register) our facial-animation handler...
				m_pxFacialAnimationHandler	= new Asura_FacialAnimation_AnimationHandler_PhysicalObject (m_uOwnerGuid, uFaceHash, "FacialAnimation", AVP_AAPC_FACIALANIMATION);

				m_pxFacialAnimationHandler->SetActive(true);
				Asura_FacialAnimation_System::AddAnimationHandler(m_uOwnerGuid, m_pxFacialAnimationHandler);
			}
		}
	}
}

/// End of function 'AVP_Human_AnimationHandler::EnableFacialAnimation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Human_AnimationHandler::DisableFacialAnimation
// Author       : Andy Tate
// Date         : 5-Sep-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Human_AnimationHandler::DisableFacialAnimation(void)
{
	if(m_pxFacialAnimationHandler != NULL)
	{
		Asura_FacialAnimation_System::RemoveAnimationHandler(m_uOwnerGuid);
		delete(m_pxFacialAnimationHandler);

		m_pxFacialAnimationHandler	= NULL;
	}
}

/// End of function 'AVP_Human_AnimationHandler::DisableFacialAnimation'. ///

#endif	//defined(ASURA_USE_FACIAL_ANIMATION_SYSTEM)


/// End of file. ///
