
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Predator_Recon.h
// Author : Callum Hope
// Date   : 21-Nov-2008
//
// Desc	  : Singleton class implementation for the system that controls and draws the predator recon mode. 
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_Predator_Recon_H__
#define __AVP_Predator_Recon_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"
#include "Core/Asura_Spline.h"
#include "Network/Asura_Guid.h"
#include "Sound/Asura_Sound_Event.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class AVP_ClientActor_Predator;
class AVP_ClientEntity_Actor;

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

class AVP_Predator_Recon
{
	public:
	// PUBLIC FUNCTIONS ////////////////////////////////////////////////////////									
		void					SetOwner( AVP_ClientActor_Predator* pxOwner );
		Asura_Guid				GetOwnerGuid();
		bool					GetOwnerSet();
		bool					GetReconIsActive()								{ return m_bReconIsActive; }
		bool					GetPotentialPrey()								{ return m_bPotentialPrey; }
		bool					GetTargetPositionValid()						{ return m_bTargetPosValid; }
		bool					GetTargetJustSelected()							{ return m_bTargetJustSelected; }
		Asura_Guid				GetPreyGuid()									{ if( m_bReconIsActive || m_bPreyTargeted ){ return m_uPreyGuid; } else{ return ASURA_GUID_UNREGISTERED; } }
		Asura_Guid				GetPotentialPreyGuid()							{ if( m_bPotentialPrey ){ return m_uPotPreyGuid; } else{ return ASURA_GUID_UNREGISTERED; } }
		const Asura_Vector_3*	GetTargetPosition()								{ return &m_xTargetPos; }
		const Asura_Vector_3*	GetPreyPosition()								{ return &m_xPreyPos; }
		void					CutAllReconActions();
		void					SetSoundHandle( Asura_Handle iSoundEventHandle);
		bool					GetDistractionSoundPlaying();
		float					GetMaxEngageDistance()							{ return m_fMaxEngageDistance; }

		void		PostPlayerCreateInitialise();
		void		UpdatePredRecon();
		void		RenderPredRecon();

		static	AVP_Predator_Recon&	GetRecon()	{ return s_xRecon; }	// Accessor function to get the single recon instance anywhere in the code. 

	// PUBLIC VARIABLES ////////////////////////////////////////////////////////

	protected:
	// PROTECTED FUNCTIONS /////////////////////////////////////////////////////
		AVP_Predator_Recon();	
		~AVP_Predator_Recon();

	// PROTECTED VARIABLES /////////////////////////////////////////////////////


	private:
	// PRIVATE FUNCTIONS ///////////////////////////////////////////////////////
		bool			IsUpdateAllowed();
		bool			TestEntityAgainstLOS( const AVP_ClientEntity_Actor& xEntity, const Asura_Vector_3& xCameraPosition, Asura_Vector_3 xCameraDirection, Asura_Vector_2& xBestScreenPos ) const;
		bool			CheckAndAdjustTargetPositionHeight( Asura_Vector_3 &xPosition );
		void			UpdatePotentialPrey();
		void			UpdatePotentialTargets();
		Asura_Vector_3	FindCenterGroundPointForPrey( Asura_Guid uGuid );
		bool			CheckForDeactivateButtonPress();
		void			ActivateRecon();
		void			DeActivateRecon();
		void			ConfirmTargetLocation();
		void			SendDistractionMessages();
		void			StopAllPfxNow( bool bStopPostTargeting );
		void			UpdateHudInfo();
		void			UpdateSelectedTarget();
		void			PlayModeActivationSound();
		void			PlayModeDeActivationSound();

	// PRIVATE VARIABLES ///////////////////////////////////////////////////////
		AVP_ClientActor_Predator	*m_pxOwner;							// Pointer to the local player using this system

		Asura_Vector_3				m_xPotPreyPos;						//	 >> 3D Positions of the prey and target locations. (for PFX placement)
		Asura_Vector_3				m_xPreyPos;							//  // 
		Asura_Vector_3				m_xTargetPos;						// // 

		Asura_Guid					m_uPotPreyGuid;						// GUID of potential prey entity. 
		Asura_Guid					m_uPreyGuid;						// GUID of prey entity. 

		u_int						m_uPathMarkPFXID;					//      >> Ids of the Particle effects used. 
		u_int						m_uPathBoldMarkPFXID;				//     //
		u_int						m_uPotentialPreyPosPFXID;			//    //
		u_int						m_uPreyPosPFXID;					//   //  
		u_int						m_uTargetPosPFXID;					//  //
		u_int						m_uTargetInvPFXID;					// //
		u_int						m_uTargetSelPFXID;					////
		u_int						m_uSelectedPreyPosPFXID;			///

		float						m_fMaxEngageDistance;				// The maximum distance a prey can be to engage the recon mode. 
		float						m_fMinEngageDistance;				// the minimum distance a target must be from the player. 
		float						m_fMaxDisplacement;					// the maximum distance a target can be from the player. 
		float						m_fArrivalDistance;					// How close the prey has to be to the target location to be considered as arrived. 

		bool						m_bReconIsActive;					// STATE Flag to keep track of active state of Recon Mode.  
		bool						m_bPotentialPrey;					// STATE Flag to say a potential prey is selected. 
		bool						m_bPreyTargeted;					// STATE Flag to say a target has been set ( recon mode is not active but still in effect )
		
		bool						m_bTargetPosValid;					// Flag to indicate if the target position is a valid selection. 
		bool						m_bTargetJustSelected;				// Flag to indicate a target has just been chosen this frame. 

		const Asura_Sound_Event		*m_pxActivateSound;					// Holds the sound to be played each time recon mode is activated (after selecting a target entity.)
		const Asura_Sound_Event		*m_pxDeActivateSound;				// and the same for deactivation without setting a target position 
		Asura_Handle				m_iDistSoundEventHandle;			// handle to check if the distraction sound is playing. 
		Asura_Handle				m_iPrevSoundEventHandle;			// handle to check if the distraction sound is playing. 

		Asura_Oriented_Spline		m_xReconSpline;						// Spline structure to connect prey and target positions. 

		static	AVP_Predator_Recon	s_xRecon;							// Singleton instance of this class. 
};   

////////////////////////////////////////////////////////////////////////////////
// End of class 'AVP_Predator_Recon'.
////////////////////////////////////////////////////////////////////////////////

#endif   // ifndef __AVP_Predator_Recon_H__

////////////////////////////////////////////////////////////////////////////////
/// End of file. ///
////////////////////////////////////////////////////////////////////////////////