//////////////////////////////////////////////////////////////////////////////
//
// AVP_External_Camera.cpp
//
// Created by Anthony Gregson - 09/11/07
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//		Generic camera for all game situations where we're not in first-person.
//		Includes spectating, AVPTV, and multiplayer limbo states.
//		Can either be a flycam, or can be attached to actors as a freely-rotated third-person camera.
//		If the user wants to spectate in first person, we'll create the normal actor first-person cameras to handle that, and go through them instead.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Camera/AVP_External_Camera.h"

// Asura headers
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Timers.h"
#include "Input/Asura_ControlManager.h"
#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
#include "Interface/Asura_CommandConsole.h"
#endif // ASURA_SUPPORT_COMMAND_CONSOLE
#include "Physics/Asura_Physics_MovementQuery.h"

// Game headers
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Core/AVP_Game_Core.h"
#ifdef AVP_FLYCAM_USES_MOVEMENT_QUERIES
#include "Core/AVP_Maths.h"
#endif // AVP_FLYCAM_USES_MOVEMENT_QUERIES
#ifdef AVP_SECOND_CONTROL_FLY_CAM
#include "Debugging/Asura_ScreenShot_System.h"
#endif
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PlayerInfo/AVP_ClientEntity_PlayerInfo.h"
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
#include "GUIMenu/AVP_GUIMenu_System.h"
#endif // ASURA_USE_GUI_MENU_SYSTEM
#include "Multiplayer/AVP_Multiplayer_GameCore.h"
#include "Multiplayer/AVP_Multiplayer_ScoreBoard.h"
#ifdef AVP_SECOND_CONTROL_FLY_CAM
#include "Visibility/Asura_Visibility_System.h"
#endif

////////////////////////////
// Constructor/destructor //
////////////////////////////

AVP_External_Camera::AVP_External_Camera()
: m_xPosition(Asura_Maths::ZeroVector3)
, m_xOrientation(Asura_Maths::IdentityMatrix3x3)
, m_uTargetEntity(ASURA_GUID_UNREGISTERED)
, m_uLastKnownTarget(ASURA_GUID_UNREGISTERED)
, m_fTargetDistance(1.0f)
, m_fCurrentDistance(m_fTargetDistance / 2.0f)
, m_bFirstPersonCamera(false)
#ifdef AVP_SECOND_CONTROL_FLY_CAM
//Set this true so that the first time we creat a control map we create the one for the fly cam
, m_bRecreateFlyCamControlMap(true)
, m_bUsingFlyCam(false)
#endif
, m_pxControlMap(NULL)
{
}

AVP_External_Camera::~AVP_External_Camera()
{
#ifdef AVP_SECOND_CONTROL_FLY_CAM
	// if we quit wile in death cam mode we need to make it so that we dont recreate the fly cam after deleating the external camera
	m_bRecreateFlyCamControlMap = false;
#endif
	DestroyControlMap();
}


////////////////////
// Initialisation //
////////////////////

void AVP_External_Camera::Initialise(const Asura_Vector_3& xInitialPosition, const Asura_Matrix_3x3 xInitialOrientation)
{
	m_xPosition = xInitialPosition;
	m_xOrientation = xInitialOrientation;
	m_uTargetEntity = ASURA_GUID_UNREGISTERED;
}

void AVP_External_Camera::Initialise(Asura_Guid uTargetEntity, float fInitialDistance)
{
#ifdef AVP_SECOND_CONTROL_FLY_CAM
	CreateControlMap( m_uControlMapIdentifier );
#endif
	m_uTargetEntity = uTargetEntity;
	m_uLastKnownTarget = uTargetEntity; 
	m_fTargetDistance = fInitialDistance; 

	ResetCameraToTarget(); 
}

//////////////////////
// Update functions //
//////////////////////

void AVP_External_Camera::Update()
{
	if(GetControlMapValue(TOGGLE_MENU) != 0.0f)
	{
		//GJDH Activate PauseMenu here
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
		if( !AVP_GUIMenu_System::IsMenuActive(AVP_GUIMenu_System::GUIMENU_ID_PAUSE) )
#endif //defined(ASURA_USE_GUI_MENU_SYSTEM)
		{
			AVP_Game_Core::RequestPause();
		}
	}

	// AJK 15-10-08 Objectives are single player only for now
	if(AVP_Settings::IsSinglePlayer())
	{
		if(GetControlMapValue(TOGGLE_OBJECTIVES) != 0.0f)
		{
#if defined(ASURA_USE_GUI_MENU_SYSTEM)
			if( !AVP_GUIMenu_System::IsMenuActive(AVP_GUIMenu_System::GUIMENU_ID_OBJECTIVES) )
#endif //defined(ASURA_USE_GUI_MENU_SYSTEM)
			{
				AVP_Game_Core::RequestObjectives();
			}
		}
	}
#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	if(GetControlMapValue(TOGGLE_DVS) != 0.0f)
	{
		Asura_DebugVariableSystem::OpenDisplayDVS();
	}
#else
	if( GetControlMapValue(SCOREBOARD) )
	{
		AVP_Multiplayer_ScoreBoard::SetScoreBoardVisible( true ); 
	}
	else
	{
		AVP_Multiplayer_ScoreBoard::SetScoreBoardVisible( false ); 
	}
#endif	// defined(ASURA_DEBUG_VARIABLE_SYSTEM)

	if (GetControlMapValue(TOGGLE_FOLLOW_PLAYER))
	{
		ToggleFollowPlayer(); 
	}

	if (GetControlMapValue(TOGGLE_FIRST_PERSON))
	{
		ToggleFirstPerson(); 
	}

#ifdef AVP_SECOND_CONTROL_FLY_CAM
	// Second control inputs
	if(GetControlMapValue(TOGGLE_FLY_CAM) != 0)
	{
		m_bUsingFlyCam = !m_bUsingFlyCam;
	}

	if(GetControlMapValue(SCREENSHOT) != 0)
	{
		Asura_ScreenShot_System::TakeNormalScreenShot();
	}

	if(GetControlMapValue(TOGGLE_FREEZE) != 0)
	{
		Asura_Timers::SetSystemPause( !Asura_Timers::GetSystemPause() );
	}

	if(GetControlMapValue(TOGGLE_RENDER_ALL) != 0)
	{
		Asura_Visibility_System::ToggleRenderEverything();
	}

	if(GetControlMapValue(TOGGLE_TIME) != 0)
	{
		float fScale = Asura_Timers::GetGameTimeScale();
		fScale += m_pxControlMap->GetValue(TOGGLE_TIME);
		// MKC: To make the scale assert not pop up I take care of it myself here, and since we decreases with 0.1 all the time
		// we should be at 0 when we are below 0.1
		if( fScale < 0.09f )
		{
			fScale = 0.0f;
		}
		Asura_Timers::SetGameTimeScale( fScale );
	}

	if(GetControlMapValue(TOGGLE_WIREFRAME))
	{
		//MKC: TODO when the Asura_Debugger::Wireframe bool have been moved to a new place
	}
#endif

	AVP_ClientEntity_Actor* pxActor = AVP_ClientEntity_Actor::FindEntity(m_uTargetEntity);
	if (pxActor)
	{
		if (GetControlMapValue(NEXT_PLAYER))
		{
			FollowNextPlayer( pxActor->GetPlayerInfoGuid() ); 
		}

		if (GetControlMapValue(PREVIOUS_PLAYER))
		{
			FollowPreviousPlayer( pxActor->GetPlayerInfoGuid() ); 
		}

		if (!m_bFirstPersonCamera)
		{
			UpdateRotationControls();
			UpdateDistanceControls(); 
		}
		SetPositionFromActor(*pxActor);
	}
	else
	{
#ifdef AVP_SECOND_CONTROL_FLY_CAM
		if( m_bUsingFlyCam )
#endif
		{
			UpdateFlyCamControls();
		}
	}
}

void AVP_External_Camera::SetCamera()
{
	// For now, we're assuming the camera to change is the one already set as the current camera (like all the actor camera handlers do).
	Asura_Camera::SetPosition(m_xPosition);
	Asura_Camera::SetOrientation(m_xOrientation);
}

void AVP_External_Camera::ResetCameraToTarget()
{
	m_fCurrentDistance = m_fTargetDistance / 2.0f; 

	if (m_uTargetEntity != ASURA_GUID_UNREGISTERED)
	{
		// Follow cam reset
		AVP_ClientEntity_Actor* pxActor = AVP_ClientEntity_Actor::FindEntity(m_uTargetEntity);
		if (!pxActor)
		{
			Asura_Assert(false, "Couldn't find actor to follow with external camera.");
			return;
		}

		SetPositionAndOrientationFromActor(*pxActor);
	}
}
///////////////////////////////
// Control manager functions //
///////////////////////////////

void AVP_External_Camera::CreateControlMap(u_int uIdentifier)
{
#ifdef AVP_SECOND_CONTROL_FLY_CAM
	//MKC - See if we are creating a death cam or fly cam, if we are creating a death cam there already is a
	//fly cam se we need to destroy the current control map. To recreate the fly cam we change a bool at the end
	//of this function.
	if( !m_bRecreateFlyCamControlMap )
	{
		DestroyControlMap();
	}
#endif
	if (m_pxControlMap)
	{
		Asura_Assert(false, "Already created a control map for this camera.");
		return;
	}

	m_pxControlMap = new Asura_ControlMap(NUM_AVP_EXTERNAL_CAMERA_MAPPINGS);

	if (m_pxControlMap && Asura_ControlManager::AddDefaultControls(uIdentifier, m_pxControlMap))
	{
		m_uControlMapIdentifier = uIdentifier;

#ifdef AVP_SECOND_CONTROL_FLY_CAM
		if( m_bRecreateFlyCamControlMap )
		{
			m_pxControlMap->Activate();
			m_uTargetEntity = ASURA_GUID_UNREGISTERED;
			Platform_FillSecondControlMap();
		}
		else
#endif
		{
			Platform_FillControlMap();
		}
	}
	else
	{
		// Failed to add to the manager, tidy up.
		Asura_Assert(false, "Failed to add camera control map to manager");
		delete m_pxControlMap;
	}
#ifdef AVP_SECOND_CONTROL_FLY_CAM
	// m_bRecreateFlyCamControlMap holds if we needs to reacreate the fly cam after destroying the death cam.
	m_bRecreateFlyCamControlMap = !m_bRecreateFlyCamControlMap;
#endif
}

void AVP_External_Camera::DestroyControlMap()
{
	if (m_pxControlMap)
	{
		Asura_ControlManager::RemoveDefaultControls(m_uControlMapIdentifier);
		delete m_pxControlMap;
		m_pxControlMap = NULL;
#ifdef AVP_SECOND_CONTROL_FLY_CAM
		if( m_bRecreateFlyCamControlMap )
		{
			CreateControlMap( m_uControlMapIdentifier );
		}
#endif
	}
}

float AVP_External_Camera::GetControlMapValue(u_int uMapping) const
{
	if (!m_pxControlMap) { return 0.0f; }

	return m_pxControlMap->GetValue(uMapping);
}

/////////////////////
// Input functions //
/////////////////////
DEBUGVAR float dbg_fExternalCameraFlyCamMovementSpeed = 5.0f;
DEBUGVAR float dbg_fExternalCameraFlyCamTurnSpeed = 1.0f; // Should be additionally controlled by sensitivity options and look inversion.
DEBUGVAR float dbg_fExternalCameraFlyCamTiltLimit = Asura_Maths::Pi * 0.45f;
DEBUGVAR float dbg_fExternalCameraCollisionSphereRadius = 0.5f;

void AVP_External_Camera::UpdateFlyCamControls()
{
	if (!HasControlMap()) { return; }

#ifndef AVP_FLYCAM_USES_MOVEMENT_QUERIES
	Asura_Vector_3	xMovement;
#else
	Asura_AVector3 	xMovement;
#endif 
	
	// TODO - probably want accerations on this, although it could be done by the control map itself.
	xMovement.SetX(m_pxControlMap->GetValue(MOVE_LEFT_RIGHT));
	xMovement.SetY(m_pxControlMap->GetValue(MOVE_UP_DOWN)); 
	xMovement.SetZ(m_pxControlMap->GetValue(MOVE_FORWARDS_BACKWARDS));

	UpdateRotationControls(); 

	xMovement = xMovement * m_xOrientation;
	xMovement *= Asura_Timers::GetFrameTime() * dbg_fExternalCameraFlyCamMovementSpeed;

#ifndef AVP_FLYCAM_USES_MOVEMENT_QUERIES
	m_xPosition += xMovement;
#else
	// JG - 07 October 2008 - The code below uses the collision report to reorientate the camera away from an obstacle. Kind of promising, but as we might not 
	// need any form of flying camera... I'll leave it for reference

	// Proceed with the movement query
	const Asura_AVector3 xStartPosition( m_xPosition ); 
	const Asura_OBB xCameraOBB(dbg_fExternalCameraCollisionSphereRadius);
	Asura_Physics_MovementQuery xCameraMovementQuery( xStartPosition, xMovement, xCameraOBB, Asura_Physics_MovementQuery::ASURA_PMQCP_OBB, 4, true, true, true, true, true ); 
	Asura_Physics_MovementQuery_Results xCameraMovementQueryResults( xCameraMovementQuery ); 
	
	// Use the result to position our camera
	m_xPosition = xCameraMovementQueryResults.GetEndPosition(); 

	// Reorientate the camera towards the collision object normal 
	Asura_AVector3 xMovementDirection( xCameraMovementQueryResults.GetEndPosition() - xStartPosition ); 
	if (!xMovement.ApproximatelyEquals( xMovementDirection ))
	{
		// The movement has been restrained, correct the camera orientation
		Asura_Euler_3 xLook;
		xLook.FromMatrix(m_xOrientation);
		Asura_Vector_3 xNormalCollision( xCameraMovementQueryResults.GetLastCollisionNormal() ); 
		Asura_Matrix_3x3 xCollisionOrientation;
		Asura_Maths::OrientationFromDirection(xCollisionOrientation, xNormalCollision); 
		Asura_Euler_3 xObstacle; 
		xObstacle.FromMatrix( xCollisionOrientation ); 
		AVP_Maths::TurnExponentiallyTowards( xLook, xObstacle, (Asura_Maths::Pi / 4.0f));
		xLook[2] = 0.0f; // No roll
		xLook.ToMatrix(m_xOrientation);
	}
#endif // AVP_FLYCAM_USES_MOVEMENT_QUERIES
}

void AVP_External_Camera::UpdateRotationControls()
{
	if (!HasControlMap()) { return; }

	Asura_Euler_3 xLookInput(m_pxControlMap->GetValue(LOOK_UP_DOWN), m_pxControlMap->GetValue(LOOK_LEFT_RIGHT), 0.0f);
	Asura_Euler_3 xLook;
	xLook.FromMatrix(m_xOrientation);
	xLook += xLookInput * Asura_Timers::GetFrameTime() * dbg_fExternalCameraFlyCamTurnSpeed;
	xLook.x = Asura_Maths::ClampToRange(xLook.x, -dbg_fExternalCameraFlyCamTiltLimit, dbg_fExternalCameraFlyCamTiltLimit);
	xLook.ToMatrix(m_xOrientation);
}

void AVP_External_Camera::UpdateDistanceControls()
{
	if (!HasControlMap()) { return; }

	float fDistanceInput = m_pxControlMap->GetValue(MOVE_FORWARDS_BACKWARDS); 
	m_fCurrentDistance -= fDistanceInput * Asura_Timers::GetFrameTime() * dbg_fExternalCameraFlyCamMovementSpeed; 
	m_fCurrentDistance = Asura_Maths::ClampToRange( m_fCurrentDistance, m_fTargetDistance / 4.0f, m_fTargetDistance ); 
}

////////////////////////////////////////////////////////////
//  Spectator mode
////////////////////////////////////////////////////////////
void AVP_External_Camera::ToggleFollowPlayer()
{
	if (m_uTargetEntity != ASURA_GUID_UNREGISTERED)
	{
		m_uLastKnownTarget = m_uTargetEntity; 
		m_uTargetEntity = ASURA_GUID_UNREGISTERED; 
	}
	else
	{
		m_uTargetEntity = m_uLastKnownTarget; 
	}
}

void AVP_External_Camera::FollowNextPlayer( Asura_Guid uPlayerGuid  )
{
	RotatePlayer(uPlayerGuid, +1); 
}

void AVP_External_Camera::FollowPreviousPlayer( Asura_Guid uPlayerGuid )
{
	RotatePlayer(uPlayerGuid, -1); 
}

void AVP_External_Camera::ToggleFirstPerson()
{
	//m_bFirstPersonCamera = !m_bFirstPersonCamera; 
}

void AVP_External_Camera::RotatePlayer( Asura_Guid uPlayerGuid, int iIncrement)
{
	if (m_uTargetEntity == ASURA_GUID_UNREGISTERED)
	{
		if (m_uLastKnownTarget == ASURA_GUID_UNREGISTERED)
		{
			Asura_Guid uFirstPlayer = AVP_Multiplayer_GameCore::FindFirstPlayerAlive();
			const AVP_ClientEntity_PlayerInfo* pxPlayerInfo = AVP_ClientEntity_PlayerInfo::FindEntity( uFirstPlayer ); 
			if (pxPlayerInfo)
			{
				m_uTargetEntity = pxPlayerInfo->GetPuppetGuid(); 
			}
			return; 
		}
	}
	
	if (m_uTargetEntity == ASURA_GUID_UNREGISTERED)
	{
		m_uTargetEntity = m_uLastKnownTarget;
	}

	if (m_uTargetEntity != ASURA_GUID_UNREGISTERED)
	{
		Asura_Guid uCurrentPlayerGuid = uPlayerGuid;
		Asura_Guid uNewPuppetGuid = ASURA_GUID_UNREGISTERED;
		int iNumberOfPlayers = AVP_Multiplayer_GameCore::GetNumberOfPlayers();
		for (int iCounter = 0; iCounter < iNumberOfPlayers; ++iCounter)
		{
			Asura_Guid uNewPlayerGuid = ASURA_GUID_UNREGISTERED;  
			if (iIncrement < 0)
			{
				uNewPlayerGuid = AVP_Multiplayer_GameCore::FindPreviousPlayer( uCurrentPlayerGuid );
			}
			else
			{
				uNewPlayerGuid = AVP_Multiplayer_GameCore::FindNextPlayer( uCurrentPlayerGuid );
			}
			AVP_ClientEntity_PlayerInfo* pxNewPlayerInfo = AVP_ClientEntity_PlayerInfo::FindEntity( uNewPlayerGuid ); 
			uCurrentPlayerGuid = uNewPlayerGuid; 
			if (pxNewPlayerInfo && 
				pxNewPlayerInfo->GetPuppetGuid() != ASURA_GUID_UNREGISTERED &&
				pxNewPlayerInfo->GetPuppetGuid() != m_uTargetEntity)
			{
				uNewPuppetGuid = pxNewPlayerInfo->GetPuppetGuid(); 
				break;
			}
		}

		if (uNewPuppetGuid != ASURA_GUID_UNREGISTERED)
		{
			m_uTargetEntity = uNewPuppetGuid; 
		}
		else
		{
			Asura_Guid uFirstPlayer = AVP_Multiplayer_GameCore::FindFirstPlayerAlive();
			const AVP_ClientEntity_PlayerInfo* pxPlayerInfo = AVP_ClientEntity_PlayerInfo::FindEntity( uFirstPlayer ); 
			if (pxPlayerInfo)
			{
				m_uTargetEntity = pxPlayerInfo->GetPuppetGuid(); 
			}
		}
	}
}

///////////////////////////////////////
// Positional data from target actor //
///////////////////////////////////////

// TODO: Some optional interpolation?
void AVP_External_Camera::SetPositionAndOrientationFromActor(AVP_ClientEntity_Actor& xActor)
{
	Asura_Euler_3 xActorOriEuler; 
	xActorOriEuler.FromMatrix(xActor.GetCurrentOrientation()); 
	xActorOriEuler.x -= Asura_Maths::Pi / 4.0f; 
	xActorOriEuler.x = Asura_Maths::ClampToRange(xActorOriEuler.x, -dbg_fExternalCameraFlyCamTiltLimit, dbg_fExternalCameraFlyCamTiltLimit);
	xActorOriEuler.ToMatrix(m_xOrientation); 

	SetPositionFromActor(xActor);
}

void AVP_External_Camera::SetPositionFromActor(AVP_ClientEntity_Actor& xActor)
{
	AVP_Actor_AnimationHandler* pxAnimHandler = xActor.GetAnimationHandler();
	if ( pxAnimHandler != NULL )
	{

		const Asura_Vector_3&			xClientPosition		= xActor.GetCurrentPosition();
		Asura_Matrix_3x3				xAnimOrientation	= xActor.GetCurrentOrientation();
		u_int	uFlags	= ASURA_HIERARCHY_RENDER_NO_FRUSTUM_TEST | ASURA_HIERARCHY_RENDER_NO_MODULELINK_TEST | ASURA_HIERARCHY_RENDER_DONTRENDER | ASURA_HIERARCHY_RENDER_UPDATEBONESALWAYS;
		if ( pxAnimHandler->IsRagdoll() )
		{
			uFlags |= ASURA_HIERARCHY_RENDER_BONES_IN_WORLDSPACE;
		}

		// Ensure that the animation has been updated this frame so their marker points etc. are up to date...
		pxAnimHandler->GetAnimation()->Update(xClientPosition, xAnimOrientation, xActor.GetShapeInstance().GetAnimTimestep(), uFlags);
	}

	if (m_bFirstPersonCamera)
	{
		static const Asura_Hash_ID uHeadTargetMarkerHashID = 0xe5a85d04; // Asura_GetHashID("target_2")
		if (!xActor.GetWorldSpaceMarkerPoint(uHeadTargetMarkerHashID, &m_xPosition, &m_xOrientation))
		{
			m_xPosition = xActor.GetCurrentPosition();
		}
	}
	else
	{
		// Collision test (movement query) to position the camera from the center of the actor
		xActor.GetChestPosition( m_xPosition );

		// Calculate the maximum displacement if the camera doesn't hit anything
		Asura_AVector3 xDisplacement(m_xOrientation.e13, m_xOrientation.e23, m_xOrientation.e33); 
		xDisplacement *= -m_fCurrentDistance; 

		// Proceed with the movement query
		const Asura_AVector3 xStartPosition( m_xPosition ); 
		const Asura_OBB xCameraOBB(dbg_fExternalCameraCollisionSphereRadius);
		Asura_Physics_MovementQuery xCameraMovementQuery( xStartPosition, xDisplacement, xCameraOBB); 
		Asura_Physics_MovementQuery_Results xCameraMovementQueryResults( xCameraMovementQuery ); 
		
		// Use the results to position our camera
		m_xPosition = xCameraMovementQueryResults.GetEndPosition(); 
	}
}

/// End of file. ///
