
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_Plane.h
// Author : Andy Tate
// Date   : 31-Aug-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_PLANE_H__
#define __AVP_PLANE_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_Plane
// Author       : Andy Tate
// Date         : 31-Aug-2007
//
// Description  : Class used to handle plane calculations.
////////////////////////////////////////////////////////////////////////////////

class AVP_Plane
{
public:
	AVP_Plane(void)																						{}
	AVP_Plane(const AVP_Plane& xOther)										  : m_xXYZD(xOther.m_xXYZD)	{}
	AVP_Plane(const float fX, const float fY, const float fZ, const float fD) : m_xXYZD(fX, fY, fZ, fD)	{}
	AVP_Plane(const Asura_AVector3& xPoint, const Asura_AVector3& xNormal);
    AVP_Plane(const Asura_Vector_3& xPoint, const Asura_Vector_3& xNormal);

	AVP_Plane& operator=(const AVP_Plane& xOther) { m_xXYZD = xOther.m_xXYZD; return *this; }

	// Asura_Vector_3 functions.
	float CalculateClosestDistance(const Asura_Vector_3& xPosition) const
	{
		return CalculateClosestDistance( Asura_AVector4( xPosition, 0.0f ) );
	}
	bool CalculateIntersectionT( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, float& fT ) const
	{
		return CalculateIntersectionT( Asura_AVector4( xRayOrigin, 0.0f ), Asura_AVector4( xRayDirection, 0.0f ), fT );
	}
	void CalculateClosestPoint( const Asura_Vector_3& xPosition, Asura_Vector_3& xClosest ) const;
	bool CalculateIntersectionPoint( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, Asura_Vector_3& xPoint ) const;

	// Asura_AVector3 functions.
	float CalculateClosestDistance(const Asura_AVector3& xPosition) const
	{
		return CalculateClosestDistance( Asura_AVector4( xPosition, 0.0f ) );
	}
	bool CalculateIntersectionT( const Asura_AVector3& xRayOrigin, const Asura_AVector3& xRayDirection, float& fT ) const
	{
		return CalculateIntersectionT( Asura_AVector4( xRayOrigin, 0.0f ), Asura_AVector4( xRayDirection, 0.0f ), fT );
	}
	void CalculateClosestPoint( const Asura_AVector3& xPosition, Asura_AVector3& xClosest ) const;
	bool CalculateIntersectionPoint( const Asura_AVector3& xRayOrigin, const Asura_AVector3& xRayDirection, Asura_AVector3& xPoint ) const;

	// Asura_AVector4 functions (most efficient).
	// Note that the distance can come out negative, depending upon what side of the plane the position is.
	float CalculateClosestDistance(const Asura_AVector4& xPosition) const;
	// Calculates 't', where the intersection point = xRayOrigin + ( t * xRayDirection ).
	// Return if the ray intersects.
	bool CalculateIntersectionT( const Asura_AVector4& xRayOrigin, const Asura_AVector4& xRayDirection, float& fT ) const;

private:
	Asura_AVector4 m_xXYZD;
};

////////////////////////////////////////////////////////////////////////////////
// Inlined functions...
////////////////////////////////////////////////////////////////////////////////

inline AVP_Plane::AVP_Plane(const Asura_AVector3& xPoint, const Asura_AVector3& xNormal)
: m_xXYZD(xNormal, -(xPoint * xNormal))
{
	////////////////////////////////////////////////////////////////////////////////
	// Using standard plane equations, where:
	// - P = xPoint
	// - n = xNormal
	//
	// P.n + d = 0  =>  d = -P.n
	////////////////////////////////////////////////////////////////////////////////
}

inline AVP_Plane::AVP_Plane(const Asura_Vector_3& xPoint, const Asura_Vector_3& xNormal)
: m_xXYZD(xNormal, -(xPoint * xNormal))
{
    ////////////////////////////////////////////////////////////////////////////////
    // Using standard plane equations, where:
    // - P = xPoint
    // - n = xNormal
    //
    // P.n + d = 0  =>  d = -P.n
    ////////////////////////////////////////////////////////////////////////////////
}

inline float AVP_Plane::CalculateClosestDistance(const Asura_AVector4& xPosition) const
{
	float	fDistance;

	////////////////////////////////////////////////////////////////////////////////
	// Using standard plane equations, where:
	// - Q = xPosition
	// - n = normal of the plane
	// - r = closest distance to plane
	// 
	// r = n.Q + d
	////////////////////////////////////////////////////////////////////////////////
	fDistance	= m_xXYZD.DotProduct3(xPosition) + m_xXYZD.GetW();
	return(fDistance);
}

inline void AVP_Plane::CalculateClosestPoint( const Asura_Vector_3& xPosition, Asura_Vector_3& xClosest ) const
{
    const float fDistance = CalculateClosestDistance( xPosition );
    xClosest = xPosition;
    xClosest.x -= fDistance * m_xXYZD.GetX();
    xClosest.y -= fDistance * m_xXYZD.GetY();
    xClosest.z -= fDistance * m_xXYZD.GetZ();
}

inline void AVP_Plane::CalculateClosestPoint( const Asura_AVector3& xPosition, Asura_AVector3& xClosest ) const
{
	const float fDistance = CalculateClosestDistance( xPosition );
	xClosest.SetX( xPosition.GetX() - fDistance * m_xXYZD.GetX() );
	xClosest.SetY( xPosition.GetY() - fDistance * m_xXYZD.GetY() );
	xClosest.SetZ( xPosition.GetZ() - fDistance * m_xXYZD.GetZ() );
}

inline bool AVP_Plane::CalculateIntersectionT( const Asura_AVector4& xRayOrigin, const Asura_AVector4& xRayDirection, float& fT ) const
{
	// c.f. Geometric Tools For Computer Graphics - Schneider, Eberly, pp 482-484.
	static const float s_fEpsilon = 0.000001f;
	const float fNum = m_xXYZD.DotProduct3( xRayOrigin ) + m_xXYZD.GetW();
	const float fDen = m_xXYZD.DotProduct3( xRayDirection );
	if ( -s_fEpsilon <= fDen && fDen <= s_fEpsilon )
	{
		fT = 0.0f;
		return ( -s_fEpsilon <= fNum && fNum < s_fEpsilon );
	}
	else
	{
		fT = -fNum / fDen;
		return true;
	}
}

inline bool AVP_Plane::CalculateIntersectionPoint( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, Asura_Vector_3& xPoint ) const
{
	float fT;
	if ( CalculateIntersectionT( xRayOrigin, xRayDirection, fT ) )
	{
		xPoint = xRayOrigin + ( xRayDirection * fT );
		return true;
	}
	else
	{
		return false;
	}
}

inline bool AVP_Plane::CalculateIntersectionPoint( const Asura_AVector3& xRayOrigin, const Asura_AVector3& xRayDirection, Asura_AVector3& xPoint ) const
{
	float fT;
	if ( CalculateIntersectionT( xRayOrigin, xRayDirection, fT ) )
	{
		xPoint = xRayOrigin + ( xRayDirection * fT );
		return true;
	}
	else
	{
		return false;
	}
}

#endif   // ifndef __AVP_PLANE_H__

/// End of file. ///
