
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_StateMachine.cpp
// Author : Andy Tate
// Date   : 12-Mar-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Core/AVP_StateMachine.h"

// Asura independent header-files...
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_Logfile.h"

// Game independent header-files...
#include "Core/States/AVP_State.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#ifdef AVP_STATEMACHINE_LOGGING
static const char* MSG_LOGFILE_AVP_STATE_CHANGE = "logfiles\\state_change.log";
#endif

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_StateMachine::AVP_StateMachine
// Author       : Andy Tate
// Date         : 12-Mar-2007
////////////////////////////////////////////////////////////////////////////////

AVP_StateMachine::AVP_StateMachine(u_int uNumStates)
: m_uFlags(0)
, m_fTimeInCurrentState(0.0f)
{
	Asura_Assert(uNumStates > 0, "State machine created without any possible states.");
	Asura_Assert(uNumStates < 256, "Too many states to fit into bitfield");

	m_uNumStates = uNumStates;

	m_ppxStates	= new AVP_State*[m_uNumStates];

	for(u_int u = 0; u < m_uNumStates; u++)
	{
		m_ppxStates[u]	= NULL;
	}
}

/// End of function 'AVP_StateMachine::AVP_StateMachine'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_StateMachine::~AVP_StateMachine
// Author       : Andy Tate
// Date         : 12-Mar-2007
////////////////////////////////////////////////////////////////////////////////

AVP_StateMachine::~AVP_StateMachine()
{
	for(u_int u = 0; u < m_uNumStates; u++)
	{
		delete(m_ppxStates[u]);
	}

	delete[](m_ppxStates);
}

/// End of function 'AVP_StateMachine::~AVP_StateMachine'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_StateMachine::RegisterState
// Author       : Andy Tate
// Date         : 14-Mar-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_StateMachine::RegisterState(const u_int uState,
									 AVP_State* pxState)
{
	if(uState < m_uNumStates)
	{
		if(m_ppxStates[uState] == NULL)
		{
			m_ppxStates[uState]	= pxState;
			return(true);
		}
		else
		{
			Asura_Assert(false, "Already have a state for this entry.");
			return(false);
		}
	}
	else
	{
		Asura_Assert(false, "Specified state is out of range.");
		return(false);
	}
}

/// End of function 'AVP_StateMachine::RegisterState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_StateMachine::ChangeState
// Author       : Andy Tate
// Date         : 12-Mar-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_StateMachine::ChangeState(const u_int uNewState)
{
	if(m_bHasNewState)
	{
		Asura_Assert(false, "Already have a state to change to.");
		return(false);
	}
	else if (uNewState < m_uNumStates)
	{
		m_uNextState	= uNewState;
		m_bHasNewState	= true;

#ifdef AVP_STATEMACHINE_LOGGING
		AVP_State* pxState = m_ppxStates[uNewState];
		const float fTime = Asura_Timers::GetGlobalGameTime();
		if (pxState)
		{
			Asura_SimpleLogfileOutput(( MSG_LOGFILE_AVP_STATE_CHANGE, "%7.2f: %s change to: %s\n", Asura_Maths::FtoD(fTime), GetStateMachineName(), pxState->GetName() ));
		}
		else
		{
			Asura_SimpleLogfileOutput(( MSG_LOGFILE_AVP_STATE_CHANGE, "%7.2f: %s change to state %d\n", Asura_Maths::FtoD(fTime), GetStateMachineName(), uNewState ));
		}
#endif //#ifdef AVP_STATEMACHINE_LOGGING
		return(true);
	}
	else
	{
		Asura_Assert(false, "Specified state is out of range.");
		return(false);
	}
}

void AVP_StateMachine::LogInitialise(void)
{
#ifdef AVP_STATEMACHINE_LOGGING
	Asura_SimpleLogfileClear( MSG_LOGFILE_AVP_STATE_CHANGE );
	Asura_SimpleLogfileOutput(( MSG_LOGFILE_AVP_STATE_CHANGE, "\n///////////////// AVP STATE MACHINE LOG\n\n\n"));
#endif
}


/// End of function 'AVP_StateMachine::ChangeState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_StateMachine::Update
// Author       : Andy Tate
// Date         : 12-Mar-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_StateMachine::Update(void)
{
	AVP_State*	pxState;

	// Do we have a state to change to?
	if(m_bHasNewState)
	{
		AVP_State*	pxOldState	= m_ppxStates[m_uCurrentState];
		AVP_State*	pxNewState	= m_ppxStates[m_uNextState];

		if(pxOldState != NULL)
		{
			pxOldState->Exit(m_uNextState);
		}

		if(pxNewState != NULL)
		{
			pxNewState->Enter(m_uCurrentState);
		}

		m_fTimeInCurrentState	= 0.0f;
		m_uCurrentState			= m_uNextState;
		m_bHasNewState			= false;
		pxState					= pxNewState;
	}
	else
	{
		m_fTimeInCurrentState	+= Asura_Timers::GetActualFrameTime();
		pxState					 = m_ppxStates[m_uCurrentState];
	}

	// Handle this state...
	if(pxState != NULL)
	{
		pxState->Update();
	}
}

/// End of function 'AVP_StateMachine::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_StateMachine::Render
// Author       : Andy Tate
// Date         : 12-Mar-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_StateMachine::Render(void) const
{
	const AVP_State*	pxState	= m_ppxStates[m_uCurrentState];

	if(pxState != NULL)
	{
		pxState->Render();
	}
}

/// End of function 'AVP_StateMachine::Render'. ///


/// End of file. ///
