///////////////////////////////////////////////////////////////////////////////
//
// File: AVP_CameraVolumeData.cpp
//
// Created by Martin G.Harris (14/9/2006)
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
// Class to contain project specific camera types and their associated data.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/CameraVolume/AVP_CameraVolumeData.h"

#ifdef ASURA_USE_CAMERA_VOLUME

#include "ChunkSystem/Asura_Chunk_Macros.h"

///////////////
// Constants //
///////////////
// Time to interpolate the camera properties when switching to this volume
// (unless we are doing a camera cut).
const float fCAMERA_DEFAULT_TIME_TO_RESET = 5.0f;

// Default FOV in game (in radians).
const float fCAMERA_DEFAULT_FOV = Asura_Maths::Deg180/3.0f; // 60 degrees


// Default values for camera properties.
const AVP_CameraVolumeData::ChunkData_Shared g_xShared_Default =
{
	static_cast<int>(AVP_CameraVolumeData::TYPE_FIXED_ANGLE), // int m_iType;
	0,									// u_int m_uFlags;
	fCAMERA_DEFAULT_TIME_TO_RESET,		// float m_fTimeToReset;
	fCAMERA_DEFAULT_FOV,				// float m_fFOV;
	ASURA_GUID_UNREGISTERED				// Asura_Guid m_uTarget;
};

// Default values for fixed angle camera.
const AVP_CameraVolumeData::ChunkData_FixedAngle g_xFixedAngle_Default =
{
	0.0f,						// float m_fRotation;
	Asura_Maths::Deg180/6.0f,	// float m_fElevation;
	6.0f,						// float m_fDistance;
};

// Sizes of chunks
const size_t AVP_CameraVolumeData::s_auChunkSize[NUM_CAMERA_TYPES] = 
{
	sizeof(ChunkData_FixedAngle),	// TYPE_FIXED_ANGLE
	sizeof(ChunkData_StaticCam),	// TYPE_STATIC_CAM
	sizeof(ChunkData_RailCam),		// TYPE_RAIL_CAM
};

//////////////////////////////
// AVP_CameraVolumeData //
//////////////////////////////
AVP_CameraVolumeData::AVP_CameraVolumeData()
{
	// Initialise to default state.
	Reset();
}

// Reset to some known default values.
void AVP_CameraVolumeData::Reset()
{
	// Shared
	m_xShared = g_xShared_Default;

	// Fixed angle
	m_xFixedAngle = g_xFixedAngle_Default;

	// Static cam
	m_xStaticCam.m_xPosition.Zero();

	// Rail cam
	m_xRailCam.m_xCameraRailStart.Zero();
	m_xRailCam.m_xCameraRailEnd.Zero();
	m_xRailCam.m_xPlayerRailStart.Zero();
	m_xRailCam.m_xPlayerRailEnd.Zero();
}


////////////////////////////
// Chunk buffer functions //
////////////////////////////
int AVP_CameraVolumeData::GetChunkBufferSize() const
{
	// Version, shared property data and second version number.
	u_int uSize = sizeof(int) + sizeof(ChunkData_Shared) + sizeof(int);

	const int iType = GetType();
	if (iType>=0 && iType<NUM_CAMERA_TYPES)
	{
		// Specific chunk data.
		uSize += s_auChunkSize[iType];
	}
	else
	{
		Asura_Assert(0, "Unknown camera volume type");
	}

	return uSize;
}

void AVP_CameraVolumeData::FillInChunkBuffer(char*& pcChunkBuffer)
{
	// Write version
	int iVersion = ChunkData_Shared::CURRENT_VERSION;
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iVersion, int);

	// Write shared camera property data.
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xShared.m_iType, int);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xShared.m_uFlags, u_int);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xShared.m_fTimeToReset, float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xShared.m_fFOV, float);
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xShared.m_uTarget, Asura_Guid);

	// Write the specific data version.
	switch (GetType())
	{
	default:
		Asura_Assert(0, "Error: Invalid camera volume type");
		return;

	case TYPE_FIXED_ANGLE:
		iVersion = ChunkData_FixedAngle::CURRENT_VERSION;
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iVersion, int);

		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xFixedAngle.m_fRotation,  float);
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xFixedAngle.m_fElevation, float);
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, m_xFixedAngle.m_fDistance,  float);
		break;

	case TYPE_STATIC_CAM:
		iVersion = ChunkData_StaticCam::CURRENT_VERSION;
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iVersion, int);

		ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer, m_xStaticCam.m_xPosition, Asura_Vector_3, float, 3);
		break;

	case TYPE_RAIL_CAM:
		iVersion = ChunkData_RailCam::CURRENT_VERSION;
		ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iVersion, int);

		ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer, m_xRailCam.m_xCameraRailStart, Asura_Vector_3, float, 3);
		ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer, m_xRailCam.m_xCameraRailEnd,   Asura_Vector_3, float, 3);
		ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer, m_xRailCam.m_xPlayerRailStart, Asura_Vector_3, float, 3);
		ASURA_CHUNK_ARRAY_FILLANDVERIFY(pcChunkBuffer, m_xRailCam.m_xPlayerRailEnd,   Asura_Vector_3, float, 3);

		break;
	}
}

void AVP_CameraVolumeData::InitialiseFromChunkBuffer(const char*& pcChunkBuffer)
{
	int iVersion;
	ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion, int);

	if (iVersion<=0 || iVersion>ChunkData_Shared::CURRENT_VERSION)
	{
		// Version 0 means that the property data is missing.
		Asura_Assert(iVersion == 0, "AVP_CameraVolumeData: Unrecognised chunk version");
		Reset(); // Reset to default values.
		return;
	}
	
	// Shared Version is greater than zero.
	if (iVersion == ChunkData_Shared::CURRENT_VERSION)
	{
		// Version 1
		ASURA_CHUNK_EXTRACT(pcChunkBuffer, m_xShared, ChunkData_Shared);
	}
	else
	{
		Asura_Assert(0,"Camera volume data unhandled chunk version");
	}

	// Extract version number for specific data.
	ASURA_CHUNK_EXTRACT( pcChunkBuffer, iVersion, int);

	bool bLoaded = false;

	switch (GetType())
	{
	default:
		break;

	case TYPE_FIXED_ANGLE:
		if (iVersion == ChunkData_FixedAngle::CURRENT_VERSION)
		{
			bLoaded = true;
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_xFixedAngle, ChunkData_FixedAngle);
		}
		break;

	case TYPE_STATIC_CAM:
		if (iVersion == ChunkData_StaticCam::CURRENT_VERSION)
		{
			bLoaded = true;
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_xStaticCam, ChunkData_StaticCam);
		}
		break;

	case TYPE_RAIL_CAM:
		if (iVersion == ChunkData_RailCam::CURRENT_VERSION)
		{
			bLoaded = true;
			ASURA_CHUNK_EXTRACT( pcChunkBuffer, m_xRailCam, ChunkData_RailCam);
		}
		break;
	}

	if (!bLoaded)
	{
		Asura_Assert(0, "Error: Invalid camera volume type.");
		Reset(); // Reset to default values.
	}
}

// Allow server entities to skip the relevant number of bytes.
void AVP_CameraVolumeData::SkipChunk(const char*& pcChunkBuffer)
{
	// Read chunk data in to a dummy variable.
	AVP_CameraVolumeData xDummy;
	xDummy.InitialiseFromChunkBuffer(pcChunkBuffer);
}

// Fallback code for when the client entity is missing (ie chunk data version = 0).
int AVP_CameraVolumeData::GetChunkBufferSize_V0()
{
	return sizeof(int);
}

void AVP_CameraVolumeData::FillInChunkBuffer_V0(char*& pcChunkBuffer)
{
	// Write a zero version number to mark that there is no property data.
	int iVersion = 0;
	ASURA_CHUNK_FILLANDVERIFY(pcChunkBuffer, iVersion, int);
}

//////////////
// Displace //
//////////////
// Displace any positions by the specified offset.
#ifdef ASURA_TOOLS
void AVP_CameraVolumeData::Displace(const Asura_Vector_3& xMove)
{
	// Static cam
	if (m_xStaticCam.m_xPosition != Asura_Maths::ZeroVector3)
	{
		m_xStaticCam.m_xPosition += xMove;
	}

	// Rail cam
	if (m_xRailCam.m_xCameraRailStart != Asura_Maths::ZeroVector3)
	{
		m_xRailCam.m_xCameraRailStart += xMove;
		m_xRailCam.m_xCameraRailEnd   += xMove;
	}

	if (m_xRailCam.m_xPlayerRailStart != Asura_Maths::ZeroVector3)
	{
		m_xRailCam.m_xPlayerRailStart += xMove;
		m_xRailCam.m_xPlayerRailEnd   += xMove;
	}
}
#endif //ASURA_TOOLS


#endif //ASURA_USE_CAMERA_VOLUME
