
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_EntityDamage_Request.h
// Author : Andy Tate
// Date   : 20-Feb-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_ENTITYDAMAGE_REQUEST_H__
#define __AVP_ENTITYDAMAGE_REQUEST_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"

#include "Entity/Damage/AVP_EntityDamage.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class Asura_LOS_InputData;
class Asura_LOS_OutputData;
class Asura_Volume;

class AVP_ServerEntity_PhysicalObject;
class AVP_ServerEntity_Container;
class Asura_Physics_RigidBodyInstance;


////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_EntityDamage_Request
// Author       : Andy Tate
// Date         : 20-Feb-2008
//
// Description  : Abstract class use to request damage for entities.
////////////////////////////////////////////////////////////////////////////////

class AVP_EntityDamage_Request : public AVP_EntityDamage
{
public:
	// Public functions...

	// Abstract functions...
	bool	Apply(AVP_ServerEntity_PhysicalObject& xFrom) const;

	// Accessor functions...
	Asura_Hash_ID	GetFromMarkerPointHash(void) const	{ return(m_uFromMarkerPointHash); }


protected:
	// Protected variables...
	Asura_Hash_ID	m_uFromMarkerPointHash;	// The hash-ID of the source marker-point for the entity responsible for this damage.

	// Protected functions...
	AVP_EntityDamage_Request(void)
	{
#if defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
		RegisterWithCommandConsole();
#endif	//defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	}
	//virtual ~AVP_EntityDamage_Request();

	AVP_EntityDamage_Request(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
							 const Asura_Hash_ID uFromMarkerPointHash, const Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED)
	: PARENT(uTypeHash, uFromGuid, uParentGuid)
	, m_uFromMarkerPointHash(uFromMarkerPointHash)
	{
#if defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
		RegisterWithCommandConsole();
#endif	//defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	}

	AVP_EntityDamage_Request(const AVP_EntityDamage_Request& xOther)
	: PARENT(xOther)
	, m_uFromMarkerPointHash(xOther.m_uFromMarkerPointHash)
	{
#if defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
		RegisterWithCommandConsole();
#endif	//defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	}

	AVP_EntityDamage_Request& operator=(const AVP_EntityDamage_Request& xOther)
	{
		m_uFromMarkerPointHash	= xOther.m_uFromMarkerPointHash;

		PARENT::operator=(xOther);
		return(*this);
	}


private:
	// Private typedefs...
	typedef AVP_EntityDamage PARENT;

	// Private variables...

	// Private functions...
#if defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	static void	RegisterWithCommandConsole(void);
#endif	//defined(AVP_USE_DAMAGE_DEBUGGING) && defined(ASURA_SUPPORT_COMMAND_CONSOLE)

};   // End of class 'AVP_EntityDamage_Request'.



////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_EntityDamage_RequestSpatial
// Author       : Andy Tate
// Date         : 20-Feb-2008
//
// Description  : Class used to request spatial-damage for entities.
////////////////////////////////////////////////////////////////////////////////

class AVP_EntityDamage_RequestSpatial : public AVP_EntityDamage_Request
{
public:
	// Public functions...

	// Accessor functions...
	float		GetMaxAmount(void) const			{ return(m_fMaxAmount); }
	float		GetMaxImpulseStrength(void) const	{ return(m_fMaxImpulseStrength); }
	Asura_Guid	GetIgnoreGuid(void) const			{ return(m_uIgnoreGuid); }


protected:
	// Protected variables...
	float		m_fMaxAmount;			// The maximum amount of damage to be caused.
	float		m_fMaxImpulseStrength;	// The maximum amount of impulse to be applied with this damage.
	Asura_Guid	m_uIgnoreGuid;			// The GUID of the entity to ignore when requesting damage.

	// Protected functions...
	AVP_EntityDamage_RequestSpatial(void)	{}
	//virtual ~AVP_EntityDamage_RequestSpatial();

	AVP_EntityDamage_RequestSpatial(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
									const Asura_Hash_ID uFromMarkerPointHash,
									const float fMaxAmount, const float fMaxImpulseStrength, const Asura_Guid uIgnoreGuid, const Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED)
	: PARENT(uTypeHash, uFromGuid, uFromMarkerPointHash, uParentGuid)
	, m_fMaxAmount(fMaxAmount)
	, m_fMaxImpulseStrength(fMaxImpulseStrength)
	, m_uIgnoreGuid(uIgnoreGuid)
	{
	}

	AVP_EntityDamage_RequestSpatial(const AVP_EntityDamage_RequestSpatial& xOther)
	: PARENT(xOther)
	, m_fMaxAmount(xOther.m_fMaxAmount)
	, m_fMaxImpulseStrength(xOther.m_fMaxImpulseStrength)
	, m_uIgnoreGuid(xOther.m_uIgnoreGuid)
	{
	}

	AVP_EntityDamage_RequestSpatial& operator=(const AVP_EntityDamage_RequestSpatial& xOther)
	{
		m_fMaxAmount			= xOther.m_fMaxAmount;
		m_fMaxImpulseStrength	= xOther.m_fMaxImpulseStrength;
		m_uIgnoreGuid			= xOther.m_uIgnoreGuid;

		PARENT::operator=(xOther);
		return(*this);
	}

	// Query functions...
	bool	LOSTest(AVP_ServerEntity_PhysicalObject& xFrom, const Asura_Vector_3& xPosition, const Asura_Vector_3& xDirection, const float fMaxRange, const float fParametric) const;

	bool	ApplyToEntity(AVP_ServerEntity_PhysicalObject& xFrom, const Asura_LOS_InputData& xLOSInput, const Asura_LOS_OutputData& xLOSOutput, const float fParametric) const;
	bool	ApplyToEnvironment(const Asura_LOS_InputData& xLOSInput, const Asura_LOS_OutputData& xLOSOutput) const;
	bool	ApplyToBotanicals( const Asura_LOS_InputData& xLOSInput ) const;

private:
	// Private typedefs...
	typedef AVP_EntityDamage_Request PARENT;

	// Private variables...

	// Private functions...

};   // End of class 'AVP_EntityDamage_RequestSpatial'.



////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_EntityDamage_RequestVictim
// Author       : Andy Tate
// Date         : 20-Feb-2008
//
// Description  : Class used to request damage for a specific "victim" entity.
////////////////////////////////////////////////////////////////////////////////

class AVP_EntityDamage_RequestVictim : public AVP_EntityDamage_Request
{
public:
	// Public functions...
	AVP_EntityDamage_RequestVictim(void)	{}
	//virtual ~AVP_EntityDamage_RequestVictim();

	AVP_EntityDamage_RequestVictim(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
								   const Asura_Hash_ID uFromMarkerPointHash,
								   const Asura_Guid uEntityToDamage, const float fAmount,
								   const Asura_Vector_3& xPointOfImpact, const Asura_Vector_3& xDirection, const Asura_Vector_3& xNormal,
								   const float fImpulseStrength = 0.0f, const int iBoneID = -1, const bool bBlocked = false,
								   const Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED,
								   const int iAttachmentID = -1,
								   Asura_Hash_ID uMaterialResponseHashID = ASURA_HASH_ID_UNSET,
								   const bool bStealth = false,
								   const bool bDoNotFlinch = false)
	: PARENT(uTypeHash, uFromGuid, uFromMarkerPointHash, uParentGuid)
	, m_uEntityToDamage(uEntityToDamage)
	, m_fAmount(fAmount)
	, m_xPointOfImpact(xPointOfImpact)
	, m_xDirection(xDirection)
	, m_xNormal(xNormal)
	, m_fImpulseStrength(fImpulseStrength)
	, m_iBoneID(iBoneID)
	, m_iAttachmentID(iAttachmentID)
	, m_uMaterialResponseHashID(uMaterialResponseHashID)
	, m_bBlocked(bBlocked)
	, m_bStealth(bStealth)
	, m_bDoNotFlinch(bDoNotFlinch)
	{
	}

	AVP_EntityDamage_RequestVictim(const AVP_EntityDamage_RequestVictim& xOther)
	: PARENT(xOther)
	, m_uEntityToDamage(xOther.m_uEntityToDamage)
	, m_fAmount(xOther.m_fAmount)
	, m_xPointOfImpact(xOther.m_xPointOfImpact)
	, m_xDirection(xOther.m_xDirection)
	, m_xNormal(xOther.m_xNormal)
	, m_fImpulseStrength(xOther.m_fImpulseStrength)
	, m_iBoneID(xOther.m_iBoneID)
	, m_iAttachmentID(xOther.m_iAttachmentID)
	, m_uMaterialResponseHashID(xOther.m_uMaterialResponseHashID)
	, m_bBlocked(xOther.m_bBlocked)
	, m_bStealth(xOther.m_bStealth)
	, m_bDoNotFlinch(xOther.m_bDoNotFlinch)
	{
	}

	AVP_EntityDamage_RequestVictim& operator=(const AVP_EntityDamage_RequestVictim& xOther)
	{
		m_uEntityToDamage	= xOther.m_uEntityToDamage;
		m_fAmount			= xOther.m_fAmount;
		m_xPointOfImpact	= xOther.m_xPointOfImpact;
		m_xDirection		= xOther.m_xDirection;
		m_xNormal			= xOther.m_xNormal;
		m_fImpulseStrength	= xOther.m_fImpulseStrength;
		m_iBoneID			= xOther.m_iBoneID;
		m_iAttachmentID		= xOther.m_iAttachmentID;
		m_uMaterialResponseHashID = xOther.m_uMaterialResponseHashID;
		m_bBlocked			= xOther.m_bBlocked;
		m_bStealth			= xOther.m_bStealth; 
		m_bDoNotFlinch      = xOther.m_bDoNotFlinch;

		PARENT::operator=(xOther);
		return(*this);
	}

	// Overridden functions...
	bool	Apply() const;

	// Accessor functions...
	Asura_Guid		GetEntityToDamage(void) const							{ return(m_uEntityToDamage); }
	float			GetAmount(void) const									{ return(m_fAmount); }
	void			GetPointOfImpact(Asura_Vector_3& xPointOfImpact) const	{ xPointOfImpact = m_xPointOfImpact; }
	void			GetDirection(Asura_Vector_3& xDirection) const			{ xDirection = m_xDirection; }
	void			GetNormal(Asura_Vector_3& xNormal) const				{ xNormal = m_xNormal; }
	float			GetImpulseStrength(void) const							{ return(m_fImpulseStrength); }
	int				GetBoneID(void) const									{ return(m_iBoneID); }
	int				GetAttachmentID() const									{ return m_iAttachmentID; }
	Asura_Hash_ID	GetMaterialResponseHashID() const						{ return m_uMaterialResponseHashID; }
	bool			GetBlocked() const										{ return m_bBlocked; }
	bool			GetStealth() const										{ return m_bStealth; }
	bool			GetDoNotFlinch() const									{ return m_bDoNotFlinch; }

	void SetAmount( float fAmount )	{ m_fAmount = fAmount; }
	void SetDoNotFlinch( bool bDoNotFlinch ) {m_bDoNotFlinch = bDoNotFlinch;}

protected:
	// Protected variables...
	Asura_Guid		m_uEntityToDamage;		// The GUID of the "victim" entity.
	float			m_fAmount;				// The amount of damage to be applied.
	Asura_Vector_3	m_xPointOfImpact;		// The position at which damage is to be applied.
	Asura_Vector_3	m_xDirection;			// The direction damage is to be applied in.
	Asura_Vector_3	m_xNormal;				// The normal of the surface at which damage is to be applied in.
	float			m_fImpulseStrength;		// The impulse to be applied along with the damage.
	int				m_iBoneID;				// The bone-ID to apply damage to on the victim.
	int				m_iAttachmentID;		// The hit attachment ID (used to store the sub-container ID).
	Asura_Hash_ID	m_uMaterialResponseHashID;	// The material response from the LOS test.
	bool			m_bBlocked;				// Flag to easily pick different flinches.
	bool			m_bStealth;				// Flag to indicate damage caused by stealth attacks
	bool			m_bDoNotFlinch;         // Flag to indicate we do not want to do a flinch move

	// Protected functions...


private:
	// Private typedefs...
	typedef AVP_EntityDamage_Request PARENT;

	// Private variables...

	// Private functions...

};   // End of class 'AVP_EntityDamage_RequestVictim'.



////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_EntityDamage_RequestLinear
// Author       : Andy Tate
// Date         : 20-Feb-2008
//
// Description  : Class used to request linear-damage for entities.
////////////////////////////////////////////////////////////////////////////////

class AVP_EntityDamage_RequestLinear : public AVP_EntityDamage_RequestSpatial
{
public:
	// Public functions...
	AVP_EntityDamage_RequestLinear(void)	{}
	//virtual ~AVP_EntityDamage_RequestLinear();

	// JG - 23 April 2009 - Deprecated both constructors, leave the code here for reference.

	//AVP_EntityDamage_RequestLinear(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
	//							   const Asura_Hash_ID uFromMarkerPointHash,
	//							   const float fMaxAmount, const float fMaxImpulseStrength, const Asura_Guid uIgnoreGuid,
	//							   const float fDistanceOutFromMarkerPoint, const Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED);

	//AVP_EntityDamage_RequestLinear(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
	//							   const Asura_Hash_ID uFromMarkerPointHash,
	//							   const float fMaxAmount, const float fMaxImpulseStrength, const Asura_Guid uIgnoreGuid,
	//							   const Asura_Vector_3& xDirection, const float fDistance, const Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED);

	AVP_EntityDamage_RequestLinear(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
								   const Asura_Hash_ID uFromMarkerPointHash,
								   const float fMaxAmount, const float fMaxImpulseStrength, const Asura_Guid uIgnoreGuid,
								   const Asura_Vector_3& xImpactPosition, const Asura_Vector_3& xDirection, const float fDistance, 
								   const Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED, 
								   const Asura_Guid uImpactGuid = ASURA_GUID_UNREGISTERED)
	: PARENT(uTypeHash, uFromGuid, uFromMarkerPointHash, fMaxAmount, fMaxImpulseStrength, uIgnoreGuid, uParentGuid)
	, m_xImpactPosition(xImpactPosition)
	, m_xDirection(xDirection)
	, m_fDistance(fDistance)
	, m_uImpactGuid(uImpactGuid)
	{
	}

	AVP_EntityDamage_RequestLinear(const AVP_EntityDamage_RequestLinear& xOther)
	: PARENT(xOther)
	, m_xImpactPosition(xOther.m_xImpactPosition)
	, m_xDirection(xOther.m_xDirection)
	, m_fDistance(xOther.m_fDistance)
	, m_uImpactGuid(xOther.m_uImpactGuid)
	{
	}

	AVP_EntityDamage_RequestLinear& operator=(const AVP_EntityDamage_RequestLinear& xOther)
	{
		m_xImpactPosition	= xOther.m_xImpactPosition;
		m_xDirection		= xOther.m_xDirection;
		m_fDistance			= xOther.m_fDistance;
		m_uImpactGuid		= xOther.m_uImpactGuid; 

		PARENT::operator=(xOther);
		return(*this);
	}

	// Overridden functions...
	bool	Apply(AVP_ServerEntity_PhysicalObject& xFrom) const;

	// Accessor functions...
	void		GetImpactPosition(Asura_Vector_3& xPosition) const		{ xPosition = m_xImpactPosition; }
	void		GetImpactDirection(Asura_Vector_3& xDirection) const	{ xDirection = m_xDirection; }
	float		GetDistance(void) const									{ return m_fDistance; }
	Asura_Guid	GetImpactGuid(void) const								{ return m_uImpactGuid; }


protected:
	// Protected variables...
	Asura_Vector_3	m_xImpactPosition;		// The end position of the line used to request damage. In world space or model space
	Asura_Vector_3	m_xDirection;			// The direction for the LOS test
	float			m_fDistance;			// The distance for the LOS test
	Asura_Guid		m_uImpactGuid;			// The entity hit by this linear damage, if any			

	// Protected functions...


private:
	// Private typedefs...
	typedef AVP_EntityDamage_RequestSpatial PARENT;

	// Private variables...

	// Private functions...

};   // End of class 'AVP_EntityDamage_RequestLinear'.



////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_EntityDamage_RequestRadial
// Author       : Andy Tate
// Date         : 20-Feb-2008
//
// Description  : Class used to request radial-damage for entities.
////////////////////////////////////////////////////////////////////////////////

class AVP_EntityDamage_RequestRadial : public AVP_EntityDamage_RequestSpatial
{
public:
	// Public functions...
	AVP_EntityDamage_RequestRadial(void)	{}
	//virtual ~AVP_EntityDamage_RequestRadial();

	AVP_EntityDamage_RequestRadial(const Asura_Hash_ID uTypeHash, const Asura_Guid uFromGuid,
								   const Asura_Hash_ID uFromMarkerPointHash,
								   const float fMaxAmount, const float fMaxImpulseStrength, const Asura_Guid uIgnoreGuid,
								   const float fRadiusFor100PercentDamage, const float fRadiusFor0PercentDamage, const bool bPerformLOSTest, Asura_Guid uParentGuid = ASURA_GUID_UNREGISTERED)
	: PARENT(uTypeHash, uFromGuid, uFromMarkerPointHash, fMaxAmount, fMaxImpulseStrength, uIgnoreGuid, uParentGuid)
	, m_fRadiusFor100PercentDamage(fRadiusFor100PercentDamage)
	, m_fRadiusFor0PercentDamage(fRadiusFor0PercentDamage)
	, m_bPerformLOSTest(bPerformLOSTest)
	{
	}

	AVP_EntityDamage_RequestRadial(const AVP_EntityDamage_RequestRadial& xOther)
	: PARENT(xOther)
	, m_fRadiusFor100PercentDamage(xOther.m_fRadiusFor100PercentDamage)
	, m_fRadiusFor0PercentDamage(xOther.m_fRadiusFor0PercentDamage)
	, m_bPerformLOSTest(xOther.m_bPerformLOSTest)
	{
	}

	AVP_EntityDamage_RequestRadial& operator=(const AVP_EntityDamage_RequestRadial& xOther)
	{
		m_fRadiusFor100PercentDamage	= xOther.m_fRadiusFor100PercentDamage;
		m_fRadiusFor0PercentDamage		= xOther.m_fRadiusFor0PercentDamage;
		m_bPerformLOSTest				= xOther.m_bPerformLOSTest;

		PARENT::operator=(xOther);
		return(*this);
	}

	// Overridden functions...
	bool	Apply(AVP_ServerEntity_PhysicalObject& xFrom) const;

	// Accessor functions...
	float	GetRadiusFor100PercentDamage(void) const	{ return(m_fRadiusFor100PercentDamage); }
	float	GetRadius0PercentDamage(void) const			{ return(m_fRadiusFor0PercentDamage); }


protected:
	// Protected variables...
	float	m_fRadiusFor100PercentDamage;	// The radius within which 100% of the damage is inflicted on entities.
	float	m_fRadiusFor0PercentDamage;		// The radius at which entities cease to be damaged.
	bool	m_bPerformLOSTest;				// Flag to say whether we need to perform a LOS test before applying damage.

	// Protected functions...
	bool	ApplyToEntity(AVP_ServerEntity_PhysicalObject& xFrom, AVP_ServerEntity_PhysicalObject& xTo) const;
	bool	ApplyToBotanicals( const Asura_Vector_3& xPosition ) const;
	
#ifdef ASURA_USE_CONTAINER_SYSTEM
	bool ApplyToContainerEntity(AVP_ServerEntity_PhysicalObject& xFrom,
								const AVP_ServerEntity_Container& xTo,
								const class AVP_PhysicsQueryResult& xResult) const;
#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

private:
	// Private typedefs...
	typedef AVP_EntityDamage_RequestSpatial PARENT;

	// Private variables...

	// Private functions...

};   // End of class 'AVP_EntityDamage_RequestRadial'.



////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_EntityDamage_RequestVolume
// Author       : Andy Tate
// Date         : 20-Feb-2008
//
// Description  : Class used to request entity damage using an arbitrary volume.
////////////////////////////////////////////////////////////////////////////////

class AVP_EntityDamage_RequestVolume : public AVP_EntityDamage_RequestSpatial
{
public:
	// Public functions...
	AVP_EntityDamage_RequestVolume(void)	: m_pxVolume(NULL)	{}
	AVP_EntityDamage_RequestVolume(const AVP_EntityDamage_RequestVolume& xOther);
	~AVP_EntityDamage_RequestVolume();

	AVP_EntityDamage_RequestVolume&	operator=(const AVP_EntityDamage_RequestVolume& xOther);

	// Overridden functions...
	bool	Apply(AVP_ServerEntity_PhysicalObject& xFrom) const;

	// Accessor functions...
	Asura_Volume*	GetVolume(void) const	{ return(m_pxVolume); }


protected:
	// Protected variables...
	Asura_Volume*	m_pxVolume;		// A pointer to the volume to use when requesting damage.

	// Protected functions...


private:
	// Private typedefs...
	typedef AVP_EntityDamage_RequestSpatial PARENT;

	// Private variables...

	// Private functions...

};   // End of class 'AVP_EntityDamage_RequestVolume'.


#endif   // ifndef __AVP_ENTITYDAMAGE_REQUEST_H__


/// End of file. ///
