//////////////////////////////////////////////////////////////////////////////
//
// AVP_ClientEntity_Actor.h
//
// Basic actor class, at this level we'll assume the shape instance is still being used.
// A subclass can set up a posable anim instead if that's needed.
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef AVP_CLIENTENTITY_ACTOR_H___
#define AVP_CLIENTENTITY_ACTOR_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Asura
#include "Network/Asura_Network.h"
#include "Rendering/Asura_Project_Skin_Elements.h"

// AVP
#include "Animation/AVP_Hierarchy_System.h"
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_PhysicsObjectHandler.h"
#include "Core/AVP_Settings.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/PhysicalObject/AVP_ClientEntity_PhysicalObject.h"

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

class Asura_Animation_IK_Chain_LOS;
class Asura_Collision_Object;
class Asura_Message;

class AVP_Actor_AttachmentHandler;
class AVP_Actor_CameraHandler;
class AVP_Actor_DynamicStat;
class AVP_Actor_HUDHandler;
class AVP_Actor_InputHandler;
class AVP_Actor_InteractionHandler;
class AVP_Actor_PlayerAnimStateHandler;
class AVP_Actor_PickupHandler;
class AVP_ClientEntity_InteractionTrigger;
class AVP_ClientEntity_Pickup;
class AVP_Direction_Interpolator;
class AVP_Inventory_Group;
class AVP_Inventory_System;
class AVP_InventoryItem_Weapon;
class AVP_Message_Actor_ClientUpdate;
class AVP_Movement_System;
class AVP_Vertebrate_Focus;
class AVP_Collision_Descriptor;

// C L A S S E S /////////////////////////////////////////////////////////////

class AVP_ClientEntity_Actor : public AVP_ClientEntity_PhysicalObject
{
public:
	enum FootIndex
	{
		FOOT_REAR_LEFT,
		FOOT_REAR_RIGHT,
		FOOT_FRONT_LEFT,
		FOOT_FRONT_RIGHT,
		MAX_NUM_FEET
	};

	enum 
	{
		AVP_WEAPON_SELECT_OPEN,
		AVP_WEAPON_SELECT_CHANGE,
		AVP_WEAPON_SELECT_CLOSE,
	};

	enum AimDirectionUsage
	{
		// if you change this enum make sure that it still fits in the union member m_eAimDirectionUsage and the same one in AVP_Move
		AIM_DIRECTION_NOT_USED,
		AIM_DIRECTION_RESET,		// lerps back to the orientation
		AIM_DIRECTION_SET_DIRECTLY, // set directly from the move
		AIM_DIRECTION_LOCK,			// locks to current relative position to the orientation
		AIM_DIRECTION_LIMIT,		// clamps to a range from the orientation forward vector

		NUM_AIM_DIRECTION_USAGES
	};

	enum LookAtPositionUsage
	{
		LOOK_AT_POSITION_NOT_USED,
		LOOK_AT_POSITION_RESET,
		LOOK_AT_POSITION_SET_DIRECTLY,
		LOOK_AT_POSITION_LOCK,
		LOOK_AT_POSITION_LIMIT,

		NUM_LOOK_AT_POSITION_USAGES
	};

	enum 
	{
		ACTOR_RAGDOLL_PRIORITY			  = 10,
		START_DEAD_ACTOR_RAGDOLL_PRIORITY = 1000  // 1000 for dead bodies, 10 for regular ragdolls
	};

	// Public functions...
	virtual	void	Update					(void);
	virtual	bool	HandleMessage			(const Asura_Message* pxMessage);
	virtual bool	CanSendMessageToServer	(void) const	{ return( (m_pxInputHandler != NULL) || ( Asura_Network::IsServer() && IsAIActor() ) ); } 
	virtual void	PostCreateInitialise(void);
	virtual void	PreDestructionShutdown();

	virtual	void	UpdateInput		(void);
	virtual	void	RenderHUD		(void) const;
	virtual	void	SetCamera		(void);

	virtual void	ReadFromClientCreateStream( Asura_Chunk_Memory_Stream& xStream, float fServerTime );

	void			BeginRenderHUD	(void) const	{ Platform_BeginRenderHUD(); }
	void			EndRenderHUD	(void) const	{ Platform_EndRenderHUD(); }

	// Virtual player manipulation functions...
	virtual	void	MakePlayer	(Asura_Guid uPlayerGuid);
	virtual	void	MakeAI		(void);
	virtual void	DisconnectPlayer(); // Similar to MakeAI, but designed just to chuck the player info out without going on to create some AI handling stuff.
	Asura_Guid		GetPlayerInfoGuid() const { return m_uPlayerInfoGuid; }
	
	AVP_Settings::Species GetSpeciesByClassification() const 	{ return AVP_Settings::GetSpeciesByClassification( static_cast<AVP_ENTITY_CLASSIFICATION>(GetClassification())); }

	virtual	void	OnBeginMoving		(void)	{ m_bMoving = true; }
	virtual	void	OnEndMoving			(void)	{ m_bMoving = false; }
	virtual	void	OnBeginMovingFast	(void)	{ m_bMovingFast = true; }
	virtual void	OnEndMovingFast		(void)	{ m_bMovingFast = false; }
	virtual	void	OnBeginMovingSlowly	(void)	{ m_bMovingSlow = true; }
	virtual void	OnEndMovingSlowly	(void)	{ m_bMovingSlow = false; }

	void			SetMoving( bool bSet );
	void			SetMovingSlow( bool bSet );
	void			SetMovingFast( bool bSet );

	float GetAnimDurationModifier() const { return m_fAnimDurationModifier; }

	bool			IsMoving()			const { return m_bMoving; }
	bool			IsMovingSlow()		const { return m_bMovingSlow; }
	bool			IsMovingMedium()	const { return IsMoving() && !m_bMovingSlow && !m_bMovingFast; }
	bool			IsMovingFast()		const { return m_bMovingFast; }
	
	virtual bool IsCrouching() const { return false; }
	virtual bool IsJumping() const { return false; }
	virtual bool IsFalling() const { return false; }
	virtual bool IsLanding() const { return false; }

	virtual AVP_Vertebrate_Focus*			GetFocus()					{ return NULL; }
	virtual const AVP_Vertebrate_Focus*		GetFocus() const			{ return NULL; }

	// true if we are currently the melee target of the player
	bool IsPlayerMeleeTarget() const;
	virtual AVP_Actor_DynamicStat*			GetDynamicEnergy()			{ return NULL; }
	virtual const AVP_Actor_DynamicStat*	GetDynamicEnergy() const	{ return NULL; }

	// Static list functions...
	static AVP_ClientEntity_Actor* FindEntity( Asura_Guid uGUID )				{ return s_xList.FindEntity( uGUID ); }
	static const Asura_Hashed_List<AVP_ClientEntity_Actor>* GetList( void )		{ return &s_xList; }

	// overrides for movement system stuff
	virtual void TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation);
	
	// Accessor functions...
	AVP_Actor_PhysicsObjectHandler::GroundType	GetGroundType(void) const	{ return(static_cast<AVP_Actor_PhysicsObjectHandler::GroundType>(m_eGroundType)); }
	AVP_Actor_PhysicsObjectHandler* GetPhysicsObjectHandler	(void)	const	{ return(m_pxPhysicsObjectHandler); }

	Asura_Guid	GetGuidOfActorBeingAttackedByUs(void) const 		{ return m_uGuidOfActorBeingAttackedByUs; }
	void		SetGuidOfActorBeingAttackedByUs(Asura_Guid uGuid)	{ m_uGuidOfActorBeingAttackedByUs = uGuid; }

	AVP_Actor_AnimationHandler*			GetAnimationHandler	(void)			{ return(m_pxAnimationHandler); }
	const AVP_Actor_AnimationHandler*	GetAnimationHandler	(void) const	{ return(m_pxAnimationHandler); }

	AVP_Actor_CameraHandler*			GetCameraHandler	(void)			{ return(m_pxCameraHandler); }
	const AVP_Actor_CameraHandler*		GetCameraHandler	(void) const	{ return(m_pxCameraHandler); }

	AVP_Actor_HUDHandler*				GetHUDHandler		(void)			{ return(m_pxHUDHandler); }
	const AVP_Actor_HUDHandler*			GetHUDHandler		(void) const	{ return(m_pxHUDHandler); }

	AVP_Actor_InputHandler*				GetInputHandler		(void)			{ return(m_pxInputHandler); }
	const AVP_Actor_InputHandler*		GetInputHandler		(void) const	{ return(m_pxInputHandler); }

	AVP_Actor_InteractionHandler*		GetInteractionHandler(void)			{ return(m_pxInteractionHandler); }
	const AVP_Actor_InteractionHandler*	GetInteractionHandler(void) const	{ return(m_pxInteractionHandler); }

	AVP_Actor_PlayerAnimStateHandler*		GetPlayerStateHandler(void)			{ return(m_pxPlayerStateHandler); }
	const AVP_Actor_PlayerAnimStateHandler*	GetPlayerStateHandler(void) const	{ return(m_pxPlayerStateHandler); }

	AVP_Actor_PickupHandler*		    GetPickupHandler(void)	        { return(m_pxPickupHandler); }
	const AVP_Actor_PickupHandler*	    GetPickupHandler(void) const	{ return(m_pxPickupHandler); }

	AVP_Inventory_System*				GetInventorySystem(void)		{ return(m_pxInventorySystem); }
	const AVP_Inventory_System*			GetInventorySystem(void) const	{ return(m_pxInventorySystem); }
	
	AVP_Actor_AttachmentHandler*		GetAttachmentHandler(void)		 { return(m_pxAttachmentHandler); }
	const AVP_Actor_AttachmentHandler*  GetAttachmentHandler(void) const { return(m_pxAttachmentHandler); }
	void								ClearAttachmentHandler(void)	 { m_pxAttachmentHandler = NULL; }

	Asura_Hash_ID			GetFloorTypeHash( u_int uFootIndex ) const	{ Asura_Assert( uFootIndex < MAX_NUM_FEET, "Bad foot index" ); return m_puFloorType[ uFootIndex ]; }
	const Asura_Vector_3&	GetFloorPosition( u_int uFootIndex ) const	{ Asura_Assert( uFootIndex < MAX_NUM_FEET, "Bad foot index" ); return m_pxFloorPos[ uFootIndex ]; }
	const Asura_Vector_3&	GetFloorNormal  ( u_int uFootIndex ) const	{ Asura_Assert( uFootIndex < MAX_NUM_FEET, "Bad foot index" ); return m_pxFloorNormal[ uFootIndex ]; }

#ifdef ASURA_USE_MOVEMENT_SYSTEM
	virtual bool			CanCurrentlyUseMoves( void ) const { return(true && (m_pxMovementSystem!=NULL) ); }
#endif

	float				GetSpeedFromAnimation();
	void				GetVelocityFromAnimation( Asura_Vector_3& xAnimVelocity );
	void				GetAngularVelocityFromAnimation( Asura_Vector_3& xAxis, Asura_Angle& fAngularSpeed );

	bool				IsCloaked(const bool bFullyCloaked) const;
	bool				IsPlayer					(void) const	{ return (m_bIsPlayer); }
	bool				IsAIActor					(void) const	{ return (m_bIsAIActor); }
	void				SetGrappleGuid				(Asura_Guid uGuid)	{ m_uGrappleGuid = uGuid; }
	Asura_Guid			GetGrappleGuid				(void) const	{ return(m_uGrappleGuid); }
	void				SetHasWonGrapple			(bool bWon) 	{ m_bHasWonGrapple = bWon; } 
	bool				HasWonGrapple				(void) const	{ return(m_bHasWonGrapple); } 
	void				SetHasLostGrapple			(bool bLost) 	{ m_bHasLostGrapple = bLost; } 
	bool				HasLostGrapple				(void) const	{ return(m_bHasLostGrapple); } 
	void				SetGrappleAnimHash			(Asura_Hash_ID uGuid)	{ m_uGrappleAnimationHash = uGuid; }
	Asura_Hash_ID		GetGrappleAnimHash			(void) const			{ return(m_uGrappleAnimationHash); }

	// weapon stuff
	virtual bool	ChangeToNewWeapon(const bool bInstant, Asura_Hash_ID uExtendAnimHash, const bool bCalculateMovement);
	virtual bool	BeginWeaponActivation(void);
	virtual bool	EndWeaponActivation(void);
	virtual bool	BeginWeaponDeactivation( Asura_Hash_ID uDeactivationAnim, const bool bCalculateMovement );
	virtual bool	EndWeaponDeactivation(void);

	virtual void	AddPermenantlyActiveItems(AVP_Inventory_Group* pxGroup) {};

	void	SetCollisionDescriptor( const Asura_Hash_ID uCollDescHash );

	float	GetInnerTargetingRadius() const { return m_fInnerTargetingRadius; }
	float	GetOuterTargetingRadius() const { return m_fOuterTargetingRadius; }
	
	virtual void GetAimDirection( Asura_AVector3& xAimDir ) const { xAimDir = m_xAimDirection; }
	virtual void SetAimDirection( const Asura_AVector3& xAimDir, const float fServerTime = -1.0f ) { m_xAimDirection = xAimDir; }

	AimDirectionUsage GetAimDirectionUsage() const;
	const Asura_AVector3 GetTargetAimDirection() const;
	float GetAimDirectionLimitAngle() const;
	float GetAimDirectionAngVel() const;

	void SetAimDirectionUsage( AimDirectionUsage eSet ){ m_eAimDirectionUsage = eSet; }
	void SetTargetAimDirection( const Asura_AVector3& xDirection ){ m_xTargetAimDirection = xDirection; }
	void SetAimDirectionLimitAngle( float fAngle ){ m_fAimDirectionLimitAngle = fAngle; }
	void SetAimDirectionAngVel( float fTimeScale ){ m_fAimDirectionAngVel = fTimeScale; }

	virtual void GetLookAtPosition( Asura_AVector3& xLookAtPos ) const { xLookAtPos = m_xLookAtPosition; }
	virtual void SetLookAtPosition( const Asura_AVector3& xLookAtPos, const float fServerTime = -1.0f ) { Asura_Assert(xLookAtPos == xLookAtPos, "bah"); m_xLookAtPosition = xLookAtPos; }

	LookAtPositionUsage GetLookAtPositionUsage() const;
	const Asura_AVector3& GetTargetLookAtPosition() const;
	float GetLookDirectionLimitAngle() const;
	float GetLookDirectionLerpTimeScale() const;
	float GetLookDirectionStoredAngle() const { return m_fLookDirectionStoredAngle; }

	void SetLookAtPositionUsage( LookAtPositionUsage eSet ){ m_eLookAtPositionUsage = eSet; }
	void SetTargetLookAtPosition( const Asura_AVector3& xPos ){ m_xTargetLookAtPosition = xPos; }
	void SetLookDirectionLimitAngle( float fAngle ){ m_fLookDirectionLimitAngle = fAngle; }
	void SetLookDirectionLerpTimeScale( float fTimeScale ){ m_fLookDirectionLerpTimeScale = fTimeScale; }
	void SetLookDirectionStoredAngle( float fAngle ){ m_fLookDirectionStoredAngle = fAngle; }

	void GetMovementDirection( Asura_AVector3& xMoveDir ) const;
	void SetMovementDirection( const Asura_AVector3& xMoveDir, const float fServerTime = -1.0f );

	float	GetWeaponEffectiveRange() const { return 20.0f; } // TODO

	static bool TestRayAgainstTargetingSpheres( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, const Asura_Guid uGuidToIgnore = ASURA_GUID_UNREGISTERED, Asura_Guid* const puFirstHitGuid = NULL, float* const pfFirstHitDistance = NULL, float* const pfHitRadiusParametric = NULL );

	bool	GetGuidAndMarkerHashForTarget(Asura_Vector_3& xCameraLocation, Asura_Matrix_3x3& xCameraOrientation, Asura_Guid& uHitEntity, Asura_Hash_ID& uHitMarker);

	void	GetHeadPosition(Asura_Vector_3& xPosition) const;
	void	GetChestPosition(Asura_Vector_3& xPosition) const;

	bool	GetPositionForDialogue(Asura_Vector_3& xPosition) const;

	bool	IsUsingFirstPersonSkin() const;

	u_int	GetFlagsForAllAnimsLastFrame() const;
	
	bool	IsEnemyTo( const Asura_Guid uGuid ) const;

	virtual bool IsInvulnerable() const;

	virtual void GetCurrentMeleeBlueprintIDs(Asura_Hash_ID& uFirstPersonID, Asura_Hash_ID& uThirdPersonID) const;

	u_int GetNumFeet() const { return m_uNumFeet; }
	virtual void	GetFootPosition( u_int uFootItr, Asura_Vector_3& xFootPos ) { xFootPos = GetCurrentPosition(); }

	u_int GetMaxHealthBlocks() const { return m_uMaxHealthBlocks; }

	float GetEnergyPerBlock() const { return m_fEnergyPerBlock; }
	float GetMinEnergy() const { return m_fMinEnergy; }
	float GetMaxEnergy() const { return m_fMaxEnergy; }

	bool  IsIdleForAutoCenter() const;
	virtual bool  IsIdleForFidget() const;
	const AVP_InventoryItem_Weapon* GetCurrentWeapon() const;

	u_short GetFlagsToIgnore() const { return m_usFlagsToIgnore; }

	u_int	CreatePFXEffect(const Asura_Hash_ID uPFXHash, const Asura_Hash_ID uMarkerPointHash, const bool bAttachToSkin);

	Asura_Hash_ID GetSurfaceType( int iBoneID, bool bFatalDamage );
	virtual u_int GetElementFromBoneID( int iBoneID ) const { return NUM_AVP_SKIN_ELEMENT_TYPES; }
	virtual const Asura_Blueprint* GetSkinElementBlueprint(u_int uElement) const { return NULL; }

	virtual bool GetCanPlayDialogue() const;

	Asura_Hash_ID GetCurrentGamesceneAnimControllerHash() const;

	void FlagMovementSystemForDestruction() { m_bDestroyMovementSystem = true; }

	const Asura_Animation_IK_Chain_LOS* GetLimbIKChainLOS( u_int uLimbItr ) const { return m_pxAnimationHandler != NULL ? m_pxAnimationHandler->GetLimbIKChainLOS( uLimbItr ) : NULL; }

	const AVP_Actor_AnimStateData* GetCurrentAnimStateData() const;
	virtual const AVP_Actor_AnimStateData* GetAnimStateDataForWeapon( const Asura_Hash_ID uWeaponBPHash ) const { return NULL; }

	void UpdateAnimStateData( const AVP_Actor_AnimStateData* pxAnimDataLBSlow, const AVP_Actor_AnimStateData* pxAnimDataLBFast, const AVP_Actor_AnimStateData* pxAnimDataUBSlow, const AVP_Actor_AnimStateData* pxAnimDataUBFast, const bool bLoop, const float fBlendLBSpeed, const float fBlendUBSpeed, const float fOSMoveDirAngle, const float fTweenTime, const bool bSynch );
	void PauseAnimState();
	void PlayAnimState();

	void OnPickup(AVP_ClientEntity_Pickup* pxPickup, float fQuantity, float fSecondaryQuantity, bool bSetAsCurrent = true);
	void RestorePreviousWeapon() const;

	u_int GetAIAnimState() const { return m_uAIAnimState; }
	void SetAIAnimState( const u_int uState ) { m_uAIAnimState = uState; }

	Asura_Hash_ID GetAnimStateBlueprintType() const { return m_uAnimStateBlueprintType; }
	virtual bool		  IsADeadBody() const { return  m_bDeadBody; }

	// These flags last for one weapon change and then get reset to true
	void				  SetAllowWeaponActivationAnim(bool bAllow)   { m_bAllowWeaponActivationAnim = bAllow;   }
	bool				  GetAllowWeaponActivationAnim() const        { return m_bAllowWeaponActivationAnim;     }

	bool IsRagdoll() const;

	void ActivateCurrentWeapon();
	void DeactivateCurrentWeapon();

	void SetEntityCurrentlyMoving(bool bMoving)						{ PARENT::SetEntityCurrentlyMoving(bMoving); }
	
	bool IsReorienting() const { return m_pxPhysicsObjectHandler ? m_pxPhysicsObjectHandler->IsReorienting() : false; }

	AVP_Settings::Team GetTeam(void) const;

	void SetTurnAnimAngularVelocityMultiplier( const float fScale ) { m_fTurnAnimAngVelScale = fScale; }

	Asura_Vector_3 GetDownVector() const { const Asura_Matrix_3x3& xOri = GetCurrentOrientation(); return Asura_Vector_3( xOri.e12, xOri.e22, xOri.e32 ); }
	Asura_Vector_3 GetForwardVector() const { const Asura_Matrix_3x3& xOri = GetCurrentOrientation(); return Asura_Vector_3( xOri.e13, xOri.e23, xOri.e33 ); }

	void UpdateLaserData( bool bUsingLaser, Asura_Vector_3 xFirePosition = Asura_Maths::ZeroVector3, Asura_Vector_3 xFireDirection = Asura_Maths::ZeroVector3, float fRange = 0.0f );
	
	Asura_Vector_3 GetLaserPosition() const			{ return m_xLaserPosition; }
	Asura_Vector_3 GetLaserDirection() const		{ return m_xLaserDirection; }
	float GetLaserRange() const						{ return m_fLaserRange; }
	bool IsUsingLaser()const						{ return m_bUsingLaser; }

	float GetMinDistanceForJumpLandLightEffect() const { return m_fMinDistanceForJumpLandLightEffect; }
	float GetMinDistanceForJumpLandNormalEffect() const { return m_fMinDistanceForJumpLandNormalEffect; }
	float GetMinDistanceForJumpLandHeavyEffect() const { return m_fMinDistanceForJumpLandHeavyEffect; }

	bool  IntersectsSphere(const Asura_Vector_3& xPos, float fRadius) const;
	bool  CoarseIntersectsSphere(const Asura_Vector_3& xPos, float fRadius) const;
	bool  IntersectsBoundingBox(const Asura_Bounding_Box& xBB) const;
	bool  CoarseIntersectsBoundingBox(const Asura_Bounding_Box& xBB) const;

#if defined(ASURA_SUPPORT_COMMAND_CONSOLE)
	static void CreateConsoleVars( void );
#endif

protected:
	// Protected variables...
	AVP_Actor_AnimationHandler*		m_pxAnimationHandler;		// The animation-handler to use for this actor.
	AVP_Actor_CameraHandler*		m_pxCameraHandler;			// The camera-handler to use for this actor.
	AVP_Actor_HUDHandler*			m_pxHUDHandler;				// The HUDhandler to use for this actor.
	AVP_Actor_InputHandler*			m_pxInputHandler;			// The Input-handler to use for this actor.
	AVP_Actor_InteractionHandler*	m_pxInteractionHandler;		// The Interaction-handler to use for this actor.
	AVP_Inventory_System*			m_pxInventorySystem;		// The inventory-system to use for this actor.
	AVP_Actor_PhysicsObjectHandler*	m_pxPhysicsObjectHandler;	// The class to use for handling the physics-object for this actor.
	AVP_Actor_PlayerAnimStateHandler*	m_pxPlayerStateHandler;		// Pointer to the player statehandler
	AVP_Actor_PickupHandler*	    m_pxPickupHandler;          // Pickup Handler
	AVP_Actor_AttachmentHandler*    m_pxAttachmentHandler;		// The attachment handler to use for this actor.

	AVP_Direction_Interpolator*		m_pxMoveDirInterpolator;

	Asura_Vector_3					m_xAimDirection;
	Asura_Vector_3					m_xLookAtPosition;
	Asura_Vector_3					m_xMovementDirection;

	Asura_Guid						m_uGuidOfActorBeingAttackedByUs;
	
    // The current anim, requested by the AI.
    u_int                           m_uCurrentAnim;
    u_int                           m_uLoopData;
    bool                            m_bShouldTween;
	
	u_int							m_uMaxHealthBlocks;

	Asura_Hash_ID					m_uAnimStateBlueprintType;

	float							m_fEnergyPerBlock;
	float							m_fMinEnergy;
	float							m_fMaxEnergy;

	Asura_Vector_3					m_xLaserPosition;
	Asura_Vector_3					m_xLaserDirection;
	float							m_fLaserRange;
	bool							m_bUsingLaser;

#ifdef ASURA_USE_AFFECTORS
	Asura_Handle					m_iHackedHeadAffector;				// ARTTODO: Remove once attachments are re-written.
	Asura_Handle					m_iHackedCompassAffector;			// ARTTODO: Remove once attachments are re-written.
#endif

	// Protected functions...
	AVP_ClientEntity_Actor(const u_short usFlagsToIgnore);
	virtual ~AVP_ClientEntity_Actor();

	void			SetNumFeet( u_int uNumFeet ) { Asura_Assertf( uNumFeet <= MAX_NUM_FEET, ( "%d <= %d", uNumFeet, MAX_NUM_FEET ) ); if ( uNumFeet <= MAX_NUM_FEET ) { m_uNumFeet = uNumFeet; } }

	const AVP_Collision_Descriptor* GetDefaultCollisionDescriptor();
	bool CanChangeCollisionObject( const Asura_Collision_Box& xCollisionObject );
	
	// Overridden event functions...
	virtual void	HealthChanged(float fHealth);

    virtual void	InitialiseFromBlueprint( const Asura_Blueprint& xBlueprint );

	virtual void	OnEntityActivate();
	virtual void	OnEntityDeactivate();

	virtual void	OnApplyHealth(const float fHealthIncrement);
	virtual void	OnApplyDamage(const AVP_EntityDamage_ApplyToEntity& xDamage);
	virtual	void	OnDeath(const AVP_EntityDamage_ApplyToEntity& xDamage);

	// Virtual initialisation functions...
	virtual bool	InitialiseInventory(void);
	virtual	bool	InitialiseMoves(void);
	virtual bool	InitialiseSoundEvents(void);

	virtual void	CreateAnimationHandler() = 0;
	// Overridden functions...
	virtual	void	Render	(void);
	virtual void    OnDecapitation() {}
	// Action functions...
	virtual void UpdateMovement() {}
	virtual void UpdateActions() {}
	virtual void UpdateAimDirection();
	void	UpdateLookAtPosition();
	void	UpdateFloorFlags(void);

    void	ProcessClientUpdate(const AVP_Message_Actor_ClientUpdate* pxData, float fServerTime, bool bProcessPosAndOri = true);

	void	Platform_BeginRenderHUD(void) const;
	void	Platform_EndRenderHUD(void) const;

	void	RenderPlayerPosition() const;

	Asura_Hash_ID	m_puFloorType[ MAX_NUM_FEET ];
	Asura_Vector_3	m_pxFloorPos[ MAX_NUM_FEET ];
	Asura_Vector_3	m_pxFloorNormal[ MAX_NUM_FEET ];

	float m_fDeathTimer;

	Asura_AVector3 m_xTargetAimDirection;
	Asura_AVector3 m_xTargetLookAtPosition;

	float	m_fAimDirectionLimitAngle;
	float	m_fAimDirectionAngVel;

	float	m_fLookDirectionLimitAngle;
	float	m_fLookDirectionStoredAngle;
	float	m_fLookDirectionLerpTimeScale;

	float	m_fAnimDurationModifier;
	float	m_fAnimDurationModifierMoving;

	float	m_fTurnAnimAngVelScale;

	float 	m_fMinDistanceForJumpLandLightEffect;
	float 	m_fMinDistanceForJumpLandNormalEffect;
	float 	m_fMinDistanceForJumpLandHeavyEffect;

	union
	{
		u_int		m_uActorFlags;				// Single access point for flags.

		struct
		{
			u_int	m_eGroundType			: 4,	// Flag to say whether we're currently on the ground.
					m_bIsPlayer				: 1,
					m_bIsAIActor			: 1,
					m_bHasWonGrapple		: 1,
					m_bHasLostGrapple		: 1,
					m_bMoving				: 1,	// these are for the AI
					m_bMovingSlow			: 1,
					m_bMovingFast			: 1,
					m_bIsDestructible		: 1,
					m_bDestroyMovementSystem: 1,
					m_bPlayingNextAnim		: 1,
					m_eAimDirectionUsage	: 3,	// if enum AimDirectionUsage changed make sure to update this 
					m_eLookAtPositionUsage	: 3,
					m_bDeadBody             : 1,
											: 0;	// 32 - 21 == 11 spare-bits.
		};
	};

	u_short	m_usFlagsToIgnore;	// Flags for the physics-objects to ignore collision.

	Asura_Guid		m_uGrappleGuid;
	Asura_Hash_ID	m_uGrappleAnimationHash;	// Stores the hash of the grapple anim so it can be updated

	Asura_Guid		m_uPlayerInfoGuid;       
	Asura_Hash_ID	m_uDeadAnimHashID;
	bool			m_bAllowWeaponActivationAnim;

	Asura_RGBA		m_xAlienPheromoneColour;
	AVP_Animation_ProjectData m_xAnimProjectData;

private:
	void DestroyMovementSystem(); // Called from our update if the destruction flag is set.

	virtual const AVP_Actor_AnimStateData* GetStateDataFromMesssageData( const u_int uState, const AVP_Message_Actor_ClientUpdate* pxData ) { return NULL; }
	virtual const AVP_Actor_AnimStateData* GetStateData( const u_int uState, const u_int uGroundType, const u_int uLimbDamage, const bool bIsAiming, const u_int uCoverType ) { return NULL; }

	// Private typedefs...
	typedef AVP_ClientEntity_PhysicalObject PARENT;

	// Private variables...	
	u_int				m_uAIAnimState;
	u_int				m_uNumFeet;
	float				m_fCorpseRemoveTime;
	float				m_fCorpseFadeTime;

	// Radius to use for targeting assist functions and the like, initialized from the BPs.
	float				m_fInnerTargetingRadius;
	float				m_fOuterTargetingRadius;

	static Asura_Hashed_List<AVP_ClientEntity_Actor> s_xList;

	// Private functions...
};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef AVP_CLIENTENTITY_ACTOR_H___
