//////////////////////////////////////////////////////////////////////////////
//
// AVP_ClientVehicle_Dropship.cpp
//
// Copyright (c) 2007 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/PhysicalObject/Actor/Vehicles/Dropship/AVP_ClientVehicle_Dropship.h"

// Engine independent.
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationAttachment.h"
#include "Core/Asura_Timers.h"
#include "Entity/Asura_PosAndOriInterpolator.h"
#include "Entity/Ragdoll/Asura_ClientEntity_Ragdoll.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Message.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_Player.h"
#include "Physics/Asura_Physics.h"
#include "Rendering/Asura_Render.h"
#include "Shapes/Asura_Shapes.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Entity_Effect.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "Core/AVP_Game_Core.h"

// Project independent.
#include "Actor/AVP_Actor_CameraHandler.h"
#include "Actor/AVP_Actor_HUDHandler.h"
#include "Actor/AVP_Actor_InputHandler.h"
#include "Animation/AVP_AnimFlags.h"
#include "Core/AVP_Settings.h"
#include "Entity/Triggers/AVP_ClientEntity_InteractionTrigger.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/Vehicles/Dropship/AVP_Messages_Dropship.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Moves/AVP_Movement_System.h"
#include "Moves/AVP_Moves.h"
#include "SpecialFX/AVP_Laser.h"


// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

Asura_Hashed_List<AVP_ClientVehicle_Dropship> AVP_ClientVehicle_Dropship::s_xList;

AVP_ClientVehicle_Dropship::AVP_ClientVehicle_Dropship()
: PARENT(0)
, m_pxLaserAttachment(NULL)
, m_fVolumeFraction(0.0)
, m_fPitchFraction(0.0)
, m_uEnginesSoundHash(ASURA_HASH_ID_UNSET)
, m_uExtraEnginesSoundHash(ASURA_HASH_ID_UNSET)
, m_iEnginesSoundHandle(ASURA_INVALID_HANDLE)
, m_iExtraEnginesSoundHandle(ASURA_INVALID_HANDLE)
, m_fMinEngineVolume(0.0)
, m_fMaxEngineVolume(1.0)
, m_fMinEnginePitch(0.0)
, m_fMaxEnginePitch(1.0)
{
	Classification = AVP_ENTITYCLASS_DROPSHIP;
	// All will be initialised later...
	s_xList.AddToList( this );
}


AVP_ClientVehicle_Dropship::~AVP_ClientVehicle_Dropship( void ) 
{
	if (m_pxLaserAttachment)
	{
		if (m_pxLaserAttachment->IsLinked())
		{
			m_pxLaserAttachment->UnlinkAttachment(true);
		}
		else
		{
			delete m_pxLaserAttachment;
			m_pxLaserAttachment = NULL;
		}
	}

	s_xList.RemoveFromList(this);
}

DEBUGVAR u_int dbg_uLaserAttachementHashID = 0x061fbecb; //"Laser"

void AVP_ClientVehicle_Dropship::PostCreateInitialise()
{
	// If there is no animation handler yet, create it.
	if( !m_pxAnimationHandler &&  GetBlueprintHashID() != ASURA_HASH_ID_UNSET )
	{
		CreateAnimationHandler();

		Asura_Blueprint* pxBluprint = GetBlueprint();
		Asura_Assert(pxBluprint, "blueprint not found");

		GetDropshipAnimatiomHandler()->PostCreateInitialise(pxBluprint);
	}

	if(!m_pxLaserAttachment)
	{
		m_pxLaserAttachment = new AVP_AnimationAttachment_Laser( new AVP_Laser(), true );
		m_pxLaserAttachment->GetLaser()->SetBeamRadius(0.2f);		
		m_pxLaserAttachment->GetLaser()->SetMaximumLength(1000.f);	
	}

	PARENT::PostCreateInitialise();
}

void AVP_ClientVehicle_Dropship::PreDestructionShutdown()
{
	if(m_iEnginesSoundHandle != ASURA_INVALID_HANDLE && Asura_Sound_Event_System::IsPlaying(m_iEnginesSoundHandle))
	{
		Asura_Sound_Event_System::Stop(m_iEnginesSoundHandle);
	}
	if(m_iExtraEnginesSoundHandle != ASURA_INVALID_HANDLE && Asura_Sound_Event_System::IsPlaying(m_iExtraEnginesSoundHandle))
	{
		Asura_Sound_Event_System::Stop(m_iExtraEnginesSoundHandle);
	}

}


bool AVP_ClientVehicle_Dropship::HandleMessage( const Asura_Message* pxMessage )
{
	switch ( pxMessage->GetMsgID() )
    {
		case ASURA_MSG_CLIENTCREATE:
        case ASURA_MSG_CLIENTUPDATE:
        {
            const AVP_Message_DropshipClientUpdate* pxData = static_cast< const AVP_Message_DropshipClientUpdate* >( pxMessage->GetData() );

            // If we are inactive and now receiving a client update this means we are becoming active.
            // We should go visible and teleport to this first pos/ori so we don't zip across the level.
            if(!IsEntityActive())
			{
                Asura_Matrix_3x3 xOrientMatrix;
                pxData->m_xOrientation.ToMatrix( xOrientMatrix );
                TeleportTo( pxData->m_xPosition, xOrientMatrix );
            }

            ProcessClientUpdate( pxData, pxMessage->GetSentTime() );
            return true;
        }

        case ASURA_MSG_AXON_GS2_ANIM_SEQUENCE_DATA:
        {
            const Axon_Gamescene_AnimSequenceData* pxData = static_cast< const Axon_Gamescene_AnimSequenceData* >( pxMessage->GetData() );
            if ( m_pxAnimationHandler ) { m_pxAnimationHandler->HandleGamesceneAnimSequenceData( *pxData ); }
            return true;
        }

        case AVP_MSGID_REQUEST_ANIMATION:
        {
            const AVP_Message_Dropship_RequestAnimation* pxData = static_cast< const AVP_Message_Dropship_RequestAnimation* >( pxMessage->GetData() );
            if ( GetDropshipAnimatiomHandler() ) { GetDropshipAnimatiomHandler()->HandleDropshipAnimationRequest( *pxData ); }
            return true;
        }

        default:
        {
            return PARENT::HandleMessage( pxMessage );
        }
    }
}

void AVP_ClientVehicle_Dropship::Update()
{
	PARENT::Update();

	const Asura_Sound_Event* pxEngineLoop = Asura_Sound_Event_Storage::FindEvent(m_uEnginesSoundHash);
	if(pxEngineLoop)
	{
		if (m_iEnginesSoundHandle != ASURA_INVALID_HANDLE && !Asura_Sound_Event_System::IsPlaying(m_iEnginesSoundHandle))
		{
			Asura_Sound_Event_System::Stop( m_iEnginesSoundHandle );
		}
		if (m_iEnginesSoundHandle == ASURA_INVALID_HANDLE)
		{
			const Asura_Vector_3& xPosition = GetCurrentPosition();
			Asura_Sound_Event_System::PlayEvent(pxEngineLoop, &m_iEnginesSoundHandle, &xPosition, 0.0f); // start silent
		}
		else
		{
			Asura_Sound_Event_System::UpdatePosition(m_iEnginesSoundHandle, GetCurrentPosition());
			Asura_Sound_Event_System::SetPitchScale(m_iEnginesSoundHandle, m_fPitchFraction);
			Asura_Sound_Event_System::SetVolumeScale(m_iEnginesSoundHandle, m_fVolumeFraction);
		}
	}
	const Asura_Sound_Event* pxExtraEngineLoop = Asura_Sound_Event_Storage::FindEvent(m_uExtraEnginesSoundHash);
	if(pxExtraEngineLoop)
	{
		if (m_iExtraEnginesSoundHandle != ASURA_INVALID_HANDLE && !Asura_Sound_Event_System::IsPlaying(m_iExtraEnginesSoundHandle))
		{
			Asura_Sound_Event_System::Stop( m_iExtraEnginesSoundHandle );
		}
		if (m_iExtraEnginesSoundHandle == ASURA_INVALID_HANDLE)
		{
			const Asura_Vector_3& xPosition = GetCurrentPosition();
			Asura_Sound_Event_System::PlayEvent(pxExtraEngineLoop, &m_iExtraEnginesSoundHandle, &xPosition, 1.0f); 
		}		
		else
		{
			Asura_Sound_Event_System::UpdatePosition(m_iExtraEnginesSoundHandle, GetCurrentPosition());
		}
	}

	
}

void AVP_ClientVehicle_Dropship::ProcessClientUpdate(const AVP_Message_DropshipClientUpdate* pxData, float fServerTime, bool bProcessPosAndOri /*= true*/)
{
#ifdef ASURA_NO_NETWORK_LAG_COMPENSATION
	const bool bAllowPosAndOrientationProcessing	= bProcessPosAndOri; 
#else
	const bool bAllowPosAndOrientationProcessing	= bProcessPosAndOri && (m_pxPosAndOriInterpolator != NULL);
#endif // #ifdef ASURA_NO_NETWORK_LAG_COMPENSATION

	// Call the base-class...
	PARENT::ProcessClientUpdate(pxData, fServerTime, bAllowPosAndOrientationProcessing);
	m_fPitchFraction = Asura_Maths::ClampToRange(pxData->GetVelocityFraction(), m_fMinEnginePitch, m_fMaxEnginePitch);
	m_fVolumeFraction = Asura_Maths::ClampToRange(pxData->GetVelocityFraction(), m_fMinEngineVolume, m_fMaxEngineVolume);
	if (m_pxLaserAttachment)
	{
		if (pxData->IsLaserOn())
		{
			if (!m_pxLaserAttachment->IsLinked() && GetAnimationHandler() && GetAnimationHandler()->GetAnimation())
			{
#ifdef ASURA_ASSERT
				bool bOK = 
#endif
					m_pxLaserAttachment->LinkAttachment(GetAnimationHandler()->GetAnimation(), dbg_uLaserAttachementHashID, false);
				Asura_Assertf( bOK, ( "Unable to link laser to marker point 0x%08x", dbg_uLaserAttachementHashID ) );

			}
			m_pxLaserAttachment->SetTarget(pxData->GetLaserTarget());
			m_pxLaserAttachment->SetTargetEnabled(true);
		}
		else
		{
			if (m_pxLaserAttachment->IsLinked())
			{
				m_pxLaserAttachment->UnlinkAttachment(false);
			}
		}
	}
}

void AVP_ClientVehicle_Dropship::CreateAnimationHandler()
{
	Asura_Assert( m_pxAnimationHandler == NULL, "Already got an animation handler" );
	m_pxAnimationHandler = new AVP_Dropship_AnimationHandler (*this);
	m_pxAnimationHandler->Initialise();
}

AVP_ClientVehicle_Dropship* AVP_ClientVehicle_Dropship::Create (const Asura_Message* pxMessage)
{
	AVP_ClientVehicle_Dropship*	pxClient	= new AVP_ClientVehicle_Dropship;

	// Initialise the entity...
	pxClient->HandleMessage(pxMessage);

	// Return the entity...
	return(pxClient);
}

void AVP_ClientVehicle_Dropship::InitialiseFromBlueprint( const Asura_Blueprint& xBlueprint )
{
	PARENT::InitialiseFromBlueprint(xBlueprint);

	Asura_Value xValue;
	if ( xBlueprint.GetParameterValueByHashID( 0x43254979, xValue ) ) //"EngineSound.Event"
	{
		m_uEnginesSoundHash = xValue.GetHashIDValue();
	}
	if ( xBlueprint.GetParameterValueByHashID( 0x43e8850b, xValue ) ) //"EngineSound.ExtraEvent"
	{
		m_uExtraEnginesSoundHash = xValue.GetHashIDValue();
	}
	if ( xBlueprint.GetParameterValueByHashID( 0x6eee08ef, xValue ) ) //"EngineSound.MinPitch"
	{
		m_fMinEnginePitch = xValue.GetFloatValue();
	}
	if ( xBlueprint.GetParameterValueByHashID( 0xd8cca51d, xValue ) ) //"EngineSound.MaxPitch"
	{
		m_fMaxEnginePitch = xValue.GetFloatValue();
	}
	if ( xBlueprint.GetParameterValueByHashID( 0x7961574b, xValue ) ) //"EngineSound.MinVolume"
	{
		m_fMinEngineVolume = xValue.GetFloatValue();
	}
	if ( xBlueprint.GetParameterValueByHashID( 0x4b5640dd, xValue ) ) //"EngineSound.MaxVolume"
	{
		m_fMaxEngineVolume = xValue.GetFloatValue();
	}

	m_fMinEnginePitch = Asura_Maths::ClampToRange(m_fMinEnginePitch, 0.0f, 1.0f);
	m_fMaxEnginePitch = Asura_Maths::ClampToRange(m_fMaxEnginePitch, 0.0f, 1.0f);
	m_fMinEngineVolume = Asura_Maths::ClampToRange(m_fMinEngineVolume, 0.0f, 1.0f);
	m_fMaxEngineVolume = Asura_Maths::ClampToRange(m_fMaxEnginePitch, 0.0f, 1.0f);
	if (m_fMaxEnginePitch < m_fMinEnginePitch)
	{
		float fTemp = m_fMaxEnginePitch; 
		m_fMaxEnginePitch = m_fMinEnginePitch;
		m_fMinEnginePitch = fTemp;
	}
	if (m_fMaxEngineVolume < m_fMinEngineVolume)
	{
		float fTemp = m_fMaxEngineVolume; 
		m_fMaxEngineVolume = m_fMinEngineVolume;
		m_fMinEngineVolume = fTemp;
	}


}

void AVP_ClientVehicle_Dropship::OnEntityActivate()
{
	PARENT::OnEntityActivate();

	const Asura_Sound_Event* pxEngineLoop = Asura_Sound_Event_Storage::FindEvent(m_uEnginesSoundHash);
	if(pxEngineLoop)
	{
		if (m_iEnginesSoundHandle != ASURA_INVALID_HANDLE && !Asura_Sound_Event_System::IsPlaying(m_iEnginesSoundHandle))
		{
			Asura_Sound_Event_System::Stop( m_iEnginesSoundHandle );
		}
		if (m_iEnginesSoundHandle == ASURA_INVALID_HANDLE)
		{
			const Asura_Vector_3& xPosition = GetCurrentPosition();
			Asura_Sound_Event_System::PlayEvent(pxEngineLoop, &m_iEnginesSoundHandle, &xPosition, 0.0f); // start silent
		}
	}

	const Asura_Sound_Event* pxExtraEngineLoop = Asura_Sound_Event_Storage::FindEvent(m_uExtraEnginesSoundHash);
	if(pxExtraEngineLoop)
	{
		if (m_iExtraEnginesSoundHandle != ASURA_INVALID_HANDLE && !Asura_Sound_Event_System::IsPlaying(m_iExtraEnginesSoundHandle))
		{
			Asura_Sound_Event_System::Stop( m_iExtraEnginesSoundHandle );
		}
		if (m_iExtraEnginesSoundHandle == ASURA_INVALID_HANDLE)
		{
			const Asura_Vector_3& xPosition = GetCurrentPosition();
			Asura_Sound_Event_System::PlayEvent(pxExtraEngineLoop, &m_iExtraEnginesSoundHandle, &xPosition, 1.0f); 
		}
	}

}

void AVP_ClientVehicle_Dropship::OnEntityDeactivate()
{
	PARENT::OnEntityDeactivate();

	if(m_iEnginesSoundHandle != ASURA_INVALID_HANDLE && Asura_Sound_Event_System::IsPlaying(m_iEnginesSoundHandle))
	{
		Asura_Sound_Event_System::Stop(m_iEnginesSoundHandle);
	}
}


/// End of function 'AVP_ClientActor_Alien::Create'. ///



/// End of file. ///
