//////////////////////////////////////////////////////////////////////////////
//
// AVP_ServerActor_Vertebrate.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ServerActor_Vertebrate.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ClientActor_Vertebrate.h"
// Engine independent.
#include "Affectors/Asura_Affector.h"
#include "Affectors/Asura_Affector_System.h"
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Mesh/Asura_Env_System.h"
#include "Network/Asura_Network.h"
#include "Network/Asura_Message.h"
#include "Shapes/Asura_Shape_System.h"
#include "Physics/Asura_Physics.h"
#include "Camera/Asura_Camera.h"

// Project independent.
#include "Actor/AVP_Actor_InputHandler.h"
#include "Actor/AVP_Actor_PhysicsObjectHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_AnimationHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Description.h"
#include "Actor/Vertebrates/AVP_Vertebrate_MeleeCounter_System.h"
#include "Actor/Vertebrates/Predator/AVP_Predator_Recon.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_Settings.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Controller/AVP_ServerEntity_Team_Marine.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/GamePlayVolume/AVP_ServerEntity_GamePlayVolume.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_ClientActor_Vertebrate.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_Vertebrate_Messages.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Moves/AVP_Movement_System.h"
#include "Moves/AVP_Moves.h"
#include "Multiplayer/AVP_Multiplayer_GameRules.h"

// Axon
#include "Core/Axon_Messages.h"
#include "Core/Axon_Project_System.h"
#include "Core/Axon_Sound_Registry.h"
#include "Core/BrainTypes/AVP_Brain_Actor.h"
#include "Behaviour/BehaviourTypes/AVP_Behaviour_Types.h"
#include "Tactical/AVP_Tactical_System.h"

// Constant variables...

DEBUGVAR float dbg_fMaxMovingOrientationAngVel = Asura_Maths::DegToRad( 720.0f ); // can rotate up to 720 degrees per second

// Static variables...
Asura_Hashed_List<AVP_ServerActor_Vertebrate> AVP_ServerActor_Vertebrate::s_xList(2);

const char* const AVP_ServerActor_Vertebrate::s_aaszFlinchParameterNames[AVP_ServerActor_Vertebrate::NUM_DIRECTIONS][AVP_ServerActor_Vertebrate::NUM_STANCES] =
{
	{// FORWARD_LEFT
		"StandingForwardLeft",
		"CrouchingForwardLeft",
	},
	{// FORWARD_RIGHT
		"StandingForwardRight",
		"CrouchingForwardRight",
	},
	{// BACKWARD_LEFT
		"StandingBackwardLeft",
		"CrouchingBackwardLeft",
	},
	{// BACKWARD_RIGHT
		"StandingBackwardRight",
		"CrouchingBackwardRight",
	},
	{// FORWARD
		"StandingForward",
		"CrouchingForward",
	},
	{// BACKWARD
		"StandingBackward",
		"CrouchingBackward",
	},
};

static const Asura_Hash_ID s_uDamageAllElementsParam	= 0x7fe33eec;	// DamageAllBodyParts

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

AVP_ServerActor_Vertebrate::AVP_ServerActor_Vertebrate(const u_short usFlagsToIgnore,
													   const Asura_Guid uGUID)
: Asura_Entity_Server(uGUID)
, PARENT(usFlagsToIgnore, uGUID)
, m_pxMeleeSystem(NULL)
, m_xStoredFlinchEffect(Asura_Maths::ZeroVector3)
, m_fStoredFlinchImpulse(0.0f)
, m_fLastFlinchMinImpulse(0.0f)
, m_fTimeSinceLastFlinch(Asura_Maths::LargeFloat)
, m_fFlinchAnimDuration(0.0f)
, m_bCanFlinch( true )
{
	Classification = AVP_ENTITYCLASS_VERTEBRATE;

	// Add to the list...
	s_xList.AddToList(this);
}

AVP_ServerActor_Vertebrate::~AVP_ServerActor_Vertebrate()
{
	delete m_pxMeleeSystem;
	// Remove from the list...
	s_xList.RemoveFromList(this);
}

void AVP_ServerActor_Vertebrate::Update()
{
	const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(GetGuid());
	if(IsDead() && pxClient)
	{
		// We need to wait for our client to ragdoll before we destroy our physics object handler.
		const AVP_Vertebrate_AnimationHandler* pxAnimHandler = static_cast<const AVP_Vertebrate_AnimationHandler*>(pxClient->GetAnimationHandler());
		if (!pxAnimHandler || pxAnimHandler->IsRagdoll())
		{
			if (m_pxPhysicsObjectHandler)
			{
				delete(m_pxPhysicsObjectHandler);
				m_pxPhysicsObjectHandler	= NULL;
				m_pxDynamicObject			= NULL;
			}
			if ( pxAnimHandler && pxAnimHandler->IsRagdoll() )
			{
				TeleportTo( pxClient->GetCurrentPosition(), pxClient->GetCurrentOrientation(), false );
			}
		}
	}

	PARENT::Update();

	if(m_pxMeleeSystem)
	{
		m_pxMeleeSystem->Update();
	}

	if (m_fTimeSinceLastFlinch < Asura_Maths::LargeFloat)
	{
		m_fTimeSinceLastFlinch += Asura_Timers::GetGameTimeSlice();
	}

	StartFlinch();
}

void AVP_ServerActor_Vertebrate::PostLoadInitialise()
{
	PARENT::PostLoadInitialise();

	if ( IsDead() )
	{
		return;
	}

	if (m_pxMeleeSystem)
	{
		m_pxMeleeSystem->PostLoadInitialise();
	}
}

void AVP_ServerActor_Vertebrate::InitialiseFromBlueprint( const Asura_Blueprint& xBlueprint )
{
    PARENT::InitialiseFromBlueprint( xBlueprint );

	Asura_Value	xValue;

	for (u_int u=0; u<NUM_AVP_SKIN_ELEMENT_TYPES; ++u )
	{
		if (xBlueprint.GetParameterValueByHashID(Asura_Project_Skin_Elements::GetElementBlueprintParameter(u), xValue))
		{
			m_axSkinElementData[u].m_uBlueprint = xValue.GetHashIDValue();
		}
		else
		{
			m_axSkinElementData[u].m_uBlueprint = ASURA_HASH_ID_UNSET;
		}

		m_axSkinElementData[u].m_fCurrentHealth = 0.0f;
		m_axSkinElementData[u].m_fDamageMultiplier = 1.0f;
		m_axSkinElementData[u].m_uFlags = 0;

		const Asura_Blueprint* pxSkinElementBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_ACTORS, m_axSkinElementData[u].m_uBlueprint);
		if (pxSkinElementBlueprint)
		{
			if (pxSkinElementBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_SKINELEMENT_INITIALHEALTH, xValue))
			{
				const float fInitialHealth = xValue.GetFloatValue();
				
				m_axSkinElementData[u].m_fCurrentHealth = fInitialHealth;

				// This signifies that the element can be destroyed, as well as how much damage it can take.
				if (fInitialHealth > 0.0f)
				{
					m_axSkinElementData[u].m_bDestructible = true;
				}
			}

			if (pxSkinElementBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_SKINELEMENT_ESSENTIAL, xValue))
			{
				m_axSkinElementData[u].m_bEssential = xValue.GetBoolValue();
			}

			if (pxSkinElementBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_SKINELEMENT_DAMAGEMULTIPLIER, xValue))
			{
				m_axSkinElementData[u].m_fDamageMultiplier = xValue.GetFloatValue();
			}
		}
	}

	if (m_pxMeleeSystem)
	{
		m_pxMeleeSystem->InitialiseFromBlueprint(xBlueprint);
	}
}

void AVP_ServerActor_Vertebrate::OnDeath(const AVP_EntityDamage_ApplyToEntity& xDamage)
{
	PARENT::OnDeath(xDamage);

	// Killed by plasma damage, disintegrated...
	if ( ExplodesOnFatalPlasmaDamage() &&
		(xDamage.GetTypeHash() == uAVP_BLUEPRINT_NAME_DAMAGE_PLASMA
		|| xDamage.GetTypeHash() == uAVP_BLUEPRINT_NAME_DAMAGE_PLASMA_DISRUPTION
		|| xDamage.GetTypeHash() == uAVP_BLUEPRINT_NAME_DAMAGE_PLASMA_IMPACT))
	{
		// if we totaly explode we dont want this acid volume that will have been created
		// due to being decapitated as our explosion removed our head.
		DestroyAcidVolume();
		SetThisAndClientsAsDestroyed();
	}
}


float AVP_ServerActor_Vertebrate::CalculateAndUpdateLocationSpecificDamage(const AVP_EntityDamage_RequestVictim& xDamage, u_int& uDestroyedSkinElements, u_int& uGibbedSkinElements, bool bForceFatal)
{
	uDestroyedSkinElements = 0;
	uGibbedSkinElements = 0;
	const float fAmount = xDamage.GetAmount();

	// Indestructible entities losing limbs is not a good idea.
	if (!IsDestructible()) { return fAmount; }

	// if damaging limb is blocked return unless the actor is dieing anyhow.
	if (fAmount < GetHealth() && !m_bAllowDamageLimb)
	{
		return fAmount;
	}

	// Damage body part
	bool bDamageAllParts = false;
	Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_DAMAGE, xDamage.GetTypeHash() );
	if ( pxBlueprint )
	{
		Asura_Value xValue;
		if ( pxBlueprint->GetParameterValueByHashID( s_uDamageAllElementsParam, xValue ) )
		{
			bDamageAllParts = xValue.GetBoolValue();
		}
	}

	if (bDamageAllParts)
	{
		// TODO - may want some special handling, similar to the old UpdateDestroyedSkinElements function.
		return fAmount;
	}

	const int iBoneID = xDamage.GetBoneID();

	// We only need to do special cases if we've got a limb to hit.
	if (iBoneID < 0) { return fAmount; }
	// Also, bites don't cause dismemberment.
	if (xDamage.GetTypeHash() == uAVP_BLUEPRINT_NAME_DAMAGE_BITE) { return fAmount; }

	const u_int uDamagedSkinElement = GetElementFromBoneID(iBoneID);

	if (uDamagedSkinElement >= NUM_AVP_SKIN_ELEMENT_TYPES) { return fAmount; }

	// Do damage to the limb itself. Use the unmodified damage for ease of setting.
	// Note that 0 health isn't destroying it - this is to allow indestructible elements to still have a way of checking if they've been removed by a parent element getting destroyed.
	if (m_axSkinElementData[uDamagedSkinElement].m_bDestructible && (m_axSkinElementData[uDamagedSkinElement].m_fCurrentHealth >= 0.0f))
	{
		m_axSkinElementData[uDamagedSkinElement].m_fCurrentHealth -= fAmount;
		if (m_axSkinElementData[uDamagedSkinElement].m_fCurrentHealth < 0.0f)
		{
			// Limb destroyed
			uDestroyedSkinElements |= 1 << uDamagedSkinElement;
			if (GetDamageCreatesGib(xDamage))
			{
				uGibbedSkinElements |= 1 << uDamagedSkinElement;
			}
			bool bFatal = m_axSkinElementData[uDamagedSkinElement].m_bEssential;

			// Also mark all child elements as destroyed.
			for (u_int u=0; u<NUM_AVP_SKIN_ELEMENT_TYPES; ++u)
			{
				if (m_axSkinElementData[u].m_fCurrentHealth >= 0.0f)
				{
					// We've already destroyed the limb above, so the next function will return false for any children of it.
					if (!IsSkinElementAvailable(u))
					{
						m_axSkinElementData[u].m_fCurrentHealth = -1.0f;
						uDestroyedSkinElements |= 1 << u;
						bFatal |= m_axSkinElementData[u].m_bEssential;
					}
				}
			}

			if (bFatal)
			{
				// Just return a fatal amount of damage.
				return 1000000.0f;
			}
		}
	}

	// Finally, multiply the overall character damage by the limb multiplier.
	return fAmount * m_axSkinElementData[uDamagedSkinElement].m_fDamageMultiplier;
}

void AVP_ServerActor_Vertebrate::DestroySkinElement(const u_int uElement)
{
	m_axSkinElementData[uElement].m_fCurrentHealth = -1.0f;
	for (u_int u=0; u<NUM_AVP_SKIN_ELEMENT_TYPES; ++u)
	{
		if (m_axSkinElementData[u].m_fCurrentHealth >= 0.0f)
		{
			// We've already destroyed the limb above, so the next function will return false for any children of it.
			if (!IsSkinElementAvailable(u))
			{
				m_axSkinElementData[u].m_fCurrentHealth = -1.0f;
			}
		}
	}
}

bool AVP_ServerActor_Vertebrate::GetDamageCreatesGib(const AVP_EntityDamage_RequestVictim& xDamage) const
{
	// Basic version - just switch on damage type.
	switch (xDamage.GetTypeHash())
	{
	case uAVP_BLUEPRINT_NAME_DAMAGE_SPEAR:	// FALL THROUGHT !!!
	case uAVP_BLUEPRINT_NAME_DAMAGE_DISC:	// FALL THROUGHT !!!
	case uAVP_BLUEPRINT_NAME_DAMAGE_BLADE:
		return true;

	default:
		return false;
	}
}

void AVP_ServerActor_Vertebrate::TakeDamageExtended(const AVP_EntityDamage_RequestVictim& xDamage,
											const bool bForceFatal, const bool bSendMessageToClient )
{
	if(IsEntityActive())
	{
		AVP_EntityDamage_RequestVictim xLocalDamage = xDamage;

		// Friendly fire checks.
		const Asura_Guid uFirerGuid = xDamage.GetFirerGuid();
		if (uFirerGuid != Guid)
		{
			const AVP_ServerActor_Vertebrate* pxAttacker = AVP_ServerActor_Vertebrate::FindEntity(xDamage.GetFirerGuid());
			if (pxAttacker && (AVP_Tactical_System::GetAllegiance(*this, *pxAttacker) == AVP_ALLEGIANCE_FRIEND))
			{
				if (AVP_Settings::IsMultiPlayer())
				{
					if (!AVP_Multiplayer_GameRules::GetFriendlyFireOn())
					{
						xLocalDamage.SetAmount(0.0f);
					}
				}
				else if (AVP_Vertebrate_MeleeCounter_System::IsMeleeDamageType(xLocalDamage.GetTypeHash()))
				{
					xLocalDamage.SetAmount(0.0f);
				}
			}
		}

		PARENT::TakeDamageExtended( xLocalDamage, bForceFatal, bSendMessageToClient );


		// DWW 12/12/08, do not do finch damage unless it is requested, (part of the alien queen behaviours, only shots from the grenades
		// will cause a flinch move) 
		if (!xLocalDamage.GetDoNotFlinch())
		{
			AddFlinchEffect(xLocalDamage);
		}
	}
}

void AVP_ServerActor_Vertebrate::RegisterMovementSoundsForAI()
{
	float	fFootstepRange = 0.0f;
	bool	bRegister = false;

	if ( IsMovingMedium() )
	{
		fFootstepRange = Axon_Project_System::GetFootstepSoundRangeMed();
		bRegister = true;
	}
	else if ( IsMovingFast() )
	{
		fFootstepRange = Axon_Project_System::GetFootstepSoundRangeHigh();
		bRegister = true;
	}

	if ( bRegister )
	{
		Axon_System::GetSoundRegistry()->RegisterSound
		(
			GetGuid(),
			GetClassification(),
			GetPosition(),
			fFootstepRange,
			AXON_SOUND_TYPE_FOOTSTEP
		);
	}
}

bool AVP_ServerActor_Vertebrate::IsSkinElementAvailable(u_int uSkinElement) const
{
	while (uSkinElement < NUM_AVP_SKIN_ELEMENT_TYPES)
	{
		if (GetSkinElementHealth(uSkinElement) < 0.0f)
		{
			return false;
		}

		uSkinElement = Asura_Project_Skin_Elements::GetParentElementType(uSkinElement);
	}

	return true;
}

u_int AVP_ServerActor_Vertebrate::GetDestroyedSkinElementFlags() const
{
	// Go through the array and fill in the bitfield.
	// Destructibility is irrelevant as an indestructible element can be the child of a destructible one. Just check the health is not negative (zero counts as alive for this, to simplify the indestructible-destroying case).
	u_int uFlags = 0;
	for (u_int u=0; u<NUM_AVP_SKIN_ELEMENT_TYPES; ++u)
	{
		if (m_axSkinElementData[u].m_fCurrentHealth < 0.0f)
		{
			uFlags |= 1 << u;
		}
	}

	return uFlags;
}


Asura_Guid AVP_ServerActor_Vertebrate::GetCurrentMeleeTarget() const
{
	if (!m_pxMeleeSystem) { return ASURA_GUID_UNREGISTERED; }

	return m_pxMeleeSystem->GetCurrentTarget();
}

void AVP_ServerActor_Vertebrate::GetMeleeOrigin(Asura_Vector_3& xOrigin) const
{
	if (!GetWorldSpaceMarkerPoint(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_MELEEORIGIN), &xOrigin, NULL))
	{
		xOrigin = GetPosition();

		const Asura_Matrix_3x3& xServerOrientation = GetOrientation();
		const Asura_Vector_3 xDown( xServerOrientation.e12, xServerOrientation.e22, xServerOrientation.e32 );
		xOrigin += xDown * GetChestPositionYOffset();
	}
}


// TODO - this probably wants to be the arbitrating inventory (once such a thing exists), rather than the master one.
AVP_Inventory_System* AVP_ServerActor_Vertebrate::GetMasterInventory()
{
	// HACK - This first version gets the server inventory for AIs, the client inventory for local players, and doesn't work for remote players.
	if (!IsPlayer())
	{
		return GetInventorySystem();
	}
	else if (AVP_Settings::IsSinglePlayer())
	{
		AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity(Guid);
		if (pxClient)
		{
			return pxClient->GetInventorySystem();
		}
	}

	return NULL;
}


bool AVP_ServerActor_Vertebrate::HandleMessage( const Asura_Message* pxMessage )
{
	if(!m_pxMeleeSystem || !m_pxMeleeSystem->HandleMessage(pxMessage))
	{
		switch( pxMessage->GetMsgID() )
		{
			case AVP_MSGID_USE_STIM:
			{
				return(true);
			}

			case AVP_MSGID_INCOMING_MELEE_STARTED:
			case AVP_MSGID_INCOMING_MELEE_FINISHED:
			case AVP_MSGID_MELEE_ATTACK_FAILED_TO_START:
			{
				if ( GetActorBrain() )
				{
					GetActorBrain()->DoHandleMessage( pxMessage );
				}
				return true;
			}

			case AVP_MSGID_COUNTER_MELEE:
			{
				if ( GetActorBrain() )
				{
					GetActorBrain()->DoHandleMessage( pxMessage );
				}
				// We have been countered so empty our melee queue;
				if(GetMeleeSystem())
				{
					GetMeleeSystem()->ClearAllRequests();
					GetMeleeSystem()->SetCurrentAttackCountered(true);
				}
				// fall through to forward the message
			}
			case AVP_MSGID_MELEE_ATTACK_HIT:
			case AVP_MSGID_MELEE_ATTACK_PREPAREGRAPPLE:
			case AVP_MSGID_GRAPPLE_ATTACKER_UPDATE:
			case AVP_MSGID_GRAPPLE_VICTIM_UPDATE:
			case AVP_MSGID_ACTIVEGRAPPLE_RESULT:				
			{
				if ( GetActorBrain() )
				{
					GetActorBrain()->DoHandleMessage( pxMessage );
				}

				if ( m_bSentClientUpdateMessage )
				{
					// If we've already sent a client update message we can forward this message straight on.
					Asura_Network::Forward( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, GetGuid() );
				}
				return true;
			}

			case AVP_MSGID_ACTIVEGRAPPLE_VALUE:
			{
				if (m_pxMeleeSystem)
				{
					m_pxMeleeSystem->HandleMessage(pxMessage);
				}

				return true;
			}

			case AVP_MSGID_DISMEMBER:
			{
				// We need to handle damage locally, so we're killed, then pass this message on to clients to create the ragdoll and gib.
				// Use a special damage type which instructs the damage system not to forward to the client.
				const AVP_Msg_Dismember* pxData = static_cast<const AVP_Msg_Dismember*>(pxMessage->GetData());
				if (pxData)
				{
					const float fDamage = pxData->GetDamage();
//					const u_int uDismemberedElement = pxData->GetElement();	// Check if the limb is essential. I wrote this part for enabling some SpearGun functionality, but since SpearGun has been removed, I am returning the code to original state. HOwever, I am keeping it here, because one day it may become usefull again.
//					bool bIsLimbEssential = m_axSkinElementData[uDismemberedElement].m_bEssential;

					if (fDamage < 0.0f /*|| bIsLimbEssential*/ ) // Automatic kill - handled in full by this message.	// ( Do Instant kill also when we are trying to remove limb that is essential)
					{
						AVP_EntityDamage_RequestVictim xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_DISMEMBER, pxData->GetFromGuid(), ASURA_HASH_ID_UNSET, Guid, 1000000.0f, GetPosition(), Asura_Vector_3(0.0f, 1.0f, 0.0f), Asura_Vector_3(0.0f, 1.0f, 0.0f));
						TakeDamageExtended(xDamage, true);
					}
					else // Not an automatic kill - let the normal damage code handle this
					{
						// Hmm, melee isn't really guaranteed to be the correct damage type. It'll do for now though.
						AVP_EntityDamage_RequestVictim xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_MELEE, pxMessage->From, ASURA_HASH_ID_UNSET, Guid, fDamage, pxData->GetDamagePosition(), pxData->GetRagdollDirection(), -pxData->GetRagdollDirection(), pxData->GetRagdollImpulse());
						TakeDamageExtended(xDamage, false);
					}

					// Whatever happens, we need to make sure this removes the chosen skin element.
					const u_int uDismemberedElement = pxData->GetElement();
					if (m_axSkinElementData[uDismemberedElement].m_fCurrentHealth >= 0.0f)
					{
						m_axSkinElementData[uDismemberedElement].m_fCurrentHealth = -1.0f;
						for (u_int u=0; u<NUM_AVP_SKIN_ELEMENT_TYPES; ++u)
						{
							if (m_axSkinElementData[u].m_fCurrentHealth >= 0.0f)
							{
								// We've already destroyed the limb above, so the next function will return false for any children of it.
								if (!IsSkinElementAvailable(u))
								{
									m_axSkinElementData[u].m_fCurrentHealth = -1.0f;
								}
							}
						}
					}

					Asura_Network::Forward(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, Guid);
				}

				return true;
			}

			case AVP_MSGID_FLINCH:
				if ( GetActorBrain() )
				{
					GetActorBrain()->DoHandleMessage( pxMessage );
				}
				if ( m_bSentClientUpdateMessage )
				{
					// If we've already sent a client update message we can forward this message straight on.
					Asura_Network::Forward( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES, GetGuid() );
				}
				return true;

			case AVP_MSGID_GRAPPLE_START:
				if ( GetActorBrain() )
				{
					GetActorBrain()->DoHandleMessage( pxMessage );
				}
				return true;

			default:
			{
				return(PARENT::HandleMessage(pxMessage));
			}
		}
	}
	
	return true; // handled by the melee system	
}

void AVP_ServerActor_Vertebrate::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	// Read the version...
	xStream >> uVersion;
	Asura_Assert((uVersion <= uCURRENT_CHUNK_VERSION), "Unsupported chunk version");

	// Read the parent data...
	PARENT::ReadFromChunkStream(xStream);

		// Read limb health data
	if (uVersion >= 2)
	{
		int iNumElementTypes;
		xStream >> iNumElementTypes;

		for (int i=0; i<iNumElementTypes; ++i)
		{
			if (i < NUM_AVP_SKIN_ELEMENT_TYPES)
			{
				xStream >> m_axSkinElementData[i].m_fCurrentHealth;
			}
			else
			{
				// This element must no longer exist, junk the data.
				float fJunk;
				xStream >> fJunk;
			}
		}
		// Fill in a default value for any gaps.
		for (int i=iNumElementTypes; i<NUM_AVP_SKIN_ELEMENT_TYPES; ++i)
		{
			m_axSkinElementData[i].m_fCurrentHealth = 0.0f;
		}
	}
	else if( uVersion == 1 )
	{
		// In version 1, this had to be 15 (and we didn't save how many there were).
		int iNumElementTypes = 15;

		for (int i=0; i<iNumElementTypes; ++i)
		{
			if (i < NUM_AVP_SKIN_ELEMENT_TYPES)
			{
				xStream >> m_axSkinElementData[i].m_fCurrentHealth;
			}
			else
			{
				// This element must no longer exist, junk the data.
				float fJunk;
				xStream >> fJunk;
			}
		}
		// Fill in a default value for any gaps.
		for (int i=iNumElementTypes; i<NUM_AVP_SKIN_ELEMENT_TYPES; ++i)
		{
			m_axSkinElementData[i].m_fCurrentHealth = 0.0f;
		}
	}
}

void AVP_ServerActor_Vertebrate::WriteToClientCreateStream(Asura_Chunk_Memory_Stream& xStream) const
{
	xStream << static_cast<u_int>(CURRENT_CLIENTCREATE_VERSION);

	PARENT::WriteToClientCreateStream(xStream);

	const AVP_Vertebrate_MeleeCounter_System* pxMeleeSystem = GetMeleeSystem();

	bool bEnemyInMeleeRange = false;
	bool bBlockActive = false;
	if (pxMeleeSystem)
	{
		bEnemyInMeleeRange = pxMeleeSystem->GetHasAttackTarget();
		bBlockActive = pxMeleeSystem->IsBlockActive();
	}

	xStream << bEnemyInMeleeRange;
	xStream << bBlockActive;
	xStream << GetDestroyedSkinElementFlags();
}

void AVP_ServerActor_Vertebrate::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// Write the version...
	xStream << uCURRENT_CHUNK_VERSION;

	// Write the parent data...
	PARENT::WriteToChunkStream(xStream);

	int iNumElementTypes = NUM_AVP_SKIN_ELEMENT_TYPES;
	xStream << iNumElementTypes;

	// Write limb health data
	for (int i=0; i<NUM_AVP_SKIN_ELEMENT_TYPES; ++i)
	{
		xStream << m_axSkinElementData[i].m_fCurrentHealth;
	}
}


void AVP_ServerActor_Vertebrate::MakePlayer(Asura_Guid uPlayerGuid)
{
	if (m_pxMeleeSystem)
	{
		m_pxMeleeSystem->MakePlayer();
	}

	PARENT::MakePlayer(uPlayerGuid);
}

void AVP_ServerActor_Vertebrate::MakeAI()
{
	if (m_pxMeleeSystem)
	{
		m_pxMeleeSystem->MakeAI();
	}

	PARENT::MakeAI();
}


void AVP_ServerActor_Vertebrate::CreateAIDistraction(const Asura_Guid uDistractionTarget/* = ASURA_GUID_UNREGISTERED*/) const
{
	// Send a message to the marine team letting them know a distraction has been activated (the distraction target guid is sent in this message.
	// TODO - do we need to do anything special for predators or aliens too? Do we need to send a message to the target too in case they aren't a marine?
	const AVP_ServerEntity_Team_Marine* pxMarineTeam = AVP_ServerEntity_Team_Marine::GetInstance();
	if (pxMarineTeam)
	{
		Asura_Message* pxAIMessage = Asura_Network::CreateMessage(AVP_MSGID_DISTRACT_AI, sizeof(AVP_Msg_DistractAI));
		if (pxAIMessage)
		{
			AVP_Msg_DistractAI* pxData = static_cast<AVP_Msg_DistractAI*>(pxAIMessage->GetData());
			pxData->SetTargetGuid(uDistractionTarget);

			// always have this false from this funciton. (because this is not a recon distraction.)
			pxData->SetReconBool( false );

			Send(pxAIMessage, pxMarineTeam->GetGuid());
		}
	}

	// Send a message to all clients to play the distraction sound.
	Asura_Message* pxSoundMessage = Asura_Network::CreateMessage( AVP_MSGID_PLAY_DISTRACTION_SOUND, sizeof( AVP_Msg_CreateSound ) );
	if (pxSoundMessage)
	{
		AVP_Msg_CreateSound* pxData = static_cast<AVP_Msg_CreateSound*>( pxSoundMessage->GetData() );

		if( pxData )
		{
			pxData->SetOriginatesFromPlayer( true );
			pxData->SetSoundHashId( 0x41F14CC2 ); // Asura_GetHash_ID("Distraction");
			Send( pxSoundMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function : AVP_ServerActor_Vertebrate::CreateAIReconDistraction
// Scope	: Public
// Author   : Callum Hope
// Date     : 27-Nov-2008
//
// Desc		: Creates the messages needed to create a distraction as a result of the Predator recon mode. 
//			  This has a seperate function ( as opposed to sharing "CreateAIDistraction" functionality ) because they need to acheive is ver dufferent
//			  This requires a distraction from a point other than the position of the player, and needs to operate independantly of any marine team
//			  To do these things without passing potentially un-needed data in every control message we have a seperate function that talks directly to the singleton Predator Recon Mode structure. 
////////////////////////////////////////////////////////////////////////////////

void AVP_ServerActor_Vertebrate::CreateAIReconDistraction( ) const
{
	// Send a message to the marine team letting them know a distraction has been activated (the distraction target guid is sent in this message.
	Asura_Message* pxAIMessage = Asura_Network::CreateMessage( AVP_MSGID_DISTRACT_AI, sizeof( AVP_Msg_DistractAI ) );
	if( pxAIMessage )
	{
		AVP_Msg_DistractAI* pxData = static_cast<AVP_Msg_DistractAI*>( pxAIMessage->GetData() );

		// the only data we need is to know if this is a recon mode distraction. 
		pxData->SetReconBool( true );

		// send the message to the AI entity being distracted, nothing else. 
		Send( pxAIMessage, AVP_Predator_Recon::GetRecon().GetPreyGuid() );
	}

	// Send a message to all clients to play the distraction sound.
	Asura_Message* pxSoundMessage = Asura_Network::CreateMessage( AVP_MSGID_PLAY_DISTRACTION_SOUND, sizeof( AVP_Msg_CreateSound ) );
	if (pxSoundMessage)
	{
		AVP_Msg_CreateSound* pxData = static_cast<AVP_Msg_CreateSound*>( pxSoundMessage->GetData() );

		pxData->SetOriginatesFromPlayer( false );
		pxData->SetSoundHashId( 0x41F14CC2 ); // Asura_GetHash_ID("Distraction");

		Send( pxSoundMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
	}
}

bool AVP_ServerActor_Vertebrate::IsFlinching() const
{
	const AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity(Guid);
	if (!pxClient) { return false; } // shouldn't happened

	const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
	if (!pxMovementSystem) { return false; } // shouldn't happened

	const AVP_Move* pxCurrentMove = static_cast<const AVP_Move*>(pxMovementSystem->GetCurrentMove());
	if (!pxCurrentMove) { return false; }  // no move so not flinching

	return pxCurrentMove->GetHash() == AVP_Move_Flinch::GetStaticHash();
}


void AVP_ServerActor_Vertebrate::ProcessVertebratePlayerCommands( const AVP_Message_VertebratePlayerCommands* pxData, float fMessageSentTime  )
{
	ProcessActorPlayerCommands( pxData, fMessageSentTime );

	if (m_pxMeleeSystem)
	{
		const Asura_Guid uTargetGuid = pxData->GetCurrentMeleeTarget();

		m_pxMeleeSystem->RequestTarget(uTargetGuid);
	}
}


void AVP_ServerActor_Vertebrate::AddFlinchEffect(const AVP_EntityDamage_RequestVictim& xDamage)
{
	Asura_Vector_3 xDirection;
	const float fImpulse = xDamage.GetImpulseStrength();
	xDamage.GetDirection(xDirection);
	m_xStoredFlinchEffect += xDirection * fImpulse;
	m_fStoredFlinchImpulse += fImpulse;
}

void AVP_ServerActor_Vertebrate::StartFlinch()
{
	// Copy the values and reset the stored ones immediately.
	const float fAttackStrength = m_fStoredFlinchImpulse;
	Asura_Vector_3 xActualDirection = m_xStoredFlinchEffect;

	m_fStoredFlinchImpulse = 0.0f;
	m_xStoredFlinchEffect.Zero();

	if (fAttackStrength < 0.001f) { return; }

	bool bFlinchInProgress;
	if (IsDead() || !GetCanFlinch(bFlinchInProgress)) { return; }

	// This can be from anywhere - a weapon, the actor, whatever the particular species requires, as long as it has parameters of the correct name.
	const Asura_Blueprint* pxTypesBlueprint = GetCurrentFlinchBlueprint();

	if (!pxTypesBlueprint) { return; }

	const Asura_Hash_ID uFlinchBlueprintType = GetMovesBlueprintType();

	AVP_Vertebrate_FlinchData auFlinchTypes[uAVP_VERTEBRATE_MAX_FLINCH_TYPES];
	u_int uNumFlinchTypes = 0;
	AVP_Blueprint_ParameterEnumerator_FlinchTypes xTypesEnumerator(auFlinchTypes, uNumFlinchTypes, uFlinchBlueprintType);

	pxTypesBlueprint->EnumerateParameters(szAVP_BLUEPRINT_PARAM_FLINCH, 2, xTypesEnumerator);
	if (uNumFlinchTypes == 0) { return; }

	// Pick the appropriate type according to the attack strength.
	u_int uBestIndex = uNumFlinchTypes;
	float fBestAttackStrength = -1.0f;
	for (u_int u=0; u<uNumFlinchTypes; ++u)
	{
		if (auFlinchTypes[u].m_fMinAttackPower > fAttackStrength) { continue; }
		if (auFlinchTypes[u].m_fMinAttackPower < fBestAttackStrength) { continue; }
		if (bFlinchInProgress && (auFlinchTypes[u].m_fMinAttackPower < m_fLastFlinchMinImpulse)) { continue; }

		uBestIndex = u;
		fBestAttackStrength = auFlinchTypes[u].m_fMinAttackPower;
	}

	if (uBestIndex == uNumFlinchTypes) { return; } // Strength too weak to cause a flinch, don't bother.

	// if this flinch matches the last one, check delays.
	if (auFlinchTypes[uBestIndex].m_fMinAttackPower == m_fLastFlinchMinImpulse)
	{
		// Check if we've got past the minimum delay yet
		if (m_fTimeSinceLastFlinch < auFlinchTypes[uBestIndex].m_fMinDelay)
		{
			return;
		}

		// Check the probability of firing if we haven't yet reached the max delay.
		if (m_fTimeSinceLastFlinch < auFlinchTypes[uBestIndex].m_fMaxDelay)
		{
			if (Asura_Random::GetFloat(0.0f, 1.0f) > auFlinchTypes[uBestIndex].m_fProbability)
			{
				return;
			}
		}
	}

	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uFlinchBlueprintType, auFlinchTypes[uBestIndex].m_uFirstPersonBlueprint);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uFlinchBlueprintType, auFlinchTypes[uBestIndex].m_uThirdPersonBlueprint);
	if (!pxFirstPersonBlueprint && !pxThirdPersonBlueprint) { return; }

	// Now we know we have sufficient impulse, turn the impulse effect vector into a direction.
	if (xActualDirection.ApproximatelyEquals(Asura_Maths::ZeroVector3))
	{
		const Asura_Matrix_3x3& xOri = GetOrientation();
		xActualDirection = Asura_Vector_3(xOri.e13, xOri.e23, xOri.e33);
	}
	else
	{
		xActualDirection.Normalise();
	}

	Asura_Vector_3 xRelativeDirection(xActualDirection);
	const Asura_Matrix_3x3& xOri = GetOrientation();
	Asura_Vector_3 xAimDir(xOri.e13, xOri.e23, xOri.e33);
	xAimDir.y = 0.0f;
	if (xAimDir.MagnitudeSquared() > 0.0001f)
	{
		xAimDir.Normalise();
	}
	else
	{
		xAimDir = Asura_Vector_3(0.0f, 0.0f, 1.0f);
	}
	Asura_Matrix_3x3 xAimOri;
	Asura_Maths::OrientationFromDirection(xAimOri, xAimDir);
	xRelativeDirection = xAimOri.ApplyTranspose(xRelativeDirection);

	// By using the impulse strength, we can use this to trigger flinches even if the damage is zero (assuming power makes a difference of course).
	const bool bOrientateToImpactDirection = auFlinchTypes[uBestIndex].m_bOrientateToImpactDirection;

	u_int uDirection;
	// The relative direction is the direction the damage is travelling, and as such is also the direction we want to move in
	if (bOrientateToImpactDirection)
	{
		// This allows the forwards and backwards directions only.
		if (xRelativeDirection.z > 0.0f)
		{
			// Move to the front
			uDirection = FORWARD;
		}
		else
		{
			// Move backwards
			uDirection = BACKWARD;
		}
	}
	else
	{
		// This allows the four diagonals only.
		if (xRelativeDirection.z > 0.0f)
		{
			// Move to the front
			if (xRelativeDirection.x > 0.0f)
			{
				// Move to the right
				uDirection = FORWARD_RIGHT;
			}
			else
			{
				uDirection = FORWARD_LEFT;
			}
		}
		else
		{
			// Move backwards
			if (xRelativeDirection.x > 0.0f)
			{
				// Move to the right
				uDirection = BACKWARD_RIGHT;
			}
			else
			{
				uDirection = BACKWARD_LEFT;
			}
		}
	}

	u_int uStance = GetIsInLowStance() ? CROUCHING : STANDING;

	const char* szBlueprintParameterBase = s_aaszFlinchParameterNames[uDirection][uStance];

	// Store of flinch variation selected by the first person blueprint, so if both first and third exist they pick the same one. Not important if only one blueprint is present.
	int iSelectedVariation = -1;

	Asura_Hash_ID uFirstPersonHash = ASURA_HASH_ID_UNSET;
	if (pxFirstPersonBlueprint)
	{
		AVP_Blueprint_ParameterEnumerator_FlinchSelector xSelectorEnumerator; // No need to pass in the selected variation this time.
		pxFirstPersonBlueprint->EnumerateParameters(szBlueprintParameterBase, 2, xSelectorEnumerator);
		uFirstPersonHash = xSelectorEnumerator.GetSelectedFlinch();
		iSelectedVariation = xSelectorEnumerator.GetSelectedVariation(); // Store this for the third person enumerator.
	}

	Asura_Hash_ID uThirdPersonHash = ASURA_HASH_ID_UNSET;
	if (pxThirdPersonBlueprint)
	{
		AVP_Blueprint_ParameterEnumerator_FlinchSelector xSelectorEnumerator(iSelectedVariation);
		pxThirdPersonBlueprint->EnumerateParameters(szBlueprintParameterBase, 2, xSelectorEnumerator);
		uThirdPersonHash = xSelectorEnumerator.GetSelectedFlinch();
	}

#if defined(ASURA_TOOLS) && defined(ASURA_ASSERT)
	// In tools we can tell people exactly which blueprints are in use.
	if ((uFirstPersonHash == ASURA_HASH_ID_UNSET) && (uThirdPersonHash == ASURA_HASH_ID_UNSET))
	{
		if (pxFirstPersonBlueprint && pxThirdPersonBlueprint)
		{
			Asura_Assertf(false, ("No flinches of type %s found in either %s or %s, has the wrong blueprint for a flinch that %s to the direction of impact been chosen?", s_aaszFlinchParameterNames[uDirection][uStance], pxFirstPersonBlueprint->GetName().GetString(), pxThirdPersonBlueprint->GetName().GetString(), bOrientateToImpactDirection ? "reorientates" : "does not reorientate"));
		}
		else // Already checked we have at least one blueprint.
		{
			Asura_Assertf(false, ("No flinches of type %s found in %s, has the wrong blueprint for a flinch that %s to the direction of impact been chosen?", s_aaszFlinchParameterNames[uDirection][uStance], pxFirstPersonBlueprint ? pxFirstPersonBlueprint->GetName().GetString() : pxThirdPersonBlueprint->GetName().GetString(), bOrientateToImpactDirection ? "reorientates" : "does not reorientate"));
		}
	}
#else
	Asura_Assertf((uFirstPersonHash != ASURA_HASH_ID_UNSET) || (uThirdPersonHash != ASURA_HASH_ID_UNSET), ("No flinches found of type %s, has the wrong blueprint for a flinch that %s to the direction of impact been chosen?", s_aaszFlinchParameterNames[uDirection][uStance], bOrientateToImpactDirection ? "reorientates" : "does not reorientate"));
#endif

	if (uFirstPersonHash == ASURA_HASH_ID_UNSET && uThirdPersonHash == ASURA_HASH_ID_UNSET) { return; }

	Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_FLINCH, sizeof(AVP_Msg_Flinch));
	if (!pxMessage) { return; }

	// Need to calculate a heading to send over to the client. To save calculating a full euler, just use the part of the algorithm for the y component.
	float fHeading = 0.0f;
	if (bOrientateToImpactDirection)
	{
		if (Asura_Maths::Abs(xActualDirection.y) > 0.999f)
		{
			// In this situation we should just base it on our own orientation.
			Asura_Euler_3 xEuler;
			xEuler.FromMatrix(GetOrientation());
			fHeading = xEuler.y;
		}
		else
		{
			fHeading = Asura_Maths::ArcTangent(xActualDirection.x, xActualDirection.z);
			if (uDirection == BACKWARD)
			{
				// We need to face the damage if we're going backwards.
				fHeading = Asura_Maths::RadiansToZeroCentredRange(fHeading + Asura_Maths::Pi);
			}
		}
	}

	m_fLastFlinchMinImpulse = auFlinchTypes[uBestIndex].m_fMinAttackPower;

	const bool bFlinchUsesMove = auFlinchTypes[uBestIndex].m_bUseFlinchMove;

	AVP_Msg_Flinch* pxData = static_cast<AVP_Msg_Flinch*>(pxMessage->GetData());
	pxData->ResetFlags();
	pxData->SetAnimHashes(uFirstPersonHash, uThirdPersonHash);
	pxData->SetTargetHeading(fHeading);
	pxData->SetUseFlinchMove(bFlinchUsesMove);
	pxData->SetAttackBlocked(false); // AJG - 29/09/08 - block flinches disabled for now.
	pxData->SetDisableSurfaceReorientation(auFlinchTypes[uBestIndex].m_bDisableSurfaceReorientation);
	pxData->SetResetAimDirection(auFlinchTypes[uBestIndex].m_bResetAimDirection);
	pxData->SetOrientateToImpactDirection(bOrientateToImpactDirection);

	pxMessage->SetToGuid(Guid);
	pxMessage->SetFromGuid(Guid);

	Asura_Network::Send(pxMessage);

	// keep flinch time
	m_fTimeSinceLastFlinch = 0.0f;
	const Asura_Hierarchy_Anim* pxAnim;
	if (uFirstPersonHash != ASURA_HASH_ID_UNSET)
	{
		pxAnim = Asura_Animation_System::GetAnimByHash(uFirstPersonHash);
	}
	else
	{
		pxAnim = Asura_Animation_System::GetAnimByHash(uThirdPersonHash);
	}

	if (pxAnim)
	{
		m_fFlinchAnimDuration = pxAnim->GetDuration();
	}
}


bool AVP_ServerActor_Vertebrate::GetCanFlinch(bool& bFlinchInProgress) const
{
	bFlinchInProgress = false;

	// Check the can flinch flag to see if we are allowed to flinch. if we are not return false
	if ( !m_bCanFlinch ) { return false; }

	// For now, peek our client and see if we have a flinch-preventing move.
	const AVP_ClientActor_Vertebrate* pxClient = AVP_ClientActor_Vertebrate::FindEntity(Guid);
	if (!pxClient) { return true; }

	const Asura_Movement_System* pxMovementSystem = pxClient->GetMovementSystem();
	if (!pxMovementSystem) { return true; }

	const AVP_Move* pxCurrentMove = static_cast<const AVP_Move*>(pxMovementSystem->GetCurrentMove());
	if (!pxCurrentMove) { return true; }

	bFlinchInProgress = pxCurrentMove->GetHash() == AVP_Move_Flinch::GetStaticHash();

	return !pxCurrentMove->GetPreventsFlinch();
}

void AVP_ServerActor_Vertebrate::UpdateMovement()
{
	// SG - linked objects should not be moving
	if (GetParentGuid() != ASURA_GUID_UNREGISTERED)
	{
		return;
	}

	if (m_pxPhysicsObjectHandler)
	{
		m_eGroundType = m_pxPhysicsObjectHandler->PerformOnGroundCheck();

		const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity(GetGuid());
		const Asura_Movement_System*	pxMovementSystem	= (pxClient == NULL) ? NULL : pxClient->GetMovementSystem();
		const AVP_Move*					pxCurrentMove		= static_cast< const AVP_Move* >( (pxMovementSystem == NULL) ? NULL : pxMovementSystem->GetCurrentMove() );
		const float						fTimeStep			= Asura_Timers::GetGameTimeSlice();

		if(pxCurrentMove == NULL)
		{
			// If we're on external control but not in a move then we're on gamescene and our position will be set by that.
			if ( m_pxPhysicsObjectHandler->GetPhysicsObject()->GetType() != ASURA_PHYSICS_OBJECT_EXTERNALCONTROL )
			{
				Asura_AVector3 xAnimVelocityOS;
				if ( pxClient && GetClientAnimVelocityOS( *pxClient, xAnimVelocityOS ) )
				{
					if ( IsMoving() )
					{
						Asura_AVector3 xAimDirWS = Asura_AVector3( GetAimDirection() );
						Asura_AVector3 xAimDirOS = Asura_AMatrix3x3( GetOrientation() ).ApplyTranspose( xAimDirWS );
						float fAimAngOS = Asura_Maths::ArcTangent( xAimDirOS.GetX(), xAimDirOS.GetZ() );
						fAimAngOS /= Asura_Maths::Max( 0.001f, Asura_Timers::GetGameTimeSlice() );

						m_xObjectSpaceAngVelocityAxis = Asura_AVector3( 0.0f, 1.0f, 0.0f );
						m_fObjectSpaceAngVelocityAngle = Asura_Maths::ClampToRange( fAimAngOS, -dbg_fMaxMovingOrientationAngVel, dbg_fMaxMovingOrientationAngVel );
					}
					else
					{
						if ( !GetClientAnimAngularVelocityOS( *pxClient, m_fObjectSpaceAngVelocityAngle, m_xObjectSpaceAngVelocityAxis ) )
						{
							m_fObjectSpaceAngVelocityAngle = 0.0f;
							m_xObjectSpaceAngVelocityAxis = Asura_AVector3( 0.0f, 1.0f, 0.0f );
						}
					}

					m_pxPhysicsObjectHandler->CalculateMovementUsingAngAndLinVelocity( fTimeStep,
						m_xObjectSpaceAngVelocityAxis,
						m_fObjectSpaceAngVelocityAngle,
						xAnimVelocityOS,
						true,
						AVP_PhysicsObjectHandler::OBJECT_SPACE );
					m_fObjectSpaceAngVelocityAngle = 0.0f;
				}
			}
		}
		else
		{
			bool bUseGravity = !pxCurrentMove->IsGravityIgnored();
			switch(pxCurrentMove->GetPositionUsage())
			{
				case Asura_Move::POSITION_NOT_USED:
				{
					Asura_AVector3 xAnimVelocityOS;
					if ( pxClient && GetClientAnimVelocityOS( *pxClient, xAnimVelocityOS ) )
					{
						if ( pxCurrentMove->GetOrientationUsage() == Asura_Move::ORIENTATION_NOT_USED )
						{
							if ( !GetClientAnimAngularVelocityOS( *pxClient, m_fObjectSpaceAngVelocityAngle, m_xObjectSpaceAngVelocityAxis ) )
							{
								m_fObjectSpaceAngVelocityAngle = 0.0f;
								m_xObjectSpaceAngVelocityAxis = Asura_AVector3( 0.0f, 1.0f, 0.0f );
							}

							m_pxPhysicsObjectHandler->CalculateMovementUsingAngAndLinVelocity( fTimeStep,
								m_xObjectSpaceAngVelocityAxis,
								m_fObjectSpaceAngVelocityAngle,
								xAnimVelocityOS,
								bUseGravity,
								AVP_PhysicsObjectHandler::OBJECT_SPACE );
							m_fObjectSpaceAngVelocityAngle = 0.0f;
						}
						else
						{
							Asura_Quat		xOrientationWS;
							Asura_AVector3	xAnimVelocityWS;
							pxMovementSystem->GetCurrentMoveOrientation(xOrientationWS);
							xAnimVelocityWS = xOrientationWS.RotateVector( xAnimVelocityOS);
							m_pxPhysicsObjectHandler->CalculateMovementUsingOrientationAndLinVelocity( fTimeStep,
								xOrientationWS,
								xAnimVelocityWS,
								bUseGravity,
								AVP_PhysicsObjectHandler::WORLD_SPACE );
							m_fObjectSpaceAngVelocityAngle = 0.0f;
						}
					}
					break;
				}

				case Asura_Move::POSITION_SET_DIRECTLY:
				{
					Asura_AVector3	xPositionWS;

					pxMovementSystem->GetCurrentMovePosition(xPositionWS);
					if ( pxCurrentMove->GetOrientationUsage() == Asura_Move::ORIENTATION_NOT_USED )
					{
						Asura_Quat		xOriWS;
						xOriWS.FromMatrix( GetOrientation() );
						if ( !GetClientAnimAngularVelocityOS( *pxClient, m_fObjectSpaceAngVelocityAngle, m_xObjectSpaceAngVelocityAxis ) )
						{
							m_fObjectSpaceAngVelocityAngle = 0.0f;
							m_xObjectSpaceAngVelocityAxis = Asura_AVector3( 0.0f, 1.0f, 0.0f );
						}
						else
						{
							Asura_Quat		xRotationOS;
							float fPhysicsTimeSlice = Asura_Maths::FloatToInt_Chop( ( Asura_Physics_System::GetTimeLeft() + Asura_Timers::GetGameTimeSlice() ) / Asura_Physics_System::GetTimeSlice() ) * Asura_Physics_System::GetTimeSlice();
							xRotationOS.FromRotation( m_fObjectSpaceAngVelocityAngle * fPhysicsTimeSlice, m_xObjectSpaceAngVelocityAxis );
							xOriWS = xOriWS * xRotationOS; // multiply this way round as the rotation is in object space. (Quat rotation is Pout = q * Pin * q'
						}
						
						m_pxPhysicsObjectHandler->CalculateMovementUsingOrientationAndPosition( fTimeStep,
							xOriWS,
							xPositionWS,
							bUseGravity,
							AVP_PhysicsObjectHandler::WORLD_SPACE );
					}
					else
					{
						Asura_Quat		xOrientationWS;
						pxMovementSystem->GetCurrentMoveOrientation(xOrientationWS);
						m_pxPhysicsObjectHandler->CalculateMovementUsingOrientationAndPosition( fTimeStep,
							xOrientationWS,
							xPositionWS,
							bUseGravity,
							AVP_PhysicsObjectHandler::WORLD_SPACE );
					}
					break;
				}

				case Asura_Move::POSITION_SET_FROM_VELOCITY:
				{
					Asura_AVector3	xVelocityWS;
					pxMovementSystem->GetCurrentMoveVelocity(xVelocityWS);
					if ( pxCurrentMove->GetOrientationUsage() == Asura_Move::ORIENTATION_NOT_USED )
					{
						Asura_AVector3 xVelocityOS = Asura_AMatrix3x3( GetOrientation() ).ApplyTranspose( xVelocityWS );
						if ( !GetClientAnimAngularVelocityOS( *pxClient, m_fObjectSpaceAngVelocityAngle, m_xObjectSpaceAngVelocityAxis ) )
						{
							m_fObjectSpaceAngVelocityAngle = 0.0f;
							m_xObjectSpaceAngVelocityAxis = Asura_AVector3( 0.0f, 1.0f, 0.0f );
						}

						m_pxPhysicsObjectHandler->CalculateMovementUsingAngAndLinVelocity( fTimeStep,
							m_xObjectSpaceAngVelocityAxis,
							m_fObjectSpaceAngVelocityAngle,
							xVelocityOS,
							bUseGravity,
							AVP_PhysicsObjectHandler::OBJECT_SPACE );
						m_fObjectSpaceAngVelocityAngle = 0.0f;
					}
					else
					{
						Asura_Quat		xOrientationWS;
						pxMovementSystem->GetCurrentMoveOrientation(xOrientationWS);

						m_pxPhysicsObjectHandler->CalculateMovementUsingOrientationAndLinVelocity( fTimeStep,
							xOrientationWS,
							xVelocityWS,
							bUseGravity,
							AVP_PhysicsObjectHandler::WORLD_SPACE );
					}
					break;
				}

				default:
				{
					Asura_Assert(false, "Unexpected position update-method.");
					break;
				}
			}

			switch ( pxCurrentMove->GetAimDirectionUsage() )
			{
				case AVP_ClientEntity_Actor::AIM_DIRECTION_RESET:
				{
					const Asura_Matrix_3x3& xOrientWS = GetOrientation();
					const Asura_Vector_3	xForwardWS( xOrientWS.e13, xOrientWS.e23, xOrientWS.e33 );
					SetAimDirection( xForwardWS );
					break;
				}
				case AVP_ClientEntity_Actor::AIM_DIRECTION_SET_DIRECTLY:
				{
					const Asura_Vector_3 xForwardWS( pxCurrentMove->GetAimDirection() );
					SetAimDirection( xForwardWS );
					break;
				}
				default:
				{
					break;
				}
			}
		}
	}
}

///////////////////////////////////////////////////////////////////
// Parameter enumerators to handle the more complex flinch data. //
///////////////////////////////////////////////////////////////////

// Enumerator to fill the flinch types array.
AVP_Blueprint_ParameterEnumerator_FlinchTypes::AVP_Blueprint_ParameterEnumerator_FlinchTypes(AVP_Vertebrate_FlinchData* pxFlinchDataStorage, u_int& uNumFlinchEntries, const Asura_Hash_ID uMovesBlueprintType)
: m_pxFlinchDataStorage(pxFlinchDataStorage)
, m_uNumFlinchEntries(uNumFlinchEntries)
, m_uMovesBlueprintType(uMovesBlueprintType)
{
}

void AVP_Blueprint_ParameterEnumerator_FlinchTypes::BeginEnumeration()
{
	m_uNumFlinchEntries = 0;
}

bool AVP_Blueprint_ParameterEnumerator_FlinchTypes::Enumerate(const u_int uIndex, char const* const szName, const Asura_Value& xValue)
{
	if (m_uNumFlinchEntries >= uAVP_VERTEBRATE_MAX_FLINCH_TYPES)
	{
		Asura_Assert(false, "Too many flinch types for this weapon, ignoring the rest.");
		return false;
	}

	const Asura_Hash_ID uBlueprintID = xValue.GetHashIDValue();
	if (uBlueprintID == ASURA_HASH_ID_UNSET)
	{
		// For now, we'll silently ignore this, it probably means we don't have any flinch data.
		// Still return true though, to prevent the Asura-side asserting.
		return true;
	}

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(m_uMovesBlueprintType, xValue.GetHashIDValue());
	if (!pxBlueprint)
	{
		Asura_Assert(false, "Failed to find blueprint for flinch type.");
		// Allow the enumeration to continue anyway.
		return true;
	}

	Asura_Value xTypeValue;
	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_FIRSTPERSON, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_uFirstPersonBlueprint = xTypeValue.GetHashIDValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_uFirstPersonBlueprint = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_THIRDPERSON, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_uThirdPersonBlueprint = xTypeValue.GetHashIDValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_uThirdPersonBlueprint = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_MINATTACKSTRENGTH, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fMinAttackPower = xTypeValue.GetFloatValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fMinAttackPower = 0.0f;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_PROBABILITY, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fProbability = xTypeValue.GetFloatValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fProbability = 1.0f;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_MINDELAY, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fMinDelay = xTypeValue.GetFloatValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fMinDelay = 0.0f;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_MAXDELAY, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fMaxDelay = xTypeValue.GetFloatValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_fMaxDelay = 0.0f;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_USESMOVE, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bUseFlinchMove = xTypeValue.GetBoolValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bUseFlinchMove = false;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_DISABLESURFACEREORIENTATION, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bDisableSurfaceReorientation = xTypeValue.GetBoolValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bDisableSurfaceReorientation = false;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_RESETAIMDIRECTION, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bResetAimDirection = xTypeValue.GetBoolValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bResetAimDirection = false;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_FLINCH_ORIENTATETOIMPACTDIRECTION, xTypeValue))
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bOrientateToImpactDirection = xTypeValue.GetBoolValue();
	}
	else
	{
		m_pxFlinchDataStorage[m_uNumFlinchEntries].m_bOrientateToImpactDirection = false;
	}

	++m_uNumFlinchEntries;

	return true;
}

void AVP_Blueprint_ParameterEnumerator_FlinchTypes::EndEnumeration(const bool bSuccess)
{
	// No need to do anything special on failure, as it just means there were too many entries.
}


// Enumerator to pick a flinch variation. Can either pick a random one or a specified one (if there are enough parameters for the specified one).
AVP_Blueprint_ParameterEnumerator_FlinchSelector::AVP_Blueprint_ParameterEnumerator_FlinchSelector(int iSpecifiedVariation/* = -1*/)
: m_iNumFlinchVariations(0)
, m_iSelectedVariation(iSpecifiedVariation)
{
}

void AVP_Blueprint_ParameterEnumerator_FlinchSelector::BeginEnumeration()
{
	m_iNumFlinchVariations = 0;
}

bool AVP_Blueprint_ParameterEnumerator_FlinchSelector::Enumerate(const u_int uIndex, char const* const szName, const Asura_Value& xValue)
{
	if (m_iNumFlinchVariations >= MAX_FLINCH_VARIATIONS)
	{
		Asura_Assert(false, "Too many flinch variations for this type, ignoring the rest.");
		return false;
	}

	m_auFlinchAnims[m_iNumFlinchVariations++] = xValue.GetHashIDValue();
	return true;
}

void AVP_Blueprint_ParameterEnumerator_FlinchSelector::EndEnumeration(const bool bSuccess)
{
	// No need to do anything special on failure, as it just means there were too many entries.

	// If we don't have any flinches, store that we have failed to select one.
	if (m_iNumFlinchVariations <= 0)
	{
		m_iSelectedVariation = -1;
	}
	// If we weren't provided with a specific variation, pick one at random.
	else if (m_iSelectedVariation < 0)
	{
		m_iSelectedVariation = Asura_Random::GetInt(0, m_iNumFlinchVariations-1);
	}
	// If the specific variation is out of range, go back to the first.
	else if (m_iSelectedVariation >= m_iNumFlinchVariations)
	{
		m_iSelectedVariation = 0;
	}
	// If none of the above, we have a valid specified variation, and we can just use that.
}

Asura_Hash_ID AVP_Blueprint_ParameterEnumerator_FlinchSelector::GetSelectedFlinch() const
{
	if ((m_iSelectedVariation < 0) || (m_iSelectedVariation >= m_iNumFlinchVariations))
	{
		return ASURA_HASH_ID_UNSET;
	}
	else
	{
		return m_auFlinchAnims[m_iSelectedVariation];
	}
}

int AVP_Blueprint_ParameterEnumerator_FlinchSelector::GetSelectedVariation() const
{
	return m_iSelectedVariation;
}

/// End of file. ///
