//////////////////////////////////////////////////////////////////////////////
//
// AVP_ClientEntity_PhysicsObject.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "Entity/PhysicalObject/PhysicsObject/AVP_ClientEntity_PhysicsObject.h"

// Engine independent.
#ifndef ASURA_NO_NETWORK_LAG_COMPENSATION
#include "Entity/Asura_PosAndOriInterpolator.h"
#endif
#include "Animation/Asura_Hierarchy.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicsObject.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Network/Asura_Message.h"
#include "Physics/Asura_Physics_Object.h"
#include "Sound/Asura_Sound_Event_System.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "SpecialFX/Asura_Fragment_System.h"

// Project independent.
#include "Core/AVP_Settings.h"
#include "Core/AVP_SurfaceEffects.h"
#include "Entity/Damage/AVP_EntityDamage_Apply.h"
#include "Entity/PhysicalObject/PhysicsObject/AVP_PhysicsObject_Messages.h"
#include "Network/AVP_Client.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// AVP_ClientEntity_PhysicsObject
//

Asura_Hashed_List<AVP_ClientEntity_PhysicsObject> AVP_ClientEntity_PhysicsObject::s_xList( 5 );


AVP_ClientEntity_PhysicsObject::AVP_ClientEntity_PhysicsObject( void )
{
	Classification = AsuraEntityClass_PhysicsObject;

	s_xList.AddToList( this );
}


AVP_ClientEntity_PhysicsObject::~AVP_ClientEntity_PhysicsObject() 
{
	// Destroy the physics-object (not forgetting to NULL the pointer to prevent
	// parent classes from also attempting to destroy it)...
	Asura_Physics_Object::Destroy(m_pxDynamicObject);
	m_pxDynamicObject	= NULL;

	s_xList.RemoveFromList( this );
}


void AVP_ClientEntity_PhysicsObject::SetupDynamicObject( u_int uPhysicalObjectFlags )
{
    if( AVP_Settings::IsSinglePlayer() || Asura_Network::IsServer() )
    {
		// We only deal with the physics client-side for multiplayer
        // client-only machines.
		return;
	}

    ASURA_PHYSICS_OBJECT_TYPE eReqdType = ( uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_STATIC
                                                ? ASURA_PHYSICS_OBJECT_STATIC
                                                : ASURA_PHYSICS_OBJECT_EXTERNALCONTROL );

	if( m_pxDynamicObject )
	{
		// Already created. Note that we don't expect the static / non-static
        // property to change.
        Asura_Assert( m_pxDynamicObject->GetType() == eReqdType, "" );
		return;
	}

	if( uPhysicalObjectFlags & (ASURA_PHYSICALOBJECT_FLAG_NOCOLLISIONS | ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV) )
	{
		return;
	}

	const Asura_Shape_Instance& xShapeInstance = GetShapeInstance();
	const Asura_Shape* pxShape = xShapeInstance.GetShape();
	
	if( pxShape && pxShape->HasNoCollisions() )
	{
		// No collisions, so no need to create a dynamics block.
		return;
	}

	//////////////////////////////////////////////////////////////////////////
	// NOTE: Get the position and orientation BEFORE we create the
	// physics-object since once we have our physics-object, this is where
	// the position and orientation are fetched from...
	//////////////////////////////////////////////////////////////////////////
	const Asura_Vector_3&	xPosition		= GetCurrentPosition();
	const Asura_Matrix_3x3&	xOrientation	= GetCurrentOrientation();

	m_pxDynamicObject = Asura_Physics_Object::Create( eReqdType );
    if( !m_pxDynamicObject ) { return; }
	
	TeleportTo(xPosition, xOrientation);
	m_pxDynamicObject->SetAssociatedEntityGuid(GetGuid());
	
	// Find a suitable collision object.
	const Asura_Collision_Object* pxCollObj = NULL;
	if( pxShape )
	{
		pxCollObj = pxShape->GetDynamicsCollisionData();
	}
	else
	{
		const Asura_Animation* pxAnim = xShapeInstance.GetAnimation();
        if( pxAnim )
        {
		    const Asura_Hierarchy_Skin* pxSkin = pxAnim->GetSkin();
		    if( pxSkin )
		    {
			    pxCollObj = pxSkin->GetCollisionObject();
		    }
        }
	}

    // Otherwise use some default box.
	if( !pxCollObj )
	{
		const Asura_Hash_ID	uShapeHash	= GetShapeInstance().GetShapeHashID();

		pxCollObj = Asura_Dynamic_Extents::GetCollisionObject( 0, uShapeHash, GetGuid() );
	    Asura_Assert( pxCollObj, "No collision objects available?" );
	}

	m_pxDynamicObject->SetCollisionObject( pxCollObj );
}


void AVP_ClientEntity_PhysicsObject::Create( const Asura_Message* pxMsg )
{
	Asura_Assert( pxMsg->GetMsgID() == ASURA_MSG_CLIENTCREATE, "" );
	Asura_Assert( pxMsg->GetEntityClassification() == AsuraEntityClass_PhysicsObject, "" );

	AVP_ClientEntity_PhysicsObject* pxPhysicsObj = new AVP_ClientEntity_PhysicsObject();

#ifndef ASURA_NO_NETWORK_LAG_COMPENSATION
	const AVP_Message_PhysicsObject_ClientCreate* pxMsgData = static_cast<const AVP_Message_PhysicsObject_ClientCreate*>( pxMsg->GetData() );
	if( pxMsgData->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_STATIC )
    {
	    // (MH - 25.03.2004) these shouldn't ever move, so free their pos interpolator
	    delete pxPhysicsObj->m_pxPosAndOriInterpolator;
	    pxPhysicsObj->m_pxPosAndOriInterpolator = NULL;
    }
#endif

	pxPhysicsObj->HandleMessage( pxMsg );
	pxPhysicsObj->InitialiseShapePositionAndOrientation();

}


bool AVP_ClientEntity_PhysicsObject::HandleMessage( const Asura_Message* pxMsg )
{
	switch(pxMsg->GetMsgID())
	{
		case ASURA_MSG_CLIENTUPDATE:
		{
			const AVP_Message_PhysicsObject_ClientUpdate*	pxData	= static_cast<const AVP_Message_PhysicsObject_ClientUpdate*>(pxMsg->GetData());
			ProcessClientUpdate(pxData, pxMsg->GetSentTime());
			return true;
		}

		case ASURA_MSG_CLIENTCREATE:
		{
			const AVP_Message_PhysicsObject_ClientCreate*	pxData	= static_cast<const AVP_Message_PhysicsObject_ClientCreate*>(pxMsg->GetData());
			ProcessClientCreate(pxData, pxMsg->GetSentTime());
			return true;
		}

		default:
		{
			return( PARENT_ASURA::HandleMessage(pxMsg) || PARENT::HandleMessage(pxMsg) );
		}
	}
}


void AVP_ClientEntity_PhysicsObject::OnDeath( const AVP_EntityDamage_ApplyToEntity& xDamage )
{
	PARENT::OnDeath(xDamage);

    // Generate fragments and switch to destroyed shape as necessary.
    const Asura_Shape_Instance& xShapeInstance = GetShapeInstance();
    const Asura_Shape* pxShape = xShapeInstance.GetShape();
	if( pxShape )
	{
		Asura_Fragment_System::GenerateFragments( &xShapeInstance, NULL, 0.01f );

		Asura_Hash_ID uFragmentHash = pxShape->GetFragmentHashID();
		if( ASURA_HASH_ID_UNSET != uFragmentHash )
		{
			Asura_Hash_ID uShapeHash = Asura_Fragment_System::GetDestroyedShapeHash( uFragmentHash );
			if( ASURA_HASH_ID_UNSET != uShapeHash )
			{
				SetShapeID( uShapeHash, Asura_Timers::GetLocalScaledTime() );

                // Don't set as destroyed or whatever. We need our destroyed
                // version to hang around, at least for a while.
				return;
			}
		}
	}

	SetAsDestroyed();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ClientEntity_PhysicsObject::OnApplyDamage
// Author       : Andy Tate
// Date         : 17-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_ClientEntity_PhysicsObject::OnApplyDamage(const AVP_EntityDamage_ApplyToEntity& xDamage)
{
	PARENT::OnApplyDamage(xDamage);

	const Asura_Hash_ID uMaterialResponseHashID = xDamage.GetMaterialResponseHashID();
	const Asura_MaterialResponse* const pxResponse = Asura_MaterialResponse_System::GetMaterialByHashID(uMaterialResponseHashID, true);

	Asura_Hash_ID uMaterialType = ASURA_HASH_ID_UNSET;
	if (pxResponse)
	{
		uMaterialType = pxResponse->GetSurfaceTypeHash();
	}

	if (uMaterialType != ASURA_HASH_ID_UNSET)
	{
		AVP_SurfaceEffect_System::PlayDamageSurfaceEffect(uMaterialType, xDamage, GetGuid());
	}
}

/// End of function 'AVP_ClientEntity_PhysicsObject::OnApplyDamage'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_ClientEntity_PhysicsObject::ProcessClientUpdate
// Author       : Andy Tate
// Date         : 13-Sep-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_ClientEntity_PhysicsObject::ProcessClientUpdate(const AVP_Message_PhysicsObject_ClientUpdate* pxData,
														 float fServerTime,
														 bool bProcessPosAndOri)
{
	PARENT::ProcessClientUpdate(pxData, fServerTime, bProcessPosAndOri);
}

/// End of function 'AVP_ClientEntity_PhysicsObject::ProcessClientUpdate'. ///

void AVP_ClientEntity_PhysicsObject::ProcessClientCreate(const AVP_Message_PhysicsObject_ClientCreate* pxData,
														 float fServerTime,
														 bool bProcessPosAndOri)
{
	PARENT::ProcessClientCreate(pxData, fServerTime, bProcessPosAndOri);

	SetupDynamicObject( pxData->m_uPhysicalObjectFlags );
}

// end of file

