
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_ServerProjectile_AcidSpit.cpp
// Author : Andy Tate
// Date   : 27-Oct-2006
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Entity/PhysicalObject/Projectile/AVP_ServerProjectile_AcidSpit.h"

// Asura independent header-files...
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"

// Game independent header-files...
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Damage/AVP_AcidDamage_Messages.h"
#include "Entity/PhysicalObject/Projectile/AVP_Projectile_Messages.h"
#include "Entity/PhysicalObject/Projectile/AVP_GenericProjectileData.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float dbg_fDoStuffSpeedSq = 100.0f;

static const u_int	g_uChunkVersion	= 0;

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

Asura_Hashed_List<AVP_ServerProjectile_AcidSpit>	AVP_ServerProjectile_AcidSpit::s_xList;

AVP_ServerProjectile_AcidSpit::AVP_ServerProjectile_AcidSpit( const Asura_Guid uGuid, const Asura_Message* pxMsg )
: Asura_Entity_Server( uGuid ) 
, PARENT( uGuid, pxMsg )
{
	Classification	= AVP_ENTITYCLASS_PROJECTILE_ACIDSPIT;
	
	// Don't forget to store us in the list...
	s_xList.AddToList( this );
}

AVP_ServerProjectile_AcidSpit::AVP_ServerProjectile_AcidSpit( const Asura_Guid uGuid )
: Asura_Entity_Server( uGuid ) 
, PARENT( uGuid )
{
	Classification	= AVP_ENTITYCLASS_PROJECTILE_ACIDSPIT;

	// Don't forget to store us in the list...
	s_xList.AddToList( this );
}

AVP_ServerProjectile_AcidSpit::~AVP_ServerProjectile_AcidSpit()
{
	s_xList.RemoveFromList( this );
}

AVP_ServerProjectile_AcidSpit*	AVP_ServerProjectile_AcidSpit::Create( const Asura_Message* pxMsg )
{
	AVP_ServerProjectile_AcidSpit*	pxServer = new AVP_ServerProjectile_AcidSpit( Asura_Entity_System::GetCreatedByGameGuid(), pxMsg );

	pxServer->InitFromBlueprint();
	const AVP_GenericProjectileData* pxProjectileData = AVP_GenericProjectileData::GetDataByHash( pxServer->GetBlueprintHashID() );
	const AVP_Message_CreateGenericProjectile* pxData = static_cast< const AVP_Message_CreateGenericProjectile* >( pxMsg != NULL ? pxMsg->GetData() : NULL );
	if( pxData != NULL )
	{
		const float				fMuzzleVel = pxProjectileData->GetMuzzleVelocity();
		Asura_Vector_3			xVelocity( 0.0f, 0.0f, fMuzzleVel );
		const Asura_Vector_3&	xPosition = pxData->GetPosition();
		Asura_Matrix_3x3		xOrientation;
		pxData->GetOrientation().ToMatrix( xOrientation );
		xVelocity = xVelocity * xOrientation;
		float fMag = xVelocity.Normalise();
		fMag += xVelocity * pxData->GetActorVelocity();
		if ( pxData->GetLaunch() )
		{
			pxServer->Launch( xPosition, xOrientation, xVelocity * fMag, pxServer->m_uOwnerGuid, pxData->GetDamageScale() ); // Launch last as this creates the client entity
		}
		else
		{
			pxServer->LinkToParent( pxServer->m_uOwnerGuid, pxProjectileData->GetLinkMarkerPoint() );
		}
	}

    return pxServer;
}

void AVP_ServerProjectile_AcidSpit::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int	uVersion;

	// Read the version...
	xStream >> uVersion;
	Asura_Assert( ( uVersion <= g_uChunkVersion ), "Unsupported chunk version");

	// Read the parent data...
	PARENT::ReadFromChunkStream( xStream );
}

void AVP_ServerProjectile_AcidSpit::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	// Write the version...
	xStream << g_uChunkVersion;

	// Write the parent data...
	PARENT::WriteToChunkStream( xStream );
}

void AVP_ServerProjectile_AcidSpit::InitialiseFromBlueprint( const Asura_Blueprint& xBlueprint )
{
	PARENT::InitialiseFromBlueprint( xBlueprint );
}

void AVP_ServerProjectile_AcidSpit::Update( void )
{
	PARENT::Update();
}

void AVP_ServerProjectile_AcidSpit::OnCollision( const Asura_LOS_InputData& xLOSInputData, const Asura_LOS_OutputData& xLOSOutputData )
{
	const Asura_Guid uVictimGuid = xLOSOutputData.m_uHitEntity;

	if( uVictimGuid == 0 )
	{
		Asura_Assert( xLOSOutputData.m_bHitEnvironment, "This shold be true" );
		// Hit environment;
		const Asura_Vector_3 xPosition(xLOSInputData.m_xOrigin + ( xLOSInputData.m_xDirection * xLOSOutputData.m_fDistanceToIntersection ) );
		OnImpact( xPosition, xLOSOutputData.m_xNormal );
		PlaceEnvironmentalAcidDamage( xPosition, xLOSOutputData.m_xNormal );
	}

	PARENT::OnCollision( xLOSInputData, xLOSOutputData );

	SetThisAndClientsAsDestroyed();
}

void AVP_ServerProjectile_AcidSpit::MakeClientUpdateMessage( void )
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_CLIENTUPDATE, sizeof( AVP_Message_GenericProjectileClientUpdate ), GetClassification());

	if( pxMessage != NULL )
	{
		// Fill in the message...
		AVP_Message_GenericProjectileClientUpdate* pxData = static_cast<AVP_Message_GenericProjectileClientUpdate*>( pxMessage->GetData() );
		pxData->FillGenericProjectile( this );

		// Send the message...
		Send( pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES );
	}
}

void AVP_ServerProjectile_AcidSpit::PlaceEnvironmentalAcidDamage( const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal )
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage( AVP_MSGID_CREATE_ACID_DAMAGE, sizeof( AVP_Message_AcidDamage_Create ) );

	if(pxMessage != NULL)
	{
		AVP_Message_AcidDamage_Create* pxData = static_cast<AVP_Message_AcidDamage_Create*>(pxMessage->GetData());

		if(pxData != NULL)
		{
			// CAH we need to set the source guid as the queen not the projectile for two reasons
			// 1 because by the time the acid damage is placed the projectile is destroyed so wont be found
			// 2 because we need to reference to queens blueprints for acid damage proporties. 
			pxData->SetSourceGuid( m_uOwnerGuid );
			pxData->SetPosition( xPosition );
			pxData->SetSurfaceNormal( xNormal );
			pxData->SetIsTrailMode( false );
		}

		Send(pxMessage, ASURA_GUID_SERVER);
	}

}
