
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_ServerProjectile_DropshipMissile.cpp
// Author : Chris Wallace
// Date   : 04-Dec-2007
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Entity/PhysicalObject/Projectile/AVP_ServerProjectile_DropshipMissile.h"

// Asura independent header-files...
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_Timers.h"
#include "Network/Asura_Network.h"
#include "SpecialFX/pfx/Asura_PFX_Entity_Effect.h"
#include "SpecialFX/pfx/Asura_PFX_System.h"

// Game independent header-files...
#include "Actor/Vertebrates/Predator/AVP_Predator_Description.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/AVP_Messages.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_Actor_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vehicles/Dropship/AVP_ServerVehicle_Dropship.h"
#include "Entity/PhysicalObject/Projectile/AVP_Projectile_Messages.h"
#include "Entity/PhysicalObject/Projectile/AVP_GenericProjectileData.h"
#include "Inventory/Items/Weapons/AVP_InventoryWeapon_DropshipMissile.h"
#include "SpecialFX/Asura_Decal_System.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const u_int	g_uChunkVersion	= 0;


Asura_Hashed_List<AVP_ServerProjectile_DropshipMissile>	AVP_ServerProjectile_DropshipMissile::s_xList;

AVP_ServerProjectile_DropshipMissile::AVP_ServerProjectile_DropshipMissile (const Asura_Guid uGuid, const Asura_Message* pxMsg)
: Asura_Entity_Server( uGuid )
, PARENT( uGuid, pxMsg )
, m_ePursuitMode(BLIND_START)
, m_fTimeSinceLaunch(0.0f)
, m_xStartPosition(Asura_Maths::ZeroVector3)
, m_fMaxVelocity(0.0f)
{
	// Update the data...
	Classification	= AVP_ENTITYCLASS_DROPSHIP_MISSILE;

	// Set the stats...
	SetHealth(100.0f);
	SetDestructible(false);

	AVP_ServerVehicle_Dropship* pxServerOwner = AVP_ServerVehicle_Dropship::FindEntity(m_uOwnerGuid);
	bool bOwnerAlive = pxServerOwner && !pxServerOwner->IsDead() && !pxServerOwner->IsDestroyed();

	if (bOwnerAlive)
	{
		pxServerOwner->MissileLaunched();
	}


	// Don't forget to store us in the list...
	s_xList.AddToList(this);
}

AVP_ServerProjectile_DropshipMissile::~AVP_ServerProjectile_DropshipMissile ()
{
	AVP_ServerVehicle_Dropship* pxServerOwner = AVP_ServerVehicle_Dropship::FindEntity(m_uOwnerGuid);
	bool bOwnerAlive = pxServerOwner && !pxServerOwner->IsDead() && !pxServerOwner->IsDestroyed();
	if (bOwnerAlive)
	{
		pxServerOwner->MissileDestroyed();
	}

	// Don't forget to remove us from the list...
	s_xList.RemoveFromList(this);
}

AVP_ServerProjectile_DropshipMissile* AVP_ServerProjectile_DropshipMissile::Create( const Asura_Message* pxMsg )
{
	AVP_ServerProjectile_DropshipMissile* pxDropshipMissile = new AVP_ServerProjectile_DropshipMissile( Asura_Entity_System::GetCreatedByGameGuid(), pxMsg);
    pxDropshipMissile->InitFromBlueprint();
	const AVP_GenericProjectileData* pxProjectileData = AVP_GenericProjectileData::GetDataByHash( pxDropshipMissile->GetBlueprintHashID() );
	const AVP_Message_CreateGenericProjectile* pxData = static_cast< const AVP_Message_CreateGenericProjectile* >( pxMsg != NULL ? pxMsg->GetData() : NULL );
	if( pxData != NULL )
	{
		const float fMuzzleVel = pxProjectileData->GetMuzzleVelocity();
		Asura_Vector_3 xVelocity( 0.0f, 0.0f, fMuzzleVel );
		const Asura_Vector_3& xPosition = pxData->GetPosition();
		Asura_Matrix_3x3 xOrientation;
		pxData->GetOrientation().ToMatrix( xOrientation );
		xVelocity = xVelocity * xOrientation;
		float fMuzVelMag = xVelocity.Normalise();
		float fActorVelMag = xVelocity * pxData->GetActorVelocity();

		if (fActorVelMag <= 0.0f)
		{
			fActorVelMag = 0.001f;
		}

		pxDropshipMissile->Launch( xPosition, xOrientation, xVelocity * fActorVelMag, pxDropshipMissile->m_uOwnerGuid, pxData->GetDamageScale() ); // Launch last as this creates the client entity
		pxDropshipMissile->SetMaxVelocity(fMuzVelMag + fActorVelMag);
	}

    return pxDropshipMissile;
}

void AVP_ServerProjectile_DropshipMissile::MakeClientUpdateMessage (void)
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_CLIENTUPDATE, sizeof(AVP_Message_GenericProjectileClientUpdate), GetClassification());
	
	if(pxMessage != NULL)
	{
		// Fill in the message...
		AVP_Message_GenericProjectileClientUpdate* pxData = static_cast<AVP_Message_GenericProjectileClientUpdate*>(pxMessage->GetData());
		pxData->FillGenericProjectile(this);

		// Send the message...
		Send(pxMessage, ASURA_GUID_MY_CLIENT_ENTITIES);
	}
}

void AVP_ServerProjectile_DropshipMissile::ReadFromChunkStream (Asura_Chunk_Stream& xStream)
{
	u_int	uVersion;

	// Read the version...
	xStream >> uVersion;
	Asura_Assert((uVersion <= g_uChunkVersion), "Unsupported chunk version");

	// Read the parent data...
	PARENT::ReadFromChunkStream(xStream);
}

void AVP_ServerProjectile_DropshipMissile::WriteToChunkStream (Asura_Chunk_Stream& xStream) const
{
	// Write the version...
	xStream << g_uChunkVersion;

	// Write the parent data...
	PARENT::WriteToChunkStream(xStream);
}
DEBUGVAR float dbg_fAcceleration = 40.0f;
DEBUGVAR float dbg_fMaxFlightTime = 10.0f;
DEBUGVAR float dbg_fStartBlindFlightDistanceSq = 10.0f * 10.0f;
DEBUGVAR float dbg_fClimbDistanceSq = 15.0f * 15.0f;
DEBUGVAR float dbg_fEndBlindFlightDistanceSq = 5.0f * 5.0f;
DEBUGVAR float dbg_fMaxAngularVelocity = Asura_Maths::Pi * 360.0f / 180.0f;
DEBUGVAR float dbg_fClimbAngularVelocity = Asura_Maths::Pi * 60.0f / 180.0f;
DEBUGVAR float dbg_fClimbAngle = Asura_Maths::Pi * 45.0f / 180.0f;

void AVP_ServerProjectile_DropshipMissile::Update( void )
{
	// Call the base-function...
	PARENT::Update();

	float fTimeStep = Asura_Timers::GetGameTimeSlice();
	

	m_fTimeSinceLaunch += fTimeStep;
	
	// limit the life time of the missile
	if (m_fTimeSinceLaunch > dbg_fMaxFlightTime)
	{
		Explode();
		return;
	}

	// Accelerate
	AVP_ServerVehicle_Dropship* pxServerOwner = AVP_ServerVehicle_Dropship::FindEntity(m_uOwnerGuid);
	bool bOwnerAlive = pxServerOwner && !pxServerOwner->IsDead() && !pxServerOwner->IsDestroyed();
	Asura_Matrix_3x3 xOrientation = m_pxDynamicObject->GetOrientation();
	Asura_Vector_3 xCurrentDirection(xOrientation.e13, xOrientation.e23, xOrientation.e33);
	float fVelocitySize = m_pxDynamicObject->GetLinVelocity().Magnitude();
	if (fVelocitySize < m_fMaxVelocity)
	{
		fVelocitySize += fTimeStep * dbg_fAcceleration;
		if (fVelocitySize > m_fMaxVelocity)
		{ 
			fVelocitySize = m_fMaxVelocity;
		}
		m_pxDynamicObject->SetLinVelocity(xCurrentDirection * fVelocitySize);
	}

	switch (m_ePursuitMode)
	{
	case BLIND_START: // let it fly in its direction until it far enough from its owner
		if (bOwnerAlive)
		{
			if ( (m_xStartPosition - m_pxDynamicObject->GetPosition()).MagnitudeSquared() > dbg_fStartBlindFlightDistanceSq)
			{
				m_ePursuitMode = CLIMB;
			}
		}
		break;
	case CLIMB:
		if(bOwnerAlive)
		{
			if ( (m_xStartPosition - m_pxDynamicObject->GetPosition()).MagnitudeSquared() > dbg_fClimbDistanceSq)
			{
				m_ePursuitMode = PURSUIT;
			}
			else
			{
				Asura_Vector_3 xNewDirection;
				Asura_Quat xTurnQuat;
				Asura_Vector_3 xDesiredDirection(xCurrentDirection);
				xDesiredDirection.y = 0.0f;
				xDesiredDirection.Normalise();
				Asura_Vector_3 xRotateVector(xOrientation.e11, xOrientation.e21,  xOrientation.e31);
				xTurnQuat.FromRotation(dbg_fClimbAngle, xRotateVector);
				xDesiredDirection = xTurnQuat.RotateVector(xDesiredDirection);
				if (xCurrentDirection * xDesiredDirection > Asura_Maths::Cosine(dbg_fClimbAngularVelocity * fTimeStep))
				{
					xTurnQuat.RotationFromAToB(xCurrentDirection, xDesiredDirection);
				}
				else
				{
					Asura_Vector_3 xTurnDirection = xCurrentDirection ^ xDesiredDirection;
					xTurnQuat.FromRotation(dbg_fClimbAngularVelocity * fTimeStep, xTurnDirection);
				}
				xNewDirection = xTurnQuat.RotateVector(xCurrentDirection);
				m_pxDynamicObject->SetLinVelocity(xNewDirection * fVelocitySize);
				Asura_Matrix_3x3 xTurnMatrix;
				xTurnQuat.ToMatrix(xTurnMatrix);
				m_pxDynamicObject->SetOrientation( xTurnMatrix * m_pxDynamicObject->GetOrientation());
			}
		}
		break;
	case PURSUIT:
		if(bOwnerAlive)
		{
			Asura_Vector_3 xTarget;
			if(pxServerOwner->GetLaserTarget(xTarget))
			{
				Asura_Vector_3 xNewDirection;
				Asura_Vector_3 xDesiredDirection = xTarget - m_pxDynamicObject->GetPosition();
				float fDistance = xDesiredDirection.Normalise();
				if (fDistance*fDistance > dbg_fEndBlindFlightDistanceSq ) 
				{
					Asura_Quat xTurnQuat;
					if (xCurrentDirection * xDesiredDirection > Asura_Maths::Cosine(dbg_fMaxAngularVelocity * fTimeStep))
					{
						xTurnQuat.RotationFromAToB(xCurrentDirection, xDesiredDirection);
					}
					else
					{
						Asura_Vector_3 xTurnDirection = xCurrentDirection ^ xDesiredDirection;
						xTurnQuat.FromRotation(dbg_fMaxAngularVelocity * fTimeStep, xTurnDirection);
					}
					xNewDirection = xTurnQuat.RotateVector(xCurrentDirection);
					m_pxDynamicObject->SetLinVelocity(xNewDirection * fVelocitySize);
					Asura_Matrix_3x3 xTurnMatrix;
					xTurnQuat.ToMatrix(xTurnMatrix);
					m_pxDynamicObject->SetOrientation( xTurnMatrix * m_pxDynamicObject->GetOrientation());
				}
				else
				{
					m_ePursuitMode = BLIND_END;
				}
			}
		}
		break;
	case BLIND_END:
		break;
		// do nothing wait for the bang
	}
}

void AVP_ServerProjectile_DropshipMissile::Launch(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation, const Asura_Vector_3& xVelocity, const Asura_Guid uGuidToIgnore, float fDamageMultiplier)
{
	m_xStartPosition = xPosition;
	PARENT::Launch(xPosition, xOrientation, xVelocity, uGuidToIgnore, fDamageMultiplier);
}


/// End of file. ///
