//////////////////////////////////////////////////////////////////////////////
//
// AVP_ServerEntity_PlayerProxy.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Entity/PlayerProxy/AVP_ServerEntity_PlayerProxy.h"

// asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Memory/Asura_Mem.h"

// avp
#include "Entity/AVP_Entity_EventIDs.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ServerActor_Predator.h"
#include "Entity/PlayerInfo/AVP_ServerEntity_PlayerInfo.h"
#include "Inventory/AVP_AmmoStore.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/AVP_InventoryItem_SubType.h"

//////////////////////////////////////////////////////////////////////////////
//
// AVP_ServerEntity_PlayerProxy
//
//////////////////////////////////////////////////////////////////////////////

Asura_Hashed_List<AVP_ServerEntity_PlayerProxy> AVP_ServerEntity_PlayerProxy::s_xList;

AVP_ServerEntity_PlayerProxy::AVP_ServerEntity_PlayerProxy( Asura_Guid uGuid )
: PARENT( uGuid )
, m_uPlayerNumber(1)
, m_uFlagsNotChunked(0)
{
	Classification = AVP_ENTITYCLASS_PLAYER_PROXY;
	s_xList.AddToList(this);
}

AVP_ServerEntity_PlayerProxy::~AVP_ServerEntity_PlayerProxy()
{
	s_xList.RemoveFromList(this);
}

void AVP_ServerEntity_PlayerProxy::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
    int iVersion;
    xStream >> iVersion;

    PARENT::ReadFromChunkStream( xStream );

    if ( iVersion == 0 )
    {
		xStream >> m_uPlayerNumber;
    }
    else
    {
        Asura_Assertf( false, ( "AVP_ServerEntity_PlayerProxy::ReadFromChunkStream - unhandled chunk version (%d)", iVersion ) );
    }
}

void AVP_ServerEntity_PlayerProxy::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
    xStream << CURRENT_CHUNK_VERSION;

    PARENT::WriteToChunkStream( xStream );    

	xStream << m_uPlayerNumber;

}

void AVP_ServerEntity_PlayerProxy::PostLoadInitialise()
{
	PARENT::PostLoadInitialise();

	m_uFlagsNotChunked = 0;

}

AVP_ServerEntity_PlayerProxy* AVP_ServerEntity_PlayerProxy::Create( const Asura_Guid uGuid )
{
	   return new AVP_ServerEntity_PlayerProxy( uGuid );
}


void AVP_ServerEntity_PlayerProxy::Update()
{
	PARENT::Update();

	if (!IsEntityActive())
	{
		return;
	}

	const AVP_ServerEntity_PlayerInfo* pxPlayerInfo = AVP_ServerEntity_PlayerInfo::FindEntity(ASURA_GUID_CLIENT_START + m_uPlayerNumber - 1);
	const AVP_ServerEntity_Actor* pxPlayerActor = pxPlayerInfo ? AVP_ServerEntity_Actor::FindEntity(pxPlayerInfo->GetPuppetGuid()) : NULL;
	if (!pxPlayerActor)
	{
		return;
	}

	bool bCurrentWeaponLowAmmo;
	bool bBothWeaponLowAmmo;
	bool bPrimaryWeaponNoAmmo;

	GetAmmoStatus(pxPlayerActor, bCurrentWeaponLowAmmo, bBothWeaponLowAmmo, bPrimaryWeaponNoAmmo);

	if (!m_bCurrentWeponLowAmmo && static_cast<bool>(m_bCurrentWeponLowAmmo) != bCurrentWeaponLowAmmo)
	{
		SendMessagesForEvent(AVP_ENTITY_EVENT_PLAYER_LOW_AMMO_CURRENT_WEAPON);
	}

	if (!m_bBothWeaponsLowAmmo && static_cast<bool>(m_bBothWeaponsLowAmmo) != bBothWeaponLowAmmo)
	{
		SendMessagesForEvent(AVP_ENTITY_EVENT_PLAYER_LOW_AMMO_BOTH_WEAPONS);
	}

	if (!m_bPrimaryWeaponNoAmmo && static_cast<bool>(m_bPrimaryWeaponNoAmmo) != bPrimaryWeaponNoAmmo)
	{
		SendMessagesForEvent(AVP_ENTITY_EVENT_PLAYER_LOW_AMMO_BOTH_WEAPONS);
	}

	m_bCurrentWeponLowAmmo = bCurrentWeaponLowAmmo;
	m_bBothWeaponsLowAmmo = bBothWeaponLowAmmo;
	m_bPrimaryWeaponNoAmmo = bPrimaryWeaponNoAmmo;

	// For predator only
	if ( pxPlayerActor->GetClassification() == AVP_ENTITYCLASS_PREDATOR )
	{
		bool bCloaked = ( static_cast< const AVP_ServerActor_Predator*> ( pxPlayerActor ) )->IsCloaked();
		if (static_cast<bool>(m_bCloaked) != bCloaked)
		{
			if ( m_bCloaked )
			{
				SendMessagesForEvent(AVP_ENTITY_EVENT_PLAYER_DECLOAKED);
			}
			else
			{
				SendMessagesForEvent(AVP_ENTITY_EVENT_PLAYER_CLOAKED);
			}
		}
		m_bCloaked = bCloaked;
	}

	// Send death message
	if ( pxPlayerActor->IsDead() )
	{
		SendMessagesForEvent( AVP_ENTITY_EVENT_PLAYER_ON_DEATH );
	}
}

void AVP_ServerEntity_PlayerProxy::OnEntityActivate()
{
	PARENT::OnEntityActivate();
	m_uFlagsNotChunked = 0;
}

void AVP_ServerEntity_PlayerProxy::OnEntityDeactivate()
{
	PARENT::OnEntityDeactivate();
}

void AVP_ServerEntity_PlayerProxy::GetAmmoStatus(const AVP_ServerEntity_Actor* pxPlayerActor, bool& bCurrentWeaponLowAmmo, bool& bBothWeaponLowAmmo, bool& bPrimaryWeaponNoAmmo)
{
	bCurrentWeaponLowAmmo = false;
	bBothWeaponLowAmmo = false;
	bPrimaryWeaponNoAmmo = false;

	bool bPrimaryLow = false;
	bool bSecondaryLow = false;

	const AVP_Inventory_System* pxInventorySystem = pxPlayerActor->GetInventorySystem();
	AVP_Inventory_Group* pxWeaponGroup = pxInventorySystem ? pxInventorySystem->GetGroup( uAVP_INVENTORY_GROUPHASH_WEAPON ) : NULL;

	Asura_Assert(pxWeaponGroup, "Player weapon group not found");
	AVP_Inventory_Item* pxCurrentWeapon = pxWeaponGroup ? pxWeaponGroup->GetCurrentInventoryItem() : NULL;
	AVP_Inventory_Item* pxPrimaryWeapon = pxWeaponGroup ? pxWeaponGroup->GetItem(uAVP_INVENTORY_WEAPONHASH_WEAPON02) : NULL;
	AVP_Inventory_Item* pxSecondaryWeapon = pxWeaponGroup ? pxWeaponGroup->GetItem(uAVP_INVENTORY_WEAPONHASH_WEAPON03) : NULL;

	if (!pxCurrentWeapon) // no current weapon weapon
	{
		bCurrentWeaponLowAmmo = false;
	}
	else if (pxCurrentWeapon->GetNumSubTypes() == 0 || 
		pxCurrentWeapon->GetSubType(0)->GetAmmoType() == ASURA_HASH_ID_UNSET) // has primary weapon but no ammo
	{
		bCurrentWeaponLowAmmo = false;
	}
	else
	{
		bCurrentWeaponLowAmmo = pxInventorySystem->GetAmmoStore()->GetAmmoQuantity(pxCurrentWeapon->GetSubType(0)->GetAmmoType()) == 0;		
	}


	if (!pxPrimaryWeapon) // no primary weapon
	{
		bPrimaryLow = true;
		bPrimaryWeaponNoAmmo = true;
	}
	else if (pxPrimaryWeapon->GetNumSubTypes() == 0 || 
			 pxPrimaryWeapon->GetSubType(0)->GetAmmoType() == ASURA_HASH_ID_UNSET) // has primary weapon but no ammo
	{
		bPrimaryLow = false;
		bPrimaryWeaponNoAmmo = false;
	}
	else
	{
		bPrimaryLow = !(pxInventorySystem->GetAmmoStore()->GetAmmoQuantity(pxPrimaryWeapon->GetSubType(0)->GetAmmoType()) > 0.0f);
		bPrimaryWeaponNoAmmo = 
			(pxInventorySystem->GetAmmoStore()->GetAmmoQuantity(pxPrimaryWeapon->GetSubType(0)->GetAmmoType()) +
			pxPrimaryWeapon->GetSubType(0)->GetCurrentQuantity()) <= 0.0001f ;
	}

	if (!pxSecondaryWeapon) // no primary weapon
	{
		bSecondaryLow = true;
	}
	else if (pxSecondaryWeapon->GetNumSubTypes() == 0 || 
		pxSecondaryWeapon->GetSubType(0)->GetAmmoType() == ASURA_HASH_ID_UNSET) // has primary weapon but no ammo
	{
		bSecondaryLow = false;
	}
	else
	{
		bSecondaryLow = !(pxInventorySystem->GetAmmoStore()->GetAmmoQuantity(pxSecondaryWeapon->GetSubType(0)->GetAmmoType()) > 0.0f);
	}

	bBothWeaponLowAmmo = (bPrimaryLow  && bSecondaryLow );
}


//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
