//////////////////////////////////////////////////////////////////////////////
//
// AVP_GameLobby.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(AVP_USE_GAME_LOBBY)

// This file's header.
#include "FrontEnd/AVP_GameLobby.h"

// Asura includes
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_HashedLocalisedText.h"
#include "Core/Resources/Asura_Resource_System.h"
#include "Fileaccess/Asura_FileAccess.h"
#include "GUIMenu/Asura_GUIMenu_System.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_NetlinkManager.h"
#include "Network/Asura_NetTransport_System.h"
#include "Network/Online/Asura_Lobby.h"
#include "Network/Online/Asura_Matchmaking_System.h"

// AVP includes
#include "Core/AVP_Core.h"
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_GameStatus.h"
#include "Core/AVP_Settings.h"
#include "Core/AVP_StateMachine_Core.h"
#include "GUIMenu/AVP_GUIMenu_System.h"
#include "Multiplayer/AVP_Multiplayer_GameRules.h"
#include "Network/AVP_Client.h"
#include "Network/AVP_Server.h"
#include "Network/AVP_VoiceChannels.h"

// C O N S T A N T S /////////////////////////////////////////////////////////
static const int iAVP_LOBBY_INVALID_SEARCH_FILTER = -1;
//static const u_int uAVP_MAX_TEAM_MEMBERS = 6;

// check num attributes is correct
ASURA_STATIC_ASSERT(AVP_NUM_LOBBY_ATTRIBUTES == ASURA_LOBBY_NUM_INT_ATTRIBUTES);

// G L O B A L   D A T A /////////////////////////////////////////////////////
static const char* szMULTIPLAYER_MAP_PATH = "Envs\\MP\\";

static bool g_bAllowRankedMatch = false;

// L O C A L   F U N C T I O N S /////////////////////////////////////////////
static Asura_Hash_ID GetSpeciesNameTextHashID(AVP_Settings::Species eSpecies)
{
	switch(eSpecies)
	{
		case AVP_Settings::AVP_SPECIES_PREDATOR:
			return 0x593E169A;	// Asura_GetHashID("Species_Predator");
		case AVP_Settings::AVP_SPECIES_ALIEN:
			return 0xF0305E04;	// Asura_GetHashID("Species_Alien");
		case AVP_Settings::AVP_SPECIES_HUMAN:
			return 0x29BEAE25;	// Asura_GetHashID("Species_Marine");
		default:
			return ASURA_HASH_ID_UNSET;
	}
}

// S T A T I C   C L A S S   D A T A /////////////////////////////////////////
bool AVP_GameLobby::s_bSetupComplete;
bool AVP_GameLobby::s_bQuitRequested;
AVP_GameLobby_GameSettings AVP_GameLobby::s_xLobbySettings;
int AVP_GameLobby::s_aiLobbySearchFilters[AVP_NUM_LOBBY_ATTRIBUTES];
bool AVP_GameLobby::s_bGametypeChanged;
int AVP_GameLobby::s_iPrevGametype;
Asura_LinkList<AVP_GameLobby_MapData*> AVP_GameLobby::s_xMapList;

// C L A S S   I M P L E M E N T A T I O N ///////////////////////////////////
void AVP_GameLobby::Initialise()
{
	// set online game to true. On PC this may be set to false by 
	// Platform_Initialise() if using LAN.
	AVP_Settings::SetOnlineGame(true);
	Platform_Initialise();

	s_bSetupComplete = false;
	s_bQuitRequested = false;
	
	const bool bLobbySystemWasInitialised = Asura_Lobby_System::IsInitialised();
	if (!bLobbySystemWasInitialised)
	{
		Asura_Lobby_System::Initialise();
		// JG - 17 December 2008 - Don't allow joining games in progress yet. It needs testing.
		Asura_Lobby_System::SetAllowJoinInProgress( false ); 

		// No need to repeatedly create the map list, just do it the first time we go into multiplayer.
		CreateMapList();
	}

	for (u_int u = 0; u < AVP_NUM_LOBBY_ATTRIBUTES; ++u)
	{
		s_aiLobbySearchFilters[u] = iAVP_LOBBY_INVALID_SEARCH_FILTER;
	}

	s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP] = s_xMapList.Size(); // Any map is signified by the limit value instead, since the menu elements don't support a -1 value.
	s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAME_VERSION] = Asura_NetTransport_System::Project_GetVersionID();

	// first change the species to make sure that they are in a team with space ( done in ChangeSpecies for the moment )
	ChangeSpecies( false );

	// open first page: either post game page or top lobby menu
	if (bLobbySystemWasInitialised)
	{
		Asura_Assert((Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_ENDING_GAME) || (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_INVALID), "unexpected lobby state");
		Asura_GUIMenu_System::OpenMenuPage( 0x74ED574D /* Asura_GetHashID("GameLobby_PostGame") */ );
	}
	else
	{
		if( ! Asura_GUIMenu_System::OpenMenuPage( 0xFE4DE780 /* Asura_GetHashID("Multiplayer_Main_Menu") */ ) )
		{
			Asura_Assert( false, "AVP_GameLobby::Initialise\n Could not open the multiplayer main menu. " );
			RequestQuit( false );
		}
	}
}

void AVP_GameLobby::AddConsoleCmdsAndVars()
{
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.AllowRankedMatch",			&g_bAllowRankedMatch,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.Quit",						RequestQuit,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.CreateGame",					CreateGame,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.RotateMap",					RotateMap,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.SearchForGame",				SearchForGame,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.StartGame",					StartGame,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.GetLocalPlayerReady", 		GetLocalPlayerReady,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.ChangeSpecies",				ChangeSpecies,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.ChangeTeam",					ChangeTeam,					Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.AllowChangingTeam",			AllowChangingTeam,			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.AllowChangingSpecies",		AllowChangingSpecies,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.IsRoleBasedGame",				IsRoleBasedGame,			Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.PopulateMapSelectorNames",	PopulateMapSelectorName,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.PopulateMapSearchNames",		PopulateMapSearchName,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddCmd( L"AVP.GameLobby.PopulateMapDisplayNames", 	PopulateMapDisplayName,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	// hack - allow hosting tiny games to help testing.
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.MaxPlayers",	&s_xLobbySettings.m_uMaxPlayers,	uAVP_MAX_NET_PLAYERS_MIN,			uAVP_MAX_NET_PLAYERS_MAX,				Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.MapIndex",	&s_xLobbySettings.m_iMapIndex,		0,									0,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.Gametype",	&s_xLobbySettings.m_iGametypeIndex,	0,									AVP_Settings::AVP_NUM_GAMETYPES - 1,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::SetPointerToModifiedFlag(L"AVP.GameLobby.Gametype", &s_bGametypeChanged);
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.Species",		&s_xLobbySettings.m_iLocalSpecies,	0,									AVP_Settings::AVP_NUM_SPECIES - 1,		Asura_CommandConsole::REMOVE_AT_SHUTDOWN );

	ASURA_STATIC_ASSERT( iAVP_LOBBY_INVALID_SEARCH_FILTER == -1 );
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.SearchFilter.Map",		&s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP],		iAVP_LOBBY_INVALID_SEARCH_FILTER,	0,										Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.SearchFilter.Gametype",	&s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_GAMETYPE],	iAVP_LOBBY_INVALID_SEARCH_FILTER,	AVP_Settings::AVP_NUM_GAMETYPES - 1,	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}

void AVP_GameLobby::Shutdown()
{
	// only shutdown the lobby if we quit from it, otherwise it remains active for the game duration.
	if (IsQuitRequested() || AVP_GameStatus::GetQuitApplication())
	{
		Asura_Lobby_System::Shutdown();

		// Also only clear the map list on quit as it only gets created when the lobby system initialises.
		ClearMapList();
	}

	Platform_Shutdown();
}

bool AVP_GameLobby::Update()
{
	if (!Asura_NetLinkManager::IsConnected())
	{
		// connection dropped, any further usage of the network functions could crash.
		// The game should display a "connection dropped" dialog. It could do this by 
		// trapping the netlink state ASURA_NETLINK_CONNECTION_DROPPED in the main menu
		// and switching to the appropriate error menu, then calling Asura_NetLinkManager::ResetDroppedStatus()
		// either directly or using the appropriate console command.
		s_bQuitRequested = true;
		return true;
	}

	Asura_CommandConsole::Update();

	// pick a server if we've done a search
	if (Asura_Lobby_System::GetSearchState() == ASURA_LOBBY_SEARCHSTATE_SEARCH_COMPLETE)
	{
		const u_int uNumSearchResults = Asura_Lobby_System::GetNumSearchResults();
		u_int uIndexToJoin = 0;
		u_int uMostPlayers = 0;
		for (u_int u = 0; u < uNumSearchResults; ++u)
		{
			// pick the one with the most players, if it has some room in it
			const u_int uNumPlayers = Asura_Lobby_System::GetSearchResultByIndex(u).m_uNumPlayers;
			const u_int uMaxPlayers = Asura_Lobby_System::GetSearchResultByIndex(u).m_uMaxPlayers;
			if (uNumPlayers < uMaxPlayers && uNumPlayers > uMostPlayers)
			{
				uIndexToJoin = u;
				uMostPlayers = uNumPlayers;
			}
		}
		if (uIndexToJoin < uNumSearchResults)
		{
			Asura_Lobby_System::JoinLobby(uIndexToJoin);
		}
	}

	// keep lobby data up to date on clients, force gametype to be valid
	UpdateLobbySettings();

	// keep team names up-to-date
	UpdateTeamNames();

	// voice channels settings for the Lobby
	SetLobbyVoiceChannelsSettings(); 

	// check the lobby gamestate to see if we should load up the game
	if (Asura_Lobby_System::IsReadyToJoinGame())
	{
		s_bSetupComplete = true;
		InitialiseGameSettings();
	}

	return s_bSetupComplete || s_bQuitRequested;
}

void AVP_GameLobby::UpdateLobbySettings()
{
	// force gametype to be allowed value
	if (s_bGametypeChanged)
	{
		s_bGametypeChanged = false;
		CheckGametypeIsAllowed();
	}
	s_iPrevGametype = s_xLobbySettings.m_iGametypeIndex;

	// keep lobby data up to date on clients
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	if (pxLobby)
	{
		if (!Asura_Lobby_System::IsHost())
		{
			s_xLobbySettings.m_iGametypeIndex = pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAMETYPE);
			s_xLobbySettings.m_iMapIndex = GetMapIndexByHash(pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_MAP));
			AVP_Settings::SetGameType( static_cast<AVP_Settings::GameType>(s_xLobbySettings.m_iGametypeIndex) );
			AVP_Multiplayer_GameRules::Refresh(); 		
		}

		// set lobby requested team to match preferred species. This will need to be more complex at some point.
		if (Asura_Lobby_System::GetState() == ASURA_LOBBY_STATE_PRE_GAME)
		{
			if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
			{
				if (AVP_Multiplayer_GameRules::GetAllowMixedTeams() == false) // Species based team deathmatch
				{
					if (AVP_Multiplayer_GameRules::GetTeamSpeciesAreMutuallyExclusive())	// No two teams with the same species
					{
						u_char ucPreferredTeam = ASURA_LOBBY_INVALID_TEAM;
						AVP_Settings::Species ePreferredSpecies = static_cast<AVP_Settings::Species>(s_xLobbySettings.m_iLocalSpecies);
						if (ePreferredSpecies == AVP_Multiplayer_GameRules::GetTeamOneSpecies())
						{
							ucPreferredTeam = 0;
						}
						else if (ePreferredSpecies == AVP_Multiplayer_GameRules::GetTeamTwoSpecies())
						{
							ucPreferredTeam = 1;
						}
						else if (ePreferredSpecies == AVP_Multiplayer_GameRules::GetTeamThreeSpecies() && AVP_Multiplayer_GameRules::GetNumberOfTeams() > 2)
						{
							ucPreferredTeam = 2;
						}
						else
						{
							// Our selected species isn't allowed. Let's rotate teams
							ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam();
							ucPreferredTeam = (ucPreferredTeam + 1) % static_cast<u_char>(AVP_Multiplayer_GameRules::GetNumberOfTeams());
							AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
							AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
							// Player species is determined by the team in this case
							AVP_Settings::Species ePlayerSpecies = AVP_Multiplayer_GameRules::GetValidPlayerSpecies(ePreferredSpecies, ePlayerTeam, AVP_Settings::AVP_ROLE_NORMAL, false); 
							s_xLobbySettings.m_iLocalSpecies = static_cast<int>(ePlayerSpecies); 
						}
						Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam);
					}
					else
					{
						// Species settings are ambiguous in this case, player should be assigned to a team
						u_char ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam();
						// Check if the team settings are correct
						if (ucPreferredTeam + 1 > AVP_Multiplayer_GameRules::GetNumberOfTeams()) 
						{
							// Assign a random team
							ucPreferredTeam = static_cast<u_char>(Asura_Random::GetIntU( 0, AVP_Multiplayer_GameRules::GetNumberOfTeams() - 1 ));  
							AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
							AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
							// Player species is determined by the team in this case
							AVP_Settings::Species ePlayerSpecies = AVP_Multiplayer_GameRules::GetValidPlayerSpecies(AVP_Settings::AVP_SPECIES_UNKNOWN, ePlayerTeam, AVP_Settings::AVP_ROLE_NORMAL, false); 
							s_xLobbySettings.m_iLocalSpecies = static_cast<int>(ePlayerSpecies); 
							Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam);
						}
						else
						{
							AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
							// Player species is determined by the team in this case
							AVP_Settings::Species ePlayerSpecies = AVP_Multiplayer_GameRules::GetValidPlayerSpecies(AVP_Settings::AVP_SPECIES_UNKNOWN, ePlayerTeam, AVP_Settings::AVP_ROLE_NORMAL, false); 
							s_xLobbySettings.m_iLocalSpecies = static_cast<int>(ePlayerSpecies); 
						}
					}
				}
				else	// Mixed teams
				{
					// Make sure we pick up a team to start with
					u_char ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam();
					// Check if the team settings are correct
					if (ucPreferredTeam + 1 > AVP_Multiplayer_GameRules::GetNumberOfTeams()) 
					{
						// Assign a random team
						ucPreferredTeam = static_cast<u_char>(Asura_Random::GetIntU( 0, AVP_Multiplayer_GameRules::GetNumberOfTeams() - 1 ));  
						AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
						AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
						Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam);
					}		
				}
			}

			// set lobby requested species
			Asura_Assert(s_xLobbySettings.m_iLocalSpecies >= 0 && s_xLobbySettings.m_iLocalSpecies < AVP_Settings::AVP_NUM_SPECIES, "Invalid species selection"); 
			u_char ucSpecies = static_cast<u_char>(s_xLobbySettings.m_iLocalSpecies);
			Asura_Lobby_System::SetLocalPlayersProjectData(&ucSpecies); 
		}
	}
}

void AVP_GameLobby::InitialiseGameSettings()
{
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	Asura_Assert(pxLobby, "need a valid lobby pointer!");
	if (pxLobby)
	{
		// get lobby attributes
		int iGametype = pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAMETYPE);
		int iLevel = GetMapIndexByHash(pxLobby->GetAttributes().GetIntAttribute(AVP_LOBBY_ATTRIBUTE_MAP));
		Asura_Assert(iLevel < static_cast<int>(s_xMapList.Size()), "invalid level index from lobby data");
		iLevel = Asura_Maths::Min(iLevel, static_cast<int>(s_xMapList.Size() - 1));
		const char* szLevelName = GetMapFilename(iLevel);

		AVP_Settings::SetMaxNetPlayers(pxLobby->GetMaxPlayers());
		AVP_Settings::SetGameType( static_cast<AVP_Settings::GameType>(iGametype) );
		AVP_Multiplayer_GameRules::Refresh(); 
		AVP_Settings::SetLevelName(szLevelName);

		const Asura_Lobby_Player* pxLocalPlayer = pxLobby->FindLocalPlayer();
		Asura_Assert(pxLocalPlayer, "couldn't get local player from lobby system");
		if (pxLocalPlayer)
		{
			if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
			{
				u_char ucTeam = pxLocalPlayer->GetTeamIndex();
				Asura_Assert(ucTeam < AVP_Multiplayer_GameRules::GetNumberOfTeams(), "invalid team");
				AVP_Settings::SetRequestedMultiplayerTeam(static_cast<AVP_Settings::Team>(ucTeam + 1)); // 0 is TEAM_NONE.
				// Voice chat is directed to our team mates only
				Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY ); 
				Asura_Lobby_System::ConnectLocalPlayerToVoiceChannel(ucTeam + 1); 
				Asura_Lobby_System::SetLocalPlayersActiveVoiceChannel(ucTeam + 1); 
			}
			else
			{
				AVP_Settings::SetRequestedMultiplayerTeam( AVP_Settings::AVP_TEAM_NONE ); 
				// Talk to everyone using the in-game global voice channel
				Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY ); 
				Asura_Lobby_System::ConnectLocalPlayerToVoiceChannel( uAVP_VOICE_CHANNEL_GLOBAL ); 
				Asura_Lobby_System::SetLocalPlayersActiveVoiceChannel( uAVP_VOICE_CHANNEL_GLOBAL ); 
			}
		}
	}

	bool bServer = Asura_Lobby_System::IsHost();
	AVP_Settings::SetIsServer( bServer );
	AVP_Settings::SetExternalNetworkActive(true);

	if (!bServer)
	{
		if (Asura_Lobby_System::GetGameAddress())
		{
			AVP_Settings::SetAddressToConnectTo( Asura_Lobby_System::GetGameAddress() );
		}
	}

	// TODO: the way these are set will change
	AVP_Settings::SetPlayerSpecies(static_cast<AVP_Settings::Species>(s_xLobbySettings.m_iLocalSpecies));
	AVP_Game_Core::SetRespawnSpecies(static_cast<AVP_Settings::Species>(s_xLobbySettings.m_iLocalSpecies));
}

void AVP_GameLobby::UpdateTeamNames()
{
	if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
	{
		const Asura_Hash_ID uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
		if (AVP_Multiplayer_GameRules::GetAllowMixedTeams())
		{
			// teams are team 1,2,3
			Asura_Lobby_System::SetTeamName(0, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920974) );// "team1"
			Asura_Lobby_System::SetTeamName(1, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920975) );// "team2"
			Asura_Lobby_System::SetTeamName(2, Asura_HashedLocalisedText::GetString(uPageHash, 0x06920976) );// "team3"
		}
		else
		{
			// teams are per species
			Asura_Lobby_System::SetTeamName(0, Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamOneSpecies())));
			Asura_Lobby_System::SetTeamName(1, Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamTwoSpecies())));
			Asura_Lobby_System::SetTeamName(2, Asura_HashedLocalisedText::GetString(uPageHash, GetSpeciesNameTextHashID(AVP_Multiplayer_GameRules::GetTeamThreeSpecies())));
		}
	}

}

void AVP_GameLobby::SetLobbyVoiceChannelsSettings()
{
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_GLOBAL ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_TEAM1 ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_TEAM2 ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_TEAM3 ); 
	Asura_Lobby_System::DisconnectLocalPlayerFromVoiceChannel( uAVP_VOICE_CHANNEL_DEATH ); 
	Asura_Lobby_System::ConnectLocalPlayerToVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY );
	Asura_Lobby_System::SetLocalPlayersActiveVoiceChannel( uAVP_VOICE_CHANNEL_LOBBY ); 
}

void AVP_GameLobby::CreateMapList()
{
	Asura_Assert(s_xMapList.Size() == 0, "Map list already created.");
	Asura_FileAccess::FindFiles(szMULTIPLAYER_MAP_PATH, AVP_EXPORTED_LEVEL_FILE_EXTENSION, CreateMapList_Callback);

	// Remove the console-variables related to the number of maps, and re-add them to make sure that the maximum is correct...
	if (Asura_CommandConsole::VarExists(L"AVP.GameLobby.MapIndex"))			Asura_CommandConsole::RemoveVar(L"AVP.GameLobby.MapIndex");
	if (Asura_CommandConsole::VarExists(L"AVP.GameLobby.SearchFilter.Map"))	Asura_CommandConsole::RemoveVar(L"AVP.GameLobby.SearchFilter.Map");

	Asura_CommandConsole::AddVar( L"AVP.GameLobby.MapIndex",			&s_xLobbySettings.m_iMapIndex,						0,									Asura_Maths::Max(static_cast<int>(s_xMapList.Size()) - 1, 0),	Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	Asura_CommandConsole::AddVar( L"AVP.GameLobby.SearchFilter.Map",	&s_aiLobbySearchFilters[AVP_LOBBY_ATTRIBUTE_MAP],	iAVP_LOBBY_INVALID_SEARCH_FILTER,	s_xMapList.Size(),												Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
}

void AVP_GameLobby::CreateMapList_Callback(const char* szFilename)
{
	AVP_GameLobby_MapData* pxData = new AVP_GameLobby_MapData(szFilename);
	if (pxData)
	{
		s_xMapList.PushBack(pxData);
	}
}

void AVP_GameLobby::ClearMapList()
{
	for (Asura_LinkList_It<AVP_GameLobby_MapData*> xIt(s_xMapList); !xIt.Done();)
	{
		// Delete the data and remove the entry from the list.
		AVP_GameLobby_MapData** ppxData = xIt.GetCurrent();
		AVP_GameLobby_MapData* pxData = ppxData ? *ppxData : NULL;

		xIt.DeleteCurrent();
		delete pxData;
	}
}

int AVP_GameLobby::GetMapIndexByHash(Asura_Hash_ID uHash)
{
	int iIndex = 0;
	for (Asura_LinkList_It<AVP_GameLobby_MapData*> xIt(s_xMapList); !xIt.Done(); xIt.Next())
	{
		AVP_GameLobby_MapData** ppxData = xIt.GetCurrent();
		if (ppxData && *ppxData && (*ppxData)->GetHash() == uHash)
		{
			// Found it, index is already correct.
			break;
		}
		else
		{
			++iIndex;
		}
	}
	return iIndex; // Intentionally returning one greater than the maximum valid index if we don't have the map, as that allows us to display unknown.
}

Asura_Hash_ID AVP_GameLobby::GetMapHashByIndex(int iIndex)
{
	const AVP_GameLobby_MapData* pxData = NULL;
	if (iIndex < static_cast<int>(s_xMapList.Size()))
	{
		const Asura_ListElt<AVP_GameLobby_MapData*>* pxElement = s_xMapList[iIndex];
		if (pxElement)
		{
			pxData = pxElement->Data();
		}
	}

	if (pxData)
	{
		return pxData->GetHash();
	}
	else
	{
		return ASURA_HASH_ID_UNSET;
	}
}

const char* AVP_GameLobby::GetMapFilename(int iIndex)
{
	const AVP_GameLobby_MapData* pxData = NULL;
	if (iIndex < static_cast<int>(s_xMapList.Size()))
	{
		const Asura_ListElt<AVP_GameLobby_MapData*>* pxElement = s_xMapList[iIndex];
		if (pxElement)
		{
			pxData = pxElement->Data();
		}
	}

	if (pxData)
	{
		return pxData->GetFilename();
	}
	else
	{
		Asura_Assert(false, "Failed to get map filename.");
		return "";
	}
}

bool AVP_GameLobby::CreateGame( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Lobby_Attributes xAttributes;
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_MAP, GetMapHashByIndex(s_xLobbySettings.m_iMapIndex));
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAMETYPE, s_xLobbySettings.m_iGametypeIndex);
		xAttributes.SetIntAttribute(AVP_LOBBY_ATTRIBUTE_GAME_VERSION, Asura_NetTransport_System::Project_GetVersionID());

		// to find the number of teams we must set the gametype and refresh MP game rules
		AVP_Settings::SetGameType( static_cast<AVP_Settings::GameType>(s_xLobbySettings.m_iGametypeIndex) );
		AVP_Multiplayer_GameRules::Refresh(); 
		int iNumTeams = AVP_Multiplayer_GameRules::GetNumberOfTeams();
		Asura_Assert(iNumTeams > 0 && iNumTeams <= UCHAR_MAX, "unexpected num teams");

		Asura_Lobby_System::CreateLobby( s_xLobbySettings.m_uMaxPlayers, static_cast<u_char>(iNumTeams), xAttributes );
	}
	return true;
}

bool AVP_GameLobby::RotateMap( bool bHelp )
{
	if (!bHelp)
	{
		const Asura_Lobby* pxLobby( Asura_Lobby_System::GetLobby() ); 
		if( pxLobby )
		{
			Asura_Lobby_Attributes xLocalAttributes( pxLobby->GetAttributes() ); 
			int iMapIndex = GetMapIndexByHash( xLocalAttributes.GetIntAttribute( AVP_LOBBY_ATTRIBUTE_MAP ));
			iMapIndex = (iMapIndex + 1) % s_xMapList.Size();
			s_xLobbySettings.m_iMapIndex = GetMapHashByIndex(iMapIndex); 
			xLocalAttributes.SetIntAttribute( AVP_LOBBY_ATTRIBUTE_MAP, s_xLobbySettings.m_iMapIndex );
			Asura_Lobby_System::SetLobbyAttributes( xLocalAttributes );

			AVP_GameStatus::SetLoadMap(true);
		}
	}
	return true; 
}

bool AVP_GameLobby::SearchForGame( bool bHelp )
{
	if (!bHelp)
	{
		// create the search filters
		Asura_FixedArray<Asura_Lobby_SearchCriterion, AVP_NUM_LOBBY_ATTRIBUTES> axSearchCriteria;
		u_int uNumSearchCriteria = 0;
		for (u_int u = 0; u < AVP_NUM_LOBBY_ATTRIBUTES; ++u)
		{
			if (s_aiLobbySearchFilters[u] != iAVP_LOBBY_INVALID_SEARCH_FILTER)
			{
				if (u == AVP_LOBBY_ATTRIBUTE_MAP)
				{
					// Special case for the map - we need to send the hash id instead of the index, and ignore the value if it's at the limit.
					if (s_aiLobbySearchFilters[u] != static_cast<int>(s_xMapList.Size()))
					{
						u_int uFilter = uNumSearchCriteria++;
						axSearchCriteria[uFilter].Set(u, ASURA_LOBBY_SEARCH_OPERATOR_EQ, GetMapHashByIndex(s_aiLobbySearchFilters[u]));
					}
				}
				else
				{
					u_int uFilter = uNumSearchCriteria++;
					axSearchCriteria[uFilter].Set(u, ASURA_LOBBY_SEARCH_OPERATOR_EQ, s_aiLobbySearchFilters[u]);
				}
			}
		}

		// start the search
		Asura_Lobby_System::SearchForLobby(&axSearchCriteria[0], uNumSearchCriteria);
	}
	return true;
}

bool AVP_GameLobby::StartGame( bool bHelp )
{
	if (!bHelp)
	{
		Asura_Lobby_System::StartGame();
	}
	return true;
}

bool AVP_GameLobby::GetLocalPlayerReady( bool bHelp )
{
	if (!bHelp)
	{
		const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();

		if( pxLobby )
		{
			const Asura_Lobby_Player* pxLocalPlayer = pxLobby->FindLocalPlayer();
			if( pxLocalPlayer )
			{
				if( pxLocalPlayer->GetReadyStatus() )
				{
					Asura_CommandConsole::Printf( L"1" );
				}
				else
				{
					Asura_CommandConsole::Printf( L"0" );
				}
			}
		}
		else 
		{
			// if we cant find the lobby return false as the text "Set Ready" can (and was) used for both conditions of the ready button. 
			return false;
		}
	}
	return true;
}

bool AVP_GameLobby::ChangeSpecies( bool bHelp )
{
	if (!bHelp)
	{
		s_xLobbySettings.m_iLocalSpecies = (s_xLobbySettings.m_iLocalSpecies + 1) % AVP_Settings::AVP_NUM_SPECIES;
		Asura_Assert(s_xLobbySettings.m_iLocalSpecies >= 0 && s_xLobbySettings.m_iLocalSpecies < AVP_Settings::AVP_NUM_SPECIES, "Invalid species selection"); 
		u_char ucSpecies = static_cast<u_char>(s_xLobbySettings.m_iLocalSpecies);
		Asura_Lobby_System::SetLocalPlayersProjectData(&ucSpecies); 
	}
	return true;
}

bool AVP_GameLobby::ChangeTeam( bool bHelp )
{
	if (!bHelp)
	{
		u_char ucPreferredTeam = Asura_Lobby_System::GetLocalPlayersRequestedTeam(); 
		ucPreferredTeam = (ucPreferredTeam + 1) % static_cast<u_char>(AVP_Multiplayer_GameRules::GetNumberOfTeams());
		Asura_Lobby_System::SetLocalPlayersRequestedTeam(ucPreferredTeam); 
		AVP_Settings::Team ePlayerTeam = static_cast<AVP_Settings::Team>(ucPreferredTeam + 1); 
		AVP_Settings::SetRequestedMultiplayerTeam(ePlayerTeam);	
	} 
	return true; 
} 

bool AVP_GameLobby::AllowChangingTeam( bool bHelp )
{
	if (!bHelp)
	{
		Asura_CommandConsole::Printf( AllowChangingTeam() ? "1" : "0" );
	}
	return true;
}

bool AVP_GameLobby::AllowChangingSpecies( bool bHelp )
{
	if (!bHelp)
	{
		Asura_CommandConsole::Printf( AllowChangingSpecies() ? "1" : "0" );
	}
	return true;
}

bool AVP_GameLobby::IsRoleBasedGame( bool bHelp )
{
	if (!bHelp)
	{
		Asura_CommandConsole::Printf( AVP_Multiplayer_GameRules::IsRoleBasedGame() ? "1" : "0" );
	}
	return true;
}

bool AVP_GameLobby::PopulateMapSelectorName(u_int uElementIndex, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		return true;
	}
	else
	{
		return PopulateMapName(uElementIndex, false, false);
	}
}

bool AVP_GameLobby::PopulateMapSearchName(u_int uElementIndex, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		return true;
	}
	else
	{
		return PopulateMapName(uElementIndex, true, false);
	}
}

bool AVP_GameLobby::PopulateMapDisplayName(u_int uElementIndex, bool bInterrogate, bool bHelp)
{
	if (bHelp || bInterrogate)
	{
		return true;
	}
	else
	{
		return PopulateMapName(uElementIndex, false, true);
	}
}

bool AVP_GameLobby::PopulateMapName(u_int uElementIndex, bool bIncludeAny, bool bIncludeUnknown)
{
	const AVP_GameLobby_MapData* pxData = NULL;
	if (uElementIndex < s_xMapList.Size())
	{
		const Asura_ListElt<AVP_GameLobby_MapData*>* pxElement = s_xMapList[uElementIndex];
		if (pxElement)
		{
			pxData = pxElement->Data();
		}
	}

	if (pxData)
	{
		Asura_CommandConsole::Printf(L"%S", pxData->GetDisplayName());
		return true;
	}
	else if (uElementIndex == s_xMapList.Size())
	{
		if (bIncludeAny)
		{
			// TODO - Localise this.
			Asura_CommandConsole::Printf(L"Any");
			return true;
		}
		else if (bIncludeUnknown)
		{
			// TODO - Localise this.
			Asura_CommandConsole::Printf(L"<UNKNOWN>");
			return true;
		}
	}
	return false;
}

bool AVP_GameLobby::AllowChangingTeam()
{
	if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
	{
		if (AVP_Multiplayer_GameRules::GetAllowMixedTeams())
		{
			return true; 
		}
		else
		{
			if (!AVP_Multiplayer_GameRules::GetTeamSpeciesAreMutuallyExclusive())
			{
				return true;
			}
		}
	}
	return false; 
}

bool AVP_GameLobby::AllowChangingSpecies()
{
	// Check for team based game rules not allowing players to select species
	if (AVP_Multiplayer_GameRules::IsTeamBasedGame())
	{
		if (!AVP_Multiplayer_GameRules::GetAllowMixedTeams())
		{
			if (!AVP_Multiplayer_GameRules::GetTeamSpeciesAreMutuallyExclusive())
			{
				return false;
			}
		}
	}
	
	// In role based games players don't select their species
	if (AVP_Multiplayer_GameRules::IsRoleBasedGame())
	{
		return false;
	}
	
	return true; 
}


void AVP_GameLobby::Render()
{
	Asura_GUIMenu_System::Render();
}


bool AVP_GameLobby::IsGametypeAllowed(int iGametype)
{
	// this is basically a hack to allow us to restrict gamemodes for milestones.
	switch (iGametype)
	{
		case AVP_Settings::AVP_GAMETYPE_DEATH_MATCH:
		case AVP_Settings::AVP_GAMETYPE_SPECIES_TEAM_DEATH_MATCH:
		//case AVP_Settings::AVP_GAMETYPE_DISRUPTION_DOMINATION:
		case AVP_Settings::AVP_GAMETYPE_HUNT:
		case AVP_Settings::AVP_GAMETYPE_SURVIVOR:
		case AVP_Settings::AVP_GAMETYPE_MIXED_SPECIES_TEAM_DEATH_MATCH:
		case AVP_Settings::AVP_GAMETYPE_DOMINATION:
			// allowed
			return true;
		default:
			// not allowed
			return false;
	}

}
void AVP_GameLobby::CheckGametypeIsAllowed()
{
	// ensure that if gametype was modified by console that it is a valid value
	// this is basically a hack to allow us to restrict gamemodes for milestones.
	int iIncrement = s_xLobbySettings.m_iGametypeIndex > s_iPrevGametype ? 1 : -1;
	while (!IsGametypeAllowed(s_xLobbySettings.m_iGametypeIndex))
	{
		if (s_xLobbySettings.m_iGametypeIndex >= AVP_Settings::AVP_NUM_GAMETYPES - 1)
		{
			iIncrement = -1;
		}
		if (s_xLobbySettings.m_iGametypeIndex <= 0)
		{
			iIncrement = 1;
		}
		s_xLobbySettings.m_iGametypeIndex += iIncrement;
	}
}

AVP_GameLobby_GameSettings::AVP_GameLobby_GameSettings()
{
	// set up default values
	m_uMaxPlayers = 9;
	m_iMapIndex = 0;
	m_iGametypeIndex = AVP_Settings::AVP_GAMETYPE_DEATH_MATCH;
	m_iLocalSpecies = static_cast<int>(Asura_Random::GetSeededIntU() % AVP_Settings::AVP_NUM_SPECIES);
}

AVP_GameLobby_MapData::AVP_GameLobby_MapData(const char* szFilename)
{
	Asura_PrintfToBfr(m_acFilename, szFilename, 256, true);
	m_uHash = Asura_GetHashID(szFilename);
	// This is messy, but since we'll have to replace this with a localised string at some point anyway it doesn't seem a major issue right now.
	Asura_String xFilename(szFilename);
	Asura_String xName;
	xFilename.SplitFilename(NULL, &xName, NULL);
	if (xName.StartsWith("MP_"))
	{
		xName = xName.SubString(3, xName.GetLength() - 3);
	}
	Asura_PrintfToBfr(m_acDisplayName, xName.GetString(), 32, true);
}

// ASURA CODE SHARED BY GAME AND DEV LOBBIES /////////////////////////////////
void Asura_Lobby_System::Project_PrintErrorCodeToConsole()
{
	Asura_Hash_ID uPageHash	= ASURA_HASH_ID_UNSET;

	const AVP_GUIMenu_System::LOBBY_TYPE	eLobbyType	= AVP_GUIMenu_System::GetLobbyType();

	switch(eLobbyType)
	{
#if defined(AVP_USE_DEV_LOBBY)
		case AVP_GUIMenu_System::LT_DEVELOPMENT:
		{
			uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
			break;
		}
#endif	//defined(AVP_USE_DEV_LOBBY)

#if defined(AVP_USE_GAME_LOBBY)
		case AVP_GUIMenu_System::LT_GAME:
		{
			uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby") //0x3EFE8AFB; // Asura_GetHashID("GameLobby_Lobby")
			break;
		}
#endif	//defined(AVP_USE_GAME_LOBBY)

		default:
		{
			Asura_Assert(false, "Unexpected lobby-type found.");
			break;
		}
	}	// End of switch.

	u_int uStringHash = ASURA_HASH_ID_UNSET;
	switch (s_iNetworkDisconnectReason)
	{
		case ASURA_NDR_TIMEOUT:
		{
			uStringHash = 0x05353E2A;//Asura_GetHashID("Error_Timeout")
			break;
		}
		case ASURA_NDR_SERVERQUIT:
		{
			uStringHash = 0x0B5B2C4E;//Asura_GetHashID("Error_HostQuit")
			break;
		}
		// PS3 error codes copied from SDK documentation.
		case 0x80022b13:
		{
			uStringHash = 0x7988D3B1; // Asura_GetHashID("Error_PS3RoomDoesNotExist")
			break;
		}
		case 0x80022b08:
		{
			uStringHash = 0x4E828A48; // Asura_GetHashID("Error_PS3UserBlocked")
			break;
		}
		case 0x80022b19:
		{
			uStringHash = 0xAA8044C3; // Asura_GetHashID("Error_PS3RoomFull")
			break;
		}
		default:
		{
			uStringHash = 0x8CE918B9;// Asura_GetHashID("Error_Code_Label")
			const Asura_UnicodeChar* wszString = Asura_HashedLocalisedText::GetString(uPageHash, uStringHash);
			Asura_CommandConsole::Printf( L"%s 0x%x", wszString, s_iNetworkDisconnectReason );
			return;
		}
	}

	const Asura_UnicodeChar* wszString = Asura_HashedLocalisedText::GetString(uPageHash, uStringHash);
	Asura_CommandConsole::Printf( L"%s", wszString );
}

void Asura_Lobby_System::Project_GetInviteText( const Asura_UnicodeChar*& wszSubject, const Asura_UnicodeChar*& wszBody )
{
	const Asura_Hash_ID uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
	const Asura_Hash_ID uStringHashSubject = 0x0B7511F6; // Asura_GetHashID("Invite_Subject")
	const Asura_Hash_ID uStringHashBody = 0xA657C118; // Asura_GetHashID("Invite_Body")
	wszSubject = Asura_HashedLocalisedText::GetString(uPageHash, uStringHashSubject);
	wszBody = Asura_HashedLocalisedText::GetString(uPageHash, uStringHashBody);
}

void Asura_Lobby_Attributes::Project_ChangeByteOrder()
{
	ASURA_NTS_CHANGE_BYTE_ORDER(m_aiAttributes[AVP_LOBBY_ATTRIBUTE_MAP]);
	ASURA_NTS_CHANGE_BYTE_ORDER(m_aiAttributes[AVP_LOBBY_ATTRIBUTE_GAMETYPE]);
	ASURA_NTS_CHANGE_BYTE_ORDER(m_aiAttributes[AVP_LOBBY_ATTRIBUTE_GAME_VERSION]);
	// KOR - these need serious attention before cross-endian networking will work, due to being bitfields.
	ASURA_NTS_CHANGE_BYTE_ORDER(m_aiAttributes[AVP_LOBBY_ATTRIBUTE_ADVANCED_SETTINGS_A]);
	ASURA_NTS_CHANGE_BYTE_ORDER(m_aiAttributes[AVP_LOBBY_ATTRIBUTE_ADVANCED_SETTINGS_B]);
}

void Asura_Matchmaking_System::Project_PrintAutomatchStatusToConsole()
{
	const Asura_Hash_ID uPageHash = 0x424C9CE1; // Asura_GetHashID("DevLobby")
	const Asura_Lobby* pxLobby = Asura_Lobby_System::GetLobby();
	if (Asura_Matchmaking_System::IsWaitingForPlayers() && pxLobby)
	{
		// display string of the form "got %d players, waiting for %d players"
		const Asura_Hash_ID uStringHash = 0xC9126D01; // Asura_GetHashID("Automatch_Waiting_For_Players")
		const Asura_UnicodeChar* wszText = Asura_HashedLocalisedText::GetString(uPageHash, uStringHash);
		Asura_CommandConsole::Printf( wszText, pxLobby->GetNumPlayers(), 
			Asura_Matchmaking_System::GetNumPlayersBeingWaitedFor() );
	}
	else
	{
		// display "Searching" or similar
		const Asura_Hash_ID uStringHash = 0x007DB878; // Asura_GetHashID("Search_For_Game_Title")
		const Asura_UnicodeChar* wszText = Asura_HashedLocalisedText::GetString(uPageHash, uStringHash);
		Asura_CommandConsole::Printf( L"%s", wszText );
	}
}

#endif //#if defined(AVP_USE_GAME_LOBBY)

// E N D   O F   F I L E /////////////////////////////////////////////////////
