
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_InventoryItem_Cloak.cpp
// Author : Andy Tate
// Date   : 28-Nov-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Inventory/Items/AVP_InventoryItem_PredatorCloak.h"

// Asura headers
#include "Animation/Asura_AnimationController_Animation.h"
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Interface/Asura_CommandConsole.h"
#include "Sound/Asura_Sound_Event_Storage.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AnimStateData.h"
#include "Actor/AVP_Actor_DynamicStat.h"
#include "Actor/Vertebrates/AVP_Vertebrate_AnimationHandler.h"
#include "Actor/Vertebrates/Predator/AVP_Predator_InputHandler.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ClientActor_Predator.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Predator/AVP_ServerActor_Predator.h"
#include "Inventory/AVP_Inventory_HashIDs.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/Items/AVP_InventoryItem_Weapon.h"
#include "SpecialFX/AVP_CloakEffect.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
DVSVAR bool dvs_bInfiniteCloak = false;
#endif

////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::UpdateEnergy
// Author       : Andy Tate
// Date         : 29-Nov-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_InventoryItem_Cloak::UpdateEnergy(AVP_ClientActor_Predator& xPred,
												   AVP_Actor_DynamicStat& xDynamicEnergy)
{
	const float	fTimeSlice	= Asura_Timers::GetGameTimeSlice();

	float		fAmount;

	// Check the actor state...
	if ( xPred.IsMovingFast() || xPred.GetGroundType() == AVP_Actor_PhysicsObjectHandler::OFF_GROUND )
	{
		fAmount = m_fCloakedEnergyDrainSprinting;
	}
	else if ( xPred.IsMovingMedium() )
	{
		fAmount = m_fCloakedEnergyDrainRunning;
	}
	else if ( xPred.IsMovingSlow() )
	{
		fAmount = m_fCloakedEnergyDrainWalking;
	}
	else if ( xPred.IsCrouching() )
	{
		fAmount = m_fCloakedEnergyDrainCrouching;
	}
	else
	{
		fAmount = m_fCloakedEnergyDrainStanding;
	}
	fAmount	*= fTimeSlice;
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	if (dvs_bInfiniteCloak)
	{
		fAmount = 0.0f;
	}
#endif

	// Update the energy...
	if(xDynamicEnergy.CanRemove(fAmount))
	{
		xDynamicEnergy.Remove(fAmount, true);
		return(false);
	}
	else
	{
		// Failed to remove energy required to cloak, so de-cloak...
		xDynamicEnergy.SetToMinimum(true);
		return(true);
	}
}

/// End of function 'AVP_InventoryItem_Cloak::UpdateEnergy'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::PlayOneShotSound
// Author       : Andy Tate
// Date         : 25-Jan-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Cloak::PlayOneShotSound(AVP_ClientEntity_Actor& xClient,
											   const Asura_Hash_ID uBlueprintParamHash)
{
	const Asura_Blueprint*	pxBlueprint	= Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, m_uBlueprintHash);

	// Extract the blueprints...
	if(pxBlueprint)
	{
		Asura_Value	xValue;

		if(pxBlueprint->GetParameterValueByHashID(uBlueprintParamHash, xValue))
		{
			const Asura_Hash_ID			uSoundEventHash	= xValue.GetHashIDValue();
			const Asura_Sound_Event*	pxSound			= Asura_Sound_Event_Storage::FindEvent(uSoundEventHash);

			if(pxSound != NULL)
			{
				if (m_iSoundHandle != ASURA_INVALID_HANDLE)
				{
					Asura_Sound_Event_System::FinishedWithHandle(m_iSoundHandle);
				}
				Asura_Sound_Event_System::PlayEvent(pxSound, &m_iSoundHandle, &xClient.GetCurrentPosition(), 1.0f, xClient.GetGuid());
			}
		}
	}
}

/// End of function 'AVP_InventoryItem_Cloak::PlayOneShotSound'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::OnStateChange
// Author       : Andy Tate
// Date         : 29-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Cloak::OnStateChange(const u_int uOldState,
											const u_int uNewState)
{
	AVP_Inventory_System&		xParentSystem	= GetParentSystem();
	AVP_ClientActor_Predator*	pxClient		= static_cast<AVP_ClientActor_Predator*>(xParentSystem.GetParentClient());

	// Call the base-function...
	PARENT::OnStateChange(uOldState, uNewState);

	// Do we need to handle client-side stuff?
	if(pxClient != NULL)
	{
		const State	eSourceState	= GetSourceState();
		const u_int	uTransitionType	= GetCurrentTransitionType();

		switch(uNewState)
		{
			case BEGIN_IDLING:
			{
				if((eSourceState == USING) && (uTransitionType != TT_INSTANT))
				{
					PlayOneShotSound(*pxClient, 0x47C68FD0);	//Asura_GetHashID("Sound.CloakOn"));
					m_pxCloakEffect->BeginUncloaking(false);
				}
				break;
			}

			case IDLING:
			{
				if((eSourceState == USING) && (uTransitionType == TT_INSTANT))
				{
					m_pxCloakEffect->BeginUncloaking(true);
				}
				break;
			}

			case END_IDLING:
			{
				const State	eDestinationState	= GetDestinationState();

				if( (eDestinationState == USING) && (uTransitionType == TT_MANUAL) )
				{
					if (GetParentSystem().IsClientSide())
					{
						PlayAnimation();
					}
				}
				break;
			}

			case BEGIN_USING:
			{
				if(uTransitionType != TT_INSTANT)
				{
					PlayOneShotSound(*pxClient, 0xB10B699E);	//Asura_GetHashID("Sound.CloakOff"));
					m_pxCloakEffect->BeginCloaking(false);
				}
				break;
			}

			case USING:
			{
				if(uTransitionType == TT_INSTANT)
				{
					m_pxCloakEffect->BeginCloaking(true);
				}
				break;
			}

			case END_USING:
			{
				if(uTransitionType == TT_MANUAL)
				{
					if (GetParentSystem().IsClientSide())
					{
						PlayAnimation();
					}
				}
				break;
			}

			default:
			{
				break;
			}
		}		// End of switch.
	}
}

/// End of function 'AVP_InventoryItem_Cloak::OnStateChange'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::UpdateClientState
// Author       : Andy Tate
// Date         : 28-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Cloak::UpdateClientState(AVP_ClientEntity_Actor& xClient)
{
	Asura_Animation* pxAnimation = xClient.GetShapeInstance().GetAnimation();
	const Asura_AnimationController_Animation* pxAnimController = (pxAnimation && (m_uAnimControllerHash != ASURA_HASH_ID_UNSET)) ? pxAnimation->GetControllerAnimationFromHash(m_uAnimControllerHash) : NULL;
	if (!pxAnimController)
	{
		m_uAnimControllerHash = ASURA_HASH_ID_UNSET;
	}
	const State eCurrentState = GetCurrentState();

	// Decide what to do based on our current state...
	switch(eCurrentState)
	{
		case INACTIVE:
		case IDLING:
		{
			break;
		}

		case BEGIN_IDLING:
		{
			const State	eSourceState	= GetSourceState();

			if(eSourceState == USING)
			{
				const u_int	uTransitionType	= GetCurrentTransitionType();

				if ((uTransitionType == TT_MANUAL) && pxAnimController)
				{
					const u_int	uAnimFlags = pxAnimController->GetFlagsForLastFrame();

					if((uAnimFlags & AVP_ANIMFLAG_COMPLETE) != 0)
					{
						RequestNextIdlingState();
					}
				}
				else
				{
					RequestNextIdlingState();
				}
			}
			else
			{
				RequestNextIdlingState();
			}
			break;
		}

		case END_IDLING:
		{
			const State	eDestinationState	= GetDestinationState();

			if(eDestinationState == USING)
			{
				const u_int	uTransitionType	= GetCurrentTransitionType();

				if ((uTransitionType == TT_MANUAL) && pxAnimController)
				{
					const u_int	uAnimFlags = pxAnimController->GetFlagsForLastFrame();

					if((uAnimFlags & AVP_ANIMFLAG_ACTION_1) != 0)
					{
						RequestNextUsingState();
					}
				}
				else
				{
					RequestNextUsingState();
				}
			}
			break;
		}

		case BEGIN_USING:
		{
			const u_int	uTransitionType	= GetCurrentTransitionType();

			if ((uTransitionType == TT_MANUAL) && pxAnimController)
			{
				const u_int	uAnimFlags = pxAnimController->GetFlagsForLastFrame();

				if((uAnimFlags & AVP_ANIMFLAG_COMPLETE) != 0)
				{
					RequestNextUsingState();
				}
			}
			else
			{
				RequestNextUsingState();
			}
			break;
		}

		case USING:
		{
			AVP_ClientActor_Predator&		xClientPredator	= static_cast<AVP_ClientActor_Predator&>(xClient);
			AVP_Actor_DynamicStat&			xDynamicEnergy	= *xClientPredator.GetDynamicEnergy();

			if ( UpdateEnergy( xClientPredator, xDynamicEnergy ) )
			{
				// We need to go to automatic, since we haven't manually requested leaving the mode.
				//RequestIdling(false, TT_AUTOMATIC);
			}
			break;
		}

		case END_USING:
		{
			const u_int	uTransitionType	= GetCurrentTransitionType();

			if ((uTransitionType == TT_MANUAL) && pxAnimController)
			{
				const u_int	uAnimFlags = pxAnimController->GetFlagsForLastFrame();

				if((uAnimFlags & AVP_ANIMFLAG_ACTION_1) != 0)
				{
					RequestNextIdlingState();
				}
			}
			else
			{
				RequestNextIdlingState();
			}
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected state found when updating master-state of weapon.");
			break;
		}
	}		// End of switch.

	// Update the Predator cloak..
	if ( pxAnimation )
	{
		if(m_pxCloakEffect->Update(xClient))
		{
			AVP_Animation_ProjectData *pxProjectData = AVP_Animation_ProjectData::FromAnimation(pxAnimation);
			if (pxProjectData) pxProjectData->SetCloakEffect(m_pxCloakEffect);
		}
		else
		{
			AVP_Animation_ProjectData *pxProjectData = AVP_Animation_ProjectData::FromAnimation(pxAnimation);
			if (pxProjectData) pxProjectData->SetCloakEffect(NULL);
		}
	}

	if ((m_iSoundHandle != ASURA_INVALID_HANDLE) && !Asura_Sound_Event_System::IsPlaying(m_iSoundHandle))
	{
		Asura_Sound_Event_System::Stop(m_iSoundHandle);
	}
}

/// End of function 'AVP_InventoryItem_Cloak::UpdateClientState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::UpdateServerState
// Author       : Andy Tate
// Date         : 28-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Cloak::UpdateServerState(AVP_ServerEntity_Actor& xServer)
{
	const AVP_ClientEntity_Actor* pxClientActor = AVP_ClientEntity_Actor::FindEntity( xServer.GetGuid() );
	const AVP_Inventory_System* pxClientSys = pxClientActor ? pxClientActor->GetInventorySystem() : NULL;
	const AVP_InventoryItem_Cloak* pxClientCloak = pxClientSys ? static_cast< const AVP_InventoryItem_Cloak* >( pxClientSys->GetItem( GetHashInParentGroup() ) ) : NULL;

	const State eCurrentState = GetCurrentState();

	// Decide what to do based on our current state...
	switch(eCurrentState)
	{
		case IDLING:
		case INACTIVE:
		{
			break;
		}

		case BEGIN_IDLING:
		{
			const State	eSourceState = GetSourceState();
			const u_int	uTransitionType	= GetCurrentTransitionType();

			if ( eSourceState != USING || uTransitionType != TT_MANUAL || pxClientCloak == NULL || pxClientCloak->GetCurrentState() == IDLING || pxClientCloak->GetDestinationState() != IDLING )
			{
				RequestNextIdlingState();
			}
			break;
		}

		case END_IDLING:
		{
			const State	eDestinationState	= GetDestinationState();

			if ( eDestinationState == USING )
			{
				const u_int	uTransitionType	= GetCurrentTransitionType();

				if ( uTransitionType != TT_MANUAL || pxClientCloak == NULL || pxClientCloak->GetCurrentState() == BEGIN_USING || pxClientCloak->GetCurrentState() == USING || pxClientCloak->GetDestinationState() != USING )
				{
					RequestNextUsingState();
				}
			}
			break;
		}

		case BEGIN_USING:
		{
			const u_int	uTransitionType	= GetCurrentTransitionType();

			if ( uTransitionType != TT_MANUAL || pxClientCloak == NULL || pxClientCloak->GetCurrentState() == USING || pxClientCloak->GetDestinationState() != USING )
			{
				RequestNextUsingState();
			}
			break;
		}

		case USING:
		{
			/*const AVP_ClientActor_Predator*	pxClientPredator	= static_cast< const AVP_ClientActor_Predator* >( pxClientActor );
			const AVP_Actor_DynamicStat*	pxDynamicEnergy		= pxClientPredator ? pxClientPredator->GetDynamicEnergy() : NULL;

			if ( pxDynamicEnergy->IsAtMinimum() )
			{
				// We need to go to automatic, since we haven't manually requested leaving the mode.
				RequestIdling(false, TT_AUTOMATIC);
			}*/
			break;
		}

		case END_USING:
		{
			const u_int	uTransitionType	= GetCurrentTransitionType();

			if ( uTransitionType != TT_MANUAL || pxClientCloak == NULL || pxClientCloak->GetCurrentState() == BEGIN_IDLING || pxClientCloak->GetCurrentState() == IDLING || pxClientCloak->GetDestinationState() != IDLING )
			{
				RequestNextIdlingState();
			}
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected state found when updating master-state of weapon.");
			break;
		}
	}		// End of switch.
}

/// End of function 'AVP_InventoryItem_Cloak::UpdateServerState'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::AVP_InventoryItem_Cloak
// Author       : Andy Tate
// Date         : 28-Nov-2007
////////////////////////////////////////////////////////////////////////////////

AVP_InventoryItem_Cloak::AVP_InventoryItem_Cloak(AVP_Inventory_Group& xParentGroup)
: PARENT(AVP_INVENTORYITEMTYPE_CLOAK, xParentGroup, uAVP_INVENTORY_ITEMHASH_CLOAK, uAVP_INVENTORY_ITEMHASH_CLOAK, szAVP_INVENTORY_ITEMNAME_CLOAK)
, m_fCloakedEnergyDrainStanding(2.0f)
, m_fCloakedEnergyDrainWalking(4.0f)
, m_fCloakedEnergyDrainRunning(20.0f)
, m_fCloakedEnergyDrainSprinting(50.0f)
, m_fCloakedEnergyDrainCrouching(1.0f)
, m_uAnimControllerHash( ASURA_HASH_ID_UNSET )
, m_iSoundHandle(ASURA_INVALID_HANDLE)
{
	AVP_Inventory_System&	xParentSystem	= GetParentSystem();

	// Initialise the valid states...
	if(xParentSystem.IsClientSide())
	{
		m_pxCloakEffect	= new AVP_CloakEffect;

		SetStateAsValid(BEGIN_IDLING);
		SetStateAsValid(END_IDLING);

		SetStateAsValid(BEGIN_USING);
		SetStateAsValid(END_USING);
	}
	else
	{
		m_pxCloakEffect	= NULL;
	}
	GetStaticDataFromBlueprint();

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if(!Asura_CommandConsole::VarExists(L"AVP.CloakEffect.Infinite"))
	{		
		Asura_CommandConsole::AddVar(L"AVP.CloakEffect.Infinite", &dvs_bInfiniteCloak);
	}
#endif //ASURA_DEBUG_VARIABLE_SYSTEM

	SetStateAsValid(USING);
}

/// End of function 'AVP_InventoryItem_Cloak::AVP_InventoryItem_Cloak'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::~AVP_InventoryItem_Cloak
// Author       : Andy Tate
// Date         : 8-Apr-2008
////////////////////////////////////////////////////////////////////////////////

AVP_InventoryItem_Cloak::~AVP_InventoryItem_Cloak()
{
	AVP_Inventory_System&	xParentSystem	= GetParentSystem();

	if(xParentSystem.IsClientSide())
	{
		AVP_ClientEntity_Actor*	pxClient		= xParentSystem.GetParentClient();
		Asura_Shape_Instance&	xShapeInstance	= pxClient->GetShapeInstance();
		Asura_Animation*		pxAnimation		= xShapeInstance.GetAnimation();

		if(pxAnimation != NULL)
		{
			AVP_Animation_ProjectData *pxProjectData = AVP_Animation_ProjectData::FromAnimation(pxAnimation);
			if (pxProjectData) pxProjectData->SetCloakEffect(NULL);
		}
	}

	delete(m_pxCloakEffect);

	Asura_Sound_Event_System::Stop(m_iSoundHandle);
}

/// End of function 'AVP_InventoryItem_Cloak::~AVP_InventoryItem_Cloak'. ///


void AVP_InventoryItem_Cloak::GetStaticDataFromBlueprint()
{
	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, GetBlueprintHash() );
	if ( pxBlueprint )
	{
		Asura_Value xValue;

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_DRAIN, xValue ) )
		{
			m_fCloakedEnergyDrainStanding = xValue.GetFloatValue();
		}
		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_DRAINWALKING, xValue ) )
		{
			m_fCloakedEnergyDrainWalking = xValue.GetFloatValue();
		}
		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_DRAINRUNNING, xValue ) )
		{
			m_fCloakedEnergyDrainRunning = xValue.GetFloatValue();
		}
		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_DRAINSPRINTING, xValue ) )
		{
			m_fCloakedEnergyDrainSprinting = xValue.GetFloatValue();
		}	
		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_DRAINCROUCHING, xValue ) )
		{
			m_fCloakedEnergyDrainCrouching = xValue.GetFloatValue();
		}	

		if ( m_pxCloakEffect )
		{
			if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_VISIBILITYDISTANCE, xValue ) )
			{
				m_pxCloakEffect->SetMaxVisibilityDistance(xValue.GetFloatValue());
			}	

			if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_CLOAK_VISIBILITYANGLE, xValue ) )
			{
				m_pxCloakEffect->SetMaxVisibilityAngle(xValue.GetFloatValue() * Asura_Maths::DegToRadConstant);
			}	
		}
	}
	else
	{
		Asura_Assertf( false, ( "AVP_InventoryItem_Cloak::GetStaticDataFromBlueprint - cannot find blueprint %u to initialise from.", GetBlueprintHash() ) );
	}
}

bool AVP_InventoryItem_Cloak::GetCanUseIdling(const bool bInstant, const u_int uTransitionType) const
{
	// With manual transitions, we have to check the same as we would for beginning using.
	if (uTransitionType == TT_MANUAL)
	{
		const AVP_Inventory_System& xParentSystem = GetParentSystem();

		const AVP_Inventory_Item* pxCurrentWeapon = xParentSystem.GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON);
		if (pxCurrentWeapon && !pxCurrentWeapon->IsIdling(true))
		{
			return false;
		}

		const AVP_Inventory_Item* pxStim = xParentSystem.GetItem(uAVP_BLUEPRINT_PARAM_STIM);
		if (pxStim && !pxStim->IsIdling(true))
		{
			return false;
		}
	}

	return PARENT::GetCanUseIdling(bInstant, uTransitionType);
}

bool AVP_InventoryItem_Cloak::GetCanUseUsageType(const u_int uUsageType, const u_int uTransitionType) const
{
	// Skip the parent definition, as we want to allow this during moves. What we need to check is the current weapon and the stim.
	const AVP_Inventory_System& xParentSystem = GetParentSystem();

	const AVP_Inventory_Item* pxCurrentWeapon = xParentSystem.GetCurrentItem(uAVP_INVENTORY_GROUPHASH_WEAPON);
	if (pxCurrentWeapon && !pxCurrentWeapon->IsInactive() && !pxCurrentWeapon->IsIdling(true))
	{
		return false;
	}

	const AVP_Inventory_Item* pxStim = xParentSystem.GetItem(uAVP_BLUEPRINT_PARAM_STIM);
	if (pxStim && !pxStim->IsInactive() && !pxStim->IsIdling(true))
	{
		return false;
	}

	return true;
}

DEBUGVAR float dbg_fCloakAnimTweenDuration = DEFAULT_TWEEN_TIME;

void AVP_InventoryItem_Cloak::PlayAnimation()
{
	const AVP_Inventory_System& xParentSystem = GetParentSystem();
	m_uAnimControllerHash = ASURA_HASH_ID_UNSET;

	Asura_Assert(xParentSystem.IsClientSide(), "Animations are a client operation.");

	AVP_ClientEntity_Actor* pxActor = xParentSystem.GetParentClient();
	if (!pxActor) { return; }

	if ( pxActor->GetCurrentAnimStateData() )
	{
		AVP_Vertebrate_AnimationHandler* pxAnimHandler = pxActor != NULL ? static_cast< AVP_Vertebrate_AnimationHandler* >( pxActor->GetAnimationHandler() ) : NULL;
		if ( pxAnimHandler != NULL )
		{
			const AVP_Actor_AnimStateData* pxStateData = pxActor->GetCurrentAnimStateData();
			if ( pxStateData )
			{
				const Asura_Hash_ID uActorAnim = pxStateData->GetCloakActorAnimHash();
				m_uAnimControllerHash = pxAnimHandler->PlayWeaponAnimation( uActorAnim, dbg_fCloakAnimTweenDuration, false );
				
				const Asura_Hash_ID uWeaponAnim = pxStateData->GetCloakWeaponAnimHash();
				if ( uWeaponAnim != ASURA_HASH_ID_UNSET )
				{
					AVP_Inventory_System* pxInvSys = pxActor->GetInventorySystem();
					AVP_InventoryItem_Weapon* pxWeapon = pxInvSys ? static_cast< AVP_InventoryItem_Weapon* >( pxInvSys->GetCurrentItem( uAVP_INVENTORY_GROUPHASH_WEAPON ) ) : NULL;
					if ( pxWeapon )
					{
						pxWeapon->PlayWeaponAnim( uWeaponAnim );
					}
				}
			}
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::ReadFromChunkStream
// Author       : Chris Wallace
// Date         : 04-Nov-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Cloak::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	u_int uVersion;
	xStream >> uVersion;

	if ( uVersion >= 1 )
	{
		PARENT::ReadFromChunkStream( xStream );
	}

	if ( uVersion == uCURRENT_CHUNK_VERSION || uVersion == 0 )
	{
		bool bInUse = false;
		xStream >> bInUse;
		if(bInUse)
		{
			if(m_xParentGroup.GetParentSystem().IsMaster())
			{
				RequestBeginUsage(UT_CLOAK, TT_INSTANT);
			}
			else
			{
				// Can't request on non-master, set it directly so can be correctly forwarded to master.
				m_uCurrentUsageType = UT_CLOAK;
				m_uCurrentTransitionType = TT_INSTANT;
				m_eCurrentState = USING;
			}
		}
	}
	else
	{
		Asura_Assertf( false, ( "AVP_InventoryItem_Cloak::ReadFromChunkStream - unhandled chunk version (%u)", uVersion ) );
	}
}

/// End of function 'AVP_InventoryItem_Cloak::ReadFromChunkStream'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Cloak::WriteToChunkStream
// Author       : Chris Wallace
// Date         : 04-Nov-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Cloak::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	// Firstly, write out the version...
	xStream << uCURRENT_CHUNK_VERSION;

	PARENT::WriteToChunkStream( xStream );

	xStream << IsUsing(true);
}

/// End of function 'AVP_InventoryItem_Cloak::WriteToChunkStream'. ///

/// End of file. ///
