
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_InventoryItem_Weapon.cpp
// Author : Andy Tate
// Date   : 13-Nov-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Inventory/Items/AVP_InventoryItem_Weapon.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Blueprint/Asura_Blueprint_System.h"

// Game independent header-files...
#include "Actor/AVP_Actor_AnimationHandler.h"
#include "Actor/AVP_Actor_AnimStateData.h"
#include "Actor/AVP_Actor_PickupHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_InputHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Focus.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Game_Core.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Inventory/AVP_Inventory_Group.h"
#include "Inventory/AVP_Inventory_System.h"
#include "Inventory/AVP_InventoryItem_SubType.h"
#include "Inventory/Items/AVP_InventoryItem_PredatorCloak.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const Asura_Hash_ID	g_uCameraMarkerPointHash	= 0x002FBC62;	//Asura_GetHashID("eyes");


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::GetClientWeapon
// Author       : Andy Tate
// Date         : 11-Dec-2007
////////////////////////////////////////////////////////////////////////////////

const AVP_InventoryItem_Weapon* AVP_InventoryItem_Weapon::GetClientWeapon(const AVP_ServerEntity_Actor& xServer) const
{
	const Asura_Hash_ID				uGroupHash			= m_xParentGroup.GetHashInParentSystem();
	const Asura_Hash_ID				uItemHash			= GetHashInParentGroup();

	const Asura_Guid				uOwnerGuid			= xServer.GetGuid();
	const AVP_ClientEntity_Actor*	pxClient			= AVP_ClientEntity_Actor::FindEntity(uOwnerGuid);
	const AVP_Inventory_System*		pxInventorySystem	= (pxClient == NULL) ? NULL : pxClient->GetInventorySystem();
	const AVP_Inventory_Group*		pxWeaponGroup		= (pxInventorySystem == NULL) ? NULL : pxInventorySystem->GetGroup(uGroupHash);
	const AVP_Inventory_Item*		pxCurrentItem		= (pxWeaponGroup == NULL) ? NULL : pxWeaponGroup->GetItem(uItemHash);
	const AVP_InventoryItem_Weapon*	pxCurrentWeapon		= static_cast<const AVP_InventoryItem_Weapon*>(pxCurrentItem);

	return(pxCurrentWeapon);
}

/// End of function 'AVP_InventoryItem_Weapon::GetClientWeapon'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::OnStateChange
// Author       : Andy Tate
// Date         : 22-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Weapon::OnStateChange(const u_int uOldState,
											 const u_int uNewState)
{
	AVP_Inventory_System&				xParentSystem		= GetParentSystem();
	AVP_ClientEntity_Actor*				pxClient			= xParentSystem.GetParentClient();
	AVP_Actor_AnimationHandler*			pxAnimationHandler	= pxClient == NULL ? NULL : pxClient->GetAnimationHandler();

	// Call the base-function...
	PARENT::OnStateChange(uOldState, uNewState);

	// Do our weapon-specific state-change work...
	if(pxAnimationHandler != NULL)
	{
		switch(uNewState)
		{
			case ACTIVATING:
			{
				// Attach weapon
				if ( m_eItemType != AVP_INVENTORYITEMTYPE_BATTERY )
				{
					pxAnimationHandler->AttachItem( this );
				}
				else
				{
					if ( pxClient->GetPickupHandler() )
					{
						pxClient->GetPickupHandler()->SetOverrideAimDirectionUsage( false );
					}
				}

				PlayActivationWeaponAnim();
				if(!pxClient->BeginWeaponActivation())
				{
					// Cannot activate this weapon - go straight to idling...
					RequestIdling(true);
				}
				break;
			}

			case IDLING:
			{
				const State	eSourceState	= GetSourceState();

				if ( eSourceState == INACTIVE )
				{
					pxClient->EndWeaponActivation();
				}
				break;
			}

			case END_IDLING:
			{
				const State	eDestinationState	= GetDestinationState();

				if( (eDestinationState == INACTIVE) )
				{
					PlayDeactivationWeaponAnim();
					if ( !pxClient->BeginWeaponDeactivation( GetDeactivationActorAnim(), IsEncumberment() ) )
					{
						RequestInactive(true, 0);
					}
				}
				break;
			}

			case INACTIVE:
			{
				AVP_Vertebrate_Focus* pxFocus = pxClient->GetFocus();
				if ( pxFocus != NULL )
				{
					pxFocus->UnSetLockOnTarget();
				}

				pxAnimationHandler->DetachItem( this );
				DestroyAnimation();

				pxClient->EndWeaponDeactivation();
				break;
			}

			default:
			{
				// Nothing to do here...
				break;
			}
		}		// End of switch.
	}
}

/// End of function 'AVP_InventoryItem_Weapon::OnStateChange'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::UpdateClientState
// Author       : Andy Tate
// Date         : 22-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Weapon::UpdateClientState(AVP_ClientEntity_Actor& xClient)
{
	AVP_Actor_AnimationHandler*	pxAnimationHandler	= xClient.GetAnimationHandler();

	if(pxAnimationHandler == NULL)
	{
		Asura_Assert(false, "Cannot find animation handler for this client.");
	}
	else
	{
		const Asura_Animation*	pxAnimation		= pxAnimationHandler->GetAnimation();
		const u_int				uAnimFlags		= pxAnimation->GetFlagsForAllAnimsLastFrame();
		const State				eCurrentState	= GetCurrentState();

		// Decide what to do based on our current state...
		switch(eCurrentState)
		{
			case INACTIVE:
			{
				break;
			}

			case ACTIVATING:
			{
				// Attach Battery
				if ( m_eItemType == AVP_INVENTORYITEMTYPE_BATTERY )
				{
					if ( ( uAnimFlags & AVP_ANIMFLAG_ACTION_1) != 0 )
					{
						pxAnimationHandler->AttachItem( this );
					}
				}
				// Wait till extended
				if ( GetTimeInState() > 0.0f && ( ( uAnimFlags & AVP_ANIMFLAG_COMPLETE) != 0 || !pxAnimationHandler->IsActivateAnimPlaying() ) )
				{
					RequestNextIdlingState();
				}
				break;
			}

			case BEGIN_IDLING:
			{
				RequestNextIdlingState();
				break;
			}

			case IDLING:
			{
				if ( !pxAnimationHandler->IsItemAttached( this ) && !IsHidden() )
				{
					pxAnimationHandler->AttachItem( this );
				}
				break;
			}

			case END_IDLING:
			{
				const State	eDestinationState	= GetDestinationState();

				if(eDestinationState == INACTIVE)
				{
					RequestNextInactiveState();
				}
				else
				{
					RequestNextUsingState();
				}
				break;
			}

			case USING:
			{
				const AVP_Inventory_System&	xParentSystem	= GetParentSystem();

				if(xParentSystem.IsMaster() && xClient.IsDead())
				{
					RequestIdling(false, TT_DEAD);
				}
				break;
			}

			case DEACTIVATING:
			{
				if ( GetTimeInState() > 0.0f && ( ( uAnimFlags & AVP_ANIMFLAG_COMPLETE ) != 0 || !pxAnimationHandler->IsDeactivateAnimPlaying() ) )
				{
					RequestNextInactiveState();
				}
				break;
			}

			default:
			{
				break;
			}
		}		// End of switch.
	}
}

/// End of function 'AVP_InventoryItem_Weapon::UpdateClientState'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::UpdateServerState
// Author       : Andy Tate
// Date         : 22-Nov-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Weapon::UpdateServerState(AVP_ServerEntity_Actor& xServer)
{
	const State	eCurrentState	= GetCurrentState();

	const AVP_ClientEntity_Actor* pxClient = AVP_ClientEntity_Actor::FindEntity( GetParentSystem().GetOwnerGuid() );
	const AVP_Actor_AnimationHandler*	pxAnimationHandler	= pxClient ? pxClient->GetAnimationHandler() : NULL;
	const Asura_Animation*	pxAnimation		= pxAnimationHandler ? pxAnimationHandler->GetAnimation() : NULL;
	const u_int				uAnimFlags		= pxAnimation ? pxAnimation->GetFlagsForAllAnimsLastFrame() : 0;

	// Decide what to do based on our current state...
	switch(eCurrentState)
	{
		case INACTIVE:
		{
			break;
		}

		case ACTIVATING:
		{
			// Wait till extended
			if ( GetTimeInState() > 0.0f && ( ( uAnimFlags & AVP_ANIMFLAG_COMPLETE) != 0 || !pxAnimationHandler->IsActivateAnimPlaying() ) )
			{
				RequestNextIdlingState();
			}
			break;
		}


		case BEGIN_IDLING:
		{
			const AVP_InventoryItem_Weapon*	pxCurrentWeapon	= GetClientWeapon(xServer);

			if( (pxCurrentWeapon != NULL) &&
				( pxCurrentWeapon->IsIdling(true) || pxCurrentWeapon->GetDestinationState() != IDLING ) )
			{
				RequestNextIdlingState();
			}
			break;
		}

		case IDLING:
		{
			break;
		}

		case BEGIN_USING:
		{
			const AVP_InventoryItem_Weapon*	pxCurrentWeapon	= GetClientWeapon(xServer);

			if( (pxCurrentWeapon != NULL) &&
				( pxCurrentWeapon->IsUsing(true) || pxCurrentWeapon->GetDestinationState() != USING ) )
			{
				RequestNextUsingState();
			}
			break;
		}

		case USING:
		{
			const AVP_Inventory_System&	xParentSystem	= GetParentSystem();

			if(xParentSystem.IsMaster() && xServer.IsDead())
			{
				RequestIdling(false, TT_DEAD);
			}
			break;
		}

		case DEACTIVATING:
		{
			if ( m_eItemType == AVP_INVENTORYITEMTYPE_BATTERY )
			{
				if ( ( uAnimFlags & AVP_ANIMFLAG_ACTION_1) != 0 )
				{
					const AVP_ServerEntity_Actor* pxServer = AVP_ServerEntity_Actor::FindEntity( GetParentSystem().GetOwnerGuid() );
					if( pxServer )
					{
						pxServer->DropTool();
					}
				}
			}
			if ( GetTimeInState() > 0.0f && ( ( uAnimFlags & AVP_ANIMFLAG_COMPLETE ) != 0 || !pxAnimationHandler->IsDeactivateAnimPlaying() ) )
			{
				RequestNextInactiveState();
			}
			break;
		}

		default:
		{
			break;
		}
	}		// End of switch.
}

/// End of function 'AVP_InventoryItem_Weapon::UpdateServerState'. ///


bool AVP_InventoryItem_Weapon::GetCanUseUsageType(const u_int uUsageType, const u_int uTransitionType) const
{
	// Weapons will never want to fire if the stim or cloak are in use.
	const AVP_Inventory_System& xParentSystem = GetParentSystem();

	const AVP_Inventory_Item* pxStim = xParentSystem.GetItem(uAVP_BLUEPRINT_PARAM_STIM);
	if (pxStim && !pxStim->IsIdling(true))
	{
		return false;
	}

	const AVP_Inventory_Item* pxCloak = xParentSystem.GetItem(uAVP_INVENTORY_ITEMHASH_CLOAK);
	if (pxCloak)
	{
		// We need to see if we're in a transition
		const bool bIdling = pxCloak->IsIdling(false);
		const bool bFullyIdling = pxCloak->IsIdling(true);
		const bool bUsing = pxCloak->IsUsing(false);
		const bool bFullyUsing = pxCloak->IsUsing(true);

		if (bIdling && !bFullyIdling)
		{
			// This is still allowed if the transition type is automatic.
			if (pxCloak->GetCurrentTransitionType() != AVP_InventoryItem_Cloak::TT_AUTOMATIC)
			{
				return false;
			}
		}

		if (bUsing && !bFullyUsing)
		{
			return false;
		}
	}

	if (uUsageType == AVP_InventoryItem_Weapon::UT_MELEE)
	{
		// Nasty - the melee usage type needs to flag the weapon in use if a move successfully starts, so we have to let that through in that case.
		return true;
	}

	// we only want to use weapons when they're not doing something else
	if ( !IsIdling( true ) )
	{
		return false;
	}

	return PARENT::GetCanUseUsageType(uUsageType, uTransitionType);
}

////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::AVP_InventoryItem_Weapon
// Author       : Andy Tate
// Date         : 13-Nov-2007
////////////////////////////////////////////////////////////////////////////////

AVP_InventoryItem_Weapon::AVP_InventoryItem_Weapon(AVPInventoryItemType eType, 
												   AVP_Inventory_Group& xParentGroup,
												   const Asura_Hash_ID uHash,
												   const Asura_Hash_ID uBlueprintHash,
												   Asura_Char const* const szName
)
: PARENT(eType, xParentGroup, uHash, uBlueprintHash, szName)
, m_bSubtypesUseSameAmmo( false )
, m_fADSFOVDecrease(0.0f)
, m_fADSNearClipDistance( 0.02f )
, m_fLockRange( 0.0f )
, m_fLockTime( 1.0f )
, m_fSqrtWeaponRecoilReturnFactor( 5.0f )
, m_fRecoilUpDownFactor( 1.0f )
, m_iFixedSlot( -1 )
, m_uFlags(0)
{
	AVP_Inventory_System&	xParentSystem	= GetParentSystem();

	// Initialise the valid states...
	if(xParentSystem.IsClientSide())
	{
		SetStateAsValid(ACTIVATING);
		SetStateAsValid(BEGIN_IDLING);
		SetStateAsValid(END_IDLING);
		SetStateAsValid(DEACTIVATING);
	}
	else
	{
		SetStateAsValid(ACTIVATING);
		SetStateAsValid(BEGIN_IDLING);
		SetStateAsValid(BEGIN_USING);
		SetStateAsValid(DEACTIVATING);
	}
	SetStateAsValid(USING);

	SetBlueprintHash(uBlueprintHash);
	GetStaticDataFromBlueprint();
}

/// End of function 'AVP_InventoryItem_Weapon::AVP_InventoryItem_Weapon'. ///

AVP_InventoryItem_Weapon::~AVP_InventoryItem_Weapon()
{
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::CreateDebugString
// Author       : Andy Tate
// Date         : 6-Dec-2007
////////////////////////////////////////////////////////////////////////////////

u_int AVP_InventoryItem_Weapon::CreateDebugString(const u_int uMaxChars,
												  char* szText) const
{
	u_int	uOffset	= PARENT::CreateDebugString(uMaxChars, szText);

	// Add each sub-type in-turn...
	for(u_int u = 0; u < m_uNumSubTypes; u++)
	{
		const AVP_InventoryItem_SubType*	pxSubType	= GetSubType(u);

		if(pxSubType != NULL)
		{
			uOffset	+= Asura_PrintfToBfr(&szText[uOffset], " - %.0f/%.0f", uMaxChars - uOffset, true, pxSubType->GetCurrentQuantity(), pxSubType->GetMaxQuantity());
		}
	}

	return(uOffset);
}

/// End of function 'AVP_InventoryItem_Weapon::CreateDebugString'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::Update
// Author       : Chris Wallace
// Date         : 11-Dec-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Weapon::Update(void)
{
	PARENT::Update();
}

/// End of function 'AVP_InventoryItem_Weapon::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::Render
// Author       : Chris Wallace
// Date         : 10-Dec-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_InventoryItem_Weapon::Render(void) const
{
}

/// End of function 'AVP_InventoryItem_Weapon::Render'. ///





////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::CalculateFiringData
// Author       : Andy Tate
// Date         : 27-Nov-2007
////////////////////////////////////////////////////////////////////////////////

float AVP_InventoryItem_Weapon::CalculateFiringData(const AVP_ClientEntity_Actor& xClient,
													const Asura_Hash_ID uMuzzleMarkerPointHash,
													const float fMaxRange,
													Asura_Vector_3& xFirePosition,
													Asura_Matrix_3x3& xFireOrientation)
{
	const Asura_Guid	uOwnerGuid	= xClient.GetGuid();

	// Check whether we can get this marker-point from the client...
	if(xClient.GetWorldSpaceMarkerPoint(uMuzzleMarkerPointHash, &xFirePosition, &xFireOrientation))
	{
		Asura_Vector_3				xFireDirection;
		Asura_AVector3				xTemp;
		xClient.GetAimDirection( xTemp );
		xFireDirection = Asura_Vector_3( xTemp );
		Asura_Maths::OrientationFromDirection( xFireOrientation, xFireDirection );
		const Asura_LOS_InputData	xInputLOS(xFirePosition, xFireDirection, uOwnerGuid, fMaxRange);
		const Asura_LOS_OutputData*	pxLOSOutputData	= Asura_LOS::TestAgainstEverything(&xInputLOS);
		const float					fRange			= pxLOSOutputData->m_fDistanceToIntersection;

		xFirePosition	+= (xFireDirection * fRange);
		return(fRange);
	}
	else
	{
		return(fMaxRange);
	}
}

/// End of function 'AVP_InventoryItem_Weapon::CalculateFiringData'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::CalculateFiringDataForCamera
// Author       : Andy Tate
// Date         : 11-Dec-2007
////////////////////////////////////////////////////////////////////////////////

float AVP_InventoryItem_Weapon::CalculateFiringDataForCamera(const AVP_ClientEntity_Actor& xClient,
															 const Asura_Hash_ID uMuzzleMarkerPointHash,
															 const float fMaxRange,
															 Asura_Vector_3& xFirePosition,
															 Asura_Matrix_3x3& xFireOrientation)
{
	Asura_Vector_3		xTargetPosition;
	Asura_Matrix_3x3	xCameraOrientation;

	float				fRange	= CalculateFiringData(xClient, g_uCameraMarkerPointHash, fMaxRange, xTargetPosition, xCameraOrientation);

	// Use this data even if we miss, otherwise binks badly when looking at the sky.

	if(xClient.GetWorldSpaceMarkerPoint(uMuzzleMarkerPointHash, &xFirePosition, NULL))
	{
		Asura_Vector_3	xFireDirection(xTargetPosition - xFirePosition);

		fRange			 = xFireDirection.Magnitude();
		xFireDirection	/= fRange;

		Asura_Maths::OrientationFromDirection(xFireOrientation, xFireDirection);
		return(fRange);
	}
	else
	{
		Asura_Assert(false, "Cannot find marker-point for weapon muzzle.");
		
		xFirePosition = xClient.GetCurrentPosition();
		xFireOrientation = xClient.GetCurrentOrientation();
		
		return(fMaxRange);
	}
}

/// End of function 'AVP_InventoryItem_Weapon::CalculateFiringDataForCamera'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_InventoryItem_Weapon::CalculateFiringDataForMovingTarget
// Author       : Andy Tate
// Date         : 3-Jan-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_InventoryItem_Weapon::CalculateFiringDataForMovingTarget(const AVP_ClientEntity_Actor& xClient,
																  const Asura_Hash_ID uSourceMarkerPointHash,
																  const float fInitialSpeed,
																  const Asura_Vector_3& xTargetPosition,
																  const Asura_Vector_3& xTargetVelocity,
																  Asura_Vector_3& xFirePosition,
																  Asura_Matrix_3x3& xFireOrientation,
																  Asura_Vector_3& xOutputFireVelocity)
{
	if(xClient.GetWorldSpaceMarkerPoint(uSourceMarkerPointHash, &xFirePosition, &xFireOrientation))
	{
		Asura_Vector_3	xTargetDistance(xTargetPosition - xFirePosition);

		if(xTargetVelocity == Asura_Maths::ZeroVector3)
		{
			xOutputFireVelocity	= xTargetDistance * (fInitialSpeed / xTargetDistance.Magnitude());
		}
		else
		{
			CalculateFiringDataForMovingTarget( xOutputFireVelocity, xFirePosition, fInitialSpeed, xTargetPosition, xTargetVelocity, 0.0f );
		}

		return(true);
	}
	else
	{
		return(false);
	}
}

void AVP_InventoryItem_Weapon::CalculateFiringDataForMovingTarget( Asura_Vector_3& xOutputVel, const Asura_Vector_3& xSourcePos, float fSourceSpeed, const Asura_Vector_3& xTargetPos, const Asura_Vector_3& xTargetVel, float fLeadingTargetHack )
{
	// Because the order of laser bolt / entity update isn't guaranteed we may need a fudge here to adjust for the order of update.
	// In general it seems that extrapolating the targets position by n frames (usually 1) corrects the error.
	// In an ideal world the entity would update then the laser bolts and this hack could be 0.

	Asura_Vector_3 xNewTargetPos = xTargetPos + xTargetVel * fLeadingTargetHack * Asura_Timers::GetGameTimeSlice();

	Asura_Vector_3 xFromShooter = xNewTargetPos - xSourcePos;
	const float fTargetSpeed = xTargetVel.Magnitude();
	const float fDistance    = xFromShooter.Magnitude();

	if ( fSourceSpeed > 0.005f && fTargetSpeed > 0.005f && fDistance > 0.005f )
	{
		float fBeta = ( -xFromShooter * xTargetVel ) / ( fDistance * fTargetSpeed );
		if ( fBeta <= 1.0f )
		{
			fBeta = Asura_Maths	::ArcCosine( fBeta );
		}
		else
		{
			fBeta = 0.0f;
		}

		const float fSinBeta = Asura_Maths::Sine( fBeta );
		const float fAlpha   = Asura_Maths::ArcSine( Asura_Maths::ClampToRange( ((fTargetSpeed / fSourceSpeed) * fSinBeta), -1.0f, 1.0f ) );
		const float fGamma	 = Asura_Maths::Deg180 - ( fAlpha + fBeta );
		const float fSineGamma = Asura_Maths::Sine( fGamma );

		if( (fSineGamma > 0.0001f) || (fSineGamma < -0.0001f) )  
		{
			const float fTimeOfFlight = ( fSinBeta / fSineGamma ) * ( fDistance / fSourceSpeed );
			// In theory flight time shouldn't be negative, in reality we may well be firing inside another entity
			// so leading it has little meaning in that case.
			if (Asura_Maths::Abs(fTimeOfFlight) > 0.0001f) // Sensible Time of Flight
			{
				xOutputVel = xFromShooter + ( xTargetVel * fTimeOfFlight );

				return;
			}
		}
	}
	// failure, just return target position.
	xOutputVel = (xTargetPos - xSourcePos);
}


Asura_Hash_ID AVP_InventoryItem_Weapon::GetActivationActorAnim() const
{
	const AVP_ClientEntity_Actor* pxClient = GetParentSystem().GetParentClient();
	if ( pxClient )
	{
		const AVP_Actor_AnimStateData* pxStateData = pxClient->GetAnimStateDataForWeapon( GetBlueprintHash() );
		if ( pxStateData )
		{
			return pxStateData->GetWeaponInActorAnimHash();
		}
	}
	// else
	return ASURA_HASH_ID_UNSET;
}

Asura_Hash_ID AVP_InventoryItem_Weapon::GetDeactivationActorAnim() const
{
	const AVP_ClientEntity_Actor* pxClient = GetParentSystem().GetParentClient();
	if ( pxClient )
	{
		const AVP_Actor_AnimStateData* pxStateData = pxClient->GetAnimStateDataForWeapon( GetBlueprintHash() );
		if ( pxStateData )
		{
			return pxStateData->GetWeaponOutActorAnimHash();
		}
	}
	// else
	return ASURA_HASH_ID_UNSET;
}

void AVP_InventoryItem_Weapon::GetMeleeBlueprintIDs(Asura_Hash_ID& uFirstPersonID, Asura_Hash_ID& uThirdPersonID) const
{
	uFirstPersonID = uThirdPersonID = ASURA_HASH_ID_UNSET;

	const Asura_Blueprint* pxBlueprint = GetBlueprint();
	if (!pxBlueprint) { return; }

	// TODO - are we better off storing these values?
	Asura_Value xValue;
	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_MELEE_FIRSTPERSON, xValue))
	{
		uFirstPersonID = xValue.GetHashIDValue();
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_MELEE_THIRDPERSON, xValue))
	{
		uThirdPersonID = xValue.GetHashIDValue();
	}
}

Asura_Hash_ID AVP_InventoryItem_Weapon::GetGrappleBlueprintID() const
{
	const Asura_Blueprint* pxBlueprint = GetBlueprint();
	if (pxBlueprint)
	{
		Asura_Value xValue;
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLESET, xValue))
		{
			return xValue.GetHashIDValue();
		}
	}

	return ASURA_HASH_ID_UNSET;
}

bool AVP_InventoryItem_Weapon::IsSelectable(void) const
{
	if(AVP_Settings::GetInfiniteAmmo()) return true; // Always selectable if ammo infinite

	return(PARENT::IsSelectable());
}

void AVP_InventoryItem_Weapon::GetStaticDataFromBlueprint()
{
	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_INVENTORY_ITEMS, GetBlueprintHash() );
	if ( pxBlueprint )
	{
		Asura_Value xValue;
		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_ADS_FOV_DECREASE, xValue ) )
		{
			m_fADSFOVDecrease = Asura_Maths::DegToRad( xValue.GetFloatValue() );
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_ADS_NEAR_CLIP_PLANE, xValue ) )
		{
			m_fADSNearClipDistance = xValue.GetFloatValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_USES_LOCKON, xValue ) )
		{
			m_bUsesLockOn = xValue.GetBoolValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_LOCKON_RANGE, xValue ) )
		{
			m_fLockRange = xValue.GetFloatValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_LOCKON_TIME, xValue ) )
		{
			m_fLockTime = xValue.GetFloatValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_WEAPON_SLOTINDEX, xValue ) )
		{
			m_iFixedSlot = xValue.GetIntValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_HASCROSSHAIR, xValue ) )
		{
			m_bHasCrossHair = xValue.GetBoolValue();
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_WEAPONRECOILRETURNTIME, xValue ) )
		{
			float fVal = xValue.GetFloatValue();
			if ( fVal > 0.2f )
			{
				m_fSqrtWeaponRecoilReturnFactor = 5.0f / ( xValue.GetFloatValue() ); // The factor of 5.0 here maps the falloff time in the blueprint to the 5 sigma falloff time. i.e. the time it takes to get 1-e^-5 ~ 99% of the way back to the target.
			}
			else
			{
				m_fSqrtWeaponRecoilReturnFactor = 25.0f;
			}
		}

		if ( pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_PARAM_RECOILUPDOWNFACTOR, xValue ) )
		{
			m_fRecoilUpDownFactor = xValue.GetFloatValue();
		}
	}
	else
	{
		Asura_Assertf( false, ( "AVP_InventoryItem_Weapon::GetStaticDataFromBlueprint - cannot find blueprint %u to initialise from.", GetBlueprintHash() ) );
	}
}

/// End of file. ///
