
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_InventoryItem_Weapon.h
// Author : Andy Tate
// Date   : 13-Nov-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_INVENTORYITEM_WEAPON_H__
#define __AVP_INVENTORYITEM_WEAPON_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"

#include "Entity/AVP_Entity_Classification.h"
#include "Inventory/AVP_Inventory_Item.h"
#include "Inventory/Groups/AVP_InventoryGroup_Weapons.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class AVP_Predator_TagHandler;
class AVP_Predator_Tag;

////////////////////////////////////////////////////////////////////////////////
// Class declaration...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_InventoryItem_Weapon
// Author		: Andy Tate
// Date			: 13-Nov-2007
//
// Description  : More specific vertebrate inventory weapon class, for common 
//				  code sharable between inventory weapons.
////////////////////////////////////////////////////////////////////////////////

class AVP_InventoryItem_Weapon : public AVP_Inventory_Item
{
public:
	// Public enumerated types...
	enum WeaponUsageType
	{
		UT_RELOAD_ALL	= NUM_ITEM_USAGE_TYPES,
		UT_MELEE,

		NUM_WEAPON_USAGE_TYPES
	};

	enum WeapoonTransitionType
	{
		TT_DEAD	= NUM_ITEM_TRANSITION_TYPES,

		NUM_WEAPON_TRANSITION_TYPES
	};

	// Public functions...
	AVP_InventoryItem_Weapon(AVPInventoryItemType eType, AVP_Inventory_Group& xParentGroup, const Asura_Hash_ID uHash, const Asura_Hash_ID uBlueprintHash, Asura_Char const* const szName);
	virtual ~AVP_InventoryItem_Weapon();

	// Overridden functions...
	virtual	void	Update(void);
	virtual void	Render(void) const;

	// Abstract manipulation functions...
	virtual void	StartUsing(const u_int uUsageType)	= 0;
	virtual void	Use(void)							= 0;
	virtual void	StopUsing(const u_int uUsageType)	= 0;

	// Debugging functions...
	virtual u_int	CreateDebugString(const u_int uMaxChars, char* szText) const;

	// Targeting functions...
	static float	CalculateFiringData(const AVP_ClientEntity_Actor& xClient, const Asura_Hash_ID uMuzzleMarkerPointHash, const float fMaxRange, Asura_Vector_3& xFirePosition, Asura_Matrix_3x3& xFireOrientation);
	static float	CalculateFiringDataForCamera(const AVP_ClientEntity_Actor& xClient, const Asura_Hash_ID uMuzzleMarkerPointHash, const float fMaxRange, Asura_Vector_3& xFirePosition, Asura_Matrix_3x3& xFireOrientation);
	static bool		CalculateFiringDataForMovingTarget(const AVP_ClientEntity_Actor& xClient, const Asura_Hash_ID uSourceMarkerPointHash, const float fInitialSpeed, const Asura_Vector_3& xTargetPosition, const Asura_Vector_3& xTargetVelocity, Asura_Vector_3& xFirePosition, Asura_Matrix_3x3& xFireOrientation, Asura_Vector_3& xOutputFireVelocity);
	static void		CalculateFiringDataForMovingTarget( Asura_Vector_3& xOutputPos, const Asura_Vector_3& xSourcePos, float fSourceSpeed, const Asura_Vector_3& xTargetPos, const Asura_Vector_3& xTargetVel, float fLeadingTargetHack );

	// Accessor functions...
	bool						GetHasCrossHair(void) const	{ return m_bHasCrossHair; }
	bool						GetUsesLockOn(void) const	{ return m_bUsesLockOn; }

	float						GetLockOnRange() const { return m_fLockRange; }
	float						GetLockOnTime() const { return m_fLockTime; }

	float						GetADSFOVDecrease() const	{ return m_fADSFOVDecrease; }
	float						GetADSNearClipDistance() const { return m_fADSNearClipDistance; }

	float						GetSqrtWeaponRecoilReturnFactor() const { return m_fSqrtWeaponRecoilReturnFactor; }
	float						GetRecoilUpDownFactor() const { return m_fRecoilUpDownFactor; }

	virtual void				PlayActivationWeaponAnim() {}
	virtual void				PlayDeactivationWeaponAnim() {}

	Asura_Hash_ID				GetActivationActorAnim() const;
	Asura_Hash_ID				GetDeactivationActorAnim() const;

	virtual float				GetSpreadParametric() const { return 0.0f; }
	virtual float				GetPrimaryMinSpread() const { return 5.0f; }
	virtual float				GetPrimaryMaxSpread() const { return 5.0f; }

	virtual float				GetSpeedRestriction() const { return 2.0f; } // no restriction

	bool GetSubtypesUseSameAmmo(void) const { return m_bSubtypesUseSameAmmo; }

	void GetMeleeBlueprintIDs(Asura_Hash_ID& uFirstPersonID, Asura_Hash_ID& uThirdPersonID) const;
	Asura_Hash_ID GetGrappleBlueprintID() const;

	virtual int GetFixedSlot() const { return m_iFixedSlot; }

	virtual	bool	IsSelectable(void) const;
	virtual void	PlayWeaponAnim( Asura_Hash_ID uHash ) {}
	virtual void	ProcessPlayerInput( const AVP_ClientEntity_Actor& xActor ) {};

	virtual bool GetCanUseUsageType(const u_int uUsageType, const u_int uTransitionType) const;

protected:
	bool  m_bSubtypesUseSameAmmo;

	float m_fADSFOVDecrease;
	float m_fADSNearClipDistance;
	float m_fLockRange;
	float m_fLockTime;
	float m_fSqrtWeaponRecoilReturnFactor;
	float m_fRecoilUpDownFactor;
	int	  m_iFixedSlot;

	const AVP_InventoryItem_Weapon*	GetClientWeapon(const AVP_ServerEntity_Actor& xServer) const;

	virtual void OnStateChange(const u_int uOldState, const u_int uNewState);
	virtual void UpdateClientState(AVP_ClientEntity_Actor& xClient);
	virtual	void UpdateServerState(AVP_ServerEntity_Actor& xServer);

private:
	typedef AVP_Inventory_Item PARENT;

	void GetStaticDataFromBlueprint();

	union
	{
		u_int		m_uFlags;				// Single access-point for.

		struct
		{
			u_int	m_bHasCrossHair		: 1,	// Flag to say whether this weapon needs a crosshair.
					m_bUsesLockOn		: 1,	// uses lock on mechanic
										: 0;	// 30 spare-bits.
		};
	};

	AVP_InventoryItem_Weapon(void);
	AVP_InventoryItem_Weapon(const AVP_InventoryItem_Weapon& xOther);
	AVP_InventoryItem_Weapon&	operator=	(const AVP_InventoryItem_Weapon& xOther);
};

#endif   // ifndef __AVP_INVENTORYITEM_WEAPON_H__

/// End of file. ///
