////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
// AVP_Profile_LoadAndSave.cpp
// Thomas Young / Callum Hope - 8,1,09
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// File functionality to deal with saving data to file on any platform
// No actual profile functionality, just profile reading saving, loading, and writing of data used in profiles. 
// Essentially this whole file is just a platform independent back bone to the save load system, 
// Its very different on all platforms so this is an interface to the three different methods; PC, PS3 and 360
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

// I N C L U D E S /////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined(AVP_SUPPORT_PROFILE_LOADANDSAVE)

// personal header
#include "LoadAndSave/AVP_Profile_LoadAndSave.h"

// asura headers
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_CommandConsole.h"

// avp headers

// G L O B A L / S T A T I C   V A R I A B L E S ///////////////////////////////

// static initialisations from within class. 
AVP_Profile*					AVP_Profile_LoadAndSave::s_pxProfile	= NULL;
AVP_LOAD_SAVE_STATES			AVP_Profile_LoadAndSave::s_eLoadingState = AVP_LOAD_SAVE_STATES_IDLE;
u_int							AVP_Profile_LoadAndSave::s_uProfileSaveSize = 0;
u_int							AVP_Profile_LoadAndSave::s_uCheckpointSize = 0;
u_int							AVP_Profile_LoadAndSave::s_uSlot = 0;
Asura_Chunk_Memory_Stream*		AVP_Profile_LoadAndSave::s_pxProfileStream = NULL;
Asura_Chunk_Memory_Stream*		AVP_Profile_LoadAndSave::s_pxCheckpointStream = NULL;
Asura_StringBuffer<128>			AVP_Profile_LoadAndSave::s_xCheckpointDisplayName[NUM_CPOINT_SAVESLOTS];
bool							AVP_Profile_LoadAndSave::s_bRequestPending = false;
AVP_LOAD_SAVE_STATES			AVP_Profile_LoadAndSave::s_ePendingLoadingState = AVP_LOAD_SAVE_STATES_IDLE;

// F U N C T I O N   D E F I N I T I O N S /////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Initialise
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 14/1/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Initialise(AVP_Profile* pxProfile)
{
	s_pxProfile	= pxProfile;

	// Write once to a stream to get the size of the profile we are saving
	Asura_Chunk_Memory_Stream xStream;
	s_pxProfile->WriteToChunkStream( xStream );
	s_uProfileSaveSize = xStream.GetPosition();

	Platform_Initialise();

#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	// Register our console commands
	//Asura_CommandConsole::AddCmd ( L"Profile.BeginProfileFlow", ConsoleCmd_BeginProfileFlow, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	
	if( !Asura_CommandConsole::CmdExists( L"AVP.Checkpoint.DisplayName" ) )
	{
		Asura_CommandConsole::AddCmd( L"AVP.Checkpoint.DisplayName", GetDisplayName, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	}
	if( !Asura_CommandConsole::CmdExists( L"AVP.Checkpoint.Load" ) )
	{
		Asura_CommandConsole::AddCmd( L"AVP.Checkpoint.Load", LoadCheckpoint, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	}
	if( !Asura_CommandConsole::CmdExists( L"AVP.Checkpoint.Save" ) )
	{
		Asura_CommandConsole::AddCmd( L"AVP.Checkpoint.Save", SaveCheckpoint, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	}
#endif
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::BeginProfileFlow
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09 
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::BeginProfileFlow( )
{
	if( s_pxProfileStream != NULL )
	{
		Asura_Assert( false, "s_pxProfileStream != NULL in Platform_BeginProfileFlow, possible memory leak!");
		delete s_pxProfileStream;
		s_pxProfileStream = NULL;
	}
	s_pxProfileStream = new Asura_Chunk_Memory_Stream;

	s_eLoadingState = AVP_LOAD_SAVE_STATES_PROFILEFLOWREQUESTED;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Update
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Update( )
{
	Platform_Update();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::Shutdown
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

void AVP_Profile_LoadAndSave::Shutdown( )
{
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	if( Asura_CommandConsole::CmdExists( L"AVP.Checkpoint.DisplayName" ) )
	{
		Asura_CommandConsole::RemoveCmd ( L"AVP.Checkpoint.DisplayName" );
	}
	if( Asura_CommandConsole::CmdExists( L"AVP.Checkpoint.Load" ) )
	{
		Asura_CommandConsole::RemoveCmd ( L"AVP.Checkpoint.Load" );
	}
	if( Asura_CommandConsole::CmdExists( L"AVP.Checkpoint.Save" ) )
	{
		Asura_CommandConsole::RemoveCmd ( L"AVP.Checkpoint.Save" );
	}
#endif

	Platform_Shutdown();

	if( s_pxProfileStream != NULL )
	{
		delete s_pxProfileStream;
	}
	if( s_pxCheckpointStream != NULL )
	{
		delete s_pxCheckpointStream;
	}
	s_pxProfile	= NULL;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::SaveProfileData
// Scope		: Public Static 
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Sets the state to load checkpoint if we are in idle mode, otherwise says that we want to load as soon as we can.
// Return		: Bool to signify a successful (or not) save. 
////////////////////////////////////////////////////////////////////////////////

bool AVP_Profile_LoadAndSave::SaveProfileData()
{
	if( s_eLoadingState == AVP_LOAD_SAVE_STATES_IDLE )
	{
		if( s_pxProfileStream != NULL )
		{
			Asura_Assert( false, "s_pxProfileStream != NULL in SaveProfileData, possible memory leak!");
			delete s_pxProfileStream;
			s_pxProfileStream = NULL;
		}
		s_pxProfileStream = new Asura_Chunk_Memory_Stream( s_uProfileSaveSize );
		s_pxProfile->WriteToChunkStream( *(s_pxProfileStream) );

		s_eLoadingState = AVP_LOAD_SAVE_STATES_PROFILESAVEREQUESTED;
	}
	else
	{
		s_bRequestPending = true;
		s_ePendingLoadingState = AVP_LOAD_SAVE_STATES_PROFILESAVEREQUESTED;
	}
	return (s_pxProfileStream != NULL );
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::LoadProfileData
// Scope		: Public Static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Call the correct platform function to load the profile data for the current profile being used. 
// Return		: Returns if the file exists or not, since this is mostly used to say if we need to create a new file or not (used for the Xbox360)
////////////////////////////////////////////////////////////////////////////////

AVP_Profile::E_PROFILE_FILE_STATES AVP_Profile_LoadAndSave::LoadProfileData()
{
	// Only used to recreate the profile file if it doesn't exist
	Asura_StringBuffer<128> xFilename;
	Platform_GetCurrentProfileFileName( xFilename, 0 );	
	
	if( !Asura_FileAccess::FileExists( xFilename.GetString() ) )
	{
		return AVP_Profile::PROFILE_FILE_NOT_FOUND;
	}
	if( s_eLoadingState == AVP_LOAD_SAVE_STATES_IDLE )
	{
		if( s_pxProfileStream != NULL )
		{
			Asura_Assert( false, "s_pxProfileStream != NULL in LoadProfileData, possible memory leak!");
			delete s_pxProfileStream;
			s_pxProfileStream = NULL;
		}
		s_pxProfileStream = new Asura_Chunk_Memory_Stream;
		
		s_eLoadingState = AVP_LOAD_SAVE_STATES_PROFILELOADREQUESTED;
	}
	else
	{
		s_bRequestPending = true;
		s_ePendingLoadingState = AVP_LOAD_SAVE_STATES_PROFILELOADREQUESTED;
	}

	return AVP_Profile::PROFILE_FILE_OK;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::SaveCheckpoint
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

bool AVP_Profile_LoadAndSave::SaveCheckpoint( int iSlot, bool bInterrogate, bool bHelp )
{
	if( !(iSlot >= 0 && iSlot < NUM_CPOINT_SAVESLOTS) )
	{
		Asura_Assert( false, "Invalid slot to be saved to in SaveCheckpoint");
		return false;
	}
	if( bInterrogate || bHelp )
	{
		Asura_CommandConsole::Printf("Save the checkpoint to a specific slot.");
	}
	else
	{
		// MkC - Might need a temporary u_int to hold iSlot in case someone presses Save, reenters the menu and saves again faster then
		// the console can get to idle state.
		s_uSlot = iSlot;
		if( s_eLoadingState == AVP_LOAD_SAVE_STATES_IDLE )
		{
			s_eLoadingState = AVP_LOAD_SAVE_STATES_CHECKPOINTSAVEREQUESTED;
		}
		else
		{
			// Save the checkpoint when opertunity gives, mostly for consoles
			s_bRequestPending = true;
			s_ePendingLoadingState = AVP_LOAD_SAVE_STATES_CHECKPOINTSAVEREQUESTED;
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::LoadCheckpoint
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

bool AVP_Profile_LoadAndSave::LoadCheckpoint( int iSlot, bool bInterrogate, bool bHelp )
{
	if( !(iSlot >= 0 && iSlot < NUM_CPOINT_SAVESLOTS) )
	{
		Asura_Assert( false, "Invalid slot to be Loaded from in LoadCheckpoint");
		return false;
	}
	if( bInterrogate || bHelp )
	{
		Asura_CommandConsole::Printf("Load the checkpoint from a specific slot.");
	}
	else
	{
		// MkC - Might need a temporary u_int to hold iSlot in case someone presses Save, reenters the menu and saves again faster then
		// the console can get to idle state.
		s_uSlot = iSlot;
		if( s_eLoadingState == AVP_LOAD_SAVE_STATES_IDLE )
		{
			s_eLoadingState = AVP_LOAD_SAVE_STATES_CHECKPOINTLOADREQUESTED;
		}
		else
		{
			// Load the checkpoint when opertunity gives, mostly for consoles
			s_bRequestPending = true;
			s_ePendingLoadingState = AVP_LOAD_SAVE_STATES_CHECKPOINTLOADREQUESTED;
		}
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::GetState
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

AVP_Profile_State AVP_Profile_LoadAndSave::GetState( )
{
	return Platform_GetState();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Profile_LoadAndSave::GetLastError
// Scope		: public static
// Author       : Mrten Carlzon
// Date         : 1/4/09
//
// Desc			: Interface to Platform dependent functionality
////////////////////////////////////////////////////////////////////////////////

//AVP_Profile_Error AVP_Profile_LoadAndSave::GetLastError(  bool bTrue  )
//{
//	return Platform_GetLastError();
//}


//CHECK THIS CODE - IF NOT USED ON XBOX DELETE IT
void AVP_Profile_LoadAndSave::WriteToChunkStream( Asura_Chunk_Memory_Stream& xStream )
{
	s_pxProfile->WriteToChunkStream( xStream );
}

void AVP_Profile_LoadAndSave::ReadFromChunkStream()
{
	s_pxProfileStream->SetPosition(0);
	s_pxProfile->ReadFromChunkStream( *(s_pxProfileStream) );
	s_pxProfile->ApplySettings();
}

bool AVP_Profile_LoadAndSave::GetDisplayName( u_int iSlot, bool bInterrogate, bool bHelp )
{
	if(bHelp || bInterrogate)
	{
		Asura_CommandConsole::Printf("Gets the name to display for the specific slot.");
	}
	else
	{
		Platform_GetSlotDisplayText( iSlot );
	}
	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : GetLevelId
// Scope		: 
// Author       : Mrten Carlzon
// Date         : 27/04/2009
//
// Desc			: Takes a levelname and returns the id of that level, 
//				  used for loading a level with AVP_GUIMenu_System::StartSpecificLevel
////////////////////////////////////////////////////////////////////////////////

u_int AVP_Profile_LoadAndSave::GetLevelId( const Asura_StringBuffer<128>* pxLevelname )
{
	int SpeciesValue;
	if( pxLevelname->Contains( "M0" ) )
	{
		SpeciesValue = 0;
	}
	else if( pxLevelname->Contains( "A0" ) )
	{
		SpeciesValue = 6;
	}
	else if( pxLevelname->Contains( "P0" ) )
	{
		SpeciesValue = 12;
	}

	if( pxLevelname->Contains( "01" ) )
	{
		return SpeciesValue;
	}
	else if( pxLevelname->Contains( "02" ) )
	{
		return SpeciesValue + 1;
	}
	else if( pxLevelname->Contains( "03" ) )
	{
		return SpeciesValue + 2;
	}
	else if( pxLevelname->Contains( "04" ) )
	{
		return SpeciesValue + 3;
	}
	else if( pxLevelname->Contains( "05" ) )
	{
		return SpeciesValue + 4;
	}
	else if( pxLevelname->Contains( "06" ) )
	{
		return SpeciesValue + 5;
	}
	
	Asura_Assertf( false, ("Invalid level name sent to GetLevelId! \nLevelname sent: %s", pxLevelname ) );
	return 0;
}

#endif	//defined(AVP_SUPPORT_PROFILE_LOADANDSAVE)