//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////
// AVP_Profile_LoadAndSave.h
// Thomas Young / Callum Hope - 8,1,09
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
// File functionality to deal with saving data to file on any platform
// No actual profile functionality, just profile reading saving, loading, and writing of data used in profiles. 
// Essentially this whole file is just a platform independent back bone to the save load system, 
// Its very different on all platforms so this is an interface to the three different methods; PC, PS3 and 360
//////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////

// F I L E   W R A P P E R ///////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifndef __AVP_PROFILE_LOADANDSAVE_H__
#define __AVP_PROFILE_LOADANDSAVE_H__

#if defined(AVP_SUPPORT_PROFILE_LOADANDSAVE)

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"

// avp headers
#include "Core/AVP_Game_Core.h"
#include "Core/AVP_Settings.h"
#include "Core/Profile/AVP_Profile.h"
#include "Core/Profile/AVP_Profile_Controls.h"
#include "Core/Profile/AVP_Profile_Graphics.h"
#include "Core/Profile/AVP_Profile_Sounds.h"

// D E F I N E S /////////////////////////////////////////////////////////////

#define INVALID_SAVE_SLOT		-1
#define NUM_CPOINT_SAVESLOTS	4

// E N U M E R A T I O N S ///////////////////////////////////////////////////

enum AVP_Profile_State
{
	AVP_PROFILE_STATE_IDLE,					// Doing nothing
	AVP_PROFILE_STATE_BUSY,					// Busy doing something not covered by the other states
	AVP_PROFILE_STATE_LOADING_PROFILE,		// Loading the profile data
	AVP_PROFILE_STATE_SAVING_PROFILE,		// Saving the profile data
	AVP_PROFILE_STATE_LOADING_CHECKPOINT,	// Loading the checkpoint data
	AVP_PROFILE_STATE_SAVING_CHECKPOINT,	// Saving the checkpoint data
	AVP_PROFILE_STATE_ERROR					// An error occured (use GetLastError() to find out what)
};

enum AVP_LOAD_SAVE_STATES
{
	AVP_LOAD_SAVE_STATES_IDLE,						// There are no loading operations going on
	AVP_LOAD_SAVE_STATES_FIRSTLAOD,					// Used on PS3 for loading the first profile, will save a profile if this failes
	AVP_LOAD_SAVE_STATES_PROFILELOADREQUESTED,		// A Load has been requested and will be dealt with when the platform can handle the request. 
	AVP_LOAD_SAVE_STATES_JUSTLOADEDPROFILE,			// Data has just been loaded but not yet transfered to the AVP_Profile class for use. 
	AVP_LOAD_SAVE_STATES_PROFILESAVEREQUESTED,		// A Save has been requested and will be dealt with when the platform can handle the request.
	AVP_LOAD_SAVE_STATES_JUSTSAVEDPROFILE,			// Data has just been saved. 
	AVP_LOAD_SAVE_STATES_PROFILEFLOWREQUESTED,		// A Profileflow has been requested and will be dealt with when the platform can handle the request.
	AVP_LOAD_SAVE_STATES_PROFILEFLOWLOADED,			// The ProfileFlow has been done
	AVP_LOAD_SAVE_STATES_CHECKPOINTSAVEREQUESTED,	// Request to save a checkpoint
	AVP_LOAD_SAVE_STATES_CHECKPOINTJUSTSAVED,		// When we have saved a checkpoint and ready the console can tidy up
	AVP_LOAD_SAVE_STATES_CHECKPOINTLOADREQUESTED,	// Request to load a checkpoint
	AVP_LOAD_SAVE_STATES_CHECKPOINTJUSTLOADED,		// When we have loaded the checkpoint and ready to load a level, used for consoles.
};

enum AVP_Profile_Error
{
	AVP_PROFILE_ERROR_NONE,
	// TODO: Fill me with error codes
};

class AVP_Profile_LoadAndSave
{
	public:

		static void Initialise			(AVP_Profile* pxProfile);
		static void Update				();
		static void Shutdown			();

		static void BeginProfileFlow	();

		static bool									SaveProfileData		();
		static AVP_Profile::E_PROFILE_FILE_STATES	LoadProfileData		();

		static void PopulateSlotNames	();

		static bool SaveCheckpoint		( int iSlot, bool bInterrogate, bool bHelp );
		static bool LoadCheckpoint		( int iSlot, bool bInterrogate, bool bHelp );

		static AVP_Profile_State GetState		();
		//static AVP_Profile_Error GetLastError	( bool bTrue );

		static int GetProfileSaveSize	(){ return s_uProfileSaveSize; }

		static void  ReadFromChunkStream();
		static void  WriteToChunkStream( Asura_Chunk_Memory_Stream& xStream );
		static bool  GetDisplayName( u_int iSlot, bool bInterrogate, bool bHelp );
		static u_int GetLevelId( const Asura_StringBuffer<128>* pxLevelname );

	private:

		static void Platform_Initialise			();
		static void Platform_Update				();
		static void Platform_Shutdown			();

		static void Platform_BeginProfileFlow	();

		static void	Platform_SaveProfileData	();
		static void	Platform_LoadProfileData	();

		static void	Platform_SetLoadSaveSlot	( const int iSlot );
		static const void Platform_GetSlotDisplayText( const int iSlot );

		static void Platform_SaveCheckpoint		( );
		static void Platform_LoadCheckpoint		( );

		static AVP_Profile_State Platform_GetState		();
		static AVP_Profile_Error Platform_GetLastError	();

		static void Platform_GetCurrentProfileFileName	( Asura_StringBuffer<128>& xFileNameAndPath, u_int iDepthToReturn = 0 );

		// Private variables...
		static AVP_Profile*				s_pxProfile;
		static u_int					s_uProfileSaveSize;
		static u_int					s_uCheckpointSize;
		static u_int					s_uSlot;

		static Asura_Chunk_Memory_Stream*		s_pxProfileStream;
		static Asura_Chunk_Memory_Stream*		s_pxCheckpointStream;

		static AVP_LOAD_SAVE_STATES		s_eLoadingState;
		static Asura_StringBuffer<128>	s_xCheckpointDisplayName[NUM_CPOINT_SAVESLOTS];

		// Used to save after just saving if we request a save while saving.
		static bool						s_bRequestPending;
		static AVP_LOAD_SAVE_STATES		s_ePendingLoadingState;
};

//////////////////////////////////////////////////////////////////////////////

#endif	//defined(AVP_SUPPORT_PROFILE_LOADANDSAVE)

#endif	//__AVP_PROFILE_LOADANDSAVE_H__

//////////////////////////////////////////////////////////////////////////////

