//////////////////////////////////////////////////////////////////////////////
//
// AVP_MoveData_Animation.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Moves/AVP_MoveData_Animation.h"

#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Core/AVP_Maths.h"

#define INVALID_BLEND_REGION 0xFFFFFFFF
#define INVALID_ANIM_INDEX 0xFFFFFFFF

DEBUGVAR float	dbg_fDefaultDirectionBlendTweenTime = 0.2f;

//////////////////////////////////////////////////////////////////////////////
//
// AVP_MoveData_Animation
//
//////////////////////////////////////////////////////////////////////////////

AVP_MoveData_Animation::AVP_MoveData_Animation( const Asura_Hash_ID uAnimHash,
                                                const u_int uFlags /*= 0*/,
                                                const u_int uPriority /*= AVP_AAPC_MOVE*/,
                                                const Asura_Hash_ID uControllerGroupHash /*= ASURA_HASH_ID_UNSET*/ )
: PARENT( uAnimHash, uFlags, uPriority, uControllerGroupHash )
{
	m_bAlwaysReCalculateFinalMovementOffset = false;
}


//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

AVP_MoveData_BlendedAnimation::AVP_MoveData_BlendedAnimation(	const Asura_Blueprint& xAnimBlendTemplateBlueprint,
																const u_int uFlags /*= 0*/,
																const u_int uPriority /*= AVP_AAPC_MOVE*/,
																const Asura_Hash_ID uControllerGroupHash /*= ASURA_HASH_ID_UNSET */ )
: PARENT( ASURA_HASH_ID_UNSET, uFlags, uPriority, uControllerGroupHash )
, m_uNumBlendRegions( 0 )
, m_fBlendAngle( 0.0f )
, m_uCurrentAnimBlendRegion( INVALID_BLEND_REGION )
, m_uCurrentACWAnimIndex( INVALID_ANIM_INDEX )
, m_uLastAnimBlendRegion( INVALID_BLEND_REGION )
{
	m_bAlwaysReCalculateFinalMovementOffset = true;

	for ( u_int uItr = 0; uItr < AVP_MOVEDATA_MAX_BLEND_REGIONS; ++uItr )
	{
		m_auNumAnims[ uItr ] = 0;
		m_afTweenAngleACW[ uItr ] = 0.0f;
		m_afTweenAngleCW[ uItr ] = 0.0f;
		for ( u_int uAnimItr = 0; uAnimItr < AVP_MOVEDATA_MAX_ANIMS_PER_REGION; ++uAnimItr )
		{
			m_aauTweenOnAnimFlags[ uItr ][ uAnimItr ] = 0;
			m_aauAnimHash[ uItr ][ uAnimItr ] = ASURA_HASH_ID_UNSET;
			m_aafAngle[ uItr ][ uAnimItr ] = 0.0f;
		}
	}

	m_auMoveAnimAACAHash[ 0 ] = ASURA_HASH_ID_UNSET;
	m_auMoveAnimAACAHash[ 1 ] = ASURA_HASH_ID_UNSET;

	InitialiseFromBlueprint( xAnimBlendTemplateBlueprint );
}

void AVP_MoveData_BlendedAnimation::Reset( void )
{
	if(m_pxAnimation != NULL && !m_bLeaveAnimAtEnd)
	{
		m_pxAnimation->DestroyController(m_auMoveAnimAACAHash[ 0 ]);
		m_pxAnimation->DestroyController(m_auMoveAnimAACAHash[ 1 ]);
	}

	m_pxAnimation				= NULL;
	m_auMoveAnimAACAHash[ 0 ]	= ASURA_HASH_ID_UNSET;
	m_auMoveAnimAACAHash[ 1 ]	= ASURA_HASH_ID_UNSET;

	ChangeState(WAITING);
}

bool AVP_MoveData_BlendedAnimation::Start( const bool bLoopAtEnd, Asura_Animation& xAnimation )
{
	m_bLoopAtEnd = bLoopAtEnd;
	m_pxAnimation = &xAnimation;
	
	UpdateBlends(); // creates the anims.

	// Ensure that we actually start the timer...
	if(!xAnimation.UpdateTimers(ANIMTIME_USEGAMETIME, false))
	{
		//Failed to update the timers because they have already been done this fame.
		//Force an update with a zero frame time to make sure that the tween data store
		//is up to date with the new animation , but without causing the timers to
		//be advanced a second time.
		//This should hopefully prevent the animation from glitching while also preventing the
		//sounds/flags from being lost for the first frame.
		//RAR 13/11/2007
		xAnimation.UpdateTimers(0.0f, true);
	}

	ChangeState(STARTING);
	return(true);
}

bool AVP_MoveData_BlendedAnimation::Update(const float fTimeStep)
{
	if(m_pxAnimation == NULL)
	{
		Asura_Assert(false, "No animation to use when updating move-data.");
		return(false);
	}
	else
	{
		UpdateBlends();
	
		// make sure our animations are up-to-date...
		m_pxAnimation->UpdateTimers(fTimeStep);

		if(m_bWaitForCompletion)
		{
			const Asura_AnimationController_Animation*	pxAnimationController	= m_pxAnimation->GetControllerAnimationFromHash(m_auMoveAnimAACAHash[ 0 ]); // anims are synced so we can just check the first one

			if(pxAnimationController == NULL)
			{
				Asura_Assert(false, "Cannot find the required animation for this move.");
				return(false);
			}
			else if(pxAnimationController->IsFinished())
			{
				return(false);
			}
			else
			{
				return(true);
			}
		}
		else
		{
			// We don't need to wait for the animation...
			return(false);
		}
	}
}

void AVP_MoveData_BlendedAnimation::StopLoopedAnims()
{
	if ( m_pxAnimation )
	{
		for ( u_int uItr = 0; uItr < 2; ++uItr )
		{
			Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ uItr ] );

			// Stop any looped animations for this move...
			if((pxAACA != NULL) && ((pxAACA->GetControllerFlags() & AACA_FLAG_LOOPED) != 0))
			{
				pxAACA->Stop();
			}
		}
	}
}

bool AVP_MoveData_BlendedAnimation::CalculateMovementOffset(	const float fParametricTimer,
																Asura_AVector3* pxPosition,
																Asura_Quat* pxOrientation) const
{
	const Asura_Hierarchy_Skin*	pxHierachySkin	= GetHierarchySkin();
	if ( m_pxAnimation == NULL || pxHierachySkin == NULL )
	{
		Asura_Assert( false, "invalid Skin/Anim in the move." );
		return false;
	}
	Asura_Vector_3	axPos[ 2 ];
	Asura_Quat		axOri[ 2 ];
	for ( u_int uItr = 0; uItr < 2; ++uItr )
	{
		const Asura_AnimationController_Animation*	pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ uItr ] );
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= ( pxAACA == NULL ) ? NULL : pxAACA->GetAnimation();
		if ( pxHierarchyAnim == NULL )
		{
			Asura_Assert(false, "Cannot find the required animation for this move.");
			return(false);
		}

		if ( !pxHierarchyAnim->GetMovementOffset( pxHierachySkin, fParametricTimer, &axPos[ uItr ], &axOri[ uItr ], false ) )
		{
			return(false);
		}
	}

	if ( pxPosition )
	{
		axPos[ 0 ].Lerp( axPos[ 1 ], m_fCurrentBlendDir );
		*pxPosition = axPos[ 0 ];
	}
	if ( pxOrientation )
	{
		pxOrientation->Slerp( axOri[ 0 ], axOri[ 1 ], m_fCurrentBlendDir );
	}
	return true;
}

bool AVP_MoveData_BlendedAnimation::CalculateVelocity(	const float fParametricTimer,
														Asura_AVector3& xVelocity) const
{
	const Asura_Hierarchy_Skin*	pxHierachySkin	= GetHierarchySkin();
	if ( m_pxAnimation == NULL || pxHierachySkin == NULL )
	{
		Asura_Assert( false, "invalid Skin/Anim in the move." );
		return false;
	}
	Asura_Vector_3	axVel[ 2 ];
	for ( u_int uItr = 0; uItr < 2; ++uItr )
	{
		const Asura_AnimationController_Animation*	pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ uItr ] );
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= pxAACA->GetAnimation();
		if ( pxHierarchyAnim == NULL )
		{
			Asura_Assert(false, "Cannot find the required animation for this move.");
			return(false);
		}
	
		if ( !pxHierarchyAnim->GetVelocity( pxHierachySkin, fParametricTimer, GetAnimationDuration(), &axVel[ uItr ] ) )
		{
			return(false);
		}
	}

	const Asura_AnimationController_Animation*	pxFirstAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ 0 ] );
	const Asura_AnimationBlend* pxDirBlend = pxFirstAACA ? pxFirstAACA->GetBlendToNextController() : NULL;
	if ( pxDirBlend && pxDirBlend->GetSlerpVelocity() )
	{
		// This could be moved into Asura_AVector3::Slerp or maybe Asura_AVector3::SlerpWithMagnitude as it's becoming a common thing to do. (Also used in Direction_Slerpolator with no magnitude)
		Asura_AVector3 xNormFrom( axVel[ 0 ] );
		Asura_AVector3 xNormTo( axVel[ 1 ] );
		const float fFromMag = xNormFrom.NormaliseReturningMagnitude();
		float fToMag = xNormTo.NormaliseReturningMagnitude();
		fToMag = Asura_Maths::Lerp( fFromMag, fToMag, m_fCurrentBlendDir );
		Asura_Quat xRot;
		Asura_AVector3 xAxis;
		xRot.RotationFromAToB( xNormFrom, xNormTo );
		float fAngle = xRot.ToRotation( xAxis );
		fAngle *= m_fCurrentBlendDir;
		xRot.FromRotation( fAngle, xAxis );
		xVelocity = xRot.RotateVector( xNormFrom);
		xVelocity *= fToMag;
	}
	else
	{
		axVel[ 0 ].Lerp( axVel[ 1 ], m_fCurrentBlendDir );
		xVelocity = axVel[ 0 ];
	}
	
	return true;
}

bool AVP_MoveData_BlendedAnimation::CalculateAngularVelocity(	const float fParametricTimer,
																Asura_AVector3& xAxis,
																Asura_Angle& fAngle) const
{
	const Asura_Hierarchy_Skin*	pxHierachySkin	= GetHierarchySkin();
	if ( m_pxAnimation == NULL || pxHierachySkin == NULL )
	{
		Asura_Assert( false, "invalid Skin/Anim in the move." );
		return false;
	}
	Asura_Vector_3	axAxis[ 2 ];
	float			afAngSpeed[ 2 ];
	for ( u_int uItr = 0; uItr < 2; ++uItr )
	{
		const Asura_AnimationController_Animation*	pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ uItr ] );
		const Asura_Hierarchy_Anim*	pxHierarchyAnim	= pxAACA->GetAnimation();
		if ( pxHierarchyAnim == NULL )
		{
			Asura_Assert(false, "Cannot find the required animation for this move.");
			return(false);
		}
	
		if(!pxHierarchyAnim->GetAngularVelocity(fParametricTimer, GetAnimationDuration(), &axAxis[ uItr ], &afAngSpeed[ uItr ]))
		{
			return(false);
		}
		axAxis[ uItr ] *= afAngSpeed[ uItr ];
	}

	xAxis.Lerp( Asura_AVector3( axAxis[ 0 ] ), Asura_AVector3( axAxis[ 1 ] ), m_fCurrentBlendDir );
	fAngle = xAxis.NormaliseReturningMagnitude();
	return true;
}

const Asura_AnimationController_Animation* AVP_MoveData_BlendedAnimation::GetAnimationController( void ) const
{
	if(m_pxAnimation == NULL)
	{
		return(NULL);
	}
	else
	{
		return(m_pxAnimation->GetControllerAnimationFromHash(m_auMoveAnimAACAHash[ 0 ]));
	}
}

Asura_AnimationController_Animation* AVP_MoveData_BlendedAnimation::GetAnimationController( void )
{
	if(m_pxAnimation == NULL)
	{
		return(NULL);
	}
	else
	{
		return(m_pxAnimation->GetControllerAnimationFromHash(m_auMoveAnimAACAHash[ 0 ]));
	}
}

const Asura_Hierarchy_Anim* AVP_MoveData_BlendedAnimation::GetHierarchyAnim( void ) const
{
	if ( HasValidAnimation() )
	{
		return Asura_Animation_System::GetAnimByHash( m_aauAnimHash[ 0 ][ 0 ] );
	}
	else
	{
		return NULL;
	}
}

void AVP_MoveData_BlendedAnimation::InitialiseFromBlueprint( const Asura_Blueprint& xAnimBlendTemplateBlueprint )
{
	AVP_Blueprint_ParameterEnumerator_MoveData_BlendedAnimation_Region xRegionEnumerator( xAnimBlendTemplateBlueprint, m_afTweenAngleACW, m_afTweenAngleCW );
	m_uNumBlendRegions = xAnimBlendTemplateBlueprint.EnumerateParameters( "Region.", 2, xRegionEnumerator );
	Asura_Assert( m_uNumBlendRegions < AVP_MOVEDATA_MAX_BLEND_REGIONS, "Too many blend regions in state blueprint" );
	if ( m_uNumBlendRegions >= AVP_MOVEDATA_MAX_BLEND_REGIONS )
	{
		m_uNumBlendRegions = AVP_MOVEDATA_MAX_BLEND_REGIONS;
	}
	for ( u_int uRegionItr = 0; uRegionItr < m_uNumBlendRegions; ++uRegionItr )
	{
		AVP_Blueprint_ParameterEnumerator_MoveData_BlendedAnimation_Anim xAnimEnumerator( xAnimBlendTemplateBlueprint, uRegionItr, &m_aauTweenOnAnimFlags[0][0], &m_aauAnimHash[0][0], &m_aafAngle[0][0] );
		m_auNumAnims[ uRegionItr ] = xAnimBlendTemplateBlueprint.EnumerateParameters( ".Anim.", 2, xAnimEnumerator );
	}
}

void AVP_MoveData_BlendedAnimation::UpdateBlends()
{
	if ( m_uNumBlendRegions > 0 )
	{
		u_int uNumInRegions = 0;
		int iCurrentRegionIndex = -1;
		int iLastRegionIndex = -1;
		int aiInRegion[ 2 ];
		int aiInRegionACWAnimIndex[ 2 ];
		float afInRegionParametric[ 2 ];
		bool bInTweenRegion[ 2 ];

		for ( u_int uRegionIndexItr = 0; uRegionIndexItr < 2; ++uRegionIndexItr )
		{
			aiInRegion[ uRegionIndexItr ] = INVALID_BLEND_REGION;
			aiInRegionACWAnimIndex[ uRegionIndexItr ] = INVALID_ANIM_INDEX;
			afInRegionParametric[ uRegionIndexItr ] = 0.0f;
			bInTweenRegion[ uRegionIndexItr ] = false;
		}

		for ( u_int uRegionItr = 0; uRegionItr < m_uNumBlendRegions && uNumInRegions < 2; ++uRegionItr )
		{
			if ( m_auNumAnims[ uRegionItr ] > 1 && AVP_Maths::IsAngleInRange( m_fBlendAngle, m_aafAngle[ uRegionItr ][ 0 ], m_aafAngle[ uRegionItr ][ m_auNumAnims[ uRegionItr ] - 1 ] ) )
			{
				bInTweenRegion[ uRegionItr ] = AVP_Maths::IsAngleInRange( m_fBlendAngle, m_afTweenAngleACW[ uRegionItr ], m_afTweenAngleCW[ uRegionItr ] );
				if ( m_uCurrentAnimBlendRegion == uRegionItr )
				{
					iCurrentRegionIndex = uNumInRegions;
				}
				if ( m_uLastAnimBlendRegion == uRegionItr )
				{
					iLastRegionIndex = uNumInRegions;
				}
				aiInRegion[ uNumInRegions ] = uRegionItr;
				for ( u_int uAnimItr = 1; uAnimItr < m_auNumAnims[ uRegionItr ]; ++uAnimItr )
				{
					float fParametric;
					if ( AVP_Maths::IsAngleInRange( m_fBlendAngle, m_aafAngle[ uRegionItr ][ uAnimItr - 1 ], m_aafAngle[ uRegionItr ][ uAnimItr ], &fParametric ) )
					{
						aiInRegionACWAnimIndex[ uNumInRegions ] = uAnimItr - 1;
						afInRegionParametric[ uNumInRegions ] = fParametric;
					}
				}
				++uNumInRegions;
			}
		}

		Asura_Assert( !( bInTweenRegion[ 0 ] && bInTweenRegion[ 1 ] ), "Can't be in two tween regions at one, error in movement template blueprint" );
		Asura_Assertf( uNumInRegions == 1 || uNumInRegions == 2, ( "Blended Anim Movement Blueprint not set up correctly we have %d valid regions at angle %.0f", uNumInRegions, Asura_Maths::RadToDeg( m_fBlendAngle ) ) );
		bool bTweenFlagPresent = true; // TODO
		u_int uNextRegionIndex = 0;
		bool bTweenNow = false;
		float fTweenTime = dbg_fDefaultDirectionBlendTweenTime;
		if ( iCurrentRegionIndex < 0 ) // if we're no longer in our current region then we must tween over to the new one
		{
			if ( m_uCurrentAnimBlendRegion == INVALID_BLEND_REGION )
			{
				fTweenTime = m_fTweenInDuration;
			}
			bTweenNow = true;
			if ( uNumInRegions == 2 && ( iLastRegionIndex == 0 || bInTweenRegion[ 1 ] ) )
			{
				uNextRegionIndex = 1;
			}
		}
		else if ( uNumInRegions == 2 && bInTweenRegion[ 1 - iCurrentRegionIndex ] && bTweenFlagPresent ) // else if we're in two regions and in the tween region of the non-current region and the tween flag is present.
		{
			bTweenNow = true;
			uNextRegionIndex = 1 - iCurrentRegionIndex;
		}
		else if ( m_uCurrentACWAnimIndex != static_cast< u_int >( aiInRegionACWAnimIndex[ iCurrentRegionIndex ] ) ) // else if we're not playing the correct anims for the current region.
		{
			bTweenNow = true;
			uNextRegionIndex = iCurrentRegionIndex;
			fTweenTime = 0.0f;
		}

		if ( bTweenNow )
		{
			const Asura_Hash_ID uAnimACW = m_aauAnimHash[ aiInRegion[ uNextRegionIndex ] ][ Asura_Maths::Max( aiInRegionACWAnimIndex[ uNextRegionIndex ], 0 ) ]; // if the region only has 1 anim then aiInRegionACWAnimIndex[ uNextRegionIndex ] will be -1 so we'll use anim 0 for both anims here
			const Asura_Hash_ID uAnimCW = m_aauAnimHash[ aiInRegion[ uNextRegionIndex ] ][ aiInRegionACWAnimIndex[ uNextRegionIndex ] + 1 ];

			m_fCurrentBlendDir = afInRegionParametric[ uNextRegionIndex ];
			m_uLastAnimBlendRegion = m_uCurrentAnimBlendRegion;
			m_uCurrentAnimBlendRegion = aiInRegion[ uNextRegionIndex ];
			m_uCurrentACWAnimIndex = aiInRegionACWAnimIndex[ uNextRegionIndex ];
			TweenToNewDirectionalAnimGroup( uAnimACW, uAnimCW, fTweenTime );
		}
		else
		{
			Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ 0 ] );
			if ( pxAACA != NULL )
			{
				Asura_AnimationBlend* pxPrevBlend = pxAACA->GetBlendToPreviousController();
				Asura_AnimationBlend* pxDirBlend = pxAACA->GetBlendToNextController();
				if ( pxPrevBlend != NULL && pxDirBlend != NULL )
				{
					m_fCurrentBlendDir = afInRegionParametric[ iCurrentRegionIndex ];
					pxDirBlend->SetWeight( pxPrevBlend->GetWeight() * m_fCurrentBlendDir );
				}
			}
			// Maybe update the blend on the old (tweening out) state anims? See how we go.
		}
	}	// if has blend regions

	// mess with the durations
	const Asura_AnimationController_Animation* pxACWAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ 0 ] );
	const Asura_Hierarchy_Anim* pxACWAnim = pxACWAACA ? pxACWAACA->GetAnimation() : NULL;
	float fACWDuration = pxACWAnim != NULL ? pxACWAnim->GetDuration() : -1.0f;

	const Asura_AnimationController_Animation* pxCWAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ 1 ] );
	const Asura_Hierarchy_Anim* pxCWAnim = pxCWAACA ? pxCWAACA->GetAnimation() : NULL;
	float fCWDuration = pxCWAnim != NULL ? pxCWAnim->GetDuration() : -1.0f;

	float fDuration = fACWDuration < 0.0f ? fCWDuration : ( fCWDuration < 0.0f ? fACWDuration : Asura_Maths::Lerp( fACWDuration, fCWDuration, m_fCurrentBlendDir ) );

	for ( u_int uDirItr = 0; uDirItr < 2; ++uDirItr )
	{
		Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ uDirItr ] );
		if ( pxAACA != NULL )
		{
			Asura_Assert( fDuration >= 0.0f, "Couldn't get a good duration" );
			if ( fDuration >= 0.0f )
			{
				pxAACA->SetDuration( fDuration );
			}
		}
	}
}

void AVP_MoveData_BlendedAnimation::TweenToNewDirectionalAnimGroup( const Asura_Hash_ID uAnimACW, const Asura_Hash_ID uAnimCW, const float fTweenTime )
{
	u_int uModifiedFlags = m_uAnimFlags | AAC_FLAG_ACTIVE | AAC_FLAG_ENSURE_UNIQUE_HASH;

	if (!m_bDontCalculateVelocity)
	{
		uModifiedFlags |= AACA_FLAG_CALCULATE_VELOCITY | AACA_FLAG_CALCULATE_MOVEMENT_OFFSET;
	}
	if (!m_bDontCalculateAngularVelocity)
	{
		uModifiedFlags |= AACA_FLAG_CALCULATE_ANGULAR_VELOCITY | AACA_FLAG_CALCULATE_MOVEMENT_OFFSET;
	}

	if(m_bLoopAtEnd)
	{
		uModifiedFlags |= AACA_FLAG_LOOPED;
	}
	else
	{
		Asura_Assert((uModifiedFlags & AACA_FLAG_LOOPED)==0,"Loop anim flag set, m_bLoopAtEnd not set. Will cause move to end leaving animation playing!");
		uModifiedFlags &= ~AACA_FLAG_LOOPED;
	}

	// Update our animation flags based upon our settings...
	if((m_fTweenOutDuration <= 0.0f) && !m_bLoopAtEnd && !m_bLeaveAnimAtEnd)
	{
		uModifiedFlags	|= AAC_FLAG_DESTROY_WHEN_FINISHED;
	}
	else
	{
		uModifiedFlags	|= AAC_FLAG_ALLOW_DESTROY_FROM_MESSAGE;
	}

	float fParametric = 0.0f;
	Asura_AnimationController_Animation* pxAACA = m_pxAnimation->GetControllerAnimationFromHash( m_auMoveAnimAACAHash[ 0 ] );
	if ( pxAACA != NULL )
	{
		fParametric = pxAACA->GetParametricTimer();
	}

	for ( u_int uDirItr = 0; uDirItr < 2; ++uDirItr )
	{
		m_auMoveAnimAACAHash[ uDirItr ] = ASURA_HASH_ID_UNSET;
	}

	Asura_AnimationController_Group* pxParentAACG = m_pxAnimation->GetControllerGroupFromHash( m_uControllerGroupHash );
	for ( u_int uDirItr = 0; uDirItr < 2; ++ uDirItr )
	{
		Asura_AnimationBlend* pxBlend;
		if ( uDirItr == 0 )
		{
			pxBlend = m_bAnimatedTween	? new Asura_AnimationBlend_AnimatedTween( fTweenTime, m_fTweenOutDuration, false )
										: new Asura_AnimationBlend_Tween( fTweenTime, m_fTweenOutDuration, false );
		}
		else
		{
			pxBlend = new Asura_AnimationBlend( fTweenTime > 0.0f ? 0.0f : m_fCurrentBlendDir );
			pxBlend->SetSlerpVelocity( true );
		}

		Asura_Hash_ID uAnimHash = uDirItr == 0 ? uAnimACW : uAnimCW;
		Asura_AnimationController_Animation* pxAACA = m_pxAnimation->CreateControllerAnimation( uModifiedFlags, m_uAnimPriority, uAnimHash, pxBlend, pxParentAACG );
		if ( pxAACA )
		{
			if ( pxAACA->GetAnimation() != NULL )
			{
				m_auMoveAnimAACAHash[ uDirItr ] = pxAACA->GetControllerHash();
				pxAACA->SetParametricTimer( fParametric, false );
				pxAACA->SetParametricStartTime( m_fParametricStartTime );
				pxAACA->SetParametricEndTime( m_fParametricEndTime );
				pxAACA->Play();
			}
			else
			{
				m_auMoveAnimAACAHash[ uDirItr ] = ASURA_HASH_ID_UNSET;
			}
		}
		else
		{
			m_auMoveAnimAACAHash[ uDirItr ] = ASURA_HASH_ID_UNSET;
		}
	}
}

bool AVP_Blueprint_ParameterEnumerator_MoveData_BlendedAnimation_Region::Enumerate( const u_int uIndex, char const* const szName, const Asura_Value& xValue )
{
	bool bOK = uIndex < AVP_MOVEDATA_MAX_BLEND_REGIONS;
	Asura_Assert( uIndex < AVP_MOVEDATA_MAX_BLEND_REGIONS, "Too many blend regions in state blueprint" );
	if ( bOK )
	{
		char szRegionNumBuffer[3];
		Asura_PrintfToBfr( szRegionNumBuffer, "%02d", 3, true, uIndex + 1 );
		Asura_String xRegionString = Asura_String( "Region." ) + szRegionNumBuffer;
		Asura_Value xNewValue;
		if ( m_xBP.GetParameterValueByHashID( ( xRegionString + ".TweenAngle.ACW" ).GetHash(), xNewValue ) )
		{
			m_pfTweenAngleACW[ uIndex ] = Asura_Maths::RadiansToZeroMinimumRange( Asura_Maths::DegToRad( xNewValue.GetFloatValue() ) );
		}
		if ( m_xBP.GetParameterValueByHashID( ( xRegionString + ".TweenAngle.CW" ).GetHash(), xNewValue ) )
		{
			m_pfTweenAngleCW[ uIndex ] = Asura_Maths::RadiansToZeroMinimumRange( Asura_Maths::DegToRad( xNewValue.GetFloatValue() ) );
		}
	}
	return bOK;
}

bool AVP_Blueprint_ParameterEnumerator_MoveData_BlendedAnimation_Anim::Enumerate( const u_int uIndex, char const* const szName, const Asura_Value& xValue )
{
	bool bOK = uIndex < AVP_MOVEDATA_MAX_ANIMS_PER_REGION;
	Asura_Assert( uIndex < AVP_MOVEDATA_MAX_ANIMS_PER_REGION, "Too many angles in blend region" );
	if ( bOK )
	{
		char szRegionNumBuffer[3];
		Asura_PrintfToBfr( szRegionNumBuffer, "%02d", 3, true, m_uRegionIndex + 1 );
		char szAnimNumBuffer[3];
		Asura_PrintfToBfr( szAnimNumBuffer, "%02d", 3, true, uIndex + 1 );
		Asura_String xRegionString = Asura_String( "Region." ) + szRegionNumBuffer;
		Asura_Value xNewValue;
		if ( m_xBP.GetParameterValueByHashID( ( xRegionString + ".Anim." + szAnimNumBuffer ).GetHash(), xNewValue ) )
		{
			m_puAnimHash[ m_uRegionIndex * AVP_MOVEDATA_MAX_ANIMS_PER_REGION + uIndex ] = xNewValue.GetHashIDValue();
		}
		if ( m_xBP.GetParameterValueByHashID( ( xRegionString + ".Angle." + szAnimNumBuffer ).GetHash(), xNewValue ) )
		{
			m_pfAngle[ m_uRegionIndex * AVP_MOVEDATA_MAX_ANIMS_PER_REGION + uIndex ] = Asura_Maths::RadiansToZeroMinimumRange( Asura_Maths::DegToRad( xNewValue.GetFloatValue() ) );
		}
	}

	return bOK;
}

const char* AVP_Blueprint_ParameterEnumerator_MoveData_BlendedAnimation_Anim::GetParameterFormat() const
{
	Asura_PrintfToBfr( m_szParameterFormat, "Region.%02d%%s%%0*d", 20, true, m_uRegionIndex + 1 );
	return m_szParameterFormat;
}
