//
// AVP_Multiplayer_SpawnSystem.cpp
//
// Created by JG - 13 October 2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Multiplayer/AVP_Multiplayer_SpawnSystem.h"

// Engine
#include "Core/Asura_HashedList.h"
#include "Core/Asura_List.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Network/Asura_Network.h"
#include "Physics/Asura_Physics_Broadphase.h"
#include "Physics/Asura_Physics_RigidBody.h"

// Project 
#include "Core/AVP_Game_Core.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PlayerInfo/AVP_ServerEntity_PlayerInfo.h"
#include "Entity/Spawners/AVP_ServerEntity_StartPoint.h"

////////////////////////////////////////////////////////////
//  Static data
////////////////////////////////////////////////////////////
Asura_Hashed_List<AVP_ServerEntity_StartPoint> AVP_Multiplayer_SpawnSystem::s_xSpeciesAlienStartPointList;
Asura_Hashed_List<AVP_ServerEntity_StartPoint> AVP_Multiplayer_SpawnSystem::s_xSpeciesHumanStartPointList;
Asura_Hashed_List<AVP_ServerEntity_StartPoint> AVP_Multiplayer_SpawnSystem::s_xSpeciesPredatorStartPointList;
Asura_Hashed_List<AVP_ServerEntity_StartPoint> AVP_Multiplayer_SpawnSystem::s_xTeamOneStartPointList; 
Asura_Hashed_List<AVP_ServerEntity_StartPoint> AVP_Multiplayer_SpawnSystem::s_xTeamTwoStartPointList;
Asura_Hashed_List<AVP_ServerEntity_StartPoint> AVP_Multiplayer_SpawnSystem::s_xTeamThreeStartPointList; 

////////////////////////////////////////////////////////////
//  Debug variables
////////////////////////////////////////////////////////////
DEBUGVAR float dbg_fBlockedDetectionRange = 1.0f; 

void AVP_Multiplayer_SpawnSystem::Initialise()
{
	// Only the server needs to worry about controlling spawn points
	if ( !Asura_Network::IsServer() )
	{
		return; 
	}
	
	// Empty lists before start adding start points
	ClearStartPointLists();

	// Build cached lists for all static start points in the map
	// Species based
	BuildStartPointList(s_xSpeciesAlienStartPointList, AVP_Game_Core::GetStartPointBlueprintHashForSpecies( AVP_Settings::AVP_SPECIES_ALIEN )); 
	BuildStartPointList(s_xSpeciesHumanStartPointList, AVP_Game_Core::GetStartPointBlueprintHashForSpecies( AVP_Settings::AVP_SPECIES_HUMAN )); 
	BuildStartPointList(s_xSpeciesPredatorStartPointList, AVP_Game_Core::GetStartPointBlueprintHashForSpecies( AVP_Settings::AVP_SPECIES_PREDATOR )); 
	// Team based
	BuildStartPointList(s_xTeamOneStartPointList, AVP_Game_Core::GetStartPointBlueprintHashForTeam( AVP_Settings::AVP_TEAM_ONE )); 
	BuildStartPointList(s_xTeamTwoStartPointList, AVP_Game_Core::GetStartPointBlueprintHashForTeam( AVP_Settings::AVP_TEAM_TWO )); 
	BuildStartPointList(s_xTeamThreeStartPointList, AVP_Game_Core::GetStartPointBlueprintHashForTeam( AVP_Settings::AVP_TEAM_THREE )); 
}

void AVP_Multiplayer_SpawnSystem::Update()
{
	// Only the server needs to worry about controlling spawn points
	if ( !Asura_Network::IsServer() )
	{
		return; 
	}
}

void AVP_Multiplayer_SpawnSystem::Render()
{
}

void AVP_Multiplayer_SpawnSystem::Shutdown()
{
	// Clear all possibly cached start points
	ClearStartPointLists(); 
}

void AVP_Multiplayer_SpawnSystem::ClearStartPointLists()
{
	// Empty all the cached spawn point lists
	s_xSpeciesAlienStartPointList.Clear(); 
	s_xSpeciesHumanStartPointList.Clear();
	s_xSpeciesPredatorStartPointList.Clear();
	s_xTeamOneStartPointList.Clear(); 
	s_xTeamTwoStartPointList.Clear(); 
	s_xTeamThreeStartPointList.Clear();
}

void AVP_Multiplayer_SpawnSystem::BuildStartPointList( Asura_Hashed_List<AVP_ServerEntity_StartPoint>& xStartPointList, Asura_Hash_ID uBlueprintParameterHash )
{
	// Build a list of start points appropriate to us.
	Asura_Hashed_List_It<AVP_ServerEntity_StartPoint> xIt(AVP_ServerEntity_StartPoint::GetList());
	
	while (!xIt.Done())
	{
		AVP_ServerEntity_StartPoint* pxStartPoint = xIt.GetCurrent();
		xIt.Next();

		if (!pxStartPoint) { continue; }

		const Asura_Hash_ID uBlueprintHashID = pxStartPoint->GetBlueprintHashID();
		const Asura_Hash_ID uBlueprintTypeHashID = pxStartPoint->GetBlueprintTypeHashID();
		Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintTypeHashID, uBlueprintHashID);
		if (pxBlueprint)
		{
			Asura_Value xValue;
			if (pxBlueprint->GetParameterValueByHashID(uBlueprintParameterHash, xValue) && xValue.GetBoolValue())
			{
				// Valid point.
				xStartPointList.AddToList(pxStartPoint);
			}
		}
		else
		{
			// Assume that if no blueprint was chosen, the start point can be used by anyone.
			xStartPointList.AddToList(pxStartPoint);
		}
	}
}


AVP_ServerEntity_StartPoint* AVP_Multiplayer_SpawnSystem::FindRandomStartPointForIndividual( Asura_Guid uPlayerGuid )
{
	AVP_ServerEntity_StartPoint* pxSelectedStartPoint = NULL; 
	AVP_ServerEntity_PlayerInfo* pxPlayerInfo = AVP_ServerEntity_PlayerInfo::FindEntity( uPlayerGuid ); 
	if (pxPlayerInfo)
	{
		AVP_Settings::Species eSpecies = pxPlayerInfo->GetPuppetSpecies();
		pxSelectedStartPoint = FindRandomStartPointForIndividual( eSpecies );
	}
	else
	{
		Asura_Assert(false, "Couldn't find player info, don't know where to spawn this player"); 
	}
	return pxSelectedStartPoint; 
}

AVP_ServerEntity_StartPoint* AVP_Multiplayer_SpawnSystem::FindRandomStartPointForIndividual( AVP_Settings::Species eSpecies )
{
	AVP_ServerEntity_StartPoint* pxSelectedStartPoint = NULL; 
	Asura_List<AVP_ServerEntity_StartPoint> xStartPointCandidates; 
	CreateSuitableStartPointList( GetStartPointListForSpecies( eSpecies ), xStartPointCandidates ); 
	int iNumberOfSuitableStartPoints = xStartPointCandidates.GetNumEntries();
	if (iNumberOfSuitableStartPoints > 0)
	{
		int iRandomStartPoint = Asura_Random::GetInt(0, iNumberOfSuitableStartPoints-1);
		pxSelectedStartPoint = xStartPointCandidates.GetEntryByIndex(iRandomStartPoint);
	}
	return pxSelectedStartPoint; 
}

AVP_ServerEntity_StartPoint* AVP_Multiplayer_SpawnSystem::FindRandomStartPointForTeam( Asura_Guid uPlayerGuid )
{
	AVP_ServerEntity_StartPoint* pxSelectedStartPoint = NULL; 
	AVP_ServerEntity_PlayerInfo* pxPlayerInfo = AVP_ServerEntity_PlayerInfo::FindEntity( uPlayerGuid ); 
	if (pxPlayerInfo)
	{
		AVP_Settings::Team eTeam = pxPlayerInfo->GetMultiplayerTeam();
		pxSelectedStartPoint = FindRandomStartPointForTeam( eTeam );
	}
	else
	{
		Asura_Assert(false, "Couldn't find player info, don't know where to spawn this player"); 
	}
	return pxSelectedStartPoint; 
}

AVP_ServerEntity_StartPoint* AVP_Multiplayer_SpawnSystem::FindRandomStartPointForTeam( AVP_Settings::Team eTeam )
{
	AVP_ServerEntity_StartPoint* pxSelectedStartPoint = NULL; 
	Asura_List<AVP_ServerEntity_StartPoint> xStartPointCandidates; 
	CreateSuitableStartPointList( GetStartPointListForTeam( eTeam ), xStartPointCandidates ); 
	int iNumberOfSuitableStartPoints = xStartPointCandidates.GetNumEntries();
	if ( iNumberOfSuitableStartPoints > 0)
	{
		int iRandomStartPoint = Asura_Random::GetInt(0, iNumberOfSuitableStartPoints-1);
		pxSelectedStartPoint = xStartPointCandidates.GetEntryByIndex(iRandomStartPoint);
	}
	return pxSelectedStartPoint; 
}

void AVP_Multiplayer_SpawnSystem::CreateSuitableStartPointList( const Asura_Hashed_List<AVP_ServerEntity_StartPoint>& xInputList, 
																Asura_List<AVP_ServerEntity_StartPoint>& xOutputList )
{
	Asura_Hashed_List_It<AVP_ServerEntity_StartPoint> xInputListConstIt( &xInputList ); 
	
	while( !xInputListConstIt.Done() )
	{
		// Check every possible start point to see if it is blocked by some other player
		AVP_ServerEntity_StartPoint* pxCurrentStartPoint = xInputListConstIt.GetCurrent();
		xInputListConstIt.Next(); 

		// Check if an actor has been spawned here recently
		Asura_Guid uLastUserGuid = pxCurrentStartPoint->GetLastActorSpawnedGuid(); 
		if (uLastUserGuid != ASURA_GUID_UNREGISTERED)
		{
			AVP_ServerEntity_Actor* pxServerActor = AVP_ServerEntity_Actor::FindEntity(uLastUserGuid); 
			AVP_ClientEntity_Actor* pxClientActor = AVP_ClientEntity_Actor::FindEntity(uLastUserGuid); 
			if (pxServerActor)
			{
				if (!pxClientActor || !pxClientActor->IsEntityActive() || !pxClientActor->GetPhysicsObject())
				{
					// An actor is about to be spawned here, but the physics system doesn't know yet
					// Don't use this spawn point
					continue;
				}
			}

			// This actor isn't around anymore or the physics system know about it
			// Don't need to track this Guid again
			pxCurrentStartPoint->SetLastActorSpawnedGuid(ASURA_GUID_UNREGISTERED); 
		}

		// This query will provide a list of physics objects in the vicinity
		Asura_Physics_Broadphase_Query xQuery;
		Asura_Bounding_Box xBox;
		xBox.SetToPoint(pxCurrentStartPoint->GetPosition());
		xBox.Grow(dbg_fBlockedDetectionRange);
		xQuery.MakeQuery(xBox, false, true, true, false);

		// Go through all the entries in turn (there shouldn't be that many).
		// Check if any of these entities is an actor blocking the spawn point
		bool bIsStartPointBlocked = false; 
		const Asura_Physics_Broadphase_Entry* pxEntry = xQuery.GetFirstEntry();
		while (pxEntry && !bIsStartPointBlocked)
		{
			const Asura_Guid uEntityGuid = pxEntry->GetInstance()->GetAssociatedEntityGuid();
			// Find if this is an actor
			AVP_ServerEntity_Actor* pxActor = AVP_ServerEntity_Actor::FindEntity( uEntityGuid ); 
			if (pxActor && !pxActor->IsDead())
			{
				bIsStartPointBlocked = true; 
			}
			pxEntry = pxEntry->GetNextEntry();
		}

		if (!bIsStartPointBlocked)
		{
			xOutputList.AddToList( pxCurrentStartPoint ); 
		}
	}

	// Check that we found at least an empty start point
	if (xOutputList.IsEmpty())
	{
		// TODO: find the best choice to spawn based on distance to other players, or number of them, proximity to buddies...
		// For now we don't exepct this to happen if the number of start points is enough, so just give the entire input list
		// as equally suitable spawn points
		Asura_Assert( false, "Couldn't find an spare start point, consider increasing the number of them in the map"); 
		
		// To mimic the previous behaviour just copy the list
		xInputListConstIt.Restart();
		while (!xInputListConstIt.Done())
		{
			xOutputList.AddToList( xInputListConstIt.GetCurrent() ); 
			xInputListConstIt.Next();
		}
	}
}

Asura_Hashed_List<AVP_ServerEntity_StartPoint>& AVP_Multiplayer_SpawnSystem::GetStartPointListForTeam( AVP_Settings::Team eTeam )
{
	switch (eTeam)
	{
		case AVP_Settings::AVP_TEAM_ONE:
		default:
		{
			return s_xTeamOneStartPointList;
		}
		case AVP_Settings::AVP_TEAM_TWO:
		{
			return s_xTeamTwoStartPointList;
		}
		case AVP_Settings::AVP_TEAM_THREE:
		{
			return s_xTeamThreeStartPointList;
		}
	}
}

Asura_Hashed_List<AVP_ServerEntity_StartPoint>& AVP_Multiplayer_SpawnSystem::GetStartPointListForSpecies( AVP_Settings::Species eSpecies )
{
	switch (eSpecies)
	{
		case AVP_Settings::AVP_SPECIES_ALIEN:
		default:
		{
			return s_xSpeciesAlienStartPointList;
		}
		case AVP_Settings::AVP_SPECIES_HUMAN:
		{
			return s_xSpeciesHumanStartPointList;
		}
		case AVP_Settings::AVP_SPECIES_PREDATOR:
		{
			return s_xSpeciesPredatorStartPointList;
		}
	}
}