
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_PhysicsObjectHandler.cpp
// Author : Andy Tate
// Date   : 4-Jun-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Physics/AVP_PhysicsObjectHandler.h"

// Asura independent header-files...
#include "Network/Asura_Network.h"
#include "Physics/Asura_Physics.h"

// Game independent header-files...
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ServerEntity_Actor.h"
#include "Mesh/AVP_PolygonFlags.h"
#include "Physics/AVP_CalculateSurfaceReorientationNormal_SphereCache.h"
#include "Physics/AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor.h"

////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

// Debug variables...
DEBUGVAR float	dbg_fLastReorientationDuration	= 0.125f;
DEBUGVAR float	dbg_fMaxClampedTimeStep			= 0.050f;


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::ValidateLocalSpaceOrientation
// Author       : Andy Tate
// Date         : 10-Mar-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::ValidateLocalSpaceOrientation(Asura_Quat& xOrientation, Space eSpaceOfOriginalOrientation) const
{
	// Validation...
	Asura_Assert((eSpaceOfOriginalOrientation == WORLD_SPACE) || ((Asura_Maths::Abs(xOrientation.x) < 0.0001f) && (Asura_Maths::Abs(xOrientation.z) < 0.0001f)), "Local-space orientation is not simply a rotation about the Y axis.");

	// Ensure that our orientation is purely around the Y axis.
	// NOTE: We need to do this to cope with numerical inaccuracies...
	xOrientation.x	= 0.0f;
	xOrientation.z	= 0.0f;

	xOrientation.Normalise();
}

/// End of function 'AVP_PhysicsObjectHandler::ValidateLocalSpaceOrientation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateRotationFromSurfaceNormal
// Author       : Andy Tate
// Date         : 29-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::CalculateRotationFromSurfaceNormal(Asura_AVector3_InParam xSurfaceNormal,
																  Asura_Quat& xRotation) const
{
	Asura_AVector3	xAxisZ;
	Asura_AVector3	xAxisX;
	Asura_AVector3	xAxisY;
	float			fDotProduct;

	// Ensure that our values are normalised...
	xAxisX.SetX(m_xSurfaceReorientationOrientation.e11);
	xAxisX.SetY(m_xSurfaceReorientationOrientation.e21);
	xAxisX.SetZ(m_xSurfaceReorientationOrientation.e31);

	xAxisY	= -xSurfaceNormal;
	xAxisY.Normalise();

	xAxisZ.SetX(m_xSurfaceReorientationOrientation.e13);
	xAxisZ.SetY(m_xSurfaceReorientationOrientation.e23);
	xAxisZ.SetZ(m_xSurfaceReorientationOrientation.e33);

	// Create our 3 orthogonal axes...
	fDotProduct	= xAxisY * xAxisX;

	if((fDotProduct > 0.866f) || (fDotProduct < -0.866f))
	{
		// Calculate our new side-vector...
		xAxisX	= xAxisY ^ xAxisZ;
		xAxisX.Normalise();

		// Calculate our new forwards-vector...
		xAxisZ	= xAxisX ^ xAxisY;
		xAxisZ.Normalise();
	}
	else
	{
		// Calculate our new forwards-vector...
		xAxisZ	= xAxisX ^ xAxisY;
		xAxisZ.Normalise();

		// Calculate our new side-vector...
		xAxisX	= xAxisY ^ xAxisZ;
		xAxisX.Normalise();

	}

	// Use these axes to calculate the orientation...
	m_xSurfaceReorientationOrientation.e11	= xAxisX.GetX();
	m_xSurfaceReorientationOrientation.e21	= xAxisX.GetY();
	m_xSurfaceReorientationOrientation.e31	= xAxisX.GetZ();

	m_xSurfaceReorientationOrientation.e12	= xAxisY.GetX();
	m_xSurfaceReorientationOrientation.e22	= xAxisY.GetY();
	m_xSurfaceReorientationOrientation.e32	= xAxisY.GetZ();

	m_xSurfaceReorientationOrientation.e13	= xAxisZ.GetX();
	m_xSurfaceReorientationOrientation.e23	= xAxisZ.GetY();
	m_xSurfaceReorientationOrientation.e33	= xAxisZ.GetZ();

	// ...and return this in a quaternion...
	xRotation.FromMatrix(m_xSurfaceReorientationOrientation);
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateRotationFromSurfaceNormal'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::UpdateSurfaceNormal
// Author       : Andy Tate
// Date         : 14-May-2007
////////////////////////////////////////////////////////////////////////////////
DEBUGVAR float dbg_fNoSurfaceToReorientateToAllowedUprightEpsilon = 0.6f;

bool AVP_PhysicsObjectHandler::UpdateSurfaceNormal(const float fTimeStep,
												   Asura_AVector3_InParam xInputVelocity,
												   Asura_AVector3_OutParam xOutputVelocity,
												   Asura_AVector3_OutParam xOutputSurfaceNormal)
{
	if ( m_fReorientingTimer > 0.0f )
	{
		m_fReorientingTimer -= fTimeStep;
	}
	Asura_AVector3 xFwd( m_pxPhysicsObject->GetOrientation().e13, m_pxPhysicsObject->GetOrientation().e23, m_pxPhysicsObject->GetOrientation().e33 );
	// Do we need to re-orientation to a surface?
	if(IsSurfaceReorientationEnabled())
	{
		// Find our normals and modify the velocity...
		if(m_bForceTargetSurfaceNormal)
		{
			m_fTimeSinceLastReoriented = 0.0f;

			AVP_SurfaceReorientationNormalInterpolator*	pxInterpolator	= m_pxFirstSurfaceReorientationNormalInterpolator;

			xOutputSurfaceNormal	= m_xForcedTargetSurfaceNormal;
			xOutputVelocity			= xInputVelocity;

			if(pxInterpolator == NULL)
			{
				return(true);
			}
			else
			{
				Asura_AVector3	xCurrentNormal(m_xCurrentSurfaceNormal);

				// Apply all of the necessary interpolators...
				do
				{
					// Update our interpolator...
					pxInterpolator->SetDesiredRotationAxis( xFwd );
					pxInterpolator->Update(fTimeStep, xOutputSurfaceNormal, xCurrentNormal);

					// Time for the next interpolator...
					pxInterpolator	= pxInterpolator->GetNext();

				} while(pxInterpolator != NULL);

				// Convert to a gravity-direction before returning...
				xOutputSurfaceNormal	= xCurrentNormal;
				return(true);
			}
		}
		else
		{
			u_int	uNumNormals;

			// Update the desired surface-normal...
			m_pxCalculateSurfaceReorientationNormal->Update(*m_pxPhysicsObject, xInputVelocity, m_xCurrentSurfaceNormal);

			// Do we have a direction to re-orient to?
			uNumNormals	= m_pxCalculateSurfaceReorientationNormal->GetNumNormalsUsed();

			if(uNumNormals > 0)
			{
				m_fTimeSinceLastReoriented = 0.0f;

				AVP_SurfaceReorientationNormalInterpolator*	pxInterpolator	= m_pxFirstSurfaceReorientationNormalInterpolator;

				m_pxCalculateSurfaceReorientationNormal->GetDesiredNormal(xOutputSurfaceNormal);
				m_pxCalculateSurfaceReorientationNormal->GetDesiredVelocity(xOutputVelocity);

				if(pxInterpolator == NULL)
				{
					return(true);
				}
				else
				{
					Asura_AVector3	xCurrentNormal(m_xCurrentSurfaceNormal);

					// Apply all of the necessary interpolators...
					do
					{
						// Update our interpolator...
						pxInterpolator->SetDesiredRotationAxis( xFwd );
						pxInterpolator->Update(fTimeStep, *m_pxCalculateSurfaceReorientationNormal, xCurrentNormal);

						// Time for the next interpolator...
						pxInterpolator	= pxInterpolator->GetNext();

					} while(pxInterpolator != NULL);

					// Convert to a gravity-direction before returning...
					xOutputSurfaceNormal	= xCurrentNormal;
					return(true);
				}
			}
			else
			{
				const bool bTimeWasLessThanThreshold = ( m_fTimeSinceLastReoriented <= dbg_fLastReorientationDuration );
				m_fTimeSinceLastReoriented	+= fTimeStep;
				if (m_fTimeSinceLastReoriented > dbg_fLastReorientationDuration)
				{
					// OK, we've been away from a surface for too long. Reset to upright.
					AVP_SurfaceReorientationNormalInterpolator*	pxInterpolator	= m_pxFirstSurfaceReorientationNormalInterpolator;

					xOutputSurfaceNormal	= -Asura_Physics_System::GetGravityDirection();
					xOutputVelocity			= xInputVelocity;

					if(pxInterpolator == NULL)
					{
						return(true);
					}
					else
					{
						Asura_AVector3	xCurrentNormal(m_xCurrentSurfaceNormal);

						// Apply all of the necessary interpolators...
						do
						{
							// Update our interpolator...
							pxInterpolator->SetDesiredRotationAxis( xFwd );
							pxInterpolator->Update(fTimeStep, xOutputSurfaceNormal, xCurrentNormal);

							// Time for the next interpolator...
							pxInterpolator	= pxInterpolator->GetNext();

						} while(pxInterpolator != NULL);

						// Convert to a gravity-direction before returning...
						xOutputSurfaceNormal	= xCurrentNormal;
					}

					if ( bTimeWasLessThanThreshold &&
						!m_xCurrentSurfaceNormal.ApproximatelyEquals( -Asura_Physics_System::GetGravityDirection(),
																	  dbg_fNoSurfaceToReorientateToAllowedUprightEpsilon ) )
					{
						AVP_ServerEntity_Actor* pxAssociatedActor = AVP_ServerEntity_Actor::FindEntity( m_pxPhysicsObject->GetAssociatedEntityGuid() );
						if ( pxAssociatedActor )
						{
							Asura_Message* pxMsg = Asura_Network::CreateMessage( AVP_MSGID_PHYSICS_REORIENTATION_RESET );
							if ( pxMsg )
							{
								pxAssociatedActor->HandleMessage( pxMsg );
								Asura_Network::DestroyMessage( pxMsg );
							}
						}
					}
				}
				else
				{
					// Allow a little leeway to reaquire a surface.
					xOutputVelocity			= xInputVelocity;
					xOutputSurfaceNormal	= m_xCurrentSurfaceNormal;
				}

				// Allow this while the timer is still increasing or if we're already upright.
				Asura_Assertf( ( m_pxPhysicsObject->GetAssociatedEntityGuid() == ASURA_GUID_UNREGISTERED ) ||
							   ( m_fTimeSinceLastReoriented <= dbg_fLastReorientationDuration ) ||
							     m_xCurrentSurfaceNormal.ApproximatelyEquals( -Asura_Physics_System::GetGravityDirection(),
																			  dbg_fNoSurfaceToReorientateToAllowedUprightEpsilon ),
							   ( "Reorientating entity %d has failed to find a surface to reorient to for too long. Resetting to upright.\n\n"
							     "Position: %.2f, %.2f, %.2f\n\nOrientation %.2f, %.2f, %.2f\n            %.2f, %.2f, %.2f\n"
								 "            %.2f, %.2f, %.2f",
								 m_pxPhysicsObject->GetAssociatedEntityGuid(),
								 m_pxPhysicsObject->GetPosition().x,
								 m_pxPhysicsObject->GetPosition().y,
								 m_pxPhysicsObject->GetPosition().z,
								 m_pxPhysicsObject->GetOrientation().e11,
								 m_pxPhysicsObject->GetOrientation().e12,
								 m_pxPhysicsObject->GetOrientation().e13,
								 m_pxPhysicsObject->GetOrientation().e21,
								 m_pxPhysicsObject->GetOrientation().e22,
								 m_pxPhysicsObject->GetOrientation().e23,
								 m_pxPhysicsObject->GetOrientation().e31,
								 m_pxPhysicsObject->GetOrientation().e32,
								 m_pxPhysicsObject->GetOrientation().e33 ) );

				return(false);
			}
		}
	}
	else
	{
		m_fTimeSinceLastReoriented	= 0.0f;
		xOutputVelocity				= xInputVelocity;
		xOutputSurfaceNormal		= m_xCurrentSurfaceNormal;
		return(true);
	}
}

/// End of function 'AVP_PhysicsObjectHandler::UpdateSurfaceNormal'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::AVP_PhysicsObjectHandler
// Author       : Andy Tate
// Date         : 8-Oct-2007
////////////////////////////////////////////////////////////////////////////////

AVP_PhysicsObjectHandler::AVP_PhysicsObjectHandler(Asura_Physics_Object* pxPhysicsObject)
: m_fReorientingTimer( 0.0f )
, m_ucExternalControlRequests( 0 )
{
	// Initialise this class...
	Initialise(pxPhysicsObject);
}

/// End of function 'AVP_PhysicsObjectHandler::AVP_PhysicsObjectHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::~AVP_PhysicsObjectHandler
// Author       : Andy Tate
// Date         : 4-Jun-2007
////////////////////////////////////////////////////////////////////////////////

AVP_PhysicsObjectHandler::~AVP_PhysicsObjectHandler()
{
	// Destroy the environment-cache (if we have one)...
	delete(m_pxCalculateSurfaceReorientationNormal);
	delete(m_pxFirstSurfaceReorientationNormalInterpolator);
}

/// End of function 'AVP_PhysicsObjectHandler::~AVP_PhysicsObjectHandler'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::Initialise
// Author       : Andy Tate
// Date         : 4-Jun-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::Initialise(Asura_Physics_Object* pxPhysicsObject)
{
	// Initialise the physics object...
	m_pxPhysicsObject	= pxPhysicsObject;

	// Initialise the variables...
	m_pxCalculateSurfaceReorientationNormal			= NULL;
	m_pxFirstSurfaceReorientationNormalInterpolator	= NULL;

	m_xSurfaceReorientationOrientation.Identity();
	m_xCurrentLocalOrientation.Identity();

	m_xCurrentSurfaceNormal	= (m_pxPhysicsObject == NULL)
							? Asura_AVector3(0.0f, -1.0f, 0.0f)
							: -m_pxPhysicsObject->GetGravityDirection();
	m_xCurrentSurfaceNormal.Normalise();

	m_xForcedTargetSurfaceNormal = Asura_Vector_3(0.0f, 1.0f, 0.0f);
	m_bForceTargetSurfaceNormal	= false;
	m_bForceGravity = false;

	m_fTimeSinceLastReoriented = 0.0f;

	m_eDefaultPhysicsObjectType = m_pxPhysicsObject ? m_pxPhysicsObject->GetType() : ASURA_PHYSICS_OBJECT_EXTERNALCONTROL;
}

/// End of function 'AVP_PhysicsObjectHandler::Initialise'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::EnableSurfaceReorientation
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::EnableSurfaceReorientation(const Asura_Angle fLimitAngle)
{
    const Asura_Matrix_3x3&	xOrientation	= m_pxPhysicsObject->GetOrientation();
	const u_short			usFlagsToIgnore	= m_pxPhysicsObject->GetFlagsOfPolysToIgnore();
	const float				fCosLimit		= Asura_Maths::Cosine(fLimitAngle);

	// Initialise the appropriate data...
	delete(m_pxCalculateSurfaceReorientationNormal);
	delete(m_pxFirstSurfaceReorientationNormalInterpolator);

	m_pxCalculateSurfaceReorientationNormal			= new AVP_CalculateSurfaceReorientationNormal_SphereCache (fCosLimit, usFlagsToIgnore | AVP_PolygonFlag_IgnoreForReorientation);
	m_pxFirstSurfaceReorientationNormalInterpolator	= new AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor;

	// Update the physics-object...
	m_pxPhysicsObject->SetReportObjectCollisions(true);
	m_pxPhysicsObject->SetUseOwnGravityDirection(true);
	m_pxPhysicsObject->SetActorIsSuperSlidy(false);
	m_pxPhysicsObject->SetAllSurfacesWalkable(true);
	m_pxPhysicsObject->SetNotWalkableSurfaceFlags(usFlagsToIgnore | AVP_PolygonFlag_IgnoreForReorientation);

	// Finally, ensure that we're starting off correctly initialised...
	ResetUsingWorldSpaceOrientation(xOrientation);
}

/// End of function 'AVP_PhysicsObjectHandler::EnableSurfaceReorientation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::DisableSurfaceReorientation
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::DisableSurfaceReorientation(void)
{
	// Update the physics-object...
	m_pxPhysicsObject->SetReportObjectCollisions(false);
	m_pxPhysicsObject->SetUseOwnGravityDirection(false);
	m_pxPhysicsObject->SetActorIsSuperSlidy(true);
	m_pxPhysicsObject->SetAllSurfacesWalkable(false);
	m_pxPhysicsObject->SetNotWalkableSurfaceFlags(0);

	// Remove the handlers...
	delete(m_pxCalculateSurfaceReorientationNormal);
	delete(m_pxFirstSurfaceReorientationNormalInterpolator);

	m_pxCalculateSurfaceReorientationNormal			= NULL;
	m_pxFirstSurfaceReorientationNormalInterpolator	= NULL;
	m_xSurfaceReorientationOrientation.Identity();
}

/// End of function 'AVP_PhysicsObjectHandler::DisableSurfaceReorientation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::EnableExternalControlPhysics
// Author       : Anthony Gregson
// Date         : 10/10/07
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::EnableExternalControlPhysics( u_char ucRequesterID )
{
	if ( m_ucExternalControlRequests == 0 && m_pxPhysicsObject != NULL )
	{
		// Init default physics object type
		if( m_eDefaultPhysicsObjectType == NUM_ASURA_PHYSICS_OBJECT_TYPES )
		{
			m_eDefaultPhysicsObjectType = m_pxPhysicsObject->GetType();
			Asura_Assert( m_eDefaultPhysicsObjectType != NUM_ASURA_PHYSICS_OBJECT_TYPES, "AVP_PhysicsObjectHandler::EnableExternalControlPhysics - bad physics object type" );
		}

		m_pxPhysicsObject->ChangeType( ASURA_PHYSICS_OBJECT_EXTERNALCONTROL );
	}
	m_ucExternalControlRequests |= ucRequesterID;
}

/// End of function 'AVP_PhysicsObjectHandler::EnableExternalControlPhysics'. ///

void AVP_PhysicsObjectHandler::SetExtCtrlRequests( const u_int uRequests )
{
	m_ucExternalControlRequests = static_cast< u_char >( uRequests );
	if ( m_ucExternalControlRequests && m_pxPhysicsObject != NULL )
	{
		// Init default physics object type
		if( m_eDefaultPhysicsObjectType == NUM_ASURA_PHYSICS_OBJECT_TYPES )
		{
			m_eDefaultPhysicsObjectType = m_pxPhysicsObject->GetType();
			Asura_Assert( m_eDefaultPhysicsObjectType != NUM_ASURA_PHYSICS_OBJECT_TYPES, "AVP_PhysicsObjectHandler::EnableExternalControlPhysics - bad physics object type" );
		}

		m_pxPhysicsObject->ChangeType( ASURA_PHYSICS_OBJECT_EXTERNALCONTROL );
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::DisableExternalControlPhysics
// Author       : Anthony Gregson
// Date         : 10/10/07
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::DisableExternalControlPhysics( u_char ucRequesterID )
{
	m_ucExternalControlRequests &= ~ucRequesterID;
	if ( m_ucExternalControlRequests == 0 && m_pxPhysicsObject != NULL )
	{
		// At this point, anything using this class should be an actor under normal circumstances
		Asura_Assert( m_eDefaultPhysicsObjectType != NUM_ASURA_PHYSICS_OBJECT_TYPES, "AVP_PhysicsObjectHandler::DisableExternalControlPhysics - bad physics object type" );
		m_pxPhysicsObject->ChangeType( m_eDefaultPhysicsObjectType );

		// talk to Kev about this, it should probably be done in the physics object
		m_pxPhysicsObject->SetLinVelocity( m_pxPhysicsObject->GetLinVelocity() );
	}
}

/// End of function 'AVP_PhysicsObjectHandler::DisableExternalControlPhysics'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::CalculateMovementUsingOrientationAndLinVelocity
// Author       : Andy Tate
// Date         : 10-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::CalculateMovementUsingOrientationAndLinVelocity(const float fTimeStep,
																			   const Asura_Quat& xCurrentOrientation,
																			   Asura_AVector3_InParam xCurrentLinVelocity,
																			   const bool bFeelsGravity,
																			   const Space eSpace)
{
	const float		fClampedTimeStep	= Asura_Maths::Min(fTimeStep, dbg_fMaxClampedTimeStep);

	Asura_Quat		xNewWSOrientation;
	Asura_AVector3	xNewWSSurfaceNormal;
	Asura_AVector3	xNewWSLinVelocity;
	Asura_AVector3	xCurrentWSLinVelocity;
	const Asura_Matrix_3x3& xOri = m_pxPhysicsObject->GetOrientation();
	// Validation...
	Asura_Assert((xCurrentLinVelocity.Magnitude() < 250.0f), "Suspiciously large velocity"); // 250m/s is roughly 558mph.

	// Calculate the velocity in world-space...
	switch(eSpace)
	{
		case LOCAL_SPACE:
		{
			const Asura_Quat	xWSOrientation(xOri);
			Asura_Quat			xInverseCurrentLocalOrientation(m_xCurrentLocalOrientation);
			Asura_Quat			xWSRotation;

			xInverseCurrentLocalOrientation.Invert();

			xWSRotation	= xWSOrientation * xInverseCurrentLocalOrientation;
			xCurrentWSLinVelocity = xWSRotation.RotateVector(xCurrentLinVelocity);

			// Calculate the gravity direction...
			// AJG - 26/06/08 - We now handle the failure case within the update function so that we can still use its reorientation. As such, its return value is now ignored.
			UpdateSurfaceNormal(fClampedTimeStep, xCurrentWSLinVelocity, xNewWSLinVelocity, xNewWSSurfaceNormal);
			break;
		}

		case WORLD_SPACE:
		{
			xNewWSLinVelocity = xCurrentLinVelocity;
			xCurrentWSLinVelocity = xCurrentLinVelocity;
			xNewWSSurfaceNormal = xCurrentOrientation.RotateVector( Asura_AVector3( 0.0f, -1.0f, 0.0f ) );
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected space used when calculating movement for physics-object.");
			return;
		}
	}		// End of switch.

	if ( !m_xCurrentSurfaceNormal.ApproximatelyEquals( xNewWSSurfaceNormal, 0.001f ) )
	{
		m_fReorientingTimer = 0.2f;
	}

	Asura_Quat	xOrientationFromSurfaceNormal;

	// Time to calculate the orientation...
	CalculateRotationFromSurfaceNormal(xNewWSSurfaceNormal, xOrientationFromSurfaceNormal);

	// Set the current local-orientation to use...
	switch(eSpace)
	{
		case LOCAL_SPACE:
		{
			m_xCurrentLocalOrientation	= xCurrentOrientation;
			xNewWSOrientation			= xOrientationFromSurfaceNormal * xCurrentOrientation;
			break;
		}

		case WORLD_SPACE:
		{
			Asura_Quat	xInverseOrientationFromSurfaceNormal(xOrientationFromSurfaceNormal);

			xInverseOrientationFromSurfaceNormal.Invert();
			m_xCurrentLocalOrientation	= xInverseOrientationFromSurfaceNormal * xCurrentOrientation;
			xNewWSOrientation			= xCurrentOrientation;
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected space used when calculating movement for physics-object.");
			return;
		}
	}		// End of switch.

	ValidateLocalSpaceOrientation(m_xCurrentLocalOrientation, eSpace);

	Asura_Assert((xNewWSLinVelocity.Magnitude() < 250.0f), "xNewWSLinVelocity is too large.");

	xNewWSSurfaceNormal.Normalise();
    xNewWSOrientation.Normalise();

	// Update the physics object...

	//little hack for now, re-orient around the bounding box centre instead of the compass
	Asura_Vector_3 xBBCentre = m_pxPhysicsObject->GetBoundingBox().CalculateCentre();
	Asura_Vector_3 xOffset = m_pxPhysicsObject->GetPosition() - xBBCentre;
	Asura_Quat xRot;
	xRot.RotationFromAToB( m_xCurrentSurfaceNormal, xNewWSSurfaceNormal );
	xOffset = xRot.RotateVector( xOffset);
	m_pxPhysicsObject->SetPosition( xBBCentre + xOffset );

	m_pxPhysicsObject->SetIgnore(false);
	m_xOrientationPassedIntoPhysics = xNewWSOrientation;
	m_pxPhysicsObject->SetOrientation(xNewWSOrientation);
	m_pxPhysicsObject->SetFeelsGravity(bFeelsGravity);
	m_pxPhysicsObject->SetLinVelocity(Asura_Vector_3(xNewWSLinVelocity));

	if(bFeelsGravity)
	{
		if(m_bForceGravity)
		{
			m_pxPhysicsObject->SetOwnGravityDirection( m_xForcedGravity);
		}
		else if(m_bForceTargetSurfaceNormal)
		{
			m_pxPhysicsObject->SetOwnGravityDirection( -m_xForcedTargetSurfaceNormal );
		}
		else if ( m_pxCalculateSurfaceReorientationNormal )
		{
			Asura_AVector3 xDesiredSurfaceNormal;
			m_pxCalculateSurfaceReorientationNormal->GetDesiredNormal( xDesiredSurfaceNormal );
			m_pxPhysicsObject->SetOwnGravityDirection( -xDesiredSurfaceNormal );
		}
		else
		{
			m_pxPhysicsObject->SetUseOwnGravityDirection( false );
		}
	}

	if (m_pxPhysicsObject->GetType() == ASURA_PHYSICS_OBJECT_EXTERNALCONTROL)
	{
		m_pxPhysicsObject->SetPosition(m_pxPhysicsObject->GetPosition() + (Asura_Vector_3(xNewWSLinVelocity) * fClampedTimeStep));
	}

	// Finally, store the current values...
	m_xCurrentSurfaceNormal	= xNewWSSurfaceNormal;
}

/// End of function 'AVP_PhysicsObjectHandler::CalculateMovementUsingOrientationAndLinVelocity'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::CalculateMovementUsingAngAndLinVelocity
// Author       : Andy Tate
// Date         : 16-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::CalculateMovementUsingAngAndLinVelocity(const float fTimeStep,
																	   Asura_AVector3_InParam xCurrentAngVelocityAxis,
																	   const Asura_Angle fCurrentAngVelocityAngle,
																	   Asura_AVector3_InParam xCurrentLinVelocity,
																	   const bool bFeelsGravity,
																	   const Space eSpace)
{
	const float			fClampedTimeStep	= Asura_Maths::Min(fTimeStep, dbg_fMaxClampedTimeStep);
	const Asura_Angle	fRotation			= fCurrentAngVelocityAngle * fClampedTimeStep;

	Asura_Quat			xRotation;
	xRotation.Identity();
	// Calculate the orientation to use for the movement...
	if ( fRotation != 0.0f )
	{
		xRotation.FromRotation(fRotation, xCurrentAngVelocityAxis);
	}

	switch(eSpace)
	{
		case OBJECT_SPACE:
		{
			const Asura_Quat	xCurrentLSOrientation(m_xCurrentLocalOrientation * xRotation);

			// Transform everything into world-space...
			Asura_AVector3 xCurrentLSLinVelocity = m_xCurrentLocalOrientation.RotateVector(xCurrentLinVelocity);

			CalculateMovementUsingOrientationAndLinVelocity(fClampedTimeStep, xCurrentLSOrientation, xCurrentLSLinVelocity, bFeelsGravity, LOCAL_SPACE);
			break;
		}

		case LOCAL_SPACE:
		{
			Asura_Quat	xCurrentLSOrientation(xRotation * m_xCurrentLocalOrientation);

			CalculateMovementUsingOrientationAndLinVelocity(fClampedTimeStep, xCurrentLSOrientation, xCurrentLinVelocity, bFeelsGravity, LOCAL_SPACE);
			break;
		}

		case WORLD_SPACE:
		{
			Asura_Quat	xCurrentWSOrientation(m_pxPhysicsObject->GetOrientation());

			xCurrentWSOrientation	= xCurrentWSOrientation * xRotation;
			CalculateMovementUsingOrientationAndLinVelocity(fClampedTimeStep, xCurrentWSOrientation, xCurrentLinVelocity, bFeelsGravity, WORLD_SPACE);
			break;
		}

		default:
		{
			Asura_Assert(false, "Unexpected space used when calculating movement for physics-object.");
			break;
		}
	}		// End of switch.
}

/// End of function 'AVP_PhysicsObjectHandler::CalculateMovementUsingAngAndLinVelocity'. ///


void AVP_PhysicsObjectHandler::CalculateMovementUsingOrientationAndPosition( const float fTimeStep,
                                                                             const Asura_Quat& xCurrentOrientation,
                                                                             Asura_AVector3_InParam xCurrentPosition,
                                                                             const bool bFeelsGravity,
                                                                             const Space eSpace )
{
    Asura_Assert( m_pxPhysicsObject->GetType() == ASURA_PHYSICS_OBJECT_EXTERNALCONTROL,
        "AVP_PhysicsObjectHandler::CalculateMovementUsingOrientationAndPosition - object should be external control" );
    Asura_Assert( eSpace == WORLD_SPACE,
        "AVP_PhysicsObjectHandler::CalculateMovementUsingOrientationAndPosition - arguments should be in world space" );

	Asura_Quat		xNewWSOrientation;
	Asura_AVector3	xNewWSSurfaceNormal;
	Asura_AVector3	xCurrentWSPosition = xCurrentPosition;

	// Calculate the gravity direction from the ws orientation...
    {
		Asura_Matrix_3x3 xWSOrientationMat;

		xCurrentOrientation.ToMatrix( xWSOrientationMat );
		xNewWSSurfaceNormal.SetX( -xWSOrientationMat.e12 );
		xNewWSSurfaceNormal.SetY( -xWSOrientationMat.e22 );
		xNewWSSurfaceNormal.SetZ( -xWSOrientationMat.e32 );
    }

    {
		Asura_Quat xOrientationFromSurfaceNormal;

		// Time to calculate the orientation...
		CalculateRotationFromSurfaceNormal(xNewWSSurfaceNormal, xOrientationFromSurfaceNormal);

		// Set the current local-orientation to use...
		Asura_Quat xInverseOrientationFromSurfaceNormal(xOrientationFromSurfaceNormal);

		xInverseOrientationFromSurfaceNormal.Invert();
		m_xCurrentLocalOrientation = xInverseOrientationFromSurfaceNormal * xCurrentOrientation;
		xNewWSOrientation		   = xCurrentOrientation;

		ValidateLocalSpaceOrientation(m_xCurrentLocalOrientation, eSpace);

		// We're re-orienting - reset our timer...
		m_fTimeSinceLastReoriented	= 0.0f;
	}

	xNewWSSurfaceNormal.Normalise();
    xNewWSOrientation.Normalise();

	// Update the physics object...
	m_pxPhysicsObject->SetIgnore(false);
	m_pxPhysicsObject->SetOrientation(xNewWSOrientation);
	m_xOrientationPassedIntoPhysics = xNewWSOrientation;
	m_pxPhysicsObject->SetFeelsGravity(bFeelsGravity);

	if(bFeelsGravity)
	{
		m_pxPhysicsObject->SetOwnGravityDirection(-xNewWSSurfaceNormal);
	}

	m_pxPhysicsObject->SetPosition( Asura_Vector_3( xCurrentWSPosition ) );

	// Finally, store the current values...
	m_xCurrentSurfaceNormal = xNewWSSurfaceNormal;
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::ResetUsingWorldSpaceOrientation
// Author       : Andy Tate
// Date         : 18-May-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_PhysicsObjectHandler::ResetUsingWorldSpaceOrientation(const Asura_Quat& xOrientation)
{
	Asura_Quat	xCurrentOrientation(xOrientation);

	// Reset the variables...
	xCurrentOrientation.Normalise();
	xCurrentOrientation.ToMatrix(m_xSurfaceReorientationOrientation);

	m_xCurrentSurfaceNormal.SetX(-m_xSurfaceReorientationOrientation.e12);
	m_xCurrentSurfaceNormal.SetY(-m_xSurfaceReorientationOrientation.e22);
	m_xCurrentSurfaceNormal.SetZ(-m_xSurfaceReorientationOrientation.e32);
	m_xCurrentSurfaceNormal.Normalise();

	if(m_pxFirstSurfaceReorientationNormalInterpolator != NULL)
	{
		m_pxFirstSurfaceReorientationNormalInterpolator->Reset(m_xCurrentSurfaceNormal);
	}

	// Local orientation starts as the identity - this assumes that the direction
	// we're facing is handled by "m_xSurfaceReorientationOrientation"...
	m_xCurrentLocalOrientation.Identity();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_PhysicsObjectHandler::SetWalkWalkingDisabled
////////////////////////////////////////////////////////////////////////////////
void AVP_PhysicsObjectHandler::SetWalkWalkingDisabled( bool bDisable )
{ 
	if (m_pxCalculateSurfaceReorientationNormal) 
	{
		m_pxCalculateSurfaceReorientationNormal->SetWalkWalkingDisabled(bDisable); 
	} 
	m_pxPhysicsObject->SetAllSurfacesWalkable(!bDisable);
}


/// End of function 'AVP_PhysicsObjectHandler::ResetUsingWorldSpaceOrientation'. ///


/// End of file. ///
