
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_PhysicsObjectHandler.h
// Author : Andy Tate
// Date   : 4-Jun-2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif   // ifdef ASURA_USE_PRAGMA_ONCE

#ifndef __AVP_PHYSICSOBJECTHANDLER_H__
#define __AVP_PHYSICSOBJECTHANDLER_H__

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura_Maths.h"
#include "Core/Asura_Timers.h"
#include "Physics/Asura_Physics_Object.h"


////////////////////////////////////////////////////////////////////////////////
// Forward class declarations...
////////////////////////////////////////////////////////////////////////////////

class AVP_CalculateSurfaceReorientationNormal;
class AVP_SurfaceReorientationNormalInterpolator;


////////////////////////////////////////////////////////////////////////////////
// Class declarations...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_PhysicsObjectHandler
// Author       : Andy Tate
// Date         : 4-Jun-2007
//
// Description  : Class to handle the manipulation of the physics-object, with
//				  particular responsibility for handling surface re-orientation.
////////////////////////////////////////////////////////////////////////////////

class AVP_PhysicsObjectHandler
{
public:
	// Public enumerated types...
	enum Space
	{
		OBJECT_SPACE,
		LOCAL_SPACE,
		WORLD_SPACE,

		MAX_SPACES
	};

	enum ExternalControlRequesters
	{
		ECREQ_GAMESCENE = 0x01,
		ECREQ_MOVE		= 0x02
		// this is a stored in a bit field so we know who's asked us to be in external control, so the next one will be:
		// BLAH			= 0x04
	};

	// Public functions...
	AVP_PhysicsObjectHandler(Asura_Physics_Object* pxPhysicsObject);
	virtual ~AVP_PhysicsObjectHandler();

	// Manipulation functions...
	void	EnableSurfaceReorientation(const Asura_Angle fLimitAngle = 2.269f);	// 130 degrees as radians.
	void	DisableSurfaceReorientation(void);

	bool	IsSurfaceReorientationEnabled(void) const	{ return(m_pxCalculateSurfaceReorientationNormal != NULL); }

	void	EnableExternalControlPhysics( u_char ucRequester ); // Note that changing to this effectively turns off gravity and colliding with the environment and other objects (other objects still collide with us though).
	void	DisableExternalControlPhysics( u_char ucRequester );

	void	CalculateMovementUsingAngAndLinVelocity(const float fTimeStep, Asura_AVector3_InParam xCurrentAngVelocityAxis, const Asura_Angle fCurrentAngVelocityAngle, Asura_AVector3_InParam xCurrentLinVelocity, const bool bFeelsGravity, const Space eSpace);
	void	CalculateMovementUsingOrientationAndLinVelocity(const float fTimeStep, const Asura_Quat& xCurrentOrientation, Asura_AVector3_InParam xCurrentLinVelocity, const bool bFeelsGravity, const Space eSpace);
    void    CalculateMovementUsingOrientationAndPosition(const float fTimeStep, const Asura_Quat& xCurrentOrientation, Asura_AVector3_InParam xCurrentPosition, const bool bFeelsGravity, const Space eSpace);

	const Asura_Quat& GetOrientationPassedIntoPhysics() const { return m_xOrientationPassedIntoPhysics; }

	// Spacial-transfer functions...
	void RotateWorldSpaceVectorIntoLocalSpace(const Asura_Vector_3& xWSVec, Asura_Vector_3& xLSVecOut) const
	{
		xLSVecOut = m_xSurfaceReorientationOrientation.ApplyTranspose( xWSVec );
	}

	void RotateLocalSpaceVectorIntoWorldSpace(const Asura_Vector_3& xLSVec, Asura_Vector_3& xWSVecOut) const
	{
		xWSVecOut = xLSVec * m_xSurfaceReorientationOrientation;
	}

	void RotateWorldSpaceOrientationIntoLocalSpace(const Asura_Quat& xWSOriIn, Asura_Quat& xLSOriOut) const
	{
		Asura_Quat	xInverseOri(m_xSurfaceReorientationOrientation);

		xInverseOri.Invert();
		xLSOriOut	= xInverseOri * xWSOriIn;

		ValidateLocalSpaceOrientation(xLSOriOut, WORLD_SPACE);
	}

	void RotateLocalSpaceOrientationIntoWorldSpace(const Asura_Quat& xLSOriIn, Asura_Quat& xWSOriOut) const
	{
		const Asura_Quat	xOri(m_xSurfaceReorientationOrientation);
		Asura_Quat			xLSOri(xLSOriIn);

		ValidateLocalSpaceOrientation(xLSOri, LOCAL_SPACE);
		xWSOriOut	= xOri * xLSOri;
	}

	// Reset functions...
	virtual void	Initialise(Asura_Physics_Object* pxPhysicsObject);
	virtual void	ResetUsingWorldSpaceOrientation(const Asura_Quat& xOrientation);

	void			ResetUsingWorldSpaceOrientation(const Asura_Matrix_3x3& xOrientation)	{ const Asura_Quat xQuat(xOrientation); ResetUsingWorldSpaceOrientation(xQuat); }

	// Accessor functions...
	Asura_Physics_Object*		GetPhysicsObject(void)			{ return(m_pxPhysicsObject); }
	const Asura_Physics_Object*	GetPhysicsObject(void) const	{ return(m_pxPhysicsObject); }

	AVP_CalculateSurfaceReorientationNormal*		GetCalculateSurfaceReorientationNormal(void)		{ return(m_pxCalculateSurfaceReorientationNormal); }
	const AVP_CalculateSurfaceReorientationNormal*	GetCalculateSurfaceReorientationNormal(void) const	{ return(m_pxCalculateSurfaceReorientationNormal); }

	AVP_SurfaceReorientationNormalInterpolator*			GetFirstSurfaceReorientationNormalInterpolator(void)		{ return(m_pxFirstSurfaceReorientationNormalInterpolator); }
	const AVP_SurfaceReorientationNormalInterpolator*	GetFirstSurfaceReorientationNormalInterpolator(void) const	{ return(m_pxFirstSurfaceReorientationNormalInterpolator); }

	void	GetCurrentGravityDirection(Asura_AVector3& xDirection) const	{ xDirection = -m_xCurrentSurfaceNormal; }
	void	GetCurrentGravityDirection(Asura_Vector_3& xDirection) const	{ xDirection = -m_xCurrentSurfaceNormal; }
	void	GetCurrentSurfaceNormal(Asura_AVector3& xNormal) const			{ xNormal = m_xCurrentSurfaceNormal; }
	void	GetCurrentSurfaceNormal(Asura_Vector_3& xNormal) const			{ xNormal = m_xCurrentSurfaceNormal; }

	void 	SetForcedTargetSurfaceNormal(const Asura_Vector_3& xTargetNormal)	{ m_bForceTargetSurfaceNormal = true; m_xForcedTargetSurfaceNormal = xTargetNormal; }
	void 	SetForcedTargetSurfaceNormal(const Asura_AVector3& xTargetNormal)	{ m_bForceTargetSurfaceNormal = true; m_xForcedTargetSurfaceNormal = xTargetNormal; }
	void 	UnsetForcedTargetSurfaceNormal(void)								{ m_bForceTargetSurfaceNormal = false; }
	void 	SetForcedGravity(const Asura_Vector_3& xGravity)	{ m_bForceGravity = true; m_xForcedGravity = xGravity; }
	void 	SetForcedGravity(const Asura_AVector3& xGravity)	{ m_bForceGravity = true; m_xForcedGravity = xGravity; }
	void 	UnsetForcedGravity(void)							{ m_bForceGravity = false; }

	bool	IsReorienting() const { return m_fReorientingTimer > 0.0f ; }

	void SetWalkWalkingDisabled( bool bDisable );

	u_int GetExtCtrlRequests() const { return m_ucExternalControlRequests; }
	// Used to re-establish the external control requests after a chunk load.
	void  SetExtCtrlRequests( const u_int uRequests );

protected:
	// Protected variables...
	Asura_Physics_Object*							m_pxPhysicsObject;									// A pointer to the physics object we're using.
	AVP_CalculateSurfaceReorientationNormal*		m_pxCalculateSurfaceReorientationNormal;			// A pointer to the class used to calculate the normal to use for surface-reorientation.
	AVP_SurfaceReorientationNormalInterpolator*		m_pxFirstSurfaceReorientationNormalInterpolator;	// A pointer to the first in a list of interpolators to use for the normal during surface-reorientation.

	float											m_fTimeSinceLastReoriented;							// The number of seconds since we last re-oriented.
	mutable Asura_Matrix_3x3						m_xSurfaceReorientationOrientation;					// The orientation used when re-orienting to a surface.

	Asura_Quat										m_xCurrentLocalOrientation;							// The current value of the local-orientation.
	Asura_AVector3									m_xCurrentSurfaceNormal;							// The current value of the surface-normal.

	Asura_AVector3									m_xForcedTargetSurfaceNormal;
	bool											m_bForceTargetSurfaceNormal;

	Asura_AVector3									m_xForcedGravity;
	bool											m_bForceGravity;

	ASURA_PHYSICS_OBJECT_TYPE						m_eDefaultPhysicsObjectType;
	Asura_Quat										m_xOrientationPassedIntoPhysics;

	// Protected functions...
	void	ValidateLocalSpaceOrientation(Asura_Quat& xOrientation, Space eSpaceOfOriginalOrientation) const;

	void	CalculateRotationFromSurfaceNormal(Asura_AVector3_InParam xSurfaceNormal, Asura_Quat& xRotation) const;
	bool	UpdateSurfaceNormal(const float fTimeStep, Asura_AVector3_InParam xInputVelocity, Asura_AVector3_OutParam xOutputVelocity, Asura_AVector3_OutParam xOutputSurfaceNormal);


private:
	// Private variables...
	float											m_fReorientingTimer;
	u_char											m_ucExternalControlRequests;

	// Private functions...
	AVP_PhysicsObjectHandler  (void);
	AVP_PhysicsObjectHandler  (const AVP_PhysicsObjectHandler& xOther);

	AVP_PhysicsObjectHandler&	operator=	(const AVP_PhysicsObjectHandler& xOther);

};   // End of class 'AVP_PhysicsObjectHandler'.


#endif   // ifndef __AVP_PHYSICSOBJECTHANDLER_H__


/// End of file. ///
