
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor.cpp
// Author : Andy Tate
// Date   : 7-Feb-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Physics/AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor.h"

// Asura independent header-files...
#include "Debugging/Asura_Logfile.h"

// Game independent header-files...
#include "Core/AVP_Maths.h"
#include "Physics/AVP_CalculateSurfaceReorientationNormal.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#if 1
#define LOG_OUTPUT(s)
#else
#define LOG_OUTPUT(s)	Asura_OutputToDebugger_Retail(s)
#endif

// Debug variables...
DEBUGVAR float			dbg_fSmoothingMaxAngularVelocity			= Asura_Maths::Pi * 3.250f;	// 360 degrees/sec
DEBUGVAR float			dbg_fSmoothingMinAngularVelocity			= Asura_Maths::Pi * 0.250f;	// 90 degrees/sec

DEBUGVAR Asura_Angle	dbg_fSmoothingMinAngularVelocityThreshold	= Asura_Maths::Pi / 36;		// 5 degrees
DEBUGVAR Asura_Angle	dbg_fSmoothingMaxAngularVelocityThreshold	= Asura_Maths::Pi / 4;		// 45 degrees


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor(void)
: m_xAngularVelocityInterceptor(0.0f)
, m_xInitialNormal(Asura_Maths::ZeroAVector3)
, m_xTargetNormal(Asura_Maths::ZeroAVector3)
, m_xAxis(Asura_Maths::ZeroAVector3)
{
	// Nothing more to initialise...
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Reset
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Reset(Asura_AVector3_InParam xNormal)
{
	// Reset the velocity interceptor...
	m_xAngularVelocityInterceptor.Reset(0.0f);

	// Reset the slerp data...
	m_xInitialNormal	= xNormal;
	m_xTargetNormal		= xNormal;

	m_xAxis.Zero();

	// Call the parent function...
	PARENT::Reset(xNormal);
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update
// Author       : Andy Tate
// Date         : 6-Mar-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update(const float fTimeStep,
																			Asura_AVector3_InParam xDesiredNormal,
																			Asura_AVector3_InOutParam xCurrentNormal)
{
	// Are we effectively at our destination?
	if(xCurrentNormal.ApproximatelyEquals(xDesiredNormal))
	{
		// Yes - ensure that our output normal is correct AND that we've finished interpolation...
		xCurrentNormal	= xDesiredNormal;

		if(m_xAngularVelocityInterceptor.IsMoving())
		{
			m_xAngularVelocityInterceptor.Reset(0.0f);
		}
	}
	else
	{
		const float	fDotProduct				= xDesiredNormal * xCurrentNormal;
		Asura_Angle	fAngleBetweenNormals	= Asura_Maths::Abs(Asura_Maths::ArcCosine(fDotProduct));

		Asura_Angle	fAngleThisFrame;
		Asura_Quat	xRotation;

		// Do we need to use the velocity interpolator?
		if( (fAngleBetweenNormals >= dbg_fSmoothingMinAngularVelocityThreshold) ||
			!m_xTargetNormal.ApproximatelyEquals(xDesiredNormal) )
		{
			const float	fScale				= (fAngleBetweenNormals - dbg_fSmoothingMinAngularVelocityThreshold) / (dbg_fSmoothingMaxAngularVelocityThreshold - dbg_fSmoothingMinAngularVelocityThreshold);
			const float	fParametric			= Asura_Maths::ClampToRange(fScale, 0.0f, 1.0f);
			Asura_Angle	fAngularVelocity	= Asura_Maths::Lerp(dbg_fSmoothingMinAngularVelocity, dbg_fSmoothingMaxAngularVelocity, fParametric);
			const float	fDuration			= fAngleBetweenNormals / fAngularVelocity;

			// Initialise the normals and axis to use...
			fAngleThisFrame		= fAngularVelocity * fTimeStep;

			m_xInitialNormal	= xCurrentNormal;
			m_xTargetNormal		= xDesiredNormal;

			m_xAxis	= m_xInitialNormal ^ m_xTargetNormal;
			if ( m_xAxis.NormaliseReturningMagnitude() < 0.01f )
			{
				m_xAxis = m_xDesiredRotationAxis;
			}

			// Set-up the interpolator in case we're to use it next frame...
			m_xAngularVelocityInterceptor.SetTarget(0.0f, fAngularVelocity, fAngleBetweenNormals, fDuration);

			// Debugging...
			LOG_OUTPUT((">>> m_xInitialNormal = (%.3f, %.3f, %.3f)", m_xInitialNormal.GetX(), m_xInitialNormal.GetY(), m_xInitialNormal.GetZ()));
			LOG_OUTPUT((", m_xTargetNormal = (%.3f, %.3f, %.3f)", m_xTargetNormal.GetX(), m_xTargetNormal.GetY(), m_xTargetNormal.GetZ()));
			LOG_OUTPUT((", m_xAxis = (%.3f, %.3f, %.3f)", m_xAxis.GetX(), m_xAxis.GetY(), m_xAxis.GetZ()));
			LOG_OUTPUT(("\n"));
			LOG_OUTPUT(("    fAngleBetweenNormals = %.3f, fParametric = %.3f, fAngularVelocity = %.3f, fDuration = %.3f", fAngleBetweenNormals, fParametric, fAngularVelocity, fDuration));
			LOG_OUTPUT(("\n"));
		}
		else
		{
			// Continue with the existing interpolation...
			m_xAngularVelocityInterceptor.AddTime(fTimeStep);

			// Calculate the angle to use...
			m_xAngularVelocityInterceptor.CalculatePosition(fAngleThisFrame);
		}

		// Calculate the new normal...
		xRotation.FromRotation(fAngleThisFrame, m_xAxis);
		xCurrentNormal = xRotation.RotateVector(m_xInitialNormal);
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update(const float fTimeStep,
																			const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																			Asura_AVector3_InOutParam xCurrentNormal)
{
	Asura_AVector3	xDesiredNormal;

	// Fetch the desired normal to use...
	xNormalCalculation.GetDesiredNormal(xDesiredNormal);

	// Call the other version of this function...
	AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update(fTimeStep, xDesiredNormal, xCurrentNormal);
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AngularVelocityInterceptor::Update'. ///


/// End of file. ///
