
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_SurfaceReorientationNormalInterpolator_AxisRotation.cpp
// Author : Andy Tate
// Date   : 7-Feb-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header...
#include "Physics/AVP_SurfaceReorientationNormalInterpolator_AxisRotation.h"

// Asura independent header-files...
#include "Debugging/Asura_Logfile.h"

// Game independent header-files...
#include "Core/AVP_Plane.h"
#include "Physics/AVP_CalculateSurfaceReorientationNormal.h"


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

#if 1
#define LOG_OUTPUT(s)
#else
#define LOG_OUTPUT(s)	Asura_OutputToDebugger_Retail(s)
#endif


////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAngleAboutAxis
// Author       : Andy Tate
// Date         : 15-Oct-2007
////////////////////////////////////////////////////////////////////////////////

Asura_Angle AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAngleAboutAxis(Asura_AVector3_InParam xAxisOfRotation,
																							 Asura_AVector3_InParam xFromNormal,
																							 Asura_AVector3_InParam xToNormal) const
{
	const float	fCosAngleBetweenNormals	= xFromNormal * xToNormal;

	if(fCosAngleBetweenNormals > 0.99999f)	// cosine(0.25 degrees)
	{
		// Normals are the same - no angle between them...
		return(0.0f);
	}
	else
	{
		Asura_Quat		xDifferenceRotation;
		Asura_AVector3	xDifferenceAxis;
		Asura_Angle		fDifferenceAngle;
		float			fCosAngleBetweenAxes;

		// Calculate the angle between these 2 vectors to ensure that we get a signed angular-velocity...
		xDifferenceRotation.RotationFromAToB(xFromNormal, xToNormal);

		fDifferenceAngle		= xDifferenceRotation.ToRotation(xDifferenceAxis);
		fCosAngleBetweenAxes	= xDifferenceAxis * xAxisOfRotation;

		LOG_OUTPUT(("--> xAxisOfRotation = (%.3f, %.3f, %.3f), xFromNormal = (%.3f, %.3f, %.3f), xToNormal = (%.3f, %.3f, %.3f), fDifferenceAngle = %.3f, fCosAngleBetweenAxes = %.3f\n", xAxisOfRotation.GetX(), xAxisOfRotation.GetY(), xAxisOfRotation.GetZ(), xFromNormal.GetX(), xFromNormal.GetY(), xFromNormal.GetZ(), xToNormal.GetX(), xToNormal.GetY(), xToNormal.GetZ(), fDifferenceAngle, fCosAngleBetweenAxes));

		// Check whether we need to flip the angles because the axis has changed direction...
		if(fCosAngleBetweenAxes > 0.999f)
		{
			return(fDifferenceAngle);
		}
		else if(fCosAngleBetweenAxes < -0.999f)
		{
			return(-fDifferenceAngle);
		}
		else
		{
			Asura_Assertf(false, ("Our axes are not parallel when calculating the angle between 2 normals (fCosAngleBetweenNormals = %.3f).", fCosAngleBetweenNormals));
			return(fDifferenceAngle);
		}
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAngleAboutAxis'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAveragedNormalIgnoringClosest
// Author       : Andy Tate
// Date         : 30-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAveragedNormalIgnoringClosest(const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																									 Asura_AVector3_InParam xAxisOfRotation,
																									 Asura_AVector3_InParam xUpVector,
																									 Asura_AVector3_OutParam xAveragedNormal) const
{
	const float		fUpVectorAlongAxis	= xAxisOfRotation * xUpVector;
	const u_int		uNumNormals			= xNormalCalculation.GetNumNormalsUsed();

	Asura_AVector3	xUpVectorPerpendicularToAxis(xUpVector - (xAxisOfRotation * fUpVectorAlongAxis));

	float			fClosestDotProduct	= -1.0f;
	u_int			uClosestNormalIndex	= 0;

	// Initialise the data...
	xUpVectorPerpendicularToAxis.Normalise();
	xAveragedNormal.Zero();

	// Find the closest normal to our up-vector...
	for(u_int u = 0; u < uNumNormals; u++)
	{
		Asura_AVector3	xNormal;
		float			fDotProduct;

		xNormalCalculation.GetNormal(u, xNormal);
		fDotProduct	= xUpVectorPerpendicularToAxis * xNormal;

		if(fClosestDotProduct < fDotProduct)
		{
			fClosestDotProduct	= fDotProduct;
			uClosestNormalIndex	= u;
		}
	}

	// Calculate the average of the remaining normals...
	for(u_int u = 0; u < uNumNormals; u++)
	{
		if(u != uClosestNormalIndex)
		{
			Asura_AVector3	xNormal;
			Asura_AVector3	xNormalPerpendicularToAxis;
			float			fNormalAlongAxis;

			xNormalCalculation.GetNormal(u, xNormal);

			fNormalAlongAxis			= xAxisOfRotation * xNormal;
			xNormalPerpendicularToAxis	= xNormal - (xAxisOfRotation * fNormalAlongAxis);

			xNormalPerpendicularToAxis.Normalise();

			xAveragedNormal	+= xNormalPerpendicularToAxis;
		}
	}

	xAveragedNormal.Normalise();
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAveragedNormalIgnoringClosest'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAveragedNormalWereRotatingTowards
// Author       : Andy Tate
// Date         : 30-Oct-2007
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAveragedNormalWereRotatingTowards(const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																										 Asura_AVector3_InParam xAxisOfRotation,
																										 Asura_AVector3_InParam xUpVector,
																										 const float fRotationSign,
																										 Asura_AVector3_OutParam xAveragedNormal) const
{
	const float		fUpVectorAlongAxis	= xAxisOfRotation * xUpVector;
	const u_int		uNumNormals			= xNormalCalculation.GetNumNormalsUsed();

	Asura_AVector3	xUpVectorPerpendicularToAxis(xUpVector - (xAxisOfRotation * fUpVectorAlongAxis));

	// Initialise the data...
	xUpVectorPerpendicularToAxis.Normalise();
	xAveragedNormal.Zero();

	// Calculate the average of the normals we're rotating towards...
	for(u_int u = 0; u < uNumNormals; u++)
	{
		Asura_AVector3	xNormal;
		Asura_AVector3	xNormalPerpendicularToAxis;
		float			fNormalAlongAxis;
		Asura_Angle		fAngleToNormal;

		xNormalCalculation.GetNormal(u, xNormal);

		fNormalAlongAxis			= xAxisOfRotation * xNormal;
		xNormalPerpendicularToAxis	= xNormal - (xAxisOfRotation * fNormalAlongAxis);

		xNormalPerpendicularToAxis.Normalise();

		fAngleToNormal	 = CalculateAngleAboutAxis(xAxisOfRotation, xUpVectorPerpendicularToAxis, xNormalPerpendicularToAxis);
		fAngleToNormal	*= fRotationSign;

		if(fAngleToNormal > 0)
		{
			xAveragedNormal	+= xNormalPerpendicularToAxis;
		}
	}

	xAveragedNormal.Normalise();
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAveragedNormalWereRotatingTowards'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAxisOfRotation
// Author       : Andy Tate
// Date         : 5-Oct-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAxisOfRotation(Asura_AVector3_InParam xDesiredNormal,
																					  Asura_AVector3_InParam xCurrentNormal,
																					  Asura_AVector3_OutParam xAxisOfRotation) const
{
	const float	fDotProduct	= xCurrentNormal * xDesiredNormal;

	if(fDotProduct < 0.999f)	// cosine(2.5 degrees)
	{
		if(fDotProduct > -0.999f)	// cosine(177.5 degrees)
		{
			// Directions are sufficiently far apart to be able to calculate a cross-product...
			xAxisOfRotation	= xCurrentNormal ^ xDesiredNormal;
			xAxisOfRotation.Normalise();

#if 1
			const float	fX				= xAxisOfRotation.GetX();
			const float	fY				= xAxisOfRotation.GetY();
			const float	fZ				= xAxisOfRotation.GetZ();
			float		fMajorAxisValue	= fX;

			// Find the major-axis...
			if(Asura_Maths::Abs(fMajorAxisValue) < Asura_Maths::Abs(fY))
			{
				fMajorAxisValue	= fY;
			}

			if(Asura_Maths::Abs(fMajorAxisValue) < Asura_Maths::Abs(fZ))
			{
				fMajorAxisValue	= fZ;
			}

			// Always ensure that the major-axis always points in the same direction...
			if(fMajorAxisValue < 0.0f)
			{
				xAxisOfRotation	= -xAxisOfRotation;
			}
#endif
			return(true);
		}
		else
		{
			// Too far apart...
			return(false);
		}
	}
	else
	{
		// Too close together...
		return(false);
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAxisOfRotation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAxisOfRotation
// Author       : Andy Tate
// Date         : 5-Oct-2007
////////////////////////////////////////////////////////////////////////////////

bool AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAxisOfRotation(const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																					  Asura_AVector3_InParam xDesiredNormal,
																					  Asura_AVector3_InParam xCurrentNormal,
																					  Asura_AVector3_OutParam xAxisOfRotation) const
{
	// Can we calculate the axis of rotation from our gravity-direction?
	if(CalculateAxisOfRotation(xDesiredNormal, xCurrentNormal, xAxisOfRotation))
	{
		return(true);
	}
	else
	{
		const Asura_AVector3	xUpVector(xCurrentNormal);

		const u_int				uNumNormals				= xNormalCalculation.GetNumNormalsUsed();

		float					fFurthestDotProduct		= 1.0f;
		u_int					uFurthestNormalIndex	= 0;

		Asura_AVector3			xNormal;

		// Find the closest normal to our up-vector...
		for(u_int u = 0; u < uNumNormals; u++)
		{
			float			fDotProduct;
			
			xNormalCalculation.GetNormal(u, xNormal);
			fDotProduct	= xUpVector * xNormal;

			if(fFurthestDotProduct > fDotProduct)
			{
				fFurthestDotProduct		= fDotProduct;
				uFurthestNormalIndex	= u;
			}
		}

		// Try the furthest normal...
		xNormalCalculation.GetNormal(uFurthestNormalIndex, xNormal);
		return(CalculateAxisOfRotation(xNormal, xCurrentNormal, xAxisOfRotation));
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAxisOfRotation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateRotationSpeed
// Author       : Andy Tate
// Date         : 5-Oct-2007
////////////////////////////////////////////////////////////////////////////////

Asura_Angle AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateRotationSpeed(const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																							Asura_AVector3_InParam xAxisOfRotation,
																							Asura_AVector3_InParam xDesiredNormal,
																							Asura_AVector3_InParam xCurrentNormal) const
{
	Asura_AVector3	xVelocity;
	float			fSpeedSquared;

	// Use the speed from our previous data...
	xNormalCalculation.GetDesiredVelocity(xVelocity);
	fSpeedSquared	= xVelocity.MagnitudeSquared();

	if(fSpeedSquared == 0.0f)
	{
		// Stationary - no angular-velocity...
		return(0.0f);
	}
	else
	{
		const Asura_AVector3	xUpVector(xCurrentNormal);

		const float				fSpeedAlongAxis		= xAxisOfRotation * xVelocity;
		const float				fUpVectorAlongAxis	= xAxisOfRotation * xUpVector;

		const Asura_AVector3	xVelocityPerpendicularToAxis(xVelocity - (xAxisOfRotation * fSpeedAlongAxis));
		Asura_AVector3			xDirectionPerpendicularToAxis(xVelocityPerpendicularToAxis);

		Asura_AVector3			xUpVectorPerpendicularToAxis(xUpVector - (xAxisOfRotation * fUpVectorAlongAxis));

		const float				fSpeedPerpendicularToAxis	= xDirectionPerpendicularToAxis.NormaliseReturningMagnitude();
		Asura_Angle				fRotationSpeed				= fSpeedPerpendicularToAxis * m_fVelocityToAngularVelocityScale;

#if 0
		Asura_AVector3			xUpVectorPlusDirection(xDirectionPerpendicularToAxis);

		Asura_Angle				fDifferenceAngle;
		float					fDifferenceAngleSign;

		// Calculate the up-vector and the up-vector with velocity added...
		xUpVectorPerpendicularToAxis.Normalise();

		xUpVectorPlusDirection	+= xUpVectorPerpendicularToAxis;
		xUpVectorPlusDirection.Normalise();

		// Calculate the angle between these 2 vectors to ensure that we get a correctly signed rotation-speed...
		fDifferenceAngle	 = CalculateAngleAboutAxis(xAxisOfRotation, xUpVectorPerpendicularToAxis, xUpVectorPlusDirection);
		fDifferenceAngleSign = Asura_Maths::Sign(fDifferenceAngle);

		fRotationSpeed		*= fDifferenceAngleSign;
#endif

#if 1
		const float		fSizeAlongAxis	= xDesiredNormal * xAxisOfRotation;

		Asura_AVector3	xSurfaceNormalPerpendicularToAxis(xDesiredNormal - (xAxisOfRotation * fSizeAlongAxis));

		Asura_AVector3	xAveragedNormalPerpendicularToAxis;

		// Calculate the averaged-normal ignoring the normal closest to our up-vector...
		xSurfaceNormalPerpendicularToAxis.Normalise();

		CalculateAveragedNormalIgnoringClosest(xNormalCalculation, xAxisOfRotation, xUpVector, xAveragedNormalPerpendicularToAxis);

#if 0
		const Asura_Angle	fAverageNormalAngleDifference	= CalculateAngleAboutAxis(xAxisOfRotation, xUpVectorPerpendicularToAxis, xAveragedNormalPerpendicularToAxis);	//xSurfaceNormalPerpendicularToAxis);	//

		LOG_OUTPUT(("- xAveragedNormalPerpendicularToAxis = (%.3f, %.3f, %.3f), xUpVectorPerpendicularToAxis = (%.3f, %.3f, %.3f), fAverageNormalAngleDifference = %.3f\n", xAveragedNormalPerpendicularToAxis.GetX(), xAveragedNormalPerpendicularToAxis.GetY(), xAveragedNormalPerpendicularToAxis.GetZ(), xUpVectorPerpendicularToAxis.GetX(), xUpVectorPerpendicularToAxis.GetY(), xUpVectorPerpendicularToAxis.GetZ(), fAverageNormalAngleDifference));
		//LOG_OUTPUT(("- xSurfaceNormalPerpendicularToAxis = (%.3f, %.3f, %.3f), xUpVectorPerpendicularToAxis = (%.3f, %.3f, %.3f), fAverageNormalAngleDifference = %.3f\n", xSurfaceNormalPerpendicularToAxis.GetX(), xSurfaceNormalPerpendicularToAxis.GetY(), xSurfaceNormalPerpendicularToAxis.GetZ(), xUpVectorPerpendicularToAxis.GetX(), xUpVectorPerpendicularToAxis.GetY(), xUpVectorPerpendicularToAxis.GetZ(), fAverageNormalAngleDifference));

		////////////////////////////////////////////////////////////////////////////////
		// This code ensures we get the correct sign depending on whether the rotation
		// to the furthest normal is clockwise or anti-clockwise around the axis...
		////////////////////////////////////////////////////////////////////////////////
		if(fAverageNormalAngleDifference < 0.000f)
		{
			fRotationSpeed	= -fRotationSpeed;
		}
#endif

#if 1
		Asura_AVector3	xDirectionPerpendicularToAveragedNormal;
		Asura_Angle		fDirectionAngleDifference;
		float			fSizeAlongAveragedNormal;

		// Calculate the angle between our 2 direction-vectors...
		fSizeAlongAveragedNormal				= xDirectionPerpendicularToAxis * xAveragedNormalPerpendicularToAxis;
		xDirectionPerpendicularToAveragedNormal	= xDirectionPerpendicularToAxis - (xAveragedNormalPerpendicularToAxis * fSizeAlongAveragedNormal);

		xDirectionPerpendicularToAveragedNormal.Normalise();

		fDirectionAngleDifference	= CalculateAngleAboutAxis(xAxisOfRotation, xDirectionPerpendicularToAveragedNormal, xDirectionPerpendicularToAxis);

		LOG_OUTPUT(("- xDirectionPerpendicularToAxis = (%.3f, %.3f, %.3f), xDirectionPerpendicularToAveragedNormal = (%.3f, %.3f, %.3f), fDirectionAngleDifference = %.3f\n", xDirectionPerpendicularToAxis.GetX(), xDirectionPerpendicularToAxis.GetY(), xDirectionPerpendicularToAxis.GetZ(), xDirectionPerpendicularToAveragedNormal.GetX(), xDirectionPerpendicularToAveragedNormal.GetY(), xDirectionPerpendicularToAveragedNormal.GetZ(), fDirectionAngleDifference));

		////////////////////////////////////////////////////////////////////////////////
		// This code ensures that we get the correct sign depending on whether our
		// velocity is moving us clockwise or anti-clockwise around the axis...
		////////////////////////////////////////////////////////////////////////////////
		if(fDirectionAngleDifference > 0.000f)
		{
			fRotationSpeed	= -fRotationSpeed;
		}
#endif
#endif

#if 0
		Asura_Angle		fDirectionAngleDifference;
		float			fSizeAlongVector;

		Asura_AVector3	xDirectionPerpendicularToAveragedNormal;
		Asura_AVector3	xDirectionPerpendicularToUpVector;
		Asura_AVector3	xAveragedNormalPerpendicularToAxis;

		// Calculate the angle between our 2 direction-vectors...
		fSizeAlongVector					= xDirectionPerpendicularToAxis * xUpVectorPerpendicularToAxis;
		xDirectionPerpendicularToUpVector	= xDirectionPerpendicularToAxis - (xUpVectorPerpendicularToAxis * fSizeAlongVector);

		xDirectionPerpendicularToUpVector.Normalise();

		fDirectionAngleDifference	= CalculateAngleAboutAxis(xAxisOfRotation, xDirectionPerpendicularToUpVector, xDirectionPerpendicularToAxis);

		LOG_OUTPUT(("- xDirectionPerpendicularToAxis = (%.3f, %.3f, %.3f), xDirectionPerpendicularToUpVector = (%.3f, %.3f, %.3f), fDirectionAngleDifference = %.3f\n", xDirectionPerpendicularToAxis.GetX(), xDirectionPerpendicularToAxis.GetY(), xDirectionPerpendicularToAxis.GetZ(), xDirectionPerpendicularToUpVector.GetX(), xDirectionPerpendicularToUpVector.GetY(), xDirectionPerpendicularToUpVector.GetZ(), fDirectionAngleDifference));

		// Calculate the average of the normals were rotating towards...
		CalculateAveragedNormalWereRotatingTowards(xAxisOfRotation, xUpVectorPerpendicularToAxis, Asura_Maths::Sign(fDirectionAngleDifference), axNormals, uNumNormals, xAveragedNormalPerpendicularToAxis);

		// Calculate the angle between our 2 direction-vectors...
		fSizeAlongVector						= xDirectionPerpendicularToAxis * xAveragedNormalPerpendicularToAxis;
		xDirectionPerpendicularToAveragedNormal	= xDirectionPerpendicularToAxis - (xAveragedNormalPerpendicularToAxis * fSizeAlongVector);

		xDirectionPerpendicularToAveragedNormal.Normalise();

		fDirectionAngleDifference	= CalculateAngleAboutAxis(xAxisOfRotation, xDirectionPerpendicularToAveragedNormal, xDirectionPerpendicularToAxis);

		LOG_OUTPUT(("- xDirectionPerpendicularToAxis = (%.3f, %.3f, %.3f), xDirectionPerpendicularToAveragedNormal = (%.3f, %.3f, %.3f), fDirectionAngleDifference = %.3f\n", xDirectionPerpendicularToAxis.GetX(), xDirectionPerpendicularToAxis.GetY(), xDirectionPerpendicularToAxis.GetZ(), xDirectionPerpendicularToAveragedNormal.GetX(), xDirectionPerpendicularToAveragedNormal.GetY(), xDirectionPerpendicularToAveragedNormal.GetZ(), fDirectionAngleDifference));

		////////////////////////////////////////////////////////////////////////////////
		// This code ensures that we get the correct sign depending on whether our
		// velocity is moving us clockwise or anti-clockwise around the axis...
		////////////////////////////////////////////////////////////////////////////////
		if(fDirectionAngleDifference > 0.000f)
		{
			fRotationSpeed	= -fRotationSpeed;
		}
#endif

#if 0
		const Asura_AVector3	xForwardsVectorFromCrossProduct(xAxisOfRotation ^ xUpVectorPerpendicularToAxis);

		const float				fDotProduct	= xForwardsVectorFromCrossProduct * xDirectionPerpendicularToAxis;

		LOG_OUTPUT(("- xForwardsVectorFromCrossProduct = (%.3f, %.3f, %.3f), xUpVectorPerpendicularToAxis = (%.3f, %.3f, %.3f), fDotProduct = %.3f\n", xForwardsVectorFromCrossProduct.GetX(), xForwardsVectorFromCrossProduct.GetY(), xForwardsVectorFromCrossProduct.GetZ(), xUpVectorPerpendicularToAxis.GetX(), xUpVectorPerpendicularToAxis.GetY(), xUpVectorPerpendicularToAxis.GetZ(), fDotProduct));

		////////////////////////////////////////////////////////////////////////////////
		// This code ensures that we get the correct sign depending on whether our
		// velocity is moving us clockwise or anti-clockwise around the axis...
		////////////////////////////////////////////////////////////////////////////////
		if(fDotProduct < 0.000f)
		{
			fRotationSpeed	= -fRotationSpeed;
		}
#endif

		return(fRotationSpeed);
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateRotationSpeed'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAngleToRotate
// Author       : Andy Tate
// Date         : 10-Oct-2007
////////////////////////////////////////////////////////////////////////////////

Asura_Angle AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAngleToRotate(const float fTimeStep,
																							const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																							Asura_AVector3_InParam xAxisOfRotation,
																							const Asura_Angle fRotationSpeed,
																							Asura_AVector3_InParam xCurrentNormal) const
{
	const Asura_Angle	fDesiredRotation	= fRotationSpeed * fTimeStep;

	// Do we need to bother rotating?
	if(fDesiredRotation == 0.0f)
	{
		return(0.0f);
	}
	else
	{
		const Asura_AVector3	xUpVector(xCurrentNormal);

		const u_int				uNumNormals			= xNormalCalculation.GetNumNormalsUsed();
		const float				fUpVectorAlongAxis	= xUpVector * xAxisOfRotation;

		Asura_Angle				fMaxRotation		= 0.0f;
		Asura_Angle				fMinRotation		= 0.0f;

		Asura_AVector3			xUpVectorPerpendicularToAxis(xUpVector - (xAxisOfRotation * fUpVectorAlongAxis));

		Asura_Angle				fRemainingRotation;

		// Ensure that our perpendicular-direction is normalised...
		xUpVectorPerpendicularToAxis.Normalise();

		// Calculate the rotation limit...
		for(u_int u = 0; u < uNumNormals; u++)
		{
			Asura_AVector3	xNormal;
			Asura_AVector3	xNormalPerpendicularToAxis;
			float			fNormalAlongAxis;
			float			fDifferenceAngle;

			// Ensure that our perpendicular-direction is normalised...
			xNormalCalculation.GetNormal(u, xNormal);

			fNormalAlongAxis			= xNormal * xAxisOfRotation;
			xNormalPerpendicularToAxis	= xNormal - (xAxisOfRotation * fNormalAlongAxis);

			xNormalPerpendicularToAxis.Normalise();

			// Calculate the angle between our 2 vectors and expand our limits accordingly...
			fDifferenceAngle	= CalculateAngleAboutAxis(xAxisOfRotation, xUpVectorPerpendicularToAxis, xNormalPerpendicularToAxis);

			fMaxRotation		= Asura_Maths::Max(fDifferenceAngle, fMaxRotation);
			fMinRotation		= Asura_Maths::Min(fDifferenceAngle, fMinRotation);
		}

		// Ensure that our remaining rotation is clamped at the limit...
		fRemainingRotation	= Asura_Maths::ClampToRange(fDesiredRotation, fMinRotation, fMaxRotation);

		if(true)	//m_fAngleToRotateScale == 0.0f)
		{
			LOG_OUTPUT(("- xAxisOfRotation = (%.3f, %.3f, %.3f), fRotationSpeed = %.3f, fMaxRotation = %.3f, fMinRotation = %.3f, fDesiredRotation = %.3f, fRemainingRotation = %.3f\n", xAxisOfRotation.GetX(), xAxisOfRotation.GetY(), xAxisOfRotation.GetZ(), fRotationSpeed, fMaxRotation, fMinRotation, fDesiredRotation, fRemainingRotation));
			return(fRemainingRotation);
		}
		else
		{
			const float	fScale				= (fDesiredRotation > 0.0f) ? (fMaxRotation / fDesiredRotation) : (fMinRotation / fDesiredRotation);
			const float	fParametricScale	= (fScale < m_fAngleToRotateScale) ? (fScale / m_fAngleToRotateScale) : 1.0f;
			const float	fRotationMultiplier	= Asura_Maths::Lerp(0.010f, 1.000f, (fParametricScale * fParametricScale));

			fRemainingRotation	*= fRotationMultiplier;

			LOG_OUTPUT(("- fRotationSpeed = %.3f, fMaxRotation = %.3f, fMinRotation = %.3f, fDesiredRotation = %.3f, fRemainingRotation = %.3f\n", fRotationSpeed, fMaxRotation, fMinRotation, fDesiredRotation, fRemainingRotation));
			return(fRemainingRotation);
		}
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::CalculateAngleToRotate'. ///



////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::AVP_SurfaceReorientationNormalInterpolator_AxisRotation
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

AVP_SurfaceReorientationNormalInterpolator_AxisRotation::AVP_SurfaceReorientationNormalInterpolator_AxisRotation(void)
: m_xCurrentAxisOfRotation(1.0f, 0.0f, 0.0f)
, m_fCurrentRotationSpeed(0.0f)
, m_fVelocityToAngularVelocityScale(1.5f)
, m_fAngleToRotateScale(5.0f)
{
	// Nothing more to initialise...
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::AVP_SurfaceReorientationNormalInterpolator_AxisRotation'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Reset
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Reset(Asura_AVector3_InParam xNormal)
{
	m_xCurrentAxisOfRotation	= Asura_AVector3(1.0f, 0.0f, 0.0f);
	m_fCurrentRotationSpeed		= 0.0f;
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Reset'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Update
// Author       : Andy Tate
// Date         : 6-Mar-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Update(const float fTimeStep,
																	 Asura_AVector3_InParam xDesiredNormal,
																	 Asura_AVector3_InOutParam xCurrentNormal)
{
	Asura_Assert(false, "This form of axis-rotation interpolator has not yet been implemented.");
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Update'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Update
// Author       : Andy Tate
// Date         : 7-Feb-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Update(const float fTimeStep,
																	 const AVP_CalculateSurfaceReorientationNormal& xNormalCalculation,
																	 Asura_AVector3_InOutParam xCurrentNormal)
{
	Asura_AVector3	xDesiredNormal;

	// Get the normal we're going to use...
	xNormalCalculation.GetDesiredNormal(xDesiredNormal);

	// Calculate the axis to rotate across...
	if(CalculateAxisOfRotation(xDesiredNormal, xCurrentNormal, m_xCurrentAxisOfRotation))
	{
		const u_int	uNumNormals	= xNormalCalculation.GetNumNormalsUsed();

		Asura_Angle	fAngleToRotate;

		LOG_OUTPUT(("Frame %d, uNumNormals = %d, m_xCurrentGravityDirection = (%.3f, %.3f, %.3f), xNewGravityDirection = (%.3f, %.3f, %.3f), xOutputVelocity = (%.3f, %.3f, %.3f)\n", Asura_Timers::GetFrameNumber(), uNumNormals, m_xCurrentGravityDirection.GetX(), m_xCurrentGravityDirection.GetY(), m_xCurrentGravityDirection.GetZ(), xNewGravityDirection.GetX(), xNewGravityDirection.GetY(), xNewGravityDirection.GetZ(), xOutputVelocity.GetX(), xOutputVelocity.GetY(), xOutputVelocity.GetZ()));

		// Calculate how much to rotate this frame...
		if(uNumNormals == 1)
		{
			fAngleToRotate	= m_fCurrentRotationSpeed * fTimeStep;
		}
		else
		{
			const float	fRotationSpeed	= CalculateRotationSpeed(xNormalCalculation, m_xCurrentAxisOfRotation, xDesiredNormal, xCurrentNormal);

			fAngleToRotate			= CalculateAngleToRotate(fTimeStep, xNormalCalculation, m_xCurrentAxisOfRotation, fRotationSpeed, xCurrentNormal);
			m_fCurrentRotationSpeed	= fAngleToRotate / fTimeStep;
		}

		// Are we actually rotating?
		if(fAngleToRotate != 0.0f)
		{
			Asura_Quat	xRotation;

			xRotation.FromRotation(fAngleToRotate, m_xCurrentAxisOfRotation);
			xDesiredNormal = xRotation.RotateVector(xCurrentNormal);
		}
	}
	else
	{
		// No axis means we can't rotate...
		m_fCurrentRotationSpeed	= 0.0f;
	}
}

/// End of function 'AVP_SurfaceReorientationNormalInterpolator_AxisRotation::Update'. ///


/// End of file. ///
