///////////////////////////////////////////////////////////////////////////////
//
// File: AVP_CollisionFX_System.cpp
//
// Created by Martin G.Harris (25/09/2007)
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Play SFX/PFX according to collision information from the physics system.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "SpecialFX/AVP_CollisionFX_System.h"

// Only compile this code if the symbol is defined.
#ifdef ASURA_USE_COLLISIONFX

// Includes
// Asura
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Entity/Asura_Entity.h"
#include "Entity/Container/Asura_Container.h"
#include "Physics/Asura_Physics_CollisionFX_Report.h"
#include "Physics/Asura_Physics_Environment.h"
#include "SpecialFX/Asura_CFX_Effect.h"

// AVP
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Core/AVP_Core.h"
#include "Core/AVP_SurfaceEffects.h"
#include "Entity/Container/AVP_ServerEntity_Container.h"
#include "Entity/PhysicalObject/AVP_ClientEntity_PhysicalObject.h"
#include "Entity/PhysicalObject/AVP_ServerEntity_PhysicalObject.h"

///////////////////////////////////////////////////////////////////////////////
// AVP_CollisionFX_System
///////////////////////////////////////////////////////////////////////////////

// Blueprint parameter
const Asura_Hash_ID uAVP_BLUEPRINT_HASH_MATERIAL_TYPE = 0x81F508E1; // Asura_GetHashID("MaterialType")
DEBUGVAR float dbg_fBodyFallThreshold =  fAVP_COLLISION_SFX_SPEED_THRESHOLD+1.0f; // all collisions less than fAVP_COLLISION_SFX_SPEED_THRESHOLD are culled in asura code

//////////////////////////////////
// GetSurfaceTypeHashFromEntity //
//////////////////////////////////
// Get the surface type for an entity.
Asura_Hash_ID AVP_CollisionFX_System::GetSurfaceTypeHashFromEntity(Asura_Guid uGuid)
{	
	// Physical objects...
	// Find the entity. Check the server entities first.
	const AVP_ServerEntity_PhysicalObject* pxServerPhysObj = AVP_ServerEntity_PhysicalObject::FindEntity(uGuid);
	if (pxServerPhysObj)
	{
		return pxServerPhysObj->GetMaterialType();
	}
	else
	{
		// Could be a client-side effect, check the client entities.
		const AVP_ClientEntity_PhysicalObject* pxClientPhysObj = AVP_ClientEntity_PhysicalObject::FindEntity(uGuid);
		if (pxClientPhysObj)
		{
			return pxClientPhysObj->GetMaterialType();
		}
	}

	Asura_Hash_ID uBlueprintHashID     = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uBlueprintTypeHashID = ASURA_HASH_ID_UNSET;

	// Containers...
	// Find the blueprint. Check both the server and client entities.
	// Note: Assume that there is a single blueprint.
	const Asura_Container_Blueprint* pxBlueprintContainer = AVP_ServerEntity_Container::FindBlueprintFromGuid(uGuid, ASURA_HASH_ID_UNSET);
	if (pxBlueprintContainer)
	{
		uBlueprintHashID	 = pxBlueprintContainer->GetBlueprintHash();
		uBlueprintTypeHashID = pxBlueprintContainer->GetBlueprintTypeHash();
	}
	else
	{
		//Asura_Assertf(0, ("Blueprint not found for entity %u", uGuid));
		return ASURA_HASH_ID_UNSET;
	}

	// For the "SurfaceEffects" blueprint type, the material type is just the hash of the blueprint.
	if (uBlueprintTypeHashID == uAVP_BLUEPRINT_TYPE_SURFACEEFFECTS)
	{
		return uBlueprintHashID;
	}

	// Look up the "MaterialType" parameter.
	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uBlueprintTypeHashID, uBlueprintHashID );
	if (pxBlueprint)
	{
		Asura_Value xValue;
		const bool bReturn = pxBlueprint->GetParameterValueByHashID( uAVP_BLUEPRINT_HASH_MATERIAL_TYPE, xValue );
		if (bReturn)
		{
			return xValue.GetHashIDValue();
		}

		Asura_Assertf(0, ("Failed to find 'MaterialType' parameter in blueprint (type: 0x%08x, hash: 0x%08x)", uBlueprintTypeHashID, uBlueprintHashID));
	}
	else
	{
		Asura_Assertf(0, ("Failed to find blueprint (type: 0x%08x, hash: 0x%08x) for entity %u", uBlueprintTypeHashID, uBlueprintHashID, uGuid));
	}

	return ASURA_HASH_ID_UNSET;
}

///////////////////////////////////
// GetSurfaceTypeHashFromPolygon //
///////////////////////////////////
// Get the surface type for an environmental polygon.
Asura_Hash_ID AVP_CollisionFX_System::GetSurfaceTypeHashFromPolygon(u_int uFaceID)
{
	const Asura_Physics_Zone& xZone = Asura_Physics_Environment::GetCurrentZone();
	const Asura_Hash_ID uMaterialResponseHashID = xZone.GetFaceMaterialResponse(uFaceID);
	
	const Asura_MaterialResponse* pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID(uMaterialResponseHashID);
		
	if (pxMaterial)
	{	
		return pxMaterial->GetSurfaceTypeHash();		
	}
	
	return ASURA_HASH_ID_UNSET;
}

//////////////////
// CreateEffect //
//////////////////
Asura_CFX_Effect* AVP_CollisionFX_System::CreateEffect(const Asura_Physics_CollisionFX_Report& xCollInfo, int iType)
{
	bool bActorBodyFall = false;
	Asura_Hash_ID uSurfaceHash1 = ASURA_HASH_ID_UNSET;

	// Look up surface types for entities/polygons....

	AVP_ClientEntity_PhysicalObject* pxClientObject = AVP_ClientEntity_PhysicalObject::FindEntity(xCollInfo.GetObject1EntityGuid());

	if (pxClientObject && pxClientObject->IsADeadBody())
	{
		// We don't have a material type, but don't assert yet, if its a ragdoll we use the effect types.
		Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID( uAVP_BLUEPRINT_TYPE_ACTORS, pxClientObject->GetBlueprintHashID() );
		if ( pxBlueprint )
		{
			Asura_Value xValue;
			Asura_Hash_ID uParamHash = (xCollInfo.GetRelVelocity().Magnitude() > dbg_fBodyFallThreshold)?0x99C639F2:0x9A007E41; // SurfaceType_BodyfallHeavy : SurfaceType_BodyfallLight
			if (pxBlueprint->GetParameterValueByHashID(uParamHash, xValue))
			{
				uSurfaceHash1 = xValue.GetHashIDValue();
				bActorBodyFall = true;
			}
		}
	}
	else
	{
		uSurfaceHash1 = GetSurfaceTypeHashFromEntity(xCollInfo.GetObject1EntityGuid());
	}

	if(uSurfaceHash1 == ASURA_HASH_ID_UNSET)
	{
		// Now assert, we dont have a material or an effect type
		Asura_Assertf(0, ("Surface type not set for entity, or Effect type not set up for ragdoll with guid: %u.", xCollInfo.GetObject1EntityGuid()));
		return 0; // No effect
	}

	const Asura_Hash_ID uSurfaceHash2 = ( xCollInfo.IsEnvironmentCollision() ?
										  GetSurfaceTypeHashFromPolygon( xCollInfo.GetEnvironmentFaceID() ) :
										  GetSurfaceTypeHashFromEntity ( xCollInfo.GetObject2EntityGuid() ) );

	if (uSurfaceHash2 == ASURA_HASH_ID_UNSET)
	{
#ifdef ASURA_ASSERT
		if (xCollInfo.IsEnvironmentCollision())
		{
			Asura_Assertf(0, ("Surface type not set for environment polygon (%u).", xCollInfo.GetEnvironmentFaceID()));
		}
		else
		{
			Asura_Assertf(0, ("Surface type not set for entity %u.", xCollInfo.GetObject2EntityGuid()));
		}
#endif //ASURA_ASSERT

		return 0; // No effect
	}


	if(bActorBodyFall)
	{
		AVP_SurfaceEffect_System::PlayEffectTypeSurfaceEffect( uSurfaceHash1, uSurfaceHash2, Asura_Vector_3(xCollInfo.GetPosition()), Asura_Vector_3(xCollInfo.GetNormal()));
	}
	else
	{
		AVP_SurfaceEffect_System::PlayCollisionSurfaceEffect( uSurfaceHash1, uSurfaceHash2, Asura_Vector_3(xCollInfo.GetPosition()),Asura_Vector_3( xCollInfo.GetNormal()));
	}

	return 0; // No effect
}

#endif //ASURA_USE_COLLISIONFX
