///////////////////////////////////////////////////////////////////////////////
//
// File: AVP_Platform_CloakEffect.cpp
//
// Created by Andrew Southgate 03/122007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Handles rendering of the predators cloak
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "SpecialFX/AVP_CloakEffect.h"
#include "SpecialFX/AVP_Platform_CloakEffect.h"

// Asura Platform-independent
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_MaterialResponse_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Assert.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_SortedRender.h"
#include "Rendering/Asura_Viewport.h"
#include "Interface/Asura_CommandConsole.h"

// PS3
#include "Rendering/Asura_PS3_CharacterRenderer.h"
#include "Rendering/Asura_PS3_FrameBuffer.h"
#include "Rendering/Asura_PS3_Graphics.h"
#include "Rendering/Shaders/Asura_PS3_GCM_ShadersLibrary.h"
#include "Rendering/Shaders/Asura_PS3_GCM_Shader.h"
#include "Rendering/Asura_PS3_MaterialResponse_System.h"
#include "Rendering/Asura_PS3_RenderingStates.h"
#include "Rendering/Asura_PS3_RSXProfiling.h"
#include "Rendering/Asura_PS3_SPU_Skinning.h"

// AVP PS3
#include "Rendering/AVP_Platform_Stencil_Setup.h"

//Shaders definition
#include "Rendering/ShadersDefinition/AVP_PS3_ProjectShaders.h"
#include "Rendering/ShadersDefinition/VD/CloakEffect_VD.h"
#include "Rendering/ShadersDefinition/VP/CloakEffect_VP.h"
#include "Rendering/ShadersDefinition/FP/CloakEffect_FP.h"
#include "Rendering/ShadersDefinition/FP/CloakHoles_FP.h"

// Partial vertex definitions, for using multiple buffers

// Position, output from SPU skinning
static Asura_PS3_VertexBufferDeclaration g_axWorldPos_VD[] = //this definition must be compatible with the shadows one (since they'll read the same buffer)
{
	//	m_szName, m_uOffset, m_uType, m_uNumberOfCoords, m_ucAttributeIndex
	{ "xIn.m_xPosition",		0,	CELL_GCM_VERTEX_F,		3, 0 }, // skinned position
};
#define NUM_WORLDPOS_VD_ELEMENTS	1

// Un-skinned values, from the input vertex buffer.
static Asura_PS3_VertexBufferDeclaration g_axObjectSpacePosFromVertex_VD[] =
{
	//	m_szName, m_uOffset, m_uType, m_uNumberOfCoords, m_ucAttributeIndex
//	{ "xIn.m_xPosition",		 0,	CELL_GCM_VERTEX_F,		3, 0 },
//	{ "xIn.m_xUV",				12,	CELL_GCM_VERTEX_SF,		4, 0 },
	{ "xIn.m_aiBoneIndices",	20,	CELL_GCM_VERTEX_UB256,	4, 0 },
	{ "xIn.m_afBoneWeights",	24,	CELL_GCM_VERTEX_UB,		4, 0 },
	{ "xIn.m_xNormal",			28,	CELL_GCM_VERTEX_CMP,	1, 0 },
//	{ "xIn.m_xTangent",			32,	CELL_GCM_VERTEX_CMP,	1, 0 },
//	{ "xIn.m_xBinormal",		36,	CELL_GCM_VERTEX_CMP,	1, 0 },
};
#define NUM_OBJPOSFROMVERTEX_VD_ELEMENTS	3


//
// Variables
//

static Asura_PS3_GCM_VertexShader*		g_pxVertexShader			= NULL;
static Asura_PS3_GCM_VertexShader*		g_pxStraightThroughVS		= NULL;
static Asura_PS3_FragmentShaderCopy*	g_pxFragmentShader			= NULL;

#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
static bool g_bShowHoles			= false;
static bool g_bEnableHoles			= true;
static bool g_bEnableUnholes		= true;
static bool g_bEnableDistortion		= true;
#else
static const bool g_bShowHoles				= false;
static const bool g_bEnableHoles			= true;
static const bool g_bEnableUnholes			= true;
static const bool g_bEnableDistortion		= true;
#endif

static float g_fCloakDepthFactor	= 0.0f;
static float g_fCloakDepthUnits		= -3.0f;

int AVP_Platform_CloakEffect::s_iNoiseTexture;

static float s_afVariations[AVP_CloakEffect::NUM_CLOAK_EFFECT_PERMUTATIONS][4] =
{
	{ 0.0f, 0.0f, 0.0f, 0.0f},
	{ 0.5f, 0.0f, 0.0f, 0.0f},
	{ 0.0f, 0.5f, 0.0f, 0.0f},
	{ 0.5f, 0.5f, 0.0f, 0.0f},
	{ 0.0f, 0.0f, 0.5f, 0.0f},
	{ 0.5f, 0.0f, 0.5f, 0.0f},
	{ 0.0f, 0.5f, 0.5f, 0.0f},
	{ 0.5f, 0.5f, 0.5f, 0.0f},
};




//
// Functions
//
void AVP_CloakEffect::Platform_Initialise()
{
	AVP_Platform_CloakEffect::Initialise();
}
void AVP_CloakEffect::Platform_Shutdown()
{
	AVP_Platform_CloakEffect::Shutdown();
}


void AVP_Platform_CloakEffect::Initialise()
{
	g_pxStraightThroughVS = new AVP_PS3_VP_CloakEffect_StraightThrough();

	g_pxStraightThroughVS->InitVertexBufferDeclaration( g_axWorldPos_VD, NUM_WORLDPOS_VD_ELEMENTS );
 	g_pxStraightThroughVS->InitVertexBufferDeclaration( g_axObjectSpacePosFromVertex_VD, NUM_OBJPOSFROMVERTEX_VD_ELEMENTS);


	g_pxVertexShader	= new AVP_PS3_VP_CloakEffect();

	g_pxVertexShader->InitVertexBufferDeclaration( g_axVertexDeclaration, VD_ELEMENTS );
	Asura_PS3_GCM_ShadersLibrary::AddFP( ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKEFFECT_SHADER, (char*)&_binary_PS3_CloakEffect_FP_bin_start, (char*)&_binary_PS3_CloakEffect_FP_bin_end - (char*)&_binary_PS3_CloakEffect_FP_bin_start, new Asura_PS3_CloakEffect_Handles, 4, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKEFFECT_SIMPLE_SHADER, (char*)&_binary_PS3_CloakEffect_Simple_FP_bin_start, (char*)&_binary_PS3_CloakEffect_Simple_FP_bin_end - (char*)&_binary_PS3_CloakEffect_Simple_FP_bin_start, new Asura_PS3_CloakEffect_Simple_Handles, 4, true );
	Asura_PS3_GCM_ShadersLibrary::AddFP( ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKHOLES_SHADER, (char*)&_binary_PS3_CloakHoles_FP_bin_start, (char*)&_binary_PS3_CloakHoles_FP_bin_end - (char*)&_binary_PS3_CloakHoles_FP_bin_start, new Asura_PS3_CloakHoles_Handles, 4, true );

#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
	Asura_CommandConsole::AddCmd(L"Render.ReloadShaders.CloakEffect", ReloadShaders, Asura_CommandConsole::REMOVE_AT_RESET);
	Asura_CommandConsole::AddVar(L"AVP.CloakEffect.Shaders.DebugHoles", &g_bShowHoles, Asura_CommandConsole::REMOVE_AT_RESET);
	Asura_CommandConsole::AddVar(L"AVP.CloakEffect.Shaders.Holes", &g_bEnableHoles, Asura_CommandConsole::REMOVE_AT_RESET);
	Asura_CommandConsole::AddVar(L"AVP.CloakEffect.Shaders.Opaque", &g_bEnableUnholes, Asura_CommandConsole::REMOVE_AT_RESET);
	Asura_CommandConsole::AddVar(L"AVP.CloakEffect.Shaders.Distort", &g_bEnableDistortion, Asura_CommandConsole::REMOVE_AT_RESET);
#endif

#if defined(ASURA_DEBUG_VARIABLE_SYSTEM)
	Asura_CommandConsole::AddVar(L"AVP.CloakEffect.DepthBias.Factor", &g_fCloakDepthFactor, -FLT_MAX, FLT_MAX, Asura_CommandConsole::REMOVE_AT_RESET);
	Asura_CommandConsole::AddVar(L"AVP.CloakEffect.DepthBias.Units", &g_fCloakDepthUnits, -FLT_MAX, FLT_MAX, Asura_CommandConsole::REMOVE_AT_RESET);
#endif

	s_iNoiseTexture = Asura_Texture_Management::AddTexture("\\SpecialFX\\AVP_PredCloak.dds");
	Asura_Texture_Management::LoadTextures();
}

void AVP_Platform_CloakEffect::Shutdown()
{
	delete g_pxVertexShader;

	g_pxVertexShader	= NULL;
}

void AVP_Platform_CloakEffect::Reset()
{

}


Asura_PS3_GCM_IndexBuffer* AVP_Platform_CloakEffect::SetupSkinning(	Asura_PS3_CharacterRenderer_Skin* pxCharacterStrips,
																	const Asura_Animation& xAnimation,
																	Asura_AMatrix4x4_InParam xTransform,
																	Asura_AMatrix4x4_InParam xInverseTransform,
																	const Asura_AMatrix4x4* const pxBoneTransforms,
																	u_int uNumBones,
																	u_int uSkinIdx,
																	float fTransition,
																	float fMaxVisibilityAngle,
																	float fMaxVisibilityDistance)
{
	// Decide if we're going to use the vertex shader or SPU for rendering
	bool bSpuVS = Asura_PS3_SPU_Skinning::IsSpuable(pxCharacterStrips);


	// fill vertex program constants
	if ( bSpuVS )
	{
		Asura_PS3_CharacterRenderer::UseSpuSkinning( pxCharacterStrips, xAnimation, SPU_SKINNING_POS_AND_CULL,
														xTransform, pxBoneTransforms, uNumBones, uSkinIdx );

		g_pxStraightThroughVS->Bind();
	}
	else
	{
		//we can use the cloaked vertex shader, it has everything we need
		g_pxVertexShader->Bind();
	}

	// fill the vertex shader constants (shared by SPU and non-SPU skinning shaders)
	const Asura_AMatrix4x4& xWorldViewProjection = Asura_PS3_Graphics::GetWorldViewProjectionMatrix( &Asura_Hierarchy_System::s_xBaseTransform );
	g_pxVertexShader->SetUniformParameter( SK_WorldViewProj, xWorldViewProjection.GetElements() );

	const Asura_AMatrix4x4 xWorldView = xTransform * Asura_PS3_Graphics::GetViewMatrix();
	g_pxVertexShader->SetUniformParameter( SK_WorldView, xWorldView.GetElements() );

	Asura_AVector4 xObjCamPos =  xInverseTransform.Apply(Asura_AVector4(*Asura_Camera::GetPosition(), 1.0f));

	float afCamPos[ 4 ]				= { xObjCamPos.GetX(), xObjCamPos.GetY(), xObjCamPos.GetZ(), 1.0f };
	g_pxVertexShader->SetUniformParameter( SK_ObjCamPos, afCamPos);


	// Set bone matrices
	{
		float axData[ 3*4*ASURA_MAX_NO_OF_BONES_PER_MATERIAL ];

		for( u_int i = 0; i < uNumBones; ++i )
		{
			// store transposed, so only need 3 float4's not 4 in shader
			pxBoneTransforms[i].StoreRowsUnaligned(&axData[i*3*4]); // this stores 4x4 to our 3x4 space, hopefully it won't break anything

			// set element 2,2
			axData[i*3*4+4*2+2] = fTransition; // pxCloakEffect->GetBoneValue(i);
			//Asura_Assertf(pxCloakEffect->GetBoneValue(i)==0,("non-zero bone offset [%d] = %.3f, I was wrong!", i, pxCloakEffect->GetBoneValue(i))); //BW
		}

		//g_pxVertexShader->SetUniformParameter( SK_BoneMatrices, axBoneMatrix[0].GetElements(), uNumBones );
		//avoid the huge copy by just setting it directly
		cellGcmSetVertexProgramParameterBlockUnsafeInline( Asura_PS3_Graphics::GetGcmCurrentContext(), g_pxVertexShader->GetRegisterIndex(SK_BoneTransforms), uNumBones*3, axData );
		
		//invalidate the cached vertex parameters, since we just set them directly
		g_pxVertexShader->InvalidateRegisterCache();

	}

	float fCosAng = Asura_Maths::Cosine(fMaxVisibilityAngle);
	float afCloakFadeOut[4] = 
	{
		1.0f / fMaxVisibilityDistance, fCosAng, fCosAng < 1.0f ? 1.0f / (1.0f - fCosAng) : 0.0f, 0.0f
	};
	g_pxVertexShader->SetUniformParameter( SK_CloakFadeOut, afCloakFadeOut);


	Asura_PS3_GCM_IndexBuffer *pIndexBuffer = NULL;

	if ( bSpuVS )
	{
		Asura_Platform_Animation_RenderCache* pxRenderCache	= xAnimation.GetRenderCache();

		Asura_PS3_GCM_VertexBuffer *pProcessedPositions = pxRenderCache->ObtainVertexBuffer( ASURA_PS3_RENDERCACHE_BUFFER_WORLDPOS, uSkinIdx );

		//use the verts processed by SPU
		//bind SPU-skinned things
		pProcessedPositions->Bind( g_axWorldPos_VD, NUM_WORLDPOS_VD_ELEMENTS, 1, true ); //true => clear other bindings

		//bind non-SPU-skinned things
		pxCharacterStrips->GetVertexBuffer()->Bind(g_axObjectSpacePosFromVertex_VD, NUM_OBJPOSFROMVERTEX_VD_ELEMENTS, 1, false);

		pIndexBuffer = pxRenderCache->ObtainIndexBuffer(ASURA_PS3_RENDERCACHE_BUFFER_VIEWCULLED,uSkinIdx);

		Asura_Assert( pIndexBuffer, "RenderCharacter: pIndexBuffer is NULL" );
		Asura_Assert( pIndexBuffer->GetIndexBufferSize() >= pxCharacterStrips->GetIndexBuffer()->GetIndexBufferSize(), "Cached buffer num indices too small" );
	}
	else
	{
		//failed to setup SPU processing, do it using a vertex shader
		pxCharacterStrips->GetVertexBuffer()->Bind( g_axVertexDeclaration, VD_ELEMENTS );

		pIndexBuffer = pxCharacterStrips->GetIndexBuffer();
	}
	
	return pIndexBuffer;
}

////////////////////////////////////////////////////////////////////////////////
//
// AVP_Platform_CloakEffect::RenderCloakedSkin()
//
// Render the cloak effect - the distorted image seen through the predator when cloaked
// and the glowing blue part of the transition effect.
//
////////////////////////////////////////////////////////////////////////////////
void AVP_Platform_CloakEffect::RenderCloakedSkin(const AVP_CloakEffect *pxCloakEffect,
												 const Asura_Animation& xAnim,
												 const Asura_Hierarchy_Skin& xSkin,
												 Asura_AMatrix4x4_InParam xTransform,
												 Asura_AMatrix4x4_InParam xInverseTransform,
												 const Asura_AMatrix4x4* const pxBoneTransforms,
												 u_int uSkinIdx )
{
	Asura_PS3_DeferredRendering::ResolveBackBuffer();

	Asura_PS3_CharacterRenderer_Skin* pxCharacterStrips = Asura_PS3_CharacterRenderer::GetCharacter( xSkin.GetHashID() );

	if( !pxCharacterStrips )
	{
		Asura_Assertf( false, ( "AVP_Platform_CloakEffect::RenderCloakedSkin - Could not get render data for smooth skin '%s'.", xSkin.GetName()) );
		return;
	}

	ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__MISC, "RenderCloakedSkin" );




	// process vertices & setup vertex shader
	Asura_PS3_GCM_IndexBuffer* pIndexBuffer = SetupSkinning( pxCharacterStrips, xAnim, xTransform, xInverseTransform, pxBoneTransforms, xSkin.NumberOfBones, uSkinIdx, pxCloakEffect->GetTransitionValue(), pxCloakEffect->GetMaxVisibilityAngle(), pxCloakEffect->GetMaxVisibilityDistance() );


	// setup pixel shader
	if ( pxCloakEffect->m_eState == AVP_CloakEffect::CLOAKED )
	{
		g_pxFragmentShader = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKEFFECT_SIMPLE_SHADER );
	}
	else
	{
		g_pxFragmentShader = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKEFFECT_SHADER );
	}

	const float fViewportWidth = static_cast< float >( Asura_Viewport::GetWidth() );
	float fScreenheight = Asura_PS3_Graphics::GetScreenHeight();

	const float fDistortionAmount = 256.0f * fViewportWidth / fScreenheight;
	float xDistortion[ 2 ] = { fDistortionAmount, -fDistortionAmount };
	g_pxFragmentShader->SetFloats( FP_DistortionScales, 2, xDistortion );


	if ( pxCloakEffect->m_eState != AVP_CloakEffect::CLOAKED )
	{
		float fTime = Asura_Timers::GetLocalScaledTime() * pxCloakEffect->m_fVariationSpeed;
		float afOffset[4] = {
			Asura_Maths::Sine(fTime) * pxCloakEffect->m_fVariationScale + s_afVariations[pxCloakEffect->m_uSeed][0], 
			Asura_Maths::Cosine(fTime * 1.53f) * pxCloakEffect->m_fVariationScale + s_afVariations[pxCloakEffect->m_uSeed][1], 
			Asura_Maths::Cosine(fTime * 1.17f) * pxCloakEffect->m_fVariationScale + s_afVariations[pxCloakEffect->m_uSeed][2], 
			1.0f
		};

		float afParams[2] = {pxCloakEffect->m_fEffectScale, 1.0f / pxCloakEffect->m_fLightningScale};

 		g_pxFragmentShader->SetFloats( FP_Params, 2, afParams);
 		g_pxFragmentShader->SetFloats( FP_Offset, 4, afOffset );
 		g_pxFragmentShader->SetFloats( FP_LightningColour, 4, &pxCloakEffect->m_xLightningColour.m_fR );
 		g_pxFragmentShader->SetFloats( FP_Time, 1, &fTime );

		g_pxFragmentShader->SetTexture( FP_NoiseTexture, s_iNoiseTexture );
	}
	
	g_pxFragmentShader->SetTexture( FP_SceneSampler, Asura_PS3_DeferredRendering::GetResolvedTexture(), false );


	g_pxFragmentShader->Bind();



	// Set render states
	{
		Asura_PS3_RenderingStates::SetStencilTestEnabled(false);

		Asura_PS3_RenderingStates::RequestAlphaTest( false );
		Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CW );
		Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_NORMAL );
		Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
		Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
		Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );

//BW irrelevant now, cloak drawn after FSFX - I'm keeping this in case we move the effect back to before FSFX
// 		//Write the alien's value when cloaked
// 		//BW - not sure if this is the best thing to do but it looks the same as PC and without it we can see the un-distorted glow doesn't
// 		// line up with the distorted image of the alien, with this flag set the glow is disabled when looking through the predator.
// 		Asura_PS3_RenderingStates::SetStencilTestEnabled( true );
// 		Asura_PS3_RenderingStates::SetStencilMask( AVP_STENCIL_MASK_VISION ); //read/write mask
// 		Asura_PS3_RenderingStates::SetStencilFunction( ASURA_PS3_STENCIL_ALWAYS, AVP_STENCIL_BIT_VISION_ALIEN, AVP_STENCIL_MASK_VISION ); //write val, read mask
// 		Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_REPLACE );
	}
	Asura_PS3_DeferredRendering::BindAsSingleRT();

	for( u_int u = 0; u < pxCharacterStrips->GetNumStrips(); ++u )
	{
#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		if (xAnim.GetSkinElementVisibility(pxCharacterStrips->GetStrip(u).m_iElementID))
#endif
		{			
			const Asura_Hash_ID uMaterialHash1 = pxCharacterStrips->GetStrip(u).m_auMaterialResponseHashIDs[0];
			const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

			if( pxMaterial1 )
			{
				if( pxMaterial1->GetDontRender() ) continue;
			}

			pIndexBuffer->RenderTriangles( pxCharacterStrips->GetStrip( u ).m_uNumIndices, pxCharacterStrips->GetStrip( u ).m_uStartIndex );
			Asura_PS3_Graphics::Flush();
		}
	}

	Asura_PS3_SPU_Skinning::NotifyBufferUsed();


	Asura_PS3_RenderingStates::SetStencilTestEnabled( false );


	ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__MISC );

}


////////////////////////////////////////////////////////////////////////////////
//
// AVP_Platform_CloakEffect::RenderHoles()
//
// This function draws a predator shape with holes cut in it. Calling this function
// to draw to the stencil buffer, then the character render testing the stencil
// will cut holes in the character without modifying the character shaders.
//
////////////////////////////////////////////////////////////////////////////////
void AVP_Platform_CloakEffect::RenderHoles(const AVP_CloakEffect *pxCloakEffect,
										   const Asura_Animation& xAnimation,
										   const Asura_Hierarchy_Skin& xSkin,
										   Asura_AMatrix4x4_InParam xTransform,
										   Asura_AMatrix4x4_InParam xInverseTransform,
										   const Asura_AMatrix4x4* const pxBoneTransforms,
										   u_int uSkinIdx )
{
	Asura_PS3_CharacterRenderer_Skin* pxCharacterStrips = Asura_PS3_CharacterRenderer::GetCharacter( xSkin.GetHashID() );

	if( !pxCharacterStrips )
	{
		Asura_Assertf( false, ( "AVP_Platform_CloakEffect::RenderHoles - Could not get render data for smooth skin '%s'.", xSkin.GetName()) );
		return;
	}


	Asura_PS3_GCM_IndexBuffer* pIndexBuffer = SetupSkinning( pxCharacterStrips, xAnimation, xTransform, xInverseTransform, pxBoneTransforms, xSkin.NumberOfBones, uSkinIdx, pxCloakEffect->GetTransitionValue(), pxCloakEffect->GetMaxVisibilityAngle(), pxCloakEffect->GetMaxVisibilityDistance() );

	// setup pixel shader
	Asura_PS3_FragmentShaderCopy* pxHoleFP = Asura_PS3_GCM_ShadersLibrary::GetPermanentFP( ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKHOLES_SHADER );

	float fTime = Asura_Timers::GetLocalScaledTime() * pxCloakEffect->m_fVariationSpeed;
	float afOffset[4] = {
		Asura_Maths::Sine(fTime) * pxCloakEffect->m_fVariationScale + s_afVariations[pxCloakEffect->m_uSeed][0], 
		Asura_Maths::Cosine(fTime * 1.53f) * pxCloakEffect->m_fVariationScale + s_afVariations[pxCloakEffect->m_uSeed][1], 
		Asura_Maths::Cosine(fTime * 1.17f) * pxCloakEffect->m_fVariationScale + s_afVariations[pxCloakEffect->m_uSeed][2], 
		1.0f
	};

	float afParams[2] = {pxCloakEffect->m_fEffectScale, 1.0f / pxCloakEffect->m_fLightningScale};

	pxHoleFP->SetFloats( CLOAK_HOLES_FP_PARAMS, 2, afParams);
	pxHoleFP->SetFloats( CLOAK_HOLES_FP_OFFSET, 4, afOffset);
	pxHoleFP->SetFloats( CLOAK_HOLES_FP_TIME, 1, &fTime );

	pxHoleFP->SetTexture( CLOAK_HOLES_FP_NoiseTexture, s_iNoiseTexture );

	pxHoleFP->Bind();
	
	for( u_int u = 0; u < pxCharacterStrips->GetNumStrips(); ++u )
	{
#ifdef ASURA_SUPPORT_SKIN_ELEMENTS
		if (xAnimation.GetSkinElementVisibility(pxCharacterStrips->GetStrip(u).m_iElementID))
#endif
		{			
			const Asura_Hash_ID uMaterialHash1 = pxCharacterStrips->GetStrip(u).m_auMaterialResponseHashIDs[0];
			const Asura_MaterialResponse* const pxMaterial1 = Asura_MaterialResponse_System::GetMaterialByHashID( uMaterialHash1 );

			if( pxMaterial1 )
			{
				if( pxMaterial1->GetDontRender() ) continue;
			}

			pIndexBuffer->RenderTriangles( pxCharacterStrips->GetStrip( u ).m_uNumIndices, pxCharacterStrips->GetStrip( u ).m_uStartIndex );
			Asura_PS3_Graphics::Flush();
		}
	}

	Asura_PS3_SPU_Skinning::NotifyBufferUsed();
}


void AVP_CloakEffect::Platform_BeginRender(void) const
{
}

void AVP_CloakEffect::Platform_RenderSkin(const Asura_Animation& xAnimation,
										  const Asura_Hierarchy_Skin& xSkin) const
{
	//BW I'm doing multiple skins in here, though really the system's set up better to handle them in the calling function
	//BW owing to the multi-stage nature of this process I expect better results by render all skins at each stage rather than the whole thing for each skin (which might show the seams)
	//BW todo: re-evaluate once we actually have a multiple-skin predator
	const float fDistance = (Asura_Hierarchy_System::GetWorldspaceBonePos(0) - *Asura_Camera::GetPosition()).Magnitude();
	char szTemp[512] = "";


	if ( Asura_Render::IsZPrepass() )
	{
		// 1st pass, draw the bits that need deferred lighting
	
		//
		// HOLES - render to stencil
		//
		if ( m_eState != CLOAKED && g_bEnableHoles )
		{
#ifdef ASURA_USE_GCM_REPLAY
			sprintf( szTemp, "AVP_CloakEffect Render Holes %s", xSkin.GetName() );
#endif
			ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__MISC, szTemp );

			// Set render states for drawing the holes
			Asura_PS3_RenderingStates::RequestAlphaTest( false );
			Asura_PS3_RenderingStates::RequestCullMode( ASURA_CULLMODE_CW );
			Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_OFF );
			Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
			Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
			Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
			Asura_PS3_RenderingStates::SetColorMaskMrt0( ASURA_PS3_COLOR_MASK_MRT0_NONE );
			
			//don't need to draw to MRT's
			Asura_PS3_DeferredRendering::BindAsSingleRT();
			
			//set stencil write
			Asura_PS3_RenderingStates::SetStencilTestEnabled( true );
			Asura_PS3_RenderingStates::SetStencilFunction( ASURA_PS3_STENCIL_ALWAYS, AVP_STENCIL_MASK_CLOAK, AVP_STENCIL_MASK_CLOAK );
			Asura_PS3_RenderingStates::SetStencilMask(AVP_STENCIL_MASK_CLOAK);
			Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_REPLACE, ASURA_PS3_STENCIL_OPERATION_REPLACE, ASURA_PS3_STENCIL_OPERATION_REPLACE );

			//BW We're gonna render the predator 3 times during the transition, this means a big overhead for vertex shading
			//BW but at some point I'll be doing skinning on SPU's, so we can just skin once and run the fragment program 3 times.

			for( int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex ) // support multiple skins per character
			{
				const Asura_Hierarchy_Skin* pxSkin = xAnimation.GetSkin(iSkinIndex);
				if (pxSkin)
				{
					const Asura_Hierarchy_Skin* const pxSkinLOD = pxSkin->GetDetailSkinFromDistance( fDistance );
				
					AVP_Platform_CloakEffect::RenderHoles(this, xAnimation, *pxSkinLOD, Asura_Hierarchy_System::s_xBaseTransform, Asura_Hierarchy_System::s_xInverseBaseTransform, Asura_Hierarchy_System::s_axBoneTransforms, iSkinIndex);
				}
			}

			Asura_PS3_RenderingStates::SetColorMaskMrt0( ASURA_PS3_COLOR_MASK_MRT0_ALL );

			ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__MISC );

		}

		//
		// CHARACTER - masked by stencil
		//
		if ( m_eState != CLOAKED && g_bEnableUnholes )
		{
#ifdef ASURA_USE_GCM_REPLAY
			sprintf( szTemp, "AVP_CloakEffect Render Character %s", xSkin.GetName() );
#endif
			ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__MISC, szTemp );

			if ( g_bEnableHoles )
			{
				//set render state to read the stencil buffer
				Asura_PS3_RenderingStates::SetStencilTestEnabled( true );
				Asura_PS3_RenderingStates::SetStencilMask(AVP_STENCIL_MASK_CLOAK);
				Asura_PS3_RenderingStates::SetStencilFunction( ASURA_PS3_STENCIL_EQUAL, AVP_STENCIL_MASK_CLOAK, AVP_STENCIL_MASK_CLOAK );
				Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_ZERO, ASURA_PS3_STENCIL_OPERATION_ZERO, ASURA_PS3_STENCIL_OPERATION_ZERO );
				
				//Unfortunately all that^ can be overridden by character render's "material context" (e.g. for heat vision mode)
				//so tell that system to combine what it does with what we want
				AVP_PlatformStencilSetup::SetTestFunction( ASURA_PS3_STENCIL_EQUAL, AVP_STENCIL_MASK_CLOAK, AVP_STENCIL_MASK_CLOAK );
			}

			for( int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex ) // support multiple skins per character
			{
				const Asura_Hierarchy_Skin* pxSkin = xAnimation.GetSkin(iSkinIndex);
				if (pxSkin)
				{
					const Asura_Hierarchy_Skin* const pxSkinLOD = pxSkin->GetDetailSkinFromDistance( fDistance );

		 			Asura_PS3_CharacterRenderer::RenderCharacter(xAnimation, xSkin, Asura_Hierarchy_System::s_xBaseTransform, Asura_Hierarchy_System::s_xInverseBaseTransform, Asura_Hierarchy_System::s_axBoneTransforms, iSkinIndex);
		 		}
		 	}
			ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__MISC );

 		}

		AVP_PlatformStencilSetup::ResetTestFunction();
		Asura_PS3_RenderingStates::SetStencilOperation( ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP, ASURA_PS3_STENCIL_OPERATION_KEEP );
		Asura_PS3_RenderingStates::SetStencilMask(0);
		Asura_PS3_RenderingStates::SetStencilTestEnabled( false );
	}
	else
	{
		// 2nd pass, draw the bits that need to happen after the scene is drawn (so it can be distorted)

		//
		// CLOAK & LIGHTNING
		//
		if( g_bEnableDistortion )
		{
#ifdef ASURA_USE_GCM_REPLAY
			sprintf( szTemp, "AVP_CloakEffect Render Cloak %s", xSkin.GetName() );
#endif
			ASURA_PS3_RSX_PROFILE__BEGIN( ASURA_PS3_RSX_PROFILE__MISC, szTemp );

			Asura_PS3_RenderingStates::SetUseDepthBias(true);
			Asura_PS3_RenderingStates::SetDepthBias(g_fCloakDepthUnits, g_fCloakDepthFactor);
		
			for( int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex ) // support multiple skins per character
			{
				const Asura_Hierarchy_Skin* pxSkin = xAnimation.GetSkin(iSkinIndex);
				if (pxSkin)
				{
					const Asura_Hierarchy_Skin* const pxSkinLOD = pxSkin->GetDetailSkinFromDistance( fDistance );

					AVP_Platform_CloakEffect::RenderCloakedSkin(this, xAnimation, *pxSkinLOD, Asura_Hierarchy_System::s_xBaseTransform, Asura_Hierarchy_System::s_xInverseBaseTransform, Asura_Hierarchy_System::s_axBoneTransforms, iSkinIndex);
				}
			}

			Asura_PS3_RenderingStates::SetUseDepthBias(false);
			Asura_PS3_RenderingStates::SetDepthBias(0.0f, 0.0f);

			ASURA_PS3_RSX_PROFILE__END( ASURA_PS3_RSX_PROFILE__MISC );

		}

		//
		// DEBUG - show where the holes are, on top of everything we just drew
		//
		if ( m_eState != CLOAKED
			&& g_bShowHoles )
		{
			// Set render states for drawing the holes
			Asura_PS3_RenderingStates::RequestAlphaTest( false );
			Asura_PS3_RenderingStates::RequestTranslucencyMode( ASURA_TRANSLUCENCY_ADD );
			Asura_PS3_RenderingStates::RequestZBufferCompareMode( ASURA_ZBUFFERCMP_LESSEQUAL );
			Asura_PS3_RenderingStates::RequestZBufferEnabled( true );
			Asura_PS3_RenderingStates::RequestZBufferWriteEnabled( false );
			Asura_PS3_RenderingStates::SetUseDepthBias(true);
			Asura_PS3_RenderingStates::SetDepthBias(-6, 0);

			//don't need to draw to MRT's
			Asura_PS3_DeferredRendering::BindAsSingleRT();

			for( int iSkinIndex = 0; iSkinIndex < ASURA_ANIMATION_SKIN_COUNT; ++iSkinIndex ) // support multiple skins per character
			{
				const Asura_Hierarchy_Skin* pxSkin = xAnimation.GetSkin(iSkinIndex);
				if (pxSkin)
				{
					const Asura_Hierarchy_Skin* const pxSkinLOD = pxSkin->GetDetailSkinFromDistance( fDistance );

					AVP_Platform_CloakEffect::RenderHoles(this, xAnimation, *pxSkinLOD, Asura_Hierarchy_System::s_xBaseTransform, Asura_Hierarchy_System::s_xInverseBaseTransform, Asura_Hierarchy_System::s_axBoneTransforms, iSkinIndex);
				}
			}

			Asura_PS3_RenderingStates::SetUseDepthBias(false);
			Asura_PS3_RenderingStates::SetDepthBias(0.0f, 0.0f);
		}
	}
}

void AVP_CloakEffect::Platform_EndRender(void) const
{
}


#if defined(ASURA_PS3_ALLOW_SHADER_RELOAD)
bool AVP_Platform_CloakEffect::ReloadShaders( bool bHelp )
{
	if ( bHelp )
	{
		Asura_CommandConsole::Printf("recompile & load predator cloak shader");
		return true;
	}

	const char *aszIncludePath[2]= { "~/Project/AVP/Game/PS3/Shaders/", NULL };
	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader( "~/Project/AVP/Game/PS3/Shaders/CloakEffect_Simple_FP.fpx",
		ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKEFFECT_SIMPLE_SHADER, NULL, aszIncludePath, new Asura_PS3_CloakEffect_Handles );


	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader( "~/Project/AVP/Game/PS3/Shaders/CloakEffect_FP.fpx",
		ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKEFFECT_SHADER, NULL, aszIncludePath, new Asura_PS3_CloakEffect_Handles );

	Asura_PS3_GCM_ShadersLibrary::LoadAndCompileShader( "~/Project/AVP/Game/PS3/Shaders/CloakHoles_FP.fpx",
		ASURA_SHADER_HASH_PROJECT, ASURA_PS3_CLOAKHOLES_SHADER, NULL, aszIncludePath, new Asura_PS3_CloakHoles_Handles );

	return true;
}
#endif

// NEXT - SPUs

/// End of file. ///
