//////////////////////////////////////////////////////////////////////////////
//
// AVP_Xbox360_Stats.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
// Description:
//
//     code for reading/writing stats to 360 Live! leaderboards
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// Asura includes
#include "Network/Online/Asura_Leaderboards.h"

// Asura 360
#include "Network/Online/Asura_Xbox360_Stats.h"

// AVP independent
#include "Core/AVP_Settings.h"
#include "Entity/PlayerInfo/AVP_ClientEntity_PlayerInfo.h"
#include "Network/AVP_Leaderboards.h"

// project XLAST include
#include "AVP3.spa.h"

//#define AVP_HACK_TRUESKILL 1 // define this to replace the score leaderboard with the TrueSkill leaderboard.
// S T A T I C   G L O B A L   D A T A ///////////////////////////////////////
static const u_int g_auLeaderboardID[AVP_NUM_LEADERBOARDS] = 
{
	STATS_VIEW_KILLS,
#ifdef AVP_HACK_TRUESKILL
	STATS_VIEW_SKILL_RANKED_NORMAL,
#else
	STATS_VIEW_SCORE,
#endif
	STATS_VIEW_MATCHES,
	STATS_VIEW_MYSTATS,
	STATS_VIEW_INTERNALSTATS,
};

static const u_int g_auKillLeaderboardProperties[AVP_NUM_KILLS_STATS] = 
{
	PROPERTY_TOTALKILLS,
	PROPERTY_KILLSASMARINE,
	PROPERTY_KILLSASALIEN,
	PROPERTY_KILLSASPREDATOR,
	PROPERTY_KILLTODEATHRATIO,
	PROPERTY_KILLSTREAK,
	PROPERTY_TOTALDEATHS,
};

static const u_int g_auScoreLeaderboardProperties[AVP_NUM_SCORE_STATS] = 
{
	PROPERTY_TOTALSCORE,
	PROPERTY_SCOREMARINE,
	PROPERTY_SCOREALIEN,
	PROPERTY_SCOREPREDATOR,
	PROPERTY_TIMEPLAYED,
};

static const u_int g_auMatchesLeaderboardProperties[AVP_NUM_MATCHES_STATS] = 
{
	PROPERTY_TOTALMATCHESPLAYED,
	PROPERTY_MATCHESPLAYEDASMARINE,
	PROPERTY_MATCHESPLAYEDASALIEN,
	PROPERTY_MATCHESPLAYEDASPREDATOR,
	PROPERTY_MATCHESWON,
	PROPERTY_MATCHESLOST,
	PROPERTY_MATCHWINNINGSTREAK,
};

static const u_int g_auMyStatsLeaderboardProperties[AVP_NUM_MYSTATS_STATS] = 
{
	PROPERTY_MYSTATSTOTALRANK,
	PROPERTY_MYSTATSMARINERANK,
	PROPERTY_MYSTATSALIENRANK,
	PROPERTY_MYSTATSPREDATORRANK,
	PROPERTY_MYSTATSMARINEHEADSHOTS,
	PROPERTY_MYSTATSALIENHEADBITES,
	PROPERTY_MYSTATSPREDATORTROPHIES,
};

static const u_int g_auInternalLeaderboardProperties[AVP_NUM_INTERNAL_STATS] = 
{
	PROPERTY_QUITEARLY,
};

static const u_short g_auKillLeaderboardColumns[AVP_NUM_KILLS_STATS] = 
{
	0, 
	STATS_COLUMN_KILLS_MARINEKILLS,
	STATS_COLUMN_KILLS_ALIENKILLS,
	STATS_COLUMN_KILLS_PREDATORKILLS,
	STATS_COLUMN_KILLS_KILLTODEATHRATIO,
	STATS_COLUMN_KILLS_KILLSTREAK,
	STATS_COLUMN_KILLS_NUM_DEATHS,
};

static const u_short g_auScoreLeaderboardColumns[AVP_NUM_SCORE_STATS] = 
{
	0,
#ifdef AVP_HACK_TRUESKILL
	X_STATS_COLUMN_SKILL_SKILL,
	X_STATS_COLUMN_SKILL_GAMESPLAYED,
	X_STATS_COLUMN_SKILL_MU,
	X_STATS_COLUMN_SKILL_SIGMA,
#else
	STATS_COLUMN_SCORE_MARINESCORE,
	STATS_COLUMN_SCORE_ALIENSCORE,
	STATS_COLUMN_SCORE_PREDATORSCORE,
	STATS_COLUMN_SCORE_TIMEPLAYED,
#endif
};

static const u_short g_auMatchesLeaderboardColumns[AVP_NUM_MATCHES_STATS] = 
{
	0,
	STATS_COLUMN_MATCHES_PLAYEDASMARINE,
	STATS_COLUMN_MATCHES_PLAYEDASALIEN,
	STATS_COLUMN_MATCHES_PLAYEDASPREDATOR,
	STATS_COLUMN_MATCHES_MATCHESWON,
	STATS_COLUMN_MATCHES_MATCHESLOST,
	STATS_COLUMN_MATCHES_WINNINGSTREAK,
};

static const u_short g_auMyStatsLeaderboardColumns[AVP_NUM_MYSTATS_STATS] = 
{
	0,
	STATS_COLUMN_MYSTATS_CURRENTRANKMARINE,
	STATS_COLUMN_MYSTATS_CURRENTRANKALIEN,
	STATS_COLUMN_MYSTATS_CURRENTRANKPREDATOR,
	STATS_COLUMN_MYSTATS_HEADSHOTS,
	STATS_COLUMN_MYSTATS_HEADBITES,
	STATS_COLUMN_MYSTATS_TROPHIES,
};

static const u_short g_auInternalLeaderboardColumns[AVP_NUM_INTERNAL_STATS] = 
{
	0,
};

static const u_int* g_auLeaderboardPropertiesArray[AVP_NUM_LEADERBOARDS] = 
{
	g_auKillLeaderboardProperties,
	g_auScoreLeaderboardProperties,
	g_auMatchesLeaderboardProperties,
	g_auMyStatsLeaderboardProperties,
	g_auInternalLeaderboardProperties,
};

static const u_short* g_auLeaderboardColumnsArray[AVP_NUM_LEADERBOARDS] = 
{
	g_auKillLeaderboardColumns,
	g_auScoreLeaderboardColumns,
	g_auMatchesLeaderboardColumns,
	g_auMyStatsLeaderboardColumns,
	g_auInternalLeaderboardColumns,
};

u_int Asura_Xbox360_Stats::Project_GetLeaderboardID( u_int uLeaderboard )
{
	// returns ID from project spa.h file
	Asura_Assert(uLeaderboard < AVP_NUM_LEADERBOARDS, "invalid leaderboard");
	return g_auLeaderboardID[uLeaderboard];
}

u_int Asura_Xbox360_Stats::Project_GetTrueSkillLeaderboardID( bool bRanked )
{
	return bRanked ? STATS_VIEW_SKILL_RANKED_NORMAL : STATS_VIEW_SKILL_STANDARD_NORMAL;
}

u_short Asura_Xbox360_Stats::Project_GetColumnID( u_int uLeaderboard, u_int uColumn )
{
	Asura_Assert(uLeaderboard < AVP_NUM_LEADERBOARDS, "invalid leaderboard");
	Asura_Assert(uColumn < Asura_Leaderboard_System::Project_GetNumLeaderboardColumns(uLeaderboard), "invalid column");
	return g_auLeaderboardColumnsArray[uLeaderboard][uColumn];
}

// Project_GetStat() must fill out the XUSER_PROPERTY with the property ID from the spa.h file
// and the XUSER_DATA value as appropriate. If it fails (e.g. player wasn't found or stat doesn't
// exist then it returns false and the xStat data must be ignored). Otherwise returns true.
bool Asura_Xbox360_Stats::Project_GetStat(  XUID xXuid, bool bDisconnectedEarly, 
											u_int uGuid, u_int uLeaderboard, u_int uColumn, 
											XUSER_PROPERTY& xStat )
{
	Asura_Assert(uLeaderboard < AVP_NUM_LEADERBOARDS, "invalid leaderboard");
	Asura_Assert(uColumn < Asura_Leaderboard_System::Project_GetNumLeaderboardColumns(uLeaderboard), "invalid stat index");
	const u_int* puPropertyArray = g_auLeaderboardPropertiesArray[uLeaderboard];
	xStat.dwPropertyId = puPropertyArray[uColumn];

	// use the client entity player info (which should exist on the server and contain the same results)
	const AVP_ClientEntity_PlayerInfo* pxPlayerInfo = AVP_ClientEntity_PlayerInfo::FindEntity(uGuid);
	Asura_Assert(pxPlayerInfo || bDisconnectedEarly, "failed to get player info to write stats");

	AVP_Multiplayer_PlayerScore xNullScore( pxPlayerInfo ? pxPlayerInfo->GetGuid() : ASURA_GUID_UNREGISTERED );
	const AVP_Multiplayer_PlayerScore& xScore = pxPlayerInfo ? pxPlayerInfo->GetPlayerScore() : xNullScore;

	const u_int uAVPStatID = AVP_Leaderboards::GetAssociatedStatID(uLeaderboard, uColumn);
	if (AVP_Leaderboards::IsAdditiveStat(uAVPStatID))
	{
		// nice and simple additive stat, only complication is setting 32-bit or 64-bit data as appropriate
		const u_int uStatResult = AVP_Leaderboards::GetStatResultById( uAVPStatID, xScore );
		if (uColumn == 0)
		{
			// ranking column, expecting int64 which all start with 0x20000000
			Asura_Assert((xStat.dwPropertyId & 0x20000000) == 0x20000000, "unexpected property ID");
			xStat.value.type = XUSER_DATA_TYPE_INT64;
			xStat.value.i64Data = uStatResult;
		}
		else
		{
			// non-ranking additive column, expecting int32 which all start with 0x10000000
			Asura_Assert((xStat.dwPropertyId & 0x10000000) == 0x10000000, "unexpected property ID");
			xStat.value.type = XUSER_DATA_TYPE_INT32;
			Asura_Assert(uStatResult <= INT_MAX, "stat result too big to fit in integer");
			xStat.value.nData = static_cast<int>(uStatResult);
		}
	}
	else
	{
		// computed stat, handle on case-by-case basis
		switch (xStat.dwPropertyId)
		{
			case PROPERTY_KILLTODEATHRATIO:
			{
				// need to read total kills and total deaths stats for this player and compute the value.
				// NB: 32-bit floats are not supported by the stats system.
				const XUSER_STATS_ROW* pxPlayerData = Asura_Xbox360_Stats_PreGameReader::GetDataForXuid(xXuid);
				double xValue = 0.0;
				if (pxPlayerData && pxPlayerData->dwNumColumns == 1 && 
					pxPlayerData->pColumns[0].wColumnId == STATS_COLUMN_KILLS_NUM_DEATHS && 
					pxPlayerData->pColumns[0].Value.type == XUSER_DATA_TYPE_INT32)
				{
					// total kills is the ranked stat we asked for in the pre-game reader
					uint64 ulTotalKills = pxPlayerData->i64Rating + xScore.GetNumberOfKills();
					u_int uTotalDeaths = pxPlayerData->pColumns[0].Value.nData + xScore.GetNumberOfTimesKilled();
					xValue = uTotalDeaths ? static_cast<double>(ulTotalKills) / static_cast<double>(uTotalDeaths) : 1.0;
				}
				else
				{
					Asura_Assert(false, "didn't get total kills and deaths from before this match; can't update k:d ratio");
				}
				xStat.value.type = XUSER_DATA_TYPE_DOUBLE;
				xStat.value.dblData = xValue;
				break;
			}

			case PROPERTY_MATCHWINNINGSTREAK:
			{
				// TODO: get this from somewhere.
				xStat.value.type = XUSER_DATA_TYPE_INT32;
				xStat.value.nData = 0;
				break;
			}

			case PROPERTY_QUITEARLY:
			{
				// NB: this is the ranked column, so it must use 64-bit
				xStat.value.type = XUSER_DATA_TYPE_INT64;
				xStat.value.i64Data = bDisconnectedEarly ? 1 : 0;
				break;
			}

			default:
			{
				Asura_Assert(false, "unknown stat type");
				return false;
			}
		}
	}

	return true;
}

// Project_GetTrueSkillRelativeScore() and Project_GetTrueSkillTeam should return values appropriate
// for TrueSkill, see "Writing to TrueSkill Leaderboards" in the XDK docs.
u_int Asura_Xbox360_Stats::Project_GetTrueSkillRelativeScore( bool bDisconnectedEarly, u_int uGuid )
{
	// hash-but-fair: players who leave early get put to the bottom of the list.
	if (bDisconnectedEarly)
	{
		return 0;
	}

	// use the client entity player info (which should exist on the server and contain the same results)
	const AVP_ClientEntity_PlayerInfo* pxPlayerInfo = AVP_ClientEntity_PlayerInfo::FindEntity(uGuid);
	Asura_Assert(pxPlayerInfo, "failed to get player info for TrueSkill");
	if (!pxPlayerInfo)
	{
		return 0;
	}

	const AVP_Multiplayer_PlayerScore& xScore = pxPlayerInfo->GetPlayerScore();

	const int iPlayerScore = xScore.GetIndividualScore();
	// relative score is (num_players - num_better_players - 1). This means that ties all score equal.

	// we iterate over the whole list for this, and this function is called once per player 
	// so this algorithm is n^2 but n <= 18 so it should be OK.
	const Asura_Hashed_List<AVP_ClientEntity_PlayerInfo>* pxPlayerInfoList = AVP_ClientEntity_PlayerInfo::GetList();
	const u_int uNumPlayers = pxPlayerInfoList->GetNumEntries();
	u_int uNumBetterPlayers = 0;
	Asura_Hashed_List_It<AVP_ClientEntity_PlayerInfo> xPlayerInfoListIt( pxPlayerInfoList ); 
	for ( ; !xPlayerInfoListIt.Done(); xPlayerInfoListIt.Next() )
	{
		AVP_ClientEntity_PlayerInfo* pxPlayerInfo = xPlayerInfoListIt.GetCurrent();
		if (pxPlayerInfo->GetPlayerScore().GetIndividualScore() > iPlayerScore)
		{
			// current player is automatically rejected by this comparison
			++uNumBetterPlayers;
		}
	}

	Asura_Assert(uNumPlayers - uNumBetterPlayers > 0, "can't be more better players than there are other players");
	return uNumPlayers - uNumBetterPlayers - 1;
}

u_int Asura_Xbox360_Stats::Project_GetTrueSkillTeam( u_int uGuid )
{
	// use the client entity player info (which should exist on the server and contain the same results)
	const AVP_ClientEntity_PlayerInfo* pxPlayerInfo = AVP_ClientEntity_PlayerInfo::FindEntity(uGuid);
	Asura_Assert(pxPlayerInfo, "failed to get player info for TrueSkill");
	if (!pxPlayerInfo)
	{
		// return invalid value
		return AVP_Settings::AVP_NUM_TEAMS;
	}

	u_int uTeam;
	AVP_Settings::Team eTeam = pxPlayerInfo->GetMultiplayerTeam();
	static u_int ls_uNextTeamID = 0;
	if (eTeam == AVP_Settings::AVP_TEAM_NONE)
	{
		uTeam = uGuid;
	}
	else
	{
		uTeam = static_cast<u_int>(eTeam);
	}
	return uTeam;
}


// Asura_Xbox360_Stats_PreGameReader Project functions - assuming all the data we need is in one leaderboard.
// Can probably ensure this by creating a new leaderboard if necessary.
// Alternatively this class could iterate over more than one leaderboard 
// using an array of stats readers.
u_int Asura_Xbox360_Stats_PreGameReader::Project_GetPreGameLeaderboardID()
{
	return STATS_VIEW_KILLS;
}

u_int Asura_Xbox360_Stats_PreGameReader::Project_GetNumColumns()
{
	// we get total kills from the main stat of the leaderboard.
	return 1;
}

u_short Asura_Xbox360_Stats_PreGameReader::Project_GetColumnID( u_int uColumnIndex )
{
	Asura_Assert(uColumnIndex == 0, "unexpected column Index");
	return g_auKillLeaderboardColumns[AVP_KILLS_TOTAL_DEATHS];
}
