///////////////////////////////////////////////////////////////////////////////
//
// File: AVP_Platform_Laser.cpp
//
// Created by Andrew Southgate 12/11/2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// Handles rendering of the predators tri-laser thingy
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"


// Asura Platform-independent
#include "Camera/Asura_Camera.h"
#include "Core/Asura_Maths.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"
#include "Rendering/Asura_Colour.h"

// Asura PC
#include "Rendering/Asura_Xbox360_Graphics.h"
#include "SpecialFX/AVP_Laser.h"
#include "SpecialFX/AVP_Platform_Laser.h"

// Shader Include Files
#include "AVP_Laser.h"

Direct3D_VertexDeclaration*	AVP_Platform_Laser_System::s_pxLaserVertexDecl			= 0;
Asura_Xbox360_Effect*		AVP_Platform_Laser_System::s_pxLaserEffect				= 0;

Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xColourHandle;
Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xParamsHandle;
Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xBeamFadeOffHandle;

Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xTimeHandle;

Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xNoiseTexHandle;
Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xSceneTextureHandle;
Asura_Xbox360_Effect_System::Handle	AVP_Platform_Laser_System::s_xDepthTextureHandle;
int	AVP_Platform_Laser_System::s_iNoiseTexture;


struct AVP_Laser_Vertex
{
	float x, y, z, w;
	float x2, y2;
	void SetPosition(const Asura_Vector_3 &xPosition)
	{
		x = xPosition.GetX();
		y = xPosition.GetY();
		z = xPosition.GetZ();
		w = 1.0f;
	}
	void SetPos2(float fx, float fy)
	{
		x2 = fx, y2 = fy;
	}
	void Set(const Asura_Vector_3 &xPosition, float fx, float fy)
	{
		SetPosition(xPosition);
		SetPos2(fx,fy);
	}
};


void AVP_Platform_Laser_System::Initialise()
{
	if( NULL == Asura_Xbox360_Graphics::GetDevice())
	{
		return;
	}

	Direct3D_VertexElement auVSD[] =
	{
		{	0,	0,	D3DDECLTYPE_FLOAT4,		D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_POSITION,	0 },		
		{	0,	16,	D3DDECLTYPE_FLOAT2,		D3DDECLMETHOD_DEFAULT,	D3DDECLUSAGE_TEXCOORD,	0 },		
		D3DDECL_END()
	};

	Asura_Xbox360_Graphics::GetDevice()->CreateVertexDeclaration( auVSD, &s_pxLaserVertexDecl );

	s_iNoiseTexture = Asura_Texture_Management::AddTexture("\\SpecialFX\\AVP_Laser.dds");
	Asura_Texture_Management::LoadTextures();
	CaptureEffect();
}

void AVP_Platform_Laser_System::CaptureEffect()
{
#define LOADPARAMETER(codeName, shaderName) { codeName = s_pxLaserEffect->GetParameterHandle(shaderName); Asura_Assert(codeName, "Could not get shader parameter '" shaderName "'."); }


	if (!s_pxLaserEffect)
	{
		s_pxLaserEffect = Asura_Xbox360_Effect_System::CreateEffectFromMemory(g_auAVP_Laser);
	}

	if (s_pxLaserEffect)
	{	
		LOADPARAMETER(s_xColourHandle, "g_xColour");
		LOADPARAMETER(s_xParamsHandle, "g_xParams");
		LOADPARAMETER(s_xBeamFadeOffHandle, "g_xBeamFadeOff");

		LOADPARAMETER(s_xTimeHandle, "g_fTime");

		LOADPARAMETER(s_xNoiseTexHandle, "g_xNoiseTextureSampler");
		LOADPARAMETER(s_xSceneTextureHandle, "g_xSceneTextureSampler");
		LOADPARAMETER(s_xDepthTextureHandle, "g_xDepthTextureSampler");
	}
}

static Asura_RGBA g_fMaxFXColour;

inline const float *_VecF(float fX, float fY, float fZ, float fW = 0.0f)
{
	static float floats[4];
	floats[0] = fX;
	floats[1] = fY;
	floats[2] = fZ;
	floats[3] = fW;
	return floats;
}


inline const float *_VecF(const Asura_Vector_3 &xVec, float fW)
{
	return _VecF(xVec.GetX(), xVec.GetY(), xVec.GetZ(), fW);
}

void AVP_Platform_Laser_System::RenderStart()
{
	PIXBeginNamedEvent( D3DCOLOR_XRGB( 0xA0, 0xFF, 0x20 ), "Laser" );

	// Setup initial renderstates
	Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_CW);
	Asura_Render::CurrentStates.RequestZBufferEnabled(true);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(false);
	Asura_Render::CurrentStates.RequestTranslucencyMode(ASURA_TRANSLUCENCY_GLOWING);
	Asura_Render::CurrentStates.RequestAlphaTest(false);

	// (AAM - 26.07.2007) Need to use high-precision blending in case HDR is on
	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, TRUE );

	Asura_Render::ApplyCameraSettings();

	s_pxLaserEffect->GetInterface()->BeginTechniqueFromIndex( 0, 0 );

	Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration(s_pxLaserVertexDecl);	
	
	g_fMaxFXColour = Asura_RGBA(0, 0);

	s_pxLaserEffect->SetFloat(s_xTimeHandle, Asura_Timers::GetLocalScaledTime());

	s_pxLaserEffect->SetTexture(s_xNoiseTexHandle, s_iNoiseTexture);

	// Set the depth-normal texture
	s_pxLaserEffect->SetTexture(s_xDepthTextureHandle, Asura_Xbox360_Graphics::GetDepthMtlTexture());

}

void AVP_Platform_Laser_System::SetupLaserParams(AVP_Laser *pxLaser)
{

	s_pxLaserEffect->SetVector(s_xParamsHandle, 
		_VecF(
		pxLaser->GetBeamRadius(), 
		Asura_Maths::Max(pxLaser->GetBeamSharpness(), 0.01f),
		pxLaser->GetBeamSpread(),
		1.0f / pxLaser->GetBeamVisibilityDistance()
		)
		);
	s_pxLaserEffect->SetVector(s_xBeamFadeOffHandle, 
		_VecF(
		pxLaser->GetMaximumLength() - pxLaser->GetFallOffLength(), 
		pxLaser->GetFallOffLength() ? 1.0f / pxLaser->GetFallOffLength() : 0.0f, 
		0.0f
		)
		);


}


void AVP_Platform_Laser_System::RenderLaser(AVP_Laser *pxLaser, const Asura_Vector_3 &xPosition, const Asura_Vector_3 &xRight, const Asura_Vector_3 &xDown, const Asura_Vector_3 &xForward, float fLength)
{
	const Asura_Matrix_3x3& xCamOri = *Asura_Camera::GetOrientation();
	const Asura_Vector_3& xCamPos = *Asura_Camera::GetPosition();

	const Asura_Vector_3 xCamRight( xCamOri.e11, xCamOri.e21, xCamOri.e31 );
	const Asura_Vector_3 xCamUp( xCamOri.e12, xCamOri.e22, xCamOri.e32 );
	const Asura_Vector_3 xCamForward( xCamOri.e13, xCamOri.e23, xCamOri.e33 );

	const Asura_Vector_3 xBeamEnd = xPosition + xForward * fLength;

	Asura_Vector_3 xCamRelativeDir = xCamPos - xPosition;

	float fDistAlongBeam = xCamRelativeDir * xForward;

	float fAttenuation;
	if (fDistAlongBeam < 0.0f)
	{
		float fDistFromOrigin = xCamRelativeDir.Normalise();

		fAttenuation = Asura_Maths::SmoothStep(fDistFromOrigin, pxLaser->GetFallOffStartDistance(), pxLaser->GetFallOffEndDistance());
	} 
	else 
	{
		float fCylinderAttenuation = 
			Asura_Maths::SmoothStep((xCamRelativeDir - xForward * Asura_Maths::Max(fDistAlongBeam, 0.0f)).Magnitude(), 
			pxLaser->GetFallOffStartDistance(), pxLaser->GetFallOffEndDistance());

		xCamRelativeDir.Normalise();
		float fConeAttenuation = 
			Asura_Maths::SmoothStep(Asura_Maths::ArcCosine(Asura_Maths::Min(xCamRelativeDir * xForward, 1.0f)), 
			pxLaser->GetFallOffStartAngle(), pxLaser->GetFallOffEndAngle());

		fAttenuation = Asura_Maths::Min(fCylinderAttenuation, fConeAttenuation);
	}

	Asura_RGBA	xColour	= pxLaser->GetColour();
	fAttenuation = 1.0f - Asura_Maths::Max(fAttenuation - pxLaser->GetFallOffAngleEndIntensity(), 0.0f);
	xColour.m_fA *= fAttenuation;
	s_pxLaserEffect->SetVector(s_xColourHandle, (float*)(&xColour));


	float fDot = xCamForward * xForward;
	if (fDot < -0.8f)
	{
		const Asura_Vector_3 xRelativeCamPos = xCamPos - xPosition;
		float fDistAlong = xRelativeCamPos * xForward;
		// check to make sure the beam isnt stopped before it reaches us
		if (fDistAlong < fLength)
		{
			const Asura_Vector_3 xProjectedPos = xForward * fDistAlong;
			float fDistFrom = (xRelativeCamPos - xProjectedPos).MagnitudeSquared(); 
			float fRadiusOuter = pxLaser->GetBeamRadius() + pxLaser->GetBeamSharpness() + pxLaser->GetBeamSpread() * fDistAlong;
			float fIntens;

			if (fDistFrom <= pxLaser->GetBeamRadius())
			{
				fIntens = 1.0f;
			} if (fDistFrom > fRadiusOuter)
			{
				fIntens = 0.0f;
			} else {
				fIntens = 1.0f - (fDistFrom - pxLaser->GetBeamRadius()) / (fRadiusOuter - pxLaser->GetBeamRadius());
			}
			fIntens *= fIntens * Asura_Maths::Abs(fDot + 0.8f) * (1.0f / 0.2f);

			if (fDistAlong > pxLaser->GetMaximumLength())
			{
				fIntens = 0.0f;
			} 
			else if (fDistAlong > pxLaser->GetMaximumLength() - pxLaser->GetFallOffLength())
			{
				fIntens *= 1.0f - (fDistAlong - (pxLaser->GetMaximumLength() - pxLaser->GetFallOffLength())) / pxLaser->GetFallOffLength();
			}


			Asura_RGBA xColour = pxLaser->GetColour() * fIntens * fAttenuation;

			if (xColour.m_fA > g_fMaxFXColour.m_fA)
			{
				g_fMaxFXColour.m_fA = xColour.m_fA;
			}
			g_fMaxFXColour.m_fR = Asura_Maths::Max(xColour.m_fR, g_fMaxFXColour.m_fR);
			g_fMaxFXColour.m_fG = Asura_Maths::Max(xColour.m_fG, g_fMaxFXColour.m_fG);
			g_fMaxFXColour.m_fB = Asura_Maths::Max(xColour.m_fB, g_fMaxFXColour.m_fB);
		}
	}

	float fNearRadiusInner = pxLaser->GetBeamRadius();
	float fNearRadiusOuter = pxLaser->GetBeamRadius() + pxLaser->GetBeamSharpness();
	float fFarRadiusInner = pxLaser->GetBeamRadius();
	float fFarRadiusOuter = pxLaser->GetBeamRadius() + pxLaser->GetBeamSharpness() + fLength * pxLaser->GetBeamSpread();

	Asura_Vector_3 xEyeVector = xCamPos;
	xEyeVector -= xPosition;
	xEyeVector.Normalise();

	// The effect doesn't work at low angles to sight so we introduce 
	// a tolerance angle in which the far end of the beam is brought
	// up to the near end - there is a tolerance to stop things
	// flipping too much with small error

	Asura_Vector_3 xPerpVector;

	HRESULT hResult;
	AVP_Laser_Vertex* pxVertex;

	// Work out if in line with camera - if in line draw two 
	// separate triangle fans with the mirrored far and near ends
	// else draw a triangle strip with flattend ends facing the camera


	s_pxLaserEffect->GetInterface()->BeginPassFromIndex(0);

	{
		// Cross product ok to use
		xPerpVector = xForward ^ xEyeVector;
		xPerpVector.Normalise();

		Asura_Vector_3 xUp = xCamForward ^ xPerpVector;
		xUp.Normalise();
		const Asura_Vector_3 xRight = xCamForward ^ xUp;

		// Vertex 0 at top and nearest
		//      + v0 
		// + v1     + v2
		// + v3     + v4
		//      + v5

		for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
		{
			if (Asura_Xbox360_Graphics::IsTiling2())
			{
				Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
				Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
			}
			s_pxLaserEffect->Commit();

			short *pusIndex;
			hResult = Asura_Xbox360_Graphics::GetDevice()->BeginIndexedVertices( D3DPT_TRIANGLESTRIP, 0, 10, 22, D3DFMT_INDEX16, sizeof( AVP_Laser_Vertex ), reinterpret_cast< void** >( &pusIndex), reinterpret_cast< void** >( &pxVertex ) );
			if( SUCCEEDED( hResult ) )
			{
				pxVertex[0].Set(xPosition	+ xUp    * fNearRadiusInner, fNearRadiusInner, 0.0f);

				pxVertex[1].Set(xPosition	- xRight * fNearRadiusOuter, fNearRadiusOuter, 0.0f);
				pxVertex[2].Set(xPosition	- xRight * fNearRadiusInner, fNearRadiusInner, 0.0f);
				pxVertex[3].Set(xPosition	+ xRight * fNearRadiusInner, fNearRadiusInner, 0.0f);
				pxVertex[4].Set(xPosition	+ xRight * fNearRadiusOuter, fNearRadiusOuter, 0.0f);

				pxVertex[5].Set(xBeamEnd	- xRight * fFarRadiusOuter, fFarRadiusOuter, fLength);
				pxVertex[6].Set(xBeamEnd	- xRight * fFarRadiusInner, fFarRadiusInner, fLength);
				pxVertex[7].Set(xBeamEnd	+ xRight * fFarRadiusInner, fFarRadiusInner, fLength);
				pxVertex[8].Set(xBeamEnd	+ xRight * fFarRadiusOuter, fFarRadiusOuter, fLength);

				pxVertex[9].Set(xBeamEnd	- xUp    * fFarRadiusInner, fFarRadiusInner, fLength);

				static unsigned short pusIndices[] = {  
					0, 1, 2, 5, 6, 9, 
					9, 0, 
					0, 2, 3, 6, 7, 9, 
					9, 0, 
					0, 3, 4, 7, 8, 9
				};
				memcpy(pusIndex, pusIndices, sizeof(pusIndices));

				Asura_Xbox360_Graphics::GetDevice()->EndIndexedVertices();
			}
		}
	}
	s_pxLaserEffect->GetInterface()->EndPass();
	if (Asura_Xbox360_Graphics::IsTiling2())
	{
		Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
	}

}

void AVP_Platform_Laser_System::RenderEnd()
{
	s_pxLaserEffect->GetInterface()->EndTechnique();
 
	if (g_fMaxFXColour.m_fA > 0.0f)
	{
		// Resolve the current RT back out so we can read the initial colours
		s_pxLaserEffect->SetTexture( s_xSceneTextureHandle, Asura_Xbox360_Graphics::GetSceneResolveTexture() );
	    
	    s_pxLaserEffect->GetInterface()->SetVectorF(s_xColourHandle, (float*)&g_fMaxFXColour);


		Asura_Xbox360_Graphics::Resolve( Asura_Xbox360_Graphics::GetSceneResolveTexture() );
		Asura_Xbox360_Graphics::GetFullScreenVB().Bind();
		Asura_Xbox360_Graphics::GetDevice()->SetVertexDeclaration( Asura_Xbox360_Graphics::GetFullScreenVD() );

		Asura_Render::CurrentStates.RequestTranslucencyMode( ASURA_TRANSLUCENCY_GLOWING );
		Asura_Render::CurrentStates.RequestCullMode(ASURA_CULLMODE_NONE);

		for (int iTiling = 0; iTiling < Asura_Xbox360_Graphics::GetNumberOfPredicatedTilePasses(); ++iTiling)
		{
			if (Asura_Xbox360_Graphics::IsTiling2())
			{
				Asura_Xbox360_Graphics::GetDevice()->SetPredication(D3DPRED_TILE(iTiling));
				Asura_Xbox360_Graphics::CorrectVPOSForTiling(iTiling);
			}
			s_pxLaserEffect->Begin(1, 0);
			s_pxLaserEffect->Commit();
			Asura_Xbox360_Graphics::GetDevice()->DrawVertices( D3DPT_RECTLIST, 0, 1 );
			s_pxLaserEffect->End();
		}
		if (Asura_Xbox360_Graphics::IsTiling2())
		{
			Asura_Xbox360_Graphics::GetDevice()->SetPredication(0);
		}

		s_pxLaserEffect->SetTexture( s_xSceneTextureHandle, NULL );
	}
	s_pxLaserEffect->SetTexture(s_xSceneTextureHandle, NULL);
	s_pxLaserEffect->SetTexture(s_xDepthTextureHandle, NULL);

	// Clean up the sampler state as we have buggered about with it in the effect file
	for (int i = 0; i < 16; i++)
	{
		Asura_Xbox360_Graphics::GetDevice()->SetSamplerState(i, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
		Asura_Xbox360_Graphics::GetDevice()->SetSamplerState(i, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);
		Asura_Xbox360_Graphics::GetDevice()->SetSamplerState(i, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR);
	}

	Asura_Xbox360_Graphics::GetDevice()->SetRenderState_Inline( D3DRS_HIGHPRECISIONBLENDENABLE, FALSE );
	PIXEndNamedEvent();
}

void AVP_Platform_Laser_System::Shutdown()
{
	SAFE_RELEASE( s_pxLaserVertexDecl);

	Asura_Xbox360_Effect_System::Destroy( s_pxLaserEffect );
	s_pxLaserEffect = NULL;


}
