//////////////////////////////////////////////////////////////////////////////
//
// FILE : ToolsAVP_Entity_Dropship.cpp
//
// Inherits from Tools_Entity_Actor
//
// AUTHORS : Chris Wallace
//
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Entity/PhysicalObject/Actor/Vehicles/Dropship/ToolsAVP_Entity_Dropship.h"

// asura
#include "Mesh/Asura_Mesh_System.h"

// tools
#include "Entity/Tools_Entity_Data.h"
#include "Entity/Locations/OrientedNode/GamesceneNode/ToolsAVP_Node_GS2Node.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Objects/ToolsAVP_Object_Type.h"
#include "Particles/Tools_Mode_PFX.h"
#include "ToolsAVP_ChooseObjects.h"

// avp
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/AVP_Messages.h"
#include "Entity/PhysicalObject/Actor/Vehicles/Dropship/AVP_ServerVehicle_Dropship.h"

//////////////////////////////////////////////////////////////////////////////
//
// ToolsAVP_Entity_Dropship
//
//////////////////////////////////////////////////////////////////////////////

ToolsAVP_Entity_Dropship* ToolsAVP_Entity_Dropship::s_pxCurrentActive = NULL;

ToolsAVP_Entity_Dropship::ToolsAVP_Entity_Dropship(AVP_ServerVehicle_Dropship* pxActor)
: Tools_Entity_PhysicalObject(pxActor)
, PARENT( pxActor )
, m_bRefreshOverlay(false)
, m_pxDropship(pxActor)
{
	Asura_String mesh_name = "ToolsShape_DropShip";
	Mesh = Asura_Mesh_System::GetMesh(mesh_name);
	SetNotSupportGS2();
}

void ToolsAVP_Entity_Dropship::Project_Create()
{
	Create( true );
}

Tools_Object* ToolsAVP_Entity_Dropship::Create( const bool bFullInit )
{
	// Get the user to pick a blueprint to initialise from.
	Asura_Hash_ID uBlueprintTypeHashID = uAVP_BLUEPRINT_TYPE_ACTORS;
	Asura_Hash_ID uBlueprintHashID     = 0xe645f9ab;
	if ( bFullInit )
	{
		BlueprintSelect_Overlay xBPOverlay( "Choose A Blueprint", &uBlueprintHashID, &uBlueprintTypeHashID );
		if ( !xBPOverlay.DoModal() ) { return NULL; }
	}

	const Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();

	// Set up tools data and auto-create a unique name (prompt for new name if there is a clash).
	Tools_Entity_Data* pxToolsData = CreateToolsDataWithDefaultName("Dropship", "Dropship", uGuid);
	if ( !pxToolsData ) { return NULL; }

	AVP_ServerVehicle_Dropship* pxServerObject = new AVP_ServerVehicle_Dropship(uGuid);
	if ( !pxServerObject )
	{
		delete pxToolsData;
		return NULL;
	}

	// Set the tools data before we create the tools object - otherwise it will have already been created!
	pxServerObject->ToolsData = pxToolsData;
	if ( bFullInit )
	{
		pxServerObject->SetPosition( pxToolsData->Position );
		pxServerObject->SetBlueprintTypeHashID( uBlueprintTypeHashID );
		pxServerObject->SetBlueprintHashID( uBlueprintHashID );
		pxServerObject->MakeAI();
		pxServerObject->CreateAndInitialiseBrain();
		pxServerObject->InitFromBlueprint();
	}

	ToolsAVP_Entity_Dropship* pxToolsObject = new ToolsAVP_Entity_Dropship( pxServerObject );
	if ( !pxToolsObject )
	{
		delete pxToolsObject;
		delete pxServerObject;
		return NULL;
	}
	Tools_Object_System::SelectObject(pxToolsObject);

	Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, "Created new Dropship." );
	Tools_SaveAsura::SetEntityModified();

	return pxToolsObject;
}

Tools_Object* ToolsAVP_Entity_Dropship::Create( Asura_Entity* pxEntity )
{
    return new ToolsAVP_Entity_Dropship( AVP_ServerVehicle_Dropship::FindEntity(pxEntity->GetGuid()) );
}

void ToolsAVP_Entity_Dropship::AddOverlayControls( Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset )
{
	PARENT::AddOverlayControls( pxOverlay, xOffset );

	Asura_Dialog* pxDialogue = new Asura_Function_Dialog( "Set GS2StartPoint", xOffset, SelectStartGS2Node );
	pxOverlay->AddDialog( pxDialogue );
	xOffset.x += 18.0f;

	pxDialogue = new Asura_Function_Dialog( "Clear GS2StartPoint", xOffset, ClearStartGS2Node );
	pxOverlay->AddDialog( pxDialogue );
	xOffset.x -= 18.0f;
	xOffset.y += Asura_Dialog::LargeSpacing();

	if (m_pxDropship->m_uGS2StartNode == ASURA_GUID_UNREGISTERED)
	{
		pxDialogue = new Asura_Dialog("No Start Point", xOffset );
	}
	else
	{
		ToolsAVP_Node_GS2Node* pxNode = ToolsAVP_Node_GS2Node::FindEntity(m_pxDropship->m_uGS2StartNode);
		Asura_Assert(pxNode, "Selected Node Does Not Exist");
		if (pxNode)
		{
			pxDialogue = new Asura_Dialog(pxNode->GetName(), xOffset );
		}
		else
		{
			pxDialogue = new Asura_Dialog("No Valid Start Point", xOffset );

		}
	}
	pxOverlay->AddDialog( pxDialogue );

}


void ToolsAVP_Entity_Dropship::OnSelection()
{
	Validate();
	m_bRefreshOverlay = false;
	PARENT::OnSelection();
	Asura_Assert(s_pxCurrentActive == NULL, "Selection without Deslection");
	s_pxCurrentActive = this;

}

void ToolsAVP_Entity_Dropship::OnDeselection()
{
	Validate();

	PARENT::OnDeselection();
	s_pxCurrentActive = NULL;
}

bool ToolsAVP_Entity_Dropship::HandlesThisMessage( int iMessageID )
{
	if ( iMessageID == AVP_MSGID_SET_DROPSHIP_ATTACK_MODE )
	{
		return true;
	}
	else
	{
		return PARENT::HandlesThisMessage( iMessageID );
	}
}


void ToolsAVP_Entity_Dropship::UpdateWhileSelected()
{
	PARENT::UpdateWhileSelected();

	if ( m_bRefreshOverlay )
	{
		DeleteOverlay();
		CreateOverlay();
		m_bRefreshOverlay = false;
	}
}


// call back functions

void ToolsAVP_Entity_Dropship::SelectStartGS2Node()
{
	ToolsAVP_ChooseObject_ByType* pxChooseObject = new ToolsAVP_ChooseObject_ByType( TOT_Entity_AVPGS2Node, SetStartGS2Node );
	Tools_Object_System::ChooseObject( "Select A GS2Node", pxChooseObject );
}


void ToolsAVP_Entity_Dropship::SetStartGS2Node(Asura_Guid uGuid)
{
	Asura_Assert(s_pxCurrentActive, "Current Dropship not set");
	if (uGuid != ASURA_GUID_UNREGISTERED && s_pxCurrentActive)
	{
		s_pxCurrentActive->m_pxDropship->m_uGS2StartNode = uGuid;
		s_pxCurrentActive->RequestRefreshOverlay();
		Tools_SaveAsura::SetEntityModified();
	}
}

void ToolsAVP_Entity_Dropship::ClearStartGS2Node()
{
	Asura_Assert(s_pxCurrentActive, "Current Dropship not set");
	if (s_pxCurrentActive)
	{
		s_pxCurrentActive->m_pxDropship->m_uGS2StartNode = ASURA_GUID_UNREGISTERED;
		s_pxCurrentActive->RequestRefreshOverlay();
		Tools_SaveAsura::SetEntityModified();
	}
}



void ToolsAVP_Entity_Dropship::Validate()
{
	if (m_pxDropship->m_uGS2StartNode != ASURA_GUID_UNREGISTERED)
	{
		const ToolsAVP_Node_GS2Node* pxNode = ToolsAVP_Node_GS2Node::FindEntity(m_pxDropship->m_uGS2StartNode);
		Asura_Assert(pxNode, "Selected Node Does Not Exist");
	}
}