//////////////////////////////////////////////////////////////////////////////
//
// Axon_BehaviourParams.cpp
//
// Copyright (C) 2008 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_AXON

// header
#include "Behaviour/Axon_BehaviourParams.h"

// asura
#include "ChunkSystem/Asura_Chunk_Stream.h"

// axon
#include "Behaviour/BehaviourParamsTypes/Axon_BehaviourParams_GS2AttractorEnRoute.h"
#include "Behaviour/BehaviourParamsTypes/Axon_BehaviourParams_GS2Node.h"
#include "Behaviour/BehaviourParamsTypes/Axon_BehaviourParams_GS2Spline.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Types.h"

// tools
#ifdef ASURA_TOOLS
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#endif

//////////////////////////////////////////////////////////////////////////////
//
// Axon_BehaviourParams
//
//////////////////////////////////////////////////////////////////////////////

Axon_BehaviourParams::Axon_BehaviourParams( const Axon_BehaviourParams& xParams )
: m_uParamsType( xParams.m_uParamsType )
, m_uBehaviourType( xParams.m_uBehaviourType )
, m_pxChild( NULL )
, m_pxSibling( NULL )
#ifdef ASURA_TOOLS
, m_pxToolsNode( xParams.m_pxToolsNode )
#endif
{
	// We don't attempt child/sibling duplication, this must happen elsewhere.
}

Axon_BehaviourParams::Axon_BehaviourParams( const u_int uBehaviourType )
: m_uParamsType( AXON_BP_TYPE_BLANK )
, m_uBehaviourType( uBehaviourType )
, m_pxChild( NULL )
, m_pxSibling( NULL )
#ifdef ASURA_TOOLS
, m_pxToolsNode( NULL )
#endif
{
}

Axon_BehaviourParams::Axon_BehaviourParams( const u_int uParamsType, const u_int uBehaviourType )
: m_uParamsType( uParamsType )
, m_uBehaviourType( uBehaviourType )
, m_pxChild( NULL )
, m_pxSibling( NULL )
#ifdef ASURA_TOOLS
, m_pxToolsNode( NULL )
#endif
{
}

Axon_BehaviourParams::~Axon_BehaviourParams()
{
	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		UnlinkChild( pxChild );
		delete pxChild;
		pxChild = GetChild();
	}
}

bool Axon_BehaviourParams::DoHandleMessage( const Asura_Message* pxMessage )
{
	bool bReturnValue = HandleMessage( pxMessage );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		bReturnValue |= pxChild->DoHandleMessage( pxMessage );
		pxChild = pxChild->GetSibling();
	}

	return bReturnValue;
}

void Axon_BehaviourParams::CreateTree( const int iBehaviourSet, Axon_BehaviourParams*& pxBehaviourParams )
{
	Asura_Assert( ( pxBehaviourParams == NULL ), "Axon_BehaviourParams::Create - old behaviour params are not free" );

	if ( Project_CreateTree( iBehaviourSet, pxBehaviourParams ) )
	{
		return;
	}

	switch ( iBehaviourSet )
	{
		case AXON_BS_ID_CORE_NODE:
		{
			pxBehaviourParams = new Axon_BehaviourParams_GS2Node( AXON_BEHAVIOUR_TYPE_GS2_NODE );
			break;
		}

		case AXON_BS_ID_CORE_SPLINE:
		{
			pxBehaviourParams = new Axon_BehaviourParams_GS2Spline( AXON_BEHAVIOUR_TYPE_GS2_SPLINE );
			break;
		}

		case AXON_BS_ID_CORE_ATTRACTOR_EN_ROUTE:
		{
			pxBehaviourParams = new Axon_BehaviourParams_GS2AttractorEnRoute( AXON_BEHAVIOUR_TYPE_GS2_ATTRACTOR_EN_ROUTE );
			break;
		}

		default:
			break;
	}
}

bool Axon_BehaviourParams::ConvertTree( const int iBehaviourSet, const Axon_BehaviourParams* pxOldBehaviourParams, Axon_BehaviourParams*& pxNewBehaviourParams )
{
	if ( Project_ConvertTree( iBehaviourSet, pxOldBehaviourParams, pxNewBehaviourParams ) )
	{
		return true;
	}

	if ( iBehaviourSet == AXON_BS_ID_CORE_ATTRACTOR_EN_ROUTE )
	{
		// Convert from the old parameter-less en-route to the new parameterised one.
		if ( pxOldBehaviourParams->GetParamsType() == AXON_BP_TYPE_BLANK )
		{
			pxNewBehaviourParams = new Axon_BehaviourParams_GS2AttractorEnRoute( AXON_BEHAVIOUR_TYPE_GS2_ATTRACTOR_EN_ROUTE );
			return true;
		}
	}

	return false;
}

void Axon_BehaviourParams::CreateTreeFromChunkStream( Asura_Chunk_Stream& xStream, Axon_BehaviourParams*& pxBehaviourParams )
{
	Asura_Assert( pxBehaviourParams == NULL, "Axon_BehaviourParams::CreateTreeFromChunkStream - should not already have parameters" );

	int iVersion;
	xStream >> iVersion;

	if ( iVersion == CURRENT_MASTER_CHUNK_VERSION )
	{
		u_int uParamType;
		xStream >> uParamType;
		u_int uBehaviourType;
		xStream >> uBehaviourType;

		pxBehaviourParams = Axon_BehaviourParams::Create( uParamType, uBehaviourType );
		if ( pxBehaviourParams )
		{
			pxBehaviourParams->ReadFromChunkStream( xStream );
		}
	}
	else
	{
		Asura_Assertf( false, ( "Axon_BehaviourParams::CreateTreeFromChunkStream - unhandled master chunk version (%d)", iVersion ) );
	}
}

void Axon_BehaviourParams::WriteTreeToChunkStream( const Axon_BehaviourParams* pxBehaviourParams, Asura_Chunk_Stream& xStream )
{
	xStream << CURRENT_MASTER_CHUNK_VERSION;

	if ( pxBehaviourParams )
	{
		xStream << pxBehaviourParams->GetParamsType();
		xStream << pxBehaviourParams->GetBehaviourType();

		pxBehaviourParams->WriteToChunkStream( xStream );
	}
	else
	{
		xStream << AXON_BP_TYPE_NONE;
		xStream << AXON_BEHAVIOUR_TYPE_UNSET;
	}
}

int Axon_BehaviourParams::GetNumChildren() const
{
	int iChildren = 0;

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		++iChildren;
		pxChild = pxChild->GetSibling();
	}

	return iChildren;
}

Axon_BehaviourParams* Axon_BehaviourParams::Create( const u_int uParamType, const u_int uBehaviourType )
{
	Axon_BehaviourParams* pxBehaviourParams = Project_Create( uParamType, uBehaviourType );
	if ( pxBehaviourParams )
	{
		return pxBehaviourParams;
	}

	switch ( uParamType )
	{
		case AXON_BP_TYPE_BLANK:
		{
			pxBehaviourParams = new Axon_BehaviourParams( uBehaviourType );
			break;
		}

		case AXON_BP_TYPE_GS2_NODE:
		{
			pxBehaviourParams = new Axon_BehaviourParams_GS2Node( uBehaviourType );
			break;
		}

		case AXON_BP_TYPE_GS2_SPLINE:
		{
			pxBehaviourParams = new Axon_BehaviourParams_GS2Spline( uBehaviourType );
			break;
		}

		case AXON_BP_TYPE_GS2_ATTRACTOR_ENROUTE:
		{
			pxBehaviourParams = new Axon_BehaviourParams_GS2AttractorEnRoute( uBehaviourType );
			break;
		}

		default:
			break;
	}

	return pxBehaviourParams;
}

void Axon_BehaviourParams::LinkChild( Axon_BehaviourParams* pxBehaviourParams )
{
	if ( m_pxChild )
	{
		m_pxChild->LinkSibling( pxBehaviourParams );
	}
	else
	{
		m_pxChild = pxBehaviourParams;
	}
}

bool Axon_BehaviourParams::UnlinkChild( Axon_BehaviourParams* pxBehaviourParams )
{
	if ( m_pxChild )
	{
		if ( m_pxChild == pxBehaviourParams )
		{
			m_pxChild = m_pxChild->GetSibling();
			pxBehaviourParams->m_pxSibling = NULL;
			return true;
		}
		else
		{
			return m_pxChild->UnlinkSibling( pxBehaviourParams );
		}
	}

	return false;
}

void Axon_BehaviourParams::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	int iVersion;
	xStream >> iVersion;

	// We'd do our parent now if we had one.

	if ( iVersion == CURRENT_CHUNK_VERSION )
	{
		int iNumChildren;
		xStream >> iNumChildren;

		while ( iNumChildren )
		{
			int iChildType;
			xStream >> iChildType;
			int iChildBehaviourType;
			xStream >> iChildBehaviourType;

			// Create and link a params class of the correct type.
			Axon_BehaviourParams* pxChild = Axon_BehaviourParams::Create( iChildType, iChildBehaviourType );
			pxChild->ReadFromChunkStream( xStream );
			LinkChild( pxChild );
		}
	}
	else
	{
		Asura_Assertf( false, ( "Axon_BehaviourParams::ReadFromChunkStream - unhandled chunk version (%d)", iVersion ) );
	}
}

void Axon_BehaviourParams::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << CURRENT_CHUNK_VERSION;

	// We'd do our parent now if we had one.

	// No need to save the type, someone else needs to do that for us.
	const int iNumChildren = GetNumChildren();
	xStream << iNumChildren;

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		xStream << pxChild->GetParamsType();
		pxChild->WriteToChunkStream( xStream );
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::LinkSibling( Axon_BehaviourParams* pxBehaviourParams )
{
	if ( m_pxSibling )
	{
		m_pxSibling->LinkSibling( pxBehaviourParams );
	}
	else
	{
		m_pxSibling = pxBehaviourParams;
	}
}

bool Axon_BehaviourParams::UnlinkSibling( Axon_BehaviourParams* pxBehaviourParams )
{
	if ( m_pxSibling )
	{
		if ( m_pxSibling == pxBehaviourParams )
		{
			m_pxSibling = m_pxSibling->GetSibling();
			pxBehaviourParams->m_pxSibling = NULL;
			return true;
		}
		else
		{
			return m_pxSibling->UnlinkSibling( pxBehaviourParams );
		}
	}

	return false;
}

#ifdef ASURA_TOOLS
void Axon_BehaviourParams::DoSetToolsNode( Tools_Entity_Node* pxToolsNode )
{
	SetToolsNode( pxToolsNode );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoSetToolsNode( pxToolsNode );
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoAddOverlayControls( Asura_Tabs_Dialog* pxTabsDialog, Asura_Vector_2& xOffset )
{
	AddOverlayControls( pxTabsDialog, xOffset );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoAddOverlayControls( pxTabsDialog, xOffset );
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoRender()
{
	Render();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoRender();
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoUpdate()
{
	Update();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoUpdate();
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoUpdateWhileSelected()
{
	UpdateWhileSelected();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoUpdateWhileSelected();
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoStartPreview()
{
	StartPreview();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->StartPreview();
		pxChild = pxChild->GetSibling();
	}
}

bool Axon_BehaviourParams::DoUpdatePreview()
{
	bool bOk = UpdatePreview();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		bOk |= pxChild->UpdatePreview();
		pxChild = pxChild->GetSibling();
	}

	return bOk;
}

void Axon_BehaviourParams::DoStopPreview()
{
	StopPreview();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->StopPreview();
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoAddFilesToExportList( List< Asura_String >& xNameList ) const
{
	AddFilesToExportList( xNameList );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoAddFilesToExportList( xNameList );
		pxChild = pxChild->GetSibling();
	}
}

void Axon_BehaviourParams::DoAddFilesToExportListFromMessage( u_short usMessageID, void* pData, List< Asura_String >& xNameList ) const
{
	AddFilesToExportListFromMessage( usMessageID, pData, xNameList );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoAddFilesToExportListFromMessage( usMessageID, pData, xNameList );
		pxChild = pxChild->GetSibling();
	}
}

bool Axon_BehaviourParams::DoHandlesThisMessage( const int iMessageID )
{
	if ( HandlesThisMessage( iMessageID ) )
	{
		return true;
	}

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		if ( pxChild->DoHandlesThisMessage( iMessageID ) )
		{
			return true;
		}
		pxChild = pxChild->GetSibling();
	}

	return false;
}

bool Axon_BehaviourParams::DoOwnsThisMessageBlock( const u_short usMessageBlockID )
{
	if ( OwnsThisMessageBlock( usMessageBlockID ) )
	{
		return true;
	}

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		if ( pxChild->DoOwnsThisMessageBlock( usMessageBlockID ) )
		{
			return true;
		}
		pxChild = pxChild->GetSibling();
	}

	return false;
}

bool Axon_BehaviourParams::DoGetParameter( const Asura_Hash_ID uParamID, Tools_Parameter* pxParam ) const
{
	if ( GetParameter( uParamID, pxParam ) )
	{
		return true;
	}

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		if ( pxChild->DoGetParameter( uParamID, pxParam ) )
		{
			return true;
		}
		pxChild = pxChild->GetSibling();
	}

	return false;
}

bool Axon_BehaviourParams::DoSetParameter( const Asura_Hash_ID uParamID, const Tools_Parameter* pxParam )
{
	if ( SetParameter( uParamID, pxParam ) )
	{
		return true;
	}

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		if ( pxChild->DoSetParameter( uParamID, pxParam ) )
		{
			return true;
		}
		pxChild = pxChild->GetSibling();
	}

	return false;
}

void Axon_BehaviourParams::DoApplyChangesForTemplateInstance( Tools_EntityTemplateInstance* pxInstance )
{
	ApplyChangesForTemplateInstance( pxInstance );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		pxChild->DoApplyChangesForTemplateInstance( pxInstance );
		pxChild = pxChild->GetSibling();
	}
}

bool Axon_BehaviourParams::DoVerify( const bool bAutomatic )
{
	bool bOk = Verify( bAutomatic );

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		bOk &= pxChild->DoVerify( bAutomatic );
		pxChild = pxChild->GetSibling();
	}

	return bOk;
}

Axon_BehaviourParams* Axon_BehaviourParams::DoClone() const
{
	Axon_BehaviourParams* pxNewParams = Clone();

	Axon_BehaviourParams* pxChild = GetChild();
	while ( pxChild )
	{
		Axon_BehaviourParams* pxNewChild = pxChild->DoClone();
		pxNewParams->LinkChild( pxNewChild );
		pxChild = pxChild->GetSibling();
	}

	return pxNewParams;
}

Axon_BehaviourParams* Axon_BehaviourParams::Clone() const
{
	return new Axon_BehaviourParams( *this );
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // ASURA_USE_AXON
