//////////////////////////////////////////////////////////////////////////////
//
// Axon_Behaviour_FollowActor.cpp
//
// Copyright (C) 2009 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_NEW_NAV_STUFF

// header
#include "Behaviour/BehaviourTypes/Axon_Behaviour_FollowActor.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#endif

// axon
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Movement.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Types.h"
#include "Core/Axon_Brain.h"

//////////////////////////////////////////////////////////////////////////////
//
// InDirectionToIdealFunctor
//
//////////////////////////////////////////////////////////////////////////////

class InDirectionToIdealFunctor : public Asura_Navigation::InDirectionFunctor
{
public:
	InDirectionToIdealFunctor() : m_fIdealDistanceSq( 0.0f ), m_uBestWPID( ASURA_INVALID_WAYPOINT_ID ) {}

	void SetIdealDistance( const float fIdealDistance ) { m_fIdealDistanceSq = fIdealDistance * fIdealDistance; }

	u_int GetBestWPID() const { return m_uBestWPID; }

protected:
	virtual bool VerifyInputData()
	{
		return ( PARENT::VerifyInputData() && m_fIdealDistanceSq > 0.0f );
	}
	virtual bool VisitNode( const u_int uWPID, const Asura_AI_NavVertex& xWP )
	{
		Asura_Vector_3 xStartToWP   = xWP.GetPosition() - GetStartPosition();
		const float    fStartToWPSq = xStartToWP.MagnitudeSquared();

		// Anything over the ideal distance will do, as we can't be sure we'll hit a waypoint near enough.
		if ( fStartToWPSq > m_fIdealDistanceSq )
		{
			m_uBestWPID = uWPID;
			return true;
		}
		return false;
	}

private:
	float m_fIdealDistanceSq;
	u_int m_uBestWPID;

	typedef Asura_Navigation::InDirectionFunctor PARENT;
};

//////////////////////////////////////////////////////////////////////////////
//
// Axon_Behaviour_FollowActor
//
//////////////////////////////////////////////////////////////////////////////

static const float s_fStandStillPeriodMin = 5.0f;
static const float s_fStandStillPeriodMax = 15.0f;

Axon_Behaviour_FollowActor::Axon_Behaviour_FollowActor( Axon_Brain* pxBrain )
: PARENT( AXON_BEHAVIOUR_TYPE_FOLLOW_ACTOR, pxBrain )
, m_uTargetActorGuid( ASURA_GUID_UNREGISTERED )
, m_fInnerRadiusSq( 0.0f )
, m_fOuterRadiusSq( 0.0f )
, m_fIdealDistance( 0.0f )
, m_fIdealDelta( 0.0f )
, m_iState( WAITING_FOR_TARGET_ACTOR_GUID )
, m_fTimer( 0.0f )
, m_pxMovementChild( NULL )
, m_pxTargetPhyObj( NULL )
, m_eRange( RANGE_IDEAL )
, m_xPathfindBasePos( Asura_Maths::ZeroVector3 )
{
	SetRadii( 5.0f, 15.0f );
}

void Axon_Behaviour_FollowActor::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;

	PARENT::ReadFromChunkStream( xStream );

	if ( uVersion == uCURRENT_CHUNK_VERSION )
	{
		xStream >> m_fInnerRadiusSq;
		xStream >> m_fOuterRadiusSq;
		xStream >> m_fIdealDistance;
		xStream >> m_fTimer;
		xStream >> m_uTargetActorGuid;
	}
	else
	{
		Asura_Assertf( false, ( "Axon_Behaviour_FollowActor::ReadFromChunkStream - unhanded chunk version (%u)", uVersion ) );
	}
}

void Axon_Behaviour_FollowActor::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << uCURRENT_CHUNK_VERSION;

	PARENT::WriteToChunkStream( xStream );

	xStream << m_fInnerRadiusSq;
	xStream << m_fOuterRadiusSq;
	xStream << m_fIdealDistance;
	xStream << m_fTimer;
	xStream << m_uTargetActorGuid;
}

void Axon_Behaviour_FollowActor::SetTargetActorGuid( const Asura_Guid uTargetActorGuid )
{
	m_uTargetActorGuid = uTargetActorGuid;
	OnNewTargetActor();
}

void Axon_Behaviour_FollowActor::SetRadii( float fInnerRadius, float fOuterRadius )
{
	Asura_Assert( fInnerRadius < fOuterRadius, "Axon_Behaviour_FollowActor::SetRadii - bad radii values" );
	m_fInnerRadiusSq = fInnerRadius * fInnerRadius;
	m_fOuterRadiusSq = fOuterRadius * fOuterRadius;

	const float fRange = fOuterRadius - fInnerRadius;
	m_fIdealDistance = fInnerRadius + fRange * Asura_Random::GetFloat( 0.3f, 0.7f );
	m_fIdealDelta    = fRange * 0.1f;
}

bool Axon_Behaviour_FollowActor::Activate()
{
	bool bOk = PARENT::Activate();
	m_pxMovementChild = static_cast< Axon_Behaviour_Movement* >( CreateChildWithType( AXON_BEHAVIOUR_TYPE_MOVEMENT ) );
	bOk &= ( m_pxMovementChild != NULL );
	return bOk;
}

void Axon_Behaviour_FollowActor::Deactivate()
{
	DestroyChildWithType( AXON_BEHAVIOUR_TYPE_MOVEMENT );
	PARENT::Deactivate();
}

void Axon_Behaviour_FollowActor::PostLoadInitialise()
{
	PARENT::PostLoadInitialise();

	if ( !m_pxMovementChild )
	{
		m_pxMovementChild = static_cast< Axon_Behaviour_Movement* >( CreateChildWithType( AXON_BEHAVIOUR_TYPE_MOVEMENT ) );
	}
	SwitchToState( WAITING );
}

void Axon_Behaviour_FollowActor::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	if ( !m_pxMovementChild )
	{
		Asura_Assert( false, "Axon_Behaviour_FollowActor::PerFrameUpdate - where did our movement child go?" );
		//## Not decided on an full error handling scheme for these core behaviours yet, but I don't think they should ever fail
		//## if they are continuous tasks like this.
		return;
	}

	Range eRange = RANGE_IDEAL;
	if ( !UpdatePhyObjPointer() )
	{
		SwitchToState( WAITING_FOR_TARGET_ACTOR_GUID );
	}
	else
	{
		eRange = UpdateRange();
	}

	switch ( m_iState )
	{
		case WAITING_FOR_ERROR_TIMEOUT:
		{
			m_fTimer -= fTimeslice;
			if ( m_fTimer <= 0.0f )
			{
				SwitchToState( WAITING );
			}

			break;
		}

		case WAITING:
		{
			m_fTimer -= fTimeslice;
			if ( eRange == RANGE_TOO_NEAR )
			{
				SwitchToState( MOVING_AWAY );
			}
			else if ( eRange == RANGE_TOO_FAR )
			{
				SwitchToState( MOVING_TOWARDS );
			}
			else if ( m_fTimer <= 0.0f )
			{
				SwitchToState( MOVING_AROUND );
			}
			else
			{
				Asura_Vector_3 xLookAtPosition;
				GetTargetLookAtPosition( xLookAtPosition );
				m_pxMovementChild->LookAtPosition( xLookAtPosition );
			}

			break;
		}

		case MOVING_TOWARDS:
		{
			m_fTimer += fTimeslice;
			if ( eRange == RANGE_IDEAL || m_pxMovementChild->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_AT_DESTINATION )
			{
				SwitchToState( WAITING );
			}
			else if ( eRange == RANGE_TOO_NEAR )
			{
				SwitchToState( MOVING_AWAY );
			}
			else if ( m_fTimer > 1.0f && !m_pxTargetPhyObj->GetPosition().ApproximatelyEquals( m_xPathfindBasePos, 1.0f ) )
			{
				SwitchToState( MOVING_TOWARDS );
			}
			else if ( m_pxMovementChild->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_IN_ERROR )
			{
				SwitchToState( WAITING_FOR_ERROR_TIMEOUT );
			}
			else
			{
				const float fPathLength = m_pxMovementChild->GetPath().GetPathRemainingLength();
				const float fSpeed = Asura_Maths::ClampToRange( fPathLength * 0.1f, 0.25f, 1.0f );
				m_pxMovementChild->SetSpeed( fSpeed );
			}

			break;
		}

		case MOVING_AWAY:
		{
			m_fTimer += fTimeslice;
			if ( eRange == RANGE_IDEAL || m_pxMovementChild->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_AT_DESTINATION )
			{
				SwitchToState( WAITING );
			}
			else if ( eRange == RANGE_TOO_FAR )
			{
				SwitchToState( MOVING_TOWARDS );
			}
			else if ( m_fTimer > 1.0f && !m_pxTargetPhyObj->GetPosition().ApproximatelyEquals( m_xPathfindBasePos, 1.0f ) )
			{
				SwitchToState( MOVING_AWAY );
			}
			else if ( m_pxMovementChild->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_IN_ERROR )
			{
				SwitchToState( WAITING_FOR_ERROR_TIMEOUT );
			}
			else
			{
				const float fPathLength = m_pxMovementChild->GetPath().GetPathRemainingLength();
				const float fSpeed = Asura_Maths::ClampToRange( fPathLength * 0.1f, 0.25f, 1.0f );
				m_pxMovementChild->SetSpeed( fSpeed );
			}

			break;
		}

		case MOVING_AROUND:
		{
			if ( eRange == RANGE_TOO_NEAR )
			{
				SwitchToState( MOVING_AWAY );
			}
			else if ( eRange == RANGE_TOO_FAR )
			{
				SwitchToState( MOVING_TOWARDS );
			}
			else if ( m_pxMovementChild->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_AT_DESTINATION )
			{
				SwitchToState( WAITING );
			}
			else if ( m_pxMovementChild->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_IN_ERROR )
			{
				SwitchToState( WAITING_FOR_ERROR_TIMEOUT );
			}

			break;
		}

		default:
			m_uTargetActorGuid = ASURA_GUID_UNREGISTERED;
			// fallthrough
		case WAITING_FOR_TARGET_ACTOR_GUID:
			break;
	}
}

void Axon_Behaviour_FollowActor::OnNewTargetActor()
{
	SwitchToState( WAITING );
}

void Axon_Behaviour_FollowActor::SwitchToState( int iNextState )
{
	switch ( m_iState )
	{
		case MOVING_TOWARDS:
		case MOVING_AWAY:
		case MOVING_AROUND:
		{
			m_pxMovementChild->StopPathTraversal();
			break;
		}

		case WAITING_FOR_TARGET_ACTOR_GUID:
		case WAITING_FOR_ERROR_TIMEOUT:
		case WAITING:
		default:
			break;
	}

	m_iState = iNextState;
	m_fTimer = 0.0f;
	switch ( m_iState )
	{
		case MOVING_TOWARDS:
		{
			m_xPathfindBasePos = m_pxTargetPhyObj->GetPosition();
			if ( m_pxMovementChild->FindPathTo( m_xPathfindBasePos ) )
			{
				const float fPathLength = m_pxMovementChild->GetPath().GetPathRemainingLength();
				const float fSpeed = Asura_Maths::ClampToRange( fPathLength * 0.1f, 0.25f, 1.0f );
				m_pxMovementChild->SetSpeed( fSpeed );
				m_pxMovementChild->StartPathTraversal();
			}
			else
			{
				SwitchToState( WAITING_FOR_ERROR_TIMEOUT );
			}
			break;
		}

		case MOVING_AWAY:
		{
			m_xPathfindBasePos = m_pxTargetPhyObj->GetPosition();
			if ( m_pxMovementChild->FindPathAwayFrom( m_xPathfindBasePos, m_fIdealDistance ) )
			{
				const float fPathLength = m_pxMovementChild->GetPath().GetPathRemainingLength();
				const float fSpeed = Asura_Maths::ClampToRange( fPathLength * 0.1f, 0.25f, 1.0f );
				m_pxMovementChild->SetSpeed( fSpeed );
				m_pxMovementChild->StartPathTraversal();
			}
			else
			{
				SwitchToState( WAITING_FOR_ERROR_TIMEOUT );
			}
			break;
		}

		case MOVING_AROUND:
		{
			// Find a position at the correct distance from the target in a random direction.
			Asura_Vector_3 xRandomDirection( Asura_Random::GetFloat( -1.0f, 1.0f ), 0.0f, Asura_Random::GetFloat( -1.0f, 1.0f ) );
			xRandomDirection.Normalise();
			InDirectionToIdealFunctor xIDTIFunctor;
			xIDTIFunctor.SetInputBuffer( Asura_AI_Navigation_System::GetWaypoints(), Asura_AI_Navigation_System::GetNumberOfWaypoints() );
			xIDTIFunctor.SetStartPosition( m_pxTargetPhyObj->GetPosition(), m_pxMovementChild->GetTestParams() );
			xIDTIFunctor.SetRange( m_fIdealDistance * 1.5f );
			xIDTIFunctor.SetDirection( xRandomDirection );
			xIDTIFunctor.SetIdealDistance( m_fIdealDistance );
			xIDTIFunctor.Search();
			if ( xIDTIFunctor.GetBestWPID() == ASURA_INVALID_WAYPOINT_ID )
			{
				// Hmph. Just have to wait a bit longer.
				SwitchToState( WAITING );
			}
			else
			{
				const Asura_AI_NavVertex* pxToWP = Asura_AI_Navigation_System::GetWaypoint( xIDTIFunctor.GetBestWPID() );

				// Find a route around the target actor that remains at roughly the same range.
				Asura_Navigation::FromToWithRangePathFindFunctor xFTWRFunctor;
				xFTWRFunctor.SetTestParams( m_pxMovementChild->GetTestParams() );
				xFTWRFunctor.SetShouldRefinePath( false );
				xFTWRFunctor.SetFromToPos( m_pxMovementChild->GetOwnerPosition(), pxToWP->GetPosition(), m_pxMovementChild->GetTestParams() );
				xFTWRFunctor.SetBasePos( m_pxTargetPhyObj->GetPosition() );
				xFTWRFunctor.SetIdealRange( m_fIdealDistance );
				if ( m_pxMovementChild->FindPath( xFTWRFunctor ) )
				{
					// Just amble over.
					m_pxMovementChild->SetSpeed( 0.25f );
					m_pxMovementChild->StartPathTraversal();
				}
				else
				{
					SwitchToState( WAITING );
				}
			}

			break;
		}

		case WAITING_FOR_ERROR_TIMEOUT:
		{
			m_fTimer = 1.0f;
			break;
		}

		case WAITING:
		{
			m_fTimer = Asura_Random::GetFloat( s_fStandStillPeriodMin, s_fStandStillPeriodMax );
			break;
		}

		case WAITING_FOR_TARGET_ACTOR_GUID:
		default:
			break;
	}
}

void Axon_Behaviour_FollowActor::GetTargetLookAtPosition( Asura_Vector_3& xPosition ) const
{
	const Asura_ServerEntity_PhysicalObject* pxTargetPhyObj = Asura_ServerEntity_PhysicalObject::FindEntity( m_uTargetActorGuid );
	if ( pxTargetPhyObj )
	{
		xPosition = pxTargetPhyObj->GetPosition();
		xPosition.y -= 1.5f;
	}
}

bool Axon_Behaviour_FollowActor::UpdatePhyObjPointer()
{
	m_pxTargetPhyObj = Asura_ServerEntity_PhysicalObject::FindEntity( m_uTargetActorGuid );
	return ( m_pxTargetPhyObj != NULL );
}

Axon_Behaviour_FollowActor::Range Axon_Behaviour_FollowActor::UpdateRange()
{
	const float fDistanceSq = ( m_pxTargetPhyObj->GetPosition() - m_pxMovementChild->GetOwnerPosition() ).MagnitudeSquared();

	const float fIdealInnerDistanceSq = ( m_fIdealDistance - m_fIdealDelta ) * ( m_fIdealDistance - m_fIdealDelta );
	const float fIdealOuterDistanceSq = ( m_fIdealDistance + m_fIdealDelta ) * ( m_fIdealDistance + m_fIdealDelta );

	if ( fDistanceSq < m_fInnerRadiusSq )
	{
		m_eRange = RANGE_TOO_NEAR;
	}
	else if ( fDistanceSq < fIdealInnerDistanceSq )
	{
		m_eRange = RANGE_NEAR;
	}
	else if ( fDistanceSq < fIdealOuterDistanceSq )
	{
		m_eRange = RANGE_IDEAL;
	}
	else if ( fDistanceSq < m_fOuterRadiusSq )
	{
		m_eRange = RANGE_FAR;
	}
	else
	{
		m_eRange = RANGE_TOO_FAR;
	}
	return m_eRange;
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
	// N.B. Keep in updated with behaviour states.
	"FollowActor - WAITING_FOR_TARGET_ACTOR_GUID",
	"FollowActor - WAITING_FOR_ERROR_TIMEOUT",
	"FollowActor - WAITING",
	"FollowActor - MOVING_TOWARDS",
	"FollowActor - MOVING_AWAY",
	"FollowActor - MOVING_AROUND",
};

void Axon_Behaviour_FollowActor::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();

	if ( GetBrain()->DebugTextPosOnScreen() )
	{
		Asura_Font_System::Print( GetBrain()->GetDebugTextPos().x, GetBrain()->GetDebugTextPos().y, 0.5f, 0xffffffff, 
			ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, s_aszStateStrings[m_iState] );
		GetBrain()->IncrementDebugTextPos();
	}
}
#endif
//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // #ifdef ASURA_USE_NEW_NAV_STUFF
