//////////////////////////////////////////////////////////////////////////////
//
// Axon_Behaviour_GS2Node.cpp
//
// Copyright (C) 2008 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_AXON

// header
#include "Behaviour/BehaviourTypes/Axon_Behaviour_GS2Node.h"

// asura
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Entity/Locations/OrientedNode/GamesceneNode/Asura_ServerNode_GS2Node.h"
#include "Entity/Locations/Spline/GamesceneSpline/Asura_ServerNode_GS2Spline.h"
#include "Network/Asura_Message.h"

// axon
#include "Behaviour/BehaviourParamsTypes/Axon_BehaviourParams_GS2Node.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_GS2Actor.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Types.h"
#include "Core/Axon_Brain.h"
#include "Core/Axon_Messages.h"

//////////////////////////////////////////////////////////////////////////////
//
// Axon_Behaviour_GS2Node
//
//////////////////////////////////////////////////////////////////////////////

Axon_Behaviour_GS2Node::Axon_Behaviour_GS2Node( Axon_Brain* pxBrain )
: PARENT( AXON_BEHAVIOUR_TYPE_GS2_NODE, pxBrain )
, m_iState( UNSET )
, m_bStopHere( false )
, m_uAnimSet( uAXON_ANIMATION_SEQUENCE_ID_NONE )
, m_iExitConditions( AXON_GS2_NEC_NONE )
, m_fExitWaitTime( 0.0f )
, m_uLinkToTake( ASURA_GUID_UNREGISTERED )
, m_fRemainingTimeslice( 0.0f )
{
}

Axon_Behaviour_GS2Node::Axon_Behaviour_GS2Node( const u_int uBehaviourType, Axon_Brain* pxBrain )
: PARENT( uBehaviourType, pxBrain )
, m_iState( UNSET )
, m_bStopHere( false )
, m_uAnimSet( uAXON_ANIMATION_SEQUENCE_ID_NONE )
, m_iExitConditions( AXON_GS2_NEC_NONE )
, m_fExitWaitTime( 0.0f )
, m_uLinkToTake( ASURA_GUID_UNREGISTERED )
, m_fRemainingTimeslice( 0.0f )
{
}

void Axon_Behaviour_GS2Node::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	int iVersion;
	xStream >> iVersion;

	PARENT::ReadFromChunkStream( xStream );

	if ( iVersion == CURRENT_CHUNK_VERSION )
	{
		xStream >> m_iState;
		xStream >> m_bStopHere;
		xStream >> m_uAnimSet;
		xStream >> m_iExitConditions;
		xStream >> m_fExitWaitTime;
		xStream >> m_uLinkToTake;
#if defined(ASURA_USE_STREAMING) && !defined(ASURA_TOOLS)
		m_uLinkToTake = Asura_Streaming_System::MakeCurrentMasterZoneGuid( m_uLinkToTake );
#endif
	}
	else
	{
		Asura_Assertf( false, ( "Axon_Behaviour_GS2Node::ReadFromChunkStream - unhandled chunk version (%d)", iVersion ) );
	}
}

void Axon_Behaviour_GS2Node::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << CURRENT_CHUNK_VERSION;

	PARENT::WriteToChunkStream( xStream );

	xStream << m_iState;
	xStream << m_bStopHere;
	xStream << m_uAnimSet;
	xStream << m_iExitConditions;
	xStream << m_fExitWaitTime;
	xStream << m_uLinkToTake;
}

void Axon_Behaviour_GS2Node::HandleParams( const Axon_BehaviourParams* pxParams )
{
	PARENT::HandleParams( pxParams );

	Asura_Assert( pxParams->IsBaseOrDerivativeOf( AXON_BP_TYPE_GS2_NODE ), "Axon_Behaviour_GS2Node::HandleParams - unexpected params type" );
	if ( pxParams->IsBaseOrDerivativeOf( AXON_BP_TYPE_GS2_NODE ) )
	{
		const Axon_BehaviourParams_GS2Node* pxNodeParams = static_cast< const Axon_BehaviourParams_GS2Node* >( pxParams );
		m_bStopHere		  = pxNodeParams->m_bStopHere;
        m_uAnimSet        = pxNodeParams->m_uAnimSet;
        m_iExitConditions = pxNodeParams->m_iExitCondition;
		m_fExitWaitTime	  = Asura_Random::GetFloat( pxNodeParams->m_fMinExitWaitTime, pxNodeParams->m_fMaxExitWaitTime );
	}
}

bool Axon_Behaviour_GS2Node::HandleMessage( const Asura_Message* pxMessage )
{
	if ( pxMessage->GetMsgID() == ASURA_MSG_AXON_GS2_ANIM_CHANGE )
	{
		const Axon_Msg_GS2_Anim_Change* pxAnimChangeMsg = static_cast< const Axon_Msg_GS2_Anim_Change* >( pxMessage->GetData() );
		m_uAnimSet = pxAnimChangeMsg->m_uAnimSet;
		GetGSActorBehaviour()->PlayAnimSet( m_uAnimSet );
		return true;
	}
	else
	{
		return PARENT::HandleMessage( pxMessage );
	}
}

bool Axon_Behaviour_GS2Node::Activate()
{
	const bool bParentOk = PARENT::Activate();

	if ( m_bStopHere )
	{
		SwitchToState( WAITING );
	}
	else
	{
		SwitchToState( FINISHED );
	}

	return bParentOk;
}

void Axon_Behaviour_GS2Node::Deactivate()
{
	SwitchToState( UNSET );
	PARENT::Deactivate();
}

void Axon_Behaviour_GS2Node::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	m_fRemainingTimeslice = fTimeslice;

	if ( IsInError() || IsFinished() )
	{
		return;
	}

	switch ( m_iState )
	{
		case WAITING:
		{
			if ( UpdateExitConditions( fTimeslice ) )
			{
				SwitchToState( FINISHED );
			}

			break;
		}

		default:
		case UNSET:
		case FINISHED:
		case NUM_STATES:
		{
			break;
		}
	}

	m_fRemainingTimeslice = 0.0f;
}

bool Axon_Behaviour_GS2Node::InitialiseExitConditions()
{
	switch ( m_iExitConditions )
	{
		case AXON_GS2_NEC_NONE:
		case AXON_GS2_NEC_TIMEOUT:
		case AXON_GS2_NEC_ANIMS_FINISHED:
		{
			return true;
		}

		case AXON_GS2_NEC_NEXT_NODE_FREE:
		{
			const Asura_ServerNode_GS2Node* pxNode = Asura_ServerNode_GS2Node::FindEntity( GetGSActorBehaviour()->GetCurrentEntityGuid() );
			if ( pxNode )
			{
				m_uLinkToTake = ASURA_GUID_UNREGISTERED;

				// Choose our link to take if necessary, otherwise we just pick whichever one is free first.
				if ( GetGSActorBehaviour()->GetRoute().Size() > 0 )
				{
					// Take the next link in our route if we can.
					const Asura_Guid uRouteLink = GetGSActorBehaviour()->GetRoute().Front()->Data();
					for ( u_int u = 0; u < pxNode->GetNumberLinks(); ++u )
					{
						if ( pxNode->GetLink( u ) == uRouteLink )
						{
							m_uLinkToTake = uRouteLink;
							GetGSActorBehaviour()->IncrementAlongRoute();
							return true;
						}
					}

					if ( m_uLinkToTake == ASURA_GUID_UNREGISTERED )
					{
						// If we have a route but cannot find the correct links we should discard the route and carry on as normal.
						GetGSActorBehaviour()->ClearRoute();
					}
				}

				if ( GetGSActorBehaviour()->GetLinksToTake().Size() > 0 )
				{
					// Search for any specific link to take.
					for ( Asura_LinkList_ConstIt< Asura_Guid > xIt( GetGSActorBehaviour()->GetLinksToTake() ); !xIt.Done(); xIt.Next() )
					{
						const Asura_Guid uLinkToTake = *xIt.GetCurrent();

						for ( u_int u = 0; u < pxNode->GetNumberLinks(); ++u )
						{
							if ( pxNode->GetLink( u ) == uLinkToTake )
							{
								m_uLinkToTake = uLinkToTake;
								return true;
							}
						}
					}
				}

				return true;
			}

			return false;
		}

		default:
		{
			return false;
		}
	}
}

bool Axon_Behaviour_GS2Node::UpdateExitConditions( const float fTimeslice )
{
	switch ( m_iExitConditions )
	{
		case AXON_GS2_NEC_NONE:
		{
			return false;
		}

		case AXON_GS2_NEC_TIMEOUT:
		{
			m_fExitWaitTime -= fTimeslice;
			return ( m_fExitWaitTime <= 0.0f );
		}

		case AXON_GS2_NEC_ANIMS_FINISHED:
		{
			return ( m_uAnimSet == uAXON_ANIMATION_SEQUENCE_ID_NONE || GetGSActorBehaviour()->AnimSetIsFinished( m_uAnimSet ) );
		}

		case AXON_GS2_NEC_NEXT_NODE_FREE:
		{
			if ( m_uLinkToTake == ASURA_GUID_UNREGISTERED )
			{
				// Just take the link to the next node we find that is free.
				if ( GetLinkToNextFreeNode( m_uLinkToTake ) )
				{
					return ( m_uLinkToTake != ASURA_GUID_UNREGISTERED );
				}
				else
				{
					m_uLinkToTake = ASURA_GUID_UNREGISTERED;
					SetInError();
					return true;
				}
			}
			else
			{
				// Find our next node.
				Asura_ServerNode_GS2Node*   pxNextNode = NULL;
				Asura_ServerNode_GS2Spline* pxLink     = Asura_ServerNode_GS2Spline::FindEntity( m_uLinkToTake );
				if ( pxLink )
				{
					pxNextNode = Asura_ServerNode_GS2Node::FindEntity( pxLink->GetTargetGuid() );
				}
				if ( pxNextNode )
				{
					return ( !pxNextNode->IsOccupied() && !pxNextNode->HasPendingActors() );
				}
				else
				{
					m_uLinkToTake = ASURA_GUID_UNREGISTERED;
					SetInError();
					return true;
				}
			}
		}

		default:
		{
			Asura_Assert( false, "Axon_Behaviour_GS2Node::UpdateExitConditions - unhandled exit condition" );
			return true;
		}
	}
}

void Axon_Behaviour_GS2Node::ShutdownExitConditions()
{
}

void Axon_Behaviour_GS2Node::SwitchToState( const State iState )
{
	// Exit the previous state.
	switch ( m_iState )
	{
		case WAITING:
		{
			ShutdownExitConditions();
			GetGSActorBehaviour()->StopAnimSet();
			break;
		}

		case FINISHED:
		{
			SetFinished( false );
			break;
		}

		default:
		case UNSET:
		case NUM_STATES:
		{
			break;
		}
	}

	// Enter the next state.
	m_iState = iState;
	switch ( m_iState )
	{
		case WAITING:
		{
			const Asura_ServerNode_GS2Node* pxNode = Asura_ServerNode_GS2Node::FindEntity( GetGSActorBehaviour()->GetCurrentEntityGuid() );
			if ( pxNode )
			{
				Asura_Matrix_3x3 xNodeOrientM;
				pxNode->GetOrientation( xNodeOrientM );
				GetGSActorBehaviour()->SetOwnerOrientation( xNodeOrientM );
				GetGSActorBehaviour()->EnableOwnerPhysicsCollisions( pxNode->ShouldTestForCollisions() );
			}

			GetGSActorBehaviour()->PlayAnimSet( m_uAnimSet );

			if ( !InitialiseExitConditions() )
			{
				SetInError();
				SwitchToState( FINISHED );
			}

			break;
		}

		case FINISHED:
		{
			SetFinished();
			break;
		}

		default:
		case UNSET:
		case NUM_STATES:
		{
			break;
		}
	}
}

bool Axon_Behaviour_GS2Node::GetLinkToNextFreeNode( Asura_Guid& uLinkGuid ) const
{
	uLinkGuid = ASURA_GUID_UNREGISTERED;

	const Asura_ServerNode_GS2Node* pxOurNode = Asura_ServerNode_GS2Node::FindEntity( GetGSActorBehaviour()->GetCurrentEntityGuid() );
	if ( !pxOurNode ) { return false; }

	// Count the number of free nodes first.
	u_int uNumFreeNodes = 0;
	for ( u_int u = 0; u < pxOurNode->GetNumberLinks(); ++u )
	{
		const Asura_ServerNode_GS2Spline* pxLink = Asura_ServerNode_GS2Spline::FindEntity( pxOurNode->GetLink( u ) );
		if ( !pxLink ) { return false; }

		if ( pxLink->IsOccupied() )
		{
			continue;
		}

		const Asura_ServerNode_GS2Node* pxNextNode = Asura_ServerNode_GS2Node::FindEntity( pxLink->GetTargetGuid() );
		if ( !pxNextNode ) { return false; }

		if ( pxNextNode->IsOccupied() || pxNextNode->HasPendingActors() )
		{
			continue;
		}

		++uNumFreeNodes;
	}
	if ( uNumFreeNodes == 0 ) { return true; }

	// Pick a free node at random.
	const u_int uRandomLink = Asura_Random::GetIntU( 0, uNumFreeNodes - 1 );
	uNumFreeNodes = 0;
	for ( u_int u = 0; u < pxOurNode->GetNumberLinks(); ++u )
	{
		const Asura_ServerNode_GS2Spline* pxLink = Asura_ServerNode_GS2Spline::FindEntity( pxOurNode->GetLink( u ) );
		if ( !pxLink ) { return false; }

		if ( pxLink->IsOccupied() )
		{
			continue;
		}

		const Asura_ServerNode_GS2Node* pxNextNode = Asura_ServerNode_GS2Node::FindEntity( pxLink->GetTargetGuid() );
		if ( !pxNextNode ) { return false; }

		if ( pxNextNode->IsOccupied() || pxNextNode->HasPendingActors() )
		{
			continue;
		}

		if ( uNumFreeNodes == uRandomLink )
		{
			uLinkGuid = pxLink->GetGuid();
			return true;
		}
		++uNumFreeNodes;
	}

	return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // ASURA_USE_AXON
