//////////////////////////////////////////////////////////////////////////////
//
// Axon_Behaviour_Wander.cpp
//
// Copyright (C) 2009 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

#ifdef ASURA_USE_NEW_NAV_STUFF

// header
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Wander.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Debugging/Asura_DebugLineRender.h"
#endif
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#ifndef ASURA_NO_DEBUG_OUTPUT
#include "Rendering/Asura_Font_System.h"
#endif

// axon
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Movement.h"
#include "Behaviour/BehaviourTypes/Axon_Behaviour_Types.h"
#include "Core/Axon_Brain.h"

//// S T A T I C S ///////////////////////////////////////////////////////////

static const float fWANDER_RANGE = 30.0f;
static const float fSTAND_STILL_PERIOD_MIN = 4.0f;
static const float fSTAND_STILL_PERIOD_MAX = 10.0f;

//////////////////////////////////////////////////////////////////////////////
//
// InDirectionToIdealFunctor
//
//////////////////////////////////////////////////////////////////////////////

class AwayFromInRangePathFindFunctor : public Asura_Navigation::AwayFromPathFindFunctor
{
public:
	AwayFromInRangePathFindFunctor() : m_fMaxDistanceFromCenterSQ( 0.0f ), m_xCentralPosition( Asura_Maths::ZeroVector3 ) {}

	void SetMaxDistanceFromCenter( const float fDistance ) { m_fMaxDistanceFromCenterSQ = fDistance * fDistance; }
	void SetCentralPosition( const Asura_Vector_3& xPosition ) { m_xCentralPosition = xPosition; }

protected:
	virtual bool VerifyInputData()
	{
		return ( PARENT::VerifyInputData() && (m_fMaxDistanceFromCenterSQ > 0.0f) );
	}
	virtual bool ShouldIgnoreLink( const u_int uSourceWPID, const Asura_AI_NavVertex& xSourceWP, const Asura_AI_NavConnection& xLink, const u_int uTargetWPID, const Asura_AI_NavVertex& xTargetWP )
	{
		// No waypoint in the path should be outside the Range,
		const Asura_Vector_3 xCentralPosToTargetWP = xTargetWP.GetPosition() - m_xCentralPosition;
		const float fCentralPosToTargetWPSQ = xCentralPosToTargetWP.MagnitudeSquared();
		if ( fCentralPosToTargetWPSQ > m_fMaxDistanceFromCenterSQ ) 
		{
			return true;
		}
		return PARENT::ShouldIgnoreLink( uSourceWPID, xSourceWP, xLink, uTargetWPID, xTargetWP );
	}
	virtual bool GoalAchieved( const u_int uWPID, const Asura_AI_NavVertex& xWP )
	{
		const Asura_Vector_3 xCentralToWP = xWP.GetPosition() - m_xCentralPosition;
		const float fCentralToWPSQ = xCentralToWP.MagnitudeSquared();
		if ( fCentralToWPSQ > m_fMaxDistanceFromCenterSQ ) 
		{
			return false;
		}
		return PARENT::GoalAchieved( uWPID, xWP );
	}

private:
	float m_fMaxDistanceFromCenterSQ; // Distance from central position that we must remain within
	Asura_Vector_3 m_xCentralPosition;

	typedef Asura_Navigation::AwayFromPathFindFunctor PARENT;
};

//////////////////////////////////////////////////////////////////////////////
//
// Axon_Behaviour_Wander
//
//////////////////////////////////////////////////////////////////////////////

Axon_Behaviour_Wander::Axon_Behaviour_Wander( Axon_Brain* pxBrain )
: PARENT( AXON_BEHAVIOUR_TYPE_WANDER, pxBrain )
, m_xCentralWanderPosition( Asura_Maths::ZeroVector3 )
, m_fWanderRange( fWANDER_RANGE )
, m_iState( WAITING_AT_POSITION )
, m_fTimer( 0.0f )
, m_fWaitAtPositionTimer( 0.0f )
, m_uWanderFlags( 0 )
, m_pxOwnerPhyObj( NULL )
{
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::Deactivate()
{
	DestroyChildWithType( AXON_BEHAVIOUR_TYPE_MOVEMENT );
	PARENT::Deactivate();
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::PostLoadInitialise()
{
	PARENT::PostLoadInitialise();

	if ( !UpdateOwnerPhyObjPtr() ) return;

	if ( m_iState != WAITING_AT_POSITION )
	{
		UpdateIsInRange();

		if ( m_bInRange )
		{
			SwitchToState( MOVING_TO_POSITION );
		}
		else
		{	
			SwitchToState( MOVING_INTO_RANGE );
		}
	}
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::ReadFromChunkStream( Asura_Chunk_Stream& xStream )
{
	u_int uVersion;
	xStream >> uVersion;

	PARENT::ReadFromChunkStream( xStream );

	if ( uVersion == uCURRENT_CHUNK_VERSION )
	{
		xStream >> m_xCentralWanderPosition;
		xStream >> m_fWanderRange;
		xStream >> m_iState;
	}
	else
	{
		Asura_Assertf( false, ( "Axon_Behaviour_Wander::ReadFromChunkStream - unhanded chunk version (%u)", uVersion ) );
	}
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::WriteToChunkStream( Asura_Chunk_Stream& xStream ) const
{
	xStream << uCURRENT_CHUNK_VERSION;

	PARENT::WriteToChunkStream( xStream );

	xStream << m_xCentralWanderPosition;
	xStream << m_fWanderRange;
	xStream << m_iState;
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::SetCentralWanderPos( const Asura_Vector_3 &xCentralPos )
{
	m_xCentralWanderPosition = xCentralPos;
	OnWanderParamChange();
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::SetWanderRange( const float fRange )
{
	m_fWanderRange = fRange;
	OnWanderParamChange();
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::OnWanderParamChange()
{
	if ( UpdateOwnerPhyObjPtr() )
	{
		UpdateIsInRange();

		if ( m_bInRange )
		{
			SwitchToState( MOVING_TO_POSITION );
		}
		else
		{
			SwitchToState( MOVING_INTO_RANGE );
		}
	}
}

//--------------------------------------------------------

bool Axon_Behaviour_Wander::UpdateOwnerPhyObjPtr() 
{
	m_pxOwnerPhyObj = Asura_ServerEntity_PhysicalObject::FindEntity( GetBrain()->GetOwnerGuid() );
	Asura_Assertf( m_pxOwnerPhyObj, ( "UpdateOwnerPhyObjPtr() - Brain Owner Guid %u, is not found in the Asura_ServerEntity_PhysicalObject list.", GetBrain()->GetOwnerGuid() ) );
	return ( m_pxOwnerPhyObj != NULL ); 
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::UpdateIsInRange() 
{
	const float fDistToCenterSQ = (GetOwnerPhyObj()->GetPosition() - m_xCentralWanderPosition).MagnitudeSquared();
	m_bInRange = ( fDistToCenterSQ < (m_fWanderRange * m_fWanderRange) ) ? true : false;
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::PerFrameUpdate( const float fTimeslice )
{
	PARENT::PerFrameUpdate( fTimeslice );

	m_fTimer += fTimeslice;

	if ( !UpdateOwnerPhyObjPtr() )
	{
		SetInError();
		return;
	}

	UpdateIsInRange();

	switch ( m_iState )
	{
	case WAITING_AT_POSITION:
		{
			if ( m_fTimer > m_fWaitAtPositionTimer )
			{
				if ( m_bInRange )
				{
					SwitchToState( MOVING_TO_POSITION );
				}
				else
				{
					SwitchToState( MOVING_INTO_RANGE );
				}
			}
			break;
		}
	case MOVING_TO_POSITION:
		{
			Axon_Behaviour_Movement* pxMovement = static_cast< Axon_Behaviour_Movement* >( GetChildWithType( this, AXON_BEHAVIOUR_TYPE_MOVEMENT ) );
			if ( !pxMovement || (pxMovement->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_AT_DESTINATION)
				|| (pxMovement->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_IN_ERROR) )
			{
				SwitchToState( WAITING_AT_POSITION );
			}
			break;
		}
		case MOVING_INTO_RANGE:
		{
			Axon_Behaviour_Movement* pxMovement = static_cast< Axon_Behaviour_Movement* >( GetChildWithType( this, AXON_BEHAVIOUR_TYPE_MOVEMENT ) );
			if ( !pxMovement || (pxMovement->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_AT_DESTINATION)
				|| (pxMovement->GetPathTraversalState() == Axon_Behaviour_Movement::PTS_IN_ERROR) )
			{
				SwitchToState( WAITING_AT_POSITION );
			}
			else if ( m_bInRange )
			{
				// Once we are in Range we start wandering in Area.
				SwitchToState( MOVING_TO_POSITION );
			}
			break;
		}
	default:
		break;
	}
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::SetWaitAtPositionTimer()
{
	m_fWaitAtPositionTimer = Asura_Random::GetFloat( fSTAND_STILL_PERIOD_MIN, fSTAND_STILL_PERIOD_MAX );
}

//--------------------------------------------------------

bool Axon_Behaviour_Wander::StartWanderMove()
{
	Axon_Behaviour_Movement* pxMovement = static_cast< Axon_Behaviour_Movement* >( CreateChildWithType( AXON_BEHAVIOUR_TYPE_MOVEMENT ) );
	if ( pxMovement )
	{
		// Find a position at the correct distance from the target in a random direction.
		AwayFromInRangePathFindFunctor xAFIRPFFunctor;
		xAFIRPFFunctor.SetInputBuffer( Asura_AI_Navigation_System::GetWaypoints(), Asura_AI_Navigation_System::GetNumberOfWaypoints() );
		xAFIRPFFunctor.SetOutputBuffer( &pxMovement->GetPath().GetPathPoints() );
		xAFIRPFFunctor.SetFromPos( GetOwnerPhyObj()->GetPosition() ,pxMovement->GetTestParams() );
		xAFIRPFFunctor.SetMaxDistanceFromCenter( m_fWanderRange );
		xAFIRPFFunctor.SetCentralPosition( m_xCentralWanderPosition );
		xAFIRPFFunctor.SetAvoidPosAndDistance( GetOwnerPhyObj()->GetPosition(), Asura_Random::GetFloat( (m_fWanderRange*0.25f), (m_fWanderRange*0.75f) ) );
		if ( pxMovement->FindPath( xAFIRPFFunctor ) )
		{
			pxMovement->StartPathTraversal();
			return true;
		}
		else
		{
			Asura_OutputToDebugger_Retail(( "Axon_Behaviour_Wander, was unable to find a path" ));
		}
	}
	return false;
}

//--------------------------------------------------------

bool Axon_Behaviour_Wander::StartMoveIntoRange()
{
	Axon_Behaviour_Movement* pxMovement = static_cast< Axon_Behaviour_Movement* >( CreateChildWithType( AXON_BEHAVIOUR_TYPE_MOVEMENT ) );
	if ( pxMovement )
	{
		// Find a position at the correct distance from the target in a random direction.
		if ( pxMovement->FindPathTo( m_xCentralWanderPosition ) )
		{
			pxMovement->StartPathTraversal();
			return true;
		}
		else
		{
			Asura_OutputToDebugger_Retail(( "Axon_Behaviour_Wander, was unable to find a path into Range" ));
		}
	}
	return false;
}

//--------------------------------------------------------

void Axon_Behaviour_Wander::SwitchToState( int iNextState )
{
	switch ( m_iState )
	{
	case WAITING_AT_POSITION:
		{
			break;
		}
	case MOVING_TO_POSITION:
	case MOVING_INTO_RANGE:
		{
			DestroyAllChildren();
			break;
		}
	default:
		break;
	}

	m_iState = iNextState;
	m_fTimer = 0.0f;
	switch ( m_iState )
	{
	case WAITING_AT_POSITION:
		{
			SetWaitAtPositionTimer();
			break;
		}

	case MOVING_TO_POSITION:
		{
			if ( !StartWanderMove() )
			{
				SwitchToState( WAITING_AT_POSITION );
			}
			break;
		}
	case MOVING_INTO_RANGE:
		{
			if ( !StartMoveIntoRange() )
			{
				SwitchToState( WAITING_AT_POSITION );
			}
			break;
		}
	default:
		break;
	}
}

#ifndef ASURA_NO_DEBUG_OUTPUT
static const char* s_aszStateStrings[] =
{
	// N.B. Keep in updated with behaviour states.
	"Wander - WAITING_AT_POSITION",
	"Wander - MOVING_TO_POSITION",
	"Wander - MOVING_INTO_RANGE",
};

void Axon_Behaviour_Wander::RenderDebugInfo()
{
	PARENT::RenderDebugInfo();
	Asura_DebugLineRender::AddSphere( m_xCentralWanderPosition, m_fWanderRange, 0xff99ff33, 12 );
	if ( GetOwnerPhyObj() )
	{
		Asura_Vector_3 xUpVec( 0.0f, -0.1f, 0.0f );
		Asura_DebugLineRender::AddLine( m_xCentralWanderPosition+xUpVec, GetOwnerPhyObj()->GetPosition()+xUpVec, 0xffcc0033 );

		if ( GetBrain()->DebugTextPosOnScreen() )
		{
			Asura_Font_System::Print( GetBrain()->GetDebugTextPos().x, GetBrain()->GetDebugTextPos().y, 0.5f, 0xffffffff, 
				ASURA_TEXTFORMAT_CENTREJUSTIFIED | ASURA_TEXTFORMAT_FORMATSTRING, s_aszStateStrings[m_iState] );
			GetBrain()->IncrementDebugTextPos();
		}
	}
}
#endif

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif // #ifdef ASURA_USE_NEW_NAV_STUFF
