//
// Tools_Attractor.cpp
//
// Created by JGP 10/05/06
//
// Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_ATTRACTORS

// This file's header
#include "Tools_Attractor.h"

#include "Entity/Attractors/Asura_ServerEntity_AttractorController.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"

#include "Attractors/Tools_Attractor_System.h"
#include "Attractors/Tools_Attractor_Node.h"
#include "Attractors/Tools_Object_Attractor_Node.h"
#include "EntityTemplate/Tools_EntityTemplate_System.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"


//If the x and z components of the attractor direction are smaller than this value , then it
//will use the special vertical case for dealing with calculating the normal from the angle.
//RAR 21/09/2007
static const float g_fNEAR_VERTICAL_TOLERANCE = 0.001f;

///////////////////////////////////////////////////////////////////////////////////////
// Construct / Destruct
///////////////////////////////////////////////////////////////////////////////////////


Tools_Attractor::Tools_Attractor(u_int uID)
{
	Initialise();

	if(!uID) uID = Tools_Attractor_System::GetInstance()->GenerateID(false);
	SetID(uID);
}

Tools_Attractor::Tools_Attractor(bool bMissionFile)
{
	Initialise();

	u_int uID = Tools_Attractor_System::GetInstance()->GenerateID(bMissionFile);
	SetID(uID);
}


void Tools_Attractor::Initialise()
{
	m_pxFirstNode = NULL;
	m_bClosed = false;
	m_uTemplateGuid = ASURA_GUID_UNREGISTERED;
	m_iTemplateVersion = 0;
	m_uControllerGuid = 0;//initialised to zero until a search can be done to see if there is a controller entity
	m_uParentEntityGuid = ASURA_GUID_UNREGISTERED;
	m_fAngleForNormal = 0.0f;
}

Tools_Attractor::~Tools_Attractor()
{
	delete [] m_puLinks;
}


u_int Tools_Attractor::CountCompiledAttractors() const
{
	u_int uNumNodes = 0;
	Tools_Attractor_Node* pxNode = m_pxFirstNode;
	while(pxNode)
	{
		uNumNodes++;
		pxNode = pxNode->GetNext();
	}
	if(uNumNodes < 2) return 0;//must have at least 2 nodes
	return IsClosed() ? uNumNodes : uNumNodes-1;
}

bool Tools_Attractor::HasAtLeastTwoNodes() const
{
	return m_pxFirstNode->GetNext() != 0;
}

void Tools_Attractor::AddLink(u_int uID)
{
	if(!uID) return;

	//ignore attempts to link to itself
	if(uID == GetID()) return;

	//make sure we don't already have a link to this attractor
	for(u_int i=0;i<m_uNumLinks;i++)
	{
		if(m_puLinks[i] == uID) return;
	}

	//Enlarge the array and add the link
	u_int* puNewLinks = new u_int[m_uNumLinks+1];
	for(u_int i=0;i<m_uNumLinks;i++)
	{
		puNewLinks[i] = m_puLinks[i];
	}
	puNewLinks[m_uNumLinks] = uID;

	delete [] m_puLinks;
	m_puLinks = puNewLinks;
	m_uNumLinks++;
}

void Tools_Attractor::RemoveLink(u_int uID)
{
	//search for the ID in the list of links
	for(u_int i=0;i<m_uNumLinks;i++)
	{
		if(m_puLinks[i] == uID)
		{
			//copy the last link down to replace this
			m_puLinks[i] = m_puLinks[m_uNumLinks-1];
			m_uNumLinks--;
			return;
		}
	}
}

void Tools_Attractor::ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance)
{
	//update links
	for(u_int i=0; i<m_uNumLinks; i++)
	{
		m_puLinks[i] = pxInstance->ConvertAttractorID(m_puLinks[i]);
	}
	m_uParentEntityGuid = pxInstance->ConvertGuid(m_uParentEntityGuid);
	m_uControllerGuid = pxInstance->ConvertGuid(m_uControllerGuid);

	for(Tools_Attractor_Node* pxNode = m_pxFirstNode;pxNode;pxNode = pxNode->GetNext())
	{
		pxNode->ApplyChangesForTemplateInstance(pxInstance);
	}
}

void Tools_Attractor::GetNormalFromDirection(Asura_Vector_3& xNormal, const Asura_Vector_3& xDirection) const
{
	// This version will cause a vertical line to produce results inconsistent with any other direction. Hopefully this won't be a serious issue.
	Asura_Matrix_3x3 xOri;
	Asura_Euler_3 xEuler;
	// Special calculation for vertical direction, as OrientationFromDirection picks an orientation inconsistent with how the rest of this is calculated.
	if (Asura_Maths::Abs(xDirection.x) < g_fNEAR_VERTICAL_TOLERANCE && Asura_Maths::Abs(xDirection.z) < g_fNEAR_VERTICAL_TOLERANCE)
	{
		Asura_Vector_3 xXVector;
		Asura_Vector_3 xYVector;
		Asura_Vector_3 xZVector = xDirection;

		xZVector.Normalise();

		xXVector.x = 0.0f;
		xXVector.y = 0.0f;
		xXVector.z = xZVector.y;

		xXVector.Normalise();

		xYVector = -xXVector^xZVector;

		xOri.e11 = xXVector.x;
		xOri.e21 = xXVector.y;
		xOri.e31 = xXVector.z;

		xOri.e12 = xYVector.x;
		xOri.e22 = xYVector.y;
		xOri.e32 = xYVector.z;

		xOri.e13 = xZVector.x;
		xOri.e23 = xZVector.y;
		xOri.e33 = xZVector.z;

		xEuler.FromMatrix(xOri);
		// This also uses a different axis affected by the euler
		xEuler.y = Asura_Maths::RadiansToZeroCentredRange(xEuler.y + m_fAngleForNormal);

		// Apply our parent, if we have one
		const Asura_ServerEntity_PhysicalObject* pxEntity = Asura_ServerEntity_PhysicalObject::FindEntity(GetParentGuid());
		if (pxEntity)
		{
			Asura_Euler_3 xEntityEuler;
			xEntityEuler.FromMatrix(pxEntity->GetOrientation());

			xEuler.y += xEntityEuler.y;
		}
	}
	else
	{
		// Standard use of OrientationFromDirection
		Asura_Maths::OrientationFromDirection(xOri, xDirection);
		// This sets our direction as the z axis, so we can add our rotation to that in an euler.
		xEuler.FromMatrix(xOri);
		xEuler.z = Asura_Maths::RadiansToZeroCentredRange(xEuler.z + m_fAngleForNormal);

		// Apply our parent, if we have one
		const Asura_ServerEntity_PhysicalObject* pxEntity = Asura_ServerEntity_PhysicalObject::FindEntity(GetParentGuid());
		if (pxEntity)
		{
			Asura_Euler_3 xEntityEuler;
			xEntityEuler.FromMatrix(pxEntity->GetOrientation());

			xEuler.z += xEntityEuler.z;
		}
	}

	xEuler.ToMatrix(xOri);

	// Arbitrarily, we'll use the x axis.
	xNormal = Asura_Vector_3(xOri.e11, xOri.e21, xOri.e31);
}

void Tools_Attractor::SetNormal(const Asura_Vector_3& xNormal,const Tools_Attractor_Node* pxSelectedNode)
{
	if (!HasAtLeastTwoNodes()) { return; }

	//Get the two nodes to use for the attractor direction
	//RAR 28/03/2007
	const Tools_Attractor_Node* pxNode1 = pxSelectedNode ? pxSelectedNode : m_pxFirstNode;
	const Tools_Attractor_Node* pxNode2 = pxNode1->GetNext();

	if(!pxNode2)
	{
		//The last node was selected ,so use the direction back to the first node if closed , or
		//use the last two nodes otherwise.
		if(m_bClosed)
		{
			pxNode2 = m_pxFirstNode;
		}
		else
		{
			pxNode2 = pxNode1;
			pxNode1 = pxNode2->GetPrev();
		}
	}
	
	

	// Get the direction between the first two nodes
	Asura_Vector_3 xDir = pxNode2->GetPosition() - pxNode1->GetPosition();
	xDir.Normalise();

	Asura_Vector_3 xConvertedNormal(xNormal);
	xConvertedNormal.Normalise();

	// Ensure the normal and the direction are not in the same/opposite direction
	if (Asura_Maths::Abs(xNormal * xDir) > 0.999f)
	{
		Tools_Timed_Message_System::AddMessage( 0xffff0000, "Cannot use a normal in the same/opposite direction to the env hook." );
		return;
	}

	if (Asura_Maths::Abs(xDir.x) < g_fNEAR_VERTICAL_TOLERANCE && Asura_Maths::Abs(xDir.z) < g_fNEAR_VERTICAL_TOLERANCE)
	{
		// Find the perpendicular to the direction and normal
		Asura_Vector_3 xPerp(xDir ^ xConvertedNormal);
		xPerp.Normalise();
		// Ensure the normal we use is perpendicular to the other two axes
		xConvertedNormal = xPerp ^ xDir;

		Asura_Matrix_3x3 xOri(	xPerp.x, xConvertedNormal.x, xDir.x,
								xPerp.y, xConvertedNormal.y, xDir.y,
								xPerp.z, xConvertedNormal.z, xDir.z);

		// Get the rotation value we want to store
		Asura_Euler_3 xEuler;
		xEuler.FromMatrix(xOri);

		m_fAngleForNormal = xEuler.y;
	}
	else
	{
		// Find the perpendicular to the direction and normal
		Asura_Vector_3 xPerp(xConvertedNormal ^ xDir);
		xPerp.Normalise();
		// Ensure the normal we use is perpendicular to the other two axes
		xConvertedNormal = xDir ^ xPerp;

		
		Asura_Matrix_3x3 xOri(	xConvertedNormal.x, xPerp.x, xDir.x,
								xConvertedNormal.y, xPerp.y, xDir.y,
								xConvertedNormal.z, xPerp.z, xDir.z);

		// Get the rotation value we want to store
		Asura_Euler_3 xEuler;
		xEuler.FromMatrix(xOri);

		m_fAngleForNormal = Asura_Maths::Pi - xEuler.z;
	}
}

void Tools_Attractor::HandleUpdateForChangeToNormalCalculationForNearVerticalEnvHooks()
{
	//This deals with updating the angles so that old attractors will continue to
	//get the same normal as they did before , to the extent that is possible.
	//The update is needed for attractors that are almost , but not quite , vertical , since
	//these now go through the special vertical case , rather than risking the numerical
	//inaccuracies caused by the general case.
	//RAR 21/09/2007
	
	
	if (!HasAtLeastTwoNodes()) { return; }

	//get the direction vector for the first part of this attractor
	const Tools_Attractor_Node* pxNode1 = m_pxFirstNode;
	const Tools_Attractor_Node* pxNode2 = pxNode1->GetNext();

	Asura_Vector_3 xDir = pxNode2->GetPosition() - pxNode1->GetPosition();
	xDir.Normalise();

	//Is this nearly , but not exactly , vertical?
	if(Asura_Maths::Abs(xDir.x) >= g_fNEAR_VERTICAL_TOLERANCE || Asura_Maths::Abs(xDir.z) >= g_fNEAR_VERTICAL_TOLERANCE) return;
	if(xDir.x == 0.0f && xDir.z == 0.0f) return;

	//Get the normal using the old method.
	Asura_Vector_3 xNormal;
	{
		Asura_Matrix_3x3 xOri;
		Asura_Euler_3 xEuler;

		Asura_Maths::OrientationFromDirection(xOri, xDir);
		// This sets our direction as the z axis, so we can add our rotation to that in an euler.
		xEuler.FromMatrix(xOri);
		xEuler.z = Asura_Maths::RadiansToZeroCentredRange(xEuler.z + m_fAngleForNormal);

		xEuler.ToMatrix(xOri);

		// Arbitrarily, we'll use the x axis.
		xNormal = Asura_Vector_3(xOri.e11, xOri.e21, xOri.e31);

	}

	//Set the normal again using the current method	, to recalculate the new angle needed
	SetNormal(xNormal, pxNode1);
}

void Tools_Attractor::SetAttractorType(Asura_Hash_ID uType, const Tools_Attractor_Node* pxSelectedNode)
{
	if(GetType() == uType) return;
	SetType(uType);
	 
	const Tools_Attractor_Type* pxType = Tools_Attractor_System::GetInstance()->GetAttractorTypeByHash(uType);
	if(pxType)
	{
		//update the mandatory properties
		UpdateFromType(pxSelectedNode);
	
		//and set any defaults
		SetRadius(pxType->m_fDefaultRadius);
	}
    	
}

void Tools_Attractor::UpdateFromType(const Tools_Attractor_Node* pxSelectedNode)
{
	const Tools_Attractor_Type* pxType = Tools_Attractor_System::GetInstance()->GetAttractorTypeByHash(GetType());
	if(pxType)
	{
		if(pxType->m_bMustBeDefaultRadius)
		{
			SetRadius(pxType->m_fDefaultRadius);
		}

		if(pxType->m_bMustBeVertical || pxType->m_bMustBeHorizontal)
		{
			//Copy x,z or y coordinate from the selected node to the others
			if(!pxSelectedNode) pxSelectedNode = GetFirstNode();
			if(pxSelectedNode)
			{
				const Asura_Vector_3 xSelectedPos = pxSelectedNode->GetPosition();
				for(Tools_Attractor_Node* pxNode = GetFirstNode(); pxNode; pxNode = pxNode->GetNext())
				{
					if(pxNode == pxSelectedNode) continue;

					Asura_Vector_3 xPos = pxNode->GetPosition();
					if(pxType->m_bMustBeVertical)
					{
						xPos.x = xSelectedPos.x;
						xPos.z = xSelectedPos.z;
					}
					if(pxType->m_bMustBeHorizontal)
					{
						xPos.y = xSelectedPos.y;
					}
					pxNode->SetPosition(xPos);

					//if there is a tools node , then update the position in that as well
					if(pxNode->GetToolsNode())
					{
						pxNode->GetToolsNode()->Position = xPos;
					}
				}
			}
		}
	}
}

bool Tools_Attractor::BelongsToMissionFile() const
{
	return (GetID() & TOOLS_ATTRACTORID_MISSIONFILE) != 0;
}

void Tools_Attractor::SetModified() const
{
	if(BelongsToMissionFile())
	{
		Tools_SaveAsura::SetMissionAttractorsModified();
	}
	else
	{
		Tools_SaveAsura::SetAttractorsModified();
	}
}

Asura_Guid Tools_Attractor::GetControllerGuid() const
{
	if(!m_uControllerGuid)
	{
		//set controller guid in case no entity exists
		m_uControllerGuid = ASURA_GUID_UNREGISTERED;

		//Search for a controller entity for this attractor
		for(Asura_Hashed_List_It<Asura_ServerEntity_AttractorController> xIt(Asura_ServerEntity_AttractorController::GetList()); !xIt.Done(); xIt.Next())
		{
			Asura_ServerEntity_AttractorController* pxEntity = xIt.GetCurrent();
		
			//any matching attractor?
			if(pxEntity->GetToolsAttractorID() == GetID())
			{
				m_uControllerGuid = pxEntity->GetGuid();
				break;
			}
		}
	}
	return m_uControllerGuid;
}

bool Tools_Attractor::AreAnyNodesSelected() const
{
	for(const Tools_Attractor_Node* pxNode = GetFirstNode(); pxNode; pxNode = pxNode->GetNext())
	{
		const Tools_Object_Attractor_Node* pxNodeObject = pxNode->GetToolsNode();
		if(pxNodeObject && pxNodeObject->IsGroupSelected())
		{
			return true;
		}
	}
	return false;
}

bool Tools_Attractor::AreAnyNodesOrTheirRootObjectsSelected() const
{
	for(const Tools_Attractor_Node* pxNode = GetFirstNode(); pxNode; pxNode = pxNode->GetNext())
	{
		const Tools_Object_Attractor_Node* pxNodeObject = pxNode->GetToolsNode();
		if(pxNodeObject && pxNodeObject->IsRootObjectGroupSelected())
		{
			return true;
		}
	}
	return false;
}


#endif // ASURA_USE_ATTRACTORS
