//
// Tools_BillboardPlant_Scatter.cpp
//
// Created by Andrew Weinkove 09/12/02
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "BillboardPlants/Tools_BillboardPlant_Scatter.h"
#include "Debugging/Asura_Assert.h"
#include "Core/Asura_Maths.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Module.h"
#include "Misc/Tools_Polygon_Selection.h"
#include "BillboardPlants/Tools_BillboardPlant_Object.h"
#include "Misc/Tools_Timed_Messages.h"
#include "BillboardPlants/Tools_Mode_BillboardPlants.h"

#ifdef ASURA_USE_BILLBOARD_PLANTS
List<Tools_BillboardPlant_Object*>Tools_BillboardPlant_Scatter::s_listCurrentRandom;
float Tools_BillboardPlant_Scatter::s_fClumpProximityMaxSquared = 9.0f;
float Tools_BillboardPlant_Scatter::s_fPlantProximityMinSquared = 0.25f;

void Tools_BillboardPlant_Scatter::Scatter(unsigned uPlantType, float fNumberToCreate, float fClumpiness)
{
	Reject();
	int iNumberToCreate = Asura_Maths::FloatToInt(fNumberToCreate);
	if (iNumberToCreate <= 0)
		return;
	if (!Tools_Polygon_Selection::NumSelectedModules)
		return;
	bool bOK = true;
	for (int i=0;i<iNumberToCreate;i++)
	{
		float fRandom = Asura_Random::GetFloat();
		if (i==0 ||
			fRandom > fClumpiness)
			bOK = CreateRandomPlant(uPlantType);
		else
			bOK = CreateClumpedPlant(uPlantType);
		if (!bOK)
			break;
	}
	if (!bOK)
		Tools_Timed_Message_System::AddMessage(0xffffff, "Error - Plant creation timed out, was there enough space for all those plants?");
}

void Tools_BillboardPlant_Scatter::Accept()
{
	while (s_listCurrentRandom.size())
	{
		s_listCurrentRandom.delete_first_entry();
	}
}

void Tools_BillboardPlant_Scatter::Reject()
{
	while (s_listCurrentRandom.size())
	{
		s_listCurrentRandom.first_entry()->Destroy();
		s_listCurrentRandom.delete_first_entry();
	}
}

void Tools_BillboardPlant_Scatter::GetRandomPositionInPoly(const PolygonID &xPolygon, Asura_Vector_3 &xPosition)
{
	//get the mesh data
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);
	//and the modules location
	Asura_Vector_3  module_loc;
	Asura_Module_System::GetPositionOfModule(xPolygon.ModuleIndex,module_loc);
	//get the polygon
	Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[xPolygon.PolyIndex];

	Asura_Assert(poly->NumberOfVertices() == 3 ||
				 poly->NumberOfVertices() == 4, "Unexpected number of vertices in polygon");
	Asura_Vector_3 xBasePoint;
	Asura_Vector_3 xEdgeVectors[3];
	bool bIsQuad = poly->NumberOfVertices() == 4;
	int iIndexToIgnore = -1;
	if (bIsQuad)
		iIndexToIgnore = Asura_Random::GetIntU() % 4;
	int iIndexCount = 0;
	Asura_Vector_3 xLastPos;
	for(int j=0; j<poly->NumberOfVertices(); j++)
	{
		if (iIndexToIgnore == j)
			continue;	// Ignore a random vertex if it's a quad.
		Asura_Vector_3 xThisPos = module_loc + mesh->VertexDataPtr[poly->VertexID[j]].Position;
		if (iIndexCount == 0)
			xBasePoint = xThisPos;
		else
		{
			xEdgeVectors[iIndexCount - 1] = xThisPos - xLastPos;
		}
		iIndexCount++;
		xLastPos = xThisPos;
	}
	xEdgeVectors[2] = xBasePoint - xLastPos;

	float fRandom = 1.0f;
	for (int i=0;i<3;i++)
	{
		fRandom = Asura_Random::GetFloat() * fRandom;
		xBasePoint += xEdgeVectors[i] * fRandom;
	}
	xPosition = xBasePoint;
}

bool Tools_BillboardPlant_Scatter::IsPositionOKToCreatePlant(const Asura_Vector_3& xPosition, const unsigned uPlantType)
{
	// Test for proximity to other plants.
	for( CLIF<Tools_BillboardPlant_Object *> it(&s_listCurrentRandom); !it.done(); it.next() )
	{
		Asura_Vector_3 xDiff = it()->Position - xPosition;
		if (xDiff * xDiff < s_fPlantProximityMinSquared)
			return false;
	}
	// Check LOS with environment.
	if (Tools_Mode_BillboardPlants::TestForCollisions())
	{
		if (Tools_Mode_BillboardPlants::DoesPlantHitEnvironment(xPosition, uPlantType))
			return false;
	}
	return true;
}

bool Tools_BillboardPlant_Scatter::IsPlantClumped(Tools_BillboardPlant_Object *pxPlant)
{
	// Only accept if plant is near another plant.
	Asura_Vector_3 xPlantPosition = pxPlant->Position;
	for( CLIF<Tools_BillboardPlant_Object *> it(&s_listCurrentRandom); !it.done(); it.next() )
	{
		if (it() == pxPlant)
			continue;
		Asura_Vector_3 xDiff = it()->Position - xPlantPosition;
		if (xDiff * xDiff < s_fClumpProximityMaxSquared)
			return true;
	}
	return false;
}


Tools_BillboardPlant_Object *Tools_BillboardPlant_Scatter::CreatePlant(const unsigned uPlantType)
{
	PolygonID xPolygon;
	bool bCreated = false;
	int iMaxAttempts = 2048;
	while (!bCreated &&
			iMaxAttempts--)
	{
		if (Tools_Polygon_Selection::GetRandomSelectedPolygon(&xPolygon))
		{
			Asura_Vector_3 xPosition;
			GetRandomPositionInPoly(xPolygon, xPosition);
			if (IsPositionOKToCreatePlant(xPosition, uPlantType))
			{
				bCreated = true;
				Tools_BillboardPlant_Object *pxToolsObject = new Tools_BillboardPlant_Object(xPosition, uPlantType);
				s_listCurrentRandom.add_entry(pxToolsObject);
				return pxToolsObject;
			}
		}
		else
			Asura_Assert(0, "Logic check");
	}
	return NULL;
}

bool Tools_BillboardPlant_Scatter::CreateRandomPlant(const unsigned uPlantType)
{
	return CreatePlant(uPlantType) != NULL;
}

bool Tools_BillboardPlant_Scatter::CreateClumpedPlant(const unsigned uPlantType)
{
	bool bGotClumpedPlant = false;
	int iMaxAttempts = 65536;
	while (!bGotClumpedPlant &&
			iMaxAttempts--)
	{
		// To be done - create near another plant.
		Tools_BillboardPlant_Object *pxToolsObject = CreatePlant(uPlantType);
		if (!pxToolsObject)
			return false;
		bGotClumpedPlant = IsPlantClumped(pxToolsObject);
		if (!bGotClumpedPlant)
		{
			s_listCurrentRandom.delete_entry(pxToolsObject);
			pxToolsObject->Destroy();
		}
	}
	if (bGotClumpedPlant)
		return true;
	return false;
}

#endif //ASURA_USE_BILLBOARD_PLANTS
