
//
// Tools_BotanicalCreation_EditLODs.h
//
// Created by Dave Reed - 05/01/2006
// 
// Copyright (C) 2005 Rebellion (Derby), All Rights Reserved.
//
//
// About the LODs:
// ------------------------------------------------------------
//
// Mesh names are named: 'ShapeName_X_Y'
// where X is the element number and Y is the LOD number
// Low LOD number = Low detail. Element and LOD numbers start at 0.
//
// On the LOD graph we have normalised LOD values, from 0.0f to 1.0f:
// 0 == Low Detail --> 1 = High Detail
//
// The LOD params for each LOD component are defined by 4 values:
//
//                 opaque ^            ______
//                        |           /.    .\
//                        |          / .    . \
//                        |         /  .    .  \
//            transparent |________/   .    .   \______
//                        o---------------------------->
//      (low detail/far) 0.0f      Fade0    Fade1     1.0f (high detail/near)
//                                 <--->    <--->
//                                 ^            ^
//                                 |            |
//                                Lod0         Lod1
//
//
//  Lod0...Lod1 - The range of normalized LOD values at which this LOD is visible (including the transitions)
//  Fade0 - Length of the fade in  (as LOD is increased)
//  Fade1 - Length of the fade out (as LOD is increased)
//
//  These values are stored in Tools_BotanicalElementLOD, as:
//      m_fLod0,  m_fLod1
//  and m_fFadeLength0, m_fFadeLength1
//  
//  Within the editor, the 0.0 -> 1.0 range is currently mapped to 100.0 -> 0.0 metres from the camera.
//  (This should really be dependent on screen resolution and/or the performance of the target platform)
//

#ifdef ASURA_USE_PRAGMA_ONCE
#pragma once
#endif

#ifdef ASURA_USE_BOTANICALS

#ifndef __TOOLS_BOTANICALCREATION_BOTANICALSHAPE_H__
#define __TOOLS_BOTANICALCREATION_BOTANICALSHAPE_H__

#include "Camera/Asura_Camera.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_LinkList.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Direct3d.h"
#include "Rendering/Asura_Viewport.h"

#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
#include "BotanicalCreation/Tools_BotanicalCreation_BotanicalStateTable.h"
#endif

class Tools_BotanicalRenderable; 

// Rendering flags
#define TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTRED 1
#define TOOLS_BOTANICALSHAPE_RENDERFLAG_BRIGHTBLUE 2

enum TOOLS_BOTANICALS_ELEMENTTYPE
{
	TOOLS_BOTANICALS_ELEMENTTYPE_1SIDED_MESH=0,		// Single-sided mesh - normal lighting (tree trunks, etc)
	TOOLS_BOTANICALS_ELEMENTTYPE_2SIDED_MESH,		// 2-sided mesh - leaves? - needs different lighting
	TOOLS_BOTANICALS_ELEMENTTYPE_BILLBOARDS,		// Billboard cluster
	TOOLS_BOTANICALS_ELEMENTTYPE_IMPOSTER,			// Set of imposters
};


enum TOOLS_BOTANICALS_LIGHTINGTYPE
{
	TOOLS_BOTANICALS_LIGHTINGTYPE_NORMALS=0,
	TOOLS_BOTANICALS_LIGHTINGTYPE_SPHERICAL,
	TOOLS_BOTANICALS_LIGHTINGTYPE_CYLINDRICAL,
	TOOLS_BOTANICALS_LIGHTINGTYPE_GROUND,
	TOOLS_BOTANICALS_LIGHTINGTYPE_CANOPY,
};



// This is likely to change... For now I'm treating imposters 3D quads, and using their normals
// to decide whether to display them. This will work nicely with tilted trees, and works consistently
// for 'top-down' imposters.

struct Tools_Botanicals_Billboard
{
	Asura_Vector_3 xPos;
	Asura_Vector_3 xNormal;
	float          fWidth;
	float          fHeight;
	float          fU[2];
	float          fV[2];
};


class Tools_Botanicals_Imposter
{
public:
	Asura_Mesh*				   m_pxMesh;
#if defined( ASURA_USE_MATERIAL_RESPONSE_SYSTEM )
	Asura_Hash_ID			   m_uMaterialResponseHashID;
#endif
	int						   m_iTextureIndexID;
	bool					   m_bValid;

	Tools_Botanicals_Billboard m_xBillboard;
	
	float					   m_fAngleRange;	  
	float					   m_fFadeRange;	
};


// A BotanicalElementLOD is a renderable component representing a single LOD level
// ---------------------------------------------------------------------
class Tools_BotanicalElementLOD
{
public:
	Tools_BotanicalElementLOD();
	~Tools_BotanicalElementLOD();

	void Render(TOOLS_BOTANICALS_ELEMENTTYPE eType,float fLOD,Asura_Vector_3* pxPosition, Asura_Matrix_3x3* pxOrientation, float fScale, uint32 uFlags);	
	void AddImposter(Asura_Mesh* pxMesh);
	int  GetNumImposters(void) const;
	Tools_Botanicals_Imposter* GetImposterByIndex(int iIndex);
	const Tools_Botanicals_Imposter* GetImposterByIndexConst(int iIndex) const;
	

//	float										m_fStart;			// LOD value at which the LOD appears
//	float										m_fEnd;				// LOD value at which the LOD disappears
//	float										m_fFadeIn;			// Length of fade-in (in normalized 0-1 range)
//	float										m_fFadeOut;			// Length of fade-out	

	float										m_fLod0;			// Lowest normalized LOD value at which this LOD is visible. Inclusive of transitions
	float										m_fLod1;			// Highest normalized LOD value at which this LOD is visible. Inclusive of transitions
	float										m_fFadeLength0;		// *Length* of 'fade in' transition (at low end of the above LOD range)
	float										m_fFadeLength1;		// *Length* of 'fade out' transition (at high end of the above LOD range)

	bool										m_bLinkToNext;		// When editing, update prev/next LOD to keep them linked
	bool										m_bLinkToPrev;

	TOOLS_BOTANICALS_LIGHTINGTYPE				m_eLightingType;

	Asura_Mesh*									m_pxMesh;
	Tools_BotanicalRenderable*					m_pxRenderable;	

	Asura_LinkList<Tools_Botanicals_Imposter*>	m_xImposterList;	
	Asura_String								m_xName;
};



// A BotanicalElement is a set of LODs for a particular part of the
// shape (e.g. trunk/branches/leaves)
// ---------------------------------------------------------------------
class Tools_BotanicalElement
{
public:	
	Tools_BotanicalElement();
	~Tools_BotanicalElement();

	int											GetNumLODs() const;
	Tools_BotanicalElementLOD*					GetLODByIndex(int iIdx);
	const Tools_BotanicalElementLOD*			GetLODByIndex(int iIdx) const;
	Tools_BotanicalElementLOD*					GetLODByName( const char* pszLODName);
	const Tools_BotanicalElementLOD*			GetLODByName( const char* pszLODName) const;
	void										UpdateRenderables(void);
	
	Asura_LinkList<Tools_BotanicalElementLOD*>  m_xLODList;	
	TOOLS_BOTANICALS_ELEMENTTYPE				m_eType;
	Asura_String								m_xName;
};

// A BotanicalShape is a collection of BotanicalElements
// ---------------------------------------------------------------------
class Tools_BotanicalVersion
{
public:
	Tools_BotanicalElement* FindElement( const char* pszElementName);
	
public:
	Asura_String m_xName;
	Asura_LinkList< Tools_BotanicalElement*> m_xList_pxElements;
	
};

// A BotanicalShape is a collection of BotanicalVersions
// ---------------------------------------------------------------------
class Tools_BotanicalSection
{
public:
	Asura_String m_xName;
	Asura_TemplateHashTree< Tools_BotanicalVersion> m_xHashTree_xVersions;
};


// A BotanicalShape is a collection of BotanicalElements
// ---------------------------------------------------------------------
class Tools_BotanicalShape
{
public:
	Tools_BotanicalShape(char *szName);
	~Tools_BotanicalShape();

	int									    GetNumElements() const;
	
	Tools_BotanicalElement*					GetElementByIndex(int iIdx);
	const Tools_BotanicalElement*			GetElementByIndex(int iIdx) const;
	Tools_BotanicalVersion*					GetVersionByIndex(int iIdx);
	const Tools_BotanicalVersion*			GetVersionByIndex(int iIdx) const;
	Tools_BotanicalSection*					GetSectionByIndex(int iIdx);
	const Tools_BotanicalSection*			GetSectionByIndex(int iIdx) const;
	
	void									SetDefaultLODs();
	bool									ContainsMesh(Asura_Mesh *pxMesh);
	void									DiscardMeshes(void);
	void									UpdateRenderables(void);	
	void									Render(Asura_Vector_3 *pxPosition, Asura_Matrix_3x3 *pxOrientation, float fScale, uint32 uFlags=0);
	const char*							    GetName() const {return m_xName.GetString();};

#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	bool									ValidateStateTable();
#endif
	
	Asura_TemplateHashTree< Tools_BotanicalSection>	 m_xHashTree_xSections;
	
#if defined( ASURA_BOTANICAL_SUPPORT_STATES) && 1 == ASURA_BOTANICAL_SUPPORT_STATES
	Asura_LinkList< Tools_Botanical_StateTable1D> m_xList_xStateTables1D;
#endif

	float									 m_fInvMaxDrawDistance;
	// DLH - At present these values are based on the Horizontal FOV and Res as that seems to be the controlling/master information in the camera system 
	float									 m_fMaxDrawDistanceFOV;
	uint32									 m_uMaxDrawDistanceRes;
	
	
#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	bool			m_bIsInteractive;
	Asura_Vector_3	m_xInteractiveBone0;
	Asura_Vector_3	m_xInteractiveBone1;
	
	float m_fMass;
	float m_fDrag;
	float m_fSpringStrength;
	float m_fSpringLength;
	float m_fBoneWeightGamma;
	bool  m_bUsePaintedColoursAsWeights;
#endif

private:	
	Asura_String							 m_xName;
};

inline Tools_BotanicalShape::Tools_BotanicalShape(char *szName) 
{
	this->m_xName=szName;	
//	this->m_fInvMaxDrawDistance = 1.0f / 100.0f;
//	this->m_fMaxDrawDistanceFOV = -90;
//	this->m_uMaxDrawDistanceRes = 0;
	
	this->m_fInvMaxDrawDistance = 0.01f;
	this->m_fMaxDrawDistanceFOV = Asura_Camera::GetHorizontalFieldOfView();
	this->m_uMaxDrawDistanceRes = Asura_Viewport::GetWidth();

#if defined( ASURA_BOTANICAL_SUPPORT_INTERACTIVE)
	this->m_bIsInteractive = false;
	this->m_xInteractiveBone0.Zero();
	this->m_xInteractiveBone1 = Asura_Vector_3( 0, -4.0f, 0);
	
	this->m_fMass = 1.0f;
	this->m_fDrag = 1.0f;
	this->m_fSpringStrength = 1.0f;
	this->m_fSpringLength = 100.0f;
	this->m_fBoneWeightGamma = 1.0f;
	this->m_bUsePaintedColoursAsWeights = false;
#endif	
	
}


// This class contains functions for dealing with Tools_BotanicalShapes
// ---------------------------------------------------------------------
class Tools_BotanicalCreation_BotanicalShape
{
public:
	//static Tools_BotanicalShape* CreateTestData();
};



#endif // __TOOLS_BOTANICALCREATION_BOTANICALSHAPE_H__
#endif // ASURA_USE_BOTANICALS
