//
// Tools_Cutscene_Subtitles.cpp
//
// Created by Richard Rice - 23/08/2005
// 
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//


#include "Core/Asura.h"
#include "Cutscene/Tools_Cutscene_Subtitles.h"

#include "Core/Asura_HashTree.h"
#include "Core/Asura_HashedLocalisedText.h"
#include "FileAccess/Asura_FileAccess.h"

#include "Misc/Tools_IniFile.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Modal_Overlay.h"

Asura_TemplateHashTree<Asura_String>* Tools_Cutscene_Subtitles::s_pxSubtitleTree;


void Tools_Cutscene_Subtitles::Initialise()
{
	s_pxSubtitleTree = new Asura_TemplateHashTree<Asura_String>;
	LoadSubtitleIDs();
}

void Tools_Cutscene_Subtitles::Shutdown()
{
	delete s_pxSubtitleTree;
	s_pxSubtitleTree = 0;
}


static void SkipToNextLine(const Asura_UnicodeChar*& pwUnicodeText)
{
	//skip to end of line
	bool bInQuotes = false;
	bool bDone = false;
	bool bCheckForEnteringQuotes = true;
	while(*pwUnicodeText && !bDone)
	{
		if(bInQuotes)
		{
			if(*pwUnicodeText == L'\"')
			{
				bInQuotes = false;
			}
			pwUnicodeText++;
				
		}
		else
		{
			if(*pwUnicodeText == L'\r' || *pwUnicodeText == L'\n')
			{
				bDone = true;
			}
			else if(*pwUnicodeText == L'\t')
			{
				pwUnicodeText++;
				bCheckForEnteringQuotes = true;
			}
			else if(*pwUnicodeText == L'\"' && bCheckForEnteringQuotes)
			{
				//entering quotes.
				//Note that this only counts as entering quotes if at a start of an entry , since in some cases
				//there may be a single " in the middle of a non-quoted string.
				bInQuotes = true;
				pwUnicodeText++;
			}
			else
			{
				pwUnicodeText++;
				bCheckForEnteringQuotes = false;
			}
		}
	}

	//and to start of next
	if(*pwUnicodeText == L'\r') pwUnicodeText++;
	if(*pwUnicodeText == L'\n') pwUnicodeText++;
}

void Tools_Cutscene_Subtitles::LoadSubtitleIDs()
{
	//open the ini file indicating the location of the unicode file
	Tools_IniFile xIni("ToolsIniFiles\\Cutscene.ini");
	if(!xIni.LoadIniFile())
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to load Cutscene.ini.");
		return;
	}

	Asura_String xFilename;
	if(!xIni.GetStringInCurrentSection("CutsceneText",xFilename))
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to find CutsceneText entry in Cutscene.ini");
		return;
	}


	char szNetFilename[MAX_PATH];
	Asura_FileAccess::PlatformGetNetworkFilename(xFilename.GetString(),szNetFilename,MAX_PATH);


	//attempt to load the unicode file
	HANDLE hFile = CreateFile(szNetFilename,GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS,0);
	if(hFile == INVALID_HANDLE_VALUE)
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to load %s",xFilename.GetString());
		return;
	}

	//read the file contents into a buffer
	int iFileSize;
	unsigned long uBytesRead;


	iFileSize = GetFileSize(hFile,0);
	char* pcBuffer = new char[iFileSize+2];
	*(Asura_UnicodeChar*)&pcBuffer[iFileSize] = L'\0';

	ReadFile(hFile,pcBuffer,iFileSize,&uBytesRead,0);
	CloseHandle(hFile);

	const Asura_UnicodeChar* pwUnicodeText = (Asura_UnicodeChar*) pcBuffer;

	//make sure this is unicode
	if(*pwUnicodeText != 0xFEFF)
	{
		delete [] pcBuffer;
		Asura_Assertf(0,("Cutscene text file '%s' isn't unicode",xFilename.GetString()));
		Tools_Timed_Message_System::AddMessage(0xff0000,"%s isn't unicode",xFilename.GetString());
		return;
	}

	//first two ines don't contain anything of interest
	SkipToNextLine(pwUnicodeText);
	SkipToNextLine(pwUnicodeText);

	//get the identifier at the start of each line
	Asura_UnicodeChar wszUnicodeID[200];
	char szID[200];
	for(;*pwUnicodeText;SkipToNextLine(pwUnicodeText))
	{
		//copy ID until tab/end of line
		Asura_UnicodeChar* pwCopyTo = wszUnicodeID;
		bool bInQuotes = false;
		bool bDone = false;
		bool bCheckForEnteringQuotes = true;

		while(*pwUnicodeText && !bDone)
		{
			if((pwCopyTo - wszUnicodeID) > 195)
			{
				Asura_Assert(0,"Unexpectedly long identifer while reading cutscene text IDs");
				break;
			}
			if(bInQuotes)
			{
				//while in quote we leave all text as is , until we exit the quotes
				if(*pwUnicodeText == L'\"')
				{
					if(pwUnicodeText[1] == L'\"')
					{
						//double quotes appear as pairs in the file , skip the first one , and copy the second
						pwUnicodeText++;
						*pwCopyTo++ = *pwUnicodeText++;
					}
					else
					{
						pwUnicodeText++;
						bInQuotes = false;
					}
				}
				else
				{
					*pwCopyTo++ = *pwUnicodeText++;
				}
			}
			else
			{
				if(*pwUnicodeText == L'\r' || *pwUnicodeText == L'\n' || *pwUnicodeText == L'\t')
				{
					bDone = true;
				}
				else if(*pwUnicodeText == L'\"' && bCheckForEnteringQuotes)
				{
					//entering quotes
					bInQuotes = true;
					pwUnicodeText++;
				}
				else
				{
					*pwCopyTo++ = *pwUnicodeText++;
				}
				bCheckForEnteringQuotes = false;
			}
		}
		//add terminator
		*pwCopyTo++ = 0;

		//convert to ascii
		sprintf(szID,"%ls",wszUnicodeID);

		if(szID[0])
		{
			//add to the hash tree
			Asura_String* pxData = s_pxSubtitleTree->AddNode(Asura_GetHashID(szID));
			Asura_Assertf(pxData, ("Unexpected NULL pointer. Does your subtitle list have a duplicate ID entry(%s)?", szID));
			if (pxData)
			{
                *pxData = szID;
			}
		}
					
	}

	delete [] pcBuffer;
}


static ListSelection_Overlay* g_pxListOverlay;
static Asura_Hash_ID g_uInitialTextHash;

static void AddToOverlayCallback(Asura_Hash_ID uHash,Asura_String* pxString)
{
	//get the full text as a help string
	const Asura_UnicodeChar* wszText = Asura_HashedLocalisedText::GetString(uASURA_HASHEDTEXTPAGE_CUTSCENE, uHash);

	char szHelpText[500] = "";
	if(wszText)
	{
		Asura_PrintfToBfr(szHelpText,"%S",500,true,wszText);
	}

	g_pxListOverlay->AddItem(pxString->GetString(),uHash,uHash==g_uInitialTextHash,szHelpText);
}

Asura_Hash_ID Tools_Cutscene_Subtitles::SelectSubtitle(Asura_Hash_ID uInitialTextHash)
{
	Asura_Assert(s_pxSubtitleTree,"Not Initialised");
	if(!s_pxSubtitleTree || !s_pxSubtitleTree->GetNumberOfNodes())
	{
		ShowMessageOverlay("Can't select subtitle , since no subtitle IDs were loaded.");
		return uInitialTextHash;
	}

	//add all the subtitle IDs to overlay
	ListSelection_Overlay xOverlay("Select Subtitle");

	g_pxListOverlay = &xOverlay;
	g_uInitialTextHash = uInitialTextHash;

	s_pxSubtitleTree->Traverse(AddToOverlayCallback);

	if(!xOverlay.DoModal()) return uInitialTextHash;

	return (Asura_Hash_ID) xOverlay.GetSelectionData();

}


const char* Tools_Cutscene_Subtitles::GetSubtitleID(Asura_Hash_ID uTextHash)
{
	Asura_Assert(s_pxSubtitleTree,"Not Initialised");
	if(uTextHash == 0) return "*None";

	if(s_pxSubtitleTree)
	{
		Asura_String* pxData = s_pxSubtitleTree->GetDataByHash(uTextHash);
		if(pxData)
		{
			return pxData->GetString();
		}
	}



	//failed to find string
	return "*Unknown*";
}
