//
// Tools_WavWriter.cpp
//
// Created by Richard Rice - 19/11/2002
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Debugging/Asura_Assert.h"

#include "Tools_WavWriter.h"
#include "Misc/Tools_Modal_Overlay.h"

#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
#include "FileAccess/Asura_FileAccess.h"
#endif


#define FAIL_WAV \
{\
	if(m_hOutputFile!=INVALID_HANDLE_VALUE)CloseHandle(m_hOutputFile);\
   	ShowMessageOverlay("Error writing file");\
	return false;\
}

#define WRITE_WAV(pBuffer,iSize)\
WriteFile(m_hOutputFile,(long*)pBuffer,(iSize),&uBytesWritten,0);\
if(uBytesWritten!=(unsigned)(iSize))\
	FAIL_WAV

typedef struct pwavchunkheader
{
	char chunkName[4];
	int chunkLength;	
} PWAVCHUNKHEADER;

typedef struct pwavriffheader
{
	char type[4];
} PWAVRIFFHEADER;

///////////////////////////////////////////////////////////////////////////////////////////////////////////////


typedef short ToolsMonoWavSample;

struct InputWavFile
{
	InputWavFile(int iStartSample,int iLastSample,const InputWavFileProp* pxProp);
	~InputWavFile();

	bool OpenWavFile(const char* pszFilename);
	void AddToOutput(Tools_WavWriter* pxOutput,int iSamplesToWrite);
	ToolsMonoWavSample GetSample(Tools_WavWriter* pxOutput,int iSampleIndex);

	const Asura_Vector_3* m_pxLocation;
	int m_iStartSample;
	int m_iLastSample;
	float m_fInnerRange;
	float m_fOuterRange;
	float m_fVolume;
	float m_fPitch;
	float m_fDuration;
	bool  m_bLooped;
	PCMWAVEFORMAT m_xWavFormat;
	
	int m_iBufferSize;
	ToolsMonoWavSample* m_psSampleBuffer;
	

	bool m_bFinished;
	
};


///////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------**
** Tools_WavWriter **
**-----------------*/

Tools_WavWriter::Tools_WavWriter()
{
	m_hOutputFile = INVALID_HANDLE_VALUE;
	m_iDataOffset = 0;
	m_fCurrentTime = 0;
	m_iBufferSize = 0;
	m_pxSampleBuffer = 0;
	m_iSamplesWritten = 0;
	m_iLastNonZeroSample = -1;

	//fill in default values for the wav format
	m_xWavFormat.wf.wFormatTag = 1 ;
	m_xWavFormat.wf.nChannels = 2;
	m_xWavFormat.wBitsPerSample = 16;
	m_xWavFormat.wf.nBlockAlign = (unsigned short)((m_xWavFormat.wf.nChannels * m_xWavFormat.wBitsPerSample)/8);
	m_xWavFormat.wf.nSamplesPerSec = 44100;
	m_xWavFormat.wf.nAvgBytesPerSec = m_xWavFormat.wf.nBlockAlign * m_xWavFormat.wf.nSamplesPerSec;

}

Tools_WavWriter::~Tools_WavWriter()
{
	if(m_hOutputFile != INVALID_HANDLE_VALUE)
	{
		CloseHandle(m_hOutputFile);
	}
	delete [] m_pxSampleBuffer;

	while(m_xWavList.size())
	{
		delete m_xWavList.first_entry();
		m_xWavList.delete_first_entry();
	}
}

bool Tools_WavWriter::StartOutput(const char* pszFilename)
{
 	unsigned long uBytesWritten;
	if(m_hOutputFile != INVALID_HANDLE_VALUE)
	{
		return false;
	}

	//initialise some values...
	m_iDataOffset = 0;
	m_fCurrentTime = 0;
	m_iSamplesWritten = 0;
	m_iLastNonZeroSample = -1;

	
	//open the output file
	m_hOutputFile = CreateFile(pszFilename,GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_FLAG_RANDOM_ACCESS, 0);
	if(m_hOutputFile == INVALID_HANDLE_VALUE)
	{
		ShowMessageOverlay("Failed to create %s",pszFilename);
		return FALSE;
	}

	//create header and write it
	PWAVCHUNKHEADER xChunkHeader = {{'R','I','F','F'},0};
	WRITE_WAV(&xChunkHeader,sizeof(xChunkHeader));

	//write RiffHeader
	PWAVRIFFHEADER xRiffHeader = {{'W','A','V','E'}};
	WRITE_WAV(&xRiffHeader,sizeof(xRiffHeader));

	//write the chunk header , and format
	xChunkHeader.chunkName[0] = 'f'; 
	xChunkHeader.chunkName[1] = 'm'; 
	xChunkHeader.chunkName[2] = 't'; 
	xChunkHeader.chunkName[3] = ' '; 
	xChunkHeader.chunkLength = sizeof(PCMWAVEFORMAT); 
	WRITE_WAV(&xChunkHeader,sizeof(xChunkHeader));
	WRITE_WAV(&m_xWavFormat,sizeof(m_xWavFormat));

	//write the header for the data block , and store the offset to the data chunk 
	m_iDataOffset = GetFileSize(m_hOutputFile,0);
	if(!m_iDataOffset) FAIL_WAV;

	xChunkHeader.chunkName[0] = 'd'; 
	xChunkHeader.chunkName[1] = 'a'; 
	xChunkHeader.chunkName[2] = 't'; 
	xChunkHeader.chunkName[3] = 'a'; 
	xChunkHeader.chunkLength = 0; 
	WRITE_WAV(&xChunkHeader,sizeof(xChunkHeader));

	return true;
}

bool Tools_WavWriter::UpdateOutput(float fTime,const Asura_Vector_3* pxLocation,const Asura_Matrix_3x3* pxOrient)
{
	Asura_Assert(pxLocation,"");
	Asura_Assert(pxOrient,"");

 	unsigned long uBytesWritten;

	if(m_hOutputFile == INVALID_HANDLE_VALUE)
	{
		return false;
	}
	Asura_Assert(fTime>=m_fCurrentTime,"");

	m_xLocation = *pxLocation;
	m_xInverseOrient = *pxOrient;

	m_xInverseOrient.Transpose();

	//determine the number of samples to write this frame
	int iSampleStart = (int)(m_fCurrentTime*m_xWavFormat.wf.nSamplesPerSec);
	int iSampleEnd = (int)(fTime*m_xWavFormat.wf.nSamplesPerSec);

	int iSamplesToWrite = iSampleEnd - iSampleStart;

	//enlarge buffer if needed
	if(m_iBufferSize<iSamplesToWrite)
	{
		delete [] m_pxSampleBuffer;
		m_iBufferSize = iSamplesToWrite+10;
		m_pxSampleBuffer = new ToolsWavSample[m_iBufferSize];
	}

	//zero output
	memset(m_pxSampleBuffer,0,iSamplesToWrite*sizeof(ToolsWavSample));

	//add in all the current samples...
	for(LIF<InputWavFile*> xWavLIF(&m_xWavList);!xWavLIF.done();)
	{
		xWavLIF()->AddToOutput(this,iSamplesToWrite);

		if(xWavLIF()->m_bFinished)
		{
			delete xWavLIF();
			xWavLIF.delete_current();
		}
		else
		{
			xWavLIF.next();
		}
	}

	//check for last non zero sample
	for(int i=iSamplesToWrite-1;i>=0;i--)
	{
		if(m_pxSampleBuffer[i].m_sLeft || m_pxSampleBuffer[i].m_sRight)
		{
			m_iLastNonZeroSample = m_iSamplesWritten+i;
			break;
		}
	}


	//write this frames samples to the file
	WRITE_WAV(m_pxSampleBuffer,iSamplesToWrite*sizeof(ToolsWavSample));

	m_iSamplesWritten += iSamplesToWrite;
	m_fCurrentTime = fTime;

	return true;
}


bool Tools_WavWriter::FinishOutput()
{
 	unsigned long uBytesWritten;
	if(m_hOutputFile == INVALID_HANDLE_VALUE)
	{
		return false;
	}
	int iLength;

	#if 0
	//truncate file to get rid of all zero samples at end
	long iTruncationAmount = (m_iSamplesWritten - (m_iLastNonZeroSample+1)) * sizeof(ToolsWavSample);

	if(iTruncationAmount)
	{
		SetFilePointer(m_hOutputFile,-iTruncationAmount,0,FILE_END);
		//set the new end of file
		SetEndOfFile(m_hOutputFile);

		//get the new number of samples written
		m_iSamplesWritten = m_iLastNonZeroSample+1;
	
	}
	#endif
	
	//get the file size
	iLength = GetFileSize(m_hOutputFile,0);
	if(!iLength) FAIL_WAV;
	iLength -= sizeof(PWAVCHUNKHEADER);
	
	//seek to RIFF chunk to write file size minus header size
	SetFilePointer(m_hOutputFile,4,0,FILE_BEGIN);
	WRITE_WAV(&iLength,sizeof(int));

	//seek to data chunk to write data size
	SetFilePointer(m_hOutputFile,m_iDataOffset+4,0,FILE_BEGIN);
	iLength = m_iSamplesWritten*m_xWavFormat.wf.nBlockAlign;
	WRITE_WAV(&iLength,sizeof(int));

	CloseHandle(m_hOutputFile);
	m_hOutputFile = INVALID_HANDLE_VALUE;

	while(m_xWavList.size())
	{
		delete m_xWavList.first_entry();
		m_xWavList.delete_first_entry();
	}
	
	return true;
}

void Tools_WavWriter::AddInputWavFile(const char* pszFilename,float fStartTime,const InputWavFileProp* pxProp)
{
	Asura_Assert(pxProp,"");

	if(m_hOutputFile == INVALID_HANDLE_VALUE)
	{
		return;
	}
	
	int iSampleStart = (int)(fStartTime*m_xWavFormat.wf.nSamplesPerSec);
	int iSampleEnd = 0;
	if(pxProp->m_fDuration)
	{
		float fEndTime = fStartTime + pxProp->m_fDuration;
		iSampleEnd =  (int)(fEndTime*m_xWavFormat.wf.nSamplesPerSec);
	}
	
	//create a new InputWavFile
	InputWavFile*  pxWavFile = new InputWavFile(iSampleStart,iSampleEnd,pxProp);
	//open the file
	if(!pxWavFile->OpenWavFile(pszFilename))
	{
		delete pxWavFile;
		return;
	}

	//add to the list of inputs
	m_xWavList.add_entry(pxWavFile);
			
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------**
** InputWavFile **
**--------------*/

#define FAIL_WAV_READ \
{\
	if(hInputFile!=INVALID_HANDLE_VALUE)CloseHandle(hInputFile);\
	ShowMessageOverlay("Error reading file '%s'",pszFilename);\
	return false;\
}

#define READ_WAV(pBuffer,iSize)\
ReadFile(hInputFile,(long*)pBuffer,(iSize),&uBytesRead,0);\
if(uBytesRead!=(unsigned)(iSize))\
	FAIL_WAV_READ

InputWavFile::InputWavFile(int iStartSample,int iLastSample,const InputWavFileProp* pxProp)
{
	m_iStartSample = iStartSample;	
	m_iLastSample = iLastSample;
	m_pxLocation = pxProp->m_pxLoc;
	m_iBufferSize = 0;
	m_psSampleBuffer = 0;
	m_bFinished = false;
	m_fInnerRange = pxProp->m_fInnerRange;
	m_fOuterRange = pxProp->m_fOuterRange;
	m_fVolume = pxProp->m_fVolume;
	m_fPitch = pxProp->m_fPitch;
	m_bLooped = pxProp->m_bLooped;
	m_fDuration = pxProp->m_fDuration;
}

InputWavFile::~InputWavFile()
{
	delete [] m_psSampleBuffer;
}

bool InputWavFile::OpenWavFile(const char* pszFilename)
{
	unsigned long uBytesRead;

	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char szFilename[MAX_PATH];

	Asura_FileAccess::PlatformGetNetworkFilename(pszFilename,szFilename,MAX_PATH);
	pszFilename = szFilename;
	#endif

	
	//open the input file
	HANDLE hInputFile = CreateFile(pszFilename,GENERIC_READ,0,0,OPEN_EXISTING,FILE_FLAG_RANDOM_ACCESS,0);
	
	if(hInputFile == INVALID_HANDLE_VALUE)
	{
		ShowMessageOverlay("Error Opening %s",pszFilename);
		return false;
	}

	//read the initial headers
	PWAVCHUNKHEADER xChunkHeader;
	READ_WAV(&xChunkHeader,sizeof(xChunkHeader));

	if((xChunkHeader.chunkName[0]!='R')||(xChunkHeader.chunkName[1]!='I')||
	   (xChunkHeader.chunkName[2]!='F')||(xChunkHeader.chunkName[3]!='F'))
	{
		FAIL_WAV_READ;
	}

	PWAVRIFFHEADER xRiffHeader;
	READ_WAV(&xRiffHeader,sizeof(xRiffHeader));

	if(xRiffHeader.type[0]!='W'||xRiffHeader.type[1]!='A'||xRiffHeader.type[2]!='V'||xRiffHeader.type[3]!='E')
	{
		FAIL_WAV_READ;
	}

	//find the format chunk
	do
	{
		READ_WAV(&xChunkHeader,sizeof(xChunkHeader));
		if((xChunkHeader.chunkName[0]=='f')&&(xChunkHeader.chunkName[1]=='m')&&
	   		(xChunkHeader.chunkName[2]=='t')&&(xChunkHeader.chunkName[3]==' '))
		{
			break;
		}

		SetFilePointer(hInputFile,xChunkHeader.chunkLength,0,FILE_CURRENT);

	}while(TRUE);


	//read the wav format chunk
	if(xChunkHeader.chunkLength==16)
	{
		/* a standard PCM wave format chunk */
		READ_WAV(&m_xWavFormat,sizeof(m_xWavFormat));
	}
	else if(xChunkHeader.chunkLength==18)
	{
		/* an extended PCM wave format chunk */
		WAVEFORMATEX xTemp;
		READ_WAV(&xTemp,sizeof(xTemp));
		
		m_xWavFormat = *(PCMWAVEFORMAT*)&xTemp;
	}
	else
	{
		ShowMessageOverlay("Unexpected size for format chunk in wav file %s",pszFilename);
		FAIL_WAV_READ;
	}

	
	//now skip until we reach the data chunk
	do
	{
		READ_WAV(&xChunkHeader,sizeof(xChunkHeader));
		if((xChunkHeader.chunkName[0]=='d')&&(xChunkHeader.chunkName[1]=='a')&&
	   		(xChunkHeader.chunkName[2]=='t')&&(xChunkHeader.chunkName[3]=='a'))
		{
			break;
		}

		SetFilePointer(hInputFile,xChunkHeader.chunkLength,0,FILE_CURRENT);

	}while(TRUE);

	unsigned char* pucTempBuffer = new unsigned char[xChunkHeader.chunkLength];
			
	//read in the data chunk
	READ_WAV(pucTempBuffer,xChunkHeader.chunkLength);

	//how many samples?
	m_iBufferSize = xChunkHeader.chunkLength/m_xWavFormat.wf.nBlockAlign;
	m_psSampleBuffer = new ToolsMonoWavSample[m_iBufferSize];

	//copy and convert from temporary buffer

	if(m_xWavFormat.wBitsPerSample == 8)
	{
		unsigned char* pucFrom = (unsigned char*)pucTempBuffer;
		ToolsMonoWavSample* psTo = m_psSampleBuffer;

		for(int i=0;i<m_iBufferSize;i++)
		{
			int iTemp = 0;
			for(int j=0;j<m_xWavFormat.wf.nChannels;j++)
			{
				iTemp += *pucFrom++;
			}	
			iTemp*=256;
			iTemp/=m_xWavFormat.wf.nChannels;
			iTemp -= 32768;
			if(iTemp>0x7fff) iTemp = 0x7fff;
			if(iTemp<-0x7fff)iTemp = -0x7fff;
		
			*psTo++ = (short)iTemp;
		}
	}
	else if(m_xWavFormat.wBitsPerSample == 16)
	{
		short* psFrom = (short*)pucTempBuffer;
		ToolsMonoWavSample* psTo = m_psSampleBuffer;

		for(int i=0;i<m_iBufferSize;i++)
		{
			int iTemp = 0;
			for(int j=0;j<m_xWavFormat.wf.nChannels;j++)
			{
				iTemp += *psFrom++;
			}	
			*psTo++ = (short)(iTemp/m_xWavFormat.wf.nChannels);
		}
	}
	else
	{
		delete [] pucTempBuffer;
		FAIL_WAV_READ;
	}

	//get rid of the temp buffer
	delete [] pucTempBuffer;
	


	//finally close the file
	CloseHandle(hInputFile);
	return true;

}


ToolsMonoWavSample InputWavFile::GetSample(Tools_WavWriter* pxOutput,int iSampleIndex)
{
	ToolsMonoWavSample sSample = 0;

	if(iSampleIndex<m_iStartSample)
	{
		//haven't reached the start of the sound yet
		return sSample;
	}
	if(m_iLastSample && iSampleIndex>=m_iLastSample)
	{
		m_bFinished = true;
		return sSample;
	}

	int iSampleOffset = (iSampleIndex-m_iStartSample);
	//apply pitch shift...
	if(m_fPitch!=1.0f)
	{
		iSampleOffset = (int) (iSampleOffset * m_fPitch);
	}

	if(m_xWavFormat.wf.nSamplesPerSec != pxOutput->m_xWavFormat.wf.nSamplesPerSec)
	{
		iSampleOffset = (int)(((float)iSampleOffset*m_xWavFormat.wf.nSamplesPerSec)/pxOutput->m_xWavFormat.wf.nSamplesPerSec);
	}

	if(iSampleOffset>=m_iBufferSize)
	{
		if(m_bLooped)
		{
			iSampleOffset%=m_iBufferSize;
		}
		else
		{
			//beyond the end of the file
			m_bFinished = TRUE;
			return sSample;
		}
	}

	return m_psSampleBuffer[iSampleOffset];
}

void InputWavFile::AddToOutput(Tools_WavWriter* pxOutput,int iSamplesToWrite)
{
	if(m_bFinished) return;
		
	int iSampleIndex = pxOutput->m_iSamplesWritten;
	

	float fDistScale = 1.0f;
	float fPan = 0;

	if(m_pxLocation)
	{
		//determine poaition relative to player
		Asura_Vector_3 xOffset = *m_pxLocation - pxOutput->m_xLocation;

		xOffset = xOffset * pxOutput->m_xInverseOrient;

		//determine distance , and work work out scaling caused by it
		float fDistance = xOffset.Magnitude();
		
		fDistScale = 1.0f;
		if(fDistance>=m_fOuterRange)
		{
			fDistScale = 0;
		}
		else if(fDistance>m_fInnerRange)
		{
			fDistScale = (m_fOuterRange-fDistance)/(m_fOuterRange-m_fInnerRange);
		}

		//if not to close then apply panning
		if(fDistance>1.0f)
		{
			xOffset.Normalise();
			fPan = xOffset.x;
		}
	}

	float fLeftScale = min((1.0f - fPan),1.0f);
	float fRightScale = min((1.0f + fPan),1.0f);
	
	
	float fScale = fDistScale* m_fVolume;
	
	fLeftScale *= fScale;
	fRightScale *= fScale;

	
	

	ToolsWavSample* pxOutputBuffer = pxOutput->m_pxSampleBuffer;

	
	for(int i=0;i<iSamplesToWrite;i++,iSampleIndex++,pxOutputBuffer++)
	{
		ToolsMonoWavSample sSample = GetSample(pxOutput,iSampleIndex);
		if(!sSample) continue;

		int iTemp;

		iTemp = pxOutputBuffer->m_sLeft + (int)(fLeftScale*sSample);
		if(iTemp>0x7fff)
		{
			iTemp = 0x7fff;
		}
		if(iTemp<-0x7fff)
		{
			iTemp = -0x7fff;
		}
		pxOutputBuffer->m_sLeft = (short)iTemp;		
		
		iTemp = pxOutputBuffer->m_sRight + (int)(fRightScale*sSample);
		if(iTemp>0x7fff)
		{
			iTemp = 0x7fff;
		}
		if(iTemp<-0x7fff)
		{
			iTemp = -0x7fff;
		}
		pxOutputBuffer->m_sRight = (short)iTemp;		
	}	
}

/////////////////////////////////////////////////////////////////////////////////////////////////
/*------------------**
** InputWavFileProp **
**------------------*/

InputWavFileProp::InputWavFileProp()
{
	m_pxLoc = 0;
	m_fInnerRange = 10.0f;
	m_fOuterRange = 20.0f;
	m_fVolume = 1.0f;
	m_fPitch = 1.0f;
	m_bLooped = false;
	m_fDuration = 0;
}
