///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Container.h
//
// Created by Martin G.Harris (6/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_CONTAINER_H__
#define __TOOLS_CONTAINER_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Asura_StringBuffer.h"
#include "Entity/Container/Asura_Container.h"

#include "Misc/list_tem.hpp"

#include "SpecialFX/PFX/Asura_PFX_Volume.h"

// Forward declarations
class Asura_Overlay;
class Asura_Tabs_Dialog;
class Asura_Chunk_Stream;
class Tools_ContainerGroup;
class Tools_ContainerFile;
class Tools_Container;
class ListSelection_Overlay;
class Tools_ContainerStateMachine;
class Tools_Container_PFX_SourceVolume_Object;
class Tools_Container_LimitingPlane_Object;

// Constants
// Maximum depth allowed for a container instance.
const int iASURA_MAX_CONTAINER_DEPTH = 16;


///////////////////////////////////
// Tools_Container_ChildIterator //
///////////////////////////////////
// Iterator for iterating over children in a Tools_Container.
// Should work for both attachments and collections, as well as
// any project specified containers that have children.
// The order of the children is arbitrary.
class Tools_Container_ChildIterator
{
public:
	Tools_Container_ChildIterator(const Tools_Container* pxContainer);
	~Tools_Container_ChildIterator();

	// Total number of children (including any repeated ones). Doesn't recurse.
	int GetNumChildren() const { return m_iTotalNumChildren;}

	// Move to first entry.
	void Begin() { m_iCurrentEntry = 0; m_iCurrentRepeat = 0; }

	// Move past last entry (Done will return true).
	void End()   { m_iCurrentEntry = m_iNumArrayEntries; m_iCurrentRepeat = 0; }

	// Finished?
	bool Done() const { return (m_iCurrentEntry >= m_iNumArrayEntries); }

	void Next();  // Move to next entry.
	void NextSkipIfSame(); // Move to next entry, but skip if hash is same as previous value.
  
	Asura_Hash_ID GetCurrentHash() const; // Return current hash
	Tools_Container* GetCurrent() const;  // Return current container or NULL if not found.

	// If there are no children then add an entry with ASURA_HASH_ID_UNSET.
	void AddUnsetEntryIfNoChildren();

	// Initialisation.
	void SetArraySize(int iMaxSize);
	void AddEntry(Asura_Hash_ID uChildHash, int iMultiplicity = 1);

	// Is the specified container a child?
	bool IsMember(Asura_Hash_ID uContainerHash) const;

protected:
	void Reset();

private:
	// Prevent copying.
	Tools_Container_ChildIterator(const Tools_Container_ChildIterator&);
	Tools_Container_ChildIterator& operator=(const Tools_Container_ChildIterator&);

protected:
	int m_iTotalNumChildren;
	int m_iMaxArraySize;
	int m_iNumArrayEntries;

	Asura_Hash_ID* m_puHash;
	int* m_piMultiplicity;

	int m_iCurrentEntry;
	int m_iCurrentRepeat;
};


/////////////////////
// Tools_Container //
/////////////////////
// Base class for all the Tools versions of the container classes.
class Tools_Container
{
public:
	Tools_Container();
	virtual ~Tools_Container() {}

	// Accessors
	u_int GetClassification() const {return m_uClassification; }

	Asura_Hash_ID GetFileHash() const		{ return m_uFileHash; }
	Asura_Hash_ID GetGroupHash() const		{ return m_uGroupHash; }
	Asura_Hash_ID GetContainerHash() const	{ return m_uContainerHash; }

	const Asura_String& GetContainerName() const { return m_xContainerName; }

	const char* GetDescription() const;

	// Only used when the container has been moved to a new file/group.
	void SetFileHash(Asura_Hash_ID uFileHash) { m_uFileHash = uFileHash; }
	void SetGroupHash(Asura_Hash_ID uGroupHash) { m_uGroupHash = uGroupHash; }

	// Flag this container as having been modified.
	// Set bMinorChange if there is no need to rebuild the instance.
	void SetModified(bool bMinorChange = false);
	
	// The state machine has been modified.
	virtual void SetModifiedStates() {}
	// Copy child hashes back to Asura for collections.
	virtual void CopyChildHashDataToAsura() {}

	// Overlay related functions.
	void CreateOverlay(bool bSortToFront);
	void DestroyOverlay();
	void ResizeOverlay();

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset) = 0;
	virtual void PerFrameOverlayUpdate() {}

	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;
	
	// Copy data from a source container of the same classification.
	void CopyDataFromContainer(const Tools_Container* pxSourceContainer);

	virtual void PostLoadInitialise() {}	// Called after all containers have been loaded.
	virtual void PostCreateInitialise() {}	// Called after new container has been created in tools, but before it is registered.
	virtual void CopyDataToAsura() {}		// Copy data to corresponding Asura_Container.
	virtual void Verify(bool bRecurse) const;	// Check for errors.
	virtual void FixUpErrors();				// Fix up certain types of error, automatically.

	// Get a short name to be used in the instance overlay (e.g. skin name).
	virtual const char* GetInstanceName() const { return 0; }

	// Render any Tools-only stuff.
	virtual void ToolsRender(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri) const {}

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

	// Build a list of textures for texture report log.
	virtual void TraverseUsedTexturesForReport(void (*pfnCallback)(int iTextureID,void* pUserData), void* pUserData) const;

	// Helper. Find texture and add it to the export list.
	void AddTextureToExportList(Asura_Hash_ID uTextureHash, const Asura_String& xTextureName) const;
	// Helper. Load required texture file.
	void LoadRequiredTextureFile(Asura_Hash_ID uTextureHash, const Asura_String& xTextureName) const;
	// Helper. Verify texture.
	void VerifyTexture(Asura_Hash_ID uTextureHash, const Asura_String& xTextureName, const char* szDescription, bool bMustBeSet = true) const;

	// Determine the depth of the tree, stopping at max value. Used to check for circular references.
	int GetDepth(const int iMax) const;

	// Recursive check to see whether this contains (or equals) the specified container.
	bool Contains(Asura_Hash_ID uSearchHash) const;
	
	// Check for excessive depth of tree. If too deep return true, and display message first time this happens.
	bool CheckIsTooDeep(bool bClearError = false) const;
	bool HasReportedDepthError() const { return m_bReportedDepthError; }

	// Remove all of the children by calling RemoveChild on each in turn.
	void RemoveAllChildren();

	// Remove specified child. Return true if succeeded.
	virtual bool RemoveChild(Asura_Hash_ID uChildHash) { return false; }

	// Add a child with the specified hash. Doesn't check that the child exists.
	// Nor does it check for circular references (calling code should prevent this).
	virtual void AddChild(Asura_Hash_ID uChildHash) {}

	// Should work for both collections and attachments.
	// Note: Must do the Add before the Remove to prevent auto-deletion of the new child in some circumstances.
	// Return true if the remove succeeded.
	virtual bool ReplaceChild(Asura_Hash_ID uOldChildHash, Asura_Hash_ID uNewChildHash)
	{
		if (uNewChildHash != ASURA_HASH_ID_UNSET)
		{
			AddChild(uNewChildHash);
		}
		return RemoveChild(uOldChildHash);
	}

	// Add a child chosen from a modal overlay.
	virtual void AddChildFromList() {}

	// Can we add a new child?
	virtual bool CanAddChild() const { return false; }

	// Set up the child iterator. Default case is to have no children.
	virtual void FillChildIterator(Tools_Container_ChildIterator& xIterator) const {}

	// Add any marker point names to the given list.
	virtual void AddMarkerPoints(ListSelection_Overlay& xList, const Asura_Hash_ID uMarkerHash) const;

	// Is this a marker attachment?
	virtual bool IsMarkerAttachment() const { return false; }
	// Is this a physics attachment?
	virtual bool IsPhysicsAttachment() const { return false; }

	// Helpers
	static void MakeCreationName(Asura_String& xName, const Asura_String& xContainerName, const Tools_ContainerGroup* pxGroup);

	void MakeFullName(Asura_String& xOutputName) const;
	void AddLogMessage(const char* szFormat,...) const;
	void AddInstanceName(Asura_String256& xBuffer, bool bAddContainerName, bool bAddDescription) const;

	// Prompt user for deletion/removal. Return true if Ok, false if cancelled.
	bool PromptForDeletion(const bool bRemoval) const;

	static bool IsEditOverlayChanged() { return s_bChangedEditOverlay || s_bChangedEditOverlay_MinorChange; }

	// State machine
	virtual		  Tools_ContainerStateMachine* GetStateMachine()	   { return 0; }
	virtual const Tools_ContainerStateMachine* GetStateMachine() const { return 0; }
	virtual u_int GetNumInputs() const { return 1; }

protected:
	u_int GetOverlayID() const { return reinterpret_cast< u_int >( this ); }

private:
	void SetCreationContainerName();

	// Recursive check to see whether this contains (or equals) the specified container.
	bool ContainsHelper(Asura_Hash_ID uSearchHash, int iDepth) const;

protected:
	u_int m_uClassification;		// From ASURA_CONTAINER_CLASSIFICATION or equivalent project enum.

	Asura_Hash_ID m_uFileHash;		// Hash of the file that this container is in.
	Asura_Hash_ID m_uGroupHash;		// Hash of the group that this container belongs to.

private:
	// Name on creation, which never changes, even when renamed, includes file and group. e.g. "file\group\container"
	Asura_String m_xCreationContainerName;
	Asura_Hash_ID m_uContainerHash;			 // Hash of m_xCreationContainerName.

protected:
	Asura_String m_xContainerName;		 // User visible name. e.g. "container"

	Asura_Container* m_pxAsuraContainer; // Corresponding Asura container.

	// Editing overlay has been modified (we need to reconstruct the instance from scratch).
	static bool s_bChangedEditOverlay;

	// Editing overlay has been modified (instance will pick up any changes on the fly).
	// Basically this only works for data that is looked up every frame by the instance.
	// Only use this if you *know* that the change will be picked up without needing
	// to destroy/recreate the instance. Otherwise just use s_bChangedEditOverlay.
	static bool s_bChangedEditOverlay_MinorChange;

private:
	mutable bool m_bReportedDepthError;  // Only report circular references once.

	// Tools-only flags, saved in tools chunk.
	enum FLAGS
	{
		FLAG_GROUP_SCOPE = 0x00000001,	// Limit scope to be within group (to make auto-deletion of orphans easier).
										// Currently there are only two scopes, namely group and global scope.
		FLAG_SHOW_NAME	 = 0x00000002,	// Show user visible name in various places (effectively this is a user-defined comment).
	};
	u_int m_uToolsFlags;

public:
	// Flags accessors
	void SetGroupScope()	{ m_uToolsFlags |= FLAG_GROUP_SCOPE; }
	void ClearGroupScope()	{ m_uToolsFlags &= ~FLAG_GROUP_SCOPE; }
	void ToggleGroupScope()	{ m_uToolsFlags ^= FLAG_GROUP_SCOPE; }
	bool IsGroupScope() const { return (m_uToolsFlags & FLAG_GROUP_SCOPE) != 0; }

	void SetShowName()		{ m_uToolsFlags |= FLAG_SHOW_NAME; }
	void ClearShowName()	{ m_uToolsFlags &= ~FLAG_SHOW_NAME; }
	void ToggleShowName()	{ m_uToolsFlags ^= FLAG_SHOW_NAME; }
	bool ShouldShowName() const { return (m_uToolsFlags & FLAG_SHOW_NAME) != 0; }	

	// Friends
	friend class Tools_Container_System;
	friend class Tools_ContainerRegistry;
	friend class Tools_Container_UndoStorage;
};


//////////////////////////
// Tools_Container_Mesh //
//////////////////////////
class Tools_Container_Mesh : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Mesh();

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);

	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	// Get a short name to be used in the instance overlay (e.g. skin name).
	virtual const char* GetInstanceName() const
	{
		return (m_xSkinName.IsBlank() ? "*None*" : m_xSkinName.GetString());
	}

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

	// Add any marker point names to the given list.
	virtual void AddMarkerPoints(ListSelection_Overlay& xList, const Asura_Hash_ID uMarkerHash) const;

	Asura_Hash_ID GetSkinHash() const { return m_xSkinName.GetHash(); }

	// Does the skin have the specified marker point?
	bool HasMarker(Asura_Hash_ID uMarkerHash) const;

	// Textures are reported directly from the loaded meshes.
	virtual void TraverseUsedTexturesForReport(void (*pfnCallback)(int iTextureID,void* pUserData), void* pUserData) const {}

	// Is the mesh container flagged to have shared dynamic physics?
	bool HasSharedDynamicPhysicsObject() const;

protected:
	void RefreshDialogue();

	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Container_Mesh*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(SelectSkinAndAnim);
	OVERLAY_VOID_CALLBACK(SelectAnim);

	#undef OVERLAY_VOID_CALLBACK

protected:
	Asura_String m_xSkinName;
	Asura_String m_xSkinFilename;
	Asura_String m_xAnimName;
	Asura_String m_xAnimFilename;

	int   m_iAnimFlags;
	u_int m_uRenderFlags;
	u_int m_uContainerMeshFlags;
	int m_iCollisionProxy;

	float m_fSelfIlluminationContribution;
	float m_fSelfIlluminationColourise;

	// Temporary strings, not saved.
	Asura_String m_xSkinNameDialogue; // String displayed on dialogue.
	Asura_String m_xAnimNameDialogue; // String displayed on dialogue.
};


////////////////////////////////
// Tools_Container_SoundEvent //
////////////////////////////////
class Tools_Container_SoundEvent : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_SoundEvent()
		: m_uSoundEventHash(ASURA_HASH_ID_UNSET)
		, m_uSEFlags(0)
	{
		RefreshDialogue();
	}

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	
	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	// Get a short name to be used in the instance overlay (e.g. sound event name).
	virtual const char* GetInstanceName() const { return m_xSoundEventName.GetString(); }

	// Render any Tools-only stuff.
	virtual void ToolsRender(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri) const;

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

protected:
	void RefreshDialogue();

	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Container_SoundEvent*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(SelectSoundEvent);

	#undef OVERLAY_VOID_CALLBACK


protected:
	Asura_Hash_ID m_uSoundEventHash;
	u_int m_uSEFlags;

	Asura_String m_xSoundEventNameDialogue; // String displayed on dialogue.
	Asura_String m_xSoundEventName;
};

/////////////////////////
// Tools_Container_PFX //
/////////////////////////
class Tools_Container_PFX : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_PFX();
	virtual ~Tools_Container_PFX();

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	virtual void PerFrameOverlayUpdate();
	
	virtual	void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual	void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void PostLoadInitialise() { RefreshDialogue(); }
	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	// Render any Tools-only stuff.
	virtual void ToolsRender(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri) const;

	// Get a short name to be used in the instance overlay (i.e. PFX name).
	virtual const char* GetInstanceName() const { return m_xPFXName.GetString(); }

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

protected:
	void RefreshDialogue();

	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Container_PFX*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(EditSourceVolume);

	#undef OVERLAY_VOID_CALLBACK


protected:
	// Properties
	Asura_Hash_ID m_uEffectTemplateHash;
	float m_fMinRetriggerTime;
	float m_fMaxRetriggerTime;

	bool m_bSkyboxEffect;
	bool m_bWaterDistorterEffect;
	bool m_bIgnoreModuleLinking;

	// PFX name
	Asura_String m_xPFXName;

	u_int m_uPFXFlags;

	// Volume editing
	Tools_Container_PFX_SourceVolume_Object* m_pxToolsSourceVolumeEdit;
	Asura_PFX_Volume m_xSourceVolume;
};


////////////////////////////
// Tools_Container_Corona //
////////////////////////////
class Tools_Container_Corona : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Corona();
	virtual ~Tools_Container_Corona() {}

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	virtual void PerFrameOverlayUpdate();
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

	// Build a list of textures for texture report log.
	virtual void TraverseUsedTexturesForReport(void (*pfnCallback)(int iTextureID,void* pUserData), void* pUserData) const;

protected:
	Asura_Container_Corona_Param m_xParam;

	Asura_String m_xCoronaTextureName;
};


////////////////////////////////
// Tools_Container_LightShaft //
////////////////////////////////
#ifdef ASURA_TARGET_SM3

class Tools_Container_LightShaft : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_LightShaft();
	virtual ~Tools_Container_LightShaft() {}

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	virtual void PerFrameOverlayUpdate();
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

	// Build a list of textures for texture report log.
	virtual void TraverseUsedTexturesForReport(void (*pfnCallback)(int iTextureID,void* pUserData), void* pUserData) const;

protected:
	Asura_Container_LightShaft_Param m_xParam;

	Asura_String m_axTextureName[2];	// Light shaft texture A,B.
	Asura_String m_xCoronaTextureName;
};

#endif //#ifdef ASURA_TARGET_SM3


///////////////////////////////
// Tools_Container_Projector //
///////////////////////////////
#ifdef ASURA_TARGET_SM3

class Tools_Container_Projector : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Projector();
	virtual ~Tools_Container_Projector();

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	virtual void PerFrameOverlayUpdate();

	// Render any Tools-only stuff.
	virtual void ToolsRender(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri) const;
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

	// Build a list of textures for texture report log.
	virtual void TraverseUsedTexturesForReport(void (*pfnCallback)(int iTextureID,void* pUserData), void* pUserData) const;

	// Add tabs for projectors and fall-off.
	void AddOverlayTab_Projector(Asura_Tabs_Dialog* pxTab, Asura_Vector_2& xOffset);

protected:
	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Container_Projector*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(EditLimitingPlane);

	#undef OVERLAY_VOID_CALLBACK

protected:
	Asura_Container_Projector_Param m_xParam;
	Asura_ShadowParameters m_xShadowParam;

	Asura_String m_xTextureName;
	Asura_String m_xDetailTextureName;

	Tools_Container_LimitingPlane_Object*  m_pxLimitingPlane;
};

#endif //#ifdef ASURA_TARGET_SM3


////////////////////////////
// Tools_Container_Photon //
////////////////////////////
#ifdef ASURA_TARGET_SM3

class Tools_Container_Photon : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Photon();
	virtual ~Tools_Container_Photon();

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	virtual void PerFrameOverlayUpdate();

	// Render any Tools-only stuff.
	virtual void ToolsRender(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri) const;
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void Verify(bool bRecurse) const;

	virtual void AddFilesToExportList(List<Asura_String>& xNameList) const;
	virtual void LoadRequiredFiles() const;

	// Build a list of textures for texture report log.
	virtual void TraverseUsedTexturesForReport(void (*pfnCallback)(int iTextureID,void* pUserData), void* pUserData) const;

	// Add tabs for photon and fall-off.
	void AddOverlayTab_Photon(Asura_Tabs_Dialog* pxTab, Asura_Vector_2& xOffset);

	Asura_Container_Photon_Param& GetParam() { return m_xParam; }

protected:
	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Container_Photon*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(EditCustomLightVolume);
	OVERLAY_VOID_CALLBACK(ResetCustomLightVolume);

	#undef OVERLAY_VOID_CALLBACK

protected:
	Asura_Container_Photon_Param m_xParam;
	Asura_ShadowParameters m_xShadowParam;

	Asura_String m_xSharpTextureName;
	Asura_String m_xBlurredTextureName;

	class Tools_Container_PhotonVolumeEdit*  m_pxLimitingVolume;
};

#endif //#ifdef ASURA_TARGET_SM3


////////////////////////////
// Tools_Container_Strobe //
////////////////////////////
class Tools_Container_Strobe : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Strobe();

	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();

protected:
	float m_fMinTimeOff;
	float m_fMaxTimeOff;
	float m_fMinTimeOn;
	float m_fMaxTimeOn;
};


/////////////////////////////////
// Tools_Container_GraphColour //
/////////////////////////////////
class Tools_Container_GraphColour : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();

protected:
	Asura_Container_GraphColour_Param m_xParam;
};


////////////////////////////
// Tools_Container_Health //
////////////////////////////
class Tools_Container_Health : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Health();
	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();
	virtual void PerFrameOverlayUpdate();

protected:
	float m_fInitialHealth;
	float m_fMaxHealth;
};

/////////////////////////////
// Tools_Container_Physics //
/////////////////////////////
class Tools_Container_Physics : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Physics() {}
	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();

protected:
	Asura_Container_Physics_Param m_xParam;
};

///////////////////////////////
// Tools_Container_Blueprint //
///////////////////////////////
class Tools_Container_Blueprint : public Tools_Container
{
	typedef Tools_Container PARENT;

public:
	Tools_Container_Blueprint();
	virtual void AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset);
	
	virtual void ReadFromChunkStream(Asura_Chunk_Stream& xStream);
	virtual void WriteToChunkStream(Asura_Chunk_Stream& xStream) const;

	virtual void CopyDataToAsura();

	void RefreshDialogue();

protected:
	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Container_Blueprint*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(SelectBlueprint);

	#undef OVERLAY_VOID_CALLBACK


protected:
	Asura_Hash_ID m_uBlueprintTypeHash;
	Asura_Hash_ID m_uBlueprintHash;

	Asura_String m_xBlueprintTypeNameDialogue;
	Asura_String m_xBlueprintNameDialogue;
};

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__TOOLS_CONTAINER_H__

