///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_ContainerFile.h
//
// Created by Martin G.Harris (3/6/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_CONTAINER_FILE_H__
#define __TOOLS_CONTAINER_FILE_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_Hash.h"
#include "Core/Asura_String.h"

// Forward declarations

/////////////////////////
// Tools_ContainerFile //
/////////////////////////
// Object representing a file containing groups of containers.
class Tools_ContainerFile
{
public:
	Tools_ContainerFile()
		: m_bModified(false)
		, m_bWriteable(false)
		, m_uCheckedWriteableFrame(0)
	{}

	void SetModified();
	void ClearModified() { m_bModified = false; }
	bool IsModified() const { return m_bModified; }

	// Is writeable? Just returns cached value.
	bool IsWriteable() const { return m_bWriteable; }

	Asura_Hash_ID GetFileHash() const { return m_xCreationShortName.GetHash(); }

	void SetFilename(const Asura_String& xFilename) { m_xFilename = xFilename; }
	void SetFilename(const Asura_Char* szString)    { m_xFilename = szString; }

	void SetShortName(bool bNewFile); // Set the short name from the filename.

	const Asura_String& GetCreationShortName() const { return m_xCreationShortName; }
	const Asura_String& GetShortName() const { return m_xShortName; }
	const Asura_String& GetFilename() const  { return m_xFilename; }

	// Warn if the file is read-only. Set flag to force a recheck even if already checked this frame.
	void WarnIfReadOnly(bool bForceCheck = false) const;

	// Refresh writeable status flag.
	void RefreshWriteableStatus() const;

	// Can we save this file? If not then display error message.
	bool CanSave();

private:
	// Short name at creation time, used for hash. Never changes, even when the file is renamed on the disk.
	Asura_String m_xCreationShortName;

	// These are derived from the actual filename being read or written on the disk.
	Asura_String m_xShortName; // Filename excluding initial directory and extension e.g. "file" or "folder\file".
	Asura_String m_xFilename;  // Filename relative to Asura_Files\ e.g. "Containers\file.asr" or "Containers\folder\file.asr".
	
	// Is this file modified?
	bool m_bModified;				
	
	mutable bool m_bWriteable; // Is the file writeable?
	mutable u_int m_uCheckedWriteableFrame; // Frame number we last checked whether file was writeable.

	// Friends
	friend class Asura_Chunk_ToolsContainerFile;
};


//////////////////////////
// Tools_ContainerGroup //
//////////////////////////
// Object representing a group of containers collected together.
class Tools_ContainerGroup
{
public:
	Tools_ContainerGroup()
		: m_uFileHash(ASURA_HASH_ID_UNSET)
		, m_uAutoGenNumber(0)
	{}

	Tools_ContainerGroup(Asura_Hash_ID uFileHash, const Asura_Char* const szGroupName)
		: m_uFileHash(uFileHash)
		, m_xGroupName(szGroupName)
		, m_uAutoGenNumber(0)
	{}

	// Flag this group as having been modified.
	void SetModified();

	void SetCreationGroupName();

	void SetFileHash(Asura_Hash_ID uFileHash) { m_uFileHash = uFileHash; }
	void SetGroupName(const Asura_String& xGroupName) { m_xGroupName = xGroupName; }

	Asura_Hash_ID GetFileHash() const { return m_uFileHash; }
	Asura_Hash_ID GetGroupHash() const { return m_xCreationGroupName.GetHash(); }

	const Asura_String& GetGroupName() const { return m_xGroupName; }
	const Asura_String& GetCreationGroupName() const { return m_xCreationGroupName; }

	u_int GetAutoGenNumber()
	{
		SetModified();
		return m_uAutoGenNumber++; 
	}

private:
	Asura_Hash_ID m_uFileHash;		   // Hash of the file this currently belongs to.

	Asura_String m_xCreationGroupName; // Name on creation, which never changes, even when renamed, includes file. e.g. "file\group"

	Asura_String m_xGroupName;		   // User visible name. e.g. "group"

	u_int m_uAutoGenNumber;			   // Number used in auto-generated names. Incremented each time used.

	// Friends
	friend class Asura_Chunk_ToolsContainerGroup;
};

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__TOOLS_CONTAINER_FILE_H__

