///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_ContainerRegistry.h
//
// Created by Martin G.Harris (23/5/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_CONTAINER_REGISTRY_H__
#define __TOOLS_CONTAINER_REGISTRY_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Entity/Container/Asura_Container.h"
#include "Entity/Container/Tools_Container.h"

// Forward declarations
class Tools_Entity_Container;
class Asura_Entity;

/////////////////////////////
// Tools_ContainerRegistry //
/////////////////////////////
// Registry for different types of container class. Allowing Tools and projects to automatically register
// container classes. Corresponding Asura classes should also be registered using Asura_ContainerRegistry.
class Tools_ContainerRegistry
{
public:
	// Typedefs for creation functions.
	typedef Tools_Container* (*PFN_TOOLS_CREATE) (void);
	typedef Tools_Entity_Container* (*PFN_TOOLS_ENTITY_CREATE) (Asura_Entity*);

	// Add entries.
	void AddEntry(u_int uClassification, PFN_TOOLS_CREATE pfnFunction, const char* szClassName, bool bCanAttach);
	void AddToolsEntity(PFN_TOOLS_ENTITY_CREATE pfnFunction, bool bProjectOverride);

	u_int GetNumEntries() const { return m_uNumEntries; }

	// Get container names.
	static const char* GetContainerClassName(u_int uClassification);

	// Creation functions.
	static Tools_Container* CreateToolsClass(u_int uClassification);
	static Tools_Entity_Container* CreateToolsEntity(Asura_Entity* pxEntity);

	// Get instance.
	static Tools_ContainerRegistry& GetInstance()
	{
		static Tools_ContainerRegistry s_xInstance;
		return s_xInstance;
	}

	// Can the specified container classification use an attachment?
	static bool CanUseAttachment(u_int uClassification);

private:
	Tools_ContainerRegistry() : m_uNumEntries(0), m_pfnToolsEntityCreateFn(0) {}

	class Entry
	{
	public:
		Entry()
			: m_uClassification(AsuraContainerClass_NotSet)
			, m_pfnToolsCreateFn(0)
			, m_szClassName(0)
			, m_bCanAttach(true)
		{}

		u_int m_uClassification;
		PFN_TOOLS_CREATE m_pfnToolsCreateFn; // Tools creation function.
		const char* m_szClassName;			 // Name of class.
		bool m_bCanAttach;					 // Can use attachment?
	};

	const Entry& GetEntry(int iIndex) const
	{
		Asura_Assert(iIndex>=0 && iIndex<iCONTAINER_TYPES_MAX_NUMBER, "Invalid Index.");
		return m_axRegistry[iIndex]; 
	}
	
	Entry m_axRegistry[iCONTAINER_TYPES_MAX_NUMBER];
	u_int m_uNumEntries;
	PFN_TOOLS_ENTITY_CREATE m_pfnToolsEntityCreateFn;

	// Friends
	friend class Tools_Container_System;
};

//////////////////////////////////
// Tools_ContainerRegisterClass //
//////////////////////////////////
// Helper to register classes derived from Tools_Container.
template <class T> class Tools_ContainerRegisterClass
{
public:
	// uClassification is from ASURA_CONTAINER_CLASSIFICATION or equivalent project enum.
	Tools_ContainerRegisterClass(u_int uClassification, const char* szClassName, bool bCanAttach = true)
	{
		Tools_ContainerRegistry::GetInstance().AddEntry(uClassification, Create, szClassName, bCanAttach);
	}

	static Tools_Container* Create() { return new T; }
};

////////////////////////////////////////
// Tools_ContainerRegisterToolsEntity //
////////////////////////////////////////
// Helper to register project-specific Tools entity derived from Tools_Entity_Container.
// bProjectOverride is true for a project-specific entity and false for an Asura entity.
template<class TOOLSENTITY> class Tools_ContainerRegisterToolsEntity
{
public:
	Tools_ContainerRegisterToolsEntity(bool bProjectOverride = true)
	{
		Tools_ContainerRegistry::GetInstance().AddToolsEntity(Create, bProjectOverride);
	}

	static Tools_Entity_Container* Create(Asura_Entity* pxEntity) { return new TOOLSENTITY(pxEntity); }
};

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__TOOLS_CONTAINER_REGISTRY_H__

