///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_ContainerState.cpp
//
// Created by Martin G.Harris (26/6/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Tools_ContainerState.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "Core/Asura_StringBuffer.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Container/Asura_ContainerState.h"

// Tools
#include "Entity/Container/Tools_Container_Attachment.h"
#include "Entity/Container/Tools_Container_System.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"

///////////////////////////////////////////////////////////////////////////////
// Tools_ContainerState
///////////////////////////////////////////////////////////////////////////////
Tools_ContainerState::Tools_ContainerState()
: m_uStateHash(ASURA_HASH_ID_UNSET)
, m_uInputSlotMask(0xffffffff)
, m_uNumChildStates(0)
, m_pxChildState(0)
, m_uNumDamageOverrides(0)
, m_pxDamageOverride(0)
{
	for (u_int u=0; u<uNUM_OLD_STATE_HASHES_TO_STORE; ++u)
	{
		m_auOldStateHashes[u] = ASURA_HASH_ID_UNSET;
	}
}

Tools_ContainerState::~Tools_ContainerState()
{
	FreeAll();
}

void Tools_ContainerState::FreeAll()
{
	delete [] m_pxChildState;
	m_pxChildState = 0;
	m_uNumChildStates = 0;

	delete [] m_pxDamageOverride;
	m_pxDamageOverride = 0;
	m_uNumDamageOverrides = 0;
}

// Chunkstream code
const int iCHUNK_VERSION_TOOLS_CONTAINER_STATE = 4;

void Tools_ContainerState::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	FreeAll();

	int iVersion = -1;
	xStream >> iVersion;

	if (iVersion < 0 || iVersion > iCHUNK_VERSION_TOOLS_CONTAINER_STATE)
	{
		Asura_Assertf(0, ("Unknown chunk version (%d) in Tools_ContainerState", iVersion));
		return;
	}

	xStream >> m_xStateName;
	m_uStateHash = m_xStateName.GetHash();

	xStream >> m_uNumChildStates;

	if (m_uNumChildStates > 0)
	{
		m_pxChildState = new Asura_ContainerStateEntry[m_uNumChildStates];
		
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			Asura_ContainerStateEntry& xEntry = m_pxChildState[u];
			xStream >> xEntry.m_uContainerHash;
			xStream >> xEntry.m_uStateHash;

			if (iVersion >= 3)
			{
				xStream >> xEntry.m_uInputIndex;
			}
		}
	}

	if (iVersion >= 1)
	{
		xStream >> m_uNumDamageOverrides;
		
		if (m_uNumDamageOverrides > 0)
		{
			m_pxDamageOverride = new Asura_ContainerState_DamageOverride[m_uNumDamageOverrides];

			for (u_int u=0; u<m_uNumDamageOverrides; ++u)
			{
				Asura_ContainerState_DamageOverride& xEntry = m_pxDamageOverride[u];
				xStream >> xEntry.m_uOverrideStateHash;
				xStream >> xEntry.m_fMaxHealthPercent;
			}
		}
	}

	if (iVersion >= 2)
	{
		for (u_int u=0; u<uNUM_OLD_STATE_HASHES_TO_STORE; ++u)
		{
			m_auOldStateHashes[u] = ASURA_HASH_ID_UNSET;
		}

		u_int uNumHashes = 0;
		xStream >> uNumHashes;

		for (u_int u=0; u<uNumHashes; ++u)
		{
			u_int uOldHash = ASURA_HASH_ID_UNSET;		
			xStream >> uOldHash;
	
			if (u<uNUM_OLD_STATE_HASHES_TO_STORE)
			{
				m_auOldStateHashes[u] = uOldHash;
			}
		}
	}

	if (iVersion >= 4)
	{
		xStream >> m_uInputSlotMask;
	}

	// On/off states are defined as all on or all off, so we don't need to save the children.
	// Fix up any historical versions that stored the children.
	if (IsInvariant())
	{
		FreeAll();
	}
}

void Tools_ContainerState::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << iCHUNK_VERSION_TOOLS_CONTAINER_STATE;
	xStream << m_xStateName;
	xStream << m_uNumChildStates;

	for (u_int u=0; u<m_uNumChildStates; ++u)
	{
		const Asura_ContainerStateEntry& xEntry = m_pxChildState[u];
		xStream << xEntry.m_uContainerHash;
		xStream << xEntry.m_uStateHash;
		xStream << xEntry.m_uInputIndex;
	}

	xStream << m_uNumDamageOverrides;
	
	for (u_int u=0; u<m_uNumDamageOverrides; ++u)
	{
		const Asura_ContainerState_DamageOverride& xEntry = m_pxDamageOverride[u];
		xStream << xEntry.m_uOverrideStateHash;
		xStream << xEntry.m_fMaxHealthPercent;
	}

	xStream << uNUM_OLD_STATE_HASHES_TO_STORE;

	for (u_int u=0; u<uNUM_OLD_STATE_HASHES_TO_STORE; ++u)
	{
		xStream << m_auOldStateHashes[u];
	}

	xStream << m_uInputSlotMask;
}

/////////////////
// IsInvariant //
/////////////////
// On/off states can't be modified.
bool Tools_ContainerState::IsInvariant() const
{
	return ( m_uStateHash == Asura_ContainerState::GetOnStateHash() ||
		     m_uStateHash == Asura_ContainerState::GetOffStateHash() );
}

//////////////
// AddChild //
//////////////
void Tools_ContainerState::AddChild(Asura_Hash_ID uChildContainerHash, Asura_Hash_ID uChildStateHash, u_int uInputIndex)
{
	// On/off states can't be modified.
	if (IsInvariant()) return;

	Asura_ContainerStateEntry* pxChildState = new Asura_ContainerStateEntry[m_uNumChildStates+1];
	Asura_Assert(pxChildState, "Null pointer");
	if (!pxChildState) return;

	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			pxChildState[u] = m_pxChildState[u];
		}
	}

	pxChildState[m_uNumChildStates].m_uContainerHash = uChildContainerHash;
	pxChildState[m_uNumChildStates].m_uStateHash	 = uChildStateHash;
	pxChildState[m_uNumChildStates].m_uInputIndex	 = uInputIndex;

	delete [] m_pxChildState;
	m_pxChildState = pxChildState;
	++m_uNumChildStates;
}

//////////////////////////
// RemoveChildAllInputs //
//////////////////////////
void Tools_ContainerState::RemoveChildAllInputs(Asura_Hash_ID uChildContainerHash)
{
	// On/off states can't be modified.
	if (IsInvariant()) return;

	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			if (m_pxChildState[u].m_uContainerHash == uChildContainerHash)
			{
				m_pxChildState[u] = m_pxChildState[m_uNumChildStates-1];
				--m_uNumChildStates;
				if (m_uNumChildStates == 0)
				{
					FreeAll();
					return;
				}
				--u;
			}
		}
	}
}

/////////////////
// RemoveChild //
/////////////////
void Tools_ContainerState::RemoveChild(Asura_Hash_ID uChildContainerHash, u_int uInputIndex)
{
	// On/off states can't be modified.
	if (IsInvariant()) return;

	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			if (m_pxChildState[u].m_uContainerHash == uChildContainerHash &&
				m_pxChildState[u].m_uInputIndex == uInputIndex)
			{
				m_pxChildState[u] = m_pxChildState[m_uNumChildStates-1];
				--m_uNumChildStates;
				if (m_uNumChildStates == 0)
				{
					FreeAll();
				}
				return;
			}
		}
	}
}

/////////////////
// ChangeChild //
/////////////////
// Add, remove or change as necessary.
void Tools_ContainerState::ChangeChild(Asura_Hash_ID uChildContainerHash, Asura_Hash_ID uChildStateHash, u_int uInputIndex)
{
	// On/off states can't be modified.
	if (IsInvariant()) return;

	const Asura_Hash_ID uOldState = GetChildState(uInputIndex, uChildContainerHash);
	if (uOldState == uChildStateHash)
	{
		return; // Nothing to do.
	}

	// Remove child.
	if (uChildStateHash == Asura_ContainerState::GetOffStateHash())
	{
		RemoveChild(uChildContainerHash, uInputIndex);
		return;
	}

	// Change child.
	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			if (m_pxChildState[u].m_uContainerHash == uChildContainerHash && 
				m_pxChildState[u].m_uInputIndex == uInputIndex)
			{
				m_pxChildState[u].m_uStateHash = uChildStateHash;
				return;
			}
		}
	}

	// Add child.
	AddChild(uChildContainerHash, uChildStateHash, uInputIndex);
}

//////////////////
// ReplaceChild //
//////////////////
// Replace the child if it exists.
void Tools_ContainerState::ReplaceChild(const Asura_Hash_ID uOldChildHash, const Asura_Hash_ID uNewChildHash)
{
	if (!m_pxChildState) return;

	if (uNewChildHash == ASURA_HASH_ID_UNSET)
	{
		RemoveChildAllInputs(uOldChildHash);
		return;
	}
	
	for (u_int u=0; u<m_uNumChildStates; ++u)
	{
		if (m_pxChildState[u].m_uContainerHash == uOldChildHash)
		{
			m_pxChildState[u].m_uContainerHash = uNewChildHash;
		}
	}
}

///////////////////
// GetChildState //
///////////////////
// Get child state, returns "off" state hash if child is disabled or unknown.
Asura_Hash_ID Tools_ContainerState::GetChildState(u_int uInputIndex, Asura_Hash_ID uChildContainerHash) const
{
	// On/off states can't be modified. Their children inherit the same state.
	if (IsInvariant()) return m_uStateHash;

	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			if (m_pxChildState[u].m_uContainerHash == uChildContainerHash &&
				m_pxChildState[u].m_uInputIndex == uInputIndex)
			{
				return m_pxChildState[u].m_uStateHash;
			}
		}
	}
	return Asura_ContainerState::GetOffStateHash();
}

///////////////////
// GetStateEntry //
///////////////////
// Access by index.
const Asura_ContainerStateEntry* Tools_ContainerState::GetStateEntry(u_int uIndex) const
{
	if (m_pxChildState && uIndex<m_uNumChildStates)
	{
		return &m_pxChildState[uIndex];
	}

	Asura_Assert(0, "Child state index out-of-range");
	return 0;
}

//////////////////
// ReplaceState //
//////////////////
// Replaced a state hash associated with a specified container collection. Fix up this state.
// Return true if modified.
bool Tools_ContainerState::ReplaceState(const Asura_Hash_ID uContainerHash,
										const Asura_Hash_ID uOldStateHash,
										const Asura_Hash_ID uNewStateHash)
{
	bool bModified = false;

	if (m_pxChildState)
	{
		for (u_int u=0; u<m_uNumChildStates; ++u)
		{
			if (m_pxChildState[u].m_uStateHash == uOldStateHash)
			{
				// Found old state hash, but is it for the correct container.
				if (m_pxChildState[u].m_uContainerHash == uContainerHash)
				{
					// Yes. Replace the state.
					m_pxChildState[u].m_uStateHash = uNewStateHash;
					bModified = true;
				}
				else
				{
					// No, but this might be an attachment, we need to check its child.
					const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_pxChildState[u].m_uContainerHash);
					if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Attachment)
					{
						for ( Tools_Container_ChildIterator xIt(pxContainer); !xIt.Done(); xIt.NextSkipIfSame() )
						{	
							const Asura_Hash_ID uHash = xIt.GetCurrentHash();
							if (uHash == uContainerHash)
							{
								// Found the collection. Replace the state.
								m_pxChildState[u].m_uStateHash = uNewStateHash;
								bModified = true;
							}
						}
					}
				}
			}
		}
	}

	return bModified;
}

////////////////////////////////
// ReplaceDamageOverrideState //
////////////////////////////////
// Replaced a state hash associated with this container collection. Fix up damage overrides.
// Return true if modified.
bool Tools_ContainerState::ReplaceDamageOverrideState(const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash)
{
	bool bModified = false;

	for (u_int u=0; u<m_uNumDamageOverrides; ++u)
	{
		if (m_pxDamageOverride[u].m_uOverrideStateHash == uOldStateHash)
		{
			m_pxDamageOverride[u].m_uOverrideStateHash = uNewStateHash;
			bModified = true;
		}
	}

	return bModified;
}

///////////////////////
// AddDamageOverride //
///////////////////////
// Add a damage override.
void Tools_ContainerState::AddDamageOverride(float fMaxHealthPercent, Asura_Hash_ID uStateHash)
{
	// On/off states can't be modified.
	if (IsInvariant()) return;

	fMaxHealthPercent = Asura_Maths::ClampToRange(fMaxHealthPercent, 0.0f, 100.0f);

	// Check for existing entry with same health value. If found then replace it.
	// Also check for whether we just need to move an existing threshold up or down.
	float fPrevious = 0.0f;
	for (u_int u=0; u<m_uNumDamageOverrides; ++u)
	{
		const float fCurrent = m_pxDamageOverride[u].m_fMaxHealthPercent;
		if (fCurrent == fMaxHealthPercent)
		{
			// Same health. Replace state.
			m_pxDamageOverride[u].m_uOverrideStateHash = uStateHash;
			return;
		}

		if (fPrevious < fMaxHealthPercent && fMaxHealthPercent < fCurrent)
		{
			if (m_pxDamageOverride[u].m_uOverrideStateHash == uStateHash)
			{
				// Adjust existing threshold downwards.
				m_pxDamageOverride[u].m_fMaxHealthPercent = fMaxHealthPercent;
				return;
			}

			if (u > 0 && m_pxDamageOverride[u-1].m_uOverrideStateHash == uStateHash)
			{
				// Adjust previous threshold upwards.
				m_pxDamageOverride[u-1].m_fMaxHealthPercent = fMaxHealthPercent;
				return;
			}
		}

		fPrevious = fCurrent;
	}

	if (m_uNumDamageOverrides > 0 &&
		m_pxDamageOverride[m_uNumDamageOverrides-1].m_fMaxHealthPercent < fMaxHealthPercent &&
		m_pxDamageOverride[m_uNumDamageOverrides-1].m_uOverrideStateHash == uStateHash)
	{
		// Adjust last threshold upwards.
		m_pxDamageOverride[m_uNumDamageOverrides-1].m_fMaxHealthPercent = fMaxHealthPercent;
		return;
	}

	// Insert in the array (which is sorted by health threshold).
	Asura_ContainerState_DamageOverride* pxOverride = new Asura_ContainerState_DamageOverride[m_uNumDamageOverrides+1];
	Asura_Assert(pxOverride, "Null pointer");
	if (!pxOverride) return;

	bool bInserted = false;
	u_int uWrite = 0;

	for (u_int u=0; u<m_uNumDamageOverrides; ++u, ++uWrite)
	{
		if (!bInserted && m_pxDamageOverride[u].m_fMaxHealthPercent > fMaxHealthPercent)
		{
			pxOverride[uWrite].m_uOverrideStateHash = uStateHash;
			pxOverride[uWrite].m_fMaxHealthPercent  = fMaxHealthPercent;
			++uWrite;
			bInserted = true;
		}
		pxOverride[uWrite] = m_pxDamageOverride[u];
	}

	if (!bInserted)
	{
		pxOverride[uWrite].m_uOverrideStateHash = uStateHash;
		pxOverride[uWrite].m_fMaxHealthPercent  = fMaxHealthPercent;
		++uWrite;
	}

	Asura_Assert(uWrite == m_uNumDamageOverrides+1, "Error: wrote wrong number of elements to array");	

	delete [] m_pxDamageOverride;
	m_pxDamageOverride = pxOverride;
	++m_uNumDamageOverrides;
}

//////////////////////////
// RemoveDamageOverride //
//////////////////////////
// Remove a damage override.
void Tools_ContainerState::RemoveDamageOverride(u_int uIndex)
{
	if (uIndex >= m_uNumDamageOverrides)
	{
		Asura_Assert(0, "Invalid array index.");
		return;
	}

	// Copy elements down one place, to overwrite the removed element.
	for (u_int u = uIndex+1; u<m_uNumDamageOverrides; ++u)
	{
		m_pxDamageOverride[u-1] = m_pxDamageOverride[u];
	}

	--m_uNumDamageOverrides;
}

///////////////////////
// GetDamageOverride //
///////////////////////
// Get damage override by index.
const Asura_ContainerState_DamageOverride* Tools_ContainerState::GetDamageOverride(u_int uIndex) const
{
	if (!m_pxDamageOverride ||
		uIndex >= m_uNumDamageOverrides)
	{
		return 0;
	}

	return &m_pxDamageOverride[uIndex];
}

//////////////////////
// SaveOldStateHash //
//////////////////////
// We've renamed this state, store the old hash.
void Tools_ContainerState::SaveOldStateHash(const Asura_Hash_ID uOldStateHash)
{
	if (uOldStateHash == ASURA_HASH_ID_UNSET) return;

	if (HasOldStateHash(uOldStateHash))
	{
		// Already stored in the array.
		return;
	}

	for (u_int u = uNUM_OLD_STATE_HASHES_TO_STORE-1; u>0; --u)
	{
		m_auOldStateHashes[u] = m_auOldStateHashes[u-1];
	}

	m_auOldStateHashes[0] = uOldStateHash;
}

/////////////////////
// HasOldStateHash //
/////////////////////
bool Tools_ContainerState::HasOldStateHash(const Asura_Hash_ID uOldStateHash) const
{
	for (u_int u = 0; u < uNUM_OLD_STATE_HASHES_TO_STORE; ++u)
	{
		if (m_auOldStateHashes[u] == uOldStateHash)
		{
			return true;
		}

		if (m_auOldStateHashes[u] == ASURA_HASH_ID_UNSET)
		{
			break;
		}
	}

	return false;
}

///////////////////////////
// IsEnabledForInputSlot //
///////////////////////////
// Is this state enabled for the specified input slot?
// The mask allows the user to specify which states appear in which input slots,
// thereby making it clearer to users how container entities should be set up.
// It should simplify setting up change state static messages too, as you need to
// know which input slot you are trying to change.
bool Tools_ContainerState::IsEnabledForInputSlot(u_int uInputIndex) const
{
	if (uInputIndex >= 32)
	{
		return true;
	}

	const u_int uTestBit = (1<<uInputIndex);
	return (m_uInputSlotMask & uTestBit) != 0;
}

////////////
// Verify //
////////////
// Check for errors. Return number of errors found.
void Tools_ContainerState::Verify(const Tools_Container* pxParent, const Tools_Container_ChildIterator& xIt) const
{
	for (u_int u=0; u<m_uNumChildStates; ++u)
	{
		Asura_Assert(m_pxChildState, "Null pointer");
		if (!m_pxChildState) return;

		// Check that the child still exists.
		const Asura_Hash_ID uChildContainerHash = m_pxChildState[u].m_uContainerHash;
		if (!xIt.IsMember(uChildContainerHash))
		{
			pxParent->AddLogMessage("child (0x%08x) for state '%s' not found", uChildContainerHash, m_xStateName.GetString());
		}

		// Check that the child has the specified state (note that "on"/"off" always exist).
		const u_int uInputIndex = m_pxChildState[u].m_uInputIndex;
		const Asura_Hash_ID uChildStateHash = m_pxChildState[u].m_uStateHash;

		if (uChildStateHash != Asura_ContainerState::GetOnStateHash() &&
			uChildStateHash != Asura_ContainerState::GetOffStateHash())
		{
			const Tools_Container* pxChild = Tools_Container_System::FindContainerByHash(uChildContainerHash);
			
			// Skip past attachments.
			if (pxChild && pxChild->GetClassification() == AsuraContainerClass_Attachment)
			{
				const Tools_Container_Attachment* pxAttachment = static_cast<const Tools_Container_Attachment*>(pxChild);
				pxChild = Tools_Container_System::FindContainerByHash(pxAttachment->GetChildContainerHash());
			}

			const Tools_ContainerStateMachine* const pxMachine = (pxChild ? pxChild->GetStateMachine() : 0);
			const Tools_ContainerState* const pxState = (pxMachine ? pxMachine->FindState(uChildStateHash) : 0); 
			if (pxState == 0)
			{
				// Child doesn't have the state.
				if (pxChild)
				{
					pxParent->AddLogMessage("state '%s' has child '%s' with state (0x%08x) which was not found", m_xStateName.GetString(), pxChild->GetContainerName().GetString(), uChildStateHash);
				}
			}
			else if (!pxState->IsEnabledForInputSlot(uInputIndex))
			{
				// State exists but isn't enabled for the specified input slot.
				pxParent->AddLogMessage("state '%s' has child '%s' with state '%s' which is not valid for input %u.", m_xStateName.GetString(), pxChild->GetContainerName().GetString(), pxState->GetStateName().GetString(), uInputIndex);
			}
		}
	}

	// Check that the damage override states exist in the parent collection.
	const Tools_ContainerStateMachine* const pxMachine = pxParent->GetStateMachine();
	if (pxMachine)
	{
		for (u_int u=0; u<m_uNumDamageOverrides; ++u)
		{
			const Asura_Hash_ID uOverrideStateHash = m_pxDamageOverride[u].m_uOverrideStateHash;
			if (!pxMachine->FindState(uOverrideStateHash))
			{
				pxParent->AddLogMessage("state '%s' has damage override state (0x%08x) which was not found", m_xStateName.GetString(), uOverrideStateHash);
			}
		}
	}
}

/////////////////
// FixUpErrors //
/////////////////
// Silently fix up certain errors that the user can't fix.
// Remove leaked references to children that no longer exist.
void Tools_ContainerState::FixUpErrors(const Tools_Container* pxParent)
{
	if (!pxParent) return;
	if (IsInvariant()) return;
	if (m_uNumChildStates == 0) return;
	if (!m_pxChildState) return;

	const Tools_Container_ChildIterator xIt(pxParent);

	for (u_int u=0; (m_pxChildState && u<m_uNumChildStates); )
	{
		// Check that the child exists.
		const Asura_Hash_ID uChildContainerHash = m_pxChildState[u].m_uContainerHash;
		
		if ( !xIt.IsMember(uChildContainerHash) )
		{
			// Child no longer exists; this is a leak. Try to clean it up.
			const u_int uOldNumChildren = m_uNumChildStates;
			RemoveChildAllInputs(uChildContainerHash);
			if (uOldNumChildren != m_uNumChildStates)
			{
				// Successfully removed a child.
				pxParent->AddLogMessage("child (0x%08x) for state '%s' not found. Fixed.", uChildContainerHash, m_xStateName.GetString());
				
				// Flag the file as being modified, so that it'll be written out.
				// There is no particular need to flag the container as modified;
				// it is such a small change that we don't need to update anything.
				Tools_Container_System::SetFileModified(pxParent->GetFileHash());

				// Repeat with the same counter.
				continue;
			}

			// Failed to remove it for some reason. Just carry on.
		}

		// Advance counter to next entry.
		++u;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Tools_ContainerStateMachine
///////////////////////////////////////////////////////////////////////////////
Tools_ContainerStateMachine::Tools_ContainerStateMachine()
: m_uNumStates(0)
, m_ppxState(0)
, m_uNumInputs(1)
{
}

Tools_ContainerStateMachine::~Tools_ContainerStateMachine()
{
	FreeAll();
}

//////////////////////
// AddDefaultStates //
//////////////////////
// Newly created state machines have an "on" and an "off" state.
bool Tools_ContainerStateMachine::AddDefaultStates()
{
	bool bAdded = false;

	// Add default "off" state.
	if (!FindState(Asura_ContainerState::GetOffStateHash()))
	{
		Asura_String xOff(Tools_ContainerState::GetOffStateName());
		AddState(xOff);
		bAdded = true;
	}

	// Add default "on" state.
	if (!FindState(Asura_ContainerState::GetOnStateHash()))
	{
		Asura_String xOn(Tools_ContainerState::GetOnStateName());
		AddState(xOn);
		bAdded = true;
	}

	return bAdded;
}

/////////////
// FreeAll //
/////////////
void Tools_ContainerStateMachine::FreeAll()
{
	if (m_ppxState)
	{
		for (u_int u=0; u<m_uNumStates; ++u)
		{
			delete m_ppxState[u];
			m_ppxState[u] = 0;
		}
	}

	delete [] m_ppxState;
	m_ppxState = 0;
	m_uNumStates = 0;

	m_uNumInputs = 1;
}

// Chunkstream code
const int iCHUNK_VERSION_TOOLS_CONTAINER_STATE_MACHINE = 1;

void Tools_ContainerStateMachine::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	FreeAll();

	int iVersion = -1;
	xStream >> iVersion;

	if (iVersion < 0 || iVersion > iCHUNK_VERSION_TOOLS_CONTAINER_STATE_MACHINE)
	{
		Asura_Assertf(0, ("Unknown chunk version (%d) in Tools_ContainerStateMachine", iVersion));
		return;
	}

	xStream >> m_uNumStates;
	if (m_uNumStates == 0) return;

	m_ppxState = new Tools_ContainerState*[m_uNumStates];
	Asura_Assert(m_ppxState, "Null pointer");
	if (!m_ppxState) return;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		m_ppxState[u] = new Tools_ContainerState;
		Asura_Assert(m_ppxState[u], "Null pointer");
		if (m_ppxState[u])
		{
			m_ppxState[u]->ReadFromChunkStream(xStream);
		}
	}

	if (iVersion >= 1)
	{
		xStream >> m_uNumInputs;
		Asura_Assert(m_uNumInputs <= uASURA_MAX_NUM_CONTAINER_INPUTS, "Too many inputs in state machine!");
	}
}

void Tools_ContainerStateMachine::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << iCHUNK_VERSION_TOOLS_CONTAINER_STATE_MACHINE;
	xStream << m_uNumStates;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		Asura_Assert(m_ppxState[u], "Error in WriteToChunkStream, null pointer.");
		if (m_ppxState[u])
		{
			m_ppxState[u]->WriteToChunkStream(xStream);
		}
	}

	xStream << m_uNumInputs;
}

//////////////
// AddState //
//////////////
void Tools_ContainerStateMachine::AddState(const Asura_String& xStateName)
{
	Tools_ContainerState** ppxState = new Tools_ContainerState*[m_uNumStates+1];
	Asura_Assert(ppxState, "Null pointer");
	if (!ppxState) return;
	
	for (u_int u=0; u<m_uNumStates; ++u)
	{
		ppxState[u] = (m_ppxState ? m_ppxState[u] : 0);
	}

	ppxState[m_uNumStates] = new Tools_ContainerState;
	Asura_Assert(ppxState[m_uNumStates], "Null pointer");
	if (ppxState[m_uNumStates])
	{
		ppxState[m_uNumStates]->SetInitialStateName(xStateName);
	}

	delete [] m_ppxState;
	m_ppxState = ppxState;
	++m_uNumStates;
}

/////////////////
// RemoveState //
/////////////////
void Tools_ContainerStateMachine::RemoveState(Asura_Hash_ID uStateHash)
{
	if (m_uNumStates == 0)
	{
		Asura_Assertf(0, ("State (0x%08x) not found for removal", uStateHash));
		return;
	}

	const int iIndex = FindStateIndex(uStateHash);
	if ( !IsValid(iIndex) )
	{
		Asura_Assertf(0, ("State (0x%08x) not found for removal", uStateHash));
		return;
	}

	if (m_uNumStates == 1)
	{
		// Removed the last state. Clear out all the data.
		FreeAll();
		return;
	}

	// Delete the state to be removed.
	delete m_ppxState[iIndex];
	m_ppxState[iIndex] = 0;

	// Move everything above the state to be removed down one place.
	for (u_int u = static_cast<u_int>(iIndex); u < (m_uNumStates-1); ++u)
	{
		m_ppxState[u] = m_ppxState[u+1];
	}

	m_ppxState[m_uNumStates-1] = 0;

	--m_uNumStates;
}

////////////////////
// FindStateIndex //
////////////////////
// Find index of state or return -1 if not found.
int Tools_ContainerStateMachine::FindStateIndex(Asura_Hash_ID uStateHash) const
{
	if (uStateHash != ASURA_HASH_ID_UNSET)
	{
		for (u_int u=0; u<m_uNumStates; ++u)
		{
			if (m_ppxState[u] && m_ppxState[u]->GetStateHash() == uStateHash)
			{
				return static_cast<int>(u);
			}
		}
	}

	return -1;
}

///////////////
// FindState //
///////////////
const Tools_ContainerState* Tools_ContainerStateMachine::FindState(Asura_Hash_ID uStateHash) const
{
	const int iIndex = FindStateIndex(uStateHash);
	if (IsValid(iIndex))
	{
		return m_ppxState[iIndex];
	}
	return 0;
}

///////////////
// FindState //
///////////////
Tools_ContainerState* Tools_ContainerStateMachine::FindState(Asura_Hash_ID uStateHash)
{
	const int iIndex = FindStateIndex(uStateHash);
	if (IsValid(iIndex))
	{
		return m_ppxState[iIndex];
	}
	return 0;
}


//////////////
// GetState //
//////////////
const Tools_ContainerState* Tools_ContainerStateMachine::GetState(u_int uIndex) const
{
	if (IsValid(uIndex))
	{
		if (m_ppxState)
		{
			return m_ppxState[uIndex];
		}
	}
	return 0;
}

//////////////
// GetState //
//////////////
Tools_ContainerState* Tools_ContainerStateMachine::GetState(u_int uIndex)
{
	if (IsValid(uIndex))
	{
		if (m_ppxState)
		{
			return m_ppxState[uIndex];
		}
	}
	return 0;
}

/////////////
// IsValid //
/////////////
bool Tools_ContainerStateMachine::IsValid(int iIndex) const
{
	if (iIndex == -1) return false;

	if (iIndex >= 0 && iIndex < static_cast<int>(m_uNumStates))
	{
		return true;
	}

	Asura_Assertf(0, ("Invalid container state index (%d)", iIndex));
	return false;
}

///////////////////
// GetChildState //
///////////////////
// Get child state, returns "off" state hash if child is disabled or unknown.
Asura_Hash_ID Tools_ContainerStateMachine::GetChildState(Asura_Hash_ID uStateHash, Asura_Hash_ID uChildContainerHash, u_int uInputIndex) const
{
	const int iIndex = FindStateIndex(uStateHash);
	if ( !IsValid(iIndex) )
	{
		Asura_Assertf(0, ("State (0x%08x) not found for GetChildState", uStateHash));
		return Asura_ContainerState::GetOffStateHash(); // invalid state
	}

	Asura_Assert(m_ppxState[iIndex], "Null pointer");
	if (m_ppxState[iIndex])
	{
		return m_ppxState[iIndex]->GetChildState(uInputIndex, uChildContainerHash);
	}
	return Asura_ContainerState::GetOffStateHash(); // invalid state
}

//////////////////
// GetStateName //
//////////////////
static Asura_String s_xInvalid("*Invalid*");

const Asura_String& Tools_ContainerStateMachine::GetStateName(u_int uIndex) const
{
	if ( IsValid(uIndex) )
	{
		if (m_ppxState && m_ppxState[uIndex])
		{
			return m_ppxState[uIndex]->GetStateName();
		}
	}
	
	return s_xInvalid; // invalid state
}

//////////////////
// GetStateHash //
//////////////////
const Asura_Hash_ID Tools_ContainerStateMachine::GetStateHash(u_int uIndex) const
{
	if ( IsValid(uIndex) )
	{
		if (m_ppxState && m_ppxState[uIndex])
		{
			return m_ppxState[uIndex]->GetStateHash();
		}
	}
	
	return ASURA_HASH_ID_UNSET; // invalid state
}

//////////////////
// ReplaceChild //
//////////////////
// Replace the child for all states.
// Typically called when we insert or remove an attachment.
void Tools_ContainerStateMachine::ReplaceChild(const Asura_Hash_ID uOldChildHash, const Asura_Hash_ID uNewChildHash)
{
	if (!m_ppxState) return;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u])
		{
			m_ppxState[u]->ReplaceChild(uOldChildHash, uNewChildHash);
		}
	}
}

//////////////////////////
// RemoveChildAllInputs //
//////////////////////////
// Remove the child for all states and all input slots.
void Tools_ContainerStateMachine::RemoveChildAllInputs(const Asura_Hash_ID uChildContainerHash)
{
	if (!m_ppxState) return;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u])
		{
			m_ppxState[u]->RemoveChildAllInputs(uChildContainerHash);
		}
	}
}

/////////////////////////
// GetNumVariantStates //
/////////////////////////
// Return number of states excluding invariant ones.
u_int Tools_ContainerStateMachine::GetNumVariantStates() const
{
	u_int uNum = 0;

	if (m_ppxState)
	{
		for (u_int u=0; u<m_uNumStates; ++u)
		{
			if (m_ppxState[u] && !m_ppxState[u]->IsInvariant())
			{
				++uNum;
			}
		}
	}
	return uNum;
}

/////////////////////
// CopyDataToAsura //
/////////////////////
// Copy the data to the Asura container collection.
void Tools_ContainerStateMachine::CopyDataToAsura(Asura_Container_Collection* pxAsuraCollection) const
{
	if (!pxAsuraCollection) return;

	const u_int uNumVariantStates = GetNumVariantStates();
		
	Asura_ContainerStateMachine& xMachine = pxAsuraCollection->GetStateMachine();
	xMachine.SetNumInputs(m_uNumInputs);
	xMachine.AllocateStates(uNumVariantStates);

	if (uNumVariantStates == 0) return;
	if (!m_ppxState) return;

	u_int uStateIndex = 0;
	for (u_int u=0; u<m_uNumStates; ++u)
	{
		const Tools_ContainerState* const pxState = m_ppxState[u];
		if (pxState && !pxState->IsInvariant())
		{
			const u_int uNumChild = pxState->GetNumChildStates();
			const u_int uNumDamageOverrides = pxState->GetNumDamageOverrides();

			xMachine.SetState(uStateIndex, pxState->GetStateHash(), uNumChild, uNumDamageOverrides);

			for (u_int v=0; v<uNumChild; ++v)
			{
				xMachine.SetStateChild(uStateIndex, v, pxState->GetStateEntry(v));
			}

			for (u_int v=0; v<uNumDamageOverrides; ++v)
			{
				const Asura_ContainerState_DamageOverride* pxOverride = pxState->GetDamageOverride(v);
				Asura_Assert(pxOverride, "Null pointer");
				if (pxOverride)
				{
					xMachine.SetStateDamageOverride(uStateIndex, v, pxOverride->m_uOverrideStateHash, pxOverride->m_fMaxHealthPercent);
				}
			}

			++uStateIndex;
		}
	}

	Asura_Assert(uNumVariantStates == uStateIndex, "Index mismatch in CopyDataToAsura.");
}

//////////////////
// ReplaceState //
//////////////////
// Replaced a state hash associated with a specified container collection. Fix up state machine.
// Return true if modified.
bool Tools_ContainerStateMachine::ReplaceState(const Asura_Hash_ID uContainerHash,
											   const Asura_Hash_ID uOldStateHash,
											   const Asura_Hash_ID uNewStateHash)
{
	if (!m_ppxState) return false;

	bool bModified = false;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u])
		{
			if (m_ppxState[u]->ReplaceState(uContainerHash, uOldStateHash, uNewStateHash))
			{
				bModified = true;
			}
		}
	}

	return bModified;
}

////////////////////////////////
// ReplaceDamageOverrideState //
////////////////////////////////
// Replace the state hash in the damage overrides.
// Return true if modified.
bool Tools_ContainerStateMachine::ReplaceDamageOverrideState(const Asura_Hash_ID uOldStateHash, const Asura_Hash_ID uNewStateHash)
{
	if (!m_ppxState) return false;

	bool bModified = false;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u])
		{
			if (m_ppxState[u]->ReplaceDamageOverrideState(uOldStateHash, uNewStateHash))
			{
				bModified = true;
			}
		}
	}

	return bModified;
}

////////////
// Verify //
////////////
// Check for errors. Return number of errors found.
void Tools_ContainerStateMachine::Verify(const Tools_Container* pxParent, const Tools_Container_ChildIterator& xIt) const
{
	if (!pxParent) return;
	if (!m_ppxState) return;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u])
		{
			m_ppxState[u]->Verify(pxParent, xIt);
		}
	}
}

/////////////////
// FixUpErrors //
/////////////////
// Silently fix up certain errors that the user can't fix.
void Tools_ContainerStateMachine::FixUpErrors(const Tools_Container* pxParent)
{
	if (!pxParent) return;
	if (!m_ppxState) return;

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u])
		{
			m_ppxState[u]->FixUpErrors(pxParent);
		}
	}
}

////////////////////////////
// GetHashForRenamedState //
////////////////////////////
// Does the hash belong to state that has been renamed? If so then return the new state hash.
Asura_Hash_ID Tools_ContainerStateMachine::GetHashForRenamedState(const Asura_Hash_ID uOldStateHash) const
{
	if (uOldStateHash == ASURA_HASH_ID_UNSET)
	{
		return ASURA_HASH_ID_UNSET;
	}

	const int iIndex = FindStateIndex(uOldStateHash);
	if (IsValid(iIndex))
	{
		// This hash is currently valid, so it's not an out-of-date hash.
		return ASURA_HASH_ID_UNSET;
	}

	for (u_int u=0; u<m_uNumStates; ++u)
	{
		if (m_ppxState[u] && m_ppxState[u]->HasOldStateHash(uOldStateHash) )
		{
			// Found it. Return the new hash for this state.
			return m_ppxState[u]->GetStateHash();
		}
	}

	// Not found. Perhaps it belongs to a state that has been deleted.
	return ASURA_HASH_ID_UNSET;
}

//////////////
// AddInput //
//////////////
// Return true if the number of inputs was changed.
bool Tools_ContainerStateMachine::AddInput()
{
	if (m_uNumInputs < uASURA_MAX_NUM_CONTAINER_INPUTS)
	{
		++m_uNumInputs;
		return true;
	}
	return false;
}

/////////////////
// RemoveInput //
/////////////////
// Return true if the number of inputs was changed.
bool Tools_ContainerStateMachine::RemoveInput()
{
	if (m_uNumInputs > 1)
	{
		--m_uNumInputs;
		return true;
	}
	return false;
}

///////////////////////////////////////////////////////////////////////////////
// Tools_ContainerSortedStateList
///////////////////////////////////////////////////////////////////////////////
Tools_ContainerSortedStateList* Tools_ContainerSortedStateList::s_pxThis = 0;

Tools_ContainerSortedStateList::Tools_ContainerSortedStateList(Tools_ContainerStateMachine& xStateMachine)
: m_xStateMachine(xStateMachine)
, m_puIndices(0)
{
	// Make a sorted index list.
	if (GetNumStates() == 0) return;
	
	m_puIndices = new u_int[GetNumStates()];
	for (u_int u=0; u<GetNumStates(); ++u)
	{
		m_puIndices[u] = u;
	}

	s_pxThis = this;
	qsort(m_puIndices, GetNumStates(), sizeof(u_int), &Compare_Callback);
	s_pxThis = 0;
}

Tools_ContainerSortedStateList::~Tools_ContainerSortedStateList()
{
	delete [] m_puIndices;
}

int Tools_ContainerSortedStateList::Compare_Callback(const void* pFirst, const void* pSecond)
{
	if (s_pxThis && pFirst && pSecond)
	{
		return s_pxThis->Compare(*static_cast<const u_int*>(pFirst), *static_cast<const u_int*>(pSecond));
	}
	return 0;
}

int Tools_ContainerSortedStateList::Compare(u_int uFirst, u_int uSecond) const
{
	const char* const szFirst  = m_xStateMachine.GetStateName(uFirst).GetString();
	const char* const szSecond = m_xStateMachine.GetStateName(uSecond).GetString();

	return stricmp(szFirst, szSecond);
}

// Return the state for the given index.
Tools_ContainerState* Tools_ContainerSortedStateList::GetState(u_int uIndex) const
{
	if (uIndex >= GetNumStates()) return 0;
	return m_xStateMachine.GetState(m_puIndices[uIndex]);
}

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM
