///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Container_Collection.cpp
//
// Created by Martin G.Harris (3/7/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Entity/Container/Tools_Container_Collection.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Asura
#include "Core/Asura_StringBuffer.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Container/Asura_Container_System.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Gadget.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Core/Asura_UserInput.h"

// Tools
#include "Entity/Container/Tools_Container_Attachment.h"
#include "Entity/Container/Tools_Container_System.h"
#include "Entity/Container/Tools_ContainerFile.h"
#include "Entity/Container/Tools_ContainerRegistry.h"
#include "Entity/Container/Tools_Entity_Container.h"
#include "Entity/Container/Tools_Mode_Container.h"
#include "Misc/Tools_Dialog.h"
#include "Misc/Tools_MessageLog.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Tools_System.h"

//
// Register the Tools container types.
//

Tools_ContainerRegisterClass<Tools_Container_Collection> xRegisterCollection(AsuraContainerClass_Collection, "Collection");


// Defined in Tools_Mode_Container
extern bool s_bIconForAttachments;

const char* s_szBlankLabelForContainer = "";

const int iTOOLS_COLLECTION_DAMAGEOVERRIDE_ID = Asura_GetHashID("DamageSubwindow");
const int iTOOLS_COLLECTION_INPUTS_ID = Asura_GetHashID("InputsSubwindow");


///////////////////////////////////////////////////////////////////////////////
// Tools_Container_ChildIterator_Sorted
///////////////////////////////////////////////////////////////////////////////

// Child iterator, but sorted in to alphabetical order. Order should match for SortTreeAlphabetically.

class Tools_Container_ChildIterator_Sorted : public Tools_Container_ChildIterator
{
	// Data used for the sorting, must match that used by SortTreeAlphabetically.
	struct SortData
	{
		SortData()
			: m_uContainerHash(ASURA_HASH_ID_UNSET)
		{
			m_xLabel.MakeBlank();
		}

		Asura_String256 m_xLabel;			// Label e.g. "GraphColour" or "Mesh (control_panel)".
		Asura_Hash_ID	m_uContainerHash;	// Container hash (NB for attachments, it's the child's hash if there is one).
	};

public:
	// If bInlineAttachments is true then we sort according to the names
	// of the attachment's children rather than the attachments themselves.
	// This is used to match the ordering in the instance overlay tree when attachments are inlined as icons.
	Tools_Container_ChildIterator_Sorted(const Tools_Container* pxContainer, bool bInlineAttachments)
		: Tools_Container_ChildIterator(pxContainer)
		, m_pxSortData(0)
	{
		if (m_iNumArrayEntries > 1)
		{
			m_pxSortData = new SortData[m_iNumArrayEntries];
			
			FillStringArray(bInlineAttachments);
			Sort();
			Begin();

			delete [] m_pxSortData;
			m_pxSortData = 0;
		}
	}

	~Tools_Container_ChildIterator_Sorted()
	{
	}

private:
	// Fill out the array with the strings used for sorting.
	void FillStringArray(bool bInlineAttachments)
	{
		if (!m_pxSortData) return;
		if (!m_puHash) return;
	
		for (int i=0; i<m_iNumArrayEntries; ++i)
		{
			const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_puHash[i]);
			if (pxContainer)
			{
				if (bInlineAttachments)
				{
					// For attachments with children, skip past the attachment.
					if (pxContainer->GetClassification() == AsuraContainerClass_Attachment)
					{
						Tools_Container_ChildIterator xAttachIt(pxContainer);
						if (xAttachIt.GetNumChildren() > 0)
						{
							const Tools_Container* pxChild = xAttachIt.GetCurrent();
							if (pxChild)
							{
								pxContainer = pxChild;
							}
						}
					}
				}

				m_pxSortData[i].m_uContainerHash = pxContainer->GetContainerHash();
				pxContainer->AddInstanceName(m_pxSortData[i].m_xLabel, false, true);
			}
		}
	}


	// Sort by alphabetical order.
	void Sort()
	{
		if (!m_pxSortData) return;
		if (!m_puHash) return;
		if (!m_piMultiplicity) return;

		// Sort the array of indices.
		int* piIndex = new int[m_iNumArrayEntries];
		for (int i=0; i<m_iNumArrayEntries; ++i)
		{
			piIndex[i] = i;
		}

		s_pxThis = this;
		qsort(&piIndex[0], m_iNumArrayEntries, sizeof(int), &Compare);
		s_pxThis = 0;

		// Now use the sorted array to fix up the data in the base class.
		Asura_Hash_ID* puNewHash = new Asura_Hash_ID[m_iNumArrayEntries];
		int* piNewMultiplicity = new int[m_iNumArrayEntries];

		for (int i=0; i<m_iNumArrayEntries; ++i)
		{
			puNewHash[i] = m_puHash[piIndex[i]];
			piNewMultiplicity[i] = m_piMultiplicity[piIndex[i]];
		}

		delete [] m_puHash;
		m_puHash = puNewHash;

		delete [] m_piMultiplicity;
		m_piMultiplicity = piNewMultiplicity;

		m_iMaxArraySize = m_iNumArrayEntries;

		// Clean up.
		delete [] piIndex;
		piIndex = 0;
	}

	// Comparison function.
	static int Compare(const void * pOne, const void * pTwo)
	{
		const int iIndex1 = * static_cast<const int*>(pOne);
		const int iIndex2 = * static_cast<const int*>(pTwo);

		if (!IsIndexValid(iIndex1) || !IsIndexValid(iIndex2))
		{
			return 0;
		}

		int iResult = stricmp( GetString(iIndex1), GetString(iIndex2) );
		if (iResult == 0)
		{
			const u_int uUserData1 = GetUserData(iIndex1);
			const u_int uUserData2 = GetUserData(iIndex2);
			if (uUserData1 == uUserData2)
			{
				iResult = 0;
			}
			else
			{
				iResult = ((uUserData1 < uUserData2) ? -1 : +1);
			}
		}
		return iResult;
	}

	static bool IsIndexValid(int iIndex)
	{
		if (iIndex < 0 || iIndex >= s_pxThis->m_iNumArrayEntries)
		{
			Asura_Assert(0, "Out of range index");
			return false;
		}
		return true;
	}

	static const char* GetString(int iIndex)
	{
		return s_pxThis->m_pxSortData[iIndex].m_xLabel.GetString();
	}

	static u_int GetUserData(int iIndex)
	{
		return s_pxThis->m_pxSortData[iIndex].m_uContainerHash;
	}

private:
	SortData* m_pxSortData;

	static Tools_Container_ChildIterator_Sorted* s_pxThis;
};

// Static data
Tools_Container_ChildIterator_Sorted* Tools_Container_ChildIterator_Sorted::s_pxThis = 0;


///////////////////////////////////////////////////////////////////////////////
// Tools_ContainerStateEdit_Dialog_Item
///////////////////////////////////////////////////////////////////////////////
class Tools_ContainerStateEdit_Dialog_Item
{
	typedef Tools_ContainerStateEdit_Dialog_Item THIS_CLASS;

public:
	Tools_ContainerStateEdit_Dialog_Item(Asura_Subwindow_Dialog* pxSubwindowDialog,
										  Asura_Vector_2& xOffset,
										  const Tools_Container* pxContainer,
										  Tools_ContainerStateEdit_Dialog_Item* pxParent)
		: m_pxParent(pxParent)
		, m_pxRoot(0)
		, m_uContainerHash(pxContainer ? pxContainer->GetContainerHash() : ASURA_HASH_ID_UNSET)
		, m_uContainerToGetStateFrom(ASURA_HASH_ID_UNSET)
		, m_uClassification(AsuraContainerClass_NotSet)
		, m_fLeftOffsetX(xOffset.x)
	{
		// Set state inputs "off" by default.
		const u_int uNumInputs = (pxContainer ? pxContainer->GetNumInputs() : 1);
		m_xInputState.SetNumInputs(uNumInputs);
		m_xInputState.SetOff();
		m_xNewInputState = m_xInputState;

		m_pxRoot = (m_pxParent ? m_pxParent->m_pxRoot : this);
		
		Asura_String256 xBuffer("*None*");
		if (pxContainer)
		{
			xBuffer.MakeBlank();
			pxContainer->AddInstanceName(xBuffer, false, true);
			m_uClassification = pxContainer->GetClassification();
		}

		m_xLabel = xBuffer.GetString();
		
		const Asura_Hash_ID uStateForEdit = Tools_Container_Collection::GetStateSelectedForEditing();
		
		// Only immediate children of the root can be changed to edit the state definition.
		bool bEnabled = (pxParent != 0 && pxParent->m_pxParent == 0);

		// "On"/"off" states can't be edited as they are invariant.
		if (uStateForEdit == Asura_ContainerState::GetOnStateHash() ||
			uStateForEdit == Asura_ContainerState::GetOffStateHash())
		{
			bEnabled = false;
		}

		// The root collection's state can be changed however to change the overall preview state.
		if (pxParent == 0)
		{
			bEnabled = true;
		}

		const bool bIsInlineAttachment = s_bIconForAttachments && m_uClassification == AsuraContainerClass_Attachment;
		const bool bIsParentInlineAttachment = s_bIconForAttachments && m_pxParent && m_pxParent->m_uClassification == AsuraContainerClass_Attachment;

		Asura_Dialog* pxDialog = 0;

		// Add drop down list for collections (or attachments with child collections),
		// add tick boxes for other containers.
		m_uContainerToGetStateFrom = GetContainerToGetStateFrom();
	
		if (!bIsParentInlineAttachment)
		{
			if (m_pxParent)
			{
				// Indent the tickbox.
				const float fIndentX = 2.5f;
				m_fLeftOffsetX = m_pxParent->m_fLeftOffsetX + fIndentX;
				xOffset.x = m_fLeftOffsetX;
			}

			// Add tickbox/dropdown, with no label.
			if (m_uContainerToGetStateFrom == ASURA_HASH_ID_UNSET)
			{
				// Only "on"/"off" states. Only a single input.
				pxDialog = new Asura_FunctionTickbox_Dialog(s_szBlankLabelForContainer, xOffset, this,
															IsStateChildBoxChecked, SetStateChildBoxChecked);

				// Grey out tickbox/dropdown if it can't be modified.
				pxDialog->SetEnabled(bEnabled);
				if (pxContainer)
				{
					pxDialog->SetHelpText(pxContainer->GetContainerName().GetString());
				}
				pxSubwindowDialog->AddDialog(pxDialog);
				
				// Move in to position for the label.
				xOffset.x = pxDialog->GetMaxCoords().x;
			}
			else
			{
				// Drop down list for collections with potentially many states. Always add "on" and "off" states.
				// May have multiple input states.
				Tools_Container_MultipleState_Dialog* pxMultiDialog = 0;

				pxMultiDialog = new Tools_Container_MultipleState_Dialog(pxSubwindowDialog,
																		bEnabled,
																		xOffset,
																		&m_uContainerToGetStateFrom,
																		m_xNewInputState.GetNumInputs(),
																		m_xNewInputState.GetStateHashArray());

				if (pxContainer)
				{
					pxMultiDialog->SetHelpTextForChildren(pxContainer->GetContainerName().GetString());
				}
				pxSubwindowDialog->AddDialog(pxMultiDialog);

				// Move in to position for the label.
				xOffset.x = pxMultiDialog->GetMaxCoordX();
			}
		}
		else if (m_pxParent)
		{
			m_fLeftOffsetX = m_pxParent->m_fLeftOffsetX;
		}

		// Now add the label on same line (added separately so that it isn't greyed out).
		Tools_Container_ChildIterator_Sorted xIt(pxContainer, s_bIconForAttachments);
		if (!bIsInlineAttachment || xIt.GetNumChildren() == 0)
		{
			pxDialog = new Asura_Dialog(m_xLabel.GetString(), xOffset);
			if (pxContainer)
			{
				pxDialog->SetHelpText(pxContainer->GetContainerName().GetString());
			}
			pxSubwindowDialog->AddDialog(pxDialog);
			
			if (m_pxParent == 0)
			{
				// Extra spacing after the root collection.
				xOffset.y += Asura_Dialog::LargeSpacing();
			}
			else
			{
				xOffset.y += Asura_Dialog::DefaultSpacing();
			}
		}

		// Sort into alphabetical order (needs to be the same order as for the tree for the instance overlay).
		for (; !xIt.Done(); xIt.Next())
		{
			const Tools_Container* pxChildContainer = xIt.GetCurrent();
			if (pxChildContainer)
			{
				m_xChildren.AddToList( new Tools_ContainerStateEdit_Dialog_Item(pxSubwindowDialog, xOffset, pxChildContainer, this) );
			}
		}
	}
									
	~Tools_ContainerStateEdit_Dialog_Item()
	{
		while (!m_xChildren.IsEmpty())
		{
			delete m_xChildren.PopHeadObject();
		}
	}

	Asura_Hash_ID GetContainerHash() const { return m_uContainerHash; }

	// ChangeState
	void ChangeState(const Asura_Container_InputState& xInputState)
	{
		if (m_xInputState != m_xNewInputState)
		{
			// The state has been modified using the drop down list.
			SetStateChildDropDownList(m_xNewInputState);
			m_xInputState = m_xNewInputState;
		}
		else
		{
			m_xInputState = xInputState;
			m_xNewInputState = xInputState;
		}

		// Pass state on to children for attachments and collections.
		Asura_Container_OutputState xOutputState;

		const Asura_Container* const pxContainer = Asura_Container_System::FindContainer(m_uContainerHash);
		Asura_Assertf(pxContainer, ("Container (0x%08x) not found", m_uContainerHash));
		if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Collection)
		{
			const Asura_Container_Collection* const pxCollection = static_cast<const Asura_Container_Collection*>(pxContainer);
			pxCollection->ConvertState(m_xInputState, 100.0f, xOutputState);	
		}

		const bool bAttachment = (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Attachment);

		for (Asura_List_It<Tools_ContainerStateEdit_Dialog_Item> xIt(m_xChildren); !xIt.Done(); xIt.Next())
		{
			Tools_ContainerStateEdit_Dialog_Item* pxChild = xIt.GetCurrent();
			if (!pxChild) continue;
			
			if (bAttachment)
			{
				// Attachments just pass the input state to their child.
				pxChild->ChangeState(m_xInputState);
			}
			else
			{
				// Collections modify the state before passing it to their children.
				const Asura_Hash_ID uChildContainerHash = pxChild->GetContainerHash();
		
				Asura_Container_InputState xChildInputState;
				xOutputState.GetChildInputState(uChildContainerHash, xChildInputState);

				pxChild->ChangeState(xChildInputState);
			}
		}
	}

	// Get the container for the root collection.
	Tools_Container* GetRootContainer() const
	{
		Asura_Assert(m_pxRoot, "Null root pointer");
		if (!m_pxRoot) return 0;

		Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_pxRoot->GetContainerHash());
		Asura_Assert(pxContainer, "Null root container pointer");
		return pxContainer;
	}

	// Get the state for the root collection that is being edited.
	Tools_ContainerState* GetRootState() const
	{
		Asura_Assert(m_pxRoot, "Null root pointer");
		if (!m_pxRoot) return 0;

		Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(m_pxRoot->GetContainerHash());
		Asura_Assert(pxContainer, "Null root container pointer");
		if (!pxContainer) return 0;
		
		Tools_ContainerStateMachine* pxStateMachine = pxContainer->GetStateMachine();
		Asura_Assert(pxStateMachine, "Null container state machine pointer");
		if (!pxStateMachine) return 0;
		
		const Asura_Hash_ID uStateForEditing = Tools_Container_Collection::GetStateSelectedForEditing();
		const Asura_Hash_ID uCheckHash = m_pxRoot->m_xInputState.GetStateHash(Tools_Container_Collection::GetStateIndexForEditing());
		if (uStateForEditing != uCheckHash)
		{
			Asura_Assert(uCheckHash == uStateForEditing, "State not set correctly");
			return 0;
		}

		Tools_ContainerState* const pxState = pxStateMachine->FindState(uStateForEditing);
		Asura_Assert(pxState, "Null state pointer");
		return pxState;
	}

	// Collections can have many states, as can attachments that have collections as children (or grandchildren...).
	// Everything else has an on/off state.
	// Note: this means that only collections can have multiple children, as only collections have state machines.
	// Anything else with children (i.e. attachments) inherits the state list from its child.

	// Get container hash for container with state to be used for drop down list
	// or return ASURA_HASH_ID_UNSET if not a collection (in which case use a tickbox).	
	Asura_Hash_ID GetContainerToGetStateFrom()
	{
		Tools_Container* pxCurrent = Tools_Container_System::FindContainerByHash(m_uContainerHash);
		while (pxCurrent)
		{
			if (pxCurrent->GetClassification() == AsuraContainerClass_Collection)
			{
				return pxCurrent->GetContainerHash();
			}

			Tools_Container_ChildIterator xIt(pxCurrent);
			if (xIt.GetNumChildren() == 0)
			{
				return ASURA_HASH_ID_UNSET;
			}

			pxCurrent = xIt.GetCurrent();
		}

		return ASURA_HASH_ID_UNSET;
	}

	// Callback for tickbox. Is box ticked?
	bool IsStateChildBoxChecked() const
	{
		return !m_xInputState.IsOff();
	}

	// Callback for tickbox. Change tick state. Always in input slot 0.
	void SetStateChildBoxChecked(bool bSet)
	{
		// Only entries immediately below the root collection can be modified.
		const bool bEnabled = (m_pxParent != 0 && m_pxParent->m_pxParent == 0);
		if (!bEnabled) return;
	
		Tools_ContainerState* const pxState = GetRootState();
		if (!pxState) return;

		const u_int uInputIndex = 0;

		const Asura_Hash_ID uChildState = pxState->GetChildState(uInputIndex, m_uContainerHash);
		const bool bWasSet = (uChildState != Asura_ContainerState::GetOffStateHash());
	
		if (bWasSet != bSet)
		{
			if (bSet)
			{
				pxState->ChangeChild(m_uContainerHash, Asura_ContainerState::GetOnStateHash(), uInputIndex);
			}
			else
			{
				pxState->RemoveChild(m_uContainerHash, uInputIndex);
			}

			// Mark as modified.
			Tools_Container* pxContainer = GetRootContainer();
			if (pxContainer)
			{
				pxContainer->SetModifiedStates();
			}
		}
	}

	// Changed selection in drop down list.
	void SetStateChildDropDownList(const Asura_Container_InputState& xNewInputState)
	{
		if (m_pxParent == 0)
		{
			// Change was made to the state for the root collection.
			// Find the state input that was changed.
			u_int uInputIndex = 0;
			Asura_Hash_ID uNewStateHash = xNewInputState.GetStateHash(0);

			for (u_int u=0; u<xNewInputState.GetNumInputs(); ++u)
			{
				if (m_xInputState.GetStateHash(u) != xNewInputState.GetStateHash(u))
				{
					uInputIndex = u;
					uNewStateHash = xNewInputState.GetStateHash(u);
				}
			}

			// Change the overall preview state.
			Tools_Container_Collection::ChangedPreviewStateFromStateDialogue(m_uContainerHash, uNewStateHash, uInputIndex);
			return;
		}

		// Only entries immediately below the root collection can be modified.
		const bool bEnabled = (m_pxParent != 0 && m_pxParent->m_pxParent == 0);
		if (!bEnabled) return;
	
		Tools_ContainerState* const pxState = GetRootState();
		if (!pxState) return;

		Asura_Assert(m_xInputState.GetNumInputs() == xNewInputState.GetNumInputs(), "Mismatch in number of inputs");

		bool bModified = false;

		for (u_int u=0; u<xNewInputState.GetNumInputs(); ++u)
		{
			const Asura_Hash_ID uChildStateHash = pxState->GetChildState(u, m_uContainerHash);
			const Asura_Hash_ID uNewStateHash = xNewInputState.GetStateHash(u);
			if (uChildStateHash != uNewStateHash)
			{
				pxState->ChangeChild(m_uContainerHash, uNewStateHash, u);
				bModified = true;
			}
		}

		// Mark as modified.
		if (bModified)
		{
			Tools_Container* pxContainer = GetRootContainer();
			if (pxContainer)
			{
				pxContainer->SetModifiedStates();
			}
		}
	}

private:
	// Helper
	static THIS_CLASS* GetThis(void* pThis) { return reinterpret_cast<THIS_CLASS*>(pThis); }

	// Callbacks
	static bool  IsStateChildBoxChecked(void* pThis) { return GetThis(pThis)->IsStateChildBoxChecked(); }
	static void SetStateChildBoxChecked(bool bSet, void* pThis) { GetThis(pThis)->SetStateChildBoxChecked(bSet); }

private:
	Tools_ContainerStateEdit_Dialog_Item* m_pxParent;
	Tools_ContainerStateEdit_Dialog_Item* m_pxRoot;

	Asura_Hash_ID m_uContainerHash;
	Asura_Hash_ID m_uContainerToGetStateFrom; // or ASURA_HASH_ID_UNSET if this is a tickbox.
	
	Asura_Container_InputState m_xInputState;		// State
	Asura_Container_InputState m_xNewInputState;	// State as set by drop down list.

	u_int m_uClassification;

	float m_fLeftOffsetX; // X value of offset for the left side of the corresponding tickbox/dropdown. 

	Asura_String m_xLabel;

	Asura_List<Tools_ContainerStateEdit_Dialog_Item> m_xChildren;
};

///////////////////////////////////////////////////////////////////////////////
//  Tools_ContainerStateEdit_Dialog
///////////////////////////////////////////////////////////////////////////////
// Dialogue for string with state name being edited, and tree of collection and sub-containers.
class Tools_ContainerStateEdit_Dialog : public Asura_StringDisplay_Dialog
{
public:
	Tools_ContainerStateEdit_Dialog(Asura_Subwindow_Dialog* pxSubwindowDialog,
									 Asura_Vector_2& xSubOffset,
									 const Tools_Container* pxContainer,
									 const char* szName)
		: Asura_StringDisplay_Dialog(szName, xSubOffset)
		, m_pxRoot(0)
	{
		xSubOffset.y += Asura_Dialog::LargeSpacing();
		m_pxRoot = new Tools_ContainerStateEdit_Dialog_Item(pxSubwindowDialog, xSubOffset, pxContainer, 0);
		RefreshStateInfo();
	}

	~Tools_ContainerStateEdit_Dialog()
	{
		delete m_pxRoot;
		m_pxRoot = 0;
	}

	virtual void PerFrameUpdate()
	{
		Asura_StringDisplay_Dialog::PerFrameUpdate();
		RefreshStateInfo();
	}

	void RefreshStateInfo()
	{
		const Asura_Hash_ID uStateForEditing = Tools_Container_Collection::GetStateSelectedForEditing();
		const u_int uInputIndex = Tools_Container_Collection::GetStateIndexForEditing();
		const u_int uNumInputs = Tools_Container_Collection::GetNumInputsForEditing();

		// Set up the input with only a single state enabled.
		Asura_Container_InputState xInputState;
		xInputState.SetNumInputs(uNumInputs);
		xInputState.SetOff();
		xInputState.SetStateHash(uInputIndex, uStateForEditing);

		if (m_pxRoot)
		{
			m_pxRoot->ChangeState(xInputState);
		}
	}

private:
	Tools_ContainerStateEdit_Dialog_Item* m_pxRoot;
};


///////////////////////////////////////////////////////////////////////////////
// Tools_Container_Collection
///////////////////////////////////////////////////////////////////////////////

// Static data
Asura_Hash_ID Tools_Container_Collection::s_uContainerHashForEdit = ASURA_HASH_ID_UNSET;
Asura_Hash_ID Tools_Container_Collection::s_uStateHashForEdit = ASURA_HASH_ID_UNSET;
u_int Tools_Container_Collection::s_uStateIndexForEdit = 0;
Tools_Container_Collection* Tools_Container_Collection::s_pxThis = 0;

Asura_Vector_2 Tools_Container_Collection::s_xDimensionsDamage(35.0f, 9.0f);
Asura_Vector_2 Tools_Container_Collection::s_xDimensionsState (35.0f, 36.0f);
Asura_Vector_2 Tools_Container_Collection::s_xDimensionsInput (35.0f, 46.0f);

Tools_Container_Collection::Tools_Container_Collection()
: m_bNeedToCopyHashes(true)
, m_bNeedToCopyStates(true)
, m_bChangedInputMask(false)
{
}

Tools_Container_Collection::~Tools_Container_Collection()
{
	if (s_uContainerHashForEdit == GetContainerHash())
	{
		s_uContainerHashForEdit = ASURA_HASH_ID_UNSET;
	}

	if (s_pxThis == this)
	{
		s_pxThis = 0;
	}
}

// Chunkstream code
const int iCHUNK_VERSION_TOOLS_CONTAINER_COLLECTION = 1;

void Tools_Container_Collection::ReadFromChunkStream(Asura_Chunk_Stream& xStream)
{
	int iVersion = -1;
	xStream >> iVersion;

	if (iVersion < 0 || iVersion > iCHUNK_VERSION_TOOLS_CONTAINER_COLLECTION)
	{
		Asura_Assertf(0, ("Unknown chunk version (%d) in Tools_Container_Collection", iVersion));
		return;
	}

	u_int uNumHashes = 0;
	xStream >> uNumHashes;

	for (u_int u=0; u<uNumHashes; ++u)
	{
		Asura_Hash_ID uChildHash = ASURA_HASH_ID_UNSET;
		int iMultiplicity = 0;
		xStream >> uChildHash;
		xStream >> iMultiplicity;

		if (!m_xChildren.AddNode(uChildHash, iMultiplicity))
		{
			Asura_Assert(0, "Failed to add node");
		}
	}

	if (iVersion >= 1)
	{
		m_xStateMachine.ReadFromChunkStream(xStream);
	}

	PARENT::ReadFromChunkStream(xStream);
}

void Tools_Container_Collection::WriteToChunkStream(Asura_Chunk_Stream& xStream) const
{
	xStream << iCHUNK_VERSION_TOOLS_CONTAINER_COLLECTION;

	const u_int uNumHashes = static_cast<u_int>(m_xChildren.GetNumberOfNodes());
	xStream << uNumHashes;

	if (uNumHashes)
	{
		Asura_TemplateHashTree_SaveIterator xIt(uNumHashes);

		for ( ; !xIt.Done() ; xIt.Next())
		{
			const Asura_Hash_ID uHash = m_xChildren.GetHashByIndex(xIt.GetCurrent());
			const int* piMultiplicity = m_xChildren.GetDataByIndex(xIt.GetCurrent());
			const int iMultiplicity = (piMultiplicity ? (*piMultiplicity) : 1);

			xStream << uHash;
			xStream << iMultiplicity;
		}
	}

	m_xStateMachine.WriteToChunkStream(xStream);

	PARENT::WriteToChunkStream(xStream);
}

////////////////////////
// SetToDefaultState //
///////////////////////
void Tools_Container_Collection::SetToDefaultState()
{
	const Asura_Hash_ID uHashOff = Asura_ContainerState::GetOffStateHash();

	if (Tools_Mode_Container::IsInstanceOverlayOpen() && 
		Tools_Mode_Container::GetContainerToEditHash() == s_uContainerHashForEdit)
	{
		// Get state from instance overlay.
		const Asura_Container_InputState& xPreviewInputState = Tools_Mode_Container::GetPreviewState();
		if (s_uStateIndexForEdit >= xPreviewInputState.GetNumInputs())
		{
			s_uStateIndexForEdit = 0;
		}
		s_uStateHashForEdit = xPreviewInputState.GetStateHash(s_uStateIndexForEdit);
	}
	else if (s_uStateHashForEdit == uHashOff)
	{
		// Leave "off" state even if it has been deleted.
	}
	else
	{
		// Default to "off" even if it has been deleted.
		s_uStateHashForEdit = uHashOff;
		
		// Use first state found.
		for (u_int u=0; u<m_xStateMachine.GetNumStates(); ++u)
		{
			// Implement any limits upon which states are in which input slots.
			const Tools_ContainerState* const pxState = m_xStateMachine.GetState(u);
			if (pxState && pxState->IsEnabledForInputSlot(s_uStateIndexForEdit))
			{
				s_uStateHashForEdit = pxState->GetStateHash();
				break;
			}
		}
	}
}

////////////////////////
// AddOverlayControls //
////////////////////////
void Tools_Container_Collection::AddOverlayControls(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset)
{
	if (!m_pxAsuraContainer) return;

	// Refresh the dynamic string labels.
	RefreshDialogue();

	if (s_uContainerHashForEdit != GetContainerHash())
	{
		// Changed the container that is being edited.
		s_uContainerHashForEdit = GetContainerHash();

		if (Tools_Mode_Container::IsInstanceOverlayOpen() && 
			Tools_Mode_Container::GetContainerToEditHash() == s_uContainerHashForEdit)
		{
			// Get state from instance overlay.
			s_uStateHashForEdit = Tools_Mode_Container::GetPreviewStateHash(s_uStateIndexForEdit);
		}
	}

	// If there is no valid state selected then choose a default.
	const Tools_ContainerState* const pxState = m_xStateMachine.FindState(s_uStateHashForEdit);
	if (!pxState)
	{
		SetToDefaultState();	
	}

	// Refresh the dialogues.
	RefreshStateDialogues();

	s_pxThis = this;

	Asura_Dialog* pxDialog = 0;

	// Tabs dialogue
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset, 0, 0);
	xNewOverlay.AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	xNewOverlay.SetCurrentParentDialog(pxTabsDialog);

	const float fTabsStartY = xOffset.y;
	float fTabsEndY = xOffset.y;

	// State tab
	pxTabsDialog->AddTab("States");
	xOffset.y = fTabsStartY;
	
	// Add controls to modify states of children for the currently selected state.
	Asura_Subwindow_Dialog* pxStateChildOverlay = AddStateChildOverlay(xNewOverlay, xOffset);

	// Add controls to modify damage overrides of the various states.
	Asura_Subwindow_Dialog* pxDamageOverrideOverlay = AddDamageOverrideOverlay(xNewOverlay, xOffset);
	pxDamageOverrideOverlay->SetRelativePosition(pxStateChildOverlay, true, false);

	// Add string with number of states
	pxDialog = new Asura_Dynamic_Dialog(&m_xNumStatesDialogue, xOffset);
	pxDialog->SetRelativePosition(pxStateChildOverlay, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	// Add buttons
	pxDialog = new Asura_Function_Dialog("Create", xOffset, CreateState, reinterpret_cast<int>(this));
	pxDialog->SetHelpText("Create an new state.");
	pxDialog->SetRelativePosition(pxStateChildOverlay, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.x += 9.0f;

	pxDialog = new Asura_Function_Dialog("Rename", xOffset, RenameState, reinterpret_cast<int>(this), 0, IsStateSelected);
	pxDialog->SetHelpText("Rename the selected state.");
	pxDialog->SetRelativePosition(pxStateChildOverlay, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.x += 10.0f;

	pxDialog = new Asura_Function_Dialog("Delete", xOffset, DeleteState, reinterpret_cast<int>(this), 0, IsStateSelected);
	pxDialog->SetHelpText("Delete the selected state.");
	pxDialog->SetRelativePosition(pxStateChildOverlay, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	xOffset.x -= 19.0f;

	if (xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	// State tab
	pxTabsDialog->AddTab("Inputs");
	xOffset.y = fTabsStartY;

	// Add input subwindow
	Asura_Subwindow_Dialog* pxSubwindowDialog = AddInputOverlay(xNewOverlay, xOffset);

	// Add string with number of states
	pxDialog = new Asura_Dynamic_Dialog(&m_xNumInputsDialogue, xOffset);
	pxDialog->SetRelativePosition(pxSubwindowDialog, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	// Add buttons
	pxDialog = new Asura_Function_Dialog("Add", xOffset, AddInput, reinterpret_cast<int>(this), 0, CanAddInput);
	pxDialog->SetHelpText("Add a new input to the collection.");
	pxDialog->SetRelativePosition(pxSubwindowDialog, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.x += 9.0f;

	pxDialog = new Asura_Function_Dialog("Remove", xOffset, RemoveInput, reinterpret_cast<int>(this), 0, CanRemoveInput);
	pxDialog->SetHelpText("Remove an input from the collection.");
	pxDialog->SetRelativePosition(pxSubwindowDialog, true, false);
	xNewOverlay.AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	xOffset.x -= 9.0f;

	if (xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	// End of tabs
	xOffset.y = fTabsEndY + 1.0f;
	xNewOverlay.SetCurrentParentDialog(0);
	pxTabsDialog->SelectTab(0);

	// Remember which tab is open, when switching between different containers.
	static u_int s_uSelectedTab = 0;
	pxTabsDialog->SetSelectedTabPointer(&s_uSelectedTab);

	// Add a resize icon.
	pxDialog = new Asura_Resize_Dialog (pxStateChildOverlay, &s_xDimensionsState);
	xNewOverlay.AddDialog(pxDialog);
}

///////////////////////////
// PerFrameOverlayUpdate //
///////////////////////////
void Tools_Container_Collection::PerFrameOverlayUpdate()
{
	// Update the x coord of the damage override sub-window.
	if (s_xDimensionsDamage.x != s_xDimensionsState.x)
	{
		s_xDimensionsDamage.x = s_xDimensionsState.x;

		Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayID());
		if (pxOverlay)
		{
			Asura_Dialog* pxDialog = pxOverlay->GetDialog(iTOOLS_COLLECTION_DAMAGEOVERRIDE_ID);
			if (pxDialog)
			{
				pxDialog->UpdateResize(ASURA_OVERLAY_RESIZE_INITIAL, Asura_Maths::ZeroVector2);
			}
		}
	}

	// Update the x,y coords of the "inputs" tab sub-window.
	const float fExpectedY =  s_xDimensionsState.y + s_xDimensionsDamage.y + 1.0f;
	if (s_xDimensionsInput.x != s_xDimensionsState.x ||
		s_xDimensionsInput.y != fExpectedY)
	{
		s_xDimensionsInput.x = s_xDimensionsState.x;
		s_xDimensionsInput.y = fExpectedY;

		Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayID());
		Asura_Dialog* pxDialog = pxOverlay->GetDialog(iTOOLS_COLLECTION_INPUTS_ID);
		if (pxDialog)
		{
			pxDialog->UpdateResize(ASURA_OVERLAY_RESIZE_INITIAL, Asura_Maths::ZeroVector2);
		}
	}

	if (m_bChangedInputMask)
	{
		// Input mask has been changed.
		m_bChangedInputMask = false;
		SetModified(true);

		// Refresh any drop down list of states.
		RefreshStateDialogues();
	}
}

//////////////
// AddChild //
//////////////
// Add a child selected from a list dialogue.
void Tools_Container_Collection::AddChild()
{
	// List all containers excluding those that would cause circular references.
	// Allow those with group-scope that are in the same group.
	Asura_Hash_ID uContainerHash = ASURA_HASH_ID_UNSET;
	const Asura_Hash_ID uGroupHash = GetGroupHash();
	const bool bOk = Tools_Container_System::SelectContainerModalOverlay(uContainerHash, GetContainerHash(), false, false, uGroupHash);

	if (bOk)
	{
		// Add hash or increase multiplicity.
		AddChild(uContainerHash);
	}
}

//////////////
// AddChild //
//////////////
// Add a child with the specified hash. Doesn't check that the child exists.
// Nor does it check for circular references (calling code should prevent this).
void Tools_Container_Collection::AddChild(Asura_Hash_ID uChildHash)
{
	// Add hash or increase multiplicity.
	int* piMultiplicity = m_xChildren.GetDataByHash(uChildHash);
	if (piMultiplicity)
	{
		++(*piMultiplicity);
	}
	else
	{
		m_xChildren.AddNode(uChildHash, 1);
	}

	m_bNeedToCopyHashes = true;
	CopyDataToAsura();
	RefreshDialogue();
	RefreshStateChildOverlay();
	SetModified();
}

/////////////////
// RemoveChild //
/////////////////
// Remove specified child. Return true if succeeded.
bool Tools_Container_Collection::RemoveChild(Asura_Hash_ID uChildHash)
{
	if (uChildHash == ASURA_HASH_ID_UNSET) return false;

	// Remove hash or decrease multiplicity.
	int* piMultiplicity = m_xChildren.GetDataByHash(uChildHash);
	if (!piMultiplicity) return false;

	const int iRepeat = *piMultiplicity;
	if (iRepeat>1)
	{
		--(*piMultiplicity);
	}
	else
	{
		const bool bRemoved = m_xChildren.DeleteNodeByHash(uChildHash);
		Asura_Assert(bRemoved, "Failed to remove child container");
		if (bRemoved)
		{
			m_xStateMachine.RemoveChildAllInputs(uChildHash);
		}
	}

	m_bNeedToCopyHashes = true;
	CopyDataToAsura();
	RefreshDialogue();
	RefreshStateChildOverlay();
	SetModified();

	if (iRepeat == 1)
	{
		// Auto-delete orphans with group-scope.
		Tools_Container_System::DeleteIfOrphaned(uChildHash);
	}

	return true;
}

//////////////////
// ReplaceChild //
//////////////////
// Return true if the remove succeeded.
bool Tools_Container_Collection::ReplaceChild(Asura_Hash_ID uOldChildHash, Asura_Hash_ID uNewChildHash)
{
	// Update the state machine. Copying the state setup from the old to new child.
	m_xStateMachine.ReplaceChild(uOldChildHash, uNewChildHash);
	m_bNeedToCopyStates = true;

	// Add the new and remove the old child.
	return PARENT::ReplaceChild(uOldChildHash, uNewChildHash);
}

/////////////////////
// RefreshDialogue //
/////////////////////
void Tools_Container_Collection::RefreshDialogue()
{
	const u_int uNumStates = m_xStateMachine.GetNumStates();
	Asura_String64 xBuffer1("Number of states: %u", uNumStates);
	m_xNumStatesDialogue = xBuffer1.GetString();

	const u_int uNumInputs = m_xStateMachine.GetNumInputs();
	Asura_String64 xBuffer2("Number of inputs: %u", uNumInputs);
	m_xNumInputsDialogue = xBuffer2.GetString();
}

/////////////////
// CanAddInput //
/////////////////
bool Tools_Container_Collection::CanAddInput()
{
	if (s_pxThis)
	{
		return s_pxThis->GetNumInputs() < uASURA_MAX_NUM_CONTAINER_INPUTS;
	}
	return false;
}

////////////////////
// CanRemoveInput //
////////////////////
bool Tools_Container_Collection::CanRemoveInput()
{
	if (s_pxThis)
	{
		return s_pxThis->GetNumInputs() > 1;
	}
	return false;
}

//////////////
// AddInput //
//////////////
void Tools_Container_Collection::AddInput()
{
	// Confirm this operation.
	Asura_String256 xBuffer("Add a new input state to the collection?");
	if (!IsGroupScope())
	{
		xBuffer += " The new input will default to *None* for any existing entities using this collection.";
	}
	YesNo_Overlay xPrompt(xBuffer.GetString());
	if (xPrompt.DoModal() == false)
	{
		return;
	}

	if (m_xStateMachine.AddInput())
	{
		UpdateNumberOfInputs();
	}
}

/////////////////
// RemoveInput //
/////////////////
void Tools_Container_Collection::RemoveInput()
{
	// Confirm this operation.
	Asura_String256 xBuffer("Remove the input state from the collection?");
	if (!IsGroupScope())
	{
		xBuffer += " The input state will be removed from any existing entities using this collection.";
	}
	YesNo_Overlay xPrompt(xBuffer.GetString());
	if (xPrompt.DoModal() == false)
	{
		return;
	}

	if (m_xStateMachine.RemoveInput())
	{
		UpdateNumberOfInputs();
	}
}

//////////////////////////
// UpdateNumberOfInputs //
//////////////////////////
void Tools_Container_Collection::UpdateNumberOfInputs()
{
	RefreshDialogue();

	m_bNeedToCopyStates = true;
	CopyDataToAsura();

	const u_int uNumInputs = GetNumInputs();
	Tools_Mode_Container::ChangedNumberOfInputs(s_uContainerHashForEdit, uNumInputs);
	Tools_Entity_Container::ChangedNumberOfInputs(s_uContainerHashForEdit, uNumInputs);

	// Refresh the state child overlay.
	RefreshStateChildOverlay();

	// Refresh anything that contains this collection.
	SetModified();
}

/////////////////
// CreateState //
/////////////////
void Tools_Container_Collection::CreateState()
{
	Tools_ContainerStateMachine* pxStateMachine = GetStateMachine();
	Asura_Assert(pxStateMachine, "Null pointer");
	if (!pxStateMachine) return;

	GetString_Overlay xGetStringOverlay( "Enter name of new State:", 0 );

	Asura_String xUserString;

	while (true)
    {
        if ( !xGetStringOverlay.DoModal() )
        {
			// Cancelled.
            return;
        }

		xUserString = xGetStringOverlay.GetString();
		xUserString.StripWhitespace(Asura_StringUtil::STRIP_BOTH_ENDS);

		if (xUserString.IsBlank())
		{
			continue;
		}

		if ( pxStateMachine->FindState(xUserString.GetHash()) )
		{
			Message_Overlay xMessageOverlay( "Please choose a name that is not already being used." );
            xMessageOverlay.DoModal();
			continue;
		}

		break;
	}

	pxStateMachine->AddState(xUserString);

	// Select the newly created state.
	s_uStateHashForEdit = xUserString.GetHash();

	m_bNeedToCopyStates = true;
	CopyDataToAsura();
	RefreshDialogue();
	RefreshStateDialogues();
	RefreshStateChildOverlay();
	SetModified();
}

/////////////////////
// IsStateSelected //
/////////////////////
bool Tools_Container_Collection::IsStateSelected()
{
	return s_uStateHashForEdit != ASURA_HASH_ID_UNSET; 
}

/////////////////
// DeleteState //
/////////////////
void Tools_Container_Collection::DeleteState()
{
	Tools_ContainerStateMachine* pxStateMachine = GetStateMachine();
	Asura_Assert(pxStateMachine, "Null pointer");
	if (!pxStateMachine) return;

	const Asura_Hash_ID uStateHash = GetStateSelectedForEditing();
	const Tools_ContainerState* pxState = pxStateMachine->FindState(uStateHash);
	if (!pxState) return;
	
	// Confirm
	const Asura_String256 xMessage("Delete state '%s' ?", pxState->GetStateName().GetString());

	YesNo_Overlay xPrompt(xMessage.GetString());
	if (xPrompt.DoModal() == false)
	{
		return;
	}
	
	pxStateMachine->RemoveState(uStateHash);

	m_bNeedToCopyStates = true;
	CopyDataToAsura();
	RefreshDialogue();
	RefreshStateDialogues();
	if (uStateHash == s_uStateHashForEdit)
	{
		RefreshStateChildOverlay();
	}
	SetModified();
}

/////////////////
// RenameState //
/////////////////
// Rename the container state.
// Note: that this can cause lots of brokenness in the entities if the state is already being used.
// However, the brokenness will be automatically fixed the next time that the entity file is saved from object mode.
void Tools_Container_Collection::RenameState()
{
	Tools_ContainerStateMachine* pxStateMachine = GetStateMachine();
	Asura_Assert(pxStateMachine, "Null pointer");
	if (!pxStateMachine) return;

	const Asura_Hash_ID uStateHash = GetStateSelectedForEditing();
	Tools_ContainerState* pxState = pxStateMachine->FindState(uStateHash);
	if (!pxState) return;
	
	if (pxState->IsInvariant())
	{
		// Disallow renaming of 'on' and 'off'.
		Message_Overlay xMessageOverlay( "The 'on' and 'off' states are defined in code and can't be renamed." );
        xMessageOverlay.DoModal();
		return;
	}

	// Prompt for new name
	const Asura_String& xOldString = pxState->GetStateName();
	Asura_String xUserString = xOldString;
	const Asura_String256 xMessage("Enter new name for state '%s':", xUserString.GetString());

	GetString_Overlay xGetStringOverlay( xMessage.GetString(), xUserString.GetString() );

	while (true)
    {
        if ( !xGetStringOverlay.DoModal() )
        {
			// Cancelled.
            return;
        }

		// Strip whitespace from ends.
		xUserString = xGetStringOverlay.GetString();
		xUserString.StripWhitespace(Asura_StringUtil::STRIP_BOTH_ENDS);

		if (xUserString.IsBlank())
		{
			// Blank string entered, reprompt with old string.
			xUserString = xOldString;
			continue;
		}

		const Asura_Hash_ID uNewHash = xUserString.GetHash();
		if (uStateHash == uNewHash)
		{
			// Hash hasn't changed. Check for change of case.
			if (xUserString.Compare(xOldString) == 0 &&
				xUserString.Compare(xOldString, false) != 0)
			{
				// Only the case has been changed, we can just modify the name without
				// it affecting the hashes.
				pxState->SetInitialStateName(xUserString);

				// Update any relevant dialogues with the new name.
				RefreshStateDialogues();
				RefreshStateChildOverlay();
				// Mark as modified so that file is saved.
				SetModified();
			}
			return;
		}

		if (uNewHash == Asura_ContainerState::GetOnStateHash() ||
			uNewHash == Asura_ContainerState::GetOffStateHash())
		{
			Message_Overlay xMessageOverlay( "The 'on' and 'off' states are reserved for use by the code." );
            xMessageOverlay.DoModal();
			// Reprompt with old string.
			xUserString = xOldString;
			continue;
		}

		if ( pxStateMachine->FindState(uNewHash) )
		{
			// Name already used, reprompt.
			Message_Overlay xMessageOverlay( "Please choose a name that is not already being used." );
            xMessageOverlay.DoModal();
			continue;
		}

		break;
	}

	// Note: unfortunately there is a lot of code assuming that the name and hash match up.
	// So we can't change the name without changing the hash too. It isn't possible to fix up
	// the entities, not least because we'd need to have *all* of the entity files checked out.
	
	// MGH 5/11/2008: the container entities and static messages are now fixed up automatically
	// the next time that they are saved out from Object mode.

	// Warn user that they may be about to mess things up in a big way.
	const Asura_String256 xWarning( "Rename state from '%s' to '%s' ?\n\nAny entity files that are using this state will need to be saved out again from Object mode and the levels re-exported.", xOldString.GetString(), xUserString.GetString() );

	YesNo_Overlay xPrompt(xWarning.GetString());
	if (xPrompt.DoModal() == false)
	{
		return;
	}

	// The most we can do is fix up the state machines so that they are consistent and
	// then refresh the dialogues, rebuild the instance and mark the file(s) as modified.
	Tools_Mode_Container::MarkDeletionStart();

	// Change the name and hash.
	pxState->SaveOldStateHash(pxState->GetStateHash());
	pxState->SetInitialStateName(xUserString);

	// Fix up all state machines that refer to this state.
	const Asura_Hash_ID uNewStateHash = xUserString.GetHash();
	Tools_Container_System::ReplaceState(GetContainerHash(), uStateHash, uNewStateHash);

	// Fix up the damage overrides.
	m_xStateMachine.ReplaceDamageOverrideState(uStateHash, uNewStateHash);

	// Copy data to Asura.
	m_bNeedToCopyStates = true;
	CopyDataToAsura();

	// Reselect the renamed state.
	if (s_uStateHashForEdit == uStateHash)
	{
		s_uStateHashForEdit = uNewStateHash;
	}

	// Update any relevant dialogues with the new name.
	RefreshStateDialogues();

	// Rebuild the edit overlay to make everything up-to-date.
	RefreshStateChildOverlay();
				
	// Mark as modified so that file is saved.
	SetModified();

	// If we have just renamed part of the previewed state then modify it.
	bool bChanged = false;
	Asura_Container_InputState xInputState = Tools_Mode_Container::GetPreviewState();
	for (u_int u = 0; u < xInputState.GetNumInputs(); ++u)
	{
		if (xInputState.GetStateHash(u) == uStateHash)
		{
			xInputState.SetStateHash(u, uNewStateHash);
			bChanged = true;
		}
	}

	if (bChanged)
	{
		Tools_Mode_Container::RequestPreviewState(xInputState, GetContainerHash());
	}

	Tools_Mode_Container::MarkDeletionStop();
}

/////////////////////////////////////
// ChangedPreviewStateFromDialogue //
/////////////////////////////////////
// Refresh the edit overlay if needed.
// The "Preview state" drop down has been modified on the instance overlay.
// If the root collection has a state selected on its edit overlay, then change to the new preview state.
void Tools_Container_Collection::ChangedPreviewStateFromDialogue(Asura_Hash_ID uContainerToEdit,
																 const Asura_Container_InputState& xInputState,
																 const Asura_Container_InputState& xInputStateOld)
{
	if (uContainerToEdit != s_uContainerHashForEdit)
	{
		return;
	}

	// Figure out which input slot was changed.
	u_int uInputIndex = s_uStateIndexForEdit;
	const u_int uNumInputs = Asura_Maths::Min(xInputState.GetNumInputs(), xInputStateOld.GetNumInputs());
	for (u_int u=0; u<uNumInputs; ++u)
	{
		if (xInputState.GetStateHash(u) != xInputStateOld.GetStateHash(u))
		{
			uInputIndex = u;
			break;
		}
	}

	const Asura_Hash_ID uStateHashToPreview = xInputState.GetStateHash(uInputIndex);

	if (s_uStateHashForEdit  != uStateHashToPreview ||
		s_uStateIndexForEdit != uInputIndex)
	{
		// Select the new preview state in the overlay.
		s_uStateHashForEdit  = uStateHashToPreview;
		s_uStateIndexForEdit = uInputIndex;

		// If the collection's edit overlay is open then rebuild it.
		Tools_Container_Overlay_Manager& xManager = Tools_Container_System::GetOverlayManager();

		if (xManager.IsEditOverlayOpen() &&
			xManager.GetEditingContainerHash() == uContainerToEdit)
		{
			xManager.RebuildEditOverlay();
		}
	}
}

//////////////////////////////////////////
// ChangedPreviewStateFromStateDialogue //
//////////////////////////////////////////
// Preview state was changed on the collection overlay's state editing dialogue.
void Tools_Container_Collection::ChangedPreviewStateFromStateDialogue(Asura_Hash_ID uContainerToEdit,
																	  Asura_Hash_ID uStateHashToPreview,
																	  u_int uInputIndex)
{
	// Rebuild the edit overlay for the newly selected state.
	if (uContainerToEdit == s_uContainerHashForEdit)
	{
		if (s_uStateHashForEdit != uStateHashToPreview ||
			s_uStateIndexForEdit != uInputIndex)
		{
			// Select the new preview state in the overlay.
			s_uStateHashForEdit = uStateHashToPreview;
			s_uStateIndexForEdit = uInputIndex;

			// If the collection's edit overlay is open then rebuild it.
			Tools_Container_Overlay_Manager& xManager = Tools_Container_System::GetOverlayManager();

			if (xManager.IsEditOverlayOpen() &&
				xManager.GetEditingContainerHash() == uContainerToEdit)
			{
				xManager.BringToFrontNextUpdate();
				xManager.RebuildEditOverlay();
			}
		}
	}

	if (s_uStateHashForEdit != ASURA_HASH_ID_UNSET)
	{
		// Change the preview state on the instance overlay if necessary.
		Asura_Container_InputState xInputState = Tools_Mode_Container::GetPreviewState();
		if (uStateHashToPreview == Asura_ContainerState::GetOnStateHash())
		{
			xInputState.SetOff();
		}
		xInputState.SetStateHash(s_uStateIndexForEdit, uStateHashToPreview);
		Tools_Mode_Container::RequestPreviewState(xInputState, uContainerToEdit);
	}
}

////////////
// Helper //
////////////
extern int SortTreeAlphabetically( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo );

///////////////////////////
// RefreshStateDialogues //
///////////////////////////
// Make sure that any other dialogues that have state lists in are refreshed,
// as states may have been added, renamed or deleted.
void Tools_Container_Collection::RefreshStateDialogues()
{
	if (s_uContainerHashForEdit == GetContainerHash())
	{
		// The instance dialogue, entity overlay or message editing dialogues may be open. Refresh them.
		Tools_ContainerState_DropDownList_Dialog::RefreshAllDialogues();
	}
}

///////////////////////
// FillChildIterator //
///////////////////////
// Set up the child iterator.
void Tools_Container_Collection::FillChildIterator(Tools_Container_ChildIterator& xIterator) const
{
	const int iNumNodes = m_xChildren.GetNumberOfNodes();
	if (iNumNodes > 0)
	{
		xIterator.SetArraySize(iNumNodes);
		m_xChildren.Traverse(AddEntry_Callback, &xIterator);
	}
}

///////////////////////
// AddEntry_Callback //
///////////////////////
bool Tools_Container_Collection::AddEntry_Callback(Asura_Hash_ID uHash, int const* piData, void* pUserData)
{
	Tools_Container_ChildIterator* pxIterator = static_cast<Tools_Container_ChildIterator*>(pUserData);
	pxIterator->AddEntry(uHash, *piData);
	return true;
}

////////////////////////
// PostLoadInitialise //
////////////////////////
void Tools_Container_Collection::PostLoadInitialise()
{
	// Need to copy hashes so that marker attachments will be sorted to the end of the array.
	m_bNeedToCopyHashes = true;
	CopyDataToAsura();
}

//////////////////////////
// PostCreateInitialise //
//////////////////////////
// Add default "on"/"off" states to all newly created collections.
// The user can delete them from the displayed list of states if they wish,
// although they still exist and can be used by the code.
void Tools_Container_Collection::PostCreateInitialise()
{
	if (m_xStateMachine.AddDefaultStates())
	{
		m_bNeedToCopyStates = true;
	}
}

/////////////////////
// CopyDataToAsura //
/////////////////////
void Tools_Container_Collection::CopyDataToAsura()
{
	Asura_Container_Collection* pxContainerCollection = static_cast<Asura_Container_Collection*>(m_pxAsuraContainer);
	
	// Copy child hashes if they have changed.
	if (pxContainerCollection && m_bNeedToCopyHashes)
	{
		// Set up and copy array of hashes.
		Tools_Container_ChildIterator xIt(this);
		const u_int uNumChildren = static_cast<u_int>(xIt.GetNumChildren());

		pxContainerCollection->AllocateSize(uNumChildren);

		u_int uEnd   = uNumChildren;
		u_int uStart = 0;

		// Note: so that meshes are updated before their marker attachments are rendered,
		// we need to ensure that the marker attachments are at the end of the array.
		// The ordering is also vital, as without it there would be a crash on unlinking attachments.
		// Also we want physics attachments to be initialised after the physics objects have been initialised.

		// So the order is:
		// 1) Containers with no attachment or a position offset attachment. (May contain shared physics objects.)
		// 2) Containers with physics attachments. (Attach to any shared physics object in group 1.)
		// 3) Containers with marker attachments. (Look for markers in groups 1, 2. Must come last to avoid crash on unlinking.)

		int iNumPhysics = 0;

		for ( ; !xIt.Done(); xIt.Next() )
		{
			const Asura_Hash_ID uHash = xIt.GetCurrentHash();

			const Tools_Container* pxChild = Tools_Container_System::FindContainerByHash(uHash);
			if (pxChild && (pxChild->IsMarkerAttachment() || pxChild->IsPhysicsAttachment()))
			{
				if (pxChild->IsMarkerAttachment())
				{
					// Add marker attachments at end of array.
					pxContainerCollection->SetChild(uHash, --uEnd);
				}
				else
				{
					++iNumPhysics;
				}
			}
			else
			{
				// Add other stuff at start of array.
				pxContainerCollection->SetChild(uHash, uStart++);
			}
		}

		if (iNumPhysics > 0)
		{
			for (xIt.Begin() ; !xIt.Done(); xIt.Next() )
			{
				const Asura_Hash_ID uHash = xIt.GetCurrentHash();

				const Tools_Container* pxChild = Tools_Container_System::FindContainerByHash(uHash);
				if (pxChild && pxChild->IsPhysicsAttachment())
				{
					// Add physics attachments to the middle of the array.
					pxContainerCollection->SetChild(uHash, --uEnd);
				}
			}
		}
		Asura_Assert(uStart == uEnd, "Failed to copy all of the children to the container.");

		m_bNeedToCopyHashes = false;	
	}

	// Copy states if they have changed.
	if (pxContainerCollection && m_bNeedToCopyStates)
	{
		m_xStateMachine.CopyDataToAsura(pxContainerCollection);
		m_bNeedToCopyStates = false;
	}

	PARENT::CopyDataToAsura();
}

//////////////////////////////
// AddDamageOverrideOverlay //
//////////////////////////////
// Overlay to display and modify the damage overrides.
Asura_Subwindow_Dialog* Tools_Container_Collection::AddDamageOverrideOverlay(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset)
{
	// Create the subwindow control for containing all the other controls.
	Asura_Subwindow_Dialog* pxSubwindowDialog = new Asura_Subwindow_Dialog(xOffset, s_xDimensionsDamage.x, s_xDimensionsDamage.y, false);
	pxSubwindowDialog->SetControlID(iTOOLS_COLLECTION_DAMAGEOVERRIDE_ID);
	pxSubwindowDialog->SetAllowResize(true, &s_xDimensionsDamage);
	xNewOverlay.AddDialog(pxSubwindowDialog);
	
	Asura_Vector_2 xSubOffset(1.0f, pxSubwindowDialog->GetChildMaxY() + 0.5f);
	
	const float fSubOffsetX = xSubOffset.x;
	const float fColumnX = 16.0f;
	const float fColumnOverride1X = 2.0f;
	const float fColumnOverride2X = 4.0f;

	const Tools_ContainerState* const pxState = m_xStateMachine.FindState(s_uStateHashForEdit);

	if (pxState && !pxState->IsInvariant())
	{
		xSubOffset.x = fSubOffsetX;
		Asura_Dialog* pxDialog = new Asura_StringDisplay_Dialog(pxState->GetStateName().GetString(), xSubOffset);
		pxSubwindowDialog->AddDialog(pxDialog);

		
		xSubOffset.x = Asura_Maths::Max(pxDialog->GetMaxCoords().x + 2.0f, fColumnX);
		pxDialog = new Asura_Function_Dialog("Add Damage Override", xSubOffset, AddDamageOverride, reinterpret_cast<int>(this), 0, 0);
		pxDialog->SetHelpText("Add a damage override. This will replace the input state of the collection with an override state according to the health percentage.");
		pxSubwindowDialog->AddDialog(pxDialog);
	

		xSubOffset.y += Asura_Dialog::DefaultSpacing();

		const u_int uNumOverrides = pxState->GetNumDamageOverrides();
		u_int uMinHealth = 0;

		for (u_int v=0; v<uNumOverrides; ++v)
		{
			const Asura_ContainerState_DamageOverride* pxOverride = pxState->GetDamageOverride(v);
			if (pxOverride)
			{				
				// Add a remove button.
				Asura_Callback_UserData<void, Tools_RemoveDamageOverride_Callback_Data> xRemove(
										Tools_RemoveDamageOverride_Callback_Data::Callback,
										Tools_RemoveDamageOverride_Callback_Data(this, v) );

				xSubOffset.x = fColumnOverride1X;
				Asura_Icon_Dialog* pxIcon = new Asura_Icon_Dialog(xSubOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_DELETE);
				pxIcon->SetClickedFunction(xRemove);
				pxIcon->SetHelpText("Remove this damage override");
				pxSubwindowDialog->AddDialog(pxIcon);

				// List the health range and state name. (e.g. "0% ...  50%  light-on")
				const Tools_ContainerState* const pxNewState = m_xStateMachine.FindState(pxOverride->m_uOverrideStateHash);
				const char* const szNewStateName = (pxNewState ? pxNewState->GetStateName().GetString() : "*Unknown*");
				const u_int uMaxHealth = static_cast<u_int>(pxOverride->m_fMaxHealthPercent);

				xSubOffset.x = fColumnOverride2X;
				Asura_String256 xBuffer("%2u%% .. %3u%%  %s", uMinHealth, uMaxHealth, szNewStateName);
				Asura_Dialog* pxDialog = new Asura_StringDisplay_Dialog(xBuffer.GetString(), xSubOffset);
				pxSubwindowDialog->AddDialog(pxDialog);
				xSubOffset.y += 1.5f*Asura_Dialog::SmallSpacing();
				uMinHealth = uMaxHealth;			
			}
		}

		// Add final entry for up to 100% health.
		if (uMinHealth != 100)
		{
			xSubOffset.x = fColumnOverride2X;
			Asura_String256 xBuffer("%2u%% .. %3u%%  %s", uMinHealth, 100, pxState->GetStateName().GetString());
			Asura_Dialog* pxDialog = new Asura_StringDisplay_Dialog(xBuffer.GetString(), xSubOffset);
			pxSubwindowDialog->AddDialog(pxDialog);
			xSubOffset.y += Asura_Dialog::LargeSpacing();
		}
	}

	xOffset.y = pxSubwindowDialog->m_xMaxCoord.y + 1.0f;
	return pxSubwindowDialog;
}

///////////////////////
// AddDamageOverride //
///////////////////////
void Tools_Container_Collection::AddDamageOverride()
{
	Tools_ContainerState* pxState = m_xStateMachine.FindState(s_uStateHashForEdit);
	if (!pxState) return;

	// Limit to 99%, as 100% should always be the original state.
	GetIntNumber_Overlay xOverlay("Health percentage threshold", 50, 0, 99);
	if (!xOverlay.DoModal())
	{
		return;
	}

	int iValue = xOverlay.GetNumber();

	ListSelection_Overlay xList("Override state below health threshold");

	const u_int uNumStates = m_xStateMachine.GetNumStates();
	for (u_int u=0; u<uNumStates; ++u)
	{
		const Tools_ContainerState* const pxTempState = m_xStateMachine.GetState(u);
		if (pxTempState)
		{
			xList.AddItem(pxTempState->GetStateName().GetString() , static_cast<int>(pxTempState->GetStateHash()));
		}
	}

	if (!xList.DoModal())
	{
		return;
	}

	const Asura_Hash_ID uStateHash = static_cast<Asura_Hash_ID>(xList.GetSelectionData());
	
	pxState->AddDamageOverride(static_cast<float>(iValue), uStateHash);

	// Mark as modified, copy data and update the overlay.
	m_bNeedToCopyStates = true;
	CopyDataToAsura();
	RefreshStateChildOverlay();
	SetModified();
}

//////////////////////////
// RemoveDamageOverride //
//////////////////////////
void Tools_Container_Collection::RemoveDamageOverride(Tools_RemoveDamageOverride_Callback_Data xData)
{
	Tools_ContainerState* pxState = m_xStateMachine.FindState(s_uStateHashForEdit);
	if (!pxState) return;

	pxState->RemoveDamageOverride(xData.m_uOverrideIndex);
	
	// Mark as modified, copy data and update the overlay.
	m_bNeedToCopyStates = true;
	CopyDataToAsura();
	RefreshStateChildOverlay();
	SetModified();
}

//////////////////////////
// AddStateChildOverlay //
//////////////////////////
Asura_Subwindow_Dialog* Tools_Container_Collection::AddStateChildOverlay(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset)
{
	// Create the subwindow control for containing all the other controls.
	Asura_Subwindow_Dialog* pxSubwindowDialog = new Asura_Subwindow_Dialog(xOffset, s_xDimensionsState.x, s_xDimensionsState.y, false);
	xNewOverlay.AddDialog(pxSubwindowDialog);
	
	Asura_Vector_2 xSubOffset(1.0f, pxSubwindowDialog->GetChildMaxY() + 0.5f);

	// Set up heading information.
	Asura_Dialog* pxDialog = 0;

	Asura_String256 xBuffer("State: ");
	const Tools_ContainerState* pxState = m_xStateMachine.FindState(s_uStateHashForEdit);
	const bool bStateOK  = (pxState && !CheckIsTooDeep());
	const bool bStateOn  = (s_uStateHashForEdit == Asura_ContainerState::GetOnStateHash());
	const bool bStateOff = (s_uStateHashForEdit == Asura_ContainerState::GetOffStateHash());

	if (bStateOK || bStateOn || bStateOff)
	{
		if (bStateOK)
		{
			// State found.
			xBuffer += pxState->GetStateName().GetString();
		}
		else if (bStateOn)
		{
			// "On" state, which has been deleted.
			xBuffer += Tools_ContainerState::GetOnStateName();
			xBuffer += " (deleted)";
		}
		else if (bStateOff)
		{
			// "Off" state, which has been deleted.
			xBuffer += Tools_ContainerState::GetOffStateName();
			xBuffer += " (deleted)";
		}

		pxDialog = new Tools_ContainerStateEdit_Dialog(pxSubwindowDialog, xSubOffset, this, xBuffer.GetString());
	}
	else
	{
		// State not found. Give up.
		xBuffer += "None";
		pxDialog = new Asura_StringDisplay_Dialog(xBuffer.GetString() , xSubOffset);
	}

	pxSubwindowDialog->AddDialog(pxDialog);

	xOffset.y = pxSubwindowDialog->m_xMaxCoord.y + 1.0f;
	return pxSubwindowDialog;
}

//////////////////////////////
// RefreshStateChildOverlay //
//////////////////////////////
void Tools_Container_Collection::RefreshStateChildOverlay()
{
	// Is the the edit overlay currently on top? If so bring it to the front on rebuilding.
	const bool bBringToFront = (Asura_Overlay_System::GetTopOverlayID() == reinterpret_cast<u_int>(this));

	// Rebuild the overlay.
	if (Tools_Container_System::IsEditOverlayOpen())
	{
		Tools_Container_Overlay_Manager& xManager = Tools_Container_System::GetOverlayManager();
		xManager.RebuildEditOverlay();
		if (bBringToFront)
		{
			xManager.BringToFrontNextUpdate();
		}
	}

	// Change the preview state.
	if (s_uStateHashForEdit != ASURA_HASH_ID_UNSET)
	{
		Asura_Container_InputState xInputState = Tools_Mode_Container::GetPreviewState();
		xInputState.SetStateHash(s_uStateIndexForEdit, s_uStateHashForEdit);
		Tools_Mode_Container::RequestPreviewState(xInputState, GetContainerHash());
	}
}

/////////////////////
// AddInputOverlay //
/////////////////////
Asura_Subwindow_Dialog* Tools_Container_Collection::AddInputOverlay(Asura_Overlay& xNewOverlay, Asura_Vector_2& xOffset)
{
	Asura_Subwindow_Dialog* pxSubwindowDialog = new Asura_Subwindow_Dialog(xOffset, s_xDimensionsInput.x, s_xDimensionsInput.y, false);
	xNewOverlay.AddDialog(pxSubwindowDialog);
	pxSubwindowDialog->SetControlID(iTOOLS_COLLECTION_INPUTS_ID);
	pxSubwindowDialog->SetAllowResize(true, &s_xDimensionsInput);

	// Add states to the overlay.
	Asura_Vector_2 xSubOffset(1.0f, pxSubwindowDialog->GetChildMaxY() + 0.5f);

	// Heading
	Asura_Dialog* pxDialog = new Asura_Dialog("Set the states enabled for each input:", xSubOffset);
	pxSubwindowDialog->AddDialog(pxDialog);
	xSubOffset.y += Asura_Dialog::LargeSpacing();

	// Display the input numbers "0", "1" etc.
	const float fIndentY = 4.0f;
	const u_int uNumStates = m_xStateMachine.GetNumStates();
	const u_int uNumInputs = Asura_Maths::Min(m_xStateMachine.GetNumInputs(), 32u);

	for (u_int v=0; v<uNumInputs; ++v)
	{
		xSubOffset.x = 1.0f + v * fIndentY;
		Asura_String64 xBuffer("%3u", v);
		Asura_Dialog* pxDialog = new Asura_Dialog(xBuffer.GetString(), xSubOffset);
		pxSubwindowDialog->AddDialog(pxDialog);
	}
	xSubOffset.x = 1.0f;
	xSubOffset.y += Asura_Dialog::LargeSpacing();

	// Sort the states in to alphabetical order.
	const Tools_ContainerSortedStateList xSortedList(m_xStateMachine);

	// Loop over the states.
	for (u_int u=0; u<uNumStates; ++u)
	{
		Tools_ContainerState* const pxState = xSortedList.GetState(u);
		Asura_Assert(pxState, "Null state");
		if (!pxState) continue;

		u_int* const puMask = pxState->GetInputMask();

		// Display a tickbox for each input.
		for (u_int v=0; v<uNumInputs; ++v)
		{
			xSubOffset.x = 1.0f + v * fIndentY;
			Asura_Dialog* pxDialog = new Asura_FlagTickbox_Dialog("", xSubOffset, puMask , 1<<v, &m_bChangedInputMask);
			pxSubwindowDialog->AddDialog(pxDialog);
		}

		// Display the state name.
		xSubOffset.x = 1.0f + uNumInputs * fIndentY;
		Asura_Dialog* pxDialog = new Asura_Dialog(pxState->GetStateName().GetString(), xSubOffset);
		pxSubwindowDialog->AddDialog(pxDialog);
		xSubOffset.y += Asura_Dialog::DefaultSpacing();
	}

	xOffset.y = pxSubwindowDialog->GetMaxCoords().y + 1.0f;

	return pxSubwindowDialog;
}

////////////
// Verify //
////////////
void Tools_Container_Collection::Verify(bool bRecurse) const
{
	PARENT::Verify(bRecurse);

	Tools_Container_ChildIterator xIt(this);

	// Check that the state machine is valid.
	m_xStateMachine.Verify(this, xIt);

	// Check that relevant marker points actually exist.
	for ( xIt.Begin(); !xIt.Done(); xIt.NextSkipIfSame() )
	{
		const Tools_Container* const pxContainer =  xIt.GetCurrent();
		
		if (pxContainer &&
			pxContainer->IsMarkerAttachment() &&
			pxContainer->GetClassification() == AsuraContainerClass_Attachment)
		{
			const Tools_Container_Attachment* const pxAttachment = static_cast<const Tools_Container_Attachment*>(pxContainer);
			const Asura_Hash_ID uMarkerHash = pxAttachment->GetMarkerHash();
			const Asura_String& xMarkerName = pxAttachment->GetMarkerName();
			const Asura_String& xAttachmentName = pxAttachment->GetContainerName();
			
			Asura_String xStateName("unknown");

			if (uMarkerHash == ASURA_HASH_ID_UNSET) continue;

			// Search for the mesh with the relevant marker point.
			const bool bFound = CheckMarkerExistsForAttachment(uMarkerHash, pxAttachment->GetContainerHash(), xStateName, pxContainer->GetNumInputs());

			if (!bFound)
			{
				AddLogMessage("mesh with marker '%s' not found in collection; used by attachment '%s' for state '%s'.", xMarkerName.GetString(), xAttachmentName.GetString(), xStateName.GetString());
			}
		}
	}

	// Check that there is only a single health container in this collection.
	u_int uNumHealth = 0;

	for ( xIt.Begin(); !xIt.Done(); xIt.Next() )
	{
		const Tools_Container* const pxContainer =  xIt.GetCurrent();
		if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Health)
		{
			++uNumHealth;
		}
	}

	if (uNumHealth > 1)
	{
		AddLogMessage("has more than one health container");
	}

	// Check that there is at most a single shared physics object active.
	VerifySharedPhysics();
}

////////////////////////////////////
// CheckMarkerExistsForAttachment //
////////////////////////////////////
bool Tools_Container_Collection::CheckMarkerExistsForAttachment(Asura_Hash_ID uMarkerHash,
																Asura_Hash_ID uAttachmentHash,
																Asura_String& xStateName,
																u_int uNumInputs) const
{
	// Loop over the states checking that each one is valid.
	const u_int uNumStates = m_xStateMachine.GetNumStates();
	for (u_int u=0; u<uNumStates; ++u)
	{
		const Tools_ContainerState* const pxState = m_xStateMachine.GetState(u);
		if (pxState)
		{
			// Check whether any of the inputs are not in the "off" state.
			bool bOff = true;
			for (u_int v=0; v<uNumInputs; ++v)
			{
				const Asura_Hash_ID uChildStateHash = pxState->GetChildState(v, uAttachmentHash);
				if (!Asura_ContainerState::IsOff(uChildStateHash))
				{
					bOff = false;
					break;
				}
			}
			
			if (!bOff)
			{
				// Attachment is enabled for this state.
				const bool bFound = CheckMarkerExistsForState(uMarkerHash, pxState);
				if (!bFound)
				{
					// Return the state name for which we failed to find an attachment.
					xStateName = pxState->GetStateName();
					return false; // Not found.
				}
			}
		}
	}

	// Ok. Found a marker for all states for which the attachment is enabled.
	return true;
}

///////////////////////////////
// CheckMarkerExistsForState //
///////////////////////////////
// Return true, if we found a mesh with the specified marker that is enabled for the given state.
bool Tools_Container_Collection::CheckMarkerExistsForState(Asura_Hash_ID uMarkerHash, const Tools_ContainerState* pxState) const
{
	if (!pxState) return false;

	// Search for the mesh with the relevant marker point.
	for ( Tools_Container_ChildIterator xIt(this) ; !xIt.Done() ; xIt.NextSkipIfSame() )
	{
		const Tools_Container* pxContainer =  xIt.GetCurrent();
		if (!pxContainer) continue;

		const Asura_Hash_ID uChildContainerHash = pxContainer->GetContainerHash();

		// Follow non-marker attachments.
		if (!pxContainer->IsMarkerAttachment() && pxContainer->GetClassification() == AsuraContainerClass_Attachment)
		{
			const Tools_Container_Attachment* pxAttachment = static_cast<const Tools_Container_Attachment*>(pxContainer);
			pxContainer = Tools_Container_System::FindContainerByHash(pxAttachment->GetChildContainerHash());
		}

		// Only meshes can have marker points.
		if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Mesh)
		{
			const Asura_Hash_ID uChildStateHash = pxState->GetChildState(0, uChildContainerHash);
			if (!Asura_ContainerState::IsOff(uChildStateHash))
			{
				// Mesh is enabled for this state.
				const Tools_Container_Mesh* const pxMesh = static_cast<const Tools_Container_Mesh*>(pxContainer);
				if (pxMesh->HasMarker(uMarkerHash))
				{
					// Mesh has the relevant marker.
					return true;
				}
			}
		}
	}

	return false;
}

/////////////////////////
// VerifySharedPhysics //
/////////////////////////
// Check that there is at most a single shared physics object enabled for each state.
// If there is an attachment with attach-to-physics ticked then we need exactly one shared physics object.
void Tools_Container_Collection::VerifySharedPhysics() const
{
	const u_int uNumStates = m_xStateMachine.GetNumStates();
	for (u_int u=0; u<uNumStates; ++u)
	{
		const Tools_ContainerState* const pxState = m_xStateMachine.GetState(u);
		if (!pxState) continue;
		const Asura_Hash_ID uStateHash = pxState->GetStateHash();

		if (uStateHash == Asura_ContainerState::GetOnStateHash())
		{
			// "On" state often breaks the limit. Don't bother checking.
			continue;
		}

		if (uStateHash == Asura_ContainerState::GetOffStateHash())
		{
			// "Off" state can never cause errors. Don't bother checking.
			continue;
		}

		const bool bOk = VerifyNumberOfSharedPhysicsObjects(pxState);
		if (!bOk)
		{
			// Stop after first error is found.
			break;
		}
	}
}

////////////////////////////////////////
// VerifyNumberOfSharedPhysicsObjects //
////////////////////////////////////////
// Check that there is at most a single shared physics object enabled for this state.
// Return true if OK; false if there is an error.
bool Tools_Container_Collection::VerifyNumberOfSharedPhysicsObjects(const Tools_ContainerState* pxState) const
{
	if (!pxState) return false;

	u_int uNumSharedPhysics = 0;
	u_int uNumAttachPhysics = 0;

	// Search for meshes with shared physics.
	for ( Tools_Container_ChildIterator xIt(this) ; !xIt.Done() ; xIt.NextSkipIfSame() )
	{
		const Tools_Container* pxContainer =  xIt.GetCurrent();
		if (!pxContainer) continue;

		const Asura_Hash_ID uChildContainerHash = pxContainer->GetContainerHash();
		const u_int uNumInputs = pxContainer->GetNumInputs();

		bool bOff = true;
		for (u_int v=0; v<uNumInputs; ++v)
		{
			const Asura_Hash_ID uChildStateHash = pxState->GetChildState(v, uChildContainerHash);
			if (!Asura_ContainerState::IsOff(uChildStateHash))
			{
				bOff = false;
				break;
			}
		}

		if (bOff)
		{
			// Disabled for this state so we can ignore it.
			continue;
		}

		if (pxContainer->GetClassification() == AsuraContainerClass_Attachment)
		{
			if (pxContainer->IsPhysicsAttachment())
			{
				++uNumAttachPhysics;	
			}
			else if (!pxContainer->IsMarkerAttachment())
			{
				// Follow non-marker, non-physics attachments.
				const Tools_Container_Attachment* pxAttachment = static_cast<const Tools_Container_Attachment*>(pxContainer);
				pxContainer = Tools_Container_System::FindContainerByHash(pxAttachment->GetChildContainerHash());
			}
		}

		// Only meshes can have physics objects.
		if (pxContainer && pxContainer->GetClassification() == AsuraContainerClass_Mesh)
		{
			// Mesh is enabled for this state.
			const Tools_Container_Mesh* const pxMesh = static_cast<const Tools_Container_Mesh*>(pxContainer);
			if (pxMesh->HasSharedDynamicPhysicsObject())
			{
				// Mesh has shared dynamic physics object.
				++uNumSharedPhysics;
			}
		}
	}

	// A maximum of one shared physics object can be enabled for any state.
	if (uNumSharedPhysics > 1)
	{
		AddLogMessage("too many shared dynamic physics objects are enabled for state '%s'.", pxState->GetStateName().GetString());
		return false;
	}

	if (uNumAttachPhysics && uNumSharedPhysics==0)
	{
		AddLogMessage("Tried to attach to a shared physics object, but none are enabled for state '%s'.", pxState->GetStateName().GetString());
		return false;
	}

	return true;
}


/////////////////
// FixUpErrors //
/////////////////
// Fix up certain types of error, automatically.
void Tools_Container_Collection::FixUpErrors()
{
	PARENT::FixUpErrors();

	m_xStateMachine.FixUpErrors(this);
}


#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM
