///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_Mode_Container.h
//
// Created by Martin G.Harris (5/6/2008)
// 
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////
// 
// Mode for editing containers and previewing container instances.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_MODE_CONTAINER_H__
#define __TOOLS_MODE_CONTAINER_H__

#ifdef ASURA_USE_CONTAINER_SYSTEM

// Includes
#include "Core/Asura_List.h"
#include "Core/Asura_Maths.h"
#include "Entity/Container/Asura_ContainerState.h"
#include "Interface/Asura_Tree_Dialog.h"
#include "Tools_Mode.h"
#include "Objects/Tools_PositionEdit_Object.h"


// Forward declarations
class Asura_ClientContainerInstance;
class Asura_Overlay;
class Tools_Container;
class Tools_Container_Attachment;
class Tools_Container_Object;

//////////////////////////////////
// Tools_Container_InstanceTree //
//////////////////////////////////
// Dialogue tree, with optional inlined highlightable attachment icons.
// Stores information about the selected entry and collapsed items, so that
// they can be restored upon rebuilding the tree.
class Tools_Container_InstanceTree
{
	// Remember which collections were collapsed when rebuilding the tree.
	enum { MAX_NUM_COLLAPSED = 16 };

public:
	Tools_Container_InstanceTree();

	void SetSelectedAttachment(Tools_Container_Object* pxHighlightedAttachment);
	void SetSelectedItem(Asura_TreeDialog_Item* pxItem);
	void SetSelectedContainer(Asura_Hash_ID uContainerHash);

	void InitialiseSelection();
	void ClearSelection();

	void DeleteAll();

	void AddItem(const char* szLabel, Asura_Hash_ID uContainerHash);

	Asura_TreeDialog_Item* InsertItem(int ( *pfnCompFunc )( const Asura_TreeDialog_Item& xOne, const Asura_TreeDialog_Item& xTwo ),
										Asura_TreeDialog_Item* pxParent,
										const char* szLabel,
										Asura_Hash_ID uContainerHash);

	Asura_Hash_ID GetSelectedContainerHash();
	Tools_Container_Object* GetSelectedToolsObject();
	Tools_Container_Object* GetParentOfSelectedToolsObject();

	Asura_TreeDialog_Item* GetRoot() { return m_xTree.GetRoot(); }
	Asura_TreeDialog_Tree& GetTree() { return m_xTree; }

	void TraverseUpdateStateInfo(Asura_TreeDialog_Item* pxItem);

protected:
	// Reset, store, restore the collapsed status of the tree.
	void ResetCollapsed();
	void StoreCollapsed();
	void RestoreCollapsed();
	void TraverseStoreCollapsedInfo(Asura_TreeDialog_Item* pxItem);

private:
	Asura_TreeDialog_Tree m_xTree;
	Tools_Container_Object* m_pxHighlightedAttachment;

	Asura_Hash_ID m_uSavedSelectedContainerHash;

	Asura_Hash_ID m_auCollapsed[MAX_NUM_COLLAPSED];
	u_int m_uNumCollapsedStored;
};

///////////////////////////////
// Tools_Container_Clipboard //
///////////////////////////////
// Class to help cut/paste/copy containers within the instance overlay.
class Tools_Container_Clipboard
{
public:
	Tools_Container_Clipboard();
	~Tools_Container_Clipboard();

	void Reset();

	void SetCut ();
	void SetCopy();
	void DoPaste (Tools_Container* pxPasteToContainer);

	bool IsEmpty() const;

	bool IsCut() const  { return !IsEmpty() && m_bShouldCut; }
	bool IsCopy() const { return !IsEmpty() && !m_bShouldCut; }

	Asura_Hash_ID GetParentContainerHash() const { return m_uParentContainerHash; }
	Asura_Hash_ID GetCopyContainerHash()   const { return m_uCopyContainerHash; }

private:
	void DoCut (Tools_Container* pxPasteToContainer);
	void DoCopy(Tools_Container* pxPasteToContainer);

private:
	Asura_Hash_ID m_uCopyContainerHash;			// Hash of container to be copied.
	Asura_Hash_ID m_uParentContainerHash;		// Hash of parent container.
	Asura_Hash_ID m_uGrandparentContainerHash;	// Hash of grandparent container.

	bool m_bShouldCut;
};

/////////////////////////////////
// Tools_Container_UndoStorage //
/////////////////////////////////
// Class used to store some limited amount of undo information.
// Currently it only stores deleted attachments, and state information.
class Tools_Container_UndoStorage
{
public:
	Tools_Container_UndoStorage();
	~Tools_Container_UndoStorage();

	// Store/restore attachments.
	void StoreData(const Tools_Container_Attachment* pxContainerToStore);
	void RestoreData(Tools_Container* pxContainerToRestore, const Asura_Hash_ID uKeyContainerHash);

	// Store/restore state information.
	void StoreStateData(const Tools_Container* pxParentContainer, const Asura_Hash_ID uKeyContainerHash, const Asura_Hash_ID uChildContainerHash);
	void RestoreStateData(Tools_Container* pxNewParentContainer, const Asura_Hash_ID uKeyContainerHash, const Asura_Hash_ID uChildContainerHash);

private:
	void PopTailEntry_Attachment();
	void PopTailEntry_State();

	// Attachment undo data.
	struct UndoData_Attachment
	{
		UndoData_Attachment()
			: m_uKeyContainerHash(ASURA_HASH_ID_UNSET)
			, m_pucStreamData(0)
			, m_uNumBytes(0)
		{}

		~UndoData_Attachment()
		{
			delete[] m_pucStreamData;
		}

		Asura_Hash_ID m_uKeyContainerHash;
		u_char* m_pucStreamData;
		u_int m_uNumBytes;
	};

	enum { MAX_NUM_STATES = 32 };

	// State undo data.
	struct UndoData_State
	{
		UndoData_State();

		void SetState(Asura_Hash_ID uStateHash, Asura_Hash_ID uChildStateHash, u_int uInputIndex);
		Asura_Hash_ID GetState(Asura_Hash_ID uStateHash, u_int uInputIndex) const;
		void RemoveState(Asura_Hash_ID uStateHash, u_int uInputIndex);
		bool IsMatch(u_int uIndex, Asura_Hash_ID uStateHash, u_int uInputIndex) const;

		Asura_Hash_ID m_uKeyContainerHash;
		Asura_Hash_ID m_auCollectionStateHash[MAX_NUM_STATES];
		Asura_Hash_ID m_auChildStateHash[MAX_NUM_STATES];
		u_int m_auInputIndex[MAX_NUM_STATES];
	};

	UndoData_State* FindState(Asura_Hash_ID uKeyContainerHash);

	// Data
	Asura_List<UndoData_Attachment> m_xAttachmentList;
	Asura_List<UndoData_State> m_xStateList;
};

//////////////////////////
// Tools_Mode_Container //
//////////////////////////
class Tools_Mode_Container : public Tools_Mode
{
	typedef Tools_Mode PARENT;

public:
	Tools_Mode_Container();
	virtual ~Tools_Mode_Container();

	virtual void InitialiseMode();
	virtual void ShutdownMode();
	virtual void PreShutdown();

	virtual void Update();
	virtual void HandleKeyboard();

	// Render
	virtual void RenderWorld_PostToneMap() {}
	virtual void RenderUI() {}
	virtual void EarlyRender();
	virtual bool ShouldRenderDuringModalOverlay() { return false; }

	// Save
	virtual void PreSave() {}
	virtual void PostSave() {}
	virtual void QuickSave();
	virtual bool IsModified();

	// So modes can hide/unhide objects as needed for editing message blocks
	virtual void PreMessageBlockEditing() {}
	virtual void PostMessageBlockEditing() {}

	// Called when a module's mesh gets changed behind the mode's back.
	// Any cached data based on the mesh will need to be fixed up as best it can.
	virtual void HandleMeshChange(int iModuleIndex) {}

	// Create child tools object.
	static Tools_Container_Object* CreateChild(Asura_Hash_ID uContainerHash);

	// Mark beginning and end of deletion process.
	static void MarkDeletionStart();
	static void MarkDeletionStop();
	static int  GetDeletionCounter() { return s_iDeletionCounter; }

	// Refresh the instance overlay when the number of collection inputs has been changed.
	static void ChangedNumberOfInputs(const Asura_Hash_ID uContainerChanged, u_int uNumInputs);

	// Refresh the instance overlay when a container has been renamed.
	static void RefreshRename(const Asura_Hash_ID uContainerRenamed);

	// Refresh the instance when a container has been modified.
	static void RefreshContainerModified(const Asura_Hash_ID uContainerChanged, bool bMinorChange);

	// Refresh the instance when a container has been modified.
	void Refresh(const Asura_Hash_ID uContainerChanged, bool bMinorChange);

	// Change the preview state.
	static void RequestPreviewState(const Asura_Container_InputState& xInputState, Asura_Hash_ID uCollectionHash);
	static void RequestPreviewStateFromDialogue();

	// Request this mode for a specific container entity.
	static void RequestModeForSelectedContainerEntity(Asura_Hash_ID uContainerHash, Asura_Guid uEntityGuid);
	static void RequestModeForSelectedContainer(Asura_Hash_ID uContainerHash);

	// Destroy both tools and client instances.
	static void DestroyAllInstances();
	// Create both tools and client instances.
	static void CreateAllInstances();

	// Display name of container.
	static void PrintInfo(float& fX, float& fY);

	// Mode data
	static void TransferModeData(int iMode);

	// Instance overlay
	static void ToggleInstanceOverlay();
	static bool IsInstanceOverlayOpen();
    static void CreateInstanceOverlay();
	static void DestroyInstanceOverlay();
	static void SelectInstanceContainer(const Asura_Hash_ID uContainerHash);

	virtual void AddInstanceOverlayItems(Asura_Overlay& xOverlay, Asura_Vector_2& xOffset);
	virtual void RefreshInstanceOverlay(bool bDeleting);
	virtual void RefreshStateToPreview();
	virtual void UpdateInstanceOverlay();

	static void UpdateInstanceOverlayStateInfo();

	static Asura_Hash_ID GetContainerToEditHash() { return s_uContainerToEdit; }

	static const Asura_Container_InputState& GetPreviewState() { return s_xStateToPreview; }
	static Asura_Hash_ID GetPreviewStateHash(u_int uIndex) { return s_xStateToPreview.GetStateHash(uIndex); }

	static Asura_ClientContainerInstance* GetClientInstance() { return (s_pxThis ? s_pxThis->m_pxInstance : 0); }

	// Insert item in instance overlay tree. Return new item.
	static Asura_TreeDialog_Item* InsertItemInTree(Asura_TreeDialog_Item* pxParent, const char* szLabel,
												   Asura_Hash_ID uContainerHash, Tools_Container_Object* pxToolsObject);

	// Find tools object
	static Tools_Container_Object* FindToolsObjectByContainerHash(const Asura_Hash_ID uContainerHash);

	static void GetSelected(Asura_Hash_ID& uSelectedHash, Asura_Hash_ID& uParentHash, Asura_Hash_ID& uGrandparentHash);

	// Selected container (or a parent) for which a child can be added.
	static Tools_Container* GetContainerSelectedParentForCanAddChild();

protected:
	// Instance
	void CreateInstance();
	void DestroyInstance();
	void CreateToolsInstance();
	void DestroyToolsInstance();

	enum { CURRENT_MODE_SAVE_VERSION = 1 };

	// Callbacks
	#define OVERLAY_VOID_CALLBACK(__FnName)\
	static void __FnName(int iThis) {((Tools_Mode_Container*)iThis)->__FnName();}\
	void __FnName();

	OVERLAY_VOID_CALLBACK(AddChild);
	OVERLAY_VOID_CALLBACK(EditChild);
	OVERLAY_VOID_CALLBACK(RemoveChild);
	OVERLAY_VOID_CALLBACK(InsertAttachment);
	OVERLAY_VOID_CALLBACK(RemoveAttachment);
	OVERLAY_VOID_CALLBACK(Retrigger);
	
	#undef OVERLAY_VOID_CALLBACK

	static int GetTreeItemIconIndex(Asura_TreeDialog_Item* pxItem);
	static void TreeIconClicked(Asura_TreeDialog_Item* pxItem, bool bDrag);
	static void DragAndDropTreeItem(Asura_TreeDialog_Item* pxDropTarget);

	static void CreateContainer(int iInsertClass);

	static Tools_Container* GetContainerSelected();
	static Tools_Container* GetContainerSelectedParent();

	// Context menu.
	static void AddContextMenu(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext);

	static void ContextMenu_Edit();
	static void ContextMenu_InsertAttachment();
	static void ContextMenu_RemoveAttachment();
	static void ContextMenu_Rename();
	static void ContextMenu_ToggleShowName();
	static void ContextMenu_Delete();

	static void ContextMenu_Cut();
	static void ContextMenu_Copy();
	static void ContextMenu_Paste();

public:
	static Asura_Hash_ID GetInstanceContainerSelected();

	// Validity (to enable/disable options).
	static bool IsInstanceContainerSelectionValid() { return GetInstanceContainerSelected() != ASURA_HASH_ID_UNSET; }
	static bool IsInstanceContainerSelectionValidForRemoval();
	static bool IsValidForCreate();
	static bool IsValidForCreateWithAttachment();
	static bool IsValidForAdd();
	static bool IsValidForInsertAttachment();
	static bool IsValidForRemoveAttachment();
	static bool IsValidForCutOrCopy();
	static bool IsValidForPaste();

protected:
	bool m_bOverlayWasOpen;
	bool m_bEditOverlayWasOpen;

	Asura_ClientContainerInstance* m_pxInstance; // Instance for rendering in tools.
	class Tools_Container_Object_Origin* m_pxMainToolsObject;

	Asura_Handle m_iMenuHandle;	// Handle for "Container Mode" menu.

	// Static
	static Tools_Mode_Container* s_pxThis;

	static Asura_Hash_ID s_uContainerToEdit;
	static Asura_Container_InputState s_xStateToPreview;
	static Asura_Container_InputState s_xStateToPreviewOld;
	static Asura_Guid s_uEntityToEdit;
	static bool s_bOpenFromFileOverlay;

	static Asura_Vector_3 s_xPosition;
	static Asura_Matrix_3x3 s_xOrientation;

	// Instance dialogue
	static Asura_Vector_2 s_xInstanceOverlayPosition;
	static Asura_Vector_2 s_xInstanceTreeDimensions;
	static bool s_bInstanceOverlayOpen;
	static Tools_Container_InstanceTree s_xInstanceTree;
	Asura_Hash_ID m_uPreviousInstanceContainerHash;

	// Deletion counter
	static int s_iDeletionCounter;

	// Clipboard
	static Tools_Container_Clipboard s_xClipboard;
	static Tools_Container_UndoStorage s_xUndoStorage;

	// Friend
	friend class Tools_Container_Object_Origin;
	friend class Tools_Overlay_ContainerInstance;
	friend class Tools_Container_Clipboard;
};


///////////////////////////////////
// Tools_Container_Object_Origin //
///////////////////////////////////
// Object for container mode. Represents the origin of the container instance.
class  Tools_Container_Object_Origin : public Tools_PositionEdit_Object
{
	typedef Tools_PositionEdit_Object PARENT;
public:
	Tools_Container_Object_Origin(Asura_Hash_ID uContainerHash);
	virtual ~Tools_Container_Object_Origin();

	// Prevent deletion by user.
	virtual bool CanBeDestroyed() { return false; }

	virtual const char* GetTypeDescription() { return "Origin"; }

	virtual void PrintInfo(float& fX, float& fY);

	virtual void UpdatePositionInAsuraData(bool bUpdateLinkedObjects);

	virtual void TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri);

	virtual bool CanBeMoved() { return true; }

	bool HasBeenMoved() const { return m_bMoved; }
	void ClearMovedFlag() { m_bMoved = false; }

	virtual void OnSelection();

	void AddToInstanceOverlayTree(Asura_TreeDialog_Item* pxParent);

	Tools_Container_Object* FindContainer(Asura_Hash_ID uContainerHash);

	void ChangeState(const Asura_Container_InputState& xInputState);
	const Asura_Container_InputState& GetInputState() const { return m_xInputState; }

protected:
	Tools_Container_Object* m_pxChild;
	Asura_Hash_ID m_uContainerHash;
	Asura_Container_InputState m_xInputState;

	bool m_bMoved; // Has been moved by the user?

	static bool s_bWasHidden; // Store hidden status.
};

////////////////////////////
// Tools_Container_Object //
////////////////////////////
// Object for container mode. Base object.
class  Tools_Container_Object : public Tools_Object
{
	typedef Tools_Object PARENT;
public:
	Tools_Container_Object(Asura_Hash_ID uContainerHash);
	virtual ~Tools_Container_Object();

	virtual void PostCreateInitialise();

	// Prevent deletion by user.
	virtual bool CanBeDestroyed() { return false; }

	virtual const char* GetName() { return m_xName.GetString(); }
	virtual const char* GetTypeDescription() { return m_szDescription; }

	virtual void PrintInfo(float& fX, float& fY);

	virtual void TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri);

	virtual bool CanBeMoved() { return false; }

	virtual void OnSelection();

	Asura_Hash_ID GetContainerHash() const { return m_uContainerHash; }
	u_int GetClassification() const { return m_uClassification; }

	void SetParent(Tools_Container_Object* pxParent) { m_pxParent = pxParent; }
	Tools_Container_Object* GetParent() const { return m_pxParent; }
	bool IsParentAnAttachment() const;

	virtual void SetUniqueID(u_int uUniqueID) { m_uUniqueID = uUniqueID; }
	u_int GetUniqueID() const { return m_uUniqueID; }

	void AddToInstanceOverlayTree(Asura_TreeDialog_Item* pxParent);

	Tools_Container_Object* FindContainer(Asura_Hash_ID uContainerHash);

	// Return the client instance if it exists (i.e. is enabled). Will always return 0 for attachments.
	const class Asura_ClientContainerInstance* GetClientInstance() const;

	virtual void ChangeState(const Asura_Container_InputState& xInputState);
	const Asura_Container_InputState& GetInputState() const { return m_xInputState; }
	bool IsOff() const { return m_xInputState.IsOff(); }

protected:
	const char* m_szDescription; // Description e.g. "Mesh", "Sound Event".
	Asura_String m_xName;		 // Name of container.

	Asura_Hash_ID m_uContainerHash;
	Asura_Container_InputState m_xInputState;

	u_int m_uClassification;
	u_int m_uUniqueID;
	u_int m_uMovedByUserFrameNumber;

	Asura_List<Tools_Container_Object> m_xChildren;
	Tools_Container_Object* m_pxParent;

	// Static
public:
	static void ResetStoredHiddenFlag();

protected:
	static void SetWasHidden(u_int uUniqueID, Asura_Hash_ID uContainerHash);
	static bool WasHidden(u_int uUniqueID, Asura_Hash_ID uContainerHash);

protected:
	// Store data about user-hidden objects, so that we can restore the flags upon rebuilding the objects.
	enum { MAX_NUM_HIDDEN_STORED = 16 };
	static u_int s_auHiddenObjectID[MAX_NUM_HIDDEN_STORED];
	static Asura_Hash_ID s_auHiddenObjectHash[MAX_NUM_HIDDEN_STORED];
};

/////////////////////////////////
// Tools_Container_Object_Misc //
/////////////////////////////////
// Object for container mode. Represents a misc object (e.g. mesh, sound event etc)
// in the container instance.
class  Tools_Container_Object_Misc : public Tools_Container_Object
{
	typedef Tools_Container_Object PARENT;

public:
	Tools_Container_Object_Misc(Asura_Hash_ID uContainerHash);

	virtual void PostCreateInitialise()
	{
		PARENT::PostCreateInitialise();
		UpdateMeshHidden();
	}

	virtual bool CanBeMoved()
	{
		return (m_pxParent && m_pxParent->Type == TOT_Container_Object_Attachment);
	}

	virtual void UpdatePositionInAsuraData(bool bUpdateLinkedObjects);
	
	virtual const char* GetTypeDescriptionForObjectType() { return "Other"; }

	virtual bool ShouldTestForLOS();
	virtual bool ObjectIsInLOS(const Asura_Vector_3& xOrigin, const Asura_Vector_3& xDir, float& fLowest);

	virtual void Update();

protected:
	void UpdateMeshHidden();

protected:
};

///////////////////////////////////////
// Tools_Container_Object_Attachment //
///////////////////////////////////////
// Object for container mode. Represents an attachment in the container instance.
class  Tools_Container_Object_Attachment : public Tools_Container_Object
{
	typedef Tools_Container_Object PARENT;

public:
	Tools_Container_Object_Attachment(Asura_Hash_ID uContainerHash);
	virtual ~Tools_Container_Object_Attachment();

	virtual void PrintInfo(float& fX, float& fY);

	virtual void UpdatePositionInAsuraData(bool bUpdateLinkedObjects);

	virtual void TeleportTo(const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOri);

	virtual bool CanBeMoved() { return true; }

	virtual void SetUniqueID(u_int uUniqueID);

	virtual void ChangeState(const Asura_Container_InputState& xInputState);

protected:
	Asura_Hash_ID m_uChildHash;

	Asura_Vector_3 m_xParentPos;
	Asura_Matrix_3x3 m_xParentOri;
};

///////////////////////////////////////
// Tools_Container_Object_Collection //
///////////////////////////////////////
// Object for container mode. Represents a collection in the container instance.
class  Tools_Container_Object_Collection : public Tools_Container_Object
{
	typedef Tools_Container_Object PARENT;

public:
	Tools_Container_Object_Collection(Asura_Hash_ID uContainerHash);
	virtual ~Tools_Container_Object_Collection() {}

	virtual bool CanBeMoved()
	{
		return (m_pxParent && m_pxParent->Type == TOT_Container_Object_Attachment);
	}

	virtual void UpdatePositionInAsuraData(bool bUpdateLinkedObjects);

	virtual void SetUniqueID(u_int uUniqueID);

	virtual void ChangeState(const Asura_Container_InputState& xInputState);
};

#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

#endif //__TOOLS_MODE_CONTAINER_H__

