//////////////////////////////////////////////////////////////////////////////
//
// Tools_Entity_DestructibleLight.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

#ifndef ASURA_TARGET_SM3

// This file's header.
#include "Entity/Tools_Entity_DestructibleLight.h"

// Engine independent.
#include "Entity/PhysicalObject/Asura_ClientEntity_DestructibleLight.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Light/Asura_Photon.h"
#include "SpecialFX/Asura_Corona_System.h"

// Tools.
#include "Interface/Tools_ColourPicker.h"
#include "Objects/Tools_Object_System.h"
#include "Objects/Tools_VolumeEdit_Object.h"
#include "Texturing/Tools_PickTexture.h"

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// Tools_Entity_DestructibleLight
//

const char* const Tools_Entity_DestructibleLight::s_szLongName  = "Destructible Light";
const char* const Tools_Entity_DestructibleLight::s_szShortName = "DestLight";
Tools_Entity_DestructibleLight* Tools_Entity_DestructibleLight::s_pxSelectedObject = NULL;
Asura_Hashed_List<Tools_Entity_DestructibleLight> Tools_Entity_DestructibleLight::s_xList;


Tools_Entity_DestructibleLight::Tools_Entity_DestructibleLight(Asura_ServerEntity_DestructibleLight* pxDestLight)
: Tools_Entity_PhysicalObject(NULL) // ignored
, m_pxPhoton(NULL)
, m_pxDestLight(pxDestLight)
, m_iCoronaHandle(ASURA_INVALID_HANDLE)
{
    Asura_Assert(pxDestLight && 
		pxDestLight->GetClassification() == AsuraEntityClass_DestructibleLight ||
		pxDestLight->GetClassification() == AsuraEntityClass_AdvancedLight, "Not a destructible light?");
	s_xList.AddToList(this);

	Type = TOT_Entity_DestructibleLight;
	if (!pxDestLight) return; // give up

	m_uVolumeEditID = 0;

	//Use the light bulb mesh by default. RAR 26/09/2006
	SetDefaultMesh("ToolsShape_Bulb");
	

	//See if the client entity for this exists
	Asura_ClientEntity_DestructibleLight* pxClientEntity = Asura_ClientEntity_DestructibleLight::FindEntity(pxDestLight->GetGuid());
	if(pxClientEntity && pxClientEntity->m_pxPhoton)
	{
		//reuse the existing photon
		m_pxPhoton = pxClientEntity->m_pxPhoton;
		m_bUsingClientPhoton = true;
	}
	else
	{
		//create a temporary photon for previewing the light in tools
		m_pxPhoton = Asura_Photon_System::CreatePhoton();
		m_bUsingClientPhoton = false;
	    Asura_Assert(m_pxPhoton, "failed to create photon");
	}

	if(pxClientEntity && pxClientEntity->m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		m_iCoronaHandle = pxClientEntity->m_iCoronaHandle;
		m_bUsingClientCorona = true;
	}
	else
	{
		m_bUsingClientCorona = false;
	}
    
    SetPhotonToValuesOfEntity();

}


Tools_Entity_DestructibleLight::~Tools_Entity_DestructibleLight()
{
	s_xList.RemoveFromList(this);
    if (m_pxPhoton && !m_bUsingClientPhoton)
    {
        Asura_Photon_System::DestroyPhoton(m_pxPhoton);
    }
	if(m_iCoronaHandle != ASURA_INVALID_HANDLE && !m_bUsingClientCorona)
	{
		Asura_Corona_System::DestroyCorona(m_iCoronaHandle);
	}

	//get rid of the bounding box editing object ,if it exists
	if(m_uVolumeEditID) Tools_VolumeEdit_Object::Delete(m_uVolumeEditID);

}

void Tools_Entity_DestructibleLight::OnSelection()
{
	s_pxSelectedObject = this;
	PARENT::OnSelection();
}

void Tools_Entity_DestructibleLight::OnDeselection()
{
	PARENT::OnDeselection();
	s_pxSelectedObject = NULL;
}



void Tools_Entity_DestructibleLight::AddOverlayControls(Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset)
{
	// No call to parent due to multiple inheritance fun

	if (!m_pxDestLight) return; // give up

    Asura_Dialog* pxNewDialog;

	// RWM - Why we have to do this is beyond me... but PhysicsObject tests whether the type is correct 
	// before adding the health overlay, so we need to do the same here to get the controls to appear and 
	// not duplicate them in subclasses...
	if (Type == TOT_Entity_DestructibleLight)
	{
		AddHealthOverlayControl(pxNewOverlay, xOffset);
	}

	xOffset.x += 14.0f;
	pxNewDialog = new Tools_ColourPicker_Dialog("Light Colour",xOffset,Asura_Vector_2(4.0f,3.0f),&(m_pxDestLight->m_fR),&(m_pxDestLight->m_fG),&(m_pxDestLight->m_fB),255.0f,NULL);
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.x -= 14.0f;

	pxNewDialog  = new Asura_Numerical_Dialog("Red",xOffset,&(m_pxDestLight->m_fR), 0.0f,255.0f,1.0f,"%0.01f");
	pxNewDialog->SetHelpText("Red component for light colour (0-255).");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

    pxNewDialog = new Asura_Numerical_Dialog("Green",xOffset,&(m_pxDestLight->m_fG), 0.0f,255.0f,1.0f,"%0.01f");
	pxNewDialog->SetHelpText("Green component for light colour (0-255).");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

    pxNewDialog = new Asura_Numerical_Dialog("Blue",xOffset,&(m_pxDestLight->m_fB), 0.0f,255.0f,1.0f,"%0.01f");
	pxNewDialog->SetHelpText("Blue component for light colour (0-255).");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

    pxNewDialog = new Asura_Numerical_Dialog("Brightness",xOffset,&(m_pxDestLight->m_fBrightness),0.0f,10.0f,0.1f,"%2.1f");
	pxNewDialog->SetHelpText("Brightness multiplier for light.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_Numerical_Dialog("Range",xOffset,&(m_pxDestLight->m_fRange), 0,999.0f,0.1f,"%0.1f");
	pxNewDialog->SetHelpText("Radius of light in metes.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_FlagTickbox_Dialog("Don't Light Env", xOffset, &(m_pxDestLight->m_uDestructibleLightFlags), AsuraDestructibleLightFlag_DontLightEnv,&Changed);
	pxNewDialog->SetHelpText("If ticked then the light won't affect the environment.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_FlagTickbox_Dialog("Has corona", xOffset, &(m_pxDestLight->m_uDestructibleLightFlags), AsuraDestructibleLightFlag_HasCorona,&Changed);
	pxNewDialog->SetHelpText("Light has a corona , and should use the settings below.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_FlagTickbox_Dialog("Corona: Dont Fade With Distance", xOffset, &(m_pxDestLight->m_uDestructibleLightFlags), AsuraDestructibleLightFlag_CoronaDontFadeWithDistance,&Changed);
	pxNewDialog->SetHelpText("Don't fade out the corona with distance.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
 
    pxNewDialog = new Asura_Numerical_Dialog("Corona Radius",xOffset,&(m_pxDestLight->m_fCoronaRadius),0.0f,99.0f,0.1f,"%0.3f",&Changed);
	pxNewDialog->SetHelpText("Corona radius in metres.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_Function_Dialog("Corona Texture",xOffset,_SelectCoronaTexture);
	pxNewDialog->SetHelpText("Choose texture for corona.");
	pxNewDialog->m_iControlID = DialogID_Texture;
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_FlagTickbox_Dialog("Use Bounding Box",xOffset, &(m_pxDestLight->m_uDestructibleLightFlags), AsuraDestructibleLightFlag_UseBoundingBox, &Changed);
	pxNewDialog->SetHelpText("Light should only affect objects and modules that intersect its bounding box..");
	pxNewOverlay->AddDialog( pxNewDialog );
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_Function_Dialog("Edit Box",xOffset,_EditBoundingBox);
	pxNewDialog->SetHelpText("Edit the box for this light.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


}

void Tools_Entity_DestructibleLight::UpdateOverlayControls(Asura_Overlay* pxOverlay)
{
	char szText[100];


	//update the select texture control
	int iTextureID = Asura_Texture_Management::GetIndexFromHashID(m_pxDestLight->m_uCoronaTexture); 
	if(iTextureID == ASURA_TEXTURE_NONE)
	{
		strcpy(szText,"Corona Texture : <Not Set>");
	}
	else
	{
		const Asura_String* pxName = Asura_Texture_Management::GetTextureName(iTextureID);
		if(pxName)
		{
			//get the name without path/extension
			Asura_String xFilename;
			pxName->SplitFilename(0,&xFilename,0);

			Asura_PrintfToBfr(szText,"Corona Texture : %s",100,true,xFilename.GetString());
		}
		else
		{
			strcpy(szText,"Corona Texture : <Unknown>");
		}

	}
	pxOverlay->UpdateDialogLabel(DialogID_Texture,szText);
}




//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_DestructibleLight::UpdateWhileSelected().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  virtual function override
// AUTHOR       Mike Healey - 04.09.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_DestructibleLight::UpdateWhileSelected()
{
    PARENT::UpdateWhileSelected();

    SetPhotonToValuesOfEntity();
}



void Tools_Entity_DestructibleLight::SetPhotonToValuesOfEntity()
{
	if (!m_pxDestLight) return; // give up

	UpdateCorona();
	
    if (m_pxPhoton)
    {
		float fScale = (m_pxDestLight->m_fBrightness / 255.0f) / 2.0f;

       	SetPhotonAndCoronaColour( m_pxDestLight->m_fR * fScale, m_pxDestLight->m_fG * fScale, m_pxDestLight->m_fB * fScale);
        m_pxPhoton->SetRange( m_pxDestLight->m_fRange );
//		m_pxPhoton->SetCorona( m_pxDestLight->HasCorona() );
		m_pxPhoton->SetPosition( m_pxDestLight->GetPosition() );
		m_pxPhoton->SetUseBoundingBox(m_pxDestLight->m_bUseBoundingBox);
		m_pxPhoton->SetBoundingBox(m_pxDestLight->m_xBoundingBox);
		m_pxPhoton->SetDontLightEnv(m_pxDestLight->m_bDontLightEnv);
    }
}

void Tools_Entity_DestructibleLight::UpdateCorona()
{
	if(m_pxDestLight->HasCorona())
	{
		if(m_iCoronaHandle == ASURA_INVALID_HANDLE)
		{
			m_iCoronaHandle = Asura_Corona_System::CreateCorona();
		}
		bool bShowCorona = m_pxPhoton ? m_pxPhoton->IsActive() : false;
		Asura_Corona_System::SetFlag_Hidden(m_iCoronaHandle,!bShowCorona);

		int iTextureIndex = Asura_Texture_Management::GetIndexFromHashID(m_pxDestLight->m_uCoronaTexture);
		//update the corona values
		Asura_Corona_System::SetPosition(m_iCoronaHandle,m_pxDestLight->GetPosition());
		Asura_Corona_System::SetRadius(m_iCoronaHandle,m_pxDestLight->m_fCoronaRadius);
		Asura_Corona_System::SetTextureIndex(m_iCoronaHandle,iTextureIndex);
		Asura_Corona_System::SetFlag_DontFadeWithDistance(m_iCoronaHandle,m_pxDestLight->m_bCorona_DontFadeWithDistance);
	}
	else if(m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Corona_System::SetFlag_Hidden(m_iCoronaHandle,true);
	}
}

void Tools_Entity_DestructibleLight::SetPhotonAndCoronaColour(float fRed,float fGreen,float fBlue)
{
    if (m_pxPhoton)
    {
        m_pxPhoton->SetColours(fRed,fGreen,fBlue);
	}

	if(m_iCoronaHandle != ASURA_INVALID_HANDLE)
	{
		float fCoronaRed = fRed*2.0f;
		if(fCoronaRed>1.0f) fCoronaRed = 1.0f;
		float fCoronaGreen = fGreen*2.0f;
		if(fCoronaGreen>1.0f) fCoronaGreen = 1.0f;
		float fCoronaBlue = fBlue*2.0f;
		if(fCoronaBlue>1.0f) fCoronaBlue = 1.0f;

		Asura_Corona_System::SetColour(m_iCoronaHandle,fCoronaRed,fCoronaGreen,fCoronaBlue);
	}

}

void Tools_Entity_DestructibleLight::SetPhotonAndCoronaActive(bool bActive)
{
	if(m_pxPhoton)
	{
		m_pxPhoton->SetAsActive(bActive);
	}

	if(m_iCoronaHandle != ASURA_INVALID_HANDLE && m_pxDestLight->HasCorona())
	{
		Asura_Corona_System::SetFlag_Hidden(m_iCoronaHandle,!bActive);
	}
}


void Tools_Entity_DestructibleLight::_EditBoundingBox()
{
	if(s_pxSelectedObject) s_pxSelectedObject->EditBoundingBox();
}

void Tools_Entity_DestructibleLight::EditBoundingBox()
{
	if (!m_pxDestLight) return; // give up
		
	//see if the volume edit object already exists
	Tools_VolumeEdit_Object* pxVolumeObject = Tools_VolumeEdit_Object::Find(m_uVolumeEditID);

	if(!pxVolumeObject)
	{
		Asura_Bounding_Box* pxBB = &m_pxDestLight->m_xBoundingBox;

		//set a default value for the bounding box if needed
		if(pxBB->GetVolume()==0)
		{
			pxBB->Reset();
			pxBB->Set(Position);
			pxBB->Grow(0.5f);
		}

		char szName[200];
		Asura_PrintfToBfr(szName,"%s Volume",200,true,GetName());

		pxVolumeObject = new Tools_VolumeEdit_Object(szName,pxBB,0,&Changed);
		m_uVolumeEditID = pxVolumeObject->GetID();
		pxVolumeObject->SetOwner(this);

		if(!m_pxDestLight->m_bUseBoundingBox)
		{
			m_pxDestLight->m_bUseBoundingBox = true;
			Changed = true;
		}
	}

	Tools_Object_System::SelectObject(pxVolumeObject);
}

void Tools_Entity_DestructibleLight::SelectCoronaTexture()
{
	int iTextureID = Asura_Texture_Management::GetIndexFromHashID(m_pxDestLight->m_uCoronaTexture); 

	int iSelected = Tools_PickTexture::PickTexture("Select Corona Texture",iTextureID,PICKTEXTURE_ALLTEXTURES);

	if(iSelected != ASURA_TEXTURE_NONE && iSelected != iTextureID)
	{
		m_pxDestLight->m_uCoronaTexture = Asura_Texture_Management::GetHashIDFromIndex(iSelected);
		Changed = true;
		UpdateOverlay();
	}
	
}


Tools_Object* Tools_Entity_DestructibleLight::CreateObjectOfThisType() const
{
	Tools_Object* pxOldSelection = Tools_Object_System::SelectedObject;
	Tools_Entity_DestructibleLight::Project_Create();
	Tools_Object* pxNewSelection = Tools_Object_System::SelectedObject;

	return (pxOldSelection!=pxNewSelection) ? pxNewSelection : 0;
}

void Tools_Entity_DestructibleLight::Copy(const Tools_Object* pxObject)
{
	const Tools_Entity_DestructibleLight* pxFromObject = s_xList.FindEntity(pxObject->GetGuid());
	if(!pxFromObject)
	{
		Asura_Assert(pxFromObject,"Failed to find light in list");
		return;
	}

	//get the entities to copy between
	Asura_ServerEntity_DestructibleLight* pxFromEntity = pxFromObject->m_pxDestLight;
	Asura_ServerEntity_DestructibleLight* pxToEntity = m_pxDestLight;

	//copy between entities
	float fRed,fGreen,fBlue;
	pxFromEntity->GetRGB(fRed,fGreen,fBlue);
	pxToEntity->SetRGB(fRed,fGreen,fBlue);
	pxToEntity->SetRange(pxFromEntity->GetRange());
	pxToEntity->SetCorona(pxFromEntity->HasCorona());
	pxToEntity->SetCoronaRadius(pxFromEntity->GetCoronaRadius());
	pxToEntity->SetCoronaTextureHash(pxFromEntity->GetCoronaTextureHash());
	pxToEntity->m_uDestructibleLightFlags = pxFromEntity->m_uDestructibleLightFlags;
	pxToEntity->m_xBoundingBox = pxFromEntity->m_xBoundingBox;


	//update the preview light values
    SetPhotonToValuesOfEntity();


}

void Tools_Entity_DestructibleLight::SelectShape( void )
{
    PARENT::SelectShape();

    // Update the target entity's dynamics block's ignore flag status.
    // Usually this gets sorted out during load.
    if( m_pxDestLight )
    {
        m_pxDestLight->SetNoCollisionsFlag( !m_pxDestLight->HasShapeOrSkin() );
    }
}

#endif // #ifndef ASURA_TARGET_SM3
