/**
*
* \file Tools_Entity_ParticleGenerator_Object.cpp
*
* Created by James Paulin - 16/09/2002 17:38:05
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/

#include "Core/Asura.h"
#include "Entity/Tools_Entity_ParticleGenerator_Object.h"

#include "Debugging/Asura_Assert.h"
#include "Core/Asura_Maths.h"

#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Camera/Asura_Camera.h"
#include "Interface/Asura_Overlay_System.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_UserInput.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_RenderStates.h"
#include "Entity/Asura_Entity.h"
#include "Entity/Asura_Entity_Classification.h"
#include "SpecialFX/Asura_Particle.h"
#include "Entity/Asura_StaticMessage_System.h"

#include "SpecialFX/Asura_Particle_TemplateStore.h"
#include "SpecialFX/Asura_ParticleGenerator.h"
#include "Entity/ParticleGenerator/Asura_ServerEntity_ParticleGenerator.h"
#include "Entity/ParticleGenerator/Asura_ClientEntity_ParticleGenerator.h"
#include "ChunkSystem/Asura_Chunk_ParticleTemplates.h"

#include "Objects/Tools_Object_System.h"
#include "Tools_Overlays.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Texturing/Tools_Texturing_Texture.h"



//STATIC MEMBER VARIABLES///////////////////////////

Tools_Entity_ParticleGenerator_Object* Tools_Entity_ParticleGenerator_Object::s_pxThisObjectForTesting = NULL;
bool Tools_Entity_ParticleGenerator_Object::m_bPreviewEffect = false;
List<Tools_Entity_ParticleGenerator_Object*> Tools_Entity_ParticleGenerator_Object::s_xParticleGeneratorList;


static const char* GetTranslucencyDescription(int iType)
{
	switch(iType)
	{
		case ASURA_TRANSLUCENCY_OFF :
			return "Off";

		case ASURA_TRANSLUCENCY_NORMAL :
			return "Normal";

		case ASURA_TRANSLUCENCY_INVCOLOUR :
			return "Inverse Colour";

		case ASURA_TRANSLUCENCY_GLOWING :
			return "Glowing";

		case ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE:
			return "Heat Haze";
		
		default :
			return "*Unknown*";
	}
}



//MEMBER FUNCTIONS//////////////////////////////////

Tools_Entity_ParticleGenerator_Object::Tools_Entity_ParticleGenerator_Object(Asura_Entity* xEntity)
:Tools_Entity_Object(xEntity)
{
	s_xParticleGeneratorList.add_entry(this);

	//Set parameters based on those supplied in the ServerEntity entity
	//This is why this class is a FRIEND of Asura_ServerEntity_ParticleGenerator...
	Asura_ServerEntity_ParticleGenerator* pxServerEntity = (Asura_ServerEntity_ParticleGenerator*) xEntity;
	Asura_Assert(pxServerEntity, "Unable to retrieve server entity");
	if(pxServerEntity)
	{
		m_xPosition			 = pxServerEntity->m_xPosition;
		m_xOrientation		 = pxServerEntity->m_xOrientation;
		m_fVelocity			 = pxServerEntity->m_fVelocity;
		m_fRate				 = pxServerEntity->m_fRate;
		m_fDuration			 = pxServerEntity->m_fGeneratorLifetime;
		m_fRandomness		 = pxServerEntity->m_fRandomness;
		m_bInfiniteLifetime	 = pxServerEntity->m_bInfiniteLifetime;
		m_bDestroyable		 = pxServerEntity->m_bDestroyable;
		m_bStartsActive		 = pxServerEntity->m_bActive;
		m_bStaccatoGenerator = pxServerEntity->m_bStaccatoGenerator;
		m_fLaunchProbability = pxServerEntity->m_fLaunchProbability;
		
		//Template store
		m_bUseTemplateFromStore = pxServerEntity->m_bUseTemplateFromStore;
		// (MH - 21.11.2002) strcpy(m_szTemplateName, pxServerEntity->m_szTemplateName);
        strncpy(m_szTemplateName, pxServerEntity->m_szTemplateName, ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1);
        m_szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1] = 0;
		m_bUseTemplateStoreStatusHasChanged = false;
		
		//Wind
		m_xWindProperties.fMinWindVelocity		= pxServerEntity->m_xWindProperties.fMinWindVelocity;
		m_xWindProperties.fMaxWindVelocity		= pxServerEntity->m_xWindProperties.fMaxWindVelocity;
		m_xWindProperties.ucUseRandomDirection	= pxServerEntity->m_xWindProperties.ucUseRandomDirection;
		m_xWindProperties.ucUseRandomVelocity	= pxServerEntity->m_xWindProperties.ucUseRandomVelocity;
		m_xWindProperties.ucUseWind				= pxServerEntity->m_xWindProperties.ucUseWind;
		m_xWindProperties.xInitialWindDirection	= pxServerEntity->m_xWindProperties.xInitialWindDirection;
		m_pxWindVectorObject = NULL;

		//Copy particle template values from the server entity
		//This may be a particle store template, ie if server entity loaded from chunk
		//but thats fine.
		Asura_Assert(pxServerEntity->m_pxParticleTemplate, "Cannot retrieve particle template");
		if(pxServerEntity->m_pxParticleTemplate)
		{
			m_xParticleTemplate.TopLeftU			= pxServerEntity->m_pxParticleTemplate->TopLeftU;
			m_xParticleTemplate.TopLeftV			= pxServerEntity->m_pxParticleTemplate->TopLeftV;
			m_xParticleTemplate.BottomRightU		= pxServerEntity->m_pxParticleTemplate->BottomRightU;
			m_xParticleTemplate.BottomRightV		= pxServerEntity->m_pxParticleTemplate->BottomRightV;
			m_xParticleTemplate.DecayRate			= pxServerEntity->m_pxParticleTemplate->DecayRate;
			m_xParticleTemplate.InitialSize			= pxServerEntity->m_pxParticleTemplate->InitialSize;
			m_xParticleTemplate.EndSize				= pxServerEntity->m_pxParticleTemplate->EndSize;
			m_xParticleTemplate.ID					= pxServerEntity->m_pxParticleTemplate->ID;
			m_xParticleTemplate.TranslucencyMode	= pxServerEntity->m_pxParticleTemplate->TranslucencyMode;
			m_xParticleTemplate.UseColourLerp		= pxServerEntity->m_pxParticleTemplate->UseColourLerp;
			m_xParticleTemplate.UseDynamics			= pxServerEntity->m_pxParticleTemplate->UseDynamics;
			m_xParticleTemplate.UseLighting			= pxServerEntity->m_pxParticleTemplate->UseLighting;
			m_xParticleTemplate.UseMotionBlur		= pxServerEntity->m_pxParticleTemplate->UseMotionBlur;
			m_xParticleTemplate.UseRandomRotation	= pxServerEntity->m_pxParticleTemplate->UseRandomRotation;
			m_xParticleTemplate.UseSizeLerp			= pxServerEntity->m_pxParticleTemplate->UseSizeLerp;
			m_xParticleTemplate.UseStandardGravity	= pxServerEntity->m_pxParticleTemplate->UseStandardGravity;
			m_xParticleTemplate.UseVelocityDecay	= pxServerEntity->m_pxParticleTemplate->UseVelocityDecay;
			m_xParticleTemplate.InitialColour		= pxServerEntity->m_pxParticleTemplate->InitialColour;
			m_xParticleTemplate.EndColour			= pxServerEntity->m_pxParticleTemplate->EndColour;
		}

		//Attachment properties
		m_uiTargetGuid	= pxServerEntity->m_uiTargetGuid;
		m_xTargetOffset	= pxServerEntity->m_xTargetOffset;
		m_xTargetOrientationOffset = pxServerEntity->m_xTargetOrientationOffset;

		//Variable gravity
		m_fGravity = pxServerEntity->m_fGravity;


		
		//Set Tools entity properties (i.e mesh and flags)

		Type = TOT_Entity_ParticleGenerator;
		Asura_String xMeshName("ToolsShape_CutsceneTrackPoint");
		Mesh = Asura_Mesh_System::GetMesh(xMeshName);
		Position = m_xPosition;
		Asura_Maths::OrientationFromDirection(Orientation, m_xOrientation);

		//Create a client entity based on the server entity
		
		//Check to see if the client entity already exists
		m_pxClientEntity = Asura_ClientEntity_ParticleGenerator::FindEntity(pxServerEntity->GetGuid());
		if( !m_pxClientEntity )
		{
			//If there isn't already a client, make a new one.
			Asura_Entity_System::SetNextClientGuid(pxServerEntity->GetGuid());
			m_pxClientEntity =  new Asura_ClientEntity_ParticleGenerator(m_xPosition, m_xOrientation, m_fVelocity, m_fRate, m_fDuration, m_fRandomness, &m_xParticleTemplate, m_bInfiniteLifetime, m_bDestroyable, m_bStartsActive, &m_xWindProperties, m_uiTargetGuid, m_xTargetOffset, m_xTargetOrientationOffset, m_bStaccatoGenerator, m_fLaunchProbability, m_bUseTemplateFromStore, m_szTemplateName, m_fGravity);
		}
		else
		{
			// If we've already got a client entity, ask it to stop
			m_pxClientEntity->m_bGeneratorActive = false;
		}
		m_bDecayRateChanged = false;

			//Set local parameters to initialise dialog boxes
		m_bDestroyable			= false; 
		m_fDecayRate			=  m_xParticleTemplate.DecayRate;
		m_bUseDynamics			= (m_xParticleTemplate.UseDynamics != 0)		? true : false;
		m_bUseStandardGravity	= (m_xParticleTemplate.UseStandardGravity != 0) ? true : false;
		m_bUseVelocityDecay		= (m_xParticleTemplate.UseVelocityDecay != 0)	? true : false;
		m_bUseColourLerp		= (m_xParticleTemplate.UseColourLerp != 0)		? true : false;
		m_bUseSizeLerp			= (m_xParticleTemplate.UseSizeLerp != 0)		? true : false;
		m_bUseLighting			= (m_xParticleTemplate.UseLighting != 0)		? true : false;
		m_bUseMotionBlur		= (m_xParticleTemplate.UseMotionBlur != 0)		? true : false;
		m_bUseRandomRotation	= (m_xParticleTemplate.UseRandomRotation != 0)	? true : false;
	
		//Wind
		m_bUseWind				= (m_xWindProperties.ucUseWind != 0)			? true : false;
		m_bUseRandomDirection	= (m_xWindProperties.ucUseRandomDirection != 0)	? true : false;
		m_bUseRandomVelocity	= (m_xWindProperties.ucUseRandomVelocity != 0)	? true : false;

	}//if able to recover server entity

	m_pxTextureEdit = 0;
	m_bGameEntitiesNeedUpdating = false;
	m_bTemplatePropertiesChanged = false;
	m_bDecayRateChanged = false;
}

Tools_Entity_ParticleGenerator_Object::~Tools_Entity_ParticleGenerator_Object()
{
	//May well require other shutdown code in here
	if (s_pxThisObjectForTesting == this)
	{
		if(s_pxThisObjectForTesting->m_pxWindVectorObject)
		{
			delete s_pxThisObjectForTesting->m_pxWindVectorObject;
			s_pxThisObjectForTesting->m_pxWindVectorObject = NULL;
		}
		s_pxThisObjectForTesting = NULL;
	}
	if(m_pxClientEntity)
	{
		//reset the active state back to the initial setting
		m_pxClientEntity->m_bGeneratorActive = m_pxClientEntity->m_bStartsActive;
		m_pxClientEntity->StopParticleGenerator();
	}
		
	s_xParticleGeneratorList.delete_entry(this);
}

//////////////////////////////////////
// Static Create function			//
//									//
//////////////////////////////////////


void Tools_Entity_ParticleGenerator_Object::Create(void)
{
	//Get a guid from the tools system
	Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();
	
	//Initialize a static particle template solely for the purposes of creating the entities
	CreateDummyParticleTemplate();
	CreateDummyWindProperties();

	Asura_Vector_3 xDirection(0.0f, -1.0f, 0.0f);
	Asura_Guid uDummyTargetGuid = ASURA_GUID_UNREGISTERED;
	Asura_Vector_3 xDummyTargetOffset(0.0f, 0.0f, 0.0f);
	Asura_Matrix_3x3 xDummyTargetOrientationOffset = Asura_Maths::IdentityMatrix3x3;
	float fDummyGravity = 10.0f;
	
	//Get position of new object from the camera
	Asura_Vector_3 xPosition = *Asura_Camera::GetPosition();

	//Create a default particle generator server entity
	Asura_ServerEntity_ParticleGenerator* pxNewPGenServer = new Asura_ServerEntity_ParticleGenerator(uGuid, xPosition, xDirection, 0.0f, 1.0f, 5.0f, 0.0f, &s_xDummyParticleTemplate, true, false, false, &s_xDummyWindProperties, uDummyTargetGuid, xDummyTargetOffset, xDummyTargetOrientationOffset, false, 1.0f, false, NULL, fDummyGravity);
	Asura_Assert(pxNewPGenServer, "Error: Unable to create particle generator object.\nPlease abort");

	//Store tools data (ie name of object)
	Tools_Entity_Data* toolsData = Tools_Entity_Object::CreateToolsData("Particle Generator",0,GetNamePrefix());
	//if(!toolsData) return;

	//Attach tools data to the server entity
	pxNewPGenServer->ToolsData = toolsData;

	//Create the corresponding tools entity
	Tools_Object* object = new Tools_Entity_ParticleGenerator_Object(pxNewPGenServer);
	Tools_Object_System::SelectObject(object);

}

/////////////////////////////////////////////////////////////////////////////////////
// Overloaded create function can create a copy of an existing particle generator ///
// at the current camera location												  ///
/////////////////////////////////////////////////////////////////////////////////////

void Tools_Entity_ParticleGenerator_Object::Create(const class Tools_Entity_ParticleGenerator_Object* pxBaseEntity)
{
	//Get a guid from the tools system
	Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();
	
	//Initialize a static particle template solely for the purposes of creting the entities
	//CreateDummyParticleTemplate();

	Asura_Vector_3 xDirection(0.0f, -1.0f, 0.0f);
	Asura_Guid uDummyTargetGuid = ASURA_GUID_UNREGISTERED;
	Asura_Vector_3 xDummyTargetOffset(0.0f, 0.0f, 0.0f);
	Asura_Matrix_3x3 xDummyTargetOrientationOffset = Asura_Maths::IdentityMatrix3x3;
	
	//Get orientation and position of new object from the camera
	Asura_Vector_3 xPosition = *Asura_Camera::GetPosition();

	//Create a default particle generator server entity
	Asura_ServerEntity_ParticleGenerator* pxNewPGenServer = 
		new Asura_ServerEntity_ParticleGenerator(uGuid, 
		xPosition, 
		xDirection, 
		pxBaseEntity->m_fVelocity, 
		pxBaseEntity->m_fRate, 
		pxBaseEntity->m_fDuration, 
		pxBaseEntity->m_fRandomness, 
		&pxBaseEntity->m_xParticleTemplate, 
		pxBaseEntity->m_bInfiniteLifetime, 
		false, 
		pxBaseEntity->m_bStartsActive, 
		&pxBaseEntity->m_xWindProperties, 
		uDummyTargetGuid, 
		xDummyTargetOffset, 
		xDummyTargetOrientationOffset, 
		pxBaseEntity->m_bStaccatoGenerator, 
		pxBaseEntity->m_fLaunchProbability, 
		pxBaseEntity->m_bUseTemplateFromStore, 
		pxBaseEntity->m_szTemplateName, 
		pxBaseEntity->m_fGravity
	);

	//Store tools data (ie name of object)
	Tools_Entity_Data* toolsData = Tools_Entity_Object::CreateToolsData("Particle Generator",0,GetNamePrefix());
	//if(!toolsData) return;

	//Attach tools data to the server entity
	pxNewPGenServer->ToolsData = toolsData;

	//Create the corresponding tools entity
	Tools_Object* object = new Tools_Entity_ParticleGenerator_Object(pxNewPGenServer);
	Tools_Object_System::SelectObject(object);


}

void Tools_Entity_ParticleGenerator_Object::UpdateWhileSelected()
{
	//Get a pointer to the currently selected object	
	s_pxThisObjectForTesting = this;


	if(!Asura_Overlay_System::OverlaysHaveKeyboardFocus())
	{
		if(!Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) && 
		   !Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT) &&
		   !Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
		{
			if(Asura_UserInput::DebouncedKeyPressed('\t'))
			{
				if(m_pxTextureEdit)
				{
					FinishUVEdit();
				}
				else
				{
					StartUVEdit();
				}
			}
		}
	}

	if(m_pxTextureEdit)
	{
		UpdateUVEdit();
	}

	Tools_Entity_Object::UpdateWhileSelected();

}

void Tools_Entity_ParticleGenerator_Object::Update()
{

	//Update wind properties with information from wind vector if neccessary.
	if(m_pxWindVectorObject)
	{
		if(m_pxWindVectorObject->m_bGameEntitiesNeedUpdating)
		{
			//Vector has moved so update wind vector (location is not required)
			m_xWindProperties.xInitialWindDirection.x = m_pxWindVectorObject->Orientation.e13;
			m_xWindProperties.xInitialWindDirection.y = m_pxWindVectorObject->Orientation.e23;
			m_xWindProperties.xInitialWindDirection.z = m_pxWindVectorObject->Orientation.e33;
			m_xWindProperties.xInitialWindDirection.Normalise();
			//reset flag
			m_pxWindVectorObject->m_bGameEntitiesNeedUpdating = false;
			//set flag so game entities get updated
			m_bGameEntitiesNeedUpdating = true;
		}

		//Remove object if neccessary...
		if(m_pxWindVectorObject->m_bReadyToDelete)
		{
			delete m_pxWindVectorObject;
			m_pxWindVectorObject = NULL;
		}
	}

	bool bUpdatePreview = m_bGameEntitiesNeedUpdating || m_bTemplatePropertiesChanged || m_bDecayRateChanged;


	//Update server and client entities with tools data
	if(m_bGameEntitiesNeedUpdating || m_bTemplatePropertiesChanged)
	{
		Changed = true;
		SetEntityModified();

		m_bGameEntitiesNeedUpdating = false;
			
		//If the user changes particle template properties but was previously using a default,
		//we need to revert to 'not using template from store'
		if( (m_bTemplatePropertiesChanged || m_bDecayRateChanged ) && !m_bUseTemplateStoreStatusHasChanged)
		{
			m_bUseTemplateFromStore = false;
		}


		//Update server entity params
		Asura_ServerEntity_ParticleGenerator* pxServerEntity = (Asura_ServerEntity_ParticleGenerator*) Entity;

		pxServerEntity->m_xPosition			 = m_xPosition;
		pxServerEntity->m_xOrientation		 = m_xOrientation;
		pxServerEntity->m_fVelocity			 = m_fVelocity;
		pxServerEntity->m_fRate				 = m_fRate;
		pxServerEntity->m_fGeneratorLifetime = m_fDuration;
		pxServerEntity->m_fRandomness		 = m_fRandomness;
		pxServerEntity->m_bInfiniteLifetime  = m_bInfiniteLifetime;
		pxServerEntity->m_bDestroyable		 = m_bDestroyable;
		pxServerEntity->m_bActive		 = m_bStartsActive;
		pxServerEntity->m_bStaccatoGenerator = m_bStaccatoGenerator;
		pxServerEntity->m_fLaunchProbability = m_fLaunchProbability;

		//Update client entity params
		m_pxClientEntity->m_xPosition			 = m_xPosition;
		m_pxClientEntity->m_xOrientation		 = m_xOrientation;
		m_pxClientEntity->m_fVelocity			 = m_fVelocity;
		m_pxClientEntity->m_fRate				 = m_fRate;
		m_pxClientEntity->m_fGeneratorLifetime	 = m_fDuration;
		m_pxClientEntity->m_fRandomness			 = m_fRandomness;
		m_pxClientEntity->m_bInfiniteLifetime	 = m_bInfiniteLifetime;
		m_pxClientEntity->m_bDestroyable		 = m_bDestroyable;
		m_pxClientEntity->m_bStartsActive		 = m_bStartsActive;
		m_pxClientEntity->m_bStaccatoGenerator   = m_bStaccatoGenerator;
		m_pxClientEntity->m_fLaunchProbability   = m_fLaunchProbability;
		

		//Server entity wind
		pxServerEntity->m_xWindProperties.fMinWindVelocity			= m_xWindProperties.fMinWindVelocity;
		pxServerEntity->m_xWindProperties.fMaxWindVelocity			= m_xWindProperties.fMaxWindVelocity;
		pxServerEntity->m_xWindProperties.ucUseRandomDirection		= m_xWindProperties.ucUseRandomDirection	 = (m_bUseRandomDirection) ? (u_char)1 : (u_char)0;
		pxServerEntity->m_xWindProperties.ucUseRandomVelocity		= m_xWindProperties.ucUseRandomVelocity		 = (m_bUseRandomVelocity)  ? (u_char)1 : (u_char)0;
		pxServerEntity->m_xWindProperties.ucUseWind					= m_xWindProperties.ucUseWind				 = (m_bUseWind)			   ? (u_char)1 : (u_char)0;
		pxServerEntity->m_xWindProperties.xInitialWindDirection		= m_xWindProperties.xInitialWindDirection;

		//Client entity wind
		m_pxClientEntity->m_xWindProperties.fMinWindVelocity		= m_xWindProperties.fMinWindVelocity;
		m_pxClientEntity->m_xWindProperties.fMaxWindVelocity		= m_xWindProperties.fMaxWindVelocity;
		m_pxClientEntity->m_xWindProperties.ucUseRandomDirection	= m_xWindProperties.ucUseRandomDirection	 = (m_bUseRandomDirection) ? (u_char)1 : (u_char)0;
		m_pxClientEntity->m_xWindProperties.ucUseRandomVelocity		= m_xWindProperties.ucUseRandomVelocity		 = (m_bUseRandomVelocity)  ? (u_char)1 : (u_char)0;
		m_pxClientEntity->m_xWindProperties.ucUseWind				= m_xWindProperties.ucUseWind				 = (m_bUseWind)			   ? (u_char)1 : (u_char)0;
		m_pxClientEntity->m_xWindProperties.xInitialWindDirection	= m_xWindProperties.xInitialWindDirection;


		//Update Client entity particle template information. Bother with server? YES!!!!
		m_pxClientEntity->m_pxParticleTemplate->TopLeftU			= m_xParticleTemplate.TopLeftU;			
		m_pxClientEntity->m_pxParticleTemplate->TopLeftV			= m_xParticleTemplate.TopLeftV;			
		m_pxClientEntity->m_pxParticleTemplate->BottomRightU		= m_xParticleTemplate.BottomRightU;		
		m_pxClientEntity->m_pxParticleTemplate->BottomRightV		= m_xParticleTemplate.BottomRightV;		
		m_pxClientEntity->m_pxParticleTemplate->DecayRate			= m_xParticleTemplate.DecayRate;			
		m_pxClientEntity->m_pxParticleTemplate->InitialSize			= m_xParticleTemplate.InitialSize;		
		m_pxClientEntity->m_pxParticleTemplate->EndSize				= m_xParticleTemplate.EndSize;			
		m_pxClientEntity->m_pxParticleTemplate->ID					= m_xParticleTemplate.ID;
		m_pxClientEntity->m_pxParticleTemplate->TranslucencyMode	= m_xParticleTemplate.TranslucencyMode;//		= ( ASURA_TRANSLUCENCY_MODE )m_iTranslucencyMode;
		//Flags
		m_pxClientEntity->m_pxParticleTemplate->UseColourLerp		= m_xParticleTemplate.UseColourLerp			= (m_bUseColourLerp)	  ? 1 : 0;		
		m_pxClientEntity->m_pxParticleTemplate->UseDynamics			= m_xParticleTemplate.UseDynamics			= (m_bUseDynamics)		  ? 1 : 0;
		m_pxClientEntity->m_pxParticleTemplate->UseLighting			= m_xParticleTemplate.UseLighting			= (m_bUseLighting)		  ? 1 : 0;		
		m_pxClientEntity->m_pxParticleTemplate->UseMotionBlur		= m_xParticleTemplate.UseMotionBlur			= (m_bUseMotionBlur)	  ? 1 : 0;		
		m_pxClientEntity->m_pxParticleTemplate->UseRandomRotation	= m_xParticleTemplate.UseRandomRotation		= (m_bUseRandomRotation)  ? 1 : 0;	
		m_pxClientEntity->m_pxParticleTemplate->UseSizeLerp			= m_xParticleTemplate.UseSizeLerp			= (m_bUseSizeLerp)		  ? 1 : 0;
		m_pxClientEntity->m_pxParticleTemplate->UseStandardGravity	= m_xParticleTemplate.UseStandardGravity	= (m_bUseStandardGravity) ? 1 : 0;	
		m_pxClientEntity->m_pxParticleTemplate->UseVelocityDecay	= m_xParticleTemplate.UseVelocityDecay		= (m_bUseVelocityDecay)   ? 1 : 0;	
		//Colours
		m_pxClientEntity->m_pxParticleTemplate->InitialColour.Red	= m_xParticleTemplate.InitialColour.Red;
		m_pxClientEntity->m_pxParticleTemplate->InitialColour.Green	= m_xParticleTemplate.InitialColour.Green;
		m_pxClientEntity->m_pxParticleTemplate->InitialColour.Blue	= m_xParticleTemplate.InitialColour.Blue;
		m_pxClientEntity->m_pxParticleTemplate->InitialColour.Alpha	= m_xParticleTemplate.InitialColour.Alpha;
		m_pxClientEntity->m_pxParticleTemplate->EndColour.Red	= m_xParticleTemplate.EndColour.Red;
		m_pxClientEntity->m_pxParticleTemplate->EndColour.Green	= m_xParticleTemplate.EndColour.Green;
		m_pxClientEntity->m_pxParticleTemplate->EndColour.Blue	= m_xParticleTemplate.EndColour.Blue;
		m_pxClientEntity->m_pxParticleTemplate->EndColour.Alpha	= m_xParticleTemplate.EndColour.Alpha;
		
		//Server entity particle template is required for saving the chunk
		pxServerEntity->m_pxParticleTemplate->TopLeftU			= m_xParticleTemplate.TopLeftU;			
		pxServerEntity->m_pxParticleTemplate->TopLeftV			= m_xParticleTemplate.TopLeftV;			
		pxServerEntity->m_pxParticleTemplate->BottomRightU		= m_xParticleTemplate.BottomRightU;		
		pxServerEntity->m_pxParticleTemplate->BottomRightV		= m_xParticleTemplate.BottomRightV;		
		pxServerEntity->m_pxParticleTemplate->DecayRate			= m_xParticleTemplate.DecayRate;			
		pxServerEntity->m_pxParticleTemplate->InitialSize		= m_xParticleTemplate.InitialSize;		
		pxServerEntity->m_pxParticleTemplate->EndSize			= m_xParticleTemplate.EndSize;			
		pxServerEntity->m_pxParticleTemplate->ID				= m_xParticleTemplate.ID;
		pxServerEntity->m_pxParticleTemplate->TranslucencyMode	= m_xParticleTemplate.TranslucencyMode;//		= ( ASURA_TRANSLUCENCY_MODE )m_iTranslucencyMode;
		//Flags
		pxServerEntity->m_pxParticleTemplate->UseColourLerp		= m_xParticleTemplate.UseColourLerp			= (m_bUseColourLerp)	  ? 1 : 0;		
		pxServerEntity->m_pxParticleTemplate->UseDynamics		= m_xParticleTemplate.UseDynamics			= (m_bUseDynamics)		  ? 1 : 0;
		pxServerEntity->m_pxParticleTemplate->UseLighting		= m_xParticleTemplate.UseLighting			= (m_bUseLighting)		  ? 1 : 0;		
		pxServerEntity->m_pxParticleTemplate->UseMotionBlur		= m_xParticleTemplate.UseMotionBlur			= (m_bUseMotionBlur)	  ? 1 : 0;		
		pxServerEntity->m_pxParticleTemplate->UseRandomRotation	= m_xParticleTemplate.UseRandomRotation		= (m_bUseRandomRotation)  ? 1 : 0;	
		pxServerEntity->m_pxParticleTemplate->UseSizeLerp		= m_xParticleTemplate.UseSizeLerp			= (m_bUseSizeLerp)		  ? 1 : 0;
		pxServerEntity->m_pxParticleTemplate->UseStandardGravity= m_xParticleTemplate.UseStandardGravity	= (m_bUseStandardGravity) ? 1 : 0;	
		pxServerEntity->m_pxParticleTemplate->UseVelocityDecay	= m_xParticleTemplate.UseVelocityDecay		= (m_bUseVelocityDecay)   ? 1 : 0;	
		//Colours
		pxServerEntity->m_pxParticleTemplate->InitialColour.Red	 = m_xParticleTemplate.InitialColour.Red;
		pxServerEntity->m_pxParticleTemplate->InitialColour.Green= m_xParticleTemplate.InitialColour.Green;
		pxServerEntity->m_pxParticleTemplate->InitialColour.Blue = m_xParticleTemplate.InitialColour.Blue;
		pxServerEntity->m_pxParticleTemplate->InitialColour.Alpha= m_xParticleTemplate.InitialColour.Alpha;
		pxServerEntity->m_pxParticleTemplate->EndColour.Red		 = m_xParticleTemplate.EndColour.Red;
		pxServerEntity->m_pxParticleTemplate->EndColour.Green	 = m_xParticleTemplate.EndColour.Green;
		pxServerEntity->m_pxParticleTemplate->EndColour.Blue	 = m_xParticleTemplate.EndColour.Blue;
		pxServerEntity->m_pxParticleTemplate->EndColour.Alpha	 = m_xParticleTemplate.EndColour.Alpha;


		//Server entity attachment parameters
		pxServerEntity->m_uiTargetGuid  = m_uiTargetGuid;
		pxServerEntity->m_xTargetOffset = m_xTargetOffset;
		pxServerEntity->m_xTargetOrientationOffset = m_xTargetOrientationOffset;

		//Client entity attachment parameters
		m_pxClientEntity->m_uiTargetGuid  = m_uiTargetGuid;
		m_pxClientEntity->m_xTargetOffset = m_xTargetOffset;
		m_pxClientEntity->m_xTargetOrientationOffset = m_xTargetOrientationOffset;

		//Update the server and client entity particle template store properties
		pxServerEntity->m_bUseTemplateFromStore = m_bUseTemplateFromStore;

		//Since strcpy is an expensize operation, only do it when the user
		//selects a template from teh store
		if(m_bUseTemplateFromStore && m_bUseTemplateStoreStatusHasChanged)
		{
			//Only need to update the server entity for chunk saving purposes...
			// (MH - 21.11.2002) strcpy(pxServerEntity->m_szTemplateName, m_szTemplateName);
            strncpy(pxServerEntity->m_szTemplateName, m_szTemplateName, ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1);
            pxServerEntity->m_szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1] = 0;
		}

		//Variable gravity
		pxServerEntity->m_fGravity   = m_fGravity;
		m_pxClientEntity->m_fGravity = m_fGravity;

	}

	if(m_bDecayRateChanged)
	{
		//Update decay rate in client and server entity
		m_xParticleTemplate.DecayRate = m_fDecayRate;
		Asura_ServerEntity_ParticleGenerator* pxServerEntity = (Asura_ServerEntity_ParticleGenerator*) Entity;
		pxServerEntity->m_pxParticleTemplate->DecayRate = m_fDecayRate;
		m_pxClientEntity->m_pxParticleTemplate->DecayRate = m_fDecayRate;

		m_bDecayRateChanged = false;
		
	}

	if(bUpdatePreview && m_pxClientEntity->m_bGeneratorActive)
	{
		m_pxClientEntity->UpdateParticleGeneratorProperties();
	}

	//Reset template store has changed status, since we'll have processed this flag by the
	//time Update gets to here.
	m_bUseTemplateStoreStatusHasChanged = false;
	m_bTemplatePropertiesChanged		= false;

	//Update Tools and the client.
	Tools_Entity_Object::Update();
	m_pxClientEntity->PerFrameUpdate();
}


void Tools_Entity_ParticleGenerator_Object::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{
	Tools_Entity_Object::UpdatePositionInAsuraData(bUpdateLinkedObjects);
	
	//Update position and orientation of tools object. Set flag so we know
	//to update the server and client entities
	
	m_bGameEntitiesNeedUpdating = true;

	m_xPosition = Position;
	m_xOrientation.x =  Orientation.e13;
	m_xOrientation.y =  Orientation.e23;
	m_xOrientation.z =  Orientation.e33;
	m_xOrientation.Normalise();

	//If we are attached to another object, update our offset and orientation
	if(m_uiTargetGuid != ASURA_GUID_UNREGISTERED)
	{
		Tools_Object* pxTargetEntity = Tools_Object_System::GetObjectFromGuid(m_uiTargetGuid);
		if(pxTargetEntity)
		{
			Asura_Matrix_3x3 xInverseTargetOrientation = pxTargetEntity->Orientation;
			xInverseTargetOrientation.Transpose();

			m_xTargetOffset = (m_xPosition - pxTargetEntity->Position) * xInverseTargetOrientation;
			//Also update our stored orientation offset
			m_xTargetOrientationOffset = xInverseTargetOrientation * Orientation;

		}
	}
}

void Tools_Entity_ParticleGenerator_Object::OnSelection(void)
{
	Tools_Entity_Object::OnSelection();

	//Close any dialog boxes relating to other objects

	//If the pg is attached to another entity, move it...
	if(m_uiTargetGuid != ASURA_GUID_UNREGISTERED)
	{
		Tools_Object* pxTargetEntity = Tools_Object_System::GetObjectFromGuid(m_uiTargetGuid);
		if(pxTargetEntity)
		{
			//Update position
			Position = m_xPosition = pxTargetEntity->Position + m_xTargetOffset * pxTargetEntity->Orientation;

			//Update orientation
			Orientation = pxTargetEntity->Orientation * m_xTargetOrientationOffset;
			m_xOrientation.x = Orientation.e13;
			m_xOrientation.y = Orientation.e23;
			m_xOrientation.z = Orientation.e33;
			m_xOrientation.Normalise();
			m_bGameEntitiesNeedUpdating = true;
		}
		else
		{
			//The user must have deleted the target entity - reset to GUID_UNREG
			m_uiTargetGuid = ASURA_GUID_UNREGISTERED;
			m_xTargetOffset = Asura_Maths::ZeroVector3;
			m_xTargetOrientationOffset.Identity();
			m_bGameEntitiesNeedUpdating = true;
		}
	}

}

void Tools_Entity_ParticleGenerator_Object::OnDeselection(void)
{
	FinishUVEdit();

//	m_pxClientEntity->m_bGeneratorActive = false;
//	m_pxClientEntity->StopParticleGenerator();
		
	//Close all overlays
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SELECTEDOBJECT);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_PARTICLEGENERATOR_SUBSIDIARY);
}


void Tools_Entity_ParticleGenerator_Object::AddOverlayControls(class Asura_Overlay* xNewOverlay,class Asura_Vector_2& xOffset)
{
	Tools_Entity_Object::AddOverlayControls(xNewOverlay,xOffset);

	Asura_Dialog* newDialog;
	Asura_Vector_2 offset = xOffset;
	xNewOverlay->ID = TOOLS_OVERLAY_MODE;
	


	//Generator parameters
	newDialog = new Asura_Numerical_Dialog("Velocity", offset, &m_fVelocity, 0.0f, 100.0f, 0.05f, "%.3f", &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Randomness", offset, &m_fRandomness, 0.0f, 100.0f, 0.05f, "%.3f", &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Release rate", offset, &m_fRate, 0.1f, 100.0f, 0.1f, "%.3f", &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Duration", offset, &m_fDuration, 0.1f, 100.0f, 0.05f, "%.3f", &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Particle decay rate", offset, &m_fDecayRate, 0.01f, 10.0f, 0.05f, "%.3f", &m_bDecayRateChanged);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Gravity", offset, &m_fGravity, 0.0f, 100.0f, 0.05f, "%.3f", &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Tickbox_Dialog("Infinite duration", offset, &m_bInfiniteLifetime, &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Initially active", offset, &m_bStartsActive, &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Staccato generation", offset, &m_bStaccatoGenerator, &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Staccato probability", offset, &m_fLaunchProbability, 0.0f, 1.0f, 0.001f, "%.3f", &m_bGameEntitiesNeedUpdating);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Function_Dialog( "Change particle properties", offset, ParticlePropertiesDialog );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Function_Dialog( "Copy this generator", offset, CopyParticleGenerator );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Choose generator template", offset, ChooseGeneratorTemplate );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Save generator template", offset, SaveGeneratorTemplate );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();


	// Particle template controls
	newDialog = new Asura_Dialog("-------------------------------",offset);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	//JP: moved preview button...
	newDialog = new Asura_Function_Dialog( "(Re)Start Selected",offset, TestParticleGenerator );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Stop Selected",offset, StopParticleGenerator );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Stop All",offset, StopAllParticleGenerators );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Dialog("-------------------------------",offset);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Change wind parameters", offset, WindPropertiesOverlay );
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Attach to another object", offset, AttachToEntity);
	xNewOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

}

void Tools_Entity_ParticleGenerator_Object::ParticlePropertiesDialog(void)
{

	if(Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_PARTICLEGENERATOR_SUBSIDIARY))
	{
		return;
	}
	
	Asura_Overlay* xOverlay = new Asura_Overlay("Particle Properties:", false);
	Asura_Dialog*  newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(2.0f, 3.0f);
	xOverlay->ID = TOOLS_OVERLAY_PARTICLEGENERATOR_SUBSIDIARY;

	newDialog = new Asura_Dialog("Particle Template properties:",offset);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();


	//Numerical attributes of templates
	newDialog = new Asura_Numerical_Dialog("Initial size", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.InitialSize, 0.0f, 100.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("End size", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.EndSize, 0.0f, 100.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	//Flags......
	newDialog = new Asura_Tickbox_Dialog("Use Dynamics:", offset, &s_pxThisObjectForTesting->m_bUseDynamics, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
#ifdef ASURA_DISABLE_DYNAMICS_WITH_PARTICLE_GENERATORS
    newDialog->SetEnabled( false );
#endif
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Use standard gravity", offset, &s_pxThisObjectForTesting->m_bUseStandardGravity, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Use velocity decay", offset, &s_pxThisObjectForTesting->m_bUseVelocityDecay, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Fade colour", offset, &s_pxThisObjectForTesting->m_bUseColourLerp, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Fade size", offset, &s_pxThisObjectForTesting->m_bUseSizeLerp, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Use Lighting", offset, &s_pxThisObjectForTesting->m_bUseLighting, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Use Motion Blur", offset, &s_pxThisObjectForTesting->m_bUseMotionBlur, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Use random rotation", offset, &s_pxThisObjectForTesting->m_bUseRandomRotation, &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

 	//Colours
	float fColourY = offset.y;
	newDialog = new Asura_Dialog("Start Colour:", offset);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Red", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.InitialColour.Red, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Green", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.InitialColour.Green, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Blue", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.InitialColour.Blue, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Alpha", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.InitialColour.Alpha, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);

	offset.x+=15.0f;
	offset.y=fColourY;

	newDialog = new Asura_Dialog("End Colour:", offset);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Red", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.EndColour.Red, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Green", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.EndColour.Green, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Blue", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.EndColour.Blue, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_ColourNumerical_Dialog("Alpha", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.EndColour.Alpha, "%d", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();
	offset.x-=15.0f;

	//Texture coordinates
	newDialog = new Asura_Dialog("Texture coordinates: (Tab)", offset);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Top Left U", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.TopLeftU, 0.0f, 1.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Top Left V", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.TopLeftV, 0.0f, 1.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Bottom Right U", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.BottomRightU, 0.0f, 1.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Numerical_Dialog("Bottom Right V", offset, &s_pxThisObjectForTesting->m_xParticleTemplate.BottomRightV, 0.0f, 1.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bTemplatePropertiesChanged);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

 	newDialog = new Asura_Function_Dialog( ""/*Choose Translucency mode*/, offset, ChooseTranslucencyMode );
	newDialog->ControlID = DialogID_Translucency;
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Dialog("-----------------------------",offset);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Choose template", offset, ChooseTemplate );
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog( "Save template", offset, SaveTemplate );
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();
	
	Asura_Overlay_System::SubmitNewOverlay(xOverlay);

	UpdateParticlePropertiesDialog();
}


void Tools_Entity_ParticleGenerator_Object::UpdateParticlePropertiesDialog()
{
	Asura_Assert(s_pxThisObjectForTesting,"");

	char szText[200];

	Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(TOOLS_OVERLAY_PARTICLEGENERATOR_SUBSIDIARY);
	if(!pxOverlay) return;

	sprintf(szText,"Translucency : %s",GetTranslucencyDescription(s_pxThisObjectForTesting->m_xParticleTemplate.TranslucencyMode));
	pxOverlay->UpdateDialogLabel(DialogID_Translucency,szText);
}





bool Tools_Entity_ParticleGenerator_Object::HandlesThisMessage(int messageID)
{
	if( messageID == ASURA_MSG_ACTIVATE || messageID == ASURA_MSG_DEACTIVATE)
	{
		return true;
	}
	return false;
}

void Tools_Entity_ParticleGenerator_Object::TestParticleGenerator()
{
	//Create an entity-less particle generator special effect

	//start all selected particle generators
	for(LIF<Tools_Entity_ParticleGenerator_Object*> xGeneratorLIF(&s_xParticleGeneratorList);!xGeneratorLIF.done();xGeneratorLIF.next())
	{
		if(!xGeneratorLIF()->IsGroupSelected()) continue;

		Asura_Assert(xGeneratorLIF()->m_pxClientEntity,"");

		xGeneratorLIF()->m_pxClientEntity->StopParticleGenerator();
		xGeneratorLIF()->m_pxClientEntity->m_bGeneratorActive = true;
	}

}

void Tools_Entity_ParticleGenerator_Object::StopParticleGenerator()
{
	//stop all selected particle generators
	for(LIF<Tools_Entity_ParticleGenerator_Object*> xGeneratorLIF(&s_xParticleGeneratorList);!xGeneratorLIF.done();xGeneratorLIF.next())
	{
		if(!xGeneratorLIF()->IsGroupSelected()) continue;

		Asura_Assert(xGeneratorLIF()->m_pxClientEntity,"");

		xGeneratorLIF()->m_pxClientEntity->m_bGeneratorActive = false;
		xGeneratorLIF()->m_pxClientEntity->StopParticleGenerator();
	}
}

void Tools_Entity_ParticleGenerator_Object::StopAllParticleGenerators()
{
	//stop all particle generators
	for(LIF<Tools_Entity_ParticleGenerator_Object*> xGeneratorLIF(&s_xParticleGeneratorList);!xGeneratorLIF.done();xGeneratorLIF.next())
	{
		Asura_Assert(xGeneratorLIF()->m_pxClientEntity,"");

		xGeneratorLIF()->m_pxClientEntity->m_bGeneratorActive = false;
		xGeneratorLIF()->m_pxClientEntity->StopParticleGenerator();
	}
}


void Tools_Entity_ParticleGenerator_Object::Destroy()
{

	if(m_pxClientEntity)
	{
		m_pxClientEntity->m_bGeneratorActive = false;
		m_pxClientEntity->StopParticleGenerator();
		m_pxClientEntity->SetAsDestroyed();
		m_pxClientEntity = NULL;
	}
	
	Tools_Entity_Object::Destroy();

	if(s_pxThisObjectForTesting == this)
	{
		Tools_Object_System::DeleteObject(s_pxThisObjectForTesting);
	}
}


Asura_Particle_Template Tools_Entity_ParticleGenerator_Object::s_xDummyParticleTemplate;

void Tools_Entity_ParticleGenerator_Object::CreateDummyParticleTemplate(void)
{
	const Asura_Particle_Template* pTemplate = Asura_ParticleGenerator::GetDefaultParticleTemplate();

	s_xDummyParticleTemplate.TopLeftU			= pTemplate->TopLeftU;
	s_xDummyParticleTemplate.TopLeftV			= pTemplate->TopLeftV;
	s_xDummyParticleTemplate.BottomRightU		= pTemplate->BottomRightU;
	s_xDummyParticleTemplate.BottomRightV		= pTemplate->BottomRightV;
	s_xDummyParticleTemplate.DecayRate			= pTemplate->DecayRate;
	s_xDummyParticleTemplate.EndColour			= pTemplate->EndColour;
	s_xDummyParticleTemplate.EndSize			= pTemplate->EndSize;
	s_xDummyParticleTemplate.ID					= pTemplate->ID;
	s_xDummyParticleTemplate.TranslucencyMode   = pTemplate->TranslucencyMode;
	s_xDummyParticleTemplate.InitialColour		= pTemplate->InitialColour;
	s_xDummyParticleTemplate.InitialSize		= pTemplate->InitialSize;
	s_xDummyParticleTemplate.UseColourLerp		= pTemplate->UseColourLerp;
	s_xDummyParticleTemplate.UseDynamics		= pTemplate->UseDynamics;
	s_xDummyParticleTemplate.UseLighting		= pTemplate->UseLighting;
	s_xDummyParticleTemplate.UseMotionBlur		= pTemplate->UseMotionBlur;
	s_xDummyParticleTemplate.UseRandomRotation	= pTemplate->UseRandomRotation;
	s_xDummyParticleTemplate.UseSizeLerp		= pTemplate->UseSizeLerp;
	s_xDummyParticleTemplate.UseStandardGravity	= pTemplate->UseStandardGravity;
	s_xDummyParticleTemplate.UseVelocityDecay	= pTemplate->UseVelocityDecay;

}

void Tools_Entity_ParticleGenerator_Object::CopyParticleGenerator(void)
{
	// Call overloaded Create based on this tools_entity_object
	Create(s_pxThisObjectForTesting);
}

void Tools_Entity_ParticleGenerator_Object::ChooseTranslucencyMode(void)
{
	//Create a modal overlay list
	ListSelection_Overlay xOverlay("Choose translucency mode:");

	//JP - 26/09/02 - Most translucency modes are disabled due to cross 
	//platform portability issues.
	xOverlay.AddItem(GetTranslucencyDescription(ASURA_TRANSLUCENCY_OFF),ASURA_TRANSLUCENCY_OFF);
	xOverlay.AddItem(GetTranslucencyDescription(ASURA_TRANSLUCENCY_NORMAL),ASURA_TRANSLUCENCY_NORMAL);
	//xOverlay.AddItem(GetTranslucencyDescription(ASURA_TRANSLUCENCY_INVCOLOUR),ASURA_TRANSLUCENCY_INVCOLOUR);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_COLOUR"		 , 3);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_DOUBLECOLOUR"	 , 4);
	xOverlay.AddItem(GetTranslucencyDescription(ASURA_TRANSLUCENCY_GLOWING),ASURA_TRANSLUCENCY_GLOWING);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_DARKENINGCOLOUR", 6);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_JUSTSETZ"		 , 7);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_ALPHACOMPARE"	 , 8);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_UGH"			 , 9);
	//xOverlay.AddItem("ASURA_TRANSLUCENCY_NOT_SET"		 ,10);
	xOverlay.AddItem(GetTranslucencyDescription(ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE),ASURA_TRANSLUCENCY_SPECIAL_HEATHAZE);

	if( xOverlay.DoModal() )
	{
		//s_pxThisObjectForTesting->m_iTranslucencyMode = xOverlay.GetSelectionData();
		s_pxThisObjectForTesting->m_xParticleTemplate.TranslucencyMode = (ASURA_TRANSLUCENCY_MODE)xOverlay.GetSelectionData();
		s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating = true;
		UpdateParticlePropertiesDialog();

	}
}


Asura_ParticleGenerator_WindProperties Tools_Entity_ParticleGenerator_Object::s_xDummyWindProperties;

void Tools_Entity_ParticleGenerator_Object::CreateDummyWindProperties()
{
	s_xDummyWindProperties.fMinWindVelocity			= 0.0f;
	s_xDummyWindProperties.fMaxWindVelocity			= 1.0f;
	s_xDummyWindProperties.ucUseRandomDirection		= (u_char)0;
	s_xDummyWindProperties.ucUseRandomVelocity		= (u_char)0;
	s_xDummyWindProperties.ucUseWind				= (u_char)0;
	s_xDummyWindProperties.xInitialWindDirection	= Asura_Vector_3(1.0f, 0.0f, 0.0f);
}

void Tools_Entity_ParticleGenerator_Object::WindPropertiesOverlay(void)
{
	//Function to open a new overlay to allow editing of the wind properties,
	//purely for tidyness purposes
	
	Asura_Overlay* xOverlay = new Asura_Overlay("Wind Properties:", false);
	Asura_Dialog*  newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(2.0f, 3.0f);
	xOverlay->ID = TOOLS_OVERLAY_PARTICLEGENERATOR_SUBSIDIARY;

	newDialog = new Asura_Numerical_Dialog("Minimum wind velocity", offset, &s_pxThisObjectForTesting->m_xWindProperties.fMinWindVelocity, 0.0f, 100.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating);
	xOverlay->AddDialog(newDialog);
	offset.y=newDialog->MaxCoord.y;
	offset.y+=0.2f;

	newDialog = new Asura_Numerical_Dialog("Maximum wind velocity", offset, &s_pxThisObjectForTesting->m_xWindProperties.fMaxWindVelocity, 0.0f, 100.0f, 0.05f, "%.3f", &s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Tickbox_Dialog("Use wind", offset, &s_pxThisObjectForTesting->m_bUseWind, &s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Wandering direction", offset, &s_pxThisObjectForTesting->m_bUseRandomDirection, &s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Tickbox_Dialog("Wandering velocity (else fixed at min)", offset, &s_pxThisObjectForTesting->m_bUseRandomVelocity, &s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Function_Dialog("Edit initial direction vector",offset, EditWindVector);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Finish editing wind vector",offset, SelectThisObject);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	Asura_Overlay_System::SubmitNewOverlay(xOverlay);
}

void Tools_Entity_ParticleGenerator_Object::EditWindVector()
{
	Asura_Assert(s_pxThisObjectForTesting, "");
	//Create a new Tools_Object so we can represent the wind direction vector
	if(!s_pxThisObjectForTesting->m_pxWindVectorObject)
	{
		s_pxThisObjectForTesting->m_pxWindVectorObject = new Tools_ParticleGenerator_Wind_Vector();
	}
		
	Asura_Assert(s_pxThisObjectForTesting->m_pxWindVectorObject, "Null wind vector");
	if(s_pxThisObjectForTesting->m_pxWindVectorObject)
	{
		//Set orientation based on current wind
		Asura_Maths::OrientationFromDirection(s_pxThisObjectForTesting->m_pxWindVectorObject->Orientation, s_pxThisObjectForTesting->m_xWindProperties.xInitialWindDirection);
		//Place vector above the generator itself
		s_pxThisObjectForTesting->m_pxWindVectorObject->Position = s_pxThisObjectForTesting->Position;
		s_pxThisObjectForTesting->m_pxWindVectorObject->Position.y -= 0.5f;
		
		//Select new object.
		Tools_Object_System::SelectObject(s_pxThisObjectForTesting->m_pxWindVectorObject);
	}
}


bool Tools_Entity_ParticleGenerator_Object::s_bDoAttach = false;

void Tools_Entity_ParticleGenerator_Object::AttachToEntity(void)
{
	//Sets a flag so we know we are in attachment editing mode
	s_bDoAttach = true;

	Tools_Object_System::DeselectAllObjects();


	Asura_Overlay* xOverlay = new Asura_Overlay("Select object to attach to:", false);
	Asura_Dialog*  newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(2.0f, 3.0f);
	xOverlay->ID = TOOLS_OVERLAY_PARTICLEGENERATOR_SUBSIDIARY;


	newDialog = new Asura_Function_Dialog("Attach",offset, FinalizeAttachment);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("UnAttach",offset, RemoveAttachment);
	xOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	Asura_Overlay_System::SubmitNewOverlay(xOverlay);

	
}

void Tools_Entity_ParticleGenerator_Object::FinalizeAttachment(void)
{
	//Function will store the guid and position of the selected object
	Tools_Object* pxObjectToAttachTo = Tools_Object_System::SelectedObject;

	if(pxObjectToAttachTo)
	{
		s_pxThisObjectForTesting->m_xTargetOffset = s_pxThisObjectForTesting->Position - pxObjectToAttachTo->Position;
		s_pxThisObjectForTesting->m_uiTargetGuid  = pxObjectToAttachTo->GetGuid();
		//Reset flag and close overlay
		s_bDoAttach = false;
		//Ensure game entities get updated
		s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating = true;
	}
}

void Tools_Entity_ParticleGenerator_Object::RemoveAttachment(void)
{
	s_pxThisObjectForTesting->m_xTargetOffset = Asura_Vector_3(0.0f, 0.0f, 0.0f);
	s_pxThisObjectForTesting->m_uiTargetGuid  = ASURA_GUID_UNREGISTERED;
	s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating = true;
}

void Tools_Entity_ParticleGenerator_Object::Render()
{

	Tools_Entity_Object::Render();

	//Draw links to any attached entities
	if(m_uiTargetGuid != ASURA_GUID_UNREGISTERED)
	{
		Tools_Object* pxTargetEntity = Tools_Object_System::GetObjectFromGuid(m_uiTargetGuid);
		if(pxTargetEntity)
		{
			Asura_Render::RenderLine(m_xPosition, pxTargetEntity->Position, 0xff00ff00);
			//Asura_Render::RenderLine(m_xPosition, Asura_Vector_3(0.0f, 0.0f, 0.0f), 0xff00ff00);
		}
	}

	if(m_pxTextureEdit) m_pxTextureEdit->Render();
}


void Tools_Entity_ParticleGenerator_Object::ChooseTemplate(void)
{
	
// (MH - 12.01.2003) this shouldn't be necessary - the data is already
//	loaded and will be saved if required.
//
////Reload the template store chunk
//	if( !LoadAsuraFile("Misc\\PcleTmps.asr") )
//	{
//		Asura_Assert(0,"Unable to reload particle template store");
//	}
//	else
	{

		//First check to see if there are any templates stored...
		if( Asura_Particle_TemplateStore::s_ppxTemplates )
		{
			ListSelection_Overlay xOverlay("Choose particle template:");

			//Add items to list
			for(int i = 0; i < Asura_Particle_TemplateStore::s_iNumberOfStoredTemplates; i++)
			{
				if(Asura_Particle_TemplateStore::s_ppxTemplates[i])
				{
					xOverlay.AddItem(Asura_Particle_TemplateStore::s_ppxTemplates[i]->szTemplateName, i);
				}
			}

			//Displat dialog box
			if( xOverlay.DoModal() )
			{
				
				#if 0
				//Stop the particlegenerator
				s_pxThisObjectForTesting->m_pxClientEntity->m_bGeneratorActive = false;
				s_pxThisObjectForTesting->m_pxClientEntity->StopParticleGeneratorInstantly();
				#endif

// (MH - 21.11.2002) how is the code below supposed to truncate exactly????
//				//STRCPY the tmepalte name in....
//				char returnedString[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS];
//				
//				//Truncate string to 16 characters...
//				strcpy(returnedString, xOverlay.GetSelectionString());
//				strcpy(s_pxThisObjectForTesting->m_szTemplateName, returnedString);
                // (MH - 21.11.2002) so try this instead
                strncpy(s_pxThisObjectForTesting->m_szTemplateName, xOverlay.GetSelectionString(), 
                    ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1);
                s_pxThisObjectForTesting->m_szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1] = 0;

				//Set a flag telling entity that it will be using a templateStore template
				s_pxThisObjectForTesting->m_bUseTemplateFromStore = true;
				s_pxThisObjectForTesting->m_bUseTemplateStoreStatusHasChanged = true;
				

				//Need to be careful here, we need to overwrite the selected object's
				//particle template pointer with the new one, as chosen from the list.
				//We therefore need to restock the particle template properties with those
				//from the store.

				Asura_Particle_Template* pxParticleTemplate = Asura_Particle_TemplateStore::GetTemplate(s_pxThisObjectForTesting->m_szTemplateName);
				s_pxThisObjectForTesting->UpdateFromParticleTemplate(pxParticleTemplate);

			}
		}
	}
}

#if 0
void Tools_Entity_ParticleGenerator_Object::SaveTemplate(void)
{
	//Add the current template to the template store
	
	//Request name for new template
	GetString_Overlay xOverlay("Enter name for new template (<16 chars)", "");
	if( xOverlay.DoModal() )
	{
		//Add template to templateStore based on name
		if( Asura_Particle_TemplateStore::AddTemplate(xOverlay.GetString(), &s_pxThisObjectForTesting->m_xParticleTemplate) )
		{
			//Save the entire template store to a chunk
			bool bOK = Asura_Chunk_ParticleTemplates::Create();
            if (!bOK)
            {
                Tools_Timed_Message_System::AddMessage(0xff0000, "ERROR: failed to save Misc/PcleTmps.asr. Is it read-only?");
            }
		
			//Ensure the pg loads from the tempalte store in future
			// (MH - 21.11.2002) strcpy(s_pxThisObjectForTesting->m_szTemplateName, xOverlay.GetString() );
            strncpy(s_pxThisObjectForTesting->m_szTemplateName, xOverlay.GetString(), 
                ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1);
            s_pxThisObjectForTesting->m_szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1] = 0;
		
			//Ensure game entities are updated
			s_pxThisObjectForTesting->m_bUseTemplateStoreStatusHasChanged = true;
			s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating = true;
			s_pxThisObjectForTesting->m_bUseTemplateFromStore = true;
		}
	}
}
#else
void Tools_Entity_ParticleGenerator_Object::SaveTemplate(void)
{
	

	Asura_Particle_Template* pxParticleTemplate = 0;
	char szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS];

	if( Asura_Particle_TemplateStore::s_ppxTemplates )
	{
		ListSelection_Overlay xOverlay("Save template as:");

		//Add items to list
		for(int i = 0; i < Asura_Particle_TemplateStore::s_iNumberOfStoredTemplates; i++)
		{
			if(Asura_Particle_TemplateStore::s_ppxTemplates[i])
			{
				xOverlay.AddItem(Asura_Particle_TemplateStore::s_ppxTemplates[i]->szTemplateName, i);
			}
		}

		//add special entry for saving under a new name
		xOverlay.AddItem("*Create New Template*",-1);

		//Display dialog box
		if(!xOverlay.DoModal()) return;

		if(xOverlay.GetSelectionData()>=0)
		{
			//get the template to be updated
			strcpy(szTemplateName,xOverlay.GetSelectionString());
	   		pxParticleTemplate = Asura_Particle_TemplateStore::GetTemplate(szTemplateName);
			Asura_Assert(pxParticleTemplate,"");
		}
	}
	if(!pxParticleTemplate)
	{
		GetString_Overlay xOverlay("Enter name for new template (<16 chars)", "");
		if(!xOverlay.DoModal()) return;

		szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS-1] = 0;
		strncpy(szTemplateName,xOverlay.GetString(),ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS);

		//create a new template
		Asura_Particle_Template xTemp;
		if(!Asura_Particle_TemplateStore::AddTemplate(szTemplateName, &xTemp)) return;

		//get the template so we can update it properly
   		pxParticleTemplate = Asura_Particle_TemplateStore::GetTemplate(szTemplateName);
		Asura_Assert(pxParticleTemplate,"");

	}

	if(!pxParticleTemplate) return;

	//update the existing/newly created template
	*pxParticleTemplate = s_pxThisObjectForTesting->m_xParticleTemplate;

		
	//update the generator to use this template
	strncpy(s_pxThisObjectForTesting->m_szTemplateName, szTemplateName,ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS);

	//Ensure game entities are updated
	s_pxThisObjectForTesting->m_bUseTemplateStoreStatusHasChanged = true;
	s_pxThisObjectForTesting->m_bGameEntitiesNeedUpdating = true;
	s_pxThisObjectForTesting->m_bUseTemplateFromStore = true;


	//Save the entire template store to a chunk
	bool bOK = Asura_Chunk_ParticleTemplates::Create();
	if (!bOK)
	{
	    Tools_Timed_Message_System::AddMessage(0xff0000, "ERROR: failed to save Misc/PcleTmps.asr. Is it read-only?");
	}

}

#endif

void Tools_Entity_ParticleGenerator_Object::ChooseGeneratorTemplate(void)
{
	//First check to see if there are any templates stored...
	if( Asura_Particle_TemplateStore::s_iNumGeneratorTemplates)
	{
		ListSelection_Overlay xOverlay("Choose particle template:");

		//Add items to list
		for(int i = 0; i < Asura_Particle_TemplateStore::s_iNumGeneratorTemplates; i++)
		{
			const Asura_String* pxName = Asura_Particle_TemplateStore::s_xGeneratorTemplateTree.FindNodeNameByIndex(i);
            Asura_Assert(pxName && !pxName->IsNull(), "");
			if (pxName && !pxName->IsNull())
            {
                xOverlay.AddItem(pxName->GetString(), i);
            }
		}

		//Display dialog box
		if( xOverlay.DoModal() )
		{
			
            // other bits copied from ChooseTemplate

            Asura_CompleteParticleGenerator_Template* pxGenTemplate = Asura_Particle_TemplateStore::GetGeneratorTemplate(xOverlay.GetSelectionString());
            Asura_Assert(pxGenTemplate, "");
            if (pxGenTemplate)
            {
                s_pxThisObjectForTesting->UpdateFromGeneratorTemplate(pxGenTemplate->m_xGeneratorTemplate);

				//update from particle template part as well if valid
				if(pxGenTemplate->m_ucParticleTemplateValid)
				{
					if(pxGenTemplate->m_szParticleTemplateName[0])
					{
						//make sure the named particle template exists
						Asura_Particle_Template* pxParticleTemplate = Asura_Particle_TemplateStore::GetTemplate(pxGenTemplate->m_szParticleTemplateName);
						if(pxParticleTemplate)
						{
							//refresh the embedded particle template
							pxGenTemplate->m_xParticleTemplate = *pxParticleTemplate;

			                strncpy(s_pxThisObjectForTesting->m_szTemplateName, pxGenTemplate->m_szParticleTemplateName,ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1);
			                s_pxThisObjectForTesting->m_szTemplateName[ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1] = 0;

							//Set a flag telling entity that it will be using a templateStore template
							s_pxThisObjectForTesting->m_bUseTemplateFromStore = true;
							s_pxThisObjectForTesting->m_bUseTemplateStoreStatusHasChanged = true;


						}
						else
						{
							pxGenTemplate->m_szParticleTemplateName[0] = 0;
						}
				
					}
					s_pxThisObjectForTesting->UpdateFromParticleTemplate(&pxGenTemplate->m_xParticleTemplate);
				}
								
            }
        }
    }
}

void Tools_Entity_ParticleGenerator_Object::SaveGeneratorTemplate(void)
{
    //copy the settings into a template to be saved
    Asura_CompleteParticleGenerator_Template xNewTemplate;
    s_pxThisObjectForTesting->FillOutParticleGeneratorTemplate(xNewTemplate.m_xGeneratorTemplate);
	xNewTemplate.m_ucParticleTemplateValid = true;
	xNewTemplate.m_xParticleTemplate = s_pxThisObjectForTesting->m_xParticleTemplate;
	if(s_pxThisObjectForTesting->m_bUseTemplateFromStore)
	{
        strncpy(xNewTemplate.m_szParticleTemplateName,s_pxThisObjectForTesting->m_szTemplateName,ASURA_PARTICLE_TEMPLATESTORE_MAX_NAME_CHARS - 1);
	}
	else
	{
		xNewTemplate.m_szParticleTemplateName[0] = 0;
	}
	
	
	bool bCreateNewTemplate = true;
	bool bSaveNeeded = false;
	//First check to see if there are any templates stored...
	if( Asura_Particle_TemplateStore::s_iNumGeneratorTemplates)
	{
		ListSelection_Overlay xOverlay("Save template as :");

		//Add items to list
		for(int i = 0; i < Asura_Particle_TemplateStore::s_iNumGeneratorTemplates; i++)
		{
			const Asura_String* pxName = Asura_Particle_TemplateStore::s_xGeneratorTemplateTree.FindNodeNameByIndex(i);
            Asura_Assert(pxName && !pxName->IsNull(), "");
			if (pxName && !pxName->IsNull())
            {
                xOverlay.AddItem(pxName->GetString(), i);
            }
		}

		//add special entry for saving under a new name
		xOverlay.AddItem("*Create New Template*",-1);

		//Display dialog box
		if(!xOverlay.DoModal()) return;

		if(xOverlay.GetSelectionData()>=0)
		{
            //update the existing template
            Asura_CompleteParticleGenerator_Template* pxGenTemplate = Asura_Particle_TemplateStore::GetGeneratorTemplate(xOverlay.GetSelectionString());
            Asura_Assert(pxGenTemplate, "");
			if(pxGenTemplate)
			{
				*pxGenTemplate = xNewTemplate;
				bSaveNeeded = true;
			}
			bCreateNewTemplate = false;
		}
	}
	

	//Request name for new template
	if(bCreateNewTemplate)
	{
		GetString_Overlay xOverlay("Enter name for new generator template", "");
		if( xOverlay.DoModal() )
		{
    	    if (Asura_Particle_TemplateStore::GetGeneratorTemplate(xOverlay.GetString()) != NULL)
    	    {
    	        Tools_Timed_Message_System::AddMessage(0xff0000, "ERROR: Generator template name %s already exists.", xOverlay.GetString());
    	    }
    	    else
    	    {

    	        bool bOK = Asura_Particle_TemplateStore::AddGeneratorTemplate(xOverlay.GetString(), xNewTemplate);
    	        Asura_Assert(bOK, "failed to add template");
    	        if (bOK)
    	        {	
    	        	bSaveNeeded = true;
    	        }
    	    }
    	}
	}

	if(bSaveNeeded)
	{
    	// save
    	bool bOK = Asura_Chunk_ParticleTemplates::Create();
    	if (!bOK)
    	{
    	    Tools_Timed_Message_System::AddMessage(0xff0000, "ERROR: failed to save Misc/PcleTmps.asr. Is it read-only?");
    	}
	}
}

void Tools_Entity_ParticleGenerator_Object::FillOutParticleGeneratorTemplate(Asura_ParticleGenerator_Template& xTemplate)
{
    xTemplate.m_fVelocity = m_fVelocity;
    xTemplate.m_fRate = m_fRate;
    xTemplate.m_fDuration = m_fDuration;
    xTemplate.m_fRandomness = m_fRandomness;
    xTemplate.m_ucInfinite = m_bInfiniteLifetime;
    xTemplate.m_xWindProperties = m_xWindProperties;
    xTemplate.m_ucStaccatoGenerator = m_bStaccatoGenerator;
    xTemplate.m_fLaunchProbability = m_fLaunchProbability;
    xTemplate.m_fGravity = m_fGravity;
}


void Tools_Entity_ParticleGenerator_Object::UpdateFromParticleTemplate(const Asura_Particle_Template* pxParticleTemplate)
{
	//A helper function that will grab values for the particle template
	//from the store. (We have to do this since the tools entity object
	//stores its own template, rather than just a pointer...

	Asura_Assert(pxParticleTemplate,"");

	m_xParticleTemplate.TopLeftU			 = pxParticleTemplate->TopLeftU;
	m_xParticleTemplate.TopLeftV			 = pxParticleTemplate->TopLeftV;
	m_xParticleTemplate.BottomRightU		 = pxParticleTemplate->BottomRightU;
	m_xParticleTemplate.BottomRightV		 = pxParticleTemplate->BottomRightV;
	m_xParticleTemplate.ID				 = pxParticleTemplate->ID;
	m_xParticleTemplate.TranslucencyMode	 = pxParticleTemplate->TranslucencyMode;
	m_fDecayRate							 = pxParticleTemplate->DecayRate;
	m_xParticleTemplate.InitialSize		 = pxParticleTemplate->InitialSize;
	m_xParticleTemplate.EndSize			 = pxParticleTemplate->EndSize;
	m_xParticleTemplate.InitialColour		 = pxParticleTemplate->InitialColour;
	m_xParticleTemplate.EndColour			 = pxParticleTemplate->EndColour;
	m_bUseColourLerp		 = (pxParticleTemplate->UseColourLerp != 0);
	m_bUseDynamics		 = (pxParticleTemplate->UseDynamics != 0);
	m_bUseLighting		 = (pxParticleTemplate->UseLighting != 0);
	m_bUseMotionBlur		 = (pxParticleTemplate->UseMotionBlur != 0);
	m_bUseRandomRotation	 = (pxParticleTemplate->UseRandomRotation != 0);
	m_bUseSizeLerp		 = (pxParticleTemplate->UseSizeLerp != 0);
	m_bUseStandardGravity  = (pxParticleTemplate->UseStandardGravity != 0);
	m_bUseVelocityDecay	 = (pxParticleTemplate->UseVelocityDecay != 0);

	m_bGameEntitiesNeedUpdating = true;
	m_bDecayRateChanged = true;


}

void Tools_Entity_ParticleGenerator_Object::UpdateFromGeneratorTemplate(const Asura_ParticleGenerator_Template& xTemplate)
{
    m_fVelocity = xTemplate.m_fVelocity; 
    m_fRate = xTemplate.m_fRate; 
    m_fDuration = xTemplate.m_fDuration; 
    m_fRandomness = xTemplate.m_fRandomness; 

    m_bInfiniteLifetime = xTemplate.m_ucInfinite != 0;
    m_bUseWind = xTemplate.m_xWindProperties.ucUseWind != 0 ;
    m_xWindProperties = xTemplate.m_xWindProperties;
    m_bStaccatoGenerator = xTemplate.m_ucStaccatoGenerator != 0;
    m_fLaunchProbability = xTemplate.m_fLaunchProbability;
    m_fGravity = xTemplate.m_fGravity;

	m_bGameEntitiesNeedUpdating = true;

}

void Tools_Entity_ParticleGenerator_Object::SelectThisObject(void)
{
	//We may call this function to reselect the pg object, after editing the wind vector for example.
	Tools_Object_System::DeselectAllObjects();
	Tools_Object_System::SelectObject(s_pxThisObjectForTesting);
}

void Tools_Entity_ParticleGenerator_Object::StartUVEdit()
{
	//stop object system mouse contorls intefering
	Tools_Object_System::EnableAllControls(false);
		
	//allocate the uv editing thing
	m_pxTextureEdit = new Tools_UV_Editor;
	int iTextureID = Asura_Texture_Management::AddTexture("\\specialfx\\partclfx.bmp");
	m_pxTextureEdit->m_iMaterialID = Asura_Material_System::GetMaterialIDForTexture(iTextureID,true);

	//initalise the uv coordinates
	m_pxTextureEdit->Nodes[0].x = m_xParticleTemplate.TopLeftU; 
	m_pxTextureEdit->Nodes[0].y = m_xParticleTemplate.TopLeftV; 
	m_pxTextureEdit->Nodes[1].x = m_xParticleTemplate.BottomRightU; 
	m_pxTextureEdit->Nodes[1].y = m_xParticleTemplate.TopLeftV; 
	m_pxTextureEdit->Nodes[2].x = m_xParticleTemplate.BottomRightU; 
	m_pxTextureEdit->Nodes[2].y = m_xParticleTemplate.BottomRightV; 
	m_pxTextureEdit->Nodes[3].x = m_xParticleTemplate.TopLeftU; 
	m_pxTextureEdit->Nodes[3].y = m_xParticleTemplate.BottomRightV; 

	m_pxTextureEdit->CreateMouseModeOverlay();

}							  

void Tools_Entity_ParticleGenerator_Object::FinishUVEdit()
{
	if(m_pxTextureEdit)
	{
		m_pxTextureEdit->CloseMouseModeOverlay();
		delete m_pxTextureEdit;
		m_pxTextureEdit = 0;
		Tools_Object_System::EnableAllControls(true);
	}
}

void Tools_Entity_ParticleGenerator_Object::UpdateUVEdit()
{
	if(!m_pxTextureEdit) return;

	m_pxTextureEdit->HandleMouse();

	if(m_pxTextureEdit->NodesMoved)
	{
		//make sure the nodes stay in rectangular formation
		m_pxTextureEdit->EnsureNodesAreStillOkayForTextureWrap();
		
		m_pxTextureEdit->Nodes[1].x = m_pxTextureEdit->Nodes[2].x;
		m_pxTextureEdit->Nodes[1].y = m_pxTextureEdit->Nodes[0].y;

		m_pxTextureEdit->Nodes[3].x = m_pxTextureEdit->Nodes[0].x;
		m_pxTextureEdit->Nodes[3].y = m_pxTextureEdit->Nodes[2].y;


		m_pxTextureEdit->NodesMoved = false;


		m_xParticleTemplate.TopLeftU = m_pxTextureEdit->Nodes[0].x; 
		m_xParticleTemplate.TopLeftV = m_pxTextureEdit->Nodes[0].y; 
		m_xParticleTemplate.BottomRightU = m_pxTextureEdit->Nodes[2].x; 
		m_xParticleTemplate.BottomRightV = m_pxTextureEdit->Nodes[2].y; 


		m_bTemplatePropertiesChanged = true;
	}
	else
	{
		//make sure the uv coordinates being edited match those on the overlay
		m_pxTextureEdit->Nodes[0].x = m_xParticleTemplate.TopLeftU; 
		m_pxTextureEdit->Nodes[0].y = m_xParticleTemplate.TopLeftV; 
		m_pxTextureEdit->Nodes[1].x = m_xParticleTemplate.BottomRightU; 
		m_pxTextureEdit->Nodes[1].y = m_xParticleTemplate.TopLeftV; 
		m_pxTextureEdit->Nodes[2].x = m_xParticleTemplate.BottomRightU; 
		m_pxTextureEdit->Nodes[2].y = m_xParticleTemplate.BottomRightV; 
		m_pxTextureEdit->Nodes[3].x = m_xParticleTemplate.TopLeftU; 
		m_pxTextureEdit->Nodes[3].y = m_xParticleTemplate.BottomRightV; 
	}
}





//////////////////////////////////////////////////////////////////////////////////////////////

// Tools_ParticleGenerator_Wind_Vector implementation
//
// JP: This is a simple class that just allows us to place a 
// vector in the world to set the wind direction.
//
//

//////////////////////////////////////////////////////////////////////////////////////////////

Tools_ParticleGenerator_Wind_Vector::Tools_ParticleGenerator_Wind_Vector()
{
	//Get a mesh to represent the vector in the world
	Asura_String xMeshName("ToolsShape_CutsceneTrackPoint");
	Mesh = Asura_Mesh_System::GetMesh(xMeshName);
	//Position = *Asura_Camera::GetPosition();
	m_bGameEntitiesNeedUpdating = false;
	m_bReadyToDelete			= false;
}

Tools_ParticleGenerator_Wind_Vector::~Tools_ParticleGenerator_Wind_Vector()
{}

void Tools_ParticleGenerator_Wind_Vector::UpdatePositionInAsuraData(bool bUpdateLinkedObjects)
{

	//Update position and orientation of tools object. Set flag so we know
	//to update the server and client entities

	//If the vector is changed, update the wind vector accordingly
	m_bGameEntitiesNeedUpdating = true;

}

void Tools_ParticleGenerator_Wind_Vector::OnDeselection()
{
	//On deselection this object should delete itself
	m_bReadyToDelete = true;

}
