//////////////////////////////////////////////////////////////////////////////////////
// 
// FILE : Tools_Entity_PhysicalObject.cpp
//
// AUTHOR : Martin G. Harris
//
// Copyright 2002 Rebellion, All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Entity/Tools_Entity_PhysicalObject.h"

// asura
#include "AI/Asura_AI_Navigation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Camera/Asura_Camera.h"
#include "Core/Asura_LOS.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_TextureManagement.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Entity/Asura_Entity_Classification.h"
#include "Entity/Asura_StaticMessage_System.h"
#include "Entity/Asura_Entity_Event.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Mesh/Asura_Env_System.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Network/Asura_Network.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Shapes/Asura_Shape_System.h"
#include "SpecialFX/Asura_Skybox.h"

// axon
#ifdef ASURA_USE_AXON
#include "Core/Axon_Brain.h"
#endif

// tools
#include "Entity/Tools_Entity_Data.h"
#include "Entity/Tools_Message_Editing.h"
#include "Hierarchy/Tools_Animation_Files.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "MeshEdit/Tools_Mode_MeshEdit.h"
#include "Misc/Tools_LOS.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_MessageLog.h"
#include "Objects/Tools_Object_System.h"
#include "Objects/Tools_ObjectSystem_Param.h"
#include "Shapes/Tools_Shapes.h"
#include "Particles/Tools_Mode_PFX.h"

bool	Tools_Entity_PhysicalObject::s_bCreating = false;
bool	Tools_Entity_PhysicalObject::s_bOffsetSkyboxObjectsWhenFlagChanged = false;
Tools_Entity_PhysicalObject* Tools_Entity_PhysicalObject::s_pxCurrentlySelectedObject = 0;
Asura_Hashed_List<Tools_Entity_PhysicalObject> Tools_Entity_PhysicalObject::s_xList;

List<Asura_Guid> Tools_Entity_PhysicalObject::s_xFailedMergeList;


TOOLS_DECLARE_SHARED_PARAM( bool,  uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_MERGEWITHENV,      PhysicalObject_MergeWithEnv      );
TOOLS_DECLARE_SHARED_PARAM( float, uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_MERGESCALEX,       PhysicalObject_MergeScaleX       );
TOOLS_DECLARE_SHARED_PARAM( float, uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_MERGESCALEY,       PhysicalObject_MergeScaleY       );
TOOLS_DECLARE_SHARED_PARAM( float, uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_MERGESCALEZ,       PhysicalObject_MergeScaleZ       );
TOOLS_DECLARE_SHARED_PARAM( bool,  uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_MERGEUNIFORMSCALE, PhysicalObject_MergeUniformScale );
TOOLS_DECLARE_SHARED_PARAM( bool,  uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_DESTRUCTIBLE,      PhysicalObject_Destructible      );
TOOLS_DECLARE_SHARED_PARAM( float, uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_HEALTH,            PhysicalObject_Health            );
TOOLS_DECLARE_SHARED_PARAM( bool,  uTOOLS_SHAREDPARAMHASH_PHYSICALOBJECT_MERGEWITHSKY,      PhysicalObject_MergeWithSky      );

/////////////////////////////////////////////////////////////////////////////////////////
// Tools_Entity_PhysicalObject
/////////////////////////////////////////////////////////////////////////////////////////
Tools_Entity_PhysicalObject::Tools_Entity_PhysicalObject(Asura_Entity* pxEntity)
: Tools_Entity_Object(pxEntity), m_pxPhysicalObject(NULL)
, m_uFileID(ASURA_HASH_ID_UNSET)
, m_uSkinID(ASURA_HASH_ID_UNSET)
, m_uAnimID(ASURA_HASH_ID_UNSET)
, m_uAnimFileID(ASURA_HASH_ID_UNSET)
, m_bNotRenderedInGame(false)
{
	Asura_Assert(pxEntity,"NULL pointer");
	m_xShapeFilename = "NULL";

	m_pxDefaultMesh = Asura_Mesh_System::GetMesh("ToolsShape_CutsceneEvent");
	m_bUsingDefaultMesh = false;

	if (pxEntity)
	{
		m_pxPhysicalObject = Asura_ServerEntity_PhysicalObject::FindEntity(pxEntity->GetGuid());
		Asura_Assert(m_pxPhysicalObject,"Couldn't find PhysicalObject");

		if (m_pxPhysicalObject)
		{
			// Copy position, orientation and shape Id from the server object.
			Position    = m_pxPhysicalObject->GetPosition();
			Orientation = m_pxPhysicalObject->GetOrientation();

			m_bIsDestructible = m_pxPhysicalObject->m_bIsDestructible;
		   	m_uPhysicalObjectFlags = m_pxPhysicalObject->GetPhysicalObjectFlags();

			
			if (m_pxPhysicalObject->GetSkinID() == ASURA_HASH_ID_UNSET)
			{
				Asura_Hash_ID uID = m_pxPhysicalObject->GetShapeFileID();
				if (uID == ASURA_HASH_ID_UNSET &&
					!s_bCreating &&
					pxEntity->GetClassification() == AsuraEntityClass_PhysicsObject)
				{
					char szText[255];
					sprintf(szText, "Object with GUID %d has no animation or shape attached to it. Please fix.", GetGuid());
					Tools_Timed_Message_System::AddMessage(0xff0000, szText);
				}
				SetToolsShape(uID);
			}
			else
			{
			    SetToolsSkinAndAnim(m_pxPhysicalObject->GetSkinFileID(), 
			                        m_pxPhysicalObject->GetSkinID(), 
			                        m_pxPhysicalObject->GetAnimID(),
			                        m_pxPhysicalObject->GetAnimFileID());
			}
          


			if ((m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY) != 0)
			{
				SetMesh(m_pxDefaultMesh);
				m_bUsingDefaultMesh = true;
			}

		}
		SetScale(pxEntity->ToolsData->m_xScale);
		m_bUniformScale = (GetScale().x == GetScale().y) && (GetScale().x == GetScale().z);
		SetUseAnimMovement((m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_USEANIMMOVEMENT)!=0);
	}


	s_xList.AddToList(this);
}

Tools_Entity_PhysicalObject::~Tools_Entity_PhysicalObject()
{
	s_xList.RemoveFromList(this);

}


bool Tools_Entity_PhysicalObject::ShouldRenderShadow()
{
	if(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_NOSHADOW)
	{
		return false;
	}
	return true;
}


void Tools_Entity_PhysicalObject::Destroy()
{
	UnmergePolygonsFromEnvironment();

	if(s_xFailedMergeList.contains(GetGuid()))
	{
		s_xFailedMergeList.delete_entry(GetGuid());
	}

	if(IsMergedWithSkybox())
	{
		Asura_Skybox::InvalidateSkyboxObjects();
	}

	if (m_pxPhysicalObject)
	{
		
		// Make sure client entities get deleted too, send them a destruct message.
		Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_DESTRUCT);
		// (MH - 04.09.2002) this fails in tools where the network is not initialised.
        if (pxMessage)
        {
            pxMessage->To = ASURA_GUID_MY_CLIENT_ENTITIES;
		    pxMessage->From = m_pxPhysicalObject->GetGuid();
		    pxMessage->UseGuaranteedDelivery();
		    pxMessage->SentTime = Asura_Timers::GetLocalScaledTime();
		    Asura_Network::Send(pxMessage);
        }
		else
		{
			// MGH 27/9/02: I haven't tested this, but it probably works...
			// Network not initialised, so try to kill the client directly.
			Asura_Guid uGuid = m_pxPhysicalObject->GetGuid();
			Asura_Entity_Client* pxClientEntity = Asura_Entity_Client::FindEntity(uGuid);
			if (pxClientEntity)
			{
				pxClientEntity->SetAsDestroyed();
			}
		}
	}
	m_pxPhysicalObject = 0;

	Tools_Entity_Object::Destroy();
}

void Tools_Entity_PhysicalObject::RenderPreFSFX()
{
	if(!m_bUsingDefaultMesh)
	{
		Tools_Entity_Object::Render();
	}
}

void Tools_Entity_PhysicalObject::Render()
{
	if(m_bUsingDefaultMesh)
	{
		Tools_Entity_Object::Render();
	}

    if ( IsSelected() )
    {
        Asura_ServerEntity_PhysicalObject* pxGameObject = Asura_ServerEntity_PhysicalObject::FindEntity( GetGuid() );
        if ( pxGameObject )
        {
            for ( u_int u = 0; u < pxGameObject->m_uNumLinksToBlock; ++u )
            {
                Asura_AI_NavVertex* pxWaypoint1 = Asura_AI_Navigation_System::GetWaypoint( pxGameObject->m_pxLinksToBlock[ u ].m_usWaypointIndex );
				if ( !pxWaypoint1 ) { continue; }
				if ( pxGameObject->m_pxLinksToBlock[ u ].m_usConnectionIndex >= pxWaypoint1->m_ucNumConnections ) { continue; }
                Asura_AI_NavVertex* pxWaypoint2 = Asura_AI_Navigation_System::GetWaypoint( pxWaypoint1->m_pxConnections[ pxGameObject->m_pxLinksToBlock[ u ].m_usConnectionIndex ].m_usIndex );
				if ( !pxWaypoint2 ) { continue; }
                Asura_Vector_3 xPos1 = pxWaypoint1->m_xPosition;
                xPos1.y -= 0.5f;
                Asura_Vector_3 xPos2 = pxWaypoint2->m_xPosition;
                xPos2.y -= 0.5f;
                Asura_Render::RenderLine( xPos1, xPos2, 0xffffffff );
            }
        }
    }
}

void Tools_Entity_PhysicalObject::UpdatePositionInAsuraData(bool /*bUpdateLinkedObjects*/)
{
	if (m_pxPhysicalObject)
	{
		Asura_Vector_3 xPosition = Position;
		
		if(xPosition != m_pxPhysicalObject->GetPosition() || Orientation != m_pxPhysicalObject->GetOrientation())
		{
			//object has moved , so need to unweld from the environment
			UnmergePolygonsFromEnvironment();
		}

		// Position has changed update the server side object.
		m_pxPhysicalObject->SetPosition(xPosition);
		m_pxPhysicalObject->SetOrientation(Orientation);

#ifdef ASURA_TOOLS_SYSTEM_UPDATE_PHYSICS
		m_pxPhysicalObject->SleepPhysicsObject();
#endif

		Changed = true; // Make sure we prompt for saving when we exit the mode.

		// Update the tools data too.
		if (m_pxPhysicalObject->ToolsData)
		{
			m_pxPhysicalObject->ToolsData->Position = xPosition;
			m_pxPhysicalObject->ToolsData->m_xOrientation = Orientation;
		}

		//flag skybox for update if needed
		if(IsMergedWithSkybox())
		{
			Asura_Skybox::InvalidateSkyboxObjects();
		}
	}
}

void Tools_Entity_PhysicalObject::PrintInfo(float& x,float& y)
{
	Tools_Entity_Object::PrintInfo(x,y);

	// Display the shape's filename including the path.
	if(GetAnimation())
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Skin : %s", m_xShapeFilename.GetString());
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Anim : %s", m_xAnimFilename.IsBlank() ? m_xAnimName.GetString() : m_xAnimFilename.GetString());
	}
	else
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"File : %s", m_xShapeFilename.GetString());
		if(m_bNotRenderedInGame)
		{
			Asura_Text_Output::Print(x,y++,0xffffffff,0,"(Not rendered in game)");
		}
	}

	if ( m_pxPhysicalObject->HasBlockWaypointLinksFlag() )
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Number Of Links Blocked : %u", m_pxPhysicalObject->GetNumLinksToBlock() );
	}
}

void Tools_Entity_PhysicalObject::OnSelection()
{
    Tools_Entity_Object::OnSelection();
    s_pxCurrentlySelectedObject = this;
}

void Tools_Entity_PhysicalObject::OnDeselection()
{
    Tools_Entity_Object::OnDeselection();
    s_pxCurrentlySelectedObject = NULL;
}

void Tools_Entity_PhysicalObject::OnGroupDeselection()
{
	//(note this gets called on a normal deselection as well as when being deselected from a group)
	if(!IsBeingDestroyed())
	{
		//In theory finished moving this object etc , so now seems a good point to lock it
		//down if needed
		MergePolygonsIntoEnvironmentIfNeeded();	

		RecalculateWaypointLinkBlocks();
	}
}

void Tools_Entity_PhysicalObject::OnUpdateFromTemplate()
{
	//Objects created from templates won't get a chance to be deselected , so do
	//the update here instead.
	//RAR 20/12/2006
 	MergePolygonsIntoEnvironmentIfNeeded();	
  	RecalculateWaypointLinkBlocks();
}

void Tools_Entity_PhysicalObject::RecalculateWaypointLinkBlocks()
{
    if ( m_pxPhysicalObject )
    {
        // Remove the old link blocks.
        m_pxPhysicalObject->UpdateWaypointLinkBlocks( true );
        // Calculate the new link blocks.
        Asura_ServerEntity_PhysicalObject::InitialiseWaypointLinkBlockCalculations();
        m_pxPhysicalObject->CalculateWaypointLinkBlocks();
        Asura_ServerEntity_PhysicalObject::ShutdownWaypointLinkBlockCalculations();
        // Apply the new link blocks.
        m_pxPhysicalObject->UpdateWaypointLinkBlocks();
    }
}


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::HandlesThisMessage().
// PARAMETERS   messageID - 
// RETURNS      
// DESCRIPTION  virtual function - check for messages we can handle that base 
//              class can't, then call the base class version.
// AUTHOR       Mike Healey - 17.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
bool Tools_Entity_PhysicalObject::HandlesThisMessage(int messageID)
{ 
    switch (messageID)
    {
		case ASURA_MSG_APPLY_LINEAR_FORCE: 
			return true;
	
        case ASURA_MSG_CHANGE_ANIMATION:
	    case ASURA_MSG_TEXTUREANIMATION:
            // handled if we have an animation and a skin
            return (GetAnimation() && GetAnimation()->GetSkin());

		case ASURA_MSG_DESTRUCT:
        case ASURA_MSG_SET_DESTRUCTIBLE:
            return true;

		#ifdef ASURA_USE_DIALOGUE_SYSTEM
		case ASURA_MSG_DIALOGUE_PLAYLINEONENTITY:
			return true;
		#endif

		case ASURA_MSG_PLAY_PFX_ON_ENTITY:
			return true;

		case ASURA_MSG_RESTORE_HEALTH:
			return true;

		#ifdef ASURA_USE_AXON
		case ASURA_MSG_AXON_GS2_ANIM_CHANGE:
		case ASURA_MSG_AXON_GS2_ATTRACTOR_CANCEL:
		case ASURA_MSG_AXON_GS2_ATTRACTOR_PULSE:
		case ASURA_MSG_AXON_GS2_DONT_STOP_AT_NODE:
		case ASURA_MSG_AXON_GS2_EXIT_GAMESCENE:
		case ASURA_MSG_AXON_GS2_LEAVE_NODE:
		case ASURA_MSG_AXON_GS2_PATHFIND_TO_NODE:
		case ASURA_MSG_AXON_GS2_TAKE_LINK:
		case ASURA_MSG_AXON_GS2_TELEPORT_TO_NODE:
			return true;
		#endif

        default:
            return Tools_Entity_Object::HandlesThisMessage(messageID);
    }
}


/////////////////////////////////////////////////////////////////////////////////////////////////////////
// CreationHelper: Helper function to add a new physical object in tools.
//
// szTypeName	: menu title in ask for name menu (eg "Pickup Object")
// szShortName	: used in auto generated names (eg "Pickup")
// pfnCreate1	: Fn ptr, creates server object (eg <Project>_ServerEntity_PickupObject) 
// pfnCreate2	: Fn ptr, creates tools object from server object (eg Tools<Project>_Entity_PickupObject)
// szNamePrefix	: optional name prefix
/////////////////////////////////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::CreationHelper(const char* szTypeName,    
												 const char* szShortName, 
												 void (*pfnCreate1)(Asura_Guid),
												 Tools_Object* (pfnCreate2)(Asura_Entity*),
												 const char* szNamePrefix /*= 0*/)
{
	// Get a new GUID
	Asura_Guid uGuid = Asura_Entity_System::GetCreatedByToolsGuid();
	
	// Set up tools data and auto-create a unique name (prompt for new name if there is a clash).
	Tools_Entity_Data* pxToolsData = CreateToolsDataWithDefaultName(szTypeName, szShortName, uGuid, szNamePrefix);
	if (!pxToolsData) return;

	// Create the game specfic server object and set its data.
	pfnCreate1(uGuid);

	Asura_ServerEntity_PhysicalObject* pxServerObject = Asura_ServerEntity_PhysicalObject::FindEntity(uGuid);
	Asura_Assert(pxServerObject,"");
	if (pxServerObject)
	{
		pxServerObject->ToolsData = pxToolsData;
		pxServerObject->SetPosition(pxToolsData->Position);
	}

	// Create the game specific tools object and select it.
	s_bCreating = true;
	Tools_Object* pxToolsObject = pfnCreate2(pxServerObject);
	Tools_Object_System::SelectObject(pxToolsObject);
	s_bCreating = false;
}


///////////////////////////////////////////////////////////////
// SetToolsShape : Set the the tools shape from the hash ID. //
///////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::SetToolsShape(Asura_Hash_ID uShapeID)
{
	// Save the hash ID.
	m_uFileID = uShapeID;
    // clear the anim ones.
    m_uSkinID = ASURA_HASH_ID_UNSET;
    m_uAnimID = ASURA_HASH_ID_UNSET;

	// Clear the tools mesh.
	m_xShapeFilename = "NULL";
	Asura_Mesh* pxMesh = NULL;
	m_bNotRenderedInGame = false;

	// Set up the new tools mesh and filename info.
	if ( ASURA_HASH_ID_UNSET != uShapeID )
	{
		const char* szShapeName		= Tools_ShapeTable::LookupObjectMeshName( uShapeID );
		const char* szShapeFilename = Tools_ShapeTable::LookupObjectFileName( uShapeID );

		if(!szShapeName || ! szShapeFilename)
		{
			char szMessagePart1[200];
			char szMessagePart2[200];
			char szMessagePart3[200];
			
			Asura_PrintfToBfr(szMessagePart1,"Failed to find shape for %s in the shape table.",200,true,GetName());
			Asura_PrintfToBfr(szMessagePart2,"The missing shape has hash ID 0x%x.",200,true,uShapeID);
						
			const char* pszFilename = Asura_HashedFilename_System::GetFilename(uShapeID,false);
			if(pszFilename)
			{
				Asura_String xErrorMessage;
				if(!Tools_ShapeTable::IsValidShapeFile(pszFilename,&xErrorMessage))
				{
					Asura_PrintfToBfr(szMessagePart3,"The shape couldn't be added because :\n%s",200,true,xErrorMessage.GetString());
				}
				else
				{
					Asura_PrintfToBfr(szMessagePart3,"This matches the ID that would be given to %s. Is there another object or environment mesh that clashes with the name used by this shape?",200,true,pszFilename);
				}
			}
			else
			{
				strcpy(szMessagePart3,"This doesn't match the ID of any shape file that could be found.");
			}
			
			
			Asura_Assertf(0,("%s\n%s\n%s",szMessagePart1,szMessagePart2,szMessagePart3));
		}

		if (szShapeName)
		{
			// Set the tools mesh for rendering here also.
			Asura_String xShapeName = szShapeName;
			pxMesh = Asura_Mesh_System::GetMesh(xShapeName);

			Asura_Assertf(Mesh, ("Mesh not found for <%s>", szShapeName) );

			if( !pxMesh &&
				Asura_Mesh_System::GetNumberOfMeshes() > 0 )
			{
				pxMesh = Asura_Mesh_System::GetMesh( Asura_Mesh_System::GetNumberOfMeshes() - 1 );
			}
		}

		if (pxMesh && szShapeFilename)
		{
			m_xShapeFilename = szShapeFilename;
		}

		Asura_Shape* pxShape = Asura_Shape_System::GetShape(uShapeID);
		if(pxShape && pxShape->IsNeverRendered())
		{
			m_bNotRenderedInGame = true;
		}

	}

	if (pxMesh && !IsMergedWithSkybox())
	{
		m_bUsingDefaultMesh = false;
	}
	else
	{	  
		// Set up some default object
		pxMesh = m_pxDefaultMesh;
		m_bUsingDefaultMesh = true;
	}

	SetMesh(pxMesh);
}


void Tools_Entity_PhysicalObject::UpdateWhileSelected(void)
{

	Tools_Entity_Object::UpdateWhileSelected();
	m_pxPhysicalObject->m_bIsDestructible = m_bIsDestructible;
	m_pxPhysicalObject->SetPhysicalObjectFlags(m_uPhysicalObjectFlags);
	
	if(Entity->ToolsData->m_xScale != GetScale())
	{
		SetScale(Entity->ToolsData->m_xScale);
		UnmergePolygonsFromEnvironment();
	}

	if(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV && !(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV))
	{
		UnmergePolygonsFromEnvironment();
	}
}



void Tools_Entity_PhysicalObject::SetToolsSkinAndAnim(Asura_Hash_ID uFileID, 
                                                      Asura_Hash_ID uSkinID, 
                                                      Asura_Hash_ID uAnimID,
                                                      Asura_Hash_ID uAnimFileID)
{
    //Asura_Assert(uFileID != ASURA_HASH_ID_UNSET, "invalid file id");
    Asura_Assertf( (uSkinID != ASURA_HASH_ID_UNSET), ("invalid skin id for entity '%s' (Guid %d)", GetName(), GetGuid()) );
    //Asura_Assert(uAnimID != ASURA_HASH_ID_UNSET, "invalid anim id");
    //Asura_Assert(uAnimFileID != ASURA_HASH_ID_UNSET, "invalid anim file id");

    // save the IDs
	m_uFileID = uFileID;
    m_uSkinID = uSkinID;
    m_uAnimID = uAnimID;
    m_uAnimFileID = uAnimFileID;

	Asura_Hierarchy_Skin* pxSkin = Tools_Animation_Files::GetSkinFromID(uSkinID,uFileID);
	Asura_Hierarchy_Anim* pxAnim = Tools_Animation_Files::GetAnimFromID(uAnimID,uAnimFileID);

    if (pxSkin)
    {
		if(!IsMergedWithSkybox())
		{
			SetSkinAndAnim(pxSkin,pxAnim);
			m_bUsingDefaultMesh = false;
		}
    }
    else
    {
		Asura_Assertf( (uSkinID == ASURA_HASH_ID_UNSET), ("Could not find tools skin (0x%08x) for Tools_Entity_PhysicalObject '%s' (%u), please check", uSkinID, GetName(), GetGuid()));
        SetToolsShape(ASURA_HASH_ID_UNSET);
    }

	m_xShapeFilename = pxSkin ? pxSkin->Name : "<Unknown>";
	if(uFileID)
	{
		//get the full path if available ,and use that instead
		const char* szSkinFilename = Asura_HashedFilename_System::GetFilename(uFileID,false);
		if(szSkinFilename)
		{
			m_xShapeFilename = szSkinFilename;
		}
	}

	if(m_uAnimID)
		m_xAnimName = pxAnim ? pxAnim->Name : "<Unknown>";
	else
		m_xAnimName = "None";

	//get the full path of the animation of available
	m_xAnimFilename.MakeNull();
	if(uAnimFileID)
	{
		const char* szAnimFilename = Asura_HashedFilename_System::GetFilename(uAnimFileID,false);
		if(szAnimFilename)
		{
			m_xAnimFilename = szAnimFilename;
		}
	}

	m_bNotRenderedInGame = false;

	// JAT 13/03/07 - if failed to find the anim and it wasn't a NULL anim then give a meaningful asset for people to fix
	Asura_Assertf( !( ( pxAnim == NULL ) && ( uAnimFileID != ASURA_HASH_ID_UNSET ) ), 
				  ( "Failed to find anim 0x%X for skin '%s' for object '%s'", uAnimFileID, m_xShapeFilename.GetString(), GetName() ) );
}

void Tools_Entity_PhysicalObject::LoadSkinsAndAnims()
{
	//check each physical object , and try getting the skin and animation , to force
	//them to be loaded if they haven't been already.
	for(Asura_Hashed_List_ConstIt<Asura_ServerEntity_PhysicalObject> xIt(Asura_ServerEntity_PhysicalObject::GetList());!xIt.Done();xIt.Next())
	{
		const Asura_ServerEntity_PhysicalObject* pxObject = xIt.GetCurrent();

		if(pxObject->GetAnimID())
		{
			Tools_Animation_Files::GetSkinFromID(pxObject->GetSkinID(),pxObject->GetSkinFileID());
			Tools_Animation_Files::GetAnimFromID(pxObject->GetAnimID(),pxObject->GetAnimFileID());
		}
	}
}


void Tools_Entity_PhysicalObject::AddFilesToExportList(List<Asura_String>& xNameList) const
{
    if (m_pxPhysicalObject && m_pxPhysicalObject->HasSkinOrHierarchy())
    {
        		
		if(m_uSkinID != ASURA_HASH_ID_UNSET)
		{
			//If there are platform specific versions of the file , then do the lookup based on the skin ID to make
			//sure the appropriate one for the platform is found.
			//It might be better to always ignore the file ID , but there could be problems with existing projects
			//if they have a file appearing in more than one place in the directories being searched.
			//RAR 25/04/2007
			Asura_Hash_ID uFileID = m_uFileID;
			if(!uFileID || Asura_HashedFilename_System::DoesStrippedFilenameHavePlatformVersions(m_uSkinID))
			{
				uFileID = m_uSkinID;
			}
			
			const char* szSkinFileName = Asura_HashedFilename_System::GetFullFilename(uFileID,false);
			Asura_Assertf(szSkinFileName, ("Failed to determine skin filename from hash ID for '%s', guid: %u",Entity->ToolsData->Name.GetString(), Entity->GetGuid()));
			if (szSkinFileName)
			{
				Tools_Level_Export::AddFileToExportList(xNameList, Asura_String(szSkinFileName), true);
			}
		}
		if(m_uAnimID != ASURA_HASH_ID_UNSET)
		{
			//ditto for anims. RAR 25/04/2007
			Asura_Hash_ID uFileID = m_uAnimFileID;
			if(!uFileID || Asura_HashedFilename_System::DoesStrippedFilenameHavePlatformVersions(m_uAnimID))
			{
				uFileID = m_uAnimID;
			}

			const char* szAnimFileName = Asura_HashedFilename_System::GetFullFilename(uFileID,false);
			Asura_Assertf(szAnimFileName, ("Failed to determine anim filename from hash ID for '%s'",Entity->ToolsData->Name.GetString()));
			if (szAnimFileName)
			{
				Tools_Level_Export::AddFileToExportList(xNameList, Asura_String(szAnimFileName), false);
			}
		}		
    }
}

void Tools_Entity_PhysicalObject::AddFilesToExportListFromMessage(u_short usMessageID, 
                                                                  void* pData, 
                                                                  List<Asura_String>& xNameList) const
{
    Asura_Assert(m_pxPhysicalObject, "");
    if (!m_pxPhysicalObject) return;

    if (usMessageID == ASURA_MSG_CHANGE_ANIMATION)
    {
        Asura_Assert(m_pxPhysicalObject->HasSkinOrHierarchy(), "this entity should not be receiving this message");
        const Asura_Message_PhysicalObject_ChangeAnimation* pxChangeAnimData = reinterpret_cast
            <Asura_Message_PhysicalObject_ChangeAnimation*>(pData);
//        const Asura_Message_PhysicalObject_ChangeAnimation* pxChangeAnimData = 
//            (const Asura_Message_PhysicalObject_ChangeAnimation*)pData;
        const char* szFileName = Asura_HashedFilename_System::GetFilename(pxChangeAnimData->m_uAnimFileID,false);
        if (szFileName)
        {
            Tools_Level_Export::AddFileToExportList(xNameList, Asura_String(szFileName), false);
        }
		else if(pxChangeAnimData->m_uAnimFileID != ASURA_HASH_ID_UNSET)
		{
			Tools_Level_Export::AddMessage(Tools_Level_Export::MT_ERROR,"Failed to determine animation for anim change message to %s",Entity->ToolsData->Name.GetString());
		}
    }
	else if (usMessageID == ASURA_MSG_PLAY_PFX_ON_ENTITY)
	{
		const Asura_Msg_PlayPFXOnEntity* pxData = static_cast<const Asura_Msg_PlayPFXOnEntity*>(pData);
		if (pxData && pxData->m_uPFXEffectHash != ASURA_HASH_ID_UNSET)
		{
			Tools_Mode_PFX::AddTemplateToExportListByHash(&xNameList, pxData->m_uPFXEffectHash, GetGuid());
		}
	}
}

bool Tools_Entity_PhysicalObject::ShouldAddFilesToExportList() const
{
	return Tools_Entity_Object::ShouldAddFilesToExportList();
}

void Tools_Entity_PhysicalObject::AddBlockLinksOverlayControl(Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset)
{
    Asura_Dialog* pxNewDialogue = new Asura_FlagTickbox_Dialog(
        "Mark waypoint links as blocked",
        xOffset,
        &m_uPhysicalObjectFlags,
        ASURA_PHYSICALOBJECT_FLAG_BLOCKWAYPOINTLINKS,
        &Changed );
    pxNewDialogue->SetHelpText( "Whether the object should mark any waypoint links going through it as blocked when it has collisions turned on." );
    pxNewOverlay->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

    // (TI - 01.03.2007) - It makse sense that whenever the blocks links tickbox is shown 
	// the block expansion should also be shown
	pxNewDialogue = new Asura_FlagTickbox_Dialog(
		"Block nav data generation",
		xOffset,
		&m_uPhysicalObjectFlags,
		ASURA_PHYSICALOBJECT_FLAG_BLOCKNAVDATAGENERATION,
		&Changed );
	pxNewDialogue->SetHelpText( "Whether the object should block waypoint and link generation, regardless of the 'No collisions' flag." );
	pxNewOverlay->AddDialog( pxNewDialogue );
	xOffset.y += Asura_Dialog::DefaultSpacing();
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::AddHealthOverlayControl().
// PARAMETERS   pxNewOverlay - 
//              xOffset      - 
// RETURNS      None.
// DESCRIPTION  Adds overlay dialog to edit the health of the object.
// AUTHOR       Mike Healey - 26.11.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::AddHealthOverlayControl(Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset, bool bAddDestructibleFlag /* = true */)
{
	m_bIsDestructible = m_pxPhysicalObject->m_bIsDestructible;

	if( bAddDestructibleFlag )
	{
        AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_Destructible );
        xOffset.y += Asura_Dialog::DefaultSpacing();
	}
    
    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_Health );
    xOffset.y += Asura_Dialog::DefaultSpacing();
}
// End function Tools_Entity_PhysicalObject::AddHealthOverlayControl().



void Tools_Entity_PhysicalObject::AddMergeWithEnvOverlayControl(Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset)
{
    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_MergeWithEnv );
	xOffset.y += Asura_Dialog::DefaultSpacing();

    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_MergeUniformScale );
	xOffset.y += Asura_Dialog::DefaultSpacing();

    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_MergeScaleX );
	xOffset.y += Asura_Dialog::DefaultSpacing();

    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_MergeScaleY );
	xOffset.y += Asura_Dialog::DefaultSpacing();

    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_MergeScaleZ );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	#if 0 //This option may not be needed after all.... leaving disabled for now. RAR 11/05/04
	xOffset.x = pxNewDialog->MaxCoord.x+1.0f;
	pxNewDialog = new Asura_Function_Dialog("Force Update",xOffset,ForceUpdateOfAllMergedObject);
	pxNewDialog->SetHelpText("Forces the merging of all objects with the 'Merge With Env' flag to be recalculate.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.x = 1.0f;
	#endif

}

void Tools_Entity_PhysicalObject::AddMergeWithSkyOverlayControl(Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset)
{
	Asura_Dialog* pxNewDialog;

    AddSharedParamDialogToOverlay( pxNewOverlay, xOffset, g_uHash_PhysicalObject_MergeWithSky );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxNewDialog = new Asura_Tickbox_Dialog("Offset When Sky-Merged", xOffset, &s_bOffsetSkyboxObjectsWhenFlagChanged);
	pxNewDialog->SetHelpText("This will cause the camera position to be added/subtracted from the object positions whenever the skybox flag is changed.\n"
							"This will mean that the object will appear to stay in the same position , and so may be convenient when creating new skybox objects.\n"
							"If not set then the position won't be altered ,which will mean that it will get drawn in a different place when the skybox flag is changed. This is probably better when setting up templates with skybox objects.\n"
							"This option will affect all skybox objects which get changed while in this mode.");
	pxNewOverlay->AddDialog(pxNewDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

}



bool Tools_Entity_PhysicalObject::HandlesThisEvent(int iEventID) const
{
	if( iEventID == ASURA_ENTITY_EVENT_ONDEATH || iEventID == ASURA_ENTITY_EVENT_ONDESTRUCTION ) 
	{
		return true;
	}
	else if( (ASURA_ENTITY_EVENT_ONHEALTHPERCENT_0 <= iEventID) && (iEventID <= ASURA_ENTITY_EVENT_ONHEALTHPERCENT_99) ) 
	{
		return true;
	}

	return Tools_Entity_Object::HandlesThisEvent( iEventID );
}



void Tools_Entity_PhysicalObject::SetPhysicalObjectFlags(u_int uFlags)
{
	m_uPhysicalObjectFlags = uFlags;
	m_pxPhysicalObject->SetPhysicalObjectFlags(uFlags);
}

void Tools_Entity_PhysicalObject::Copy(const Tools_Object* pxObject)
{
	Tools_Entity_Object::Copy(pxObject);

	UnmergePolygonsFromEnvironment();

	const Tools_Entity_PhysicalObject* pxFromObject = s_xList.FindEntity(pxObject->GetGuid());
	if(!pxFromObject) return;

	//get the entities to copy from and to
	const Asura_ServerEntity_PhysicalObject* pxFromEntity = pxFromObject->m_pxPhysicalObject;
	Asura_ServerEntity_PhysicalObject* pxToEntity = m_pxPhysicalObject;

	//copy everything between the entities
	pxToEntity->SetDestructible(pxFromEntity->IsDestructible());
	pxToEntity->SetHealth(pxFromEntity->GetHealth());
	pxToEntity->SetMaxHealth(pxFromEntity->GetMaxHealth());

	if(pxFromEntity->GetSkinID() != ASURA_HASH_ID_UNSET)
	{
		pxToEntity->SetSkinAndAnimID(pxFromEntity->GetSkinID(),pxFromEntity->GetAnimID());
		pxToEntity->SetSkinAndAnimFileID(pxFromEntity->GetSkinFileID(), pxFromEntity->GetAnimFileID());
	}
	else
	{
		pxToEntity->SetShapeFileID(pxFromEntity->GetShapeFileID());
	}

	//Copy physical object flags.
	//Reset the merged with environment flag
	u_int uPhysicalFlags = pxFromEntity->GetPhysicalObjectFlags() &~ ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
	//Preserve the static object flag , since this should only change through SetupPhysicsObject();
	if(pxToEntity->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_STATIC)
		uPhysicalFlags |= ASURA_PHYSICALOBJECT_FLAG_STATIC;
	else
		uPhysicalFlags &=~ (ASURA_PHYSICALOBJECT_FLAG_STATIC|ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV|ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY);
	pxToEntity->SetPhysicalObjectFlags(uPhysicalFlags);

	//scaling can only be copied onto static objects
	if(m_uPhysicalObjectFlags  & ASURA_PHYSICALOBJECT_FLAG_STATIC)
	{
		pxToEntity->ToolsData->m_xScale = pxFromEntity->ToolsData->m_xScale;
	}

	#ifdef ASURA_USE_AXON
	pxToEntity->ShutdownAndDestroyBrain();
	if ( pxFromEntity->GetBrain() )
	{
		pxToEntity->CreateAndInitialiseBrain();
		if ( pxToEntity->GetBrain() )
		{
			pxToEntity->GetBrain()->Copy( *pxFromEntity->GetBrain() );
		}
	}
	#endif

	//and copy between the tools objects
	m_bIsDestructible = pxFromObject->m_bIsDestructible;
	m_uPhysicalObjectFlags = uPhysicalFlags;
	if(pxFromObject->m_uSkinID)
	{
		SetToolsSkinAndAnim(pxFromObject->m_uFileID,pxFromObject->m_uSkinID,pxFromObject->m_uAnimID,pxFromObject->m_uAnimFileID);
	}
	else
	{
		SetToolsShape(pxFromObject->m_uFileID);
	}
	//scaling can only be copied onto static objects
	if(m_uPhysicalObjectFlags  & ASURA_PHYSICALOBJECT_FLAG_STATIC)
	{
		SetScale(Entity->ToolsData->m_xScale);
		m_bUniformScale = pxFromObject->m_bUniformScale;
	}
	SetUseAnimMovement((m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_USEANIMMOVEMENT)!=0);
}


void Tools_Entity_PhysicalObject::UnmergePolygonsFromEnvironment()
{
	if(!m_pxPhysicalObject) return;

	if(!(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV)) return;
	
	Tools_Mode_MeshEdit::DeleteEntityPolygons(GetGuid());
	m_uPhysicalObjectFlags &=~ ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
	m_pxPhysicalObject->SetPhysicalObjectFlags(m_uPhysicalObjectFlags);

	SetEntityModified();

}


void Tools_Entity_PhysicalObject::MergePolygonsIntoEnvironmentIfNeeded()
{
	if(!m_pxPhysicalObject) return;

	//make sure not aready merged
	if(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV) return;

	//make sure we want to be merged
	if(!(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV)) return;

	//must have a skin or shape
	if(!m_pxPhysicalObject->HasShapeOrSkin()) return;

	Asura_Hierarchy_Skin* pxSkin = 0;
	if(m_pxPhysicalObject->GetSkinID())
	{
		//make sure the skin can be found.
		pxSkin = Tools_Animation_Files::GetSkinFromID(m_pxPhysicalObject->GetSkinID(), m_pxPhysicalObject->GetSkinFileID());
		if(!pxSkin) return;
	}
	else
	{
		if(!Mesh) return;
		//don't bother for tools shapes
		if(!strnicmp(Mesh->Name.GetString(),"ToolsShape_",11)) return;
	}


	//Determine module to merge to
	Asura_Bounding_Box xObjectBB;
	GetBoundingBox(xObjectBB);
	Asura_Vector_3 xCentre = Position;
	xCentre += xObjectBB.CalculateCentre()*Orientation;

	int iModule = Tools_Mode_MeshEdit::GetModuleForObjectMerge(Position,xCentre);

	Asura_Guid uGuid = GetGuid();
	if(iModule < 0)
	{
		if(!s_xFailedMergeList.contains(uGuid))
		{
			s_xFailedMergeList.add_entry(uGuid);
			//Warn about failed merge

			static u_int s_uSkipErrorFrameStamp = 0xffffffff;
			if(s_uSkipErrorFrameStamp != Asura_Timers::GetFrameNumber())
			{
				bool bSkip = !ShowYesNoOverlay("Failed to find a suitable module to merge object to.\nName : %s\nGuid : %d\n\nPress 'Yes' to continue.\nPress 'No' to skip this batch of error messages.\n(A list of these objects can be found in the message log)",GetName(),uGuid);
				if(bSkip)
				{
					//Skip any remaining errors this frame.
					//May need an alternative plan if anything else causes message overlays to be shown in
					//his batch , since they will up the frame number.
					//RAR 04/04/2006
					s_uSkipErrorFrameStamp = Asura_Timers::GetFrameNumber();
				}
			}

			Tools_MessageLog::AddMessage("Entity","Failed Merge",new Tools_MessageLogAction_GotoEntity(uGuid),"Failed to find module to merge '%s' to",GetName());

		}
		return;
	}
	else
	{
		if(s_xFailedMergeList.contains(uGuid))
		{
			s_xFailedMergeList.delete_entry(uGuid);
		}
	}

	//check for flags that should be converted to polygon flags
	u_int uPolygonFlagsToAdd = 0;
	if(m_pxPhysicalObject->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_NOCOLLISIONS)
	{
		uPolygonFlagsToAdd |= AsuraPolygonFlag_NoCollisions;
	}
	if(m_pxPhysicalObject->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_INVISIBLE)
	{
		uPolygonFlagsToAdd |= AsuraPolygonFlag_DontRender|AsuraPolygonFlag_IgnoreLOS;
	}
	if(m_pxPhysicalObject->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_IGNOREFORLOS)
	{
		uPolygonFlagsToAdd |= AsuraPolygonFlag_IgnoreLOS;
	}

	//do the merge
	bool bMerged = false;
	if(pxSkin)
	{
		//ignoring scale for now , since there isn't currently anyway of seeing the scale in the pre-merged state
		Asura_Vector_3 xScale(1.0f,1.0f,1.0f);

		bMerged = Tools_Mode_MeshEdit::MergeSkinIntoModule(iModule,pxSkin,Orientation,Position,GetGuid(),GetChecksumForMerge(),xScale,uPolygonFlagsToAdd,0);
	}
	else
	{
		bMerged = Tools_Mode_MeshEdit::MergeObjectIntoModule(iModule,Mesh,Orientation,Position,GetGuid(),GetChecksumForMerge(),GetScale(),uPolygonFlagsToAdd,0);
	}


	if(bMerged)
	{
		//and update the flags
		m_uPhysicalObjectFlags |= ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
		m_pxPhysicalObject->SetPhysicalObjectFlags(m_uPhysicalObjectFlags);

		SetEntityModified();
	}
	else
	{
		if(!s_xFailedMergeList.contains(uGuid))
		{
			s_xFailedMergeList.add_entry(uGuid);
			//Warn about failed merge

			static u_int s_uSkipErrorFrameStamp = 0xffffffff;
			if(s_uSkipErrorFrameStamp != Asura_Timers::GetFrameNumber())
			{
				bool bSkip = !ShowYesNoOverlay("Failed to merge object into module , since it would result in the module having too many triangles or vertices.\nName : %s\nGuid : %d\n\nPress 'Yes' to continue.\nPress 'No' to skip this batch of error messages.\n(A list of these objects can be found in the message log)",GetName(),uGuid);
				if(bSkip)
				{
					//Skip any remaining errors this frame.
					//May need an alternative plan if anything else causes message overlays to be shown in
					//his batch , since they will up the frame number.
					//RAR 04/04/2006
					s_uSkipErrorFrameStamp = Asura_Timers::GetFrameNumber();
				}
			}

			Tools_MessageLog::AddMessage("Entity","Failed Merge",new Tools_MessageLogAction_GotoEntity(uGuid),"Error merging '%s' to module",GetName());
		}
	}

}

void Tools_Entity_PhysicalObject::MergeObjectsIntoEnvironmentIfNeeded(bool bForceRecalculateAll)
{
	if(bForceRecalculateAll)
	{
		Tools_Mode_MeshEdit::DeleteAllEntityPolygons();
	}

	//check each object to see if it needs to be merged into the environment
	for(Asura_Hashed_List_It<Tools_Entity_PhysicalObject> xIt(&s_xList);!xIt.Done();xIt.Next())
	{
		Tools_Entity_PhysicalObject* pxObject = xIt.GetCurrent();

		if(bForceRecalculateAll)
		{
			pxObject->m_uPhysicalObjectFlags &=~ ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
			pxObject->m_pxPhysicalObject->SetPhysicalObjectFlags(pxObject->m_uPhysicalObjectFlags);
		}
		pxObject->MergePolygonsIntoEnvironmentIfNeeded();
	}

}

void Tools_Entity_PhysicalObject::ForceUpdateOfAllMergedObject()
{
	//Confirm whether this should be done
	if(ShowYesNoOverlay("Recalculate the merging for all objects with the environment?\nThis should fix any problems with the files being out of sync , but will result in a lot of lighting etc needing to be recalculated."))
	{
		MergeObjectsIntoEnvironmentIfNeeded(true);
	}
}


void Tools_Entity_PhysicalObject::CheckIsMergedWithEnvironmentFlags()
{
	//get a list of all entity guids attached to the environment
	List<Asura_Mesh_MergedEntity> xMergeList;
	LIF<Asura_Mesh_MergedEntity> xMergeLIF(&xMergeList);

	for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);

		bool bResetMesh = false;
		if(pxMesh->m_iNumberOfMergedEntities && !pxMesh->m_pxMergedEntityCollisionMesh)
		{
			//Need to reattach all objects again for meshes that don't have a merged
			//entity collision mesh yet.		
			bResetMesh = true;
		}

		for(int j=0;j<pxMesh->m_iNumberOfMergedEntities;j++)
		{
			Asura_Mesh_MergedEntity xMerge = pxMesh->m_pxMergedEntities[j];
			
			//Only checking merge entries that have guids in tools range here , since those outside that
			//range will be checked elsewhere (such as botanicals). RAR 17/03/2006
			if(xMerge.m_uGuid >= ASURA_GUID_TOOLSCREATEDENTITY_START && xMerge.m_uGuid <= ASURA_GUID_TOOLSCREATEDENTITY_END)
			{
				if(bResetMesh) xMerge.m_uChecksum = 0;//change the checksum so that mesh gets reattached
				xMergeList.add_entry(xMerge);
			}
		}
	}


	//check each entity to see if the IsMergedWithEnv flag corresponds with the guid list
	for(Asura_Hashed_List_It<Tools_Entity_PhysicalObject> xIt(&s_xList);!xIt.Done();xIt.Next())
	{
		Tools_Entity_PhysicalObject* pxObject = xIt.GetCurrent();

		//search for guid in list
		bool bIsMerged = false;
		for(xMergeLIF.restart();!xMergeLIF.done();xMergeLIF.next())
		{
			if(xMergeLIF().m_uGuid == pxObject->GetGuid())
			{
				bIsMerged =  true;
				break;
			}
		}
	

		if(bIsMerged)
		{
			
			if(pxObject->m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV)
			{
				pxObject->m_uPhysicalObjectFlags |= ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
				pxObject->m_pxPhysicalObject->SetPhysicalObjectFlags(pxObject->m_uPhysicalObjectFlags);
				if(xMergeLIF().m_uChecksum != pxObject->GetChecksumForMerge())
				{
					//object out of date... unmerge the polygons , and remerge below
					pxObject->UnmergePolygonsFromEnvironment();
				}
			}
			else
			{
				//object is merged with environment , but shouldn't be... remove the polygons
				pxObject->m_uPhysicalObjectFlags &=~ ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
				pxObject->m_pxPhysicalObject->SetPhysicalObjectFlags(pxObject->m_uPhysicalObjectFlags);
				Tools_Mode_MeshEdit::DeleteEntityPolygons(pxObject->GetGuid());
				
			}

			xMergeLIF.delete_current();
		}
		else
		{
			pxObject->m_uPhysicalObjectFlags &=~ ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV;
			pxObject->m_pxPhysicalObject->SetPhysicalObjectFlags(pxObject->m_uPhysicalObjectFlags);
		}
		pxObject->MergePolygonsIntoEnvironmentIfNeeded();
	}

	//Delete the polygons for any guids left in the list , since their entites no longer exist
	while(xMergeList.size())
	{
		Tools_Mode_MeshEdit::DeleteEntityPolygons(xMergeList.first_entry().m_uGuid);
		xMergeList.delete_first_entry();
	}

}

bool Tools_Entity_PhysicalObject::ShouldRenderSelectionOutlineOnly()
{
	return (m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV)!=0;
}

u_int Tools_Entity_PhysicalObject::GetChecksumForMerge() const
{
	if(!m_pxPhysicalObject) return 0;


	u_int uChecksum = 0;

	if(m_pxPhysicalObject->GetSkinID())
	{
		Asura_Hierarchy_Skin* pxSkin = Tools_Animation_Files::GetSkinFromID(m_pxPhysicalObject->GetSkinID(), m_pxPhysicalObject->GetSkinFileID());
		if(!pxSkin) return 0;

		//get the value from the skin's modified time
		uChecksum = pxSkin->m_xModifiedTime.dwLowDateTime;
	}
	else
	{
		Asura_Hash_ID uShapeID = m_pxPhysicalObject->GetShapeFileID();
		if(uShapeID == ASURA_HASH_ID_UNSET) return 0;
	
		//get the value for the shape (based on modification time)
		uChecksum = Tools_ShapeTable::GetChecksumForShape(uShapeID);
	}

	uChecksum += m_pxPhysicalObject->GetPositionAndOrientChecksum();

	//adjust checksum for polygon flags that are to be transferred
	if(m_pxPhysicalObject->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_NOCOLLISIONS)
	{
		uChecksum += 0x221A0779;//taken from flag hash value , but doesn't make any difference really
	}
	if(m_pxPhysicalObject->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_INVISIBLE)
	{
		uChecksum += 0x5EA7F396;
	}
	if(m_pxPhysicalObject->GetPhysicalObjectFlags() & ASURA_PHYSICALOBJECT_FLAG_IGNOREFORLOS)
	{
		uChecksum += 0xCB4CA5B0;
	}
		
	return uChecksum;
}


Asura_Handle Tools_Entity_PhysicalObject::GetEntityPhysicsObjectHandle() const
{
	return m_pxPhysicalObject->GetPhysicsObjectHandle();
}

bool Tools_Entity_PhysicalObject::DoesObjectNeedToBeAbleToMove() const
{

	return false;
}

Asura_String Tools_Entity_PhysicalObject::GetShapeOrSkinName() const
{
	if(m_xShapeFilename.IsBlank()) return "";

	Asura_String xName;
	m_xShapeFilename.SplitFilename(0,&xName,0);
	return xName;
}


bool Tools_Entity_PhysicalObject::GetParameter(Asura_Hash_ID uParamID,class Tools_Parameter* pxParam) const
{
	switch(uParamID)
	{
		case g_uHash_PhysicalObject_MergeWithEnv :
			if(!SupportsMerging()) return false;
			pxParam->SetBool((m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV)!=0);
			break;

		case g_uHash_PhysicalObject_MergeUniformScale :
			if(!SupportsMerging()) return false;
			pxParam->SetBool(m_bUniformScale);
			break;

		case g_uHash_PhysicalObject_MergeScaleX :
			if(!SupportsMerging()) return false;
			pxParam->SetFloat(Entity->ToolsData->m_xScale.x);
			break;

		case g_uHash_PhysicalObject_MergeScaleY :
			if(!SupportsMerging()) return false;
			pxParam->SetFloat(Entity->ToolsData->m_xScale.y);
			break;

		case g_uHash_PhysicalObject_MergeScaleZ :
			if(!SupportsMerging()) return false;
			pxParam->SetFloat(Entity->ToolsData->m_xScale.z);
			break;


		case g_uHash_PhysicalObject_Destructible :
			pxParam->SetBool(m_bIsDestructible);
			break;

		case g_uHash_PhysicalObject_Health :
			pxParam->SetFloat(m_pxPhysicalObject->m_fHealth);
			break;

		case g_uHash_PhysicalObject_MergeWithSky :
			pxParam->SetBool((m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY)!=0);
			break;

		default :
			return Tools_Entity_Object::GetParameter(uParamID,pxParam);
	}
	return true;
}

bool Tools_Entity_PhysicalObject::SetParameter(Asura_Hash_ID uParamID,const class Tools_Parameter* pxParam)
{
	bool bChangedPhysicalFlags = false;

	switch(uParamID)
	{
		case g_uHash_PhysicalObject_MergeWithEnv :
			if(!SupportsMerging()) return false;
			if(pxParam->GetBool())
			{
				m_uPhysicalObjectFlags |= ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV;

                // These flags cannot be set if we are merged with the environment.
                if ( m_pxPhysicalObject->HasBlockWaypointLinksFlag() )
                {
                    m_uPhysicalObjectFlags &= ~ASURA_PHYSICALOBJECT_FLAG_BLOCKWAYPOINTLINKS;
                    m_pxPhysicalObject->SetPhysicalObjectFlags( m_uPhysicalObjectFlags );
                    // Remove the old link blocks.
                    m_pxPhysicalObject->UpdateWaypointLinkBlocks( true );
                    // Forget that we ever had any.
                    m_pxPhysicalObject->DiscardWaypointLinkBlockInformation();
                }
                if ( m_pxPhysicalObject->HasBlockNavDataGenerationFlag() )
                {
                    m_uPhysicalObjectFlags &= ~ASURA_PHYSICALOBJECT_FLAG_BLOCKNAVDATAGENERATION;
                    m_pxPhysicalObject->SetPhysicalObjectFlags( m_uPhysicalObjectFlags );
                }
			}
			else
			{
				m_uPhysicalObjectFlags &=~ ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV;
			}
			bChangedPhysicalFlags = true;
			break;

		case g_uHash_PhysicalObject_MergeUniformScale :
			if(!SupportsMerging()) return false;
			m_bUniformScale = pxParam->GetBool();
			if(m_bUniformScale)
			{
				//set all the scale values to be the same
				Entity->ToolsData->m_xScale.z = Entity->ToolsData->m_xScale.y = Entity->ToolsData->m_xScale.x;
			}
			if(Entity->ToolsData->m_xScale != GetScale())
			{
				SetScale(Entity->ToolsData->m_xScale);
				UnmergePolygonsFromEnvironment();
			}

			break;

		case g_uHash_PhysicalObject_MergeScaleX :
			if(!SupportsMerging()) return false;
			Entity->ToolsData->m_xScale.x = pxParam->GetFloat();
			if(m_bUniformScale)
			{
				Entity->ToolsData->m_xScale.y = pxParam->GetFloat();
				Entity->ToolsData->m_xScale.z = pxParam->GetFloat();
			}

			if(Entity->ToolsData->m_xScale != GetScale())
			{
				SetScale(Entity->ToolsData->m_xScale);
				UnmergePolygonsFromEnvironment();
			}
			break;

		case g_uHash_PhysicalObject_MergeScaleY :
			if(!SupportsMerging()) return false;
			Entity->ToolsData->m_xScale.y = pxParam->GetFloat();
			if(m_bUniformScale)
			{
				Entity->ToolsData->m_xScale.x = pxParam->GetFloat();
				Entity->ToolsData->m_xScale.z = pxParam->GetFloat();
			}
			if(Entity->ToolsData->m_xScale != GetScale())
			{
				SetScale(Entity->ToolsData->m_xScale);
				UnmergePolygonsFromEnvironment();
			}
			break;

		case g_uHash_PhysicalObject_MergeScaleZ :
			if(!SupportsMerging()) return false;
			Entity->ToolsData->m_xScale.z = pxParam->GetFloat();
			if(m_bUniformScale)
			{
				Entity->ToolsData->m_xScale.x = pxParam->GetFloat();
				Entity->ToolsData->m_xScale.y = pxParam->GetFloat();
			}
			if(Entity->ToolsData->m_xScale != GetScale())
			{
				SetScale(Entity->ToolsData->m_xScale);
				UnmergePolygonsFromEnvironment();
			}
			break;


		case g_uHash_PhysicalObject_Destructible :
			m_bIsDestructible = pxParam->GetBool();
			m_pxPhysicalObject->m_bIsDestructible = m_bIsDestructible;
			break;

		case g_uHash_PhysicalObject_Health :
			m_pxPhysicalObject->m_fHealth = pxParam->GetFloat();
			m_pxPhysicalObject->m_fMaxHealth = pxParam->GetFloat();
			break;

		case g_uHash_PhysicalObject_MergeWithSky :
			if(pxParam->GetBool())
			{
				if(!(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY))
				{
					m_uPhysicalObjectFlags |= ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY;

					SetMesh(m_pxDefaultMesh);
					m_bUsingDefaultMesh = true;

					if(s_bOffsetSkyboxObjectsWhenFlagChanged)
					{
						//adjust position so that it looks the same as skybox relative
						Position -= *Asura_Camera::GetPosition();
						UpdatePositionInAsuraData(false);
					}
				}
			}
			else
			{
				if(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY)
				{
					m_uPhysicalObjectFlags &=~ ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHSKY;
					if(this->m_uSkinID != ASURA_HASH_ID_UNSET)
					{
						SetToolsSkinAndAnim(m_uFileID, m_uSkinID, m_uAnimID, m_uAnimFileID);
					}
					else
					{
						SetToolsShape(m_uFileID);
					}

					if(s_bOffsetSkyboxObjectsWhenFlagChanged)
					{
						//adjust position so that it looks the same as world relative
						Position += *Asura_Camera::GetPosition();
						UpdatePositionInAsuraData(false);
					}

				}
			}
			bChangedPhysicalFlags = true;
			break;

		case g_uHash_Entity_StartsActive:
		{
			PARENT::SetParameter( uParamID, pxParam );
			#ifdef ASURA_USE_AXON
			m_pxPhysicalObject->SetBrainActive( pxParam->GetBool() );
			#endif
			break;
		}

		default :
			return PARENT::SetParameter(uParamID,pxParam);
	}

	if(bChangedPhysicalFlags)
	{
		//update flags in entity
		m_pxPhysicalObject->SetPhysicalObjectFlags(m_uPhysicalObjectFlags);

		//check to see if object should be unmerged
		if(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_ISMERGEDWITHENV && !(m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV))
		{
			UnmergePolygonsFromEnvironment();
		}

		// refresh skybox object list
		Asura_Skybox::InvalidateSkyboxObjects();
	}

	SetEntityModified();

	return true;
}


bool Tools_Entity_PhysicalObject::CanBeScaled() const
{
	return (m_uPhysicalObjectFlags & ASURA_PHYSICALOBJECT_FLAG_SHOULDMERGEWITHENV) && !m_pxPhysicalObject->GetSkinID();
}

void Tools_Entity_PhysicalObject::UpdateScaleInAsuraData()
{
	if(Entity->ToolsData->m_xScale != GetScale())
	{
		Entity->ToolsData->m_xScale = GetScale();

		UnmergePolygonsFromEnvironment();
		SetEntityModified();
	}		
}


void Tools_Entity_PhysicalObject::SetDefaultMesh(const char* szMeshName)
{
	m_pxDefaultMesh = Asura_Mesh_System::GetMesh(szMeshName);
	if(m_bUsingDefaultMesh) SetMesh(m_pxDefaultMesh);
}





void Tools_Entity_PhysicalObject::SelectShape_Dialog()
{
	Asura_Assert(s_pxCurrentlySelectedObject, "Tools_Entity_PhysicalObject::SelectShape_Dialog() called with no object selected");
	if (s_pxCurrentlySelectedObject)
	{
		s_pxCurrentlySelectedObject->SelectShape();
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::SelectSkinAndAnim_Dialog().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Static member function called when this option is selected in 
//              the tools overlay,
// AUTHOR       Mike Healey - 17.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::SelectSkinAndAnim_Dialog()
{
	Asura_Assert(s_pxCurrentlySelectedObject, "Tools_Entity_PhysicalObject::SelectSkinAndAnim_Dialog() called with no object selected");
	if (s_pxCurrentlySelectedObject)
	{
		s_pxCurrentlySelectedObject->SelectSkinAndAnim();
	}
}
// End function Tools_Entity_PhysicalObject::SelectSkinAndAnim_Dialog().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::SelectSkinAndAnim_Dialog().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Static member function called when this option is selected in 
//              the tools overlay,
// AUTHOR       Mike Healey - 17.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::SelectSkinOnly_Dialog()
{
	Asura_Assert(s_pxCurrentlySelectedObject, "Tools_Entity_PhysicalObject::SelectSkinOnly_Dialog() called with no object selected");
	if (s_pxCurrentlySelectedObject)
	{
		s_pxCurrentlySelectedObject->SelectSkinAndAnim(true);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::SelectAnim_Dialog().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Static member function called when this option is selected in 
//              the tools overlay,
// AUTHOR       Mike Healey - 18.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::SelectAnim_Dialog()
{
	Asura_Assert(s_pxCurrentlySelectedObject, "Tools_Entity_PhysicalObject::SelectAnim_Dialog() called with no object selected");
	if (s_pxCurrentlySelectedObject)
	{
		s_pxCurrentlySelectedObject->SelectAnim();
	}
}
// End function Tools_Entity_PhysicalObject::SelectAnim_Dialog().


void Tools_Entity_PhysicalObject::SelectShape()
{
	if (!IsSelected()) return;

	Asura_Hash_ID uShapeHashID = Tools_ShapeTable::SelectShape();

	if( ASURA_HASH_ID_UNSET != uShapeHashID )
	{
		Changed = true;

		if (m_pxPhysicalObject)
		{
			// Update server object's shape ID.
			m_pxPhysicalObject->SetShapeFileID( uShapeHashID );

			// Update the tools shape also
			SetToolsShape( uShapeHashID );
			

			//Shape has changed , so if merged to environment , we need to unmerge now
			UnmergePolygonsFromEnvironment();

		}

		// (MH - 28.11.2002) attempt to get the textures in for this object
		Asura_Texture_Management::LoadTextures();
	}
}


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::SelectSkinAndAnim().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Displays a skin selection dialog, and if that gets a skin, 
//              displays an anim selection dialog. If both are valid, updates
//              the tools object and server entity with the new values.
// AUTHOR       Mike Healey - 17.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::SelectSkinAndAnim(bool bForceNoAnimation /* = false */)
{
	Asura_Assert(IsSelected(), "SelectSkinAndAnim() called on non-selected object");
	if (!IsSelected()) return;

	Asura_Assert(m_pxPhysicalObject, "Tools_Entity_PhysicalObject has no m_pxPhysicalObject");

	Asura_String xSkinName;
	Asura_String xSkinFilename;
	Asura_String xAnimName;
	Asura_String xAnimFilename;

	if(m_uSkinID)
	{
		//get current skin name so that it will be selected as the default option
		xSkinName = GetShapeOrSkinName();
	}

	if (!Tools_Animation_Files::ChooseSkin(&xSkinName, &xSkinFilename))
	{
		return;
	}
	
	if(bForceNoAnimation)
	{
		xAnimName.MakeNull();
		xAnimFilename.MakeNull();	
	}
	else
	{
		Tools_Animation_Files::ChooseAnimation(xSkinName.GetString(), NULL, &xAnimName, &xAnimFilename,0,true);
	}

	Asura_Hash_ID uSkinFileHashID   = Asura_GetHashID(xSkinFilename.GetString());
	Asura_Hash_ID uSkinHashID       = Asura_GetHashID(xSkinName.GetString());
	Asura_Hash_ID uAnimFileHashID   = Asura_GetHashID(xAnimFilename.GetString());
	Asura_Hash_ID uAnimHashID       = Asura_GetHashID(xAnimName.GetString());
	SetToolsSkinAndAnim(uSkinFileHashID, uSkinHashID, uAnimHashID, uAnimFileHashID);
	m_pxPhysicalObject->SetSkinAndAnimID(uSkinHashID, uAnimHashID);
	m_pxPhysicalObject->SetSkinAndAnimFileID(uSkinFileHashID, uAnimFileHashID);

	//Shape has changed , so if merged to environment , we need to unmerge now
	UnmergePolygonsFromEnvironment();

	SetEntityModified();

	// (MH - 28.11.2002) attempt to get the textures in for this skin
	Asura_Texture_Management::LoadTextures();
}
// End function Tools_Entity_PhysicalObject::SelectSkinAndAnim().


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     Tools_Entity_PhysicalObject::SelectAnim().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Displays an anim selection dialog. If valid, updates
//              the tools object and server entity with the new value.
// AUTHOR       Mike Healey - 18.10.2002.
//
//////////////////////////////////////////////////////////////////////////////
void Tools_Entity_PhysicalObject::SelectAnim()
{
	Asura_Assert(IsSelected(), "SelectAnim() called on non-selected object");
	if (!IsSelected()) return;

	Asura_Assert(m_pxPhysicalObject, "Tools_Entity_PhysicalObject has no m_pxPhysicalObject");
	if (!m_pxPhysicalObject) return;

	if (!m_pxPhysicalObject->HasSkinOrHierarchy())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff, "object is not animating");
		return;
	}

	const Asura_Hierarchy_Skin* pxSkin = Asura_Animation_System::GetSkinByHash(m_pxPhysicalObject->GetSkinID());
	Asura_Assert(pxSkin, "skin hash ID failed to find skin");
	if (!pxSkin) return;

	Asura_String xAnimName;
	Asura_String xAnimFilename;
	Asura_Hierarchy_Anim* pxInitialAnim = 0;
	if(m_pxPhysicalObject->GetAnimID())
	{
		pxInitialAnim = Asura_Animation_System::GetAnimByHash(m_pxPhysicalObject->GetAnimID());
	}


	if(Tools_Animation_Files::ChooseAnimation(pxSkin->Name.GetString(), NULL, &xAnimName, &xAnimFilename,pxInitialAnim,true) == pxInitialAnim)
	{
		//If user cancels and we get back the original animation ,then the anim name and filename won't have been set , 
		//so return here instead. RAR 27/06/2006
		return;
	}

	Asura_Hash_ID uSkinFileHashID   = m_pxPhysicalObject->GetSkinFileID();
	Asura_Hash_ID uSkinHashID       = m_pxPhysicalObject->GetSkinID();
	Asura_Hash_ID uAnimFileHashID   = Asura_GetHashID(xAnimFilename.GetString());
	Asura_Hash_ID uAnimHashID       = Asura_GetHashID(xAnimName.GetString());
	SetToolsSkinAndAnim(uSkinFileHashID, uSkinHashID, uAnimHashID, uAnimFileHashID);
	m_pxPhysicalObject->SetSkinAndAnimID(uSkinHashID, uAnimHashID);
	m_pxPhysicalObject->SetSkinAndAnimFileID(uSkinFileHashID, uAnimFileHashID);

	Changed = true;
}
// End function Tools_Entity_PhysicalObject::SelectAnim().

bool Tools_Entity_PhysicalObject::AddSharedParamDialogToOverlay( Asura_Overlay* pxOverlay, const Asura_Vector_2& xOffset, Asura_Hash_ID uParamID )
{
    Asura_Dialog* pxDialog = NULL;

    switch( uParamID )
    {
        case g_uHash_PhysicalObject_MergeWithEnv:
	        pxDialog = new Asura_SharedTickbox_Dialog( "Merge With Env", xOffset, &g_xParam_PhysicalObject_MergeWithEnv );
	        pxDialog->SetHelpText( "The mesh for this object should be merged into the environment mesh." );
            break;

        case g_uHash_PhysicalObject_MergeScaleX:
            pxDialog = new Asura_Numerical_Dialog( "Merge Scale X", xOffset, &g_xParam_PhysicalObject_MergeScaleX, 0.001f, 99.0f, 0.1f, "%.3f" );
            pxDialog->SetHelpText( "Scale factor to use for shapes being merged into the environment." );
            break;

        case g_uHash_PhysicalObject_MergeScaleY:
            pxDialog = new Asura_Numerical_Dialog( "Merge Scale Y", xOffset, &g_xParam_PhysicalObject_MergeScaleY, 0.001f, 99.0f, 0.1f, "%.3f" );
            pxDialog->SetHelpText( "Scale factor to use for shapes being merged into the environment." );
            break;

        case g_uHash_PhysicalObject_MergeScaleZ:
            pxDialog = new Asura_Numerical_Dialog( "Merge Scale Z", xOffset, &g_xParam_PhysicalObject_MergeScaleZ, 0.001f, 99.0f, 0.1f, "%.3f" );
            pxDialog->SetHelpText( "Scale factor to use for shapes being merged into the environment." );
            break;

        case g_uHash_PhysicalObject_MergeUniformScale:
            pxDialog = new Asura_SharedTickbox_Dialog( "Uniform Scale", xOffset, &g_xParam_PhysicalObject_MergeUniformScale );
            pxDialog->SetHelpText( "Forces the X,Y and Z scale values to be the same." );
            break;

        case g_uHash_PhysicalObject_Destructible:
            pxDialog = new Asura_SharedTickbox_Dialog( "Destructible", xOffset, &g_xParam_PhysicalObject_Destructible );
            pxDialog->SetHelpText( "Sets whether or not this entity is destructible." );
            break;

        case g_uHash_PhysicalObject_Health:
            pxDialog = new Asura_Numerical_Dialog( "Health Points", xOffset, &g_xParam_PhysicalObject_Health, 0.0f, 9999.9f, 1.0f, "%0.0f" );
            break;

        case g_uHash_PhysicalObject_MergeWithSky:
            pxDialog = new Asura_SharedTickbox_Dialog( "Merge With Skybox", xOffset, &g_xParam_PhysicalObject_MergeWithSky );
            pxDialog->SetHelpText( "Signal that this object should be rendered as part of the skybox." );
            break;

        default:
            break;
    }

	if( pxDialog )
	{
		pxOverlay->AddDialog( pxDialog );
        return true;
	}
    else
    {
        return PARENT::AddSharedParamDialogToOverlay( pxOverlay, xOffset, uParamID );
    }
}

void Tools_Entity_PhysicalObject::SetEntityModified()
{
	if(IsMergedWithSkybox())
	{
		//Ensure skybox object list is up to date if one of the shapes that belong to it has been edited
		Asura_Skybox::InvalidateSkyboxObjects();
	}

	PARENT::SetEntityModified();
}

bool Tools_Entity_PhysicalObject::Verify( const bool bAutomatic )
{
	bool bOk = PARENT::Verify( bAutomatic );

	// Check for corrupted pos/ori.
	if ( m_pxPhysicalObject->GetPosition() != m_pxPhysicalObject->GetPosition() )
	{
		Tools_MessageLog::AddMessage( Tools_MessageLog::VERIFICATION, "Objects", new Tools_MessageLogAction_GotoEntity( GetGuid() ),
			"Object '%s' has a corrupt position.", GetName() );
		bOk = false;
	}
	if ( m_pxPhysicalObject->GetOrientation() != m_pxPhysicalObject->GetOrientation() )
	{
		Tools_MessageLog::AddMessage( Tools_MessageLog::VERIFICATION, "Objects", new Tools_MessageLogAction_GotoEntity( GetGuid() ),
			"Object '%s' has a corrupt orientation.", GetName() );
		bOk = false;
	}

	return bOk;
}
