//
// Tools_PaletteColourMap.cpp
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include <math.h>
#include "ImageProcessing/Tools_ColourMap.h"

#ifndef NULL
	#define NULL 0
#endif

#define CM_QUEUESIZE 0

#if CM_TREEMETHOD
	#define CM_MAXLISTSIZE 32U // best values seem to be in range 16-40 for random images
	#define CM_MAXSHIFT 8U // bits per colour int
	#define CM_MINSHIFT 2U
	#define CM_SHIFTSTEP 2U
	#define CM_MUSTBEINCUBE 0
#else
	#define CM_SUBCUBE_SIZE_BITS 4U
	#define CM_HITS 0
#endif

namespace CM
{
	class SquaresTable
	{
		public:
			SquaresTable()
			{
				for (signed i=-255; i<=255; ++i)
				{
					valueA[i+255] = static_cast<unsigned>(i*i);
				}
			}
			unsigned valueA[511];
	}
		squareTable;
		
	#define CM_SQUARED(x) (squareTable.valueA[(x)+255])
	
	inline unsigned Distance2(CMColour const & _c1R, CMColour const & _c2R)
	{
		signed rd = static_cast<signed>(_c1R.r) - static_cast<signed>(_c2R.r);
		signed gd = static_cast<signed>(_c1R.g) - static_cast<signed>(_c2R.g);
		signed bd = static_cast<signed>(_c1R.b) - static_cast<signed>(_c2R.b);
		return static_cast<unsigned>(CM_SQUARED(rd)+CM_SQUARED(gd)+CM_SQUARED(bd));
	}
	
	#if CM_QUEUESIZE
	class QueueNode
	{
		public:
			void AddTail();
			void Remove();
			
			static void Check();
			
			Subcube * * refP;
		
		private:
		
			QueueNode * nextP;
			QueueNode * prevP;
	};
	extern QueueNode queueTail;
	QueueNode queueHead = { NULL, &queueTail, NULL };
	QueueNode queueTail = { NULL, NULL, &queueHead };
	unsigned queueSize = 0;
	
	inline void QueueNode::AddTail()
	{
		nextP = &queueTail;
		prevP = queueTail.prevP;
		prevP->nextP = this;
		queueTail.prevP = this;
	}
	
	inline void QueueNode::Remove()
	{
		nextP->prevP = prevP;
		prevP->nextP = nextP;
	}
	#endif
		
	class Subcube
	{
		private:
			struct PalEntry : public CMColour
			{
				PalEntry(CMColour const & _colourR, unsigned _index);
				unsigned index;
			};
			struct Node : public PalEntry
			{
				Node(PalEntry const & _palEntryR, Node * _nextP);
				Node(CMColour const & _colourR, unsigned _index, Node * _nextP);
				~Node();
				Node * nextP;
				unsigned CountMemoryAllocated() const;
			}
				* colourListP;
				
			unsigned listSize;
			
			#if CM_TREEMETHOD
			Subcube * nextLevelPA[1<<CM_SHIFTSTEP*3];
			bool leafB;
			#elif CM_HITS
			unsigned short hitsA[1<<CM_SUBCUBE_SIZE_BITS*3];
			#endif
			
		public:
			#if CM_QUEUESIZE
			QueueNode queueNode;
			#endif
			
			Subcube();
			~Subcube();
			void AddColour(CMColour const & _colourR, unsigned _index);
			void AddColour(PalEntry const & _palEntryR);
			#if CM_TREEMETHOD
			unsigned NearestColour(CMColour const & _colourR, unsigned _shift);
			#else
			unsigned NearestColour(CMColour const & _colourR);
			#endif
			unsigned CountMemoryAllocated() const;
	};

	#if CM_QUEUESIZE
	inline void QueueNode::Check()
	{
		if (queueSize >= CM_QUEUESIZE)
			delete *queueHead.nextP->refP;
	}
	#endif
	
	inline Subcube::PalEntry::PalEntry(CMColour const & _colourR, unsigned _index)
		: CMColour(_colourR)
		, index(_index)
	{
	}
	
	inline Subcube::Node::Node(CMColour const & _colourR, unsigned _index, Node * _nextP)
		: PalEntry(_colourR,_index)
		, nextP(_nextP)
	{
	}
	
	inline Subcube::Node::Node(PalEntry const & _palEntryR, Node * _nextP)
		: PalEntry(_palEntryR)
		, nextP(_nextP)
	{
	}
	
	inline Subcube::Node::~Node()
	{
		Node * nodeP = nextP;
		while (nodeP)
		{
			Node * nnP = nodeP->nextP;
			nodeP->nextP = NULL;
			delete nodeP;
			nodeP = nnP;
		}
	}
	
	unsigned Subcube::Node::CountMemoryAllocated() const
	{
		return sizeof(Node) + (nextP ? nextP->CountMemoryAllocated() : 0U);
	}
	
	inline Subcube::Subcube()
		: colourListP(NULL)
		, listSize(0)
		#if CM_TREEMETHOD
		, leafB(false)
	{
		for (unsigned i=0; i<1<<CM_SHIFTSTEP*3; ++i)
			nextLevelPA[i]=NULL;
		#else
	{
		#if CM_HITS
		for (unsigned i=0; i<1<<CM_SUBCUBE_SIZE_BITS*3; ++i)
		{
			hitsA[i]=static_cast<unsigned short>(-1);
		}
		#endif
		#endif
		#if CM_QUEUESIZE
		queueNode.refP = NULL;
		#endif
	}
	
	/*inline*/ Subcube::~Subcube()
	{
		#if CM_QUEUESIZE
		if (queueNode.refP)
		{
			-- queueSize;
			queueNode.Remove();
			*queueNode.refP = NULL;
		}
		#endif
		if (colourListP) delete colourListP;
		#if CM_TREEMETHOD
		for (unsigned i=0; i<1<<CM_SHIFTSTEP*3; ++i)
			if (nextLevelPA[i]) delete nextLevelPA[i];
		#endif
	}
	
	inline void Subcube::AddColour(CMColour const & _colourR, unsigned _index)
	{
		colourListP = new Node(_colourR,_index,colourListP);
		++ listSize;
	}
	
	inline void Subcube::AddColour(PalEntry const & _palEntryR)
	{
		colourListP = new Node(_palEntryR,colourListP);
		++ listSize;
	}
	
	#if CM_TREEMETHOD
	unsigned Subcube::NearestColour(CMColour const & _colourR, unsigned _shift)
	#else
	unsigned Subcube::NearestColour(CMColour const & _colourR)
	#endif
	{
		#if CM_QUEUESIZE
		queueNode.Remove();
		queueNode.AddTail();
		#endif
		
		#if CM_TREEMETHOD
		if (leafB)
		{
		#elif CM_HITS
			unsigned hit_index = static_cast<unsigned>(_colourR.r) & (1<<CM_SUBCUBE_SIZE_BITS)-1;
			hit_index |= (static_cast<unsigned>(_colourR.g) & (1<<CM_SUBCUBE_SIZE_BITS)-1)*(1<<CM_SUBCUBE_SIZE_BITS);
			hit_index |= (static_cast<unsigned>(_colourR.b) & (1<<CM_SUBCUBE_SIZE_BITS)-1)*(1<<CM_SUBCUBE_SIZE_BITS*2);
			if (static_cast<unsigned short>(-1) != hitsA[hit_index])
				return hitsA[hit_index];
		#endif
			unsigned min_dist = static_cast<unsigned>(-1);
			unsigned md_colour = 0;
			for (Node * listP = colourListP; listP; listP = listP->nextP)
			{
				unsigned this_dist = Distance2(_colourR,*listP);
				if (this_dist<min_dist)
				{
					min_dist = this_dist;
					md_colour = listP->index;
				}
			}
			#if !CM_TREEMETHOD && CM_HITS
			hitsA[hit_index] = static_cast<unsigned short>(md_colour);
			#endif
			return md_colour;
		#if CM_TREEMETHOD
		}
		else
		{
			_shift -= CM_SHIFTSTEP;
			// which child subcube next?
			unsigned rb = _colourR.r>>_shift & (1<<CM_SHIFTSTEP)-1;
			unsigned gb = _colourR.g>>_shift & (1<<CM_SHIFTSTEP)-1;
			unsigned bb = _colourR.b>>_shift & (1<<CM_SHIFTSTEP)-1;
			unsigned childidx = rb+(gb<<CM_SHIFTSTEP*1)+(bb<<CM_SHIFTSTEP*2);
			if (!nextLevelPA[childidx])
			{
				Subcube * childP = nextLevelPA[childidx] = new Subcube;
				
				#if CM_QUEUESIZE
				childP->queueNode.refP = &nextLevelPA[childidx];
				childP->queueNode.AddTail();
				++ queueSize;
				QueueNode::Check();
				#endif
				
				// build this subcube...
				unsigned corner_mask = static_cast<unsigned>(-1)<<_shift;
				unsigned centre_add = (1<<_shift)>>1;
				
				// work out centre of cube
				CMColour sc_centre
				(
					static_cast<unsigned char>(_colourR.r & corner_mask | centre_add),
					static_cast<unsigned char>(_colourR.g & corner_mask | centre_add),
					static_cast<unsigned char>(_colourR.b & corner_mask | centre_add)
				);
				
				// how far is nearest colour to centre
				unsigned min_dist = static_cast<unsigned>(-1);
				CMColour md_colour;
				for (Node * listP = colourListP; listP; listP = listP->nextP)
				{
					unsigned this_dist = Distance2(sc_centre,*listP);
					if (this_dist<min_dist)
					{
						min_dist = this_dist;
						md_colour = *listP;
					}
				}
				
				#if 0
				unsigned far_corner_add = (centre_add<<1)-1;
				CMColour opp_corner;
				opp_corner.r = (md_colour.r < sc_centre.r ? far_corner_add : 0) + (_colourR.r & corner_mask);
				opp_corner.g = (md_colour.g < sc_centre.g ? far_corner_add : 0) + (_colourR.g & corner_mask);
				opp_corner.b = (md_colour.b < sc_centre.b ? far_corner_add : 0) + (_colourR.b & corner_mask);
				min_dist = Distance2(opp_corner,md_colour);
				#elif 0
				double dbl_dist = sqrt(static_cast<double>(min_dist)) + sqrt(static_cast<double>(centre_add)*centre_add*12) + 0;
				min_dist = static_cast<unsigned>(dbl_dist*dbl_dist)+1;
				#elif 0
				min_dist += ((1<<_shift)+1)*((1<<_shift)+1)*3+1;
				#else
				unsigned r_off = md_colour.r < sc_centre.r ? sc_centre.r - md_colour.r : md_colour.r - sc_centre.r;
				unsigned g_off = md_colour.g < sc_centre.g ? sc_centre.g - md_colour.g : md_colour.g - sc_centre.g;
				unsigned b_off = md_colour.b < sc_centre.b ? sc_centre.b - md_colour.b : md_colour.b - sc_centre.b;
				min_dist = CM_SQUARED(r_off+(1<<_shift))+CM_SQUARED(g_off+(1<<_shift))+CM_SQUARED(b_off+(1<<_shift));
				#endif
				
				// which palette colours could possibly map to a colour in the cube
				for (listP = colourListP; listP; listP = listP->nextP)
				{
					unsigned this_dist = Distance2(sc_centre,*listP);
					if (this_dist<min_dist)
						childP->AddColour(*listP);
				}
				
				if (childP->listSize <= CM_MAXLISTSIZE || _shift<=CM_MINSHIFT
					#if CM_MUSTBEINCUBE
					|| Distance2(sc_centre,md_colour)>3*CM_SQUARED(centre_add)
					#endif
				)
					childP->leafB = true;
			}
			return nextLevelPA[childidx]->NearestColour(_colourR,_shift);
		}
		#endif
	}
	
	unsigned Subcube::CountMemoryAllocated() const
	{
		unsigned size = colourListP ? colourListP->CountMemoryAllocated() : 0U;
		#if CM_TREEMETHOD
		for (unsigned i=0; i<1<<CM_SHIFTSTEP*3; ++i)
			if (nextLevelPA[i]) size += nextLevelPA[i]->CountMemoryAllocated();
		#endif
		return size + sizeof(Subcube);
	}
}

#if !CM_TREEMETHOD
#define CM_SUBCUBE_COUNT (1<<(8-CM_SUBCUBE_SIZE_BITS))
#endif

#ifdef __GNUC__ // bug in GNU C
#define CM_SUBCUBE_T typeof(*firstLevelP)
#else
#define CM_SUBCUBE_T CM::Subcube
#endif

Tools_ColourMap::Tools_ColourMap(CMColour const * _paletteA, unsigned _paletteSize)
#if CM_TREEMETHOD
	: firstLevelP ( new CM_SUBCUBE_T )
{
	#if CM_QUEUESIZE
	firstLevelP->queueNode.refP = NULL;
	#endif
	for (unsigned i=0; i<_paletteSize; ++i)
		firstLevelP->AddColour(_paletteA[i],i);
}
#else
	: firstLevelPA ( new CM_SUBCUBE_T * [CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT] )
	, paletteA( new CMColour[_paletteSize] )
	, paletteSize(_paletteSize)
{
	for (unsigned i=0; i<_paletteSize; ++i)
		paletteA[i] = _paletteA[i];
		
	for (unsigned i=0; i<CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT; ++i)
		firstLevelPA[i] = NULL;
}
#endif

Tools_ColourMap::~Tools_ColourMap()
{
	#if CM_TREEMETHOD
	delete firstLevelP;
	#else
	delete[] paletteA;
	for (unsigned i=0; i<CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT; ++i)
		if (firstLevelPA[i]) delete firstLevelPA[i];
	delete[] firstLevelPA;
	#endif
}

unsigned Tools_ColourMap::NearestColour(CMColour const & _colourR) const
{
	#if CM_TREEMETHOD
	return firstLevelP->NearestColour(_colourR, CM_MAXSHIFT);
	#else
	unsigned rc=_colourR.r >> CM_SUBCUBE_SIZE_BITS;
	unsigned gc=_colourR.g >> CM_SUBCUBE_SIZE_BITS;
	unsigned bc=_colourR.b >> CM_SUBCUBE_SIZE_BITS;
	unsigned childidx=rc*CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT + gc*CM_SUBCUBE_COUNT + bc;
	if (!firstLevelPA[childidx])
	{
		firstLevelPA[childidx] = new CM_SUBCUBE_T;
				
		using namespace CM;
		
		#if CM_QUEUESIZE
		firstLevelPA[childidx]->queueNode.refP = &firstLevelPA[childidx];
		firstLevelPA[childidx]->queueNode.AddTail();
		++ queueSize;
		QueueNode::Check();
		#endif
		
		// build this subcube...
		#define corner_mask (static_cast<unsigned>(-1)<<CM_SUBCUBE_SIZE_BITS)
		#define centre_add ((1<<CM_SUBCUBE_SIZE_BITS)>>1)
		#define _shift CM_SUBCUBE_SIZE_BITS
		
		// work out centre of cube
		CMColour sc_centre
		(
			static_cast<unsigned char>(_colourR.r & corner_mask | centre_add),
			static_cast<unsigned char>(_colourR.g & corner_mask | centre_add),
			static_cast<unsigned char>(_colourR.b & corner_mask | centre_add)
		);
		
		// how far is nearest colour to centre
		unsigned min_dist = static_cast<unsigned>(-1);
		CMColour md_colour;
		for (unsigned i = 0; i<paletteSize; ++i)
		{
			unsigned this_dist = Distance2(sc_centre,paletteA[i]);
			if (this_dist<min_dist)
			{
				min_dist = this_dist;
				md_colour = paletteA[i];
			}
		}
		
		#if 0
		unsigned far_corner_add = (centre_add<<1)-1;
		CMColour opp_corner;
		opp_corner.r = (md_colour.r < sc_centre.r ? far_corner_add : 0) + (_colourR.r & corner_mask);
		opp_corner.g = (md_colour.g < sc_centre.g ? far_corner_add : 0) + (_colourR.g & corner_mask);
		opp_corner.b = (md_colour.b < sc_centre.b ? far_corner_add : 0) + (_colourR.b & corner_mask);
		min_dist = Distance2(opp_corner,md_colour);
		#elif 0
		double dbl_dist = sqrt(static_cast<double>(min_dist)) + sqrt(static_cast<double>(centre_add)*centre_add*12) + 0;
		min_dist = static_cast<unsigned>(dbl_dist*dbl_dist)+1;
		#elif 0
		min_dist += ((1<<_shift)+1)*((1<<_shift)+1)*3+1;
		#else
		unsigned r_off = md_colour.r < sc_centre.r ? sc_centre.r - md_colour.r : md_colour.r - sc_centre.r;
		unsigned g_off = md_colour.g < sc_centre.g ? sc_centre.g - md_colour.g : md_colour.g - sc_centre.g;
		unsigned b_off = md_colour.b < sc_centre.b ? sc_centre.b - md_colour.b : md_colour.b - sc_centre.b;
		min_dist = CM_SQUARED(r_off+(1<<_shift))+CM_SQUARED(g_off+(1<<_shift))+CM_SQUARED(b_off+(1<<_shift));
		#endif
		
		// which palette colours could possibly map to a colour in the cube
		for (unsigned i = 0; i<paletteSize; ++i)
		{
			unsigned this_dist = Distance2(sc_centre,paletteA[i]);
			if (this_dist<min_dist)
				firstLevelPA[childidx]->AddColour(paletteA[i],i);
		}
	}
	return firstLevelPA[childidx]->NearestColour(_colourR);
	#endif
}

unsigned Tools_ColourMap::CountMemoryAllocated() const
{
	#if CM_TREEMETHOD
	return sizeof(Tools_ColourMap) + firstLevelP->CountMemoryAllocated();
	#else
	unsigned size = sizeof(Tools_ColourMap)+paletteSize*sizeof(CMColour)+CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT*sizeof(CM_SUBCUBE_T *);
	for (unsigned i=0; i<CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT*CM_SUBCUBE_COUNT; ++i)
		if (firstLevelPA[i]) size += firstLevelPA[i]->CountMemoryAllocated();
	return size;
	#endif
}
