//
// Tools_IPImage.h
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _INCLUDED_IPIMAGE_HPP_
#define _INCLUDED_IPIMAGE_HPP_

#include "Tools_RGBTypes.h"
#include <stddef.h>

typedef RGBTriple<unsigned char> IPColour;

typedef unsigned IPIndex;

typedef unsigned IPAlphaValue;

class Tools_IPImage
{
	public:
		Tools_IPImage();
		Tools_IPImage(Tools_IPImage const & r);
		Tools_IPImage const & operator = (Tools_IPImage const & r);
		~Tools_IPImage();
		
		enum { NOT_PALETTIZED = 0 };
		
		void CreateNew(unsigned w, unsigned h, unsigned psize);
		
		void MoveTo(Tools_IPImage * pDest);
		
		bool IsValid() const;
		
		bool IsPalettized() const;
		unsigned GetPaletteSize() const;
		bool SetPaletteSize(unsigned);
		unsigned GetWidth() const;
		unsigned GetHeight() const;
		
		bool operator == (Tools_IPImage const &) const;
		bool operator != (Tools_IPImage const &) const;
		
		IPIndex * GetIndexRow(unsigned row);
		IPIndex const * GetIndexRow(unsigned row) const;
		
		IPColour * GetColourRow(unsigned row);
		IPColour const * GetColourRow(unsigned row) const;
		
		IPColour * GetPalette();
		IPColour const * GetPalette() const;
		
		void SetMaskColour(IPColour colour);
		void SetMaskIndex(IPIndex index);
		IPColour GetMaskColour() const;
		IPIndex GetMaskIndex() const;
		void RemoveMask();
		bool HasMask() const;
		
		bool SwapIndeces(IPIndex index1, IPIndex index2);
		
		void CreateAlphaChannel();
		void RemoveAlphaChannel();
		bool HasAlphaChannel() const;
		IPAlphaValue * GetAlphaRow(unsigned row);
		IPAlphaValue const * GetAlphaRow(unsigned row) const;
		
		void MoveAlphaTo(Tools_IPImage * pDest);
		
	protected:
		mutable bool bModified;
	
	private:
		void Destroy();
		void Init();
		void CopyFrom(Tools_IPImage const &);
		void Create();
		
		IPColour * aColourData;
		IPColour * * aColourRows;
		
		IPIndex * aIndexData;
		IPIndex * * aIndexRows;
		
		IPAlphaValue * aAlphaData;
		IPAlphaValue * * aAlphaRows;
		
		unsigned palette_size;
		
		unsigned width;
		unsigned height;
		
		bool bHasMask;
		IPColour colourMask;
		IPIndex indexMask;
};

inline Tools_IPImage::Tools_IPImage()
{
	Init();
}

inline Tools_IPImage::Tools_IPImage(Tools_IPImage const & r)
{
	Init();
	CopyFrom(r);
}

inline Tools_IPImage const & Tools_IPImage::operator = (Tools_IPImage const & r)
{
	if (&r != this)
	{
		Destroy();
		Init();
		CopyFrom(r);
	}
	return *this;
}

inline Tools_IPImage::~Tools_IPImage()
{
	Destroy();
}

inline void Tools_IPImage::CreateNew(unsigned w, unsigned h, unsigned psize)
{
	Destroy();
	Init();
	
	width = w;
	height = h;
	palette_size = psize;
	
	Create();
}

inline bool Tools_IPImage::IsValid() const
{
	return aColourData ? true : false;
}

inline void Tools_IPImage::MoveTo(Tools_IPImage * pDest)
{
	pDest->Destroy();
	pDest->aColourData = aColourData;
	pDest->aColourRows = aColourRows;
	pDest->aIndexData = aIndexData;
	pDest->aIndexRows = aIndexRows;
	pDest->aAlphaData = aAlphaData;
	pDest->aAlphaRows = aAlphaRows;
	pDest->palette_size = palette_size;
	pDest->width = width;
	pDest->height = height;
	pDest->bHasMask = bHasMask;
	pDest->colourMask = colourMask;
	pDest->indexMask = indexMask;
	Init();
}

inline void Tools_IPImage::MoveAlphaTo(Tools_IPImage * pDest)
{
	if (pDest->aAlphaData) delete[] pDest->aAlphaData;
	if (pDest->aAlphaRows) delete[] pDest->aAlphaRows;
	pDest->aAlphaData = aAlphaData;
	pDest->aAlphaRows = aAlphaRows;
	aAlphaData = NULL;
	aAlphaRows = NULL;
}

inline bool Tools_IPImage::IsPalettized() const
{
	return NOT_PALETTIZED != palette_size;
}

inline unsigned Tools_IPImage::GetPaletteSize() const
{
	return palette_size;
}

inline bool Tools_IPImage::SetPaletteSize(unsigned newsize)
{
	if (newsize > palette_size || NOT_PALETTIZED == newsize || NOT_PALETTIZED == palette_size)
		return false;
		
	// only allow this if the new size is less or equal to the old size
	palette_size = newsize;
	return true;
}

inline unsigned Tools_IPImage::GetWidth() const
{
	return width;
}

inline unsigned Tools_IPImage::GetHeight() const
{
	return height;
}

inline bool Tools_IPImage::operator != (Tools_IPImage const & r) const
{
	return ! operator == (r);
}

inline IPIndex const * Tools_IPImage::GetIndexRow(unsigned row) const
{
	return aIndexRows[row];
}

inline IPIndex * Tools_IPImage::GetIndexRow(unsigned row)
{
	bModified = true;
	return aIndexRows[row];
}

inline IPColour const * Tools_IPImage::GetColourRow(unsigned row) const
{
	return aColourRows[row];
}

inline IPColour * Tools_IPImage::GetColourRow(unsigned row)
{
	bModified = true;
	return aColourRows[row];
}

inline IPColour const * Tools_IPImage::GetPalette() const
{
	return aColourData;
}

inline IPColour * Tools_IPImage::GetPalette()
{
	bModified = true;
	return aColourData;
}
		
inline void Tools_IPImage::SetMaskColour(IPColour colour)
{
	colourMask = colour;
	bHasMask = true;
	bModified = true;
}
inline void Tools_IPImage::SetMaskIndex(IPIndex index)
{
	indexMask = index;
	bHasMask = true;
	bModified = true;
}
inline IPColour Tools_IPImage::GetMaskColour() const
{
	return colourMask;
}
inline IPIndex Tools_IPImage::GetMaskIndex() const
{
	return indexMask;
}
inline void Tools_IPImage::RemoveMask()
{
	bHasMask = false;
	bModified = true;
}
inline bool Tools_IPImage::HasMask() const
{
	return bHasMask;
}

inline IPAlphaValue * Tools_IPImage::GetAlphaRow(unsigned row)
{
	bModified = true;
	return aAlphaRows[row];
}

inline IPAlphaValue const * Tools_IPImage::GetAlphaRow(unsigned row) const
{
	return aAlphaRows[row];
}

inline bool Tools_IPImage::HasAlphaChannel() const
{
	return aAlphaRows ? true : false;
}

#endif // ! _INCLUDED_IPIMAGE_HPP_
