//
// Tools_Paletize.h
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _TOOLS_PALETIZE_H_
#define _TOOLS_PALETIZE_H_

/*

Compute a palette using OcTrees.
Author: Jake Hotson

Based on Jamie's implementation of the 'Wu' algorithm,
and an idea I think he had to improve it, which I didn't understand at the time

However, the main goal of this program is to provide an algorithm which
can generate a part-palette given any number preset colours which define the rest.
It knows about the preset colours, so will use this knowledge
to complete the palette which gives the best spread over the input colours.

If the number of preset colours exceeds the desired palette size, then
the output palette will be the best representation of the preset colours.

The other advantage over Wu's algorithm is the ability to split
colour space boxes down to the unit level, though will only do
this when the boxes become sufficiently small.

I am using doubles for the statistics, because the sums of squares can easily
exceed the bound for a 32-bit integer, and floats do not provide enough precision
when small sumbers are added to or subtracted from much larger numbers.

On the negative side, the algorithm is slower (though still quite quick),
particularly so when the input data covers a lot of colour space.
The time taken varies logarithmically with the number of palette colours required.

'darken-flag' has to be implemented by the caller who inserts colours,
by inserting darker versions of colours into the tree in your desired manner.
There is no 'unique-flag' support, though this would be fairly easy to add.

There is potential for this to be adapted for more general purposes, such as
finite vector spaces of any dimension...

To use, first declare an object of type PG_ColourOcTree:

	PG_ColourOcTree t;

Insert image colours with a pixel count using the Insert member function:

	t.Insert(PG_Pixel(RED_VAL,GREEN_VAL,BLUE_VAL),COUNT);

Insert any preset colours using the constant PG_PRESETCOLOUR instead of a count:

	t.Insert(PG_Pixel(RED_VAL,GREEN_VAL,BLUE_VAL),PG_PRESETCOLOUR);

Allocate space for the palette, and fill it using Compute_Palette:

	PG_Pixel * p = new PG_Pixel[256];
	int palette_size = t.Compute_Palette(&p[1],255); // reserves colour 0 for transparency
	// palette_size will only be less than 255 if there were fewer than 255 original colours

Note:

The Insert function accumulates statistics each time it is called, and would perform
unnecessary calculations if the same colour pixel were inserted more than once.
One optimizing solution would be to build up some kind of Hash-table Histogram, and then
only call 'Insert' with the total count for each different colour pixel.

Alteranively the function Store could be used in place of Insert:

	t.Store(PG_Pixel(RED_VAL,GREEN_VAL,BLUE_VAL),COUNT);

And after all Store operations have been done,
the total statistics should be calculated with Correct_Statistics:

	t.Correct_Statistics();

If you want to include darker/brighter versions of colours, eg. for lighting effects,
I recommend adding only the original colours to one tree, using 'Store'.
Then create another PG_ColourOcTree, and unload the colours from the first tree,
using 'Remove_Any_Colour' which returns a PG_PixelCount structure with members r,g,b and n.
You can then manipulate the r,g,b values as required and add each new colour to the new
tree using 'Store'. You can use 'Num_Colours' to find out how many times you should call
'Remove_Any_Colour' and should also first call 'Correct_Counts' to correct the counts of
colours at nodes of the tree (otherwise Remove_Any_Colour will fail). 'Correct_Counts'
behaves similarly to 'Correct_Statistics' but does not aggregate the sums and sums of
squares, which is unncessary until the second tree has been built using 'Store'.

Here is an example:

	t.Correct_Counts();
	PG_ColourOcTree t2;
	for (int count = t.Num_Colours(); count; --count)
	{
		PG_PixelCount p = t.Remove_Any_Colour();

		for (int shades = NUM_SHADES; shades; --shades)
		{
			int r = TRANSFORM_R(p.r,shades);
			int g = TRANSFORM_G(p.g,shades);
			int b = TRANSFORM_B(p.b,shades);

			t2.Store(PG_Pixel(r,g,b),p.n);
		}
	}
	t2.Correct_Statistics();

You can include preset colours in t, before including darker/brighter colours,
though darker/brighter colours will be added as presets, as n in PG_PixelCount will
hold the constant PG_PRESETCOLOUR whenever a preset colour is obtained.

Alternatively, you can add the preset colours into t2 (recommended), just before calling
'Correct_Statistics' or after if you add them with 'Insert'.

*/

#if defined(cencon) && cencon
	#define PROGRESS_COMPUTE
	// #define COUNT_NODES_ALLOCATED
#endif

#if defined(COUNT_NODES_ALLOCATED) || defined(PROGRESS_COMPUTE)
#include "palette.h"
#endif

#define PG_PRESETCOLOUR (-1)

#define PG_COLOURDEPTHBITS 10

#include "ImageProcessing/Tools_RGBTypes.h"

#define PG_RGBData RGBTriple // cannot have a template typedef!

typedef RGBTriple<int> PG_Pixel;

typedef RGBQuad<int,unsigned int> PG_PixelCount;

struct PG_ColourStatistic
{
	PG_ColourStatistic() : n(0), sum(0,0,0), sum_sq(0,0,0) {}
//	PG_ColourStatistic(PG_ColourStatistic const & s) : n(s.n), sum(s.sum), sum_sq(s.sum_sq) {}

	inline void Reset(void)
	{
		n=0;
		sum=PG_RGBData<float>(0,0,0);
		sum_sq=PG_RGBData<double>(0,0,0);
	}
	
	PG_RGBData<float> sum;
	PG_RGBData<double> sum_sq;
	unsigned int n;

	inline void Include(PG_Pixel const & p, int const count)
	{
		n += count;
		sum += (PG_RGBData<float>)p * (float)count;
		sum_sq += (PG_RGBData<double>)p * p * count;
	}

	inline PG_ColourStatistic & operator += (PG_ColourStatistic const & s)
	{
		n += s.n;
		sum += s.sum;
		sum_sq += s.sum_sq;

		return *this;
	}

	inline PG_ColourStatistic & operator -= (PG_ColourStatistic const & s)
	{
		n -= s.n;
		sum -= s.sum;
		sum_sq -= s.sum_sq;

		return *this;
	}

	inline PG_ColourStatistic operator + (PG_ColourStatistic const & s) const
	{
		PG_ColourStatistic retval;
		retval.n = n + s.n;
		retval.sum = sum + s.sum;
		retval.sum_sq = sum_sq + s.sum_sq;
		return retval;
	}

	inline PG_Pixel Mean(void) const
	{
		if (n)
			return sum/(float)n+PG_RGBData<float>(0.5,0.5,0.5);
		else
			return PG_Pixel(0,0,0);
	}

	inline PG_RGBData<double> Sum_D_Sq(void) const
	{
		if (n)
			return sum_sq - (PG_RGBData<double>)sum * sum / n;
		else
			return PG_RGBData<double>(0,0,0);
	}
		
	inline PG_RGBData<double> Sum_D_Sq(PG_RGBData<float> const & mean) const
	{
		return sum_sq - (PG_RGBData<double>)mean*sum*2 + (PG_RGBData<double>)mean*mean*n;
	}
		
};

enum PG_Primary { PG_RED, PG_GREEN, PG_BLUE, PG_N_PRIMARIES };

class PG_NodeData
{
protected:
#ifdef COUNT_NODES_ALLOCATED
	static size_t memory_allocated;
	static int num_allocated_nodes;
#endif

	PG_ColourStatistic helpful;
	PG_ColourStatistic * important;

public:
	PG_NodeData() : important(0) {}
	PG_NodeData(PG_NodeData const & s) : helpful(s.helpful), important(0)
	{
		if (s.important)
		{
#ifdef COUNT_NODES_ALLOCATED
			memory_allocated += sizeof(PG_ColourStatistic);
#endif
			important = new PG_ColourStatistic(*s.important);
		}
	}
	PG_NodeData & operator = (PG_NodeData const & s)
	{
		if (&s != this)
		{
			helpful = s.helpful;
			if (important)
			{
#ifdef COUNT_NODES_ALLOCATED
				memory_allocated -= sizeof(PG_ColourStatistic);
#endif
				delete important;
			}
#ifdef COUNT_NODES_ALLOCATED
			if (s.important)
				memory_allocated += sizeof(PG_ColourStatistic);
#endif
			important = s.important ? new PG_ColourStatistic(*s.important) : 0;
		}
		return *this;
	}
	virtual ~PG_NodeData()
	{
		if (important)
		{
#ifdef COUNT_NODES_ALLOCATED
			memory_allocated -= sizeof(PG_ColourStatistic);
#endif
			delete important;
		}
	}
	
	inline PG_NodeData & operator += (PG_NodeData const & s)
	{
		helpful += s.helpful;
		if (s.important)
		{
			if (important)
				*important += *s.important;
			else
			{
#ifdef COUNT_NODES_ALLOCATED
				memory_allocated += sizeof(PG_ColourStatistic);
#endif
				important = new PG_ColourStatistic(*s.important);
			}
		}
		return *this;
	}
	inline PG_NodeData & operator -= (PG_NodeData const & s)
	{
		helpful -= s.helpful;
		if (s.important)
		{
			if (!important)
			{
#ifdef COUNT_NODES_ALLOCATED
				memory_allocated += sizeof(PG_ColourStatistic);
#endif
				important = new PG_ColourStatistic;
			}
			*important -= *s.important;
			if (!important->n)
			{
#ifdef COUNT_NODES_ALLOCATED
				memory_allocated -= sizeof(PG_ColourStatistic);
#endif
				delete important;
				important = 0;
			}
		}
		return *this;
	}
	inline PG_NodeData operator + (PG_NodeData const & s) const
	{
		PG_NodeData retval;
		retval.helpful = helpful + s.helpful;
		if (important)
		{
#ifdef COUNT_NODES_ALLOCATED
			memory_allocated += sizeof(PG_ColourStatistic);
#endif
			retval.important = new PG_ColourStatistic(*important);
			if (s.important)
				*retval.important += *s.important;
		}
		else if (s.important)
		{
#ifdef COUNT_NODES_ALLOCATED
			memory_allocated += sizeof(PG_ColourStatistic);
#endif
			retval.important = new PG_ColourStatistic(*s.important);
		}
		return retval;
	}
	inline operator unsigned int (void) { return helpful.n + (important ? important->n : 0); }
	inline unsigned int Important(void) { return important ? important->n : 0; }

protected:
	inline void Reset(void)
	{
		helpful.Reset();
		if (important)
		{
#ifdef COUNT_NODES_ALLOCATED
			memory_allocated -= sizeof(PG_ColourStatistic);
#endif
			delete important;
			important = 0;
		}
	}

public:
	inline PG_Pixel Mean(void) const
	{
		if (important)
			if (important->n)
				return important->Mean();
		return helpful.Mean();
	}

	friend class PG_Variance;
	friend class PG_ColourOcTreeLeaf;
};


class PG_SplitNodeData
{
public:
	PG_NodeData cumul_low, cumul_high;

	PG_SplitNodeData() {}
	PG_SplitNodeData(PG_NodeData const & low, PG_NodeData const & high) : cumul_low(low), cumul_high(high) {}
};

class PG_TreeNodeData : public PG_NodeData
{
protected:
	virtual int Insert_Rc(PG_Pixel const &, int const, int const, PG_Pixel &) = 0;

	virtual void Store_Rc(int const, int const, PG_Pixel &) = 0;

	virtual void Correct_Statistics_Rc(PG_Pixel const &, int) = 0;

	virtual void Correct_Counts(void) = 0;

	virtual void Remove_Rc(PG_Pixel const &, PG_Pixel &) = 0;

	virtual PG_NodeData StatisticForRange_Rc(PG_Pixel const &, PG_Primary const, int const, int const, int) const = 0;

	virtual int Num_Colours(void) const = 0;

	virtual PG_PixelCount Any_Colour_Rc(int) const = 0;

	virtual PG_PixelCount Remove_Any_Colour_Rc(PG_Pixel const &, int) = 0;

	friend class PG_ColourOcTreeLeaf;
	friend class PG_ColourOcTreeNode;
};


class PG_ColourOcTreeLeaf : public PG_TreeNodeData
{
protected:
	unsigned int leafcounts [1<<PG_N_PRIMARIES];
	#define PGL_IMPORTANT 0x80000000

public:
	PG_ColourOcTreeLeaf()
	{
		InitializeChildren();
#ifdef COUNT_NODES_ALLOCATED
		++ num_allocated_nodes;
		memory_allocated += sizeof(PG_ColourOcTreeLeaf);
		if (!(num_allocated_nodes & 0x3fff))
			your_printf("PG_Alloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
	~PG_ColourOcTreeLeaf()
	{
#ifdef COUNT_NODES_ALLOCATED
		-- num_allocated_nodes;
		memory_allocated -= sizeof(PG_ColourOcTreeLeaf);
		if ((num_allocated_nodes & 0x3fff)==0x2000)
			your_printf("PG_Dealloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
private:
	void MakeBySplitting( PG_Pixel const & base, PG_ColourOcTreeLeaf & tree, PG_Primary const p, int const split);

	// operator = not supported
	inline PG_ColourOcTreeLeaf & operator = (PG_ColourOcTreeLeaf const &) { return *this; }

protected:
	inline void InitializeChildren(void)
	{
		unsigned int * ptr = leafcounts;
		for (int i=1<<PG_N_PRIMARIES; i; --i, ++ptr) *ptr = 0;
	}

	PG_ColourOcTreeLeaf( PG_Pixel const & base, PG_ColourOcTreeLeaf & tree, PG_Primary const p, int const split)
	{
		InitializeChildren();
		MakeBySplitting(base,tree,p,split);
#ifdef COUNT_NODES_ALLOCATED
		++ num_allocated_nodes;
		memory_allocated += sizeof(PG_ColourOcTreeLeaf);
		if (!(num_allocated_nodes & 0x3fff))
			your_printf("PG_Alloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}

	virtual int Insert_Rc(PG_Pixel const & p, int const count, int const depth, PG_Pixel & bits);

	virtual void Store_Rc(int const count, int const depth, PG_Pixel & bits);

	virtual void Correct_Statistics_Rc(PG_Pixel const & base, int depth);

	virtual void Correct_Counts(void);

	virtual void Remove_Rc(PG_Pixel const & p, PG_Pixel & bits);

	virtual PG_NodeData StatisticForRange_Rc(PG_Pixel const & base, PG_Primary const p, int const start, int const end, int depth) const;

	virtual int Num_Colours(void) const;

	virtual PG_PixelCount Any_Colour_Rc(int depth) const;

	virtual PG_PixelCount Remove_Any_Colour_Rc(PG_Pixel const & base, int depth);

	friend class PG_ColourOcTreeNode;
	friend class PG_ColourOcTree;
};


class PG_ColourOcTreeNode : public PG_TreeNodeData
{
protected:
	PG_TreeNodeData * children [1<<PG_N_PRIMARIES];

public:
	PG_ColourOcTreeNode()
	{
		InitializeChildren();
#ifdef COUNT_NODES_ALLOCATED
		++ num_allocated_nodes;
		memory_allocated += sizeof(PG_ColourOcTreeNode);
		if (!(num_allocated_nodes & 0x3fff))
			your_printf("PG_Alloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
	~PG_ColourOcTreeNode()
	{
		DeleteChildren();
#ifdef COUNT_NODES_ALLOCATED
		-- num_allocated_nodes;
		memory_allocated -= sizeof(PG_ColourOcTreeNode);
		if ((num_allocated_nodes & 0x3fff)==0x2000)
			your_printf("PG_Dealloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
private:
	void MakeBySplitting( PG_Pixel const & base, PG_ColourOcTreeNode & tree, PG_Primary const p, int const split, int depth);

	// operator = not supported
	inline PG_ColourOcTreeNode & operator = (PG_ColourOcTreeNode const &) { return *this; }

protected:
	inline void InitializeChildren(void)
	{
		PG_TreeNodeData * * ptr = children;
		for (int i=1<<PG_N_PRIMARIES; i; --i, ++ptr) *ptr = 0;
	}
	inline void DeleteChildren(void)
	{
		PG_TreeNodeData * * ptr = children;
		for (int i=1<<PG_N_PRIMARIES; i; --i, ++ptr) if (*ptr) { delete *ptr; *ptr=0; }
	}

	PG_ColourOcTreeNode( PG_Pixel const & base, PG_ColourOcTreeNode & tree, PG_Primary const p, int const split, int const depth)
	{
		InitializeChildren();
		MakeBySplitting(base,tree,p,split,depth);
#ifdef COUNT_NODES_ALLOCATED
		++ num_allocated_nodes;
		memory_allocated += sizeof(PG_ColourOcTreeNode);
		if (!(num_allocated_nodes & 0x3fff))
			your_printf("PG_Alloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
	
	virtual int Insert_Rc(PG_Pixel const & p, int const count, int const depth, PG_Pixel & bits);

	virtual void Store_Rc(int const count, int const depth, PG_Pixel & bits);

	virtual void Correct_Statistics_Rc(PG_Pixel const & base, int depth);

	virtual void Correct_Counts(void);

	virtual void Remove_Rc(PG_Pixel const & p, PG_Pixel & bits);

	virtual PG_NodeData StatisticForRange_Rc(PG_Pixel const & base, PG_Primary const p, int const start, int const end, int depth) const;

	virtual int Num_Colours(void) const;

	virtual PG_PixelCount Any_Colour_Rc(int depth) const;

	virtual PG_PixelCount Remove_Any_Colour_Rc(PG_Pixel const & base, int depth);

	friend class PG_ColourOcTree;
};


class PG_ColourOcTree : public PG_ColourOcTreeNode
{
public:
	int starts[PG_N_PRIMARIES];
	int ends[PG_N_PRIMARIES];

	PG_ColourOcTree()
	{
		InitializeRange(PG_COLOURDEPTHBITS);
#ifdef COUNT_NODES_ALLOCATED
		++ num_allocated_nodes;
		memory_allocated += sizeof(PG_ColourOcTree) - sizeof(PG_ColourOcTreeNode);
		if (!(num_allocated_nodes & 0x3fff))
			your_printf("PG_Alloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
	~PG_ColourOcTree()
	{
#ifdef COUNT_NODES_ALLOCATED
		-- num_allocated_nodes;
		memory_allocated -= sizeof(PG_ColourOcTree) - sizeof(PG_ColourOcTreeNode);
		if ((num_allocated_nodes & 0x3fff)==0x2000)
			your_printf("PG_Dealloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}
private:
	void MakeBySplitting( PG_Pixel const & base, PG_ColourOcTree & tree, PG_Primary const p, int const split, int depth);

	int BestSplit(PG_Primary const p, int start, int end, PG_SplitNodeData const * start_stats, PG_SplitNodeData const * end_stats);
	
	// operator = not supported
	inline PG_ColourOcTree & operator = (PG_ColourOcTree const &) { return *this; }

private:
	inline void InitializeRange(int const depth)
	{
		//int const end = (1<<depth);
		for (int i=0; i<PG_N_PRIMARIES; ++i)
		{
			starts[i]=0;
			ends[i]=0; // impossible, but insertion of colours will adjust range
		}
	}

public:
	PG_ColourOcTree( PG_ColourOcTree & tree, PG_Primary const p, int const split)
	{
		InitializeChildren();
		MakeBySplitting(PG_Pixel(0,0,0),tree,p,split,PG_COLOURDEPTHBITS);
#ifdef COUNT_NODES_ALLOCATED
		++ num_allocated_nodes;
		memory_allocated += sizeof(PG_ColourOcTree) - sizeof(PG_ColourOcTreeNode);
		if (!(num_allocated_nodes & 0x3fff))
			your_printf("PG_Alloc_Info: %d nodes allocated (%uK)\n",num_allocated_nodes,memory_allocated/1024);
#endif
	}

public:
	void Insert(PG_Pixel const & p, int const count);

	void Store(PG_Pixel const & p, int const count);

	void Correct_Statistics(void) { PG_ColourOcTreeNode::Correct_Statistics_Rc(PG_Pixel(0,0,0),PG_COLOURDEPTHBITS); }

	void Correct_Counts(void) { PG_ColourOcTreeNode::Correct_Counts(); }

	void Remove(PG_Pixel const & p);

	PG_NodeData StatisticForRange(PG_Primary const p, int const start, int const end) const;

	int BestSplit(PG_Primary const p) { return BestSplit(p,starts[p]+1,ends[p]-1,0,0); }

	int Num_Colours(void) const { return PG_ColourOcTreeNode::Num_Colours(); }

	PG_PixelCount Any_Colour(void) const { return PG_ColourOcTreeNode::Any_Colour_Rc(PG_COLOURDEPTHBITS); }

	PG_PixelCount Remove_Any_Colour(void) { return PG_ColourOcTreeNode::Remove_Any_Colour_Rc(PG_Pixel(0,0,0),PG_COLOURDEPTHBITS); }

	unsigned Compute_Palette(PG_Pixel * palette, unsigned const ncols);
};


class PG_Variance
{
private:
	int cnt_imp;
	double vari;
protected:
	PG_Primary var_worst;
public:
	int operator < (PG_Variance const & g) const;
	inline PG_Variance & operator += (PG_Variance const & a)
	{
		if (cnt_imp > 1)
		{
			if (a.cnt_imp > 1)
				vari += a.vari;
		}
		else
		{
			if (a.cnt_imp > 1)
				*this = a;
			else
				vari += a.vari;
		}
		return *this;
	}
	PG_Variance(PG_NodeData const & s);
};	


class PG_ColourOcTree_Header : public PG_Variance
{
private:
	PG_ColourOcTree * tree;

public:
	PG_ColourOcTree_Header(PG_ColourOcTree * const p) : PG_Variance(*p), tree(p) {}
	inline operator PG_ColourOcTree * (void) const { return tree; }
	inline operator PG_Primary (void) const { return var_worst; }
};


#endif //_TOOLS_PALETIZE_H_
