//
// Tools_ContextMenu_System.h
//
// Created by Richard Rice - 30/10/2007
// 
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//


#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif


#ifndef _TOOLS_CONTEXTMENU_SYSTEM_H_
#define _TOOLS_CONTEXTMENU_SYSTEM_H_

#include "Core/Asura_String.h"

#include "Misc/list_tem.hpp"

/*
Context menu system that is roughly based on the menus in Max

Right clicking will cause a menu to appear with up to 4 main sections (one per direction) , 
which are currently Display,Selection,Mode and Object.

Right clicking on an overlay will cause items to be added for that overlay if an appropriate
callback has been added using Asura_Overlay::SetContextMenuCallback()

Right clicking an object will add options for that object via Tools_Object::AddContextMenuItems().

Also regardless of where the click is done , extra items can be added for the current mode via Tools_Mode::AddContextMenuItems()

In each case the callback should use Tools_ContextMenu::GetSection() to get the menu section to add menu items to , and then
create instances of Tools_ContextMenu_Item_Function , Tools_ContextMenu_Item_Tick and Tools_ContextMenu_Item_Submenu to add
to this section (or to a submenu).

If a mode is using the right mouse button for something else , it can use Tools_ContextMenu_System::SetRightClickUsed() to
prevent the menus from being opened as a result of clicks on that frame.

RAR 7/11/2007
*/

class Tools_ContextMenu;
class Tools_ContextMenu_Section;
class Tools_ContextMenu_Item;
class Tools_ContextMenu_Item_SectionHeading;
struct Tools_ContextMenu_DefaultAction;

class Tools_ContextMenu_System
{
public :
	//Per-frame update for the system. Will run the menus if the
	//right button is pressed.
	static void Update();

	//Prevents right click from being used by context menus this frame.
	//For existing systems that already use it for other purposes.
	static void SetRightClickUsed(){s_bRightClickUsed = true;}

	//for dealing with the list of default actions to trigger when clicking on the main headings
	static void AddToDefaultActionList(Asura_Hash_ID uHeadingID, Asura_Hash_ID uTargetID);
	static Asura_Hash_ID GetDefaultAction(Asura_Hash_ID uHeadingID);

private :
	static List<Tools_ContextMenu_DefaultAction> s_xDefaultActionList;

	static bool s_bRightClickUsed;

};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//Class passed to the menu population functions to help them determine the context
class Tools_ContextMenu_Context
{
public :
	Tools_ContextMenu_Context()
	{
		m_pxOverlay = 0;
		m_bShiftHeld = false;
		m_bControlHeld = false;
		m_bAltHeld = false;
		m_pOverlayUserData = 0;
	}
	class Asura_Overlay* m_pxOverlay;//the overlay clicked on
	bool m_bShiftHeld;
	bool m_bControlHeld;
	bool m_bAltHeld;

	void* m_pOverlayUserData;//pointer passed into Asura_Overlay::SetContextMenuCallback() if clicking on an overlay
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//the 4 main sections available in the context menu
enum
{
	TOOLS_CONTEXTMENU_SECTION_SELECTION,
	TOOLS_CONTEXTMENU_SECTION_DISPLAY,
	TOOLS_CONTEXTMENU_SECTION_OBJECT,
	TOOLS_CONTEXTMENU_SECTION_MODE,
};

//Class containing the whole context menu.
class Tools_ContextMenu
{
public :
	Tools_ContextMenu();
	~Tools_ContextMenu();

	//Get/create the specified menu section , using values from the enumeration above
	Tools_ContextMenu_Section* GetSection(int iSection);


	//returns true while displaying any menu
	static bool IsMenuActive() {return s_bMenuActive;}



	////////////////////////////////////////////////////////////////////
	//Mainly internal menu system stuff
	////////////////////////////////////////////////////////////////////

	//Runs the menu , and quits when the user clicks somewhere
	void ShowMenu();

	void Update();
	void Render(const Tools_ContextMenu& xContextMenu);

	//determines the width and height of all the bits of menu
	void CalculateSize();

	void SetPosition(int iCursorX,int iCursorY);

	void InitialiseBeforeDisplay();

	//returns true if anything has been added to any of the sections other than the default
	//section heading menu items
	bool HasAnyEntries() const;

	const Tools_ContextMenu_Section* GetCurrentSection () const { return m_pxCurrentSection; }

private :
	void RestoreCursorPosition();

	List<Tools_ContextMenu_Section*> m_xSectionList;

	//position where the child sections meet
	int m_iPosX;
	int m_iPosY;

	//for undoing the offset to the cursor when the menu closes
	int m_iCursorOffsetX;
	int m_iCursorOffsetY;

	bool m_bReadyToClose;

	//menu section that the cursor has most recently been over
	Tools_ContextMenu_Section* m_pxCurrentSection;
	Tools_ContextMenu_Item* m_pxClickedItem;

	const char* m_szHelpText;

	static bool s_bMenuActive;
};


///////////////////////////////////////////////////////////////////////////////////////////////////////////

//one 'layer' of the context menu
class Tools_ContextMenu_Section
{
public :
	Tools_ContextMenu_Section(const char* szName, bool bCreateHeadingItem);
	~Tools_ContextMenu_Section();

	//Add a menu item to this section
	void AddItem(Tools_ContextMenu_Item* pxItem);
	
	//Causes a separator to be added before the next menu item
	void AddSeparator() {m_bNeedToAddSeparator = true;}

	//Allows the name of the section heading to be changed from the default
	//Only applies to the 4 top level sections in the menus
	void SetHeading(const char* szHeading);
	
	////////////////////////////////////////////////////////////////////
	//Mainly internal menu system stuff
	////////////////////////////////////////////////////////////////////
	
	void Render(const Tools_ContextMenu& xContextMenu);

	const char* GetName() const {return m_xName.GetString();}

	//recursively calculates the sizes of items in the menu
	void CalculateSize();

	//Sets the corner position.. exactly which depends on which
	//direction the menu goes
	void SetPosition(int iX,int iY);

	void SetCurrentSection(bool bSet);

	//get the dimensions in pixels
	int GetWidth() const {return m_iWidth;}
	int GetHeight() const {return m_iHeight;}

	//Determines which way the menu items are positioned
	void SetExpandsRight(bool bSet) {m_bRight = bSet;}
	void SetExpandsDown(bool bSet) {m_bDown = bSet;}

	bool GetExpandsRight() const {return m_bRight;}
	bool GetExpandsDown() const {return m_bDown;}

	//does the menu section contain this point
	bool Contains(int iX, int iY) const;

	//Find the menu item under the mouse , opening/closing submenus as needd
	Tools_ContextMenu_Item* CheckCursorPosition(int iX, int iY);

	//Gets the ID of section heading menu item
	Asura_Hash_ID GetHeadingID() const;

	//Sets the default action to use when the heading is clicked on , and colours things appropriately
	void SetupDefaultActionForHeading();
	//finds the item with the given ID , and flags the items in the path to it so
	//as to use the default action colour
	Tools_ContextMenu_Item* FindDefaultAction(Asura_Hash_ID uTargetID);

	//resets various bits and pieces before the menu gets displayed
	void InitialiseBeforeDisplay();

	int GetNumItems() const {return m_xItemList.size();}

private :
	Asura_String m_xName;

	bool m_bNeedToAddSeparator;

	//top left of menu section
	int m_iPosX;
	int m_iPosY;

	int m_iWidth;
	int m_iHeight;

	bool m_bRight;
	bool m_bDown;

	Tools_ContextMenu_Item_SectionHeading* m_pxHeading;

	List<Tools_ContextMenu_Item*> m_xItemList;

	Tools_ContextMenu_Item* m_pxCurrentItem;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//base class for an entry (one line) in the context menu
class Tools_ContextMenu_Item
{
public :
	Tools_ContextMenu_Item(const char* szName);
	virtual ~Tools_ContextMenu_Item() {}

	//allows a menu item to be greyed out
	void SetEnabled(bool bEnabled) {m_bEnabled = bEnabled;}
	bool IsEnabled() const {return m_bEnabled;}

	//some sort of tool tip thing , (to be implemented)
	void SetHelpText(const char* szText){m_xHelpText = szText;}
	const char* GetHelpText() const {return m_xHelpText.GetString();}

	//Sets an alternative name to use for generating the ID
	//For menu items where the text can vary depending on exactly what is clicked on , but
	//where they want to be considered the same for the purpose of clicking on the section heading
	void SetID(const char* szIDName);

	////////////////////////////////////////////////////////////////////
	//Mainly internal menu system stuff
	////////////////////////////////////////////////////////////////////

	virtual void Render(const Tools_ContextMenu& xContextMenu);
	virtual void RenderSubmenus(){};

	virtual bool IsHeading() { return false; }

	virtual void SetCurrentItem(bool bSet){m_bHighlight = bSet;}
	bool IsHighlight() const {return m_bHighlight;}
	
	//recursively calculates the sizes of items in the menu
	virtual void CalculateSize(){};

	//set the top-left position of this menu item
	void SetPosition(int iX, int iY);

	int GetPosX() const {return m_iPosX;}
	int GetPosY() const {return m_iPosY;}

	//get/set the dimensions in pixels
	void SetWidth(int iWidth);
	void SetHeight(int iHeight) {m_iHeight = iHeight;}

	int GetWidth() const {return m_iWidth;}
	int GetMinWidth() const {return m_iMinWidth;}
	int GetHeight() const {return m_iHeight;}

	bool Contains(int iX, int iY) const;

	//Searches this and children for the item that the cursor is currently over , and 
	//opens/closes submenus as needed.
	virtual Tools_ContextMenu_Item* CheckCursorPosition(int iX, int iY);

	void SetParentSection(Tools_ContextMenu_Section* pxParent);
	Tools_ContextMenu_Section* GetParentSection() {return m_pxParent;}

	//get the colour to use as the background for this menu item's rectangle
	virtual u_int GetBackgroundColour() const;

	//Indicates whether clicking on this should result in the menu being closed and HandleClick() being called for it
	virtual bool CanBeClicked() const{return false;}

	//Called when this item is clicked on.
	virtual void HandleClick(){}

	Asura_Hash_ID GetID() const {return m_uID;}

	void SetupDefaultActionForHeading();

	//finds the item with the given ID , and flags the items in the path to it so
	//as to use the default action colour
	virtual Tools_ContextMenu_Item* FindDefaultAction(Asura_Hash_ID uTargetID);
	void SetDefaultAction(bool bSet) {m_bDefaultAction = true;}

	//resets various bits and pieces before the menu gets displayed
	virtual void InitialiseBeforeDisplay();

	//changes the name but not the ID
	void SetName(const char* szName);

	//determines width based on the name
	void CalculateWidth();

private :
	Tools_ContextMenu_Section* m_pxParent;

	Asura_String m_xName;
	Asura_Hash_ID m_uID;

	Asura_String m_xHelpText;

	bool m_bHighlight;
	bool m_bEnabled;
	bool m_bDefaultAction;//Used to colour the item that is the default action when clicking on a heading , and the submenus along the way
	
	int m_iPosX;
	int m_iPosY;

	int m_iMinWidth;
	int m_iWidth;
	int m_iHeight;

};


///////////////////////////////////////////////////////////////////////////////////////////////////////////

//Menu item that has another menu layer/section as a child, that will be shown when the cursor goes over it
class Tools_ContextMenu_Item_Submenu : public Tools_ContextMenu_Item
{
	typedef Tools_ContextMenu_Item PARENT;

public :
	Tools_ContextMenu_Item_Submenu(const char* szName);
	~Tools_ContextMenu_Item_Submenu();

	//Adds a new menu item to this submenu
	void AddItem(Tools_ContextMenu_Item* pxItem) {m_pxMenuSection->AddItem(pxItem);}

	//Causes a separator to be added before the next menu item
	void AddSeparator() {m_pxMenuSection->AddSeparator();}

protected :
	virtual void Render(const Tools_ContextMenu& xContextMenu);
	virtual void RenderSubmenus(const Tools_ContextMenu& xContextMenu);

	virtual void SetCurrentItem(bool bSet);
	
	virtual void CalculateSize();

	virtual Tools_ContextMenu_Item* CheckCursorPosition(int iX, int iY);

	virtual Tools_ContextMenu_Item* FindDefaultAction(Asura_Hash_ID uTargetID);

	virtual void InitialiseBeforeDisplay();
private :

	bool m_bSubmenuOpen;	
	Tools_ContextMenu_Section* m_pxMenuSection;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//The main type of menu item , that will call some function when clicked on.
class Tools_ContextMenu_Item_Function : public Tools_ContextMenu_Item
{
public :
	//for callback with no arguments
	Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(void));
	//for callback with void*
	Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(void*), void* pUserData);
	//for callback with void* and u_int
	Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(void*,u_int), void* pUserData, u_int uUserData);
	//for callback with int
	Tools_ContextMenu_Item_Function(const char* szName, void(*pfnFunction)(int), int iUserData);

protected :

	virtual bool CanBeClicked() const {return IsEnabled();}	
	virtual void HandleClick();

private :
	void InitialiseFunctionPointers();

	void (*m_pfnFunction)(void);	
	void (*m_pfnFunction_Ptr)(void *);	
	void (*m_pfnFunction_PtrUint)(void *, u_int);	
	void (*m_pfnFunction_Int)(int);	

	void* m_pUserData;
	u_int m_uUserData;
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//A 'tickbox' menu item.
//Currently just a variety that modifies a bool. Other variations can be done as and when this actually gets used
class Tools_ContextMenu_Item_Tick  : public Tools_ContextMenu_Item
{
	typedef Tools_ContextMenu_Item PARENT;
public :
	//pbTicked is the bool to read/write the current tickedness from
	//pbChanged gets set to true when the menu item is clicked
	Tools_ContextMenu_Item_Tick(const char* szName, bool* pbTicked, bool* pbChanged = 0);

protected :
	virtual bool CanBeClicked() const {return IsEnabled();}	
	virtual void HandleClick();

	virtual void Render(const Tools_ContextMenu& xContextMenu);


private :	
	bool m_bTicked;

	bool* m_pbTicked;
	bool* m_pbChanged;



};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//Seperator bar on the menu.
//Should probably just be created via Tools_ContextMenu_Section::AddSeparator()
//and Tools_ContextMenu_Item_Submenu::AddSeparator()

class Tools_ContextMenu_Item_Separator : public Tools_ContextMenu_Item
{

public :
	Tools_ContextMenu_Item_Separator();

	virtual void SetCurrentItem(bool bSet){}
	virtual void Render(const Tools_ContextMenu& xContextMenu);

};

///////////////////////////////////////////////////////////////////////////////////////////////////////////

//Special heading that when clicked on will duplicate the last function used from that section
//(generated automatically by the top level sections , and shouldn't be created beyond that).
class Tools_ContextMenu_Item_SectionHeading : public Tools_ContextMenu_Item
{
public :
	Tools_ContextMenu_Item_SectionHeading(const char* szName);

	void SetDefaultAction(Tools_ContextMenu_Item* pxItem) {m_pxDefaultActionItem = pxItem;}

	virtual bool IsHeading() { return true; }

protected :
	virtual u_int GetBackgroundColour() const;
	virtual bool CanBeClicked() const {return m_pxDefaultActionItem ? m_pxDefaultActionItem->CanBeClicked() : false;}	
	virtual void HandleClick(){if(m_pxDefaultActionItem) m_pxDefaultActionItem->HandleClick();}

private :
	//the item to be triggered when this is clicked on
	Tools_ContextMenu_Item* m_pxDefaultActionItem;



};


///////////////////////////////////////////////////////////////////////////////////////////////////////////

//For internal use , for tracking what actions should be done when clicking on the section headings
struct Tools_ContextMenu_DefaultAction
{
	Asura_Hash_ID m_uHeadingID;//ID of heading this belongs with
	Asura_Hash_ID m_uTargetID;//ID of target button to trigger
};

#endif //_TOOLS_CONTEXTMENU_SYSTEM_H_