//
// Tools_Object_LOD.cpp
//
// Created by Richard Rice - 26/09/2002
// 
// Copyright (C) 2002 Rebellion, All Rights Reserved.
//


#include "Core/Asura.h"
#include "Debugging/Asura_Assert.h"

#include "LevelOfDetail/Tools_Object_LOD.h"

#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Text_Output.h"
#include "Camera/Asura_Camera.h"

#include "Tools_Overlays.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Load&Save/Tools_SaveAsura.h"


Asura_Vector_2 Tools_Object_LOD::s_xOverlayPosition(10,10);
Tools_Object_LOD* Tools_Object_LOD::s_pxThis = 0;

Tools_Object_LOD::Tools_Object_LOD(Asura_Mesh* pxMesh,const Asura_Vector_3* pxPosition)
{
	Asura_Assert(pxMesh,"");

	Mesh = pxMesh;
	Type = TOT_LevelOfDetail;
	Position = *pxPosition;
	Orientation.Identity();

	m_bAutoLOD = false;
	m_bDetailLevelChanged = false;
	m_iDetailLevel = 0;
}

const char* Tools_Object_LOD::GetName()
{
	return Mesh->Name.GetString();
}

void Tools_Object_LOD::PrintInfo(float& x,float& y)
{
	Tools_Object::PrintInfo(x,y);

	//show current distance away
	float fDistance = (*Asura_Camera::GetPosition() - Position).Magnitude();
	Asura_Text_Output::Print(x,y++,0xffffffff,0,"Current Distance : %.3f",fDistance);

	//and current min visible distance
	Asura_Mesh* pxMesh = Mesh->GetDetailMeshFromIndex(m_iDetailLevel);
	Asura_Text_Output::Print(x,y++,0xffffffff,0,"Min Vis Distance : %.3f",pxMesh->m_fMinDistance);




	
}

void Tools_Object_LOD::OnSelection()
{
	s_pxThis = this;
	CreateOverlay();
}

void Tools_Object_LOD::OnDeselection()
{
	DeleteOverlay();
	s_pxThis = 0;
}


void Tools_Object_LOD::CreateOverlay()
{
	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay(GetTypeDescription(),false,&s_xOverlayPosition);
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);


	pxDialog = new Asura_Numerical_Dialog("Detail Level",xOffset,&m_iDetailLevel,0,Mesh->GetNumberOfDetailLevels()-1,1,"%d",&m_bDetailLevelChanged);
	pxDialog->SetHelpText("Set detail level to view/edit.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Autochange LOD",xOffset,&m_bAutoLOD);
	pxDialog->SetHelpText("Automatically set current detail level according to distance from shape.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Set Distance",xOffset,SetMinimumVisibleDistance);
	pxDialog->SetHelpText("Minimum distance at which this detail level is used.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	
	pxOverlay->m_uID = TOOLS_OVERLAY_SELECTEDOBJECT;

 	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}

void Tools_Object_LOD::DeleteOverlay()
{
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SELECTEDOBJECT);
}

void Tools_Object_LOD::UpdateWhileSelected()
{
	if(m_bDetailLevelChanged)
	{
		//if the detail level is changed by hand, stop changing it accoding to distance
		m_bAutoLOD = false;
		m_bDetailLevelChanged = false;
	}	
}


Asura_Mesh* Tools_Object_LOD::GetLODMesh()
{
	if(!m_bAutoLOD || m_bDetailLevelChanged)
	{
		return Mesh->GetDetailMeshFromIndex(m_iDetailLevel);
	}

	Asura_Mesh* pxDetailMesh = Tools_Object::GetLODMesh();
	//work out which index mesh this is
	m_iDetailLevel = 0;
	Asura_Mesh* pxMesh = Mesh;
	while(pxMesh!=pxDetailMesh)
	{
		m_iDetailLevel++;
		pxMesh = pxMesh->m_pxNextLowerMesh;
		Asura_Assert(pxMesh,"");
	}
	return pxDetailMesh;
}

void Tools_Object_LOD::SetMinimumVisibleDistance()
{
	Asura_Assert(s_pxThis,"");

	if(s_pxThis->m_iDetailLevel==0)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"The distance for the highest detail level cannot be changed.");
		return;
	}

	//get the current distance
	float fDistance = (*Asura_Camera::GetPosition() - s_pxThis->Position).Magnitude();
	if(fDistance<0.1f) fDistance = 0.1f;

	//get an array of detail meshes
	Asura_Mesh* pxMainMesh = s_pxThis->Mesh;
	int iNumLevels = pxMainMesh->GetNumberOfDetailLevels();
	
	//set the distance for the current detail level
	Asura_Mesh* pxCurrentMesh = pxMainMesh->GetDetailMeshFromIndex(s_pxThis->m_iDetailLevel);
	pxCurrentMesh->m_fMinDistance = fDistance;

	//make sure lower details have a further distance
	for(int i=s_pxThis->m_iDetailLevel;i<iNumLevels-1;i++)
	{
		Asura_Mesh* pxMesh1 = pxMainMesh->GetDetailMeshFromIndex(i);
		Asura_Mesh* pxMesh2 = pxMesh1->m_pxNextLowerMesh;
		if(pxMesh1->m_fMinDistance>=pxMesh2->m_fMinDistance)
		{
			pxMesh2->m_fMinDistance = pxMesh1->m_fMinDistance+0.01f;
		}
	}

	//make sure higher detail levels have a nearer distance
	for(int i=s_pxThis->m_iDetailLevel;i;i--)
	{
		Asura_Mesh* pxMesh1 = pxMainMesh->GetDetailMeshFromIndex(i-1);
		Asura_Mesh* pxMesh2 = pxMesh1->m_pxNextLowerMesh;
		if(pxMesh1->m_fMinDistance>=pxMesh2->m_fMinDistance)
		{
			pxMesh1->m_fMinDistance = pxMesh2->m_fMinDistance-0.01f;
		}
	}

  	Tools_Timed_Message_System::AddMessage(0xffffff,"LOD Distance set to %.3f",fDistance);
	Tools_SaveAsura::SetMainFileModified();
}






