//
// Tools_LightSets.h
//
// Created by Richard Rice - 24/03/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef _TOOLS_LIGHTSETS_H_
#define _TOOLS_LIGHTSETS_H_

#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"

#include "Interface/Asura_Dialog.h"


//Maximum number of light sets that can be created.
//Because of the way lights indicate membership into the sets via bitfields ,this can't really be set higher than 32
//without a fair bit of work.
//(And changing this value at all would need some change to the chunk loaders).
#define TOOLS_MAX_LIGHT_SETS (32)

#define TOOLS_MAX_LIGHT_SET_GROUPS (2)

class Tools_LightSetGroup;

class Tools_LightSets
{
public :
	Tools_LightSets();
	~Tools_LightSets();

	static Tools_LightSets* GetThis() {return s_pxThis;}


	//initialise/shutdown for use from Asura_Core initialise/shutdown
	static void Initialise();
	static void Shutdown();

	//Initialise/s,hutdown used from Tools_System initialise/shutdown
	static void ToolsSystem_Initialise();
	static void ToolsSystem_Shutdown();

	static const char* GetLightSetNameByIndex(u_int uIndex)
	{
		Asura_Assert(uIndex<TOOLS_MAX_LIGHT_SETS,"Invalid light set");
		return s_axNames[uIndex].IsBlank() ? 0 : s_axNames[uIndex].GetString();
	}
	static const char* GetLightSetNameByValue(u_int uSet)
	{
		return GetLightSetNameByIndex(Asura_Maths::GetFloorLog2(uSet));
	}


	//Returns true if index corresponds to a light set
	static bool IsUsedLightSetIndex(u_int uSetIndex){return IsValidLightSetIndex(uSetIndex) && (s_uUsedLightSets & (1<<uSetIndex));}
	static bool IsValidLightSetIndex(u_int uSetIndex){return uSetIndex<TOOLS_MAX_LIGHT_SETS && !s_axNames[uSetIndex].IsBlank();}

	static bool IsUsedLightSetCombination(u_int uSets);

	static void SetUsedLightSets(u_int uSets);
	static u_int GetUsedLightSets(){return s_uUsedLightSets;}
	static u_int GetValidLightSets(){return s_uValidLightSets;}
	

	static void SetCurrentLightSets(u_int uSets,bool bRelight = true);
	static u_int GetCurrentLightSets() {return s_uCurrentLightSets;}

	//multiples the number of used light sets in each group to get the total number
	//of combinations in use
	static u_int GetNumUsedCombinations();

	//Returns true if an object with the given light set flags should belong to the current light set
	static bool IsMemberOfCurrentLightSets(u_int uSetFlags) {return (uSetFlags & s_uCurrentLightSets) == s_uCurrentLightSets;}

	static void GetCurrentLightSetNames(Asura_String& xOutput);

	#if 0
	//temp
	static void SetCurrentLightSet(u_int uSets,bool bRelight = true){}
	static u_int GetCurrentLightSet() {return 1;}

	static u_int GetCurrentLightSetBit() {return 1;}
	#endif

	//Returns true if some light set definitions have been read in for this project
	static bool IsUsingLightSets(){return s_bUsingLightSets;}

	static void ToggleOverlay();
	static bool OverlayIsOn();
	static void CloseOverlay();

	//Gets light sets to apply to a newly created object at the given position
	static u_int GetLightSetsForNewObject();
	static u_int GetLightSetsForNewObject(const Asura_Vector_3& xPosition);
	static u_int GetLightSetsForNewObject(int iModuleIndex);

	static bool ShouldHideModulesInOtherSets(){return s_bHideModulesInOtherSets;}
	static bool ShouldHideObjectsInOtherSets(){return s_bHideObjectsInOtherSets;}
	static bool IsModuleInCurrentLightSet(int iModuleIndex);

	//So GI mode can have its own tickbox for this
	static bool* GetHideObjectsInOtherSetsPtr() {return &s_bHideObjectsInOtherSets;}

	//access to groups of light sets
	static u_int GetNumGroups() {return s_uNumGroups;}
	static const Tools_LightSetGroup* GetGroup(u_int uIndex);

	static const char* GetGroupName(u_int uIndex);

	//returns true if an option for this light set should appear on the level export overlay
	static bool ShouldIncludeLightSetIndexInExportUI(u_int uIndex);

protected :
	//Allows project to add extra controls to the light set overlay
	virtual void AddOverlayControls(class Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset){};

private :
	static void TransferModeData(int iLoadOrSave);
	

	static void CreateOverlay();
	static void UsedLightSetsChanged();
	static bool IsDefaultingToSelectedSets(){return s_bDefaultSet_SelectedSets;}

	static u_int s_uNumGroups;
	static Tools_LightSetGroup s_axGroups[TOOLS_MAX_LIGHT_SET_GROUPS];

	static Asura_String s_axNames[TOOLS_MAX_LIGHT_SETS];
	static u_int s_uNumLightSets;

	static u_int s_uCurrentLightSets;

	static u_int s_uUsedLightSets;//bitfield of sets used by this level.
	static u_int s_uValidLightSets;//bitfield of light sets that could be used.

	static u_int s_uUsedLightSetsForOverlay;//opy of used light sets used by the tickboxes

	static bool s_bUsingLightSets;

	static bool s_bHideModulesInOtherSets;
	static bool s_bHideObjectsInOtherSets;

	//defaults to use for newly created objects
	static bool s_bDefaultSet_Current;
	static bool s_bDefaultSet_All;
	static bool s_bDefaultSet_NearModule;
	static bool s_bDefaultSet_SelectedSets;
	static u_int s_uDefaultLightSets;//default light sets for new objects if using s_bDefaultSet_SelectedSets

	static Asura_Vector_2 s_xOverlayPosition;
	static bool s_bOverlayShouldBeOpen;
	static u_int s_uOverlayMinimised;

	friend class Tools_UsedLightSetIterator;

	static Tools_LightSets* s_pxThis;

};

class Tools_LightSetGroup
{
public :
	Tools_LightSetGroup();

	void SetName(const char* szName) {m_xName = szName;}
	const char* GetName() {return m_xName.IsBlank() ? "Light Set" : m_xName.GetString();}

	void SetLightSets(u_int uSets);
	u_int GetLightSets() const {return m_uSets;}
	u_int GetUsedLightSets() const {return m_uUsedSets;}

	u_int GetNumValidSets() const {return m_uNumValidSets;}
	u_int GetValidSetByIndex(u_int uIndex) const {return m_auValidSets[uIndex];}

	u_int GetNumUsedSets() const {return m_uNumUsedSets;}
	u_int GetUsedSetByIndex(u_int uIndex) const {return m_auUsedSets[uIndex];}

	//determines whether this group should be iterated over when doing exports
	bool GetIncludeInExportUI() const {return m_bIncludeInExportUI;}
	void SetIncludeInExportUI(bool bSet) {m_bIncludeInExportUI = bSet;}

	//updates the record of the used light sets within this group
	void UpdateUsedLightSets();

	//Modifies the passed in light sets so that exactly one used light set from this group is set
	void EnsureExactlyOneUsedLightSetFromGroup(u_int& uSets) const;

private :
	Asura_String m_xName;

	u_int m_uSets;
	u_int m_uUsedSets;
	u_int m_uNumValidSets;
	u_int m_uNumUsedSets;

	bool m_bIncludeInExportUI;

	//the bit values of all the light sets that are part of this group
	u_int m_auValidSets[TOOLS_MAX_LIGHT_SETS];
	u_int m_auUsedSets[TOOLS_MAX_LIGHT_SETS];
};

enum TOOLS_LIGHTSETITERATOR_STYLE
{
	TOOLS_LIGHTSETITERATOR_DEFAULT,
	TOOLS_LIGHTSETITERATOR_EXPORT,//export iterator may ignore some of the light set groups (BF3 deals with one of its groups in project code)
};

//Iterates over the valid combinations for the current light sets
class Tools_UsedLightSetIterator
{
public :
	Tools_UsedLightSetIterator(u_int uStyle = TOOLS_LIGHTSETITERATOR_DEFAULT);

	u_int GetCurrent() const;

	//Gets the current value of the iteration , with bits in groups that aren't covered by the iteration masked out
	u_int GetMaskedCurrent() const {return m_uCurrentValue & m_uMask;}

	//moves to next combination
	void Next();
	//returns true if iteration has finished
	bool Done() const {return m_bDone;}
	//restarts iteration
	void Restart();

private :
	int m_iIndex;
	u_int m_uCurrentValue;
	u_int m_uStyle;
	u_int m_uMask;//light sets being covered by the iteration
	bool m_bDone;

};


//For creating a drop down list containing the light set names
//If puLightSets is null , then the control will directly set the current light sets in Tools_LightSets instead
class Tools_LightSet_Dialog : public Asura_DropDown_List_Dialog
{
public :
	Tools_LightSet_Dialog(const Asura_Vector_2& xOffset,u_int uGroup,u_int* puLightSets = 0,bool* pbChangedFlag = 0,const char* szName = NULL);

	void Update(const Asura_Vector_2& xOrigin);
	void PerFrameUpdate();

private :
	void UpdateIndexFromSet();
	void UpdateNameList(bool bUpdateDialog);

	u_int m_uGroup;

	u_int* m_puLightSets;
	bool* m_pbChangedFlag;

	u_int m_uLightSetsCopy;
	bool m_bChanged;
	int m_iSelectedIndex;

	int m_iNameListSize;
	const char* m_aszNameList[TOOLS_MAX_LIGHT_SETS];
	u_int m_auLightSetValue[TOOLS_MAX_LIGHT_SETS];
	u_int m_uUsedLightSets;

};


#endif //_TOOLS_LIGHTSETS_H_
