//
// Tools_PhotonMapping_LightSource.h
//
// Created by KJL - 15/01/2004 12:18:15
//
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __TOOLS_PHOTONMAPPING_LIGHTSOURCE_H__
#define __TOOLS_PHOTONMAPPING_LIGHTSOURCE_H__

#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Network/Asura_Guid.h"

//! Manages the properties of an individual Photon Mapping light source.
class Tools_PhotonMapping_LightSource
{
	public:
		Tools_PhotonMapping_LightSource()
		{
			ResetToDefaults();
		}
		virtual ~Tools_PhotonMapping_LightSource()
		{
		}

		void ResetToDefaults(void)
		{
			m_xPosition.Zero();
			m_xOrientation.Identity();
			
			m_xColour.x = m_xColour.y = m_xColour.z = 255.0f;
			m_fPower = 1.0f;
			m_uAccuracy = 1;
			m_fAngle = 360.0f;

			m_uPhotonMapValidity = 0;
			m_bHasPhotonMap = false;
			
			m_bParallel = false;
			m_bBleached = true;

			m_bShortRange = false;

			m_bUseRange = false;
			m_fInnerRange = 0.0f;
			m_fOuterRange = 0.0f;
			m_xName.MakeBlank();

			m_fReflectionChance = 0.25f;

			m_uLightSets = 1;

		}

		// Access functions for light parameters
		const Asura_Vector_3& GetPosition(void) const
		{
			return m_xPosition;
		}
		virtual void SetPosition(const Asura_Vector_3& xPosition)
		{
			m_xPosition = xPosition;
			m_uPhotonMapValidity = 0;
		}

		const Asura_Matrix_3x3& GetOrientation(void) const
		{
			return m_xOrientation;
		}
		virtual void SetOrientation(const Asura_Matrix_3x3& xOrientation)
		{
			m_xOrientation = xOrientation;
			m_uPhotonMapValidity = 0;
		}


		const Asura_Vector_3& GetColour(void) const
		{
			return m_xColour;
		}
		void SetColour(const Asura_Vector_3& xColour)
		{
			m_xColour = xColour;
		}

		float GetPower(void) const
		{
			return m_fPower;
		}
		void SetPower(const float fPower)
		{
			m_fPower = fPower;
		}

		u_int GetAccuracy(void) const
		{
			return m_uAccuracy;
		}
		void SetAccuracy(const u_int uAccuracy)
		{
			m_uAccuracy = uAccuracy;
		}

		const bool IsParallel(void) const
		{
			return m_bParallel;
		}
		void SetParallel(const bool b)
		{
			m_bParallel = b;
		}

		const bool IsBleached(void) const
		{
			return m_bBleached;
		}
		void SetBleached(const bool b)
		{
			m_bBleached = b;
		}

		const bool IsShortRange(void) const
		{
			return m_bShortRange;
		}
		void SetShortRange(const bool b)
		{
			if(b != (m_bShortRange ? true : false))
			{
				m_bShortRange = b;
				m_uPhotonMapValidity = 0;
			}
		}


		bool IsPhotonMapCompletelyValid(void) const
		{
			return m_uPhotonMapValidity >= m_uAccuracy;
		}

		bool IsPhotonMapPartiallyValid(void) const
		{
			return m_uPhotonMapValidity!=0;
		}
		
		//get accuracy level at which photon map is valid
		u_int GetPhotonMapValidity() const
		{
			return m_uPhotonMapValidity;
		}

		void SetPhotonMapValidity(u_int uAccuracy)
		{
			m_uPhotonMapValidity = uAccuracy;
		}

		//Indicates whether light has a (possibly invalid) photon map
		bool HasPhotonMap() const
		{
			return m_bHasPhotonMap;
		}

		void SetHasPhotonMap(const bool b)
		{
			m_bHasPhotonMap = b;
		}

		float GetAngle() const
		{
			return m_fAngle;
		}

		void SetAngle(float fAngle)
		{
			if(m_fAngle != fAngle)
			{
				m_fAngle = fAngle;
				m_uPhotonMapValidity = 0;
			}
		}

		float GetInnerRange() const
		{
			return m_fInnerRange;
		}

		void SetInnerRange(float fRange)
		{
			if(m_fInnerRange != fRange)
			{
				m_fInnerRange = fRange;
				m_uPhotonMapValidity = 0;
			}
		}

		float GetOuterRange() const
		{
			return m_fOuterRange;
		}

		void SetOuterRange(float fRange)
		{
			if(m_fOuterRange != fRange)
			{
				m_fOuterRange = fRange;
				m_uPhotonMapValidity = 0;
			}
		}

		bool GetUseRange() const
		{
			return m_bUseRange;
		}
		
		void SetUseRange(bool bUse)
		{
			if((m_bUseRange ? true: false) != bUse)
			{
				m_bUseRange = bUse;
				m_uPhotonMapValidity = 0;
			}
		}

		//gets the name (if one has been assigned by user)
		const char* GetName() const
		{
			return m_xName.GetString();
		}

		void SetName(const char* pszName)
		{
			m_xName = pszName ? pszName : "";
		}

		float GetReflectionChance() const
		{
			return m_fReflectionChance;
		}

		void SetReflectionChance(float fChance)
		{
			if(m_fReflectionChance != fChance)
			{
				m_fReflectionChance = fChance;
				m_uPhotonMapValidity = 0;
			}
		}

		void SetLightSets(u_int uLightSets)
		{
			m_uLightSets = uLightSets;
		}

		u_int GetLightSets() const
		{
			return m_uLightSets;		
		}

		virtual void SetStratificationLevels(u_int uNumberOfSamples);

		//! Returns a possible photon position and direction based on the light's position and shape
		virtual void GetStratifiedPhotonPositionAndDirection(Asura_Vector_3& xPosition, Asura_Vector_3& xDirection, u_int uX, u_int uY);

		virtual float GetStratifiedPhotonEnergy(void);

		//Used after a light has been created from the template's chunk buffers to adjust positions etc
		//as needed. RAR 27/04/2004
		void ApplyChangesForTemplateInstance(class Tools_EntityTemplateInstance* pxInstance);

		virtual bool IsPolygonLight(){return false;}

		//indicates whether this light should  be ignored for lighting purposes
		//(for handling polygon lights that lack polygons)
		virtual bool ShouldIgnoreLight(){return false;}

		
	protected:
		
		Asura_Vector_3		m_xPosition;

		Asura_Matrix_3x3	m_xOrientation;

		Asura_Vector_3		m_xColour;
							
		float				m_fPower;
		
		u_int				m_uAccuracy;

		float				m_fAngle;//angle of light in degrees

		u_int				m_uPhotonMapValidity;//accuracy level that has been calculated.. zero if invalid

		float				m_fInnerRange;
		float				m_fOuterRange;

		float				m_fReflectionChance;

		u_int				m_bParallel :1,	
							m_bBleached :1,  
							m_bHasPhotonMap : 1, 
							m_bShortRange :1, 
							m_bUseRange : 1, :0;

		u_int				m_uLightSets;//Bitfield indicating which lighting sets this is used by.

		Asura_String m_xName;//optional name override.

		


		friend class Asura_Chunk_GlobalIllumination;
};

//Value ORed in with the light IDs for lights that belong to template files
//(to distinguish from those that are just part of the test environment)
#define TOOLS_PHOTONMAPPING_TEMPLATELIGHT	0x80000000

//For marking lights that are to be saved out to the mission light file , instead
//of the common .gi file.
#define TOOLS_PHOTONMAPPING_MISSIONLIGHT	0x40000000

#define TOOLS_PHOTONMAPPING_IDMASK (~(TOOLS_PHOTONMAPPING_TEMPLATELIGHT|TOOLS_PHOTONMAPPING_MISSIONLIGHT))

class Tools_PhotonMapping_TemplateData
{
public :
	Tools_PhotonMapping_TemplateData()
	{
		m_uID = 0;
		m_uTemplateGuid = ASURA_GUID_UNREGISTERED;
		m_iTemplateVersion = 0;
	}

	void GenerateID()
	{
		m_uID = s_uNextID++;
	}

	bool IsTemplateLight() const {return (m_uID & TOOLS_PHOTONMAPPING_TEMPLATELIGHT) !=0;}
	bool IsMissionLight() const {return (m_uID & TOOLS_PHOTONMAPPING_MISSIONLIGHT) !=0;}

	u_int 				m_uID;// ID of this light
	Asura_Guid 			m_uTemplateGuid;//guid of template instance that owns this , if any
	int					m_iTemplateVersion;//version number of owning template instance

	static u_int		s_uNextID;
};


#endif//__TOOLS_PHOTONMAPPING_LIGHTSOURCE_H__
