//
// Tools_Photon_Object.cpp
//
// Created by Richard Rice - 12/03/2003
// 
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//


#include "Core/Asura.h"
#include "Lighting/Tools_Photon_Object.h"


#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Tools_ColourPicker.h"
#include "Light/Asura_Photon.h"
#include "Mesh/Asura_Mesh_System.h"


#include "Tools_Overlays.h"
#include "ChunkSystem/Tools_Chunk_ToolsLights.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Render.h"
#include "Objects/Tools_Object_System.h"


Asura_Vector_2 Tools_Photon_Object::s_xPropertiesOverlayPosition(10,10);

bool Tools_Photon_Object::s_bGeneralOverlayOpen = false;
u_int Tools_Photon_Object::s_uGeneralOverlayMinimised = 0;
Asura_Vector_2 Tools_Photon_Object::s_xGeneralOverlayPosition( 23.0f, 13.0f );

bool Tools_Photon_Object::s_bShowLights = true;


Tools_Photon_Object::Tools_Photon_Object(const Asura_Vector_3* pxPosition,const Asura_Matrix_3x3* pxOrientation)
:Tools_Object(TOT_ToolsPhoton)
{
	Position = *pxPosition;
	Orientation = *pxOrientation;

	Mesh = Asura_Mesh_System::GetMesh("ToolsShape_Light");


	//set defaults for the light values
	m_fRed = 1.0f;
	m_fGreen = 1.0f;
	m_fBlue = 1.0f;
	m_fBrightness = 1.0f;
	m_fRange = 20.0f;
	m_bActive = true;

	m_bChanged = false;

	//create the photon
	m_pxPhoton = Asura_Photon_System::CreatePhoton();
	if(!m_pxPhoton)
	{
		Tools_Timed_Message_System::AddMessage( 0xff0000,"Failed to allocate photon. This light won't work."); 
	}	

	//initialise the photon
	UpdatePhoton();
}

Tools_Photon_Object::~Tools_Photon_Object()
{
	if(m_pxPhoton) Asura_Photon_System::DestroyPhoton(m_pxPhoton);		
}


void Tools_Photon_Object::UpdatePositionInAsuraData(bool /*bUpdateLinkedObjects*/)
{
	//update the position in the asura photon
	UpdatePhoton();	
}

void Tools_Photon_Object::Render()
{
	Tools_Object::Render();

	if(Mesh)
	{
		//draw some of the sides using the colour of the light
		//get the colour to use
		int iRed = (int)(m_fRed*255.0f);
		int iGreen = (int)(m_fGreen*255.0f);
		int iBlue = (int)(m_fBlue*255.0f);
		int iColour = 0xff000000|(iRed<<16)|(iGreen<<8)|iBlue;

		Tools_Render::RenderPolygonOutline(Mesh,1,&Position,&Orientation,iColour,Tools_Render::POM_ForceSolid);
		Tools_Render::RenderPolygonOutline(Mesh,3,&Position,&Orientation,iColour,Tools_Render::POM_ForceSolid);
	}
}

void Tools_Photon_Object::UpdatePhoton()
{
	if(!m_pxPhoton) return;
	

	m_pxPhoton->SetPosition(Position);
	m_pxPhoton->SetColours(m_fRed*m_fBrightness,m_fGreen*m_fBrightness,m_fBlue*m_fBrightness);
	m_pxPhoton->SetRange(m_fRange);
	m_pxPhoton->SetAsActive(m_bActive);

	m_bChanged = false;
}

void Tools_Photon_Object::OnSelection()
{
	CreatePropertiesOverlay();
}

void Tools_Photon_Object::OnDeselection()
{
	DeletePropertiesOverlay();
}

void Tools_Photon_Object::UpdateWhileSelected()
{
	if(m_bChanged)
	{
		UpdatePhoton();
	}
}

void Tools_Photon_Object::CreatePropertiesOverlay()
{
	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay(GetTypeDescription(),false,&s_xPropertiesOverlayPosition);
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);

	xOffset.x += 12.0f;
	pxDialog = new Tools_ColourPicker_Dialog("Light Colour",xOffset,Asura_Vector_2(4.0f,3.0f),&m_fRed,&m_fGreen,&m_fBlue,1.0f,&m_bChanged);
	pxOverlay->AddDialog(pxDialog);
	xOffset.x -= 12.0f;

	pxDialog = new Asura_ColourNumerical_Dialog("Red",xOffset,&m_fRed,"%d",&m_bChanged);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_ColourNumerical_Dialog("Green",xOffset,&m_fGreen,"%d",&m_bChanged);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_ColourNumerical_Dialog("Blue",xOffset,&m_fBlue,"%d",&m_bChanged);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	#ifdef ASURA_USE_HDR
	pxDialog = new Asura_Numerical_Dialog("Brightness",xOffset,&m_fBrightness, 0,999.0f,0.1f,"%0.1f",&m_bChanged);
	#else
	pxDialog = new Asura_Numerical_Dialog("Brightness",xOffset,&m_fBrightness, 0,2.0f,0.1f,"%0.1f",&m_bChanged);
	#endif
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Range",xOffset,&m_fRange, 0.1f,999.0f,0.1f,"%0.1f",&m_bChanged);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Active",xOffset,&m_bActive,&m_bChanged);
	pxOverlay->AddDialog( pxDialog );
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	
	pxOverlay->m_uID = TOOLS_OVERLAY_SELECTEDOBJECT;

 	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
	
}

void Tools_Photon_Object::DeletePropertiesOverlay()
{
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SELECTEDOBJECT);
}

// (AAM - 02.02.2007) Moved the management of these light objects here from Tools_Mode_Texturing

Tools_Photon_Object* Tools_Photon_Object::CreateLight( const Asura_Vector_3& xPosition, const Asura_Matrix_3x3& xOrientation )
{
	Tools_Photon_Object* const pxNewLight = new Tools_Photon_Object( &xPosition, &xOrientation );
	return pxNewLight;
}

void Tools_Photon_Object::CreateToolsObjects()
{
	Asura_String xFilename = Tools_SaveAsura::GetNameWithoutExtension();
	xFilename += ".ToolsLights";

	Asura_ChunkLoading_System::LoadAsuraFile( xFilename.GetString() );
}

void Tools_Photon_Object::SaveToolsObjects()
{
	Asura_String xFilename = Tools_SaveAsura::GetNameWithoutExtension();
	xFilename += ".ToolsLights";

	List< Tools_Object* > xLightList;
	Tools_Object_System::ListObjectsOfType( TOT_ToolsPhoton, xLightList );

	if( xLightList.size() > 0 )
	{
		Asura_SaveBlock xSaveBlock;
		Asura_Chunk_SaveBlock_Stream xStream( &xSaveBlock );

		Tools_Chunk_ToolsLights::Create( xStream );
		xStream.WriteAsuraTerminator();

		if( !xSaveBlock.SaveFile( xFilename.GetString(), false ) )
		{
			Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Failed to open '%s' for writing to save Tools light data.", xFilename.GetString() );
		}
	}
	else if( Tools_Chunk_ToolsLights::LoadedAnyLights() )
	{
		// If we loaded some lights initially and the user has now deleted them, remove the file
		DeleteFile( xFilename.GetString() );
	}
}

void Tools_Photon_Object::DeleteToolsObjects()
{
	List< Tools_Object* > xObjectList;
	Tools_Object_System::ListObjectsOfType( TOT_ToolsPhoton, xObjectList );

	for( LIF< Tools_Object* > xObjectLIF( &xObjectList ); !xObjectLIF.done(); xObjectLIF.next() )
	{
		Tools_Object_System::DeleteObject( xObjectLIF() );
	}
}

void Tools_Photon_Object::CreateGeneralOverlay()
{
	if( !s_bGeneralOverlayOpen ) return;

	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay( "Lights", false, &s_xGeneralOverlayPosition, &s_uGeneralOverlayMinimised );

	Asura_Dialog* pxDialog;
	Asura_Tickbox_Dialog* pxTickbox;
	Asura_Vector_2 xOffset( 1.0f, 3.0f );

	const char* const szInfoString = "To select and edit lights, make sure\n'Enable Object Controls' is ticked\n in the Utility menu.";
	pxDialog = new Asura_StringDisplay_Dialog( szInfoString, xOffset );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += 4.0f;

	pxDialog = new Asura_Function_Dialog( "Add Light", xOffset, CreateLight );
	pxDialog->SetHelpText( "Create a test light.\nMainly for use when editing skin/object files.\nThese lights will only work using the 'F7' or 'F8' renderers" );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	pxTickbox = new Asura_Tickbox_Dialog( "Show Lights", xOffset, &s_bShowLights );
	pxTickbox->SetHelpText( "Show the temporary light objects." );
	pxTickbox->SetClickedFunction( ShowLightsClicked );
	pxOverlay->AddDialog( pxTickbox );
	xOffset.y += Asura_Tickbox_Dialog::LargeSpacing();

	pxOverlay->m_uID = TOOLS_OVERLAY_TOOLS_LIGHTS;

	Asura_Overlay_System::SubmitNewOverlay( pxOverlay, false );
}

void Tools_Photon_Object::ToggleGeneralOverlay()
{
	//make sure the overlay is allowed to be open
	//if( TexturingMode == TexturingMode_NodeEdit )
	//{
	//	Tools_Timed_Message_System::AddMessage( 0xffffff, "Overlay cannot be opened in this mode" );
	//	return;
	//}

	s_bGeneralOverlayOpen = !s_bGeneralOverlayOpen;

	if( s_bGeneralOverlayOpen)
	{
		s_uGeneralOverlayMinimised = 0;
		CreateGeneralOverlay();
	}
	else
	{
		Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_TOOLS_LIGHTS );
	}
}

void Tools_Photon_Object::SetLightColour(const Asura_Vector_3& xColour)
{
	m_fRed = xColour.x;
	m_fGreen = xColour.y;
	m_fBlue = xColour.z;
}

void Tools_Photon_Object::CreateLight()
{
	//make sure object selection is on
	if( !Tools_Object_System::AreControlsEnabled() )
	{
		Tools_Object_System::EnableAllControls( true );
	}

	//create a light at the current camera position
	Tools_Photon_Object* const pxPhoton = new Tools_Photon_Object( Asura_Camera::GetPosition(), Asura_Camera::GetOrientation() );

	//and select it
	Tools_Object_System::SelectObject( pxPhoton );
}

void Tools_Photon_Object::ShowLightsClicked()
{
	Tools_Object_System::SetTypeHidden( TOT_ToolsPhoton, ObjectHideFlag_User, !s_bShowLights );
}
