//////////////////////////////////////////////////////////////////////////////
//
// Tools_StreamingSounds_Export.cpp
//
// Copyright (c) 2008 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

#if defined ASURA_USE_AUDIO_SYSTEM

#include "Load&Save/LevelExport/Tools_Level_Export.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_Stream_WindowsFile.h"
#include "ChunkSystem/Asura_Chunk_StreamingSounds.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_List.h"
#include "Core/Asura_String.h"
#include "Core/Asura_StringBuffer.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Sound_Event.h"

// Tools
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Platform.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Sound/Tools_Sound_System.h"

/******************************************************************************/
/* Tools_Level_Export Static Class Functions                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Tools_Level_Export::ExportStreamingSounds()
//
// Main entrypoint for exporting streaming sounds. Confirms with the user which
// platforms they want to export to
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Level_Export::ExportStreamingSounds()
{
	if( !This ) return;

	// Ask which platform to export for

	ListSelection_Overlay xOverlay( "Export streaming sounds for :" );
	xOverlay.AddItem( "*All*", -1 );

	for( int i = 0; i < ExportType_NumPlatforms; ++i )
	{
		const EXPORT_TYPE ePlatform = static_cast< EXPORT_TYPE >(i);
		if( !This->IsPlatformAllowed( ePlatform ) ) continue;

		xOverlay.AddItem( GetPlatformName(i), i );
	}

	if( !xOverlay.DoModal() ) return;

	const int iSelectedPlatform = xOverlay.GetSelectionData();

	// If doing all platforms, check to make sure they are all writeable before starting

	if( iSelectedPlatform == -1 )
	{
		bool bAbort = false;

		for( int i = 0; i < ExportType_NumPlatforms; ++i )
		{
			const EXPORT_TYPE ePlatform = static_cast< EXPORT_TYPE >(i);
			if( !This->IsPlatformAllowed( ePlatform ) ) continue;

			s_eExportType = ePlatform;

			Asura_String xExportName;
			This->GetStreamingSoundsFilename( xExportName );

			if( !This->IsWriteableIncludingLanguageFiles( xExportName.GetString(), true ) )
			{
				bAbort = true;
			}
		}

		s_eExportType = Export_PC;

		if( bAbort )
		{
			AddMessage( MT_ERROR, "Aborting export." );
			return;
		}
	}

	// Export for the selected platform(s)

	for( int i = 0; i < ExportType_NumPlatforms; ++i )
	{
		if( iSelectedPlatform == i || iSelectedPlatform == -1 )
		{
			const EXPORT_TYPE ePlatform = static_cast< EXPORT_TYPE >(i);
			if( !This->IsPlatformAllowed( ePlatform ) ) continue;

			s_eExportType = ePlatform;
			This->ExportStreamingSoundsMain();
		}
	}

	s_eExportType = Export_PC;
}

////////////////////////////////////////////////////////////////////////////////
//
// Tools_Level_Export::ExportStreamingSounds()
//
// Exports streaming sounds to a given platform. Called by batch export process
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Level_Export::ExportStreamingSounds( const int iPlatform )
{
	Asura_Assert( This, "Invalid This pointer." );
	if( !This ) return;

	s_eExportType = static_cast< EXPORT_TYPE >( iPlatform );
	This->ExportStreamingSoundsMain();
	s_eExportType = Export_PC;
}

////////////////////////////////////////////////////////////////////////////////
//
// Tools_Level_Export::GetStreamingSoundsFilename()
//
// Constructs the filename for the streaming sounds export, based on platform
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Level_Export::GetStreamingSoundsFilename( Asura_String& xOutputFilename )
{
	Asura_StringBuffer< MAX_PATH > xFilename;

	const char* const szName = "StreamingSounds";
	const char* const szPlatformName = GetPlatformName( s_eExportType );

	if( szPlatformName )
	{
		const char* const szExtension = Tools_Platform::IsBigEndian( s_eExportType )
									  ? ASURA_STANDARD_FILE_EXTENSION_BE
									  : ASURA_STANDARD_FILE_EXTENSION_LE;

		xFilename.Print( "..\\%s_Release\\Sounds\\%s.%s", szPlatformName, szName, szExtension );
		xOutputFilename = xFilename.GetString();
	}
	else
	{
		xOutputFilename = szName;
		xOutputFilename += ".unknown_platform";
	}
}

////////////////////////////////////////////////////////////////////////////////
//
// Tools_Level_Export::ExportStreamingSoundsMain()
//
// Workhorse for the streaming sounds export. Converts files if needed and
// creates the reatil chunk in an output file
//
////////////////////////////////////////////////////////////////////////////////

void Tools_Level_Export::ExportStreamingSoundsMain()
{
	if( Tools_Sound_System::IsStreamingModified() )
	{
		Tools_Sound_System::SaveStreaming();
	}

	Asura_String xExportName;
	GetStreamingSoundsFilename( xExportName );

	if( !IsWriteableIncludingLanguageFiles( xExportName.GetString(), true ) ) return;

	// SetMainHeading() just stores a pointer, must clear it after this is removed from stack
	Asura_String xHeading = "Streaming Sounds Export - ";
	xHeading += GetPlatformName( s_eExportType );

	StartExport();
	OpenExportLog( "StreamingSounds" );
	SetMainHeading( xHeading.GetString() );
	StartFileExport();

	Tools_FileSystem::EnsureDirectoryExists( xExportName.GetString() );

	const HANDLE xSaveHandle =
		CreateFile( xExportName.GetString(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL );

	if( xSaveHandle == INVALID_HANDLE_VALUE )
	{
		AddMessage( MT_ERROR, "Failed to open '%s' for writing. Export aborted.",
			xExportName.GetString() );
	}
	else
	{
		Asura_Chunk_Stream_WindowsFile xStream( xSaveHandle );
		xStream.SetSwapEndian( ShouldExportBigEndian() );

		xStream.Write( ASURA_FILE_ID_NUM_CHARS, ASURA_UNCOMPRESSED_FILE_ID );

		Asura_List< Tools_StreamingSoundExportInfo > xExportList;

		const Asura_TemplateHashTree< Asura_Audio_StreamInfo >& xTree =
			Asura_Audio_System::GetStreamInfoTree();

		for( Asura_TemplateHashTree_ConstIterator< Asura_Audio_StreamInfo > xIter( xTree );
			!xIter.Done(); xIter.Next() )
		{
			const Asura_Audio_StreamInfo* const pxStreamInfo =
				xIter.GetCurrent();

			// Streaming dialogue will get added to the tree through playback in Tools. Don't
			// export it in this package though, it will go in the <level>.<lang>[BE] files
			if( !pxStreamInfo->m_bDialogue )
			{
				Tools_StreamingSoundExportInfo* const pxExportInfo =
					new Tools_StreamingSoundExportInfo;

				// No localisation, so export name is same as filename
				pxExportInfo->m_xFilename = pxStreamInfo->m_szFilename;
				pxExportInfo->m_xExportName = pxStreamInfo->m_szFilename;
				pxExportInfo->m_bLooping = pxStreamInfo->m_bLooping;

				xExportList.AddToList( pxExportInfo );
			}
		}

		CreateStreamingSoundsChunk( xStream, xExportList, false );

		// Asura_List doesn't own its contents, so we have to delete them manually
		for( Asura_List_It< Tools_StreamingSoundExportInfo > xIter( xExportList ); !xIter.Done(); )
		{
			// Delete the export info struct
			delete xIter.GetCurrent();

			// Delete the node itself and advance to next
			xIter.DeleteCurrent();
		}

		xStream.FlushOutput();
		xStream.WriteAsuraTerminator();
	}

	// Clean up

	if( xSaveHandle != INVALID_HANDLE_VALUE )
	{
		CloseHandle( xSaveHandle );
	}

	s_xHeading = NULL;

	if( !IsAbortFlagSet() )
	{
		AddMessage( MT_INFORMATION, "Finished writing '%s'.", xExportName.GetString() );
	}
	else
	{
		DeleteFile( xExportName.GetString() );
		AddMessage( MT_INFORMATION, "Aborted." );
	}

	EndFileExport();
	SetMainHeading( "" );
	CloseExportLog();
	EndExport( false );
}

void Tools_Level_Export::CreateStreamingSoundsChunk( Asura_Chunk_Stream& xStream,
	const Asura_List< Tools_StreamingSoundExportInfo >& xList, const bool bDialogue )
{
	// Write chunk header

	xStream.WriteChunk_Start( ASURA_CHUNK_STREAMING_SOUNDS,
		Asura_Chunk_StreamingSounds::iCURRENT_VERSION, 0 );

	const u_int uNumStreamingSounds = xList.GetNumEntries();
	xStream << uNumStreamingSounds;

	u_int uCurrSound = 0;

	// Write each entry in the list

	for( Asura_List_ConstIt< Tools_StreamingSoundExportInfo > xIter( xList );
		!xIter.Done() && !IsAbortFlagSet(); xIter.Next() )
	{
		if( bDialogue )
		{
			// Do dialogue this way to match Tools_Level_Export::AddStreamedSoundsFromList()
			s_iPercent = 90 + (uCurrSound++) * 10 / uNumStreamingSounds;
			UpdateProgress();
		}
		else
		{
			const u_int uPercent = (uCurrSound++) * 100 / uNumStreamingSounds;
			UpdateStatus( EXPORT_STAGE_STREAMING_SOUNDS, "Streaming sounds...", uPercent );
		}

		const Tools_StreamingSoundExportInfo* const pxCurrInfo = xIter.GetCurrent();

		// Name in localised export will always be English, but file is language-specific
		const Asura_String& xExportName = pxCurrInfo->m_xExportName;
		xStream << xExportName;

		const bool bLooping = pxCurrInfo->m_bLooping;
		xStream << bLooping;

		Asura_String xConvertedFile;
		bool bSuccess;
		
		const char* szFilename = pxCurrInfo->m_xFilename.GetString();
		bSuccess = ConvertSound( szFilename, TSS_QUALITY_MEDIUM, false, &xConvertedFile, true );

		if( bSuccess )
		{
			const Asura_Handle iFileHandle =
				Asura_FileAccess::Open( xConvertedFile.GetString() );

			if( iFileHandle != ASURA_INVALID_HANDLE )
			{
				const u_int uFileSize = Asura_FileAccess::GetFileSize( iFileHandle );
				xStream << uFileSize;

				void* const pFileContents = Asura_Allocate( uFileSize );

				Asura_FileAccess::ReadData( iFileHandle, pFileContents, uFileSize );
				Asura_FileAccess::Close( iFileHandle );

				xStream.Write( uFileSize, pFileContents );

				Asura_Deallocate( pFileContents );
			}
			else
			{
				Asura_Assertf( false, ("Failed to open converted sound file '%s'.",
					xConvertedFile.GetString()) );

				bSuccess = false;
			}
		}

		if( !bSuccess )
		{
			// Ensure the zero occupies 4 bytes
			const u_int uFileSize = 0;
			xStream << uFileSize;
		}
	}

	// Finish the chunk, then terminate the file

	xStream.WriteChunk_End();
}

#endif // defined ASURA_USE_AUDIO_SYSTEM