//////////////////////////////////////////////////////////////////////////////
//
// Tools_MaterialResponse_System.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#include "Materials/Tools_MaterialResponse_System.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_FixedArray.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_List.h"
#include "Core/Asura_TextureFlags.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_Debug_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "Interface/Asura_Tree_Dialog.h"

// Tools
#include "ChunkSystem/Tools_Chunk_MaterialResponse.h"
#include "Texturing/Tools_Texturing_MaterialResponse.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#ifdef ASURA_MAIN
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_ProgressIndicator.h"
#endif

#ifdef ASURA_USE_DECAL_MANAGER
#include "SpecialFX/Asura_Decal_Manager.h"
#endif

#ifndef ASURAIMPORT_EXPORTS
#include "SpecialFX/Asura_StaticCorona_System.h"
#endif // ASURAIMPORT_EXPORTS

#include "Rendering/Asura_Direct3D_MaterialResponse.h"

static const char* g_aszBlendModeNames[Asura_MaterialResponse::NUMBER_OF_BLEND_MODES] = 
{
	"None (Opaque, other settings ignored)",
	"Additive",
	"Background Blend"
};


static const char* g_aszSpecialFXModeNames[Asura_MaterialResponse::NUMBER_OF_SPECIALFX_MODES] = 
{
	"None",
	"Water"	
};

/******************************************************************************/
/* Tools_MaterialResponse_System Static Class Data                            */
/******************************************************************************/

Asura_TemplateHashTree< Tools_MaterialResponse > Tools_MaterialResponse_System::s_xMaterialTree;
Asura_TemplateHashTree< Asura_String >           Tools_MaterialResponse_System::s_xMaterialNameTree;
uint32										     Tools_MaterialResponse_System::s_uNextResponseRowIndex = 0;
const char*										 Tools_MaterialResponse_System::s_aszTextureSlotDescription[ Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS ] = 
{
	"ALBEDO_TEXTURE",
	"NORMAL_MAP",
	"SPECULAR_MAP",
	"TRANSPARENCY_MAP",
	"SELFILLUMINATION_MAP",
	"ENVIRONMENT_MAP",
	"REFLECTIVITY_MASK",
	"PARALLAX_MAP",
	"DETAIL_MAP",
	"DETAIL_NORMAL_MAP"
};

bool											Tools_MaterialResponse_System::s_bHasLoadedAllMaterials = false;

/******************************************************************************/
/* Tools_MaterialResponse_System Static Class Functions                       */
/******************************************************************************/

void Tools_MaterialResponse_System::Initialise()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::Initialise");
		
	LoadMaterials();
}

void Tools_MaterialResponse_System::Reset()
{
	s_xMaterialTree.Reset();
	s_uNextResponseRowIndex = 0;
	s_bHasLoadedAllMaterials = false;

	s_xMaterialNameTree.Reset();
}

void Tools_MaterialResponse_System::Shutdown()
{
	Reset();
}

void Tools_MaterialResponse_System::LoadMaterials()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LoadMaterials");
	
	// KJFL - 20 December 2006 - TODO: We won't want to load all the materials like this with the full system in place -
	// we only need to use materials in use (but we'll need a name list of materials etc so artists can select from materials on disk,
	// like they can with textures).
	List< Asura_String > xFilenameList;
	{
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LoadMaterials FindFilesLocal");
		Tools_FileSystem::FindFiles( szMATERIALS_DIRECTORY "*" szMATERIALS_EXTENSION, xFilenameList, TOOLS_FILEFIND_RECURSIVE | TOOLS_FILEFIND_IGNOREBACKUPDIR );
	}
	
	// Load the files we found
	{
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LoadMaterials LoadFilesLocal");
		for( CLIF< Asura_String > xIter( &xFilenameList ); !xIter.done(); xIter.next() )
		{
			const Asura_String& xCurrFilename = xIter();

			{
				// This does assume there's only one material per file
				Asura_String xMaterialName;
				Tools_MaterialResponse_System::MakeMaterialNameFromPath( xCurrFilename, xMaterialName );

				Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddMaterial( xMaterialName.GetHash() );

				if( pxNewMaterial )
				{
					pxNewMaterial->SetName( xMaterialName );
					pxNewMaterial->MarkAsUnloaded();
				}
			}

		}
	}

	// then load any files that exist on the network that so far we've missed (i.e. they don't exist locally)
	#if defined ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	List< Asura_String > xNetworkFilenameList;
	
	{
		// Enumerate files on the network

		char szNetworkDir[ MAX_PATH ];
		Asura_FileAccess::PlatformGetNetworkDir( szMATERIALS_DIRECTORY, szNetworkDir, MAX_PATH - 20 );
		strncat( szNetworkDir, "*" szMATERIALS_EXTENSION, 20 );

		{
			ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LoadMaterials FindFilesNet");
			Tools_FileSystem::FindFiles( szNetworkDir, xNetworkFilenameList, TOOLS_FILEFIND_RECURSIVE | TOOLS_FILEFIND_IGNOREBACKUPDIR );
		}

	}
	{
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LoadMaterials - load files only on network");
		for( CLIF< Asura_String > xIter( &xNetworkFilenameList ); !xIter.done(); xIter.next() )
		{
			const Asura_String& xCurrFilename = xIter();

			Asura_String xMaterialName;
			Tools_MaterialResponse_System::MakeMaterialNameFromPath( xCurrFilename, xMaterialName );

			{
				// This does assume there's only one material per file
				Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddMaterial( xMaterialName.GetHash() );

				if( pxNewMaterial )
				{
					pxNewMaterial->SetName( xMaterialName );
					pxNewMaterial->MarkAsUnloaded();
				}
			}
			
			
		}
	}
	#endif // defined ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	s_bHasLoadedAllMaterials = false;

}

void Tools_MaterialResponse_System::LogMaterials( void )
{
	ASURA_STATIC_ASSERT( Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS == 10 );
	static const char* const s_aszTextureSlotNames[Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS] = 
		{
			"ALBEDO",
			"NORMAL_MAP",
			"SPECULAR_MAP",
			"TRANSPARENCY_MAP",
			"SELFILLUMINATION_MAP",
			"ENVIRONMENT_MAP",
			"REFLECTIVITY_MASK",
			"PARALLAX_MAP"
			"ALBEDO_DETAIL_MAP"
			"NORMAL_DETAIL_MAP"
		};

	EnsureAllMaterialsAreLoaded();
	
	{
		// JAT 12/09/07 - loop over the materials logging the textures used by each material
		const char* szMaterialsLogFile = "Logfiles\\Materials.txt";

		// JAT 02/10/07 - loop over the materials and output any duplicate texture names - to try and find redundancy
		const char* szDuplicateTexturesLogFile = "Logfiles\\DuplicateTextures.txt";

		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LogMaterials" );

		// init logfile
		Asura_SimpleLogfileClear( szMaterialsLogFile );
		Asura_SimpleLogfileClear( szDuplicateTexturesLogFile );
		
		// Store the texture names by hash but also store the material index mangled with the texture index in the tree 
		Asura_TemplateHashTree< u_int > s_xTextureNames;
		const u_int uNumMaterials = Tools_MaterialResponse_System::GetNumMaterials();
		for( u_int u = 0; u < uNumMaterials; u++ )
		{
			Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex( u );
			Asura_SimpleLogfileOutput( ( szMaterialsLogFile,"\nMaterial: %s \n", pxCurrMaterial->GetName().GetString() ) );
			for ( int i = 0; i < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; i ++ )
			{
				if ( pxCurrMaterial->GetTextureHash( i ) != ASURA_HASH_ID_UNSET )
				{
					Asura_String xTextureFileName( pxCurrMaterial->GetTextureFilename( i ) );
					Asura_SimpleLogfileOutput( ( szMaterialsLogFile , "%s Texture: %s \n", s_aszTextureSlotNames[i], xTextureFileName.GetString() ) );

					Asura_String xName;
					Asura_String xExt;
					xTextureFileName.SplitFilename( NULL, &xName, &xExt );
					Asura_String xTextureName;
					xTextureName = xName;
					xTextureName += ".";
					xTextureName += xExt;
					const Asura_Hash_ID xHash( xTextureName.GetHash() );
					// Texture slot = lower 4-bits of index
					// Material Index = upper bits of the index
					const u_int uMangledIndex = ( u << 4 ) + i;
					if ( s_xTextureNames.AddNode( xHash, uMangledIndex ) == NULL )
					{
						const u_int* const puMangledIndex = s_xTextureNames.GetDataByHash( xHash );
						const u_int uMangledIndex = *puMangledIndex;
						const u_int uUsedByMaterialIndex = uMangledIndex >> 4;
						const u_int uUsedByTextureSlot = uMangledIndex & 0xF;

						Tools_MaterialResponse* const pxUsedByMaterial = Tools_MaterialResponse_System::GetMaterialByIndex( uUsedByMaterialIndex );

						// Compare the long hashes of the texture filenames and if the long hashes don't match then we have a duplicate texture
						const Asura_Hash_ID xThisHash( pxCurrMaterial->GetTextureHash( i ) );
						const Asura_Hash_ID xUsedByHash( pxUsedByMaterial->GetTextureHash( uUsedByTextureSlot ) );
						if ( xThisHash != xUsedByHash )
						{
							// So the texture name short hash is already in the list output the conflict
							// e.g. /environments/jake.tga would match /objects/jake.tga
							Asura_SimpleLogfileOutput( ( szDuplicateTexturesLogFile , "\nMaterial '%s' \n", pxCurrMaterial->GetName().GetString() ) );

							Asura_SimpleLogfileOutput( ( szDuplicateTexturesLogFile , "%s Texture: '%s' is\n",
													   s_aszTextureSlotNames[i], xTextureFileName.GetString() ) );

							Asura_SimpleLogfileOutput( ( szDuplicateTexturesLogFile , "already used by Material '%s'\n", pxUsedByMaterial->GetName().GetString() ) );

							Asura_SimpleLogfileOutput( ( szDuplicateTexturesLogFile , "the previous usage is %s Texture: '%s'\n",
													   s_aszTextureSlotNames[uUsedByTextureSlot], pxUsedByMaterial->GetTextureFilename( uUsedByTextureSlot )
													 ) );
						}
					}
				}
			}
		}
	}
}



void Tools_MaterialResponse_System::MakeMaterialNameFromPath( const Asura_String& xPath, Asura_String& xResult )
{
	const char* szRelativePath = Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir( xPath.GetString() );
	szRelativePath += strlen( szMATERIALS_DIRECTORY );

	// Remove file extension

	Asura_String xDirectory, xFilename;
	Asura_String( szRelativePath ).SplitFilename( &xDirectory, &xFilename, NULL );

	xResult = xDirectory;

	if( !xDirectory.IsBlank() )
	{
		xResult += "\\";
	}

	xResult += xFilename;
}

void Tools_MaterialResponse_System::MakePathFromMaterialName( const Asura_String& xName, Asura_String& xResult )
{
	xResult = szMATERIALS_DIRECTORY;
	xResult += xName;
	xResult += szMATERIALS_EXTENSION;
}

int Tools_MaterialResponse_System::CompareMaterials( const Asura_Hash_ID uHash1, const Asura_Hash_ID uHash2 )
{
	if( uHash1 == ASURA_HASH_ID_UNSET && uHash2 != ASURA_HASH_ID_UNSET ) return -1;
	if( uHash1 != ASURA_HASH_ID_UNSET && uHash2 == ASURA_HASH_ID_UNSET ) return 1;
	if( uHash1 == ASURA_HASH_ID_UNSET && uHash2 == ASURA_HASH_ID_UNSET ) return 0;

	const Tools_MaterialResponse* const pxMaterial1 = GetMaterialByHashID( uHash1 );
	const Tools_MaterialResponse* const pxMaterial2 = GetMaterialByHashID( uHash2 );

	#if defined ASURA_ASSERT
		const char* const szMissingMaterialText = "Some polygons are using missing material %s. Is it checked in?";
	#endif

	Asura_Assertf( pxMaterial1, (szMissingMaterialText, GetMissingMaterialDescription(uHash1)) );
	Asura_Assertf( pxMaterial2, (szMissingMaterialText, GetMissingMaterialDescription(uHash2)) );
	if( !pxMaterial1 || !pxMaterial2 ) return 0;

	return CompareMaterials( *pxMaterial1, *pxMaterial2 );
}

int Tools_MaterialResponse_System::CompareMaterials( const Tools_MaterialResponse& xMaterial1, const Tools_MaterialResponse& xMaterial2 )
{
	const Asura_Hash_ID uHash1 = xMaterial1.GetHash();
	const Asura_Hash_ID uHash2 = xMaterial2.GetHash();

	// Early out - are these the same material?
	if( uHash1 == uHash2 ) return 0;

	// 1. Render order

	// A. Additive-alpha vs not (the former go last)

	if( xMaterial1.GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_ADDITIVE
	 && xMaterial2.GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_ADDITIVE )
	{
		return 1;
	}
	if( xMaterial1.GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_ADDITIVE
	 && xMaterial2.GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_ADDITIVE )
	{
		return -1;
	}

	// B. Transparent vs. non-transparent (the former go last)

	if( xMaterial1.GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_BACKGROUNDBLEND
	 && xMaterial2.GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_BACKGROUNDBLEND )
	{
		return 1;
	}
	if( xMaterial1.GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_BACKGROUNDBLEND
	 && xMaterial2.GetBlendMode() == Asura_MaterialResponse::BLEND_MODE_BACKGROUNDBLEND )
	{
		return -1;
	}

	// Couldn't make a decision by render order, let's try
	// 2. Texture

	// A. Sort by albedo texture

	{
		const int iTexture1 = xMaterial1.GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE );
		const int iTexture2 = xMaterial2.GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE );

		if( iTexture1 < iTexture2 ) return 1;
		if( iTexture1 > iTexture2 ) return -1;
	}

	
	// TODO: Normal maps? Reflectance maps? etc.

	// Couldn't make a decision so far; let's just separate them based on material name
	// Material names

	if( uHash1 < uHash2 ) return 1;
	if( uHash1 > uHash2 ) return -1;

	return 0;
}

Tools_MaterialResponse* Tools_MaterialResponse_System::GetMaterialByHashID( const Asura_Hash_ID uNameHashID, int* const piMaterialIndex /*= NULL*/ )
{
	if( uNameHashID == ASURA_HASH_ID_UNSET ) return NULL;

	Tools_MaterialResponse* const pxMaterial = s_xMaterialTree.GetDataByHash( uNameHashID );

	if (pxMaterial) pxMaterial->LoadDataIfRequired();

	if( piMaterialIndex )
	{
		// Returns -1 if not found
		*piMaterialIndex = s_xMaterialTree.GetIndexOfHash( uNameHashID );
	}

	return pxMaterial;
}

Tools_MaterialResponse* Tools_MaterialResponse_System::GetMaterialByIndex( const u_int uIndex )
{
	Tools_MaterialResponse* const pxMaterial = s_xMaterialTree.GetDataByIndex( uIndex );

	if (pxMaterial) pxMaterial->LoadDataIfRequired();

	return pxMaterial;
}

Tools_MaterialResponse* const Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading( const u_int uIndex )
{
	Tools_MaterialResponse* const pxMaterial = s_xMaterialTree.GetDataByIndex( uIndex );

	return pxMaterial;
}



const char* Tools_MaterialResponse_System::GetMaterialNameByIndex( const u_int uIndex )
{
	Tools_MaterialResponse* const pxMaterial = s_xMaterialTree.GetDataByIndex( uIndex );

	return pxMaterial ? pxMaterial->GetName().GetString() : 0;
}



static int GetMaterialNames_SortCallback( const void* pData1, const void* pData2)
{
	// The data are pointers to string pointers
	const char* const szString1 = *static_cast< const char* const* >( pData1 );
	const char* const szString2 = *static_cast< const char* const* >( pData2 );

	//The purpose of this somewhat overly complicated sorting method is to make sure
	//that within any given branch of the tree any folder names end up earlier in
	//the list than any material names.
	//RAR 13/09/2007

	
	//Get the last slash in each string to determine the path part of the strings
	const char* pcLastSlash1 = strrchr(szString1,'\\');
	const char* pcLastSlash2 = strrchr(szString2,'\\');

	int iDistToLastSlash1 = pcLastSlash1 ? int(pcLastSlash1 - szString1) : 0;
	int iDistToLastSlash2 = pcLastSlash2 ? int(pcLastSlash2 - szString2) : 0;

	if(iDistToLastSlash1 != iDistToLastSlash2)
	{
		//See if the shorter path is a substring of the longer path
		
		int iMinDistToLastSlash = Asura_Maths::Min(iDistToLastSlash1,iDistToLastSlash2);
		if(!_strnicmp(szString1,szString2,iMinDistToLastSlash))
		{
			//sort so that the string with the shorter path appears later.
			return (iDistToLastSlash1 < iDistToLastSlash2) ? 1 : -1;
		}
	}

	//One string isn't in a subpath of the other string's path , so just sort alphabetically.
	return _stricmp( szString1, szString2 );
}

void Tools_MaterialResponse_System::GetMaterialNames( int& iNumMaterials, const char**& pszMaterialNames )
{
	iNumMaterials = GetNumMaterials();
	pszMaterialNames = new const char*[ iNumMaterials ];

	if( !pszMaterialNames )
	{
		iNumMaterials = 0;
		return;
	}

	const char** pszCurrName = pszMaterialNames;
	for( int i = 0; i < iNumMaterials; ++i )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = s_xMaterialTree.GetDataByIndex(i);
		*pszCurrName++ = pxCurrMaterial->GetName().GetString();
	}

	qsort( pszMaterialNames, iNumMaterials, sizeof( const char* ), GetMaterialNames_SortCallback );
}

u_int Tools_MaterialResponse_System::GetNumMaterials()
{
	return s_xMaterialTree.GetNumberOfNodes();
}

void Tools_MaterialResponse_System::CreateMaterialNameTree( Asura_TreeDialog_Tree& xTree,
														    MaterialTreeFilterByMaterialFn* const pfnFilter /*= NULL*/,
															const void* const pUserData /*= NULL*/ )
{
	CreateMaterialNameTreeInternal(xTree, pfnFilter, NULL, pUserData);
}

void Tools_MaterialResponse_System::CreateMaterialNameTree( Asura_TreeDialog_Tree& xTree,
														    MaterialTreeFilterByHashFn* const pfnFilter,
															const void* const pUserData)
{
	CreateMaterialNameTreeInternal(xTree, NULL, pfnFilter, pUserData);
}


void Tools_MaterialResponse_System::CreateMaterialNameTreeInternal( Asura_TreeDialog_Tree& xTree,
																    MaterialTreeFilterByMaterialFn* const pfnFilterByMaterial,
																    MaterialTreeFilterByHashFn* const pfnFilterByHash,
																	const void* const pUserData)
{
#ifdef ASURA_MAIN	
	xTree.DeleteAll();

	if(pfnFilterByMaterial)
	{
		Tools_MaterialResponse_System::EnsureAllMaterialsAreLoaded();
	}

	for( Asura_TemplateHashTree_Iterator< Tools_MaterialResponse > xIter( s_xMaterialTree );
		!xIter.Done(); xIter.Next() )
	{
		Tools_MaterialResponse* const pxCurrMaterial = xIter.GetCurrent();

		// KJFL - 26 October 2007 - delay getting a material pointer as long as possible, since this will
		// require loading the material fully.
		if ( pfnFilterByMaterial )
		{
			// Callback may require data loaded
			pxCurrMaterial->LoadDataIfRequired();

			// Filter based on the supplied callback
			if( pxCurrMaterial && !pfnFilterByMaterial( *pxCurrMaterial, pUserData ) ) continue;
		}
		else if(pfnFilterByHash)
		{
			if( !pfnFilterByHash( pxCurrMaterial->GetHash(), pUserData ) ) continue;
		}

		Asura_TreeDialog_Item* const pxNewItem =
			xTree.AddFilePath( pxCurrMaterial->GetName().GetString() );

		// Ensure the user data associated with the node is the full path
		if( pxNewItem )
		{
			// We use the user data because the per-node hash ID was added after this was written
			// TODO: could switch to using the hash ID now
			pxNewItem->SetUserData( pxCurrMaterial->GetName().GetHash() );
		}
	}

	if( xTree.GetRoot() )
	{
		xTree.GetRoot()->Expand( true );
	}
#endif //ASURA_MAIN
}

u_int Tools_MaterialResponse_System::GetTextureUsage( const Asura_Hash_ID uTextureNameHash )
{
	// Can't store flags for more than 32 texture types in a u_int
	ASURA_STATIC_ASSERT( Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS / 2 <= 32 );

	u_int uUsage = 0;

	// Loop over all materials. If the texture is used as either the primary or secondary
	// for one of the slot types, set the corresponding flag in the return value

	const u_int uNumMaterials = s_xMaterialTree.GetNumberOfNodes();
	for( u_int u = 0; u < uNumMaterials; ++u )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = GetMaterialByIndex_WithoutLoading(u);
		if ( !pxCurrMaterial->IsLoaded() ) continue;

		for( u_int v = 0, uUsageFlag = TU_ALBEDO; v < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++v )
		{
			if( pxCurrMaterial->GetTextureHash(v) == uTextureNameHash )
			{
				uUsage |= uUsageFlag;
			}

			uUsageFlag <<= 1;
		}
	}


#ifdef ASURA_USE_DECAL_MANAGER
	// DMR 16/08/07
	// This is not ideal, hopefully a temporary solution - to fix up the decal textures
	uUsage |= Asura_Decal_Manager::GetTextureUsage( uTextureNameHash );
#endif

#ifndef ASURAIMPORT_EXPORTS
	uUsage |= Asura_StaticCorona_System::GetTextureUsage( uTextureNameHash );
#endif // ASURAIMPORT_EXPORTS


	return uUsage;
}

void Tools_MaterialResponse_System::AddToMaterialNameTree(const char* szName)
{
	Asura_Assert(szName && *szName,"");
	if(!szName || !*szName) return;

	Asura_Hash_ID uHash = Asura_GetHashID(szName);
	
	Asura_String* pxString = s_xMaterialNameTree.AddNode(uHash);
	if(pxString)
	{
		*pxString = szName;		
	}
}

const char* Tools_MaterialResponse_System::GetMaterialNameByHashID(Asura_Hash_ID uHash, bool bCheckAllKnownMaterialNames /*= true*/)
{
	//First try the material list , to see if there is a matching material
	const Tools_MaterialResponse* pxMaterial = s_xMaterialTree.GetDataByHash(uHash);
	if(pxMaterial) return pxMaterial->GetName().GetString();

	if(bCheckAllKnownMaterialNames)
	{
		//Failing that , try the previously known material names
		const Asura_String* pxName = s_xMaterialNameTree.GetDataByHash(uHash);
		if(pxName) return pxName->GetString();
	}

	return 0;
}

const char* Tools_MaterialResponse_System::GetMissingMaterialDescription(Asura_Hash_ID uHash)
{
	static char s_acBuffer[200];

	const Asura_String* pxName = s_xMaterialNameTree.GetDataByHash(uHash);
	if(pxName)
	{
		Asura_PrintfToBfr(s_acBuffer,"0x%x (%s)", sizeof(s_acBuffer), true, uHash, pxName->GetString());
	}
	else
	{
		Asura_PrintfToBfr(s_acBuffer,"0x%x (Name Not Available)", sizeof(s_acBuffer), true, uHash);
	}

	return s_acBuffer;
}

void Tools_MaterialResponse_System::EnsureAllMaterialsAreLoaded()
{
	if(s_bHasLoadedAllMaterials) return;
	s_bHasLoadedAllMaterials = true;

	#ifdef ASURA_MAIN
	Tools_ProgressIndicator xProgress("Loading Materials");
	#endif

	//Just get each material in turn , this will force it to be loaded
	u_int uNumMaterials = GetNumMaterials();
	for(u_int i=0; i<uNumMaterials; i++)
	{
		#ifdef ASURA_MAIN
		xProgress.Update(i, uNumMaterials);
		#endif
		
		GetMaterialByIndex(i);
	}
}




/******************************************************************************/
/* Tools_MaterialResponse Class Functions                                     */
/******************************************************************************/

Tools_MaterialResponse::Tools_MaterialResponse()
	: m_xName("")
	, m_uParentHash( ASURA_HASH_ID_UNSET )

	, m_fAmbientContribution( 100.0f )
	, m_fColouriseAmbient( 0.0f )
	, m_xAmbientColour( 1.0f, 1.0f, 1.0f, 0.0f )

	, m_xDiffuseGraph( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH, 0.0f, 100.0f, -1.0f, 1.0f )
	, m_fColouriseDiffuse( 0.0f )
	, m_xDiffuseColour( 0.0f, 0.0f, 0.0f, 0.0f )

	, m_fMetalicity( 0.0f )
	, m_xSpecularGraph( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH, 0.0f, 100.0f, -1.0f, 1.0f )
	, m_fColouriseSpecular( 0.0f )
	, m_xSpecularColour( 0.0f, 0.0f, 0.0f, 0.0f )

	, m_xTransparencyGraph( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH, 100.0f, 100.0f, -1.0f, 1.0f )

	, m_fSelfIlluminationContribution( 0.0f )
	, m_fColouriseSelfIllumination( 0.0f )
	, m_xSelfIlluminationColour( 0.0f, 0.0f, 0.0f, 0.0f )

	, m_fReflectivity( 0.0f )

	, m_fParallaxScale( 4.0f )
	, m_fAlbedoDetailMapUVScale( 4.0f )
	, m_fAlbedoDetailMapValueScale( 0.5f )
	, m_fNormalDetailMapUVScale( 4.0f )
	, m_fNormalDetailMapValueScale( 0.5f )
	, m_puOldTextureHashIDs(0)
	, m_pxOldTextureFilenames(0)
	, m_puOldTextureAnimHashIDs(0)

	, m_eBlendMode( Asura_MaterialResponse::BLEND_MODE_OPAQUE )
	, m_eRenderType( Asura_MaterialResponse::RENDER_TYPE_DEFAULT )

	, m_bModified( false )
	, m_bNeedsSave( false )
	, m_bIsLoaded( true )

	, m_uChunkedFlags(0)
	, m_uProjectContext(0)

	, m_eSpecialFXMode( Asura_MaterialResponse::SPECIALFX_MODE_NONE )
	, m_xWaterReflectionColour( 0.5f, 0.5f, 0.5f, 0.5f )
	, m_xWaterRefractionColour( 0.5f, 0.5f, 0.5f, 0.5f )
	, m_xWaterFogColour( 0.25f, 0.25f, 0.25f, 0.25f )
	, m_fWaterFogStrength( 1.0f )	
	, m_fWaterRippleScale( 1.0f )
	, m_fWaterDistortionScale( 1.0f )
	, m_fWaterEdgeSoftness( 1.0f )

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
	, m_uSurfaceTypeHash(Asura_MaterialResponse_System::GetDefaultSurfaceTypeHash())
#else
	, m_uSurfaceType(0)
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
	, m_uSecondMaterialID(0)

{
	for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++u )
	{
		m_auTextureHashIDs[u] = ASURA_HASH_ID_UNSET;
		m_axTextureFilenames[u].MakeBlank();
	}

	m_uTextureAnimHashID = ASURA_HASH_ID_UNSET;

	for ( u_int u = 0; u <2; ++u )
	{
		m_afWaterUVControl[ u*3 + 0 ]	= 0.0f;	  // u scroll speed
		m_afWaterUVControl[ u*3 + 1 ]	= 0.0f;   // v scroll speed
		m_afWaterUVControl[ u*3 + 2 ]	= 1.0f;   // uv scale

		m_afWaterWaveControl[ u*3 + 0 ] = 0.0f;	  // amplitude
		m_afWaterWaveControl[ u*3 + 1 ] = 1.0f;   // frequency
		m_afWaterWaveControl[ u*3 + 2 ] = 1.0f;   // speed
	}

	// Graph values in percentage points
	m_xDiffuseGraph.SetClampingValues( true, true, 0.0f, fTOOLS_MATERIAL_RESPONSE_GRAPH_MAX );
	m_xSpecularGraph.SetClampingValues( true, true, 0.0f, fTOOLS_MATERIAL_RESPONSE_GRAPH_MAX );
	// Transparency only from 0-100%
	m_xTransparencyGraph.SetClampingValues( true, true, 0.0f, 100.0f );

	// Don't want to use Tools' dynamic approximation for the discrete samples
	m_xDiffuseGraph.SetSampleTolerance( 0.0f );
	m_xSpecularGraph.SetSampleTolerance( 0.0f );
	m_xTransparencyGraph.SetSampleTolerance( 0.0f );

	// Graphs' domains run from -1.0 to 1.0 meaning back-facing to forward-facing.
	// Add a point in the middle to default back-facing stuff to 0.0 value
	m_xDiffuseGraph.AddPoint( 0.0f, 0.0f );

	// For specular, default to roughly power 16 curve
	//	ApplySpecularSharpness( 16.0f, false );

	// KJFL - 15 April 2008 - commented out the above code which was making a 256 point specular graph,
	// which was insane overkill. It actually added several seconds to the load time of Tools, as well
	// as increasing the time it takes to quit to the OS (since there are so many deallocations etc).
	// Instead I'm using adding just five points that give something visually identical to the power 
	// 16 curve, and the above fn gets called when creating a new material in Tools - but setting that 
	// optimise option to be true.
	
	m_xSpecularGraph.AddPoint( 1.1595092f, 0.75686276f );
	m_xSpecularGraph.AddPoint( 5.6160483f, 0.83529413f );
	m_xSpecularGraph.AddPoint( 20.565952f, 0.90588236f );
	m_xSpecularGraph.AddPoint( 40.516304f, 0.94509804f );
	m_xSpecularGraph.AddPoint( 77.649139f, 0.98431373f );


	// ApplySpecularSharpness() sets modified flag
	m_bModified = false;

	// If a row hasn't been allocated, we point it off the bottom row of the texture - to mark it as unset
	m_uResponseRowIndex = ASURA_PC_MATERIALS_TEXTURE_HEIGHT;
}


Tools_MaterialResponse::Tools_MaterialResponse( const Tools_MaterialResponse& xOther )
	: m_puOldTextureHashIDs(0)
	, m_pxOldTextureFilenames(0)
	, m_puOldTextureAnimHashIDs(0)
	, m_uSecondMaterialID(0)
	, m_uResponseRowIndex( ASURA_PC_MATERIALS_TEXTURE_HEIGHT )
{
	//Leaving all the 'Old' data used by the conversion process as zero, since I don't think there is any reason to copy
	//them when creating a new material.
	//RAR 13/06/2007

	// We don't need all the duplication, do we?....
	*this = xOther;
}

Tools_MaterialResponse::~Tools_MaterialResponse()
{
	delete [] m_puOldTextureHashIDs;
	delete [] m_pxOldTextureFilenames;
	delete [] m_puOldTextureAnimHashIDs;
}

const Tools_MaterialResponse& Tools_MaterialResponse::operator =( const Tools_MaterialResponse& xOther )
{
	m_xName									= xOther.m_xName;
	m_uParentHash							= xOther.m_uParentHash;

	m_fAmbientContribution					= xOther.m_fAmbientContribution;
	m_fColouriseAmbient						= xOther.m_fColouriseAmbient;
	m_xAmbientColour						= xOther.m_xAmbientColour;

	m_xDiffuseGraph							= xOther.m_xDiffuseGraph;
	m_fColouriseDiffuse						= xOther.m_fColouriseDiffuse;
	m_xDiffuseColour						= xOther.m_xDiffuseColour;

	m_fMetalicity							= xOther.m_fMetalicity;
	m_xSpecularGraph						= xOther.m_xSpecularGraph;
	m_fColouriseSpecular					= xOther.m_fColouriseSpecular;
	m_xSpecularColour						= xOther.m_xSpecularColour;

	m_xTransparencyGraph					= xOther.m_xTransparencyGraph;

	m_fSelfIlluminationContribution			= xOther.m_fSelfIlluminationContribution;
	m_fColouriseSelfIllumination			= xOther.m_fColouriseSelfIllumination;
	m_xSelfIlluminationColour				= xOther.m_xSelfIlluminationColour;

	m_fReflectivity							= xOther.m_fReflectivity;

	m_fParallaxScale						= xOther.m_fParallaxScale;

	m_fAlbedoDetailMapUVScale				= xOther.m_fAlbedoDetailMapUVScale;
	m_fAlbedoDetailMapValueScale			= xOther.m_fAlbedoDetailMapValueScale;
	m_fNormalDetailMapUVScale				= xOther.m_fNormalDetailMapUVScale;
	m_fNormalDetailMapValueScale			= xOther.m_fNormalDetailMapValueScale;



	for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++u )
	{
		m_auTextureHashIDs[u]				= xOther.m_auTextureHashIDs[u];
		m_axTextureFilenames[u]				= xOther.m_axTextureFilenames[u];
	}

	m_uTextureAnimHashID					= xOther.m_uTextureAnimHashID;

	m_eBlendMode							= xOther.m_eBlendMode;
	m_eRenderType							= xOther.m_eRenderType;

	m_uChunkedFlags							= xOther.m_uChunkedFlags;
	m_uProjectContext						= xOther.m_uProjectContext;

	m_eSpecialFXMode						= xOther.m_eSpecialFXMode;	
	m_xWaterReflectionColour				= xOther.m_xWaterReflectionColour;
	m_xWaterRefractionColour				= xOther.m_xWaterRefractionColour;
	m_xWaterFogColour						= xOther.m_xWaterFogColour;
	m_fWaterFogStrength						= xOther.m_fWaterFogStrength;	
	m_fWaterRippleScale						= xOther.m_fWaterRippleScale;
	m_fWaterDistortionScale					= xOther.m_fWaterDistortionScale;
	m_fWaterEdgeSoftness					= xOther.m_fWaterEdgeSoftness;

	for ( u_int u = 0; u < 6; ++u )
	{
		m_afWaterUVControl[u]				= xOther.m_afWaterUVControl[u];
		m_afWaterWaveControl[u]				= xOther.m_afWaterWaveControl[u];
	}

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
	m_uSurfaceTypeHash	= xOther.m_uSurfaceTypeHash;
#else
	m_uSurfaceType = xOther.m_uSurfaceType;
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

	// Don't copy modification state
	m_bModified								= false;
	m_bNeedsSave							= false;
	m_bIsLoaded								= true;

	return *this;
}

void Tools_MaterialResponse::LoadDataIfRequired( void )
{
	if ( !m_bIsLoaded )
	{
		LoadData();

		// force a parent to be loaded at the same time
		if ( GetParentHash() != ASURA_HASH_ID_UNSET )
		{
			Tools_MaterialResponse_System::GetMaterialByHashID( GetParentHash() );
		}
	}
}
void Tools_MaterialResponse::LoadData( void )
{
	// KJFL - 10 April 2008 - commented this line out now the load on demand seems to be working nicely.
	// For debugging it's good to see when the calls happen but you wouldn't want it on all the time.
	//ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_MaterialResponse_System::LoadData  - Just-in-time material load" );

	if ( m_bIsLoaded )
	{
		Asura_Assert( !m_bIsLoaded, "Code Bug: data already loaded" );
		return;
	}

	Asura_String xFilename;

	Tools_MaterialResponse_System::MakePathFromMaterialName( m_xName, xFilename );

	Tools_Chunk_MaterialResponse::SetLoadingFilename( xFilename.GetString() );
	LoadAsuraFile( xFilename.GetString() );
}


int Tools_MaterialResponse::GetTexture( u_int uTextureSlot ) const
{
	Asura_Assert( uTextureSlot < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS, "Code bug: invalid texture slot" );

	Asura_Hash_ID uHash = m_auTextureHashIDs[uTextureSlot];
	// early out case for an unset texture
	if ( uHash == ASURA_HASH_ID_UNSET ) return ASURA_TEXTURE_NONE;

	// see if the texture is in memory
	int iTextureIndex = Asura_Texture_Management::GetIndexFromHashID( uHash );
	
	// return the index if we found one
	if ( iTextureIndex != ASURA_TEXTURE_NONE )
	{
		return iTextureIndex;
	}

	// otherwise try and load the texture
	bool bWasNewTexture = false;
	iTextureIndex = Asura_Texture_Management::AddTexture( m_axTextureFilenames[uTextureSlot], &bWasNewTexture);
	if( true == bWasNewTexture)
	{
		// DLH - 26 June 2007 16:57:48 - The texture was new and is needed NOW so load the resource
		Asura_Texture_Management::LoadTextures();

		#if defined ASURA_ASSERT

			if( uTextureSlot == Asura_MaterialResponse::ENVIRONMENT_MAP )
			{
				const bool bCubemap = Asura_Texture_Management::IsTextureCubemap( iTextureIndex );

				// (AAM - 12.09.2007) Water still uses spheremaps for environment mapping

				if( GetSpecialFXMode() == Asura_MaterialResponse::SPECIALFX_MODE_WATER )
				{
					Asura_Assertf( !bCubemap,
						("Water '%s' is using texture '%s' as an environment map, but it is a cubemap. Water still needs spheremaps.",
						GetName().GetString(), m_axTextureFilenames[ uTextureSlot ]) );
				}
				else
				{
					Asura_Assertf( bCubemap,
						("Material '%s' is using texture '%s' as an environment map, but it isn't a cubemap.",
						GetName().GetString(), m_axTextureFilenames[ uTextureSlot ]) );
				}
			}

		#endif // defined ASURA_ASSERT
	}
	return( iTextureIndex);
}

Asura_Hash_ID Tools_MaterialResponse::GetTextureHash( u_int uTextureSlot ) const
{
	Asura_Assert( uTextureSlot < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS, "Code bug: invalid texture slot" );

	return m_auTextureHashIDs[uTextureSlot];
}

const char* Tools_MaterialResponse::GetTextureFilename( u_int uTextureSlot ) const
{
	Asura_Assert( uTextureSlot < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS, "Code bug: invalid texture slot" );

	return m_axTextureFilenames[uTextureSlot].GetString();
}

void Tools_MaterialResponse::SetTexture( u_int uTextureSlot, const int iTextureIndex )
{
	Asura_Assert( uTextureSlot < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS, "Code bug: invalid texture slot" );

	m_auTextureHashIDs[ uTextureSlot ] = Asura_Texture_Management::GetHashIDFromIndex( iTextureIndex );
	
	const Asura_String* pxName = Asura_Texture_Management::GetTextureName( iTextureIndex );

	if (pxName)
	{
		m_axTextureFilenames[ uTextureSlot ] = *pxName;
	}
	else
	{
		m_axTextureFilenames[ uTextureSlot ].MakeBlank();
	}

	m_bModified = true;
}

void Tools_MaterialResponse::ClearTexture( u_int uTextureSlot )
{
	Asura_Assert( uTextureSlot < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS, "Code bug: invalid texture slot" );
	
	m_auTextureHashIDs[ uTextureSlot ] = ASURA_HASH_ID_UNSET;
	m_axTextureFilenames[ uTextureSlot ].MakeBlank();
	m_bModified = true;
}

Asura_Hash_ID Tools_MaterialResponse::GetTextureAnim() const
{
	return m_uTextureAnimHashID;
}

void Tools_MaterialResponse::SetTextureAnim(Asura_Hash_ID uAnimID)
{
	if(m_uTextureAnimHashID != uAnimID)
	{
		m_uTextureAnimHashID = uAnimID;
		#ifdef ASURA_MAIN
		SetNeedsSave(true);
		#endif
	}
}


const char* const* Tools_MaterialResponse::GetBlendModeTextStrings( void ) 
{
	return g_aszBlendModeNames;
}


const char* const* Tools_MaterialResponse::GetSpecialFXModeTextStrings( void ) 
{
	return g_aszSpecialFXModeNames;
}


void Tools_MaterialResponse::CopyTextures( const bool bPerformCopy, const Tools_MaterialResponse& xFrom,
										   const Asura_MaterialResponse::TEXTURE_SLOT_ID eSlot)
{
	if( bPerformCopy )
	{
		m_auTextureHashIDs[ eSlot ] = xFrom.m_auTextureHashIDs[ eSlot ];
		m_axTextureFilenames[ eSlot ] = xFrom.m_axTextureFilenames[ eSlot ];

		m_bModified = true;
	}
}

void Tools_MaterialResponse::CopyPropertiesFrom( const Tools_MaterialResponse& xOther, PropertyCopyFlags xCopyFlags )
{
	if( xCopyFlags.m_uEverything )
	{
		xCopyFlags.m_uResponses    = true;
		xCopyFlags.m_uTextures     = true;
		xCopyFlags.m_uTextureAnims = true;
		xCopyFlags.m_uSurfaceType = true;
	}

	if( xCopyFlags.m_uResponses )
	{
		xCopyFlags.m_uAmbientResponse          = true;
		xCopyFlags.m_uDiffuseResponse          = true;
		xCopyFlags.m_uSpecularResponse         = true;
		xCopyFlags.m_uTransparencyResponse     = true;
		xCopyFlags.m_uSelfIlluminationResponse = true;
		xCopyFlags.m_uReflectivityResponse     = true;
	}

	if( xCopyFlags.m_uTextures )
	{
		xCopyFlags.m_uAlbedoMaps           = true;
		xCopyFlags.m_uNormalMaps           = true;
		xCopyFlags.m_uSpecularMaps         = true;
		xCopyFlags.m_uTransparencyMaps     = true;
		xCopyFlags.m_uSelfIlluminationMaps = true;
		xCopyFlags.m_uReflectionTextures   = true;
	};

	if( xCopyFlags.m_uAmbientResponse )
	{
		m_fAmbientContribution = xOther.m_fAmbientContribution;
		m_fColouriseAmbient = xOther.m_fColouriseAmbient;
		m_xAmbientColour = xOther.m_xAmbientColour;
		m_bModified = true;
	}

	if( xCopyFlags.m_uDiffuseResponse )
	{
		m_xDiffuseGraph = xOther.m_xDiffuseGraph;
		m_fColouriseDiffuse = xOther.m_fColouriseDiffuse;
		m_xDiffuseColour = xOther.m_xDiffuseColour;
		m_fAlbedoDetailMapUVScale				= xOther.m_fAlbedoDetailMapUVScale;
		m_fAlbedoDetailMapValueScale			= xOther.m_fAlbedoDetailMapValueScale;
		m_bModified = true;
	}

	if( xCopyFlags.m_uSpecularResponse )
	{
		m_fMetalicity = xOther.m_fMetalicity;
		m_xSpecularGraph = xOther.m_xSpecularGraph;
		m_fColouriseSpecular = xOther.m_fColouriseSpecular;
		m_xSpecularColour = xOther.m_xSpecularColour;

		if( xOther.GetIgnoreSpecularPrelighting() )
		{
			m_uChunkedFlags |= BITMASK_IGNORE_SPECULAR_PRELIGHTING;
		}
		else
		{
			m_uChunkedFlags &= ~BITMASK_IGNORE_SPECULAR_PRELIGHTING;
		}

		m_bModified = true;
	}

	if( xCopyFlags.m_uTransparencyResponse )
	{
		m_xTransparencyGraph = xOther.m_xTransparencyGraph;
		m_eBlendMode = xOther.m_eBlendMode;
		if(xOther.GetIsDecal())
		{
			m_uChunkedFlags	|= BITMASK_DECAL;
		}
		else
		{
			m_uChunkedFlags	&= ~BITMASK_DECAL;
		}
		if(xOther.GetIsZPrePass())
		{
			m_uChunkedFlags	|= BITMASK_ZPREPASS;
		}
		else
		{
			m_uChunkedFlags	&= ~BITMASK_ZPREPASS;
		}

		m_bModified = true;
	}

	if( xCopyFlags.m_uSelfIlluminationResponse )
	{
		m_fSelfIlluminationContribution = xOther.m_fSelfIlluminationContribution;
		m_fColouriseSelfIllumination = xOther.m_fColouriseSelfIllumination;
		m_xSelfIlluminationColour = xOther.m_xSelfIlluminationColour;
		m_bModified = true;
	}

	if( xCopyFlags.m_uReflectivityResponse )
	{
		m_fReflectivity = xOther.m_fReflectivity;
		m_bModified = true;
	}

	if( xCopyFlags.m_uNormalMaps )
	{
		m_fParallaxScale = xOther.m_fParallaxScale;
		m_fNormalDetailMapUVScale = xOther.m_fNormalDetailMapUVScale;
		m_fNormalDetailMapValueScale = xOther.m_fNormalDetailMapValueScale;

		m_bModified = true;
	}

	if( xCopyFlags.m_uSpecialFX )
	{
		
		m_eSpecialFXMode		 = xOther.m_eSpecialFXMode;

		// SpecialFX mode: Water
		m_xWaterReflectionColour = xOther.m_xWaterReflectionColour;
		m_xWaterRefractionColour = xOther.m_xWaterRefractionColour; 
		m_xWaterFogColour		 = xOther.m_xWaterFogColour;
		m_fWaterFogStrength		 = xOther.m_fWaterFogStrength;
		m_fWaterRippleScale		 = xOther.m_fWaterRippleScale;
		m_fWaterDistortionScale	 = xOther.m_fWaterDistortionScale;
		m_fWaterEdgeSoftness	 = xOther.m_fWaterEdgeSoftness;

		for(int i=0; i<6; i++)
		{		
			m_afWaterUVControl[i]   = xOther.m_afWaterUVControl[i];
			m_afWaterWaveControl[i] = xOther.m_afWaterWaveControl[i];
		}

		m_uProjectContext		= xOther.m_uProjectContext;
		m_bModified = true;
	}

	CopyTextures( xCopyFlags.m_uAlbedoMaps, xOther, Asura_MaterialResponse::ALBEDO_TEXTURE );
	CopyTextures( xCopyFlags.m_uAlbedoMaps, xOther, Asura_MaterialResponse::DETAIL_MAP);
	CopyTextures( xCopyFlags.m_uNormalMaps, xOther, Asura_MaterialResponse::NORMAL_MAP);
	CopyTextures( xCopyFlags.m_uNormalMaps, xOther, Asura_MaterialResponse::DETAIL_NORMAL_MAP);
	CopyTextures( xCopyFlags.m_uSpecularMaps, xOther, Asura_MaterialResponse::SPECULAR_MAP);
	CopyTextures( xCopyFlags.m_uTransparencyMaps, xOther, Asura_MaterialResponse::TRANSPARENCY_MAP);
	CopyTextures( xCopyFlags.m_uSelfIlluminationMaps, xOther, Asura_MaterialResponse::SELFILLUMINATION_MAP);
	CopyTextures( xCopyFlags.m_uReflectionTextures, xOther, Asura_MaterialResponse::ENVIRONMENT_MAP);
	CopyTextures( xCopyFlags.m_uReflectionTextures, xOther, Asura_MaterialResponse::REFLECTIVITY_MASK);
	CopyTextures( xCopyFlags.m_uNormalMaps, xOther, Asura_MaterialResponse::PARALLAX_MAP);

	// The textures in the other material might not have been loaded yet, in which case they won't
	// show up if this material's properties are being edited at the moment
	for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++u )
	{
		GetTexture(u);
	}

	if( xCopyFlags.m_uTextureAnims )
	{
		m_uTextureAnimHashID = xOther.m_uTextureAnimHashID;
	 
		m_bModified = true;
	}
	
	if( xCopyFlags.m_uSurfaceType )
	{
#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
		m_uSurfaceTypeHash	= xOther.m_uSurfaceTypeHash;
#else
		m_uSurfaceType = xOther.m_uSurfaceType;
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
		m_bModified = true;
	}
}

void Tools_MaterialResponse::WriteTextureLineToStream( Asura_Chunk_Stream& xStream,
													   const bool bScale /*= false*/ ) const
{
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//
	// JAT 16/05/07 
	// if the data being output changes then Asura_Material_System::EXPORTED_TEXTURE_VERSION needs to be changed
	// also the PC, PS3, 360 resource file loading code needs updating to match the data that is written out by this function
	//
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	// Some platforms want to scale the graphs so they fit into [0,1]
	const float fDiffuseMax = m_xDiffuseGraph.GetMaxDiscreteValue();
	const float fSpecularMax = m_xSpecularGraph.GetMaxDiscreteValue();

	float fDiffuseScale = 1.0f / (bScale && fDiffuseMax > 0.0f ? fDiffuseMax : 100.0f);
	float fSpecularScale = 1.0f / (bScale && fSpecularMax > 0.0f ? fSpecularMax : 100.0f);

	Asura_RGBA xTexel;

	// First we'll have the graphs, one in each colour channel
	for ( u_int uElement=0; uElement<ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH; ++uElement )
	{
		xTexel.m_fR = m_xDiffuseGraph.GetDiscreteSample( uElement ) * fDiffuseScale;
		xTexel.m_fG = m_xSpecularGraph.GetDiscreteSample( uElement ) * fSpecularScale;
		xTexel.m_fB = m_xTransparencyGraph.GetDiscreteSample( uElement ) / 100.0f;
		xTexel.m_fA = 0.0f;

		xStream << xTexel;
	}

	// Write the last graph texel again. If any errors in the floating-point maths take the graph
	// read co-ords slightly out of bounds, we don't want to sample the colourise data
	xStream << xTexel;

	// Then the rest of the material's properties

	xStream << m_xDiffuseColour.m_fR;
	xStream << m_xDiffuseColour.m_fG;
	xStream << m_xDiffuseColour.m_fB;
	xStream << m_fColouriseDiffuse / 100.0f;

	xStream << m_xSpecularColour.m_fR;
	xStream << m_xSpecularColour.m_fG;
	xStream << m_xSpecularColour.m_fB;
	xStream << m_fColouriseSpecular / 100.0f;

	// Graph scales need to fit in [0,1]
	xStream << m_fMetalicity / 100.0f;
	xStream << fDiffuseMax / 200.0f;
	xStream << fSpecularMax / 200.0f;
	xStream << 0.0f;
}

bool Tools_MaterialResponse::HasSlaves() const
{
	const Asura_Hash_ID uHash = GetHash();

	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading(u);
		if ( !pxCurrMaterial->IsLoaded() ) continue;

		if( pxCurrMaterial->GetParentHash() == uHash )
		{
			return true;
		}
	}

	return false;
}

void Tools_MaterialResponse::ApplySpecularSharpness( const float fSpecularSharpness, const bool bOptimize )
{
	Asura_Graph<float> xTemp( ASURA_MATERIAL_RESPONSE_GRAPH_WIDTH, 0.0f, 100.0f, -1.0f, 1.0f );
	xTemp.SetSampleTolerance( 0.0f );

	m_xSpecularGraph = xTemp;
	m_xSpecularGraph.SetClampingValues( true, true, 0.0f, fTOOLS_MATERIAL_RESPONSE_GRAPH_MAX );
	m_xSpecularGraph.AddPoint( 0.0f, 0.0f );

	Asura_FixedArray<float, 256> afTimes;
	Asura_FixedArray<float, 256> afValues;

	for( int i = 0; i < 256; ++i )
	{
		const float fTime = static_cast< float >(i) / 255.0f;
		const float fValue = 100.0f * powf( fTime, fSpecularSharpness );
		afTimes[i] = fTime;
		afValues[i] = fValue;
		m_xSpecularGraph.AddPoint( fValue, fTime );
	}

	m_bModified = true;
	if ( bOptimize  == false )
	{
		m_xSpecularGraph.Update();//causes the discrete samples to be recalculated.
		return;
	}

	// Remove points based on a threshold between the straight line from prev point to next point e.g. error compared to bi-linear sampling
	// Don't use relative error because that sucks when the numbers are close to 0
	// The scale of values are 0->100
	const float fMaxError = 1.0f;

	// Don't try to optmize point 0.0 or point 1.0 away
	for( int i = 1; i < 255; ++i )
	{
		const float fTime = afTimes[i];
		const float fValue = afValues[i];

		bool bNeedPoint = false;
		// Skip first
		Asura_Graph_Point<float>* pxPoint = m_xSpecularGraph.GetControlPoints()->m_pxNext;
		// Skip last
		bool bFoundIt = false;
		while ( pxPoint->m_pxNext )
		{
			if ( ( pxPoint->m_fTime == fTime ) && ( pxPoint->m_xValue == fValue ) )
			{
				m_xSpecularGraph.RemovePoint( pxPoint );
				bFoundIt = true;
				break;
			}
			pxPoint = pxPoint->m_pxNext;
		}
		if ( bFoundIt == true )
		{
			for ( int i = 0; i < 256; ++i )
			{
				const float fTime = afTimes[i];
				const float fRealValue = afValues[i];

				const float fCurveValue = m_xSpecularGraph.GetValueAtTime( fTime );

				const float fError = Asura_Maths::Abs( fCurveValue - fRealValue );
				if ( fError > fMaxError )
				{
					bNeedPoint = true;
					break;
				}
			}
			if ( bNeedPoint == true )
			{
				m_xSpecularGraph.AddPoint( fValue, fTime );
			}
		}
		else
		{
			Asura_Assertf( false, ( "Didn't find point Time %f Value %f on the specular graph", fTime, fValue ) );
		}
	}
	m_xSpecularGraph.Update();//causes the discrete samples to be recalculated.
}

bool Tools_MaterialResponse::AllocateResponseRowIndex(void)
{
	if (GetHash()==ASURA_DEFAULT_MATERIAL_RESPONSE)
	{
		// It's the default material - put it in the last row
		m_uResponseRowIndex = ASURA_PC_MATERIALS_TEXTURE_HEIGHT-1;
		return true;
	}
	else
	{	
		m_uResponseRowIndex = Tools_MaterialResponse_System::GetNextResponseRowIndex();
	}

	// Out of space - point to the default material
	if (m_uResponseRowIndex >= ASURA_PC_MATERIALS_TEXTURE_HEIGHT - 1)
	{
		Asura_Assert( false, "Out of space on material response texture - Increase ASURA_PC_MATERIALS_TEXTURE_HEIGHT!" );

		m_uResponseRowIndex = ASURA_PC_MATERIALS_TEXTURE_HEIGHT - 1;
		return false;
	}
	else
	{
		return true;
	}	
};

Asura_Hash_ID Tools_MaterialResponse::CalculatePartialShaderHash() const
{
	const Tools_MaterialResponse* const pxParent =
		Tools_MaterialResponse_System::GetMaterialByHashID( GetParentHash() );

	Asura_Hash_ID uHash = 0;
	u_int uBit = 1;

	// Start with one bit per texture. Store them in the primary slots here. If combined as a
	// secondary material, these will get shifted up by one into the combined hash

	for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++u )
	{
		if( GetTexture(u) != ASURA_TEXTURE_NONE )
		{
			uHash |= uBit;
		}

		// Skip over the secondary texture bit
		uBit <<= 2;
	}

	// Now two bits each for colourise diffuse, colourise specular, and metalicity

	Asura_FixedArray< float, 3 > afPercents;

	afPercents[0] = (pxParent ? pxParent->GetColouriseDiffuse() : GetColouriseDiffuse());
	afPercents[1] = (pxParent ? pxParent->GetColouriseSpecular() : GetColouriseSpecular());
	afPercents[2] = (pxParent ? pxParent->GetMetalicity() : GetMetalicity());

	for( u_int u = 0; u < afPercents.GetSize(); ++u )
	{
		if( afPercents[u] <= 1.0f )
		{
			// Don't set the two bits
			uBit <<= 2;
		}
		else if( afPercents[u] >= 99.0f )
		{
			// Set the bits to 01
			uHash |= uBit;
			uBit <<= 2;
		}
		else
		{
			// Set the bits to 11
			uHash |= uBit;
			uBit <<= 1;
			uHash |= uBit;
			uBit <<= 1;
		}
	}

	// Now a bit for 'blend into albedo'

	const bool bDecal = (pxParent ? pxParent->GetIsDecal() : GetIsDecal());

	const Asura_MaterialResponse::BLEND_MODE eBlendMode =
		(pxParent ? pxParent->GetBlendMode() : GetBlendMode());

	if( bDecal && eBlendMode != Asura_MaterialResponse::BLEND_MODE_OPAQUE )
	{
		uHash |= uASURA_MRPS_HASH_BITMASK_BLEND_INTO_ALBEDO;
	}

	// And one for 'ignore specular'

	const bool bIgnoreSpecular =
		(pxParent ? pxParent->GetIgnoreSpecularPrelighting() : GetIgnoreSpecularPrelighting());

	if( bIgnoreSpecular )
	{
		uHash |= uASURA_MRPS_HASH_BITMASK_IGNORE_SPECULAR;
	}

	// Then one bit for opaque (deferred lighting) vs. transparent (forward render) strips

	const bool bTransparent = (pxParent ? pxParent->IsTransparent() : IsTransparent());

	if( bTransparent )
	{
		uHash |= uASURA_MRPS_HASH_BITMASK_TRANSPARENT;
	}

	// Finally a bit for skin vs. environment, set at runtime

	return uHash;
}

#ifdef ASURA_MAIN
void Tools_MaterialResponse::SetNeedsSave( bool bNeedsSave)
{
	if(bNeedsSave && !m_bNeedsSave)
	{
		//Warn if the file is read-only
		char szFilename[MAX_PATH];
		Asura_PrintfToBfr(szFilename,szMATERIALS_DIRECTORY "%s" szMATERIALS_EXTENSION,sizeof(szFilename),true, m_xName.GetString());

		Tools_SaveAsura::WarnIfReadOnly(szFilename);		
	}
	m_bNeedsSave = bNeedsSave;
}

bool Tools_MaterialResponse::CanSaveFile(bool bShowWarning)
{
	char szFilename[MAX_PATH];
	Asura_PrintfToBfr(szFilename,szMATERIALS_DIRECTORY "%s" szMATERIALS_EXTENSION,sizeof(szFilename),true, m_xName.GetString());

	return Tools_SaveAsura::CanSaveFile(szFilename, bShowWarning);
}

#endif //ASURA_MAIN


#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
