//
// Tools_VertexMove_Object.cpp
//
// Created by Richard Rice - 27/01/04
// 
// Copyright (C) 2004 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "MeshEdit/Tools_VertexMove_Object.h"


#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Module.h"

#include "Load&Save/Tools_SaveAsura.h"
#include "MeshEdit/Tools_Mode_MeshEdit.h"
#include "Objects/Tools_Object_System.h"
#include "Rendering/Tools_RenderModules.h"
#include "Tools_Overlays.h"


List<Tools_VertexMove_Object*> Tools_VertexMove_Object::s_xList;
static Asura_Vector_2 g_xOverlayPosition(20.0f,10.0f);

bool Tools_VertexMove_Object::s_bDontUpdateMeshes = false;

Tools_VertexMove_Object::Tools_VertexMove_Object(const List<ModuleVertexPair>& xVertexList)
:Tools_Object(TOT_VertexMove)
{
	s_xList.add_entry(this);
	Mesh = Asura_Mesh_System::GetMesh("ToolsShape_CutsceneTrackPoint");

	//copy vertex list into array , and determine the meshes used
	Asura_Assert(xVertexList.size(),"");
	m_iNumVertices = xVertexList.size();
	m_pxVertexList = new ModuleVertexPair[m_iNumVertices];
	int iIndex = 0;

	List<int> xMeshList;
	for(CLIF<ModuleVertexPair> xVertLIF(&xVertexList);!xVertLIF.done();xVertLIF.next())
	{
		m_pxVertexList[iIndex++] = xVertLIF();

		if(!xMeshList.contains(xVertLIF().iModuleIndex))
		{
			xMeshList.add_entry(xVertLIF().iModuleIndex);
		}
	}

	m_iNumMeshes = xMeshList.size();
	m_piMeshList = new int[m_iNumMeshes];
	iIndex = 0;
	for(LIF<int> xMeshLIF(&xMeshList);!xMeshLIF.done();xMeshLIF.next())
	{
		m_piMeshList[iIndex++] = xMeshLIF();
	}

	//choose a default position and orientation
	Orientation.Identity();
	//get position from first vertex
	Asura_Module_System::GetPositionOfModule(m_pxVertexList[0].iModuleIndex,Position);
	Asura_Mesh* m_pxModuleMesh = Asura_Module_System::GetMeshForModule(m_pxVertexList[0].iModuleIndex);
	Position += m_pxModuleMesh->VertexDataPtr[m_pxVertexList[0].iVertexIndex].Position;

	m_xPreviousPosition = Position;
	m_xPreviousOrientation = Orientation;

	m_bUseOrientation = true;
	m_bUnmerged = false;
	m_bMeshesNeedToBeInvalidated = false;
}

Tools_VertexMove_Object::~Tools_VertexMove_Object()
{
	if(m_bMeshesNeedToBeInvalidated) InvalidateMeshes();

	delete [] m_pxVertexList;
	delete [] m_piMeshList;
	s_xList.delete_entry(this);
}

void Tools_VertexMove_Object::DeleteAll()
{
	while(s_xList.size())
	{
		Tools_Object_System::DeleteObject(s_xList.first_entry());
	}
}

void Tools_VertexMove_Object::UpdatePositionInAsuraData(bool /*bUpdateLinkedObjects*/)
{
	//determine the rotation to apply
	Asura_Matrix_3x3 xRelativeOrient;
	xRelativeOrient.Identity();
	if(m_bUseOrientation)
	{
		if(m_xPreviousOrientation != Orientation)
		{ 
			xRelativeOrient = m_xPreviousOrientation;
			xRelativeOrient.Transpose();
			xRelativeOrient =  Orientation*xRelativeOrient;
		}
	}

	if(Position != m_xPreviousPosition || xRelativeOrient!= Asura_Maths::IdentityMatrix3x3)
	{
		for(int i=0;i<m_iNumVertices;i++)
		{
			int iModuleIndex = m_pxVertexList[i].iModuleIndex;
			int iVertexIndex = m_pxVertexList[i].iVertexIndex;
			//get the module's mesh and position
			Asura_Assert(iModuleIndex<Asura_Module_System::GetNumberOfModules(),"");
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModuleIndex);
			Asura_Vector_3 xModulePosition;
			Asura_Module_System::GetPositionOfModule(iModuleIndex,xModulePosition);

			//get the vertex's position relative to the start point of the movement
			Asura_Assert(iVertexIndex<pxMesh->NumberOfVertices,"");
			Asura_Vector_3 xRelativePosition = xModulePosition+pxMesh->VertexDataPtr[iVertexIndex].Position - m_xPreviousPosition;

			//rotate if needed
			if(m_bUseOrientation)
			{
				xRelativePosition = xRelativePosition*xRelativeOrient;
			}

			//Use this as a position relative to the end point , and convert back to a vertex position
			pxMesh->VertexDataPtr[iVertexIndex].Position = xRelativePosition+Position - xModulePosition;
		}

		UnmergePolygonsIfNeeded();

		//update the meshes
		m_bMeshesNeedToBeInvalidated = true;

	}

	m_xPreviousPosition = Position;
	m_xPreviousOrientation = Orientation;

}

void Tools_VertexMove_Object::SetYPositionForVertices(float fY)
{
	for(int i=0;i<m_iNumVertices;i++)
	{
		int iModuleIndex = m_pxVertexList[i].iModuleIndex;
		int iVertexIndex = m_pxVertexList[i].iVertexIndex;
		//get the module's mesh and position
		Asura_Assert(iModuleIndex<Asura_Module_System::GetNumberOfModules(),"");
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModuleIndex);
		Asura_Vector_3 xModulePosition;
		Asura_Module_System::GetPositionOfModule(iModuleIndex,xModulePosition);

		//set the new position
		pxMesh->VertexDataPtr[iVertexIndex].Position.y = fY - xModulePosition.y;
	}

	UnmergePolygonsIfNeeded();

	m_bMeshesNeedToBeInvalidated = true;	

}


void Tools_VertexMove_Object::OnSelection()
{
	Tools_Object_System::EnableAllControls(true);
	CreateOverlay();

}

void Tools_VertexMove_Object::OnDeselection()
{
	DeleteOverlay();
	Tools_Object_System::EnableAllControls(false);
}

void Tools_VertexMove_Object::CreateOverlay()
{
	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Vertex Movement",false,&g_xOverlayPosition);
	
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Tickbox_Dialog("Use Orientation",xOffset,&m_bUseOrientation);
	pxDialog->SetHelpText("If this is ticked then rotating the vertex movement object will cause the attached vertices to be rotated as well.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Don't Update Meshes",xOffset,&s_bDontUpdateMeshes);
	pxDialog->SetHelpText("Prevents the continual update of the meshes as the vertices move ,in an attempt to get a slightly better frame rate\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxOverlay->m_uID = TOOLS_OVERLAY_SELECTEDOBJECT;
 	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}

void Tools_VertexMove_Object::DeleteOverlay()
{
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_SELECTEDOBJECT);
}

void Tools_VertexMove_Object::Update()
{
	Tools_Object::Update();

	if(m_bMeshesNeedToBeInvalidated && !s_bDontUpdateMeshes)
	{
		InvalidateMeshes();
	}
}

Asura_Vector_3 Tools_VertexMove_Object::GetSnapPoint(const Asura_Vector_3& xPos) const
{
	Asura_Vector_3 xSnapPos = Position;
	float fMinDist = ASURA_LOS_MAX_DISTANCE;

	Asura_Matrix_3x3 xRelativeOrient;
	xRelativeOrient.Identity();
	if(m_bUseOrientation)
	{
		if(m_xPreviousOrientation != Orientation)
		{ 
			xRelativeOrient = m_xPreviousOrientation;
			xRelativeOrient.Transpose();
			xRelativeOrient =  Orientation*xRelativeOrient;
		}
	}
	

	//Get the nearest vertex to the point.
	//Take into account any changes to position/orientation that haven't been applied yet
	for(int i = 0;i<m_iNumVertices;i++)
	{
		int iModuleIndex = m_pxVertexList[i].iModuleIndex;
		int iVertexIndex = m_pxVertexList[i].iVertexIndex;
		//get the module's mesh and position
		Asura_Assert(iModuleIndex<Asura_Module_System::GetNumberOfModules(),"");
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModuleIndex);
		Asura_Vector_3 xModulePosition;
		Asura_Module_System::GetPositionOfModule(iModuleIndex,xModulePosition);

		//get the vertex's position relative to the start point of the movement
		Asura_Assert(iVertexIndex<pxMesh->NumberOfVertices,"");
		Asura_Vector_3 xRelativePosition = xModulePosition+pxMesh->VertexDataPtr[iVertexIndex].Position - m_xPreviousPosition;

		//rotate if needed
		if(m_bUseOrientation)
		{
			xRelativePosition = xRelativePosition*xRelativeOrient;
		}

		//Use this as a position relative to the end point
		Asura_Vector_3 xVertPos = xRelativePosition+Position;

		float fDist = (xVertPos-xPos).Magnitude();
		if(fDist<fMinDist)
		{
			fMinDist = fDist;
			xSnapPos = xVertPos;
		}
	}
	return xSnapPos;
}

void Tools_VertexMove_Object::UnmergePolygonsIfNeeded()
{
	if(m_bUnmerged) return;
	m_bUnmerged = true;
	//check for any quads that have had just some of their vertices moved , and unmerge them

	for(int i=0;i<m_iNumMeshes;i++)
	{
		int iModuleIndex = m_piMeshList[i];
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModuleIndex);

		//build array of vertices moved in this mesh
		bool* pbVertMoved = new bool[pxMesh->NumberOfVertices];
		for(int j=0;j<pxMesh->NumberOfVertices;j++)
		{
			pbVertMoved[j] = false;
		}
		for(int j=0;j<m_iNumVertices;j++)
		{
			if(m_pxVertexList[j].iModuleIndex == iModuleIndex)
			{
				Asura_Assert(m_pxVertexList[j].iVertexIndex<pxMesh->NumberOfVertices,"");
				pbVertMoved[m_pxVertexList[j].iVertexIndex] = true;
			}
		}

		//now check each polygon , and find those that have been partially moved
		bool* pbUnmergeArray = new bool[pxMesh->NumberOfPolygons];
		int iNumToUnmerge = 0;
		for(int j=0;j<pxMesh->NumberOfPolygons;j++)
		{
			pbUnmergeArray[j] = false;
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[j];
			if(pxPoly->NumberOfVertices() != 4) continue;

			int iMoved = 0;
			for(int k=0;k<4;k++)
			{
				if(pbVertMoved[pxPoly->VertexID[k]])
				{
					iMoved ++;
				}
			}
			if(iMoved>0 && iMoved<4)
			{
				//this polygon will need unmerging
				//(well , it might be safe not to , but easiest just to do so anyway)
				pbUnmergeArray[j] = true;
				iNumToUnmerge++;
			}
		}
		delete [] pbVertMoved;
		pbVertMoved = 0;

		if(iNumToUnmerge)
		{
			//unmerge the polygons
			Tools_Mode_MeshEdit::UnmergePolygons(iModuleIndex,pbUnmergeArray);
		}
		delete [] pbUnmergeArray;

	}

}

void Tools_VertexMove_Object::InvalidateMeshes()
{
	//update the meshes
	for(int i=0;i<m_iNumMeshes;i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(m_piMeshList[i]);

		pxMesh->InvalidateNormals();
		pxMesh->BoundingBoxDataValid = false;
		
		pxMesh->CreateNormals();
		pxMesh->CalculateBoundingBox();

		Asura_Mesh_System::PlatformResetRenderMesh(pxMesh);
		Tools_RenderModule_System::UpdateModuleMesh(pxMesh->Index);
		Tools_Mode_MeshEdit::InvalidateModuleCollisionMesh(m_piMeshList[i]);

		Tools_SaveAsura::SetMeshModified(pxMesh);
		
	}


	m_bMeshesNeedToBeInvalidated = false;
}