//
// // Tools_AAEdges.cpp
//
// Copyright (C) 2009 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "MeshProcessing/Tools_AAEdges.h"

// Independent
#include "Rendering/Asura_Module.h"

// Tools
#include "MeshProcessing/Tools_EnvironmentProcessor.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Misc/Tools_Adjacency.h"


Tools_AAEdges::Tools_AAEdges()
{
	m_iTotalNumEdgeIndices	= 0;
	m_iNumEdgeIndices		= NULL;
	m_puEdgeIndices			= NULL;

	m_axAllocatedEdges		= NULL;
	m_apxSpatialHash		= NULL;
	m_puNumAAEdgesPerStrip	= NULL;

	m_pxSortedEdges			= NULL;
	m_iMappedEdgeCount		= -1;
}

Tools_AAEdges::~Tools_AAEdges()
{
	for ( int i=0; i < Asura_Module_System::GetNumberOfModules(); i++ )
	{
		delete[] m_puEdgeIndices[i];
	}
	delete[] m_puEdgeIndices;
	delete[] m_iNumEdgeIndices;
	delete[] m_axAllocatedEdges;
	delete[] m_apxSpatialHash;
	delete[] m_puNumAAEdgesPerStrip;
}

void Tools_AAEdges::Create( const Asura_Bounding_Box& xBB, const u_int uNumberOfStrips )
{
	Asura_Assert( m_iTotalNumEdgeIndices == 0 && m_iNumEdgeIndices == NULL && m_puEdgeIndices == NULL, "Calling Create twice?" );

	m_iTotalNumEdgeIndices	= 0;
	m_iNumEdgeIndices		= new int[Asura_Module_System::GetNumberOfModules()];
	m_puEdgeIndices			= new uint16*[Asura_Module_System::GetNumberOfModules()];

	for ( int iModule = 0; iModule < Asura_Module_System::GetNumberOfModules(); iModule++ )
	{		
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModule);

		Asura_Assert( pxMesh, "FindAntiAliasEdges - this mesh is NULL, invalid object perhaps?" );

		m_iNumEdgeIndices[iModule] = FindAntiAliasEdges( pxMesh, &m_puEdgeIndices[iModule] );

		m_iTotalNumEdgeIndices += m_iNumEdgeIndices[iModule];
	}

	// Find which strip each anti-alias edge belongs to (some will be on 2 strips, just pick the first we encounter)
	m_puNumAAEdgesPerStrip = new u_int[uNumberOfStrips];
	Asura_Mem::Set( m_puNumAAEdgesPerStrip, 0, sizeof(int)*uNumberOfStrips );

	Asura_Vector_3 xBBMin = xBB.GetMin();
	Asura_Vector_3 xBBRange = xBB.GetMax()-xBBMin;

	m_iGridDim = (int)pow( (float)m_iTotalNumEdgeIndices/2.0f, 1.0f/3.0f ) + 1; // a 3D grid with approximately as many cells as there are points (should give pretty efficient result so long as they're fairly evenly distributed)

	m_xSpatialOffset	= xBBMin;
	m_xSpatialInvRange	= Asura_Vector_3( ((float)m_iGridDim - 0.001f) / xBBRange.GetX(), ((float)m_iGridDim - 0.001f) / xBBRange.GetY(), ((float)m_iGridDim - 0.001f) / xBBRange.GetZ() );

	m_axAllocatedEdges	= new Tools_LinkedAntiAliasEdge[m_iTotalNumEdgeIndices/2];
	m_apxSpatialHash	= new Tools_LinkedAntiAliasEdge*[m_iGridDim*m_iGridDim*m_iGridDim];

	Asura_Mem::Set( m_apxSpatialHash, NULL, sizeof(Tools_LinkedAntiAliasEdge*)*m_iGridDim*m_iGridDim*m_iGridDim );

	// build a list of where all the edges are, so we can efficiently look them up by position
	Tools_LinkedAntiAliasEdge* pxEdge = m_axAllocatedEdges;
	for ( int iModule = 0; iModule < Asura_Module_System::GetNumberOfModules(); iModule++ )
	{		
		Asura_Vector_3 xModulePosition = Asura_Module_System::GetPositionOfModule(iModule);

		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModule);
		for ( int iEdge = 0; iEdge < m_iNumEdgeIndices[iModule]; iEdge += 2 )
		{
			// get positions of vertices
			// indices are different between strips & meshes, but positions are exactly equal
			uint16 uEdgeIdx0 = m_puEdgeIndices[iModule][iEdge];
			uint16 uEdgeIdx1 = m_puEdgeIndices[iModule][iEdge+1];

			//this has partly fixed the problem, it got up to 7100 edges (10%) before hitting problems => precision.
			// positions must be offset to the world
			Asura_Vector_3 xEdgeP0 = pxMesh->VertexDataPtr[uEdgeIdx0].Position + xModulePosition;
			Asura_Vector_3 xEdgeP1 = pxMesh->VertexDataPtr[uEdgeIdx1].Position + xModulePosition;

			// use centre point of edges, so we can efficiently look things up by position
			Asura_Vector_3 xCentre = (xEdgeP0 + xEdgeP1)/2.0f;

			// add it to the spatial hash				
			xCentre -= m_xSpatialOffset;
			xCentre = xCentre.Vector_Multiply(m_xSpatialInvRange);

			// to avoid possible crashes from rounding errors or extraneous data, clamp it to the range.
			int x = (int)xCentre.x; int y = (int)xCentre.y; int z = (int)xCentre.z;
			x = min(m_iGridDim-1,max(0,x));
			y = min(m_iGridDim-1,max(0,y));
			z = min(m_iGridDim-1,max(0,z));

			int iSpatialIndex = x + y*m_iGridDim + z*m_iGridDim*m_iGridDim;

			pxEdge->m_iParentIndex = iModule;
			pxEdge->m_auIdx[0] = uEdgeIdx0;
			pxEdge->m_auIdx[1] = uEdgeIdx1;
			pxEdge->m_pxNext = m_apxSpatialHash[iSpatialIndex];
			m_apxSpatialHash[iSpatialIndex] = pxEdge;

			pxEdge++;
		}
	}
}

void Tools_AAEdges::BeginStripsMapping()
{
	Asura_Assert( m_iMappedEdgeCount == -1, "BeginStripsMapping can be called only once" );
	m_pxSortedEdges = NULL;
	m_iMappedEdgeCount = 0;
}

void Tools_AAEdges::MapStrip( const Tools_EnvironmentProcessor_Strip& xStrip, int iStripId )
{
	for (int k = 1; k <= 2; k++ )
	{
		// k=1 => zig-zag edges
		// k=2 => parallel (outer) edges
		for (u_int j = k; j < xStrip.m_uNumberOfIndices; j++)
		{
			uint16 uStripIdx0 = xStrip.m_pusIndices[j];
			uint16 uStripIdx1 = xStrip.m_pusIndices[j-k];

			if ( uStripIdx0 == 0xffff ) continue;
			if ( uStripIdx1 == 0xffff ) continue;

			Asura_Vector_3 xStripP0 = xStrip.m_pxVertices[uStripIdx0].m_xPosition;
			Asura_Vector_3 xStripP1 = xStrip.m_pxVertices[uStripIdx1].m_xPosition;

			// use centre point of edges, so we can efficiently look things up by position
			Asura_Vector_3 xCentre = (xStripP0 + xStripP1)/2.0f;


			// Find it if it exists in the table.
			xCentre -= m_xSpatialOffset;
			xCentre = xCentre.Vector_Multiply(m_xSpatialInvRange);

			int iSpatialIndex = (int)xCentre.x+(int)xCentre.y*m_iGridDim+(int)xCentre.z*m_iGridDim*m_iGridDim;

			Tools_LinkedAntiAliasEdge* pxEdge = m_apxSpatialHash[iSpatialIndex];
			Tools_LinkedAntiAliasEdge** pxPrevNext = &m_apxSpatialHash[iSpatialIndex];

			while ( pxEdge )
			{						
				Tools_LinkedAntiAliasEdge* pxNextEdge = pxEdge->m_pxNext;

				// this lookup is a bit wasteful, but so would be storing the positions in the Asura_Xbox360_Env_LinkedAntiAliasEdge struct
				Asura_Vector_3 xModulePosition = Asura_Module_System::GetPositionOfModule(pxEdge->m_iParentIndex);
				Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(pxEdge->m_iParentIndex);
				Asura_Vector_3 xEdgeP0 = pxMesh->VertexDataPtr[pxEdge->m_auIdx[0]].Position + xModulePosition;
				Asura_Vector_3 xEdgeP1 = pxMesh->VertexDataPtr[pxEdge->m_auIdx[1]].Position + xModulePosition;

				if ( ( xEdgeP0 == xStripP0 && xEdgeP1 == xStripP1 ) || ( xEdgeP0 == xStripP1 && xEdgeP1 == xStripP0 ) )
				{
					// mark edge as belonging to this strip
					pxEdge->m_iParentIndex = iStripId;

					// change indices to be in same range as strip's indices (then we can maintain them whenever we change the strip's ones)
					pxEdge->m_auIdx[0] = uStripIdx0;
					pxEdge->m_auIdx[1] = uStripIdx1;

					// 									pxSortedEdges[iEdgeId].m_xVerify0 = xStripP0;
					// 									pxSortedEdges[iEdgeId].m_xVerify1 =	xStripP1;

					// increment strip's aa edge counter. so offset can be computed later.
					m_puNumAAEdgesPerStrip[iStripId]++;

					// remove from hash list (so we only draw each AA edge once)
					*pxPrevNext = pxEdge->m_pxNext;

					// add to output list
					pxEdge->m_pxNext = m_pxSortedEdges; // list is built in reverse order which is why loop counters count backwards
					m_pxSortedEdges = pxEdge;

					m_iMappedEdgeCount++;
				}
				else
				{
					pxPrevNext = &pxEdge->m_pxNext;
				}

				pxEdge = pxNextEdge;
			}

		}
	}
}

void Tools_AAEdges::EndStripsMapping()
{
	Asura_Assertf( m_pxSortedEdges != NULL || m_iTotalNumEdgeIndices == 0, ( "Couldn't map any anti-alias edges, there should be %d", m_iTotalNumEdgeIndices/2 ) );

	Tools_Level_Export::AddMessage( Tools_Level_Export::MT_INFORMATION, "Environment has %d anti-aliased edges [from %d potential edges]", m_iMappedEdgeCount, m_iTotalNumEdgeIndices/2 );

	// if some edges couldn't be found, forget them
	m_iTotalNumEdgeIndices = m_iMappedEdgeCount*2;
}

int Tools_AAEdges::FindAntiAliasEdges( const Asura_Mesh* pxMesh, uint16** ppuOutput ) 
{
	if ( !pxMesh )
	{
		// marine's pistol has a null mesh - this doesn't seem to cause a problem.
		// EP.9.DEC.2008 I blocked this assert and put message cast before calling of this function in one case.
		// Asura_Assert( 0, "FindAntiAliasEdges - this mesh is NULL, invalid object perhaps?" );
		return 0;
	}

	if (!Tools_Adjacency::IsInitialised())
	{
		Tools_Adjacency::Initialise();
	}

	Tools_MeshAdjacency xMeshAdjancy; 
	xMeshAdjancy.Initialise( pxMesh );

	uint16* puFoundEdges = new uint16[pxMesh->NumberOfPolygons*4*2]; //enough for every poly to be a quad with it's own edges
	uint16* puOutEdge = puFoundEdges;
	int iTotalNumEdges = 0;

	for ( int iPolyA=0; iPolyA < pxMesh->NumberOfPolygons; iPolyA++ )
	{
		const Tools_Polygon_Adjacency* pxPolyAdj = xMeshAdjancy.GetPolygonAdjacencies( iPolyA );

		uint16 uPointsA[] =
		{
			pxMesh->PolygonDataPtr[iPolyA].VertexID[0],
			pxMesh->PolygonDataPtr[iPolyA].VertexID[1],
			pxMesh->PolygonDataPtr[iPolyA].VertexID[2],
			pxMesh->PolygonDataPtr[iPolyA].VertexID[3], // 0xffff if it's a triangle
		};

		u_int uNumPointsA = pxMesh->PolygonDataPtr[iPolyA].NumberOfVertices();

		bool abRawEdges[4] = { true,true,true,true }; //flag per edge, assume they're raw until an adjacent poly for each is found. Numbering: [0] = point0 to point1, ... [n] = pointn to point0, where n is 3 or 4

		int iNumAdjacentEdgePolys = 0;
		//int iNumberOfAdjacentPolys = (pxPolyAdj->GetNumAdjacentPolygons());
		for(const Tools_Adjacent_Polygon* pxAdj = pxPolyAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
		{
			// ignore polys that only meet at a vertex rather than an edge
			if (pxAdj->GetNumSharedVertices() < 2) continue;

			iNumAdjacentEdgePolys++;

			const int iPolyB = pxAdj->GetPolygonIndex();

			iTotalNumEdges++;

			uint16 uPointsB[] =
			{
				pxMesh->PolygonDataPtr[iPolyB].VertexID[0],
				pxMesh->PolygonDataPtr[iPolyB].VertexID[1],
				pxMesh->PolygonDataPtr[iPolyB].VertexID[2],
				pxMesh->PolygonDataPtr[iPolyB].VertexID[3], // 0xffff if it's a triangle
			};

			u_int uNumPointsB = pxMesh->PolygonDataPtr[iPolyB].NumberOfVertices();


			// every edge where these polys meet is not a raw edge
			for ( u_int i=0; i < uNumPointsA; i++ )
			{
				if ( abRawEdges[i] ) // skip the ones we've already marked as not-raw
				{
					uint16 p0 = uPointsA[i];
					uint16 p1 = uPointsA[(i+1)%uNumPointsA];
					for ( u_int j=0; j < uNumPointsB; j++ )
					{
						if ( ( p0 == uPointsB[j] && p1 == uPointsB[(j+1)%uNumPointsB] )
							|| ( p1 == uPointsB[j] && p0 == uPointsB[(j+1)%uNumPointsB] )  )
						{
							abRawEdges[i] = false;
						}
					}
				}
			}


			if ( iPolyB < iPolyA ) continue; // we've already done this poly 


			// decide if this edges needs anti-aliasing, using several criteria
			bool bAAEdge = false;

			// are materials or normals discontinuous?
			bAAEdge = bAAEdge || pxMesh->GetSmoothingGroup( iPolyA ) != pxMesh->GetSmoothingGroup( iPolyB );

			// this test is invalid, there's a table of per-poly data. It's not vital that we AA the edges between textures (artists will usually make them look seamless anyway)
			// 			bAAEdge = bAAEdge || pxMesh->GetTextureInfoForExport( iPolyA ) != pxMesh->GetTextureInfoForExport( iPolyB );

			// is this edge convex?
			if ( !bAAEdge )
			{
				Asura_Vector_3 xA1 = pxMesh->VertexDataPtr[uPointsA[1]].Position - pxMesh->VertexDataPtr[uPointsA[0]].Position;
				Asura_Vector_3 xA2 = pxMesh->VertexDataPtr[uPointsA[2]].Position - pxMesh->VertexDataPtr[uPointsA[0]].Position;
				xA1.Normalise();
				xA2.Normalise();
				Asura_Vector_3 xAFaceNormal = (xA1^xA2);
				xAFaceNormal.Normalise();

				//120,000 from 144,000, meaning this is not removing the 50% I expect.
				// still getting all the damn edges in the cave
				// test convex (points in B are behind points in A)
				Asura_Vector_3 xAB = pxMesh->VertexDataPtr[uPointsB[0]].Position + pxMesh->VertexDataPtr[uPointsB[1]].Position + pxMesh->VertexDataPtr[uPointsB[2]].Position;
				xAB /= 3.0f;
				xAB -= pxMesh->VertexDataPtr[uPointsA[0]].Position;
				float fDot = xAB * xAFaceNormal;
				if ( fDot < 0.0f )
				{
					Asura_Vector_3 xB1 = pxMesh->VertexDataPtr[uPointsB[1]].Position - pxMesh->VertexDataPtr[uPointsB[0]].Position;
					Asura_Vector_3 xB2 = pxMesh->VertexDataPtr[uPointsB[2]].Position - pxMesh->VertexDataPtr[uPointsB[0]].Position;
					xB1.Normalise();
					xB2.Normalise();
					Asura_Vector_3 xBFaceNormal = (xB1^xB2);
					xBFaceNormal.Normalise();

					// test angle
					fDot = xAFaceNormal*xBFaceNormal;
					if (fDot < 0.996f)	bAAEdge	= true; //about 5 degrees
				}

			}



			if (!bAAEdge) continue;

			// add every shared edge to our list
			for ( u_int i=0; i < uNumPointsA; i++ )
			{
				uint16 p0 = uPointsA[i];
				uint16 p1 = uPointsA[(i+1)%uNumPointsA];
				for ( u_int j=0; j < uNumPointsB; j++ )
				{
					if ( ( p0 == uPointsB[j] && p1 == uPointsB[(j+1)%uNumPointsB] )
						|| ( p1 == uPointsB[j] && p0 == uPointsB[(j+1)%uNumPointsB] )  )
					{
						*(puOutEdge++) = p0;
						*(puOutEdge++) = p1;
					}
				}
			}

		}

		// output any raw edges
		for ( u_int i=0; i < uNumPointsA; i++ )
		{
			if ( abRawEdges[i] )
			{
				*(puOutEdge++) = uPointsA[i];
				*(puOutEdge++) = uPointsA[(i+1)%uNumPointsA];
			}
		}

	}

	int iNumIndices = (puOutEdge - puFoundEdges);

	//	Tools_Level_Export::AddMessage( Tools_Level_Export::MT_INFORMATION, "Found %d/%d edges for anti-aliasing", iNumIndices/2, iTotalNumEdges );

	//v1 - just output the line list, we'll switch to line strips later.
	*ppuOutput = puFoundEdges;

	return iNumIndices;
}
