//
// Tools_EnvironmentProcessor.cpp
//
// Created by KJL - 27/02/2003 11:27:32
//
// Copyright (C) 2003 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "MeshProcessing/Tools_EnvironmentProcessor.h"

#include "Core/Asura_Material_System.h"
#include "Core/Asura_Timers.h"
#include "Core/Memory/Asura_Mem.h"
#include "Debugging/Asura_Debug_Timers.h"
#include "Light/Asura_Light.h"
#include "Mesh/Asura_Mesh.h"
#include "MeshProcessing/Asura_TriStripper.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Surface.h"

#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Materials/Tools_MaterialResponse_System.h"
#include "MeshProcessing/Tools_MeshRestartIndex.h"
#include "MeshProcessing/Tools_SubdividedMesh.h"
#include "MeshProcessing/Tools_TangentSpace.h"
#include "Misc/Tools_HashTable.h"

Tools_EnvironmentProcessor_Mesh*	Tools_EnvironmentProcessor::s_pxProcessedMesh;
u_int								Tools_EnvironmentProcessor::s_uTotalNumberOfVertices;
u_int								Tools_EnvironmentProcessor::s_uTotalNumberOfIndices;
u_int								Tools_EnvironmentProcessor::s_uTotalNumberOfStrips;
u_int								Tools_EnvironmentProcessor::s_uNumberOfMeshes;
bool								Tools_EnvironmentProcessor::s_bBeenInitialised = false;

Tools_EnvironmentProcessor_PosMesh**	Tools_EnvironmentProcessor::s_ppxPosMesh = NULL;
bool								Tools_EnvironmentProcessor::s_bBeenGrouped = false;
u_int								Tools_EnvironmentProcessor::s_uTotalNumberOfPositions;

Tools_EnvironmentProcessor_Strip*	Tools_EnvironmentProcessor::s_pxStripsSortedByMaterial;

u_int								Tools_EnvironmentProcessor::s_uTotalDegenerateSourceTris = 0;
u_int								Tools_EnvironmentProcessor::s_uTotalDuplicateSourceTris = 0;
PROJECT_LIGHTING_FN					Tools_EnvironmentProcessor::s_pfnProjectLightingCallback = 0;

Tools_EnvironmentProcessor::ColourAdjustmentCallback*		Tools_EnvironmentProcessor::s_pfnColourAdjustmentCallback = NULL;
Tools_EnvironmentProcessor::PacketSelectionCallback*		Tools_EnvironmentProcessor::s_pfnPacketSelectionCallback = NULL;
Tools_EnvironmentProcessor::StripFlagCallback*				Tools_EnvironmentProcessor::s_pfnStripFlagCallback = NULL;

void Tools_EnvironmentProcessor::Initialise(bool bShouldRemoveDuplicateVertices /* = false */, bool bShouldStripify /* = true */, bool bShouldSubdivide /* = true */, const u_int uPolygonFlagsToIgnor)
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_EnvironmentProcessor::Initialise");
	
	Asura_Assert( !s_bBeenInitialised, "Already Initialised!" );
	
    Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY, "Processing environment..." );
    
    // Record how long it takes.
    const u_int uStartTime = Asura_Timers::GetMSecsSinceStart_Continuous();

	s_pxStripsSortedByMaterial = 0;

	s_uNumberOfMeshes = Asura_Module_System::GetNumberOfModules();

	if (s_uNumberOfMeshes==0)
	{
		Asura_Assert( false, "No modules found");
		return;
	}

	s_pxProcessedMesh = new Tools_EnvironmentProcessor_Mesh[s_uNumberOfMeshes];
	Asura_Assert( s_pxProcessedMesh, "Out of memory" );

	//Force all the surface meshes to be invalidated , since they may contain out of date UV data etc
	//RAR 31/08/2005
	{
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_EnvironmentProcessor::Initialise InvalidateAll" );
		Asura_Surface_Mesh_System::InvalidateAll();
	}

	// (AAM - 23.06.2004) Don't want to force this on all the time! Get it from the export options instead
//	Tools_SubdividedMesh::s_bRecalculatePrelighting = true;

	{
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_EnvironmentProcessor::Initialise CalculatePrelighting" );
		if(s_pfnProjectLightingCallback)
		{
			s_pfnProjectLightingCallback();
		}
		else
		{
			// KJFL - 12 May 2005 - if we're doing a quick trilist 'export' make sure the lighting is up to date.
			for(int iModule = 0; iModule < Asura_Module_System::GetNumberOfModules(); iModule++)
			{
				Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(iModule);
				const Asura_Vector_3& xPosition = Asura_Module_System::GetPositionOfModule(iModule);
				if(!pxMesh->PrelightingDataValid)
				{
					Asura_Lighting_System::CalculatePrelighting(*pxMesh, xPosition, iModule);
				}
			}
		}
	}

	s_uTotalNumberOfVertices = 0;
	s_uTotalNumberOfIndices = 0;
	s_uTotalNumberOfStrips = 0;

    long iTotalVerticesBeforeOptimisation = 0;
	long iTotalVerticesOptimisedAway = 0;
	s_uTotalDuplicateSourceTris = 0;
	s_uTotalDegenerateSourceTris = 0;

    if( bShouldStripify )
    {
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_EnvironmentProcessor::Initialise Tools_TriStripper::Initialise" );
        Tools_TriStripper::Initialise();
        Tools_TriStripper::SetOutputType( Tools_TriStripper::OUTPUT_TRISTRIP_WITH_RESET_INDICES );
    }
	


	// KJL - 28/02/2003 09:40:04 - stripify each module, after subdivision and (if necessary) prelighting
	{
		ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_EnvironmentProcessor::Initialise StripifyEachModule" );
		for (u_int u=0; u<s_uNumberOfMeshes && !Tools_Level_Export::IsAbortFlagSet(); u++)
		{
			Tools_SubdividedMesh* pxMesh = 0;

			//During exports , ignore any modules that don't match the current light set
			//RAR 12/12/2007
			if(!Tools_Level_Export::ShouldIncludeModuleInExport(u))
			{
				s_pxProcessedMesh[u].m_uTotalNumberOfVertices = 0;
				s_pxProcessedMesh[u].m_uTotalNumberOfIndices = 0;
				s_pxProcessedMesh[u].m_uNumberOfStrips = 0;
				continue;
			}
			
			if(bShouldRemoveDuplicateVertices)
			{
				Tools_Level_Export::UpdateStatus(Tools_Level_Export::EXPORT_STAGE_STRIPIFYING_ENVIRONMENT, "Subdividing, optimizing, stripifying environment...", (u*3 + 0) * 100 / (s_uNumberOfMeshes * 3));
				pxMesh = new Tools_SubdividedMesh(u,true,true,bShouldSubdivide, uPolygonFlagsToIgnor);

				#if defined( ASURA_ASSERT)
				{
					u_int uPacketCount = pxMesh->NumberOfPackets;
					for( u_int uPacketIndex = 0; uPacketIndex < uPacketCount; uPacketIndex += 1)
					{
						const Tools_SubdividedMesh_Packet& xPacket = pxMesh->Packets[ uPacketIndex];
						const u_int uIndexCount = 3 * xPacket.NumberOfTris;
						
						for( u_int uIndexIndex = 0; uIndexIndex < uIndexCount; uIndexIndex += 1)
						{
							const u_int uVertexIndex = xPacket.TriList[ uIndexIndex];
							Asura_Assertf( uVertexIndex < static_cast< u_int>( xPacket.NumberOfVertices), ("Vertex Index %d out of %d range. Mesh '%s'", uVertexIndex, xPacket.NumberOfVertices, pxMesh->m_pxMesh->Name.GetString()));
						}
					}
				}
				#endif

				Tools_Level_Export::UpdateStatus(Tools_Level_Export::EXPORT_STAGE_STRIPIFYING_ENVIRONMENT, "Subdividing, optimizing, stripifying environment...", (u*3 + 1) * 100 / (s_uNumberOfMeshes * 3));
				int iBefore = pxMesh->TotalVertices;
				ProcessVerticesAndEliminateDuplicates(*pxMesh);
				
				#if defined( ASURA_ASSERT)
				{
					u_int uPacketCount = pxMesh->NumberOfPackets;
					for( u_int uPacketIndex = 0; uPacketIndex < uPacketCount; uPacketIndex += 1)
					{
						const Tools_SubdividedMesh_Packet& xPacket = pxMesh->Packets[ uPacketIndex];
						const u_int uIndexCount = 3 * xPacket.NumberOfTris;
						
						for( u_int uIndexIndex = 0; uIndexIndex < uIndexCount; uIndexIndex += 1)
						{
							const u_int uVertexIndex = xPacket.TriList[ uIndexIndex];
							Asura_Assertf( uVertexIndex < static_cast< u_int>( xPacket.NumberOfVertices), ("Vertex Index %d out of %d range. Mesh '%s'", uVertexIndex, xPacket.NumberOfVertices, pxMesh->m_pxMesh->Name.GetString()));
						}
					}
				}
				#endif

				Asura_Assert(pxMesh->TotalVertices <= iBefore, "Optimized mesh has more vertices than before?");
				iTotalVerticesBeforeOptimisation += iBefore;
				iTotalVerticesOptimisedAway += iBefore - pxMesh->TotalVertices;

				Tools_Level_Export::UpdateStatus(Tools_Level_Export::EXPORT_STAGE_STRIPIFYING_ENVIRONMENT, "Subdividing, optimizing, stripifying environment...", (u*3 + 2) * 100 / (s_uNumberOfMeshes * 3));
				
				if (bShouldStripify)
				{
					StripifySubdividedMesh(pxMesh, s_pxProcessedMesh[u]);
				}
				else
				{
					TriListSubdividedMesh( pxMesh, s_pxProcessedMesh[u]); 
				}
			}
			else
			{
				Tools_Level_Export::UpdateStatus(Tools_Level_Export::EXPORT_STAGE_STRIPIFYING_ENVIRONMENT, "Stripifying environment...", u * 100 / s_uNumberOfMeshes);
				pxMesh = new Tools_SubdividedMesh(u,true,true, bShouldSubdivide, uPolygonFlagsToIgnor);

				if (bShouldStripify)
				{
					StripifySubdividedMesh(pxMesh, s_pxProcessedMesh[u]);
				}
				else
				{
					TriListSubdividedMesh( pxMesh, s_pxProcessedMesh[u]); 
	#if defined(ASURA_TARGET_SM3)
					// JAT 25/02/07 only do this on next-gen builds for now (just for safety) because zero-sized polys make tangents & binormals become infinite
					// which makes lighting go mad
					TriListRemoveSmallTriangles( s_pxProcessedMesh[u], 1.0e-6f );
	#endif // #if defined( ASURA_TARGET_SM3 )
				}
			}

			s_uTotalNumberOfVertices += s_pxProcessedMesh[u].m_uTotalNumberOfVertices;
			s_uTotalNumberOfIndices += s_pxProcessedMesh[u].m_uTotalNumberOfIndices;
			s_uTotalNumberOfStrips += s_pxProcessedMesh[u].m_uNumberOfStrips;

			Asura_Assert( NULL != pxMesh && NULL != pxMesh->m_pxMesh, "No source Mesh Pointer");
			s_pxProcessedMesh[u].m_xMeshName = pxMesh->m_pxMesh->Name;

			delete pxMesh;
		}
	}

	Tools_Level_Export::UpdateStatus(Tools_Level_Export::EXPORT_STAGE_STRIPIFYING_ENVIRONMENT, 0, 100);

	if( bShouldRemoveDuplicateVertices )
	{
		float fPercentRemoved = (100.0f * iTotalVerticesOptimisedAway) / iTotalVerticesBeforeOptimisation;
		Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                        "    %li vertices (%.1f%%) removed by the optimization step (tris: %i degenerates, %i duplicates)",
                                        iTotalVerticesOptimisedAway, fPercentRemoved, s_uTotalDegenerateSourceTris, s_uTotalDuplicateSourceTris );
	}

    if( bShouldStripify )
    {
        // Output some statistics.
        {
            Tools_TriStripper::Stats xStats;
            Tools_TriStripper::GetStats( &xStats );

            float fPercentSingle = (100.0f * xStats.uNumRealSingleTriStrips) / xStats.uNumRealStrips;
            Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                            "    Generated %u real strips - avg %.2f indices; %.1f%% (%u) single tri strips.\n"
                                            "    Concatenated to %u strips - avg %.2f indices.",
                                            xStats.uNumRealStrips, xStats.fAvgRealStripIndices, fPercentSingle, xStats.uNumRealSingleTriStrips,
                                            xStats.uNumConcatdStrips, xStats.fAvgConcatdStripIndices );

            if( Tools_MeshRestartIndex::IsValidForCurrentPlatform() )
            {
                Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                                "    %.1f%% of indices were restarts in concatenated strips.",
                                                xStats.fIdxRedundancyProportion * 100.0f );
            }
            else
            {
                Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                                "    %.1f%% of indices were part of degenerates in concatenated strips\n"
                                                "    and %.1f%% of the tris were degenerates.",
                                                xStats.fIdxRedundancyProportion * 100.0f, xStats.fTriRedundancyProportion * 100.0f );
            }
        }

        Tools_TriStripper::Shutdown();
    }


	// KJL - 07/03/2003 12:20:55 - convert material ids
	if(!Tools_Level_Export::IsAbortFlagSet())
	{
		#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

			int iConversionListSize = ASURA_MAX_NO_OF_MATERIALS;

			int* piMaterialIDToIndexConversion = new int[iConversionListSize];

			//get list of meshes fo building conversion list
			List<Asura_Mesh*> xMeshList;
			int iNumMeshes = Asura_Module_System::GetNumberOfModules();
			for( int iMesh = 0; iMesh < iNumMeshes; ++iMesh )
			{
				xMeshList.add_entry(Asura_Module_System::GetMeshForModule( iMesh ));
			}

			Tools_SaveAsura::PrepareMeshMaterialConversionLists(xMeshList,piMaterialIDToIndexConversion,0);

		#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

		//apply conversion to the strips
		for (u_int u=0; u<s_uNumberOfMeshes; u++)
		{
			const Tools_EnvironmentProcessor_Mesh& xMesh = s_pxProcessedMesh[u];

			for(u_int s=0; s<xMesh.m_uNumberOfStrips; s++)
			{
				Tools_EnvironmentProcessor_Strip& xStrip = xMesh.m_pxStrips[s];

				#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
					xStrip.m_iMaterialID = Asura_Material_System::ConvertIDToIndex(xStrip.m_iMaterialID,piMaterialIDToIndexConversion);
				#endif

				xStrip.m_uModule = u;
			}
		}

		#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
			delete [] piMaterialIDToIndexConversion;
		#endif
    }


	//During level exports may  need to apply an offset to recentre all the verts
	//RAR 08/03/2006
	Asura_Vector_3 xPositionOffset = Tools_Level_Export::GetExportCentre();
	if(!Tools_Level_Export::IsAbortFlagSet())
	{
		if(xPositionOffset != Asura_Maths::ZeroVector3)
		{
			for (u_int u=0; u<s_uNumberOfMeshes; u++)
			{
				const Tools_EnvironmentProcessor_Mesh& xMesh = s_pxProcessedMesh[u];

				for(u_int s=0; s<xMesh.m_uNumberOfStrips; s++)
				{
					Tools_EnvironmentProcessor_Strip& xStrip = xMesh.m_pxStrips[s];

					for(u_int v=0; v<xStrip.m_uNumberOfVertices; v++)
					{
						xStrip.m_pxVertices[v].m_xPosition -= xPositionOffset;
					}
				}
			}
		}
	}

	s_bBeenInitialised = true;

	if(Tools_Level_Export::IsAbortFlagSet())
	{
		Shutdown();
	}
    else
    {
        const u_int uEndTime = Asura_Timers::GetMSecsSinceStart_Continuous();
        Tools_Level_Export::AddMessage( Tools_Level_Export::MT_LOGONLY,
                                        "    Processing took %.1fs in total.",
                                        static_cast<float>(uEndTime - uStartTime) / 1000.0f );
    }
}

void Tools_EnvironmentProcessor::Shutdown(void)
{
	Asura_Assert( s_bBeenInitialised, "Not yet Initialised!" );

	Asura_Assert( s_pxProcessedMesh, "Where have the meshes gone?" );

	delete [] s_pxProcessedMesh;
	s_pxProcessedMesh = 0;

    if( s_pxStripsSortedByMaterial )
    {
	    for ( u_int u=0; u<s_uTotalNumberOfStrips; ++u)
	    {
		    s_pxStripsSortedByMaterial[u].m_pxVertices = 0;
		    s_pxStripsSortedByMaterial[u].m_pusIndices = 0;
	    }

	    delete [] s_pxStripsSortedByMaterial;
	    s_pxStripsSortedByMaterial = 0;
    }

	if(s_bBeenGrouped)
	{
		Asura_Assert( s_ppxPosMesh, "Where have the meshes gone?" );

		for(u_int u = 0; u < s_uNumberOfMeshes; u++)
		{
			delete s_ppxPosMesh[u];
		}
		delete [] s_ppxPosMesh;
		s_ppxPosMesh = NULL;
	}
	s_bBeenGrouped = false;

	s_bBeenInitialised = false;

	//Invalidate surface meshes again , so they can be recalculated as needed for normal tools use.
	//RAR 31/08/2005
	Asura_Surface_Mesh_System::InvalidateAll();

}

const Tools_EnvironmentProcessor_Mesh* Tools_EnvironmentProcessor::GetProcessedMesh(u_int uModuleIndex)
{
	if (s_bBeenInitialised)
	{
		if (uModuleIndex<s_uNumberOfMeshes)
		{
			return &s_pxProcessedMesh[uModuleIndex];
		}
		else
		{
			Asura_Assert( false, "Index out of range" );
			return 0;
		}

	}
	else
	{
		Asura_Assert( s_bBeenInitialised, "Not yet Initialised!" );
		return 0;
	}
}

static int StripSortCompare(const void *p1, const void *p2)
{
	const Tools_EnvironmentProcessor_Strip* pxStrip1 = static_cast<const Tools_EnvironmentProcessor_Strip*>( p1 );
	const Tools_EnvironmentProcessor_Strip* pxStrip2 = static_cast<const Tools_EnvironmentProcessor_Strip*>( p2 );

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

	// Crude comparison for new materials for now
	for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
	{
		if(pxStrip1->m_auMaterialResponseHashID[iMatSlot] < pxStrip2->m_auMaterialResponseHashID[iMatSlot]) return -1;
		if(pxStrip1->m_auMaterialResponseHashID[iMatSlot] > pxStrip2->m_auMaterialResponseHashID[iMatSlot]) return 1;
	}
	return 0;

#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

	// Order of importance:
	//
	// 1. Shader type (this includes transparent polys - they go last)
	// 2. Texture (to group materials of the same texture together)
	// 3. Material ID
	//
	// Obviously (1) is actually dependent on the material properties.

	const int iMat1 = pxStrip1->m_iUnalteredMaterialID;
	const int iMat2 = pxStrip2->m_iUnalteredMaterialID;

	return Asura_Material_System::CompareMaterials( iMat1, iMat2 );

#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
}

const Tools_EnvironmentProcessor_Strip* Tools_EnvironmentProcessor::GetStripsSortedByMaterial( STRIP_CMP_FN pfnStripCmp )
{
	if (!s_bBeenInitialised)
	{
		Asura_Assert( s_bBeenInitialised, "Not yet Initialised!" );
		return 0;
	}

	if (!s_uTotalNumberOfStrips)
	{
		Asura_Assert( s_uTotalNumberOfStrips, "No strips present" );
		return 0;
	}

    // (MP - 29.06.2006) TODO: I've added this since previously we'd have
    // just leaked memory. However, it would probably be better to regard
    // this as lazily evaluated and just return the pointer if we've already
    // done this once.
    Asura_Assert( !s_pxStripsSortedByMaterial, "Already done this!" );

	s_pxStripsSortedByMaterial = new Tools_EnvironmentProcessor_Strip[ s_uTotalNumberOfStrips ];

	u_int uStripIndex = 0;

	for (u_int u=0; u<s_uNumberOfMeshes; u++)
	{
		const Tools_EnvironmentProcessor_Mesh& xMesh = s_pxProcessedMesh[u];

		for(u_int s=0; s<xMesh.m_uNumberOfStrips; s++)
		{
			Tools_EnvironmentProcessor_Strip& xStrip = xMesh.m_pxStrips[s];

			s_pxStripsSortedByMaterial[ uStripIndex++ ] = xStrip;
		}
	}

    // Default comparison function for when caller hasn't supplied one.
    if( !pfnStripCmp )
    {
        pfnStripCmp = StripSortCompare;
    }

    qsort( s_pxStripsSortedByMaterial, s_uTotalNumberOfStrips, sizeof(Tools_EnvironmentProcessor_Strip), pfnStripCmp );

	return s_pxStripsSortedByMaterial;
}

const Tools_EnvironmentProcessor_PosMesh* Tools_EnvironmentProcessor::GetPositionMesh(u_int uModuleIndex)
{
	if (s_bBeenInitialised)
	{
		if(s_bBeenGrouped)
		{
			if (uModuleIndex < s_uNumberOfMeshes)
			{
				return s_ppxPosMesh[uModuleIndex];
			}
			else
			{
				Asura_Assert( false, "Index out of range" );
				return 0;
			}
		}
		else
		{
			Asura_Assert( s_bBeenGrouped, "Not yet grouped!" );
			return 0;
		}
	}
	else
	{
		Asura_Assert( s_bBeenInitialised, "Not yet Initialised!" );
		return 0;
	}
}
class StripManager
{
	class OutputStripInfo {
		u_int m_uID;
		u_int m_uStripIndex;

		u_int m_uNumVerts;
		u_int m_uNumTris;
		u_int m_uMaxNumVerts;
		u_int m_uMaxNumTris;

		u_int * m_pVerts;
		u_short* m_pusTriList;

		template <typename T> static void Xchg(T &a, T &b)
		{
			T t = a;
			a = b;
			b = t;
		}
		void operator =(const OutputStripInfo&)
		{

		}
		OutputStripInfo(const OutputStripInfo&)
		{
		}

	public:
		OutputStripInfo(u_int uNumVerts = 0, u_int uNumTris = 0)
		{
			m_pVerts = (m_uMaxNumVerts = uNumVerts) > 0 ? new u_int[uNumVerts] : NULL;
			m_pusTriList = (m_uMaxNumTris = uNumTris) > 0 ? new u_short[uNumTris] : NULL;
			Asura_Assert(m_uMaxNumVerts == 0 || m_pVerts, "Out of mem");
			Asura_Assert(m_uMaxNumTris == 0 || m_pusTriList, "Out of mem");
		}
		~OutputStripInfo()
		{
			delete[] m_pVerts;
			delete[] m_pusTriList;
		}

		void Init(u_int uStripID, u_int uStripIndex)
		{
			m_uID = uStripID;
			m_uStripIndex = uStripIndex;

			m_uNumVerts = 0;
			m_uNumTris = 0;
		}
		void Reserve(u_int uNumVerts = 0, u_int uNumTris = 0)
		{
			if (m_uMaxNumVerts < uNumVerts)
			{
				delete [] m_pVerts;
				m_pVerts = new u_int[m_uMaxNumVerts = uNumVerts];
				Asura_Assert(m_pVerts, "Out of mem");
			}
			if (m_uMaxNumTris < uNumTris)
			{
				delete [] m_pusTriList;
				m_pusTriList = new u_short[m_uMaxNumTris = uNumTris];
				Asura_Assert(m_pusTriList, "Out of mem");
			}
		}
		u_int AddVert(int uIndex)
		{
			if (m_uMaxNumVerts <= m_uNumVerts)
			{
				u_int *pVerts = new u_int[m_uMaxNumVerts = (m_uNumVerts+1)*2];
				if (m_uNumVerts)
					memcpy(pVerts, m_pVerts, m_uNumVerts * sizeof(u_int));
				delete[] m_pVerts;
				m_pVerts = pVerts;
				Asura_Assert(m_pVerts, "Out of mem");
			} 
			Asura_Assert(m_pVerts, "");

			u_int uVertIndex = m_uNumVerts++;
			m_pVerts[uVertIndex] = uIndex;
			return uVertIndex;
		}

		void AddTri(u_int VertA, u_int VertB, u_int VertC)
		{
			if (m_uMaxNumTris <= m_uNumTris+2)
			{
				u_short *pusTris = new u_short[m_uMaxNumTris = (m_uNumTris+3)*2];
				if (m_uNumVerts)
					memcpy(pusTris, m_pusTriList, m_uNumTris * sizeof(u_short));
				delete[] m_pusTriList;
				m_pusTriList = pusTris;
				Asura_Assert(m_pusTriList, "Out of mem");
			}
			Asura_Assert(m_pusTriList, "");

			m_pusTriList[m_uNumTris + 0] = static_cast<u_short>(VertA); 
			m_pusTriList[m_uNumTris + 1] = static_cast<u_short>(VertB); 
			m_pusTriList[m_uNumTris + 2] = static_cast<u_short>(VertC); 
			m_uNumTris += 3;
		}
		u_int GetID() const {return m_uID;}
		u_int GetNumVerts() const {return m_uNumVerts;}
		u_int GetNumTris() const {return m_uNumTris/3;}
		u_int GetStripIndex() const {return m_uStripIndex;}
		const u_short *GetTriList() const {return m_pusTriList;}
		const u_int *GetVertList() const {return m_pVerts;}

		u_int GetVert(u_int uIndex) const { return m_pVerts[uIndex];}



		static void Swap(OutputStripInfo &xOne, OutputStripInfo &xTwo)
		{
			Xchg(xOne.m_uID, xTwo.m_uID);
			Xchg(xOne.m_uStripIndex, xTwo.m_uStripIndex);
			Xchg(xOne.m_uNumVerts, xTwo.m_uNumVerts);
			Xchg(xOne.m_uNumTris, xTwo.m_uNumTris);
			Xchg(xOne.m_pVerts, xTwo.m_pVerts);
			Xchg(xOne.m_pusTriList, xTwo.m_pusTriList);
			Xchg(xOne.m_uMaxNumTris, xTwo.m_uMaxNumTris);
			Xchg(xOne.m_uMaxNumVerts, xTwo.m_uMaxNumVerts);
		}
		static void Move(OutputStripInfo &xOne, OutputStripInfo &xTwo)
		{
			if(xOne.m_pusTriList) delete [] xOne.m_pusTriList;
			if(xOne.m_pVerts) delete [] xOne.m_pVerts;
			
			xOne.m_uID = xTwo.m_uID;
			xOne.m_uStripIndex = xTwo.m_uStripIndex;
			xOne.m_uNumVerts = xTwo.m_uNumVerts;
			xOne.m_uNumTris = xTwo.m_uNumTris;
			xOne.m_pVerts = xTwo.m_pVerts;
			xOne.m_pusTriList = xTwo.m_pusTriList;
			xOne.m_uMaxNumTris = xTwo.m_uMaxNumTris;
			xOne.m_uMaxNumVerts = xTwo.m_uMaxNumVerts;

			xTwo.m_pVerts = NULL;
			xTwo.m_pusTriList = NULL;
			xTwo.m_uMaxNumTris = 0;
			xTwo.m_uMaxNumVerts = 0;
		}


	};

	class StripInfoCollection
	{
		OutputStripInfo *m_pStripInfos;
		u_int m_uMaxNumStripInfos;
		u_int m_uNumStripInfos;

		u_int Sort(u_int uNewIndex, u_int uNewID)
		{
			for (u_int uStrip = 0; uStrip < m_uNumStripInfos-1; /*empty*/)
			{
				if (uStrip != uNewIndex)
				{
					u_int uID = m_pStripInfos[uStrip].GetID();
					if (uID != uStrip && uID < m_uNumStripInfos)
					{
						if (uID == uNewIndex)
						{
							// since we want to do a swap with new one which is uninited we can do a move not a swap
							OutputStripInfo::Move(m_pStripInfos[uID], m_pStripInfos[uNewIndex = uStrip]);
						} else {
							// need to do a full swap
							OutputStripInfo::Swap(m_pStripInfos[uID], m_pStripInfos[uStrip]);
						}
					} else {
						++uStrip;
					}
				} else {
					if (uNewID != uStrip && uNewID < m_uNumStripInfos)
					{
						// since we want to do a swap with new one which is uninited we can do a copy not a swap
						OutputStripInfo::Move(m_pStripInfos[uNewIndex = uStrip], m_pStripInfos[uNewID]);
					} else {
						++uStrip;
					}
				}
			}
			return uNewIndex;
		}

	public:
		StripInfoCollection()
		{
			m_pStripInfos = NULL;
			m_uNumStripInfos = 0;
			m_uMaxNumStripInfos = 0;
		}
		~StripInfoCollection()
		{
			delete[] m_pStripInfos;
		}

		void Reset()
		{
			m_uNumStripInfos = 0;
		}
		u_int AllocateStripInfo(u_int uStripID)
		{
			if (m_uMaxNumStripInfos <= m_uNumStripInfos)
			{
				m_uMaxNumStripInfos = (m_uNumStripInfos+1)*2;
				OutputStripInfo *pNewStripInfos = new OutputStripInfo[m_uMaxNumStripInfos];

				for (u_int i = 0; i < m_uNumStripInfos; ++i)
				{
					OutputStripInfo::Move(pNewStripInfos[i], m_pStripInfos[i]);
				}

				delete[] m_pStripInfos;
				m_pStripInfos = pNewStripInfos;
			}
			return Sort(m_uNumStripInfos++, uStripID) + 1;
		}

		OutputStripInfo *GetStripInfoFromIndex(u_int uStripIndex)
		{
			if (uStripIndex == 0)
				return NULL;
			--uStripIndex;
			return (uStripIndex < m_uNumStripInfos) ? (m_pStripInfos + uStripIndex) : NULL;
		}

		u_int LookupStripIndex(u_int uStripID)
		{
			if (uStripID < m_uNumStripInfos)
			{
				if (m_pStripInfos[uStripID].GetID() == uStripID)
					return uStripID+1;
			} else {
				for (u_int uStrip = 0; uStrip < m_uNumStripInfos; ++uStrip)
				{
					if (m_pStripInfos[uStrip].GetID() == uStripID)
					{
						return uStrip+1;
					}
				}

			}
			return 0;
		}


		u_int GetNumInfos() const {return m_uNumStripInfos;}
	};

	struct OutputVertInfo
	{
		u_int uStripInfoIndex;
		u_int uStripVertIndex;
	};

	class StripCollection
	{
		Tools_EnvironmentProcessor_Strip *m_pStrips;
		u_int m_uMaxNumStrips;
		u_int m_uNumStrips;
	public:
		StripCollection()
		{
			m_pStrips = NULL;
			m_uMaxNumStrips = 0;
			m_uNumStrips = 0;
		}
		~StripCollection()
		{
			delete[] m_pStrips;
		}
		Tools_EnvironmentProcessor_Strip *GetStrip(u_int uIndex)
		{
			if (uIndex < m_uNumStrips)
				return m_pStrips + uIndex;
			return NULL;
		}

		u_int AllocateStrip()
		{
			if (m_uMaxNumStrips <= m_uNumStrips)
			{
				Tools_EnvironmentProcessor_Strip *pNewStrips = new Tools_EnvironmentProcessor_Strip[m_uMaxNumStrips = (m_uNumStrips+1)*2];
				Asura_Assert( pNewStrips, "Out of memory" );

				for (u_int i = 0; i < m_uNumStrips; ++i)
				{
					Tools_EnvironmentProcessor_Strip::Move(pNewStrips[i], m_pStrips[i]);
				}
				delete[] m_pStrips;
				m_pStrips = pNewStrips;
			}
			return m_uNumStrips++;
		}

		void ClaimBuffer(Tools_EnvironmentProcessor_Strip *&pStrips, u_int &uNumStrips)
		{
			pStrips = m_pStrips;
			uNumStrips = m_uNumStrips;
			m_uMaxNumStrips = 0;
			m_uNumStrips = 0;
			m_pStrips = NULL;
		}


	};

	StripCollection m_xStrips;
	StripInfoCollection m_xStripInfos;


	u_int m_uNumVerts;
	u_int m_uNumIndicies;





public:
	StripManager()
	{
		m_uNumVerts = 0;
		m_uNumIndicies = 0;

	}
	~StripManager()
	{
	}

	void ProcessPacket(const Tools_SubdividedMesh &xMesh, const Tools_SubdividedMesh_Packet& xPacket, Tools_EnvironmentProcessor::PacketSelectionCallback pfnPacketSelectionCallback, Tools_EnvironmentProcessor::ColourAdjustmentCallback pfnColourAdjustmentCallback, Tools_EnvironmentProcessor::StripFlagCallback pfnStripFlagCallback, bool bUseRestartIndex)
	{
		m_xStripInfos.Reset();

		if (pfnPacketSelectionCallback != NULL)
		{
			OutputVertInfo* pVertInfos = new OutputVertInfo[xPacket.NumberOfVertices];


			for (int v = 0; v < xPacket.NumberOfVertices; ++v)
			{
				u_int uStripID = pfnPacketSelectionCallback(xMesh, xPacket, v);
				u_int uSelectedStrip = m_xStripInfos.LookupStripIndex(uStripID);

				if (uSelectedStrip == 0)
				{
					// create new strip
					uSelectedStrip = m_xStripInfos.AllocateStripInfo(uStripID);
					m_xStripInfos.GetStripInfoFromIndex(uSelectedStrip)->Init(uStripID, m_xStrips.AllocateStrip());
				}
				OutputStripInfo *pSelectedStrip = m_xStripInfos.GetStripInfoFromIndex(uSelectedStrip);

				pVertInfos[v].uStripInfoIndex = uStripID;
				pVertInfos[v].uStripVertIndex = pSelectedStrip->AddVert(v);
			}

			int TriVertIndex = 0;
			for (u_int uTriIndex=0; uTriIndex<(u_int)xPacket.NumberOfTris; uTriIndex++)
			{
				u_int uIndexA = xPacket.TriList[TriVertIndex++];
				u_int uIndexB = xPacket.TriList[TriVertIndex++];
				u_int uIndexC = xPacket.TriList[TriVertIndex++];

				u_int uTriStripID = pVertInfos[uIndexA].uStripInfoIndex;
				Asura_Assertf( uTriStripID == pVertInfos[uIndexB].uStripInfoIndex, ("Not all verticies of a triangle ended up in the same packet on mesh '%s'", xMesh.m_pxMesh->Name.GetString()));
				Asura_Assertf( uTriStripID == pVertInfos[uIndexC].uStripInfoIndex, ("Not all verticies of a same triangle ended up in the same packet on mesh '%s'", xMesh.m_pxMesh->Name.GetString()));

				m_xStripInfos.GetStripInfoFromIndex(m_xStripInfos.LookupStripIndex(uTriStripID))->AddTri(pVertInfos[uIndexA].uStripVertIndex, pVertInfos[uIndexB].uStripVertIndex, pVertInfos[uIndexC].uStripVertIndex);
			}
			delete [] pVertInfos;
			pVertInfos = NULL;
		} else {

			OutputStripInfo *pSelectedStrip = m_xStripInfos.GetStripInfoFromIndex(m_xStripInfos.AllocateStripInfo(0));

			pSelectedStrip->Init(0, m_xStrips.AllocateStrip());
			pSelectedStrip->Reserve(xPacket.NumberOfVertices, xPacket.NumberOfTris*3);

			for (int v = 0; v < xPacket.NumberOfVertices; ++v)
			{
				pSelectedStrip->AddVert(v);
			}
			for (u_int uTriIndex=0; uTriIndex<(u_int)xPacket.NumberOfTris; uTriIndex++)
			{
				pSelectedStrip->AddTri(xPacket.TriList[uTriIndex*3 + 0], xPacket.TriList[uTriIndex*3 + 1], xPacket.TriList[uTriIndex*3 + 2]);
			}
		}

		for (u_int uStrip = 0; uStrip < m_xStripInfos.GetNumInfos(); ++uStrip)
		{
			OutputStripInfo &xStripInfo = *m_xStripInfos.GetStripInfoFromIndex(uStrip+1);
			Tools_EnvironmentProcessor_Strip &xStrip = *m_xStrips.GetStrip(xStripInfo.GetStripIndex());

			u_int uNumberOfStrippedIndices = Tools_TriStripper::CreateStrips( xStripInfo.GetNumTris(), xStripInfo.GetTriList(), &xStrip.m_pusIndices );

			u_short* pusPermutationTable;
			Tools_TriStripper::ComputeVertexPermutation( uNumberOfStrippedIndices, xStrip.m_pusIndices, xStripInfo.GetNumVerts(), &pusPermutationTable );

			xStrip.m_pxVertices = new Tools_EnvironmentProcessor_Vertex[xStripInfo.GetNumVerts()];
			Asura_Assert(  xStrip.m_pxVertices, "Out of memory" );

			// reorder vertices based on new permutation
			for (u_int v=0; v < xStripInfo.GetNumVerts(); v++)
			{	
				Tools_SubdividedMesh_Vertex &xSrcVertex = xPacket.Vertices[xStripInfo.GetVert(pusPermutationTable[v])];
				Tools_EnvironmentProcessor_Vertex &xDestVertex = xStrip.m_pxVertices[v];

				if (v == 0) {
					xStrip.m_uPolygonFlags = xSrcVertex.m_iPolygonFlags & AsuraPolygonFlag_Darken;
				} else {
					Asura_Assert(xStrip.m_uPolygonFlags == (u_int)(xSrcVertex.m_iPolygonFlags & AsuraPolygonFlag_Darken), "");
				}

				xDestVertex.m_xPosition = xSrcVertex.Position;
				xDestVertex.m_xNormal = xSrcVertex.Normal;
				xDestVertex.m_fU = xSrcVertex.U;
				xDestVertex.m_fV = xSrcVertex.V;
				xDestVertex.m_fU2 = xSrcVertex.U2;
				xDestVertex.m_fV2 = xSrcVertex.V2;


				// KJL - 15/04/2003 09:12:55 - make sure alpha is set to full bright on each vertex
				// (AAM - 16.09.2004) Not any more - need alpha for dual-texturing

				// (SB - 06/03/07) Allow colour to be adjusted - for example to set certain
				// components to constants for blending operations gameside
				if (pfnColourAdjustmentCallback != NULL)
				{
					xDestVertex.m_uDiffuseRGBA = pfnColourAdjustmentCallback(xSrcVertex.Colour);
				} else {
					xDestVertex.m_uDiffuseRGBA = xSrcVertex.Colour;

				}

#if defined(ASURA_TARGET_SM3)
				xDestVertex.m_xLightTransportHarmonic = xSrcVertex.m_xLightTransportHarmonic;
				xDestVertex.m_xDiffuseSH = xSrcVertex.m_xDiffuseSH;

				xDestVertex.m_xBrightestLightVector = xSrcVertex.m_xBrightestLightVector;

				xDestVertex.m_xTangent = xSrcVertex.m_xTangent;
				xDestVertex.m_xBinormal = xSrcVertex.m_xBinormal;
				xDestVertex.m_xTangent2 = xSrcVertex.m_xTangent2;
				xDestVertex.m_xBinormal2 = xSrcVertex.m_xBinormal2;
#endif
			}

			if (pfnStripFlagCallback != NULL)
			{
				xStrip.m_uPolygonFlags = pfnStripFlagCallback(xMesh, xPacket, xStrip, xStripInfo.GetVertList(), xStripInfo.GetTriList(), xStripInfo.GetNumTris());
			} else {
				xStrip.m_uPolygonFlags = 0;
			}



			// finished with the stripifier's data
			delete [] pusPermutationTable;

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
			for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				xStrip.m_auMaterialResponseHashID[iMatSlot] = xPacket.m_auMaterialResponseHashID[iMatSlot];
			}
#else
			xStrip.m_iMaterialID = xPacket.m_iMaterialID;
#endif

			// material IDs get converted for export, so keep a copy around for tests later
			xStrip.m_iUnalteredMaterialID = xStrip.m_iMaterialID;

			xStrip.m_uNumberOfIndices = uNumberOfStrippedIndices;
			xStrip.m_uNumberOfTriangles = bUseRestartIndex ? xStripInfo.GetNumTris() : uNumberOfStrippedIndices - 2;
			xStrip.m_uNumberOfVertices = xStripInfo.GetNumVerts();

			m_uNumIndicies += uNumberOfStrippedIndices;
			m_uNumVerts += xStrip.m_uNumberOfVertices;
		}


	}

	void Populate(Tools_EnvironmentProcessor_Mesh &xMesh)
	{
		m_xStrips.ClaimBuffer(xMesh.m_pxStrips, xMesh.m_uNumberOfStrips);
		xMesh.m_uTotalNumberOfIndices = m_uNumIndicies;
		xMesh.m_uTotalNumberOfVertices	= m_uNumVerts;
	}
};


void Tools_EnvironmentProcessor::StripifySubdividedMesh(const Tools_SubdividedMesh* pxMesh, Tools_EnvironmentProcessor_Mesh& xProcessedMesh)
{
	if ( pxMesh->NumberOfPackets == 0 || pxMesh->TotalTris == 0 || pxMesh->TotalVertices == 0 )
	{
		xProcessedMesh.m_pxStrips = 0;
		xProcessedMesh.m_uNumberOfStrips = 0;
		xProcessedMesh.m_uTotalNumberOfIndices = 0;
		xProcessedMesh.m_uTotalNumberOfVertices = 0;
		return;
	}

	StripManager Strips;

	const bool bUseRestartIndex = Tools_MeshRestartIndex::IsValidForCurrentPlatform( );

	for (int i=0; i<pxMesh->NumberOfPackets; i++)
	{
		Strips.ProcessPacket(*pxMesh, pxMesh->Packets[i], s_pfnPacketSelectionCallback, s_pfnColourAdjustmentCallback, s_pfnStripFlagCallback, bUseRestartIndex);
	}

	Strips.Populate(xProcessedMesh);
}

// JAT 25/02/07 loop over the polygons removing degnerate triangles
void Tools_EnvironmentProcessor::TriListRemoveSmallTriangles( Tools_EnvironmentProcessor_Mesh& xProcessedMesh, const float fMinArea )
{
	const u_int uNumStrips = xProcessedMesh.m_uNumberOfStrips;
	if ( ( xProcessedMesh.m_pxStrips == 0 ) ||
		 ( uNumStrips == 0 ) ||
		 ( xProcessedMesh.m_uTotalNumberOfIndices == 0 ) ||
		 ( xProcessedMesh.m_uTotalNumberOfVertices == 0 ) )
	{
		return;
	}

	// Loop over the triangles removing them if the polygon is degenerate (very small area)
	for ( u_int uStrip = 0; uStrip < uNumStrips; uStrip++ )
	{
		Tools_DynamicArray<u_int> uTrisToRemove( 16 );
		Tools_EnvironmentProcessor_Strip& xStrip = xProcessedMesh.m_pxStrips[uStrip];
		const u_int uNumTriangles = xStrip.m_uNumberOfTriangles;
		u_int uIndex = 0;
		for ( u_int uTri = 0; uTri < uNumTriangles; uTri++ )
		{
			const u_short usIndex0 = xStrip.m_pusIndices[uIndex+0];
			const u_short usIndex1 = xStrip.m_pusIndices[uIndex+1];
			const u_short usIndex2 = xStrip.m_pusIndices[uIndex+2];

			// TODO: cope with restart indecies
			Tools_EnvironmentProcessor_Vertex& xVertex0	= xStrip.m_pxVertices[usIndex0];
			Tools_EnvironmentProcessor_Vertex& xVertex1	= xStrip.m_pxVertices[usIndex1];
			Tools_EnvironmentProcessor_Vertex& xVertex2	= xStrip.m_pxVertices[usIndex2];

            const float fArea = Asura_Maths::CalculateTriArea( xVertex0.m_xPosition, xVertex1.m_xPosition, xVertex2.m_xPosition );
			// <= to cope with 0.0f
			if ( fArea <= fMinArea )
			{
				// Ignore this triangle
				uTrisToRemove += uTri;
			}
			uIndex += 3;
		}
		const u_int uNumTrisToRemove = uTrisToRemove.GetSize();
		if ( uNumTrisToRemove > 0 )
		{
			// Now loop over removing the flagged small triangles from this strip
			const u_int uNumNewTris = uNumTriangles - uNumTrisToRemove;
			const u_int uNumNewIndices = uNumNewTris * 3;
			u_short* const pusNewIndices = new u_short[uNumNewIndices];
			u_short* const pusOldIndices =  xStrip.m_pusIndices;
			u_int uRemoveTriIndex = 0;
			u_int uTriToRemove = uTrisToRemove[uRemoveTriIndex];
			uRemoveTriIndex++;

			u_int uOldIndex = 0;
			u_int uNewIndex = 0;
			// This is a slow way of doing it but much easier to understand and debug and code quickly (sorry)
			for ( u_int uTri = 0; uTri < uNumTriangles; uTri++ )
			{
				if ( uTri != uTriToRemove )
				{
					// Just output it
					pusNewIndices[uNewIndex+0] = pusOldIndices[uOldIndex+0];
					pusNewIndices[uNewIndex+1] = pusOldIndices[uOldIndex+1];
					pusNewIndices[uNewIndex+2] = pusOldIndices[uOldIndex+2];
					uNewIndex += 3;
				}
				else
				{
					// Move to the next triangle to remove
					if ( uRemoveTriIndex < uNumTrisToRemove )
					{
						uTriToRemove = uTrisToRemove[uRemoveTriIndex];
						uRemoveTriIndex++;
					}
					else
					{
						// No more triangles to remove
						uTriToRemove = INT_MAX;
					}
				}
				// Advance to the next old triangle
				uOldIndex += 3;
			}
			// Need to attach the new indices to the strip and delete the old indices
			xStrip.m_pusIndices = pusNewIndices;
			delete[] pusOldIndices;

			// Also update the index and triangle count in the strip
			xStrip.m_uNumberOfTriangles = uNumNewTris;
			xStrip.m_uNumberOfIndices = uNumNewIndices;
			Asura_Assert( xStrip.m_uNumberOfIndices == uNewIndex, "" );
		}
	}
}

void Tools_EnvironmentProcessor::TriListSubdividedMesh( const Tools_SubdividedMesh* pxMesh, Tools_EnvironmentProcessor_Mesh& xProcessedMesh )
{
	if ( pxMesh->NumberOfPackets == 0 || pxMesh->TotalTris == 0 || pxMesh->TotalVertices == 0 )
	{
		xProcessedMesh.m_pxStrips = 0;
		xProcessedMesh.m_uNumberOfStrips = 0;
		xProcessedMesh.m_uTotalNumberOfIndices = 0;
		xProcessedMesh.m_uTotalNumberOfVertices = 0;
		return;
	}

	xProcessedMesh.m_uNumberOfStrips = pxMesh->NumberOfPackets;
	xProcessedMesh.m_pxStrips = new Tools_EnvironmentProcessor_Strip[pxMesh->NumberOfPackets];
	Asura_Assert( xProcessedMesh.m_pxStrips, "Out of memory" );
	
	xProcessedMesh.m_uTotalNumberOfIndices = 0;
	xProcessedMesh.m_uTotalNumberOfVertices = 0;

	for (int i=0; i<pxMesh->NumberOfPackets; i++)
	{
		Tools_SubdividedMesh_Packet& xPacket = pxMesh->Packets[i];
		Tools_EnvironmentProcessor_Strip& xStrip = xProcessedMesh.m_pxStrips[i];

		const u_int uNumberOfIndices = xPacket.NumberOfTris*3;

		xStrip.m_pxVertices = new Tools_EnvironmentProcessor_Vertex[xPacket.NumberOfVertices];
		xStrip.m_pusIndices = new u_short[uNumberOfIndices];

		Asura_Assert( xStrip.m_pxVertices, "Out of memory" );
		Asura_Assert( xStrip.m_pusIndices, "Out of memory" );

		for (int v=0; v<xPacket.NumberOfVertices; v++)
		{	
			xStrip.m_pxVertices[v].m_xPosition = xPacket.Vertices[v].Position;
			xStrip.m_pxVertices[v].m_xNormal = xPacket.Vertices[v].Normal;
			xStrip.m_pxVertices[v].m_fU = xPacket.Vertices[v].U;
			xStrip.m_pxVertices[v].m_fV = xPacket.Vertices[v].V;
			xStrip.m_pxVertices[v].m_fU2 = xPacket.Vertices[v].U2;
			xStrip.m_pxVertices[v].m_fV2 = xPacket.Vertices[v].V2;
			xStrip.m_pxVertices[v].m_uDiffuseRGBA = xPacket.Vertices[v].Colour;

			#if defined(ASURA_TARGET_SM3)
			xStrip.m_pxVertices[v].m_xLightTransportHarmonic = xPacket.Vertices[v].m_xLightTransportHarmonic;
			xStrip.m_pxVertices[v].m_xDiffuseSH = xPacket.Vertices[v].m_xDiffuseSH;

			// VGP 13Feb2006
			xStrip.m_pxVertices[v].m_xBrightestLightVector = xPacket.Vertices[v].m_xBrightestLightVector;

			xStrip.m_pxVertices[v].m_xTangent = xPacket.Vertices[v].m_xTangent;
			xStrip.m_pxVertices[v].m_xBinormal = xPacket.Vertices[v].m_xBinormal;
			#endif

		}

		for (u_int u=0; u<uNumberOfIndices; u++)
		{
			xStrip.m_pusIndices[u] = (unsigned short) xPacket.TriList[u];
		}

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
		for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
		{
			xStrip.m_auMaterialResponseHashID[iMatSlot] = xPacket.m_auMaterialResponseHashID[iMatSlot];
		}
#else
		xStrip.m_iMaterialID = xPacket.m_iMaterialID;
#endif

		// material IDs get converted for export, so keep a copy around for tests later
		xStrip.m_iUnalteredMaterialID = xStrip.m_iMaterialID;

		xStrip.m_uNumberOfIndices = uNumberOfIndices;
		xStrip.m_uNumberOfTriangles = xPacket.NumberOfTris;
		xStrip.m_uNumberOfVertices = xPacket.NumberOfVertices;

		xProcessedMesh.m_uTotalNumberOfIndices += uNumberOfIndices;
		xProcessedMesh.m_uTotalNumberOfVertices += xPacket.NumberOfVertices;
	}
}

void Tools_EnvironmentProcessor::GroupByPosition(void)
{
	if(s_bBeenInitialised)
	{
		s_uTotalNumberOfPositions = 0;
		// Allocate pointers to each mesh
		s_ppxPosMesh = new Tools_EnvironmentProcessor_PosMesh*[s_uNumberOfMeshes];
		for(u_int u = 0; u < s_uNumberOfMeshes; u++)
		{
			Tools_Level_Export::UpdateStatus(Tools_Level_Export::EXPORT_STAGE_GROUP_ENVIRONMENT_VERTS, "Grouping environment vertices by position", u * 100 / s_uNumberOfMeshes);
			// Create the position mesh from the processed mesh, which will do the grouping in the constructor
			s_ppxPosMesh[u] = new Tools_EnvironmentProcessor_PosMesh(&s_pxProcessedMesh[u]);
			s_uTotalNumberOfPositions += s_ppxPosMesh[u]->m_uNumberOfPositions;
		}
		s_bBeenGrouped = true;
	}
	else
	{
		Asura_Assert(s_bBeenInitialised, "Cannot group before initialisation.");
	}
}

Tools_EnvironmentProcessor_PosMesh::Tools_EnvironmentProcessor_PosMesh(Tools_EnvironmentProcessor_Mesh* pxMesh)
{
	Asura_Assert(pxMesh, "No original mesh to create grouped position mesh");
	m_uNumberOfPositions = 0;
	m_uNumberOfStrips = 0;
	if(pxMesh)
	{
		// Allocate as many positions as there are vertices, as this is the upper limit
		// (the actual number will be approximately a third of this, but best to play it safe)
		m_pxPositions = new Asura_Vector_3[pxMesh->m_uTotalNumberOfVertices];
		// Allocate pointers for each strip
		m_ppusIndices = new unsigned short*[pxMesh->m_uNumberOfStrips];
		// Allocate index map
		m_ppusPositionMap = new unsigned short*[pxMesh->m_uNumberOfStrips];
		// Iterate through strips, adding to mesh, which will add positions as necessary
		for(u_int u = 0; u < pxMesh->m_uNumberOfStrips; u++)
		{
			AddStrip(&pxMesh->m_pxStrips[u], u);
		}
		m_uNumberOfStrips = pxMesh->m_uNumberOfStrips;
	}
}

void Tools_EnvironmentProcessor_PosMesh::AddStrip(Tools_EnvironmentProcessor_Strip* pxStrip, unsigned int uStripIndex)
{
	Asura_Assert(pxStrip, "No strip to add to grouped position mesh.");
	if(pxStrip)
	{
		// Allocate new indices
		unsigned short* pusIndices = m_ppusIndices[uStripIndex] = new unsigned short[pxStrip->m_uNumberOfIndices];
		// Allocate mapping from old indices to new positions
		unsigned short* pusIndexMap = m_ppusPositionMap[uStripIndex] = new unsigned short[pxStrip->m_uNumberOfVertices];
		for(u_int u = 0; u < pxStrip->m_uNumberOfVertices; u++)
		{
			// Get new index
			pusIndexMap[u] = AddVertex(&pxStrip->m_pxVertices[u]);
		}
		for(u_int u = 0; u < pxStrip->m_uNumberOfIndices; u++)
		{
			// Remap to position index
			pusIndices[u] = pusIndexMap[pxStrip->m_pusIndices[u]];
		}
	}
	return;
}

unsigned short Tools_EnvironmentProcessor_PosMesh::AddVertex(Tools_EnvironmentProcessor_Vertex* pxVertex)
{
	// Rather naive/slow way of doing it at the moment - should really use a tree or sumfink.
	// We'll see how long it ends up taking
	// KOR - 23/05/2003 12:03:24 - Usually takes of the order of 10 seconds, which we can live with.
	unsigned short u;
	for(u = 0; u < m_uNumberOfPositions; u++)
	{
		// Found the same position. Return that one's index
		if(pxVertex->m_xPosition == m_pxPositions[u])
		{
			return u;
		}
	}
	// Didn't find it. Copy in.
	m_pxPositions[u] = pxVertex->m_xPosition;
	m_uNumberOfPositions++;
	return u;
}

//------------------------------------------------------------------------------
////////////////////////////////////////////////////////////////////////////////
//
// ProcessVerticesAndEliminateDuplicates function and supporting code
//
////////////////////////////////////////////////////////////////////////////////
//------------------------------------------------------------------------------


class Tools_EnvironmentProcessor_TriangleInfo
{

    public:

        // O P E R A T O R S /////////////////////////////////////////////////

        bool operator==( const Tools_EnvironmentProcessor_TriangleInfo& xOther ) const
        {
            if( m_uHashID != xOther.m_uHashID ) { return false; }

            for( u_int u = 0; u < 3; ++u )
            {
                if( m_auV[u] != xOther.m_auV[u] ) { return false; }
            }
            
            return true;
        }

        bool operator!=( const Tools_EnvironmentProcessor_TriangleInfo& xOther ) const
        {
            return !(*this == xOther);
        }

        // M E T H O D S /////////////////////////////////////////////////////

        Asura_Hash_ID GetHashID( void ) const { return m_uHashID; }

        u_int GetVtxIdx( u_int uIdxWithinTri ) const { return m_auV[uIdxWithinTri]; }

        void SetVtxIndices( u_int uV0, u_int uV1, u_int uV2 )
        {
            m_auV[0]  = uV0;
            m_auV[1]  = uV1;
            m_auV[2]  = uV2;
            m_uHashID = Asura_GetHashID_Binary( m_auV, 3 * sizeof(u_int) );
        }

        void ChangeVtxIdx( u_int uIdxWithinTri, u_int uNewVtxIdx )
        {
            m_auV[uIdxWithinTri] = uNewVtxIdx;
            m_uHashID            = Asura_GetHashID_Binary( m_auV, 3 * sizeof(u_int) );
        }

    private:
        
        // S T A T E /////////////////////////////////////////////////////////

        // Vertex indices.
        // Typically indices into processed data's vertex triplets.
        u_int m_auV[3];

        Asura_Hash_ID m_uHashID;

};

//Note that this is no longer a quadlet , but I can't be bothered to rename everything using it.
//RAR 04/05/2006
class Tools_EnvironmentProcessor_VertexIndexQuadlet
{

    public:

        // M E T H O D S /////////////////////////////////////////////////////

        u_int GetVertIdx( void ) const { return m_uVertIdx; }
        u_int GetPosIdx( void ) const { return m_uPosIdx; }

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY


		void SetIndices( u_int uPosIdx, u_int uVertIdx)
		{
			m_uPosIdx					= uPosIdx;
			m_uVertIdx					= uVertIdx;
		}

		Asura_Hash_ID GetMaterialResponseHashID(int iMatSlot) const { return m_auMaterialResponseHashID[iMatSlot]; }
		void SetMaterialResponseHashID(int iMatSlot, Asura_Hash_ID uID)	{ m_auMaterialResponseHashID[iMatSlot] = uID;	}

#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

        int GetMaterial( void ) const { return m_iMaterial; }

		void SetIndices( u_int uPosIdx, u_int uVertIdx, int iMaterial )
        {
            m_uPosIdx     = uPosIdx;
            m_uVertIdx    = uVertIdx;
			m_iMaterial   = iMaterial;
        }

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

    private:

        // S T A T E /////////////////////////////////////////////////////////////

        u_int m_uPosIdx;     // Index into list/tree of unique positions.
        u_int m_uVertIdx;    //          "            "        vertices.

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
		Asura_Hash_ID m_auMaterialResponseHashID[TOOLS_NUM_MATERIALRESPONSE_PER_POLY];
#else
        int m_iMaterial;     // Id of material used
#endif

};

void ASdump(bool bDo, char *pMsg, ...)
{
	if (bDo)
	{
		char buf[10240];
		va_list vl;
		va_start( vl, pMsg );     /* Initialize variable arguments. */
		vsprintf(buf, pMsg, vl);
		strcat(buf, "\n");
		OutputDebugString(buf);
	}
}

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
//Class for use in hash table for storing different material combinations
class Tools_EnvironmentProcessor_MaterialIDs
{
	public :
		Tools_EnvironmentProcessor_MaterialIDs(const Tools_EnvironmentProcessor_VertexIndexQuadlet* pxQuadlet)
		{
			for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				m_auMaterialResponseHashID[iMatSlot] = pxQuadlet->GetMaterialResponseHashID(iMatSlot);
				m_uNumTriangles = 0;
			}
		}

		Asura_Hash_ID GetHashID() const
		{
			Asura_Hash_ID uHash = m_auMaterialResponseHashID[0];
			for(int iMatSlot=1; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				uHash += m_auMaterialResponseHashID[iMatSlot];
			}
			return uHash;
		}

		bool operator == (const Tools_EnvironmentProcessor_MaterialIDs& xOther) const
		{
			for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				if(m_auMaterialResponseHashID[iMatSlot] != xOther.m_auMaterialResponseHashID[iMatSlot]) return false;
			}
			return true;
		}

	
		Asura_Hash_ID m_auMaterialResponseHashID[TOOLS_NUM_MATERIALRESPONSE_PER_POLY];
		u_int m_uNumTriangles;
	
		
};
#endif //defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY


// Name stolen shamelessly from the PS2 exporter, as well as much of the code
void Tools_EnvironmentProcessor::ProcessVerticesAndEliminateDuplicates(Tools_SubdividedMesh& xMesh)
{
	if(!xMesh.TotalTris || !xMesh.TotalVertices || !xMesh.NumberOfPackets) return;

	int iPacket = 0;
	Tools_SubdividedMesh_Packet* pPacket = &xMesh.Packets[0];
	Asura_Assert(pPacket, "No packets");
	const u_int uNumSrcTris  = xMesh.TotalTris + 1;
	const u_int uNumSrcVerts = 3* uNumSrcTris;

    // Just allocate worst cases for arrays. Not really going to be that vast
    // and we're on PC anyway.

    Tools_SubdividedMesh_Vertex*					axMeshVertices						= new Tools_SubdividedMesh_Vertex[uNumSrcVerts];
    u_int											uNumMeshVertices					= 0;
	Tools_EnvironmentProcessor_VertexIndexQuadlet*	axUniqueMeshVtxIdxQuadlets			= new Tools_EnvironmentProcessor_VertexIndexQuadlet[uNumSrcVerts];
	u_int											uNumUniqueMeshVtxIdxQuadlets		= 0;
	u_int											uNumUnusedUniqueMeshVtxIdxQuadlets	= 0;
	Tools_EnvironmentProcessor_TriangleInfo*		axUniqueMeshTris					= new Tools_EnvironmentProcessor_TriangleInfo[uNumSrcTris];
	u_int											uNumUniqueMeshTris					= 0;
	u_int*											auDeletedUniqueMeshVtxIdxQuadlets	= new u_int[(uNumSrcVerts + 31) / 32];

	memset(auDeletedUniqueMeshVtxIdxQuadlets, 0, ((uNumSrcVerts + 31) / 32) * sizeof(u_int));

	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
//		const u_int									uUniqueTrisArraySize				= 1 + Tools_MaterialResponse_System::GetNumMaterials();
	#else
		const u_int									uUniqueTrisArraySize				= 1 + ASURA_MAX_NO_OF_MATERIALS;
	#endif

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	Tools_HashTable<Tools_EnvironmentProcessor_MaterialIDs> xHashTableUniqueMeshTrisPerMaterial(256);
	List<Tools_EnvironmentProcessor_MaterialIDs*> xListUniqueMeshTrisPerMaterial;
#else
	u_int*											auUniqueMeshTrisPerMaterial			= new u_int[ uUniqueTrisArraySize ];
#endif
	Asura_Vector_3*									axMeshPositions						= new Asura_Vector_3[uNumSrcVerts];
    u_int											uNumMeshPositions					= 0;

	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
//		Asura_Mem::Set( auUniqueMeshTrisPerMaterial, 0, uUniqueTrisArraySize * sizeof(u_int) );
	#else
		Asura_Mem::Set( auUniqueMeshTrisPerMaterial, 0, uUniqueTrisArraySize * sizeof(u_int) );
	#endif

	const u_int uMAX_POTENTIAL_DUP_VERTS = 64; // 32 should be higher than any vaguely sensible mesh should ever produce. But SS2 has a bit  
											   // of valid geometry that exceeds that, to I've upped it to 64....

    u_int auPotentialDupVerts[uMAX_POTENTIAL_DUP_VERTS];

	// Tolerance values
	// These are used to decide if the properties of two vertices are close enough that they can be merged
	// Position tolerance is the squared maximum distance between positions
	const float fPosTol = 0.0025f * 0.0025f; // +/- 0.25cm.

	bool bLog = false;//xMesh.m_pxMesh->Name.Compare("polySurface734") == 0;
	int iPoly = -1;
    u_int auCurrentPolyVertexQuadletIndices[4];
	while(true)
	{
		u_int uNumVertsInInputPoly;

		// next poly
		iPoly++;
		uNumVertsInInputPoly = 3;
		if(iPoly >= pPacket->NumberOfTris)
		{
			// go to the next packet
			iPoly = 0;
			iPacket++;
			if(iPacket >= xMesh.NumberOfPackets) break;
			pPacket = &xMesh.Packets[iPacket];
			Asura_Assert(pPacket, "Run out of packets");
			Asura_Assert(pPacket->NumberOfTris, "No triangles in packet");
		}
		ASdump(bLog, "Poly %d-%d", iPoly, iPacket);

		for( u_int uVertexWithinInputPoly = 0; uVertexWithinInputPoly < uNumVertsInInputPoly; ++uVertexWithinInputPoly )
		{
			Tools_SubdividedMesh_Vertex* pxSubdividedVertex = &pPacket->Vertices[pPacket->TriList[iPoly*3 + uVertexWithinInputPoly]];
			ASdump(bLog, "Vertex %d", uVertexWithinInputPoly);

			//////////////////////////////////
			// P O S I T I O N
			//////////////////

			u_int uPosIndex;
			const Asura_Vector_3* pxPos = &pxSubdividedVertex->Position;

			// search for previously matching position
			uPosIndex = UINT_MAX;
			for( u_int u = 0; u < uNumMeshPositions; ++u)
			{
				if(pxPos->ApproximatelyEquals(axMeshPositions[u], fPosTol))
				{
					uPosIndex = u;
					break;
				}
			}

			// Unique?
			if( UINT_MAX == uPosIndex )
			{
				uPosIndex = uNumMeshPositions++;
				Asura_Assert(uPosIndex < uNumSrcVerts, "More positions than we started with?!");
				axMeshPositions[uPosIndex] = *pxPos;
				ASdump(bLog, "position unique %d", uPosIndex);
			} else {
				ASdump(bLog, "position shared %d", uPosIndex);

			}

			// Create list of potential duplicated (based on position index only)
			u_int uNumPotentialDupVerts = 0;
			for( u_int u = 0; u < uNumUniqueMeshVtxIdxQuadlets; ++u)
			{
				if((auDeletedUniqueMeshVtxIdxQuadlets[u/32] & (1 << (u & 31))) == 0 && axUniqueMeshVtxIdxQuadlets[u].GetPosIdx() == uPosIndex)
				{
					ASdump(bLog, "potential dup with %d", u);
					auPotentialDupVerts[uNumPotentialDupVerts++] = u;
					if( uNumPotentialDupVerts == uMAX_POTENTIAL_DUP_VERTS )
					{
						Asura_Assertf( false, ("Mesh %s: Too many vertices are duplicates of each other - uMAX_POTENTIAL_DUP_VERTS needs increasing, tolerance needs lowering, or your mesh needs fixing",
											   xMesh.m_pxMesh ? xMesh.m_pxMesh->Name.GetString() : "<Unknown>" ) );
						break;
					}
				}
			}
			//////////////////////////////////
			//Compare the rest of the vertex data for the potential duplicates

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

				for( u_int u = 0; u < uNumPotentialDupVerts; /* No iterative term */ )
				{
					bool bMaterialMatch = true;
					for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
					{
						const Asura_Hash_ID uOtherMaterial = axUniqueMeshVtxIdxQuadlets[ auPotentialDupVerts[u] ].GetMaterialResponseHashID(iMatSlot);
						if(uOtherMaterial != pPacket->m_auMaterialResponseHashID[iMatSlot])
						{
							bMaterialMatch = false;
							break;
						}
					}
					
					const u_int uOtherVertIndex = axUniqueMeshVtxIdxQuadlets[ auPotentialDupVerts[u] ].GetVertIdx();

					bool bDualMaterial = pPacket->m_auMaterialResponseHashID[1] != 0;


					if( bMaterialMatch && Tools_SubdividedMesh_Vertex::ApproximatelyEquals( pxSubdividedVertex, &axMeshVertices[ uOtherVertIndex ], bDualMaterial, false ) )
					{
						// Still a viable duplicate
						++u;
					}
					else
					{
						// Not viable - remove it from the list
						auPotentialDupVerts[u] = auPotentialDupVerts[ --uNumPotentialDupVerts ];
					}
				}

			#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

				const int iThisMaterial = pPacket->m_iMaterialID;

				for(u_int u = 0; u < uNumPotentialDupVerts; /* no iterative term */ )
				{
					const int iOtherMaterial = axUniqueMeshVtxIdxQuadlets[auPotentialDupVerts[u]].GetMaterial();
					const u_int uOtherVertIndex = axUniqueMeshVtxIdxQuadlets[auPotentialDupVerts[u]].GetVertIdx();

					if(iThisMaterial == iOtherMaterial && Tools_SubdividedMesh_Vertex::ApproximatelyEquals(pxSubdividedVertex,&axMeshVertices[uOtherVertIndex],iThisMaterial,false))
					{
						ASdump(bLog, "still a potential dup with %d", auPotentialDupVerts[u]);
						// still a viable duplicate
						++u;
					}
					else
					{
						ASdump(bLog, "not a potential dup with %d", auPotentialDupVerts[u]);
						// not viable - remove it from the list
						auPotentialDupVerts[u] = auPotentialDupVerts[--uNumPotentialDupVerts];
					}
				}

			#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

			////////////////////////////////////////////////////////
			// Q U A D L E T   P R O C E S S I N G
			//////////////////////////////////////

			u_int uQuadletIndex;
			bool bIsNewQuadlet;

			// Check that we've actually got duplicate vertices
			if( uNumPotentialDupVerts )
			{
				ASdump(bLog, "got dup %d", uNumPotentialDupVerts);
				// It's possible to get to this point with more than 1
                // potential duplicate. This is because we might have a
                // sequence of vertices {A, B, C}  such that A ~ C and B ~ C
                // but !(A ~ B) (where ~ denotes approximate equality.) In
                // this case we'll reach C with both A and B as potential
                // duplicates.
				if(1 == uNumPotentialDupVerts)
				{
					// Just point the duplicate at the unique vertex instead.
					uQuadletIndex = auPotentialDupVerts[0];
					bIsNewQuadlet = false;
				}
				else
				{
					// We're actually going to create a new quadlet
					uQuadletIndex = uNumUniqueMeshVtxIdxQuadlets++;
					bIsNewQuadlet = true;
					ASdump(bLog, "uQuadletIndex = %d uNumUniqueMeshVtxIdxQuadlets = %d", uQuadletIndex, uNumUniqueMeshVtxIdxQuadlets);

					// we can't remove existing quadlets because that would screw over our indices
					// we'll just count how many are unused instead
					uNumUnusedUniqueMeshVtxIdxQuadlets += uNumPotentialDupVerts;
					ASdump(bLog, "uNumUnusedUniqueMeshVtxIdxQuadlets = %d", uNumUnusedUniqueMeshVtxIdxQuadlets);
					Asura_Assert( uNumUnusedUniqueMeshVtxIdxQuadlets < uNumUniqueMeshVtxIdxQuadlets, "More unused quadlets than there are quadlets in total?");

					// Fix up triangles to use the new quadlet
					for(u_int uTri = 0; uTri < uNumUniqueMeshTris; ++uTri )
					{
						Tools_EnvironmentProcessor_TriangleInfo& xTri = axUniqueMeshTris[uTri];
						for(u_int uIdxWithinTri = 0; uIdxWithinTri < 3; ++uIdxWithinTri )
						{
							for(u_int uDup = 0; uDup < uNumPotentialDupVerts; ++uDup)
							{
								if(xTri.GetVtxIdx(uIdxWithinTri) == auPotentialDupVerts[uDup] )
								{
									xTri.ChangeVtxIdx( uIdxWithinTri, uQuadletIndex );
									break;
								}
							}
						}
					}

					// Fix up any earlier vertices within this current triangle
					for( u_int uEarlierVert = 0; uEarlierVert < uVertexWithinInputPoly; ++uEarlierVert )
					{
						for( u_int uDup = 0; uDup < uNumPotentialDupVerts; ++uDup )
						{
							if( auCurrentPolyVertexQuadletIndices[uEarlierVert] == auPotentialDupVerts[uDup])
							{
								auCurrentPolyVertexQuadletIndices[uEarlierVert] = uQuadletIndex;
								break;
							}
						}
					}

					// mark these as gone!
					for( u_int uDup = 0; uDup < uNumPotentialDupVerts; ++uDup )
					{
						u_int uIndex = auPotentialDupVerts[uDup];
						
						auDeletedUniqueMeshVtxIdxQuadlets[uIndex/32] |= (1 << (uIndex & 31));
					}
				}

			}
			else
			{
				// No duplicate vertices. Just create a new quadlet for this vertex
				uQuadletIndex = uNumUniqueMeshVtxIdxQuadlets++;
				ASdump(bLog, "uQuadletIndex = %d uNumUniqueMeshVtxIdxQuadlets = %d", uQuadletIndex, uNumUniqueMeshVtxIdxQuadlets);
				bIsNewQuadlet = true;
			}

			// If the quadlet is new, set it up
			if( bIsNewQuadlet )
			{
				u_int uVertIndex = uNumMeshVertices++;
				ASdump(bLog, "bIsNewQuadlet uNumMeshVertices = %d (%d)", uNumMeshVertices, uNumSrcVerts);
				Asura_Assert(uVertIndex < uNumSrcVerts, "More vertices than we started with?!");
				axMeshVertices[uVertIndex] = *pxSubdividedVertex;
				ASdump(bLog, "axMeshVertices[%d].index = %d", uVertIndex, pxSubdividedVertex->m_iIndexInPacket);
				ASdump(bLog, "axMeshVertices[%d].Pos = {%f,%f,%f}", uVertIndex, pxSubdividedVertex->Position.GetX(), pxSubdividedVertex->Position.GetY(), pxSubdividedVertex->Position.GetZ());
				ASdump(bLog, "axMeshVertices[%d].Nor = {%f,%f,%f}", uVertIndex, pxSubdividedVertex->Normal.GetX(), pxSubdividedVertex->Normal.GetY(), pxSubdividedVertex->Normal.GetZ());
				ASdump(bLog, "axMeshVertices[%d].argb = %x", uVertIndex, pxSubdividedVertex->Colour);
				ASdump(bLog, "axMeshVertices[%d].uv = {%f,%f}", uVertIndex, pxSubdividedVertex->U, pxSubdividedVertex->V);
				ASdump(bLog, "axMeshVertices[%d].flags = %x", uVertIndex, pxSubdividedVertex->m_iPolygonFlags);

				#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
					axUniqueMeshVtxIdxQuadlets[uQuadletIndex].SetIndices( uPosIndex, uVertIndex );
					for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
					{
						axUniqueMeshVtxIdxQuadlets[uQuadletIndex].SetMaterialResponseHashID(iMatSlot, pPacket->m_auMaterialResponseHashID[iMatSlot]);
					}
				#else
					axUniqueMeshVtxIdxQuadlets[uQuadletIndex].SetIndices( uPosIndex, uVertIndex, pPacket->m_iMaterialID );
				#endif
			}

			auCurrentPolyVertexQuadletIndices[uVertexWithinInputPoly] = uQuadletIndex;
			ASdump(bLog, "auCurrentPolyVertexQuadletIndices[%d] = %d", uVertexWithinInputPoly, uQuadletIndex);
		} // end loop through vertices

		ASdump(bLog, "axUniqueMeshTris[%d].SetVtxIndicies(%d,%d,%d)", uNumUniqueMeshTris, auCurrentPolyVertexQuadletIndices[0],auCurrentPolyVertexQuadletIndices[1],auCurrentPolyVertexQuadletIndices[2]);
		// Create triangle
		axUniqueMeshTris[uNumUniqueMeshTris++].SetVtxIndices( auCurrentPolyVertexQuadletIndices[0],
															  auCurrentPolyVertexQuadletIndices[1],
															  auCurrentPolyVertexQuadletIndices[2] );
	} // End loop through triangles

	// Now that we've finished removing duplicate vertices, 
	// process our triangle list and remove degenerates/duplicates/etc.

	// Use a hash table for finding duplicates quickly
	const u_int uTriTableSize = ( uNumUniqueMeshTris > 1 ? Asura_Maths::Max(1, 1 << (Asura_Maths::GetCeilLog2(uNumUniqueMeshTris) - 1)) : 1 );
	Tools_HashTable<Tools_EnvironmentProcessor_TriangleInfo> xTriHashTable( uTriTableSize );

	const u_int uNumTrisToCheck = uNumUniqueMeshTris;
	uNumUniqueMeshTris = 0; // reset our counter

	for( u_int uSrcTri = 0; uSrcTri < uNumTrisToCheck; ++uSrcTri )
	{
		const Tools_EnvironmentProcessor_TriangleInfo& xNext = axUniqueMeshTris[uSrcTri];

		bool bDupOrDegen = false;

		// Try adding this triangle to the hash table. If what we get back isn't what we put in, this is duplicate.
		const Tools_EnvironmentProcessor_TriangleInfo* pxInsertedTri = xTriHashTable.Insert( &xNext );
		if( pxInsertedTri != &xNext )
		{
			++s_uTotalDuplicateSourceTris;
			bDupOrDegen = true;
		}

		// Now test for degeneracy
		const u_int uPosIndices[3] = {	axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(0)].GetPosIdx(),
										axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(1)].GetPosIdx(),
										axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(2)].GetPosIdx() };
		if(uPosIndices[0] == uPosIndices[1] || uPosIndices[0] == uPosIndices[2] || uPosIndices[1] == uPosIndices[2] )
		{
			++s_uTotalDegenerateSourceTris;
			bDupOrDegen = true;
		}

		if( !bDupOrDegen )
		{
			// Copy this triangle back in the array (over a degen/dupl entry)
			if( uNumUniqueMeshTris != uSrcTri )
			{
				axUniqueMeshTris[uNumUniqueMeshTris] = xNext;
			}

			++uNumUniqueMeshTris;

			const Tools_EnvironmentProcessor_VertexIndexQuadlet& xFirstQuadlet = axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(0)];

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

				//Is there an existing entry in the hash table for this material combo?
				const Tools_EnvironmentProcessor_MaterialIDs* pxExisting = xHashTableUniqueMeshTrisPerMaterial.Find(Tools_EnvironmentProcessor_MaterialIDs(&xFirstQuadlet));
				if(pxExisting)
				{
					//bleh , stupid hash table has const pointers.... tough!
					//RAR 14/06/2007
					const_cast<Tools_EnvironmentProcessor_MaterialIDs*>(pxExisting)->m_uNumTriangles++;
				}
				else
				{
					//create a new entry then
					Tools_EnvironmentProcessor_MaterialIDs* pxMaterialCombo = new Tools_EnvironmentProcessor_MaterialIDs(&xFirstQuadlet);
					xHashTableUniqueMeshTrisPerMaterial.Insert(pxMaterialCombo);
					xListUniqueMeshTrisPerMaterial.add_entry(pxMaterialCombo);
					pxMaterialCombo->m_uNumTriangles++;
					
				}

			#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

				int iMaterialID = xFirstQuadlet.GetMaterial();
				++auUniqueMeshTrisPerMaterial[Asura_Material_System::TranslateMaterialIDToArrayIdx(iMaterialID) + 1];

			#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
		}
	}

	// Count number of unique materials in this mesh.
    u_int uNumUniqueMeshMaterials = 0;

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	{
		u_int uMatCount = xListUniqueMeshTrisPerMaterial.size();
		for( u_int uMatIndex = 0; uMatIndex < uMatCount; uMatIndex += 1)
		{
			if( xListUniqueMeshTrisPerMaterial[uMatIndex]->m_uNumTriangles ) 
			{ 
				++uNumUniqueMeshMaterials; 
			}
		}
	}
#else
	for( u_int u = 0; u < uUniqueTrisArraySize; ++u )
    {
        if( auUniqueMeshTrisPerMaterial[u] ) 
        { 
			++uNumUniqueMeshMaterials; 
		}
    }
#endif

	// Done working. Now we need to rebuild the packets from our input mesh, so nobody knows we were here
	delete[] xMesh.Packets;
	xMesh.Packets = new Tools_SubdividedMesh_Packet[uNumUniqueMeshMaterials];
	xMesh.NumberOfPackets = uNumUniqueMeshMaterials;
	xMesh.TotalTris = 0;
	xMesh.TotalVertices = 0;

	// Build each packet
	iPacket = 0;
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	u_int uMatCount = xListUniqueMeshTrisPerMaterial.size();
	for( u_int uMatIndex = 0; uMatIndex < uMatCount; uMatIndex += 1)
	{
		const Tools_EnvironmentProcessor_MaterialIDs* pxMatCombo = xListUniqueMeshTrisPerMaterial[uMatIndex];
		
		const u_int uUniqueMeshTrisPerMaterial = pxMatCombo->m_uNumTriangles;
		
		if( 0 == uUniqueMeshTrisPerMaterial) 
		{
			continue;
		}
		
		Tools_SubdividedMesh_Packet& xPacket = xMesh.Packets[iPacket++];
		
		for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
		{
			xPacket.m_auMaterialResponseHashID[iMatSlot] = pxMatCombo->m_auMaterialResponseHashID[iMatSlot];
		}
		
		xPacket.NumberOfTris = uUniqueMeshTrisPerMaterial;
		
#else
	for(u_int u = 0; u < uUniqueTrisArraySize; ++u)
	{
		if(!auUniqueMeshTrisPerMaterial[u]) continue;
		
		Tools_SubdividedMesh_Packet& xPacket = xMesh.Packets[iPacket++];
		
		xPacket.m_iMaterialID = Asura_Material_System::TranslateArrayIdxToMaterialID(u - 1);
		
		xPacket.NumberOfTris = auUniqueMeshTrisPerMaterial[u];
#endif

		xPacket.TriList = new int[xPacket.NumberOfTris * 3];

		List<int> xUsedVertexIndices;
		for(u_int uTri = 0; uTri < uNumUniqueMeshTris; ++uTri)
		{
			// check that this triangle matches our current material
			const Tools_EnvironmentProcessor_TriangleInfo& xTri = axUniqueMeshTris[uTri];
			const Tools_EnvironmentProcessor_VertexIndexQuadlet& xFirstQuadlet = axUniqueMeshVtxIdxQuadlets[xTri.GetVtxIdx(0)];

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
				bool bMaterialMatch = true;
				for(int iMatSlot=0; iMatSlot < TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
				{
					if( xFirstQuadlet.GetMaterialResponseHashID(iMatSlot) != xPacket.m_auMaterialResponseHashID[iMatSlot] )
					{
						bMaterialMatch = false;
						break;
					}
				}
				if(!bMaterialMatch) continue;

			#else
				if(xFirstQuadlet.GetMaterial() != xPacket.m_iMaterialID)
				{
					continue;
				}
			#endif

			// add the vertices from this triangle to the list if they are not already there
			if(!xUsedVertexIndices.contains(xTri.GetVtxIdx(0)))
			{
				xUsedVertexIndices.add_entry(xTri.GetVtxIdx(0));
			}
			if(!xUsedVertexIndices.contains(xTri.GetVtxIdx(1)))
			{
				xUsedVertexIndices.add_entry(xTri.GetVtxIdx(1));
			}
			if(!xUsedVertexIndices.contains(xTri.GetVtxIdx(2)))
			{
				xUsedVertexIndices.add_entry(xTri.GetVtxIdx(2));
			}
		}

		xPacket.NumberOfVertices = xUsedVertexIndices.size();
		xPacket.Vertices = new Tools_SubdividedMesh_Vertex[xPacket.NumberOfVertices];

		// Work out which indices will contain which quadlets
		int* piVertConv = new int[xPacket.NumberOfVertices];
		int iNextIndex = 0;
		for(CLIF<int> xIt(&xUsedVertexIndices); !xIt.done(); xIt.next())
		{
			piVertConv[iNextIndex++] = xIt();
		}

		// Fill out the vertex data
		for(int iVertex = 0; iVertex < xPacket.NumberOfVertices; ++iVertex)
		{
			const Tools_EnvironmentProcessor_VertexIndexQuadlet& xQuadlet = axUniqueMeshVtxIdxQuadlets[piVertConv[iVertex]];
			Tools_SubdividedMesh_Vertex* pxOut = &xPacket.Vertices[iVertex];

			*pxOut = axMeshVertices[xQuadlet.GetVertIdx()];
		}

		// Build the triangle list, remapping vertex indices as we go
		int iNextTriangle = 0;
		for(u_int uTri = 0; uTri < uNumUniqueMeshTris; ++uTri)
		{
			const Tools_EnvironmentProcessor_TriangleInfo& xTri = axUniqueMeshTris[uTri];
			const Tools_EnvironmentProcessor_VertexIndexQuadlet& xFirstQuadlet = axUniqueMeshVtxIdxQuadlets[xTri.GetVtxIdx(0)];

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
				bool bMaterialMatch = true;
				for(int iMatSlot=0; iMatSlot < TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
				{
					if( xFirstQuadlet.GetMaterialResponseHashID(iMatSlot) != xPacket.m_auMaterialResponseHashID[iMatSlot] )
					{
						bMaterialMatch = false;
						break;
					}				
				}
				if(!bMaterialMatch) continue;
			#else
				if(xFirstQuadlet.GetMaterial() != xPacket.m_iMaterialID) continue;
			#endif

			int iQuadletIndices[3] = { xTri.GetVtxIdx(0), xTri.GetVtxIdx(1), xTri.GetVtxIdx(2) };
			for(int i = 0; i < 3; ++i)
			{
				int newIndex;
				for(newIndex = 0; newIndex < xPacket.NumberOfVertices; ++newIndex)
				{
					if(piVertConv[newIndex] == iQuadletIndices[i]) break;
				}
				Asura_Assert(newIndex != xPacket.NumberOfVertices, "Should never happen"); // we added the vertices from this triangle to the list earlier, but they don't seem to be there now?
				Asura_Assert(newIndex < xPacket.NumberOfVertices, ("Index %d is out of %d range. Mesh '%s'", newIndex, xPacket.NumberOfVertices, xMesh.m_pxMesh->Name.GetString())); 
				const u_int uTriangleIndex = iNextTriangle * 3 + i;
				Asura_Assert( uTriangleIndex < static_cast< u_int>(3 * xPacket.NumberOfTris), ("Triangle index %d is out of %d range. Mesh '%s'", uTriangleIndex, 3 * xPacket.NumberOfTris, xMesh.m_pxMesh->Name.GetString())); // we added the vertices from this triangle to the list earlier, but they don't seem to be there now?
				xPacket.TriList[ uTriangleIndex] = newIndex;
			}
			++iNextTriangle;
			
			Asura_Assertf( iNextTriangle <= xPacket.NumberOfTris, ("Not all triangles have been setup correctly. Mesh '%s'", xMesh.m_pxMesh->Name.GetString()));
		}

		Asura_Assertf( iNextTriangle == xPacket.NumberOfTris, ("Not all triangles have been setup correctly. Mesh '%s'", xMesh.m_pxMesh->Name.GetString()));
		
		delete[] piVertConv;

		xMesh.TotalTris += xPacket.NumberOfTris;
		xMesh.TotalVertices += xPacket.NumberOfVertices;
	}

	// Finished! Clean up the crapload of arrays we created :-)

	delete[] axMeshPositions;
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	while(xListUniqueMeshTrisPerMaterial.size())
	{
		delete xListUniqueMeshTrisPerMaterial.first_entry();
		xListUniqueMeshTrisPerMaterial.delete_first_entry();
	}
#else
	delete[] auUniqueMeshTrisPerMaterial;
#endif
	delete[] axUniqueMeshTris;
	delete[] axUniqueMeshVtxIdxQuadlets;
	delete[] axMeshVertices;
	delete[] auDeletedUniqueMeshVtxIdxQuadlets;

	#if defined( ASURA_ASSERT)
	{
		u_int uPacketCount = xMesh.NumberOfPackets;
		for( u_int uPacketIndex = 0; uPacketIndex < uPacketCount; uPacketIndex += 1)
		{
			const Tools_SubdividedMesh_Packet& xPacket = xMesh.Packets[ uPacketIndex];
			const u_int uIndexCount = 3 * xPacket.NumberOfTris;
			
			for( u_int uIndexIndex = 0; uIndexIndex < uIndexCount; uIndexIndex += 1)
			{
				const u_int uVertexIndex = xPacket.TriList[ uIndexIndex];
				Asura_Assertf( uVertexIndex < static_cast< u_int>( xPacket.NumberOfVertices), ("Vertex Index %d out of %d range. Mesh '%s'", uVertexIndex, xPacket.NumberOfVertices, xMesh.m_pxMesh->Name.GetString()));
			}
		}
	}
	#endif

}





#if defined(ASURA_TARGET_SM3)

////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_EnvironmentProcessor::CalculateTangentSpaceVectors
// Author       : Andy Tate
// Date         : 13-May-2005
//
// Description  : 
//
// Parameters   : References to the 3 vectors to use for the triangle calculation.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Tools_EnvironmentProcessor::CalculateTangentSpaceVectors (const Tools_EnvironmentProcessor_Vertex& rxVertex0,
															   const Tools_EnvironmentProcessor_Vertex& rxVertex1,
															   const Tools_EnvironmentProcessor_Vertex& rxVertex2,
																u_int uTextureIndex,
															   Asura_Vector_3& xTangent,
															   Asura_Vector_3& xBinormal)
{
	Asura_Vector_3	axPositions [3];
	Asura_Vector_2	axTextureCoords [3];


	// Initialise the vertex-data...

	axPositions [0]			= rxVertex0.m_xPosition;
	axPositions [1]			= rxVertex1.m_xPosition;
	axPositions [2]			= rxVertex2.m_xPosition;

	switch (uTextureIndex)
	{
	case 0:
		axTextureCoords [0].x	= rxVertex0.m_fU;
		axTextureCoords [0].y	= 1.000f - rxVertex0.m_fV;

		axTextureCoords [1].x	= rxVertex1.m_fU;
		axTextureCoords [1].y	= 1.000f - rxVertex1.m_fV;

		axTextureCoords [2].x	= rxVertex2.m_fU;
		axTextureCoords [2].y	= 1.000f - rxVertex2.m_fV;
		break;
	case 1:
		axTextureCoords [0].x	= rxVertex0.m_fU2;
		axTextureCoords [0].y	= 1.000f - rxVertex0.m_fV2;

		axTextureCoords [1].x	= rxVertex1.m_fU2;
		axTextureCoords [1].y	= 1.000f - rxVertex1.m_fV2;

		axTextureCoords [2].x	= rxVertex2.m_fU2;
		axTextureCoords [2].y	= 1.000f - rxVertex2.m_fV2;
		break;
	default:
		Asura_Assertf(0, ("Unknown texture index %d\n this is a programmer error so slap whomever added more than 2 UV sets and didnt update this assert", uTextureIndex));
	}


	// Calculate the tangent AND binormal...

	Tools_TangentSpace::CalculateTangentAndBinormal	(axPositions, axTextureCoords, xTangent, xBinormal);
}

/// End of function 'Tools_EnvironmentProcessor::CalculateTangentSpaceVectors'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_EnvironmentProcessor::CalculateTriListTangentSpaceVectors
// Author       : Andy Tate
// Date         : 13-May-2005
//
// Description  : Runs through the specfified environment mesh and calculates
//				  tangent space vectors for ALL of the triangles within it.
//
// Parameters   : A reference to the environment mesh to use.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Tools_EnvironmentProcessor::CalculateTriListTangentSpaceVectors (Tools_EnvironmentProcessor_Mesh& rxProcessedMesh)
{
	u_short	usRestartIndex;

	bool	bHasRestartIndex	= Tools_MeshRestartIndex::IsValidForCurrentPlatform (usRestartIndex);

	// Step through each strip in-turn...
	for (u_int s = 0; s < rxProcessedMesh.m_uNumberOfStrips; s++)
	{
		Tools_EnvironmentProcessor_Strip&	rxStrip	= rxProcessedMesh.m_pxStrips [s];
		int									iIndex	= 0;

		// Go through each triangle within this strip...
		for (u_int t = 0; t < rxStrip.m_uNumberOfTriangles; t++)
		{
			const u_short	usIndex0	= rxStrip.m_pusIndices [iIndex++];
			const u_short	usIndex1	= rxStrip.m_pusIndices [iIndex++];
			const u_short	usIndex2	= rxStrip.m_pusIndices [iIndex++];

			// Are any of our indices the restart one?
			if ( !bHasRestartIndex ||
				 ((usIndex0 != usRestartIndex) && (usIndex1 != usRestartIndex) && (usIndex2 != usRestartIndex)) )
			{
				Tools_EnvironmentProcessor_Vertex&	rxVertex0	= rxStrip.m_pxVertices [usIndex0];
				Tools_EnvironmentProcessor_Vertex&	rxVertex1	= rxStrip.m_pxVertices [usIndex1];
				Tools_EnvironmentProcessor_Vertex&	rxVertex2	= rxStrip.m_pxVertices [usIndex2];

				Asura_Vector_3						xBinormal;
				Asura_Vector_3						xTangent;
				Asura_Vector_3						xBinormal2;
				Asura_Vector_3						xTangent2;

				// Calculate the tangent-space vectors for this triangle...
				CalculateTangentSpaceVectors	(rxVertex0, rxVertex1, rxVertex2, 0, xTangent, xBinormal);
				CalculateTangentSpaceVectors	(rxVertex0, rxVertex1, rxVertex2, 1, xTangent2, xBinormal2);

				// Finally, update the vertices themselves...
				rxVertex0.m_xBinormal	= xBinormal;
				rxVertex0.m_xTangent	= xTangent;
				rxVertex0.m_xBinormal2	= xBinormal2;
				rxVertex0.m_xTangent2	= xTangent2;

				rxVertex1.m_xBinormal	= xBinormal;
				rxVertex1.m_xTangent	= xTangent;
				rxVertex1.m_xBinormal2	= xBinormal2;
				rxVertex1.m_xTangent2	= xTangent2;

				rxVertex2.m_xBinormal	= xBinormal;
				rxVertex2.m_xTangent	= xTangent;
				rxVertex2.m_xBinormal2	= xBinormal2;
				rxVertex2.m_xTangent2	= xTangent2;
			}
		}

		// Go through each vertex and fix the tangent and binormal...
		for (u_int v = 0; v < rxStrip.m_uNumberOfVertices; v++)
		{
			Tools_EnvironmentProcessor_Vertex&	rxVertex	= rxStrip.m_pxVertices [v];

			const Asura_Vector_3 xOldTangent = rxVertex.m_xTangent;
			const Asura_Vector_3 xOldBinormal = rxVertex.m_xBinormal;

			// Clamp the tangent and binormal to the normal...
			rxVertex.m_xTangent		= Tools_TangentSpace::ClampPerpendicularToNormal (rxVertex.m_xNormal, rxVertex.m_xTangent, rxProcessedMesh.m_xMeshName.GetString());
			rxVertex.m_xBinormal	= Tools_TangentSpace::ClampPerpendicularToNormal (rxVertex.m_xNormal, rxVertex.m_xBinormal, rxProcessedMesh.m_xMeshName.GetString());
			
			if ( rxVertex.m_xTangent.MagnitudeSquared() <= 1.0e-3f )
			{
				rxVertex.m_xTangent		= Tools_TangentSpace::ClampPerpendicularToNormal( rxVertex.m_xNormal, xOldTangent, rxProcessedMesh.m_xMeshName.GetString() );
			}
			if ( rxVertex.m_xBinormal.MagnitudeSquared() <= 1.0e-3f )
			{
				rxVertex.m_xBinormal	= Tools_TangentSpace::ClampPerpendicularToNormal( rxVertex.m_xNormal, xOldBinormal, rxProcessedMesh.m_xMeshName.GetString() );
			}
		}
	}
}

/// End of function 'Tools_EnvironmentProcessor::CalculateTriListTangentSpaceVectors'. ///



////////////////////////////////////////////////////////////////////////////////
// Function     : Tools_EnvironmentProcessor::CalculateTriStripTangentSpaceVectors
// Author       : Andy Tate
// Date         : 20-Sep-2005
//
// Description  : Runs through the specfified environment mesh and calculates
//				  tangent space vectors for ALL of the triangle-strips within it.
//
// Parameters   : A reference to the environment mesh to use.
//
// Return value : None.
////////////////////////////////////////////////////////////////////////////////

void Tools_EnvironmentProcessor::CalculateTriStripTangentSpaceVectors (Tools_EnvironmentProcessor_Mesh& rxProcessedMesh)
{
	u_short	usRestartIndex;

	bool	bHasRestartIndex	= Tools_MeshRestartIndex::IsValidForCurrentPlatform (usRestartIndex);

#if defined( ASURA_ASSERT)
	u_int uNumDegenerateTriangles = 0;
	u_int uNumShortTangents = 0;
#endif

	// Step through each strip in-turn...
	for (u_int s = 0; s < rxProcessedMesh.m_uNumberOfStrips; s++)
	{
		Tools_EnvironmentProcessor_Strip&	xStrip	= rxProcessedMesh.m_pxStrips [s];

		// Go through each triangle within this strip...
		for (u_int i = 0; i < xStrip.m_uNumberOfIndices - 2; i++)
		{
			const u_short	usIndex0	= xStrip.m_pusIndices [i + 0];
			const u_short	usIndex1	= xStrip.m_pusIndices [i + 1];
			const u_short	usIndex2	= xStrip.m_pusIndices [i + 2];

			// (AAM - 11.05.2006) Skip degenerates, otherwise we'll calculate garbage data
			if( usIndex0 == usIndex1 || usIndex0 == usIndex2 || usIndex1 == usIndex2 ) continue;

			// Are any of our indices the restart one?
			if ( !bHasRestartIndex ||
				 ((usIndex0 != usRestartIndex) && (usIndex1 != usRestartIndex) && (usIndex2 != usRestartIndex)) )
			{
				Tools_EnvironmentProcessor_Vertex&	xVertex0	= xStrip.m_pxVertices [usIndex0];
				Tools_EnvironmentProcessor_Vertex&	xVertex1	= xStrip.m_pxVertices [usIndex1];
				Tools_EnvironmentProcessor_Vertex&	xVertex2	= xStrip.m_pxVertices [usIndex2];

				Asura_Vector_3						xBinormal;
				Asura_Vector_3						xTangent;
				Asura_Vector_3						xBinormal2;
				Asura_Vector_3						xTangent2;

				// Calculate the tangent-space vectors for this triangle...
				CalculateTangentSpaceVectors	(xVertex0, xVertex1, xVertex2, 0, xTangent, xBinormal);
				CalculateTangentSpaceVectors	(xVertex0, xVertex1, xVertex2, 1, xTangent2, xBinormal2);

				// Clamp the tangent and binormal to the normal...
				xVertex0.m_xTangent		= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex0.m_xNormal, xTangent, rxProcessedMesh.m_xMeshName.GetString());
				xVertex0.m_xBinormal	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex0.m_xNormal, xBinormal, rxProcessedMesh.m_xMeshName.GetString());
				xVertex0.m_xTangent2	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex0.m_xNormal, xTangent2, rxProcessedMesh.m_xMeshName.GetString());
				xVertex0.m_xBinormal2	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex0.m_xNormal, xBinormal2, rxProcessedMesh.m_xMeshName.GetString());

				xVertex1.m_xTangent		= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex1.m_xNormal, xTangent, rxProcessedMesh.m_xMeshName.GetString());
				xVertex1.m_xBinormal	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex1.m_xNormal, xBinormal, rxProcessedMesh.m_xMeshName.GetString());
				xVertex1.m_xTangent2	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex1.m_xNormal, xTangent2, rxProcessedMesh.m_xMeshName.GetString());
				xVertex1.m_xBinormal2	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex1.m_xNormal, xBinormal2, rxProcessedMesh.m_xMeshName.GetString());

				xVertex2.m_xTangent		= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex2.m_xNormal, xTangent, rxProcessedMesh.m_xMeshName.GetString());
				xVertex2.m_xBinormal	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex2.m_xNormal, xBinormal, rxProcessedMesh.m_xMeshName.GetString());
				xVertex2.m_xTangent2	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex2.m_xNormal, xTangent2, rxProcessedMesh.m_xMeshName.GetString());
				xVertex2.m_xBinormal2	= Tools_TangentSpace::ClampPerpendicularToNormal (xVertex2.m_xNormal, xBinormal2, rxProcessedMesh.m_xMeshName.GetString());

#if defined( ASURA_ASSERT)
				if( xVertex0.m_xPosition == xVertex1.m_xPosition || xVertex0.m_xPosition == xVertex2.m_xPosition || xVertex1.m_xPosition == xVertex2.m_xPosition)
				{
					uNumDegenerateTriangles++;
				}
				if ( ( xVertex0.m_xTangent.MagnitudeSquared() < 1.0e-3f ) || ( xVertex0.m_xBinormal.MagnitudeSquared() < 1.0e-3f ) || 
					 ( xVertex1.m_xTangent.MagnitudeSquared() < 1.0e-3f ) || ( xVertex1.m_xBinormal.MagnitudeSquared() < 1.0e-3f ) || 
					 ( xVertex2.m_xTangent.MagnitudeSquared() < 1.0e-3f ) || ( xVertex2.m_xBinormal.MagnitudeSquared() < 1.0e-3f ) )
				{
					uNumShortTangents++;
					
					// DLH - 12 February 2007 16:39:02 - StuartM Requested this goes in so stop bad values getting into the 360 export
					Asura_Vector_3 xTangent = Asura_Vector_3( 1.0f, 2.0f, 3.0f );
					xTangent.Normalise();
					
					xVertex0.m_xTangent = xTangent ^ xVertex0.m_xNormal;
					xVertex1.m_xTangent = xTangent ^ xVertex1.m_xNormal;
					xVertex2.m_xTangent = xTangent ^ xVertex2.m_xNormal;
					
					xVertex0.m_xBinormal = xVertex0.m_xTangent ^ xVertex0.m_xNormal;
					xVertex1.m_xBinormal = xVertex1.m_xTangent ^ xVertex1.m_xNormal;
					xVertex2.m_xBinormal = xVertex2.m_xTangent ^ xVertex2.m_xNormal;
				}
				if ( ( xVertex0.m_xTangent2.MagnitudeSquared() < 1.0e-3f ) || ( xVertex0.m_xBinormal.MagnitudeSquared() < 1.0e-3f ) || 
					( xVertex1.m_xTangent2.MagnitudeSquared() < 1.0e-3f ) || ( xVertex1.m_xBinormal.MagnitudeSquared() < 1.0e-3f ) || 
					( xVertex2.m_xTangent.MagnitudeSquared() < 1.0e-3f ) || ( xVertex2.m_xBinormal.MagnitudeSquared() < 1.0e-3f ) )
				{
					uNumShortTangents++;

					// DLH - 12 February 2007 16:39:02 - StuartM Requested this goes in so stop bad values getting into the 360 export
					Asura_Vector_3 xTangent = Asura_Vector_3( 1.0f, 2.0f, 3.0f );
					xTangent.Normalise();

					xVertex0.m_xTangent = xTangent ^ xVertex0.m_xNormal;
					xVertex1.m_xTangent = xTangent ^ xVertex1.m_xNormal;
					xVertex2.m_xTangent = xTangent ^ xVertex2.m_xNormal;

					xVertex0.m_xBinormal = xVertex0.m_xTangent ^ xVertex0.m_xNormal;
					xVertex1.m_xBinormal = xVertex1.m_xTangent ^ xVertex1.m_xNormal;
					xVertex2.m_xBinormal = xVertex2.m_xTangent ^ xVertex2.m_xNormal;
				}

#endif // #if defined( ASURA_ASSERT)
			}
		}
	}
	
	Asura_Assertf( 0 == uNumDegenerateTriangles, ( "Mesh '%s' contains %d degenerate triangles. This will need fixing", rxProcessedMesh.m_xMeshName.GetString(), uNumDegenerateTriangles ) );
#if defined( ASURA_ASSERT)
	if ( uNumShortTangents > 0 )
	{
		Tools_Level_Export::AddMessage( Tools_Level_Export::MT_WARNING, "Mesh '%s' contains %d triangles with bad texture tangent data.", rxProcessedMesh.m_xMeshName.GetString(), uNumShortTangents );
	}
#endif // #if defined( ASURA_ASSERT)
}

/// End of function 'Tools_EnvironmentProcessor::CalculateTriStripTangentSpaceVectors'. ///

#endif	// defined(ASURA_TARGET_SM3)


/// End of file. ///
