//
// Tools_SmoothSkin_Processor_PC.cpp
//
// Created by RJF - 17 June 2005
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "MeshProcessing/Tools_SmoothSkin_Processor_PC.h"

// Asura independent headers
#include "Animation/Asura_Hierarchy.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Memory/Asura_Mem.h"
#include "Mesh/Asura_Mesh.h"

// Asura PC headers
#include "ChunkSystem/Asura_Chunk_ResourceFile_PC.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile_PC_Character.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile_PC_Skin.h"
#if !defined(ASURA_PC_DISABLE_PRESM3_SMOOTH_SKIN_RENDERER) || !ASURA_PC_DISABLE_PRESM3_SMOOTH_SKIN_RENDERER
	#include "Rendering/Asura_PlatformRenderCharacter.h"
#endif

#if defined(ASURA_PC_USE_SHADER_MODEL_3_SMOOTH_SKIN_RENDERER) && ASURA_PC_USE_SHADER_MODEL_3_SMOOTH_SKIN_RENDERER
	#include "ChunkSystem/Asura_Chunk_ResourceFile_PC_SmoothSkin.h"
	#include "Rendering/Asura_Direct3D_SmoothSkinRenderer.h"
#endif

// Tools headers
#include "MeshProcessing/Asura_TriStripper.h"
#include "MeshProcessing/Tools_HierarchyProcessor_MultiMaterial_NormalMapped.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor.h"
#include "Misc/Tools_DynamicArray.h"
#include "PC/Tools_PC_Export.h"


////////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_PC_SubElement
//
// Holds stripified data for a sub-element
//
////////////////////////////////////////////////////////////////////////////////

class Tools_SmoothSkin_PC_SubElement
{
	friend class Tools_SmoothSkin_PC_Processor;

public:

	Tools_SmoothSkin_PC_SubElement(): m_uNumStripIndices(0), m_pusStripIndices(NULL) {}
	~Tools_SmoothSkin_PC_SubElement() { delete[] m_pusStripIndices; }

protected:

	u_int		m_uNumStripIndices;
	u_short*	m_pusStripIndices;
	u_int		m_uFirstIndexPosition;
};

////////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_PC_Material
//
// Holds stripified data for a material
//
////////////////////////////////////////////////////////////////////////////////

class Tools_SmoothSkin_PC_Material
{
	friend class Tools_SmoothSkin_PC_Processor;

public:

	Tools_SmoothSkin_PC_Material(): m_pxSubElement(NULL) {}
	~Tools_SmoothSkin_PC_Material() { delete[] m_pxSubElement; }

protected:

	// Array of sub-elements. There are as many entries as there are elements in the main skin
	Tools_SmoothSkin_PC_SubElement* m_pxSubElement;
};


#if defined(ASURA_TARGET_SM3)
////////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_PC_Processor::ProcessedStrip
//
// Holds information on a strip as it's prepared by the skin processor
//
////////////////////////////////////////////////////////////////////////////////

class Tools_SmoothSkin_PC_Processor::ProcessedStrip
{
public:

	// Constructor / destructor
	ProcessedStrip();
	~ProcessedStrip();

	// Functions
	u_int GetMaxWeights() const;

	// Data

	Tools_SmoothSkin_Vertex* m_pxVertices;
	u_int    m_uNumVertices;
	u_short* m_pusIndices;
	u_int    m_uNumIndices;
	Asura_Hash_ID m_auMaterialResponseHashID[2];
	int      m_iElementID;

	u_int	m_uBoneMapID;
};


////////////////////////////////////////////////////////////////////////////////
// Tools_SmoothSkin_PC_Processor::ProcessedStrip Class Functions		  //
////////////////////////////////////////////////////////////////////////////////

Tools_SmoothSkin_PC_Processor::ProcessedStrip::ProcessedStrip()
: m_pxVertices( NULL )
, m_uNumVertices(0)
, m_pusIndices( NULL )
, m_uNumIndices(0)
, m_iElementID(0)
, m_uBoneMapID(0)
{
}

Tools_SmoothSkin_PC_Processor::ProcessedStrip::~ProcessedStrip()
{
	delete[] m_pxVertices;
	delete[] m_pusIndices;
}

u_int Tools_SmoothSkin_PC_Processor::ProcessedStrip::GetMaxWeights() const
{
	// Find the maximum number of weightings used by verts in this strip

	u_int uMax = 0;

	for( u_int u = 0; u < m_uNumVertices; ++u )
	{
		uMax = Asura_Maths::Max( uMax, m_pxVertices[u].GetNumWeightings() );
	}

	return uMax;
}
#endif // #if defined(ASURA_TARGET_SM3)

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


Tools_SmoothSkin_PC_Processor::Tools_SmoothSkin_PC_Processor(const Tools_SmoothSkin_Processor& xProcessor) 
: m_xSrcProcessor(xProcessor), m_pxMaterial(NULL)
#if defined(ASURA_TARGET_SM3)
	, m_szSrcName( NULL )
	, m_pxProcessedStrips( NULL )
	, m_uNumProcessedStrips( 0 )
	, m_uTotalVertices( 0 )
	, m_uTotalIndices( 0 )
	, m_uTotalTriangles( 0 )
#endif 
{

}

Tools_SmoothSkin_PC_Processor::~Tools_SmoothSkin_PC_Processor()
{
	delete[] m_pxMaterial;
#if defined(ASURA_TARGET_SM3)
	delete[] m_pxProcessedStrips;
#endif
}

bool Tools_SmoothSkin_PC_Processor::Initialise(bool bStripify /*=true*/)
{
	const u_int uNumMaterials = m_xSrcProcessor.GetNumMaterials();

#if defined(ASURA_TARGET_SM3)
	m_szSrcName = m_xSrcProcessor.GetInputSkin()->GetName();
#endif 

	// The 'elements' stuff isn't very well commented, but I asked Kev and he's suggested that it's probably for hacking pieces off of skins
	// at runtime (removing limbs etc). As such we want to keep elements seperate instead of stripifying them all together. Could be a bit 
	// wasteful if the skin isn't actually used in that manner but we'll trust the artists to unify their meshes where appropriate.
	const u_int uNumElements = m_xSrcProcessor.GetNumElements();

	// Allocate storage for the stripified materials

	m_pxMaterial = new Tools_SmoothSkin_PC_Material[uNumMaterials];
	if(!m_pxMaterial) return false;

	// Loop through each material in the skin

	for (u_int u = 0; u < uNumMaterials; ++u)
	{
		const Tools_SmoothSkin_Material& xMaterial(m_xSrcProcessor.GetMaterial(u));

		// Allocate storage for the stripified sub-elements
		m_pxMaterial[u].m_pxSubElement = new Tools_SmoothSkin_PC_SubElement[uNumElements];
		if(!m_pxMaterial[u].m_pxSubElement) return false;

		// Loop through each sub-element
		for (u_int v = 0; v < uNumElements; ++v)
		{
			const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement(v);

			if(xSubElement.GetNumTris() == 0) continue;

            if(bStripify)
			{
            	Tools_TriStripper::Initialise();

				u_short* pusStripIndices;
            	u_int uNumStripIndices = Tools_TriStripper::CreateStrips( xSubElement.GetNumTris(), xSubElement.GetIndices(), &pusStripIndices );

            	Tools_TriStripper::Shutdown();

				// Store the results
				Tools_SmoothSkin_PC_SubElement& xDestSubElement = m_pxMaterial[u].m_pxSubElement[v];

				xDestSubElement.m_uNumStripIndices = uNumStripIndices;
				xDestSubElement.m_pusStripIndices = pusStripIndices;
			}
			else
			{
				//not stripifying , so just copy across the triangle indices. RAR 05/01/2007
				Tools_SmoothSkin_PC_SubElement& xDestSubElement = m_pxMaterial[u].m_pxSubElement[v];

				xDestSubElement.m_uNumStripIndices = xSubElement.GetNumTris()*3;
				xDestSubElement.m_pusStripIndices = new unsigned short[xDestSubElement.m_uNumStripIndices];

				memcpy(xDestSubElement.m_pusStripIndices,xSubElement.GetIndices(),sizeof(unsigned short)*xDestSubElement.m_uNumStripIndices);
			}
		}
	}

	m_bTriLists = !bStripify;

	return true;
}

bool Tools_SmoothSkin_PC_Processor::OutputToChunk(Asura_Chunk_Stream& xStream, bool bConvertMaterialIDs) const
{
	//#define WRITEOUTDATA(a,b) WriteFile( hFile, a, b, &bytesWritten, 0 );
	#define WRITEOUTDATA(pAddress, uSize) xStream.Write(uSize, pAddress);

	const bool bIsNormalMapped = m_xSrcProcessor.ShouldIncludeNormalMapData();

	// count the indices
	u_int uNumNonEmptyElements = 0;
	u_int uTotalIndices = 0;
	for(u_int uMaterial = 0; uMaterial < m_xSrcProcessor.GetNumMaterials(); ++uMaterial)
	{
		for(u_int uElement = 0; uElement < m_xSrcProcessor.GetNumElements(); ++uElement)
		{
			if(m_xSrcProcessor.GetMaterial(uMaterial).GetSubElement(uElement).GetNumTris() > 0)
			{
				++uNumNonEmptyElements;
				uTotalIndices += m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices;
			}
		}
	}

	// VGP 20Mar2006 Count render bones and build bone map
	int* auBoneMap = new int[m_xSrcProcessor.GetInputSkin()->NumberOfBones];
	Tools_DynamicArray<u_char> auRenderBoneIndices;
	ZeroMemory(auBoneMap, m_xSrcProcessor.GetInputSkin()->NumberOfBones * sizeof(int));

	// RJF 27Jun2006 Always include bone 0. It will never actually be used, but it's required
	// for some constant math to work out. Check out mantis #0027148 for info.
	auRenderBoneIndices += 0;

	// Count bones with verts attached to them, these will be the render bones
	for(u_int uVertIndex = 0; uVertIndex != m_xSrcProcessor.GetNumVertices(); ++uVertIndex)
	{
		const Tools_SmoothSkin_Vertex& xVert = m_xSrcProcessor.GetVertex(uVertIndex);

		for(u_int i = 0; i != xVert.GetNumWeightings(); ++i)
		{
			u_char uBoneIndex = xVert.GetBoneIndices()[i];
			++auBoneMap[uBoneIndex];

			if (auBoneMap[uBoneIndex] == 1)
			{
				auRenderBoneIndices += uBoneIndex;
			}
		}
	}

//DeadCode VGP 20/03/2006 	auRenderBoneIndices.Sort();
	for(int uBone = 0; uBone != m_xSrcProcessor.GetInputSkin()->NumberOfBones; ++uBone)
	{
		auBoneMap[uBone] = -1;
	}

	for (u_int uBone = 0; uBone != auRenderBoneIndices.GetSize(); ++uBone)
	{
		auBoneMap[auRenderBoneIndices[uBone]] = uBone;
	}

	u_int uFileNameSize = (m_xSrcProcessor.GetInputSkin()->Name.GetSizeInInts()<<2);
	
	u_int uFileSize = 0;
	
	// String to tell us the filename of the character (uFileNameSize)
	uFileSize += uFileNameSize;

	// for each strip:
	// NumberOfTris (u_int)
	// NumberOfVertices (u_int)
	// NumberOfIndices (u_int)
	// MaterialID (int)
	// ElementID (int)
	// strip terminator (u_char)
	uFileSize += (sizeof(u_int) * 5 + sizeof(u_char)) * uNumNonEmptyElements;

	// And then each vertex will be (until optimised!)
	// sizeof(Tools_CharacterProcessor_Vertex)
	uFileSize += m_xSrcProcessor.GetNumVertices() * ( sizeof(float) * 3 // Position
													+ sizeof(float) * 3 // Normal
													+ sizeof(float) * 2 // UV
													+ sizeof(float) * 4 // Bone indices
													+ sizeof(float) * 4 // Bone weightings
													);

	if (bIsNormalMapped)
	{
		uFileSize += m_xSrcProcessor.GetNumVertices() * ( sizeof(float) * 3 // Tangent
														+ sizeof(float) * 3 // Binormal
														);
	}

	// and the render bone map
	uFileSize +=	sizeof(u_char)										// render bone count
					+ auRenderBoneIndices.GetSize() * sizeof(u_char);	// render bone indices

	// and then indices
	uFileSize += uTotalIndices * sizeof(u_short);

	// Chunk size is all the data, 'filename' & header
	int uChunkSize = uFileSize + uFileNameSize + sizeof(Asura_Chunk_ResourceFile);

	Asura_Chunk_ResourceFile_PC_Character xResourceFile;

	u_int uFlags = 0;
	if (bIsNormalMapped)
	{
		uFlags |= uASURA_CHUNK_PC_CHARACTER_FLAG_HASNORMALMAPDATA;
	}
	if(m_bTriLists)
	{	
		uFlags |= uASURA_CHUNK_PC_CHARACTER_FLAG_TRILIST;
	}
	
	xResourceFile.Create
	(
		uChunkSize,
		uFlags,
		ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC,
		ASURA_RESOURCEFILE_TYPE_PC_CHARACTER,
		uFileSize
	);

	xResourceFile.Version = iASURA_CHUNK_PC_CHARACTER_VERSION;

	const u_int uChunkStartPos = xStream.GetPosition();

	WRITEOUTDATA( &xResourceFile, sizeof(Asura_Chunk_ResourceFile) );
	WRITEOUTDATA( m_xSrcProcessor.GetInputSkin()->Name.GetString(),uFileNameSize );	 // chunk name
	WRITEOUTDATA( m_xSrcProcessor.GetInputSkin()->Name.GetString(),uFileNameSize );	 // skin name

	// Strip data
	bool bIsFirstElement = true;
	for(u_int uMaterial = 0; uMaterial < m_xSrcProcessor.GetNumMaterials(); ++uMaterial)
	{
		for(u_int uElement = 0; uElement < m_xSrcProcessor.GetNumElements(); ++uElement)
		{
			if(m_xSrcProcessor.GetMaterial(uMaterial).GetSubElement(uElement).GetNumTris() > 0)
			{
				if(!bIsFirstElement)
				{
					// write out the 'more strips' flag
					u_char ucMoreStrips = 1;
					WRITEOUTDATA( &ucMoreStrips, sizeof(u_char));
				}

				if(m_bTriLists)
				{
					u_int u = m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices/3;
					WRITEOUTDATA( &u, sizeof(u_int) );
				}
				else
				{
					u_int u = m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices -2;
					WRITEOUTDATA( &u, sizeof(u_int) );
				}

				u_int uNumVertices = m_xSrcProcessor.GetNumVertices();
				WRITEOUTDATA( &uNumVertices, sizeof(u_int) );

				WRITEOUTDATA( &m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices, sizeof(u_int));

				#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
					int iMaterialID = ASURA_MATERIAL_NONE;
					WRITEOUTDATA( &iMaterialID, sizeof( int ) );
				#else
					int iMaterial;
					if (bConvertMaterialIDs)
					{
						iMaterial = m_xSrcProcessor.GetMaterial(uMaterial).GetMaterialIDWithinSkin();
					}
					else
					{
						iMaterial = m_xSrcProcessor.GetMaterial(uMaterial).GetMaterialID();
					}
					WRITEOUTDATA( &iMaterial, sizeof(int) );
				#endif

				// Write out element ID
				int iElementID = m_xSrcProcessor.GetElementID(uElement);
				WRITEOUTDATA( &iElementID, sizeof(int) );

				if(bIsFirstElement)
				{
					// Vertex data
					for(u_int vertexIndex = 0; vertexIndex < uNumVertices; ++vertexIndex)
					{
						const Tools_SmoothSkin_Vertex& v = m_xSrcProcessor.GetVertex(vertexIndex);

						WRITEOUTDATA( &v.GetPos().x, sizeof(float));
						WRITEOUTDATA( &v.GetPos().y, sizeof(float));
						WRITEOUTDATA( &v.GetPos().z, sizeof(float));

						WRITEOUTDATA( &v.GetNormal().x, sizeof(float));
						WRITEOUTDATA( &v.GetNormal().y, sizeof(float));
						WRITEOUTDATA( &v.GetNormal().z, sizeof(float));

						float fU = v.GetU();
						WRITEOUTDATA( &fU, sizeof(float));
						float fV = v.GetV();
						WRITEOUTDATA( &fV, sizeof(float));

						u_int i;
						for(i = 0; i < v.GetNumWeightings(); ++i)
						{
							// VGP Remap bone indices to render bones
							// (MH - 01.06.2006) normal-mapped renderer doesn't use remapping
							Asura_Assert(auBoneMap[v.GetBoneIndices()[i]] != -1, "Wrong bone index!");
							u_char ucRenderBoneIndex = static_cast<u_char>
								( bIsNormalMapped ? (v.GetBoneIndices()[i]) : (auBoneMap[v.GetBoneIndices()[i]]) );

							float f = ucRenderBoneIndex * 3.f; // *3 because each bone will use 3 registers
							WRITEOUTDATA( &f, sizeof(float));
						}
						for( ; i < 4; ++i)
						{
							float f = 0;
							WRITEOUTDATA( &f, sizeof(float));
						}
						for(i = 0; i < v.GetNumWeightings(); ++i)
						{
							WRITEOUTDATA( &v.GetBoneWeights()[i], sizeof(float));
						}
						for( ; i < 4; ++i)
						{
							float f = 0;
							WRITEOUTDATA( &f, sizeof(float));
						}

						// tangent and binormal?
						if (bIsNormalMapped)
						{
							WRITEOUTDATA( &v.GetTangent().x, sizeof(float));
							WRITEOUTDATA( &v.GetTangent().y, sizeof(float));
							WRITEOUTDATA( &v.GetTangent().z, sizeof(float));

							WRITEOUTDATA( &v.GetBinormal().x, sizeof(float));
							WRITEOUTDATA( &v.GetBinormal().y, sizeof(float));
							WRITEOUTDATA( &v.GetBinormal().z, sizeof(float));
						}
					}

					// VGP 20Mar2006 Render bone map
					Asura_Assert(auRenderBoneIndices.GetSize() <= 0xFF, "Too many render bones!");
					u_char ucRenderBoneCount = static_cast<u_char>(auRenderBoneIndices.GetSize());
					WRITEOUTDATA(&ucRenderBoneCount, sizeof(u_char));
					for (u_int uBone=0; uBone!=auRenderBoneIndices.GetSize(); ++uBone)
					{
						u_char ucBoneIndex = auRenderBoneIndices[uBone];
						WRITEOUTDATA(&ucBoneIndex, sizeof(u_char));
					}


				} // if bIsFirstElement

				bIsFirstElement = false;

				// Index data
				WRITEOUTDATA( m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_pusStripIndices, sizeof(u_short) * m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices);
			}
		}
	}
	// write out the 'no more strips' flag
	u_char ucMoreStrips = 0;
	WRITEOUTDATA(&ucMoreStrips, sizeof(u_char));

	#undef WRITEOUTDATA

	const u_int uChunkEndPos = xStream.GetPosition();
	Asura_Assert(uChunkEndPos-uChunkStartPos == (u_int)uChunkSize,"Chunk size incorrect");

	delete[] auBoneMap;

	return true;
}

#if defined(ASURA_PC_USE_SHADER_MODEL_3_SMOOTH_SKIN_RENDERER) && ASURA_PC_USE_SHADER_MODEL_3_SMOOTH_SKIN_RENDERER

bool Tools_SmoothSkin_PC_Processor::OutputToChunk_SmoothSkinSM3(Asura_Chunk_Stream& xStream, bool bConvertMaterialIDs) const
{

	// count the indices
	u_int uNumNonEmptyElements = 0;
	u_int uTotalIndices = 0;
	for(u_int uMaterial = 0; uMaterial < m_xSrcProcessor.GetNumMaterials(); ++uMaterial)
	{
		for(u_int uElement = 0; uElement < m_xSrcProcessor.GetNumElements(); ++uElement)
		{
			if(m_xSrcProcessor.GetMaterial(uMaterial).GetSubElement(uElement).GetNumTris() > 0)
			{
				++uNumNonEmptyElements;
				uTotalIndices += m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices;
			}
		}
	}

	u_int uFlags = 0;
	if(m_bTriLists)
	{
		uFlags |= uASURA_CHUNK_PC_SMOOTHSKIN_FLAG_TRILIST;
	}
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	uFlags |= uASURA_CHUNK_PC_SMOOTHSKIN_FLAG_MATERIALRESPONSE;
	#endif


	const Asura_String& xName = m_xSrcProcessor.GetInputSkin()->Name;
	xStream.WriteResourceChunk_Start( ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC, ASURA_RESOURCEFILE_TYPE_PC_SMOOTHSKIN, iASURA_CHUNK_PC_SMOOTHSKIN_VERSION, uFlags, xName);
	
	xStream << xName;

	// Number of strips
	xStream << uNumNonEmptyElements;

	xStream << m_xSrcProcessor.GetNumBoneMaps();

	// Strip data
	for(u_int uMaterial = 0; uMaterial < m_xSrcProcessor.GetNumMaterials(); ++uMaterial)
	{
		for(u_int uElement = 0; uElement < m_xSrcProcessor.GetNumElements(); ++uElement)
		{
			if(m_xSrcProcessor.GetMaterial(uMaterial).GetSubElement(uElement).GetNumTris() > 0)
			{
				#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

					const Asura_Hash_ID uMaterialResponseHashID1 = m_xSrcProcessor.GetMaterial( uMaterial ).GetMaterialResponseHashID(0);
					const Asura_Hash_ID uMaterialResponseHashID2 = m_xSrcProcessor.GetMaterial( uMaterial ).GetMaterialResponseHashID(1);
					xStream << uMaterialResponseHashID1;
					xStream << uMaterialResponseHashID2;

				#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

					int iMaterial;
					if (bConvertMaterialIDs)
					{
						iMaterial = m_xSrcProcessor.GetMaterial(uMaterial).GetMaterialIDWithinSkin();
					}
					else
					{
						iMaterial = m_xSrcProcessor.GetMaterial(uMaterial).GetMaterialID();
					}
					xStream << iMaterial;

				#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

				const u_int uNumStripIndices = m_pxMaterial[uMaterial].m_pxSubElement[uElement].m_uNumStripIndices;
				xStream << uNumStripIndices;

				int iElementID = m_xSrcProcessor.GetElementID(uElement);
				xStream << iElementID;

				if (m_xSrcProcessor.GetNumBoneMaps())
				{
					xStream << m_xSrcProcessor.GetBoneMapIndex(uMaterial, uElement);
				}
			}
		}
	}

	// Vertex data
	const u_int uNumVertices = m_xSrcProcessor.GetNumVertices();
	xStream << uNumVertices;
			
	// Vertex data
	for(u_int vertexIndex = 0; vertexIndex < uNumVertices; ++vertexIndex)
	{
		const Tools_SmoothSkin_Vertex& v = m_xSrcProcessor.GetVertex(vertexIndex);

		Asura_Direct3D_SmoothSkin::Vertex vConv; vConv = v;
		const Tools_SmoothSkin_BoneMap *pxBoneMap = m_xSrcProcessor.GetBoneMap(m_xSrcProcessor.GetVertexBoneMapIndex(vertexIndex));
		if (pxBoneMap)
		{
			for (u_int i = 0; i < v.GetNumWeightings(); ++i)
			{
				u_int uBoneMapEntry = pxBoneMap->GetEntryForBone(v.GetBoneIndices()[i]);
				Asura_Assert(uBoneMapEntry != 0xFFFFFFFF, "Something bad has happened to the bonemap on skin");
				vConv.m_afIndex[i] = (u_char)uBoneMapEntry;
				Asura_Assert(uBoneMapEntry < pxBoneMap->GetNumEntries(), "Something bad has happened to the bonemap on skin");
				Asura_Assert(v.GetBoneIndices()[i] == (u_char)pxBoneMap->GetBone(uBoneMapEntry), "Something bad has happened to the bonemap on skin");
			}
		} 

		xStream.Write( sizeof( vConv), &vConv);
	}

	// Index data
	for(u_int uMaterial = 0; uMaterial < m_xSrcProcessor.GetNumMaterials(); ++uMaterial)
	{
		for(u_int uElement = 0; uElement < m_xSrcProcessor.GetNumElements(); ++uElement)
		{
			const Tools_SmoothSkin_PC_SubElement& xSubElement = m_pxMaterial[uMaterial].m_pxSubElement[uElement];
			const u_int uIndexCount = xSubElement.m_uNumStripIndices;
			
			xStream.Write( sizeof(u_short) * uIndexCount, xSubElement.m_pusStripIndices);
		}
	}
	for(u_int uBoneMap = 0; uBoneMap < m_xSrcProcessor.GetNumBoneMaps(); ++uBoneMap)
	{
		const Tools_SmoothSkin_BoneMap *pxBoneMap = m_xSrcProcessor.GetBoneMap(uBoneMap);
		xStream << pxBoneMap->GetNumEntries();
		xStream.Write( sizeof(u_int) * pxBoneMap->GetNumEntries(), pxBoneMap->GetMap());
	}


	xStream.WriteResourceChunk_End();
	
	return true;
}

#endif

#if defined(ASURA_TARGET_SM3)

bool Tools_SmoothSkin_PC_Processor::InitialiseV2(bool bStripify)
{
	m_szSrcName = m_xSrcProcessor.GetInputSkin()->Name.GetString();
	m_bTriLists = !bStripify;

	// (MP - 10.05.2006) I haven't bothered coping with multiple elements
	// since that functionality isn't used and I suspect never will be. It
	// would be fairly straightforward to extend the code below if necessary.
	// Asura_Assert( m_xSrcProcessor.GetNumElements() == 1, "Multiple elements currently not supported" );

	// (DMR - 12.06.2006) Multiple elements now supported. Note that everything that
	// was called a 'material' before, is called a 'strip' now, which is consistent
	// with the PC version


	// First, count the number of non-empty elements
	u_int uNumStrips = 0; 
	for( u_int u = 0; u < m_xSrcProcessor.GetNumMaterials(); ++u )
	{
		for( u_int v = 0; v < m_xSrcProcessor.GetNumElements(); ++v )
		{
			const Tools_SmoothSkin_Material& xMaterial = m_xSrcProcessor.GetMaterial( u );
			const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement( v );

			if (xSubElement.GetNumTris()>0)
			{
				uNumStrips++;
			}
		}
	}

	m_uNumProcessedStrips = uNumStrips;
	m_pxProcessedStrips   = new ProcessedStrip[uNumStrips];
	m_uTotalVertices	  = 0;
	m_uTotalIndices		  = 0;
	m_uTotalTriangles	  = 0;
	m_uNumProcessedStrips = 0;

	for( u_int u = 0; u < m_xSrcProcessor.GetNumMaterials(); ++u )
	{
		for( u_int v = 0; v < m_xSrcProcessor.GetNumElements(); ++v )
		{
			const Tools_SmoothSkin_Material& xMaterial = m_xSrcProcessor.GetMaterial( u );
			const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement( v );

			if (xSubElement.GetNumTris()>0)
			{			
				ProcessedStrip& xProcessedStrip = m_pxProcessedStrips[m_uNumProcessedStrips];
				m_uNumProcessedStrips++;

				xProcessedStrip.m_iElementID  = m_xSrcProcessor.GetElementID(v);
				xProcessedStrip.m_auMaterialResponseHashID[0] = m_xSrcProcessor.GetMaterial(u).GetMaterialResponseHashID(0);
				xProcessedStrip.m_auMaterialResponseHashID[1] = m_xSrcProcessor.GetMaterial(u).GetMaterialResponseHashID(1);
				xProcessedStrip.m_uBoneMapID = m_xSrcProcessor.GetBoneMapIndex(u, v);

				Tools_TriStripper::Initialise();
				if( m_bTriLists )
				{
					xProcessedStrip.m_uNumIndices = xSubElement.GetNumTris() * 3;
					xProcessedStrip.m_pusIndices = new unsigned short[xProcessedStrip.m_uNumIndices];
					Asura_Mem::Copy( xProcessedStrip.m_pusIndices, xSubElement.GetIndices(), sizeof(unsigned short) * xProcessedStrip.m_uNumIndices );
				}
				else
				{
					Tools_TriStripper::SetOutputType( Tools_TriStripper::OUTPUT_TRISTRIP );
					xProcessedStrip.m_uNumIndices = Tools_TriStripper::CreateStrips( xSubElement.GetNumTris(), xSubElement.GetIndices(),
						&xProcessedStrip.m_pusIndices );
				}

				u_short* pusVtxPermutations;
				// TODO: Set cache size?
				Tools_TriStripper::ComputeVertexPermutation( xProcessedStrip.m_uNumIndices, xProcessedStrip.m_pusIndices,
					m_xSrcProcessor.GetNumVertices(), &pusVtxPermutations,
					&xProcessedStrip.m_uNumVertices );

				Tools_TriStripper::Shutdown();

				// Create new list of vertices just containing the ones used by
				// this material's strips, matching the newly modified index list.
				xProcessedStrip.m_pxVertices = new Tools_SmoothSkin_Vertex[xProcessedStrip.m_uNumVertices];
				for( u_int u = 0; u < xProcessedStrip.m_uNumVertices; ++u )
				{
					const Tools_SmoothSkin_Vertex &xSrcVtx = m_xSrcProcessor.GetVertex( pusVtxPermutations[u] );

					const Tools_SmoothSkin_BoneMap *pxBoneMap = m_xSrcProcessor.GetBoneMap(m_xSrcProcessor.GetVertexBoneMapIndex(pusVtxPermutations[u]));
					if (pxBoneMap)
					{
						short axBoneIndicies[4];
						axBoneIndicies[0] = (short)pxBoneMap->GetEntryForBone(xSrcVtx.GetBoneIndices()[0]);
						axBoneIndicies[1] = (short)pxBoneMap->GetEntryForBone(xSrcVtx.GetBoneIndices()[1]);
						axBoneIndicies[2] = (short)pxBoneMap->GetEntryForBone(xSrcVtx.GetBoneIndices()[2]);
						axBoneIndicies[3] = (short)pxBoneMap->GetEntryForBone(xSrcVtx.GetBoneIndices()[3]);

						xProcessedStrip.m_pxVertices[u].Set(xSrcVtx.GetMaterialIdx(), xSrcVtx.GetPos(), xSrcVtx.GetNormal(),
							xSrcVtx.GetU(), xSrcVtx.GetV(), 
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_VERTEX_FLAGS
							xSrcVtx.GetFlags(), 
#endif
#ifdef TOOLS_SMOOTHSKIN_PROCESSOR_USE_ELEMENT
							xSrcVtx.GetElement(), 
#endif
							axBoneIndicies, xSrcVtx.GetBoneWeights(), xSrcVtx.GetColour());

						xProcessedStrip.m_pxVertices[u].SetTangentAndBinormal( xSrcVtx.GetTangent(), xSrcVtx.GetBinormal() );
						xProcessedStrip.m_pxVertices[u].SetDualPass( Asura_Vector_2(xSrcVtx.GetU2(), xSrcVtx.GetV2()), xSrcVtx.GetAlpha() );

					} 
					else 
					{
						xProcessedStrip.m_pxVertices[u] =  xSrcVtx;
					}
				}

				delete[] pusVtxPermutations;

				m_uTotalVertices  += xProcessedStrip.m_uNumVertices;
				m_uTotalIndices   += xProcessedStrip.m_uNumIndices;
				m_uTotalTriangles += xSubElement.GetNumTris();
			}
		}
	}

	// Also, check we can put all of the indices in a single buffer. Max
	// vertex index is 0xFFFFFF (otherwise we would have to use 32bit index buffers).
	Asura_Assert( m_uTotalIndices <= 0xFFFFFF, "" );

	return true;
}

bool Tools_SmoothSkin_PC_Processor::InitialiseV2FromHierarchyProcessor()
{
	// Don't need to do much, just copy the hierarchy processor's data into our 'processed strips'

	m_bTriLists = false;

	m_szSrcName = Tools_HierarchyProcessor_MultiMaterial_NormalMapped::GetCurrentSkin()->Name.GetString();

	m_uTotalTriangles		= 0; // Will be incremented in the loop
	m_uTotalVertices		= Tools_HierarchyProcessor_MultiMaterial_NormalMapped::GetTotalNumberOfVertices();
	m_uTotalIndices			= Tools_HierarchyProcessor_MultiMaterial_NormalMapped::GetTotalNumberOfIndices();
	m_uNumProcessedStrips	= Tools_HierarchyProcessor_MultiMaterial_NormalMapped::GetTotalNumberOfStrips();

	m_pxProcessedStrips		= new ProcessedStrip[ m_uNumProcessedStrips ];

	Asura_Assert( m_pxProcessedStrips, "Unable to allocate array for skin processor." );
	if( !m_pxProcessedStrips ) return false;

	for( u_int u = 0; u < m_uNumProcessedStrips; ++u )
	{
		const Tools_HierarchyProcessor_MultiMaterial_NormalMapped_Strip& xSrcStrip =
			*Tools_HierarchyProcessor_MultiMaterial_NormalMapped::GetProcessedStrip(u);

		ProcessedStrip& xProcessedStrip = m_pxProcessedStrips[u];

		xProcessedStrip.m_auMaterialResponseHashID[0] = xSrcStrip.m_auMaterialResponseHashID[0];
		xProcessedStrip.m_auMaterialResponseHashID[1] = xSrcStrip.m_auMaterialResponseHashID[1];

		// Indices

		const u_int uNumIndices = xSrcStrip.m_uNumberOfIndices;

		xProcessedStrip.m_uNumIndices = uNumIndices;
		xProcessedStrip.m_pusIndices = new u_short[ uNumIndices ];

		if( !xProcessedStrip.m_pusIndices )
		{
			Asura_Assert( false, "Unable to allocate index array for skin processor." );
			return false;
		}

		Asura_Mem::Copy( xProcessedStrip.m_pusIndices, xSrcStrip.m_pusIndices,
			sizeof( u_short ) * uNumIndices );

		xProcessedStrip.m_iElementID = 0; // TODO: can hierarchies be split into chunking elements?
		xProcessedStrip.m_uBoneMapID = 0; // TODO: can remove bone maps after this

		// Vertices

		const u_int uNumVertices = xSrcStrip.m_uNumberOfVertices;

		xProcessedStrip.m_uNumVertices = uNumVertices;
		xProcessedStrip.m_pxVertices = new Tools_SmoothSkin_Vertex[ uNumVertices ];

		if( !xProcessedStrip.m_pxVertices )
		{
			Asura_Assert( false, "Unable to allocate vertex array for skin processor." );
			return false;
		}

		for( u_int v = 0; v < uNumVertices; ++v )
		{
			const Tools_HierarchyProcessor_MultiMaterial_NormalMapped_Vertex& xSrcVtx =
				xSrcStrip.m_pxVertices[v];

			Tools_SmoothSkin_Vertex& xDstVtx = xProcessedStrip.m_pxVertices[v];

			// Hierarchy verts weighted to just one bone each
			const short ausBoneIndices[4] = { xSrcVtx.m_usBoneIndex, 0, 0, 0 };
			const float afBoneWeights[4] = { 1.0f, 0.0f, 0.0f, 0.0f };

			xDstVtx.Set( 0, xSrcVtx.m_xPosition, xSrcVtx.m_xNormal, xSrcVtx.m_fU0, xSrcVtx.m_fV0,
				ausBoneIndices, afBoneWeights );

			xDstVtx.SetTangentAndBinormal( xSrcVtx.m_xTangent, xSrcVtx.m_xBinormal );

			xDstVtx.SetDualPass( Asura_Vector_2( xSrcVtx.m_fU1, xSrcVtx.m_fV1 ), xSrcVtx.m_fAlpha );
		}

		m_uTotalTriangles += xSrcStrip.m_uNumberOfTriangles;
	}

	return true;
}

static void WriteHalfFloat4Vector( Asura_Chunk_Stream& xStream, const Asura_Vector_3& xVec )
{
	xStream << Asura_HalfFloat( xVec.x );
	xStream << Asura_HalfFloat( xVec.y );
	xStream << Asura_HalfFloat( xVec.z );
	xStream << Asura_HalfFloat( 0.0f );
}

static void WriteSHORT4N( Asura_Chunk_Stream& xStream, float fX, float fY, float fZ, float fW )
{
	short sX = (short)Asura_Maths::FloatToInt_Near( Asura_Maths::ClampToRange( fX * 32767.0f, -32767.0f, 32767.0f ) );
	short sY = (short)Asura_Maths::FloatToInt_Near( Asura_Maths::ClampToRange( fY * 32767.0f, -32767.0f, 32767.0f ) );
	short sZ = (short)Asura_Maths::FloatToInt_Near( Asura_Maths::ClampToRange( fZ * 32767.0f, -32767.0f, 32767.0f ) );
	short sW = (short)Asura_Maths::FloatToInt_Near( Asura_Maths::ClampToRange( fW * 32767.0f, -32767.0f, 32767.0f ) );
	xStream << sX << sY << sZ << sW;
}

bool Tools_SmoothSkin_PC_Processor::OutputToChunkV2( Asura_Chunk_Stream& xStream ) const
{
	xStream.WriteResourceChunk_Start( ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC,
		ASURA_RESOURCEFILE_TYPE_PC_SKIN,
		Asura_Chunk_ResourceFile_PC_Skin::iCURRENT_VERSION, 0, m_szSrcName );


	xStream << m_uNumProcessedStrips;
	xStream << m_uTotalVertices;
	xStream << m_uTotalIndices;
	xStream << m_uTotalTriangles;

	xStream << m_xSrcProcessor.GetNumBoneMaps();

	xStream << (bool)m_bTriLists;

	// Strip headers
	for( u_int u = 0; u < m_uNumProcessedStrips; ++u )
	{
		const ProcessedStrip& xStrip = m_pxProcessedStrips[u];

		const Asura_Hash_ID auMaterials[2] =
		{ xStrip.m_auMaterialResponseHashID[0], xStrip.m_auMaterialResponseHashID[1] };

		Tools_Level_Export::AddMRPShaders( auMaterials[0], auMaterials[1], true );

		xStream << auMaterials[0] << auMaterials[1];

		xStream << xStrip.m_uNumIndices;
		xStream << xStrip.m_iElementID;

		const u_int uMaxWeights = xStrip.GetMaxWeights();
		xStream << uMaxWeights;

		xStream << xStrip.m_uBoneMapID;
	}

	u_int uTotalVert = 0;

	// Vertex data
	for( u_int u = 0; u < m_uNumProcessedStrips; ++u )
	{
		const ProcessedStrip& xStrip = m_pxProcessedStrips[u];

		for( u_int v = 0; v < xStrip.m_uNumVertices; ++v )
		{
			//WriteVtxToStream( xStrip.m_pxVertices[v], xStream );

			const Tools_SmoothSkin_Vertex& xVtx = xStrip.m_pxVertices[v];

			// Position
			//xStream << Asura_HalfFloat( xVtx.GetPos().x ) << Asura_HalfFloat( xVtx.GetPos().y );
			//xStream << Asura_HalfFloat( xVtx.GetPos().z ) << Asura_HalfFloat( 1.0f );

			// using D3DDECLTYPE_FLOAT3
			xStream << xVtx.GetPos().x << xVtx.GetPos().y;
			xStream << xVtx.GetPos().z;

			// Blend indices

			const u_char* const pucIndices = xVtx.GetBoneIndices();
			const u_int uIndices = pucIndices[0] + ( pucIndices[1] << 8 ) + ( pucIndices[2] << 16 ) + ( pucIndices[3] << 24 );

			xStream << uIndices;

			// Blend weights
			//
			const float* const pfWeights = xVtx.GetBoneWeights();
			//
			// Save weights as D3DDECLTYPE_SHORT4N
			WriteSHORT4N( xStream, pfWeights[0], pfWeights[1], pfWeights[2], pfWeights[3] );


			// Tangent, binormal, normal
			WriteHalfFloat4Vector( xStream, xVtx.GetTangent() );
			WriteHalfFloat4Vector( xStream, xVtx.GetBinormal() );
			WriteHalfFloat4Vector( xStream, xVtx.GetNormal() );

			// UV pair
			xStream << Asura_HalfFloat( xVtx.GetU() ) << Asura_HalfFloat( xVtx.GetV() );
			xStream << Asura_HalfFloat( xVtx.GetU2() ) << Asura_HalfFloat( xVtx.GetV2() );

			// Dual texture lerp factor
			xStream << xVtx.GetAlpha();

			++uTotalVert;
		}
	}

	// Index data
	{
		//u_short usRestartIndex;
		//Tools_MeshRestartIndex::IsValidForCurrentPlatform( usRestartIndex );

		// Within each processed strip, indices start at 0. We want one single VB/IB in export
		u_int uIndexOffset = 0;

		for( u_int u = 0; u < m_uNumProcessedStrips; ++u )
		{
			const ProcessedStrip& xStrip = m_pxProcessedStrips[u];

			for( u_int v = 0; v < xStrip.m_uNumIndices; ++v )
			{
				u_short usNextIdx = xStrip.m_pusIndices[v];

				// Offset into cumulative vertex buffer as we go
				//if( usNextIdx != usRestartIndex )
				{
					usNextIdx = static_cast< u_short >( usNextIdx + uIndexOffset );
				}

				xStream << usNextIdx;
			}

			uIndexOffset += xStrip.m_uNumVertices;
		}
	}


	// Render bone map
	for(u_int uBoneMap = 0; uBoneMap < m_xSrcProcessor.GetNumBoneMaps(); ++uBoneMap)
	{
		const Tools_SmoothSkin_BoneMap *pxBoneMap = m_xSrcProcessor.GetBoneMap(uBoneMap);
		xStream << pxBoneMap->GetNumEntries();
		xStream.Write( sizeof(u_int) * pxBoneMap->GetNumEntries(), pxBoneMap->GetMap());
	}


	xStream.WriteResourceChunk_End();

	return true;
}

#endif // #if defined(ASURA_TARGET_SM3)

