//
// Tools_Edge_Selection.cpp
//
// Created by JGP 09/05/06
//
// Copyright (C) 2006 Rebellion (Derby), All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header
#include "Tools_Edge_Selection.h"

// Asura headers
#include "Rendering/Asura_Module.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_Cursor.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Rendering/Asura_Viewport.h"
#include "Camera/Asura_Camera.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/Tools_LOS.h"
#include "Misc/Tools_Prompted_Selection.h"
#include "Misc/Tools_Adjacency.h"
#include "Debugging/Asura_Assert.h"
#include "Tools_Overlays.h"
#include "Misc/Tools_Help.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Render_Options.h"


// static variables


bool Tools_Edge_Selection::s_bShowSelection; 

bool Tools_Edge_Selection::s_bSelectionLocked;
int Tools_Edge_Selection::s_iSelectionMode;								
int Tools_Edge_Selection::s_iSelectionModeListOffset;						

Asura_Vector_2 Tools_Edge_Selection::s_xSelectionOverlayPosition = Asura_Vector_2( 3.0f,6.0f );
bool Tools_Edge_Selection::s_bSelectionOverlayOpen;
u_int Tools_Edge_Selection::s_uSelectionOverlayMinimised = 0;

List<ModuleEdgeTriplet> Tools_Edge_Selection::s_xSelectionList;

EdgeSelectionUndo Tools_Edge_Selection::s_xSelectionUndo;

bool Tools_Edge_Selection::s_bButtonDownForSelection;						
ModuleEdgeTriplet Tools_Edge_Selection::s_xHighlitEdge;					
PolygonID Tools_Edge_Selection::s_xHighlitPolygon;						

bool Tools_Edge_Selection::s_bActive;

float Tools_Edge_Selection::s_fEdgePlaneTolerance = Asura_Maths::DegToRad(20.0f);
float Tools_Edge_Selection::s_fEdgeJoinTolerance = Asura_Maths::DegToRad(45.0f);


static const char* g_aszEdgeSelectionHelpStrings[] =
{
	"LMB :         Select Edge",
	"Shift+LMB :   Deselect Edge",
	"",
	"L :           Clear Selection",
	"",
	"Shift+Z :     Undo Selection",
	"Shift+Y :     Redo Selection",
	"Shift+< :     Undo Selection",
	"Shift+> :     Redo Selection",
	NULL,
};


///////////////////////////////////////////////////////////////////////////////////////
// EdgeSelectionUndo
///////////////////////////////////////////////////////////////////////////////////////


void EdgeSelectionUndo::ProcessUndoData( void* pData )
{
	EdgeSelectionUndoData* pxUndoData = static_cast< EdgeSelectionUndoData* >( pData );

	switch( pxUndoData->m_eUndoType )
	{

		case EdgeSelectionUndoData::UndoType_Added:

			// unselect selected edge
			Tools_Edge_Selection::DeselectEdge( pxUndoData->m_iModuleIndex, pxUndoData->m_aiVertexIndex[0], pxUndoData->m_aiVertexIndex[1], false );

			// reverse the undo data
			pxUndoData->m_eUndoType = EdgeSelectionUndoData::UndoType_Removed;

			break;

		case EdgeSelectionUndoData::UndoType_Removed:

			// select unselected edge
			Tools_Edge_Selection::SelectEdge( pxUndoData->m_iModuleIndex, pxUndoData->m_aiVertexIndex[0], pxUndoData->m_aiVertexIndex[1], false );

			// reverse the undo data
			pxUndoData->m_eUndoType = EdgeSelectionUndoData::UndoType_Added;

			break;

		default:

			Asura_Assert(0, "Unexpected undo type in EdgeSelectionUndo::ProcessUndoData" );
			break;
	}
}


void EdgeSelectionUndo::LogEdgeSelected( int iModule, int iVert1, int iVert2 )
{
	EdgeSelectionUndoData* pxUndo = static_cast< EdgeSelectionUndoData* >( AllocateNextUndoData() );
	if ( !pxUndo ) 
	{
		return;
	}

	pxUndo->m_eUndoType = EdgeSelectionUndoData::UndoType_Added;
	pxUndo->m_iModuleIndex = iModule;
	pxUndo->m_aiVertexIndex[0] = iVert1;
	pxUndo->m_aiVertexIndex[1] = iVert2;

}


void EdgeSelectionUndo::LogEdgeDeselected( int iModule, int iVert1, int iVert2 )
{
	EdgeSelectionUndoData* pxUndo = static_cast< EdgeSelectionUndoData* >( AllocateNextUndoData() );
	if ( !pxUndo ) 
	{
		return;
	}

	pxUndo->m_eUndoType = EdgeSelectionUndoData::UndoType_Removed;
	pxUndo->m_iModuleIndex = iModule;
	pxUndo->m_aiVertexIndex[0] = iVert1;
	pxUndo->m_aiVertexIndex[1] = iVert2;

}


void Tools_Edge_Selection::Undo()
{
	if ( !s_bSelectionLocked )
	{
		if ( s_xSelectionUndo.Undo() )
		{
			Tools_Timed_Message_System::AddMessage( 0xffffff, "Selection Undo" );
		}
	}
}


void Tools_Edge_Selection::Redo()
{
	if ( !s_bSelectionLocked )
	{
		if( s_xSelectionUndo.Redo() )
		{
			Tools_Timed_Message_System::AddMessage( 0xffffff, "Selection Redo" );
		}
	}
}


///////////////////////////////////////////////////////////////////////////////////////
// Tools_Edge_Selection Initialise / Shutdown
///////////////////////////////////////////////////////////////////////////////////////


void Tools_Edge_Selection::Initialise()
{

	s_bSelectionLocked = false;
	s_bShowSelection = true;
	s_iSelectionMode = ESelectionMode_Vertex;
	s_iSelectionModeListOffset = 0;
	s_bButtonDownForSelection = false;
	s_xHighlitEdge.m_iModuleIndex = -1;
	s_xHighlitPolygon.ModuleIndex = -1;

	//load values from .dat file
	TransferModeData( TOOLSMODEDATAFILE_LOAD );
	
	s_bSelectionOverlayOpen = true;

	s_xSelectionUndo.Initialise();

}


void Tools_Edge_Selection::Shutdown()
{

	//save values to .dat file
	TransferModeData(TOOLSMODEDATAFILE_SAVE);

    s_xSelectionList.empty();	

	s_xSelectionUndo.Shutdown();

}


void Tools_Edge_Selection::Activate()
{
	s_bActive = true;

	s_xHighlitEdge.m_iModuleIndex = -1;
	s_xHighlitPolygon.ModuleIndex = -1;

	CreateSelectionOverlay();

	Tools_Help::AddHelpTopic( "Edge Selection", g_aszEdgeSelectionHelpStrings );
}


void Tools_Edge_Selection::Deactivate()
{
	
	// currently it is reset every time it is deactivated
	DeselectAll();

	Tools_Help::RemoveHelpTopic( "Edge Selection" );
	CloseSelectionOverlay();

	s_bActive = false;
}


void Tools_Edge_Selection::TransferModeData( int iMode )
{
	//open the file
	Tools_ModeDataFile xFile( "EdgeSelection.dat", iMode );

	bool bJunk = false;
	// order is important here
	if( false == xFile.GetError())
	{
		xFile.Transfer( s_xSelectionOverlayPosition );
		xFile.Transfer( bJunk );
		xFile.Transfer(s_uSelectionOverlayMinimised);
	}
}


void Tools_Edge_Selection::Update( int eUpdateMode )
{

	//Flag the selection undo to start a new undo batch
	s_xSelectionUndo.MarkUndoStart();

	//if the mouse button has been release , reset the ButtonHeldForSelection flag
	if( !Asura_UserInput::KeyPressed( ASURA_KEYID_MOUSEBUTTON1 ) )
	{
		s_bButtonDownForSelection = false;
	}

	//reset the highlit vertex
	s_xHighlitEdge.m_iModuleIndex = -1;
	s_xHighlitPolygon.ModuleIndex = -1;
	
	if( eUpdateMode != EdgeSelectionUpdate_NoSelection && !Tools_Prompted_Selection::IsSelecting() )
	{
		//act upon keyboard input
		HandleKeyboard();

		if ( eUpdateMode == EdgeSelectionUpdate_Standard )
		{
			//and handle selection using the mouse
			HandleMouseSelection();
		}
	}
	
}


void Tools_Edge_Selection::Render()
{
	//Show the currently selected polygons
	if( s_bShowSelection )
	{
		HighlightSelection();
	}

	if ( s_iSelectionMode == ESelectionMode_Polygon )
	{
		if( s_xHighlitPolygon.ModuleIndex >= 0 )
		{
			HighlightPolygon( s_xHighlitPolygon );
		}
	}
	else
	{
		if( s_xHighlitEdge.m_iModuleIndex >= 0 )
		{
			HighlightEdge( s_xHighlitEdge );
		}
	}

}


///////////////////////////////////////////////////////////////////////////////////////
// Tools_Edge_Selection Overlay
///////////////////////////////////////////////////////////////////////////////////////


void Tools_Edge_Selection::ToggleSelectionOverlay()
{
	s_bSelectionOverlayOpen = !s_bSelectionOverlayOpen;

	if ( s_bSelectionOverlayOpen )
	{
		s_uSelectionOverlayMinimised = 0;
		CreateSelectionOverlay();
	}
	else
	{
		CloseSelectionOverlay();
	}
}


bool Tools_Edge_Selection::IsSelectionOverlayOn()
{
	return s_bSelectionOverlayOpen;
}


void Tools_Edge_Selection::CreateSelectionOverlay()
{
	if ( ( !s_bSelectionOverlayOpen ) || ( Asura_Overlay_System::DoesOverlayExist( TOOLS_OVERLAY_EDGE_SELECTION ) ) )
	{
		return;
	}

	//strings for selection mode options
	static char* s_aszSelectionModeStrings[] =
	{
		"1)Edge",
		"2)Polygon",
		"3)Object",
		"4)Horizontal Edges",
		"5)Connected Edges",
	};

	static char* s_aszSelectionModeHelpStrings[] =
	{
		"Selects a single edge.",
		"Selects all the edges in a polygon.",
		"Selects all the edges in a module.",
		"Selects connected concave edges that are rougly horizontal.",
		"Selects connected edges that are sufficiently convex/concave up until a branch is reached.",
	};


	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay( "Selection", false, &s_xSelectionOverlayPosition, &s_uSelectionOverlayMinimised );
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2( 1.0f, 3.0f );

	//heading for selection mode list box
	pxDialog = new Asura_Dialog( "Selection Mode :", xOffset );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += 1.0f;

	//selection mode list box
	Asura_List_Dialog* pxList = new Asura_List_Dialog( xOffset, s_aszSelectionModeStrings, &s_iSelectionModeListOffset, &s_iSelectionMode, ESelectionMode_NumModes, ESelectionMode_NumModes, NULL );
	pxDialog = pxList;
	pxList->SetListHelpStrings( s_aszSelectionModeHelpStrings );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += 1.2f*ESelectionMode_NumModes + 1.0f;


	
	pxDialog = new Asura_Tickbox_Dialog( "Lock Selection", xOffset, &s_bSelectionLocked );
	pxDialog->SetHelpText( "Disables all controls for changing the current selection." );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog( "Show Selection", xOffset, &s_bShowSelection );
	pxDialog->SetHelpText( "Highlight currently selected edges with a red outline." );
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Function_Dialog( "Select All", xOffset, SelectAll);
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Clear Selection (L)", xOffset, DeselectAll);
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();
	
	pxDialog = new Asura_Function_Dialog( "Undo (Shift+Z)", xOffset, Undo);
	pxDialog->SetHelpText( "Edge selection undo." );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Redo (Shift+Y)", xOffset, Redo);
	pxDialog->SetHelpText( "Edge selection redo." );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();
	
	pxOverlay->m_uID = TOOLS_OVERLAY_EDGE_SELECTION;

	Asura_Overlay_System::SubmitNewOverlay( pxOverlay, false );
}


void Tools_Edge_Selection::CloseSelectionOverlay()
{
	if ( Asura_Overlay_System::DoesOverlayExist( TOOLS_OVERLAY_EDGE_SELECTION ) ) 
	{
		Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_EDGE_SELECTION );
	}
}


///////////////////////////////////////////////////////////////////////////////////////
// Tools_Edge_Selection Overlay
///////////////////////////////////////////////////////////////////////////////////////


const int Tools_Edge_Selection::CountSelectedEdges()
{
	int iCount = 0;
	for( LIF<ModuleEdgeTriplet> xLIF( &s_xSelectionList ); !xLIF.done(); xLIF.next() )
	{
		iCount++;
	}
	return iCount;
}


void Tools_Edge_Selection::SelectEdge( int iModuleIndex, int iVertexIndex1, int iVertexIndex2, bool bAllowUndo )
{
	ModuleEdgeTriplet xEdge( iModuleIndex, iVertexIndex1, iVertexIndex2 );
	if ( !s_xSelectionList.contains( xEdge ) )
	{
		s_xSelectionList.add_entry( xEdge );
		if ( bAllowUndo )
		{
			s_xSelectionUndo.LogEdgeSelected( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1] );
		}
	}
}


void Tools_Edge_Selection::DeselectEdge( int iModuleIndex, int iVertexIndex1, int iVertexIndex2, const bool bAllowUndo )
{
	ModuleEdgeTriplet xEdge( iModuleIndex, iVertexIndex1, iVertexIndex2 );
	
	if ( s_xSelectionList.contains( xEdge ) )
	{
		s_xSelectionList.delete_entry( xEdge );
		if ( bAllowUndo )
		{
			s_xSelectionUndo.LogEdgeDeselected( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1] );
		}
	}
}


void Tools_Edge_Selection::SelectPolygon( int iModuleIndex, int iPolygonIndex )
{
	const Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule( iModuleIndex );
	Asura_Assert( iPolygonIndex >= 0 && iPolygonIndex < pxMesh->NumberOfPolygons, "" );

	const Asura_Mesh_Polygon* const pxPoly = &pxMesh->PolygonDataPtr[ iPolygonIndex ];
	const int iNumVerts = pxPoly->NumberOfVertices();

	for( int i = 0; i < iNumVerts; i++ )
	{
		SelectEdge( iModuleIndex, pxPoly->VertexID[ i ], pxPoly->VertexID[ ( i + 1 ) % iNumVerts ] );
	}
}


void Tools_Edge_Selection::DeselectPolygon( int iModuleIndex, int iPolygonIndex )
{
	const Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule( iModuleIndex );
	Asura_Assert( iPolygonIndex >= 0 && iPolygonIndex < pxMesh->NumberOfPolygons, "" );

	const Asura_Mesh_Polygon* const pxPoly = &pxMesh->PolygonDataPtr[ iPolygonIndex ];
	const int iNumVerts = pxPoly->NumberOfVertices();

	for( int i = 0; i < iNumVerts; i++ )
	{
		DeselectEdge( iModuleIndex, pxPoly->VertexID[ i ], pxPoly->VertexID[ ( i + 1 ) % iNumVerts ] );
	}
}


void Tools_Edge_Selection::SelectObject( int iModuleIndex )
{
	const Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule( iModuleIndex );
	for ( int i = 0; i < pxMesh->NumberOfPolygons; i++ )
	{
		SelectPolygon( iModuleIndex, i );
	}
}



void Tools_Edge_Selection::DeselectObject( int iModuleIndex )
{
	const Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule( iModuleIndex );
	for ( int i = 0; i < pxMesh->NumberOfPolygons; i++ )
	{
		DeselectPolygon( iModuleIndex, i );
	}
}

void Tools_Edge_Selection::SelectHorizontalEdge(const ModuleEdgeTriplet& xEdge)
{
	List<ModuleEdgeTriplet> xEdgeList;
	ListHorizontalEdge(xEdge, xEdgeList);

	for(LIF<ModuleEdgeTriplet> xEdgeLIF(&xEdgeList); !xEdgeLIF.done(); xEdgeLIF.next())
	{
		SelectEdge(xEdgeLIF().m_iModuleIndex,xEdgeLIF().m_aiVertexIndex[0],xEdgeLIF().m_aiVertexIndex[1]);
	}
}

void Tools_Edge_Selection::DeselectHorizontalEdge(const ModuleEdgeTriplet& xEdge)
{
	List<ModuleEdgeTriplet> xEdgeList;
	ListHorizontalEdge(xEdge, xEdgeList);

	for(LIF<ModuleEdgeTriplet> xEdgeLIF(&xEdgeList); !xEdgeLIF.done(); xEdgeLIF.next())
	{
		DeselectEdge(xEdgeLIF().m_iModuleIndex,xEdgeLIF().m_aiVertexIndex[0],xEdgeLIF().m_aiVertexIndex[1]);
	}
}

void Tools_Edge_Selection::SelectContinuousEdge(const ModuleEdgeTriplet& xEdge)
{
	List<ModuleEdgeTriplet> xEdgeList;
	ListContinuousEdge(xEdge, xEdgeList);

	for(LIF<ModuleEdgeTriplet> xEdgeLIF(&xEdgeList); !xEdgeLIF.done(); xEdgeLIF.next())
	{
		SelectEdge(xEdgeLIF().m_iModuleIndex,xEdgeLIF().m_aiVertexIndex[0],xEdgeLIF().m_aiVertexIndex[1]);
	}
}

void Tools_Edge_Selection::DeselectContinuousEdge(const ModuleEdgeTriplet& xEdge)
{
	List<ModuleEdgeTriplet> xEdgeList;
	ListContinuousEdge(xEdge, xEdgeList);

	for(LIF<ModuleEdgeTriplet> xEdgeLIF(&xEdgeList); !xEdgeLIF.done(); xEdgeLIF.next())
	{
		DeselectEdge(xEdgeLIF().m_iModuleIndex,xEdgeLIF().m_aiVertexIndex[0],xEdgeLIF().m_aiVertexIndex[1]);
	}
}



void Tools_Edge_Selection::SelectAll()
{
	const int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for ( int i = 0; i < iNumberOfModules; i++ )
	{
		SelectObject( i );
	}
}


void Tools_Edge_Selection::DeselectAll()
{
	while ( s_xSelectionList.size() > 0 )
	{
		ModuleEdgeTriplet& xEdge = s_xSelectionList.first_entry();
		DeselectEdge( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1] );
	}
}


///////////////////////////////////////////////////////////////////////////////////////
// Tools_Edge_Selection Update functions
///////////////////////////////////////////////////////////////////////////////////////


void Tools_Edge_Selection::HandleKeyboard()
{

	if ( ( Asura_Overlay_System::OverlaysHaveKeyboardFocus() ) || ( Asura_UserInput::KeyPressed( ASURA_KEYID_ALT ) ) )
	{
		return;
	}
	
	if( Asura_UserInput::KeyPressed( ASURA_KEYID_CTRL ) )
	{
	}
	else if( Asura_UserInput::KeyPressed( ASURA_KEYID_SHIFT ) )
	{
		if( Asura_UserInput::DebouncedKeyPressed( 'z' ) || Asura_UserInput::DebouncedKeyPressed( ',' )) // 'shift+z' or 'shift+<' for selection undo
		{
			Undo();
		}
		if ( Asura_UserInput::DebouncedKeyPressed( 'y' ) || Asura_UserInput::DebouncedKeyPressed( '.' )) // 'shift+y' or 'shift+>' for selection redo
		{
			Redo();
		}
	}
	else
	{
		if ( Asura_UserInput::DebouncedKeyPressed( 'l' ) ) // 'L' for clear selection
		{
			DeselectAll();
		}

		//number keys are shorcuts for selecting the mode
		for(int i=0;i<ESelectionMode_NumModes;i++)
		{
			if(Asura_UserInput::DebouncedKeyPressed(i+'1'))
			{
				s_iSelectionMode = i;
			}
		}

	}
}


void Tools_Edge_Selection::HandleMouseSelection()
{
	// If the user is doing anything with overlays or menus,
	// ignore any mousebutton/key presses etc. 
	// or locked
	if ( ( Asura_Interface::UsingMenusOrOverlays() ) || ( s_bSelectionLocked ) )
	{
		return;
	}
	
	//see if the mouse is over a vertex
	ModuleEdgeTriplet xEdge;
	const bool bOverEdge = GetEdgeUnderMouse( xEdge );

	PolygonID xPoly;
	const bool bOverPoly = Tools_Polygon_Selection::GetPolygonUnderMouse( xPoly );
	Asura_Assert( bOverPoly == bOverEdge, "" );

	if ( bOverEdge ) 		
	{
		s_xHighlitEdge = xEdge;
	}

	if ( bOverPoly ) 		
	{
		s_xHighlitPolygon = xPoly;
	}

	if ( ( !Asura_UserInput::DebouncedKeyPressed( ASURA_KEYID_MOUSEBUTTON1 ) ) && ( !s_bButtonDownForSelection ) )
	{
		return;
	}

	s_bButtonDownForSelection = true;

	if ( !bOverEdge )
	{
		return;
	}

	if ( Asura_UserInput::KeyPressed( ASURA_KEYID_SHIFT ) )
	{
		switch( s_iSelectionMode + s_iSelectionModeListOffset )
		{
			case ESelectionMode_Vertex :
				DeselectEdge( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1] );
				break;
			case ESelectionMode_Polygon :
				DeselectPolygon( xPoly.ModuleIndex, xPoly.PolyIndex );
				break;
			case ESelectionMode_Object :
				DeselectObject( xPoly.ModuleIndex );
				break;

			case ESelectionMode_HorizontalEdge :
				DeselectHorizontalEdge( xEdge);
				break;

			case ESelectionMode_ContinuousEdge :
				DeselectContinuousEdge( xEdge);
				break;

		}
	}
	else
	{
		switch( s_iSelectionMode + s_iSelectionModeListOffset )
		{
			case ESelectionMode_Vertex :
				SelectEdge( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1] );
				break;
			case ESelectionMode_Polygon :
				SelectPolygon( xPoly.ModuleIndex, xPoly.PolyIndex );
				break;
			case ESelectionMode_Object :
				SelectObject( xPoly.ModuleIndex );
				break;

			case ESelectionMode_HorizontalEdge :
				SelectHorizontalEdge( xEdge);
				break;

			case ESelectionMode_ContinuousEdge :
				SelectContinuousEdge( xEdge);
				break;


		}
	}
}


bool Tools_Edge_Selection::GetEdgeUnderMouse( ModuleEdgeTriplet& xEdge )
{
	Tools_LOS_EdgeOutput xLOSOut;
	const bool bDone = Tools_LOS::FindEdgeUnderCursor( xLOSOut );
	if ( !bDone )
	{
		return false;
	}

	xEdge.m_iModuleIndex = xLOSOut.m_axEdge[0].m_iModuleIndex;
	xEdge.m_aiVertexIndex[0] = Asura_Maths::Min(xLOSOut.m_axEdge[0].m_aiVertexIndex[ 0 ], xLOSOut.m_axEdge[0].m_aiVertexIndex[ 1 ]);
	xEdge.m_aiVertexIndex[1] = Asura_Maths::Max(xLOSOut.m_axEdge[0].m_aiVertexIndex[ 0 ], xLOSOut.m_axEdge[0].m_aiVertexIndex[ 1 ]);
	return true;
}


bool Tools_Edge_Selection::GetHighlitEdge( ModuleEdgeTriplet& xEdge )
{
	if ( s_xHighlitEdge.m_iModuleIndex < 0 )
	{
		return false;
	}
	
	xEdge = s_xHighlitEdge;
	return true;
}


///////////////////////////////////////////////////////////////////////////////////////
// Tools_Vertex_Selection render
///////////////////////////////////////////////////////////////////////////////////////


void Tools_Edge_Selection::HighlightSelection()
{
	for( LIF<ModuleEdgeTriplet> xLIF( &s_xSelectionList ); !xLIF.done(); xLIF.next() )
	{
		ModuleEdgeTriplet& xEdge = xLIF();
		Tools_Render::RenderEdge( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1], 0xffff0000 );
	}
}


void Tools_Edge_Selection::HighlightEdge( ModuleEdgeTriplet& xEdge )
{
	if ( !Tools_Render::HighlightPolyOrObjectUnderCursor ) 
	{
		return;
	}

	float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
	fAlpha = 255.0f*fAlpha*fAlpha;
	const unsigned int uColour = 0x00ffffff | ( Asura_Maths::FloatToInt( fAlpha ) << 24 );
	
	Tools_Render::RenderEdge( xEdge.m_iModuleIndex, xEdge.m_aiVertexIndex[0], xEdge.m_aiVertexIndex[1], uColour );
}


void Tools_Edge_Selection::HighlightPolygon( PolygonID& xPolygon )
{
	if ( !Tools_Render::HighlightPolyOrObjectUnderCursor ) 
	{
		return;
	}

	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule( xPolygon.ModuleIndex );
	if ( xPolygon.PolyIndex >= pxMesh->NumberOfPolygons ) 
	{
		return;
	}

	float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
	fAlpha = 255.0f*fAlpha*fAlpha;
	const unsigned int uColour = 0x00ffffff | ( Asura_Maths::FloatToInt( fAlpha ) << 24 );
	
	Tools_Render::RenderPolygonOutline( xPolygon.PolyIndex, xPolygon.ModuleIndex, uColour, true );
}


void Tools_Edge_Selection::HandleMeshChange(int iModuleIndex)
{
	//Just going to clear the selection and undo for now. RAR 14/03/2007
	while(s_xSelectionList.size()) s_xSelectionList.delete_first_entry();
	s_xSelectionUndo.FlushUndoBuffer();
	s_xHighlitEdge.m_iModuleIndex = -1;
	
}

void Tools_Edge_Selection::DeleteModule(int iModuleIndex)
{
	//Just going to clear the selection and undo for now. RAR 14/03/2007
	while(s_xSelectionList.size()) s_xSelectionList.delete_first_entry();
	s_xSelectionUndo.FlushUndoBuffer();
	s_xHighlitEdge.m_iModuleIndex = -1;
}


void Tools_Edge_Selection::ListHorizontalEdge(const ModuleEdgeTriplet& xStartEdge,List<ModuleEdgeTriplet>& xEdgeList)
{
	Asura_Vector_3 xPlaneNormal(0.0f,1.0f,0.0f);
	const float fSinPlaneTolerance = Asura_Maths::Sine(Asura_Maths::DegToRad(20.0f));//how close the edge needs to be to horizontal
	const Asura_Angle fEdgeAngleTolerance = Asura_Maths::DegToRad(180.0f + 45.0f);//accept edges with angle greater than this 

	
	//get a direction vector for the starting edge
	int iStartModule = xStartEdge.m_iModuleIndex;
	const Asura_Mesh* pxStartMesh = Asura_Module_System::GetMeshForModule(iStartModule);

	Asura_Vector_3 xStartDir = pxStartMesh->VertexDataPtr[xStartEdge.m_aiVertexIndex[1]].Position - pxStartMesh->VertexDataPtr[xStartEdge.m_aiVertexIndex[0]].Position;
	if(xStartDir == Asura_Maths::ZeroVector3) return;
	xStartDir.Normalise();


	xEdgeList.add_entry(xStartEdge);

	for(LIF<ModuleEdgeTriplet> xEdgeLIF(&xEdgeList);!xEdgeLIF.done(); xEdgeLIF.next())
	{
		ModuleEdgeTriplet xEdge1 = xEdgeLIF();

		int iModule1 = xEdge1.m_iModuleIndex;
		const Asura_Mesh* pxMesh1 = Asura_Module_System::GetMeshForModule(iModule1);
		Asura_Vector_3 xModulePos1 = Asura_Module_System::GetPositionOfModule(iModule1);
		
		//check for edges that connect to either of the vertices in this edge
		for(int i=0;i<2;i++)
		{
			int iVert1 = i ? xEdge1.m_aiVertexIndex[1] : xEdge1.m_aiVertexIndex[0];
			Asura_Vector_3 xVertPos1 = xModulePos1 + pxMesh1->VertexDataPtr[iVert1].Position;

			
			//check each polygon that is using this vertex
			 const Tools_Polygon_Adjacency* pxPolyAdj = Tools_Adjacency::GetVertexAdjacencies(iModule1,iVert1);

			for(const Tools_Adjacent_Polygon* pxAdj = pxPolyAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
			{
				//ignore adjacencies between zones
				if(pxAdj->IsBetweenZones()) continue;

				int iModule2 = pxAdj->GetModuleIndex();
				const Asura_Mesh* pxMesh2 = Asura_Module_System::GetMeshForModule(iModule2);
				Asura_Vector_3 xModulePos2 = Asura_Module_System::GetPositionOfModule(iModule2);

				//Ignore hidden polygons. This allows the function to act on render/collision mesh depending on what
				//is being viewed.
				int iPoly2 = pxAdj->GetPolygonIndex();
				if(Tools_Render_Options::IsHidden(iModule2,iPoly2)) continue;


				//find the equivalent vertex in the adjacent polygon
				const Asura_Mesh_Polygon* pxPoly2 = &pxMesh2->PolygonDataPtr[iPoly2];

				int iEquivVertIndex = -1;
				int iNumPolyVerts2 = pxPoly2->NumberOfVertices();
				for(int k=0;k<iNumPolyVerts2;k++)
				{
					Asura_Vector_3 xVertPos2 = xModulePos2 + pxMesh2->VertexDataPtr[pxPoly2->VertexID[k]].Position;

					if(Tools_Adjacency::IsDistanceWithinTolerance(xVertPos2 - xVertPos1))
					{
						iEquivVertIndex = k;
						break;
					}
				}
				if(iEquivVertIndex == -1) continue;

				//need to consider the edge either side of the vertex in the polygon
				for(int l=0;l<2;l++)
				{
					int iEdgeVert1 = pxPoly2->VertexID[iEquivVertIndex];
					int iEdgeVert2 = l 	? pxPoly2->VertexID[(iEquivVertIndex + 1) % iNumPolyVerts2] 
										: pxPoly2->VertexID[(iEquivVertIndex + iNumPolyVerts2 - 1) % iNumPolyVerts2];

					ModuleEdgeTriplet xEdge2(iModule2,iEdgeVert1,iEdgeVert2);

					//make sure this edge hasn't already been done
					if(xEdgeList.contains(xEdge2)) continue;

					//see if this is at an edge on the mesh
					Asura_Angle fEdgeAngle = -Asura_Maths::Pi;
					GetAngleAtEdge(xEdge2,fEdgeAngle);
					if(fEdgeAngle < fEdgeAngleTolerance) continue;
 
					//get the direction of this edge
					Asura_Vector_3 xEdgeDir2 = pxMesh2->VertexDataPtr[iEdgeVert2].Position - pxMesh2->VertexDataPtr[iEdgeVert1].Position;
					if(xEdgeDir2 == Asura_Maths::ZeroVector3) continue;
					xEdgeDir2.Normalise();

					
					
					//Determine if its close enough to horizontal
					
					float fDot = xPlaneNormal*xEdgeDir2;
					if(Asura_Maths::Abs(fDot) > fSinPlaneTolerance) continue;
					

					xEdgeList.add_entry(xEdge2);
				}
			}
		}	
	}

	//finally make sure the first edge was acceptable after all
	{
		bool bStartEdgeOkay = true;

		//check at an edge in mesh
		Asura_Angle fEdgeAngle = -Asura_Maths::Pi;
		GetAngleAtEdge(xStartEdge,fEdgeAngle);
		if(fEdgeAngle <= fEdgeAngleTolerance) bStartEdgeOkay = false;

		//check if horizontal
		float fDot = xPlaneNormal*xStartDir;
		if(Asura_Maths::Abs(fDot) > fSinPlaneTolerance) bStartEdgeOkay = false;

		if(!bStartEdgeOkay) xEdgeList.delete_first_entry();

	}

}

bool Tools_Edge_Selection::GetAngleAtEdge(const ModuleEdgeTriplet& xEdge,Asura_Angle& fAngle)
{
	//Get a list of all polygons that use the first vertex
	const Tools_Polygon_Adjacency* pxPolyAdj1 = Tools_Adjacency::GetVertexAdjacencies(xEdge.m_iModuleIndex,xEdge.m_aiVertexIndex[0]);
	const Tools_Polygon_Adjacency* pxPolyAdj2 = Tools_Adjacency::GetVertexAdjacencies(xEdge.m_iModuleIndex,xEdge.m_aiVertexIndex[1]);

	//search for polygons that use bother vertices
	int iNumPoly = 0;
	PolygonID axPoly[2];

	for(const Tools_Adjacent_Polygon* pxAdj1 = pxPolyAdj1->GetFirstAdjacency(); pxAdj1 && iNumPoly<2; pxAdj1 = pxAdj1->GetNext())
	{
		if(pxAdj1->IsBetweenZones()) continue;

		int iModule = pxAdj1->GetModuleIndex();
		int iPoly = pxAdj1->GetPolygonIndex();
 
 		//Ignore hidden polygons. This allows the function to act on render/collision mesh depending on what
 		//is being viewed.
		if(Tools_Render_Options::IsHidden(iModule,iPoly)) continue;


		//check if this polygon uses the second vertex
		for(const Tools_Adjacent_Polygon* pxAdj2 = pxPolyAdj2->GetFirstAdjacency(); pxAdj2; pxAdj2 = pxAdj2->GetNext())
		{
			if(pxAdj2->IsBetweenZones()) continue;

			if((pxAdj2->GetModuleIndex() == iModule) && (pxAdj2->GetPolygonIndex() == iPoly))
			{
				axPoly[iNumPoly].ModuleIndex = iModule;
				axPoly[iNumPoly].PolyIndex = iPoly;
				iNumPoly++;
				break;
			}
		}
	}

	if(iNumPoly<2) return false;

	//get world positions of both edge vertices
	const Asura_Mesh* pxEdgeMesh = Asura_Module_System::GetMeshForModule(xEdge.m_iModuleIndex);
	Asura_Vector_3 axEdgePos[2];
	
	axEdgePos[0] = pxEdgeMesh->VertexDataPtr[xEdge.m_aiVertexIndex[0]].Position + Asura_Module_System::GetPositionOfModule(xEdge.m_iModuleIndex);
	axEdgePos[1] = pxEdgeMesh->VertexDataPtr[xEdge.m_aiVertexIndex[1]].Position + Asura_Module_System::GetPositionOfModule(xEdge.m_iModuleIndex);
	


	//get direction vector along the edge
	Asura_Vector_3 xEdgeDir = axEdgePos[1] - axEdgePos[0];
	if(xEdgeDir == Asura_Maths::ZeroVector3) return false;
	xEdgeDir.Normalise();




	//get direction vector from the edge in each polygon , and make it perpendicular to the edge
	Asura_Vector_3 axDir[2];
	for(int i=0;i<2;i++)
	{
		const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(axPoly[i].ModuleIndex);
		Asura_Vector_3 xModulePos = Asura_Module_System::GetPositionOfModule(axPoly[i].ModuleIndex);
		const Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[axPoly[i].PolyIndex];

		Asura_Vector_3 xBestDir;
		float fBestDirLength = 0.0f;

		//check each vertex for the one furthest from the edge
		for(int j=0;j<pxPoly->NumberOfVertices();j++)
		{
			Asura_Vector_3 xVertPos = pxMesh->VertexDataPtr[pxPoly->VertexID[j]].Position + xModulePos;

			//compare againt the two edge positions
			bool bVertOkay = true;
			for(int k=0;k<2;k++)
			{
				if(Tools_Adjacency::IsDistanceWithinTolerance(xVertPos - axEdgePos[k]))
				{
					bVertOkay = false;
					break;
				}				
			}
			if(!bVertOkay) continue;

			//get direction from edge to the vertex
			Asura_Vector_3 xDir = xVertPos - axEdgePos[0];
			//make perpendicular to the edge
			xDir -= xEdgeDir * (xDir*xEdgeDir);

			//use this vector if the longest found
			float fLength = xDir.Magnitude();
			if(fLength > fBestDirLength)
			{
				xBestDir = xDir;
				fBestDirLength = fLength;
			}
		}

		if(!fBestDirLength) return false;

		axDir[i] = xBestDir;
		axDir[i].Normalise();
	}

	//determine the angle between these two directions
	float fCosAngle = axDir[0] * axDir[1];

	fAngle = Asura_Maths::ArcCosine(Asura_Maths::ClampToRange(fCosAngle,-1.0f,1.0f));

	//compare the edge on the second polygon to the first polygon's normal to determine if the angle is more than 180 degrees
	Asura_Mesh* pxMesh1 = Asura_Module_System::GetMeshForModule(axPoly[0].ModuleIndex);
	if(!pxMesh1->m_bFaceNormalsValid) pxMesh1->CreateFaceNormals();

	float fDot = axDir[1] * pxMesh1->NormalsDataPtr[axPoly[0].PolyIndex].FaceNormal;
	if(fDot < 0.0f)
	{
		fAngle = 2 * Asura_Maths::Pi - fAngle;
	}

	return true;

}



void Tools_Edge_Selection::ListContinuousEdge(const ModuleEdgeTriplet& xStartEdge,List<ModuleEdgeTriplet>& xEdgeList)
{
	Asura_Vector_3 xPlaneNormal(0.0f,1.0f,0.0f);
	const Asura_Angle fConcaveEdgeAngleTolerance = Asura_Maths::DegToRad(180.0f + 45.0f);//accept edges with angle greater than this 
	const Asura_Angle fConvexEdgeAngleTolerance = Asura_Maths::DegToRad(180.0f - 45.0f);//accept edges with angle lower than this 

	bool bWantConcaveEdges;	

	//check to see if the starting edge is okay
	//and determine whether to go for convex or concanve edges
	//check at an edge in mesh
	Asura_Angle fEdgeAngle = -Asura_Maths::Pi;
	GetAngleAtEdge(xStartEdge,fEdgeAngle);

	if(fEdgeAngle >= fConcaveEdgeAngleTolerance)
	{
		bWantConcaveEdges = true;
	}
	else if(fEdgeAngle <= fConvexEdgeAngleTolerance)
	{
		bWantConcaveEdges = false;
	}
	else
	{
		//not sufficiently one way or the other
		return;
	}

	xEdgeList.add_entry(xStartEdge);

	
	
	//head out in both directions from the starting edge
	for(int i=0;i<2; i++)
	{
		ModuleEdgeTriplet xPrevEdge = xStartEdge;
		int iPrevDir = i;
		
		while(1)
		{
			int iPrevModule = xPrevEdge.m_iModuleIndex;
			const Asura_Mesh* pxPrevMesh = Asura_Module_System::GetMeshForModule(iPrevModule);
			Asura_Vector_3 xPrevModulePos = Asura_Module_System::GetPositionOfModule(iPrevModule);
			
			//search for edges connected to the vertex in the previous edge
			
			int iPrevVert = xPrevEdge.m_aiVertexIndex[iPrevDir];
			Asura_Vector_3 xPrevVertPos = xPrevModulePos + pxPrevMesh->VertexDataPtr[iPrevVert].Position;


			List<ModuleEdgeTriplet> xConnectedEdgeList;
			
			//check each polygon that is using this vertex
			const Tools_Polygon_Adjacency* pxVertAdj = Tools_Adjacency::GetVertexAdjacencies(iPrevModule,iPrevVert);

			for(const Tools_Adjacent_Polygon* pxAdj = pxVertAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
			{
				//ignore adjacencies between zones
				if(pxAdj->IsBetweenZones()) continue;

				int iModule2 = pxAdj->GetModuleIndex();
				const Asura_Mesh* pxMesh2 = Asura_Module_System::GetMeshForModule(iModule2);
				Asura_Vector_3 xModulePos2 = Asura_Module_System::GetPositionOfModule(iModule2);

				//Ignore hidden polygons. This allows the function to act on render/collision mesh depending on what
				//is being viewed.
				int iPoly2 = pxAdj->GetPolygonIndex();
				if(Tools_Render_Options::IsHidden(iModule2,iPoly2)) continue;


				//find the equivalent vertex in the adjacent polygon
				const Asura_Mesh_Polygon* pxPoly2 = &pxMesh2->PolygonDataPtr[iPoly2];

				int iEquivVertIndex = -1;
				int iNumPolyVerts2 = pxPoly2->NumberOfVertices();
				for(int k=0;k<iNumPolyVerts2;k++)
				{
					Asura_Vector_3 xVertPos2 = xModulePos2 + pxMesh2->VertexDataPtr[pxPoly2->VertexID[k]].Position;

					if(Tools_Adjacency::IsDistanceWithinTolerance(xVertPos2 - xPrevVertPos))
					{
						iEquivVertIndex = k;
						break;
					}
				}
				if(iEquivVertIndex == -1) continue;

				//need to consider the edge either side of the vertex in the polygon
				for(int l=0;l<2;l++)
				{
					int iEdgeVert1 = pxPoly2->VertexID[iEquivVertIndex];
					int iEdgeVert2 = l 	? pxPoly2->VertexID[(iEquivVertIndex + 1) % iNumPolyVerts2] 
										: pxPoly2->VertexID[(iEquivVertIndex + iNumPolyVerts2 - 1) % iNumPolyVerts2];

					ModuleEdgeTriplet xEdge2(iModule2,iEdgeVert1,iEdgeVert2);

					//make sure this edge hasn't already been done
					if(xConnectedEdgeList.contains(xEdge2)) continue;
					//and this this isn't the previous edge
					if(xEdge2 == xPrevEdge) continue;

					//see if this is at an edge on the mesh
					Asura_Angle fEdgeAngle = -Asura_Maths::Pi;
					GetAngleAtEdge(xEdge2,fEdgeAngle);
					if(bWantConcaveEdges)
					{
						if(fEdgeAngle < fConcaveEdgeAngleTolerance) continue;
					}
					else
					{
						if(fEdgeAngle > fConvexEdgeAngleTolerance) continue;
					}
 

					xConnectedEdgeList.add_entry(xEdge2);
				}
			}

			//If there is exactly one connected edge we can continue along it.
			//Might be worth doing something about choosing the best direction for more than 1 , but I'm not
			//sure it would be an improvement.
			if(xConnectedEdgeList.size() != 1) break;

			ModuleEdgeTriplet xConnEdge = xConnectedEdgeList.first_entry();
			//stop if we already have the edge (looped back to start maybe)
			if(xEdgeList.contains(xConnEdge)) break;

			xEdgeList.add_entry(xConnEdge);


			//set this edge as the one to continue from
			xPrevEdge = xConnEdge;

			//work out which vertex in the connected edge doesn't match the vertex we were going from
			Asura_Mesh* pxConnMesh = Asura_Module_System::GetMeshForModule(xConnEdge.m_iModuleIndex);
			Asura_Vector_3 xConnVertPos = pxConnMesh->VertexDataPtr[xConnEdge.m_aiVertexIndex[0]].Position + Asura_Module_System::GetPositionOfModule(xConnEdge.m_iModuleIndex);

			iPrevDir = Tools_Adjacency::IsDistanceWithinTolerance(xConnVertPos - xPrevVertPos) ? 1 : 0;
		}
	}

}
