//
// Tools_HashTable.h
//
// Created by KJFL - 2004-07-12 17:30:13
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//
// KJFL - 2004-07-12 - this class was originally written by Morgan for use in
// the smooth-skin character processor, but it seemed worthy of Asurification.
//
// MGH 2007-10-17: This class was only used in Tools. I've renamed it
// from Asura_HashTable to Tools_HashTable, in order to make space in core
// Asura for a shiny new hash table class that has more functionality.
//

#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __TOOLS_HASH_TABLE_H__
#define __TOOLS_HASH_TABLE_H__

// Includes
#include "Debugging/Asura_Assert.h"

// Generic hash table class that uses hashing with chaining and keeps
// pointers to the inserted data, rather than copying. Requires that the type
// implements a == operator and a method GetHashID(), returning an
// Asura_Hash_ID value. The table size is a power of two, for quick index
// calculation, and so obviously this requires a well distributed hash
// function.
template< typename T >
class Tools_HashTable
{

    public:

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        Tools_HashTable( u_int uSize );
        ~Tools_HashTable( void );

        // M E T H O D S /////////////////////////////////////////////////////

        void Clear( void );
        bool Delete( const T* pxData );
        u_int GetNumEntries( void ) const { return m_uNumEntries; }
        const T* Insert( const T* pxData );
        bool Contains( const T& xData ) const;

		//Finds a matching entry in the table if it exists
		const T* Find(const T& xData);

    private:

        // T Y P E S /////////////////////////////////////////////////////////

        struct Entry
        {
            // In main table, both pointers will be NULL iff there's no chain
            // at this index. Beyond the head of the chain, pxVertex will
            // always be non-NULL and pxNext will be non-NULL except for the
            // last entry in the chain.
            Entry*   pxNext;
            const T* pxData;
        };

        // S T A T E /////////////////////////////////////////////////////////

        // Main chain table.
        Entry* m_pxTable;
        u_int  m_uMask;

        // Stats.
        u_int  m_uNumEntries;
#if _DEBUG
        u_int  m_uNumEmptyChains;
        u_int  m_uMaxChainLength;
#endif

};



template< typename T >
Tools_HashTable<T>::Tools_HashTable( u_int uSize )
{
    // Must be a power of two, so we can just mask.
    Asura_Assert( uSize == (1u << Asura_Maths::GetCeilLog2(uSize)), "" );

    m_pxTable = new Entry[uSize];
    for( u_int u = 0; u < uSize; ++u )
    {
        m_pxTable[u].pxNext = NULL;
        m_pxTable[u].pxData = NULL;
    }
    m_uMask = uSize - 1;

    m_uNumEntries = 0;
    
#if _DEBUG
    m_uNumEmptyChains = uSize;
    m_uMaxChainLength = 0;
#endif
}

template< typename T >
Tools_HashTable<T>::~Tools_HashTable( void )
{
    Clear();

    // Free main table.
    delete[] m_pxTable;
}

template< typename T >
void Tools_HashTable<T>::Clear( void )
{
    // Delete any chains we've accumulated.
    for( u_int u = 0; u <= m_uMask; ++u )
    {
	    Entry& xChainHead = m_pxTable[ u];

#if _DEBUG
        if( NULL != xChainHead.pxData)
        {
            Asura_Assert( this->m_uNumEmptyChains <= (this->m_uMask + 1), "Debug Count of Empty Chains is wrong");
            this->m_uNumEmptyChains += 1;
        }
#endif

		{
			Entry* pxEntry = xChainHead.pxNext;
			while( pxEntry )
			{
				Entry* pxToDelete = pxEntry;
				pxEntry = pxEntry->pxNext;
				delete pxToDelete;
			}
		}
        
        xChainHead.pxNext = NULL;
        xChainHead.pxData = NULL;
    }

    m_uNumEntries = 0;
#if _DEBUG
    Asura_Assert( this->m_uNumEmptyChains == (this->m_uMask + 1), "Debug Count of Empty Chains is wrong");
#endif
}

// Returns true iff the data was found and successfully removed.
template< typename T >
bool Tools_HashTable<T>::Delete( const T* pxData )
{
    u_int uIdx = pxData->GetHashID() & m_uMask;

    if( m_pxTable[uIdx].pxData )
    {
        Entry* pxEntry     = &m_pxTable[uIdx];
        Entry* pxPrevEntry = NULL;
    
        while( pxEntry )
        {
            if( *pxData == *pxEntry->pxData ) { break; }

            pxPrevEntry = pxEntry;
            pxEntry = pxEntry->pxNext;
        }

        if( pxEntry )
        {
            // Ok, this is the entry we're looking for. Is this actually
            // the head of the chain?
            if( pxPrevEntry )
            {
                // No. We need to fix up the previous entry and then
                // deallocate the current one.
                pxPrevEntry->pxNext = pxEntry->pxNext;
                delete pxEntry;
            }
            else
            {
                // Yes. Just lose pointer to data and we're done.
                Asura_Assert( !pxEntry->pxNext, "" );
                pxEntry->pxData = NULL;
            }
            
            return true;
        }
    }

    return false;
}

// Attempts to insert the specific data into the table. If identical data
// already exists then the table is not modified and a pointer to that other
// data is returned. Returns NULL on failure. Otherwise returns the original
// input pointer, on successfully inserting new distinct data.
template< typename T >
const T* Tools_HashTable<T>::Insert( const T* pxData )
{
    u_int uIdx = pxData->GetHashID() & m_uMask;

    Entry& xChainHead = m_pxTable[uIdx];
    if( xChainHead.pxData )
    {
        // Search the existing chain.
        Entry* pxSearch = &xChainHead;
        while( true )
        {
            if( *pxData == *pxSearch->pxData ) { return pxSearch->pxData; }

            if( !pxSearch->pxNext ) { break; }

            pxSearch = pxSearch->pxNext;
        }

        // Ok, no match found so append a new entry.
        Entry* pxNewEntry = new Entry;
        if( !pxNewEntry ) { return NULL; }
        pxNewEntry->pxNext = NULL;
        pxNewEntry->pxData = pxData;
        pxSearch->pxNext = pxNewEntry;
    }
    else
    {
        // Start new chain.
        Asura_Assert( !xChainHead.pxNext, "" );
        xChainHead.pxData = pxData;
    }

    // If we get here then we've successfully inserted a new entry so we need
    // update our count.
    ++m_uNumEntries;

#if _DEBUG
    {
        u_int uChainLength = 0;
        for( const Entry* pxEntry = &xChainHead; pxEntry; pxEntry = pxEntry->pxNext )
        {
            ++uChainLength;
        }
        m_uMaxChainLength = Asura_Maths::Max( uChainLength, m_uMaxChainLength );

        // Was it the first entry in the chain?
        if( !xChainHead.pxNext )
        {
            Asura_Assert( m_uNumEmptyChains > 0, "" );
            --m_uNumEmptyChains;
        }
    }
#endif

    return pxData;
}

template< typename T >
bool Tools_HashTable<T>::Contains( const T& xData ) const
{
    u_int uIdx = xData.GetHashID() & m_uMask;

    Entry& xChainHead = m_pxTable[uIdx];
    if ( xChainHead.pxData )
    {
        // Search the existing chain.
        Entry* pxSearch = &xChainHead;
        while ( pxSearch )
        {
            if( xData == *pxSearch->pxData ) { return true; }
            pxSearch = pxSearch->pxNext;
        }

        return false;
    }
    else
    {
        return false;
    }
}

template< typename T >
const T* Tools_HashTable<T>::Find( const T& xData )
{
    u_int uIdx = xData.GetHashID() & m_uMask;

    Entry& xChainHead = m_pxTable[uIdx];
    if ( xChainHead.pxData )
    {
        // Search the existing chain.
        Entry* pxSearch = &xChainHead;
        while ( pxSearch )
        {
            if( xData == *pxSearch->pxData ) { return pxSearch->pxData; }
            pxSearch = pxSearch->pxNext;
        }

        return 0;
    }
    else
    {
        return 0;
    }
}



// End class Tools_HashTable.

#endif//__TOOLS_HASH_TABLE_H__
