/**
*
* \file Tools_Large_AABB_Tree.h
*
* Created by Kevin Lea - 21/06/2002 14:41:38
*
* Copyright (C) 2002 Rebellion, All Rights Reserved.
*
* \brief
*
**/
#ifdef ASURA_USE_PRAGMA_ONCE
	#pragma once
#endif

#ifndef __TOOLS_LARGE_AABB_TREE_H__
#define __TOOLS_LARGE_AABB_TREE_H__


#include "Core/Asura_Maths.h"

struct Tools_Large_AABB_Tree_Node
{
	// KJL - 08/05/2002 19:37:27 - Child Extents
	//
	// These extents are shared between the left and right nodes,
	// such that the bounding box of a child node is constructed
	// by a mixture of the parent's extents and these new ones,
	// as determined by the flags below. This halves the space
	// required.
	//
	// Further optimisations could be made by compressing this
	// data - e.g. using unsigned chars which indicate the scale
	// of the new extents relative to the parent bounding box,
	// rather than just using floats.
	Asura_Bounding_Box ChildExtents;

	// KJL - 08/05/2002 19:41:25 - node index numbers
	// if node is a leaf then index is to a triangle/module/etc,
	// otherwise it's an index to its child node
	unsigned int LeftIndex;
	unsigned int RightIndex;


	unsigned char MinXBelongsToLeft :1;
	unsigned char MinYBelongsToLeft :1;
	unsigned char MinZBelongsToLeft :1;
	unsigned char MaxXBelongsToLeft :1;
	unsigned char MaxYBelongsToLeft :1;
	unsigned char MaxZBelongsToLeft :1;
	unsigned char LeftIsLeaf		:1;
	unsigned char RightIsLeaf		:1;

	// KJL - 16/05/2002 16:12:41 - hmm, this structure may actually
	// be irritating to some processors since it's an odd number of
	// bytes.
};

// Data passed into threadsafe version - class may be derived by calling methods
class Tools_Large_AABB_Tree_AABBData
{
	public:
		typedef void (*TLAT_AABBTestCallback)( u_int uIndex, Tools_Large_AABB_Tree_AABBData* pxData );

		Tools_Large_AABB_Tree_AABBData( const Asura_Bounding_Box& xTestAABB, TLAT_AABBTestCallback pfnCallback )
		{
			m_xTestAABB = xTestAABB;
			m_pfnCallback = pfnCallback;
		}

		const Asura_Bounding_Box& GetTestAABB( void ) const
		{
			return m_xTestAABB;
		}

		TLAT_AABBTestCallback GetCallback( void ) const 
		{
			return m_pfnCallback;
		}

		u_int GetModuleIndex( void ) const
		{
			return m_uModuleIndex;
		}

		void SetModuleIndex( u_int uIndex )
		{
			m_uModuleIndex = uIndex;
		}
	private:

		Asura_Bounding_Box m_xTestAABB;

		TLAT_AABBTestCallback m_pfnCallback;

		u_int m_uModuleIndex;
		
};

// Data passed into threadsafe version - class may be derived by calling methods
class Tools_Large_AABB_Tree_RayData
{
	public:

		Tools_Large_AABB_Tree_RayData( const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection )
		: m_xRayOrigin( xRayOrigin )
		, m_fDistance( Asura_Maths::LargeFloat )
		{
			SetRayDirection(xRayDirection);
		}

				
		inline bool RayIntersectsAABB( const Asura_Bounding_Box& xAABB, float& fDistance ) const
		{
			bool bInside = true;

			Asura_Vector_3 xMinCorner(xAABB.MinX, xAABB.MinY, xAABB.MinZ);
			Asura_Vector_3 xMaxCorner(xAABB.MaxX, xAABB.MaxY, xAABB.MaxZ);

			Asura_Vector_3 xLambda;
			
			// KOR - 14/11/2005 - we were getting overflows when very small (denormalised) betas were passed in
			// This may cause false negatives when *just* skimming a BB, but hopefully an error of just 1e-30 should
			// make any problems negligible.
			const float fBetaEpsilon = 1.0e-30f;
			for(int i=0; i<3; i++)
			{
				if ((m_xRayOrigin)[i] < xMinCorner[i])
				{
					if ((m_xRayDirection)[i] <= fBetaEpsilon) return false;
					bInside = false;
				
					xLambda[i] = (xMinCorner[i] - (m_xRayOrigin)[i]) * m_xOneOverRayDirection[i];
				}
				else if ((m_xRayOrigin)[i] > xMaxCorner[i])
				{
					if ((m_xRayDirection)[i] >= -fBetaEpsilon) return false;
					bInside = false;
					
					xLambda[i] = (xMaxCorner[i] - (m_xRayOrigin)[i]) * m_xOneOverRayDirection[i];	
				}
				else
				{
					xLambda[i] = -1.0f;
				}
			}

			// Ray origin inside bounding box? Return as a hit.
			// (NOTE: also a hit if the ray starts on the surface of the BB.)
			if (bInside)
			{
				fDistance = 0.0f;

				return true;
			}

			// Find largest lambda value
			int iPlane = 0;
			if (xLambda[1] > xLambda[iPlane]) iPlane = 1;
			if (xLambda[2] > xLambda[iPlane]) iPlane = 2;

			// KJFL - in theory this check shouldn't be necessary, but someone did find an assert
			// firing in the original Asura_Maths code. So worth checking for now, to see if it
			// comes up.
			// KJFL - 14 May 2008 - this happened when the initial ray direction values were NANs. 
			// Worth keeping an assert around for if it happens again.
			if ( xLambda[iPlane] < 0.0f )
			{
				Asura_Assertf
				(
					false, 
					("Code bug: negative distance (%f) to a bounding box.\n Ray origin was (%f,%f,%f)\n - direction and recip were (%f,%f,%f) (%f,%f,%f).\n Box was Min(%f,%f,%f) Max(%f,%f,%f)",
					xLambda[iPlane],
					m_xRayOrigin[0],m_xRayOrigin[1],m_xRayOrigin[2],
					m_xRayDirection[0],m_xRayDirection[1],m_xRayDirection[2],
					m_xOneOverRayDirection[0],m_xOneOverRayDirection[1],m_xOneOverRayDirection[2],
					xMinCorner[0],xMinCorner[1],xMinCorner[2],
					xMaxCorner[0],xMaxCorner[1],xMaxCorner[2])
				);
				return false; 
			}

			// bigger distance than the one passed in?
			if ( xLambda[iPlane] > fDistance ) return false;

			for (int i=0; i<3; i++)
			{
				if (i != iPlane )
				{
					float fIntersection = m_xRayOrigin[i] + xLambda[iPlane] * m_xRayDirection[i];
					if (fIntersection < xMinCorner[i] || fIntersection > xMaxCorner[i]) return false;
				}
			}

			// We have an intersection then.
			fDistance    = xLambda[iPlane];

			return true;
		}


		const Asura_Vector_3& GetRayOrigin( void ) const { return m_xRayOrigin; }
		
		const Asura_Vector_3& GetRayDirection( void ) const { return m_xRayDirection; }
		const Asura_Vector_3& GetOneOverRayDirection( void ) const { return m_xOneOverRayDirection; }
	
		float GetDistance( void ) const { return m_fDistance; }

		void SetDistance( float fDistance )
		{
			Asura_Assert( m_fDistance >= 0.0f, "Code bug: Setting a negative distance" );
			m_fDistance = fDistance;
		}

		void SetRayOrigin( const Asura_Vector_3& xRayOrigin) {m_xRayOrigin = xRayOrigin;}
		void SetRayDirection( const Asura_Vector_3& xRayDirection)
		{
			m_xRayDirection = xRayDirection;

			const float fEPSILON = 1.0e-30f;
			m_xOneOverRayDirection.x = ( Asura_Maths::Abs( xRayDirection.x ) < fEPSILON ) ? 0.0f : ( 1.0f / xRayDirection.x );
			m_xOneOverRayDirection.y = ( Asura_Maths::Abs( xRayDirection.y ) < fEPSILON ) ? 0.0f : ( 1.0f / xRayDirection.y );
			m_xOneOverRayDirection.z = ( Asura_Maths::Abs( xRayDirection.z ) < fEPSILON ) ? 0.0f : ( 1.0f / xRayDirection.z );

		}

	private: 

		Asura_Vector_3			m_xRayOrigin;
		Asura_Vector_3			m_xRayDirection;

		float					m_fDistance;
		Asura_Vector_3			m_xOneOverRayDirection;


};

class Tools_Large_AABB_Tree
{
	public:
		Tools_Large_AABB_Tree();
		~Tools_Large_AABB_Tree();

		void Tools_Large_AABB_Tree::TestTree(const Asura_Bounding_Box& testBB, void(*outputCallback)(unsigned int index));


		void TestTreeWithRay(const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection, bool(*outputCallback)(unsigned int index, float fDistance), const float* pfMaxDistance);

		void TestTreeWithRay_ThreadSafe( Tools_Large_AABB_Tree_RayData* pxData, bool(*pfnCallback)(u_int uIndex, float fAABBDistance, Tools_Large_AABB_Tree_RayData* pxData) ) const;

		void TestTreeWithAABB_ThreadSafe( Tools_Large_AABB_Tree_AABBData* pxData, void(*pfnCallback)(u_int uIndex, Tools_Large_AABB_Tree_AABBData* pxData) ) const;

		u_int GetNumberOfObjects( void ) const { return m_uNumberOfObjects; }

	private:
		friend class Tools_Large_AABB_Tree_Creator;
		friend class Asura_Visibility_System;

		void Tools_Large_AABB_Tree::TestNode(int nodeIndex, Asura_Bounding_Box ParentBB) const;
		void Tools_Large_AABB_Tree::OutputAll(int nodeIndex) const;

		// Returns true iff the search should continue.
		bool Tools_Large_AABB_Tree::TestNodeWithRay(const int iNodeIndex, Asura_Bounding_Box ParentBB, const Asura_Vector_3& xRayOrigin, const Asura_Vector_3& xRayDirection);


		Tools_Large_AABB_Tree_Node* Nodes;
		Asura_Bounding_Box RootBoundingBox;

		/// Kevin Lea - 15/05/2002 17:27:06 - data used in testing
		/// I originally thought to make this static i.e. shared between
		/// all trees, but there are occasions where multiple tests
		/// of different trees (e.g. tree within tree) are needed.
		void(*OutputCallback)(unsigned int index);
		const Asura_Bounding_Box*  TestBB;

		bool(*OutputCallback2)(unsigned int index, float fDistance);
		const float* m_pfMaxDistance;

		u_int m_uNumberOfObjects;
};

class Tools_Large_AABB_Tree_Creator
{
	public: 
		static Tools_Large_AABB_Tree* Create(const int numberOfObjects, const Asura_Bounding_Box* objectAABBs, const Asura_Vector_3* objectCentres=0);
	protected:
		static void SortObjectsIntoNodes(int start, int end, Asura_Bounding_Box& parentBB);
			
		static unsigned int*				OldObjectSet;
		static unsigned int*				NewObjectSet;

		static const Asura_Vector_3*		LeafCentre;
		static const Asura_Bounding_Box*	LeafAABB;

		static Tools_Large_AABB_Tree*		Tree;
		static unsigned int					NodeIndex;
	
};


#endif//__TOOLS_LARGE_AABB_TREE_H__
