//
// Tools_Module_Element.cpp
//
// Created by Kevin Lea - 05/10/2001 16:52:54
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Rendering/Asura_Module.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Debugging/Asura_Assert.h"

#include "Misc/list_tem.hpp"
#include "Misc/Tools_Module_Element.h"
#include "Misc/Tools_Adjacency.h"

///////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------**
** Tools_Module_Elements **
**-----------------------*/

Tools_Module_Elements::Tools_Module_Elements(int module_index)
{
	ModuleIndex = module_index;
	//we don't bother calculating the elements until something actually
	//needs them
	NumPolygons = 0;
	ElementArray = NULL;
}

Tools_Module_Elements::~Tools_Module_Elements()
{
	delete [] ElementArray;
}


int Tools_Module_Elements::GetNextPolygonInElement(int poly_index)
{
	//calculate the element data if we haven't done so already
	if(!ElementArray) CalculateElements();

	Asura_Assert(poly_index>=0 && poly_index<NumPolygons,"polygon index is out of bounds");

	//look up the next polygon in the array and return it
	return ElementArray[poly_index];
}

void Tools_Module_Elements::Reset()
{
	delete [] ElementArray;
	ElementArray = 0;
	NumPolygons = 0;
}


#if 1
void Tools_Module_Elements::CalculateElements()
{
	Asura_Assert(!ElementArray,"Elements have already been calculated");

	//get the mesh data for the module
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(ModuleIndex);
	Asura_Assert(mesh,"Failed to get the mesh for the module");


	//allocate and initialise the ElementArray
	NumPolygons = mesh->NumberOfPolygons;
	ElementArray = new int[mesh->NumberOfPolygons];
	for(int i=0;i<mesh->NumberOfPolygons;i++)
	{
		ElementArray[i] = -1;
	}


	for(int i=0;i<mesh->NumberOfPolygons;i++)
	{
		//has this polygon been assigned to an element?
		if(ElementArray[i] !=-1 ) continue;

		//this polygon is the start of a new element
		//work out through the adjacencies from this polygon , adding polygons
		//to the element
		int current = i;//current member of the element being looked at
		int last_polygon = i;//the last polygon in the element
		do
		{
			const Tools_Polygon_Adjacency* pxPolyAdj = Tools_Adjacency::GetPolygonAdjacencies(ModuleIndex,current);
			for(const Tools_Adjacent_Polygon* pxAdj = pxPolyAdj->GetFirstAdjacency(); pxAdj; pxAdj = pxAdj->GetNext())
			{
				if(pxAdj->GetNumSharedVertices()<2)continue;//want a shared side
				if(pxAdj->IsBetweenModules())continue;//only want internal connections
				int polyIndex = pxAdj->GetPolygonIndex();

				if(ElementArray[polyIndex]!=-1) continue;
				if(polyIndex == last_polygon) continue;
				//link this polygon on to the end of the array
				ElementArray[last_polygon] = polyIndex;
				last_polygon = polyIndex;
			}
			current = ElementArray[current];
		}while(current!=-1);
		//done all the polygons in the element , link the last polygon back to the
		//first polygon
		ElementArray[last_polygon] = i;
		
	}
	
}
#else
void Tools_Module_Elements::CalculateElements()
{
	Asura_Assert(!ElementArray,"Elements have already been calculated");

	//get the mesh data for the module
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(ModuleIndex);
	Asura_Assert(mesh,"Failed to get the mesh for the module");

	//Step 1 - Work out the equivalent vertices
	int * vert_conv = new int[mesh->NumberOfVertices];
	
	int i;
	for(i=0;i<mesh->NumberOfVertices;i++)
	{
		//see if this vertex is the same as any earlier vertex
		int j;
		for(j=0;j<i;j++)
		{
			if(vert_conv[j]!=j) continue;

			//is vertex j the same as vertex i?
			Asura_Vector_3 diff = mesh->VertexDataPtr[i].Position - mesh->VertexDataPtr[j].Position;
			if(diff.x>=-0.001f && diff.x<=0.001f &&
			   diff.y>=-0.001f && diff.y<=0.001f &&	
			   diff.z>=-0.001f && diff.z<=0.001f)	
			{
				break;
			}
		}

		vert_conv[i] = j;
	}
	
	//Step 2 - Determine the polygons attached to each vertex
	List<int>* vert_poly_list = new List<int>[mesh->NumberOfVertices];

	for(i=0;i<mesh->NumberOfPolygons;i++)
	{
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[i];

		int num_verts = poly->NumberOfVertices();
		for(int j=0;j<num_verts;j++)
		{
			//add this polygon to the list for this vertex
			int vertex = vert_conv[poly->VertexID[j]];
			if(!vert_poly_list[vertex].contains(i))
			{
				vert_poly_list[vertex].add_entry(i);
			}
		}
	}

	//Step 3 - For each polygon determine all adjacent polygons

	List<int>* adjacency_list = new List<int>[mesh->NumberOfPolygons];
	for(i=0;i<mesh->NumberOfPolygons;i++)
	{
		//list of polygons that share a vertex with this one
		List<int> shared_vertex_list;

		//look at each vertex used by this polygon , and find the polygons using it
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[i];

		int num_verts = poly->NumberOfVertices();
		for(int j=0;j<num_verts;j++)
		{
			int vertex = vert_conv[poly->VertexID[j]];
			for(LIF<int> plif(&vert_poly_list[vertex]);!plif.done();plif.next())
			{
				if(plif() == i) continue;//don't want to count a polygon as adjacent to itself

				//add the polygon to the list if not already there
				if(!shared_vertex_list.contains(plif()))
				{
					shared_vertex_list.add_entry(plif());
					continue;
				}
				//the polygon was already in the list.. so it must share at least 2 vertices
				//so we can add it to the adjacency list
				if(!adjacency_list[i].contains(plif()))
				{
					adjacency_list[i].add_entry(plif());
				}
			}

		}
	}

	//finished with the vertex conversion array and vert_poly_list
	delete [] vert_conv;
	delete [] vert_poly_list;


	//allocate and initialise the ElementArray
	NumPolygons = mesh->NumberOfPolygons;
	ElementArray = new int[mesh->NumberOfPolygons];
	for(i=0;i<mesh->NumberOfPolygons;i++)
	{
		ElementArray[i] = -1;
	}

	//Step 4 - Use the adjacenecy list to determine the polygons in each element

	for(i=0;i<mesh->NumberOfPolygons;i++)
	{
		//has this polygon been assigned to an element?
		if(ElementArray[i] !=-1 ) continue;

		//this polygon is the start of a new element
		//work out through the adjacencies from this polygon , adding polygons
		//to the element
		int current = i;//current member of the element being looked at
		int last_polygon = i;//the last polygon in the element
		do
		{
			for(LIF<int> adj_lif(&adjacency_list[current]);!adj_lif.done();adj_lif.next())
			{
				if(ElementArray[adj_lif()]!=-1) continue;
				if(adj_lif() == last_polygon) continue;
				//link this polygon on to the end of the array
				ElementArray[last_polygon] = adj_lif();
				last_polygon = adj_lif();
			}
			current = ElementArray[current];
		}while(current!=-1);
		//done all the polygons in the element , link the last polygon back to the
		//first polygon
		ElementArray[last_polygon] = i;
		
	}
	delete [] adjacency_list;
	
}
#endif
