//
// Tools_Polygon_Selection.h
//
// Created by Kevin Lea - 19/07/2001 11:39:02
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_POLYGON_SELECTION_H__
#define __TOOLS_POLYGON_SELECTION_H__

#include "Tools_Undo.h"
#include "Core/Asura_Maths.h"
#include "Misc/list_tem.hpp"
///////////////////////////////////////////////////////////////////////////////////////
class Tools_Module_Elements;

#define TOOLS_MAX_FLOOR_NORMAL -0.71f

///////////////////////////////////////////////////////////////////////////////////////
/*
Class for managing selection of an index array of items.
Requires a fair amount of memory , but is done this way to make it fast
determining whether an item is selected , and finding a list of the currently
selected items
*/
class ItemSelectionList
{
public :
	ItemSelectionList(int numitems);
	virtual ~ItemSelectionList();

	//adds the index to the selected list
	void Add(int index);
	//removes from the selected list
	void Remove(int index);

	//start iteration through the selection
	void StartIteration()	 {iterationPos = 0;}
	//get the index of the current item in the selection iteration
	int Current();
	//has the iteration finished?
	bool Done() 		{return iterationPos >= numSelected;}
	//next position in the iteration
	void Next()				{iterationPos++;}

	//get the number of items selected
	int NumSelected(){return numSelected;}
	int NumItems(){return numItems;}
	
	bool IsSelected(int index);
	bool IsSelectable(){return Selectable;};

	//mark everything as selected
	void SelectAll();
	//mark everything as unselected
	void ClearList();

	int GetIndex(const int iIndex) const {return SelectedItems[iIndex];}


protected :
	unsigned int numItems; //total number of items that may be slected
	unsigned int numSelected; //number of items currently selected
	int* SelectedItems; //array containing indeces of selected items
	int* ItemIsSelected; //array indicating which items are selected
	bool Selectable; //is selection currently allowed?

	unsigned int iterationPos; //index of current position for iterating through selection

	virtual void ItemBeingSelected(int /*item*/){};
	virtual void ItemBeingDeselected(int /*item*/){};
};

///////////////////////////////////////////////////////////////////////////////////////
/*
For keeping track of the selected polygons in one module. Also logs selection/deselection for undo
*/
class PolygonSelectionList : public ItemSelectionList
{
public :
	PolygonSelectionList(int module,int numpoly);

private :
	int module_index;

	virtual void ItemBeingSelected(int index);
	virtual void ItemBeingDeselected(int index);
};

///////////////////////////////////////////////////////////////////////////////////////
struct PolygonSelectionUndoData
{
	enum UndoType
	{
		UndoType_Added,
		UndoType_Removed,
		UndoType_ForceInt = 0xffffffff,
	};
	
	enum UndoType undo_type;
	int module_index;
	int poly_index;
};

class PolygonSelectionUndo : public FixedSizeUndo
{
public :
	PolygonSelectionUndo()
	:FixedSizeUndo(10000,sizeof(PolygonSelectionUndoData)){}

	void ProcessUndoData(void*);
	
	void LogPolySelected(int module,int poly);
	void LogPolyDeselected(int module,int poly);
};
///////////////////////////////////////////////////////////////////////////////////////

//simple structure for identifying a given polygon from its module and polygon index
struct PolygonID
{
	PolygonID(){};
	PolygonID(int m,int p){ModuleIndex=m,PolyIndex=p;}

	bool operator == (const PolygonID& p){return p.ModuleIndex==ModuleIndex && p.PolyIndex==PolyIndex;}
	bool operator != (const PolygonID& p){return p.ModuleIndex!=ModuleIndex || p.PolyIndex!=PolyIndex;}
	
	int ModuleIndex;
	int PolyIndex;
};

///////////////////////////////////////////////////////////////////////////////////////

/*
Keeps track of all polygons selected in all modules
*/
enum PolygonSelectionMode
{
	SelectionMode_Polygon,
	SelectionMode_Object,
	SelectionMode_Element,
	SelectionMode_Plane,
	SelectionMode_Curve,
	SelectionMode_Floor,
	SelectionMode_UV,
	SelectionMode_Invert,
};

#define PolygonSelectionUpdate_Standard 0
#define PolygonSelectionUpdate_NoMouseSelection 1
#define PolygonSelectionUpdate_NoSelection 2

class Tools_Polygon_Selection
{
public : 
	static void Initialise();
	static void Shutdown();
	
	static void Activate();
	static void Deactivate();

	static void Update(int update_mode = PolygonSelectionUpdate_Standard);
	static void Render();
	static void RenderUI();

	static unsigned int CountSelectedPolygons();

//member variables
	static int NumberOfModules;
	static int NumSelectedModules;//number of modules with at least 1 poly selected

	static bool IsActive(){return s_bActive;}

	static bool ShowSelection; 
private :
	static bool SelectionLocked;
	static int SelectionMode;//value from enum PolygonSelectionMode
	static bool s_bLimitSelectionToSameMaterial;
	static bool s_bAllowBackfaceSelection;
	static bool s_bLimitSelectionToSameObject;

	static float s_fPlaneAngleTolerance;//tolerance level for plane selection (degrees)
	static float s_fCurveAngleTolerance;//tolerance level for curve selection (degrees)

	static bool s_bMaterialValid;

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	static Asura_Hash_ID s_uMaterialResponseHashID;
#else
	static int s_iMaterial;
#endif

	static Asura_Vector_2 SelectionOverlayPosition;
	static bool SelectionOverlayOpen;
	static u_int s_uSelectionOverlayMinimised;
	static PolygonSelectionList** Selection;
	static Tools_Module_Elements** ElementData;
	static PolygonSelectionUndo SelectionUndo;
	static int CurrentIterationModule;
	static void (*SelectionClearedCallback)();//function to be called when all polgons have been deselected

	//used to determine whether the mouse button was pressed to start polygon selection
	//(allows use of non-debounced button , excepet when it would be underiable)
	static bool ButtonDownForSelection;
	static bool s_bDragSelection;
	static Asura_Vector_2 s_xDragStart;
	static Asura_Vector_2 s_xDragCurrent;
	
	static PolygonID HighlitPolygon;//polygon mouse is currently over

	static bool s_bActive;

	// (SB - 12/12/06) Support for 'Select by polygon data' option
	static unsigned int s_uPolyData;

public :

	// (SB - 12/12/06) Support for 'Select by polygon data' option - public to expose to project Tools dialogs
	static bool	s_bLimitSelectionToSamePolygonData;
	// (SB - 29/01/07) Support for 'Require shared edge' option - public to expose to Flags mode
	static bool s_bRequireSharedEdge;

//selection/deselection functions
	static void SelectPolygon(int module_index,int polygon_index); 
	static void DeselectPolygon(int module_index,int polygon_index);
	static void SelectElement(int module_index,int polygon_index); 
	static void DeselectElement(int module_index,int polygon_index);
	static void SelectObject(int module_index); 
	static void DeselectObject(int module_index);
	static void SelectPlane(const PolygonID& polygon,bool bCurve);
	static void DeselectPlane(const PolygonID& polygon,bool bCurve);
	static void SelectFloor(const PolygonID& polygon);
	static void DeselectFloor(const PolygonID& polygon);
	static void SelectUVShell(const PolygonID& polygon);
	static void DeselectUVShell(const PolygonID& polygon);
	static void InvertSelection(int iModuleIndex);
	static void InvertAll();
	static void SelectAll();
	static void DeselectAll();
	static void SelectByName();
	static void GrowSelection();
	static void ShrinkSelection();

//setting callback function... used by texturing mode
	static void SetSelectionClearedCallback(void (*function)()){SelectionClearedCallback = function;}

//Functions for hiding/unhiding polygons
//These should possibly be put elsewhere , but they are going here for now , since they
//are intended for all modes that support polygon selection
	static void HideSelection();
	static void HideUnselected();
	static void Unhide();
	static void UnhideAndSelect();


//iterating through the selection
//(note this is a bit dubious since only 1 iteration can happen at a time... but I haven't needed
//to run several concurrent iterations yet)
	static void StartIteration();
	static bool GetNextSelected(PolygonID* polygon);
	static bool GetRandomSelectedPolygon(PolygonID *pxPolygon);

	static bool IsSelected(const PolygonID& polygon);
	static int CountSelectedInModule(int moduleIndex);
	static bool IsModuleCompletelySelected(int iModuleIndex);
	static void ListSelectedInModule(int iModuleIndex,List<int>& xOutputList);

	//Checks to see if a polygon should be allowed for selection (is hidden , etc)
	static bool IsSelectable(int iModuleIndex,int iPolygonIndex);

	//get the centre of the BB around the selected polygons
	//(shouldn't be used while iterating over selection)
	static Asura_Vector_3 GetCentreOfSelection();

	//gets the frame in which the selection last changed
	static u_int GetSelectionFrameStamp();


//selection overlay functions
	static void ToggleSelectionOverlay(void);
	static bool SelectionOverlayIsOn(void);
	static void CreateSelectionOverlay(void);
	static void CloseSelectionOverlay(void);
	
	static bool GetPolygonUnderMouse(PolygonID& polygon);
	static bool GetHighlitPolygon(PolygonID& polygon);

//handling changes that result in changes to the number of polygons
	static void HandlePolygonConversion(int module_index,int* PolyConv1,int* PolyConv2 = 0);
	static void DeleteModule(int iModuleIndex);
//Notifies polygon selection that the mesh has been replaced by another that may or may not be completely different
	static void HandleMeshChange(int iModuleIndex);

	static void ClearUndo(){SelectionUndo.FlushUndoBuffer();}

	//helper function for selection by plane
	static void ListPolygonsInPlane(const PolygonID& startPolygon,List<PolygonID>& outputList,bool bCurve,float fToleranceDeg);
	//and for selecting floor polygons
	static void ListConnectedFloorPolygons(const PolygonID& startPolygon,List<PolygonID>& outputList);
	//gets polygons for UV shell selection
	static void ListPolygonsInUVShell(int iChannel, const PolygonID& xStartPolygon,List<PolygonID>& xOutputList);

private :

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && !defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	static Asura_Hash_ID GetMaterialResponseHashID( const int iModule, const int iPolygon );
#else
	static int GetMaterialID(int iModule,int iPolygon);
#endif

	// (SB - 12/12/06) Support for 'Select by polygon data' option
	static unsigned int Tools_Polygon_Selection::GetPolyData(int iModule,int iPolygon);

//highlighting polygons
	static void HighlightSelection();//red outline
   	static void HighlightPolygon(PolygonID& polygon);//pulsing white outline

	static void Undo();	
	static void Redo();	

	//checks for keyboard shortcuts being pressed
	static void HandleKeyboard();
	//for handling general polygon selection using the mouse
	static void HandleMouseSelection();

	//loads/saves data for this mode
	static void TransferModeData(int mode);

	static void DragPolygonSelection(const Asura_Vector_2& xStart,const Asura_Vector_2& xEnd,bool bAllVertsInRect,bool bSelect);
	static void DragObjectSelection(const Asura_Vector_2& xStart,const Asura_Vector_2& xEnd,bool bAllVertsInRect,bool bSelect);


	friend class PolygonSelectionList;

};


#endif//__TOOLS_POLYGON_SELECTION_H__
