//
// Tools_Timed_Messages.cpp
//
// Created by Kevin Lea - 05/10/2001 16:51:41
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

// This file's header.
#include "Misc/Tools_Timed_Messages.h"

// Stdlib independent.
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

// Engine independent.
#include "Core/Asura_String.h"
#include "Core/Asura_Timers.h"
#include "Interface/Asura_Overlay_System.h"
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
#include "Interface/Asura_CommandConsole.h"
#include "Interface/Asura_CommandConsole_IO.h"
#endif
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Viewport.h"
// Tools.
#include "Misc/list_tem.hpp"
#include "Misc/Tools_MessageLog.h"
#include "Tools_System.h"

#define MAX_TIMED_MESSAGES 12
#define TIMED_MESSAGE_DURATION 20.0f


struct Timed_Message
{
	Timed_Message(const char* _Text,int _Colour);
	~Timed_Message();
	
	float TimeLeft;
	char* Text;
	int Colour;
};

static List<Timed_Message*> MessageList;

///////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Tools_Timed_Message_System **
**----------------------------*/


//declaration of static data
bool Tools_Timed_Message_System::TimerEnabled;
bool Tools_Timed_Message_System::s_bRedirectToConsole = false;
bool Tools_Timed_Message_System::s_bInitialised = false;
bool Tools_Timed_Message_System::s_bForceHideMessages = false;

void Tools_Timed_Message_System::Initialise()
{
	s_bInitialised = true;
	TimerEnabled = true;
}

void Tools_Timed_Message_System::Shutdown()
{
	//empty the list
	while(MessageList.size())
	{
		delete MessageList.first_entry();
		MessageList.delete_first_entry();
	}
	s_bInitialised = false;
}

void Tools_Timed_Message_System::Render()
{
	if (s_bForceHideMessages)
	{
		return;
	}

	//determine coordinates for bottom (most recent) message
	float x = (float)Asura_Viewport::GetWidth()/2.0f;
	float y = (float)Asura_Viewport::GetHeight();
	Asura_Text_Output::ConvertPixelsToPrintCoords(x,y);
	x-=0.5f;
	y-=3.0f;

	//may have to move the text up a bit if there are some overlays in a panel at the screen bottom
	if (Asura_Overlay_System::GetPanelMinY()-1.0f<y) y = Asura_Overlay_System::GetPanelMinY()-1.0f;
	 
	//iterate through the messages backwards , drawing them
	LIB<Timed_Message*> mlib(&MessageList);
	for(;!mlib.done();mlib.next())
	{
		//arbitrary alpha picking function...
		int alpha = (int)((mlib()->TimeLeft/TIMED_MESSAGE_DURATION)*350);
		if(alpha>255) alpha = 255;
		int colour = mlib()->Colour | (alpha<<24);

		Asura_Text_Output::Print(x,y--,colour,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,mlib()->Text);
	}

	//now update timers ,and clear out old messages
	if(TimerEnabled)
	{
		float time = Asura_Timers::GetFrameTime();
		for(mlib.restart();!mlib.done();)
		{
			mlib()->TimeLeft -= time;
			if(mlib()->TimeLeft<=0)	
			{
				delete mlib();
				mlib.delete_current();
			}
			else
			{
				mlib.next();
			}
		}
	}
}

void Tools_Timed_Message_System::AddMessage(int iColour, const char* szFormat, ...)
{
	va_list xArgs;
	va_start(xArgs, szFormat);
	AddMessageInternal(true,iColour,false,szFormat,xArgs);
	va_end(xArgs);
}

void Tools_Timed_Message_System::AddUnloggedMessage(int iColour, const char* szFormat, ...)
{
	va_list xArgs;
	va_start(xArgs, szFormat);
	AddMessageInternal(false,iColour,false,szFormat,xArgs);
	va_end(xArgs);
}

void Tools_Timed_Message_System::AddMessageRemovingDuplicate(int iColour, const char* szFormat, ...)
{
	va_list xArgs;
	va_start(xArgs, szFormat);
	AddMessageInternal(true,iColour,true,szFormat,xArgs);
	va_end(xArgs);
}


void Tools_Timed_Message_System::AddMessageInternal(bool bLog,int iColour,bool bRemoveDuplicate,const char* szFormat, va_list xArgs)
{
	if (!s_bInitialised && !s_bRedirectToConsole)
	{
		// (AW - 08.11.2004) - Occasionally tools code is called which generates timed messages
		// but we are not actually in tools (e.g. The PC env renderer calls tools code to check
		// validity of lighting, this code can generate tools messages). Ignore these to fix a mem-leak.
		return;
	}
	char szBuffer[255];

	// format string and place in buffer
    Asura_PrintfToBfr( szBuffer, szFormat, xArgs, sizeof(szBuffer), true );

	if(s_bInitialised)
	{
		if(bRemoveDuplicate)
		{
			for(LIF<Timed_Message*> xMessageLIF(&MessageList);!xMessageLIF.done();)
			{
				 if(!stricmp(xMessageLIF()->Text,szBuffer))
				 {
					//Delete the matching message that already exists.
					delete xMessageLIF();
				 	xMessageLIF.delete_current();
				 }
				 else
				 {
				 	xMessageLIF.next();
				 }
			}
		}
		
		//add message to the list
		MessageList.add_entry(new Timed_Message(szBuffer,iColour));

		//if we have too many messages delete one
		if(MessageList.size()>MAX_TIMED_MESSAGES)
		{
			if(TimerEnabled)
			{
				//Search for a suitable message to ditch.
				//Try to preserve messages that might be errors or warnings if they haven't been displayed for long
				int iMessageIndex = 0;
				for(LIF<Timed_Message*> xMessageLIF(&MessageList);!xMessageLIF.done();xMessageLIF.next(),iMessageIndex++)
				{
					if(iMessageIndex >= MAX_TIMED_MESSAGES - 4)
					{
						//Don't attempt to delete the most recent messages in this way
						//If we get this far , we can just delete the oldest message instead
						break;
					}
					
					Timed_Message* pxMessage = xMessageLIF();
					if(TIMED_MESSAGE_DURATION - pxMessage->TimeLeft < 5.0f)
					{
						//message has been on for less than 5 seconds. Preserve it if it appears to be red or yellow
						u_int uR = (pxMessage->Colour & 0xff0000) >> 16;
						u_int uG = (pxMessage->Colour & 0x00ff00) >> 8;
						u_int uB = (pxMessage->Colour & 0x0000ff);

						if(uR > 200 && uG<100 && uB<100) continue; //redish
						if(uR > 200 && uG>200 && uB<100) continue; //yellowish
					}
					//okay to delete this one
					delete pxMessage;
					xMessageLIF.delete_current();
					break;
				}
				//If no suitable candidate was found , the first message will be deleted below
			}
			
			
			if(MessageList.size()>MAX_TIMED_MESSAGES)
			{
				delete MessageList.first_entry();
				MessageList.delete_first_entry();
			}
		}



		//log message if required
		if(bLog)
		{
			Tools_MessageLog::AddMessage("Messages","","%s",szBuffer);
		}

	}
	else
	{
		//Not in tools , so display in the command console instead
		#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
		Asura_CommandConsole::Printf("%s",szBuffer);
		Asura_CommandConsole::TimedShowOutputLine(5.0f);
		#endif
	}

	
}

void Tools_Timed_Message_System::DisableTimer()
{
	TimerEnabled = false;
}
void Tools_Timed_Message_System::EnableTimer()
{
	TimerEnabled = true;
}


///////////////////////////////////////////////////////////////////////////////////
/*----------------**
** Timed_Message  **
**----------------*/

Timed_Message::Timed_Message(const char* _Text, int _Colour)
{
	Text = new char[strlen(_Text)+1];
	strcpy(Text,_Text);
	Colour = _Colour & 0xffffff;//clear alpha , since this is calcuated based on time
	TimeLeft = TIMED_MESSAGE_DURATION;
}

Timed_Message::~Timed_Message()
{
	delete [] Text;
}
