///////////////////////////////////////////////////////////////////////////////
//
// File: Tools_LineEdit_Object.h
//
// Created by Martin G.Harris (15/9/2006)
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//
// Tools object for editing a line by positioning the start and end points. 
// Currently these are used to set camera rails for the camera volume system,
// but they may have other applications.
//
///////////////////////////////////////////////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_LINE_EDIT_OBJECT_H__
#define __TOOLS_LINE_EDIT_OBJECT_H__

// Includes
#include "Objects/Tools_Object.h"
#include "Objects/Tools_PositionEdit_Object.h"

// Forward declarations
class Asura_Overlay;

///////////////////////////
// Tools_LineEdit_Object //
///////////////////////////
// Main line editing object.
class Tools_LineEdit_Object : public Tools_Object
{
	typedef Tools_Object PARENT;

public:
	Tools_LineEdit_Object(Asura_Vector_3* pxStartPos,
						  Asura_Vector_3* pxEndPos,
						  const char* pszName = 0,
						  const char* pszStartMessage = 0,
						  const char* pszEndMessage   = 0,
						  bool* pbModifiedFlag = 0);

	virtual ~Tools_LineEdit_Object();

	virtual void Render();
	virtual void Update();

	virtual void OnSelection();
	virtual void OnDeselection();
	virtual void OnReselection();

	virtual void PrintInfo(float& fX,float& fY);
	virtual bool ConfirmDestruction() { return false; }
	virtual bool IsTemporaryEditObject() const { return true; }

	void SetLineColour(const u_int uColour) { m_uLineColour = uColour; }

	// If something external has changed the position of one or both ends, then this
	// will make everything match up again.
	void RefreshPosition() { m_bUpdatedVectorDialog = true; }

	// Deletes the line edit object with this ID, if it still exists.
	static void Delete(const int iToolsID);

	// Find the line edit object with this ID, if it still exists.
	static Tools_LineEdit_Object* Find(const int iToolsID);

	// A sub-object has been destroyed, deal with it.
	static void SubObjectDestroyed(const int iParent, const u_int uDestroyed);

protected:
	virtual const char* GetName() { return m_xName.GetString(); }
	virtual const char* GetTypeDescription() { return "Line Edit"; }

	void CreateOverlay();
	void DeleteOverlay();

	virtual void AddToOverlay(Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset) {}

public:
	// MGH 21/11/2006: Made this public as it may be more generally useful.
	// Add a 3 component vector dialog to the overlay.
	static void AddVectorDialogToOverlay(Asura_Overlay* pxOverlay,
										 Asura_Vector_3* pxVector,
										 const char* szName,
										 Asura_Vector_2& xOffset,
										 float fMin, float fMax, float fDelta,
										 const char* szPrintFormat,
										 bool* pbChanged = 0);

protected:
	// Update the position of the tools object.
	static void UpdateToolsPositionEditObject(u_int uID, const Asura_Vector_3* pxPosition);

protected:
	u_int m_uStartObject;	// Start object ID
	u_int m_uEndObject;		// End object ID

	u_int m_uLineColour;	// Colour of line.
	
	Asura_Vector_3* m_pxStartPos;
	Asura_Vector_3* m_pxEndPos;

	Asura_String m_xName;
	bool* m_pbModifiedFlag;
	bool m_bUpdatedVectorDialog;

	// Static 
	static Tools_LineEdit_Object* s_pxThisLineEditObject;
	static Asura_Vector_2 s_xOverlayPosition;
};

////////////////////////////////////
// Tools_LineEdit_EndPoint_Object //
////////////////////////////////////
// Object used by the line editing object to mark the endpoints of the line.
class Tools_LineEdit_EndPoint_Object : public Tools_PositionEdit_Object
{
	typedef Tools_PositionEdit_Object PARENT;

private:
	Tools_LineEdit_EndPoint_Object( Asura_Vector_3* pxPosition,
									const char* pszName,
									const char* pszMessage,
									bool* pbModifiedFlag,
									int iParent,
									bool bStart)
	: Tools_PositionEdit_Object(pxPosition, 0, pszName, pszMessage, pbModifiedFlag)
	, m_iParent(iParent)
	, m_bStart(bStart)
	{}

public:
	virtual ~Tools_LineEdit_EndPoint_Object()
	{
		// Inform the parent that we have been deleted.
		Tools_LineEdit_Object::SubObjectDestroyed(m_iParent, GetID());
	}

	virtual const char* GetTypeDescription() { return "Line Edit point"; }
	
	virtual void PrintInfo(float& fX,float& fY);


protected:
	int m_iParent; // Parent tools ID
	bool m_bStart; // Start or end?

	// Friends
	friend class Tools_LineEdit_Object;
};

#endif //__TOOLS_LINE_EDIT_OBJECT_H__

