#include "Core/Asura.h"
#include "Objects/Tools_Object_System.h"

#include "Animation/Asura_Animation.h"
#include "Camera/Asura_Camera.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"    // LoadAsuraFile()
#include "Core/Asura_UserInput.h"
#include "Core/Asura_Los.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Debugging/Asura_Assert.h"
#include "Entity/Asura_Entity.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Dialog.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Direct3D.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Font_System.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Viewport.h"


#include "Tools_System.h"
#include "Tools_Mode.h"
#include "Entity/Tools_Message_Editing.h"
#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
#include "Entity/Tools_MessageEditor.h"
#endif
#include "Entity/Tools_Entity_Object.h"
#include "Entity/Tools_Entity_Data.h"
#include "Entity/Tools_SecondaryEntityFile_System.h"
#include "EntityTemplate/Tools_Entity_Template_Object.h"
#include "HelperObject/Tools_HelperObject_Object.h"
#include "Hierarchy/Tools_Mode_Hierarchy.h"
#include "Interface/Tools_ContextMenu_System.h"
#include "Lighting/Tools_LightSets.h"
#include "Lighting/Tools_Photon_Object.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "MeshEdit/Tools_Mode_MeshEdit.h"//for merging objects into environment
#include "Movement/Tools_Movement.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Misc/Tools_Help.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_LOS.h"
#ifdef ASURA_USE_DIALOGUE_SYSTEM
#include "Sound/Tools_Dialogue_System.h"
#endif

#include "Objects/Tools_Object.h"
#include "Objects/Tools_ObjectSystem_Param.h"
#include "Objects/Tools_Object_Gizmo.h"
#include "Misc/Tools_Prompted_Selection.h"
#include "Misc/Tools_Render_Options.h"
#include "Misc/Tools_Render.h"
#include "PC/FileSystem/Tools_FileSystem.h"

#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "ChunkSystem/Asura_Chunk_Entity.h"

#include <math.h>

///////////////////////////////////////////////////////////////////////////////////////
/*--------------**
** Help Strings **
**--------------*/
static const char* ObjectSystemHelpStrings [] =
{
	"LMB :           Select object (deselecting everything else)",
	"Ctrl+LMB :      Select object (add to current selection)",
	"Shift+LMB :     Deselect object",
	"LMB+Drag :      Select objects in rectangle (shift and control keys modify as above)",
	"RMB on object:  Select multiple related objects (deselecting everything else)",
	"Ctrl+RMB :      As above , but added to current selection",
	"Shift+RMB :     As above , but objects are deselected",
	"L :             Deselect all",
	"",
	"F :             Toggle flying",
	"C :             Attach camera to object (Only while flying)",
	"E :             Edit selected object",
	"V :             Snap to vertex/surface",
	"Delete :        Delete selected object",
	"Alt+F :         Find object",
	"Alt+G :         Goto selected object",
	"Alt+H :         Find object by guid",
	"Alt+< :         Select previous object in Find list",
	"Alt+> :         Select next object in Find list",
	"Alt+F12 :       Scale positions of selected objects",
	"Alt+V :         Set visible object types",
	"Ctrl+D :        Duplicate all selected objects (if they support copying)",
	"Ctrl+C :        Copy selected object to clipboard (if it supports copying)",
	"Ctrl+V :        Pastes the clipboard object at the current camera position",
	"Ctrl+Shift+V :  Pastes the properties of the clipboard object over the current selection",
	"Ctrl+T :        Create the selected object type (object mode only)",
	"",
	"Ctrl+Z :        Undo (Movement)",
	"Ctrl+Y :        Redo (Movement)",
	"< :             Undo (Movement)",
	"> :             Redo (Movement)",
	"",
	"Shift+Z :       Undo (Selection)",
	"Shift+Y :       Redo (Selection)",
	"Shift+< :       Undo (Selection)",
	"Shift+> :       Redo (Selection)",
	"",
	"                **Movement Gizmos**",
	"Shift+Q :       Toggle movement gizmos for selected object",
	"Shift+W :       Opens gizmo properties overlay",
	"Ctrl+Drag :     If control is held down while dragging a gizmo then movement will be snapped to discrete values.",
	NULL
};
//////////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------**
** Tools_ObjectTypeVisibility **
**----------------------------*/

//stores the visibility of all the object types. 
//Moved out of Tools_Object_System to allow it to persist after quitting to game
class Tools_ObjectTypeVisibility
{
public :
	Tools_ObjectTypeVisibility()
	{
		m_iTypeHiddenArraySize = 0;
		m_piTypeHiddenFlags = 0;
	}
	~Tools_ObjectTypeVisibility()
	{
		delete [] m_piTypeHiddenFlags;
	}

	bool TypeIsHidden(int iType) const
	{
		if(iType>=m_iTypeHiddenArraySize || iType<0) return false;
		return m_piTypeHiddenFlags[iType] ? true : false;
	}


	bool GetTypeHiddenFlag(int iType,int iFlags) const
	{
		if(iType>=m_iTypeHiddenArraySize || iType<0) return false;
		return (m_piTypeHiddenFlags[iType] & iFlags)? true : false;
	}

	void SetTypeHidden(int iType,int iFlags,bool bHide)
	{
		Asura_Assert(iType>=0,"");
		if(iType>=m_iTypeHiddenArraySize)
		{
			//need to enlarge the array
			Asura_Assert(iType<10000,"Unexpectedly large iType index");
		
			//allocate new array
			int iNewArraySize = Asura_Maths::Max(iType+1,static_cast<int>(TOT_NumCommonTypes));
			int* piNewTypeHiddenFlags = new int[iNewArraySize];

			//copy old values ,and initialise new entries
			memcpy(piNewTypeHiddenFlags,m_piTypeHiddenFlags,sizeof(int)*m_iTypeHiddenArraySize);

			memset(piNewTypeHiddenFlags+m_iTypeHiddenArraySize,0,(iNewArraySize-m_iTypeHiddenArraySize)*sizeof(int));
		
			//replace old array
			delete [] m_piTypeHiddenFlags;
			m_piTypeHiddenFlags = piNewTypeHiddenFlags;
			m_iTypeHiddenArraySize = iNewArraySize;
		}

		if(bHide)
		{
			m_piTypeHiddenFlags[iType] |= iFlags;
		}
		else
		{
			m_piTypeHiddenFlags[iType] &=~ iFlags;
		}
	}

	void RemoveHideFlagForAllTypes(int iFlags)
	{
		for(int i=0;i<m_iTypeHiddenArraySize;i++)
		{
			m_piTypeHiddenFlags[i] &=~ iFlags;
		}
	}


private :

	int m_iTypeHiddenArraySize;
	int* m_piTypeHiddenFlags;
};

static Tools_ObjectTypeVisibility g_xObjectTypeVisibility;


//////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Tools_Overlay_MergeObjectToModule **
**-----------------------------------*/

class Tools_Overlay_MergeObjectToModule : public Tools_BaseMessage_Overlay
{
public:
	Tools_Overlay_MergeObjectToModule(const char* szText);
	bool DoModal();

	bool ShouldMergeToNearest(){return s_iReturnValue==1;}

protected :
	virtual void InitialiseOverlay(class Asura_Overlay* pxOverlay,Asura_Vector_2& xOffset);

	static void Cancel(){s_iReturnValue = 0;}
	static void MergeToNearest(){s_iReturnValue = 1;}
	static void MergeToChosen(){s_iReturnValue = 2;}

	static int s_iReturnValue;

};
int Tools_Overlay_MergeObjectToModule::s_iReturnValue;

Tools_Overlay_MergeObjectToModule::Tools_Overlay_MergeObjectToModule(const char* szText)
:Tools_BaseMessage_Overlay(szText)
{
}

bool Tools_Overlay_MergeObjectToModule::DoModal()
{
	s_iReturnValue = 0;
	ShowOverlay();
	return s_iReturnValue != 0;
}

void Tools_Overlay_MergeObjectToModule::InitialiseOverlay(class Asura_Overlay* pxOverlay,Asura_Vector_2& xOffset)
{
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Function_Dialog("Merge to nearest module",xOffset, MergeToNearest, &pxOverlay->m_bShouldClose);
	pxDialog->SetHelpText("Merge each object to the module that is nearest to it");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=2.0f;

	pxDialog = new Asura_Function_Dialog("Merge to chosen module",xOffset, MergeToChosen, &pxOverlay->m_bShouldClose);
	pxDialog->SetHelpText("Prompts for a module to be selected. All objects will be merged to this module");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=2.0f;

	pxDialog = new Asura_Function_Dialog("Cancel",xOffset, Cancel, &pxOverlay->m_bShouldClose);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=2.0f;

}

//////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------**
** Tools_Object_System **
**---------------------*/

List<Tools_Object*> Tools_Object_System::ObjectList;
Tools_Object* Tools_Object_System::SelectedObject = 0;
List<Tools_Object*> Tools_Object_System::SelectedObjectList;
List<Tools_Object*> Tools_Object_System::s_xFindObjectList;
Tools_Object* Tools_Object_System::s_pxLastFoundObject = NULL;
List<Tools_Object*> Tools_Object_System::s_xDestroyObjectList;

List<Tools_Object*> Tools_Object_System::ObjectUnderCursorList;
bool Tools_Object_System::ControlsEnabled = false;
int Tools_Object_System::s_iHideAllFlags = 0;

bool Tools_Object_System::DoingSnapToVertex = false;
Asura_Vector_3 Tools_Object_System::SnapToVertexPositionStore;
Asura_Matrix_3x3 Tools_Object_System::SnapToVertexOrientationStore;
bool Tools_Object_System::s_bSnapToVertex_SourceVertexLocked;
bool Tools_Object_System::s_bSnapToVertex_PickingSourceVertex;
Asura_Vector_3 Tools_Object_System::s_xSnapToVertex_SourceVertex;
Asura_Vector_3 Tools_Object_System::s_xSnapToVertex_HighlightPosition;
int Tools_Object_System::s_iSnapToVertex_Rotate;
bool Tools_Object_System::s_bSnapToVertex_UseCollisionMesh = false;
Asura_Bounding_Box Tools_Object_System::s_xSnapToVertex_SourceBB;
int Tools_Object_System::s_iSnapToVertex_SetOrientState;
int Tools_Object_System::s_iSnapToVertex_DragStartX;
int Tools_Object_System::s_iSnapToVertex_DragStartY;
bool Tools_Object_System::s_bSnapToVertex_AlignMode_AlignYAxis = true;


bool Tools_Object_System::MessageBlockEditing = false;
Tools_Object* Tools_Object_System::MessageBlockObject = 0;

int		Tools_Object_System::s_iNumSelectionFilters	= 0;
int*	Tools_Object_System::s_piSelectionFilters	= NULL;

Tools_ChooseObject* Tools_Object_System::s_pxChooseObject = 0;
Asura_String Tools_Object_System::s_xChooseObjectMessage;

ObjectSystemUndo* Tools_Object_System::s_pxUndo = 0;
ObjectSelectionUndo* Tools_Object_System::s_pxSelectionUndo = 0;

Asura_Vector_2 Tools_Object_System::s_xTextPos;
float Tools_Object_System::s_fMaxTextY;


Tools_Object_System::SelectionState Tools_Object_System::s_xSelectionState;

Asura_List <Tools_Object> Tools_Object_System::s_xClipboardList;

enum ToolsCopyObjectType Tools_Object_System::s_eCopyType = TOOLS_COPYOBJECT_NONE;

bool Tools_Object_System::s_bShowComment = true;
Asura_UnicodeChar* Tools_Object_System::s_pszCommentText = 0;
void* Tools_Object_System::s_pCommentOwner = 0;

bool Tools_Object_System::s_bPreserveMessageEditing = false;

u_int Tools_Object_System::s_uVisiblePlatforms = 0xffffffff;
u_int Tools_Object_System::s_uVisibleLightSets = 0xffffffff;
bool Tools_Object_System::s_bHideByPlatformEnabled = false;
bool Tools_Object_System::s_bHideByLightSetEnabled = false;

Asura_Guid Tools_Object_System::s_uFindObjectGuid = 1000;

bool Tools_Object_System::s_bCreateNewObjectsInMissionFile = true;//Defaults to true , but only used if actually editing a mission file.

bool Tools_Object_System::s_bShowPhysicalObjectsInOtherModes = false;

u_int Tools_Object_System::s_uSelectionChangedStamp = 0;

Asura_TemplateHashTree<int> Tools_Object_System::s_xDistantIconTree;


void Tools_Object_System::Initialise()
{
	TransferModeData(TOOLSMODEDATAFILE_LOAD);

	//check to see if the tools shapes have been loaded yet by attempting to get one of them
	Asura_String test_mesh_name = "ToolsShape_Cutscene";
	if(!Asura_Mesh_System::GetMesh(test_mesh_name))
	{
		LoadAsuraFile("Objects\\Tools\\AsuraToolsShapes.asr");
	}
	SelectedObject = NULL;
	

	SetVisibleLightSets(s_uVisibleLightSets);
	SetVisiblePlatforms(s_uVisiblePlatforms);


	ClearSelectionFilter();


	s_xSelectionState.Reset();

	s_pxUndo = new ObjectSystemUndo;
	s_pxUndo->Initialise();

	s_pxSelectionUndo = new ObjectSelectionUndo;
	s_pxSelectionUndo->Initialise();

	ResetTextRenderPos();

	Tools_HelperObject_Object::CreateToolsObjects();
	Tools_Photon_Object::CreateToolsObjects();
}

void Tools_Object_System::Shutdown()
{
	TransferModeData(TOOLSMODEDATAFILE_SAVE);

	Tools_Object_System::DeleteAllModeSpecific();

	Tools_HelperObject_Object::DeleteToolsObjects();

	Tools_Photon_Object::SaveToolsObjects();
	Tools_Photon_Object::DeleteToolsObjects();

	if( Tools_Photon_Object::GeneralOverlayIsOpen() )
	{
		Tools_Photon_Object::ToggleGeneralOverlay();
	}


	ObjectHideByType_Overlay::Close();
	Tools_Gizmo_Handler::CloseOverlay();

// Delete the array of selection filters
	ClearSelectionFilter();

	//hopefully all the objects have been destroyed
	Asura_Assert(!ObjectList.size(),"Expecting all objects to be destroyed by now");
	SelectedObject = NULL;
	EnableAllControls(false);


	ClearCommentText(0);

	delete s_pxUndo;
	s_pxUndo = NULL;

	delete s_pxSelectionUndo;
	s_pxSelectionUndo = NULL;

	s_xDistantIconTree.Reset();

}

void Tools_Object_System::TransferModeData(int iMode)
{
	//open the xFile
	Tools_ModeDataFile xFile("ObjectSystem.dat",iMode);

	//tranfer the data
	//(none of these should be removed , and new data should go on the end, otherwise
	//the xFiles will get screwed up)
	if( false == xFile.GetError())
	{
		xFile.Transfer(Tools_HelperObject_Object::s_xOverlayPosition);
		xFile.Transfer(Tools_Gizmo_Handler::s_xOverlayPosition);
		xFile.Transfer(Tools_Gizmo_Handler::s_fSnapMoveGranularity);
		xFile.Transfer(Tools_Gizmo_Handler::s_fSnapRotateGranularity);
		xFile.Transfer(Tools_Gizmo_Handler::s_fGizmoScale);
		xFile.Transfer(s_bSnapToVertex_AlignMode_AlignYAxis);
	}
}

void Tools_Object_System::Update()
{
	//Create the tools objects for physical entities if needed
	if(ShouldShowPhysicalObjectsInOtherModes() && !Tools_Entity_Object::HasCreatedPhysicalObjects())
	{
		Tools_Entity_Object::CreatePhysicalObjects();
	}

	UpdateObjectUnderCursorList();
	if(ControlsEnabled)
	{
		//Flag the undo to start a new undo batch
		s_pxUndo->MarkUndoStartIfEndOfBatch();
		s_pxSelectionUndo->MarkUndoStartIfEndOfBatch();

		//check for any movement to be logged for undo 
		if(!DoingSnapToVertex)
		{
			for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
			{
				xObjectLIF()->UpdateMovementUndo();
			}
		}
		
		//handle object selection
		if (!Asura_Interface::UsingMenusOrOverlays() && !Tools_Prompted_Selection::IsSelecting())
		{
			if(DoingSnapToVertex)
			{
				HandleSnapToVertex();
			}
			else
			{
				bool bMessageSelection = false;
				if(IsMessageEditing() && !s_xSelectionState.m_uSelectionFlags)
				{
					//see if selecting a message
					bMessageSelection = Tools_Message_Editing::HandleMouseSelection();
				}
				if(!bMessageSelection)
				{
					HandleMouseSelection();
				}
			}
		}

		//per frame maintenance of template overlay added here , since it may be needed in several object modes
		if(Tools_Entity_TemplateInstance_Object::TemplateOverlayIsOn())
		{
			Tools_Entity_TemplateInstance_Object::UpdateTemplateOverlay();
		}

	}
	//deal with shortcut keys
	HandleKeyboard();

	if(!IsChoosingObject() && !IsMessageEditing())
	{
		if(SelectedObject) SelectedObject->UpdateWhileSelected();
	}

	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		xObjectLIF()->Update();
	}

	if(SelectedObject && SelectedObject->GetAnimation() && SelectedObject->GetAnimation()->GetSkin())
	{
		//Store the skin being used by the selected object ,so that it is the default when entering hierarchy mode
		Tools_Mode_Hierarchy::SetDefaultSkin(SelectedObject->GetAnimation()->GetSkin());
	}

#ifndef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	if(IsMessageEditing())
	{
		Tools_Message_Editing::Update();
	}
#else
	Tools_Message_Editing::Update();
#endif

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::Update();
	#endif

	// Clean up any objects set to be destroyed (without asking for user confirmation).
	if ( s_xDestroyObjectList.size() )
	{
		DestroyObjects( s_xDestroyObjectList, false );
		s_xDestroyObjectList.empty();
	}

	if(s_pxChooseObject)
	{
		s_pxChooseObject->HideOverlays(true);
	}
}

void Tools_Object_System::RenderPreFSFX()
{
	SCOPEDPIXEVENT(( 0xFF837B4A, L"Tools_Object_System::EarlyRender()"));

	//render all the objects in the list
	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		Tools_Object* pxObject = olif();
		//make sure objects are scaled correctly first
		if(pxObject->IsScaledWhenClose()) pxObject->UpdateScale();

		if(!pxObject->GetShouldBeRendered(true)) continue;
		if(pxObject->ShouldRenderAsUI()) continue;

		pxObject->RenderPreFSFX();
	}
	Tools_Render::RenderBatchedIcons();
}



void Tools_Object_System::Render()
{
	SCOPEDPIXEVENT(( 0xFF837B4A, L"Tools_Object_System::Render()"));
	
	//render all the objects in the list
	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		Tools_Object* pxObject = olif();
		
		if(!pxObject->GetShouldBeRendered(false)) continue;
		if(pxObject->ShouldRenderAsUI()) continue;

		pxObject->Render();
		
	}
	Tools_Render::RenderBatchedIcons();

	if(MessageBlockObject)
	{
		SCOPEDPIXEVENT(( 0xFF000000, L"Message block object links" ));
		MessageBlockObject->RenderMessageBlockLinks();
	}
	else if(SelectedObject)
	{
		SCOPEDPIXEVENT(( 0xFF000000, L"Selected object message links" ));
		SelectedObject->RenderMessageBlockLinks();
	}

}

void Tools_Object_System::RenderAsUI()
{
	SCOPEDPIXEVENT(( 0xFF837B4A, L"Tools_Object_System::RenderAsUI()"));
	
	//Render objects that are always on top.
	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		Tools_Object* pxObject = olif();

		if(!pxObject->GetShouldBeRendered(false)) continue;
		if(!pxObject->ShouldRenderAsUI()) continue;

		pxObject->Render();
	}
	Tools_Render::RenderBatchedIcons();

	s_fMaxTextY = 0.0f;

	
	if(ControlsEnabled)
	{
		if(s_xSelectionState.m_bDragSelect)
		{
			//determine the colour for the lines
			float fBright = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
			int iBright = Asura_Maths::FloatToInt(255.0f*fBright*fBright);
			unsigned int uColour = 0xff000000 + (iBright) + (iBright<<8) + (iBright<<16);

			//draw drag rectangle

			Asura_Vector_2 xStart,xEnd;
			xEnd = s_xSelectionState.m_xDragStart;
		
			xStart = xEnd;
			xEnd.x = s_xSelectionState.m_xDragCurrent.x;
			Asura_Render::Render2DLine(xStart,xEnd,uColour);

			xStart = xEnd;
			xEnd.y = s_xSelectionState.m_xDragCurrent.y;
			Asura_Render::Render2DLine(xStart,xEnd,uColour);

			xStart = xEnd;
			xEnd.x = s_xSelectionState.m_xDragStart.x;
			Asura_Render::Render2DLine(xStart,xEnd,uColour);

			xStart = xEnd;
			xEnd.y = s_xSelectionState.m_xDragStart.y;
			Asura_Render::Render2DLine(xStart,xEnd,uColour);
		}
		
		if(s_pxChooseObject)
		{
			float x = s_xTextPos.x;
			float y = s_xTextPos.y;
			Asura_Text_Output::Print(x,y++,0xffffffff,0,"%s",s_xChooseObjectMessage.GetString());
			Asura_Text_Output::Print(x,y++,0xffffffff,0,"Space : Accept");
			Asura_Text_Output::Print(x,y++,0xffffffff,0,"Escape : Cancel");
			y++;

			if(s_pxChooseObject->SupportMultipleSelection())
			{
				Asura_Text_Output::Print(x,y++,0xffffffff,0,"Objects selected : %d",s_pxChooseObject->GetSelectionList()->size());
			}
			Tools_Object* pxSelectedObject = s_pxChooseObject->GetMainSelection();
			if(pxSelectedObject)
			{
 				pxSelectedObject->Tools_Object::PrintInfo(x,y);//just display basic information
			}
			s_fMaxTextY = y;
		}
		else
		{
			float x = s_xTextPos.x;
			float y = s_xTextPos.y;
			if(!DoingSnapToVertex && !MessageBlockEditing)
			{
				Asura_Text_Output::Print(x,y++,0xffffffff,0,"Objects selected : %d",SelectedObjectList.size());

				//Show dialogue subtitles
				//(doesn't really belong here , but it needs to be able to cooperate with the object system text)
				#ifdef ASURA_USE_DIALOGUE_SYSTEM
				Tools_Dialogue_System::RenderSubtitle(x,y);
				#endif
			}


			if(SelectedObject)
			{
				if(DoingSnapToVertex)
				{
					//highlight the current snap position
					u_int uColour = s_bSnapToVertex_PickingSourceVertex ?  0x7fffffff : 0x7f00ff00;
					Asura_Bounding_Box xBox;

					float fSize = 0.1f;
					xBox.MinX = s_xSnapToVertex_HighlightPosition.x - fSize;
					xBox.MinY = s_xSnapToVertex_HighlightPosition.y - fSize;
					xBox.MinZ = s_xSnapToVertex_HighlightPosition.z - fSize;
					xBox.MaxX = s_xSnapToVertex_HighlightPosition.x + fSize;
					xBox.MaxY = s_xSnapToVertex_HighlightPosition.y + fSize;
					xBox.MaxZ = s_xSnapToVertex_HighlightPosition.z + fSize;

					Tools_Render::RenderCube(xBox,uColour,false);

					
					if(s_bSnapToVertex_PickingSourceVertex)
					{
						Asura_Text_Output::Print(x,y++,0xff00ff00,0,"Snap To Vertex/Surface - Picking Source Vertex :");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"LMB : Accept current source vertex");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Release V to cancel");
					}
					else if(s_iSnapToVertex_SetOrientState)
					{
						Asura_Text_Output::Print(x,y++,0xff00ff00,0,"Snap To Vertex/Surface - Set Facing");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Release LMB : Accept facing");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold L Alt : Face towards object");
						y++;
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Snap to collision mesh (C) : %s",s_bSnapToVertex_UseCollisionMesh ? "Yes" : "No");
					}
					else
					{
						Asura_Text_Output::Print(x,y++,0xff00ff00,0,"Snap To Vertex/Surface :");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"LMB : Accept position");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Escape : Cancel");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold Shift : Align to surface");
						if(!s_bSnapToVertex_SourceVertexLocked)
						{
							Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold L Control : Snap vertex to vertex");
							Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold R Control : Snap pivot to vertex");
							Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold R Alt : Snap pivot to cursor");
						}
						else
						{
							Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold Control : Snap vertex to vertex");
						}
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold L Alt : Snap to object");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold V : Select source vertex for snapping");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Hold Z : Keep initial pos/orient");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"R : Rotate by 90 degrees");
						y++;
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Snap to collision mesh (C) : %s",s_bSnapToVertex_UseCollisionMesh ? "Yes" : "No");
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Align mode (I) : %s",s_bSnapToVertex_AlignMode_AlignYAxis ? "Align Y Axis" : "Align BB");
					}
				}
				else
				{
					//display information about the selected object
					if(MessageBlockEditing)
						SelectedObject->Tools_Object::PrintInfo(x,y);//just display basic information
					else
						SelectedObject->PrintInfo(x,y);
				}
			}
			s_fMaxTextY = y;
		}

		if(MessageBlockEditing)
		{
			float x = s_xTextPos.x;
			float y = s_xTextPos.y+3.0f;
			Tools_Message_Editing::Render(x,y);

			s_fMaxTextY = y;
		}
		
		//show the name of the objects under the cursor
		if (!Asura_Interface::UsingMenusOrOverlays() && !Tools_ContextMenu::IsMenuActive())
		{
			float x,y;
			Asura_Cursor::GetScreenCoords(x,y);
			x++;
			y--;
			for(LIF<Tools_Object*> olif(&ObjectUnderCursorList);!olif.done();olif.next())
			{
				if(olif()->GetName())
				{
					Asura_Text_Output::Print(x,y++,0xff00ff00,0,"%s",olif()->GetName());

					if(ObjectUnderCursorList.size()==1 && olif()->IsEntityObject() && Tools_SecondaryEntityFile_System::AnySecondaryFiles())
					{
						Tools_Entity_Object* pxEntity = (Tools_Entity_Object*) olif();
						Asura_Text_Output::Print(x,y++,0xff00ff00,0,"File : %s",pxEntity->GetEntityFileName());

					}
				}
			}
		}
	}

	if(s_bShowComment)
	{
		RenderComment();
	}

	// JAT 27/11/06 - flush the display to get any lines displayed correctly
	Asura_Render::FlushDisplay();
}


Tools_Object* Tools_Object_System::FindObjectInLOS(const Asura_Vector_3& alpha, const Asura_Vector_3& beta)
{
	float fLowDistanceWorld = ASURA_LOS_MAX_DISTANCE;
	float fLowDistanceUI = ASURA_LOS_MAX_DISTANCE;
	Tools_Object* pxClosestObjectWorld = 0;
	Tools_Object* pxClosestObjectUI = 0;

	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		if(!olif()->ShouldTestForLOS()) continue;
		
		if(olif()->ShouldRenderAsUI())
		{
			if(olif()->ObjectIsInLOS(alpha,beta,fLowDistanceUI))
			{
				pxClosestObjectUI = olif();
			}
		}
		else
		{
			if(olif()->ObjectIsInLOS(alpha,beta,fLowDistanceWorld))
			{
				pxClosestObjectWorld = olif();
			}
		}
	}

	//Prefer objects that are being drawn on top as UI over those in the world.
	return pxClosestObjectUI ? pxClosestObjectUI : pxClosestObjectWorld;
}

Tools_Object* Tools_Object_System::FindObjectUnderCursor(Tools_Object* pxObjectToIgnore /*= 0*/)
{
	const Asura_Vector_3* cameraPosition = Asura_Camera::GetPosition();
	const Asura_Matrix_3x3* cameraOrientation = Asura_Camera::GetOrientation();


	//get the z direction for the mouse
	Asura_Vector_3 zvec;
	Asura_Cursor::GetLOSDirection(zvec);

	// Kevin Lea - 22/06/2001 12:49:20 - and transform this direction into worldspace
	zvec = zvec * (*cameraOrientation);

	if(pxObjectToIgnore)
	{
		//hide the object so it doesn't get picked up by the LOS test
		pxObjectToIgnore->HiddenFlags |= ObjectHideFlag_ObjectSystem;
	}

	Tools_Object* pxObject = FindObjectInLOS(*cameraPosition,zvec);

	if(pxObjectToIgnore)
	{
		pxObjectToIgnore->HiddenFlags &=~ ObjectHideFlag_ObjectSystem;
	}

	return pxObject;
}

Tools_Object* Tools_Object_System::FindObjectUnderCursor(const List<Tools_Object*>& xObjectsToIgnore)
{
	const Asura_Vector_3* cameraPosition = Asura_Camera::GetPosition();
	const Asura_Matrix_3x3* cameraOrientation = Asura_Camera::GetOrientation();


	//get the z direction for the mouse
	Asura_Vector_3 zvec;
	Asura_Cursor::GetLOSDirection(zvec);

	// Kevin Lea - 22/06/2001 12:49:20 - and transform this direction into worldspace
	zvec = zvec * (*cameraOrientation);

	for(CLIF<Tools_Object*> xObjectLIF(&xObjectsToIgnore);!xObjectLIF.done();xObjectLIF.next())
	{
		//hide the object so it doesn't get picked up by the LOS test
		xObjectLIF()->HiddenFlags |= ObjectHideFlag_ObjectSystem;
	}

	Tools_Object* pxObject = FindObjectInLOS(*cameraPosition,zvec);

	for(CLIF<Tools_Object*> xObjectLIF(&xObjectsToIgnore);!xObjectLIF.done();xObjectLIF.next())
	{
		xObjectLIF()->HiddenFlags &=~ ObjectHideFlag_ObjectSystem;
	}

	return pxObject;
}


Tools_Object* Tools_Object_System::GetObjectFromGuid(Asura_Guid guid,bool bIncludeSubEntities /*=false*/)
{
	if(guid==ASURA_GUID_UNREGISTERED) return 0;
	

	Asura_Entity_Server* entity = Asura_Entity_Server::FindEntity(guid);
	if(entity && entity->ToolsData && entity->ToolsData->ToolsObject)
	{
		return entity->ToolsData->ToolsObject;
	}

	//failed to find in the asura entity system.. check the tools objects directly
	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		if(olif()->GetGuid()==guid) return olif();

		if(bIncludeSubEntities)
		{
			if(olif()->IsSubEntity(guid)) return olif();
		}
	}

	return 0;
}

Tools_Object* Tools_Object_System::GetObjectFromToolsID(int iID)
{
	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		if(xObjectLIF()->GetToolsID() == iID)
		{
			return xObjectLIF();
		}
	}
	return 0;
}

Tools_Object* Tools_Object_System::GetOwnerOfMessageBlock(unsigned short usBlockID)
{
	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		if(xObjectLIF()->OwnsThisMessageBlock(usBlockID)) return xObjectLIF();

	}

	return 0;
}

void Tools_Object_System::CheckForMessageBlockOwnershipProblems(bool bDeleteOrphans)
{
	static List<u_short> s_xProblemMessageBlocks;

	// Statistics for the orphaned message blocks.
	static int s_iSavedNumberOfOrphans = -1;
	int iNumOrphans = 0;
	int iNumOrphanMessages = 0;
	int iMemoryWasteBytes = 0;

	//The checks for orphaned messages won't be valid unless objects have been created for all the entities.
	//Which means it won't be safe to attempt to delete them
	//RAR 01/12/2008
	if(bDeleteOrphans && !Tools_Entity_Object::HasCreatedAllObjects())
	{
		ShowMessageOverlay("Can't delete orphaned messages since tools objects haven't been created for all entities etc.");
		bDeleteOrphans = false;
	}

	if(bDeleteOrphans)
	{
		//Confirm deletion
		const char* szMessage = "This will delete any message blocks that no entity claims ownership of , and are probably just bugged and wasting memory.\n"
								"For new instances of this error check with a programmer first to confirm that they are indeed being orphaned rather than mis-reported , and so "
								"that the bug can be fixed\n\n"
								"Go ahead and delete any orphaned messages?";
		if(!ShowYesNoOverlay(szMessage)) return;
	}


	//Prepare all objects for the checks. (Allow cutscenes to refresh their message block lists)
	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		xObjectLIF()->PreMessageBlockOwnershipCheck();
	}

	for(int iSet = 0;iSet<ASURA_STATICMESSAGE_NUMSETS;iSet++)
	{
		Asura_StaticMessageSetIndex eSet=(Asura_StaticMessageSetIndex)iSet;
		int iNumberOfMessageBlocks = Asura_StaticMessage_System::GetNumberOfMessageBlocks(eSet);

		for(int iBlockIndex = 0;iBlockIndex<iNumberOfMessageBlocks;iBlockIndex++)
		{
			unsigned short usBlockID = Asura_StaticMessage_System::ConvertIndexToID(eSet,iBlockIndex);
			int iNumberOfMessages = 0;
			bool bMessageBlockValid = Asura_StaticMessage_System::IsMessageBlockValid(usBlockID);
			if(bMessageBlockValid)
			{
				iNumberOfMessages = Asura_StaticMessage_System::GetNumberOfMessagesInBlock(usBlockID);
			}

			//find objects owning the message block
			Tools_Object* axFoundObjects[2];
			int iNumObjectsFound = 0;
			for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
			{
				if(xObjectLIF()->OwnsThisMessageBlock(usBlockID))
				{
					axFoundObjects[iNumObjectsFound++] = xObjectLIF();
					if(iNumObjectsFound==2) break;

				}
			}

			if(iNumObjectsFound == 2)
			{
				if(!s_xProblemMessageBlocks.contains(usBlockID))
				{
					s_xProblemMessageBlocks.add_entry(usBlockID);
					//two objects sharing the same message block
					ShowMessageOverlay("Error.\n"
										"More than one object owns message block %d\n."
										"The offending objects are :\n%s\n%s\n\n"
										"Please make sure a programmer knows about these errors , so the cause can be found - They can cause major problems for levels if left unchecked.\n"
										"(See Richard Rice for more details)",
										(int)usBlockID,axFoundObjects[0]->GetName(),axFoundObjects[1]->GetName());
				}
			}
			else if(iNumObjectsFound == 1 && !bMessageBlockValid)
			{
				if(!s_xProblemMessageBlocks.contains(usBlockID))
				{
					s_xProblemMessageBlocks.add_entry(usBlockID);
					ShowMessageOverlay("Error.\n"
										"%s owns message block %d which is invalid.This may cause multiple objects to share the same message block (which is bad).\n\n"
										"Please make sure a programmer knows about these errors , so the cause can be found - They can cause major problems for levels if left unchecked.\n"
										"(See Richard Rice for more details)",
										axFoundObjects[0]->GetName(),(int)usBlockID);
				}
			}
			else
			{
				if (iNumObjectsFound == 0 && bMessageBlockValid)
				{
					// MGH 15/2/2008: Found an orphaned block, collect some statistics.
					++iNumOrphans;
					iNumOrphanMessages += iNumberOfMessages;
					iMemoryWasteBytes += sizeof(Asura_StaticMessage*) + sizeof(u_int);

					for (int iMessage = 0; iMessage<iNumberOfMessages; ++iMessage)
					{
						const Asura_StaticMessage* pxStaticMessage = Asura_StaticMessage_System::GetMessagePointer(eSet, iBlockIndex, iMessage);
						if (pxStaticMessage)
						{
							iMemoryWasteBytes += sizeof(Asura_StaticMessage) + pxStaticMessage->m_usDataSize;
						}
					}
					if(bDeleteOrphans)
					{
						Asura_StaticMessage_System::DeleteMessageBlock(usBlockID);
						switch(iSet)
						{
							case ASURA_STATICMESSAGE_ENTITYSET :
								Tools_SaveAsura::SetEntityModified();
								break;
							case ASURA_STATICMESSAGE_CUTSCENESET :
								Tools_SaveAsura::SetCutscenesModified();
								break;

							default :
								Asura_Assert(0,"Unknown message set");
								break;
			
						}

					}
				}

				if(s_xProblemMessageBlocks.contains(usBlockID))
				{
					s_xProblemMessageBlocks.delete_entry(usBlockID);
				}
			}
		}
	}

	//The checks for orphaned messages won't be valid unless objects have been created for all the entities.
	//RAR 30/06/2008
	if(Tools_Entity_Object::HasCreatedAllObjects())
	{
		if(bDeleteOrphans)
		{
			ShowMessageOverlay("Found %d orphaned static message blocks (with %d messages, wasting approx %d bytes).",iNumOrphans,iNumOrphanMessages,iMemoryWasteBytes);
			s_iSavedNumberOfOrphans = 0;
		}
		else
		{
			if (iNumOrphans != s_iSavedNumberOfOrphans && iNumOrphans>0)
			{
				//Disabled warning when doing template editing , since the 'export template' option will
				//result in orphaned messages in the template file.
				//In this case it is harmless though ,since only messages that are actually used will get
				//copied into a level when updating from a template.
				//RAR 27/02/2008
				if(!Tools_SaveAsura::IsTemplateEditing())
				{
					// MGH 15/2/2008: Number of orphans has changed.
					Tools_MessageLog::AddMessage( "Entities", "Static Messages",
																  "Found %d orphaned static message blocks (with %d messages, wasting approx %d bytes).",
																  iNumOrphans,
																  iNumOrphanMessages,
																  iMemoryWasteBytes);

					Tools_MessageLog::ShowOverlay();

					if (s_iSavedNumberOfOrphans != -1 && iNumOrphans>s_iSavedNumberOfOrphans)
					{
						// MGH 15/2/2008: Assert if the number of orphans has increased since the last check.
						// Could be caused by a missing or incorrect implementation of OwnsThisMessageBlock(), or we might
						// be leaking static message blocks due to missing clean up code in a Destroy() function.
						Asura_Assert(0, "Number of static message block orphans has increased!");
					}
				}
			}

			s_iSavedNumberOfOrphans = iNumOrphans;
		}
	}
}

void Tools_Object_System::UpdateObjectUnderCursorList()
{
	while(ObjectUnderCursorList.size()) ObjectUnderCursorList.delete_first_entry();
	
	Asura_Vector_3 alpha = *Asura_Camera::GetPosition();
	const Asura_Matrix_3x3* cameraOrientation = Asura_Camera::GetOrientation();

	//get the z direction for the mouse
	Asura_Vector_3 beta;
	Asura_Cursor::GetLOSDirection(beta);

	// Kevin Lea - 22/06/2001 12:49:20 - and transform this direction into worldspace
	beta = beta * (*cameraOrientation);

	//first find the closest object under the cursor
	float fLowDistanceWorld = ASURA_LOS_MAX_DISTANCE;
	float fLowDistanceUI = ASURA_LOS_MAX_DISTANCE;
	Tools_Object* pxClosestObjectWorld = 0;
	Tools_Object* pxClosestObjectUI = 0;

	const int iObjectInLOSArraySize = 128;
	Tools_Object* apxObjectInLOS[128];
	float afObjectDistance[128];
	int iNumObjectInLOS = 0;

	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		Tools_Object* pxObject = olif();

		if(!pxObject->ShouldTestForLOS()) continue;

		if(pxObject->ShouldRenderAsUI())
		{
			if(pxObject->ObjectIsInLOS(alpha,beta,fLowDistanceUI))
			{
				pxClosestObjectUI = pxObject;
			}
		}
		else
		{
			//check to see if this object is within half metre of the closest object so far
			float fDistanceCopy = fLowDistanceWorld + 0.5f;
			if(pxObject->ObjectIsInLOS(alpha,beta,fDistanceCopy))
			{
				if(iNumObjectInLOS < iObjectInLOSArraySize)
				{
					//store the result for this object
					afObjectDistance[iNumObjectInLOS] = fDistanceCopy;
					apxObjectInLOS[iNumObjectInLOS++] = pxObject;
				}
				else
				{
					Asura_Assert(0,"Unexpectedly large number of objects in LOS under cursor.");
				}
				
				if(fDistanceCopy < fLowDistanceWorld)
				{
					pxClosestObjectWorld = pxObject;
					fLowDistanceWorld = fDistanceCopy;
				}
			}
		}
	}
	if(!pxClosestObjectWorld && !pxClosestObjectUI) return;

	if(pxClosestObjectUI)
	{
		//If over a UI object , don't bother adding other objects that are under the cursor
		pxClosestObjectUI = pxClosestObjectUI->GetObjectToSelect();
		ObjectUnderCursorList.add_entry(pxClosestObjectUI);
	}
	else
	{
		pxClosestObjectWorld = pxClosestObjectWorld->GetObjectToSelect();
		ObjectUnderCursorList.add_entry(pxClosestObjectWorld);

		//add all objects near to that in LOS
		//go through the array of objects found on the first pass , and get all those that were within
		//half a metre of the closest distance.
		for(int i=0; i<iNumObjectInLOS; i++)
		{
			if(afObjectDistance[i] < fLowDistanceWorld + 0.5f)
			{
				Tools_Object* pxObject = apxObjectInLOS[i]->GetObjectToSelect();
				if(!ObjectUnderCursorList.contains(pxObject))
				{
					ObjectUnderCursorList.add_entry(pxObject);
				}
			}
		}
	}
}




void Tools_Object_System::DeleteAllModeSpecific()
{
	//make sure nothing is selected
	EndMessageBlockEditing();
	ChooseObject_Finish();
	DeselectAllObjects();

	
	List<Tools_Object*> xPreservedObjects;

	//delete everything but the camera and tools helper objects
	while(ObjectList.size())
	{
		Tools_Object* pxObject = ObjectList.first_entry();
		if(pxObject->ShouldPersistBetweenModes())
		{
			//copy this object into the preserved list
			xPreservedObjects.add_entry(pxObject);
			ObjectList.delete_first_entry();
		}
		else
		{
			delete ObjectList.first_entry();
		}
	}

	//Copy the list of preserved objects back
	ObjectList = xPreservedObjects;

	//Entity objects will no longer exist, so update the state of those
	Tools_Entity_Object::PostDeleteAllObjects();

}

void Tools_Object_System::DeleteObject(Tools_Object* object)
{
	EndMessageBlockEditing();

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::HandleDeletedObject(object);
	#endif

	Tools_SecondaryEntityFile_System::RemoveFromHistoryList(object);

	ChooseObject_Finish();
	DeselectObject(object);
	delete object;
}


void Tools_Object_System::DestroySelected()
{
	EndMessageBlockEditing();
	if(!SelectedObject) return;
	if(!SelectedObject->CanBeDestroyed()) return;
	

	//confirm deletion
	if(SelectedObject->ConfirmDestruction())
	{
		char Question[200];
		sprintf(Question,"Delete %s?",SelectedObject->GetName());
		YesNo_Overlay dlg(Question);
		if(!dlg.DoModal()) return;
	}
	 

	Tools_Object* object = SelectedObject;

	//deselect before killing the object
	object->SetBeingDestroyed();
	DeselectObject(object);
	object->Destroy();

	CheckForMessageBlockOwnershipProblems();

}

void Tools_Object_System::DestroyAllSelected()
{
	DestroyObjects(SelectedObjectList);
}

void Tools_Object_System::DestroyObjects(const List<Tools_Object*>& xObjectList, const bool bAskForConfirmation /*= true*/)
{
	EndMessageBlockEditing();
	if(!xObjectList.size()) return;

	List<Tools_Object*> xObjectDestroyList = xObjectList;

	//confirm deletion
	if ( bAskForConfirmation )
	{
		if(xObjectDestroyList.size() == 1)
		{
			Tools_Object* pxFirstToolsObject = xObjectDestroyList.first_entry();
			Asura_Assert( NULL != pxFirstToolsObject, "The first tools object should not be NULL");
			if(!(pxFirstToolsObject->CanBeDestroyed())) 
			{
				return;
			}

			if(pxFirstToolsObject->ConfirmDestruction())
			{
				char Question[200];
				sprintf(Question,"Delete %s?",pxFirstToolsObject->GetName());
				YesNo_Overlay dlg(Question);
				if(!dlg.DoModal()) 
				{	
					return;
				}
			}
		}
		else
		{
			char Question[200];
			sprintf(Question,"Delete the %d objects selected?",xObjectDestroyList.size());
			YesNo_Overlay dlg(Question);
			if(!dlg.DoModal()) return;
		}

		if(!Tools_Message_Editing::CheckOkayToDeleteIncomingMessages(xObjectDestroyList))
		{
			return;
		}
	}

	bool bMultipleObjectSelection = xObjectDestroyList.size()>1;
	List<int> xTypesToDelete;
	List<int> xTypesToSave;
	u_int uDeletedMessages = 0;

	//destroy all the objects in the list
	while(xObjectDestroyList.size())
	{
		//get the next object to destroy
		Tools_Object* pxObject = xObjectDestroyList.first_entry();
		xObjectDestroyList.delete_first_entry();

		//make sure the object still exists , since it might be deleted as a side effect
		//of destroying something else
		if(!ObjectList.contains(pxObject)) continue;
		if(!pxObject->CanBeDestroyed()) continue;

		if(bMultipleObjectSelection && bAskForConfirmation && pxObject->ExtraConfirmationForMultiDestruction())
		{
			bool bDelete = false;
			//This object type requires confirmation before being destroyed
			//Have we seen this type before?
			if(xTypesToDelete.contains(pxObject->Type))
			{
				bDelete = true;
			}
			else if(xTypesToSave.contains(pxObject->Type))
			{
				//don't delete this object type
				bDelete = false;
			}
			else
			{
				//Need to ask about this still.
				//Count the number of objects of this type
				int iCount = 1;
				for(LIF<Tools_Object*> xObjectLIF(&xObjectDestroyList);!xObjectLIF.done();xObjectLIF.next())
				{
					if(xObjectLIF()->Type == pxObject->Type) iCount++;
				}
				//and ask if they should be deleted
				char szMessage[200];
				Asura_PrintfToBfr(szMessage,"Warning.\n%d object of type '%s' have been selected.\nAre you sure you want to delete these?",200,true,iCount,pxObject->GetTypeDescription());
				if(ShowYesNoOverlay(szMessage))
				{
					xTypesToDelete.add_entry(pxObject->Type);
					bDelete = true;
				}
				else
				{
					xTypesToSave.add_entry(pxObject->Type);
					bDelete = false;
				}
			}
			if(!bDelete) continue;
		}

		Asura_Guid uGuid = pxObject->GetGuid();
		if(uGuid != ASURA_GUID_UNREGISTERED)
		{
			uDeletedMessages += Tools_Message_Editing::DeleteIncomingMessages(uGuid);
		}

		//deselect before killing the object
		pxObject->SetBeingDestroyed();
		DeselectObject(pxObject);
		pxObject->Destroy();
	
	}

	if(uDeletedMessages)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Deleted %d message(s) targeting the deleted object(s)", uDeletedMessages);
	}

	CheckForMessageBlockOwnershipProblems();
}

bool Tools_Object_System::SelectObject(Tools_Object* object,bool multipleSelect /*=false*/, bool bChangePrimarySelection /*= true*/)
{
	if(s_pxChooseObject)
	{
		s_uSelectionChangedStamp++;
		return s_pxChooseObject->SelectObject(object,multipleSelect);
	}

	if(SelectedObject && SelectedObject->DisallowSelectionOfOtherTypes())
	{
		if(!object || object->Type != SelectedObject->Type)
		{
			return false;
		}
	}

	
	if(multipleSelect)
	{
		//if doing multiple select , and the object passed isn't selectable
		//then exit here , so the primary selection isn't lost
		if(!object) return false;
		if(!object->IsSelectable()) return false;
		if( IsFilteredOut( object->Type ) ) return false;
	}
	else
	{
		//empty the group selection list
		while(SelectedObjectList.size())
		{
			s_pxSelectionUndo->LogObjectDeselected(SelectedObjectList.first_entry());
			SelectedObjectList.first_entry()->OnGroupDeselection();
			SelectedObjectList.first_entry()->SetGroupSelected(false);
			SelectedObjectList.delete_first_entry();
			s_uSelectionChangedStamp++;
		}
	}
	

	if(bChangePrimarySelection)
	{
		if(object == SelectedObject)
		{
			//re-selecting the same object as before
			if(SelectedObject && !SelectedObject->IsGroupSelected())
			{
				SelectedObjectList.add_entry(SelectedObject);
				SelectedObject->SetGroupSelected(true);
				s_pxSelectionUndo->LogObjectSelected(SelectedObject);
			}
			if(object)
			{
				object->OnReselection();
			}
			return true;
		}
		
		//deselect the previous object
		if(SelectedObject)
		{
			//stop any flying
			Tools_Movement::SetFlyingObject(NULL);
			//stop snap to vertex
			CancelSnapToVertex();

			if(!MessageBlockEditing)
			{
				SelectedObject->OnDeselection();
			}
			s_pxSelectionUndo->LogObjectDeselectedPrimary(SelectedObject);
			SelectedObject = NULL;
		}
	}

	bool bObjectExists = false;
	if(object)
	{
		//Was previously just filtering out invalid objects silently.
		//I can't remember whether there was any good reason for that , so
		//changing it to an assert to see what happens.
		//RAR 12/11/2007
		bObjectExists = ObjectList.contains(object);
		Asura_Assert(bObjectExists,"Attempting to select an object that has been deleted?");
	}

	//select the new one , if it can be selected
	if(bObjectExists && object->IsSelectable())
	{
		// check against any selection filters that have been set up
		if( IsFilteredOut( object->Type ) )
		{
			return false;
		}

		if(bChangePrimarySelection)
		{
			SelectedObject = object;
			s_pxSelectionUndo->LogObjectSelectedPrimary(SelectedObject);
		}
		if(!object->IsGroupSelected())
		{
			SelectedObjectList.add_entry(object);
			object->SetGroupSelected(true);
			s_pxSelectionUndo->LogObjectSelected(object);
		}
		if(bChangePrimarySelection)
		{
			if(!MessageBlockEditing)
			{
				SelectedObject->OnSelection();
				Tools_Gizmo_Handler::SelectionChanged(SelectedObject);
			}
		}
		s_uSelectionChangedStamp++;
		return true;
	}
	return false;
}

void Tools_Object_System::SelectObjects(const List<Tools_Object*>& objectList,bool multipleSelect)
{
	Tools_Object* pxPreviousSelection = SelectedObject;

	if(!multipleSelect) DeselectAllObjects();

	Tools_Object* pxLastSelection = 0;

	for(CLIF<Tools_Object*> olif(&objectList);!olif.done();olif.next())
	{
		//Don't select the objects as the primary selection to avoid creating/deleting multiple overlays for them.
		//Keep track of the last one that was successfully selected though.
		if( SelectObject(olif(),true,false))
		{
			pxLastSelection = olif();
		}
	}
	//Select the last selected object again ,and make it the primary selection
	if(pxLastSelection)
	{
		SelectObject(pxLastSelection, true, true);
	}

	if(pxPreviousSelection && pxPreviousSelection->DisallowSelectionOfOtherTypes())
	{
		if(!SelectedObject || SelectedObject->Type != pxPreviousSelection->Type)
		{
			SelectObject(pxPreviousSelection, false, true);
		}
	}
}

void Tools_Object_System::DeselectObject(Tools_Object* object)
{
	if(s_pxChooseObject)
	{
		s_pxChooseObject->DeselectObject(object);
		s_uSelectionChangedStamp++;
		return;
	}

	
	if(object && object==SelectedObject)
	{
		//stop any flying
		Tools_Movement::SetFlyingObject(NULL);
		//stop snap to vertex
		CancelSnapToVertex();


		if(!MessageBlockEditing)
		{
			SelectedObject->OnDeselection();
		}
		s_pxSelectionUndo->LogObjectDeselectedPrimary(object);
		SelectedObject = NULL;
	}

	if(SelectedObjectList.contains(object))
	{
		s_pxSelectionUndo->LogObjectDeselected(object);
		object->OnGroupDeselection();
		object->SetGroupSelected(false);
		SelectedObjectList.delete_entry(object);
		s_uSelectionChangedStamp++;
	}
}

void Tools_Object_System::DeselectObjects(const List<Tools_Object*>& objectList)
{
	for(CLIF<Tools_Object*> olif(&objectList);!olif.done();olif.next())
	{
		DeselectObject(olif());
	}
}

void Tools_Object_System::DeselectAllObjects()
{
	if(s_pxChooseObject)
	{
		s_pxChooseObject->DeselectAllObjects();
		s_uSelectionChangedStamp++;
	}
	else
	{
		DeselectObject(SelectedObject);
		while(SelectedObjectList.size())
		{
			
			Tools_Object* pxObject = SelectedObjectList.first_entry();
			s_pxSelectionUndo->LogObjectDeselected(pxObject);
			pxObject->OnGroupDeselection();
			pxObject->SetGroupSelected(false);
			SelectedObjectList.delete_first_entry();
			s_uSelectionChangedStamp++;
		}
	}
}

void Tools_Object_System::GetSelectionListForCurrentSelectionMode(List<Tools_Object*>& xOutputList)
{
	if(s_pxChooseObject)
	{
		xOutputList = *s_pxChooseObject->GetSelectionList();
	}
	else
	{
		xOutputList = SelectedObjectList;
	}
}

Tools_Object* Tools_Object_System::GetMainSelectionForCurrentSelectionMode()
{
	if(s_pxChooseObject)
	{
		return s_pxChooseObject->GetMainSelection();
	}
	else
	{
		return SelectedObject;
	}
}



void Tools_Object_System::AddContextMenuItems(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext)
{

	Tools_Object* pxObject = 0;
	//determine which object under the cursor to base the menus on
	if(ObjectUnderCursorList.size()>1)
	{
		//ask which object to select
		ListSelection_Overlay overlay("Menus for which object?");

		//add all the object names to the overlay
		for(LIF<Tools_Object*> olif(&ObjectUnderCursorList);!olif.done();olif.next())
		{
			if(!olif()->GetName()) continue;

			overlay.AddItem(olif()->GetName(),(int)olif());
		}
		if(overlay.GetNumItems())
		{
			if(overlay.DoModal())
			{
				pxObject = (Tools_Object*)overlay.GetSelectionData();
			}
		}
		else
		{
			pxObject = ObjectUnderCursorList.first_entry();
		}
	}
	else if(ObjectUnderCursorList.size()==1)
	{
		pxObject = ObjectUnderCursorList.first_entry();
	}

	AddContextMenuItems(xMenu, xContext, pxObject);
}

void Tools_Object_System::AddContextMenuItems(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext, Tools_Object* pxObject)
{

	//Add options that are available regardless of selection

	Tools_ContextMenu_Section* pxSection = xMenu.GetSection(TOOLS_CONTEXTMENU_SECTION_SELECTION);
	Tools_ContextMenu_Item* pxItem;

	if(xContext.m_bShiftHeld)
	{

		pxItem = new Tools_ContextMenu_Item_Function("Deselect all", DeselectAllObjects);
		pxItem->SetID("Select all");
		pxSection->AddItem(pxItem);

		pxItem = new Tools_ContextMenu_Item_Function("Deselect visible", ContextMenu_DeselectAllVisible);
		pxItem->SetID("Select visible");
		pxSection->AddItem(pxItem);

	}
	else if(xContext.m_bControlHeld)
	{
		pxSection->SetHeading("Selection (add)");

		pxItem = new Tools_ContextMenu_Item_Function("Select all", ContextMenu_SelectAll, pxObject);
		pxItem->SetID("Select all");
		pxSection->AddItem(pxItem);

		pxItem = new Tools_ContextMenu_Item_Function("Select visible", ContextMenu_SelectAllVisible, pxObject);
		pxItem->SetID("Select visible");
		pxSection->AddItem(pxItem);

	}
	else
	{
		pxSection->SetHeading("Selection (replace)");

		pxItem = new Tools_ContextMenu_Item_Function("Select all", ContextMenu_SelectAll, pxObject);
		pxItem->SetID("Select all");
		pxSection->AddItem(pxItem);

		pxItem = new Tools_ContextMenu_Item_Function("Select visible", ContextMenu_SelectAllVisible, pxObject);
		pxItem->SetID("Select visible");
		pxSection->AddItem(pxItem);

	}
	
	pxSection = xMenu.GetSection(TOOLS_CONTEXTMENU_SECTION_DISPLAY);

	pxItem = new Tools_ContextMenu_Item_Function("Hide selected", ContextMenu_HideSelected);
	pxItem->SetEnabled(SelectedObjectList.size() > 0);
	pxItem->SetHelpText("Hide and deselect all selected objects");
	pxSection->AddItem(pxItem);

	pxItem = new Tools_ContextMenu_Item_Function("Hide unselected", ContextMenu_HideUnselected);
	pxItem->SetHelpText("Hide all unselected objects");
	pxSection->AddItem(pxItem);

	pxItem = new Tools_ContextMenu_Item_Function("Unhide all objects", ContextMenu_UnhideAll);
	pxItem->SetHelpText("Unhide all hidden objects.\nDoesn't unhide objects that have been hidden by type.");
	pxSection->AddItem(pxItem);

	pxItem = new Tools_ContextMenu_Item_Function("Unhide all types", ContextMenu_UnhideAllByType);
	pxItem->SetHelpText("Unhide all objects that have been hidden by type");
	pxSection->AddItem(pxItem);


	//Add object specific options
	if(pxObject)
	{
		pxObject->AddContextMenuItems(xMenu,xContext);
	}
}

void Tools_Object_System::AddContextMenuItemsFromOverlay(class Tools_ContextMenu& xMenu, const class Tools_ContextMenu_Context& xContext)
{
	Tools_Object* pxObject = (Tools_Object*) xContext.m_pOverlayUserData;
	if(pxObject)
	{
		//confirm this is an existing object
		if(!ObjectList.contains(pxObject))
		{
			Asura_Assert(0,"AddContextMenuItemsFromOverlay() - object doesn't exist");
			return;
		}
	}

	AddContextMenuItems(xMenu, xContext, pxObject);
}



void Tools_Object_System::ContextMenu_SelectAll(void* pObject)
{
	Tools_Object* pxObject = (Tools_Object*) pObject;

	List<Tools_Object*> xSelectionList = ObjectList;
	//Move the object clicked on to the end of the list so it becomes the primary selection
	if(pxObject && xSelectionList.contains(pxObject))
	{
		xSelectionList.delete_entry(pxObject);
		xSelectionList.add_entry(pxObject);
	}
	SelectObjects(xSelectionList,false);
}

void Tools_Object_System::ContextMenu_SelectAllVisible(void* pObject)
{
	Tools_Object* pxObject = (Tools_Object*) pObject;

	
	List<Tools_Object*> xSelectionList = ObjectList;
	//Move the object clicked on to the end of the list so it becomes the primary selection
	if(pxObject && xSelectionList.contains(pxObject))
	{
		xSelectionList.delete_entry(pxObject);
		xSelectionList.add_entry(pxObject);
	}
	//remove any hidden objects from the selection list
	for(LIF<Tools_Object*> xSelectionLIF(&xSelectionList); !xSelectionLIF.done();)
	{
		if(xSelectionLIF()->IsHidden())
		{
			xSelectionLIF.delete_current();
		}
		else
		{
			xSelectionLIF.next();
		}
	}

	SelectObjects(xSelectionList,false);
}

void Tools_Object_System::ContextMenu_DeselectAllVisible()
{

	List<Tools_Object*> xSelectedList = SelectedObjectList;
	for(LIF<Tools_Object*> xSelectedLIF(&xSelectedList); !xSelectedLIF.done(); xSelectedLIF.next())
	{
		if(!xSelectedLIF()->IsHidden())
		{
			DeselectObject(xSelectedLIF());
		}
	}
}

void Tools_Object_System::ContextMenu_HideSelected()
{
	for(LIF<Tools_Object*> xObjectLIF(&SelectedObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		xObjectLIF()->SetHiddenFlagIncludingChildren(ObjectHideFlag_User,true);
	}
	DeselectAllObjects();
}

void Tools_Object_System::ContextMenu_HideUnselected()
{
	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		if ( !xObjectLIF()->IsGroupSelected() )
		{
			xObjectLIF()->SetHiddenFlagIncludingChildren(ObjectHideFlag_User,true);
		}
	}
}

void Tools_Object_System::ContextMenu_UnhideAll()
{
	//Unhide individually hidden objects (which use the user hide flag)
	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		xObjectLIF()->SetHiddenFlagIncludingChildren(ObjectHideFlag_User,false);
	}
}

void Tools_Object_System::ContextMenu_UnhideAllByType()
{
	//Unhide objects that have been hidden by type (which share the hide flag from the hide by type overlay)
	g_xObjectTypeVisibility.RemoveHideFlagForAllTypes(ObjectHideFlag_HideByTypeOverlay);
}

void Tools_Object_System::HandleKeyboard()
{
	if(Asura_Overlay_System::OverlaysHaveKeyboardFocus()) return;



	//common keys to all object system modes
	{
 		if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
		{
			if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_F12))
			{
				ScaleSelectedPositions();
			}
			if (Asura_UserInput::DebouncedKeyPressed('v'))
			{
				ObjectHideByType_Overlay::ToggleOverlay();
			}
			if (Asura_UserInput::DebouncedKeyPressed('f'))
			{
				FindObject();
			}
			if (Asura_UserInput::DebouncedKeyPressed('g'))
			{
				GotoSelected();
			}
			if (Asura_UserInput::DebouncedKeyPressed('h'))
			{
				FindObjectByGuid();
			}
			if (Asura_UserInput::DebouncedKeyPressed(','))
			{
				SelectPreviousFoundObject();
			}
			if (Asura_UserInput::DebouncedKeyPressed('.'))
			{
				SelectNextFoundObject();
			}
		}
		else
		{
			#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR

			switch(Tools_System::GetCurrentMode())
			{
				case TOOLS_MODE_OBJECT:
				case TOOLS_MODE_CUTSCENE:
				case TOOLS_MODE_SOUND:
				case TOOLS_MODE_ATTRACTORS:
				{
			 		if(!Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) && !Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
					{
						if (Asura_UserInput::DebouncedKeyPressed('m'))
						{
							Tools_MessageEditor::StartMessageEditingMode();
						}
					}

			 		if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) && !Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
					{
						if (Asura_UserInput::DebouncedKeyPressed('r'))
						{
							Tools_MessageEditor::ToggleOverlay();
						}
					}

					break;
				}
			}

			#endif
		}
	}


	//keys for 'choose object' mode
	if(s_pxChooseObject)
	{
		s_pxChooseObject->HandleKeyboard();

 		if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
		{
		}
 		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
 		{
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
		{
		}
		else
		{
			if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_SPACE))
			{
				ChooseObject_Accept();

			}
			else if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ESCAPE))
			{
				ChooseObject_Finish();
			}
		}
	}
	//keys for message block editing
	else if(MessageBlockEditing)
	{
 		if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
		{
		}
 		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
 		{
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
		{
		}
		else
		{
			if (Asura_UserInput::DebouncedKeyPressed('l'))
			{
				DeselectAllObjects();
			}
			if(ControlsEnabled)
			{
				if (Asura_UserInput::DebouncedKeyPressed('f'))
				{
					ToggleFlying();
				}
				if (Asura_UserInput::DebouncedKeyPressed('v'))
				{
					StartSnapToVertex();
				}

			}


		}
		Tools_Message_Editing::HandleKeyboard();
	}
	//standard object system keys
	else
	{
 		if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
		{
			
		}
 		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
 		{
			if(ControlsEnabled)
			{
				Asura_Overlay* pxTopOverlay = Asura_Overlay_System::GetTopOverlay();

				if(!pxTopOverlay || !Asura_Overlay_System::WasLastMouseClickOnOverlay() || !pxTopOverlay->DoesOverlayOverrideCutAndPasteKey())
				{
					if(Asura_UserInput::DebouncedKeyPressed('c'))
					{
						CopyToClipboard();
					}

					if(Asura_UserInput::DebouncedKeyPressed('v'))
					{
						if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
							PasteOver();
						else
							Paste();
					}
				}

				if(Asura_UserInput::DebouncedKeyPressed('d'))
				{
					DuplicateSelected();
				}
				if(Asura_UserInput::DebouncedKeyPressed('z'))// 'Ctrl+Z' for  undo
				{
					if(s_pxUndo->Undo())
					{
						Tools_Timed_Message_System::AddMessage(0xffffff,"Undo");
					}
				}
				if(Asura_UserInput::DebouncedKeyPressed('y'))// 'Ctrl+Y' for  redo
				{
					if(s_pxUndo->Redo())
					{
						Tools_Timed_Message_System::AddMessage(0xffffff,"Redo");
					}
				}
			}
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
		{

			if(ControlsEnabled)
			{
				if (Asura_UserInput::DebouncedKeyPressed('q'))
				{
					Tools_Gizmo_Handler::ToggleGizmos(SelectedObject);
				}
				if (Asura_UserInput::DebouncedKeyPressed('w'))
				{
					Tools_Gizmo_Handler::ShowOverlay();

				}
				if(Asura_UserInput::DebouncedKeyPressed('z') || Asura_UserInput::DebouncedKeyPressed(','))// 'Shift+Z' or 'Shift+<' for  selection undo
				{
					if(s_pxSelectionUndo->Undo())
					{
						Tools_Timed_Message_System::AddMessage(0xffffff,"Selection Undo");
					}
				}
				if(Asura_UserInput::DebouncedKeyPressed('y') || Asura_UserInput::DebouncedKeyPressed('.'))// 'Shift+Y' or 'Shift+>'for  selection redo
				{
					if(s_pxSelectionUndo->Redo())
					{
						Tools_Timed_Message_System::AddMessage(0xffffff,"Selection Redo");
					}
				}
			}
		}
		else
		{
			if(ControlsEnabled)
			{
				if (Asura_UserInput::DebouncedKeyPressed('f'))
				{
					ToggleFlying();
				}

				if(Asura_UserInput::DebouncedKeyPressed('e') && !Tools_Movement::IsUsingDynamics())//(E is used for another purpose when dynamics is on)
				{
					if(SelectedObject) SelectedObject->Edit();
				}
				if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_DEL))
				{
					Asura_Overlay* pxTopOverlay = Asura_Overlay_System::GetTopOverlay();

					if(!pxTopOverlay || !Asura_Overlay_System::WasLastMouseClickOnOverlay() || !pxTopOverlay->DoesOverlayOverrideDeleteKey())
					{
						DestroyAllSelected();
					}
				}
				if (Asura_UserInput::DebouncedKeyPressed('v'))
				{
					StartSnapToVertex();
				}
				if(Asura_UserInput::DebouncedKeyPressed(','))// '<' for  undo
				{
					if(s_pxUndo->Undo())
					{
						Tools_Timed_Message_System::AddMessage(0xffffff,"Undo");
					}
				}
				if(Asura_UserInput::DebouncedKeyPressed('.'))// '>' for  redo
				{
					if(s_pxUndo->Redo())
					{
						Tools_Timed_Message_System::AddMessage(0xffffff,"Redo");
					}
				}

			}
			if (Asura_UserInput::DebouncedKeyPressed('l'))
			{
				DeselectAllObjects();
			}


		}
	}
}


void Tools_Object_System::ToggleFlying()
{
	if(Tools_Movement::GetFlyingObject())
	{
		//currently flying something , so stop
		Tools_Movement::SetFlyingObject(NULL);
		Tools_Timed_Message_System::AddMessage(0xffffff,"Flying off.");
	}
	else if(SelectedObject)
	{

		if(SelectedObject->CanBeMoved())
		{
			CancelSnapToVertex();
			//start flying the selected object
			Tools_Movement::SetFlyingObject(SelectedObject);
			Tools_Timed_Message_System::AddMessage(0xffffff,"Flying on.");

			//calculate the relative positions for any other selected objects
			UpdateRelativePositions();
			
		}
		else
		{
			Tools_Timed_Message_System::AddMessage(0xffffff,"This object can't be moved.");
		}
	}
}

void Tools_Object_System::UpdateGroupMovement()
{
	if(!SelectedObject)
	{
		//Its possible to have everything deselected while still moving an object via the movement gizmos.
		//In that case just ignore this.
		return;
	}
	//calculate the relative positions for any other selected objects
	for(LIF<Tools_Object*> olif(&SelectedObjectList);!olif.done();olif.next())
	{
		if(olif()->CanBeMoved())
		{
			olif()->UpdateRelative(SelectedObject);
		}
	}
}

void Tools_Object_System::UpdateRelativePositions()
{
	for(LIF<Tools_Object*> olif(&SelectedObjectList);!olif.done();olif.next())
	{
		olif()->CalculateRelativePosition(SelectedObject);
	}
}


void Tools_Object_System::ListObjectsOfType(int type,List<Tools_Object*>& outputList)
{
	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		if(type == olif()->Type)
		{
			outputList.add_entry(olif());
		}
	}
}

void Tools_Object_System::ListSelectedObjectsOfType(int iType,List<Tools_Object*>& xOutputList)
{
	for(LIF<Tools_Object*> olif(&SelectedObjectList);!olif.done();olif.next())
	{
		if(iType == olif()->Type)
		{
			xOutputList.add_entry(olif());
		}
	}
}


void Tools_Object_System::EditMessageBlock(const char* prompt,unsigned short* messageBlock,int iMessageSet,bool bClientMessages)
{
	if(MessageBlockEditing) EndMessageBlockEditing();

	CheckForMessageBlockOwnershipProblems();

	Asura_Assert(iMessageSet>=0 && iMessageSet<ASURA_STATICMESSAGE_NUMSETS,"Invalid message set");

	Tools_System::CurrentModeClass->PreMessageBlockEditing();

	//Asura_Assert(SelectedObject,"");
	MessageBlockObject = SelectedObject;

	DeselectAllObjects();
	s_xSelectionState.Reset();

	MessageBlockEditing = true;
	Tools_Message_Editing::StartEdit(prompt,messageBlock,(Asura_StaticMessageSetIndex)iMessageSet,bClientMessages,MessageBlockObject);

	
	Tools_Timed_Message_System::AddMessage(0xffffff,"Editing messages.");

}

void Tools_Object_System::EditMessageBlock(const char* szPrompt,unsigned short usMessageBlockID,int iMessageIndex,Tools_Object* pxSenderObject)
{
	if(MessageBlockEditing) EndMessageBlockEditing();
	//can't edit an empty block via this function
	Asura_Assert(usMessageBlockID != ASURA_STATICMESSAGE_EMPTYBLOCK,"");
	if(usMessageBlockID == ASURA_STATICMESSAGE_EMPTYBLOCK) return;


	CheckForMessageBlockOwnershipProblems();

	//determine the messages set
	enum Asura_StaticMessageSetIndex eSet;
	int iBlockIndex;
	if(!Asura_StaticMessage_System::ConvertIDToIndex(usMessageBlockID, &iBlockIndex, &eSet)) return;

	Tools_System::CurrentModeClass->PreMessageBlockEditing();

	//Asura_Assert(SelectedObject,"");
	MessageBlockObject = NULL;

	DeselectAllObjects();
	s_xSelectionState.Reset();

	MessageBlockEditing = true;
	Tools_Message_Editing::StartEdit(szPrompt,usMessageBlockID,iMessageIndex,eSet,pxSenderObject);

	
	Tools_Timed_Message_System::AddMessage(0xffffff,"Editing messages.");
}

void Tools_Object_System::EditMessageBlock(const char* prompt,unsigned short* messageBlock,bool bClientMessages)
{
	if(!SelectedObject)
	{
		Asura_Assert(SelectedObject,"No object selected. Aborting message editing.");	
		return;
	}
	
	//Determine which message set should be edited , according to whether the object's guid
	//is in the entity or cutscene range.
	Asura_Guid uGuid = SelectedObject->GetGuid();

	if(uGuid == ASURA_GUID_UNREGISTERED)
	{
		Asura_Assert(false,"Selected object doesn't have a guid. Aborting message editing.");
		return;
	}

	Asura_StaticMessageSetIndex eSet;
	if(uGuid>=ASURA_GUID_TOOLS_ENTITYFILE_START && uGuid <=ASURA_GUID_TOOLS_ENTITYFILE_END)
	{
		eSet = ASURA_STATICMESSAGE_ENTITYSET;
	}
	else if(uGuid>=ASURA_GUID_TOOLS_CUTSCENEFILE_START && uGuid <=ASURA_GUID_TOOLS_CUTSCENEFILE_END)
	{
		eSet = ASURA_STATICMESSAGE_CUTSCENESET;
	}
	else
	{
		Asura_Assert(0,"Object doesn't have a guid in the tools allocated range. Aborting message editing");
		return;
	}

	EditMessageBlock(prompt,messageBlock,eSet,bClientMessages);

}

void Tools_Object_System::EditMessageBlock(const char* szPrompt,Asura_StaticMessageBlock* pxMessageBlock,Tools_Object* pxSenderObject,bool* pbModifiedFlag)
{
	if(MessageBlockEditing) EndMessageBlockEditing();
	
	Tools_System::CurrentModeClass->PreMessageBlockEditing();

	MessageBlockObject = NULL;

	DeselectAllObjects();
	s_xSelectionState.Reset();

	MessageBlockEditing = true;
	Tools_Message_Editing::StartEdit(szPrompt,pxMessageBlock,false,pxSenderObject,pbModifiedFlag);

	
	Tools_Timed_Message_System::AddMessage(0xffffff,"Editing messages.");
}



void Tools_Object_System::EndMessageBlockEditing()
{
	if ( MessageBlockEditing && s_bPreserveMessageEditing == false )
	{
		//reset all the vars
		SelectObject(0);
		Tools_Message_Editing::EndEdit();
		MessageBlockEditing = false;

		//revert back to the originally selected object
		SelectObject(MessageBlockObject);
		MessageBlockObject = 0;

		Tools_System::CurrentModeClass->PostMessageBlockEditing();
	}
}

bool Tools_Object_System::AskForName(Asura_String& name,const char* title,const Tools_Object* objectToIgnore /*= 0*/,bool bAllowBlank /*=false*/,const char* szForcePrefix /*= 0*/)
{
	if(name.IsNull()) name = "";
	Asura_String nameCopy = name;

	if( !nameCopy.StartsWith( szForcePrefix ) )
	{
		nameCopy.Prepend( szForcePrefix );
	}

	bool okay;
	do
	{
		okay = true;
		//bring up the overlay
		GetString_Overlay overlay(title,nameCopy.GetString());
		if(!overlay.DoModal()) return false;

		//get the string
		nameCopy = overlay.GetString();
		if(!nameCopy.GetLength())
		{
			if(bAllowBlank)
			{
				//blank names allowed... so just accept it
				break;
			}
			else
			{
				//not allowed , so ask again
				okay = false;
				continue;
			}
		}
		else
		if( szForcePrefix )
		{
			if( !nameCopy.StartsWith( szForcePrefix ) )
			{
				// prefix missing, ask again
				okay = false;
				nameCopy.Prepend( szForcePrefix );
				continue;
			}
		}

		//see if the name is already in use
		if(IsNameInUse(nameCopy, objectToIgnore))
		{
			okay = false;//name already in use
			Tools_Timed_Message_System::AddMessage(0xff0000,"That name is already in use");
		}
	}while(!okay);
	name = nameCopy;

	return true;
}

bool Tools_Object_System::IsNameInUse(const Asura_String& xName, const Tools_Object* objectToIgnore /* = 0 */)
{
	for(LIF<Tools_Object*> olif(&ObjectList);!olif.done();olif.next())
	{
		if(olif()==objectToIgnore) continue;
		if(!stricmp(xName.GetString(),olif()->GetName()))
		{
			return true;
		}
	}
	return false;
}

void Tools_Object_System::EnableAllControls(bool enable)
{
	if(enable == ControlsEnabled) return;
	ControlsEnabled = enable;
	if(ControlsEnabled)
	{
		//add help
		Tools_Help::AddHelpTopic("Object Editing",ObjectSystemHelpStrings);
	}
	else
	{
		Tools_Help::RemoveHelpTopic("Object Editing");
	}
	s_xSelectionState.Reset();

	//Update visiblity of helper objects. RAR 04/09/2006
	Tools_HelperObject_Object::UpdateVisibility(enable);
}

// Needed for menu bar
void Tools_Object_System::ToggleObjectControls()
{
	EnableAllControls( !AreControlsEnabled() );
}

bool Tools_Object_System::TypeIsHidden(int type)
{
	return g_xObjectTypeVisibility.TypeIsHidden(type);
}


bool Tools_Object_System::GetTypeHiddenFlag(int iType,int iFlags)
{
	return g_xObjectTypeVisibility.GetTypeHiddenFlag(iType,iFlags);
	
}

void Tools_Object_System::SetTypeHidden(int type,int flags,bool hide)
{
	#ifdef ASURA_USE_BOTANICALS
	//For botanicals , update the Tools_System flag as well
	if(type == TOT_Botanicals && (flags == ObjectHideFlag_HideByTypeOverlay))
	{
		if(Tools_System::IsHideBotanicalsChecked() != hide)
		{
			Tools_System::ToggleHideBotanicals();
		}
	}
	#endif
	g_xObjectTypeVisibility.SetTypeHidden(type,flags,hide);
}

void Tools_Object_System::SetVisibleLightSets(u_int uLightSets)
{
	s_uVisibleLightSets = uLightSets & Tools_LightSets::GetValidLightSets();
}


void Tools_Object_System::SetLightSetVisible(int iLightSet,bool bVisible)
{
	int iBit = 1 << iLightSet;
	if(bVisible)
	{
		SetVisibleLightSets(s_uVisibleLightSets | iBit);
	}
	else
	{
		SetVisibleLightSets(s_uVisibleLightSets &~ iBit);
	}
}

void Tools_Object_System::SetVisiblePlatforms(u_int uPlatforms)
{
	int iPlatformMask = (1<<(Tools_Level_Export::ExportType_NumPlatforms+1))-1;

	s_uVisiblePlatforms = uPlatforms & iPlatformMask;
}

void Tools_Object_System::SetPlatformVisible(int iPlatform,bool bVisible)
{
	int iBit = 1 << iPlatform;
	if(bVisible)
	{
		SetVisiblePlatforms(s_uVisiblePlatforms | iBit);
	}
	else
	{
		SetVisiblePlatforms(s_uVisiblePlatforms &~ iBit);
	}
}


void Tools_Object_System::SetHideAll(int iFlags,bool bHide)
{
	if(bHide)
	{
		s_iHideAllFlags |= iFlags;
	}
	else
	{
		s_iHideAllFlags &=~ iFlags;
	}
}

void Tools_Object_System::ShowUserHideByTypeOverlay()
{
	ObjectHideByType_Overlay::Create(&ObjectList);
}




Tools_Object* Tools_Object_System::ChooseObjectFromList(const char* message,bool includeHiddenObjects/*=false*/)
{
	ObjectSelection_Overlay overlay(message,&ObjectList,includeHiddenObjects);

	
	if(!overlay.GetNumItems())
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"No suitable objects for selection.");
		return 0;
	}

	//show the overlay
	if(overlay.DoModal())
	{
		return (Tools_Object*) overlay.GetSelectionData();
	}
	return 0;
}

void Tools_Object_System::FindObject()
{
	ObjectFind_Overlay xOverlay("Find Object :",&ObjectList,false);

	if(!xOverlay.GetNumItems())
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"No suitable objects for selection.");
		return;
	}

	//show the overlay
	if(!xOverlay.DoModal()) return;
	
	Tools_Object* pxObject =  (Tools_Object*) xOverlay.GetSelectionData();
	if(!pxObject) return;
	
	//select this object
	SelectObject(pxObject->GetObjectToSelect());

	//and goto if needed	
	if(xOverlay.ShouldGotoSelected())
	{
		GotoObject(pxObject);
	}

}


void Tools_Object_System::FindObjectByGuid()
{
    GetIntNumber_Overlay xOverlay( "Enter Guid", s_uFindObjectGuid, 0, 9999999 );

	if (xOverlay.DoModal())
	{
		s_uFindObjectGuid = static_cast< Asura_Guid >( xOverlay.GetNumber() );

		Tools_Object *pxToolsObject = GetObjectFromGuid(s_uFindObjectGuid);

		if (pxToolsObject)
		{
			Tools_Object* pxObjectToSelect = pxToolsObject->GetObjectToSelect();
			if(pxObjectToSelect != pxToolsObject)
			{
				if(pxToolsObject->IsEntityObject() && ((Tools_Entity_Object*)pxToolsObject)->IsMemberOfTemplateInstance())
				{
					Tools_Timed_Message_System::AddMessage(0xffffff,"Entity %d (%s) is part of a template. Selecting the template instance instead.",s_uFindObjectGuid, pxToolsObject->GetName());
				}
				else
				{
					Tools_Timed_Message_System::AddMessage(0xffffff,"Object with guid %d (%s) isn't directly selectable. Selecting the parent instead.",s_uFindObjectGuid, pxToolsObject->GetName());
				}
				
			}

			
			SelectObject(pxObjectToSelect);
			Asura_Vector_3 xPosition = pxToolsObject->Position;
			Asura_Vector_3 xFwdVec;
			Asura_Camera::GetForwardVector(xFwdVec);
			xPosition -= xFwdVec * 2.0f;
			Asura_Matrix_3x3 xMat = *Asura_Camera::GetOrientation();
			Tools_Movement::TeleportCameraToPosition(&xPosition, &xMat);
		}
		else
		{
			Tools_Timed_Message_System::AddMessage(0xffffff,"Could not find specified object.");
			return;
		}
	}
}


void Tools_Object_System::GotoAndSelect(Tools_Object* pxObject)
{
	if(!pxObject)
	{
		Asura_Assert(pxObject, "GotoAnSelect() - NULL object pointer");
		return;
	}

	//select this object
	SelectObject(pxObject);


	//and go to it
	GotoObject(pxObject);
}

void Tools_Object_System::GotoObject(Tools_Object* pxObject)
{
	if(!pxObject)
	{
		Asura_Assert(pxObject, "GotoObject() - NULL object pointer");
		return;
	}

	const Asura_Matrix_3x3* cameraOrientation = Asura_Camera::GetOrientation();

	Asura_Vector_3 position = pxObject->Position;
	position.x -= cameraOrientation->e13*2.0f;
	position.y -= cameraOrientation->e23*2.0f;
	position.z -= cameraOrientation->e33*2.0f;

	Asura_Camera::SetPosition(position);

}

void Tools_Object_System::GotoSelected()
{
	if(s_pxChooseObject)
	{
		if(s_pxChooseObject->GetMainSelection()) GotoObject(s_pxChooseObject->GetMainSelection());
	}
	else
	{
		if(SelectedObject) GotoObject(SelectedObject);
	}
}


void Tools_Object_System::StartSnapToVertex()
{
	if(DoingSnapToVertex) return;
	if(!SelectedObject) return;

	if(!SelectedObject->CanBeMoved())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"This object can't be moved.");
		return;
	}

	//make sure the object isn't being flown
	Tools_Movement::SetFlyingObject(NULL);

	//calculate the relative positions for any other selected objects
	UpdateRelativePositions();
	


	//store the current position and orientation
	SnapToVertexPositionStore = SelectedObject->Position;
	SnapToVertexOrientationStore = SelectedObject->Orientation;

	DoingSnapToVertex = true;

	s_bSnapToVertex_SourceVertexLocked = false;
	s_bSnapToVertex_PickingSourceVertex = false;
	s_xSnapToVertex_SourceVertex = SelectedObject->Position;
	s_xSnapToVertex_HighlightPosition = SelectedObject->Position;
	s_iSnapToVertex_Rotate = 0;
	s_iSnapToVertex_SetOrientState = 0;
	//get bounding box now , to avoid problems with animating objects
	SelectedObject->GetBoundingBoxForSnapToVertex(s_xSnapToVertex_SourceBB);


	//reset selection state ,since we don't want to be selecting while doing this
	s_xSelectionState.Reset();
}

void Tools_Object_System::CancelSnapToVertex()
{
	if(!DoingSnapToVertex) return;
	Asura_Assert(SelectedObject,"");

	//restore the original position/orientation
	SelectedObject->Position = SnapToVertexPositionStore;
	SelectedObject->Orientation = SnapToVertexOrientationStore;

	SelectedObject->UpdatePositionInAsuraData();

	UpdateGroupMovement();

	DoingSnapToVertex = false;
}


void Tools_Object_System::HandleSnapToVertex()
{
	if(!DoingSnapToVertex) return;
	Asura_Assert(SelectedObject,"");

	if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ESCAPE))
	{
		CancelSnapToVertex();
		return;
	}


	SelectedObject->Position = SnapToVertexPositionStore;
	SelectedObject->Orientation = SnapToVertexOrientationStore;

	//Need to do this for to allow snap point calculation for templates to work correctly
	//hopefully it shouldn't cause any undesirable side effects. RAR 09/07/04
	SelectedObject->UpdatePositionInAsuraData();

	if(Asura_UserInput::DebouncedKeyPressed('c'))
	{
		s_bSnapToVertex_UseCollisionMesh = !s_bSnapToVertex_UseCollisionMesh;
	}
	if(Asura_UserInput::DebouncedKeyPressed('i'))
	{
		s_bSnapToVertex_AlignMode_AlignYAxis = !s_bSnapToVertex_AlignMode_AlignYAxis;
	}

	
	//Check to see if we are picking the source vertex for snapping.
	//V must be held down , and it must be debounced unless we have already started ,or if
	//no vertex has been selected yet
	if((Asura_UserInput::KeyPressed('v') && (!s_bSnapToVertex_SourceVertexLocked || s_bSnapToVertex_PickingSourceVertex))
	  ||Asura_UserInput::DebouncedKeyPressed('v'))
	{
		s_bSnapToVertex_PickingSourceVertex = true;

		//holding V down... so selecting snap point in source object
		Asura_Vector_3 xSnapPoint = s_xSnapToVertex_SourceVertex;

		//find object in LOS , and get position of cursor over that object
		Tools_Object* pxObject = FindObjectUnderCursor();
		if(pxObject)
		{
			
			//get mesh (if it has one) , position and orient of object cursor is over
			int iDestPolygon;
			Asura_Vector_3 xIntersection;
			if(!pxObject->GetPolygonAndPositionUnderCursor(iDestPolygon,xIntersection))
			{
				xIntersection = pxObject->Position;
			}

			//find the nearest snap point in the source object
			xSnapPoint = SelectedObject->GetSnapPoint(xIntersection);
			s_xSnapToVertex_HighlightPosition = xSnapPoint;
		}

		if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
		{
			//accept this snap point
			s_bSnapToVertex_SourceVertexLocked = true;
			s_bSnapToVertex_PickingSourceVertex = false;
			s_xSnapToVertex_SourceVertex = s_xSnapToVertex_HighlightPosition;
		}
	}
	else
	{
		s_bSnapToVertex_PickingSourceVertex = false;

		enum ObjectSnapSource
		{
			ObjectSnapSource_BoundingBox,
			ObjectSnapSource_Vertex,
			ObjectSnapSource_Pivot,
		};
		enum ObjectSnapDest
		{
			ObjectSnapDest_Cursor,
			ObjectSnapDest_Vertex,
			ObjectSnapDest_Pivot,
		};
		enum SetOrientState
		{
			SetOrientState_Off,
			SetOrientState_Pending,
			SetOrientState_On,
		};
		

   		ObjectSnapSource eSnapSource = ObjectSnapSource_BoundingBox;
		ObjectSnapDest eSnapDest = ObjectSnapDest_Cursor;

		bool bAlign = Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT);
		bool bToObject = Asura_UserInput::KeyPressed(ASURA_KEYID_LEFTALT);

		if(SelectedObject->UsePivotForSnapToVertex())
		{
			eSnapSource = ObjectSnapSource_Pivot;
		}

		if(bToObject)
		{
			eSnapSource = ObjectSnapSource_Pivot;
			eSnapDest = ObjectSnapDest_Pivot;
		}
		if(Asura_UserInput::KeyPressed(ASURA_KEYID_RIGHTALT))
		{
			//Giving Right Alt precedence to avoid problems when using UK key settings ,
			//where pressing Right Alt results in both Right Alt and Left Ctrl.

			eSnapSource = ObjectSnapSource_Pivot;
			eSnapDest = ObjectSnapDest_Cursor;
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_LEFTCTRL))
		{
			eSnapSource = ObjectSnapSource_Vertex;
			eSnapDest = ObjectSnapDest_Vertex;
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_RIGHTCTRL))
		{
			eSnapSource = ObjectSnapSource_Pivot;
			eSnapDest = ObjectSnapDest_Vertex;
		}


		if(Asura_UserInput::DebouncedKeyPressed('r'))
		{
			s_iSnapToVertex_Rotate++;
			if(s_iSnapToVertex_Rotate==5) s_iSnapToVertex_Rotate = 1;
		}





		Asura_Mesh* pxDestMesh = 0;
		const Asura_CollisionMesh* pxDestCollMesh = 0;
		Tools_Object* pxDestObject = 0;
		Asura_Vector_3 xDestScale(1.0f,1.0f,1.0f);
		Asura_Vector_3 xDestNormal(0.0f,0.0f,0.0f);
		Asura_Vector_3 xDestPos; //location of mesh under cursor
		Asura_Matrix_3x3 xDestOrient;//and orient
		Asura_Vector_3 xIntersection;//world position that cursor is over
		bool bDestValid = false;

		if(bToObject)
		{
			//find the object under the cursor
			//want to ignore the object being snapped , and any of its child objects
			List<Tools_Object*> xObjectsToIgnore;
			SelectedObject->ListChildObjects(xObjectsToIgnore,true);
			xObjectsToIgnore.add_entry(SelectedObject);

			Tools_Object* pxObject = FindObjectUnderCursor(xObjectsToIgnore);
			if(pxObject)
			{
				
				//get mesh (if it has one) , position and orient of object cursor is over
				pxDestMesh = pxObject->Mesh;
				pxDestObject = pxObject;
				xDestPos = pxObject->Position;
				xDestOrient = pxObject->Orientation;
				xDestScale = pxObject->GetScale();
				if(!pxObject->GetPolygonNormalAndPositionUnderCursor(xDestNormal,xIntersection))
				{
					xIntersection = pxObject->Position;
					pxDestMesh = 0;
					xDestNormal.Zero();
					xDestScale.x = 1.0f;
					xDestScale.y = 1.0f;
					xDestScale.z = 1.0f;
				}
				bDestValid = true;
			}
		}
		else
		{
			//get the position under the mouse
			int iModuleIndex;
			Asura_Hash_ID uZoneHash;

			if(s_bSnapToVertex_UseCollisionMesh)
			{
				Tools_LOS::SetPolygonFlagsToIgnore(AsuraPolygonFlag_NoCollisions);
				if(Tools_LOS::FindCollModuleAndPositionUnderCursor(iModuleIndex,xIntersection,&xDestNormal,&uZoneHash))
				{
					#ifdef ASURA_USE_STREAMING
					Asura_Module_Handler* pxHandler = Asura_Module_System::GetHandler(uZoneHash);
					Asura_Assert(pxHandler,"Failed to get module handler for zone");
					#else
					Asura_Module_Handler* pxHandler = &Asura_Module_System::GetCurrentHandler();
					#endif
					
					//get collision mesh, position and orient of module cursor is over
					pxDestCollMesh = pxHandler->GetCollisionMeshForModule(iModuleIndex);
					pxHandler->GetPositionOfModule(iModuleIndex,xDestPos);
					xDestOrient.Identity();
					bDestValid = true;
				}
				Tools_LOS::SetPolygonFlagsToIgnore(0);
			}
			else
			{
				int iDestPolygon;
				if(Tools_LOS::FindPolyAndPositionUnderCursor(iDestPolygon,iModuleIndex,xIntersection,&uZoneHash))
				{
					#ifdef ASURA_USE_STREAMING
					Asura_Module_Handler* pxHandler = Asura_Module_System::GetHandler(uZoneHash);
					Asura_Assert(pxHandler,"Failed to get module handler for zone");
					#else
					Asura_Module_Handler* pxHandler = &Asura_Module_System::GetCurrentHandler();
					#endif
					
					//get mesh, position and orient of module cursor is over
					pxDestMesh = pxHandler->GetMeshForModule(iModuleIndex);
					pxHandler->GetPositionOfModule(iModuleIndex,xDestPos);
					xDestOrient.Identity();
					bDestValid = true;

					if(!pxDestMesh->AreNormalsValid()) pxDestMesh->CreateNormals();
					xDestNormal = pxDestMesh->NormalsDataPtr[iDestPolygon].FaceNormal;
				}
			}

		}
	
		if(s_iSnapToVertex_SetOrientState != SetOrientState_Off)
		{
			//attempt to turn the object so that the z vector points towards the point under the cursor
			if(s_iSnapToVertex_SetOrientState == SetOrientState_Pending)
			{
				//See if the cursor is being dragged while the mouse is held.
				//If so then we start with the orientation setting for real
				int iX,iY;
				Asura_Cursor::GetScreenSpaceCoords(iX,iY);

				int iOffset = Asura_Maths::Abs(iX-s_iSnapToVertex_DragStartX) + Asura_Maths::Abs(iY-s_iSnapToVertex_DragStartY);
				if(iOffset > Asura_Viewport::GetWidth()/20)
				{
					s_iSnapToVertex_SetOrientState = SetOrientState_On;
				}
			}
			
			if(bDestValid && s_iSnapToVertex_SetOrientState == SetOrientState_On)
			{
				//get a direction vector from the object to the cursor
				Asura_Vector_3 xZDir = xIntersection - SelectedObject->Position;

				//Try to keep the current Y direction , so resolve the Z dir to be perpendicular
				Asura_Vector_3 xYDir;
				xYDir.x = SelectedObject->Orientation.e12;
				xYDir.y = SelectedObject->Orientation.e22;
				xYDir.z = SelectedObject->Orientation.e32;

				float fYDistance = xYDir*xZDir;
				xZDir -= xYDir*fYDistance;

				if(xZDir.Magnitude() > 0.01f)
				{
					xZDir.Normalise();

					//and set the x dir to be perpendicular to the the original Y and the Z we have just got
					Asura_Vector_3 xXDir = xYDir^xZDir;
					xXDir.Normalise();


					//fill in the new orientation
					SelectedObject->Orientation.e11 = xXDir.x;
					SelectedObject->Orientation.e21 = xXDir.y;
					SelectedObject->Orientation.e31 = xXDir.z;
					SelectedObject->Orientation.e12 = xYDir.x;
					SelectedObject->Orientation.e22 = xYDir.y;
					SelectedObject->Orientation.e32 = xYDir.z;
					SelectedObject->Orientation.e13 = xZDir.x;
					SelectedObject->Orientation.e23 = xZDir.y;
					SelectedObject->Orientation.e33 = xZDir.z;
				}
			}
			
			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				//Finish when the mouse is release
				DoingSnapToVertex = false;
			}
		}
		else
		{
			if(bDestValid && !Asura_UserInput::KeyPressed('z'))
			{
				//need to determine the position to snap to
				Asura_Vector_3 xSnapPos = xDestPos;

				
				if(eSnapDest == ObjectSnapDest_Cursor)
				{
					xSnapPos = xIntersection;

					//offset slightly away from the snap point
					//Mainly so that objects snapped to the ground are moved slightly up to avoid problems
					//RAR 12/12/2005
					if(!bToObject)
					{
						const float g_fOffset = 0.0005f;//half mm

						xSnapPos += xDestNormal*g_fOffset;
					}


				}
				else if(eSnapDest == ObjectSnapDest_Vertex)
				{
					//check each vertex in the destination mesh , and find the one closest to the intersection point
					if(pxDestObject)
					{
						xSnapPos = pxDestObject->GetSnapPoint(xIntersection);
					}
					else if(pxDestMesh)
					{
						float fLowest = ASURA_LOS_MAX_DISTANCE;

						for(int i=0;i<pxDestMesh->NumberOfVertices;i++)
						{
							Asura_Vector_3 xVertPos = pxDestMesh->VertexDataPtr[i].Position;
							xVertPos.x *= xDestScale.x;
							xVertPos.y *= xDestScale.y;
							xVertPos.z *= xDestScale.z;
							xVertPos = xDestPos + (xVertPos*xDestOrient);
							
							float d = (xVertPos - xIntersection).Magnitude();
							if(d<fLowest)
							{
								fLowest = d;
								xSnapPos = xVertPos;
							}
						}
					}
					else if(pxDestCollMesh)
					{
						float fLowest = ASURA_LOS_MAX_DISTANCE;

						for(u_int i=0;i<pxDestCollMesh->GetNumberOfVertices();i++)
						{
							Asura_Vector_3 xVertPos = pxDestCollMesh->GetVertex(i);
							xVertPos.x *= xDestScale.x;
							xVertPos.y *= xDestScale.y;
							xVertPos.z *= xDestScale.z;
							xVertPos = xDestPos + (xVertPos*xDestOrient);
							
							float d = (xVertPos - xIntersection).Magnitude();
							if(d<fLowest)
							{
								fLowest = d;
								xSnapPos = xVertPos;
							}
						}
					}
				}
				

				if(bAlign)
				{
					if(eSnapDest == ObjectSnapDest_Pivot || xDestNormal == Asura_Maths::ZeroVector3)
					{
						SelectedObject->Orientation = xDestOrient;
					}
					else
					{

						if(s_bSnapToVertex_AlignMode_AlignYAxis)
						{
							//rotate so that the (negative) y direction is aligned with the normal
							Asura_Vector_3 xNegYDir;
							xNegYDir.x = -SelectedObject->Orientation.e12;
							xNegYDir.y = -SelectedObject->Orientation.e22;
							xNegYDir.z = -SelectedObject->Orientation.e32;

							Asura_Quat xQuat;
							xQuat.RotationFromAToB(xNegYDir,xDestNormal);

							Asura_Matrix_3x3 xMat;
							xQuat.ToMatrix(xMat);

							SelectedObject->Orientation = xMat*SelectedObject->Orientation;
						}
						else
						{
							//generate a rotation matrix that has one of its axes in the direction of the normal
							Asura_Matrix_3x3 xMat,xMatInverse;
							Asura_Maths::OrientationFromDirection(xMat,xDestNormal);
					
							xMatInverse = xMat;
							xMatInverse.Transpose();

							//Rotate the matrix of the object being moved so that it is relative the
							//polygon normals matrix
							SelectedObject->Orientation = xMatInverse*SelectedObject->Orientation;

							//snap the rotated matrix to the nearest set of axes
							SelectedObject->SnapToNearestAxes(false);

							//and rotate back
							SelectedObject->Orientation = xMat*SelectedObject->Orientation;
						}
					}

				}

				if(s_iSnapToVertex_Rotate)
				{
					//snap to nearest axes
					SelectedObject->SnapToNearestAxes(false);

					//rotate by 90 degrees as needed
					for(int i=0;i<s_iSnapToVertex_Rotate;i++)
					{
						Asura_Matrix_3x3 xMat90(0,0,1,
												0,1,0,
												-1,0,0);

						SelectedObject->Orientation = SelectedObject->Orientation * xMat90;
						
					}

					if(bAlign && eSnapDest != ObjectSnapDest_Pivot && xDestNormal != Asura_Maths::ZeroVector3)
					{
						//align to surface again

						//generate a rotation matrix that has one of its axes in the direction of the normal
						Asura_Matrix_3x3 xMat,xMatInverse;
						Asura_Maths::OrientationFromDirection(xMat,xDestNormal);

						Asura_Matrix_3x3 xMat90(1,0,0,
												0,0,1,
												0,-1,0);
						xMat = xMat * xMat90;
					
						xMatInverse = xMat;
						xMatInverse.Transpose();

						//Rotate the matrix of the object being moved so that it is relative the
						//polygon normals matrix
						SelectedObject->Orientation = xMatInverse*SelectedObject->Orientation;

						//snap the rotated matrix to level
						SelectedObject->SnapToLevel(false);

						//and rotate back
						SelectedObject->Orientation = xMat*SelectedObject->Orientation;
					}
				}

				//Now we look to see which vertex in the object being moved is closest to the vertex
				//that we have just found
				Asura_Vector_3 xSourceSnapPos = SelectedObject->Position;
				
				if(s_bSnapToVertex_SourceVertexLocked)
				{
					//use the predefined source position
					xSourceSnapPos = s_xSnapToVertex_SourceVertex;
					//if aligning to surface ,need to take into account that object has been rotated
					if(bAlign)
					{
						Asura_Vector_3 xOffset = xSourceSnapPos - SnapToVertexPositionStore;
						xOffset = SnapToVertexOrientationStore.ApplyTranspose(xOffset);
						xOffset = xOffset * SelectedObject->Orientation;
						xSourceSnapPos = SnapToVertexPositionStore + xOffset;
					}
				}
				else if(eSnapSource == ObjectSnapSource_Vertex)
				{
					xSourceSnapPos = SelectedObject->GetSnapPoint(xSnapPos);
				}
				else if(eSnapSource == ObjectSnapSource_BoundingBox && xDestNormal != Asura_Maths::ZeroVector3)
				{
					//work out which side is most aligned to the polygon...
					//Get the polygon normal
					Asura_Vector_3 xNormal = xDestNormal;
					
					//rotate into the local space of the object being moved
					Asura_Matrix_3x3 xInvMat = SelectedObject->Orientation;
					xInvMat.Transpose();
					xNormal = xNormal*xInvMat;
	

					const Asura_Bounding_Box& xBB = s_xSnapToVertex_SourceBB;

					//determine which side of the boundary box we are interested in...
					Asura_Vector_3 xCentreOfSide = xBB.CalculateCentre();

					if(fabsf(xNormal.x) > fabsf(xNormal.y) && fabsf(xNormal.x) > fabsf(xNormal.z))
					{
						if(xNormal.x > 0)
							xCentreOfSide.x = xBB.MinX;
						else
							xCentreOfSide.x = xBB.MaxX;
						
					}
					else if(fabsf(xNormal.y) > fabsf(xNormal.z))
					{
						if(xNormal.y > 0)
							xCentreOfSide.y = xBB.MinY;
						else
							xCentreOfSide.y = xBB.MaxY;
					}
					else
					{
						if(xNormal.z > 0)
							xCentreOfSide.z = xBB.MinZ;
						else
							xCentreOfSide.z = xBB.MaxZ;
					}
					//take this position and rotate into world space
					xSourceSnapPos = xCentreOfSide;
					xSourceSnapPos = xSourceSnapPos * SelectedObject->Orientation;
					xSourceSnapPos += SelectedObject->Position;
				}

				SelectedObject->Position -= xSourceSnapPos;
				SelectedObject->Position += xSnapPos;

				s_xSnapToVertex_HighlightPosition = xSnapPos;

				//resolve object out of floor if needed
				if(SelectedObject->DoesObjectNeedToBeAbleToMove())
				{
					SelectedObject->ResolveOutOfFloor(2.0f,false);
				}
				
			}
			if (Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				//Lock in the current position , and move onto setting the orientation
				s_iSnapToVertex_SetOrientState = SetOrientState_Pending;
				SnapToVertexPositionStore = SelectedObject->Position;
				SnapToVertexOrientationStore = SelectedObject->Orientation;

				Asura_Cursor::GetScreenSpaceCoords(s_iSnapToVertex_DragStartX, s_iSnapToVertex_DragStartY);

			}
		}
	}
	
	SelectedObject->UpdatePositionInAsuraData();
	UpdateGroupMovement();
}




/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Tools_Object_System::SetSelectionFilter
// PARAMETERS	Enum of the object type to filter on
// RETURNS		None
// DESCRIPTION	Adds a new filter to the selection filter list (i.e. only objects of the
//				specified types will be selectable)
// AUTHOR		Richard M. Smith (21 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Tools_Object_System::SetSelectionFilter( int iToolsObjectType )
{
// Allocate a new array
	int*	piNewArray = new int[ s_iNumSelectionFilters + 1 ];

// Copy the old array contents across
	for( int i = 0; i < s_iNumSelectionFilters; ++i )
	{
		piNewArray[ i ] = s_piSelectionFilters[ i ];
	}

// Add the new entry
	piNewArray[ s_iNumSelectionFilters ] = iToolsObjectType;

// Delete the old array
	if( s_piSelectionFilters )
	{
		delete [] s_piSelectionFilters;
		s_piSelectionFilters = NULL;
	}

// Set the new one up
	s_piSelectionFilters = piNewArray;
	s_iNumSelectionFilters++;
	
} // End Function 'Tools_Object_System::SetSelectionFilter'






/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Tools_Object_System::ClearSelectionFilter
// PARAMETERS	None
// RETURNS		None
// DESCRIPTION	Removes all selection filters
// AUTHOR		Richard M. Smith (21 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
void Tools_Object_System::ClearSelectionFilter( void )
{
	if( s_piSelectionFilters )
	{
		delete [] s_piSelectionFilters;
		s_piSelectionFilters = NULL;
	}

	s_iNumSelectionFilters = 0;

} // End Function 'Tools_Object_System::ClearSelectionFilter'





/////////////////////////////////////////////////////////////////////////////////////////
// FUNCTION		Tools_Object_System::IsFilteredOut
// PARAMETERS	Type to check
// RETURNS		TRUE if the type is filtered out; FALSE otherwise
// DESCRIPTION	See returns
// AUTHOR		Richard M. Smith (21 August 2002)
/////////////////////////////////////////////////////////////////////////////////////////
bool Tools_Object_System::IsFilteredOut( const int iType )
{
	if( 0 == s_iNumSelectionFilters )
	{
	// There are no filters therefore nothing is filtered out...
		return false;
	}

// See if this type is one of the filtered types
	for( int i = 0; i < s_iNumSelectionFilters; i++ )
	{
		if( iType == s_piSelectionFilters[ i ] )
		{
		// This one is allowed
			return false;
		}
	}

	return true;

} // End Function 'Tools_Object_System::IsFilteredOut'




void Tools_Object_System::ChooseObject(const char* pszMessage,Tools_ChooseObject* pxChooseObject,Tools_Object* pxInitialSelection)
{
	if(s_pxChooseObject)
	{
		ChooseObject_Finish();
	}

	Asura_Assert(pxChooseObject,"");
	//stop any flying
	Tools_Movement::SetFlyingObject(NULL);
	//stop snap to vertex
	CancelSnapToVertex();
	//make sure not currently in the middle of a drag-select
	s_xSelectionState.Reset();



	s_pxChooseObject = pxChooseObject;
	s_pxChooseObject->ClearList();
	s_xChooseObjectMessage = pszMessage ? pszMessage : "Choose object";

	//hide anything that shouldn't be selected
	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		Tools_Object* pxObjectToCheck = xObjectLIF()->GetObjectToSelect();
		if(!s_pxChooseObject->CanBeSelected(pxObjectToCheck) || !pxObjectToCheck->IsSelectable())
		{
			xObjectLIF()->HiddenFlags |= ObjectHideFlag_ChooseObject;
		}
	}
	if(pxInitialSelection && s_pxChooseObject->CanBeSelected(pxInitialSelection))
	{
		s_pxChooseObject->SelectObject(pxInitialSelection,false);
	}
	s_uSelectionChangedStamp++;

	s_pxChooseObject->OnStart();


	s_pxChooseObject->HideOverlays(true);
}


void Tools_Object_System::ChooseObject_Accept()
{
	// (TI - 13.06.2007) - Sometimes we want nothing to be selected, i've fixed up where this was causing a crash
	// everything should be able to deal with nothing being selected.
	// if(!s_pxChosenObject) return; // HE - 23/05/2007 - don't do anything if no object is selected

	s_pxChooseObject->ObjectSelected(s_pxChooseObject->GetMainSelection());
	ChooseObject_Finish();
}

void Tools_Object_System::ChooseObject_Finish()
{
	if(!s_pxChooseObject) return;

	s_pxChooseObject->HideOverlays(false);
	
	Asura_Overlay* pxEntityManagerOverlay = Asura_Overlay_System::GetOverlay(TOOLS_OVERLAY_SECONDARYENTFILE);
	if(pxEntityManagerOverlay)
	{
		pxEntityManagerOverlay->SetOverrideOverlaysHiddenFlag(false);
	}


	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		xObjectLIF()->HiddenFlags &=~ ObjectHideFlag_ChooseObject;
	}
	s_pxChooseObject->ClearSelectionFlags();
	s_pxChooseObject->OnFinish();
	s_pxChooseObject = 0;

	//This will force the entity manager to update the displayed selection , to cope with the switch
	//from showing 'choose object' selection and normal selection.
	s_uSelectionChangedStamp++;
}


void Tools_Object_System::HandleMouseSelection()
{
	if (Asura_Interface::UsingMenusOrOverlays()) return;

	int x,y;
	Asura_Cursor::GetScreenSpaceCoords(x,y);
	s_xSelectionState.m_xDragCurrent.x = (float)x;
	s_xSelectionState.m_xDragCurrent.y = (float)y;

	{
		if(!IsChoosingObject() && !IsMessageEditing())
		{
			if(SelectedObject && SelectedObject->UsesLeftButton())
			{
				//selected object makes use of left mouse button for its own purposes
				s_xSelectionState.Reset();
				return;
			}
		}

		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1))
		{
			//mouse just pressed.
			s_xSelectionState.m_uSelectionFlags = 0;
			s_xSelectionState.m_bLeftClick = true;
			s_xSelectionState.m_bSelect = !Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT);
			s_xSelectionState.m_bMultiSelect = Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL);
			s_xSelectionState.m_xDragStart = s_xSelectionState.m_xDragCurrent;

			//Was the click on a draggable object?
			if(ObjectUnderCursorList.size())
			{
				Tools_Object* pxObject = ObjectUnderCursorList.first_entry();
				if(pxObject->IsDraggable())
				{
					//Start dragging this object then.
					//This does prevent the object from being selected in a normal manner , but I guess it
					//can cause itself to be selected if it wants to be in StartDrag()
					//RAR 07/09/2006
					pxObject->StartDrag(s_xSelectionState.m_xDragStart);
					s_xSelectionState.m_pxDragObject = pxObject;
				}
				else if(!pxObject->AllowSelectionUsingLMB())
				{
					s_xSelectionState.Reset();
					return;
				}
			}
		}
		else if(s_xSelectionState.m_bLeftClick)
		{
			if(!s_pxChooseObject || s_pxChooseObject->SupportMultipleSelection())
			{
				if(s_xSelectionState.m_pxDragObject)
				{
					s_xSelectionState.m_pxDragObject->UpdateDrag(s_xSelectionState.m_xDragCurrent);
				}
				else
				{
					//has the cursor moved enough for this to be a drag operation?
					if((s_xSelectionState.m_xDragCurrent-s_xSelectionState.m_xDragStart).Magnitude()>15.0f)
					{
						s_xSelectionState.m_bDragSelect = true;
					}
				}
			}

			if(!Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1))
			{
				//mouse released , do the selection
				if(s_xSelectionState.m_pxDragObject)
				{
					s_xSelectionState.m_pxDragObject->EndDrag(s_xSelectionState.m_xDragCurrent);
				}
				else if(s_xSelectionState.m_bDragSelect)
				{
					HandleDragSelect();
				}
				else
				{
					HandleNonDragSelect();
				}
				s_xSelectionState.Reset();

			}
		}
	}
}

void Tools_Object_System::HandleDragSelect()
{
	//get the min/max values for the rect
	Asura_Vector_2 xDragMin,xDragMax;
	xDragMin.x = min(s_xSelectionState.m_xDragStart.x,s_xSelectionState.m_xDragCurrent.x);
	xDragMin.y = min(s_xSelectionState.m_xDragStart.y,s_xSelectionState.m_xDragCurrent.y);
	xDragMax.x = max(s_xSelectionState.m_xDragStart.x,s_xSelectionState.m_xDragCurrent.x);
	xDragMax.y = max(s_xSelectionState.m_xDragStart.y,s_xSelectionState.m_xDragCurrent.y);

	//get a list of objects contained in the rect
	List<Tools_Object*> xSelectionList;

	for(LIF<Tools_Object*> xObjectLIF(&ObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		//ignore hidden objects
		if(xObjectLIF()->IsHidden()) continue;

		//ignore objects that are out of range
		if(!Tools_Render_Options::IsPointWithinViewDistance(xObjectLIF()->Position)) continue;
		

		//get the screen coordinates for the object
		Asura_Vector_3 xScreenPos = xObjectLIF()->Position;
		Asura_Camera::ProjectWorldCoordsToScreenSpace(xScreenPos, true);
		if(xScreenPos.z<=0) continue;

		//is the point in the bounding rect?
		if(xScreenPos.x>=xDragMin.x && xScreenPos.x<=xDragMax.x &&
		   xScreenPos.y>=xDragMin.y && xScreenPos.y<=xDragMax.y)
		{
			Tools_Object* pxObjectToSelect = xObjectLIF()->GetObjectToSelect();
			if(pxObjectToSelect->IsSelectable() && !xSelectionList.contains(pxObjectToSelect))
			{
				xSelectionList.add_entry(pxObjectToSelect);
			}
		}
	}

	//select/deselect the objects in the list
	if(s_xSelectionState.m_bSelect)
	{
		SelectObjects(xSelectionList,s_xSelectionState.m_bMultiSelect);
	}
	else
	{
		DeselectObjects(xSelectionList);
	}


}

void Tools_Object_System::HandleNonDragSelect()
{
	//select the object under the mouse
	Tools_Object* pxObjectToSelect = 0;
	if(ObjectUnderCursorList.size()>1)
	{
		//ask which object to select
		ListSelection_Overlay overlay("Select which object?");

		//add all the object names to the overlay
		for(LIF<Tools_Object*> olif(&ObjectUnderCursorList);!olif.done();olif.next())
		{
			if(!olif()->GetName()) continue;

			overlay.AddItem(olif()->GetName(),(int)olif());
		}
		if(overlay.GetNumItems())
		{
			if(overlay.DoModal())
			{
				pxObjectToSelect = (Tools_Object*)overlay.GetSelectionData();
			}
		}
		else
		{
			pxObjectToSelect = ObjectUnderCursorList.first_entry();
		}
	}
	else if(ObjectUnderCursorList.size()==1)
	{
		pxObjectToSelect = ObjectUnderCursorList.first_entry();
	}

	
	if(s_xSelectionState.m_bSelect)
	{
		SelectObject(pxObjectToSelect,s_xSelectionState.m_bMultiSelect);
	}
	else
	{
		DeselectObject(pxObjectToSelect);
	}
}


void Tools_Object_System::SetFindObjectList(const List<Tools_Object*>& xObjectList)
{
	s_pxLastFoundObject = NULL;
	s_xFindObjectList = xObjectList;

	if(xObjectList.size())
		Tools_Timed_Message_System::AddMessage(0xffffff,"%d object(s) found. Use Alt+< and Alt+> to cycle through them.",s_xFindObjectList.size());
	else
		Tools_Timed_Message_System::AddMessage(0xffffff,"No objects found");
}


void Tools_Object_System::SelectNextFoundObject()
{
	if(!s_xFindObjectList.size())
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"No objects in 'Find' list , so unable to select next one.");
		return;
	}
	if(s_pxChooseObject) return; //disallow normal object selection while doing 'choose object'

	int iIndex = GetCurrentIndexInFindList();
	iIndex++;
	if(iIndex >= s_xFindObjectList.size()) iIndex = 0;

	s_pxLastFoundObject = s_xFindObjectList[iIndex];
	GotoAndSelect(s_pxLastFoundObject);

}

void Tools_Object_System::SelectPreviousFoundObject()
{
	if(!s_xFindObjectList.size())
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"No objects in 'Find' list , so unable to select previous one.");
		return;
	}
	if(s_pxChooseObject) return; //disallow normal object selection while doing 'choose object'

	int iIndex = GetCurrentIndexInFindList();
	iIndex--;
	if(iIndex < 0) iIndex = s_xFindObjectList.size()-1;

	s_pxLastFoundObject = s_xFindObjectList[iIndex];
	GotoAndSelect(s_pxLastFoundObject);
}

int Tools_Object_System::GetCurrentIndexInFindList()
{
	int iSelectedIndex = -1;

	//find the indeces of the last found object
	for(int i=0;i<s_xFindObjectList.size();i++)
	{
		if(s_xFindObjectList[i] == s_pxLastFoundObject)
		{
			return i;
		}

		if(s_xFindObjectList[i] == SelectedObject)
		{
			iSelectedIndex = i;
		}
	}
	//the last found object is no longer in the list , return -1 so it should start again from
	//the beginning
	return -1;

}


void Tools_Object_System::ScaleSelectedPositions(float fScale)
{
	for(LIF<Tools_Object*> xSelectedLIF(&SelectedObjectList);!xSelectedLIF.done();xSelectedLIF.next())
	{
		if(xSelectedLIF()->CanBeMoved())
		{
			xSelectedLIF()->ScalePosition(fScale);
		}
	}

	//the relative positions will have changed
	UpdateRelativePositions();
}

void Tools_Object_System::ScaleSelectedPositions()
{
	GetNumber_Overlay xOverlay("Enter scale value for selected objects",1.0f,0.0f,1000.0f,"%.4f");
	if(xOverlay.DoModal())
	{
		float fScale = xOverlay.GetNumber();
		if(fScale != 1.0f)
		{
			ScaleSelectedPositions(fScale);
			Tools_Timed_Message_System::AddMessage(0xffffff,"Positions scaled by %f",fScale);

		}
	}
}

void Tools_Object_System::EnableUndo()
{
	if(!s_pxUndo->IsInitialised())
	{
		s_pxUndo->Initialise();
	}	
}

void Tools_Object_System::DisableUndo()
{
	if(s_pxUndo->IsInitialised())
	{
		s_pxUndo->Shutdown();
	}
}

void Tools_Object_System::AddCopyOverlayControls(Asura_Overlay* pxOverlay,Asura_Vector_2& xOffset)
{
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Function_Dialog("Duplicate Selected",xOffset,DuplicateSelected,0,IsSelectedCopiable);
	pxDialog->SetHelpText("Creates a copy of all selected objects (if they can be copied) , and puts them slightly offset from the originals.\nShorcut key : Ctrl+D");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Copy",xOffset,CopyToClipboard,0,IsSelectedCopiable);
	pxDialog->SetHelpText("Copies the selected object to the clipboard.\n(Note that the object properties will be taking at the time of pasting , not as they are at the time of copying.)\nShorcut key : Ctrl+C");
	pxOverlay->AddDialog(pxDialog);

	float x = xOffset.x;
	xOffset.x+= pxDialog->m_xMaxCoord.x + 1.0f;
	pxDialog = new Asura_Function_Dialog("Paste",xOffset,Paste,0,CanPaste);
	pxDialog->SetHelpText("Pastes a copy of the clipboard object at the current position.\nShortcut key : Ctrl+V");
	pxOverlay->AddDialog(pxDialog);
	xOffset.x = x;
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Paste Over",xOffset,PasteOver,0,CanPasteOver);
	pxDialog->SetHelpText("Pastes the properties of the copied object over the selected object.\nShortcut key : Ctrl+Shift+V");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();
}

void Tools_Object_System::DuplicateSelected()
{
	if(!SelectedObject) return;
	if(!SelectedObject->CanBeCopied())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"%s cannot be copied",SelectedObject->GetName());
		return;
	}

	//get the list of selected object
	List<Tools_Object*> xSelectedList = SelectedObjectList;
	//make sure the primary selection is at the end of the list.
	//this should ensure that the clone of this object should be the primary
	//selection once we are done
	if(xSelectedList.contains(SelectedObject))
	{
		xSelectedList.delete_entry(SelectedObject);
	}
	xSelectedList.add_entry(SelectedObject);

	List<Tools_Object*> xCreatedList;

	Tools_Object* pxCopy;

	Tools_SecondaryEntityFile_System::PreCreateObject();
	
	s_eCopyType = TOOLS_COPYOBJECT_DUPLICATE;
	for(LIF<Tools_Object*> xSelectedLIF(&xSelectedList);!xSelectedLIF.done();xSelectedLIF.next())
	{
		Tools_Object* pxToolsObject = xSelectedLIF();

		//make sure this object can be copied
		if(!pxToolsObject->CanBeCopied()) continue;

		// New method of duplicating objects makes use of the serialisation functionality
		// Ensure that in the tools object, CanBeCopied and UseStreamCopy return true.
		if (pxToolsObject->UseStreamCopy())
		{
			pxCopy = DuplicateToolsObject(pxToolsObject);
		}
		else	// old method of copying - lots of work done in each virtual Copy function
		{

			// Create a tools object (and associated entity) of the same type
			pxCopy = pxToolsObject->CreateObjectOfThisType();
			if(!pxCopy) continue;

			// Copy over it
			pxCopy->Copy(pxToolsObject);
		}

		//also set position and orientation
		pxCopy->Position = pxToolsObject->Position;
		pxCopy->Orientation = pxToolsObject->Orientation;
		pxCopy->UpdatePositionInAsuraData();

		//add the copy to the list of newly created objects
		xCreatedList.add_entry(pxCopy);
	}
	s_eCopyType = TOOLS_COPYOBJECT_NONE;

	Tools_SecondaryEntityFile_System::PostCreateObject();


	//select the newly created objects
	Tools_Object_System::SelectObjects(xCreatedList,false);

	Tools_Timed_Message_System::AddMessage(0xffffff,"%d objects duplicated",xCreatedList.size());

}

Tools_Object* Tools_Object_System::DuplicateToolsObject(Tools_Object* pxToolsObject)
{
	Asura_SaveBlock			xSaveBlock;

	Asura_Entity_Server*	pxSourceEntity		= Asura_Entity_Server::FindEntity(pxToolsObject->GetGuid());
	const Asura_Guid		uCreatedGuid		= Asura_Entity_System::GetCreatedByToolsGuid();
	const char*				pcEntityChunkStart	= static_cast<const char*>(xSaveBlock.GetCurrentPosition());
	const char*				szName				= pxToolsObject->GetName();

	// Create the tools data. Do this first as we want to avoid creating the entity if this fails.
	Tools_Entity_Data* pxToolsData = Tools_Entity_Object::CreateToolsDataWithDuplicateName(szName, szName, uCreatedGuid);

	if (!pxToolsData)
	{
		return NULL;
	}

	// Fill out our save-block with the current state of our source entity...
	Asura_Chunk_Entity::Create(pxSourceEntity, xSaveBlock);

	Asura_Chunk_Entity*	pxChunkEntity	= const_cast<Asura_Chunk_Entity*>(reinterpret_cast<const Asura_Chunk_Entity*>(pcEntityChunkStart));
	const char*			pcChunkBuffer	= pcEntityChunkStart + sizeof(*pxChunkEntity);

	// Patch the GUID in the chunk-entity (from the save-block) and process this
	// as if it were a brand new entity...
	pxChunkEntity->Guid	= uCreatedGuid;
	pxChunkEntity->Process(pcChunkBuffer);

	Asura_Entity_Server*	pxDestEntity	= Asura_Entity_Server::FindEntity(uCreatedGuid);

	// Convert message blocks
	Tools_EntityTemplateInstance_MessageBlocks xConvertMessageBlocks(pxToolsObject->GetGuid(), uCreatedGuid);
	pxDestEntity->ApplyChangesForTemplateInstance(&xConvertMessageBlocks);

	// Create the Tools object, set the entity tools data and find the tools object
	Tools_Entity_Object::CreateObject(pxDestEntity);

	delete pxDestEntity->ToolsData;

	pxDestEntity->ToolsData = pxToolsData;
	Tools_Object* pxCopy		= Tools_Entity_Object::FindEntity(uCreatedGuid);

	return pxCopy;
}

void Tools_Object_System::CopyToClipboard()
{
	s_xClipboardList.ClearList();

	for(LIF<Tools_Object*> xObjectLIF(&SelectedObjectList); !xObjectLIF.done(); xObjectLIF.next())
	{
		CopyToClipboard(xObjectLIF(), false);
	}
}

void Tools_Object_System::CopyToClipboard(Tools_Object* pxObject, bool bClearClipboard /* = true */)
{
	if(!pxObject) return;
	if(!pxObject->CanBeCopied())
	{
		Tools_Timed_Message_System::AddMessage(0xffffff, "%s cannot be copied", pxObject->GetName());
		return;
	}

	if(bClearClipboard)
	{
		s_xClipboardList.ClearList();
	}

	s_xClipboardList.AddToList(pxObject);

	Tools_Timed_Message_System::AddMessage(0xffffff, "%s copied", pxObject->GetName());
}


void Tools_Object_System::PasteOver()
{
	if(!CanPasteOver()) return;

	int iCopied = 0;
	s_eCopyType = TOOLS_COPYOBJECT_PASTEOVER;

	Tools_Object* pxObjectToCopy = s_xClipboardList.GetHeadObject();

	// Since we are deleting and re-creating to do the paste-over, we need a copy
	// of the selected object list, otherwise it will change while we are iterating through it!
	List<Tools_Object*> SelectedObjectListCopy = SelectedObjectList;
	for(LIF<Tools_Object*> xObjectLIF(&SelectedObjectListCopy);!xObjectLIF.done();xObjectLIF.next())
	{
		// Are we allowed to paste over this entity, and does the object still exist?
		if(CanPasteOver(xObjectLIF()) && ObjectList.contains(xObjectLIF()))
		{
			Tools_Object* pxToolsObject = xObjectLIF();

			// New method of pasting makes use of the serialisation functionality
			if (pxToolsObject->UseStreamCopy())
			{
				Asura_SaveBlock			xSaveBlock;
				const Asura_Guid		uDestGuid			= pxToolsObject->GetGuid();

				Asura_Entity_Server*	pxSourceEntity		= Asura_Entity_Server::FindEntity(pxObjectToCopy->GetGuid());
				Asura_Entity_Server*	pxDestEntity		= Asura_Entity_Server::FindEntity(pxToolsObject->GetGuid());

				const char*				pcEntityChunkStart	= static_cast<const char*>(xSaveBlock.GetCurrentPosition());

				// Remember where the tools entity data is so we can point the new entity to it
				Tools_Entity_Data* pxToolsEntityData = pxDestEntity->ToolsData;
				pxDestEntity->ToolsData = NULL;

				// Also remember the position and orientation so we can set them later
				Asura_Vector_3 xPosition = pxToolsObject->Position;
				Asura_Matrix_3x3 xOrientation = pxToolsObject->Orientation;

				// Now we can deselect the tools object we want to paste over and destroy it
				DeselectObject(pxToolsObject);
				pxToolsObject->Destroy();

				// Ensure that the deleted Guid is removed from all lists
				Asura_Entity_System::PruneEntityLists();

				// Fill out our save-block with the current state of our source entity...
				Asura_Chunk_Entity::Create(pxSourceEntity, xSaveBlock);

				Asura_Chunk_Entity*	pxChunkEntity	= const_cast<Asura_Chunk_Entity*>(reinterpret_cast<const Asura_Chunk_Entity*>(pcEntityChunkStart));
				const char*			pcChunkBuffer	= pcEntityChunkStart + sizeof(*pxChunkEntity);

				// Keep the same Guid as the entity we deleted
				pxChunkEntity->Guid	= uDestGuid;
				pxChunkEntity->Process(pcChunkBuffer);

				// Find the pasted entity
				Asura_Entity_Server*	pxPastedEntity	= Asura_Entity_Server::FindEntity(uDestGuid);

				// Convert message blocks
				Tools_EntityTemplateInstance_MessageBlocks xConvertMessageBlocks(pxObjectToCopy->GetGuid(), uDestGuid);
				pxPastedEntity->ApplyChangesForTemplateInstance(&xConvertMessageBlocks);

				// Point the new entity at the tools entity data we remembered the location of
				pxPastedEntity->ToolsData = pxToolsEntityData;

				// Create the Tools object, find it and select it
				Tools_Entity_Object::CreateObject(pxPastedEntity);
				Tools_Object* pxCopy = Tools_Entity_Object::FindEntity(uDestGuid);
				SelectObject(pxCopy, true);

				// Set the position
				pxCopy->Position = xPosition;
				pxCopy->Orientation = xOrientation;
				pxCopy->UpdatePositionInAsuraData();
			}
			else // old method where the work is done in each virtual Copy function
			{
				xObjectLIF()->Copy(pxObjectToCopy);
			}

			iCopied++;
		}
	}
	s_eCopyType = TOOLS_COPYOBJECT_NONE;

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::UpdateMessageEditorTable();
	#endif

	//deselect and select again to make sure overlay is up to date
	Tools_Object* pxObject = SelectedObject;
	DeselectObject(pxObject);
	SelectObject(pxObject,true);

	if(SelectedObjectList.size() == 1)
	{
		if(iCopied)
		{
			Tools_Timed_Message_System::AddMessage(0xffffff,"%s pasted over %s",pxObjectToCopy->GetName(),SelectedObject->GetName());
		}
	}
	else
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"%s pasted over %d objects",pxObjectToCopy->GetName(),iCopied);
	}
	CheckForMessageBlockOwnershipProblems();
}

void Tools_Object_System::Paste()
{
	if(!CanPaste()) return;

	s_eCopyType = TOOLS_COPYOBJECT_PASTENEW;

	Tools_Object_System::DeselectAllObjects();

	for (Asura_List_It<Tools_Object> xIter(s_xClipboardList); !xIter.Done(); xIter.Next())
	{
		Tools_Object* pxObjectToCopy = xIter.GetCurrent();

		// If the object to copy no longer exists or cannot be pasted then skip it.
		if ( !ObjectList.contains(pxObjectToCopy) || !pxObjectToCopy->CanBePasted() )
		{
			Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_ERROR,"Could not paste %s",pxObjectToCopy->GetName());
			continue;
		}

		Tools_SecondaryEntityFile_System::PreCreateObject();

		// new method of copying makes use of the serialisation functionality
		Tools_Object* pxCopy;
		if (pxObjectToCopy->UseStreamCopy())
		{
			//Duplicate the tools object
			pxCopy = DuplicateToolsObject(pxObjectToCopy);
		}
		else	// old method where lots of work is done in each virtual Copy function
		{
			//create a new object of the type being copied
			pxCopy = pxObjectToCopy->CreateObjectOfThisType();
			if(pxCopy)
			{
				//paste the properties over it
				pxCopy->Copy(pxObjectToCopy);
			}
		}

		if(!pxCopy)
		{
			//Need to call PostCreateObject() even though nothing was created to prevent a later assert. RAR 05/06/2008
			Tools_SecondaryEntityFile_System::PostCreateObject();
			continue;
		}

		//Copy orientation for the new object
		pxCopy->Orientation = pxObjectToCopy->Orientation;
		pxCopy->UpdatePositionInAsuraData();

	  	//assign to correct entity file
		Tools_SecondaryEntityFile_System::PostCreateObject();

		//Select the newlt created object
		Tools_Object_System::SelectObject(pxCopy, true);

		Tools_Timed_Message_System::AddMessage(Tools_Timed_Message_System::COLOUR_DEFAULT,"%s pasted",pxCopy->GetName());
	}

	s_eCopyType = TOOLS_COPYOBJECT_NONE;
	
	CheckForMessageBlockOwnershipProblems();
}


bool Tools_Object_System::IsSelectedCopiable()
{
	if(MessageBlockEditing || s_pxChooseObject) return false;
	if(!SelectedObject) return false;
	if(!SelectedObject->CanBeCopied()) return false;
	return true;
}


bool Tools_Object_System::CanPasteOver()
{
	if(MessageBlockEditing || s_pxChooseObject) return false;

	Tools_Object* pxPasteOverObject = s_xClipboardList.GetHeadObject();

	// We should have exactly one entry in the clipboard that hasn't been deleted and can be pasted over another object.
	if ( !pxPasteOverObject ||
		( s_xClipboardList.GetNumEntries() != 1 ) ||
		!ObjectList.contains( pxPasteOverObject ) ||
		!pxPasteOverObject->CanBePastedOverAnother() )
	{
		return false;
	}

	// Check that at least one of the selected objects can be pasted over.
	for(LIF<Tools_Object*> xObjectLIF(&SelectedObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		if(CanPasteOver(xObjectLIF()))
		{
			return true;
		}
	}

	return false;
}


bool Tools_Object_System::CanPasteOver(Tools_Object* pxObject)
{
	Tools_Object* pxPasteOverObject = s_xClipboardList.GetHeadObject();

	//there needs to be a selected object of the same type
	if(!pxObject) return false;
	if(pxObject->Type != pxPasteOverObject->Type) return false;
	//but it must be a different object
	if(pxObject == pxPasteOverObject) return false;

	return true;
}

bool Tools_Object_System::CanPaste()
{
	if(MessageBlockEditing || s_pxChooseObject) return false;

	// Determine whether there are any entries in the clipboard that can be pasted.
	int iNumClipboardEntries = s_xClipboardList.GetNumEntries();
	for(Asura_List_It<Tools_Object> xClipboardListIt(s_xClipboardList); !xClipboardListIt.Done(); xClipboardListIt.Next())
	{
		Tools_Object* pxClipboardObject = xClipboardListIt.GetCurrent();
		if (!ObjectList.contains(pxClipboardObject) || !pxClipboardObject->CanBePasted())
		{
			--iNumClipboardEntries;
		}
	}

	return ( iNumClipboardEntries > 0 );
}

void Tools_Object_System::MergeSelectedIntoModules()
{
	if(!SelectedObjectList.size())
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Please select the objects to be merged first.");
		return;
	}

	//first get confirmation on doing this.
	char szQuestion[200];
	if(SelectedObjectList.size() > 1)
	{
		Asura_PrintfToBfr(szQuestion,"Would you like to merge the meshes of the %d selected objects into the environment?\nThe objects will be destroyed if you do this.",200,true,SelectedObjectList.size());
	}
	else
	{
		Asura_PrintfToBfr(szQuestion,"Would you like to merge the mesh for '%s' into the environment?\nIt will be destroyed if you do this.",200,true,SelectedObjectList.first_entry()->GetName());
	}

	Tools_Overlay_MergeObjectToModule xOverlay(szQuestion);
	if(!xOverlay.DoModal()) return;
	
	
	if(xOverlay.ShouldMergeToNearest())
	{
		MergeSelectedIntoModules_ModuleSelected(-1);
	}
	else
	{
		Tools_Prompted_Selection::ChooseModule("Select module to merge with.", MergeSelectedIntoModules_ModuleSelected);
	}


}

void Tools_Object_System::MergeSelectedIntoModules_ModuleSelected(int iSelectedModule)
{
	//take a copy of the selected object list
	List<Tools_Object*> xMergeList = SelectedObjectList;

	int iNotDeletableCount = 0;
	int iNotInModuleCount = 0;
	int iNotMergeableCount = 0;
	int iMergeCount = 0;
	int iTooManyPolyCount = 0;
	for(LIF<Tools_Object*> xMergeLIF(&xMergeList);!xMergeLIF.done();xMergeLIF.next())
	{
		Tools_Object* pxObject = xMergeLIF();

		//make sure this object still exists(it may have been deleted as a side effect of
		//another object being deleted)
		if(!ObjectList.contains(pxObject))
		{
			continue;
		}

		//ignore objects that can't be destroyed
		if(!pxObject->CanBeDestroyed())
		{
			iNotDeletableCount++;
			continue;
		}

		//or that don't have a mesh or skin
		if(!pxObject->Mesh && !pxObject->GetAnimation())
		{
			iNotMergeableCount++;
			continue;
		}

		//disallow objects that use tools shapes
		if(pxObject->Mesh && strstr(pxObject->Mesh->Name.GetString(),"ToolsShape"))
		{
			iNotMergeableCount++;
			continue;
		}

		//find out which module the object is in
		int iModule = (iSelectedModule>=0) ? iSelectedModule : Asura_Module_System::GetModuleIndexFromPosition(pxObject->Position,-1);
		if(iModule == -1)
		{
			iNotInModuleCount++;
			continue;
		}

		//do the merge
		bool bMergeOkay = false;
		if(pxObject->GetAnimation())
		{
			Asura_Hierarchy_Skin* pxSkin = pxObject->GetAnimation()->GetSkin();
			if(pxSkin)
			{
				bMergeOkay = Tools_Mode_MeshEdit::MergeSkinIntoModule(iModule,pxSkin,pxObject->Orientation,pxObject->Position,ASURA_GUID_UNREGISTERED,0,pxObject->GetScale(),0,0);
			}
			else
			{
				iNotMergeableCount++;
				continue;
			}
		}
		else
		{
			bMergeOkay = Tools_Mode_MeshEdit::MergeObjectIntoModule(iModule,pxObject->Mesh,pxObject->Orientation,pxObject->Position,ASURA_GUID_UNREGISTERED,0,pxObject->GetScale(),0,0);
		}
		if(bMergeOkay)
		{
			pxObject->Destroy();
			iMergeCount++;
		}
		else
		{
			iTooManyPolyCount++;
		}
	}

	
	Tools_Timed_Message_System::AddMessage(0xffffff,"%d objects merged into the environment",iMergeCount);

	if(iNotDeletableCount)
	{
		Tools_Timed_Message_System::AddMessage(0xffff00,"%d objects skipped because they can't be destroyed",iNotDeletableCount);
	}
	if(iNotInModuleCount)
	{
		Tools_Timed_Message_System::AddMessage(0xffff00,"%d objects skipped because they aren't in any module",iNotInModuleCount);
	}
	if(iNotMergeableCount)
	{
		Tools_Timed_Message_System::AddMessage(0xffff00,"%d objects skipped because they aren't allowed to be merged.",iNotMergeableCount);
	}
	if(iTooManyPolyCount)
	{
		Tools_Timed_Message_System::AddMessage(0xffff00,"%d objects skipped because it would have resulted in meshes with too many verts/polys.",iTooManyPolyCount);
	}

}


void Tools_Object_System::SetTextRenderPos(float x,float y)
{
	s_xTextPos.x = x;
	s_xTextPos.y = y;
}

void Tools_Object_System::ResetTextRenderPos()
{
	s_xTextPos.x = 1.0f;
	s_xTextPos.y = 3.0f;
}

void Tools_Object_System::SetCommentText(const char* pszText,void* pOwner)
{
	if(!pszText || !*pszText)
	{
		//if the text is blank , just clear the comment text
		ClearCommentText(pOwner);
		return;
	}

	//delete existing comment text
	delete [] s_pszCommentText;
	s_pszCommentText = 0;


	//copy into unicode buffer
	int iLength = strlen(pszText);
	s_pszCommentText = new Asura_UnicodeChar[iLength+12];
	Asura_PrintfToBfr(s_pszCommentText,L"Comment : %S",iLength+12,true,pszText);

	s_pCommentOwner = pOwner;
}

void Tools_Object_System::ClearCommentText(void* pOwner)
{
	if(pOwner == 0 || pOwner == s_pCommentOwner)
	{
		delete [] s_pszCommentText;
		s_pszCommentText = 0;
		s_pCommentOwner = 0;
		
	}
}

void Tools_Object_System::RenderComment()
{
	if(!s_pszCommentText) return;

	Asura_Handle iFont = Asura_Text_Output::GetSmallFontHandle();
	if(iFont == ASURA_INVALID_HANDLE) return;


	u_int uTextColour = 0xffffffff;
	u_int uFormat = ASURA_TEXTFORMAT_WORDWRAP|ASURA_TEXTFORMAT_RELATIVETORECT;

	float fScreenWidth = (float)Asura_Viewport::GetWidth();
	float fScreenHeight = (float)Asura_Viewport::GetHeight();


	//first determine the size of box needed for the text
	Asura_Bounding_Rectangle xRect;
	xRect.MinX = 0;
	xRect.MaxX = 350.0f;
	xRect.MinY = 0;
	xRect.MaxY = fScreenHeight;

	Asura_Font_System::SetFont(iFont);
	Asura_Font_System::PrintLongString(s_pszCommentText,&xRect,0,0,1.0f,1.0f,uTextColour,uFormat | ASURA_TEXTFORMAT_DONTRENDER);

	//shrink the box so just large enough for the text
	xRect.MaxY = Asura_Font_System::GetNextLinesYCoordinate()+2.0f;
	if(Asura_Font_System::GetNextLinesYCoordinate() <= Asura_Font_System::GetFontHeight())
	{
		xRect.MaxX = Asura_Font_System::GetLastPrintedXCoordinate();
	}

	//reposition box on side of screen
	xRect.MinX += fScreenWidth-xRect.MaxX-10.0f;
	xRect.MaxX = fScreenWidth;
	xRect.MinY += 30.0f;
	xRect.MaxY += 30.0f;


	
	//draw a background rectangle for the text
	Asura_Render::CurrentStates.RequestTextureOpMode(ASURA_TEXTUREOP_SINGLE_PASS);
	Asura_Render::CurrentStates.RequestZBufferWriteEnabled(true);
	
	Asura_Render::Overlay.Colour = 0x7f000000;
	Asura_Render::Overlay.TextureIndex = ASURA_TEXTURE_NONE;
	Asura_Render::Overlay.TranslucencyMode = ASURA_TRANSLUCENCY_NORMAL;
	
	Asura_Render::Overlay.SetRectXY(xRect.MinX-2.0f,xRect.MinY,(xRect.MaxX-xRect.MinX)+2.0f,(xRect.MaxY-xRect.MinY));
	Asura_Render::Overlay.Render();
	Asura_Render::FlushDisplay();

	xRect.MaxX -= 10.0f;
	//and draw the text
	Asura_Font_System::PrintLongString(s_pszCommentText,&xRect,0,0,1.0f,1.0f,uTextColour,uFormat);
	Asura_Font_System::PrintLongString(L"Comment :",&xRect,0,0,1.0f,1.0f,0xffffff7f,uFormat);
}

bool Tools_Object_System::GetParameterFromSelectedObjects(Asura_Hash_ID uParamID,class Tools_Parameter* pxParam,bool* pbMixed)
{
	bool bAnyGet = false;
	pxParam->m_bChanged = false;
	int iChangedCount = 0;

	for(LIF<Tools_Object*>  xObjectLIF(&SelectedObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		if(xObjectLIF() == SelectedObject) continue;//primary selection will be done last
		if(xObjectLIF()->GetParameter(uParamID,pxParam))
		{
			//Keep track of how many times the value changes.
			//If this is the first time the value was got , then count as changed , even
			//if it wasn't changed from the default value
			if(pxParam->m_bChanged || !bAnyGet)
			{
				iChangedCount++;
				pxParam->m_bChanged = false;
			}	
			bAnyGet = true;
		}
	}
	if(SelectedObject)
	{
		//do primary selection , so the param wil take this value in preference
		if(SelectedObject->GetParameter(uParamID,pxParam))
		{
			if(pxParam->m_bChanged || !bAnyGet)
			{
				iChangedCount++;
				pxParam->m_bChanged = false;
			}	
			bAnyGet = true;
		}
	}

	if(pbMixed) *pbMixed = (iChangedCount>1);

	return bAnyGet;
}

bool Tools_Object_System::SetParameterForSelectedObjects(Asura_Hash_ID uParamID,const class Tools_Parameter* pxParam)
{
	bool bAnySet = false;
	for(LIF<Tools_Object*>  xObjectLIF(&SelectedObjectList);!xObjectLIF.done();xObjectLIF.next())
	{
		if(xObjectLIF()->SetParameter(uParamID,pxParam)) bAnySet = true;
	}
	return bAnySet;
}


void Tools_Object_System::ObjectDeleted(Tools_Object* pxObject)
{
	//Remove the deleted object from various lists
	ObjectList.delete_entry(pxObject);

	if(ObjectUnderCursorList.contains(pxObject))
	{
		ObjectUnderCursorList.delete_entry(pxObject);
	}

	if(s_xFindObjectList.contains(pxObject))
	{
		s_xFindObjectList.delete_entry(pxObject);
	}

	if(s_xSelectionState.m_pxDragObject == pxObject) s_xSelectionState.Reset();

	Tools_Gizmo_Handler::ObjectDeleted(pxObject);

	#ifdef ASURA_TOOLS_USE_NEW_MESSAGE_EDITOR
	Tools_MessageEditor::HandleDeletedObject(pxObject);
	#endif


	Tools_SecondaryEntityFile_System::RemoveFromHistoryList(pxObject);

}


void Tools_Object_System::AddMissionFileTickbox(Asura_Overlay* pxOverlay, Asura_Vector_2& xOffset)
{
	if(!Tools_SaveAsura::IsEditingMission()) return;

	Asura_Dialog* pxDialog = new Asura_Tickbox_Dialog("Create In Mission",xOffset,&s_bCreateNewObjectsInMissionFile);	
	pxDialog->SetHelpText("Determines whether new lights,coronas,phonons and env hooks should be created in the mission or the common files.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

}

bool Tools_Object_System::ShouldCreateNewObjectsInMissionFile()
{
	return s_bCreateNewObjectsInMissionFile && Tools_SaveAsura::IsEditingMission();
}

int Tools_Object_System::FindDistantIconTexture(const char* szName)
{
	if(!szName || !*szName) return ASURA_TEXTURE_NONE;

	Asura_Hash_ID uHash = Asura_GetHashID(szName);

	//Search for an existing entry for this
	bool bAlreadyExists  = false;
	int* piTextureID = s_xDistantIconTree.AddOrFindNode(uHash, &bAlreadyExists);

	if(bAlreadyExists)
	{
		return *piTextureID;
	}

	//Haven't done this one yet , so check for the file as a tga and dds , both locally and on the network.
	//Doing it this way so as to avoid being spammed by errors from the texture management system for textures
	//that don't exist
	char acFilename[MAX_PATH];
	Asura_PrintfToBfr(acFilename, "graphics\\Tools\\ObjectIcons\\%s.tga", sizeof(acFilename), true, szName);
	if(!Tools_FileSystem::FileExists_CheckNetwork(acFilename))
	{
		//try as .dds
		Asura_PrintfToBfr(acFilename, "graphics\\Tools\\ObjectIcons\\%s.dds", sizeof(acFilename), true, szName);
		if(!Tools_FileSystem::FileExists_CheckNetwork(acFilename))
		{
			//No texture for this then
			*piTextureID = ASURA_TEXTURE_NONE;			
			return *piTextureID;
		}
	}

	//Texture appears to exist , so add to the texture management system
	*piTextureID = Asura_Texture_Management::AddTexture(&acFilename[8]);//(the 8 is there to skip past 'graphics' in the filename)
	Asura_Texture_Management::LoadTextures();
	return *piTextureID;
}


//////////////////////////////////////////////////////////////////////////////////////////////////
void ObjectSystemUndo::PreUndoProcessing()
{	
	//FR took out the cancelling of any flying thing. 
	//It does not seems to be necessary and is less confusing

	//stop snap to vertex
	Tools_Object_System::CancelSnapToVertex();
}



void ObjectSelectionUndo::ProcessUndoData(void* pUndo)
{
	ObjectSelectionUndoData* pxUndoData = (ObjectSelectionUndoData*) pUndo;
	Tools_Object* pxObject = Tools_Object_System::GetObjectFromToolsID(pxUndoData->m_iObjectID);

	switch(pxUndoData->m_eUndoType)
	{
		case ObjectSelectionUndoData::UndoType_SelectedPrimary :
			if(pxObject)
			{
				bool bInList = Tools_Object_System::SelectedObjectList.contains(pxObject);
				Tools_Object_System::DeselectObject(pxObject);
				if(bInList && !pxObject->IsGroupSelected())
				{
					//only wanted to remove as the primary selection , not from the selected object list
					//so add back in
					Tools_Object_System::SelectedObjectList.add_entry(pxObject);
					pxObject->SetGroupSelected(true);
				}
			}

			pxUndoData->m_eUndoType = ObjectSelectionUndoData::UndoType_DeselectedPrimary;
			break;

		case ObjectSelectionUndoData::UndoType_DeselectedPrimary :
			if(pxObject)
			{
				Tools_Object_System::SelectObject(pxObject,true);
			}
			pxUndoData->m_eUndoType = ObjectSelectionUndoData::UndoType_SelectedPrimary;
			break;

		case ObjectSelectionUndoData::UndoType_Selected :
			if(pxObject && Tools_Object_System::SelectedObjectList.contains(pxObject))
			{
				pxObject->OnGroupDeselection();
				pxObject->SetGroupSelected(false);
				Tools_Object_System::SelectedObjectList.delete_entry(pxObject);
			}

			pxUndoData->m_eUndoType = ObjectSelectionUndoData::UndoType_Deselected;
			break;

		case ObjectSelectionUndoData::UndoType_Deselected :
			if(pxObject && !Tools_Object_System::SelectedObjectList.contains(pxObject))
			{
				Tools_Object_System::SelectedObjectList.add_entry(pxObject);
				pxObject->SetGroupSelected(true);
			}
			pxUndoData->m_eUndoType = ObjectSelectionUndoData::UndoType_Selected;
			break;
	}
}


void ObjectSelectionUndo::LogObjectSelectedPrimary(const Tools_Object* pxObject)
{
	if(Tools_Object_System::IsChoosingObject()) return;
	ObjectSelectionUndoData* pxUndo = (ObjectSelectionUndoData*) AllocateNextUndoData();
	if(!pxUndo) return;

	pxUndo->m_eUndoType = ObjectSelectionUndoData::UndoType_SelectedPrimary;
	pxUndo->m_iObjectID = pxObject->GetToolsID();

}

void ObjectSelectionUndo::LogObjectDeselectedPrimary(const Tools_Object* pxObject)
{
	if(Tools_Object_System::IsChoosingObject()) return;
	ObjectSelectionUndoData* pxUndo = (ObjectSelectionUndoData*) AllocateNextUndoData();
	if(!pxUndo) return;

	pxUndo->m_eUndoType = ObjectSelectionUndoData::UndoType_DeselectedPrimary;
	pxUndo->m_iObjectID = pxObject->GetToolsID();
}


void ObjectSelectionUndo::LogObjectSelected(const Tools_Object* pxObject)
{
	if(Tools_Object_System::IsChoosingObject()) return;
	ObjectSelectionUndoData* pxUndo = (ObjectSelectionUndoData*) AllocateNextUndoData();
	if(!pxUndo) return;

	pxUndo->m_eUndoType = ObjectSelectionUndoData::UndoType_Selected;
	pxUndo->m_iObjectID = pxObject->GetToolsID();

}

void ObjectSelectionUndo::LogObjectDeselected(const Tools_Object* pxObject)
{
	if(Tools_Object_System::IsChoosingObject()) return;
	ObjectSelectionUndoData* pxUndo = (ObjectSelectionUndoData*) AllocateNextUndoData();
	if(!pxUndo) return;

	pxUndo->m_eUndoType = ObjectSelectionUndoData::UndoType_Deselected;
	pxUndo->m_iObjectID = pxObject->GetToolsID();
}

/////////////////////////////////////////////////////////////////////////////////////////
/*----------------------------------**
** Tools_ObjectSystem_ParamRegistry **
**----------------------------------*/

//Probably ought to go into Tools_ObjectSystem_Param.cpp , but I'm not sure I can be bothered to create
//a new file just for this,
Asura_TemplateHashTree<Tools_ObjectSystem_Param>* Tools_ObjectSystem_ParamRegistry::s_pxTree;

void Tools_ObjectSystem_ParamRegistry::RegisterParam(Asura_Hash_ID uParamID, const char* szName, int iType)
{
	static Asura_TemplateHashTree<Tools_ObjectSystem_Param> s_xTree;
	s_pxTree = &s_xTree;

	Asura_Hash_ID uHashID = Asura_GetHashID(szName);

	if(!s_xTree.GetDataByHash(uHashID))
	{
		Tools_ObjectSystem_Param* pxParam = s_xTree.AddNode(uHashID);
		pxParam->m_uParamID = uParamID;
		pxParam->m_uHashID = uHashID;
		pxParam->m_szName = szName;
		pxParam->m_iType = iType;
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------**
** Tools_ObjectSystemParam_Dialog **
**--------------------------------*/

//For selecting one of the object parameters from a drop down list.
class Tools_ObjectSystemParam_Dialog : public Asura_Hash_DropDownList_Dialog
{
public :
	Tools_ObjectSystemParam_Dialog(Asura_Vector_2& xOffset,int iNumberOfLines,Asura_Hash_ID* puHashID, bool* pbChangedFlag, const char* szName)
	:Asura_Hash_DropDownList_Dialog(xOffset)
	{
		Initialise(xOffset,iNumberOfLines,puHashID,pbChangedFlag,szName);
		SetOnlyUpdateSelectionIndexOnClick(true);
	}
protected :
	virtual u_int GetNumberOfHashValues() const 
	{
		return Tools_ObjectSystem_ParamRegistry::GetNumParam();
	}
	virtual const char* GetNameByIndex(int iIndex) const
	{
		const Tools_ObjectSystem_Param* pxParam = Tools_ObjectSystem_ParamRegistry::GetParamByIndex(iIndex);
		return pxParam ? pxParam->m_szName : "*Error*";
	}

	virtual const char* GetZeroHashText() const {return "*None*";}
};


//////////////////////////////////////////////////////////////////////////////////////////////

int ObjectSelection_Overlay::s_iLastFilterType = ObjectSelection_Overlay::FILTER_OBJECTTYPE;
int ObjectSelection_Overlay::s_iLastTypeSelected = TOT_Anon;
Asura_String ObjectSelection_Overlay::s_xLastShapeSelected = "";
Asura_String ObjectSelection_Overlay::s_xLastBlueprintTypeSelected = "";
Asura_String ObjectSelection_Overlay::s_xLastBlueprintSelected = "";

Asura_Hash_ID ObjectSelection_Overlay::s_uFilterParam = 0;
const Tools_ObjectSystem_Param* ObjectSelection_Overlay::s_pxFilterParam;
bool ObjectSelection_Overlay::s_bFilterParam_Bool;
int ObjectSelection_Overlay::s_iFilterParam_MinInt;
int ObjectSelection_Overlay::s_iFilterParam_MaxInt;
u_int ObjectSelection_Overlay::s_uFilterParam_MinUInt;
u_int ObjectSelection_Overlay::s_uFilterParam_MaxUInt;
float ObjectSelection_Overlay::s_fFilterParam_MinFloat;
float ObjectSelection_Overlay::s_fFilterParam_MaxFloat;
Asura_String ObjectSelection_Overlay::s_xFilterParam_String;


ObjectSelection_Overlay::ObjectSelection_Overlay(Asura_String xTitle,const List<Tools_Object*>* pxObjectList,bool bIncludeHidden)
:ListSelection_Overlay(xTitle)
{
	m_pxFilterDialog = 0;
	m_pxObjectList = pxObjectList;
	m_bIncludeHidden = bIncludeHidden;
	m_bOnlyShowSelectable = true;

	m_ppszFilterStrings = 0;

	m_iFilterType = s_iLastFilterType;
	m_iFilterIndex = 0;

	//forces the lists to updated on the first frame
	m_bFilterTypeChanged = true;
	m_bFilterIndexChanged = true;
	m_bFilterParamChanged = true;
	m_bFilterParamMinChanged = false;
	m_bFilterParamMaxChanged = false;

	s_pxFilterParam = Tools_ObjectSystem_ParamRegistry::GetParamByHash(s_uFilterParam);
	

	//add all the object names to the overlay
	for(CLIF<Tools_Object*> olif(pxObjectList);!olif.done();olif.next())
	{
		if(olif()->Type == TOT_Camera) continue;
		if(!m_bIncludeHidden && olif()->IsHidden()) continue;

		AddItem(olif()->GetName(),(int)olif());

	}
}

ObjectSelection_Overlay::~ObjectSelection_Overlay()
{
	DeleteFilterList();
}

void ObjectSelection_Overlay::DeleteFilterList()
{
	while(m_xFilterList.size())
	{
		delete m_xFilterList.first_entry();
		m_xFilterList.delete_first_entry();
	}

	delete [] m_ppszFilterStrings;
	m_ppszFilterStrings = 0;
}

void ObjectSelection_Overlay::UpdateFilterList()
{
	//clear out old strings
	DeleteFilterList();

	//add special entry to filter list
	AddToFilterList("*All*",0,false);

	
	//get list of object types/skins according to filter type
	for(CLIF<Tools_Object*> olif(m_pxObjectList);!olif.done();olif.next())
	{
		if(olif()->Type == TOT_Camera) continue;
		if(!m_bIncludeHidden && olif()->IsHidden()) continue;

		switch(m_iFilterType)
		{
			case FILTER_OBJECTTYPE :
				AddToFilterList(olif()->GetTypeDescriptionForObjectType(),olif()->Type,false);
				break;

			case FILTER_SHAPE :
				AddToFilterList(olif()->GetShapeOrSkinName().GetString(),0,true);
				break;

            case FILTER_BLUEPRINTTYPE :
                AddToFilterList(olif()->GetBlueprintTypeName().GetString(),0,true);
                break;

            case FILTER_BLUEPRINT :
                AddToFilterList(olif()->GetBlueprintName().GetString(),0,true);
                break;

			default :
				Asura_Assert(0,"Unexpected filter type");
				break;
		}
	}

	//build array of strings
	m_ppszFilterStrings = new char*[m_xFilterList.size()];
	for(int i=0;i<m_xFilterList.size();i++)
	{
		m_ppszFilterStrings[i] = m_xFilterList[i]->xString.GetString();
	}

	//determine initial selection
	m_iFilterIndex = 0;
	m_bFilterIndexChanged = true;
	for(int i=0;i<m_xFilterList.size();i++)
	{
		bool bMatch = false;
		switch(m_iFilterType)
		{
			case FILTER_OBJECTTYPE :
				bMatch = (m_xFilterList[i]->iData == s_iLastTypeSelected);
				break;

			case FILTER_SHAPE :
				bMatch = !s_xLastShapeSelected.Compare(m_xFilterList[i]->xString);
				break;

            case FILTER_BLUEPRINTTYPE :
                bMatch = !s_xLastBlueprintTypeSelected.Compare(m_xFilterList[i]->xString);
                break;

            case FILTER_BLUEPRINT :
                bMatch = !s_xLastBlueprintSelected.Compare(m_xFilterList[i]->xString);
                break;

			default :
				Asura_Assert(0,"Unexpected filter type");
				break;
		}
		if(bMatch)
		{
			m_iFilterIndex = i;
			break;
		}

	}

	//update list dialog
	m_pxFilterDialog->ReplaceStrings(m_xFilterList.size(),m_ppszFilterStrings);	
	
}

void ObjectSelection_Overlay::AddToFilterList(const char* pszName,int iData,bool bMatchByName)
{
	if(!pszName || !strlen(pszName)) return;

	//check for this type already being added
	LIF<ListSelectionItem*> xFilterLIF(&m_xFilterList);

	for(xFilterLIF.restart();!xFilterLIF.done();xFilterLIF.next())
	{
		if(bMatchByName)
		{
			if(!stricmp(pszName,xFilterLIF()->xString.GetString())) return;
		}
		else
		{
			if(iData == xFilterLIF()->iData) return;
		}
	}


	//create a ListSelectionItem
	ListSelectionItem* pxItem = new ListSelectionItem;
	pxItem->xString = pszName;
	pxItem->iData = iData;
	pxItem->bSelectedInitially = false;
	pxItem->bHidden = false;

	//add it into the list (sorted)

	for(xFilterLIF.restart();!xFilterLIF.done();xFilterLIF.next())
	{
		if(pxItem->xString.Compare(xFilterLIF()->xString,true)<=0)
		{
			m_xFilterList.add_entry_before(pxItem,xFilterLIF());
			break;
		}
	}
	if(xFilterLIF.done())
	{
		m_xFilterList.add_entry(pxItem);
	}
}



void ObjectSelection_Overlay::AddExtraOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset)
{
	Asura_Dialog* pxDialog;

	//Add drop down for selecting a parameter to filter the list by
	pxDialog = new Asura_Dialog( "Filter by Param :", xOffset);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Tools_ObjectSystemParam_Dialog(xOffset, 12, &s_uFilterParam, &m_bFilterParamChanged, NULL );
	pxDialog->SetHelpText("Allows list of objects to be filtered by the value of a specified parameter");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	//Add controls for specfifying the filter value for each of the types of parameter
	//bool
	pxDialog = new Asura_Tickbox_Dialog("Value", xOffset, &s_bFilterParam_Bool, &m_bFilterParamChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it matches this value");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamBool,0);
	pxOverlay->AddDialog(pxDialog);

	//int	
	pxDialog = new Asura_Numerical_Dialog("Min", xOffset, &s_iFilterParam_MinInt, -999999, 999999, 1, "%d",&m_bFilterParamMinChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it lies within this range");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamInt,0);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Max", xOffset, &s_iFilterParam_MaxInt, -999999, 999999, 1, "%d",&m_bFilterParamMaxChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it lies within this range");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamInt,0);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y -= Asura_Dialog::DefaultSpacing();

	//u_int
	pxDialog = new Asura_Numerical_Dialog("Min", xOffset, &s_uFilterParam_MinUInt, 0, 9999999, 1, "%d",&m_bFilterParamMinChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it lies within this range");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamUInt,0);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Max", xOffset, &s_uFilterParam_MaxUInt, 0, 9999999, 1, "%d",&m_bFilterParamMaxChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it lies within this range");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamUInt,0);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y -= Asura_Dialog::DefaultSpacing();

	//float
	pxDialog = new Asura_Numerical_Dialog("Min", xOffset, &s_fFilterParam_MinFloat, -999999.0f, 999999.0f, 1, "%.3f",&m_bFilterParamMinChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it lies within this range");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamFloat,0);
	pxOverlay->AddDialog(pxDialog);					   
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Max", xOffset, &s_fFilterParam_MaxFloat, -999999.0f, 999999.0f, 1, "%.3f",&m_bFilterParamMaxChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it lies within this range");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamFloat,0);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y -= Asura_Dialog::DefaultSpacing();

	//asura_string
	pxDialog = new Asura_String_Dialog("Value", xOffset, &s_xFilterParam_String, 20, &m_bFilterParamChanged);
	pxDialog->SetHelpText("Objects will appear in the list if they have the parameter , and it starts with this string");
	pxDialog->SetIsHiddenFunction(IsNotFilterParamString,0);
	pxOverlay->AddDialog(pxDialog);

	xOffset.y += Asura_Dialog::DefaultSpacing();
	xOffset.y += Asura_Dialog::LargeSpacing();


	//add filter types
	static const char* s_apszFilterNames [FILTER_NUMTYPES] =
	{
		"Type",
		"Shape/Skin",
        "Blueprint Type",
        "Blueprint"
	};
	

	pxDialog = new Asura_DropDown_List_Dialog(xOffset,s_apszFilterNames,&m_iFilterType,FILTER_NUMTYPES,FILTER_NUMTYPES,NULL,&m_bFilterTypeChanged,"Filter By");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();


	//add controls for the list of object types

	pxDialog = new Asura_Dialog("Filter List :",xOffset);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::SmallSpacing();

	static const char* s_apszDummyList[1] = {"*All*"};

	//filter strings will be filled in properly later
	pxDialog = m_pxFilterDialog = new Asura_List_Dialog(xOffset,s_apszDummyList,&m_iFilterIndex,10,1,NULL,&m_bFilterIndexChanged);
	pxDialog->SetHelpText("Select type to filter main object list by.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y=pxDialog->m_xMaxCoord.y;
	xOffset.y+=1.0f;


}

void ObjectSelection_Overlay::Update()
{
	bool bNeedUpdateVisibility = false;
	
	if(m_bFilterTypeChanged)
	{
		UpdateFilterList();
		s_iLastFilterType = m_iFilterType;
		m_bFilterTypeChanged = false;
	}

	if(m_bFilterParamMinChanged)
	{
		s_iFilterParam_MaxInt = Asura_Maths::Max(s_iFilterParam_MaxInt, s_iFilterParam_MinInt);
		s_uFilterParam_MaxUInt = Asura_Maths::Max(s_uFilterParam_MaxUInt, s_uFilterParam_MinUInt);
		s_fFilterParam_MaxFloat = Asura_Maths::Max(s_fFilterParam_MaxFloat, s_fFilterParam_MinFloat);
		m_bFilterParamMinChanged = false;
		m_bFilterParamChanged = true;
	}

	if(m_bFilterParamMaxChanged)
	{
		s_iFilterParam_MinInt = Asura_Maths::Min(s_iFilterParam_MaxInt, s_iFilterParam_MinInt);
		s_uFilterParam_MinUInt = Asura_Maths::Min(s_uFilterParam_MaxUInt, s_uFilterParam_MinUInt);
		s_fFilterParam_MinFloat = Asura_Maths::Min(s_fFilterParam_MaxFloat, s_fFilterParam_MinFloat);
		m_bFilterParamMaxChanged = false;
		m_bFilterParamChanged = true;
	}


	if(m_bFilterParamChanged)
	{
		s_pxFilterParam = Tools_ObjectSystem_ParamRegistry::GetParamByHash(s_uFilterParam);
		bNeedUpdateVisibility = true;
		m_bFilterParamChanged = false;
	}
	
	if(m_bFilterIndexChanged)
	{
		bNeedUpdateVisibility = true;
		//update last filter type
		switch(m_iFilterType)
		{
			case FILTER_OBJECTTYPE :
				s_iLastTypeSelected = m_xFilterList[m_iFilterIndex]->iData;
				break;
			case FILTER_SHAPE :
				s_xLastShapeSelected = m_xFilterList[m_iFilterIndex]->xString;
				break;
            case FILTER_BLUEPRINTTYPE :
                s_xLastBlueprintTypeSelected = m_xFilterList[m_iFilterIndex]->xString;
                break;
            case FILTER_BLUEPRINT :
                s_xLastBlueprintSelected = m_xFilterList[m_iFilterIndex]->xString;
                break;
			default :
				Asura_Assert(0,"Unexpected filter type");
				break;
		}
		m_bFilterIndexChanged = false;
	}

	if(bNeedUpdateVisibility)
	{
		UpdateVisibilities();
	}


	ListSelection_Overlay::Update();
}

bool ObjectSelection_Overlay::ShouldItemBeHidden(const char* pszString,int iData,bool bWasHidden)
{
	Tools_Object* pxObject = (Tools_Object*) iData;

	if(m_bOnlyShowSelectable)
	{
		if(!pxObject->IsSelectable()) return true;
	}

	
	if(s_pxFilterParam)
	{
		bool bMatch = false;

		//Check to see if the object has the parameter that is being checked , and that it
		//has the correct value.
		switch(s_pxFilterParam->m_iType)
		{
			case TOOLS_OBJECTSYSTEM_PARAM_BOOL :
			{
				Tools_ParameterDerived<bool> xParam;
				if(pxObject->GetParameter(s_pxFilterParam->m_uParamID, &xParam))
				{
					bMatch = xParam.GetBool() == s_bFilterParam_Bool;
				}
			}
			break;

			case TOOLS_OBJECTSYSTEM_PARAM_INT :
			{
				Tools_ParameterDerived<int> xParam;
				if(pxObject->GetParameter(s_pxFilterParam->m_uParamID, &xParam))
				{
					bMatch = xParam.GetInt() >= s_iFilterParam_MinInt && xParam.GetInt() <= s_iFilterParam_MaxInt;
				}
			}
			break;

			case TOOLS_OBJECTSYSTEM_PARAM_UINT :
			{
				Tools_ParameterDerived<u_int> xParam;
				if(pxObject->GetParameter(s_pxFilterParam->m_uParamID, &xParam))
				{
					bMatch = xParam.GetUInt() >= s_uFilterParam_MinUInt && xParam.GetUInt() <= s_uFilterParam_MaxUInt;
				}
			}
			break;

			case TOOLS_OBJECTSYSTEM_PARAM_FLOAT :
			{
				Tools_ParameterDerived<float> xParam;
				if(pxObject->GetParameter(s_pxFilterParam->m_uParamID, &xParam))
				{
					bMatch = xParam.GetFloat() >= s_fFilterParam_MinFloat && xParam.GetFloat() <= s_fFilterParam_MaxFloat;
				}
			}
			break;

			case TOOLS_OBJECTSYSTEM_PARAM_STRING :
			{
				Tools_ParameterDerived<Asura_String> xParam;
				if(pxObject->GetParameter(s_pxFilterParam->m_uParamID, &xParam))
				{
					Asura_String xString = xParam.GetString();
					if(xString.IsBlank())
					{
						bMatch = s_xFilterParam_String.IsBlank();
					}
					else if(s_xFilterParam_String.IsBlank())
					{
						bMatch = true;
					}
					else
					{
						bMatch = xString.StartsWith(s_xFilterParam_String.GetString());
					}

				}
			}
			break;

		}

		if(!bMatch) return true;
	}

	if(m_iFilterIndex == 0) return false;//first entry is for showing everything
	
	switch(m_iFilterType)
	{
		case FILTER_OBJECTTYPE :
			return m_xFilterList[m_iFilterIndex]->iData != pxObject->Type;

		case FILTER_SHAPE :
			return pxObject->GetShapeOrSkinName().Compare(m_ppszFilterStrings[m_iFilterIndex])!=0;

        case FILTER_BLUEPRINTTYPE :
            return pxObject->GetBlueprintTypeName().Compare(m_ppszFilterStrings[m_iFilterIndex])!=0;

        case FILTER_BLUEPRINT :
            return pxObject->GetBlueprintName().Compare(m_ppszFilterStrings[m_iFilterIndex])!=0;

		default :
			Asura_Assert(0,"Unexpected filter type");
			break;
	}
	return false;
}

void ObjectSelection_Overlay::SetOnlyShowSelectable(bool bSet)
{
	if(m_bOnlyShowSelectable != bSet)
	{
		m_bOnlyShowSelectable = bSet;
		m_bFilterIndexChanged = true;//force the list to be redone during the next update;
	}
}

bool ObjectSelection_Overlay::IsNotFilterParamBool(int iJunk)
{
	return !(s_pxFilterParam && s_pxFilterParam->m_iType == TOOLS_OBJECTSYSTEM_PARAM_BOOL);
}

bool ObjectSelection_Overlay::IsNotFilterParamInt(int iJunk)
{
	return !(s_pxFilterParam && s_pxFilterParam->m_iType == TOOLS_OBJECTSYSTEM_PARAM_INT);
}

bool ObjectSelection_Overlay::IsNotFilterParamUInt(int iJunk)
{
	return !(s_pxFilterParam && s_pxFilterParam->m_iType == TOOLS_OBJECTSYSTEM_PARAM_UINT);
}

bool ObjectSelection_Overlay::IsNotFilterParamFloat(int iJunk)
{
	return !(s_pxFilterParam && s_pxFilterParam->m_iType == TOOLS_OBJECTSYSTEM_PARAM_FLOAT);
}

bool ObjectSelection_Overlay::IsNotFilterParamString(int iJunk)
{
	return !(s_pxFilterParam && s_pxFilterParam->m_iType == TOOLS_OBJECTSYSTEM_PARAM_STRING);
}




//////////////////////////////////////////////////////////////////////////////////////////////
bool ObjectFind_Overlay::s_bGotoSelected = true;
bool ObjectFind_Overlay::s_bOnlyShowSelectable = true;

void ObjectFind_Overlay::AddExtraOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset)
{
	ObjectSelection_Overlay::AddExtraOverlayControls(pxOverlay,xOffset);

	Asura_Dialog* pxDialog = new Asura_Tickbox_Dialog("Goto Selected",xOffset,&s_bGotoSelected);
	pxDialog->SetHelpText("If ticked then the camera will be moved to the object that is chosen.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Only Selectable",xOffset,&s_bOnlyShowSelectable);
	pxDialog->SetHelpText("If ticked then only objects that can be selected will be shown in the list.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

}

void ObjectFind_Overlay::Update()
{
	SetOnlyShowSelectable(s_bOnlyShowSelectable);

	ObjectSelection_Overlay::Update();
}

//////////////////////////////////////////////////////////////////////////////////////////////


u_int ObjectHideByType_Overlay::s_uSelectedTab = 0;
Asura_Vector_2 ObjectHideByType_Overlay::s_xPosition(20.0f,10.0f);

void ObjectHideByType_Overlay::Create(const List<Tools_Object*>* pxObjectList)
{
	//close any existing overlay
	Close();
	
	//create a new overlay
	ObjectHideByType_Overlay* pxOverlay = new ObjectHideByType_Overlay(pxObjectList);
	if(pxOverlay->m_xTypeList.size())
	{
		pxOverlay->ShowOverlay();
	}
	else
	{
		delete pxOverlay;
	}
}

bool ObjectHideByType_Overlay::IsOverlayOn( void )
{
	return( Asura_Overlay_System::DoesOverlayExist( TOOLS_OVERLAY_HIDEBYTYPE ) );
}
void ObjectHideByType_Overlay::ToggleOverlay( void )
{
	if( IsOverlayOn() )
	{
		Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_HIDEBYTYPE );
	}
	else
	{
		Tools_Object_System::ShowUserHideByTypeOverlay();
	}
}

void ObjectHideByType_Overlay::Close()
{
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_HIDEBYTYPE);
}


ObjectHideByType_Overlay::ObjectHideByType_Overlay(const List<Tools_Object*>* pxObjectList)
:Asura_Overlay("Visible Object Types :",false,&s_xPosition)
{
	m_ppszTypeStrings = 0;
	m_ppszPlatformStrings = 0;
	m_ppszLightSetStrings = 0;
	
	
	#ifdef ASURA_USE_BOTANICALS
	//Include botanicals in the object type list regardless
	AddToTypeList("Botanical", TOT_Botanicals);
	#endif
	
	
	//build list of type names
	for(CLIF<Tools_Object*> olif(pxObjectList);!olif.done();olif.next())
	{
		if(olif()->Type == TOT_Camera) continue;

		//add this object's type to the type list
		AddToTypeList(olif()->GetTypeDescriptionForObjectType(),olif()->Type);
	}


	m_ppszTypeStrings = new char*[m_xTypeList.size()];
	for(int i=0;i<m_xTypeList.size();i++)
	{
		m_ppszTypeStrings[i] = m_xTypeList[i]->xString.GetString();
	}

	//platform names
	for(int i=0;i<Tools_Level_Export::ExportType_NumPlatforms;i++)
	{
		const char* szName = Tools_Level_Export::GetPlatformName(i);
		if(!szName) continue;

		ListSelectionItem* pxItem = new ListSelectionItem;
		pxItem->xString = szName;
		pxItem->iData = i;
		pxItem->bSelectedInitially = false;
		pxItem->bHidden = false;
		
		m_xPlatformList.add_entry(pxItem);

	}

	m_ppszPlatformStrings = new char*[m_xPlatformList.size()];
	for(int i=0;i<m_xPlatformList.size();i++)
	{
		m_ppszPlatformStrings[i] = m_xPlatformList[i]->xString.GetString();
	}


	//light set names
	for(int i=0;i<TOOLS_MAX_LIGHT_SETS;i++)
	{
		const char* szName = Tools_LightSets::GetLightSetNameByIndex(i);
		if(!szName) continue;

		ListSelectionItem* pxItem = new ListSelectionItem;
		pxItem->xString = szName;
		pxItem->iData = i;
		pxItem->bSelectedInitially = false;
		pxItem->bHidden = false;
		
		m_xLightSetList.add_entry(pxItem);

	}

	m_ppszLightSetStrings = new char*[m_xLightSetList.size()];
	for(int i=0;i<m_xLightSetList.size();i++)
	{
		m_ppszLightSetStrings[i] = m_xLightSetList[i]->xString.GetString();
	}

}

ObjectHideByType_Overlay::~ObjectHideByType_Overlay()
{
	while(m_xTypeList.size())
	{
		delete m_xTypeList.first_entry();
		m_xTypeList.delete_first_entry();
	}

	delete [] m_ppszTypeStrings;

	while(m_xLightSetList.size())
	{
		delete m_xLightSetList.first_entry();
		m_xLightSetList.delete_first_entry();
	}

	delete [] m_ppszLightSetStrings;

	while(m_xPlatformList.size())
	{
		delete m_xPlatformList.first_entry();
		m_xPlatformList.delete_first_entry();
	}

	delete [] m_ppszPlatformStrings;

}

void ObjectHideByType_Overlay::AddToTypeList(const char* pszTypeName,int iType)
{
	if(!pszTypeName || !strlen(pszTypeName)) return;

	//check for this type already being added
	LIF<ListSelectionItem*> xTypeLIF(&m_xTypeList);

	for(xTypeLIF.restart();!xTypeLIF.done();xTypeLIF.next())
	{
		if(iType == xTypeLIF()->iData) return;
	}


	//create a ListSelectionItem
	ListSelectionItem* pxItem = new ListSelectionItem;
	pxItem->xString = pszTypeName;
	pxItem->iData = iType;
	pxItem->bSelectedInitially = false;
	pxItem->bHidden = false;

	//add it into the list (sorted)

	for(xTypeLIF.restart();!xTypeLIF.done();xTypeLIF.next())
	{
		if(pxItem->xString.Compare(xTypeLIF()->xString,true)<=0)
		{
			m_xTypeList.add_entry_before(pxItem,xTypeLIF());
			break;
		}
	}
	if(xTypeLIF.done())
	{
		m_xTypeList.add_entry(pxItem);
	}
}



void ObjectHideByType_Overlay::ShowOverlay()
{
	m_iTypeSelectionIndex = 0;


	//add controls for the list of object types
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);

//	pxDialog = new Asura_Dialog("Visible Types :",xOffset);
//	AddDialog(pxDialog);
//	xOffset.y += Asura_Dialog::SmallSpacing();

	//Tab control
	static char* s_aszTabs[3] = {"Type","Platform","Light Set"};
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset,s_aszTabs,3);
	AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	float fTabsStartY = xOffset.y;
	float fTabsEndY = xOffset.y;

	SetCurrentParentDialog(pxTabsDialog);	

	//Type tab
	pxTabsDialog->SelectTab(0);
	xOffset.y = fTabsStartY;

	int iNumLines = min(m_xTypeList.size(),10);

	pxDialog = new Asura_List_Dialog(xOffset,m_ppszTypeStrings,&m_iTypeSelectionIndex,iNumLines,m_xTypeList.size(),_IsVisible,_ToggleVisible,(int)this);
	AddDialog(pxDialog);
	xOffset.y=pxDialog->m_xMaxCoord.y;
	xOffset.y+=1.0f;
	

	//End of type tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//Platform tab
	pxTabsDialog->SelectTab(1);
	xOffset.y = fTabsStartY;
	
 	iNumLines = min(m_xPlatformList.size(),10);

	pxDialog = new Asura_List_Dialog(xOffset,m_ppszPlatformStrings,&m_iPlatformSelectionIndex,iNumLines,m_xPlatformList.size(),_IsVisible,_ToggleVisible,(int)this);
	pxDialog->SetIsEnabledFunction(Tools_Object_System::IsHideByPlatformEnabled);
	AddDialog(pxDialog);
	xOffset.y=pxDialog->m_xMaxCoord.y;
	xOffset.y+=1.0f;

	m_bPlatformEnabled = Tools_Object_System::IsHideByPlatformEnabled();
	pxDialog = new Asura_Tickbox_Dialog("Enabled",xOffset,&m_bPlatformEnabled);
	pxDialog->SetHelpText("Enables hiding of objects based on platform");
	AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();
	
	//End of platform tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//Light set tab
	pxTabsDialog->SelectTab(2);
	xOffset.y = fTabsStartY;

 	iNumLines = min(m_xLightSetList.size(),10);

	pxDialog = new Asura_List_Dialog(xOffset,m_ppszLightSetStrings,&m_iLightSetSelectionIndex,iNumLines,m_xLightSetList.size(),_IsVisible,_ToggleVisible,(int)this);
	pxDialog->SetIsEnabledFunction(Tools_Object_System::IsHideByLightSetEnabled);
	AddDialog(pxDialog);
	xOffset.y=pxDialog->m_xMaxCoord.y;
	xOffset.y+=1.0f;

	m_bLightSetEnabled = Tools_Object_System::IsHideByLightSetEnabled();
	pxDialog = new Asura_Tickbox_Dialog("Enabled",xOffset,&m_bLightSetEnabled);
	pxDialog->SetHelpText("Enables hiding of objects based on light set");
	AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();


	
	//End of light set tab
	if(xOffset.y > fTabsEndY) fTabsEndY = xOffset.y;

	//End of tabs
	SetCurrentParentDialog(0);
	pxTabsDialog->SetSelectedTabPointer(&s_uSelectedTab);
	xOffset.y = fTabsEndY + 1.0f;


	pxDialog = Tools_Entity_TemplateInstance_Object::CreateHideTemplateTriggersTickBox( xOffset );
	AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FunctionTickbox_Dialog("Show phys objs in other modes", xOffset, Tools_Object_System::ShouldShowPhysicalObjectsInOtherModes, Tools_Object_System::SetShowPhysicalObjectsInOtherModes);
	pxDialog->SetHelpText("Causes tools objects for physical entities to be created (and be visible depending on other flags) in modes where they wouldn't normally be edited.\nThis won't do anything if a mode has already created the objects and is hiding them for its own diabolical schemes.");
	AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::LargeSpacing();


	pxDialog = new Asura_Function_Dialog("Hide All",xOffset,_HideAll,(int)this);
	pxDialog->SetHelpText("Unticks all the boxes on the current tab");
	AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Show All",xOffset,_ShowAll,(int)this);
	pxDialog->SetHelpText("Ticks all the boxes on the current tab");
	AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Invert",xOffset,_InvertVisible,(int)this);
	pxDialog->SetHelpText("Inverts all the boxes on the current tab");
	AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	m_uID = TOOLS_OVERLAY_HIDEBYTYPE;
	Asura_Overlay_System::SubmitNewOverlay(this);
	
}

void ObjectHideByType_Overlay::PerFrameUpdate()
{
	Asura_Overlay::PerFrameUpdate();

	Tools_Object_System::SetHideByPlatformEnabled(m_bPlatformEnabled);
	Tools_Object_System::SetHideByLightSetEnabled(m_bLightSetEnabled);
}

bool ObjectHideByType_Overlay::IsVisible(int iIndex)
{
	switch(s_uSelectedTab)
	{
		case SELECTEDTAB_TYPE :
		{
			int iType = m_xTypeList[iIndex]->iData;
			return !Tools_Object_System::GetTypeHiddenFlag(iType,ObjectHideFlag_HideByTypeOverlay);
		}
		case SELECTEDTAB_PLATFORM :
		{
			int iPlatform = m_xPlatformList[iIndex]->iData;
			return Tools_Object_System::IsPlatformVisible(iPlatform);
		}
		case SELECTEDTAB_LIGHTSET :
		{
			int iLightSet = m_xLightSetList[iIndex]->iData;
			return Tools_Object_System::IsLightSetVisible(iLightSet);
		}
		default :
		{
			Asura_Assert(0,"Unexpected tab");
			return false;
		}
	}
}

void ObjectHideByType_Overlay::ToggleVisible(int iIndex)
{
	switch(s_uSelectedTab)
	{
		case SELECTEDTAB_TYPE :
		{
			int iType = m_xTypeList[iIndex]->iData;
			bool bWasHidden = Tools_Object_System::GetTypeHiddenFlag(iType,ObjectHideFlag_HideByTypeOverlay);
			Tools_Object_System::SetTypeHidden(iType,ObjectHideFlag_HideByTypeOverlay,!bWasHidden);
			break;
		}
		case SELECTEDTAB_PLATFORM :
		{
			int iPlatform = m_xPlatformList[iIndex]->iData;
			bool bWasVisible = Tools_Object_System::IsPlatformVisible(iPlatform);
			Tools_Object_System::SetPlatformVisible(iPlatform,!bWasVisible);
			break;
		}
		case SELECTEDTAB_LIGHTSET :
		{
			int iLightSet = m_xLightSetList[iIndex]->iData;
			bool bWasVisible = Tools_Object_System::IsLightSetVisible(iLightSet);
			Tools_Object_System::SetLightSetVisible(iLightSet,!bWasVisible);
			break;
		}
		default :
		{
			Asura_Assert(0,"Unexpected tab");
		}
	}
	
}

void ObjectHideByType_Overlay::HideAll()
{
	switch(s_uSelectedTab)
	{
		case SELECTEDTAB_TYPE :
		{
			for(int i=0;i<m_xTypeList.size();i++)
			{
				int iType = m_xTypeList[i]->iData;
				Tools_Object_System::SetTypeHidden(iType,ObjectHideFlag_HideByTypeOverlay,true);
			}
			break;
		}
		case SELECTEDTAB_PLATFORM :
		{
			Tools_Object_System::SetVisiblePlatforms(0);
			break;
		}
		case SELECTEDTAB_LIGHTSET :
		{
			Tools_Object_System::SetVisibleLightSets(0);
			break;
		}
		default :
		{
			Asura_Assert(0,"Unexpected tab");
			break;
		}
	}
}

void ObjectHideByType_Overlay::ShowAll()
{
	switch(s_uSelectedTab)
	{
		case SELECTEDTAB_TYPE :
		{
			for(int i=0;i<m_xTypeList.size();i++)
			{
				int iType = m_xTypeList[i]->iData;
				Tools_Object_System::SetTypeHidden(iType,ObjectHideFlag_HideByTypeOverlay,false);
			}
			break;
		}
		case SELECTEDTAB_PLATFORM :
		{
			Tools_Object_System::SetVisiblePlatforms(0xffffffff);
			break;
		}
		case SELECTEDTAB_LIGHTSET :
		{
			Tools_Object_System::SetVisibleLightSets(0xffffffff);
			break;
		}
		default :
		{
			Asura_Assert(0,"Unexpected tab");
			break;
		}
	}
}

void ObjectHideByType_Overlay::InvertVisible()
{
	int iListSize = 0;
	switch(s_uSelectedTab)
	{
		case SELECTEDTAB_TYPE :
		{
			iListSize = m_xTypeList.size();
			break;
		}
		case SELECTEDTAB_PLATFORM :
		{
			iListSize = m_xPlatformList.size();
			break;
		}
		case SELECTEDTAB_LIGHTSET :
		{
			iListSize = m_xLightSetList.size();
			break;
		}
		default :
		{
			Asura_Assert(0,"Unexpected tab");
			break;
		}
	}
	for(int i=0;i<iListSize;i++)
	{
		ToggleVisible(i);
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_ChooseObject::SelectObject(Tools_Object* pxObject, bool bMultipleSelect)
{
	if(!pxObject || !CanBeSelected(pxObject)) return false;

	if(!bMultipleSelect || !SupportMultipleSelection())
	{
		//clear the existing selection
		DeselectAllObjects();
	}
	//remove from the list if already selected ,so it can be added to the end
	if(pxObject->IsSelectedByChooseObject())
	{
		s_xSelectedObjectList.delete_entry(pxObject);
	}
	s_xSelectedObjectList.add_entry(pxObject);
	pxObject->SetSelectedByChooseObject(true);

	return true;
}

void Tools_ChooseObject::DeselectObject(Tools_Object* pxObject)
{
	if(pxObject && pxObject->IsSelectedByChooseObject())
	{
		s_xSelectedObjectList.delete_entry(pxObject);
		pxObject->SetSelectedByChooseObject(false);
	}
}

void Tools_ChooseObject::DeselectAllObjects()
{
	while(s_xSelectedObjectList.size())
	{
		s_xSelectedObjectList.first_entry()->SetSelectedByChooseObject(false);
		s_xSelectedObjectList.delete_first_entry();
	}
}

void Tools_ChooseObject::ClearList()
{
	while(s_xSelectedObjectList.size())
	{
		s_xSelectedObjectList.delete_first_entry();
	}
}

void Tools_ChooseObject::ClearSelectionFlags()
{
	for(LIF<Tools_Object*> xObjectLIF(&s_xSelectedObjectList); !xObjectLIF.done(); xObjectLIF.next())
	{
		xObjectLIF()->SetSelectedByChooseObject(false);
	}
}

void Tools_ChooseObject::HideOverlays(bool bHide)
{
	Asura_Overlay_System::SetHideFlagOnCurrentOverlays(ASURA_OVERLAY_HIDEFLAG_OBJECTSYSTEM, bHide);

	if(bHide)
	{
		//don't hide the entity manager during choose object
		Asura_Overlay* pxEntityManagerOverlay = Asura_Overlay_System::GetOverlay(TOOLS_OVERLAY_SECONDARYENTFILE);
		if(pxEntityManagerOverlay)
		{
			pxEntityManagerOverlay->SetHideFlag(ASURA_OVERLAY_HIDEFLAG_OBJECTSYSTEM, false);
		}
	}

}


bool Tools_ChooseObject_ByType::CanBeSelected( Tools_Object* pxObject )
{
	if ( !pxObject ) { return false; }
	if ( m_xTypes.Find( pxObject->Type ) == NULL ) { return false; }
	if ( m_xGuidsToIgnore.Find( pxObject->GetGuid() ) ) { return false; }
	return true;
}

void Tools_ChooseObject_ByType::ObjectSelected( Tools_Object* pxObject )
{
	if ( SupportMultipleSelection() )
	{
		for ( CLIF< Tools_Object* > xIter( GetSelectionList() ); !xIter.done(); xIter.next() )
		{
			if ( m_puGuidToSet )
			{
				*m_puGuidToSet = xIter()->GetGuid();
			}
			if ( m_pfnOnChosen )
			{
				m_pfnOnChosen( xIter()->GetGuid() );
			}
			if ( m_pbChanged )
			{
				*m_pbChanged = true;
			}
		}
	}
	else
	{
		const Asura_Guid uObjectGuid = pxObject ? pxObject->GetGuid() : ASURA_GUID_UNREGISTERED;
		if ( m_puGuidToSet )
		{
			*m_puGuidToSet = uObjectGuid;
		}
		if ( m_pfnOnChosen )
		{
			m_pfnOnChosen( uObjectGuid );
		}
		if ( m_pbChanged )
		{
			*m_pbChanged = true;
		}
	}
}

