/*
	DEBUG functions

	James Russell - SCEE Technology Group

	===================================================
	*Please* send bug reports to James_Russell@scee.net
	If you fix it yourself but don't tell me, your fix
	may be wiped out when you d/l the next release!
	===================================================

	These functions are used to handle any output generated by the debug modules.
	They implement a very rudimentary STDOUT and STDERR channel.

	Version history:

	Ver		Author		Comments
	===		======		========

	2.12	J Russell	25/10/2002 - Fixed parsing bug for '-B' option.
	2.11	J Russell	03/10/2002 - Added option to strip colours from output
						Added option to specify line wrapping length.
	2.07    J Russell   8/1/2001 - Now explicitly tells you if it halted because
						there were too many errors/warnings.
	0.95	J Russell	11/7/2000 - Added bounds checks on the output size.
						Revamped config string parsing.
						Added variable buffer size and scrolling.
	0.9		J Russell	8/7/2000 - Added a check to see if we were being
						used without being initialised first.
	0.7		J Russell	1/6/2000 - First release

*/

/*
	debug.c

	Error handling and debugging functions
	======================================

	The error system is fairly simple - there is one large output buffer
	called gOutputBuffer (equivalent to STDOUT).

	Errors/Warnings are added using AddError and AddWarning, which take an
	error/warning number (defined in Debug.h) and a string to be added to the
	gOutputBuffer.

	These error messages are stored in memory rather than being sent straight to
	the console output, because the console has been known to drop some output
	lines, and you might want to implement a more reliable protocol for getting
	the output back the the user (ie via the PS2 screen or some TCP connection).
	
	When the error system is initialised, you can specify a string which has some
	options on it. The parser is very simple, so it's probably easy to break, but
	basically it ignores other options on that line, so you can pass options to
	modules which use this error subsystem and be safe in the knowledge that the
	error subsystem will ignore those options.

	Initialisation options:
		-w			Ignore all warnings
		-Wx			Ignore warning 'x', where 'x' is a number. EG -W36 ignores warning 36.
		-Werror		Treat warnings as errors, i.e halt if a warning is encountered.
		-Ex			Stop processing after 'x' errors encountered. I.e. -E6 means stop
					after encountering 6 errors.
		-Bx			Change buffer size. By default it is 100KB. Specify a new size with 'x'.
					E.G. -B200 allocates a 200KB buffer. (The error buffer is always
					fixed in size, you cannot change it)
		-noscroll	The buffer can either fill and stop (causing a fatal error
					to be returned), or it can 'scroll', that is, once information goes off
					one end, information at the start is removed to make space (this is the
					default behaviour). Specifying this option causes the module to
					disable deleting the first 20 lines whenever it fills up.
		-Lx			Specify the line wrapping width to 'x' (default is 100, set to 0 for
					no wrapping)
		-nocols		Strip colour codes from the output.

	You'd initialise the subsystem like so: InitialiseErrorSystem("-W36 -W22 -W1 -E8");
	which would ignore warnings 36, 22 and 1, and stop processing after 8 errors.

	Pass in NULL or the empty string if you don't need to set any options.

	----------

	Because there is a maximum number of errors/warnings that can occur before the error
	system decides enough is enough, the error system must be able to inform the calling
	procedure that it should halt.

	Thus, when the limit of errors/warnings is reached, AddError and AddWarning will
	return TRUE.

	However, it is up to the calling function to actually recognise this and halt.

	In general, warnings are things that won't cause a crash, but are a bit suspect,
	like having bits set in unused parts of a DMA tag. Errors are things which are
	definitely right out, like setting reserved bits and invalid memory accesses.
	There are two types of errors though (just like in a C compiler) - those where
	you can still continue on for a bit and try to find any more errors (like setting
	reserved bits in some address), and those which mean that the module just can't go
	on, like null addresses in the 'next tag' field.
*/

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "PS2/PS2Strip/DMADebug/Debug.h"

// Stdlib independent.
#include <ctype.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

// Engine independent.
#include "Debugging/Asura_Assert.h"
#include "Core/Asura_String.h"

// DMADebug.
#include "PS2/PS2Strip/DMADebug/DMADebug.h"

// B E G I N   N A M E S P A C E /////////////////////////////////////////////

namespace DMADebug
{

// C O N S T A N T S /////////////////////////////////////////////////////////

static const int DEFAULT_LINE_WIDTH             = 100;
static const int MAX_WARNINGS_MASKED            = 100;
static const int DEFAULT_MAX_ERRORS_BEFORE_HALT = 6;
static const int MAX_WARNINGS_BEFORE_HALT       = 25;

//// Specified in KB
//static const int ERROR_BUFFER_SIZE              = 10;

// G L O B A L S /////////////////////////////////////////////////////////////

static bool initialised = false;		// Has this module been initialised?

static int gLineWidth;
static bool gStripColours;
static int gMaxErrors;				// The maximum number of errors we're prepared to encounter before
									// returning HALT.
static bool gTreatWarningsAsErrors;

static int gErrCount;				// Number of errors/warnings encountered so far.
static int gWarnCount;

static int numWarningsIgnored = 0;
static int warningsIgnored[MAX_WARNINGS_MASKED];	// Array holding all the warning numbers to ignore.

// (MP - 16.11.2002) Yuck.
char gTempString[1024];				// Just a temporary string for random sprintf calls.
static char gTempStringBuffer[1024];	// Temporary string for use in this modules only

char *gOutputBuffer = NULL;		// The big output buffer
u_int gOutputBufferSize = 0;

static char *gpOutputBuffer;		// optimisation - a pointer showing the current end of the output
									// buffer, so we can add strings onto the end faster.

static bool gScroll = true;		// Boolean, should we scroll when the buffer fills, or fail with an error.

// F U N C T I O N   I M P L E M E N T A T I O N S ///////////////////////////

// Initialises the Error System, using a configuration string (much like command line options)

// -w = ignore all warnings
// -Werror = treat warnings as errors
// -Wx = mask warnings for warning number 'x' (e.g. -W36 masks error 36, see Debug.h for errnums)
// -Ex = allow 'x' errors before quitting
// -Lx = set line wrapping to this amount (0 means no line wrapping)
// -nocol = strip colour output (remove ANSI colour chars)

	// Helper function - converts a string to an int.
static int parseNumber( const char *p, const char *paramName )
 {
	int num = 0;

    Asura_Assertf( isdigit(*p), ("Invalid debug initialisation string at '%s'!", paramName) );
	while(isdigit(*p))
    {
		num = num * 10 + (*p - 0x30);
		p++;
	}

	return num;
}

bool DEBUG_parseConfig( const char* initString )
{
	int maskWarning;
	const char *p;

//    if( gOutputBuffer ) { delete[] gOutputBuffer; }

	gErrCount = 0;
	gWarnCount = 0;

	gMaxErrors = DEFAULT_MAX_ERRORS_BEFORE_HALT;
	gTreatWarningsAsErrors = false;
	numWarningsIgnored = 0;
	gLineWidth = DEFAULT_LINE_WIDTH;
	gStripColours = false;

    if( !gOutputBufferSize )
    {
	    gOutputBufferSize = 100;
    }
	gScroll = true;

	if(initString != NULL) {
				// Parse the init string
		if(strstr(initString, "-w") != NULL) {
			numWarningsIgnored = -1;	// -1 means ignore all warnings
			DMADebug_Puts("Ignoring all warnings");
		}

		if((p = strstr(initString, "-L")) != NULL) {
			gLineWidth = parseNumber(p+2, "-L");
			Asura_Assert( gLineWidth > 25 || gLineWidth == 0, "Invalid debug initialisation, Line Width too small (must be >25)!" );
			DMADebug_Printf( ("Using a line width of %d.\n", gLineWidth) );
		}

		if((p = strstr(initString, "-E")) != NULL) {
			gMaxErrors = parseNumber(p+2, "-E");
			DMADebug_Printf( ("Maximum number of errors displayed: %d\n", gMaxErrors) );
		}

		p = initString;
		while( (p = strstr(p, "-W")) != NULL) {
			p+=2;
			Asura_Assert( *p, "Invalid debug initialisation string at '-W'!" );

			if(*p == 'e') {		// Check for 'error' (actually just check for 'e');
				gTreatWarningsAsErrors = true;
			} else {
						// Else we assume its a decimal number and mask that warning
				maskWarning = parseNumber(p, "-W");
				Asura_Assert( numWarningsIgnored < MAX_WARNINGS_MASKED, "Invalid initialisation string at '-W'! Too many warning masks!" );
				warningsIgnored[numWarningsIgnored++] = maskWarning;
				DMADebug_Printf( ("Will ignore warning number %d\n", maskWarning ) );
			}
		}

		if((p=strstr(initString, "-B")) != NULL)
        {
            u_int uOldOutputBufferSize = gOutputBufferSize;

			gOutputBufferSize = parseNumber(p+2, "-B");
			Asura_Assert( gOutputBufferSize != 0, "Invalid parameter for buffer size (-B)! Must be a number > 0" );
			DMADebug_Printf( ("Setting buffer size to %dkb\n", gOutputBufferSize) );

            // Need to allocate a new buffer if a new size is required.
            if( gOutputBuffer && gOutputBufferSize != uOldOutputBufferSize )
            {
                delete[] gOutputBuffer;
                gOutputBuffer = NULL;
            }
		}

		if(strstr(initString, "-noscroll") != NULL) {
			gScroll = false;
			DMADebug_Puts("Disabling scrolling.");
		}

		if(strstr(initString, "-nocols") != NULL) {
			gStripColours = true;
			DMADebug_Puts("Will strip colour codes from output.");
		}
	}

    if( !gOutputBuffer )
    {
	    gOutputBuffer = new char[gOutputBufferSize * 1024];
        if( !gOutputBuffer )
        {
            return false;
        }
    }

	initialised = true;

	DEBUG_reset();

    return true;
}

//// Strips colour character codes from the string (in place)
//static void stripColours( char* szDest, const char* szSrc )
//{
//    if( !gStripColours ) { return; }
//
//    // (MP - 16.11.2002) TODO: Fix this to work for RTF!
//	while( *szSrc )
//    {
//		if( 0x1b == *szSrc )
//        {
//            // If find escape char, skip to just past 'm' and try again.
//            while( 'm' != *szSrc++ ) { }
//		}
//        else
//        {
//			*(szDest++) = *(szSrc++);
//		}
//	}
//
//	*szDest = '\0';
//}

// once initialised, this will clear the output buffers
void DEBUG_reset(void)
{
	Asura_Assert( initialised, "An attempt was made to use the DEBUG module without initialising it first." );

	gOutputBuffer[0] = '\0';
	gpOutputBuffer = gOutputBuffer;
}


static void ConvertNewlines( char* szDest, const char* szSrc, int iLimit )
{
    Asura_Assert( iLimit > 0, "" );

    char* szOut = szDest;
    while( szOut - szDest < iLimit )
    {
        if( '\n' == *szSrc )
        {
            // Check we can fit "\par" and a nul terminator in. Give up if
            // not.
            if( (szOut + 4) - szDest < iLimit )
            {
                *szOut++ = '\\';
                *szOut++ = 'p';
                *szOut++ = 'a';
                *szOut++ = 'r';
                ++szSrc;
            }
            else
            {
                *szOut++ = '\0';
                break;
            }
        }
        else
        {
            if( '\0' == (*szOut++ = *szSrc++) )
            {
                break;
            }
        }
    }
    Asura_Assert( szOut - szDest <= iLimit, "" );
    
    // Append nul terminator if necessary.
    if( iLimit && *(szOut - 1) )
    {
        *(szOut - 1) = '\0';
    }
}


// =====================================================
// Adds a string to the Output Buffer

int DEBUG_addToOutputBuffer( const char* szToAdd )
{
	int foundError = DMA_ERR_NO_ERROR;
	int i;

	Asura_Assert( initialised, "An attempt was made to use the DEBUG module without initialising it first." );

//	stripColours( gTempStringBuffer, szToAdd );
//	i = strlen( gTempStringBuffer );

    ConvertNewlines( gTempStringBuffer, szToAdd, 1024 );
	i = strlen( gTempStringBuffer );

    // if adding this string will overflow the buffer...
	if((gpOutputBuffer + i + 100) > (gOutputBuffer + gOutputBufferSize * 1024))
    {
		if( !gScroll )
        {
			// Exit with an error;
            Asura_Assert( false, "Make buffer bigger!" );
			strcat(gpOutputBuffer, "\n\n" DMADEBUG_RTF_COLOUR_ERROR "DEBUG MODULE - OUTPUT BUFFER FULL" DMADEBUG_RTF_COLOUR_NORMAL "\n");
			return DMA_ERR_FATAL_HALT;
		}
        else
        {
				// Delete the first 20 lines of output and retry.
			char *p = gOutputBuffer;
			int numNewlines = 20;
			while(numNewlines > 0) {
				if(*p == '\n') {
					numNewlines--;
				}
				p++;
				if(p >= gpOutputBuffer) {
					break;
				}
			}
			if(numNewlines > 0 || p >= gpOutputBuffer) {
				p = gpOutputBuffer;
			}
				// p should now point to the char after the first char of the 20th line.
			memmove(gOutputBuffer, p, strlen(p)+1);
			gpOutputBuffer = gOutputBuffer + strlen(gOutputBuffer);
				// If it's still larger then we have to quit.
			Asura_Assert( gpOutputBuffer + i + 100 <= gOutputBuffer + gOutputBufferSize * 1024, "Error - could not fit string into Output Buffer, even after scrolling." );
		}
	}

	strcat( gpOutputBuffer, gTempStringBuffer );
	strcat( gpOutputBuffer, DMADEBUG_RTF_TEXT_RESET "\\par" );
	gpOutputBuffer = gpOutputBuffer + strlen(gpOutputBuffer);

	return foundError;
}







// ==========================================
// Adds an error to the Error Buffer

int DEBUG_addError( int errNum, int severity, const char* errString )
{
	int err = severity;
	int i;

	Asura_Assert( initialised, "An attempt was made to use the DEBUG module without initialising it first." );

	if( severity == DMA_ERR_WARNING )
    {
		if(numWarningsIgnored == -1) {			// -1 means ignore all warnings
			DMADebug_Puts("ignoring all warnings");
			return 0;
		}
		for(i = 0; i < numWarningsIgnored; i++) {	// else check to see if we should ignore this warning
			if(warningsIgnored[i] == errNum) {
				DMADebug_Printf( ("Ignoring warning %d\n", errNum) );
				return 0;
			}
		}
		sprintf(gTempStringBuffer, DMADEBUG_RTF_COLOUR_WARNING"warning (%d) : %s" DMADEBUG_RTF_COLOUR_NORMAL, errNum, errString);
		if(gTreatWarningsAsErrors) {
			gErrCount++;
			if(gErrCount > gMaxErrors) {
				puts(DMADEBUG_RTF_COLOUR_ERROR "DMAdebug - too many errors (treating warnings as errors), exiting."DMADEBUG_RTF_COLOUR_NORMAL);
				err |= DMA_ERR_FATAL_HALT;
			}
		} else {
			gWarnCount++;
			if(gWarnCount > MAX_WARNINGS_BEFORE_HALT) {
				puts(DMADEBUG_RTF_COLOUR_ERROR "DMAdebug - too many warnings, exiting."DMADEBUG_RTF_COLOUR_NORMAL);
				err |= DMA_ERR_FATAL_HALT;
			}
		}

	} else {	// Errors
		sprintf(gTempStringBuffer, DMADEBUG_RTF_COLOUR_ERROR"ERROR (%d) : %s"DMADEBUG_RTF_COLOUR_NORMAL, errNum, errString);
		gErrCount++;
		if(gErrCount > gMaxErrors) {
			puts(DMADEBUG_RTF_COLOUR_ERROR "DMAdebug - too many errors, exiting."DMADEBUG_RTF_COLOUR_NORMAL);
			err |= DMA_ERR_FATAL_HALT;
		}
	}

//	stripColours( gTempStringBuffer, gTempStringBuffer );

	err |= DEBUG_addToOutputBuffer(gTempStringBuffer);

	return err;
}

int DEBUG_strlen( const char* szStr )
{
	// This returns the length of the string in characters, i.e. the length without
	// ANSI colour codes.

	int iLen = strlen( szStr );

    // Don't count RTF ctrl words.
	while( *szStr )
    {
        if( '\\' == *szStr )
        {
            if( '\\' == *(szStr + 1) )
            {
                szStr += 2;
            }
            else
            {
                // Skip slash.
                ++szStr;
                --iLen;
                
                // Find end of control word.
                bool bIsDelimiter  = false;
                bool bPartOfCtrlWd = false; // If it's a delimiter, is it part of the control word?
                bool bSeenHyphen   = false;
                while( true )
                {
                    if( ' ' == *szStr )
                    {
                        bIsDelimiter  = true;
                        bPartOfCtrlWd = true;
                    }
                    else if( '-' == *szStr )
                    {
                        if( bSeenHyphen )
                        {
                            bIsDelimiter = true;
                        }
                        else
                        {
                            bSeenHyphen = true;
                        }
                    }
                    else if( !isalnum(*szStr) )
                    {
                        bIsDelimiter = true;
                    }

                    if( !bIsDelimiter || bPartOfCtrlWd )
                    {
                        ++szStr;
                        --iLen;
                    }

                    if( bIsDelimiter )
                    {
                        break;
                    }
                }
            }
        }
        else
        {
            ++szStr;
        }
	}

    Asura_Assert( iLen >= 0, "" );
	return iLen;
}





	// 'final' is a string containing a bit field's disassembly. addField will add this string
	// to the output. It checks to see that adding the field
	// will go off the end of the current line (there's a pre-defined line width), and
	// if so it will add a new line and indent.

	// This function essentially duplicates the work of sprintf.

	// There's some complex stuff in here because we'd ideally want to pass all the va_args
	// off to sprintf, but that's a little difficult without writing some assembler or
	// using the (nonexistant) varargs macros.

void DEBUG_addField( DBGMSG eType, char* final, const char* szFormat, ... )
{
	char szTemp[1024];

    {
	    va_list xArgs;
	    va_start( xArgs, szFormat );
        Asura_PrintfToBfr( szTemp, szFormat, xArgs, 1024, true );
        va_end( xArgs );
    }

    // Now add the 'temp' string to 'final'.
	char* p = final + strlen(final);
    int i;
	switch( eType )
    {
		case DBGMSG_ENTRY_TITLE:
			DMADebug_Puts("-=-=-=-=-= DEBUG - Adding new TITLE");
			i = (DMADEBUG_INDENT_SIZE - DEBUG_strlen(szTemp)) - 1;
			strcat(p, szTemp);
			while(i-- > 0) {
				strcat(p, " ");
			}
			strcat(p, ":");
			break;

		case DBGMSG_ENTRY_FIELD:
			DMADebug_Puts("-=-=-=-=-= DEBUG - Adding new FIELD");
			while(p > final && *p != '\n') {
				p--;
			}
			if(*p == '\n') p++;

			i = DEBUG_strlen(p) + DEBUG_strlen(szTemp);
					// This will add a new line and indent if the field will go off the end... but won't
					// if it's the first field on the line.
			if(gLineWidth != 0 && i > gLineWidth && DEBUG_strlen(p) > DMADEBUG_INDENT_SIZE)
            {
				strcat( p, "\n" );
                strcat( p, DMADEBUG_FMT_INDENT );
			}
			strcat(p, " ");
			strcat(p, szTemp);
			break;

		case DBGMSG_ENTRY_FIELD_NL:
			DMADebug_Puts("-=-=-=-=-= DEBUG - Adding new FIELD_NL");
			if(p > final && *(p - 1) != '\n') {		// add a NL if the last char was not a new line.
				strcat(p, "\n");
			}
			strcat( p, DMADEBUG_FMT_INDENT );
            strcat( p, " " );
			strcat( p, szTemp );
			break;

		default:
			Asura_Assert( false, "LOGIC ERROR in DEBUG_addField - this type does not exist." );
			break;
	}
}

// Save the output buffers to a file
void DEBUG_saveBuffers( const char *szFilename, bool bAppend )
{
    FILE* pxFile = fopen( szFilename, (bAppend ? "a" : "w") );
    if( pxFile )
    {
        fwrite( DMADEBUG_RTF_STD_HEADER, strlen(DMADEBUG_RTF_STD_HEADER), 1, pxFile );
        fwrite( gOutputBuffer, strlen(gOutputBuffer), 1, pxFile );
        fwrite( DMADEBUG_RTF_STD_FOOTER, strlen(DMADEBUG_RTF_STD_FOOTER), 1, pxFile );
        fclose( pxFile );
    }
    else
    {
        // TODO
    }
}

// initialises all the modules
void DEBUG_init( const char* cs )
{
	DEBUG_parseConfig(cs);
	DMA_parseConfig(cs);
	GIF_parseConfig(cs);
	VIF_parseConfig(cs);
}

void DEBUG_FreeAll( void )
{
    if( gOutputBuffer )
    {
        delete[] gOutputBuffer;
        gOutputBuffer = NULL;
    }
}

// E N D   N A M E S P A C E /////////////////////////////////////////////////

}

