//////////////////////////////////////////////////////////////////////////////
//
// PS2Strip_GeometryTypes.h
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
// Description:
//
//     Geometry-related data structures for PS2Strip code.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___PS2STRIP_GEOMETRYTYPES_H___
#define ___PS2STRIP_GEOMETRYTYPES_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

// Engine independent.
#include "Core/Asura_Hash.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureFlags.h"

// T Y P E S /////////////////////////////////////////////////////////////////

class PS2Strip_Tri
{

    public:

        // O P E R A T O R S /////////////////////////////////////////////////

        bool operator==( const PS2Strip_Tri& xOther ) const
        {
            if( m_uHashID != xOther.m_uHashID ) { return false; }

            for( u_int u = 0; u < 3; ++u )
            {
                if( m_auV[u] != xOther.m_auV[u] ) { return false; }
            }
            
            return true;
        }

        bool operator!=( const PS2Strip_Tri& xOther ) const
        {
            return !(*this == xOther);
        }

        // M E T H O D S /////////////////////////////////////////////////////

        Asura_Hash_ID GetHashID( void ) const { return m_uHashID; }

        u_int GetVtxIdx( u_int uIdxWithinTri ) const { return m_auV[uIdxWithinTri]; }

        void SetVtxIndices( u_int uV0, u_int uV1, u_int uV2 )
        {
            m_auV[0]  = uV0;
            m_auV[1]  = uV1;
            m_auV[2]  = uV2;
            m_uHashID = Asura_GetHashID_Binary( m_auV, 3 * sizeof(u_int) );
        }

        void ChangeVtxIdx( u_int uIdxWithinTri, u_int uNewVtxIdx )
        {
            m_auV[uIdxWithinTri] = uNewVtxIdx;
            m_uHashID            = Asura_GetHashID_Binary( m_auV, 3 * sizeof(u_int) );
        }

    private:
        
        // S T A T E /////////////////////////////////////////////////////////

        // Vertex indices.
        // Typically indices into processed data's vertex triplets.
        u_int m_auV[3];

        Asura_Hash_ID m_uHashID;

};
// End class PS2Strip_Tri.

class PS2Strip_MatInfo
{

    public:

        // O P E R A T O R S /////////////////////////////////////////////////

        bool operator==( const PS2Strip_MatInfo& xOther ) const
        {
            if( m_iMaterialID != xOther.m_iMaterialID ) { return false; }

            // Compare base texture coordinates.
            if( m_afU[0] != xOther.m_afU[0] || m_afV[0] != xOther.m_afV[0] ) { return false; }

            // Only bother comparing secondary texture coordinates if it's
            // a dual texture material. Note that we know at this point that
            // the materials match.
            if( m_bDualTextured )
            {
                if( m_afU[1] != xOther.m_afU[1] || m_afV[1] != xOther.m_afV[1] ) { return false; }
            }

            return true;
        }

        // M E T H O D S /////////////////////////////////////////////////////////

        bool ApproximatelyEquals( const PS2Strip_MatInfo& xOther, float fCoordTol = 0.005f ) const
        {
            if( m_iMaterialID != xOther.m_iMaterialID ) { return false; }

            // Compare base texture coordinates.
            if( Asura_Maths::Abs(m_afU[0] - xOther.m_afU[0]) > fCoordTol ) { return false; }
            if( Asura_Maths::Abs(m_afV[0] - xOther.m_afV[0]) > fCoordTol ) { return false; }

            // Only bother comparing secondary texture coordinates if it's
            // a dual texture material. Note that we know at this point that
            // the materials match.
            if( m_bDualTextured )
            {
                if( Asura_Maths::Abs(m_afU[1] - xOther.m_afU[1]) > fCoordTol ) { return false; }
                if( Asura_Maths::Abs(m_afV[1] - xOther.m_afV[1]) > fCoordTol ) { return false; }
            }

            return true;
        }

        bool CheckTexCoords( void ) const
        {
            for( u_int u = 0; u < (m_bDualTextured ? 2u : 1u); ++u )
            {
                if( m_afU[u] <= -8.0f || m_afU[u] >= 8.0f || m_afV[u] <= -8.0f || m_afV[u] >= 8.0f )
                {
                    return false;
                }
            }
            return true;
        }

        // S T A T E /////////////////////////////////////////////////////////

        int  m_iMaterialID;
        bool m_bDualTextured;

        union
        {
            float m_afUVs[2][ASURA_MATERIAL_MAXTEXTURES];

            struct
            {
                float m_afU[ASURA_MATERIAL_MAXTEXTURES];
                float m_afV[ASURA_MATERIAL_MAXTEXTURES];
            };
        };

};
// End class PS2Strip_MatInfo.

class PS2Strip_VtxIdxQuadlet
{

    public:

        // M E T H O D S /////////////////////////////////////////////////////

        u_int GetNrmIdx( void ) const { return m_uNrmIdx; }
        u_int GetPosIdx( void ) const { return m_uPosIdx; }
        u_int GetColourIdx( void ) const { return m_uColourIdx; }
        u_int GetMatInfoIdx( void ) const { return m_uMatInfoIdx; }

        void SetIndices( u_int uPosIdx, u_int uNrmIdx, u_int uMatInfoIdx, u_int uColourIdx )
        {
            m_uPosIdx     = uPosIdx;
            m_uNrmIdx     = uNrmIdx;
            m_uMatInfoIdx = uMatInfoIdx;
            m_uColourIdx  = uColourIdx;
        }


    private:

        // S T A T E /////////////////////////////////////////////////////////////

        u_int m_uPosIdx;     // Index into list/tree of unique positions.
        u_int m_uNrmIdx;     //          "            "        normals.
        u_int m_uMatInfoIdx; //          "            "        material data.
        u_int m_uColourIdx;  //          "            "        colour data.

};
// End class PS2Strip_VtxIdxQuadlet.

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___PS2STRIP_GEOMETRYTYPES_H___
