//////////////////////////////////////////////////////////////////////////////
//
// PS2Strip_ToolsInterface.cpp
//
// By: Morgan Parry.
//
// Copyright (c) 2002 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "PS2/PS2Strip/PS2Strip_ToolsInterface.h"

// Stdlib independent.
#include <stdarg.h>
#include <stdio.h>

// Engine independent.
#include "Animation/Asura_Hierarchy.h"
#include "Debugging/Asura_Logfile.h"
#ifdef ASURA_USE_BOTANICALS
#include "PS2/Botanicals/Asura_PS2_Botanicals_Shared.h"
#include "PS2/Rendering/Asura_PS2_BotanicalsDefines.h"
#endif // #ifdef ASURA_USE_BOTANICALS
#include "PS2/Rendering/Asura_PS2_GSMDefines.h"
#include "PS2/Rendering/Asura_PS2_SmoothSkinDefines.h"

// Stdlib platform-specific.
#include <windows.h>

// Tools.
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "MeshProcessing/Tools_SubdividedMesh.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor_PS2.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/OS/Tools_OS.h"

// PS2Strip.
#include "PS2/PS2Strip/PS2Strip_Exporter.h"

// C O N S T A N T S /////////////////////////////////////////////////////////

static const unsigned MAX_LOG_LINE_LENGTH = 2048;

// G L O B A L S /////////////////////////////////////////////////////////////

static PS2Strip::EXPORTTYPE g_eCurExportType;
static int g_iCompletionPercentage;
static Asura_String g_xCompletionStatus;
static HANDLE g_pvOutputFile;
static const char* g_szCurAssetName;
static u_int g_uCurSkinLOD;

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

static void StripTrailingNewlines( char* szStr, u_int uLen )
{
    for( char* szIt = szStr + (uLen - 1); szIt >= szStr; --szIt )
    {
        if( '\r' == *szIt || '\n' == *szIt )
        {
            *szIt = '\0';
        }
        else
        {
            break;
        }
    }
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

void PS2Strip::ExportSkinLODThresholdChunk( HANDLE hOut, const Asura_Hierarchy_Skin* pxSkin )
{
    PS2Strip_Exporter::OutputSkinLODThresholdChunk( hOut, pxSkin );
}

#if defined( ASURA_USE_CARPET )
bool PS2Strip::ExportCarpetMeshChunk( HANDLE pOutput )
{
	return PS2Strip_Exporter::OutputCarpetMeshChunk( pOutput );
}
#endif //#if defined( ASURA_USE_CARPET )

#if defined( ASURA_USE_BOTANICALS )
bool PS2Strip::ExportBotanicalsChunk( HANDLE pOutput, const int* const piMaterialConversionList  )
{
	return PS2Strip_Exporter::OutputBotanicalsChunk( pOutput, piMaterialConversionList );
}
#endif //#if defined( ASURA_USE_CARPET )

bool PS2Strip::ExportSkin( HANDLE pvFile, Asura_Hierarchy_Skin* pxSkin, float fUScale, float fVScale )
{
    if( !pxSkin ) { return false; }

    if( pxSkin->IsSkinnedModel() && !pxSkin->GetMesh() )
    {
        LogError( "ERROR: Mesh for skin \"%s\" not loaded.\n", pxSkin->Name.GetString() );
        return false;
    }

    g_eCurExportType        = ( pxSkin->IsSkinnedModel() ? EXPORTTYPE_SKIN_SMOOTH : EXPORTTYPE_SKIN_HIER );
    g_iCompletionPercentage = 0;
    g_pvOutputFile          = pvFile;
    g_szCurAssetName        = pxSkin->Name.GetString();
    g_uCurSkinLOD           = 0;
  
    // Count skin LODs.
    Asura_Hierarchy_Skin* pxLOD = pxSkin;
    u_int uNumLODs = 0;
    while( pxLOD )
    {
        ++uNumLODs;
		pxLOD = pxLOD->m_pxNextLowerSkin;
    }

    float fPercentPerLOD = 100.0f / uNumLODs;
    float fCurPercent = 0.0f;

    // Export LODs in descending order of detail.
    bool bOK = true;
    pxLOD = pxSkin;
    while( bOK && pxLOD )
    {
        // Update status display.
        {
            char szTmpBfr[256];
            Asura_PrintfToBfr( szTmpBfr, "Exporting skin \"%s\"...", 256, true, pxSkin->Name.GetString() );
            g_xCompletionStatus = szTmpBfr;

            LogInfo( szTmpBfr );
            LogInfo( "\n" );

            UpdateCompletionStatus( fCurPercent );
        }

        Asura_Assert( pxLOD->IsSkinnedModel() == pxSkin->IsSkinnedModel(), "" );
        if( pxLOD->IsSkinnedModel() )
        {
		    Tools_SmoothSkin_Processor xProcessor;
			xProcessor.SetIgnoreDoubleSidedFlag(true);
		    if( !xProcessor.Initialise(pxLOD) )
            {
                bOK = false;
                break;
            }

            UpdateCompletionStatus( fCurPercent + fPercentPerLOD * (1.0f / 3.0f) );

		    Tools_SmoothSkin_PS2_Processor xPS2Processor;
		    if( !xPS2Processor.Initialise(&xProcessor) )
            {
                bOK = false;
                break;
            }

            UpdateCompletionStatus( fCurPercent + fPercentPerLOD * (2.0f / 3.0f) );

		    xPS2Processor.SetUVScales( fUScale, fVScale );

		    if( !xPS2Processor.OutputToChunk(pvFile, g_szCurAssetName ) )
            {
                bOK = false;
                break;
            }

            fCurPercent += fPercentPerLOD;
        }
        else
        {
            PS2Strip_Exporter::BeginHierSkinExport( pxLOD, fCurPercent, fCurPercent + fPercentPerLOD );
            PS2Strip_Exporter::SetUVScale( fUScale, fVScale );

            PS2Strip_Exporter::UPDATERESULT eUpdateResult;
            do
            {
                // Check if the user has asked for the export to be aborted.
                if( Tools_Level_Export::IsAbortFlagSet() ) { break; }

                eUpdateResult = PS2Strip_Exporter::Update();
            }
            while( PS2Strip_Exporter::UPDATERESULT_PENDING == eUpdateResult );

            bOK = ( PS2Strip_Exporter::UPDATERESULT_COMPLETE == eUpdateResult );
            fCurPercent += fPercentPerLOD;
        }

        pxLOD = pxLOD->m_pxNextLowerSkin;
        ++g_uCurSkinLOD;
    }

    if( bOK )
    {
        Asura_Assert( g_uCurSkinLOD == uNumLODs, "" );
        UpdateCompletionStatus( 100.0f );
    }

    return bOK;
}

void PS2Strip::ExportEnvironment( HANDLE pvFile, bool bSubdivide )
{
    g_eCurExportType        = EXPORTTYPE_ENV;
    g_iCompletionPercentage = 0;
    g_pvOutputFile          = pvFile;
    g_szCurAssetName        = NULL;

    // Change the message displayed depending on exactly how much work we've
    // been asked to do.
    if( bSubdivide )
    {
        if( Tools_SubdividedMesh::s_bRecalculatePrelighting )
        {
            g_xCompletionStatus = "Subdividing, relighting and stripifying environment...";
        }
        else
        {
            g_xCompletionStatus = "Subdividing and stripifying environment...";
        }
    }
    else
    {
        g_xCompletionStatus = "Stripifying environment...";
    }

    UpdateCompletionStatus( 0.0f );

    PS2Strip_Exporter::BeginEnvExport();
	PS2Strip_Exporter::SetSubdivideFlag(bSubdivide);
    
    PS2Strip_Exporter::UPDATERESULT eUpdateResult;
    while( (eUpdateResult = PS2Strip_Exporter::Update()) == PS2Strip_Exporter::UPDATERESULT_PENDING )
    {
// (MP - 01.05.2002) Not sure this is terribly helpful and it prevents the
// artists from seeing warnings, errors, etc.
//        Tools_Level_Export::AddMessage(
//            Tools_Level_Export::MT_INFORMATION,
//            "Stripified module \"%s\".",
//            PS2Strip_Exporter::GetNameOfLastMesh()
//        );

        // Check if the user has requested for the export to be aborted.
        if( PS2Strip_Exporter::UPDATERESULT_PENDING == eUpdateResult && Tools_Level_Export::IsAbortFlagSet() )
        {
            PS2Strip_Exporter::Abort();
            return;
        }
    }

    if( PS2Strip_Exporter::UPDATERESULT_COMPLETE == eUpdateResult )
    {
        // Report how long it took.
        float fElapsed = PS2Strip_Exporter::Timing_GetTotalElapsedTime();
        Tools_Level_Export::AddMessage( Tools_Level_Export::MT_INFORMATION, "Env stripified and written in %.1f seconds.", fElapsed );
        
        PS2Strip_Exporter::Timing_PrintReport();
    }
    else
    {
        // Export failed.
        Tools_Level_Export::SetAbortFlag();
    }
}

bool PS2Strip::ExportObject( HANDLE pvFile, Asura_Mesh* pxMesh, const char* szName )
{
    // Just to be awkward, we don't deal with multiple LODs here; that's
    // done in the independent calling tools code.

    if( !pxMesh ) { return false; }

    g_eCurExportType        = EXPORTTYPE_OBJECT;
    g_iCompletionPercentage = 0;
    g_pvOutputFile          = pvFile;
    g_szCurAssetName        = szName;
    
    g_xCompletionStatus = "Exporting object \"";
    g_xCompletionStatus += szName;
    g_xCompletionStatus += "\"...";
    UpdateCompletionStatus( 0.0f );

    LogInfo( g_xCompletionStatus.GetString() );
    LogInfo( "\n" );

    PS2Strip_Exporter::BeginObjectExport( pxMesh );

    PS2Strip_Exporter::UPDATERESULT eUpdateResult;
    do
    {
        // Check if the user has asked for the export to be aborted.
        if( Tools_Level_Export::IsAbortFlagSet() ) { break; }

        eUpdateResult = PS2Strip_Exporter::Update();
    }
    while( PS2Strip_Exporter::UPDATERESULT_PENDING == eUpdateResult );

    if( PS2Strip_Exporter::UPDATERESULT_COMPLETE == eUpdateResult )
    {
        UpdateCompletionStatus( 100.0f );
        return true;
    }
    else
    {
        return false;
    }
}

const char* PS2Strip::GetLastSystemError( void )
{
    return Tools_OS::GetLastSystemError();
}

unsigned PS2Strip::GetMaxStripLength( const Asura_Hierarchy_Skin* pxInputSkin )
{
    if( EXPORTTYPE_SKIN_SMOOTH == g_eCurExportType )
    {
	    // Calculate number of verts per strip from number of bones.
	    return ASURA_PS2_SS__MAX_VERTS_PER_STRIP( pxInputSkin->NumberOfBones );
    }
#ifdef ASURA_USE_BOTANICALS
    else if( EXPORTTYPE_BOTANICAL_OBJECT == g_eCurExportType )
	{
	    return ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP;
	}
#endif // #ifdef ASURA_USE_BOTANICALS
	else
    {
	    return ASURA_PS2_GSM__MAX_VERTS_PER_STRIP;
    }
}

HANDLE PS2Strip::GetOutputHandle( void )
{
    return g_pvOutputFile;
}

PS2Strip::EXPORTTYPE PS2Strip::GetCurExportType( void )
{
    return g_eCurExportType;
}

const char* PS2Strip::GetCurExportDescription( void )
{
    switch( g_eCurExportType )
    {
        case EXPORTTYPE_ENV:         return "environment";
        case EXPORTTYPE_OBJECT:      return "object";
        case EXPORTTYPE_SKIN_HIER:   return "hier skin";
        case EXPORTTYPE_SKIN_SMOOTH: return "smooth skin";
		case EXPORTTYPE_BOTANICAL_OBJECT: return "botanical object";

        default:
            Asura_Assert( false, "" );
            return "";
    }
}

void PS2Strip::LogError( const char* szFormat, ... )
{
    char szBuffer[MAX_LOG_LINE_LENGTH];

	va_list xArgList;
	va_start( xArgList, szFormat );
	u_int uNumChars = Asura_PrintfToBfr( szBuffer, szFormat, xArgList, MAX_LOG_LINE_LENGTH, true );
	va_end( xArgList );

    if( !uNumChars ) { return; }
    
    // Write to logfile.
    Asura_SimpleLogfileOutput(( Tools_Level_Export::s_szLogPath, "%s", szBuffer ));
	Asura_SimpleLogfileOutput(( Tools_Level_Export::s_szErrorLogPath, "%s", szBuffer ));
    
    // Display in tools window.
    StripTrailingNewlines( szBuffer, uNumChars );
	Tools_Timed_Message_System::AddMessage( 0xff0000, "%s", szBuffer );

    // Update export error count.
    Tools_Level_Export::RecordError();
}

void PS2Strip::LogWarning( const char* szFormat, ... )
{
    char szBuffer[MAX_LOG_LINE_LENGTH];

	va_list xArgList;
	va_start( xArgList, szFormat );
	u_int uNumChars = Asura_PrintfToBfr( szBuffer, szFormat, xArgList, MAX_LOG_LINE_LENGTH, true );
	va_end( xArgList );
    
    if( !uNumChars ) { return; }
       
    // Write to logfile.
    Asura_SimpleLogfileOutput(( Tools_Level_Export::s_szLogPath, "%s", szBuffer ));

    // Display in tools window.
    StripTrailingNewlines( szBuffer, uNumChars );
	Tools_Timed_Message_System::AddMessage( 0xffff00, "%s", szBuffer );
}

void PS2Strip::LogInfo( const char* szFormat, ... )
{
    char szBuffer[MAX_LOG_LINE_LENGTH];

	va_list xArgList;
	va_start( xArgList, szFormat );
	u_int uNumChars = Asura_PrintfToBfr( szBuffer, szFormat, xArgList, MAX_LOG_LINE_LENGTH, true );
	va_end( xArgList );
    
    if( !uNumChars ) { return; }
    
    // Write to logfile.
    Asura_SimpleLogfileOutput(( Tools_Level_Export::s_szLogPath, "%s", szBuffer ));
}

void PS2Strip::UpdateCompletionStatus( float fPercent, const char* szNewStatus /*= NULL*/, bool bRefreshDisplay /*= true*/ )
{
    g_iCompletionPercentage = Asura_Maths::FloatToInt_Near( fPercent );
    
    if( szNewStatus )
    {
        g_xCompletionStatus = szNewStatus;
    }

    if( bRefreshDisplay )
    {
        Tools_Level_Export::UpdateStatus( Tools_Level_Export::EXPORT_STAGE_PS2_WRITING_STRIPPED_ENV, g_xCompletionStatus.GetString(), g_iCompletionPercentage );
    }
}

const char* PS2Strip::GetCurAssetName( void )
{
    return g_szCurAssetName;
}

u_int PS2Strip::GetCurSkinLOD( void )
{
    return g_uCurSkinLOD;
}

#if defined( ASURA_USE_BOTANICALS )
// JAT 24/03/06 Export the botanicals shape data
// Very like the ExportObject() function and code route
// but different strip length (smaller than GSM because of instance data in VU memory)
bool PS2Strip::ExportBotanicalObject( HANDLE pvFile, Asura_Mesh* pxMesh, const char* szName, enum Asura_PS2_Botanicals_Mesh_Type eType )
{
	if( !pxMesh ) 
	{ 
		return false; 
	}
	if ( eType == ASURA_PS2_BOTANICALS_IMPOSTER )
	{
		return false;
	}

	g_eCurExportType        = EXPORTTYPE_BOTANICAL_OBJECT;
	g_iCompletionPercentage = 0;
	g_pvOutputFile          = pvFile;
	g_szCurAssetName        = szName;

	g_xCompletionStatus = "Exporting botanical object \"";
	g_xCompletionStatus += szName;
	g_xCompletionStatus += ":";
	switch ( eType )
	{
		case ASURA_PS2_BOTANICALS_NORMAL_MESH:
		{
			g_xCompletionStatus += "NORMAL_MESH";
			break;
		}
		case ASURA_PS2_BOTANICALS_2SIDED_MESH:
		{
			g_xCompletionStatus += "2SIDED_MESH";
			break;
		}
		case ASURA_PS2_BOTANICALS_BILLBOARDS:
		{
			g_xCompletionStatus += "BILLBOARDS";
			break;
		}
		default:
		{
			return false;
			break;
		}
	}
	g_xCompletionStatus += "...";
	UpdateCompletionStatus( 0.0f );

	LogInfo( g_xCompletionStatus.GetString() );
	LogInfo( "\n" );

	PS2Strip_Exporter::BeginBotanicalObjectExport( pxMesh, eType );

	PS2Strip_Exporter::UPDATERESULT eUpdateResult;
	do
	{
		// Check if the user has asked for the export to be aborted.
		if( Tools_Level_Export::IsAbortFlagSet() ) { break; }

		if ( ( eType == ASURA_PS2_BOTANICALS_NORMAL_MESH ) || ( eType == ASURA_PS2_BOTANICALS_2SIDED_MESH ) )
		{
			eUpdateResult = PS2Strip_Exporter::Update();
		}
		else
		{
			// For billboards output in a different way
			bool bOK = PS2Strip_Exporter::OutputBotanicalObjectBillboards();
			eUpdateResult = ( bOK ) ? PS2Strip_Exporter::UPDATERESULT_COMPLETE : PS2Strip_Exporter::UPDATERESULT_ERROR;

		}
	}
	while( PS2Strip_Exporter::UPDATERESULT_PENDING == eUpdateResult );

	if( PS2Strip_Exporter::UPDATERESULT_COMPLETE == eUpdateResult )
	{
		UpdateCompletionStatus( 100.0f );
		return true;
	}
	else
	{
		return false;
	}
}

#endif // #if defined( ASURA_USE_BOTANICALS )
