//////////////////////////////////////////////////////////////////////////////
//
// PS2Strip_Tree.h
//
// By: Morgan Parry.
//
// Copyright (c) 2001 Rebellion. All rights reserved.
//
// This tree class does not currently support deletions other than a complete
// deallocation of all nodes. Therefore all indices are in the range 
// [0, tree->m_uNumNodes) and all (integer) values in that range are actually
// used. Also, indices are monotonic increasing as you descend the tree, ie.
// all nodes in a subtree have indices greater than those in their parent
// nodes.
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___PS2STRIP_TREE_H___
#define ___PS2STRIP_TREE_H___ 1

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Debugging/Asura_Assert.h"
#include "Core/Asura_Maths.h"

// It would be nice to avoid this but it allows us to provide inline Cmp()
// implementations for the appropriate types. This in turn allows the
// compiler to optimise wrt the bApprox argument.
#include "PS2/PS2Strip/PS2Strip_GeometryTypes.h"

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

template< typename T >
int PS2Strip_Cmp( const T& x1, const T& x2, bool bApprox );

template<>
static inline int PS2Strip_Cmp<int>( const int& i1, const int& i2, bool bApprox )
{
    Asura_Assert( !bApprox, "This version of PS2Strip_Cmp() doesn't support approximate comparison." );
    return( i1 < i2 ? -1 : (i1 == i2 ? 0 : 1) );
}

template<>
static inline int PS2Strip_Cmp<u_int>( const u_int& u1, const u_int& u2, bool bApprox )
{
    Asura_Assert( !bApprox, "This version of PS2Strip_Cmp() doesn't support approximate comparison." );
    return( u1 < u2 ? -1 : (u1 == u2 ? 0 : 1) );
}

template<>
static inline int PS2Strip_Cmp<const Asura_Vector_3*>( const Asura_Vector_3* const & pxVec1, const Asura_Vector_3* const & pxVec2, bool bApprox )
{
    Asura_Assert( bApprox, "Vec Cmp() only supports approximate comparison." );

    if( pxVec1->ApproximatelyEquals(*pxVec2) )
    {
        return 0;
    }
    else
    {
        if( pxVec1->x == pxVec2->x )
        {
            if( pxVec1->y == pxVec2->y )
            {
                if( pxVec1->z == pxVec2->z )
                {
                    return 0;
                }
                else
                {
                    return( pxVec1->z < pxVec2->z ? -1 : 1 );
                }
            }
            else
            {
                return( pxVec1->y < pxVec2->y ? -1 : 1 );
            }
        }
        else
        {
            return( pxVec1->x < pxVec2->x ? -1 : 1 );
        }
    }
}

template<>
static inline int PS2Strip_Cmp<PS2Strip_TexInfo*>( PS2Strip_TexInfo* const & pxTexInfo1, PS2Strip_TexInfo* const & pxTexInfo2, bool bApprox )
{
    Asura_Assert( bApprox, "This version of PS2Strip_Cmp() only supports approximate comparison." );

    // Check material ID.
    if( pxTexInfo1->m_iMaterialID != pxTexInfo2->m_iMaterialID )
    {
        return( pxTexInfo1->m_iMaterialID < pxTexInfo2->m_iMaterialID ? -1 : 1 );
    }

    // Check UVs.
    const float TEXTURE_EPSILON = 0.005f;
    for( u_int uTexPass = 0; uTexPass < ASURA_MATERIAL_MAXTEXTURES; ++uTexPass )
    {
        // Only bother comparing secondary texture coordinates if there's a
        // secondary texture specified. Note that we know at this point that
        // the materials match.
        if( uTexPass && ASURA_TEXTURE_NONE == Asura_Material_System::GetTextureID(m_iMaterialID, 1) )
        {
            break;
        }

        for( u_int uAxis = 0; uAxis < 2; ++uAxis )
        {
            if( Asura_Maths::Abs(pxTexInfo1->m_afUVs[uAxis][uTexPass] - pxTexInfo2->m_afUVs[uAxis][uTexPass]) >= TEXTURE_EPSILON )
            {
                return( pxTexInfo1->m_afUVs[uAxis][uTexPass] < pxTexInfo2->m_afUVs[uAxis][uTexPass] ? -1 : 1 );
            }
        }
    }

    // Ok, guess they match.
    return 0;
}

static inline u_int PS2Strip_CmpHash( u_int u )
{
	return( u ^ (u >> 4) ^ (u >> 9) ^ (u >> 15) ^ (u >> 22) );
}

template<>
static inline int PS2Strip_Cmp<PS2Strip_VtxIdxQuadlet*>( PS2Strip_VtxIdxQuadlet* const & pxQuadlet1,
                                                         PS2Strip_VtxIdxQuadlet* const & pxQuadlet2,
                                                         bool bApprox )
{
    Asura_Assert( !bApprox, "This version of PS2Strip_Cmp() doesn't support approximate comparison." );

    //If the quadlets being compared aren't equivalent ,use a hash value to do the sorting.
	//Otherwise a very unbalanced tree will be formed , since the values arrive in more-or-less ascending order
    const u_int uHash1 = PS2Strip_CmpHash( pxQuadlet1->m_uPosIdx + pxQuadlet1->m_uTexInfoIdx + pxQuadlet1->m_uNrmIdx + pxQuadlet1->m_uPrelightIdx );
    const u_int uHash2 = PS2Strip_CmpHash( pxQuadlet2->m_uPosIdx + pxQuadlet2->m_uTexInfoIdx + pxQuadlet2->m_uNrmIdx + pxQuadlet2->m_uPrelightIdx );
    
    for( u_int u = 0; u < 4; ++u )
    {
        if( pxQuadlet1->m_auIndices[u] != pxQuadlet2->m_auIndices[u] && UINT_MAX != pxQuadlet1->m_auIndices[u] && UINT_MAX != pxQuadlet2->m_auIndices[u] )
        {
            if( uHash1 == uHash2 )
            {
                return( pxQuadlet1->m_auIndices[u] < pxQuadlet2->m_auIndices[u] ? -1 : 1 );
            }
            else
            {
                return( uHash1 < uHash2 ? -1 : 1 );
            }
        }
    }

    return 0;
}

template<>
static inline int PS2Strip_Cmp<PS2Strip_Tri*>( PS2Strip_Tri* const & pxTri1, PS2Strip_Tri* const & pxTri2, bool bApprox )
{
    Asura_Assert( !bApprox, "This version of PS2Strip_Cmp() doesn't support approximate comparison." );

    const u_int uHash1 = PS2Strip_CmpHash( pxTri1->m_auV[0] + pxTri1->m_auV[1] + pxTri1->m_auV[2] );
	const u_int uHash2 = PS2Strip_CmpHash( pxTri2->m_auV[0] + pxTri2->m_auV[1] + pxTri2->m_auV[2] );	
	
    for( u_int u = 0; u < 3; ++u )
    {
        if( pxTri1->m_auV[u] != pxTri2->m_auV[u] )
        {
            if( uHash1 == uHash2 )
            {
                return( pxTri1->m_auV[u] < pxTri2->m_auV[u] ? -1 : 1 );
            }
            else
            {
                return( uHash1 < uHash2 ? -1 : 1 );
            }
        }
    }

    return 0;
}

// F O R W A R D   D E C L A R A T I O N S ///////////////////////////////////

template< typename T, bool ALLOW_APPROX_EQUALITY_TESTS >
class PS2Strip_Tree;

// C L A S S E S /////////////////////////////////////////////////////////////

template< typename T, bool ALLOW_APPROX_EQUALITY_TESTS >
class PS2Strip_TreeNode
{

    // NB: Entire class is private since it is only for PS2Strip_Tree's use.
    private:

        // F R I E N D S /////////////////////////////////////////////////////

        friend PS2Strip_Tree<T, ALLOW_APPROX_EQUALITY_TESTS>;

        // T Y P E S /////////////////////////////////////////////////////////

// (MP - 21.11.2001) Not used at the moment.
//        typedef void (*FUNCTION)( T& );

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        PS2Strip_TreeNode( const T& xData, unsigned uIndex )
            : m_xData( xData )
            , m_uIndex( uIndex )
            , m_pxLesserChild( NULL )
            , m_pxGreaterChild( NULL )
        {
        }

        ~PS2Strip_TreeNode( void )
        {
            Asura_Assert( !m_pxLesserChild && !m_pxGreaterChild, "Child nodes haven't been deallocated!" );
        }

        // M E T H O D S /////////////////////////////////////////////////////

// (MP - 21.11.2001) Not used at the moment.
//        //////////////////////////////////////////////////////////////////////
//        //
//        // FUNCTION     ApplyFunction().
//        // PARAMETERS   The function to be applied.
//        // RETURNS      None.
//        // DESCRIPTION  Applies the specified function to the left subtree's
//        //              data, this node's data, and then the right subtree's
//        //              data, ie. an inorder walk of the subtree. NB: The
//        //              walk of the tree is not in the order of the nodes'
//        //              indices, rather in ascending order of the data, as
//        //              defined by the data type's comparison operators.
//        // AUTHOR       Morgan Parry - 17.11.2001.
//        //
//        //////////////////////////////////////////////////////////////////////
//        void ApplyFunction( FUNCTION fnFunction )
//        {
//            if( m_pxLesserChild ) { m_pxLesserChild->ApplyFunction( fnFunction ); }
//
//            fnFunction( m_xData );
//
//            if( m_pxGreaterChild ) { m_pxGreaterChild->ApplyFunction( fnFunction ); }
//        }
//        // End function ApplyFunction().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     Deallocate().
        // PARAMETERS   None.
        // RETURNS      None.
        // DESCRIPTION  Deallocates this node, all child nodes, and,
        //              optionally, the data contained within them, depending
        //              on the template parameter bDeallocateData.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        void Deallocate( void )
        {
            if( m_pxLesserChild )
            {
                m_pxLesserChild->Deallocate();
                m_pxLesserChild = NULL;
            }

            if( m_pxGreaterChild )
            {
                m_pxGreaterChild->Deallocate();
                m_pxGreaterChild = NULL;
            }

            delete this;
        }
        // End function Deallocate().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     GetIndex().
        // PARAMETERS   The data to be found.
        // RETURNS      The (non-negative) index of the data within the tree
        //              or else -1 if it wasn't found.
        // DESCRIPTION  Searches for the specified data within the subtree
        //              starting at this node.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        //lint -esym(1763,PS2Strip_TreeNode<*>::GetIndex) [suppress "member fn marked as const indirectly modifies class"; seems to be a Lint bug]
        int GetIndex( const T& xData ) const
        {
            const int iCmp = PS2Strip_Cmp<T>( xData, m_xData, ALLOW_APPROX_EQUALITY_TESTS );
            if( iCmp < 0 )
            {
                return( m_pxLesserChild ? m_pxLesserChild->GetIndex(xData) : -1 );
            }
            else if( iCmp > 0 )
            {
                return( m_pxGreaterChild ? m_pxGreaterChild->GetIndex(xData) : -1 );
            }
            else
            {
                return m_uIndex;
            }
        }
        // End function GetIndex().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     Insert().
        // PARAMETERS   The data to be inserted and the index to be given to
        //              the data if it's new.
        // RETURNS      The (non-negative) index the data was inserted at,
        //              if successful, or else -1. Failure will be due to
        //              lack of memory.
        // DESCRIPTION  Attempts to insert the specified data into the
        //              subtree begining at this node.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        int Insert( const T& xData, unsigned uIndexIfNew )
        {
            const int iCmp = PS2Strip_Cmp<T>( xData, m_xData, ALLOW_APPROX_EQUALITY_TESTS );
            if( iCmp < 0 )
            {
                if( m_pxLesserChild )
                {
                    // Recurse into left subtree.
                    return m_pxLesserChild->Insert( xData, uIndexIfNew );
                }
                else
                {
                    // No left subtree so create one now.
                    m_pxLesserChild = new PS2Strip_TreeNode<T, ALLOW_APPROX_EQUALITY_TESTS>( xData, uIndexIfNew );
                    return( m_pxLesserChild ? uIndexIfNew : -1 );
                }
            }
            else if( iCmp > 0 )
            {
                if( m_pxGreaterChild )
                {
                    // Recurse into right subtree.
                    return m_pxGreaterChild->Insert( xData, uIndexIfNew );
                }
                else
                {
                    // No right subtree so create one now.
                    m_pxGreaterChild = new PS2Strip_TreeNode<T, ALLOW_APPROX_EQUALITY_TESTS>( xData, uIndexIfNew );
                    return( m_pxGreaterChild ? uIndexIfNew : -1 );
                }
            }
            else
            {
                // The data already exists in the tree.
                return m_uIndex;
            }
        }
        // End function Insert().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     PopulateArray().
        // PARAMETERS   Array to be populated.
        // RETURNS      None.
        // DESCRIPTION  Places data in the subtree beginning at this node
        //              into the specified array at each node's index, ie.
        //              a node with an index of i will place its data into
        //              array entry i.
        // AUTHOR       Morgan Parry - 21.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        void PopulateArray( T* axDataArray ) const
        {
            Asura_Assert( axDataArray, "" );
            axDataArray[m_uIndex] = m_xData;

            // Recurse.
            if( m_pxLesserChild ) { m_pxLesserChild->PopulateArray( axDataArray ); }
            if( m_pxGreaterChild ) { m_pxGreaterChild->PopulateArray( axDataArray ); }
        }
        // End function PopulateArray().

        // S T A T E /////////////////////////////////////////////////////////

        // Node data.
        T m_xData;
        const unsigned m_uIndex;

        // Subtrees. NULL iff no subtree in that direction.
        PS2Strip_TreeNode* m_pxLesserChild;
        PS2Strip_TreeNode* m_pxGreaterChild;


};

template< typename T, bool ALLOW_APPROX_EQUALITY_TESTS >
class PS2Strip_Tree
{

    public:

        // C O N S T R U C T O R S   /   D E S T R U C T O R S ///////////////

        PS2Strip_Tree( void )
            : m_pxRoot( NULL )
            , m_uNumNodes( 0 )
        {
        }

        ~PS2Strip_Tree( void )
        {
            if( m_pxRoot )
            {
                DeallocateAll();
            }
            else
            {
                Asura_Assert( m_uNumNodes == 0, "" );
            }
        }

        // M E T H O D S /////////////////////////////////////////////////////

// (MP - 21.11.2001) Not used at the moment.
//        //////////////////////////////////////////////////////////////////////
//        //
//        // FUNCTION     ApplyFunction().
//        // PARAMETERS   The function to be applied to each node's data.
//        // RETURNS      None.
//        // DESCRIPTION  Applies the specified function to each node during an
//        //              inorder walk of the tree, ie. the function is applied
//        //              to the data in the left (lesser) subtree, then the
//        //              data in the node itself, then in the right (greater)
//        //              subtree. NB: The walk of the tree is not in the order
//        //              of the nodes' indices, rather in ascending order of
//        //              the data, as defined by the data type's comparison
//        //              operators.
//        // AUTHOR       Morgan Parry - 17.11.2001.
//        //
//        //////////////////////////////////////////////////////////////////////
//        void ApplyFunction( PS2Strip_TreeNode<T, ALLOW_APPROX_EQUALITY_TESTS>::FUNCTION fnFunction )
//        {
//            if( m_pxRoot ) { m_pxRoot->ApplyFunction( fnFunction ); }
//        }
//        // End function ApplyFunction().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     DeallocateAll().
        // PARAMETERS   None.
        // RETURNS      None.
        // DESCRIPTION  Deallocates all tree nodes and, optionally, the data
        //              contained within them, depending on the value of the
        //              template parameter bDeallocateData.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        void DeallocateAll( void )
        {
            if( m_pxRoot )
            {
                m_pxRoot->Deallocate();
                m_pxRoot = NULL;
                m_uNumNodes = 0;
            }
            else
            {
                Asura_Assert( m_uNumNodes == 0, "" );
            }
        }
        // End function DeallocateAll().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     GetIndex().
        // PARAMETERS   The data to be found.
        // RETURNS      The (non-negative) index of the data within the tree
        //              or else -1 if it wasn't found.
        // DESCRIPTION  Searches for the specified data within the tree.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        int GetIndex( const T& xSearchData ) const
        {
            return( m_pxRoot ? m_pxRoot->GetIndex(xSearchData) : -1 );
        }
        // End function GetIndex().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     GetNumEntries().
        // PARAMETERS   None.
        // RETURNS      The number of data items stored in the tree.
        // DESCRIPTION  See RETURNS.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        unsigned GetNumEntries( void ) const
        {
            return m_uNumNodes;
        }
        // End function GetNumEntries().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     Insert().
        // PARAMETERS   The data to be inserted and a pointer to a boolean.
        //              The pointer may be NULL if the caller is not
        //              interested in whether or not the data was new.
        // RETURNS      The (non-negative) index the data was inserted at,
        //              if successful, or else -1.
        // DESCRIPTION  Attempts to insert the specified data into the tree.
        //              If successful, the specified boolean will be set to
        //              true iff the data was new, ie. iff it did not 
        //              previously exist within the tree. Note that in the
        //              event of insertion failure, the value of the boolean
        //              is undefined. In the event that insertion succeeds
        //              but the data was already in the tree, the pointer
        //              passed in will not be stored, and may therefore be
        //              safely deleted, assuming it is not also pointing to
        //              the data already in the tree.
        // AUTHOR       Morgan Parry - 17.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        int Insert( const T& xData, bool* pbDataWasNew = NULL )
        {
            if( m_pxRoot )
            {
                const int iIndex = m_pxRoot->Insert( xData, m_uNumNodes );
                const bool bDataWasNew = ( iIndex == static_cast<int>(m_uNumNodes) );
                    
                // Update node count if the data was new.
                if( bDataWasNew )
                {
                    ++m_uNumNodes;
                }

                // Set the boolean if necessary.
                if( pbDataWasNew )
                {
                    *pbDataWasNew = bDataWasNew;
                }

                return iIndex;
            }
            else
            {
                Asura_Assert( m_uNumNodes == 0, "" );

                m_pxRoot = new PS2Strip_TreeNode<T, ALLOW_APPROX_EQUALITY_TESTS>( xData, 0 );
                if( m_pxRoot )
                {
                    if( pbDataWasNew )
                    {
                        *pbDataWasNew = true;
                    }
                    return m_uNumNodes++;
                }
                else
                {
                    if( pbDataWasNew )
                    {
                        *pbDataWasNew = true;
                    }
                    return -1;
                }
            }
        }
        // End function Insert().

        //////////////////////////////////////////////////////////////////////
        //
        // FUNCTION     ToArray().
        // PARAMETERS   A pointer to a location where a pointer to the array
        //              should be stored.
        // RETURNS      True iff successful.
        // DESCRIPTION  Stores pointers to node data into an array, ordered
        //              by node indices. This method will allocate an array
        //              and store a pointer to it at the specified location.
        //              It is up to the caller to free the array once it is
        //              no longer required.
        // AUTHOR       Morgan Parry - 21.11.2001.
        //
        //////////////////////////////////////////////////////////////////////
        bool ToArray( T** paxDataArray ) const
        {
            Asura_Assert( paxDataArray, "" );

            // Allocate array.
            *paxDataArray = new T[m_uNumNodes];
            if( !paxDataArray ) { return false; }

            // Populate array.
            if( m_pxRoot ) { m_pxRoot->PopulateArray( *paxDataArray ); }

            // Done.
            return true;
        }
        // End function ToArray().

    private:

        // S T A T E /////////////////////////////////////////////////////////

        PS2Strip_TreeNode<T, ALLOW_APPROX_EQUALITY_TESTS>* m_pxRoot;
        unsigned m_uNumNodes;

};

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

#endif // #ifndef ___PS2STRIP_TREE_H___
