//
// Tools_PFX_Import.h
//
// Created by Andrew Weinkove - 03/08/05
//
// Code for importing PFX from another project. User will select a PFX effect they want
// to import and all of the related PFX files will be copied over.
//
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"


#include "Particles/Tools_PFX_Import.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"

#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"


class Tools_PFX_Importer
{
public:
	static void ImportPFXEffect();
private:
	static bool BrowseForEffectFile(char* pszTitle,Asura_String* pxFilename);
	static void AddFileToList(const Asura_String& xFilename);
	static void ClearList(bool bCopyFiles);
	static void UpOneDirectory(Asura_String *pxPath);

	static List<Asura_String*> s_xFileList;
};

List<Asura_String*> Tools_PFX_Importer::s_xFileList;

void Tools_PFX_Importer::ImportPFXEffect()
{
	// Select the file.
	Asura_String xEffectFilename;
	if (!BrowseForEffectFile("Select Effect File to Import", &xEffectFilename))
	{
		return;
	}
	// Load the effect file. This gets a bit messy because the loading code expects us
	// to have loaded the source and particle files already.
	AsuraAssert_SetIgnoreAll(true);
	Asura_ChunkLoading_System::LoadAsuraFile(xEffectFilename.GetString());
	AsuraAssert_SetIgnoreAll(false);
	// Grab a pointer to the new template.
	Asura_String xEffectName;
	Asura_String xPath;
	xEffectFilename.SplitFilename(&xPath, &xEffectName, NULL);
	const Asura_Hash_ID uEffectHash = Asura_GetHashID(xEffectName.GetString());
	Asura_PFX_Effect_Template* pxEffectTemplate = Asura_PFX_System::GetEffectTempByHash(uEffectHash);
	if (!pxEffectTemplate)
	{
		return;
	}
	AddFileToList(xEffectFilename);
	// Build a list of all the source templates.
	UpOneDirectory(&xPath);
	List<Asura_String> xSourceTemplates;
	Asura_String xSourceTemplateSearchPath = xPath;
	xSourceTemplateSearchPath += "\\*.src";
	Tools_FileSystem::FindFiles(xSourceTemplateSearchPath.GetString(), xSourceTemplates, TOOLS_FILEFIND_RECURSIVE|TOOLS_FILEFIND_IGNOREBACKUPDIR);
	// Build a list of all the particle templates.
	List<Asura_String> xParticleTemplates;
	Asura_String xParticleTemplateSearchPath = xPath;
	xParticleTemplateSearchPath += "\\*.pcl";
	Tools_FileSystem::FindFiles(xParticleTemplateSearchPath.GetString(), xParticleTemplates, TOOLS_FILEFIND_RECURSIVE|TOOLS_FILEFIND_IGNOREBACKUPDIR);
	// Find the source templates used (very slow algorithm could be improved in lots of ways!)
	for (u_int i=0;i<pxEffectTemplate->m_uNumSources;i++)
	{
		const Asura_Hash_ID uSourceHash = pxEffectTemplate->m_pxSources[i].m_uTemplate;
		for(LIF<Asura_String> xFileLIF(&xSourceTemplates);!xFileLIF.done();xFileLIF.next())
		{
			Asura_String xSourceFilename;
			xFileLIF().SplitFilename(NULL, &xSourceFilename, NULL);
			if (Asura_GetHashID(xSourceFilename.GetString()) == uSourceHash)
			{
				// Found a match. We're going to have to load this source to find what particle
				// template it uses. Ignore asserts because the system expects us to the particle 
				// effects already loaded.
				AsuraAssert_SetIgnoreAll(true);
				LoadAsuraFile(xFileLIF().GetString());
				AsuraAssert_SetIgnoreAll(false);
				Asura_PFX_Source_Template* pxSourceTemplate = Asura_PFX_System::GetSourceTempByHash(uSourceHash);
				// Search for the particle template.
				if (pxSourceTemplate)
				{
					for(LIF<Asura_String> xFileLIF2(&xParticleTemplates);!xFileLIF2.done();xFileLIF2.next())
					{			
						Asura_String xParticleFilename;
						xFileLIF2().SplitFilename(NULL, &xParticleFilename, NULL);
						if (Asura_GetHashID(xParticleFilename.GetString()) == pxSourceTemplate->m_uParticleTemplateHash)
						{
							// Add the particle file to the list we'll have to copy across.
							AddFileToList(xFileLIF2());
							break;
						}
					}
				}
				// Add this file to the list we'll have to copy across.
				AddFileToList(xFileLIF());
				Asura_PFX_System::DeleteSourceTemplate(uSourceHash);
				break;
			}
		}		
	}
	Asura_PFX_System::DeleteEffectTemplate(uEffectHash);
	// Copy everything across and clear our list.
	ClearList(true);
}

bool Tools_PFX_Importer::BrowseForEffectFile(char* pszTitle,Asura_String* pxFilename)
{
	Asura_Assert(pxFilename,"");
	Asura_Assert(pszTitle,"");

	static Asura_String s_xDefaultPath = "L:\\";

	Tools_FileSystem_DialogInfo	xFileSelector;

	xFileSelector.DefaultExtension		= ".fct";
	xFileSelector.DefaultFilename		= "";
	xFileSelector.DefaultPath			= s_xDefaultPath.GetString();
	xFileSelector.FileFilter			= "Asura Effect Files\0*.fct;\0";
	xFileSelector.WindowTitle			= pszTitle;
	xFileSelector.AllowMultipleSelect	= false;

	*pxFilename = Tools_FileSystem::GetLoadFilename( xFileSelector );

	if(pxFilename->IsNull()) return false;

	pxFilename->SplitFilename(&s_xDefaultPath,0,0);
	return true;
}

void Tools_PFX_Importer::AddFileToList(const Asura_String& xFilename)
{
	// Build a list of files we'll need to copy across.
	Asura_String *pxNewString = new Asura_String(xFilename);
	s_xFileList.add_entry(pxNewString);
}

void Tools_PFX_Importer::ClearList(bool bCopyFiles)
{
	if (bCopyFiles)
	{
		// Copy all of the relevant files across to our project.
		u_int uNumberOfFilesCopied = 0;
		bool bYesAll = false;
		bool bNoAll = false;
		for(LIF<Asura_String*> xFileLIF(&s_xFileList);!xFileLIF.done();xFileLIF.next())
		{
			Asura_String xFilename;
			Asura_String xExt;
			xFileLIF()->SplitFilename(NULL, &xFilename, &xExt);
			Asura_String xDestination("PFXTemps\\");
			if (!xExt.Compare("fct"))
			{
				xDestination += "Effect\\";
			}
			else if (!xExt.Compare("src"))
			{
				xDestination += "Source\\";
			}
			else if (!xExt.Compare("pcl"))
			{
				xDestination += "Particle\\";
			}
			else 
			{
				Asura_Assert(false, "Unhandled case");
			}
			xDestination += xFilename;
			xDestination += ".";
			xDestination += xExt;
			if (CopyFile(xFileLIF()->GetString(), xDestination.GetString(), false))
			{
				Tools_Timed_Message_System::AddMessage(0x00ff00, "Imported %s", xFileLIF()->GetString());
				uNumberOfFilesCopied++;
			}
			else
			{
				bool bOverwrite;
				if (!bYesAll &&
					!bNoAll)
				{
					Asura_String xQuestion(xDestination.GetString());
					xQuestion += " already exists. Overwrite?";
					YesNoAll_Overlay xQuery(xQuestion);
					if (xQuery.DoModal())
					{
						if (xQuery.AllSelected())
						{
							bYesAll = true;
						}
						bOverwrite = true;
					}
					else
					{
						if (xQuery.AllSelected())
						{
							bNoAll = true;
						}
						bOverwrite = false;
					}
				}
				else 
				{
					bOverwrite = bYesAll;
				}
				if (bOverwrite &&
					!CopyFile(xFileLIF()->GetString(), xDestination.GetString(), true))
				{
					Tools_Timed_Message_System::AddMessage(0xff0000, "Failed to overwrite %s", xDestination.GetString());
					uNumberOfFilesCopied++;
				}
			}
		}
		if (uNumberOfFilesCopied)
		{
			Tools_Timed_Message_System::AddMessage(0xffffff, "%d PFX files imported. \"Reload all Templates\" to use the new files", uNumberOfFilesCopied);
		}
	}
	// Free up the memory used by our list.
	while (s_xFileList.size())
	{
		delete s_xFileList.first_entry();
		s_xFileList.delete_first_entry();
	}
}

void Tools_PFX_Importer::UpOneDirectory(Asura_String *pxPath)
{
	// Strips the last directory off our path.
	// This is a bit dirty - I'm fooling Asura_String::SplitFilename into thinking
	// the final directory is a file (don't think it's actually necessary but this 
	// should insure against any future changes to the implementation of SplitFilename).
	(*pxPath) += ".ext";
	pxPath->SplitFilename(pxPath, NULL, NULL);
}

void Tools_PFX_Import::ImportPFXEffect()
{
	Tools_PFX_Importer::ImportPFXEffect();
}



/// End of file. ///
