///////////////////////////////////////////////////////////////////////////////
//
// File      : Tools_MSF_Conversion.cpp
// Author    : Philip Hall
// Created   : 27/09/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "Sound/Conversion/Tools_MSF_Conversion.h"

// Asura Platform-independent
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_Sound_Event.h"

// Tools
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_Sound_Export.h"
#include "Sound/Conversion/Tools_WAV_Reader.h"

// Misc
#include <windows.h>


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Tools_ConvertMSFFile
// Description : Converts a named WAV file into a PS3 friendly MSF using
//				 the MSEnc.exe tool provided with the multistream tools.
//
// Parameters  : 
//   const char* const szInputFileName                   - Full input file path
//   const char* const szOutputFileName                  - Full output file path
//   const ToolsSoundConversionData* const pxConvertData - Conversion options
//
// Returns     : int - Returns 0 on success, nonzero error code on failure
///////////////////////////////////////////////////////////////////////////////
int Tools_ConvertMSFFile( const char* const szInputFileName,
						   const char* const szOutputFileName,
						   const ToolsSoundConversionData* const pxConvertData )
{
	// Parameter check
	if( !szInputFileName || !szOutputFileName )
	{
		Asura_Assert( false, "NULL parameter passed to Tools_ConvertMSFFile()." );

		return Tools_ConvertMSFFile_Success;
	}

	Tools_WAV_Reader xWAVReader;
	if ( !xWAVReader.ReadHeader(szInputFileName) )
	{
		#if defined ASURA_MAIN
			Tools_Level_Export::AddMessage(Tools_Level_Export::MT_ERROR, "Failed to read %s for MSF conversion",szInputFileName);
		#else
			Asura_Assertf(0,("Failed to read %s for MSF conversion",szInputFileName));
		#endif
		return Tools_ConvertMSFFile_FailedToReadWav;
	}

	// Premade 'channel string' - it's easier to destroy than create
	Asura_Char szChannels[20];
	strncpy(szChannels, "0 1 2 3 4 5 6 7 ", 20);
	
	// Terminate the string early to make the input channel parameters 
	szChannels[ (xWAVReader.GetHeader().nChannels * 2) - 1 ] = 0; 

	// Execute the MSF encoder tool.
	// (MH - 15.01.2008) we use MP3 for all sounds now, except tiny ones
	Asura_Char szCommand[MAX_PATH * 2];
	// compute num samples, assuming 16 bit samples (2 bytes per sample)
	u_int uNumSamples = xWAVReader.GetDataSize() / 2;
	// if num samples is less than double the MP3 packet size (1152) use ADPCM
	bool bTinyFile = uNumSamples < (2 * 1152);
	Asura_String xFormat;
	if( bTinyFile )
	{
		xFormat = "-adpcm";
	}
	else
	{
		xFormat = "-mp3 ";
		
		Asura_String xQuality;
		Asura_StringUtil::ToString( pxConvertData->m_iMaxFrequency, xQuality );

		xFormat += xQuality;

		xFormat += " -fscomp";
	}
	Asura_PrintfToBfr	( szCommand
						, "MSEnc.exe -in \"%s\" %s -out \"%s\" -x -loop %s"
						, (MAX_PATH * 2 ) - 1 
						, true
						, szInputFileName
						, szChannels
						, szOutputFileName
						, xFormat.GetString() );

	PROCESS_INFORMATION xProcessInfo;
	STARTUPINFO xStartupInfo = {0};
	xStartupInfo.cb = sizeof( STARTUPINFO );

	if( !CreateProcess( NULL, szCommand, NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &xStartupInfo, &xProcessInfo ) )
	{
		#if defined ASURA_MAIN
			Tools_Level_Export::AddMessage( Tools_Level_Export::MT_ERROR, "Failed to encode sound. Check that MSEnc.exe is present." );
		#else
			Asura_Assert( false, "Failed to encode sound. Check that MSEnc.exe is present." );
		#endif

		return Tools_ConvertMSFFile_FailedToLoadExe;
	}

	WaitForSingleObject( xProcessInfo.hProcess, INFINITE );

	DWORD uProcessExitCode = static_cast< u_int>( ~0);
	GetExitCodeProcess( xProcessInfo.hProcess, &uProcessExitCode);

	CloseHandle( xProcessInfo.hThread );
	CloseHandle( xProcessInfo.hProcess );

	if(uProcessExitCode != 0)
	{
		Asura_Assertf( 0, ("MSEnc.exe returned an error (%d).", uProcessExitCode ));
		return Tools_ConvertMSFFile_Failed;
	}

	return Tools_ConvertMSFFile_Success;
}


