///////////////////////////////////////////////////////////////////////////////
//
// File      : Tools_WAV_Reader.cpp
// Author    : Philip Hall
// Created   : 01/10/2007
//
// Copyright (C) 2007 Rebellion. All Rights Reserved
//
///////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header
#include "Sound/Conversion/Tools_WAV_Reader.h"

// Asura Platform-independent
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Debugging/Asura_LogFile.h"

// Tools
#include "PC/FileSystem/Tools_File_Wrapper.h"


// WAFE format chunk IDs
static const uint32 CHUNK_RIFF = 'FFIR';
static const uint32 CHUNK_WAVE = 'EVAW';
static const uint32 CHUNK_PAD  = ' DAP';
static const uint32 CHUNK_fmt  = ' tmf';
static const uint32 CHUNK_data = 'atad';


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Tools_WAV_Reader::ReadHeader
// Description : Read a WAV header from the given file
//
// Parameters  : 
//   const char* szFilename - Full path to WAV file to read
//
// Returns     : bool - Returns true on success, false on failure
///////////////////////////////////////////////////////////////////////////////
bool Tools_WAV_Reader::ReadHeader(const char* szFilename)
{
	Tools_File_Wrapper xWAVFile;

	// Open WAV file for reading
	if ( !xWAVFile.Open(szFilename, true) )
	{
		Asura_OutputToDebugger(("[Tools_WAV_Reader::ReadWAVHeader] - Unable to open %s for reading\n", szFilename));
		return false;
	}

	// Buffer the start of the file (quicker and easier than progressive reads though might not catch the 'data' section)
	char acBuffer[8192];
	const int iBufferSize = Asura_Maths::Min(sizeof(acBuffer), xWAVFile.FileSize());

	u_int uBytesRead = xWAVFile.Read( acBuffer, iBufferSize );
	if ( uBytesRead < sizeof(WAVEFORMAT) )
	{
		Asura_OutputToDebugger(("[Tools_WAV_Reader::ReadWAVHeader] - Error reading data from %s\n", szFilename));
		return false;
	}

	// Check for wave header
	if ( FindChunkInData(acBuffer, iBufferSize, CHUNK_WAVE) < 0 )
	{
		Asura_OutputToDebugger(("[Tools_WAV_Reader::ReadWAVHeader] - Missing 'WAVE' chunk\n"));
		return false;
	}

	// Check for format header
	int iOffset = FindChunkInData(acBuffer, iBufferSize, CHUNK_fmt);
	if ( iOffset < 0 )
	{
		Asura_OutputToDebugger(("[Tools_WAV_Reader::ReadWAVHeader] - Missing 'fmt ' chunk\n"));
		return false;
	}
	
	uint32 uFmtSize = *((uint32*)(acBuffer + iOffset));
	uFmtSize;   // unused

	// Extract ATRAC wave information
	m_xWaveFormat = *(reinterpret_cast<WAVEFORMAT*>( acBuffer + iOffset + sizeof(uint32) ));

	// Seek to data
	iOffset	= FindChunkInData(acBuffer, iBufferSize, CHUNK_data);
	if ( iOffset < 0 )
	{
		// (If you get this assert we may not have read the 'data' chunk in - make the read buffer bigger)
		Asura_OutputToDebugger(("[Asura_PS3_Sound_Loader::LoadWAVInfo] - Missing 'data' chunk\n"));
		return false;
	}
	
	m_uDataSize = *(reinterpret_cast<uint32*>(acBuffer + iOffset));
	
	xWAVFile.Read( acBuffer, iBufferSize );
	
	m_iDataOffset = iOffset + sizeof(uint32);
	return true;
}


////PGH////////////////////////////////////////////////////////////////////////
// Function    : Tools_WAV_Reader::FindChunkInData
// Description : Find a named chunk within the given data 
//
// Parameters  : 
//   const char* pcData  - Data to search  in
//   int iDataSize       - Size of searchable data
//   uint32 uChunkName   - 'Name' of chunk to locate
//
// Returns     : nothing
///////////////////////////////////////////////////////////////////////////////
int	Tools_WAV_Reader::FindChunkInData(const char* pcData, int iDataSize, uint32 uChunkName)
{
	const char* pcRead = pcData;

	// Must start with RIFF
	if ( reinterpret_cast<const uint32*>(pcRead)[0] == CHUNK_RIFF )
	{
		pcRead += sizeof(uint32) * 2;  // Skip RIFF & Size field
	}
	else
	{
		return -1;
	}

	// Seach all data until chunk name located
	while ( iDataSize > 0 )
	{
		if ( reinterpret_cast<const uint32*>(pcRead)[0] == uChunkName )
		{
			return sizeof(uint32) + ( int(pcRead) - int(pcData) );
		}
		
		pcRead++;
		iDataSize--;
	}

	return -1;
}
