//////////////////////////////////////////////////////////////////////////////
//
// Tools_XMA_Conversion.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#include "Sound/Conversion/Tools_XMA_Conversion.h"

// Asura Platform-independent
#include "Core/Asura_String.h"
#include "Debugging/Asura_Assert.h"
#include "Sound/Asura_Sound_Event.h"

// Tools
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_Sound_Export.h"

// Misc
#include <windows.h>

/******************************************************************************/
/* Functions                                                                  */
/******************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
// Tools_ConvertXMAFile()
//
// Converts a wave file into an XMA. Does nothing if the file already exists
// and doesn't need reconverting
//
// Parameters:		szInputFileName - Path of input file
//					szOutputFileName - Path of output file
//					pxConvertData - Pointer to conversion options structure
//					puFailureCode - Optional pointer to output failure code
//
// Return value:	true for success, false for failure
//
////////////////////////////////////////////////////////////////////////////////

bool Tools_ConvertXMAFile( const char* const szInputFileName,
						   const char* const szOutputFileName,
						   const ToolsSoundConversionData* const pxConvertData,
						   u_int* const puFailureCode /*= 0*/ )
{
	// Parameter check

	if( !szInputFileName || !szOutputFileName )
	{
		Asura_Assert( false, "NULL parameter passed to Tools_ConvertXMAFile()." );

		if( puFailureCode )
		{
			*puFailureCode = Tools_ConvertXMAFile_InvalidParameter;
		}

		return false;
	}

	// Execute the XMA encoder tool

	Asura_String xQuality;
	Asura_StringUtil::ToString( pxConvertData->m_iMaxFrequency, xQuality );
	Asura_String xCommand( "XMAEncode.exe \"" );
	xCommand += Asura_String( szInputFileName );
	xCommand += "\" /l /b 8 /Q ";
	xCommand += xQuality;
	xCommand += " /t \"";
	xCommand += Asura_String( szOutputFileName );
	xCommand += "\"";

	#if defined ASURA_USE_AUDIO_SYSTEM
		// With XAudio2, need to pack the XMA2 in a standard 'FMT ' chunk instead of 'XMA2'
		xCommand += " /p";
	#endif

	PROCESS_INFORMATION xProcessInfo;
	STARTUPINFO xStartupInfo = {0};
	xStartupInfo.cb = sizeof( STARTUPINFO );

	if( !CreateProcess( NULL, xCommand.GetString(), NULL, NULL, FALSE, CREATE_NO_WINDOW, NULL, NULL, &xStartupInfo, &xProcessInfo ) )
	{
		#if defined ASURA_MAIN
			Tools_Level_Export::AddMessage( Tools_Level_Export::MT_ERROR, "Failed to encode sound. Check that XMAEncode.exe is present." );
		#else
			Asura_Assert( false, "Failed to encode sound. Check that XMAEncode.exe is present." );
		#endif

		if( puFailureCode )
		{
			*puFailureCode = Tools_ConvertXMAFile_FailedToLoadExe;
		}

		return false;
	}

	WaitForSingleObject( xProcessInfo.hProcess, INFINITE );

	// Check the exit code for failure

	DWORD uExitCode = 0;
	if( !GetExitCodeProcess( xProcessInfo.hProcess, &uExitCode ) )
	{
		Asura_Assert( false, "Unable to query exit code for XMAEncode process." );
	}

	CloseHandle( xProcessInfo.hThread );
	CloseHandle( xProcessInfo.hProcess );

	if( uExitCode != 0 )
	{
		const char* const szMsg = "XMAEncode tool reported an error for file %s. Exit code %X.";

		#if defined ASURA_MAIN
			Tools_Level_Export::AddMessage( Tools_Level_Export::MT_ERROR, szMsg, szInputFileName, uExitCode );
		#else
			Asura_Assertf( false, (szMsg, szInputFileName, uExitCode) );
		#endif

		if( puFailureCode )
		{
			*puFailureCode = Tools_ConvertXMAFile_FailedToConvertXMA;
		}
		
		return false;
	}

	if( puFailureCode )
	{
		*puFailureCode = Tools_ConvertXMAFile_Success;
	}

	return true;
}