//
// Tools_SoundEvent_System.cpp
//
// Created by Richard Rice - 06/07/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Sound/Tools_SoundEvent_System.h"

#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Chunk_SoundEvent.h"
#include "Core/Asura_HashTree.h"
#include "Core/Asura_NameTree.h"
#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_Timers.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Cutscene/Asura_Cutscene_Actor.h"
#include "Cutscene/Asura_Cutscene_Attachment.h"
#include "Debugging/Asura_Debug_Timers.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Audio_Voice.h"
#include "Sound/Asura_Sound_Event.h"
#include "Sound/Asura_Sound_Event_System.h"
#include "Sound/Asura_Sound_Event_Storage.h"
#include "Sound/Asura_Sound.h"
#include "Sound/Asura_Phonon.h"
#include "SpecialFX/PFX/Asura_PFX_System.h"
#include "SpecialFX/PFX/Asura_PFX_Templates.h"


//tools
#include "Blueprint/Tools_Blueprint_Edit_Object.h"
#include "Blueprint/Tools_Mode_Blueprint.h"
#include "Hierarchy/Tools_Mode_Hierarchy.h"
#include "Load&Save/Tools_AsuraFileContents.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Load&Save/Tools_LoadedFileList.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Misc/list_tem.hpp"
#include "Misc/Tools_ProgressIndicator.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Sound/Tools_PickSoundEvent_Overlay.h"
#include "Sound/Tools_PickSoundSample_Overlay.h"
#include "Sound/Tools_SoundEvent_Overlay.h"
#include "Sound/Tools_PickSound.h"
#include "Sound/Tools_Phonon_Object.h"



List<Asura_String> Tools_SoundEvent_System::s_xModifiedFiles;
Asura_HashTree Tools_SoundEvent_System::s_xNewEventTree;
Asura_HashTree Tools_SoundEvent_System::s_xNewSampleTree;
bool Tools_SoundEvent_System::s_bMixersModified = false;
bool Tools_SoundEvent_System::s_bGlobalSettingsModified = false;
bool Tools_SoundEvent_System::s_bEverHadMixers = false;
bool Tools_SoundEvent_System::s_bEverHadEventsOrSamples = false;

static Asura_StringTree* g_pxCodeEventStringTree = 0;
static Asura_HashTree* g_pxCodeEventHashTree = 0;

static const char* g_szMixersFilename = "Sounds\\SoundEvents\\Mixers.asr";
static const char* g_szGlobalSettingsFilename = "Sounds\\SoundEvents\\GlobalSettings.asr";

static const char* g_aszDefaultCategory[] =
{
	"Weapons",
	"Environmentals",
	"Foley",
	"Menu",
	"Vehicles",
};
static int g_iNumDefaultCategory = sizeof(g_aszDefaultCategory)/sizeof(const char*);

int Tools_SoundEvent_System::s_aiPlatformSampleRates[TOOLS_PLATFORM_NUM_PLATFORMS];
int Tools_SoundEvent_System::s_aiPlatformCompressionQuality[TOOLS_PLATFORM_NUM_PLATFORMS];
int Tools_SoundEvent_System::s_aiPlatformCompressionRates[TOOLS_PLATFORM_NUM_PLATFORMS][TSS_QUALITY_NUM_ENTRIES];

u_int Tools_SoundEvent_System::s_uMemoryLimit;

Asura_TemplateHashTree<Asura_Hash_ID>* Tools_SoundEvent_System::s_pxWavToSampleTree;

bool Tools_SoundEvent_System::s_bDefaultMaxInstancesChanged = false;
u_int Tools_SoundEvent_System::s_uDefaultMaxInstances = 0;

int Tools_SoundEvent_System::s_iCurrentPlatformDirectory = -1;

List<Asura_Hash_ID> Tools_SoundEvent_System::s_xUnusedEventList;
List<Asura_Hash_ID> Tools_SoundEvent_System::s_xUnusedSampleList;

//This fixes a memory leak with the code event stuff , if tools quits too
//early for the sound event system to have been initialised and shutdown.
class Tools_SoundEvent_DeallocateCodeEventTrees
{
public :
	~Tools_SoundEvent_DeallocateCodeEventTrees()
	{
		delete g_pxCodeEventStringTree;
		g_pxCodeEventStringTree = 0;
		delete g_pxCodeEventHashTree;
		g_pxCodeEventHashTree = 0;
	}
}; 
static Tools_SoundEvent_DeallocateCodeEventTrees s_xInstance;




void Tools_SoundEvent_System::Initialise()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_SoundEvent_System::Initialise()" );

	for( int iPlat = 0; iPlat < TOOLS_PLATFORM_NUM_PLATFORMS; iPlat++ )
	{
		s_aiPlatformSampleRates[iPlat] = 22050;
		s_aiPlatformCompressionQuality[iPlat] = TSS_QUALITY_MEDIUM;

		// Pick some sensible defaults for compression quality (hopefully overwritten when loading global settings below)
		for( int iQual = 0; iQual < TSS_QUALITY_NUM_ENTRIES; ++iQual )
		{
			if( iPlat == TOOLS_PLATFORM_PS3 )
			{
				switch ( iQual )
				{
					case TSS_QUALITY_PERFECT:	s_aiPlatformCompressionRates[iPlat][iQual] = 256;	break;
					case TSS_QUALITY_HIGH:		s_aiPlatformCompressionRates[iPlat][iQual] = 128;	break;
					case TSS_QUALITY_MEDIUM:	s_aiPlatformCompressionRates[iPlat][iQual] = 96;	break;
					case TSS_QUALITY_LOW:		s_aiPlatformCompressionRates[iPlat][iQual] = 64;	break;
					case TSS_QUALITY_VERY_LOW:	s_aiPlatformCompressionRates[iPlat][iQual] = 32;	break;
				}
			}
			else if ( iPlat == TOOLS_PLATFORM_XBOX360 )
			{
				switch ( iQual )
				{
					case TSS_QUALITY_PERFECT:	s_aiPlatformCompressionRates[iPlat][iQual] = 80;	break;
					case TSS_QUALITY_HIGH:		s_aiPlatformCompressionRates[iPlat][iQual] = 60;	break;
					case TSS_QUALITY_MEDIUM:	s_aiPlatformCompressionRates[iPlat][iQual] = 50;	break;
					case TSS_QUALITY_LOW:		s_aiPlatformCompressionRates[iPlat][iQual] = 30;	break;
					case TSS_QUALITY_VERY_LOW:	s_aiPlatformCompressionRates[iPlat][iQual] = 10;	break;
				}
			}
			else
			{
				// Other platforms dont use compression quality settings
				Asura_Assert(	!Tools_SoundEvent_System::DoesPlatformUseCompressionQuality( iPlat ), 
								"Didnt expect this platform to use compression quality settings, fill out some defaults here" );
				break;
			}
			
		}
	}

	s_uMemoryLimit = 30;

	LoadAll();

	s_uDefaultMaxInstances = Asura_Sound_Event_System::GetDefaultMaxInstances();

	Tools_SoundEvent_Overlay::TransferModeData(TOOLSMODEDATAFILE_LOAD);
}

void Tools_SoundEvent_System::Shutdown()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_SoundEvent_System::Shutdown()" );

	Tools_SoundEvent_Overlay::TransferModeData(TOOLSMODEDATAFILE_SAVE);
	
	while(s_xModifiedFiles.size()) s_xModifiedFiles.delete_first_entry();
	s_xNewEventTree.Reset();
	s_xNewSampleTree.Reset();

	delete g_pxCodeEventStringTree;
	g_pxCodeEventStringTree = 0;
	delete g_pxCodeEventHashTree;
	g_pxCodeEventHashTree = 0;

	delete s_pxWavToSampleTree;
	s_pxWavToSampleTree = 0;

	s_xUnusedEventList.empty();
	s_xUnusedSampleList.empty();
}


void Tools_SoundEvent_System::Update()
{
	if(s_bDefaultMaxInstancesChanged)
	{
		s_bDefaultMaxInstancesChanged = false;
		Asura_Sound_Event_System::SetDefaultMaxInstances(s_uDefaultMaxInstances);
		SetGlobalSettingsModified();
	}
}



void Tools_SoundEvent_System::SetModified(Asura_Sound_Sample* pxSample)
{
	Asura_Assert(pxSample,"");
	SetModified(pxSample->GetFilename());

	if(s_pxWavToSampleTree)
	{
		delete s_pxWavToSampleTree;
		s_pxWavToSampleTree = 0;
	}
}

void Tools_SoundEvent_System::SetModified(Asura_Sound_Event* pxEvent)
{
	Asura_Assert(pxEvent,"");
	SetModified(pxEvent->GetFilename());
}


void Tools_SoundEvent_System::SetModified(const char* szFilename)
{
	Asura_Assert(szFilename && *szFilename,"Invalid filename");

	//file already in list?
	for(LIF<Asura_String> xModifiedLIF(&s_xModifiedFiles);!xModifiedLIF.done();xModifiedLIF.next())
	{
		if(!xModifiedLIF().Compare(szFilename)) return;
	}

	//File wasn't already flagged as modified ,so check whether its going to be okay to write to it
	Asura_String xFullFilename;
	GetFullFilename(szFilename,xFullFilename);

	Tools_SaveAsura::WarnIfReadOnly(xFullFilename.GetString());	

	s_xModifiedFiles.add_entry(szFilename);
}

void Tools_SoundEvent_System::SetMixersModified()
{
	if(s_bMixersModified) return;

	Tools_SaveAsura::WarnIfReadOnly(g_szMixersFilename);	
	s_bMixersModified = true;
	
}

void Tools_SoundEvent_System::SetGlobalSettingsModified()
{
	if(s_bGlobalSettingsModified) return;

	Tools_SaveAsura::WarnIfReadOnly(g_szGlobalSettingsFilename);	
	s_bGlobalSettingsModified = true;
	
}


void Tools_SoundEvent_System::GetFullFilename(const char* szFilename,Asura_String& xFullFilename)
{
	char szTemp[MAX_PATH];
	if(s_iCurrentPlatformDirectory>=0)
	{
		Asura_PrintfToBfr(szTemp,"Sounds\\SoundEvents\\%s\\%s.SoundEvent",sizeof(szTemp),true,Tools_Platform::GetPlatformName(s_iCurrentPlatformDirectory),szFilename);
	}
	else
	{
		Asura_PrintfToBfr(szTemp,"Sounds\\SoundEvents\\%s.SoundEvent",sizeof(szTemp),true,szFilename);
	}
	xFullFilename = szTemp;
}


bool Tools_SoundEvent_System::IsModified()
{
	return s_xModifiedFiles.size() > 0 || s_bMixersModified || s_bGlobalSettingsModified;
}

void Tools_SoundEvent_System::SaveAll()
{
	//Sound events don't get saved in this way during level exports.RAR 14/07/2006
	if(Tools_Level_Export::IsExport()) return;
	

	int iJunkIndex;
	Asura_NameTree xFileTree;

	//attempt to save all the modified files
	//(this ensures that any files that have lost all their samples/events will get saved)
	for(LIF<Asura_String> xModifiedLIF(&s_xModifiedFiles);!xModifiedLIF.done();)
	{
		//keep track of the files done
		xFileTree.FindNodeIndexByName(xModifiedLIF().GetString(),true,iJunkIndex);

		if(SaveFile(xModifiedLIF().GetString()))
		{
			xModifiedLIF.delete_current();
		}
		else
		{
			xModifiedLIF.next();
		}
	}

	//scan through all samples/events for other files to save
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		const char* szFilename = pxEvent->GetFilename();

		if(!xFileTree.FindNodeIndexByName(szFilename,true,iJunkIndex))
		{
			SaveFile(szFilename);
		}
	}

	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);
		const char* szFilename = pxSample->GetFilename();

		if(!xFileTree.FindNodeIndexByName(szFilename,true,iJunkIndex))
		{
			SaveFile(szFilename);
		}
	}

	SaveMixers();

	SaveGlobalSettings();
}

void Tools_SoundEvent_System::SaveModified()
{
	//Sound events don't get saved in this way during level exports.RAR 14/07/2006
	if(Tools_Level_Export::IsExport()) return;

	for(LIF<Asura_String> xModifiedLIF(&s_xModifiedFiles);!xModifiedLIF.done();)
	{
		if(SaveFile(xModifiedLIF().GetString()))
		{
			xModifiedLIF.delete_current();
		}
		else
		{
			xModifiedLIF.next();
		}
	}

	if(s_bMixersModified)
	{
		SaveMixers();
	}

	if(s_bGlobalSettingsModified)
	{
		SaveGlobalSettings();
	}

}

bool Tools_SoundEvent_System::SaveFile(const char* szFilename)
{
	//Sound events don't get saved in this way during level exports.RAR 14/07/2006
	if(Tools_Level_Export::IsExport()) return false;

	Asura_String xFullFilename;
	GetFullFilename(szFilename,xFullFilename);
	if(!Tools_SaveAsura::CanSaveFile(xFullFilename.GetString(),true))return false;

	//TBD Do something about making backups?


	Asura_SaveBlock xSaveBlock;

	//Attempt to save the samples in an order that will result in a balanced tree on reload. RAR 17/08/2006
	if(Asura_Sound_Event_Storage::GetNumberOfSamples())
	{
		for(Asura_TemplateHashTree_SaveIterator xIt(Asura_Sound_Event_Storage::GetNumberOfSamples());!xIt.Done();xIt.Next())
		{
			const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(xIt.GetCurrent());
			if(!stricmp(pxSample->GetFilename(),szFilename))
			{
				Asura_Chunk_SoundEvent_Sample::Create(pxSample,&xSaveBlock);
			}
		}
	}

	//And save the events
	if(Asura_Sound_Event_Storage::GetNumberOfEvents())
	{
		for(Asura_TemplateHashTree_SaveIterator xIt(Asura_Sound_Event_Storage::GetNumberOfEvents());!xIt.Done();xIt.Next())
		{
			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(xIt.GetCurrent());
			if(!stricmp(pxEvent->GetFilename(),szFilename))
			{
				Asura_Chunk_SoundEvent_Event::Create(pxEvent,&xSaveBlock);
			}
		}
	}
 
	//write out the save block

	xSaveBlock.AddAsuraTerminator();
	Tools_FileSystem::EnsureDirectoryExists(xFullFilename.GetString());
	bool bSuccess = xSaveBlock.SaveFile(xFullFilename.GetString(),true);
    if(bSuccess)
	{
		Tools_Timed_Message_System::AddMessage(0x00ff00,"Saved %s.",xFullFilename.GetString());
		Tools_LoadedFileList::UpdateModifiedTime(xFullFilename.GetString());
	}
	else
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to save %s",xFullFilename.GetString());
	}

	return bSuccess;	
}


bool Tools_SoundEvent_System::SaveMixers()
{
	//Sound mixers don't get saved in this way during level exports.RAR 14/07/2006
	if(Tools_Level_Export::IsExport()) return false;

	if(!Tools_SaveAsura::CanSaveFile(g_szMixersFilename,true))return false;

	//TBD Do something about making backups?

	if(!s_bEverHadMixers)
	{
		//Don't bother creating the mixers file if there have never been any mixers.
		//(Prevents people getting writeable local copies of essentially empty files).
		//RAR 26/10/2006
		s_bMixersModified = false;
		return false;
	}

	Asura_SaveBlock xSaveBlock;

	Asura_Chunk_SoundEvent_Mixer::Create(&xSaveBlock);
 
	//write out the save block

	xSaveBlock.AddAsuraTerminator();
	Tools_FileSystem::EnsureDirectoryExists(g_szMixersFilename);
	bool bSuccess = xSaveBlock.SaveFile(g_szMixersFilename,true);
    if(bSuccess)
	{
		Tools_Timed_Message_System::AddMessage(0x00ff00,"Saved %s.",g_szMixersFilename);
		Tools_LoadedFileList::UpdateModifiedTime(g_szMixersFilename);
		s_bMixersModified = false;
	}
	else
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to save %s",g_szMixersFilename);
	}

	return bSuccess;	
}

bool Tools_SoundEvent_System::SaveGlobalSettings()
{
	//These aren't needed for level exports.RAR 11/08/2006
	if(Tools_Level_Export::IsExport()) return false;

	if(!Tools_SaveAsura::CanSaveFile(g_szGlobalSettingsFilename,true))return false;

	//TBD Do something about making backups?


	if(!s_bEverHadEventsOrSamples)
	{
		//Don't bother creating the global sound events file if there have never been any sound events.
		//(Prevents people getting writeable local copies of essentially empty files).
		//RAR 26/10/2006
		s_bGlobalSettingsModified = false;
		return false;
	}

	Asura_SaveBlock xSaveBlock;
	Asura_Chunk_SoundEvent_ToolsGlobalSettings::Create(&xSaveBlock);
	Asura_Chunk_SoundEvent_GlobalSettings::Create(&xSaveBlock);

	//write out the save block

	xSaveBlock.AddAsuraTerminator();
	Tools_FileSystem::EnsureDirectoryExists(g_szGlobalSettingsFilename);
	bool bSuccess = xSaveBlock.SaveFile(g_szGlobalSettingsFilename,true);
    if(bSuccess)
	{
		Tools_Timed_Message_System::AddMessage(0x00ff00,"Saved %s.",g_szGlobalSettingsFilename);
		Tools_LoadedFileList::UpdateModifiedTime(g_szGlobalSettingsFilename);
		s_bGlobalSettingsModified = false;
	}
	else
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to save %s",g_szGlobalSettingsFilename);
	}

	return bSuccess;	
}



void Tools_SoundEvent_System::LoadAll()
{
	//Global settings
	LoadAsuraFile(g_szGlobalSettingsFilename);

	//Mixers
	LoadAsuraFile(g_szMixersFilename);

	//sound events
	Tools_FileSystem::LoadAsuraFiles("Sounds\\SoundEvents\\*.SoundEvent");

	//Now need to go through all the events and sort out the pointers to the playables
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		pxEvent->UpdatePlayableListFromIDList();
	}
}

void Tools_SoundEvent_System::SetFilesToCheckForExternalModification()
{
	int iJunkIndex;
	Asura_NameTree xFileTree;

	//scan through all samples/events for files being used
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		const char* szFilename = pxEvent->GetFilename();

		if(!xFileTree.FindNodeIndexByName(szFilename,true,iJunkIndex))
		{
			Asura_String xFullFilename;
			GetFullFilename(szFilename,xFullFilename);

			Tools_LoadedFileList::TrackExternalModification(xFullFilename.GetString());
		}
	}

	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);
		const char* szFilename = pxSample->GetFilename();

		if(!xFileTree.FindNodeIndexByName(szFilename,true,iJunkIndex))
		{
			Asura_String xFullFilename;
			GetFullFilename(szFilename,xFullFilename);

			Tools_LoadedFileList::TrackExternalModification(xFullFilename.GetString());
		}
	}
	
	Tools_LoadedFileList::TrackExternalModification(g_szMixersFilename);
	Tools_LoadedFileList::TrackExternalModification(g_szGlobalSettingsFilename);
}

void Tools_SoundEvent_System::RegisterNewEvent(const class Asura_Sound_Event* pxEvent)
{
	int iJunk;
	s_xNewEventTree.FindNodeIndexByHash(pxEvent->GetHashID(),true,iJunk);
	s_bEverHadEventsOrSamples = true;
}

void Tools_SoundEvent_System::RegisterNewSample(const class Asura_Sound_Sample* pxSample)
{
	int iJunk;
	s_xNewSampleTree.FindNodeIndexByHash(pxSample->GetHashID(),true,iJunk);
	s_bEverHadEventsOrSamples = true;

}

bool Tools_SoundEvent_System::CanTransferEvent(const class Asura_Sound_Event* pxEvent,const char* szDestFile)
{
	//If this is a new event , then allow it
	int iJunk;
	if(s_xNewEventTree.FindNodeIndexByHash(pxEvent->GetHashID(),false,iJunk)) return true;

	const char* szSourceFile = pxEvent->GetFilename();

	//Ignore if the source and dest are the same
	if(!stricmp(szSourceFile,szDestFile)) return true;

	//make sure the source and destination files can be written to
	Asura_String xFullFilename;
	GetFullFilename(szSourceFile,xFullFilename);
	if(!Tools_SaveAsura::CanSaveFile(xFullFilename.GetString(),false))return false;

	GetFullFilename(szDestFile,xFullFilename);
	if(!Tools_SaveAsura::CanSaveFile(xFullFilename.GetString(),false))return false;

	return true;
}

bool Tools_SoundEvent_System::CanTransferSample(const class Asura_Sound_Sample* pxSample,const char* szDestFile)
{
	//If this is a new sample , then allow it
	int iJunk;
	if(s_xNewSampleTree.FindNodeIndexByHash(pxSample->GetHashID(),false,iJunk)) return true;

	const char* szSourceFile = pxSample->GetFilename();

	//Ignore if the source and dest are the same
	if(!stricmp(szSourceFile,szDestFile)) return true;

	//make sure the source and destination files can be written to
	Asura_String xFullFilename;
	GetFullFilename(szSourceFile,xFullFilename);
	if(!Tools_SaveAsura::CanSaveFile(xFullFilename.GetString(),false))return false;

	GetFullFilename(szDestFile,xFullFilename);
	if(!Tools_SaveAsura::CanSaveFile(xFullFilename.GetString(),false))return false;

	return true;
}


void Tools_SoundEvent_System::AddOverlayControls(class Asura_Overlay* pxOverlay,class Asura_Vector_2& xOffset)
{
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Function_Dialog("Edit Sample",xOffset,EditSample);
	pxDialog->SetHelpText("Edit or create a sound event sample");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Edit Event",xOffset,EditEvent);
	pxDialog->SetHelpText("Edit or create a sound event");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Combine Overlays",xOffset,&Tools_SoundEvent_Overlay::s_bCombineEventAndSampleOverlay);
	pxDialog->SetHelpText("If ticked then the event and sample overlays will be combined into one.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Default Max Instances", xOffset, &s_uDefaultMaxInstances, 0, 999, 1, "%d", &s_bDefaultMaxInstancesChanged);
	pxDialog->SetHelpText("Default maximum number of instances of each sound event to play.\nIf this is set to zero then there is no maximum.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y += Asura_Dialog::DefaultSpacing();
}


void Tools_SoundEvent_System::EditSample()
{
	Tools_PickSoundSample_Overlay xOverlay("Edit/Create Samples",true);
	xOverlay.DoModal(0);

	for(int i=0;i<xOverlay.GetNumSelected();i++)
	{
		Tools_SoundEvent_Overlay::EditSample(xOverlay.GetSelection(i));
	}
}

void Tools_SoundEvent_System::EditEvent()
{
	Tools_PickSoundEvent_Overlay xOverlay("Edit/Create Events",true);
	xOverlay.DoModal(0);

	for(int i=0;i<xOverlay.GetNumSelected();i++)
	{
		Tools_SoundEvent_Overlay::EditEvent(xOverlay.GetSelection(i));
	}
}

int Tools_SoundEvent_System::GetNumDefaultCategories()
{
	return g_iNumDefaultCategory;	
}

const char* Tools_SoundEvent_System::GetDefaultCategory(int iIndex)
{
	Asura_Assert(iIndex>=0 && iIndex<g_iNumDefaultCategory,"");
	return g_aszDefaultCategory[iIndex];
}



const char* Tools_SoundEvent_System::GetDefaultCategoryForWavFile(const char* szFilename)
{
	static const char* s_aaszDefaults[][2] =
	{
		{"sounds\\weapons\\",			"Weapons"},
		{"sounds\\vehicles\\",		"Vehicles"},
		{"sounds\\characters\\",		"Foley"},
		{"sounds\\menu\\",			"Menu"},
		{"sounds\\environmentals\\",	"Environmentals"},
		{0,0}
	};
	
	//Replace /'s with \'s
	Asura_String xFilenameCopy = szFilename;
	xFilenameCopy.ReplaceAllChars('/','\\');
	szFilename = xFilenameCopy.GetString();


	for(int i=0;s_aaszDefaults[i][0]!=0;i++)
	{
		int iLength = strlen(s_aaszDefaults[i][0]);
		if(!strnicmp(szFilename,s_aaszDefaults[i][0],iLength))
		{
			return s_aaszDefaults[i][1];
		}
	}
	return 0;
}



static ListSelection_Overlay* g_pxCreateCodeEventListOverlay;
static void CreateCodeEventCallback(const Asura_String& xName,int iIndex)
{
	Asura_Hash_ID uHash = Asura_GetHashID(xName.GetString());
	if(!Asura_Sound_Event_Storage::FindEvent(uHash))
	{
		//This event doesn't exist yet
		g_pxCreateCodeEventListOverlay->AddItem(xName.GetString());
	}
}

Asura_Hash_ID Tools_SoundEvent_System::CreateCodeEvent(const char* szFile /*= 0*/,const char* szCategory /*= 0*/,const char* szSubCategory /*= 0*/)
{
	//build a list of all code events that still need to be created
	ListSelection_Overlay xOverlay("Select event to create");

	if(g_pxCodeEventStringTree)
	{
		g_pxCreateCodeEventListOverlay = &xOverlay;
		g_pxCodeEventStringTree->Traverse(CreateCodeEventCallback);
	}

	if(!xOverlay.GetNumItems())
	{
		ShowMessageOverlay("All code events have already been created.");
		return 0;
	}

	if(!xOverlay.DoModal()) return 0;

	//create the selected event
	return CreateEvent(szFile,szCategory,szSubCategory,xOverlay.GetSelectionString());
}


Asura_Hash_ID Tools_SoundEvent_System::CreateEvent(const char* szFile,const char* szCategory,const char* szSubCategory,const char* szEventName)
{
	
	Asura_Hash_ID uEventHash = 0;
	GetString_Overlay xStringOverlay("Name new event","");

	if(szEventName && *szEventName)
	{
		//confirm that the event doesn't already exist
		uEventHash = Asura_GetHashID(szEventName);
		if(Asura_Sound_Event_Storage::GetEventByHash(uEventHash))
		{
			ShowMessageOverlay("An event already exists using the same hash ID.");
			return 0;
		}
	}
	else
	{
		//no name has been given , so ask for one
		do
		{
			if(!xStringOverlay.DoModal()) return 0;
			
			szEventName = xStringOverlay.GetString();
			if(!szEventName || !*szEventName) return 0;

			//make sure this event doesn't already exist
			uEventHash = Asura_GetHashID(szEventName);
			if(Asura_Sound_Event_Storage::GetEventByHash(uEventHash))
			{
				ShowMessageOverlay("An event already exists using the same hash ID.");
				continue;
			}

			break;
			
		}while(true);
	}

	bool bAlreadyExists;
	Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::CreateEvent(szEventName, bAlreadyExists);
	Asura_Assert(pxEvent && !bAlreadyExists, "Error creating sound event");
	//Set the initial category and file
	if(szCategory && *szCategory)
	{
		pxEvent->SetEventType(szCategory);
	}
	if(szSubCategory && *szSubCategory)
	{
		pxEvent->SetEventSubType(szSubCategory);
	}
	if(szFile && *szFile)
	{
		pxEvent->SetFilename(szFile);
	}
	else
	{
		pxEvent->SetFilename(GetDefaultFile());
	}

	SetModified(pxEvent);
	RegisterNewEvent(pxEvent);

	//edit the new event
	Tools_SoundEvent_Overlay::EditEvent(uEventHash);
	return uEventHash;
		
}

void Tools_SoundEvent_System::CreateSample(const char* szFile,const char* szCategory,const char* szSubCategory,List<Asura_Hash_ID>* pxOutputSampleList)
{

	//pick wav files for the new samples
#if defined ASURA_USE_AUDIO_SYSTEM
	List<Asura_Hash_ID> xSoundList;
	Tools_PickSound::PickSoundForSoundEventSample(ASURA_HASH_ID_UNSET,&xSoundList);
#else
	List<Asura_Sound_ID> xSoundList;
	Tools_PickSound::PickSoundForSoundEventSample(ASURA_SID_NO_SOUND,&xSoundList);
#endif

	if(!xSoundList.size()) return;

	//create a sample for each wav selected

	#if defined ASURA_USE_AUDIO_SYSTEM
		LIF<Asura_Hash_ID> xSoundLIF(&xSoundList);
	#else
		LIF<Asura_Sound_ID> xSoundLIF(&xSoundList);
	#endif

	for(;!xSoundLIF.done();xSoundLIF.next())
	{
		#if defined ASURA_USE_AUDIO_SYSTEM
			Asura_Hash_ID uWave = xSoundLIF();
			Asura_String xWavFilename = Asura_Audio_System::GetWaveFilename(uWave);
		#else
			Asura_Sound_ID eSoundID = xSoundLIF();
			Asura_String xWavFilename = Asura_Sound_System::GetFileNameFromSoundID(eSoundID);
		#endif

		if(xWavFilename.IsBlank()) continue;

		//get a default name for the sample
		Asura_String xWavName;
		xWavFilename.SplitFilename(0,&xWavName,0);

		char szSampleName[200];
		Asura_Hash_ID xSampleNameHash = ASURA_HASH_ID_UNSET;

		for(int iIndex=0;;iIndex++)
		{
			if(iIndex)
				Asura_PrintfToBfr(szSampleName,"%s%d",200,false,xWavName.GetString(),iIndex);
			else
				Asura_PrintfToBfr(szSampleName,"%s",200,false,xWavName.GetString());

			xSampleNameHash = Asura_GetHashID( szSampleName );

			if ( !Asura_Sound_Event_Storage::GetSampleByHash( xSampleNameHash ) ) break;
		}

		//Wav files in some directories get assigned default categories
		const char* szDefaultCategory = GetDefaultCategoryForWavFile(xWavFilename.GetString());
		if(szDefaultCategory)
		{
			szCategory = szDefaultCategory;
		}

		bool bAlreadyExists;
		Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::CreateSample(szSampleName, bAlreadyExists);
		Asura_Assert(pxSample && !bAlreadyExists, "Error creating new sample");

		//Set the initial category and file
		if(szCategory && *szCategory)
		{
			pxSample->SetSoundType(szCategory);
		}
		if(szSubCategory && *szSubCategory)
		{
			pxSample->SetSoundSubType(szSubCategory);
		}
		if(szFile && *szFile)
		{
			pxSample->SetFilename(szFile);
		}
		else
		{
			pxSample->SetFilename(Tools_SoundEvent_System::GetDefaultFile());
		}

		#if defined ASURA_USE_AUDIO_SYSTEM
			pxSample->m_uWave = uWave;
		#else
			pxSample->m_eSoundID = eSoundID;
		#endif
		
		SetModified(pxSample);
		RegisterNewSample(pxSample);

		//edit the new sample
		Tools_SoundEvent_Overlay::EditSample( xSampleNameHash );

		//add to the output list if one was provided
		if(pxOutputSampleList) pxOutputSampleList->add_entry(xSampleNameHash);
	}




}


void Tools_SoundEvent_System::RegisterCodeEvent(const char* szName)
{
	if(!szName || !*szName) return;

	//Allocate the trees if this is the first time
	if(!g_pxCodeEventStringTree)
	{
		g_pxCodeEventStringTree = new Asura_StringTree;
		g_pxCodeEventHashTree = new Asura_HashTree;
	}
	
	int iJunk;
	if(!g_pxCodeEventStringTree->FindNodeIndexByName(szName,true,iJunk))
	{
		//was a new name , so add to the hash tree as well
		Asura_Hash_ID uHash = Asura_GetHashID(szName);
		if(g_pxCodeEventHashTree->FindNodeIndexByHash(uHash,true,iJunk))
		{
			Asura_Assertf(0,("Code sound event '%s' has the same hash ID as a previous event",szName));
		}
	}
}

bool Tools_SoundEvent_System::IsCodeEvent(Asura_Hash_ID uEventHash)
{
	//hash tree won't exist if there are no code events
	if(!g_pxCodeEventHashTree) return false;

	int iJunk;
	return g_pxCodeEventHashTree->FindNodeIndexByHash(uEventHash,false,iJunk);
}


#ifdef ASURA_USE_AUDIO_SYSTEM
void Tools_SoundEvent_System::ListSamplesUsedByEvent(const class Asura_Sound_Event* pxEvent,List<Asura_Hash_ID>& xSoundList)
#else
void Tools_SoundEvent_System::ListSoundIDsUsedByEvent(const class Asura_Sound_Event* pxEvent,List<int>& xSoundList)
#endif
{
	List<const Asura_Sound_Event*> xEventList;
	xEventList.add_entry(pxEvent);

	// This is a dynamically changing list - as we iterate through it, events may get added
	for(LIF<const Asura_Sound_Event*> xEventLIF(&xEventList);!xEventLIF.done();xEventLIF.next())
	{
		//check all playables used by this event
		for(u_int i=0;i<xEventLIF()->GetNumPlayables();i++)
		{
			const Asura_Sound_Playable* pxPlayable = xEventLIF()->GetPlayableSound(i);
			if(pxPlayable->IsEvent())
			{
				//add any events to the event list to be checked in turn
				const Asura_Sound_Event* pxChildEvent = (const Asura_Sound_Event*) pxPlayable;
				if(!xEventList.contains(pxChildEvent))
				{
					xEventList.add_entry(pxChildEvent);
				}
			}
			if(pxPlayable->IsSample())
			{
				const Asura_Sound_Sample* pxSample = (const Asura_Sound_Sample*) pxPlayable;

				#if defined ASURA_USE_AUDIO_SYSTEM
				int iSoundID = pxSample->GetHashID();	
				#else
				int iSoundID = pxSample->GetSoundID();	
				#endif

				//add any sound ID used by this sample
				if(!xSoundList.contains(iSoundID))
				{
					xSoundList.add_entry(iSoundID);
				}
			}

		}
	}
}

int Tools_SoundEvent_System::GetPlatformSampleRate(int iPlatform)
{
	Asura_Assert(iPlatform<TOOLS_PLATFORM_NUM_PLATFORMS,"Unknown platform");
	
	return s_aiPlatformSampleRates[iPlatform];
}

void Tools_SoundEvent_System::SetPlatformSampleRate(int iPlatform , int iSampleRate)
{
	Asura_Assert(iPlatform<TOOLS_PLATFORM_NUM_PLATFORMS,"Unknown platform");
	
	s_aiPlatformSampleRates[iPlatform] = iSampleRate;
}

int Tools_SoundEvent_System::GetDefaultPlatformCompressionQuality(int iPlatform)
{
	Asura_Assert(iPlatform<TOOLS_PLATFORM_NUM_PLATFORMS,"Unknown platform");
	
	return s_aiPlatformCompressionQuality[iPlatform];
}

void Tools_SoundEvent_System::SetDefaultPlatformCompressionQuality(int iPlatform , int iQuality)
{
	Asura_Assert(iPlatform<TOOLS_PLATFORM_NUM_PLATFORMS,"Unknown platform");
	Asura_Assert( iQuality >=0 && iQuality < TSS_QUALITY_NUM_ENTRIES, "Illegal quality enum");
	
	s_aiPlatformCompressionQuality[iPlatform] = iQuality;
}



int Tools_SoundEvent_System::GetPlatformCompressionQuality( const int iPlatform, const int iQuality )
{
	Asura_Assert( iPlatform < TOOLS_PLATFORM_NUM_PLATFORMS, "Unknown platform" );
	Asura_Assert( iQuality >= TSS_QUALITY_USE_DEFAULT && iQuality < TSS_QUALITY_NUM_ENTRIES, "Illegal quality enum");

	if(iQuality == TSS_QUALITY_USE_DEFAULT)
	{
		return s_aiPlatformCompressionRates[ iPlatform ][ GetDefaultPlatformCompressionQuality(iPlatform) ];
	}
	else
	{
		return s_aiPlatformCompressionRates[ iPlatform ][ iQuality ];
	}
}


void Tools_SoundEvent_System::SetPlatformCompressionQuality( const int iPlatform, const int iQuality, const int iValue )
{
	Asura_Assert( iPlatform < TOOLS_PLATFORM_NUM_PLATFORMS, "Unknown platform" );
	Asura_Assert( iQuality >= 0 && iQuality < TSS_QUALITY_NUM_ENTRIES, "Illegal quality enum");

	s_aiPlatformCompressionRates[ iPlatform ][ iQuality ] = iValue;
}


int Tools_SoundEvent_System::GetSampleRate(int iPlatform,int iSoundID)
{
	// AUDIOTODO: Do we need these functions?
	#ifdef ASURA_USE_AUDIO_SYSTEM
	return GetPlatformSampleRate(iPlatform);
	#else
	const char* szWavName = Asura_Sound_System::GetFileNameFromSoundID((Asura_Sound_ID)iSoundID);
	Asura_Assert(szWavName,"");

	return szWavName ? GetSampleRate(iPlatform,szWavName) : 0;
	#endif
}

int Tools_SoundEvent_System::GetSampleRate(int iPlatform,const char* szWavName)
{
	// AUDIOTODO: Do we need these functions?
	#ifdef ASURA_USE_AUDIO_SYSTEM
	return GetPlatformSampleRate(iPlatform);
	#else
	//make sure wav to sample tree is up to date
	if(!s_pxWavToSampleTree) BuildWavToSampleTree();

	//find the sample from the wav name
	Asura_Hash_ID* puSampleHash = s_pxWavToSampleTree->GetDataByName(szWavName);
	if(!puSampleHash) return 0;
	Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByHash(*puSampleHash);

	//Get the sample rate , returning the default value for the platform if it isn't set
	int iSampleRate = pxSample->GetSampleRate(iPlatform);
	return iSampleRate ? iSampleRate : GetPlatformSampleRate(iPlatform);
	#endif
}

void Tools_SoundEvent_System::BuildWavToSampleTree()
{
	Asura_Assert(!s_pxWavToSampleTree,"Already allocated");

	s_pxWavToSampleTree = new Asura_TemplateHashTree<Asura_Hash_ID>;

	//Add all the samples to the tree
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);

		#if defined ASURA_USE_AUDIO_SYSTEM

			if(pxSample->GetWaveHash() == ASURA_HASH_ID_UNSET) continue;
			s_pxWavToSampleTree->AddNode(pxSample->GetWaveHash(),pxSample->GetHashID());

		#else

			if(pxSample->m_eSoundID == ASURA_SID_NO_SOUND) continue;

			const char* szWavName = Asura_Sound_System::GetFileNameFromSoundID(pxSample->m_eSoundID);
			if(!szWavName) continue;

			s_pxWavToSampleTree->AddNode(Asura_GetHashID(szWavName),pxSample->GetHashID());

		#endif
	}
}


int Tools_SoundEvent_System::GetMaxSamplesPlayed(const Asura_Sound_Event* pxEvent)
{
	Asura_Assert(pxEvent,"");
	pxEvent = pxEvent->GetRemappedEvent();
	if(!pxEvent) return 0;

	int iMaxSamples = 0;
	bool bSimultaneous = (pxEvent->GetPlaybackType() == ASURA_PT_SIMULTANEOUS);

	//Add up or take the max of the samples used by each child event ,according to the simultaneous flag
	u_int uNumPlayable = pxEvent->GetNumPlayables();
	for(u_int i=0;i<uNumPlayable;i++)
	{
		const Asura_Sound_Playable* pxPlayable = pxEvent->GetPlayableSound(i);
		int iMaxForPlayable = 0;
		if(pxPlayable->IsEvent())
		{
			iMaxForPlayable = GetMaxSamplesPlayed((const Asura_Sound_Event* )pxPlayable);
		}
		else
		{
			//samples just play one sample
			iMaxForPlayable = 1;
		}

		if(bSimultaneous)
		{
			iMaxSamples += iMaxForPlayable;
		}
		else
		{
			iMaxSamples = Asura_Maths::Max(iMaxForPlayable,iMaxSamples);
		}
	}

	return iMaxSamples;
}


void Tools_SoundEvent_System::Verify(bool bAutomatic)
{
	if(bAutomatic && !Tools_MessageLog::ShouldShowForSound())
	{
		return;
	}
	
	//Clear out existing messages
	Tools_MessageLog::DeleteMessages(Tools_MessageLog::SOUND_EVENT,0);
	//Get current position to see if any errors were logged
	u_int uInitialPos = Tools_MessageLog::GetNextID();

	//check each event
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		Asura_Hash_ID uEventHash = pxEvent->GetHashID();
		
		//Check max simultaneous samples played
		int iMaxSamples = GetMaxSamplesPlayed(pxEvent);
		if(iMaxSamples > uASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY)
		{
			Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"",new Tools_MessageLogAction_SoundEvent(uEventHash),
										"'%s' may play up to %d simultaneous samples. Max allowed is %d",
										pxEvent->GetName(),iMaxSamples,uASURA_SOUND_EVENT_MAX_FOR_SIMULTANEOUS_PLAY);
		}

		//make sure variable events only contain samples in their playable list
		if(pxEvent->GetPlaybackType() == ASURA_PT_VARIABLE)
		{
			for(u_int i=0;i<pxEvent->GetNumPlayableID();i++)
			{
				if(pxEvent->GetPlayableID(i).m_eType != ASURA_SOUND_PLAYABLE_SAMPLE)
				{
					Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"",new Tools_MessageLogAction_SoundEvent(uEventHash),
												"Variable sound event '%s' contains some events in its play list. It should only have samples.",
												pxEvent->GetName());
					break;
				}
			}
		}
		//check for event mapped events that don't have exactly one event in their playable list
		if(pxEvent->IsEventMapping())
		{
			bool bMappingOkay = false;
			if(pxEvent->GetNumPlayableID() == 1)
			{
				if(pxEvent->GetPlayableID(0).m_eType == ASURA_SOUND_PLAYABLE_EVENT)
				{
					bMappingOkay = true;
				}
			}
			if(!bMappingOkay)
			{
				Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"",new Tools_MessageLogAction_SoundEvent(uEventHash),
											"Sound event '%s' has the Event Mapping flag, but doesn't have exactly one event in the playable list.",
											pxEvent->GetName());
			}
		}

		//check for completely empty events
		if(pxEvent->GetNumPlayableID() == 0)
		{
			Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"",new Tools_MessageLogAction_SoundEvent(uEventHash),
										"Sound event '%s' has an empty playable list.",
										pxEvent->GetName());
		
		}
		
	}

	//check samples
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);
		Asura_Hash_ID uSampleHash = pxSample->GetHashID();
	
		//make sure the wav file can be found
		//(don't do for the automatic check , since it is fairly slow)
		if(!bAutomatic)
		{
			#if defined ASURA_USE_AUDIO_SYSTEM
				Asura_Hash_ID uWave = pxSample->GetWaveHash();
				if(uWave == ASURA_HASH_ID_UNSET) continue;
				const char* szFilename = Asura_Audio_System::GetWaveFilename(uWave);
			#else
				Asura_Sound_ID eSoundID = pxSample->GetSoundID();
				if(eSoundID == ASURA_SID_NO_SOUND) continue;
				const char* szFilename = Asura_Sound_System::GetFileNameFromSoundID(eSoundID);
			#endif

			Asura_Assert(szFilename,"Error getting wav filename for sound sample");
			if(szFilename)
			{
				if(!Tools_FileSystem::FileExists_CheckNetwork(szFilename))
				{
					Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"",new Tools_MessageLogAction_SoundSample(uSampleHash),
												"Wav file %s for sample '%s' doesn't exists.",
												szFilename,pxSample->GetName());
					
				}
			}
		}

	}

	if(Tools_MessageLog::AreThereAnyMessages(Tools_MessageLog::SOUND_EVENT, 0, uInitialPos))
	{
		//Found some errors ,so show message log
		Tools_MessageLog::ShowOverlay(Tools_MessageLog::SOUND_EVENT,0,uInitialPos);
	}
	else if(!bAutomatic)
	{
		ShowMessageOverlay("No sound event system errors found.");
	}
}


bool Tools_SoundEvent_System::DeleteEvent(Asura_Hash_ID uHash,bool bConfirm)
{
	Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByHash(uHash);
	if(!pxEvent) return false;

	Asura_Sound_PlayableID xID;
	xID.m_uHashID = uHash;
	xID.m_eType = ASURA_SOUND_PLAYABLE_EVENT;

	if(bConfirm)
	{
		//See how many events use this
		int iEventCount = 0;
		for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
		{
			Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);

			u_int uNumPlayable = pxEvent->GetNumPlayableID();
			for(u_int j=0;j<uNumPlayable;j++)
			{
				if(pxEvent->GetPlayableID(j) == xID)
				{
					iEventCount++;
					break;
				}
			}
		}

		//Confirm deletion
		char szEventCount[100];
		Asura_PrintfToBfr(szEventCount,"It is used by %d event(s).",sizeof(szEventCount),true,iEventCount);

		if(!ShowYesNoOverlay("Delete event '%s'?\n%s",pxEvent->GetName(),iEventCount ? szEventCount : ""))
		{
			return false;
		}
	}

	//Delete the tab for this event (if any)
	if(!Tools_SoundEvent_Overlay::CloseEventTab(uHash))
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to delete %s",pxEvent->GetName());
		return false;
	}


	//Go through the events again removing any references to this event
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		if(pxEvent->RemovePlayableID(xID))
		{
			Tools_SoundEvent_System::SetModified(pxEvent);

			//Update the tab for this event if it was being edited
			Tools_SoundEvent_Overlay::UpdateEventFromAsuraData(pxEvent->GetHashID());
		}
	}

	//Remove references to this from any phonons
	Tools_Phonon_Object::SoundEventDeleted(pxEvent);
	Asura_Phonon_System::SoundEventDeleted(pxEvent);

	//And from any currently playing sounds
	//(now dealt with in the destructor)
	//Asura_Sound_Event_System::SoundEventDeleted(pxEvent);

	Tools_SoundEvent_System::SetModified(pxEvent);
	
	Asura_Sound_Event_Storage::DeleteEvent(pxEvent);
	return true;

}

bool Tools_SoundEvent_System::DeleteSample(Asura_Hash_ID uHash,bool bConfirm)
{
	Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByHash(uHash);
	if(!pxSample) return false;

	Asura_Sound_PlayableID xID;
	xID.m_uHashID = uHash;
	xID.m_eType = ASURA_SOUND_PLAYABLE_SAMPLE;

	if(bConfirm)
	{
		//See how many events use this
		int iEventCount = 0;
		for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
		{
			Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);

			u_int uNumPlayable = pxEvent->GetNumPlayableID();
			for(u_int j=0;j<uNumPlayable;j++)
			{
				if(pxEvent->GetPlayableID(j) == xID)
				{
					iEventCount++;
					break;
				}
			}
		}

		//Confirm deletion
		char szEventCount[100];
		Asura_PrintfToBfr(szEventCount,"It is used by %d event(s).",sizeof(szEventCount),true,iEventCount);

		if(!ShowYesNoOverlay("Delete sample '%s'?\n%s",pxSample->GetName(),iEventCount ? szEventCount : ""))
		{
			return false;
		}
	}



	//Delete the tab for this sample (if any)
	if(!Tools_SoundEvent_Overlay::CloseSampleTab(uHash))
	{
		Tools_Timed_Message_System::AddMessage(0xff0000,"Failed to delete %s",pxSample->GetName());
		return false;
	}

	//Go through the events removing any references to the sample

	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		if(pxEvent->RemovePlayableID(xID))
		{
			Tools_SoundEvent_System::SetModified(pxEvent);

			//Update the tab for this event if it was being edited
			Tools_SoundEvent_Overlay::UpdateEventFromAsuraData(pxEvent->GetHashID());
		}
	}

	Tools_SoundEvent_System::SetModified(pxSample);
	Asura_Sound_Event_Storage::DeleteSample(pxSample);

    return true;	
}


void Tools_SoundEvent_System::FindWavFiles()
{
	//Clear out existing messages
	Tools_MessageLog::DeleteMessages(Tools_MessageLog::SOUND_EVENT,0);
	//Get current position to see if any errors were logged
	u_int uInitialPos = Tools_MessageLog::GetNextID();

	Tools_WavFinder xWavFinder;
	
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);
		Asura_Hash_ID uSampleHash = pxSample->GetHashID();
	
		
		//see if the wav for this sample is still where it was
		#if defined ASURA_USE_AUDIO_SYSTEM
			Asura_Hash_ID uWave = pxSample->GetWaveHash();
			if(uWave == ASURA_HASH_ID_UNSET) continue;
			const char* szFilename = Asura_Audio_System::GetWaveFilename(uWave);
		#else
			Asura_Sound_ID eSoundID = pxSample->GetSoundID();
			if(eSoundID == ASURA_SID_NO_SOUND) continue;
			const char* szFilename = Asura_Sound_System::GetFileNameFromSoundID(eSoundID);
		#endif

		Asura_Assert(szFilename,"Error getting wav filename for sound sample");
		if(!szFilename) continue;
		if(Tools_FileSystem::FileExists_CheckNetwork(szFilename)) continue;

		//Search for the file elsewhere then
		const char* szNewFilename = xWavFinder.FindFile(szFilename);
		if(szNewFilename)
		{
			Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"",new Tools_MessageLogAction_SoundSample(uSampleHash),
										"Wav file %s for sample '%s' moved from %s to %s.",
										szFilename,pxSample->GetName(),szFilename,szNewFilename);

			#if defined ASURA_USE_AUDIO_SYSTEM
				pxSample->m_uWave = Asura_GetHashID(szNewFilename);
				Asura_Audio_System::LoadWave(szNewFilename,pxSample->GetLooping());
			#else
				pxSample->m_eSoundID = Tools_PickSound::GetSoundIDForSoundEventFilename(szNewFilename);
				Asura_Assert(pxSample->m_eSoundID != ASURA_SID_NO_SOUND,"Failed to get sound ID for sound");
			#endif
		
			SetModified(pxSample);

			Tools_SoundEvent_Overlay::UpdateSampleFromAsuraData(pxSample->m_uHashID);

			
		
		}
		else
		{
				Tools_MessageLog::AddMessage(Tools_MessageLog::SOUND_EVENT,"Error",new Tools_MessageLogAction_SoundSample(uSampleHash),
											"Wav file %s for sample '%s' couldn't be found in anywhere beneath the sounds directory.",
											szFilename,pxSample->GetName());
		}
	}

	if(Tools_MessageLog::AreThereAnyMessages(Tools_MessageLog::SOUND_EVENT, 0, uInitialPos))
	{
		//Found some errors ,so show message log
		Tools_MessageLog::ShowOverlay(Tools_MessageLog::SOUND_EVENT,0,uInitialPos);
	}
	else
	{
		ShowMessageOverlay("No samples were missing wav files.");
	}

}

void Tools_SoundEvent_System::ReloadWavsForSample(Asura_Hash_ID uSampleHash,bool bForceReloadNow)
{
	const Asura_Sound_Sample* pxSample =  Asura_Sound_Event_Storage::GetSampleByHash(uSampleHash);
	if(!pxSample) return;

	#if defined ASURA_USE_AUDIO_SYSTEM
		const Asura_Hash_ID uWave = pxSample->GetWaveHash();
		if(uWave != ASURA_HASH_ID_UNSET)
		{
			Asura_Audio_System::ReloadWave(uWave, bForceReloadNow);
		}
	#else
		Asura_Sound_ID eSoundID = pxSample->GetSoundID();
		if(eSoundID != ASURA_SID_NO_SOUND)
		{
			Asura_Sound_System::ReloadSound(eSoundID);
			if(bForceReloadNow) Asura_Sound_System::ForceSoundToBeLoaded(eSoundID);
		}
	#endif
}

void Tools_SoundEvent_System::ReloadWavsForEvent(Asura_Hash_ID uEventHash,bool bForceReloadNow)
{
	const Asura_Sound_Event* pxEvent =  Asura_Sound_Event_Storage::GetEventByHash(uEventHash);
	if(!pxEvent) return;

	List<const Asura_Sound_Event*> xEventList;
	xEventList.add_entry(pxEvent);

	for(LIF<const Asura_Sound_Event*> xEventLIF(&xEventList); !xEventLIF.done(); xEventLIF.next())
	{
		for(u_int i=0;i<pxEvent->GetNumPlayables();i++)
		{
			const Asura_Sound_Playable* pxPlayable = pxEvent->GetPlayableSound(i);
			if(pxPlayable->IsEvent())
			{
				//add event to list to be checked
				const Asura_Sound_Event* pxChildEvent = (const Asura_Sound_Event*) pxPlayable;
				if(!xEventList.contains(pxChildEvent))
				{
					xEventList.add_entry(pxChildEvent);
				}
			}
			if(pxPlayable->IsSample())
			{
				const Asura_Sound_Sample* pxSample = (const Asura_Sound_Sample*) pxPlayable;
				ReloadWavsForSample(pxSample->GetHashID(), bForceReloadNow);
			}
		}
	}
}

void Tools_SoundEvent_System::FindReferencesToWav()
{
	#if defined ASURA_USE_AUDIO_SYSTEM
		Asura_Hash_ID uWave = Tools_PickSound::PickSoundForSoundEventSample(ASURA_HASH_ID_UNSET, NULL, "Find samples/events using which wav?");
		if(uWave != ASURA_HASH_ID_UNSET) FindReferencesToWav(uWave);
	#else
		Asura_Sound_ID eSoundID = Tools_PickSound::PickSoundForSoundEventSample(ASURA_SID_NO_SOUND, NULL, "Find samples/events using which wav?");
		if(eSoundID != ASURA_SID_NO_SOUND) FindReferencesToWav(eSoundID);
	#endif
}

void Tools_SoundEvent_System::FindReferencesToEvent( const Asura_Sound_Event* pxEvent )
{
	// finds references to sound events and outputs them to the sound event log
	// currently only searching through the blueprints but may be extended to entities, pfx, etc.
	if( !pxEvent ) return;
	const Asura_Hash_ID uEventHash = pxEvent->GetHashID();

	Tools_MessageLog::Category eCat = Tools_MessageLog::SOUND_EVENT;
	const char* szSubCat = "Find";


	//search for phonons using the event
	for(int i=0; i<Asura_Phonon_System::GetNumberOfPhonons(); i++)
	{
		Asura_Phonon* pxPhonon = Asura_Phonon_System::GetPhononByIndex(i);
		if(pxEvent == pxPhonon->GetEvent())
		{
			const char* szName = pxPhonon->GetName()->GetString();
			char acNameBuffer[100];
			if(!szName || !*szName)
			{
				int iID = pxPhonon->GetGuid()-ASURA_GUID_TOOLSCREATEDENTITY_START;
				Asura_PrintfToBfr(acNameBuffer, "%s (%d)", sizeof(acNameBuffer), true, pxEvent->GetName(), iID);
				szName = acNameBuffer;
			}
		
			
			Tools_MessageLog::AddMessage( eCat, szSubCat, new Tools_MessageLogAction_Phonon(pxPhonon),
				"%s is used by Phonon '%s'.",
				pxEvent->GetName(),szName);
		} 
	}


	const Tools_Blueprint_Edit_Object* pxTempEditObj = Tools_Mode_Blueprint::CreateBlueprintEditObject_Static();

	// run through the blueprints searching for parameters with the sound event template
	for ( Asura_Blueprint_Parameter_ConstIterator xBPParamIter; !xBPParamIter.Done(); xBPParamIter.Next() )
	{
		const Asura_Parameter* pxParam = xBPParamIter.GetCurrent();
		// does the parameter use the TVT_SOUND_EVENT template?
		// if yes, check the hash ID of the parameter against the event we're searching for
		const int iTVType = pxTempEditObj->GetTemplatedTypeFromTemplateID( pxParam->GetTemplate() );

		if( (iTVType == Tools_Blueprint_Edit_Object::TVT_SOUNDEVENT)
			&& (pxParam->GetValue().GetHashIDValue() == pxEvent->GetHashID()) )
		{
			// add output
			const Asura_Hash_ID uTypeHashID = xBPParamIter.GetCurrentBlueprintTypeHash();
			const Asura_Blueprint* pxBP = xBPParamIter.GetCurrentBlueprint();

			Tools_MessageLog::AddMessage( eCat, szSubCat, new Tools_MessageLogAction_Blueprint(uTypeHashID,pxBP->GetHashID(),pxParam->GetHashID()),
				"%s is referenced in blueprint '%s' by parameter '%s'.",
				pxEvent->GetName(),pxBP->GetName().GetString(),pxParam->GetName().GetString());
		}
	}
	delete pxTempEditObj;

	//check for PFX templates using the sound event
	{
		//effect templates
		{
			const Asura_TemplateHashTree<Asura_String>&	xTemplateNameTree = Asura_PFX_System::GetEffectTempNames();
			for(int i=0; i< xTemplateNameTree.GetNumberOfNodes(); i++)
			{
				const Asura_String*	pxFullTemplateName	= xTemplateNameTree.GetDataByIndex(i);

				const Asura_PFX_Effect_Template* pxTemplate = Asura_PFX_System::GetEffectTempByHash(pxFullTemplateName->GetHash());
				if(pxTemplate && pxTemplate->m_uSoundEvent == pxEvent->GetHashID())
				{
					Tools_MessageLog::AddMessage( eCat, szSubCat, NULL,
						"%s is used by PFX effect template '%s'.",
						pxEvent->GetName(),pxFullTemplateName->GetString());
				}
			}
		}
		//particle templates
		{
			const Asura_TemplateHashTree<Asura_String>&	xTemplateNameTree = Asura_PFX_System::GetParticleTempNames();
			for(int i=0; i< xTemplateNameTree.GetNumberOfNodes(); i++)
			{
				const Asura_String*	pxFullTemplateName	= xTemplateNameTree.GetDataByIndex(i);

				const Asura_PFX_Particle_Template* pxTemplate = Asura_PFX_System::GetParticleTempByHash(pxFullTemplateName->GetHash());
				if(pxTemplate && pxTemplate->m_uSoundEvent == pxEvent->GetHashID())
				{
					Tools_MessageLog::AddMessage( eCat, szSubCat, NULL,
						"%s is used by PFX particle template '%s'.",
						pxEvent->GetName(),pxFullTemplateName->GetString());
				}
			}
		}

	}

	//check loaded skins
	for(int i=0; i<Asura_Animation_System::GetNumberOfSkins(); i++)
	{
		const Asura_Hierarchy_Skin* pxSkin = Asura_Animation_System::GetSkin(i);

		for(int j=0; j<pxSkin->m_iNumberOfSoundDefns; j++)
		{
			const Asura_Hierarchy_Sound* pxSound = &pxSkin->m_pxSoundDefns[j];
			if(pxSound->m_uSoundEvent == uEventHash)
			{
				Tools_MessageLog::AddMessage( eCat, szSubCat, new Tools_MessageLogAction_SelectSkin(pxSkin),
					"%s is used by skin '%s'.",
					pxEvent->GetName(),pxSkin->GetName());
			}
		}
	}

	//check loaded anims
	for(int i=0; i<Asura_Animation_System::GetNumberOfAnims(); i++)
	{
		const Asura_Hierarchy_Anim* pxAnim = Asura_Animation_System::GetAnim(i);

		for(int j=0; j<pxAnim->m_iNumberOfSoundDefns; j++)
		{
			const Asura_Hierarchy_Sound* pxSound = &pxAnim->m_pxSoundDefns[j];
			if(pxSound->m_uSoundEvent == uEventHash)
			{
				Tools_MessageLog::AddMessage( eCat, szSubCat, new Tools_MessageLogAction_SelectAnim(pxAnim),
					"%s is used by anim '%s'.",
					pxEvent->GetName(),pxAnim->GetName());
			}
		}
	}

	//check cutscenes
	for(int i=0; i< Asura_Cutscene_System::GetNumberOfCutscenes(); i++)
	{
		Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByIndex(i);
		for(int j=0; j<pxCutscene->NumberOfActors; j++)
		{
			Asura_Cutscene_Actor* pxActor = pxCutscene->Actors[j];
			if(!pxActor->IsHierarchyActor()) continue;
			
			Asura_Cutscene_Actor_Hierarchy* pxHierActor = (Asura_Cutscene_Actor_Hierarchy*) pxActor;

			for(int k=0; k<pxHierActor->m_iNumberOfAttachments; k++)
			{
				Asura_Cutscene_Attachment* pxAttachment = pxHierActor->m_ppxAttachments[k];
				if(pxAttachment->GetType() != AsuraCutsceneAttachment_SoundEvent) continue;

				Asura_Cutscene_Attachment_SoundEvent* pxSoundAttachment = (Asura_Cutscene_Attachment_SoundEvent*) pxAttachment;
				if(pxSoundAttachment->GetSoundEventHash() == uEventHash)
				{
					Tools_MessageLog::AddMessage( eCat, szSubCat, NULL,
						"%s is used by an attachment in cutscene %s.",
						pxEvent->GetName(),pxCutscene->GetName());
					
				}
			}
		}
	}

}

#if defined ASURA_USE_AUDIO_SYSTEM

void Tools_SoundEvent_System::FindReferencesToWav(Asura_Hash_ID uWave)
{
	if(uWave == ASURA_HASH_ID_UNSET) return;
	const char* szFilename = Asura_Audio_System::GetWaveFilename(uWave);

#else

void Tools_SoundEvent_System::FindReferencesToWav(int iSoundID)
{
	if(iSoundID == ASURA_SID_NO_SOUND) return;
	const char* szFilename = Asura_Sound_System::GetFileNameFromSoundID((Asura_Sound_ID)iSoundID);

#endif

	Asura_Assert(szFilename,"Error getting sound filename");

	szFilename = Tools_FileSystem::StripPath(szFilename);


	//Get current position to see if any messages get logged
	u_int uInitialPos = Tools_MessageLog::GetNextID();

	Tools_MessageLog::Category eCat = Tools_MessageLog::SOUND_EVENT;
	const char* szSubCat = "Find";

	//get a list of all samples that make use of the wav file
	List<const Asura_Sound_Sample*> xSampleList;

	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);

		#if defined ASURA_USE_AUDIO_SYSTEM
			const bool bFound = (pxSample->GetWaveHash() == uWave);
		#else
			const bool bFound = (pxSample->GetSoundID() == iSoundID);
		#endif

		if(bFound)
		{
			xSampleList.add_entry(pxSample);

			Tools_MessageLog::AddMessage(eCat, szSubCat,new Tools_MessageLogAction_SoundSample(pxSample->GetHashID()),
										"%s is used by sample %s.",
										szFilename,pxSample->GetName());

		}
	}

	//see which events use those samples directly
	List<const Asura_Sound_Event*> xEventList;

	for(LIF<const Asura_Sound_Sample*> xSampleLIF(&xSampleList); !xSampleLIF.done() ; xSampleLIF.next())
	{
		const Asura_Sound_Sample* pxSample = xSampleLIF();
		for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
		{
			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);

			for(u_int j=0;j<pxEvent->GetNumPlayableID();j++)
			{
				if(pxEvent->GetPlayableID(j).m_uHashID == pxSample->GetHashID() && 
				   pxEvent->GetPlayableID(j).m_eType == ASURA_SOUND_PLAYABLE_SAMPLE)
				{
					xEventList.add_entry(pxEvent);
					Tools_MessageLog::AddMessage(eCat, szSubCat,new Tools_MessageLogAction_SoundEvent(pxEvent->GetHashID()),
												"%s is used by event %s.",
												pxSample->GetName(), pxEvent->GetName());
					FindReferencesToEvent( pxEvent );
					break;
				}
			}
		}
	}

	//check for events using those events
	for(LIF<const Asura_Sound_Event*> xEventLIF(&xEventList); !xEventLIF.done(); xEventLIF.next())
	{
		const Asura_Sound_Event* pxEvent1 = xEventLIF();
		for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
		{
			const Asura_Sound_Event* pxEvent2 = Asura_Sound_Event_Storage::GetEventByIndex(i);
			if(xEventList.contains(pxEvent2)) continue;

			for(u_int j=0;j<pxEvent2->GetNumPlayableID();j++)
			{
				if(pxEvent2->GetPlayableID(j).m_uHashID == pxEvent1->GetHashID() && 
				   pxEvent2->GetPlayableID(j).m_eType == ASURA_SOUND_PLAYABLE_EVENT)
				{					
					Tools_MessageLog::AddMessage(eCat, szSubCat,new Tools_MessageLogAction_SoundEvent(pxEvent2->GetHashID()),
												"%s is used by event %s.",
												pxEvent1->GetName(), pxEvent2->GetName());
					FindReferencesToEvent( pxEvent2 );
					break;
				}
			}
		}
	}

	if(Tools_MessageLog::AreThereAnyMessages(eCat, szSubCat, uInitialPos))
	{
		//show the log of found samples/events
		Tools_MessageLog::ShowOverlay(eCat,szSubCat,uInitialPos);
	}
	else
	{
		//nothing found
		ShowMessageOverlay("Failed to find any samples using this wav file");
	}
}

void Tools_SoundEvent_System::AddSoundEventsAndSamplesToTree(const Asura_Sound_Event* pxEvent, Asura_HashTree& xEventTree, Asura_HashTree& xSampleTree)
{
	int iJunkIndex;
	bool bAlreadyPresent = xEventTree.FindNodeIndexByHash(pxEvent->GetHashID(), true, iJunkIndex);

	//If the event wasn't previously added ,then recurse down and add any events/samples used by it
	if(!bAlreadyPresent)
	{
		for(u_int i=0;i<pxEvent->GetNumPlayables();i++)
		{
			const Asura_Sound_Playable* pxPlayable = pxEvent->GetPlayableSound(i);
			if(pxPlayable->IsSample())
			{
				const Asura_Sound_Sample* pxChildSample = (const Asura_Sound_Sample*) pxPlayable;
				xSampleTree.FindNodeIndexByHash(pxChildSample->GetHashID(), true, iJunkIndex);
			}
			else
			{
				const Asura_Sound_Event* pxChildEvent = (const Asura_Sound_Event*) pxPlayable;
				AddSoundEventsAndSamplesToTree(pxChildEvent, xEventTree, xSampleTree);
			}
		}
	}
}


void Tools_SoundEvent_System::FindUnusedEventsAndSamples()
{
	Tools_MessageLog::Category eCat = Tools_MessageLog::SOUND_EVENT;
	const char* szSubCat = "Unused";

	//get the directory to scan
	Tools_FileSystem_SelectFolderInfo xSelectFolder;
	xSelectFolder.InitialSelection = "";
	xSelectFolder.DisplayText = "Scan the exports in which directory?";

	if(!Tools_FileSystem::SelectFolder(xSelectFolder)) return;

	//get a list of all the files.
	char szSearch[MAX_PATH];
	Asura_PrintfToBfr(szSearch,"%s\\*.*",sizeof(szSearch),true,xSelectFolder.SelectedFolder);
	List<Asura_String> xFileList;
	Tools_FileSystem::FindFiles(szSearch,xFileList,TOOLS_FILEFIND_RECURSIVE|TOOLS_FILEFIND_IGNOREBACKUPDIR);

	if(!xFileList.size())
	{
		ShowMessageOverlay("Failed to find any files");
		return;
	}

	//trees containing all events/samples that are in use
	Asura_HashTree xEventTree;
	Asura_HashTree xSampleTree;

	//start by adding all global events to the tree
	if(Asura_Sound_Event_Storage::GetNumberOfEvents())
	{
		for(Asura_TemplateHashTree_SaveIterator xIt(Asura_Sound_Event_Storage::GetNumberOfEvents());!xIt.Done();xIt.Next())
		{
			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(xIt.GetCurrent());
			if(pxEvent->IsGlobalSound())
			{
				AddSoundEventsAndSamplesToTree(pxEvent, xEventTree, xSampleTree);
			}
		}
	}

	Tools_ProgressIndicator xProgress("Scanning files");
	Tools_MessageLog::DeleteMessages(eCat, szSubCat);
	
	//load each of the files found and check for events/samples used by them
	Tools_MessageLog::AddMessage(eCat, szSubCat, "Files searched:");
	int iFileIndex = 0;
	for(LIF<Asura_String> xFileLIF(&xFileList);!xFileLIF.done();xFileLIF.next(),iFileIndex++)
	{
		const char* szExportedFile = xFileLIF().GetString();

		xProgress.Update(iFileIndex, xFileList.size());

#if 0
		//ignore files that don't have one of the platform extensions
		const char* szExtension = strrchr(szExportedFile,'.');
		bool bSkipFile = true;
		if(szExtension)
		{
			for(int iPlatform=0;iPlatform<TOOLS_PLATFORM_NUM_PLATFORMS;iPlatform++)
			{
				if(!stricmp(szExtension,Tools_Platform::GetPlatformExtension(iPlatform)))
				{
					bSkipFile = false;
				}
			}
		}
		if(bSkipFile) continue;
#endif

		if(Asura_UserInput::KeyPressed(ASURA_KEYID_ESCAPE)) return;

		Tools_AsuraFileChunkLister xChunkLister;
		if(!xChunkLister.LoadAsuraFile(szExportedFile,false,false))
		{
			//TBD , complain , probably
			continue;
		}
		bool bBigEndian = xChunkLister.IsBigEndian();

		// Output this file to the message-log...
		Tools_MessageLog::AddMessage(eCat, szSubCat, szExportedFile);

		//scan for sound event chunks
		List<const Asura_Chunk_Header*> xChunkList;
		xChunkLister.ListChunksByID(bBigEndian ? Asura_Endian::SwapBytes(ASURA_CHUNK_SOUNDEVENT_EVENT) : ASURA_CHUNK_SOUNDEVENT_EVENT,xChunkList);

		for(LIF<const Asura_Chunk_Header*> xChunkLIF(&xChunkList);!xChunkLIF.done(); xChunkLIF.next())
		{
			//the name should be just beyond the header
			const char* szEventID = (const char*)(xChunkLIF()+1);

			const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByHash(Asura_GetHashID(szEventID));
			if(pxEvent)
			{
				//add to the list of events
				AddSoundEventsAndSamplesToTree(pxEvent, xEventTree, xSampleTree);
			}
		}

		//and for sound sample chunks
		xChunkList.empty();
		xChunkLister.ListChunksByID(bBigEndian ? Asura_Endian::SwapBytes(ASURA_CHUNK_SOUNDEVENT_SAMPLE) : ASURA_CHUNK_SOUNDEVENT_SAMPLE,xChunkList);

		for(LIF<const Asura_Chunk_Header*> xChunkLIF(&xChunkList);!xChunkLIF.done(); xChunkLIF.next())
		{
			//the name should be just beyond the header
			const char* szSampleID = (const char*)(xChunkLIF()+1);

			//add to the list of samples
			int iJunkIndex;
			xSampleTree.FindNodeIndexByHash(Asura_GetHashID(szSampleID), true, iJunkIndex);
		}
	}
	xProgress.Finish();

	s_xUnusedEventList.empty();
	s_xUnusedSampleList.empty();


	//get the names of all events and samples that haven't been found as used
	List<const char*> xEventNameList;
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		
		int iJunkIndex;
		if(!xEventTree.FindNodeIndexByHash(pxEvent->GetHashID(), false, iJunkIndex))
		{
			xEventNameList.add_entry(pxEvent->GetName());
			s_xUnusedEventList.add_entry(pxEvent->GetHashID());
		}
	}
	
	
	List<const char*> xSampleNameList;
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);
		
		int iJunkIndex;
		if(!xSampleTree.FindNodeIndexByHash(pxSample->GetHashID(), false, iJunkIndex))
		{
			xSampleNameList.add_entry(pxSample->GetName());
			s_xUnusedSampleList.add_entry(pxSample->GetHashID());
		}
	}

	if(!xEventNameList.size() && !xSampleNameList.size())
	{
		ShowMessageOverlay("Failed to find any unused events and samples.");
		return;
	}

	//output to the message log
	if(xEventNameList.size())
	{
		Tools_MessageLog::AddMessage(eCat, szSubCat, "No references could be found to %d sound events :", xEventNameList.size());

		//get a sorted list of names
		const char** pszNameList = new const char*[xEventNameList.size()];
		int iIndex = 0;
		for(LIF<const char*> xNameLIF(&xEventNameList); !xNameLIF.done(); xNameLIF.next(),iIndex++)
		{
			pszNameList[iIndex] = xNameLIF();
		}

		Asura_SortAndSearch::Quicksort( pszNameList, xEventNameList.size(), Asura_SortAndSearch::SortStringsCaseInsensitive);

		for(int i=0 ; i<xEventNameList.size(); i++)
		{
			Asura_Hash_ID uHash = Asura_GetHashID(pszNameList[i]);
			Tools_MessageLog::AddMessage(eCat, szSubCat, new Tools_MessageLogAction_SoundEvent(uHash), "%s", pszNameList[i]);
		}

		delete [] pszNameList;
	}

	if(xSampleNameList.size())
	{
		Tools_MessageLog::AddMessage(eCat, szSubCat, "No references could be found to %d sound samples :", xSampleNameList.size());

		//get a sorted list of names
		const char** pszNameList = new const char*[xSampleNameList.size()];
		int iIndex = 0;
		for(LIF<const char*> xNameLIF(&xSampleNameList); !xNameLIF.done(); xNameLIF.next(),iIndex++)
		{
			pszNameList[iIndex] = xNameLIF();
		}

		Asura_SortAndSearch::Quicksort( pszNameList, xSampleNameList.size(), Asura_SortAndSearch::SortStringsCaseInsensitive);

		for(int i=0 ; i<xSampleNameList.size(); i++)
		{
			Asura_Hash_ID uHash = Asura_GetHashID(pszNameList[i]);
			Tools_MessageLog::AddMessage(eCat, szSubCat, new Tools_MessageLogAction_SoundSample(uHash), "%s", pszNameList[i]);
		}

		//list the wav files used by those samples
		Tools_MessageLog::AddMessage(eCat, szSubCat, "Wav files used :");
		for(int i=0 ; i<xSampleNameList.size(); i++)
		{
			Asura_Hash_ID uHash = Asura_GetHashID(pszNameList[i]);
			Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByHash(uHash);
			if(pxSample)
			{
				#if defined ASURA_USE_AUDIO_SYSTEM
					Asura_Hash_ID uWave = pxSample->GetWaveHash();
					if(uWave == ASURA_HASH_ID_UNSET) continue;
					const char* szWav = Asura_Audio_System::GetWaveFilename(uWave);
				#else
					Asura_Sound_ID eSoundID = pxSample->GetSoundID();
					if(eSoundID == ASURA_SID_NO_SOUND) continue;
					const char* szWav = Asura_Sound_System::GetFileNameFromSoundID(eSoundID);
				#endif

				if(szWav)
				{
					Tools_MessageLog::AddMessage(eCat, szSubCat, 0, "%s", szWav);
				}
			}
		}


		delete [] pszNameList;
	}

	Tools_MessageLog::ShowOverlay(eCat, szSubCat);
}

void Tools_SoundEvent_System::DeleteUnusedEventsAndSamples()
{
	if(!s_xUnusedEventList.size() && !s_xUnusedSampleList.size()) return;
	
	bool bOkay = ShowYesNoOverlay("Delete the %d unused events and %d unused samples?",s_xUnusedEventList.size(),s_xUnusedSampleList.size());	
	if(!bOkay) return;

	while(s_xUnusedEventList.size())
	{
		DeleteEvent(s_xUnusedEventList.first_entry(),false);
		s_xUnusedEventList.delete_first_entry();
	}		

	while(s_xUnusedSampleList.size())
	{
		DeleteSample(s_xUnusedSampleList.first_entry(),false);
		s_xUnusedSampleList.delete_first_entry();
	}		
}

bool Tools_SoundEvent_System::IsEnabled_DeleteUnusedEventsAndSamples()
{
	return s_xUnusedEventList.size() || s_xUnusedSampleList.size();
}

void Tools_SoundEvent_System::DeleteUnusedSamples()
{
	Asura_HashTree xUsedSampleTree;

	//scan all the events and find the samples they use
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		const Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);

		for(u_int j=0;j<pxEvent->GetNumPlayableID();j++)
		{
			if(pxEvent->GetPlayableID(j).m_eType == ASURA_SOUND_PLAYABLE_SAMPLE)
			{
				int iJunkIndex;
				xUsedSampleTree.FindNodeIndexByHash(pxEvent->GetPlayableID(j).m_uHashID, true, iJunkIndex);
			}
		}
	}

	//Now go through the samples and find any that haven't been added to the tree
	List<Asura_Hash_ID> xUnusedSampleList;

	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfSamples();i++)
	{
		const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::GetSampleByIndex(i);

		Asura_Hash_ID uSampleHash = pxSample->GetHashID();
		
		int iJunkIndex;
		if(!xUsedSampleTree.FindNodeIndexByHash(uSampleHash, false, iJunkIndex))		
		{
			xUnusedSampleList.add_entry(uSampleHash);
		}
	}

	if(!xUnusedSampleList.size())
	{
		ShowMessageOverlay("No unused samples were found");
		return;
	}

	bool bDelete = ShowYesNoOverlay("Found %d samples that aren't used by any sound events.\nWould you like to delete them?", xUnusedSampleList.size());
	if(!bDelete) return;

	for(LIF<Asura_Hash_ID> xSampleLIF(&xUnusedSampleList); !xSampleLIF.done(); xSampleLIF.next())
	{
		DeleteSample(xSampleLIF(), false);
	}
}

void Tools_SoundEvent_System::LoadSoundEventsForPlatform(int iPlatform)
{
	int iPlatformDir = -1;
	char szSearch[200];

	if(iPlatform >= 0)
	{
		//first search the platform directory for sound event files , to determine whether
		//to use the platform directory or the root directory
		Asura_PrintfToBfr(szSearch, "Sounds\\SoundEvents\\%s\\*.SoundEvent", sizeof(szSearch), true, Tools_Platform::GetPlatformName(iPlatform));

		List<Asura_String> xFileList;
		Tools_FileSystem::FindFilesLocalAndNetwork(szSearch, xFileList, 0);
		if(xFileList.size())
		{
			iPlatformDir = iPlatform;
		}
	}

	//has this resulted in a change in directory?
	if(s_iCurrentPlatformDirectory == iPlatformDir) return;

	//see if there are any modified sound events to save , before those changes get lost
	if(s_xModifiedFiles.size() > 0)
	{
		//ask whether to save
		if(ShowYesNoOverlay("%d sound event files are marked as modified.\nIf these aren't saved now ,the changes will be lost because of loading files for a different platform.\nWould you like to save?",s_xModifiedFiles.size()))
		{
			SaveModified();
		}

		//clear out the list of modified files regardless of success
		while(s_xModifiedFiles.size()) s_xModifiedFiles.delete_first_entry();
	}

	//clear out the old sound events and samples
	Asura_Sound_Event_Storage::Shutdown();
//	Asura_Sound_Event_Storage::Initialise();

	//load the sounds from the new directory
	if(iPlatformDir >= 0)
	{
		Asura_PrintfToBfr(szSearch, "Sounds\\SoundEvents\\%s\\*.SoundEvent", sizeof(szSearch), true, Tools_Platform::GetPlatformName(iPlatformDir));
	}
	else
	{
		Asura_PrintfToBfr(szSearch, "Sounds\\SoundEvents\\*.SoundEvent", sizeof(szSearch), true);
	}

	s_iCurrentPlatformDirectory = iPlatform;

	Tools_FileSystem::LoadAsuraFiles(szSearch);

	//Now need to go through all the events and sort out the pointers to the playables
	for(u_int i=0;i<Asura_Sound_Event_Storage::GetNumberOfEvents();i++)
	{
		Asura_Sound_Event* pxEvent = Asura_Sound_Event_Storage::GetEventByIndex(i);
		pxEvent->UpdatePlayableListFromIDList();
	}

	//refresh the sound event and sample editing tabs
	Tools_SoundEvent_Overlay::HandleReload();

	//get the phonons to find their events again
	Tools_Phonon_Object::HandleSoundEventReload();

	SetFilesToCheckForExternalModification();
}


/////////////////////////////////////////////////////////////////////////////////////////

void Tools_MessageLogAction_SoundEvent::Execute()
{
	Tools_SoundEvent_Overlay::EditEvent(m_uHashID);	
}

void Tools_MessageLogAction_SoundSample::Execute()
{
	Tools_SoundEvent_Overlay::EditSample(m_uHashID);	
}

/////////////////////////////////////////////////////////////////////////////////////////

Tools_WavFinder::~Tools_WavFinder()
{
	delete m_pxWavTree;
	delete m_pxFileList;
}

const char* Tools_WavFinder::FindFile(const char* szOldFilename)
{
	//initialise tree if not done yet
	if(!m_pxWavTree) BuildTree();

	const char* szOldStrippedFilename = Tools_FileSystem::StripPath(szOldFilename);
	Asura_Hash_ID uHash = Asura_GetHashID(szOldStrippedFilename);

	//get a list of filenames with that hash
	const List<const char*>* pxList = m_pxWavTree->GetDataByHash(uHash);
	if(!pxList) return 0;

	//See if any matches
	for(CLIF<const char*> xFileLIF(pxList);!xFileLIF.done();xFileLIF.next())
	{
		const char* szNewStrippedFilename = Tools_FileSystem::StripPath(xFileLIF());
		if(!stricmp(szOldStrippedFilename,szNewStrippedFilename))
		{
			//found a file , return the relative path
			return Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir(xFileLIF());
		}
	}

	return 0;
}

void Tools_WavFinder::BuildTree()
{
	m_pxWavTree = new Asura_TemplateHashTree<List<const char*> >;
	m_pxFileList = new List<Asura_String>;

	//serach for all wav files locally and across the network

	char szSearch[MAX_PATH];
	
	strcpy(szSearch,"Sounds\\*.wav");
	Tools_FileSystem::FindFiles(szSearch,*m_pxFileList,TOOLS_FILEFIND_RECURSIVE|TOOLS_FILEFIND_IGNOREBACKUPDIR);

	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	const char* szNetDir = Asura_FileAccess::PlatformGetNetworkPath();

	Asura_PrintfToBfr(szSearch,"%s\\Sounds\\*.wav",sizeof(szSearch),true,szNetDir);
	Tools_FileSystem::FindFiles(szSearch,*m_pxFileList,TOOLS_FILEFIND_RECURSIVE|TOOLS_FILEFIND_IGNOREBACKUPDIR);
	#endif


	//Add all the files to the hash tree , based on the hash value of the name without path
	for(LIF<Asura_String> xFileLIF(m_pxFileList);!xFileLIF.done();xFileLIF.next())
	{
		const char* szStrippedFilename = Tools_FileSystem::StripPath(xFileLIF().GetString());
		Asura_Hash_ID uHash = Asura_GetHashID(szStrippedFilename);

		//add/fine node in tree for this hash value
		List<const char*>* pxList = m_pxWavTree->GetDataByHash(uHash);
		if(!pxList) pxList = m_pxWavTree->AddNode(uHash);
		Asura_Assert(pxList,"Error adding to hash tree");

		//add to the list of names with that hash
		pxList->add_entry(xFileLIF().GetString());
	}
}

/////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------**
** Tools_SoundSamplePreview **
**---------------------------*/

Tools_SoundSamplePreview::Tools_SoundSamplePreview()
{
	m_iSoundHandle = ASURA_INVALID_HANDLE;
	m_fTimer = 0.0f;
	m_uSampleHash =0;
}

Tools_SoundSamplePreview::~Tools_SoundSamplePreview()
{
	Stop();
}

void Tools_SoundSamplePreview::Play(Asura_Hash_ID uSampleHash)
{
	Stop();

	const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::FindSample(uSampleHash);
	if(!pxSample) return;

	#if defined ASURA_USE_AUDIO_SYSTEM

		if(pxSample->GetWaveHash() == ASURA_HASH_ID_UNSET) return;

		m_iSoundHandle = Asura_Audio_System::CreateVoice(pxSample->GetWaveHash(),false,false);

		if(m_iSoundHandle != ASURA_INVALID_HANDLE)
		{
			Asura_Audio_Voice* const pxVoice = Asura_Audio_System::GetVoice(m_iSoundHandle);

			pxVoice->SetSpeakerMappingType( ASURA_ASMT_DIRECT );
			pxVoice->SetVolume(pxSample->m_fVolume);
			pxVoice->SetPitch(pxSample->m_fPitch);

			pxVoice->Start();

			m_fTimer = 0.0f;
			m_uSampleHash = uSampleHash;
		}

	#else

		if(pxSample->GetSoundID() == ASURA_SID_NO_SOUND) return;

		m_iSoundHandle = Asura_Sound_System::CreateSound(pxSample->GetSoundID());	
		if(m_iSoundHandle != ASURA_INVALID_HANDLE)
		{
			Asura_Sound_System::SetSoundSpatial(m_iSoundHandle,false);
			Asura_Sound_System::SetSoundVolume(m_iSoundHandle,pxSample->m_fVolume);
			Asura_Sound_System::SetSoundPitch(m_iSoundHandle,pxSample->m_fPitch);

			Asura_Sound_System::PlaySound(m_iSoundHandle,pxSample->m_bLooping);

			m_fTimer = 0.0f;
			m_uSampleHash = uSampleHash;
		}

	#endif
}

void Tools_SoundSamplePreview::Update()
{
	if(m_iSoundHandle == ASURA_INVALID_HANDLE) return;

	const Asura_Sound_Sample* pxSample = Asura_Sound_Event_Storage::FindSample(m_uSampleHash);
	if(!pxSample)
	{
		Stop();
		return;
	}

	//check timer
	m_fTimer += Asura_Timers::GetFrameTime();
	if(pxSample->m_fDuration && pxSample->m_fDuration < m_fTimer)
	{
		Stop();
		return;
	}

	#if defined ASURA_USE_AUDIO_SYSTEM

		Asura_Audio_Voice* const pxVoice = Asura_Audio_System::GetVoice(m_iSoundHandle);
		if(pxVoice)
		{
			pxVoice->SetVolume(pxSample->m_fVolume);
			pxVoice->SetPitch(pxSample->m_fPitch);
		}

	#else

		//update volume and pitch
		Asura_Sound_System::SetSoundVolume(m_iSoundHandle,pxSample->m_fVolume);
		Asura_Sound_System::SetSoundPitch(m_iSoundHandle,pxSample->m_fPitch);

	#endif
}

void Tools_SoundSamplePreview::Stop()
{
	#if defined ASURA_USE_AUDIO_SYSTEM
		Asura_Audio_System::DestroyVoice(m_iSoundHandle);
	#else
		Asura_Sound_System::DestroySound(m_iSoundHandle);
	#endif
}

//eof
