//
// Tools_Sound_System.cpp
//
// Created by Richard Rice - 19/08/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Sound/Tools_Sound_System.h"

#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_ReverbPresets.h"
#include "Core/Asura_SortAndSearch.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_UserInput.h"
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Key_Combination.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_WAVAlign_Dialog.h"
#include "Sound/Asura_Audio_System.h"
#include "Sound/Asura_Audio_Voice.h"
#include "Sound/Asura_Phonon.h"
#include "Sound/Asura_Sound.h"
#include "Sound/Asura_Sound_Mixer.h"
#include "Sound/Asura_StreamingBackgroundSound_System.h"
#include "Sound/Asura_Wav.h"
#include "Sound/Asura_XAudio2.h"
#include "Sound/Asura_XAudio2_Buffer.h"

#include "Tools_Overlays.h"
#include "ChunkSystem/Tools_Chunk_StreamingSounds.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Misc/Tools_Dialog.h"
#include "Misc/Tools_IniFile.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/ChunkSystem/Asura_Chunk_Stream_WindowsFile.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Sound/Tools_ReverbPresets.h"
#include "Texturing/Tools_Texture_Management.h"

bool Tools_Sound_System::s_bMuteStreaming = false;
bool Tools_Sound_System::s_bMutePhonons = false;
bool Tools_Sound_System::s_bDontLoadWavs = false;
bool Tools_Sound_System::s_bLoadWavsOnQuitToGame = true;

List<Tools_Mixer_Setting> Tools_Sound_System::s_xMixerSettings;
List<Tools_Mixer_Setting*> Tools_Sound_System::s_xSortedMixerSettings;
bool Tools_Sound_System::s_bMuteMixersChanged = false;

#ifdef ASURA_USE_AUDIO_SYSTEM
bool Tools_Sound_System::s_bStreamingChanged = false;
bool Tools_Sound_System::s_bReverbChanged = false;
#endif

static const char* g_szMixerOptionsFile = "ToolsIniData\\MixerOptions.txt";


void Tools_Sound_System::Initialise()
{
	//apply the mixer settings for use in tools
	LoadMixerSettings();
	ApplyMixerSettings(true);
}

void Tools_Sound_System::Shutdown()
{
#ifdef ASURA_USE_AUDIO_SYSTEM
	DestroyStreamingOverlay();
	DestroyReverbOverlay();
#endif

	Asura_Phonon_System::EnablePhonons();//make sure sounds don't remain disabled
	Asura_StreamingBackgroundSound_System::SetDisabled(false);

	//apply the mixer settings for use in game
	SaveMixerSettings();
	ApplyMixerSettings(false);
}

void Tools_Sound_System::Update()
{
	//update sound muteness
    if(s_bMutePhonons)
	{
		Asura_Phonon_System::DisablePhonons();
	}
	else
	{
    	Asura_Phonon_System::EnablePhonons();
	}

	Asura_StreamingBackgroundSound_System::SetDisabled(s_bMuteStreaming);
        
	if(s_bMuteMixersChanged)
	{
		ApplyMixerSettings(true);
		s_bMuteMixersChanged = false;
	}
}


void Tools_Sound_System::AddOverlayControls(Asura_Overlay* pxOverlay,Asura_Vector_2& xOffset)
{
	Asura_Dialog* pxDialog;


	pxDialog = new Asura_Tickbox_Dialog("Mute Phonons",xOffset,&s_bMutePhonons);
	pxDialog->SetHelpText("Disables phonons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Mute Streaming",xOffset,&s_bMuteStreaming);
	pxDialog->SetHelpText("Disables streaming background sounds.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Dont Load Wavs",xOffset,&s_bDontLoadWavs);
	pxDialog->SetHelpText("Prevents wav files from being loaded.\nOnce a wav file has been prevented from loading in this way , it will remain unloaded until you next load tools even if you change this option.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Load Wavs On Quit To Game",xOffset,&s_bLoadWavsOnQuitToGame);
	pxDialog->SetHelpText("Causes all wavs used by sound events to be loaded when quitting to game.\nDisabling this will speed up quitting to game at the cost of delays in game when sounds are loaded as required.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	CreateMixerSettings();

	//add tickboxes for muting the mixers in game and in tools
	pxDialog = new Asura_Dialog("Mute Mixers :", xOffset);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	static char* s_apszTabs[2] = {"Tools","Game"};
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset,s_apszTabs,2);
	pxOverlay->AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();
	pxOverlay->PushCurrentParentDialog(pxTabsDialog);

	float fTabsStartY = xOffset.y;
	float fTabsEndY = xOffset.y;

	for(int i=0;i<2;i++)
	{
		pxTabsDialog->SelectTab(i);
		xOffset.y = fTabsStartY;
		const char* szHelpText = (i==0) ? "Mutes this mixer while in tools"
										: "Mutes this mixer when quitting to game";

		for(LIF<Tools_Mixer_Setting*> xSettingLIF(&s_xSortedMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
		{
			Tools_Mixer_Setting* pxSetting = xSettingLIF();

			const char* szName = pxSetting->m_xName.GetString();
			bool* pbMute = (i==0) ? &pxSetting->m_bMuteInTools : &pxSetting->m_bMuteInGame;

			xOffset.x = 1.0f + 1.0f*pxSetting->m_iIndent;

			pxDialog = new Asura_Tickbox_Dialog(szName, xOffset, pbMute, &s_bMuteMixersChanged);
			pxDialog->SetHelpText(szHelpText);
			pxOverlay->AddDialog(pxDialog);
			xOffset.y += Asura_Dialog::DefaultSpacing();

		}

		fTabsEndY = Asura_Maths::Max(xOffset.y,fTabsEndY);
	}
	xOffset.y = fTabsEndY;
	xOffset.x = 1.0f;
	pxOverlay->PopCurrentParentDialog();

	static u_int s_uSelectedTab = 0;
	pxTabsDialog->SetSelectedTabPointer(&s_uSelectedTab);
}


#ifdef ASURA_USE_AUDIO_SYSTEM

void Tools_Sound_System::CreateStreamingOverlay()
{
	Tools_Audio_StreamingOverlay* const pxOverlay = new Tools_Audio_StreamingOverlay( false );
	Asura_Overlay_System::SubmitNewOverlay( pxOverlay );
}

void Tools_Sound_System::DestroyStreamingOverlay()
{
	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_STREAMING_SOUNDS );
}

void Tools_Sound_System::ToggleStreamingOverlay()
{
	if( Tools_Audio_StreamingOverlay::IsOpen() )
	{
		DestroyStreamingOverlay();
	}
	else
	{
		CreateStreamingOverlay();
	}
}

void Tools_Sound_System::SaveStreaming()
{
	const HANDLE xFileHandle = CreateFile( szASURA_AUDIO_STREAMING_FILENAME, GENERIC_WRITE, 0, NULL,
		CREATE_ALWAYS, 0, NULL );

	if( xFileHandle == INVALID_HANDLE_VALUE )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR,
			"Couldn't open '%s' for writing. Is it checked out?",
			szASURA_AUDIO_STREAMING_FILENAME );

		return;
	}

	{
		Asura_Chunk_Stream_WindowsFile xStream( xFileHandle );

		xStream.Write( ASURA_FILE_ID_NUM_CHARS, ASURA_UNCOMPRESSED_FILE_ID );

		Tools_Chunk_StreamingSounds::Create( xStream );

		xStream.FlushOutput();
		xStream.WriteAsuraTerminator();
	}

	CloseHandle( xFileHandle );

	s_bStreamingChanged = false;

	Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_SUCCESS,
		"Saved streaming sounds successfully." );
}

void Tools_Sound_System::CreateReverbOverlay()
{
	Tools_Audio_ReverbOverlay* const pxOverlay = new Tools_Audio_ReverbOverlay();
	Asura_Overlay_System::SubmitNewOverlay( pxOverlay );
}

void Tools_Sound_System::DestroyReverbOverlay()
{
	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_REVERB );
}

void Tools_Sound_System::ToggleReverbOverlay()
{
	if( Tools_Audio_ReverbOverlay::IsOpen() )
	{
		DestroyReverbOverlay();
	}
	else
	{
		CreateReverbOverlay();
	}
}

void Tools_Sound_System::SaveReverb()
{
	const HANDLE xFileHandle = CreateFile( szTOOLS_AUDIO_REVERB_FILENAME, GENERIC_WRITE, 0, NULL,
		CREATE_ALWAYS, 0, NULL );

	if( xFileHandle == INVALID_HANDLE_VALUE )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR,
			"Couldn't open '%s' for writing. Is it checked out?",
			szTOOLS_AUDIO_REVERB_FILENAME );

		return;
	}

	{
		Asura_Chunk_Stream_WindowsFile xStream( xFileHandle );

		xStream.Write( ASURA_FILE_ID_NUM_CHARS, ASURA_UNCOMPRESSED_FILE_ID );

		Asura_Chunk_ReverbPresets::Create( xStream );

		xStream.FlushOutput();
		xStream.WriteAsuraTerminator();
	}

	CloseHandle( xFileHandle );

	s_bReverbChanged = false;

	Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_SUCCESS,
		"Saved reverb presets successfully." );
}

void Tools_Sound_System::ResetReverb( Asura_Audio_Reverb& xReverb, const u_int uPresetIndex )
{
	if( uPresetIndex >= uTOOLS_NUM_REVERB_PRESETS )
	{
		Asura_Assertf( false, ("Reverb preset index (%u) is too high.", uPresetIndex) );
		return;
	}

	const Tools_ReverbPreset xPreset = axTOOLS_REVERB_PRESETS[ uPresetIndex ];

	xReverb.m_iRoom              = xPreset.m_iRoom;
	xReverb.m_iRoomHF            = xPreset.m_iRoomHF;
	xReverb.m_fRoomRolloffFactor = xPreset.m_fRoomRolloffFactor;
	xReverb.m_fDecayTime         = xPreset.m_fDecayTime;
	xReverb.m_fDecayHFRatio      = xPreset.m_fDecayHFRatio;
	xReverb.m_iReflections       = xPreset.m_iReflections;
	xReverb.m_fReflectionsDelay  = xPreset.m_fReflectionsDelay;
	xReverb.m_iReverb            = xPreset.m_iReverb;
	xReverb.m_fReverbDelay       = xPreset.m_fReverbDelay;
	xReverb.m_fDiffusion         = xPreset.m_fDiffusion;
	xReverb.m_fDensity           = xPreset.m_fDensity;
	xReverb.m_fHFReference       = xPreset.m_fHFReference;
}

#endif // ASURA_USE_AUDIO_SYSTEM


float Tools_Sound_System::GetSoundLength(const char* szFilename)
{
	#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char szNetFilename[MAX_PATH];
	Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy(szFilename,szNetFilename,MAX_PATH);
	szFilename = szNetFilename;
	#endif

	CWaveSoundRead xWavRead;
	if(FAILED (xWavRead.Open(szFilename))) return 0.0f;
	if(!xWavRead.m_pwfx->nAvgBytesPerSec) return 0.0f;

	return xWavRead.m_ckIn.cksize/(float)xWavRead.m_pwfx->nAvgBytesPerSec;
}


float Tools_Sound_System::GetSoundLength( const Asura_Hash_ID uWave )
{
	#ifdef ASURA_USE_AUDIO_SYSTEM
	Asura_XAudio2_Buffer* pxBuffer = Asura_XAudio2::GetBuffer( uWave );
	if( pxBuffer )
	{
		pxBuffer->LoadIfDelayed();
		return pxBuffer->GetDuration();
	}
	#endif
	return 0.0f;
}


#if defined ASURA_USE_AUDIO_SYSTEM
bool Tools_Sound_System::GetSoundWaveData( Asura_Hash_ID uWave, WAVEFORMATEX& xWaveData, MMCKINFO& xWaveCkInfo )
#else
bool Tools_Sound_System::GetSoundWaveData( Asura_Sound_ID eSoundID, WAVEFORMATEX& xWaveData, MMCKINFO& xWaveCkInfo )
#endif
{
	// Find out file name (if stored)
	const char* szWaveFileName = NULL;
#if defined ASURA_USE_AUDIO_SYSTEM
	szWaveFileName = Asura_XAudio2::GetWaveFilename( uWave );
#else
	if( eSoundID == ASURA_SID_NO_SOUND )
	{
		szWaveFileName = "";
	}
	else
	{
		szWaveFileName = Asura_Sound_System::GetFileNameFromSoundID( eSoundID );
	}
#endif

	// Retrieve from network
#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
	char szNetFilename[MAX_PATH];
	Asura_FileAccess::PlatformGetNetworkFilenameOrGetLocalCopy( szWaveFileName, szNetFilename, MAX_PATH );
	szWaveFileName = szNetFilename;
#endif

	// Read file
	CWaveSoundRead xWavRead;
	if( FAILED ( xWavRead.Open( szWaveFileName ) ) ) return false;

	xWaveData = *(xWavRead.m_pwfx);
	xWaveCkInfo = xWavRead.m_ckIn;

	return true;
}


void Tools_Sound_System::AddMixerSetting(const char* szName)
{
	//make sure the mixer isn't already in the list
	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
	{
		if(!stricmp(szName,xSettingLIF().m_xName.GetString()))
		{
			return;
		}
	}

	s_xMixerSettings.add_entry(Tools_Mixer_Setting(szName));
}

void Tools_Sound_System::CreateMixerSettings()
{
	AddMixerSetting("Master");

	for(u_int i=0; i<Asura_Sound_Mixer_System::GetNumberOfChannels(); i++)
	{
		Asura_Sound_Mixer_Channel* pxMixer = Asura_Sound_Mixer_System::GetChannelByIndex(i);

		AddMixerSetting(pxMixer->GetName());
	}

	//delete any settings for mixers that no longer exist , and find the parents
	//of the ones that do
	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done();)
	{
		bool bDelete = false;
		
		Tools_Mixer_Setting& xSetting = xSettingLIF();
		const char* szName = xSetting.m_xName.GetString();

		xSetting.m_xParent.MakeNull();
		
		if(stricmp("Master",szName))
		{
			//not the master mixer , so check for a channel for this
			const Asura_Sound_Mixer_Channel* pxMixerChannel = Asura_Sound_Mixer_System::FindChannel(Asura_GetHashID(szName));
			if(pxMixerChannel)
			{
				xSetting.m_xParent = pxMixerChannel->GetOutputMixerName();
				if(xSetting.m_xParent.IsBlank()) xSetting.m_xParent = "Master";
			}
			else
			{
				bDelete = true;
			}
		}

		if(bDelete)
		{
			xSettingLIF.delete_current();
		}
		else
		{
			xSettingLIF.next();
		}
	}

	//sort the mixers
	while(s_xSortedMixerSettings.size()) s_xSortedMixerSettings.delete_first_entry();
	AddToSortedMixerSettings(&s_xMixerSettings.first_entry(),0);
}


void Tools_Sound_System::LoadMixerSettings()
{
	if(s_xMixerSettings.size()) return;

	//make sure we have settings for all existing mixers;
	CreateMixerSettings();

	Tools_IniFile xIni(g_szMixerOptionsFile,false);
	if(!xIni.LoadIniFile()) return;

	//read in the settings for each mixer
	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
	{
		Tools_Mixer_Setting& xSetting = xSettingLIF();

		if(xIni.GetIniTextSection(xSetting.m_xName.GetString()))
		{
			xIni.GetBoolInCurrentSection("MuteInGame", xSetting.m_bMuteInGame);
			xIni.GetBoolInCurrentSection("MuteInTools", xSetting.m_bMuteInTools);
		}
	}
}

void Tools_Sound_System::SaveMixerSettings()
{
	if(!s_xMixerSettings.size()) return;

	FILE* pFile = fopen(g_szMixerOptionsFile,"wt");
	if(!pFile) return;

	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
	{
		Tools_Mixer_Setting& xSetting = xSettingLIF();

		fprintf(pFile,"[%s]\n",xSetting.m_xName.GetString());
		
		fprintf(pFile,"MuteInGame = %s\n", xSetting.m_bMuteInGame ? "true" : "false");
		fprintf(pFile,"MuteInTools = %s\n", xSetting.m_bMuteInTools ? "true" : "false");
		fprintf(pFile,"\n");
	}

	fclose(pFile);
}

void Tools_Sound_System::ApplyMixerSettings(bool bTools)
{
	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
	{
		Tools_Mixer_Setting& xSetting = xSettingLIF();

		const char* szName = xSetting.m_xName.GetString();
		bool bMute = bTools ? xSetting.m_bMuteInTools : xSetting.m_bMuteInGame;

		Asura_Sound_Mixer_System::SetToolsMute(Asura_GetHashID(szName), bMute);
	}
}

void Tools_Sound_System::AddToSortedMixerSettings(Tools_Mixer_Setting* pxSetting, int iIndent)
{
	pxSetting->m_iIndent = iIndent;
	s_xSortedMixerSettings.add_entry(pxSetting);

	//Get a sorted list of children
	List<Tools_Mixer_Setting*> xChildList;
	LIF<Tools_Mixer_Setting*> xChildLIF(&xChildList);


	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
	{
		if(!pxSetting->m_xName.Compare(xSettingLIF().m_xParent))
		{
			//found a child , so add to the list
			Tools_Mixer_Setting* pxChild = &xSettingLIF();
			
			for(xChildLIF.restart(); !xChildLIF.done(); xChildLIF.next())
			{
				if(pxChild->m_xName.Compare(xChildLIF()->m_xName) < 0)
				{
					xChildList.add_entry_before(pxChild, xChildLIF());
					break;
				}
			}
			if(xChildLIF.done())
			{
				xChildList.add_entry(pxChild);
			}
			
		}
	}

	//Add these to the sorted list
	for(xChildLIF.restart(); !xChildLIF.done(); xChildLIF.next())
	{
		AddToSortedMixerSettings(xChildLIF(), iIndent+1);
	}
}


bool Tools_Sound_System::AreAnyMixersMuted()
{
	for(LIF<Tools_Mixer_Setting> xSettingLIF(&s_xMixerSettings); !xSettingLIF.done(); xSettingLIF.next())
	{
		if(xSettingLIF().m_bMuteInTools) return true;
	}
	return false;
}

bool Tools_Sound_System::IsMusicMixerMuted()
{
	return Asura_Sound_Mixer_System::GetMixerVolume(Asura_Sound_Mixer_System::FindChannel(uMIXER_HASHID_MUSIC)) == 0.0f;
}


#ifdef ASURA_USE_AUDIO_SYSTEM

Tools_Audio_StreamingOverlay* Tools_Audio_StreamingOverlay::s_pxThis = NULL;
Asura_Hash_ID Tools_Audio_StreamingOverlay::s_uModalResult = ASURA_HASH_ID_UNSET;

Tools_Audio_StreamingOverlay::Tools_Audio_StreamingOverlay( const bool bSelectionMode )
	: Asura_Overlay( "Streaming Sounds", false )
	, m_pxFilenameCaption( NULL )
	, m_pxPathCaption( NULL )
	, m_pxWaveform( NULL )
	, m_pxLoopControl( NULL )
	, m_iPreviewHandle( ASURA_INVALID_HANDLE )
	, m_bSelectionChanged( false )
	, m_bSoundsChanged( false )
{
	m_uID = TOOLS_OVERLAY_STREAMING_SOUNDS;

	// Add dialogs
	Asura_Vector_2 xOffset( 1.0f, 2.0f );
	Asura_Dialog* pxDialog;

	bool bWasNew;
	int iTexID, iTexWidth = 0, iTexHeight = 0;
	const Asura_Vector_2 xBtnSize( 24.0f, 24.0f );

	// Data for the button dialogs
	{
		iTexID = Asura_Texture_Management::AddTextureWithResetGroup( "\\Tools\\Buttons.bmp",
			ASURA_TEXTURE_LOADED_PERMANENT, &bWasNew );

		if( bWasNew )
		{
			Asura_Texture_Management::LoadTextures();
		}

		Tools_Texture_Management::GetTextureSize( iTexID, iTexWidth, iTexHeight );
	}

	const int iIconTexture = Tools_Dialog::GetIconTexture();

	// Add button
	pxDialog = new Asura_Function_Dialog( "Add New", xOffset, OnAddFile );
	pxDialog->SetHelpText( "Adds a new streaming sound file." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();

	// Filename tree
	pxDialog =
		new Asura_Tree_Dialog( xOffset, &m_xFilenameTree, 25.0f, 30.0f, &m_bSelectionChanged );
	AddDialog( pxDialog );
	xOffset.y = pxDialog->m_xMaxCoord.y + 1.0f;

	// Separator
	pxDialog = new Asura_Divider_Dialog( xOffset );
	AddDialog( pxDialog );
	xOffset.y += Asura_Divider_Dialog::DefaultSpacing();

	// Filename caption
	m_pxFilenameCaption = new Asura_StringDisplay_Dialog( "No file selected", xOffset );
	AddDialog( m_pxFilenameCaption );

	// Remove button
	Asura_Icon_Dialog* const pxIconDlg =
		new Asura_Icon_Dialog( Asura_Maths::ZeroVector2, iIconTexture, TOOLS_ICON_DELETE );
	pxIconDlg->SetClickedFunction( OnRemoveFile );
	pxIconDlg->SetHelpText( "Removes the selected sound from the project." );
	pxIconDlg->SetOverlayRelativePosition( Asura_Vector_2( -3.0f, xOffset.y ), false, true );
	pxIconDlg->SetIsEnabledFunction( IsLeafNodeSelected );
	AddDialog( pxIconDlg );

	xOffset.y += Asura_Dialog::DefaultSpacing();

	// Directory caption
	m_pxPathCaption = new Asura_StringDisplay_Dialog( "", xOffset );
	AddDialog( m_pxPathCaption );
	xOffset.y += Asura_StringDisplay_Dialog::LargeSpacing();

	// Looping tickbox
	m_pxLoopControl = new Asura_Tickbox_Dialog( "Looping", xOffset, NULL, &m_bSoundsChanged );
	m_pxLoopControl->SetHelpText( "Sets whether this sound loops or not." );
	m_pxLoopControl->SetIsEnabledFunction( IsLeafNodeSelected );
	AddDialog( m_pxLoopControl );
	xOffset.y += Asura_Tickbox_Dialog::LargeSpacing();

	// Waveform display
	m_pxWaveform = new Asura_WAVAlign_Dialog( xOffset, ASURA_WAVALIGN_DIALOG_DEFAULT_WIDTH, 48.0f );
	m_pxWaveform->AddWAVEntry();
	m_pxWaveform->SetRenderFlags( ASURA_WAVALIGN_RENDER_PLAYBACK_MARKER );
	AddDialog( m_pxWaveform );

	float fOldX = xOffset.x;
	xOffset.x = (m_pxWaveform->m_xMaxCoord.x - m_pxWaveform->m_xOffset.x) * 0.5f - 2.0f;
	xOffset.y = m_pxWaveform->m_xMaxCoord.y;

	const Asura_Vector_2 xPlayPos( 54.0f, 0.0f );
	const Asura_Vector_2 xStopPos( 102.0f, 0.0f );

	// Play button
	pxDialog =
		new Asura_Button_Dialog( xOffset, iTexID, iTexWidth, xPlayPos, xBtnSize, OnPreviewPlay );
	pxDialog->SetIsEnabledFunction( IsLeafNodeSelected );
	pxDialog->SetHelpText( "Begins playback of the selected sound." );
	AddDialog( pxDialog );

	xOffset.x = pxDialog->m_xMaxCoord.x;

	// Stop button
	pxDialog =
		new Asura_Button_Dialog( xOffset, iTexID, iTexWidth, xStopPos, xBtnSize, OnPreviewStop );
	pxDialog->SetIsEnabledFunction( IsLeafNodeSelected );
	pxDialog->SetHelpText( "Stops the preview." );
	AddDialog( pxDialog );

	xOffset.x = fOldX;
	xOffset.y = pxDialog->m_xMaxCoord.y + 1.5f;

	// Separator
	pxDialog = new Asura_Divider_Dialog( xOffset );
	AddDialog( pxDialog );
	xOffset.y += Asura_Divider_Dialog::DefaultSpacing();

	if( bSelectionMode )
	{
		// Select button
		pxDialog = new Asura_FunctionWithKeybinding_Dialog( "Select", xOffset, OnSelect,
			ASURA_KEYID_ENTER );
		pxDialog->SetIsEnabledFunction( IsLeafNodeSelected );
		pxDialog->SetHelpText( "Closes this overlay and accepts the selection." );
		AddDialog( pxDialog );

		fOldX = xOffset.x;
		xOffset.x = pxDialog->m_xMaxCoord.x + 2.0f;

		// Cancel button
		pxDialog = new Asura_FunctionWithKeybinding_Dialog( "Cancel", xOffset, OnCancel,
			ASURA_KEYID_ESCAPE );
		pxDialog->SetHelpText( "Closes this overlay and cancels the operation." );
		AddDialog( pxDialog );

		xOffset.x = fOldX;
		xOffset.y += Asura_FunctionWithKeybinding_Dialog::DefaultSpacing();
	}
	else
	{
		// Export button
		pxDialog = new Asura_Function_Dialog( "Export", xOffset,
			Tools_Level_Export::ExportStreamingSounds );
		pxDialog->SetHelpText( "Exports all streaming sounds to retail platforms." );
		AddDialog( pxDialog );
		xOffset.y += Asura_Function_Dialog::DefaultSpacing();
	}

	// Build the initial tree
	{
		const Asura_TemplateHashTree< Asura_Audio_StreamInfo >& xDataTree = 
			Asura_Audio_System::GetStreamInfoTree();

		for( Asura_TemplateHashTree_ConstIterator< Asura_Audio_StreamInfo > xIter( xDataTree );
			!xIter.Done(); xIter.Next() )
		{
			const Asura_Audio_StreamInfo* const pxCurrStream = xIter.GetCurrent();

			// Skip streaming dialogue which has been added through playback
			if( pxCurrStream->m_bDialogue ) continue;

			Asura_TreeDialog_Item* const pxItem =
				m_xFilenameTree.AddFilePath( pxCurrStream->m_szFilename );

			if( pxItem )
			{
				// Node hash will default to just the filename. Include the whole path here
				pxItem->SetHash( Asura_GetHashID( pxCurrStream->m_szFilename ) );
			}
		}
	}

	// Store the singleton pointer
	Asura_Assert( !s_pxThis, "Created more than one streaming sounds overlay." );
	s_pxThis = this;
}

Tools_Audio_StreamingOverlay::~Tools_Audio_StreamingOverlay()
{
	OnPreviewStop();
	s_pxThis = NULL;
}

Asura_Hash_ID Tools_Audio_StreamingOverlay::SelectStream()
{
	// Ditch any existing copy, since class is supposed to be singleton
	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_STREAMING_SOUNDS );

	Tools_Audio_StreamingOverlay* const pxOverlay = new Tools_Audio_StreamingOverlay( true );
	Asura_Overlay_System::SubmitNewOverlay( pxOverlay, false, false, true );

	Tools_Modal_Overlay::DoModal( pxOverlay );

	return s_uModalResult;
}

void Tools_Audio_StreamingOverlay::PerFrameUpdate()
{
	Asura_Overlay::PerFrameUpdate();

	if( m_bSoundsChanged )
	{
		// Trap things like looping being toggled
		Tools_Sound_System::SetStreamingModified();
		m_bSoundsChanged = false;
	}

	if( m_bSelectionChanged )
	{
		OnSelectionChange();
		m_bSelectionChanged = false;
	}

	if( m_pxWaveform )
	{
		float fPreviewTime = 0.0f;

		if( m_iPreviewHandle != ASURA_INVALID_HANDLE )
		{
			Asura_Audio_Voice* const pxVoice = Asura_Audio_System::GetVoice( m_iPreviewHandle );
			if( pxVoice )
			{
				fPreviewTime = pxVoice->GetPlayPosition();
			}
		}

		m_pxWaveform->SetPlaybackMarkerTime( fPreviewTime );
	}
}

bool Tools_Audio_StreamingOverlay::IsLeafNodeSelected()
{
	Asura_Assert( s_pxThis, "No streaming overlay exists." );
	if( !s_pxThis ) return false;

	// Leaf nodes represent actual waves, others just directories

	const Asura_TreeDialog_Item* const pxSelectedItem = s_pxThis->m_xFilenameTree.GetSelectedItem();
	return pxSelectedItem ? pxSelectedItem->GetNumChildren() == 0 : false;
}

void Tools_Audio_StreamingOverlay::OnAddFile()
{
	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	Tools_FileSystem_DialogInfo xInfo;

	xInfo.DefaultExtension = "";

	char szDefaultName[ MAX_PATH ] = "";
	xInfo.DefaultFilename = szDefaultName;

	char szDefaultPath[ MAX_PATH ] = "";
	xInfo.DefaultPath = szDefaultPath;

	xInfo.FileFilter = "Sound Files (*.wav)\0*.wav\0All Files\0*.*\0";
	xInfo.WindowTitle = "Add Streaming Sound";
	xInfo.AllowMultipleSelect = true;

	Asura_String xFilename = Tools_FileSystem::GetLoadFilename( xInfo );

	while( !xFilename.IsBlank() )
	{
		const char* const szRelativePath = 
			Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir( xFilename.GetString() );

		// Check if the sound is already registered. Don't use accessor function, as that will
		// assert if it's not present!
		const Asura_Hash_ID uHash = Asura_GetHashID( szRelativePath );
		if( !Asura_Audio_System::GetStreamInfoTree().GetDataByHash( uHash ) )
		{
			// In Tools, the package is the wave file itself
			Asura_Audio_System::AddStreamInfo( szRelativePath, szRelativePath, 0, false );

			Asura_TreeDialog_Item* const pxItem =
				s_pxThis->m_xFilenameTree.AddFilePath( szRelativePath );

			if( pxItem )
			{
				// Node hash will default to just the filename. Include the whole path here
				pxItem->SetHash( uHash );
				pxItem->MakeVisible();

				// Select the new item
				s_pxThis->m_xFilenameTree.SetSelectedItem( pxItem );

				// SetSelection() doesn't invoke the 'selection changed' callback
				OnSelectionChange();
			}
		}

		xFilename = Tools_FileSystem::GetNextLoadFilename();
	}

	Tools_Sound_System::SetStreamingModified();
}

void Tools_Audio_StreamingOverlay::OnRemoveFile()
{
	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	Asura_TreeDialog_Item* const pxSelectedItem = s_pxThis->m_xFilenameTree.GetSelectedItem();
	if( !pxSelectedItem ) return;

	int iVisIndex = pxSelectedItem->GetVisibleIndex();

	OnPreviewStop();
	Asura_Audio_System::DeleteStreamInfo( pxSelectedItem->GetHash() );

	// Delete the item, and its parents all the way up if they're empty dirs now
	for( Asura_TreeDialog_Item* pxNode = pxSelectedItem; pxNode; )
	{
		Asura_TreeDialog_Item* const pxParent = pxNode->GetParent();
		const bool bOnlyChild = (pxParent ? pxParent->GetNumChildren() == 1 : false);

		s_pxThis->m_xFilenameTree.DeleteItem( pxNode );

		pxNode = (bOnlyChild && !pxParent->IsRoot() ? pxParent : NULL );
	}

	// Select the next visible item
	{
		// Clamp to the end if we're deleting the last one
		iVisIndex = Asura_Maths::Min( iVisIndex, s_pxThis->m_xFilenameTree.GetNumVisibleItems() - 1 );

		Asura_TreeDialog_Item* const pxNewItem =
			s_pxThis->m_xFilenameTree.GetVisibleItem( iVisIndex );

		if( pxNewItem )
		{
			s_pxThis->m_xFilenameTree.SetSelectedItem( pxNewItem );
			pxNewItem->MakeVisible();
		}

		OnSelectionChange();
	}
	
	Tools_Sound_System::SetStreamingModified();
}

void Tools_Audio_StreamingOverlay::OnSelect()
{
	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	const Asura_TreeDialog_Item* const pxSelectedItem = s_pxThis->m_xFilenameTree.GetSelectedItem();
	s_uModalResult = (pxSelectedItem ? pxSelectedItem->GetHash() : ASURA_HASH_ID_UNSET);

	// Must do this here, rather than with the flag, since the flag would close the overlay before
	// this gets executed
	s_pxThis->m_bShouldClose = true;
}

void Tools_Audio_StreamingOverlay::OnCancel()
{
	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	s_uModalResult = ASURA_HASH_ID_UNSET;
	s_pxThis->m_bShouldClose = true;
}

void Tools_Audio_StreamingOverlay::OnSelectionChange()
{
	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	// Ditch the preview if it's playing
	OnPreviewStop();

	Asura_String xPath, xFilename, xExtension;
	Asura_String xNameCaption = "No file selected", xPathCaption;
	bool* pbLooping = NULL;
	const char* szFilename = NULL;

	const Asura_TreeDialog_Item* const pxSelection = s_pxThis->m_xFilenameTree.GetSelectedItem();

	// Only leaf nodes represent wave files, the rest form the directory hierarchy
	if( pxSelection && pxSelection->GetNumChildren() == 0 )
	{
		const Asura_Hash_ID uHash = pxSelection->GetHash();
		if( uHash == ASURA_HASH_ID_UNSET ) return;

		Asura_Audio_StreamInfo* const pxInfo = Asura_Audio_System::GetStreamInfo( uHash );
		if( !pxInfo ) return;

		szFilename = pxInfo->m_szFilename;
		Asura_String( szFilename ).SplitFilename( &xPath, &xFilename, &xExtension );

		xNameCaption = xFilename;
		xNameCaption += ".";
		xNameCaption += xExtension;

		xPathCaption = "(";
		xPathCaption += xPath;
		xPathCaption += ")";

		pbLooping = &pxInfo->m_bLooping;
	}

	if( s_pxThis->m_pxFilenameCaption )
	{
		s_pxThis->m_pxFilenameCaption->UpdateLabel( xNameCaption.GetString() );
	}

	if( s_pxThis->m_pxPathCaption )
	{
		s_pxThis->m_pxPathCaption->UpdateLabel( xPathCaption.GetString() );
	}

	if( s_pxThis->m_pxLoopControl )
	{
		s_pxThis->m_pxLoopControl->SetValue( pbLooping );
	}

	if( s_pxThis->m_pxWaveform )
	{
		s_pxThis->m_pxWaveform->LoadWAV( szFilename );
	}
}

void Tools_Audio_StreamingOverlay::OnPreviewPlay()
{
	OnPreviewStop();

	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	const Asura_TreeDialog_Item* const pxSelectedItem = s_pxThis->m_xFilenameTree.GetSelectedItem();
	if( !pxSelectedItem ) return;

	const Asura_Hash_ID uHash = pxSelectedItem->GetHash();
	if( uHash == ASURA_HASH_ID_UNSET ) return;

	s_pxThis->m_iPreviewHandle = Asura_Audio_System::CreateStreamingVoice( uHash, false, false );
	if( s_pxThis->m_iPreviewHandle == ASURA_INVALID_HANDLE ) return;

	Asura_Audio_Voice* const pxVoice = Asura_Audio_System::GetVoice( s_pxThis->m_iPreviewHandle );
	if( !pxVoice ) return;

	pxVoice->Start();
}

void Tools_Audio_StreamingOverlay::OnPreviewStop()
{
	Asura_Assert( s_pxThis, "No streaming sound overlay exists." );
	if( !s_pxThis ) return;

	// DestroyVoice() sets the handle to invalid for us
	Asura_Audio_System::DestroyVoice( s_pxThis->m_iPreviewHandle );
}

Tools_Audio_ReverbOverlay* Tools_Audio_ReverbOverlay::s_pxThis = NULL;

Tools_Audio_ReverbOverlay::Tools_Audio_ReverbOverlay()
	: Asura_Overlay( "Reverbs", false )
	, m_pszReverbNames( NULL )
	, m_uNumReverbNames(0)
	, m_iSelectedReverb(0)
	, m_iSelectedPreset(0)
	, m_bModified( false )
{
	m_uID = TOOLS_OVERLAY_REVERB;

	// Add dialogs
	Asura_Vector_2 xOffset( 1.0f, 2.0f );
	Asura_Dialog* pxDialog;

	static Asura_Vector_2 xReverbListDlgSize( 10.0f, 24.0f );

	// Reverb list
	m_pxReverbListDlg = new Asura_List_Dialog( xOffset, m_pszReverbNames,
		&m_iSelectedReverb, 20, 0, OnSelectionChange, NULL, true );
	m_pxReverbListDlg->SetExpandWidthToFillOverlay( false );
	m_pxReverbListDlg->SetAllowResize( false, &xReverbListDlgSize );
	m_pxReverbListDlg->SetSingleClickFunction( Asura_Callback_Void< void >( OnSelectionChange ) );
	AddDialog( m_pxReverbListDlg );
	xOffset.y = m_pxReverbListDlg->m_xMaxCoord.y;

	// Create button
	pxDialog = new Asura_Function_Dialog( "Create Reverb", xOffset, OnCreateReverb );
	pxDialog->SetHelpText( "Creates a new reverb." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Delete button
	pxDialog = new Asura_Function_Dialog( "Delete Selected Reverb", xOffset, OnDeleteReverb );
	pxDialog->SetHelpText( "Deletes the selected reverb." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Reset to default button
	pxDialog = new Asura_Function_Dialog( "Reset Selected To:", xOffset, OnResetReverb );
	pxDialog->SetHelpText( "Resets the selected reverb to the chosen I3DL2 default values." );
	AddDialog( pxDialog );

	const float fOldX = xOffset.x;
	xOffset.x = pxDialog->m_xMaxCoord.x + 2.0f;

	// Preset list
	pxDialog = new Asura_DropDown_List_Dialog( xOffset, aszTOOLS_REVERB_PRESET_NAMES,
		&m_iSelectedPreset, 10, uTOOLS_NUM_REVERB_PRESET_NAMES, NULL );
	AddDialog( pxDialog );
	xOffset.x = fOldX;
	xOffset.y += Asura_Hash_DropDownList_Dialog::DefaultSpacing();

	UpdateDialogs();

	Asura_Assert( !s_pxThis, "Created more than one reverb overlay." );
	s_pxThis = this;
}

Tools_Audio_ReverbOverlay::~Tools_Audio_ReverbOverlay()
{
	delete[] m_pszReverbNames;
	s_pxThis = NULL;
}

void Tools_Audio_ReverbOverlay::PerFrameUpdate()
{
	Asura_Overlay::PerFrameUpdate();

	if( m_bModified )
	{
		Tools_Sound_System::SetReverbModified();
		m_bModified = false;
	}
}

Asura_Audio_Reverb* Tools_Audio_ReverbOverlay::GetSelectedReverb()
{
	if( static_cast< u_int >( m_iSelectedReverb ) >= m_uNumReverbNames ) return NULL;

	const char* const szSelectedReverb = m_pszReverbNames[ m_iSelectedReverb ];
	const Asura_Hash_ID uSelectedReverb = Asura_GetHashID( szSelectedReverb );

	return Asura_Audio_System::GetReverb( uSelectedReverb );
}

void Tools_Audio_ReverbOverlay::ResetReverbDialogs()
{
	Asura_Vector_2 xOffset;
	xOffset.x = m_pxReverbListDlg ? m_pxReverbListDlg->m_xMaxCoord.x : 0.0f;
	xOffset.y = 2.0f;

	Asura_Dialog* pxDialog;

	// Delete all existing reverb property dialogs
	while( true )
	{
		pxDialog = GetDialog( DID_REVERB_PROPERTY );
		if( !pxDialog ) break;

		DeleteSingleDialog( pxDialog );
	}

	Asura_Audio_Reverb* const pxReverb = GetSelectedReverb();
	if( !pxReverb ) return;

	// Min & max values taken from the I3DL2 spec

	// Room
	pxDialog = new Asura_Numerical_Dialog( "Room Effect Attenuation", xOffset, &pxReverb->m_iRoom,
		-10000, 0, 1, "%d", &m_bModified );
	pxDialog->SetHelpText( "Intensity of room effect, in mB." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Room HF
	pxDialog = new Asura_Numerical_Dialog( "Room Effect HF Attenuation", xOffset,
		&pxReverb->m_iRoomHF, -10000, 0, 1, "%d", &m_bModified );
	pxDialog->SetHelpText( "Additional attenuation at high frequencies (relative to Room Effect Attenuation), in mB." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Room rolloff factor
	pxDialog = new Asura_Numerical_Dialog( "Room Rolloff Factor", xOffset,
		&pxReverb->m_fRoomRolloffFactor, 0.0f, 10.0f, 0.1f, "%.1f", &m_bModified );
	pxDialog->SetHelpText( "Rolloff factor for reflected signals." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Decay time
	pxDialog = new Asura_Numerical_Dialog( "Decay Time", xOffset, &pxReverb->m_fDecayTime, 0.1f,
		20.0f, 0.01f, "%.2f", &m_bModified );
	pxDialog->SetHelpText( "Reverberation decay time at low frequencies, in seconds." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Decay HF ratio
	pxDialog = new Asura_Numerical_Dialog( "Decay HF Ratio", xOffset,
		&pxReverb->m_fDecayHFRatio, 0.1f, 2.0f, 0.01f, "%.2f", &m_bModified );
	pxDialog->SetHelpText( "Ratio of high-frequency decay time to that at low frequencies." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Reflections
	pxDialog = new Asura_Numerical_Dialog( "Reflections", xOffset, &pxReverb->m_iReflections,
		-10000, 1000, 1, "%d", &m_bModified );
	pxDialog->SetHelpText( "Extra attenuation of early reflections (relative to Room Effect Attenuation), in mB." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Reflections delay
	pxDialog = new Asura_Numerical_Dialog( "Reflections Delay", xOffset,
		&pxReverb->m_fReflectionsDelay, 0.0f, 0.3f, 0.001f, "%.3f", &m_bModified );
	pxDialog->SetHelpText( "Delay time of first reflection relative to direct path, in seconds." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Reverb
	pxDialog = new Asura_Numerical_Dialog( "Reverb", xOffset, &pxReverb->m_iReverb, -10000, 2000, 1,
		"%d", &m_bModified );
	pxDialog->SetHelpText( "Extra attenuation of late reverberation (relative to Room Effect Attenuation), in mB." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Reverb delay
	pxDialog = new Asura_Numerical_Dialog( "Reverb Delay", xOffset, &pxReverb->m_fReverbDelay, 0.0f,
		0.1f, 0.001f, "%.3f", &m_bModified );
	pxDialog->SetHelpText( "Delay time of reverberations relative to early reflections, in seconds." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Diffusion
	pxDialog = new Asura_Numerical_Dialog( "% Diffusion", xOffset, &pxReverb->m_fDiffusion, 0.0f,
		100.0f, 1.0f, "%.0f", &m_bModified );
	pxDialog->SetHelpText( "Echo density in late reverberation decay." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Density
	pxDialog = new Asura_Numerical_Dialog( "% Density", xOffset, &pxReverb->m_fDensity, 0.0f,
		100.0f, 1.0f, "%.0f", &m_bModified );
	pxDialog->SetHelpText( "Modal density in late reverberation decay." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	// Reference higher frequency
	pxDialog = new Asura_Numerical_Dialog( "Reference High Frequency", xOffset,
		&pxReverb->m_fHFReference, 20.0f, 20000.0, 0.1f, "%.1f", &m_bModified );
	pxDialog->SetHelpText( "Frequency relative to which HF effects above take place, in Hz." );
	pxDialog->SetControlID( DID_REVERB_PROPERTY );
	AddDialog( pxDialog );
	xOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

	CalculateOverlaySize();
}

void Tools_Audio_ReverbOverlay::UpdateDialogs()
{
	// We're about to delete the name list, so store a pointer to the name in the
	// object itself instead
	const Asura_Audio_Reverb* const pxReverb = GetSelectedReverb();
	const char* const szSelectedName = (pxReverb ? pxReverb->m_xName.GetString() : NULL);

	// Rebuild the name list

	const u_int uNumReverbs = Asura_Audio_System::GetNumReverbs();
	const char** pszReverbNames = new const char*[ uNumReverbs ];

	Asura_Assert( pszReverbNames, "Failed to allocate DSP chain name array." );
	if( !pszReverbNames ) return;

	Asura_Audio_System::GetReverbNames( pszReverbNames );

	delete[] m_pszReverbNames;
	m_pszReverbNames = pszReverbNames;
	m_uNumReverbNames = uNumReverbs;

	m_pxReverbListDlg->ReplaceStrings( uNumReverbs, pszReverbNames );

	// Find the new index of our selection

	for( u_int u = 0; u < uNumReverbs; ++u )
	{
		const char* const szReverbName = pszReverbNames[u];
		if( Asura_StringUtil::CompareAgainstWildcardString( szReverbName, szSelectedName, false ) )
		{
			// The list dialog will automatically update following this variable
			m_iSelectedReverb = u;
			break;
		}
	}

	ResetReverbDialogs();
}

void Tools_Audio_ReverbOverlay::OnCreateReverb()
{
	Asura_Assert( s_pxThis, "No reverb overlay exists." );
	if( !s_pxThis ) return;

	GetString_Overlay xNamePicker( "New Reverb", "" );
	if( !xNamePicker.DoModal() ) return;

	const char* const szName = xNamePicker.GetString();
	const Asura_Hash_ID uNameHash = Asura_GetHashID( szName );

	if( uNameHash == ASURA_HASH_ID_UNSET ) return;

	if( Asura_Audio_System::GetReverb( uNameHash, false ) )
	{
		ShowMessageOverlay( "Reverb '%s' already exists.", szName );
		return;
	}

	bool bAlreadyExists;
	Asura_Audio_System::AddReverb( szName, bAlreadyExists );
	Asura_Assert(!bAlreadyExists, "Error creating reverb");
	Tools_Sound_System::SetReverbModified();

	s_pxThis->UpdateDialogs();
	Tools_Reverb_DropDownList_Dialog::RefreshReverbs();
}

void Tools_Audio_ReverbOverlay::OnDeleteReverb()
{
	Asura_Assert( s_pxThis, "No reverb overlay exists." );
	if( !s_pxThis ) return;

	if( static_cast< u_int >( s_pxThis->m_iSelectedReverb ) >= s_pxThis->m_uNumReverbNames ) return;

	const char* const szSelectedReverb = s_pxThis->m_pszReverbNames[ s_pxThis->m_iSelectedReverb ];
	const Asura_Hash_ID uSelectedReverb = Asura_GetHashID( szSelectedReverb );

	Asura_Audio_System::DeleteReverb( uSelectedReverb );
	Tools_Sound_System::SetReverbModified();

	s_pxThis->UpdateDialogs();
	Tools_Reverb_DropDownList_Dialog::RefreshReverbs();
}

void Tools_Audio_ReverbOverlay::OnResetReverb()
{
	Asura_Assert( s_pxThis, "No reverb overlay exists." );
	if( !s_pxThis ) return;

	if( static_cast< u_int >( s_pxThis->m_iSelectedReverb ) >= s_pxThis->m_uNumReverbNames ) return;

	const char* const szSelectedReverb = s_pxThis->m_pszReverbNames[ s_pxThis->m_iSelectedReverb ];
	const Asura_Hash_ID uSelectedReverb = Asura_GetHashID( szSelectedReverb );

	Asura_Audio_Reverb* const pxReverb = Asura_Audio_System::GetReverb( uSelectedReverb );
	if( !pxReverb ) return;

	Tools_Sound_System::ResetReverb( *pxReverb, s_pxThis->m_iSelectedPreset );
}

void Tools_Audio_ReverbOverlay::OnSelectionChange()
{
	Asura_Assert( s_pxThis, "No reverb overlay exists." );
	if( !s_pxThis ) return;

	s_pxThis->ResetReverbDialogs();
}

int Tools_Reverb_DropDownList_Dialog::s_iListValidID = 0;

Tools_Reverb_DropDownList_Dialog::Tools_Reverb_DropDownList_Dialog(
	const Asura_Vector_2& xOffset, const int iNumLines, Asura_Hash_ID* puHashID,
	bool* const pbChanged, const char* const szName /*= NULL*/ )

	: Asura_Hash_DropDownList_Dialog( xOffset )
{
	SetOnlyUpdateSelectionIndexOnClick( true );
	Initialise( xOffset, iNumLines, puHashID, pbChanged, szName );
}

u_int Tools_Reverb_DropDownList_Dialog::GetNumberOfHashValues() const
{
	return Asura_Audio_System::GetNumReverbs();
}

const char* Tools_Reverb_DropDownList_Dialog::GetNameByIndex( int iIndex ) const
{
	const Asura_Audio_Reverb* const pxReverb = Asura_Audio_System::GetReverbByIndex( iIndex );
	return pxReverb ? pxReverb->m_xName.GetString() : "*Unknown*";
}

void Tools_Reverb_DropDownList_Dialog::RefreshReverbs()
{
	// Changing this value causes all dialogs of this class to automatically regenerate their lists
	++s_iListValidID;
}

#endif // ASURA_USE_AUDIO_SYSTEM
