//
// Tools_Streaming_System.cpp
//
// Created by Richard Rice - 06/04/2006
// 
// Copyright (C) 2006 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#ifdef ASURA_USE_STREAMING
#include "Streaming/Tools_Streaming_System.h"

//asura
#include "Debugging/Asura_Assert.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Streaming/Asura_Streaming_System.h"
#include "Streaming/Asura_Streaming_Zone.h"
#include "Rendering/Asura_Module.h"

//tools
#include "Lighting/Tools_EnvironmentalLightTransport.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Adjacency.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Rendering/Tools_RenderModules.h"
#include "Streaming/Tools_StreamingModuleNames.h"



Asura_Hash_ID Tools_Streaming_System::s_uHashForThisLevel = 0;



void Tools_Streaming_System::Initialise()
{
	if(!s_uHashForThisLevel)
	{
		//just need to make sure that the zone has been setup for the main level , and we know
		//what our level hash ID is
		GetStreamingZoneForThisLevel();
	}
	//Make sure that the streaming system is set to the handlers for the main level
	//(It might not be after returning from game)
	Asura_Streaming_System::SetCurrentZone(s_uHashForThisLevel);
}

Asura_Streaming_Zone* Tools_Streaming_System::GetStreamingZoneForThisLevel()
{
	if(!s_uHashForThisLevel)
	{
		//determine name for zone
		//(Was getting filename from Tools_SaveAsura.. but unfortunately there is a possiblity
		//of this being called before tools is initialised)
		extern Asura_String EnvironmentFilename;
		Asura_String xFilename = EnvironmentFilename;
		Asura_String xName;
		xFilename.SplitFilename(0,&xName,0);
		

		s_uHashForThisLevel =  Asura_GetHashID(xName.GetString());

		//already exists?
		Asura_Streaming_Zone* pxZone = Asura_Streaming_System::FindZone(xName.GetString());
		if(!pxZone)
		{
			pxZone = Asura_Streaming_System::CreateZone(Asura_Streaming_System::ZONE_DATA, xFilename.GetString());
			Asura_Assert(pxZone,"");
		}

		//Determine a default centre position for the zone.
		Asura_Bounding_Box xEnvBox;
		xEnvBox.Reset();
		for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
		{
			Asura_Bounding_Box xModuleBox = Asura_Module_System::GetBoundingBoxOfModule(i);
			Asura_Vector_3 xPosition = Asura_Module_System::GetPositionOfModule(i);
			xModuleBox.Displace(xPosition);
			xEnvBox.Set(xModuleBox);
		}

		pxZone->SetPosition(xEnvBox.CalculateCentre());

		return pxZone;
	}
	else
	{
		Asura_Streaming_Zone* pxZone = Asura_Streaming_System::FindZone(s_uHashForThisLevel);
		Asura_Assert(pxZone,"");

		return pxZone;
	}
}

void Tools_Streaming_System::LoadConnectedZones()
{
	Asura_Streaming_Zone* pxMainZone = GetStreamingZoneForThisLevel();

	bool bLoadedAny = false;

	//load all connected zones
	for(u_int i=0;i<pxMainZone->GetNumConnection();i++)
	{
		const Asura_Streaming_Connection* pxConnection = pxMainZone->GetConnection(i);

		const char* szName = pxConnection->GetToZoneName();
		Asura_Assert(szName && *szName,"All connections should have a name associated with them");

		//create connected zone if it doesn't exist yet
		if(Asura_Streaming_System::FindZone(szName)) continue;

		const char* szFilename = pxConnection->GetToolsFilename();
		if(!szFilename || !*szFilename)
		{
			//old files may not have the filename set ,so sort it out now
			UpdateConnectionFilename(i);
			szFilename = pxConnection->GetToolsFilename();
			if(!szFilename || !*szFilename) continue;
		}
		
		

		Asura_Streaming_Zone* pxZone = Asura_Streaming_System::CreateZone(Asura_Streaming_System::ZONE_DATA,szFilename);
		Asura_Assert(pxZone,"Failed to create zone");
		Asura_Hash_ID uZoneHash = pxZone->GetHashID();

		//load file into this zone
		if(Asura_Streaming_System::LoadZone(Asura_Streaming_System::ZONE_DATA, uZoneHash,false))
		{
			Asura_Module_Handler*	pxHandler	= Asura_Module_System::GetHandler(uZoneHash);

			pxHandler->PreLoad();
			InitialiseZoneDataForCurrentZone();
			bLoadedAny = true; 
		}
		else
		{
			Asura_Streaming_System::DestroyZone(Asura_Streaming_System::ZONE_DATA, uZoneHash);
		}

		//make sure main zone is the current one again
 		Asura_Streaming_System::SetCurrentZone(pxMainZone->GetHashID());
	}
	if(bLoadedAny)
	{
		//invalidate adjacencies , so that inter-zone adjacencies for the new files get included. RAR 02/05/2006
		Tools_Adjacency::Invalidate();
	}

	//Sort out any reindexing for module linking
	Tools_StreamingModuleNames::ResyncAll();
}

void Tools_Streaming_System::UnloadZone(Asura_Hash_ID uHash)
{
	//Unloading zones doesn't work correctly without the memory banks turned on at the moment.
	//So just leave them hanging around for now. RAR 06/04/2006
#if  0
	if(uHash == s_uHashForThisLevel)
	{	
		Asura_Assert(0,"Can't unload the zone corresponding to the level being edited");
		return;
	}

	if(Asura_Streaming_System::FindZone(uHash))
	{
		Asura_Streaming_System::DestroyZone(Asura_Streaming_System::ZONE_DATA,uHash);

		//make sure main zone is the current one again
		Asura_Streaming_System::SetCurrentZone(s_uHashForThisLevel);
	}
#endif
}



void Tools_Streaming_System::CreateZoneData(Asura_Hash_ID uHash)
{
	Tools_RenderModule_System::CreateAndSetHandler(uHash);
	Tools_EnvironmentalLightTransport::CreateAndSetHandler(uHash);
	Tools_Adjacency::CreateAndSetHandler(uHash);
}

void Tools_Streaming_System::DestroyZoneData(Asura_Hash_ID uHash)
{
	Tools_RenderModule_System::DestroyHandler(uHash);
	Tools_EnvironmentalLightTransport::DestroyHandler(uHash);
	Tools_Adjacency::DestroyHandler(uHash);
}

void Tools_Streaming_System::SetZoneData(Asura_Hash_ID uHash)
{
	Tools_RenderModule_System::SetCurrentHandler(uHash);
	Tools_EnvironmentalLightTransport::SetCurrentHandler(uHash);
	Tools_Adjacency::SetCurrentHandler(uHash);
}

void Tools_Streaming_System::InitialiseZoneDataForCurrentZone()
{
	Tools_RenderModule_System::InitialiseCurrentZone();
}

void Tools_Streaming_System::UpdateConnectionFilename(int iIndex)
{
	Asura_Streaming_Zone* pxMainZone = GetStreamingZoneForThisLevel();
	Asura_Assert(pxMainZone,"");
	Asura_Assert(iIndex < (int)pxMainZone->GetNumConnection(),"");

	Asura_Streaming_Connection* pxConnection = pxMainZone->GetConnection(iIndex);
	const char* szZoneName = pxConnection->GetToZoneName();


	//Sort out old levels that don't have filenames stored with the connections.
	//This is very much Gun specific , since only Gun will have connections that
	//are old enough to be a problem. RAR 15/05/2006

	//Try Environments\ZoneName\ZoneName.asr
	{
		char szFilename[MAX_PATH];
		Asura_PrintfToBfr(szFilename,"Environments\\%s\\%s.asr",MAX_PATH,true,szZoneName,szZoneName);

		char szNetworkFilename[MAX_PATH];
		Asura_FileAccess::PlatformGetNetworkFilename(szFilename,szNetworkFilename,MAX_PATH);

		if(Tools_FileSystem::FileExists(szNetworkFilename))
		{
			pxConnection->SetToolsFilename(szFilename);
			return;
		}
	}

	//Try Environments\LODs\ZoneName\ZoneName.asr
	{
		char szFilename[MAX_PATH];
		Asura_PrintfToBfr(szFilename,"Environments\\LODs\\%s\\%s.asr",MAX_PATH,true,szZoneName,szZoneName);

		char szNetworkFilename[MAX_PATH];
		Asura_FileAccess::PlatformGetNetworkFilename(szFilename,szNetworkFilename,MAX_PATH);

		if(Tools_FileSystem::FileExists(szNetworkFilename))
		{
			pxConnection->SetToolsFilename(szFilename);
			return;
		}
	}
	
	//Try searching entire Environments directory tree
	{
		char szSearch[500];
		//search on network
		char szTemp[MAX_PATH];
		Asura_FileAccess::PlatformGetNetworkDir("Environments",szTemp,MAX_PATH);
		Asura_PrintfToBfr(szSearch,"%s\\%s.asr",sizeof(szSearch),true,szTemp,szZoneName);
		List<Asura_String> xFilenames;
		Tools_FileSystem::FindFiles(szSearch,xFilenames,TOOLS_FILEFIND_RECURSIVE | TOOLS_FILEFIND_IGNOREBACKUPDIR);
    	   	
		if(xFilenames.size())
		{
			pxConnection->SetToolsFilename(Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir(xFilenames.first_entry().GetString()));
			return;
		}
	}

	///No joy
	Asura_Assertf(0,("Failed to determine filename for zone '%s'",szZoneName));
}


//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------**
** Tools_Streaming_Iterator **
**--------------------------*/

Tools_Streaming_Iterator::Tools_Streaming_Iterator(bool bSetZone)
{
	m_bSetZone = bSetZone;
	m_bInitialZoneValid = false;
	m_pxMainZone = Tools_Streaming_System::GetStreamingZoneForThisLevel();

	Restart();
}


void Tools_Streaming_Iterator::Restart()
{
	if(m_bInitialZoneValid) End();

	m_pxInitialZone = Asura_Streaming_System::GetCurrentZone();
	m_bInitialZoneValid = true;

	//-1th index is the main zone
	m_iIndex = -1;
	m_pxCurrentZone = m_pxMainZone;
	if(m_bSetZone)
	{
		Asura_Streaming_System::SetCurrentZone(m_pxCurrentZone->GetHashID());
	}
}

void Tools_Streaming_Iterator::End()
{
	if(m_bInitialZoneValid && m_bSetZone)
	{
		//restore original zone
		Asura_Streaming_System::SetCurrentZone(m_pxInitialZone->GetHashID());
	}
	m_bInitialZoneValid = false;
	m_pxInitialZone = 0;

	m_iIndex = m_pxMainZone->GetNumConnection();

}

void Tools_Streaming_Iterator::Next()
{
	for(m_iIndex++;;m_iIndex++)
	{
		if(m_iIndex >= (int)m_pxMainZone->GetNumConnection())
		{
			//have finished
			End();
			return;
		}
		else
		{
			const Asura_Streaming_Connection* pxConnection = m_pxMainZone->GetConnection(m_iIndex);
			m_pxCurrentZone = Asura_Streaming_System::FindZone(pxConnection->GetToZoneHashID());
			if(!m_pxCurrentZone) continue;//skip past unloaded zones

			if(m_bSetZone)
			{
				Asura_Streaming_System::SetCurrentZone(m_pxCurrentZone->GetHashID());
			}
			return;
		}
	}
}

bool Tools_Streaming_Iterator::Done() const
{
	return (m_iIndex>=(int)m_pxMainZone->GetNumConnection());
}

Asura_Hash_ID Tools_Streaming_Iterator::GetHash() const
{
	Asura_Assert(m_pxCurrentZone,"");
	return m_pxCurrentZone->GetHashID();

}

#endif //ASURA_USE_STREAMING