//
// Tools_Mode_TextureTiler.h
//
// Created by Nick Brett (NDB) as a direct port from 
// the stand alone Texture_Tiler application written 
// by Niel Massey.
//
// 07/09/2007 
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef __TOOLS_MODE_TEXTURETILER_H_
#define __TOOLS_MODE_TEXTURETILER_H_

#include "Tools_Mode.h"

#include "Core/Asura_TextureManagement.h"
#include "Rendering/Asura_Viewport.h"

#include "TextureTiler/Tools_TextureTiler_Render.h"
#include "TextureTiler/Tools_TextureTiler_TTL.h"

// Texture Tiler Tools Mode
//
// Creates Tiled Textures from any number of textures. One of three different 
// algortihms is used to position the textures within the tiled texture. The
// completed Tiled textures can be exported as a pair of ausra chunks or as a
// texture file in several different formats - NDB.
class Tools_Mode_TextureTiler : public Tools_Mode
{
	enum TOOLS_TT_ALPHA_MODE
	{
		TOOLS_TT_ALPHA_OPAQUE,
		TOOLS_TT_ALPHA_TRANSPARENT_BLACK,
		TOOLS_TT_ALPHA_TRANSPARENT_MAGENTA,
	};

	enum TOOLS_TT_OPTIONS_FLAGS
	{
		TOOLS_TT_AUTOTILE,
		TOOLS_TT_COMPRESS_CHUNKS,
		TOOLS_TT_FIX_TTL_DIMENSIONS_TO_POWER_OF_TWO,
		TOOLS_TT_HAS_PADDING,
		TOOLS_TT_PREFIX_IMAGE_DATA_NAME,   // options prepends a string (Tools_TT_TiledTextureList::s_xDataNamePrefix) to the name of the texture stored in the resource chunk associated with the TTL
		TOOLS_TT_RELOAD_CHUNK_AFTER_EXPORT,
		TOOLS_TT_SHOW_REMOVE_CONFIRMATION,
		TOOLS_TT_CREATE_TEXTURE_ON_EXPORT,
		
		TOOLS_TT_NUM_OPTIONS_FLAGS,
	};

public:
	Tools_Mode_TextureTiler();
	virtual ~Tools_Mode_TextureTiler();

	virtual void Update();
	virtual void RenderWorld_PostToneMap();
	virtual void RenderUI();

	virtual void InitialiseMode();
	virtual void ShutdownMode();

	virtual bool IsModified();
	virtual void QuickSave();

	// Prepare mode for save - required by SaveAsura in order to use Tools' "is modified" system
 	virtual void PreSave();
	virtual void PostSave();

	// sets default file name which will be loaded on initialisation and offered as the default 
	// in appropriate file pickers. Default file is always the most recently saved|loaded file
	static void SetDefaultFileToLoad( const Asura_String& xDefaultFile );
	
private:
	// <-- private methods -->

	// non virtual init and shutdown methods can safely be called in ctor
	void InitialiseTTMode();
	void ShutdownTTMode();
	
	// change state of mode based on user input
	void ProcessUserInput();

	// load textures from disk and return a container of the id's of the loaded textures
	const Tools_DynamicArray<int> LoadTextureFromDisk_Implimentation();
	
	// retile TTL and update the screen positions of all subtextures
	void UpdateTTL();

	// update the BPP displayed in the texture overlay
	void UpdateBPP();

	// Update UI variables to match those in the current TTL
	void SynchronizeUIWithCurrentTTL();

	// Resize all ui components to account for a change in viewport size
	void ResizeUI();

	// modifies the menu bar (top of screen) for TT mode 
	void InitialiseMenubarForMode();
	void ShutdownMenubarForMode();

	// returns true if the selected texture index is less that the size of the TTL
	bool IsSelectedTextureValid() const;

	// attempts to load a TTL from an asr file with the name provided 
	void LoadChunkFromDisk_Implimentation( const Asura_String& xFileName,const bool bExtractTextures = false );

	// exports a chunk and then reloads it if TOOLS_TT_RELOAD_CHUNK_AFTER_EXPORT option is true 
	void Export_Implimentation();
	void Export_Implimentation( const int iPlatformType,const Asura_String& xExportPath );
	void Export_ImplimentationForPlatforms();
	void Export_ImplimentationForTools();
	
	// creates a tile hash id from each of the texture names in the current TTL and deletes the 
	// corresponding tiles from the tile manager. Used before reloading an exported chunk file.
	void DeleteTilesFromManager();
	void CreateTemporaryTiledTexture();
	void DestroyTemporaryTiledTexture();

	// write asura chunk file containing the details of the current TTL
	const bool WriteChunk( const Asura_String& xOutputFileName,const Tools_Platform_Type eTargetPlatform );
	const bool WriteMultipleChunks( const Asura_String& xOutputFileName,const Tools_Platform_Type eTargetPlatform );
	
	const Asura_String CreateExportPathForPlatform( const int iPlatformType );
	// <-- private static methods -->

	// methods must be static so that they can be passed to certain Asura methods
	static bool IsChunkNameValid() { return !s_pxPtrToSelf->m_xTTLChunkName.IsBlank(); }

	static void RemoveSelectedTextureFromTTL();
	static void RemoveSelectedTextureFromTTLWithConfirmation();
	static void RemFromTextureOverlay();	

	static void NewTTL();
	static void NewTTLWithConfirmation();
	static void LoadTTLFromDisk();
	static void LoadTTLFromDiskWithConfirmation();
	static void SaveTTL();
	static void SaveTTLAs();
	
	static void LoadTextureFromDisk();
	static void AddTextureFromDisk();
	static void SaveTextureAs();

	static void ChooseBPP();
	static void ChooseTilingAlgorithm();
	static void ChooseExportOptions();

	// create an overlay with choice of target platforms for which asr file could be created
	static void CreateExportOverlay();

	// ask user for asr save file name before opening export overlay
	static void CreateExportOverlayAs();
	static void ToggleTextureOverlay();
	static void TogglePreviewOverlay();
	static void ToggleOptionsOverlay();
	
	static bool TextureOverlayCreated();
	static bool PreviewOverlayCreated();
	static bool OptionsOverlayCreated();

	static void CloseOverlay( int iOverlayID );

	// export asr file of the type specified by the arg (as defined in Tools_TT_Platform)
	static void Export();
	static void ExportAs();

	// Attempt to load/save a tiled texture from/to an asr file
	static void SaveChunk();
	static void SaveChunkAs();
	static void LoadChunkFromDisk();
	static void LoadChunkFromDiskWithConfirmation();
	static void ReloadChunkAfterExport();

	static void CopySelectedTileHashStringToClipboard();

	// Attempt to load TTL from an asr file and extract any missing (not on disk) subtextures from the TTL image data stored in the chunk
	static void LoadChunkAndExtractTextures();

	// set all platform export flags to false
	static void ResetExportFlags();

	// open a file picker and set the chunk name from the string returned
	void SetChunkNameFromFilePicker();

	// returns the index of the subtexture in tbe current TTL which is contains the screen coordinates supplied as arguments
	const u_int FindIndexOfTTLSubtextureUnderCursor( const int iCursorX, const int iCursorY ) const;

	// Draw the current tiled texture list (TTL)
	void RenderCurrentTTL(void) const;

	// change the texture and details shown in the preview overlay to the currently selected texture
	void UpdatePreviewOverlay( const bool bForceUpdate = false );
	
	// find the position of the texture with the supplied ID in the current TTL's array
	const u_int GetTextureIndexFromID( const int iTextureID ) const;
	
	// <-- private data members -->
	Tools_DynamicArray<int> m_xLoadedTextureCollection; // collection of ID's of all the textures loaded in this session (they are destoryed on shutdown)
	Tools_TT_UIComponent    m_xCurrentTTL;              // box in which current TTL is displayed on screen

	// texture overlay
	Tools_TT_OverlayNumericalDialog m_xTextureOverlayTTLWidth;
	Tools_TT_OverlayNumericalDialog m_xTextureOverlayTTLHeight;
	Tools_TT_OverlayListDialog<Asura_List_Dialog> m_xTextureOverlayTTLSubtextureList;        // list of all textures in TTL
	Tools_TT_OverlayListDialog<Asura_DropDown_List_Dialog> m_xTextureOverlayTilingAlgorithm; // list of all tiling algs
	Tools_TT_OverlayListDialog<Asura_DropDown_List_Dialog> m_xTextureOverlayBPP;			 // list of all Bits per pixel modes

	// export overlay
	Tools_TT_OverlayListDialog<Asura_DropDown_List_Dialog> m_xExportOverlayDDSFormat;        // list of all dds compression formats availible during export
	Tools_TT_OverlayListDialog<Asura_DropDown_List_Dialog> m_xExportOverlayPSPPixelFormat;   // list of all PSP pixel formats availible during export
	Tools_TT_OverlayListDialog<Asura_DropDown_List_Dialog> m_xExportOverlayTPLAlphaBits;     // list of all TPL Alpha bit levels availible during export
	
	Asura_String m_xTTLFileName;  // file name for TTL with no file extension
	Asura_String m_xTTLChunkName; // file name for TTL with no file extension
	
	float m_fPreviewOverlayWidth;
	bool m_bTextureOverlayCreated;

	Tools_TT_ConfirmationOverlay m_xConfirmationOverlay;
	
	Asura_Handle m_xMenuHandle;
	
	// width and height of viewport the last time ResizeUI() was called
	int m_iLastKnownViewportWidth;
	int m_iLastKnownViewportHeight;

	Tools_TT_TiledTextureList m_xTTL; // the current TTL

	u_int m_uSelectedTextureIndex;      // position of selected texture in thumbnail list
	u_int m_uTTLSelectedTextureIndex;
	u_int m_uFirstThumbnailDisplayed;
	u_int m_uNumberOfThumbnailsDisplayed;
	u_int m_uOverlayColour;  // background colour
	u_int m_uUIBorderColour;

	TOOLS_TT_ALPHA_MODE m_eAlphaMode;
	TOOLS_TT_ALPHA_MODE m_ePreviewDisplayedAlphaMode;

	Asura_String m_xTemporaryTextureName;
	bool		 m_bTemporaryTiledTexturesExist;

	// <-- private static data members -->
	
	// require a static pointer to the only instance of this class for use with static functions such as SaveTTL()
	static Tools_Mode_TextureTiler* s_pxPtrToSelf; 

	// Mode will attempt to load a file with this name at initialisation
	static Asura_String s_xDefaultFileToLoad;

	static bool s_abOptionFlag[TOOLS_TT_NUM_OPTIONS_FLAGS];
	static bool	s_abTextureExportFlag[Tools_TT_FileType::TOOLS_TT_FILE_NAME_EXTENSION_NUM];
	static bool s_abFileExtensionFilterFlag[Tools_TT_FileType::TOOLS_TT_FILE_NAME_EXTENSION_NUM];
	
	static Tools_TT_OverlayStatus s_xOptionsOverlayStatus;
	static Tools_TT_OverlayStatus s_xTextureOverlayStatus;
	static Tools_TT_OverlayStatus s_xPreviewOverlayStatus;
	static Tools_TT_OverlayStatus s_xExportOverlayStatus;

	static u_int s_uPreviewDisplayedTextureIndex;
	static int	 s_iTextureFlags;

	static bool s_abExportPlatformFlags[TOOLS_PLATFORM_NUM_PLATFORMS];
};

void Tools_TT_CopyToClipboard( const Asura_String& xString );


#endif // __TOOLS_MODE_TEXTURETILER_H_

