
#include "Core/Asura.h"

#include "TextureTiler/Tools_TextureTiler_TTL.h"

// Core
#include "ChunkSystem/Asura_Chunk_FileInfo.h"
#include "ChunkSystem/Asura_Chunk_Loaders.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_Chunk_TiledTexture.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_TiledTextureManagement.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Load&Save/Tools_AsuraFileContents.h"
#include "Xbox360/ChunkSystem/Asura_Chunk_ResourceFile_Xbox360.h"

// Tools
#include "Misc/Tools_DynamicArray.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Texturing/Tools_Texture_Management.h"

const char* Tools_TT_ListToken::s_aszTools_TT_List_Token[TOOLS_TT_LIST_TOKEN_NUM_TOKENS] =
{
	"NULL",				// TOOLS_TT_LIST_TOKEN_NULL = 0,
	"HEIGHT",			// TOOLS_TT_LIST_TOKEN_HEIGHT,
	"WIDTH",			// TOOLS_TT_LIST_TOKEN_WIDTH,
	"BPP",				// TOOLS_TT_LIST_TOKEN_BPP,
	"TEXTURE",			// TOOLS_TT_LIST_TOKEN_TEXTURE,
	"DDS_FORMAT",		// TOOLS_TT_LIST_TOKEN_DDSFORMAT,
	"PSP_PIXEL_FORMAT",	// TOOLS_TT_LIST_TOKEN_PSP_PIXEL_FORMAT,
	"TPL_ALPHA_BITS",	// TOOLS_TT_LIST_TOKEN_TPL_ALPHA_BITS,
	"PADDING",			// TOOLS_TT_LIST_TOKEN_PADDING,
	"ALGORITHM",        // TOOLS_TT_LIST_TOKEN_ALGORITHM,
};

// <-- class Tiled Texture List -->
const Asura_String Tools_TT_TiledTextureList::s_xDataNamePrefix = "tile_";

Tools_TT_TiledTextureList::Tools_TT_TiledTextureList( const u_int uReserve /* = 20 */ )
:m_xTextureInfoCollection()
,m_bHasPadding(false)
,m_bPrefixDataName(false)
,m_bModified(false)
,m_iWidth(256)
,m_iHeight(256)
,m_iBytesPP(4)
,m_eTilingAlgorithm(TOOLS_TT_ALGORITHM_CLASSIC)
,m_eDDSFormat( Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_UNCOMPRESSED )
,m_eTPLAlphaBits( Tools_TT_TPLAlphaBits::TOOLS_TT_EXPORT_TPL_ALPHA_BITS_8 )
,m_iPSPPixelFormat( Tools_TT_PSPPixelFormat::GetValue(1) )
,m_iTextureFlags(0)
,m_uTiledTextureCount(0)
{}

Tools_TT_TiledTextureList::Tools_TT_TiledTextureList( const Tools_TT_TiledTextureList& xOther )
{
	*this = xOther;
}

Tools_TT_TiledTextureList::~Tools_TT_TiledTextureList()
{}

const Tools_TT_TiledTextureList& Tools_TT_TiledTextureList::operator=( const Tools_TT_TiledTextureList& xOther )
{
	m_xTextureInfoCollection    = xOther.m_xTextureInfoCollection;

	m_bHasPadding               = xOther.HasPadding();
	m_bPrefixDataName			= xOther.HasDataNamePrefix();
	m_bModified 	  			= xOther.IsModified();
	m_iWidth		  			= xOther.GetWidth(); 
	m_iHeight         			= xOther.GetHeight(); 
	m_iBytesPP      	  		= xOther.GetBytesPP();
	m_eTilingAlgorithm			= xOther.GetTilingAlgorithm();
	m_eDDSFormat				= xOther.GetDDSFormat();
	m_iPSPPixelFormat			= xOther.GetPSPPixelFormat();
	m_eTPLAlphaBits				= xOther.GetTPLAlphaBits();
	m_iTextureFlags				= xOther.GetTextureFlags();
	m_uTiledTextureCount		= xOther.m_uTiledTextureCount;

	return *this;
}

// add a texture to the collection
const bool Tools_TT_TiledTextureList::Insert( const int iTextureID )
{
	// ensure textures in collection are unique
	if( ContainsTexture(iTextureID) )
	{
		return false;
	}
	
	Tools_TT_Tile xTile = Tools_TT_Tile(iTextureID);
	if(xTile.IsValid())
	{
		m_xTextureInfoCollection += xTile;
	}
	else
	{
		return false;
	}

	m_bModified = true;

	return true;
}

const bool Tools_TT_TiledTextureList::Insert( const Asura_String& xTextureFileName )
{
	bool bShouldLoadTexture(false);
	const int iTextureID( Asura_Texture_Management::AddTexture( xTextureFileName.GetString(),&bShouldLoadTexture ) );

	if( iTextureID == ASURA_TEXTURE_NONE )
	{
		return false;
	}

	if( bShouldLoadTexture )
	{
		Asura_Texture_Management::LoadTextures();
	}

	return Insert( iTextureID );
}


const bool Tools_TT_TiledTextureList::Insert( const int* piTextureIDCollection, const int iSize )
{
	bool bSuccess(true);

	for( int i(0);i<iSize;++i )
	{
		bSuccess &= Insert( piTextureIDCollection[i] );
	}

	return bSuccess;
}

const bool Tools_TT_TiledTextureList::Insert( const Tools_TT_Tile& xTile )
{
	bool bSuccess( false );

	if( xTile.IsValid() )
	{
		m_xTextureInfoCollection += xTile;
	
		if( xTile.IsTiled() )
		{
			++m_uTiledTextureCount;
		}

		m_bModified = true;
		bSuccess    = true;
	}

	return bSuccess;
}

const bool Tools_TT_TiledTextureList::Erase( const int iTextureID )
{
	// find index of target in collection
	u_int uTargetIndex(0);
	bool  bTargetNotFound(true);
	
	for(;(bTargetNotFound) && (uTargetIndex<Size());++uTargetIndex )
	{
		if( m_xTextureInfoCollection[uTargetIndex].GetTextureID() == iTextureID )
		{
			bTargetNotFound = false;
			break;
		}
	}

	// if texture is found then remove it from collection
	if( !bTargetNotFound )
	{
		m_bModified = true;
		
		// if this texture has been tiled then decrement tiled texture counter
		if( m_uTiledTextureCount && (m_xTextureInfoCollection[uTargetIndex].IsTiled()) )
		{
			--m_uTiledTextureCount;
		}

		m_xTextureInfoCollection.Remove( uTargetIndex,1 );
	}

	return !bTargetNotFound;
}

const bool Tools_TT_TiledTextureList::ContainsTexture( const int iTextureID ) const
{
	u_int u(0);

	for(;u<Size();++u )
	{
		if( m_xTextureInfoCollection[u].GetTextureID() == iTextureID )
			break;
	}

	return u<Size();
}

const Tools_TT_Tile* const Tools_TT_TiledTextureList::GetTextureFromID( const int iTextureID ) const
{
	const u_int uIndex( GetTextureIndexFromID(iTextureID) );
	
	if( uIndex == ASURA_TEXTURE_NONE )
	{
		return NULL;
	}

	return m_xTextureInfoCollection + uIndex;
}

const u_int Tools_TT_TiledTextureList::GetTextureIndexFromID( const int iTextureID ) const
{
	u_int u(0);

	for(;u<Size();++u )
	{
		if( m_xTextureInfoCollection[u].GetTextureID() == iTextureID )
			break;
	}

	return (u<Size())? u:ASURA_TEXTURE_NONE;
}

const u_int Tools_TT_TiledTextureList::GetTextureIndexFromName( const char* szTextureName ) const
{
	return GetTextureIndexFromID( Asura_Texture_Management::GetTextureIndex( szTextureName ) );
}

// rearange textures within TTL using the specified algorithm
const int Tools_TT_TiledTextureList::Regenerate()
{
	m_uTiledTextureCount = 0;

	// don't do anything if there are no textures to tile
	if( Size() )
	{
		m_bModified = true;

		UnsetAllTiledStatus();

		if( HasPadding() )
		{
			//increase the width of	all	the	textures to	leave room for the padding
			for( u_int u(0);u<Size();++u )
			{
				this->operator[](u).SetWidth( this->operator[](u).GetWidth() + 2 );
				this->operator[](u).SetHeight( this->operator[](u).GetHeight() + 2 );
			}
		}

		m_uTiledTextureCount = Tools_TT_DoTiling( m_xTextureInfoCollection.GetArray(),Size(),GetDimension(), m_eTilingAlgorithm );
		
		if( HasPadding() )
		{
			// correct the texture size and move textures to center them within the padded area
			for( u_int u(0); u<Size();++u )
			{
				this->operator[](u).SetWidth( this->operator[](u).GetWidth() - 2 );
				this->operator[](u).SetHeight( this->operator[](u).GetHeight() - 2 );
				this->operator[](u).SetPosition( Tools_TT_Vector_2i(this->operator[](u).GetPosition().x() + 1,this->operator[](u).GetPosition().y() + 1) );
			}
		}

		Tools_TT_UpdateTTLUVCoordinates( *this );
	}

	return m_uTiledTextureCount;
}

void Tools_TT_TiledTextureList::UnsetAllTiledStatus()
{
	m_bModified = true;

	for( u_int u(0);u<Size();++u )
	{
		m_xTextureInfoCollection[u].SetTiled( false );
	}
}
const bool Tools_TT_TiledTextureList::ReadFile( const Asura_String& xFileName )
{
	// remove all elements from TTL
	Clear();

	FILE* pxFile( fopen(xFileName.GetString(),"r") );

	if(!pxFile)
	{
		return false;
	}

	int iWidth(0);
	int iHeight(0);
	int iBytesPP(0);                 // stored in file as bits per pixel, but stord in TTL object as bytes per pixel
	int iDDSFormat(0);
	int iPSPPixelFormat(0);
	int iTPLAlphaBits(0);
	int iAlgorithm(0);
	bool bPadding(false);

	char* pcBufferPtr( NULL );
	bool  bNewTexture( false );
	
	const u_int uBufferSize(256);
	char* pcBuffer( new char[uBufferSize] );

	Tools_DynamicArray<int> xTextureID; // list of all texture ids returned by AddTexture()

	while( !feof(pxFile) )
	{
		fgets( pcBuffer,uBufferSize,pxFile );
		
		// load textures if required
		if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_TEXTURE] ) )
		{
			// set trailing end of line character to NULL (effectively removing it from string)
			pcBufferPtr = strrchr(pcBuffer,'\n');
			if( pcBufferPtr )
			{
				*pcBufferPtr = NULL;
			}

			// find start of texture name
			pcBufferPtr = strstr( pcBuffer," = " );
			pcBufferPtr += 3;

			// strip any path upto and including \\graphics
			char* pcPathEnd( strstr( pcBufferPtr,"\\graphics" ) );

			// try searching for capitalised version (should still work on case sensitive file systems - NDB)
			if( pcPathEnd == NULL )
			{
				pcPathEnd = strstr( pcBufferPtr,"\\Graphics" );
			}

			if( pcPathEnd )
			{
				pcBufferPtr = pcPathEnd+strlen("\\graphics");
			}

			// if a texture name has been found then add it to texture management and record its ID
			if( pcBufferPtr )
			{
				xTextureID += Asura_Texture_Management::AddTexture( pcBufferPtr,&bNewTexture );
			}
			else
			{
				Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red," ERROR could not load texture %s required by TTL %s",pcBuffer,xFileName.GetString() );
			}
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_WIDTH] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iWidth );
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_HEIGHT] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iHeight );
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_BPP] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iBytesPP );
			iBytesPP /= 8;
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_PADDING] ) )
		{
			// must cope with padding stored as a string (True|False) or as an int (0|>0)

			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;

			if( (*pcBufferPtr == 'F') | (*pcBufferPtr == 'f') | (*pcBufferPtr == '0') )
			{
				bPadding = false;
			}
			else if( (*pcBufferPtr == 'T') | (*pcBufferPtr == 't') )
			{
				bPadding = true;
			}
			else
			{
				const int i(*pcBufferPtr - '0');
				if( (i>0) && (i<10) ) {bPadding = true;}
			}
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_DDSFORMAT] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iDDSFormat );
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_PSP_PIXEL_FORMAT] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iPSPPixelFormat );
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_TPL_ALPHA_BITS] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iTPLAlphaBits );
		}
		else if( strstr( pcBuffer,Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_ALGORITHM] ) )
		{
			pcBufferPtr = strstr( pcBuffer," = " );	
			pcBufferPtr += 3;
			Asura_StringUtil::GetIntValue( pcBufferPtr,&iAlgorithm );
		}
	}

	delete [] pcBuffer;

	// set parameters read from file
	SetWidth( iWidth );
	SetHeight( iHeight );
	SetPadding( bPadding );

	if( (iBytesPP>0) && (iBytesPP<5) )
	{
		SetBytesPP( iBytesPP );
	}
	else
	{
		SetBytesPP( 4 );
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red, "WARNING %s NOT recognised, set to default",Tools_TT_ListToken::GetName(Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_BPP) );
	}

	if( static_cast<Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT>(iDDSFormat) < Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT_NUM )
	{
		SetDDSFormat( static_cast<Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_FORMAT>(iDDSFormat) );
	}
	else
	{
		SetDDSFormat( Tools_TT_DDSCompressionFormat::TOOLS_TT_EXPORT_DDS_COMPRESSION_UNCOMPRESSED );
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red, "WARNING %s NOT recognised, set to default",Tools_TT_ListToken::GetName(Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_DDSFORMAT) );
	}
	
	if( (iPSPPixelFormat == Tools_TT_PSPPixelFormat::GetValue(0)) || (iPSPPixelFormat == Tools_TT_PSPPixelFormat::GetValue(1)) )
	{
		SetPSPPixelFormat( Tools_TT_PSPPixelFormat::GetValue(iPSPPixelFormat) );		
	}
	else
	{
		SetPSPPixelFormat( Tools_TT_PSPPixelFormat::GetValue(Tools_TT_PSPPixelFormat::TOOLS_TT_EXPORT_PSP_PIXEL_8) );
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red, "WARNING %s NOT recognised, set to default",Tools_TT_ListToken::GetName(Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_PSP_PIXEL_FORMAT) );
	}

	if( static_cast<Tools_TT_TPLAlphaBits::TOOLS_TT_EXPORT_TPL_ALPHA_BITS>(iTPLAlphaBits) < Tools_TT_TPLAlphaBits::TOOLS_TT_EXPORT_TPL_ALPHA_BITS_NUM )
	{
		SetTPLAlphaBits( static_cast<Tools_TT_TPLAlphaBits::TOOLS_TT_EXPORT_TPL_ALPHA_BITS>(iTPLAlphaBits) );	
	}
	else
	{
		SetTPLAlphaBits( static_cast<Tools_TT_TPLAlphaBits::TOOLS_TT_EXPORT_TPL_ALPHA_BITS>(iTPLAlphaBits) );
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red, "WARNING %s NOT recognised, set to default",Tools_TT_ListToken::GetName(Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_TPL_ALPHA_BITS) );
	}

	if( static_cast<TOOLS_TT_ALGORITHM_TYPE>(iAlgorithm) < TOOLS_TT_NUM_ALGORITHM )
	{
		SetTilingAlgorithm( static_cast<TOOLS_TT_ALGORITHM_TYPE>(iAlgorithm) );	
	}
	else
	{
		SetTilingAlgorithm( static_cast<TOOLS_TT_ALGORITHM_TYPE>(iAlgorithm) );
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red, "WARNING %s NOT recognised, set to default",Tools_TT_ListToken::GetName(Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_ALGORITHM) );
	}

	// load requested textures
	Asura_Texture_Management::LoadTextures();

	// add textures to TTL
	for( u_int u(0);u<xTextureID.GetSize();++u )
	{
		Insert( xTextureID[u] );
	}

	// run tiling algorithm
	Regenerate();

	Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::White,"loaded %s",xFileName.GetString() );

	return true;
}

const bool Tools_TT_TiledTextureList::WriteFile( const Asura_String& xFileName )
{
	FILE *pFile = fopen( xFileName.GetString(), "w" );

	if(!pFile)
		return false;

	// Write a comment header first
	fprintf(pFile, "; Texture Tiler file list\n");

	// Version number
	fprintf(pFile, "; Version 2\n");
	// Next write the target size details
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_WIDTH],static_cast<int>(GetWidth()) );	
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_HEIGHT],static_cast<int>(GetHeight()) );
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_BPP],GetBytesPP()*8 );
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_PADDING],(HasPadding())?1:0 );
			
	// DDS format
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_DDSFORMAT],static_cast<int>(GetDDSFormat()) );
	
	// PSP format
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_PSP_PIXEL_FORMAT], static_cast<int>(GetPSPPixelFormat()) );
	
	// TPL alpha bits
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_TPL_ALPHA_BITS], static_cast<int>(GetTPLAlphaBits()) );
	
	// Tiling algorithm
	fprintf( pFile,"%s = %i\n",Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_ALGORITHM], static_cast<int>(GetTilingAlgorithm()) );

	// Now write the file names	
	for( u_int u(0);u<Size();++u )
	{
		fprintf(pFile, "%s = %s\n", Tools_TT_ListToken::s_aszTools_TT_List_Token[Tools_TT_ListToken::TOOLS_TT_LIST_TOKEN_TEXTURE],this->operator[](u).GetName() );
	}

	fclose(pFile);

	m_bModified = false;
	
	return true;
}

const TOOLS_TT_ERROR_CODE Tools_TT_TiledTextureList::ReadChunk( const Asura_String& xInputFileName,const bool bExtractMissingTextures /* = false */ )
{
	TOOLS_TT_ERROR_CODE eError(TOOLS_TT_FAIL);

	// check that this is indeed a TTL chunk
	if( Tools_AsuraFileContents::GetAsuraFileType( xInputFileName.GetString() ) != AsuraFile_TextureTiler )
	{
		return eError;
	}

	Asura_Handle xFileHandle = Asura_FileAccess::Open( xInputFileName.GetString() );
    if( xFileHandle == ASURA_INVALID_HANDLE )
    {
        return eError;
    }

	if( !Asura_FileAccess::ProcessAsuraFileHeader(xFileHandle) )
	{
		return eError;
	}
	
	// collection of tile names
	Tools_DynamicArray< Asura_String > xTextureNameCollection;

	// allocate memory to hold chunk headers read from file
	char* pcChunkBuffer( new char[512] );
	Asura_Chunk_Header*	pxChunkHeader;

	// instanciate local variables
	bool bTiledTextureChunkRead(false);
	bool bResourceChunkRead(false);

	Asura_Chunk_File_Stream xStream( xFileHandle );

	while( xStream.ReadChunk_GetNextID(pcChunkBuffer) )
	{
		pxChunkHeader = reinterpret_cast<Asura_Chunk_Header*>( pcChunkBuffer );
		
		// if this is a TiledTexture chunk read it in
		if( (pxChunkHeader->ID == ASURA_CHUNK_TILEDTEXTURE) && !bTiledTextureChunkRead )
		{
			// read rest of tiled texture chunk header
			xStream.ReadChunk_GetNextHeader( pcChunkBuffer,sizeof(Asura_Chunk_TiledTexture));

			// recast to tiled texture chunk header
			Asura_Chunk_TiledTexture* pxTiledTextureHeader( reinterpret_cast<Asura_Chunk_TiledTexture*>( pcChunkBuffer ) );
			
			Asura_Assertf( pxTiledTextureHeader->m_iNTiles, ( "Tiled texture in file '%s' does not contain any tiles", xInputFileName.GetString() ) );

			if( pxTiledTextureHeader->m_iNTiles )
			{
				// allocate memory to hold names of texture files used to produce each file (these are different from the 
				// NameOfTextureUsed returned by a tile after it has been read into the tile manager, instead they are the 
				// strings stored in the chunk file) - NDB

				xTextureNameCollection.SetSize( pxTiledTextureHeader->m_iNTiles );

				Asura_String* pxTextureNameBegin( xTextureNameCollection.GetArray() );
				Asura_String* pxTextureNameEnd(   xTextureNameCollection.GetArray() + xTextureNameCollection.GetSize() );

				// read contents of chunk into tiled texture manager
				pxTiledTextureHeader->Process( xStream, pxTextureNameBegin, pxTextureNameEnd );

				bTiledTextureChunkRead = true;
			}

		}
		else if( pxChunkHeader->ID == ASURA_CHUNK_RESOURCEFILE )
		{
			// read rest of resource chunk header
			xStream.ReadChunk_GetNextHeader( pcChunkBuffer,sizeof(Asura_Chunk_ResourceFile));

			// recast to resource chunk header
			Asura_Chunk_ResourceFile* pxResourceHeader( reinterpret_cast<Asura_Chunk_ResourceFile*>( pcChunkBuffer ) );
			
			pxResourceHeader->Process( xStream.GetFileHandle() );

			bResourceChunkRead = true;
		}
		else
		{
			// read rest of chunk header
			xStream.ReadChunk_GetNextHeader( pcChunkBuffer,sizeof(Asura_Chunk_Header) );
		}
		
		xStream.ReadChunk_End( bResourceChunkRead|bTiledTextureChunkRead );

		// if a tiled texture chunk and a resource chunk have been read in then break
		if( bResourceChunkRead && bTiledTextureChunkRead )
		{
			break;
		}

	}
	
	delete [] pcChunkBuffer;

	Asura_FileAccess::Close( xFileHandle );
		
	if( bTiledTextureChunkRead  )
	{
		// required to fix up texture ID's on old v1 TT files
		Asura_TiledTextureManagement::SetTextureHandlesFromHash();

		// allocate storage for tiles to be read back out of tile manager
		Tools_DynamicArray< Asura_TiledTextureInfo > xTextureInfoCollection;
		xTextureInfoCollection.SetSize( xTextureNameCollection.GetSize() );
		Asura_TiledTextureInfo* pxTextureBegin( xTextureInfoCollection.GetArray() );

		// read tiles out of tile manager and replace the name of the parent texture with the names of the original textures used to produce the tiles		
		Asura_TiledTextureInfo* pxTextureItr( xTextureInfoCollection.GetArray() );
		Asura_TiledTextureInfo* pxTextureEnd( xTextureInfoCollection.GetArray() + xTextureInfoCollection.GetSize() );
		
		Asura_String* pxTextureNameItr( xTextureNameCollection.GetArray() );
		Asura_String* pxTextureNameEnd( xTextureNameCollection.GetArray() + xTextureNameCollection.GetSize() );
		
		const Asura_TiledTextureInfo* pxTileInfo(NULL);
		
		// strip preceeding path
		const char* pszHashString( Tools_FileSystem::StripPath( pxTextureNameItr->GetString() ) );

		if( pszHashString )
		{
			pxTileInfo = Asura_TiledTextureManagement::GetTileInfoByHash( Asura_GetHashID( pszHashString ) );
		}

		if( pxTileInfo )
		{
			// save a copy of the parent texture name
			const Asura_String xParentTextureName( pxTileInfo->m_xNameOfTextureUsed );

			// change the name of texture used in each tile info from the name of the parent texture back to the name of the original texture used to create the tile
			for( ;(pxTextureNameItr != pxTextureNameEnd) && (pxTextureItr != pxTextureEnd); ++pxTextureNameItr, ++pxTextureItr )
			{
				// strip path
				pszHashString = Tools_FileSystem::StripPath( pxTextureNameItr->GetString() );

				pxTileInfo = Asura_TiledTextureManagement::GetTileInfoByHash( Asura_GetHashID( pszHashString ) );
				pxTextureItr->m_xNameOfTextureUsed.MakeBlank();

				if( pxTileInfo )
				{
					*pxTextureItr = *pxTileInfo;
					pxTextureItr->m_xNameOfTextureUsed = *pxTextureNameItr;
				}
			}

			if(bExtractMissingTextures)
			{
				Asura_TiledTextureInfo* const pxNonExistantFileBegin( Tools_TT_ShiftNonExistantFilesToEnd( pxTextureBegin,pxTextureEnd ) );
				Tools_TT_ExtractTexturesFromResourceChunk( xInputFileName, xParentTextureName, pxNonExistantFileBegin, pxTextureEnd );
			}

			// delete current contents of *this
			Clear();

			// assign tiles to *this
			eError = ReadTiledTextureInfoArray( pxTextureBegin,pxTextureEnd );
		}
		else
		{
			Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR failed to retrive first tile from Tiled Texture chunk" );
		}
	}

	SetModified(false);

	return eError;
}

const bool Tools_TT_TiledTextureList::WriteChunk( const Asura_String& xOutputFileName,const Tools_Platform_Type eTargetPlatform, const bool bCompressChunk /* = false */ )
{
	return WriteChunk( xOutputFileName,xOutputFileName,eTargetPlatform,bCompressChunk );
}

const bool Tools_TT_TiledTextureList::WriteChunk( const Asura_String& xOutputFileName,const Asura_String& xResourceChunkName,const Tools_Platform_Type eTargetPlatform, const bool bCompressChunk /* = false */ )
{
	// instanciate chunk stream
	Asura_SaveBlock				 xSaveBlock;
	Asura_Chunk_SaveBlock_Stream xStream(&xSaveBlock);

	// Set the endian-ness of the stream...
	xStream.SetSwapEndian( Tools_Platform::IsBigEndian( eTargetPlatform ) );

	// write chunk to stream
	bool bSuccess( WriteChunk( xStream,xResourceChunkName,eTargetPlatform,bCompressChunk ) );
	
	if( !bSuccess )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"\"%s\" chunk write FAILED",xOutputFileName.GetString());
		return bSuccess;
	}

	xSaveBlock.AddAsuraTerminator();
	
	// save stream to disk
	bSuccess = xSaveBlock.SaveFile( xOutputFileName.GetString(),bCompressChunk );
	
	if( !bSuccess )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"\"%s\" chunk save FAILED",xOutputFileName.GetString());
		return bSuccess;
	}

	Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::White,"\"%s\" chunk saved sucessfully",xOutputFileName.GetString());
	
	return bSuccess;
}

const bool Tools_TT_TiledTextureList::WriteChunk( Asura_Chunk_Stream& xStream,const Asura_String& xResourceChunkName,const Tools_Platform_Type eTargetPlatform,const bool bCompressChunk /* = false */ )
{
	bool bSuccess( false );

	if( xResourceChunkName.IsNull() || xResourceChunkName.IsBlank() )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR chunk output filename is NULL or blank" );	
	}
	else
	{
		// Create an array of the tiled texture info
		Asura_TiledTextureInfo* pxTTLInfo( new Asura_TiledTextureInfo[GetTiledTextureCount()] );
		
		for( u_int u(0);u<Size();++u ) 
		{
			if( operator[](u).IsTiled() )
			{
				pxTTLInfo[u].m_fu1 = operator[](u).GetUVRectangle().MinX;
				pxTTLInfo[u].m_fv1 = operator[](u).GetUVRectangle().MinY; 
				pxTTLInfo[u].m_fu2 = operator[](u).GetUVRectangle().MaxX; 
				pxTTLInfo[u].m_fv2 = operator[](u).GetUVRectangle().MaxY; 
				pxTTLInfo[u].m_TiledToWidth  = static_cast<float>(GetWidth());
				pxTTLInfo[u].m_TiledToHeight = static_cast<float>(GetHeight());
				pxTTLInfo[u].m_xNameOfTextureUsed = operator[](u).GetName().GetString();
			}
		}

		// create temporary texture file
		Tools_TT_FileType::TOOLS_TT_FILE_NAME_EXTENSION eFileType( Tools_TT_FileType::GetTextureFileTypeForPlatform( eTargetPlatform ) );

		Asura_String xFileName;
		xResourceChunkName.SplitFilename( NULL, &xFileName, NULL );
		Asura_String xTemporaryTextureName = xFileName + Tools_TT_FileType::GetName( eFileType );

		if( GetBytesPP() < 4 )
		{
			Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR 8 bit export not yet implimented, forcing 32 bit export" );
		}

		// currently doesn't support any type other than R8G8B8A8 for export so force 4 Bytes PP - NDB 250408
		Tools_TT_Export_RawPixelData xRawPixelData( static_cast<u_int>(GetWidth()),static_cast<u_int>(GetHeight()),4 ); 
		GetRawPixelData( xRawPixelData );

		if( eTargetPlatform == TOOLS_PLATFORM_PSP )
		{
			Tools_TT_Export_SavePSPTx( xRawPixelData,xTemporaryTextureName );
		}
		else if( eTargetPlatform == TOOLS_PLATFORM_PS3 )
		{
			Tools_TT_Export_SavePS3Tx( xRawPixelData,xTemporaryTextureName );
		}
		else if( eFileType == Tools_TT_FileType::TOOLS_TT_DDS )
		{
			Tools_TT_Export_SaveDDS( xRawPixelData,xTemporaryTextureName,GetDDSFormat() );
		}
		else if( eFileType == Tools_TT_FileType::TOOLS_TT_XPR )
		{
			Tools_TT_Export_SaveXPR( xRawPixelData,xTemporaryTextureName,GetDDSFormat() );
		}
		else if( eFileType == Tools_TT_FileType::TOOLS_TT_TPL )
		{
			Tools_TT_Export_SaveTPL( xRawPixelData,xTemporaryTextureName,GetTPLAlphaBits() );
		}
		else
		{
			Tools_TT_Export_SaveTextureAs( xRawPixelData,xTemporaryTextureName );	
		}

		// read temporary texture file into buffer
		HANDLE xTextureFileHandle = CreateFile( xTemporaryTextureName.GetString(),GENERIC_READ,FILE_SHARE_READ,0,OPEN_EXISTING,FILE_ATTRIBUTE_NORMAL,0 );
		
		Asura_Assertf( xTextureFileHandle != reinterpret_cast<HANDLE>(-1),( "Failed to create texture file \"%s\"", xTemporaryTextureName.GetString() ) );
		
		if( xTextureFileHandle != reinterpret_cast<HANDLE>(-1) )
		{
			const int iTrueFileSize( GetFileSize(xTextureFileHandle,0) );
			const int iTextureBufferSize( ((iTrueFileSize+3)/4)*4 );
			char* pcTextureBuffer( new char[iTextureBufferSize] );
			unsigned long bytesRead(0);

			// ReadFile member method maskes global so must sepcify global scope here
			::ReadFile( xTextureFileHandle,pcTextureBuffer,iTrueFileSize,&bytesRead,0 );

			// close and delete temporary texture file
			CloseHandle( xTextureFileHandle );
			DeleteFile( xTemporaryTextureName.GetString() );
			Asura_String xTextureFileName;
			Tools_TT_CreateResourceChunkName( xTextureFileName,xResourceChunkName,eTargetPlatform,HasDataNamePrefix() );

			// <-- write TTL info to chunk -->

			// include offset in u/v positions if exporting for any platform other than PSP
			const u_int uTTLInfoSize( GetTiledTextureCount() );
			
			if( eTargetPlatform != TOOLS_PLATFORM_PSP )
			{
				const float	fOffset( 0.5f );
				
				for( u_int u(0);u<uTTLInfoSize;++u ) 
				{
					pxTTLInfo[u].m_fu1 += fOffset / pxTTLInfo[u].m_TiledToWidth;
					pxTTLInfo[u].m_fv1 += fOffset / pxTTLInfo[u].m_TiledToHeight;
					pxTTLInfo[u].m_fu2 += fOffset / pxTTLInfo[u].m_TiledToWidth;
					pxTTLInfo[u].m_fv2 += fOffset / pxTTLInfo[u].m_TiledToHeight;
				}
			}

			// prefix name of resource chunk with "graphics\\" string as this is required by LoadTextureFromResouceChunk(), which will remove the prefix ( by chopping off the first 9 characters in the string ) without actually checking wheather or not its there
			Asura_String xResourceName( "graphics\\" );
			xResourceName += xTextureFileName;

			// write resource chunk
			Asura_Assert( iTextureBufferSize > 0, "Texture buffer size is zero" );

			if( iTextureBufferSize > 0 )
			{
				ASURA_RESOURCEFILE_TYPE_ID eResourceType( ASURA_RESOURCEFILE_TYPE_TEXTURE );
				int iSubType(0);

				// Write the texture data, along with header info
				xStream.WriteChunk_Start( ASURA_CHUNK_RESOURCEFILE,0,0 );

				// xStream << eTypeEnum;
				xStream << Asura_WriteToStream_Enum_Wrapper<unsigned int>( eResourceType );
				xStream << iSubType;
				xStream << iTextureBufferSize;
				xStream << xResourceName;

				xStream.Write( iTextureBufferSize,pcTextureBuffer );

				xStream.WriteChunk_End();
			
			
				// write TTL data chunk 
				Asura_Chunk_TiledTexture::Create( xTextureFileName,uTTLInfoSize,pxTTLInfo,xStream );
			
				bSuccess = true;
			}

			// tidy up
			delete [] pcTextureBuffer;
			delete [] pxTTLInfo;
		}
		
	}

	return bSuccess;	
}


const bool Tools_TT_TiledTextureList::WriteMultipleChunks( const Asura_String& xOutputFileName,
														   const Tools_Platform_Type eTargetPlatform,const bool bCompressChunk /* = false */,
														   Tools_DynamicArray<Asura_String>* pxOutput /* = NULL */ )
{
	// instanciate chunk stream
	Asura_SaveBlock				 xSaveBlock;
	Asura_Chunk_SaveBlock_Stream xStream(&xSaveBlock);

	// Set the endian-ness of the stream...
	xStream.SetSwapEndian( Tools_Platform::IsBigEndian( eTargetPlatform ) );

	// create copy of tiled texture	
	Tools_TT_TiledTextureList xCopy( *this );

	// write chunk to stream
	xCopy.WriteMultipleChunks( xStream,xOutputFileName,eTargetPlatform,bCompressChunk,pxOutput );
	
	bool bSuccess( xSaveBlock.AddAsuraTerminator() );
	
	// save stream to disk
	bSuccess = xSaveBlock.SaveFile( xOutputFileName.GetString(),bCompressChunk );
	
	if( !bSuccess )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"\"%s\" chunk save FAILED",xOutputFileName.GetString());
		return bSuccess;
	}

	Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::White,"\"%s\" chunk saved sucessfully",xOutputFileName.GetString());
	
	return bSuccess;
}

const bool Tools_TT_TiledTextureList::WriteMultipleChunks( Asura_Chunk_Stream& xStream,const Asura_String& xOutputFileName,
														   const Tools_Platform_Type eTargetPlatform,const bool bCompressChunk /* = false */,
														   Tools_DynamicArray<Asura_String>* pxOutput /* = NULL */ )
{
	// tile TTL
	const u_int uTiledTextureCount( Regenerate() );

	// clip any free space
	Tools_TT_ClipTTLFreeSpace( *this );

	// ensure texture dims are a power of two after resize
	Tools_TT_ExpandTTLToNearestPowerOfTwo( *this );

	// update UVs to compensate for any change in texture size
	Tools_TT_UpdateTTLUVCoordinates( *this );

	// Fail if no textures have been tiled. If no textures have been tiled at this point then either there 
	// wern't any to begin with or they're never going to get tiled with current TTL settings - NDB. 
	if( uTiledTextureCount < 1 )
	{
		return false;
	}

	// write chunk file
	bool bSuccess( WriteChunk( xStream,xOutputFileName,eTargetPlatform,bCompressChunk ) );

	if( !bSuccess )
	{
		return false;
	}
	
	// if TiledTexture chunks were successfully written to disk then add chunk name to pxOutput	
	if(	pxOutput )
	{
		*pxOutput += xOutputFileName;
	}

	Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::White,"\"%s\" saved to disk",xOutputFileName.GetString() );

	// remove tiled textures from TTL
	while( GetTiledTextureCount() )
	{
		Erase( this->operator[](0).GetTextureID() );
	}

	// if there are still textures that could be tiled then recursively call this method untill there are none left
	if( Size() )
	{
		// increment output chunk name 
		Asura_String xNextTiledTextureName(xOutputFileName);

		Tools_TT_IncrementFileName(xNextTiledTextureName);		

		// call this method with remaining untiled textures
		bSuccess = WriteMultipleChunks( xStream,xNextTiledTextureName,eTargetPlatform,bCompressChunk,pxOutput );
	}
	
	// return true if all recursive calls to AutoTile were successful, false otherwise
	return bSuccess;
}

const bool Tools_TT_TiledTextureList::WriteTexture( const Asura_String& xOutputFileName )
{
	if( xOutputFileName.IsNull() || xOutputFileName.IsBlank() )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red," ERROR invlaid file name" );
		
		return false;
	}

	Tools_TT_Export_RawPixelData xRawPixelData( static_cast<u_int>(GetWidth()),static_cast<u_int>(GetHeight()),GetBytesPP() );
	GetRawPixelData( xRawPixelData );

	return	Tools_TT_Export_SaveTextureAs( xRawPixelData,xOutputFileName );
}


const TOOLS_TT_ERROR_CODE Tools_TT_TiledTextureList::ReadTiledTextureInfoArray( Asura_TiledTextureInfo* const pxTextureInfoBegin,Asura_TiledTextureInfo* const pxTextureInfoEnd )
{
	TOOLS_TT_ERROR_CODE eError(TOOLS_TT_FAIL);

	if( !(pxTextureInfoBegin < pxTextureInfoEnd) )
	{
		return eError;
	}

	eError = TOOLS_TT_OK;

	// set TTL dimensions using first texture info object
	SetWidth(  static_cast<int>(pxTextureInfoBegin->m_TiledToWidth) );
	SetHeight( static_cast<int>(pxTextureInfoBegin->m_TiledToHeight) );
	SetPadding(false); // padding is not stored in chunk files

	Asura_TiledTextureInfo* const pxFirstNonExistantTexture( Tools_TT_ShiftNonExistantFilesToEnd( pxTextureInfoBegin,pxTextureInfoEnd ) );

	if( pxFirstNonExistantTexture!=pxTextureInfoEnd )
	{
		eError = TOOLS_TT_FILE_NOT_FOUND;
	}

	int iTextureID(0);

	for( Asura_TiledTextureInfo* pxItr(pxTextureInfoBegin);pxItr!=pxFirstNonExistantTexture;++pxItr )
	{
		// add textures to Asura texture management system
		iTextureID = Asura_Texture_Management::AddTexture( pxItr->m_xNameOfTextureUsed.GetString() );
		
		if( iTextureID != ASURA_TEXTURE_NONE )
		{
			// insert tile into TTL
			Insert( Tools_TT_Tile( *pxItr, iTextureID ) );
		}
		else
		{
			Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"\"%s\" Texture file required by the TTL cannot be loaded",pxItr->m_xNameOfTextureUsed.GetString() );	
		}
	}

	SetModified(false);

	return eError;
}

const Tools_TT_Export_RawPixelData& Tools_TT_TiledTextureList::GetRawPixelData( Tools_TT_Export_RawPixelData& xRawPixelData ) const
{	
	if( (xRawPixelData.GetWidth()   != static_cast<u_int>(this->GetWidth()))  || 
		(xRawPixelData.GetHeight()  != static_cast<u_int>(this->GetHeight()))  )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red,"ERROR Raw pixel data container's dimensions don't match TTL's dimensions, aborting pixel data extraction" );
		return xRawPixelData;
	}
	
	u_int uTopLeftX; 
	u_int uTopLeftY;
	
	Asura_String xFileName;
	
	for( u_int u(0);u<Size();++u )
	{
		const Tools_TT_Tile& xTextureInfo( operator[](u) );

		if( xTextureInfo.IsTiled() )
		{
			uTopLeftX = static_cast<u_int>( xTextureInfo.GetPosition().x() );
			uTopLeftY = static_cast<u_int>( xTextureInfo.GetPosition().y() );

			// set window on bit array
			xRawPixelData.SetWindow( uTopLeftX,uTopLeftY,
									 uTopLeftX+xTextureInfo.GetWidth(),uTopLeftY+xTextureInfo.GetHeight() );

			// retrive bits from texture
			Tools_TextureTiler_Export_RipBitsFromARGBTexture( xTextureInfo.GetName().GetString(),xRawPixelData );
		}
	}

	// reset window to full size of raw pixel data
	xRawPixelData.SetWindowToMax();

	// copy texture flags to output
	xRawPixelData.SetTextureFlags( GetTextureFlags() );

	return xRawPixelData;
}

void Tools_TT_TiledTextureList::SetTextureFlag( const int iFlag,const bool bState /*= true */ )
{
	if( bState )
	{
		m_iTextureFlags |= iFlag;
	}
	else
	{
		m_iTextureFlags &= ~iFlag;
	}
}

void Tools_TT_UpdateTTLUVCoordinates( Tools_TT_TiledTextureList& xTTL )
{
	for( u_int u(0);u<xTTL.Size();++u )
	{	
		xTTL[u].UpdateUV( xTTL.GetWidth(),xTTL.GetHeight() );
	}
}

void Tools_TT_ClipTTLFreeSpace( Tools_TT_TiledTextureList& xTTL )
{
	// free space is always toward the right and bottom of a TTL, at least with current algorithms - NDB 111007

	// find max u2 and v2
	Tools_TT_Vector_2i xBottomRightMax;
	Tools_TT_Vector_2i xBottomRight;

	for( u_int u(0);u<xTTL.Size();++u )
	{
		if( xTTL[u].IsTiled() )
		{
			xBottomRight = xTTL[u].GetPosition();
			xBottomRight.x() += xTTL[u].GetWidth();
			xBottomRight.y() += xTTL[u].GetHeight();

			if( xBottomRight.x() > xBottomRightMax.x() ) { xBottomRightMax.x() = xBottomRight.x(); }
			if( xBottomRight.y() > xBottomRightMax.y() ) { xBottomRightMax.y() = xBottomRight.y(); }
		}
	}

	xTTL.SetDimension( xBottomRightMax );
}

void Tools_TT_ExpandTTLToNearestPowerOfTwo( Tools_TT_TiledTextureList& xTTL )
{
	xTTL.SetWidth( Tools_TT_FindNextLargerPowerOfTwo(xTTL.GetWidth()) );
	xTTL.SetHeight( Tools_TT_FindNextLargerPowerOfTwo(xTTL.GetHeight()) );
}

const bool Tools_TT_SaveTextureForPlatform( const Tools_TT_TiledTextureList& xTTL,const Asura_String& xOutputFileName,const Tools_Platform_Type eTargetPlatform )
{
	Asura_Assert( !xOutputFileName.IsBlank(),"output file name is blank" );

	// rip pixels from TTL
	Tools_TT_Export_RawPixelData xRawPixelData( static_cast<u_int>(xTTL.GetWidth()),static_cast<u_int>(xTTL.GetHeight()),xTTL.GetBytesPP() );
	xTTL.GetRawPixelData( xRawPixelData );

	// save pixles as a texture
	return Tools_TT_Export_SaveTextureForPlatform( xRawPixelData,xOutputFileName,eTargetPlatform );
}

const int Tools_TT_ExtractTexturesFromResourceChunk( const Asura_String& xInputFileName,const Asura_String& xTargetTextureName,Asura_TiledTextureInfo* const pxTextureInfoBegin,Asura_TiledTextureInfo* const pxTextureInfoEnd )
{
	// if there are no tiles to extract then do nothing
	if( !(pxTextureInfoBegin < pxTextureInfoEnd ) )
	{
		return 0;
	}

	Asura_Handle xFileHandle( Asura_FileAccess::Open( xInputFileName.GetString() ) );

	if( !Asura_FileAccess::ProcessAsuraFileHeader(xFileHandle) )
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red," ERROR could not process asura file header" );
		return 0;
	}
	
	bool bTargetNotFound(true);
	bool bTargetExtracted(false);
	Tools_TT_Export_RawPixelData xBitData;

	// resource chunk name is prefixed with "graphics\\" for compatability with texture resouce chunk loader
	Asura_String xFullTargetTextureName( "graphics\\" );
	xFullTargetTextureName += xTargetTextureName;

	// allocate memory to hold chunk headers read from file
	char* pcChunkBuffer( new char[512] );
	Asura_Chunk_Header*	pxChunkHeader(NULL);

	Asura_Chunk_File_Stream xStream( xFileHandle );

	while( bTargetNotFound && (xStream.ReadChunk_GetNextID(pcChunkBuffer)) )
	{
		// read chunk header
		xStream.ReadChunk_GetNextHeader( pcChunkBuffer,sizeof(Asura_Chunk_ResourceFile));
		pxChunkHeader = reinterpret_cast<Asura_Chunk_Header*>( pcChunkBuffer );
		
		if( pxChunkHeader->ID == ASURA_CHUNK_RESOURCEFILE )
		{
			Asura_Chunk_ResourceFile* pxChunkResourceFileHeader( reinterpret_cast<Asura_Chunk_ResourceFile*>( pcChunkBuffer ) );

			Asura_String xTextureFilename;
			xStream >> xTextureFilename;

			// if this is the texture resouce chunk we're looking for then read out the data
			if( xTextureFilename == xFullTargetTextureName )
			{
				// read out texture data into buffer
				bTargetNotFound = false;
				bTargetExtracted = Tools_TextureTiler_Export_RipAllARGBBitsFromFileInMemory( xFileHandle,pxChunkResourceFileHeader->FileSize,xBitData );
			}		
		}

		xStream.ReadChunk_End(false);
	}

	delete [] pcChunkBuffer;

	Asura_FileAccess::Close( xFileHandle );

	if(bTargetNotFound)
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red," ERROR could not find texture resource chunk \"%s\" in file \"%s\"",xFullTargetTextureName.GetString(),xInputFileName.GetString() );
		return 0;
	}

	if(!bTargetExtracted)
	{
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red," ERROR texture resource chunk \"%s\"",xFullTargetTextureName.GetString() );
		Tools_Timed_Message_System::AddMessage( Tools_TT_Colour::Red," found in file \"%s\" but could not be read",xInputFileName.GetString() );
		return 0;
	}

	// save required sub textures to disk
	Asura_String xTargetPath;
	int iExtractedTextureCount(0);

	for( Asura_TiledTextureInfo* pxItr(pxTextureInfoBegin);pxItr!=pxTextureInfoEnd;++pxItr )
	{
		xBitData.SetWindow( static_cast<u_int>(pxItr->m_fu1*pxItr->m_TiledToWidth),
							static_cast<u_int>(pxItr->m_fv1*pxItr->m_TiledToHeight),
							static_cast<u_int>(pxItr->m_fu2*pxItr->m_TiledToWidth),
							static_cast<u_int>(pxItr->m_fv2*pxItr->m_TiledToHeight) );

		
		xTargetPath = pxItr->m_xNameOfTextureUsed;
		Tools_TT_PrefixWithGraphicsPath( xTargetPath );

		if( Tools_TT_Export_SaveTextureAs( xBitData,xTargetPath ) )
		{
			++iExtractedTextureCount;
		}
	}

	return iExtractedTextureCount;
}

const Asura_String& Tools_TT_CreateResourceChunkName( Asura_String& xReturnString, const Asura_String& xOutputFileName,const Tools_Platform_Type eTargetPlatform,const bool bPrefixName /* = false */ )
{
	xReturnString.MakeBlank();

	Asura_String xName;
	xOutputFileName.SplitFilename( NULL,&xName,NULL );

	if( bPrefixName )
	{
		xReturnString += Tools_TT_TiledTextureList::GetDataNamePrefix();
	}
	
	xReturnString += xName;	
	xReturnString += Tools_TT_FileType::GetName( Tools_TT_FileType::GetTextureFileTypeForPlatform( eTargetPlatform ) );

	return xReturnString;
}

// <-- class IsTextureContainedInTTL -->
IsTextureTiled::IsTextureTiled( const Tools_TT_TiledTextureList* pxTTL ):
	m_pxTTL(pxTTL)
{}
