////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//
// Tools_Cubemap_Generator.cpp
//
// Created by Dan Clarke - 07/08/2008
//
// Copyright (C) 2008 Rebellion.  All Rights Reserved.
//

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"
#include "Tools_Cubemap_Generator.h"
#include "Debugging/Asura_Screenshot_System.h"
#include "Camera/Asura_Camera.h"

#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Tools_Overlays.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Rendering/Asura_Render.h"

#ifndef ASURA_USE_SCREENSHOT_SYSTEM
#error This relies upon the screenshot system being enabled.
#endif

#include "../../Middleware/NVidia/DDS_Utilities/dxtlib.h"

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

struct ToolCubeMapGen_QuadVertex
{
	float fX, fY, fZ, fRHW;
	float fU, fV;
};

#define TCUBEMAPGEN_FVF		(D3DFVF_XYZRHW | D3DFVF_TEX1)

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

Tools_Cubemap_Generator::StateEnum Tools_Cubemap_Generator::s_eState = Tools_Cubemap_Generator::STATE_IDLE;
Asura_Matrix_3x3 Tools_Cubemap_Generator::s_xOriginalOrientation;
Asura_Angle Tools_Cubemap_Generator::s_xOriginalFOV;
Direct3D_VertexBuffer* Tools_Cubemap_Generator::s_pxVertexBuffer = NULL;
u_int Tools_Cubemap_Generator::s_uScreenshotWidth = 0;
u_int Tools_Cubemap_Generator::s_uScreenshotHeight = 0;
bool Tools_Cubemap_Generator::s_bIsOverlayVisible = false;
Asura_Overlay* Tools_Cubemap_Generator::s_pxOverlay;
float Tools_Cubemap_Generator::s_fCubeMapDimension = 256;
bool Tools_Cubemap_Generator::s_bOriginalOverlaysHidden;
bool Tools_Cubemap_Generator::s_bHideOverlay = true;
bool Tools_Cubemap_Generator::s_bGenerateHDR = false;
Asura_String Tools_Cubemap_Generator::s_szFileLocationString = "Graphics/Cubemap.dds";
Direct3D_CubeTexture* Tools_Cubemap_Generator::s_pxCubemapTexture = NULL;

Direct3D_Surface* Tools_Cubemap_Generator::s_pxOriginalRenderTarget = NULL;
Direct3D_VertexBuffer* Tools_Cubemap_Generator::s_pxOriginalVertexBuffer = NULL;
u_int Tools_Cubemap_Generator::s_uOriginalVBOffset, Tools_Cubemap_Generator::s_uOriginalVBStride;
DWORD Tools_Cubemap_Generator::s_lOriginalFVF;

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tools_Cubemap_Generator::ToggleOverlay ()
{
	if (!s_bIsOverlayVisible)
	{
		CreateOverlay ();
	}
	else
	{
		CloseOverlay ();
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tools_Cubemap_Generator::CreateOverlay()
{
	Asura_Assert (!s_bIsOverlayVisible, "Overlay is already visible!");

	Asura_Vector_2 xPosition (6.0f, 6.0f);

	s_pxOverlay = new Asura_Overlay ("Cubemap Generator", false, &xPosition);
	s_pxOverlay->m_uID = TOOLS_OVERLAY_CUBEMAP_GENERATOR;

	// Add a button to generate the cubemap

	Asura_Vector_2 xOffset (1.0f, 3.0f);

	Asura_Dialog* pxGenerateButton = new Asura_Function_Dialog ("Generate Cubemap", xOffset, Tools_Cubemap_Generator::Create);

	s_pxOverlay->AddDialog (pxGenerateButton);

	// Add cubemap resolution options

	xOffset [1] += 2.0f;
	Asura_Dialog* pxCubeMapDimension = new Asura_Numerical_Dialog ("Dimension", xOffset, &s_fCubeMapDimension, 32.0f, 1024.0f, 1.0f, "%.0f");

	s_pxOverlay->AddDialog (pxCubeMapDimension);

	// Add other toggle options

	xOffset [1] += 2.0f;
	Asura_Dialog* pxOverlayHideToggle = new Asura_Tickbox_Dialog ("Hide Overlays", xOffset, &s_bHideOverlay);

	s_pxOverlay->AddDialog (pxOverlayHideToggle);
	xOffset [1] += 2.0f;

	// Add an HDR tickbox

	Asura_Dialog* pxHDRToggle = new Asura_Tickbox_Dialog("Generate HDR cubemap", xOffset, &s_bGenerateHDR);

	s_pxOverlay->AddDialog (pxHDRToggle);
	xOffset [1] += 2.0f;

	// Submit this dialog to the overlay system

	Asura_Overlay_System::SubmitNewOverlay (s_pxOverlay);

	s_bIsOverlayVisible = true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tools_Cubemap_Generator::CloseOverlay ()
{
	Asura_Overlay_System::CloseOverlays (TOOLS_OVERLAY_CUBEMAP_GENERATOR);

	s_bIsOverlayVisible = false;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tools_Cubemap_Generator::Create ()
{
	// Verify that the dimension is a power of 2

	u_int uDimension = (u_int) s_fCubeMapDimension;
	u_int uDimensionMinusOne = uDimension - 1;

	if ((uDimension & uDimensionMinusOne) != 0)
	{
		Tools_Timed_Message_System::AddMessage (0xff0000, "ERROR: Dimension must be a power of two!");
		return;
	}

	// Ask the user where they want the cubemap saved

	if (!ShowSaveAsDialogBox ())
	{
		return;
	}

	// Verify that the path exists

	Asura_String xPathOnly;

	s_szFileLocationString.SplitFilename (&xPathOnly, NULL, NULL);

	if (!Tools_FileSystem::IsDirectory (xPathOnly.GetString ()))
	{
		Tools_Timed_Message_System::AddMessage (0xff0000, "ERROR: Target path doesn't exist!");
		return;
	}

	// Show/Hide the overlays

	s_bOriginalOverlaysHidden = Asura_Overlay_System::AreOverlaysHidden ();

	Asura_Overlay_System::SetHideOverlays (s_bHideOverlay);

	// Create HDR cubemap if requested

	if(s_bGenerateHDR)
	{
		// If we're generating an HDR cubemap, then create the cubemap now, rather than at the end.

		HRESULT ret = Asura_Direct3D::D3DDevice->CreateCubeTexture((int)s_fCubeMapDimension, 0, D3DUSAGE_RENDERTARGET, D3DFMT_A16B16G16R16F, D3DPOOL_DEFAULT, &s_pxCubemapTexture, NULL);
		Asura_Assert (SUCCEEDED (ret), "Failed to create cubemap texture");
		if(FAILED(ret)) { return; }
	}

	Tools_Timed_Message_System::SetForceHideMessageSystem (true);

	// Kick off the cubemap generation

	s_eState = s_bGenerateHDR ? STATE_START_CAPTURES : STATE_SNAP_FRONT;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tools_Cubemap_Generator::Update()
{
	char szBuffer [256];
	char szTempPath [256];

	// During the cubemap generation process, each case block in this function will be called one at a time.  As we need the
	// screenshot system to do it's stuff (and the frame to actually render itself) - we can only do one screenshot per frame.
	// The last case block (STATE_GENERATE_CUBEMAP) is the final stage, which actually generates the cubemap.

	switch (s_eState)
	{
		case STATE_SNAP_FRONT:
		{
			s_xOriginalOrientation = *Asura_Camera::GetOrientation ();
			s_xOriginalFOV = Asura_Camera::GetVerticalFieldOfView ();
			Asura_Camera::SetVerticalFieldOfView (PI_OVER_2);

			Asura_Euler_3 xEulerAngle (0.0f, 0.0f, 0.0f);
			Asura_Camera::SetOrientation (xEulerAngle);

			if(!s_bGenerateHDR)
			{
				GetTempPath (256, szTempPath);
				Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, D3DCUBEMAP_FACE_POSITIVE_Z);
				Asura_ScreenShot_System::SetOverrideFileName (szBuffer);
				Asura_ScreenShot_System::TakeNormalScreenShot ();

				s_eState = STATE_SNAP_RIGHT;
			}

			break;
		}

		case STATE_SNAP_RIGHT:
		{
			Asura_Euler_3 xEulerAngle (0.0f, PI_OVER_2, 0.0f);
			Asura_Camera::SetOrientation (xEulerAngle);

			if(!s_bGenerateHDR)
			{
				GetTempPath (256, szTempPath);
				Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, D3DCUBEMAP_FACE_POSITIVE_X);
				Asura_ScreenShot_System::SetOverrideFileName (szBuffer);
				Asura_ScreenShot_System::TakeNormalScreenShot ();

				s_eState = STATE_SNAP_BACK;
			}

			break;
		}

		case STATE_SNAP_BACK:
		{
			Asura_Euler_3 xEulerAngle (0.0f, PI_OVER_2 * 2.0f, 0.0f);
			Asura_Camera::SetOrientation (xEulerAngle);

			if(!s_bGenerateHDR)
			{
				GetTempPath (256, szTempPath);
				Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, D3DCUBEMAP_FACE_NEGATIVE_Z);
				Asura_ScreenShot_System::SetOverrideFileName (szBuffer);
				Asura_ScreenShot_System::TakeNormalScreenShot ();

				s_eState = STATE_SNAP_LEFT;
			}

			break;
		}

		case STATE_SNAP_LEFT:
		{
			Asura_Euler_3 xEulerAngle (0.0f, PI_OVER_2 * 3.0f, 0.0f);
			Asura_Camera::SetOrientation (xEulerAngle);

			if(!s_bGenerateHDR)
			{
				GetTempPath (256, szTempPath);
				Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, D3DCUBEMAP_FACE_NEGATIVE_X);
				Asura_ScreenShot_System::SetOverrideFileName (szBuffer);
				Asura_ScreenShot_System::TakeNormalScreenShot ();

				s_eState = STATE_SNAP_UP;
			}

			break;
		}

		case STATE_SNAP_UP:
		{
			Asura_Euler_3 xEulerAngle (PI_OVER_2, 0.0f, 0.0f);
			Asura_Camera::SetOrientation (xEulerAngle);

			if(!s_bGenerateHDR)
			{
				GetTempPath (256, szTempPath);
				Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, D3DCUBEMAP_FACE_POSITIVE_Y);
				Asura_ScreenShot_System::SetOverrideFileName (szBuffer);
				Asura_ScreenShot_System::TakeNormalScreenShot ();

				s_eState = STATE_SNAP_DOWN;
			}

			break;
		}

		case STATE_SNAP_DOWN:
		{
			Asura_Euler_3 xEulerAngle (-PI_OVER_2, 0.0f, 0.0f);
			Asura_Camera::SetOrientation (xEulerAngle);

			if(!s_bGenerateHDR)
			{
				GetTempPath (256, szTempPath);
				Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, D3DCUBEMAP_FACE_NEGATIVE_Y);
				Asura_ScreenShot_System::SetOverrideFileName (szBuffer);
				Asura_ScreenShot_System::TakeNormalScreenShot ();

				s_eState = STATE_GENERATE_CUBEMAP;
			}

			break;
		}

		case STATE_GENERATE_CUBEMAP:
		{
			Asura_Camera::SetOrientation (s_xOriginalOrientation);
			Asura_Camera::SetVerticalFieldOfView (s_xOriginalFOV);
			s_eState = STATE_IDLE;

			// Take the 6 screenshots and generate the cubemap

			bool bRet;

			if(s_bGenerateHDR)
			{
				bRet = GenerateHDRCubeMap();
			}
			else
			{
				bRet = GenerateCubeMap();
			}

			// Display a 'success' message to the user

			if (bRet)
			{
				Tools_Timed_Message_System::AddMessage (0x00ff00, "Generated cubemap texture successfully.");
			}

			if(s_pxCubemapTexture)
			{
				s_pxCubemapTexture->Release();
				s_pxCubemapTexture = NULL;
			}

			// Restore the overlay hidden state

			Asura_Overlay_System::SetHideOverlays (s_bOriginalOverlaysHidden);

			Tools_Timed_Message_System::SetForceHideMessageSystem (false);

			Asura_ScreenShot_System::SetOverrideFileName ("");

			break;
		}
	}

	// Check if the overlay has been closed by the X button on the overlay

	if (s_bIsOverlayVisible && !Asura_Overlay_System::DoesOverlayExist (TOOLS_OVERLAY_CUBEMAP_GENERATOR))
	{
		s_bIsOverlayVisible = false;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::GenerateCubeMap ()
{
	HRESULT ret;
	char szBuffer [256];
	Direct3D_Texture* pxaTextures [6];
	char szTempPath [256];
	bool bRet = true;

	GetTempPath (256, szTempPath);

	s_uScreenshotWidth = Asura_Direct3D::GetPresentParameters ().BackBufferWidth;
	s_uScreenshotHeight = Asura_Direct3D::GetPresentParameters ().BackBufferHeight;

	// Init directx stuff

	if(!InitCubeMapGeneration())
	{
		Tools_Timed_Message_System::AddMessage (0xff0000, "ERROR: Failed in InitCubeMapGeneration.");
	}

	// Create the cubemap texture

	IDirect3DCubeTexture9* pxCubeTexture;

	ret = Asura_Direct3D::D3DDevice->CreateCubeTexture ((int) s_fCubeMapDimension, 0, 0, D3DFMT_DXT1, D3DPOOL_SYSTEMMEM, &pxCubeTexture, NULL);
	Asura_Assert (SUCCEEDED (ret), "Failed to create cubemap texture");
	if(FAILED(ret)) { return false; }

	// Load the screenshots into offscreen direct3d surfaces

	for (u_int n = 0; n < 6; n ++)
	{
		Asura_PrintfToBfr (szBuffer, "%scubemap_%02u.bmp", 256, true, szTempPath, n);
		ret = D3DXCreateTextureFromFile (Asura_Direct3D::D3DDevice, szBuffer, &pxaTextures [n]);
		Asura_Assert (SUCCEEDED (ret), "Failed to create texture from cubemap sub texture");
		if(FAILED(ret)) { return false; }

		// Crop the surfaces to a square aspect ratio

		Direct3D_Texture* pxTexture;
		Direct3D_Surface* pxSurface;

		ret = Asura_Direct3D::D3DDevice->CreateTexture((u_int)s_fCubeMapDimension, (u_int)s_fCubeMapDimension, 1, D3DUSAGE_RENDERTARGET, D3DFMT_X8R8G8B8, D3DPOOL_DEFAULT, &pxTexture, NULL);
		Asura_Assert (SUCCEEDED (ret), "Failed to create texture");
		if(FAILED(ret)) { return false; }

		ret = pxTexture->GetSurfaceLevel(0, &pxSurface);
		Asura_Assert (SUCCEEDED (ret), "Failed to get surface");
		if(FAILED(ret)) { return false; }

		CropImageToSquareAspect (pxaTextures[n], pxSurface);

		if(!WriteTextureToDXT1CubeTextureFace(pxSurface, pxCubeTexture, n))
		{
			Tools_Timed_Message_System::AddMessage (0xff0000, "ERROR: Failed when generating cubemap.");
			return false;
		}

		pxTexture->Release();
		pxSurface->Release();
	}

	// Write out the cubemap to a file

	ret = D3DXSaveTextureToFile (s_szFileLocationString.GetString (), D3DXIFF_DDS, pxCubeTexture, NULL);
	if (FAILED (ret))
	{
		Tools_Timed_Message_System::AddMessage (0xff0000, "ERROR: Failed to write cubemap texture to disc.  Is this file in use by another program?");
		return false;
	}

	// Clean-up

	for (u_int n = 0; n < 6; n ++)
	{
		pxaTextures [n]->Release ();
	}

	pxCubeTexture->Release ();

	CleanupCubeMapGeneration ();

	return bRet;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::GenerateHDRCubeMap()
{
	// Write out the cubemap to a file

	HRESULT ret = D3DXSaveTextureToFile(s_szFileLocationString.GetString (), D3DXIFF_DDS, s_pxCubemapTexture, NULL);
	if (FAILED (ret))
	{
		Tools_Timed_Message_System::AddMessage (0xff0000, "ERROR: Failed to write cubemap texture to disc.  Is this file in use by another program?");
		return false;
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::CropImageToSquareAspect (Direct3D_Texture* pxTexture, Direct3D_Surface* pxTargetSurface)
{
	HRESULT ret;

	// Render the image to the offscreen surface

	ret = Asura_Direct3D::D3DDevice->SetRenderTarget (0, pxTargetSurface);
	Asura_Assert (SUCCEEDED (ret), "Failed to set render target to offscreen surface");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->SetStreamSource (0, s_pxVertexBuffer, 0, sizeof (ToolCubeMapGen_QuadVertex));
	Asura_Assert (SUCCEEDED (ret), "Failed to set stream source");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->SetTexture (0, (IDirect3DBaseTexture9 *) pxTexture);
	Asura_Assert (SUCCEEDED (ret), "Failed to set texture");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->SetVertexShader (NULL);
	Asura_Assert (SUCCEEDED (ret), "Failed to disable vertex shader");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->SetPixelShader (NULL);
	Asura_Assert (SUCCEEDED (ret), "Failed to disable pixel shader");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->SetFVF (TCUBEMAPGEN_FVF);
	Asura_Assert (SUCCEEDED (ret), "Failed to set fvf");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->BeginScene ();
	Asura_Assert (SUCCEEDED (ret), "Failed to begin scene");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->Clear (0, NULL, D3DCLEAR_TARGET, 0xff000000, 0.0f, 0);
	Asura_Assert (SUCCEEDED (ret), "Failed to clear render target");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->DrawPrimitive (D3DPT_TRIANGLELIST, 0, 2);
	Asura_Assert (SUCCEEDED (ret), "Failed to draw quad to offscreen surface");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->EndScene ();
	Asura_Assert (SUCCEEDED (ret), "Failed to end scene");
	if(FAILED(ret)) { return false; }

	// Output temp debug texture
	//ret = D3DXSaveSurfaceToFile ("d:\\dump\\debug.tga", D3DXIFF_TGA, pxNewSurface, NULL, NULL);
	//Asura_Assert (SUCCEEDED (ret), "Failed to write debug texture to file");

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::CreateVertexBuffer ()
{
	ToolCubeMapGen_QuadVertex* pxVertices;
	HRESULT ret;

	if (s_pxVertexBuffer != NULL)
	{
		s_pxVertexBuffer->Release ();
	}

	// Create

	ret = Asura_Direct3D::D3DDevice->CreateVertexBuffer (sizeof (ToolCubeMapGen_QuadVertex) * 6, D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY, TCUBEMAPGEN_FVF, D3DPOOL_DEFAULT, &s_pxVertexBuffer, NULL);
	Asura_Assert (SUCCEEDED (ret), "Failed to create vertex buffer");
	if(FAILED(ret)) { return false; }

	// Lock

	ret = s_pxVertexBuffer->Lock (0, 0, (void **) &pxVertices, D3DLOCK_DISCARD);
	Asura_Assert (SUCCEEDED (ret), "Failed to lock vertex buffer");
	if(FAILED(ret)) { return false; }

	// Calculate the horizontal scale to make it crop the left and right edges

	const float fCoEff = ((float) s_uScreenshotWidth / (float) s_uScreenshotHeight) * s_fCubeMapDimension;
	const float fXOffset = (fCoEff - s_fCubeMapDimension) * 0.5f;

	// Fill in

	const float fXMin = -fXOffset;
	const float fXMax = s_fCubeMapDimension + fXOffset;
	const float fYMin = 0.0f;
	float fYMax = s_fCubeMapDimension;

	pxVertices [0].fX = fXMax;
	pxVertices [0].fY = fYMin;
	pxVertices [0].fZ = 0.01f;
	pxVertices [0].fRHW = 1.0f;
	pxVertices [0].fU = 1.0f;
	pxVertices [0].fV = 0.0f;

	pxVertices [1].fX = fXMin;
	pxVertices [1].fY = fYMin;
	pxVertices [1].fZ = 0.01f;
	pxVertices [1].fRHW = 1.0f;
	pxVertices [1].fU = 0.0f;
	pxVertices [1].fV = 0.0f;

	pxVertices [2].fX = fXMin;
	pxVertices [2].fY = fYMax;
	pxVertices [2].fZ = 0.01f;
	pxVertices [2].fRHW = 1.0f;
	pxVertices [2].fU = 0.0f;
	pxVertices [2].fV = 1.0f;

	pxVertices [3].fX = fXMax;
	pxVertices [3].fY = fYMin;
	pxVertices [3].fZ = 0.01f;
	pxVertices [3].fRHW = 1.0f;
	pxVertices [3].fU = 1.0f;
	pxVertices [3].fV = 0.0f;

	pxVertices [4].fX = fXMin;
	pxVertices [4].fY = fYMax;
	pxVertices [4].fZ = 0.01f;
	pxVertices [4].fRHW = 1.0f;
	pxVertices [4].fU = 0.0f;
	pxVertices [4].fV = 1.0f;

	pxVertices [5].fX = fXMax;
	pxVertices [5].fY = fYMax;
	pxVertices [5].fZ = 0.01f;
	pxVertices [5].fRHW = 1.0f;
	pxVertices [5].fU = 1.0f;
	pxVertices [5].fV = 1.0f;

	// Unlock

	ret = s_pxVertexBuffer->Unlock ();
	Asura_Assert (SUCCEEDED (ret), "Failed unlock vertex buffer");
	if(FAILED(ret)) { return false; }

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::InitCubeMapGeneration ()
{
	HRESULT ret;

	CreateVertexBuffer ();

	// Store the original values of any states that we change

	ret = Asura_Direct3D::D3DDevice->GetRenderTarget (0, &s_pxOriginalRenderTarget);
	Asura_Assert (SUCCEEDED (ret), "Failed to get render target");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->GetStreamSource (0, &s_pxOriginalVertexBuffer, &s_uOriginalVBOffset, &s_uOriginalVBStride);
	Asura_Assert (SUCCEEDED (ret), "Failed to get stream source");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->GetFVF (&s_lOriginalFVF);
	Asura_Assert (SUCCEEDED (ret), "Failed to get original fvf");
	if(FAILED(ret)) { return false; }

	// Set the render states

	Asura_Render::CurrentStates.RequestCullMode (ASURA_CULLMODE_NONE);
	Asura_Render::CurrentStates.RequestTranslucencyMode (ASURA_TRANSLUCENCY_OFF);
	Asura_Render::CurrentStates.RequestDynamicLights (false);
	Asura_Render::CurrentStates.RequestZBufferEnabled (false);
	Asura_Render::CurrentStates.RequestTextureOpMode (ASURA_TEXTUREOP_OPAQUE);
	Asura_Render::CurrentStates.RequestDitheringMode (false);

	Asura_Direct3D::D3DDevice->SetRenderState (D3DRS_SRGBWRITEENABLE, FALSE);

	Asura_Direct3D::D3DDevice->SetSamplerState (0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
	Asura_Direct3D::D3DDevice->SetSamplerState (0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);
	Asura_Direct3D::D3DDevice->SetSamplerState (0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	Asura_Direct3D::D3DDevice->SetSamplerState (0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Tools_Cubemap_Generator::CleanupCubeMapGeneration ()
{
	HRESULT ret;

	// Destroy Direct3D objects

	if (s_pxVertexBuffer != NULL)
	{
		s_pxVertexBuffer->Release ();
		s_pxVertexBuffer = NULL;
	}

	// Restore the original states of any states that we changed

	ret = Asura_Direct3D::D3DDevice->SetRenderTarget (0, s_pxOriginalRenderTarget);
	Asura_Assert (SUCCEEDED (ret), "Failed to restore original render target");

	ret = Asura_Direct3D::D3DDevice->SetStreamSource (0, s_pxOriginalVertexBuffer, s_uOriginalVBOffset, s_uOriginalVBStride);
	Asura_Assert (SUCCEEDED (ret), "Failed to restore original stream source");

	ret = Asura_Direct3D::D3DDevice->SetFVF (s_lOriginalFVF);
	Asura_Assert (SUCCEEDED (ret), "Failed to restore original fvf");

	if (s_pxOriginalRenderTarget)
	{
		s_pxOriginalRenderTarget->Release ();
		s_pxOriginalRenderTarget = NULL;
	}

	if (s_pxOriginalVertexBuffer)
	{
		s_pxOriginalVertexBuffer->Release ();
		s_pxOriginalVertexBuffer = NULL;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::ShowSaveAsDialogBox ()
{
	Tools_FileSystem_DialogInfo xInfo;

	xInfo.DefaultExtension = "dds";

	xInfo.DefaultFilename = "CubeMap.dds";
	xInfo.DefaultPath = "Graphics";
	xInfo.FileFilter = "Cubemap Files\0*.dds\0";
	xInfo.WindowTitle = "Save Cubemap Texture as...";

	Asura_Char* szName = Tools_FileSystem::GetSaveFilename (xInfo);

	if (!szName)
	{
		return false;
	}

	//Make sure the user hasn't selected a DDS directory.
	Asura_String xPath;
	Asura_String(szName).SplitFilename(&xPath,0,0);
	const char* pcSlashPos = strrchr(xPath.GetString(),'\\');
	const char* szLastDir = pcSlashPos ? pcSlashPos+1 : xPath.GetString();
	if(!stricmp(szLastDir,"dds"))
	{
		ShowMessageOverlay("Textures shouldn't be created in the DDS directory , since that is used for the automatically generated files.");
		return false;
	}

	s_szFileLocationString = szName;

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

static int DDSWrite_OutputNumber = 0;
static u_char** papucDXTDataPtr;

static NV_ERROR_CODE DxtMipMapCallback(const void* pPixelData, size_t uCount, const MIPMapData* pxMipMapData, void* pUserData)
{
	if(DDSWrite_OutputNumber > 1)
	{
		u_int uMipmapLevel = pxMipMapData->mipLevel;

		memcpy(papucDXTDataPtr[uMipmapLevel], pPixelData, uCount);
		papucDXTDataPtr[uMipmapLevel] += uCount;
	}

	DDSWrite_OutputNumber++;
	return NV_OK;
}

bool Tools_Cubemap_Generator::WriteTextureToDXT1CubeTextureFace(Direct3D_Surface* pxSourceSurface, Direct3D_CubeTexture* pxDestCubeTexture, u_int uFaceIndex)
{
	D3DLOCKED_RECT xLockedRect;
	D3DLOCKED_RECT xLockedRect2;
	nvCompressionOptions xOptions;
	HRESULT ret;
	u_int n;

	// First we need to use GetRenderTargetData to copy the data from the render target surface to a normal lockable surface

	Direct3D_Surface* pxPlainSourceSurface;

	ret = Asura_Direct3D::D3DDevice->CreateOffscreenPlainSurface((u_int)s_fCubeMapDimension, (u_int)s_fCubeMapDimension, D3DFMT_X8R8G8B8, D3DPOOL_SYSTEMMEM, &pxPlainSourceSurface, NULL);
	Asura_Assert(SUCCEEDED(ret), "Failed to create offscreen surface.");
	if(FAILED(ret)) { return false; }

	ret = Asura_Direct3D::D3DDevice->GetRenderTargetData(pxSourceSurface, pxPlainSourceSurface);
	Asura_Assert(SUCCEEDED(ret), "Failed to copy render target data.");
	if(FAILED(ret)) { return false; }

	// Lock the textures

	ret = pxPlainSourceSurface->LockRect(&xLockedRect, NULL, D3DLOCK_READONLY);
	Asura_Assert(SUCCEEDED(ret), "Failed to lock suface.");
	if(FAILED(ret)) { return false; }

	u_int uLevelCount = pxDestCubeTexture->GetLevelCount();

	papucDXTDataPtr = new u_char*[uLevelCount];

	for(n = 0; n < uLevelCount; n++)
	{
		ret = pxDestCubeTexture->LockRect((D3DCUBEMAP_FACES)uFaceIndex, n, &xLockedRect2, NULL, 0);
		Asura_Assert(SUCCEEDED(ret), "Failed to lock surface.");
		if(FAILED(ret)) { return false; }

		papucDXTDataPtr[n] = static_cast<u_char*>(xLockedRect2.pBits);
	}

	// Convert the texture data to dxt1

	DDSWrite_OutputNumber = 0;

	xOptions.SetDefaultOptions();
	xOptions.mipMapGeneration = kGenerateMipMaps;
	xOptions.textureFormat = kDXT1;

	DDSWrite_OutputNumber = 0;

	NV_ERROR_CODE eError = Asura_PC_CompressDXT(
		reinterpret_cast<const u_char*>(xLockedRect.pBits),
		256,
		256,
		xLockedRect.Pitch,
		nvBGRA,	// pixel order
		&xOptions,//compression options
		DxtMipMapCallback,	// mip callback
		NULL);	// subrect (use defaults).
	Asura_Assert(eError == NV_OK, "Failed to convert texture data to DXT");
	if(FAILED(ret)) { return false; }

	// Unlock the textures
	pxPlainSourceSurface->UnlockRect();

	for(n = 0; n < uLevelCount; n++)
	{
		pxDestCubeTexture->UnlockRect((D3DCUBEMAP_FACES)uFaceIndex, n);
	}

	delete [] papucDXTDataPtr;

	pxPlainSourceSurface->Release();

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::Render()
{
	if(s_bGenerateHDR)
	{
		switch (s_eState)
		{
			case STATE_START_CAPTURES:
			{
				s_eState = STATE_SNAP_FRONT;
				break;
			}

			case STATE_SNAP_FRONT:
			{
				CopyHDRBackbufferToSurface(D3DCUBEMAP_FACE_POSITIVE_Z);
				s_eState = STATE_SNAP_RIGHT;
				break;
			}

			case STATE_SNAP_RIGHT:
			{
				CopyHDRBackbufferToSurface(D3DCUBEMAP_FACE_POSITIVE_X);
				s_eState = STATE_SNAP_BACK;
				break;
			}

			case STATE_SNAP_BACK:
			{
				CopyHDRBackbufferToSurface(D3DCUBEMAP_FACE_NEGATIVE_Z);
				s_eState = STATE_SNAP_LEFT;
				break;
			}

			case STATE_SNAP_LEFT:
			{
				CopyHDRBackbufferToSurface(D3DCUBEMAP_FACE_NEGATIVE_X);
				s_eState = STATE_SNAP_UP;
				break;
			}

			case STATE_SNAP_UP:
			{
				CopyHDRBackbufferToSurface(D3DCUBEMAP_FACE_POSITIVE_Y);
				s_eState = STATE_SNAP_DOWN;
				break;
			}

			case STATE_SNAP_DOWN:
			{
				CopyHDRBackbufferToSurface(D3DCUBEMAP_FACE_NEGATIVE_Y);
				s_eState = STATE_GENERATE_CUBEMAP;
				break;
			}
		}
	}

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool Tools_Cubemap_Generator::CopyHDRBackbufferToSurface(D3DCUBEMAP_FACES eFace)
{
	Direct3D_Surface* pxBackBufferSurface;
	Direct3D_Surface* pxCubemapSurface;
	Direct3D_Surface* pxOffscreenSurface;
	D3DSURFACE_DESC xSurfaceDesc;
	HRESULT ret;

	// Ensure that we've finished rendering
	Asura_Render::FlushDisplay();

	// Get the back-buffer as a surface...
	ret = Asura_Direct3D::D3DDevice->GetRenderTarget (0, &pxBackBufferSurface);
	Asura_Assert(SUCCEEDED(ret), "Failed to get back buffer for screenshot");
	if(FAILED(ret)) { return false; }

	// Get the backbuffer surface description
	ret = pxBackBufferSurface->GetDesc(&xSurfaceDesc);
	Asura_Assert(SUCCEEDED(ret), "Failed to get backbuffer surface description");
	if(FAILED(ret)) { return false; }

	// Calculate the horizontal scale to make it crop the left and right edges

	int iXOffset = (int)((float)(xSurfaceDesc.Width - xSurfaceDesc.Height) * 0.5f);

	if(iXOffset < 0)
	{
		iXOffset = 0;
	}

	RECT xSourceRect;

	xSourceRect.left = iXOffset;
	xSourceRect.top = 0;
	xSourceRect.right = xSurfaceDesc.Width - iXOffset;
	xSourceRect.bottom = xSurfaceDesc.Height;

	u_int uNumMipmapLevels = s_pxCubemapTexture->GetLevelCount();

	// Downsample the render target (do once per each mipmap level)

	for(u_int n = 0; n < uNumMipmapLevels; n++)
	{
		ret = s_pxCubemapTexture->GetCubeMapSurface(eFace, n, &pxCubemapSurface);
		Asura_Assert(SUCCEEDED(ret), "Failed to get cubemap surface");

		if(SUCCEEDED(ret))
		{
			// Copy the backbuffer onto the cubemap surface

			ret = Asura_Direct3D::D3DDevice->StretchRect(pxBackBufferSurface, &xSourceRect, pxCubemapSurface, NULL, D3DTEXF_LINEAR);
			Asura_Assert(SUCCEEDED(ret), "StretchRect failed");

			// Make sure that the alpha channel is completely white (AvP seems to have black in the back buffer's alpha channel).
			// As I cannot lock a render target, then I have to use GetRenderTargetData to copy the data to another lockable surface,
			// then modify that surface, then use UpdateSurface to copy the data back to the cubemap surface.

			ret = pxCubemapSurface->GetDesc(&xSurfaceDesc);
			Asura_Assert(SUCCEEDED(ret), "Failed to get surface description.");

			if(SUCCEEDED(ret))
			{
				ret = Asura_Direct3D::D3DDevice->CreateOffscreenPlainSurface(xSurfaceDesc.Width, xSurfaceDesc.Height, D3DFMT_A16B16G16R16F, D3DPOOL_SYSTEMMEM, &pxOffscreenSurface, NULL);
				Asura_Assert(SUCCEEDED(ret), "Failed to create offscreen surface.");

				if(SUCCEEDED(ret))
				{
					ret = Asura_Direct3D::D3DDevice->GetRenderTargetData(pxCubemapSurface, pxOffscreenSurface);
					Asura_Assert(SUCCEEDED(ret), "Failed to get render target data.");

					if(SUCCEEDED(ret))
					{
						D3DLOCKED_RECT xLockedRect;
						ret = pxOffscreenSurface->LockRect(&xLockedRect, NULL, 0);
						Asura_Assert(SUCCEEDED(ret), "Failed to lock surface");

						if(SUCCEEDED(ret))
						{
							for(u_int y = 0; y < xSurfaceDesc.Height; y++)
							{
								for(u_int x = 0; x < xSurfaceDesc.Height; x++)
								{
									D3DXFLOAT16* pfFloatBuffer = (D3DXFLOAT16*)((u_char*)xLockedRect.pBits + (y * xLockedRect.Pitch) + (x * 8));

									pfFloatBuffer[3] = 1.0f;
								}
							}

							pxOffscreenSurface->UnlockRect();

							ret = Asura_Direct3D::D3DDevice->UpdateSurface(pxOffscreenSurface, NULL, pxCubemapSurface, NULL);
							Asura_Assert(SUCCEEDED(ret), "UpdateSurface failed");
						}
					}
				}
			}

			pxCubemapSurface->Release();
		}
	}

	pxBackBufferSurface->Release();

	return true;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
