//
// Tools_TextureReport.cpp
//
// Created by Richard Rice - 28/10/2005
// 
// Copyright (C) 2005 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"
#include "Tools_TextureReport.h"

#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Material_System.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Entity/Asura_Entity.h"
#include "Debugging/Asura_Assert.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"

#include "Entity/Tools_Entity_Data.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "PC/FileSystem/Tools_FileSystem.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM
#include "Entity/Container/Asura_ServerEntity_Container.h"
#include "Entity/Container/Tools_Container_System.h"
#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

struct Tools_TextureReportEntry
{
	const void* m_pObject;
	Asura_String m_xDesc;
	int m_iCount;
};

class Tools_TextureReportMain
{
public :
	Tools_TextureReportMain();
	~Tools_TextureReportMain();
	void DoReport();

private :
	void AddToLists(int iTextureID,const void* pObject,const char* pszDesc);

	static void EntityCallback(int iTextureID,void* pUserData);

	List<Tools_TextureReportEntry*>** m_ppxLists;
	const Asura_Entity_Server* m_pxCurrentEntity;
};

void Tools_TextureReport::OutputReport()
{
	Tools_TextureReportMain xReport;
	xReport.DoReport();
}



Tools_TextureReportMain::Tools_TextureReportMain()
{
	//initialise lists
	m_ppxLists = new List<Tools_TextureReportEntry*>*[ASURA_MAX_NO_OF_TEXTURES];
	for(int i=0;i<ASURA_MAX_NO_OF_TEXTURES;i++)
	{
		m_ppxLists[i] = 0;
	}
}

Tools_TextureReportMain::~Tools_TextureReportMain()
{
	for(int i=0;i<ASURA_MAX_NO_OF_TEXTURES;i++)
	{
		if(m_ppxLists[i])
		{
			while(m_ppxLists[i]->size())
			{
				delete m_ppxLists[i]->first_entry();
				m_ppxLists[i]->delete_first_entry();
			}
		}
		
		delete m_ppxLists[i];
	}
	delete [] m_ppxLists;
}


void Tools_TextureReportMain::AddToLists(int iTextureID,const void* pObject,const char* pszDesc)
{
	if(iTextureID<0) return;
	Asura_Assert(iTextureID<ASURA_MAX_NO_OF_TEXTURES,"");

	//allocate list for texture if not done yet
	if(!m_ppxLists[iTextureID])
	{
		m_ppxLists[iTextureID] = new List<Tools_TextureReportEntry*>;
	}

	//search for an existing entry for this texture/object
	for(LIF<Tools_TextureReportEntry*> xLIF(m_ppxLists[iTextureID]);!xLIF.done();xLIF.next())
	{
		if(xLIF()->m_pObject == pObject)
		{
			//just increment the count for this object
			xLIF()->m_iCount++;
			return;
		}
	}

	//add a new entry
	Tools_TextureReportEntry* pxEntry = new Tools_TextureReportEntry;
	pxEntry->m_pObject = pObject;
	pxEntry->m_xDesc = pszDesc;
	pxEntry->m_iCount = 1;

	m_ppxLists[iTextureID]->add_entry(pxEntry);
}

void Tools_TextureReportMain::DoReport()
{

	//Get filename and open file now to confirm that this should be done
	Tools_FileSystem_DialogInfo xInfo;

	xInfo.DefaultExtension = "txt";

	Asura_String xLevelFilename = Tools_SaveAsura::GetMainFilename();
	Asura_String xPath;
	xLevelFilename.SplitFilename(&xPath,0,0);

	xInfo.DefaultFilename = "TextureReport.txt";
	xInfo.DefaultPath = xLevelFilename.GetString();
	xInfo.FileFilter = "Text Files\0*.txt\0";
	xInfo.WindowTitle = "Write texture report as :";

	Asura_Char* pszFileName = Tools_FileSystem::GetSaveFilename(xInfo);
	if(!pszFileName) return;

	FILE* pFile = fopen(pszFileName,"wt");
	if(!pFile)
	{
		ShowMessageOverlay("Failed to open file for writing");
		return;
	}



	//check all meshes
	for(int i=0;i<Asura_Mesh_System::GetNumberOfMeshes();i++)
	{
		Asura_Mesh* pxMesh = Asura_Mesh_System::GetMesh(i);
		if(!pxMesh->TextureDataPtr) continue;

		for(int j=0;j<pxMesh->NumberOfPolygons;j++)
		{
			int iMat = pxMesh->TextureDataPtr[j].m_iMaterialID;

			for(int k=0;k<ASURA_MATERIAL_MAXTEXTURES;k++)
			{
				AddToLists(Asura_Material_System::GetTextureID(iMat,k),pxMesh,pxMesh->Name.GetString());
				AddToLists(Asura_Material_System::GetNormalMapID(iMat,k),pxMesh,pxMesh->Name.GetString());

			}
		}
	}

	//cutscenes
	for(int i=0;i<Asura_Cutscene_System::GetNumberOfCutscenes();i++)
	{
		Asura_Cutscene* pxCutscene = Asura_Cutscene_System::GetCutsceneByIndex(i);
		Asura_NameTree xNameTree;
		pxCutscene->BuildUsedTextureTree(&xNameTree);
		if(!xNameTree.GetNumberOfNodes()) continue;

		//cutscene uses some textures , so get a desc for this
		char szDesc[200];
		Asura_PrintfToBfr(szDesc,"Cutscene - %s",200,true,pxCutscene->GetName().GetString());

		for(int i=0;i<xNameTree.GetNumberOfNodes();i++)
		{
			const Asura_String* pxTextureName = xNameTree.FindNodeNameByIndex(i);
			Asura_Assert(pxTextureName,"");

			//using AddTexture in case the textures haven't been assigned indeces yet
			int iTextureID = Asura_Texture_Management::AddTexture(pxTextureName->GetString());

			AddToLists(iTextureID,pxCutscene,szDesc);
	
		}
	}

	//entities
	for(Asura_Classified_List_It<Asura_Entity_Server> it(Asura_Entity_Server::GetList());!it.Done();it.Next())
	{
		Asura_Entity_Server* pxEntity = it.GetCurrent();
		//ignore entities marked for destruction
		if(pxEntity->IsDestroyed()) continue;

		//add textures used by this entity to the list
		//need to store the pointer ,since it won't be given by the callback
		m_pxCurrentEntity = pxEntity;
		pxEntity->TraverseUsedTextures(EntityCallback,this);
	}


#ifdef ASURA_USE_CONTAINER_SYSTEM
	// Container entities. Add textures for projector, photon, corona, lightshaft etc containers.
	for (Asura_Hashed_List_It<Asura_ServerEntity_Container> xIt(Asura_ServerEntity_Container::GetList()); !xIt.Done(); xIt.Next())
	{
		Asura_ServerEntity_Container* pxEntity = xIt.GetCurrent();

		// Ignore entities marked for destruction
		if (pxEntity->IsDestroyed()) continue;

		// Add textures used by this entity to the list.
		// We need to store the pointer, since it won't be given by the callback.
		m_pxCurrentEntity = pxEntity;

		const Asura_Hash_ID uContainerHash = pxEntity->GetContainerHash();
		const Tools_Container* pxContainer = Tools_Container_System::FindContainerByHash(uContainerHash);
		if (pxContainer && !pxContainer->HasReportedDepthError())
		{
			pxContainer->TraverseUsedTexturesForReport(EntityCallback, this);
		}
	}
#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM


	fprintf(pFile,"This file lists textures used by all currently loaded meshes and entities.\n");
	fprintf(pFile,"Note that this may include some objects that aren't actually used by the level.\n");
	fprintf(pFile,"Also note that it may be necessary to do a level export or quit to game to ensure that all the used meshes have been loaded.\n");
	fprintf(pFile,"The numbers after the names indicate the number of times each texture is used. This only really has any meaning for meshes.\n\n\n");

	for(int i=0;i<ASURA_MAX_NO_OF_TEXTURES;i++)
	{
		if(!m_ppxLists[i]) continue;
		const Asura_String* pxName = Asura_Texture_Management::GetTextureName(i);

		fprintf(pFile,"## %s ##\n",pxName ? pxName->GetString() : "*Unknown*");
		for(LIF<Tools_TextureReportEntry*> xLIF(m_ppxLists[i]);!xLIF.done();xLIF.next())
		{
			fprintf(pFile,"%-50s %d\n",xLIF()->m_xDesc.GetString(),xLIF()->m_iCount);
		}

		fprintf(pFile,"\n\n");
		
	}

	fclose(pFile);

}

void Tools_TextureReportMain::EntityCallback(int iTextureID,void* pUserData)
{
	Tools_TextureReportMain* pxThis = (Tools_TextureReportMain*) pUserData;
	if(iTextureID == ASURA_TEXTURE_NONE) return;

	//get a description for the entity
	char szDesc[200];
	const Asura_Entity_Server* pxEntity = pxThis->m_pxCurrentEntity;
	
	if(pxEntity->ToolsData && !pxEntity->ToolsData->Name.IsBlank())
	{
		Asura_PrintfToBfr(szDesc,"Entity - %s",200,true,pxEntity->ToolsData->Name.GetString());
	}
	else
	{
		Asura_PrintfToBfr(szDesc,"Entity %d",200,true,pxEntity->GetGuid());
	}

	pxThis->AddToLists(iTextureID,pxEntity,szDesc);
}
