//
// Tools_Undo.cpp
//
// Created by Kevin Lea - 05/10/2001 16:54:20
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Tools_Undo.h"
#include <string.h>

#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Interface/Asura_Overlay_System.h"

#include "Misc/Tools_Timed_Messages.h"

////////////////////////////////////////////////////////////////////////////////////////
RootUndo::RootUndo(int buffer_size)
{
	Position = 0;
	StartPosition = 0;
	NumUndo = 0;
	NumRedo = 0;
	UndoInProcess = false;
	UndoInitialised = false;

	BufferSize = buffer_size;

	m_bVerbose = false;

	m_fMinTimeToKeepBatchOpen = 0.0f;
	m_fLastUndoLogTime = 0.0f;
}

RootUndo::~RootUndo()
{
}

void RootUndo::Initialise()
{
	Position = 0;
	StartPosition = 0;
	NumUndo = 0;
	NumRedo = 0;
	UndoInProcess = false;
	NewUndoBatch = false;
	UndoInitialised = true;
	UndoLoggedThisFrame = false;
}

void RootUndo::Shutdown()
{
	Position = 0;
	StartPosition = 0;
	NumUndo = 0;
	NumRedo = 0;
	UndoInProcess = false;
	NewUndoBatch = false;
	UndoInitialised = false;
}

bool RootUndo::Undo()
{
	if(!UndoInitialised) return false;
	if(UndoInProcess) return false;
	if(!NumUndo) return false;

	PreUndoProcessing();
	UndoInProcess = true;

	//initialise undo description
	if(m_bVerbose)
	{
		m_xUndoDescription = "";
		m_iUndoDescriptionPriority = -1;
	}


	Prev(Position);

	while(!IsNullSlot(Position) && !IsUndoStart(Position))
	{
		ProcessUndo(Position);
		Prev(Position);
	}


	NumUndo--;
	NumRedo++;	

	PostUndoProcessing();

	//display message if needed
	if(m_bVerbose)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Undo %s",m_xUndoDescription.GetString());
	}

	UndoInProcess = false;

	return true;
}

bool RootUndo::Redo()
{
	if(!UndoInitialised) return false;
	if(UndoInProcess) return false;
	if(!NumRedo) return false;

	PreUndoProcessing();
	UndoInProcess = true;

	//initialise undo description
	if(m_bVerbose)
	{
		m_xUndoDescription = "";
		m_iUndoDescriptionPriority = -1;
	}


	Next(Position);

	while(!IsNullSlot(Position) && !IsUndoStart(Position))
	{
		ProcessUndo(Position);
		Next(Position);
	}


	NumUndo++;
	NumRedo--;
	
	PostUndoProcessing();	

	//display message if needed
	if(m_bVerbose)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Redo %s",m_xUndoDescription.GetString());
	}


	UndoInProcess = false;

	return true;

}

void RootUndo::FlushUndoBuffer()
{
	if(UndoInitialised)
	{
		Initialise();
	}
}

void RootUndo::MarkUndoStart()
{
	if(!UndoInitialised) return;
	if(UndoInProcess) return;

	//the next time undo data is allocated , it will be put at the start of
	//a new undo batch.
	//However redo data won't be cleared until that time.
	NewUndoBatch = true;
}

void RootUndo::MarkUndoStartIfEndOfBatch()
{
	if(!UndoInitialised) return;

	bool bEnoughTimePassed = false;
	if((Asura_Timers::GetLocalUnscaledTime() - m_fLastUndoLogTime) >= m_fMinTimeToKeepBatchOpen)
	{
		//Some time has passed since the last undo was logged
		bEnoughTimePassed = true;
	}

	//Mark start of a new batch if enough time has passed since the last undo , and if no undo was
	//logged in the last frame.
	//(The latter may also happen if the batch was kept open voa KeepUndoBatchOpenThisFrame())
	if(!UndoLoggedThisFrame && bEnoughTimePassed)
	{
		MarkUndoStart();
	}
	
	UndoLoggedThisFrame = false;
	
}

void RootUndo::KeepUndoBatchOpenThisFrame()
{
	UndoLoggedThisFrame = true;
}

void RootUndo::SetStartOfNewUndoIfNeeded()
{
	if(!UndoInitialised) return;
	if(UndoInProcess) return;
	/*
	Just about to allocate some space for a new undo.
	We need to mark the start of a new undo batch if :
	1.This is our very first undo
	2.There is some redo data outstanding (which will need to be cleared)
	3.The flag for starting a new undo has been set
	*/
	if(NumRedo >0 || !NumUndo || NewUndoBatch)
	{
		int pos = Position;
		//need to overwrite all the redo data with nulls
		while(NumRedo)
		{
			NumRedo--;
			
			MarkSlotAsNull(pos);
			Next(pos);

			while(!IsNullSlot(pos) && !IsUndoStart(pos))
			{
				MarkSlotAsNull(pos);
				Next(pos);
			}
		}

		//mark the start of the next undo block
		MarkSlotAsUndoStart(Position);
		Next(Position);

		NumUndo++;

		//Reset the request for a new undo batch
		NewUndoBatch = false;
		
		if(!IsNullSlot(Position))
		{
			//if we have got here then we have wrapped round and hit the start of the undo data.
			//therefore need to eliminate the oldest undo to make room
			if(NumUndo>2)
			{
				NumUndo--;
				int pos = Position;
			
				MarkSlotAsNull(pos);
				Next(pos);
			
				while(!IsNullSlot(pos) && !IsUndoStart(pos))
				{
					MarkSlotAsNull(pos);
					Next(pos);
				}
				StartPosition = pos;
			}
			else
			{
				//resize if needed
				Prev(Position);
				Resize(BufferSize*2);
				Next(Position);
			}
			
		}
	}
}


////////////////////////////////////////////////////////////////////////////////////////
FixedSizeUndo::FixedSizeUndo(int buffer_size,int undo_size)
:RootUndo(buffer_size)
{
	UndoDataSize = ((undo_size+3)&~3)/4;
	UndoBuffer = NULL;
}

FixedSizeUndo::~FixedSizeUndo()
{
	Shutdown();
}

void FixedSizeUndo::Initialise()
{
	RootUndo::Initialise();
	
	delete [] UndoBuffer;
	UndoBuffer = NULL;

	UndoBuffer = new int[BufferSize*UndoDataSize];

	for(int pos=0;pos<(BufferSize*UndoDataSize);pos+=UndoDataSize)
	{
		UndoBuffer[pos] = FixedSizeUndo_Null;
	}	
}

void FixedSizeUndo::Shutdown()
{
	RootUndo::Shutdown();

	delete [] UndoBuffer;
	UndoBuffer = 0;
}


bool FixedSizeUndo::IsNullSlot(int slot)
{
	return UndoBuffer[slot*UndoDataSize] == FixedSizeUndo_Null;
}
bool FixedSizeUndo::IsUndoStart(int slot)
{
	return UndoBuffer[slot*UndoDataSize] == FixedSizeUndo_StartOfUndo;
}
void FixedSizeUndo::ProcessUndo(int slot)
{
	ProcessUndoData((void*)&UndoBuffer[slot*UndoDataSize]);	
}

void FixedSizeUndo::MarkSlotAsNull(int slot)
{
	UndoBuffer[slot*UndoDataSize] = FixedSizeUndo_Null;
}

void FixedSizeUndo::MarkSlotAsUndoStart(int slot)
{
	UndoBuffer[slot*UndoDataSize] = FixedSizeUndo_StartOfUndo;
}


void* FixedSizeUndo::AllocateNextUndoData()
{
	if(!UndoBuffer) return 0;
	if(UndoInProcess) return 0;

	UndoLoggedThisFrame = true;
	m_fLastUndoLogTime = Asura_Timers::GetLocalUnscaledTime();
	
	//check to see if anew undo batch neds to be started
	SetStartOfNewUndoIfNeeded();
	
	void* retval = (void*)&UndoBuffer[Position*UndoDataSize];
	Next(Position);


	if(!IsNullSlot(Position))
	{
		//if we have got here then we have wrapped round and hit the start of the undo data.
		//therefore need to eliminate the oldest undo to make room
		if(NumUndo>2)
		{
			NumUndo--;
			int pos = Position;
		
			MarkSlotAsNull(pos);
			Next(pos);
		
			while(!IsNullSlot(Position) && !IsUndoStart(Position))
			{
				MarkSlotAsNull(pos);
				Next(pos);
			}
			StartPosition = pos;
		
		}
		else
		{
			//resize if needed
			Prev(Position);
			Resize(BufferSize*2);
			return(AllocateNextUndoData());
		}
		
	}
	return retval;
}



void FixedSizeUndo::Resize(int NewBufferSize)
{
	if(!UndoBuffer) return;

	//allocate the new undo buffer
	int* NewUndoBuffer = new int[NewBufferSize*UndoDataSize];

	//mark all the slots as null
	for(int pos=0;pos<(NewBufferSize*UndoDataSize);pos+=UndoDataSize)
	{
		NewUndoBuffer[pos] = FixedSizeUndo_Null;
	}	
	

	//copy the contents of the old buffer from the StartPosition to the start of the undo buffer
	int ints_after_start = (BufferSize-StartPosition)*UndoDataSize;
	memcpy(NewUndoBuffer,&UndoBuffer[StartPosition*UndoDataSize],ints_after_start*4);

	//if the StartPosition wasn't right at the start of the old buffer , copy the beginning of the buffer
	//over as well.
	if(StartPosition>0)
	{
		memcpy(NewUndoBuffer+ints_after_start,UndoBuffer,StartPosition*4*UndoDataSize);
	}

	//since StartPosition will be going back to 0 , we need to adjust Position accordingly , 
	//so that it stays the same relative to it
	Position = Position - StartPosition;
	if(Position<0)
	{
		Position+=BufferSize;
	}
	//dump the old buffer
	delete [] UndoBuffer;

	BufferSize = NewBufferSize;
	UndoBuffer = NewUndoBuffer;
	StartPosition = 0;
}
//////////////////////////////////////////////////////////////////////////////////////



VariableSizeUndo::VariableSizeUndo(int buffer_size)
:RootUndo(buffer_size)
{
	UndoBuffer = NULL;
}

VariableSizeUndo::~VariableSizeUndo()
{
	Shutdown();
}


void VariableSizeUndo::Initialise()
{
	RootUndo::Initialise();
	
	if(UndoBuffer)
	{
		for(int i=0;i<BufferSize;i++)
		{
			MarkSlotAsNull(i);
		}		
	}
	delete [] UndoBuffer;
	UndoBuffer = 0;

	UndoBuffer = new RootUndoData*[BufferSize];

	for(int pos=0;pos<BufferSize;pos++)
	{
		UndoBuffer[pos] = VariableSizeUndo_Null;
	}	
}

void VariableSizeUndo::Shutdown()
{
	RootUndo::Shutdown();
	
	if(UndoBuffer)
	{
		for(int i=0;i<BufferSize;i++)
		{
			MarkSlotAsNull(i);
		}		
	}
	delete [] UndoBuffer;
	UndoBuffer = NULL;
}


bool VariableSizeUndo::IsNullSlot(int slot)
{
	return UndoBuffer[slot] == VariableSizeUndo_Null;
}
bool VariableSizeUndo::IsUndoStart(int slot)
{
	return UndoBuffer[slot] == VariableSizeUndo_StartOfUndo;
}
void VariableSizeUndo::ProcessUndo(int slot)
{
	if(m_bVerbose)
	{
		//store description of undo if higher priority than previous
		if(UndoBuffer[slot]->GetPriority() > m_iUndoDescriptionPriority)
		{
			const char* pszText = UndoBuffer[slot]->GetDescription();
			if(pszText)
			{
				m_iUndoDescriptionPriority = UndoBuffer[slot]->GetPriority();
				m_xUndoDescription = pszText;
			}
		}
	}

	RootUndoData* undo_data = UndoBuffer[slot]->ProcessUndoData();	
	
	if(undo_data!=UndoBuffer[slot])
	{
		delete UndoBuffer[Position];
		UndoBuffer[Position] = undo_data;
	}
}



void VariableSizeUndo::MarkSlotAsNull(int slot)
{
	if(UndoBuffer[slot] != VariableSizeUndo_StartOfUndo && UndoBuffer[slot] != VariableSizeUndo_Null)
	{
		delete UndoBuffer[slot];
	}
	UndoBuffer[slot] = VariableSizeUndo_Null;
}

void VariableSizeUndo::MarkSlotAsUndoStart(int slot)
{
	if(UndoBuffer[slot] != VariableSizeUndo_StartOfUndo && UndoBuffer[slot] != VariableSizeUndo_Null)
	{
		delete UndoBuffer[slot];
	}
	UndoBuffer[slot] = VariableSizeUndo_StartOfUndo;
}

RootUndoData** VariableSizeUndo::AllocateNextUndoData()
{
	if(!UndoInitialised) return 0;
	if(!UndoBuffer) return 0;
	if(UndoInProcess) return 0;

	UndoLoggedThisFrame = true;
	m_fLastUndoLogTime = Asura_Timers::GetLocalUnscaledTime();

	
	//check to see if anew undo batch neds to be started
	SetStartOfNewUndoIfNeeded();
	
	RootUndoData** retval = &UndoBuffer[Position];
	Next(Position);


	if(UndoBuffer[Position]!=VariableSizeUndo_Null)
	{
		//if we have got here then we have wrapped round and hit the start of the undo data.
		//therefore need to eliminate the oldest undo to make room
		if(NumUndo>2)
		{
			NumUndo--;
			int pos = Position;
		
			MarkSlotAsNull(pos);
			Next(pos);
		
			while(!IsNullSlot(Position) && !IsUndoStart(Position))
			{
				MarkSlotAsNull(pos);
				Next(pos);
			}
			StartPosition = pos;
		}
		else
		{
			//resize if needed
			Prev(Position);
			Resize(BufferSize*2);
			return(AllocateNextUndoData());
		}
		
	}
	return retval;
}



void VariableSizeUndo::Resize(int NewBufferSize)
{
	if(!UndoBuffer) return;

	RootUndoData** NewUndoBuffer = new RootUndoData*[NewBufferSize];

	for(int pos=0;pos<NewBufferSize;pos++)
	{
		NewUndoBuffer[pos] = VariableSizeUndo_Null;
	}	
	

	int ints_after_start = BufferSize-StartPosition;

	memcpy(NewUndoBuffer,&UndoBuffer[StartPosition],ints_after_start*4);

	if(StartPosition>0)
	{
		memcpy(NewUndoBuffer+ints_after_start,UndoBuffer,StartPosition*4);
	}

	Position = Position - StartPosition;
	if(Position<0)
	{
		Position+=BufferSize;
	}

	delete [] UndoBuffer;

	BufferSize = NewBufferSize;
	UndoBuffer = NewUndoBuffer;
	StartPosition = 0;
}

void RootUndoData::SetDescription(const char* pszFormat,...)
{
	char szBuffer[255];

	// format string and place in buffer
	va_list ap;
	va_start(ap, pszFormat);
    Asura_PrintfToBfr( szBuffer, pszFormat, ap, 255, true );
	va_end(ap);

	m_xUndoDescription = szBuffer;
}

////////////////////////////////////////////////////////////////////////////////////////
/*-------------------**
** Tools_Undo_System **
**-------------------*/

VariableSizeUndo Tools_Undo_System::s_xUndo(10000);


void Tools_Undo_System::Initialise()
{
	s_xUndo.Initialise();
	s_xUndo.SetVerbose(true);
}

void Tools_Undo_System::Update()
{
	s_xUndo.MarkUndoStartIfEndOfBatch();

	//TBD Sort out issues with key clashes with other systems/modes.
	//Currently shortcut keys are just dealt with directly by texturing mode , 
	//which is the only mode that really cares about this instance of the undo system
	//at the moment.
	//RAR 04/06/2007
	#if 0
	//check for keyboard shortcuts for undo/redo.
	if(!Asura_Overlay_System::OverlaysHaveKeyboardFocus())
	{
		if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
		{
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
		{
			if(Asura_UserInput::DebouncedKeyPressed('z'))
			{
				Undo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('y'))
			{
				Redo();
			}
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
		{
		}
		else
		{
			if(Asura_UserInput::DebouncedKeyPressed(','))
			{
				Undo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('.'))
			{
				Redo();
			}
		}
	}
	#endif
}