//
// Tools_ViewCell.cpp
//
// Created by KJFL - 01 June 2007
//
// Copyright (C) 2007 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM

// This file's header.
#include "Visibility/Tools_ViewCell.h"

// Engine independent.
#include "Core/Memory/Asura_Mem.h"
#include "Core/Asura_LinkList.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Mempool.h"
#include "Core/Asura_RingBuffer.h"
#include "Debugging/Asura_Logfile.h"
#include "Rendering/Asura_Render.h"

// Engine tools.
#include "Visibility/Tools_Visibility_System.h"


#define ASURA_VISIBILITY_LOGFILE "logfiles/visibility.txt"

static const u_int uMIN_NUMBER_OF_DIFFERENCES_TO_ALLOW_SPLIT = 3;
static const float fNORMAL_FEATURE_SCORE = 1.0f;
static const float fSKYBOX_SCORE = 5.0f;
static const u_int uNUM_VIS_SEGMENTS_PER_POOL = 1024;
static const u_int uMAX_VIS_SEGMENTS_TO_STORE_FROM_REFINEMENT = 100000; // (MP - 13.03.2009) This happens to be the minimum currently generated by any create step and should therefore be quite reasonable.

static Asura_ResizingMempool g_xVisibilitySegmentPool( uNUM_VIS_SEGMENTS_PER_POOL, sizeof(Tools_VisibilitySegment), __FILE__, __LINE__ );
static Asura_RingBuffer<Tools_VisibilitySegment> g_xRefinementVisSegmentBfr( uMAX_VIS_SEGMENTS_TO_STORE_FROM_REFINEMENT );

static Tools_ViewCell::FIND_SPLIT_PLANE_CALLBACK g_pfnFindSplitPlaneCallback = NULL;
static const Asura_LinkList<Tools_Visibility_SplitHintQuad>* g_pxSplitHintQuadList = NULL;

Tools_VisibilitySegment* Tools_VisibilitySegment::Create( bool bFromRefinement, const Asura_Vector_3& xP0, const Asura_Vector_3& xP1, u_int uVisibleFeatureID0, u_int uVisibleFeatureID1 )
{
	Tools_VisibilitySegment* pxSegment;
    if( bFromRefinement )
    {
        pxSegment = g_xRefinementVisSegmentBfr.Alloc_Overwrite();
    }
    else
    {
        pxSegment = static_cast<Tools_VisibilitySegment*>( g_xVisibilitySegmentPool.Allocate() );
    }

	if (pxSegment)
	{
		pxSegment->m_xPoint[0] = xP0;
		pxSegment->m_xPoint[1] = xP1;
		pxSegment->m_uVisibleFeatureID[0] = uVisibleFeatureID0;
		pxSegment->m_uVisibleFeatureID[1] = uVisibleFeatureID1;
	}

	return pxSegment;
}


void Tools_VisibilitySegment::DestroyAll( void )
{
    g_xVisibilitySegmentPool.ForceReset();
    g_xRefinementVisSegmentBfr.Reset();
}


const Asura_RingBuffer<Tools_VisibilitySegment>& Tools_VisibilitySegment::GetRefinementSegments( void )
{
    return g_xRefinementVisSegmentBfr;
}


Tools_ViewCell* Tools_ViewCell::Create( const Asura_Bounding_Box& xAABB, const Tools_VisibilitySegment** ppxCandidateList, u_int uNumberOfCandidates, u_int uDepth )
{
	Tools_ViewCell* pxCell = new Tools_ViewCell;

	pxCell->m_xAABB = xAABB;
	pxCell->m_pxLeftChild = 0;
	pxCell->m_pxRightChild = 0;
	pxCell->m_uDepth = uDepth;

	
	pxCell->m_uNumberOfSegments = 0;
	pxCell->m_ppxSegmentList = 0;

	pxCell->m_uLeafIndex = 0xffffffff;
	pxCell->m_pbVisibleFeatures = 0;
	pxCell->m_pbVisibleViewCells = 0;


	pxCell->m_bHasInternalLinks = false;
	
	if ( uNumberOfCandidates )
	{
		pxCell->m_ppxSegmentList = new const Tools_VisibilitySegment* [uNumberOfCandidates];
	}

	Asura_Assert( uNumberOfCandidates, "No candidates - an empty viewcell?" );

	if ( uNumberOfCandidates && ppxCandidateList )
	{
		for( u_int u = 0; u<uNumberOfCandidates; ++u )
		{
			const Tools_VisibilitySegment* pxSegment = ppxCandidateList[u];
	            	
			if ( pxCell->AABBContainsVisibilitySegment( pxSegment ) )
			{
				pxCell->AddVisibilitySegment( pxSegment );

				if (! pxCell->m_bHasInternalLinks )
				{
					if ( Tools_Visibility_System::IsExternalFeature( pxSegment->GetVisibleFeatureID(0) ) ) continue;
					if ( Tools_Visibility_System::IsExternalFeature( pxSegment->GetVisibleFeatureID(1) ) ) continue;

					pxCell->m_bHasInternalLinks = true;
				}
			}
		}
	}


	return pxCell;
}

void Tools_ViewCell::Destroy( Tools_ViewCell* pxCell )
{
	if (!pxCell) return;

	delete [] pxCell->m_ppxSegmentList;

	Destroy( pxCell->m_pxLeftChild );
	Destroy( pxCell->m_pxRightChild );

	delete [] pxCell->m_pbVisibleFeatures;

	delete [] pxCell->m_pbVisibleViewCells;

	delete pxCell;

}

void Tools_ViewCell::SplitBoundingBox( u_int uAxis, float fPlane, Asura_Bounding_Box& xLeft, Asura_Bounding_Box& xRight )
{
	xRight = GetAABB();
	xLeft = GetAABB();

	xLeft.m_fElement[ uAxis*2 +1 ] = fPlane;
	xRight.m_fElement[ uAxis*2  ] = fPlane;

}

void Tools_ViewCell::SetVisibleFeaturesArray( void )
{
	m_pbVisibleFeatures = new bool[ Tools_Visibility_System::GetNumberOfVisibilityFeatures() ];

	Asura_Mem::Set( m_pbVisibleFeatures, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(bool) );

	for ( u_int uInner = 0; uInner<m_uNumberOfSegments; ++uInner )
	{
		const Tools_VisibilitySegment* pxSegment = m_ppxSegmentList[uInner];

		for (u_int u=0; u<2; ++u)
		{
			u_int uFeatureID = pxSegment->GetVisibleFeatureID(u);
			m_pbVisibleFeatures[uFeatureID] = true;
		}
	}
}

bool Tools_ViewCell::AABBContainsVisibilitySegment( const Tools_VisibilitySegment* pxSegment )
{
	Asura_Vector_3 xAlpha( pxSegment->GetPoint(0) );
	Asura_Vector_3 xBeta( pxSegment->GetPoint(1) );

	return RaySegmentIntersectsViewCell( xAlpha, xBeta );
}

bool Tools_ViewCell::RaySegmentIntersectsViewCell( const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd ) const
{
	Asura_Vector_3 xBeta = xEnd - xStart;

	float fLimit = xBeta.Normalise();
	float fDistance = 2.0f;
	bool bHit = Asura_Maths::RayIntersectsBoundingBox( xStart, xBeta, m_xAABB, &fDistance );

	return ( bHit && fDistance <= fLimit );
}

void Tools_ViewCell::AddVisibilitySegment( const Tools_VisibilitySegment* pxSegment )
{
	m_ppxSegmentList[ m_uNumberOfSegments++ ] = pxSegment;
}


void Tools_ViewCell::CreateLeftChild( const Asura_Bounding_Box& xAABB )
{
	m_pxLeftChild = Create( xAABB, m_ppxSegmentList, m_uNumberOfSegments, m_uDepth+1 );
}

void Tools_ViewCell::CreateRightChild( const Asura_Bounding_Box& xAABB )
{
	m_pxRightChild = Create( xAABB, m_ppxSegmentList, m_uNumberOfSegments, m_uDepth+1 );
}

void Tools_ViewCell::DebugRender( void )
{
	
	if (GetLeftChild()) GetLeftChild()->DebugRender();

	if (GetRightChild())
	{
		GetRightChild()->DebugRender();
	}
	else
	{

		Asura_Vector_3 xPos; xPos.Zero();
		Asura_Matrix_3x3 xOri; xOri.Identity();
	    
		u_int uColour = 0xff007f00;

		Asura_Render::RenderBB( GetAABB(), xPos, xOri, uColour );
		//Asura_DebugLineRender::AddBoundingBox( GetAABB(), xPos, xOri, uColour );
	}

	return;
}


void Tools_ViewCell::DebugLogFile( void )
{
#if 0
	if (!GetLeftChild() && !GetRightChild() )
	{

		/////

		u_int* puVisibilities = new u_int[ Tools_Visibility_System::GetNumberOfVisibilityFeatures() ];

		Asura_Mem::Set( puVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );

		for ( u_int uInner = 0; uInner<m_uNumberOfSegments; ++uInner )
		{
			const Tools_VisibilitySegment* pxSegment = m_ppxSegmentList[uInner];

			for (u_int u=0; u<2; ++u)
			{
				u_int uFeatureID = pxSegment->GetVisibleFeatureID(u);
				++(puVisibilities[uFeatureID]);
			}
		}

		u_int uVisibleStuff = 0;
		for ( u_int u=0; u<Tools_Visibility_System::GetNumberOfVisibilityFeatures(); ++u )
		{
			if (puVisibilities[u]) uVisibleStuff++;
		}

		delete [] puVisibilities;

		///
		if (uVisibleStuff > 3)
		{
			Asura_Logfile_Retail(( ASURA_VISIBILITY_LOGFILE, "Node depth: %d. Visible features: %d", m_uDepth, uVisibleStuff ));
		}
	}

	

	if (GetLeftChild()) GetLeftChild()->DebugLogFile();

	if (GetRightChild()) GetRightChild()->DebugLogFile();

#endif
	return;
}
/* 

static u_int g_uSortAxis;
int SortOnMinValue( const void *pSegment1, const void *pSegment2 )
{
	// sort so lowest min value comes first

	const Tools_VisibilitySegment* pxSegment1 = *((const Tools_VisibilitySegment**) pSegment1);
	const Tools_VisibilitySegment* pxSegment2 = *((const Tools_VisibilitySegment**) pSegment2);

	if ( pxSegment1->GetMin(g_uSortAxis) > pxSegment2->GetMin(g_uSortAxis) )
	{
		return 1;
	}
	else if ( pxSegment1->GetMin(g_uSortAxis) < pxSegment2->GetMin(g_uSortAxis) )
	{
		return -1;
	}
	return 0;
}
int SortOnMaxValue( const void *pSegment1, const void *pSegment2 )
{
	// sort so highest max value comes first

	const Tools_VisibilitySegment* pxSegment1 = *((const Tools_VisibilitySegment**) pSegment1);
	const Tools_VisibilitySegment* pxSegment2 = *((const Tools_VisibilitySegment**) pSegment2);

	if ( pxSegment1->GetMax(g_uSortAxis) < pxSegment2->GetMax(g_uSortAxis) )
	{
		return 1;
	}
	else if ( pxSegment1->GetMax(g_uSortAxis) > pxSegment2->GetMax(g_uSortAxis) )
	{
		return -1;
	}
	return 0;
}
void SortOnMinOrder( const Tools_VisibilitySegment** ppxSegments, u_int uNumberOfSegments, u_int uSortAxis )
{
	g_uSortAxis = uSortAxis;
	qsort( ppxSegments, uNumberOfSegments, sizeof(const Tools_VisibilitySegment*), SortOnMinValue );
}
void SortOnMaxOrder( const Tools_VisibilitySegment** ppxSegments, u_int uNumberOfSegments, u_int uSortAxis )
{
	g_uSortAxis = uSortAxis;
	qsort( ppxSegments, uNumberOfSegments, sizeof(const Tools_VisibilitySegment*), SortOnMaxValue );
}

*/


void Tools_ViewCell::SetFindSplitPlaneCallback( FIND_SPLIT_PLANE_CALLBACK pfnFindSplitPlaneCallback )
{
    g_pfnFindSplitPlaneCallback = pfnFindSplitPlaneCallback;
}


void Tools_ViewCell::SetSplitHintQuadList( const Asura_LinkList<Tools_Visibility_SplitHintQuad>* pxList )
{
    g_pxSplitHintQuadList = pxList;
}


bool Tools_ViewCell::FindBestSplittingPlane( u_int& uBestAxis, float& fBestPlane, bool& bForceSplit )
{
#if 01

	uBestAxis   = 0;
	fBestPlane  = 0.0f;
    bForceSplit = false;

    // Check if any callback wants to force a specific plane to be used.
    if( g_pfnFindSplitPlaneCallback && g_pfnFindSplitPlaneCallback(this, uBestAxis, fBestPlane, bForceSplit) )
    {
        return true;
    }

    // Next, check if there are any initial splits that are relevant for
    // this node.
    if( CheckSplitHintQuadList(uBestAxis, fBestPlane, bForceSplit) )
    {
        return true;
    }

	bool bSolution = false;

	float fBestScore = Asura_Maths::LargeFloat;
	
	// KJFL - 6 Aug 2007
	// Allocating these arrays here rather than passing in two pointers (i.e. continually reusing the same buffers)
	// creates faster code - presumably to do with the compiler knowing it can avoid aliasing.
	u_int* puLeftVisibilities = new u_int [Tools_Visibility_System::GetNumberOfVisibilityFeatures()];
	u_int* puRightVisibilities = new u_int [Tools_Visibility_System::GetNumberOfVisibilityFeatures()];

//	Asura_Mem::Set( puLeftVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );
//	Asura_Mem::Set( puRightVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );

	const u_int uDivisions = 100;//m_uNumberOfSegments;
	for ( u_int uOuter = 0; uOuter<uDivisions; ++uOuter )
	{
		// first test is actually against the midpoint, to get a fair upper score limit to aid 
		// early rejections (and testing 0 is pointless anyway since it's not a split)
		const float fScale = (uOuter==0) ? 0.5f : ((float)uOuter)/((float)uDivisions);
		
		for ( u_int uAxis=0; uAxis<3; ++uAxis )
		{	
			const float fMin = m_xAABB.m_fElement[ uAxis*2 + 0 ];
			const float fMax = m_xAABB.m_fElement[ uAxis*2 + 1 ];

			// don't split small across small extents
			if ( fMax - fMin < fTOOLS_VIEWCELL_MIN_EXTENT * 2.0f ) continue;

			Asura_Mem::Set( puLeftVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );
			Asura_Mem::Set( puRightVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );

			const float fTestPlane = (fMax - fMin) * fScale + fMin;


			Asura_Bounding_Box xLeft;
			Asura_Bounding_Box xRight;
			SplitBoundingBox( uAxis, fTestPlane, xLeft, xRight );

			const float fLeftScoreScale = xLeft.GetSurfaceArea();
			const float fRightScoreScale = xRight.GetSurfaceArea();

			float fLeftScore = 0;
			float fRightScore = 0;
			float fOverallScore = 0;
		
            for ( u_int uInner = 0; uInner<m_uNumberOfSegments; ++uInner )
			{
				const Tools_VisibilitySegment* pxSegment = m_ppxSegmentList[uInner];

				const u_int uFeatureID0 = pxSegment->GetVisibleFeatureID(0);
				const u_int uFeatureID1 = pxSegment->GetVisibleFeatureID(1);


				float fFeature0Score = ( uFeatureID0 != Tools_Visibility_System::GetSkyBoxFeatureID() ) ? fNORMAL_FEATURE_SCORE : fSKYBOX_SCORE;
				float fFeature1Score = ( uFeatureID1 != Tools_Visibility_System::GetSkyBoxFeatureID() ) ? fNORMAL_FEATURE_SCORE : fSKYBOX_SCORE;
	
				if ( pxSegment->GetMin(uAxis) < fTestPlane )
				{
					if ( puLeftVisibilities[uFeatureID0] == 0 )
					{
						fLeftScore += fFeature0Score;
					}
					++puLeftVisibilities[uFeatureID0];

					if ( puLeftVisibilities[uFeatureID1] == 0 )
					{
						fLeftScore += fFeature1Score;
					}
					++puLeftVisibilities[uFeatureID1];
				}

				if ( pxSegment->GetMax(uAxis) > fTestPlane )
				{
					if ( puRightVisibilities[uFeatureID0] == 0 )
					{
						fRightScore += fFeature0Score;
					}
					++puRightVisibilities[uFeatureID0];

					if ( puRightVisibilities[uFeatureID1] == 0 )
					{
						fRightScore += fFeature1Score;
					}
					++puRightVisibilities[uFeatureID1];
				}

				const float fLeftScore2 = (fLeftScore) * fLeftScoreScale;
				const float fRightScore2 = (fRightScore) * fRightScoreScale;

				fOverallScore  = fLeftScore2*fLeftScore2 + fRightScore2*fRightScore2;

				if (fOverallScore > fBestScore)
				{
					// give up, it can only get worse
					break;
				}
			}

			if ( fOverallScore < fBestScore )
			{
				bool bStopSplit = true;

				// don't make empty viewcells
				if ( fLeftScore == 0.0f || fRightScore == 0.0f )
				{
			
				}
				else
				{
					// don't split cells if the vis data is identical
					u_int uDifferences = 0;
					for ( u_int u=0; u<Tools_Visibility_System::GetNumberOfVisibilityFeatures(); ++u )
					{
						if ( (puLeftVisibilities[u]!=0) != (puRightVisibilities[u]!=0) )
						{
							++uDifferences;

							if ( uDifferences >= uMIN_NUMBER_OF_DIFFERENCES_TO_ALLOW_SPLIT )
							{
								bStopSplit = false;
								break;
							}
						}
					}


				}

				if ( !bStopSplit  )
				{
					fBestScore = fOverallScore;
					uBestAxis = uAxis;
					fBestPlane = fTestPlane; 
					bSolution = true;			
				}
			}
		}
	}
	
	delete [] puLeftVisibilities;
	delete [] puRightVisibilities;

	return bSolution;
#else
	bool bSolution = false;

	float fBestScore = Asura_Maths::LargeFloat;
	uBestAxis = 0;
	fBestPlane = 0.0f;
	
	// KJFL - 6 Aug 2007
	// Allocating these arrays here rather than passing in two pointers (i.e. continually reusing the same buffers)
	// creates faster code - presumably to do with the compiler knowing it can avoid aliasing.
	u_int* puLeftVisibilities = new u_int [Tools_Visibility_System::GetNumberOfVisibilityFeatures()];
	u_int* puRightVisibilities = new u_int [Tools_Visibility_System::GetNumberOfVisibilityFeatures()];

	Asura_Mem::Set( puLeftVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );
	Asura_Mem::Set( puRightVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );

	for ( u_int uAxis=0; uAxis<3; ++uAxis )
	{
		const float fMin = m_xAABB.m_fElement[ uAxis*2 + 0 ];
		const float fMax = m_xAABB.m_fElement[ uAxis*2 + 1 ];

		const u_int uDivisions = 100;//m_uNumberOfSegments;
		for ( u_int uOuter = 0; uOuter<uDivisions; ++uOuter )
		{	
			const float fScale = ((float)uOuter)/((float)uDivisions);

			const float fTestPlane = (fMax - fMin) * fScale + fMin;

			for ( u_int uInner = 0; uInner<m_uNumberOfSegments; ++uInner )
			{
				const Tools_VisibilitySegment* pxSegment = m_ppxSegmentList[uInner];

				const u_int uFeatureID0 = pxSegment->GetVisibleFeatureID(0);
				const u_int uFeatureID1 = pxSegment->GetVisibleFeatureID(1);

				if ( pxSegment->GetMin(uAxis) < fTestPlane )
				{
					++puLeftVisibilities[uFeatureID0];
					++puLeftVisibilities[uFeatureID1];
				}

				if ( pxSegment->GetMax(uAxis) > fTestPlane )
				{
					++puRightVisibilities[uFeatureID0];
					++puRightVisibilities[uFeatureID1];
				}
			}

			u_int uLeftScore = 0;
			u_int uRightScore = 0;
			bool bMatches = true;

			for ( u_int u=0; u<Tools_Visibility_System::GetNumberOfVisibilityFeatures(); ++u )
			{
				if ( (puLeftVisibilities[u]!=0) != (puRightVisibilities[u]!=0) )
				{
					bMatches = false;
				}
	#if 01
				if ( puLeftVisibilities[u] ) uLeftScore += 1;
				if ( puRightVisibilities[u] ) uRightScore += 1;
	#else
				uLeftScore += puLeftVisibilities[u];
				uRightScore += puRightVisibilities[u];
	#endif

			}
			Asura_Mem::Set( puLeftVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );
			Asura_Mem::Set( puRightVisibilities, 0, Tools_Visibility_System::GetNumberOfVisibilityFeatures() * sizeof(u_int) );


			// KJFL - I originally thought of if (!bMatches) at this point, but it turned out slower!
			// Generally the !bMatches is true, so it seems to generate faster code if we test it
			// at the last minute rather than stalling and checking it earlier.
			{
				Asura_Bounding_Box xLeft;
				Asura_Bounding_Box xRight;

				SplitBoundingBox( uAxis, fTestPlane, xLeft, xRight );

				const float fLeftScore = (float)uLeftScore * xLeft.GetSurfaceArea();
				const float fRightScore = (float)uRightScore * xRight.GetSurfaceArea();

				const float fOverallScore = fLeftScore*fLeftScore + fRightScore*fRightScore;

				if ( !bMatches && fOverallScore < fBestScore )
				{
					fBestScore = fOverallScore;
					uBestAxis = uAxis;
					fBestPlane = fTestPlane; 
					bSolution = true;			
				}
			}
		}
	}
	
	delete [] puLeftVisibilities;
	delete [] puRightVisibilities;

	return bSolution;
#endif
}


bool Tools_ViewCell::CheckSplitHintQuadList( u_int& uBestAxis, float& fBestPlane, bool& bForceSplit )
{
    if( !g_pxSplitHintQuadList )
    {
        return false;
    }

    float fBestScore = 0.0f;
    for( Asura_LinkList_ConstIt<Tools_Visibility_SplitHintQuad> xQuadIt(g_pxSplitHintQuadList); !xQuadIt.Done(); xQuadIt.Next() )
    {
        const Tools_Visibility_SplitHintQuad* pxHintQuad = xQuadIt.GetCurrent();

        if( !m_xAABB.Intersects(pxHintQuad->m_xAABB) )
        {
            continue;
        }

        // Don't split where the node we're examining extends a long way
        // outside the hint quad in the axes other than the one the hint
        // applies to. These hints are evaluated before anything else and we
        // don't want to end up splitting the whole initial environment AABB
        // against a small hint quad in the corner, for example.
        {
            bool bTooFarOutside = false;

            for( u_int uAxis = 0; uAxis < 3; ++uAxis )
            {
                if( uAxis == pxHintQuad->m_uAxis ) { continue; }

                const float fMAX_NODE_EXTENT_OUTSIDE = 50.0f;
                if( (pxHintQuad->m_xAABB.m_fElement[uAxis * 2] - m_xAABB.m_fElement[uAxis * 2] > fMAX_NODE_EXTENT_OUTSIDE)
                    ||
                    (m_xAABB.m_fElement[uAxis * 2 + 1] - pxHintQuad->m_xAABB.m_fElement[uAxis * 2 + 1] > fMAX_NODE_EXTENT_OUTSIDE) )
                {
                    bTooFarOutside = true;
                    break;
                }

            }

            if( bTooFarOutside ) { continue; }
        }

		const float fBBMin = m_xAABB.m_fElement[pxHintQuad->m_uAxis * 2 + 0];
		const float fBBMax = m_xAABB.m_fElement[pxHintQuad->m_uAxis * 2 + 1];

        // Ensure we don't end up splitting at the same plane twice due to
        // numerical inaccuracies.
        const float fINTERSECTION_EPSILON = 0.01f;
        if( (fBBMin < pxHintQuad->m_fAxisValue - fINTERSECTION_EPSILON) && (fBBMax > pxHintQuad->m_fAxisValue + fINTERSECTION_EPSILON) )
        {
            // Take the plane that causes the most significant split if
            // there's more than one candidate.
            float fScore = Asura_Maths::Min( pxHintQuad->m_fAxisValue - fBBMin, fBBMax - pxHintQuad->m_fAxisValue );
            if( fScore > fBestScore )
            {
                uBestAxis   = pxHintQuad->m_uAxis;
                fBestPlane  = pxHintQuad->m_fAxisValue;
                bForceSplit = true;
                fBestScore  = fScore;
            }
        }
    }

    return( fBestScore > 0.0f );
}


void Tools_ViewCell::GetVisibleFeatures( u_char* pucVisibleFeaturesArray, u_char ucBitPattern ) const
{
	if (!m_pbVisibleFeatures) return;

	for (u_int u=0; u<Tools_Visibility_System::GetSkyBoxFeatureID(); ++u)
	{
		if (m_pbVisibleFeatures[u])
		{
			pucVisibleFeaturesArray[u] |= ucBitPattern;
		}
	}
#if 0
	for ( u_int uInner = 0; uInner<m_uNumberOfSegments; ++uInner )
	{
		const Tools_VisibilitySegment* pxSegment = m_ppxSegmentList[uInner];

		bool bInsideLink = (pxSegment->GetVisibleFeatureID(0) != Tools_Visibility_System::GetSkyBoxFeatureID() && pxSegment->GetVisibleFeatureID(1) != Tools_Visibility_System::GetSkyBoxFeatureID());
		u_int uColour = bInsideLink ? 0x1fffff00 : 0x1fff0000;
		Asura_Render::RenderLine( pxSegment->GetPoint(0), pxSegment->GetPoint(1), uColour );
	}
#endif

	Asura_Vector_3 xPos; xPos.Zero();
	Asura_Matrix_3x3 xOri; xOri.Identity();
    
	u_int uColour = 0xff00ff00;

	Asura_Render::RenderBB( GetAABB(), xPos, xOri, uColour );

}

void Tools_ViewCell::RefineVisibility( u_int uFeatureID0, u_int uFeatureID1 )
{
	if ( m_pbVisibleFeatures && !(m_pxLeftChild||m_pxRightChild) )
	{
		m_pbVisibleFeatures[uFeatureID0] = true;
		m_pbVisibleFeatures[uFeatureID1] = true;
	}
	return;
}

void Tools_ViewCell::SetLeafIndex( u_int uLeafIndex )
{
	// TODO - could check not already set if required
	m_uLeafIndex = uLeafIndex;
}


void Tools_ViewCell::SetLeavesAsVisible( Tools_ViewCell** ppxLeafCells, u_int uNumberOfLeaves, u_int uTotalNumberOfLeaves )
{
	if ( !m_pbVisibleViewCells )
	{
		m_pbVisibleViewCells = new bool [ uTotalNumberOfLeaves ];
		Asura_Mem::Set( m_pbVisibleViewCells, 0, uTotalNumberOfLeaves * sizeof(bool) );
	}

	for ( u_int u=0; u<uNumberOfLeaves; ++u )
	{
		u_int uLeafIndex = ppxLeafCells[u]->GetLeafIndex();

		Asura_Assert( uLeafIndex < uTotalNumberOfLeaves, "Invalid leaf index" );

		if ( uLeafIndex < uTotalNumberOfLeaves )
		{
			m_pbVisibleViewCells[ uLeafIndex ] = true;
		}
	}
}


#endif // #ifdef ASURA_USE_NEW_VISIBILITY_SYSTEM

