//////////////////////////////////////////////////////////////////////////////
//
// Tools_Mode_Waypointing.h
//
// Copyright (c) 2005 Rebellion. All rights reserved.
//
// Nick Davis
//
//////////////////////////////////////////////////////////////////////////////

// B E G I N   F I L E   W R A P P E R ///////////////////////////////////////

#ifdef ASURA_USE_PRAGMA_ONCE
    #pragma once
#endif

#ifndef ___TOOLS_MODE_WAYPOINTING_H___
#define ___TOOLS_MODE_WAYPOINTING_H___ 1

// E N D   F I L E   W R A P P E R ///////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////////
//
// includes
//
//////////////////////////////////////////////////////////////////////////////

// asura
#include "AI/Asura_AI_Navigation.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Hash.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_Multithread.h"
#include "Mesh/Asura_Mesh_PolygonFlags.h"
#include "Physics/Asura_Physics_Object.h"

// tools
#include "Tools_Mode.h"
#include "Misc/Tools_DynamicArray.h"

//////////////////////////////////////////////////////////////////////////////
//
// forward declarations
//
//////////////////////////////////////////////////////////////////////////////

class Asura_Collision_Box;
class Asura_Collision_Object;
class Asura_Dialog;
template < class T > class Asura_List;
class Asura_Overlay;
class Asura_Physics_Object;
class Asura_Physics_RigidBodyInstance;
class Asura_ServerEntity_PhysicalObject;
class Asura_Tabs_Dialog;

class Tools_ModeDataFile;
class Tools_Waypoint_Analysis;
class Tools_Waypoint_Editing;
class Tools_Waypoint_Object;

//////////////////////////////////////////////////////////////////////////////
//
// support classes
//
//////////////////////////////////////////////////////////////////////////////

// These flags are only used during the waypointing process in tools.
// When the navigation data is copied across to the navigation system these are discarded.
// These must have values only in the top 16 bits of the int, so they don't clash with the Asura connection flags.
enum TOOLS_CONNECTION_FLAGS
{
    // If we need engine flags, they should start from here:
    // TOOLS_CONNECTION_FLAG_SOME_ENGINE_FLAG = 0x00010000,

    TOOLS_CONNECTION_FLAG_PROJECT_SPECIFIC  = 0x01000000,
};

class Tools_Nav_Link
{
public:
    Tools_Nav_Link()
    : m_iIndex( 0 )
    , m_uFlags( 0 )
    , m_fLength( 0.0f )
    {}

public:
    int          m_iIndex;
    unsigned int m_uFlags;
    float        m_fLength;
};

// These flags are only used during the waypointing process in tools.
// When the navigation data is copied across to the navigation system these are discarded.
// These must have values only in the top 16 bits of the int, so they don't clash with the Asura nav vertex flags.
enum TOOLS_NAVVERTEX_FLAGS
{
    TOOLS_NAVVERTEX_FLAG_DENSE					= 0x00010000,
    TOOLS_NAVVERTEX_FLAG_ALREADY_VISITED		= 0x00020000,
    TOOLS_NAVVERTEX_FLAG_BELONGS_TO_THIS_ENV	= 0x00040000, // Only relevant in streaming builds
	TOOLS_NAVVERTEX_FLAG_FLOOR					= 0x00080000,
	
	TOOLS_NAVVERTEX_FLAG_PROJECT_SPECIFIC 		= 0x01000000,
};

class Tools_Nav_Vertex
{
public:
    Tools_Nav_Vertex()
    : m_xPosition( Asura_Maths::ZeroVector3 )
	, m_xNormal(0.0f, -1.0f, 0.0f)
    , m_fRadius( Asura_Maths::LargeFloat )
    , m_xLinks()
    , m_pxCoverpointIndices( NULL )
    , m_iNumCoverpointIndices( 0 )
    , m_iModuleIndex( 0 )
    , m_iFlags( TOOLS_NAVVERTEX_FLAG_BELONGS_TO_THIS_ENV )
    , m_iZoneID( 0 )
	, m_uUDWID( Asura_AI_Navigation_System::UserDefinedWaypoint::uINVALID_ID )
#ifdef ASURA_USE_STREAMING
	, m_iInterZonePairIndex( -1 )
#endif
    {}

    virtual ~Tools_Nav_Vertex()
    {
        delete [] m_pxCoverpointIndices;
    }

	bool IsFloor()				const { return ( ( m_iFlags & TOOLS_NAVVERTEX_FLAG_FLOOR	   			) != 0 ); }
    bool IsSlope()              const { return ( ( m_iFlags & ASURA_AI_NAVVERTEX_FLAG_SLOPE   			) != 0 ); }
    bool IsSteps()              const { return ( ( m_iFlags & ASURA_AI_NAVVERTEX_FLAG_STEPS   			) != 0 ); }
    bool IsCeiling()            const { return ( ( m_iFlags & ASURA_AI_NAVVERTEX_FLAG_CEILING   		) != 0 ); }
    bool IsWall()               const { return ( ( m_iFlags & ASURA_AI_NAVVERTEX_FLAG_WALL   			) != 0 ); }
    bool IsDense()				const { return ( ( m_iFlags & TOOLS_NAVVERTEX_FLAG_DENSE				) != 0 ); }
	bool GetBelongsToThisEnv()	const { return ( ( m_iFlags & TOOLS_NAVVERTEX_FLAG_BELONGS_TO_THIS_ENV	) != 0 ); }
	int	 GetNumLinks()			const { return m_xLinks.GetSize(); }
	u_int GetUDWPID()			const { return m_uUDWID; }
	bool  IsUserDefined()		const { return ( m_uUDWID != Asura_AI_Navigation_System::UserDefinedWaypoint::uINVALID_ID ); }

    Tools_Nav_Vertex& operator=( const Tools_Nav_Vertex& xRhs )
    {
        if ( this != &xRhs )
        {
            delete [] m_pxCoverpointIndices;
            m_pxCoverpointIndices = NULL;

            m_xPosition = xRhs.m_xPosition;
			m_xNormal	= xRhs.m_xNormal;
            m_fRadius   = xRhs.m_fRadius;
			m_xLinks	= xRhs.m_xLinks;

            m_iNumCoverpointIndices = xRhs.m_iNumCoverpointIndices;
            if ( m_iNumCoverpointIndices )
            {
                m_pxCoverpointIndices = new unsigned short[ m_iNumCoverpointIndices ];
                for ( int i = 0; i < m_iNumCoverpointIndices; ++i )
                {
                    m_pxCoverpointIndices[ i ] = xRhs.m_pxCoverpointIndices[ i ];
                }
            }
            m_iModuleIndex = xRhs.m_iModuleIndex;
            m_iFlags       = xRhs.m_iFlags;
            m_iZoneID      = xRhs.m_iZoneID;
			m_uUDWID       = xRhs.m_uUDWID;
#ifdef ASURA_USE_STREAMING
			m_iInterZonePairIndex = xRhs.m_iInterZonePairIndex;
#endif
        }

        return *this;
    }

public:
    Asura_Vector_3							m_xPosition;
	Asura_Vector_3							m_xNormal;
    float									m_fRadius;
    Tools_DynamicArray< Tools_Nav_Link >	m_xLinks;
    unsigned short*							m_pxCoverpointIndices;
    int										m_iNumCoverpointIndices;
    int										m_iModuleIndex;
	// cf ASURA_AI_NAVVERTEX_FLAG and TOOLS_NAVVERTEX_FLAG enums.
    int										m_iFlags;
    int										m_iZoneID;
	// User-defined waypoint ID that this waypoint has been generated from (if any).
	u_int									m_uUDWID;

#ifdef ASURA_USE_STREAMING
	int										m_iInterZonePairIndex;
#endif
};

class Tools_Nav_Coverpoint
{
public:
    Tools_Nav_Coverpoint()
    : m_bValid( true ),
      m_xCentre( Asura_Maths::ZeroVector3 ),
      m_fRadius( 0.0f ),
      m_usFlags( 0 ),
      m_ucShortExits( 0 ),
      m_ucLongExits( 0 ),
      m_iNearestWaypoint( -1 )
    {}

public:
    bool           m_bValid;
    Asura_Vector_3 m_xCentre;
    float          m_fRadius;
    unsigned short m_usFlags;
    unsigned char  m_ucShortExits;
    unsigned char  m_ucLongExits;
    int            m_iNearestWaypoint;
};

//////////////////////////////////////////////////////////////////////////////
//
// Tools_Mode_Waypointing
//
// To create a project specific version of the class add code to your version
// of Tools_System::Project_ChangeToNextMode to handle the TOOLS_MODE_WAYPOINTING
// case.
//
//////////////////////////////////////////////////////////////////////////////

class Tools_Mode_Waypointing : public Tools_Mode
{
    friend class Tools_Batch_Waypointing;
    friend class Tools_Waypoint_Analysis;
    friend class Tools_Waypoint_Editing;
    friend class Tools_Waypoint_Selection;
	friend class Asura_Chunk_WaypointingSettings;
	friend class Asura_Container_Functor_RemoveBlockedWaypoints;

public:
    Tools_Mode_Waypointing()  { s_pxThis = this; }
    virtual ~Tools_Mode_Waypointing() { s_pxThis = NULL; }

	virtual void InitialiseMode();
	virtual void ShutdownMode();

    virtual void Update();
    virtual void RenderWorld_PostToneMap();
	virtual void RenderUI();
    virtual bool IsModified();
    virtual void QuickSave();

    virtual void HandleMeshChange(int iModuleIndex);

    static Tools_Mode_Waypointing* GetInstance() { return s_pxThis; }

public:
    // call these functions to set the waypoint/coverpoint generation parameters
    // these parameters must be set up before Initialise is called on this class -
    // somewhere like Tools_System::Project_Initialise()

    static void SetIsBatchWaypointing( const bool bIsBatchWaypointing )         { s_bIsBatchWaypointing = bIsBatchWaypointing; }
    // this is the square of the min size that polys are subdivided to when creating waypoints from vertices
	static void SetMinSubdivideSizeSq( const float fSize )                      { s_fMinSubdivideSizeSq = fSize; }
    static void SetMinDenseSubdivideSizeSq( const float fSize )                 { s_fMinDenseSubdivideSizeSq = fSize; }
    // this is the max level of recursion allowed when subdividing polys and creating waypoints
    static void SetMaxSubdivideRecursionLevel( const int iLevel )               { s_iMaxSubdivideRecursionLevel = iLevel; }
	// this is the min clearance radius between a waypoint and an obstructing static object or door
    // it should normally be set to approx half the width of the smallest waypoint user's BB in game
    static void SetMinWaypointClearanceRadius( const float fRadius )            { s_fMinWaypointClearanceRadius = fRadius; }
    static void SetMinWaypointRadius( const float fRadius )                     { s_fMinWaypointRadius = fRadius; }
    static void SetMaxWaypointRadius( const float fRadius )                     { s_fMaxWaypointRadius = fRadius; }
    static void SetMinDenseWaypointRadius( const float fRadius )                { s_fMinDenseWaypointRadius = fRadius; }
    static void SetMaxDenseWaypointRadius( const float fRadius )                { s_fMaxDenseWaypointRadius = fRadius; }
    // the amount by which waypoints are allowed to overlap adjacent waypoints as a fraction of the
    // overlapping waypoint's radius
    // (note that there will always be some waypoint overlapping regardless of this value)
    static void SetWaypointOverlapFraction( const float fFraction )                 { s_fWaypointOverlapFraction = fFraction; }
    // Waypoints will not expand onto adjacent polys greater than this dot away from the waypoint's normal, even if they are flagged for waypoints.
    static void SetMaxDotProductBetweenWaypointPolygonNormals(const float fValue)   { s_fMaxDotProductBetweenWaypointPolygonNormals = fValue; }
    // Allows projects to set their own thread function to calculate the waypoint volumes. This must handle the thread work as the standard function does.
    static void SetWaypointVolumeCalcThreadFunction( void ( *pfnFunction )( const void* pParams ) ) { s_pfnWaypointCalcThreadFn = pfnFunction; }

    // The collision box to use when culling physically invalid waypoints and coverpoints, and culling links with physics.
	static void SetCollisionProxy( Asura_Physics_Object::COLLISION_PROXY eProxy )	{ s_eCollisionProxy = eProxy; }
    static void SetTestCollisionBox( const Asura_Collision_Box& xCollisionBox );
    static void SetTestCollisionBox( const Asura_Bounding_Box& xBoundingBox );
	static void SetCrouchCollisionBox( const Asura_Collision_Box& xCollisionBox );
    static void SetCrouchCollisionBox( const Asura_Bounding_Box& xBoundingBox );
	static void SetProneCollisionBox( const Asura_Collision_Box& xCollisionBox );
    static void SetProneCollisionBox( const Asura_Bounding_Box& xBoundingBox );
	static void EnableCrouch()												    { s_bCrouchEnabled = true; }
	static void EnableProne()												    { s_bProneEnabled = true; }
    // Set the speed at which the physics travels along a link to test it for validity.
    static void SetLinkTraversalSpeed( const float fSpeed )                     { s_fLinkTraversalSpeed = fSpeed; }
    // Sets whether the physics block should keep its forward momentum during a fall or fall directly downwards when traversing links.
    static void SetFallDirectlyDownwards( const bool bFallDirectlyDownwards )   { s_bFallDirectlyDownwards = bFallDirectlyDownwards; }
    // The direction an object placed on a waypoint will move if it needs to be resolved out of the environment.
    static void SetPlacedObjectMovementScale( const float fScale )				{ s_fPlacedObjectMovementScale = fScale; }
    // The distance an object placed on a waypoint is allowed to move without collision before the waypoint is culled.
    static void SetPlacedObjectAllowedMovement( const float fDistance )         { s_fPlacedObjectAllowedMovement = fDistance; }
    static void SetMinLinkLength( const float fLength )                         { s_fMinLinkLength = fLength; }
    static void SetMaxLinkLength( const float fLength )                         
	{ 
		if ( fLength > ASURA_AI_MAX_LINK_LENGTH )
		{
			Asura_Assert( false, "Tools_Mode_Waypointing::SetMaxLinkLength - Attempting to set max length to greater than possible limit. Setting to limit." );
			s_fMaxLinkLength = ASURA_AI_MAX_LINK_LENGTH;
		}
		else
		{
			s_fMaxLinkLength = fLength; 
		}
	}
    // the amount by which links are allowed to overlap intervening waypoints en route from one
    // waypoint to the next as a fraction of the overlapped waypoint's radius
    static void SetLinkOverlapFraction( const float fFraction )                 { s_fLinkOverlapFraction = fFraction; }
    // the distance beyond which a link is considered to be a fall link
    static void SetFallLinkDistance( const float fDistance )                    { s_fFallLinkDistance = fDistance; }
    // the distance beyond which a link is considered to be a long fall link
    static void SetLongFallLinkDistance( const float fDistance )                { s_fLongFallLinkDistance = fDistance; }
    // The maximum distance physics blocks are allowed to fall along links before the links are culled.
    static void SetMaxLinkFallDistance( const float fDistance )                 
	{ 
		if ( fDistance > ASURA_AI_MAX_LINK_LENGTH )
		{
			Asura_Assert( false, "Tools_Mode_Waypointing::SetMaxLinkFallDistance - Attempting to set max length to greater than possible limit. Setting to limit" );
			s_fMaxLinkLength = ASURA_AI_MAX_LINK_LENGTH;
		}
		else
		{
			s_fMaxLinkFallDistance = fDistance; 
		}
	}
    // The distance we allow the 'can reach place' physics test to be out by and still pass.
    static void SetCanReachDistance( const float fDistance )                        { s_fCanReachDistance = fDistance; }
    // The distance (along the three axes) we allow the 'MoveToDestination' test to be out by and still pass.
    static void SetMovedObjectAllowedOffset( const Asura_Vector_3& xAllowed )       { s_xMovedObjectAllowedOffset = xAllowed; }
    // Whether the gravity of the link testing physics object is set into the start position's suface normal or not.
    static void SetGravityFollowsSurfaceNormals(const bool bSet)				    { s_bGravityFollowsSurfaceNormals = bSet; }
    // Links between waypoints whose normals differ by more than this dot will be automatically culled.
    static void SetMaxDotProductForLinkBetweenWaypointNormals(const float fValue)   { s_fMaxDotProductForLinkBetweenWaypointNormals = fValue; }

    static void SetShouldGenerateCoverpoints( const bool bShouldGen )           { s_bShouldGenCoverpoints = bShouldGen; }
    // the maximum radius of waypoints that can become coverpoints
    static void SetMaxCoverpointRadius( const float fMaxRadius )                { s_fMaxCoverpointRadius = fMaxRadius; }
    static void SetMinCoverpointRadius( const float fMinRadius )                { s_fMinCoverpointRadius = fMinRadius; }
	static void SetShouldCalculateWaypointZoningData( const bool bShouldCalc )  { s_bShouldCalcZoningData = bShouldCalc; }
	static void SetShouldCullOneWayLinksWithNoReturn( const bool bValue )		{ s_bShouldCullOneWayLinksWithNoReturn = bValue; }

	static void SetWallBlockHeight( const float fHeight )						{ s_fWallBlockHeight = fHeight; }
	static void SetMaxCoverpointDistance( const float fDist )					{ s_fMaxCoverPointDistanceSq = fDist * fDist; }
	static void SetMaxCoverpointsPerWaypoint( const int iMaxCPs )				{ s_iMaxCoverpointsPerWaypoint = iMaxCPs; }
	static void SetMinLoneWaypointRadius( const float fRadius )					{ s_fMinLoneWaypointRadius = fRadius; }
	static void SetMaxJumpImpulse( const float fMaxJumpImpulse )				{ s_fMaxJumpImpulse = fMaxJumpImpulse; }
	static void SetWaypointToCoverpointMaxDist( const float fDist )				{ s_fWaypointToCoverpointMaxDist = fDist; }

	static void SetCalculateCoverpointExits( const bool bValue )				{ s_bCalculateCoverpointExits = bValue; }
	static void SetCliffStopWaypoints( const bool bValue )						{ s_bCliffStopWaypoints = bValue; }
	static void SetDontCullLinkedToWaypoints( const bool bValue )				{ s_bDontCullLinkedToWaypoints = bValue; }


	static float GetMinSubdivideSizeSq()										{ return s_fMinSubdivideSizeSq; }
    static float GetMinDenseSubdivideSizeSq()									{ return s_fMinDenseSubdivideSizeSq; }
    static int   GetMaxSubdivideRecursionLevel()								{ return s_iMaxSubdivideRecursionLevel; }
	static float GetMinWaypointClearanceRadius()								{ return s_fMinWaypointClearanceRadius; }
	static float GetMinWaypointRadius()											{ return s_fMinWaypointRadius; }
	static float GetMaxWaypointRadius()											{ return s_fMaxWaypointRadius; }
    static float GetMinDenseWaypointRadius()                                    { return s_fMinDenseWaypointRadius; }
    static float GetMaxDenseWaypointRadius()                                    { return s_fMaxDenseWaypointRadius; }

    static const Asura_Collision_Box& GetTestCollisionBox()                     { return s_xTestCollisionBox; }
	static float GetMinLinkLength()												{ return s_fMinLinkLength; }
	static float GetMaxLinkLength()												{ return s_fMaxLinkLength; }
	static float GetMaxLinkFallDistance()										{ return s_fMaxLinkFallDistance; }
    static float GetPlacedObjectAllowedMovement()								{ return s_fPlacedObjectAllowedMovement; }
    static const Asura_Vector_3& GetMovedObjectAllowedOffset()                  { return s_xMovedObjectAllowedOffset; }
	static float GetMaxCoverpointRadius()										{ return s_fMaxCoverpointRadius; }
    static bool  GetShouldGenCoverpoints()                                      { return s_bShouldGenCoverpoints; }
    static bool  GetShouldCalculateWaypointZoningData()                         { return s_bShouldCalcZoningData; }
	static float GetMaxCoverPointDistanceSq()									{ return s_fMaxCoverPointDistanceSq; }
	static int   GetMaxCoverpointsPerWaypoint()									{ return s_iMaxCoverpointsPerWaypoint; }

	static bool  GetGenerateTestWaypointsOnly()									{ return s_bGenerateTestWaypointsOnly; }

    static int  GetNumZones()                                                   { return s_iNumZones; }
    static int  GetNumCoverpoints()												{ return s_iNumCoverpoints; }
	static Tools_Nav_Coverpoint* GetCoverpoint(int i)							{ return &s_pxCoverpointStorage[ i ]; }


    // Iteratively attempts to place an object, moving it by the movement vector each time up to the allowed offset.
    // Returns whether it was successful or not.
    static bool PlaceObjectAt( const Asura_Vector_3& xPosition, const Asura_Vector_3& xMovementVector, const float fAllowedOffsetSq,
                               const Asura_Collision_Object& xCollisionObject, const u_short usFlagsOfPolysToIgnore,
                               const Asura_Matrix_3x3& xOrientMatrix, Asura_Vector_3& xFinalPosition,
							   Asura_Physics_Object::COLLISION_PROXY eCollisionProxy = Asura_Physics_Object::COLLISION_PROXY_OBB );
    // Uses a LOS test to get the surface normal at the desired position.
    static bool GetSurfaceNormalAtPosition( const Asura_Vector_3& xPosition, Asura_Vector_3& xSurfaceNormal, u_int uLOSFlags = AsuraPolygonFlag_NoCollisions );
    // Creates an orientation matrix from the surface normal at the desired position.
    static bool GetSurfaceOrientAtPosition( const Asura_Vector_3& xPosition, Asura_Matrix_3x3& xOrientMatrix );
    // Returns whether we made it or not. Also, the physics object will be left at the final position.
    // Will write the intermediate positions to an array, if provided.
    static bool MoveToDestination( const Asura_Vector_3& xStart, const Asura_Vector_3& xFinish, const Asura_Vector_3& xAllowedOffsetVector,
                                   Asura_Physics_Object* pxPhysicsObject, float& fMaxFallDistance,
                                   Asura_Vector_3* pxPositionArray = NULL, u_int* puPositionCount = NULL, const u_int uMaxPositions = 0 );

    static bool UpdateProgress( const int iNum, const int iDen, const char* pszHeading );

    static void CalculateOrientationFromNormalAndDirection(const Asura_Vector_3& xNormal, const Asura_Vector_3& xDirection, Asura_Matrix_3x3& xOrientation);

	static bool HasCreatedToolsEntities() { return s_bToolsEntitiesCreated; }

protected:
    void         TransferFileData( const int iMode );
    virtual void TransferToolsModeData( const int iMode, Tools_ModeDataFile& xFile );
    virtual void AddOverlayControls( Asura_Overlay* pxNewOverlay, Asura_Vector_2& xOffset );
    virtual Tools_Waypoint_Analysis* CreateWaypointAnalysis();
    virtual Tools_Waypoint_Editing*  CreateWaypointEditing();

    virtual void GenerateWaypoints( const bool bTestWaypointsOnly );
    virtual void PreGenerateWaypoints();
    virtual void GenerateWaypointsMain();
    virtual void GenerateWaypointsMainIncremental();
    virtual bool GenerateWaypointsSingleStep(int iPhase,bool& bFinished);
    virtual void PostGenerateWaypoints();

    virtual void VerifyWaypointingData();

    // If you want your own project specific user waypoint object then override this function to create the correct type.
	virtual Tools_Waypoint_Object* CreateUserWaypointToolsObject( Asura_AI_Navigation_System::UserDefinedWaypoint* pxUDW );

    // adds a user defined waypoint to the in-generation waypoints
    virtual bool AddUserDefinedNewWaypoint( const Tools_Waypoint_Object* const pxUserWaypoint );

    // Initialising physics objects for waypointing/linking.
    virtual void RememberPhysicsObjectsStatus();
    virtual void SetRequiredPhysicsObjectsAsIgnored();
    virtual bool CreateWaypointingPhysicsObjects();
    // Resetting physics objects after waypointing/linking.
    virtual void DestroyWaypointingPhysicsObjects();
    virtual void RestorePhysicsObjectsStatus();
    virtual void ForgetPhysicsObjectsStatus();

	virtual int  GetFlagsForNewWaypoint( const int iPolyFlags, const Asura_Vector_3& xPolyNormal ) const;
    virtual bool IsSlopePoly( const Asura_Vector_3& xFaceNormal )   const { return ( xFaceNormal.y > -0.99f ); }
    virtual bool IsCeilingPoly( const Asura_Vector_3& xFaceNormal ) const { return ( xFaceNormal.y >= 0.99f ); }
    virtual bool IsWallPoly( const Asura_Vector_3& xFaceNormal )	const { return ( xFaceNormal.y < 0.1f && xFaceNormal.y > -0.1f); }

    // Returns whether the link is ok or not (and so should be culled).
    // Note that, if called from outside the normal waypoint data generation code (from Tools_Waypoint_Analysis or Tools_Waypoint_Editing),
    // the s_pxWaypoints array will be NULL. In this case the data in the Asura_AI_Navigation_System must be used.
	virtual bool CullAndFlagLinkWithPhysics( unsigned int&         uFlags,
                                             const Asura_Vector_3& xStart,
											 const Asura_Vector_3& xStartNormal,
                                             const int             iStartWPIndex,
                                             const Asura_Vector_3& xFinish,
											 const Asura_Vector_3& xFinishNormal,
                                             const int             iFinishWPIndex );

    // Verifying waypointing data.
    virtual void WarnAboutNonCollisionMeshWaypointingFlags();
	virtual void WarnAboutBadPhysicalObjectPosOri();
	virtual void WarnAboutBadPhysicalObjectWPFlags();

    // render the poly flags for waypointing on the meshes
    virtual void RenderLinks() const;
    virtual void RenderLinksHelpText( float& fX, float& fY ) const;
    virtual void RenderWaypoints() const;
	virtual void RenderTestLinks() const;
	virtual void RenderTestWaypoints() const;
	virtual void RenderWaypointNormals() const;

    virtual void GetLinkColour( const unsigned int uFlags, const bool bTwoWayLink, Asura_Vector_3& xStartColour, Asura_Vector_3& xEndColour ) const;
    virtual void GetWaypointColour( const unsigned int uFlags, int& iColour ) const;

    virtual void RenderWaypointInfo( const int iWPIndex, float& fPrintX, float& fPrintY ) const;
    virtual void RenderIntermediateWaypointInfo( float& fPrintX, float& fPrintY ) const;

protected:
    // All functions return whether they completed successfully or not.
    // initialising
    virtual bool VerifyWaypointingParameters();
    virtual bool InitialiseAndAllocateStorage();
    virtual bool InitialisePhysicsDataForWaypointing();

    // waypointing
    virtual bool AddNewWaypoints();
    virtual bool ReleaseHashTableStorage();
    virtual bool CullWaypointsUnderObjectsAndDoors();
	virtual bool CullWaypointsUnderContainers();
    virtual bool CalculateWaypointVolumes();
    virtual bool FindPotentialCoverpoints();
    virtual bool AllocateCoverpointStorage();
    virtual bool GenerateCoverpointData();
    virtual bool CullWaypointsBySize();
    virtual bool SortWaypointsPerBin();
    virtual bool CullRedundantWaypoints();
    virtual bool AllocateReducedSizeWaypointStorage();
    virtual bool CopyWaypointsIntoReducedStorage();
    virtual bool ReleaseInitialWaypointStorage();
    virtual bool CullPhysicallyInvalidWaypoints();

    // linking
    virtual bool GenerateWaypointLinks();
    virtual bool CullRedundantLinks();
    virtual bool MakeAllLinksTwoWay();
    virtual bool CullAndFlagLinksWithPhysics();
    virtual bool CullTestWaypointsWithNoLinks();
	virtual bool AddInterZoneWaypointLinks();

    // coverpointing
    virtual bool CheckCoverpointsAreNearWaypoints();
    virtual bool CullRedundantCoverpoints();
    virtual bool AssignCoverpointsToWaypoints();
    virtual bool FindCoverpointExits();

    // setting up navigation system
    virtual bool CopyDataToNavigationSystem();
    virtual bool CopyDataToTestWaypointStorage();
#ifdef ASURA_USE_STREAMING
	void FindInterZoneWaypointsInNavigationSystem();
#endif

    // zoning
    virtual bool CullOneWayLinksWithNoReturn();
    virtual bool CalculateWaypointZoneIDs();

    // final steps to creating new nav data
    virtual bool CullWaypointsWithNoLinks();
	virtual bool AllowWaypointCull( int iFlags ) const					{ return true; }
    virtual bool CreateNavigationSystemWaypointTree();
    virtual bool CalculatePhysicalObjectWaypointBlocks();

    // cleaning up
    virtual bool ResetPhysicsDataAfterWaypointing();
    virtual bool ReleaseStorageAndReset( const bool bRenderProgress = true );

    virtual Asura_Char* GetPhaseName( const int iPhase );

protected:
    static void HandleKeyboard();
    static void ToggleWaypointingOverlay();
    static bool WaypointingOverlayIsOn();
    static void CreateWaypointingOverlay();
    static void RenderCoverpoints();
    static void RenderIntermediateWaypoints();
    static void RenderIntermediateLinks();
	static void RenderWaypointBlockerVolumes();
    static void CreateUserWaypoint();
    static void CreateUserWaypointToolsObjects();
    static void DestroyUserWaypointToolsObjects();
    static void CalculateWaypointDataStatistics();
    static void FindWaypointByIndex();

	static void CreateToolsEntities();
	static void UpdateStaticObjects();

	static void RemoveWaypointsBlockedByCollisionObject (const Asura_Collision_Object* pxCollisionData,
														 const Asura_Vector_3& xPosition,
														 const Asura_Matrix_3x3& xOrientation);

    struct CalcWPVolumesInBin_ThreadParams
    {
        CalcWPVolumesInBin_ThreadParams( u_int uThreadCount )
        : m_uThreadCount( uThreadCount )
        , m_xCurrentBinMutex()
        , m_uCurrentBin( 0 )
        , m_xFinishedWaypointCountMutex()
        , m_uFinishedWaypointCount( 0 )
        , m_xIteratorMutex()
		, m_xContainerIteratorMutex()
        , m_xThreadFinishedCountMutex()
        , m_uThreadFinishedCount( 0 )
        , m_xCancelEvent( "CalcWPVolumesInBin Cancelled", false )
        , m_xFinishedEvent( "CalcWPVolumesInBin Finished", false )
        {}

        const u_int     m_uThreadCount;
        Asura_Mutex     m_xFinishedWaypointCountMutex;
        volatile u_int  m_uFinishedWaypointCount;
        Asura_Mutex     m_xCurrentBinMutex;
        volatile u_int  m_uCurrentBin;
        Asura_Mutex     m_xIteratorMutex;
		Asura_Mutex     m_xContainerIteratorMutex;
        Asura_Mutex     m_xThreadFinishedCountMutex;
        volatile u_int  m_uThreadFinishedCount;
        Asura_SyncEvent m_xCancelEvent;
        Asura_SyncEvent m_xFinishedEvent;
    };
    static void CalcWPVolumesInBin_ThreadFunction( const void* pParams );

    // searches the waypoint position hash table to see if this entry already exists
    static bool  FindInHashTable( const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal );
    static bool  AddToHashTable( const int iBinIndex, const int iWaypointIndex, const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal );
    static u_int GetHashTableKey( const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal )
    {
        u_int uKey = static_cast< u_int >( Asura_Maths::FloatToInt( xPosition.x + xPosition.y + xPosition.z ) );
		uKey += static_cast< u_int >( Asura_Maths::FloatToInt( xNormal.x + xNormal.y + xNormal.z ) );
        return ( uKey & ( NUM_HASH_BINS - 1 ) );
    }

	static bool CanGenerateTestWaypointsStatic()	{ return ( s_iGenerateIncrementalWaypointPhase < 0 ); }
	static bool HasTestWaypointsStatic()            { return ( s_pxTestWaypointStorage != NULL ); }
    static void GenerateWaypointsStatic()           { s_pxThis->GenerateWaypoints( false ); }
    static void GenerateTestWaypointsStatic()       { s_pxThis->GenerateWaypoints( true ); }
    static void VerifyWaypointingDataStatic();
	static void ShowUDWsChanged();
    static void DeleteTestWaypoints();

    // adds a waypoint to the in-generation waypoints
    static bool AddNewWaypoint( const int iModuleIndex, const Asura_Vector_3& xPosition, const Asura_Vector_3& xNormal, const int iWaypointFlags );
#ifdef ASURA_USE_STREAMING
	static bool AddInterZoneNewWaypoint( const Asura_Vector_3& xPosition, int iPairIndex, bool bBelongsToThisEnv );
#endif
    static bool SubdivideAndAddWaypoints( const Asura_Vector_3& xVertex0,
                                          const Asura_Vector_3& xVertex1,
                                          const Asura_Vector_3& xVertex2,
                                          const Asura_Vector_3& xVertex3,
									      int                   iRecursionLevel );
    static bool SubdivideAndAddWaypoints( const Asura_Vector_3& xVertex0,
                                          const Asura_Vector_3& xVertex1,
                                          const Asura_Vector_3& xVertex2,
										  int                   iRecursionLevel );

    static bool RadiusIsLessThanMinRadius( const int iWaypointFlags, const float fRadius )
    {
        if ( 0 != ( iWaypointFlags & TOOLS_NAVVERTEX_FLAG_DENSE ) )
        {
            return ( fRadius < s_fMinDenseWaypointRadius );
        }
        else
        {
            return ( fRadius < s_fMinWaypointRadius );
        }
    }

    // NOTE: This function will cause the bin pointers to overshoot the end of the bins. Do not use the bin pointers once this function returns false.
    static bool GetNextLargestWaypoint( Tools_Nav_Vertex* apxBinPointers[], Tools_Nav_Vertex*& pxLargestWaypoint, int& iBinIndex, int& iIntraBinIndex );
    static void AABBTreePerBinSizeTestCallback( u_int uIndex );
    static bool CalculateBinAndIntraBinIndex( const u_int uIndex, int& iBinIndex, int& iIntraBinIndex );
	static void CalculateGlobalIndex( int iBinIndex, const int iIntraBinIndex, u_int& uGlobalIndex );
    static void WaypointVisibilityCheckCallback( const u_int uIndex );
    static void CoverpointVisibilityCheckCallback( const u_int uIndex );

    static bool GetWaypointIndexUnderMouse( int& iIndex,
                                            const class Tools_Nav_Vertex* pxVerts,
                                            const int iNumVerts );

    virtual unsigned short LOSExistsFromCoverpointToWaypoint( const int iCoverpoint, const int iWaypoint );
    static void           FindCoverpointExits( const int iCoverpoint,
                                             unsigned char& ucShortExits,
                                             unsigned char& ucLongExits );
    static bool FloodFillWaypointGraph( const int iStartWaypointIndex, const u_short usZoneID );
    static int  GetEnclosingWaypointIndex( const Asura_Vector_3& xPosition );

    static void RenderLine( const Asura_Vector_3& xStart,       const Asura_Vector_3& xEnd,
                            const Asura_Vector_3& xStartColour, const Asura_Vector_3& xEndColour, 
                            const int             iNumSubDivisions = 5 );

    // Can be used mid-waypoint generation to check if we still have potential waypoints.
    static bool SomeBinWaypointsExist()
    {
        for ( int i = 0; i < NUM_WAYPOINT_BINS; ++i )
        {
            if ( s_aiNumWaypoints[ i ] > 0 )
            {
                return true;
            }
        }
        return false;
    }
    static int GetTotalBinWaypointCount()
    {
        int iCount = 0;
        for ( int i = 0; i < NUM_WAYPOINT_BINS; ++i ) { iCount += s_aiNumWaypoints[ i ]; }
        return iCount;
    }

    static bool LOSTestHit( const Asura_Vector_3& xStart, const Asura_Vector_3& xEnd );

	// Incremental waypoint functions...
	static void FindNextIncrementalWaypointPhase();

	static void PreGenerateWaypointsIncrementalStatic()		{ FindNextIncrementalWaypointPhase(); s_pxThis->PreGenerateWaypoints(); }
	static void GenerateWaypointsIncrementalStatic()		{ s_pxThis->GenerateWaypointsMainIncremental(); }
	static void PostGenerateWaypointsIncrementalStatic()	{ s_iGenerateIncrementalWaypointPhase = -1; s_pxThis->PostGenerateWaypoints(); }

	static bool HasIntermediateWaypointsStatic()			{ return ( SomeBinWaypointsExist() || ( s_iNumWaypoints > 0 ) ); }
	static bool CanPreGenerateWaypointsIncrementalStatic()	{ return(s_iGenerateIncrementalWaypointPhase < 0); }
	static bool CanGenerateWaypointsIncrementalStatic()		{ return((s_iGenerateIncrementalWaypointPhase >= 0) && (s_iGenerateIncrementalWaypointPhase < INT_MAX)); }
	static bool CanPostGenerateWaypointsIncrementalStatic()	{ return(s_iGenerateIncrementalWaypointPhase == INT_MAX); }

protected:
	Asura_Dialog*					   m_pxAssignCoverpointsTickbox;

    static Tools_Mode_Waypointing*     s_pxThis;
    static Tools_Waypoint_Analysis*    s_pxWaypointAnalysis;
    static Tools_Waypoint_Editing*     s_pxWaypointEditing;
    static bool                        s_bIsBatchWaypointing;

    // Members for the overlay controls.
    enum TAB_NAMES
    {
        TAB_SETTINGS,
        TAB_RENDERING,
        TAB_SELECTION,
        TAB_ANALYSIS,
        TAB_EDITING,
        TAB_TESTING,
        TAB_NUM_TAB_NAMES
    };
    static const char*                 s_aszTabNames[ TAB_NUM_TAB_NAMES ];
    static Asura_Hash_ID               s_iTabDialogueControlID;
    static bool                        s_bShowFloorFlag;
    static bool                        s_bShowWaypoints;
    static bool                        s_bShowWaypointNormals;
    static bool                        s_bShowLinks;
    static bool                        s_bShowOnlyInZone;
    static bool                        s_bShowOnlyInRange;
    static int                         s_iZoneIDToRender;
    static float                       s_fRangeToRender;
    static bool                        s_bShowCoverpoints;
    static bool                        s_bShowCoverpointExits;
    static bool                        s_bShowCoverpointFlags;
    static bool                        s_bShowCoverpointDistances;
    static bool                        s_bShowCoverpointGradientColour;
    static float                       s_fShowCoverpointDistanceScale;
    static bool                        s_bShowTestWaypoints;
    static bool                        s_bShowTestLinks;
    static bool                        s_bShowIntermediateWaypoints;
    static bool                        s_bShowIntermediateLinks;
	static bool						   s_bShowUserDefinedWaypoints;
	static bool                        s_bShowWaypointBlockerVolumes;
    static int                         s_iNumIntermediateWaypointsInBatch;
    static int                         s_iIntermediateWaypointBatchIndex;
    static bool                        s_bShowWaypointInfoUnderCursor;
    static bool	                       s_bWaypointingOverlayOpen;
    static Asura_Vector_2              s_xWaypointingOverlayPosition;
    static Asura_Handle                s_iMenuHandle;
    static bool                        s_bGenerateTestWaypointsOnly;
    static int                         s_iNumZones;
    static unsigned short              s_usNumWaypointsInZone[ ASURA_NAVVERTEX_MAX_ZONE_ID + 1 ];
    static float                       s_fCoverpointReferencesPerWaypoint;

    // storage for in-generation waypoints
    enum
    {
        NUM_WAYPOINTS_PER_BIN = 1024 * 32,
        NUM_WAYPOINT_BINS     = 256
    };
    static Tools_Nav_Vertex*           s_apxWaypointStorage[ NUM_WAYPOINT_BINS ];
    static int                         s_aiNumWaypoints[ NUM_WAYPOINT_BINS ];
    static Tools_Nav_Vertex*           s_pxWaypointStorage;
    static int                         s_iNumWaypoints;

    static Tools_Nav_Coverpoint*       s_pxCoverpointStorage;
    static int                         s_iNumCoverpoints;

    // table to store the waypoint positions already tagged, to avoid duplication
    // NUM_HASH_BINS must be a power of two
    enum { NUM_HASH_BINS = 512, NUM_HASH_BIN_ENTRIES = 20000 };
    static bool                        s_bHashStorageAllocated;
    static int**                       s_ppiHashTable;
    static int*                        s_piNumHashEntries;

    // storage for waypoint visibility tests during coverpoint generation
    static bool*                       s_bVisibleModules;
    // Used by the callbacks for Asura_Module_System::TestModuleTree().
	static bool                        s_bBBInNonExteriorModule;

    // storage for test waypoints
    static Tools_Nav_Vertex*           s_pxTestWaypointStorage;
    static int                         s_iNumTestWaypoints;

    // Storage for physics MoveToDestination test, to aid culling of links.
    enum { MAX_STORED_PHYSICS_POSITIONS = 400 };
    static Asura_Vector_3*             s_pxPhysicsPositions;
    static u_int                       s_uNumPhysicsPositions;

    // Storage for remembering which physics objects were ignored and which not at the start of waypointing mode, so this can be restored at the end.
    static bool											 s_bPhysicsDataInitialised;
	static Asura_List< Asura_Physics_RigidBodyInstance > s_xRigidBodyIgnoreList;
	static Asura_List< Asura_Physics_RigidBodyInstance > s_xRigidBodyNonIgnoreList;

    // waypoint, link and coverpoint generation parameters
    // c.f. comments by the accessor functions
	static float                       s_fMinSubdivideSizeSq;
    static float                       s_fMinDenseSubdivideSizeSq;
    static int                         s_iMaxSubdivideRecursionLevel;
    static float                       s_fMinWaypointClearanceRadius;
    static float                       s_fMinWaypointRadius;
    static float                       s_fMaxWaypointRadius;
    static float                       s_fMinDenseWaypointRadius;
    static float                       s_fMaxDenseWaypointRadius;
	static float					   s_fMinLoneWaypointRadius;
    static bool						   s_bCliffStopWaypoints;
    static float					   s_fWallBlockHeight;
    static float                       s_fWaypointOverlapFraction;
    static void ( *s_pfnWaypointCalcThreadFn )( const void* pParams );

	static Asura_Physics_Object::COLLISION_PROXY	s_eCollisionProxy;
    static Asura_Physics_Object*       s_pxStandardPhysicsObject;
    static Asura_Collision_Box         s_xTestCollisionBox;
	static bool						   s_bCrouchEnabled;
    static Asura_Physics_Object*       s_pxCrouchPhysicsObject;
	static Asura_Collision_Box         s_xCrouchCollisionBox;
	static bool						   s_bProneEnabled;
    static Asura_Physics_Object*       s_pxPronePhysicsObject;
	static Asura_Collision_Box         s_xProneCollisionBox;
    static float                       s_fLinkTraversalSpeed;
    static bool                        s_bFallDirectlyDownwards;
    static float                       s_fPlacedObjectMovementScale;
    static float                       s_fPlacedObjectAllowedMovement;
    static float                       s_fMinLinkLength;
    static float                       s_fMaxLinkLength;
    static float                       s_fLinkOverlapDistance;
    static float                       s_fLinkOverlapFraction;
    static float                       s_fFallLinkDistance;
    static float                       s_fLongFallLinkDistance;
    static float                       s_fMaxLinkFallDistance;
	static float					   s_fMaxJumpImpulse;
    static float                       s_fCanReachDistance;
    static Asura_Vector_3              s_xMovedObjectAllowedOffset;
    static bool						   s_bDontCullLinkedToWaypoints;
    static bool						   s_bShouldCullOneWayLinksWithNoReturn;

    static bool                        s_bShouldGenCoverpoints;
	static float					   s_fMinCoverpointRadius;
    static float                       s_fMaxCoverpointRadius;
    static bool                        s_bShouldCalcZoningData;
	static float					   s_fMaxCoverPointDistanceSq;
	static int						   s_iMaxCoverpointsPerWaypoint;
	static float					   s_fWaypointToCoverpointMaxDist;
	static bool						   s_bCalculateCoverpointExits;

    // Allow gravity for physics objects to change based on the normal of the polygon under the waypoint.
	static bool                        s_bGravityFollowsSurfaceNormals;
    // Polys whose dot-product is less than this will act to constrict waypoint volumes.
	static float                       s_fMaxDotProductBetweenWaypointPolygonNormals;
    // Links to a poly whose dot-product is less than this will be culled.
	static float                       s_fMaxDotProductForLinkBetweenWaypointNormals;

	static bool						   s_bToolsEntitiesCreated;
	static bool						   s_bStaticObjectsVisible;
	static bool						   s_bShowStaticObjects;
	static bool						   s_bWaypointingSettingsModified;

	//Used by waypoint generation so that it knows if the data needs to be deleted if aborted halfway
	static bool                        s_bDataInNavSystem;

	// Used during adding waypoints to avoid having to pass parameters around on the stack.
	static int						   s_iAWCurrentModuleIndex;
	static Asura_Vector_3			   s_xAWCurrentNormal;
	static int						   s_iAWCurrentWPFlags;
	static float					   s_fAWCurrentMinSubdivideSizeSq;
	static bool						   s_bAWDoubleSidedPoly;

    // Indices for any callback functions necessary.
    static int                         s_iBinTestIndex;
    static u_int                       s_uTestIndex;

	static Asura_Tabs_Dialog*		   s_pxWaypointingTabs;

	//for restoring the collision mesh rendering mode when leaving waypointing mode
	static int                         s_iInitialRenderMode;

	// Incremental waypoint generation...
	static int                         s_iGenerateIncrementalWaypointPhase;

	static bool						   s_bShowMessageLogOnCompletion;

protected:
    static const u_short s_usWhiteZoneID;
    static const u_short s_usGrayZoneID;
    static const u_short s_usBlackZoneID;

private:
    enum { CURRENT_MODE_DATA_VERSION = 5 };
    static const char* s_pszWaypointingModeDataFilename;
};

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

#endif //___TOOLS_MODE_WAYPOINTING_H___
