//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Editing.cpp
//
// Copyright (C) 2007 Rebellion. All Rights Reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

// core
#include "Core/Asura.h"

// header
#include "Tools_Waypoint_Editing.h"

// asura
#include "Core/Asura_Maths.h"
#include "Core/Asura_String.h"
#include "Core/Memory/Asura_ResourceAllocator.h"
#include "Entity/PhysicalObject/Asura_ServerEntity_PhysicalObject.h"
#include "Entity/Triggers/Asura_ServerEntity_WaypointBlockerVolume.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Overlay_System.h"

// tools
#include "Entity/Tools_Entity_Object.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Objects/Tools_Object_System.h"
#include "Waypointing/Tools_Mode_Waypointing.h"
#include "Waypointing/Tools_Waypoint_Object.h"
#include "Waypointing/Tools_Waypoint_Selection.h"

#ifdef ASURA_USE_CONTAINER_SYSTEM
#include "Entity/Container/Tools_Entity_Container.h"
#include "Entity/Container/Asura_ServerEntity_Container.h"
#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

//////////////////////////////////////////////////////////////////////////////
//
// Tools_Waypoint_Editing
//
//////////////////////////////////////////////////////////////////////////////

Tools_Waypoint_Editing* Tools_Waypoint_Editing::s_pxThis = NULL;

Tools_Waypoint_Editing::Tools_Waypoint_Editing()
: m_bAdjustingWaypointRadius( false )
, m_xRadiusAdjustOverlayPosition( 3.0f, 6.0f )
{
    s_pxThis = this;
}

Tools_Waypoint_Editing::~Tools_Waypoint_Editing()
{
    s_pxThis = NULL;
}

void Tools_Waypoint_Editing::Shutdown()
{
    if ( m_bAdjustingWaypointRadius )
    {
        StopAdjustingWaypointRadius();
    }
}

void Tools_Waypoint_Editing::AddOverlayItemsToTabsDialogue( Asura_Tabs_Dialog* pxTabsDialogue, Asura_Vector_2& xOffset )
{
    Asura_Dialog* pxNewDialogue = NULL;
	
	pxNewDialogue = new Asura_Function_Dialog( "Adjust Selected Waypoint Radius", xOffset,
        AdjustSelectedWaypointRadiusStatic, NULL, AllowedToAdjustWaypointRadius );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Add Link Between Currently Selected Waypoints", xOffset,
        AddLinkBetweenSelectedWaypointsStatic, 1, NULL, AllowedToEditLinking );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

	pxNewDialogue = new Asura_Function_Dialog( "Add One-Way Link Between Currently Selected Waypoints", xOffset,
		AddLinkBetweenSelectedWaypointsStatic, 0, NULL, AllowedToEditLinking );
	pxNewDialogue->SetHelpText( "Adds a link between the first waypoint selected and the second, but not vice versa." );
	pxTabsDialogue->AddDialog( pxNewDialogue );
	xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Delete Link Between Currently Selected Waypoints", xOffset,
        DeleteLinkBetweenSelectedWaypointsStatic, 1, NULL, AllowedToEditLinking );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();

	pxNewDialogue = new Asura_Function_Dialog( "Delete Link One-Way Between Currently Selected Waypoints", xOffset,
		DeleteLinkBetweenSelectedWaypointsStatic, 0, NULL, AllowedToEditLinking );
	pxNewDialogue->SetHelpText( "Deletes the link between the first waypoint selected and the second, if it exists, but not vice versa." );
	pxTabsDialogue->AddDialog( pxNewDialogue );
	xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialogue = new Asura_Function_Dialog( "Delete Currently Selected Waypoints And All Associated Links", xOffset,
        DeleteSelectedWaypointsAndLinksStatic, NULL, AllowedToDeleteWaypoints );
    pxTabsDialogue->AddDialog( pxNewDialogue );
    xOffset.y += Asura_Dialog::DefaultSpacing();
}

void Tools_Waypoint_Editing::AddLink( const int iWaypointOne, const int iWaypointTwo, const bool bRecalcDependentData, const bool bTwoWay /*= true*/ )
{
    Asura_AI_NavVertex* pxWaypointOne = Asura_AI_Navigation_System::GetWaypoint( iWaypointOne );
    Asura_AI_NavVertex* pxWaypointTwo = Asura_AI_Navigation_System::GetWaypoint( iWaypointTwo );
    if ( !pxWaypointOne && !pxWaypointTwo ) { return; }

    bool bLinkAdded = false;

    const float fLinkDistance = ( pxWaypointOne->m_xPosition - pxWaypointTwo->m_xPosition ).Magnitude();

    // check that a link doesn't already exist between A and B
    bool bConnectionExists = false;
    for ( int i = 0; i < pxWaypointOne->m_ucNumConnections; ++i )
    {
        if ( pxWaypointOne->m_pxConnections[ i ].m_usIndex == iWaypointTwo )
        {
            bConnectionExists = true;
            break;
        }
    }

    if ( bConnectionExists )
    {
        char acString[ 64 ];
        Asura_PrintfToBfr( acString, "Link already exists between %d and %d.", 64, true, iWaypointOne, iWaypointTwo );
        Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, acString );
    }
    else
    {
        // Try and add a link.
        Tools_Mode_Waypointing::GetInstance()->InitialisePhysicsDataForWaypointing();

        Asura_Vector_3 xWaypointOneNormal = pxWaypointOne->GetNormal();
        Asura_Vector_3 xWaypointTwoNormal = pxWaypointTwo->GetNormal();

        unsigned int uAToBFlags  = 0;
        const bool   bAToBLinkOk = Tools_Mode_Waypointing::GetInstance()->CullAndFlagLinkWithPhysics(
            uAToBFlags,
            pxWaypointOne->m_xPosition,
            xWaypointOneNormal,
            iWaypointOne,
            pxWaypointTwo->m_xPosition,
            xWaypointTwoNormal,
            iWaypointTwo );

        Tools_Mode_Waypointing::GetInstance()->ResetPhysicsDataAfterWaypointing();

        if ( bAToBLinkOk )
        {
            // Add the link to waypoint A.
            InsertLinkIntoWaypointInfo( iWaypointOne, iWaypointTwo, uAToBFlags, fLinkDistance );
            bLinkAdded = true;
        }
        else
        {
            char acString[ 64 ];
            Asura_PrintfToBfr( acString, "Physics test fails between %d and %d.", 64, true, iWaypointOne, iWaypointTwo );
            Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, acString );
        }
    }

	if ( bTwoWay )
	{
		// check that a link doesn't already exist between B and A
		bConnectionExists = false;
		for ( int i = 0; i < pxWaypointTwo->m_ucNumConnections; ++i )
		{
			if ( pxWaypointTwo->m_pxConnections[ i ].m_usIndex == iWaypointOne )
			{
				bConnectionExists = true;
				break;
			}
		}

		if ( bConnectionExists )
		{
			char acString[ 64 ];
			Asura_PrintfToBfr( acString, "Link already exists between %d and %d.", 64, true, iWaypointTwo, iWaypointOne );
			Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, acString );
		}
		else
		{
			// Try and add a link.
			Tools_Mode_Waypointing::GetInstance()->InitialisePhysicsDataForWaypointing();

			Asura_Vector_3 xWaypointOneNormal = pxWaypointOne->GetNormal();
			Asura_Vector_3 xWaypointTwoNormal = pxWaypointTwo->GetNormal();

			unsigned int uBToAFlags  = 0;
			const bool   bBToALinkOk = Tools_Mode_Waypointing::GetInstance()->CullAndFlagLinkWithPhysics(
				uBToAFlags,
				pxWaypointTwo->m_xPosition,
				xWaypointTwoNormal,
				iWaypointTwo,
				pxWaypointOne->m_xPosition,
				xWaypointOneNormal,
				iWaypointOne );

			Tools_Mode_Waypointing::GetInstance()->ResetPhysicsDataAfterWaypointing();

			if ( bBToALinkOk )
			{
				// add the link to the end of waypoint B
				InsertLinkIntoWaypointInfo( iWaypointTwo, iWaypointOne, uBToAFlags, fLinkDistance );
				bLinkAdded = true;
			}
			else
			{
				char acString[ 64 ];
				Asura_PrintfToBfr( acString, "Physics test fails between %d and %d.", 64, true, iWaypointTwo, iWaypointOne );
				Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, acString );
			}
		}
	}

    // If we did add a link we need to update the physical object waypoint blocks.
    if ( bLinkAdded && bRecalcDependentData )
    {
        UpdateWaypointLinkBlocks();
    }
}

void Tools_Waypoint_Editing::DeleteLink( const int iWaypointOne, const int iWaypointTwo, const bool bRecalcDependentData, const bool bTwoWay /*= true*/ )
{
    Asura_AI_NavVertex* pxWaypointOne = Asura_AI_Navigation_System::GetWaypoint( iWaypointOne );
    Asura_AI_NavVertex* pxWaypointTwo = Asura_AI_Navigation_System::GetWaypoint( iWaypointTwo );
    if ( !pxWaypointOne && !pxWaypointTwo ) { return; }

    bool bLinkDeleted = false;

    // check that a link exists between A and B
    bool bConnectionExists = false;
    for ( int i = 0; i < pxWaypointOne->m_ucNumConnections; ++i )
    {
        if ( pxWaypointOne->m_pxConnections[ i ].m_usIndex == iWaypointTwo )
        {
            bConnectionExists = true;
            break;
        }
    }

    if ( bConnectionExists )
    {
        RemoveLinkFromWaypointInfo( iWaypointOne, iWaypointTwo );
        bLinkDeleted = true;
    }
    else
    {
        char acString[ 64 ];
        Asura_PrintfToBfr( acString, "No link found between %d and %d.", 64, true, iWaypointOne, iWaypointTwo );
        Tools_Timed_Message_System::AddMessage( 0xffffffff, acString );
    }

	if ( bTwoWay )
	{
		// check that a link exists between B and A
		bConnectionExists = false;
		for ( int i = 0; i < pxWaypointTwo->m_ucNumConnections; ++i )
		{
			if ( pxWaypointTwo->m_pxConnections[ i ].m_usIndex == iWaypointOne )
			{
				bConnectionExists = true;
				break;
			}
		}

		if ( bConnectionExists )
		{
			RemoveLinkFromWaypointInfo( iWaypointTwo, iWaypointOne );
			bLinkDeleted = true;
		}
		else
		{
			char acString[ 64 ];
			Asura_PrintfToBfr( acString, "No link found between %d and %d.", 64, true, iWaypointTwo, iWaypointOne );
			Tools_Timed_Message_System::AddMessage( 0xffffffff, acString );
		}
	}

    // If we did delete any links we need to update the physical object waypoint blocks.
    if ( bLinkDeleted && bRecalcDependentData )
    {
        UpdateWaypointLinkBlocks();
    }
}

void Tools_Waypoint_Editing::DeleteWaypointAndLinks( const int iWaypointIndex, const bool bRecalcDependentData /*= true*/ )
{
    Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( iWaypointIndex );
    if ( !pxWaypoint ) { return; }

    // delete all the links of this waypoint
    while ( pxWaypoint->m_ucNumConnections )
    {
        // deleting the link will decrement pxWaypoint->NumberOfConnections
        RemoveLinkFromWaypointInfo( iWaypointIndex, pxWaypoint->m_pxConnections[ 0 ].m_usIndex );
    }

    // delete all links pointing to this waypoint
    for ( int i = 0; i < Asura_AI_Navigation_System::GetNumberOfWaypoints(); ++i )
    {
        Asura_AI_NavVertex* pxCurrWaypoint = Asura_AI_Navigation_System::GetWaypoint( i );
        for ( int c = 0; c < pxCurrWaypoint->m_ucNumConnections;  )
        {
            if ( pxCurrWaypoint->m_pxConnections[ c ].m_usIndex == iWaypointIndex )
            {
                RemoveLinkFromWaypointInfo( i, iWaypointIndex );
            }
            else
            {
                ++c;
            }
        }
    }

    // delete the waypoint itself
    RemoveWaypointFromWaypointInfo( iWaypointIndex );

    // We need to update the physical object waypoint blocks and the waypoint tree.
    if ( bRecalcDependentData )
    {
        Asura_AI_Navigation_System::CreateWaypointTree();
        UpdateWaypointLinkBlocks();
    }
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

void Tools_Waypoint_Editing::CreateWaypoint()
{
    //## This should be changed to use the full physics tests.

	bool bChanged = false;

	List<Tools_Object*> xSelectedList;
	Tools_Object_System::ListSelectedObjectsOfType(TOT_UserDefinedWaypointPosition,xSelectedList);
	
	for (LIF<Tools_Object*> xSelectedLIF(&xSelectedList);!xSelectedLIF.done();xSelectedLIF.next())
	{
		Tools_Waypoint_Object* pxObject = (Tools_Waypoint_Object*) xSelectedLIF();

		Asura_AI_Navigation_System::CreateWaypoint( pxObject->Position, 1.f );

		bChanged = true;
	}

	if ( bChanged )
	{
		Asura_AI_Navigation_System::CreateWaypointTree();
		Tools_Waypoint_Selection::OnWaypointDataChanged();
		Tools_SaveAsura::SetWaypointingModified();
	}
}

void Tools_Waypoint_Editing::AdjustSelectedWaypointRadius()
{
    // Find the waypoint index.
    int iSelectedWaypoint;
    if ( !Tools_Waypoint_Selection::GetSelectedWaypoint( iSelectedWaypoint ) ) { return; }

    Asura_AI_NavVertex* pxWaypoint = Asura_AI_Navigation_System::GetWaypoint( iSelectedWaypoint );
    if ( pxWaypoint == NULL ) { return; }

    m_bAdjustingWaypointRadius = true;

    Asura_Overlay* pxNewOverlay;
    pxNewOverlay = new Asura_Overlay( "Set Waypoint Radius", false, &m_xRadiusAdjustOverlayPosition );

    Asura_Dialog* pxNewDialog;
    Asura_Vector_2 xOffset = Asura_Vector_2( 1.0f, 3.0f );

    //pxNewDialog = new Asura_Numerical_Dialog( "Waypoint Radius", xOffset, &pxWaypoint->m_fRadius, 0.0f, 100.0f, 0.05f, "%.2f" );
    //pxNewOverlay->AddDialog( pxNewDialog );
    //xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewDialog = new Asura_Function_Dialog( "Done", xOffset, StopAdjustingWaypointRadiusStatic );
    pxNewOverlay->AddDialog( pxNewDialog );
    xOffset.y += Asura_Dialog::DefaultSpacing();

    pxNewOverlay->m_uID = reinterpret_cast< int >( this );

    Asura_Overlay_System::SubmitNewOverlay( pxNewOverlay, false );
}

void Tools_Waypoint_Editing::StopAdjustingWaypointRadius()
{
    Asura_Overlay_System::CloseOverlays( reinterpret_cast< int >( this ) );
    m_bAdjustingWaypointRadius = false;
}

void Tools_Waypoint_Editing::AddLinkBetweenSelectedWaypoints( int iUserData )
{
    int iWaypointOne, iWaypointTwo;
    if ( !Tools_Waypoint_Selection::GetFirstTwoSelectedWaypoints( iWaypointOne, iWaypointTwo ) ) { return; }

	const bool bTwoWay = ( iUserData != 0 );
    AddLink( iWaypointOne, iWaypointTwo, true, bTwoWay );
}

void Tools_Waypoint_Editing::DeleteLinkBetweenSelectedWaypoints( int iUserData )
{
    int iWaypointOne, iWaypointTwo;
    if ( !Tools_Waypoint_Selection::GetFirstTwoSelectedWaypoints( iWaypointOne, iWaypointTwo ) ) { return; }

	const bool bTwoWay = ( iUserData != 0 );
    DeleteLink( iWaypointOne, iWaypointTwo, true, bTwoWay );
}

void Tools_Waypoint_Editing::DeleteSelectedWaypointsAndLinks()
{
    // Delete the waypoints and all associated links, from back to front so that the indices of the remaining waypoints do not change mid-way.
    Tools_Waypoint_Selection::SelectionData* pxSelectionArray = Tools_Waypoint_Selection::GetSelectionArray();
    for ( int i = Tools_Waypoint_Selection::GetSelectionArrayLength(); i > 0; --i )
    {
        if ( pxSelectionArray[ i - 1 ].m_bIsSelected )
        {
            DeleteWaypointAndLinks( ( i - 1 ), false );
        }
    }

    // Recalculate the dependent data down here, to avoid having to do it repeatedly above.
    Asura_AI_Navigation_System::CreateWaypointTree();
    UpdateWaypointLinkBlocks();

    Tools_Waypoint_Selection::OnWaypointDataChanged();
}

void Tools_Waypoint_Editing::InsertLinkIntoWaypointInfo( const int iFromWaypoint, const int iToWaypoint, const unsigned int uFlags, const float fDistance )
{
	Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
	{
		// allocate new storage for the connections
		Asura_AI_NavConnection* pxConnectionStorage;
		Asura_ResourceAllocateArray(pxConnectionStorage, Asura_AI_Navigation_System::s_uNumConnections + 1);
		if ( pxConnectionStorage == NULL )
		{
            Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Could not allocate enough storage." );
			Asura_ResourcePopType();
			return;
		}

		// copy all of the links across and move the new, larger array back into the nav system
		unsigned int u;
		for ( u = 0; u < Asura_AI_Navigation_System::s_uNumConnections; ++u )
		{
			pxConnectionStorage[ u ] = Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u ];
		}
		Asura_ResourceDeallocateArray(Asura_AI_Navigation_System::s_pxNavConnectionStorage);
		Asura_AI_Navigation_System::s_pxNavConnectionStorage = pxConnectionStorage;
		Asura_AI_Navigation_System::s_uNumConnections += 1;

		// now count the number of connections up to and including the from waypoint
		unsigned int uConnectionCount = 0;
		int i;
		for ( i = 0; i < iFromWaypoint + 1; ++i )
		{
			uConnectionCount += Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections;
		}

		// move all the connections after this count up by one
		// (starting from the end of the array and working back)
		for ( u = Asura_AI_Navigation_System::s_uNumConnections - 1; u > uConnectionCount; --u )
		{
			Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u ] =
				Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u - 1 ];
		}

		// add the connection to the end of the from waypoint's list and increment its connection count
		Asura_AI_Navigation_System::s_pxNavConnectionStorage[ uConnectionCount ].m_usIndex =
			static_cast< unsigned short >( iToWaypoint );
		Asura_AI_Navigation_System::s_pxNavConnectionStorage[ uConnectionCount ].m_usFlags =
			static_cast< unsigned short >( uFlags );
		Asura_AI_Navigation_System::s_pxNavConnectionStorage[ uConnectionCount ].SetLength(fDistance);
		Asura_AI_Navigation_System::s_pxWaypointStorage[ iFromWaypoint ].m_ucNumConnections += 1;

		// fix up the connection pointers in the waypoints again - this time to the new array
		// (this will automatically take into account the new number of connections in the from waypoint)
		{
		Asura_AI_NavConnection* pxConnectionListPtr = Asura_AI_Navigation_System::s_pxNavConnectionStorage;
		for ( i = 0; i < Asura_AI_Navigation_System::s_usNumWaypoints; ++i )
		{
			if ( Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections )
			{
				Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_pxConnections = pxConnectionListPtr;
				pxConnectionListPtr += Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections;
			}
		}
		}

		char acString[ 64 ];
		Asura_PrintfToBfr( acString, "Added a link from waypoint %d to waypoint %d.", 64, true, iFromWaypoint, iToWaypoint );
        Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, acString );

		Tools_SaveAsura::SetWaypointingModified();
	}
	Asura_ResourcePopType();
}

void Tools_Waypoint_Editing::RemoveLinkFromWaypointInfo( const int iFromWaypoint, const int iToWaypoint )
{
    Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
    {
        // allocate new storage for the connections first, so we can fail if needs be
        Asura_AI_NavConnection* pxConnectionStorage = NULL;
        if ( Asura_AI_Navigation_System::s_uNumConnections > 1 )
        {
            Asura_ResourceAllocateArray(pxConnectionStorage,  Asura_AI_Navigation_System::s_uNumConnections - 1);
            if ( pxConnectionStorage == NULL )
            {
                Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Could not allocate enough storage." );
                Asura_ResourcePopType();
                return;
            }
        }

        // find this connection's index in the connection array
        unsigned int uConnectionCount = 0;
        int i;
        for ( i = 0; i < iFromWaypoint; ++i )
        {
            uConnectionCount += Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections;
        }
        Asura_AI_NavVertex* pxFromWaypoint = Asura_AI_Navigation_System::GetWaypoint( iFromWaypoint );
        for ( i = 0; i < pxFromWaypoint->m_ucNumConnections; ++i )
        {
            if ( pxFromWaypoint->m_pxConnections[ i ].m_usIndex == iToWaypoint )
            {
                break;
            }
            else
            {
                ++uConnectionCount;
            }
        }

        // now move the connections down, overwriting this index
        unsigned int u;
        for ( u = uConnectionCount; u < Asura_AI_Navigation_System::s_uNumConnections - 1; ++u )
        {
            Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u ] =
                Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u + 1 ];
        }

        // decrement the number of connections the from waypoint has
        pxFromWaypoint->m_ucNumConnections -= 1;
        if ( pxFromWaypoint->m_ucNumConnections == 0 )
        {
            pxFromWaypoint->m_pxConnections = NULL;
        }

        // copy all of the links across and move the new, smaller array back into the nav system
        for ( u = 0; u < Asura_AI_Navigation_System::s_uNumConnections - 1; ++u )
        {
            pxConnectionStorage[ u ] = Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u ];
        }
        Asura_ResourceDeallocateArray(Asura_AI_Navigation_System::s_pxNavConnectionStorage);
        Asura_AI_Navigation_System::s_pxNavConnectionStorage = pxConnectionStorage;
        Asura_AI_Navigation_System::s_uNumConnections -= 1;

        // fix up the connection pointers in the waypoints again - this time to the new array
        // (this will automatically take into account the new number of connections in the from waypoint)
        {
            Asura_AI_NavConnection* pxConnectionListPtr = Asura_AI_Navigation_System::s_pxNavConnectionStorage;
            for ( i = 0; i < Asura_AI_Navigation_System::s_usNumWaypoints; ++i )
            {
                if ( Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections )
                {
                    Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_pxConnections = pxConnectionListPtr;
                    pxConnectionListPtr += Asura_AI_Navigation_System::s_pxWaypointStorage[ i ].m_ucNumConnections;
                }
            }
        }

        char acString[ 64 ];
        Asura_PrintfToBfr( acString, "Removed the link from waypoint %d to waypoint %d.", 64, true, iFromWaypoint, iToWaypoint );
        Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, acString );

        Tools_SaveAsura::SetWaypointingModified();
    }
    Asura_ResourcePopType();
}

void Tools_Waypoint_Editing::RemoveWaypointFromWaypointInfo( const int iWaypoint )
{
    Asura_ResourcePushType(ASURA_RESOURCE_WAYPOINT);
    {
        const bool	bIncludeCoverpointData	= (Asura_AI_Navigation_System::s_pxPerVertexCoverpointData != NULL);
        const int	iNumWaypoints			= Asura_AI_Navigation_System::s_usNumWaypoints;

        if(iNumWaypoints == 1)
        {
            Asura_ResourceDeallocateArray(Asura_AI_Navigation_System::s_pxWaypointStorage);
            Asura_AI_Navigation_System::s_pxWaypointStorage = NULL;

            if ( bIncludeCoverpointData )
            {
                Asura_ResourceDeallocateArray(Asura_AI_Navigation_System::s_pxPerVertexCoverpointData);
                Asura_AI_Navigation_System::s_pxPerVertexCoverpointData = NULL;
            }
        }
        else
        {
            // allocate new storage
            Asura_AI_NavVertex* pxWaypoints;
            Asura_ResourceAllocateArray(pxWaypoints, iNumWaypoints - 1);
            if ( pxWaypoints == NULL )
            {
                Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Out of memory before we could finish." );
                Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR,
                    "IMPORTANT: waypoint data is now broken - you should close tools without saving the waypointing data." );
                Asura_ResourcePopType();
                return;
            }

            // copy across all of the waypoints except the one to remove and move the new, smaller array back into the nav system
            Asura_AI_NavVertex* pxWaypointDest = pxWaypoints;
            for ( int i = 0; i < iNumWaypoints; ++i )
            {
                if ( i != iWaypoint )
                {
                    *pxWaypointDest++ = Asura_AI_Navigation_System::s_pxWaypointStorage[ i ];
                }
            }
            Asura_ResourceDeallocateArray(Asura_AI_Navigation_System::s_pxWaypointStorage);
            Asura_AI_Navigation_System::s_pxWaypointStorage = pxWaypoints;

            // also remove the waypoint from the coverpoint information, if necessary
            if ( bIncludeCoverpointData )
            {
                Asura_AI_PerVertexCoverpointData* pxCoverpoints;
                Asura_ResourceAllocateArray(pxCoverpoints,  iNumWaypoints - 1);
                if ( pxCoverpoints == NULL )
                {
                    Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Out of memory before we could finish." );
                    Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR,
                        "IMPORTANT: waypoint data is now broken - you should close tools without saving the waypointing data." );
                    Asura_ResourcePopType();
                    return;
                }

                Asura_AI_PerVertexCoverpointData* pxCoverpointDest = pxCoverpoints;
                for ( int i = 0; i < iNumWaypoints; ++i )
                {
                    if ( i != iWaypoint )
                    {
                        *pxCoverpointDest++ = Asura_AI_Navigation_System::s_pxPerVertexCoverpointData[ i ];
                    }
                }
                Asura_ResourceDeallocateArray(Asura_AI_Navigation_System::s_pxPerVertexCoverpointData);
                Asura_AI_Navigation_System::s_pxPerVertexCoverpointData = pxCoverpoints;
            }
        }

        // update the waypoint count in the main system
        Asura_AI_Navigation_System::s_usNumWaypoints -= 1;

        // go through the link info and decrement any link pointing to a waypoint id greater than this one
        for ( unsigned int u = 0; u < Asura_AI_Navigation_System::s_uNumConnections; ++u )
        {
            if ( Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u ].m_usIndex > iWaypoint )
            {
                Asura_AI_Navigation_System::s_pxNavConnectionStorage[ u ].m_usIndex -= 1;
            }
        }

        char acString[ 32 ];
        Asura_PrintfToBfr( acString, "Removed waypoint %d.", 32, true, iWaypoint );
        Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_DEFAULT, acString );

        Tools_SaveAsura::SetWaypointingModified();
    }
    Asura_ResourcePopType();
}

void Tools_Waypoint_Editing::UpdateWaypointLinkBlocks()
{
	const bool bAlreadyCreated = Tools_Mode_Waypointing::HasCreatedToolsEntities();
	if (!bAlreadyCreated)
	{
	    Tools_Entity_Object::CreateAllObjects();
	}

    bool bLinkBlocksChanged = false;
    Asura_ServerEntity_PhysicalObject::InitialiseWaypointLinkBlockCalculations();
	Asura_ServerEntity_WaypointBlockerVolume::InitialiseWaypointLinkBlockCalculations();
    
	Asura_Hashed_List_It< Asura_ServerEntity_PhysicalObject > xPhyObjectIter( Asura_ServerEntity_PhysicalObject::GetList() );
    while ( !xPhyObjectIter.Done() )
    {
        // Remove the old link blocks.
        xPhyObjectIter.GetCurrent()->UpdateWaypointLinkBlocks( true );
        // Calculate the new link blocks.
        bLinkBlocksChanged |= xPhyObjectIter.GetCurrent()->CalculateWaypointLinkBlocks();
        // Apply the new link blocks.
        xPhyObjectIter.GetCurrent()->UpdateWaypointLinkBlocks();

        xPhyObjectIter.Next();
    }
	Asura_Hashed_List_It< Asura_ServerEntity_WaypointBlockerVolume > xWBVIter( Asura_ServerEntity_WaypointBlockerVolume::GetList() );
	while ( !xWBVIter.Done() )
	{
		// Calculate the new link blocks.
		bLinkBlocksChanged |= xWBVIter.GetCurrent()->CalculateWaypointLinkBlocks();
		// Apply the new link blocks.
		xWBVIter.GetCurrent()->UpdateWaypointLinkBlocks();

		xWBVIter.Next();
	}

    Asura_ServerEntity_PhysicalObject::ShutdownWaypointLinkBlockCalculations();
	Asura_ServerEntity_WaypointBlockerVolume::ShutdownWaypointLinkBlockCalculations();

#ifdef ASURA_USE_CONTAINER_SYSTEM
	Asura_ServerEntity_Container::InitialiseWaypointLinkBlockCalculations();
	Asura_Hashed_List_It< Asura_ServerEntity_Container > xContIter( Asura_ServerEntity_Container::GetList() );

	for ( ; !xContIter.Done() ; xContIter.Next() )
	{
		Asura_ServerEntity_Container* pxServerEntity = xContIter.GetCurrent();

		// Remove the old link blocks.
		pxServerEntity->UpdateWaypointLinkBlocks( true );

		// Calculate the new link blocks.
        bLinkBlocksChanged |= pxServerEntity->CalculateWaypointLinkBlocks();

        // Apply the new link blocks.
        pxServerEntity->UpdateWaypointLinkBlocks();
	}
    Asura_ServerEntity_Container::ShutdownWaypointLinkBlockCalculations();
#endif //#ifdef ASURA_USE_CONTAINER_SYSTEM

    if ( bLinkBlocksChanged )
    {
        Tools_SaveAsura::SetEntityModified();
    }

	if (!bAlreadyCreated)
	{
	    Tools_Object_System::DeleteAllModeSpecific();
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////

bool Tools_Waypoint_Editing::AllowedToCreateWaypoint()
{
	List<Tools_Object*> xSelectedList;
	Tools_Object_System::ListSelectedObjectsOfType(TOT_UserDefinedWaypointPosition,xSelectedList);
	
	if ( xSelectedList.size() > 0 )
		return true;

	return false;
}

bool Tools_Waypoint_Editing::AllowedToAdjustWaypointRadius()
{
    return ( Tools_Waypoint_Selection::IsEnabled() && Tools_Waypoint_Selection::GetNumSelectedWaypoints() == 1 );
}

bool Tools_Waypoint_Editing::AllowedToEditLinking()
{
    return ( Tools_Waypoint_Selection::IsEnabled() && Tools_Waypoint_Selection::GetNumSelectedWaypoints() == 2 );
}

bool Tools_Waypoint_Editing::AllowedToDeleteWaypoints()
{
    return ( Tools_Waypoint_Selection::IsEnabled() && Tools_Waypoint_Selection::GetNumSelectedWaypoints() > 0 );
}

//////////////////////////////////////////////////////////////////////////////
//
// 
//
//////////////////////////////////////////////////////////////////////////////
