//////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_Processor_PS3.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////

#include "Core/Asura.h"

// This file's header.
#include "MeshProcessing/Tools_SmoothSkin_Processor_PS3.h"

// Engine independent.
#include "Animation/Asura_Hierarchy.h"
#include "ChunkSystem/Asura_Chunk_ResourceFile.h"
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "Core/Asura_HalfFloat.h"

// Engine PS3.
#include "PS3/ChunkSystem/Asura_Chunk_ResourceFile_PS3.h"
#include "PS3/ChunkSystem/Asura_Chunk_ResourceFile_PS3_Character.h"

// Tools.
#include "Load&Save/Tools_Level_Export.h"
#include "MeshProcessing/Asura_TriStripper.h"
#include "MeshProcessing/Tools_MeshRestartIndex.h"
#include "MeshProcessing/Tools_SmoothSkin_Processor.h"
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
#include "Materials/Tools_MaterialResponse_System.h"
#include "PS3/Tools_PS3_Export.h"
#endif

// C L A S S E S /////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
//
// Tools_SmoothSkin_Processor_PS3::ProcessedStrip
//
// Holds information on a strip as it's prepared by the skin processor
//
////////////////////////////////////////////////////////////////////////////////

class Tools_SmoothSkin_Processor_PS3::ProcessedStrip
{
public:

    // Constructor / destructor
	ProcessedStrip();
    ~ProcessedStrip();

    // Data

	Tools_SmoothSkin_Vertex*	m_pxVertices;
    u_int						m_uNumVertices;
    u_short*					m_pusIndices;
    u_int						m_uNumIndices;
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	Asura_Hash_ID				m_auMaterialResponseHashID[2];
	#else
	u_int						m_uMaterialID;
	#endif
	int							m_iElementID;
};

/******************************************************************************/
/* Tools_SmoothSkin_Processor_PS3::ProcessedStrip Class Functions		  */
/******************************************************************************/

Tools_SmoothSkin_Processor_PS3::ProcessedStrip::ProcessedStrip()
	: m_pxVertices( NULL )
	, m_uNumVertices(0)
	, m_pusIndices( NULL )
	, m_uNumIndices(0)
	, m_iElementID(0)
{
}


Tools_SmoothSkin_Processor_PS3::ProcessedStrip::~ProcessedStrip()
{
	delete[] m_pxVertices;
	delete[] m_pusIndices;
}
 
//
// Tools_SmoothSkin_Processor_PS3
//

Tools_SmoothSkin_Processor_PS3::Tools_SmoothSkin_Processor_PS3( void )
: m_pxSrcProcessor( NULL )
, m_pxProcessedStrips( NULL )
{
}

Tools_SmoothSkin_Processor_PS3::~Tools_SmoothSkin_Processor_PS3( void )
{
    delete[] m_pxProcessedStrips;
}

//BW put this in a shared header ffs...
// for now SPU code only knows how to load a fixed number of indices, and for safety we'll assume no shared verts when allocing this (i.e. we're using a batch size that's waaaaay smaller than we could use)
const int iSPU_SKINNING_BATCH_NUM_TRIS = (64*1024-12)/((68+2)*3);
//const int iSPU_SKINNING_BATCH_SIZE = 64*1024-12; //BW todo: use actual memory size so trade-offs between index & vertex data.

int MaxPartitions( int iNumTris )
{
	// get an upper bound for the number of partitions
	// assume no verts are shared

	return (iNumTris + iSPU_SKINNING_BATCH_NUM_TRIS-1) / iSPU_SKINNING_BATCH_NUM_TRIS;

	//BW todo use memory not fixed num tris	
	//return (iNumTris*3*(iVertSize+sizeof(u_short)) + iSpuBufferSize-1) / iSpuBufferSize;
}

bool Tools_SmoothSkin_Processor_PS3::Initialise( const Tools_SmoothSkin_Processor& xSrcProcessor )
{
    m_pxSrcProcessor = &xSrcProcessor;

    // (MP - 10.05.2006) I haven't bothered coping with multiple elements
    // since that functionality isn't used and I suspect never will be. It
    // would be fairly straightforward to extend the code below if necessary.
    //Asura_Assert( m_pxSrcProcessor->GetNumElements() == 1, "Multiple elements currently not supported" );

	// (DMR - 12.06.2006) Multiple elements now supported. Note that everything that
	// was called a 'material' before, is called a 'strip' now, which is consistent
	// with the PC version

	// First, count the number of non-empty elements
	u_int uNumStrips = 0; 
	for( u_int u = 0; u < m_pxSrcProcessor->GetNumMaterials(); ++u )
	{
		for( u_int v = 0; v < m_pxSrcProcessor->GetNumElements(); ++v )
		{
			const Tools_SmoothSkin_Material& xMaterial = m_pxSrcProcessor->GetMaterial( u );
			const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement( v );

			if ( xSubElement.GetNumTris() > 0 )
			{
				uNumStrips += MaxPartitions( xSubElement.GetNumTris() );
			}
		}
	}

	m_uNumProcessedStrips = uNumStrips;
	m_pxProcessedStrips   = new ProcessedStrip[uNumStrips];
	m_uTotalVertices	  = 0;
	m_uTotalIndices		  = 0;
	m_uTotalTriangles	  = 0;
	m_uNumProcessedStrips = 0;

	for( u_int u = 0; u < m_pxSrcProcessor->GetNumMaterials(); ++u )
	{
		for( u_int v = 0; v < m_pxSrcProcessor->GetNumElements(); ++v )
		{
			const Tools_SmoothSkin_Material& xMaterial = m_pxSrcProcessor->GetMaterial( u );
			const Tools_SmoothSkin_SubElement& xSubElement = xMaterial.GetSubElement( v );
			if (xSubElement.GetNumTris()>0)
			{
			
				// chop up into localised chunks that will fit in SPU
				Tri_List* aPartitionList = PartitionMesh(xSubElement.GetIndices(), xSubElement.GetNumTris(), m_pxSrcProcessor->GetNumVertices());
				Tri_List* pPartition = aPartitionList;
				
				while ( pPartition->GetIndices() )
				{
					ProcessedStrip& xProcessedStrip = m_pxProcessedStrips[m_uNumProcessedStrips];
					m_uNumProcessedStrips++;

					Tools_TriStripper::Initialise();
					Tools_TriStripper::SetOutputType( Tools_TriStripper::OUTPUT_TRISTRIP_WITH_RESET_INDICES );
					// TODO: Set cache size?
					
					xProcessedStrip.m_iElementID  = m_pxSrcProcessor->GetElementID(v);

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
					xProcessedStrip.m_auMaterialResponseHashID[0] = m_pxSrcProcessor->GetMaterial(u).GetMaterialResponseHashID(0);
					xProcessedStrip.m_auMaterialResponseHashID[1] = m_pxSrcProcessor->GetMaterial(u).GetMaterialResponseHashID(1);
#else
					xProcessedStrip.m_uMaterialID = m_pxSrcProcessor->GetMaterial(u).GetMaterialIDWithinSkin();
#endif
// 					xProcessedStrip.m_uNumIndices = Tools_TriStripper::CreateStrips( pPartition->GetNumTris(), pPartition->GetIndices(),
// 																					&xProcessedStrip.m_pusIndices );
//BW SPU skinning needs tri-list, for culling

					// just copy tri list
					xProcessedStrip.m_pusIndices = new u_short[pPartition->GetNumTris()*3];
					Asura_Mem::Copy( xProcessedStrip.m_pusIndices, pPartition->GetIndices(), pPartition->GetNumTris()*3*sizeof(u_short) );
					xProcessedStrip.m_uNumIndices = pPartition->GetNumTris()*3;
					
					//todo - optimise tri-list for RSX cache usage (there's a function in the edge libraries)
					//n.b. this is really unimportant because character's verts are processed by SPUs, so reloading the vertex cache doesn't involve re-shading verts.

					u_short* pusVtxPermutations;
					Tools_TriStripper::ComputeVertexPermutation( xProcessedStrip.m_uNumIndices, xProcessedStrip.m_pusIndices,
																 m_pxSrcProcessor->GetNumVertices(), &pusVtxPermutations,
																 &xProcessedStrip.m_uNumVertices );

					Tools_TriStripper::Shutdown();

					// Create new list of vertices just containing the ones used by
					// this material's strips, matching the newly modified index list.
					xProcessedStrip.m_pxVertices = new Tools_SmoothSkin_Vertex[xProcessedStrip.m_uNumVertices];
					for( u_int u = 0; u < xProcessedStrip.m_uNumVertices; ++u )
					{
						xProcessedStrip.m_pxVertices[u] = m_pxSrcProcessor->GetVertex( pusVtxPermutations[u] );
					}

					delete[] pusVtxPermutations;

					m_uTotalVertices  += xProcessedStrip.m_uNumVertices;
					m_uTotalIndices   += xProcessedStrip.m_uNumIndices;
					m_uTotalTriangles += pPartition->GetNumTris();
					
					pPartition++;
				}
				
				delete[] aPartitionList;   //this calls the destructors right?
			}
		}
	}

	// Check we're going to be able to put all of these vertices into a
	// single buffer using only 16-bit indices.
	{
		u_short usRestartIndex;
		Tools_MeshRestartIndex::IsValidForCurrentPlatform( usRestartIndex );
		Asura_Assert( m_uTotalVertices != usRestartIndex && m_uTotalVertices < (1 << 16), "Too many vertices" );
	}

	// Also, check we can put all of the indices in a single buffer. Max
	// vertex index is 0xFFFFFF on 360.
	Asura_Assert( m_uTotalIndices <= 0xFFFFFF, "" );

    return true;
}


inline int TriVertIdxOffset( int iVert, int iOffset )
{
	// where iVert is an index into a triangle list, with 3 indices per triangle, return the index of one of the other verts on this triangle.
	// if iOffset is 0 it will return iVert, if it is 1 or 2 it will be one of the other verts of this triangle (iVert +-1 or +-2)
	return (iVert/3)*3 + ((iVert+iOffset)%3);
}

// PartitionMesh
//
// Returns a NULL terminated list of pointers to subelements, which can then be stripped
Tri_List* Tools_SmoothSkin_Processor_PS3::PartitionMesh( const u_short* puIndices, int iNumTris, int iNumVertices )
{
	// build adjacency
	// Could use Tools_MeshAdjacency, but this is just triangles so it's somewhat simpler
	
	// build list of tris on each vertex
	u_short* auVertexAdjacency = new u_short[iNumVertices];
	u_short* auNextTriOnVertex = new u_short[iNumTris*3]; //linked list of polys, linked to next poly on each vertex, using indices instead of pointer
	
	Asura_Mem::Set( auVertexAdjacency, 0xff, iNumVertices*sizeof(u_short) );
	Asura_Mem::Set( auNextTriOnVertex, 0xff, iNumTris*3*sizeof(u_short) );
	
	int iMaxIdx = 0;
	for ( int i=0; i < iNumTris*3; i++ )
	{
		int iVert = static_cast<int>(puIndices[i]);
		
		// add poly to the linked list for this vert
		//note: we store the index of the vert of the poly (so poly is i/3), this way we can have each poly in 3 linked lists
		auNextTriOnVertex[i] = auVertexAdjacency[iVert];
		auVertexAdjacency[iVert] = static_cast<u_short>(i);
		
		iMaxIdx = (iMaxIdx>iVert)?iMaxIdx:iVert;
	}
	Asura_Assert( iMaxIdx < iNumVertices, "Tools_SmoothSkin_Processor_PS3::PartitionMesh - Vertex index out of range - max index is >= iNumVertices" );


	// now get list of adjacent triangles for each triangle
	u_short* auNextTriOnEdge = new u_short[iNumTris*3]; //linked list per edge per tri, edge 0 is from vert 0 to vert 1.

	Asura_Mem::Set( auNextTriOnEdge, 0xff, iNumTris*3*sizeof(u_short) );

	for ( int i=0; i < iNumTris*3; i++ )
	{
		// build linked list per edge
		u_short a = puIndices[i];
		u_short b = puIndices[TriVertIdxOffset(i,1)];

		// check all adjacent triangles on vert a to find one that uses b
		u_short uSearch = auVertexAdjacency[a];
		u_short uFoundEdge = 0xffff;
		while ( uSearch != 0xffff )
		{
			if ( uSearch != i )
			{
				if ( b == puIndices[TriVertIdxOffset(uSearch,1)] )
				{
					uFoundEdge = uSearch;
					break;
				}
				else if ( b == puIndices[TriVertIdxOffset(uSearch,2)] )
				{
					uFoundEdge = static_cast<u_short>(TriVertIdxOffset(uSearch,2)); //index of edge is same as that of preceding vert, i.e. edge0=vert0-vert1, edge1=vert1-vert2, edge2=vert2-vert0.
					break;
				}
			}
			uSearch = auNextTriOnVertex[uSearch];
		}
		
		if ( uFoundEdge != 0xffff )
		{
			// it is possible multiple tris share one edge, so make linked list behave
			// determine whether it should link on a or b of other tri...
			
			if ( auNextTriOnEdge[uFoundEdge] == 0xffff )
			{
				// make it a ring list
				auNextTriOnEdge[i] = uFoundEdge;
			}
			else
			{
				// edge has more than 2 polys
				auNextTriOnEdge[i] = auNextTriOnEdge[uFoundEdge];
			}
			auNextTriOnEdge[uFoundEdge] = static_cast<u_short>(i);
		}
	}
//BW todo put that^ in it's own func - in fact put all this partitioning in it's own file if it works

	int iMaxPartitions = MaxPartitions(iNumTris);
	
	Tri_List* axTriList = new Tri_List[iMaxPartitions+1];

	Tri_List* pxWriteTriList = axTriList;
	int iNumUnassignedTris = iNumTris;
	while ( iNumUnassignedTris > 0 )
	{
		int iNumTrisToAdd = Asura_Maths::Min( iNumUnassignedTris, iSPU_SKINNING_BATCH_NUM_TRIS );

		pxWriteTriList->m_puIndices = new u_short[iNumTrisToAdd*3];
		u_short* puWrite = pxWriteTriList->m_puIndices;
		
		pxWriteTriList->m_uNumTris = iNumTrisToAdd;
		
		while ( iNumTrisToAdd > 0 )
		{
			//BW todo - add best ones (greatest adjacentness) - see notes.
			// for now, just add in the same order they came in!
			for ( int i=0; i < 3; i++ )
			{
				*(puWrite++) = *(puIndices++);
			}
			
			iNumTrisToAdd--;
			iNumUnassignedTris--;
		}
		
		pxWriteTriList++;
	}
	pxWriteTriList->m_puIndices = NULL;
	
	delete[] auVertexAdjacency;
	delete[] auNextTriOnVertex;
	delete[] auNextTriOnEdge;
	
	return axTriList;
}

bool Tools_SmoothSkin_Processor_PS3::OutputToChunk( HANDLE pvFile ) const
{
    const Asura_Hierarchy_Skin* pxInputSkin = m_pxSrcProcessor->GetInputSkin();

    Asura_Chunk_Memory_Stream xStream;//( uChunkSize, pucData );
    xStream.SetSwapEndian( Tools_Level_Export::ShouldExportBigEndian() );

    // Chunk header.
    {
		const Asura_String	xName				= pxInputSkin->Name;
		xStream.WriteResourceChunk_Start( ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC, ASURA_RESOURCEFILE_TYPE_PS3_CHARACTER, ASURA_PS3_CHARACTER_CHUNK_VERSION, 0, xName );
    }

    // Skin header.
    xStream << pxInputSkin->m_uNameHashID;
	xStream << m_uNumProcessedStrips;
	xStream << m_uTotalVertices;
	xStream << m_uTotalIndices;
	xStream << m_uTotalTriangles;

	// Material headers.
	for( u_int uMaterial = 0; uMaterial < m_uNumProcessedStrips; ++uMaterial )
	{
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		Asura_Hash_ID uHash = Tools_PS3_ShaderExport::AddSmoothSkinShader( m_pxProcessedStrips[uMaterial].m_auMaterialResponseHashID[0], m_pxProcessedStrips[uMaterial].m_auMaterialResponseHashID[1] );
		xStream << uHash;
		xStream << m_pxProcessedStrips[uMaterial].m_auMaterialResponseHashID[0];
		xStream << m_pxProcessedStrips[uMaterial].m_auMaterialResponseHashID[1];
	#endif
		xStream << m_pxProcessedStrips[uMaterial].m_uNumIndices;
		xStream << m_pxProcessedStrips[uMaterial].m_iElementID;
	}

	// Vertex data.
	for( u_int uMaterial = 0; uMaterial < m_uNumProcessedStrips; ++uMaterial)
	{
		const ProcessedStrip& xProcessedStrip = m_pxProcessedStrips[uMaterial];
		for( u_int uVtx = 0; uVtx < xProcessedStrip.m_uNumVertices; ++uVtx )
		{
			WriteVtxToStream( xProcessedStrip.m_pxVertices[uVtx], xStream );
		}
	}

    // Index data.
    {
		u_short usRestartIndex;
		Tools_MeshRestartIndex::IsValidForCurrentPlatform( usRestartIndex );

		u_int uIndexOffset = 0;
		for( u_int uMaterial = 0; uMaterial < m_uNumProcessedStrips; ++uMaterial)
		{
			const ProcessedStrip& xProcessedStrip = m_pxProcessedStrips[uMaterial];
			for( u_int u = 0; u < xProcessedStrip.m_uNumIndices; ++u )
			{
				u_short usNextIdx = xProcessedStrip.m_pusIndices[u];

				// Offset into cumulative vertex buffer as we go.
				if( usNextIdx != usRestartIndex ) { usNextIdx = static_cast< u_short >( usNextIdx + uIndexOffset ); }

				xStream << usNextIdx;
			}

			uIndexOffset += xProcessedStrip.m_uNumVertices;
		}
    }

	xStream.WriteResourceChunk_End();
	const void* pData = xStream.GetStartAddress();
	u_int uChunkSize = xStream.GetPosition();
    // Write it all out.
    DWORD uBytesWritten;
    bool bOK = ( WriteFile(pvFile, pData, uChunkSize, &uBytesWritten, NULL)
                 &&
                 uBytesWritten == uChunkSize );

    return bOK;
}


void Tools_SmoothSkin_Processor_PS3::WriteVtxToStream( const Tools_SmoothSkin_Vertex& xVtx,
                                                       Asura_Chunk_Memory_Stream& xStream ) const
{
    #define WRITE_FLOAT_AS_HALF( fValue ) \
    { \
        Asura_HalfFloat xHalfVersion( fValue ); \
        xStream << xHalfVersion; \
    }
	#define WRITE_PACKED_BYTES( pucValues ) \
	{ \
		xStream << pucValues[0]; \
		xStream << pucValues[1]; \
		xStream << pucValues[2]; \
		xStream << pucValues[3]; \
	}

	#define WRITE_VECTOR_TO_UINT_PS3( xVector )														\
	{																								\
		const u_int uIntVersion = ( ( static_cast< u_int >( xVector.x * 511.0f ) & 0x7FF ) <<  0 )	\
								+ ( ( static_cast< u_int >( xVector.y * 511.0f ) & 0x7FF ) << 11 )	\
								+ ( ( static_cast< u_int >( xVector.z * 511.0f ) & 0x3FF ) << 22 );	\
		xStream << uIntVersion;																		\
	}

	#define WRITE_NORMALISED_VECTOR( fX, fY, fZ )\
	{ \
		const u_int uIntVersion = (static_cast< u_int >( fX * 511.0f ) <<  0) \
								+ (static_cast< u_int >( fY * 511.0f ) << 11) \
								+ (static_cast< u_int >( fZ * 511.0f ) << 22); \
		xStream << uIntVersion; \
	}

    const Asura_Vector_3& xPos = xVtx.GetPos();
    xStream << xPos.x;
    xStream << xPos.y;
    xStream << xPos.z;
	xStream <<  xVtx.GetAlpha();
    WRITE_FLOAT_AS_HALF( xVtx.GetU() );
    WRITE_FLOAT_AS_HALF( xVtx.GetV() );
    WRITE_FLOAT_AS_HALF( xVtx.GetU2() );
    WRITE_FLOAT_AS_HALF( xVtx.GetV2() );

    WRITE_PACKED_BYTES( xVtx.GetBoneIndices() );

    const float* pfBoneWeights = xVtx.GetBoneWeights();
    WRITE_FLOAT_AS_HALF( pfBoneWeights[0] );
    WRITE_FLOAT_AS_HALF( pfBoneWeights[1] );
    WRITE_FLOAT_AS_HALF( pfBoneWeights[2] );
    WRITE_FLOAT_AS_HALF( pfBoneWeights[3] );

    const Asura_Vector_3& xNormal = xVtx.GetNormal();
    WRITE_VECTOR_TO_UINT_PS3( xNormal );

    const Asura_Vector_3& xTangent = xVtx.GetTangent();
    WRITE_VECTOR_TO_UINT_PS3( xTangent);

    const Asura_Vector_3& xBinormal = xVtx.GetBinormal();
	WRITE_VECTOR_TO_UINT_PS3( xBinormal );

	#undef WRITE_FLOAT_AS_HALF
	#undef WRITE_PACKED_BYTES
	#undef WRITE_NORMALISED_VECTOR
	#undef WRITE_VECTOR_TO_UINT_PS3
}
