
///////////////////////////////////////////////////////////////////////////////
//
// File   : AVP_MovesMelee.cpp
// Author : Chris Wallace
// Date   : 23-Jan-2008
//
// Copyright (C) 2008 Rebellion, All Rights Reserved.
//
///////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Include files...
////////////////////////////////////////////////////////////////////////////////

#include "Core/Asura.h"


// This file's header...
#include "Moves/Vertebrates/AVP_MovesMelee.h"

// Asura independent header-files...
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_AnimationController_Animation.h"
#include "Blueprint/Asura_Blueprint.h"
#include "Blueprint/Asura_Blueprint_System.h"
#include "Interface/Asura_CommandConsole.h"
#include "Network/Asura_Message.h"

// Asura platform header-files...

// Game independent header-files...
#include "Actor/AVP_Actor_InputHandler.h"
#include "Actor/AVP_Actor_InteractionHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_Description.h"
#include "Actor/Vertebrates/AVP_Vertebrate_HUDHandler.h"
#include "Actor/Vertebrates/AVP_Vertebrate_MeleeCounter_System.h"
#include "Actor/Vertebrates/Human/AVP_Human_AnimationHandler.h"
#include "Actor/Vertebrates/Human/AVP_Human_CameraHandler.h"
#include "Actor/Vertebrates/Human/AVP_Human_InputHandler.h"
#include "Animation/AVP_AnimFlags.h"
#include "Blueprint/AVP_Blueprint_HashIDs.h"
#include "Entity/AVP_Messages.h"
#include "Entity/AVP_Entity_Classification.h"
#include "Entity/Damage/AVP_EntityDamage_Messages.h"
#include "Entity/PhysicalObject/Actor/AVP_ClientEntity_Actor.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/Human/AVP_ClientActor_Human.h"
#include "Entity/PhysicalObject/Actor/Vertebrates/AVP_Vertebrate_Messages.h"
#include "Entity/PhysicalObject/Projectile/AVP_ClientProjectile_HumanFlare.h"

#include "Moves/AVP_Movement_System.h"
#include "Rendering/AVP_Overlay_Render.h"

// Game platform header-files...

// Tools header-files...


////////////////////////////////////////////////////////////////////////////////
// Constant declarations...
////////////////////////////////////////////////////////////////////////////////

static const float g_fTweenInDuration = 0.05f;
static const float g_fTweenOutDuration = 0.2f;

////////////////////////////////////////////////////////////////////////////////
// Static variable declarations...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Private functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Protected functions...
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////////////
// Public functions...
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::AVP_Move_Melee
// Author       : Chris Wallace
// Date         : 23-Jan-2008
////////////////////////////////////////////////////////////////////////////////
DVSVAR float dvs_fDefaultMeleeAimDirectionAngVel = Asura_Maths::Deg360; //360 degrees per second
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
u_int AVP_Move_Melee::s_uConsolevarCount = 0;
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

AVP_Move_Melee::AVP_Move_Melee(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash)
: PARENT(xParentSystem, uHash)
, m_uTargetGuid(ASURA_GUID_UNREGISTERED)
, m_uNumFirstPersonVariations(0)
, m_uNumThirdPersonVariations(0)
, m_uVariation(0)
, m_uHitEntity(ASURA_GUID_UNREGISTERED)
, m_uPendingGrappleBlueprint(ASURA_HASH_ID_UNSET)
, m_uPendingVariation(0)
, m_uPendingAttackType(AVP_MELEEATTACK_TYPE_NONE)
, m_uMeleeFlags(0)
{
	// Since only moves triggered by ShouldStart ever send move messages, we should be fine to set this on all melee moves.
	m_bUseServerStartMessage = true;

	for (u_int u=0; u<uAVP_MOVE_MELEE_MAX_VARIATIONS; ++u)
	{
		m_auFirstPersonAnimHashes[u] = ASURA_HASH_ID_UNSET;
		m_auThirdPersonAnimHashes[u] = ASURA_HASH_ID_UNSET;
	}

	// Set a sensible default lerp time scale
	SetAimDirectionAngVel(dvs_fDefaultMeleeAimDirectionAngVel);

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_uConsolevarCount == 0)
	{
		Asura_CommandConsole::AddVar(L"AVP.Melee.AimDirectionAngVel", &dvs_fDefaultMeleeAimDirectionAngVel, 0.0f, Asura_Maths::LargeFloat, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	}
	++s_uConsolevarCount;
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
}

/// End of function 'AVP_Move_Melee::AVP_Move_Melee'. ///

AVP_Move_Melee::~AVP_Move_Melee()
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	--s_uConsolevarCount;
	if (s_uConsolevarCount == 0)
	{
		Asura_CommandConsole::RemoveVar(L"AVP.Melee.AimDirectionAngVel");
	}
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
}

void AVP_Move_Melee::SetVariation(u_int uVariation)
{
	u_int uMax = Asura_Maths::Max(m_uNumFirstPersonVariations, m_uNumThirdPersonVariations);

	Asura_Assert(uVariation < uMax, "Variation out of range.");
	if (uMax == 0) { return; }

	m_uVariation = Asura_Maths::Min(uVariation, uMax-1);
	SetAnimHashes(m_auFirstPersonAnimHashes[m_uVariation], m_auThirdPersonAnimHashes[m_uVariation]);
	// By default, follow-up moves use the same variation, so set the pending to that.
	m_uPendingVariation = m_uVariation;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::SetAnimHashesForVariations
// Author       : Chris Wallace
// Date         : 29-Jan-2008
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee::SetFirstPersonAnimHashesForVariations(const Asura_Hash_ID* puHashes, u_int uNumVariations)
{
	Asura_Assert(uNumVariations <= uAVP_MOVE_MELEE_MAX_VARIATIONS, "Melee move variation limit too low, ask a coder to increase it.");
	uNumVariations = Asura_Maths::Min(uNumVariations, uAVP_MOVE_MELEE_MAX_VARIATIONS);

	m_uVariation = 0;
	m_uNumFirstPersonVariations = uNumVariations;

	u_int u = 0;
	for (; u<uNumVariations; ++u)
	{
		m_auFirstPersonAnimHashes[u] = puHashes[u];
	}
	for (; u<uAVP_MOVE_MELEE_MAX_VARIATIONS; ++u)
	{
		m_auFirstPersonAnimHashes[u] = ASURA_HASH_ID_UNSET;
	}
}

void AVP_Move_Melee::SetThirdPersonAnimHashesForVariations(const Asura_Hash_ID* puHashes, u_int uNumVariations)
{
	Asura_Assert(uNumVariations <= uAVP_MOVE_MELEE_MAX_VARIATIONS, "Melee move variation limit too low, ask a coder to increase it.");
	uNumVariations = Asura_Maths::Min(uNumVariations, uAVP_MOVE_MELEE_MAX_VARIATIONS);

	m_uVariation = 0;
	m_uNumThirdPersonVariations = uNumVariations;

	u_int u = 0;
	for (; u<uNumVariations; ++u)
	{
		m_auThirdPersonAnimHashes[u] = puHashes[u];
	}
	for (; u<uAVP_MOVE_MELEE_MAX_VARIATIONS; ++u)
	{
		m_auThirdPersonAnimHashes[u] = ASURA_HASH_ID_UNSET;
	}
}

void AVP_Move_Melee::SetupAnimHashesFromBlueprint(const Asura_Blueprint* pxFirstPersonBlueprint, const Asura_Blueprint* pxThirdPersonBlueprint, const char* szParameterName)
{
	if (!pxFirstPersonBlueprint && !pxThirdPersonBlueprint) { return; }

	// First check if there's no enumerated set.
	Asura_Value xValue;
	if (pxFirstPersonBlueprint)
	{
		if (pxFirstPersonBlueprint->GetParameterValueByHashID(Asura_GetHashID(szParameterName), xValue))
		{
			// No need to bother enumerating this then.
			Asura_Hash_ID uHash = xValue.GetHashIDValue();
			SetFirstPersonAnimHashesForVariations(&uHash, 1);
		}
		else
		{
			// The parameter doesn't exist with the exact name, try enumerating it.
			AVP_Blueprint_ParameterEnumerator_MeleeVariations xEnumerator(*this, false);

			pxFirstPersonBlueprint->EnumerateParameters(szParameterName, 2, xEnumerator);
		}
	}

	if (pxThirdPersonBlueprint)
	{
		if (pxThirdPersonBlueprint->GetParameterValueByHashID(Asura_GetHashID(szParameterName), xValue))
		{
			// No need to bother enumerating this then.
			Asura_Hash_ID uHash = xValue.GetHashIDValue();
			SetThirdPersonAnimHashesForVariations(&uHash, 1);
		}
		else
		{
			// The parameter doesn't exist with the exact name, try enumerating it.
			AVP_Blueprint_ParameterEnumerator_MeleeVariations xEnumerator(*this, true);

			pxThirdPersonBlueprint->EnumerateParameters(szParameterName, 2, xEnumerator);
		}
	}

	SetVariation(0);
}

void AVP_Move_Melee::SetupAnimHashesFromBlueprint(const Asura_Blueprint* pxFirstPersonBlueprint, const Asura_Blueprint* pxThirdPersonBlueprint, const Asura_Hash_ID uParameterHash)
{
	if (!pxFirstPersonBlueprint && !pxThirdPersonBlueprint) { return; }

	// If we have a parameter hash ID, we're not going through the enumeration.
	Asura_Value xValue;
	if (pxFirstPersonBlueprint && pxFirstPersonBlueprint->GetParameterValueByHashID(uParameterHash, xValue))
	{
		Asura_Hash_ID uHash = xValue.GetHashIDValue();
		SetFirstPersonAnimHashesForVariations(&uHash, 1);
	}

	if (pxThirdPersonBlueprint && pxThirdPersonBlueprint->GetParameterValueByHashID(uParameterHash, xValue))
	{
		Asura_Hash_ID uHash = xValue.GetHashIDValue();
		SetThirdPersonAnimHashesForVariations(&uHash, 1);
	}

	SetVariation(0);
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::HandleMessage
// Author       : Chris Wallace
// Date         : 23-Jan-2008
////////////////////////////////////////////////////////////////////////////////

bool AVP_Move_Melee::HandleMessage(const Asura_Message* pxMessage)
{
	switch (pxMessage->GetMsgID())
	{
	case AVP_MSGID_COUNTER_MELEE:
		{
			SetHasBeenCountered(true);
			return true;
		}

	case AVP_MSGID_MELEE_ATTACK_HIT:
		{
			const AVP_Msg_Melee_Attack_Hit* pxData = static_cast<const AVP_Msg_Melee_Attack_Hit*>(pxMessage->GetData());
			// We can't afford to skip a valid hit entity, so only set the flag if we have data.
			if (pxData)
			{
				m_uHitEntity = pxData->GetHitEntity();
				m_uPendingVariation = pxData->GetPendingVariation();
				m_uPendingAttackType = pxData->GetPendingAttackType();
				SetHasDoneDamage(true);
			}
			return true;
		}

	case AVP_MSGID_MELEE_ATTACK_PREPAREGRAPPLE:
		{
			const AVP_Msg_Melee_Attack_PrepareGrapple* pxData = static_cast<const AVP_Msg_Melee_Attack_PrepareGrapple*>(pxMessage->GetData());
			// We can't afford to skip a valid hit entity, so only set the flag if we have data.
			if (pxData)
			{
				m_uHitEntity = pxData->GetHitEntity();
				m_uPendingGrappleBlueprint = pxData->GetGrappleDescriptionBlueprint();
				SetHasDoneDamage(true);
			}
			return true;
		}

	default:
		{
			return PARENT::HandleMessage(pxMessage);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::Start
// Author       : Chris Wallace
// Date         : 23-Jan-2008
////////////////////////////////////////////////////////////////////////////////

bool AVP_Move_Melee::Start(const Asura_Msg_Move* pxStartMsg)
{
	SetHasBeenCountered(false);
	SetHasDoneDamage(false);

	m_uHitEntity = ASURA_GUID_UNREGISTERED;

	if (!m_bDontClearPendingGrappleBlueprint)
	{
		m_uPendingGrappleBlueprint = ASURA_HASH_ID_UNSET;
	}

	m_uTargetGuid = ASURA_GUID_UNREGISTERED;
	if (pxStartMsg)
	{
		const AVP_Msg_Melee_Move* pxMeleeStart = static_cast<const AVP_Msg_Melee_Move*>(pxStartMsg);
		m_uTargetGuid = pxMeleeStart->GetTargetGuid();
		m_uPendingAttackType = pxMeleeStart->GetPendingAttackType();
		SetVariation( pxMeleeStart->GetVariation() );
	}
	else
	{
		const AVP_Move* pxCurrentMove = static_cast<const AVP_Move*>(GetParentSystem().GetCurrentMove());
		if (pxCurrentMove)
		{
			m_uTargetGuid = pxCurrentMove->GetTargetGuid();
			// We can't propagate the pending attack type here, that'll have to be done in subclasses which can guarantee the predecessor was a melee move.
		}
	}

	// Set a sensible default lerp time scale
	SetAimDirectionAngVel(dvs_fDefaultMeleeAimDirectionAngVel);

	return(PARENT::Start(pxStartMsg));
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::Abort
// Author       : Anthony Gregson
// Date         : 16/05/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee::Abort()
{
	PARENT::Abort();

	if (m_bLookAtEnemyByAnimFlag)
	{
		m_bLookAtEnemyByAnimFlag = false;
		SetAimDirectionUsage(static_cast<AVP_ClientEntity_Actor::AimDirectionUsage>(m_eAimDirectionUsageBackup));
	}
}


////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::Finish
// Author       : Anthony Gregson
// Date         : 16/05/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee::Finish()
{
	PARENT::Finish();

	if (m_bLookAtEnemyByAnimFlag)
	{
		m_bLookAtEnemyByAnimFlag = false;
		SetAimDirectionUsage(static_cast<AVP_ClientEntity_Actor::AimDirectionUsage>(m_eAimDirectionUsageBackup));
	}
}


////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  AVP_Move_Melee
//  NAME     :  SendStartMsg()
//  CREATED  :  JG (12.8.2008)
////////////////////////////////////////////////////////////////////////////////
void AVP_Move_Melee::SendStartMsg()
{
	if (!m_bUseServerStartMessage)
	{
		if (!m_bSuppressStartMessage)
		{
			Asura_Message*	pxMessage	= CreateStartMsg();

			if(pxMessage != NULL)
			{
				Asura_Network::Send(pxMessage);
			}
		}
		return;
	}

	if (AVP_Settings::GetIsServer())
	{
		Asura_Message*	pxMessage	= CreateStartMsg();

		if(pxMessage != NULL)
		{
			Asura_Network::Send(pxMessage);
		}
	}
}


////////////////////////////////////////////////////////////////////////////////
//  CLASS    :  AVP_Move_Melee
//  NAME     :  CreateStartMsg()
//  CREATED  :  JG (13.8.2008)
////////////////////////////////////////////////////////////////////////////////
Asura_Message* AVP_Move_Melee::CreateStartMsg()
{
	Asura_Message* pxMessage = Asura_Network::CreateMessage(ASURA_MSG_MOVE_START, sizeof(AVP_Msg_Melee_Move));

	if (pxMessage)
	{
		AVP_Msg_Melee_Move* pxData = static_cast<AVP_Msg_Melee_Move*>(pxMessage->GetData());
		const Asura_Guid uOwnerGuid	= m_xParentSystem.GetOwnerGuid();

		if(pxData != NULL)
		{
			// Parent data
			pxData->ClearFlags();
			pxData->SetMoveID(m_uHash);
			pxData->SetIgnoredByOwner(true);

			// Our data
			pxData->SetForceAbort(true);
			pxData->SetVariation(GetVariation());
			pxData->SetPendingAttackType(GetPendingAttackType());
			pxData->SetTargetEntityIsPuppet(false);
			pxData->SetTargetGuid(GetTargetGuid()); 
			pxData->SetSuckPermitted(true); 
		}

		pxMessage->SetToGuid(uOwnerGuid);
		pxMessage->SetFromGuid(uOwnerGuid);
		pxMessage->UseGuaranteedDelivery();
	}

	return pxMessage; 
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee::UpdateStates
// Author       : Anthony Gregson
// Date         : 16/05/08
////////////////////////////////////////////////////////////////////////////////

DEBUGVAR float dbg_fMeleeAimDirectionLimitAngle = 90.0f * Asura_Maths::DegToRadConstant;

void AVP_Move_Melee::UpdateStates(const float fTimeStep)
{
	PARENT::UpdateStates(fTimeStep);

	// Check for the look-at anim flag.
	const AVP_ClientActor_Vertebrate* pxEnemy = AVP_ClientActor_Vertebrate::FindEntity(m_uTargetGuid);
	if (pxEnemy && ((GetAnimFlagsForLastFrame() & AVP_ANIMFLAG_ACTIVATE_TARGET_FOLLOW) != 0))
	{
		if (!m_bLookAtEnemyByAnimFlag)
		{
			m_bLookAtEnemyByAnimFlag = true;
			m_eAimDirectionUsageBackup = GetAimDirectionUsage();
			SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_SET_DIRECTLY);
		}

		// Update to our target entity
		const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
		Asura_Vector_3 xTargetPos;
		Asura_Vector_3 xOwnerPos;
		if (xOwner.GetMarkerLocationFromHash(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_EYES), &xOwnerPos, NULL)
			&& pxEnemy->GetMarkerLocationFromHash(AVP_Vertebrate_Description::GetMarkerPointHashID(AVP_Vertebrate_Description::MPID_TARGET_2), &xTargetPos, NULL))
		{
			m_xAimDirection = Asura_AVector3(xTargetPos - xOwnerPos);
			m_xAimDirection.Normalise();

			// Keep this within a sensible limit.
			const Asura_Matrix_3x3& xOri = xOwner.GetCurrentOrientation();
			const Asura_AVector3 xForward(xOri.e13, xOri.e23, xOri.e33);

			Asura_Quat xRot;
			xRot.RotationFromAToB(xForward, m_xAimDirection);
			Asura_AVector3 xAxis;
			float fAngle = xRot.ToRotation(xAxis);
			if (fAngle < 0.0f)
			{
				xAxis *= -1.0f;
				fAngle = -fAngle;
			}

			if (fAngle > dbg_fMeleeAimDirectionLimitAngle)
			{
				xRot.FromRotation(dbg_fMeleeAimDirectionLimitAngle, xAxis);
				m_xAimDirection = xRot.RotateVector(xForward);
			}
		}
	}
	else
	{
		if (m_bLookAtEnemyByAnimFlag)
		{
			m_bLookAtEnemyByAnimFlag = false;
			SetAimDirectionUsage(static_cast<AVP_ClientEntity_Actor::AimDirectionUsage>(m_eAimDirectionUsageBackup));
		}
	}

}


////////////////////////////////////////////////////////////////////////////////
// Generic melee attack sequence (start -> hit/miss/countered).
// Since so many attacks will follow this pattern we can share a lot of the code.
////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
// Class	    : AVP_Move_Melee_GenericAttack_Start
// Author       : Anthony Gregson
// Date         : 15/02/08
////////////////////////////////////////////////////////////////////////////////

// TODO - this should probably move into character-specific blueprints, and be accessed through the client actor.
float AVP_Move_Melee_GenericAttack_Start::s_fMinSuckDistance = 0.0f;
float AVP_Move_Melee_GenericAttack_Start::s_fSuckTargetDistance = 1.5f;

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
u_int AVP_Move_Melee_GenericAttack_Start::s_uDVSVarCount = 0;
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM

AVP_Move_Melee_GenericAttack_Start::AVP_Move_Melee_GenericAttack_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																		const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const char* szBaseParameterName,
																		const Asura_Hash_ID uHitReturnMove, const Asura_Hash_ID uCounteredReturnMove, const Asura_Hash_ID uMissReturnMove)
: PARENT(xParentSystem, uHash)
, m_bTargetChaseActive(false)
, m_bAllowSuckToEnemy(true)
{
	SetAlwaysCheckNextMoves(false);

	AddNextMove(uHitReturnMove);
	AddNextMove(uCounteredReturnMove);
	AddNextMove(uMissReturnMove);

	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, szBaseParameterName);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, 0.0f, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_uDVSVarCount == 0)
	{
		Asura_CommandConsole::AddVar(L"AVP.Melee.MinSuckDistance", &s_fMinSuckDistance, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
		Asura_CommandConsole::AddVar(L"AVP.Melee.SuckTargetDistance", &s_fSuckTargetDistance, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	}
	++s_uDVSVarCount;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}

AVP_Move_Melee_GenericAttack_Start::AVP_Move_Melee_GenericAttack_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																		const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const Asura_Hash_ID uParameterHash,
																		const Asura_Hash_ID uHitReturnMove, const Asura_Hash_ID uCounteredReturnMove, const Asura_Hash_ID uMissReturnMove)
: PARENT(xParentSystem, uHash)
, m_bTargetChaseActive(false)
, m_bAllowSuckToEnemy(true)
{
	SetAlwaysCheckNextMoves(false);

	AddNextMove(uHitReturnMove);
	AddNextMove(uCounteredReturnMove);
	AddNextMove(uMissReturnMove);

	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, uParameterHash);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, 0.0f, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_NOT_USED);

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_uDVSVarCount == 0)
	{
		Asura_CommandConsole::AddVar(L"AVP.Melee.MinSuckDistance", &s_fMinSuckDistance, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
		Asura_CommandConsole::AddVar(L"AVP.Melee.SuckTargetDistance", &s_fSuckTargetDistance, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN);
	}
	++s_uDVSVarCount;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}

AVP_Move_Melee_GenericAttack_Start::AVP_Move_Melee_GenericAttack_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																		const Asura_Hash_ID uHitReturnMove, const Asura_Hash_ID uCounteredReturnMove, const Asura_Hash_ID uMissReturnMove)
: PARENT(xParentSystem, uHash)
, m_bTargetChaseActive(false)
, m_bAllowSuckToEnemy(true)
{
	SetAlwaysCheckNextMoves(false);

	AddNextMove(uHitReturnMove);
	AddNextMove(uCounteredReturnMove);
	AddNextMove(uMissReturnMove);

	// By default, these moves will try to aim towards our melee target
	SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_SET_DIRECTLY);

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_uDVSVarCount == 0)
	{
		Asura_CommandConsole::AddVar(L"AVP.Melee.MinSuckDistance", &s_fMinSuckDistance, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
		Asura_CommandConsole::AddVar(L"AVP.Melee.SuckTargetDistance", &s_fSuckTargetDistance, 0.0f, 10.0f, Asura_CommandConsole::REMOVE_AT_SHUTDOWN );
	}
	++s_uDVSVarCount;
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}

AVP_Move_Melee_GenericAttack_Start::~AVP_Move_Melee_GenericAttack_Start()
{
#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_uDVSVarCount > 0)
	{
		--s_uDVSVarCount;
		if (s_uDVSVarCount == 0)
		{
			Asura_CommandConsole::RemoveVar(L"AVP.Melee.MinSuckDistance");
			Asura_CommandConsole::RemoveVar(L"AVP.Melee.SuckTargetDistance");
		}
	}
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}


bool AVP_Move_Melee_GenericAttack_Start::Start(const Asura_Msg_Move* pxStartMsg)
{
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	xOwner.GetAimDirection(m_xAimDirection);
	m_bTargetChaseActive = false;
	m_xPositionInterpolation.SetVelocityOperation(Asura_MoveData_PositionInterpolation::ADD_VELOCITY);

	// Start moves get a variation from message
	if (pxStartMsg)
	{
		const AVP_Msg_Melee_Move* pxMeleeStart = static_cast<const AVP_Msg_Melee_Move*>(pxStartMsg);

		SetVariation(pxMeleeStart->GetVariation());

		m_uTargetGuid = pxMeleeStart->GetTargetGuid();
		// If we have a target, aim towards it.
		const AVP_ClientEntity_Actor* pxEnemy = AVP_ClientEntity_Actor::FindEntity(m_uTargetGuid);
		if (pxEnemy)
		{
			if (m_pxAVPAnimationData)
			{
				// We need to ensure the correct anim hash is set up, as it's about to get used.
				if (xOwner.IsUsingFirstPersonSkin())
				{
					m_pxAVPAnimationData->SetAnimHash(m_uFirstPersonAnimHash);
				}
				else
				{
					m_pxAVPAnimationData->SetAnimHash(m_uThirdPersonAnimHash);
				}
			}

			// Also suck us towards the enemy if we're close enough (but not too close).
			// Only do this if the subclass hasn't already set up some kind of position interpolation.
			if (GetPositionUsage() == POSITION_NOT_USED && pxMeleeStart->GetSuckPermitted() && m_bAllowSuckToEnemy)
			{
				bool bHasAnimation = m_pxAVPAnimationData && m_pxAVPAnimationData->HasValidAnimation();
				float fDuration = bHasAnimation ? m_pxAVPAnimationData->GetAnimationDuration() : 0.3f;

				const Asura_Vector_3& xEnemyPos = pxEnemy->GetCurrentPosition();
				const Asura_Vector_3& xOwnerPos = xOwner.GetCurrentPosition();
				Asura_Vector_3 xFromEnemy = xOwnerPos - xEnemyPos;
				const float fMagSquared = xFromEnemy.MagnitudeSquared();

				if (fMagSquared > 0.001f)
				{
					xFromEnemy.Normalise();
					Asura_Vector_3 xTargetPos = xEnemyPos + (xFromEnemy * s_fSuckTargetDistance);
					Asura_Vector_3 xEnemyVel;
					pxEnemy->GetCurrentVelocity(xEnemyVel);
					xTargetPos += xEnemyVel * Asura_Timers::GetGameTimeSlice();
					m_xPositionInterpolation.SetInitial(xOwnerPos);
					m_xPositionInterpolation.SetFinal(xTargetPos);

					m_xPositionInterpolation.SetDuration(fDuration);
					m_xPositionInterpolation.SetVelocityOperation(Asura_MoveData_PositionInterpolation::IGNORE_VELOCITY);

					m_xParentSystem.SetCurrentMovePosition(xOwnerPos);
					InitialisePositionInterpolation(m_xPositionInterpolation);
					// We just want to use the velocity.
					SetPositionUsage(POSITION_SET_FROM_VELOCITY);
					m_bTargetChaseActive = true;
				}
			}
		}
	}
	else
	{
		// no message? variation 1 it is then.
		SetVariation(0);
	}

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_GenericAttack_Start::UpdateStates(const float fTimeStep)
{
	// If we're using the target chasing, update the target position to where they now are.
	if (m_bTargetChaseActive)
	{
		const AVP_ClientEntity_Actor* pxEnemy = AVP_ClientEntity_Actor::FindEntity(m_uTargetGuid);
		if (pxEnemy)
		{
			Asura_Vector_3 xTargetPos;
			const Asura_Vector_3& xEnemyPos = pxEnemy->GetCurrentPosition();
			const Asura_Vector_3& xOwnerPos = GetAVPParentSystem().GetActorOwner().GetCurrentPosition();
			Asura_Vector_3 xFromEnemy = xOwnerPos - xEnemyPos;

			if (xFromEnemy.MagnitudeSquared() > 0.00001f)
			{
				xFromEnemy.Normalise();
				xTargetPos = xEnemyPos + (xFromEnemy * s_fSuckTargetDistance);
				Asura_Vector_3 xEnemyVel;
				pxEnemy->GetCurrentVelocity(xEnemyVel);
				xTargetPos += xEnemyVel * fTimeStep;
				m_xPositionInterpolation.SetFinal(xTargetPos);
			}
		}
	}

	PARENT::UpdateStates(fTimeStep);
}

void AVP_Move_Melee_GenericAttack_Start::Finish()
{
	PARENT::Finish();

	if (m_bTargetChaseActive)
	{
		SetPositionUsage(POSITION_NOT_USED);
		m_bTargetChaseActive = false;
	}
}

void AVP_Move_Melee_GenericAttack_Start::Abort()
{
	PARENT::Abort();

	if (m_bTargetChaseActive)
	{
		SetPositionUsage(POSITION_NOT_USED);
		m_bTargetChaseActive = false;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Class	    : AVP_Move_Melee_GenericAttack_Return_Hit
// Author       : Anthony Gregson
// Date         : 15/02/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_GenericAttack_Return_Hit::AVP_Move_Melee_GenericAttack_Return_Hit(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																				const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const char* szBaseParameterName)
: PARENT(xParentSystem, uHash)
{
	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, szBaseParameterName);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}
}

AVP_Move_Melee_GenericAttack_Return_Hit::AVP_Move_Melee_GenericAttack_Return_Hit(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																				const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const Asura_Hash_ID uParameterHash)
: PARENT(xParentSystem, uHash)
{
	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, uParameterHash);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}
}

AVP_Move_Melee_GenericAttack_Return_Hit::AVP_Move_Melee_GenericAttack_Return_Hit(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash)
: PARENT(xParentSystem, uHash)
{
}

bool AVP_Move_Melee_GenericAttack_Return_Hit::ShouldStart() const
{
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		if (pxCurrentMove->HasBeenCountered())
		{
			return false;
		}

		return pxCurrentMove->HasDoneDamage();
	}

	return false;
}

bool AVP_Move_Melee_GenericAttack_Return_Hit::Start(const Asura_Msg_Move* pxStartMsg)
{
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		SetVariation(pxCurrentMove->GetPendingVariation()); // Get variation here
		m_uPendingAttackType = pxCurrentMove->GetPendingAttackType();
	}
	
	return PARENT::Start(pxStartMsg);
}

////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_GenericAttack_Return_Countered
// Author       : Anthony Gregson
// Date         : 15/02/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_GenericAttack_Return_Countered::AVP_Move_Melee_GenericAttack_Return_Countered(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																							const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const char* szBaseParameterName)
: PARENT(xParentSystem, uHash)
{
	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, szBaseParameterName);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	m_bClearQueuedMeleeRequests = true;
}

AVP_Move_Melee_GenericAttack_Return_Countered::AVP_Move_Melee_GenericAttack_Return_Countered(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																							const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const Asura_Hash_ID uParameterHash)
: PARENT(xParentSystem, uHash)
{
	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, uParameterHash);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	m_bClearQueuedMeleeRequests = true;
}

AVP_Move_Melee_GenericAttack_Return_Countered::AVP_Move_Melee_GenericAttack_Return_Countered(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash)
: PARENT(xParentSystem, uHash)
{
	m_bClearQueuedMeleeRequests = true;
}

bool AVP_Move_Melee_GenericAttack_Return_Countered::ShouldStart() const
{
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		if (pxCurrentMove->HasBeenCountered())
		{
			return true;
		}
	}
	return false;
}

bool AVP_Move_Melee_GenericAttack_Return_Countered::Start(const Asura_Msg_Move* pxStartMsg)
{
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		SetVariation(pxCurrentMove->GetPendingVariation()); // Get variation here
		m_uPendingAttackType = pxCurrentMove->GetPendingAttackType();
	}

	return PARENT::Start(pxStartMsg);
}

////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_GenericAttack_Return_Miss
// Author       : Anthony Gregson
// Date         : 15/02/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_GenericAttack_Return_Miss::AVP_Move_Melee_GenericAttack_Return_Miss(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																					const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const char* szBaseParameterName)
: PARENT(xParentSystem, uHash)
{
	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, szBaseParameterName);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	m_bClearQueuedMeleeRequests = true;
}

AVP_Move_Melee_GenericAttack_Return_Miss::AVP_Move_Melee_GenericAttack_Return_Miss(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash,
																					const Asura_Hash_ID uBlueprintType, const Asura_Hash_ID uFirstPersonBlueprintHash, const Asura_Hash_ID uThirdPersonBlueprintHash, const Asura_Hash_ID uParameterHash)
: PARENT(xParentSystem, uHash)
{
	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uBlueprintType, uThirdPersonBlueprintHash);
	if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
	{
		SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, uParameterHash);
	}

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	m_bClearQueuedMeleeRequests = true;
}

AVP_Move_Melee_GenericAttack_Return_Miss::AVP_Move_Melee_GenericAttack_Return_Miss(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uHash)
: PARENT(xParentSystem, uHash)
{
	m_bClearQueuedMeleeRequests = true;
}

bool AVP_Move_Melee_GenericAttack_Return_Miss::ShouldStart() const
{
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		if (pxCurrentMove->HasBeenCountered())
		{
			return false;
		}

		return !pxCurrentMove->HasDoneDamage();
	}

	return false;
}

bool AVP_Move_Melee_GenericAttack_Return_Miss::Start(const Asura_Msg_Move* pxStartMsg)
{
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		SetVariation(pxCurrentMove->GetPendingVariation()); // Get variation here
		m_uPendingAttackType = pxCurrentMove->GetPendingAttackType();
	}

	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_DefaultAttack_Start
// Author       : Anthony Gregson
// Date         : 24/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_DefaultAttack_Start::AVP_Move_Melee_DefaultAttack_Start(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash(),
		 AVP_Move_Melee_DefaultAttack_Return_Hit::GetStaticHash(), AVP_Move_Melee_DefaultAttack_Return_Countered::GetStaticHash(), AVP_Move_Melee_DefaultAttack_Return_Miss::GetStaticHash())
, m_uCurrentFirstPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
, m_uCurrentThirdPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
{
	m_bClearQueuedMeleeRequests = true;
	m_bForceStanding = true;
}

bool AVP_Move_Melee_DefaultAttack_Start::Start(const Asura_Msg_Move* pxStartMsg)
{
	// The default attack gets a standard set of parameters from the current melee blueprint, and since that can change we may have to regenerate our animation list here.
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	Asura_Hash_ID uFirstPersonHash, uThirdPersonHash;
	xOwner.GetCurrentMeleeBlueprintIDs(uFirstPersonHash, uThirdPersonHash);

	if ((uFirstPersonHash != m_uCurrentFirstPersonMeleeBlueprint) || (uThirdPersonHash != m_uCurrentThirdPersonMeleeBlueprint))
	{
		// Yep, we do need to regenerate, get the appropriate blueprints
		const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uFirstPersonHash);
		const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uThirdPersonHash);
		if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
		{
			SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, "DefaultAttackStart");
		}

		m_uCurrentFirstPersonMeleeBlueprint = uFirstPersonHash;
		m_uCurrentThirdPersonMeleeBlueprint = uThirdPersonHash;

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, 0.0f, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
			m_pxAVPAnimationData->SetDontCalculateVelocity(true);
			m_pxAVPAnimationData->SetDontCalculateAngularVelocity(true);
		}
	}

	// Having regenerated the list, we can run through the generic start function.
	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_DefaultAttack_Return_Hit
// Author       : Anthony Gregson
// Date         : 24/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_DefaultAttack_Return_Hit::AVP_Move_Melee_DefaultAttack_Return_Hit(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_uCurrentFirstPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
, m_uCurrentThirdPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
{
	m_bClearQueuedMeleeRequests = true;
	m_bForceStanding = true;
}

bool AVP_Move_Melee_DefaultAttack_Return_Hit::Start(const Asura_Msg_Move* pxStartMsg)
{
	// The default attack gets a standard set of parameters from the current melee blueprint, and since that can change we may have to regenerate our animation list here.
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	Asura_Hash_ID uFirstPersonBlueprintHash, uThirdPersonBlueprintHash;
	xOwner.GetCurrentMeleeBlueprintIDs(uFirstPersonBlueprintHash, uThirdPersonBlueprintHash);

	if ((uFirstPersonBlueprintHash != m_uCurrentFirstPersonMeleeBlueprint) || (uThirdPersonBlueprintHash != m_uCurrentThirdPersonMeleeBlueprint))
	{
		// Yep, we do need to regenerate, get the appropriate blueprints
		const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uFirstPersonBlueprintHash);
		const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uThirdPersonBlueprintHash);
		if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
		{
			SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, "DefaultAttackReturnHit");
		}

		m_uCurrentFirstPersonMeleeBlueprint = uThirdPersonBlueprintHash;
		m_uCurrentThirdPersonMeleeBlueprint = uThirdPersonBlueprintHash;

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
			m_pxAVPAnimationData->SetDontCalculateVelocity(true);
			m_pxAVPAnimationData->SetDontCalculateAngularVelocity(true);
		}
	}

	// Having regenerated the list, we can run through the generic start function.
	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_DefaultAttack_Return_Countered
// Author       : Anthony Gregson
// Date         : 24/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_DefaultAttack_Return_Countered::AVP_Move_Melee_DefaultAttack_Return_Countered(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_uCurrentFirstPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
, m_uCurrentThirdPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
{
	m_bForceStanding = true;
}

bool AVP_Move_Melee_DefaultAttack_Return_Countered::Start(const Asura_Msg_Move* pxStartMsg)
{
	// The default attack gets a standard set of parameters from the current melee blueprint, and since that can change we may have to regenerate our animation list here.
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	Asura_Hash_ID uFirstPersonBlueprintHash, uThirdPersonBlueprintHash;
	xOwner.GetCurrentMeleeBlueprintIDs(uFirstPersonBlueprintHash, uThirdPersonBlueprintHash);

	if ((uFirstPersonBlueprintHash != m_uCurrentFirstPersonMeleeBlueprint) || (uThirdPersonBlueprintHash != m_uCurrentThirdPersonMeleeBlueprint))
	{
		// Yep, we do need to regenerate, get the appropriate blueprints
		const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uFirstPersonBlueprintHash);
		const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uThirdPersonBlueprintHash);
		if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
		{
			SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, "DefaultAttackReturnCountered");
		}

		m_uCurrentFirstPersonMeleeBlueprint = uFirstPersonBlueprintHash;
		m_uCurrentThirdPersonMeleeBlueprint = uThirdPersonBlueprintHash;

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
		}
	}

	// Having regenerated the list, we can run through the generic start function.
	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_DefaultAttack_Return_Miss
// Author       : Anthony Gregson
// Date         : 24/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_DefaultAttack_Return_Miss::AVP_Move_Melee_DefaultAttack_Return_Miss(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_uCurrentFirstPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
, m_uCurrentThirdPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
{
	m_bForceStanding = true;
}

bool AVP_Move_Melee_DefaultAttack_Return_Miss::Start(const Asura_Msg_Move* pxStartMsg)
{
	// The default attack gets a standard set of parameters from the current melee blueprint, and since that can change we may have to regenerate our animation list here.
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	Asura_Hash_ID uFirstPersonBlueprintHash, uThirdPersonBlueprintHash;
	xOwner.GetCurrentMeleeBlueprintIDs(uFirstPersonBlueprintHash, uThirdPersonBlueprintHash);

	if ((uFirstPersonBlueprintHash != m_uCurrentFirstPersonMeleeBlueprint) || (uThirdPersonBlueprintHash != m_uCurrentThirdPersonMeleeBlueprint))
	{
		// Yep, we do need to regenerate, get the appropriate blueprints
		const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uFirstPersonBlueprintHash);
		const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uThirdPersonBlueprintHash);
		if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
		{
			SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, "DefaultAttackReturnMiss");
		}

		m_uCurrentFirstPersonMeleeBlueprint = uFirstPersonBlueprintHash;
		m_uCurrentThirdPersonMeleeBlueprint = uThirdPersonBlueprintHash;

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
			m_pxAVPAnimationData->SetDontCalculateVelocity(true);
			m_pxAVPAnimationData->SetDontCalculateAngularVelocity(true);
		}
	}

	// Having regenerated the list, we can run through the generic start function.
	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_DefaultCounter
// Author       : Anthony Gregson
// Date         : 24/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_DefaultCounter::AVP_Move_Melee_DefaultCounter(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_uCurrentFirstPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
, m_uCurrentThirdPersonMeleeBlueprint(ASURA_HASH_ID_UNSET)
{
	m_bClearQueuedMeleeRequests = true;
}

bool AVP_Move_Melee_DefaultCounter::Start(const Asura_Msg_Move* pxStartMsg)
{
	// The default counter gets a standard set of parameters from the current melee blueprint, and since that can change we may have to regenerate our animation list here.
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	Asura_Hash_ID uFirstPersonBlueprintHash, uThirdPersonBlueprintHash;
	xOwner.GetCurrentMeleeBlueprintIDs(uFirstPersonBlueprintHash, uThirdPersonBlueprintHash);

	if ((uFirstPersonBlueprintHash != m_uCurrentFirstPersonMeleeBlueprint) || (uThirdPersonBlueprintHash != m_uCurrentThirdPersonMeleeBlueprint))
	{
		// Yep, we do need to regenerate, get the appropriate blueprints
		const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uFirstPersonBlueprintHash);
		const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uThirdPersonBlueprintHash);
		if ( pxFirstPersonBlueprint || pxThirdPersonBlueprint )
		{
			SetupAnimHashesFromBlueprint(pxFirstPersonBlueprint, pxThirdPersonBlueprint, "DefaultCounter");
		}

		m_uCurrentFirstPersonMeleeBlueprint = uFirstPersonBlueprintHash;
		m_uCurrentThirdPersonMeleeBlueprint = uThirdPersonBlueprintHash;

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
		}
	}

	// Having regenerated the list, we can run through the generic start function.
	return PARENT::Start(pxStartMsg);
}


DEBUGVAR float dbg_fBlockTweenInTime = 0.1f;
DEBUGVAR float dbg_fBlockTweenOutTime = 0.2f;

////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_BlockStart
// Author       : Anthony Gregson
// Date         : 26/03/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_BlockStart::AVP_Move_BlockStart(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
{
	AddNextMove(AVP_Move_BlockLoop::GetStaticHash());
	AddNextMove(AVP_Move_BlockEnd::GetStaticHash());

	SetAlwaysCheckNextMoves(true);

	m_bAllowsBlock = true;
	m_bRestrictedMovement = true;
	m_bAllowMeleeTargetUpdate = true;
}


bool AVP_Move_BlockStart::Start(const Asura_Msg_Move* pxStartMsg)
{
	// Get the block anims to play from our entity.
	const AVP_ClientActor_Vertebrate& xOwner = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());
	Asura_Hash_ID uFirstPersonHash, uThirdPersonHash;
	xOwner.GetBlockStartAnims(uFirstPersonHash, uThirdPersonHash);
	if ((uFirstPersonHash == ASURA_HASH_ID_UNSET) && (uThirdPersonHash == ASURA_HASH_ID_UNSET)) { return false; }

	SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(dbg_fBlockTweenInTime, 0.0f, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_BlockLoop
// Author       : Anthony Gregson
// Date         : 26/03/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_BlockLoop::AVP_Move_BlockLoop(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_bFinished(false)
{
	AddNextMove(AVP_Move_BlockEnd::GetStaticHash());

	// TODO - we still need to change the name of this flag to something more sensible (or remove it and just use the animation flag).
	m_bLoopAtEnd = true;

	m_bAllowsBlock = true;
	m_bRestrictedMovement = true;
	m_bAllowMeleeTargetUpdate = true;
}

bool AVP_Move_BlockLoop::HasFinished() const
{
	return m_bFinished;
}

bool AVP_Move_BlockLoop::ShouldStart() const
{
	// Start if the previous move is finished, and we're still blocking.
	const Asura_Move* pxPrevMove = GetParentSystem().GetCurrentMove();
	if (pxPrevMove && !pxPrevMove->HasFinished()) { return false; }

	const AVP_ClientActor_Vertebrate& xVert = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());
	if (!xVert.IsBlockActive()) { return false; }

	return PARENT::ShouldStart();
}

bool AVP_Move_BlockLoop::Start(const Asura_Msg_Move* pxStartMsg)
{
	// Get the block anims to play from our entity.
	const AVP_ClientActor_Vertebrate& xOwner = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());
	Asura_Hash_ID uFirstPersonHash, uThirdPersonHash;
	xOwner.GetBlockLoopAnims(uFirstPersonHash, uThirdPersonHash);
	if ((uFirstPersonHash == ASURA_HASH_ID_UNSET) && (uThirdPersonHash == ASURA_HASH_ID_UNSET)) { return false; }

	SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

	if (m_pxAVPAnimationData)
	{
		// As the name suggests, this anim loops
		m_pxAVPAnimationData->SetAnimFlags(AACA_FLAG_LOOPED);
		m_pxAVPAnimationData->SetTweenData(dbg_fBlockTweenInTime, 0.0f, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	m_bFinished = false;

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_BlockLoop::Finish()
{
	// If we've finished but haven't picked a new move, we want to make sure we tween out.
	if (m_pxAVPAnimationData && !GetNextMove())
	{
		m_pxAVPAnimationData->SetTweenData(dbg_fBlockTweenInTime, dbg_fBlockTweenOutTime, true);
		m_pxAVPAnimationData->UpdateTweenData();
	}

	PARENT::Finish();
}

void AVP_Move_BlockLoop::UpdateStates(const float fTimeStep)
{
	const AVP_ClientActor_Vertebrate& xVert = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());
	m_bFinished |= !xVert.IsBlockActive();

	PARENT::UpdateStates(fTimeStep);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_BlockEnd
// Author       : Anthony Gregson
// Date         : 26/03/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_BlockEnd::AVP_Move_BlockEnd(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
{
	m_bAllowsBlock = true;
	m_bAllowMeleeTargetUpdate = true;
}

bool AVP_Move_BlockEnd::ShouldStart() const
{
	const AVP_ClientActor_Vertebrate& xVert = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());
	if (xVert.IsBlockActive()) { return false; }

	return PARENT::ShouldStart();
}

bool AVP_Move_BlockEnd::Start(const Asura_Msg_Move* pxStartMsg)
{
	// Get the block anims to play from our entity.
	const AVP_ClientActor_Vertebrate& xOwner = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());
	Asura_Hash_ID uFirstPersonHash, uThirdPersonHash;
	xOwner.GetBlockEndAnims(uFirstPersonHash, uThirdPersonHash);
	if ((uFirstPersonHash == ASURA_HASH_ID_UNSET) && (uThirdPersonHash == ASURA_HASH_ID_UNSET)) { return false; }

	SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(dbg_fBlockTweenInTime, dbg_fBlockTweenOutTime, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_BlockFlinch
// Author       : Anthony Gregson
// Date         : 26/03/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_BlockFlinch::AVP_Move_BlockFlinch(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
{
	AddNextMove(AVP_Move_BlockLoop::GetStaticHash());
	AddNextMove(AVP_Move_BlockEnd::GetStaticHash());

	SetAlwaysCheckNextMoves(true);

	m_bAllowsBlock = true;
	m_bClearQueuedMeleeRequests = true;
}

bool AVP_Move_BlockFlinch::Start(const Asura_Msg_Move* pxStartMsg)
{
	// This move can only be started by message.
	if (!pxStartMsg) { return false; }

	// We need our own special message.
	const AVP_Msg_Flinch_Move_Start* pxFlinchMsg = static_cast<const AVP_Msg_Flinch_Move_Start*>(pxStartMsg);

	SetAnimHashes(pxFlinchMsg->GetFirstPersonAnim(), pxFlinchMsg->GetThirdPersonAnim());

	// Make sure the anim data was successfully set up.
	if (!m_pxAVPAnimationData) { return false; }

	m_pxAVPAnimationData->SetTweenData(dbg_fBlockTweenInTime, 0.0f, true);
	m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);

	return PARENT::Start(pxStartMsg);
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_GenericPuppeteer_Start
// Author       : Anthony Gregson
// Date         : 01/04/08
////////////////////////////////////////////////////////////////////////////////
DEBUGVAR float dbg_fGenericPuppeteerPositionInterpolationTime = 0.2f;
DEBUGVAR float dbg_fGenericPuppeteerTweenInDuration = 0.1f;
DEBUGVAR float dbg_fGenericPuppeteerTweenOutDuration = 0.2f;

AVP_Move_Melee_GenericPuppeteer_Start::AVP_Move_Melee_GenericPuppeteer_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash,
																			 const AVP_Settings::Species ePuppeteerSpecies, const AVP_Settings::Species eVictimSpecies, const Asura_Hash_ID uBlueprintNameHash,
																			 const Asura_Hash_ID uSuccessMove, const Asura_Hash_ID uCounteredMove)
: PARENT(xParentSystem, uMoveHash)
, m_uFirstPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_uThirdPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_eVictimSpecies(eVictimSpecies)
, m_uPuppeteerPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppeteerPFXMarker(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXMarker(ASURA_HASH_ID_UNSET)
, m_fInitialVictimForwardOffset(0.0f)
, m_uFinalAttackType(AVP_MELEEATTACK_TYPE_NONE)
, m_bCountered(false)
, m_bAlwaysFatal(false)
, m_bNoBranching(false)
{
	Asura_Assert(uSuccessMove != ASURA_HASH_ID_UNSET || uCounteredMove != ASURA_HASH_ID_UNSET, "This constructor should only be used with at least one follow-up move");
	if (uSuccessMove != ASURA_HASH_ID_UNSET)
	{
		AddNextMove(uSuccessMove);
	}
	if (uCounteredMove != ASURA_HASH_ID_UNSET)
	{
		AddNextMove(uCounteredMove);
	}

	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(GetBlueprintType(ePuppeteerSpecies), uBlueprintNameHash);
	if (pxBlueprint)
	{
		Asura_Hash_ID uFirstPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Hash_ID uThirdPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Value xValue;
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerFirstPersonParam(ePuppeteerSpecies, true), xValue))
		{
			uFirstPersonHash = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerThirdPersonParam(ePuppeteerSpecies, true), xValue))
		{
			uThirdPersonHash = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_INITIALOFFSET, xValue))
		{
			m_fInitialVictimForwardOffset = xValue.GetFloatValue();
		}

		SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(dbg_fGenericPuppeteerTweenInDuration, 0.0f, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
		}

		if (pxBlueprint->GetParameterValueByHashID(GetVictimFirstPersonParam(eVictimSpecies, true), xValue))
		{
			m_uFirstPersonVictimAnimID = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetVictimThirdPersonParam(eVictimSpecies, true), xValue))
		{
			m_uThirdPersonVictimAnimID = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_FSFX, xValue))
		{
			m_uPuppeteerFSFX = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_FSFX, xValue))
		{
			m_uPuppetFSFX = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_ACTIVEGRAPPLE, xValue))
		{
			m_uActiveGrappleAttackerBlueprint = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_ACTIVEGRAPPLE, xValue))
		{
			m_uActiveGrappleVictimBlueprint = xValue.GetHashIDValue();
		}

		// With follow-up moves, we don't read in the PFX effects here.
	}

	// We use external control, but use movement queries to keep ourselves from falling out of the world.
	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

AVP_Move_Melee_GenericPuppeteer_Start::AVP_Move_Melee_GenericPuppeteer_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash,
																			 const AVP_Settings::Species ePuppeteerSpecies, const AVP_Settings::Species eVictimSpecies, const Asura_Hash_ID uBlueprintNameHash)
: PARENT(xParentSystem, uMoveHash)
, m_uFirstPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_uThirdPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_eVictimSpecies(eVictimSpecies)
, m_uPuppeteerPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppeteerPFXMarker(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXMarker(ASURA_HASH_ID_UNSET)
, m_fInitialVictimForwardOffset(0.0f)
, m_uFinalAttackType(AVP_MELEEATTACK_TYPE_NONE)
, m_bCountered(false)
, m_bAlwaysFatal(false)
, m_bNoBranching(true)
{
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(GetBlueprintType(ePuppeteerSpecies), uBlueprintNameHash);
	if (pxBlueprint)
	{
		Asura_Hash_ID uFirstPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Hash_ID uThirdPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Value xValue;
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerFirstPersonParam(ePuppeteerSpecies, false), xValue))
		{
			uFirstPersonHash = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerThirdPersonParam(ePuppeteerSpecies, false), xValue))
		{
			uThirdPersonHash = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_INITIALOFFSET, xValue))
		{
			m_fInitialVictimForwardOffset = xValue.GetFloatValue();
		}

		SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

		if (m_pxAVPAnimationData)
		{
			// No next move, so set up a tween out.
			m_pxAVPAnimationData->SetTweenData(dbg_fGenericPuppeteerTweenInDuration, dbg_fGenericPuppeteerTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
		}

		if (pxBlueprint->GetParameterValueByHashID(GetVictimFirstPersonParam(eVictimSpecies, false), xValue))
		{
			m_uFirstPersonVictimAnimID = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetVictimThirdPersonParam(eVictimSpecies, false), xValue))
		{
			m_uThirdPersonVictimAnimID = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_FSFX, xValue))
		{
			m_uPuppeteerFSFX = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_FSFX, xValue))
		{
			m_uPuppetFSFX = xValue.GetHashIDValue();
		}

		// As this has no follow-up moves, this is the one that needs the PFX.
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_PFX_EFFECT, xValue))
		{
			m_uPuppeteerPFXEffect = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_PFX_EFFECT, xValue))
		{
			m_uPuppetPFXEffect = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_PFX_MARKER, xValue))
		{
			m_uPuppeteerPFXMarker = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_PFX_MARKER, xValue))
		{
			m_uPuppetPFXMarker = xValue.GetHashIDValue();
		}
	}

	// We use external control, but use movement queries to keep ourselves from falling out of the world.
	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}


AVP_Move_Melee_GenericPuppeteer_Start::AVP_Move_Melee_GenericPuppeteer_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash)
: PARENT(xParentSystem, uMoveHash)
, m_uFirstPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_uThirdPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_eVictimSpecies(AVP_Settings::AVP_NUM_SPECIES) // This means ShouldStart will always fail unless the subclass sets this variable or provides its own ShouldStart override.
, m_uPuppeteerPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppeteerPFXMarker(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXMarker(ASURA_HASH_ID_UNSET)
, m_fInitialVictimForwardOffset(0.0f)
, m_uFinalAttackType(AVP_MELEEATTACK_TYPE_NONE)
, m_bCountered(false)
, m_bAlwaysFatal(false)
, m_bNoBranching(false)
{
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	// We use external control, but use movement queries to keep ourselves from falling out of the world.
	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

bool AVP_Move_Melee_GenericPuppeteer_Start::HandleMessage(const Asura_Message* pxMessage)
{
	if (pxMessage->GetMsgID() == AVP_MSGID_COUNTER_MELEE)
	{
		m_bCountered = true;
		m_uFinalAttackType = AVP_MELEEATTACK_TYPE_NONE;

		return true;
	}
	else if (pxMessage->GetMsgID() == AVP_MSGID_ACTIVEGRAPPLE_RESULT)
	{
		// This can be corrected (unless it's final).
		const AVP_Message_Vertebrate_ActiveGrappleResult* pxData = static_cast<const AVP_Message_Vertebrate_ActiveGrappleResult*>(pxMessage->GetData());
		if (pxData->GetAttackerVictorious())
		{
			m_bCountered = false;
			m_uFinalAttackType = AVP_MELEEATTACK_TYPE_NORMAL;
		}
		else
		{
			m_bCountered = true;
			m_uFinalAttackType = AVP_MELEEATTACK_TYPE_NONE;
		}

		return true;
	}
	else
	{
		return PARENT::HandleMessage(pxMessage);
	}
}

bool AVP_Move_Melee_GenericPuppeteer_Start::ShouldStart() const
{
	// If the parent says we can start, check that the entity is of a type that we don't attach on to.
	if (!PARENT::ShouldStart()) { return false; }

	// There's only one move we can have come from.
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (!pxCurrentMove) { return false; }

	if (pxCurrentMove->GetHash() == AVP_Move_Melee_Pounce_Attack_Start::GetStaticHash())
	{
		// Extra check for whether this is a pounce grapple.
		const AVP_Move_Melee_Pounce_Attack_Start* pxPounceMove = static_cast<const AVP_Move_Melee_Pounce_Attack_Start*>(GetParentSystem().GetCurrentMove());
		if (!pxPounceMove->GetIsGrapplePounce()) { return false; }
	}

	const Asura_Guid uHitGuid = pxCurrentMove->GetHitEntity();
	const AVP_ClientEntity_Actor* pxVictim = AVP_ClientEntity_Actor::FindEntity(uHitGuid);

	if (!pxVictim) { return false; }
	if (pxVictim->IsDead()) { return false; } // TODO - Actually, we want to allow this to continue, so we can get full deaths even if the initial move killed the entity.
	// If the enemy is already someone's puppet, don't steal them.
	// If they're ours or no-one's though, allow it.
	const Asura_Movement_System* pxVictimSystem = pxVictim->GetMovementSystem();
	if (pxVictimSystem)
	{
		const Asura_Move* pxVictimMove = pxVictimSystem->GetCurrentMove();
		if (pxVictimMove && (pxVictimMove->GetHash() == AVP_Move_Puppet::GetStaticHash()))
		{
			const AVP_Move_Puppet* pxVictimPuppetMove = static_cast<const AVP_Move_Puppet*>(pxVictimMove);
			const Asura_Guid uVictimPuppeteer = pxVictimPuppetMove->GetCurrentPuppeteerGuid();
			if ((uVictimPuppeteer != GetParentSystem().GetOwnerGuid()) && (uVictimPuppeteer != ASURA_GUID_UNREGISTERED))
			{
				return false;
			}
		}
	}

	// Entities that are interacting are immune.
	if (AVP_Actor_InteractionHandler::IsEntityInteracting(uHitGuid)) { return false; }

	// If this is also the kill move, we need the additional check for whether they're invulnerable here.
	if (m_bNoBranching)
	{
		if (pxVictim->IsInvulnerable()) { return false; }
	}

	switch (pxVictim->GetClassification())
	{
	default:
		return false;

	case AVP_ENTITYCLASS_ALIEN:
	case AVP_ENTITYCLASS_PREDALIEN:
		return m_eVictimSpecies == AVP_Settings::AVP_SPECIES_ALIEN;

	case AVP_ENTITYCLASS_ANDROID:
	case AVP_ENTITYCLASS_CIVILIAN:
	case AVP_ENTITYCLASS_MARINE:
		return m_eVictimSpecies == AVP_Settings::AVP_SPECIES_HUMAN;

	case AVP_ENTITYCLASS_PREDATOR:
		return m_eVictimSpecies == AVP_Settings::AVP_SPECIES_PREDATOR;
	}
}

bool AVP_Move_Melee_GenericPuppeteer_Start::Start(const Asura_Msg_Move* pxStartMsg)
{
	// We need to set up the puppet move now.
	if (pxStartMsg)
	{
		// Started from the server start message, get the enemy we pounced on here.
		const AVP_Msg_Melee_Move* pxMeleeMsg = static_cast<const AVP_Msg_Melee_Move*>(pxStartMsg);
		m_uCurrentPuppet = pxMeleeMsg->GetTargetGuid();
	}
	else
	{
		// Started by ShouldStart, get the enemy we hit from the previous move.
		const AVP_Move_Melee* pxPreviousMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
		if (!pxPreviousMove) { return false; }

		m_uCurrentPuppet = pxPreviousMove->GetHitEntity();
	}

	AVP_ClientActor_Vertebrate* pxVictim = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentPuppet);
	if (!pxVictim) { return false; }

	Asura_Movement_System* pxMovementSystem = pxVictim->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	pxPuppetMove->SetCurrentPuppeteerGuid(GetParentSystem().GetOwnerGuid());
	pxPuppetMove->SetRelativeToPuppeteer(true);
	pxPuppetMove->SetClearQueuedMeleeRequests(true);
	pxPuppetMove->SetClearQueuedCounterRequest(true);
	pxPuppetMove->SetAlwaysCreateRagdollOnDeath(false);
	pxPuppetMove->SetControlDeathInternally(true);
	pxPuppetMove->SetPreventsFlinch(true);
	pxPuppetMove->SetDisableSurfaceReorientation(true);
	float fTweenInDuration, fTweenOutDuration;
	if (m_pxAVPAnimationData)
	{
		fTweenInDuration = m_pxAVPAnimationData->GetTweenInDuration();
		fTweenOutDuration = m_pxAVPAnimationData->GetTweenOutDuration();
	}
	else
	{
		fTweenInDuration = dbg_fGenericPuppeteerTweenInDuration;
		fTweenOutDuration = dbg_fGenericPuppeteerTweenOutDuration;
	}
	if (m_bAlwaysFatal)
	{
		fTweenOutDuration = 0.0f;
	}
	SetPuppetAnimationData(m_uFirstPersonVictimAnimID, m_uThirdPersonVictimAnimID, fTweenInDuration, fTweenOutDuration, true, true);

	// We need to set the position interpolation each time, as Reset clears everything.
	m_xPuppetPositionInterpolation.SetDuration(dbg_fGenericPuppeteerPositionInterpolationTime);
	m_xPuppetPositionInterpolation.SetFinal(Asura_AVector3(0.0f, 0.0f, m_fInitialVictimForwardOffset));
	m_xPuppetOrientationInterpolation.SetDuration(dbg_fGenericPuppeteerPositionInterpolationTime);
	m_xPuppetOrientationInterpolation.SetFinal(Asura_Quat(0.0f, 1.0f, 0.0f, 0.0f));

	m_uFinalAttackType = AVP_MELEEATTACK_TYPE_NONE;
	m_bCountered = false;

	AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();

	// Create all PFX effects we need to.
	if ((m_uPuppeteerPFXEffect != ASURA_HASH_ID_UNSET) && (m_uPuppeteerPFXMarker != ASURA_HASH_ID_UNSET))
	{
		xOwner.CreatePFXEffect(m_uPuppeteerPFXEffect, m_uPuppeteerPFXMarker, true);
	}

	if ((m_uPuppetPFXEffect != ASURA_HASH_ID_UNSET) && (m_uPuppetPFXMarker != ASURA_HASH_ID_UNSET))
	{
		pxVictim->CreatePFXEffect(m_uPuppetPFXEffect, m_uPuppetPFXMarker, true);
	}

	// Default our counter/kill state to what's appropriate.
	// The attacker always gets the advantage unless the attacker is a player and the victim is an AI.
	if (xOwner.IsPlayer() && !pxVictim->IsPlayer())
	{
		m_bCountered = true;
		m_uFinalAttackType = AVP_MELEEATTACK_TYPE_NONE;
	}
	else
	{
		m_bCountered = false;
		m_uFinalAttackType = AVP_MELEEATTACK_TYPE_NORMAL;
	}

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_GenericPuppeteer_Start::Reset()
{
	PARENT::Reset();

	SetPuppetRelativeToOwner(true);
}

void AVP_Move_Melee_GenericPuppeteer_Start::Abort()
{
	PARENT::Abort();

	const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();
	// Kill our puppet if we're aborted because our owner has died and this is a fatal move
	if (m_bAlwaysFatal && m_xParentSystem.GetOwner().IsDead())
	{
		// Check we have the master entity on this machine. That's the local player or the server AI.
		if (m_xParentSystem.IsControlledByLocalPlayer(uOwnerGuid) || (!m_xParentSystem.IsControlledByAnyPlayer(uOwnerGuid) && Asura_Network::IsServer()))
		{
			// Make sure our puppet isn't already dead.
			const AVP_ClientEntity_Actor* pxPuppet = AVP_ClientEntity_Actor::FindEntity(m_uCurrentPuppet);
			if (pxPuppet && !pxPuppet->IsDead())
			{
				// Send a damage message.
				Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
				if(pxMessage)
				{
					AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
					const AVP_EntityDamage_RequestVictim	xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_EXPLOSIVE,
																	uOwnerGuid,
																	ASURA_HASH_ID_UNSET,
																	m_uCurrentPuppet,
																	100000.0f,
																	pxPuppet->GetCurrentPosition(),
																	Asura_Vector_3(0.0f, -1.0f, 0.0f),
																	Asura_Vector_3(0.0f, 0.0f, 0.0f));
					pxData->SetDamage(xDamage);

					pxMessage->SetFromGuid(uOwnerGuid);
					pxMessage->SetToGuid(m_uCurrentPuppet);
					Asura_Network::Send(pxMessage);
				}
			}
		}
	}
}

void AVP_Move_Melee_GenericPuppeteer_Start::SendStartMsg()
{
	if (!m_bUseServerStartMessage)
	{
		PARENT::SendStartMsg();
		return;
	}

	if (AVP_Settings::GetIsServer())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_MOVESTART_SERVER, sizeof(AVP_Msg_Melee_Move));

		if(pxMessage != NULL)
		{
			AVP_Msg_Melee_Move* pxData = static_cast<AVP_Msg_Melee_Move*>(pxMessage->GetData());
			const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(false);
				pxData->SetTargetGuid(m_uCurrentPuppet);
				pxData->SetTargetEntityIsPuppet(true);
				pxData->SetVariation(0);
				pxData->SetPendingAttackType(AVP_MELEEATTACK_TYPE_NONE);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Start::GetBlueprintType(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_TYPE_GRAPPLEALIEN;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_TYPE_GRAPPLEPREDATOR;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Start::GetPuppeteerFirstPersonParam(AVP_Settings::Species ePuppeteerSpecies, bool bHasNextMove) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_ALIENSTART_FIRSTPERSON : uAVP_BLUEPRINT_PARAM_ALIEN_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_PREDATORSTART_FIRSTPERSON : uAVP_BLUEPRINT_PARAM_PREDATOR_FIRSTPERSON;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Start::GetPuppeteerThirdPersonParam(AVP_Settings::Species ePuppeteerSpecies, bool bHasNextMove) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_ALIENSTART_THIRDPERSON : uAVP_BLUEPRINT_PARAM_ALIEN_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_PREDATORSTART_THIRDPERSON : uAVP_BLUEPRINT_PARAM_PREDATOR_THIRDPERSON;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Start::GetVictimFirstPersonParam(AVP_Settings::Species eVictimSpecies, bool bHasNextMove) const
{
	switch (eVictimSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_ALIENSTART_FIRSTPERSON : uAVP_BLUEPRINT_PARAM_ALIEN_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_HUMAN:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_HUMANSTART_FIRSTPERSON : uAVP_BLUEPRINT_PARAM_HUMAN_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_PREDATORSTART_FIRSTPERSON : uAVP_BLUEPRINT_PARAM_PREDATOR_FIRSTPERSON;

	default:
		Asura_Assert(false, "Unexpected puppet species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Start::GetVictimThirdPersonParam(AVP_Settings::Species eVictimSpecies, bool bHasNextMove) const
{
	switch (eVictimSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_ALIENSTART_THIRDPERSON : uAVP_BLUEPRINT_PARAM_ALIEN_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_HUMAN:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_HUMANSTART_THIRDPERSON : uAVP_BLUEPRINT_PARAM_HUMAN_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return bHasNextMove ? uAVP_BLUEPRINT_PARAM_PREDATORSTART_THIRDPERSON : uAVP_BLUEPRINT_PARAM_PREDATOR_THIRDPERSON;

	default:
		Asura_Assert(false, "Unexpected puppet species.");
		return ASURA_HASH_ID_UNSET;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_Move_Melee_GenericPuppeteer_Success
// Author       : Anthony Gregson
// Date         : 01/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_GenericPuppeteer_Success::AVP_Move_Melee_GenericPuppeteer_Success(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash,
																				 const AVP_Settings::Species ePuppeteerSpecies, const AVP_Settings::Species eVictimSpecies, const Asura_Hash_ID uBlueprintNameHash)
: PARENT(xParentSystem, uMoveHash)
, m_uFirstPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_uThirdPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_uPuppeteerPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppeteerPFXMarker(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXEffect(ASURA_HASH_ID_UNSET)
, m_uPuppetPFXMarker(ASURA_HASH_ID_UNSET)
, m_bAlwaysFatal(false)
{
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(GetBlueprintType(ePuppeteerSpecies), uBlueprintNameHash);
	if (pxBlueprint)
	{
		Asura_Hash_ID uFirstPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Hash_ID uThirdPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Value xValue;
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerFirstPersonParam(ePuppeteerSpecies), xValue))
		{
			uFirstPersonHash = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerThirdPersonParam(ePuppeteerSpecies), xValue))
		{
			uThirdPersonHash = xValue.GetHashIDValue();
		}

		SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(dbg_fGenericPuppeteerTweenInDuration, dbg_fGenericPuppeteerTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
		}

		if (pxBlueprint->GetParameterValueByHashID(GetVictimFirstPersonParam(eVictimSpecies), xValue))
		{
			m_uFirstPersonVictimAnimID = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetVictimThirdPersonParam(eVictimSpecies), xValue))
		{
			m_uThirdPersonVictimAnimID = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_FSFX, xValue))
		{
			m_uPuppeteerFSFX = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_FSFX, xValue))
		{
			m_uPuppetFSFX = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_PFX_EFFECT, xValue))
		{
			m_uPuppeteerPFXEffect = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_PFX_EFFECT, xValue))
		{
			m_uPuppetPFXEffect = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_PFX_MARKER, xValue))
		{
			m_uPuppeteerPFXMarker = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_PFX_MARKER, xValue))
		{
			m_uPuppetPFXMarker = xValue.GetHashIDValue();
		}
	}

	// In this first version, we'll use an external control object to move.
	// Eventually, we'll need something that actually collides with walls.
	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

bool AVP_Move_Melee_GenericPuppeteer_Success::ShouldStart() const
{
	// This assumes it follows a generic puppeteer start.
	const AVP_Move_Melee_GenericPuppeteer_Start* pxCurrentMove = static_cast<const AVP_Move_Melee_GenericPuppeteer_Start*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		// Make sure the enemy isn't invulnerable, as a kill move is a bit broken in that situation.
		// This means finding the puppet entity connected to the previous move.
		const AVP_ClientEntity_Actor* pxPuppet = AVP_ClientEntity_Actor::FindEntity(pxCurrentMove->GetCurrentPuppet());
		if (!pxPuppet || pxPuppet->IsInvulnerable()) { return false; }

		// Check dead enemy immediately, as that beats the enemy having countered or missing the window.
		if (pxPuppet->IsDead()) { return true; }

		if (pxCurrentMove->GetCountered()) { return false; }

		// The default version accepts any final attack type.
		const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
		if (!xOwner.IsAIActor() &&  pxCurrentMove->GetFinalAttackType() == AVP_MELEEATTACK_TYPE_NONE) { return false; }

		return true;
	}

	return false;
}

bool AVP_Move_Melee_GenericPuppeteer_Success::Start(const Asura_Msg_Move* pxStartMsg)
{
	// We need to set up the puppet move now.
	if (pxStartMsg)
	{
		// Started from the server start message, get the enemy we connected to here.
		const AVP_Msg_Melee_Move* pxMeleeMsg = static_cast<const AVP_Msg_Melee_Move*>(pxStartMsg);
		m_uCurrentPuppet = pxMeleeMsg->GetTargetGuid();
	}
	else
	{
		// Started by ShouldStart, get the enemy from the generic puppeteer start move (which will be the previous move - if it isn't, things go horribly wrong).
		const AVP_Move_Melee_GenericPuppeteer_Start* pxPreviousMove = static_cast<const AVP_Move_Melee_GenericPuppeteer_Start*>(GetParentSystem().GetCurrentMove());
		if (!pxPreviousMove) { return false; }

		m_uCurrentPuppet = pxPreviousMove->GetCurrentPuppet();
	}

	AVP_ClientActor_Vertebrate* pxVictim = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentPuppet);
	if (!pxVictim) { return false; }

	Asura_Movement_System* pxMovementSystem = pxVictim->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	pxPuppetMove->SetCurrentPuppeteerGuid(GetParentSystem().GetOwnerGuid());
	pxPuppetMove->SetRelativeToPuppeteer(true);
	pxPuppetMove->SetClearQueuedMeleeRequests(true);
	pxPuppetMove->SetClearQueuedCounterRequest(true);
	pxPuppetMove->SetAlwaysCreateRagdollOnDeath(false);
	pxPuppetMove->SetControlDeathInternally(true);
	pxPuppetMove->SetDisableSurfaceReorientation(true);
	SetPuppetAnimationData(m_uFirstPersonVictimAnimID, m_uThirdPersonVictimAnimID, dbg_fGenericPuppeteerTweenInDuration, m_bAlwaysFatal ? 0.0f : dbg_fGenericPuppeteerTweenOutDuration, true, true);

	// Create all PFX effects we need to.
	if ((m_uPuppeteerPFXEffect != ASURA_HASH_ID_UNSET) && (m_uPuppeteerPFXMarker != ASURA_HASH_ID_UNSET))
	{
		AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
		xOwner.CreatePFXEffect(m_uPuppeteerPFXEffect, m_uPuppeteerPFXMarker, true);
	}

	if ((m_uPuppetPFXEffect != ASURA_HASH_ID_UNSET) && (m_uPuppetPFXMarker != ASURA_HASH_ID_UNSET))
	{
		pxVictim->CreatePFXEffect(m_uPuppetPFXEffect, m_uPuppetPFXMarker, true);
	}

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_GenericPuppeteer_Success::Reset()
{
	PARENT::Reset();

	SetPuppetRelativeToOwner(true);
}

void AVP_Move_Melee_GenericPuppeteer_Success::Abort()
{
	PARENT::Abort();

	const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();
	// Kill our puppet if we're aborted because our owner has died and this is a fatal move
	if (m_bAlwaysFatal && m_xParentSystem.GetOwner().IsDead())
	{
		// Only do this on the server, since that's where all damage is controlled so will know first whether either character is dead.
		if (Asura_Network::IsServer())
		{
			// Make sure our puppet isn't already dead.
			const AVP_ClientEntity_Actor* pxPuppet = AVP_ClientEntity_Actor::FindEntity(m_uCurrentPuppet);
			if (pxPuppet && !pxPuppet->IsDead())
			{
				// Send a damage message.
				Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
				if(pxMessage)
				{
					AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
					const AVP_EntityDamage_RequestVictim xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_EXPLOSIVE, // Forcing ragdoll.
																	uOwnerGuid,
																	ASURA_HASH_ID_UNSET,
																	m_uCurrentPuppet,
																	100000.0f,
																	pxPuppet->GetCurrentPosition(),
																	Asura_Vector_3(0.0f, -1.0f, 0.0f),
																	Asura_Vector_3(0.0f, 0.0f, 0.0f));
					pxData->SetDamage(xDamage);

					pxMessage->SetFromGuid(uOwnerGuid);
					pxMessage->SetToGuid(m_uCurrentPuppet);
					Asura_Network::Send(pxMessage);
				}
			}
		}
	}
}

void AVP_Move_Melee_GenericPuppeteer_Success::SendStartMsg()
{
	if (!m_bUseServerStartMessage)
	{
		PARENT::SendStartMsg();
		return;
	}

	if (AVP_Settings::GetIsServer())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_MOVESTART_SERVER, sizeof(AVP_Msg_Melee_Move));

		if(pxMessage != NULL)
		{
			AVP_Msg_Melee_Move* pxData = static_cast<AVP_Msg_Melee_Move*>(pxMessage->GetData());
			const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(false);
				pxData->SetTargetGuid(m_uCurrentPuppet);
				pxData->SetTargetEntityIsPuppet(true);
				pxData->SetVariation(0);
				pxData->SetPendingAttackType(AVP_MELEEATTACK_TYPE_NONE);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Success::GetBlueprintType(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_TYPE_GRAPPLEALIEN;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_TYPE_GRAPPLEPREDATOR;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Success::GetPuppeteerFirstPersonParam(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENKILL_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORKILL_FIRSTPERSON;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Success::GetPuppeteerThirdPersonParam(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENKILL_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORKILL_THIRDPERSON;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Success::GetVictimFirstPersonParam(AVP_Settings::Species eVictimSpecies) const
{
	switch (eVictimSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENDIE_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_HUMAN:
		return uAVP_BLUEPRINT_PARAM_HUMANDIE_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORDIE_FIRSTPERSON;

	default:
		Asura_Assert(false, "Unexpected puppet species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Success::GetVictimThirdPersonParam(AVP_Settings::Species eVictimSpecies) const
{
	switch (eVictimSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENDIE_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_HUMAN:
		return uAVP_BLUEPRINT_PARAM_HUMANDIE_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORDIE_THIRDPERSON;

	default:
		Asura_Assert(false, "Unexpected puppet species.");
		return ASURA_HASH_ID_UNSET;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Class        : AVP_Move_Melee_GenericPuppeteer_Countered
// Author       : Anthony Gregson
// Date         : 01/04/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_GenericPuppeteer_Countered::AVP_Move_Melee_GenericPuppeteer_Countered(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash,
																					 const AVP_Settings::Species ePuppeteerSpecies, const AVP_Settings::Species eVictimSpecies, const Asura_Hash_ID uBlueprintNameHash)
: PARENT(xParentSystem, uMoveHash)
, m_uFirstPersonVictimAnimID(ASURA_HASH_ID_UNSET)
, m_uThirdPersonVictimAnimID(ASURA_HASH_ID_UNSET)
{
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(GetBlueprintType(ePuppeteerSpecies), uBlueprintNameHash);
	if (pxBlueprint)
	{
		Asura_Hash_ID uFirstPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Hash_ID uThirdPersonHash = ASURA_HASH_ID_UNSET;
		Asura_Value xValue;
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerFirstPersonParam(ePuppeteerSpecies), xValue))
		{
			uFirstPersonHash = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetPuppeteerThirdPersonParam(ePuppeteerSpecies), xValue))
		{
			uThirdPersonHash = xValue.GetHashIDValue();
		}

		SetAnimHashes(uFirstPersonHash, uThirdPersonHash);

		if (m_pxAVPAnimationData)
		{
			m_pxAVPAnimationData->SetTweenData(dbg_fGenericPuppeteerTweenInDuration, dbg_fGenericPuppeteerTweenOutDuration, true);
			m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
		}

		if (pxBlueprint->GetParameterValueByHashID(GetVictimFirstPersonParam(eVictimSpecies), xValue))
		{
			m_uFirstPersonVictimAnimID = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(GetVictimThirdPersonParam(eVictimSpecies), xValue))
		{
			m_uThirdPersonVictimAnimID = xValue.GetHashIDValue();
		}

		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_ATTACKER_FSFX, xValue))
		{
			m_uPuppeteerFSFX = xValue.GetHashIDValue();
		}
		if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_VICTIM_FSFX, xValue))
		{
			m_uPuppetFSFX = xValue.GetHashIDValue();
		}
	}

	// In this first version, we'll use an external control object to move.
	// Eventually, we'll need something that actually collides with walls.
	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

bool AVP_Move_Melee_GenericPuppeteer_Countered::ShouldStart() const
{
	const AVP_Move_Melee_GenericPuppeteer_Start* pxCurrentMove = static_cast<const AVP_Move_Melee_GenericPuppeteer_Start*>(GetParentSystem().GetCurrentMove());
	if (pxCurrentMove)
	{
		if (pxCurrentMove->GetCountered()) { return true; }

		// The default success move accepts any final attack type as the final blow.
		const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
		if (!xOwner.IsAIActor() && pxCurrentMove->GetFinalAttackType() == AVP_MELEEATTACK_TYPE_NONE) { return true; }

		// Make sure this starts if the enemy is invulnerable, as a kill move is a bit broken in that situation.
		// This means finding the puppet entity connected to the previous move.
		const AVP_ClientEntity_Actor* pxPuppet = AVP_ClientEntity_Actor::FindEntity(pxCurrentMove->GetCurrentPuppet());
		if (!pxPuppet || pxPuppet->IsInvulnerable()) { return true; }

		return false;
	}

	return false;
}

bool AVP_Move_Melee_GenericPuppeteer_Countered::Start(const Asura_Msg_Move* pxStartMsg)
{
	// We need to set up the puppet move now.
	if (pxStartMsg)
	{
		// Started from the server start message, get the enemy we pounced on here.
		const AVP_Msg_Melee_Move* pxMeleeMsg = static_cast<const AVP_Msg_Melee_Move*>(pxStartMsg);
		m_uCurrentPuppet = pxMeleeMsg->GetTargetGuid();
	}
	else
	{
		// Started by ShouldStart, get the enemy from the generic puppeteer start move (which will be the previous move - if it isn't, things go horribly wrong).
		const AVP_Move_Melee_GenericPuppeteer_Start* pxPreviousMove = static_cast<const AVP_Move_Melee_GenericPuppeteer_Start*>(GetParentSystem().GetCurrentMove());
		if (!pxPreviousMove) { return false; }

		m_uCurrentPuppet = pxPreviousMove->GetCurrentPuppet();
	}

	AVP_ClientActor_Vertebrate* pxVictim = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentPuppet);
	if (!pxVictim) { return false; }

	Asura_Movement_System* pxMovementSystem = pxVictim->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	pxPuppetMove->SetCurrentPuppeteerGuid(GetParentSystem().GetOwnerGuid());
	pxPuppetMove->SetRelativeToPuppeteer(true);
	pxPuppetMove->SetClearQueuedMeleeRequests(true);
	pxPuppetMove->SetClearQueuedCounterRequest(true);
	pxPuppetMove->SetAlwaysCreateRagdollOnDeath(true);
	pxPuppetMove->SetControlDeathInternally(false);
	pxPuppetMove->SetDisableSurfaceReorientation(true);
	SetPuppetAnimationData(m_uFirstPersonVictimAnimID, m_uThirdPersonVictimAnimID, dbg_fGenericPuppeteerTweenInDuration, dbg_fGenericPuppeteerTweenOutDuration, true, true);

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_GenericPuppeteer_Countered::Reset()
{
	PARENT::Reset();

	SetPuppetRelativeToOwner(true);
}

void AVP_Move_Melee_GenericPuppeteer_Countered::SendStartMsg()
{
	if (!m_bUseServerStartMessage)
	{
		PARENT::SendStartMsg();
		return;
	}

	if (AVP_Settings::GetIsServer())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_MOVESTART_SERVER, sizeof(AVP_Msg_Melee_Move));

		if(pxMessage != NULL)
		{
			AVP_Msg_Melee_Move* pxData = static_cast<AVP_Msg_Melee_Move*>(pxMessage->GetData());
			const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(false);
				pxData->SetTargetGuid(m_uCurrentPuppet);
				pxData->SetTargetEntityIsPuppet(true);
				pxData->SetVariation(0);
				pxData->SetPendingAttackType(AVP_MELEEATTACK_TYPE_NONE);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Countered::GetBlueprintType(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_TYPE_GRAPPLEALIEN;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_TYPE_GRAPPLEPREDATOR;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Countered::GetPuppeteerFirstPersonParam(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENCOUNTERED_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORCOUNTERED_FIRSTPERSON;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Countered::GetPuppeteerThirdPersonParam(AVP_Settings::Species ePuppeteerSpecies) const
{
	switch (ePuppeteerSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENCOUNTERED_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORCOUNTERED_THIRDPERSON;

	default:
		Asura_Assert(false, "Unexpected puppeteer species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Countered::GetVictimFirstPersonParam(AVP_Settings::Species eVictimSpecies) const
{
	switch (eVictimSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENCOUNTER_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_HUMAN:
		return uAVP_BLUEPRINT_PARAM_HUMANCOUNTER_FIRSTPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORCOUNTER_FIRSTPERSON;

	default:
		Asura_Assert(false, "Unexpected puppet species.");
		return ASURA_HASH_ID_UNSET;
	}
}

Asura_Hash_ID AVP_Move_Melee_GenericPuppeteer_Countered::GetVictimThirdPersonParam(AVP_Settings::Species eVictimSpecies) const
{
	switch (eVictimSpecies)
	{
	case AVP_Settings::AVP_SPECIES_ALIEN:
		return uAVP_BLUEPRINT_PARAM_ALIENCOUNTER_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_HUMAN:
		return uAVP_BLUEPRINT_PARAM_HUMANCOUNTER_THIRDPERSON;

	case AVP_Settings::AVP_SPECIES_PREDATOR:
		return uAVP_BLUEPRINT_PARAM_PREDATORCOUNTER_THIRDPERSON;

	default:
		Asura_Assert(false, "Unexpected puppet species.");
		return ASURA_HASH_ID_UNSET;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Class	    : AVP_Move_Melee_Pounce_Attack_Start
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////
DVSVAR bool dvs_bUsePouncePositionPrediction = true;
DVSVAR float dvs_fPouncePositionPredictionDurationMultiplier = 0.3f;
DVSVAR float dvs_fPouncePositionPredictionYModifier = 0.5f;

DEBUGVAR float dbg_fDefaultPounceTrackingMaxAcceleration = 20.0f;

int AVP_Move_Melee_Pounce_Attack_Start::s_iConsoleVarsCount = 0;

AVP_Move_Melee_Pounce_Attack_Start::AVP_Move_Melee_Pounce_Attack_Start(Asura_Movement_System& xParentSystem, Asura_Hash_ID uFirstPersonBlueprintHash, Asura_Hash_ID uThirdPersonBlueprintHash)
: PARENT(xParentSystem, GetStaticHash(),
		 AVP_Move_Melee_Pounce_Attack_Return_Hit::GetStaticHash(), AVP_Move_Melee_Pounce_Attack_Return_Countered::GetStaticHash(), AVP_Move_Melee_Pounce_Attack_Return_Miss::GetStaticHash())
, m_xCurrentVelocity(Asura_Maths::ZeroAVector3)
, m_fEnemyPositionForwardOffset(0.0f)
, m_fTrackingMaxAcceleration(dbg_fDefaultPounceTrackingMaxAcceleration)
, m_bHasAlienGrapple(false)
, m_bHasHumanGrapple(false)
, m_bHasPredatorGrapple(false)
, m_bIsGrapplePounce(false)
{
	AddNextMove(AVP_Move_Melee_Grapple_Start::GetStaticHash());

	Asura_Value xValue;

	const Asura_Blueprint* pxFirstPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uFirstPersonBlueprintHash);
	const Asura_Blueprint* pxThirdPersonBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_MELEE, uThirdPersonBlueprintHash);
	if (pxFirstPersonBlueprint || pxThirdPersonBlueprint)
	{
		const Asura_Blueprint* pxBlueprint = pxFirstPersonBlueprint ? pxFirstPersonBlueprint : pxThirdPersonBlueprint;
		if (pxBlueprint->GetParameterValueByHashID(0x2E819496, xValue)) // Asura_GetHashID("Pounce_EnemyOffset")
		{
			m_fEnemyPositionForwardOffset = xValue.GetFloatValue();
		}
	}

	// We also need the actor blueprint to get the rate of tracking.
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	const Asura_Blueprint* pxActorBlueprint = xOwner.GetBlueprint();
	if (pxActorBlueprint)
	{
		if (pxActorBlueprint->GetParameterValueByHashID(0x535D0C9C, xValue)) // Asura_GetHashID("Pounce_TrackingAcceleration")
		{
			m_fTrackingMaxAcceleration = xValue.GetFloatValue();
		}
	}

	m_bForceStanding = true;
	m_bPreventsFlinch = true;
	SetGravityIgnored(true);

	// Remove the default generic start aiming behaviour
	SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_NOT_USED);
	m_xPositionInterpolationData.SetVelocityOperation(Asura_MoveData_PositionInterpolation::IGNORE_VELOCITY);

#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
	if (s_iConsoleVarsCount == 0)
	{
		if (!Asura_CommandConsole::VarExists(L"AVP.Pounce.UsePrediction"))
		{
			Asura_CommandConsole::AddVar(L"AVP.Pounce.UsePrediction", &dvs_bUsePouncePositionPrediction);
		}
		if (!Asura_CommandConsole::VarExists(L"AVP.Pounce.PredictedDistanceMultiplier"))
		{
			Asura_CommandConsole::AddVar(L"AVP.Pounce.PredictedDistanceMultiplier", &dvs_fPouncePositionPredictionDurationMultiplier, 0.0f, 1.0f);
		}
		if (!Asura_CommandConsole::VarExists(L"AVP.Pounce.PredictedHeightChangeMultiplier"))
		{
			Asura_CommandConsole::AddVar(L"AVP.Pounce.PredictedHeightChangeMultiplier", &dvs_fPouncePositionPredictionYModifier, 0.0f, 1.0f);
		}
	}
	++s_iConsoleVarsCount;
#endif //#ifdef ASURA_DEBUG_VARIABLE_SYSTEM
}

AVP_Move_Melee_Pounce_Attack_Start::~AVP_Move_Melee_Pounce_Attack_Start()
{
#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
	--s_iConsoleVarsCount;
	if (s_iConsoleVarsCount == 0)
	{
		if (Asura_CommandConsole::VarExists(L"AVP.Pounce.UsePrediction"))
		{
			Asura_CommandConsole::RemoveVar(L"AVP.Pounce.UsePrediction");
		}
		if (Asura_CommandConsole::VarExists(L"AVP.Pounce.PredictedDistanceMultiplier"))
		{
			Asura_CommandConsole::RemoveVar(L"AVP.Pounce.PredictedDistanceMultiplier");
		}
		if (Asura_CommandConsole::VarExists(L"AVP.Pounce.PredictedHeightChangeMultiplier"))
		{
			Asura_CommandConsole::RemoveVar(L"AVP.Pounce.PredictedHeightChangeMultiplier");
		}
	}
#endif //#ifdef ASURA_SUPPORT_COMMAND_CONSOLE
}

void AVP_Move_Melee_Pounce_Attack_Start::AddAlienGrapple(AVP_Move_Puppeteer* pxMove)
{
	AddNextMove(pxMove);
	m_bHasAlienGrapple = true;
}

void AVP_Move_Melee_Pounce_Attack_Start::AddHumanGrapple(AVP_Move_Puppeteer* pxMove)
{
	AddNextMove(pxMove);
	m_bHasHumanGrapple = true;
}

void AVP_Move_Melee_Pounce_Attack_Start::AddPredatorGrapple(AVP_Move_Puppeteer* pxMove)
{
	AddNextMove(pxMove);
	m_bHasPredatorGrapple = true;
}

void AVP_Move_Melee_Pounce_Attack_Start::Reset()
{
	PARENT::Reset();

	m_xPositionInterpolationData.Reset();
	m_bIsGrapplePounce = false;
}

DEBUGVAR float dbg_fPounceOrientationInterpolationTime = 0.3f;

bool AVP_Move_Melee_Pounce_Attack_Start::Start(const Asura_Msg_Move* pxStartMsg)
{
	// There's not really much we can do without the message in this case, as we need a target position.
	if(!pxStartMsg) { return false; }

	const AVP_Msg_Melee_Move* pxMeleeStart = static_cast<const AVP_Msg_Melee_Move*>(pxStartMsg);
	const AVP_ClientEntity_PhysicalObject* pxTarget = AVP_ClientEntity_PhysicalObject::FindEntity(pxMeleeStart->GetTargetGuid());
	if (!pxTarget) { return false; }

	AVP_ClientActor_Vertebrate& xVertebrateOwner = static_cast<AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());

	SetVariation(pxMeleeStart->GetVariation());

	if (m_pxAVPAnimationData)
	{
		// We need to ensure the correct anim hash is set up, as it's about to get used.
		if (xVertebrateOwner.IsUsingFirstPersonSkin())
		{
			m_pxAVPAnimationData->SetAnimHash(m_uFirstPersonAnimHash);
		}
		else
		{
			m_pxAVPAnimationData->SetAnimHash(m_uThirdPersonAnimHash);
		}

		// Since these can't be set in the constructor, make sure they get set up correctly.
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, 0.0f, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	m_bIsGrapplePounce = pxMeleeStart->GetPendingAttackType() == AVP_MELEEATTACK_TYPE_GRAPPLEPOUNCE;

	bool bHasAnimation = m_pxAVPAnimationData && m_pxAVPAnimationData->HasValidAnimation();
	float fDuration = bHasAnimation ? m_pxAVPAnimationData->GetAnimationDuration() : 1.000f;

	Asura_Matrix_3x3 xTargetOri;
	SetupTargetPosition(*pxTarget, fDuration, &xTargetOri);

	m_xPositionInterpolationData.SetInitial(xVertebrateOwner.GetCurrentPosition());
	m_xOrientationInterpolationData.SetFinal(xTargetOri);

	m_xPositionInterpolationData.SetDuration(fDuration);
	m_xOrientationInterpolationData.SetDuration(dbg_fPounceOrientationInterpolationTime);

	// Calculate the initial velocity for getting to the target location.
	m_xCurrentVelocity = m_xPositionInterpolationData.GetFinal() - m_xPositionInterpolationData.GetInitial();
	m_xCurrentVelocity /= fDuration;

	InitialiseOrientationInterpolation(m_xOrientationInterpolationData);

	SetPositionUsage(POSITION_SET_FROM_VELOCITY);

	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(xVertebrateOwner.GetHUDHandler());

	if(pxHUDHandler != NULL)
	{
		pxHUDHandler->OnBeginMeleeJump();
	}

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_Pounce_Attack_Start::UpdateStates(const float fTimeStep)
{
	UpdateTracking(fTimeStep);

	PARENT::UpdateStates(fTimeStep);
}

DEBUGVAR float dbg_fPounceTrackingSpeedSquaredThreshold = 0.01f;

bool AVP_Move_Melee_Pounce_Attack_Start::CalculateVelocity(const float fTimeStep, Asura_AVector3_OutParam xVelocity) const
{
	PARENT::CalculateVelocity(fTimeStep, xVelocity);

	// Only use our tracking velocity if the animation is actually moving.
	if (xVelocity.MagnitudeSquared() >= dbg_fPounceTrackingSpeedSquaredThreshold)
	{
		xVelocity = m_xCurrentVelocity;
	}

	return false;
}

DEBUGVAR float dbg_fMaxPounceSpeed = 40.0f;
DEBUGVAR float dbg_fMaxPounceSpeedSquared = dbg_fMaxPounceSpeed * dbg_fMaxPounceSpeed;

void AVP_Move_Melee_Pounce_Attack_Start::UpdateTracking(const float fTimeStep)
{
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();

	// If we have the appropriate anim flag, we can update the target position based on the latest info from the target.
	if ((GetAnimFlagsForLastFrame() & AVP_ANIMFLAG_TRACK_TARGET) != 0)
	{
		const AVP_ClientEntity_PhysicalObject* pxTarget = AVP_ClientEntity_PhysicalObject::FindEntity(m_uTargetGuid);
		if (pxTarget)
		{
			SetupTargetPosition(*pxTarget, m_xPositionInterpolationData.GetDuration(), NULL);
		}
	}

	// Always set up the velocity sensibly.
	m_xPositionInterpolationData.Update(fTimeStep);

	Asura_AVector3 xTargetVel = m_xPositionInterpolationData.GetFinal() - Asura_AVector3(xOwner.GetCurrentPosition());
	xTargetVel /= m_xPositionInterpolationData.GetDuration() - m_xPositionInterpolationData.GetTimer();
	if (xTargetVel.MagnitudeSquared() > dbg_fMaxPounceSpeedSquared)
	{
		xTargetVel.Normalise();
		xTargetVel *= dbg_fMaxPounceSpeed;
	}

	Asura_AVector3 xDiff = xTargetVel - m_xCurrentVelocity;
	const float fMaxSpeedChange = m_fTrackingMaxAcceleration * fTimeStep;
	if (xDiff.MagnitudeSquared() <= (fMaxSpeedChange * fMaxSpeedChange))
	{
		m_xCurrentVelocity = xTargetVel;
	}
	else
	{
		xDiff.Normalise();
		m_xCurrentVelocity += xDiff * fMaxSpeedChange;
	}
}

void AVP_Move_Melee_Pounce_Attack_Start::SetupTargetPosition(const AVP_ClientEntity_PhysicalObject& xTargetEnemy, const float fDuration, Asura_Matrix_3x3* pxOrientationToEnemy)
{
	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();

	Asura_Vector_3 xTargetPosition;
	xTargetEnemy.GetMeleeOrigin(xTargetPosition);
	xTargetPosition.y = xTargetEnemy.GetCurrentPosition().y;
	Asura_Vector_3 xFlatDirection;

	if (dvs_bUsePouncePositionPrediction)
	{
		// We need to estimate the final position of the enemy, so add on their velocity to the target position.
		Asura_Vector_3 xActorVel;
		xTargetEnemy.GetCurrentVelocity(xActorVel);
		xActorVel.y *= dvs_fPouncePositionPredictionYModifier;
		xTargetPosition += xActorVel * fDuration * dvs_fPouncePositionPredictionDurationMultiplier;

		Asura_Vector_3 xBaseDirection = xTargetPosition - xOwner.GetCurrentPosition();
		xFlatDirection = xBaseDirection;
		xFlatDirection.y = 0.0f;
		xFlatDirection.Normalise();
		// We need to move back by the forward offset
		xTargetPosition -= xFlatDirection * m_fEnemyPositionForwardOffset;

		// Test if this position is reachable, and if not fall back to where the enemy currently is.
		const float fMag = xBaseDirection.Normalise();
		Asura_Vector_3 xMeleeOrigin;
		xOwner.GetMeleeOrigin(xMeleeOrigin);
		const Asura_LOS_InputData xInput(xMeleeOrigin, xBaseDirection, xOwner.GetGuid(), fMag);
		const Asura_LOS_OutputData* pxOutput = Asura_LOS::TestAgainstEverything(&xInput);
		if ((pxOutput->m_fDistanceToIntersection < fMag) && (pxOutput->m_uHitEntity != xTargetEnemy.GetGuid()))
		{
			// Failed to find a clear path to the predicted location, go to the fallback.
			xTargetEnemy.GetMeleeOrigin(xTargetPosition);
			xTargetPosition.y = xTargetEnemy.GetCurrentPosition().y;
			xFlatDirection = xTargetPosition - xOwner.GetCurrentPosition();
			xFlatDirection.y = 0.0f;
			xFlatDirection.Normalise();
			// We need to move back by the forward offset
			xTargetPosition -= xFlatDirection * m_fEnemyPositionForwardOffset;
		}
	}
	else
	{
		// Just offset this position by the blueprinted offset.
		xFlatDirection = xTargetPosition - xOwner.GetCurrentPosition();;
		xFlatDirection.y = 0.0f;
		xFlatDirection.Normalise();
		// We need to move back by the forward offset
		xTargetPosition -= xFlatDirection * m_fEnemyPositionForwardOffset;
	}

	m_xPositionInterpolationData.SetFinal(xTargetPosition);

	if (pxOrientationToEnemy)
	{
		Asura_Maths::OrientationFromDirection(*pxOrientationToEnemy, xFlatDirection);
	}
}

////////////////////////////////////////////////////////////////////////////////
// Class	    : AVP_Move_Melee_Pounce_Attack_Return_Hit
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_Pounce_Attack_Return_Hit::AVP_Move_Melee_Pounce_Attack_Return_Hit(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
{
	m_bForceStanding = true;
	m_bClearQueuedMeleeRequests = true;
	SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_NOT_USED);
	SetPositionUsage(Asura_Move::POSITION_SET_FROM_VELOCITY);
}

bool AVP_Move_Melee_Pounce_Attack_Return_Hit::ShouldStart() const
{
	// Check the parent first.
	if (!PARENT::ShouldStart()) { return false; }

	// Grapple pounces need some extra rules.
	const AVP_Move_Melee_Pounce_Attack_Start* pxCurrentMove = static_cast<const AVP_Move_Melee_Pounce_Attack_Start*>(GetParentSystem().GetCurrentMove());
	if (!pxCurrentMove)
	{
		return false;
	}

	if (pxCurrentMove->GetIsGrapplePounce())
	{
		const AVP_ClientEntity_Actor* pxVictim = AVP_ClientEntity_Actor::FindEntity(pxCurrentMove->GetHitEntity());
		// TODO - These next two checks take the safe option of using this move if our victim has gone missing or died, but we should possibly be going into miss in that case.
		if (!pxVictim) { return true; }
		if (pxVictim->IsDead()) { return true; }

		switch (pxVictim->GetClassification())
		{
		default:
			return true;

		case AVP_ENTITYCLASS_PREDALIEN:
		case AVP_ENTITYCLASS_ALIEN:
			return !pxCurrentMove->GetHasAlienGrapple();

		case AVP_ENTITYCLASS_MARINE:
		case AVP_ENTITYCLASS_CIVILIAN:
		case AVP_ENTITYCLASS_ANDROID:
			return !pxCurrentMove->GetHasHumanGrapple();

		case AVP_ENTITYCLASS_PREDATOR:
			return !pxCurrentMove->GetHasPredatorGrapple();
		}
	}

	return true;
}

bool AVP_Move_Melee_Pounce_Attack_Return_Hit::Start(const Asura_Msg_Move* pxStartMsg)
{
	bool bSuccess = PARENT::Start(pxStartMsg);

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	return bSuccess;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee_Pounce_Attack_Return_Hit::Abort
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee_Pounce_Attack_Return_Hit::Abort(void)
{
	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(GetAVPParentSystem().GetActorOwner().GetHUDHandler());

	if (pxHUDHandler)
	{
		pxHUDHandler->OnEndMeleeJump();
	}

	PARENT::Abort();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee_Pounce_Attack_Return_Hit::Finish
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee_Pounce_Attack_Return_Hit::Finish(void)
{
	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(GetAVPParentSystem().GetActorOwner().GetHUDHandler());

	if (pxHUDHandler)
	{
		pxHUDHandler->OnEndMeleeJump();
	}

	PARENT::Finish();
}

////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_Pounce_Attack_Return_Countered
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_Pounce_Attack_Return_Countered::AVP_Move_Melee_Pounce_Attack_Return_Countered(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
{
	m_bForceStanding = true;
	SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_RESET);
	SetPositionUsage(Asura_Move::POSITION_SET_FROM_VELOCITY);
}

bool AVP_Move_Melee_Pounce_Attack_Return_Countered::Start(const Asura_Msg_Move* pxStartMsg)
{
	bool bSuccess = PARENT::Start(pxStartMsg);

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	return bSuccess;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee_Pounce_Attack_Return_Countered::Abort
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee_Pounce_Attack_Return_Countered::Abort(void)
{
	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(GetAVPParentSystem().GetActorOwner().GetHUDHandler());

	if (pxHUDHandler)
	{
		pxHUDHandler->OnEndMeleeJump();
	}

	PARENT::Abort();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee_Pounce_Attack_Return_Countered::Finish
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee_Pounce_Attack_Return_Countered::Finish(void)
{
	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(GetAVPParentSystem().GetActorOwner().GetHUDHandler());

	if (pxHUDHandler)
	{
		pxHUDHandler->OnEndMeleeJump();
	}

	PARENT::Finish();
}

////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_Pounce_Attack_Return_Miss
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_Pounce_Attack_Return_Miss::AVP_Move_Melee_Pounce_Attack_Return_Miss(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
{
	m_bForceStanding = true;
	SetAimDirectionUsage(AVP_ClientEntity_Actor::AIM_DIRECTION_RESET);
	SetPositionUsage(Asura_Move::POSITION_SET_FROM_VELOCITY);
}

bool AVP_Move_Melee_Pounce_Attack_Return_Miss::Start(const Asura_Msg_Move* pxStartMsg)
{
	bool bSuccess = PARENT::Start(pxStartMsg);

	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(g_fTweenInDuration, g_fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	return bSuccess;
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee_Pounce_Attack_Return_Miss::Abort
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee_Pounce_Attack_Return_Miss::Abort(void)
{
	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(GetAVPParentSystem().GetActorOwner().GetHUDHandler());

	if (pxHUDHandler)
	{
		pxHUDHandler->OnEndMeleeJump();
	}

	PARENT::Abort();
}

////////////////////////////////////////////////////////////////////////////////
// Function     : AVP_Move_Melee_Pounce_Attack_Return_Miss::Finish
// Author       : Anthony Gregson
// Date         : 11/08/08
////////////////////////////////////////////////////////////////////////////////

void AVP_Move_Melee_Pounce_Attack_Return_Miss::Finish(void)
{
	AVP_Vertebrate_HUDHandler* pxHUDHandler = static_cast<AVP_Vertebrate_HUDHandler*>(GetAVPParentSystem().GetActorOwner().GetHUDHandler());

	if (pxHUDHandler)
	{
		pxHUDHandler->OnEndMeleeJump();
	}

	PARENT::Finish();
}


////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_Grapple_Start
// Author       : Anthony Gregson
// Date         : 02/04/09
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_Grapple_Start::AVP_Move_Melee_Grapple_Start(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash, const Asura_Hash_ID uFinishMove)
: PARENT(xParentSystem, uMoveHash)
, m_uGrappleDescriptionBlueprint(ASURA_HASH_ID_UNSET)
, m_xFinalPuppetPositionOffset(Asura_Maths::ZeroVector3)
, m_xFinalPuppetOrientationOffset(Asura_Maths::IdentityQuat)
, m_uGrappleFlags(0)
{
	if (uFinishMove != ASURA_HASH_ID_UNSET)
	{
		AddNextMove(uFinishMove);
	}

	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

AVP_Move_Melee_Grapple_Start::AVP_Move_Melee_Grapple_Start(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_uGrappleDescriptionBlueprint(ASURA_HASH_ID_UNSET)
, m_xFinalPuppetPositionOffset(Asura_Maths::ZeroVector3)
, m_xFinalPuppetOrientationOffset(Asura_Maths::IdentityQuat)
, m_uGrappleFlags(0)
{
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);

	AddNextMove(AVP_Move_Melee_Grapple_Finish::GetStaticHash());
}

bool AVP_Move_Melee_Grapple_Start::HandleMessage(const Asura_Message* pxMessage)
{
	if (pxMessage->GetMsgID() == AVP_MSGID_ACTIVEGRAPPLE_RESULT)
	{
		// This can be corrected (unless it's final).
		if (!m_bResultFinalised)
		{
			const AVP_Message_Vertebrate_ActiveGrappleResult* pxData = static_cast<const AVP_Message_Vertebrate_ActiveGrappleResult*>(pxMessage->GetData());
			m_bAttackerVictorious = pxData->GetAttackerVictorious();
			m_bCriticalVictory = pxData->GetCriticalVictory();
			m_bResultFinalised = pxData->IsFinalResult();
		}

		return true;
	}
	else
	{
		return PARENT::HandleMessage(pxMessage);
	}
}

bool AVP_Move_Melee_Grapple_Start::HasFinished() const
{
	// We cannot be finished if the frame stamps don't match, since that means we've been updated externally and haven't gone through our own update yet.
	if (m_uLastUpdateFrameStamp != m_uLastStatesUpdateFrameStamp) { return false; }
	// Also make sure the matching update is this frame.
	if (m_uLastUpdateFrameStamp != Asura_Timers::GetFrameNumber()) { return false; }
	// Finally, check the bool which confirms we have got past UpdateStates in our own call of Update.
	if (!m_bCanFinish) { return false; }

	// If we've reached a point in the animation where we can finish and the active grapple result has been finalised, we've finished.
	// The m_bGrappleCanFinish flag is updated with both of those pieces of information during UpdateStates.
	if (m_bGrappleCanFinish)
	{
		return true;
	}

	return PARENT::HasFinished();
}

bool AVP_Move_Melee_Grapple_Start::GetFinishMovePuppetStartOffset(Asura_AVector3& xPosition, Asura_Quat& xOrientation) const
{
	if (m_bHasFinishMoveOffset)
	{
		xPosition = m_xFinalPuppetPositionOffset;
		xOrientation = m_xFinalPuppetOrientationOffset;
	}
	return m_bHasFinishMoveOffset;
}

bool AVP_Move_Melee_Grapple_Start::ShouldStart() const
{
	// Check the parent details first.
	if (!PARENT::ShouldStart()) { return false; }

	// Making an assumption that we must be coming from a melee move.
	const AVP_Move_Melee* pxCurrentMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
	if (!pxCurrentMove) { return false; }

	const Asura_Hash_ID uBlueprint = pxCurrentMove->GetPendingGrappleBlueprint();
	if (uBlueprint == ASURA_HASH_ID_UNSET) { return false; } // If the previous move doesn't have a grapple to come, we can't start.

	const Asura_Guid uHitGuid = pxCurrentMove->GetHitEntity();
	const AVP_ClientEntity_Actor* pxVictim = AVP_ClientEntity_Actor::FindEntity(uHitGuid);

	if (!pxVictim) { return false; }
	if (pxVictim->IsRagdoll()) { return false; }

	// We need a puppet move in order to go on, make sure there is one.
	const Asura_Movement_System* pxVictimSystem = pxVictim->GetMovementSystem();
	if (!pxVictimSystem) { return false; }

	const AVP_Move_Puppet* pxPuppetMove = static_cast<const AVP_Move_Puppet*>(pxVictimSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	// If the enemy is already someone's puppet, don't steal them.
	// If they're ours or no-one's though, allow it.
	const Asura_Move* pxVictimMove = pxVictimSystem->GetCurrentMove();
	if (pxVictimMove && (pxVictimMove->GetHash() == AVP_Move_Puppet::GetStaticHash()))
	{
		const AVP_Move_Puppet* pxVictimPuppetMove = static_cast<const AVP_Move_Puppet*>(pxVictimMove);
		const Asura_Guid uVictimPuppeteer = pxVictimPuppetMove->GetCurrentPuppeteerGuid();
		if ((uVictimPuppeteer != GetParentSystem().GetOwnerGuid()) && (uVictimPuppeteer != ASURA_GUID_UNREGISTERED))
		{
			return false;
		}
	}

	// Entities that are interacting are immune.
	if (AVP_Actor_InteractionHandler::IsEntityInteracting(uHitGuid)) { return false; }

	// If this grapple has no separate kill animation, we need to skip immediately if the enemy is invulnerable.
	const Asura_Blueprint* pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, uBlueprint);
	if (!pxGrappleDescription) { return false; }

	const Asura_Hash_ID uAnimSetBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_ANIMSET, ASURA_HASH_ID_UNSET);
	if (uAnimSetBlueprint == ASURA_HASH_ID_UNSET) { return false; }

	const Asura_Blueprint* pxAnimSet = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimSetBlueprint);
	if (!pxAnimSet) { return false; }

	// The most common parameter to exist is the third-person attacker 
	Asura_Hash_ID uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_ATTACKER_TP, ASURA_HASH_ID_UNSET);
	if (uAnimationsBlueprint == ASURA_HASH_ID_UNSET)
	{
		// However, if that doesn't exist, the first-person set really should.
		uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_ATTACKER_HUD, ASURA_HASH_ID_UNSET);
		if (uAnimationsBlueprint == ASURA_HASH_ID_UNSET)
		{
			return false;
		}
	}

	const Asura_Blueprint* pxAnimations = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint);
	if (!pxAnimations) { return false; }

	// Check for the lack of a win animation.
	Asura_Value xValue;
	if (!pxAnimations->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN, xValue))
	{
		if (pxVictim->IsInvulnerable())
		{
			return false;
		}
	}

	return true;
}

bool AVP_Move_Melee_Grapple_Start::Start(const Asura_Msg_Move* pxStartMsg)
{
	// We need to set up the puppet move now.
	if (pxStartMsg)
	{
		// Started from the server start message, get the enemy we've grabbed here.
		const AVP_Msg_Grapple_Move* pxMeleeMsg = static_cast<const AVP_Msg_Grapple_Move*>(pxStartMsg);
		m_uCurrentPuppet = pxMeleeMsg->GetTargetGuid();
		m_uGrappleDescriptionBlueprint = pxMeleeMsg->GetGrappleDescriptionBlueprint();
	}
	else
	{
		// Started by ShouldStart, get the enemy we hit from the previous move.
		const AVP_Move_Melee* pxPreviousMove = static_cast<const AVP_Move_Melee*>(GetParentSystem().GetCurrentMove());
		if (!pxPreviousMove) { return false; }

		m_uCurrentPuppet = pxPreviousMove->GetHitEntity();
		m_uGrappleDescriptionBlueprint = pxPreviousMove->GetPendingGrappleBlueprint();
	}

	// We'll need our owner as a vertebrate. Only vertebrates support melee anyway.
	const AVP_ClientActor_Vertebrate& xOwner = static_cast<const AVP_ClientActor_Vertebrate&>(GetAVPParentSystem().GetActorOwner());

	// Get the victim's puppet move.
	AVP_ClientActor_Vertebrate* pxVictim = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentPuppet);
	if (!pxVictim) { return false; }

	Asura_Movement_System* pxMovementSystem = pxVictim->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	// Get the grapple blueprint
	const Asura_Blueprint* pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, m_uGrappleDescriptionBlueprint);
	if (!pxGrappleDescription) { return false; }

	const Asura_Hash_ID uAnimSetBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_ANIMSET, ASURA_HASH_ID_UNSET);
	if (uAnimSetBlueprint == ASURA_HASH_ID_UNSET) { return false; }

	const Asura_Blueprint* pxAnimSet = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimSetBlueprint);
	if (!pxAnimSet) { return false; }

	m_uGrappleFlags = 0;
	m_xFinalPuppetPositionOffset.Zero();
	m_xFinalPuppetOrientationOffset.Identity();

	// Get all the animations.
	Asura_Value xValue;
	m_bBranching = false;
	Asura_Hash_ID uFirstPersonAttackerAnimID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uThirdPersonAttackerAnimID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uFirstPersonVictimAnimID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uThirdPersonVictimAnimID = ASURA_HASH_ID_UNSET;

	Asura_Hash_ID uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_ATTACKER_HUD, ASURA_HASH_ID_UNSET);
	const Asura_Blueprint* pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uFirstPersonAttackerAnimID = pxAnimations->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START, ASURA_HASH_ID_UNSET);
		// Check for a win anim to see if this is a branching set. This is also checked in the TP blueprint.
		m_bBranching |= pxAnimations->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN, xValue);

		// Events
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(m_axFirstPersonGrappleEvents, m_iNumFirstPersonGrappleEvents);
		pxAnimations->EnumerateParameters(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START_EVENT, 2, xEnumerator);
	}

	uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_ATTACKER_TP, ASURA_HASH_ID_UNSET);
	pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uThirdPersonAttackerAnimID = pxAnimations->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START, ASURA_HASH_ID_UNSET);
		// Check for a win anim to see if this is a branching set. This is also checked in the HUD blueprint.
		m_bBranching |= pxAnimations->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN, xValue);

		// Events
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(m_axThirdPersonGrappleEvents, m_iNumThirdPersonGrappleEvents);
		pxAnimations->EnumerateParameters(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START_EVENT, 2, xEnumerator);
	}

	uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_VICTIM_HUD, ASURA_HASH_ID_UNSET);
	pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uFirstPersonVictimAnimID = pxAnimations->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START, ASURA_HASH_ID_UNSET);

		// Events
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(pxPuppetMove->GetFirstPersonGrappleEventArray(), pxPuppetMove->GetNumFirstPersonGrappleEventsReference());
		pxAnimations->EnumerateParameters(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START_EVENT, 2, xEnumerator);
	}

	uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_VICTIM_TP, ASURA_HASH_ID_UNSET);
	pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uThirdPersonVictimAnimID = pxAnimations->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START, ASURA_HASH_ID_UNSET);

		// Events
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(pxPuppetMove->GetThirdPersonGrappleEventArray(), pxPuppetMove->GetNumThirdPersonGrappleEventsReference());
		pxAnimations->EnumerateParameters(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_START_EVENT, 2, xEnumerator);
	}

	float fTweenInDuration, fTweenOutDuration;
	fTweenInDuration = dbg_fGenericPuppeteerTweenInDuration;
	fTweenOutDuration = m_bBranching ? 0.0f : dbg_fGenericPuppeteerTweenOutDuration;

	// Our anim data.
	SetAnimHashes(uFirstPersonAttackerAnimID, uThirdPersonAttackerAnimID);
	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(fTweenInDuration, fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	// Active grapple difficulty settings.
	int iAttackerDifficulty = pxGrappleDescription->GetParameterAsInt(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DIFFICULTY_ATTACKER, 0);
	int iVictimDifficulty = pxGrappleDescription->GetParameterAsInt(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DIFFICULTY_VICTIM, 0);
	m_uActiveGrappleAttackerBlueprint = xOwner.GetActiveGrappleBlueprintID(iAttackerDifficulty);
	m_uActiveGrappleVictimBlueprint = pxVictim->GetActiveGrappleBlueprintID(iVictimDifficulty);
	m_uActiveGrappleBlueprintType = uAVP_BLUEPRINT_TYPE_GRAPPLEDESIGN;

	// Active grapple sounds.
	if (pxGrappleDescription->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_SOUNDSET_ATTACKER, xValue))
	{
		m_uGrappleSoundsAttackerBlueprint = xValue.GetHashIDValue();
	}
	else
	{
		m_uGrappleSoundsAttackerBlueprint = ASURA_HASH_ID_UNSET;
	}
	if (pxGrappleDescription->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_SOUNDSET_VICTIM, xValue))
	{
		m_uGrappleSoundsVictimBlueprint = xValue.GetHashIDValue();
	}
	else
	{
		m_uGrappleSoundsVictimBlueprint = ASURA_HASH_ID_UNSET;
	}

	// Check if this is an always-fatal attack as the puppet should immediately ragdoll on death if it isn't. Only relevant in this class if this is a non-branching grapple.
	bool bRagdollOnDeath = false;
	if (!m_bBranching)
	{
		Asura_Hash_ID uDamageBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DAMAGE, ASURA_HASH_ID_UNSET);
		const Asura_Blueprint* pxDamage = (uDamageBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEDESIGN, uDamageBlueprint) : NULL;
		if (pxDamage)
		{
			const float fDamage = pxDamage->GetParameterAsFloat(uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERWIN_AMOUNT, 0.0f);
			const float fInitialHealth = pxVictim->GetInitialHealth();
			bRagdollOnDeath = fDamage < fInitialHealth;
		}
	}

	// All puppet data, including anims.
	pxPuppetMove->SetCurrentPuppeteerGuid(GetParentSystem().GetOwnerGuid());
	pxPuppetMove->SetRelativeToPuppeteer(true);
	pxPuppetMove->SetClearQueuedMeleeRequests(true);
	pxPuppetMove->SetClearQueuedCounterRequest(true);
	pxPuppetMove->SetAlwaysCreateRagdollOnDeath(bRagdollOnDeath);
	pxPuppetMove->SetControlDeathInternally(!bRagdollOnDeath);
	pxPuppetMove->SetPreventsFlinch(true);
	pxPuppetMove->SetDisableSurfaceReorientation(true);
	SetPuppetAnimationData(uFirstPersonVictimAnimID, uThirdPersonVictimAnimID, fTweenInDuration, (!m_bBranching && !bRagdollOnDeath) ? 0.0f : fTweenOutDuration, true, true);

	// Puppet interpolation data.
	const float fInitialVictimForwardOffset = pxAnimSet->GetParameterAsFloat(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_FORWARDOFFSET, 0.0f);
	const float fInitialVictimUpOffset = pxAnimSet->GetParameterAsFloat(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_UPOFFSET, 0.0f);
	const bool bFrontAttack = pxAnimSet->GetParameterAsBool(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_FRONTATTACK, true);

	m_xPuppetPositionInterpolation.SetDuration(dbg_fGenericPuppeteerPositionInterpolationTime);
	m_xPuppetPositionInterpolation.SetFinal(Asura_AVector3(0.0f, -fInitialVictimUpOffset, fInitialVictimForwardOffset));
	m_xPuppetOrientationInterpolation.SetDuration(dbg_fGenericPuppeteerPositionInterpolationTime);
	m_xPuppetOrientationInterpolation.SetFinal(bFrontAttack ? Asura_Quat(0.0f, 1.0f, 0.0f, 0.0f) : Asura_Maths::IdentityQuat);

	// Default our victory state to what's appropriate.
	// Check which actors are players, and whether one or both have an active grapple difficulty setting. The attacker gets the advantage unless they're the only player with a difficulty, or if the victim is the only player and neither have a difficulty.
	bool bOwnerIsPlayer = xOwner.IsPlayer();
	bool bVictimIsPlayer = pxVictim->IsPlayer();
	bool bAttackerHasAG = bOwnerIsPlayer && (iAttackerDifficulty > 0);
	bool bVictimHasAG = bVictimIsPlayer && (iVictimDifficulty > 0);
	if (!bAttackerHasAG && !bVictimHasAG)
	{
		m_bAttackerVictorious = bOwnerIsPlayer || !bVictimIsPlayer;
	}
	else
	{
		m_bAttackerVictorious = !bAttackerHasAG || bVictimHasAG;
	}

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_Grapple_Start::Reset()
{
	PARENT::Reset();

	SetPuppetRelativeToOwner(true);
}

void AVP_Move_Melee_Grapple_Start::Finish()
{
	// If this is branching, make sure the next move is actually able to start - if it can't, make sure our anim will be cleared up.
	if (m_bBranching && !GetNextMove())
	{
		EnsureAnimationFinishes();
	}

	m_bHasFinishMoveOffset = CalculateFinishMovePuppetStartOffset();

	PARENT::Finish();
}

void AVP_Move_Melee_Grapple_Start::Abort()
{
	// Make sure our puppet isn't stuck.
	AVP_ClientEntity_Actor* pxPuppet = AVP_ClientEntity_Actor::FindEntity(GetCurrentPuppet());
	if (pxPuppet)
	{
		Asura_Movement_System* pxPuppetMovementSystem = pxPuppet->GetMovementSystem();
		if (pxPuppetMovementSystem)
		{
			AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxPuppetMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
			if (pxPuppetMove)
			{
				pxPuppetMove->SetControlDeathInternally(false);
				pxPuppetMove->SetAlwaysCreateRagdollOnDeath(true);

				if (!pxPuppetMove->GetAVPAnimationData() || !pxPuppetMove->GetAVPAnimationData()->GetAnimationController())
				{
					pxPuppetMovementSystem->AbortCurrentMove(true, true);
				}
				else
				{
					pxPuppetMove->EnsureAnimationFinishes(true);
				}
			}
		}
	}

	PARENT::Abort();

	if (!m_bBranching && pxPuppet)
	{
		AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
		const Asura_Guid uOwnerGuid = xOwner.GetGuid();
		// Check we have the master entity on this machine. That's the local player or the server AI.
		if (m_xParentSystem.IsControlledByLocalPlayer(uOwnerGuid) || (!m_xParentSystem.IsControlledByAnyPlayer(uOwnerGuid) && Asura_Network::IsServer()))
		{
			const Asura_Blueprint* pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, m_uGrappleDescriptionBlueprint);
			if (pxGrappleDescription)
			{
				Asura_Hash_ID uDamageBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DAMAGE, ASURA_HASH_ID_UNSET);
				const Asura_Blueprint* pxDamage = (uDamageBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEDESIGN, uDamageBlueprint) : NULL;
				if (pxDamage)
				{
					// Who's taking damage
					if (m_bAttackerVictorious)
					{
						// Make sure our puppet isn't already dead.
						if (!pxPuppet->IsDead())
						{
							const float fDamage = pxDamage->GetParameterAsFloat(m_bCriticalVictory ? uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERCRITICALWIN_AMOUNT : uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERWIN_AMOUNT, 0.0f);
							if (fDamage > 0.0f)
							{
								// Send a damage message.
								Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
								if(pxMessage)
								{
									AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
									const AVP_EntityDamage_RequestVictim	xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_EXPLOSIVE,
																					uOwnerGuid,
																					ASURA_HASH_ID_UNSET,
																					m_uCurrentPuppet,
																					fDamage,
																					pxPuppet->GetCurrentPosition(),
																					Asura_Vector_3(0.0f, -1.0f, 0.0f),
																					Asura_Vector_3(0.0f, 0.0f, 0.0f));
									pxData->SetDamage(xDamage);

									pxMessage->SetFromGuid(uOwnerGuid);
									pxMessage->SetToGuid(m_uCurrentPuppet);
									Asura_Network::Send(pxMessage);
								}
							}
						}
					}
					else
					{
						// Make sure we're not already dead.
						if (!xOwner.IsDead())
						{
							const float fDamage = pxDamage->GetParameterAsFloat(m_bCriticalVictory ? uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMCRITICALWIN_AMOUNT : uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMWIN_AMOUNT, 0.0f);
							if (fDamage > 0.0f)
							{
								// Send a damage message.
								Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
								if(pxMessage)
								{
									AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
									const AVP_EntityDamage_RequestVictim	xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_EXPLOSIVE,
																					m_uCurrentPuppet,
																					ASURA_HASH_ID_UNSET,
																					uOwnerGuid,
																					fDamage,
																					pxPuppet->GetCurrentPosition(),
																					Asura_Vector_3(0.0f, -1.0f, 0.0f),
																					Asura_Vector_3(0.0f, 0.0f, 0.0f));
									pxData->SetDamage(xDamage);

									pxMessage->SetFromGuid(m_uCurrentPuppet);
									pxMessage->SetToGuid(uOwnerGuid);
									Asura_Network::Send(pxMessage);
								}
							}
						}
					}
				}
			}
		}
	}
}

void AVP_Move_Melee_Grapple_Start::UpdateStates(const float fTimeStep)
{
	PARENT::UpdateStates(fTimeStep);

	const AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
	if ((xOwner.GetFlagsForAllAnimsLastFrame() & AVP_ANIMFLAG_COMPLETE) != 0)
	{
		m_bAnimCompleteFlagHit = true;
	}

	if (m_bAnimCompleteFlagHit && m_bResultFinalised)
	{
		m_bGrappleCanFinish = true;
	}
}

void AVP_Move_Melee_Grapple_Start::SendStartMsg()
{
	if (!m_bUseServerStartMessage)
	{
		PARENT::SendStartMsg();
		return;
	}

	if (AVP_Settings::GetIsServer())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_MOVESTART_SERVER, sizeof(AVP_Msg_Grapple_Move));

		if(pxMessage != NULL)
		{
			AVP_Msg_Grapple_Move* pxData = static_cast<AVP_Msg_Grapple_Move*>(pxMessage->GetData());
			const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(false);
				pxData->SetTargetGuid(m_uCurrentPuppet);
				pxData->SetTargetEntityIsPuppet(true);
				pxData->SetGrappleDescriptionBlueprint(m_uGrappleDescriptionBlueprint);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}
}

bool AVP_Move_Melee_Grapple_Start::CalculateFinishMovePuppetStartOffset()
{
	// Get the puppet's target initial position and orientation, and apply the animation offsets.
	if (!m_bPuppetRelativeToOwner) { return false; }
	if (!m_pxAVPAnimationData) { return false; }

	const AVP_ClientActor_Vertebrate* pxVictim = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentPuppet);
	if (!pxVictim) { return false; }

	const Asura_Movement_System* pxMovementSystem = pxVictim->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	const AVP_Move_Puppet* pxPuppetMove = static_cast<const AVP_Move_Puppet*>(pxMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	const Asura_MoveData_Animation* pxPuppetAnimData = pxPuppetMove->GetAVPAnimationData();
	if (!pxPuppetAnimData) { return false; }

	m_xFinalPuppetPositionOffset = m_xPuppetPositionInterpolation.GetFinal();
	m_xFinalPuppetOrientationOffset = m_xPuppetOrientationInterpolation.GetFinal();

	// We need the offsets on the final frame.
	Asura_AVector3 xOwnerOffsetPos, xPuppetOffsetPos;
	Asura_Quat xOwnerOffsetOri, xPuppetOffsetOri;

	if (!m_pxAVPAnimationData->CalculateMovementOffset(1.0f, &xOwnerOffsetPos, &xOwnerOffsetOri))
	{
		xOwnerOffsetPos.Zero();
		xOwnerOffsetOri.Identity();
	}
	
	if (!pxPuppetAnimData->CalculateMovementOffset(1.0f, &xPuppetOffsetPos, &xPuppetOffsetOri))
	{
		xPuppetOffsetPos.Zero();
		xPuppetOffsetOri.Identity();
	}

	xPuppetOffsetPos = m_xFinalPuppetOrientationOffset.RotateVector(xPuppetOffsetPos);

	m_xFinalPuppetPositionOffset = m_xFinalPuppetPositionOffset + xPuppetOffsetPos;
	// As we're relative, we also need to take off our owner's movement to keep the offset correct.
	m_xFinalPuppetPositionOffset -= xOwnerOffsetPos;

	Asura_Quat xInvertOwnerOffsetOri(xOwnerOffsetOri);
	xInvertOwnerOffsetOri.Invert();
	m_xFinalPuppetPositionOffset = xInvertOwnerOffsetOri.RotateVector(m_xFinalPuppetPositionOffset);

	m_xFinalPuppetOrientationOffset = m_xFinalPuppetOrientationOffset * xInvertOwnerOffsetOri;

	return true;
}

////////////////////////////////////////////////////////////////////////////////
// Class		: AVP_Move_Melee_Grapple_Finish
// Author       : Anthony Gregson
// Date         : 17/04/09
////////////////////////////////////////////////////////////////////////////////

AVP_Move_Melee_Grapple_Finish::AVP_Move_Melee_Grapple_Finish(Asura_Movement_System& xParentSystem, const Asura_Hash_ID uMoveHash)
: PARENT(xParentSystem, uMoveHash)
, m_uGrappleDescriptionBlueprint(ASURA_HASH_ID_UNSET)
, m_xInitialPuppetPositionOffset(Asura_Maths::ZeroVector3)
, m_xInitialPuppetOrientationOffset(Asura_Maths::IdentityQuat)
, m_uGrappleFlags(0)
{
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

AVP_Move_Melee_Grapple_Finish::AVP_Move_Melee_Grapple_Finish(Asura_Movement_System& xParentSystem)
: PARENT(xParentSystem, GetStaticHash())
, m_uGrappleDescriptionBlueprint(ASURA_HASH_ID_UNSET)
, m_xInitialPuppetPositionOffset(Asura_Maths::ZeroVector3)
, m_xInitialPuppetOrientationOffset(Asura_Maths::IdentityQuat)
, m_uGrappleFlags(0)
{
	m_bClearQueuedMeleeRequests = true;
	m_bClearQueuedCounterRequest = true;
	m_bUseServerStartMessage = true;
	SetAlwaysCheckNextMoves(false);
	SetPuppetRelativeToOwner(true);
	m_bInterpolateInitialPuppetPos = true;
	m_bInterpolateInitialPuppetOri = true;

	m_bForceStanding = true;
	m_bPreventsFlinch = true;

	SetPositionUsage(POSITION_SET_DIRECTLY);
	SetOrientationUsage(ORIENTATION_SET_DIRECTLY);
}

bool AVP_Move_Melee_Grapple_Finish::ShouldStart() const
{
	// Check the parent details first.
	if (!PARENT::ShouldStart()) { return false; }

	// This has to come from a grapple start
	const AVP_Move_Melee_Grapple_Start* pxCurrentMove = static_cast<const AVP_Move_Melee_Grapple_Start*>(GetParentSystem().GetCurrentMove());
	if (!pxCurrentMove) { return false; }

	// No finish move for unbranching grapples.
	if (!pxCurrentMove->IsBranching()) { return false; }

	const Asura_Guid uPuppetGuid = pxCurrentMove->GetCurrentPuppet();
	const AVP_ClientEntity_Actor* pxVictim = AVP_ClientEntity_Actor::FindEntity(uPuppetGuid);

	if (!pxVictim) { return false; }
	if (pxVictim->IsRagdoll()) { return false; }

	return true;
}

bool AVP_Move_Melee_Grapple_Finish::Start(const Asura_Msg_Move* pxStartMsg)
{
	bool bAttackerWin;
	bool bCriticalWin;

	// We need to set up the puppet move now.
	if (pxStartMsg)
	{
		// TODO - need some way to set up these moves to start from message.
		Asura_Assert(false, "Starting a grapple finish by message not yet implemented.");
		return false;
		// Started from the server start message, get the enemy we've grabbed here.
		//const AVP_Msg_Grapple_Move* pxMeleeMsg = static_cast<const AVP_Msg_Grapple_Move*>(pxStartMsg);
		//m_uCurrentPuppet = pxMeleeMsg->GetTargetGuid();
		//m_uGrappleDescriptionBlueprint = pxMeleeMsg->GetGrappleDescriptionBlueprint();
	}
	else
	{
		// Started by ShouldStart, get the enemy we've been grappling from the previous move.
		const AVP_Move_Melee_Grapple_Start* pxPreviousMove = static_cast<const AVP_Move_Melee_Grapple_Start*>(GetParentSystem().GetCurrentMove());
		if (!pxPreviousMove) { return false; }

		m_uCurrentPuppet = pxPreviousMove->GetCurrentPuppet();
		m_uGrappleDescriptionBlueprint = pxPreviousMove->GetGrappleDescriptionBlueprint();
		bAttackerWin = pxPreviousMove->IsAttackerVictorious();
		bCriticalWin = pxPreviousMove->IsCriticalVictory();

		// So we can branch into this move without waiting for the previous animation to finish we need to set up a position interpolation for the puppet.
		bool bInterpolatePuppet = pxPreviousMove->GetFinishMovePuppetStartOffset(m_xInitialPuppetPositionOffset, m_xInitialPuppetOrientationOffset);
		if (bInterpolatePuppet)
		{
			m_xPuppetPositionInterpolation.SetDuration(dbg_fGenericPuppeteerPositionInterpolationTime);
			m_xPuppetPositionInterpolation.SetFinal(m_xInitialPuppetPositionOffset);
			m_xPuppetOrientationInterpolation.SetDuration(dbg_fGenericPuppeteerPositionInterpolationTime);
			m_xPuppetOrientationInterpolation.SetFinal(m_xInitialPuppetOrientationOffset);
		}
		m_bInterpolateInitialPuppetPos = bInterpolatePuppet;
		m_bInterpolateInitialPuppetOri = bInterpolatePuppet;
	}

	// Get the victim's puppet move.
	AVP_ClientActor_Vertebrate* pxVictim = AVP_ClientActor_Vertebrate::FindEntity(m_uCurrentPuppet);
	if (!pxVictim) { return false; }

	Asura_Movement_System* pxMovementSystem = pxVictim->GetMovementSystem();
	if (!pxMovementSystem) { return false; }

	AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
	if (!pxPuppetMove) { return false; }

	// Get the grapple blueprint
	const Asura_Blueprint* pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, m_uGrappleDescriptionBlueprint);
	if (!pxGrappleDescription) { return false; }

	const Asura_Hash_ID uAnimSetBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_ANIMSET, ASURA_HASH_ID_UNSET);
	if (uAnimSetBlueprint == ASURA_HASH_ID_UNSET) { return false; }

	const Asura_Blueprint* pxAnimSet = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimSetBlueprint);
	if (!pxAnimSet) { return false; }

	m_uGrappleFlags = 0;

	// Get all the animations.
	Asura_Hash_ID uAttackerAnimParamID;
	Asura_Hash_ID uAttackerEventParamID;
	Asura_Hash_ID uVictimAnimParamID;
	Asura_Hash_ID uVictimEventParamID;

	if (bAttackerWin)
	{
		if (bCriticalWin)
		{
			uAttackerAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALWIN;
			uAttackerEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALWIN_EVENT;
			uVictimAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALLOSE;
			uVictimEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALLOSE_EVENT;
		}
		else
		{
			uAttackerAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN;
			uAttackerEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN_EVENT;
			uVictimAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_LOSE;
			uVictimEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_LOSE_EVENT;
		}
	}
	else
	{
		if (bCriticalWin)
		{
			uAttackerAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALLOSE;
			uAttackerEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALLOSE_EVENT;
			uVictimAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALWIN;
			uVictimEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_CRITICALWIN_EVENT;
		}
		else
		{
			uAttackerAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_LOSE;
			uAttackerEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_LOSE_EVENT;
			uVictimAnimParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN;
			uVictimEventParamID = uAVP_BLUEPRINT_PARAM_GRAPPLEANIMS_WIN_EVENT;
		}
	}

	Asura_Hash_ID uFirstPersonAttackerAnimID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uThirdPersonAttackerAnimID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uFirstPersonVictimAnimID = ASURA_HASH_ID_UNSET;
	Asura_Hash_ID uThirdPersonVictimAnimID = ASURA_HASH_ID_UNSET;

	Asura_Hash_ID uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_ATTACKER_HUD, ASURA_HASH_ID_UNSET);
	const Asura_Blueprint* pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uFirstPersonAttackerAnimID = pxAnimations->GetParameterAsHash(uAttackerAnimParamID, ASURA_HASH_ID_UNSET);

		// Events.
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(m_axFirstPersonGrappleEvents, m_iNumFirstPersonGrappleEvents);
		pxAnimations->EnumerateParameters(uAttackerEventParamID, 2, xEnumerator);
	}

	uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_ATTACKER_TP, ASURA_HASH_ID_UNSET);
	pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uThirdPersonAttackerAnimID = pxAnimations->GetParameterAsHash(uAttackerAnimParamID, ASURA_HASH_ID_UNSET);

		// Events.
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(m_axThirdPersonGrappleEvents, m_iNumThirdPersonGrappleEvents);
		pxAnimations->EnumerateParameters(uAttackerEventParamID, 2, xEnumerator);
	}

	pxPuppetMove->ResetGrappleEvents();

	uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_VICTIM_HUD, ASURA_HASH_ID_UNSET);
	pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uFirstPersonVictimAnimID = pxAnimations->GetParameterAsHash(uVictimAnimParamID, ASURA_HASH_ID_UNSET);

		// Events.
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(pxPuppetMove->GetFirstPersonGrappleEventArray(), pxPuppetMove->GetNumFirstPersonGrappleEventsReference());
		pxAnimations->EnumerateParameters(uVictimEventParamID, 2, xEnumerator);
	}

	uAnimationsBlueprint = pxAnimSet->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEANIMSET_VICTIM_TP, ASURA_HASH_ID_UNSET);
	pxAnimations = (uAnimationsBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uAnimationsBlueprint) : NULL;
	if (pxAnimations)
	{
		uThirdPersonVictimAnimID = pxAnimations->GetParameterAsHash(uVictimAnimParamID, ASURA_HASH_ID_UNSET);

		// Events.
		AVP_Blueprint_ParameterEnumerator_GrappleEvents xEnumerator(pxPuppetMove->GetThirdPersonGrappleEventArray(), pxPuppetMove->GetNumThirdPersonGrappleEventsReference());
		pxAnimations->EnumerateParameters(uVictimEventParamID, 2, xEnumerator);
	}

	float fTweenInDuration, fTweenOutDuration;
	fTweenInDuration = dbg_fGenericPuppeteerTweenInDuration;
	fTweenOutDuration = dbg_fGenericPuppeteerTweenOutDuration;

	// Our anim data.
	SetAnimHashes(uFirstPersonAttackerAnimID, uThirdPersonAttackerAnimID);
	if (m_pxAVPAnimationData)
	{
		m_pxAVPAnimationData->SetTweenData(fTweenInDuration, fTweenOutDuration, true);
		m_pxAVPAnimationData->SetLeaveAnimAtEnd(true);
	}

	// Check if this is an always-fatal attack as the puppet should immediately ragdoll on death if it isn't. Only relevant if the attacker won.
	bool bRagdollOnDeath = true; // Unlike the start move, there are more situations where the puppet should immediately ragdoll than not, since it includes the counters which will never be doing damage to the puppet internally.
	if (bAttackerWin)
	{
		Asura_Hash_ID uDamageBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DAMAGE, ASURA_HASH_ID_UNSET);
		const Asura_Blueprint* pxDamage = (uDamageBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEDESIGN, uDamageBlueprint) : NULL;
		if (pxDamage)
		{
			const float fDamage = pxDamage->GetParameterAsFloat(bCriticalWin ? uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERCRITICALWIN_AMOUNT : uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERWIN_AMOUNT, 0.0f);
			const float fInitialHealth = pxVictim->GetInitialHealth();
			bRagdollOnDeath = fDamage < fInitialHealth;
		}
	}

	// All puppet data, including anims.
	pxPuppetMove->SetCurrentPuppeteerGuid(GetParentSystem().GetOwnerGuid());
	pxPuppetMove->SetRelativeToPuppeteer(true);
	pxPuppetMove->SetClearQueuedMeleeRequests(true);
	pxPuppetMove->SetClearQueuedCounterRequest(true);
	pxPuppetMove->SetAlwaysCreateRagdollOnDeath(bRagdollOnDeath);
	pxPuppetMove->SetControlDeathInternally(!bRagdollOnDeath);
	pxPuppetMove->SetPreventsFlinch(true);
	pxPuppetMove->SetDisableSurfaceReorientation(true);
	SetPuppetAnimationData(uFirstPersonVictimAnimID, uThirdPersonVictimAnimID, fTweenInDuration, (bAttackerWin && !bRagdollOnDeath) ? 0.0f : fTweenOutDuration, true, true);

	return PARENT::Start(pxStartMsg);
}

void AVP_Move_Melee_Grapple_Finish::Reset()
{
	PARENT::Reset();

	SetPuppetRelativeToOwner(true);
}

void AVP_Move_Melee_Grapple_Finish::Abort()
{
	// Make sure our puppet isn't stuck.
	AVP_ClientEntity_Actor* pxPuppet = AVP_ClientEntity_Actor::FindEntity(GetCurrentPuppet());
	if (pxPuppet)
	{
		Asura_Movement_System* pxPuppetMovementSystem = pxPuppet->GetMovementSystem();
		if (pxPuppetMovementSystem)
		{
			AVP_Move_Puppet* pxPuppetMove = static_cast<AVP_Move_Puppet*>(pxPuppetMovementSystem->GetMove(AVP_Move_Puppet::GetStaticHash()));
			if (pxPuppetMove)
			{
				pxPuppetMove->SetControlDeathInternally(false);
				pxPuppetMove->SetAlwaysCreateRagdollOnDeath(true);

				if (!pxPuppetMove->GetAVPAnimationData() || !pxPuppetMove->GetAVPAnimationData()->GetAnimationController())
				{
					pxPuppetMovementSystem->AbortCurrentMove(true, true);
				}
				else
				{
					pxPuppetMove->EnsureAnimationFinishes(true);
				}
			}
		}
	}

	PARENT::Abort();

	if (pxPuppet)
	{
		AVP_ClientEntity_Actor& xOwner = GetAVPParentSystem().GetActorOwner();
		const Asura_Guid uOwnerGuid = xOwner.GetGuid();
		// Check we have the master entity on this machine. That's the local player or the server AI.
		if (m_xParentSystem.IsControlledByLocalPlayer(uOwnerGuid) || (!m_xParentSystem.IsControlledByAnyPlayer(uOwnerGuid) && Asura_Network::IsServer()))
		{
			const Asura_Blueprint* pxGrappleDescription = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLE, m_uGrappleDescriptionBlueprint);
			if (pxGrappleDescription)
			{
				Asura_Hash_ID uDamageBlueprint = pxGrappleDescription->GetParameterAsHash(uAVP_BLUEPRINT_PARAM_GRAPPLEDESCRIPTION_DAMAGE, ASURA_HASH_ID_UNSET);
				const Asura_Blueprint* pxDamage = (uDamageBlueprint != ASURA_HASH_ID_UNSET) ? Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEDESIGN, uDamageBlueprint) : NULL;
				if (pxDamage)
				{
					// Who's taking damage
					if (m_bAttackerVictorious)
					{
						// Make sure our puppet isn't already dead.
						if (xOwner.IsDead() && !pxPuppet->IsDead())
						{
							const float fDamage = pxDamage->GetParameterAsFloat(m_bCriticalVictory ? uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERCRITICALWIN_AMOUNT : uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_ATTACKERWIN_AMOUNT, 0.0f);
							if (fDamage > 0.0f)
							{
								// Send a damage message.
								Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
								if(pxMessage)
								{
									AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
									const AVP_EntityDamage_RequestVictim	xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_EXPLOSIVE,
																					uOwnerGuid,
																					ASURA_HASH_ID_UNSET,
																					m_uCurrentPuppet,
																					fDamage,
																					pxPuppet->GetCurrentPosition(),
																					Asura_Vector_3(0.0f, -1.0f, 0.0f),
																					Asura_Vector_3(0.0f, 0.0f, 0.0f));
									pxData->SetDamage(xDamage);

									pxMessage->SetFromGuid(uOwnerGuid);
									pxMessage->SetToGuid(m_uCurrentPuppet);
									Asura_Network::Send(pxMessage);
								}
							}
						}
					}
					else
					{
						// Make sure we're not already dead.
						if (pxPuppet->IsDead() && !xOwner.IsDead())
						{
							const float fDamage = pxDamage->GetParameterAsFloat(m_bCriticalVictory ? uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMCRITICALWIN_AMOUNT : uAVP_BLUEPRINT_PARAM_GRAPPLEDAMAGE_VICTIMWIN_AMOUNT, 0.0f);
							if (fDamage > 0.0f)
							{
								// Send a damage message.
								Asura_Message* pxMessage = Asura_Network::CreateMessage(AVP_MSGID_ENTITYDAMAGE_REQUEST_VICTIM, sizeof(AVP_Message_EntityDamage_RequestVictim));
								if(pxMessage)
								{
									AVP_Message_EntityDamage_RequestVictim*	pxData = static_cast<AVP_Message_EntityDamage_RequestVictim*>(pxMessage->GetData());
									const AVP_EntityDamage_RequestVictim	xDamage(uAVP_BLUEPRINT_NAME_DAMAGE_EXPLOSIVE,
																					m_uCurrentPuppet,
																					ASURA_HASH_ID_UNSET,
																					uOwnerGuid,
																					fDamage,
																					pxPuppet->GetCurrentPosition(),
																					Asura_Vector_3(0.0f, -1.0f, 0.0f),
																					Asura_Vector_3(0.0f, 0.0f, 0.0f));
									pxData->SetDamage(xDamage);

									pxMessage->SetFromGuid(m_uCurrentPuppet);
									pxMessage->SetToGuid(uOwnerGuid);
									Asura_Network::Send(pxMessage);
								}
							}
						}
					}
				}
			}
		}
	}
}

void AVP_Move_Melee_Grapple_Finish::SendStartMsg()
{
	if (!m_bUseServerStartMessage)
	{
		PARENT::SendStartMsg();
		return;
	}

	if (AVP_Settings::GetIsServer())
	{
		Asura_Message*	pxMessage	= Asura_Network::CreateMessage(AVP_MSGID_MOVESTART_SERVER, sizeof(AVP_Msg_Grapple_Move));

		if(pxMessage != NULL)
		{
			AVP_Msg_Grapple_Move* pxData = static_cast<AVP_Msg_Grapple_Move*>(pxMessage->GetData());
			const Asura_Guid uOwnerGuid = m_xParentSystem.GetOwnerGuid();

			if(pxData != NULL)
			{
				pxData->ClearFlags();
				pxData->SetMoveID(m_uHash);
				pxData->SetIgnoredByOwner(false);
				pxData->SetTargetGuid(m_uCurrentPuppet);
				pxData->SetTargetEntityIsPuppet(true);
				pxData->SetGrappleDescriptionBlueprint(m_uGrappleDescriptionBlueprint);
			}

			pxMessage->SetToGuid(uOwnerGuid);
			pxMessage->SetFromGuid(uOwnerGuid);
			pxMessage->UseGuaranteedDelivery();

			Asura_Network::Send(pxMessage);
		}
	}
}

////////////////////////////////////////////////////////////////////////////////
// Parameter enumerator to set up melee move variations easily.
////////////////////////////////////////////////////////////////////////////////

AVP_Blueprint_ParameterEnumerator_MeleeVariations::AVP_Blueprint_ParameterEnumerator_MeleeVariations(AVP_Move_Melee& xMove, bool bThirdPersonSet)
: m_xMove(xMove)
, m_uNumVariations(0)
, m_bThirdPersonSet(bThirdPersonSet)
{
}

void AVP_Blueprint_ParameterEnumerator_MeleeVariations::BeginEnumeration()
{
}

bool AVP_Blueprint_ParameterEnumerator_MeleeVariations::Enumerate(const u_int uIndex, const char* const szName, const Asura_Value& xValue)
{
	Asura_Assert(m_uNumVariations < MAX_MELEE_MOVE_VARIATIONS, "Too many melee attack variations, ignoring the rest.");
	if (m_uNumVariations >= MAX_MELEE_MOVE_VARIATIONS) { return false; }

	m_auVariationHashes[m_uNumVariations++] = xValue.GetHashIDValue();

	return true;
}

void AVP_Blueprint_ParameterEnumerator_MeleeVariations::EndEnumeration(const bool bSuccess)
{
	// Failure just means we had too many variations, set what we have regardless.
	if (m_uNumVariations > 0)
	{
		if (m_bThirdPersonSet)
		{
			m_xMove.SetThirdPersonAnimHashesForVariations(m_auVariationHashes, m_uNumVariations);
		}
		else
		{
			m_xMove.SetFirstPersonAnimHashesForVariations(m_auVariationHashes, m_uNumVariations);
		}
	}
}


// AVP_Blueprint_ParameterEnumerator_MeleeVariations_MultipleParameters

AVP_Blueprint_ParameterEnumerator_MeleeVariations_MultipleParameters::AVP_Blueprint_ParameterEnumerator_MeleeVariations_MultipleParameters(AVP_Move_Melee& xMove, bool bThirdPersonSet)
: PARENT(xMove, bThirdPersonSet)
, m_bSuccess(true)
{
}

void AVP_Blueprint_ParameterEnumerator_MeleeVariations_MultipleParameters::EnumeratedAllParameters()
{
	// This is when we do what the parent class did to end the enumeration.
	AVP_Blueprint_ParameterEnumerator_MeleeVariations::EndEnumeration(m_bSuccess);
}


////////////////////////////////////////////////////////////////////////////////
// Enumerator for grapple events, storing all the details for the current move.
////////////////////////////////////////////////////////////////////////////////

AVP_Blueprint_ParameterEnumerator_GrappleEvents::AVP_Blueprint_ParameterEnumerator_GrappleEvents(AVP_GrappleEvent_Entry* pxEntries, int& iNumEntries)
: m_pxEntries(pxEntries)
, m_iNumEntries(iNumEntries)
{
	Asura_Assert(m_pxEntries, "Must have an array of entries to use this enumerator.");
	Asura_Assert(m_iNumEntries == 0, "Should be no current grapple event entries stored, you run the risk of events getting left behind if this happens.");
}

void AVP_Blueprint_ParameterEnumerator_GrappleEvents::BeginEnumeration()
{
	m_iNumEntries = 0;
}

bool AVP_Blueprint_ParameterEnumerator_GrappleEvents::Enumerate(const u_int uIndex, const Asura_Value& xValue)
{
	Asura_Assert(m_iNumEntries < iAVP_MAX_GRAPPLE_EVENTS, "Too many grapple events for a single anim (maximum is 8), ignoring the rest.");
	if (m_iNumEntries >= iAVP_MAX_GRAPPLE_EVENTS) { return false; }

	Asura_Hash_ID uEventEntryBlueprintID = xValue.GetHashIDValue();
	const Asura_Blueprint* pxBlueprint = Asura_Blueprint_System::GetBlueprintByHashID(uAVP_BLUEPRINT_TYPE_GRAPPLEANIMATION, uEventEntryBlueprintID);
	if (!pxBlueprint) { return true; }

	Asura_Value xInternalValue;
	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEEVENTENTRY_EVENTDATA, xInternalValue))
	{
		m_pxEntries[m_iNumEntries].m_uEventDataBlueprintID = xInternalValue.GetHashIDValue();
	}
	else
	{
		m_pxEntries[m_iNumEntries].m_uEventDataBlueprintID = ASURA_HASH_ID_UNSET;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEEVENTENTRY_STARTFRAME, xInternalValue))
	{
		m_pxEntries[m_iNumEntries].m_iStartFrame = xInternalValue.GetIntValue();
	}
	else
	{
		m_pxEntries[m_iNumEntries].m_iStartFrame = -1;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEEVENTENTRY_STARTANIMFLAG, xInternalValue))
	{
		m_pxEntries[m_iNumEntries].m_iStartAnimFlag = xInternalValue.GetIntValue();
	}
	else
	{
		m_pxEntries[m_iNumEntries].m_iStartAnimFlag = 0;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEEVENTENTRY_ENDFRAME, xInternalValue))
	{
		m_pxEntries[m_iNumEntries].m_iEndFrame = xInternalValue.GetIntValue();
	}
	else
	{
		m_pxEntries[m_iNumEntries].m_iEndFrame = -1;
	}

	if (pxBlueprint->GetParameterValueByHashID(uAVP_BLUEPRINT_PARAM_GRAPPLEEVENTENTRY_ENDANIMFLAG, xInternalValue))
	{
		m_pxEntries[m_iNumEntries].m_iEndAnimFlag = xInternalValue.GetIntValue();
	}
	else
	{
		m_pxEntries[m_iNumEntries].m_iEndAnimFlag = 0;
	}

	m_pxEntries[m_iNumEntries].m_uTypeData = 0;
	m_pxEntries[m_iNumEntries].m_bRunning = false;
	m_pxEntries[m_iNumEntries].m_bFinished = false;

	++m_iNumEntries;

	return true;
}

void AVP_Blueprint_ParameterEnumerator_GrappleEvents::EndEnumeration(const bool bSuccess)
{
}


/// End of file. ///
