//////////////////////////////////////////////////////////////////////////////
//
// PS2Strip_Exporter.cpp
//
// By: Morgan Parry.
//
// Copyright (c) 2001 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

// I N C L U D E S ///////////////////////////////////////////////////////////
#include "Core/Asura.h"

// This file's header.
#include "PS2/PS2Strip/PS2Strip_Exporter.h"

// Stdlib independent.
#include "Core/Asura_MemoryManagement_DebugOff.h" // Std headers define their own new and delete operators and our macros break them.
#pragma warning( push )
#pragma warning( disable : 4702 ) // unreachable code
#include <algorithm>
#pragma warning( pop )
#include "Core/Asura_MemoryManagement_DebugOn.h"

// Stdlib PC.
#include <windows.h>

// Engine independent.
#include "Animation/Asura_Animation.h"
#include "Animation/Asura_Hierarchy.h"
#if defined( ASURA_USE_BOTANICALS )
#include "Botanicals/Asura_Botanicals_InstanceManager.h"
#include "Botanicals/Asura_Botanicals_TemplateManager.h"
#endif // #if defined( ASURA_USE_BOTANICALS )
#if defined( ASURA_USE_CARPET )
#include "Carpet/Asura_Carpet_MeshManager.h"
#include "Carpet/Asura_Carpet_StyleManager.h"
#endif // #if defined( ASURA_USE_CARPET )

#include "Misc/Tools_HashTable.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Cutscene/Asura_Cutscene.h"
#include "Cutscene/Asura_Cutscene_System.h"
#include "Debugging/Asura_Version.h"
#include "Debugging/Asura_Logfile.h"
#include "Light/Asura_GI_Solution.h"
#include "Light/Asura_Light.h"
#include "Mesh/Asura_Env.h"
#include "Mesh/Asura_Env_System.h"
#include "Mesh/Asura_Mesh.h"
#include "Rendering/Asura_Colour.h"
#include "Rendering/Asura_Module.h"

// Engine PS2.
#include "PS2/Animation/Asura_PS2_SkinTypes.h"
#if defined( ASURA_USE_BOTANICALS )
#include "PS2/Botanicals/Asura_PS2_Botanicals_Shared.h"
#endif // #if defined( ASURA_USE_BOTANICALS )
#if defined( ASURA_USE_CARPET )
#include "PS2/Carpet/Asura_PS2_Carpet_Mesh_Shared.h"
#endif // #if defined( ASURA_USE_CARPET )
#include "PS2/ChunkSystem/Asura_Chunk_ResourceFile_PS2.h"
#include "PS2/ChunkSystem/Asura_Chunk_ResourceFile_PS2_SkinLODThresholds.h"
#include "PS2/ChunkSystem/Asura_Chunk_ResourceFile_PS2_StrippedHierSkin.h"
#include "PS2/ChunkSystem/Asura_Chunk_ResourceFile_PS2_StrippedObject.h"
#include "PS2/Mesh/Asura_PS2_Env_Shared.h"

#ifdef ASURA_USE_BOTANICALS
#include "PS2/Rendering/Asura_PS2_BotanicalsDefines.h"
#endif // #ifdef ASURA_USE_BOTANICALS

#if defined( ASURA_USE_CARPET )
#include "PS2/Rendering/Asura_PS2_CarpetMeshDefines.h"
#endif // #if defined( ASURA_USE_CARPET )

#include "PS2/Rendering/Asura_PS2_GSMDefines.h"

// Tools.
#include "Lighting/Tools_Environment_LOS.h"
#include "Load&Save/LevelExport/Tools_Level_Export.h"
#include "Load&Save/Tools_SaveAsura.h"
#if defined( ASURA_USE_CARPET ) || defined( ASURA_USE_BOTANICALS )
#include "Misc/Tools_DynamicArray.h"
#endif //  #if defined( ASURA_USE_CARPET ) || defined( ASURA_USE_BOTANICALS )
#include "MeshProcessing/Tools_SubdividedMesh.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Regions/Tools_Mode_Regions.h"

// PS2Strip headers.
#include "PS2/PS2Strip/PS2Strip_GeometryTypes.h"
#include "PS2/PS2Strip/PS2Strip_PS2Stuff.h"
#include "PS2/PS2Strip/PS2Strip_ToolsInterface.h"
#include "PS2/PS2Strip/Tools_PS2_DMAChainWriter.h"

// NVTriStrip headers.
#include "PS2/PS2Strip/NVTriStrip/NvTriStripObjects.h"

// M A C R O S ///////////////////////////////////////////////////////////////

// Check that the sizes of various types match the PS2. We require binary
// compatibility in the export.
#if USHRT_MAX != 0xffff
    #error The size of the short integer type on this platform does not match the PS2.
#endif
#if UINT_MAX != 0xffffffff
    #error The size of the integer type on this platform does not match the PS2.
#endif
// Ought to check float here too.

#define WRITE_DATA( file_handle, cumulative_count, data_ptr, num_bytes ) \
{ \
    const int iNumBytes = static_cast<int>( num_bytes ); \
    Asura_Assert( iNumBytes > 0, "" ); \
    \
    DWORD uNumBytesWritten; \
    BOOL bRet = WriteFile( (file_handle), (data_ptr), iNumBytes, &uNumBytesWritten, NULL ); \
    if( !bRet || uNumBytesWritten != static_cast<DWORD>(iNumBytes) ) \
    { \
        PS2Strip::LogError( "ERROR: Failed to write all bytes to hierarchical skin file. Reason: %s\n", PS2Strip::GetLastSystemError() ); \
        return false; \
    } \
    \
    (cumulative_count) += iNumBytes; \
}



#define SKIP_DATA( file_handle, cumulative_count, num_bytes ) \
{ \
    const int iNumBytes = static_cast<int>( num_bytes ); \
    Asura_Assert( iNumBytes > 0, "" ); \
    \
    SetFilePointer( (file_handle), iNumBytes, NULL, FILE_CURRENT ); \
    SetEndOfFile( (file_handle) ); \
    (cumulative_count) += iNumBytes; \
}

#define PAD_TO_ALIGNMENT( file_handle, alignment, cumulative_count, reference_point ) \
{ \
    int iBytesSinceRefPoint = (cumulative_count) - (reference_point); \
    Asura_Assert( iBytesSinceRefPoint >= 0, "" ); \
    if( iBytesSinceRefPoint & ((alignment) - 1) ) /* alignment is assumed to be a power of 2 */ \
    { \
        int iPadBytes = (alignment) - (iBytesSinceRefPoint & ((alignment) - 1)); \
        for( ; iPadBytes > 0; iPadBytes -= 4 ) \
        { \
            WRITE_DATA( (file_handle), (cumulative_count), "\0\0\0\0", Asura_Maths::Min(4, iPadBytes) ); \
        } \
    } \
}

float PS2Strip_Exporter::s_fBotanicalBillboardLightingContrastFactor = 1.0f;

// T Y P E S /////////////////////////////////////////////////////////////////

// This enumeration is just used for ordering the materials. Not all
// categories apply to all mesh types but the ordering is pretty much
// universal, or at least the parts that matter are.
enum RENDERPASS
{
    RENDERPASS_OPAQUE,
    RENDERPASS_FMV,
    RENDERPASS_ON_OFF_TRANSPARENCY,
    RENDERPASS_TRANSLUCENT,
    RENDERPASS_ADDITIVE_ALPHA,
    RENDERPASS_NUM_RENDERPASSES,
};

enum TIMING_PROFILEAREA
{
    TIMING_PROFILEAREA_NVTRISTRIP,
    TIMING_PROFILEAREA_OTHER,
    TIMING_PROFILEAREA_PS2STRIP_ELIM_DUP_VTX_INFO,
    TIMING_PROFILEAREA_PS2STRIP_SUBDIV_AND_PRELIGHT,
    TIMING_PROFILEAREA_PS2STRIP_OUTPUT,
    TIMING_PROFILEAREA_PS2STRIP_OTHER,
    TIMING_PROFILEAREA_REFRESH_DISPLAY,
    TIMING_PROFILEAREA_NUMAREAS,
};

// C O N S T A N T S /////////////////////////////////////////////////////////

//// Winding of output tris: true => CW, false => CCW.
//static const bool USE_CW_WINDING = true;

static const unsigned MAX_MESH_NAME_LENGTH = 256;

// (MH - 19.09.2002) this is partially dependent on texture size, but also on the fact
// that we send texture coordinates to the VU in fixed point 1.3.12 format. For 256x256
// the limit on the GS is (-8,+8). Any 512x512s that might end up here (source 1024x1024)
// will need to have been adjusted to (-4,+4). 
// Note that because of the fixed point limitation we can't use bnigger ranges on textures
// smaller that 256x256.
// Until I figure out how to get the current texture size from the place it is used, 
// we can't actually check that 512s have been adjusted.
static const float PS2_TEXTURE_COORD_LIMIT = 8.0f;    // for 256x256 texture

static const char* ENV_RESOURCE_FILENAME       = "PS2Env ";
static const char* HIER_SKIN_RESOURCE_FILENAME = "PS2Hier";
static const char* OBJECT_RESOURCE_FILENAME    = "PS2Obj ";

// G L O B A L S /////////////////////////////////////////////////////////////

static Asura_Char g_szNameOfCurMesh[MAX_MESH_NAME_LENGTH]  = "";
static Asura_Char g_szNameOfLastMesh[MAX_MESH_NAME_LENGTH] = "";

static float g_fWeldTolerance = 0.0025f * 0.0025f;

// C L A S S E S /////////////////////////////////////////////////////////////

struct PostProcessedStrip
{

    // C L A S S E S /////////////////////////////////////////////////////////

    class LengthCmp
    {
    public:
        bool operator()( const PostProcessedStrip* pxStripA, const PostProcessedStrip* pxStripB ) const
        {
            return( pxStripA->xQuadletIndices.size() < pxStripB->xQuadletIndices.size() );
        }
    };

    class MaterialCmp
    {
    public:
        bool operator()( const PostProcessedStrip* pxStripA, const PostProcessedStrip* pxStripB ) const;
    };

    class EnvMaterialCmp
    {
    public:
        bool operator()( const PostProcessedStrip* pxStripA, const PostProcessedStrip* pxStripB ) const;
        bool operator()( int iMaterialA, int iMaterialB ) const;
    };

    // C O N S T A N T S /////////////////////////////////////////////////////

    enum
    {
        NEEDS_ADC_BIT = 0x80000000,
    };

    // D A T A ///////////////////////////////////////////////////////////////

    int iMaterialID;

    // Actual index is (value & ~NEEDS_ADC_BIT). The index will be OR'd with
    // NEEDS_ADC_BIT if the ADC bit should set when the PS2 file is written.
    std::vector<unsigned> xQuadletIndices;

    unsigned uTriCount;          // This will not necessarily be (vertex count - 2), as you might expect, because of vertices with ADC bits set.

};

class EnvStrip
{

public:

    EnvStrip( void )
    {
        m_pfPositions       = NULL;
        m_pusNrmsAndADCBits = NULL;
        m_piTexCoords       = NULL;
        m_pucRGBs           = NULL;
    }

    ~EnvStrip( void )
    {
        delete[] m_pucRGBs;
        delete[] m_piTexCoords;
        delete[] m_pusNrmsAndADCBits;
        delete[] m_pfPositions;
    }

    // These strips will have a different amount of data if they're dual
    // textured. In the comments below, the first possibility is for singly
    // textured strips while the second is for dual textured ones.

    u_int    m_uVertexCount;
    u_int    m_uTriCount;         // This will not necessarily be (vertex count - 2), as you might expect, because of vertices with ADC bits set.
    float*   m_pfPositions;       // 3 * uVertexCount entries.
    u_short* m_pusNrmsAndADCBits; // 1 * uVertexCount entries.
    short*   m_piTexCoords;       // {2,4} * uVertexCount entries.
    u_char*  m_pucRGBs;           // {3,4} * uVertexCount entries.

};

struct ModuleStripsWithSameMaterial
{
    int                    iModuleIdx;
    std::vector<EnvStrip*> xStrips;
};

struct MaterialIDToModuleStripsMap
{
    int                                        iMaterialID;
    std::vector<ModuleStripsWithSameMaterial*> xModuleData;
};

struct PS2HierSkinVertex
{
    Asura_Vector_3   xPos;
    Asura_Vector_3   xNrm;
    PS2Strip_MatInfo xMat;
    float            fAlpha;
    bool             bNeedsADC;
};

class PS2HierSkinStrip
{
public:

    PS2HierSkinStrip( void )
    {
        m_iMaterialID  = ASURA_MATERIAL_NONE;
        m_uNumVertices = 0;
        m_pxVerts      = NULL;
        m_uTriCount    = 0;
    }

    ~PS2HierSkinStrip( void )
    {
        delete[] m_pxVerts;
    }

    int                m_iMaterialID;
    u_int              m_uNumVertices;
    PS2HierSkinVertex* m_pxVerts;
    u_int              m_uTriCount;
};

struct PS2HierSkinSectionMaterial
{
public:

    PS2HierSkinSectionMaterial( void )
    {
        m_iMaterialID        = ASURA_MATERIAL_NONE;
        m_uNumStrips         = 0;
        m_pxStrips           = NULL;
        m_uTriCount          = 0;
        m_bAnyNonOpaqueVerts = false;
    }

    ~PS2HierSkinSectionMaterial( void )
    {
        delete[] m_pxStrips;
    }

    int               m_iMaterialID;
    u_int             m_uNumStrips;
    PS2HierSkinStrip* m_pxStrips;
    u_int             m_uTriCount;
    bool              m_bAnyNonOpaqueVerts;
};

class PS2HierSkinSection
{
public:

    PS2HierSkinSection( void )
    {
        m_uNumMaterials = 0;
        m_pxMaterials = NULL;
    }

    ~PS2HierSkinSection( void )
    {
        delete[] m_pxMaterials;
    }

    u_int                       m_uNumMaterials;
    PS2HierSkinSectionMaterial* m_pxMaterials;
};

class PS2HierSkin
{
public:

    PS2HierSkin( void )
    {
        m_uNumSections           = 0;
        m_pxSections             = NULL;
        m_uTotalSectionMaterials = 0;
        m_uTotalStrips           = 0;
        m_bAnyNonOpaqueVerts     = false;
    }
    ~PS2HierSkin( void )
    {
        delete[] m_pxSections;
    }

    u_int               m_uNumSections;
    PS2HierSkinSection* m_pxSections;
    u_int               m_uTotalSectionMaterials;
    u_int               m_uTotalStrips;
    bool                m_bAnyNonOpaqueVerts;
};

class PS2Strip_ExporterImplementation
{

    public:

        // M E T H O D S /////////////////////////////////////////////////////

        static void BeginEnvExport( void );
        static void BeginHierSkinExport( const Asura_Hierarchy_Skin* pxSkin, float fStartProgressPercentage, float fEndProgressPercentage );
        static void BeginObjectExport( Asura_Mesh* pxMesh );

        static void Abort( void );
        static PS2Strip_Exporter::UPDATERESULT Update( void );
		
		static void SetSubdivideFlag(bool bSubdivide){s_bSubdivideMesh = bSubdivide;}
		static void SetUVScale(float fUScale,float fVScale){s_fTexUScale = fUScale,s_fTexVScale = fVScale;}

        static float Timing_GetTotalElapsedTime( void );
        static void Timing_PrintReport( void );

		#if defined( ASURA_USE_CARPET )
		static bool OutputCarpetMeshChunk( HANDLE pOutput );
		#endif //#if defined( ASURA_USE_CARPET )

		#if defined( ASURA_USE_BOTANICALS )
		static bool OutputBotanicalsChunk( HANDLE pOutput, const int* const piMaterialConversionList  );
		static bool OutputBotanicalObjectBillboards( void );
        static void BeginBotanicalObjectExport( Asura_Mesh* pxMesh, enum Asura_PS2_Botanicals_Mesh_Type eType );
		#endif //#if defined( ASURA_USE_BOTANICALS )

    private:

        // C L A S S E S /////////////////////////////////////////////////////

//      class Predicate_OrderByTextureID_NVTS;
//      friend Predicate_OrderByTextureID_NVTS;

        // M E T H O D S /////////////////////////////////////////////////////

        static void BeginExportInternal( void );

        static void Timing_SetProfileArea( TIMING_PROFILEAREA eArea );
        static void Timing_Reset( void );

        // Export steps used by all file types.
        static bool GenerateRequiredVertexData( void );
        static bool ProcessVerticesAndEliminateDuplicates( void );

        // Extra steps if using NVTriStrip stripifier.
        static bool NVTS_Stripify( void );
        static bool NVTS_PostProcessStrips( void );

        // Additional export steps for hierarchical character data. Again,
        // each step returns true iff successful. These steps are interspersed
        // with the shared steps.
        static bool AddHierSkinData( void );
        static bool OutputHierSkin( void );

        // Additional export steps for environment data. Again, each step
        // returns true iff successful and these additional ones are
        // interspersed with the shared steps.
        static bool BuildEnvMaterialIDList( void );
        static bool GenerateEnvPrelighting( void );
        static bool AddEnvData( void );
        static bool OutputEnv( void );
        static bool WriteEnvDmaDataForModule( HANDLE hFile,
                                              Tools_PS2_DMAChainWriter& xChainWriter,
                                              u_int uModuleOffset,
                                              int iMaterialID,
                                              const ModuleStripsWithSameMaterial* pxModuleStrips,
                                              u_int* puNumDmaChainPieces,
                                              u_int* puTotalBytesWritten );
        static bool WriteEnvDmaDataForStrip( const EnvStrip* pxStrip, Tools_PS2_DMAChainWriter& xChainWriter, bool bIsDualTexture );
        static bool EnvDmaDataEndChainCallback( const Tools_PS2_DMAChainWriter::EndChainInfo& xEndChainInfo, void** ppvDataToBeWritten, u_int* puNumBytesToBeWritten );

        // Additional export steps for object data.
        static bool OutputObject( void );

		#if defined( ASURA_USE_BOTANICALS )
        static bool OutputBotanicalObject( void );
        static int GetMaterialsFromMesh( const Asura_Mesh* const pxMesh, Tools_DynamicArray<int>& xUsedMaterialIDs );
		#endif //#if defined( ASURA_USE_BOTANICALS )

        // Utility methods for NVTriStrip stripifier.
        static void NVTS_AppendStripperStripToPostProcessedStrip( PostProcessedStrip* pxPPStrip, const NVTriStrip::NvStripInfo* pxInputStripToAppend );
        static bool NVTS_CanAppend( const PostProcessedStrip* pxPPStrip, const NVTriStrip::NvStripInfo* pxToAppend );
        static int NVTS_ExtractMaterialID( const NVTriStrip::NvStripInfo* pxStrip );
//      static bool NVTS_IsNextTriCW( const PostProcessedStrip* pxCurPPStrip );
//      static bool NVTS_IsTriCW( const NVTriStrip::NvFaceInfo* pxTri, int iV0, int iV1 );

        // General utility methods.
        static void CleanUp( bool bDeallocateAccumulatedData );
        static void OutputCumulativeWarnings( void );
        static bool SortPostProcessedStripsOnLength( u_int uStartIdx, u_int uEndIdx );
        static bool SortPostProcessedStripsOnMaterial( u_int uStartIdx, u_int uEndIdx );

        // S T A T E /////////////////////////////////////////////////////////

        static Asura_Mesh* s_pxCurMesh;

        static float s_fStartProgressPercentage;
        static float s_fProgressPercentageDelta;

		static bool s_bSubdivideMesh;
		static Tools_SubdividedMesh* s_pSubdividedMesh;

		static float s_fTexUScale;
		static float s_fTexVScale;

        // Some stats on current mesh.
        static u_int s_uNumUniqueMeshMaterials;

        // Accumulated data over all meshes.
        static u_int s_uTotalDegenerateSourceTris;
        static u_int s_uTotalDuplicateSourceTris;
		static u_int s_uTotalEdgesWithTooManyTriangles;

        // Arrays of distinct values in each category of vertex data.
        // (MP - 29.11.2001) Unique mesh position stuff is not used for
        // characters - we use the position indices from the source mesh.
        static Asura_Vector_3* s_axMeshPositions;
        static u_int s_uNumMeshPositions;
        static Asura_Vector_3* s_axMeshNormals;
        static u_int s_uNumMeshNormals;
        static PS2Strip_MatInfo* s_axMeshMatInfo;
        static u_int s_uNumMeshMatInfos;
        static u_int* s_auMeshColourValues;
        static u_int s_uNumMeshColourValues;

        // These are the unique quadlets of indices into the above four
        // arrays, ie. each distinct vertex has an entry in here.
        static PS2Strip_VtxIdxQuadlet* s_axUniqueMeshVtxIdxQuadlets;
        static u_int s_uNumUniqueMeshVtxIdxQuadlets;
        static u_int s_uNumUnusedUniqueMeshVtxIdxQuadlets;

        // Indices for each tri into the array above. These tris are sorted
        // by material.
        static PS2Strip_Tri* s_axUniqueMeshTris;
        static u_int s_uNumUniqueMeshTris;
        static u_int* s_auUniqueMeshTrisPerMaterial;
        static u_int* s_auNumPostProcessedStripsPerMaterial;

        static NVTriStrip::NvStripInfoVec s_xNVTSStripiferOutput;

        // Post-processed stripifier output. These strips are sorted by
        // material.
        static std::vector<PostProcessedStrip*> s_xPostProcessedStrips;
        static unsigned s_uNumOutputTris;
        static unsigned s_uNumDegenerateOutputTris;

        // Additional data for hier skin exports.
        static const Asura_Hierarchy_Skin* s_pxInputSkin;
        static int s_iCurSectionIdx;
        static PS2HierSkin* s_pxOutputHierSkin;
		// Backup valuse for undoing changes during export
		static int s_iBackupNumberOfBones;
		static Asura_Hierarchy_Skin_Vertex* s_pxBackupVertexData;

        // Additional data for environment files. This is cumulative over the
        // entire environment, not just the individual mesh we're currently
        // working on.
        static int s_iEnvCurModuleIdx;
        static u_int s_uEnvNumUniqueMaterialIDs;
        static int* s_aiEnvUniqueMaterialIDs;
        static MaterialIDToModuleStripsMap* s_axMaterialIDToModuleStripsMaps;
        static unsigned s_uEnvNumTris;
        static unsigned s_uEnvNumVertices;
        static unsigned s_uEnvNumStrips;
        static unsigned s_uEnvNumSingleTriStrips;
        static unsigned s_uEnvNumStripsFromStripifier;
        static unsigned s_uEnvNumSingleTriStripsFromStripifier;
        static unsigned s_uEnvMaxLengthFromStripifier;
        static unsigned s_uEnvNumVerticesWithDodgyTexCoords;

        static LONGLONG s_iStartTime;
        static LONGLONG s_iTimeElapsedInCodeNotBeingTimed;
        static LONGLONG s_iStartTimeForCurCodeNotBeingTimed;

		#if defined( ASURA_USE_BOTANICALS )
		// How many bytes written out by OutputBotanicalObject function
		static u_int s_uNumBytesWrittenBotanicalObject;
		// The type of the botanical mesh being exported
		static enum Asura_PS2_Botanicals_Mesh_Type s_eBotanicalType;
		#endif // #if defined( ASURA_USE_BOTANICALS )
};

// P R O T O T Y P E S   A N D   I N L I N E   F U N C T I O N S /////////////

#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

// (AAM - 08.12.2006) This function is only used in ProcessVerticesAndEliminateDuplicates().
// Remove here too to avoid warning

static bool IsABGRApproximatelyEqual( u_int uABGR1, u_int uABGR2, u_int uTol )
{
    if( uABGR1 != uABGR2 )
    {
        // Compare channels individually.
        for( u_int u = 4; u--; )
        {
            int iChannel1 = uABGR1 & 0xFFu;
            int iChannel2 = uABGR2 & 0xFFu;

            if( Asura_Maths::Abs(iChannel1 - iChannel2) > static_cast<int>(uTol) )
            {
                return false;
            }
        
            uABGR1 >>= 8u;
            uABGR2 >>= 8u;
        }
    }

    return true;
}

#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

static bool IsMaterialDualTexture( int iMaterialID )
{
    return( Tools_Level_Export::ShouldIncludeDualTextureData() && Asura_Material_System::IsMaterialDualTexture(iMaterialID) );
}

static bool Has8BitAlpha( int iMatID )
{
    bool bHas = false;

    Tools_Level_Export::DoesTIM2Have8BitAlpha( Asura_Material_System::GetTextureID(iMatID), &bHas );

    return bHas;
}

static RENDERPASS GetRenderPassFromMaterialID( int iMatID )
{
    RENDERPASS ePass;

    // The prioritisation here reflects that done at runtime on the PS2, in
    // terms of which flags override others.

    const u_int uMaterialFlags = Asura_Material_System::GetMaterialFlags( iMatID );
    if( uMaterialFlags & AsuraMaterialFlag_PS2AdditiveAlpha )
    {
        ePass = RENDERPASS_ADDITIVE_ALPHA;
    }
    else if( uMaterialFlags & AsuraMaterialFlag_Transparent )
    {
        ePass = ( Has8BitAlpha(iMatID) ? RENDERPASS_TRANSLUCENT : RENDERPASS_ON_OFF_TRANSPARENCY );
    }
    else if( Asura_Texture_Management::IsTextureFMV(Asura_Material_System::GetTextureID(iMatID)) )
    {
        ePass = RENDERPASS_FMV;
    }
    else
    {
        ePass = RENDERPASS_OPAQUE;
    }

    return ePass;
}

// C L A S S   I M P L E M E N T A T I O N S /////////////////////////////////

//
// PS2Strip_Exporter
//

float PS2Strip_Exporter::Timing_GetTotalElapsedTime( void )
{
    return PS2Strip_ExporterImplementation::Timing_GetTotalElapsedTime();
}

void PS2Strip_Exporter::Timing_PrintReport( void )
{
    PS2Strip_ExporterImplementation::Timing_PrintReport();
}

void PS2Strip_Exporter::OutputSkinLODThresholdChunk( HANDLE hOut, const Asura_Hierarchy_Skin* pxSkin )
{
    u_int uNumLODs = 0;
    const Asura_Hierarchy_Skin* pxHighestLOD = NULL;

    // Count LODs.
    {
        // Find highest LOD.
        pxHighestLOD = pxSkin;
        while( pxHighestLOD->m_pxNextHigherSkin )
        {
            pxHighestLOD = pxHighestLOD->m_pxNextHigherSkin;
        }

        // Count.
        for( const Asura_Hierarchy_Skin* pxSkinTemp = pxHighestLOD; pxSkinTemp; pxSkinTemp = pxSkinTemp->m_pxNextLowerSkin )
        {
            ++uNumLODs;
        }
    }
	if(uNumLODs > 4)
	{
		PS2Strip::LogError( "ERROR: hierarchy skin \"%s\" has too many LODs (%u, maximum 4).\n",
			pxSkin->Name.GetString(), uNumLODs);
	}

    const Asura_Char* LOD_THRESHOLDS_RESOURCE_FILENAME = "PS2LODThresholds";
    const unsigned uSizeOfResourceName = (strlen(LOD_THRESHOLDS_RESOURCE_FILENAME) + 4) & ~3;
    const unsigned uSizeOfSkinName     = (pxSkin->Name.GetSizeInInts() << 2);

    u_int uTotalNumBytesWritten = 0;
    #define WRITE( data_ptr, num_bytes ) \
    { \
        /* Store the ptr value and byte count so we don't evaluate them \
           multiple times. */ \
        const void* const pvData = (data_ptr); \
        const unsigned uNumBytes = static_cast<unsigned>( (num_bytes) ); \
        \
        Asura_Assert( pvData, "" ); \
        Asura_Assert( uNumBytes > 0, "" ); \
         \
        DWORD uNumBytesWritten; \
        BOOL bRet = WriteFile( hOut, pvData, uNumBytes, &uNumBytesWritten, NULL ); \
        if( !bRet || uNumBytesWritten != uNumBytes ) \
        { \
            /* (MP - 19.09.2002) Get sys error before closing file because \
               that may fail as well. */ \
            PS2Strip::LogError( "ERROR: Failed to write all bytes to file. Reason: %s\n", PS2Strip::GetLastSystemError() ); \
            return; \
        } \
         \
        uTotalNumBytesWritten += uNumBytes; \
    }

    const u_int uExpectedDataSize  = uSizeOfSkinName + (uNumLODs << 2);
    const u_int uExpectedTotalSize = sizeof(Asura_Chunk_ResourceFile_PS2_SkinLODThresholds) + uSizeOfResourceName + uExpectedDataSize;

    // Output chunk header.
    Asura_Chunk_ResourceFile_PS2_SkinLODThresholds xChunkHdr;
    xChunkHdr.ID       = ASURA_CHUNK_RESOURCEFILE;
    xChunkHdr.Size     = uExpectedTotalSize;
    xChunkHdr.Version  = 0;
    xChunkHdr.Flags    = 0;
    xChunkHdr.Type     = ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC;
    xChunkHdr.SubType  = ASURA_RESOURCEFILE_TYPE_PS2_SKINLODTHRESHOLDS;
    xChunkHdr.FileSize = uExpectedDataSize;
    WRITE( &xChunkHdr, sizeof(Asura_Chunk_ResourceFile_PS2_SkinLODThresholds) );

    // Output resource filename.
    WRITE( LOD_THRESHOLDS_RESOURCE_FILENAME, uSizeOfResourceName );

#ifdef ASURA_ASSERT
    const u_int uNumBytesWrittenBeforeData = uTotalNumBytesWritten;
#endif // #ifdef ASURA_ASSERT

    // Output skin name.
    WRITE( pxSkin->Name.GetString(), uSizeOfSkinName );

    // Output thresholds.
    for( const Asura_Hierarchy_Skin* pxSkinTemp = pxHighestLOD; pxSkinTemp; pxSkinTemp = pxSkinTemp->m_pxNextLowerSkin )
    {
        const float fMinDistSqd = pxSkinTemp->m_fMinDistance * pxSkinTemp->m_fMinDistance;
        WRITE( &fMinDistSqd, 4 );
    }

    Asura_Assert( uExpectedTotalSize == uTotalNumBytesWritten, "" );
    Asura_Assert( uExpectedDataSize == uTotalNumBytesWritten - uNumBytesWrittenBeforeData, "" );

    #undef WRITE
}

#if defined( ASURA_USE_CARPET )
bool PS2Strip_Exporter::OutputCarpetMeshChunk( HANDLE pOutput )
{
	bool bReturn = PS2Strip_ExporterImplementation::OutputCarpetMeshChunk( pOutput );
	return bReturn;
}
#endif // #if defined( ASURA_USE_CARPET )

#if defined( ASURA_USE_BOTANICALS )

void PS2Strip_Exporter::BeginBotanicalObjectExport( Asura_Mesh* pxMesh, enum Asura_PS2_Botanicals_Mesh_Type eType )
{
    PS2Strip_ExporterImplementation::BeginBotanicalObjectExport( pxMesh, eType );
}

bool PS2Strip_Exporter::OutputBotanicalsChunk( HANDLE pOutput, const int* const piMaterialConversionList  )
{
	bool bReturn = PS2Strip_ExporterImplementation::OutputBotanicalsChunk( pOutput, piMaterialConversionList );
	return bReturn;
}

bool PS2Strip_Exporter::OutputBotanicalObjectBillboards()
{
	bool bReturn = PS2Strip_ExporterImplementation::OutputBotanicalObjectBillboards();
	return bReturn;
}

#endif // #if defined( ASURA_USE_BOTANICALS )


//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     PS2Strip_Exporter::Abort().
// PARAMETERS   None.
// RETURNS      None.
// DESCRIPTION  Aborts the export process.
// AUTHOR       Morgan Parry - 01.05.2002.
//
//////////////////////////////////////////////////////////////////////////////
void PS2Strip_Exporter::Abort( void )
{
    PS2Strip_ExporterImplementation::Abort();
}
// End function PS2Strip_Exporter::Abort().

void PS2Strip_Exporter::BeginEnvExport( void )
{
    PS2Strip_ExporterImplementation::BeginEnvExport();
}

void PS2Strip_Exporter::BeginHierSkinExport( const Asura_Hierarchy_Skin* pxSkin, float fStartProgressPercentage, float fEndProgressPercentage )
{
    PS2Strip_ExporterImplementation::BeginHierSkinExport( pxSkin, fStartProgressPercentage, fEndProgressPercentage );
}

void PS2Strip_Exporter::BeginObjectExport( Asura_Mesh* pxMesh )
{
    PS2Strip_ExporterImplementation::BeginObjectExport( pxMesh );
}

const char* PS2Strip_Exporter::GetNameOfCurMesh( void )
{
    return g_szNameOfCurMesh;
}

const char* PS2Strip_Exporter::GetNameOfLastMesh( void )
{
    return g_szNameOfLastMesh;
}

PS2Strip_Exporter::UPDATERESULT  PS2Strip_Exporter::Update( void )
{
    return PS2Strip_ExporterImplementation::Update();
}

void PS2Strip_Exporter::SetSubdivideFlag(bool bSubdivide)
{
	PS2Strip_ExporterImplementation::SetSubdivideFlag(bSubdivide);
}

void PS2Strip_Exporter::SetUVScale(float fUScale,float fVScale)
{
	PS2Strip_ExporterImplementation::SetUVScale(fUScale,fVScale);
}

void PS2Strip_Exporter::SetWeldTolerance(float fTolerance)
{
	g_fWeldTolerance = fTolerance*fTolerance;
}


//
// PostProcessedStrip::MaterialCmp.
//

bool PostProcessedStrip::MaterialCmp::operator()( const PostProcessedStrip* pxStripA, const PostProcessedStrip* pxStripB ) const
{
    const int iMatA = pxStripA->iMaterialID;
    const int iMatB = pxStripB->iMaterialID;

    if( iMatA == iMatB ) { return false; }

    // First compare by render pass.
    {
        const RENDERPASS ePassA = GetRenderPassFromMaterialID( iMatA );
        const RENDERPASS ePassB = GetRenderPassFromMaterialID( iMatB );

        if( ePassA != ePassB )
        {
            return( ePassA < ePassB );
        }
    }

    // Next by base pass texture.
    {
        const int iBaseTexA = Asura_Material_System::GetTextureID( iMatA );
        const int iBaseTexB = Asura_Material_System::GetTextureID( iMatB );

        if( iBaseTexA != iBaseTexB )
        {
            return( iBaseTexA < iBaseTexB );
        }
    }
    
    // Next by material flags.
    {
        const u_int uFlagsA = Asura_Material_System::GetMaterialFlags( iMatA );
        const u_int uFlagsB = Asura_Material_System::GetMaterialFlags( iMatB );

        if( uFlagsA != uFlagsB )
        {
            return( uFlagsA < uFlagsB );
        }
    }

    // Finally by material ID.
    return( iMatA < iMatB );
}

//
// PostProcessedStrip::EnvMaterialCmp.
//

// Essentially the same as PostProcessedStrip::MaterialCmp above except we
// take account of any dual pass stuff setup.
bool PostProcessedStrip::EnvMaterialCmp::operator()( int iMatA, int iMatB ) const
{
    if( iMatA == iMatB ) { return false; }

    // First compare by render pass.
    {
        const RENDERPASS ePassA = GetRenderPassFromMaterialID( iMatA );
        const RENDERPASS ePassB = GetRenderPassFromMaterialID( iMatB );

        if( ePassA != ePassB )
        {
            return( ePassA < ePassB );
        }
    }

    // Next by base pass texture.
    {
        const int iBaseTexA = Asura_Material_System::GetTextureID( iMatA, 0 );
        const int iBaseTexB = Asura_Material_System::GetTextureID( iMatB, 0 );

        if( iBaseTexA != iBaseTexB )
        {
            return( iBaseTexA < iBaseTexB );
        }
    }

    // Next by second pass texture, if applicable.
    {
        const int iSecondTexA = Asura_Material_System::GetTextureID( iMatA, 1 );
        const int iSecondTexB = Asura_Material_System::GetTextureID( iMatB, 1 );

        if( iSecondTexA != iSecondTexB )
        {
            return( iSecondTexA < iSecondTexB );
        }
    }

    // Next by material flags.
    {
        const u_int uFlagsA = Asura_Material_System::GetMaterialFlags( iMatA );
        const u_int uFlagsB = Asura_Material_System::GetMaterialFlags( iMatB );

        if( uFlagsA != uFlagsB )
        {
            return( uFlagsA < uFlagsB );
        }
    }

    // Finally by material ID.
    return( iMatA < iMatB );
}

bool PostProcessedStrip::EnvMaterialCmp::operator()( const PostProcessedStrip* pxStripA, const PostProcessedStrip* pxStripB ) const
{
    const int iMatA = pxStripA->iMaterialID;
    const int iMatB = pxStripB->iMaterialID;

    return (*this)( iMatA, iMatB );
}

//
// PS2Strip_ExporterImplementation.
//

Asura_Mesh* PS2Strip_ExporterImplementation::s_pxCurMesh;
float PS2Strip_ExporterImplementation::s_fStartProgressPercentage;
float PS2Strip_ExporterImplementation::s_fProgressPercentageDelta;

bool PS2Strip_ExporterImplementation::s_bSubdivideMesh = false;
Tools_SubdividedMesh* PS2Strip_ExporterImplementation::s_pSubdividedMesh = 0;

u_int PS2Strip_ExporterImplementation::s_uNumUniqueMeshMaterials;

u_int PS2Strip_ExporterImplementation::s_uTotalDegenerateSourceTris;
u_int PS2Strip_ExporterImplementation::s_uTotalDuplicateSourceTris;
u_int PS2Strip_ExporterImplementation::s_uTotalEdgesWithTooManyTriangles;

Asura_Vector_3* PS2Strip_ExporterImplementation::s_axMeshPositions = NULL;
u_int PS2Strip_ExporterImplementation::s_uNumMeshPositions = 0;
Asura_Vector_3* PS2Strip_ExporterImplementation::s_axMeshNormals = NULL;
u_int PS2Strip_ExporterImplementation::s_uNumMeshNormals = 0;
PS2Strip_MatInfo* PS2Strip_ExporterImplementation::s_axMeshMatInfo = NULL;
u_int PS2Strip_ExporterImplementation::s_uNumMeshMatInfos = 0;
u_int* PS2Strip_ExporterImplementation::s_auMeshColourValues = NULL;
u_int PS2Strip_ExporterImplementation::s_uNumMeshColourValues = 0;

PS2Strip_VtxIdxQuadlet* PS2Strip_ExporterImplementation::s_axUniqueMeshVtxIdxQuadlets;
u_int PS2Strip_ExporterImplementation::s_uNumUniqueMeshVtxIdxQuadlets;
u_int PS2Strip_ExporterImplementation::s_uNumUnusedUniqueMeshVtxIdxQuadlets;

PS2Strip_Tri* PS2Strip_ExporterImplementation::s_axUniqueMeshTris;
unsigned PS2Strip_ExporterImplementation::s_uNumUniqueMeshTris;
u_int* PS2Strip_ExporterImplementation::s_auUniqueMeshTrisPerMaterial;
u_int* PS2Strip_ExporterImplementation::s_auNumPostProcessedStripsPerMaterial;

NVTriStrip::NvStripInfoVec PS2Strip_ExporterImplementation::s_xNVTSStripiferOutput;

std::vector<PostProcessedStrip*> PS2Strip_ExporterImplementation::s_xPostProcessedStrips;
unsigned PS2Strip_ExporterImplementation::s_uNumOutputTris;
unsigned PS2Strip_ExporterImplementation::s_uNumDegenerateOutputTris;

int PS2Strip_ExporterImplementation::s_iCurSectionIdx = -1;
const Asura_Hierarchy_Skin* PS2Strip_ExporterImplementation::s_pxInputSkin = NULL;
PS2HierSkin* PS2Strip_ExporterImplementation::s_pxOutputHierSkin;

int PS2Strip_ExporterImplementation::s_iBackupNumberOfBones = 0;
Asura_Hierarchy_Skin_Vertex* PS2Strip_ExporterImplementation::s_pxBackupVertexData = 0;

int PS2Strip_ExporterImplementation::s_iEnvCurModuleIdx = -1;
u_int PS2Strip_ExporterImplementation::s_uEnvNumUniqueMaterialIDs;
int* PS2Strip_ExporterImplementation::s_aiEnvUniqueMaterialIDs;
MaterialIDToModuleStripsMap* PS2Strip_ExporterImplementation::s_axMaterialIDToModuleStripsMaps;
unsigned PS2Strip_ExporterImplementation::s_uEnvNumTris;
unsigned PS2Strip_ExporterImplementation::s_uEnvNumVertices;
unsigned PS2Strip_ExporterImplementation::s_uEnvNumStrips;
unsigned PS2Strip_ExporterImplementation::s_uEnvNumSingleTriStrips;
unsigned PS2Strip_ExporterImplementation::s_uEnvNumStripsFromStripifier;
unsigned PS2Strip_ExporterImplementation::s_uEnvNumSingleTriStripsFromStripifier;
unsigned PS2Strip_ExporterImplementation::s_uEnvMaxLengthFromStripifier;

float PS2Strip_ExporterImplementation::s_fTexUScale = 1.0f;
float PS2Strip_ExporterImplementation::s_fTexVScale = 1.0f;

unsigned PS2Strip_ExporterImplementation::s_uEnvNumVerticesWithDodgyTexCoords;

LONGLONG PS2Strip_ExporterImplementation::s_iStartTime;
LONGLONG PS2Strip_ExporterImplementation::s_iTimeElapsedInCodeNotBeingTimed;
LONGLONG PS2Strip_ExporterImplementation::s_iStartTimeForCurCodeNotBeingTimed;

#if defined( ASURA_USE_BOTANICALS )
// How many bytes written out by OutputBotanicalObject function
u_int PS2Strip_ExporterImplementation::s_uNumBytesWrittenBotanicalObject;
enum Asura_PS2_Botanicals_Mesh_Type PS2Strip_ExporterImplementation::s_eBotanicalType;
#endif // #if defined( ASURA_USE_BOTANICALS )

void PS2Strip_ExporterImplementation::Abort( void )
{
    CleanUp( true );
}

bool PS2Strip_ExporterImplementation::AddEnvData( void )
{
    Asura_Assert( s_aiEnvUniqueMaterialIDs != NULL, "" );
    Asura_Assert( s_uEnvNumUniqueMaterialIDs > 0, "" );

    // If this is the first mesh in the file, we need to allocate our map
    // array now.
    if( s_axMaterialIDToModuleStripsMaps == NULL )
    {
// (MP - 05.06.2003) This is no longer necessarily true. The first module may
// not have any data, eg. if it's entirely flagged with "don't render",
// meaning we won't call this method for it.
//        Asura_Assert( s_iEnvCurModuleIdx == 0, "" );

        s_axMaterialIDToModuleStripsMaps = new MaterialIDToModuleStripsMap[s_uEnvNumUniqueMaterialIDs];
        if( s_axMaterialIDToModuleStripsMaps == NULL )
        {
            PS2Strip::LogError( "ERROR: Out of memory!\n" );
            return false;
        }

        // Initialise each mapping's texture ID.
        for( unsigned u = 0; u < s_uEnvNumUniqueMaterialIDs; ++u )
        {
            s_axMaterialIDToModuleStripsMaps[u].iMaterialID = s_aiEnvUniqueMaterialIDs[u];
        }
    }

    const u_int uNumStrips = s_xPostProcessedStrips.size();

    MaterialIDToModuleStripsMap*  pxCurMap       = NULL;
    ModuleStripsWithSameMaterial* pxCurStripList = NULL;

    // Iterate through the mesh's strips and store data.
    for( u_int uStrip = 0; uStrip < uNumStrips; ++uStrip )
    {
        const PostProcessedStrip* pxPPStrip = s_xPostProcessedStrips[uStrip];

        const bool bIsDualTexture = IsMaterialDualTexture( pxPPStrip->iMaterialID );

        // Allocate an environment strip header.
        EnvStrip* pxEnvStrip = new EnvStrip();
        if( !pxEnvStrip )
        {
            PS2Strip::LogError( "ERROR: Out of memory!\n" );
            return false;
        }

        // Copy counts.
        pxEnvStrip->m_uVertexCount = static_cast<unsigned short>( pxPPStrip->xQuadletIndices.size() );
        pxEnvStrip->m_uTriCount    = pxPPStrip->uTriCount;

        // Update global stats.
        {
            s_uEnvNumVertices += pxEnvStrip->m_uVertexCount;
            s_uEnvNumTris     += pxEnvStrip->m_uTriCount;

            if( 1 == pxEnvStrip->m_uTriCount )
            {
                ++s_uEnvNumSingleTriStrips;
            }
        }

        const u_int uTexCoordsPerVtx        = ( bIsDualTexture ? 4 : 2 );
        const u_int uColourComponentsPerVtx = ( bIsDualTexture ? 4 : 3 );

        // Allocate storage for the strip data.
        {
            pxEnvStrip->m_pfPositions       = new float[3 * pxEnvStrip->m_uVertexCount];
            pxEnvStrip->m_pusNrmsAndADCBits = new u_short[1 * pxEnvStrip->m_uVertexCount];
            pxEnvStrip->m_piTexCoords       = new short[uTexCoordsPerVtx * pxEnvStrip->m_uVertexCount];
            pxEnvStrip->m_pucRGBs           = new u_char[uColourComponentsPerVtx * pxEnvStrip->m_uVertexCount];

            if( !pxEnvStrip->m_pfPositions || !pxEnvStrip->m_pusNrmsAndADCBits || !pxEnvStrip->m_piTexCoords || !pxEnvStrip->m_pucRGBs )
            {
                PS2Strip::LogError( "ERROR: Out of memory!\n" );
                delete pxEnvStrip;
                return false;
            }
        }

        // Get module's position so we can translate the vertex positions
        // below.
        Asura_Vector_3 xTempVec;
        Asura_Module_System::GetPositionOfModule( s_iEnvCurModuleIdx, xTempVec );
        const Asura_Vector_3 xModulePos( xTempVec );

        // Iterate through vertices and add their data into the env strip.
        for( u_int uVertex = 0; uVertex < pxEnvStrip->m_uVertexCount; ++uVertex )
        {
            u_int uQuadletIdx = pxPPStrip->xQuadletIndices[uVertex];
            bool bNeedsADC = ( (uQuadletIdx & PostProcessedStrip::NEEDS_ADC_BIT) != 0 );
            uQuadletIdx &= ~PostProcessedStrip::NEEDS_ADC_BIT;

            Asura_Assert( uQuadletIdx < s_uNumUniqueMeshVtxIdxQuadlets, "" );
            const PS2Strip_VtxIdxQuadlet& xQuadlet = s_axUniqueMeshVtxIdxQuadlets[uQuadletIdx];

            // Fetch individual vertex components.
            const Asura_Vector_3&   xPos  = s_axMeshPositions[xQuadlet.GetPosIdx()];
            const Asura_Vector_3&   xNrm  = s_axMeshNormals[xQuadlet.GetNrmIdx()];
            const PS2Strip_MatInfo& xMat  = s_axMeshMatInfo[xQuadlet.GetMatInfoIdx()];
            const u_int             uABGR = s_auMeshColourValues[xQuadlet.GetColourIdx()];

            Asura_Assert( xMat.m_iMaterialID == pxPPStrip->iMaterialID, "" );

            // Copy and convert as necessary.
            {
                // (MP - 02.01.2002) We now export environment mesh positions
                // in world space rather than object space.
                xTempVec = xPos;
				if(!s_pSubdividedMesh)//The positions get transformed into worldspace when the subdivision takes place
				{
					xTempVec += xModulePos;
				}
                pxEnvStrip->m_pfPositions[3 * uVertex + 0] = xTempVec.x;
                pxEnvStrip->m_pfPositions[3 * uVertex + 1] = xTempVec.y;
                pxEnvStrip->m_pfPositions[3 * uVertex + 2] = xTempVec.z;

                // (MP - 09.01.2003) NB: Due to the way V4_5 is unpacked,
                // these normals will be biased by +16 and then scaled by 8.
                // We compensate for this at runtime.
                {
                    PS2Strip_VIF1_V4_5_Data xNrmAndADC;

                    xNrmAndADC.usX = Tools_PS2Strip_ConvertFloatForV45( xNrm.x );
                    xNrmAndADC.usY = Tools_PS2Strip_ConvertFloatForV45( xNrm.y );
                    xNrmAndADC.usZ = Tools_PS2Strip_ConvertFloatForV45( xNrm.z );
                    xNrmAndADC.usW = bNeedsADC;
                    
                    pxEnvStrip->m_pusNrmsAndADCBits[uVertex] = xNrmAndADC.usValue;
                }

                pxEnvStrip->m_piTexCoords[uTexCoordsPerVtx * uVertex + 0] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMat.m_afU[0] );
                pxEnvStrip->m_piTexCoords[uTexCoordsPerVtx * uVertex + 1] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMat.m_afV[0] );

                pxEnvStrip->m_pucRGBs[uColourComponentsPerVtx * uVertex + 0] = static_cast<u_char>( (uABGR & 0x000000FFu) >>  0u );
                pxEnvStrip->m_pucRGBs[uColourComponentsPerVtx * uVertex + 1] = static_cast<u_char>( (uABGR & 0x0000FF00u) >>  8u );
                pxEnvStrip->m_pucRGBs[uColourComponentsPerVtx * uVertex + 2] = static_cast<u_char>( (uABGR & 0x00FF0000u) >> 16u );

                if( bIsDualTexture )
                {
                    pxEnvStrip->m_piTexCoords[uTexCoordsPerVtx * uVertex + 2] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMat.m_afU[1] );
                    pxEnvStrip->m_piTexCoords[uTexCoordsPerVtx * uVertex + 3] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMat.m_afV[1] );

                    pxEnvStrip->m_pucRGBs[uColourComponentsPerVtx * uVertex + 3] = static_cast<u_char>( (uABGR & 0xFF000000u) >> 24u );
                }
            }

            // Update texture coordinate statistics. Ignore ridiculous
            // coordinates, of which we seem to have a few for some reason.
            {
                bool bDodgyTexCoords = false;

                for( u_int j = 0; j < (bIsDualTexture ? 2u : 1u) && !bDodgyTexCoords; ++j )
                {
                    for( u_int k = 0; k < 2; ++k )
                    {
                        if( Asura_Maths::Abs(xMat.m_afUVs[k][j]) >= PS2_TEXTURE_COORD_LIMIT )
                        {
                            bDodgyTexCoords = true;
                            break;
                        }
                    }
                }

                if( bDodgyTexCoords )
                {
                    ++s_uEnvNumVerticesWithDodgyTexCoords;
                }
            }
        }
        // End loop through strip's vertices.

        // Update current map and strip list ptrs, if necessary.
        if( !pxCurMap || pxCurMap->iMaterialID != pxPPStrip->iMaterialID )
        {
            int iMapIndex = -1;
            for( u_int u = 0; u < s_uEnvNumUniqueMaterialIDs; ++u )
            {
                if( s_aiEnvUniqueMaterialIDs[u] == pxPPStrip->iMaterialID )
                {
                    iMapIndex = u;
                    break;
                }
            }

            Asura_Assert( iMapIndex != -1, "" );
            pxCurMap = &s_axMaterialIDToModuleStripsMaps[iMapIndex];

            // Now add a strip list for the strips in this module using this
            // material. This won't already exist since we're processing the
            // incoming strips in order of material, and obviously we won't
            // previously have encountered anything to do with this module
            // anywhere else.
            {
#ifdef ASURA_ASSERT
                for( u_int u = 0; u < pxCurMap->xModuleData.size(); ++u )
                {
                    Asura_Assert( pxCurMap->xModuleData[u]->iModuleIdx != s_iEnvCurModuleIdx, "" );
                }
#endif

                pxCurStripList = new ModuleStripsWithSameMaterial;
                if( !pxCurStripList )
                {
                    delete pxEnvStrip;
                    return false;
                }

                pxCurStripList->iModuleIdx = s_iEnvCurModuleIdx;

                pxCurMap->xModuleData.push_back( pxCurStripList );
            }
        }

        Asura_Assert( pxCurStripList && pxCurStripList->iModuleIdx == s_iEnvCurModuleIdx, "" );

        pxCurStripList->xStrips.push_back( pxEnvStrip );
    }
    // End loop through strips.

    s_uEnvNumStrips += uNumStrips;

    return true;
}

bool PS2Strip_ExporterImplementation::AddHierSkinData( void )
{
    Asura_Assert( PS2Strip::EXPORTTYPE_SKIN_HIER == PS2Strip::GetCurExportType(), "" );
    Asura_Assert( s_pxInputSkin, "" );
    Asura_Assert( !s_pxInputSkin->IsSkinnedModel(), "" );
    Asura_Assert( s_pxOutputHierSkin, "" );

    const u_int uNumPPStrips = s_xPostProcessedStrips.size();
    Asura_Assert( uNumPPStrips, "" );

    PS2HierSkinSection& xSection = s_pxOutputHierSkin->m_pxSections[s_iCurSectionIdx];
    
    xSection.m_uNumMaterials = s_uNumUniqueMeshMaterials;
    xSection.m_pxMaterials   = new PS2HierSkinSectionMaterial[s_uNumUniqueMeshMaterials];

    // Iterate through the processed strips and copy data.
    {
        int iCurMaterialID = INT_MIN;
        u_int uNextMaterialIdx = 0;
        PS2HierSkinSectionMaterial* pxCurSectionMat = NULL;
        u_int uStripIdxWithinCurSectionMat = 0;
        
        for( u_int uPPStrip = 0; uPPStrip < uNumPPStrips; ++uPPStrip )
        {
            const PostProcessedStrip* const pxPPStrip = s_xPostProcessedStrips[uPPStrip];

#if _DEBUG
            // Verify the materials are in ascending order (as dictated by
            // our comparison class.)
            if( uPPStrip + 1 < uNumPPStrips )
            {               
                const PostProcessedStrip* const pxNextPPStrip = s_xPostProcessedStrips[uPPStrip + 1];

                PostProcessedStrip::MaterialCmp xCmp;
                Asura_Assert( xCmp(pxPPStrip, pxNextPPStrip) || !xCmp(pxNextPPStrip, pxPPStrip), "" ); // ie. (this < next) || !(next > this)
            }
#endif

            if( pxPPStrip->iMaterialID != iCurMaterialID )
            {
                if( pxCurSectionMat )
                {
                    // Verify that we completely filled the section material
                    // we're just finishing.
                    Asura_Assert( uStripIdxWithinCurSectionMat == pxCurSectionMat->m_uNumStrips, "" );
                }

                // Move onto next section material.
                pxCurSectionMat = &xSection.m_pxMaterials[uNextMaterialIdx++];
                uStripIdxWithinCurSectionMat = 0;

                iCurMaterialID = pxPPStrip->iMaterialID;

                // Initialise this next section material.
                {
                    int iMaterialIdx = Asura_Material_System::TranslateMaterialIDToArrayIdx( iCurMaterialID );
                    
                    pxCurSectionMat->m_iMaterialID = iCurMaterialID;
                    pxCurSectionMat->m_uNumStrips  = s_auNumPostProcessedStripsPerMaterial[iMaterialIdx + 1];
                    pxCurSectionMat->m_pxStrips    = new PS2HierSkinStrip[pxCurSectionMat->m_uNumStrips];
                }
            }
        
            const u_int uNumVertices = pxPPStrip->xQuadletIndices.size();

            PS2HierSkinStrip& xHierSkinStrip = pxCurSectionMat->m_pxStrips[uStripIdxWithinCurSectionMat++];
            
            xHierSkinStrip.m_iMaterialID  = iCurMaterialID;
            xHierSkinStrip.m_uNumVertices = uNumVertices;
            xHierSkinStrip.m_pxVerts      = new PS2HierSkinVertex[uNumVertices];
            xHierSkinStrip.m_uTriCount    = pxPPStrip->uTriCount;

            // Copy individual vertices.
            for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
            {
                PS2HierSkinVertex& xOutVtx = xHierSkinStrip.m_pxVerts[uVtx];

                u_int uQuadletIdx = pxPPStrip->xQuadletIndices[uVtx];
                bool bNeedsADC = ( (uQuadletIdx & PostProcessedStrip::NEEDS_ADC_BIT) != 0 );
                uQuadletIdx &= ~PostProcessedStrip::NEEDS_ADC_BIT;

                Asura_Assert( uQuadletIdx < s_uNumUniqueMeshVtxIdxQuadlets, "" );
                const PS2Strip_VtxIdxQuadlet& xQuadlet = s_axUniqueMeshVtxIdxQuadlets[uQuadletIdx];

                xOutVtx.xPos      = s_axMeshPositions[xQuadlet.GetPosIdx()];
                xOutVtx.xNrm      = s_axMeshNormals[xQuadlet.GetNrmIdx()];
                xOutVtx.xMat      = s_axMeshMatInfo[xQuadlet.GetMatInfoIdx()];
                xOutVtx.bNeedsADC = bNeedsADC;
                xOutVtx.fAlpha    = (s_auMeshColourValues[xQuadlet.GetColourIdx()] >> 24) / 128.0f;

                if( xOutVtx.fAlpha < 1.0f )
                {
                    pxCurSectionMat->m_bAnyNonOpaqueVerts = true;
                    s_pxOutputHierSkin->m_bAnyNonOpaqueVerts = true;
                }
            }

            pxCurSectionMat->m_uTriCount += xHierSkinStrip.m_uTriCount;
        }

        Asura_Assert( uNextMaterialIdx == s_uNumUniqueMeshMaterials, "" );
    }

    // Update some stats.
    s_pxOutputHierSkin->m_uTotalSectionMaterials += s_uNumUniqueMeshMaterials;
    s_pxOutputHierSkin->m_uTotalStrips           += uNumPPStrips;

    // Done.
    return true;
}

bool PS2Strip_ExporterImplementation::BuildEnvMaterialIDList( void )
{
    Asura_Assert( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_ENV, "" );

    // Output progress message.
    PS2Strip::LogInfo( "Building material ID list for environment... " );

    bool abUsedMaterials[1 + ASURA_MAX_NO_OF_MATERIALS];
    memset( abUsedMaterials, 0, (1 + ASURA_MAX_NO_OF_MATERIALS) * sizeof(bool) );

    // Iterate through modules.
    const int iNumModules = Asura_Module_System::GetNumberOfModules();
    Asura_Assert( iNumModules > 0, "" );
    for( int iModule = 0; iModule < iNumModules; ++iModule )
    {
        const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule( iModule );

        // Iterate through polys.
        const int iNumPolys = pxMesh->NumberOfPolygons;
        Asura_Assert( iNumPolys > 0, "" );
        Asura_Assert( pxMesh->TextureDataValid, "" );
        for( int iPoly = 0; iPoly < iNumPolys; ++iPoly )
        {
            // Check this polygon should be exported.
            if( pxMesh->PolygonFlagDataValid && (pxMesh->PolygonFlagDataPtr[iPoly].Flags & (AsuraPolygonFlag_DontExport| AsuraPolygonFlag_DontRender| AsuraPolygonFlag_Water)) )
            {
                continue;
            }

            const int iNextMaterialArrayIdx = Asura_Material_System::TranslateMaterialIDToArrayIdx( pxMesh->TextureDataPtr[iPoly].m_iMaterialID );
            Asura_Assert( ASURA_MATERIAL_NONE <= iNextMaterialArrayIdx && iNextMaterialArrayIdx < ASURA_MAX_NO_OF_MATERIALS, "" );
            abUsedMaterials[iNextMaterialArrayIdx + 1] = true;
        }
        // End loop through polys.
    }
    // End loop through modules.

    // Count.
    s_uEnvNumUniqueMaterialIDs = 0;
    for( u_int u = 0; u < 1 + ASURA_MAX_NO_OF_MATERIALS; ++u )
    {
        if( abUsedMaterials[u] ) { ++s_uEnvNumUniqueMaterialIDs; }
    }

    s_aiEnvUniqueMaterialIDs = new int[s_uEnvNumUniqueMaterialIDs];
    if( !s_aiEnvUniqueMaterialIDs ) { return false; }

    int* piOut = &s_aiEnvUniqueMaterialIDs[0];
    for( int i = 0; i < 1 + ASURA_MAX_NO_OF_MATERIALS; ++i )
    {
        if( abUsedMaterials[i] ) { *piOut++ = Asura_Material_System::TranslateArrayIdxToMaterialID( i - 1 ); }
    }

    // Sort materials.
    std::sort( &s_aiEnvUniqueMaterialIDs[0], &s_aiEnvUniqueMaterialIDs[s_uEnvNumUniqueMaterialIDs], PostProcessedStrip::EnvMaterialCmp() );

	//Check the used materials to make sure that...
	//1)No material is both dual-texture and reflective
	//2)There are at most 2 types out of reflective,rainsplash , dual-texture and 512 texture
	//RAR 11/1/2004
	int iBadMaterialCount = 0;
	bool bHasReflective = false;
	bool bHasRainsplash = false;
	bool bHasDualTexture = false;
	bool bHas512 = false;

	for(u_int u=0;u<s_uEnvNumUniqueMaterialIDs;u++)
	{
		//check for reflective + dual-texture combination
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(s_aiEnvUniqueMaterialIDs[u]);
		if(!pxMaterial) continue;
		if(pxMaterial->m_iTextureID[1] != ASURA_TEXTURE_NONE && (pxMaterial->m_uMaterialFlags & AsuraMaterialFlag_Reflective))
		{
			iBadMaterialCount++;
		}

		//check reflective,rainsplash,dual-texture and 512 texture
		if(pxMaterial->m_uMaterialFlags & AsuraMaterialFlag_Reflective)
		{
			bHasReflective = true;
		}
		if(pxMaterial->m_uMaterialFlags & AsuraMaterialFlag_RainSplashes)
		{
			bHasRainsplash = true;
		}
		if(pxMaterial->m_iTextureID[1] != ASURA_TEXTURE_NONE)
		{
			bHasDualTexture = true;
		}
		for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
		{
			int iTextureID = pxMaterial->m_iTextureID[i];
			if(iTextureID == ASURA_TEXTURE_NONE) continue;
			if(!(Asura_Texture_Management::GetTextureFlags(iTextureID) & AsuraTextureFlag_PS2Allow512)) continue;

			//texture has flag to be 512... is it in fact this size?
			u_int uWidth,uHeight;
			if(Tools_Level_Export::GetTIM2Proportions(iTextureID,&uWidth,&uHeight,false))
			{
				if(uWidth>256 || uHeight>256)
				{
					bHas512 = true;
				}
			}

		}
	}

	if( iBadMaterialCount )
	{
		PS2Strip::LogError( "Found %d environment materials that were both dual-texture and reflective:\n",iBadMaterialCount );

        u_int uNumListed = 0;
	    for( u_int u = 0; u < s_uEnvNumUniqueMaterialIDs; ++u )
	    {
		    const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial( s_aiEnvUniqueMaterialIDs[u] );
		    if( pxMaterial && pxMaterial->m_iTextureID[1] != ASURA_TEXTURE_NONE && (pxMaterial->m_uMaterialFlags & AsuraMaterialFlag_Reflective) )
		    {
                if( uNumListed >= 2 )
                {
                    // Just indicate that there were more. (Don't want to
                    // lose the header error message above due to listing too
                    // many.)
                    PS2Strip::LogError( "...\n" );
                    break;
                }

                const char* szTexName;
                const Asura_String* pxName = Asura_Texture_Management::GetTextureName( pxMaterial->m_iTextureID[0] );
                if( pxName )
                {
                    szTexName = pxName->GetString();
                }
                else
                {
                    szTexName = "<none>";
                }

                PS2Strip::LogError( "material %i ('%s', using base texture '%s')\n", s_aiEnvUniqueMaterialIDs[u], pxMaterial->m_xMaterialName.GetString(), szTexName );
                ++uNumListed;
            }
        }
	}
    
	int iThingsUsed = 0;
	if(bHasReflective) iThingsUsed++;
	if(bHasRainsplash) iThingsUsed++;
	if(bHasDualTexture) iThingsUsed++;
	if(bHas512) iThingsUsed++;

	if( iThingsUsed > 2 )
	{
		PS2Strip::LogError( "More than 2 out of dual-texture, reflective, rainsplash and 512 texture used on environment.\n" );
	}

    if( iBadMaterialCount || iThingsUsed > 2 )
    {
        return false;
    }

    PS2Strip::LogInfo( "done.\n    (Found %u unique materials.)\n", s_uEnvNumUniqueMaterialIDs );
    return true;
}

void PS2Strip_ExporterImplementation::CleanUp( bool bDeallocateAccumulatedData )
{
    s_pxCurMesh = NULL;

    delete[] s_axMeshPositions;
    s_axMeshPositions = NULL;
    delete[] s_axMeshNormals;
    s_axMeshNormals = NULL;
    delete[] s_axMeshMatInfo;
    s_axMeshMatInfo = NULL;
    delete[] s_auMeshColourValues;
    s_auMeshColourValues = NULL;

    delete[] s_axUniqueMeshVtxIdxQuadlets;
    s_axUniqueMeshVtxIdxQuadlets = NULL;

    delete[] s_axUniqueMeshTris;
    s_axUniqueMeshTris = NULL;
    delete[] s_auUniqueMeshTrisPerMaterial;
    s_auUniqueMeshTrisPerMaterial = NULL;
    delete[] s_auNumPostProcessedStripsPerMaterial;
    s_auNumPostProcessedStripsPerMaterial = NULL;

    // Stripper-specific stuff.
    if( !s_xNVTSStripiferOutput.empty() )
    {
        for( unsigned u = 0; u < s_xNVTSStripiferOutput.size(); ++u )
        {
            for( unsigned v = 0; v < s_xNVTSStripiferOutput[u]->m_faces.size(); ++v )
            {
                delete s_xNVTSStripiferOutput[u]->m_faces[v];
            }

            delete s_xNVTSStripiferOutput[u];
        }

        s_xNVTSStripiferOutput.clear();
    }

    if( !s_xPostProcessedStrips.empty() )
    {
        for( unsigned u = 0; u < s_xPostProcessedStrips.size(); ++u )
        {
            delete s_xPostProcessedStrips[u];
        }

        s_xPostProcessedStrips.clear();
    }

    // Character-specific stuff.
    if( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_SKIN_HIER )
    {     
        if( bDeallocateAccumulatedData )
        {
            s_pxInputSkin = NULL;

            delete s_pxOutputHierSkin;
            s_pxOutputHierSkin = NULL;
        }
    }

    // Environment-specific stuff.
    if( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_ENV )
    {
        if( bDeallocateAccumulatedData )
        {
            delete[] s_aiEnvUniqueMaterialIDs;
            s_aiEnvUniqueMaterialIDs = NULL;

            if( s_axMaterialIDToModuleStripsMaps )
            {
                Asura_Assert( s_uEnvNumUniqueMaterialIDs, "" );

                for( unsigned uMap = 0; uMap < s_uEnvNumUniqueMaterialIDs; ++uMap )
                {
                    MaterialIDToModuleStripsMap& xMap = s_axMaterialIDToModuleStripsMaps[uMap];
                    const unsigned uNumModulesInMap = xMap.xModuleData.size();

                    for( unsigned uModule = 0; uModule < uNumModulesInMap; ++uModule )
                    {
                        ModuleStripsWithSameMaterial* pxStrips = xMap.xModuleData[uModule];
                        const unsigned uNumStripsInModule = pxStrips->xStrips.size();

                        for( u_int uStrip = 0; uStrip < uNumStripsInModule; ++uStrip )
                        {
                            delete pxStrips->xStrips[uStrip];
                        }

                        delete pxStrips;
                    }
                }

                delete[] s_axMaterialIDToModuleStripsMaps;
                s_axMaterialIDToModuleStripsMaps = NULL;
            }
        }

		delete s_pSubdividedMesh;
		s_pSubdividedMesh = NULL;
    }
}

bool PS2Strip_ExporterImplementation::ProcessVerticesAndEliminateDuplicates( void )
{
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	// (AAM - 08.12.2006) Gnh, this access bits of the old material system. Unlikely that we'll
	// want to export to PS2 with this on anyway

	Asura_Assert( false, "Can't perform PS2 mesh processing if material response system is switched on." );
	return false;

#else

    const PS2Strip::EXPORTTYPE eExportType = PS2Strip::GetCurExportType();

    Asura_Assert( s_pxCurMesh->AreNormalsValid(), "" );
    Asura_Assert( eExportType != PS2Strip::EXPORTTYPE_ENV || s_pxCurMesh->PrelightingDataValid, "" );
    Asura_Assert( eExportType != PS2Strip::EXPORTTYPE_SKIN_HIER || s_pxInputSkin, "" );
    Asura_Assert( s_pxCurMesh->PolygonDataValid, "" );
    Asura_Assert( s_pxCurMesh->TextureDataValid, "" );
    Asura_Assert( s_pxCurMesh->VertexDataValid,  "" );

    // Output progress message.
    PS2Strip::LogInfo( "Processing vertices and eliminating duplicates...\n" );

    int iPacket = 0;
	Tools_SubdividedMesh_Packet* pPacket = 0;

    Asura_Assert( !s_pSubdividedMesh, "" );
	if( s_bSubdivideMesh )
	{
        Asura_Assert( eExportType == PS2Strip::EXPORTTYPE_ENV, "" );

        Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_SUBDIV_AND_PRELIGHT );
        s_pSubdividedMesh = new Tools_SubdividedMesh( s_iEnvCurModuleIdx, true );
        Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_ELIM_DUP_VTX_INFO );
        if( !s_pSubdividedMesh )
        {
            PS2Strip::LogError( "ERROR: Out of memory!\n" );
            return false;
        }
        
        if( !s_pSubdividedMesh->NumberOfPackets )
        {
            PS2Strip::LogError( "ERROR: Mesh \"%s\" generated too many polys when subdivided.\n", s_pxCurMesh->Name.GetString() );
            return false;
        }

		pPacket = &s_pSubdividedMesh->Packets[0];
	}

    u_int uNumSrcTris;
    u_int uNumSrcVerts;
    if( s_pSubdividedMesh )
    {
        uNumSrcTris  = s_pSubdividedMesh->TotalTris;
        uNumSrcVerts = s_pSubdividedMesh->TotalVertices;
    }
    else
    {
        uNumSrcTris  = s_pxCurMesh->NumberOfTriangles;
        uNumSrcVerts = 3 * s_pxCurMesh->NumberOfTriangles;
    }

    u_int uNumSrcTrisWithNonExportFlags = 0;

    // Just allocate worst cases for arrays. Not really going to be that vast
    // and we're on PC anyway. Note that we're allowing additional space for
    // a 1 vertex overflow for each component, so that we don't overflow any
    // buffers when we're processing the last vertex in mesh that has no
    // duplicate data.
    s_axMeshNormals                      = new Asura_Vector_3[uNumSrcVerts + 1];
    s_uNumMeshNormals                    = 0;
    s_axMeshMatInfo                      = new PS2Strip_MatInfo[uNumSrcVerts + 1];
    s_uNumMeshMatInfos                   = 0;
    s_auMeshColourValues                 = new u_int[uNumSrcVerts + 1];
    s_uNumMeshColourValues               = 0;
    s_axUniqueMeshVtxIdxQuadlets         = new PS2Strip_VtxIdxQuadlet[uNumSrcVerts + 1];
    s_uNumUniqueMeshVtxIdxQuadlets       = 0;
    s_uNumUnusedUniqueMeshVtxIdxQuadlets = 0;
    s_axUniqueMeshTris                   = new PS2Strip_Tri[uNumSrcTris];
    s_uNumUniqueMeshTris                 = 0;
    s_auUniqueMeshTrisPerMaterial        = new u_int[1 + ASURA_MAX_NO_OF_MATERIALS];

    // Special handling for mesh positions. In the case of non-subdivided
    // environments or objects, just use the source position indices. The
    // higher level Asura_Mesh stuff doesn't seem to leave any duplicate
    // positions lying around for that case.
    if( eExportType != PS2Strip::EXPORTTYPE_ENV || s_pSubdividedMesh )
    {
        s_axMeshPositions   = new Asura_Vector_3[uNumSrcVerts + 1];
        s_uNumMeshPositions = 0;
    }
    else
    {
        s_uNumMeshPositions = s_pxCurMesh->NumberOfVertices;
        s_axMeshPositions   = new Asura_Vector_3[s_uNumMeshPositions];

        if( s_axMeshPositions )
        {
            Asura_Assert( sizeof(Asura_Mesh_Vertex) == sizeof(Asura_Vector_3), "" );
            memcpy( s_axMeshPositions, s_pxCurMesh->VertexDataPtr, s_uNumMeshPositions * sizeof(Asura_Vector_3) );
        }
    }

    if( !s_axMeshPositions || !s_axMeshNormals || !s_axMeshMatInfo || !s_auMeshColourValues || !s_axUniqueMeshVtxIdxQuadlets || !s_axUniqueMeshTris || !s_auUniqueMeshTrisPerMaterial )
    {
        PS2Strip::LogError( "ERROR: Out of memory!\n" );
        return false;
    }

    memset( s_auUniqueMeshTrisPerMaterial, 0, (1 + ASURA_MAX_NO_OF_MATERIALS) * sizeof(u_int) );

    // (MP - 10.03.2005) Small change to the scheme for weeding out duplicate
    // vertices. Previously we were individually approximately comparing each
    // piece of vertex information. However, this is a bit pointless if any
    // earlier vertex information made the current vertex distinct from any
    // previously encountered one, since obviously it's going to end up being
    // a new vertex anyway and that's what we ultimately care about. So now
    // we maintain a list of potential duplicate vertices as we go. The main
    // benefits of this are: (i) it prevents quantisation of the vertex
    // colours (perhaps particularly the alpha) unless it's going to be a
    // shared vertex anyway; (ii) it allows us to use a larger tolerances for
    // the approximate comparisons when we know all of the other components
    // must match.
    const u_int uMAX_POTENTIAL_DUP_VERTS = 32; // 32 should be higher than any vaguely sensible mesh should ever produce.
    u_int auPotentialDupVerts[uMAX_POTENTIAL_DUP_VERTS];
	

	float fPosTol     = g_fWeldTolerance;  // +/- 0.25cm.

	const float fNrmTol     = 0.996194698092f;   // +/- 5 (where this value is cos 5)
    const float fZeroNrmTol = 0.01f * 0.01f;
    const float fUVTol      = 0.01f;             // +/- 0.01 in any channel.
    const u_int uColTol     = 2;                 // +/- 2 in any channel.

	int iPoly = -1;
    u_int auCurrentPolyVertexQuadletIndices[4];

	while( true )
	{
		u_int uNumVertsInInputPoly;
        
		//advance to the next polygon
		if(s_pSubdividedMesh)
		{
            Asura_Assert( !IsMaterialDualTexture(pPacket->m_iMaterialID), "Subdivision in the PS2 export and dual texturing are currently mutually exclusive." );

            iPoly++;
			uNumVertsInInputPoly = 3;
			if(iPoly >= pPacket->NumberOfTris)
			{
				//try the next packet
				iPoly = 0;
				iPacket++;
				if(iPacket>=s_pSubdividedMesh->NumberOfPackets) break;//finished
				pPacket = &s_pSubdividedMesh->Packets[iPacket];
				Asura_Assert(pPacket->NumberOfTris,"");
			}
		}
		else
		{
			iPoly++;
			if(iPoly >= s_pxCurMesh->NumberOfPolygons) break;
	        uNumVertsInInputPoly = s_pxCurMesh->PolygonDataPtr[iPoly].NumberOfVertices();

            // Check this poly should be included in our render data.
		    if( s_pxCurMesh->PolygonFlagDataValid )
            {
                const int iFlags = s_pxCurMesh->PolygonFlagDataPtr[iPoly].Flags;
                if( iFlags & (AsuraPolygonFlag_DontExport | AsuraPolygonFlag_DontRender | AsuraPolygonFlag_Water) )
			    {
                    uNumSrcTrisWithNonExportFlags += ( 3 == uNumVertsInInputPoly ? 1 : 2 );
				    continue;
			    }
            }
        }

        Asura_Assert( uNumVertsInInputPoly == 3 || uNumVertsInInputPoly == 4, "" );
           
        for( u_int uVtxWithinInputPoly = 0; uVtxWithinInputPoly < uNumVertsInInputPoly; ++uVtxWithinInputPoly )
        {
			Tools_SubdividedMesh_Vertex* pSubdividedVertex = NULL;
			if(pPacket) pSubdividedVertex = &pPacket->Vertices[pPacket->TriList[iPoly*3+uVtxWithinInputPoly]];

            const int  iMaterialID  = ( s_pSubdividedMesh ? pPacket->m_iMaterialID : s_pxCurMesh->TextureDataPtr[iPoly].m_iMaterialID );
            const bool bDualTexture = ( eExportType == PS2Strip::EXPORTTYPE_ENV && IsMaterialDualTexture(iMaterialID) );

            u_int uPosIdx;

            // Position.
            if( eExportType != PS2Strip::EXPORTTYPE_ENV || s_pSubdividedMesh )
            {
                const Asura_Vector_3* pxPos;
                if( s_pSubdividedMesh )
                {
                    pxPos = &pSubdividedVertex->Position;
                }
                else
                {
                    int iVertexID = s_pxCurMesh->PolygonDataPtr[iPoly].VertexID[uVtxWithinInputPoly];
                    pxPos = &s_pxCurMesh->VertexDataPtr[iVertexID].Position;
                }

                // Search for previous matching position.
                uPosIdx = UINT_MAX;
                for( u_int u = 0; u < s_uNumMeshPositions; ++u )
                {
                    if( pxPos->ApproximatelyEquals(s_axMeshPositions[u], fPosTol) )
                    {
                        uPosIdx = u;
                        break;
                    }
                }

                // New?
                if( UINT_MAX == uPosIdx )
                {
                    Asura_Assert( s_uNumMeshPositions < uNumSrcVerts, "" );
                    uPosIdx = s_uNumMeshPositions++;
                    s_axMeshPositions[uPosIdx] = *pxPos;
                }
			}
			else
			{
                uPosIdx = s_pxCurMesh->PolygonDataPtr[iPoly].VertexID[uVtxWithinInputPoly];
			}

            // Create initial list of potential duplicates.
            u_int uNumPotentialDupVerts = 0;
            for( u_int u = 0; u < s_uNumUniqueMeshVtxIdxQuadlets; ++u )
            {
                if( s_axUniqueMeshVtxIdxQuadlets[u].GetPosIdx() == uPosIdx )
                {
                    auPotentialDupVerts[uNumPotentialDupVerts++] = u;
                    if( uNumPotentialDupVerts == uMAX_POTENTIAL_DUP_VERTS )
                    {
                        Asura_Assert( false, "TELL MORGAN ABOUT THIS" ); // Just need to increase uMAX_POTENTIAL_DUP_VERTS if this ever fires.
                        break;
                    }
                }
            }

            const Asura_Vector_3* pxNrm;
			const Asura_Vector_3 xUp(0.f, -1.f, 0.f);

            // Normal.
            {
				if( s_pSubdividedMesh )
				{
	    	       	pxNrm = &pSubdividedVertex->Normal;
				}
				else
                {
                    pxNrm = &s_pxCurMesh->NormalsDataPtr[iPoly].VertexNormal[uVtxWithinInputPoly];
                }
				if(eExportType == PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT)
				{
					// Fiddle the normal to be directly up for botanicals flagged with ignore normal for lighting
					if(s_pxCurMesh->PolygonFlagDataValid && s_pxCurMesh->PolygonFlagDataPtr)
					{
						if(s_pxCurMesh->PolygonFlagDataPtr[iPoly].Flags & AsuraPolygonFlag_IgnoreNormalForLighting)
						{
							pxNrm = &xUp;
						}
					}
				}


                // Update potential duplicate vert list.
                for( u_int u = 0; u < uNumPotentialDupVerts; )
                {
                    // (MP - 25.05.2005) Added a check here for zero-length
                    // normals. Obviously in this case our angle check will
                    // fail horribly and we'll regard two zero-length normals
                    // as distinct. This can be a problem in extreme cases
                    // since then it's possible to end up generating rather
                    // more vertices than were in the input mesh.
                    const Asura_Vector_3& xOtherNrm = s_axMeshNormals[ s_axUniqueMeshVtxIdxQuadlets[auPotentialDupVerts[u]].GetNrmIdx() ];
                    if( *pxNrm * xOtherNrm > fNrmTol || (pxNrm->MagnitudeSquared() < fZeroNrmTol && xOtherNrm.MagnitudeSquared() < fZeroNrmTol) )
                    {
                        ++u;
                    }
                    else
                    {
                        auPotentialDupVerts[u] = auPotentialDupVerts[--uNumPotentialDupVerts];
                    }
                }
            }

            PS2Strip_MatInfo* pxMatInfo;

            // Texture data.
            {
                pxMatInfo = &s_axMeshMatInfo[s_uNumMeshMatInfos];

                pxMatInfo->m_iMaterialID   = iMaterialID;
                pxMatInfo->m_bDualTextured = bDualTexture;

   				if( s_pSubdividedMesh )
				{
                    pxMatInfo->m_afU[0] = pSubdividedVertex->U * s_fTexUScale;
                    pxMatInfo->m_afV[0] = pSubdividedVertex->V * s_fTexVScale;
				}
				else
				{
					const Asura_Mesh_TextureInfo* pxTexInfo = s_pxCurMesh->GetTextureInfoForExport(iPoly);
                    
                    for( u_int u = 0; u < (bDualTexture ? 2u : 1u); ++u )
                    {
                    	pxMatInfo->m_afU[u] = pxTexInfo->U[u][uVtxWithinInputPoly] * s_fTexUScale;
                    	pxMatInfo->m_afV[u] = pxTexInfo->V[u][uVtxWithinInputPoly] * s_fTexVScale;
                    }
				}

                // Search for duplicate tex info and update potential
                // duplicate vert list.
                for( u_int u = 0; u < uNumPotentialDupVerts; )
                {
                    u_int uOtherMatInfoIdx = s_axUniqueMeshVtxIdxQuadlets[auPotentialDupVerts[u]].GetMatInfoIdx();
                    if( pxMatInfo->ApproximatelyEquals(s_axMeshMatInfo[uOtherMatInfoIdx], fUVTol) )
                    {
                        ++u;
                    }
                    else
                    {
                        auPotentialDupVerts[u] = auPotentialDupVerts[--uNumPotentialDupVerts];
                    }
                }
            }

            u_int uABGR;

            // Colour value.
            {
                switch( eExportType )
                {
					case PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT:
                    case PS2Strip::EXPORTTYPE_OBJECT:
                        // No support for vertex alpha on objects ATM so fix
                        // to be opaque.
                        uABGR = (128u << 24u) | 0u;
                        break;

                    case PS2Strip::EXPORTTYPE_SKIN_HIER:
                    {
                        u_int uA;
                    
                        // If the material isn't flagged as transparent or
						// additive alpha then we'll ignore the vertex alpha
						// on PS2 anyway so just fix it to 128.
                        const u_int uMaterialFlags = Asura_Material_System::GetMaterialFlags( iMaterialID );
                        if( uMaterialFlags & (AsuraMaterialFlag_Transparent | AsuraMaterialFlag_PS2AdditiveAlpha) )
                        {
                            uA = Asura_Maths::FloatToInt_Near( s_pxCurMesh->TextureDataPtr[iPoly].m_afAlpha[uVtxWithinInputPoly] * 128.0f );
                            uA = Asura_Maths::Min( uA, 128u );
                        }
                        else
                        {
                            uA = 128;
                        }

                        uABGR = (uA << 24u) | 0u;                       
                        break;
                    }

                    case PS2Strip::EXPORTTYPE_ENV:
                    {
                        u_int uRGB;
                        if( s_pSubdividedMesh )
				        {
                            uRGB = pSubdividedVertex->Colour;
				        }
				        else
				        {
                            uRGB = Tools_Level_Export::ScalePrelightValue( s_pxCurMesh->PrelightingDataPtr[iPoly].Colour[uVtxWithinInputPoly] );
				        }

             	        // Kevin Lea - 12/02/2003 11:22:43 - although a
                        // lighting value of 128 is normal brightness on the
                        // PS2, allow full range (0-255) for over-saturation
                        // effects
                        u_int uR = (uRGB & 0x00FF0000u) >> 16u;
                        u_int uG = (uRGB & 0x0000FF00u) >>  8u;
                        u_int uB = (uRGB & 0x000000FFu) >>  0u;

                        u_int uA;
                        if( bDualTexture )
                        {
                            uA = Asura_Maths::FloatToInt_Near( s_pxCurMesh->TextureDataPtr[iPoly].m_afAlpha[uVtxWithinInputPoly] * 128.0f );
                            uA = Asura_Maths::Min( uA, 128u );
                        }
                        else
                        {
                            // Just fix to be opaque.
                            uA = 128u;
                        }

                        uABGR = (uA << 24u) | (uB << 16u) | (uG << 8u) | (uR << 0u);
                        
                        break;
                    }
                }

                // Update potential duplicate vert list.
                for( u_int u = 0; u < uNumPotentialDupVerts; )
                {
                    u_int uOtherColourIdx = s_axUniqueMeshVtxIdxQuadlets[auPotentialDupVerts[u]].GetColourIdx();
                    if( IsABGRApproximatelyEqual(uABGR, s_auMeshColourValues[uOtherColourIdx], uColTol) )
                    {
                        ++u;
                    }
                    else
                    {
                        auPotentialDupVerts[u] = auPotentialDupVerts[--uNumPotentialDupVerts];
                    }
                }
            }

            u_int uQuadletIdx;
            bool  bIsNewQuadlet;

            // Quadlet.
            // If there's a remaining potential duplicate then it is actually
            // duplicate.
            if( uNumPotentialDupVerts )
            {
                // It's possible to get to this point with more than 1
                // potential duplicate. This is because we might have a
                // sequence of vertices {A, B, C}  such that A ~ C and B ~ C
                // but !(A ~ B) (where ~ denotes approximate equality.) In
                // this case we'll reach C with both A and B as potential
                // duplicates.
                if( 1 == uNumPotentialDupVerts )
                {
                    // Simple case. Just point at the duplicate vertex
                    // instead.
                    uQuadletIdx   = auPotentialDupVerts[0];
                    bIsNewQuadlet = false;
                }
                else
                {
                    // This is the more interesting case. What we want to do
                    // is use this new vertex to replace all of the potential
                    // duplicates, even in tris we've already processed, since
                    // it's an acceptable replacement for all of them.
                    
                    // Get the new quadlet index.
                    uQuadletIdx   = s_uNumUniqueMeshVtxIdxQuadlets++;
                    bIsNewQuadlet = true;

                    // Note that we can't remove the existing quadlets that
                    // we're replacing from the main quadlet list because
                    // that will affect a huge number of other indices.
                    // We just leave them in and record how many are unused.
                    s_uNumUnusedUniqueMeshVtxIdxQuadlets += uNumPotentialDupVerts;
                    Asura_Assert( s_uNumUnusedUniqueMeshVtxIdxQuadlets < s_uNumUniqueMeshVtxIdxQuadlets, "" );

                    // Fix up existing tris.
                    for( u_int uTri = 0; uTri < s_uNumUniqueMeshTris; ++uTri )
                    {
                        PS2Strip_Tri& xTri = s_axUniqueMeshTris[uTri];
                        for( u_int uIdxWithinTri = 0; uIdxWithinTri < 3; ++uIdxWithinTri )
                        {
                            for( u_int uDup = 0; uDup < uNumPotentialDupVerts; ++uDup )
                            {
                                if( xTri.GetVtxIdx(uIdxWithinTri) == auPotentialDupVerts[uDup] )
                                {
                                    xTri.ChangeVtxIdx( uIdxWithinTri, uQuadletIdx );
                                    break;
                                }
                            }
                        }
                    }

                    // Fix up earlier vertices in this tri.
                    for( u_int uEarlierVert = 0; uEarlierVert < uVtxWithinInputPoly; ++uEarlierVert )
                    {
                        for( u_int uDup = 0; uDup < uNumPotentialDupVerts; ++uDup )
                        {
                            if( auCurrentPolyVertexQuadletIndices[uEarlierVert] == auPotentialDupVerts[uDup] )
                            {
                                auCurrentPolyVertexQuadletIndices[uEarlierVert] = uQuadletIdx;
                                break;
                            }
                        }
                    }
                }
            }
            else
            {
                Asura_Assert( s_uNumUniqueMeshVtxIdxQuadlets < uNumSrcVerts, "" );

//              // TEMP
//              // Search through preceeding vertices in the subdivided mesh
//              // making sure we haven't missed a match.
//              if( s_pSubdividedMesh )
//              {
//              	const Tools_SubdividedMesh_Packet* pxSearchPacket = &s_pSubdividedMesh->Packets[iPacket];
//                  int iSearchPoly = -1;
//
//                  while( true )
//                  {
//                      iSearchPoly++;
//                      if( iSearchPoly >= pxSearchPacket->NumberOfTris )
//                      {
//                          break;
//                      }
//
//                      if( iSearchPoly > iPoly )
//                      {
//                          break;
//                      }
//
//                      for( u_int uSearchVtx = 0; uSearchVtx < 3; ++uSearchVtx )
//                      {
//                          if( iSearchPoly == iPoly && uSearchVtx >= uVtxWithinInputPoly )
//                          {
//                              break;
//                          }
//
//		                    const Tools_SubdividedMesh_Vertex* pxSearchVertex = &pxSearchPacket->Vertices[pxSearchPacket->TriList[iSearchPoly * 3 + uSearchVtx]];
//                          if( *pxSearchVertex == *pSubdividedVertex )
//                          {
//                              Asura_Assert( false, "hello" );
//                              int i = 0; ++i;
//                          }
//                      }
//                  }
//              }
                
                uQuadletIdx   = s_uNumUniqueMeshVtxIdxQuadlets++;
                bIsNewQuadlet = true;
            }
            // End quadlet index calculation.

            // Setup new quadlet?
            if( bIsNewQuadlet )
            {
                Asura_Assert( s_uNumMeshNormals < uNumSrcVerts, "" );
                u_int uNrmIdx = s_uNumMeshNormals++;
                s_axMeshNormals[uNrmIdx] = *pxNrm;
                
                Asura_Assert( s_uNumMeshMatInfos < uNumSrcVerts, "" );
                u_int uMatInfoIdx = s_uNumMeshMatInfos++;
                // No need to copy since we used the first unused entry
                // as a temporary anyway.

                Asura_Assert( s_uNumMeshColourValues < uNumSrcVerts, "" );
                u_int uColourIdx = s_uNumMeshColourValues++;
                s_auMeshColourValues[uColourIdx] = uABGR;

                s_axUniqueMeshVtxIdxQuadlets[uQuadletIdx].SetIndices( uPosIdx, uNrmIdx, uMatInfoIdx, uColourIdx );
            }

            auCurrentPolyVertexQuadletIndices[uVtxWithinInputPoly] = uQuadletIdx;
        }
        // End loop through vertices.

        // Create tris.
        {
            // NB: use 'unique' counter for now, even though they're not yet.

            s_axUniqueMeshTris[s_uNumUniqueMeshTris++].SetVtxIndices( auCurrentPolyVertexQuadletIndices[0],
                                                                      auCurrentPolyVertexQuadletIndices[1],
                                                                      auCurrentPolyVertexQuadletIndices[2] );
            
            if( 4 == uNumVertsInInputPoly )
            {
                s_axUniqueMeshTris[s_uNumUniqueMeshTris++].SetVtxIndices( auCurrentPolyVertexQuadletIndices[2],
                                                                          auCurrentPolyVertexQuadletIndices[3],
                                                                          auCurrentPolyVertexQuadletIndices[0] );
            }

            Asura_Assert( s_uNumUniqueMeshTris <= uNumSrcTris, "" );
        }
	}
    // End loop through tris/polys.

    // Go through tris and weed out degenerates and duplicates and update
    // stats. Note that we don't do this during the above process to keep
    // things simple since as we modify quadlet indices, etc., we may
    // introduce extra degenerates or duplicates.
    {
        // Hashtables to speed up duplicate tri checks below.
        const u_int uTriTableSize = ( s_uNumUniqueMeshTris > 1 ? Asura_Maths::Max(1u, 1u << (Asura_Maths::GetCeilLog2(s_uNumUniqueMeshTris) - 1)) : 1 );
        Tools_HashTable<PS2Strip_Tri> xTriHashTable( uTriTableSize );

        Asura_Assert( s_uNumUniqueMeshTris == uNumSrcTris - uNumSrcTrisWithNonExportFlags, "" );
        const u_int uNumTrisToCheck = s_uNumUniqueMeshTris;
        s_uNumUniqueMeshTris = 0;

        const u_int uNumDupBeforeThisMesh   = s_uTotalDuplicateSourceTris;
        const u_int uNumDegenBeforeThisMesh = s_uTotalDegenerateSourceTris;

        for( u_int uSrcTri = 0; uSrcTri < uNumTrisToCheck; ++uSrcTri )
        {
            const PS2Strip_Tri& xNext = s_axUniqueMeshTris[uSrcTri];

            bool bDupOrDegen = false;

            // Duplicate?
            const PS2Strip_Tri* pxInsertedTri = xTriHashTable.Insert( &xNext );
            if( !pxInsertedTri )
            {
                PS2Strip::LogError( "ERROR: Out of memory!\n" );
                return false;
            }
            else if( pxInsertedTri != &xNext )
            {
                ++s_uTotalDuplicateSourceTris;
                bDupOrDegen = true;
            }

            // Degenerate?
            {
                // Get the three position indices.
                const u_int uPosIdx0 = s_axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(0)].GetPosIdx();
                const u_int uPosIdx1 = s_axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(1)].GetPosIdx();
                const u_int uPosIdx2 = s_axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(2)].GetPosIdx();

                if( uPosIdx0 == uPosIdx1 || uPosIdx0 == uPosIdx2 || uPosIdx1 == uPosIdx2 )
                {
                    ++s_uTotalDegenerateSourceTris;
                    bDupOrDegen = true;
                }
            }

            if( !bDupOrDegen )
            {
                // Copy down if required.
                if( s_uNumUniqueMeshTris != uSrcTri )
                {
                    s_axUniqueMeshTris[s_uNumUniqueMeshTris] = xNext;
                }

                ++s_uNumUniqueMeshTris;

                const PS2Strip_VtxIdxQuadlet& xFirstQuadlet = s_axUniqueMeshVtxIdxQuadlets[xNext.GetVtxIdx(0)];
                int iMaterialID = s_axMeshMatInfo[xFirstQuadlet.GetMatInfoIdx()].m_iMaterialID;
                ++s_auUniqueMeshTrisPerMaterial[Asura_Material_System::TranslateMaterialIDToArrayIdx(iMaterialID) + 1];
            }
        }

        // Extra info 
        if( s_uTotalDuplicateSourceTris > uNumDupBeforeThisMesh )
        {
            PS2Strip::LogInfo( "    Found %u duplicate tris.\n", s_uTotalDuplicateSourceTris - uNumDupBeforeThisMesh );
        }
        if( s_uTotalDegenerateSourceTris > uNumDegenBeforeThisMesh )
        {
            PS2Strip::LogInfo( "    Found %u degenerate tris.\n", s_uTotalDegenerateSourceTris - uNumDegenBeforeThisMesh );
        }
    }

    // Count number of unique materials in this mesh.
    s_uNumUniqueMeshMaterials = 0;
    for( u_int u = 0; u < 1 + ASURA_MAX_NO_OF_MATERIALS; ++u )
    {
        if( s_auUniqueMeshTrisPerMaterial[u] ) { ++s_uNumUniqueMeshMaterials; }
    }

    // Check we found some tris. // NB: Don't regard this case an error; for
    // example, we might have a mesh consisting entirely of degenerate tris.
    if( !s_uNumUniqueMeshTris )
    {
        PS2Strip::LogInfo( "    Done.\n    No polys for export.\n" );
        return true;
    }

    // Sort tris by material.
    {
        PS2Strip_Tri* pxNewTriList = new PS2Strip_Tri[s_uNumUniqueMeshTris];
        if( !pxNewTriList )
        {
            PS2Strip::LogError( "ERROR: Out of memory!\n" );
            return false;
        }

        u_int auNextMaterialIdx[1 + ASURA_MAX_NO_OF_MATERIALS];

        u_int uOutIdx = 0;
        for( u_int u = 0; u < 1 + ASURA_MAX_NO_OF_MATERIALS; ++u )
        {
            auNextMaterialIdx[u] = uOutIdx;
            uOutIdx += s_auUniqueMeshTrisPerMaterial[u];
        }
        Asura_Assert( uOutIdx == s_uNumUniqueMeshTris, "" );

        for( u_int u = 0; u < s_uNumUniqueMeshTris; ++u )
        {
            const PS2Strip_Tri& xSrcTri = s_axUniqueMeshTris[u];

            const PS2Strip_VtxIdxQuadlet& xFirstVtxQuadlet = s_axUniqueMeshVtxIdxQuadlets[xSrcTri.GetVtxIdx(0)];
            const PS2Strip_MatInfo& xFirstVtxMatInfo = s_axMeshMatInfo[xFirstVtxQuadlet.GetMatInfoIdx()];

            int iArrayIdx = 1 + Asura_Material_System::TranslateMaterialIDToArrayIdx(xFirstVtxMatInfo.m_iMaterialID);
            
            uOutIdx = auNextMaterialIdx[iArrayIdx]++;
            pxNewTriList[uOutIdx] = xSrcTri;
        }
        
        delete[] s_axUniqueMeshTris;
        s_axUniqueMeshTris = pxNewTriList;
    }   

    // Report progress.
    {
        u_int uNumMaterials            = 0;
        u_int uNumDualTextureMaterials = 0;

        for( int i = 0; i < 1 + ASURA_MAX_NO_OF_MATERIALS; ++i )
        {
            if( s_auUniqueMeshTrisPerMaterial[i] )
            {
                ++uNumMaterials;

                int iMaterialID = Asura_Material_System::TranslateArrayIdxToMaterialID( i - 1 );
                if( IsMaterialDualTexture(iMaterialID) )
                {
                    ++uNumDualTextureMaterials;
                }
            }
        }

        const Asura_Char* szFormat =
            "    Done.\n"
            "    Mesh contained %u materials (%u dual texture.)\n"
            "    %u unique vertices from %u in input mesh.\n"
            "    %u unique tris from %u in input mesh.\n";

        PS2Strip::LogInfo( szFormat,
                           uNumMaterials, uNumDualTextureMaterials,
                           (s_uNumUniqueMeshVtxIdxQuadlets - s_uNumUnusedUniqueMeshVtxIdxQuadlets), uNumSrcVerts,
                           s_uNumUniqueMeshTris, uNumSrcTris - uNumSrcTrisWithNonExportFlags );
    }

    return true;

#endif
}

bool PS2Strip_ExporterImplementation::GenerateRequiredVertexData( void )
{
    Asura_Assert( s_pxCurMesh, "" );
    Asura_Assert( PS2Strip::GetCurExportType() != PS2Strip::EXPORTTYPE_SKIN_HIER || (s_pxInputSkin && s_pxOutputHierSkin), "" );

    PS2Strip::LogInfo( "Generating required vertex data... " );

	if( !s_pxCurMesh->AreNormalsValid() )
    {
        s_pxCurMesh->CreateNormals();
	}

    PS2Strip::LogInfo( "done.\n" );
    return true;
}

bool PS2Strip_ExporterImplementation::GenerateEnvPrelighting( void )
{
    Asura_Assert( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_ENV, "" );
    Asura_Assert( s_pxCurMesh, "" );
    Asura_Assert( s_pxCurMesh->AreNormalsValid(), "Call GenerateRequiredVertexData() before GenerateEnvPrelighting()." );

    PS2Strip::LogInfo( "Generating prelighting data... " );

    Asura_Vector_3 xModulePos;
    Asura_Module_System::GetPositionOfModule( s_iEnvCurModuleIdx, xModulePos );
    Asura_Lighting_System::CalculatePrelighting( *s_pxCurMesh, xModulePos, s_iEnvCurModuleIdx );

    PS2Strip::LogInfo( "done.\n" );
    return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     PS2Strip_ExporterImplementation::NVTS_AppendStripperStripToPostProcessedStrip().
// PARAMETERS   pxPPStrip   - The strip to be appended to.
//              pxNVTSStrip - The output from the stripping algorithm.
// RETURNS      None.
// DESCRIPTION  Appends the specified NVTS strip to the specified output
//              strip.
// AUTHOR       Morgan Parry - 02.04.2002.
//
//////////////////////////////////////////////////////////////////////////////
void PS2Strip_ExporterImplementation::NVTS_AppendStripperStripToPostProcessedStrip( PostProcessedStrip* pxPPStrip, const NVTriStrip::NvStripInfo* pxNVTSStrip )
{
    // Get input strip's tri count.
    const unsigned uNumTrisInInputStrip = pxNVTSStrip->m_faces.size();
    Asura_Assert( uNumTrisInInputStrip > 0, "Stripifier output an empty strip." );
    Asura_Assert( uNumTrisInInputStrip + 2 <= PS2Strip::GetMaxStripLength(s_pxInputSkin), "Stripifier output a strip that was too long." );

    // Determine whether or not this is a new output strip.
    const bool bPPStripWasNonEmpty = ( pxPPStrip->xQuadletIndices.size() > 0 );

    // Deal with first tri.
    // Append this tri's vertices to the output strip. If the output strip
    // is non-empty, the first two vertices will need the ADC bit to be
    // set for them on the PS2.
    // (MP - 03.04.2002) TODO: This code will need updating if we start
    // BFCing on the PS2 again because it doesn't deal with alternating
    // winding order correctly.
    NVTriStrip::NvFaceInfo xLastFace = *pxNVTSStrip->m_faces[0];
    for( unsigned u = 0; u < 3; ++u )
    {
        // Determine whether the ADC bit will be required or not.
        unsigned uToPush = xLastFace.m_av[u];
        if( bPPStripWasNonEmpty && u < 2 )
        {
            uToPush |= PostProcessedStrip::NEEDS_ADC_BIT;
        }

        // Store.
        pxPPStrip->xQuadletIndices.push_back( uToPush );
    }
    ++s_uNumOutputTris;

    // Process the remaining tris in this input strip.
    for( unsigned uTri = 1; uTri < uNumTrisInInputStrip; ++uTri )
    {
        // Find unique vertex in this next tri compared to previous
        // tri.
        int iUniqueVertex = NVTriStrip::NvStripifier::GetUniqueVertexInB( &xLastFace, pxNVTSStrip->m_faces[uTri] );

        // If there's no unique vertex, the tri is redundant and we can
        // discard it.
        // (MP - 26.11.2001) TODO: The Nvidia stuff does this, but doesn't it
        // break winding?
        if( iUniqueVertex == -1 ) { continue; }

        // Output the unique vertex.
        pxPPStrip->xQuadletIndices.push_back( iUniqueVertex );
        ++s_uNumOutputTris;

        // Update last tri record.
        xLastFace.m_v0 = xLastFace.m_v1;
        xLastFace.m_v1 = xLastFace.m_v2;
        xLastFace.m_v2 = iUniqueVertex;
    }

    // Update output strip's tri count.
    pxPPStrip->uTriCount += uNumTrisInInputStrip;

    // Update stats.
    if( uNumTrisInInputStrip > s_uEnvMaxLengthFromStripifier )
    {
        s_uEnvMaxLengthFromStripifier = uNumTrisInInputStrip;
    }
}
// End function PS2Strip_ExporterImplementation::NVTS_AppendStripperStripToPostProcessedStrip().

//////////////////////////////////////////////////////////////////////////////
// (MP - 03.09.2004) //
// FUNCTION     PS2Strip_ExporterImplementation::NVTS_CanAppend().
// PARAMETERS
// RETURNS      True iff can append.
// DESCRIPTION  Tests whether or not the specified NVTS strip can be appended
//              to the specified output strip.
// AUTHOR       Morgan Parry - 08.04.2002.
//
//////////////////////////////////////////////////////////////////////////////
bool PS2Strip_ExporterImplementation::NVTS_CanAppend( const PostProcessedStrip* pxPPStrip, const NVTriStrip::NvStripInfo* pxToAppend )
{
    // Calculate num vertices in NVTS strip.
    unsigned uNumToAppend = 3 + (pxToAppend->m_faces.size() - 1);

    // Calculate potential new length.
    unsigned uNewLength = pxPPStrip->xQuadletIndices.size() + uNumToAppend;

    return( uNewLength <= PS2Strip::GetMaxStripLength(s_pxInputSkin) );
}
// End function PS2Strip_ExporterImplementation::NVTS_CanAppend().

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     PS2Strip_ExporterImplementation::NVTS_ExtractMaterialID().
// PARAMETERS   __pxStrip - The strip for which the texture index is required.
// RETURNS      The material index used by the specified strip.
// DESCRIPTION  See RETURNS.
// AUTHOR       Morgan Parry - 08.04.2002.
//
//////////////////////////////////////////////////////////////////////////////
int PS2Strip_ExporterImplementation::NVTS_ExtractMaterialID( const NVTriStrip::NvStripInfo* pxStrip )
{
    Asura_Assert( pxStrip->m_faces.size() > 0, "" );
    int iFirstQuadletIdx = pxStrip->m_faces[0]->m_v0;

    Asura_Assert( iFirstQuadletIdx < static_cast<int>(s_uNumUniqueMeshVtxIdxQuadlets), "" );
    const PS2Strip_VtxIdxQuadlet& xFirstQuadlet = s_axUniqueMeshVtxIdxQuadlets[iFirstQuadletIdx];

    Asura_Assert( xFirstQuadlet.GetMatInfoIdx() < s_uNumMeshMatInfos, "" );
    const PS2Strip_MatInfo& xMatInfo = s_axMeshMatInfo[xFirstQuadlet.GetMatInfoIdx()];

    const int iMaterialID = xMatInfo.m_iMaterialID;

#if _DEBUG
    // Verify that all other vertices in the strip use the same material.
    for( unsigned u = 0; u < pxStrip->m_faces.size(); ++u )
    {
        for( unsigned v = 0; v < 3; ++v )
        {
            int iQuadletIdx = pxStrip->m_faces[u]->m_av[v];

            Asura_Assert( iQuadletIdx < static_cast<int>(s_uNumUniqueMeshVtxIdxQuadlets), "" );
            const PS2Strip_VtxIdxQuadlet& xQuadlet = s_axUniqueMeshVtxIdxQuadlets[iQuadletIdx];

            Asura_Assert( xQuadlet.GetMatInfoIdx() < s_uNumMeshMatInfos, "" );
            const PS2Strip_MatInfo& xMatInfo = s_axMeshMatInfo[xQuadlet.GetMatInfoIdx()];

            Asura_Assert( xMatInfo.m_iMaterialID == iMaterialID, "Multiple material IDs in single strip!" );
        }
    }
#endif

    return iMaterialID;
}
// End function PS2Strip_ExporterImplementation::NVTS_ExtractMaterialID().

////////////////////////////////////////////////////////////////////////////////
////
//// FUNCTION     PS2Strip_ExporterImplementation::NVTS_IsNextTriCW().
//// PARAMETERS   The current output tri strip.
//// RETURNS      True iff the next tri to be added to the strip should have
////              CW winding.
//// DESCRIPTION  See RETURNS.
//// AUTHOR       Morgan Parry - 22.11.2001.
////
////////////////////////////////////////////////////////////////////////////////
//bool PS2Strip_ExporterImplementation::NVTS_IsNextTriCW( const PostProcessedStrip* __pxCurOutputStrip )
//{
//    // The first tri is CW iff USE_CW_WINDING is true. Therefore the next
//    // tri to be added to the strip should be CW iff the size
//    const unsigned uStripLength = __pxCurOutputStrip->xQuadletIndices.size();
//    bool bSameWindingAsFirstTri = ( (uStripLength & 1) == 0 );
//    return( USE_CW_WINDING ? bSameWindingAsFirstTri : !bSameWindingAsFirstTri );
//}
//// End function PS2Strip_ExporterImplementation::NVTS_IsNextTriCW().

////////////////////////////////////////////////////////////////////////////////
////
//// FUNCTION     PS2Strip_ExporterImplementation::NVTS_IsTriCW().
//// PARAMETERS   The original tri as output by the stripifier and the first
////              two indices of the modified triangle (ie. as will be output
////              in the strip in order to preserve alternating winding order.)
//// RETURNS      True iff the modified triangle has CW winding.
//// DESCRIPTION  See RETURNS.
//// AUTHOR       Morgan Parry - 21.11.2001.
////
////////////////////////////////////////////////////////////////////////////////
//bool PS2Strip_ExporterImplementation::NVTS_IsTriCW( const NVTriStrip::NvFaceInfo* __pxTri, int __iV0, int __iV1 )
//{
//  Asura_Assert( __pxTri, "" );
//  Asura_Assert( __iV0 >= 0, "" );
//  Asura_Assert( __iV1 >= 0, "" );
//
//  if( __pxTri->m_v0 == __iV0 )
//  {
//      return( __pxTri->m_v1 == __iV1 );
//  }
//  else if( __pxTri->m_v1 == __iV0 )
//  {
//      return( __pxTri->m_v2 == __iV1 );
//  }
//  else
//  {
//      Asura_Assert( __pxTri->m_v2 == __iV0, "" );
//      return( __pxTri->m_v0 == __iV1 );
//  }
//}
//// End function PS2Strip_ExporterImplementation::NVTS_IsTriCW().

// Currently it makes sense for our output strips to be sorted so that the
// shorter strips are at either end, with strips of similar length being
// grouped together as much as possible (ie. so it would end up looking like
// a normal distribution.) This should minimise stalls at runtime. This
// method helps with that. The indices specified are inclusive.
bool PS2Strip_ExporterImplementation::SortPostProcessedStripsOnLength( u_int uStartIdx, u_int uEndIdx )
{
    Asura_Assert( uStartIdx <= uEndIdx && uEndIdx < s_xPostProcessedStrips.size(), "" );
	std::vector<PostProcessedStrip*>::iterator xStart = s_xPostProcessedStrips.begin() + uStartIdx;
	std::vector<PostProcessedStrip*>::iterator xEnd   = s_xPostProcessedStrips.begin() + (uEndIdx + 1);

    const u_int uNumToSort = (uEndIdx + 1) - uStartIdx;
    if( uNumToSort <= 1 ) { return true; }

    // First sort the strips into ascending order.
    std::sort( xStart, xEnd, PostProcessedStrip::LengthCmp() );

    // And sort into our intended order.
    {
        // Create a temporary copy so we can overwrite the input.
        std::vector<PostProcessedStrip*> xCopy;
        xCopy.resize( uNumToSort );
        std::copy( xStart, xEnd, &xCopy[0] );

        // To get our desired ordering, alternate between the ends of the
        // output array as we write out the ptrs.
        for( u_int u = 0; u < uNumToSort; ++u )
        {
            u_int uOdd    = u & 1;
            u_int uOffset = (u >> 1) + uOdd;
            u_int uOutIdx = uStartIdx + (uOdd ? uNumToSort - uOffset : uOffset);

            s_xPostProcessedStrips[uOutIdx] = xCopy[u];
        }
    }

    return true;
}

// Sorts by material ID in ascending order. The indices specified are
// inclusive.
bool PS2Strip_ExporterImplementation::SortPostProcessedStripsOnMaterial( u_int uStartIdx, u_int uEndIdx )
{
    Asura_Assert( uStartIdx <= uEndIdx && uEndIdx < s_xPostProcessedStrips.size(), "" );
	std::vector<PostProcessedStrip*>::iterator xStart = s_xPostProcessedStrips.begin() + uStartIdx;
	std::vector<PostProcessedStrip*>::iterator xEnd   = s_xPostProcessedStrips.begin() + (uEndIdx + 1);

    if( PS2Strip::EXPORTTYPE_ENV == PS2Strip::GetCurExportType() )
    {
        std::sort( xStart, xEnd, PostProcessedStrip::EnvMaterialCmp() );
    }
    else
    {
        std::sort( xStart, xEnd, PostProcessedStrip::MaterialCmp() );
    }

    return true;
}

bool PS2Strip_ExporterImplementation::NVTS_PostProcessStrips( void )
{
    Asura_Assert( s_xNVTSStripiferOutput.size() > 0, "" );
    Asura_Assert( s_xPostProcessedStrips.size() == 0, "" );

    // Output progress message.
    PS2Strip::LogInfo( "Post-processing strips... " );

    // Reset statistics.
    s_uNumOutputTris = 0; //s_uNumDegenerateOutputTris = 0;

    const unsigned uNumStrips = s_xNVTSStripiferOutput.size();

    // Allocate an array to record which NVTS strips we've dealt with.
    bool* abAppended = new bool[uNumStrips];
    if( !abAppended )
    {
        PS2Strip::LogError( "ERROR: Out of memory!\n" );
        return false;
    }

    // Initialise it.
    for( unsigned u = 0; u < uNumStrips; ++u )
    {
        abAppended[u] = false;
    }

    PostProcessedStrip* pxCurPPStrip = NULL;
    for( u_int uNumStripsToAppend = uNumStrips; uNumStripsToAppend--; )
    {
        const NVTriStrip::NvStripInfo* pxToAppend = NULL;
        unsigned uIndexToAppend                   = UINT_MAX;

        bool bStartNewPPStrip = true;
        if( pxCurPPStrip )
        {
            // Find next longest strip that (a) hasn't been dealt with yet,
            // (b) has the correct texture ID, and (c) will fit on the end of
            // the current output strip.
            unsigned uLongestLength = 0;
            for( unsigned uStrip = 0; uStrip < uNumStrips; ++uStrip )
            {
                // Have we dealt with this NVTS strip already?
                if( abAppended[uStrip] ) { continue; }

                const NVTriStrip::NvStripInfo* pxNVTSStrip = s_xNVTSStripiferOutput[uStrip];

                // Is it the longest so far?
                if( pxNVTSStrip->m_faces.size() <= uLongestLength ) { continue; }

                // Does it have the correct material ID?
                if( NVTS_ExtractMaterialID(pxNVTSStrip) != pxCurPPStrip->iMaterialID ) { continue; }

                // Will it fit on the end of the current output strip?
                if( !NVTS_CanAppend(pxCurPPStrip, pxNVTSStrip) ) { continue; }

                // Best candidate so far.
                pxToAppend      = pxNVTSStrip;
                uLongestLength  = pxNVTSStrip->m_faces.size();
                uIndexToAppend  = uStrip;
            }

            // We need to start a new output strip iff we didn't find an
            // appropriate candidate.
            if( pxToAppend )
            {
                bStartNewPPStrip = false;
            }
            else
            {
                s_xPostProcessedStrips.push_back( pxCurPPStrip );
                pxCurPPStrip = NULL;
            }
        }

        if( bStartNewPPStrip )
        {
            // Find next longest strip that hasn't yet been dealt with.
            unsigned uLongestLength = 0;
            for( unsigned uStrip = 0; uStrip < uNumStrips; ++uStrip )
            {
                if( abAppended[uStrip] ) { continue; }

                const NVTriStrip::NvStripInfo* pxNVTSStrip = s_xNVTSStripiferOutput[uStrip];
                if( pxNVTSStrip->m_faces.size() > uLongestLength )
                {
                    pxToAppend      = pxNVTSStrip;
                    uLongestLength  = pxNVTSStrip->m_faces.size();
                    uIndexToAppend  = uStrip;
                }
            }

            Asura_Assert( pxToAppend, "" );

            // Allocate a new output strip and initialise it with this
            // texture ID.
            pxCurPPStrip = new PostProcessedStrip();
            if( !pxCurPPStrip )
            {
                PS2Strip::LogError( "ERROR: Out of memory!\n" );
                delete[] abAppended;
                return false;
            }
            pxCurPPStrip->iMaterialID = NVTS_ExtractMaterialID( pxToAppend );
            pxCurPPStrip->uTriCount   = 0;
        }

        // Append this strip.
        Asura_Assert( pxCurPPStrip, "" );
        Asura_Assert( pxToAppend, "" );
        NVTS_AppendStripperStripToPostProcessedStrip( pxCurPPStrip, pxToAppend );

        // Record that we've dealt with it.
        Asura_Assert( uIndexToAppend != UINT_MAX, "" );
        abAppended[uIndexToAppend] = true;
    }

    // Assuming we had some strips in the first place, we'll have a left-
    // over output strip now.
    Asura_Assert( pxCurPPStrip, "" );
    s_xPostProcessedStrips.push_back( pxCurPPStrip );

#if _DEBUG
    // Paranoia...
    for( unsigned u = 0; u < uNumStrips; ++u )
    {
        Asura_Assert( abAppended[u], "" );
    }
#endif

    delete[] abAppended;
    abAppended = NULL;

    // Output progress message.
    PS2Strip::LogInfo( "done.\n" );

    u_int uMinPerStrip         = UINT_MAX;
    u_int uMaxPerStrip         = 0;
    u_int uNumSingleTriStrips  = 0;
    u_int uTotalOutputVertices = 0;

    s_auNumPostProcessedStripsPerMaterial = new u_int[1 + ASURA_MAX_NO_OF_MATERIALS];
    memset( s_auNumPostProcessedStripsPerMaterial, 0, (1 + ASURA_MAX_NO_OF_MATERIALS) * sizeof(u_int) );

    // Gather stats.
    const unsigned uNumPPStrips = s_xPostProcessedStrips.size();
    for( unsigned u = 0; u < uNumPPStrips; ++u )
    {
        const unsigned uNumVerticesInStrip = s_xPostProcessedStrips[u]->xQuadletIndices.size();
        Asura_Assert( uNumVerticesInStrip >= 3, "" );

        uTotalOutputVertices += uNumVerticesInStrip;

        if( 3 == uNumVerticesInStrip ) { ++uNumSingleTriStrips; }

        if( uNumVerticesInStrip > uMaxPerStrip ) { uMaxPerStrip = uNumVerticesInStrip; }
        if( uNumVerticesInStrip < uMinPerStrip ) { uMinPerStrip = uNumVerticesInStrip; }

        int iMaterialArrayIdx = Asura_Material_System::TranslateMaterialIDToArrayIdx( s_xPostProcessedStrips[u]->iMaterialID );
        ++s_auNumPostProcessedStripsPerMaterial[iMaterialArrayIdx + 1];
    }

    if( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_ENV )
    {
        s_uEnvNumStripsFromStripifier          += uNumStrips;
        s_uEnvNumSingleTriStripsFromStripifier += uNumSingleTriStrips;
    }

    //  Output statistics.
    PS2Strip::LogInfo(
        "    (Generated %u strips; avg %.1f vertices per strip; min %u; max %u; %u single tri strips.)\n",
        uNumPPStrips,
        static_cast<float>(uTotalOutputVertices) / static_cast<float>(uNumPPStrips),
        uMinPerStrip,
        uMaxPerStrip,
        uNumSingleTriStrips
    );

    return true;
}

bool PS2Strip_ExporterImplementation::NVTS_Stripify( void )
{
    // Output progress message.
    PS2Strip::LogInfo( "Stripifying (using NVTriStrip algorithm)... " );

    u_int uNumberOfEdgesWithTooManyTriangles = 0;

    {
        NVTriStrip::NvStripifier xStripifier;

        u_int uStartIdx = 0;
        for( u_int u = 0; u < 1 + ASURA_MAX_NO_OF_MATERIALS; ++u )
        {
            u_int uNumTrisForThisMaterial = s_auUniqueMeshTrisPerMaterial[u];
            if( !uNumTrisForThisMaterial ) { continue; }

            xStripifier.Stripify( &s_axUniqueMeshTris[uStartIdx],
                                  uNumTrisForThisMaterial,
                                  PS2Strip::GetMaxStripLength(s_pxInputSkin) - 2,
                                  s_uNumUniqueMeshVtxIdxQuadlets - 1,
                                  false,
                                  &s_xNVTSStripiferOutput,
                                  NULL );

            uNumberOfEdgesWithTooManyTriangles += xStripifier.GetNumberOfEdgesWithTooManyTriangles();

            uStartIdx += uNumTrisForThisMaterial;
        }

        xStripifier.ReorderFirstFaceOfStrips( &s_xNVTSStripiferOutput );
    }

	s_uTotalEdgesWithTooManyTriangles += uNumberOfEdgesWithTooManyTriangles;
    
    // Output progress message.
    PS2Strip::LogInfo( "done.\n" );

    // Output statistics.
    {
        float fAvgVerticesPerStrip;
        unsigned uMin = UINT_MAX, uMax = 0;

        // Gather stats.
        {
            unsigned uTotalNumVertices = 0;

            const unsigned uNumStrips = s_xNVTSStripiferOutput.size();
            for( unsigned u = 0; u < uNumStrips; ++u )
            {
                const unsigned uNumVertices = s_xNVTSStripiferOutput[u]->m_faces.size() + 2;

                if( uNumVertices > uMax ) { uMax = uNumVertices; }
                if( uNumVertices < uMin ) { uMin = uNumVertices; }

                uTotalNumVertices += uNumVertices;
            }

            fAvgVerticesPerStrip = static_cast<float>(uTotalNumVertices) / static_cast<float>(uNumStrips);
        }

        PS2Strip::LogInfo( "    (Generated %u strips; avg %.1f vertices per strip; min %u; max %u.)\n",
                           s_xNVTSStripiferOutput.size(),
                           fAvgVerticesPerStrip,
                           uMin,
                           uMax );

 		if(uNumberOfEdgesWithTooManyTriangles)
		{
        	PS2Strip::LogInfo( "    (Warning: found %d edge(s) with more than 2 triangles)\n",uNumberOfEdgesWithTooManyTriangles );
		}
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     PS2Strip_ExporterImplementation::WriteEnvDmaDataForStrip().
// PARAMETERS   pxStrip        - Strip whose data should be written.
//              xChainWriter   - DMA chain writer instance for outputting /
//                               splitting data. It's assumed that the
//                               instance has already been initialised.
//              bIsDualTexture - 
// RETURNS      True iff successful.
// DESCRIPTION  Writes out a DMA chain for the specified strip.
// AUTHOR       MP - 08.01.2003.
//
//////////////////////////////////////////////////////////////////////////////
bool PS2Strip_ExporterImplementation::WriteEnvDmaDataForStrip( const EnvStrip* pxStrip, Tools_PS2_DMAChainWriter& xChainWriter, bool bIsDualTexture )
{
    #define ADD_VIFCODE( vif_code ) \
        if( !xChainWriter.AddVifCode(vif_code) ) { return false; }

    #define ADD_DATA( data, num_bytes ) \
        if( !xChainWriter.AddData(data, num_bytes) ) { return false; }

    const u_int uNumVertices = pxStrip->m_uVertexCount;

    // Positions.
    ADD_VIFCODE( PS2STRIP_VIF1_SET_UNPACKR(0, uNumVertices, PS2STRIP_VIF1_CMD_V3_32, false) );
    ADD_DATA( pxStrip->m_pfPositions, uNumVertices * 12 );
    
    // UVs.
    if( bIsDualTexture )
    {
        ADD_VIFCODE( PS2STRIP_VIF1_SET_UNPACKR(1, uNumVertices, PS2STRIP_VIF1_CMD_V4_16, false) );
        ADD_DATA( pxStrip->m_piTexCoords, uNumVertices << 3 );
    }
    else
    {
        ADD_VIFCODE( PS2STRIP_VIF1_SET_UNPACKR(1, uNumVertices, PS2STRIP_VIF1_CMD_V2_16, false) );
        ADD_DATA( pxStrip->m_piTexCoords, uNumVertices << 2 );
    }

    // RGB(A)s. Note that STMASK, etc., will already have been setup, before
    // the first strip, if required.
    if( bIsDualTexture )
    {
        ADD_VIFCODE( PS2STRIP_VIF1_SET_UNPACKR(2, uNumVertices, PS2STRIP_VIF1_CMD_V4_8, false) | PS2STRIP_VIF1_UNPACK_USN_FLAG );
        ADD_DATA( pxStrip->m_pucRGBs, uNumVertices << 2 );
    }
    else
    {
        ADD_VIFCODE( PS2STRIP_VIF1_SET_UNPACKR(2, uNumVertices, PS2STRIP_VIF1_CMD_V3_8 | PS2STRIP_VIF1_MASK_PROCESS_ENABLE, false) | PS2STRIP_VIF1_UNPACK_USN_FLAG );
        ADD_DATA( pxStrip->m_pucRGBs, uNumVertices * 3 );
    }

    // Normals and ADC bits.
    ADD_VIFCODE( PS2STRIP_VIF1_SET_UNPACKR(3, uNumVertices, PS2STRIP_VIF1_CMD_V4_5, false) );
    ADD_DATA( pxStrip->m_pusNrmsAndADCBits, uNumVertices << 1 );

    // Render this strip. We need to specify the vertex count in the ITOP
    // register.
    ADD_VIFCODE( PS2STRIP_VIF1_SET_ITOP(uNumVertices, false) );
    ADD_VIFCODE( PS2STRIP_VIF1_SET_MSCNT(false) );

    return true;

    #undef ADD_DATA
    #undef ADD_VIFCODE
}
// End function PS2Strip_ExporterImplementation::WriteEnvDmaDataForStrip().

static u_int g_uModuleOffsetForCallback = 0xFFFFFFFFu;
bool PS2Strip_ExporterImplementation::EnvDmaDataEndChainCallback( const Tools_PS2_DMAChainWriter::EndChainInfo& xEndChainInfo,
                                                                  void** ppvDataToBeWritten,
                                                                  u_int* puNumBytesToBeWritten )
{
    static Asura_PS2_Env_ExportData_DmaChainHdr xHdr;

    Asura_Assert( UINT_MAX != g_uModuleOffsetForCallback, "" );

    xHdr.uModuleOffset       = g_uModuleOffsetForCallback;
    xHdr.uNumPieces          = xEndChainInfo.uNumPiecesUsed;
    xHdr.uNumQWsInFirstPiece = xEndChainInfo.uQWsInFirstPiece;
    xHdr.uNumQWsInLastPiece  = xEndChainInfo.uQWsInLastPiece;

    *ppvDataToBeWritten = &xHdr;
    *puNumBytesToBeWritten = sizeof(Asura_PS2_Env_ExportData_DmaChainHdr );
    return true;
}

//////////////////////////////////////////////////////////////////////////////
//
// FUNCTION     PS2Strip_ExporterImplementation::WriteEnvDmaDataForModule().
// PARAMETERS   hFile               - File to write data to.
//              xChainWriter        - DMA chain writer instance for
//                                    outputting / splitting data. This
//                                    method assumes the instance has already
//                                    been initialised.
//              uModuleOffset       - Value to be written to the texture hdr's
//                                    uModuleOffset field.
//              iMaterialID         - Material used by (all) strips.
//              pxModuleStrips      - Module whose strips should be written
//                                    out.
//              puNumDmaChainPieces - 
//              puTotalBytesWritten -
// RETURNS      True iff successful.
// DESCRIPTION  Writes out DMA chain for a given module's strips (of a given
//              texture.)
// AUTHOR       MP - 12.11.2002.
//
//////////////////////////////////////////////////////////////////////////////
bool PS2Strip_ExporterImplementation::WriteEnvDmaDataForModule( HANDLE hFile,
                                                                Tools_PS2_DMAChainWriter& xChainWriter,
                                                                u_int uModuleOffset,
                                                                int iMaterialID,
                                                                const ModuleStripsWithSameMaterial* pxModuleStrips,
                                                                u_int* puNumDmaChainPieces,
                                                                u_int* puTotalBytesWritten )
{
    #define CHECK( e ) { if( !(e) ) { return false; } }

    const u_int uNumStrips = pxModuleStrips->xStrips.size();
    if( !uNumStrips )
    {
        Asura_Assert( false, "No strips in this module with this texture." );
        return false;
    }

    const bool bIsDualTexture = IsMaterialDualTexture( iMaterialID );

#ifdef ASURA_ASSERT
    const DWORD uInitialNumBytesWritten = SetFilePointer( hFile, 0, NULL, FILE_CURRENT );
    if( INVALID_SET_FILE_POINTER == uInitialNumBytesWritten )
    {
        PS2Strip::LogError( "ERROR: Couldn't determine num bytes written to file so far. Reason: %s\n", PS2Strip::GetLastSystemError() );
        return false;
    }
#endif

    CHECK( xChainWriter.BeginChain() );

    // To save doing it per-strip, setup STCYCL, STMASK and STCOL here. The
    // latter two are used just to get 128 into the alpha components.
    {
        CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(1, 4, false)) );

        // If this is a dual texture material, we don't need to bother
        // playing around with the MASK register and so on since we'll
        // actually be filling in the other components.
        if( !bIsDualTexture )
        {
            const PS2Strip_VIF1_Mask xMask =
            {
                0, 0, 0, 2, // Write 128 (from STCOL - see below) into alpha component of RGBA (and use UNPACK data for RGB.)
                3, 3, 3, 3, // Only the first matters since WL (in STCYCL) is 1. Just set to mask.
                3, 3, 3, 3, //                          "
                3, 3, 3, 3, //                          "
            };
            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STMASK(false)) );
            CHECK( xChainWriter.AddDataWord(reinterpret_cast<const u_int&>(xMask)) );

            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCOL(false)) );
            CHECK( xChainWriter.AddDataWord(0x00000080) );
            CHECK( xChainWriter.AddDataWord(0) ); // Only the
            CHECK( xChainWriter.AddDataWord(0) ); // first one
            CHECK( xChainWriter.AddDataWord(0) ); // is used.
        }
    }

    for( u_int uStrip = 0; uStrip < uNumStrips; ++uStrip )
    {
        CHECK( WriteEnvDmaDataForStrip(pxModuleStrips->xStrips[uStrip], xChainWriter, bIsDualTexture) );
    }

    g_uModuleOffsetForCallback = uModuleOffset;
    CHECK( xChainWriter.EndChain() );

#ifdef ASURA_ASSERT
    const DWORD uNewNumBytesWritten = SetFilePointer( hFile, 0, NULL, FILE_CURRENT );
    if( INVALID_SET_FILE_POINTER == uNewNumBytesWritten )
    {
        PS2Strip::LogError( "ERROR: Couldn't determine num bytes written to file so far. Reason: %s\n", PS2Strip::GetLastSystemError() );
        return false;
    }
#endif

    // Each chain section - including the hdr - should be QW aligned.
    Asura_Assert( !(xChainWriter.GetTotalBytesWrittenForLastChain() & 15), "Chain data wasn't QW aligned." );

    // Verify we wrote the number of bytes claimed by the env writer.
    Asura_Assert( uNewNumBytesWritten > uInitialNumBytesWritten, "" );
    Asura_Assert( uNewNumBytesWritten - uInitialNumBytesWritten == xChainWriter.GetTotalBytesWrittenForLastChain(), "Env writer reported incorrect byte count." );

    // Report number of DMA chain pieces required by this module.
    *puNumDmaChainPieces = xChainWriter.GetNumPiecesForLastChain();

    // Update total byte count.
    *puTotalBytesWritten += xChainWriter.GetTotalBytesWrittenForLastChain();

    // Verify that we're QW aligned.
    Asura_Assert( !(*puTotalBytesWritten & 15), "Should be QW aligned at this point." );

    return true;

    #undef CHECK
}
// End function PS2Strip_ExporterImplementation::WriteEnvDmaDataForModule().

bool PS2Strip_ExporterImplementation::OutputEnv( void )
{
    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OUTPUT );

    Asura_Assert( s_uEnvNumUniqueMaterialIDs > 0, "" );
    Asura_Assert( s_axMaterialIDToModuleStripsMaps != NULL, "" );

    // Update export status.
    Timing_SetProfileArea( TIMING_PROFILEAREA_REFRESH_DISPLAY );
    PS2Strip::UpdateCompletionStatus( 0.0f, "Writing stripped env..." );
    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OUTPUT );

    // Output progress message to log.
    PS2Strip::LogInfo( "\nWriting env data... " );

    const Asura_Environment* const pxEnv = Asura_Environment_System::GetEnvironment();
    Asura_Assert( pxEnv, "How'd we get this far without an environment?" );

    // Open file.
    HANDLE hPS2EnvFile;
    hPS2EnvFile = PS2Strip::GetOutputHandle();
    Asura_Assert( INVALID_HANDLE_VALUE != hPS2EnvFile, "" ); // Shouldn't have got this far if failed to open file.

#ifdef ASURA_ASSERT
    // Record initial file ptr.
    const u_int uInitialFilePtr = SetFilePointer( hPS2EnvFile, 0, NULL, FILE_CURRENT );
    if( INVALID_SET_FILE_POINTER == uInitialFilePtr )
    {
        PS2Strip::LogError( "ERROR: Couldn't get file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
        return false;
    }
	// (AW - 24.02.2005) - Stripped environments now need to start on QW boundaries. This is because
	// BHD's technique of asynchronously loading the front-end environment into a memory buffer
	// before processing requires it. If this assertion fires then investigate Tools_Level_Export::PadToQWordBoundary.
	Asura_Assert(!(uInitialFilePtr & 15), "Stripped environment not starting on QW boundary. This will break levels that load asynchronously");
#endif

    unsigned uTotalNumBytesWritten = 0;

    #define PAD_TO_QW_ALIGNMENT( num_bytes_of_padding_reqd ) \
    { \
        /* Copy pad count to prevent evaluating an expression with side- \
           effects multiple times. */ \
        const int iPad = (num_bytes_of_padding_reqd); \
        \
        Asura_Assert( 0 < iPad && !(iPad & 3) && iPad <= 12, "Invalid num pad bytes." ); \
        \
        /* Should be always be word aligned in the file. */ \
        Asura_Assert( !(uTotalNumBytesWritten & 3), "Should always be word aligned." ); \
        \
        /* Output padding. */ \
        WRITE_DATA( hPS2EnvFile, uTotalNumBytesWritten, "            ", iPad ); \
    }
      
    #define RECORD_FILE_PTR( var ) \
    { \
        const u_int uFilePtr = SetFilePointer( hPS2EnvFile, 0, NULL, FILE_CURRENT ); \
        if( INVALID_SET_FILE_POINTER == uFilePtr ) \
        { \
            PS2Strip::LogError( "ERROR: Couldn't get file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() ); \
            return false; \
        } \
        \
        /* At each point, our total-bytes-written count should be the same \
           as the current file pointer (assuming this is always called when \
           we're at EOF. */ \
        Asura_Assert( uFilePtr == uInitialFilePtr + uTotalNumBytesWritten, "Total bytes written count is incorrect." ); \
        \
        (var) = uFilePtr; \
    }

    // Leave a gap for the chunk header. Record the cur file ptr first so we
    // can return to this point and fill it out later.
    u_int uFilePtrForChunkHdr;
    RECORD_FILE_PTR( uFilePtrForChunkHdr );
    SKIP_DATA( hPS2EnvFile, uTotalNumBytesWritten, sizeof(Asura_Chunk_ResourceFile) );

    const u_int uSizeOfResourceFilenameInBytes = strlen(ENV_RESOURCE_FILENAME) + 1;
    Asura_Assert( !(uSizeOfResourceFilenameInBytes & 3), "" );

    u_int uSizeOfResourceFilenamePlusPaddingInBytes = uSizeOfResourceFilenameInBytes;
    
    // We're about to write the resource filename. Before doing so, write out
    // padding so that the data will be QW aligned to speed up loading at
    // runtime. We'll do this by prepending spaces to the resource filename.
    {
        const u_int uRem = ((uTotalNumBytesWritten + uSizeOfResourceFilenameInBytes) & 15);
        if( uRem )
        {
            const u_int uPad = 16 - uRem;
            PAD_TO_QW_ALIGNMENT( uPad );
            uSizeOfResourceFilenamePlusPaddingInBytes += uPad;
        }
    }

    // Write actual resource filename.
    WRITE_DATA( hPS2EnvFile, uTotalNumBytesWritten, ENV_RESOURCE_FILENAME, uSizeOfResourceFilenameInBytes );
    Asura_Assert( !(uTotalNumBytesWritten & 15), "Should be QW aligned at this point." );

    const u_int uRgnOffsetOrigin = uTotalNumBytesWritten;

    // Leave a gap for the env header. Record the cur file ptr first so we
    // can return to this point and fill it out later.
    u_int uFilePtrForEnvHdr;
    RECORD_FILE_PTR( uFilePtrForEnvHdr );
    SKIP_DATA( hPS2EnvFile, uTotalNumBytesWritten, sizeof(Asura_PS2_Env_ExportData_EnvHdr) );

	int iUntexturedMaterialID = Tools_Level_Export::GetUntexturedMaterialIndex();

    int aiToEnvConversionList[ASURA_MAX_NO_OF_MATERIALS];

    // Prepare conversion list from all modules in environment. This needs to
    // match the process in the main .asr save, which is where the conversion
    // list used at runtime will get generated.
    {
        List<Asura_Mesh*> xMeshList;
        for( int i = 0; i < Asura_Module_System::GetNumberOfModules(); ++i )
        {
            xMeshList.add_entry_end( Asura_Module_System::GetMeshForModule(i) );
        }

        Tools_SaveAsura::PrepareMeshMaterialConversionLists( xMeshList, aiToEnvConversionList, NULL );
    }

    // Output texture and module headers.
    for( unsigned uMap = 0; uMap < s_uEnvNumUniqueMaterialIDs; ++uMap )
    {
        // Each texture header should be on a QW boundary.
        Asura_Assert( !(uTotalNumBytesWritten & 15), "" );

        const MaterialIDToModuleStripsMap& xMap = s_axMaterialIDToModuleStripsMaps[uMap];
        const int iMaterialID = xMap.iMaterialID;

        // Verify env is sorted by texture according to PS2 render order.
        if( uMap )
        {
            Asura_Assert( GetRenderPassFromMaterialID(iMaterialID) >= GetRenderPassFromMaterialID(s_axMaterialIDToModuleStripsMaps[uMap - 1].iMaterialID), "Env isn't sorted by material." );
        }

        // Extract texture ID.
		bool bTextureAvailable = false;
		if(ASURA_MATERIAL_NONE != xMap.iMaterialID)
		{
			//check to make sure the texture beingused by this material exists. If not
			//we will use the untextured material instead. RAR 08/09/03
			if(Asura_Texture_Management::IsTextureLoaded(Asura_Material_System::GetTextureID(iMaterialID)))
			{
				bTextureAvailable = true;				
			}
		}

        int iMaterialArrayIdx = Asura_Material_System::TranslateMaterialIDToArrayIdx( iMaterialID );
        Asura_Assert( ASURA_MATERIAL_NONE == iMaterialArrayIdx || (0 <= iMaterialArrayIdx && iMaterialArrayIdx < ASURA_MAX_NO_OF_MATERIALS), "" );

        int iChunkMatID = ( bTextureAvailable ? aiToEnvConversionList[iMaterialArrayIdx] : iUntexturedMaterialID );

        // If we have untextured polys then we should already have setup the
        // fake material by now.
        Asura_Assert( ASURA_MATERIAL_NONE != iChunkMatID, "Should have setup a fake untextured material index by this point." );

        // Extract module count.
        const unsigned uNumModulesWithThisMaterial = xMap.xModuleData.size();

        // Output hdr.
        {
            // We use shorts to store the material ID and module count. This
            // should still be way more than we'll ever need but let's check.
            Asura_Assert( iChunkMatID <= 32767, "Too many unique materials in this environment." );
            Asura_Assert( uNumModulesWithThisMaterial <= 65535, "Too many modules using this material." );

            Asura_PS2_Env_ExportData_MaterialHdr xMatHdr;
            xMatHdr.sMaterialID  = static_cast<short>( iChunkMatID );
            xMatHdr.usNumModules = static_cast<u_short>( uNumModulesWithThisMaterial );

            WRITE_DATA( hPS2EnvFile, uTotalNumBytesWritten, &xMatHdr, sizeof(xMatHdr) );
        }

        // Output module hdrs.
        {
            Asura_PS2_Env_ExportData_ModuleHdr xTmpModuleHdr;
            for( u_int u = 0; u < uNumModulesWithThisMaterial; ++u )
            {
                const ModuleStripsWithSameMaterial* pxModuleStrips = xMap.xModuleData[u];

                // Calculate total tri count for the strips in this module
                // that use the current texture.
                u_int uTotalTris = 0;
                for( u_int v = 0; v < pxModuleStrips->xStrips.size(); ++v )
                {
                    uTotalTris += pxModuleStrips->xStrips[v]->m_uTriCount;
                }

                // We use shorts to store the module index and tri count.
                // This should still be way more than we'll ever need but
                // let's check.
                Asura_Assert( pxModuleStrips->iModuleIdx <= 32767, "Module idx is too high." );
                Asura_Assert( uTotalTris <= 65535, "Too many tris in this module for cur material." );

                // Fill out hdr.
                xTmpModuleHdr.sModuleIdx = static_cast<short>(pxModuleStrips->iModuleIdx);
                xTmpModuleHdr.usNumTris  = static_cast<u_short>(uTotalTris);

                WRITE_DATA( hPS2EnvFile, uTotalNumBytesWritten, &xTmpModuleHdr, sizeof(xTmpModuleHdr) );
            }
        }

        // Pad to QW alignment if necessary.
        if( uTotalNumBytesWritten & 15 )
        {
            PAD_TO_QW_ALIGNMENT( 16 - (uTotalNumBytesWritten & 15) );
        }
    }
    // End texture header / module index list output.

    // Leave a gap for the region offset table. Record the cur file ptr first
    // so we can return to this point and fill it out later.
    u_int uFilePtrForRgnOffsetTbl;
    RECORD_FILE_PTR( uFilePtrForRgnOffsetTbl );
    SKIP_DATA( hPS2EnvFile, uTotalNumBytesWritten, ASURA_ENV_MAX_REGIONS * sizeof(u_int) );

    // We need to keep a record of each region's QW offset from the beginning
    // of the file and also how many DMA chain pieces are required for each.
    u_int auRegionOffsets[ASURA_ENV_MAX_REGIONS];
    u_int auNumDmaChainPiecesPerRegion[ASURA_ENV_MAX_REGIONS];
    for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
    {
        auNumDmaChainPiecesPerRegion[u] = 0;
    }

    // Prepare env writer instance.
    Tools_PS2_DMAChainWriter xChainWriter;
    if( !xChainWriter.Init(hPS2EnvFile) )
    {
        return false;
    }
    xChainWriter.SetMaxQWCPerTag( ASURA_PS2_ENV_NUM_QWS_PER_DMA_PIECE - 1 );
    xChainWriter.SetTagTypeForNonFinalPieces( PS2STRIP_DMA_TAG_TYPE_NEXT );
    xChainWriter.SetEndChainCallback( EnvDmaDataEndChainCallback );

    // Some stuff for the progress percentage. Note that there will be lots
    // of duplicates counted here because we visit each modules multiple
    // times.
    u_int uTotalModules = 0;
    for( unsigned uMap = 0; uMap < s_uEnvNumUniqueMaterialIDs; ++uMap )
    {
        uTotalModules += s_axMaterialIDToModuleStripsMaps[uMap].xModuleData.size();
    }

    // Iterate through regions.
    u_int uCumModules = 0;
    for( u_int uRegion = 0; uRegion < ASURA_ENV_MAX_REGIONS; ++uRegion )
    {
        // Count number of unique materials in this region.
        u_int uNumUniqueMaterialsInRegion = 0;
        for( unsigned uMap = 0; uMap < s_uEnvNumUniqueMaterialIDs; ++uMap )
        {
            const MaterialIDToModuleStripsMap& xMap = s_axMaterialIDToModuleStripsMaps[uMap];
            
            for( u_int uModule = 0; uModule < xMap.xModuleData.size(); ++uModule )
            {
                if( pxEnv->GetModuleRegion(xMap.xModuleData[uModule]->iModuleIdx) == uRegion )
                {
                    ++uNumUniqueMaterialsInRegion;
                    break;
                }
            }
        }

        // If this region is empty, record this and then continue to the next
        // one.
        if( !uNumUniqueMaterialsInRegion )
        {
            auRegionOffsets[uRegion] = 0xFFFFFFFFu;
            continue;
        }

        // Pad to QW alignment if necessary.
        if( uTotalNumBytesWritten & 15 )
        {
            PAD_TO_QW_ALIGNMENT( 16 - (uTotalNumBytesWritten & 15) );
        }

        // Update offset table.
        Asura_Assert( !(uTotalNumBytesWritten & 15), "Should be QW aligned at this point." );
        Asura_Assert( !((uTotalNumBytesWritten - uRgnOffsetOrigin) & 15), "Should be QW aligned at this point." );
        auRegionOffsets[uRegion] = ((uTotalNumBytesWritten - uRgnOffsetOrigin) >> 4);

        // Record file ptr and leave space for region hdr.
        u_int uFilePtrForRgnHdr;
        RECORD_FILE_PTR( uFilePtrForRgnHdr );
        SKIP_DATA( hPS2EnvFile, uTotalNumBytesWritten, sizeof(Asura_PS2_Env_ExportData_RgnHdr) );

        // Pad to QW alignment.
        Asura_Assert( uTotalNumBytesWritten & 15, "" );
        PAD_TO_QW_ALIGNMENT( 16 - (uTotalNumBytesWritten & 15) );

        // Iterate through this region's materials.
        for( unsigned uMap = 0; uMap < s_uEnvNumUniqueMaterialIDs; ++uMap )
        {
            const MaterialIDToModuleStripsMap& xMap = s_axMaterialIDToModuleStripsMaps[uMap];
            
            // Count modules in this region that use this material.
            u_int uNumModulesWithThisMaterialInThisRegion = 0;
            for( u_int uModule = 0; uModule < xMap.xModuleData.size(); ++uModule )
            {
                if( pxEnv->GetModuleRegion(xMap.xModuleData[uModule]->iModuleIdx) == uRegion )
                {
                    ++uNumModulesWithThisMaterialInThisRegion;
                }
            }

            if( !uNumModulesWithThisMaterialInThisRegion ) { continue; }

            // Output material header.
            {
                // Each material header should be at a QW boundary. This is
                // because we pad after this region's header and then the end
                // of each DMA chain will be at a QW boundary anyway.
                Asura_Assert( !(uTotalNumBytesWritten & 15), "" );

                Asura_PS2_Env_ExportData_RgnMaterialHdr xMatHdr;
                xMatHdr.uMaterialOffset = uMap;
                xMatHdr.uNumModules     = uNumModulesWithThisMaterialInThisRegion;
                
                WRITE_DATA( hPS2EnvFile, uTotalNumBytesWritten, &xMatHdr, sizeof(xMatHdr) );

                // Pad to QW boundary. Each module's DMA chain header needs
                // to be on one.
                PAD_TO_QW_ALIGNMENT( 16 - sizeof(xMatHdr) );
            }
            // End material header output.

            // Output each module's DMA chain.
            for( u_int uModule = 0; uModule < xMap.xModuleData.size(); ++uModule )
            {
                if( pxEnv->GetModuleRegion(xMap.xModuleData[uModule]->iModuleIdx) == uRegion )
                {
                    // We should be at a QW boundary before each module's
                    // header. This is because we pad after the material hdr
                    // and then each section of DMA data ends at a QW
                    // boundary.
                    Asura_Assert( !(uTotalNumBytesWritten & 15), "Should be QW aligned at this point." );

                    u_int uNumDmaChainPieces;
                    if( !WriteEnvDmaDataForModule(hPS2EnvFile,
                                                  xChainWriter,
                                                  uModule,
                                                  xMap.iMaterialID,
                                                  xMap.xModuleData[uModule],
                                                  &uNumDmaChainPieces,
                                                  &uTotalNumBytesWritten) )
                    {
                        return false;
                    }

                    // We should be at QW alignment after each module's DMA
                    // chain.
                    Asura_Assert( !(uTotalNumBytesWritten & 15), "Should be QW aligned at this point." );

                    // Update DMA chain piece count for this region.
                    Asura_Assert( uNumDmaChainPieces, "" );
                    auNumDmaChainPiecesPerRegion[uRegion] += uNumDmaChainPieces;

                    // Time to update progress msg?
                    if( !(++uCumModules & 15) || uCumModules == uTotalModules )
                    {
                        float fPercent = Asura_Maths::Min( 99.0f, (100.0f * uCumModules) / uTotalModules );
                        Timing_SetProfileArea( TIMING_PROFILEAREA_REFRESH_DISPLAY );
                        PS2Strip::UpdateCompletionStatus( fPercent );
                        Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OUTPUT );
                    }
                }
            }
            // End module DMA chain output.
        }
        // End loop through this region's materials.

        // Seek back to region hdr.
        if( INVALID_SET_FILE_POINTER == SetFilePointer(hPS2EnvFile, uFilePtrForRgnHdr, NULL, FILE_BEGIN) )
        {
            PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
            return false;
        }

        // Write it.
        {
            Asura_PS2_Env_ExportData_RgnHdr xRgnHdr;
            xRgnHdr.uNumUniqueMaterials = uNumUniqueMaterialsInRegion;
            xRgnHdr.uNumDmaChainPieces  = auNumDmaChainPiecesPerRegion[uRegion];

            int iIgnore = 0;
            WRITE_DATA( hPS2EnvFile, iIgnore, &xRgnHdr, sizeof(xRgnHdr) );
        }

        // And seek back to EOF.
        if( INVALID_SET_FILE_POINTER == SetFilePointer(hPS2EnvFile, 0, NULL, FILE_END) )
        {
            PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
            return false;
        }
    }
    // End loop through regions.

    Asura_Assert( uCumModules == uTotalModules, "" );

    Timing_SetProfileArea( TIMING_PROFILEAREA_REFRESH_DISPLAY );
    PS2Strip::UpdateCompletionStatus( 99.0f );
    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OUTPUT );

    u_int     uTotalDmaChainPieces;
    u_int     uMaxDmaChainPiecesReqd;
    u_int     uRegionsForMaxChainPieces;
    List<int> xRequiringModuleList;
    u_int     uOtherRegionsForMaxChainPieces; // Cooperative.
    List<int> xOtherRequiringModuleList;      //      "

    // Go back and fill out env hdr.
    {
        // Seek back to beginning of env hdr bytes.
        if( INVALID_SET_FILE_POINTER == SetFilePointer(hPS2EnvFile, uFilePtrForEnvHdr, NULL, FILE_BEGIN) )
        {
            PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
            return false;
        }

        Asura_PS2_Env_ExportData_EnvHdr xEnvHdr;

        // Setup hdr.
        {
            xEnvHdr.uNumUniqueMaterials = s_uEnvNumUniqueMaterialIDs;
            xEnvHdr.uVersionID          = ASURA_PS2_ENV_FILE_FORMAT_VERSION_NUMBER;

            {
                u_int uNumFMVs    = 0;
                u_int uIdxOfFirst = 0xFFFFFFFFu;
                for( u_int uMaterial = 0; uMaterial < s_uEnvNumUniqueMaterialIDs; ++uMaterial )
                {
                    const int iTexID = Asura_Material_System::GetTextureID( s_aiEnvUniqueMaterialIDs[uMaterial] );
                    if( Asura_Texture_Management::IsTextureFMV(iTexID) )
                    {
                        if( !uNumFMVs++ )
                        {
                            uIdxOfFirst = uMaterial;
                            break;
                        }
                    }
                }

                xEnvHdr.uNumFmvMaterials       = uNumFMVs;
                xEnvHdr.uIdxOfFirstFmvMaterial = uIdxOfFirst;
            }

            {
                // We'll also provide some helpful statistics about the data.
                u_int uNumDmaChains    = 0;
                uTotalDmaChainPieces   = 0;
                uMaxDmaChainPiecesReqd = 0;
                
                // Count total number of DMA chains.
                for( unsigned uMap = 0; uMap < s_uEnvNumUniqueMaterialIDs; ++uMap )
                {
                    uNumDmaChains += s_axMaterialIDToModuleStripsMaps[uMap].xModuleData.size();
                }

                Asura_Vector_3        xCutsceneStartPos;
                const Asura_Cutscene* pxCutscene;
                Asura_Bounding_Box    xPlayerBB;

                bool bUseCutsceneStartPos = Asura_Cutscene_System::GetCutsceneStartPosition( &xCutsceneStartPos, &pxCutscene );
                if( bUseCutsceneStartPos )
                {
                    if( pxCutscene->m_iCutsceneFlags & AsuraCutsceneFlag_DontExportToPS2 )
                    {
                        bUseCutsceneStartPos = false;
                    }
                }

                if( !bUseCutsceneStartPos )
                {
                    // Setup a BB for use when considering start positions.
                    if( !PS2Strip_Exporter::Project_SetupPlayerBB(&xPlayerBB) )
                    {
                        PS2Strip::LogError( "ERROR: Couldn't get BB for player.\n" );
                        return false;
                    }
                }

                // Count total DMA chain pieces.
                for( u_int uRegion = 0; uRegion < ASURA_ENV_MAX_REGIONS; ++uRegion )
                {
                    uTotalDmaChainPieces += auNumDmaChainPiecesPerRegion[uRegion];
                }

                // Calculate the maximum number of DMA chain pieces that are
                // required simultaneously.
                uMaxDmaChainPiecesReqd = Tools_Mode_Regions::CalculateMaxReqdDmaChainPieces( auNumDmaChainPiecesPerRegion,
                                                                                             &uRegionsForMaxChainPieces,
                                                                                             &xRequiringModuleList,
                                                                                             &uOtherRegionsForMaxChainPieces,
                                                                                             &xOtherRequiringModuleList );

                if( !uMaxDmaChainPiecesReqd ) 
				{ 
                    PS2Strip::LogError( "ERROR: Failed to calculate max simultaneously required DMA chains.\n" );
					return false; 
				}

                // Determine initially required regions.
                // There are issues at runtime with determining which regions
                // are required initially so we'll provide a hint.
                u_int uInitReqdRgns = 0;
                for( int iModule = 0; iModule < Asura_Module_System::GetNumberOfModules(); ++iModule )
                {
                    bool bUse;

                    // (MP - 24.06.2003) If the max reqd chain pieces is
                    // actually all of them, we might as well keep everything
                    // loaded at all times, seeing as we're having to reserve
                    // the memory anyway.
                    if( uMaxDmaChainPiecesReqd == uTotalDmaChainPieces )
                    {
                        bUse = true;
                    }
                    else
                    {
                        const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule( iModule );
                        Asura_Assert( pxMesh->BoundingBoxDataValid, "Module's BB hasn't been setup." );

                        const Asura_Vector_3& xModulePos = Asura_Module_System::GetPositionOfModule( iModule );
                    
                        if( bUseCutsceneStartPos )
                        {
                            bUse = pxMesh->BoundingBox.Contains( xCutsceneStartPos - xModulePos );
                        }
                        else
                        {
                            // See if any of the start positions lie inside this
                            // module.
                            bUse = false;
                            for( int i = 0; i < pxEnv->NumberOfStartPositions; ++i )
                            {
                                if( pxMesh->BoundingBox.Intersects(xPlayerBB, pxEnv->StartPositions[i].Position - xModulePos) )
                                {
                                    bUse = true;
                                    break;
                                }
                            }
                        }
                    }

                    if( bUse ) { uInitReqdRgns |= pxEnv->GetRequiredRegions( iModule ); }
                }
                // End loop through modules to determine initially required regions.

                xEnvHdr.uNumDmaChains          = uNumDmaChains;
                xEnvHdr.uTotalDmaChainPieces   = uTotalDmaChainPieces;
                xEnvHdr.uMaxDmaChainPiecesReqd = uMaxDmaChainPiecesReqd;
                xEnvHdr.uInitReqdRgns          = uInitReqdRgns;
            }
        }
        // End hdr setup.

        // Output. Note that we've already included these bytes in the
        // cumulative count when we skipped the bytes.
        int iIgnore = 0;
        WRITE_DATA( hPS2EnvFile, iIgnore, &xEnvHdr, sizeof(Asura_PS2_Env_ExportData_EnvHdr) );
    }
    // End env hdr write.

    // Go back and fill out region offset table.
    {
        // Seek back to beginning of region hdr bytes.
        if( INVALID_SET_FILE_POINTER == SetFilePointer(hPS2EnvFile, uFilePtrForRgnOffsetTbl, NULL, FILE_BEGIN) )
        {
            PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
            return false;
        }
        
        // Output. Note that we've already included these bytes in the
        // cumulative count when we skipped the bytes.
        int iIgnore = 0;
        WRITE_DATA( hPS2EnvFile, iIgnore, auRegionOffsets, ASURA_ENV_MAX_REGIONS * sizeof(u_int) );
    }
    // End region offset table write.

    // Go back and fill out chunk hdr.
    {
        // Seek back to beginning of chunk hdr bytes.
        if( INVALID_SET_FILE_POINTER == SetFilePointer(hPS2EnvFile, uFilePtrForChunkHdr, NULL, FILE_BEGIN) )
        {
            PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
            return false;
        }

        // Prepare hdr.
        Asura_Chunk_ResourceFile xChunkHdr;
        xChunkHdr.ID       = ASURA_CHUNK_RESOURCEFILE;
        xChunkHdr.Size     = uTotalNumBytesWritten;
        xChunkHdr.Version  = 0;
        xChunkHdr.Flags    = 0;
        xChunkHdr.Type     = ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC;
        xChunkHdr.SubType  = ASURA_RESOURCEFILE_TYPE_PS2_STRIPPEDENV;
        xChunkHdr.FileSize = uTotalNumBytesWritten - (sizeof(xChunkHdr) + uSizeOfResourceFilenamePlusPaddingInBytes);
        
        // Output. Note that we've already included these bytes in the
        // cumulative count when we skipped the bytes.
        int iIgnore = 0;
        WRITE_DATA( hPS2EnvFile, iIgnore, &xChunkHdr, sizeof(xChunkHdr) );
    }
    // End chunk hdr write.

    // Seek back to EOF.
    {
        const u_int uEOF = SetFilePointer( hPS2EnvFile, 0, NULL, FILE_END );
        if( INVALID_SET_FILE_POINTER == uEOF )
        {
            PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
            return false;
        }

        // Check our total byte count.
        Asura_Assert( uEOF == uInitialFilePtr + uTotalNumBytesWritten, "Total byte count is incorrect." );
    }

    Timing_SetProfileArea( TIMING_PROFILEAREA_REFRESH_DISPLAY );
    PS2Strip::UpdateCompletionStatus( 100.0f );
    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OUTPUT );

    #undef RECORD_FILE_PTR
    #undef PAD_TO_QW_ALIGNMENT

    // Output progress message.
    {
        const char* szFmt =
            "done.\n"
            "(Output a total of %u strips, averaging %.1f tris (+ %.1f degenerate) / %.1f vertices per strip; %u single tri strips.)\n"
            "(Stripifier produced %u strips, averaging %.1f tris per strip; %u max tris; %u single tri strips.\n"
            "(Max. allowed strip length was %u tris.)\n";

        const float fAvgTris       = static_cast<float>(s_uEnvNumTris) / static_cast<float>(s_uEnvNumStrips);
        const float fAvgVertices   = static_cast<float>(s_uEnvNumVertices) / static_cast<float>(s_uEnvNumStrips);
        const float fAvgDegenerate = ( fAvgVertices - 2.0f > fAvgTris ? (fAvgVertices - 2.0f) - fAvgTris : 0.0f );

        PS2Strip::LogInfo(
            szFmt,
            s_uEnvNumStrips,
            fAvgTris,
            fAvgDegenerate,
            fAvgVertices,
            s_uEnvNumSingleTriStrips,
            s_uEnvNumStripsFromStripifier,
            static_cast<float>(s_uEnvNumTris) / static_cast<float>(s_uEnvNumStripsFromStripifier),
            s_uEnvMaxLengthFromStripifier,
            s_uEnvNumSingleTriStripsFromStripifier,
            PS2Strip::GetMaxStripLength(s_pxInputSkin) - 2
        );
    }
    // End progress msg output.

    // Output some info about the regions.
    {
        const char* szIntro = "(Num DMA chain pieces required per region: ";
        const char* szSpace = "                                           ";

        char szTmpBfr[1024];
        char* szOut = &szTmpBfr[0];
        
        szOut += sprintf( szOut, "%s", szIntro );
        for( u_int u = 0; u < ASURA_ENV_MAX_REGIONS; ++u )
        {
            szOut += sprintf( szOut, "%4u ", auNumDmaChainPiecesPerRegion[u] );
            if( ((u + 1) & 7) == 0 )
            {
                if( u + 1 < ASURA_ENV_MAX_REGIONS )
                {
                    szOut += sprintf( szOut, "\n%s", szSpace );
                }
                else
                {
                    szOut += sprintf( szOut, ")\n" );
                }
            }
        }

        PS2Strip::LogInfo( szTmpBfr );

        PS2Strip::LogInfo( "(Total chain pieces: %u.)\n"
                           "(Max chain pieces required simultaneously: %u.)\n"
                           "(Reqd rgns for max piece count: ",
                           uTotalDmaChainPieces,
                           uMaxDmaChainPiecesReqd );

        bool bFirst = true;
        for( u_int uRgn = 0; uRgn < ASURA_ENV_MAX_REGIONS; ++uRgn )
        {
            if( uRegionsForMaxChainPieces & (1 << uRgn) )
            {
                PS2Strip::LogInfo( "%s%u", (bFirst ? "" : ", "), uRgn );
                bFirst = false;
            }
        }

        PS2Strip::LogInfo( ".)\n"
                           "(Module(s) that caused this max piece count: " );

        bFirst = true;
        for( CLIF<int> xIt(&xRequiringModuleList); !xIt.done(); xIt.next() )
        {
            PS2Strip::LogInfo( "%s%i (%s)", (bFirst ? "" : ", "), xIt(), Asura_Module_System::GetModuleData(xIt())->Mesh->Name.GetString() );
            bFirst = false;
        }

        PS2Strip::LogInfo( ".)\n" );

        if( Tools_Level_Export::IsPS2MultiplayerExport() && uOtherRegionsForMaxChainPieces )
        {
            PS2Strip::LogInfo( "(Second camera's additional reqd rgns for max piece count: " );

            bFirst = true;
            for( u_int uRgn = 0; uRgn < ASURA_ENV_MAX_REGIONS; ++uRgn )
            {
                if( uOtherRegionsForMaxChainPieces & (1 << uRgn) )
                {
                    PS2Strip::LogInfo( "%s%u", (bFirst ? "" : ", "), uRgn );
                    bFirst = false;
                }
            }

            PS2Strip::LogInfo( ".)\n"
                               "(Module(s) that added these extra reqd regions: " );

            bFirst = true;
            for( CLIF<int> xIt(&xOtherRequiringModuleList); !xIt.done(); xIt.next() )
            {
                PS2Strip::LogInfo( "%s%i (%s)", (bFirst ? "" : ", "), xIt(), Asura_Module_System::GetModuleData(xIt())->Mesh->Name.GetString() );
                bFirst = false;
            }

            PS2Strip::LogInfo( ".)\n" );
        }
    }
    // End region info output.

    return true;
}

void PS2Strip_ExporterImplementation::OutputCumulativeWarnings( void )
{
    const char* szExportDesc = PS2Strip::GetCurExportDescription();
    const char* szAssetName  = PS2Strip::GetCurAssetName();

    Asura_String xDesc = szExportDesc;
    if( szAssetName )
    {
        xDesc += " \"";
        xDesc += szAssetName;
        xDesc += "\"";
    }

    if( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_ENV && s_uEnvNumVerticesWithDodgyTexCoords )
    {
        PS2Strip::LogWarning( "WARNING: %s has %u vert%s with illegal texture coordinates.\n",
                              xDesc.GetString(),
                              s_uEnvNumVerticesWithDodgyTexCoords,
                              (1 == s_uEnvNumVerticesWithDodgyTexCoords ? "ex" : "ices") );
    }

	if( s_uTotalEdgesWithTooManyTriangles )
	{
        PS2Strip::LogWarning( "WARNING: %s has %u edge%s shared by more than 2 triangles.\n",
                              xDesc.GetString(), 
                              s_uTotalEdgesWithTooManyTriangles,
                              (1 == s_uTotalEdgesWithTooManyTriangles ? "" : "s") );
	}

    if( s_uTotalDegenerateSourceTris )
    {
        PS2Strip::LogWarning( "WARNING: %s has %u degenerate tris.\n", xDesc.GetString(), s_uTotalDegenerateSourceTris );
    }

    if( s_uTotalDuplicateSourceTris )
    {
        PS2Strip::LogWarning( "WARNING: %s has %u duplicate tris.\n", xDesc.GetString(), s_uTotalDuplicateSourceTris );
    }
}

bool PS2Strip_ExporterImplementation::OutputObject( void )
{
    #define CHECK( e ) { if( !(e) ) { return false; } }

    // Output progress message.
    PS2Strip::LogInfo( "\nWriting object data... " );

    HANDLE pvOutputFile = PS2Strip::GetOutputHandle();
    Asura_Assert( INVALID_HANDLE_VALUE != pvOutputFile, "" ); // Shouldn't have got this far if failed to open file.

    int   iTotalBytesWritten    = 0;   
    bool  bDoneMultipleMatError = false;
    bool  bDoneTexCoordError    = false;
    bool  bDoneUntexturedError  = false;

    // Leave space for chunk header, which we'll write out at the end once we
    // know how much data various bits required.
    SKIP_DATA( pvOutputFile, iTotalBytesWritten, sizeof(Asura_Chunk_ResourceFile_PS2_StrippedObject) );

    // Resource file name.
    const u_int uResourceFilenameSize = strlen(OBJECT_RESOURCE_FILENAME) + 1;
    Asura_Assert( !(uResourceFilenameSize & 3), "" );
    WRITE_DATA( pvOutputFile, iTotalBytesWritten, OBJECT_RESOURCE_FILENAME, uResourceFilenameSize );

    const int iTotalBytesWrittenBeforeResourceData = iTotalBytesWritten;

    // Object name.
    const int iBytesWrittenBeforeObjName = iTotalBytesWritten;
    const char* szObjectName = PS2Strip::GetCurAssetName();
    const u_int uObjectNameSize = strlen(szObjectName) + 1;
    WRITE_DATA( pvOutputFile, iTotalBytesWritten, szObjectName, uObjectNameSize );
    PAD_TO_ALIGNMENT( pvOutputFile, 4, iTotalBytesWritten, iBytesWrittenBeforeObjName );

    // Shape hdr.
    {
        Asura_PS2_Shape_ShapeHdr xShapeHdr;

        xShapeHdr.usTotalTris   = static_cast<u_short>( s_uNumUniqueMeshTris );
        xShapeHdr.usNumStrips   = static_cast<u_short>( s_xPostProcessedStrips.size() );
        xShapeHdr.fLODThreshold = s_pxCurMesh->m_fMinDistance;

        WRITE_DATA( pvOutputFile, iTotalBytesWritten, &xShapeHdr, sizeof(Asura_PS2_Shape_ShapeHdr) );
    }

    // The DMA chain.
    if( s_uNumUniqueMeshTris )
    {
        Tools_PS2_DMAChainWriter xChainWriter;

        CHECK( xChainWriter.Init(pvOutputFile) );

        CHECK( xChainWriter.BeginChain() );

        int iObjMaterialID = INT_MIN;

        for( u_int uStrip = 0; uStrip < s_xPostProcessedStrips.size(); ++uStrip )
        {
            // For the first two strips we need to fill alpha values. We
            // don't actually support any vertex alpha on objects right now
            // but the render path is the same. However, once we've ensured
            // both buffers are full, we can ignore it.
            if( uStrip < 2 )
            {
                // Temporarily change CYCLE.
                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(4, 0, false)) );

                // On the first strip setup STMASK for the alpha filling.
                if( !uStrip )
                {
                    // Just write to w of colour QW...
                    CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STMASK(false)) );
                    const PS2Strip_VIF1_Mask xMask =
                    {
                        3, 3, 3, 3,
                        3, 3, 3, 3,
                        3, 3, 3, 1,
                        3, 3, 3, 3,
                    };
                    CHECK( xChainWriter.AddDataWord(reinterpret_cast<const u_int&>(xMask)) );

                    // ... and fill it with 1.
                    CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STROW(false)) );
                    const float afFillRow[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
                    CHECK( xChainWriter.AddData(afFillRow, 16) );
                }

                // NB: The CMD here is irrelevant since we're
                // just using fill values and not sending any
                // data.
                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(0,
                                                                         ASURA_PS2_GSM__MAX_VERTS_PER_STRIP * 4,
                                                                         PS2STRIP_VIF1_CMD_V4_32 | PS2STRIP_VIF1_MASK_PROCESS_ENABLE,
                                                                         false)) );

                // Set CYCLE ready for proper strip data.
                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(1, 4, false)) );
            }

            const PostProcessedStrip* pxPPStrip = s_xPostProcessedStrips[uStrip];

            const u_int uNumVertices = pxPPStrip->xQuadletIndices.size();
            Asura_Assert( uNumVertices <= ASURA_PS2_GSM__MAX_VERTS_PER_STRIP, "" );

            #define GET_QUADLET(idx_within_strip) (s_axUniqueMeshVtxIdxQuadlets[pxPPStrip->xQuadletIndices[uVtx] & ~PostProcessedStrip::NEEDS_ADC_BIT])

            // Positions.
            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(0, uNumVertices, PS2STRIP_VIF1_CMD_V3_32, false)) );
            for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
            {
                const u_int uPosIdx = GET_QUADLET(uVtx).GetPosIdx();
                CHECK( xChainWriter.AddData(&s_axMeshPositions[uPosIdx], 12) );
            }

            // STs.
            {
                short aiCoords[ASURA_PS2_GSM__MAX_VERTS_PER_STRIP * 2];
                
                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(1, uNumVertices, PS2STRIP_VIF1_CMD_V2_16, false)) );

                for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
                {
                    const u_int uMatInfoIdx = GET_QUADLET(uVtx).GetMatInfoIdx();
                    const PS2Strip_MatInfo& xMatInfo = s_axMeshMatInfo[uMatInfoIdx];

                    if( !bDoneTexCoordError && !xMatInfo.CheckTexCoords() )
                    {
                        PS2Strip::LogError( "ERROR: Object \"%s\" has texture coordinate outside valid PS2 range.", szObjectName );
                        bDoneTexCoordError = true;
                    }

                    const int iVtxMatID = xMatInfo.m_iMaterialID;
                    if( ASURA_MATERIAL_NONE == iVtxMatID || ASURA_TEXTURE_NONE == Asura_Material_System::GetTextureID(iVtxMatID) )
                    {
                        if( !bDoneUntexturedError )
                        {
                            PS2Strip::LogError( "ERROR: Object \"%s\" has untextured polys.", szObjectName );
                            bDoneUntexturedError = true;
                        }
                    }
                    else
                    {
                        if( INT_MIN == iObjMaterialID )
                        {
                            iObjMaterialID = iVtxMatID;
                        }
                        else if( !bDoneMultipleMatError && iVtxMatID != iObjMaterialID )
                        {
                            PS2Strip::LogError( "ERROR: Object \"%s\" uses multiple materials and/or textures.", szObjectName );
                            bDoneMultipleMatError = true;
                        }
                    }

                    aiCoords[(uVtx << 1) + 0] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMatInfo.m_afU[0] );
                    aiCoords[(uVtx << 1) + 1] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMatInfo.m_afV[0] );
                }
                
                CHECK( xChainWriter.AddData(aiCoords, uNumVertices * 4) );
            }
           
            // Normals and ADC bits.
            {
                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(3, uNumVertices, PS2STRIP_VIF1_CMD_V4_5, false)) );

                PS2Strip_VIF1_V4_5_Data axNrmsAndADCs[ASURA_PS2_GSM__MAX_VERTS_PER_STRIP];
                for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
                {
                    const Asura_Vector_3& xNormal = s_axMeshNormals[GET_QUADLET(uVtx).GetNrmIdx()];
                    axNrmsAndADCs[uVtx].usX = Tools_PS2Strip_ConvertFloatForV45( xNormal.x );
                    axNrmsAndADCs[uVtx].usY = Tools_PS2Strip_ConvertFloatForV45( xNormal.y );
                    axNrmsAndADCs[uVtx].usZ = Tools_PS2Strip_ConvertFloatForV45( xNormal.z );

                    axNrmsAndADCs[uVtx].usW = ( 0 != (pxPPStrip->xQuadletIndices[uVtx] & PostProcessedStrip::NEEDS_ADC_BIT) );
                }

                // NB: This deals with padding to word alignment on
                // our behalf.
                CHECK( xChainWriter.AddData(axNrmsAndADCs, uNumVertices * 2) );
            }

            #undef GET_QUADLET
            
            // Pass vertex count in ITOPS.
            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_ITOP(uNumVertices, false)) );

            // And kick off the microprogram to render this strip.
            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_MSCNT(false)) );
        }
        // End loop through strips.

        CHECK( xChainWriter.EndChain() );        

        // Update cumulative byte count.
        const u_int uBytesWrittenForChain = xChainWriter.GetTotalBytesWrittenForLastChain();
        Asura_Assert( !(uBytesWrittenForChain & 15), "" );
        iTotalBytesWritten += uBytesWrittenForChain;
    }
    // End DMA chain write.

    // Seek back and write out the chunk header.
    {
        SetFilePointer( pvOutputFile, -iTotalBytesWritten, NULL, FILE_CURRENT );

        Asura_Chunk_ResourceFile_PS2_StrippedObject xHdr;

        xHdr.ID       = ASURA_CHUNK_RESOURCEFILE;
        xHdr.Size     = iTotalBytesWritten;
        xHdr.Version  = iASURA_PS2_STRIPPEDOBJECT_CUR_VERSION_NUMBER;
        xHdr.Flags    = 0;
        xHdr.Type     = ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC;
        xHdr.SubType  = ASURA_RESOURCEFILE_TYPE_PS2_STRIPPEDOBJECT;
        xHdr.FileSize = iTotalBytesWritten - iTotalBytesWrittenBeforeResourceData;

        int iIgnore = 0;
        WRITE_DATA( pvOutputFile, iIgnore, &xHdr, sizeof(Asura_Chunk_ResourceFile_PS2_StrippedObject) );

        // Return to EOF.
        SetFilePointer( pvOutputFile, 0, NULL, FILE_END );
    }

    PS2Strip::LogInfo( "done.\n" );

    return true;

    #undef CHECK
}

bool PS2Strip_ExporterImplementation::OutputHierSkin( void )
{
    #define CHECK( e ) { if( !(e) ) { return false; } }

    // See /Asura/PS2/Animation/Asura_PS2_SkinTypes.h for data format overview.

    // Output progress message.
    PS2Strip::LogInfo( "\nWriting hierarchical skin data... " );

    HANDLE hOutputFile = PS2Strip::GetOutputHandle();
    Asura_Assert( INVALID_HANDLE_VALUE != hOutputFile, "" ); // Shouldn't have got this far if failed to open file.

    u_int uNumSections          = s_pxOutputHierSkin->m_uNumSections;
    int   iTotalBytesWritten    = 0;   
    bool  bDoneMultipleTexError = false;
    bool  bDoneTexCoordError    = false;
    bool  bDoneUntexturedError  = false;

    // Special case for when there are no polys in the skin that we care
    // about, eg. if it's being used for some special purpose and has no
    // polys to render.
    if( !s_pxOutputHierSkin->m_uTotalStrips )
    {
        uNumSections = 0;
    }
    Asura_Assert( !s_pxOutputHierSkin->m_uTotalStrips == !s_pxOutputHierSkin->m_uTotalSectionMaterials, "" );

    // Leave space for chunk header, which we'll write out at the end once we
    // know how much data various bits required.
    SKIP_DATA( hOutputFile, iTotalBytesWritten, sizeof(Asura_Chunk_ResourceFile_PS2_StrippedHierSkin) );

    // Output resource filename.
    const u_int uSizeOfResourceFilename = strlen(HIER_SKIN_RESOURCE_FILENAME) + 1;
    Asura_Assert( !(uSizeOfResourceFilename & 3), "" );
    WRITE_DATA( hOutputFile, iTotalBytesWritten, HIER_SKIN_RESOURCE_FILENAME, uSizeOfResourceFilename );

    const int iTotalBytesWrittenBeforeResourceData = iTotalBytesWritten;

    // Write skin name and pad to int alignment.
    const char* szSkinName = PS2Strip::GetCurAssetName();
    const u_int uSkinNameSize = strlen(szSkinName) + 1;
    WRITE_DATA( hOutputFile, iTotalBytesWritten, szSkinName, uSkinNameSize );
    PAD_TO_ALIGNMENT( hOutputFile, 4, iTotalBytesWritten, iTotalBytesWrittenBeforeResourceData );

    const int iNumBytesWrittenBeforeSkinHdr = iTotalBytesWritten;

    // Write skin header.
    {
        Asura_PS2_HierSkin_SkinHdr xSkinHdr;

        xSkinHdr.uAllMaterialFlags = 0;
        xSkinHdr.sTextureID        = 0; // Only expect a single texture per skin so we'll always need to use entry 0 in the conversion list at runtime.
        xSkinHdr.usSectionCount    = static_cast<u_short>( uNumSections );

        int iFirstTexID = ASURA_TEXTURE_NONE;
        for( u_int uSection = 0; uSection < uNumSections; ++uSection )
        {
            const PS2HierSkinSection& xSection = s_pxOutputHierSkin->m_pxSections[uSection];
            for( u_int uMaterial = 0; uMaterial < xSection.m_uNumMaterials; ++uMaterial )
            {
                int iMaterialID = xSection.m_pxMaterials[uMaterial].m_iMaterialID;
                
                int iTextureID = Asura_Material_System::GetTextureID( iMaterialID );
                if( ASURA_TEXTURE_NONE == iTextureID )
                {
                    if( !bDoneUntexturedError )
                    {
                        PS2Strip::LogError( "ERROR: Skin \"%s\" (LOD %u) contains untextured polys.\n", szSkinName, PS2Strip::GetCurSkinLOD() );
                        bDoneUntexturedError = true;
                    }
                }
                else
                {
                    if( ASURA_TEXTURE_NONE == iFirstTexID )
                    {
                        iFirstTexID = iTextureID;
                    }
                    else if( !bDoneMultipleTexError && iTextureID != iFirstTexID )
                    {
                        PS2Strip::LogError( "ERROR: Skin \"%s\" (LOD %u) uses multiple textures.\n", szSkinName, PS2Strip::GetCurSkinLOD() );
                        bDoneMultipleTexError = true;
                    }
                }

                xSkinHdr.uAllMaterialFlags |= Asura_Material_System::GetMaterialFlags( iMaterialID );
            }
        }

        int iByteOffset = sizeof(Asura_PS2_HierSkin_SkinHdr);                    // Skin hdr.
        iByteOffset += uNumSections * sizeof(u_short);                           // Section material counts.
        if( uNumSections & 1 ) { iByteOffset += sizeof(u_short); }               // Pad to word alignment.
        *(reinterpret_cast<int*>(&xSkinHdr.pxSectionMatHdrStart)) = iByteOffset;
        
        iByteOffset += s_pxOutputHierSkin->m_uTotalSectionMaterials * sizeof(Asura_PS2_HierSkin_SectionMatHdr); // Section material hdrs.
        xSkinHdr.uStripDataStart = ((iByteOffset + 127) & ~127) >> 4;                                           // Padding to 8QW alignment.

        WRITE_DATA( hOutputFile, iTotalBytesWritten, &xSkinHdr, sizeof(Asura_PS2_HierSkin_SkinHdr) );
    }

    int iBytesWrittenBeforeSectionMaterialHdrs = -1;
    Asura_PS2_HierSkin_SectionMatHdr axSectionMatHdrs[256];
    Asura_Assert( s_pxOutputHierSkin->m_uTotalSectionMaterials <= 256, "" ); // Just increase the (arbitrary) limit if this ever fires.

    // We can skip a lot of stuff for special empty skins...
    if( uNumSections )
    {
        // Write section material counts.
        for( u_int uSection = 0; uSection < uNumSections; ++uSection )
        {
            u_short usCount = static_cast<u_short>( s_pxOutputHierSkin->m_pxSections[uSection].m_uNumMaterials );
            WRITE_DATA( hOutputFile, iTotalBytesWritten, &usCount, sizeof(u_short) );
        }
        // Pad to word alignment.
        if( uNumSections & 1 )
        {
            const u_short usPad = 0;
            WRITE_DATA( hOutputFile, iTotalBytesWritten, &usPad, sizeof(u_short) );
        }

        // Leave a gap for the section material headers. We'll fill them in
        // once we've written out the DMA chains below and know the QWCs.
        iBytesWrittenBeforeSectionMaterialHdrs = iTotalBytesWritten;
        SKIP_DATA( hOutputFile, iTotalBytesWritten, s_pxOutputHierSkin->m_uTotalSectionMaterials * sizeof(Asura_PS2_HierSkin_SectionMatHdr) );

        // Pad to 8QW alignment before the start of the vertex data.
        PAD_TO_ALIGNMENT( hOutputFile, 128, iTotalBytesWritten, iNumBytesWrittenBeforeSkinHdr );

        int aiToSkinMaterialIDConversionList[ASURA_MAX_NO_OF_MATERIALS];

        // Prepare material ID conversion list ready for use with the section
        // material headers.
        {
            List<Asura_Mesh*> xMeshList;
            for( int i = 0; i < s_pxInputSkin->NumberOfBones; ++i )
            {
                if( s_pxInputSkin->SectionData[i].Mesh )
                {
                    xMeshList.add_entry_end( s_pxInputSkin->SectionData[i].Mesh );
                }
            }

            Tools_SaveAsura::PrepareMeshMaterialConversionLists( xMeshList, aiToSkinMaterialIDConversionList, NULL );
        }

        // Write out the chains.
        {
            Tools_PS2_DMAChainWriter xChainWriter;
            CHECK( xChainWriter.Init(hOutputFile) );

            u_int uCumulativeSectionMatIdx = 0;
            u_int uCumulativeStripIdx      = 0;

            for( u_int uSection = 0; uSection < uNumSections; ++uSection )
            {
                const PS2HierSkinSection& xSection = s_pxOutputHierSkin->m_pxSections[uSection];
                for( u_int uMaterial = 0; uMaterial < xSection.m_uNumMaterials; ++uMaterial )
                {
                    const int iBytesWrittenBeforeThisSectionMaterial = iTotalBytesWritten;

                    const PS2HierSkinSectionMaterial& xSectionMaterial = xSection.m_pxMaterials[uMaterial];

                    CHECK( xChainWriter.BeginChain() );

                    for( u_int uStrip = 0; uStrip < xSectionMaterial.m_uNumStrips; ++uStrip )
                    {
                        // Deal with setting up CYCLE and filling alpha
                        // values in advance.
                        {
                            bool bJustFillOpaqueAlphas;
                            bool bSetMASKandROW; // Only valid if bJustFillOpaqueAlphas is true.
                            if( s_pxOutputHierSkin->m_bAnyNonOpaqueVerts )
                            {
                                bJustFillOpaqueAlphas = !xSectionMaterial.m_bAnyNonOpaqueVerts && (uStrip < 2);
                                bSetMASKandROW        = !uStrip;
                            }
                            else
                            {
                                bJustFillOpaqueAlphas = ( uCumulativeStripIdx < 2 );
                                bSetMASKandROW        = ( 0 == uCumulativeStripIdx || (1 == uCumulativeStripIdx && !uStrip) );
                            }

                            if( bJustFillOpaqueAlphas )
                            {
                                // Fill alpha in both buffers for maximum
                                // number of vertices, so we don't need to
                                // worry about it for the duration of this
                                // skin / section material.

                                // Temporarily change CYCLE.
                                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(4, 0, false)) );

                                // Set STMASK, if required, so we 
                                if( bSetMASKandROW )
                                {
                                    // Just write to w of colour QW...
                                    CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STMASK(false)) );
                                    const PS2Strip_VIF1_Mask xMask =
                                    {
                                        3, 3, 3, 3,
                                        3, 3, 3, 3,
                                        3, 3, 3, 1,
                                        3, 3, 3, 3,
                                    };
                                    CHECK( xChainWriter.AddDataWord(reinterpret_cast<const u_int&>(xMask)) );

                                    // ... and fill it with 1.
                                    CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STROW(false)) );
                                    const float afFillRow[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
                                    CHECK( xChainWriter.AddData(afFillRow, 16) );
                                }

                                // NB: The CMD here is irrelevant since we're
                                // just using fill values and not sending any
                                // data.
                                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(0,
                                                                                         ASURA_PS2_GSM__MAX_VERTS_PER_STRIP * 4,
                                                                                         PS2STRIP_VIF1_CMD_V4_32 | PS2STRIP_VIF1_MASK_PROCESS_ENABLE,
                                                                                         false)) );

                                // Reset CYCLE.
                                CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(1, 4, false)) );
                            }
                            else
                            {
                                // Ensure we've setup CYCLE before the first
                                // strip in each chain.
                                if( !uStrip )
                                {
                                    CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(1, 4, false)) );
                                }
                            }
                        }

                        const PS2HierSkinStrip& xStrip = xSectionMaterial.m_pxStrips[uStrip];

                        const u_int uNumVertices = xStrip.m_uNumVertices;
                        Asura_Assert( uNumVertices <= ASURA_PS2_GSM__MAX_VERTS_PER_STRIP, "" );

                        // Positions.
                        CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(0, uNumVertices, PS2STRIP_VIF1_CMD_V3_32, false)) );
                        for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
                        {
                            CHECK( xChainWriter.AddData(&xStrip.m_pxVerts[uVtx].xPos, 12) );
                        }

                        // STs.
                        {
                            short aiCoords[ASURA_PS2_GSM__MAX_VERTS_PER_STRIP * 2];
                        
                            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(1, uNumVertices, PS2STRIP_VIF1_CMD_V2_16, false)) );

                            for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
                            {
                                const PS2Strip_MatInfo& xMatInfo = xStrip.m_pxVerts[uVtx].xMat;

                                if( !bDoneTexCoordError && !xMatInfo.CheckTexCoords() )
                                {
                                    PS2Strip::LogError( "ERROR: Skin \"%s\" (LOD %u) has texture coordinate outside valid PS2 range.", szSkinName, PS2Strip::GetCurSkinLOD() );
                                    bDoneTexCoordError = true;
                                }

                                aiCoords[(uVtx << 1) + 0] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMatInfo.m_afU[0] );
                                aiCoords[(uVtx << 1) + 1] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMatInfo.m_afV[0] );
                            }
                        
                            CHECK( xChainWriter.AddData(aiCoords, uNumVertices * 4) );
                        }

                        // Alpha, if required.
                        if( xSectionMaterial.m_bAnyNonOpaqueVerts )
                        {
                            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(2, uNumVertices, PS2STRIP_VIF1_CMD_S_32, false)) );

                            float afAlphas[ASURA_PS2_GSM__MAX_VERTS_PER_STRIP];
                            for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
                            {
                                afAlphas[uVtx] = xStrip.m_pxVerts[uVtx].fAlpha;
                            }

                            // NB: This deals with padding to word alignment
                            // on our behalf.
                            CHECK( xChainWriter.AddData(afAlphas, uNumVertices * 4) );
                        }
                    
                        // Normals and ADC bits.
                        {
                            CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(3, uNumVertices, PS2STRIP_VIF1_CMD_V4_5, false)) );

                            PS2Strip_VIF1_V4_5_Data axNrmsAndADCs[ASURA_PS2_GSM__MAX_VERTS_PER_STRIP];
                            for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
                            {
                                const Asura_Vector_3& xNormal = xStrip.m_pxVerts[uVtx].xNrm;
                                axNrmsAndADCs[uVtx].usX = Tools_PS2Strip_ConvertFloatForV45( xNormal.x );
                                axNrmsAndADCs[uVtx].usY = Tools_PS2Strip_ConvertFloatForV45( xNormal.y );
                                axNrmsAndADCs[uVtx].usZ = Tools_PS2Strip_ConvertFloatForV45( xNormal.z );

                                axNrmsAndADCs[uVtx].usW = xStrip.m_pxVerts[uVtx].bNeedsADC;
                            }

                            // NB: This deals with padding to word alignment
                            // on our behalf.
                            CHECK( xChainWriter.AddData(axNrmsAndADCs, uNumVertices * 2) );
                        }

                        // Pass vertex count in ITOPS.
                        CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_ITOP(uNumVertices, false)) );

                        // And kick off the microprogram to render this strip.
                        CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_MSCNT(false)) );

                        ++uCumulativeStripIdx;
                    }
                    // End loop through strips.

                    CHECK( xChainWriter.EndChain() );

                    // Update cumulative byte count.
                    const u_int uBytesForThisMaterial = xChainWriter.GetTotalBytesWrittenForLastChain();
                    Asura_Assert( !(uBytesForThisMaterial & 15), "" );
                    iTotalBytesWritten += uBytesForThisMaterial;

                    // If this isn't the last section material, pad to 8QW
                    // alignment before the next one.
                    if( uCumulativeSectionMatIdx + 1 < s_pxOutputHierSkin->m_uTotalSectionMaterials )
                    {
                        PAD_TO_ALIGNMENT( hOutputFile, 128, iTotalBytesWritten, iNumBytesWrittenBeforeSkinHdr );
                    }

                    // Record details for this section material.
                    {
                        int iMaterialIDWithinSkin;
                        if( ASURA_MATERIAL_NONE == xSectionMaterial.m_iMaterialID )
                        {
                            if( !bDoneUntexturedError )
                            {
                                PS2Strip::LogError( "ERROR: Skin \"%s\" (LOD %u) contains untextured polys.\n", szSkinName, PS2Strip::GetCurSkinLOD() );
                                bDoneUntexturedError = true;
                            }

                            iMaterialIDWithinSkin = ASURA_MATERIAL_NONE;
                        }
                        else
                        {
                            int iArrayIdx = Asura_Material_System::TranslateMaterialIDToArrayIdx( xSectionMaterial.m_iMaterialID );
                            iMaterialIDWithinSkin = aiToSkinMaterialIDConversionList[iArrayIdx];
                            Asura_Assert( ASURA_MATERIAL_NONE != iMaterialIDWithinSkin, "" );
                        }

                        int iNumBytesForSecMat = iTotalBytesWritten - iBytesWrittenBeforeThisSectionMaterial;
                        Asura_Assert( !(iNumBytesForSecMat & 15), "" );
                    
                        axSectionMatHdrs[uCumulativeSectionMatIdx].usTriCount             = static_cast<u_short>( xSectionMaterial.m_uTriCount );
                        axSectionMatHdrs[uCumulativeSectionMatIdx].sMaterialID            = static_cast<short>( iMaterialIDWithinSkin );
                        axSectionMatHdrs[uCumulativeSectionMatIdx].uByteOffsetToNextChain = iNumBytesForSecMat;

                        ++uCumulativeSectionMatIdx;
                    }
                }
                // End loop through section materials.
            }
            // End loop through sections.

            Asura_Assert( uCumulativeSectionMatIdx == s_pxOutputHierSkin->m_uTotalSectionMaterials, "" );
            Asura_Assert( uCumulativeStripIdx == s_pxOutputHierSkin->m_uTotalStrips, "" );
        }
        // End chain writing.
    }
    // End if( uNum Sections )

    // Go back and fill in chunk header.
    {
        SetFilePointer( hOutputFile, -iTotalBytesWritten, NULL, FILE_CURRENT );

        Asura_Chunk_ResourceFile_PS2_StrippedHierSkin xHdr;

        xHdr.ID       = ASURA_CHUNK_RESOURCEFILE;
        xHdr.Size     = iTotalBytesWritten;
        xHdr.Version  = ASURA_PS2_HIERSKIN_CUR_VERSION_NUMBER;
        xHdr.Flags    = 0;
        xHdr.Type     = ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC;
        xHdr.SubType  = ASURA_RESOURCEFILE_TYPE_PS2_STRIPPEDHIERSKIN;
        xHdr.FileSize = iTotalBytesWritten - iTotalBytesWrittenBeforeResourceData;

        int iIgnore = 0;
        WRITE_DATA( hOutputFile, iIgnore, &xHdr, sizeof(Asura_Chunk_ResourceFile_PS2_StrippedHierSkin) );
    }

    // Seek forward to the section material headers and fill those in, if
    // relevant.
    if( s_pxOutputHierSkin->m_uTotalSectionMaterials )
    {
        Asura_Assert( -1 != iBytesWrittenBeforeSectionMaterialHdrs, "" );

        const int iSeek = iBytesWrittenBeforeSectionMaterialHdrs - sizeof(Asura_Chunk_ResourceFile_PS2_StrippedHierSkin);
        Asura_Assert( iSeek > 0, "" );
        SetFilePointer( hOutputFile, iSeek, NULL, FILE_CURRENT );

        int iIgnore = 0;
        WRITE_DATA( hOutputFile, iIgnore, axSectionMatHdrs, s_pxOutputHierSkin->m_uTotalSectionMaterials * sizeof(Asura_PS2_HierSkin_SectionMatHdr) );
    }

    // Seek back to the end of the file.
    SetFilePointer( hOutputFile, 0, NULL, FILE_END );

    // Output progress message.
    // (MP - 25.05.2002) TODO: Add some stats here.
    PS2Strip::LogInfo( "done.\n" );

    return true;

    #undef CHECK
}

void PS2Strip_ExporterImplementation::BeginEnvExport( void )
{
    Asura_Assert( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_ENV, "" );

    s_fStartProgressPercentage = 0.0f;
    s_fProgressPercentageDelta = 100.0f;
    
    s_iEnvCurModuleIdx                     = 0;
    s_uEnvNumTris                          = 0;
    s_uEnvNumVertices                      = 0;
    s_uEnvNumStrips                        = 0;
    s_uEnvNumSingleTriStrips               = 0;
    s_uEnvNumStripsFromStripifier          = 0;
    s_uEnvNumSingleTriStripsFromStripifier = 0;
    s_uEnvMaxLengthFromStripifier          = 0;
    s_uEnvNumVerticesWithDodgyTexCoords    = 0;

    BeginExportInternal();
}

void PS2Strip_ExporterImplementation::BeginHierSkinExport( const Asura_Hierarchy_Skin* pxSkin, float fStartProgressPercentage, float fEndProgressPercentage )
{
    Asura_Assert( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_SKIN_HIER, "" );

    s_fStartProgressPercentage = fStartProgressPercentage;
    s_fProgressPercentageDelta = fEndProgressPercentage - fStartProgressPercentage;

    s_pxInputSkin    = pxSkin;
    s_iCurSectionIdx = 0;
    
    BeginExportInternal();
}

void PS2Strip_ExporterImplementation::BeginObjectExport( Asura_Mesh* pxMesh )
{
    Asura_Assert( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_OBJECT, "" );

    s_fStartProgressPercentage = 0.0f;
    s_fProgressPercentageDelta = 50.0f;

    s_pxCurMesh = pxMesh;

    BeginExportInternal();
}


void PS2Strip_ExporterImplementation::BeginExportInternal( void )
{
    // Output header message.
    {
        const Asura_Char* szFormat =
            "Exporting PS2 stripped %s.\n"
            "File opened at %02i:%02i:%02i on %02i.%02i.%04i.\n"
            "Exported by Asura built %s\n"
            "\n";

        SYSTEMTIME xLocalTime;
        GetLocalTime( &xLocalTime );

        PS2Strip::LogInfo(
            szFormat,
            PS2Strip::GetCurExportDescription(),
            xLocalTime.wHour,
            xLocalTime.wMinute,
            xLocalTime.wSecond,
            xLocalTime.wDay,
            xLocalTime.wMonth,
            xLocalTime.wYear,
            Asura_BuildTimeAndDate
        );
    }

    s_bSubdivideMesh                          = false;
	s_fTexUScale                              = 1.0f;
	s_fTexVScale                              = 1.0f;

    s_uTotalDegenerateSourceTris      = 0;
    s_uTotalDuplicateSourceTris       = 0;
	s_uTotalEdgesWithTooManyTriangles = 0;

    Timing_Reset();
}

PS2Strip_Exporter::UPDATERESULT PS2Strip_ExporterImplementation::Update( void )
{
    const PS2Strip::EXPORTTYPE eExportType = PS2Strip::GetCurExportType();

    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OTHER );

    // Env-specific bits.
    if( eExportType == PS2Strip::EXPORTTYPE_ENV )
    {
        // Check export index is valid.
        Asura_Assert( 0 <= s_iEnvCurModuleIdx && s_iEnvCurModuleIdx < Asura_Module_System::GetNumberOfModules(), "" );

        // Before processing the meshes individually, we need to build a list
        // of all the unique materials in the environment.
        if( !s_iEnvCurModuleIdx )
        {
            bool bOK = BuildEnvMaterialIDList();
            if( !bOK )
            {
                CleanUp( true );
                return PS2Strip_Exporter::UPDATERESULT_ERROR;
            }
        }
    }

    PS2Strip::LogInfo( "\n" );

    bool bMeshProcessedOK = true;
    #define MESH_PROCESSING_FAILED() \
    { \
        bMeshProcessedOK = false; \
        goto PROCESSED_MESH; \
    }
    #define CHECK( x ) \
    { \
        if( !(x) ) { MESH_PROCESSING_FAILED(); } \
    }

    // Get next mesh.
    if ( ( eExportType != PS2Strip::EXPORTTYPE_OBJECT ) && ( eExportType != PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT ) )
    {
        Asura_Assert( s_pxCurMesh == NULL, "" );

        if( eExportType == PS2Strip::EXPORTTYPE_SKIN_HIER )
        {
            // Setup input mesh and output skin.
            {
                // Allocate output skin if this is the first section.
                if( !s_pxOutputHierSkin )
                {
                    Asura_Assert( 0 == s_iCurSectionIdx, "" );

                    s_pxOutputHierSkin = new PS2HierSkin();
                    if( !s_pxOutputHierSkin )
                    {
                        PS2Strip::LogError( "ERROR: Out of memory!\n" );
                        MESH_PROCESSING_FAILED();
                    }

                    s_pxOutputHierSkin->m_uNumSections = s_pxInputSkin->NumberOfBones;
                    s_pxOutputHierSkin->m_pxSections   = new PS2HierSkinSection[s_pxOutputHierSkin->m_uNumSections];
                }

                s_pxCurMesh = s_pxInputSkin->SectionData[s_iCurSectionIdx].Mesh;
                if( !s_pxCurMesh )
                {
                    goto PROCESSED_MESH;
                }
            }
        }
        else
        {
            s_pxCurMesh = Asura_Module_System::GetMeshForModule( s_iEnvCurModuleIdx );
        }
    }
    Asura_Assert( s_pxCurMesh, "" );

	if(eExportType == PS2Strip::EXPORTTYPE_ENV)
	{
		//Skip modules that aren't part of the current light set
		//RAR 12/12/2007
		if(!Tools_Level_Export::ShouldIncludeModuleInExport(s_iEnvCurModuleIdx))
		{
			goto PROCESSED_MESH;
		}
	}

    // Check mesh is sensible.
    if( !s_pxCurMesh->NumberOfPolygons )
    {
        switch( eExportType )
        {
            case PS2Strip::EXPORTTYPE_SKIN_HIER:
                PS2Strip::LogError( "ERROR: Hierarchy section %s of skin %s has no polys.\n", s_pxCurMesh->Name.GetString(), s_pxInputSkin->Name.GetString() );
                break;

            case PS2Strip::EXPORTTYPE_ENV:
                PS2Strip::LogError( "ERROR: Module \"%s\" has no polys in its mesh.\n", s_pxCurMesh->Name.GetString() );
                break;

            case PS2Strip::EXPORTTYPE_OBJECT:
                PS2Strip::LogError( "ERROR: Object \"%s\" has no polys.\n", s_pxCurMesh->Name.GetString() );
                break;

			case PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT:
                PS2Strip::LogError( "ERROR: Botanical Object \"%s\" has no polys.\n", s_pxCurMesh->Name.GetString() );
                break;
        }

        MESH_PROCESSING_FAILED();
    }

    // Output status message.
    if ( ( eExportType != PS2Strip::EXPORTTYPE_OBJECT ) && ( eExportType != PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT ) )
    {
        PS2Strip::LogInfo( "Mesh %i - \"%s\".\n",
                           (eExportType == PS2Strip::EXPORTTYPE_SKIN_HIER ? s_iCurSectionIdx : s_iEnvCurModuleIdx),
                           s_pxCurMesh->Name.GetString() );
    }

    // Record name of this next mesh for GetNameOfCurMesh()'s purposes.
    strncpy( g_szNameOfCurMesh, s_pxCurMesh->Name.GetString(), MAX_MESH_NAME_LENGTH );

    // Generate any required mesh vertex data. This means normals for both
    // characters and environments and also modelspace bone offsets for
    // characters.
    CHECK( GenerateRequiredVertexData() );

    // If this is an environment mesh, create prelighting data now.
    if( eExportType == PS2Strip::EXPORTTYPE_ENV )
    {
        CHECK( GenerateEnvPrelighting() );
    }

    // Process vertex data ready for stripping.
    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_ELIM_DUP_VTX_INFO );
    CHECK( ProcessVerticesAndEliminateDuplicates() );
    Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OTHER );

	if( s_uNumUniqueMeshTris )
	{
        Timing_SetProfileArea( TIMING_PROFILEAREA_NVTRISTRIP );
    	CHECK( NVTS_Stripify() );
        Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OTHER );
       	CHECK( NVTS_PostProcessStrips() );

        // Sort the strips. Currently the requirements of everything we
        // export here are the same: sort by material and then by strip
        // length within that.
        {
            const u_int uNumStrips = s_xPostProcessedStrips.size();

            // So, firstly sort by material...
            SortPostProcessedStripsOnMaterial( 0, uNumStrips - 1 );

            int iCurMaterialID = s_xPostProcessedStrips[0]->iMaterialID;
            u_int uFirstStripWithThisMaterial = 0;

            // ... and now by length. See SortPostProcessedStripsOnLength()'s
            // notes for details on what exactly we do here.
            for( u_int uStrip = 1; uStrip < uNumStrips; ++uStrip )
            {
                const PostProcessedStrip* pxPPStrip = s_xPostProcessedStrips[uStrip];

                // End of current material's strips?
                if( pxPPStrip->iMaterialID != iCurMaterialID )
                {
                    // Ok, sort this lot by length.
                    SortPostProcessedStripsOnLength( uFirstStripWithThisMaterial, uStrip - 1 );

                    iCurMaterialID = pxPPStrip->iMaterialID;
                    uFirstStripWithThisMaterial = uStrip;
                }
            }

            // Deal with last material.
            SortPostProcessedStripsOnLength( uFirstStripWithThisMaterial, uNumStrips - 1 );
        }


        switch( eExportType )
        {
            case PS2Strip::EXPORTTYPE_SKIN_HIER:
                CHECK( AddHierSkinData() );
                break;

            case PS2Strip::EXPORTTYPE_ENV:
                CHECK( AddEnvData() );
                break;

			case PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT:
            case PS2Strip::EXPORTTYPE_OBJECT:
                // Just deal with it below.
                break;
        }
    }

    #undef CHECK
    #undef MESH_PROCESSING_FAILED

    // Output progress message.
    PS2Strip::LogInfo( "Successfully stripified mesh \"%s\".\n", s_pxCurMesh->Name.GetString() );

    // Record name of this mesh for GetNameOfLastMesh()'s purposes now we're
    // done with it.
    strncpy( g_szNameOfLastMesh, s_pxCurMesh->Name.GetString(), MAX_MESH_NAME_LENGTH );

    // Clear name of current mesh while we're not processing one. (This will
    // get set up again at the beginning of the next update cycle.)
    g_szNameOfCurMesh[0] = '\0';

PROCESSED_MESH:

    PS2Strip_Exporter::UPDATERESULT eRetVal;
    if( bMeshProcessedOK )
    {
        u_int uNumMeshesExported;
        u_int uNumMeshesToExport;

        switch( eExportType )
        {
            case PS2Strip::EXPORTTYPE_SKIN_HIER:
                uNumMeshesExported = ++s_iCurSectionIdx;
                uNumMeshesToExport = s_pxInputSkin->NumberOfBones;
                break;

            case PS2Strip::EXPORTTYPE_ENV:
                uNumMeshesExported = ++s_iEnvCurModuleIdx;
                uNumMeshesToExport = Asura_Module_System::GetNumberOfModules();
                break;

			case PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT:
            case PS2Strip::EXPORTTYPE_OBJECT:
                uNumMeshesExported = uNumMeshesToExport = 1;
                break;
        }
        
        Asura_Assert( uNumMeshesExported <= uNumMeshesToExport, "" );

        if( uNumMeshesExported == uNumMeshesToExport )
        {
            OutputCumulativeWarnings();

            bool bOK;
            switch( eExportType )
            {
                case PS2Strip::EXPORTTYPE_SKIN_HIER:
                    bOK = OutputHierSkin(); break;

                case PS2Strip::EXPORTTYPE_ENV:
                    bOK = OutputEnv(); break;

                case PS2Strip::EXPORTTYPE_OBJECT:
                    bOK = OutputObject(); break;

#if defined( ASURA_USE_BOTANICALS )
				case PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT:
				{
                    bOK = OutputBotanicalObject(); 
					break;
				}
#endif // #if defined( ASURA_USE_BOTANICALS )
            }

            Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OTHER );
            eRetVal = ( bOK ? PS2Strip_Exporter::UPDATERESULT_COMPLETE : PS2Strip_Exporter::UPDATERESULT_ERROR );
        }
        else
        {
            Timing_SetProfileArea( TIMING_PROFILEAREA_REFRESH_DISPLAY );
            PS2Strip::UpdateCompletionStatus( s_fStartProgressPercentage + s_fProgressPercentageDelta * uNumMeshesExported / uNumMeshesToExport );
            Timing_SetProfileArea( TIMING_PROFILEAREA_PS2STRIP_OTHER );
            eRetVal = PS2Strip_Exporter::UPDATERESULT_PENDING;
        }
    }
    else
    {
        eRetVal = PS2Strip_Exporter::UPDATERESULT_ERROR;
    }

	CleanUp( PS2Strip_Exporter::UPDATERESULT_PENDING != eRetVal );

    Timing_SetProfileArea( TIMING_PROFILEAREA_OTHER );

    return eRetVal;
}

static inline float Timing_GetFrequency( void )
{
    LARGE_INTEGER xFreq;
    QueryPerformanceFrequency( &xFreq );
    return( static_cast<float>(xFreq.QuadPart) );
}

static inline LONGLONG Timing_GetTicks( void )
{
    LARGE_INTEGER xTicks;
    QueryPerformanceCounter( &xTicks );
    return xTicks.QuadPart;
}

static LONGLONG g_aiElapsedTimes[TIMING_PROFILEAREA_NUMAREAS];
static TIMING_PROFILEAREA g_eCurProfileArea;
static LONGLONG g_iTimeEnteredThisArea;

float PS2Strip_ExporterImplementation::Timing_GetTotalElapsedTime( void )
{
    LONGLONG iCurTicks = Timing_GetTicks();

    // Bit of a hack: make a final update to the separate timers.
    if( TIMING_PROFILEAREA_NUMAREAS != g_eCurProfileArea )
    {
        g_aiElapsedTimes[g_eCurProfileArea] += iCurTicks - g_iTimeEnteredThisArea;
        g_eCurProfileArea = TIMING_PROFILEAREA_NUMAREAS;
    }

    return( static_cast<float>(iCurTicks - s_iStartTime) / Timing_GetFrequency() );
}

void PS2Strip_ExporterImplementation::Timing_SetProfileArea( TIMING_PROFILEAREA eArea )
{
    Asura_Assert( TIMING_PROFILEAREA_NUMAREAS != eArea, "Invalid area." );

    LONGLONG iCurTicks = Timing_GetTicks();

    Asura_Assert( TIMING_PROFILEAREA_NUMAREAS != g_eCurProfileArea, "Did you reset?" );
    g_aiElapsedTimes[g_eCurProfileArea] += iCurTicks - g_iTimeEnteredThisArea;

    g_eCurProfileArea      = eArea;
    g_iTimeEnteredThisArea = iCurTicks;
}

void PS2Strip_ExporterImplementation::Timing_Reset( void )
{
    s_iStartTime = Timing_GetTicks();

    for( int i = 0; i < TIMING_PROFILEAREA_NUMAREAS; ++i )
    {
        g_aiElapsedTimes[i] = 0;
    }
    g_eCurProfileArea      = TIMING_PROFILEAREA_OTHER;
    g_iTimeEnteredThisArea = s_iStartTime;
}

void PS2Strip_ExporterImplementation::Timing_PrintReport( void )
{
    float fOneOverFreq = 1.0f / Timing_GetFrequency();

    float afTimes[TIMING_PROFILEAREA_NUMAREAS];
    for( int i = 0; i < TIMING_PROFILEAREA_NUMAREAS; ++i )
    {
        afTimes[i] = static_cast<float>(g_aiElapsedTimes[i]) * fOneOverFreq;
    }

    PS2Strip::LogInfo( "(Timing breakdown: PS2Strip [elim dup vtx]     : %3.1f\n"
                       "                   PS2Strip [subdiv & prelight]: %3.1f\n"
                       "                   PS2Strip [output]           : %3.1f\n"
                       "                   PS2Strip [other]            : %3.1f\n"
                       "                   NVTriStrip                  : %3.1f\n"
                       "                   Display                     : %3.1f\n"
                       "                   Other                       : %3.1f )\n",
                       afTimes[TIMING_PROFILEAREA_PS2STRIP_ELIM_DUP_VTX_INFO],
                       afTimes[TIMING_PROFILEAREA_PS2STRIP_SUBDIV_AND_PRELIGHT],
                       afTimes[TIMING_PROFILEAREA_PS2STRIP_OUTPUT],
                       afTimes[TIMING_PROFILEAREA_PS2STRIP_OTHER],
                       afTimes[TIMING_PROFILEAREA_NVTRISTRIP],
                       afTimes[TIMING_PROFILEAREA_REFRESH_DISPLAY],
                       afTimes[TIMING_PROFILEAREA_OTHER] );
}

#if defined( ASURA_USE_CARPET )

struct Asura_PS2_Carpet_Triangle_Info
{
	u_int uStyleIndex;	
	u_int uModuleIndex;
    u_int uSeed;
    u_int uExpectedNumBillboards;
};


bool PS2Strip_ExporterImplementation::OutputCarpetMeshChunk( HANDLE pOutput )
{
// JAT 010306
#define CARPET_PAD_TO_QW_ALIGNMENT( handle, total_num_bytes_written, num_bytes_of_padding_reqd ) \
	{ \
	/* Copy pad count to prevent evaluating an expression with side- \
effects multiple times. */ \
	const int iPad = (num_bytes_of_padding_reqd); \
	\
	Asura_Assert( 0 < iPad && !(iPad & 3) && iPad <= 12, "Invalid num pad bytes." ); \
	\
	/* Should be always be word aligned in the file. */ \
	Asura_Assert( !((total_num_bytes_written) & 3), "Should always be word aligned." ); \
	\
	/* Output padding. */ \
	WRITE_DATA( (handle), (total_num_bytes_written), "            ", iPad ); \
	}

#define CARPET_RECORD_FILE_PTR( handle, initial_file_ptr, total_num_bytes_written, var ) \
	{ \
	const u_int uFilePtr = SetFilePointer( (handle), 0, NULL, FILE_CURRENT ); \
	if( INVALID_SET_FILE_POINTER == uFilePtr ) \
	{ \
	PS2Strip::LogError( "ERROR: Couldn't get file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() ); \
	return false; \
	} \
	\
	/* At each point, our total-bytes-written count should be the same \
as the current file pointer (assuming this is always called when \
we're at EOF. */ \
	Asura_Assert( uFilePtr == (initial_file_ptr) + (total_num_bytes_written), "Total bytes written count is incorrect." ); \
	\
	(var) = uFilePtr; \
	}

#ifdef ASURA_ASSERT
	// Record initial file ptr.
	const u_int uInitialFilePtr = SetFilePointer( pOutput, 0, NULL, FILE_CURRENT );
	if( INVALID_SET_FILE_POINTER == uInitialFilePtr )
	{
		PS2Strip::LogError( "ERROR: Couldn't get file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
		return false;
	}
	Asura_Assert(!(uInitialFilePtr & 15), "Carpet mesh not starting on QW boundary. This will break levels that load asynchronously");
#endif

	////////////////////////////////////////////////////////////////////////////
	//
	// Process the carpet mesh triangle array
    //
    // See Asura_PS2_Carpet_Mesh_Shared.h for the format details
	//
	///////////////////////////////////////////////////////////////////////////

	// Make an array of the active carpet mesh triangles: a valid style, >0 quads to draw
	// So make an array of DMA triangle data (sort it by style 1st then module second)
	const u_int uMaxNumCarpetTris = Asura_Carpet_MeshManager::GetNumTriangles();

	// Don't output the chunk if no data to export
	if ( uMaxNumCarpetTris == 0 )
	{
		return true;
	}

	const u_int uMaxNumCarpetStyles = Asura_Carpet_StyleManager::GetNumStyles();
	Tools_DynamicArray<Asura_PS2_Carpet_Style_DMA_Data> xDMAStyles( uMaxNumCarpetStyles );
	Asura_Assert( xDMAStyles, "" );

	Tools_DynamicArray<Asura_PS2_Carpet_Triangle_Info> xTriInfos( uMaxNumCarpetTris );
	Asura_Assert( xTriInfos, "" );

	Tools_DynamicArray<Asura_PS2_Carpet_Triangle_DMA_Data> xDMATris( uMaxNumCarpetTris );
	Asura_Assert( xDMATris, "" );

	u_int uNumTriangles = 0;
	u_int uNumStyles = 0;

	Tools_DynamicArray<bool> xActiveStyles( uMaxNumCarpetStyles );
	memset( xActiveStyles, false, sizeof( bool ) * uMaxNumCarpetStyles );
	Asura_LinkList<Asura_Carpet_Triangle>& xTriangleList = Asura_Carpet_MeshManager::GetTriangleList();

	ASURA_STATIC_ASSERT( ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH == 16 );
	ASURA_STATIC_ASSERT( ASURA_CARPET_BILLBOARDS_PER_STYLE == 8 );

	// JAT 13/02/06 : ASURA_CARPET_MESH_DATA_FRAME_SIZE_WAVEX must be a power of 2
	ASURA_STATIC_ASSERT( ( ASURA_CARPET_MESH_DATA_FRAME_SIZE_WAVEX & ( ASURA_CARPET_MESH_DATA_FRAME_SIZE_WAVEX - 1 ) ) == 0 );

	// JAT 13/02/06 : ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH must be a power of 2
	ASURA_STATIC_ASSERT( ( ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH & ( ASURA_CARPET_BILLBOARD_SEQUENCE_LENGTH - 1 ) ) == 0 );

	// Need to remap the styles
	Tools_DynamicArray<int> xStylesRemap( uMaxNumCarpetStyles );
	for ( u_int uI = 0; uI < uMaxNumCarpetStyles; uI++ )
	{
		xStylesRemap[uI] = uMaxNumCarpetStyles;
	}

	// Loop over the carpet mesh editor triangles making PS2 data from them
	for (Asura_ListElt<Asura_Carpet_Triangle>* pxElt = xTriangleList.Front(); pxElt; pxElt = pxElt->Next() )
	{
		const Asura_Carpet_Triangle* const pxTri = &( pxElt->Data() );
		const int iStyleUID = pxTri->iStyleUID;
		const int iStyleIndex = Asura_Carpet_StyleManager::FindStyleFromUID( iStyleUID );
		// Ignore if the style doesn't exist
		if ( iStyleIndex < 0 )
		{
			continue;
		}
		const Asura_Carpet_Style* const pxStyle = Asura_Carpet_StyleManager::GetStyle( iStyleIndex );
		const float fTriDensity = ( pxTri->xVerts[0].fDensity + pxTri->xVerts[1].fDensity + pxTri->xVerts[2].fDensity ) * 0.3333f;
		int iNumQuads = Asura_Maths::FloatToInt( pxTri->fArea * pxStyle->fDensity * fTriDensity );
		if ( iNumQuads <= 0 )
		{
			continue;
		}
		const Asura_Vector_3 xLOSPos(pxTri->xCentre.x, pxTri->xCentre.y - 0.05f, pxTri->xCentre.z);
		Tools_Environment_LOS_Data xLOSData(xLOSPos, Asura_Vector_3(0.f, 1.f, 0.f));
		xLOSData.SetReturnOnFirstHit(true);
		xLOSData.SetPolyFlagsToIgnore(AsuraPolygonFlag_IgnoreLOS);

		int iModuleIndex = -1;
		if (Tools_Environment_LOS::TestWithRay(xLOSData))		
		{
			iModuleIndex = xLOSData.GetModuleIndex();
		}
		else
		{
			// Fall back on simple position test
			iModuleIndex = Asura_Module_System::GetModuleIndexFromPosition( pxTri->xCentre );
		}
		// Ignore triangle if it is not in a valid module
		if ( iModuleIndex < 0 )
		{
			Asura_Assert(0, "Could not find module for carpet tri");
			continue;
		}
		if ( xActiveStyles[iStyleIndex] == false )
		{
			xStylesRemap[iStyleIndex] = uNumStyles;
			xActiveStyles[iStyleIndex] = true;
			// Copy the style data over to the PS2 structure
			// 16 QW's of
			//       x            y            z            w
			//  sequence[0]   movement[0]    width[0]    height[0]
			//  sequence[1]   movement[1]    width[1]    height[1]
			//  sequence[2]   movement[2]    width[2]    height[2]
			//  sequence[3]   movement[3]    width[3]    height[3]
			//  sequence[4]   movement[4]    width[4]    height[4]
			//  sequence[5]   movement[5]    width[5]    height[5]
			//  sequence[6]   movement[6]    width[6]    height[6]
			//  sequence[7]   movement[7]    width[7]    height[7]
			//  sequence[8]   movement[8]   minScale[0]    iTextureID 
			//  sequence[9]   movement[9]   minScale[1]    fLoDFadeNear
			// sequence[10]  movement[10]   minScale[2]    fLoDFadeFar
			// sequence[11]  movement[11]   minScale[3]    free
			// sequence[12]  movement[12]   minScale[4]    free
			// sequence[13]  movement[13]   minScale[5]    free
			// sequence[14]  movement[14]   minScale[6]    free
			// sequence[15]  movement[15]   minScale[7]    free
			//
			// 8 QW's of
			//    x       y        z      w
			//   u0[0]   v0[0]   u1[0]   v1[0]
			//   u0[1]   v0[1]   u1[1]   v1[1]
			//   u0[2]   v0[2]   u1[2]   v1[2]
			//   u0[3]   v0[3]   u1[3]   v1[3]
			//   u0[4]   v0[4]   u1[4]   v1[4]
			//   u0[5]   v0[5]   u1[5]   v1[5]
			//   u0[6]   v0[6]   u1[6]   v1[6]
			//   u0[7]   v0[7]   u1[7]   v1[7]

			Asura_PS2_Carpet_Style_DMA_Data& xPS2Style = xDMAStyles[uNumStyles];

			// The return tag
			Asura_PS2_Carpet_QW* const pxRETtag = reinterpret_cast<Asura_PS2_Carpet_QW* const>( &( xPS2Style.xRETtag ) ); // contains the UNPACK command
			// Zero it out
			pxRETtag->u32s[0] = 0;
			pxRETtag->u32s[1] = 0;
			pxRETtag->u32s[2] = 0;
			pxRETtag->u32s[3] = 0;

			// 0->15 = QWC, 26:27 = PCE , 28:30 = ID , 31 = IRQ
			pxRETtag->u32s[0] = 24;
			// PCE_NOCHANGE = 0 (bits 26,27)
			pxRETtag->u32s[0] |= ( 0 << 26 );
			// ID_RET = 6 (bits 28:30)
			pxRETtag->u32s[0] |= ( 6 << 28 );
			// ADDR
			pxRETtag->u32s[1] = 0;
			// TTE data0
			pxRETtag->u32s[2] = PS2STRIP_VIF1_SET_STCYCL( 1, 1, 0 );
			// TTE data1
// (MP - 14.05.2008) This gets setup at runtime (and the version that was
// written here was completely wrong).
//			pxRETtag->u32s[3] = PS2STRIP_VIF1_SET_UNPACKR( 0, 24, PS2STRIP_VIF1_CMD_V4_32, 0 );

			const Asura_Hash_ID xTextureHash = Asura_Texture_Management::GetHashIDFromIndex( pxStyle->iTextureID );
            *(u_int*)&xPS2Style.iTextureID = xTextureHash;

			//  sequence[8]   movement[8]   minScale[0]    iTextureID 
			//  sequence[9]   movement[9]   minScale[1]    fLoDFadeNear
			// sequence[10]  movement[10]   minScale[2]    fLoDFadeFar
			// QW[8].i32s[3] = style density
			xPS2Style.xData[8].f32s[3] = pxStyle->fDensity;
			// QW[9].f32s[3] = fLoDFadeNear
			xPS2Style.xData[9].f32s[3] = pxStyle->fLodFadeNear;
			// QW[10].f32s[3] = fLoDFadeFar
			xPS2Style.xData[10].f32s[3] = pxStyle->fLodFadeFar;

			// QW[i].u32s[0] = sequence[i]
			for ( int i = 0; i < 16; i++ )
			{
				xPS2Style.xData[i].u32s[0] =  pxStyle->aiSequence[i];
			}

			for ( int i = 0; i < 8; i++ )
			{
				// QW[16+i] = u0[i], v0[i], u1[i], v1[i]
				// U0, V0, U1, V1
				xPS2Style.xData[16+i].f32s[0] =  pxStyle->axBillboards[i].fU[0];
				// Invert v
				xPS2Style.xData[16+i].f32s[1] =  pxStyle->axBillboards[i].fV[1];
				xPS2Style.xData[16+i].f32s[2] =  pxStyle->axBillboards[i].fU[1];
				xPS2Style.xData[16+i].f32s[3] =  pxStyle->axBillboards[i].fV[0];

				// QW[i] = sequence movement width[i] height[i]
				xPS2Style.xData[i].f32s[2] = pxStyle->axBillboards[i].fWidth * 0.5f;
				xPS2Style.xData[i].f32s[3] =  pxStyle->axBillboards[i].fHeight;

				// QW[8+i] = sequence movement minScale[i] 
				xPS2Style.xData[8+i].f32s[2] = pxStyle->axBillboards[i].fMinScale;
			}
			uNumStyles++;
		}
		// Copy the triangle data over to PS2 structure
		Asura_PS2_Carpet_Triangle_Info& xTriInfo = xTriInfos[uNumTriangles];

		// TODO: Remap the old editor style index to the new game data style index
		xTriInfo.uStyleIndex            = iStyleIndex;
		xTriInfo.uModuleIndex           = iModuleIndex;
        xTriInfo.uSeed                  = pxTri->uSeed;
        xTriInfo.uExpectedNumBillboards = iNumQuads;

        const Asura_Vector_3& xModulePos = Asura_Module_System::GetPositionOfModule( iModuleIndex );

        // Setup the UNPACKs for this triangle here. However, we'll deal with
        // adding the DMA tag, STCYCL and anything else later.
		Asura_PS2_Carpet_Triangle_DMA_Data& xDmaTri = xDMATris[uNumTriangles];

        xDmaTri.uUnpackXYZW = PS2STRIP_VIF1_SET_UNPACKR( 0, 3, PS2STRIP_VIF1_CMD_V4_16, false );
        for( u_int u = 0; u < 3; ++u )
        {
            // To allow us to compress the positions into 1:11:4 fixed point
            // format we store a position relative to the module origin.
            Asura_Vector_3 xRelativePos = pxTri->xVerts[u].xPos - xModulePos;

            // We have a range of approximately +/- 2^11 metres, ie. 2048m,
            // which should hopefully be plenty.
            Asura_Assertf( Asura_Maths::Abs(xRelativePos.x) < 2048.0f
                           && Asura_Maths::Abs(xRelativePos.y) < 2048.0f
                           && Asura_Maths::Abs(xRelativePos.z) < 2048.0f,
                           ("Module (%s) containing carpet is too large", Asura_Module_System::GetMeshForModule(iModuleIndex)->Name.GetString()) );

            xDmaTri.ausXYZWs[u * 4 + 0] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 4, xRelativePos.x );
            xDmaTri.ausXYZWs[u * 4 + 1] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 4, xRelativePos.y );
            xDmaTri.ausXYZWs[u * 4 + 2] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 4, xRelativePos.z );
            xDmaTri.ausXYZWs[u * 4 + 3] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, pxTri->xVerts[u].fDensity );
        }

        xDmaTri.uUnpackRGBA = ( PS2STRIP_VIF1_SET_UNPACKR(3, 3, PS2STRIP_VIF1_CMD_V4_8, false) | PS2STRIP_VIF1_UNPACK_USN_FLAG );

        for( u_int u = 0; u < 3; ++u )
		{
#if defined( ASURA_DIRECT3D_BGR_SUPPORT_SM30)
			xDmaTri.aucRGBAs[u * 4 + 0] = static_cast<u_char>( Asura_Maths::ClampToRange( pxTri->xVerts[u].afColour[0], 0.0f, 1.0f ) );
			xDmaTri.aucRGBAs[u * 4 + 1] = static_cast<u_char>( Asura_Maths::ClampToRange( pxTri->xVerts[u].afColour[1], 0.0f, 1.0f ) );
			xDmaTri.aucRGBAs[u * 4 + 2] = static_cast<u_char>( Asura_Maths::ClampToRange( pxTri->xVerts[u].afColour[2], 0.0f, 1.0f ) );
			xDmaTri.aucRGBAs[u * 4 + 3] = static_cast<u_char>( Asura_Maths::ClampToRange( pxTri->xVerts[u].afColour[3], 0.0f, 1.0f ) );
#else
			// the PC colour data is BGRA but need RGBA
            xDmaTri.aucRGBAs[u * 4 + 0] = static_cast<u_char>( (pxTri->xVerts[u].uColour >> 16) & 0xFF );
            xDmaTri.aucRGBAs[u * 4 + 1] = static_cast<u_char>( (pxTri->xVerts[u].uColour >>  8) & 0xFF );
            xDmaTri.aucRGBAs[u * 4 + 2] = static_cast<u_char>( (pxTri->xVerts[u].uColour >>  0) & 0xFF );
            xDmaTri.aucRGBAs[u * 4 + 3] = 128;
#endif
		}

        xDmaTri.uMSCNT = PS2STRIP_VIF1_SET_MSCNT( false );

		uNumTriangles++;
	}

	// Loop over the triangle infos remapping the styles in case there are some
	// styles which didn't have any triangles that get exported
	for ( u_int uI = 0; uI < uNumTriangles; uI++ )
	{
		Asura_PS2_Carpet_Triangle_Info& xTriInfo = xTriInfos[uI];
		const u_int iOldStyleIndex = xTriInfo.uStyleIndex;
		const u_int iNewStyleIndex = xStylesRemap[iOldStyleIndex];
		Asura_Assert( iNewStyleIndex < uNumStyles, "" );
		xTriInfo.uStyleIndex = iNewStyleIndex;
	}
	// Bubble sort the triangle infos by style index as 1st key and module index as 2nd key
	// So end up with a list of triangles sorted by module: and within a module sorted by style
	// This is so when rendering triangle references are near each other to save cache thashing
	for ( u_int uI = 0; uI < uNumTriangles; uI++ )
	{
		for ( u_int uJ = uI + 1; uJ < uNumTriangles; uJ++ )
		{
			const int iIndex1 = xTriInfos[uI].uModuleIndex;
			const int iIndex2 = xTriInfos[uJ].uModuleIndex;
			const int iStyleID1 = xTriInfos[uI].uStyleIndex;
			const int iStyleID2 = xTriInfos[uJ].uStyleIndex;
			// 1st Key is style ID index
			// 2nd Key is module index
			if ( ( iStyleID2 < iStyleID1 ) || ( ( iStyleID2 == iStyleID1 ) && ( iIndex2 < iIndex1 ) ) )
			{
				// Swap them over
				const Asura_PS2_Carpet_Triangle_Info xTemp = xTriInfos[uI];
				xTriInfos[uI] = xTriInfos[uJ];
				xTriInfos[uJ] = xTemp;
				
				// Also swap the DMA data over
				const Asura_PS2_Carpet_Triangle_DMA_Data xTemp2 = xDMATris[uI];
				xDMATris[uI] = xDMATris[uJ];
				xDMATris[uJ] = xTemp2;
			}
		}
	}

	Tools_DynamicArray<Asura_PS2_Carpet_Mesh_Style_Header> xStyleHeaders;
	// a 2D array:  array of Module_Header's per style
	Tools_DynamicArray< Tools_DynamicArray< Tools_DynamicArray<int> > > xTrisPerModulePerStyle( uNumStyles );
	Tools_DynamicArray< Tools_DynamicArray<Asura_PS2_Carpet_Mesh_Module_Header> > xModuleHeadersPerStyle( uNumStyles );

	const u_int uNumModules = Asura_Module_System::GetNumberOfModules();

	// Loop over the triangle infos adding the triangles to the module headers
	// They are sorted by style and then by module
	u_int uOldModuleIndex = uNumModules;
	u_int uOldStyleIndex = uNumStyles;
	int iCurrentModulePlace = -1;
    u_int uTotalModuleHdrs = 0;
	for ( u_int uI = 0; uI < uNumTriangles; uI++ )
	{
		Asura_PS2_Carpet_Triangle_Info& xInfo = xTriInfos[uI];
		const u_int uModuleIndex = xInfo.uModuleIndex;
		const u_int uStyleIndex = xInfo.uStyleIndex;

		// Is this a new style
		if ( uStyleIndex != uOldStyleIndex )
		{
			xStyleHeaders[uStyleIndex].usNumModules = 0;
			xStyleHeaders[uStyleIndex].usStyleID = static_cast<u_short>( uStyleIndex );
			xStyleHeaders[uStyleIndex].uNumTris = 0;
			Tools_DynamicArray<Asura_PS2_Carpet_Mesh_Module_Header>& xModules = xModuleHeadersPerStyle[uStyleIndex];
			// Zero out the module array for this new style
			xModules.Reset();
			uOldStyleIndex = uStyleIndex;
			uOldModuleIndex = uNumModules;
			iCurrentModulePlace = -1;
		}

		// Is this a new module
		if ( uModuleIndex != uOldModuleIndex )
		{
			xStyleHeaders[uStyleIndex].usNumModules++;
			uOldModuleIndex = uModuleIndex;

			Asura_PS2_Carpet_Mesh_Module_Header xModHdr;
			xModHdr.usModuleID = static_cast<u_short>( uModuleIndex );
			xModHdr.usExpectedNumBillboards = 0;

            // Use a single seed for all of this module's tris. It's still
            // deterministic (for a single export at least) and saves a bunch
            // of data. The results between PC / PS2 / any other platform
            // would have been different even when doing it per tri so it
            // doesn't matter much.
            xModHdr.uSeed = xInfo.uSeed;

			Tools_DynamicArray<Asura_PS2_Carpet_Mesh_Module_Header>& xModules = xModuleHeadersPerStyle[uStyleIndex];
			// Add the new module header to the array
			xModules += xModHdr;
			iCurrentModulePlace++;
			Asura_Assert( iCurrentModulePlace+1 == static_cast<int>( xModules.GetSize() ), "" );

			Tools_DynamicArray<int>& xTrisPerModule = xTrisPerModulePerStyle[uStyleIndex].operator[]( iCurrentModulePlace );
			// Zero out the triangle array for this new module
			// Prime it with an allocation of 64 triangles
			xTrisPerModule.Reset( 64 );

            ++uTotalModuleHdrs;
		}

		// Add this triangle index to the correct place
		Tools_DynamicArray<int>& xTrisPerModule = xTrisPerModulePerStyle[uStyleIndex].operator[]( iCurrentModulePlace );
		xTrisPerModule += uI;

		// Update statistics
		Asura_PS2_Carpet_Mesh_Module_Header& xModHdr =  xModuleHeadersPerStyle[uStyleIndex].operator[] ( iCurrentModulePlace );
		// One more triangle in this module
        u_int uNewExpectedNumBillboards = xModHdr.usExpectedNumBillboards + xInfo.uExpectedNumBillboards;
        Asura_Assert( uNewExpectedNumBillboards <= USHRT_MAX, "Too many billboards on tris in this module" ); // Not actually very fatal - this is only used to calculate rendered tri count at runtime.
        xModHdr.usExpectedNumBillboards = static_cast<u_short>(uNewExpectedNumBillboards);

		// One more triangle in this style
		xStyleHeaders[uStyleIndex].uNumTris++;
	}

	Asura_PS2_Carpet_Mesh_Header xHeader;
	xHeader.uNumStyles = uNumStyles;
	xHeader.uTotalNumTris = uNumTriangles;
    xHeader.uTotalNumModuleHdrs = uTotalModuleHdrs;
    xHeader.uTotalDMAChainSize = 0; // Will be calculated incrementally later.
	xHeader.uTotalHeadersSizeInBytes = uNumStyles * sizeof(Asura_PS2_Carpet_Mesh_Style_Header)
                                     + uTotalModuleHdrs * sizeof(Asura_PS2_Carpet_Mesh_Module_Header);

	////////////////////////////////////////////////////////////////////////////
	//
	// Start of output code
	//
	////////////////////////////////////////////////////////////////////////////

	// Output header message.
	{
		const Asura_Char* szFormat =
			"Started exporting PS2 Carpet Mesh at %02i:%02i:%02i on %02i.%02i.%04i.\n"
			"Exported by Asura built %s\n"
			"\n";

		SYSTEMTIME xLocalTime;
		GetLocalTime( &xLocalTime );

		PS2Strip::LogInfo(
			szFormat,
			xLocalTime.wHour,
			xLocalTime.wMinute,
			xLocalTime.wSecond,
			xLocalTime.wDay,
			xLocalTime.wMonth,
			xLocalTime.wYear,
			Asura_BuildTimeAndDate
			);
	}
	u_int uTotalNumBytesWritten = 0;

	// Leave a gap for the chunk header. Record the cur file ptr first so we
	// can return to this point and fill it out later.
	u_int uFilePtrForChunkHdr;
	CARPET_RECORD_FILE_PTR( pOutput, uInitialFilePtr, uTotalNumBytesWritten, uFilePtrForChunkHdr );
	SKIP_DATA( pOutput, uTotalNumBytesWritten, sizeof(Asura_Chunk_ResourceFile) );

	// Write out the resource filename
	// Make the name multiple of 4 bytes - 1
	const Asura_Char* const CARPET_MESH_DATA_RESOURCE_FILENAME = "PS2CarpetMesh  ";
	const u_int uSizeOfResourceFilenameInBytes = strlen(CARPET_MESH_DATA_RESOURCE_FILENAME) + 1;
	Asura_Assert( !(uSizeOfResourceFilenameInBytes & 3), "" );

	u_int uSizeOfResourceFilenamePlusPaddingInBytes = uSizeOfResourceFilenameInBytes;

	// We're about to write the resource filename. Before doing so, write out
	// padding so that the data will be QW aligned to speed up loading at
	// runtime. We'll do this by prepending spaces to the resource filename.
	{
		const u_int uRem = ((uTotalNumBytesWritten + uSizeOfResourceFilenameInBytes) & 15);
		if( uRem )
		{
			const u_int uPad = 16 - uRem;
			CARPET_PAD_TO_QW_ALIGNMENT( pOutput, uTotalNumBytesWritten, uPad );
			uSizeOfResourceFilenamePlusPaddingInBytes += uPad;
		}
	}

	// Write actual resource filename.
	WRITE_DATA( pOutput, uTotalNumBytesWritten, CARPET_MESH_DATA_RESOURCE_FILENAME, uSizeOfResourceFilenameInBytes );
	Asura_Assert( !(uTotalNumBytesWritten & 15), "Should be QW aligned at this point." );

	///////////////////////////////////////////////////////////////////////////
	//
	// Write out the carpet mesh data
    //
    // See Asura_PS2_Carpet_Mesh_Shared.h for the format details
	//
	///////////////////////////////////////////////////////////////////////////

    // Leave a gap for the various headers.
	u_int uFilePtrForModuleHdrs;
	CARPET_RECORD_FILE_PTR( pOutput, uInitialFilePtr, uTotalNumBytesWritten, uFilePtrForModuleHdrs );
	SKIP_DATA( pOutput, uTotalNumBytesWritten, sizeof(Asura_PS2_Carpet_Mesh_Header) + xHeader.uTotalHeadersSizeInBytes );

    // Write out the DMA chains.
    {
        Tools_PS2_DMAChainWriter xDMAChainWriter;
        xDMAChainWriter.Init( pOutput );

	    u_int uTotalNumTris = 0; 
	    for ( u_int uStyleIndex = 0; uStyleIndex < uNumStyles; uStyleIndex++ )
	    {
		    u_int uNumTrisInStyle = 0;

		    Tools_DynamicArray< Asura_PS2_Carpet_Mesh_Module_Header >& xModuleHeaders = xModuleHeadersPerStyle[uStyleIndex];
		    const Tools_DynamicArray< Tools_DynamicArray<int> >& xTrisPerStyle = xTrisPerModulePerStyle[uStyleIndex];

		    for ( u_int uModuleIndex = 0; uModuleIndex < xStyleHeaders[uStyleIndex].usNumModules; uModuleIndex++ )
		    {
			    const Tools_DynamicArray<int>& xTrisPerModule = xTrisPerStyle[uModuleIndex];
			    const u_int uNumTris = xTrisPerModule.GetSize();

                // Ensure we don't spend ages incrementally growing the
                // buffer.
                xDMAChainWriter.EnsureMinimumCapacity( uNumTris * sizeof(Asura_PS2_Carpet_Triangle_DMA_Data) );

                xDMAChainWriter.BeginChain();

// (MP - 20.05.2008) Currently this is the only UNPACK mode this renderer 
// uses and so this is dealt with when we're setting stuff up initially at
// runtime.
//              xDMAChainWriter.AddVifCode( PS2STRIP_VIF1_SET_STCYCL(1, 1, false) );

			    for ( u_int uI = 0; uI < uNumTris; uI++ )
			    {
                    const Asura_PS2_Carpet_Triangle_DMA_Data& xTriDMAData = xDMATris[xTrisPerModule[uI]];
                    xDMAChainWriter.AddData( &xTriDMAData, sizeof(Asura_PS2_Carpet_Triangle_DMA_Data) );
			    }

                xDMAChainWriter.EndChain();

                u_int uBytesWritten = xDMAChainWriter.GetTotalBytesWrittenForLastChain();
                Asura_Assert( !(uBytesWritten & 15), "" );

                xModuleHeaders[uModuleIndex].uDMAChainSizeOrPtr = uBytesWritten;
                uTotalNumBytesWritten += uBytesWritten;
                xHeader.uTotalDMAChainSize += uBytesWritten;

                uNumTrisInStyle += uNumTris;
			    uTotalNumTris   += uNumTris;
		    }

		    Asura_Assert( uNumTrisInStyle == xStyleHeaders[uStyleIndex].uNumTris, "" );
	    }

	    Asura_Assert( uTotalNumTris == xHeader.uTotalNumTris, "" );
    }

    // Seek back and fill in the gap for the various headers.
    {
		if( INVALID_SET_FILE_POINTER == SetFilePointer(pOutput, uFilePtrForModuleHdrs, NULL, FILE_BEGIN) )
		{
			PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
			return false;
		}

        // We've already included all of these bytes in our output count when
        // we skipped this space earlier.
        int iIgnoredByteCount = 0;

	    // Main header.
        WRITE_DATA( pOutput, iIgnoredByteCount, &xHeader, sizeof(Asura_PS2_Carpet_Mesh_Header) );

        // Style headers.
	    for( u_int uStyle = 0; uStyle < uNumStyles; ++uStyle )
	    {
		    WRITE_DATA( pOutput, iIgnoredByteCount, &xStyleHeaders[uStyle], sizeof(Asura_PS2_Carpet_Mesh_Style_Header) );
        }

        // Module headers.
	    for( u_int uStyle = 0; uStyle < uNumStyles; ++uStyle )
	    {
		    const Tools_DynamicArray<Asura_PS2_Carpet_Mesh_Module_Header>& xModuleHdrs = xModuleHeadersPerStyle[uStyle];

		    WRITE_DATA( pOutput, iIgnoredByteCount, xModuleHdrs.GetArray(), xStyleHeaders[uStyle].usNumModules * sizeof(Asura_PS2_Carpet_Mesh_Module_Header) );
        }

        // Seek back to EOF.
		u_int uEOFCount = SetFilePointer( pOutput, 0, NULL, FILE_END );
		if( INVALID_SET_FILE_POINTER == uEOFCount )
		{
			PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
			return false;
		}
		Asura_Assert( uEOFCount == uInitialFilePtr + uTotalNumBytesWritten, "Total byte count is incorrect." );
    }

	// Asura_PS2_Carpet_Style_DMA_Data pxStyles[uNumStyles];
	WRITE_DATA( pOutput, uTotalNumBytesWritten, xDMAStyles, uNumStyles * sizeof(Asura_PS2_Carpet_Style_DMA_Data) );

	// Go back and fill out chunk hdr.
	{
		// Seek back to beginning of chunk hdr bytes.
		if( INVALID_SET_FILE_POINTER == SetFilePointer( pOutput, uFilePtrForChunkHdr, NULL, FILE_BEGIN) )
		{
			PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
			return false;
		}

		// Prepare hdr.
		Asura_Chunk_ResourceFile xChunkHdr;
		xChunkHdr.ID       = ASURA_CHUNK_RESOURCEFILE;
		xChunkHdr.Size     = uTotalNumBytesWritten;
		xChunkHdr.Version  = ASURA_PS2_CARPET_MESH_CHUNK_VERSION;
		xChunkHdr.Flags    = 0;
		xChunkHdr.Type     = ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC;
		xChunkHdr.SubType  = ASURA_RESOURCEFILE_TYPE_PS2_CARPET;
		xChunkHdr.FileSize = uTotalNumBytesWritten - (sizeof(xChunkHdr) + uSizeOfResourceFilenamePlusPaddingInBytes);

        u_int uExpectedFileSize = sizeof(Asura_PS2_Carpet_Mesh_Header)
                                + xHeader.uTotalHeadersSizeInBytes
                                + xHeader.uTotalDMAChainSize
                                + uNumStyles * sizeof(Asura_PS2_Carpet_Style_DMA_Data);
        Asura_Assert( xChunkHdr.FileSize == uExpectedFileSize, "" );

		// Output. Note that we've already included these bytes in the
		// cumulative count when we skipped the bytes.
		int iIgnore = 0;
		WRITE_DATA( pOutput, iIgnore, &xChunkHdr, sizeof(xChunkHdr) );
	}
	// End chunk hdr write.

	// Seek back to EOF.
	{
		const u_int uEOF = SetFilePointer( pOutput, 0, NULL, FILE_END );
		if( INVALID_SET_FILE_POINTER == uEOF )
		{
			PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
			return false;
		}

		// Check our total byte count.
		Asura_Assert( uEOF == uInitialFilePtr + uTotalNumBytesWritten, "Total byte count is incorrect." );
	}

	// Output end log message
	{
		const Asura_Char* szFormat =
			"Finished exporting PS2 Carpet Mesh at %02i:%02i:%02i on %02i.%02i.%04i.\n"
			"numModules %d Tris %d Styles %d Modules %d\n"
			"\n";

		SYSTEMTIME xLocalTime;
		GetLocalTime( &xLocalTime );

		PS2Strip::LogInfo(
			szFormat,
			xLocalTime.wHour,
			xLocalTime.wMinute,
			xLocalTime.wSecond,
			xLocalTime.wDay,
			xLocalTime.wMonth,
			xLocalTime.wYear,
			uNumModules, uNumTriangles, uNumStyles, uTotalModuleHdrs
			);
        Tools_Level_Export::AddMessage( Tools_Level_Export::MT_INFORMATION,  "%d bytes written PS2 Carpet Mesh (%d tris %d styles %d modules)", 
										uTotalNumBytesWritten, uNumTriangles, uNumStyles, uTotalModuleHdrs );
					
	}
#undef CARPET_PAD_TO_QW_ALIGNMENT
#undef CARPET_RECORD_FILE_PTR
	return true;
}

#endif //#if defined( ASURA_USE_CARPET )

#if defined( ASURA_USE_BOTANICALS )

// Loop over the polygons in a mesh making an array of the used material ID's in the mesh
int PS2Strip_ExporterImplementation::GetMaterialsFromMesh( const Asura_Mesh* const pxMesh, Tools_DynamicArray<int>& xUsedMaterialIDs )
{
	// Set to no material ID's
	xUsedMaterialIDs.Reset();
	if ( !pxMesh ) 
	{
		return 0;
	}

	// Loop over the polygons
	u_int uNumberOfMaterialsUsed = 0;
	for( int iP = 0; iP < pxMesh->NumberOfPolygons; iP++ )
	{
		if ( pxMesh->PolygonFlagDataValid && ( pxMesh->PolygonFlagDataPtr[iP].Flags & ( AsuraPolygonFlag_DontExport | AsuraPolygonFlag_DontRender ) ) )
		{
			//ignore textures for polygons that won't be rendered/exported
			continue;
		}

		const int iMaterialID = pxMesh->TextureDataPtr[iP].m_iMaterialID;
		// Check to see if this in the list already
		bool bFound = false;
		for ( u_int uI = 0; uI < xUsedMaterialIDs.GetSize(); uI++ )
		{
			if ( xUsedMaterialIDs[uI] == iMaterialID )
			{
				bFound = true;
				break;
			}
		}
		if ( bFound == false )
		{
			xUsedMaterialIDs += iMaterialID;
			uNumberOfMaterialsUsed++;
		}
	}
	Asura_Assert( uNumberOfMaterialsUsed == xUsedMaterialIDs.GetSize(), "" );
	return uNumberOfMaterialsUsed;
}

struct Asura_PS2_Botanicals_Instance_Info
{
	u_int uTypeIndex;	
	u_int uModuleIndex;	
};

// Each LoD chain is an array of components
// Each component will be of the same type
// BBOX of the chain

// A component is an Asura_Mesh, distance to switch
// BBOX of the mesh
struct Asura_PS2_Botanicals_LoD_Component
{
	Asura_PS2_Botanicals_Mesh_Type eType;
	u_int uNumTris;
	u_int uMaterialID;
	float fLoDMin;
	float fLoDMax;
	float fLoDFadeNear;
	float fLoDFadeFar;
	Asura_Mesh* pxMesh;
	Asura_Bounding_Box xBB;
};

struct Asura_PS2_Botanicals_LoD_Chain
{
	Tools_DynamicArray<u_int> xComponentIndexes;
	Asura_PS2_Botanicals_Mesh_Type eType;
	Asura_Bounding_Box xBB;
};

// Each Type is an array of LoD chains
// BBOX of the type
struct Asura_PS2_Botanicals_Type_Data
{
	u_int uOldTypeIndex;
	u_int uNewTypeIndex;
	u_int uNumTotalComponents;
	Tools_DynamicArray<Asura_PS2_Botanicals_LoD_Chain> xChains;
	Asura_Bounding_Box xBB;
	Asura_String xName;
	float fInvMaxDrawDistance;
};

// JAT 210306
bool PS2Strip_ExporterImplementation::OutputBotanicalsChunk( HANDLE pOutput, const int* const piMaterialConversionList )
{
#define BOTANICALS_PAD_TO_QW_ALIGNMENT( handle, total_num_bytes_written, num_bytes_of_padding_reqd ) \
	{ \
	/* Copy pad count to prevent evaluating an expression with side- \
effects multiple times. */ \
	const int iPad = (num_bytes_of_padding_reqd); \
	\
	Asura_Assert( 0 < iPad && !(iPad & 3) && iPad <= 12, "Invalid num pad bytes." ); \
	\
	/* Should be always be word aligned in the file. */ \
	Asura_Assert( !((total_num_bytes_written) & 3), "Should always be word aligned." ); \
	\
	/* Output padding. */ \
	WRITE_DATA( (handle), (total_num_bytes_written), "            ", iPad ); \
	}

#define BOTANICALS_RECORD_FILE_PTR( handle, initial_file_ptr, total_num_bytes_written, var ) \
	{ \
	const u_int uFilePtr = SetFilePointer( (handle), 0, NULL, FILE_CURRENT ); \
	if( INVALID_SET_FILE_POINTER == uFilePtr ) \
	{ \
	PS2Strip::LogError( "ERROR: Couldn't get file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() ); \
	return false; \
	} \
	\
	/* At each point, our total-bytes-written count should be the same \
as the current file pointer (assuming this is always called when \
we're at EOF. */ \
	Asura_Assert( uFilePtr == (initial_file_ptr) + (total_num_bytes_written), "Total bytes written count is incorrect." ); \
	\
	(var) = uFilePtr; \
	}

#ifdef ASURA_ASSERT
	// Record initial file ptr.
	const u_int uInitialFilePtr = SetFilePointer( pOutput, 0, NULL, FILE_CURRENT );
	if( INVALID_SET_FILE_POINTER == uInitialFilePtr )
	{
		PS2Strip::LogError( "ERROR: Couldn't get file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
		return false;
	}
	Asura_Assert(!(uInitialFilePtr & 15), "Botanicals not starting on QW boundary. This will break levels that load asynchronously");
#endif

	////////////////////////////////////////////////////////////////////////////
	//
	// Process the botanicals instance array
	//
	////////////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////////
	//
	// The PS2 Botanicals mesh data format Version 6 is:
	//
	// Asura_PS2_Botanicals_Header
	// {
	//    u_int uNumTypes;
	//    u_int uNumTotalComponents;
	//    u_int uNumTotalInstances;
	//    u_int uTotalHeadersSizeInBytes;
	//    u_int uTotalTypeInfosSizeInBytes;
	// };
	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Header
	//    {
	//       u_short usTypeID;
	//   	 u_short usNumModules;
	//       u_int uNumInstances;
	//    };
	//    Array of usNumModules
	//    {
	//       struct Asura_PS2_Botanicals_Module_Header
	//       {
	//       	u_short usModuleID;
	//       	u_short usNumInstances;
	//       };
	//	     Array of usNumInstances;
	//       {
	//	        struct Asura_PS2_Botanicals_Instance
	//		    {
	//			   This first part is EE used data only
	//			   u_short usGlobalInstanceIndex;
	//			   u_short usPadding;
	//
	//			   float fSphereBoundX;
	//			   float fSphereBoundY;
	//			   float fSphereBoundZ;
	//			   float fSphereBoundR;
	//
	//			   float fScaleXZZ;
	//
	//             This should be DMA friendly packet data
	//		       float fX;
	//			   float fY;
	//			   float fZ;
	//
	//			   Asura_Matrix_3x3 xLocal2World;
	//
	//			   u_int uWorldSH0;
	//			   u_int uWorldSH1;
	//			   u_int uWorldSH2;
	//			   u_int uWorldSH3;
	//
	//			   u_int uObjectSH0;
	//			   u_int uObjectSH1;
	//			   u_int uObjectSH2;
	//			   u_int uObjectSH3;
	//		    };
	//       }
	//    }
	// }
	//
	// This data is separate to the Asura_PS2_Botanicals_Type_Header data because
	// one day this type info data might be passed to VU1 for LoD selection
	//
	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Info
	//    {
	//       Asura_Bounding_Box xBB;
	// 	     u_int uNumComponents;
	//		 Array of uNumComponents;
	//       {
	//          struct Asura_PS2_Botanicals_Component_Info
	//          {
	//			   int iTextureID;					// On export it is a material ID which is converted to a texture ID & material flags
	//			   u_int uMaterialFlags;
	//			   u_short usType;					// enum Asura_PS2_Botanicals_Mesh_Type
	//			   u_short usNumTris;					   
	//			   float fLoDMin;
	//			   float fLoDMax;
	// 	           float fLoDConstant1;
	// 	           float fLoDConstant2;
	// 	           u_int uComponentIndexAndType; // u_int : top 8-bits is type, bottom 24-bits is component index into component DMA data
	//			};
	//       }
	//    }
	// }
	//
	// Array of uNumTotalComponents
	// {
	//    DMA data for the mesh component: referenced by a DMA CALL tag so the DMA data includes the DMA RET tag
	// }
	//
	///////////////////////////////////////////////////////////////////////////

	// Don't output the chunk if no data to export
	const u_int uMaxNumBotanicalsInstances = Asura_Botanicals_InstanceManager::GetNumInstances();
	if ( uMaxNumBotanicalsInstances == 0 )
	{
		return true;
	}

	// Make an array of the active botanicals instances: a valid type, not marked for merge with render env
	const u_int uMaxNumBotanicalsTypes = Asura_Botanicals_TemplateManager::GetNumTemplates();

	Tools_DynamicArray<Asura_PS2_Botanicals_Instance_Info> xInstanceInfos( uMaxNumBotanicalsInstances );
	Asura_Assert( xInstanceInfos, "" );

	Tools_DynamicArray<Asura_PS2_Botanicals_LoD_Component> xComponents( uMaxNumBotanicalsTypes );
	Asura_Assert( xComponents, "" );

	Tools_DynamicArray<Asura_PS2_Botanicals_Instance> xInstances( uMaxNumBotanicalsInstances );
	Asura_Assert( xInstances, "" );

	u_int uNumInstances = 0;
	u_int uNumTypes = 0;
	u_int uNumComponents = 0;

	Tools_DynamicArray<bool> xActiveTypes( uMaxNumBotanicalsTypes );
	Asura_Assert( xActiveTypes, "" );
	memset( xActiveTypes, false, sizeof( bool ) * uMaxNumBotanicalsTypes );
	Asura_LinkList<Asura_Botanicals_Instance*>& xInstanceList = Asura_Botanicals_InstanceManager::GetInstanceList();

	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Header ) == 5*4 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Type_Header ) == 8 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Module_Header ) == 4 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Component_Info ) == 8*4 );
	ASURA_STATIC_ASSERT( sizeof( Asura_PS2_Botanicals_Instance ) == 26*4 );

	// Make the info about the ordinary 3D geometry render shapes for each type
	Tools_DynamicArray<Asura_PS2_Botanicals_Type_Data> xTypeDatas( uMaxNumBotanicalsTypes );
	Asura_Assert( xTypeDatas, "" );

	// Need to remap the styles
	Tools_DynamicArray<int> xTypesRemap( uMaxNumBotanicalsTypes );
	for ( u_int uI = 0; uI < uMaxNumBotanicalsTypes; uI++ )
	{
		xTypesRemap[uI] = uMaxNumBotanicalsTypes;
	}

	// When making PS2 tree type - make 1 tree type per material used in a component mesh
	// So each input tree template will generate n tree types (1 per material used in the template)
	// Loop over the botanicals editor instances making PS2 data from them
	for ( Asura_ListElt<Asura_Botanicals_Instance*>* pxElt = xInstanceList.Front(); pxElt; pxElt = pxElt->Next() )
	{
		const Asura_Botanicals_Instance* const pxInstance = pxElt->Data();
		const int iTypeID = pxInstance->iTemplateUID;
		const int iTypeIndex = Asura_Botanicals_TemplateManager::GetTemplateIndexFromUID( iTypeID );
		// Ignore if the type doesn't exist
		if ( iTypeIndex < 0 )
		{
			continue;
		}
		const Asura_Botanicals_Template* const pxTemplate = Asura_Botanicals_TemplateManager::GetTemplate( iTypeIndex );
		// Ignore templates which are marked to be merged with the environment
		if ( pxTemplate->bMergeRenderMeshWithEnv == true )
		{
			continue;
		}
		const Asura_Vector_3& xPos = pxInstance->xPosition;
		// Ignore instance if it is not in a valid module
		int iModuleIndex = Asura_Botanicals_InstanceManager::GetModuleIndexFromPosition( xPos, pxInstance->xOrientation );
		if ( iModuleIndex < 0 )
		{
			continue;
		}

		// Loop over the existing types to see if this template index has already been created
		if ( xActiveTypes[iTypeIndex] == false )
		{
			bool bValidType = true;
			// So is it a standard shape or a special botanical shape?
			switch ( pxTemplate->eType )
			{
				case ASURA_BOTANICALS_TEMPLATE_SHAPE:
				{
					if ( pxTemplate->uShapeID == ASURA_HASH_ID_UNSET )
					{
						// This is an invalid type
						Asura_Assert( false, "TEMPLATE_SHAPE but uShapeID not set" );
						bValidType = false;
					}
					break;
				}
				case ASURA_BOTANICALS_TEMPLATE_BOTANICALSHAPE:
				{
					if ( pxTemplate->pxBotanicalShape == NULL )
					{
						// This is an invalid type
						Asura_Assert( false, "TEMPLATE_BOTANICALSHAPE but pxBotanicalsShape is NULL" );
						bValidType = false;
					}
					if ( pxTemplate->pxBotanicalShape->GetToolsShape() == NULL )
					{
						// This is an invalid type
						Asura_Assert( false, "TEMPLATE_BOTANICALSHAPE but tools shape is NULL" );
						bValidType = false;
					}
					break;
				}
				default:
				{
					// Unknown template type (should never happen)
					bValidType = false;
					break;
				}
			}
			Asura_Assert( bValidType, "Unknown template type" );
			// Ignore invalid template types
			if ( bValidType == false )
			{
				continue;
			}

			// A new type so add its meshes to the list of meshes to export
			Asura_PS2_Botanicals_Type_Data& xTypeData = xTypeDatas[uNumTypes];
			if ( pxTemplate->eType == ASURA_BOTANICALS_TEMPLATE_SHAPE )
			{
				// For simple shapes just one mesh, just one LoD!
				xTypeData.xChains.Reset( 1 );
				xTypeData.xChains[0].xComponentIndexes.Reset( 1 );

				// Get the mesh for the standard shape (i.e. not a specific botanical shape)
				Asura_String xShapeName( pxTemplate->xShapeName );
				Asura_String xMeshName;
				Asura_String xExtension;
				Asura_String xPath;
				xShapeName.SplitFilename( &xPath, &xMeshName, &xExtension );
				Asura_Mesh* pxMesh = Asura_Mesh_System::GetMesh( xMeshName.GetString(), false );
				Asura_Assert( pxMesh, "" );
				if ( pxMesh == NULL )
				{
					bValidType = false;
				}
				else
				{
					// Find the material ID for this component
					Tools_DynamicArray<int> xUsedMaterialIDs;
					const int iNumMaterials = GetMaterialsFromMesh( pxMesh, xUsedMaterialIDs );
					Asura_Assertf( iNumMaterials == 1, ("Mesh %s has multiple materials, which is not supported by PS2 botanicals", pxMesh->Name ));
					if ( iNumMaterials != 1 )
					{
						bValidType = false;
						continue;
					}

					Asura_PS2_Botanicals_LoD_Component& xComponent = xComponents[uNumComponents];
					xComponent.fLoDMin = 0.0f;
					xComponent.fLoDMax = Asura_Maths::LargeFloat;
					xComponent.fLoDFadeFar = Asura_Maths::LargeFloat;
					xComponent.fLoDFadeFar = Asura_Maths::LargeFloat;
					xComponent.pxMesh = pxMesh;
					xComponent.xBB =  pxMesh->GetBoundingBox();
					xComponent.eType = ASURA_PS2_BOTANICALS_NORMAL_MESH;
					xComponent.uNumTris = pxMesh->NumberOfTriangles;
					xComponent.uMaterialID = Asura_Material_System::ConvertIDToIndex( xUsedMaterialIDs[0], piMaterialConversionList );

					xTypeData.xChains[0].xComponentIndexes[0] = uNumComponents;
					xTypeData.xChains[0].eType = ASURA_PS2_BOTANICALS_NORMAL_MESH;
					xTypeData.xName = xMeshName;
					uNumComponents++;
				}
			}	
			else
			{
				// Must be a special botanicals shape
				// Know these are not NULL because it has been tested in previous validation code
				Asura_BotanicalShape* const	pxBotanicalShape = pxTemplate->pxBotanicalShape;
				Tools_BotanicalShape* const pxToolsShape = pxBotanicalShape->GetToolsShape(); 
				// Create its LoD chains from the Tools_BotanicalShape data
				xTypeData.xName = pxToolsShape->GetName();
				xTypeData.fInvMaxDrawDistance = pxToolsShape->m_fInvMaxDrawDistance;
				const u_int uNumElements = pxToolsShape->GetNumElements();
				xTypeData.xChains.Reset( uNumElements );
				// Loop over the elements in the botanical shape
				for ( u_int uI = 0; uI < uNumElements; uI++ )
				{
					Tools_BotanicalElement*	const pxElement = pxToolsShape->GetElementByIndex( uI );
					Asura_Assert( pxElement, "" ); 
					if ( pxElement == NULL )
					{
						bValidType = false;
						break;
					}
					Asura_PS2_Botanicals_LoD_Chain& xChain = xTypeData.xChains[uI];

					// Convert element type to our type
					Asura_PS2_Botanicals_Mesh_Type eType;
					switch ( pxElement->m_eType )
					{
						case TOOLS_BOTANICALS_ELEMENTTYPE_1SIDED_MESH:
						{
							eType = ASURA_PS2_BOTANICALS_NORMAL_MESH;
							break;
						}
						case TOOLS_BOTANICALS_ELEMENTTYPE_2SIDED_MESH:
						{
							eType = ASURA_PS2_BOTANICALS_2SIDED_MESH;
							break;
						}
						case TOOLS_BOTANICALS_ELEMENTTYPE_BILLBOARDS:
						{	
							eType = ASURA_PS2_BOTANICALS_BILLBOARDS;
							break;
						}
						case TOOLS_BOTANICALS_ELEMENTTYPE_IMPOSTER:
						{
							eType = ASURA_PS2_BOTANICALS_IMPOSTER;
							break;
						}
						default:
						{
							Asura_Assert( false, "Unknown element type" );
							bValidType = false;
							break;
						}
					}
					if ( bValidType == false )
					{
						break;
					}
					// Ignore imposters
					if ( eType == ASURA_PS2_BOTANICALS_IMPOSTER )
					{
						continue;
					}
					xChain.eType = eType;
					const int iNumLoDs = pxElement->GetNumLODs();
					xChain.xComponentIndexes.Reset( iNumLoDs );
					// Get to each LoD mesh - the lowest level data
					for ( int iLoD = 0; iLoD < iNumLoDs; iLoD++ )
					{
						Tools_BotanicalElementLOD* const pxLoD = pxElement->GetLODByIndex( iLoD );
						Asura_Assert( pxLoD, "" ); 
						if ( pxLoD == NULL )
						{
							bValidType = false;
							break;
						}
						Asura_Mesh* const pxMesh = pxLoD->m_pxMesh;
						// For imposters get the mesh from the imposter array
						if ( eType == ASURA_PS2_BOTANICALS_IMPOSTER )
						{
						}
						Asura_Assert( pxMesh, "" );
						if ( pxMesh == NULL )
						{
							continue;
						}
						// Find the material ID for this component
						Tools_DynamicArray<int> xUsedMaterialIDs;
						const int iNumMaterials = GetMaterialsFromMesh( pxMesh, xUsedMaterialIDs );
						Asura_Assertf( iNumMaterials == 1, ("Mesh %s has multiple materials, which is not supported by PS2 botanicals", pxMesh->Name ));
						if ( iNumMaterials != 1 )
						{
							bValidType = false;
							continue;
						}

						Asura_PS2_Botanicals_LoD_Component& xComponent = xComponents[uNumComponents];
						// 4 LoD values: 0, 1, 2, 3
						//   |
						//   |       +------------+
						//   |      /|            |\
						//   |     / |            | \
						//   |    /  |            |  \
						//   +---+--------------------+------------
						//       3   2            1   0
						//   1.0 -----------------------> 0.0

						// Note in pxLoD the axis is 1->0 not 0->1
						const float fLoD0 = pxLoD->m_fLod0;
						const float fLoD3 = pxLoD->m_fLod1;
						const float fLoD1 = fLoD0 + pxLoD->m_fFadeLength0;
						const float fLoD2 = fLoD3 - pxLoD->m_fFadeLength1;
						xComponent.fLoDMin = 1.0f - fLoD3;
						xComponent.fLoDMax = 1.0f - fLoD0;
						xComponent.fLoDFadeNear = 1.0f - fLoD2;
						xComponent.fLoDFadeFar = 1.0f - fLoD1;
						// Clamp all the values between 0 & 1
						xComponent.fLoDMin = Asura_Maths::Max( 0.0f, xComponent.fLoDMin );
						xComponent.fLoDMin = Asura_Maths::Min( 1.0f, xComponent.fLoDMin );

						xComponent.fLoDMax = Asura_Maths::Max( 0.0f, xComponent.fLoDMax );
						xComponent.fLoDMax = Asura_Maths::Min( 1.0f, xComponent.fLoDMax );

						xComponent.fLoDFadeNear = Asura_Maths::Max( 0.0f, xComponent.fLoDFadeNear );
						xComponent.fLoDFadeNear = Asura_Maths::Min( 1.0f, xComponent.fLoDFadeNear );

						xComponent.fLoDFadeFar = Asura_Maths::Max( 0.0f, xComponent.fLoDFadeFar );
						xComponent.fLoDFadeFar = Asura_Maths::Min( 1.0f, xComponent.fLoDFadeFar );

						Asura_Assert( xComponent.fLoDMin <= xComponent.fLoDFadeNear, "LOD Ranges out of range");
						Asura_Assert( xComponent.fLoDFadeNear <= xComponent.fLoDFadeFar, "LOD Ranges out of range");
						Asura_Assert( xComponent.fLoDFadeFar <= xComponent.fLoDMax, "LOD Ranges out of range");

						xComponent.pxMesh = pxMesh;
						xComponent.xBB = pxMesh->GetBoundingBox();
						xComponent.eType = eType;

						xComponent.uMaterialID = Asura_Material_System::ConvertIDToIndex( xUsedMaterialIDs[0], piMaterialConversionList );

						xChain.xComponentIndexes[iLoD] = uNumComponents;
						uNumComponents++;
					}
					if ( bValidType == false )
					{
						break;
					}
				}
			}

			// Ignore invalid template types
			if ( bValidType == false )
			{
				continue;
			}

			// Need to compute the BB for the chains and then for the type
			const u_int uNumChains = xTypeData.xChains.GetSize();
			u_int uNumTotalComponents = 0;
			Asura_Bounding_Box xTypeBB;
			xTypeBB.Reset();
			for ( u_int uI = 0; uI < uNumChains; uI++ )
			{
				Asura_Bounding_Box xChainBB;
				xChainBB.Reset();
				Asura_PS2_Botanicals_LoD_Chain& xChain = xTypeData.xChains[uI];
				const u_int uNumComponents = xChain.xComponentIndexes.GetSize();
				for ( u_int uComp = 0; uComp < uNumComponents; uComp++ )
				{
					const u_int uIndex = xChain.xComponentIndexes[uComp]; 
					const Asura_PS2_Botanicals_LoD_Component& xComponent = xComponents[uIndex]; 
					const Asura_Bounding_Box xComponentBB( xComponent.xBB );
					xChainBB.Set( xComponentBB );
					uNumTotalComponents++;
				}
				xChain.xBB = xChainBB;
				xTypeBB.Set( xChainBB );
			}
			xTypeData.uNumTotalComponents = uNumTotalComponents;
			xTypeData.xBB = xTypeBB;

			xTypeData.uOldTypeIndex = iTypeIndex;
			xTypeData.uNewTypeIndex = uNumTypes;

			// For remapping type indexes in the instance (tools template index -> game type index )
			xTypesRemap[iTypeIndex] = uNumTypes;
			xActiveTypes[iTypeIndex] = true;
			uNumTypes++;
		}

		Asura_PS2_Botanicals_Instance_Info& xInstanceInfo = xInstanceInfos[uNumInstances];

		xInstanceInfo.uModuleIndex = iModuleIndex;
		xInstanceInfo.uTypeIndex = iTypeIndex;

		// Updated and used by the PS2 renderer at runtime
		Asura_PS2_Botanicals_Instance& xInstance = xInstances[uNumInstances];

		// Make the lcoal2world matrix for this instance
		const Asura_Quat xOrientQuat( pxInstance->xOrientation );
		Asura_Matrix_3x3 xLocal2World;
		xOrientQuat.ToMatrix( xLocal2World );
		xInstance.xLocal2World = xLocal2World * pxInstance->fScale;

		xInstance.fX = xPos.x;
		xInstance.fY = xPos.y;
		xInstance.fZ = xPos.z;

		xInstance.fScaleXYZ = pxInstance->fScale;

		// Need to compute the spherical bounding sphere for this instance
		Asura_PS2_Botanicals_Type_Data& xTypeData = xTypeDatas[ xTypesRemap[ iTypeIndex ] ];
		Asura_Bounding_Box xLocalBB = xTypeData.xBB;

		// Make a world bbox from the local bbox by applying the local2world matrix
		// and then displacing it by the position of the object
		Asura_Bounding_Box xWorldBB;
		xWorldBB.FromOBB( xLocalBB, xLocal2World );
		xWorldBB.Displace( xPos );

		const Asura_Vector_3 xCentre = xWorldBB.CalculateCentre();
		const float fRadius = xWorldBB.CalculateRadius();
		xInstance.fSphereBoundX = xCentre.x;
		xInstance.fSphereBoundY = xCentre.y;
		xInstance.fSphereBoundZ = xCentre.z;
		xInstance.fSphereBoundR = fRadius;

		Asura_SphericalHarmonics_Order2_RGB xWorldSH;
		Asura_GI_Solution::GetStaticIrradiance( xWorldSH, xPos );

		// Rotate to the current instance orientation
		Asura_SphericalHarmonics_Order2_RGB xObjectSH( xWorldSH );
		xObjectSH.RotateToModelspace( xInstance.xLocal2World );

//CSB 24/05/2006	vvvvv	Reduce contrast of SH for billboards and imposters
		const float fSH_ContrastFactor = PS2Strip_Exporter::GetBotanicalBillboardLightingContrastFactor();
		xWorldSH.m_afRedCoefficients[1]   *= fSH_ContrastFactor;
		xWorldSH.m_afRedCoefficients[2]   *= fSH_ContrastFactor;
		xWorldSH.m_afRedCoefficients[3]   *= fSH_ContrastFactor;
		xWorldSH.m_afGreenCoefficients[1] *= fSH_ContrastFactor;
		xWorldSH.m_afGreenCoefficients[2] *= fSH_ContrastFactor;
		xWorldSH.m_afGreenCoefficients[3] *= fSH_ContrastFactor;
		xWorldSH.m_afBlueCoefficients[1]  *= fSH_ContrastFactor;
		xWorldSH.m_afBlueCoefficients[2]  *= fSH_ContrastFactor;
		xWorldSH.m_afBlueCoefficients[3]  *= fSH_ContrastFactor;
//CSB 24/05/2006	^^^^^	Reduce contrast of SH for billboards and imposters

		// Convert the co-efficients to fixed point notation
		const float fFloatScale = 64.0f;
		for ( u_int uI = 0; uI< Asura_SphericalHarmonics_Order2_RGB::NUMBER_OF_COEFFICIENTS; uI++ )
		{
			float fCoeff;

			fCoeff = xObjectSH.m_afCoefficient[uI];
			fCoeff *= fFloatScale;
			// Translate from -127->+128 to 0->255
			fCoeff += 127.0f;
			// Keep within int limits
			fCoeff = Asura_Maths::Max( fCoeff, 0.0f );
			fCoeff = Asura_Maths::Min( fCoeff, +255.0f );
			xObjectSH.m_afCoefficient[uI] = fCoeff;

			fCoeff = xWorldSH.m_afCoefficient[uI];
			fCoeff *= fFloatScale;
			// Translate from -127->+128 to 0->255
			fCoeff += 127.0f;
			// Keep within int limits
			fCoeff = Asura_Maths::Max( fCoeff, 0.0f );
			fCoeff = Asura_Maths::Min( fCoeff, +255.0f );
			xWorldSH.m_afCoefficient[uI] = fCoeff;
		}

		// 0 co-efficents of RGB
		xInstance.uObjectSH0 = ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afRedCoefficients[0] ) & 0xFF ) << 0 ) |
							   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afGreenCoefficients[0] ) & 0xFF ) << 8 ) |
						 	   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afBlueCoefficients[0] ) & 0xFF ) << 16 );

		// 1st co-efficents of RGB
		xInstance.uObjectSH1 = ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afRedCoefficients[1] ) & 0xFF ) << 0 ) |
							   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afGreenCoefficients[1] ) & 0xFF ) << 8 ) |
						 	   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afBlueCoefficients[1] ) & 0xFF ) << 16 );

		// 2nd co-efficents of RGB
		xInstance.uObjectSH2 = ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afRedCoefficients[2] ) & 0xFF ) << 0 ) |
						 	   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afGreenCoefficients[2] ) & 0xFF ) << 8 ) |
						 	   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afBlueCoefficients[2] ) & 0xFF ) << 16 );

		// 3rd co-efficents of RGB
		xInstance.uObjectSH3 = ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afRedCoefficients[3] ) & 0xFF ) << 0 ) |
						 	   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afGreenCoefficients[3] ) & 0xFF ) << 8 ) |
						 	   ( ( Asura_Maths::FloatToInt_Near( xObjectSH.m_afBlueCoefficients[3] ) & 0xFF ) << 16 );
		
		// 0 co-efficents of RGB
		xInstance.uWorldSH0 = ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afRedCoefficients[0] ) & 0xFF ) << 0 ) |
							  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afGreenCoefficients[0] ) & 0xFF ) << 8 ) |
						 	  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afBlueCoefficients[0] ) & 0xFF ) << 16 );

		// 1st co-efficents of RGB
		xInstance.uWorldSH1 = ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afRedCoefficients[1] ) & 0xFF ) << 0 ) |
							  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afGreenCoefficients[1] ) & 0xFF ) << 8 ) |
						 	  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afBlueCoefficients[1] ) & 0xFF ) << 16 );

		// 2nd co-efficents of RGB
		xInstance.uWorldSH2 = ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afRedCoefficients[2] ) & 0xFF ) << 0 ) |
						 	  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afGreenCoefficients[2] ) & 0xFF ) << 8 ) |
						 	  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afBlueCoefficients[2] ) & 0xFF ) << 16 );

		// 3rd co-efficents of RGB
		xInstance.uWorldSH3 = ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afRedCoefficients[3] ) & 0xFF ) << 0 ) |
						 	  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afGreenCoefficients[3] ) & 0xFF ) << 8 ) |
						 	  ( ( Asura_Maths::FloatToInt_Near( xWorldSH.m_afBlueCoefficients[3] ) & 0xFF ) << 16 );
		
		Asura_Assert( uNumInstances < 0xFFFF, "" );
		xInstance.usGlobalInstanceIndex = 0xFFFF;
		xInstance.usPadding = 0xBEAF;

		uNumInstances++;
	}

	// Loop over the instance infos remapping the types in case there are some
	// types which didn't have any instances that did not get exported
	for ( u_int uI = 0; uI < uNumInstances; uI++ )
	{
		Asura_PS2_Botanicals_Instance_Info& xInstanceInfo = xInstanceInfos[uI];
		const u_int iOldTypeIndex = xInstanceInfo.uTypeIndex;
		const u_int iNewTypeIndex = xTypesRemap[iOldTypeIndex];
		Asura_Assert( iNewTypeIndex < uNumTypes, "" );
		xInstanceInfo.uTypeIndex = iNewTypeIndex;
	}

	// Bubble sort the instance infos by type index as 1st key and module index as 2nd key
	// So end up with a list of instances sorted by module: and within a module sorted by type
	// This is so when rendering instances references are near each other to save cache thrashing
	for ( u_int uI = 0; uI < uNumInstances; uI++ )
	{
		for ( u_int uJ = uI + 1; uJ < uNumInstances; uJ++ )
		{
			const int iIndex1 = xInstanceInfos[uI].uModuleIndex;
			const int iIndex2 = xInstanceInfos[uJ].uModuleIndex;
			const int iTypeID1 = xInstanceInfos[uI].uTypeIndex;
			const int iTypeID2 = xInstanceInfos[uJ].uTypeIndex;
			// 1st Key is type ID index
			// 2nd Key is module index
			if ( ( iTypeID2 < iTypeID1 ) || ( ( iTypeID2 == iTypeID1 ) && ( iIndex2 < iIndex1 ) ) )
			{
				// Swap them over
				const Asura_PS2_Botanicals_Instance_Info xTemp = xInstanceInfos[uI];
				xInstanceInfos[uI] = xInstanceInfos[uJ];
				xInstanceInfos[uJ] = xTemp;

				// Swap over the instance runtime data
				const Asura_PS2_Botanicals_Instance xTemp2 = xInstances[uI];
				xInstances[uI] = xInstances[uJ];
				xInstances[uJ] = xTemp2;
			}
		}
	}

	Tools_DynamicArray<Asura_PS2_Botanicals_Type_Header> xTypeHeaders;
	// a 2D array:  array of Module_Header's per style
	Tools_DynamicArray< Tools_DynamicArray<Asura_PS2_Botanicals_Module_Header> > xModuleHeadersPerType( uNumTypes );

	// index of which instance in xInstances array to render
	Tools_DynamicArray< Tools_DynamicArray< Tools_DynamicArray<int> > > xInstancesPerModulePerType( uNumTypes );

	const u_int uNumModules = Asura_Module_System::GetNumberOfModules();

	// Loop over the botanicals infos adding the instances to the module headers
	// They are sorted by type and then by module
	u_int uOldModuleIndex = uNumModules;
	u_int uOldTypeIndex = uNumTypes;
	int iTotalHeadersSizeInBytes = 0;
	int iCurrentModulePlace = -1;
	for ( u_int uI = 0; uI < uNumInstances; uI++ )
	{
		Asura_PS2_Botanicals_Instance_Info& xInfo = xInstanceInfos[uI];
		const u_int uModuleIndex = xInfo.uModuleIndex;
		const u_int uTypeIndex = xInfo.uTypeIndex;

		// Is this a new style
		if ( uTypeIndex != uOldTypeIndex )
		{
			xTypeHeaders[uTypeIndex].usNumModules = 0;
			xTypeHeaders[uTypeIndex].usTypeID = static_cast<u_short>( uTypeIndex );
			xTypeHeaders[uTypeIndex].uNumInstances = 0;
			Tools_DynamicArray<Asura_PS2_Botanicals_Module_Header>& xModules = xModuleHeadersPerType[uTypeIndex];
			// Zero out the module array for this new style
			xModules.Reset();
			uOldTypeIndex = uTypeIndex;
			uOldModuleIndex = uNumModules;
			iCurrentModulePlace = -1;
			iTotalHeadersSizeInBytes += sizeof( Asura_PS2_Botanicals_Type_Header );
		}
		// Is this a new module
		if ( uModuleIndex != uOldModuleIndex )
		{
			xTypeHeaders[uTypeIndex].usNumModules++;
			uOldModuleIndex = uModuleIndex;
			Asura_PS2_Botanicals_Module_Header xModHdr;
			xModHdr.usModuleID = static_cast<u_short>( uModuleIndex );
			xModHdr.usNumInstances = 0;
			Tools_DynamicArray<Asura_PS2_Botanicals_Module_Header>& xModules = xModuleHeadersPerType[uTypeIndex];
			// Add the new module header to the array
			xModules += xModHdr;
			iCurrentModulePlace++;
			Asura_Assert( iCurrentModulePlace+1 == static_cast<int>( xModules.GetSize() ), "" );

			Tools_DynamicArray<int>& xInstancesPerModule = xInstancesPerModulePerType[uTypeIndex].operator[]( iCurrentModulePlace );
			// Zero out the instance array for this new module
			// Prime it with an allocation of 64 instances
			xInstancesPerModule.Reset( 64 );
			iTotalHeadersSizeInBytes += sizeof( Asura_PS2_Botanicals_Module_Header );
		}

		// Add this instance data to the correct place
		Tools_DynamicArray<int>& xInstancesPerModule = xInstancesPerModulePerType[uTypeIndex].operator[]( iCurrentModulePlace );
		xInstancesPerModule += uI;

		// Update statistics
		Asura_PS2_Botanicals_Module_Header& xModHdr =  xModuleHeadersPerType[uTypeIndex].operator[] ( iCurrentModulePlace );
		// One more instance in this module
		xModHdr.usNumInstances++;

		Asura_Assert( xInstancesPerModule.GetSize() == xModHdr.usNumInstances, "" );

		// One more instance in this style
		xTypeHeaders[uTypeIndex].uNumInstances++;
		iTotalHeadersSizeInBytes += sizeof( Asura_PS2_Botanicals_Instance );
	}

	Asura_PS2_Botanicals_Header xHeader;
	xHeader.uNumTypes = uNumTypes;
	xHeader.uNumTotalComponents = uNumComponents;
	xHeader.uNumTotalInstances = uNumInstances;
	xHeader.uTotalHeadersSizeInBytes = iTotalHeadersSizeInBytes;
	// Unknown at the moment: seek back through file to fill this data in
	xHeader.uTotalTypeInfosSizeInBytes = 0;

	////////////////////////////////////////////////////////////////////////////
	//
	// Start of output code
	//
	////////////////////////////////////////////////////////////////////////////

	// Output header message.
	{
		const Asura_Char* szFormat =
			"Started exporting PS2 Botanicals at %02i:%02i:%02i on %02i.%02i.%04i.\n"
			"Exported by Asura built %s\n"
			"\n";

		SYSTEMTIME xLocalTime;
		GetLocalTime( &xLocalTime );

		PS2Strip::LogInfo(
			szFormat,
			xLocalTime.wHour,
			xLocalTime.wMinute,
			xLocalTime.wSecond,
			xLocalTime.wDay,
			xLocalTime.wMonth,
			xLocalTime.wYear,
			Asura_BuildTimeAndDate
			);
	}
	u_int uTotalNumBytesWritten = 0;

	// Leave a gap for the chunk header. Record the cur file ptr first so we
	// can return to this point and fill it out later.
	u_int uFilePtrForChunkHdr;
	BOTANICALS_RECORD_FILE_PTR( pOutput, uInitialFilePtr, uTotalNumBytesWritten, uFilePtrForChunkHdr );
	SKIP_DATA( pOutput, uTotalNumBytesWritten, sizeof(Asura_Chunk_ResourceFile) );

	// Write out the resource filename
	// Make the name multiple of 4 bytes - 1
	const Asura_Char* const BOTANICALS_DATA_RESOURCE_FILENAME = "PS2Botanicals  ";
	const u_int uSizeOfResourceFilenameInBytes = strlen(BOTANICALS_DATA_RESOURCE_FILENAME) + 1;
	Asura_Assert( !(uSizeOfResourceFilenameInBytes & 3), "" );

	u_int uSizeOfResourceFilenamePlusPaddingInBytes = uSizeOfResourceFilenameInBytes;

	// We're about to write the resource filename. Before doing so, write out
	// padding so that the data will be QW aligned to speed up loading at
	// runtime. We'll do this by prepending spaces to the resource filename.
	{
		const u_int uRem = ((uTotalNumBytesWritten + uSizeOfResourceFilenameInBytes) & 15);
		if( uRem )
		{
			const u_int uPad = 16 - uRem;
			BOTANICALS_PAD_TO_QW_ALIGNMENT( pOutput, uTotalNumBytesWritten, uPad );
			uSizeOfResourceFilenamePlusPaddingInBytes += uPad;
		}
	}

	// Write actual resource filename.
	WRITE_DATA( pOutput, uTotalNumBytesWritten, BOTANICALS_DATA_RESOURCE_FILENAME, uSizeOfResourceFilenameInBytes );
	Asura_Assert( !(uTotalNumBytesWritten & 15), "Should be QW aligned at this point." );

	///////////////////////////////////////////////////////////////////////////
	//
	// Write out the botanicals data
	//
	///////////////////////////////////////////////////////////////////////////

	///////////////////////////////////////////////////////////////////////////
	//
	// The PS2 Botanicals mesh data format Version 6 is:
	//
	// Asura_PS2_Botanicals_Header
	// {
	//    u_int uNumTypes;
	//    u_int uNumTotalComponents;
	//    u_int uNumTotalInstances;
	//    u_int uTotalHeadersSizeInBytes;
	//    u_int uTotalTypeInfosSizeInBytes;
	// };
	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Header
	//    {
	//       u_short usTypeID;
	//   	 u_short usNumModules;
	//       u_int uNumInstances;
	//    };
	//    Array of usNumModules
	//    {
	//       struct Asura_PS2_Botanicals_Module_Header
	//       {
	//       	u_short usModuleID;
	//       	u_short usNumInstances;
	//       };
	//		 Array of usNumInstances;
	//       {
	//		    struct Asura_PS2_Botanicals_Instance
	//			{
	//			   This first part is EE used data only
	//			   u_short usGlobalInstanceIndex;
	//			   u_short usPadding;
	//
	//			   float fSphereBoundX;
	//			   float fSphereBoundY;
	//			   float fSphereBoundZ;
	//			   float fSphereBoundR;
	//
	//			   float fScaleXYZ;
	//
	//			   This should be DMA friendly packet data
	//			   float fX;
	//			   float fY;
	//			   float fZ;
	//
	//			   Asura_Matrix_3x3 xLocal2World;
	//
	//			   u_int uWorldSH0;
	//			   u_int uWorldSH1;
	//			   u_int uWorldSH2;
	//			   u_int uWorldSH3;
	//
	//			   u_int uObjectSH0;
	//			   u_int uObjectSH1;
	//			   u_int uObjectSH2;
	//			   u_int uObjectSH3;
	//			};
	//       }
	//    }
	// }
	//
	// This data is separate to the Asura_PS2_Botanicals_Type_Header data because
	// one day this type info data might be passed to VU1 for LoD selection
	//
	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Info
	//    {
	//       Asura_Bounding_Box xBB;
	//		 float fInvMaxDrawDistance;
	// 	     u_int uNumComponents;
	//		 Array of uNumComponents;
	//       {
	//          struct Asura_PS2_Botanicals_Component_Info
	//          {
	//			   int iTextureID;					// On export it is a material ID which is converted to a texture ID & material flags
	//			   u_int uMaterialFlags;
	//			   u_short usType;					// enum Asura_PS2_Botanicals_Mesh_Type
	//			   u_short usNumTris;					
	//			   float fLoDMin;
	//			   float fLoDMax;
	// 	           float fLoDConstant1;
	// 	           float fLoDConstant2;
	// 	           u_int uComponentIndexAndType; // u_int : top 8-bits is type, bottom 24-bits is component index into component DMA data
	//			};
	//       }
	//    };
	// }
	//
	// Array of uNumTotalComponents
	// {
	//    DMA data for the mesh component: referenced by a DMA CALL tag so the DMA data includes the DMA RET tag
	// }
	//
	///////////////////////////////////////////////////////////////////////////

	// Asura_PS2_Botanicals_Header xHeader
	// Record position of the header so we can seek back through the file and fill in the details
	u_int uFilePtrForHeader;
	BOTANICALS_RECORD_FILE_PTR( pOutput, uInitialFilePtr, uTotalNumBytesWritten, uFilePtrForHeader );
	WRITE_DATA( pOutput, uTotalNumBytesWritten, &xHeader, sizeof( xHeader ) );

	u_int uNumTotalInstances = 0; 
	u_int uNumTotalMods = 0;
	for ( u_int uTypeIndex = 0; uTypeIndex < uNumTypes; uTypeIndex++ )
	{
		u_int uNumInstancesInType = 0;

		// Write out the style header
		const Asura_PS2_Botanicals_Type_Header& xTypeHdr = xTypeHeaders[uTypeIndex];
		WRITE_DATA( pOutput, uTotalNumBytesWritten, &xTypeHdr, sizeof( xTypeHdr ) );

		const Tools_DynamicArray< Asura_PS2_Botanicals_Module_Header >& xModuleHeaders = xModuleHeadersPerType[uTypeIndex];
		const Tools_DynamicArray< Tools_DynamicArray<int> >& xInstancesPerType = xInstancesPerModulePerType[uTypeIndex];

		const u_int uNumModsInType = xTypeHdr.usNumModules;
		uNumTotalMods += uNumModsInType;
		for ( u_int uModuleIndex = 0; uModuleIndex < uNumModsInType; uModuleIndex++ )
		{
			u_int uNumInstancesInModule = 0;

			// Write out the module header
			const Asura_PS2_Botanicals_Module_Header& xModuleHdr = xModuleHeaders[uModuleIndex];
			WRITE_DATA( pOutput, uTotalNumBytesWritten, &xModuleHdr, sizeof( xModuleHdr ) );

			const Tools_DynamicArray<int>& xInstancesPerModule = xInstancesPerType[uModuleIndex];

			const u_int uNumInstances = xModuleHdr.usNumInstances;
			// Asura_PS2_Botanicals_Instance xInstanceInfo[usNumInstances];
			for ( u_int uI = 0; uI < uNumInstances; uI++ )
			{
				const u_int uInstanceIndex = xInstancesPerModule[uI];
				Asura_PS2_Botanicals_Instance& xInstance = xInstances[uInstanceIndex];
				xInstance.usGlobalInstanceIndex = static_cast<u_short>( uNumTotalInstances & 0xFFFF );

				WRITE_DATA( pOutput, uTotalNumBytesWritten, &xInstance, sizeof( xInstance ) );
				uNumInstancesInModule++;
				uNumInstancesInType++;
				uNumTotalInstances++;
			}
			Asura_Assert( uNumInstancesInModule == xModuleHdr.usNumInstances, "" );
		}
		Asura_Assert( uNumInstancesInType == xTypeHdr.uNumInstances, "" );
	}
	Asura_Assert( uNumTotalInstances == xHeader.uNumTotalInstances, "" );

	// Array of uNumTypes
	// {
	//    struct Asura_PS2_Botanicals_Type_Info
	//    {
	//       Asura_Bounding_Box xBB;
	//		 float fInvMaxDrawDistance;
	// 	     u_int uNumComponents;
	//		 Array of uNumComponents;
	//       {
	//          struct Asura_PS2_Botanicals_Component_Info
	//          {
	// 	           float fLoDFadeNear;
	// 	           float fLoDFadeFar;
	// 	           u_int uComponentIndexAndType; // u_int : top 8-bits is type, bottom 24-bits is component index into component DMA data
	//			};
	//       }
	//    };
	// }
	PS2Strip::LogInfo( "\nWriting botanical type info data... " );
	u_int uTotalTypeInfosSizeInBytes = uTotalNumBytesWritten;
	for ( u_int uTypeIndex = 0; uTypeIndex < uNumTypes; uTypeIndex++ )
	{
		u_int uOutputSizeInBytes = uTotalNumBytesWritten;
		const Asura_PS2_Botanicals_Type_Data& xTypeData = xTypeDatas[uTypeIndex];
		Asura_Assert( xTypeData.uNewTypeIndex == uTypeIndex, "" );

		const Asura_Bounding_Box& xBB = xTypeData.xBB;
		// Asura_Bounding_Box xBB;
		WRITE_DATA( pOutput, uTotalNumBytesWritten, &xBB, sizeof( xBB ) );

		// float fInvMaxDrawDistance;
		WRITE_DATA( pOutput, uTotalNumBytesWritten, &xTypeData.fInvMaxDrawDistance, sizeof(float) );

		const u_int uNumComponents = xTypeData.uNumTotalComponents;
		// u_int uNumComponents;
		WRITE_DATA( pOutput, uTotalNumBytesWritten, &uNumComponents, sizeof( uNumComponents ) );

		// Asura_PS2_Botanicals_Component_Info xComponentInfo[uNumComponents];
		u_int uOutputComp = 0;
		const u_int uNumChains = xTypeData.xChains.GetSize();
		for ( u_int uI = 0; uI < uNumChains; uI++ )
		{
			const Asura_PS2_Botanicals_LoD_Chain& xChain = xTypeData.xChains[uI];
			const u_int uNumComps = xChain.xComponentIndexes.GetSize();
			const u_int uTypeValue = ( xChain.eType & 0xFF ) << 24;
			for ( u_int uComp = 0; uComp < uNumComps; uComp++ )
			{
				const u_int uIndex = ( xChain.xComponentIndexes[uComp] ) & 0x00FFFFFF;
				Asura_Assert( uIndex == xChain.xComponentIndexes[uComp], "" );
				const Asura_PS2_Botanicals_LoD_Component& xComponent = xComponents[uIndex]; 

				Asura_PS2_Botanicals_Component_Info xComponentInfo;
				xComponentInfo.iTextureID = xComponent.uMaterialID;
				xComponentInfo.uMaterialFlags = 0xDEADBEAF;
				// This data is over-written when the data is loaded into game
				// with the actual triangle count from the strips in the PS2 mesh
				xComponentInfo.usNumTris = static_cast<u_short>( xComponent.uNumTris );
				xComponentInfo.usType = static_cast<u_short>( xComponent.eType );
				xComponentInfo.fLoDMin = xComponent.fLoDMin;
				xComponentInfo.fLoDMax = xComponent.fLoDMax;
				xComponentInfo.fLoDConstant1 = xComponent.fLoDFadeNear - xComponent.fLoDMin;
				xComponentInfo.fLoDConstant2 = xComponent.fLoDMax - xComponent.fLoDFadeFar;

				// Stop divide by 0 problems
				if ( Asura_Maths::Abs( xComponentInfo.fLoDConstant1 ) > 0.0f )
				{
					xComponentInfo.fLoDConstant1 = 1.0f / xComponentInfo.fLoDConstant1;
				}
				else
				{
					xComponentInfo.fLoDConstant1 = Asura_Maths::LargeFloat;
				}

				if ( Asura_Maths::Abs( xComponentInfo.fLoDConstant2 ) > 0.0f )
				{
					xComponentInfo.fLoDConstant2 = 1.0f / xComponentInfo.fLoDConstant2;
				}
				else
				{
					xComponentInfo.fLoDConstant2 = Asura_Maths::LargeFloat;
				}

				xComponentInfo.uComponentIndexAndType = uTypeValue | uIndex;

				// Asura_PS2_Botanicals_Component_Info
				WRITE_DATA( pOutput, uTotalNumBytesWritten, &xComponentInfo, sizeof( xComponentInfo ) );
				uOutputComp++;
			}
		}
		Asura_Assert( uOutputComp == uNumComponents, "" );

		// Compute how much data in this type and output debug info to the log file
		uOutputSizeInBytes = uTotalNumBytesWritten - uOutputSizeInBytes;
		PS2Strip::LogInfo( "\nBotanicalType[%d] '%s' numComps %d BB %f %f %f -> %f %f %f Range %f %d bytes",
						   uTypeIndex, 
						   xTypeData.xName.GetString(), 
						   uNumComponents, 
						   xBB.GetMin().x, xBB.GetMin().y, xBB.GetMin().z,
						   xBB.GetMax().x, xBB.GetMax().y, xBB.GetMax().z,
						   1.0f / xTypeData.fInvMaxDrawDistance,
						   uOutputSizeInBytes
						 );
	}
	uTotalTypeInfosSizeInBytes = uTotalNumBytesWritten - uTotalTypeInfosSizeInBytes;
	// Per type:
	//	Asura_Bounding_Box xBB;
	//	u_int uNumComponents;
	// Per component:
	//  int iTextureID;					// On export it is a material ID which is converted to a texture ID & material flags
	//  u_int uMaterialFlags;
	//  u_int uType;					    // enum Asura_PS2_Botanicals_Mesh_Type
	//	float fLoDFadeNear;
	//	float fLoDFadeFar;
	//  float fLoDConstant1;
	//  float fLoDConstant2;
	//	u_int uComponentIndexAndType; 
#ifdef ASURA_ASSERT
	const u_int uTypeInfosSizeInBytes = ( 7 * 4 + 4 ) * xHeader.uNumTypes + ( 8 * 4 * xHeader.uNumTotalComponents );
	Asura_Assert( uTotalTypeInfosSizeInBytes == uTypeInfosSizeInBytes, "" );
#endif // #ifdef ASURA_ASSERT

	// Write out the DMA data for the mesh components (different formats for different mesh types)
	// Each set of data is referenced by a DMA CALL tag, so the DMA data contains the RET tag and the number of qwords
	// The DMA RET tag nQwords is used to work out the size of each set of DMA data
	for ( u_int uI = 0; uI < uNumComponents; uI++ )
	{
		const Asura_PS2_Botanicals_LoD_Component& xComponent = xComponents[uI];
		Asura_Mesh* pxMesh = xComponent.pxMesh;
		const char* const szMeshName = pxMesh->Name.GetString();
		// Convert the Asura_Mesh data to PS2 DMA data
		if ( PS2Strip::ExportBotanicalObject( pOutput, pxMesh, szMeshName, xComponent.eType ) == false )
		{
			return false;
		}
		uTotalNumBytesWritten += s_uNumBytesWrittenBotanicalObject;
	}

	// Go back and fill in the header details
	if( INVALID_SET_FILE_POINTER == SetFilePointer( pOutput, uFilePtrForHeader, NULL, FILE_BEGIN) )
	{
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
		return false;
	}
	int iIgnore = 0;
	xHeader.uTotalTypeInfosSizeInBytes = uTotalTypeInfosSizeInBytes;
	WRITE_DATA( pOutput, iIgnore, &xHeader, sizeof( xHeader ) );

	// Don't seek back to end of file because next operation is going to seek through the file again!

	// Go back and fill out chunk hdr.
	// Seek back to beginning of chunk hdr bytes.
	if( INVALID_SET_FILE_POINTER == SetFilePointer( pOutput, uFilePtrForChunkHdr, NULL, FILE_BEGIN) )
	{
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
		return false;
	}

	// Prepare hdr.
	Asura_Chunk_ResourceFile xChunkHdr;
	xChunkHdr.ID       = ASURA_CHUNK_RESOURCEFILE;
	xChunkHdr.Size     = uTotalNumBytesWritten;
	xChunkHdr.Version  = ASURA_PS2_BOTANICALS_CHUNK_VERSION;
	xChunkHdr.Flags    = 0;
	xChunkHdr.Type     = ASURA_RESOURCEFILE_TYPE_PLATFORMSPECIFIC;
	xChunkHdr.SubType  = ASURA_RESOURCEFILE_TYPE_PS2_BOTANICALS;
	xChunkHdr.FileSize = uTotalNumBytesWritten - (sizeof(xChunkHdr) + uSizeOfResourceFilenamePlusPaddingInBytes);

	// Output. Note that we've already included these bytes in the
	// cumulative count when we skipped the bytes.
	WRITE_DATA( pOutput, iIgnore, &xChunkHdr, sizeof(xChunkHdr) );

	// End chunk hdr write.

	// Seek back to EOF.
	const u_int uEOF = SetFilePointer( pOutput, 0, NULL, FILE_END );
	if( INVALID_SET_FILE_POINTER == uEOF )
	{
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. Reason: %s\n", PS2Strip::GetLastSystemError() );
		return false;
	}

	// Check our total byte count.
	Asura_Assert( uEOF == uInitialFilePtr + uTotalNumBytesWritten, "Total byte count is incorrect." );

	// Output end log message
	const Asura_Char* szFormat =
		"Finished exporting PS2 Botanicals at %02i:%02i:%02i on %02i.%02i.%04i.\n"
		"numModules %d Instances %d Types %d Modules %d\n"
		"\n";

	SYSTEMTIME xLocalTime;
	GetLocalTime( &xLocalTime );

	PS2Strip::LogInfo(
						szFormat,
						xLocalTime.wHour,
						xLocalTime.wMinute,
						xLocalTime.wSecond,
						xLocalTime.wDay,
						xLocalTime.wMonth,
						xLocalTime.wYear,
						uNumModules, uNumInstances, uNumTypes, uNumTotalMods
					);

	Tools_Level_Export::AddMessage( Tools_Level_Export::MT_INFORMATION,  "%d bytes written PS2 Botanicals (%d instances %d types %d modules)", 
			uTotalNumBytesWritten, uNumInstances, uNumTypes, uNumTotalMods );

#undef BOTANICALS_PAD_TO_QW_ALIGNMENT
#undef BOTANICALS_RECORD_FILE_PTR
	return true;
}

// JAT 24/03/06 - based on BeginObjectExport
void PS2Strip_ExporterImplementation::BeginBotanicalObjectExport( Asura_Mesh* pxMesh, enum Asura_PS2_Botanicals_Mesh_Type eType )
{
	Asura_Assert( PS2Strip::GetCurExportType() == PS2Strip::EXPORTTYPE_BOTANICAL_OBJECT, "" );

	s_uNumBytesWrittenBotanicalObject = 0;

	s_fStartProgressPercentage = 0.0f;
	s_fProgressPercentageDelta = 50.0f;

	s_pxCurMesh = pxMesh;
	s_eBotanicalType = eType;

	BeginExportInternal();
}

// JAT 24/03/06 - based on OutputObject()
bool PS2Strip_ExporterImplementation::OutputBotanicalObject( void )
{
#define CHECK( e ) { if( !(e) ) { return false; } }

	// Output progress message.
	PS2Strip::LogInfo( "\nWriting botanical object data... " );

	HANDLE pvOutputFile = PS2Strip::GetOutputHandle();
	Asura_Assert( INVALID_HANDLE_VALUE != pvOutputFile, "" ); // Shouldn't have got this far if failed to open file.

	s_uNumBytesWrittenBotanicalObject = 0;

	int   iTotalBytesWritten    = 0;   
	bool  bDoneMultipleMatError = false;
	bool  bDoneTexCoordError    = false;
	bool  bDoneUntexturedError  = false;

	// Object name.
	const int iBytesWrittenBeforeObjName = iTotalBytesWritten;
	const char* szObjectName = PS2Strip::GetCurAssetName();
	const u_int uObjectNameSize = strlen(szObjectName) + 1;
	WRITE_DATA( pvOutputFile, iTotalBytesWritten, szObjectName, uObjectNameSize );
	PAD_TO_ALIGNMENT( pvOutputFile, 4, iTotalBytesWritten, iBytesWrittenBeforeObjName );

	// Shape hdr.
	Asura_PS2_Shape_ShapeHdr xShapeHdr;

	xShapeHdr.usTotalTris   = static_cast<u_short>( s_uNumUniqueMeshTris );
	xShapeHdr.usNumStrips   = static_cast<u_short>( s_xPostProcessedStrips.size() );
	xShapeHdr.fLODThreshold = s_pxCurMesh->m_fMinDistance;

	WRITE_DATA( pvOutputFile, iTotalBytesWritten, &xShapeHdr, sizeof(Asura_PS2_Shape_ShapeHdr) );

	// u_int uNumBytes;
	u_int uNumBytes = 0;
	const u_int uNumBytesFilePtr = SetFilePointer( pvOutputFile, 0, NULL, FILE_CURRENT );
	WRITE_DATA( pvOutputFile, iTotalBytesWritten, &uNumBytes, sizeof(u_int) );

	uNumBytes = iTotalBytesWritten;

	// The DMA chain.
	if( s_uNumUniqueMeshTris )
	{
		Tools_PS2_DMAChainWriter xChainWriter;

		CHECK( xChainWriter.Init(pvOutputFile) );

		CHECK( xChainWriter.BeginChain() );

		int iObjMaterialID = INT_MIN;

		for( u_int uStrip = 0; uStrip < s_xPostProcessedStrips.size(); ++uStrip )
		{
			// For the first two strips we need to fill alpha values. We
			// don't actually support any vertex alpha on objects right now
			// but the render path is the same. However, once we've ensured
			// both buffers are full, we can ignore it.
			if( uStrip < 2 )
			{
				// Temporarily change CYCLE.
				CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(4, 0, false)) );

				// On the first strip setup STMASK for the alpha filling.
				if( !uStrip )
				{
					// Just write to w of colour QW...
					CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STMASK(false)) );
					const PS2Strip_VIF1_Mask xMask =
					{
						3, 3, 3, 3,
							3, 3, 3, 3,
							3, 3, 3, 1,
							3, 3, 3, 3,
					};
					CHECK( xChainWriter.AddDataWord(reinterpret_cast<const u_int&>(xMask)) );

					// ... and fill it with 1.
					CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STROW(false)) );
					const float afFillRow[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
					CHECK( xChainWriter.AddData(afFillRow, 16) );
				}

				// NB: The CMD here is irrelevant since we're
				// just using fill values and not sending any
				// data.
				CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(0,
					ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP * 4,
					PS2STRIP_VIF1_CMD_V4_32 | PS2STRIP_VIF1_MASK_PROCESS_ENABLE,
					false)) );

				// Set CYCLE ready for proper strip data.
				CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_STCYCL(1, 4, false)) );
			}

			const PostProcessedStrip* pxPPStrip = s_xPostProcessedStrips[uStrip];

			const u_int uNumVertices = pxPPStrip->xQuadletIndices.size();
			Asura_Assert( uNumVertices <= ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP, "" );

#define GET_QUADLET(idx_within_strip) (s_axUniqueMeshVtxIdxQuadlets[pxPPStrip->xQuadletIndices[uVtx] & ~PostProcessedStrip::NEEDS_ADC_BIT])

			// Positions.
			CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(0, uNumVertices, PS2STRIP_VIF1_CMD_V3_32, false)) );
			for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
			{
				const u_int uPosIdx = GET_QUADLET(uVtx).GetPosIdx();
				CHECK( xChainWriter.AddData(&s_axMeshPositions[uPosIdx], 12) );
			}

			// STs.
			{
				short aiCoords[ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP * 2];

				CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(1, uNumVertices, PS2STRIP_VIF1_CMD_V2_16, false)) );

				for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
				{
					const u_int uMatInfoIdx = GET_QUADLET(uVtx).GetMatInfoIdx();
					const PS2Strip_MatInfo& xMatInfo = s_axMeshMatInfo[uMatInfoIdx];

					if( !bDoneTexCoordError && !xMatInfo.CheckTexCoords() )
					{
						PS2Strip::LogError( "ERROR: Object \"%s\" has texture coordinate outside valid PS2 range.", szObjectName );
						bDoneTexCoordError = true;
					}

					const int iVtxMatID = xMatInfo.m_iMaterialID;
					if( ASURA_MATERIAL_NONE == iVtxMatID || ASURA_TEXTURE_NONE == Asura_Material_System::GetTextureID(iVtxMatID) )
					{
						if( !bDoneUntexturedError )
						{
							PS2Strip::LogError( "ERROR: Object \"%s\" has untextured polys.", szObjectName );
							bDoneUntexturedError = true;
						}
					}
					else
					{
						if( INT_MIN == iObjMaterialID )
						{
							iObjMaterialID = iVtxMatID;
						}
						else if( !bDoneMultipleMatError && iVtxMatID != iObjMaterialID )
						{
							PS2Strip::LogError( "ERROR: Object \"%s\" uses multiple materials and/or textures.", szObjectName );
							bDoneMultipleMatError = true;
						}
					}

					aiCoords[(uVtx << 1) + 0] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMatInfo.m_afU[0] );
					aiCoords[(uVtx << 1) + 1] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, xMatInfo.m_afV[0] );
				}

				CHECK( xChainWriter.AddData(aiCoords, uNumVertices * 4) );
			}

			// Normals and ADC bits.
			{
				CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_UNPACKR(3, uNumVertices, PS2STRIP_VIF1_CMD_V4_5, false)) );

				PS2Strip_VIF1_V4_5_Data axNrmsAndADCs[ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP];
				for( u_int uVtx = 0; uVtx < uNumVertices; ++uVtx )
				{
					const Asura_Vector_3& xNormal = s_axMeshNormals[GET_QUADLET(uVtx).GetNrmIdx()];
					axNrmsAndADCs[uVtx].usX = Tools_PS2Strip_ConvertFloatForV45( xNormal.x );
					axNrmsAndADCs[uVtx].usY = Tools_PS2Strip_ConvertFloatForV45( xNormal.y );
					axNrmsAndADCs[uVtx].usZ = Tools_PS2Strip_ConvertFloatForV45( xNormal.z );

					axNrmsAndADCs[uVtx].usW = ( 0 != (pxPPStrip->xQuadletIndices[uVtx] & PostProcessedStrip::NEEDS_ADC_BIT) );
				}

				// NB: This deals with padding to word alignment on
				// our behalf.
				CHECK( xChainWriter.AddData(axNrmsAndADCs, uNumVertices * 2) );
			}

#undef GET_QUADLET

			// Pass vertex count in ITOPS.
			CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_ITOP(uNumVertices, false)) );

			// And kick off the microprogram to render this strip.
			CHECK( xChainWriter.AddVifCode(PS2STRIP_VIF1_SET_MSCNT(false)) );
		}
		// End loop through strips.

		CHECK( xChainWriter.EndChain() );        

		// Update cumulative byte count.
		const u_int uBytesWrittenForChain = xChainWriter.GetTotalBytesWrittenForLastChain();
		Asura_Assert( !(uBytesWrittenForChain & 15), "" );
		iTotalBytesWritten += uBytesWrittenForChain;
	}
	// End DMA chain write.
	uNumBytes = iTotalBytesWritten - uNumBytes;
	// Go back and fill in the size of the dma data
	if ( INVALID_SET_FILE_POINTER == SetFilePointer( pvOutputFile, uNumBytesFilePtr, NULL, FILE_BEGIN ) )
	{
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr (botanical object '%s'). Reason: %s\n", szObjectName, PS2Strip::GetLastSystemError() );
		return false;
	}
	int iIgnore = 0;
	WRITE_DATA( pvOutputFile, iIgnore, &uNumBytes, sizeof( u_int ) );

    // And seek back to EOF.
    if( INVALID_SET_FILE_POINTER == SetFilePointer( pvOutputFile, 0, NULL, FILE_END) )
    {
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. (botanical object '%s'. Reason: %s\n", szObjectName, PS2Strip::GetLastSystemError() );
		return false;
    }

	s_uNumBytesWrittenBotanicalObject = iTotalBytesWritten;

	PS2Strip::LogInfo( "botancial object done.\n" );

	return true;

#undef CHECK
}

// JAT 25/04/06 - based on OutputBotanicalObject()
// But doesn't need stripped data instead get the billboard data direct from the polygon mesh'
struct Tools_PS2_Billboard_Data
{
	Asura_Vector_3 xCentre;
	float fWidth;
	float fHeight;
	float afUVs[4][2];
};

bool PS2Strip_ExporterImplementation::OutputBotanicalObjectBillboards( void )
{
    bool bProcessedOK = true;

    #define PROCESSING_FAILED() \
    { \
        bProcessedOK = false; \
        goto PROCESSED_BILLBOARD_OBJECT; \
    }
    #define CHECK( x ) \
    { \
        if( !(x) ) { PROCESSING_FAILED(); } \
    }

	// First thing is to do some error checking on the mesh
	Asura_Mesh* const pxElementMesh = s_pxCurMesh;
    Asura_Assert( pxElementMesh, "" );
	Asura_Assert( true == pxElementMesh->VertexDataValid, "Vertex Data not Valid");
	Asura_Assert( true == pxElementMesh->TextureDataValid, "Texture Data not Valid");
	Asura_Assert( NULL != pxElementMesh->PolygonDataPtr, "Polygon Data Pointer is NULL");
	Asura_Assert( NULL != pxElementMesh->VertexDataPtr, "Position Vertex Data Pointer is NULL");
	Asura_Assert( NULL != pxElementMesh->TextureDataPtr, "Texture Vertex Data Pointer is NULL");

	const char* szObjectName = PS2Strip::GetCurAssetName();
	const u_int uVertexCount = pxElementMesh->NumberOfVertices;
	if ( uVertexCount < 4 )
	{
		PS2Strip::LogError( "\nNot enough vertices in the mesh %d '%s'", uVertexCount, szObjectName );
		return false;
	}

	// Must be a multiple of 4 vertices
	if ( uVertexCount % 4 != 0 )
	{
		PS2Strip::LogError( "\nWrong number of vertices in the mesh %d remainder %d '%s'", uVertexCount, uVertexCount%4, szObjectName );
		return false;
	}

	Tools_DynamicArray<Tools_PS2_Billboard_Data> xBillboards( uVertexCount / 4 );

	// Generate any required vertex data
    CHECK( GenerateRequiredVertexData() );

	int iMaterialID = ASURA_MATERIAL_NONE;
	u_int uNumBillboards = 0;

	bool  bDoneTexCoordError    = false;
	bool  bDoneMultipleMatError = false;
	bool  bDoneUntexturedError  = false;
	int iObjectMaterialID = INT_MIN;

	const u_int uNumPolygons = pxElementMesh->NumberOfPolygons;
	for( u_int uPolyIndex = 0; uPolyIndex < uNumPolygons; uPolyIndex++ )
	{
		const Asura_Mesh_Polygon& xPoly = pxElementMesh->PolygonDataPtr[ uPolyIndex ];
		const Asura_Mesh_TextureInfo* pxTextureInfo = pxElementMesh->GetTextureInfoForExport(uPolyIndex);

		const bool bIsQuad = xPoly.IsQuad();
		// Ignore non quads
		if( bIsQuad == false )
		{
			continue;
		}
		iMaterialID = pxTextureInfo->m_iMaterialID;
		if ( iObjectMaterialID == INT_MIN )
		{
			iObjectMaterialID = iMaterialID;
		}
		else if( !bDoneMultipleMatError &&  ( iObjectMaterialID != iMaterialID ) )
		{
			PS2Strip::LogError( "ERROR: Botanical Billboard Object '%s' uses multiple materials and/or textures.", szObjectName );
			bDoneMultipleMatError = true;
		}
		if( !bDoneUntexturedError )
		{
			if ( ( iMaterialID == ASURA_MATERIAL_NONE ) || ASURA_TEXTURE_NONE == Asura_Material_System::GetTextureID( iMaterialID ) )
			{
				PS2Strip::LogError( "ERROR: Botanical Billboard Object '%s' has untexture polys.", szObjectName );
				bDoneUntexturedError = true;
			}
		}

		const Asura_Vector_3 xDirectionWorldWidth( 1.0f, 0.0f, 0.0f);
		const Asura_Vector_3 xDirectionWorldHeight( 0.0f, 1.0f, 0.0f);

		float afTexCoords[4][2] = { 0 };
		Asura_Vector_3 axPositions[4];
		Asura_Vector_3 xPositionCentre;
		xPositionCentre.Zero();

		for( u_int uPolygonVertex = 0; uPolygonVertex < 4; uPolygonVertex++ )
		{
			uint32 uVertexIndex = xPoly.VertexID[ uPolygonVertex ];
			Asura_Assert( uVertexIndex < uVertexCount, "Element Vertex Index Out of Range" );
			const Asura_Vector_3& xVertexPosition = pxElementMesh->VertexDataPtr[uVertexIndex].Position;
			axPositions[uPolygonVertex] = xVertexPosition;

			afTexCoords[uPolygonVertex][0] = pxTextureInfo->U[0][uPolygonVertex];
			afTexCoords[uPolygonVertex][1] = pxTextureInfo->V[0][uPolygonVertex];

			xPositionCentre += xVertexPosition;
		}

		xPositionCentre /= 4.0f;

		const u_int uIndexBiggestInvalid = static_cast<u_int>( -1 );

		// Find the natural Directions (NDs)
		const Asura_Vector_3 axNaturalDirections[3] =
		{
			axPositions[0] - axPositions[1],
			axPositions[0] - axPositions[2],
			axPositions[0] - axPositions[3]
		};
		// Find the angles between the NDs
		const float afDotNDs[3] = 
		{
			axNaturalDirections[0] * axNaturalDirections[1],
			axNaturalDirections[0] * axNaturalDirections[2],
			axNaturalDirections[2] * axNaturalDirections[1]
		};

		const float afAbsDotNDs[ 3] = 
		{
			Asura_Maths::Abs( afDotNDs[0] ),
			Asura_Maths::Abs( afDotNDs[1] ),
			Asura_Maths::Abs( afDotNDs[2] )
		};

		// Find the outer two NDs
		u_int uUseNDIndex0 = uIndexBiggestInvalid;
		u_int uUseNDIndex1 = uIndexBiggestInvalid;

		if( afAbsDotNDs[0] < afAbsDotNDs[1])
		{
			if( afAbsDotNDs[0] < afAbsDotNDs[2])
			{
				uUseNDIndex0 = 0;
				uUseNDIndex1 = 1;
			}
			else
			{
				if( afAbsDotNDs[1] < afAbsDotNDs[2])
				{
					uUseNDIndex0 = 0;
					uUseNDIndex1 = 2;
				}
				else
				{
					uUseNDIndex0 = 2;
					uUseNDIndex1 = 1;
				}
			}
		}
		else
		{
			if( afAbsDotNDs[1] < afAbsDotNDs[2])
			{
				uUseNDIndex0 = 0;
				uUseNDIndex1 = 2;
			}
			else
			{
				if( afAbsDotNDs[0] < afAbsDotNDs[2])
				{
					uUseNDIndex0 = 0;
					uUseNDIndex1 = 1;
				}
				else
				{
					uUseNDIndex0 = 2;
					uUseNDIndex1 = 1;
				}
			}
		}
		Asura_Assert( uIndexBiggestInvalid != uUseNDIndex0, "Outter NDs not found!" );
		Asura_Assert( uIndexBiggestInvalid != uUseNDIndex1, "Outter NDs not found!" );

		// Find one that is nice and close to the Height Direction
		const float afDotUp[ 2] = 
		{
			xDirectionWorldHeight * axNaturalDirections[ uUseNDIndex0],
				xDirectionWorldHeight * axNaturalDirections[ uUseNDIndex1],
		};
		const float afAbsDotUp[ 2] = 
		{
			Asura_Maths::Abs( afDotUp[0]),
				Asura_Maths::Abs( afDotUp[1]),
		};

		uint32 uIndexBiggestDotUp = uIndexBiggestInvalid;
		uint32 uIndexBiggestDotSide = uIndexBiggestInvalid;

		if( afAbsDotUp[ 0] < afAbsDotUp[ 1])
		{
			uIndexBiggestDotUp = uUseNDIndex1;
			uIndexBiggestDotSide = uUseNDIndex0;
		}
		else
		{
			uIndexBiggestDotUp = uUseNDIndex0;
			uIndexBiggestDotSide = uUseNDIndex1;
		}

		Asura_Assert( uIndexBiggestInvalid != uIndexBiggestDotUp, "Biggest Up Vector not found!");
		Asura_Assert( uIndexBiggestInvalid != uIndexBiggestDotSide, "Biggest Side Vector not found!");

		// Does the Direction need Inverting
		Asura_Vector_3 xDirectionHeight( axNaturalDirections[uIndexBiggestDotUp] );
		xDirectionHeight.Normalise();

		// Does the Direction need Inverting
		Asura_Vector_3 xDirectionWidth( axNaturalDirections[uIndexBiggestDotSide] );
		xDirectionWidth.Normalise();

		const float afHeights[3] = 
		{
			axNaturalDirections[0] * xDirectionHeight,
			axNaturalDirections[1] * xDirectionHeight,
			axNaturalDirections[2] * xDirectionHeight
		};
		const float afWidths[3] = 
		{
			axNaturalDirections[0] * xDirectionWidth,
			axNaturalDirections[1] * xDirectionWidth,
			axNaturalDirections[2] * xDirectionWidth
		};

		const float fWidth = 0.5f * Asura_Maths::Max( afWidths[0], Asura_Maths::Max( afWidths[1], afWidths[2]));
		const float fHeight = 0.5f * Asura_Maths::Max( afHeights[0], Asura_Maths::Max( afHeights[1], afHeights[2]));

		if ( bDoneTexCoordError == false )
		{
			for ( u_int uI = 0; uI < 4; uI++ )
			{
				for ( u_int uJ = 0; uJ < 2; uJ++ )
				{
					const float fTexCoord = afTexCoords[uI][uJ];
					if ( ( fTexCoord <= -8.0f ) || ( fTexCoord >= 8.0f ) )
					{
						PS2Strip::LogError( "ERROR: Botanical Billboard Object \"%s\" has texture coordinate outside valid PS2 range.", szObjectName );
						bDoneTexCoordError = true;
        			}
				}
			}
		}

		// Save the billboard data
		Tools_PS2_Billboard_Data& xBillboard = xBillboards[uNumBillboards];
		xBillboard.fWidth = fWidth;
		xBillboard.fHeight = fHeight;
		xBillboard.xCentre = xPositionCentre;

		if ( ( fWidth < -10.0f ) || ( fWidth > +10.0f ) )
		{
			PS2Strip::LogError( "ERROR: Botanical Billboard Object \"%s\" too big a width", szObjectName );
			Asura_Assert( false, "" );
		}
		if ( ( fHeight < -10.0f ) || ( fHeight > +10.0f ) )
		{
			PS2Strip::LogError( "ERROR: Botanical Billboard Object \"%s\" too big a height", szObjectName );
			Asura_Assert( false, "" );
		}

		// Input quad is 
		//
		// 0---3
		// |   |
		// |   |
		// 1---2
		//
		// PS2 quad is
		//
		// 0---1
		// |   |
		// |   |
		// 2---3
		//
		xBillboard.afUVs[0][0] = afTexCoords[0][0];
		xBillboard.afUVs[0][1] = afTexCoords[0][1];
		xBillboard.afUVs[1][0] = afTexCoords[3][0];
		xBillboard.afUVs[1][1] = afTexCoords[3][1];
		xBillboard.afUVs[2][0] = afTexCoords[1][0];
		xBillboard.afUVs[2][1] = afTexCoords[1][1];
		xBillboard.afUVs[3][0] = afTexCoords[2][0];
		xBillboard.afUVs[3][1] = afTexCoords[2][1];

		uNumBillboards++;
	}
	Asura_Assert( uNumBillboards > 0, "" );
	if ( uNumBillboards == 0 )
	{
		return false;
	}

	// Output progress message.
	PS2Strip::LogInfo( "\nWriting botanical object billboard data... " );

	HANDLE pvOutputFile = PS2Strip::GetOutputHandle();
	Asura_Assert( INVALID_HANDLE_VALUE != pvOutputFile, "" ); // Shouldn't have got this far if failed to open file.

	s_uNumBytesWrittenBotanicalObject = 0;

	int   iTotalBytesWritten    = 0;   

	// Object name.
	const int iBytesWrittenBeforeObjName = iTotalBytesWritten;
	const u_int uObjectNameSize = strlen(szObjectName) + 1;
	WRITE_DATA( pvOutputFile, iTotalBytesWritten, szObjectName, uObjectNameSize );
	PAD_TO_ALIGNMENT( pvOutputFile, 4, iTotalBytesWritten, iBytesWrittenBeforeObjName );

	// Shape hdr.
	Asura_PS2_Shape_ShapeHdr xShapeHdr;

	const u_int uMaxNumBillboardsPerStrip = ( ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP / 4 );
	const u_int uNumTriangles = uNumBillboards * 2;
	const u_int uNumStrips = ( uNumBillboards + uMaxNumBillboardsPerStrip - 1 ) / uMaxNumBillboardsPerStrip;
	xShapeHdr.usTotalTris   = static_cast<u_short>( uNumTriangles );
	xShapeHdr.usNumStrips   = static_cast<u_short>( uNumStrips );
	xShapeHdr.fLODThreshold = s_pxCurMesh->m_fMinDistance;

	WRITE_DATA( pvOutputFile, iTotalBytesWritten, &xShapeHdr, sizeof(Asura_PS2_Shape_ShapeHdr) );

	// u_int uNumBytes;
	u_int uNumBytes = 0;
	const u_int uNumBytesFilePtr = SetFilePointer( pvOutputFile, 0, NULL, FILE_CURRENT );
	WRITE_DATA( pvOutputFile, iTotalBytesWritten, &uNumBytes, sizeof(u_int) );

	uNumBytes = iTotalBytesWritten;

	// The DMA chain.
	if( uNumBillboards )
	{
		Tools_PS2_DMAChainWriter xChainWriter;

		CHECK( xChainWriter.Init( pvOutputFile ) );

		CHECK( xChainWriter.BeginChain() );

		u_int uBillboard = 0;
		u_int uNumBillboardsLeft = uNumBillboards;
		for( u_int uStrip = 0; uStrip < uNumStrips; uStrip++ )
		{
			// For the first two strips we need to fill alpha values. We
			// don't actually support any vertex alpha on objects right now
			// but the render path is the same. However, once we've ensured
			// both buffers are full, we can ignore it.

			// Output data per billboard (4 vertices)
			// Pos[0] = Billboard Position: x, y, z
			// UV[0]
			// rgba[0] - no data sent
			// normals[0] - no data sent
			// Pos[1] = Billboard width, height
			// UV[1]
			// rgba[1] - no data sent
			// normals[1] - no data sent
			// Pos[2] - no data sent
			// UV[2]
			// rgba[2] - no data sent
			// normals[2] - no data sent
			// Pos[3] - no data sent
			// UV[3]
			// rgba[3] - no data sent
			// normals[3] - no data sent

			if ( uStrip < 2 )
			{
				// Temporarily change CYCLE.
				CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_STCYCL( 4, 0, false ) ) );

				// On the first strip setup STMASK for the alpha filling.
				if( !uStrip )
				{
					// Just write to w of colour QW...
					CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_STMASK( false ) ) );
					const PS2Strip_VIF1_Mask xMask =
					{
						3, 3, 3, 3,
						3, 3, 3, 3,
						3, 3, 3, 1,
						3, 3, 3, 3,
					};
					CHECK( xChainWriter.AddDataWord( reinterpret_cast<const u_int&>( xMask ) ) );

					// ... and fill it with 1.
					CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_STROW( false ) ) );
					const float afFillRow[4] = { 0.0f, 0.0f, 0.0f, 1.0f };
					CHECK( xChainWriter.AddData( afFillRow, 16 ) );
				}

				// NB: The CMD here is irrelevant since we're
				// just using fill values and not sending any
				// data.
				CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_UNPACKR( 0, 
												ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP * 4,
												PS2STRIP_VIF1_CMD_V4_32 | PS2STRIP_VIF1_MASK_PROCESS_ENABLE,
												false) ) );
			}

			u_int uNumBillboardsInStrip = uNumBillboardsLeft;
			if ( uNumBillboardsInStrip > uMaxNumBillboardsPerStrip ) uNumBillboardsInStrip = uMaxNumBillboardsPerStrip;
			uNumBillboardsLeft -= uNumBillboardsInStrip;

			// Billboard centre positions
			// Stride of 16 because 1 position for every 4 vertices (1 per billboard)
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_STCYCL( 1, 16, false ) ) );
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_UNPACKR( 0, uNumBillboardsInStrip, PS2STRIP_VIF1_CMD_V3_32, false ) ) );
			for( u_int uBB = 0; uBB < uNumBillboardsInStrip; uBB++ )
			{
				const u_int uBBIdx = uBillboard + uBB;
				CHECK( xChainWriter.AddData( &( xBillboards[uBBIdx].xCentre ), 12 ) );
			}
			// Billboard width & heights
			// Uses a stride of 16 because 1 width,height for every 4 vertices (1 per billboard)
			// Start at address 4 because that is VertexPos[1]
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_UNPACKR( 4, uNumBillboardsInStrip, PS2STRIP_VIF1_CMD_V2_32, false ) ) );
			for( u_int uBB = 0; uBB < uNumBillboardsInStrip; uBB++ )
			{
				const u_int uBBIdx = uBillboard + uBB;
				const float fWidth = xBillboards[uBBIdx].fWidth;
				const float fHeight = xBillboards[uBBIdx].fHeight;
				CHECK( xChainWriter.AddData( &fWidth, 4 ) );
				CHECK( xChainWriter.AddData( &fHeight, 4 ) );

				if ( ( fWidth < -10.0f ) || ( fWidth > +10.0f ) )
				{
					PS2Strip::LogError( "ERROR: Botanical Billboard Object \"%s\" too big a width", szObjectName );
					Asura_Assert( false, "" );
				}
				if ( ( fHeight < -10.0f ) || ( fHeight > +10.0f ) )
				{
					PS2Strip::LogError( "ERROR: Botanical Billboard Object \"%s\" too big a height", szObjectName );
					Asura_Assert( false, "" );
				}
			}

			// Stride of 4 because 1 ST per vertex (4 per billboard)
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_STCYCL( 1, 4, false ) ) );

			const u_int uNumVertices = uNumBillboardsInStrip * 4;
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_UNPACKR( 1, uNumVertices, PS2STRIP_VIF1_CMD_V2_16, false ) ) );

			short asTexCoords[ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP * 2];
			Asura_Assert( uNumVertices <= ASURA_PS2_BOTANICALS__MAX_VERTS_PER_STRIP, "" );

			u_int uVertex = 0;
			for( u_int uBB = 0; uBB < uNumBillboardsInStrip; uBB++ )
			{
				const u_int uBBIdx = uBillboard + uBB;
				for ( u_int uI = 0; uI < 4; uI++ )
				{
					const float fU = xBillboards[uBBIdx].afUVs[uI][0];
					const float fV = xBillboards[uBBIdx].afUVs[uI][1];
					asTexCoords[( uVertex << 1 ) + 0] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, fU );
					asTexCoords[( uVertex << 1 ) + 1] = Tools_PS2Strip_ConvertFloatTo16BitFixed( 12, fV );
					uVertex++;
				}
			}
			Asura_Assert( uVertex == uNumVertices, "" );
			CHECK( xChainWriter.AddData( asTexCoords, uNumVertices * ( 2 + 2 ) ) );

			// Pass billboard count in ITOPS.
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_ITOP( uNumBillboardsInStrip, false ) ) );

			// And kick off the microprogram to render this strip.
			CHECK( xChainWriter.AddVifCode( PS2STRIP_VIF1_SET_MSCNT( false ) ) );

			uBillboard += uNumBillboardsInStrip;
		}
		// End loop through strips.
		CHECK( xChainWriter.EndChain() );        

		Asura_Assert( uNumBillboardsLeft == 0, "" );
		Asura_Assert( uBillboard == uNumBillboards, "" );

		// Update cumulative byte count.
		const u_int uBytesWrittenForChain = xChainWriter.GetTotalBytesWrittenForLastChain();
		Asura_Assert( !( uBytesWrittenForChain & 15 ), "" );
		iTotalBytesWritten += uBytesWrittenForChain;
	}
	// End DMA chain write.
	uNumBytes = iTotalBytesWritten - uNumBytes;
	// Go back and fill in the size of the dma data
	if ( INVALID_SET_FILE_POINTER == SetFilePointer( pvOutputFile, uNumBytesFilePtr, NULL, FILE_BEGIN ) )
	{
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr (botanical object '%s'). Reason: %s\n", szObjectName, PS2Strip::GetLastSystemError() );
		return false;
	}
	int iIgnore = 0;
	WRITE_DATA( pvOutputFile, iIgnore, &uNumBytes, sizeof( u_int ) );

	// And seek back to EOF.
	if( INVALID_SET_FILE_POINTER == SetFilePointer( pvOutputFile, 0, NULL, FILE_END) )
	{
		PS2Strip::LogError( "ERROR: Couldn't adjust file ptr. (botanical object '%s'. Reason: %s\n", szObjectName, PS2Strip::GetLastSystemError() );
		return false;
	}

	s_uNumBytesWrittenBotanicalObject = iTotalBytesWritten;

	PS2Strip::LogInfo( "botancial object billboard done.\n" );

PROCESSED_BILLBOARD_OBJECT:
	PS2Strip_ExporterImplementation::CleanUp( true );

	return true;

#undef CHECK
}

#endif //#if defined( ASURA_USE_BOTANICALS )

