//
// Tools_Mode_Texturing.cpp
//
// Created by Kevin Lea - 20/06/2001 14:52:09
// 
// Copyright (C) 2001 Rebellion, All Rights Reserved.
//

#include "Core/Asura.h"

#include "Texturing/Tools_Mode_Texturing.h"

#include "Animation/Asura_TextureAnimation.h"
#include "Core/Asura_Material_System.h"
#include "Core/Asura_TextureManagement.h"
#include "Core/Asura_Timers.h"
#include "Core/Asura_UserInput.h"
#include "Debugging/Asura_Assert.h"
#include "Interface/Asura_Cursor.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Interface.h"
#include "Interface/Asura_Menubar.h"
#include "Interface/Asura_Menubar_System.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Module.h"
#include "Rendering/Asura_Render.h"
#include "Rendering/Asura_Text_Output.h"
#include "Rendering/Asura_Viewport.h"

#include "Tools_Overlays.h"
#include "Tools_System.h"
#include "Lighting/Tools_Photon_Object.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Materials/Tools_MaterialResponse_System.h"
#include "Materials/Tools_Overlay_PickMaterialResponse.h"
#include "Materials/Tools_Overlay_ViewMaterialResponse.h"
#include "MeshEdit/Tools_Mode_MeshEdit.h"
#include "Misc/Tools_Help.h"
#include "Misc/Tools_LOS.h"
#include "Misc/Tools_Render.h"
#include "Misc/Tools_Timed_Messages.h"
#include "Movement/Tools_Movement.h"
#include "Objects/Tools_Object_System.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Texturing/Tools_PickTexture.h"
#include "Texturing/Tools_Multiple_Node.h"
#include "Texturing/Tools_Texturing_Copy.h"
#include "Texturing/Tools_Texturing_Find.h"
#include "Texturing/Tools_Texturing_MaterialResponse.h"
#include "Texturing/Tools_Texturing_Texture.h"
#include "Texturing/Tools_Texture_Management.h"
#include "Texturing/Tools_TextureAnimation.h"
#include "Texturing/Tools_UVEdit.h"


#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM && defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	#include "Animation/Asura_Hierarchy.h"
#endif

#define ABS(x)	(((x) > 0) ? (x) : -(x))
//////////////////////////////////////////////////////////////////////////////////////////////
/*--------------**
** Help Strings **
**--------------*/
static const char* TexturingSelectionHelpStrings [] =
{
	"R :                Rotate UV coordinates 90 degrees",
	"X :                X-Flip UV coordinates",
	"Y :                Y-Flip UV coordinates",
	"O :                Cycle nodes",
	"V :                Reverse nodes",
	"",
	"T :                Texture wrap",
	"Shift+T :          Texture tile",
	"B :                Box texture wrap",
	"Shift+B :          Tiled box texture wrap",
	"Ctrl+T :           Stop texture wrap/tile mode",
	"",
	"C :                Choose texture (and apply to selection)",
	"Shift + M :        Edits the material currently used by a polygon",
	"U :                Copy texturing from a single polygon",
	"I :                Align texturing on 2 adjacent polygons",
	"M :                Toggle 'All nodes shown'",
	"Tab :              Switch to node editing",
	"Ctrl+Q :           Toggle new UV editor",
	"F :                Edit texture flags",
	"",
	"P :                Toggle alpha painting (while overlay is open)",
	"",
	"Ctrl+Z :           Undo texturing changes",
	"Ctrl+Y :           Redo texturing changes",
	"< :                Undo texturing changes",
	"> :                Redo texturing changes",
	NULL
};

static const char* TexturingNodeEditHelpStrings [] =
{
	"LMB edits nodes according to current left button mode",
	"In Select/Move :",
	"Left click and drag on a node to move it",
	"Hold shift to snap the node to the cursor",
	"Left click and drag on a line to move the nodes at either end",
	"Left click and drag within the nodes to move all nodes",
	"Right click and drag to rotate nodes",
	"Right click and drag (with Control) to scale nodes",
	"Shift              Hold to restrict movement to one axis",
	"Control            Hold to snap to nearest texel while moving nodes"
	"",
	"N                  Snap cursor to nearest node",
	"",
	"Hold Z             Left button zooms",
	"Hold Space         Left button edits the shown area",
	"",
	"R :                Rotate UV coordinates 90 degrees",
	"X :                X-Flip UV coordinates",
	"Y :                Y-Flip UV coordinates",
	"O :                Cycle nodes",
	"V :                Reverse nodes",
	"",
	"C :                Choose texture",
	"Tab :              Switch to polygon selection",
	"",
	"Ctrl+Z :           Undo texturing changes",
	"Ctrl+Y :           Redo texturing changes",
	"< :                Undo texturing changes",
	"> :                Redo texturing changes",
	"",
	"While tiling a texture :",
	"4,6                Alter U Offset",
	"2,8                Alter V Offset",
	"7,9                Alter U Scale",
	"1,3                Alter V Scale",
	"Hold 0             Alter offsets and scales faster",
	NULL
};


//////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------**
** Tools_Texturing_UndoData_TextureInfo **
**--------------------------------------*/


Tools_Texturing_UndoData_TextureInfo::Tools_Texturing_UndoData_TextureInfo(const PolygonID& xPoly)
{
	m_xPoly = xPoly;

	Asura_Assert(xPoly.ModuleIndex < Asura_Module_System::GetNumberOfModules(),"");

	const Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);
	Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
	Asura_Assert(xPoly.PolyIndex < pxMesh->NumberOfPolygons,"");

	//store a copy of the current texturing
	m_xTextureInfo = pxMesh->TextureDataPtr[xPoly.PolyIndex];
}

void Tools_Texturing_UndoData_TextureInfo::Create(const PolygonID& xPoly)
{
	RootUndoData** const ppxUndoData = Tools_Undo_System::AllocateNextUndoData();
	if( !ppxUndoData ) return;

	*ppxUndoData = new Tools_Texturing_UndoData_TextureInfo( xPoly );
}

void Tools_Texturing_UndoData_TextureInfo::Create(const PolygonID& xPoly,VariableSizeUndo* pxUndoSystem)
{
	RootUndoData** const ppxUndoData = pxUndoSystem->AllocateNextUndoData();
	if( !ppxUndoData ) return;

	*ppxUndoData = new Tools_Texturing_UndoData_TextureInfo( xPoly );
}


RootUndoData* Tools_Texturing_UndoData_TextureInfo::ProcessUndoData()
{

	//apply undo data
	Asura_Assert(m_xPoly.ModuleIndex < Asura_Module_System::GetNumberOfModules(),"");

	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(m_xPoly.ModuleIndex);
	Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
	Asura_Assert(m_xPoly.PolyIndex < pxMesh->NumberOfPolygons,"");

	//Swap the texturing between the undo data and the mesh
	Asura_Mesh_TextureInfo xTemp = pxMesh->TextureDataPtr[m_xPoly.PolyIndex];
	pxMesh->TextureDataPtr[m_xPoly.PolyIndex] = m_xTextureInfo;
	m_xTextureInfo = xTemp;


	//make sure the mesh gets updated in the hardware
	Tools_UpdateTexturingOnSurfaces(pxMesh);
	Tools_SaveAsura::SetMeshModified(pxMesh);


	return this;
}

//////////////////////////////////////////////////////////////////////////////////////////////
/*----------------------**
** Tools_Mode_Texturing **
**----------------------*/
// Kevin Lea - 22/06/2001 16:09:36 - declaration of static data

Asura_Handle					Tools_Mode_Texturing::MenuHandle;
Asura_Vector_2					Tools_Mode_Texturing::OptionsOverlayPosition = Asura_Vector_2(3.0f,28.0f);
bool							Tools_Mode_Texturing::OptionsOverlayOpen = true;
u_int							Tools_Mode_Texturing::s_uOptionsOverlayMinimised = 0;
Asura_Vector_2					Tools_Mode_Texturing::TileTextureOverlayPosition = Asura_Vector_2(23.0f,10.0f);
Asura_Vector_2					Tools_Mode_Texturing::CopyOverlayPosition = Asura_Vector_2(23.0f,10.0f);
bool							Tools_Mode_Texturing::CopyOverlayOpen = false;
u_int							Tools_Mode_Texturing::s_uCopyOverlayMinimised = 0;
Asura_Vector_2					Tools_Mode_Texturing::NodeEditOverlayPosition = Asura_Vector_2(3.0f,3.0f);
u_int							Tools_Mode_Texturing::s_uNodeEditOverlayMinimised = 0;
bool							Tools_Mode_Texturing::StandardInputDisabledForModalOverlay;
Asura_Vector_2					Tools_Mode_Texturing::s_xAlphaOverlayPosition = Asura_Vector_2(23.0f,23.0f);
bool							Tools_Mode_Texturing::s_bAlphaOverlayOpen = false;
u_int							Tools_Mode_Texturing::s_uAlphaOverlayMinimised = 0;

Asura_Vector_2					Tools_Mode_Texturing::s_xTextureRepeatsOverlayPosition = Asura_Vector_2(17.0f,10.0f);
bool							Tools_Mode_Texturing::s_bTextureRepeatsOverlayOpen = false;
u_int							Tools_Mode_Texturing::s_uTextureRepeatsOverlayMinimised = 0;

enum TexturingModeEnum 			Tools_Mode_Texturing::TexturingMode;
bool							Tools_Mode_Texturing::MultipleNodeMode;

enum PromptedSelectionModeEnum  Tools_Mode_Texturing::PromptedSelectionMode;

Tools_UV_Editor*				Tools_Mode_Texturing::s_pxTextureEdit;
Asura_Vector_3					Tools_Mode_Texturing::TextureWrap_X;
Asura_Vector_3					Tools_Mode_Texturing::TextureWrap_Y;
Asura_Vector_3					Tools_Mode_Texturing::TextureWrap_Z;
enum TextureWrapMode			Tools_Mode_Texturing::WrapMode;

PolygonID						Tools_Mode_Texturing::BasePolygon;
PolygonID						Tools_Mode_Texturing::SelectedPolygon1;
PolygonID						Tools_Mode_Texturing::SelectedPolygon2;

Asura_Vector_2					Tools_Mode_Texturing::TextureTile_Scale;
Asura_Vector_2					Tools_Mode_Texturing::TextureTile_Offset;
float							Tools_Mode_Texturing::TextureTile_OffsetScale = 512.0f;
bool							Tools_Mode_Texturing::TextureTile_Changed = false;

Asura_Vector_2					Tools_Mode_Texturing::TextureRepeats;
Asura_Vector_2					Tools_Mode_Texturing::TextureRepeatsOrigin;
bool							Tools_Mode_Texturing::TextureRepeatsChanged = false;
int								Tools_Mode_Texturing::s_iTextureRepeat_NodeAdjustment = Tools_Mode_Texturing::TEXTUREREPEAT_NODESTOCORNER;

bool							Tools_Mode_Texturing::s_bShowTextureUnderCursor;

u_int							Tools_Mode_Texturing::s_uEditedChannels = 0xffffffff;
float							Tools_Mode_Texturing::s_fSecondaryUVScale = 1.0f;
bool							Tools_Mode_Texturing::s_bLimitUVEditingToSelectedTexture = false;
int								Tools_Mode_Texturing::s_iTextureForUVEditing = ASURA_TEXTURE_NONE;

bool							Tools_Mode_Texturing::s_bAlphaPainting = false;
bool							Tools_Mode_Texturing::s_bAlphaPaint_AffectNearestVert;
float							Tools_Mode_Texturing::s_fAlphaPaint_Radius = 5.0f;
float							Tools_Mode_Texturing::s_fAlphaPaint_Alpha[2] = {0.0f,1.0f};
bool							Tools_Mode_Texturing::s_bAlphaPaint_UseHoming = false;
float							Tools_Mode_Texturing::s_fAlphaPaint_HomingRate = 1.0f;
bool							Tools_Mode_Texturing::s_bAlphaPaint_SelectedOnly = true;
bool							Tools_Mode_Texturing::s_bAlphaPaint_ShowSphere = true;
bool							Tools_Mode_Texturing::s_bAlphaPaint_ButtonClickedForPaint = false;
bool							Tools_Mode_Texturing::s_bAlphaPaint_SetUVs = false;

bool							Tools_Mode_Texturing::s_bAlphaPaint_ApplyTexture = false;
bool							Tools_Mode_Texturing::s_bAlphaPaint_AllowTextureOverwrite = false;
#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
int								Tools_Mode_Texturing::s_iAlphaPaint_TextureSlot = 0;
int								Tools_Mode_Texturing::s_iAlphaPaint_TextureSlot_ListOffset = 0;
int								Tools_Mode_Texturing::s_iAlphaPaint_TextureSlot_ListSelection = 0;
int								Tools_Mode_Texturing::s_aiSecondaryTextures[3][TOOLS_MAX_SECONDARY_TEXTURES+1];
char*							Tools_Mode_Texturing::s_apszSecondaryTextureNames[3][TOOLS_MAX_SECONDARY_TEXTURES+1];
#endif

bool							Tools_Mode_Texturing::s_bShowCoordsInTexels = false;

int 							Tools_Mode_Texturing::s_iSelectedMaterialID = ASURA_MATERIAL_NONE;
#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
Asura_Hash_ID		 			Tools_Mode_Texturing::s_uSelectedMaterialResponse;
Asura_Hash_ID		 			Tools_Mode_Texturing::s_uSelectedMaterialResponse2;
#endif

bool							Tools_Mode_Texturing::s_bApplyMaterialOnTextureWrap = true;

Asura_Angle						Tools_Mode_Texturing::s_fRotationAngle = Asura_Maths::Deg90;

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
bool							Tools_Mode_Texturing::s_bMaterialResponseOverlayWasOpenBeforeMode = false;
#endif

Tools_UVEdit*					Tools_Mode_Texturing::s_pxUVEdit;



void Tools_Mode_Texturing::Initialise(void)
{

	//initialise multiple copy
	Tools_Texturing_Copy::Initialise();
	Texturing_Multiple_Nodes::Initialise();

	// Kevin Lea - 25/06/2001 15:15:11 - basic initialisation
	
	//load values from .dat file
	TransferModeData(TOOLSMODEDATAFILE_LOAD);
	
	s_pxTextureEdit = new Tools_UV_Editor;
	s_pxUVEdit = new Tools_UVEdit;


	//set lots of defaults for things
	TexturingMode = TexturingMode_Selection;
	MultipleNodeMode = false;
	PromptedSelectionMode = PSM_None;
	BasePolygon.ModuleIndex = -1;
	BasePolygon.PolyIndex = -1;
	WrapMode = TextureWrapMode_None;
	TextureTile_Scale.x = 1.0f;
	TextureTile_Scale.y = 1.0f;
	TextureTile_Offset.x = 0.0f;
	TextureTile_Offset.y = 0.0f;
	TextureTile_Changed = false;
	TextureRepeats.x = 1.0f;
	TextureRepeats.y = 1.0f;
	TextureRepeatsOrigin.x = 0.0f;
	TextureRepeatsOrigin.y = 0.0f;
	TextureRepeatsChanged = false;

	StandardInputDisabledForModalOverlay = false;

	Tools_PickTexture::RebuildMaterialList();
	//check for dodgy secondary texture settings
	FixSecondaryTextureAnomalies();
	
	#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	//get list of used secondary textures
	InitialiseSecondaryTextureArray();
	BuildSecondaryTextureNames();
	#endif

	
	//create the menu
	MenuHandle = ASURA_INVALID_HANDLE;
	SetMenubarForMode();

	// Kevin Lea - 25/06/2001 15:14:57 - create required overlays

	#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	s_bMaterialResponseOverlayWasOpenBeforeMode = Tools_Texturing_MaterialResponse::OverlayIsOn();
	if(!Tools_Texturing_MaterialResponse::OverlayIsOn()) Tools_Texturing_MaterialResponse::ToggleOverlay();
	#endif

	CreateOptionsOverlay();
	#ifndef ASURA_DISABLE_VERTEX_ALPHA
	CreateAlphaOverlay();
	#endif
	Tools_TextureAnimation::Initialise();

	if(s_bTextureRepeatsOverlayOpen) CreateTextureRepeatsOverlay();

	

	#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		//set the current texture to the first environment texture
		//(do before undo gets initialised , so it can't be undone)
		NextMaterial();
	#endif

	//Initalise polygon selection
	Tools_Polygon_Selection::Activate();
	Tools_Polygon_Selection::SetSelectionClearedCallback(SelectionIsNowEmpty);

	s_pxTextureEdit->NodeUndo.Initialise();

	//add help
	Tools_Help::AddHelpTopic("Texturing - Selection Mode",TexturingSelectionHelpStrings);
	Tools_Help::AddHelpTopic("Texturing - Node Editing",TexturingNodeEditHelpStrings);


}

void Tools_Mode_Texturing::Shutdown(void)
{
	if(TexturingMode == TexturingMode_NodeEdit)
	{
		//switch out of node edit mode , so things get cleared up properly
		SwitchToSelectionMode();
	}

	Tools_Help::RemoveHelpTopic("Texturing - Selection Mode");
	Tools_Help::RemoveHelpTopic("Texturing - Node Editing");

	//Tools_Undo_System::ClearUndo();
	s_pxTextureEdit->NodeUndo.Shutdown();

	Tools_Texturing_Copy::Shutdown();
	Texturing_Multiple_Nodes::Shutdown();

	Tools_Polygon_Selection::Deactivate();

	//save values to .dat file
	TransferModeData(TOOLSMODEDATAFILE_SAVE);

	// Kevin Lea - 25/06/2001 15:17:58 - release menu and overlays
	if(MenuHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Menubar_System::DestroyMenuSection(MenuHandle);
		MenuHandle = ASURA_INVALID_HANDLE;
	}
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_OPTIONS);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_TILING);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_COPY);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_ALPHA);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_NODEEDIT);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_MOUSEMODE);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_MODAL);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_REPEATS);
	Tools_TextureAnimation::Shutdown();

	#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	//Close the material response overlay if it was opened as a result of entering the mode
	if(Tools_Texturing_MaterialResponse::OverlayIsOn() && !s_bMaterialResponseOverlayWasOpenBeforeMode)
	{
		Tools_Texturing_MaterialResponse::ToggleOverlay();
	}
	#endif


	#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	DeleteSecondaryTextureNames();
	#endif

	delete s_pxTextureEdit;
	s_pxTextureEdit = 0;

	delete s_pxUVEdit;
	s_pxUVEdit = 0;

	//clear up any temporary lights
	Tools_Object_System::DeleteAllModeSpecific();
	Tools_Object_System::EnableAllControls(false);


}

void Tools_Mode_Texturing::TransferModeData(int mode)
{
	//open the file
	Tools_ModeDataFile file("TextureMode.dat",mode,1);

	bool bJunk = false;
	//tranfer the data
	//(none of these should be removed , and new data should go on the end, otherwise
	//the files will get screwed up)
	file.Transfer(OptionsOverlayPosition);
	file.Transfer(OptionsOverlayOpen);
	file.Transfer(bJunk);

	file.Transfer(TileTextureOverlayPosition);

	file.Transfer(CopyOverlayPosition);
	file.Transfer(CopyOverlayOpen);
	file.Transfer(bJunk);

	file.Transfer(NodeEditOverlayPosition);
	file.Transfer(bJunk);

	file.Transfer(Tools_UV_Editor::MouseModeOverlayPosition);
	file.Transfer(bJunk);

	file.Transfer(s_xAlphaOverlayPosition);
	file.Transfer(s_bAlphaOverlayOpen);
	file.Transfer(bJunk);

	file.Transfer(s_fAlphaPaint_Radius);
	file.Transfer(s_fAlphaPaint_Alpha[0]);
	file.Transfer(s_fAlphaPaint_Alpha[1]);
	file.Transfer(s_bAlphaPaint_UseHoming);
	file.Transfer(s_fAlphaPaint_HomingRate);
	file.Transfer(s_bAlphaPaint_SelectedOnly);
	file.Transfer(s_bAlphaPaint_ShowSphere);
	file.Transfer(s_bAlphaPaint_ApplyTexture);
	file.Transfer(s_bAlphaPaint_AllowTextureOverwrite);

	//Can't be sure of exact data beyond this point in earlier versions. RAR 06/03/2007
	if(file.GetVersion() < 1) return;

	file.Transfer(s_bAlphaPaint_SetUVs);
	file.Transfer(s_fSecondaryUVScale);

	file.Transfer(Tools_TextureAnimation::s_xOverlayPosition);
	file.Transfer(Tools_TextureAnimation::s_xEditAnimOverlayPosition);

	file.Transfer(s_bAlphaPaint_AffectNearestVert);
	file.Transfer(s_bApplyMaterialOnTextureWrap);

	file.Transfer(s_xTextureRepeatsOverlayPosition);
	file.Transfer(s_bTextureRepeatsOverlayOpen);
	file.Transfer(bJunk);

	file.Transfer(s_uOptionsOverlayMinimised);
	file.Transfer(s_uCopyOverlayMinimised);
	file.Transfer(s_uNodeEditOverlayMinimised);
	file.Transfer(Tools_UV_Editor::s_uMouseModeOverlayMinimised);
	file.Transfer(s_uAlphaOverlayMinimised);
	file.Transfer(s_uTextureRepeatsOverlayMinimised);
}

void Tools_Mode_Texturing::Update()
{
	//Keep undo batch open while alpha painting
	if(s_bAlphaPaint_ButtonClickedForPaint)
	{
		Tools_Undo_System::KeepUndoBatchOpenThisFrame();
	}

	
	//if the texture repeats overlay is open , check for any changes
	s_bTextureRepeatsOverlayOpen = Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_TEXTURING_REPEATS);
	if(s_bTextureRepeatsOverlayOpen)
	{
		if(TextureRepeatsChanged)
		{
			UpdateTextureRepeats();
		}
	}
	
	s_pxTextureEdit->SetTextureChannel(GetPrimaryEditedChannel());

	if(TexturingMode == TexturingMode_Selection && !Tools_Object_System::AreControlsEnabled()) 
	{
		if (StandardInputDisabledForModalOverlay || IsAlphaPainting())
		{
			//most other input disabled when the modal overlay is up
			Tools_Polygon_Selection::Update(PolygonSelectionUpdate_NoSelection);
		}
		else if(PromptedSelectionMode == PSM_None)
		{
			Tools_Polygon_Selection::Update(PolygonSelectionUpdate_Standard);
		}
		else
		{
			Tools_Polygon_Selection::Update(PolygonSelectionUpdate_NoMouseSelection);
			UpdatePromptedSelection();
		}
	}
	else
	{
		Tools_Polygon_Selection::Update(PolygonSelectionUpdate_NoSelection);
	}

	#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	//update alpha texture selection from list values
	s_iAlphaPaint_TextureSlot = s_iAlphaPaint_TextureSlot_ListOffset + s_iAlphaPaint_TextureSlot_ListSelection;
	#endif
	
	//act upon keyboard input
	Tools_Mode_Texturing::HandleKeyboard();

	if(!Tools_Object_System::AreControlsEnabled())
	{
		if(TexturingMode != TexturingMode_Selection && (WrapMode == TextureWrapMode_Tile || WrapMode == TextureWrapMode_BoxTile))
		{
			//numberpad keys are used for adjusting the texture tile parameters in this case
			Tools_Movement::DisableNumberPadControls = true;
		}
		else
		{
			Tools_Movement::DisableNumberPadControls = false;
		}


		if(StandardInputDisabledForModalOverlay)
		{
			if (!Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_TEXTURING_MODAL))
			{
				StandardInputDisabledForModalOverlay = false;
			}
		}

		UpdateAlphaPainting();

		//update alpha overlay if open
		UpdateAlphaOverlay();
	}

	//update texture anim editing
	Tools_TextureAnimation::Update();
}

void Tools_Mode_Texturing::RenderWorld_PostToneMap(void)
{
	//Show the currently selected polygons
	if(!Tools_Object_System::AreControlsEnabled())
	{
		Tools_Polygon_Selection::Render();
		
		if(TexturingMode == TexturingMode_Selection)
		{
			//handle selection
			if(PromptedSelectionMode != PSM_None)
			{
				RenderPromptedSelection();
			}
		}

		if(IsAlphaPainting())
		{
			RenderAlphaPainting();
		}
	}

	if(s_pxUVEdit && s_pxUVEdit->IsOverlayOpen())
	{
		s_pxUVEdit->RenderWorld();
	}
}

void Tools_Mode_Texturing::RenderUI(void)
{
	/*
	Render also contains some part of the more 'update' like code , because they
	need to draw text , or whatever.
	*/

	if(!Tools_Object_System::AreControlsEnabled())
	{
		Tools_Polygon_Selection::RenderUI();
	}

	
	
	//Show the currently selected polygons
	if(!Tools_Object_System::AreControlsEnabled())
	{
		if(TexturingMode != TexturingMode_Selection)
		{
			HandleMouseForNodeEdit();
		}
	
		//if doing texture tileing , need to see if the values have changed
		if(WrapMode == TextureWrapMode_Tile || WrapMode == TextureWrapMode_BoxTile)
		{
			if(TextureTile_Changed)
			{
				TextureWrap_Apply();
			}
		}


		//display status information
		Tools_Mode_Texturing::ShowStatusText();

		if(TexturingMode == TexturingMode_NodeEdit)
		{
			//draw the texture and nodes
			s_pxTextureEdit->Render();
		}
	}
	
}

bool Tools_Mode_Texturing::IsModified()
{
	return Tools_SaveAsura::IsMainFileModified();
}

void Tools_Mode_Texturing::QuickSave()
{
}

// Kevin Lea - 22/06/2001 15:29:50
// 
// Overlays & their associated function hooks


void Tools_Mode_Texturing::ToggleOptionsOverlay(void)
{
	//make sure the overlay is allowed to be open
	if(TexturingMode == TexturingMode_NodeEdit)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Overlay cannot be opened in this mode");
		return;
	}
	OptionsOverlayOpen = !OptionsOverlayOpen;

	if (OptionsOverlayOpen)
	{
		s_uOptionsOverlayMinimised = 0;
		CreateOptionsOverlay();
	}
	else
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_OPTIONS);
	}
}
bool Tools_Mode_Texturing::OptionsOverlayIsOn(void)
{
	return OptionsOverlayOpen;
}

void Tools_Mode_Texturing::CreateOptionsOverlay(void)
{
	if(!OptionsOverlayOpen) return;

	// Kevin Lea - 25/06/2001 16:32:41 - making overlays is a bit iffy at the moment
	// it needs a system a bit more like the menubars
	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Texturing Mode",false,&OptionsOverlayPosition,&s_uOptionsOverlayMinimised);
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
			
	pxDialog = new Asura_Tickbox_Dialog("All Nodes Shown (M)",xOffset,&(Tools_Mode_Texturing::MultipleNodeMode));
	pxDialog->SetHelpText("For use in node editing mode (Tab).\n"
							"Causes one node to be drawn on the texture for each distinct vertex in the selection.\n"
							"Otherwise just 4 nodes will be shown , taken from one polygon.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Show Texture Under Cursor",xOffset,&(Tools_Mode_Texturing::s_bShowTextureUnderCursor));
	pxDialog->SetHelpText("Displays the name of the texture used by the polygon that the mouse cursor is currently over.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	Asura_Tickbox_Dialog* pxTickbox;
	pxDialog = pxTickbox = new Asura_Tickbox_Dialog("Limit UV Editing to Texture",xOffset,&s_bLimitUVEditingToSelectedTexture);
	pxTickbox->SetClickedFunction(LimitUVEditingToSelectedTextureClicked);
	pxDialog->SetHelpText("This limits UV editing operations to affecting those channels that use a specified texture.\nTurning this on will cause the Edit Texture option above to be ignored.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Apply Mat On Texture Wrap",xOffset,&(Tools_Mode_Texturing::s_bApplyMaterialOnTextureWrap));
	pxDialog->SetHelpText("If ticked then the current material will be applied to the selected polygons when using the texture wrap/tile or box wrap/tile options.\nThis only applies if editing the primary UVs");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_FlagTickbox_Dialog("Edit Primary UVs",xOffset,&s_uEditedChannels,0x1);
	pxDialog->SetHelpText("If ticked then texturing operations will affect the UV coordinates for the main texture");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_FlagTickbox_Dialog("Edit Secondary UVs",xOffset,&s_uEditedChannels,0x2);
	pxDialog->SetHelpText("If ticked then texturing operations will affect the UV coordinates for the second texture");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Numerical_Dialog("Secondary UV Scale",xOffset,&s_fSecondaryUVScale,1.0f,99.0f,1.0f,"%.0f");
	pxDialog->SetHelpText("When editing both primary and secondary UV coordinates at the same time , this scaling factor is used for calculating the secondary coordinates.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();


	pxDialog = new Asura_Function_Dialog("Apply Material",xOffset,Tools_Mode_Texturing::ApplyTexture,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Applies the properties from the current material to all selected polygons , but leaves the secondary textures alone.");
	pxDialog->SetIsEnabledFunction(IsEditingPrimaryChannel);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Get Material",xOffset,Tools_Mode_Texturing::GetMaterialOnPolygon);
	pxDialog->SetHelpText("Gets the material from a chosen polygon and sets it as the current material.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("Edit Material (Shift+M)",xOffset,Tools_Mode_Texturing::EditMaterialOnPolygon);
	pxDialog->SetHelpText("Edits the material currently applied to a chosen polygon.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Wrap Texture (T)",xOffset,Tools_Mode_Texturing::TextureWrapSpread_Start,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Spread the selected number of tiles of the current texture over the selected polygons.\nCoordinates are chosen by projecting the vertices onto a plane defined by one of the polygons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Tile Texture (Shift + T)",xOffset,Tools_Mode_Texturing::TextureWrapTile_Start,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Spread the current texture over the selected polygons using a specified number of tiles per metre.\nCoordinates are chosen by projecting the vertices onto a plane defined by one of the polygons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("X-Flip UVs (X)",xOffset,Tools_Mode_Texturing::AdjustUV_XFlip,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Mirror the 'U'coordinate of all selected polygons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Y-Flip UVs (Y)",xOffset,Tools_Mode_Texturing::AdjustUV_YFlip,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Mirror the 'V'coordinate of all selected polygons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Cycle UVs (O)",xOffset,Tools_Mode_Texturing::AdjustUV_Cycle,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Cycle the UV coordinates for the selected polygons , so that each vertex gets the coordinates from the next vertex in the same polygon.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Reverse UVs (V)",xOffset,Tools_Mode_Texturing::AdjustUV_Reverse,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Reverse the ordering of the UV coordinates in the selected polygons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Rotate UVs (R)",xOffset,Tools_Mode_Texturing::AdjustUV_Rotate,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Rotate the UV coordinates on the selected polygons by the angle specified below.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_AngleNumerical_Dialog("Angle", xOffset, &s_fRotationAngle, -360.0f,360.0f,1.0f,"%.1f");
	pxDialog->SetHelpText("Angle to use for rotation.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Shift UVs",xOffset,Tools_Mode_Texturing::AdjustUV_Shift,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Apply a specified xOffset to the UV coordinates of all selected polygons.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Scale UVs",xOffset,Tools_Mode_Texturing::AdjustUV_Scale,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Multiply the UV coordinates of all selected  polygons by a specified amount.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		pxDialog = new Asura_Function_Dialog("Choose Material (C)",xOffset,Tools_Mode_Texturing::ChooseMaterial,NULL,Tools_Object_System::AreControlsDisabled);
		pxDialog->SetHelpText("Selects a new material to use as the current material and applies its properties to all selected polygons.\nThe secondary textures will be unchanged.\n");
	#else
		pxDialog = new Asura_Function_Dialog("Choose Material (C)",xOffset,Tools_Mode_Texturing::ChooseMaterialResponse,NULL,Tools_Object_System::AreControlsDisabled);
		pxDialog->SetHelpText("Selects a new material to use as the primary material and applies it to the selected polygons.\n");
	#endif
		pxDialog->SetIsEnabledFunction(IsEditingPrimaryChannel);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("Texture Flags (F)",xOffset,Tools_Mode_Texturing::EditTextureFlags,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Brings up the texture flag editing screen.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Align Texturing (I)",xOffset,Tools_Mode_Texturing::TextureAlign_Start,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Prompts for 2 adjacent polygons to be chosen.\nThe UV coordinates will be changed to match at any shared vertices.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Sync UV Channels",xOffset,Tools_Mode_Texturing::SyncUVChannelsPriToSec,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("This sets the secondary UV coordinates on all selected polygons to be a chosen multiple of the primary UV coordinates.");
	pxOverlay->AddDialog(pxDialog);

	xOffset.x = pxDialog->m_xMaxCoord.x+ 1.0f;
	pxDialog = new Asura_Function_Dialog("Sec->Pri",xOffset,Tools_Mode_Texturing::SyncUVChannelsSecToPri,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Similar to Sync UV Channels except that the primary UVs are altered based on the secondary UVs.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.x = 1.0f;

	xOffset.y+=Asura_Dialog::LargeSpacing();


	pxDialog = new Asura_Function_Dialog("Undo (Ctrl+Z)",xOffset,Undo,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Undo changes to texturing.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Redo (Ctrl+Y)",xOffset,Redo,NULL,Tools_Object_System::AreControlsDisabled);
	pxDialog->SetHelpText("Redo changes to texturing.");
	pxOverlay->AddDialog(pxDialog);
	
	
	pxOverlay->m_uID = TOOLS_OVERLAY_TEXTURING_OPTIONS;

	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}
void Tools_Mode_Texturing::ToggleCopyOverlay(void)
{
	//make sure the overlay is allowed to be open
	if(TexturingMode == TexturingMode_NodeEdit)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Overlay cannot be opened in this mode");
		return;
	}
	CopyOverlayOpen = !CopyOverlayOpen;

	if (CopyOverlayOpen)
	{
		s_uCopyOverlayMinimised = 0;
		CreateCopyOverlay();
	}
	else
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_COPY);
	}
}
bool Tools_Mode_Texturing::CopyOverlayIsOn(void)
{
	return CopyOverlayOpen;
}

void Tools_Mode_Texturing::CreateCopyOverlay(void)
{
	if(!CopyOverlayOpen) return;

	Asura_Overlay* newOverlay;
	newOverlay = new Asura_Overlay("Texture Copying",false,&CopyOverlayPosition,&s_uCopyOverlayMinimised);
	
	Asura_Dialog* newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(1.0f,3.0f);
			
	newDialog = new Asura_Function_Dialog("Polygon (U)",offset,Tools_Mode_Texturing::TextureCopyPolygon_Start);
	newDialog->SetHelpText("Prompts for a polygon to be chosen.\nThe texturing will be copied from this polygon to all selected polygons.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Object",offset,Tools_Mode_Texturing::TextureCopyObject_Start);
	newDialog->SetHelpText("Prompts for two similar objects to be chosen.\nThe texturing will be copied from the first to the second.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("From Multiple",offset,Tools_Texturing_Copy::CopyFromMultiple_Start);
	newDialog->SetHelpText("Stores the currently selected polygons for a multiple copy operation.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("To Multiple",offset,Tools_Texturing_Copy::CopyToMultiple_Start);
	newDialog->SetHelpText("Copies the texturing from the polygons selected with 'From Multiple' to the current selection.\n"
						   "The two sets of polygons must be approximately equivalent (topologically or spacially) for this to work.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();
	
	
	newOverlay->m_uID = TOOLS_OVERLAY_TEXTURING_COPY;

	Asura_Overlay_System::SubmitNewOverlay(newOverlay,false);
}

void Tools_Mode_Texturing::ToggleAlphaOverlay(void)
{
	//make sure the overlay is allowed to be open
	if(TexturingMode == TexturingMode_NodeEdit)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Overlay cannot be opened in this mode");
		return;
	}
	s_bAlphaOverlayOpen = !s_bAlphaOverlayOpen;

	if (s_bAlphaOverlayOpen)
	{
		s_uAlphaOverlayMinimised = 0;
		CreateAlphaOverlay();
	}
	else
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_ALPHA);
	}
}
bool Tools_Mode_Texturing::AlphaOverlayIsOn(void)
{
	return s_bAlphaOverlayOpen;
}

void Tools_Mode_Texturing::CreateAlphaOverlay(void)
{
	if(!s_bAlphaOverlayOpen) return;


	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Alpha Painting",false,&s_xAlphaOverlayPosition,&s_uAlphaOverlayMinimised);
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
			
	
	pxDialog = new Asura_Tickbox_Dialog("Active [P]",xOffset,&s_bAlphaPainting);
	pxDialog->SetHelpText("If ticked then the mouse buttons will be used for painting alpha values , instead of polygon selection.\nShortcut key : P");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Selected Only",xOffset,&s_bAlphaPaint_SelectedOnly);
	pxDialog->SetHelpText("Limits any alpha editing to vertices belonging to the currently selected polygons");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Show Sphere",xOffset,&s_bAlphaPaint_ShowSphere);
	pxDialog->SetHelpText("Draws a sphere to show the radius affected");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Numerical_Dialog("Alpha (Left)",xOffset,&s_fAlphaPaint_Alpha[0],0,1.0f,0.1f,"%.3f");
	pxDialog->SetHelpText("The alpha value to be applied when the left button is pressed");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Alpha (Right)",xOffset,&s_fAlphaPaint_Alpha[1],0,1.0f,0.1f,"%.3f");
	pxDialog->SetHelpText("The alpha value to be applied when the right button is pressed");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Nearest Vert",xOffset,&s_bAlphaPaint_AffectNearestVert);
	pxDialog->SetHelpText("Alpha painting will just be applied to the vertex nearest the cursor");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Radius",xOffset,&s_fAlphaPaint_Radius,0,9999.0f,1.0f,"%.3f");
	pxDialog->SetHelpText("The radius around the mouse cursor that will be affected");
	pxDialog->SetIsEnabledFunction(IsEnabled_AlphaPaintRadius);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Use Spray Rate",xOffset,&s_bAlphaPaint_UseHoming);
	pxDialog->SetHelpText("If ticked then the alpha value will be gradually 'sprayed' on instead of changing instantly.");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("Spray Rate",xOffset,&s_fAlphaPaint_HomingRate,0,99.0f,0.1f,"%.2f");
	pxDialog->SetHelpText("The rate at which the alpha on a vertex will head towards the desired value , in units per second.");
	pxDialog->SetIsEnabledFunction(IsUsingAlphaHoming);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Apply Textures",xOffset,&s_bAlphaPaint_ApplyTexture);
	pxDialog->SetHelpText("If ticked then the selected texture will be applied at the same time as painting the alpha values.");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Allow Overwrite",xOffset,&s_bAlphaPaint_AllowTextureOverwrite);
	pxDialog->SetHelpText("Allows existing texture to be overwritten when applying textures.\nIf not ticked then textures can only be applied to polygons that don't yet have a second texture.\nThis flag is ignored when the selected texture is *None*.");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Set Second UVs",xOffset,&s_bAlphaPaint_SetUVs);
	pxDialog->SetHelpText("If checked then the secondary UV coordinates of any polygons altered will be set based on the primary coordinates.\nThe 'Secondary UV Scale' setting on the main texture mode overlay will be used when calculating the coordinates.");
	pxDialog->SetIsEnabledFunction(IsAlphaPainting);
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	xOffset.y += Asura_Dialog::LargeSpacing() - Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Get 1st Material",xOffset,Tools_Mode_Texturing::GetFirstMaterialOnPolygonForSecondary);
	pxDialog->SetHelpText("Gets the first material from a chosen polygon and sets it as the current secondary material.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Get 2nd Material",xOffset,Tools_Mode_Texturing::GetSecondMaterialOnPolygonForSecondary);
	pxDialog->SetHelpText("Gets the secondary material from a chosen polygon and sets it as the current secondary material.\n");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Mat",xOffset,SelectSecondMaterial);
	pxDialog->SetHelpText("The current secondary material to apply");
	pxDialog->m_iControlID = DialogID_SecondaryMaterial;
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	#else 

	pxDialog = new Asura_Dialog("PS2 Usage",xOffset);
	pxDialog->SetHelpText("Percentage of PS2 secondary texture memory used.\nUp to the equivalent of ten 128x128 textures are allowed.\nTextures used by polygons flagged as don't render or don't export are ignored.");
	pxDialog->m_iControlID = DialogID_PS2Usage;
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	//Add tab control
	static char* s_apszTabs[3] = {"Standard","Normal", "Specular"};
	Asura_Tabs_Dialog* pxTabsDialog = new Asura_Tabs_Dialog(xOffset,s_apszTabs,2);
	pxOverlay->AddDialog(pxTabsDialog);
	xOffset.y += Asura_Dialog::LargeSpacing();

	float fTabsStartY = xOffset.y;
	
	const int iNumTextureDisplayLines = Asura_Maths::Min( TOOLS_MAX_SECONDARY_TEXTURES + 1, 10 );
	
	//Standard textures tab	
	pxTabsDialog->SelectTab(0);
	
	pxDialog = new Asura_Dialog("Textures:",xOffset);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::SmallSpacing();
	
	pxDialog = new Asura_List_Dialog(xOffset,s_apszSecondaryTextureNames[0],&s_iAlphaPaint_TextureSlot_ListOffset,&s_iAlphaPaint_TextureSlot_ListSelection,iNumTextureDisplayLines,TOOLS_MAX_SECONDARY_TEXTURES+1,NULL);
	pxDialog->SetHelpText("Select texture to be applied when alpha painting.");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y = pxDialog->m_xMaxCoord.y + 1.0f;


	pxDialog = new Asura_Function_Dialog("Change Texture",xOffset,EditSecondaryTexture);
	pxDialog->SetHelpText("Change the texture for the current slot selected in the list above.");
	pxDialog->SetIsEnabledFunction(CanEditSecondaryTexture);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	//Normal textures tab
	pxTabsDialog->SelectTab(1);
	xOffset.y = fTabsStartY;

	pxDialog = new Asura_Dialog("Normal Maps:",xOffset);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::SmallSpacing();
	
	pxDialog = new Asura_List_Dialog(xOffset,s_apszSecondaryTextureNames[1],&s_iAlphaPaint_TextureSlot_ListOffset,&s_iAlphaPaint_TextureSlot_ListSelection,iNumTextureDisplayLines,TOOLS_MAX_SECONDARY_TEXTURES+1,NULL);
	pxDialog->SetHelpText("Select texture to be applied when alpha painting.");
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y = pxDialog->m_xMaxCoord.y + 1.0f;


	pxDialog = new Asura_Function_Dialog("Change Texture",xOffset,EditSecondaryBumpTexture);
	pxDialog->SetHelpText("Change the normal map texture for the current slot selected in the list above.");
	pxDialog->SetIsEnabledFunction(CanEditSecondaryTexture);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Clear Texture",xOffset,ClearSecondaryBumpTexture);
	pxDialog->SetHelpText("Removes the secondary normal map texture from the current slot selected in the list above.");
	pxDialog->SetIsEnabledFunction(CanEditSecondaryTexture);
	pxTabsDialog->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	pxTabsDialog->SelectTab(0);

	#endif //ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM


	pxDialog = new Asura_Function_Dialog("Set Alpha",xOffset,SetAlphaOnSelected);
	pxDialog->SetHelpText("Sets alpha on all selected values to a chosen value");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	pxDialog = new Asura_Function_Dialog("Set Second Material",xOffset,SetSecondMaterialOnSelected);
	pxDialog->SetHelpText("Sets the current secondary material on all selected polygons");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Remove Second Material",xOffset,RemoveSecondMaterialFromSelected);
	pxDialog->SetHelpText("Removes the secondary material from all selected polygons.\nDoing this will also reset the alpha to the default value.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	#else
	pxDialog = new Asura_Function_Dialog("Set Second Texture",xOffset,SetSecondTextureOnSelected);
	pxDialog->SetHelpText("Sets the current secondary texture on all selected polygons");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("Remove Second Texture",xOffset,RemoveSecondTextureFromSelected);
	pxDialog->SetHelpText("Removes the secondary texture from all selected polygons.\nDoing this will also reset the alpha to the default value.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();
	#endif



	
	pxOverlay->m_uID = TOOLS_OVERLAY_TEXTURING_ALPHA;

	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}

void Tools_Mode_Texturing::CreateTextureTileOverlay()
{
	//create an overlay for adjusting the uv scales and offsets for texture tiling
	Asura_Overlay* newOverlay;
	newOverlay = new Asura_Overlay("Texture Tile",false,&TileTextureOverlayPosition);
	
	Asura_Dialog* newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(1.0f,3.0f);
			
	newDialog  = new Asura_Numerical_Dialog("U Scale",offset,&(TextureTile_Scale.x),0.1f,1000000.0f,0.1f,"%0.2f",&TextureTile_Changed);
	newDialog->SetHelpText("Number of metres that one tile of the texture should be spread over.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog  = new Asura_Numerical_Dialog("V Scale",offset,&(TextureTile_Scale.y),0.1f,1000000.0f,0.1f,"%0.2f",&TextureTile_Changed);
	newDialog->SetHelpText("Number of metres that one tile of the texture should be spread over.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();
	
	newDialog  = new Asura_Numerical_Dialog("U Offset",offset,&(TextureTile_Offset.x),-1024.0f,1024.0f,1.0f,"%0.0f",&TextureTile_Changed);
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog  = new Asura_Numerical_Dialog("V Offset",offset,&(TextureTile_Offset.y),-1024.0f,1024.0f,1.0f,"%0.0f",&TextureTile_Changed);
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog  = new Asura_Numerical_Dialog("Offset Scale",offset,&TextureTile_OffsetScale,1.0f,1024.0f,1.0f,"%0.0f",&TextureTile_Changed);
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::LargeSpacing();

	newDialog = new Asura_Function_Dialog("Stop (Ctrl+T)",offset,TextureWrap_Stop);
	newOverlay->AddDialog(newDialog);
	
	
	newOverlay->m_uID = TOOLS_OVERLAY_TEXTURING_TILING;

	Asura_Overlay_System::SubmitNewOverlay(newOverlay,false);
	
	TextureTile_Changed = false;

}

void Tools_Mode_Texturing::CreateNodeEditOverlay(void)
{
	Asura_Overlay* pxOverlay;
	pxOverlay = new Asura_Overlay("Node Edit",false,&NodeEditOverlayPosition,&s_uNodeEditOverlayMinimised);
	
	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
			
	//pxDialog = new Asura_Tickbox_Dialog("Edit Shown Area",xOffset,&(Tools_Mode_Texturing::ResizingTexture));
//	pxOverlay->AddDialog(pxDialog);
	
//	xOffset.y+=2.5f;
	pxDialog = new Asura_Function_Dialog("X-Flip (X)",xOffset,Tools_Mode_Texturing::AdjustUV_XFlip);
	pxDialog->SetHelpText("Mirror the 'U'coordinate of the nodes.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Y-Flip (Y)",xOffset,Tools_Mode_Texturing::AdjustUV_YFlip);
	pxDialog->SetHelpText("Mirror the 'V'coordinate of the nodes.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Cycle (O)",xOffset,Tools_Mode_Texturing::AdjustUV_Cycle);
	pxDialog->SetHelpText("Cycle the UV coordinates for the nodes , so that each node gets the coordinates from the next node in turn.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Reverse UVs (V)",xOffset,Tools_Mode_Texturing::AdjustUV_Reverse);
	pxDialog->SetHelpText("Reverse the ordering of the UV coordinates in the nodes.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Rotate (R)",xOffset,Tools_Mode_Texturing::AdjustUV_Rotate);
	pxDialog->SetHelpText("Rotate the nodes by the angle specified below.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Rotate about node centre",xOffset,Tools_Mode_Texturing::AdjustUV_RotateNodeCentre);
	pxDialog->SetHelpText("Rotate the nodes by the angle below around their centre.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_AngleNumerical_Dialog("Angle", xOffset, &s_fRotationAngle, -360.0f,360.0f,1.0f,"%.1f");
	pxDialog->SetHelpText("Angle to use for rotation.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("To Corners",xOffset,Tools_Mode_Texturing::NodesToCorners);
	pxDialog->SetHelpText("Moves the nodes to the corners of the area of the texture being used.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Clamp Nodes",xOffset,Tools_Mode_Texturing::ClampNodes);
	pxDialog->SetHelpText("Clamps the nodes to the area of the texture currently being used.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Evenly  Distribute UVs",xOffset,Tools_Mode_Texturing::EvenlyDistributeUVs);
	pxDialog->SetHelpText("Attempts to distribute the nodes on a regular grid across the texture.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog("Snap To Near Texel",xOffset,Tools_Mode_Texturing::SnapNodesToNearestTexel);
	pxDialog->SetHelpText("Snaps all nodes to the nearest texel boundary.\nControl can be held down instead of using this to snap any nodes as they get moved.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		pxDialog = new Asura_Function_Dialog("Next Material",xOffset,Tools_Mode_Texturing::NextMaterial);
		pxDialog->SetHelpText("Selects the next material and applies its properties to the selected polygons.\nThe secondary textures will be unchanged.");
		pxDialog->SetIsEnabledFunction(IsEditingPrimaryChannel);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();
		
		pxDialog = new Asura_Function_Dialog("Previous Material",xOffset,Tools_Mode_Texturing::PreviousMaterial);
		pxDialog->SetHelpText("Selects the previous material and applies its properties to the selected polygons.\nThe secondary textures will be unchanged.");
		pxDialog->SetIsEnabledFunction(IsEditingPrimaryChannel);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();

		pxDialog = new Asura_Function_Dialog("Choose Material (C)",xOffset,Tools_Mode_Texturing::ChooseMaterial);
		pxDialog->SetHelpText("Selects a new material to use as the current material and applies its properties to all selected polygons.\nThe secondary textures will be unchanged.\n");
		pxDialog->SetIsEnabledFunction(IsEditingPrimaryChannel);
		pxOverlay->AddDialog(pxDialog);
		xOffset.y+=Asura_Dialog::DefaultSpacing();

	#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	pxDialog = new Asura_Tickbox_Dialog("Show Selection",xOffset,&(Tools_Polygon_Selection::ShowSelection));
	pxDialog->SetHelpText("Show selected polygons with a red outline.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Show Texel Coords",xOffset,&s_bShowCoordsInTexels);
	pxDialog->SetHelpText("If ticked then the UV coordinate display will be in texels.\nIf not used , then the UVs will be displayed in the range 0-1 (for one repeat of the texture)");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	
	pxOverlay->m_uID = TOOLS_OVERLAY_TEXTURING_NODEEDIT;

	Asura_Overlay_System::SubmitNewOverlay(pxOverlay,false);
}


void Tools_Mode_Texturing::ToggleTextureRepeatsOverlay()
{
	if(Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_TEXTURING_REPEATS))
	{
		s_bTextureRepeatsOverlayOpen = false;
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_REPEATS);
	}
	else
	{
		s_bTextureRepeatsOverlayOpen = true;
		s_uTextureRepeatsOverlayMinimised = 0;
		CreateTextureRepeatsOverlay();
	}
}

bool Tools_Mode_Texturing::IsTextureRepeatsOverlayOpen()
{
	return s_bTextureRepeatsOverlayOpen;
}

void Tools_Mode_Texturing::CreateTextureRepeatsOverlay()
{
	//don't bother , if the overlay already exists
	if(Asura_Overlay_System::DoesOverlayExist(TOOLS_OVERLAY_TEXTURING_REPEATS)) return;

	
	Asura_Overlay* newOverlay;
	newOverlay = new Asura_Overlay("Texture Repeats",false,&s_xTextureRepeatsOverlayPosition,&s_uTextureRepeatsOverlayMinimised);
	
	Asura_Dialog* newDialog;
	Asura_Vector_2 offset = Asura_Vector_2(1.0f,3.0f);

	static const char* s_apszListStrings[] =
	{
		"No Change",
		"To Corners",
		"Scale",
	};

	static const char* s_apszHelpListStrings[] =
	{
		"The nodes are left unchanged when the number of repeats are changed.",
		"Whenever the number of repeats are changed  , the nodes get moved to cover the entire area",
		"Whenever the number of repeats are changed , the nodes get scaled proportionally",
	};


	Asura_DropDown_List_Dialog* pxDropDownDialog = new Asura_DropDown_List_Dialog(offset,s_apszListStrings,&s_iTextureRepeat_NodeAdjustment,3,3,NULL,NULL,"Node Adjustment");
	pxDropDownDialog->SetHelpText("Determines how the nodes get changed whenever the the texture repeats are changed.");
	pxDropDownDialog->SetListHelpStrings(s_apszHelpListStrings);
	newOverlay->AddDialog(pxDropDownDialog);		   
	offset.y+=Asura_Dialog::LargeSpacing();

			
	newDialog  = new Asura_Numerical_Dialog("U Origin",offset,&(TextureRepeatsOrigin.x),-1000.0f,1000.0f,1.0f,"%.0f",&TextureRepeatsChanged);
	newDialog->SetHelpText("Top left coordinate of texture to use for editing.");
	newOverlay->AddDialog(newDialog);		   
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog  = new Asura_Numerical_Dialog("V Origin",offset,&(TextureRepeatsOrigin.y),-1000.0f,1000.0f,1.0f,"%.0f",&TextureRepeatsChanged);
	newDialog->SetHelpText("Top left coordinate of texture to use for editing.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog  = new Asura_Numerical_Dialog("U Repeats",offset,&(TextureRepeats.x),1.0f,1000.0f,1.0f,"%.0f",&TextureRepeatsChanged);
	newDialog->SetHelpText("Number of repeats of the texture to use horizontally for editing.");
	newOverlay->AddDialog(newDialog);		   
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog  = new Asura_Numerical_Dialog("V Repeats",offset,&(TextureRepeats.y),1.0f,1000.0f,1.0f,"%.0f",&TextureRepeatsChanged);
	newDialog->SetHelpText("Number of repeats of the texture to use vertically for editing.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Set to 1x1",offset,Tools_Mode_Texturing::SetSingleTextureRepeat);
	newDialog->SetHelpText("Set to using just one repeat of the texture.");
	newOverlay->AddDialog(newDialog);
	offset.y+=Asura_Dialog::DefaultSpacing();

	newDialog = new Asura_Function_Dialog("Fit to contain nodes",offset,Tools_Mode_Texturing::FitUsedAreaToNodes);
	newDialog->SetHelpText("Expands the number of repeats of the teture shown for editing to include all the nodes.");
	newOverlay->AddDialog(newDialog);
	
	newOverlay->m_uID = TOOLS_OVERLAY_TEXTURING_REPEATS;

	Asura_Overlay_System::SubmitNewOverlay(newOverlay,true);

	TextureRepeatsChanged = false;
}

void Tools_Mode_Texturing::UpdateTextureRepeats()
{
	if(TexturingMode == TexturingMode_NodeEdit)
	{
		s_pxTextureEdit->NodeUndo.Log_UsedAreaChange(s_pxTextureEdit->UsedArea_UV);
	}
	
	float fUOldOrigin = s_pxTextureEdit->UsedArea_UV.left;
	float fVOldOrigin = s_pxTextureEdit->UsedArea_UV.top;
	float fUScale = TextureRepeats.x / (s_pxTextureEdit->UsedArea_UV.right - s_pxTextureEdit->UsedArea_UV.left);
	float fVScale = TextureRepeats.y / (s_pxTextureEdit->UsedArea_UV.bottom - s_pxTextureEdit->UsedArea_UV.top);

	s_pxTextureEdit->UsedArea_UV.left = TextureRepeatsOrigin.x;
	s_pxTextureEdit->UsedArea_UV.top = TextureRepeatsOrigin.y;
	
	s_pxTextureEdit->UsedArea_UV.right = s_pxTextureEdit->UsedArea_UV.left+TextureRepeats.x;
	s_pxTextureEdit->UsedArea_UV.bottom = s_pxTextureEdit->UsedArea_UV.top+TextureRepeats.y;

	s_pxTextureEdit->DetermineSuitableShownArea();


	if(s_iTextureRepeat_NodeAdjustment == TEXTUREREPEAT_NODESTOCORNER)
	{
		if(s_pxTextureEdit->NumNodes==4)
		{
			s_pxTextureEdit->NodesToCorners();	
			//If in node-edit mode the uvs will be applied later.. in selection mode ,
			//apply the coordinates as we can,
			if(TexturingMode == TexturingMode_Selection)
			{
				if(WrapMode == TextureWrapMode_Spread || WrapMode == TextureWrapMode_Box)
				{
					TextureWrap_Apply();
				}
			}
		}
	}
	else if(s_iTextureRepeat_NodeAdjustment == TEXTUREREPEAT_SCALENODES)
	{
		s_pxTextureEdit->ShiftAndScaleNodes(fUOldOrigin,fVOldOrigin,TextureRepeatsOrigin.x,TextureRepeatsOrigin.y,fUScale,fVScale);
		//If in node-edit mode the uvs will be applied later.. in selection mode ,
		//apply the coordinates as we can,
		if(TexturingMode == TexturingMode_Selection)
		{
			if(WrapMode == TextureWrapMode_Spread || WrapMode == TextureWrapMode_Box)
			{
				TextureWrap_Apply();
			}
		}
	}


	
	TextureRepeatsChanged = false;
}

void Tools_Mode_Texturing::SetSingleTextureRepeat()
{
	TextureRepeatsOrigin.x = 0;
	TextureRepeatsOrigin.y = 0;
	TextureRepeats.x = 1.0f;
	TextureRepeats.y = 1.0f;
	UpdateTextureRepeats();
}


void Tools_Mode_Texturing::SetMenubarForMode()
{
	//destroy the old menubar
	if(MenuHandle != ASURA_INVALID_HANDLE)
	{
		Asura_Menubar_System::DestroyMenuSection(MenuHandle);
		MenuHandle = ASURA_INVALID_HANDLE;
   	}
	
	//create a new menubar
	if(TexturingMode == TexturingMode_Selection)
	{
		MenuHandle = Asura_Menubar_System::CreateMenuSection();
		Asura_Menubar_Title* title;

		title = Asura_Menubar_System::AddTitle(MenuHandle,"Texturing");
		title->AddItem("Options",ToggleOptionsOverlay,OptionsOverlayIsOn);
		title->AddItem("Selection",Tools_Polygon_Selection::ToggleSelectionOverlay,Tools_Polygon_Selection::SelectionOverlayIsOn);
		title->AddItem("Copy",ToggleCopyOverlay,CopyOverlayIsOn);
		title->AddItem("Texture Repeats",ToggleTextureRepeatsOverlay,IsTextureRepeatsOverlayOpen);
		title->AddItem("Find",Tools_Texturing_Find::ToggleFindOverlay,Tools_Texturing_Find::FindOverlayIsOn);
		#ifndef ASURA_DISABLE_VERTEX_ALPHA
		title->AddItem("Alpha",ToggleAlphaOverlay,AlphaOverlayIsOn);
		#endif
		title->AddItem("Texture Animation",Tools_TextureAnimation::ToggleOverlay,Tools_TextureAnimation::OverlayIsOn);

		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			title->AddItem("Material Response",Tools_Texturing_MaterialResponse::ToggleOverlay,Tools_Texturing_MaterialResponse::OverlayIsOn);
		#endif

		title->AddItem("UV Edit (Ctrl+Q)", ToggleUVEditOverlay, IsUVEditOverlayOpen);
	}
	else if(TexturingMode == TexturingMode_NodeEdit)
	{
		MenuHandle = Asura_Menubar_System::CreateMenuSection();
		Asura_Menubar_Title* title;

		title = Asura_Menubar_System::AddTitle(MenuHandle,"Texturing");
		title->AddItem("Texture Repeats",CreateTextureRepeatsOverlay);
		title->AddItem("Texture Animation",Tools_TextureAnimation::ToggleOverlay,Tools_TextureAnimation::OverlayIsOn);
	}


}

void Tools_Mode_Texturing::ApplyTextureAndUVs()
{
	ApplyTextureAndUVs(true);
}

void Tools_Mode_Texturing::ApplyTextureAndUVs(bool bPrepareUndo)
{
	if(!Tools_Polygon_Selection::IsActive()) return;
	
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}

	if(BasePolygon.ModuleIndex<0)
	{
		//need to select a base polygon first
		SetPromptedSelectionMode(PSM_BasePolygon_ApplyTextureAndUVs);
		return;
	}


	if(IsEditingChannel(0))
	{
		ApplyTexture(bPrepareUndo);
	}
	ApplyUVs(bPrepareUndo);
}

void Tools_Mode_Texturing::ApplyUVs()
{
	ApplyUVs(true);
}

void Tools_Mode_Texturing::ApplyUVs(bool bPrepareUndo)
{
	if(!Tools_Polygon_Selection::IsActive()) return;
	
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	if(BasePolygon.ModuleIndex<0)
	{
		//need to select a base polygon first
		SetPromptedSelectionMode(PSM_BasePolygon_ApplyUVs);
		return;
	}
	
	Asura_Assert(BasePolygon.ModuleIndex>=0,"No polygon has been chosen to be the base polygon");
	
	//log all selecte polygons for undo
	if(TexturingMode == TexturingMode_Selection && bPrepareUndo)
	{
		PrepareUndo();
	}

	int num_verts,base_num_verts;
	Asura_Vector_3 side[4],base_side[4];

	//get the polygon data for the base polygon
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(BasePolygon.ModuleIndex);
	Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[BasePolygon.PolyIndex];

	//Get the (normalised)sides for the base polygon
	//This is used for determing which sides of the polygons correspond
	base_num_verts = poly->NumberOfVertices();
	int i;
	for(i=0;i<base_num_verts;i++)
	{
		base_side[i] = mesh->VertexDataPtr[poly->VertexID[(i+1)%base_num_verts]].Position;		
		base_side[i] -= mesh->VertexDataPtr[poly->VertexID[i]].Position;
		base_side[i].Normalise();
	}

	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		
		//get the polygon
		poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		//get the sides for this polygon
		num_verts = poly->NumberOfVertices();
		for(i=0;i<num_verts;i++)
		{
			side[i] = mesh->VertexDataPtr[poly->VertexID[(i+1)%num_verts]].Position;		
			side[i] -= mesh->VertexDataPtr[poly->VertexID[i]].Position;
			side[i].Normalise();
		}

		//search for the pair of sides that is closest to being in the same direction
		float best_dot = -10;
		int base_best_side = 0;
		int best_side = 0;
		for(i=0;i<base_num_verts;i++)
		{
			for(int j=0;j<num_verts;j++)
			{
				float dot = base_side[i] * side[j];
				if(dot>best_dot)
				{
					best_dot = dot;
					base_best_side = i;
					best_side = j;
				}
			}
		}
		//build vertex conversion array
		int vert_conv[4];
		//the vertices for the 2 sides correspond
		vert_conv[best_side] =base_best_side;
		vert_conv[(best_side+1)%num_verts] = (base_best_side+1)%base_num_verts;
		
		if(base_num_verts == num_verts)
		{
			//base polygon and this polygon have the same number of vertices
			//just continue matchig vertices in order
			for(i=2;i<num_verts;i++)
			{
				vert_conv[(best_side+i)%num_verts] = (base_best_side+i)%base_num_verts;
			}
		}
		else if (num_verts == 3)
		{
			//this polygon is a triangle and the base is a quad
			//see which of the other sides have the closest match to determine
			//which vertex to skip
			float after_dot = base_side[(base_best_side+1)%base_num_verts] * side[(best_side+1)%num_verts];
			float before_dot = base_side[(base_best_side+base_num_verts-1)%base_num_verts] * side[(best_side+num_verts-1)%num_verts];
			if(after_dot > before_dot)
			{
				//sides after the best side match
				vert_conv[(best_side+2)%num_verts] = (base_best_side+2)%base_num_verts;
				
			}
			else
			{
				//sides before the best side match
				vert_conv[(best_side+num_verts-1)%num_verts] = (base_best_side+base_num_verts-1)%base_num_verts;
			}

		}
		else
		{
			//This polygon is a quad , base is a triangle
			//This is going to be fairly doomed anyway...
			
			for(i=2;i<num_verts;i++)
			{
				vert_conv[(best_side+i)%num_verts] = (base_best_side+i)%base_num_verts;
			}
		}


		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;
			float fUVScale = GetUVScaleForChannel(iChan);

			//and copy the uv coordinates
			for(i=0;i<num_verts;i++)
			{
				tinfo->U[iChan][i] = s_pxTextureEdit->Nodes[vert_conv[i]].x * fUVScale;
				tinfo->V[iChan][i] = s_pxTextureEdit->Nodes[vert_conv[i]].y * fUVScale;
			}
		}
		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(mesh,false);

	}
	
	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();
		
}


void Tools_Mode_Texturing::ApplyTexture()
{
	ApplyTexture(true);
}

void Tools_Mode_Texturing::ApplyTexture(bool bPrepareUndo)
{
	//places the currently selected texture on all selected polygons
	if(!Tools_Polygon_Selection::IsActive()) return;
	
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		return;
	}
	
	{

	
		//log all selected polygons for undo
		if(TexturingMode == TexturingMode_Selection && bPrepareUndo)
		{
			PrepareUndo();
		}


		PolygonID polygon;
		Tools_Polygon_Selection::StartIteration();

		while(Tools_Polygon_Selection::GetNextSelected(&polygon))
		{
			//get the mesh data
			Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//and get the texture info
			Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];


			//apply the current material
			
			// In the old system, this can actually create a new material with the selected textures. That's
			// not so easy to do with the responses, so we'll just apply whatever material we're editing

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			tinfo->m_auMaterialResponseHashID[0] = s_uSelectedMaterialResponse;
			#endif

			int iTextureID = Asura_Material_System::GetTextureID(tinfo->m_iMaterialID,1);
			int iBumpTextureID = Asura_Material_System::GetNormalMapID(tinfo->m_iMaterialID,1);
			int iSpecularTextureID = Asura_Material_System::GetSpecularMapID(tinfo->m_iMaterialID,1);
			tinfo->m_iMaterialID = GetMaterialWithTexture(s_iSelectedMaterialID,iTextureID,iBumpTextureID,iSpecularTextureID,1);
			


			//make sure the mesh gets updated in the hardware
			Tools_UpdateTexturingOnSurfaces(mesh);

		}
	}
	if(s_pxUVEdit)
	{
		s_pxUVEdit->UpdateMaterial();
	}

}


void Tools_Mode_Texturing::SwitchToNodeEditMode()
{
	Asura_Assert(TexturingMode == TexturingMode_Selection,"Changing mode from the wrong mode");
	
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	
	
	if(MultipleNodeMode)
	{
		//cease texture wrap mode , if curently doing that
		TextureWrap_Stop();
		
		PrepareUndo();
		//make sure everything has the correct texture?

		//generate the nodes
		if(!Texturing_Multiple_Nodes::DetermineNodesForSelection(s_pxTextureEdit))
		{
			//can't enter the mode after all
			Tools_Timed_Message_System::AddMessage(0xffffff,"Cannot use multiple node mode with the current polygons/texturing");
		 	return;
		}
		SetMaterialFromSelection();

	}
	else if(WrapMode == TextureWrapMode_None)
	{
		//need to make sure we have a 'base' polygon to get texturing from
		if(BasePolygon.ModuleIndex<0)
		{
			SetPromptedSelectionMode(PSM_BasePolygon_NodeEdit);
			return;			
		}

		//get the texturing from the base poly
		SetMaterialFromPolygon(BasePolygon);
		s_pxTextureEdit->CopyFromPolygon(&BasePolygon,GetPrimaryEditedChannel());

		//apply the UVs to the selection
		ApplyUVs(true);

	}
	else
	{
		PrepareUndo();
		SetMaterialFromSelection();
	}


	TexturingMode = TexturingMode_NodeEdit;
	//close down some overlays
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_OPTIONS);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_SELECTION);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_COPY);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_ALPHA);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_MODAL);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_FIND);

	//open overlay for this mode
	CreateNodeEditOverlay();
	s_pxTextureEdit->CreateMouseModeOverlay();

	//update the menubar
	SetMenubarForMode();

	//make sure some stuff is properly reinitialised
	s_pxTextureEdit->DragMode = 0;
	s_pxTextureEdit->NodeUndo.SetNumNodes(s_pxTextureEdit->NumNodes);

}

void Tools_Mode_Texturing::SwitchToSelectionMode()
{
	Asura_Assert(TexturingMode != TexturingMode_Selection,"Changing mode from the wrong mode");
	TexturingMode = TexturingMode_Selection;

	if(MultipleNodeMode)
	{
		//revert back to 4 nodes
		Texturing_Multiple_Nodes::ClearupNodes(s_pxTextureEdit);

	}
	//close some overlays
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_NODEEDIT);
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_MOUSEMODE);

	//reopen some overlays
	//(they will only actually get created if they are flagged as open)
	CreateOptionsOverlay();
	CreateCopyOverlay();
	CreateAlphaOverlay();
	Tools_Texturing_Find::CreateFindOverlay();
	Tools_Polygon_Selection::CreateSelectionOverlay();

	//update the menubar
	SetMenubarForMode();

}


void Tools_Mode_Texturing::HighlightPolygon(PolygonID& polygon)
{
//	Asura_Text_Output::Print(30,10,0xffff0000,0,"Module: %d  Poly: %d",polygon.ModuleIndex,polygon.PolyIndex);
	
	float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
	fAlpha = 255.0f * fAlpha * fAlpha;

	unsigned int colour = 0x00ffffff + (Asura_Maths::FloatToInt( fAlpha )<<24);
	
	Tools_Render::RenderPolygonOutline(polygon.PolyIndex,polygon.ModuleIndex,colour);
}


void Tools_Mode_Texturing::HandleKeyboard()
{
	//ignore other key presses if we have one of the modal overlays up
	if(StandardInputDisabledForModalOverlay) return;
	if(Asura_Overlay_System::OverlaysHaveKeyboardFocus()) return;
	if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT)) return;

 
	if(TexturingMode == TexturingMode_Selection)
	{
		if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL) && !Tools_Object_System::AreControlsEnabled())
		{
			if(Asura_UserInput::DebouncedKeyPressed('t'))// 'Ctrl + T' stops texture wrapping
			{
				TextureWrap_Stop();
			}
			if(Asura_UserInput::DebouncedKeyPressed('z'))// 'ctrl+z' for texturing undo
			{
				Undo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('y'))// 'ctrl+y' for texturing redo
			{
				Redo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('q'))
			{
				ToggleUVEditOverlay();
			}
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
		{
			// These are OK if object controls are on, because they'll switch them off

			if(Asura_UserInput::DebouncedKeyPressed('t'))// 'Shift +T' for texture tile
			{
				TextureWrapTile_Start();
			}
			if(Asura_UserInput::DebouncedKeyPressed('b'))// 'Shift + B' for tiled box wrap
			{
				TextureWrapBoxTile_Start();
			}
			if(Asura_UserInput::DebouncedKeyPressed('m'))// 'Shift +M' edit material
			{
				EditMaterialOnPolygon();
			}

		}
		else if(!Tools_Object_System::AreControlsEnabled())
		{
			if(Asura_UserInput::DebouncedKeyPressed('t'))// 'T' for texture wrap
			{
				TextureWrapSpread_Start();
			}
			if(Asura_UserInput::DebouncedKeyPressed('b'))// 'B' for box wrap
			{
				TextureWrapBox_Start();
			}
			if(Asura_UserInput::DebouncedKeyPressed('r'))// 'R' rotate 90
			{
				AdjustUV_Rotate();
			}
			if(Asura_UserInput::DebouncedKeyPressed('y'))// 'Y' for Y Flip
			{
				AdjustUV_YFlip();
			}
			if(Asura_UserInput::DebouncedKeyPressed('x'))// 'X' for X Flip
			{
				AdjustUV_XFlip();
			}
			if(Asura_UserInput::DebouncedKeyPressed('o'))// 'O' for cycle nodes
			{
				AdjustUV_Cycle();
			}
			if(Asura_UserInput::DebouncedKeyPressed('v'))// 'V' for reverse nodes
			{
				AdjustUV_Reverse();
			}
			#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
				if(Asura_UserInput::DebouncedKeyPressed('c')) // 'C' for choose texture 
				{
					ChooseMaterial();
				}
			#else
				//test
				if(Asura_UserInput::DebouncedKeyPressed('c')) 
				{
					ChooseMaterialResponse();
				}
			#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			if(Asura_UserInput::DebouncedKeyPressed('f')) // 'F' for editing texture flags
			{
				Tools_PickTexture::EditTextureFlags(PICKTEXTURE_DEFAULTFLAGEDIT);
			}
			if(Asura_UserInput::DebouncedKeyPressed('\t')) // Tab to switch modes
			{
				SwitchToNodeEditMode();
			}
			if(Asura_UserInput::DebouncedKeyPressed('u')) // 'U' to copy texturing
			{
				TextureCopyPolygon_Start();
			}
			if(Asura_UserInput::DebouncedKeyPressed('m')) // 'M' toggle multiple node mode
			{
				MultipleNodeMode = !MultipleNodeMode;
			}
			if(Asura_UserInput::DebouncedKeyPressed('i'))// 'I' to align textures
			{
				TextureAlign_Start();
			}
			if(Asura_UserInput::DebouncedKeyPressed('p'))// 'P' Toggle alpha painting
			{
				if(s_bAlphaOverlayOpen)
				{
					s_bAlphaPainting = !s_bAlphaPainting;
				}
			}
			if(Asura_UserInput::DebouncedKeyPressed(','))// '<' for texturing undo
			{
				Undo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('.'))// '>' for texturing redo
			{
				Redo();
			}
			#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			if(Asura_UserInput::DebouncedKeyPressed('g')) //apply selected material from material response overlay
			{
				//note , this only actually does anything if the overlay is open
				Tools_Texturing_MaterialResponse::ApplyEditedOrSelectedMaterial();
			}
			#endif //ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		}
	}
	else if(!Tools_Object_System::AreControlsEnabled())
	{
		if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
		{
			if(Asura_UserInput::DebouncedKeyPressed('z'))// 'ctrl+z' for  undo
			{
				s_pxTextureEdit->NodeUndo.Undo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('y'))// 'ctrl+y' for  redo
			{
				s_pxTextureEdit->NodeUndo.Redo();
			}
		}
		else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
		{
		}
		else
		{
			if(Asura_UserInput::DebouncedKeyPressed('\t')) // Tab to switch modes
			{
				SwitchToSelectionMode();
			}
			if(Asura_UserInput::DebouncedKeyPressed('r'))// 'R' rotate 90
			{
				AdjustUV_Rotate();
			}
			if(Asura_UserInput::DebouncedKeyPressed('y'))// 'Y' for Y Flip
			{
				AdjustUV_YFlip();
			}
			if(Asura_UserInput::DebouncedKeyPressed('x'))// 'X' for X Flip
			{
				AdjustUV_XFlip();
			}
			if(Asura_UserInput::DebouncedKeyPressed('o'))// 'O' for cycle nodes
			{
				AdjustUV_Cycle();
			}
			if(Asura_UserInput::DebouncedKeyPressed('v'))// 'V' for reverse nodes
			{
				AdjustUV_Reverse();
			}
			#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
				if(Asura_UserInput::DebouncedKeyPressed('c')) // 'C' for choose texture 
				{
					ChooseMaterial();
				}
			#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			if(Asura_UserInput::DebouncedKeyPressed('n')) // 'N' for snap cursor to node
			{
				s_pxTextureEdit->SnapCursorToNearestNode();
			}
			if(Asura_UserInput::DebouncedKeyPressed(','))// '<' for undo
			{
				s_pxTextureEdit->NodeUndo.Undo();
			}
			if(Asura_UserInput::DebouncedKeyPressed('.'))// '>' for redo
			{
				s_pxTextureEdit->NodeUndo.Redo();
			}

			
			#if 0 //key already used
			if(Asura_UserInput::DebouncedKeyPressed('q'))// 'Q' move the nodes to the corners
			{
				NodesToCorners();
			}
			#endif

			if(WrapMode == TextureWrapMode_Tile || WrapMode == TextureWrapMode_BoxTile)
			{
				//counter for making the scaling fame rate independent
				static float fri_counter = 0;
				fri_counter+= Asura_Timers::GetFrameTime()*20.0f;
				
				//check for keys altering texture tile paramaters
				float scale = (float)(int)fri_counter;
				if(scale)
				{
					fri_counter -= scale;
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD0)) scale = 10.0f;

					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD4))
					{
						TextureTile_Offset.x-=scale;
						if(TextureTile_Offset.x<-1024)TextureTile_Offset.x+=1024;
						TextureTile_Changed = true;
					}
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD6))
					{
						TextureTile_Offset.x+=scale;
						if(TextureTile_Offset.x>1024)TextureTile_Offset.x-=1024;
						TextureTile_Changed = true;
					}
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD8))
					{
						TextureTile_Offset.y-=scale;
						if(TextureTile_Offset.y<-1024)TextureTile_Offset.y+=1024;
						TextureTile_Changed = true;
					}
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD2))
					{
						TextureTile_Offset.y+=scale;
						if(TextureTile_Offset.y>1024)TextureTile_Offset.y-=1024;
						TextureTile_Changed = true;
					}
			
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD7))
					{
						TextureTile_Scale.x-=scale*0.1f;
						TextureTile_Scale.x=max(TextureTile_Scale.x,0.1f);
						TextureTile_Changed = true;
					}
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD9))
					{
						TextureTile_Scale.x+=scale*0.1f;
						TextureTile_Scale.x=min(TextureTile_Scale.x,1000.0f);
						TextureTile_Changed = true;
					}
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD1))
					{
						TextureTile_Scale.y-=scale*0.1f;
						TextureTile_Scale.y=max(TextureTile_Scale.y,0.1f);
						TextureTile_Changed = true;
					}
					if(Asura_UserInput::KeyPressed(ASURA_KEYID_NUMPAD3))
					{
						TextureTile_Scale.y+=scale*0.1f;
						TextureTile_Scale.y=min(TextureTile_Scale.y,1000.0f);
						TextureTile_Changed = true;
					}
				}
			}
		}
	}
}


void Tools_Mode_Texturing::UpdatePromptedSelection()
{
	if(PromptedSelectionMode == PSM_None) return;
	
	//find the polygon under the cursor
	PolygonID polygon;
	bool overPoly = false;

	// Kevin Lea - 22/06/2001 12:52:58 - if the user is doing anything with overlays or menus,
	// ignore any mousebutton/key presses etc. 
	if (!Asura_Interface::UsingMenusOrOverlays())
	{
		overPoly = Tools_Polygon_Selection::GetPolygonUnderMouse(polygon);
	}

 	//if a polygon is clicked on we need to do something dependent on the mode
	if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) && overPoly)
	{
		switch(PromptedSelectionMode)
		{
			case PSM_PolygonForTextureWrapBox :
			case PSM_PolygonForTextureWrapBoxTile :
			case PSM_PolygonForTextureWrap :
			case PSM_PolygonForTextureTile :
				TextureWrap_PolygonSelected(polygon);
				break;

			case PSM_BasePolygon_ApplyTextureAndUVs :
				BasePolygon = polygon;
				SetPromptedSelectionMode(PSM_None);
				//now have a base polygon , try applying the texture again
				ApplyTextureAndUVs(true);
				break;

			case PSM_BasePolygon_ApplyUVs :
				BasePolygon = polygon;
				SetPromptedSelectionMode(PSM_None);
				//now have a base polygon , try applying the uvs again
				ApplyUVs(true);
				break;


			case PSM_BasePolygon_NodeEdit :
				BasePolygon = polygon;
				SetPromptedSelectionMode(PSM_None);
				//now have a base polygon , try switching modes again
				SwitchToNodeEditMode();
				break;

			case PSM_PolygonForTextureCopy :
				TextureCopyPolygon_Apply(polygon);
				break;

			case PSM_ObjectCopy_From :
				//note the polygon , and ask for an object to copy to
				SelectedPolygon1 = polygon;
				SetPromptedSelectionMode(PSM_ObjectCopy_To);
				break;

			case PSM_ObjectCopy_To :
				//note the polygon , and do the object copy
				SelectedPolygon2 = polygon;
				TextureCopyObject_Apply();
				break;

			case PSM_MultipleCopy_From1 :
			case PSM_MultipleCopy_From2 :
			case PSM_MultipleCopy_To1 :
			case PSM_MultipleCopy_To2 :
				Tools_Texturing_Copy::PolygonSelected(polygon);
				break;

			case PSM_AlignTexture_PolyToAlter :
			case PSM_AlignTexture_PolyToAlignWith :
				TextureAlign_PolygonSelected(polygon);
				break;

			case PSM_EditMaterialOnPolygon :
				EditMaterialOnPolygon_PolygonSelected(polygon);
				break;

			case PSM_GetMaterialOnPolygon :
				GetMaterialOnPolygon_PolygonSelected(polygon);
				break;

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			case PSM_GetFirstMaterialOnPolygonForSecondary :
				GetFirstMaterialOnPolygonForSecondary_PolygonSelected(polygon);
				break;

			case PSM_GetSecondMaterialOnPolygonForSecondary :
				GetSecondMaterialOnPolygonForSecondary_PolygonSelected(polygon);
				break;
			#endif

			default :
				Asura_Assert(0,"Invalid selection mode");
				break;
		}
	}

	//if escape is pressed abort the prompted selection
	//(it may be necessary to put some clean up code here for some modes)
	if (!Asura_Interface::UsingMenusOrOverlays())
	{
		if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_ESCAPE))
		{
			PromptedSelectionMode = PSM_None;
		}
	}
		
}

void Tools_Mode_Texturing::RenderPromptedSelection()
{
	if(PromptedSelectionMode == PSM_None) return;
	
	//find the polygon under the cursor
	PolygonID polygon;
	bool overPoly = false;

	// Kevin Lea - 22/06/2001 12:52:58 - if the user is doing anything with overlays or menus,
	// ignore any mousebutton/key presses etc. 
	if (!Asura_Interface::UsingMenusOrOverlays())
	{
		overPoly = Tools_Polygon_Selection::GetPolygonUnderMouse(polygon);
	
		if (overPoly) // Kevin Lea - 20/06/2001 14:53:13 - highlight poly for user
		{
			HighlightPolygon(polygon);
		}
	}

	//work out print coordinates
	float print_x = Asura_Viewport::GetWidth()/2.0f;
	float print_y = Asura_Viewport::GetHeight()/2.0f;
	Asura_Text_Output::ConvertPixelsToPrintCoords(print_x,print_y);
	print_x-=0.5f;
	print_y-=0.5f;


	//display a prompt dependent on the mode
	switch(PromptedSelectionMode)
	{
		case PSM_PolygonForTextureTile :
			Asura_Text_Output::Print(print_x,print_y-2.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texture Tile :");
			Asura_Text_Output::Print(print_x,print_y-1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon for plane");
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold shift for partial axis alignment");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold control to snap to world axes");
			Asura_Text_Output::Print(print_x,print_y+1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold alt to take normal as average from selected polygons");
			break;

		case PSM_PolygonForTextureWrap :
			Asura_Text_Output::Print(print_x,print_y-2.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texture Wrap :");
			Asura_Text_Output::Print(print_x,print_y-1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon for plane");
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold shift for partial axis alignment");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold control to snap to world axes");
			Asura_Text_Output::Print(print_x,print_y+1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold alt to take normal as average from selected polygons");
			break;

		case PSM_PolygonForTextureWrapBox :
			Asura_Text_Output::Print(print_x,print_y-2.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Box Wrap :");
			Asura_Text_Output::Print(print_x,print_y-1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon for plane");
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold shift for partial axis alignment");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold control to snap to world axes");
			Asura_Text_Output::Print(print_x,print_y+1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold alt to take normal as average from selected polygons");
			break;

		case PSM_PolygonForTextureWrapBoxTile :
			Asura_Text_Output::Print(print_x,print_y-2.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Tiled Box Wrap :");
			Asura_Text_Output::Print(print_x,print_y-1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon for plane");
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold shift for partial axis alignment");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold control to snap to world axes");
			Asura_Text_Output::Print(print_x,print_y+1.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Hold alt to take normal as average from selected polygons");
			break;

		case PSM_BasePolygon_ApplyTextureAndUVs :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Apply Texture :");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select Base polygon");
			break;

		case PSM_BasePolygon_ApplyUVs :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Apply UVs :");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select Base polygon");
			break;


		case PSM_BasePolygon_NodeEdit :
			Asura_Text_Output::Print(print_x,print_y,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select Base Polygon to take coordinates from");
			break;

		case PSM_PolygonForTextureCopy :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - Polygon:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon to copy from");
			break;

		case PSM_ObjectCopy_From :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - Object:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select object to copy from");
			break;


		case PSM_ObjectCopy_To :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - Object:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select object to copy to");
			break;

		case PSM_MultipleCopy_From1 :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - From Multiple:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Choose the first reference polygon");
			break;
		case PSM_MultipleCopy_From2 :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - From Multiple:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Choose the second reference polygon");
			break;
		case PSM_MultipleCopy_To1 :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - To Multiple:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Choose the first reference polygon");
			break;
		case PSM_MultipleCopy_To2 :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing Copy - To Multiple:");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Choose the second reference polygon");
			break;

		case PSM_AlignTexture_PolyToAlter :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing alignment");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select the polygon to be altered");
			break;
		case PSM_AlignTexture_PolyToAlignWith :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Texturing alignment");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select the polygon that it should be aligned with");
			break;

		case PSM_EditMaterialOnPolygon :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Edit Material :");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon whose material should be edited.");
			break;
 		
		case PSM_GetMaterialOnPolygon :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Get Material :");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon to take the material from.");
			break;

		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		case PSM_GetFirstMaterialOnPolygonForSecondary :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Get secondary material :");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon to take the 1st material from.");
			break;
		case PSM_GetSecondMaterialOnPolygonForSecondary :
			Asura_Text_Output::Print(print_x,print_y-0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Get secondary material :");
			Asura_Text_Output::Print(print_x,print_y+0.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Select polygon to take the 2nd material from.");
			break;
		#endif

		default :
			break;
	}
	if(PromptedSelectionMode != PSM_None)
	{
		Asura_Text_Output::Print(print_x,print_y+2.5f,0xff00ff00,ASURA_TEXT_FORMAT_CENTREJUSTIFIED,"Press escape to cancel");
	}
			
}


void Tools_Mode_Texturing::HandleMouseForNodeEdit()
{
	s_pxTextureEdit->HandleMouse();


	if((WrapMode == TextureWrapMode_Tile || WrapMode == TextureWrapMode_BoxTile))
	{
		if((s_pxTextureEdit->DragMode &~ (TexturingDragMode_RotateSelected|TexturingDragMode_RotateResize)))
		{
			//only rotation of nodes is allowed in texture tile mode
			Tools_Timed_Message_System::AddMessage(0xffffff,"Only the orientation of the nodes is relevant in 'Texture Tile' mode");
			s_pxTextureEdit->DragMode = 0;
		}
	}

	//if the nodes have moved , update the texturing
	if(s_pxTextureEdit->NodesMoved)
	{
		if(MultipleNodeMode)
		{
			Texturing_Multiple_Nodes::ApplyUVCoordinates(s_pxTextureEdit);
		}
		else if(WrapMode == TextureWrapMode_None)
		{
			ApplyUVs();
		}
		else
		{
			///make sure the nodes stay in a rectangular formation
			s_pxTextureEdit->EnsureNodesAreStillOkayForTextureWrap();

			//apply the changes
			TextureWrap_Apply();
		}
		s_pxTextureEdit->NodesMoved = false;
	}
}


void Tools_Mode_Texturing::ShowStatusText()
{
	float x = 2;
	float y = 5;

	Asura_Text_Output::Print(x,y++,0xffffffff,0,"Modules selected : %d",Tools_Polygon_Selection::NumSelectedModules);
	
	if(WrapMode == TextureWrapMode_Spread)
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Texture Wrap");
	}
	else if(WrapMode == TextureWrapMode_Box)
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Box Wrap");
	}
	else if(WrapMode == TextureWrapMode_Tile)
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Texture Tile");
	}
	else if(WrapMode == TextureWrapMode_BoxTile)
	{
		Asura_Text_Output::Print(x,y++,0xffffffff,0,"Tiled Box Wrap");
	}

	if(TexturingMode == TexturingMode_NodeEdit)
	{
		//show name of current material
		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( s_uSelectedMaterialResponse );
			if( pxMaterial )
			{
				// Materials always have names in the new system
				Asura_Text_Output::Print( x, y++, 0xFFFFFFFF, 0, "Material : %s", pxMaterial->GetName().GetString() );
			}

		#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

			const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(s_iSelectedMaterialID);
			if(pxMaterial)
			{
				if(pxMaterial->m_xMaterialName.IsBlank())
				{
					Asura_Text_Output::Print(x,y++,0xffffffff,0,"Material : <Unnamed>");
				}
				else
				{
					Asura_Text_Output::Print(x,y++,0xffffffff,0,"Material : %s",pxMaterial->m_xMaterialName.GetString());
				}
			}

		#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		
		if(s_pxTextureEdit->NumNodes<10)
		{
			int iTextureWidth=0;
			int iTextureHeight=0;

			if(s_bShowCoordsInTexels)
			{
				int iTextureID = ASURA_TEXTURE_NONE;
				#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

					// With MR, if we don't have a response assigned, fall back to the old system's assignment
					if( s_uSelectedMaterialResponse != ASURA_HASH_ID_UNSET )
					{
						const Asura_MaterialResponse* const pxMaterial = Asura_MaterialResponse_System::GetMaterialByHashID( s_uSelectedMaterialResponse );
						if( pxMaterial )
						{
							iTextureID = pxMaterial->GetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE );
						}
					}
					else

				#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

				{
					iTextureID = Asura_Material_System::GetTextureID(s_iSelectedMaterialID,s_pxTextureEdit->m_iTextureChannel);
				}
				if(iTextureID != ASURA_TEXTURE_NONE)
				{
					Tools_Texture_Management::GetTextureSize(iTextureID,iTextureWidth,iTextureHeight);
				}

			}
			
			//show uv coordinates for the nodes
			for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
			{
				if(s_bShowCoordsInTexels && iTextureWidth && iTextureHeight)
					Asura_Text_Output::Print(x,y++,0xffffffff,0,"UV%d : %8.1f %8.1f",i,s_pxTextureEdit->Nodes[i].x*iTextureWidth,s_pxTextureEdit->Nodes[i].y*iTextureHeight);
				else
					Asura_Text_Output::Print(x,y++,0xffffffff,0,"UV%d : %8.3f %8.3f",i,s_pxTextureEdit->Nodes[i].x,s_pxTextureEdit->Nodes[i].y);
			}
		}
	}
	else if(s_bShowTextureUnderCursor)
	{
		PolygonID xPoly;
		if(Tools_Polygon_Selection::GetHighlitPolygon(xPoly))
		{
			//show the matieral/texture on the highlighted polygon

			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);

			#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

				for(int iMatSlot = 0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
				{
					const Asura_Hash_ID uMaterialResponseHashID = pxMesh->TextureDataPtr[ xPoly.PolyIndex ].m_auMaterialResponseHashID[iMatSlot];
					const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialResponseHashID );

					if( pxMaterial )
					{
						const char* const szMaterialName = pxMaterial->GetName().GetString();
						Asura_Text_Output::Print( x, y++, 0xFFFFFFFF, 0, "Material Under Cursor (%d) : %s", iMatSlot,szMaterialName );

						for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++u )
						{
							const int iTextureID = pxMaterial->GetTexture(u);
							if( iTextureID != ASURA_TEXTURE_NONE )
							{
								static const char* const aszTEXTURE_SLOT_NAMES[] =
								{
									"Albedo",
									"Normal",
									"Specular",
									"Transparency",
									"Self-illumination",
									"Environment",
									"Reflectivity",
									"Parallax",
									"Albedo Detail",
									"Normal Detail",
								};

								const u_int uARRAY_SIZE = sizeof( aszTEXTURE_SLOT_NAMES ) / sizeof( aszTEXTURE_SLOT_NAMES[0] );

								const char* szSlotName;
								if( u < uARRAY_SIZE )
								{
									szSlotName = aszTEXTURE_SLOT_NAMES[u];
								}
								else
								{
									Asura_Assert( false, "Texture slot names array needs to be updated." );
									szSlotName = "Unknown";
								}

								const char* const szTextureName = Asura_Texture_Management::GetTextureName( iTextureID )->GetString();
								Asura_Text_Output::Print( x, y++, 0xFFFFFFFF, 0, "[%s] %s", szSlotName, szTextureName );
							}
						}
					}
				}

			#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

				int iMaterialID = pxMesh->TextureDataPtr[xPoly.PolyIndex].m_iMaterialID;
				if(iMaterialID != ASURA_MATERIAL_NONE)
				{
					const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(iMaterialID);
					if(pxMaterial->m_iTextureID[0] != ASURA_TEXTURE_NONE)
					{
						
						Asura_Text_Output::Print(x,y++,0xffffffff,0,"Textures Under Cursor :");

						// Diffuse textures
						for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
						{
							if(pxMaterial->m_iTextureID[i] != ASURA_TEXTURE_NONE)
							{
								Asura_Text_Output::Print(x,y++,0xffffffff,0,"[Diffuse] %s",Asura_Texture_Management::GetTextureName(pxMaterial->m_iTextureID[i])->GetString());
							}
						}

						// Bump maps
						for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
						{
							if(pxMaterial->m_iNormalMapID[i] != ASURA_TEXTURE_NONE)
							{
								Asura_Text_Output::Print(x,y++,0xffffffff,0,"[NormalMap] %s",Asura_Texture_Management::GetTextureName(pxMaterial->m_iNormalMapID[i])->GetString());
							}
						}

						// Specular maps
						for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
						{
							if(pxMaterial->m_iSpecularMapID[i] != ASURA_TEXTURE_NONE)
							{
								Asura_Text_Output::Print(x,y++,0xffffffff,0,"[Specular] %s",Asura_Texture_Management::GetTextureName(pxMaterial->m_iSpecularMapID[i])->GetString());
							}
						}

						// Material
						if(!pxMaterial->m_xMaterialName.IsBlank())
						{
							Asura_Text_Output::Print(x,y++,0xffffffff,0,"Material %s",pxMaterial->m_xMaterialName.GetString());
						}

						// Anim
						if(pxMaterial->m_uAnimHash)
						{
							const Asura_TextureAnimation* pxAnim = Asura_TextureAnimation_System::GetAnimationByHash(pxMaterial->m_uAnimHash);
							Asura_Text_Output::Print(x,y++,0xffffffff,0,"Anim : %s",pxAnim ? pxAnim->GetName() : "*Unknown*");
						}
					}
				}

			#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
		}
	}
}

void Tools_Mode_Texturing::TextureWrapSpread_Start()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	//if we are already in the middle of doing texture wrapping , just update the texture
	//coordinates now
	if(WrapMode == TextureWrapMode_Spread)
	{
		TextureWrap_Apply();
	}
	else
	{
		TextureWrap_Stop();//clearup old mode
		//need to ask for a polygon , so set the selection mode
		SetPromptedSelectionMode(PSM_PolygonForTextureWrap);
	}
}

void Tools_Mode_Texturing::TextureWrapBox_Start()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	//if we are already in the middle of doing texture wrapping , just update the texture
	//coordinates now
	if(WrapMode == TextureWrapMode_Box)
	{
		TextureWrapBox_Apply();
	}
	else
	{
		TextureWrap_Stop();//clearup old mode
		//need to ask for a polygon , so set the selection mode
		SetPromptedSelectionMode(PSM_PolygonForTextureWrapBox);
	}
}

void Tools_Mode_Texturing::TextureWrapBoxTile_Start()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	//if we are already in the middle of doing texture wrapping , just update the texture
	//coordinates now
	if(WrapMode == TextureWrapMode_BoxTile)
	{
		TextureWrapBox_Apply();
	}
	else
	{
		TextureWrap_Stop();//clearup old mode
		//need to ask for a polygon , so set the selection mode
		SetPromptedSelectionMode(PSM_PolygonForTextureWrapBoxTile);
	}
}




void Tools_Mode_Texturing::TextureWrapTile_Start()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}
	//if we are already in the middle of doing texture wrapping , just update the texture
	//coordinates now
	if(WrapMode == TextureWrapMode_Tile)
	{
		TextureWrap_Apply();
	}
	else
	{
		TextureWrap_Stop();//clearup old mode
		//need to ask for a polygon , so set the selection mode
		SetPromptedSelectionMode(PSM_PolygonForTextureTile);
	}
}

void Tools_Mode_Texturing::TextureWrap_PolygonSelected(PolygonID polygon)
{
	//Need to work out x and y axes for the plane to be projected onto.
	//Look at the corners of the selected polygon , and search for the side closest
	//to being a right angle

	//get the mesh and polygon data
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
	Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
	Asura_Vector_3 normal = mesh->NormalsDataPtr[polygon.PolyIndex].FaceNormal;

	if(Asura_UserInput::KeyPressed(ASURA_KEYID_ALT))
	{
		//instead of using the normal from the polygon , take the average of all selected polygon normals
		PolygonID xPoly;
		Tools_Polygon_Selection::StartIteration();

		normal = Asura_Maths::ZeroVector3;
		while(Tools_Polygon_Selection::GetNextSelected(&xPoly))
		{
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);
			if(!pxMesh->m_bFaceNormalsValid) pxMesh->CreateFaceNormals();
			normal += pxMesh->NormalsDataPtr[xPoly.PolyIndex].FaceNormal;
		}
		normal.Normalise();
	}


	if(Asura_UserInput::KeyPressed(ASURA_KEYID_CTRL))
	{
		//use the nearest set of world axes
		float abs_x = ABS(normal.x);
		float abs_y = ABS(normal.y);
		float abs_z = ABS(normal.z);
		
		TextureWrap_X.Zero();
		TextureWrap_Y.Zero();

		if(abs_x>abs_y && abs_x>abs_z)
		{
			TextureWrap_X.y = 1.0f;
			TextureWrap_Y.z = 1.0f;
		}	
		else if(abs_y>abs_z)
		{
			TextureWrap_X.z = 1.0f;
			TextureWrap_Y.x = 1.0f;
		}
		else
		{
			TextureWrap_X.x = 1.0f;
			TextureWrap_Y.y = 1.0f;
		}
	}

	else if(Asura_UserInput::KeyPressed(ASURA_KEYID_SHIFT))
	{
		//choose axes that arein the plane of the polygon , but as aligned to the world
		//axes as possible
		float abs_x = ABS(normal.x);
		float abs_y = ABS(normal.y);
		float abs_z = ABS(normal.z);
		
		TextureWrap_Y.Zero();
		if(abs_y>abs_x && abs_y>abs_z)
		{
			if(abs_x<abs_z)
				TextureWrap_Y.x = 1.0f;
			else
				TextureWrap_Y.z = 1.0f;
		
		}
		else
		{
			TextureWrap_Y.y = 1.0f;
		}
		TextureWrap_X = TextureWrap_Y ^ normal;
		TextureWrap_X.Normalise();

		TextureWrap_Y = normal ^ TextureWrap_X;
		TextureWrap_Y.Normalise();
		

	}
	else
	{
		int num_verts = poly->NumberOfVertices();
		float low_dot = 1000.0f;//(some suitably large number...)
		for(int i=0;i<num_verts ;i++)
		{
			//get the 2 sides
			Asura_Vector_3 side1 = mesh->VertexDataPtr[poly->VertexID[(i+1)%num_verts]].Position - mesh->VertexDataPtr[poly->VertexID[i]].Position;
			Asura_Vector_3 side2 = mesh->VertexDataPtr[poly->VertexID[(num_verts+i-1)%num_verts]].Position - mesh->VertexDataPtr[poly->VertexID[i]].Position;
			//normalise
			side1.Normalise();
			side2.Normalise();

			float dot = side1 * side2;
			if(dot<0) dot = -dot;

			//is this closer to being 90 degrees?
			if(dot<low_dot)
			{
				low_dot = dot;
				TextureWrap_X = side1;
			}
		}

	

		//Now need to make sure the y vector is perpendicular to the x vector , but in the same plane
		TextureWrap_Y = normal ^ TextureWrap_X;
		TextureWrap_Y.Normalise();
	}

	if(PromptedSelectionMode == PSM_PolygonForTextureWrapBox || PromptedSelectionMode == PSM_PolygonForTextureWrapBoxTile)
	{
		//for box wrapping , prefer a vertical y vector
		if(ABS(TextureWrap_X.y)>ABS(TextureWrap_Y.y))
		{
			Asura_Vector_3 temp = TextureWrap_X;
			TextureWrap_X = TextureWrap_Y;
			TextureWrap_Y = temp;
		}
		if(TextureWrap_Y.y<0)
		{
			TextureWrap_Y = TextureWrap_Y*-1;
		}

		TextureWrap_Z = TextureWrap_X ^ TextureWrap_Y;

		//pick the z vector so that it points the other way from the normal
		if(TextureWrap_Z*normal>0)
		{
			TextureWrap_X =TextureWrap_X*-1;
			TextureWrap_Z =TextureWrap_Z*-1;
		}


	}
	
	
	//set the current texture wrap mode
	if(PromptedSelectionMode == PSM_PolygonForTextureWrap)
	{
		WrapMode = TextureWrapMode_Spread;
	}
	else if(PromptedSelectionMode == PSM_PolygonForTextureWrapBox)
	{
		WrapMode = TextureWrapMode_Box;
	}
	else if(PromptedSelectionMode == PSM_PolygonForTextureWrapBoxTile)
	{
		WrapMode = TextureWrapMode_BoxTile;
		CreateTextureTileOverlay();
	}
	else
	{
		WrapMode = TextureWrapMode_Tile;
		CreateTextureTileOverlay();
	}

	SetPromptedSelectionMode(PSM_None);


	//need to check that the nodes are positioned in a nice rectangle
	s_pxTextureEdit->InitaliseNodesForTextureWrap();

	//now we have the 2 normals on the plane.. go ahead and project the texture onto the polygons
	TextureWrap_Apply();
}

void Tools_Mode_Texturing::TextureWrap_Apply()
{
	Asura_Assert(WrapMode!=TextureWrapMode_None,"Should be in one of the texture wrap modes");

	if(WrapMode==TextureWrapMode_Box || WrapMode==TextureWrapMode_BoxTile)
	{
		//redirect to the box wrapping function
		TextureWrapBox_Apply();
		return;
	}

	//log all selected polygons for undo
	if(TexturingMode == TexturingMode_Selection)
	{
		PrepareUndo();
	}

	if(s_bApplyMaterialOnTextureWrap)
	{
		ApplyTexture(false);
	}


	//First pass through the selection...
	//Need to get every vertex's coordinates in terms of the basis vectors TextureWrap_X
	//and TextureWrap_Y.
	//Also need to get the min and maximum value for these coordinates (so we can scale the	
	//texture coordinates appropriately)

	Asura_Bounding_Rectangle bounds;
	bounds.Reset();

	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		//and the modules location
		Asura_Vector_3  module_loc;
		Asura_Module_System::GetPositionOfModule(polygon.ModuleIndex,module_loc);

		//get the polygon
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;

			
			for(int j=0; j<poly->NumberOfVertices(); j++)
			{
				//convert the vertex into world cordinates
				Asura_Vector_3 pos = module_loc + mesh->VertexDataPtr[poly->VertexID[j]].Position;

				//take the dot product with TextureWrap_X and TextureWrap_Y to get coordinates
				//using those vectors as a basis.
				//Write the output into the texure info for the moment
				tinfo->U[iChan][j] = pos * Tools_Mode_Texturing::TextureWrap_X;
				tinfo->V[iChan][j] = pos * Tools_Mode_Texturing::TextureWrap_Y;

				//include in the bounding rectangle
				Asura_Vector_2 temp(tinfo->U[iChan][j],tinfo->V[iChan][j]);
				bounds.Set(temp);
			}
		}
	}

	Asura_Assert(Tools_Mode_Texturing::s_pxTextureEdit->NumNodes == 4,"Expecting 4 texture nodes (supposed to be in a rectangle)");
	Asura_Vector_2* Nodes = Tools_Mode_Texturing::s_pxTextureEdit->Nodes;

	//The coordinates are supposed to be projected into the rectangle made up by the Nodes
	
	//get the sides of the node rectangle
	Asura_Vector_2 Nodes_U = Nodes[1] - Nodes[0];
	Asura_Vector_2 Nodes_V = Nodes[3] - Nodes[0];
	float Scale_U,Scale_V;

	
	if(WrapMode == TextureWrapMode_Spread)
	{
		//Get a scaling factor by comparing the extents of the bounding rectangle , with the width and height
		//(cooridinates will get saled up correctly based on the area used by the nodes , since the sides haven't
		//been normalised)
		Scale_U = 1.0f / (bounds.MaxX - bounds.MinX);
		Scale_V = 1.0f / (bounds.MaxY - bounds.MinY);
	}
	else
	{
		//normalise the sides
		Nodes_U.Normalise();
		Nodes_V.Normalise();
		
		Scale_U = 1.0f / TextureTile_Scale.x;
		Scale_V = 1.0f / TextureTile_Scale.y;
	}

	//Iterate through the points again mapping the coordinates into the area
	//bounds by the Nodes
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");

		//get the polygon
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;
			
			float fUVScale = GetUVScaleForChannel(iChan);

			for(int j=0; j<poly->NumberOfVertices(); j++)
			{
				if(WrapMode == TextureWrapMode_Spread)
				{
					//make relative to the top left corner of the bounding rectangle
					tinfo->U[iChan][j] -= bounds.MinX;
					tinfo->V[iChan][j] -= bounds.MinY;
					//scale
					tinfo->U[iChan][j] *= Scale_U;
					tinfo->V[iChan][j] *= Scale_V;

					//rotate
					float temp_u = tinfo->U[iChan][j];
					float temp_v = tinfo->V[iChan][j];
					tinfo->U[iChan][j] = Nodes_U.x * temp_u + Nodes_V.x * temp_v;
					tinfo->V[iChan][j] = Nodes_U.y * temp_u + Nodes_V.y * temp_v;

					//and offset relative to the first node
					tinfo->U[iChan][j] += Nodes[0].x;
					tinfo->V[iChan][j] += Nodes[0].y;

				}
				else
				{
					//scale
					tinfo->U[iChan][j] *= Scale_U;
					tinfo->V[iChan][j] *= Scale_V;

					//rotate
					float temp_u = tinfo->U[iChan][j];
					float temp_v = tinfo->V[iChan][j];
					tinfo->U[iChan][j] = Nodes_U.x * temp_u + Nodes_V.x * temp_v;
					tinfo->V[iChan][j] = Nodes_U.y * temp_u + Nodes_V.y * temp_v;
					
					//apply uv offset
					tinfo->U[iChan][j] += TextureTile_Offset.x/TextureTile_OffsetScale;
					tinfo->V[iChan][j] += TextureTile_Offset.y/TextureTile_OffsetScale;
				}
				
				tinfo->U[iChan][j] *= fUVScale;
				tinfo->V[iChan][j] *= fUVScale;
			}
		}

		
		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(mesh);
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();
	

	TextureTile_Changed = false;
}

void Tools_Mode_Texturing::TextureWrapBox_Apply()
{
	Asura_Assert(WrapMode==TextureWrapMode_Box || WrapMode==TextureWrapMode_BoxTile,"");
	if(!Tools_Polygon_Selection::NumSelectedModules) return;

	//log all selected polygons for undo
	if(TexturingMode == TexturingMode_Selection)
	{
		PrepareUndo();
	}

	if(s_bApplyMaterialOnTextureWrap)
	{
		ApplyTexture(false);
	}
	
	Asura_Assert(Tools_Mode_Texturing::s_pxTextureEdit->NumNodes == 4,"Expecting 4 texture nodes (supposed to be in a rectangle)");
	Asura_Vector_2* Nodes = Tools_Mode_Texturing::s_pxTextureEdit->Nodes;

	//The coordinates are supposed to be projected into the rectangle made up by the Nodes
	
	//get the sides of the node rectangle
	Asura_Vector_2 Nodes_U = Nodes[1] - Nodes[0];
	Asura_Vector_2 Nodes_V = Nodes[3] - Nodes[0];

	if(WrapMode == TextureWrapMode_BoxTile)
	{
		//normalises when tiling ,since they effectively get used for a 2d matrix
		Nodes_U.Normalise();
		Nodes_V.Normalise();
	}

	
	//get a matrix to transform the points so that they use the axes defined by TextureWrap_X,Y and Z
	Asura_Matrix_3x3 inverseMat( TextureWrap_X.x,TextureWrap_X.y,TextureWrap_X.z,
								 TextureWrap_Y.x,TextureWrap_Y.y,TextureWrap_Y.z,
								 TextureWrap_Z.x,TextureWrap_Z.y,TextureWrap_Z.z );

	Asura_Bounding_Box bounds;
	bounds.Reset();


	enum
	{
		BoxMinX,
		BoxMaxX,
		BoxMinY,
		BoxMaxY,
		BoxMinZ,
		BoxMaxZ,
	};


	//Go through all the polygons working out the extennts in the axes defined by TextureWrap_X,Y and Z
	PolygonID polygon;
 	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		if(!mesh->AreNormalsValid()) mesh->CreateNormals();
		//and the modules location
		Asura_Vector_3  module_loc;
		Asura_Module_System::GetPositionOfModule(polygon.ModuleIndex,module_loc);

		//get the polygon normal , and transform into the space of the box matrix
		Asura_Vector_3 normal = mesh->NormalsDataPtr[polygon.PolyIndex].FaceNormal;
		normal = normal * inverseMat;

		//work out which side the normal is closest to
		int side;
		if(ABS(normal.x)>ABS(normal.y) && ABS(normal.x)>ABS(normal.z))
		{
			side = (normal.x>0) ? BoxMaxX : BoxMinX;
		}
		else if(ABS(normal.y)>ABS(normal.z))
		{
			side = (normal.y>0) ? BoxMaxY : BoxMinY;
		}
		else
		{
			side = (normal.z>0) ? BoxMaxZ : BoxMinZ;
		}


		//get the polygon
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		for(int j=0; j<poly->NumberOfVertices(); j++)
		{
			//convert the vertex into world cordinates
			Asura_Vector_3 pos = module_loc + mesh->VertexDataPtr[poly->VertexID[j]].Position;

			pos = pos * inverseMat;

			bounds.Set(pos);

			//rotate acoording to the side faced , and put the x and y coordinates in the texture coords
			float temp;
			switch (side)
			{
				case BoxMinX :
					temp = pos.x;
					pos.x = -pos.z;
					pos.z = temp;
					break;
				case BoxMaxX :
					temp = -pos.x;
					pos.x = pos.z;
					pos.z = temp;
					break;
				case BoxMinY :
					temp = pos.y;
					pos.y = -pos.z;
					pos.z = temp;
					break;
				case BoxMaxY :
					temp = -pos.y;
					pos.y = pos.z;
					pos.z = temp;
					break;
				case BoxMinZ :
					break;//do nothing
				case BoxMaxZ :
					pos.x = -pos.x;
					pos.z = -pos.z;
					break;

			}

			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;

				tinfo->U[iChan][j] = pos.x;
				tinfo->V[iChan][j] = pos.y;
			}
		}

		//write the side into the texture id for now , since it will be needed again for the second pass
		short sMaterial = (short) tinfo->m_iMaterialID;
		tinfo->m_iMaterialID = (side<<16)|(u_short)sMaterial;

}

	//get the centre , and the size of the box
	Asura_Vector_3 centre = bounds.CalculateCentre();
	float maxExtent = bounds.MaxX-bounds.MinX;
	maxExtent = max(maxExtent,bounds.MaxY-bounds.MinY);
	maxExtent = max(maxExtent,bounds.MaxZ-bounds.MinZ);

	if(maxExtent==0) maxExtent = 1;

	//get the centre with each of the side rotations applied
	//(2d vector , since only x and y will matter
	Asura_Vector_2 rotatedCentre[6] =
	{
		Asura_Vector_2(-centre.z,centre.y),
		Asura_Vector_2(centre.z,centre.y),
		Asura_Vector_2(centre.x,-centre.z),
		Asura_Vector_2(centre.x,centre.z),
		Asura_Vector_2(centre.x,centre.y),
		Asura_Vector_2(-centre.x,centre.y),
	};


	
	//Now go through the selection again working out the texture coords
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");

		//get the polygon
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		//and the modules location
		Asura_Vector_3  module_loc;
		Asura_Module_System::GetPositionOfModule(polygon.ModuleIndex,module_loc);

		//recover the side the normal was facing
		int side = tinfo->m_iMaterialID>>16;
		tinfo->m_iMaterialID = (short)(tinfo->m_iMaterialID & 0xffff);
		Asura_Assert(side>=0 && side<6,"");



		for(int iChan=0 ;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;
			float fUVScale = GetUVScaleForChannel(iChan);

			for(int j=0; j<poly->NumberOfVertices(); j++)
			{
				float x = tinfo->U[iChan][j];
				float y = tinfo->V[iChan][j];

				if(WrapMode == TextureWrapMode_Box)
				{
					//apply the offset
					x -= rotatedCentre[side].x;
					y -= rotatedCentre[side].y;

				
					//apply scale  to get x and y in range 0 to 1
					x = (x/maxExtent) + 0.5f;
					y = (y/maxExtent) + 0.5f;

					//make sure values are in the expected range
					if(x<0) x = 0;
					if(x>1) x = 1;
					if(y<0) y = 0;
					if(y>1) y = 1;
				
		
					tinfo->U[iChan][j] = (Nodes_U.x*x + Nodes_V.x*y) + Nodes[0].x;
					tinfo->V[iChan][j] = (Nodes_U.y*x + Nodes_V.y*y) + Nodes[0].y;

					tinfo->U[iChan][j] *= fUVScale;
					tinfo->V[iChan][j] *= fUVScale;
				}
				else
				{
					Asura_Assert(WrapMode == TextureWrapMode_BoxTile,"Unexpected texture wrap mode");
					//scale
					x *= fUVScale/TextureTile_Scale.x;
					y *= fUVScale/TextureTile_Scale.y;

					//rotate
				   	tinfo->U[iChan][j] = Nodes_U.x * x + Nodes_V.x * y;
				   	tinfo->V[iChan][j] = Nodes_U.y * x + Nodes_V.y * y;

					//apply uv offset
					tinfo->U[iChan][j] += TextureTile_Offset.x/TextureTile_OffsetScale;
					tinfo->V[iChan][j] += TextureTile_Offset.y/TextureTile_OffsetScale;

					
				}
			}
		}

		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(mesh);

	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

	TextureTile_Changed = false;
}

void Tools_Mode_Texturing::TextureWrap_Stop()
{
	if(TexturingMode == TexturingMode_NodeEdit)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"Cannot stop texture wrapping while in node edit mode");
		return;
	}
	
	//reset the mode
	WrapMode = TextureWrapMode_None;
	//and close the overlay
	Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_TEXTURING_TILING);
}


void Tools_Mode_Texturing::TextureCopyPolygon_Start()
{
	if(!Tools_Polygon_Selection::NumSelectedModules)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"No polygons have been selected");
		return;
	}

	//ask the user to select the polygon to copy from
	SetPromptedSelectionMode(PSM_PolygonForTextureCopy);
}

void Tools_Mode_Texturing::TextureCopyPolygon_Apply(PolygonID polygon)
{
	u_int uEditedChannelsBackup = s_uEditedChannels;
	BasePolygon = polygon;

	PrepareUndo();
	
	//copy each of the currently edited channels in turn
	for(int iChan = ASURA_MATERIAL_MAXTEXTURES-1;iChan>=0;iChan--)
	{
		if((1<<iChan) & uEditedChannelsBackup)
		{
			s_uEditedChannels = (1<<iChan);
			//get the texturing from the selected polygon
			SetMaterialFromPolygon(BasePolygon);
			s_pxTextureEdit->CopyFromPolygon(&BasePolygon,iChan);

			//and apply the texturing to the selection
			ApplyTextureAndUVs(false);

			//switch off the prompt mode
			SetPromptedSelectionMode(PSM_None);
		}
	}

	s_uEditedChannels = uEditedChannelsBackup;
}

void Tools_Mode_Texturing::TextureCopyObject_Start()
{
	//ask the user to select the polygon to copy from
	SetPromptedSelectionMode(PSM_ObjectCopy_From);
}

void Tools_Mode_Texturing::TextureCopyObject_Apply()
{
	if(SelectedPolygon1.ModuleIndex == SelectedPolygon2.ModuleIndex)
	{
		//no point in copying from the object to itself...
		return;
	}

	//get the mesh data for the 2 objects
	Asura_Mesh* mesh1 = Asura_Module_System::GetMeshForModule(SelectedPolygon1.ModuleIndex);
	Asura_Mesh* mesh2 = Asura_Module_System::GetMeshForModule(SelectedPolygon2.ModuleIndex);

	//check number of polygons
	if(mesh1->NumberOfPolygons != mesh2->NumberOfPolygons)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"The objects contain differing numbers of polygons");
		return;
	}

	//check that the triangles/quads match up
	int i;
	for(i=0;i<mesh1->NumberOfPolygons;i++)
	{
		Asura_Mesh_Polygon* poly1 = &mesh1->PolygonDataPtr[i];
		Asura_Mesh_Polygon* poly2 = &mesh2->PolygonDataPtr[i];

		if(poly1->IsQuad() != poly2->IsQuad())
		{
			Tools_Timed_Message_System::AddMessage(0xffffff,"The merging for the objects doesn't match");
			return;
  		}
	}

	//we can go ahead with the copy.. prepare undo for the object
	PolygonID polygon = SelectedPolygon2;
	for(i=0;i<mesh2->NumberOfPolygons;i++)
	{
		polygon.PolyIndex = i;
		Tools_Texturing_UndoData_TextureInfo::Create(polygon);
	}

	//and do the copy
	for(i=0;i<mesh1->NumberOfPolygons;i++)
	{
		Asura_Mesh_Polygon* poly1 = &mesh1->PolygonDataPtr[i];
		Asura_Mesh_Polygon* poly2 = &mesh2->PolygonDataPtr[i];
		
		//check for matching vertex indces on the 2 polygons
		int num_verts = poly1->NumberOfVertices();
		int VertConv[4];
	 
		int j;
		for(j=0;j<num_verts;j++)
		{
			int k;
			for(k=0;k<num_verts;k++)
			{
				if (poly1->VertexID[j] == poly2->VertexID[k])
				{
					VertConv[j] = k;
					break;
				}
			}
			if(k==num_verts)break;
		}
		if(j<num_verts)continue;//polygons have different numbered verts;
		
		//get the texture info for the polygons
		Asura_Mesh_TextureInfo* tinfo1 = &mesh1->TextureDataPtr[i];
		Asura_Mesh_TextureInfo* tinfo2 = &mesh2->TextureDataPtr[i];
		
		//copy the texture index
		if(IsEditingChannel(0) && !s_bLimitUVEditingToSelectedTexture)
		{
			tinfo2->m_iMaterialID = tinfo1->m_iMaterialID;
		}
		
		//and the uvs
		for(int iChan = 0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo2->m_iMaterialID)) continue;

			for(j=0;j<num_verts;j++)
			{
				tinfo2->U[iChan][VertConv[j]] = tinfo1->U[iChan][j];
				tinfo2->V[iChan][VertConv[j]] = tinfo1->V[iChan][j];
			}
		}
	}

	//make sure the mesh gets updated in the hardware
	Tools_UpdateTexturingOnSurfaces(mesh2);

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();


}

void Tools_Mode_Texturing::TextureAlign_Start()
{
	//ask for the polygon to be affected
	SetPromptedSelectionMode(PSM_AlignTexture_PolyToAlter);
}

void Tools_Mode_Texturing::TextureAlign_PolygonSelected(PolygonID polygon)
{
	//make sure the selected polygon is textured
	Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
	Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
	Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

	if(tinfo->m_iMaterialID == ASURA_MATERIAL_NONE)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"That polygon isn't textured");
		return;
	}

	if(PromptedSelectionMode == PSM_AlignTexture_PolyToAlter)
	{
		//note this polygon
		SelectedPolygon1 = polygon;
		//and ask for the polygon to align with
		SetPromptedSelectionMode(PSM_AlignTexture_PolyToAlignWith);
	}
	else
	{
		Asura_Assert(PromptedSelectionMode == PSM_AlignTexture_PolyToAlignWith,"Not a valid selection mode for this function");

		//note this polygon
		SelectedPolygon2 = polygon;
		//if the texture alignment works , switch off the selection mode
		if(TextureAlign_Apply())
		{
			SetPromptedSelectionMode(PSM_None);
		}
	}
}

bool Tools_Mode_Texturing::TextureAlign_Apply()
{
	//get the polygon and texturing data for the 2 polygons
	Asura_Mesh* mesh1 = Asura_Module_System::GetMeshForModule(SelectedPolygon1.ModuleIndex);
	Asura_Assert(mesh1->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
	Asura_Mesh_Polygon* poly1 = &mesh1->PolygonDataPtr[SelectedPolygon1.PolyIndex];
	Asura_Mesh_TextureInfo* tinfo1 = &mesh1->TextureDataPtr[SelectedPolygon1.PolyIndex];
	Asura_Mesh* mesh2 = Asura_Module_System::GetMeshForModule(SelectedPolygon2.ModuleIndex);
	Asura_Assert(mesh2->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
	Asura_Mesh_Polygon* poly2 = &mesh2->PolygonDataPtr[SelectedPolygon2.PolyIndex];
	Asura_Mesh_TextureInfo* tinfo2 = &mesh2->TextureDataPtr[SelectedPolygon2.PolyIndex];

	//make sure both polygons are still textured
	if(tinfo1->m_iMaterialID < 0 || tinfo2->m_iMaterialID < 0)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"One of the polygon's isn't textured");
		return true;//while this hasn't actually worked , we don't want to continue asking for polygons in this case
	}

	bool changed = false;

	
	Asura_Vector_3  module_loc1;
	Asura_Vector_3  module_loc2;
	Asura_Module_System::GetPositionOfModule(SelectedPolygon1.ModuleIndex,module_loc1);
	Asura_Module_System::GetPositionOfModule(SelectedPolygon2.ModuleIndex,module_loc2);
	
	//check each vertex in the first polygon for a matching one in the second polygon
	int num_verts1 = poly1->NumberOfVertices();
	int num_verts2 = poly2->NumberOfVertices();

	for(int i=0;i<num_verts1;i++)
	{
		Asura_Vector_3 pos1 = module_loc1 + mesh1->VertexDataPtr[poly1->VertexID[i]].Position;
		int j;
		for( j=0;j<num_verts2;j++)
		{
			Asura_Vector_3 pos2 = module_loc2 + mesh2->VertexDataPtr[poly2->VertexID[j]].Position;

			if((pos1-pos2).Magnitude()<0.004f)
			{
				//found a matching vert
				break;
			}
		}

		if(j<num_verts2)
		{
			if(!changed)
			{
				//log for undo
				Tools_Texturing_UndoData_TextureInfo::Create(SelectedPolygon1);
				changed = true;
			}

			for(int iChan = 0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,tinfo2->m_iMaterialID)) continue;

				//get the offset required to reach the desired UV coordinates , and adjust
				//them by integer multiples so that the offset is between -0.5 and 0.5
				float fUOffset = tinfo2->U[iChan][j]-tinfo1->U[iChan][i];
				float fVOffset = tinfo2->V[iChan][j]-tinfo1->V[iChan][i];

				fUOffset -= Asura_Maths::FloatToInt_Chop(fUOffset);
				if(fUOffset>0.5f) fUOffset-=1.0f;
				else if(fUOffset<-0.5f) fUOffset+=1.0f;

				fVOffset -= Asura_Maths::FloatToInt_Chop(fVOffset);
				if(fVOffset>0.5f) fVOffset-=1.0f;
				else if(fVOffset<-0.5f) fVOffset+=1.0f;
				
				//shift the UVS so that they match(give or take integer multiples)
				tinfo1->U[iChan][i] += fUOffset;
				tinfo1->V[iChan][i] += fVOffset;
			}
		}
	}

	if(!changed)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"That polygon doesn't touch the first polygon");
		return false;
	}
	else
	{
		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(mesh1,false);
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

	return true;
		
}

void Tools_Mode_Texturing::AdjustUV_XFlip()
{
	if(TexturingMode == TexturingMode_Selection)
	{
		//log all selected polygons for undo
		PrepareUndo();
		PolygonID xPolyID;

		//get the extents of the UVs of the selected polygons
		float afMin[ASURA_MATERIAL_MAXTEXTURES];
		float afMax[ASURA_MATERIAL_MAXTEXTURES];
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			afMin[iChan] = Asura_Maths::LargeFloat;
			afMax[iChan] = -Asura_Maths::LargeFloat;
		}
		
		Tools_Polygon_Selection::StartIteration();
		while(Tools_Polygon_Selection::GetNextSelected(&xPolyID))
		{
			//get the mesh data
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolyID.ModuleIndex);
			Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPolyID.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* pxTexInfo = &pxMesh->TextureDataPtr[xPolyID.PolyIndex];

			//get the min/max U coordinate
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,pxTexInfo->m_iMaterialID)) continue;

				for(int j=0; j<pxPoly->NumberOfVertices(); j++)
				{
   					afMin[iChan] = Asura_Maths::Min(afMin[iChan], pxTexInfo->U[iChan][j]);
   					afMax[iChan] = Asura_Maths::Max(afMax[iChan], pxTexInfo->U[iChan][j]);
				}
			}
		}

		float afCentre[ASURA_MATERIAL_MAXTEXTURES];
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			afCentre[iChan] = (afMin[iChan] + afMax[iChan])/2.0f;
		}
		

		Tools_Polygon_Selection::StartIteration();
		while(Tools_Polygon_Selection::GetNextSelected(&xPolyID))
		{
			//get the mesh data
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolyID.ModuleIndex);
			Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPolyID.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* pxTexInfo = &pxMesh->TextureDataPtr[xPolyID.PolyIndex];

			//adjust the uv coordinates
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,pxTexInfo->m_iMaterialID)) continue;

				for(int j=0; j<pxPoly->NumberOfVertices(); j++)
				{
   					pxTexInfo->U[iChan][j] = 2.0f*afCentre[iChan] - pxTexInfo->U[iChan][j];
				}
			}
	
			//make sure the mesh gets updated in the hardware
			Tools_UpdateTexturingOnSurfaces(pxMesh,false);
		}
		//apply the change to the nodes as well
		for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
		{
			s_pxTextureEdit->Nodes[i].x = (s_pxTextureEdit->UsedArea_UV.left+s_pxTextureEdit->UsedArea_UV.right)-s_pxTextureEdit->Nodes[i].x;
		}
	}
	else
	{
		s_pxTextureEdit->PrepareNodeUndo();
		s_pxTextureEdit->NodesMoved = true;

		//do the flip about the centre of the nodes
		//get min and max x
		float min = s_pxTextureEdit->Nodes[0].x;
		float max = s_pxTextureEdit->Nodes[0].x;
		for(int i=1;i<s_pxTextureEdit->NumNodes;i++)
		{
			if(s_pxTextureEdit->Nodes[i].x<min) min = s_pxTextureEdit->Nodes[i].x;
			if(s_pxTextureEdit->Nodes[i].x>max) max = s_pxTextureEdit->Nodes[i].x;
		}
		float centre = (min+max)/2;

		
		for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
		{
			s_pxTextureEdit->Nodes[i].x = 2*centre - s_pxTextureEdit->Nodes[i].x;
		}
		
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_YFlip()
{
	if(TexturingMode == TexturingMode_Selection)
	{
		//log all selected polygons for undo
		PrepareUndo();
		PolygonID xPolyID;

		//get the extents of the UVs of the selected polygons
		float afMin[ASURA_MATERIAL_MAXTEXTURES];
		float afMax[ASURA_MATERIAL_MAXTEXTURES];
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			afMin[iChan] = Asura_Maths::LargeFloat;
			afMax[iChan] = -Asura_Maths::LargeFloat;
		}
		
		Tools_Polygon_Selection::StartIteration();
		while(Tools_Polygon_Selection::GetNextSelected(&xPolyID))
		{
			//get the mesh data
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolyID.ModuleIndex);
			Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPolyID.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* pxTexInfo = &pxMesh->TextureDataPtr[xPolyID.PolyIndex];

			//get the min/max V coordinate
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,pxTexInfo->m_iMaterialID)) continue;

				for(int j=0; j<pxPoly->NumberOfVertices(); j++)
				{
   					afMin[iChan] = Asura_Maths::Min(afMin[iChan], pxTexInfo->V[iChan][j]);
   					afMax[iChan] = Asura_Maths::Max(afMax[iChan], pxTexInfo->V[iChan][j]);
				}
			}
		}

		float afCentre[ASURA_MATERIAL_MAXTEXTURES];
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			afCentre[iChan] = (afMin[iChan] + afMax[iChan])/2.0f;
		}
		

		Tools_Polygon_Selection::StartIteration();
		while(Tools_Polygon_Selection::GetNextSelected(&xPolyID))
		{
			//get the mesh data
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolyID.ModuleIndex);
			Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPolyID.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* pxTexInfo = &pxMesh->TextureDataPtr[xPolyID.PolyIndex];

			//adjust the uv coordinates
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,pxTexInfo->m_iMaterialID)) continue;

				for(int j=0; j<pxPoly->NumberOfVertices(); j++)
				{
   					pxTexInfo->V[iChan][j] = 2.0f*afCentre[iChan] - pxTexInfo->V[iChan][j];
				}
			}
	
			//make sure the mesh gets updated in the hardware
			Tools_UpdateTexturingOnSurfaces(pxMesh,false);
		}
		//apply the change to the nodes as well
		for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
		{
			s_pxTextureEdit->Nodes[i].x = (s_pxTextureEdit->UsedArea_UV.left+s_pxTextureEdit->UsedArea_UV.right)-s_pxTextureEdit->Nodes[i].x;
		}
	}
	else
	{
		s_pxTextureEdit->PrepareNodeUndo();
		s_pxTextureEdit->NodesMoved = true;

		//do the flip about the centre of the nodes
		//get min and max y
		float min = s_pxTextureEdit->Nodes[0].y;
		float max = s_pxTextureEdit->Nodes[0].y;
		for(int i=1;i<s_pxTextureEdit->NumNodes;i++)
		{
			if(s_pxTextureEdit->Nodes[i].y<min) min = s_pxTextureEdit->Nodes[i].y;
			if(s_pxTextureEdit->Nodes[i].y>max) max = s_pxTextureEdit->Nodes[i].y;
		}
		float centre = (min+max)/2;

		
		for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
		{
			s_pxTextureEdit->Nodes[i].y = 2*centre - s_pxTextureEdit->Nodes[i].y;
		}
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_Rotate()
{
	float fSinAngle = Asura_Maths::Sine(s_fRotationAngle);
	float fCosAngle = Asura_Maths::Cosine(s_fRotationAngle);

	if(TexturingMode == TexturingMode_Selection)
	{
		//log all selecte polygons for undo
		PrepareUndo();
		PolygonID xPolyID;

		//get the extents of the UVs of the selected polygons
		Asura_Bounding_Rectangle axRect[ASURA_MATERIAL_MAXTEXTURES];
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			axRect[iChan].Reset();
		}

		Tools_Polygon_Selection::StartIteration();
		while(Tools_Polygon_Selection::GetNextSelected(&xPolyID))
		{
			//get the mesh data
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolyID.ModuleIndex);
			Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPolyID.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* pxTexInfo = &pxMesh->TextureDataPtr[xPolyID.PolyIndex];

			//get the min/max U coordinate
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,pxTexInfo->m_iMaterialID)) continue;

				for(int j=0; j<pxPoly->NumberOfVertices(); j++)
				{
					axRect[iChan].Set(Asura_Vector_2(pxTexInfo->U[iChan][j], pxTexInfo->V[iChan][j]));
				}
			}
		}

		Asura_Vector_2 axCentre[ASURA_MATERIAL_MAXTEXTURES];
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			axCentre[iChan] = axRect[iChan].CalculateCentre();
		}


		Tools_Polygon_Selection::StartIteration();
		while(Tools_Polygon_Selection::GetNextSelected(&xPolyID))
		{
			//get the mesh data
			Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolyID.ModuleIndex);
			Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPolyID.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* pxTextInfo = &pxMesh->TextureDataPtr[xPolyID.PolyIndex];

			//adjust the uv coordinates
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,pxTextInfo->m_iMaterialID)) continue;

				for(int j=0; j<pxPoly->NumberOfVertices(); j++)
				{
					float fU = pxTextInfo->U[iChan][j] - axCentre[iChan].x;
					float fV = pxTextInfo->V[iChan][j] - axCentre[iChan].y;
										
					pxTextInfo->U[iChan][j] = fCosAngle*fU - fSinAngle*fV + axCentre[iChan].x;
					pxTextInfo->V[iChan][j] = fSinAngle*fU + fCosAngle*fV + axCentre[iChan].y;
				}
			}
	
			//make sure the mesh gets updated in the hardware
			Tools_UpdateTexturingOnSurfaces(pxMesh,false);
		}
	}
	else
	{
		s_pxTextureEdit->PrepareNodeUndo();
		s_pxTextureEdit->NodesMoved = true;
	}
	
	//apply the change to the nodes as well
	float fCentreU = (s_pxTextureEdit->UsedArea_UV.right+s_pxTextureEdit->UsedArea_UV.left)/2.0f;
	float fCentreV = (s_pxTextureEdit->UsedArea_UV.bottom+s_pxTextureEdit->UsedArea_UV.top)/2.0f;

	for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
	{
		float fUOffset = s_pxTextureEdit->Nodes[i].x - fCentreU;
		float fVOffset = s_pxTextureEdit->Nodes[i].y - fCentreV;

		s_pxTextureEdit->Nodes[i].x = fCosAngle*fUOffset - fSinAngle*fVOffset + fCentreU;
		s_pxTextureEdit->Nodes[i].y	= fSinAngle*fUOffset + fCosAngle*fVOffset + fCentreV;
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_Cycle()
{
	
	if(TexturingMode == TexturingMode_Selection)
	{
		//log all selecte polygons for undo
		PrepareUndo();
		PolygonID polygon;
		Tools_Polygon_Selection::StartIteration();

		while(Tools_Polygon_Selection::GetNextSelected(&polygon))
		{
			//get the mesh data
			Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

			//adjust the uv coordinates
			int num_verts = poly->NumberOfVertices();
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;

				float temp_u = tinfo->U[iChan][0];
				float temp_v = tinfo->V[iChan][0];
			
				for(int j=0;j<num_verts-1;j++)
				{
					tinfo->U[iChan][j] = tinfo->U[iChan][j+1];
					tinfo->V[iChan][j] = tinfo->V[iChan][j+1];
				}
				tinfo->U[iChan][num_verts-1] = temp_u;
				tinfo->V[iChan][num_verts-1] = temp_v;
			}
	
			//make sure the mesh gets updated in the hardware
			Tools_UpdateTexturingOnSurfaces(mesh,false);
		}
	}
	else if(MultipleNodeMode)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"UV cycle doesn't work in all nodes shown mode");
		return;
	}
	else
	{
		s_pxTextureEdit->PrepareNodeUndo();
		s_pxTextureEdit->NodesMoved = true;
	}
	
	//apply the change to the nodes as well
	Asura_Vector_2 temp = s_pxTextureEdit->Nodes[0];
	for(int i=0;i<3;i++)
	{
		s_pxTextureEdit->Nodes[i] = s_pxTextureEdit->Nodes[i+1];
	}
	s_pxTextureEdit->Nodes[3] = temp;

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_Reverse()
{
	
	if(TexturingMode == TexturingMode_Selection)
	{
		//log all selected polygons for undo
		PrepareUndo();
		PolygonID polygon;
		Tools_Polygon_Selection::StartIteration();

		while(Tools_Polygon_Selection::GetNextSelected(&polygon))
		{
			//get the mesh data
			Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
			Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
			
			//get the polygon
			Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
			//and get the texture info
			Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];
			
			//adjust the uv coordinates
			int num_verts = poly->NumberOfVertices();
			float temp;
			
			for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
			{
				if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;

				for(int j=0;j<num_verts/2;j++)
				{
					temp = tinfo->U[iChan][j];
					tinfo->U[iChan][j] = tinfo->U[iChan][num_verts-j-1];
					tinfo->U[iChan][num_verts-j-1] = temp;

					temp = tinfo->V[iChan][j];
					tinfo->V[iChan][j] = tinfo->V[iChan][num_verts-j-1];
					tinfo->V[iChan][num_verts-j-1] = temp;
			
				}
			}
			
	
			//make sure the mesh gets updated in the hardware
			Tools_UpdateTexturingOnSurfaces(mesh,false);
		}
	}
	else if(MultipleNodeMode)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"UV reverse doesn't work in all nodes shown mode");
		return;
	}
	else
	{
		s_pxTextureEdit->PrepareNodeUndo();
		s_pxTextureEdit->NodesMoved = true;
	}
	
	//apply the change to the nodes as well
	for(int i=0;i<2;i++)
	{
		Asura_Vector_2 temp = s_pxTextureEdit->Nodes[i];
		s_pxTextureEdit->Nodes[i] = s_pxTextureEdit->Nodes[3-i];
		s_pxTextureEdit->Nodes[3-i] = temp;
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_RotateNodeCentre()
{
	Asura_Assert(TexturingMode == TexturingMode_NodeEdit,"");

	//prepare undo
	s_pxTextureEdit->PrepareNodeUndo();
	s_pxTextureEdit->NodesMoved = true;
	
	
	//get the bounding rectangle for the nodes
	Asura_Bounding_Rectangle bounds;
	bounds.Reset();

	for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
	{
		s_pxTextureEdit->m_pxInitialNodes[i] = s_pxTextureEdit->Nodes[i];
		bounds.Set(s_pxTextureEdit->Nodes[i]);
	}

	//use the centre of the rectange as the centre of the rotation
	Asura_Vector_2 centre;
	centre.x = (bounds.MinX + bounds.MaxX)/2.0f;
	centre.y = (bounds.MinY + bounds.MaxY)/2.0f;


	//rotate by the chosen amount
	s_pxTextureEdit->RotateNodes(centre,s_fRotationAngle,false);

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_Shift()
{
	Asura_Assert(TexturingMode == TexturingMode_Selection,"");

	PolygonID polygon;
	
	//find the width and height of texture on the polygons
	int iWidth = 512;
	int iHeight = 512;

	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		int iTextureID = Asura_Material_System::GetTextureID(tinfo->m_iMaterialID,GetPrimaryEditedChannel());
		
		if(iTextureID == ASURA_TEXTURE_NONE) continue;

		if(Tools_Texture_Management::GetTextureSize(iTextureID,iWidth,iHeight))
		{
			break;
		}

	}
	
	//ask for shift values...
	UVShiftOverlay xOverlay(iWidth,iHeight);
	if(!xOverlay.DoModal()) return;
	

	float fShiftU = xOverlay.GetUShift();
	float fShiftV = xOverlay.GetVShift();

	//log all selected polygons for undo
	PrepareUndo();
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		
		//get the polygon
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		//adjust the uv coordinates
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;
			float fUVScale = GetUVScaleForChannel(iChan);

			for(int j=0; j<poly->NumberOfVertices(); j++)
			{
   				tinfo->U[iChan][j] += fShiftU * fUVScale;
   				tinfo->V[iChan][j] += fShiftV * fUVScale;
			}
		}
	
		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(mesh,false);
	}
	//apply the change to the nodes as well
	for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
	{
		s_pxTextureEdit->Nodes[i].x += fShiftU;
		s_pxTextureEdit->Nodes[i].y += fShiftV;
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();

}

void Tools_Mode_Texturing::AdjustUV_Scale()
{
	Asura_Assert(TexturingMode == TexturingMode_Selection,"");
	
	//ask for shift values...

	UVScaleOverlay xOverlay;
	if(!xOverlay.DoModal()) return;
		
	
	float fScaleU = xOverlay.m_fScaleU;
	float fScaleV = xOverlay.m_fScaleV;

	//log all selected polygons for undo
	PrepareUndo();
	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		//get the mesh data
		Asura_Mesh* mesh = Asura_Module_System::GetMeshForModule(polygon.ModuleIndex);
		Asura_Assert(mesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		
		//get the polygon
		Asura_Mesh_Polygon* poly = &mesh->PolygonDataPtr[polygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &mesh->TextureDataPtr[polygon.PolyIndex];

		//adjust the uv coordinates
		for(int iChan=0;iChan<ASURA_MATERIAL_MAXTEXTURES;iChan++)
		{
			if(!ShouldEditUVsForChannel(iChan,tinfo->m_iMaterialID)) continue;

			for(int j=0; j<poly->NumberOfVertices(); j++)
			{
   				tinfo->U[iChan][j] *= fScaleU;
   				tinfo->V[iChan][j] *= fScaleV;
			}
		}
	
		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(mesh,false);
	}
	//apply the change to the nodes as well
	for(int i=0;i<s_pxTextureEdit->NumNodes;i++)
	{
		s_pxTextureEdit->Nodes[i].x *= fScaleU;
		s_pxTextureEdit->Nodes[i].y *= fScaleV;
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();
}

void Tools_Mode_Texturing::SyncUVChannelsPriToSec()
{
	//ask for the scaling to use
	GetNumber_Overlay xOverlay("Altering secondary UVs. Choose scale value for secondary coordinates relative to primary.",s_fSecondaryUVScale,0.0f,99.0f,"%.0f");
	if(!xOverlay.DoModal()) return;

	float fScale = xOverlay.GetNumber();

	//log for undo
	PrepareUndo();

	//and go over all the selected polygons , adjusting the second coordinates

	PolygonID xPolygon;
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&xPolygon))
	{
		//get the mesh data
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);
		Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		
		//get the polygon
		Asura_Mesh_Polygon* poly = &pxMesh->PolygonDataPtr[xPolygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &pxMesh->TextureDataPtr[xPolygon.PolyIndex];

		//adjust the uv coordinates

		for(int j=0; j<poly->NumberOfVertices(); j++)
		{
   			tinfo->U[1][j] = tinfo->U[0][j] * fScale;
   			tinfo->V[1][j] = tinfo->V[0][j] * fScale;
		}

		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(pxMesh,false);
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();
}

void Tools_Mode_Texturing::SyncUVChannelsSecToPri()
{
	//ask for the scaling to use
	GetNumber_Overlay xOverlay("Altering primary UVs. Choose scale value for secondary coordinates relative to primary.",s_fSecondaryUVScale,0.0f,99.0f,"%.0f");
	if(!xOverlay.DoModal()) return;

	float fScale = xOverlay.GetNumber();
	if(fScale == 0.0f) return;

	//log for undo
	PrepareUndo();

	//and go over all the selected polygons , adjusting the primary coordinates

	PolygonID xPolygon;
	Tools_Polygon_Selection::StartIteration();

	while(Tools_Polygon_Selection::GetNextSelected(&xPolygon))
	{
		//get the mesh data
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);
		Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		
		//get the polygon
		Asura_Mesh_Polygon* poly = &pxMesh->PolygonDataPtr[xPolygon.PolyIndex];
		//and get the texture info
		Asura_Mesh_TextureInfo* tinfo = &pxMesh->TextureDataPtr[xPolygon.PolyIndex];

		//adjust the uv coordinates

		for(int j=0; j<poly->NumberOfVertices(); j++)
		{
   			tinfo->U[0][j] = tinfo->U[1][j] / fScale;
   			tinfo->V[0][j] = tinfo->V[1][j] / fScale;
		}

		//make sure the mesh gets updated in the hardware
		Tools_UpdateTexturingOnSurfaces(pxMesh,false);
	}

	//Keep UV editor overlay in sync
	s_pxUVEdit->UpdateNodesFromSource();
}



void Tools_Mode_Texturing::NodesToCorners()
{
	Asura_Assert(TexturingMode == TexturingMode_NodeEdit,"NodesToCorners should only be used in node edit mode");
	if(s_pxTextureEdit->NumNodes!=4)
	{
		Tools_Timed_Message_System::AddMessage(0xffffff,"This only works when there are 4 nodes");
		return;
	}
	s_pxTextureEdit->NodesToCorners();

}

void Tools_Mode_Texturing::ClampNodes()
{
	s_pxTextureEdit->ClampNodesToUsedArea();
}

void Tools_Mode_Texturing::FitUsedAreaToNodes()
{
	s_pxTextureEdit->FitUsedAreaToNodes();
}

void Tools_Mode_Texturing::EvenlyDistributeUVs()
{
	s_pxTextureEdit->EvenlyDistributeUVs();
}

void Tools_Mode_Texturing::SnapNodesToNearestTexel()
{
	s_pxTextureEdit->SnapNodesToNearestTexel();
}

#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

// Can't change material in this way in the response system

void Tools_Mode_Texturing::NextMaterial()
{

	int iMaterialID = Tools_PickTexture::GetNextMaterial(s_iSelectedMaterialID);
	if((iMaterialID != ASURA_MATERIAL_NONE) && (iMaterialID != s_iSelectedMaterialID))
	{
		SetMaterialID(iMaterialID);
		ApplyTexture();
	}
}
void Tools_Mode_Texturing::PreviousMaterial()
{

	int iMaterialID = Tools_PickTexture::GetPreviousMaterial(s_iSelectedMaterialID);
	if((iMaterialID != ASURA_MATERIAL_NONE) && (iMaterialID != s_iSelectedMaterialID))
	{
		SetMaterialID(iMaterialID);
		ApplyTexture();
	}
}


void Tools_Mode_Texturing::ChooseMaterial()
{

	int iNewMaterialID = Tools_PickTexture::PickMaterial(0,s_iSelectedMaterialID,true);
	if(iNewMaterialID!=ASURA_MATERIAL_NONE)
	{
	
		SetMaterialID(iNewMaterialID);
		ApplyTexture();
	}
}

#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::ChooseMaterialResponse()
{
	
	Asura_Hash_ID uSelectedMaterial = Tools_Overlay_ViewMaterialResponse::PickMaterial("Select Material", s_uSelectedMaterialResponse);

	if(uSelectedMaterial && !Tools_Overlay_ViewMaterialResponse::WasCancelled())
	{
	
		SetMaterialResponse(uSelectedMaterial);
		Tools_Texturing_MaterialResponse::SetSelectedMaterial(uSelectedMaterial);

		ApplyTexture();

	}
}


void Tools_Mode_Texturing::SetMaterialResponse(Asura_Hash_ID uMaterialResponse)
{
	s_uSelectedMaterialResponse = uMaterialResponse;
	if(s_pxTextureEdit)
	{
		s_pxTextureEdit->SetMaterialResponse(0,uMaterialResponse);
	}
	if(s_pxUVEdit)
	{
		s_pxUVEdit->UpdateMaterial();
	}

}

void Tools_Mode_Texturing::SetSecondMaterialResponse(Asura_Hash_ID uMaterialResponse)
{
	s_uSelectedMaterialResponse2 = uMaterialResponse;
	if(s_pxTextureEdit)
	{
		s_pxTextureEdit->SetMaterialResponse(1,uMaterialResponse);
	}
	if(s_pxUVEdit)
	{
		s_pxUVEdit->UpdateMaterial();
	}
}

#endif //defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

void Tools_Mode_Texturing::SetMaterialID(int iMaterialID)
{
	s_iSelectedMaterialID = iMaterialID;
	#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	//In material response builds the UV editor will just have the material response hashes set instead RAR 01/06/2007
	if(s_pxTextureEdit)
	{
		s_pxTextureEdit->SetMaterialID(iMaterialID);
	}
	if(s_pxUVEdit)
	{
		s_pxUVEdit->UpdateMaterial();
	}
	#endif
}

void Tools_Mode_Texturing::SetMaterialFromPolygon(const PolygonID& xPoly)
{
	//get the mesh data
	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);
	Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");

	//get the texture info
	Asura_Mesh_TextureInfo* pxTex  = &pxMesh->TextureDataPtr[xPoly.PolyIndex];

	//set the current material response and/or material ID
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	SetMaterialResponse(pxTex->m_auMaterialResponseHashID[0]);
	#endif

	SetMaterialID(pxTex->m_iMaterialID);
}


void Tools_Mode_Texturing::EditTextureFlags()
{
	Tools_PickTexture::EditTextureFlags(PICKTEXTURE_DEFAULTFLAGEDIT);
}

void Tools_Mode_Texturing::SelectionIsNowEmpty()
{
	//reset the texture wrap mode
	TextureWrap_Stop();

	//clear the base polygon
	BasePolygon.ModuleIndex = -1;
	BasePolygon.PolyIndex = -1;

}

void Tools_Mode_Texturing::PrepareUndo()
{

	PolygonID polygon;
	Tools_Polygon_Selection::StartIteration();

	int iPrevModuleIndex = -1;
	while(Tools_Polygon_Selection::GetNextSelected(&polygon))
	{
		Tools_Texturing_UndoData_TextureInfo::Create(polygon);
		if(iPrevModuleIndex != polygon.ModuleIndex)
		{
			Tools_SaveAsura::SetModuleModified(polygon.ModuleIndex);
			iPrevModuleIndex = polygon.ModuleIndex;
		}
	}
}



bool Tools_Mode_Texturing::IsEditingChannel(int iTextureChannel)
{
	return (s_uEditedChannels & (1<<iTextureChannel)) ? true : false;
}

int Tools_Mode_Texturing::GetPrimaryEditedChannel()
{
	for(int i=0;i<ASURA_MATERIAL_MAXTEXTURES;i++)
	{
		if(s_uEditedChannels & (1<<i)) return i;
	}

	//None selected , this may not be allowed... however assume the first one is the main
	//channel in this case
	return 0;
}

bool Tools_Mode_Texturing::ShouldEditUVsForChannel(int iTextureChannel,int iMaterialID)
{
	if(!IsEditingChannel(iTextureChannel)) return false;
	if(!s_bLimitUVEditingToSelectedTexture) return true;

	//does the material use the selected texture in this channel?
	const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(iMaterialID);
	if(!pxMaterial) return false;

	if(pxMaterial->m_iTextureID[iTextureChannel] == s_iTextureForUVEditing) return true;
	if(pxMaterial->m_iNormalMapID[iTextureChannel] == s_iTextureForUVEditing) return true;
	if(pxMaterial->m_iSpecularMapID[iTextureChannel] == s_iTextureForUVEditing) return true;

	return false;
}

float Tools_Mode_Texturing::GetUVScaleForChannel(int iTextureChannel)
{
	//If editing both primary and seocndary UVs , then the second texture channel (channel 1)
	//should be scaled
	if(iTextureChannel == 1 && GetPrimaryEditedChannel() == 0)
	{
		return s_fSecondaryUVScale;
	}

	return 1.0f;	
}

bool Tools_Mode_Texturing::IsEditingAllChannels()
{
	const u_int uMask = (1<<ASURA_MATERIAL_MAXTEXTURES)-1;

	return (s_uEditedChannels & uMask) == uMask;
}

bool Tools_Mode_Texturing::IsAlphaPainting()
{
	return s_bAlphaPainting && s_bAlphaOverlayOpen && TexturingMode==TexturingMode_Selection;
}

bool Tools_Mode_Texturing::IsUsingAlphaHoming()
{
	return IsAlphaPainting() && s_bAlphaPaint_UseHoming;
}

bool Tools_Mode_Texturing::IsEnabled_AlphaPaintRadius()
{
	return IsAlphaPainting() && !s_bAlphaPaint_AffectNearestVert;
}

void Tools_Mode_Texturing::UpdateAlphaOverlay()
{
	Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(TOOLS_OVERLAY_TEXTURING_ALPHA);
	if(!pxOverlay) return;

	char szText[100];

	#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	if(!s_uSelectedMaterialResponse2)
	{
		strcpy(szText,"Mat : *None*");
	}
	else
	{
		const Tools_MaterialResponse* pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID(s_uSelectedMaterialResponse2);
		if(pxMaterial)
		{
			const char* szName = pxMaterial->GetName().GetString();
			if(szName) szName = Tools_FileSystem::StripPath(szName);

			Asura_PrintfToBfr(szText,"Mat : %s",sizeof(szText),true,szName);
		}
		else
		{
			strcpy(szText,"Mat : *Unknown*");
		}
	}
	pxOverlay->UpdateDialogLabel(DialogID_SecondaryMaterial,szText);

	#else
	int iTexturesUsed;
	float fTextureUse ;
	GetPS2SecondaryTextureUsage(iTexturesUsed,fTextureUse);
	sprintf(szText,"PS2 Usage : %.1f%%",fTextureUse*100.0f);
	pxOverlay->UpdateDialogLabel(DialogID_PS2Usage,szText);
	#endif

}

void Tools_Mode_Texturing::UpdateAlphaPainting()
{
	if (Asura_Interface::UsingMenusOrOverlays() || !IsAlphaPainting())
	{
		s_bAlphaPaint_ButtonClickedForPaint = false;
		return;
	}

	float fDestAlpha = -1.0f;

	//get the alpha value to head for
		
	if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON1) ||
		(Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON1) && s_bAlphaPaint_ButtonClickedForPaint))
	{
		fDestAlpha = s_fAlphaPaint_Alpha[0];
		s_bAlphaPaint_ButtonClickedForPaint = true;
	}
	else if(Asura_UserInput::DebouncedKeyPressed(ASURA_KEYID_MOUSEBUTTON2) ||
		(Asura_UserInput::KeyPressed(ASURA_KEYID_MOUSEBUTTON2) && s_bAlphaPaint_ButtonClickedForPaint))
	{
		fDestAlpha = s_fAlphaPaint_Alpha[1];
		s_bAlphaPaint_ButtonClickedForPaint = true;
	}
	else
	{
		s_bAlphaPaint_ButtonClickedForPaint = false;
	}
	

	#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	int iTextureID = s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot];
	int iBumpTextureID = s_aiSecondaryTextures[1][s_iAlphaPaint_TextureSlot];
	int iSpecularTextureID = s_aiSecondaryTextures[2][s_iAlphaPaint_TextureSlot];
	#endif
	

	if(fDestAlpha >= 0 && s_fAlphaPaint_Radius > 0)
	{
		float fHomingAmount = s_fAlphaPaint_HomingRate * Asura_Timers::GetFrameTime();

		//get the position under the cursor
		int iJunk1,iJunk2;
		Asura_Vector_3 xCursorPos;
		if(Tools_LOS::FindPolyAndPositionUnderCursor(iJunk1,iJunk2,xCursorPos))
		{
			int iNearModule = -1;
			int iNearVert = -1;
			if(s_bAlphaPaint_AffectNearestVert)
			{
				//find the vertex to be affected
				if(!Tools_LOS::FindVertUnderCursor(iNearVert,iNearModule)) return;
			}
			
			
			//search for modules that are in range
			for(int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
			{
				Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
				Asura_Vector_3 xModulePos = Asura_Module_System::GetPositionOfModule(i);
				Asura_Vector_3 xModuleCentre = Asura_Module_System::GetCentreOfModule(i);
				float fModuleRadius = Asura_Module_System::GetRadiusOfModule(i);

				if(s_bAlphaPaint_AffectNearestVert)
				{
					if(i != iNearModule) continue;
				}
				else
				{
					//module in range?
					if((xModuleCentre-xCursorPos).Magnitude() > (fModuleRadius + s_fAlphaPaint_Radius)) continue;
				}


				Asura_Vector_3 xRelPos = xCursorPos - xModulePos;
				bool bModuleChanged = false;
				//check for polygons/vertice in range
				for(int j=0;j<pxMesh->NumberOfPolygons;j++)
				{
					//check for 'selected only'
					if(s_bAlphaPaint_SelectedOnly && !Tools_Polygon_Selection::IsSelected(PolygonID(i,j)))
					{
						continue;
					}
					//even if not limiting to selection , only allow polygons that could be selected
					if(!Tools_Polygon_Selection::IsSelectable(i,j))
					{
						continue;
					}

					Asura_Mesh_TextureInfo* pxTex = &pxMesh->TextureDataPtr[j];

					#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
					//determine what the current and desired secondary materials are
					Asura_Hash_ID uCurrentSecondaryMaterial = pxTex->m_auMaterialResponseHashID[1];
					Asura_Hash_ID uNewSecondaryMaterial =  uCurrentSecondaryMaterial;
					if(s_bAlphaPaint_ApplyTexture)
					{
						if(s_uSelectedMaterialResponse2 == 0)
						{
							uNewSecondaryMaterial = 0;
						}
						else
						{
							if(uCurrentSecondaryMaterial == 0 || s_bAlphaPaint_AllowTextureOverwrite)
							{
								uNewSecondaryMaterial = s_uSelectedMaterialResponse2;
							}
						}
					}
					#else
					int iSecondaryTex = Asura_Material_System::GetTextureID(pxTex->m_iMaterialID,1);
					int iSecondaryBumpTex = Asura_Material_System::GetNormalMapID(pxTex->m_iMaterialID,1);
					int iSecondarySpecularTex = Asura_Material_System::GetSpecularMapID(pxTex->m_iMaterialID,1);
					int iNewSecondaryTex = iSecondaryTex;
					int iNewSecondaryBumpTex = iSecondaryBumpTex;
					int iNewSecondarySpecularTex = iSecondarySpecularTex;

					//determine what the secondary texture would be , to work out
					//if we should allow alpha to be applied
					if(s_bAlphaPaint_ApplyTexture)
					{
						if(s_iAlphaPaint_TextureSlot == TOOLS_MAX_SECONDARY_TEXTURES)
						{
							iNewSecondaryTex = ASURA_TEXTURE_NONE;
						}
						else
						{
							if(iSecondaryTex == ASURA_TEXTURE_NONE || s_bAlphaPaint_AllowTextureOverwrite)
							{
								iNewSecondaryTex = iTextureID;
								iNewSecondaryBumpTex = iBumpTextureID;
								iNewSecondarySpecularTex = iSpecularTextureID;
							}
						}
					}
					#endif 


					//check each vertex used by this polygon
					const Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[j];
					bool bPolyChanged = false;
					for(int k=0;k<pxPoly->NumberOfVertices();k++)
					{
						//get the distance of this vertex from the cursor
						if(s_bAlphaPaint_AffectNearestVert)
						{
							if(pxPoly->VertexID[k] != iNearVert) continue;
						}
						else
						{
							const Asura_Vector_3& xVertPos = pxMesh->VertexDataPtr[pxPoly->VertexID[k]].Position;
							float fDist = (xVertPos-xRelPos).Magnitude();

							if(fDist > s_fAlphaPaint_Radius) continue;
						}

						float& fVertAlpha = pxMesh->TextureDataPtr[j].m_afAlpha[k];

						if(!bPolyChanged)
						{
							//log for undo
							Tools_Texturing_UndoData_TextureInfo::Create(PolygonID(i,j));

							bModuleChanged = true;
							bPolyChanged = true;
						}
						
						#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
						if(uNewSecondaryMaterial == 0 && uCurrentSecondaryMaterial != 0)
						#else
						if(iNewSecondaryTex == ASURA_TEXTURE_NONE && iSecondaryTex != ASURA_TEXTURE_NONE)
						#endif
						{
							//Reset the alpha value if moving from having a secondary texture to not having one
							fVertAlpha = 1.0f;
						}
						else if(s_bAlphaPaint_UseHoming)
						{
							float fHomingScale = 1.0f;
							//float fHomingScale = 0.3f+((s_fAlphaPaint_Radius-fDist)/s_fAlphaPaint_Radius);
							//if(fHomingScale>1.0f) fHomingScale = 1.0f;
							
							//home towards desired value
							if(fVertAlpha > fDestAlpha)
							{
								fVertAlpha -= fHomingAmount*fHomingScale;
								if(fVertAlpha < fDestAlpha) fVertAlpha = fDestAlpha;
							}
							else
							{
								fVertAlpha += fHomingAmount*fHomingScale;
								if(fVertAlpha > fDestAlpha) fVertAlpha = fDestAlpha;
							}
						}
						else
						{
							fVertAlpha = fDestAlpha;
						}
					}

					#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
					if(bPolyChanged)
					{
						//apply the current material to this polygon as well
						pxTex->m_auMaterialResponseHashID[1] = uNewSecondaryMaterial;
					}
					#else
					if(bPolyChanged && (iSecondaryTex != iNewSecondaryTex || iSecondaryBumpTex != iNewSecondaryBumpTex || iSecondarySpecularTex != iNewSecondarySpecularTex))
					{
						//apply the current texture to this polygon as well
 						pxTex->m_iMaterialID = GetMaterialWithTexture(pxTex->m_iMaterialID,iNewSecondaryTex,iNewSecondaryBumpTex,iNewSecondarySpecularTex,1);
					}
					#endif
					if(bPolyChanged && s_bAlphaPaint_SetUVs)
					{
						//set secondary coordinates from first
						for(int k=0;k<pxPoly->NumberOfVertices();k++)
						{
							pxTex->U[1][k] = pxTex->U[0][k] * s_fSecondaryUVScale;							
							pxTex->V[1][k] = pxTex->V[0][k] * s_fSecondaryUVScale;							
						}
					}
				}
				if(bModuleChanged)
				{
					//this module has been modified
					Tools_UpdateTexturingOnSurfaces(pxMesh,false);
					Tools_SaveAsura::SetMeshModified(pxMesh);
				}
			}
		}
	}
}

void Tools_Mode_Texturing::RenderAlphaPainting()
{
	if(s_bAlphaPaint_ShowSphere && !Asura_Interface::UsingMenusOrOverlays())
	{
		if(s_bAlphaPaint_AffectNearestVert)
		{
			//highlight the nearest vertex
			int iVert,iModule;
			if(Tools_LOS::FindVertUnderCursor(iVert,iModule))
			{
				float fAlpha = Asura_Maths::Sine( Asura_Timers::GetLocalUnscaledTime() * 2.0f );
				fAlpha = 255.0f * fAlpha * fAlpha;

				unsigned int uColour = 0x00ffffff + (Asura_Maths::FloatToInt( fAlpha )<<24);

				Tools_Render::RenderVertexOutline(iModule,iVert,uColour);
			}
		}
		else
		{
			//draw a sphere where the cursor is
			int iJunk1,iJunk2;
			Asura_Vector_3 xPosition;
			if(Tools_LOS::FindPolyAndPositionUnderCursor(iJunk1,iJunk2,xPosition))
			{
				extern void Render_Sphere( const Asura_Vector_3& xCentre, float fRadius, int iColour );
				Render_Sphere(xPosition,s_fAlphaPaint_Radius,0x7f0000ff);
			}
		}
	}
}

void Tools_Mode_Texturing::SetAlphaOnSelected()
{
	//ask for alpha value
	GetNumber_Overlay xOverlay("Apply what alpha value to selected polygons?",s_fAlphaPaint_Alpha[0],0,1.0f,"%.3f");
	if(!xOverlay.DoModal()) return;

	float fAlpha = xOverlay.GetNumber();

	//log for undo
	PrepareUndo();
	
	//apply to all selected polygons
	PolygonID xPoly;
	Tools_Polygon_Selection::StartIteration();

	bool bChanged = false;
	
	while(Tools_Polygon_Selection::GetNextSelected(&xPoly))
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);
		Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[xPoly.PolyIndex];
		Asura_Mesh_TextureInfo* pxTex = &pxMesh->TextureDataPtr[xPoly.PolyIndex];

		for(int i=0;i<pxPoly->NumberOfVertices();i++)
		{
			pxTex->m_afAlpha[i] = fAlpha;
		}

		bChanged = true;
		Tools_UpdateTexturingOnSurfaces(pxMesh,false);
	}

}

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::SelectSecondMaterial()
{
	Tools_Overlay_PickMaterialResponse xOverlay("Select secondary material");
	if(xOverlay.DoModal(s_uSelectedMaterialResponse2))
	{
		SetSecondMaterialResponse(xOverlay.GetSelection());
	}
}
#endif


#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::SetSecondMaterialOnSelected()
{
	SetSecondMaterialOnSelected(s_uSelectedMaterialResponse2);
}
#else
void Tools_Mode_Texturing::SetSecondTextureOnSelected()
{
	PolygonID xPoly;
	Tools_Polygon_Selection::StartIteration();

	int iTextureID = s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot];

	if(iTextureID == ASURA_TEXTURE_NONE && s_iAlphaPaint_TextureSlot !=	TOOLS_MAX_SECONDARY_TEXTURES)
	{
		ShowMessageOverlay("A non-empty secondary texture slot must be chosen first");
		return;
	}

	
	if(iTextureID == ASURA_TEXTURE_NONE)
	{
		if(!ShowYesNoOverlay("Remove the secondary texture from all selected polygons?"))
		{
			return;
		}
	}
	else
	{
		const Asura_String* pxName = Asura_Texture_Management::GetTextureName(iTextureID);
		if(!pxName || pxName->IsBlank())
		{
			Asura_Assert(0,"Failed to get texture name");
			return;
		}
		const char* szName = Tools_FileSystem::StripPath(pxName->GetString());
	
		if(!ShowYesNoOverlay("Set %s as the secondary texture on all selected polygons?",szName))
		{
			return;
		}
	}

	SetSecondTextureOnSelected(s_iAlphaPaint_TextureSlot);

}
#endif //ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::RemoveSecondMaterialFromSelected()
{
	if(!ShowYesNoOverlay("Remove the secondary material from all selected polygons?"))
	{
		return;
	}
	SetSecondMaterialOnSelected(0);
	
}
#else
void Tools_Mode_Texturing::RemoveSecondTextureFromSelected()
{
	if(!ShowYesNoOverlay("Remove the secondary texture from all selected polygons?"))
	{
		return;
	}
	SetSecondTextureOnSelected(TOOLS_MAX_SECONDARY_TEXTURES);
}
#endif

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::SetSecondMaterialOnSelected(Asura_Hash_ID uMaterial)
{


	PrepareUndo();


	bool bChanged = false;

	PolygonID xPoly;
	Tools_Polygon_Selection::StartIteration();

	
	while(Tools_Polygon_Selection::GetNextSelected(&xPoly))
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);
		Asura_Mesh_TextureInfo* pxTex = &pxMesh->TextureDataPtr[xPoly.PolyIndex];

		if(pxTex->m_auMaterialResponseHashID[0])
		{
	
			
			if(pxTex->m_auMaterialResponseHashID[1] && !uMaterial)
			{
				//Removing second material ,so reset alpha
				for(int i=0;i<4;i++)
				{
					pxTex->m_afAlpha[i] = 1.0f;
				}
				
			}
			
			pxTex->m_auMaterialResponseHashID[1] = uMaterial;
			bChanged = true;
			Tools_UpdateTexturingOnSurfaces(pxMesh);
			
		}
 	}

}
#else //else not material response
void Tools_Mode_Texturing::SetSecondTextureOnSelected(int iSlot)
{

	int iTextureID = s_aiSecondaryTextures[0][iSlot];
	int iBumpTextureID = s_aiSecondaryTextures[1][iSlot];
	int iSpecularTextureID = s_aiSecondaryTextures[2][iSlot];

	if(iTextureID == ASURA_TEXTURE_NONE && iSlot != TOOLS_MAX_SECONDARY_TEXTURES)
	{
		return;
	}


	PrepareUndo();


	bool bChanged = false;

	PolygonID xPoly;
	Tools_Polygon_Selection::StartIteration();

	
	while(Tools_Polygon_Selection::GetNextSelected(&xPoly))
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPoly.ModuleIndex);
		Asura_Mesh_TextureInfo* pxTex = &pxMesh->TextureDataPtr[xPoly.PolyIndex];

		if(pxTex->m_iMaterialID != ASURA_MATERIAL_NONE)
		{
	
			int iSecondaryTextureID = Asura_Material_System::GetTextureID(pxTex->m_iMaterialID,1);
			
			if(iSlot == TOOLS_MAX_SECONDARY_TEXTURES)
			{
				//removing second texture
				pxTex->m_iMaterialID = GetMaterialWithTexture(pxTex->m_iMaterialID,ASURA_TEXTURE_NONE,ASURA_TEXTURE_NONE,ASURA_TEXTURE_NONE,1);
				if(iSecondaryTextureID != ASURA_TEXTURE_NONE)
				{
					//need to reset alpha as well
					for(int i=0;i<4;i++)
					{
						pxTex->m_afAlpha[i] = 1.0f;
					}
				}

				bChanged = true;
				Tools_UpdateTexturingOnSurfaces(pxMesh);
			}
			else if(iTextureID != ASURA_TEXTURE_NONE)
			{
				pxTex->m_iMaterialID = GetMaterialWithTexture(pxTex->m_iMaterialID,iTextureID,iBumpTextureID,iSpecularTextureID,1);
				bChanged = true;
				Tools_UpdateTexturingOnSurfaces(pxMesh);
			}
		}
 	}
}
#endif //end of not material response


#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::InitialiseSecondaryTextureArray()
{
	//initialise array
	for(int i=0;i<=TOOLS_MAX_SECONDARY_TEXTURES;i++)
	{
		for(int j=0;j<3;j++)
		{
			s_aiSecondaryTextures[j][i] = ASURA_TEXTURE_NONE;
		}
	}

	//search through all materials used on the environment ,and look for those
	//that use a second texture
	List<int> xUnplacedTextures;
	List<int> xUnplacedBumpTextures;
	List<int> xUnplacedSpecularTextures;

	for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
	{
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
		Asura_Assert(pxMaterial,"");

		//got a second texture?
		int iTextureID = pxMaterial->m_iTextureID[1];
		int iBumpTextureID = pxMaterial->m_iNormalMapID[1];
		int iSpecularTextureID = pxMaterial->m_iSpecularMapID[1];
		if(iTextureID == ASURA_TEXTURE_NONE) continue;

		//see if this texture already has a slot
		bool bFound = false;
		for(int i=0;i<TOOLS_MAX_SECONDARY_TEXTURES;i++)
		{
			if(s_aiSecondaryTextures[0][i] == iTextureID)
			{
				bFound = true;

				if(iBumpTextureID != ASURA_TEXTURE_NONE)
				{
					s_aiSecondaryTextures[1][i] = iBumpTextureID;
				}

				if(iSpecularTextureID != ASURA_TEXTURE_NONE)
				{
					s_aiSecondaryTextures[2][i] = iSpecularTextureID;
				}
				break;
			}
		}
		if(bFound) continue;

		//See if the material has a name to allow us to deduce its slot
		const char* pszName = pxMaterial->m_xMaterialName.GetString();
		if(pszName)
		{
			const char* pcPos = strstr(pszName,"[Tx");
			if(pcPos)
			{
				int iSlot = 0;
				if(sscanf(pcPos,"[Tx%d]",&iSlot))
				{
					if(iSlot>=0 && iSlot < TOOLS_MAX_SECONDARY_TEXTURES && s_aiSecondaryTextures[0][iSlot] == ASURA_TEXTURE_NONE)
					{
						//okay , found the slot for this texture
						s_aiSecondaryTextures[0][iSlot] = iTextureID;
						s_aiSecondaryTextures[1][iSlot] = iBumpTextureID;
						s_aiSecondaryTextures[2][iSlot] = iSpecularTextureID;
						if(xUnplacedTextures.contains(iTextureID))
						{
							LIF<int> xTextLIF(&xUnplacedTextures);
							LIF<int> xBumpLIF(&xUnplacedBumpTextures);
							LIF<int> xSpecularLIF(&xUnplacedSpecularTextures);
							for(;!xTextLIF.done();xTextLIF.next(),xBumpLIF.next(),xSpecularLIF.next())
							{
								if(xTextLIF() == iTextureID)
								{
									xTextLIF.delete_current();
									xBumpLIF.delete_current();
									xSpecularLIF.delete_current();
									break;
								}
							}
						}
						continue;
					}	
				}
			}
		}
		//failed to determine slot for texture , just add to a list to put in whatever slots are leftover
		//at the end
		if(!xUnplacedTextures.contains(iTextureID))
		{
			xUnplacedTextures.add_entry(iTextureID);
			xUnplacedBumpTextures.add_entry(iBumpTextureID);
			xUnplacedSpecularTextures.add_entry(iSpecularTextureID);
		}
	}

	//Finally , if there were any secondary textures that haven't been given a slot yet , 
	//see if there is space available
	Asura_Assert(xUnplacedTextures.size() == xUnplacedBumpTextures.size(),"");
	Asura_Assert(xUnplacedTextures.size() == xUnplacedSpecularTextures.size(),"");
	for(int i=0;i<TOOLS_MAX_SECONDARY_TEXTURES && xUnplacedTextures.size();i++)
	{
		if(s_aiSecondaryTextures[0][i] == ASURA_TEXTURE_NONE)
		{
			s_aiSecondaryTextures[0][i] = xUnplacedTextures.first_entry();
			s_aiSecondaryTextures[1][i] = xUnplacedBumpTextures.first_entry();
			s_aiSecondaryTextures[2][i] = xUnplacedSpecularTextures.first_entry();
			xUnplacedTextures.delete_first_entry();
			xUnplacedBumpTextures.delete_first_entry();
			xUnplacedSpecularTextures.delete_first_entry();
		}
	}
#if 1
	if(xUnplacedTextures.size())
	{
		//warn about excess secondary textures
		char szMessage[1000];
		strcpy(szMessage,"Warning... more than 4 secondary textures found in use.\nThese textures couldn't be assigned a slot :\n");
		char* pcPos = szMessage + strlen(szMessage);
		
		while(xUnplacedTextures.size())
		{
			const char* pszName = Asura_Texture_Management::GetTextureName(xUnplacedTextures.first_entry())->GetString();
			xUnplacedTextures.delete_first_entry();
			pszName = Tools_FileSystem::StripPath(pszName);

			if((pcPos - szMessage) + strlen(pszName) + 10 > 1000)
			{
				break;
			}
			sprintf(pcPos,"%s\n",pszName);
			pcPos += strlen(pcPos);
		}

		ShowMessageOverlay(szMessage);
	}
#endif
}


void Tools_Mode_Texturing::BuildSecondaryTextureNames()
{
	DeleteSecondaryTextureNames();

	for(int i=0;i<TOOLS_MAX_SECONDARY_TEXTURES;i++)
	{
		for(int j=0;j<3;j++)
		{
			const char* pszName;
			if(s_aiSecondaryTextures[j][i] == ASURA_TEXTURE_NONE)
			{
				pszName = j ? "*None*" : "*Empty Slot*";
			}
			else
			{
				pszName = Asura_Texture_Management::GetTextureName(s_aiSecondaryTextures[j][i])->GetString();
				pszName = Tools_FileSystem::StripPath(pszName);
			}
			int iLength = strlen(pszName)+10;
			s_apszSecondaryTextureNames[j][i] = new char[iLength];
			Asura_PrintfToBfr(s_apszSecondaryTextureNames[j][i],"%-3d %s",iLength,true,i,pszName);
			
		}

	}
	s_apszSecondaryTextureNames[0][TOOLS_MAX_SECONDARY_TEXTURES] = "*None*";
	s_apszSecondaryTextureNames[1][TOOLS_MAX_SECONDARY_TEXTURES] = "*None*";
	s_apszSecondaryTextureNames[2][TOOLS_MAX_SECONDARY_TEXTURES] = "*None*";
}


void Tools_Mode_Texturing::DeleteSecondaryTextureNames()
{
	for(int i=0;i<TOOLS_MAX_SECONDARY_TEXTURES;i++)
	{
		for(int j=0;j<3;j++)
		{
			delete [] s_apszSecondaryTextureNames[j][i];
			s_apszSecondaryTextureNames[j][i] = 0;
		}
	}
	//don't need to delete the 'no texture' slot , since it isn't allocated as such
}

bool Tools_Mode_Texturing::CanEditSecondaryTexture()
{
	return(s_iAlphaPaint_TextureSlot>=0 && s_iAlphaPaint_TextureSlot<TOOLS_MAX_SECONDARY_TEXTURES);
}

void Tools_Mode_Texturing::EditSecondaryTexture()
{
	if(!CanEditSecondaryTexture()) return;

	//ask for texture for this slot
	int iCurrentTextureID = s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot];
	int iTextureID = Tools_PickTexture::PickTexture("Pick texture for secondary texture slot",iCurrentTextureID,PICKTEXTURE_SECONDARY);

	//prevent click used to choose texture from being used to alpha paint
	s_bAlphaPaint_ButtonClickedForPaint = false;

	if(iTextureID == iCurrentTextureID) return;
	if(iTextureID == ASURA_TEXTURE_NONE) return;

	//make sure this texture isn't in another slot
	for(int i=0;i<TOOLS_MAX_SECONDARY_TEXTURES;i++)
	{
		if(s_aiSecondaryTextures[0][i] == iTextureID)
		{
			ShowMessageOverlay("That texture is already used by another secondary texture slot");
			return;
		}
	}

	//search for all materials that were using the old secondary texture , and replace
	bool bChanged = false;
	if(iCurrentTextureID != ASURA_TEXTURE_NONE)
	{
		for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
		{
			const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
			Asura_Assert(pxMaterial,"");

			if(pxMaterial->m_iTextureID[1] == iCurrentTextureID)
			{
				Asura_Material xNewMat = *pxMaterial;
				xNewMat.m_iTextureID[1] = iTextureID;
				Asura_Material_System::SetMaterial(xMatLIF(),xNewMat);
				bChanged = true;
			}
		}
	}

	//update our secondary texture list
	s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot] = iTextureID;
	//and the texture names
	BuildSecondaryTextureNames();

	if(bChanged)
	{
		Tools_SaveAsura::SetMainFileModified();
	}

}

void Tools_Mode_Texturing::EditSecondaryBumpTexture()
{
	if(!CanEditSecondaryTexture()) return;

	//ask for texture for this slot
	int iCurrentTextureID = s_aiSecondaryTextures[1][s_iAlphaPaint_TextureSlot];
	int iTextureID = Tools_PickTexture::PickTexture("Pick texture for secondary normal map slot",iCurrentTextureID,PICKTEXTURE_BUMP);

	//prevent click used to choose texture from being used to alpha paint
	s_bAlphaPaint_ButtonClickedForPaint = false;

	if(iTextureID == iCurrentTextureID) return;
	if(iTextureID == ASURA_TEXTURE_NONE) return;

	//search for all materials that were using the old secondary texture , and replace
	bool bChanged = false;
	
	
	for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
	{
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
		Asura_Assert(pxMaterial,"");

		if(pxMaterial->m_iNormalMapID[1] == iCurrentTextureID && pxMaterial->m_iTextureID[1] == s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot])
		{
			Asura_Material xNewMat = *pxMaterial;
			xNewMat.m_iNormalMapID[1] = iTextureID;
			Asura_Material_System::SetMaterial(xMatLIF(),xNewMat);

			bChanged = true;
		}
	}
	

	//update our secondary texture list
	s_aiSecondaryTextures[1][s_iAlphaPaint_TextureSlot] = iTextureID;
	//and the texture names
	BuildSecondaryTextureNames();

	if(bChanged)
	{
		Tools_SaveAsura::SetMainFileModified();
	}

}
void Tools_Mode_Texturing::ClearSecondaryBumpTexture()
{
	if(!CanEditSecondaryTexture()) return;

	//get the current texture
	int iCurrentTextureID = s_aiSecondaryTextures[1][s_iAlphaPaint_TextureSlot];
	if(iCurrentTextureID == ASURA_TEXTURE_NONE) return;


	//search for all materials that were using the old secondary texture , and replace
	bool bChanged = false;
	
	
	for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
	{
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
		Asura_Assert(pxMaterial,"");

		if(pxMaterial->m_iNormalMapID[1] == iCurrentTextureID && pxMaterial->m_iTextureID[1] == s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot])
		{
			Asura_Material xNewMat = *pxMaterial;
			xNewMat.m_iNormalMapID[1] = ASURA_TEXTURE_NONE;
			Asura_Material_System::SetMaterial(xMatLIF(),xNewMat);

			bChanged = true;
		}
	}
	

	//update our secondary texture list
	s_aiSecondaryTextures[1][s_iAlphaPaint_TextureSlot] = ASURA_TEXTURE_NONE;
	//and the texture names
	BuildSecondaryTextureNames();

	if(bChanged)
	{
		Tools_SaveAsura::SetMainFileModified();
	}

}

void Tools_Mode_Texturing::EditSecondarySpecularTexture()
{
	if(!CanEditSecondaryTexture()) return;

	//ask for texture for this slot
	int iCurrentTextureID = s_aiSecondaryTextures[2][s_iAlphaPaint_TextureSlot];
	int iTextureID = Tools_PickTexture::PickTexture("Pick texture for secondary specular texture slot",iCurrentTextureID);

	//prevent click used to choose texture from being used to alpha paint
	s_bAlphaPaint_ButtonClickedForPaint = false;

	if(iTextureID == iCurrentTextureID) return;
	if(iTextureID == ASURA_TEXTURE_NONE) return;

	//search for all materials that were using the old secondary texture , and replace
	bool bChanged = false;
		
	for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
	{
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
		Asura_Assert(pxMaterial,"");

		if(pxMaterial->m_iSpecularMapID[1] == iCurrentTextureID && pxMaterial->m_iTextureID[1] == s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot])
		{
			Asura_Material xNewMat = *pxMaterial;
			xNewMat.m_iSpecularMapID[1] = iTextureID;
			Asura_Material_System::SetMaterial(xMatLIF(),xNewMat);
			bChanged = true;
		}
	}
	

	//update our secondary texture list
	s_aiSecondaryTextures[2][s_iAlphaPaint_TextureSlot] = iTextureID;
	//and the texture names
	BuildSecondaryTextureNames();

	if(bChanged)
	{
		Tools_SaveAsura::SetMainFileModified();
	}

}
void Tools_Mode_Texturing::ClearSecondarySpecularTexture()
{
	if(!CanEditSecondaryTexture()) return;

	//get the current texture
	int iCurrentTextureID = s_aiSecondaryTextures[2][s_iAlphaPaint_TextureSlot];
	if(iCurrentTextureID == ASURA_TEXTURE_NONE) return;


	//search for all materials that were using the old secondary texture , and replace
	bool bChanged = false;
	
	
	for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
	{
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
		Asura_Assert(pxMaterial,"");

		if(pxMaterial->m_iSpecularMapID[1] == iCurrentTextureID && pxMaterial->m_iTextureID[1] == s_aiSecondaryTextures[0][s_iAlphaPaint_TextureSlot])
		{
			Asura_Material xNewMat = *pxMaterial;
			xNewMat.m_iSpecularMapID[1] = ASURA_TEXTURE_NONE;
			Asura_Material_System::SetMaterial(xMatLIF(),xNewMat);
			bChanged = true;
		}
	}
	

	//update our secondary texture list
	s_aiSecondaryTextures[2][s_iAlphaPaint_TextureSlot] = ASURA_TEXTURE_NONE;
	//and the texture names
	BuildSecondaryTextureNames();

	if(bChanged)
	{
		Tools_SaveAsura::SetMainFileModified();
	}

}
#endif // ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM

void Tools_Mode_Texturing::GetPS2SecondaryTextureUsage(int& iNumTextures,float& fTextureUse,List<int>* pxTextureList)
{
	#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	iNumTextures = 0;
	fTextureUse = 0.0f;
	#else
	//scan environment for all secondary textures used
	List<int> xTextureList;

	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();

	for(int i=0;i<iNumberOfModules;i++)
	{
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		Asura_Assert(pxMesh,"");

		for(int j=0;j<pxMesh->NumberOfPolygons;j++)
		{
			int iPolyFlags = pxMesh->PolygonFlagDataPtr ? pxMesh->PolygonFlagDataPtr[j].Flags : 0;
			//ignore polygons that won't be exported
			if(iPolyFlags & (AsuraPolygonFlag_DontExport|AsuraPolygonFlag_DontRender))
			{
				continue;
			}
			
			
			int iMaterialID = pxMesh->TextureDataPtr[j].m_iMaterialID;
			if(iMaterialID == ASURA_MATERIAL_NONE) continue;

			int iTextureID = Asura_Material_System::GetTextureID(iMaterialID,1);
			if(iTextureID == ASURA_TEXTURE_NONE) continue;
			
			if(!xTextureList.contains(iTextureID))
			{
				xTextureList.add_entry(iTextureID);
			}
		}
	}

	iNumTextures = xTextureList.size();
	u_int uTextureSize = 0;
	
	for(LIF<int> xTextureLIF(&xTextureList);!xTextureLIF.done();xTextureLIF.next())
	{
		int iTextureID = xTextureLIF();
		int iWidth,iHeight;

		if(!Tools_Texture_Management::GetTextureSize(iTextureID,iWidth,iHeight))
		{
			continue;
		}

		//check texture flags to determine how much it will be reduced
		int iFlags = Asura_Texture_Management::GetTextureFlags(iTextureID);
		if(!(iFlags & AsuraTextureFlag_PS2DontHalveSize))
		{
			iWidth >>= 1 ;
			iHeight >>= 1;
		}
		if(iFlags & AsuraTextureFlag_ReduceExtraForConsoles)
		{
			iWidth >>= 1;
			iHeight >>= 1;
		}

		if(!(iFlags & AsuraTextureFlag_PS2Allow512))
		{
			while(iWidth>256 || iHeight>256)
			{
				iWidth >>= 1;
				iHeight >>= 1;
			}
		}

		uTextureSize += (u_int) (iWidth*iHeight);
	}

	if(pxTextureList)
	{
		*pxTextureList = xTextureList;
	}

	#if TOOLS_MAX_PS2_SECONDARY_TEXTURE_SIZE
	fTextureUse = (float)uTextureSize / (float)(TOOLS_MAX_PS2_SECONDARY_TEXTURE_SIZE);
	#else
	fTextureUse = 0;
	#endif
	
	#endif
}


void Tools_Mode_Texturing::FixSecondaryTextureAnomalies()
{
	int iNumberOfModules = Asura_Module_System::GetNumberOfModules();
	for(int i=0;i<iNumberOfModules;i++)
	{
		bool bChanged = false;
		
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(i);
		Asura_Assert(pxMesh,"");

		for(int j=0;j<pxMesh->NumberOfPolygons;j++)
		{
			Asura_Mesh_TextureInfo* pxTex = &pxMesh->TextureDataPtr[j];
			const Asura_Mesh_Polygon* pxPoly = &pxMesh->PolygonDataPtr[j];
			
			int iMaterialID = pxTex->m_iMaterialID;
			

			int iPrimaryTextureID = ASURA_TEXTURE_NONE;
			int iSecondaryTextureID = ASURA_TEXTURE_NONE;
			if(iMaterialID != ASURA_MATERIAL_NONE)
			{
				iPrimaryTextureID = Asura_Material_System::GetTextureID(iMaterialID,0);
				iSecondaryTextureID = Asura_Material_System::GetTextureID(iMaterialID,1);
			}
		
			if(iSecondaryTextureID != ASURA_TEXTURE_NONE && iPrimaryTextureID != ASURA_TEXTURE_NONE)
			{
				bool bRemoveSecondTexture = false;

				#if 0
				//Automatic removal of second texture from merged objects now disabled. RAR 01/06/2006
				if(pxMesh->IsMergedObjectPolygon(j))
				{
					//polygons merged in from entity shouldn't have a second texture
					bRemoveSecondTexture = true;
				}
				else
				#endif
				{
					//nor polygons that don't have any non-one alpha values
					bool bHasAlpha = false;
					for(int k=0;k<pxPoly->NumberOfVertices();k++)
					{
						if(pxTex->m_afAlpha[k] != 1.0f)
						{
							bHasAlpha = true;
						}
					}
					if(!bHasAlpha) bRemoveSecondTexture = true;

#ifdef ASURA_USE_DAMAGE_MAPS
					// If we're using damage maps, we need to keep second textures on skins, even
					// if there's no vertex alphas present, so override the above test.
					if (Tools_SaveAsura::GetFileType() == AsuraFile_Skin)
					{					
						bRemoveSecondTexture = false;
					}
#endif
				}

				
				if(bRemoveSecondTexture)
				{
					pxTex->m_iMaterialID = GetMaterialWithTexture(pxTex->m_iMaterialID,ASURA_TEXTURE_NONE,ASURA_TEXTURE_NONE,ASURA_TEXTURE_NONE,1);
					iSecondaryTextureID = ASURA_TEXTURE_NONE;
					bChanged = true;

					for(int k=0;k<pxPoly->NumberOfVertices();k++)
					{
						if(pxTex->m_afAlpha[k] != 1.0f)
						{
							pxTex->m_afAlpha[k] = 1.0f;
							bChanged = true;
						}
					}
				}
			}


		}
		if(bChanged)
		{
			//this module has been modified
			Tools_UpdateTexturingOnSurfaces(pxMesh);
			Tools_SaveAsura::SetMeshModified(pxMesh);
		}

	}
	
}


int Tools_Mode_Texturing::GetMaterialWithTexture(int iBaseMaterialID,int iTextureID,int iBumpTextureID,int iSpecularTextureID,int iChannel)
{
	if(iBaseMaterialID == ASURA_MATERIAL_NONE)
	{
		//no original material... , can only generate a material if the texture is for the first slot then
		if(iTextureID != ASURA_TEXTURE_NONE && iChannel == 0)
		{
			return Asura_Material_System::GetMaterialIDForTexture(iTextureID,true);
		}
		else
		{
			return ASURA_MATERIAL_NONE;
		}
	}

	//get the existing material
	const Asura_Material* pxBaseMaterial = Asura_Material_System::GetMaterial(iBaseMaterialID);
	Asura_Assert(pxBaseMaterial,"");

	//does the current material meet our requirements?
	if(	pxBaseMaterial->m_iTextureID[iChannel] == iTextureID
		&& pxBaseMaterial->m_iNormalMapID[iChannel] == iBumpTextureID
		&& pxBaseMaterial->m_iSpecularMapID[iChannel] == iSpecularTextureID) return iBaseMaterialID;

	Asura_Material xDesiredMat = *pxBaseMaterial;
	xDesiredMat.m_iTextureID[iChannel] = iTextureID;
	xDesiredMat.m_iNormalMapID[iChannel] = iBumpTextureID;
	xDesiredMat.m_iSpecularMapID[iChannel] = iSpecularTextureID;


	//search for an existing material that will do
	for(CLIF<int> xMatLIF(Tools_PickTexture::GetMaterialList()) ; !xMatLIF.done();xMatLIF.next())
	{
		const Asura_Material* pxMaterial = Asura_Material_System::GetMaterial(xMatLIF());
		Asura_Assert(pxMaterial,"");

		if(Asura_Material::MaterialsAreEquivalent(&xDesiredMat,pxMaterial,false))
		{
			//this material works
			return xMatLIF();
		}
	}

	//okay... need to create a new material then
	int iMaterialID = Asura_Material_System::CreateMaterial();
	if(iMaterialID == ASURA_MATERIAL_NONE) return iBaseMaterialID;


	#ifndef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	if(iChannel == 1)
	{
		//alter the name of the material to reflect the texture slot used

		//first get the name without the [Tx] tacked on the end
		char szName[200] = "";
		if(xDesiredMat.m_xMaterialName.GetString())
		{
			strcpy(szName,xDesiredMat.m_xMaterialName.GetString());
			char* pcEnd = strstr(szName,"[Tx");
			if(pcEnd) *pcEnd = 0;
		}

		if(iTextureID != ASURA_TEXTURE_NONE)
		{
			//determine the texture slot
			for(int i=0;i<TOOLS_MAX_SECONDARY_TEXTURES;i++)
			{
				if(iTextureID == s_aiSecondaryTextures[0][i])
				{
					char* pcEnd = &szName[strlen(szName)];
					sprintf(pcEnd,"[Tx%d]",i);
					break;
				}
			}
		}
		//update the name
		xDesiredMat.m_xMaterialName = szName;
	}
	#endif

	//initialise new material
	Asura_Material_System::SetMaterial(iMaterialID,xDesiredMat);

	Tools_PickTexture::AddToMaterialList(iMaterialID);

	return iMaterialID;	
}

void Tools_Mode_Texturing::EditMaterialOnPolygon()
{
	Tools_Object_System::EnableAllControls( false );
	SetPromptedSelectionMode(PSM_EditMaterialOnPolygon);
}
void Tools_Mode_Texturing::EditMaterialOnPolygon_PolygonSelected(PolygonID xPolygon)
{
	SetPromptedSelectionMode(PSM_None);

	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

		#if defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
			if( Tools_SaveAsura::GetFileType() == AsuraFile_Skin )
		#endif

		{
			const Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule( xPolygon.ModuleIndex );

			Asura_Hash_ID uMaterialResponseHashID = ASURA_HASH_ID_UNSET;

			//ask which material to take
			ListSelection_Overlay xOverlay( "Edit which material?" );
			xOverlay.SetSortList(false);
			
			
			for(int iMatSlot=0;iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY;iMatSlot++)
			{
				Asura_Hash_ID uHash = pxMesh->TextureDataPtr[ xPolygon.PolyIndex ].m_auMaterialResponseHashID[iMatSlot];
				Tools_MaterialResponse* pxMat = Tools_MaterialResponse_System::GetMaterialByHashID(uHash);
				
				if(pxMat)
				{
					char szText[100];
					Asura_PrintfToBfr(szText,"%s - %s",sizeof(szText),true,iMatSlot ? "Secondary" : "Primary",pxMat->GetName().GetString());

					xOverlay.AddItem(szText, uHash);
				}

				//check to see if an override is currently being used , and offer that as well
				Asura_Hash_ID uOverrideHash = Asura_MaterialResponse_System::GetOverriddenHash( uHash );
                if( uOverrideHash != uHash )
                {
				    pxMat = Tools_MaterialResponse_System::GetMaterialByHashID( uOverrideHash );
				    if(pxMat)
				    {
					    char szText[100];
					    Asura_PrintfToBfr(szText,"%s (Override) - %s",sizeof(szText),true,iMatSlot ? "Secondary" : "Primary",pxMat->GetName().GetString());

					    xOverlay.AddItem( szText, uOverrideHash );
				    }
			    }
            }

			if(xOverlay.GetNumItems() >= 2)
			{
				if(!xOverlay.DoModal()) return;
				uMaterialResponseHashID = xOverlay.GetSelectionData();
			}
			else if(xOverlay.GetNumItems() == 1)
			{
				uMaterialResponseHashID = xOverlay.GetItemData(0);
			}
		

			if( uMaterialResponseHashID == ASURA_HASH_ID_UNSET )
			{
				ShowMessageOverlay( "Can't edit material on polygons that have yet to have a material assigned." );
				return;
			}

			Tools_Texturing_MaterialResponse::EditMaterial( uMaterialResponseHashID );
		}

		#if defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
			else
		#endif

	#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

	#if !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY

	{
		//get the material from the selected polygon
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);
		int iMaterialID = pxMesh->TextureDataPtr[xPolygon.PolyIndex].m_iMaterialID;

		if(iMaterialID == ASURA_MATERIAL_NONE)
		{
			ShowMessageOverlay("Can't edit material on polygons that have yet to have a material assigned.");
			return;
		}

		Tools_PickTexture::EditMaterial(iMaterialID);
	}

	#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM || defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
}

void Tools_Mode_Texturing::GetMaterialOnPolygon()
{
	Tools_Object_System::EnableAllControls( false );
	SetPromptedSelectionMode(PSM_GetMaterialOnPolygon);
}
void Tools_Mode_Texturing::GetMaterialOnPolygon_PolygonSelected(PolygonID xPolygon)
{
	SetPromptedSelectionMode(PSM_None);

	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);

	//get the old style material
	int iMaterialID = pxMesh->TextureDataPtr[xPolygon.PolyIndex].m_iMaterialID;

	if(iMaterialID != ASURA_MATERIAL_NONE)
	{
		SetMaterialID(iMaterialID);
	}


	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	#if defined ASURA_MATERIAL_RESPONSE_CHARACTERS_ONLY
	if( Tools_SaveAsura::GetFileType() == AsuraFile_Skin )
	#endif
	{
		const Asura_Hash_ID uMaterialResponseHashID = pxMesh->TextureDataPtr[ xPolygon.PolyIndex ].m_auMaterialResponseHashID[0];

		if( uMaterialResponseHashID != ASURA_HASH_ID_UNSET )
		{
			SetMaterialResponse(uMaterialResponseHashID);
			Tools_Texturing_MaterialResponse::SetSelectedMaterial(uMaterialResponseHashID);
		}
	}
	#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

}

#ifdef ASURA_USE_MATERIAL_RESPONSE_SYSTEM
void Tools_Mode_Texturing::GetFirstMaterialOnPolygonForSecondary()
{
	Tools_Object_System::EnableAllControls( false );
	s_bAlphaPainting = false;
	SetPromptedSelectionMode(PSM_GetFirstMaterialOnPolygonForSecondary);
}
void Tools_Mode_Texturing::GetFirstMaterialOnPolygonForSecondary_PolygonSelected(PolygonID xPolygon)
{
	SetPromptedSelectionMode(PSM_None);

	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);



	const Asura_Hash_ID uMaterialResponseHashID = pxMesh->TextureDataPtr[ xPolygon.PolyIndex ].m_auMaterialResponseHashID[0];

	SetSecondMaterialResponse(uMaterialResponseHashID);
}

void Tools_Mode_Texturing::GetSecondMaterialOnPolygonForSecondary()
{
	Tools_Object_System::EnableAllControls( false );
	s_bAlphaPainting = false;
	SetPromptedSelectionMode(PSM_GetSecondMaterialOnPolygonForSecondary);
}
void Tools_Mode_Texturing::GetSecondMaterialOnPolygonForSecondary_PolygonSelected(PolygonID xPolygon)
{
	SetPromptedSelectionMode(PSM_None);

	Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);



	const Asura_Hash_ID uMaterialResponseHashID = pxMesh->TextureDataPtr[ xPolygon.PolyIndex ].m_auMaterialResponseHashID[1];

	SetSecondMaterialResponse(uMaterialResponseHashID);
}
#endif //ASURA_USE_MATERIAL_RESPONSE_SYSTEM



void Tools_Mode_Texturing::LimitUVEditingToSelectedTextureClicked()
{
	if(!s_bLimitUVEditingToSelectedTexture) return;

	//select the texture to limit the editing to
	s_iTextureForUVEditing = Tools_PickTexture::PickTexture("Limit UV editing to which texture?",s_iTextureForUVEditing,PICKTEXTURE_PRIMARY|PICKTEXTURE_SECONDARY|PICKTEXTURE_BUMP);
}

bool Tools_Mode_Texturing::IsEnabled_EditTexture()
{
	return !s_bLimitUVEditingToSelectedTexture;
}

void Tools_Mode_Texturing::HandleMeshChange(int iModuleIndex)
{

	if(BasePolygon.ModuleIndex == iModuleIndex) BasePolygon.ModuleIndex = -1;
	if(SelectedPolygon1.ModuleIndex == iModuleIndex) SelectedPolygon1.ModuleIndex = -1;
	if(SelectedPolygon2.ModuleIndex == iModuleIndex) SelectedPolygon2.ModuleIndex = -1;

	Tools_Texturing_Copy::ClearLists();
}


void Tools_Mode_Texturing::SetMaterialFromSelection()
{
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	Asura_Hash_ID uNewMaterialResponseHashID = ASURA_HASH_ID_UNSET;
	#else
	int iNewMaterialID = ASURA_MATERIAL_NONE;
	#endif

	PolygonID xPolygon;
	Tools_Polygon_Selection::StartIteration();
	while(Tools_Polygon_Selection::GetNextSelected(&xPolygon))
	{
		//get the mesh data
		Asura_Mesh* pxMesh = Asura_Module_System::GetMeshForModule(xPolygon.ModuleIndex);
		
		//get the texturing info
		Asura_Assert(pxMesh->TextureDataPtr,"Texture data hasn't been allocated for the mesh");
		Asura_Mesh_TextureInfo* pxTexInfo = &pxMesh->TextureDataPtr[xPolygon.PolyIndex];
		/*
		If we haven't found a texture to use yet , go with the one in this poly regardless
		Otherwise we go with the texture on this polygon , unless we have already found a polygon
		using the current texture.
		*/
		#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			if( uNewMaterialResponseHashID == ASURA_HASH_ID_UNSET
			 || (pxTexInfo->m_auMaterialResponseHashID[0] != ASURA_HASH_ID_UNSET
			 && s_pxTextureEdit->GetMaterialResponse(0) != uNewMaterialResponseHashID) )
			{
				uNewMaterialResponseHashID = pxTexInfo->m_auMaterialResponseHashID[0];
			}
		#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
			if((iNewMaterialID == ASURA_MATERIAL_NONE) || (pxTexInfo->m_iMaterialID!=ASURA_MATERIAL_NONE && s_pxTextureEdit->GetMaterialID()!=iNewMaterialID))
			{
				iNewMaterialID = pxTexInfo->m_iMaterialID;
			}
		#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	}

	//update the current material
	#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	if( uNewMaterialResponseHashID != ASURA_HASH_ID_UNSET )
	{
		SetMaterialResponse( uNewMaterialResponseHashID );
	}
	#else // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
	if(iNewMaterialID!=ASURA_MATERIAL_NONE)
	{
		SetMaterialID(iNewMaterialID);
	}
	#endif // !defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
}

void Tools_Mode_Texturing::ToggleUVEditOverlay()
{
	s_pxUVEdit->ToggleOverlay();
}

bool Tools_Mode_Texturing::IsUVEditOverlayOpen()
{
	return s_pxUVEdit->IsOverlayOpen();
}

void Tools_Mode_Texturing::Undo()
{
	if(Tools_Undo_System::Undo())
	{
		s_pxUVEdit->UpdateNodesFromSource();
	}
}

void Tools_Mode_Texturing::Redo()
{
	if(Tools_Undo_System::Redo())
	{
		s_pxUVEdit->UpdateNodesFromSource();
	}
}


//////////////////////////////////////////////////////////////////////////////////////////////////

#include "Rendering/Asura_Surface.h"
#include "Rendering/Tools_RenderModules.h"
void Tools_UpdateTexturingOnSurfaces(Asura_Mesh* mesh, bool bInvalidateCollisionMesh /*=true*/)
{
	mesh->DeleteProcessedTextureInfo();

	Asura_Surface_Mesh* surfaceMesh = Asura_Surface_Mesh_System::GetSurfaceMesh(mesh->Index);
	if (surfaceMesh) surfaceMesh->Invalidate();
	
	if(bInvalidateCollisionMesh)
	{
		// Have to update the module's collision mesh.
		for (int i=0;i<Asura_Module_System::GetNumberOfModules();i++)
		{
			if (Asura_Module_System::GetMeshForModule(i) == mesh)
			{
				Tools_Mode_MeshEdit::InvalidateModuleCollisionMesh(i);
			}
		}
	}
	Asura_Mesh_System::PlatformResetRenderMeshOnly(mesh);
	Tools_RenderModule_System::UpdateModuleMesh(mesh->Index);
}







//////////////////////////////////////////////////////////////////////////////////////////
/*----------------**
** UVShiftOverlay **
**----------------*/

UVShiftOverlay::UVShiftOverlay(int iWidth,int iHeight)
:Tools_Modal_Overlay("Shift UV")
{
	Asura_Assert(iWidth,"");
	Asura_Assert(iHeight,"");

	m_iWidth = iWidth;
	m_iHeight = iHeight;
	m_fShiftU = 0;
	m_fShiftV = 0;
}

bool UVShiftOverlay::DoModal()
{
	m_bReturnValue = false;
	ShowOverlay();
	return m_bReturnValue;
}

void UVShiftOverlay::OkaySelected(int iData)
{
	Asura_Assert(iData,"");
	*(bool*)iData = true;
}

void UVShiftOverlay::CancelSelected(int iData)
{
	Asura_Assert(iData,"");
	*(bool*)iData = false;
}

void UVShiftOverlay::InitialiseOverlay(class Asura_Overlay* pxOverlay)
{
	//setup the controls on the overlay
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Numerical_Dialog("U Pixels",xOffset,&m_fShiftU,-(float)m_iWidth,(float)m_iWidth,1.0f,"%.1f");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("V Pixels",xOffset,&m_fShiftV,-(float)m_iHeight,(float)m_iHeight,1.0f,"%.1f");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Function_Dialog("Okay",xOffset,OkaySelected,(int)&m_bReturnValue,&pxOverlay->m_bShouldClose);
	pxOverlay->AddDialog(pxDialog);
	xOffset.x+=7.0f;

	pxDialog = new Asura_Function_Dialog("Cancel",xOffset,CancelSelected,(int)&m_bReturnValue,&pxOverlay->m_bShouldClose);
	pxOverlay->AddDialog(pxDialog);
}

//////////////////////////////////////////////////////////////////////////////////////////
/*----------------**
** UVScaleOverlay **
**----------------*/

UVScaleOverlay::UVScaleOverlay()
:Tools_Modal_Overlay("Scale UV")
{
	m_fScaleU = 1.0f;
	m_fScaleV = 1.0f;
	m_bUniformScale = true;
	m_bUChanged = false;
	m_bVChanged = false;
}

bool UVScaleOverlay::DoModal()
{
	m_bReturnValue = false;
	ShowOverlay();
	return m_bReturnValue;
}

void UVScaleOverlay::OkaySelected(int iData)
{
	Asura_Assert(iData,"");
	*(bool*)iData = true;
}

void UVScaleOverlay::CancelSelected(int iData)
{
	Asura_Assert(iData,"");
	*(bool*)iData = false;
}

void UVScaleOverlay::InitialiseOverlay(class Asura_Overlay* pxOverlay)
{
	//setup the controls on the overlay
	Asura_Vector_2 xOffset = Asura_Vector_2(1.0f,3.0f);
	Asura_Dialog* pxDialog;

	pxDialog = new Asura_Numerical_Dialog("U",xOffset,&m_fScaleU,-1000.0f,1000.0f,0.1f,"%.3f",&m_bUChanged);
	pxDialog->SetHelpText("Horizontal multiplier.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Numerical_Dialog("V",xOffset,&m_fScaleV,-1000.0f,1000.0f,0.1f,"%.3f",&m_bVChanged);
	pxDialog->SetHelpText("Vertical multiplier.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Uniform Scale",xOffset,&m_bUniformScale);
	pxDialog->SetHelpText("Force the U and V scales to be the same.");
	pxOverlay->AddDialog(pxDialog);
	xOffset.y+=Asura_Dialog::DefaultSpacing();


	pxDialog = new Asura_Function_Dialog("Okay",xOffset,OkaySelected,(int)&m_bReturnValue,&pxOverlay->m_bShouldClose);
	pxOverlay->AddDialog(pxDialog);
	xOffset.x+=7.0f;

	pxDialog = new Asura_Function_Dialog("Cancel",xOffset,CancelSelected,(int)&m_bReturnValue,&pxOverlay->m_bShouldClose);
	pxOverlay->AddDialog(pxDialog);
}

void UVScaleOverlay::Update()
{
	if(m_bUniformScale)
	{
		if(m_bUChanged) m_fScaleV = m_fScaleU;
		if(m_bVChanged) m_fScaleU = m_fScaleV;
	}
	m_bUChanged = false;
	m_bVChanged = false;
}
