//////////////////////////////////////////////////////////////////////////////
//
// Tools_Texturing_MaterialResponse.cpp
//
// Copyright (c) 2006 Rebellion. All rights reserved.
//
//////////////////////////////////////////////////////////////////////////////

/******************************************************************************/
/* Included Files                                                             */
/******************************************************************************/

#include "Core/Asura.h"

#if defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM

#include "Texturing/Tools_Texturing_MaterialResponse.h"

// Asura Platform-independent
#include "ChunkSystem/Asura_Chunk_Stream.h"
#include "ChunkSystem/Asura_SaveBlock.h"
#include "Core/Asura_Handle.h"
#include "Core/Asura_Graph.h"
#include "Core/Asura_List.h"
#include "Core/Asura_String.h"
#include "Core/Asura_Maths.h"
#include "Core/Asura_TextureManagement.h"
#include "Debugging/Asura_CPUProfile.h"
#include "Debugging/Asura_Debug_Timers.h"
#include "Debugging/Asura_Logfile.h"
#include "FileAccess/Asura_FileAccess.h"
#include "Interface/Asura_Dialog.h"
#include "Interface/Asura_Graph_Dialog.h"
#include "Interface/Asura_Overlay.h"
#include "Interface/Asura_Overlay_System.h"
#include "Interface/Asura_TexturePicker_Dialog.h"
#include "Interface/Asura_Tree_Dialog.h"
#include "Mesh/Asura_Mesh.h"
#include "Mesh/Asura_Mesh_System.h"
#include "Rendering/Asura_Module.h"

// Asura PC
#include "Rendering/Asura_Direct3D_EnvironmentRenderer.h"
#include "Rendering/Asura_Direct3D_SkinRenderer.h"

// Tools
#include "Tools_Overlays.h"
#include "ChunkSystem/Tools_Chunk_MaterialResponse.h"
#include "Interface/Tools_ColourPicker.h"
#include "Load&Save/Tools_ModeDataFile.h"
#include "Load&Save/Tools_SaveAsura.h"
#include "Materials/Tools_MaterialResponse_System.h"
#include "Misc/Tools_HashTable.h"
#include "Misc/Tools_MessageLog.h"
#include "Misc/Tools_Modal_Overlay.h"
#include "Misc/Tools_Polygon_Selection.h"
#include "Misc/Tools_Timed_Messages.h"
#include "PC/FileSystem/Tools_FileSystem.h"
#include "Texturing/Tools_Mode_Texturing.h"
#include "Texturing/Tools_PickTexture.h"
#include "Texturing/Tools_Texturing_Find.h"
#include "Materials/Tools_Overlay_ImportMaterialResponse.h"
#include "Materials/Tools_Overlay_MaterialResponseOverrides.h"
#include "Materials/Tools_Overlay_ViewMaterialResponse.h"

#ifdef ASURA_USE_DECAL_MANAGER
#include "SpecialFX/Asura_Decal_Manager.h"
#endif

#ifndef ASURAIMPORT_EXPORTS
#include "SpecialFX/Asura_StaticCorona_System.h"
#endif // ASURAIMPORT_EXPORTS

/******************************************************************************/
/* Constants                                                                  */
/******************************************************************************/

const float fTEXTURE_PICKER_SIZE = 64.0f;

const int iDAT_FILE_VERSION = 1;

/******************************************************************************/
/* Types                                                                      */
/******************************************************************************/

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*--------------------------------------------**
** Tools_MaterialResponse_DropDownList_Dialog **
**--------------------------------------------*/


//Drop down list for selecting materials , with an (external) option for filtering the list to
//the materials used by the environment
class Tools_MaterialResponse_DropDownList_Dialog : public Asura_Hash_DropDownList_Dialog
{
public :
	Tools_MaterialResponse_DropDownList_Dialog(const Asura_Vector_2& xOffset, int iNumberOfLines, bool bFilterByEnv, Asura_Hash_ID* puSelectedMaterial, bool* pbChangedFlag)
	:Asura_Hash_DropDownList_Dialog(xOffset)
	{
		m_bFilterByEnv = bFilterByEnv;
		m_iHashListSize = 0;
		m_puHashList = 0;

		Initialise(xOffset, iNumberOfLines, puSelectedMaterial, pbChangedFlag, 0);
	}

	~Tools_MaterialResponse_DropDownList_Dialog()
	{
		delete [] m_puHashList;
	}

	void SetFilterByEnv(bool bSet)
	{
		if(bSet != m_bFilterByEnv)
		{
			m_bFilterByEnv = bSet;
			m_iListValidID++;
		}
	}

protected :
	virtual u_int GetNumberOfHashValues() const;
	virtual const char* GetNameByIndex(int iIndex) const;

	virtual int GetListValidID() const {return m_iListValidID;}
	virtual const char* GetUnknownHashText() const {return "*Not In List*";}

private :
	bool m_bFilterByEnv;
	int m_iListValidID;

	mutable int m_iHashListSize;
	mutable Asura_Hash_ID* m_puHashList;
};

u_int Tools_MaterialResponse_DropDownList_Dialog::GetNumberOfHashValues() const
{
	if(m_bFilterByEnv)
	{
		if(!m_iHashListSize)
		{
			//find the materials being used
			Asura_HashTree xMaterialTree;
			Tools_Texturing_MaterialResponse::FindMaterialsUsedByEnv(xMaterialTree);

			m_iHashListSize = xMaterialTree.GetNumberOfNodes();
			if(m_iHashListSize)
			{
				m_puHashList = new Asura_Hash_ID[m_iHashListSize];
				xMaterialTree.MakeHashList(m_puHashList);
			}
		}
		return m_iHashListSize;
	}
	else
	{
		return Tools_MaterialResponse_System::GetNumMaterials();
	}
}

const char* Tools_MaterialResponse_DropDownList_Dialog::GetNameByIndex(int iIndex) const
{
	if(m_bFilterByEnv)
	{
		return Tools_MaterialResponse_System::GetMaterialNameByHashID(m_puHashList[iIndex], false);
	}
	else
	{
		return Tools_MaterialResponse_System::GetMaterialNameByIndex(iIndex);
	}
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*---------------------------------**
** Tools_MaterialResponseHashAndID **
**---------------------------------*/

//Helper class for updating old style materials to match textures used by material response system
//RAR 30/07/2007
class Tools_MaterialResponseHashAndID
{
public :
	Asura_Hash_ID m_auMaterialResponse[2];
	int m_iMaterialID;

	Asura_Hash_ID GetHashID() const {return m_auMaterialResponse[0]+m_auMaterialResponse[1];}

	bool operator == (const Tools_MaterialResponseHashAndID& xOther) const
	{
		return (m_auMaterialResponse[0] == xOther.m_auMaterialResponse[0] &&
		        m_auMaterialResponse[1] == xOther.m_auMaterialResponse[1]);
	}
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*-----------------------------------**
** Tools_ConvertMaterialResponseToID **
**-----------------------------------*/

//Helper class for updating old style materials to match textures used by material response system
//RAR 30/07/2007
class Tools_ConvertMaterialResponseToID
{
public :
	Tools_ConvertMaterialResponseToID();
	~Tools_ConvertMaterialResponseToID();

	int GetMaterialID(Asura_Hash_ID uMat1,Asura_Hash_ID uMat2);


private :
	Tools_HashTable<Tools_MaterialResponseHashAndID> m_xMaterialConvTable;
	List<Tools_MaterialResponseHashAndID*> m_xMaterialConvList;
};

Tools_ConvertMaterialResponseToID::Tools_ConvertMaterialResponseToID()
:m_xMaterialConvTable(256)
{
}

Tools_ConvertMaterialResponseToID::~Tools_ConvertMaterialResponseToID()
{
	while(m_xMaterialConvList.size())
	{
		delete m_xMaterialConvList.first_entry();
		m_xMaterialConvList.delete_first_entry();
	}	
}

int Tools_ConvertMaterialResponseToID::GetMaterialID(Asura_Hash_ID uMatHash1,Asura_Hash_ID uMatHash2)
{
	//search for an entry for this material combination
	Tools_MaterialResponseHashAndID xTest;
	xTest.m_auMaterialResponse[0] = uMatHash1;	
	xTest.m_auMaterialResponse[1] = uMatHash2;

	const Tools_MaterialResponseHashAndID* pxExisting = m_xMaterialConvTable.Find(xTest);
	if(pxExisting) return pxExisting->m_iMaterialID;

	int iNewMaterialID = ASURA_MATERIAL_NONE;
	
	//No existing entry for this.
	//Generate a temporary old-style material that has appropriate textures
	Asura_Material xMat;
	xMat.Initialise();

	for(int iSlot=0; iSlot<2; iSlot++)
	{
		const Tools_MaterialResponse* pxMaterialResponse = Tools_MaterialResponse_System::GetMaterialByHashID(iSlot ? uMatHash2 : uMatHash1);
		if(!pxMaterialResponse) continue;

		xMat.m_iTextureID[iSlot] = pxMaterialResponse->GetTexture(Asura_MaterialResponse::ALBEDO_TEXTURE);
		xMat.m_iNormalMapID[iSlot] = pxMaterialResponse->GetTexture(Asura_MaterialResponse::NORMAL_MAP);
		xMat.m_iSpecularMapID[iSlot] = pxMaterialResponse->GetTexture(Asura_MaterialResponse::SPECULAR_MAP);
	}

	if(xMat.m_iTextureID[0] != ASURA_TEXTURE_NONE)
	{
		//Search for an existing matching material
		//Note that this will result in the material being added to Tools_PickTexture's list of environment materials , 
		//regardless of whether that is appropriate. 
		//However that list ought to be redundant when using material response anyway.
		Tools_PickTexture::CreateOrFindExistingMaterial(&xMat, false, iNewMaterialID, NULL);
	}

	//fill in a new entry for this conversion and add to the list and tree
	Tools_MaterialResponseHashAndID* pxNewEntry = new Tools_MaterialResponseHashAndID;
	pxNewEntry->m_auMaterialResponse[0] = uMatHash1;
	pxNewEntry->m_auMaterialResponse[1] = uMatHash2;
	pxNewEntry->m_iMaterialID = iNewMaterialID;

	m_xMaterialConvTable.Insert(pxNewEntry);
	m_xMaterialConvList.add_entry(pxNewEntry);

	return  iNewMaterialID;


}


////////////////////////////////////////////////////////////////////////////////
//
// Tools_MaterialResponse_UndoData_ApplyMaterial
//
// Represents one undoable "apply material to polygon" action
//
////////////////////////////////////////////////////////////////////////////////

class Tools_MaterialResponse_UndoData_ApplyMaterial : public RootUndoData
{
public:

	Tools_MaterialResponse_UndoData_ApplyMaterial( const u_int uMeshIndex, const u_int uPolygonIndex );

	virtual int GetPriority() const { return 0; }
	virtual RootUndoData* ProcessUndoData();

	virtual const char* GetDescription() const{return "Apply Material";}
protected:

	const u_int m_uMeshIndex;
	const u_int m_uPolygonIndex;

	u_int m_auPreviousMaterialHash[TOOLS_NUM_MATERIALRESPONSE_PER_POLY];
};

////////////////////////////////////////////////////////////////////////////////
//
// Tools_MaterialResponse_UndoData_ImportMaterial
//
// Represents one undoable "import old polygon material" action
//
////////////////////////////////////////////////////////////////////////////////

class Tools_MaterialResponse_UndoData_ImportMaterial : public Tools_MaterialResponse_UndoData_ApplyMaterial
{
public:

	Tools_MaterialResponse_UndoData_ImportMaterial( const u_int uMeshIndex, const u_int uPolygonIndex );

	virtual const char* GetDescription() const{return "Import Old Materials";}

};

////////////////////////////////////////////////////////////////////////////////
//
// Tools_MaterialResponse_UndoData_RemoveMaterial
//
// Represents one undoable "remove polygon material" action
//
////////////////////////////////////////////////////////////////////////////////

class Tools_MaterialResponse_UndoData_RemoveMaterial : public Tools_MaterialResponse_UndoData_ApplyMaterial
{
public:

	Tools_MaterialResponse_UndoData_RemoveMaterial( const u_int uMeshIndex, const u_int uPolygonIndex );

	virtual const char* GetDescription() const{return "Remove Materials From Mesh";}
};

////////////////////////////////////////////////////////////////////////////////
//
// Tools_MaterialResponse_Overlay
//
// Simple subclass of Asura_Overlay to manage the list of material names, and
// ensure the tick in the menu bar is removed if the user closes the overlay
// with the X button
//
////////////////////////////////////////////////////////////////////////////////


class Tools_MaterialResponse_Overlay : public Asura_Overlay
{
public:

	// Constructors / destructor
	Tools_MaterialResponse_Overlay();
	virtual ~Tools_MaterialResponse_Overlay();

	// Functions
	void AddDialogs();
	void UpdateDialogs();
	void UpdateImageDialog();

	virtual void PerFrameUpdate();

	// Static functions
	static Asura_Hash_ID GetSelectedMaterialHash();
	static void SetSelectedMaterialHash(Asura_Hash_ID uMaterialHash);
	static Tools_MaterialResponse* GetSelectedMaterial();
	static bool GetResetSelectedMaterialsOnly();
	static bool OverlayExists();

protected:
	

	// Types
	enum DialogID
	{
		DialogID_MaterialList,
	};

	// Static functions
	static bool IsMaterialNodeSelected();

	// Data
	Asura_TreeDialog_Tree m_xMaterialNameTree;

	bool m_bResetInvalidMaterialsOnly;

	Asura_Image_Dialog* m_pxImageDialog;
	bool m_bSelectionChanged;

	static bool s_bFilterTreeByEnv;
	bool m_bFilterTreeByEnvChanged;


};

bool Tools_MaterialResponse_Overlay::s_bFilterTreeByEnv = false;

////////////////////////////////////////////////////////////////////////////////
//
// Tools_MaterialProperties_Overlay
//
// Overlay to allow editing the properties of a material response
//
////////////////////////////////////////////////////////////////////////////////

class Tools_MaterialProperties_Overlay : public Asura_Overlay
{
public:

	// Constructor
	explicit Tools_MaterialProperties_Overlay( const char* const szName, const Asura_Hash_ID uMaterialHash, int iOverlayIndex );
	virtual ~Tools_MaterialProperties_Overlay();

	static void EditMaterial(const Asura_Hash_ID uMaterialHash,int iOverlayIndex = -1);

	// Functions

	void AddDialogs();
	void UpdateDialogs();

	virtual void PerFrameUpdate();

	Tools_MaterialResponse* GetMaterial() { return m_pxMaterialResponse; }
	const Tools_MaterialResponse* GetMaterial() const { return m_pxMaterialResponse; }
	u_int GetMaterialHash() const { return m_uMaterialHash; }
	const char* GetSelectedSlave() const { return m_pszSlaveNames[ m_iSelectedSlave ]; }
	int GetOverlayIndex() const {return m_iOverlayIndex;}

	//Searches for an existing overlay for a material
	static Tools_MaterialProperties_Overlay* FindOverlayForMaterial(Asura_Hash_ID uMaterialHash);
	//Gets the material overlay that has most recently been used
	static Tools_MaterialProperties_Overlay* GetMostRecentlySelectedOverlay();

	//Calls UpdateDialogs() on all overlays
	static void UpdateAllOverlayDialogs();

	//close all overlays of this type
	static void CloseOverlays();

	// Types
	enum DialogID
	{
		DialogID_EnvironmentMap,

		DialogID_ReadOnlyIcon,
		DialogID_MaterialName,

		DialogID_ParentName,
		DialogID_BreakFromParent,

		DialogID_AddSlave,
		DialogID_RemoveSlave,
		DialogID_SlaveList,
		DialogID_Enslave,
	};

protected:
	// Static functions
	static bool EnableResponseControls( int iMaterialHash );
	static bool EnableZPrePassControls( int iMaterialHash );
	static void RevertChanges( int iUserData );
	static void RenderTypeChanged( int iThis );
	static void SurfaceTypeChanged( int iThis );
	
	static void RemoveClosedOverlaysFromList();//Removes any overlays that no longer exist from s_xOverlayIndexList	
	static int GetOverlayIDFromIndex(int iIndex);//converts material properties overlay index into ID used by material system

	// Data
	Tools_MaterialResponse m_xInitialProperties;
	Tools_MaterialResponse* const m_pxMaterialResponse;
	const char** m_pszSlaveNames;
	const Asura_Hash_ID m_uMaterialHash;
	int m_iSelectedSlave;
	int m_iOverlayIndex;
	int m_iRenderType;

#if !defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
	//for surface type drop-down list
	u_int m_uSurfaceTypeListSize;
	char** m_pszSurfaceTypes;
#endif	// !defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

	static const char* s_aszRenderTypeNames[Asura_MaterialResponse::NUMBER_OF_RENDER_TYPES];

	static bool s_bSetSurfaceTypeByTexture;
	//list of index numbers of open overlays , sorted so that the most recently selected one is at the start
	static List<int> s_xOverlayIndexList;

	Tools_Overlay_MaterialResponseOverrides m_xMaterialResponseOverrides;

public :
	//some public static data for loading/saving preferences
	static int s_iMaxOpenOverlay;//user requested limit on the number of overlays to show
	static Asura_Vector_2 s_axOverlayPosition[iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY];

};

const char*	Tools_MaterialProperties_Overlay::s_aszRenderTypeNames[] =
{
	"Default",						// Asura_MaterialResponse::RENDER_TYPE_DEFAULT
	"Render but don't cast shadow",	// Asura_MaterialResponse::RENDER_TYPE_RENDER_BUT_DONT_CAST_SHADOW
	"Don't render but cast shadow",	// Asura_MaterialResponse::RENDER_TYPE_DONT_RENDER_BUT_CAST_SHADOW
};

////////////////////////////////////////////////////////////////////////////////
//
// Tools_CopyMaterialProperties_Overlay
//
// Overlay to allow selective copying of properties from one material to another
//
////////////////////////////////////////////////////////////////////////////////

class Tools_CopyMaterialProperties_Overlay : public Tools_Modal_Overlay
{
public:

	// Constructors / destructor
	Tools_CopyMaterialProperties_Overlay( const Asura_Hash_ID uInitialCopyFromHash, const Asura_Hash_ID uInitialCopyToHash );

protected:

	// Types
	enum DialogID
	{
		DialogID_Everything,

		DialogID_Responses,
		DialogID_AmbientResponse,
		DialogID_DiffuseResponse,
		DialogID_SpecularResponse,
		DialogID_TransparencyResponse,
		DialogID_SelfIlluminationResponse,
		DialogID_ReflectivityResponse,

		DialogID_Textures,
		DialogID_AlbedoMaps,
		DialogID_NormalMaps,
		DialogID_SpecularMaps,
		DialogID_TransparencyMaps,
		DialogID_SelfIlluminationMaps,
		DialogID_ReflectivityTextures,

		DialogID_TextureAnims,

		DialogID_SpecialFX,

		DialogID_SurfaceType,
	};

	// Virtual overrides
	virtual void Update();

	// Functions
	Tools_MaterialResponse* GetFromMaterial();
	Tools_MaterialResponse* GetToMaterial();

private:

	// Virtual overrides
	virtual void InitialiseOverlay( Asura_Overlay* pxOverlay );

	// Static functions
	static void OnOK( int iThis );

	// Data
	Asura_Hash_ID m_uFromHash;
	Asura_Hash_ID m_uToHash;
	Tools_MaterialResponse::PropertyCopyFlags m_xCopyFlags;

	Tools_MaterialResponse_DropDownList_Dialog* m_pxFromMaterialDialog;
	Tools_MaterialResponse_DropDownList_Dialog* m_pxToMaterialDialog;

	static bool s_bFilterFromListByEnv;
	static bool s_bFilterToListByEnv;

};

bool Tools_CopyMaterialProperties_Overlay::s_bFilterFromListByEnv = false;
bool Tools_CopyMaterialProperties_Overlay::s_bFilterToListByEnv = false;

////////////////////////////////////////////////////////////////////////////////
//
// Tools_AddMaterialSlaves_Overlay
//
// Overlay for selecting materials to become slaves
//
////////////////////////////////////////////////////////////////////////////////

class Tools_AddMaterialSlaves_Overlay : public ListSelection_Overlay
{
public:

	explicit Tools_AddMaterialSlaves_Overlay( const Asura_String& xTitle )
		: ListSelection_Overlay( xTitle )
	{
	}

protected:

	virtual bool IsTickboxList() const { return true; }
};

////////////////////////////////////////////////////////////////////////////////
//
// Tools_EnslaveMaterialToOther_Overlay
//
// Overlay for enslaving one material to another
//
////////////////////////////////////////////////////////////////////////////////

class Tools_EnslaveMaterialToOther_Overlay : public Tools_Modal_Overlay
{
public:

	// Constructors / destructor
	Tools_EnslaveMaterialToOther_Overlay( const Asura_String& xTitle, Tools_MaterialResponse& xMaterial );
	virtual ~Tools_EnslaveMaterialToOther_Overlay();

private:

	// Functions
	Asura_TreeDialog_Tree& GetTree() { return m_xMaterialNameTree; }
	Tools_MaterialResponse& GetMaterial() { return m_xMaterial; }
	void UpdateTreeDialog();

	// Virtual overrides
	virtual void InitialiseOverlay( Asura_Overlay* pxOverlay );
	virtual void Update();

	// Static functions
	static bool IsMaterialNodeSelected();
	static void OnOK();
	static bool EnslaveTreeFilter( const Tools_MaterialResponse& xMaterial, const void* const pUserData );


	// Data
	Asura_TreeDialog_Tree m_xMaterialNameTree;
	Tools_MaterialResponse& m_xMaterial;
	Asura_Tree_Dialog* m_pxTreeDialog;
	bool m_bFilterTreeByEnvChanged;

	Asura_HashTree m_xUsedMaterialTree;

	// Static data
	static Tools_EnslaveMaterialToOther_Overlay* s_pxThis;
	static bool s_bFilterTreeByEnv;
};

/******************************************************************************/
/* Tools_MaterialResponse_Overlay Class Functions                             */
/******************************************************************************/

Tools_MaterialResponse_Overlay::Tools_MaterialResponse_Overlay()
	: Asura_Overlay( "Material Response List", false, &Tools_Texturing_MaterialResponse::s_xOverlayPosition, &Tools_Texturing_MaterialResponse::s_uOverlayMinimised )
	, m_bResetInvalidMaterialsOnly( false )
	, m_pxImageDialog(0)
	, m_bSelectionChanged(false)
	, m_bFilterTreeByEnvChanged(false)
{
	m_uID = TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST;
	Tools_Texturing_MaterialResponse::s_bOverlayOpen = true;

	
}

Tools_MaterialResponse_Overlay::~Tools_MaterialResponse_Overlay()
{
	Tools_Texturing_MaterialResponse::s_bOverlayOpen = false;

}

void Tools_MaterialResponse_Overlay::AddDialogs()
{
	Asura_Dialog* pxDialog = NULL;
	Asura_Vector_2 xOffset( 1.0f, 3.0f );

	// Show material statistics
	pxDialog = new Asura_Function_Dialog( "Show Statistics", xOffset, Tools_Texturing_MaterialResponse::ShowStatistics );
	pxDialog->SetHelpText( "Displays statistics about material response usage." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();
	
	// logfile option
	pxDialog = new Asura_Function_Dialog( "Output to Logfile", xOffset, Tools_MaterialResponse_System::LogMaterials );
	pxDialog->SetHelpText( "Scans through all materials and writes out details such as texture usage to logfiles/Materials.txt and logfiles/DuplicateTextures.txt" );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Find Missing Materials", xOffset, Tools_Texturing_MaterialResponse::FindMissingMaterials );
	pxDialog->SetHelpText( "Displays a list of materials that are missing or that aren't checked in." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Show Env Mat Thumbnails", xOffset, Tools_Overlay_ViewMaterialResponse::CreateOverlay );
	pxDialog->SetHelpText( "Creates an overlay that shows thumbnails of the materials used on the environment." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();

	pxDialog = new Asura_Function_Dialog( "Show Bad Texture Usage report", xOffset, Tools_Texturing_MaterialResponse::ShowBadTextureUsageReport);
	pxDialog->SetHelpText( "Creates an overlay that shows thumbnails of the materials used on the environment." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();


	// Reset all materials on mesh
	pxDialog = new Asura_Function_Dialog( "Remove Material Assignments From Mesh", xOffset, Tools_Texturing_MaterialResponse::RemoveMaterialsFromMesh );
	pxDialog->SetHelpText( "Clears material assignments from loaded meshes." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Flag to allow removing invalid material assignments only
	pxDialog = new Asura_Tickbox_Dialog( "Invalid/Missing Materials Only", xOffset, &m_bResetInvalidMaterialsOnly );
	pxDialog->SetHelpText( "Tick this to reset only invalid or missing assignments, rather than all polygons in the mesh." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Tickbox_Dialog::LargeSpacing();

	// Import old materials
	pxDialog = new Asura_Function_Dialog( "Import Materials From Old System", xOffset, Tools_Overlay_ImportMaterialResponse::CreateOverlay );
	pxDialog->SetHelpText( "Creates new materials from the old system's data, and applies them to all meshes." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();


	// Creation button
	pxDialog = new Asura_Function_Dialog( "Create New Material", xOffset, Tools_Texturing_MaterialResponse::CreateMaterialResponse );
	pxDialog->SetHelpText( "Creates a new material." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();

	pxDialog = new Asura_Tickbox_Dialog("Filter By Env", xOffset, &s_bFilterTreeByEnv, &m_bFilterTreeByEnvChanged);
	pxDialog->SetHelpText( "Filters the list of materials so that it only shows the materials used by the 'environment'." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();


	// Material name tree
	pxDialog = new Asura_Tree_Dialog( xOffset, &m_xMaterialNameTree, 30.0f, 30.0f, &m_bSelectionChanged );
	pxDialog->SetHelpText( "Selects a material to edit." );
	((Asura_Tree_Dialog*)pxDialog)->SetDoubleClickedFunction(Tools_Texturing_MaterialResponse::EditSelectedMaterial);
	pxDialog->m_iControlID = DialogID_MaterialList;
	AddDialog( pxDialog );
	xOffset.y += 30.0f + 1.0f;

	//Add control for resizing the tree
	pxDialog = new Asura_Resize_Dialog(pxDialog ,&Tools_Texturing_MaterialResponse::s_xUserSetTreeSize);
	AddDialog(pxDialog);

	//Add thumbnail to right
	xOffset.x += 19.0f;
	m_pxImageDialog = new Asura_Image_Dialog(xOffset,ASURA_TEXTURE_NONE,96.0f,96.0f);
	m_pxImageDialog->SetHelpText("Albedo texture from the selected material");
	m_pxImageDialog->SetTranslucent(false);
	AddDialog( m_pxImageDialog );
	xOffset.x -= 19.0f;
	
	pxDialog = new Asura_Numerical_Dialog("Max Property Overlays", xOffset, &Tools_MaterialProperties_Overlay::s_iMaxOpenOverlay,1,iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY,1,"%d");
	pxDialog->SetHelpText( "Determines the maximum number of material properties overlays that should be opened before starting to close old ones." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();


	// Edit material button
	pxDialog = new Asura_Function_Dialog( "Edit Selected Material", xOffset, Tools_Texturing_MaterialResponse::EditSelectedMaterial, NULL, IsMaterialNodeSelected );
	pxDialog->SetHelpText( "Opens the properties overlay for the selected material." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Duplicate material button
	pxDialog = new Asura_Function_Dialog( "Duplicate Selected Material", xOffset, Tools_Texturing_MaterialResponse::DuplicateSelectedMaterial, NULL, IsMaterialNodeSelected );
	pxDialog->SetHelpText( "Duplicates the selected material." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();

	// Apply button
	pxDialog = new Asura_Function_Dialog( "Apply As Primary Mat (G)", xOffset, Tools_Texturing_MaterialResponse::ApplySelectedMaterialAsPrimary, NULL, IsMaterialNodeSelected );
	pxDialog->SetHelpText( "Applies the selected material as the primary material to all currently selected mesh polygons.\nThe shortcut key only works while in texturing mode.\nIf editing a material , then the shortcut will apply that material instead." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Apply As Secondary Mat", xOffset, Tools_Texturing_MaterialResponse::ApplySelectedMaterialAsSecondary, NULL, IsMaterialNodeSelected );
	pxDialog->SetHelpText( "Applies the selected material as the secondary material to all currently selected mesh polygons.\n" );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();


	// Undo button
	pxDialog = new Asura_Function_Dialog( "Undo", xOffset, Tools_Undo_System::UndoForOverlay, NULL, Tools_Undo_System::CanUndo  );
	pxDialog->SetHelpText( "Undoes the last undoable operation." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Redo button
	pxDialog = new Asura_Function_Dialog( "Redo", xOffset, Tools_Undo_System::RedoForOverlay, NULL, Tools_Undo_System::CanRedo );
	pxDialog->SetHelpText( "Redoes the last undoable operation." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	UpdateDialogs();

	MakeDialogsRelativeToResizableDialog();
}

void Tools_MaterialResponse_Overlay::UpdateDialogs()
{
	Asura_Hash_ID uCurrentSelection = 0;
	if(Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST ))
	{
		uCurrentSelection = GetSelectedMaterialHash();
	}

	if(s_bFilterTreeByEnv)
	{
		Tools_Texturing_MaterialResponse::CreateMaterialNameTree_FilteredByEnv( m_xMaterialNameTree );
	}
	else
	{
		Tools_MaterialResponse_System::CreateMaterialNameTree( m_xMaterialNameTree );
	}
	m_xMaterialNameTree.GetRoot()->Expand( true );

	//preserve the initial selection (if its still there)
	SetSelectedMaterialHash(uCurrentSelection);
	
	UpdateImageDialog();
}

void Tools_MaterialResponse_Overlay::UpdateImageDialog()
{
	Asura_Assert(m_pxImageDialog,"");

	int iTexture = ASURA_TEXTURE_NONE;

	//get the albedo texture from the selected material
	const Asura_TreeDialog_Item* pxSelected = m_xMaterialNameTree.GetSelectedItem();
	if(pxSelected)
	{
		Asura_Hash_ID uMaterialID = (Asura_Hash_ID)pxSelected->GetUserData();
		const Tools_MaterialResponse* pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID(uMaterialID);
		if(pxMaterial)
		{
			iTexture = pxMaterial->GetTexture(Asura_MaterialResponse::ALBEDO_TEXTURE);
		}
	}

	m_pxImageDialog->SetTexture(iTexture);
	m_pxImageDialog->SetHidden(iTexture==ASURA_TEXTURE_NONE);

}

void Tools_MaterialResponse_Overlay::PerFrameUpdate()
{
	Asura_Overlay::PerFrameUpdate();

	if(m_bFilterTreeByEnvChanged)
	{
		UpdateDialogs();
		m_bFilterTreeByEnvChanged = false;
	}

	if(m_bSelectionChanged)
	{
		UpdateImageDialog();
		m_bSelectionChanged = false;
	}
}



/******************************************************************************/
/* Tools_MaterialResponse_Overlay Static Class Functions                      */
/******************************************************************************/

bool Tools_MaterialResponse_Overlay::OverlayExists()
{
	return Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST ) != 0;
}

Asura_Hash_ID Tools_MaterialResponse_Overlay::GetSelectedMaterialHash()
{
	const Asura_Overlay* const pxOverlay = Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
	if( !pxOverlay )
	{
		Asura_Assert( false, "Cannot get selected material if overlay doesn't exist." );
		return ASURA_HASH_ID_UNSET;
	}

	const Tools_MaterialResponse_Overlay* const pxMaterialListOverlay = static_cast< const Tools_MaterialResponse_Overlay* >( pxOverlay );

	const Asura_TreeDialog_Item* const pxSelectedItem = pxMaterialListOverlay->m_xMaterialNameTree.GetSelectedItem();
	return pxSelectedItem ? pxSelectedItem->GetUserData() : ASURA_HASH_ID_UNSET;
}

void Tools_MaterialResponse_Overlay::SetSelectedMaterialHash(Asura_Hash_ID uMaterialHash)
{
	Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
	if( !pxOverlay ) return;

	Tools_MaterialResponse_Overlay* pxMaterialListOverlay = static_cast< Tools_MaterialResponse_Overlay* >( pxOverlay );

	//find the node for the material ,and select it if it exists
	Asura_TreeDialog_Item* pxItem = pxMaterialListOverlay->m_xMaterialNameTree.FindByData(uMaterialHash);
	if(pxItem)
	{
		pxMaterialListOverlay->m_xMaterialNameTree.SetSelectedItem(pxItem);		

		Asura_Dialog* pxDialog = pxOverlay->GetDialog(DialogID_MaterialList);
		if(pxDialog)
		{
			static_cast<Asura_Tree_Dialog*>(pxDialog)->MakeSelectionVisible();
		}

		pxMaterialListOverlay->UpdateImageDialog();
	}
}

Tools_MaterialResponse* Tools_MaterialResponse_Overlay::GetSelectedMaterial()
{
	return Tools_MaterialResponse_System::GetMaterialByHashID( GetSelectedMaterialHash() );
}

bool Tools_MaterialResponse_Overlay::GetResetSelectedMaterialsOnly()
{
	const Asura_Overlay* const pxOverlay = Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
	if( !pxOverlay )
	{
		Asura_Assert( false, "Overlay doesn't exist." );
		return false;
	}

	const Tools_MaterialResponse_Overlay* const pxMaterialListOverlay = static_cast< const Tools_MaterialResponse_Overlay* >( pxOverlay );
	return pxMaterialListOverlay->m_bResetInvalidMaterialsOnly;
}

bool Tools_MaterialResponse_Overlay::IsMaterialNodeSelected()
{
	const Asura_Overlay* const pxOverlay = Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
	if( !pxOverlay )
	{
		// Can't assert here. Once the overlay is scheduled for destruction, the above call won't
		// find it. However the dialogs' per-frame update functions continue to be called
		return false;
	}

	const Tools_MaterialResponse_Overlay* const pxMaterialListOverlay = static_cast< const Tools_MaterialResponse_Overlay* >( pxOverlay );

	// Only leaf nodes are actual materials, the others represent directories
	const Asura_TreeDialog_Item* const pxSelectedNode = pxMaterialListOverlay->m_xMaterialNameTree.GetSelectedItem();
	return (pxSelectedNode && pxSelectedNode->GetNumChildren() == 0);
}

/******************************************************************************/
/* Tools_MaterialProperties_Overlay Class Functions                           */
/******************************************************************************/

bool Tools_MaterialProperties_Overlay::s_bSetSurfaceTypeByTexture = false;
List<int> Tools_MaterialProperties_Overlay::s_xOverlayIndexList;

Asura_Vector_2 Tools_MaterialProperties_Overlay::s_axOverlayPosition[iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY];
int Tools_MaterialProperties_Overlay::s_iMaxOpenOverlay = 1;


Tools_MaterialProperties_Overlay::Tools_MaterialProperties_Overlay( const char* const szName, const Asura_Hash_ID uMaterialHash, int iOverlayIndex )
	: Asura_Overlay( szName, false, &s_axOverlayPosition[iOverlayIndex] )
	, m_pxMaterialResponse( Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash ) )
	, m_pszSlaveNames( NULL )
	, m_uMaterialHash( uMaterialHash )
	, m_iSelectedSlave(0)
	, m_iOverlayIndex(iOverlayIndex)
	, m_iRenderType(Asura_MaterialResponse::RENDER_TYPE_DEFAULT)
{
	Asura_Assert(!s_xOverlayIndexList.contains(m_iOverlayIndex),"Overlay index already in list");
	s_xOverlayIndexList.add_entry_start(m_iOverlayIndex);

	if( m_pxMaterialResponse )
	{
		// Store a copy of the material for reverting later
		m_xInitialProperties = *m_pxMaterialResponse;

		// Ensure all the textures are loaded
		for( u_int u = 0; u < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++u )
		{
			// GetTexture() will load if not present
			m_pxMaterialResponse->GetTexture(u);
		}

		// Decide what type of shadows to cast...
		m_iRenderType	= m_pxMaterialResponse->m_eRenderType;
	}
	else
	{
		Asura_Assert( false, "Selected material does not exist within the list." );
		m_bShouldClose = true;
	}

#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
#else
	//build list for drop down list of surface types
	m_uSurfaceTypeListSize = Tools_Texturing_MaterialResponse::GetNumSurfaceTypes();
	if(m_uSurfaceTypeListSize)
	{
		m_pszSurfaceTypes = new char*[m_uSurfaceTypeListSize];
		for(u_int i=0;i<m_uSurfaceTypeListSize;i++)
		{
			const char* szName = Tools_Texturing_MaterialResponse::GetSurfaceTypeName(i);
			Asura_Assert(szName,"Failed to get surface type name");

			m_pszSurfaceTypes[i] = new char[strlen(szName)+1];
			strcpy(m_pszSurfaceTypes[i],szName);
		}
		
	}
	else
	{
		m_uSurfaceTypeListSize = 1;
		m_pszSurfaceTypes = new char*[1];
		m_pszSurfaceTypes[0] = new char[20];
		strcpy(m_pszSurfaceTypes[0],"Default");
	}
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

	m_uID = GetOverlayIDFromIndex(iOverlayIndex);
}

Tools_MaterialProperties_Overlay::~Tools_MaterialProperties_Overlay()
{
	delete[] m_pszSlaveNames;

#if !defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
	for(u_int i=0;i<m_uSurfaceTypeListSize;i++)
	{
		delete [] m_pszSurfaceTypes[i];
	}
	delete [] m_pszSurfaceTypes;
#endif	// !defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
}

void Tools_MaterialProperties_Overlay::AddDialogs()
{
	// Labels of dialogs will be updated in UpdateDialogs()

	Asura_Dialog* pxDialog = NULL;
	Asura_Icon_Dialog* pxIconDialog = NULL;
	Asura_Graph_Dialog< float >* pxGraphDialog = NULL;
	Asura_TexturePicker_Dialog* pxTexturePickerDialog = NULL;
	Asura_Vector_2 xOffset( 1.0f, 3.0f );

	// Read-only icon
	pxIconDialog = new Asura_Icon_Dialog( xOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_PADLOCK );
	pxIconDialog->SetHelpText( "This material is read-only. Check it out before saving." );
	pxIconDialog->m_iControlID = DialogID_ReadOnlyIcon;
	AddDialog( pxIconDialog );
	xOffset.x += 2.0f;

	// Full path
	pxDialog = new Asura_Dialog( "", xOffset );
	pxDialog->m_iControlID = DialogID_MaterialName;
	AddDialog( pxDialog );
	xOffset.x -= 2.0f;
	xOffset.y += Asura_Dialog::LargeSpacing();

	// Move / rename
	pxDialog = new Asura_Function_Dialog( "Move/Rename Material", xOffset, Tools_Texturing_MaterialResponse::MoveMaterial, m_uMaterialHash );
	pxDialog->SetHelpText( "Moves this material to another folder, or renames it." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Duplicate
	pxDialog = new Asura_Function_Dialog( "Duplicate Material", xOffset, Tools_Texturing_MaterialResponse::DuplicateMaterial, m_uMaterialHash );
	pxDialog->SetHelpText( "Duplicates this material." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();

	// Parent details and slave list go in the same place, since they're mutually exclusive
	const float fSlaveStartY = xOffset.y;

	// Parent name
	pxDialog = new Asura_Dialog( "", xOffset );
	pxDialog->m_iControlID = DialogID_ParentName;
	AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	// Break from parent
	pxDialog = new Asura_Function_Dialog( "Break From Parent", xOffset, Tools_Texturing_MaterialResponse::BreakFromParent, m_uMaterialHash );
	pxDialog->SetHelpText( "Breaks this material from its parent, allowing separate response control." );
	pxDialog->m_iControlID = DialogID_BreakFromParent;
	AddDialog( pxDialog );

	xOffset.y = fSlaveStartY;

	// Add slave
	pxDialog = new Asura_Function_Dialog( "Add Slaves", xOffset, Tools_Texturing_MaterialResponse::AddSlaves, m_uMaterialHash );
	pxDialog->SetHelpText( "Sets other materials as slaves of this one." );
	pxDialog->m_iControlID = DialogID_AddSlave;
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	// Remove slave
	pxIconDialog = new Asura_Icon_Dialog( xOffset, Tools_Dialog::GetIconTexture(), TOOLS_ICON_DELETE );
	pxIconDialog->SetClickedFunction( Tools_Texturing_MaterialResponse::RemoveSlave, m_uMaterialHash );
	pxIconDialog->SetHelpText( "Breaks the connection to the selected slave." );
	pxIconDialog->m_iControlID = DialogID_RemoveSlave;
	AddDialog( pxIconDialog );
	xOffset.x += 3.0f;

	// Slave list
	m_pszSlaveNames = new const char*[1];
	m_pszSlaveNames[0] = "";
	pxDialog = new Asura_DropDown_List_Dialog( xOffset, m_pszSlaveNames, &m_iSelectedSlave, 10, 1, NULL );
	pxDialog->m_iControlID = DialogID_SlaveList;
	AddDialog( pxDialog );
	xOffset.x -= 3.0f;

	// Enslave to other - goes in the same place as slave list
	pxDialog = new Asura_Function_Dialog( "Enslave To Other", xOffset, Tools_Texturing_MaterialResponse::EnslaveToOther, m_uMaterialHash );
	pxDialog->SetHelpText( "Enslaves this material to another." );
	pxDialog->m_iControlID = DialogID_Enslave;
	AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::LargeSpacing();

	// Apply to selected polygons
	pxDialog = new Asura_Function_Dialog( "Apply As Primary Material  (G)", xOffset, Tools_Texturing_MaterialResponse::ApplyPrimaryMaterial, m_uMaterialHash );
	pxDialog->SetHelpText( "Applies this material to all selected polygons.\nThe shortcut key only works while in texturing mode." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Apply As Secondary Material", xOffset, Tools_Texturing_MaterialResponse::ApplySecondaryMaterial, m_uMaterialHash );
	pxDialog->SetHelpText( "Applies this material as the secondary material to all selected polygons." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Dialog::LargeSpacing();


	const char* aszTAB_NAMES[] =
	{
		"Diffuse",
		"Specular",
		"Self-Illumination",
		"Transparency",
		"Reflectivity",
		"Misc",
		"Special FX",
		"Overrides",
		"Project",
	};

	const u_int uNUM_TABS = sizeof( aszTAB_NAMES ) / sizeof( aszTAB_NAMES[0] );

	Asura_Tabs_Dialog* const pxTabsDialog = new Asura_Tabs_Dialog( xOffset, aszTAB_NAMES, uNUM_TABS, 45 );
	AddDialog( pxTabsDialog );

	xOffset.y += 4.0f;
	{
		pxTabsDialog->SelectTab(0);
		Asura_Vector_2 xTabOffset = xOffset;
	
		// Albedo texture
		Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::ALBEDO_TEXTURE ];
		Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::ALBEDO_TEXTURE ];
		pxDialog = new Asura_TexturePicker_Dialog( "Albedo Texture", xTabOffset, NULL, puTextureHash, pxTextureName,
												   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
												   ASURA_TEXTUREOP_VIEWTEXTURECOLOUR );

		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 10.0f;

		// Albedo detail map
		puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::DETAIL_MAP ];
		pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::DETAIL_MAP ];
		pxDialog = new Asura_TexturePicker_Dialog( "Albedo Detail Map", xTabOffset, NULL, puTextureHash, pxTextureName,
			fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
			ASURA_TEXTUREOP_VIEWTEXTUREALPHA);

		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x -= 10.0f;
		xTabOffset.y += 7.0f;

		// Albedo detail map UV scale
		pxDialog = new Asura_Numerical_Dialog( "Detail UVMap Scale", xTabOffset, &m_pxMaterialResponse->m_fAlbedoDetailMapUVScale,
			1.0f, 128.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "Sets how many times the detail map tiles to one tile of the Albedo Map." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 20.f;

		// Albedo detail map value scale
		pxDialog = new Asura_Numerical_Dialog( "Detail Value Scale", xTabOffset, &m_pxMaterialResponse->m_fAlbedoDetailMapValueScale,
			0.0f, 5.0f, 1.0f, "%.1f", &m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "Sets how much influence the detail map has." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x -= 20.f;
		xTabOffset.y += 2.0f;

		// Diffuse graph
		pxGraphDialog = new Asura_Graph_Dialog< float >( "Diffuse Graph", xTabOffset, &m_pxMaterialResponse->m_xDiffuseGraph,
														 0.0f, 100.0f, &m_pxMaterialResponse->m_bModified );

		pxGraphDialog->SetHelpText( "Specifies the diffuse response curve." );
		pxGraphDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxGraphDialog->SetShowPointControls( false );
		pxGraphDialog->SetShowDiscreteSamples( false );
		pxGraphDialog->SetGridlineSpacingY( 100.0f );
		pxTabsDialog->AddDialog( pxGraphDialog );
		xTabOffset.y += 14.0f;

		// Diffuse colour
		pxDialog = new Tools_ColourPicker_Dialog( "Diffuse Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
												  &m_pxMaterialResponse->m_xDiffuseColour.m_fR,
												  &m_pxMaterialResponse->m_xDiffuseColour.m_fG,
												  &m_pxMaterialResponse->m_xDiffuseColour.m_fB,
												  1.0f, &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Picks the diffuse colour override when colourise is non-zero." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 2.5f;

		// Colourise diffuse
		pxDialog = new Asura_Numerical_Dialog( "% Colourise Diffuse", xTabOffset, &m_pxMaterialResponse->m_fColouriseDiffuse,
											   0.0f, 100.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Sets the proportion of the picked diffuse colour to use." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
	}

	{
		pxTabsDialog->SelectTab(1);
		Asura_Vector_2 xTabOffset = xOffset;

		// Specular map
		Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::SPECULAR_MAP ];
		Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::SPECULAR_MAP ];
		pxDialog = new Asura_TexturePicker_Dialog( "Specular Map", xTabOffset, NULL, puTextureHash, pxTextureName,
												   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
												   ASURA_TEXTUREOP_VIEWTEXTUREALPHA );

		pxTabsDialog->AddDialog( pxDialog );

		const Asura_Vector_2 xOldOffset = xTabOffset;
		xTabOffset.x += 9.0f;
		xTabOffset.y += 2.0f;

		// Ignore specular prelighting
		pxDialog = new Asura_FlagTickbox_Dialog( "Ignore Specular Prelighting", xTabOffset,
			&m_pxMaterialResponse->m_uChunkedFlags,
			Tools_MaterialResponse::BITMASK_IGNORE_SPECULAR_PRELIGHTING,
			&m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "If ticked, the material will not reflect specular light from prelighting. This does not affect its response from dynamic lights." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );

		xTabOffset = xOldOffset;
		xTabOffset.y += 7.0f;

		// Metalicity
		pxDialog = new Asura_Numerical_Dialog( "% Metalicity", xTabOffset, &m_pxMaterialResponse->m_fMetalicity,
											   0.0f, 100.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Sets the proportion of the texture colour that comes through in the specular illumination." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

		// Specular graph
		pxGraphDialog = new Asura_Graph_Dialog< float >( "Specular Graph", xTabOffset, &m_pxMaterialResponse->m_xSpecularGraph,
														 0.0f, 100.0f, &m_pxMaterialResponse->m_bModified );

		pxGraphDialog->SetHelpText( "Specifies the specular response curve." );
		pxGraphDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxGraphDialog->SetShowPointControls( false );
		pxGraphDialog->SetShowDiscreteSamples( false );
		pxGraphDialog->SetGridlineSpacingY( 100.0f );
		pxTabsDialog->AddDialog( pxGraphDialog );
		xTabOffset.y += 14.0f;

		// Function to put a curve in the graph
		pxDialog = new Asura_Numerical_Dialog( "Power", xTabOffset, &Tools_Texturing_MaterialResponse::s_fSpecularSharpness, 0.0f, 200.0f, 1.0f, "%.0f");
		pxDialog->SetHelpText( "Create a curve with the specified specular power" );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 16.0f;

		// Apply
		pxDialog = new Asura_Function_Dialog( "Apply", xTabOffset, Tools_Texturing_MaterialResponse::ApplySpecularSharpness, m_uMaterialHash );
		pxDialog->SetHelpText( "Create a curve with the specified specular power" );		
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x -= 16.0f;
		xTabOffset.y += Asura_Function_Dialog::DefaultSpacing();

		// Specular colour
		pxDialog = new Tools_ColourPicker_Dialog( "Specular Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
												  &m_pxMaterialResponse->m_xSpecularColour.m_fR,
												  &m_pxMaterialResponse->m_xSpecularColour.m_fG,
												  &m_pxMaterialResponse->m_xSpecularColour.m_fB,
												  1.0f, &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Picks the specular colour override when colourise is non-zero." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 2.5f;

		// Colourise specular
		pxDialog = new Asura_Numerical_Dialog( "% Colourise Specular", xTabOffset, &m_pxMaterialResponse->m_fColouriseSpecular,
											   0.0f, 100.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Sets the proportion of the picked specular colour to use." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );

		xTabOffset.x = 1.0f;
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();
	}

	{
		pxTabsDialog->SelectTab(2);
		Asura_Vector_2 xTabOffset = xOffset;
	
		// Self illumination contribution
		pxDialog = new Asura_Numerical_Dialog( "% Self Illumination", xTabOffset,
			&m_pxMaterialResponse->m_fSelfIlluminationContribution, 0.0f, 10000.0f, 1.0f, "%.0f",
			&m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Sets the amount of self illumination to add to the overall image, in percent." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();

		// Self illumination map
		Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::SELFILLUMINATION_MAP ];
		Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::SELFILLUMINATION_MAP ];
		pxDialog = new Asura_TexturePicker_Dialog( "Self Illumination Map", xTabOffset, NULL, puTextureHash, pxTextureName,
												   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
												   ASURA_TEXTUREOP_VIEWTEXTUREALPHA );

		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += 7.0f;

		// Self illumination colour
		pxDialog = new Tools_ColourPicker_Dialog( "Self Illumination Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
												  &m_pxMaterialResponse->m_xSelfIlluminationColour.m_fR,
												  &m_pxMaterialResponse->m_xSelfIlluminationColour.m_fG,
												  &m_pxMaterialResponse->m_xSelfIlluminationColour.m_fB,
												  1.0f, &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Picks the self illumination colour override when colourise is non-zero." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 2.5f;

		// Colourise self illumination
		pxDialog = new Asura_Numerical_Dialog( "% Colourise Self Illumination", xTabOffset, &m_pxMaterialResponse->m_fColouriseSelfIllumination,
											   0.0f, 100.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Sets the proportion of the picked self illumination colour to use." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );

		xTabOffset.x = 1.0f;
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();
	}

	{
		pxTabsDialog->SelectTab(3);
		Asura_Vector_2 xTabOffset = xOffset;
	
		// Blend mode
		const u_int uNumberOfBlendModes = Tools_MaterialResponse::GetNumberOfBlendModes();
		pxDialog = new Asura_DropDown_List_Dialog( xTabOffset, Tools_MaterialResponse::GetBlendModeTextStrings(),
												   (int*)( &m_pxMaterialResponse->m_eBlendMode ), uNumberOfBlendModes,
												   uNumberOfBlendModes, NULL, &m_pxMaterialResponse->m_bModified,
												   "Blend Mode" );

		pxDialog->SetHelpText( "The blending mode to use when rendering." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();

		// Transparency map
		Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::TRANSPARENCY_MAP ];
		Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::TRANSPARENCY_MAP ];
		pxDialog = new Asura_TexturePicker_Dialog( "Transparency Map", xTabOffset, NULL, puTextureHash, pxTextureName,
												   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
												   ASURA_TEXTUREOP_VIEWTEXTUREALPHA );

		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 15.0f;

		// Decal flag
		pxDialog = new Asura_FlagTickbox_Dialog( "Decal", xTabOffset,
			&m_pxMaterialResponse->m_uChunkedFlags, Tools_MaterialResponse::BITMASK_DECAL,
			&m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "If ticked, the material will display on top of other geometry. It won't be lit separately." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );

		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();;

		// ZPrePass
		pxDialog = new Asura_FlagTickbox_Dialog( "Z PrePass", xTabOffset,
			&m_pxMaterialResponse->m_uChunkedFlags, Tools_MaterialResponse::BITMASK_ZPREPASS,
			&m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "If ticked, the material will cause a Z only prepass meaning you cant see parts of the same model through itself." );
		pxDialog->SetIsEnabledFunction( EnableZPrePassControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );

		xTabOffset.x -= 15.0f;
		xTabOffset.y += 7.0f - Asura_Numerical_Dialog::LargeSpacing();

		// Transparency graph
		pxGraphDialog = new Asura_Graph_Dialog< float >( "Transparency Graph", xTabOffset, &m_pxMaterialResponse->m_xTransparencyGraph,
														 0.0f, 100.0f, &m_pxMaterialResponse->m_bModified );

		pxGraphDialog->SetHelpText( "Specifies the transparency response curve. Lower values mean more transparent." );
		pxGraphDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxGraphDialog->SetShowPointControls( false );
		pxGraphDialog->SetShowDiscreteSamples( false );
		pxTabsDialog->AddDialog( pxGraphDialog );
		xTabOffset.y += 14.0f + Asura_Dialog::SmallSpacing();
	}
	
	{
		pxTabsDialog->SelectTab(4);
		Asura_Vector_2 xTabOffset = xOffset;
	
		// Reflectivity
		pxDialog = new Asura_Numerical_Dialog( "% Reflectivity", xTabOffset, &m_pxMaterialResponse->m_fReflectivity, 0.0f, 100.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "Specifies the reflectivity scale for the whole material. Modulates the reflectivity mask texture. Zero means no reflections." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();

		// Reflection mask
		Asura_Hash_ID* puReflectivityTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::REFLECTIVITY_MASK ];
		Asura_String* pxReflectivityTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::REFLECTIVITY_MASK ];
		pxDialog = new Asura_TexturePicker_Dialog( "Reflectivity Mask", xTabOffset, NULL, puReflectivityTextureHash, pxReflectivityTextureName,
												   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
												   ASURA_TEXTUREOP_VIEWTEXTUREALPHA );

		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 15.0f;

		// Environment map
		Asura_Hash_ID* puEnvironmentMapHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::ENVIRONMENT_MAP ];
		Asura_String* pxEnvironmentMapName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::ENVIRONMENT_MAP ];
		pxTexturePickerDialog = new Asura_TexturePicker_Dialog( "Environment Map", xTabOffset, NULL,
																puEnvironmentMapHash,
																pxEnvironmentMapName,
																fTEXTURE_PICKER_SIZE,
																fTEXTURE_PICKER_SIZE,
																&m_pxMaterialResponse->m_bModified,
																ASURA_TEXTUREOP_VIEWTEXTURECOLOUR );

		// Cubemap options will be set as appropriate in PerFrameUpdate()
		pxTexturePickerDialog->SetControlID( DialogID_EnvironmentMap );
		pxTabsDialog->AddDialog( pxTexturePickerDialog );

		xTabOffset.y += 14.0f + Asura_Dialog::LargeSpacing();
	}

	{
		pxTabsDialog->SelectTab(5);
		Asura_Vector_2 xTabOffset = xOffset;

		// Copy from other
		pxDialog = new Asura_Function_Dialog( "Copy Properties From Other Material", xTabOffset, Tools_Texturing_MaterialResponse::CopyPropertiesFromOther, m_uMaterialHash );
		pxDialog->SetHelpText( "Copies selected properties from another material." );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Function_Dialog::DefaultSpacing();

		// Copy to other
		pxDialog = new Asura_Function_Dialog( "Copy Properties To Other Material", xTabOffset, Tools_Texturing_MaterialResponse::CopyPropertiesToOther, m_uMaterialHash );
		pxDialog->SetHelpText( "Copies selected properties to another material." );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Function_Dialog::LargeSpacing();

		{
			// Normal map
			Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::NORMAL_MAP ];
			Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::NORMAL_MAP ];
			pxDialog = new Asura_TexturePicker_Dialog( "Normal Map", xTabOffset, NULL, puTextureHash, pxTextureName,
													   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
													   ASURA_TEXTUREOP_VIEWTEXTURECOLOUR );

			pxTabsDialog->AddDialog( pxDialog );
		}

		xTabOffset.x += 10.0f;
		{
			// Normal detail map
			Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::DETAIL_NORMAL_MAP ];
			Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::DETAIL_NORMAL_MAP ];
			pxDialog = new Asura_TexturePicker_Dialog( "Normal Detail Map", xTabOffset, NULL, puTextureHash, pxTextureName,
				fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
				ASURA_TEXTUREOP_VIEWTEXTURECOLOUR );

			pxTabsDialog->AddDialog( pxDialog );
		}

		xTabOffset.x += 15.0f;
		
		{
			// Parallax map
			Asura_Hash_ID* puTextureHash = &m_pxMaterialResponse->m_auTextureHashIDs[ Asura_MaterialResponse::PARALLAX_MAP ];
			Asura_String* pxTextureName = &m_pxMaterialResponse->m_axTextureFilenames[ Asura_MaterialResponse::PARALLAX_MAP ];
			pxDialog = new Asura_TexturePicker_Dialog( "Parallax Map", xTabOffset, NULL, puTextureHash, pxTextureName,
													   fTEXTURE_PICKER_SIZE, fTEXTURE_PICKER_SIZE, &m_pxMaterialResponse->m_bModified,
													   ASURA_TEXTUREOP_VIEWTEXTUREALPHA );

			pxTabsDialog->AddDialog( pxDialog );
			xTabOffset.y += 7.0f;
		}

		xTabOffset.x -= 15.0f;
		xTabOffset.x -= 10.0f;

		// Normal detail map UV scale
		pxDialog = new Asura_Numerical_Dialog( "Detail UVMap Scale", xTabOffset, &m_pxMaterialResponse->m_fNormalDetailMapUVScale,
			1.0f, 128.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "Sets how many times the detail map tiles to one tile of the Normal Map." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x += 20.f;

		// Normal detail map value scale
		pxDialog = new Asura_Numerical_Dialog( "Detail Value Scale", xTabOffset, &m_pxMaterialResponse->m_fNormalDetailMapValueScale,
			0.0f, 5.0f, 1.0f, "%.1f", &m_pxMaterialResponse->m_bModified );
		pxDialog->SetHelpText( "Sets how much influence the detail map has." );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.x -= 20.f;
		xTabOffset.y += 2.0f;

		// Parallax scale
		pxDialog = new Asura_Numerical_Dialog( "Parallax Scale", xTabOffset, &m_pxMaterialResponse->m_fParallaxScale,
											   0.0f, 10.0f, 0.1f, "%.1f", &m_pxMaterialResponse->m_bModified );

		pxDialog->SetHelpText( "Sets the scale of the parallax mapping effect. About 4 should give good results." );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();

		{
			// Ambient contribution
			pxDialog = new Asura_Numerical_Dialog( "% Prelighting Contribution", xTabOffset, &m_pxMaterialResponse->m_fAmbientContribution,
												0.0f, 200.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );

			pxDialog->SetHelpText( "Sets the contribution of the prelighting to the overall image, in percent." );
			pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
			pxTabsDialog->AddDialog( pxDialog );
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

			// Ambient colour
			pxDialog = new Tools_ColourPicker_Dialog( "Prelighting Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
													&m_pxMaterialResponse->m_xAmbientColour.m_fR,
													&m_pxMaterialResponse->m_xAmbientColour.m_fG,
													&m_pxMaterialResponse->m_xAmbientColour.m_fB,
													1.0f, &m_pxMaterialResponse->m_bModified );

			pxDialog->SetHelpText( "Picks the prelighting colour override when colourise is non-zero." );
			pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
			pxTabsDialog->AddDialog( pxDialog );
			xTabOffset.x += 2.5f;

			// Colourise ambient
			pxDialog = new Asura_Numerical_Dialog( "% Colourise Prelighting", xTabOffset, &m_pxMaterialResponse->m_fColouriseAmbient,
												0.0f, 100.0f, 1.0f, "%.0f", &m_pxMaterialResponse->m_bModified );

			pxDialog->SetHelpText( "Sets the proportion of the picked prelighting colour to use." );
			pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
			pxTabsDialog->AddDialog( pxDialog );

			xTabOffset.x = 1.0f;
			xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();
		}

		Asura_Hash_ID* puTexAnimHash = &m_pxMaterialResponse->m_uTextureAnimHashID;
		pxDialog = new Tools_TextureAnim_DropDownList_Dialog( xTabOffset, 10, puTexAnimHash, &m_pxMaterialResponse->m_bModified, "Texture Anim" );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Dialog::LargeSpacing();

		//add surface type drop down list
#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
		Asura_DropDown_List_Dialog*	pxDropDownListDialog	= Tools_Texturing_MaterialResponse::Project_CreateSurfaceTypeDialog(xTabOffset, 12, &m_pxMaterialResponse->m_uSurfaceTypeHash, NULL, "Surface Type");

		Tools_Texturing_MaterialResponse::Project_ValidateSurfaceTypeHash(m_pxMaterialResponse->m_uSurfaceTypeHash);
#else
		if(m_pxMaterialResponse->m_uSurfaceType >= m_uSurfaceTypeListSize)
		{
			Asura_Assertf(0,("Unknown surface type %d.\nResetting to 0",m_pxMaterialResponse->m_uSurfaceType));
			m_pxMaterialResponse->m_uSurfaceType = 0;
		}

		Asura_DropDown_List_Dialog* pxDropDownListDialog = new Asura_DropDown_List_Dialog(xTabOffset, m_pszSurfaceTypes, (int*)&m_pxMaterialResponse->m_uSurfaceType, 12, (int)m_uSurfaceTypeListSize, NULL, NULL, "Surface Type");
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)

		if(pxDropDownListDialog != NULL)
		{
			pxDropDownListDialog->SetOnSelectCallback(SurfaceTypeChanged,(int)this);
			pxDropDownListDialog->SetOnlyUpdateSelectionIndexOnClick(true);
			pxDropDownListDialog->SetHelpText( "Project defined surface type." );
			pxTabsDialog->AddDialog( pxDropDownListDialog );

			xTabOffset.x = pxDropDownListDialog->m_xMaxCoord.x + 2.0f;
			pxDialog = new Asura_Tickbox_Dialog("Apply by texture", xTabOffset, &s_bSetSurfaceTypeByTexture);
			pxDialog->SetHelpText("If ticked ,then changes to the surface type will be copied to all materials that share the same primary albedo/diffuse texture.");
			pxTabsDialog->AddDialog( pxDialog );
			xTabOffset.x = pxDropDownListDialog->m_xOffset.x;
			xTabOffset.y += Asura_Dialog::LargeSpacing();
		}

		pxDropDownListDialog = new Asura_DropDown_List_Dialog( xTabOffset, s_aszRenderTypeNames, &m_iRenderType, Asura_MaterialResponse::NUMBER_OF_RENDER_TYPES, Asura_MaterialResponse::NUMBER_OF_RENDER_TYPES, NULL, &m_pxMaterialResponse->m_bModified, "Render Type");
		pxDropDownListDialog->SetOnSelectCallback(RenderTypeChanged,(int)this);
		pxDropDownListDialog->SetOnlyUpdateSelectionIndexOnClick(true);
		pxDropDownListDialog->SetHelpText( "Specifies how this material is rendered (with respect to shadowing)." );
		pxDropDownListDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDropDownListDialog );
		xTabOffset.y += Asura_Tickbox_Dialog::DefaultSpacing();
	}

	{
		pxTabsDialog->SelectTab(6);
		Asura_Vector_2 xTabOffset = xOffset;

		// SFX mode
		const u_int uNumberOfSpecialFXModes = Tools_MaterialResponse::GetNumberOfSpecialFXModes();

		pxDialog = new Asura_DropDown_List_Dialog( xTabOffset, Tools_MaterialResponse::GetSpecialFXModeTextStrings(),
			(int*)( &m_pxMaterialResponse->m_eSpecialFXMode ), uNumberOfSpecialFXModes, uNumberOfSpecialFXModes, NULL, &m_pxMaterialResponse->m_bModified, "Special FX Mode");

		pxDialog->SetHelpText( "The special fx mode to use" );
		pxDialog->SetIsEnabledFunction( EnableResponseControls, m_uMaterialHash );
		pxTabsDialog->AddDialog( pxDialog );
		xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();

		// Add the controls for each SFX type. I would much prefer to only show the controls for the current effect type, but can't
		// figure out a sensible way to make this particular overlay destroy/recreate itself (the only way of adding/removing dialogs?) 
		// when the dropdown is changed. So it's going to have to be another set of tabs.
		//
		// We've only got one SFX type for now, water, so it doesn't matter much for now.

		const char* aszSFX_TAB_NAMES[] =
		{
			"Water"			
		};

		const u_int uNUM_SFX_TABS = sizeof( aszSFX_TAB_NAMES ) / sizeof( aszSFX_TAB_NAMES[0] );

		Asura_Tabs_Dialog* const pxSubTabsDialog = new Asura_Tabs_Dialog( xTabOffset, aszSFX_TAB_NAMES, uNUM_SFX_TABS, 38 );
		pxTabsDialog->AddDialog(pxSubTabsDialog);

		// Add the water parameters - Subtab 0
		pxSubTabsDialog->SelectTab(0);

		{
			Tools_ColourPicker_Dialog*	pxColourPickerDialog;

			
			xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();		

			// Reflection colour
			pxColourPickerDialog = new Tools_ColourPicker_Dialog( "Reflection Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
														&m_pxMaterialResponse->m_xWaterReflectionColour.m_fR,
														&m_pxMaterialResponse->m_xWaterReflectionColour.m_fG,
														&m_pxMaterialResponse->m_xWaterReflectionColour.m_fB,
														1.0f, &m_pxMaterialResponse->m_bModified );		
#if defined(ASURA_USE_HDR)
			pxColourPickerDialog->EnableHDR(true);
#endif	//defined(ASURA_USE_HDR)
			pxSubTabsDialog->AddDialog(pxColourPickerDialog);
			pxDialog = new Asura_Dialog( "Reflection Colour", xTabOffset + Asura_Vector_2(4,0));
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

			// Refraction colour
			pxColourPickerDialog = new Tools_ColourPicker_Dialog( "Refraction Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
														&m_pxMaterialResponse->m_xWaterRefractionColour.m_fR,
														&m_pxMaterialResponse->m_xWaterRefractionColour.m_fG,
														&m_pxMaterialResponse->m_xWaterRefractionColour.m_fB,
														1.0f, &m_pxMaterialResponse->m_bModified );		
#if defined(ASURA_USE_HDR)
			pxColourPickerDialog->EnableHDR(true);
#endif	//defined(ASURA_USE_HDR)
			pxSubTabsDialog->AddDialog(pxColourPickerDialog);
			pxDialog = new Asura_Dialog( "Refraction Colour", xTabOffset + Asura_Vector_2(4,0));
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

			// Fog colour
			pxColourPickerDialog = new Tools_ColourPicker_Dialog( "Fogging Colour", xTabOffset, Asura_Vector_2( 2.0f, 1.0f ),
														&m_pxMaterialResponse->m_xWaterFogColour.m_fR,
														&m_pxMaterialResponse->m_xWaterFogColour.m_fG,
														&m_pxMaterialResponse->m_xWaterFogColour.m_fB,
														1.0f, &m_pxMaterialResponse->m_bModified );	
#if defined(ASURA_USE_HDR)
			pxColourPickerDialog->EnableHDR(true);
#endif	//defined(ASURA_USE_HDR)
			pxSubTabsDialog->AddDialog(pxColourPickerDialog);
			pxDialog = new Asura_Dialog( "Fogging Colour", xTabOffset + Asura_Vector_2(4,0));
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			// Ripple Scale
			xTabOffset.x +=20;
			pxDialog = new Asura_Numerical_Dialog("Ripple Scale", xTabOffset, &m_pxMaterialResponse->m_fWaterRippleScale, 0.0f, 1.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

			// Distortion Scale			
			pxDialog = new Asura_Numerical_Dialog("Distortion Scale", xTabOffset, &m_pxMaterialResponse->m_fWaterDistortionScale, 0.0f, 5.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

			// Edge Softness
			pxDialog = new Asura_Numerical_Dialog("Edge Softness", xTabOffset, &m_pxMaterialResponse->m_fWaterEdgeSoftness, 0.01f, 10.0f, 0.01f, "%0.02f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

			// Fog Strength		
			pxDialog = new Asura_Numerical_Dialog("Fog Strength", xTabOffset, &m_pxMaterialResponse->m_fWaterFogStrength, 0.0f, 10.0f, 0.01f, "%0.02f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();

		
			xTabOffset.x -=20;
			xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();
			

			// UV scroll/scale control
			pxDialog = new Asura_Dialog( "UV Scrolling", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Dialog( "Layer 0", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Dialog( "Layer 1", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);

			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			xTabOffset.x+=10;
			pxDialog = new Asura_Dialog( "   U Speed", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterUVControl[0], -9.99f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterUVControl[3], -9.99f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			xTabOffset.x+=10;
			pxDialog = new Asura_Dialog( "   V Speed", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterUVControl[1], -9.99f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterUVControl[4], -9.99f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			xTabOffset.x+=10;
			pxDialog = new Asura_Dialog( "   UV Scale", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterUVControl[2], 0.1f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterUVControl[5], 0.1f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();
			xTabOffset.x -=30;


			// Wave control
			pxDialog = new Asura_Dialog( "Vertex Waves", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Dialog( "X Axis", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Dialog( "Z Axis", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);

			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			xTabOffset.x+=10;
			pxDialog = new Asura_Dialog( "   Amplitude", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterWaveControl[0], -9.99f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterWaveControl[3], -9.99f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			xTabOffset.x+=10;
			pxDialog = new Asura_Dialog( "   Frequency", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterWaveControl[1], -99.99f, 100.0f, 0.01f, "%0.02f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterWaveControl[4], -99.99f, 100.0f, 0.01f, "%0.02f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y -= 2*Asura_Numerical_Dialog::DefaultSpacing();

			xTabOffset.x+=10;
			pxDialog = new Asura_Dialog( "   Speed", xTabOffset);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterWaveControl[2], -9.9f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::DefaultSpacing();
			pxDialog = new Asura_Numerical_Dialog("", xTabOffset, &m_pxMaterialResponse->m_afWaterWaveControl[5], -9.9f, 10.0f, 0.01f, "%0.03f", &m_pxMaterialResponse->m_bModified);
			pxSubTabsDialog->AddDialog(pxDialog);
			xTabOffset.y += Asura_Numerical_Dialog::LargeSpacing();			
		}
	}
	{
		pxTabsDialog->SelectTab( 7 );
		Asura_Vector_2 xTabOffset = xOffset;
		Asura_Vector_2 xTableSize( 43.0f, 7.5f );

		m_xMaterialResponseOverrides.TabsDialog( m_pxMaterialResponse->GetName().GetString(), pxTabsDialog, xTabOffset, xTableSize );
	}

	if (Tools_Texturing_MaterialResponse::GetProjectUICallback())
	{
		pxTabsDialog->SelectTab( 8 );
		Asura_Vector_2 xTabOffset = xOffset;

		Tools_Texturing_MaterialResponse::GetProjectUICallback()(m_pxMaterialResponse, pxTabsDialog, xOffset);
	}


	pxTabsDialog->SelectTab(0);

	xOffset.y = pxTabsDialog->GetMaxCoords().y + Asura_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Revert Changes (Except Slaves)", xOffset, RevertChanges, reinterpret_cast< int >( this ) );
	pxDialog->SetHelpText( "Reverts all changes to this material since this overlay was opened. Does not free any slaves that were added." );
	AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	UpdateDialogs();
}

static int CompareStrings( const void* pData1, const void* pData2)
{
	// The data are pointers to string pointers
	const char* const szString1 = *static_cast< const char* const* >( pData1 );
	const char* const szString2 = *static_cast< const char* const* >( pData2 );

	return _stricmp( szString1, szString2 );
}

void Tools_MaterialProperties_Overlay::UpdateDialogs()
{
	if( !m_pxMaterialResponse ) return;

	// Update read-only icon
	{
		Asura_String xMaterialPath;
		Tools_MaterialResponse_System::MakePathFromMaterialName( m_pxMaterialResponse->GetName(), xMaterialPath );

		const bool bReadOnly = Tools_FileSystem::IsReadOnly( xMaterialPath.GetString() );

		Asura_Dialog* const pxReadOnlyIcon = GetDialog( DialogID_ReadOnlyIcon );
		if( pxReadOnlyIcon )
		{
			pxReadOnlyIcon->SetHidden( !bReadOnly );
		}
	}

	// Names might have changed through being moved
	{
		Asura_String xFilename;
		m_pxMaterialResponse->GetName().SplitFilename( NULL, &xFilename, NULL );
		SetTitle( xFilename.GetString() );

		UpdateDialogLabel( DialogID_MaterialName, m_pxMaterialResponse->GetName().GetString() );
	}

	// Calculate the slave list

	const Asura_Hash_ID uMaterialHash = m_pxMaterialResponse->GetHash();
	Asura_List< const Tools_MaterialResponse > xSlaveList;

	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading(u);
		if ( !pxCurrMaterial->IsLoaded() ) continue;

		if( pxCurrMaterial->GetParentHash() == uMaterialHash )
		{
			xSlaveList.AddToList( pxCurrMaterial );
		}
	}

	const int iNumSlaves = xSlaveList.GetNumEntries();

	// Update parent/slave dialogs

	Asura_Dialog* const pxParentNameDialog = GetDialog( DialogID_ParentName );
	Asura_Dialog* const pxBreakFromParentDialog = GetDialog( DialogID_BreakFromParent );
	Asura_Dialog* const pxAddSlaveDialog = GetDialog( DialogID_AddSlave );
	Asura_Dialog* const pxRemoveSlaveDialog = GetDialog( DialogID_RemoveSlave );
	Asura_DropDown_List_Dialog* const pxSlaveListDialog = static_cast< Asura_DropDown_List_Dialog* >( GetDialog( DialogID_SlaveList ) );
	Asura_Dialog* const pxEnslaveDialog = GetDialog( DialogID_Enslave );

	const Tools_MaterialResponse* const pxParentMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( m_pxMaterialResponse->GetParentHash() );

	if( pxParentMaterial )
	{
		char szParentString[100];
		Asura_PrintfToBfr( szParentString, "Parent: %s", 100, true, pxParentMaterial->GetName().GetString() );
		pxParentNameDialog->UpdateLabel( szParentString );

		pxParentNameDialog->SetHidden( false );
		pxBreakFromParentDialog->SetHidden( false );

		pxAddSlaveDialog->SetHidden( true );
		pxRemoveSlaveDialog->SetHidden( true );
		pxSlaveListDialog->SetHidden( true );
		pxEnslaveDialog->SetHidden( true );
	}
	else if( iNumSlaves > 0 )
	{
		// Update slave list

		delete[] m_pszSlaveNames;

		m_pszSlaveNames = new const char*[ iNumSlaves ];

		u_int uCurrSlave = 0;

		for( Asura_List_ConstIt< const Tools_MaterialResponse > xIter( &xSlaveList ); !xIter.Done(); xIter.Next() )
		{
			const Tools_MaterialResponse* const pxCurrSlave = xIter.GetCurrent();
			m_pszSlaveNames[ uCurrSlave++ ] = pxCurrSlave->GetName().GetString();
		}

		qsort( m_pszSlaveNames, iNumSlaves, sizeof( const char* ), CompareStrings );

		pxSlaveListDialog->ReplaceStrings( iNumSlaves, m_pszSlaveNames );

		pxParentNameDialog->SetHidden( true );
		pxBreakFromParentDialog->SetHidden( true );

		pxAddSlaveDialog->SetHidden( false );
		pxRemoveSlaveDialog->SetHidden( false );
		pxSlaveListDialog->SetHidden( false );
		pxEnslaveDialog->SetHidden( true );

		m_iSelectedSlave = 0;
	}
	else
	{
		pxParentNameDialog->SetHidden( true );
		pxBreakFromParentDialog->SetHidden( true );

		pxAddSlaveDialog->SetHidden( false );
		pxRemoveSlaveDialog->SetHidden( true );
		pxSlaveListDialog->SetHidden( true );
		pxEnslaveDialog->SetHidden( false );
	}

	ResizeOverlay();
}


/******************************************************************************/
/* Tools_MaterialProperties_Overlay Static Class Functions                    */
/******************************************************************************/

bool Tools_MaterialProperties_Overlay::EnableResponseControls( int iMaterialHash )
{
	const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( iMaterialHash );
	if( !pxMaterial ) return true;

	return pxMaterial->GetParentHash() == ASURA_HASH_ID_UNSET;
}

bool Tools_MaterialProperties_Overlay::EnableZPrePassControls( int iMaterialHash )
{
	const Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( iMaterialHash );
	if( !pxMaterial ) return true;

	return pxMaterial->GetParentHash() == ASURA_HASH_ID_UNSET && pxMaterial->GetBlendMode() != Asura_MaterialResponse::BLEND_MODE_OPAQUE;

}


void Tools_MaterialProperties_Overlay::RevertChanges( int iUserData )
{
	Tools_MaterialProperties_Overlay* const pxOverlay = reinterpret_cast< Tools_MaterialProperties_Overlay* >( iUserData );
	if( !pxOverlay )
	{
		Asura_Assert( false, "Invalid parameter passed to Tools_MaterialProperties_Overlay::Revert()." );
		return;
	}

	*pxOverlay->m_pxMaterialResponse = pxOverlay->m_xInitialProperties;
	pxOverlay->m_pxMaterialResponse->m_bModified = true;

	// Note we don't know which slaves we've added this editing session, so can't undo that
	// change here. Probably not worth keeping track of this, since it's easy to fix up

	pxOverlay->UpdateDialogs();
}

void Tools_MaterialProperties_Overlay::RenderTypeChanged(int iThis)
{
	Tools_MaterialProperties_Overlay*	pxOverlay		= (Tools_MaterialProperties_Overlay*) iThis;
	Tools_MaterialResponse*				pxThisMaterial	= pxOverlay->m_pxMaterialResponse;

	pxThisMaterial->m_eRenderType	= static_cast<Asura_MaterialResponse::RENDER_TYPE>(pxOverlay->m_iRenderType);
	pxThisMaterial->m_bModified		= true;
}

void Tools_MaterialProperties_Overlay::SurfaceTypeChanged(int iThis)
{
	Tools_MaterialProperties_Overlay* pxOverlay = (Tools_MaterialProperties_Overlay*) iThis;

	Tools_MaterialResponse* pxThisMaterial = pxOverlay->m_pxMaterialResponse;
	pxThisMaterial->m_bModified = true;

	if(s_bSetSurfaceTypeByTexture)
	{
		const char* szThisTexture = pxThisMaterial->GetTextureFilename(Asura_MaterialResponse::ALBEDO_TEXTURE);
		if(!szThisTexture) szThisTexture = "";

		//search for all materials with the same albedo texture
		for(u_int i=0; i<Tools_MaterialResponse_System::GetNumMaterials(); i++)
		{
			Tools_MaterialResponse* const pxOtherMaterial = Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading(i);
			if ( !pxOtherMaterial->IsLoaded() ) continue;
			
			if(pxOtherMaterial == pxThisMaterial) continue;

			const char* szOtherTexture = pxOtherMaterial->GetTextureFilename(Asura_MaterialResponse::ALBEDO_TEXTURE);
			if(!szOtherTexture) szOtherTexture = "";

			if(!stricmp(szThisTexture,szOtherTexture))
			{
#if defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
				pxOtherMaterial->m_uSurfaceTypeHash	= pxThisMaterial->m_uSurfaceTypeHash;
#else
				pxOtherMaterial->m_uSurfaceType = pxThisMaterial->m_uSurfaceType;
#endif	// defined(ASURA_MATERIAL_RESPONSE_SURFACES_USE_HASH_IDS)
				pxOtherMaterial->m_bModified = true;	
			}
		}
	}
}

void Tools_MaterialProperties_Overlay::RemoveClosedOverlaysFromList()
{
	for(LIF<int> xIndexLIF(&s_xOverlayIndexList); !xIndexLIF.done();)
	{
		if(Asura_Overlay_System::GetOverlay(GetOverlayIDFromIndex(xIndexLIF())))		
		{
			xIndexLIF.next();
		}
		else
		{
			xIndexLIF.delete_current();
		}
	}
}

int Tools_MaterialProperties_Overlay::GetOverlayIDFromIndex(int iIndex)
{
	Asura_Assert(iIndex>=0 && iIndex < iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY,"Invalid index");
	return TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_PROPERTIES1 + iIndex;
}

void Tools_MaterialProperties_Overlay::EditMaterial(const Asura_Hash_ID uMaterialHash,int iOverlayIndex /*=-1*/)
{
	//If an overlay index has been provided then close that one first
	if(iOverlayIndex != -1)
	{
		Asura_Overlay_System::CloseOverlays(GetOverlayIDFromIndex(iOverlayIndex));
	}


	//Check for an existing overlay for editing this material
	Tools_MaterialProperties_Overlay* pxOverlay = FindOverlayForMaterial(uMaterialHash);
	if(pxOverlay)
	{
		//just bring it to the foreground
		Asura_Overlay_System::BringOverlayToTop(pxOverlay->m_uID);
		//move the overlay index to the start of the list (will get done during next update anyway , but may as
		//well ensure it gets done now to be safe).
		if(s_xOverlayIndexList.contains(pxOverlay->m_iOverlayIndex)) s_xOverlayIndexList.delete_entry(pxOverlay->m_iOverlayIndex);
		s_xOverlayIndexList.add_entry_start(pxOverlay->m_iOverlayIndex);
		return;
	}

	//Will need to create an overlay , but just confirm the material exists first
	Tools_MaterialResponse* const pxSelectedMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash );
	if( !pxSelectedMaterial )
	{
		Asura_Assertf( false, ( "Selected material 0x%08X doesn't exist in the list.", uMaterialHash ) );
		return;
	}

	//Check to see whether any overlays need to be closed to make room
	RemoveClosedOverlaysFromList();
	if(iOverlayIndex == -1)
	{
		if(s_xOverlayIndexList.size() >= s_iMaxOpenOverlay)
		{
			int iOverlayID = GetOverlayIDFromIndex(s_xOverlayIndexList.last_entry());
			Asura_Overlay_System::CloseOverlays(iOverlayID);
			s_xOverlayIndexList.delete_last_entry();
		}
	

		//Find the first free overlay index
		for(iOverlayIndex = 0; iOverlayIndex<iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY ; iOverlayIndex++)
		{
			if(!s_xOverlayIndexList.contains(iOverlayIndex))
			{
				break;
			}
		}
	}
	Asura_Assert(iOverlayIndex<iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY, "Failed to find free overlay index??");

	// Create the overlay
	Asura_String xOverlayCaption;
	pxSelectedMaterial->GetName().SplitFilename( NULL, &xOverlayCaption, NULL );
	xOverlayCaption += " Properties";

	pxOverlay = new Tools_MaterialProperties_Overlay( xOverlayCaption.GetString(), uMaterialHash, iOverlayIndex );
	if( !pxOverlay )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Failed to create material response properties overlay." );
		return;
	}

	// Add controls
	pxOverlay->AddDialogs();

	// Submit to the interface system
	Asura_Overlay_System::SubmitNewOverlay( pxOverlay );
		
}

Tools_MaterialProperties_Overlay* Tools_MaterialProperties_Overlay::FindOverlayForMaterial(Asura_Hash_ID uMaterialHash)
{
	for(LIF<int> xIndexLIF(&s_xOverlayIndexList); !xIndexLIF.done(); xIndexLIF.next())
	{
		Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayIDFromIndex(xIndexLIF()));
		if(!pxOverlay) continue;

		Tools_MaterialProperties_Overlay* pxMaterialOverlay = (Tools_MaterialProperties_Overlay*) pxOverlay;
		if(pxMaterialOverlay->m_uMaterialHash == uMaterialHash)
		{
			return pxMaterialOverlay;
		}
	}
	return 0;
}

Tools_MaterialProperties_Overlay* Tools_MaterialProperties_Overlay::GetMostRecentlySelectedOverlay()
{
	for(LIF<int> xIndexLIF(&s_xOverlayIndexList); !xIndexLIF.done(); xIndexLIF.next())
	{
		Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayIDFromIndex(xIndexLIF()));
		if(!pxOverlay) continue;

		return (Tools_MaterialProperties_Overlay*) pxOverlay;
	}
	return 0;
}


void Tools_MaterialProperties_Overlay::CloseOverlays()
{
	while(s_xOverlayIndexList.size())
	{
		int iOverlayID = GetOverlayIDFromIndex(s_xOverlayIndexList.last_entry());
		Asura_Overlay_System::CloseOverlays(iOverlayID);
		s_xOverlayIndexList.delete_last_entry();
	}
}

void Tools_MaterialProperties_Overlay::UpdateAllOverlayDialogs()
{
	for(LIF<int> xIndexLIF(&s_xOverlayIndexList); !xIndexLIF.done(); xIndexLIF.next())
	{
		Asura_Overlay* pxOverlay = Asura_Overlay_System::GetOverlay(GetOverlayIDFromIndex(xIndexLIF()));
		if(!pxOverlay) continue;

		((Tools_MaterialProperties_Overlay*)pxOverlay)->UpdateDialogs();
	}
}

void Tools_MaterialProperties_Overlay::PerFrameUpdate()
{
	Asura_Overlay::PerFrameUpdate();

	if(m_bSortToFront)
	{
		//move to the beginning of our list of overlays
		if(s_xOverlayIndexList.contains(m_iOverlayIndex)) s_xOverlayIndexList.delete_entry(m_iOverlayIndex);
		s_xOverlayIndexList.add_entry_start(m_iOverlayIndex);
	}

	if( m_pxMaterialResponse )
	{
		Asura_TexturePicker_Dialog* const pxEnvMapPicker
			= static_cast< Asura_TexturePicker_Dialog* >( GetDialog( DialogID_EnvironmentMap ) );

		const Asura_MaterialResponse::SPECIALFX_MODE eSFXMode
			= m_pxMaterialResponse->GetSpecialFXMode();

		if( eSFXMode  == Asura_MaterialResponse::SPECIALFX_MODE_WATER )
		{
			// Should really be all textures except cubemaps... this interface is getting a bit
			// complex though, since it mixes texture type with old material usage
			pxEnvMapPicker->SetPickerFlags( PICKTEXTURE_ALLTEXTURES );
			pxEnvMapPicker->SetValidateCubemap( false );
		}
		else
		{
			pxEnvMapPicker->SetPickerFlags( PICKTEXTURE_CUBEMAP );
			pxEnvMapPicker->SetValidateCubemap( true );
		}
	}
}


/******************************************************************************/
/* Tools_CopyMaterialProperties_Overlay Class Functions                       */
/******************************************************************************/

Tools_CopyMaterialProperties_Overlay::Tools_CopyMaterialProperties_Overlay( const Asura_Hash_ID uInitialCopyFromHash, const Asura_Hash_ID uInitialCopyToHash )
	: Tools_Modal_Overlay( "Copy Material Properties" )
	, m_uFromHash( uInitialCopyFromHash )
	, m_uToHash( uInitialCopyToHash )
	, m_pxFromMaterialDialog(0)
	, m_pxToMaterialDialog(0)
{
	m_xCopyFlags.m_uAllFlags = 0;
}


void Tools_CopyMaterialProperties_Overlay::Update()
{
	Asura_Overlay* const pxOverlay = GetOverlay();
	Asura_Assert( pxOverlay, "Copy Material Properties has lost its overlay." );
	if( !pxOverlay ) return;

	// Set subcontrols' enabled state based on parents

	Asura_Dialog* pxDialog;

	// Responses

	pxDialog = pxOverlay->GetDialog( DialogID_Responses );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything );

	pxDialog = pxOverlay->GetDialog( DialogID_AmbientResponse );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uResponses );

	pxDialog = pxOverlay->GetDialog( DialogID_DiffuseResponse );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uResponses );

	pxDialog = pxOverlay->GetDialog( DialogID_SpecularResponse );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uResponses );

	pxDialog = pxOverlay->GetDialog( DialogID_TransparencyResponse );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uResponses );

	pxDialog = pxOverlay->GetDialog( DialogID_SelfIlluminationResponse );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uResponses );

	pxDialog = pxOverlay->GetDialog( DialogID_ReflectivityResponse );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uResponses );

	// Textures

	pxDialog = pxOverlay->GetDialog( DialogID_Textures );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything );

	pxDialog = pxOverlay->GetDialog( DialogID_AlbedoMaps );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uTextures );

	pxDialog = pxOverlay->GetDialog( DialogID_NormalMaps );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uTextures );

	pxDialog = pxOverlay->GetDialog( DialogID_SpecularMaps );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uTextures );

	pxDialog = pxOverlay->GetDialog( DialogID_TransparencyMaps );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uTextures );

	pxDialog = pxOverlay->GetDialog( DialogID_SelfIlluminationMaps );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uTextures );

	pxDialog = pxOverlay->GetDialog( DialogID_ReflectivityTextures );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything && !m_xCopyFlags.m_uTextures );

	// Texture anims
	pxDialog = pxOverlay->GetDialog( DialogID_TextureAnims );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything );


	// Special FX
	pxDialog = pxOverlay->GetDialog( DialogID_SpecialFX );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything );

	// Surface type
	pxDialog = pxOverlay->GetDialog( DialogID_SurfaceType );
	pxDialog->SetEnabled( !m_xCopyFlags.m_uEverything );

	m_pxFromMaterialDialog->SetFilterByEnv(s_bFilterFromListByEnv);
	m_pxToMaterialDialog->SetFilterByEnv(s_bFilterToListByEnv);


	
}

Tools_MaterialResponse* Tools_CopyMaterialProperties_Overlay::GetToMaterial()
{

	return Tools_MaterialResponse_System::GetMaterialByHashID( m_uToHash );
}

Tools_MaterialResponse* Tools_CopyMaterialProperties_Overlay::GetFromMaterial()
{
	return Tools_MaterialResponse_System::GetMaterialByHashID( m_uFromHash );
}

void Tools_CopyMaterialProperties_Overlay::InitialiseOverlay( Asura_Overlay* pxOverlay )
{
	Asura_Assert( pxOverlay, "NULL parameter passed to Tools_CopyMaterialProperties_Overlay::InitialiseOverlay()." );
	if( !pxOverlay ) return;

	Asura_Vector_2 xOffset( 1.0f, 3.0f );
	Asura_Dialog* pxDialog;

	// "Copy from" selector
	{
		pxDialog = new Asura_Dialog( "Copy From:", xOffset );
		pxOverlay->AddDialog( pxDialog );
		
		xOffset.x += 8.0f;
		pxDialog = new Asura_Tickbox_Dialog("Filter By Env", xOffset, &s_bFilterFromListByEnv);
		pxDialog->SetHelpText("Limits list of materials to those used on the environment");
		pxOverlay->AddDialog( pxDialog );
		xOffset.x -= 8.0f;
		
		xOffset.y += Asura_Dialog::DefaultSpacing();

		pxDialog = m_pxFromMaterialDialog = new Tools_MaterialResponse_DropDownList_Dialog( xOffset, 15, s_bFilterFromListByEnv, &m_uFromHash, NULL );
		pxDialog->SetHelpText( "Selects the material from which to copy properties." );
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_List_Dialog::LargeSpacing();
	}

	// "Copy to" selector
	{
		pxDialog = new Asura_Dialog( "Copy To:", xOffset );
		pxOverlay->AddDialog( pxDialog );

		xOffset.x += 8.0f;
		pxDialog = new Asura_Tickbox_Dialog("Filter By Env", xOffset, &s_bFilterToListByEnv);
		pxDialog->SetHelpText("Limits list of materials to those used on the environment");
		pxOverlay->AddDialog( pxDialog );
		xOffset.x -= 8.0f;

		xOffset.y += Asura_Dialog::DefaultSpacing();

		pxDialog = m_pxToMaterialDialog = new Tools_MaterialResponse_DropDownList_Dialog( xOffset, 15, s_bFilterToListByEnv, &m_uToHash, NULL );
		pxDialog->SetHelpText( "Selects the material to which to copy properties." );
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_List_Dialog::LargeSpacing();
	}

	// All properties
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "All Properties", xOffset, &m_xCopyFlags.m_uAllFlags, 1 );
		pxDialog->SetHelpText( "Selects all properties." );
		pxDialog->m_iControlID = DialogID_Everything;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	xOffset.x += 1.0f;

	// All responses
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "All Responses", xOffset, &m_xCopyFlags.m_uAllFlags, 2 );
		pxDialog->SetHelpText( "Selects all light responses." );
		pxDialog->m_iControlID = DialogID_Responses;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	xOffset.x += 1.0f;

	// Ambient response
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Ambient", xOffset, &m_xCopyFlags.m_uAllFlags, 4 );
		pxDialog->SetHelpText( "Selects all ambient properties." );
		pxDialog->m_iControlID = DialogID_AmbientResponse;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Diffuse response
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Diffuse", xOffset, &m_xCopyFlags.m_uAllFlags, 8 );
		pxDialog->SetHelpText( "Selects all diffuse properties." );
		pxDialog->m_iControlID = DialogID_DiffuseResponse;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Specular response
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Specular", xOffset, &m_xCopyFlags.m_uAllFlags, 16 );
		pxDialog->SetHelpText( "Selects all specular properties." );
		pxDialog->m_iControlID = DialogID_SpecularResponse;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Transparency response
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Transparency", xOffset, &m_xCopyFlags.m_uAllFlags, 32 );
		pxDialog->SetHelpText( "Selects all transparency properties." );
		pxDialog->m_iControlID = DialogID_TransparencyResponse;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Self illumination response
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Self Illumination", xOffset, &m_xCopyFlags.m_uAllFlags, 64 );
		pxDialog->SetHelpText( "Selects all self illumination properties." );
		pxDialog->m_iControlID = DialogID_SelfIlluminationResponse;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Reflectivity response
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Reflectivity", xOffset, &m_xCopyFlags.m_uAllFlags, 128 );
		pxDialog->SetHelpText( "Selects all reflectivity properties." );
		pxDialog->m_iControlID = DialogID_ReflectivityResponse;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	xOffset.x -= 1.0f;

	// Textures
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Textures", xOffset, &m_xCopyFlags.m_uAllFlags, 256 );
		pxDialog->SetHelpText( "Selects all texture slots." );
		pxDialog->m_iControlID = DialogID_Textures;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	xOffset.x += 1.0f;

	// Albedo maps
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Albedo Maps", xOffset, &m_xCopyFlags.m_uAllFlags, 512 );
		pxDialog->SetHelpText( "Selects all albedo maps." );
		pxDialog->m_iControlID = DialogID_AlbedoMaps;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Normal maps
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Normal Maps", xOffset, &m_xCopyFlags.m_uAllFlags, 1024 );
		pxDialog->SetHelpText( "Selects all normal maps." );
		pxDialog->m_iControlID = DialogID_NormalMaps;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Specular maps
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Specular Maps", xOffset, &m_xCopyFlags.m_uAllFlags, 2048 );
		pxDialog->SetHelpText( "Selects all specular maps." );
		pxDialog->m_iControlID = DialogID_SpecularMaps;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Transparency maps
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Transparency", xOffset, &m_xCopyFlags.m_uAllFlags, 4096 );
		pxDialog->SetHelpText( "Selects all transparency maps." );
		pxDialog->m_iControlID = DialogID_TransparencyMaps;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Self illumination maps
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Self Illumination", xOffset, &m_xCopyFlags.m_uAllFlags, 8192 );
		pxDialog->SetHelpText( "Selects all self illumination maps." );
		pxDialog->m_iControlID = DialogID_SelfIlluminationMaps;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	// Reflectivity textures
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Reflectivity", xOffset, &m_xCopyFlags.m_uAllFlags, 16384 );
		pxDialog->SetHelpText( "Selects all reflectivity textures." );
		pxDialog->m_iControlID = DialogID_ReflectivityTextures;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::DefaultSpacing();
	}

	xOffset.x -= 1.0f;

	// Texture animations
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Texture Animations", xOffset, &m_xCopyFlags.m_uAllFlags, 32768 );
		pxDialog->SetHelpText( "Selects all texture animation sets." );
		pxDialog->m_iControlID = DialogID_TextureAnims;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::LargeSpacing();
	}

	// Special FX
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Special FX", xOffset, &m_xCopyFlags.m_uAllFlags, 65536 );
		pxDialog->SetHelpText( "Selects all special FX properties." );
		pxDialog->m_iControlID = DialogID_SpecialFX;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::LargeSpacing();
	}

	// Surface type
	{
		pxDialog = new Asura_FlagTickbox_Dialog( "Surface Type", xOffset, &m_xCopyFlags.m_uAllFlags, 0x00020000 );
		pxDialog->SetHelpText( "Selects the surface type." );
		pxDialog->m_iControlID = DialogID_SurfaceType;
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_FlagTickbox_Dialog::LargeSpacing();
	}


	// OK button
	{
		pxDialog = new Asura_Function_Dialog( "OK", xOffset, OnOK, reinterpret_cast< int >( this ), &pxOverlay->m_bShouldClose );
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_Function_Dialog::DefaultSpacing();
	}

	// Cancel button
	{
		pxDialog = new Asura_Function_Dialog( "Cancel", xOffset, NULL, &pxOverlay->m_bShouldClose );
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_Function_Dialog::DefaultSpacing();
	}
}

/******************************************************************************/
/* Tools_CopyMaterialProperties_Overlay Static Class Functions                */
/******************************************************************************/

void Tools_CopyMaterialProperties_Overlay::OnOK( int iThis )
{
	// Urgh... but that's the way these overlays work
	Tools_CopyMaterialProperties_Overlay* const pxThis = reinterpret_cast< Tools_CopyMaterialProperties_Overlay* >( iThis );

	if(!pxThis->m_uFromHash || !pxThis->m_uToHash) return;


	const Tools_MaterialResponse* const pxCopyFromMaterial = pxThis->GetFromMaterial();
	if( !pxCopyFromMaterial )
	{
		Asura_Assert( false, "Selected From material is not in the material list." );
		return;
	}

	Tools_MaterialResponse* const pxCopyToMaterial = pxThis->GetToMaterial();
	if( !pxCopyToMaterial )
	{
		Asura_Assert( false, "Selected To material is not in the material list." );
		return;
	}

	pxCopyToMaterial->CopyPropertiesFrom( *pxCopyFromMaterial, pxThis->m_xCopyFlags );

	//update the properties overlay for the materials involved
	Tools_MaterialProperties_Overlay* pxOverlay = Tools_MaterialProperties_Overlay::FindOverlayForMaterial(pxCopyFromMaterial->GetHash());
	if(pxOverlay)
	{
		pxOverlay->UpdateDialogs();
	}
	pxOverlay = Tools_MaterialProperties_Overlay::FindOverlayForMaterial(pxCopyToMaterial->GetHash());
	if(pxOverlay)
	{
		pxOverlay->UpdateDialogs();
	}
}

/******************************************************************************/
/* Tools_EnslaveMaterialToOther_Overlay Static Class Data                     */
/******************************************************************************/

Tools_EnslaveMaterialToOther_Overlay* Tools_EnslaveMaterialToOther_Overlay::s_pxThis = NULL;
bool Tools_EnslaveMaterialToOther_Overlay::s_bFilterTreeByEnv = false;

/******************************************************************************/
/* Tools_EnslaveMaterialToOther_Overlay Class Functions                       */
/******************************************************************************/

Tools_EnslaveMaterialToOther_Overlay::Tools_EnslaveMaterialToOther_Overlay( const Asura_String& xTitle, Tools_MaterialResponse& xMaterial )
	: Tools_Modal_Overlay( xTitle )
	, m_xMaterial( xMaterial )
	, m_pxTreeDialog(NULL)
	, m_bFilterTreeByEnvChanged(false)
{
	Asura_Assert( !s_pxThis, "More than one 'enslave to other material' overlay is active at once." );
	s_pxThis = this;
}

Tools_EnslaveMaterialToOther_Overlay::~Tools_EnslaveMaterialToOther_Overlay()
{
	s_pxThis = NULL;
}

bool Tools_EnslaveMaterialToOther_Overlay::EnslaveTreeFilter( const Tools_MaterialResponse& xMaterial, const void* const pUserData )
{
	const Tools_EnslaveMaterialToOther_Overlay* pxOverlay = static_cast< const Tools_EnslaveMaterialToOther_Overlay* >( pUserData );
	
	// User data is the material we're enslaving
	const Tools_MaterialResponse* const pxSlave = &pxOverlay->m_xMaterial;

	if(s_bFilterTreeByEnv)
	{
		//Need to cast constness away on the hash tree , because of the multi-purpose nature of FindNodeIndexByHash()
		Asura_HashTree* pxHashTree = const_cast<Asura_HashTree*>(&pxOverlay->m_xUsedMaterialTree);
		
		//check against the tree of materials used on the environment
		int iJunkIndex;
		if(!pxHashTree->FindNodeIndexByHash(xMaterial.GetHash(), false, iJunkIndex))
		{	
			return false;
		}
	}

	// Don't include ourselves, or materials which are already slaves to something else
	const bool bThisMaterial = (xMaterial.GetHash() == pxSlave->GetHash());
	const bool bNoParent = (xMaterial.GetParentHash() == ASURA_HASH_ID_UNSET);
	return !bThisMaterial && bNoParent;
}

void Tools_EnslaveMaterialToOther_Overlay::InitialiseOverlay( Asura_Overlay* pxOverlay )
{
	if( !pxOverlay )
	{
		Asura_Assert( false, "NULL parameter passed to Tools_EnslaveMaterialToOther_Overlay::InitialiseOverlay()." );
		return;
	}

	//Get a list of materials used by the environment if needed by the filter
	if(s_bFilterTreeByEnv && !m_xUsedMaterialTree.GetNumberOfNodes())
	{
		Tools_Texturing_MaterialResponse::FindMaterialsUsedByEnv(m_xUsedMaterialTree);
	}

	Tools_MaterialResponse_System::CreateMaterialNameTree( m_xMaterialNameTree, EnslaveTreeFilter, this );

	Asura_Dialog* pxDialog;
	Asura_Vector_2 xOffset( 1.0f, 3.0f );

	pxDialog = new Asura_Tickbox_Dialog("Filter By Env", xOffset, &s_bFilterTreeByEnv, &m_bFilterTreeByEnvChanged);
	pxDialog->SetHelpText( "Filters the list of materials so that it only shows the materials used by the 'environment'." );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::LargeSpacing();

	pxDialog = m_pxTreeDialog = new Asura_Tree_Dialog( xOffset, &m_xMaterialNameTree, 30.0f, 30.0f );
	((Asura_Tree_Dialog*)pxDialog)->SetDoubleClickedFunction(OnOK);
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += 30.0f + Asura_Dialog::LargeSpacing();

	pxDialog = new Asura_Function_Dialog( "OK", xOffset, OnOK, NULL, IsMaterialNodeSelected );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();

	pxDialog = new Asura_Function_Dialog( "Cancel", xOffset, NULL, &pxOverlay->m_bShouldClose );
	pxOverlay->AddDialog( pxDialog );
	xOffset.y += Asura_Function_Dialog::DefaultSpacing();
}

void Tools_EnslaveMaterialToOther_Overlay::UpdateTreeDialog()
{
	Asura_TreeDialog_Item* pxPrevSelection = m_xMaterialNameTree.GetSelectedItem();
	Asura_Hash_ID uSelectedMaterial = pxPrevSelection ? pxPrevSelection->GetUserData() : 0;

	//Get a list of materials used by the environment if needed by the filter
	if(s_bFilterTreeByEnv && !m_xUsedMaterialTree.GetNumberOfNodes())
	{
		Tools_Texturing_MaterialResponse::FindMaterialsUsedByEnv(m_xUsedMaterialTree);
	}
	
	//rebuild the tree
	Tools_MaterialResponse_System::CreateMaterialNameTree( m_xMaterialNameTree, EnslaveTreeFilter, this );

	//set the previous selection again if it is still available
	Asura_TreeDialog_Item* pxItem = m_xMaterialNameTree.FindByData(uSelectedMaterial);
	if(pxItem)
	{
		m_xMaterialNameTree.SetSelectedItem(pxItem);		
		m_pxTreeDialog->MakeSelectionVisible();
	}
}

void Tools_EnslaveMaterialToOther_Overlay::Update()
{
	if(m_bFilterTreeByEnvChanged)
	{
		m_bFilterTreeByEnvChanged = false;
		UpdateTreeDialog();
	}
}

/******************************************************************************/
/* Tools_EnslaveMaterialToOther_Overlay Static Class Functions                */
/******************************************************************************/

bool Tools_EnslaveMaterialToOther_Overlay::IsMaterialNodeSelected()
{
	if( !s_pxThis )
	{
		Asura_Assert( false, "No 'enslave to other material' dialogue exists." );
		return false;
	}

	// Only leaf nodes are actual materials, the others represent directories
	const Asura_TreeDialog_Item* const pxSelectedNode = s_pxThis->GetTree().GetSelectedItem();
	return (pxSelectedNode && pxSelectedNode->GetNumChildren() == 0);
}

void Tools_EnslaveMaterialToOther_Overlay::OnOK()
{
	if( !s_pxThis )
	{
		Asura_Assert( false, "No 'enslave to other material' dialogue exists." );
		return;
	}
	s_pxThis->CloseOverlay();

	Asura_TreeDialog_Item* const pxSelectedNode = s_pxThis->GetTree().GetSelectedItem();
	const Asura_Hash_ID uSelectedHash = pxSelectedNode->GetUserData();

	Tools_MaterialResponse* const pxSelectedParent = Tools_MaterialResponse_System::GetMaterialByHashID( uSelectedHash );
	if( pxSelectedParent )
	{
		s_pxThis->m_xMaterial.SetParentHash( uSelectedHash );
		s_pxThis->m_xMaterial.SetModified( true );
	}
	else
	{
		Asura_Assert( false, "Selected material couldn't be found." );
	}
}


/******************************************************************************/
/* Tools_MaterialResponse_UndoData_ApplyMaterial Class Functions              */
/******************************************************************************/

Tools_MaterialResponse_UndoData_ApplyMaterial::Tools_MaterialResponse_UndoData_ApplyMaterial( const u_int uMeshIndex, const u_int uPolygonIndex )
	: m_uMeshIndex( uMeshIndex )
	, m_uPolygonIndex( uPolygonIndex )
{
	Asura_Mesh* const pxMesh = Asura_Mesh_System::GetMesh( uMeshIndex );

	if( pxMesh && pxMesh->TextureDataValid && pxMesh->TextureDataPtr )
	{
		for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
		{
			m_auPreviousMaterialHash[iMatSlot] = pxMesh->TextureDataPtr[ m_uPolygonIndex ].m_auMaterialResponseHashID[iMatSlot];
		}
	}
	else
	{
		Asura_Assertf( pxMesh, ("Apply materials undo: cannot find mesh with index %u.", uMeshIndex) );
		for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
		{
			m_auPreviousMaterialHash[iMatSlot] = ASURA_HASH_ID_UNSET;
		}
	}
}

RootUndoData* Tools_MaterialResponse_UndoData_ApplyMaterial::ProcessUndoData()
{
	Asura_Mesh* const pxMesh = Asura_Mesh_System::GetMesh( m_uMeshIndex );

	if( !pxMesh || !pxMesh->TextureDataValid || !pxMesh->TextureDataPtr )
	{
		// Mesh has been deleted or something, best ignore
		return this;
	}

  	for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
	{
		Asura_Hash_ID& uPolyMaterialHash = pxMesh->TextureDataPtr[ m_uPolygonIndex ].m_auMaterialResponseHashID[iMatSlot];

		const Asura_Hash_ID uTempHash = uPolyMaterialHash;
		uPolyMaterialHash = m_auPreviousMaterialHash[iMatSlot];

		// Convert this object into a redo
		m_auPreviousMaterialHash[iMatSlot] = uTempHash;
	}

	Tools_UpdateTexturingOnSurfaces( pxMesh );
	Tools_Texturing_MaterialResponse::SetMeshModified( pxMesh, true );

	return this;
}

/******************************************************************************/
/* Tools_MaterialResponse_UndoData_ImportMaterial Class Functions             */
/******************************************************************************/

Tools_MaterialResponse_UndoData_ImportMaterial::Tools_MaterialResponse_UndoData_ImportMaterial( const u_int uMeshIndex, const u_int uPolygonIndex )
	: Tools_MaterialResponse_UndoData_ApplyMaterial( uMeshIndex, uPolygonIndex )
{
}

/******************************************************************************/
/* Tools_MaterialResponse_UndoData_RemoveMaterial Class Functions             */
/******************************************************************************/

Tools_MaterialResponse_UndoData_RemoveMaterial::Tools_MaterialResponse_UndoData_RemoveMaterial( const u_int uMeshIndex, const u_int uPolygonIndex )
	: Tools_MaterialResponse_UndoData_ApplyMaterial( uMeshIndex, uPolygonIndex )
{
}

/******************************************************************************/
/* Tools_Texturing_MaterialResponse Static Class Data                         */
/******************************************************************************/

float Tools_Texturing_MaterialResponse::s_fSpecularSharpness = 16.0f;

// This state is stored outside of the overlay itself so it can be saved when Tools is quit,
// regardless of whether the overlay is currently open

Asura_Vector_2 Tools_Texturing_MaterialResponse::s_xOverlayPosition( 0.0f, 0.0f );
Asura_Vector_2 Tools_Texturing_MaterialResponse::s_xUserSetTreeSize( 0.0f, 0.0f );
bool Tools_Texturing_MaterialResponse::s_bOverlayOpen = false;
u_int Tools_Texturing_MaterialResponse::s_uOverlayMinimised = 0;

bool Tools_Texturing_MaterialResponse::s_bMeshModified = false;

Tools_Texturing_MaterialResponse* Tools_Texturing_MaterialResponse::s_pxThis = 0;

int Tools_Texturing_MaterialResponse::s_iMaterialConversion_NumChanged = 0;
int Tools_Texturing_MaterialResponse::s_iMaterialConversion_NumCreated = 0;
bool Tools_Texturing_MaterialResponse::s_bMaterialConversion_ShowMessage = false;

Tools_Texturing_MaterialResponse::MaterialProjectUICallback Tools_Texturing_MaterialResponse::s_pfnProjectUICallback = NULL;

/******************************************************************************/
/* Tools_Texturing_MaterialResponse Static Class Functions                    */
/******************************************************************************/

Tools_Texturing_MaterialResponse::Tools_Texturing_MaterialResponse()
{
	Asura_Assert(!s_pxThis,"Tools_Texturing_MaterialResponse already allocated");
	s_pxThis = this;
}

Tools_Texturing_MaterialResponse::~Tools_Texturing_MaterialResponse()
{
	Asura_Assert(s_pxThis,"Tools_Texturing_MaterialResponse not allocated");
	s_pxThis = 0;
}

void Tools_Texturing_MaterialResponse::Initialise()
{
	TransferModeData( TOOLSMODEDATAFILE_LOAD );
}

void Tools_Texturing_MaterialResponse::Shutdown()
{
	TransferModeData( TOOLSMODEDATAFILE_SAVE );

	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
	Tools_MaterialProperties_Overlay::CloseOverlays();
	Tools_Overlay_ViewMaterialResponse::CloseOverlay();
	Tools_Overlay_ImportMaterialResponse::CloseOverlay();
}

void Tools_Texturing_MaterialResponse::CreateMaterialResponse()
{
	// Create the new material
	Asura_String xFilename;
	if( !ShowNewFileDialogue( xFilename ) ) return;

	const Asura_Hash_ID uNewMaterialHash = xFilename.GetHash();

	Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddMaterial( uNewMaterialHash );
	if( !pxNewMaterial )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Failed to create new material." );
		return;
	}

	pxNewMaterial->m_xName = xFilename;
	pxNewMaterial->SetNeedsSave(true);

	// No need to set m_bModified, just update the texture now
	Asura_PC_MaterialResponse_UpdateTexture();

	// Regenerate the name list in the main overlay
	RegenerateOverlay();

	// Open the properties overlay immediately
	EditMaterial( uNewMaterialHash );
}

void Tools_Texturing_MaterialResponse::CreateOverlay()
{
	if( !OverlayIsOn() ) return;

	// Create the overlay
	Tools_MaterialResponse_Overlay* const pxOverlay = new Tools_MaterialResponse_Overlay;
	if( !pxOverlay )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Failed to create material response overlay." );
		return;
	}

	// Add controls
	pxOverlay->AddDialogs();

	// Submit to the interface system
	Asura_Overlay_System::SubmitNewOverlay( pxOverlay );
}

void Tools_Texturing_MaterialResponse::DestroyOverlay()
{
	Asura_Overlay_System::CloseOverlays( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
}

void Tools_Texturing_MaterialResponse::RegenerateOverlay()
{
	Asura_Overlay* const pxOverlay = Asura_Overlay_System::GetOverlay( TOOLS_OVERLAY_TEXTURING_MATERIAL_RESPONSE_LIST );
	if( pxOverlay )
	{
		Tools_MaterialResponse_Overlay* const pxMaterialListOverlay = static_cast< Tools_MaterialResponse_Overlay* >( pxOverlay );
		pxMaterialListOverlay->UpdateDialogs();
	}
}

void Tools_Texturing_MaterialResponse::SetSelectedMaterial(Asura_Hash_ID uMaterialHash)
{
	Tools_MaterialResponse_Overlay::SetSelectedMaterialHash(uMaterialHash);
}


void Tools_Texturing_MaterialResponse::ApplySelectedMaterial(int iMatSlot)
{
	if(Tools_MaterialResponse_Overlay::OverlayExists())
	{
		ApplyMaterial( Tools_MaterialResponse_Overlay::GetSelectedMaterialHash(), iMatSlot );
	}
}

void Tools_Texturing_MaterialResponse::ApplyEditedOrSelectedMaterial()
{
	//Apply material from material properties overlay if it exists
	//RAR 12/07/2007
	Tools_MaterialProperties_Overlay* pxOverlay = Tools_MaterialProperties_Overlay::GetMostRecentlySelectedOverlay();
	if(pxOverlay)
	{
		ApplyPrimaryMaterial(pxOverlay->GetMaterialHash());
	}
	else
	{
		//Otherwise apply the selected material
		ApplySelectedMaterial(0);
	}
}

void Tools_Texturing_MaterialResponse::ApplyMaterial( int iMaterialHash,int iMatSlot )
{
	Asura_Assert(iMatSlot>=0 && iMatSlot < TOOLS_NUM_MATERIALRESPONSE_PER_POLY,"");


	
	if( Tools_Polygon_Selection::IsActive() && Tools_Polygon_Selection::NumSelectedModules )
	{

		// Iterate over selected polygons and apply the material

		Tools_Polygon_Selection::StartIteration();

		PolygonID xPolygon;
		while( Tools_Polygon_Selection::GetNextSelected( &xPolygon ) )
		{
			Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule( xPolygon.ModuleIndex );
			if( !pxMesh || !pxMesh->TextureDataValid ) continue;

			LogApplyMaterial( pxMesh->Index, xPolygon.PolyIndex );

			pxMesh->TextureDataPtr[ xPolygon.PolyIndex ].m_auMaterialResponseHashID[iMatSlot] = iMaterialHash;

			Tools_UpdateTexturingOnSurfaces( pxMesh );
			SetMeshModified( pxMesh, true );
		}

		// Kick off the next undo batch, so the whole apply will be undone at once
		Tools_Undo_System::MarkUndoStart();
	}

	//Update the selected material in texturing mode
	//(Doesn't matter if we aren't currently in that mode)
	//RAR 01/06/2007
	Tools_Mode_Texturing::SetMaterialResponse(iMaterialHash);

}


void Tools_Texturing_MaterialResponse::DuplicateSelectedMaterial()
{
	DuplicateMaterial( Tools_MaterialResponse_Overlay::GetSelectedMaterialHash(), true );
}

Asura_Hash_ID Tools_Texturing_MaterialResponse::DuplicateMaterial( Asura_Hash_ID uMaterialHash, bool bOpenOverlay )
{
	Tools_MaterialResponse* const pxSelectedMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash );
	if( !pxSelectedMaterial ) return 0;

	Asura_String xFilename = pxSelectedMaterial->GetName();
	if( !ShowNewFileDialogue( xFilename ) ) return 0;

	const Asura_Hash_ID uNewMaterialHash = xFilename.GetHash();

	// AddDuplicateMaterial will copy-construct a new material
	Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddDuplicateMaterial( uNewMaterialHash, *pxSelectedMaterial );
	if( !pxNewMaterial )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Failed to duplicate material." );
		return 0;
	}

	pxNewMaterial->m_xName = xFilename;
	pxNewMaterial->SetNeedsSave(true);

	// No need to set m_bModified, just update the texture now
	Asura_PC_MaterialResponse_UpdateTexture();

	// Recreate the name list in the main overlay
	RegenerateOverlay();

	if(bOpenOverlay)
	{
		// Open the properties overlay immediately
		EditMaterial( uNewMaterialHash );
	}

	return uNewMaterialHash;
}

void Tools_Texturing_MaterialResponse::EditSelectedMaterial()
{
	EditMaterial( Tools_MaterialResponse_Overlay::GetSelectedMaterialHash() );
}

void Tools_Texturing_MaterialResponse::EditMaterial( const Asura_Hash_ID uMaterialHash )
{
	Tools_MaterialProperties_Overlay::EditMaterial(uMaterialHash);
}


Tools_MaterialResponse* Tools_Texturing_MaterialResponse::CreateMaterialResponseFromOldMaterial(int iOldMaterialID, int iMatSlot, bool bAutomatic)
{
	// Retrieve the old material
	const Asura_Material* const pxOldMaterial = Asura_Material_System::GetMaterial( iOldMaterialID );
	Asura_Assert(pxOldMaterial, "");
	if(!pxOldMaterial) return 0;


	//ignore old materials that don't have an albedo texture
	if (pxOldMaterial->m_iTextureID[iMatSlot] == ASURA_TEXTURE_NONE) return 0;

	
	Asura_String xFullMaterialName;

	static Asura_String s_xMaterialPath;

	if(s_xMaterialPath.IsBlank())
	{
		const Asura_String xLoadedFile = Tools_SaveAsura::GetRealFilename();

		Asura_String xPath, xFilename;
		xLoadedFile.SplitFilename( &xPath, &xFilename, NULL );

		s_xMaterialPath = Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir( xPath.GetString() );
		s_xMaterialPath += "\\";
		s_xMaterialPath += xFilename;
		s_xMaterialPath += "\\";
	}


	//generate a material name
	Asura_String xTextureFilename;
	const Asura_String* const pxTextureName = Asura_Texture_Management::GetTextureName( pxOldMaterial->m_iTextureID[iMatSlot] );
	if( pxTextureName )
	{
		pxTextureName->SplitFilename( NULL, &xTextureFilename, NULL );
	}
	else
	{
		Asura_Assertf( false,
						   ("Material %s (ID %d) has unknown texture name (ID %d).",
						    pxOldMaterial->m_xMaterialName.GetString(),
							iOldMaterialID, pxOldMaterial->m_iTextureID[0]) );

		xTextureFilename = "Untextured";
	}

	char szMaterialName[256];
	if( pxOldMaterial->m_xMaterialName.IsBlank() )
	{
		Asura_PrintfToBfr( szMaterialName, "%s_%d", 256, true, xTextureFilename.GetString(), iOldMaterialID );
	}
	else
	{
		Asura_PrintfToBfr( szMaterialName, "%s_%s", 256, true, xTextureFilename.GetString(),
						   pxOldMaterial->m_xMaterialName.GetString() );
	}

	// We're going to save these as filenames, so make sure we strip out invalid chars
	{
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '\\', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '/', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, ':', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '*', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '?', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '\"', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '<', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '>', '_' );
		Asura_StringUtil::ReplaceAllChars( szMaterialName, '|', '_' );
	}

	xFullMaterialName = s_xMaterialPath;
	xFullMaterialName += szMaterialName;
	
	if(!bAutomatic)
	{
		//ask for a material name
		if(!ShowNewFileDialogue(xFullMaterialName)) return 0;
	}

	// Check for existing material with this name
	Asura_Hash_ID uHashID = xFullMaterialName.GetHash();
	Tools_MaterialResponse* pxNewMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( uHashID );
	if(pxNewMaterial) return pxNewMaterial;

	// Create a new material, matching the old appearance
	{
		pxNewMaterial = Tools_MaterialResponse_System::AddMaterial( xFullMaterialName.GetHash() );
		if( !pxNewMaterial )
		{
			Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Ran out of memory." );
			return 0;
		}

		pxNewMaterial->m_xName = xFullMaterialName;

		pxNewMaterial->SetTexture( Asura_MaterialResponse::ALBEDO_TEXTURE, pxOldMaterial->m_iTextureID[iMatSlot] );
		pxNewMaterial->SetTexture( Asura_MaterialResponse::NORMAL_MAP, pxOldMaterial->m_iNormalMapID[iMatSlot] );

		if( pxOldMaterial->m_uMaterialFlags & (AsuraMaterialFlag_Specular_Soft | AsuraMaterialFlag_Specular_Sharp) )
		{
			if( pxOldMaterial->m_iSpecularMapID[0] != ASURA_TEXTURE_NONE )
			{
				pxNewMaterial->SetTexture( Asura_MaterialResponse::SPECULAR_MAP, pxOldMaterial->m_iSpecularMapID[iMatSlot] );
			}
			else
			{
				pxNewMaterial->SetTexture( Asura_MaterialResponse::SPECULAR_MAP, pxOldMaterial->m_iTextureID[iMatSlot] );
			}

			const float fSpecularPower = (pxOldMaterial->m_uMaterialFlags & AsuraMaterialFlag_Specular_Soft ? 16.0f : 32.0f);
			pxNewMaterial->ApplySpecularSharpness( fSpecularPower, true );
		}
		else
		{
			// Remove points, leaving a line at 0
			Asura_Graph_Point< float >* pxControlPoint = pxNewMaterial->m_xSpecularGraph.GetControlPoints()->m_pxNext;
			Asura_Graph_Point< float >* pxNextPoint = pxControlPoint->m_pxNext;

			while( pxNextPoint  )
			{
				pxNewMaterial->m_xSpecularGraph.RemovePoint( pxControlPoint );
				pxControlPoint = pxNextPoint;
				pxNextPoint = pxControlPoint->m_pxNext;
			}

			pxControlPoint->m_pxPrev->m_xValue = 0.0f;
			pxNewMaterial->m_xSpecularGraph.Update();
		}

		if( pxOldMaterial->m_uMaterialFlags & AsuraMaterialFlag_Transparent )
		{
			// Problem here: environment strips have only 1-bit transparency. The old texture
			// might well be 8-bit though for use with specular at the same time
			pxNewMaterial->SetTexture( Asura_MaterialResponse::TRANSPARENCY_MAP, pxOldMaterial->m_iTextureID[iMatSlot] );
		}

		if( pxOldMaterial->m_uMaterialFlags & AsuraMaterialFlag_SelfIlluminate )
		{
			pxNewMaterial->SetTexture( Asura_MaterialResponse::SELFILLUMINATION_MAP, pxOldMaterial->m_iTextureID[iMatSlot] );
			pxNewMaterial->m_fSelfIlluminationContribution = 100.0f;
		}

		if( pxOldMaterial->m_uMaterialFlags & AsuraMaterialFlag_TextureAnim )
		{
			pxNewMaterial->m_uTextureAnimHashID = pxOldMaterial->m_uAnimHash;
		}

		if( pxOldMaterial->m_uMaterialFlags & AsuraMaterialFlag_AlphaDecal )
		{
			pxNewMaterial->m_uChunkedFlags	|= Tools_MaterialResponse::BITMASK_DECAL;
		}

		pxNewMaterial->m_bModified = true;
	}

	OutputToMaterialLogFile("Created %s\n",pxNewMaterial->GetName().GetString());
	



	// Rebuild the material list in the main overlay
	RegenerateOverlay();

	Asura_PC_MaterialResponse_UpdateTexture();

	return pxNewMaterial;
}

void Tools_Texturing_MaterialResponse::ConvertOldMaterials()
{
	ASURA_CPU_PROFILE_SCOPED_TIMER( "Tools_Texturing_MaterialResponse::ConvertOldMaterials");
	
	//Get a list of material response materials that need updating
	List<Tools_MaterialResponse*> xMatList;
	
	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		Tools_MaterialResponse* pxMaterial = Tools_MaterialResponse_System::GetMaterialByIndex(u);

		if(pxMaterial->m_puOldTextureHashIDs)
		{
			xMatList.add_entry(pxMaterial);
		}
	}

	if(!xMatList.size()) return;

	OutputToMaterialLogFile("Material conversion updated %d materials :\n\n",xMatList.size());
	for(LIF<Tools_MaterialResponse*> xMatLIF(&xMatList); !xMatLIF.done(); xMatLIF.next())
	{
		OutputToMaterialLogFile("%s\n",xMatLIF()->GetName().GetString());
	}
	OutputToMaterialLogFile("\n");

	List<Tools_MaterialResponse*> xNewMatList;
	
	for(LIF<Tools_MaterialResponse*> xMatLIF(&xMatList); !xMatLIF.done(); xMatLIF.next())
	{
		Tools_MaterialResponse* pxMaterial = xMatLIF();

		//Does this material use any secondary textures?
		bool bAnySecondary = false;
		for(int iSlot=1;iSlot<=Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_2;iSlot+=2)
		{
			if(pxMaterial->m_puOldTextureHashIDs[iSlot])
			{
				bAnySecondary = true;
				break;
			}			
		}

		if(bAnySecondary)
		{
			//Search for an existing material that uses the uses textures that match those in the secondary slot for this one

			for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
			{
				Tools_MaterialResponse* pxOtherMaterial = Tools_MaterialResponse_System::GetMaterialByIndex(u);

				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::ALBEDO_TEXTURE] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::NORMAL_MAP] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_NORMAL_MAP_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::SPECULAR_MAP] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_SPECULAR_MAP_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::TRANSPARENCY_MAP] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_TRANSPARENCY_MAP_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::SELFILLUMINATION_MAP] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::ENVIRONMENT_MAP] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::REFLECTIVITY_MASK] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_2]) continue;
				if(pxOtherMaterial->m_auTextureHashIDs[Asura_MaterialResponse::PARALLAX_MAP] != pxMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_PARALLAX_MAP_2]) continue;

				if(pxOtherMaterial->m_uTextureAnimHashID != pxMaterial->m_puOldTextureAnimHashIDs[1]) continue;

				//this material should do
				pxMaterial->m_uSecondMaterialID = pxOtherMaterial->GetHash();
			}

			if(!pxMaterial->m_uSecondMaterialID)
			{
				//need to generate a new material then
				Tools_MaterialResponse* pxSecondaryMaterial = GenerateSecondaryMaterialFromOldMaterial(pxMaterial);
				if(pxSecondaryMaterial)
				{
					pxMaterial->m_uSecondMaterialID = pxSecondaryMaterial->GetHash();
					s_iMaterialConversion_NumCreated++;
					xNewMatList.add_entry(pxSecondaryMaterial);
				}
			}
		}

		//No longer need the old material array
		delete [] pxMaterial->m_puOldTextureHashIDs;
		pxMaterial->m_puOldTextureHashIDs = 0;
		delete [] pxMaterial->m_pxOldTextureFilenames;
		pxMaterial->m_pxOldTextureFilenames = 0;
		delete [] pxMaterial->m_puOldTextureAnimHashIDs;
		pxMaterial->m_puOldTextureAnimHashIDs = 0;

		pxMaterial->SetNeedsSave(true);
		s_iMaterialConversion_NumChanged++;
	}
	Asura_PC_MaterialResponse_UpdateTexture();

	if(xNewMatList.size())
	{
		OutputToMaterialLogFile("Material conversion created %d materials :\n\n",xNewMatList.size());
		for(LIF<Tools_MaterialResponse*> xMatLIF(&xNewMatList); !xMatLIF.done(); xMatLIF.next())
		{
			OutputToMaterialLogFile("%s\n",xMatLIF()->GetName().GetString());
		}

		OutputToMaterialLogFile("\n");
	}


	s_bMaterialConversion_ShowMessage = (s_iMaterialConversion_NumChanged || s_iMaterialConversion_NumCreated);

}

Tools_MaterialResponse* Tools_Texturing_MaterialResponse::GenerateSecondaryMaterialFromOldMaterial(const Tools_MaterialResponse* pxOldMaterial)
{
	//Determine a name for the new material
	//Take the albedo texture name if available , otherwise base it on the original material's name
	char szNewName[200];
	if(pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_2].IsBlank())
	{
		Asura_PrintfToBfr(szNewName,"%s_Secondary",sizeof(szNewName)-10,true,pxOldMaterial->GetName().GetString());
	}
	else
	{
		//use the path from the old material and the name from the albedo texture
		Asura_String xPath,xName;
		pxOldMaterial->GetName().SplitFilename(&xPath,0,0);
		pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_2].SplitFilename(0,&xName,0);

		if(xPath.IsBlank())
			Asura_PrintfToBfr(szNewName,"%s",sizeof(szNewName)-10,true,xName.GetString());
		else
			Asura_PrintfToBfr(szNewName,"%s\\%s",sizeof(szNewName)-10,true,xPath.GetString(),xName.GetString());
	}

	//Make sure the name is unique
	char* szPostFixPos = szNewName;
	szPostFixPos += strlen(szPostFixPos);

	Asura_Hash_ID uNewMaterialHash;
	for(int iPostFix=0;;iPostFix++)
	{
		if(iPostFix)
		{
			//add a number on the end
			sprintf(szPostFixPos,"%d",iPostFix);
		}

		//any material with this hash?
		uNewMaterialHash = Asura_GetHashID(szNewName);
		if(!Tools_MaterialResponse_System::GetMaterialByHashID(uNewMaterialHash))
		{
			break;
		}
	}

	//Create the new material by duplicating the old one
	Tools_MaterialResponse* const pxNewMaterial = Tools_MaterialResponse_System::AddDuplicateMaterial( uNewMaterialHash, *pxOldMaterial );
	if( !pxNewMaterial )
	{
		Asura_Assert(0,"Error creating new material in material conversion process");
		return 0;
	}

	pxNewMaterial->m_xName = szNewName;
	pxNewMaterial->SetNeedsSave(true);

	//copy the old secondary textures into the primary slots
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::ALBEDO_TEXTURE] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::NORMAL_MAP] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_NORMAL_MAP_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::SPECULAR_MAP] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_SPECULAR_MAP_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::TRANSPARENCY_MAP] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_TRANSPARENCY_MAP_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::SELFILLUMINATION_MAP] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::ENVIRONMENT_MAP] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::REFLECTIVITY_MASK] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_2];
	pxNewMaterial->m_auTextureHashIDs[Asura_MaterialResponse::PARALLAX_MAP] = pxOldMaterial->m_puOldTextureHashIDs[Tools_MaterialResponse::OLD_PARALLAX_MAP_2];

	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::ALBEDO_TEXTURE] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_ALBEDO_TEXTURE_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::NORMAL_MAP] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_NORMAL_MAP_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::SPECULAR_MAP] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_SPECULAR_MAP_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::TRANSPARENCY_MAP] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_TRANSPARENCY_MAP_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::SELFILLUMINATION_MAP] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_SELFILLUMINATION_MAP_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::ENVIRONMENT_MAP] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_ENVIRONMENT_MAP_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::REFLECTIVITY_MASK] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_REFLECTIVITY_MASK_2];
	pxNewMaterial->m_axTextureFilenames[Asura_MaterialResponse::PARALLAX_MAP] = pxOldMaterial->m_pxOldTextureFilenames[Tools_MaterialResponse::OLD_PARALLAX_MAP_2];


	pxNewMaterial->m_uTextureAnimHashID = pxOldMaterial->m_puOldTextureAnimHashIDs[1];

	pxNewMaterial->m_uSecondMaterialID = 0;
	
	return pxNewMaterial;
}

void Tools_Texturing_MaterialResponse::ApplySpecularSharpness( int iMaterialHash )
{
	Tools_MaterialResponse* const pxSelectedMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( iMaterialHash );

	if (!pxSelectedMaterial)
	{
		return;
	}
	
	pxSelectedMaterial->ApplySpecularSharpness( s_fSpecularSharpness, true );	
}

bool Tools_Texturing_MaterialResponse::IsModified()
{
	if( s_bMeshModified ) return true;

	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading(u);
		if (!pxCurrMaterial->IsLoaded()) continue;

		if( pxCurrMaterial->m_bNeedsSave ) return true;
	}

	return false;
}

void Tools_Texturing_MaterialResponse::SaveChangedMaterials()
{
	bool bSavedAnyMaterials = false;

	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading(u);
		if (!pxCurrMaterial->IsLoaded()) continue;

		if( !pxCurrMaterial->m_bNeedsSave ) continue;

		if( !bSavedAnyMaterials )
		{
			Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_INFO, "Saving changed materials..." );
		}

		if( pxCurrMaterial->CanSaveFile(true))
		{
			if( SaveMaterial( *pxCurrMaterial ) )
			{
				pxCurrMaterial->SetNeedsSave(false);
				Tools_Timed_Message_System::AddMessage( 0xFF00FF00, "Saved %s", pxCurrMaterial->GetName().GetString() );
			}
			else
			{
				Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "Failed to save material %s. Is it checked out?", pxCurrMaterial->GetName().GetString() );
			}
		}

		bSavedAnyMaterials = true;
	}

	if( bSavedAnyMaterials )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_INFO, "Finished saving materials" );
	}

	// If the main file is no longer tagged as modified, must have been successfully saved since we
	// did any material application modifications

	if( !Tools_SaveAsura::IsMainFileModified() )
	{
		s_bMeshModified = false;
	}
}

bool Tools_Texturing_MaterialResponse::SaveMaterial( Tools_MaterialResponse& xMaterial )
{
	Asura_List< Tools_MaterialResponse > xList;
	xList.AddToList( &xMaterial );

	Asura_SaveBlock xSaveBlock;
	Asura_Chunk_SaveBlock_Stream xStream( &xSaveBlock );

	Tools_Chunk_MaterialResponse::Create( xStream, xList );

	xStream.WriteAsuraTerminator();

	Asura_String xMaterialDirectory, xMaterialFilename;
	xMaterial.GetName().SplitFilename( &xMaterialDirectory, &xMaterialFilename, NULL );

	Asura_String xFullPath = szMATERIALS_DIRECTORY;
	xFullPath += xMaterialDirectory;
	xFullPath += "\\";

	Tools_FileSystem::EnsureDirectoryExists( xFullPath.GetString() );

	xFullPath += xMaterialFilename;
	xFullPath += szMATERIALS_EXTENSION;

	return xSaveBlock.SaveFile( xFullPath.GetString(), false );
}

void Tools_Texturing_MaterialResponse::ToggleOverlay()
{
	s_bOverlayOpen = !s_bOverlayOpen;

	if( OverlayIsOn() )
	{
		s_uOverlayMinimised = 0;
		CreateOverlay();
	}
	else
	{
		DestroyOverlay();
	}
}

class Tools_TextureUsageEntry {
	Asura_Hash_ID m_uHashID;
	unsigned int m_uUsage;
	unsigned int m_uDecalManagerUsage;
	unsigned int m_uStaticCoronaUsage;
	const char* m_pTextureName;
	Asura_List<Tools_MaterialResponse> m_axMaterials[Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS];

	void Init(Asura_Hash_ID uHashID, const char* pxTextureName);
public:
	Tools_TextureUsageEntry() {}
	Tools_TextureUsageEntry(Asura_Hash_ID uHashID, const char* pxTextureName)
	{
		Init(uHashID, pxTextureName);
	}

	static Tools_TextureUsageEntry *CreateEntry(Asura_TemplateHashTree<Tools_TextureUsageEntry>* pxAllUsage, Asura_Hash_ID uHashID, const char* pxTextureName);
	static void GenerateUsageTree(Asura_TemplateHashTree<Tools_TextureUsageEntry>& xAllUsage);

	const char* GetTextureName() {return m_pTextureName;}

	const Asura_List<Tools_MaterialResponse>& GetMaterialList(Asura_MaterialResponse::TEXTURE_SLOT_ID uUsage) {return m_axMaterials[uUsage];}


	void SetUsedByMaterial(Tools_MaterialResponse* pxMaterial, Asura_MaterialResponse::TEXTURE_SLOT_ID uUsage);
	void SetUsedByDecalMaterial(u_int uUsageMask);
	void SetUsedByStaticCoronaMaterial(u_int uUsageMask);

	unsigned int GetUsage() {return m_uUsage | m_uDecalManagerUsage | m_uStaticCoronaUsage;}
	unsigned int GetDecalManagerUsage() {return m_uDecalManagerUsage;}
	unsigned int GetStaticCoronaUsage() {return m_uStaticCoronaUsage;}

	void BruteForce();

	static bool IsUsedBadly(unsigned int uUsageMask);
	bool IsUsedBadly();
};

bool Tools_TextureUsageEntry::IsUsedBadly(unsigned int uUsageMask)
{
	const Tools_MaterialResponse_System::TextureUsage uUsageNormal = Tools_MaterialResponse_System::TU_NORMAL_CHANNELS;
	const Tools_MaterialResponse_System::TextureUsage uUsageElse = 	(Tools_MaterialResponse_System::TextureUsage)(Tools_MaterialResponse_System::TU_SINGLE_CHANNEL|Tools_MaterialResponse_System::TU_COLOUR_CHANNELS);

	return (uUsageNormal & uUsageMask) && (uUsageElse & uUsageMask);
}

bool Tools_TextureUsageEntry::IsUsedBadly()
{
	return IsUsedBadly(GetUsage());
}


void Tools_TextureUsageEntry::Init(Asura_Hash_ID uHashID, const char* pxTextureName)
{
	m_uHashID = uHashID;
	m_pTextureName = pxTextureName;
	m_uUsage = 0;
	m_uDecalManagerUsage = 0;
#ifndef ASURAIMPORT_EXPORTS
	m_uStaticCoronaUsage = Asura_StaticCorona_System::GetTextureUsage( uHashID );
#else
	m_uStaticCoronaUsage = 0;

#endif
}

void Tools_TextureUsageEntry::SetUsedByMaterial(Tools_MaterialResponse* pxMaterial, Asura_MaterialResponse::TEXTURE_SLOT_ID uUsage)
{
	m_axMaterials[uUsage].AddToList(pxMaterial);
	m_uUsage |= (1<<uUsage);
}

void Tools_TextureUsageEntry::SetUsedByDecalMaterial(u_int uUsageMask)
{
	m_uDecalManagerUsage |= uUsageMask;
}
void Tools_TextureUsageEntry::SetUsedByStaticCoronaMaterial(u_int uUsageMask)
{
	m_uStaticCoronaUsage |= uUsageMask;
}


void Tools_TextureUsageEntry::BruteForce()
{
#ifdef ASURA_USE_DECAL_MANAGER
	this->SetUsedByDecalMaterial(Asura_Decal_Manager::GetTextureUsage(m_uHashID));
#endif
	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex(u);

		for (int i = 0; i < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++i)
		{
			if (pxCurrMaterial->GetTextureHash(i) == m_uHashID)
			{
				this->SetUsedByMaterial(pxCurrMaterial, static_cast<Asura_MaterialResponse::TEXTURE_SLOT_ID>(i));
			}
		}
	}
}


Tools_TextureUsageEntry *Tools_TextureUsageEntry::CreateEntry(Asura_TemplateHashTree<Tools_TextureUsageEntry> *pxAllUsage, Asura_Hash_ID uHashID, const char* pxTextureName)
{
	if (uHashID != ASURA_HASH_ID_UNSET)
	{
		if (pxAllUsage)
		{
			bool bDidExist;
			Tools_TextureUsageEntry *pxEntry = pxAllUsage->AddOrFindNode(uHashID, &bDidExist);
			if (!bDidExist)
			{
				pxEntry->Init(uHashID, pxTextureName);
			}
			return pxEntry;
		}
		Tools_TextureUsageEntry *pxEntry = new Tools_TextureUsageEntry();
		pxEntry->Init(uHashID, pxTextureName);

		return pxEntry;
	}
	return NULL;

}

void Tools_TextureUsageEntry::GenerateUsageTree(Asura_TemplateHashTree<Tools_TextureUsageEntry> &xAllUsage)
{
	// Loop through each material getting every texture
	for( u_int u = 0, uNumMats = Tools_MaterialResponse_System::GetNumMaterials(); u < uNumMats; ++u )
	{
		Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex(u);

		for (int i = 0; i < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++i)
		{
			Tools_TextureUsageEntry *pEntry = Tools_TextureUsageEntry::CreateEntry(&xAllUsage, pxCurrMaterial->GetTextureHash(i), pxCurrMaterial->GetTextureFilename(i));
			if (pEntry)
			{
				pEntry->SetUsedByMaterial(pxCurrMaterial, static_cast<Asura_MaterialResponse::TEXTURE_SLOT_ID>(i));
			}
		}
	}
#ifdef ASURA_USE_DECAL_MANAGER
#if !defined(ASURA_TARGET_SM3) // SM3 uses materialresponse for materials so is already handled
	for (int i=0; i<Asura_Decal_Manager::GetNumDecalMaterials(); i++)
	{
		const Asura_DecalManager_Material *pxMtl = Asura_Decal_Manager::GetDecalMaterialByIndex(i);
		Tools_TextureUsageEntry *pEntry;

		pEntry = Tools_TextureUsageEntry::CreateEntry(&xAllUsage, pxMtl->m_xTextureHash, pxMtl->m_xTextureName.GetString()	);
		if (pEntry) 
		{
			pEntry->SetUsedByDecalMaterial(Tools_MaterialResponse_System::TU_ALBEDO);
		}

	}
#endif
#endif

}


class Tools_TextureUsageReport_Overlay : public Asura_Overlay
{
	static Asura_Vector_2 s_xPosition;
	static const char* Tools_TextureUsageReport_Overlay::s_pcMapNames[Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS];

	Asura_Subwindow_Dialog* m_pxSubWindowDialog;

	Tools_TextureUsageReport_Overlay(const bool bAlwaysOnTop = false);

	Asura_Vector_2 m_xNextChildOffset;


	static void SelectMaterialCB(int iID);





public:
	void GenerateSingleReport(Tools_TextureUsageEntry &xEntry);
	void GenerateFullReport();

	static bool CheckForBadTextureUsage(u_int uUsage);

	static Tools_TextureUsageReport_Overlay* Create()
	{
		Asura_Overlay_System::CloseOverlays(TOOLS_OVERLAY_MATERIALRESPONSE_TEXTUREUSAGEREPORT);

		Tools_TextureUsageReport_Overlay* pxOverlay = new Tools_TextureUsageReport_Overlay();

		return pxOverlay;
	}

};



Asura_Vector_2 Tools_TextureUsageReport_Overlay::s_xPosition;

Tools_TextureUsageReport_Overlay::Tools_TextureUsageReport_Overlay(const bool bAlwaysOnTop) :
Asura_Overlay( "Bad Texture Usage Report", bAlwaysOnTop, &s_xPosition)
{
	m_uID = TOOLS_OVERLAY_MATERIALRESPONSE_TEXTUREUSAGEREPORT;
	Asura_Vector_2 xOffset( 1.0f, 3.0f );
	m_xNextChildOffset = Asura_Vector_2(1.0f, 1.0f);

	m_pxSubWindowDialog = new Asura_Subwindow_Dialog(xOffset, 50, 20, false);
	AddDialog(m_pxSubWindowDialog);
}
void Tools_TextureUsageReport_Overlay::SelectMaterialCB(int iID)
{
	Tools_Texturing_MaterialResponse::EditMaterial(static_cast<Asura_Hash_ID>(iID));
}



const char* Tools_TextureUsageReport_Overlay::s_pcMapNames[Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS] = {
	"albedo texture",
	"normal map",
	"specular map",
	"transparency map",
	"self illumination map",
	"environment map",
	"reflectivity mask",
	"parallax map"
};



static void HandleWordWrap(const char* pMsg, float fWidth, Asura_Subwindow_Dialog* pxOverlay, Asura_Vector_2& xOffset, void (*pCBClick)(int), int iClickContext)
{
	if (Asura_Text_Output::WidthOfString(pMsg, ASURA_TEXT_FORMAT_SMALLFONT) < (fWidth - xOffset.x))
	{
		if (pCBClick)
		{
			Asura_ClickableString_Dialog *pxDialog = new Asura_ClickableString_Dialog(pMsg, xOffset);
			pxDialog->SetClickFunction(pCBClick, iClickContext);
			pxOverlay->AddDialog(pxDialog);
		} else {
			Asura_StringDisplay_Dialog *pxDialog = new Asura_StringDisplay_Dialog(pMsg, xOffset);
			pxOverlay->AddDialog(pxDialog);
		}
		xOffset.y +=  Asura_Function_Dialog::DefaultSpacing();
		return;
	}
	unsigned int uSize = strlen(pMsg)+1;
	Asura_Char *pBuf = new Asura_Char[uSize];
	memcpy(pBuf, pMsg, uSize);
	Asura_String xRet;

	Asura_Char* pSpace;
	Asura_Char* pPrevSpace = pBuf-1;
	Asura_Char* pStart = pBuf;
	while ((pSpace = strchr(pPrevSpace+1, ' ')) != NULL)
	{
		*pSpace = 0;
		if (Asura_Text_Output::WidthOfString(pStart, ASURA_TEXT_FORMAT_SMALLFONT) > (fWidth - xOffset.x))
		{
			if (pPrevSpace <= pStart)
			{
				if (pCBClick)
				{
					Asura_ClickableString_Dialog *pxDialog = new Asura_ClickableString_Dialog(pStart, xOffset);
					pxDialog->SetClickFunction(pCBClick, iClickContext);
					pxOverlay->AddDialog(pxDialog);
				} else {
					Asura_StringDisplay_Dialog *pxDialog = new Asura_StringDisplay_Dialog(pStart, xOffset);
					pxOverlay->AddDialog(pxDialog);
				}
				xOffset.y +=  Asura_Function_Dialog::DefaultSpacing();
				pPrevSpace = pStart = pSpace + 1;
			} else {
				*pPrevSpace = 0;
				if (pCBClick)
				{
					Asura_ClickableString_Dialog *pxDialog = new Asura_ClickableString_Dialog(pStart, xOffset);
					pxDialog->SetClickFunction(pCBClick, iClickContext);
					pxOverlay->AddDialog(pxDialog);
				} else {
					Asura_StringDisplay_Dialog *pxDialog = new Asura_StringDisplay_Dialog(pStart, xOffset);
					pxOverlay->AddDialog(pxDialog);
				}
				xOffset.y +=  Asura_Function_Dialog::DefaultSpacing();
				pStart = pPrevSpace + 1;
				pPrevSpace = pStart;
				*pSpace = ' ';
			}
		} else {
			pPrevSpace = pSpace;
			*pSpace = ' ';
		}
	}
	if (pCBClick)
	{
		Asura_ClickableString_Dialog *pxDialog = new Asura_ClickableString_Dialog(pStart, xOffset);
		pxDialog->SetClickFunction(pCBClick, iClickContext);
		pxOverlay->AddDialog(pxDialog);
	} else {
		Asura_StringDisplay_Dialog *pxDialog = new Asura_StringDisplay_Dialog(pStart, xOffset);
		pxOverlay->AddDialog(pxDialog);
	}
	xOffset.y +=  Asura_Function_Dialog::DefaultSpacing();
	delete[] pBuf;
}

void Tools_TextureUsageReport_Overlay::GenerateSingleReport(Tools_TextureUsageEntry& xEntry)
{
	Asura_String xFileName = xEntry.GetTextureName();
	xFileName += " is used by :-";

	HandleWordWrap(xFileName.GetString(), 45.0f, m_pxSubWindowDialog, m_xNextChildOffset, NULL, 0);
	m_xNextChildOffset.x += Asura_Function_Dialog::DefaultSpacing() * 2;

	for (Asura_MaterialResponse::TEXTURE_SLOT_ID i = Asura_MaterialResponse::ALBEDO_TEXTURE; 
		i < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; 
		i = Asura_MaterialResponse::TEXTURE_SLOT_ID(i+1))
	{
		for (Asura_List_ConstIt<Tools_MaterialResponse> xMatEntries(&xEntry.GetMaterialList(i)); !xMatEntries.Done();xMatEntries.Next())
		{
			Asura_String xMaterialName = xMatEntries.GetCurrent()->GetName();
			xMaterialName += " as a ";
			xMaterialName += s_pcMapNames[i];

			m_xNextChildOffset.x -= Asura_Function_Dialog::DefaultSpacing();
			HandleWordWrap("*", 45.0f, m_pxSubWindowDialog, m_xNextChildOffset, NULL, 0);
			m_xNextChildOffset.y -= Asura_Function_Dialog::DefaultSpacing();
			m_xNextChildOffset.x += Asura_Function_Dialog::DefaultSpacing();
			HandleWordWrap(xMaterialName.GetString(), 45.0f, m_pxSubWindowDialog, m_xNextChildOffset, SelectMaterialCB, (int)xMatEntries.GetCurrent()->GetHash());
		}
#ifdef ASURA_USE_DECAL_MANAGER
		if (xEntry.GetDecalManagerUsage() & (1<<i))
		{
			Asura_String xMaterialName = "Asura_Decal_Manager as a ";
			xMaterialName += s_pcMapNames[i];
			m_xNextChildOffset.x -= Asura_Function_Dialog::DefaultSpacing();
			HandleWordWrap("*", 45.0f, m_pxSubWindowDialog, m_xNextChildOffset, NULL, 0);
			m_xNextChildOffset.y -= Asura_Function_Dialog::DefaultSpacing();
			m_xNextChildOffset.x += Asura_Function_Dialog::DefaultSpacing();
			HandleWordWrap(xMaterialName.GetString(), 45, m_pxSubWindowDialog, m_xNextChildOffset, NULL, 0);
		}
#endif
#ifndef ASURAIMPORT_EXPORTS
		if (xEntry.GetStaticCoronaUsage() & (1<<i))
		{
			Asura_String xMaterialName = "Asura_StaticCorona_System as a ";
			xMaterialName += s_pcMapNames[i];
			m_xNextChildOffset.x -= Asura_Function_Dialog::DefaultSpacing();
			HandleWordWrap("*", 45.0f, m_pxSubWindowDialog, m_xNextChildOffset, NULL, 0);
			m_xNextChildOffset.y -= Asura_Function_Dialog::DefaultSpacing();
			m_xNextChildOffset.x += Asura_Function_Dialog::DefaultSpacing();
			HandleWordWrap(xMaterialName.GetString(), 45, m_pxSubWindowDialog, m_xNextChildOffset, NULL, 0);
		}
#endif // ASURAIMPORT_EXPORTS
	}
	m_pxSubWindowDialog->UpdateSize();
	m_xNextChildOffset.x -= Asura_Function_Dialog::DefaultSpacing() * 2.0f;
}


void Tools_TextureUsageReport_Overlay::GenerateFullReport()
{
	Asura_TemplateHashTree<Tools_TextureUsageEntry> xAllUsage;

	Tools_TextureUsageEntry::GenerateUsageTree(xAllUsage);

	for (Asura_TemplateHashTree_Iterator<Tools_TextureUsageEntry> xEntryIt(xAllUsage); !xEntryIt.Done();xEntryIt.Next())
	{
		Tools_TextureUsageEntry* pxEntry = xEntryIt.GetCurrent();

		if (pxEntry->IsUsedBadly())
		{
			GenerateSingleReport(*pxEntry);
		}
	}
}

void Tools_Texturing_MaterialResponse::ShowBadTextureUsageReport(void)
{
	Tools_TextureUsageReport_Overlay* pxOverlay = Tools_TextureUsageReport_Overlay::Create();
	pxOverlay->GenerateFullReport();
	Asura_Overlay_System::SubmitNewOverlay(pxOverlay, true, true, true);


}


void Tools_Texturing_MaterialResponse::Update()
{
	ASURA_CPU_PROFILE__START_LABEL(ASURA_CPU_PROFILE__PROJECT_PALE_BLACK, "Tools_Texturing_MaterialResponse::Update");

	//Warn if the material conversion was done during the loading process.
	//RAR 19/06/2007
	if(s_bMaterialConversion_ShowMessage)
	{
		ShowMessageOverlay("%d materials have been converted to allow for two materials per polygon.\n"
							"%d new materials were created in the process for use as secondary materials.\n"
							"Make sure you save these changes and check in the new materials",s_iMaterialConversion_NumChanged,s_iMaterialConversion_NumCreated);
		s_bMaterialConversion_ShowMessage = false;
	}


	bool bUpdateTexture = false;

	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex_WithoutLoading(u);
		if ( ! pxCurrMaterial->IsLoaded() ) continue;

		if( pxCurrMaterial->m_bModified )
		{
			pxCurrMaterial->m_bModified = false;

			Tools_TextureUsageReport_Overlay* pxOverlay = NULL;

			for (int i = 0; i < Asura_MaterialResponse::NUMBER_OF_TEXTURE_SLOTS; ++i)
			{
				if (pxCurrMaterial->m_auTextureHashIDs[i] != ASURA_HASH_ID_UNSET && 
					Tools_TextureUsageEntry::IsUsedBadly(Tools_MaterialResponse_System::GetTextureUsage(pxCurrMaterial->m_auTextureHashIDs[i])))
				{
					Tools_TextureUsageEntry xEntry = Tools_TextureUsageEntry(pxCurrMaterial->m_auTextureHashIDs[i], pxCurrMaterial->m_axTextureFilenames[i].GetString());
					xEntry.BruteForce();

					if ( xEntry.IsUsedBadly())
					{
						if (!pxOverlay)
						{
							pxOverlay = Tools_TextureUsageReport_Overlay::Create();;
						}
						pxOverlay->GenerateSingleReport(xEntry);
					}
				}

			}
			if (pxOverlay)
			{
				Asura_Overlay_System::SubmitNewOverlay(pxOverlay, true, true, true);
			}

			pxCurrMaterial->SetNeedsSave(true);
			bUpdateTexture = true;
		}
	}

	if( bUpdateTexture )
	{
		Asura_PC_MaterialResponse_UpdateTexture();
	}


	ASURA_CPU_PROFILE__END(ASURA_CPU_PROFILE__PROJECT_PALE_BLACK);
}

void Tools_Texturing_MaterialResponse::TransferModeData( const int iMode )
{
	Tools_ModeDataFile xFile( "MaterialResponse.dat", iMode, iDAT_FILE_VERSION );

	// N.B. Don't change the ordering here
	if( false == xFile.GetError())
	{
		xFile.Transfer( s_xOverlayPosition );
		xFile.Transfer( Tools_MaterialProperties_Overlay::s_axOverlayPosition[0] );
		xFile.Transfer( Tools_MaterialProperties_Overlay::s_axOverlayPosition[1] );
		xFile.Transfer( Tools_MaterialProperties_Overlay::s_axOverlayPosition[2] );
		xFile.Transfer( Tools_MaterialProperties_Overlay::s_axOverlayPosition[3] );
		xFile.Transfer( Tools_MaterialProperties_Overlay::s_iMaxOpenOverlay );
		xFile.Transfer( s_xUserSetTreeSize );
	}

	Tools_MaterialProperties_Overlay::s_iMaxOpenOverlay = Asura_Maths::ClampToRange(Tools_MaterialProperties_Overlay::s_iMaxOpenOverlay, 1, iTOOLS_MAX_EDIT_MATERIAL_RESPONSE_OVERLAY);
}

bool Tools_Texturing_MaterialResponse::ShowNewFileDialogue( Asura_String& xFilename )
{
	const char* szChosenFilename = NULL;

	static Asura_String s_xLastUsedPath;

	// Prompt the user with the standard Windows save dialogue, so they can create folders as required
	{
		Tools_FileSystem_DialogInfo xSaveDialogInfo;

		Asura_String xRelFilename, xRelPath;
		if(!xFilename.IsBlank())
		{
			//use the path from the filename passed in
			xRelFilename = szMATERIALS_DIRECTORY;
			xRelFilename += xFilename.GetString();
			xRelFilename.SplitFilename(&xRelPath, 0, 0);

			Tools_FileSystem::EnsureDirectoryExists(xRelPath.GetString());

			xSaveDialogInfo.DefaultPath = xRelPath.GetString();
			xSaveDialogInfo.DefaultFilename = Tools_FileSystem::StripPath(xRelFilename.GetString());
		}
		else
		{
			//use the most recently selected path
			xSaveDialogInfo.DefaultPath = s_xLastUsedPath.IsBlank() ? szMATERIALS_DIRECTORY : s_xLastUsedPath.GetString();
			xSaveDialogInfo.DefaultFilename = "";
		}
		xSaveDialogInfo.FileFilter = "Asura Materials\0*" szMATERIALS_EXTENSION "\0\0";
		xSaveDialogInfo.DefaultExtension = szMATERIALS_EXTENSION + 1; // Strip off the .
		xSaveDialogInfo.WindowTitle = "Enter New Material Name";

		szChosenFilename = Tools_FileSystem::GetSaveFilename( xSaveDialogInfo );
		if( !szChosenFilename ) return false;
	}

	// Check for correct extension
	if( !Asura_StringUtil::EndsWith( szChosenFilename, szMATERIALS_EXTENSION, true ) )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "The filename must have a " szMATERIALS_EXTENSION " extension." );
		return false;
	}

	// Strip off Asura_Files\Materials from the start of the path
	{
		szChosenFilename = Asura_FileAccess::PlatformRemoveNetworkOrExecutableDir( szChosenFilename );

		if( !Asura_StringUtil::StartsWith( szChosenFilename, szMATERIALS_DIRECTORY, true ) )
		{
			Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "The filename is not within the " szMATERIALS_DIRECTORY " folder." );
			return false;
		}

		//store the most recently used path , since it appears to be valid
		Asura_String(szChosenFilename).SplitFilename(&s_xLastUsedPath, 0, 0);

		szChosenFilename += strlen( szMATERIALS_DIRECTORY );
	}

	// Remove extension from the path
	Asura_String xMaterialName;
	xFilename.MakeNull();
	Asura_String( szChosenFilename ).SplitFilename( &xFilename, &xMaterialName, NULL );

	if( !xFilename.IsBlank() )
	{
		xFilename += "\\";
	}

	xFilename += xMaterialName;

	// Ensure there are no existing materials with this name
	if( Tools_MaterialResponse_System::GetMaterialByHashID( xFilename.GetHash() ) )
	{
		Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_ERROR, "A material with that name already exists." );
		return false;
	}

	return true;
}

void Tools_Texturing_MaterialResponse::CopyPropertiesFromOther( int iMaterialHash )
{
	Tools_CopyMaterialProperties_Overlay* const pxOverlay = new Tools_CopyMaterialProperties_Overlay( 0, iMaterialHash );
	pxOverlay->ShowOverlay();

	delete pxOverlay;
}

void Tools_Texturing_MaterialResponse::CopyPropertiesToOther( int iMaterialHash )
{
	Tools_CopyMaterialProperties_Overlay* const pxOverlay = new Tools_CopyMaterialProperties_Overlay( iMaterialHash, 0 );
	pxOverlay->ShowOverlay();

	delete pxOverlay;
}

void Tools_Texturing_MaterialResponse::MoveMaterial( int iMaterialHash )
{
	// Parameter check

	Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( iMaterialHash );
	Asura_Assert( pxMaterial, "Selected material does not exist." );
	if( !pxMaterial ) return;

	const Asura_Hash_ID uOldHash = pxMaterial->GetHash();

	Asura_String xNewName = pxMaterial->GetName();
	if( !ShowNewFileDialogue( xNewName ) ) return;

	const Asura_Hash_ID uNewHash = xNewName.GetHash();
	if( uNewHash == uOldHash ) return;

	//store the old name in the list of known material names, so it can be used in error messages should anything
	//later be found to refer to it.
	//RAR 21/09/2007
	Tools_MaterialResponse_System::AddToMaterialNameTree(pxMaterial->m_xName.GetString());

	pxMaterial->m_xName = xNewName;
	pxMaterial->m_bModified = true;

	Tools_MaterialResponse_System::ReplaceMaterialHash( uOldHash, uNewHash );

	// Material name pointer has changed, so update property sheets & material list if open
	{
		Tools_MaterialProperties_Overlay* pxOverlay = Tools_MaterialProperties_Overlay::FindOverlayForMaterial(uOldHash);
		if( pxOverlay )
		{
			//Need to force the overlay to be regenerated from scratch , since many of the controls will have the old material
			//hash embedded in them.
			Tools_MaterialProperties_Overlay::EditMaterial(uNewHash,pxOverlay->GetOverlayIndex());
		}
		//Also just force an update of all material properties overlays , since other overlays may have references to this
		Tools_MaterialProperties_Overlay::UpdateAllOverlayDialogs();
		
		RegenerateOverlay();
	}

	// Fix up any polygons this is applied to
	{
		const u_int uNumModules = Asura_Module_System::GetNumberOfModules();

		for( u_int u = 0; u < uNumModules; ++u )
		{
			Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule(u);
			Asura_Assert( pxMesh && pxMesh->TextureDataValid, "Expected module texture data to be valid." );
			if( !pxMesh || !pxMesh->TextureDataValid ) continue;

			bool bNeedToUpdateMesh = false;

			const u_int uNumPolygons = pxMesh->NumberOfPolygons;
			for( u_int v = 0; v < uNumPolygons; ++v )
			{
				if( pxMesh->TextureDataPtr[v].m_auMaterialResponseHashID[0] == uOldHash )
				{
					pxMesh->TextureDataPtr[v].m_auMaterialResponseHashID[0] = uNewHash;
					bNeedToUpdateMesh = true;
				}
				if( pxMesh->TextureDataPtr[v].m_auMaterialResponseHashID[1] == uOldHash )
				{
					pxMesh->TextureDataPtr[v].m_auMaterialResponseHashID[1] = uNewHash;
					bNeedToUpdateMesh = true;
				}
			}

			if( bNeedToUpdateMesh )
			{
				Tools_UpdateTexturingOnSurfaces( pxMesh );
				SetMeshModified( pxMesh, true );
			}
		}
	}

	// Fix up any slaves to point to the new master
	{
		for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
		{
			Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex(u);

			if( pxCurrMaterial->GetParentHash() == uOldHash )
			{
				pxCurrMaterial->m_uParentHash = uNewHash;
				pxCurrMaterial->m_bModified = true;
			}
		}
	}

	// We can't delete the old file from source control here, so notify the user to do it.
	// The new one will be created in the right place when they next save

	Tools_Timed_Message_System::AddMessage( Tools_Timed_Message_System::COLOUR_INFO,
											"Material moved to %s. Remember to delete the old file from source control.",
											xNewName.GetString() );
}

void Tools_Texturing_MaterialResponse::SetMeshModified( const Asura_Mesh* pxMesh, const bool bModified )
{
	s_bMeshModified = bModified;

	if( bModified )
	{
		if(pxMesh)
			Tools_SaveAsura::SetMeshModified(pxMesh);
		else
			Tools_SaveAsura::SetMainFileModified();
	}
}

void Tools_Texturing_MaterialResponse::AddSlaves( int iMaterialHash )
{
	// To avoid sign warnings
	const Asura_Hash_ID uMaterialHash = iMaterialHash;

	Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash );
	Asura_Assert( pxMaterial, "Selected material not found in the list." );
	if( !pxMaterial ) return;

	Asura_String xTitle( "Add slaves to " );
	xTitle += pxMaterial->GetName();

	Tools_AddMaterialSlaves_Overlay xOverlay( xTitle );

	for( u_int u = 0; u < Tools_MaterialResponse_System::GetNumMaterials(); ++u )
	{
		const Tools_MaterialResponse* const pxCurrMaterial = Tools_MaterialResponse_System::GetMaterialByIndex(u);
		const bool bAlreadyEnslaved = (pxCurrMaterial->GetParentHash() == uMaterialHash);

		// Don't allow us to enslave ourselves, our existing slaves, or other master materials
		if( pxCurrMaterial != pxMaterial && !bAlreadyEnslaved && !pxCurrMaterial->HasSlaves() )
		{
			xOverlay.AddItem( pxCurrMaterial->GetName().GetString(), pxCurrMaterial->GetHash() );
		}
	}

	if( !xOverlay.DoModal() ) return;

	for( int i = 0; i < xOverlay.GetNumItems(); ++i )
	{
		if( xOverlay.GetItemTicked(i) )
		{
			const Asura_Hash_ID uSlaveHash = xOverlay.GetItemData(i);

			Tools_MaterialResponse* const pxSlave = Tools_MaterialResponse_System::GetMaterialByHashID( uSlaveHash );
			if( pxSlave )
			{
				pxSlave->m_uParentHash = uMaterialHash;
				pxSlave->m_bModified = true;
			}
			else
			{
				Asura_Assertf( false, ("Selected material with hash ID 0x%X not found.", uSlaveHash) );
			}
		}
	}

	Tools_MaterialProperties_Overlay::UpdateAllOverlayDialogs();
}

void Tools_Texturing_MaterialResponse::RemoveSlave( int iMaterialHash )
{
	const Tools_MaterialProperties_Overlay* const pxOverlay = Tools_MaterialProperties_Overlay::FindOverlayForMaterial(iMaterialHash);
	if( !pxOverlay ) return;

	const Asura_Hash_ID uSlaveHash = Asura_GetHashID( pxOverlay->GetSelectedSlave() );

	BreakFromParent( uSlaveHash );
}

void Tools_Texturing_MaterialResponse::EnslaveToOther( int iMaterialHash )
{
	Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( iMaterialHash );
	Asura_Assert( pxMaterial, "Selected material not found in the list." );
	if( !pxMaterial ) return;

	{
		Asura_String xTitle( "Select parent for " );
		xTitle += pxMaterial->GetName();

		Tools_EnslaveMaterialToOther_Overlay xOverlay( xTitle, *pxMaterial );

		xOverlay.ShowOverlay();
	}

	Tools_MaterialProperties_Overlay::UpdateAllOverlayDialogs();
}

void Tools_Texturing_MaterialResponse::BreakFromParent( int iMaterialHash )
{
	Tools_MaterialResponse* const pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID( iMaterialHash );
	Asura_Assert( pxMaterial, "Selected material not found in the list." );
	if( !pxMaterial ) return;

	pxMaterial->m_uParentHash = ASURA_HASH_ID_UNSET;
	pxMaterial->m_bModified = true;

	Tools_MaterialProperties_Overlay::UpdateAllOverlayDialogs();
}

void Tools_Texturing_MaterialResponse::RemoveMaterialsFromMesh()
{
	const bool bResetInvalidOnly = Tools_MaterialResponse_Overlay::GetResetSelectedMaterialsOnly();

	bool bOkay = ShowYesNoOverlay("Remove materials from mesh :\nThis will remove all %smaterial response assignments for meshes used by this file.\nAre you sure you want to do this?",bResetInvalidOnly ? "invalid " : "");
	if(!bOkay) return;


	for( int i = 0; i < Asura_Module_System::GetNumberOfModules(); ++i )
	{
		Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule(i);
		if( !pxMesh->TextureDataValid )
		{
			Asura_Assertf( false, ("Mesh '%s' has invalid texture data.", pxMesh->Name.GetString()) );
			continue;
		}

		for( int j = 0; j < pxMesh->NumberOfPolygons; ++j )
		{
			bool bLoggedUndo = false;
			
			for(int iMatSlot=0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				if(!pxMesh->TextureDataPtr[j].m_auMaterialResponseHashID[iMatSlot]) continue;

				if( bResetInvalidOnly )
				{
					const Asura_Hash_ID uMaterialHash = pxMesh->TextureDataPtr[j].m_auMaterialResponseHashID[iMatSlot];
					if( Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash ) ) continue;
				}

				if(!bLoggedUndo)
				{
					bLoggedUndo = true;
					LogRemoveMaterial( pxMesh->Index, j );
				}

				pxMesh->TextureDataPtr[j].m_auMaterialResponseHashID[iMatSlot] = ASURA_HASH_ID_UNSET;
			}
		}

		Tools_UpdateTexturingOnSurfaces( pxMesh );
	}

	SetMeshModified( NULL, true );

	Tools_Undo_System::MarkUndoStart();
}

void Tools_Texturing_MaterialResponse::ShowStatistics()
{
	Asura_TreeDialog_Tree xInfoTree;

	// Add the main loaded meshes at the top
	{
		Asura_TreeDialog_Item* pxMainFileNode = xInfoTree.AddItem( Tools_SaveAsura::GetMainFilename() );

		const int iNumModules = Asura_Module_System::GetNumberOfModules();
		const char** pszModuleNames = new const char*[ iNumModules ];
		if( !iNumModules ) return;

		for( int i = 0; i < iNumModules; ++i )
		{
			const Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule(i);
			pszModuleNames[i] = pxMesh->Name.GetString();
		}

		qsort( pszModuleNames, iNumModules, sizeof( const char * ), CompareStrings );

		for( int i = 0; i < iNumModules; ++i )
		{
			const Asura_Mesh* const pxMesh = Asura_Mesh_System::GetMesh( pszModuleNames[i] );
			AddMeshStatsToTree( xInfoTree, *pxMesh, pxMainFileNode );
		}

		delete[] pszModuleNames;
	}

	// Run through the rest of the loaded meshes
	{
		Asura_TreeDialog_Item* pxOtherMeshesNode = xInfoTree.AddItem( "Other loaded meshes" );

		const int iNumMeshes = Asura_Mesh_System::GetNumberOfMeshes() - Asura_Module_System::GetNumberOfModules();
		const char** pszMeshNames = new const char*[ iNumMeshes ];

		int iCurrMeshName = 0;

		for( int i = 0; i < Asura_Mesh_System::GetNumberOfMeshes(); ++i )
		{
			const Asura_Mesh* const pxMesh = Asura_Mesh_System::GetMesh(i);
			const char* const szMeshName = pxMesh->Name.GetString();

			// Skip modules, we've already shown them
			if( !xInfoTree.FindByData( pxMesh->Name.GetHash() ) )
			{
				pszMeshNames[ iCurrMeshName++ ] = szMeshName;

				if( iCurrMeshName > iNumMeshes )
				{
					Asura_Assert( false, "Added too many non-module meshes to list." );
					delete[] pszMeshNames;

					return;
				}
			}
		}

		qsort( pszMeshNames, iNumMeshes, sizeof( const char* ), CompareStrings );

		for( int i = 0; i < iNumMeshes; ++i )
		{
			const Asura_Mesh* const pxMesh = Asura_Mesh_System::GetMesh( pszMeshNames[i] );

			// Skip Tools meshes
			if( !Asura_StringUtil::StartsWith( pszMeshNames[i], "ToolsShape_", true ) )
			{
				AddMeshStatsToTree( xInfoTree, *pxMesh, pxOtherMeshesNode );
			}
		}

		delete[] pszMeshNames;
	}

	// Display the results in a modal overlay
	{
		Asura_Overlay* const pxOverlay = new Asura_Overlay( "Material Response Statistics", false );
		if( !pxOverlay )
		{
			Asura_Assert( false, "Unable to allocate material statistics overlay." );
			return;
		}

		Asura_Vector_2 xOffset( 1.0f, 3.0f );
		Asura_Dialog* pxDialog;

		pxDialog = new Asura_Tree_Dialog( xOffset, &xInfoTree, 40.0f, 40.0f );
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += 40.0f + Asura_Dialog::LargeSpacing();

		pxDialog = new Asura_Function_Dialog( "Close", xOffset, NULL, &pxOverlay->m_bShouldClose );
		pxOverlay->AddDialog( pxDialog );
		xOffset.y += Asura_Function_Dialog::DefaultSpacing();

		Asura_Overlay_System::SubmitNewOverlay( pxOverlay, true, false );

		Tools_Modal_Overlay::DoModal( pxOverlay );
	}
}

void Tools_Texturing_MaterialResponse::AddMeshStatsToTree( Asura_TreeDialog_Tree& xInfoTree, const Asura_Mesh& xMesh, Asura_TreeDialog_Item* pxRootNode /*= NULL*/ )
{
	Asura_HashTree xFoundMaterialsTree;
	u_int uNumUnassignedPolygons = 0;
	u_int uNumInvalidPolygons = 0;

	for( int j = 0; j < xMesh.NumberOfPolygons; ++j )
	{
		const Asura_Hash_ID uMaterialHash = xMesh.TextureDataPtr[j].m_auMaterialResponseHashID[0];

		if( uMaterialHash == ASURA_HASH_ID_UNSET )
		{
			++uNumUnassignedPolygons;
		}
		else if( Tools_MaterialResponse_System::GetMaterialByHashID( uMaterialHash ) )
		{
			int iDummy;
			xFoundMaterialsTree.FindNodeIndexByHash( uMaterialHash, true, iDummy );
		}
		else
		{
			++uNumInvalidPolygons;
		}
	}

	const int iNumValidMaterials = xFoundMaterialsTree.GetNumberOfNodes();

	char szText[256];

	Asura_PrintfToBfr( szText, "%s        (Mtls: %u Invalid: %u Blank: %u)", 256, true, xMesh.Name.GetString(),
					   iNumValidMaterials, uNumInvalidPolygons, uNumUnassignedPolygons );

	Asura_TreeDialog_Item* const pxMeshNode = xInfoTree.AddItem( pxRootNode, szText, xMesh.Name.GetHash() );

	const char* szPlural;

	szPlural = (iNumValidMaterials == 1 ? "" : "s");
	Asura_PrintfToBfr( szText, "%d valid material%s used", 256, true, iNumValidMaterials, szPlural );
	xInfoTree.AddItem( pxMeshNode, szText );

	szPlural = (uNumInvalidPolygons == 1 ? "" : "s");
	Asura_PrintfToBfr( szText, "%u polygon%s with invalid/missing material assigned", 256, true, uNumInvalidPolygons, szPlural );
	xInfoTree.AddItem( pxMeshNode, szText );

	szPlural = (uNumUnassignedPolygons ? "" : "s");
	Asura_PrintfToBfr( szText, "%u polygon%s without a material assigned", 256, true, uNumUnassignedPolygons, szPlural );
	xInfoTree.AddItem( pxMeshNode, szText );
}

void Tools_Texturing_MaterialResponse::UpdatePropertiesOverlay()
{
	Tools_MaterialProperties_Overlay::UpdateAllOverlayDialogs();
}

void Tools_Texturing_MaterialResponse::LogApplyMaterial( const u_int uMeshIndex, const u_int uPolygonIndex )
{
	RootUndoData** const ppxUndoData = Tools_Undo_System::AllocateNextUndoData();
	if( !ppxUndoData ) return;

	*ppxUndoData = new Tools_MaterialResponse_UndoData_ApplyMaterial( uMeshIndex, uPolygonIndex );
}

void Tools_Texturing_MaterialResponse::LogImportMaterial( const u_int uMeshIndex, const u_int uPolygonIndex )
{
	RootUndoData** const ppxUndoData = Tools_Undo_System::AllocateNextUndoData();
	if( !ppxUndoData ) return;

	*ppxUndoData = new Tools_MaterialResponse_UndoData_ImportMaterial( uMeshIndex, uPolygonIndex );
}

void Tools_Texturing_MaterialResponse::LogRemoveMaterial( const u_int uMeshIndex, const u_int uPolygonIndex )
{
	RootUndoData** const ppxUndoData = Tools_Undo_System::AllocateNextUndoData();
	if( !ppxUndoData ) return;

	*ppxUndoData = new Tools_MaterialResponse_UndoData_RemoveMaterial( uMeshIndex, uPolygonIndex );
}

void Tools_Texturing_MaterialResponse::OutputToMaterialLogFile(const char* szFormat,...)
{
	static bool s_bClearedFile = false;
	if(!s_bClearedFile)
	{
		Asura_SimpleLogfileClear("Logfiles\\MaterialResponse.txt");	
		s_bClearedFile = true;
	}

	va_list xArgs;
	va_start(xArgs, szFormat);
	char szBuffer[512];
	Asura_PrintfToBfr(szBuffer,szFormat,xArgs,sizeof(szBuffer),true);
	va_end(xArgs);

	Asura_SimpleLogfileOutput_PreFormatted("Logfiles\\MaterialResponse.txt",szBuffer,strlen(szBuffer));
}



void Tools_Texturing_MaterialResponse::UpdateOldMaterialsFromMaterialResponseAndGetUsedMaterials(const List<Asura_Mesh*>& xMeshList,Asura_HashTree& xOutputMaterialResponseTree)
{

	//This updates all the old-style materials to use appropriate materials based on the current material responses.
	//This is to avoid having references to old textures hanging around.
	//RAR 30/07/2007
	Tools_ConvertMaterialResponseToID xConvert;
	int iJunkIndex;

    for( CLIF<Asura_Mesh*> xIt(&xMeshList); !xIt.done(); xIt.next() )
    {
        //update the render mesh
        Asura_Mesh* pxMesh = xIt();
        for( int iPoly = 0; iPoly < pxMesh->NumberOfPolygons; ++iPoly )
        {
			Asura_Mesh_TextureInfo& xTexInfo = pxMesh->TextureDataPtr[iPoly];

            if(xTexInfo.m_auMaterialResponseHashID[0] || xTexInfo.m_auMaterialResponseHashID[1])
			{
            	xTexInfo.m_iMaterialID = xConvert.GetMaterialID(xTexInfo.m_auMaterialResponseHashID[0],xTexInfo.m_auMaterialResponseHashID[1]);
			}
			else
			{
				xTexInfo.m_iMaterialID = ASURA_MATERIAL_NONE;
			}
			if(xTexInfo.m_auMaterialResponseHashID[0]) xOutputMaterialResponseTree.FindNodeIndexByHash(xTexInfo.m_auMaterialResponseHashID[0],true,iJunkIndex);
			if(xTexInfo.m_auMaterialResponseHashID[1]) xOutputMaterialResponseTree.FindNodeIndexByHash(xTexInfo.m_auMaterialResponseHashID[1],true,iJunkIndex);
        }
		//and the collision mesh
		for(int iColl=0;iColl<2;iColl++)
		{
			Asura_CollisionMesh* pxCollMesh = iColl ? pxMesh->m_pxCollisionMesh : pxMesh->m_pxMergedEntityCollisionMesh;
			if(!pxCollMesh) continue;

			//update the overall material
			if(pxCollMesh->GetOverallMaterialResponse())
			{
				pxCollMesh->SetOverallMaterialID((short)xConvert.GetMaterialID(pxCollMesh->GetOverallMaterialResponse(),0));
			}
			else
			{
				pxCollMesh->SetOverallMaterialID(ASURA_MATERIAL_NONE);
			}


			//and the per-poly IDs if they exist
			if(pxCollMesh->HasPerPolyMaterialID())
			{
        		for( u_int iPoly = 0; iPoly < pxCollMesh->GetNumberOfPolys(); ++iPoly )
        		{
					Asura_Hash_ID uMaterialResponse = pxCollMesh->GetMaterialResponseHashID(iPoly);
					
					if(uMaterialResponse)
					{
						int iMaterialID = xConvert.GetMaterialID(uMaterialResponse,0);
						pxCollMesh->SetMaterialID(iPoly, (short)iMaterialID);

						if(uMaterialResponse) xOutputMaterialResponseTree.FindNodeIndexByHash(uMaterialResponse,true,iJunkIndex);
					}
					else
					{
						pxCollMesh->SetMaterialID(iPoly, ASURA_MATERIAL_NONE);
					}

				}
	       	}
		
		}
	}
}

static bool FilterMaterialByEnvCallback(Asura_Hash_ID uMaterialHash, const void* const pUserData)
{
	//brush the constness under the carpet , since the Asura_HashTree doesn't have any const methods for checking
	//its contents
	//RAR 31/08//2007
	Asura_HashTree* pxTree = (Asura_HashTree*) pUserData;

	int iJunkIndex;
	return pxTree->FindNodeIndexByHash(uMaterialHash, false, iJunkIndex);
}

void Tools_Texturing_MaterialResponse::CreateMaterialNameTree_FilteredByEnv(Asura_TreeDialog_Tree& xTree)
{
	//build a hash tree containing all materials used in the environment
	Asura_HashTree xUsedMaterialTree;
	FindMaterialsUsedByEnv(xUsedMaterialTree);


	//build the material tree , filtering by the ones used on the environment
	Tools_MaterialResponse_System::CreateMaterialNameTree(xTree, FilterMaterialByEnvCallback, &xUsedMaterialTree);

	
}

void Tools_Texturing_MaterialResponse::FindMaterialsUsedByEnv(Asura_HashTree& xOutputMaterialTree)
{
	//build a hash tree containing all materials used in the environment
	for( int i = 0; i < Asura_Module_System::GetNumberOfModules(); ++i )
	{
		Asura_Mesh* const pxMesh = Asura_Module_System::GetMeshForModule(i);
		if( !pxMesh->TextureDataValid )
		{
			Asura_Assertf( false, ("Mesh '%s' has invalid texture data.", pxMesh->Name.GetString()) );
			continue;
		}

		for( int j = 0; j < pxMesh->NumberOfPolygons; ++j )
		{
			for(int iMatSlot=0 ; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				Asura_Hash_ID uMaterial = pxMesh->TextureDataPtr[j].m_auMaterialResponseHashID[iMatSlot];
				
				if(uMaterial)
				{
					int iJunkIndex;
					xOutputMaterialTree.FindNodeIndexByHash(uMaterial, true, iJunkIndex);
				}
			}
		}
	}
}

class Tools_MessageLogAction_SelectMissingMaterials : public Tools_MessageLogAction
{
public:

	Tools_MessageLogAction_SelectMissingMaterials( const Asura_Hash_ID uMaterialHash )
		: m_uMaterialHash( uMaterialHash )
	{
	}

	virtual void Execute()
	{
		// Look for the material in all channels

		u_int uChannels = 0;
		for( u_int u = 0; u < TOOLS_NUM_MATERIALRESPONSE_PER_POLY; ++u )
		{
			uChannels |= (1 << u);
		}

		Tools_PolygonFind_Material xFind( m_uMaterialHash, uChannels );
		Tools_Texturing_Find::FindPolygons( &xFind );
	}

protected:

	const Asura_Hash_ID m_uMaterialHash;
};

void Tools_Texturing_MaterialResponse::FindMissingMaterials()
{
	
	//Clear out existing messages
	Tools_MessageLog::DeleteMessages(Tools_MessageLog::MATERIAL_RESPONSE,0);
	//Get current position to see if any errors were logged
	u_int uInitialPos = Tools_MessageLog::GetNextID();
	
	
	Asura_HashTree xMaterialTree;

	//Search all loaded meshes and check for the materials being used
	for(int i=0; i<Asura_Mesh_System::GetNumberOfMeshes(); i++)
	{
		const Asura_Mesh* pxMesh = Asura_Mesh_System::GetMesh(i);
		if(!pxMesh) continue;
		if(!pxMesh->TextureDataPtr) continue;

		for(int j=0; j<pxMesh->NumberOfPolygons; j++)
		{
			for(int iMatSlot = 0; iMatSlot<TOOLS_NUM_MATERIALRESPONSE_PER_POLY; iMatSlot++)
			{
				Asura_Hash_ID uMaterial = pxMesh->TextureDataPtr[j].m_auMaterialResponseHashID[iMatSlot];
				if(!uMaterial) continue;

				//checked this material yet?
				int iJunkIndex;
				if(xMaterialTree.FindNodeIndexByHash(uMaterial, true, iJunkIndex)) continue;

				//does this material exist?
				const Tools_MaterialResponse* pxMaterial = Tools_MaterialResponse_System::GetMaterialByHashID(uMaterial);
				if(pxMaterial)
				{
					//check for this material existing on the network
					#ifdef ASURA_PC_ALLOW_LOAD_FROM_NETWORK
					const char* szNetworkPath = Asura_FileAccess::PlatformGetNetworkPath();
					if(!szNetworkPath || !*szNetworkPath) continue;

					char szFilename[MAX_PATH];
					Asura_PrintfToBfr(szFilename,"%s\\%s%s%s", sizeof(szFilename), true, szNetworkPath, szMATERIALS_DIRECTORY, pxMaterial->GetName().GetString(), szMATERIALS_EXTENSION);

					if(!Tools_FileSystem::FileExists(szFilename))
					{
						Tools_MessageLog::AddMessage(Tools_MessageLog::MATERIAL_RESPONSE,"",
							new Tools_MessageLogAction_SelectMissingMaterials( uMaterial ),
							"'%s' can't be found on the network (used by %s)",
							pxMaterial->GetName().GetString(), pxMesh->Name.GetString());
					}
					#endif

				}
				else
				{
					const char* szMaterialDesc = Tools_MaterialResponse_System::GetMissingMaterialDescription(uMaterial);

					Tools_MessageLog::AddMessage(Tools_MessageLog::MATERIAL_RESPONSE,"",
						new Tools_MessageLogAction_SelectMissingMaterials( uMaterial ),
						"'%s' is missing (used by %s)", szMaterialDesc, pxMesh->Name.GetString());
				}
			}
		}
	}

	if(Tools_MessageLog::AreThereAnyMessages(Tools_MessageLog::MATERIAL_RESPONSE, 0, uInitialPos))
	{
		//Found some errors ,so show message log
		Tools_MessageLog::ShowOverlay(Tools_MessageLog::MATERIAL_RESPONSE,0,uInitialPos);
	}
	else
	{
		ShowMessageOverlay("No missing materials found.");
	}

}

#endif // defined ASURA_USE_MATERIAL_RESPONSE_SYSTEM
